/*!
 * Copyright (c) 2012 - 2021, Anaconda, Inc., and Bokeh Contributors
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * Neither the name of Anaconda nor the names of any contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
(function(root, factory) {
  const bokeh = factory();
  bokeh.__bokeh__ = true;
  if (typeof root.Bokeh === "undefined" || typeof root.Bokeh.__bokeh__ === "undefined") {
    root.Bokeh = bokeh;
  }
  const Bokeh = root.Bokeh;
  Bokeh[bokeh.version] = bokeh;
})(this, function() {
  let define;
  const parent_require = typeof require === "function" && require
  return (function(modules, entry, aliases, externals) {
  if (aliases === undefined) aliases = {};
  if (externals === undefined) externals = {};

  const cache = {};

  const normalize = function(name) {
    if (typeof name === "number")
      return name;

    if (name === "bokehjs")
      return entry;

    if (!externals[name]) {
      const prefix = "@bokehjs/"
      if (name.slice(0, prefix.length) === prefix)
        name = name.slice(prefix.length)
    }

    const alias = aliases[name]
    if (alias != null)
      return alias;

    const trailing = name.length > 0 && name[name.lenght-1] === "/";
    const index = aliases[name + (trailing ? "" : "/") + "index"];
    if (index != null)
      return index;

    return name;
  }

  const require = function(name) {
    let mod = cache[name];
    if (!mod) {
      const id = normalize(name);

      mod = cache[id];
      if (!mod) {
        if (!modules[id]) {
          if (externals[id] === false || (externals[id] == true && parent_require)) {
            try {
              mod = {exports: externals[id] ? parent_require(id) : {}};
              cache[id] = cache[name] = mod;
              return mod.exports;
            } catch (e) {}
          }

          const err = new Error("Cannot find module '" + name + "'");
          err.code = 'MODULE_NOT_FOUND';
          throw err;
        }

        mod = {exports: {}};
        cache[id] = cache[name] = mod;

        function __esModule() {
          Object.defineProperty(mod.exports, "__esModule", {value: true});
        }

        function __esExport(name, value) {
          Object.defineProperty(mod.exports, name, {
            enumerable: true, get: function () { return value; }
          });
        }

        modules[id].call(mod.exports, require, mod, mod.exports, __esModule, __esExport);
      } else {
        cache[name] = mod;
      }
    }

    return mod.exports;
  }
  require.resolve = function(name) {
    return ""
  }

  const main = require(entry);
  main.require = require;

  if (typeof Proxy !== "undefined") {
    // allow Bokeh.loader["@bokehjs/module/name"] syntax
    main.loader = new Proxy({}, {
      get: function(_obj, module) {
        return require(module);
      }
    });
  }

  main.register_plugin = function(plugin_modules, plugin_entry, plugin_aliases, plugin_externals) {
    if (plugin_aliases === undefined) plugin_aliases = {};
    if (plugin_externals === undefined) plugin_externals = {};

    for (let name in plugin_modules) {
      modules[name] = plugin_modules[name];
    }

    for (let name in plugin_aliases) {
      aliases[name] = plugin_aliases[name];
    }

    for (let name in plugin_externals) {
      externals[name] = plugin_externals[name];
    }

    const plugin = require(plugin_entry);

    for (let name in plugin) {
      main[name] = plugin[name];
    }

    return plugin;
  }

  return main;
})
([
/* main.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const yargs_1 = require(1) /* yargs */;
    const path_1 = require("path");
    const sys_1 = require(20) /* ./sys */;
    const build_1 = require(35) /* ./build */;
    const compile_1 = require(1217) /* ./compile */;
    async function read_stdin() {
        const stdin = process.stdin;
        stdin.setEncoding("utf-8");
        stdin.resume();
        let data = "";
        for await (const chunk of stdin) {
            data += chunk;
        }
        return data;
    }
    function reply(data) {
        process.stdout.write(JSON.stringify(data));
        process.stdout.write("\n");
    }
    async function compile() {
        if (yargs_1.argv.file != null) {
            const input = {
                code: yargs_1.argv.code != null ? yargs_1.argv.code : (0, sys_1.read)(yargs_1.argv.file),
                lang: yargs_1.argv.lang ?? "typescript",
                file: yargs_1.argv.file,
                bokehjs_dir: yargs_1.argv.bokehjsDir ?? "./build", // this is what bokeh.settings defaults to
            };
            return await (0, compile_1.compile_and_resolve_deps)(input);
        }
        else {
            const input = JSON.parse(await read_stdin());
            return await (0, compile_1.compile_and_resolve_deps)(input);
        }
    }
    async function main() {
        const cmd = yargs_1.argv._[0];
        if (cmd == "build") {
            try {
                const base_dir = (0, path_1.resolve)(yargs_1.argv.baseDir);
                const bokehjs_dir = (0, path_1.resolve)(yargs_1.argv.bokehjsDir);
                const rebuild = yargs_1.argv.rebuild;
                const bokeh_version = yargs_1.argv.bokehVersion;
                const result = await (0, build_1.build)(base_dir, bokehjs_dir, { rebuild, bokeh_version });
                process.exit(result ? 0 : 1);
            }
            catch (error) {
                const msg = error instanceof Error && error.stack ? error.stack : `${error}`;
                console.log(msg);
                process.exit(1);
            }
        }
        else if (cmd == "init") {
            try {
                const base_dir = (0, path_1.resolve)(yargs_1.argv.baseDir);
                const bokehjs_dir = (0, path_1.resolve)(yargs_1.argv.bokehjsDir);
                const interactive = yargs_1.argv.interactive;
                const bokehjs_version = yargs_1.argv.bokehjsVersion;
                const bokeh_version = yargs_1.argv.bokehVersion;
                const result = await (0, build_1.init)(base_dir, bokehjs_dir, { interactive, bokehjs_version, bokeh_version });
                process.exit(result ? 0 : 1);
            }
            catch (error) {
                const msg = error instanceof Error && error.stack ? error.stack : `${error}`;
                console.log(msg);
                process.exit(1);
            }
        }
        else {
            try {
                reply(await compile());
            }
            catch (error) {
                const msg = error instanceof Error && error.stack ? error.stack : `${error}`;
                reply({ error: msg });
            }
        }
    }
    main();
},
/* yargs/index.cjs */ function _(require, module, exports, __esModule, __esExport) {
    // classic singleton yargs API, to use yargs
    // without running as a singleton do:
    // require('yargs/yargs')(process.argv.slice(2))
    const { Yargs, processArgv } = require(2) /* ./build/index.cjs */;
    Argv(processArgv.hideBin(process.argv));
    module.exports = Argv;
    function Argv(processArgs, cwd) {
        const argv = Yargs(processArgs, cwd, require);
        singletonify(argv);
        // TODO(bcoe): warn if argv.parse() or argv.argv is used directly.
        return argv;
    }
    /*  Hack an instance of Argv with process.argv into Argv
        so people can do
        require('yargs')(['--beeble=1','-z','zizzle']).argv
        to parse a list of args and
        require('yargs').argv
        to get a parsed version of process.argv.
    */
    function singletonify(inst) {
        [
            ...Object.keys(inst),
            ...Object.getOwnPropertyNames(inst.constructor.prototype),
        ].forEach(key => {
            if (key === 'argv') {
                Argv.__defineGetter__(key, inst.__lookupGetter__(key));
            }
            else if (typeof inst[key] === 'function') {
                Argv[key] = inst[key].bind(inst);
            }
            else {
                Argv.__defineGetter__('$0', () => {
                    return inst.$0;
                });
                Argv.__defineGetter__('parsed', () => {
                    return inst.parsed;
                });
            }
        });
    }
},
/* yargs/build/index.cjs */ function _(require, module, exports, __esModule, __esExport) {
    var t = require("assert");
    class e extends Error {
        constructor(t) { super(t || "yargs error"), this.name = "YError", Error.captureStackTrace(this, e); }
    }
    let s, n = [];
    function i(t, o, a, h) { s = h; let l = {}; if (Object.prototype.hasOwnProperty.call(t, "extends")) {
        if ("string" != typeof t.extends)
            return l;
        const r = /\.json|\..*rc$/.test(t.extends);
        let h = null;
        if (r)
            h = function (t, e) { return s.path.resolve(t, e); }(o, t.extends);
        else
            try {
                h = require.resolve(t.extends);
            }
            catch (e) {
                return t;
            }
        !function (t) { if (n.indexOf(t) > -1)
            throw new e(`Circular extended configurations: '${t}'.`); }(h), n.push(h), l = r ? JSON.parse(s.readFileSync(h, "utf8")) : require(t.extends), delete t.extends, l = i(l, s.path.dirname(h), a, s);
    } return n = [], a ? r(l, t) : Object.assign({}, l, t); }
    function r(t, e) { const s = {}; function n(t) { return t && "object" == typeof t && !Array.isArray(t); } Object.assign(s, t); for (const i of Object.keys(e))
        n(e[i]) && n(s[i]) ? s[i] = r(t[i], e[i]) : s[i] = e[i]; return s; }
    function o(t) { const e = t.replace(/\s{2,}/g, " ").split(/\s+(?![^[]*]|[^<]*>)/), s = /\.*[\][<>]/g, n = e.shift(); if (!n)
        throw new Error(`No command found in: ${t}`); const i = { cmd: n.replace(s, ""), demanded: [], optional: [] }; return e.forEach(((t, n) => { let r = !1; t = t.replace(/\s/g, ""), /\.+[\]>]/.test(t) && n === e.length - 1 && (r = !0), /^\[/.test(t) ? i.optional.push({ cmd: t.replace(s, "").split("|"), variadic: r }) : i.demanded.push({ cmd: t.replace(s, "").split("|"), variadic: r }); })), i; }
    const a = ["first", "second", "third", "fourth", "fifth", "sixth"];
    function h(t, s, n) { try {
        let i = 0;
        const [r, a, h] = "object" == typeof t ? [{ demanded: [], optional: [] }, t, s] : [o(`cmd ${t}`), s, n], f = [].slice.call(a);
        for (; f.length && void 0 === f[f.length - 1];)
            f.pop();
        const d = h || f.length;
        if (d < r.demanded.length)
            throw new e(`Not enough arguments provided. Expected ${r.demanded.length} but received ${f.length}.`);
        const u = r.demanded.length + r.optional.length;
        if (d > u)
            throw new e(`Too many arguments provided. Expected max ${u} but received ${d}.`);
        r.demanded.forEach((t => { const e = l(f.shift()); 0 === t.cmd.filter((t => t === e || "*" === t)).length && c(e, t.cmd, i), i += 1; })), r.optional.forEach((t => { if (0 === f.length)
            return; const e = l(f.shift()); 0 === t.cmd.filter((t => t === e || "*" === t)).length && c(e, t.cmd, i), i += 1; }));
    }
    catch (t) {
        console.warn(t.stack);
    } }
    function l(t) { return Array.isArray(t) ? "array" : null === t ? "null" : typeof t; }
    function c(t, s, n) { throw new e(`Invalid ${a[n] || "manyith"} argument. Expected ${s.join(" or ")} but received ${t}.`); }
    function f(t) { return !!t && !!t.then && "function" == typeof t.then; }
    function d(t, e, s, n) { s.assert.notStrictEqual(t, e, n); }
    function u(t, e) { e.assert.strictEqual(typeof t, "string"); }
    function p(t) { return Object.keys(t); }
    function g(t = {}, e = (() => !0)) { const s = {}; return p(t).forEach((n => { e(n, t[n]) && (s[n] = t[n]); })), s; }
    function m() { return process.versions.electron && !process.defaultApp ? 0 : 1; }
    function y() { return process.argv[m()]; }
    var b = Object.freeze({ __proto__: null, hideBin: function (t) { return t.slice(m() + 1); }, getProcessArgvBin: y });
    function v(t, e, s, n) { if ("a" === s && !n)
        throw new TypeError("Private accessor was defined without a getter"); if ("function" == typeof e ? t !== e || !n : !e.has(t))
        throw new TypeError("Cannot read private member from an object whose class did not declare it"); return "m" === s ? n : "a" === s ? n.call(t) : n ? n.value : e.get(t); }
    function w(t, e, s, n, i) { if ("m" === n)
        throw new TypeError("Private method is not writable"); if ("a" === n && !i)
        throw new TypeError("Private accessor was defined without a setter"); if ("function" == typeof e ? t !== e || !i : !e.has(t))
        throw new TypeError("Cannot write private member to an object whose class did not declare it"); return "a" === n ? i.call(t, s) : i ? i.value = s : e.set(t, s), s; }
    class O {
        constructor(t) { this.globalMiddleware = [], this.frozens = [], this.yargs = t; }
        addMiddleware(t, e, s = !0, n = !1) { if (h("<array|function> [boolean] [boolean] [boolean]", [t, e, s], arguments.length), Array.isArray(t)) {
            for (let n = 0; n < t.length; n++) {
                if ("function" != typeof t[n])
                    throw Error("middleware must be a function");
                const i = t[n];
                i.applyBeforeValidation = e, i.global = s;
            }
            Array.prototype.push.apply(this.globalMiddleware, t);
        }
        else if ("function" == typeof t) {
            const i = t;
            i.applyBeforeValidation = e, i.global = s, i.mutates = n, this.globalMiddleware.push(t);
        } return this.yargs; }
        addCoerceMiddleware(t, e) { const s = this.yargs.getAliases(); return this.globalMiddleware = this.globalMiddleware.filter((t => { const n = [...s[e] || [], e]; return !t.option || !n.includes(t.option); })), t.option = e, this.addMiddleware(t, !0, !0, !0); }
        getMiddleware() { return this.globalMiddleware; }
        freeze() { this.frozens.push([...this.globalMiddleware]); }
        unfreeze() { const t = this.frozens.pop(); void 0 !== t && (this.globalMiddleware = t); }
        reset() { this.globalMiddleware = this.globalMiddleware.filter((t => t.global)); }
    }
    function C(t, e, s, n) { return s.reduce(((t, s) => { if (s.applyBeforeValidation !== n)
        return t; if (s.mutates) {
        if (s.applied)
            return t;
        s.applied = !0;
    } if (f(t))
        return t.then((t => Promise.all([t, s(t, e)]))).then((([t, e]) => Object.assign(t, e))); {
        const n = s(t, e);
        return f(n) ? n.then((e => Object.assign(t, e))) : Object.assign(t, n);
    } }), t); }
    function j(t, e, s = (t => { throw t; })) { try {
        const s = "function" == typeof t ? t() : t;
        return f(s) ? s.then((t => e(t))) : e(s);
    }
    catch (t) {
        return s(t);
    } }
    const _ = /(^\*)|(^\$0)/;
    class M {
        constructor(t, e, s, n) { this.requireCache = new Set, this.handlers = {}, this.aliasMap = {}, this.frozens = [], this.shim = n, this.usage = t, this.globalMiddleware = s, this.validation = e; }
        addDirectory(t, e, s, n) { "boolean" != typeof (n = n || {}).recurse && (n.recurse = !1), Array.isArray(n.extensions) || (n.extensions = ["js"]); const i = "function" == typeof n.visit ? n.visit : t => t; n.visit = (t, e, s) => { const n = i(t, e, s); if (n) {
            if (this.requireCache.has(e))
                return n;
            this.requireCache.add(e), this.addHandler(n);
        } return n; }, this.shim.requireDirectory({ require: e, filename: s }, t, n); }
        addHandler(t, e, s, n, i, r) { let a = []; const h = function (t) { return t ? t.map((t => (t.applyBeforeValidation = !1, t))) : []; }(i); if (n = n || (() => { }), Array.isArray(t))
            if (function (t) { return t.every((t => "string" == typeof t)); }(t))
                [t, ...a] = t;
            else
                for (const e of t)
                    this.addHandler(e);
        else {
            if (function (t) { return "object" == typeof t && !Array.isArray(t); }(t)) {
                let e = Array.isArray(t.command) || "string" == typeof t.command ? t.command : this.moduleName(t);
                return t.aliases && (e = [].concat(e).concat(t.aliases)), void this.addHandler(e, this.extractDesc(t), t.builder, t.handler, t.middlewares, t.deprecated);
            }
            if (k(s))
                return void this.addHandler([t].concat(a), e, s.builder, s.handler, s.middlewares, s.deprecated);
        } if ("string" == typeof t) {
            const i = o(t);
            a = a.map((t => o(t).cmd));
            let l = !1;
            const c = [i.cmd].concat(a).filter((t => !_.test(t) || (l = !0, !1)));
            0 === c.length && l && c.push("$0"), l && (i.cmd = c[0], a = c.slice(1), t = t.replace(_, i.cmd)), a.forEach((t => { this.aliasMap[t] = i.cmd; })), !1 !== e && this.usage.command(t, e, l, a, r), this.handlers[i.cmd] = { original: t, description: e, handler: n, builder: s || {}, middlewares: h, deprecated: r, demanded: i.demanded, optional: i.optional }, l && (this.defaultCommand = this.handlers[i.cmd]);
        } }
        getCommandHandlers() { return this.handlers; }
        getCommands() { return Object.keys(this.handlers).concat(Object.keys(this.aliasMap)); }
        hasDefaultCommand() { return !!this.defaultCommand; }
        runCommand(t, e, s, n, i, r) { const o = this.handlers[t] || this.handlers[this.aliasMap[t]] || this.defaultCommand, a = e.getInternalMethods().getContext(), h = a.commands.slice(), l = !t; t && (a.commands.push(t), a.fullCommands.push(o.original)); const c = this.applyBuilderUpdateUsageAndParse(l, o, e, s.aliases, h, n, i, r); return f(c) ? c.then((t => this.applyMiddlewareAndGetResult(l, o, t.innerArgv, a, i, t.aliases, e))) : this.applyMiddlewareAndGetResult(l, o, c.innerArgv, a, i, c.aliases, e); }
        applyBuilderUpdateUsageAndParse(t, e, s, n, i, r, o, a) { const h = e.builder; let l = s; if (E(h)) {
            const c = h(s.getInternalMethods().reset(n), a);
            if (f(c))
                return c.then((n => { var a; return l = (a = n) && "function" == typeof a.getInternalMethods ? n : s, this.parseAndUpdateUsage(t, e, l, i, r, o); }));
        }
        else
            (function (t) { return "object" == typeof t; })(h) && (l = s.getInternalMethods().reset(n), Object.keys(e.builder).forEach((t => { l.option(t, h[t]); }))); return this.parseAndUpdateUsage(t, e, l, i, r, o); }
        parseAndUpdateUsage(t, e, s, n, i, r) { t && s.getInternalMethods().getUsageInstance().unfreeze(), this.shouldUpdateUsage(s) && s.getInternalMethods().getUsageInstance().usage(this.usageFromParentCommandsCommandHandler(n, e), e.description); const o = s.getInternalMethods().runYargsParserAndExecuteCommands(null, void 0, !0, i, r); return f(o) ? o.then((t => ({ aliases: s.parsed.aliases, innerArgv: t }))) : { aliases: s.parsed.aliases, innerArgv: o }; }
        shouldUpdateUsage(t) { return !t.getInternalMethods().getUsageInstance().getUsageDisabled() && 0 === t.getInternalMethods().getUsageInstance().getUsage().length; }
        usageFromParentCommandsCommandHandler(t, e) { const s = _.test(e.original) ? e.original.replace(_, "").trim() : e.original, n = t.filter((t => !_.test(t))); return n.push(s), `$0 ${n.join(" ")}`; }
        applyMiddlewareAndGetResult(t, e, s, n, i, r, o) { let a = {}; if (i)
            return s; o.getInternalMethods().getHasOutput() || (a = this.populatePositionals(e, s, n, o)); const h = this.globalMiddleware.getMiddleware().slice(0).concat(e.middlewares); if (s = C(s, o, h, !0), !o.getInternalMethods().getHasOutput()) {
            const e = o.getInternalMethods().runValidation(r, a, o.parsed.error, t);
            s = j(s, (t => (e(t), t)));
        } if (e.handler && !o.getInternalMethods().getHasOutput()) {
            o.getInternalMethods().setHasOutput();
            const n = !!o.getOptions().configuration["populate--"];
            o.getInternalMethods().postProcess(s, n, !1, !1), s = j(s = C(s, o, h, !1), (t => { const s = e.handler(t); return f(s) ? s.then((() => t)) : t; })), t || o.getInternalMethods().getUsageInstance().cacheHelpMessage(), f(s) && !o.getInternalMethods().hasParseCallback() && s.catch((t => { try {
                o.getInternalMethods().getUsageInstance().fail(null, t);
            }
            catch (t) { } }));
        } return t || (n.commands.pop(), n.fullCommands.pop()), s; }
        populatePositionals(t, e, s, n) { e._ = e._.slice(s.commands.length); const i = t.demanded.slice(0), r = t.optional.slice(0), o = {}; for (this.validation.positionalCount(i.length, e._.length); i.length;) {
            const t = i.shift();
            this.populatePositional(t, e, o);
        } for (; r.length;) {
            const t = r.shift();
            this.populatePositional(t, e, o);
        } return e._ = s.commands.concat(e._.map((t => "" + t))), this.postProcessPositionals(e, o, this.cmdToParseOptions(t.original), n), o; }
        populatePositional(t, e, s) { const n = t.cmd[0]; t.variadic ? s[n] = e._.splice(0).map(String) : e._.length && (s[n] = [String(e._.shift())]); }
        cmdToParseOptions(t) { const e = { array: [], default: {}, alias: {}, demand: {} }, s = o(t); return s.demanded.forEach((t => { const [s, ...n] = t.cmd; t.variadic && (e.array.push(s), e.default[s] = []), e.alias[s] = n, e.demand[s] = !0; })), s.optional.forEach((t => { const [s, ...n] = t.cmd; t.variadic && (e.array.push(s), e.default[s] = []), e.alias[s] = n; })), e; }
        postProcessPositionals(t, e, s, n) { const i = Object.assign({}, n.getOptions()); i.default = Object.assign(s.default, i.default); for (const t of Object.keys(s.alias))
            i.alias[t] = (i.alias[t] || []).concat(s.alias[t]); i.array = i.array.concat(s.array), i.config = {}; const r = []; if (Object.keys(e).forEach((t => { e[t].map((e => { i.configuration["unknown-options-as-args"] && (i.key[t] = !0), r.push(`--${t}`), r.push(e); })); })), !r.length)
            return; const o = Object.assign({}, i.configuration, { "populate--": !1 }), a = this.shim.Parser.detailed(r, Object.assign({}, i, { configuration: o })); if (a.error)
            n.getInternalMethods().getUsageInstance().fail(a.error.message, a.error);
        else {
            const s = Object.keys(e);
            Object.keys(e).forEach((t => { s.push(...a.aliases[t]); })), Object.keys(a.argv).forEach((n => { s.includes(n) && (e[n] || (e[n] = a.argv[n]), Object.prototype.hasOwnProperty.call(t, n) && Object.prototype.hasOwnProperty.call(a.argv, n) && (Array.isArray(t[n]) || Array.isArray(a.argv[n])) ? t[n] = [].concat(t[n], a.argv[n]) : t[n] = a.argv[n]); }));
        } }
        runDefaultBuilderOn(t) { if (!this.defaultCommand)
            return; if (this.shouldUpdateUsage(t)) {
            const e = _.test(this.defaultCommand.original) ? this.defaultCommand.original : this.defaultCommand.original.replace(/^[^[\]<>]*/, "$0 ");
            t.getInternalMethods().getUsageInstance().usage(e, this.defaultCommand.description);
        } const e = this.defaultCommand.builder; if (E(e))
            return e(t, !0); k(e) || Object.keys(e).forEach((s => { t.option(s, e[s]); })); }
        moduleName(t) { const e = function (t) { if ("undefined" == typeof require)
            return null; for (let e, s = 0, n = Object.keys(require.cache); s < n.length; s++)
            if (e = require.cache[n[s]], e.exports === t)
                return e; return null; }(t); if (!e)
            throw new Error(`No command name given for module: ${this.shim.inspect(t)}`); return this.commandFromFilename(e.filename); }
        commandFromFilename(t) { return this.shim.path.basename(t, this.shim.path.extname(t)); }
        extractDesc({ describe: t, description: e, desc: s }) { for (const n of [t, e, s]) {
            if ("string" == typeof n || !1 === n)
                return n;
            d(n, !0, this.shim);
        } return !1; }
        freeze() { this.frozens.push({ handlers: this.handlers, aliasMap: this.aliasMap, defaultCommand: this.defaultCommand }); }
        unfreeze() { const t = this.frozens.pop(); d(t, void 0, this.shim), ({ handlers: this.handlers, aliasMap: this.aliasMap, defaultCommand: this.defaultCommand } = t); }
        reset() { return this.handlers = {}, this.aliasMap = {}, this.defaultCommand = void 0, this.requireCache = new Set, this; }
    }
    function k(t) { return "object" == typeof t && !!t.builder && "function" == typeof t.handler; }
    function E(t) { return "function" == typeof t; }
    function x(t) { "undefined" != typeof process && [process.stdout, process.stderr].forEach((e => { const s = e; s._handle && s.isTTY && "function" == typeof s._handle.setBlocking && s._handle.setBlocking(t); })); }
    function S(t) { return "boolean" == typeof t; }
    function $(t, s) { const n = s.y18n.__, i = {}, r = []; i.failFn = function (t) { r.push(t); }; let o = null, a = !0; i.showHelpOnFail = function (t = !0, e) { const [s, n] = "string" == typeof t ? [!0, t] : [t, e]; return o = n, a = s, i; }; let h = !1; i.fail = function (s, n) { const l = t.getInternalMethods().getLoggerInstance(); if (!r.length) {
        if (t.getExitProcess() && x(!0), h || (h = !0, a && (t.showHelp("error"), l.error()), (s || n) && l.error(s || n), o && ((s || n) && l.error(""), l.error(o))), n = n || new e(s), t.getExitProcess())
            return t.exit(1);
        if (t.getInternalMethods().hasParseCallback())
            return t.exit(1, n);
        throw n;
    } for (let t = r.length - 1; t >= 0; --t) {
        const e = r[t];
        if (S(e)) {
            if (n)
                throw n;
            if (s)
                throw Error(s);
        }
        else
            e(s, n, i);
    } }; let l = [], c = !1; i.usage = (t, e) => null === t ? (c = !0, l = [], i) : (c = !1, l.push([t, e || ""]), i), i.getUsage = () => l, i.getUsageDisabled = () => c, i.getPositionalGroupName = () => n("Positionals:"); let f = []; i.example = (t, e) => { f.push([t, e || ""]); }; let d = []; i.command = function (t, e, s, n, i = !1) { s && (d = d.map((t => (t[2] = !1, t)))), d.push([t, e || "", s, n, i]); }, i.getCommands = () => d; let u = {}; i.describe = function (t, e) { Array.isArray(t) ? t.forEach((t => { i.describe(t, e); })) : "object" == typeof t ? Object.keys(t).forEach((e => { i.describe(e, t[e]); })) : u[t] = e; }, i.getDescriptions = () => u; let p = []; i.epilog = t => { p.push(t); }; let m, y = !1; function b() { return y || (m = function () { const t = 80; return s.process.stdColumns ? Math.min(t, s.process.stdColumns) : t; }(), y = !0), m; } i.wrap = t => { y = !0, m = t; }; const v = "__yargsString__:"; function w(t, e, n) { let i = 0; return Array.isArray(t) || (t = Object.values(t).map((t => [t]))), t.forEach((t => { i = Math.max(s.stringWidth(n ? `${n} ${I(t[0])}` : I(t[0])) + P(t[0]), i); })), e && (i = Math.min(i, parseInt((.5 * e).toString(), 10))), i; } let O; function C(e) { return t.getOptions().hiddenOptions.indexOf(e) < 0 || t.parsed.argv[t.getOptions().showHiddenOpt]; } function j(t, e) { let s = `[${n("default:")} `; if (void 0 === t && !e)
        return null; if (e)
        s += e;
    else
        switch (typeof t) {
            case "string":
                s += `"${t}"`;
                break;
            case "object":
                s += JSON.stringify(t);
                break;
            default: s += t;
        } return `${s}]`; } i.deferY18nLookup = t => v + t, i.help = function () { if (O)
        return O; !function () { const e = t.getDemandedOptions(), s = t.getOptions(); (Object.keys(s.alias) || []).forEach((n => { s.alias[n].forEach((r => { u[r] && i.describe(n, u[r]), r in e && t.demandOption(n, e[r]), s.boolean.includes(r) && t.boolean(n), s.count.includes(r) && t.count(n), s.string.includes(r) && t.string(n), s.normalize.includes(r) && t.normalize(n), s.array.includes(r) && t.array(n), s.number.includes(r) && t.number(n); })); })); }(); const e = t.customScriptName ? t.$0 : s.path.basename(t.$0), r = t.getDemandedOptions(), o = t.getDemandedCommands(), a = t.getDeprecatedOptions(), h = t.getGroups(), g = t.getOptions(); let m = []; m = m.concat(Object.keys(u)), m = m.concat(Object.keys(r)), m = m.concat(Object.keys(o)), m = m.concat(Object.keys(g.default)), m = m.filter(C), m = Object.keys(m.reduce(((t, e) => ("_" !== e && (t[e] = !0), t)), {})); const y = b(), _ = s.cliui({ width: y, wrap: !!y }); if (!c)
        if (l.length)
            l.forEach((t => { _.div({ text: `${t[0].replace(/\$0/g, e)}` }), t[1] && _.div({ text: `${t[1]}`, padding: [1, 0, 0, 0] }); })), _.div();
        else if (d.length) {
            let t = null;
            t = o._ ? `${e} <${n("command")}>\n` : `${e} [${n("command")}]\n`, _.div(`${t}`);
        } if (d.length > 1 || 1 === d.length && !d[0][2]) {
        _.div(n("Commands:"));
        const s = t.getInternalMethods().getContext(), i = s.commands.length ? `${s.commands.join(" ")} ` : "";
        !0 === t.getInternalMethods().getParserConfiguration()["sort-commands"] && (d = d.sort(((t, e) => t[0].localeCompare(e[0])))), d.forEach((t => { const s = `${e} ${i}${t[0].replace(/^\$0 ?/, "")}`; _.span({ text: s, padding: [0, 2, 0, 2], width: w(d, y, `${e}${i}`) + 4 }, { text: t[1] }); const r = []; t[2] && r.push(`[${n("default")}]`), t[3] && t[3].length && r.push(`[${n("aliases:")} ${t[3].join(", ")}]`), t[4] && ("string" == typeof t[4] ? r.push(`[${n("deprecated: %s", t[4])}]`) : r.push(`[${n("deprecated")}]`)), r.length ? _.div({ text: r.join(" "), padding: [0, 0, 0, 2], align: "right" }) : _.div(); })), _.div();
    } const M = (Object.keys(g.alias) || []).concat(Object.keys(t.parsed.newAliases) || []); m = m.filter((e => !t.parsed.newAliases[e] && M.every((t => -1 === (g.alias[t] || []).indexOf(e))))); const k = n("Options:"); h[k] || (h[k] = []), function (t, e, s, n) { let i = [], r = null; Object.keys(s).forEach((t => { i = i.concat(s[t]); })), t.forEach((t => { r = [t].concat(e[t]), r.some((t => -1 !== i.indexOf(t))) || s[n].push(t); })); }(m, g.alias, h, k); const E = t => /^--/.test(I(t)), x = Object.keys(h).filter((t => h[t].length > 0)).map((t => ({ groupName: t, normalizedKeys: h[t].filter(C).map((t => { if (M.includes(t))
            return t; for (let e, s = 0; void 0 !== (e = M[s]); s++)
            if ((g.alias[e] || []).includes(t))
                return e; return t; })) }))).filter((({ normalizedKeys: t }) => t.length > 0)).map((({ groupName: t, normalizedKeys: e }) => { const s = e.reduce(((e, s) => (e[s] = [s].concat(g.alias[s] || []).map((e => t === i.getPositionalGroupName() ? e : (/^[0-9]$/.test(e) ? g.boolean.includes(s) ? "-" : "--" : e.length > 1 ? "--" : "-") + e)).sort(((t, e) => E(t) === E(e) ? 0 : E(t) ? 1 : -1)).join(", "), e)), {}); return { groupName: t, normalizedKeys: e, switches: s }; })); if (x.filter((({ groupName: t }) => t !== i.getPositionalGroupName())).some((({ normalizedKeys: t, switches: e }) => !t.every((t => E(e[t]))))) && x.filter((({ groupName: t }) => t !== i.getPositionalGroupName())).forEach((({ normalizedKeys: t, switches: e }) => { t.forEach((t => { var s, n; E(e[t]) && (e[t] = (s = e[t], n = "-x, ".length, A(s) ? { text: s.text, indentation: s.indentation + n } : { text: s, indentation: n })); })); })), x.forEach((({ groupName: t, normalizedKeys: e, switches: s }) => { _.div(t), e.forEach((t => { const e = s[t]; let o = u[t] || "", h = null; o.includes(v) && (o = n(o.substring(v.length))), g.boolean.includes(t) && (h = `[${n("boolean")}]`), g.count.includes(t) && (h = `[${n("count")}]`), g.string.includes(t) && (h = `[${n("string")}]`), g.normalize.includes(t) && (h = `[${n("string")}]`), g.array.includes(t) && (h = `[${n("array")}]`), g.number.includes(t) && (h = `[${n("number")}]`); const l = [t in a ? (c = a[t], "string" == typeof c ? `[${n("deprecated: %s", c)}]` : `[${n("deprecated")}]`) : null, h, t in r ? `[${n("required")}]` : null, g.choices && g.choices[t] ? `[${n("choices:")} ${i.stringifiedValues(g.choices[t])}]` : null, j(g.default[t], g.defaultDescription[t])].filter(Boolean).join(" "); var c; _.span({ text: I(e), padding: [0, 2, 0, 2 + P(e)], width: w(s, y) + 4 }, o), l ? _.div({ text: l, padding: [0, 0, 0, 2], align: "right" }) : _.div(); })), _.div(); })), f.length && (_.div(n("Examples:")), f.forEach((t => { t[0] = t[0].replace(/\$0/g, e); })), f.forEach((t => { "" === t[1] ? _.div({ text: t[0], padding: [0, 2, 0, 2] }) : _.div({ text: t[0], padding: [0, 2, 0, 2], width: w(f, y) + 4 }, { text: t[1] }); })), _.div()), p.length > 0) {
        const t = p.map((t => t.replace(/\$0/g, e))).join("\n");
        _.div(`${t}\n`);
    } return _.toString().replace(/\s*$/, ""); }, i.cacheHelpMessage = function () { O = this.help(); }, i.clearCachedHelpMessage = function () { O = void 0; }, i.hasCachedHelpMessage = function () { return !!O; }, i.showHelp = e => { const s = t.getInternalMethods().getLoggerInstance(); e || (e = "error"); ("function" == typeof e ? e : s[e])(i.help()); }, i.functionDescription = t => ["(", t.name ? s.Parser.decamelize(t.name, "-") : n("generated-value"), ")"].join(""), i.stringifiedValues = function (t, e) { let s = ""; const n = e || ", ", i = [].concat(t); return t && i.length ? (i.forEach((t => { s.length && (s += n), s += JSON.stringify(t); })), s) : s; }; let _ = null; i.version = t => { _ = t; }, i.showVersion = e => { const s = t.getInternalMethods().getLoggerInstance(); e || (e = "error"); ("function" == typeof e ? e : s[e])(_); }, i.reset = function (t) { return o = null, h = !1, l = [], c = !1, p = [], f = [], d = [], u = g(u, (e => !t[e])), i; }; const M = []; return i.freeze = function () { M.push({ failMessage: o, failureOutput: h, usages: l, usageDisabled: c, epilogs: p, examples: f, commands: d, descriptions: u }); }, i.unfreeze = function () { const t = M.pop(); t && ({ failMessage: o, failureOutput: h, usages: l, usageDisabled: c, epilogs: p, examples: f, commands: d, descriptions: u } = t); }, i; }
    function A(t) { return "object" == typeof t; }
    function P(t) { return A(t) ? t.indentation : 0; }
    function I(t) { return A(t) ? t.text : t; }
    class N {
        constructor(t, e, s, n) { var i, r, o; this.yargs = t, this.usage = e, this.command = s, this.shim = n, this.completionKey = "get-yargs-completions", this.aliases = null, this.customCompletionFunction = null, this.zshShell = null !== (o = (null === (i = this.shim.getEnv("SHELL")) || void 0 === i ? void 0 : i.includes("zsh")) || (null === (r = this.shim.getEnv("ZSH_NAME")) || void 0 === r ? void 0 : r.includes("zsh"))) && void 0 !== o && o; }
        defaultCompletion(t, e, s, n) { const i = this.command.getCommandHandlers(); for (let e = 0, s = t.length; e < s; ++e)
            if (i[t[e]] && i[t[e]].builder) {
                const s = i[t[e]].builder;
                if (E(s)) {
                    const t = this.yargs.getInternalMethods().reset();
                    return s(t, !0), t.argv;
                }
            } const r = []; this.commandCompletions(r, t, s), this.optionCompletions(r, t, e, s), n(null, r); }
        commandCompletions(t, e, s) { const n = this.yargs.getInternalMethods().getContext().commands; s.match(/^-/) || n[n.length - 1] === s || this.usage.getCommands().forEach((s => { const n = o(s[0]).cmd; if (-1 === e.indexOf(n))
            if (this.zshShell) {
                const e = s[1] || "";
                t.push(n.replace(/:/g, "\\:") + ":" + e);
            }
            else
                t.push(n); })); }
        optionCompletions(t, e, s, n) { if (n.match(/^-/) || "" === n && 0 === t.length) {
            const i = this.yargs.getOptions(), r = this.yargs.getGroups()[this.usage.getPositionalGroupName()] || [];
            Object.keys(i.key).forEach((o => { const a = !!i.configuration["boolean-negation"] && i.boolean.includes(o); r.includes(o) || this.argsContainKey(e, s, o, a) || (this.completeOptionKey(o, t, n), a && i.default[o] && this.completeOptionKey(`no-${o}`, t, n)); }));
        } }
        argsContainKey(t, e, s, n) { if (-1 !== t.indexOf(`--${s}`))
            return !0; if (n && -1 !== t.indexOf(`--no-${s}`))
            return !0; if (this.aliases)
            for (const t of this.aliases[s])
                if (void 0 !== e[t])
                    return !0; return !1; }
        completeOptionKey(t, e, s) { const n = this.usage.getDescriptions(), i = !/^--/.test(s) && (t => /^[^0-9]$/.test(t))(t) ? "-" : "--"; if (this.zshShell) {
            const s = n[t] || "";
            e.push(i + `${t.replace(/:/g, "\\:")}:${s.replace("__yargsString__:", "")}`);
        }
        else
            e.push(i + t); }
        customCompletion(t, e, s, n) { if (d(this.customCompletionFunction, null, this.shim), this.customCompletionFunction.length < 3) {
            const t = this.customCompletionFunction(s, e);
            return f(t) ? t.then((t => { this.shim.process.nextTick((() => { n(null, t); })); })).catch((t => { this.shim.process.nextTick((() => { n(t, void 0); })); })) : n(null, t);
        } return function (t) { return t.length > 3; }(this.customCompletionFunction) ? this.customCompletionFunction(s, e, ((i = n) => this.defaultCompletion(t, e, s, i)), (t => { n(null, t); })) : this.customCompletionFunction(s, e, (t => { n(null, t); })); }
        getCompletion(t, e) { const s = t.length ? t[t.length - 1] : "", n = this.yargs.parse(t, !0), i = this.customCompletionFunction ? n => this.customCompletion(t, n, s, e) : n => this.defaultCompletion(t, n, s, e); return f(n) ? n.then(i) : i(n); }
        generateCompletionScript(t, e) { let s = this.zshShell ? '#compdef {{app_name}}\n###-begin-{{app_name}}-completions-###\n#\n# yargs command completion script\n#\n# Installation: {{app_path}} {{completion_command}} >> ~/.zshrc\n#    or {{app_path}} {{completion_command}} >> ~/.zsh_profile on OSX.\n#\n_{{app_name}}_yargs_completions()\n{\n  local reply\n  local si=$IFS\n  IFS=$\'\n\' reply=($(COMP_CWORD="$((CURRENT-1))" COMP_LINE="$BUFFER" COMP_POINT="$CURSOR" {{app_path}} --get-yargs-completions "${words[@]}"))\n  IFS=$si\n  _describe \'values\' reply\n}\ncompdef _{{app_name}}_yargs_completions {{app_name}}\n###-end-{{app_name}}-completions-###\n' : '###-begin-{{app_name}}-completions-###\n#\n# yargs command completion script\n#\n# Installation: {{app_path}} {{completion_command}} >> ~/.bashrc\n#    or {{app_path}} {{completion_command}} >> ~/.bash_profile on OSX.\n#\n_{{app_name}}_yargs_completions()\n{\n    local cur_word args type_list\n\n    cur_word="${COMP_WORDS[COMP_CWORD]}"\n    args=("${COMP_WORDS[@]}")\n\n    # ask yargs to generate completions.\n    type_list=$({{app_path}} --get-yargs-completions "${args[@]}")\n\n    COMPREPLY=( $(compgen -W "${type_list}" -- ${cur_word}) )\n\n    # if no match was found, fall back to filename completion\n    if [ ${#COMPREPLY[@]} -eq 0 ]; then\n      COMPREPLY=()\n    fi\n\n    return 0\n}\ncomplete -o default -F _{{app_name}}_yargs_completions {{app_name}}\n###-end-{{app_name}}-completions-###\n'; const n = this.shim.path.basename(t); return t.match(/\.js$/) && (t = `./${t}`), s = s.replace(/{{app_name}}/g, n), s = s.replace(/{{completion_command}}/g, e), s.replace(/{{app_path}}/g, t); }
        registerFunction(t) { this.customCompletionFunction = t; }
        setParsed(t) { this.aliases = t.aliases; }
    }
    function D(t, e) { if (0 === t.length)
        return e.length; if (0 === e.length)
        return t.length; const s = []; let n, i; for (n = 0; n <= e.length; n++)
        s[n] = [n]; for (i = 0; i <= t.length; i++)
        s[0][i] = i; for (n = 1; n <= e.length; n++)
        for (i = 1; i <= t.length; i++)
            e.charAt(n - 1) === t.charAt(i - 1) ? s[n][i] = s[n - 1][i - 1] : n > 1 && i > 1 && e.charAt(n - 2) === t.charAt(i - 1) && e.charAt(n - 1) === t.charAt(i - 2) ? s[n][i] = s[n - 2][i - 2] + 1 : s[n][i] = Math.min(s[n - 1][i - 1] + 1, Math.min(s[n][i - 1] + 1, s[n - 1][i] + 1)); return s[e.length][t.length]; }
    const z = ["$0", "--", "_"];
    var q, H, F, U, W, L, V, R, T, G, B, K, Y, J, Z, X, Q, tt, et, st, nt, it, rt, ot, at, ht, lt, ct, ft, dt, ut, pt;
    const gt = Symbol("copyDoubleDash"), mt = Symbol("copyDoubleDash"), yt = Symbol("deleteFromParserHintObject"), bt = Symbol("freeze"), vt = Symbol("getDollarZero"), wt = Symbol("getParserConfiguration"), Ot = Symbol("guessLocale"), Ct = Symbol("guessVersion"), jt = Symbol("parsePositionalNumbers"), _t = Symbol("pkgUp"), Mt = Symbol("populateParserHintArray"), kt = Symbol("populateParserHintSingleValueDictionary"), Et = Symbol("populateParserHintArrayDictionary"), xt = Symbol("populateParserHintDictionary"), St = Symbol("sanitizeKey"), $t = Symbol("setKey"), At = Symbol("unfreeze"), Pt = Symbol("validateAsync"), It = Symbol("getCommandInstance"), Nt = Symbol("getContext"), Dt = Symbol("getHasOutput"), zt = Symbol("getLoggerInstance"), qt = Symbol("getParseContext"), Ht = Symbol("getUsageInstance"), Ft = Symbol("getValidationInstance"), Ut = Symbol("hasParseCallback"), Wt = Symbol("postProcess"), Lt = Symbol("rebase"), Vt = Symbol("reset"), Rt = Symbol("runYargsParserAndExecuteCommands"), Tt = Symbol("runValidation"), Gt = Symbol("setHasOutput");
    class Bt {
        constructor(t = [], e, s, n) { this.customScriptName = !1, this.parsed = !1, q.set(this, void 0), H.set(this, void 0), F.set(this, { commands: [], fullCommands: [] }), U.set(this, null), W.set(this, null), L.set(this, "show-hidden"), V.set(this, null), R.set(this, !0), T.set(this, !0), G.set(this, []), B.set(this, void 0), K.set(this, {}), Y.set(this, !1), J.set(this, null), Z.set(this, void 0), X.set(this, ""), Q.set(this, void 0), tt.set(this, void 0), et.set(this, {}), st.set(this, null), nt.set(this, null), it.set(this, {}), rt.set(this, {}), ot.set(this, void 0), at.set(this, !1), ht.set(this, void 0), lt.set(this, !1), ct.set(this, !1), ft.set(this, !1), dt.set(this, void 0), ut.set(this, null), pt.set(this, void 0), w(this, ht, n, "f"), w(this, ot, t, "f"), w(this, H, e, "f"), w(this, tt, s, "f"), w(this, B, new O(this), "f"), this.$0 = this[vt](), this[Vt](), w(this, q, v(this, q, "f"), "f"), w(this, dt, v(this, dt, "f"), "f"), w(this, pt, v(this, pt, "f"), "f"), w(this, Q, v(this, Q, "f"), "f"), v(this, Q, "f").showHiddenOpt = v(this, L, "f"), w(this, Z, this[mt](), "f"); }
        addHelpOpt(t, e) { return h("[string|boolean] [string]", [t, e], arguments.length), v(this, J, "f") && (this[yt](v(this, J, "f")), w(this, J, null, "f")), !1 === t && void 0 === e || (w(this, J, "string" == typeof t ? t : "help", "f"), this.boolean(v(this, J, "f")), this.describe(v(this, J, "f"), e || v(this, dt, "f").deferY18nLookup("Show help"))), this; }
        help(t, e) { return this.addHelpOpt(t, e); }
        addShowHiddenOpt(t, e) { if (h("[string|boolean] [string]", [t, e], arguments.length), !1 === t && void 0 === e)
            return this; const s = "string" == typeof t ? t : v(this, L, "f"); return this.boolean(s), this.describe(s, e || v(this, dt, "f").deferY18nLookup("Show hidden options")), v(this, Q, "f").showHiddenOpt = s, this; }
        showHidden(t, e) { return this.addShowHiddenOpt(t, e); }
        alias(t, e) { return h("<object|string|array> [string|array]", [t, e], arguments.length), this[Et](this.alias.bind(this), "alias", t, e), this; }
        array(t) { return h("<array|string>", [t], arguments.length), this[Mt]("array", t), this; }
        boolean(t) { return h("<array|string>", [t], arguments.length), this[Mt]("boolean", t), this; }
        check(t, e) { return h("<function> [boolean]", [t, e], arguments.length), this.middleware(((e, s) => j((() => t(e)), (s => (s ? ("string" == typeof s || s instanceof Error) && v(this, dt, "f").fail(s.toString(), s) : v(this, dt, "f").fail(v(this, ht, "f").y18n.__("Argument check failed: %s", t.toString())), e)), (t => (v(this, dt, "f").fail(t.message ? t.message : t.toString(), t), e)))), !1, e), this; }
        choices(t, e) { return h("<object|string|array> [string|array]", [t, e], arguments.length), this[Et](this.choices.bind(this), "choices", t, e), this; }
        coerce(t, s) { if (h("<object|string|array> [function]", [t, s], arguments.length), Array.isArray(t)) {
            if (!s)
                throw new e("coerce callback must be provided");
            for (const e of t)
                this.coerce(e, s);
            return this;
        } if ("object" == typeof t) {
            for (const e of Object.keys(t))
                this.coerce(e, t[e]);
            return this;
        } if (!s)
            throw new e("coerce callback must be provided"); return v(this, Q, "f").key[t] = !0, v(this, B, "f").addCoerceMiddleware(((n, i) => { let r; return j((() => (r = i.getAliases(), s(n[t]))), (e => { if (n[t] = e, r[t])
            for (const s of r[t])
                n[s] = e; return n; }), (t => { throw new e(t.message); })); }), t), this; }
        conflicts(t, e) { return h("<string|object> [string|array]", [t, e], arguments.length), v(this, pt, "f").conflicts(t, e), this; }
        config(t = "config", e, s) { return h("[object|string] [string|function] [function]", [t, e, s], arguments.length), "object" != typeof t || Array.isArray(t) ? ("function" == typeof e && (s = e, e = void 0), this.describe(t, e || v(this, dt, "f").deferY18nLookup("Path to JSON config file")), (Array.isArray(t) ? t : [t]).forEach((t => { v(this, Q, "f").config[t] = s || !0; })), this) : (t = i(t, v(this, H, "f"), this[wt]()["deep-merge-config"] || !1, v(this, ht, "f")), v(this, Q, "f").configObjects = (v(this, Q, "f").configObjects || []).concat(t), this); }
        completion(t, e, s) { return h("[string] [string|boolean|function] [function]", [t, e, s], arguments.length), "function" == typeof e && (s = e, e = void 0), w(this, W, t || v(this, W, "f") || "completion", "f"), e || !1 === e || (e = "generate completion script"), this.command(v(this, W, "f"), e), s && v(this, U, "f").registerFunction(s), this; }
        command(t, e, s, n, i, r) { return h("<string|array|object> [string|boolean] [function|object] [function] [array] [boolean|string]", [t, e, s, n, i, r], arguments.length), v(this, q, "f").addHandler(t, e, s, n, i, r), this; }
        commands(t, e, s, n, i, r) { return this.command(t, e, s, n, i, r); }
        commandDir(t, e) { h("<string> [object]", [t, e], arguments.length); const s = v(this, tt, "f") || v(this, ht, "f").require; return v(this, q, "f").addDirectory(t, s, v(this, ht, "f").getCallerFile(), e), this; }
        count(t) { return h("<array|string>", [t], arguments.length), this[Mt]("count", t), this; }
        default(t, e, s) { return h("<object|string|array> [*] [string]", [t, e, s], arguments.length), s && (u(t, v(this, ht, "f")), v(this, Q, "f").defaultDescription[t] = s), "function" == typeof e && (u(t, v(this, ht, "f")), v(this, Q, "f").defaultDescription[t] || (v(this, Q, "f").defaultDescription[t] = v(this, dt, "f").functionDescription(e)), e = e.call()), this[kt](this.default.bind(this), "default", t, e), this; }
        defaults(t, e, s) { return this.default(t, e, s); }
        demandCommand(t = 1, e, s, n) { return h("[number] [number|string] [string|null|undefined] [string|null|undefined]", [t, e, s, n], arguments.length), "number" != typeof e && (s = e, e = 1 / 0), this.global("_", !1), v(this, Q, "f").demandedCommands._ = { min: t, max: e, minMsg: s, maxMsg: n }, this; }
        demand(t, e, s) { return Array.isArray(e) ? (e.forEach((t => { d(s, !0, v(this, ht, "f")), this.demandOption(t, s); })), e = 1 / 0) : "number" != typeof e && (s = e, e = 1 / 0), "number" == typeof t ? (d(s, !0, v(this, ht, "f")), this.demandCommand(t, e, s, s)) : Array.isArray(t) ? t.forEach((t => { d(s, !0, v(this, ht, "f")), this.demandOption(t, s); })) : "string" == typeof s ? this.demandOption(t, s) : !0 !== s && void 0 !== s || this.demandOption(t), this; }
        demandOption(t, e) { return h("<object|string|array> [string]", [t, e], arguments.length), this[kt](this.demandOption.bind(this), "demandedOptions", t, e), this; }
        deprecateOption(t, e) { return h("<string> [string|boolean]", [t, e], arguments.length), v(this, Q, "f").deprecatedOptions[t] = e, this; }
        describe(t, e) { return h("<object|string|array> [string]", [t, e], arguments.length), this[$t](t, !0), v(this, dt, "f").describe(t, e), this; }
        detectLocale(t) { return h("<boolean>", [t], arguments.length), w(this, R, t, "f"), this; }
        env(t) { return h("[string|boolean]", [t], arguments.length), !1 === t ? delete v(this, Q, "f").envPrefix : v(this, Q, "f").envPrefix = t || "", this; }
        epilogue(t) { return h("<string>", [t], arguments.length), v(this, dt, "f").epilog(t), this; }
        epilog(t) { return this.epilogue(t); }
        example(t, e) { return h("<string|array> [string]", [t, e], arguments.length), Array.isArray(t) ? t.forEach((t => this.example(...t))) : v(this, dt, "f").example(t, e), this; }
        exit(t, e) { w(this, Y, !0, "f"), w(this, V, e, "f"), v(this, T, "f") && v(this, ht, "f").process.exit(t); }
        exitProcess(t = !0) { return h("[boolean]", [t], arguments.length), w(this, T, t, "f"), this; }
        fail(t) { if (h("<function|boolean>", [t], arguments.length), "boolean" == typeof t && !1 !== t)
            throw new e("Invalid first argument. Expected function or boolean 'false'"); return v(this, dt, "f").failFn(t), this; }
        getAliases() { return this.parsed ? this.parsed.aliases : {}; }
        async getCompletion(t, e) { return h("<array> [function]", [t, e], arguments.length), e ? v(this, U, "f").getCompletion(t, e) : new Promise(((e, s) => { v(this, U, "f").getCompletion(t, ((t, n) => { t ? s(t) : e(n); })); })); }
        getDemandedOptions() { return h([], 0), v(this, Q, "f").demandedOptions; }
        getDemandedCommands() { return h([], 0), v(this, Q, "f").demandedCommands; }
        getDeprecatedOptions() { return h([], 0), v(this, Q, "f").deprecatedOptions; }
        getDetectLocale() { return v(this, R, "f"); }
        getExitProcess() { return v(this, T, "f"); }
        getGroups() { return Object.assign({}, v(this, K, "f"), v(this, rt, "f")); }
        getHelp() { if (w(this, Y, !0, "f"), !v(this, dt, "f").hasCachedHelpMessage()) {
            if (!this.parsed) {
                const t = this[Rt](v(this, ot, "f"), void 0, void 0, 0, !0);
                if (f(t))
                    return t.then((() => v(this, dt, "f").help()));
            }
            const t = v(this, q, "f").runDefaultBuilderOn(this);
            if (f(t))
                return t.then((() => v(this, dt, "f").help()));
        } return Promise.resolve(v(this, dt, "f").help()); }
        getOptions() { return v(this, Q, "f"); }
        getStrict() { return v(this, lt, "f"); }
        getStrictCommands() { return v(this, ct, "f"); }
        getStrictOptions() { return v(this, ft, "f"); }
        global(t, e) { return h("<string|array> [boolean]", [t, e], arguments.length), t = [].concat(t), !1 !== e ? v(this, Q, "f").local = v(this, Q, "f").local.filter((e => -1 === t.indexOf(e))) : t.forEach((t => { v(this, Q, "f").local.includes(t) || v(this, Q, "f").local.push(t); })), this; }
        group(t, e) { h("<string|array> <string>", [t, e], arguments.length); const s = v(this, rt, "f")[e] || v(this, K, "f")[e]; v(this, rt, "f")[e] && delete v(this, rt, "f")[e]; const n = {}; return v(this, K, "f")[e] = (s || []).concat(t).filter((t => !n[t] && (n[t] = !0))), this; }
        hide(t) { return h("<string>", [t], arguments.length), v(this, Q, "f").hiddenOptions.push(t), this; }
        implies(t, e) { return h("<string|object> [number|string|array]", [t, e], arguments.length), v(this, pt, "f").implies(t, e), this; }
        locale(t) { return h("[string]", [t], arguments.length), t ? (w(this, R, !1, "f"), v(this, ht, "f").y18n.setLocale(t), this) : (this[Ot](), v(this, ht, "f").y18n.getLocale()); }
        middleware(t, e, s) { return v(this, B, "f").addMiddleware(t, !!e, s); }
        nargs(t, e) { return h("<string|object|array> [number]", [t, e], arguments.length), this[kt](this.nargs.bind(this), "narg", t, e), this; }
        normalize(t) { return h("<array|string>", [t], arguments.length), this[Mt]("normalize", t), this; }
        number(t) { return h("<array|string>", [t], arguments.length), this[Mt]("number", t), this; }
        option(t, e) { if (h("<string|object> [object]", [t, e], arguments.length), "object" == typeof t)
            Object.keys(t).forEach((e => { this.options(e, t[e]); }));
        else {
            "object" != typeof e && (e = {}), v(this, Q, "f").key[t] = !0, e.alias && this.alias(t, e.alias);
            const s = e.deprecate || e.deprecated;
            s && this.deprecateOption(t, s);
            const n = e.demand || e.required || e.require;
            n && this.demand(t, n), e.demandOption && this.demandOption(t, "string" == typeof e.demandOption ? e.demandOption : void 0), e.conflicts && this.conflicts(t, e.conflicts), "default" in e && this.default(t, e.default), void 0 !== e.implies && this.implies(t, e.implies), void 0 !== e.nargs && this.nargs(t, e.nargs), e.config && this.config(t, e.configParser), e.normalize && this.normalize(t), e.choices && this.choices(t, e.choices), e.coerce && this.coerce(t, e.coerce), e.group && this.group(t, e.group), (e.boolean || "boolean" === e.type) && (this.boolean(t), e.alias && this.boolean(e.alias)), (e.array || "array" === e.type) && (this.array(t), e.alias && this.array(e.alias)), (e.number || "number" === e.type) && (this.number(t), e.alias && this.number(e.alias)), (e.string || "string" === e.type) && (this.string(t), e.alias && this.string(e.alias)), (e.count || "count" === e.type) && this.count(t), "boolean" == typeof e.global && this.global(t, e.global), e.defaultDescription && (v(this, Q, "f").defaultDescription[t] = e.defaultDescription), e.skipValidation && this.skipValidation(t);
            const i = e.describe || e.description || e.desc;
            this.describe(t, i), e.hidden && this.hide(t), e.requiresArg && this.requiresArg(t);
        } return this; }
        options(t, e) { return this.option(t, e); }
        parse(t, e, s) { h("[string|array] [function|boolean|object] [function]", [t, e, s], arguments.length), this[bt](), void 0 === t && (t = v(this, ot, "f")), "object" == typeof e && (w(this, nt, e, "f"), e = s), "function" == typeof e && (w(this, st, e, "f"), e = !1), e || w(this, ot, t, "f"), v(this, st, "f") && w(this, T, !1, "f"); const n = this[Rt](t, !!e), i = this.parsed; return v(this, U, "f").setParsed(this.parsed), f(n) ? n.then((t => (v(this, st, "f") && v(this, st, "f").call(this, v(this, V, "f"), t, v(this, X, "f")), t))).catch((t => { throw v(this, st, "f") && v(this, st, "f")(t, this.parsed.argv, v(this, X, "f")), t; })).finally((() => { this[At](), this.parsed = i; })) : (v(this, st, "f") && v(this, st, "f").call(this, v(this, V, "f"), n, v(this, X, "f")), this[At](), this.parsed = i, n); }
        parseAsync(t, e, s) { const n = this.parse(t, e, s); return f(n) ? n : Promise.resolve(n); }
        parseSync(t, s, n) { const i = this.parse(t, s, n); if (f(i))
            throw new e(".parseSync() must not be used with asynchronous builders, handlers, or middleware"); return i; }
        parserConfiguration(t) { return h("<object>", [t], arguments.length), w(this, et, t, "f"), this; }
        pkgConf(t, e) { h("<string> [string]", [t, e], arguments.length); let s = null; const n = this[_t](e || v(this, H, "f")); return n[t] && "object" == typeof n[t] && (s = i(n[t], e || v(this, H, "f"), this[wt]()["deep-merge-config"] || !1, v(this, ht, "f")), v(this, Q, "f").configObjects = (v(this, Q, "f").configObjects || []).concat(s)), this; }
        positional(t, e) { h("<string> <object>", [t, e], arguments.length); const s = ["default", "defaultDescription", "implies", "normalize", "choices", "conflicts", "coerce", "type", "describe", "desc", "description", "alias"]; e = g(e, ((t, e) => !("type" === t && !["string", "number", "boolean"].includes(e)) && s.includes(t))); const n = v(this, F, "f").fullCommands[v(this, F, "f").fullCommands.length - 1], i = n ? v(this, q, "f").cmdToParseOptions(n) : { array: [], alias: {}, default: {}, demand: {} }; return p(i).forEach((s => { const n = i[s]; Array.isArray(n) ? -1 !== n.indexOf(t) && (e[s] = !0) : n[t] && !(s in e) && (e[s] = n[t]); })), this.group(t, v(this, dt, "f").getPositionalGroupName()), this.option(t, e); }
        recommendCommands(t = !0) { return h("[boolean]", [t], arguments.length), w(this, at, t, "f"), this; }
        required(t, e, s) { return this.demand(t, e, s); }
        require(t, e, s) { return this.demand(t, e, s); }
        requiresArg(t) { return h("<array|string|object> [number]", [t], arguments.length), "string" == typeof t && v(this, Q, "f").narg[t] || this[kt](this.requiresArg.bind(this), "narg", t, NaN), this; }
        showCompletionScript(t, e) { return h("[string] [string]", [t, e], arguments.length), t = t || this.$0, v(this, Z, "f").log(v(this, U, "f").generateCompletionScript(t, e || v(this, W, "f") || "completion")), this; }
        showHelp(t) { if (h("[string|function]", [t], arguments.length), w(this, Y, !0, "f"), !v(this, dt, "f").hasCachedHelpMessage()) {
            if (!this.parsed) {
                const e = this[Rt](v(this, ot, "f"), void 0, void 0, 0, !0);
                if (f(e))
                    return e.then((() => { v(this, dt, "f").showHelp(t); })), this;
            }
            const e = v(this, q, "f").runDefaultBuilderOn(this);
            if (f(e))
                return e.then((() => { v(this, dt, "f").showHelp(t); })), this;
        } return v(this, dt, "f").showHelp(t), this; }
        scriptName(t) { return this.customScriptName = !0, this.$0 = t, this; }
        showHelpOnFail(t, e) { return h("[boolean|string] [string]", [t, e], arguments.length), v(this, dt, "f").showHelpOnFail(t, e), this; }
        showVersion(t) { return h("[string|function]", [t], arguments.length), v(this, dt, "f").showVersion(t), this; }
        skipValidation(t) { return h("<array|string>", [t], arguments.length), this[Mt]("skipValidation", t), this; }
        strict(t) { return h("[boolean]", [t], arguments.length), w(this, lt, !1 !== t, "f"), this; }
        strictCommands(t) { return h("[boolean]", [t], arguments.length), w(this, ct, !1 !== t, "f"), this; }
        strictOptions(t) { return h("[boolean]", [t], arguments.length), w(this, ft, !1 !== t, "f"), this; }
        string(t) { return h("<array|string>", [t], arguments.length), this[Mt]("string", t), this; }
        terminalWidth() { return h([], 0), v(this, ht, "f").process.stdColumns; }
        updateLocale(t) { return this.updateStrings(t); }
        updateStrings(t) { return h("<object>", [t], arguments.length), w(this, R, !1, "f"), v(this, ht, "f").y18n.updateLocale(t), this; }
        usage(t, s, n, i) { if (h("<string|null|undefined> [string|boolean] [function|object] [function]", [t, s, n, i], arguments.length), void 0 !== s) {
            if (d(t, null, v(this, ht, "f")), (t || "").match(/^\$0( |$)/))
                return this.command(t, s, n, i);
            throw new e(".usage() description must start with $0 if being used as alias for .command()");
        } return v(this, dt, "f").usage(t), this; }
        version(t, e, s) { const n = "version"; if (h("[boolean|string] [string] [string]", [t, e, s], arguments.length), v(this, ut, "f") && (this[yt](v(this, ut, "f")), v(this, dt, "f").version(void 0), w(this, ut, null, "f")), 0 === arguments.length)
            s = this[Ct](), t = n;
        else if (1 === arguments.length) {
            if (!1 === t)
                return this;
            s = t, t = n;
        }
        else
            2 === arguments.length && (s = e, e = void 0); return w(this, ut, "string" == typeof t ? t : n, "f"), e = e || v(this, dt, "f").deferY18nLookup("Show version number"), v(this, dt, "f").version(s || void 0), this.boolean(v(this, ut, "f")), this.describe(v(this, ut, "f"), e), this; }
        wrap(t) { return h("<number|null|undefined>", [t], arguments.length), v(this, dt, "f").wrap(t), this; }
        [(q = new WeakMap, H = new WeakMap, F = new WeakMap, U = new WeakMap, W = new WeakMap, L = new WeakMap, V = new WeakMap, R = new WeakMap, T = new WeakMap, G = new WeakMap, B = new WeakMap, K = new WeakMap, Y = new WeakMap, J = new WeakMap, Z = new WeakMap, X = new WeakMap, Q = new WeakMap, tt = new WeakMap, et = new WeakMap, st = new WeakMap, nt = new WeakMap, it = new WeakMap, rt = new WeakMap, ot = new WeakMap, at = new WeakMap, ht = new WeakMap, lt = new WeakMap, ct = new WeakMap, ft = new WeakMap, dt = new WeakMap, ut = new WeakMap, pt = new WeakMap, gt)](t) { if (!t._ || !t["--"])
            return t; t._.push.apply(t._, t["--"]); try {
            delete t["--"];
        }
        catch (t) { } return t; }
        [mt]() { return { log: (...t) => { this[Ut]() || console.log(...t), w(this, Y, !0, "f"), v(this, X, "f").length && w(this, X, v(this, X, "f") + "\n", "f"), w(this, X, v(this, X, "f") + t.join(" "), "f"); }, error: (...t) => { this[Ut]() || console.error(...t), w(this, Y, !0, "f"), v(this, X, "f").length && w(this, X, v(this, X, "f") + "\n", "f"), w(this, X, v(this, X, "f") + t.join(" "), "f"); } }; }
        [yt](t) { p(v(this, Q, "f")).forEach((e => { if ("configObjects" === e)
            return; const s = v(this, Q, "f")[e]; Array.isArray(s) ? s.includes(t) && s.splice(s.indexOf(t), 1) : "object" == typeof s && delete s[t]; })), delete v(this, dt, "f").getDescriptions()[t]; }
        [bt]() { v(this, G, "f").push({ options: v(this, Q, "f"), configObjects: v(this, Q, "f").configObjects.slice(0), exitProcess: v(this, T, "f"), groups: v(this, K, "f"), strict: v(this, lt, "f"), strictCommands: v(this, ct, "f"), strictOptions: v(this, ft, "f"), completionCommand: v(this, W, "f"), output: v(this, X, "f"), exitError: v(this, V, "f"), hasOutput: v(this, Y, "f"), parsed: this.parsed, parseFn: v(this, st, "f"), parseContext: v(this, nt, "f") }), v(this, dt, "f").freeze(), v(this, pt, "f").freeze(), v(this, q, "f").freeze(), v(this, B, "f").freeze(); }
        [vt]() { let t, e = ""; return t = /\b(node|iojs|electron)(\.exe)?$/.test(v(this, ht, "f").process.argv()[0]) ? v(this, ht, "f").process.argv().slice(1, 2) : v(this, ht, "f").process.argv().slice(0, 1), e = t.map((t => { const e = this[Lt](v(this, H, "f"), t); return t.match(/^(\/|([a-zA-Z]:)?\\)/) && e.length < t.length ? e : t; })).join(" ").trim(), v(this, ht, "f").getEnv("_") && v(this, ht, "f").getProcessArgvBin() === v(this, ht, "f").getEnv("_") && (e = v(this, ht, "f").getEnv("_").replace(`${v(this, ht, "f").path.dirname(v(this, ht, "f").process.execPath())}/`, "")), e; }
        [wt]() { return v(this, et, "f"); }
        [Ot]() { if (!v(this, R, "f"))
            return; const t = v(this, ht, "f").getEnv("LC_ALL") || v(this, ht, "f").getEnv("LC_MESSAGES") || v(this, ht, "f").getEnv("LANG") || v(this, ht, "f").getEnv("LANGUAGE") || "en_US"; this.locale(t.replace(/[.:].*/, "")); }
        [Ct]() { return this[_t]().version || "unknown"; }
        [jt](t) { const e = t["--"] ? t["--"] : t._; for (let t, s = 0; void 0 !== (t = e[s]); s++)
            v(this, ht, "f").Parser.looksLikeNumber(t) && Number.isSafeInteger(Math.floor(parseFloat(`${t}`))) && (e[s] = Number(t)); return t; }
        [_t](t) { const e = t || "*"; if (v(this, it, "f")[e])
            return v(this, it, "f")[e]; let s = {}; try {
            let e = t || v(this, ht, "f").mainFilename;
            !t && v(this, ht, "f").path.extname(e) && (e = v(this, ht, "f").path.dirname(e));
            const n = v(this, ht, "f").findUp(e, ((t, e) => e.includes("package.json") ? "package.json" : void 0));
            d(n, void 0, v(this, ht, "f")), s = JSON.parse(v(this, ht, "f").readFileSync(n, "utf8"));
        }
        catch (t) { } return v(this, it, "f")[e] = s || {}, v(this, it, "f")[e]; }
        [Mt](t, e) { (e = [].concat(e)).forEach((e => { e = this[St](e), v(this, Q, "f")[t].push(e); })); }
        [kt](t, e, s, n) { this[xt](t, e, s, n, ((t, e, s) => { v(this, Q, "f")[t][e] = s; })); }
        [Et](t, e, s, n) { this[xt](t, e, s, n, ((t, e, s) => { v(this, Q, "f")[t][e] = (v(this, Q, "f")[t][e] || []).concat(s); })); }
        [xt](t, e, s, n, i) { if (Array.isArray(s))
            s.forEach((e => { t(e, n); }));
        else if ((t => "object" == typeof t)(s))
            for (const e of p(s))
                t(e, s[e]);
        else
            i(e, this[St](s), n); }
        [St](t) { return "__proto__" === t ? "___proto___" : t; }
        [$t](t, e) { return this[kt](this[$t].bind(this), "key", t, e), this; }
        [At]() { var t, e, s, n, i, r, o, a, h, l, c, f; const u = v(this, G, "f").pop(); let p; d(u, void 0, v(this, ht, "f")), t = this, e = this, s = this, n = this, i = this, r = this, o = this, a = this, h = this, l = this, c = this, f = this, ({ options: { set value(e) { w(t, Q, e, "f"); } }.value, configObjects: p, exitProcess: { set value(t) { w(e, T, t, "f"); } }.value, groups: { set value(t) { w(s, K, t, "f"); } }.value, output: { set value(t) { w(n, X, t, "f"); } }.value, exitError: { set value(t) { w(i, V, t, "f"); } }.value, hasOutput: { set value(t) { w(r, Y, t, "f"); } }.value, parsed: this.parsed, strict: { set value(t) { w(o, lt, t, "f"); } }.value, strictCommands: { set value(t) { w(a, ct, t, "f"); } }.value, strictOptions: { set value(t) { w(h, ft, t, "f"); } }.value, completionCommand: { set value(t) { w(l, W, t, "f"); } }.value, parseFn: { set value(t) { w(c, st, t, "f"); } }.value, parseContext: { set value(t) { w(f, nt, t, "f"); } }.value } = u), v(this, Q, "f").configObjects = p, v(this, dt, "f").unfreeze(), v(this, pt, "f").unfreeze(), v(this, q, "f").unfreeze(), v(this, B, "f").unfreeze(); }
        [Pt](t, e) { return j(e, (e => (t(e), e))); }
        getInternalMethods() { return { getCommandInstance: this[It].bind(this), getContext: this[Nt].bind(this), getHasOutput: this[Dt].bind(this), getLoggerInstance: this[zt].bind(this), getParseContext: this[qt].bind(this), getParserConfiguration: this[wt].bind(this), getUsageInstance: this[Ht].bind(this), getValidationInstance: this[Ft].bind(this), hasParseCallback: this[Ut].bind(this), postProcess: this[Wt].bind(this), reset: this[Vt].bind(this), runValidation: this[Tt].bind(this), runYargsParserAndExecuteCommands: this[Rt].bind(this), setHasOutput: this[Gt].bind(this) }; }
        [It]() { return v(this, q, "f"); }
        [Nt]() { return v(this, F, "f"); }
        [Dt]() { return v(this, Y, "f"); }
        [zt]() { return v(this, Z, "f"); }
        [qt]() { return v(this, nt, "f") || {}; }
        [Ht]() { return v(this, dt, "f"); }
        [Ft]() { return v(this, pt, "f"); }
        [Ut]() { return !!v(this, st, "f"); }
        [Wt](t, e, s, n) { if (s)
            return t; if (f(t))
            return t; e || (t = this[gt](t)); return (this[wt]()["parse-positional-numbers"] || void 0 === this[wt]()["parse-positional-numbers"]) && (t = this[jt](t)), n && (t = C(t, this, v(this, B, "f").getMiddleware(), !1)), t; }
        [Vt](t = {}) { w(this, Q, v(this, Q, "f") || {}, "f"); const e = {}; e.local = v(this, Q, "f").local || [], e.configObjects = v(this, Q, "f").configObjects || []; const s = {}; e.local.forEach((e => { s[e] = !0, (t[e] || []).forEach((t => { s[t] = !0; })); })), Object.assign(v(this, rt, "f"), Object.keys(v(this, K, "f")).reduce(((t, e) => { const n = v(this, K, "f")[e].filter((t => !(t in s))); return n.length > 0 && (t[e] = n), t; }), {})), w(this, K, {}, "f"); return ["array", "boolean", "string", "skipValidation", "count", "normalize", "number", "hiddenOptions"].forEach((t => { e[t] = (v(this, Q, "f")[t] || []).filter((t => !s[t])); })), ["narg", "key", "alias", "default", "defaultDescription", "config", "choices", "demandedOptions", "demandedCommands", "deprecatedOptions"].forEach((t => { e[t] = g(v(this, Q, "f")[t], (t => !s[t])); })), e.envPrefix = v(this, Q, "f").envPrefix, w(this, Q, e, "f"), w(this, dt, v(this, dt, "f") ? v(this, dt, "f").reset(s) : $(this, v(this, ht, "f")), "f"), w(this, pt, v(this, pt, "f") ? v(this, pt, "f").reset(s) : function (t, e, s) { const n = s.y18n.__, i = s.y18n.__n, r = { nonOptionCount: function (s) { const n = t.getDemandedCommands(), r = s._.length + (s["--"] ? s["--"].length : 0) - t.getInternalMethods().getContext().commands.length; n._ && (r < n._.min || r > n._.max) && (r < n._.min ? void 0 !== n._.minMsg ? e.fail(n._.minMsg ? n._.minMsg.replace(/\$0/g, r.toString()).replace(/\$1/, n._.min.toString()) : null) : e.fail(i("Not enough non-option arguments: got %s, need at least %s", "Not enough non-option arguments: got %s, need at least %s", r, r.toString(), n._.min.toString())) : r > n._.max && (void 0 !== n._.maxMsg ? e.fail(n._.maxMsg ? n._.maxMsg.replace(/\$0/g, r.toString()).replace(/\$1/, n._.max.toString()) : null) : e.fail(i("Too many non-option arguments: got %s, maximum of %s", "Too many non-option arguments: got %s, maximum of %s", r, r.toString(), n._.max.toString())))); }, positionalCount: function (t, s) { s < t && e.fail(i("Not enough non-option arguments: got %s, need at least %s", "Not enough non-option arguments: got %s, need at least %s", s, s + "", t + "")); }, requiredArguments: function (t, s) { let n = null; for (const e of Object.keys(s))
                Object.prototype.hasOwnProperty.call(t, e) && void 0 !== t[e] || (n = n || {}, n[e] = s[e]); if (n) {
                const t = [];
                for (const e of Object.keys(n)) {
                    const s = n[e];
                    s && t.indexOf(s) < 0 && t.push(s);
                }
                const s = t.length ? `\n${t.join("\n")}` : "";
                e.fail(i("Missing required argument: %s", "Missing required arguments: %s", Object.keys(n).length, Object.keys(n).join(", ") + s));
            } }, unknownArguments: function (s, n, o, a, h = !0) { var l; const c = t.getInternalMethods().getCommandInstance().getCommands(), f = [], d = t.getInternalMethods().getContext(); if (Object.keys(s).forEach((e => { z.includes(e) || Object.prototype.hasOwnProperty.call(o, e) || Object.prototype.hasOwnProperty.call(t.getInternalMethods().getParseContext(), e) || r.isValidAndSomeAliasIsNotNew(e, n) || f.push(e); })), h && (d.commands.length > 0 || c.length > 0 || a) && s._.slice(d.commands.length).forEach((t => { c.includes("" + t) || f.push("" + t); })), h) {
                const e = (null === (l = t.getDemandedCommands()._) || void 0 === l ? void 0 : l.max) || 0, n = d.commands.length + e;
                n < s._.length && s._.slice(n).forEach((t => { t = String(t), d.commands.includes(t) || f.includes(t) || f.push(t); }));
            } f.length && e.fail(i("Unknown argument: %s", "Unknown arguments: %s", f.length, f.join(", "))); }, unknownCommands: function (s) { const n = t.getInternalMethods().getCommandInstance().getCommands(), r = [], o = t.getInternalMethods().getContext(); return (o.commands.length > 0 || n.length > 0) && s._.slice(o.commands.length).forEach((t => { n.includes("" + t) || r.push("" + t); })), r.length > 0 && (e.fail(i("Unknown command: %s", "Unknown commands: %s", r.length, r.join(", "))), !0); }, isValidAndSomeAliasIsNotNew: function (e, s) { if (!Object.prototype.hasOwnProperty.call(s, e))
                return !1; const n = t.parsed.newAliases; return [e, ...s[e]].some((t => !Object.prototype.hasOwnProperty.call(n, t) || !n[e])); }, limitedChoices: function (s) { const i = t.getOptions(), r = {}; if (!Object.keys(i.choices).length)
                return; Object.keys(s).forEach((t => { -1 === z.indexOf(t) && Object.prototype.hasOwnProperty.call(i.choices, t) && [].concat(s[t]).forEach((e => { -1 === i.choices[t].indexOf(e) && void 0 !== e && (r[t] = (r[t] || []).concat(e)); })); })); const o = Object.keys(r); if (!o.length)
                return; let a = n("Invalid values:"); o.forEach((t => { a += `\n  ${n("Argument: %s, Given: %s, Choices: %s", t, e.stringifiedValues(r[t]), e.stringifiedValues(i.choices[t]))}`; })), e.fail(a); } }; let o = {}; function a(t, e) { const s = Number(e); return "number" == typeof (e = isNaN(s) ? e : s) ? e = t._.length >= e : e.match(/^--no-.+/) ? (e = e.match(/^--no-(.+)/)[1], e = !Object.prototype.hasOwnProperty.call(t, e)) : e = Object.prototype.hasOwnProperty.call(t, e), e; } r.implies = function (e, n) { h("<string|object> [array|number|string]", [e, n], arguments.length), "object" == typeof e ? Object.keys(e).forEach((t => { r.implies(t, e[t]); })) : (t.global(e), o[e] || (o[e] = []), Array.isArray(n) ? n.forEach((t => r.implies(e, t))) : (d(n, void 0, s), o[e].push(n))); }, r.getImplied = function () { return o; }, r.implications = function (t) { const s = []; if (Object.keys(o).forEach((e => { const n = e; (o[e] || []).forEach((e => { let i = n; const r = e; i = a(t, i), e = a(t, e), i && !e && s.push(` ${n} -> ${r}`); })); })), s.length) {
            let t = `${n("Implications failed:")}\n`;
            s.forEach((e => { t += e; })), e.fail(t);
        } }; let l = {}; r.conflicts = function (e, s) { h("<string|object> [array|string]", [e, s], arguments.length), "object" == typeof e ? Object.keys(e).forEach((t => { r.conflicts(t, e[t]); })) : (t.global(e), l[e] || (l[e] = []), Array.isArray(s) ? s.forEach((t => r.conflicts(e, t))) : l[e].push(s)); }, r.getConflicting = () => l, r.conflicting = function (t) { Object.keys(t).forEach((s => { l[s] && l[s].forEach((i => { i && void 0 !== t[s] && void 0 !== t[i] && e.fail(n("Arguments %s and %s are mutually exclusive", s, i)); })); })); }, r.recommendCommands = function (t, s) { s = s.sort(((t, e) => e.length - t.length)); let i = null, r = 1 / 0; for (let e, n = 0; void 0 !== (e = s[n]); n++) {
            const s = D(t, e);
            s <= 3 && s < r && (r = s, i = e);
        } i && e.fail(n("Did you mean %s?", i)); }, r.reset = function (t) { return o = g(o, (e => !t[e])), l = g(l, (e => !t[e])), r; }; const c = []; return r.freeze = function () { c.push({ implied: o, conflicting: l }); }, r.unfreeze = function () { const t = c.pop(); d(t, void 0, s), ({ implied: o, conflicting: l } = t); }, r; }(this, v(this, dt, "f"), v(this, ht, "f")), "f"), w(this, q, v(this, q, "f") ? v(this, q, "f").reset() : function (t, e, s, n) { return new M(t, e, s, n); }(v(this, dt, "f"), v(this, pt, "f"), v(this, B, "f"), v(this, ht, "f")), "f"), v(this, U, "f") || w(this, U, function (t, e, s, n) { return new N(t, e, s, n); }(this, v(this, dt, "f"), v(this, q, "f"), v(this, ht, "f")), "f"), v(this, B, "f").reset(), w(this, W, null, "f"), w(this, X, "", "f"), w(this, V, null, "f"), w(this, Y, !1, "f"), this.parsed = !1, this; }
        [Lt](t, e) { return v(this, ht, "f").path.relative(t, e); }
        [Rt](t, s, n, i = 0, r = !1) { let o = !!n || r; t = t || v(this, ot, "f"), v(this, Q, "f").__ = v(this, ht, "f").y18n.__, v(this, Q, "f").configuration = this[wt](); const a = !!v(this, Q, "f").configuration["populate--"], h = Object.assign({}, v(this, Q, "f").configuration, { "populate--": !0 }), l = v(this, ht, "f").Parser.detailed(t, Object.assign({}, v(this, Q, "f"), { configuration: { "parse-positional-numbers": !1, ...h } })), c = Object.assign(l.argv, v(this, nt, "f")); let d; const u = l.aliases; let p = !1, g = !1; Object.keys(c).forEach((t => { t === v(this, J, "f") && c[t] ? p = !0 : t === v(this, ut, "f") && c[t] && (g = !0); })), c.$0 = this.$0, this.parsed = l, 0 === i && v(this, dt, "f").clearCachedHelpMessage(); try {
            if (this[Ot](), s)
                return this[Wt](c, a, !!n, !1);
            if (v(this, J, "f")) {
                [v(this, J, "f")].concat(u[v(this, J, "f")] || []).filter((t => t.length > 1)).includes("" + c._[c._.length - 1]) && (c._.pop(), p = !0);
            }
            const h = v(this, q, "f").getCommands(), m = v(this, U, "f").completionKey in c, y = p || m || r;
            if (c._.length) {
                if (h.length) {
                    let t;
                    for (let e, s = i || 0; void 0 !== c._[s]; s++) {
                        if (e = String(c._[s]), h.includes(e) && e !== v(this, W, "f")) {
                            const t = v(this, q, "f").runCommand(e, this, l, s + 1, r, p || g || r);
                            return this[Wt](t, a, !!n, !1);
                        }
                        if (!t && e !== v(this, W, "f")) {
                            t = e;
                            break;
                        }
                    }
                    !v(this, q, "f").hasDefaultCommand() && v(this, at, "f") && t && !y && v(this, pt, "f").recommendCommands(t, h);
                }
                v(this, W, "f") && c._.includes(v(this, W, "f")) && !m && (v(this, T, "f") && x(!0), this.showCompletionScript(), this.exit(0));
            }
            if (v(this, q, "f").hasDefaultCommand() && !y) {
                const t = v(this, q, "f").runCommand(null, this, l, 0, r, p || g || r);
                return this[Wt](t, a, !!n, !1);
            }
            if (m) {
                v(this, T, "f") && x(!0);
                const s = (t = [].concat(t)).slice(t.indexOf(`--${v(this, U, "f").completionKey}`) + 1);
                return v(this, U, "f").getCompletion(s, ((t, s) => { if (t)
                    throw new e(t.message); (s || []).forEach((t => { v(this, Z, "f").log(t); })), this.exit(0); })), this[Wt](c, !a, !!n, !1);
            }
            if (v(this, Y, "f") || (p ? (v(this, T, "f") && x(!0), o = !0, this.showHelp("log"), this.exit(0)) : g && (v(this, T, "f") && x(!0), o = !0, v(this, dt, "f").showVersion("log"), this.exit(0))), !o && v(this, Q, "f").skipValidation.length > 0 && (o = Object.keys(c).some((t => v(this, Q, "f").skipValidation.indexOf(t) >= 0 && !0 === c[t]))), !o) {
                if (l.error)
                    throw new e(l.error.message);
                if (!m) {
                    const t = this[Tt](u, {}, l.error);
                    n || (d = C(c, this, v(this, B, "f").getMiddleware(), !0)), d = this[Pt](t, null != d ? d : c), f(d) && !n && (d = d.then((() => C(c, this, v(this, B, "f").getMiddleware(), !1))));
                }
            }
        }
        catch (t) {
            if (!(t instanceof e))
                throw t;
            v(this, dt, "f").fail(t.message, t);
        } return this[Wt](null != d ? d : c, a, !!n, !0); }
        [Tt](t, s, n, i) { const r = { ...this.getDemandedOptions() }; return o => { if (n)
            throw new e(n.message); v(this, pt, "f").nonOptionCount(o), v(this, pt, "f").requiredArguments(o, r); let a = !1; v(this, ct, "f") && (a = v(this, pt, "f").unknownCommands(o)), v(this, lt, "f") && !a ? v(this, pt, "f").unknownArguments(o, t, s, !!i) : v(this, ft, "f") && v(this, pt, "f").unknownArguments(o, t, {}, !1, !1), v(this, pt, "f").limitedChoices(o), v(this, pt, "f").implications(o), v(this, pt, "f").conflicting(o); }; }
        [Gt]() { w(this, Y, !0, "f"); }
    }
    var Kt, Yt;
    const { readFileSync: Jt } = require("fs"), { inspect: Zt } = require("util"), { resolve: Xt } = require("path"), Qt = require(3) /* y18n */, te = require(4) /* yargs-parser */;
    var ee, se = { assert: { notStrictEqual: t.notStrictEqual, strictEqual: t.strictEqual }, cliui: require(5) /* cliui */, findUp: require(17) /* escalade/sync */, getEnv: t => process.env[t], getCallerFile: require(18) /* get-caller-file */, getProcessArgvBin: y, inspect: Zt, mainFilename: null !== (Yt = null === (Kt = null === require || void 0 === require ? void 0 : require.main) || void 0 === Kt ? void 0 : Kt.filename) && void 0 !== Yt ? Yt : process.cwd(), Parser: te, path: require("path"), process: { argv: () => process.argv, cwd: process.cwd, execPath: () => process.execPath, exit: t => { process.exit(t); }, nextTick: process.nextTick, stdColumns: void 0 !== process.stdout.columns ? process.stdout.columns : null }, readFileSync: Jt, require: require, requireDirectory: require(19) /* require-directory */, stringWidth: require(6) /* string-width */, y18n: Qt({ directory: Xt(__dirname, "../locales"), updateFiles: !1 }) };
    const ne = (null === (ee = null === process || void 0 === process ? void 0 : process.env) || void 0 === ee ? void 0 : ee.YARGS_MIN_NODE_VERSION) ? Number(process.env.YARGS_MIN_NODE_VERSION) : 12;
    if (process && process.version) {
        if (Number(process.version.match(/v([^.]+)/)[1]) < ne)
            throw Error(`yargs supports a minimum Node.js version of ${ne}. Read our version support policy: https://github.com/yargs/yargs#supported-nodejs-versions`);
    }
    const ie = require(4) /* yargs-parser */;
    var re, oe = { applyExtends: i, cjsPlatformShim: se, Yargs: (re = se, (t = [], e = re.process.cwd(), s) => { const n = new Bt(t, e, s, re); return Object.defineProperty(n, "argv", { get: () => n.parse(), enumerable: !0 }), n.help(), n.version(), n; }), argsert: h, isPromise: f, objFilter: g, parseCommand: o, Parser: ie, processArgv: b, YError: e };
    module.exports = oe;
},
/* y18n/build/index.cjs */ function _(require, module, exports, __esModule, __esExport) {
    var fs = require('fs');
    var util = require('util');
    var path = require('path');
    let shim;
    class Y18N {
        constructor(opts) {
            // configurable options.
            opts = opts || {};
            this.directory = opts.directory || './locales';
            this.updateFiles = typeof opts.updateFiles === 'boolean' ? opts.updateFiles : true;
            this.locale = opts.locale || 'en';
            this.fallbackToLanguage = typeof opts.fallbackToLanguage === 'boolean' ? opts.fallbackToLanguage : true;
            // internal stuff.
            this.cache = Object.create(null);
            this.writeQueue = [];
        }
        __(...args) {
            if (typeof arguments[0] !== 'string') {
                return this._taggedLiteral(arguments[0], ...arguments);
            }
            const str = args.shift();
            let cb = function () { }; // start with noop.
            if (typeof args[args.length - 1] === 'function')
                cb = args.pop();
            cb = cb || function () { }; // noop.
            if (!this.cache[this.locale])
                this._readLocaleFile();
            // we've observed a new string, update the language file.
            if (!this.cache[this.locale][str] && this.updateFiles) {
                this.cache[this.locale][str] = str;
                // include the current directory and locale,
                // since these values could change before the
                // write is performed.
                this._enqueueWrite({
                    directory: this.directory,
                    locale: this.locale,
                    cb
                });
            }
            else {
                cb();
            }
            return shim.format.apply(shim.format, [this.cache[this.locale][str] || str].concat(args));
        }
        __n() {
            const args = Array.prototype.slice.call(arguments);
            const singular = args.shift();
            const plural = args.shift();
            const quantity = args.shift();
            let cb = function () { }; // start with noop.
            if (typeof args[args.length - 1] === 'function')
                cb = args.pop();
            if (!this.cache[this.locale])
                this._readLocaleFile();
            let str = quantity === 1 ? singular : plural;
            if (this.cache[this.locale][singular]) {
                const entry = this.cache[this.locale][singular];
                str = entry[quantity === 1 ? 'one' : 'other'];
            }
            // we've observed a new string, update the language file.
            if (!this.cache[this.locale][singular] && this.updateFiles) {
                this.cache[this.locale][singular] = {
                    one: singular,
                    other: plural
                };
                // include the current directory and locale,
                // since these values could change before the
                // write is performed.
                this._enqueueWrite({
                    directory: this.directory,
                    locale: this.locale,
                    cb
                });
            }
            else {
                cb();
            }
            // if a %d placeholder is provided, add quantity
            // to the arguments expanded by util.format.
            const values = [str];
            if (~str.indexOf('%d'))
                values.push(quantity);
            return shim.format.apply(shim.format, values.concat(args));
        }
        setLocale(locale) {
            this.locale = locale;
        }
        getLocale() {
            return this.locale;
        }
        updateLocale(obj) {
            if (!this.cache[this.locale])
                this._readLocaleFile();
            for (const key in obj) {
                if (Object.prototype.hasOwnProperty.call(obj, key)) {
                    this.cache[this.locale][key] = obj[key];
                }
            }
        }
        _taggedLiteral(parts, ...args) {
            let str = '';
            parts.forEach(function (part, i) {
                const arg = args[i + 1];
                str += part;
                if (typeof arg !== 'undefined') {
                    str += '%s';
                }
            });
            return this.__.apply(this, [str].concat([].slice.call(args, 1)));
        }
        _enqueueWrite(work) {
            this.writeQueue.push(work);
            if (this.writeQueue.length === 1)
                this._processWriteQueue();
        }
        _processWriteQueue() {
            const _this = this;
            const work = this.writeQueue[0];
            // destructure the enqueued work.
            const directory = work.directory;
            const locale = work.locale;
            const cb = work.cb;
            const languageFile = this._resolveLocaleFile(directory, locale);
            const serializedLocale = JSON.stringify(this.cache[locale], null, 2);
            shim.fs.writeFile(languageFile, serializedLocale, 'utf-8', function (err) {
                _this.writeQueue.shift();
                if (_this.writeQueue.length > 0)
                    _this._processWriteQueue();
                cb(err);
            });
        }
        _readLocaleFile() {
            let localeLookup = {};
            const languageFile = this._resolveLocaleFile(this.directory, this.locale);
            try {
                // When using a bundler such as webpack, readFileSync may not be defined:
                if (shim.fs.readFileSync) {
                    localeLookup = JSON.parse(shim.fs.readFileSync(languageFile, 'utf-8'));
                }
            }
            catch (err) {
                if (err instanceof SyntaxError) {
                    err.message = 'syntax error in ' + languageFile;
                }
                if (err.code === 'ENOENT')
                    localeLookup = {};
                else
                    throw err;
            }
            this.cache[this.locale] = localeLookup;
        }
        _resolveLocaleFile(directory, locale) {
            let file = shim.resolve(directory, './', locale + '.json');
            if (this.fallbackToLanguage && !this._fileExistsSync(file) && ~locale.lastIndexOf('_')) {
                // attempt fallback to language only
                const languageFile = shim.resolve(directory, './', locale.split('_')[0] + '.json');
                if (this._fileExistsSync(languageFile))
                    file = languageFile;
            }
            return file;
        }
        _fileExistsSync(file) {
            return shim.exists(file);
        }
    }
    function y18n$1(opts, _shim) {
        shim = _shim;
        const y18n = new Y18N(opts);
        return {
            __: y18n.__.bind(y18n),
            __n: y18n.__n.bind(y18n),
            setLocale: y18n.setLocale.bind(y18n),
            getLocale: y18n.getLocale.bind(y18n),
            updateLocale: y18n.updateLocale.bind(y18n),
            locale: y18n.locale
        };
    }
    var nodePlatformShim = {
        fs: {
            readFileSync: fs.readFileSync,
            writeFile: fs.writeFile
        },
        format: util.format,
        resolve: path.resolve,
        exists: (file) => {
            try {
                return fs.statSync(file).isFile();
            }
            catch (err) {
                return false;
            }
        }
    };
    const y18n = (opts) => {
        return y18n$1(opts, nodePlatformShim);
    };
    module.exports = y18n;
},
/* yargs-parser/build/index.cjs */ function _(require, module, exports, __esModule, __esExport) {
    var util = require('util');
    var fs = require('fs');
    var path = require('path');
    function camelCase(str) {
        const isCamelCase = str !== str.toLowerCase() && str !== str.toUpperCase();
        if (!isCamelCase) {
            str = str.toLocaleLowerCase();
        }
        if (str.indexOf('-') === -1 && str.indexOf('_') === -1) {
            return str;
        }
        else {
            let camelcase = '';
            let nextChrUpper = false;
            const leadingHyphens = str.match(/^-+/);
            for (let i = leadingHyphens ? leadingHyphens[0].length : 0; i < str.length; i++) {
                let chr = str.charAt(i);
                if (nextChrUpper) {
                    nextChrUpper = false;
                    chr = chr.toLocaleUpperCase();
                }
                if (i !== 0 && (chr === '-' || chr === '_')) {
                    nextChrUpper = true;
                }
                else if (chr !== '-' && chr !== '_') {
                    camelcase += chr;
                }
            }
            return camelcase;
        }
    }
    function decamelize(str, joinString) {
        const lowercase = str.toLocaleLowerCase();
        joinString = joinString || '-';
        let notCamelcase = '';
        for (let i = 0; i < str.length; i++) {
            const chrLower = lowercase.charAt(i);
            const chrString = str.charAt(i);
            if (chrLower !== chrString && i > 0) {
                notCamelcase += `${joinString}${lowercase.charAt(i)}`;
            }
            else {
                notCamelcase += chrString;
            }
        }
        return notCamelcase;
    }
    function looksLikeNumber(x) {
        if (x === null || x === undefined)
            return false;
        if (typeof x === 'number')
            return true;
        if (/^0x[0-9a-f]+$/i.test(x))
            return true;
        if (x.length > 1 && x[0] === '0')
            return false;
        return /^[-]?(?:\d+(?:\.\d*)?|\.\d+)(e[-+]?\d+)?$/.test(x);
    }
    function tokenizeArgString(argString) {
        if (Array.isArray(argString)) {
            return argString.map(e => typeof e !== 'string' ? e + '' : e);
        }
        argString = argString.trim();
        let i = 0;
        let prevC = null;
        let c = null;
        let opening = null;
        const args = [];
        for (let ii = 0; ii < argString.length; ii++) {
            prevC = c;
            c = argString.charAt(ii);
            if (c === ' ' && !opening) {
                if (!(prevC === ' ')) {
                    i++;
                }
                continue;
            }
            if (c === opening) {
                opening = null;
            }
            else if ((c === "'" || c === '"') && !opening) {
                opening = c;
            }
            if (!args[i])
                args[i] = '';
            args[i] += c;
        }
        return args;
    }
    let mixin;
    class YargsParser {
        constructor(_mixin) {
            mixin = _mixin;
        }
        parse(argsInput, options) {
            const opts = Object.assign({
                alias: undefined,
                array: undefined,
                boolean: undefined,
                config: undefined,
                configObjects: undefined,
                configuration: undefined,
                coerce: undefined,
                count: undefined,
                default: undefined,
                envPrefix: undefined,
                narg: undefined,
                normalize: undefined,
                string: undefined,
                number: undefined,
                __: undefined,
                key: undefined
            }, options);
            const args = tokenizeArgString(argsInput);
            const aliases = combineAliases(Object.assign(Object.create(null), opts.alias));
            const configuration = Object.assign({
                'boolean-negation': true,
                'camel-case-expansion': true,
                'combine-arrays': false,
                'dot-notation': true,
                'duplicate-arguments-array': true,
                'flatten-duplicate-arrays': true,
                'greedy-arrays': true,
                'halt-at-non-option': false,
                'nargs-eats-options': false,
                'negation-prefix': 'no-',
                'parse-numbers': true,
                'parse-positional-numbers': true,
                'populate--': false,
                'set-placeholder-key': false,
                'short-option-groups': true,
                'strip-aliased': false,
                'strip-dashed': false,
                'unknown-options-as-args': false
            }, opts.configuration);
            const defaults = Object.assign(Object.create(null), opts.default);
            const configObjects = opts.configObjects || [];
            const envPrefix = opts.envPrefix;
            const notFlagsOption = configuration['populate--'];
            const notFlagsArgv = notFlagsOption ? '--' : '_';
            const newAliases = Object.create(null);
            const defaulted = Object.create(null);
            const __ = opts.__ || mixin.format;
            const flags = {
                aliases: Object.create(null),
                arrays: Object.create(null),
                bools: Object.create(null),
                strings: Object.create(null),
                numbers: Object.create(null),
                counts: Object.create(null),
                normalize: Object.create(null),
                configs: Object.create(null),
                nargs: Object.create(null),
                coercions: Object.create(null),
                keys: []
            };
            const negative = /^-([0-9]+(\.[0-9]+)?|\.[0-9]+)$/;
            const negatedBoolean = new RegExp('^--' + configuration['negation-prefix'] + '(.+)');
            [].concat(opts.array || []).filter(Boolean).forEach(function (opt) {
                const key = typeof opt === 'object' ? opt.key : opt;
                const assignment = Object.keys(opt).map(function (key) {
                    const arrayFlagKeys = {
                        boolean: 'bools',
                        string: 'strings',
                        number: 'numbers'
                    };
                    return arrayFlagKeys[key];
                }).filter(Boolean).pop();
                if (assignment) {
                    flags[assignment][key] = true;
                }
                flags.arrays[key] = true;
                flags.keys.push(key);
            });
            [].concat(opts.boolean || []).filter(Boolean).forEach(function (key) {
                flags.bools[key] = true;
                flags.keys.push(key);
            });
            [].concat(opts.string || []).filter(Boolean).forEach(function (key) {
                flags.strings[key] = true;
                flags.keys.push(key);
            });
            [].concat(opts.number || []).filter(Boolean).forEach(function (key) {
                flags.numbers[key] = true;
                flags.keys.push(key);
            });
            [].concat(opts.count || []).filter(Boolean).forEach(function (key) {
                flags.counts[key] = true;
                flags.keys.push(key);
            });
            [].concat(opts.normalize || []).filter(Boolean).forEach(function (key) {
                flags.normalize[key] = true;
                flags.keys.push(key);
            });
            if (typeof opts.narg === 'object') {
                Object.entries(opts.narg).forEach(([key, value]) => {
                    if (typeof value === 'number') {
                        flags.nargs[key] = value;
                        flags.keys.push(key);
                    }
                });
            }
            if (typeof opts.coerce === 'object') {
                Object.entries(opts.coerce).forEach(([key, value]) => {
                    if (typeof value === 'function') {
                        flags.coercions[key] = value;
                        flags.keys.push(key);
                    }
                });
            }
            if (typeof opts.config !== 'undefined') {
                if (Array.isArray(opts.config) || typeof opts.config === 'string') {
                    [].concat(opts.config).filter(Boolean).forEach(function (key) {
                        flags.configs[key] = true;
                    });
                }
                else if (typeof opts.config === 'object') {
                    Object.entries(opts.config).forEach(([key, value]) => {
                        if (typeof value === 'boolean' || typeof value === 'function') {
                            flags.configs[key] = value;
                        }
                    });
                }
            }
            extendAliases(opts.key, aliases, opts.default, flags.arrays);
            Object.keys(defaults).forEach(function (key) {
                (flags.aliases[key] || []).forEach(function (alias) {
                    defaults[alias] = defaults[key];
                });
            });
            let error = null;
            checkConfiguration();
            let notFlags = [];
            const argv = Object.assign(Object.create(null), { _: [] });
            const argvReturn = {};
            for (let i = 0; i < args.length; i++) {
                const arg = args[i];
                let broken;
                let key;
                let letters;
                let m;
                let next;
                let value;
                if (arg !== '--' && isUnknownOptionAsArg(arg)) {
                    pushPositional(arg);
                }
                else if (arg.match(/---+(=|$)/)) {
                    pushPositional(arg);
                    continue;
                }
                else if (arg.match(/^--.+=/) || (!configuration['short-option-groups'] && arg.match(/^-.+=/))) {
                    m = arg.match(/^--?([^=]+)=([\s\S]*)$/);
                    if (m !== null && Array.isArray(m) && m.length >= 3) {
                        if (checkAllAliases(m[1], flags.arrays)) {
                            i = eatArray(i, m[1], args, m[2]);
                        }
                        else if (checkAllAliases(m[1], flags.nargs) !== false) {
                            i = eatNargs(i, m[1], args, m[2]);
                        }
                        else {
                            setArg(m[1], m[2]);
                        }
                    }
                }
                else if (arg.match(negatedBoolean) && configuration['boolean-negation']) {
                    m = arg.match(negatedBoolean);
                    if (m !== null && Array.isArray(m) && m.length >= 2) {
                        key = m[1];
                        setArg(key, checkAllAliases(key, flags.arrays) ? [false] : false);
                    }
                }
                else if (arg.match(/^--.+/) || (!configuration['short-option-groups'] && arg.match(/^-[^-]+/))) {
                    m = arg.match(/^--?(.+)/);
                    if (m !== null && Array.isArray(m) && m.length >= 2) {
                        key = m[1];
                        if (checkAllAliases(key, flags.arrays)) {
                            i = eatArray(i, key, args);
                        }
                        else if (checkAllAliases(key, flags.nargs) !== false) {
                            i = eatNargs(i, key, args);
                        }
                        else {
                            next = args[i + 1];
                            if (next !== undefined && (!next.match(/^-/) ||
                                next.match(negative)) &&
                                !checkAllAliases(key, flags.bools) &&
                                !checkAllAliases(key, flags.counts)) {
                                setArg(key, next);
                                i++;
                            }
                            else if (/^(true|false)$/.test(next)) {
                                setArg(key, next);
                                i++;
                            }
                            else {
                                setArg(key, defaultValue(key));
                            }
                        }
                    }
                }
                else if (arg.match(/^-.\..+=/)) {
                    m = arg.match(/^-([^=]+)=([\s\S]*)$/);
                    if (m !== null && Array.isArray(m) && m.length >= 3) {
                        setArg(m[1], m[2]);
                    }
                }
                else if (arg.match(/^-.\..+/) && !arg.match(negative)) {
                    next = args[i + 1];
                    m = arg.match(/^-(.\..+)/);
                    if (m !== null && Array.isArray(m) && m.length >= 2) {
                        key = m[1];
                        if (next !== undefined && !next.match(/^-/) &&
                            !checkAllAliases(key, flags.bools) &&
                            !checkAllAliases(key, flags.counts)) {
                            setArg(key, next);
                            i++;
                        }
                        else {
                            setArg(key, defaultValue(key));
                        }
                    }
                }
                else if (arg.match(/^-[^-]+/) && !arg.match(negative)) {
                    letters = arg.slice(1, -1).split('');
                    broken = false;
                    for (let j = 0; j < letters.length; j++) {
                        next = arg.slice(j + 2);
                        if (letters[j + 1] && letters[j + 1] === '=') {
                            value = arg.slice(j + 3);
                            key = letters[j];
                            if (checkAllAliases(key, flags.arrays)) {
                                i = eatArray(i, key, args, value);
                            }
                            else if (checkAllAliases(key, flags.nargs) !== false) {
                                i = eatNargs(i, key, args, value);
                            }
                            else {
                                setArg(key, value);
                            }
                            broken = true;
                            break;
                        }
                        if (next === '-') {
                            setArg(letters[j], next);
                            continue;
                        }
                        if (/[A-Za-z]/.test(letters[j]) &&
                            /^-?\d+(\.\d*)?(e-?\d+)?$/.test(next) &&
                            checkAllAliases(next, flags.bools) === false) {
                            setArg(letters[j], next);
                            broken = true;
                            break;
                        }
                        if (letters[j + 1] && letters[j + 1].match(/\W/)) {
                            setArg(letters[j], next);
                            broken = true;
                            break;
                        }
                        else {
                            setArg(letters[j], defaultValue(letters[j]));
                        }
                    }
                    key = arg.slice(-1)[0];
                    if (!broken && key !== '-') {
                        if (checkAllAliases(key, flags.arrays)) {
                            i = eatArray(i, key, args);
                        }
                        else if (checkAllAliases(key, flags.nargs) !== false) {
                            i = eatNargs(i, key, args);
                        }
                        else {
                            next = args[i + 1];
                            if (next !== undefined && (!/^(-|--)[^-]/.test(next) ||
                                next.match(negative)) &&
                                !checkAllAliases(key, flags.bools) &&
                                !checkAllAliases(key, flags.counts)) {
                                setArg(key, next);
                                i++;
                            }
                            else if (/^(true|false)$/.test(next)) {
                                setArg(key, next);
                                i++;
                            }
                            else {
                                setArg(key, defaultValue(key));
                            }
                        }
                    }
                }
                else if (arg.match(/^-[0-9]$/) &&
                    arg.match(negative) &&
                    checkAllAliases(arg.slice(1), flags.bools)) {
                    key = arg.slice(1);
                    setArg(key, defaultValue(key));
                }
                else if (arg === '--') {
                    notFlags = args.slice(i + 1);
                    break;
                }
                else if (configuration['halt-at-non-option']) {
                    notFlags = args.slice(i);
                    break;
                }
                else {
                    pushPositional(arg);
                }
            }
            applyEnvVars(argv, true);
            applyEnvVars(argv, false);
            setConfig(argv);
            setConfigObjects();
            applyDefaultsAndAliases(argv, flags.aliases, defaults, true);
            applyCoercions(argv);
            if (configuration['set-placeholder-key'])
                setPlaceholderKeys(argv);
            Object.keys(flags.counts).forEach(function (key) {
                if (!hasKey(argv, key.split('.')))
                    setArg(key, 0);
            });
            if (notFlagsOption && notFlags.length)
                argv[notFlagsArgv] = [];
            notFlags.forEach(function (key) {
                argv[notFlagsArgv].push(key);
            });
            if (configuration['camel-case-expansion'] && configuration['strip-dashed']) {
                Object.keys(argv).filter(key => key !== '--' && key.includes('-')).forEach(key => {
                    delete argv[key];
                });
            }
            if (configuration['strip-aliased']) {
                [].concat(...Object.keys(aliases).map(k => aliases[k])).forEach(alias => {
                    if (configuration['camel-case-expansion'] && alias.includes('-')) {
                        delete argv[alias.split('.').map(prop => camelCase(prop)).join('.')];
                    }
                    delete argv[alias];
                });
            }
            function pushPositional(arg) {
                const maybeCoercedNumber = maybeCoerceNumber('_', arg);
                if (typeof maybeCoercedNumber === 'string' || typeof maybeCoercedNumber === 'number') {
                    argv._.push(maybeCoercedNumber);
                }
            }
            function eatNargs(i, key, args, argAfterEqualSign) {
                let ii;
                let toEat = checkAllAliases(key, flags.nargs);
                toEat = typeof toEat !== 'number' || isNaN(toEat) ? 1 : toEat;
                if (toEat === 0) {
                    if (!isUndefined(argAfterEqualSign)) {
                        error = Error(__('Argument unexpected for: %s', key));
                    }
                    setArg(key, defaultValue(key));
                    return i;
                }
                let available = isUndefined(argAfterEqualSign) ? 0 : 1;
                if (configuration['nargs-eats-options']) {
                    if (args.length - (i + 1) + available < toEat) {
                        error = Error(__('Not enough arguments following: %s', key));
                    }
                    available = toEat;
                }
                else {
                    for (ii = i + 1; ii < args.length; ii++) {
                        if (!args[ii].match(/^-[^0-9]/) || args[ii].match(negative) || isUnknownOptionAsArg(args[ii]))
                            available++;
                        else
                            break;
                    }
                    if (available < toEat)
                        error = Error(__('Not enough arguments following: %s', key));
                }
                let consumed = Math.min(available, toEat);
                if (!isUndefined(argAfterEqualSign) && consumed > 0) {
                    setArg(key, argAfterEqualSign);
                    consumed--;
                }
                for (ii = i + 1; ii < (consumed + i + 1); ii++) {
                    setArg(key, args[ii]);
                }
                return (i + consumed);
            }
            function eatArray(i, key, args, argAfterEqualSign) {
                let argsToSet = [];
                let next = argAfterEqualSign || args[i + 1];
                const nargsCount = checkAllAliases(key, flags.nargs);
                if (checkAllAliases(key, flags.bools) && !(/^(true|false)$/.test(next))) {
                    argsToSet.push(true);
                }
                else if (isUndefined(next) ||
                    (isUndefined(argAfterEqualSign) && /^-/.test(next) && !negative.test(next) && !isUnknownOptionAsArg(next))) {
                    if (defaults[key] !== undefined) {
                        const defVal = defaults[key];
                        argsToSet = Array.isArray(defVal) ? defVal : [defVal];
                    }
                }
                else {
                    if (!isUndefined(argAfterEqualSign)) {
                        argsToSet.push(processValue(key, argAfterEqualSign));
                    }
                    for (let ii = i + 1; ii < args.length; ii++) {
                        if ((!configuration['greedy-arrays'] && argsToSet.length > 0) ||
                            (nargsCount && typeof nargsCount === 'number' && argsToSet.length >= nargsCount))
                            break;
                        next = args[ii];
                        if (/^-/.test(next) && !negative.test(next) && !isUnknownOptionAsArg(next))
                            break;
                        i = ii;
                        argsToSet.push(processValue(key, next));
                    }
                }
                if (typeof nargsCount === 'number' && ((nargsCount && argsToSet.length < nargsCount) ||
                    (isNaN(nargsCount) && argsToSet.length === 0))) {
                    error = Error(__('Not enough arguments following: %s', key));
                }
                setArg(key, argsToSet);
                return i;
            }
            function setArg(key, val) {
                if (/-/.test(key) && configuration['camel-case-expansion']) {
                    const alias = key.split('.').map(function (prop) {
                        return camelCase(prop);
                    }).join('.');
                    addNewAlias(key, alias);
                }
                const value = processValue(key, val);
                const splitKey = key.split('.');
                setKey(argv, splitKey, value);
                if (flags.aliases[key]) {
                    flags.aliases[key].forEach(function (x) {
                        const keyProperties = x.split('.');
                        setKey(argv, keyProperties, value);
                    });
                }
                if (splitKey.length > 1 && configuration['dot-notation']) {
                    (flags.aliases[splitKey[0]] || []).forEach(function (x) {
                        let keyProperties = x.split('.');
                        const a = [].concat(splitKey);
                        a.shift();
                        keyProperties = keyProperties.concat(a);
                        if (!(flags.aliases[key] || []).includes(keyProperties.join('.'))) {
                            setKey(argv, keyProperties, value);
                        }
                    });
                }
                if (checkAllAliases(key, flags.normalize) && !checkAllAliases(key, flags.arrays)) {
                    const keys = [key].concat(flags.aliases[key] || []);
                    keys.forEach(function (key) {
                        Object.defineProperty(argvReturn, key, {
                            enumerable: true,
                            get() {
                                return val;
                            },
                            set(value) {
                                val = typeof value === 'string' ? mixin.normalize(value) : value;
                            }
                        });
                    });
                }
            }
            function addNewAlias(key, alias) {
                if (!(flags.aliases[key] && flags.aliases[key].length)) {
                    flags.aliases[key] = [alias];
                    newAliases[alias] = true;
                }
                if (!(flags.aliases[alias] && flags.aliases[alias].length)) {
                    addNewAlias(alias, key);
                }
            }
            function processValue(key, val) {
                if (typeof val === 'string' &&
                    (val[0] === "'" || val[0] === '"') &&
                    val[val.length - 1] === val[0]) {
                    val = val.substring(1, val.length - 1);
                }
                if (checkAllAliases(key, flags.bools) || checkAllAliases(key, flags.counts)) {
                    if (typeof val === 'string')
                        val = val === 'true';
                }
                let value = Array.isArray(val)
                    ? val.map(function (v) { return maybeCoerceNumber(key, v); })
                    : maybeCoerceNumber(key, val);
                if (checkAllAliases(key, flags.counts) && (isUndefined(value) || typeof value === 'boolean')) {
                    value = increment();
                }
                if (checkAllAliases(key, flags.normalize) && checkAllAliases(key, flags.arrays)) {
                    if (Array.isArray(val))
                        value = val.map((val) => { return mixin.normalize(val); });
                    else
                        value = mixin.normalize(val);
                }
                return value;
            }
            function maybeCoerceNumber(key, value) {
                if (!configuration['parse-positional-numbers'] && key === '_')
                    return value;
                if (!checkAllAliases(key, flags.strings) && !checkAllAliases(key, flags.bools) && !Array.isArray(value)) {
                    const shouldCoerceNumber = looksLikeNumber(value) && configuration['parse-numbers'] && (Number.isSafeInteger(Math.floor(parseFloat(`${value}`))));
                    if (shouldCoerceNumber || (!isUndefined(value) && checkAllAliases(key, flags.numbers))) {
                        value = Number(value);
                    }
                }
                return value;
            }
            function setConfig(argv) {
                const configLookup = Object.create(null);
                applyDefaultsAndAliases(configLookup, flags.aliases, defaults);
                Object.keys(flags.configs).forEach(function (configKey) {
                    const configPath = argv[configKey] || configLookup[configKey];
                    if (configPath) {
                        try {
                            let config = null;
                            const resolvedConfigPath = mixin.resolve(mixin.cwd(), configPath);
                            const resolveConfig = flags.configs[configKey];
                            if (typeof resolveConfig === 'function') {
                                try {
                                    config = resolveConfig(resolvedConfigPath);
                                }
                                catch (e) {
                                    config = e;
                                }
                                if (config instanceof Error) {
                                    error = config;
                                    return;
                                }
                            }
                            else {
                                config = mixin.require(resolvedConfigPath);
                            }
                            setConfigObject(config);
                        }
                        catch (ex) {
                            if (ex.name === 'PermissionDenied')
                                error = ex;
                            else if (argv[configKey])
                                error = Error(__('Invalid JSON config file: %s', configPath));
                        }
                    }
                });
            }
            function setConfigObject(config, prev) {
                Object.keys(config).forEach(function (key) {
                    const value = config[key];
                    const fullKey = prev ? prev + '.' + key : key;
                    if (typeof value === 'object' && value !== null && !Array.isArray(value) && configuration['dot-notation']) {
                        setConfigObject(value, fullKey);
                    }
                    else {
                        if (!hasKey(argv, fullKey.split('.')) || (checkAllAliases(fullKey, flags.arrays) && configuration['combine-arrays'])) {
                            setArg(fullKey, value);
                        }
                    }
                });
            }
            function setConfigObjects() {
                if (typeof configObjects !== 'undefined') {
                    configObjects.forEach(function (configObject) {
                        setConfigObject(configObject);
                    });
                }
            }
            function applyEnvVars(argv, configOnly) {
                if (typeof envPrefix === 'undefined')
                    return;
                const prefix = typeof envPrefix === 'string' ? envPrefix : '';
                const env = mixin.env();
                Object.keys(env).forEach(function (envVar) {
                    if (prefix === '' || envVar.lastIndexOf(prefix, 0) === 0) {
                        const keys = envVar.split('__').map(function (key, i) {
                            if (i === 0) {
                                key = key.substring(prefix.length);
                            }
                            return camelCase(key);
                        });
                        if (((configOnly && flags.configs[keys.join('.')]) || !configOnly) && !hasKey(argv, keys)) {
                            setArg(keys.join('.'), env[envVar]);
                        }
                    }
                });
            }
            function applyCoercions(argv) {
                let coerce;
                const applied = new Set();
                Object.keys(argv).forEach(function (key) {
                    if (!applied.has(key)) {
                        coerce = checkAllAliases(key, flags.coercions);
                        if (typeof coerce === 'function') {
                            try {
                                const value = maybeCoerceNumber(key, coerce(argv[key]));
                                ([].concat(flags.aliases[key] || [], key)).forEach(ali => {
                                    applied.add(ali);
                                    argv[ali] = value;
                                });
                            }
                            catch (err) {
                                error = err;
                            }
                        }
                    }
                });
            }
            function setPlaceholderKeys(argv) {
                flags.keys.forEach((key) => {
                    if (~key.indexOf('.'))
                        return;
                    if (typeof argv[key] === 'undefined')
                        argv[key] = undefined;
                });
                return argv;
            }
            function applyDefaultsAndAliases(obj, aliases, defaults, canLog = false) {
                Object.keys(defaults).forEach(function (key) {
                    if (!hasKey(obj, key.split('.'))) {
                        setKey(obj, key.split('.'), defaults[key]);
                        if (canLog)
                            defaulted[key] = true;
                        (aliases[key] || []).forEach(function (x) {
                            if (hasKey(obj, x.split('.')))
                                return;
                            setKey(obj, x.split('.'), defaults[key]);
                        });
                    }
                });
            }
            function hasKey(obj, keys) {
                let o = obj;
                if (!configuration['dot-notation'])
                    keys = [keys.join('.')];
                keys.slice(0, -1).forEach(function (key) {
                    o = (o[key] || {});
                });
                const key = keys[keys.length - 1];
                if (typeof o !== 'object')
                    return false;
                else
                    return key in o;
            }
            function setKey(obj, keys, value) {
                let o = obj;
                if (!configuration['dot-notation'])
                    keys = [keys.join('.')];
                keys.slice(0, -1).forEach(function (key) {
                    key = sanitizeKey(key);
                    if (typeof o === 'object' && o[key] === undefined) {
                        o[key] = {};
                    }
                    if (typeof o[key] !== 'object' || Array.isArray(o[key])) {
                        if (Array.isArray(o[key])) {
                            o[key].push({});
                        }
                        else {
                            o[key] = [o[key], {}];
                        }
                        o = o[key][o[key].length - 1];
                    }
                    else {
                        o = o[key];
                    }
                });
                const key = sanitizeKey(keys[keys.length - 1]);
                const isTypeArray = checkAllAliases(keys.join('.'), flags.arrays);
                const isValueArray = Array.isArray(value);
                let duplicate = configuration['duplicate-arguments-array'];
                if (!duplicate && checkAllAliases(key, flags.nargs)) {
                    duplicate = true;
                    if ((!isUndefined(o[key]) && flags.nargs[key] === 1) || (Array.isArray(o[key]) && o[key].length === flags.nargs[key])) {
                        o[key] = undefined;
                    }
                }
                if (value === increment()) {
                    o[key] = increment(o[key]);
                }
                else if (Array.isArray(o[key])) {
                    if (duplicate && isTypeArray && isValueArray) {
                        o[key] = configuration['flatten-duplicate-arrays'] ? o[key].concat(value) : (Array.isArray(o[key][0]) ? o[key] : [o[key]]).concat([value]);
                    }
                    else if (!duplicate && Boolean(isTypeArray) === Boolean(isValueArray)) {
                        o[key] = value;
                    }
                    else {
                        o[key] = o[key].concat([value]);
                    }
                }
                else if (o[key] === undefined && isTypeArray) {
                    o[key] = isValueArray ? value : [value];
                }
                else if (duplicate && !(o[key] === undefined ||
                    checkAllAliases(key, flags.counts) ||
                    checkAllAliases(key, flags.bools))) {
                    o[key] = [o[key], value];
                }
                else {
                    o[key] = value;
                }
            }
            function extendAliases(...args) {
                args.forEach(function (obj) {
                    Object.keys(obj || {}).forEach(function (key) {
                        if (flags.aliases[key])
                            return;
                        flags.aliases[key] = [].concat(aliases[key] || []);
                        flags.aliases[key].concat(key).forEach(function (x) {
                            if (/-/.test(x) && configuration['camel-case-expansion']) {
                                const c = camelCase(x);
                                if (c !== key && flags.aliases[key].indexOf(c) === -1) {
                                    flags.aliases[key].push(c);
                                    newAliases[c] = true;
                                }
                            }
                        });
                        flags.aliases[key].concat(key).forEach(function (x) {
                            if (x.length > 1 && /[A-Z]/.test(x) && configuration['camel-case-expansion']) {
                                const c = decamelize(x, '-');
                                if (c !== key && flags.aliases[key].indexOf(c) === -1) {
                                    flags.aliases[key].push(c);
                                    newAliases[c] = true;
                                }
                            }
                        });
                        flags.aliases[key].forEach(function (x) {
                            flags.aliases[x] = [key].concat(flags.aliases[key].filter(function (y) {
                                return x !== y;
                            }));
                        });
                    });
                });
            }
            function checkAllAliases(key, flag) {
                const toCheck = [].concat(flags.aliases[key] || [], key);
                const keys = Object.keys(flag);
                const setAlias = toCheck.find(key => keys.includes(key));
                return setAlias ? flag[setAlias] : false;
            }
            function hasAnyFlag(key) {
                const flagsKeys = Object.keys(flags);
                const toCheck = [].concat(flagsKeys.map(k => flags[k]));
                return toCheck.some(function (flag) {
                    return Array.isArray(flag) ? flag.includes(key) : flag[key];
                });
            }
            function hasFlagsMatching(arg, ...patterns) {
                const toCheck = [].concat(...patterns);
                return toCheck.some(function (pattern) {
                    const match = arg.match(pattern);
                    return match && hasAnyFlag(match[1]);
                });
            }
            function hasAllShortFlags(arg) {
                if (arg.match(negative) || !arg.match(/^-[^-]+/)) {
                    return false;
                }
                let hasAllFlags = true;
                let next;
                const letters = arg.slice(1).split('');
                for (let j = 0; j < letters.length; j++) {
                    next = arg.slice(j + 2);
                    if (!hasAnyFlag(letters[j])) {
                        hasAllFlags = false;
                        break;
                    }
                    if ((letters[j + 1] && letters[j + 1] === '=') ||
                        next === '-' ||
                        (/[A-Za-z]/.test(letters[j]) && /^-?\d+(\.\d*)?(e-?\d+)?$/.test(next)) ||
                        (letters[j + 1] && letters[j + 1].match(/\W/))) {
                        break;
                    }
                }
                return hasAllFlags;
            }
            function isUnknownOptionAsArg(arg) {
                return configuration['unknown-options-as-args'] && isUnknownOption(arg);
            }
            function isUnknownOption(arg) {
                if (arg.match(negative)) {
                    return false;
                }
                if (hasAllShortFlags(arg)) {
                    return false;
                }
                const flagWithEquals = /^-+([^=]+?)=[\s\S]*$/;
                const normalFlag = /^-+([^=]+?)$/;
                const flagEndingInHyphen = /^-+([^=]+?)-$/;
                const flagEndingInDigits = /^-+([^=]+?\d+)$/;
                const flagEndingInNonWordCharacters = /^-+([^=]+?)\W+.*$/;
                return !hasFlagsMatching(arg, flagWithEquals, negatedBoolean, normalFlag, flagEndingInHyphen, flagEndingInDigits, flagEndingInNonWordCharacters);
            }
            function defaultValue(key) {
                if (!checkAllAliases(key, flags.bools) &&
                    !checkAllAliases(key, flags.counts) &&
                    `${key}` in defaults) {
                    return defaults[key];
                }
                else {
                    return defaultForType(guessType(key));
                }
            }
            function defaultForType(type) {
                const def = {
                    boolean: true,
                    string: '',
                    number: undefined,
                    array: []
                };
                return def[type];
            }
            function guessType(key) {
                let type = 'boolean';
                if (checkAllAliases(key, flags.strings))
                    type = 'string';
                else if (checkAllAliases(key, flags.numbers))
                    type = 'number';
                else if (checkAllAliases(key, flags.bools))
                    type = 'boolean';
                else if (checkAllAliases(key, flags.arrays))
                    type = 'array';
                return type;
            }
            function isUndefined(num) {
                return num === undefined;
            }
            function checkConfiguration() {
                Object.keys(flags.counts).find(key => {
                    if (checkAllAliases(key, flags.arrays)) {
                        error = Error(__('Invalid configuration: %s, opts.count excludes opts.array.', key));
                        return true;
                    }
                    else if (checkAllAliases(key, flags.nargs)) {
                        error = Error(__('Invalid configuration: %s, opts.count excludes opts.narg.', key));
                        return true;
                    }
                    return false;
                });
            }
            return {
                aliases: Object.assign({}, flags.aliases),
                argv: Object.assign(argvReturn, argv),
                configuration: configuration,
                defaulted: Object.assign({}, defaulted),
                error: error,
                newAliases: Object.assign({}, newAliases)
            };
        }
    }
    function combineAliases(aliases) {
        const aliasArrays = [];
        const combined = Object.create(null);
        let change = true;
        Object.keys(aliases).forEach(function (key) {
            aliasArrays.push([].concat(aliases[key], key));
        });
        while (change) {
            change = false;
            for (let i = 0; i < aliasArrays.length; i++) {
                for (let ii = i + 1; ii < aliasArrays.length; ii++) {
                    const intersect = aliasArrays[i].filter(function (v) {
                        return aliasArrays[ii].indexOf(v) !== -1;
                    });
                    if (intersect.length) {
                        aliasArrays[i] = aliasArrays[i].concat(aliasArrays[ii]);
                        aliasArrays.splice(ii, 1);
                        change = true;
                        break;
                    }
                }
            }
        }
        aliasArrays.forEach(function (aliasArray) {
            aliasArray = aliasArray.filter(function (v, i, self) {
                return self.indexOf(v) === i;
            });
            const lastAlias = aliasArray.pop();
            if (lastAlias !== undefined && typeof lastAlias === 'string') {
                combined[lastAlias] = aliasArray;
            }
        });
        return combined;
    }
    function increment(orig) {
        return orig !== undefined ? orig + 1 : 1;
    }
    function sanitizeKey(key) {
        if (key === '__proto__')
            return '___proto___';
        return key;
    }
    const minNodeVersion = (process && process.env && process.env.YARGS_MIN_NODE_VERSION)
        ? Number(process.env.YARGS_MIN_NODE_VERSION)
        : 10;
    if (process && process.version) {
        const major = Number(process.version.match(/v([^.]+)/)[1]);
        if (major < minNodeVersion) {
            throw Error(`yargs parser supports a minimum Node.js version of ${minNodeVersion}. Read our version support policy: https://github.com/yargs/yargs-parser#supported-nodejs-versions`);
        }
    }
    const env = process ? process.env : {};
    const parser = new YargsParser({
        cwd: process.cwd,
        env: () => {
            return env;
        },
        format: util.format,
        normalize: path.normalize,
        resolve: path.resolve,
        require: (path) => {
            if (typeof require !== 'undefined') {
                return require(path);
            }
            else if (path.match(/\.json$/)) {
                return fs.readFileSync(path, 'utf8');
            }
            else {
                throw Error('only .json config files are supported in ESM');
            }
        }
    });
    const yargsParser = function Parser(args, opts) {
        const result = parser.parse(args.slice(), opts);
        return result.argv;
    };
    yargsParser.detailed = function (args, opts) {
        return parser.parse(args.slice(), opts);
    };
    yargsParser.camelCase = camelCase;
    yargsParser.decamelize = decamelize;
    yargsParser.looksLikeNumber = looksLikeNumber;
    module.exports = yargsParser;
},
/* cliui/build/index.cjs */ function _(require, module, exports, __esModule, __esExport) {
    const align = {
        right: alignRight,
        center: alignCenter
    };
    const top = 0;
    const right = 1;
    const bottom = 2;
    const left = 3;
    class UI {
        constructor(opts) {
            var _a;
            this.width = opts.width;
            this.wrap = (_a = opts.wrap) !== null && _a !== void 0 ? _a : true;
            this.rows = [];
        }
        span(...args) {
            const cols = this.div(...args);
            cols.span = true;
        }
        resetOutput() {
            this.rows = [];
        }
        div(...args) {
            if (args.length === 0) {
                this.div('');
            }
            if (this.wrap && this.shouldApplyLayoutDSL(...args) && typeof args[0] === 'string') {
                return this.applyLayoutDSL(args[0]);
            }
            const cols = args.map(arg => {
                if (typeof arg === 'string') {
                    return this.colFromString(arg);
                }
                return arg;
            });
            this.rows.push(cols);
            return cols;
        }
        shouldApplyLayoutDSL(...args) {
            return args.length === 1 && typeof args[0] === 'string' &&
                /[\t\n]/.test(args[0]);
        }
        applyLayoutDSL(str) {
            const rows = str.split('\n').map(row => row.split('\t'));
            let leftColumnWidth = 0;
            // simple heuristic for layout, make sure the
            // second column lines up along the left-hand.
            // don't allow the first column to take up more
            // than 50% of the screen.
            rows.forEach(columns => {
                if (columns.length > 1 && mixin.stringWidth(columns[0]) > leftColumnWidth) {
                    leftColumnWidth = Math.min(Math.floor(this.width * 0.5), mixin.stringWidth(columns[0]));
                }
            });
            // generate a table:
            //  replacing ' ' with padding calculations.
            //  using the algorithmically generated width.
            rows.forEach(columns => {
                this.div(...columns.map((r, i) => {
                    return {
                        text: r.trim(),
                        padding: this.measurePadding(r),
                        width: (i === 0 && columns.length > 1) ? leftColumnWidth : undefined
                    };
                }));
            });
            return this.rows[this.rows.length - 1];
        }
        colFromString(text) {
            return {
                text,
                padding: this.measurePadding(text)
            };
        }
        measurePadding(str) {
            // measure padding without ansi escape codes
            const noAnsi = mixin.stripAnsi(str);
            return [0, noAnsi.match(/\s*$/)[0].length, 0, noAnsi.match(/^\s*/)[0].length];
        }
        toString() {
            const lines = [];
            this.rows.forEach(row => {
                this.rowToString(row, lines);
            });
            // don't display any lines with the
            // hidden flag set.
            return lines
                .filter(line => !line.hidden)
                .map(line => line.text)
                .join('\n');
        }
        rowToString(row, lines) {
            this.rasterize(row).forEach((rrow, r) => {
                let str = '';
                rrow.forEach((col, c) => {
                    const { width } = row[c]; // the width with padding.
                    const wrapWidth = this.negatePadding(row[c]); // the width without padding.
                    let ts = col; // temporary string used during alignment/padding.
                    if (wrapWidth > mixin.stringWidth(col)) {
                        ts += ' '.repeat(wrapWidth - mixin.stringWidth(col));
                    }
                    // align the string within its column.
                    if (row[c].align && row[c].align !== 'left' && this.wrap) {
                        const fn = align[row[c].align];
                        ts = fn(ts, wrapWidth);
                        if (mixin.stringWidth(ts) < wrapWidth) {
                            ts += ' '.repeat((width || 0) - mixin.stringWidth(ts) - 1);
                        }
                    }
                    // apply border and padding to string.
                    const padding = row[c].padding || [0, 0, 0, 0];
                    if (padding[left]) {
                        str += ' '.repeat(padding[left]);
                    }
                    str += addBorder(row[c], ts, '| ');
                    str += ts;
                    str += addBorder(row[c], ts, ' |');
                    if (padding[right]) {
                        str += ' '.repeat(padding[right]);
                    }
                    // if prior row is span, try to render the
                    // current row on the prior line.
                    if (r === 0 && lines.length > 0) {
                        str = this.renderInline(str, lines[lines.length - 1]);
                    }
                });
                // remove trailing whitespace.
                lines.push({
                    text: str.replace(/ +$/, ''),
                    span: row.span
                });
            });
            return lines;
        }
        // if the full 'source' can render in
        // the target line, do so.
        renderInline(source, previousLine) {
            const match = source.match(/^ */);
            const leadingWhitespace = match ? match[0].length : 0;
            const target = previousLine.text;
            const targetTextWidth = mixin.stringWidth(target.trimRight());
            if (!previousLine.span) {
                return source;
            }
            // if we're not applying wrapping logic,
            // just always append to the span.
            if (!this.wrap) {
                previousLine.hidden = true;
                return target + source;
            }
            if (leadingWhitespace < targetTextWidth) {
                return source;
            }
            previousLine.hidden = true;
            return target.trimRight() + ' '.repeat(leadingWhitespace - targetTextWidth) + source.trimLeft();
        }
        rasterize(row) {
            const rrows = [];
            const widths = this.columnWidths(row);
            let wrapped;
            // word wrap all columns, and create
            // a data-structure that is easy to rasterize.
            row.forEach((col, c) => {
                // leave room for left and right padding.
                col.width = widths[c];
                if (this.wrap) {
                    wrapped = mixin.wrap(col.text, this.negatePadding(col), { hard: true }).split('\n');
                }
                else {
                    wrapped = col.text.split('\n');
                }
                if (col.border) {
                    wrapped.unshift('.' + '-'.repeat(this.negatePadding(col) + 2) + '.');
                    wrapped.push("'" + '-'.repeat(this.negatePadding(col) + 2) + "'");
                }
                // add top and bottom padding.
                if (col.padding) {
                    wrapped.unshift(...new Array(col.padding[top] || 0).fill(''));
                    wrapped.push(...new Array(col.padding[bottom] || 0).fill(''));
                }
                wrapped.forEach((str, r) => {
                    if (!rrows[r]) {
                        rrows.push([]);
                    }
                    const rrow = rrows[r];
                    for (let i = 0; i < c; i++) {
                        if (rrow[i] === undefined) {
                            rrow.push('');
                        }
                    }
                    rrow.push(str);
                });
            });
            return rrows;
        }
        negatePadding(col) {
            let wrapWidth = col.width || 0;
            if (col.padding) {
                wrapWidth -= (col.padding[left] || 0) + (col.padding[right] || 0);
            }
            if (col.border) {
                wrapWidth -= 4;
            }
            return wrapWidth;
        }
        columnWidths(row) {
            if (!this.wrap) {
                return row.map(col => {
                    return col.width || mixin.stringWidth(col.text);
                });
            }
            let unset = row.length;
            let remainingWidth = this.width;
            // column widths can be set in config.
            const widths = row.map(col => {
                if (col.width) {
                    unset--;
                    remainingWidth -= col.width;
                    return col.width;
                }
                return undefined;
            });
            // any unset widths should be calculated.
            const unsetWidth = unset ? Math.floor(remainingWidth / unset) : 0;
            return widths.map((w, i) => {
                if (w === undefined) {
                    return Math.max(unsetWidth, _minWidth(row[i]));
                }
                return w;
            });
        }
    }
    function addBorder(col, ts, style) {
        if (col.border) {
            if (/[.']-+[.']/.test(ts)) {
                return '';
            }
            if (ts.trim().length !== 0) {
                return style;
            }
            return '  ';
        }
        return '';
    }
    // calculates the minimum width of
    // a column, based on padding preferences.
    function _minWidth(col) {
        const padding = col.padding || [];
        const minWidth = 1 + (padding[left] || 0) + (padding[right] || 0);
        if (col.border) {
            return minWidth + 4;
        }
        return minWidth;
    }
    function getWindowWidth() {
        /* istanbul ignore next: depends on terminal */
        if (typeof process === 'object' && process.stdout && process.stdout.columns) {
            return process.stdout.columns;
        }
        return 80;
    }
    function alignRight(str, width) {
        str = str.trim();
        const strWidth = mixin.stringWidth(str);
        if (strWidth < width) {
            return ' '.repeat(width - strWidth) + str;
        }
        return str;
    }
    function alignCenter(str, width) {
        str = str.trim();
        const strWidth = mixin.stringWidth(str);
        /* istanbul ignore next */
        if (strWidth >= width) {
            return str;
        }
        return ' '.repeat((width - strWidth) >> 1) + str;
    }
    let mixin;
    function cliui(opts, _mixin) {
        mixin = _mixin;
        return new UI({
            width: (opts === null || opts === void 0 ? void 0 : opts.width) || getWindowWidth(),
            wrap: opts === null || opts === void 0 ? void 0 : opts.wrap
        });
    }
    // Bootstrap cliui with CommonJS dependencies:
    const stringWidth = require(6) /* string-width */;
    const stripAnsi = require(7) /* strip-ansi */;
    const wrap = require(11) /* wrap-ansi */;
    function ui(opts) {
        return cliui(opts, {
            stringWidth,
            stripAnsi,
            wrap
        });
    }
    module.exports = ui;
},
/* string-width/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const stripAnsi = require(7) /* strip-ansi */;
    const isFullwidthCodePoint = require(9) /* is-fullwidth-code-point */;
    const emojiRegex = require(10) /* emoji-regex */;
    const stringWidth = string => {
        if (typeof string !== 'string' || string.length === 0) {
            return 0;
        }
        string = stripAnsi(string);
        if (string.length === 0) {
            return 0;
        }
        string = string.replace(emojiRegex(), '  ');
        let width = 0;
        for (let i = 0; i < string.length; i++) {
            const code = string.codePointAt(i);
            // Ignore control characters
            if (code <= 0x1F || (code >= 0x7F && code <= 0x9F)) {
                continue;
            }
            // Ignore combining characters
            if (code >= 0x300 && code <= 0x36F) {
                continue;
            }
            // Surrogates
            if (code > 0xFFFF) {
                i++;
            }
            width += isFullwidthCodePoint(code) ? 2 : 1;
        }
        return width;
    };
    module.exports = stringWidth;
    // TODO: remove this in the next major version
    module.exports.default = stringWidth;
},
/* strip-ansi/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const ansiRegex = require(8) /* ansi-regex */;
    module.exports = string => typeof string === 'string' ? string.replace(ansiRegex(), '') : string;
},
/* ansi-regex/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = ({ onlyFirst = false } = {}) => {
        const pattern = [
            '[\\u001B\\u009B][[\\]()#;?]*(?:(?:(?:[a-zA-Z\\d]*(?:;[-a-zA-Z\\d\\/#&.:=?%@~_]*)*)?\\u0007)',
            '(?:(?:\\d{1,4}(?:;\\d{0,4})*)?[\\dA-PR-TZcf-ntqry=><~]))'
        ].join('|');
        return new RegExp(pattern, onlyFirst ? undefined : 'g');
    };
},
/* is-fullwidth-code-point/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const isFullwidthCodePoint = codePoint => {
        if (Number.isNaN(codePoint)) {
            return false;
        }
        // Code points are derived from:
        // http://www.unix.org/Public/UNIDATA/EastAsianWidth.txt
        if (codePoint >= 0x1100 && (codePoint <= 0x115F || // Hangul Jamo
            codePoint === 0x2329 || // LEFT-POINTING ANGLE BRACKET
            codePoint === 0x232A || // RIGHT-POINTING ANGLE BRACKET
            // CJK Radicals Supplement .. Enclosed CJK Letters and Months
            (0x2E80 <= codePoint && codePoint <= 0x3247 && codePoint !== 0x303F) ||
            // Enclosed CJK Letters and Months .. CJK Unified Ideographs Extension A
            (0x3250 <= codePoint && codePoint <= 0x4DBF) ||
            // CJK Unified Ideographs .. Yi Radicals
            (0x4E00 <= codePoint && codePoint <= 0xA4C6) ||
            // Hangul Jamo Extended-A
            (0xA960 <= codePoint && codePoint <= 0xA97C) ||
            // Hangul Syllables
            (0xAC00 <= codePoint && codePoint <= 0xD7A3) ||
            // CJK Compatibility Ideographs
            (0xF900 <= codePoint && codePoint <= 0xFAFF) ||
            // Vertical Forms
            (0xFE10 <= codePoint && codePoint <= 0xFE19) ||
            // CJK Compatibility Forms .. Small Form Variants
            (0xFE30 <= codePoint && codePoint <= 0xFE6B) ||
            // Halfwidth and Fullwidth Forms
            (0xFF01 <= codePoint && codePoint <= 0xFF60) ||
            (0xFFE0 <= codePoint && codePoint <= 0xFFE6) ||
            // Kana Supplement
            (0x1B000 <= codePoint && codePoint <= 0x1B001) ||
            // Enclosed Ideographic Supplement
            (0x1F200 <= codePoint && codePoint <= 0x1F251) ||
            // CJK Unified Ideographs Extension B .. Tertiary Ideographic Plane
            (0x20000 <= codePoint && codePoint <= 0x3FFFD))) {
            return true;
        }
        return false;
    };
    module.exports = isFullwidthCodePoint;
    module.exports.default = isFullwidthCodePoint;
},
/* emoji-regex/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function () {
        // https://mths.be/emoji
        return /\uD83C\uDFF4\uDB40\uDC67\uDB40\uDC62(?:\uDB40\uDC65\uDB40\uDC6E\uDB40\uDC67|\uDB40\uDC73\uDB40\uDC63\uDB40\uDC74|\uDB40\uDC77\uDB40\uDC6C\uDB40\uDC73)\uDB40\uDC7F|\uD83D\uDC68(?:\uD83C\uDFFC\u200D(?:\uD83E\uDD1D\u200D\uD83D\uDC68\uD83C\uDFFB|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\uD83C\uDFFF\u200D(?:\uD83E\uDD1D\u200D\uD83D\uDC68(?:\uD83C[\uDFFB-\uDFFE])|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\uD83C\uDFFE\u200D(?:\uD83E\uDD1D\u200D\uD83D\uDC68(?:\uD83C[\uDFFB-\uDFFD])|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\uD83C\uDFFD\u200D(?:\uD83E\uDD1D\u200D\uD83D\uDC68(?:\uD83C[\uDFFB\uDFFC])|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\u200D(?:\u2764\uFE0F\u200D(?:\uD83D\uDC8B\u200D)?\uD83D\uDC68|(?:\uD83D[\uDC68\uDC69])\u200D(?:\uD83D\uDC66\u200D\uD83D\uDC66|\uD83D\uDC67\u200D(?:\uD83D[\uDC66\uDC67]))|\uD83D\uDC66\u200D\uD83D\uDC66|\uD83D\uDC67\u200D(?:\uD83D[\uDC66\uDC67])|(?:\uD83D[\uDC68\uDC69])\u200D(?:\uD83D[\uDC66\uDC67])|[\u2695\u2696\u2708]\uFE0F|\uD83D[\uDC66\uDC67]|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|(?:\uD83C\uDFFB\u200D[\u2695\u2696\u2708]|\uD83C\uDFFF\u200D[\u2695\u2696\u2708]|\uD83C\uDFFE\u200D[\u2695\u2696\u2708]|\uD83C\uDFFD\u200D[\u2695\u2696\u2708]|\uD83C\uDFFC\u200D[\u2695\u2696\u2708])\uFE0F|\uD83C\uDFFB\u200D(?:\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\uD83C[\uDFFB-\uDFFF])|(?:\uD83E\uDDD1\uD83C\uDFFB\u200D\uD83E\uDD1D\u200D\uD83E\uDDD1|\uD83D\uDC69\uD83C\uDFFC\u200D\uD83E\uDD1D\u200D\uD83D\uDC69)\uD83C\uDFFB|\uD83E\uDDD1(?:\uD83C\uDFFF\u200D\uD83E\uDD1D\u200D\uD83E\uDDD1(?:\uD83C[\uDFFB-\uDFFF])|\u200D\uD83E\uDD1D\u200D\uD83E\uDDD1)|(?:\uD83E\uDDD1\uD83C\uDFFE\u200D\uD83E\uDD1D\u200D\uD83E\uDDD1|\uD83D\uDC69\uD83C\uDFFF\u200D\uD83E\uDD1D\u200D(?:\uD83D[\uDC68\uDC69]))(?:\uD83C[\uDFFB-\uDFFE])|(?:\uD83E\uDDD1\uD83C\uDFFC\u200D\uD83E\uDD1D\u200D\uD83E\uDDD1|\uD83D\uDC69\uD83C\uDFFD\u200D\uD83E\uDD1D\u200D\uD83D\uDC69)(?:\uD83C[\uDFFB\uDFFC])|\uD83D\uDC69(?:\uD83C\uDFFE\u200D(?:\uD83E\uDD1D\u200D\uD83D\uDC68(?:\uD83C[\uDFFB-\uDFFD\uDFFF])|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\uD83C\uDFFC\u200D(?:\uD83E\uDD1D\u200D\uD83D\uDC68(?:\uD83C[\uDFFB\uDFFD-\uDFFF])|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\uD83C\uDFFB\u200D(?:\uD83E\uDD1D\u200D\uD83D\uDC68(?:\uD83C[\uDFFC-\uDFFF])|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\uD83C\uDFFD\u200D(?:\uD83E\uDD1D\u200D\uD83D\uDC68(?:\uD83C[\uDFFB\uDFFC\uDFFE\uDFFF])|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\u200D(?:\u2764\uFE0F\u200D(?:\uD83D\uDC8B\u200D(?:\uD83D[\uDC68\uDC69])|\uD83D[\uDC68\uDC69])|\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD])|\uD83C\uDFFF\u200D(?:\uD83C[\uDF3E\uDF73\uDF93\uDFA4\uDFA8\uDFEB\uDFED]|\uD83D[\uDCBB\uDCBC\uDD27\uDD2C\uDE80\uDE92]|\uD83E[\uDDAF-\uDDB3\uDDBC\uDDBD]))|\uD83D\uDC69\u200D\uD83D\uDC69\u200D(?:\uD83D\uDC66\u200D\uD83D\uDC66|\uD83D\uDC67\u200D(?:\uD83D[\uDC66\uDC67]))|(?:\uD83E\uDDD1\uD83C\uDFFD\u200D\uD83E\uDD1D\u200D\uD83E\uDDD1|\uD83D\uDC69\uD83C\uDFFE\u200D\uD83E\uDD1D\u200D\uD83D\uDC69)(?:\uD83C[\uDFFB-\uDFFD])|\uD83D\uDC69\u200D\uD83D\uDC66\u200D\uD83D\uDC66|\uD83D\uDC69\u200D\uD83D\uDC69\u200D(?:\uD83D[\uDC66\uDC67])|(?:\uD83D\uDC41\uFE0F\u200D\uD83D\uDDE8|\uD83D\uDC69(?:\uD83C\uDFFF\u200D[\u2695\u2696\u2708]|\uD83C\uDFFE\u200D[\u2695\u2696\u2708]|\uD83C\uDFFC\u200D[\u2695\u2696\u2708]|\uD83C\uDFFB\u200D[\u2695\u2696\u2708]|\uD83C\uDFFD\u200D[\u2695\u2696\u2708]|\u200D[\u2695\u2696\u2708])|(?:(?:\u26F9|\uD83C[\uDFCB\uDFCC]|\uD83D\uDD75)\uFE0F|\uD83D\uDC6F|\uD83E[\uDD3C\uDDDE\uDDDF])\u200D[\u2640\u2642]|(?:\u26F9|\uD83C[\uDFCB\uDFCC]|\uD83D\uDD75)(?:\uD83C[\uDFFB-\uDFFF])\u200D[\u2640\u2642]|(?:\uD83C[\uDFC3\uDFC4\uDFCA]|\uD83D[\uDC6E\uDC71\uDC73\uDC77\uDC81\uDC82\uDC86\uDC87\uDE45-\uDE47\uDE4B\uDE4D\uDE4E\uDEA3\uDEB4-\uDEB6]|\uD83E[\uDD26\uDD37-\uDD39\uDD3D\uDD3E\uDDB8\uDDB9\uDDCD-\uDDCF\uDDD6-\uDDDD])(?:(?:\uD83C[\uDFFB-\uDFFF])\u200D[\u2640\u2642]|\u200D[\u2640\u2642])|\uD83C\uDFF4\u200D\u2620)\uFE0F|\uD83D\uDC69\u200D\uD83D\uDC67\u200D(?:\uD83D[\uDC66\uDC67])|\uD83C\uDFF3\uFE0F\u200D\uD83C\uDF08|\uD83D\uDC15\u200D\uD83E\uDDBA|\uD83D\uDC69\u200D\uD83D\uDC66|\uD83D\uDC69\u200D\uD83D\uDC67|\uD83C\uDDFD\uD83C\uDDF0|\uD83C\uDDF4\uD83C\uDDF2|\uD83C\uDDF6\uD83C\uDDE6|[#\*0-9]\uFE0F\u20E3|\uD83C\uDDE7(?:\uD83C[\uDDE6\uDDE7\uDDE9-\uDDEF\uDDF1-\uDDF4\uDDF6-\uDDF9\uDDFB\uDDFC\uDDFE\uDDFF])|\uD83C\uDDF9(?:\uD83C[\uDDE6\uDDE8\uDDE9\uDDEB-\uDDED\uDDEF-\uDDF4\uDDF7\uDDF9\uDDFB\uDDFC\uDDFF])|\uD83C\uDDEA(?:\uD83C[\uDDE6\uDDE8\uDDEA\uDDEC\uDDED\uDDF7-\uDDFA])|\uD83E\uDDD1(?:\uD83C[\uDFFB-\uDFFF])|\uD83C\uDDF7(?:\uD83C[\uDDEA\uDDF4\uDDF8\uDDFA\uDDFC])|\uD83D\uDC69(?:\uD83C[\uDFFB-\uDFFF])|\uD83C\uDDF2(?:\uD83C[\uDDE6\uDDE8-\uDDED\uDDF0-\uDDFF])|\uD83C\uDDE6(?:\uD83C[\uDDE8-\uDDEC\uDDEE\uDDF1\uDDF2\uDDF4\uDDF6-\uDDFA\uDDFC\uDDFD\uDDFF])|\uD83C\uDDF0(?:\uD83C[\uDDEA\uDDEC-\uDDEE\uDDF2\uDDF3\uDDF5\uDDF7\uDDFC\uDDFE\uDDFF])|\uD83C\uDDED(?:\uD83C[\uDDF0\uDDF2\uDDF3\uDDF7\uDDF9\uDDFA])|\uD83C\uDDE9(?:\uD83C[\uDDEA\uDDEC\uDDEF\uDDF0\uDDF2\uDDF4\uDDFF])|\uD83C\uDDFE(?:\uD83C[\uDDEA\uDDF9])|\uD83C\uDDEC(?:\uD83C[\uDDE6\uDDE7\uDDE9-\uDDEE\uDDF1-\uDDF3\uDDF5-\uDDFA\uDDFC\uDDFE])|\uD83C\uDDF8(?:\uD83C[\uDDE6-\uDDEA\uDDEC-\uDDF4\uDDF7-\uDDF9\uDDFB\uDDFD-\uDDFF])|\uD83C\uDDEB(?:\uD83C[\uDDEE-\uDDF0\uDDF2\uDDF4\uDDF7])|\uD83C\uDDF5(?:\uD83C[\uDDE6\uDDEA-\uDDED\uDDF0-\uDDF3\uDDF7-\uDDF9\uDDFC\uDDFE])|\uD83C\uDDFB(?:\uD83C[\uDDE6\uDDE8\uDDEA\uDDEC\uDDEE\uDDF3\uDDFA])|\uD83C\uDDF3(?:\uD83C[\uDDE6\uDDE8\uDDEA-\uDDEC\uDDEE\uDDF1\uDDF4\uDDF5\uDDF7\uDDFA\uDDFF])|\uD83C\uDDE8(?:\uD83C[\uDDE6\uDDE8\uDDE9\uDDEB-\uDDEE\uDDF0-\uDDF5\uDDF7\uDDFA-\uDDFF])|\uD83C\uDDF1(?:\uD83C[\uDDE6-\uDDE8\uDDEE\uDDF0\uDDF7-\uDDFB\uDDFE])|\uD83C\uDDFF(?:\uD83C[\uDDE6\uDDF2\uDDFC])|\uD83C\uDDFC(?:\uD83C[\uDDEB\uDDF8])|\uD83C\uDDFA(?:\uD83C[\uDDE6\uDDEC\uDDF2\uDDF3\uDDF8\uDDFE\uDDFF])|\uD83C\uDDEE(?:\uD83C[\uDDE8-\uDDEA\uDDF1-\uDDF4\uDDF6-\uDDF9])|\uD83C\uDDEF(?:\uD83C[\uDDEA\uDDF2\uDDF4\uDDF5])|(?:\uD83C[\uDFC3\uDFC4\uDFCA]|\uD83D[\uDC6E\uDC71\uDC73\uDC77\uDC81\uDC82\uDC86\uDC87\uDE45-\uDE47\uDE4B\uDE4D\uDE4E\uDEA3\uDEB4-\uDEB6]|\uD83E[\uDD26\uDD37-\uDD39\uDD3D\uDD3E\uDDB8\uDDB9\uDDCD-\uDDCF\uDDD6-\uDDDD])(?:\uD83C[\uDFFB-\uDFFF])|(?:\u26F9|\uD83C[\uDFCB\uDFCC]|\uD83D\uDD75)(?:\uD83C[\uDFFB-\uDFFF])|(?:[\u261D\u270A-\u270D]|\uD83C[\uDF85\uDFC2\uDFC7]|\uD83D[\uDC42\uDC43\uDC46-\uDC50\uDC66\uDC67\uDC6B-\uDC6D\uDC70\uDC72\uDC74-\uDC76\uDC78\uDC7C\uDC83\uDC85\uDCAA\uDD74\uDD7A\uDD90\uDD95\uDD96\uDE4C\uDE4F\uDEC0\uDECC]|\uD83E[\uDD0F\uDD18-\uDD1C\uDD1E\uDD1F\uDD30-\uDD36\uDDB5\uDDB6\uDDBB\uDDD2-\uDDD5])(?:\uD83C[\uDFFB-\uDFFF])|(?:[\u231A\u231B\u23E9-\u23EC\u23F0\u23F3\u25FD\u25FE\u2614\u2615\u2648-\u2653\u267F\u2693\u26A1\u26AA\u26AB\u26BD\u26BE\u26C4\u26C5\u26CE\u26D4\u26EA\u26F2\u26F3\u26F5\u26FA\u26FD\u2705\u270A\u270B\u2728\u274C\u274E\u2753-\u2755\u2757\u2795-\u2797\u27B0\u27BF\u2B1B\u2B1C\u2B50\u2B55]|\uD83C[\uDC04\uDCCF\uDD8E\uDD91-\uDD9A\uDDE6-\uDDFF\uDE01\uDE1A\uDE2F\uDE32-\uDE36\uDE38-\uDE3A\uDE50\uDE51\uDF00-\uDF20\uDF2D-\uDF35\uDF37-\uDF7C\uDF7E-\uDF93\uDFA0-\uDFCA\uDFCF-\uDFD3\uDFE0-\uDFF0\uDFF4\uDFF8-\uDFFF]|\uD83D[\uDC00-\uDC3E\uDC40\uDC42-\uDCFC\uDCFF-\uDD3D\uDD4B-\uDD4E\uDD50-\uDD67\uDD7A\uDD95\uDD96\uDDA4\uDDFB-\uDE4F\uDE80-\uDEC5\uDECC\uDED0-\uDED2\uDED5\uDEEB\uDEEC\uDEF4-\uDEFA\uDFE0-\uDFEB]|\uD83E[\uDD0D-\uDD3A\uDD3C-\uDD45\uDD47-\uDD71\uDD73-\uDD76\uDD7A-\uDDA2\uDDA5-\uDDAA\uDDAE-\uDDCA\uDDCD-\uDDFF\uDE70-\uDE73\uDE78-\uDE7A\uDE80-\uDE82\uDE90-\uDE95])|(?:[#\*0-9\xA9\xAE\u203C\u2049\u2122\u2139\u2194-\u2199\u21A9\u21AA\u231A\u231B\u2328\u23CF\u23E9-\u23F3\u23F8-\u23FA\u24C2\u25AA\u25AB\u25B6\u25C0\u25FB-\u25FE\u2600-\u2604\u260E\u2611\u2614\u2615\u2618\u261D\u2620\u2622\u2623\u2626\u262A\u262E\u262F\u2638-\u263A\u2640\u2642\u2648-\u2653\u265F\u2660\u2663\u2665\u2666\u2668\u267B\u267E\u267F\u2692-\u2697\u2699\u269B\u269C\u26A0\u26A1\u26AA\u26AB\u26B0\u26B1\u26BD\u26BE\u26C4\u26C5\u26C8\u26CE\u26CF\u26D1\u26D3\u26D4\u26E9\u26EA\u26F0-\u26F5\u26F7-\u26FA\u26FD\u2702\u2705\u2708-\u270D\u270F\u2712\u2714\u2716\u271D\u2721\u2728\u2733\u2734\u2744\u2747\u274C\u274E\u2753-\u2755\u2757\u2763\u2764\u2795-\u2797\u27A1\u27B0\u27BF\u2934\u2935\u2B05-\u2B07\u2B1B\u2B1C\u2B50\u2B55\u3030\u303D\u3297\u3299]|\uD83C[\uDC04\uDCCF\uDD70\uDD71\uDD7E\uDD7F\uDD8E\uDD91-\uDD9A\uDDE6-\uDDFF\uDE01\uDE02\uDE1A\uDE2F\uDE32-\uDE3A\uDE50\uDE51\uDF00-\uDF21\uDF24-\uDF93\uDF96\uDF97\uDF99-\uDF9B\uDF9E-\uDFF0\uDFF3-\uDFF5\uDFF7-\uDFFF]|\uD83D[\uDC00-\uDCFD\uDCFF-\uDD3D\uDD49-\uDD4E\uDD50-\uDD67\uDD6F\uDD70\uDD73-\uDD7A\uDD87\uDD8A-\uDD8D\uDD90\uDD95\uDD96\uDDA4\uDDA5\uDDA8\uDDB1\uDDB2\uDDBC\uDDC2-\uDDC4\uDDD1-\uDDD3\uDDDC-\uDDDE\uDDE1\uDDE3\uDDE8\uDDEF\uDDF3\uDDFA-\uDE4F\uDE80-\uDEC5\uDECB-\uDED2\uDED5\uDEE0-\uDEE5\uDEE9\uDEEB\uDEEC\uDEF0\uDEF3-\uDEFA\uDFE0-\uDFEB]|\uD83E[\uDD0D-\uDD3A\uDD3C-\uDD45\uDD47-\uDD71\uDD73-\uDD76\uDD7A-\uDDA2\uDDA5-\uDDAA\uDDAE-\uDDCA\uDDCD-\uDDFF\uDE70-\uDE73\uDE78-\uDE7A\uDE80-\uDE82\uDE90-\uDE95])\uFE0F|(?:[\u261D\u26F9\u270A-\u270D]|\uD83C[\uDF85\uDFC2-\uDFC4\uDFC7\uDFCA-\uDFCC]|\uD83D[\uDC42\uDC43\uDC46-\uDC50\uDC66-\uDC78\uDC7C\uDC81-\uDC83\uDC85-\uDC87\uDC8F\uDC91\uDCAA\uDD74\uDD75\uDD7A\uDD90\uDD95\uDD96\uDE45-\uDE47\uDE4B-\uDE4F\uDEA3\uDEB4-\uDEB6\uDEC0\uDECC]|\uD83E[\uDD0F\uDD18-\uDD1F\uDD26\uDD30-\uDD39\uDD3C-\uDD3E\uDDB5\uDDB6\uDDB8\uDDB9\uDDBB\uDDCD-\uDDCF\uDDD1-\uDDDD])/g;
    };
},
/* wrap-ansi/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const stringWidth = require(6) /* string-width */;
    const stripAnsi = require(7) /* strip-ansi */;
    const ansiStyles = require(12) /* ansi-styles */;
    const ESCAPES = new Set([
        '\u001B',
        '\u009B'
    ]);
    const END_CODE = 39;
    const ANSI_ESCAPE_BELL = '\u0007';
    const ANSI_CSI = '[';
    const ANSI_OSC = ']';
    const ANSI_SGR_TERMINATOR = 'm';
    const ANSI_ESCAPE_LINK = `${ANSI_OSC}8;;`;
    const wrapAnsi = code => `${ESCAPES.values().next().value}${ANSI_CSI}${code}${ANSI_SGR_TERMINATOR}`;
    const wrapAnsiHyperlink = uri => `${ESCAPES.values().next().value}${ANSI_ESCAPE_LINK}${uri}${ANSI_ESCAPE_BELL}`;
    // Calculate the length of words split on ' ', ignoring
    // the extra characters added by ansi escape codes
    const wordLengths = string => string.split(' ').map(character => stringWidth(character));
    // Wrap a long word across multiple rows
    // Ansi escape codes do not count towards length
    const wrapWord = (rows, word, columns) => {
        const characters = [...word];
        let isInsideEscape = false;
        let isInsideLinkEscape = false;
        let visible = stringWidth(stripAnsi(rows[rows.length - 1]));
        for (const [index, character] of characters.entries()) {
            const characterLength = stringWidth(character);
            if (visible + characterLength <= columns) {
                rows[rows.length - 1] += character;
            }
            else {
                rows.push(character);
                visible = 0;
            }
            if (ESCAPES.has(character)) {
                isInsideEscape = true;
                isInsideLinkEscape = characters.slice(index + 1).join('').startsWith(ANSI_ESCAPE_LINK);
            }
            if (isInsideEscape) {
                if (isInsideLinkEscape) {
                    if (character === ANSI_ESCAPE_BELL) {
                        isInsideEscape = false;
                        isInsideLinkEscape = false;
                    }
                }
                else if (character === ANSI_SGR_TERMINATOR) {
                    isInsideEscape = false;
                }
                continue;
            }
            visible += characterLength;
            if (visible === columns && index < characters.length - 1) {
                rows.push('');
                visible = 0;
            }
        }
        // It's possible that the last row we copy over is only
        // ansi escape characters, handle this edge-case
        if (!visible && rows[rows.length - 1].length > 0 && rows.length > 1) {
            rows[rows.length - 2] += rows.pop();
        }
    };
    // Trims spaces from a string ignoring invisible sequences
    const stringVisibleTrimSpacesRight = string => {
        const words = string.split(' ');
        let last = words.length;
        while (last > 0) {
            if (stringWidth(words[last - 1]) > 0) {
                break;
            }
            last--;
        }
        if (last === words.length) {
            return string;
        }
        return words.slice(0, last).join(' ') + words.slice(last).join('');
    };
    // The wrap-ansi module can be invoked in either 'hard' or 'soft' wrap mode
    //
    // 'hard' will never allow a string to take up more than columns characters
    //
    // 'soft' allows long words to expand past the column length
    const exec = (string, columns, options = {}) => {
        if (options.trim !== false && string.trim() === '') {
            return '';
        }
        let returnValue = '';
        let escapeCode;
        let escapeUrl;
        const lengths = wordLengths(string);
        let rows = [''];
        for (const [index, word] of string.split(' ').entries()) {
            if (options.trim !== false) {
                rows[rows.length - 1] = rows[rows.length - 1].trimStart();
            }
            let rowLength = stringWidth(rows[rows.length - 1]);
            if (index !== 0) {
                if (rowLength >= columns && (options.wordWrap === false || options.trim === false)) {
                    // If we start with a new word but the current row length equals the length of the columns, add a new row
                    rows.push('');
                    rowLength = 0;
                }
                if (rowLength > 0 || options.trim === false) {
                    rows[rows.length - 1] += ' ';
                    rowLength++;
                }
            }
            // In 'hard' wrap mode, the length of a line is never allowed to extend past 'columns'
            if (options.hard && lengths[index] > columns) {
                const remainingColumns = (columns - rowLength);
                const breaksStartingThisLine = 1 + Math.floor((lengths[index] - remainingColumns - 1) / columns);
                const breaksStartingNextLine = Math.floor((lengths[index] - 1) / columns);
                if (breaksStartingNextLine < breaksStartingThisLine) {
                    rows.push('');
                }
                wrapWord(rows, word, columns);
                continue;
            }
            if (rowLength + lengths[index] > columns && rowLength > 0 && lengths[index] > 0) {
                if (options.wordWrap === false && rowLength < columns) {
                    wrapWord(rows, word, columns);
                    continue;
                }
                rows.push('');
            }
            if (rowLength + lengths[index] > columns && options.wordWrap === false) {
                wrapWord(rows, word, columns);
                continue;
            }
            rows[rows.length - 1] += word;
        }
        if (options.trim !== false) {
            rows = rows.map(stringVisibleTrimSpacesRight);
        }
        const pre = [...rows.join('\n')];
        for (const [index, character] of pre.entries()) {
            returnValue += character;
            if (ESCAPES.has(character)) {
                const { groups } = new RegExp(`(?:\\${ANSI_CSI}(?<code>\\d+)m|\\${ANSI_ESCAPE_LINK}(?<uri>.*)${ANSI_ESCAPE_BELL})`).exec(pre.slice(index).join('')) || { groups: {} };
                if (groups.code !== undefined) {
                    const code = Number.parseFloat(groups.code);
                    escapeCode = code === END_CODE ? undefined : code;
                }
                else if (groups.uri !== undefined) {
                    escapeUrl = groups.uri.length === 0 ? undefined : groups.uri;
                }
            }
            const code = ansiStyles.codes.get(Number(escapeCode));
            if (pre[index + 1] === '\n') {
                if (escapeUrl) {
                    returnValue += wrapAnsiHyperlink('');
                }
                if (escapeCode && code) {
                    returnValue += wrapAnsi(code);
                }
            }
            else if (character === '\n') {
                if (escapeCode && code) {
                    returnValue += wrapAnsi(escapeCode);
                }
                if (escapeUrl) {
                    returnValue += wrapAnsiHyperlink(escapeUrl);
                }
            }
        }
        return returnValue;
    };
    // For each newline, invoke the method separately
    module.exports = (string, columns, options) => {
        return String(string)
            .normalize()
            .replace(/\r\n/g, '\n')
            .split('\n')
            .map(line => exec(line, columns, options))
            .join('\n');
    };
},
/* ansi-styles/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const wrapAnsi16 = (fn, offset) => (...args) => {
        const code = fn(...args);
        return `\u001B[${code + offset}m`;
    };
    const wrapAnsi256 = (fn, offset) => (...args) => {
        const code = fn(...args);
        return `\u001B[${38 + offset};5;${code}m`;
    };
    const wrapAnsi16m = (fn, offset) => (...args) => {
        const rgb = fn(...args);
        return `\u001B[${38 + offset};2;${rgb[0]};${rgb[1]};${rgb[2]}m`;
    };
    const ansi2ansi = n => n;
    const rgb2rgb = (r, g, b) => [r, g, b];
    const setLazyProperty = (object, property, get) => {
        Object.defineProperty(object, property, {
            get: () => {
                const value = get();
                Object.defineProperty(object, property, {
                    value,
                    enumerable: true,
                    configurable: true
                });
                return value;
            },
            enumerable: true,
            configurable: true
        });
    };
    /** @type {typeof import('color-convert')} */
    let colorConvert;
    const makeDynamicStyles = (wrap, targetSpace, identity, isBackground) => {
        if (colorConvert === undefined) {
            colorConvert = require(13) /* color-convert */;
        }
        const offset = isBackground ? 10 : 0;
        const styles = {};
        for (const [sourceSpace, suite] of Object.entries(colorConvert)) {
            const name = sourceSpace === 'ansi16' ? 'ansi' : sourceSpace;
            if (sourceSpace === targetSpace) {
                styles[name] = wrap(identity, offset);
            }
            else if (typeof suite === 'object') {
                styles[name] = wrap(suite[targetSpace], offset);
            }
        }
        return styles;
    };
    function assembleStyles() {
        const codes = new Map();
        const styles = {
            modifier: {
                reset: [0, 0],
                // 21 isn't widely supported and 22 does the same thing
                bold: [1, 22],
                dim: [2, 22],
                italic: [3, 23],
                underline: [4, 24],
                inverse: [7, 27],
                hidden: [8, 28],
                strikethrough: [9, 29]
            },
            color: {
                black: [30, 39],
                red: [31, 39],
                green: [32, 39],
                yellow: [33, 39],
                blue: [34, 39],
                magenta: [35, 39],
                cyan: [36, 39],
                white: [37, 39],
                // Bright color
                blackBright: [90, 39],
                redBright: [91, 39],
                greenBright: [92, 39],
                yellowBright: [93, 39],
                blueBright: [94, 39],
                magentaBright: [95, 39],
                cyanBright: [96, 39],
                whiteBright: [97, 39]
            },
            bgColor: {
                bgBlack: [40, 49],
                bgRed: [41, 49],
                bgGreen: [42, 49],
                bgYellow: [43, 49],
                bgBlue: [44, 49],
                bgMagenta: [45, 49],
                bgCyan: [46, 49],
                bgWhite: [47, 49],
                // Bright color
                bgBlackBright: [100, 49],
                bgRedBright: [101, 49],
                bgGreenBright: [102, 49],
                bgYellowBright: [103, 49],
                bgBlueBright: [104, 49],
                bgMagentaBright: [105, 49],
                bgCyanBright: [106, 49],
                bgWhiteBright: [107, 49]
            }
        };
        // Alias bright black as gray (and grey)
        styles.color.gray = styles.color.blackBright;
        styles.bgColor.bgGray = styles.bgColor.bgBlackBright;
        styles.color.grey = styles.color.blackBright;
        styles.bgColor.bgGrey = styles.bgColor.bgBlackBright;
        for (const [groupName, group] of Object.entries(styles)) {
            for (const [styleName, style] of Object.entries(group)) {
                styles[styleName] = {
                    open: `\u001B[${style[0]}m`,
                    close: `\u001B[${style[1]}m`
                };
                group[styleName] = styles[styleName];
                codes.set(style[0], style[1]);
            }
            Object.defineProperty(styles, groupName, {
                value: group,
                enumerable: false
            });
        }
        Object.defineProperty(styles, 'codes', {
            value: codes,
            enumerable: false
        });
        styles.color.close = '\u001B[39m';
        styles.bgColor.close = '\u001B[49m';
        setLazyProperty(styles.color, 'ansi', () => makeDynamicStyles(wrapAnsi16, 'ansi16', ansi2ansi, false));
        setLazyProperty(styles.color, 'ansi256', () => makeDynamicStyles(wrapAnsi256, 'ansi256', ansi2ansi, false));
        setLazyProperty(styles.color, 'ansi16m', () => makeDynamicStyles(wrapAnsi16m, 'rgb', rgb2rgb, false));
        setLazyProperty(styles.bgColor, 'ansi', () => makeDynamicStyles(wrapAnsi16, 'ansi16', ansi2ansi, true));
        setLazyProperty(styles.bgColor, 'ansi256', () => makeDynamicStyles(wrapAnsi256, 'ansi256', ansi2ansi, true));
        setLazyProperty(styles.bgColor, 'ansi16m', () => makeDynamicStyles(wrapAnsi16m, 'rgb', rgb2rgb, true));
        return styles;
    }
    // Make the export immutable
    Object.defineProperty(module, 'exports', {
        enumerable: true,
        get: assembleStyles
    });
},
/* color-convert/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const conversions = require(14) /* ./conversions */;
    const route = require(16) /* ./route */;
    const convert = {};
    const models = Object.keys(conversions);
    function wrapRaw(fn) {
        const wrappedFn = function (...args) {
            const arg0 = args[0];
            if (arg0 === undefined || arg0 === null) {
                return arg0;
            }
            if (arg0.length > 1) {
                args = arg0;
            }
            return fn(args);
        };
        // Preserve .conversion property if there is one
        if ('conversion' in fn) {
            wrappedFn.conversion = fn.conversion;
        }
        return wrappedFn;
    }
    function wrapRounded(fn) {
        const wrappedFn = function (...args) {
            const arg0 = args[0];
            if (arg0 === undefined || arg0 === null) {
                return arg0;
            }
            if (arg0.length > 1) {
                args = arg0;
            }
            const result = fn(args);
            // We're assuming the result is an array here.
            // see notice in conversions.js; don't use box types
            // in conversion functions.
            if (typeof result === 'object') {
                for (let len = result.length, i = 0; i < len; i++) {
                    result[i] = Math.round(result[i]);
                }
            }
            return result;
        };
        // Preserve .conversion property if there is one
        if ('conversion' in fn) {
            wrappedFn.conversion = fn.conversion;
        }
        return wrappedFn;
    }
    models.forEach(fromModel => {
        convert[fromModel] = {};
        Object.defineProperty(convert[fromModel], 'channels', { value: conversions[fromModel].channels });
        Object.defineProperty(convert[fromModel], 'labels', { value: conversions[fromModel].labels });
        const routes = route(fromModel);
        const routeModels = Object.keys(routes);
        routeModels.forEach(toModel => {
            const fn = routes[toModel];
            convert[fromModel][toModel] = wrapRounded(fn);
            convert[fromModel][toModel].raw = wrapRaw(fn);
        });
    });
    module.exports = convert;
},
/* color-convert/conversions.js */ function _(require, module, exports, __esModule, __esExport) {
    /* MIT license */
    /* eslint-disable no-mixed-operators */
    const cssKeywords = require(15) /* color-name */;
    // NOTE: conversions should only return primitive values (i.e. arrays, or
    //       values that give correct `typeof` results).
    //       do not use box values types (i.e. Number(), String(), etc.)
    const reverseKeywords = {};
    for (const key of Object.keys(cssKeywords)) {
        reverseKeywords[cssKeywords[key]] = key;
    }
    const convert = {
        rgb: { channels: 3, labels: 'rgb' },
        hsl: { channels: 3, labels: 'hsl' },
        hsv: { channels: 3, labels: 'hsv' },
        hwb: { channels: 3, labels: 'hwb' },
        cmyk: { channels: 4, labels: 'cmyk' },
        xyz: { channels: 3, labels: 'xyz' },
        lab: { channels: 3, labels: 'lab' },
        lch: { channels: 3, labels: 'lch' },
        hex: { channels: 1, labels: ['hex'] },
        keyword: { channels: 1, labels: ['keyword'] },
        ansi16: { channels: 1, labels: ['ansi16'] },
        ansi256: { channels: 1, labels: ['ansi256'] },
        hcg: { channels: 3, labels: ['h', 'c', 'g'] },
        apple: { channels: 3, labels: ['r16', 'g16', 'b16'] },
        gray: { channels: 1, labels: ['gray'] }
    };
    module.exports = convert;
    // Hide .channels and .labels properties
    for (const model of Object.keys(convert)) {
        if (!('channels' in convert[model])) {
            throw new Error('missing channels property: ' + model);
        }
        if (!('labels' in convert[model])) {
            throw new Error('missing channel labels property: ' + model);
        }
        if (convert[model].labels.length !== convert[model].channels) {
            throw new Error('channel and label counts mismatch: ' + model);
        }
        const { channels, labels } = convert[model];
        delete convert[model].channels;
        delete convert[model].labels;
        Object.defineProperty(convert[model], 'channels', { value: channels });
        Object.defineProperty(convert[model], 'labels', { value: labels });
    }
    convert.rgb.hsl = function (rgb) {
        const r = rgb[0] / 255;
        const g = rgb[1] / 255;
        const b = rgb[2] / 255;
        const min = Math.min(r, g, b);
        const max = Math.max(r, g, b);
        const delta = max - min;
        let h;
        let s;
        if (max === min) {
            h = 0;
        }
        else if (r === max) {
            h = (g - b) / delta;
        }
        else if (g === max) {
            h = 2 + (b - r) / delta;
        }
        else if (b === max) {
            h = 4 + (r - g) / delta;
        }
        h = Math.min(h * 60, 360);
        if (h < 0) {
            h += 360;
        }
        const l = (min + max) / 2;
        if (max === min) {
            s = 0;
        }
        else if (l <= 0.5) {
            s = delta / (max + min);
        }
        else {
            s = delta / (2 - max - min);
        }
        return [h, s * 100, l * 100];
    };
    convert.rgb.hsv = function (rgb) {
        let rdif;
        let gdif;
        let bdif;
        let h;
        let s;
        const r = rgb[0] / 255;
        const g = rgb[1] / 255;
        const b = rgb[2] / 255;
        const v = Math.max(r, g, b);
        const diff = v - Math.min(r, g, b);
        const diffc = function (c) {
            return (v - c) / 6 / diff + 1 / 2;
        };
        if (diff === 0) {
            h = 0;
            s = 0;
        }
        else {
            s = diff / v;
            rdif = diffc(r);
            gdif = diffc(g);
            bdif = diffc(b);
            if (r === v) {
                h = bdif - gdif;
            }
            else if (g === v) {
                h = (1 / 3) + rdif - bdif;
            }
            else if (b === v) {
                h = (2 / 3) + gdif - rdif;
            }
            if (h < 0) {
                h += 1;
            }
            else if (h > 1) {
                h -= 1;
            }
        }
        return [
            h * 360,
            s * 100,
            v * 100
        ];
    };
    convert.rgb.hwb = function (rgb) {
        const r = rgb[0];
        const g = rgb[1];
        let b = rgb[2];
        const h = convert.rgb.hsl(rgb)[0];
        const w = 1 / 255 * Math.min(r, Math.min(g, b));
        b = 1 - 1 / 255 * Math.max(r, Math.max(g, b));
        return [h, w * 100, b * 100];
    };
    convert.rgb.cmyk = function (rgb) {
        const r = rgb[0] / 255;
        const g = rgb[1] / 255;
        const b = rgb[2] / 255;
        const k = Math.min(1 - r, 1 - g, 1 - b);
        const c = (1 - r - k) / (1 - k) || 0;
        const m = (1 - g - k) / (1 - k) || 0;
        const y = (1 - b - k) / (1 - k) || 0;
        return [c * 100, m * 100, y * 100, k * 100];
    };
    function comparativeDistance(x, y) {
        /*
            See https://en.m.wikipedia.org/wiki/Euclidean_distance#Squared_Euclidean_distance
        */
        return (((x[0] - y[0]) ** 2) +
            ((x[1] - y[1]) ** 2) +
            ((x[2] - y[2]) ** 2));
    }
    convert.rgb.keyword = function (rgb) {
        const reversed = reverseKeywords[rgb];
        if (reversed) {
            return reversed;
        }
        let currentClosestDistance = Infinity;
        let currentClosestKeyword;
        for (const keyword of Object.keys(cssKeywords)) {
            const value = cssKeywords[keyword];
            // Compute comparative distance
            const distance = comparativeDistance(rgb, value);
            // Check if its less, if so set as closest
            if (distance < currentClosestDistance) {
                currentClosestDistance = distance;
                currentClosestKeyword = keyword;
            }
        }
        return currentClosestKeyword;
    };
    convert.keyword.rgb = function (keyword) {
        return cssKeywords[keyword];
    };
    convert.rgb.xyz = function (rgb) {
        let r = rgb[0] / 255;
        let g = rgb[1] / 255;
        let b = rgb[2] / 255;
        // Assume sRGB
        r = r > 0.04045 ? (((r + 0.055) / 1.055) ** 2.4) : (r / 12.92);
        g = g > 0.04045 ? (((g + 0.055) / 1.055) ** 2.4) : (g / 12.92);
        b = b > 0.04045 ? (((b + 0.055) / 1.055) ** 2.4) : (b / 12.92);
        const x = (r * 0.4124) + (g * 0.3576) + (b * 0.1805);
        const y = (r * 0.2126) + (g * 0.7152) + (b * 0.0722);
        const z = (r * 0.0193) + (g * 0.1192) + (b * 0.9505);
        return [x * 100, y * 100, z * 100];
    };
    convert.rgb.lab = function (rgb) {
        const xyz = convert.rgb.xyz(rgb);
        let x = xyz[0];
        let y = xyz[1];
        let z = xyz[2];
        x /= 95.047;
        y /= 100;
        z /= 108.883;
        x = x > 0.008856 ? (x ** (1 / 3)) : (7.787 * x) + (16 / 116);
        y = y > 0.008856 ? (y ** (1 / 3)) : (7.787 * y) + (16 / 116);
        z = z > 0.008856 ? (z ** (1 / 3)) : (7.787 * z) + (16 / 116);
        const l = (116 * y) - 16;
        const a = 500 * (x - y);
        const b = 200 * (y - z);
        return [l, a, b];
    };
    convert.hsl.rgb = function (hsl) {
        const h = hsl[0] / 360;
        const s = hsl[1] / 100;
        const l = hsl[2] / 100;
        let t2;
        let t3;
        let val;
        if (s === 0) {
            val = l * 255;
            return [val, val, val];
        }
        if (l < 0.5) {
            t2 = l * (1 + s);
        }
        else {
            t2 = l + s - l * s;
        }
        const t1 = 2 * l - t2;
        const rgb = [0, 0, 0];
        for (let i = 0; i < 3; i++) {
            t3 = h + 1 / 3 * -(i - 1);
            if (t3 < 0) {
                t3++;
            }
            if (t3 > 1) {
                t3--;
            }
            if (6 * t3 < 1) {
                val = t1 + (t2 - t1) * 6 * t3;
            }
            else if (2 * t3 < 1) {
                val = t2;
            }
            else if (3 * t3 < 2) {
                val = t1 + (t2 - t1) * (2 / 3 - t3) * 6;
            }
            else {
                val = t1;
            }
            rgb[i] = val * 255;
        }
        return rgb;
    };
    convert.hsl.hsv = function (hsl) {
        const h = hsl[0];
        let s = hsl[1] / 100;
        let l = hsl[2] / 100;
        let smin = s;
        const lmin = Math.max(l, 0.01);
        l *= 2;
        s *= (l <= 1) ? l : 2 - l;
        smin *= lmin <= 1 ? lmin : 2 - lmin;
        const v = (l + s) / 2;
        const sv = l === 0 ? (2 * smin) / (lmin + smin) : (2 * s) / (l + s);
        return [h, sv * 100, v * 100];
    };
    convert.hsv.rgb = function (hsv) {
        const h = hsv[0] / 60;
        const s = hsv[1] / 100;
        let v = hsv[2] / 100;
        const hi = Math.floor(h) % 6;
        const f = h - Math.floor(h);
        const p = 255 * v * (1 - s);
        const q = 255 * v * (1 - (s * f));
        const t = 255 * v * (1 - (s * (1 - f)));
        v *= 255;
        switch (hi) {
            case 0:
                return [v, t, p];
            case 1:
                return [q, v, p];
            case 2:
                return [p, v, t];
            case 3:
                return [p, q, v];
            case 4:
                return [t, p, v];
            case 5:
                return [v, p, q];
        }
    };
    convert.hsv.hsl = function (hsv) {
        const h = hsv[0];
        const s = hsv[1] / 100;
        const v = hsv[2] / 100;
        const vmin = Math.max(v, 0.01);
        let sl;
        let l;
        l = (2 - s) * v;
        const lmin = (2 - s) * vmin;
        sl = s * vmin;
        sl /= (lmin <= 1) ? lmin : 2 - lmin;
        sl = sl || 0;
        l /= 2;
        return [h, sl * 100, l * 100];
    };
    // http://dev.w3.org/csswg/css-color/#hwb-to-rgb
    convert.hwb.rgb = function (hwb) {
        const h = hwb[0] / 360;
        let wh = hwb[1] / 100;
        let bl = hwb[2] / 100;
        const ratio = wh + bl;
        let f;
        // Wh + bl cant be > 1
        if (ratio > 1) {
            wh /= ratio;
            bl /= ratio;
        }
        const i = Math.floor(6 * h);
        const v = 1 - bl;
        f = 6 * h - i;
        if ((i & 0x01) !== 0) {
            f = 1 - f;
        }
        const n = wh + f * (v - wh); // Linear interpolation
        let r;
        let g;
        let b;
        /* eslint-disable max-statements-per-line,no-multi-spaces */
        switch (i) {
            default:
            case 6:
            case 0:
                r = v;
                g = n;
                b = wh;
                break;
            case 1:
                r = n;
                g = v;
                b = wh;
                break;
            case 2:
                r = wh;
                g = v;
                b = n;
                break;
            case 3:
                r = wh;
                g = n;
                b = v;
                break;
            case 4:
                r = n;
                g = wh;
                b = v;
                break;
            case 5:
                r = v;
                g = wh;
                b = n;
                break;
        }
        /* eslint-enable max-statements-per-line,no-multi-spaces */
        return [r * 255, g * 255, b * 255];
    };
    convert.cmyk.rgb = function (cmyk) {
        const c = cmyk[0] / 100;
        const m = cmyk[1] / 100;
        const y = cmyk[2] / 100;
        const k = cmyk[3] / 100;
        const r = 1 - Math.min(1, c * (1 - k) + k);
        const g = 1 - Math.min(1, m * (1 - k) + k);
        const b = 1 - Math.min(1, y * (1 - k) + k);
        return [r * 255, g * 255, b * 255];
    };
    convert.xyz.rgb = function (xyz) {
        const x = xyz[0] / 100;
        const y = xyz[1] / 100;
        const z = xyz[2] / 100;
        let r;
        let g;
        let b;
        r = (x * 3.2406) + (y * -1.5372) + (z * -0.4986);
        g = (x * -0.9689) + (y * 1.8758) + (z * 0.0415);
        b = (x * 0.0557) + (y * -0.2040) + (z * 1.0570);
        // Assume sRGB
        r = r > 0.0031308
            ? ((1.055 * (r ** (1.0 / 2.4))) - 0.055)
            : r * 12.92;
        g = g > 0.0031308
            ? ((1.055 * (g ** (1.0 / 2.4))) - 0.055)
            : g * 12.92;
        b = b > 0.0031308
            ? ((1.055 * (b ** (1.0 / 2.4))) - 0.055)
            : b * 12.92;
        r = Math.min(Math.max(0, r), 1);
        g = Math.min(Math.max(0, g), 1);
        b = Math.min(Math.max(0, b), 1);
        return [r * 255, g * 255, b * 255];
    };
    convert.xyz.lab = function (xyz) {
        let x = xyz[0];
        let y = xyz[1];
        let z = xyz[2];
        x /= 95.047;
        y /= 100;
        z /= 108.883;
        x = x > 0.008856 ? (x ** (1 / 3)) : (7.787 * x) + (16 / 116);
        y = y > 0.008856 ? (y ** (1 / 3)) : (7.787 * y) + (16 / 116);
        z = z > 0.008856 ? (z ** (1 / 3)) : (7.787 * z) + (16 / 116);
        const l = (116 * y) - 16;
        const a = 500 * (x - y);
        const b = 200 * (y - z);
        return [l, a, b];
    };
    convert.lab.xyz = function (lab) {
        const l = lab[0];
        const a = lab[1];
        const b = lab[2];
        let x;
        let y;
        let z;
        y = (l + 16) / 116;
        x = a / 500 + y;
        z = y - b / 200;
        const y2 = y ** 3;
        const x2 = x ** 3;
        const z2 = z ** 3;
        y = y2 > 0.008856 ? y2 : (y - 16 / 116) / 7.787;
        x = x2 > 0.008856 ? x2 : (x - 16 / 116) / 7.787;
        z = z2 > 0.008856 ? z2 : (z - 16 / 116) / 7.787;
        x *= 95.047;
        y *= 100;
        z *= 108.883;
        return [x, y, z];
    };
    convert.lab.lch = function (lab) {
        const l = lab[0];
        const a = lab[1];
        const b = lab[2];
        let h;
        const hr = Math.atan2(b, a);
        h = hr * 360 / 2 / Math.PI;
        if (h < 0) {
            h += 360;
        }
        const c = Math.sqrt(a * a + b * b);
        return [l, c, h];
    };
    convert.lch.lab = function (lch) {
        const l = lch[0];
        const c = lch[1];
        const h = lch[2];
        const hr = h / 360 * 2 * Math.PI;
        const a = c * Math.cos(hr);
        const b = c * Math.sin(hr);
        return [l, a, b];
    };
    convert.rgb.ansi16 = function (args, saturation = null) {
        const [r, g, b] = args;
        let value = saturation === null ? convert.rgb.hsv(args)[2] : saturation; // Hsv -> ansi16 optimization
        value = Math.round(value / 50);
        if (value === 0) {
            return 30;
        }
        let ansi = 30
            + ((Math.round(b / 255) << 2)
                | (Math.round(g / 255) << 1)
                | Math.round(r / 255));
        if (value === 2) {
            ansi += 60;
        }
        return ansi;
    };
    convert.hsv.ansi16 = function (args) {
        // Optimization here; we already know the value and don't need to get
        // it converted for us.
        return convert.rgb.ansi16(convert.hsv.rgb(args), args[2]);
    };
    convert.rgb.ansi256 = function (args) {
        const r = args[0];
        const g = args[1];
        const b = args[2];
        // We use the extended greyscale palette here, with the exception of
        // black and white. normal palette only has 4 greyscale shades.
        if (r === g && g === b) {
            if (r < 8) {
                return 16;
            }
            if (r > 248) {
                return 231;
            }
            return Math.round(((r - 8) / 247) * 24) + 232;
        }
        const ansi = 16
            + (36 * Math.round(r / 255 * 5))
            + (6 * Math.round(g / 255 * 5))
            + Math.round(b / 255 * 5);
        return ansi;
    };
    convert.ansi16.rgb = function (args) {
        let color = args % 10;
        // Handle greyscale
        if (color === 0 || color === 7) {
            if (args > 50) {
                color += 3.5;
            }
            color = color / 10.5 * 255;
            return [color, color, color];
        }
        const mult = (~~(args > 50) + 1) * 0.5;
        const r = ((color & 1) * mult) * 255;
        const g = (((color >> 1) & 1) * mult) * 255;
        const b = (((color >> 2) & 1) * mult) * 255;
        return [r, g, b];
    };
    convert.ansi256.rgb = function (args) {
        // Handle greyscale
        if (args >= 232) {
            const c = (args - 232) * 10 + 8;
            return [c, c, c];
        }
        args -= 16;
        let rem;
        const r = Math.floor(args / 36) / 5 * 255;
        const g = Math.floor((rem = args % 36) / 6) / 5 * 255;
        const b = (rem % 6) / 5 * 255;
        return [r, g, b];
    };
    convert.rgb.hex = function (args) {
        const integer = ((Math.round(args[0]) & 0xFF) << 16)
            + ((Math.round(args[1]) & 0xFF) << 8)
            + (Math.round(args[2]) & 0xFF);
        const string = integer.toString(16).toUpperCase();
        return '000000'.substring(string.length) + string;
    };
    convert.hex.rgb = function (args) {
        const match = args.toString(16).match(/[a-f0-9]{6}|[a-f0-9]{3}/i);
        if (!match) {
            return [0, 0, 0];
        }
        let colorString = match[0];
        if (match[0].length === 3) {
            colorString = colorString.split('').map(char => {
                return char + char;
            }).join('');
        }
        const integer = parseInt(colorString, 16);
        const r = (integer >> 16) & 0xFF;
        const g = (integer >> 8) & 0xFF;
        const b = integer & 0xFF;
        return [r, g, b];
    };
    convert.rgb.hcg = function (rgb) {
        const r = rgb[0] / 255;
        const g = rgb[1] / 255;
        const b = rgb[2] / 255;
        const max = Math.max(Math.max(r, g), b);
        const min = Math.min(Math.min(r, g), b);
        const chroma = (max - min);
        let grayscale;
        let hue;
        if (chroma < 1) {
            grayscale = min / (1 - chroma);
        }
        else {
            grayscale = 0;
        }
        if (chroma <= 0) {
            hue = 0;
        }
        else if (max === r) {
            hue = ((g - b) / chroma) % 6;
        }
        else if (max === g) {
            hue = 2 + (b - r) / chroma;
        }
        else {
            hue = 4 + (r - g) / chroma;
        }
        hue /= 6;
        hue %= 1;
        return [hue * 360, chroma * 100, grayscale * 100];
    };
    convert.hsl.hcg = function (hsl) {
        const s = hsl[1] / 100;
        const l = hsl[2] / 100;
        const c = l < 0.5 ? (2.0 * s * l) : (2.0 * s * (1.0 - l));
        let f = 0;
        if (c < 1.0) {
            f = (l - 0.5 * c) / (1.0 - c);
        }
        return [hsl[0], c * 100, f * 100];
    };
    convert.hsv.hcg = function (hsv) {
        const s = hsv[1] / 100;
        const v = hsv[2] / 100;
        const c = s * v;
        let f = 0;
        if (c < 1.0) {
            f = (v - c) / (1 - c);
        }
        return [hsv[0], c * 100, f * 100];
    };
    convert.hcg.rgb = function (hcg) {
        const h = hcg[0] / 360;
        const c = hcg[1] / 100;
        const g = hcg[2] / 100;
        if (c === 0.0) {
            return [g * 255, g * 255, g * 255];
        }
        const pure = [0, 0, 0];
        const hi = (h % 1) * 6;
        const v = hi % 1;
        const w = 1 - v;
        let mg = 0;
        /* eslint-disable max-statements-per-line */
        switch (Math.floor(hi)) {
            case 0:
                pure[0] = 1;
                pure[1] = v;
                pure[2] = 0;
                break;
            case 1:
                pure[0] = w;
                pure[1] = 1;
                pure[2] = 0;
                break;
            case 2:
                pure[0] = 0;
                pure[1] = 1;
                pure[2] = v;
                break;
            case 3:
                pure[0] = 0;
                pure[1] = w;
                pure[2] = 1;
                break;
            case 4:
                pure[0] = v;
                pure[1] = 0;
                pure[2] = 1;
                break;
            default:
                pure[0] = 1;
                pure[1] = 0;
                pure[2] = w;
        }
        /* eslint-enable max-statements-per-line */
        mg = (1.0 - c) * g;
        return [
            (c * pure[0] + mg) * 255,
            (c * pure[1] + mg) * 255,
            (c * pure[2] + mg) * 255
        ];
    };
    convert.hcg.hsv = function (hcg) {
        const c = hcg[1] / 100;
        const g = hcg[2] / 100;
        const v = c + g * (1.0 - c);
        let f = 0;
        if (v > 0.0) {
            f = c / v;
        }
        return [hcg[0], f * 100, v * 100];
    };
    convert.hcg.hsl = function (hcg) {
        const c = hcg[1] / 100;
        const g = hcg[2] / 100;
        const l = g * (1.0 - c) + 0.5 * c;
        let s = 0;
        if (l > 0.0 && l < 0.5) {
            s = c / (2 * l);
        }
        else if (l >= 0.5 && l < 1.0) {
            s = c / (2 * (1 - l));
        }
        return [hcg[0], s * 100, l * 100];
    };
    convert.hcg.hwb = function (hcg) {
        const c = hcg[1] / 100;
        const g = hcg[2] / 100;
        const v = c + g * (1.0 - c);
        return [hcg[0], (v - c) * 100, (1 - v) * 100];
    };
    convert.hwb.hcg = function (hwb) {
        const w = hwb[1] / 100;
        const b = hwb[2] / 100;
        const v = 1 - b;
        const c = v - w;
        let g = 0;
        if (c < 1) {
            g = (v - c) / (1 - c);
        }
        return [hwb[0], c * 100, g * 100];
    };
    convert.apple.rgb = function (apple) {
        return [(apple[0] / 65535) * 255, (apple[1] / 65535) * 255, (apple[2] / 65535) * 255];
    };
    convert.rgb.apple = function (rgb) {
        return [(rgb[0] / 255) * 65535, (rgb[1] / 255) * 65535, (rgb[2] / 255) * 65535];
    };
    convert.gray.rgb = function (args) {
        return [args[0] / 100 * 255, args[0] / 100 * 255, args[0] / 100 * 255];
    };
    convert.gray.hsl = function (args) {
        return [0, 0, args[0]];
    };
    convert.gray.hsv = convert.gray.hsl;
    convert.gray.hwb = function (gray) {
        return [0, 100, gray[0]];
    };
    convert.gray.cmyk = function (gray) {
        return [0, 0, 0, gray[0]];
    };
    convert.gray.lab = function (gray) {
        return [gray[0], 0, 0];
    };
    convert.gray.hex = function (gray) {
        const val = Math.round(gray[0] / 100 * 255) & 0xFF;
        const integer = (val << 16) + (val << 8) + val;
        const string = integer.toString(16).toUpperCase();
        return '000000'.substring(string.length) + string;
    };
    convert.rgb.gray = function (rgb) {
        const val = (rgb[0] + rgb[1] + rgb[2]) / 3;
        return [val / 255 * 100];
    };
},
/* color-name/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        "aliceblue": [240, 248, 255],
        "antiquewhite": [250, 235, 215],
        "aqua": [0, 255, 255],
        "aquamarine": [127, 255, 212],
        "azure": [240, 255, 255],
        "beige": [245, 245, 220],
        "bisque": [255, 228, 196],
        "black": [0, 0, 0],
        "blanchedalmond": [255, 235, 205],
        "blue": [0, 0, 255],
        "blueviolet": [138, 43, 226],
        "brown": [165, 42, 42],
        "burlywood": [222, 184, 135],
        "cadetblue": [95, 158, 160],
        "chartreuse": [127, 255, 0],
        "chocolate": [210, 105, 30],
        "coral": [255, 127, 80],
        "cornflowerblue": [100, 149, 237],
        "cornsilk": [255, 248, 220],
        "crimson": [220, 20, 60],
        "cyan": [0, 255, 255],
        "darkblue": [0, 0, 139],
        "darkcyan": [0, 139, 139],
        "darkgoldenrod": [184, 134, 11],
        "darkgray": [169, 169, 169],
        "darkgreen": [0, 100, 0],
        "darkgrey": [169, 169, 169],
        "darkkhaki": [189, 183, 107],
        "darkmagenta": [139, 0, 139],
        "darkolivegreen": [85, 107, 47],
        "darkorange": [255, 140, 0],
        "darkorchid": [153, 50, 204],
        "darkred": [139, 0, 0],
        "darksalmon": [233, 150, 122],
        "darkseagreen": [143, 188, 143],
        "darkslateblue": [72, 61, 139],
        "darkslategray": [47, 79, 79],
        "darkslategrey": [47, 79, 79],
        "darkturquoise": [0, 206, 209],
        "darkviolet": [148, 0, 211],
        "deeppink": [255, 20, 147],
        "deepskyblue": [0, 191, 255],
        "dimgray": [105, 105, 105],
        "dimgrey": [105, 105, 105],
        "dodgerblue": [30, 144, 255],
        "firebrick": [178, 34, 34],
        "floralwhite": [255, 250, 240],
        "forestgreen": [34, 139, 34],
        "fuchsia": [255, 0, 255],
        "gainsboro": [220, 220, 220],
        "ghostwhite": [248, 248, 255],
        "gold": [255, 215, 0],
        "goldenrod": [218, 165, 32],
        "gray": [128, 128, 128],
        "green": [0, 128, 0],
        "greenyellow": [173, 255, 47],
        "grey": [128, 128, 128],
        "honeydew": [240, 255, 240],
        "hotpink": [255, 105, 180],
        "indianred": [205, 92, 92],
        "indigo": [75, 0, 130],
        "ivory": [255, 255, 240],
        "khaki": [240, 230, 140],
        "lavender": [230, 230, 250],
        "lavenderblush": [255, 240, 245],
        "lawngreen": [124, 252, 0],
        "lemonchiffon": [255, 250, 205],
        "lightblue": [173, 216, 230],
        "lightcoral": [240, 128, 128],
        "lightcyan": [224, 255, 255],
        "lightgoldenrodyellow": [250, 250, 210],
        "lightgray": [211, 211, 211],
        "lightgreen": [144, 238, 144],
        "lightgrey": [211, 211, 211],
        "lightpink": [255, 182, 193],
        "lightsalmon": [255, 160, 122],
        "lightseagreen": [32, 178, 170],
        "lightskyblue": [135, 206, 250],
        "lightslategray": [119, 136, 153],
        "lightslategrey": [119, 136, 153],
        "lightsteelblue": [176, 196, 222],
        "lightyellow": [255, 255, 224],
        "lime": [0, 255, 0],
        "limegreen": [50, 205, 50],
        "linen": [250, 240, 230],
        "magenta": [255, 0, 255],
        "maroon": [128, 0, 0],
        "mediumaquamarine": [102, 205, 170],
        "mediumblue": [0, 0, 205],
        "mediumorchid": [186, 85, 211],
        "mediumpurple": [147, 112, 219],
        "mediumseagreen": [60, 179, 113],
        "mediumslateblue": [123, 104, 238],
        "mediumspringgreen": [0, 250, 154],
        "mediumturquoise": [72, 209, 204],
        "mediumvioletred": [199, 21, 133],
        "midnightblue": [25, 25, 112],
        "mintcream": [245, 255, 250],
        "mistyrose": [255, 228, 225],
        "moccasin": [255, 228, 181],
        "navajowhite": [255, 222, 173],
        "navy": [0, 0, 128],
        "oldlace": [253, 245, 230],
        "olive": [128, 128, 0],
        "olivedrab": [107, 142, 35],
        "orange": [255, 165, 0],
        "orangered": [255, 69, 0],
        "orchid": [218, 112, 214],
        "palegoldenrod": [238, 232, 170],
        "palegreen": [152, 251, 152],
        "paleturquoise": [175, 238, 238],
        "palevioletred": [219, 112, 147],
        "papayawhip": [255, 239, 213],
        "peachpuff": [255, 218, 185],
        "peru": [205, 133, 63],
        "pink": [255, 192, 203],
        "plum": [221, 160, 221],
        "powderblue": [176, 224, 230],
        "purple": [128, 0, 128],
        "rebeccapurple": [102, 51, 153],
        "red": [255, 0, 0],
        "rosybrown": [188, 143, 143],
        "royalblue": [65, 105, 225],
        "saddlebrown": [139, 69, 19],
        "salmon": [250, 128, 114],
        "sandybrown": [244, 164, 96],
        "seagreen": [46, 139, 87],
        "seashell": [255, 245, 238],
        "sienna": [160, 82, 45],
        "silver": [192, 192, 192],
        "skyblue": [135, 206, 235],
        "slateblue": [106, 90, 205],
        "slategray": [112, 128, 144],
        "slategrey": [112, 128, 144],
        "snow": [255, 250, 250],
        "springgreen": [0, 255, 127],
        "steelblue": [70, 130, 180],
        "tan": [210, 180, 140],
        "teal": [0, 128, 128],
        "thistle": [216, 191, 216],
        "tomato": [255, 99, 71],
        "turquoise": [64, 224, 208],
        "violet": [238, 130, 238],
        "wheat": [245, 222, 179],
        "white": [255, 255, 255],
        "whitesmoke": [245, 245, 245],
        "yellow": [255, 255, 0],
        "yellowgreen": [154, 205, 50]
    };
},
/* color-convert/route.js */ function _(require, module, exports, __esModule, __esExport) {
    const conversions = require(14) /* ./conversions */;
    /*
        This function routes a model to all other models.
    
        all functions that are routed have a property `.conversion` attached
        to the returned synthetic function. This property is an array
        of strings, each with the steps in between the 'from' and 'to'
        color models (inclusive).
    
        conversions that are not possible simply are not included.
    */
    function buildGraph() {
        const graph = {};
        // https://jsperf.com/object-keys-vs-for-in-with-closure/3
        const models = Object.keys(conversions);
        for (let len = models.length, i = 0; i < len; i++) {
            graph[models[i]] = {
                // http://jsperf.com/1-vs-infinity
                // micro-opt, but this is simple.
                distance: -1,
                parent: null
            };
        }
        return graph;
    }
    // https://en.wikipedia.org/wiki/Breadth-first_search
    function deriveBFS(fromModel) {
        const graph = buildGraph();
        const queue = [fromModel]; // Unshift -> queue -> pop
        graph[fromModel].distance = 0;
        while (queue.length) {
            const current = queue.pop();
            const adjacents = Object.keys(conversions[current]);
            for (let len = adjacents.length, i = 0; i < len; i++) {
                const adjacent = adjacents[i];
                const node = graph[adjacent];
                if (node.distance === -1) {
                    node.distance = graph[current].distance + 1;
                    node.parent = current;
                    queue.unshift(adjacent);
                }
            }
        }
        return graph;
    }
    function link(from, to) {
        return function (args) {
            return to(from(args));
        };
    }
    function wrapConversion(toModel, graph) {
        const path = [graph[toModel].parent, toModel];
        let fn = conversions[graph[toModel].parent][toModel];
        let cur = graph[toModel].parent;
        while (graph[cur].parent) {
            path.unshift(graph[cur].parent);
            fn = link(conversions[graph[cur].parent][cur], fn);
            cur = graph[cur].parent;
        }
        fn.conversion = path;
        return fn;
    }
    module.exports = function (fromModel) {
        const graph = deriveBFS(fromModel);
        const conversion = {};
        const models = Object.keys(graph);
        for (let len = models.length, i = 0; i < len; i++) {
            const toModel = models[i];
            const node = graph[toModel];
            if (node.parent === null) {
                // No possible conversion, or this node is the source model.
                continue;
            }
            conversion[toModel] = wrapConversion(toModel, graph);
        }
        return conversion;
    };
},
/* escalade/sync/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const { dirname, resolve } = require('path');
    const { readdirSync, statSync } = require('fs');
    module.exports = function (start, callback) {
        let dir = resolve('.', start);
        let tmp, stats = statSync(dir);
        if (!stats.isDirectory()) {
            dir = dirname(dir);
        }
        while (true) {
            tmp = callback(dir, readdirSync(dir));
            if (tmp)
                return resolve(dir, tmp);
            dir = dirname(tmp = dir);
            if (tmp === dir)
                break;
        }
    };
},
/* get-caller-file/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // Call this function in a another function to find out the file from
    // which that function was called from. (Inspects the v8 stack trace)
    //
    // Inspired by http://stackoverflow.com/questions/13227489
    module.exports = function getCallerFile(position) {
        if (position === void 0) {
            position = 2;
        }
        if (position >= Error.stackTraceLimit) {
            throw new TypeError('getCallerFile(position) requires position be less then Error.stackTraceLimit but position was: `' + position + '` and Error.stackTraceLimit was: `' + Error.stackTraceLimit + '`');
        }
        var oldPrepareStackTrace = Error.prepareStackTrace;
        Error.prepareStackTrace = function (_, stack) { return stack; };
        var stack = new Error().stack;
        Error.prepareStackTrace = oldPrepareStackTrace;
        if (stack !== null && typeof stack === 'object') {
            // stack[0] holds this file
            // stack[1] holds where this function was called
            // stack[2] holds the file we're interested in
            return stack[position] ? stack[position].getFileName() : undefined;
        }
    };
},
/* require-directory/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var fs = require('fs'), join = require('path').join, resolve = require('path').resolve, dirname = require('path').dirname, defaultOptions = {
        extensions: ['js', 'json', 'coffee'],
        recurse: true,
        rename: function (name) {
            return name;
        },
        visit: function (obj) {
            return obj;
        }
    };
    function checkFileInclusion(path, filename, options) {
        return (
        // verify file has valid extension
        (new RegExp('\\.(' + options.extensions.join('|') + ')$', 'i').test(filename)) &&
            // if options.include is a RegExp, evaluate it and make sure the path passes
            !(options.include && options.include instanceof RegExp && !options.include.test(path)) &&
            // if options.include is a function, evaluate it and make sure the path passes
            !(options.include && typeof options.include === 'function' && !options.include(path, filename)) &&
            // if options.exclude is a RegExp, evaluate it and make sure the path doesn't pass
            !(options.exclude && options.exclude instanceof RegExp && options.exclude.test(path)) &&
            // if options.exclude is a function, evaluate it and make sure the path doesn't pass
            !(options.exclude && typeof options.exclude === 'function' && options.exclude(path, filename)));
    }
    function requireDirectory(m, path, options) {
        var retval = {};
        // path is optional
        if (path && !options && typeof path !== 'string') {
            options = path;
            path = null;
        }
        // default options
        options = options || {};
        for (var prop in defaultOptions) {
            if (typeof options[prop] === 'undefined') {
                options[prop] = defaultOptions[prop];
            }
        }
        // if no path was passed in, assume the equivelant of __dirname from caller
        // otherwise, resolve path relative to the equivalent of __dirname
        path = !path ? dirname(m.filename) : resolve(dirname(m.filename), path);
        // get the path of each file in specified directory, append to current tree node, recurse
        fs.readdirSync(path).forEach(function (filename) {
            var joined = join(path, filename), files, key, obj;
            if (fs.statSync(joined).isDirectory() && options.recurse) {
                // this node is a directory; recurse
                files = requireDirectory(m, joined, options);
                // exclude empty directories
                if (Object.keys(files).length) {
                    retval[options.rename(filename, joined, filename)] = files;
                }
            }
            else {
                if (joined !== m.filename && checkFileInclusion(joined, filename, options)) {
                    // hash node key shouldn't include file extension
                    key = filename.substring(0, filename.lastIndexOf('.'));
                    obj = m.require(joined);
                    retval[options.rename(key, joined, filename)] = options.visit(obj, joined, filename) || obj;
                }
            }
        });
        return retval;
    }
    module.exports = requireDirectory;
    module.exports.defaults = defaultOptions;
},
/* sys.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const crypto_1 = __importDefault(require("crypto"));
    const path_1 = require("path");
    function scan(path, extensions, exclude, include, depth) {
        return ts.sys.readDirectory(path, extensions, exclude, include, depth).map((p) => (0, path_1.normalize)(p));
    }
    exports.scan = scan;
    function glob(...patterns) {
        return scan(".", undefined, undefined, patterns);
    }
    exports.glob = glob;
    exports.read = ts.sys.readFile;
    exports.write = ts.sys.writeFile;
    exports.file_exists = ts.sys.fileExists;
    exports.directory_exists = ts.sys.directoryExists;
    function rename(path, options) {
        let { dir, name, ext } = (0, path_1.parse)(path);
        if (options.dir != null) {
            if (options.base != null)
                dir = dir.replace(options.base, options.dir);
            else
                dir = options.dir;
        }
        if (options.name != null)
            name = options.name(name);
        if (options.ext != null)
            ext = options.ext;
        return (0, path_1.format)({ dir, name, ext });
    }
    exports.rename = rename;
    function hash(data) {
        return crypto_1.default.createHash("sha256").update(data).digest("hex");
    }
    exports.hash = hash;
    function hash_file(path) {
        const contents = (0, exports.read)(path);
        return contents != null ? hash(contents) : null;
    }
    exports.hash_file = hash_file;
    function read_json(path) {
        const data = (0, exports.read)(path);
        if (data == null)
            return undefined;
        else {
            try {
                return JSON.parse(data);
            }
            catch {
                return undefined;
            }
        }
    }
    exports.read_json = read_json;
},
/* typescript/lib/typescript.js */ function _(require, module, exports, __esModule, __esExport) {
    var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
        if (pack || arguments.length === 2)
            for (var i = 0, l = from.length, ar; i < l; i++) {
                if (ar || !(i in from)) {
                    if (!ar)
                        ar = Array.prototype.slice.call(from, 0, i);
                    ar[i] = from[i];
                }
            }
        return to.concat(ar || Array.prototype.slice.call(from));
    };
    var __assign = (this && this.__assign) || function () {
        __assign = Object.assign || function (t) {
            for (var s, i = 1, n = arguments.length; i < n; i++) {
                s = arguments[i];
                for (var p in s)
                    if (Object.prototype.hasOwnProperty.call(s, p))
                        t[p] = s[p];
            }
            return t;
        };
        return __assign.apply(this, arguments);
    };
    var __makeTemplateObject = (this && this.__makeTemplateObject) || function (cooked, raw) {
        if (Object.defineProperty) {
            Object.defineProperty(cooked, "raw", { value: raw });
        }
        else {
            cooked.raw = raw;
        }
        return cooked;
    };
    var __generator = (this && this.__generator) || function (thisArg, body) {
        var _ = { label: 0, sent: function () { if (t[0] & 1)
                throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
        return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function () { return this; }), g;
        function verb(n) { return function (v) { return step([n, v]); }; }
        function step(op) {
            if (f)
                throw new TypeError("Generator is already executing.");
            while (_)
                try {
                    if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done)
                        return t;
                    if (y = 0, t)
                        op = [op[0] & 2, t.value];
                    switch (op[0]) {
                        case 0:
                        case 1:
                            t = op;
                            break;
                        case 4:
                            _.label++;
                            return { value: op[1], done: false };
                        case 5:
                            _.label++;
                            y = op[1];
                            op = [0];
                            continue;
                        case 7:
                            op = _.ops.pop();
                            _.trys.pop();
                            continue;
                        default:
                            if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) {
                                _ = 0;
                                continue;
                            }
                            if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) {
                                _.label = op[1];
                                break;
                            }
                            if (op[0] === 6 && _.label < t[1]) {
                                _.label = t[1];
                                t = op;
                                break;
                            }
                            if (t && _.label < t[2]) {
                                _.label = t[2];
                                _.ops.push(op);
                                break;
                            }
                            if (t[2])
                                _.ops.pop();
                            _.trys.pop();
                            continue;
                    }
                    op = body.call(thisArg, _);
                }
                catch (e) {
                    op = [6, e];
                    y = 0;
                }
                finally {
                    f = t = 0;
                }
            if (op[0] & 5)
                throw op[1];
            return { value: op[0] ? op[1] : void 0, done: true };
        }
    };
    var __rest = (this && this.__rest) || function (s, e) {
        var t = {};
        for (var p in s)
            if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
                t[p] = s[p];
        if (s != null && typeof Object.getOwnPropertySymbols === "function")
            for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
                if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                    t[p[i]] = s[p[i]];
            }
        return t;
    };
    var __extends = (this && this.__extends) || (function () {
        var extendStatics = function (d, b) {
            extendStatics = Object.setPrototypeOf ||
                ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
                function (d, b) { for (var p in b)
                    if (Object.prototype.hasOwnProperty.call(b, p))
                        d[p] = b[p]; };
            return extendStatics(d, b);
        };
        return function (d, b) {
            if (typeof b !== "function" && b !== null)
                throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
            extendStatics(d, b);
            function __() { this.constructor = d; }
            d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
        };
    })();
    /* @internal */
    var ts;
    (function (ts) {
        function createMapData() {
            var sentinel = {};
            sentinel.prev = sentinel;
            return { head: sentinel, tail: sentinel, size: 0 };
        }
        function createMapEntry(key, value) {
            return { key: key, value: value, next: undefined, prev: undefined };
        }
        function sameValueZero(x, y) {
            // Treats -0 === 0 and NaN === NaN
            return x === y || x !== x && y !== y;
        }
        function getPrev(entry) {
            var prev = entry.prev;
            // Entries without a 'prev' have been removed from the map.
            // An entry whose 'prev' points to itself is the head of the list and is invalid here.
            if (!prev || prev === entry)
                throw new Error("Illegal state");
            return prev;
        }
        function getNext(entry) {
            while (entry) {
                // Entries without a 'prev' have been removed from the map. Their 'next'
                // pointer should point to the previous entry prior to deletion and
                // that entry should be skipped to resume iteration.
                var skipNext = !entry.prev;
                entry = entry.next;
                if (skipNext) {
                    continue;
                }
                return entry;
            }
        }
        function getEntry(data, key) {
            // We walk backwards from 'tail' to prioritize recently added entries.
            // We skip 'head' because it is an empty entry used to track iteration start.
            for (var entry = data.tail; entry !== data.head; entry = getPrev(entry)) {
                if (sameValueZero(entry.key, key)) {
                    return entry;
                }
            }
        }
        function addOrUpdateEntry(data, key, value) {
            var existing = getEntry(data, key);
            if (existing) {
                existing.value = value;
                return;
            }
            var entry = createMapEntry(key, value);
            entry.prev = data.tail;
            data.tail.next = entry;
            data.tail = entry;
            data.size++;
            return entry;
        }
        function deleteEntry(data, key) {
            // We walk backwards from 'tail' to prioritize recently added entries.
            // We skip 'head' because it is an empty entry used to track iteration start.
            for (var entry = data.tail; entry !== data.head; entry = getPrev(entry)) {
                // all entries in the map should have a 'prev' pointer.
                if (entry.prev === undefined)
                    throw new Error("Illegal state");
                if (sameValueZero(entry.key, key)) {
                    if (entry.next) {
                        entry.next.prev = entry.prev;
                    }
                    else {
                        // an entry in the map without a 'next' pointer must be the 'tail'.
                        if (data.tail !== entry)
                            throw new Error("Illegal state");
                        data.tail = entry.prev;
                    }
                    entry.prev.next = entry.next;
                    entry.next = entry.prev;
                    entry.prev = undefined;
                    data.size--;
                    return entry;
                }
            }
        }
        function clearEntries(data) {
            var node = data.tail;
            while (node !== data.head) {
                var prev = getPrev(node);
                node.next = data.head;
                node.prev = undefined;
                node = prev;
            }
            data.head.next = undefined;
            data.tail = data.head;
            data.size = 0;
        }
        function forEachEntry(data, action) {
            var entry = data.head;
            while (entry) {
                entry = getNext(entry);
                if (entry) {
                    action(entry.value, entry.key);
                }
            }
        }
        function forEachIteration(iterator, action) {
            if (iterator) {
                for (var step = iterator.next(); !step.done; step = iterator.next()) {
                    action(step.value);
                }
            }
        }
        function createIteratorData(data, selector) {
            return { current: data.head, selector: selector };
        }
        function iteratorNext(data) {
            // Navigate to the next entry.
            data.current = getNext(data.current);
            if (data.current) {
                return { value: data.selector(data.current.key, data.current.value), done: false };
            }
            else {
                return { value: undefined, done: true };
            }
        }
        /* @internal */
        var ShimCollections;
        (function (ShimCollections) {
            function createMapShim(getIterator) {
                var MapIterator = /** @class */ (function () {
                    function MapIterator(data, selector) {
                        this._data = createIteratorData(data, selector);
                    }
                    MapIterator.prototype.next = function () { return iteratorNext(this._data); };
                    return MapIterator;
                }());
                return /** @class */ (function () {
                    function Map(iterable) {
                        var _this = this;
                        this._mapData = createMapData();
                        forEachIteration(getIterator(iterable), function (_a) {
                            var key = _a[0], value = _a[1];
                            return _this.set(key, value);
                        });
                    }
                    Object.defineProperty(Map.prototype, "size", {
                        get: function () { return this._mapData.size; },
                        enumerable: false,
                        configurable: true
                    });
                    Map.prototype.get = function (key) { var _a; return (_a = getEntry(this._mapData, key)) === null || _a === void 0 ? void 0 : _a.value; };
                    Map.prototype.set = function (key, value) { return addOrUpdateEntry(this._mapData, key, value), this; };
                    Map.prototype.has = function (key) { return !!getEntry(this._mapData, key); };
                    Map.prototype.delete = function (key) { return !!deleteEntry(this._mapData, key); };
                    Map.prototype.clear = function () { clearEntries(this._mapData); };
                    Map.prototype.keys = function () { return new MapIterator(this._mapData, function (key, _value) { return key; }); };
                    Map.prototype.values = function () { return new MapIterator(this._mapData, function (_key, value) { return value; }); };
                    Map.prototype.entries = function () { return new MapIterator(this._mapData, function (key, value) { return [key, value]; }); };
                    Map.prototype.forEach = function (action) { forEachEntry(this._mapData, action); };
                    return Map;
                }());
            }
            ShimCollections.createMapShim = createMapShim;
            function createSetShim(getIterator) {
                var SetIterator = /** @class */ (function () {
                    function SetIterator(data, selector) {
                        this._data = createIteratorData(data, selector);
                    }
                    SetIterator.prototype.next = function () { return iteratorNext(this._data); };
                    return SetIterator;
                }());
                return /** @class */ (function () {
                    function Set(iterable) {
                        var _this = this;
                        this._mapData = createMapData();
                        forEachIteration(getIterator(iterable), function (value) { return _this.add(value); });
                    }
                    Object.defineProperty(Set.prototype, "size", {
                        get: function () { return this._mapData.size; },
                        enumerable: false,
                        configurable: true
                    });
                    Set.prototype.add = function (value) { return addOrUpdateEntry(this._mapData, value, value), this; };
                    Set.prototype.has = function (value) { return !!getEntry(this._mapData, value); };
                    Set.prototype.delete = function (value) { return !!deleteEntry(this._mapData, value); };
                    Set.prototype.clear = function () { clearEntries(this._mapData); };
                    Set.prototype.keys = function () { return new SetIterator(this._mapData, function (key, _value) { return key; }); };
                    Set.prototype.values = function () { return new SetIterator(this._mapData, function (_key, value) { return value; }); };
                    Set.prototype.entries = function () { return new SetIterator(this._mapData, function (key, value) { return [key, value]; }); };
                    Set.prototype.forEach = function (action) { forEachEntry(this._mapData, action); };
                    return Set;
                }());
            }
            ShimCollections.createSetShim = createSetShim;
        })(ShimCollections = ts.ShimCollections || (ts.ShimCollections = {}));
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        // WARNING: The script `configurePrerelease.ts` uses a regexp to parse out these values.
        // If changing the text in this section, be sure to test `configurePrerelease` too.
        ts.versionMajorMinor = "4.4";
        // The following is baselined as a literal template type without intervention
        /** The version of the TypeScript compiler release */
        // eslint-disable-next-line @typescript-eslint/no-inferrable-types
        ts.version = "4.4.1-rc";
        /* @internal */
        var Comparison;
        (function (Comparison) {
            Comparison[Comparison["LessThan"] = -1] = "LessThan";
            Comparison[Comparison["EqualTo"] = 0] = "EqualTo";
            Comparison[Comparison["GreaterThan"] = 1] = "GreaterThan";
        })(Comparison = ts.Comparison || (ts.Comparison = {}));
        /* @internal */
        var NativeCollections;
        (function (NativeCollections) {
            /**
             * Returns the native Map implementation if it is available and compatible (i.e. supports iteration).
             */
            function tryGetNativeMap() {
                // Internet Explorer's Map doesn't support iteration, so don't use it.
                // eslint-disable-next-line no-in-operator
                return typeof Map !== "undefined" && "entries" in Map.prototype && new Map([[0, 0]]).size === 1 ? Map : undefined;
            }
            NativeCollections.tryGetNativeMap = tryGetNativeMap;
            /**
             * Returns the native Set implementation if it is available and compatible (i.e. supports iteration).
             */
            function tryGetNativeSet() {
                // Internet Explorer's Set doesn't support iteration, so don't use it.
                // eslint-disable-next-line no-in-operator
                return typeof Set !== "undefined" && "entries" in Set.prototype && new Set([0]).size === 1 ? Set : undefined;
            }
            NativeCollections.tryGetNativeSet = tryGetNativeSet;
        })(NativeCollections = ts.NativeCollections || (ts.NativeCollections = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function getCollectionImplementation(name, nativeFactory, shimFactory) {
            var _a;
            // NOTE: ts.ShimCollections will be defined for typescriptServices.js but not for tsc.js, so we must test for it.
            var constructor = (_a = ts.NativeCollections[nativeFactory]()) !== null && _a !== void 0 ? _a : ts.ShimCollections === null || ts.ShimCollections === void 0 ? void 0 : ts.ShimCollections[shimFactory](getIterator);
            if (constructor)
                return constructor;
            throw new Error("TypeScript requires an environment that provides a compatible native " + name + " implementation.");
        }
        ts.Map = getCollectionImplementation("Map", "tryGetNativeMap", "createMapShim");
        ts.Set = getCollectionImplementation("Set", "tryGetNativeSet", "createSetShim");
        function getIterator(iterable) {
            if (iterable) {
                if (isArray(iterable))
                    return arrayIterator(iterable);
                if (iterable instanceof ts.Map)
                    return iterable.entries();
                if (iterable instanceof ts.Set)
                    return iterable.values();
                throw new Error("Iteration not supported.");
            }
        }
        ts.getIterator = getIterator;
        ts.emptyArray = [];
        ts.emptyMap = new ts.Map();
        ts.emptySet = new ts.Set();
        function createMap() {
            return new ts.Map();
        }
        ts.createMap = createMap;
        /**
         * Create a new map from a template object is provided, the map will copy entries from it.
         * @deprecated Use `new Map(getEntries(template))` instead.
         */
        function createMapFromTemplate(template) {
            var map = new ts.Map();
            // Copies keys/values from template. Note that for..in will not throw if
            // template is undefined, and instead will just exit the loop.
            for (var key in template) {
                if (hasOwnProperty.call(template, key)) {
                    map.set(key, template[key]);
                }
            }
            return map;
        }
        ts.createMapFromTemplate = createMapFromTemplate;
        function length(array) {
            return array ? array.length : 0;
        }
        ts.length = length;
        /**
         * Iterates through 'array' by index and performs the callback on each element of array until the callback
         * returns a truthy value, then returns that value.
         * If no such value is found, the callback is applied to each element of array and undefined is returned.
         */
        function forEach(array, callback) {
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    var result = callback(array[i], i);
                    if (result) {
                        return result;
                    }
                }
            }
            return undefined;
        }
        ts.forEach = forEach;
        /**
         * Like `forEach`, but iterates in reverse order.
         */
        function forEachRight(array, callback) {
            if (array) {
                for (var i = array.length - 1; i >= 0; i--) {
                    var result = callback(array[i], i);
                    if (result) {
                        return result;
                    }
                }
            }
            return undefined;
        }
        ts.forEachRight = forEachRight;
        /** Like `forEach`, but suitable for use with numbers and strings (which may be falsy). */
        function firstDefined(array, callback) {
            if (array === undefined) {
                return undefined;
            }
            for (var i = 0; i < array.length; i++) {
                var result = callback(array[i], i);
                if (result !== undefined) {
                    return result;
                }
            }
            return undefined;
        }
        ts.firstDefined = firstDefined;
        function firstDefinedIterator(iter, callback) {
            while (true) {
                var iterResult = iter.next();
                if (iterResult.done) {
                    return undefined;
                }
                var result = callback(iterResult.value);
                if (result !== undefined) {
                    return result;
                }
            }
        }
        ts.firstDefinedIterator = firstDefinedIterator;
        function reduceLeftIterator(iterator, f, initial) {
            var result = initial;
            if (iterator) {
                for (var step = iterator.next(), pos = 0; !step.done; step = iterator.next(), pos++) {
                    result = f(result, step.value, pos);
                }
            }
            return result;
        }
        ts.reduceLeftIterator = reduceLeftIterator;
        function zipWith(arrayA, arrayB, callback) {
            var result = [];
            ts.Debug.assertEqual(arrayA.length, arrayB.length);
            for (var i = 0; i < arrayA.length; i++) {
                result.push(callback(arrayA[i], arrayB[i], i));
            }
            return result;
        }
        ts.zipWith = zipWith;
        function zipToIterator(arrayA, arrayB) {
            ts.Debug.assertEqual(arrayA.length, arrayB.length);
            var i = 0;
            return {
                next: function () {
                    if (i === arrayA.length) {
                        return { value: undefined, done: true };
                    }
                    i++;
                    return { value: [arrayA[i - 1], arrayB[i - 1]], done: false };
                }
            };
        }
        ts.zipToIterator = zipToIterator;
        function zipToMap(keys, values) {
            ts.Debug.assert(keys.length === values.length);
            var map = new ts.Map();
            for (var i = 0; i < keys.length; ++i) {
                map.set(keys[i], values[i]);
            }
            return map;
        }
        ts.zipToMap = zipToMap;
        /**
         * Creates a new array with `element` interspersed in between each element of `input`
         * if there is more than 1 value in `input`. Otherwise, returns the existing array.
         */
        function intersperse(input, element) {
            if (input.length <= 1) {
                return input;
            }
            var result = [];
            for (var i = 0, n = input.length; i < n; i++) {
                if (i)
                    result.push(element);
                result.push(input[i]);
            }
            return result;
        }
        ts.intersperse = intersperse;
        /**
         * Iterates through `array` by index and performs the callback on each element of array until the callback
         * returns a falsey value, then returns false.
         * If no such value is found, the callback is applied to each element of array and `true` is returned.
         */
        function every(array, callback) {
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    if (!callback(array[i], i)) {
                        return false;
                    }
                }
            }
            return true;
        }
        ts.every = every;
        function find(array, predicate) {
            for (var i = 0; i < array.length; i++) {
                var value = array[i];
                if (predicate(value, i)) {
                    return value;
                }
            }
            return undefined;
        }
        ts.find = find;
        function findLast(array, predicate) {
            for (var i = array.length - 1; i >= 0; i--) {
                var value = array[i];
                if (predicate(value, i)) {
                    return value;
                }
            }
            return undefined;
        }
        ts.findLast = findLast;
        /** Works like Array.prototype.findIndex, returning `-1` if no element satisfying the predicate is found. */
        function findIndex(array, predicate, startIndex) {
            for (var i = startIndex || 0; i < array.length; i++) {
                if (predicate(array[i], i)) {
                    return i;
                }
            }
            return -1;
        }
        ts.findIndex = findIndex;
        function findLastIndex(array, predicate, startIndex) {
            for (var i = startIndex === undefined ? array.length - 1 : startIndex; i >= 0; i--) {
                if (predicate(array[i], i)) {
                    return i;
                }
            }
            return -1;
        }
        ts.findLastIndex = findLastIndex;
        /**
         * Returns the first truthy result of `callback`, or else fails.
         * This is like `forEach`, but never returns undefined.
         */
        function findMap(array, callback) {
            for (var i = 0; i < array.length; i++) {
                var result = callback(array[i], i);
                if (result) {
                    return result;
                }
            }
            return ts.Debug.fail();
        }
        ts.findMap = findMap;
        function contains(array, value, equalityComparer) {
            if (equalityComparer === void 0) {
                equalityComparer = equateValues;
            }
            if (array) {
                for (var _i = 0, array_1 = array; _i < array_1.length; _i++) {
                    var v = array_1[_i];
                    if (equalityComparer(v, value)) {
                        return true;
                    }
                }
            }
            return false;
        }
        ts.contains = contains;
        function arraysEqual(a, b, equalityComparer) {
            if (equalityComparer === void 0) {
                equalityComparer = equateValues;
            }
            return a.length === b.length && a.every(function (x, i) { return equalityComparer(x, b[i]); });
        }
        ts.arraysEqual = arraysEqual;
        function indexOfAnyCharCode(text, charCodes, start) {
            for (var i = start || 0; i < text.length; i++) {
                if (contains(charCodes, text.charCodeAt(i))) {
                    return i;
                }
            }
            return -1;
        }
        ts.indexOfAnyCharCode = indexOfAnyCharCode;
        function countWhere(array, predicate) {
            var count = 0;
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    var v = array[i];
                    if (predicate(v, i)) {
                        count++;
                    }
                }
            }
            return count;
        }
        ts.countWhere = countWhere;
        function filter(array, f) {
            if (array) {
                var len = array.length;
                var i = 0;
                while (i < len && f(array[i]))
                    i++;
                if (i < len) {
                    var result = array.slice(0, i);
                    i++;
                    while (i < len) {
                        var item = array[i];
                        if (f(item)) {
                            result.push(item);
                        }
                        i++;
                    }
                    return result;
                }
            }
            return array;
        }
        ts.filter = filter;
        function filterMutate(array, f) {
            var outIndex = 0;
            for (var i = 0; i < array.length; i++) {
                if (f(array[i], i, array)) {
                    array[outIndex] = array[i];
                    outIndex++;
                }
            }
            array.length = outIndex;
        }
        ts.filterMutate = filterMutate;
        function clear(array) {
            array.length = 0;
        }
        ts.clear = clear;
        function map(array, f) {
            var result;
            if (array) {
                result = [];
                for (var i = 0; i < array.length; i++) {
                    result.push(f(array[i], i));
                }
            }
            return result;
        }
        ts.map = map;
        function mapIterator(iter, mapFn) {
            return {
                next: function () {
                    var iterRes = iter.next();
                    return iterRes.done ? iterRes : { value: mapFn(iterRes.value), done: false };
                }
            };
        }
        ts.mapIterator = mapIterator;
        function sameMap(array, f) {
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    var item = array[i];
                    var mapped = f(item, i);
                    if (item !== mapped) {
                        var result = array.slice(0, i);
                        result.push(mapped);
                        for (i++; i < array.length; i++) {
                            result.push(f(array[i], i));
                        }
                        return result;
                    }
                }
            }
            return array;
        }
        ts.sameMap = sameMap;
        /**
         * Flattens an array containing a mix of array or non-array elements.
         *
         * @param array The array to flatten.
         */
        function flatten(array) {
            var result = [];
            for (var _i = 0, array_2 = array; _i < array_2.length; _i++) {
                var v = array_2[_i];
                if (v) {
                    if (isArray(v)) {
                        addRange(result, v);
                    }
                    else {
                        result.push(v);
                    }
                }
            }
            return result;
        }
        ts.flatten = flatten;
        /**
         * Maps an array. If the mapped value is an array, it is spread into the result.
         *
         * @param array The array to map.
         * @param mapfn The callback used to map the result into one or more values.
         */
        function flatMap(array, mapfn) {
            var result;
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    var v = mapfn(array[i], i);
                    if (v) {
                        if (isArray(v)) {
                            result = addRange(result, v);
                        }
                        else {
                            result = append(result, v);
                        }
                    }
                }
            }
            return result || ts.emptyArray;
        }
        ts.flatMap = flatMap;
        function flatMapToMutable(array, mapfn) {
            var result = [];
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    var v = mapfn(array[i], i);
                    if (v) {
                        if (isArray(v)) {
                            addRange(result, v);
                        }
                        else {
                            result.push(v);
                        }
                    }
                }
            }
            return result;
        }
        ts.flatMapToMutable = flatMapToMutable;
        function flatMapIterator(iter, mapfn) {
            var first = iter.next();
            if (first.done) {
                return ts.emptyIterator;
            }
            var currentIter = getIterator(first.value);
            return {
                next: function () {
                    while (true) {
                        var currentRes = currentIter.next();
                        if (!currentRes.done) {
                            return currentRes;
                        }
                        var iterRes = iter.next();
                        if (iterRes.done) {
                            return iterRes;
                        }
                        currentIter = getIterator(iterRes.value);
                    }
                },
            };
            function getIterator(x) {
                var res = mapfn(x);
                return res === undefined ? ts.emptyIterator : isArray(res) ? arrayIterator(res) : res;
            }
        }
        ts.flatMapIterator = flatMapIterator;
        function sameFlatMap(array, mapfn) {
            var result;
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    var item = array[i];
                    var mapped = mapfn(item, i);
                    if (result || item !== mapped || isArray(mapped)) {
                        if (!result) {
                            result = array.slice(0, i);
                        }
                        if (isArray(mapped)) {
                            addRange(result, mapped);
                        }
                        else {
                            result.push(mapped);
                        }
                    }
                }
            }
            return result || array;
        }
        ts.sameFlatMap = sameFlatMap;
        function mapAllOrFail(array, mapFn) {
            var result = [];
            for (var i = 0; i < array.length; i++) {
                var mapped = mapFn(array[i], i);
                if (mapped === undefined) {
                    return undefined;
                }
                result.push(mapped);
            }
            return result;
        }
        ts.mapAllOrFail = mapAllOrFail;
        function mapDefined(array, mapFn) {
            var result = [];
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    var mapped = mapFn(array[i], i);
                    if (mapped !== undefined) {
                        result.push(mapped);
                    }
                }
            }
            return result;
        }
        ts.mapDefined = mapDefined;
        function mapDefinedIterator(iter, mapFn) {
            return {
                next: function () {
                    while (true) {
                        var res = iter.next();
                        if (res.done) {
                            return res;
                        }
                        var value = mapFn(res.value);
                        if (value !== undefined) {
                            return { value: value, done: false };
                        }
                    }
                }
            };
        }
        ts.mapDefinedIterator = mapDefinedIterator;
        function mapDefinedEntries(map, f) {
            if (!map) {
                return undefined;
            }
            var result = new ts.Map();
            map.forEach(function (value, key) {
                var entry = f(key, value);
                if (entry !== undefined) {
                    var newKey = entry[0], newValue = entry[1];
                    if (newKey !== undefined && newValue !== undefined) {
                        result.set(newKey, newValue);
                    }
                }
            });
            return result;
        }
        ts.mapDefinedEntries = mapDefinedEntries;
        function mapDefinedValues(set, f) {
            if (set) {
                var result_1 = new ts.Set();
                set.forEach(function (value) {
                    var newValue = f(value);
                    if (newValue !== undefined) {
                        result_1.add(newValue);
                    }
                });
                return result_1;
            }
        }
        ts.mapDefinedValues = mapDefinedValues;
        function getOrUpdate(map, key, callback) {
            if (map.has(key)) {
                return map.get(key);
            }
            var value = callback();
            map.set(key, value);
            return value;
        }
        ts.getOrUpdate = getOrUpdate;
        function tryAddToSet(set, value) {
            if (!set.has(value)) {
                set.add(value);
                return true;
            }
            return false;
        }
        ts.tryAddToSet = tryAddToSet;
        ts.emptyIterator = { next: function () { return ({ value: undefined, done: true }); } };
        function singleIterator(value) {
            var done = false;
            return {
                next: function () {
                    var wasDone = done;
                    done = true;
                    return wasDone ? { value: undefined, done: true } : { value: value, done: false };
                }
            };
        }
        ts.singleIterator = singleIterator;
        function spanMap(array, keyfn, mapfn) {
            var result;
            if (array) {
                result = [];
                var len = array.length;
                var previousKey = void 0;
                var key = void 0;
                var start = 0;
                var pos = 0;
                while (start < len) {
                    while (pos < len) {
                        var value = array[pos];
                        key = keyfn(value, pos);
                        if (pos === 0) {
                            previousKey = key;
                        }
                        else if (key !== previousKey) {
                            break;
                        }
                        pos++;
                    }
                    if (start < pos) {
                        var v = mapfn(array.slice(start, pos), previousKey, start, pos);
                        if (v) {
                            result.push(v);
                        }
                        start = pos;
                    }
                    previousKey = key;
                    pos++;
                }
            }
            return result;
        }
        ts.spanMap = spanMap;
        function mapEntries(map, f) {
            if (!map) {
                return undefined;
            }
            var result = new ts.Map();
            map.forEach(function (value, key) {
                var _a = f(key, value), newKey = _a[0], newValue = _a[1];
                result.set(newKey, newValue);
            });
            return result;
        }
        ts.mapEntries = mapEntries;
        function some(array, predicate) {
            if (array) {
                if (predicate) {
                    for (var _i = 0, array_3 = array; _i < array_3.length; _i++) {
                        var v = array_3[_i];
                        if (predicate(v)) {
                            return true;
                        }
                    }
                }
                else {
                    return array.length > 0;
                }
            }
            return false;
        }
        ts.some = some;
        /** Calls the callback with (start, afterEnd) index pairs for each range where 'pred' is true. */
        function getRangesWhere(arr, pred, cb) {
            var start;
            for (var i = 0; i < arr.length; i++) {
                if (pred(arr[i])) {
                    start = start === undefined ? i : start;
                }
                else {
                    if (start !== undefined) {
                        cb(start, i);
                        start = undefined;
                    }
                }
            }
            if (start !== undefined)
                cb(start, arr.length);
        }
        ts.getRangesWhere = getRangesWhere;
        function concatenate(array1, array2) {
            if (!some(array2))
                return array1;
            if (!some(array1))
                return array2;
            return __spreadArray(__spreadArray([], array1, true), array2, true);
        }
        ts.concatenate = concatenate;
        function selectIndex(_, i) {
            return i;
        }
        function indicesOf(array) {
            return array.map(selectIndex);
        }
        ts.indicesOf = indicesOf;
        function deduplicateRelational(array, equalityComparer, comparer) {
            // Perform a stable sort of the array. This ensures the first entry in a list of
            // duplicates remains the first entry in the result.
            var indices = indicesOf(array);
            stableSortIndices(array, indices, comparer);
            var last = array[indices[0]];
            var deduplicated = [indices[0]];
            for (var i = 1; i < indices.length; i++) {
                var index = indices[i];
                var item = array[index];
                if (!equalityComparer(last, item)) {
                    deduplicated.push(index);
                    last = item;
                }
            }
            // restore original order
            deduplicated.sort();
            return deduplicated.map(function (i) { return array[i]; });
        }
        function deduplicateEquality(array, equalityComparer) {
            var result = [];
            for (var _i = 0, array_4 = array; _i < array_4.length; _i++) {
                var item = array_4[_i];
                pushIfUnique(result, item, equalityComparer);
            }
            return result;
        }
        /**
         * Deduplicates an unsorted array.
         * @param equalityComparer An `EqualityComparer` used to determine if two values are duplicates.
         * @param comparer An optional `Comparer` used to sort entries before comparison, though the
         * result will remain in the original order in `array`.
         */
        function deduplicate(array, equalityComparer, comparer) {
            return array.length === 0 ? [] :
                array.length === 1 ? array.slice() :
                    comparer ? deduplicateRelational(array, equalityComparer, comparer) :
                        deduplicateEquality(array, equalityComparer);
        }
        ts.deduplicate = deduplicate;
        /**
         * Deduplicates an array that has already been sorted.
         */
        function deduplicateSorted(array, comparer) {
            if (array.length === 0)
                return ts.emptyArray;
            var last = array[0];
            var deduplicated = [last];
            for (var i = 1; i < array.length; i++) {
                var next = array[i];
                switch (comparer(next, last)) {
                    // equality comparison
                    case true:
                    // relational comparison
                    // falls through
                    case 0 /* EqualTo */:
                        continue;
                    case -1 /* LessThan */:
                        // If `array` is sorted, `next` should **never** be less than `last`.
                        return ts.Debug.fail("Array is unsorted.");
                }
                deduplicated.push(last = next);
            }
            return deduplicated;
        }
        function insertSorted(array, insert, compare) {
            if (array.length === 0) {
                array.push(insert);
                return;
            }
            var insertIndex = binarySearch(array, insert, identity, compare);
            if (insertIndex < 0) {
                array.splice(~insertIndex, 0, insert);
            }
        }
        ts.insertSorted = insertSorted;
        function sortAndDeduplicate(array, comparer, equalityComparer) {
            return deduplicateSorted(sort(array, comparer), equalityComparer || comparer || compareStringsCaseSensitive);
        }
        ts.sortAndDeduplicate = sortAndDeduplicate;
        function arrayIsSorted(array, comparer) {
            if (array.length < 2)
                return true;
            var prevElement = array[0];
            for (var _i = 0, _a = array.slice(1); _i < _a.length; _i++) {
                var element = _a[_i];
                if (comparer(prevElement, element) === 1 /* GreaterThan */) {
                    return false;
                }
                prevElement = element;
            }
            return true;
        }
        ts.arrayIsSorted = arrayIsSorted;
        function arrayIsEqualTo(array1, array2, equalityComparer) {
            if (equalityComparer === void 0) {
                equalityComparer = equateValues;
            }
            if (!array1 || !array2) {
                return array1 === array2;
            }
            if (array1.length !== array2.length) {
                return false;
            }
            for (var i = 0; i < array1.length; i++) {
                if (!equalityComparer(array1[i], array2[i], i)) {
                    return false;
                }
            }
            return true;
        }
        ts.arrayIsEqualTo = arrayIsEqualTo;
        function compact(array) {
            var result;
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    var v = array[i];
                    if (result || !v) {
                        if (!result) {
                            result = array.slice(0, i);
                        }
                        if (v) {
                            result.push(v);
                        }
                    }
                }
            }
            return result || array;
        }
        ts.compact = compact;
        /**
         * Gets the relative complement of `arrayA` with respect to `arrayB`, returning the elements that
         * are not present in `arrayA` but are present in `arrayB`. Assumes both arrays are sorted
         * based on the provided comparer.
         */
        function relativeComplement(arrayA, arrayB, comparer) {
            if (!arrayB || !arrayA || arrayB.length === 0 || arrayA.length === 0)
                return arrayB;
            var result = [];
            loopB: for (var offsetA = 0, offsetB = 0; offsetB < arrayB.length; offsetB++) {
                if (offsetB > 0) {
                    // Ensure `arrayB` is properly sorted.
                    ts.Debug.assertGreaterThanOrEqual(comparer(arrayB[offsetB], arrayB[offsetB - 1]), 0 /* EqualTo */);
                }
                loopA: for (var startA = offsetA; offsetA < arrayA.length; offsetA++) {
                    if (offsetA > startA) {
                        // Ensure `arrayA` is properly sorted. We only need to perform this check if
                        // `offsetA` has changed since we entered the loop.
                        ts.Debug.assertGreaterThanOrEqual(comparer(arrayA[offsetA], arrayA[offsetA - 1]), 0 /* EqualTo */);
                    }
                    switch (comparer(arrayB[offsetB], arrayA[offsetA])) {
                        case -1 /* LessThan */:
                            // If B is less than A, B does not exist in arrayA. Add B to the result and
                            // move to the next element in arrayB without changing the current position
                            // in arrayA.
                            result.push(arrayB[offsetB]);
                            continue loopB;
                        case 0 /* EqualTo */:
                            // If B is equal to A, B exists in arrayA. Move to the next element in
                            // arrayB without adding B to the result or changing the current position
                            // in arrayA.
                            continue loopB;
                        case 1 /* GreaterThan */:
                            // If B is greater than A, we need to keep looking for B in arrayA. Move to
                            // the next element in arrayA and recheck.
                            continue loopA;
                    }
                }
            }
            return result;
        }
        ts.relativeComplement = relativeComplement;
        function sum(array, prop) {
            var result = 0;
            for (var _i = 0, array_5 = array; _i < array_5.length; _i++) {
                var v = array_5[_i];
                result += v[prop];
            }
            return result;
        }
        ts.sum = sum;
        function append(to, value) {
            if (value === undefined)
                return to;
            if (to === undefined)
                return [value];
            to.push(value);
            return to;
        }
        ts.append = append;
        function combine(xs, ys) {
            if (xs === undefined)
                return ys;
            if (ys === undefined)
                return xs;
            if (isArray(xs))
                return isArray(ys) ? concatenate(xs, ys) : append(xs, ys);
            if (isArray(ys))
                return append(ys, xs);
            return [xs, ys];
        }
        ts.combine = combine;
        /**
         * Gets the actual offset into an array for a relative offset. Negative offsets indicate a
         * position offset from the end of the array.
         */
        function toOffset(array, offset) {
            return offset < 0 ? array.length + offset : offset;
        }
        function addRange(to, from, start, end) {
            if (from === undefined || from.length === 0)
                return to;
            if (to === undefined)
                return from.slice(start, end);
            start = start === undefined ? 0 : toOffset(from, start);
            end = end === undefined ? from.length : toOffset(from, end);
            for (var i = start; i < end && i < from.length; i++) {
                if (from[i] !== undefined) {
                    to.push(from[i]);
                }
            }
            return to;
        }
        ts.addRange = addRange;
        /**
         * @return Whether the value was added.
         */
        function pushIfUnique(array, toAdd, equalityComparer) {
            if (contains(array, toAdd, equalityComparer)) {
                return false;
            }
            else {
                array.push(toAdd);
                return true;
            }
        }
        ts.pushIfUnique = pushIfUnique;
        /**
         * Unlike `pushIfUnique`, this can take `undefined` as an input, and returns a new array.
         */
        function appendIfUnique(array, toAdd, equalityComparer) {
            if (array) {
                pushIfUnique(array, toAdd, equalityComparer);
                return array;
            }
            else {
                return [toAdd];
            }
        }
        ts.appendIfUnique = appendIfUnique;
        function stableSortIndices(array, indices, comparer) {
            // sort indices by value then position
            indices.sort(function (x, y) { return comparer(array[x], array[y]) || compareValues(x, y); });
        }
        /**
         * Returns a new sorted array.
         */
        function sort(array, comparer) {
            return (array.length === 0 ? array : array.slice().sort(comparer));
        }
        ts.sort = sort;
        function arrayIterator(array) {
            var i = 0;
            return { next: function () {
                    if (i === array.length) {
                        return { value: undefined, done: true };
                    }
                    else {
                        i++;
                        return { value: array[i - 1], done: false };
                    }
                } };
        }
        ts.arrayIterator = arrayIterator;
        function arrayReverseIterator(array) {
            var i = array.length;
            return {
                next: function () {
                    if (i === 0) {
                        return { value: undefined, done: true };
                    }
                    else {
                        i--;
                        return { value: array[i], done: false };
                    }
                }
            };
        }
        ts.arrayReverseIterator = arrayReverseIterator;
        /**
         * Stable sort of an array. Elements equal to each other maintain their relative position in the array.
         */
        function stableSort(array, comparer) {
            var indices = indicesOf(array);
            stableSortIndices(array, indices, comparer);
            return indices.map(function (i) { return array[i]; });
        }
        ts.stableSort = stableSort;
        function rangeEquals(array1, array2, pos, end) {
            while (pos < end) {
                if (array1[pos] !== array2[pos]) {
                    return false;
                }
                pos++;
            }
            return true;
        }
        ts.rangeEquals = rangeEquals;
        /**
         * Returns the element at a specific offset in an array if non-empty, `undefined` otherwise.
         * A negative offset indicates the element should be retrieved from the end of the array.
         */
        function elementAt(array, offset) {
            if (array) {
                offset = toOffset(array, offset);
                if (offset < array.length) {
                    return array[offset];
                }
            }
            return undefined;
        }
        ts.elementAt = elementAt;
        /**
         * Returns the first element of an array if non-empty, `undefined` otherwise.
         */
        function firstOrUndefined(array) {
            return array.length === 0 ? undefined : array[0];
        }
        ts.firstOrUndefined = firstOrUndefined;
        function first(array) {
            ts.Debug.assert(array.length !== 0);
            return array[0];
        }
        ts.first = first;
        /**
         * Returns the last element of an array if non-empty, `undefined` otherwise.
         */
        function lastOrUndefined(array) {
            return array.length === 0 ? undefined : array[array.length - 1];
        }
        ts.lastOrUndefined = lastOrUndefined;
        function last(array) {
            ts.Debug.assert(array.length !== 0);
            return array[array.length - 1];
        }
        ts.last = last;
        /**
         * Returns the only element of an array if it contains only one element, `undefined` otherwise.
         */
        function singleOrUndefined(array) {
            return array && array.length === 1
                ? array[0]
                : undefined;
        }
        ts.singleOrUndefined = singleOrUndefined;
        function singleOrMany(array) {
            return array && array.length === 1
                ? array[0]
                : array;
        }
        ts.singleOrMany = singleOrMany;
        function replaceElement(array, index, value) {
            var result = array.slice(0);
            result[index] = value;
            return result;
        }
        ts.replaceElement = replaceElement;
        /**
         * Performs a binary search, finding the index at which `value` occurs in `array`.
         * If no such index is found, returns the 2's-complement of first index at which
         * `array[index]` exceeds `value`.
         * @param array A sorted array whose first element must be no larger than number
         * @param value The value to be searched for in the array.
         * @param keySelector A callback used to select the search key from `value` and each element of
         * `array`.
         * @param keyComparer A callback used to compare two keys in a sorted array.
         * @param offset An offset into `array` at which to start the search.
         */
        function binarySearch(array, value, keySelector, keyComparer, offset) {
            return binarySearchKey(array, keySelector(value), keySelector, keyComparer, offset);
        }
        ts.binarySearch = binarySearch;
        /**
         * Performs a binary search, finding the index at which an object with `key` occurs in `array`.
         * If no such index is found, returns the 2's-complement of first index at which
         * `array[index]` exceeds `key`.
         * @param array A sorted array whose first element must be no larger than number
         * @param key The key to be searched for in the array.
         * @param keySelector A callback used to select the search key from each element of `array`.
         * @param keyComparer A callback used to compare two keys in a sorted array.
         * @param offset An offset into `array` at which to start the search.
         */
        function binarySearchKey(array, key, keySelector, keyComparer, offset) {
            if (!some(array)) {
                return -1;
            }
            var low = offset || 0;
            var high = array.length - 1;
            while (low <= high) {
                var middle = low + ((high - low) >> 1);
                var midKey = keySelector(array[middle], middle);
                switch (keyComparer(midKey, key)) {
                    case -1 /* LessThan */:
                        low = middle + 1;
                        break;
                    case 0 /* EqualTo */:
                        return middle;
                    case 1 /* GreaterThan */:
                        high = middle - 1;
                        break;
                }
            }
            return ~low;
        }
        ts.binarySearchKey = binarySearchKey;
        function reduceLeft(array, f, initial, start, count) {
            if (array && array.length > 0) {
                var size = array.length;
                if (size > 0) {
                    var pos = start === undefined || start < 0 ? 0 : start;
                    var end = count === undefined || pos + count > size - 1 ? size - 1 : pos + count;
                    var result = void 0;
                    if (arguments.length <= 2) {
                        result = array[pos];
                        pos++;
                    }
                    else {
                        result = initial;
                    }
                    while (pos <= end) {
                        result = f(result, array[pos], pos);
                        pos++;
                    }
                    return result;
                }
            }
            return initial;
        }
        ts.reduceLeft = reduceLeft;
        var hasOwnProperty = Object.prototype.hasOwnProperty;
        /**
         * Indicates whether a map-like contains an own property with the specified key.
         *
         * @param map A map-like.
         * @param key A property key.
         */
        function hasProperty(map, key) {
            return hasOwnProperty.call(map, key);
        }
        ts.hasProperty = hasProperty;
        /**
         * Gets the value of an owned property in a map-like.
         *
         * @param map A map-like.
         * @param key A property key.
         */
        function getProperty(map, key) {
            return hasOwnProperty.call(map, key) ? map[key] : undefined;
        }
        ts.getProperty = getProperty;
        /**
         * Gets the owned, enumerable property keys of a map-like.
         */
        function getOwnKeys(map) {
            var keys = [];
            for (var key in map) {
                if (hasOwnProperty.call(map, key)) {
                    keys.push(key);
                }
            }
            return keys;
        }
        ts.getOwnKeys = getOwnKeys;
        function getAllKeys(obj) {
            var result = [];
            do {
                var names = Object.getOwnPropertyNames(obj);
                for (var _i = 0, names_1 = names; _i < names_1.length; _i++) {
                    var name = names_1[_i];
                    pushIfUnique(result, name);
                }
            } while (obj = Object.getPrototypeOf(obj));
            return result;
        }
        ts.getAllKeys = getAllKeys;
        function getOwnValues(sparseArray) {
            var values = [];
            for (var key in sparseArray) {
                if (hasOwnProperty.call(sparseArray, key)) {
                    values.push(sparseArray[key]);
                }
            }
            return values;
        }
        ts.getOwnValues = getOwnValues;
        var _entries = Object.entries || (function (obj) {
            var keys = getOwnKeys(obj);
            var result = Array(keys.length);
            for (var i = 0; i < keys.length; i++) {
                result[i] = [keys[i], obj[keys[i]]];
            }
            return result;
        });
        function getEntries(obj) {
            return obj ? _entries(obj) : [];
        }
        ts.getEntries = getEntries;
        function arrayOf(count, f) {
            var result = new Array(count);
            for (var i = 0; i < count; i++) {
                result[i] = f(i);
            }
            return result;
        }
        ts.arrayOf = arrayOf;
        function arrayFrom(iterator, map) {
            var result = [];
            for (var iterResult = iterator.next(); !iterResult.done; iterResult = iterator.next()) {
                result.push(map ? map(iterResult.value) : iterResult.value);
            }
            return result;
        }
        ts.arrayFrom = arrayFrom;
        function assign(t) {
            var args = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                args[_i - 1] = arguments[_i];
            }
            for (var _a = 0, args_1 = args; _a < args_1.length; _a++) {
                var arg = args_1[_a];
                if (arg === undefined)
                    continue;
                for (var p in arg) {
                    if (hasProperty(arg, p)) {
                        t[p] = arg[p];
                    }
                }
            }
            return t;
        }
        ts.assign = assign;
        /**
         * Performs a shallow equality comparison of the contents of two map-likes.
         *
         * @param left A map-like whose properties should be compared.
         * @param right A map-like whose properties should be compared.
         */
        function equalOwnProperties(left, right, equalityComparer) {
            if (equalityComparer === void 0) {
                equalityComparer = equateValues;
            }
            if (left === right)
                return true;
            if (!left || !right)
                return false;
            for (var key in left) {
                if (hasOwnProperty.call(left, key)) {
                    if (!hasOwnProperty.call(right, key))
                        return false;
                    if (!equalityComparer(left[key], right[key]))
                        return false;
                }
            }
            for (var key in right) {
                if (hasOwnProperty.call(right, key)) {
                    if (!hasOwnProperty.call(left, key))
                        return false;
                }
            }
            return true;
        }
        ts.equalOwnProperties = equalOwnProperties;
        function arrayToMap(array, makeKey, makeValue) {
            if (makeValue === void 0) {
                makeValue = identity;
            }
            var result = new ts.Map();
            for (var _i = 0, array_6 = array; _i < array_6.length; _i++) {
                var value = array_6[_i];
                var key = makeKey(value);
                if (key !== undefined)
                    result.set(key, makeValue(value));
            }
            return result;
        }
        ts.arrayToMap = arrayToMap;
        function arrayToNumericMap(array, makeKey, makeValue) {
            if (makeValue === void 0) {
                makeValue = identity;
            }
            var result = [];
            for (var _i = 0, array_7 = array; _i < array_7.length; _i++) {
                var value = array_7[_i];
                result[makeKey(value)] = makeValue(value);
            }
            return result;
        }
        ts.arrayToNumericMap = arrayToNumericMap;
        function arrayToMultiMap(values, makeKey, makeValue) {
            if (makeValue === void 0) {
                makeValue = identity;
            }
            var result = createMultiMap();
            for (var _i = 0, values_1 = values; _i < values_1.length; _i++) {
                var value = values_1[_i];
                result.add(makeKey(value), makeValue(value));
            }
            return result;
        }
        ts.arrayToMultiMap = arrayToMultiMap;
        function group(values, getGroupId, resultSelector) {
            if (resultSelector === void 0) {
                resultSelector = identity;
            }
            return arrayFrom(arrayToMultiMap(values, getGroupId).values(), resultSelector);
        }
        ts.group = group;
        function clone(object) {
            var result = {};
            for (var id in object) {
                if (hasOwnProperty.call(object, id)) {
                    result[id] = object[id];
                }
            }
            return result;
        }
        ts.clone = clone;
        /**
         * Creates a new object by adding the own properties of `second`, then the own properties of `first`.
         *
         * NOTE: This means that if a property exists in both `first` and `second`, the property in `first` will be chosen.
         */
        function extend(first, second) {
            var result = {};
            for (var id in second) {
                if (hasOwnProperty.call(second, id)) {
                    result[id] = second[id];
                }
            }
            for (var id in first) {
                if (hasOwnProperty.call(first, id)) {
                    result[id] = first[id];
                }
            }
            return result;
        }
        ts.extend = extend;
        function copyProperties(first, second) {
            for (var id in second) {
                if (hasOwnProperty.call(second, id)) {
                    first[id] = second[id];
                }
            }
        }
        ts.copyProperties = copyProperties;
        function maybeBind(obj, fn) {
            return fn ? fn.bind(obj) : undefined;
        }
        ts.maybeBind = maybeBind;
        function createMultiMap() {
            var map = new ts.Map();
            map.add = multiMapAdd;
            map.remove = multiMapRemove;
            return map;
        }
        ts.createMultiMap = createMultiMap;
        function multiMapAdd(key, value) {
            var values = this.get(key);
            if (values) {
                values.push(value);
            }
            else {
                this.set(key, values = [value]);
            }
            return values;
        }
        function multiMapRemove(key, value) {
            var values = this.get(key);
            if (values) {
                unorderedRemoveItem(values, value);
                if (!values.length) {
                    this.delete(key);
                }
            }
        }
        function createUnderscoreEscapedMultiMap() {
            return createMultiMap();
        }
        ts.createUnderscoreEscapedMultiMap = createUnderscoreEscapedMultiMap;
        /**
         * Tests whether a value is an array.
         */
        function isArray(value) {
            return Array.isArray ? Array.isArray(value) : value instanceof Array;
        }
        ts.isArray = isArray;
        function toArray(value) {
            return isArray(value) ? value : [value];
        }
        ts.toArray = toArray;
        /**
         * Tests whether a value is string
         */
        function isString(text) {
            return typeof text === "string";
        }
        ts.isString = isString;
        function isNumber(x) {
            return typeof x === "number";
        }
        ts.isNumber = isNumber;
        function tryCast(value, test) {
            return value !== undefined && test(value) ? value : undefined;
        }
        ts.tryCast = tryCast;
        function cast(value, test) {
            if (value !== undefined && test(value))
                return value;
            return ts.Debug.fail("Invalid cast. The supplied value " + value + " did not pass the test '" + ts.Debug.getFunctionName(test) + "'.");
        }
        ts.cast = cast;
        /** Does nothing. */
        function noop(_) { }
        ts.noop = noop;
        /** Do nothing and return false */
        function returnFalse() { return false; }
        ts.returnFalse = returnFalse;
        /** Do nothing and return true */
        function returnTrue() { return true; }
        ts.returnTrue = returnTrue;
        /** Do nothing and return undefined */
        function returnUndefined() { return undefined; }
        ts.returnUndefined = returnUndefined;
        /** Returns its argument. */
        function identity(x) { return x; }
        ts.identity = identity;
        /** Returns lower case string */
        function toLowerCase(x) { return x.toLowerCase(); }
        ts.toLowerCase = toLowerCase;
        // We convert the file names to lower case as key for file name on case insensitive file system
        // While doing so we need to handle special characters (eg \u0130) to ensure that we dont convert
        // it to lower case, fileName with its lowercase form can exist along side it.
        // Handle special characters and make those case sensitive instead
        //
        // |-#--|-Unicode--|-Char code-|-Desc-------------------------------------------------------------------|
        // | 1. | i        | 105       | Ascii i                                                                |
        // | 2. | I        | 73        | Ascii I                                                                |
        // |-------- Special characters ------------------------------------------------------------------------|
        // | 3. | \u0130   | 304       | Upper case I with dot above                                            |
        // | 4. | i,\u0307 | 105,775   | i, followed by 775: Lower case of (3rd item)                           |
        // | 5. | I,\u0307 | 73,775    | I, followed by 775: Upper case of (4th item), lower case is (4th item) |
        // | 6. | \u0131   | 305       | Lower case i without dot, upper case is I (2nd item)                   |
        // | 7. | \u00DF   | 223       | Lower case sharp s                                                     |
        //
        // Because item 3 is special where in its lowercase character has its own
        // upper case form we cant convert its case.
        // Rest special characters are either already in lower case format or
        // they have corresponding upper case character so they dont need special handling
        //
        // But to avoid having to do string building for most common cases, also ignore
        // a-z, 0-9, \u0131, \u00DF, \, /, ., : and space
        var fileNameLowerCaseRegExp = /[^\u0130\u0131\u00DFa-z0-9\\/:\-_\. ]+/g;
        /**
         * Case insensitive file systems have descripencies in how they handle some characters (eg. turkish Upper case I with dot on top - \u0130)
         * This function is used in places where we want to make file name as a key on these systems
         * It is possible on mac to be able to refer to file name with I with dot on top as a fileName with its lower case form
         * But on windows we cannot. Windows can have fileName with I with dot on top next to its lower case and they can not each be referred with the lowercase forms
         * Technically we would want this function to be platform sepcific as well but
         * our api has till now only taken caseSensitive as the only input and just for some characters we dont want to update API and ensure all customers use those api
         * We could use upper case and we would still need to deal with the descripencies but
         * we want to continue using lower case since in most cases filenames are lowercasewe and wont need any case changes and avoid having to store another string for the key
         * So for this function purpose, we go ahead and assume character I with dot on top it as case sensitive since its very unlikely to use lower case form of that special character
         */
        function toFileNameLowerCase(x) {
            return fileNameLowerCaseRegExp.test(x) ?
                x.replace(fileNameLowerCaseRegExp, toLowerCase) :
                x;
        }
        ts.toFileNameLowerCase = toFileNameLowerCase;
        /** Throws an error because a function is not implemented. */
        function notImplemented() {
            throw new Error("Not implemented");
        }
        ts.notImplemented = notImplemented;
        function memoize(callback) {
            var value;
            return function () {
                if (callback) {
                    value = callback();
                    callback = undefined;
                }
                return value;
            };
        }
        ts.memoize = memoize;
        /** A version of `memoize` that supports a single primitive argument */
        function memoizeOne(callback) {
            var map = new ts.Map();
            return function (arg) {
                var key = typeof arg + ":" + arg;
                var value = map.get(key);
                if (value === undefined && !map.has(key)) {
                    value = callback(arg);
                    map.set(key, value);
                }
                return value;
            };
        }
        ts.memoizeOne = memoizeOne;
        function compose(a, b, c, d, e) {
            if (!!e) {
                var args_2 = [];
                for (var i = 0; i < arguments.length; i++) {
                    args_2[i] = arguments[i];
                }
                return function (t) { return reduceLeft(args_2, function (u, f) { return f(u); }, t); };
            }
            else if (d) {
                return function (t) { return d(c(b(a(t)))); };
            }
            else if (c) {
                return function (t) { return c(b(a(t))); };
            }
            else if (b) {
                return function (t) { return b(a(t)); };
            }
            else if (a) {
                return function (t) { return a(t); };
            }
            else {
                return function (t) { return t; };
            }
        }
        ts.compose = compose;
        var AssertionLevel;
        (function (AssertionLevel) {
            AssertionLevel[AssertionLevel["None"] = 0] = "None";
            AssertionLevel[AssertionLevel["Normal"] = 1] = "Normal";
            AssertionLevel[AssertionLevel["Aggressive"] = 2] = "Aggressive";
            AssertionLevel[AssertionLevel["VeryAggressive"] = 3] = "VeryAggressive";
        })(AssertionLevel = ts.AssertionLevel || (ts.AssertionLevel = {}));
        function equateValues(a, b) {
            return a === b;
        }
        ts.equateValues = equateValues;
        /**
         * Compare the equality of two strings using a case-sensitive ordinal comparison.
         *
         * Case-sensitive comparisons compare both strings one code-point at a time using the integer
         * value of each code-point after applying `toUpperCase` to each string. We always map both
         * strings to their upper-case form as some unicode characters do not properly round-trip to
         * lowercase (such as `ẞ` (German sharp capital s)).
         */
        function equateStringsCaseInsensitive(a, b) {
            return a === b
                || a !== undefined
                    && b !== undefined
                    && a.toUpperCase() === b.toUpperCase();
        }
        ts.equateStringsCaseInsensitive = equateStringsCaseInsensitive;
        /**
         * Compare the equality of two strings using a case-sensitive ordinal comparison.
         *
         * Case-sensitive comparisons compare both strings one code-point at a time using the
         * integer value of each code-point.
         */
        function equateStringsCaseSensitive(a, b) {
            return equateValues(a, b);
        }
        ts.equateStringsCaseSensitive = equateStringsCaseSensitive;
        function compareComparableValues(a, b) {
            return a === b ? 0 /* EqualTo */ :
                a === undefined ? -1 /* LessThan */ :
                    b === undefined ? 1 /* GreaterThan */ :
                        a < b ? -1 /* LessThan */ :
                            1 /* GreaterThan */;
        }
        /**
         * Compare two numeric values for their order relative to each other.
         * To compare strings, use any of the `compareStrings` functions.
         */
        function compareValues(a, b) {
            return compareComparableValues(a, b);
        }
        ts.compareValues = compareValues;
        /**
         * Compare two TextSpans, first by `start`, then by `length`.
         */
        function compareTextSpans(a, b) {
            return compareValues(a === null || a === void 0 ? void 0 : a.start, b === null || b === void 0 ? void 0 : b.start) || compareValues(a === null || a === void 0 ? void 0 : a.length, b === null || b === void 0 ? void 0 : b.length);
        }
        ts.compareTextSpans = compareTextSpans;
        function min(a, b, compare) {
            return compare(a, b) === -1 /* LessThan */ ? a : b;
        }
        ts.min = min;
        /**
         * Compare two strings using a case-insensitive ordinal comparison.
         *
         * Ordinal comparisons are based on the difference between the unicode code points of both
         * strings. Characters with multiple unicode representations are considered unequal. Ordinal
         * comparisons provide predictable ordering, but place "a" after "B".
         *
         * Case-insensitive comparisons compare both strings one code-point at a time using the integer
         * value of each code-point after applying `toUpperCase` to each string. We always map both
         * strings to their upper-case form as some unicode characters do not properly round-trip to
         * lowercase (such as `ẞ` (German sharp capital s)).
         */
        function compareStringsCaseInsensitive(a, b) {
            if (a === b)
                return 0 /* EqualTo */;
            if (a === undefined)
                return -1 /* LessThan */;
            if (b === undefined)
                return 1 /* GreaterThan */;
            a = a.toUpperCase();
            b = b.toUpperCase();
            return a < b ? -1 /* LessThan */ : a > b ? 1 /* GreaterThan */ : 0 /* EqualTo */;
        }
        ts.compareStringsCaseInsensitive = compareStringsCaseInsensitive;
        /**
         * Compare two strings using a case-sensitive ordinal comparison.
         *
         * Ordinal comparisons are based on the difference between the unicode code points of both
         * strings. Characters with multiple unicode representations are considered unequal. Ordinal
         * comparisons provide predictable ordering, but place "a" after "B".
         *
         * Case-sensitive comparisons compare both strings one code-point at a time using the integer
         * value of each code-point.
         */
        function compareStringsCaseSensitive(a, b) {
            return compareComparableValues(a, b);
        }
        ts.compareStringsCaseSensitive = compareStringsCaseSensitive;
        function getStringComparer(ignoreCase) {
            return ignoreCase ? compareStringsCaseInsensitive : compareStringsCaseSensitive;
        }
        ts.getStringComparer = getStringComparer;
        /**
         * Creates a string comparer for use with string collation in the UI.
         */
        var createUIStringComparer = (function () {
            var defaultComparer;
            var enUSComparer;
            var stringComparerFactory = getStringComparerFactory();
            return createStringComparer;
            function compareWithCallback(a, b, comparer) {
                if (a === b)
                    return 0 /* EqualTo */;
                if (a === undefined)
                    return -1 /* LessThan */;
                if (b === undefined)
                    return 1 /* GreaterThan */;
                var value = comparer(a, b);
                return value < 0 ? -1 /* LessThan */ : value > 0 ? 1 /* GreaterThan */ : 0 /* EqualTo */;
            }
            function createIntlCollatorStringComparer(locale) {
                // Intl.Collator.prototype.compare is bound to the collator. See NOTE in
                // http://www.ecma-international.org/ecma-402/2.0/#sec-Intl.Collator.prototype.compare
                var comparer = new Intl.Collator(locale, { usage: "sort", sensitivity: "variant" }).compare;
                return function (a, b) { return compareWithCallback(a, b, comparer); };
            }
            function createLocaleCompareStringComparer(locale) {
                // if the locale is not the default locale (`undefined`), use the fallback comparer.
                if (locale !== undefined)
                    return createFallbackStringComparer();
                return function (a, b) { return compareWithCallback(a, b, compareStrings); };
                function compareStrings(a, b) {
                    return a.localeCompare(b);
                }
            }
            function createFallbackStringComparer() {
                // An ordinal comparison puts "A" after "b", but for the UI we want "A" before "b".
                // We first sort case insensitively.  So "Aaa" will come before "baa".
                // Then we sort case sensitively, so "aaa" will come before "Aaa".
                //
                // For case insensitive comparisons we always map both strings to their
                // upper-case form as some unicode characters do not properly round-trip to
                // lowercase (such as `Ã¡ÂºÅ¾` (German sharp capital s)).
                return function (a, b) { return compareWithCallback(a, b, compareDictionaryOrder); };
                function compareDictionaryOrder(a, b) {
                    return compareStrings(a.toUpperCase(), b.toUpperCase()) || compareStrings(a, b);
                }
                function compareStrings(a, b) {
                    return a < b ? -1 /* LessThan */ : a > b ? 1 /* GreaterThan */ : 0 /* EqualTo */;
                }
            }
            function getStringComparerFactory() {
                // If the host supports Intl, we use it for comparisons using the default locale.
                if (typeof Intl === "object" && typeof Intl.Collator === "function") {
                    return createIntlCollatorStringComparer;
                }
                // If the host does not support Intl, we fall back to localeCompare.
                // localeCompare in Node v0.10 is just an ordinal comparison, so don't use it.
                if (typeof String.prototype.localeCompare === "function" &&
                    typeof String.prototype.toLocaleUpperCase === "function" &&
                    "a".localeCompare("B") < 0) {
                    return createLocaleCompareStringComparer;
                }
                // Otherwise, fall back to ordinal comparison:
                return createFallbackStringComparer;
            }
            function createStringComparer(locale) {
                // Hold onto common string comparers. This avoids constantly reallocating comparers during
                // tests.
                if (locale === undefined) {
                    return defaultComparer || (defaultComparer = stringComparerFactory(locale));
                }
                else if (locale === "en-US") {
                    return enUSComparer || (enUSComparer = stringComparerFactory(locale));
                }
                else {
                    return stringComparerFactory(locale);
                }
            }
        })();
        var uiComparerCaseSensitive;
        var uiLocale;
        function getUILocale() {
            return uiLocale;
        }
        ts.getUILocale = getUILocale;
        function setUILocale(value) {
            if (uiLocale !== value) {
                uiLocale = value;
                uiComparerCaseSensitive = undefined;
            }
        }
        ts.setUILocale = setUILocale;
        /**
         * Compare two strings in a using the case-sensitive sort behavior of the UI locale.
         *
         * Ordering is not predictable between different host locales, but is best for displaying
         * ordered data for UI presentation. Characters with multiple unicode representations may
         * be considered equal.
         *
         * Case-sensitive comparisons compare strings that differ in base characters, or
         * accents/diacritic marks, or case as unequal.
         */
        function compareStringsCaseSensitiveUI(a, b) {
            var comparer = uiComparerCaseSensitive || (uiComparerCaseSensitive = createUIStringComparer(uiLocale));
            return comparer(a, b);
        }
        ts.compareStringsCaseSensitiveUI = compareStringsCaseSensitiveUI;
        function compareProperties(a, b, key, comparer) {
            return a === b ? 0 /* EqualTo */ :
                a === undefined ? -1 /* LessThan */ :
                    b === undefined ? 1 /* GreaterThan */ :
                        comparer(a[key], b[key]);
        }
        ts.compareProperties = compareProperties;
        /** True is greater than false. */
        function compareBooleans(a, b) {
            return compareValues(a ? 1 : 0, b ? 1 : 0);
        }
        ts.compareBooleans = compareBooleans;
        /**
         * Given a name and a list of names that are *not* equal to the name, return a spelling suggestion if there is one that is close enough.
         * Names less than length 3 only check for case-insensitive equality.
         *
         * find the candidate with the smallest Levenshtein distance,
         *    except for candidates:
         *      * With no name
         *      * Whose length differs from the target name by more than 0.34 of the length of the name.
         *      * Whose levenshtein distance is more than 0.4 of the length of the name
         *        (0.4 allows 1 substitution/transposition for every 5 characters,
         *         and 1 insertion/deletion at 3 characters)
         */
        function getSpellingSuggestion(name, candidates, getName) {
            var maximumLengthDifference = Math.min(2, Math.floor(name.length * 0.34));
            var bestDistance = Math.floor(name.length * 0.4) + 1; // If the best result is worse than this, don't bother.
            var bestCandidate;
            for (var _i = 0, candidates_1 = candidates; _i < candidates_1.length; _i++) {
                var candidate = candidates_1[_i];
                var candidateName = getName(candidate);
                if (candidateName !== undefined && Math.abs(candidateName.length - name.length) <= maximumLengthDifference) {
                    if (candidateName === name) {
                        continue;
                    }
                    // Only consider candidates less than 3 characters long when they differ by case.
                    // Otherwise, don't bother, since a user would usually notice differences of a 2-character name.
                    if (candidateName.length < 3 && candidateName.toLowerCase() !== name.toLowerCase()) {
                        continue;
                    }
                    var distance = levenshteinWithMax(name, candidateName, bestDistance - 0.1);
                    if (distance === undefined) {
                        continue;
                    }
                    ts.Debug.assert(distance < bestDistance); // Else `levenshteinWithMax` should return undefined
                    bestDistance = distance;
                    bestCandidate = candidate;
                }
            }
            return bestCandidate;
        }
        ts.getSpellingSuggestion = getSpellingSuggestion;
        function levenshteinWithMax(s1, s2, max) {
            var previous = new Array(s2.length + 1);
            var current = new Array(s2.length + 1);
            /** Represents any value > max. We don't care about the particular value. */
            var big = max + 0.01;
            for (var i = 0; i <= s2.length; i++) {
                previous[i] = i;
            }
            for (var i = 1; i <= s1.length; i++) {
                var c1 = s1.charCodeAt(i - 1);
                var minJ = Math.ceil(i > max ? i - max : 1);
                var maxJ = Math.floor(s2.length > max + i ? max + i : s2.length);
                current[0] = i;
                /** Smallest value of the matrix in the ith column. */
                var colMin = i;
                for (var j = 1; j < minJ; j++) {
                    current[j] = big;
                }
                for (var j = minJ; j <= maxJ; j++) {
                    // case difference should be significantly cheaper than other differences
                    var substitutionDistance = s1[i - 1].toLowerCase() === s2[j - 1].toLowerCase()
                        ? (previous[j - 1] + 0.1)
                        : (previous[j - 1] + 2);
                    var dist = c1 === s2.charCodeAt(j - 1)
                        ? previous[j - 1]
                        : Math.min(/*delete*/ previous[j] + 1, /*insert*/ current[j - 1] + 1, /*substitute*/ substitutionDistance);
                    current[j] = dist;
                    colMin = Math.min(colMin, dist);
                }
                for (var j = maxJ + 1; j <= s2.length; j++) {
                    current[j] = big;
                }
                if (colMin > max) {
                    // Give up -- everything in this column is > max and it can't get better in future columns.
                    return undefined;
                }
                var temp = previous;
                previous = current;
                current = temp;
            }
            var res = previous[s2.length];
            return res > max ? undefined : res;
        }
        function endsWith(str, suffix) {
            var expectedPos = str.length - suffix.length;
            return expectedPos >= 0 && str.indexOf(suffix, expectedPos) === expectedPos;
        }
        ts.endsWith = endsWith;
        function removeSuffix(str, suffix) {
            return endsWith(str, suffix) ? str.slice(0, str.length - suffix.length) : str;
        }
        ts.removeSuffix = removeSuffix;
        function tryRemoveSuffix(str, suffix) {
            return endsWith(str, suffix) ? str.slice(0, str.length - suffix.length) : undefined;
        }
        ts.tryRemoveSuffix = tryRemoveSuffix;
        function stringContains(str, substring) {
            return str.indexOf(substring) !== -1;
        }
        ts.stringContains = stringContains;
        /**
         * Takes a string like "jquery-min.4.2.3" and returns "jquery"
         */
        function removeMinAndVersionNumbers(fileName) {
            // We used to use the regex /[.-]((min)|(\d+(\.\d+)*))$/ and would just .replace it twice.
            // Unfortunately, that regex has O(n^2) performance because v8 doesn't match from the end of the string.
            // Instead, we now essentially scan the filename (backwards) ourselves.
            var end = fileName.length;
            for (var pos = end - 1; pos > 0; pos--) {
                var ch = fileName.charCodeAt(pos);
                if (ch >= 48 /* _0 */ && ch <= 57 /* _9 */) {
                    // Match a \d+ segment
                    do {
                        --pos;
                        ch = fileName.charCodeAt(pos);
                    } while (pos > 0 && ch >= 48 /* _0 */ && ch <= 57 /* _9 */);
                }
                else if (pos > 4 && (ch === 110 /* n */ || ch === 78 /* N */)) {
                    // Looking for "min" or "min"
                    // Already matched the 'n'
                    --pos;
                    ch = fileName.charCodeAt(pos);
                    if (ch !== 105 /* i */ && ch !== 73 /* I */) {
                        break;
                    }
                    --pos;
                    ch = fileName.charCodeAt(pos);
                    if (ch !== 109 /* m */ && ch !== 77 /* M */) {
                        break;
                    }
                    --pos;
                    ch = fileName.charCodeAt(pos);
                }
                else {
                    // This character is not part of either suffix pattern
                    break;
                }
                if (ch !== 45 /* minus */ && ch !== 46 /* dot */) {
                    break;
                }
                end = pos;
            }
            // end might be fileName.length, in which case this should internally no-op
            return end === fileName.length ? fileName : fileName.slice(0, end);
        }
        ts.removeMinAndVersionNumbers = removeMinAndVersionNumbers;
        /** Remove an item from an array, moving everything to its right one space left. */
        function orderedRemoveItem(array, item) {
            for (var i = 0; i < array.length; i++) {
                if (array[i] === item) {
                    orderedRemoveItemAt(array, i);
                    return true;
                }
            }
            return false;
        }
        ts.orderedRemoveItem = orderedRemoveItem;
        /** Remove an item by index from an array, moving everything to its right one space left. */
        function orderedRemoveItemAt(array, index) {
            // This seems to be faster than either `array.splice(i, 1)` or `array.copyWithin(i, i+ 1)`.
            for (var i = index; i < array.length - 1; i++) {
                array[i] = array[i + 1];
            }
            array.pop();
        }
        ts.orderedRemoveItemAt = orderedRemoveItemAt;
        function unorderedRemoveItemAt(array, index) {
            // Fill in the "hole" left at `index`.
            array[index] = array[array.length - 1];
            array.pop();
        }
        ts.unorderedRemoveItemAt = unorderedRemoveItemAt;
        /** Remove the *first* occurrence of `item` from the array. */
        function unorderedRemoveItem(array, item) {
            return unorderedRemoveFirstItemWhere(array, function (element) { return element === item; });
        }
        ts.unorderedRemoveItem = unorderedRemoveItem;
        /** Remove the *first* element satisfying `predicate`. */
        function unorderedRemoveFirstItemWhere(array, predicate) {
            for (var i = 0; i < array.length; i++) {
                if (predicate(array[i])) {
                    unorderedRemoveItemAt(array, i);
                    return true;
                }
            }
            return false;
        }
        function createGetCanonicalFileName(useCaseSensitiveFileNames) {
            return useCaseSensitiveFileNames ? identity : toFileNameLowerCase;
        }
        ts.createGetCanonicalFileName = createGetCanonicalFileName;
        function patternText(_a) {
            var prefix = _a.prefix, suffix = _a.suffix;
            return prefix + "*" + suffix;
        }
        ts.patternText = patternText;
        /**
         * Given that candidate matches pattern, returns the text matching the '*'.
         * E.g.: matchedText(tryParsePattern("foo*baz"), "foobarbaz") === "bar"
         */
        function matchedText(pattern, candidate) {
            ts.Debug.assert(isPatternMatch(pattern, candidate));
            return candidate.substring(pattern.prefix.length, candidate.length - pattern.suffix.length);
        }
        ts.matchedText = matchedText;
        /** Return the object corresponding to the best pattern to match `candidate`. */
        function findBestPatternMatch(values, getPattern, candidate) {
            var matchedValue;
            // use length of prefix as betterness criteria
            var longestMatchPrefixLength = -1;
            for (var _i = 0, values_2 = values; _i < values_2.length; _i++) {
                var v = values_2[_i];
                var pattern = getPattern(v);
                if (isPatternMatch(pattern, candidate) && pattern.prefix.length > longestMatchPrefixLength) {
                    longestMatchPrefixLength = pattern.prefix.length;
                    matchedValue = v;
                }
            }
            return matchedValue;
        }
        ts.findBestPatternMatch = findBestPatternMatch;
        function startsWith(str, prefix) {
            return str.lastIndexOf(prefix, 0) === 0;
        }
        ts.startsWith = startsWith;
        function removePrefix(str, prefix) {
            return startsWith(str, prefix) ? str.substr(prefix.length) : str;
        }
        ts.removePrefix = removePrefix;
        function tryRemovePrefix(str, prefix, getCanonicalFileName) {
            if (getCanonicalFileName === void 0) {
                getCanonicalFileName = identity;
            }
            return startsWith(getCanonicalFileName(str), getCanonicalFileName(prefix)) ? str.substring(prefix.length) : undefined;
        }
        ts.tryRemovePrefix = tryRemovePrefix;
        function isPatternMatch(_a, candidate) {
            var prefix = _a.prefix, suffix = _a.suffix;
            return candidate.length >= prefix.length + suffix.length &&
                startsWith(candidate, prefix) &&
                endsWith(candidate, suffix);
        }
        function and(f, g) {
            return function (arg) { return f(arg) && g(arg); };
        }
        ts.and = and;
        function or() {
            var fs = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                fs[_i] = arguments[_i];
            }
            return function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                for (var _a = 0, fs_1 = fs; _a < fs_1.length; _a++) {
                    var f = fs_1[_a];
                    if (f.apply(void 0, args)) {
                        return true;
                    }
                }
                return false;
            };
        }
        ts.or = or;
        function not(fn) {
            return function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return !fn.apply(void 0, args);
            };
        }
        ts.not = not;
        function assertType(_) { }
        ts.assertType = assertType;
        function singleElementArray(t) {
            return t === undefined ? undefined : [t];
        }
        ts.singleElementArray = singleElementArray;
        function enumerateInsertsAndDeletes(newItems, oldItems, comparer, inserted, deleted, unchanged) {
            unchanged = unchanged || noop;
            var newIndex = 0;
            var oldIndex = 0;
            var newLen = newItems.length;
            var oldLen = oldItems.length;
            var hasChanges = false;
            while (newIndex < newLen && oldIndex < oldLen) {
                var newItem = newItems[newIndex];
                var oldItem = oldItems[oldIndex];
                var compareResult = comparer(newItem, oldItem);
                if (compareResult === -1 /* LessThan */) {
                    inserted(newItem);
                    newIndex++;
                    hasChanges = true;
                }
                else if (compareResult === 1 /* GreaterThan */) {
                    deleted(oldItem);
                    oldIndex++;
                    hasChanges = true;
                }
                else {
                    unchanged(oldItem, newItem);
                    newIndex++;
                    oldIndex++;
                }
            }
            while (newIndex < newLen) {
                inserted(newItems[newIndex++]);
                hasChanges = true;
            }
            while (oldIndex < oldLen) {
                deleted(oldItems[oldIndex++]);
                hasChanges = true;
            }
            return hasChanges;
        }
        ts.enumerateInsertsAndDeletes = enumerateInsertsAndDeletes;
        function fill(length, cb) {
            var result = Array(length);
            for (var i = 0; i < length; i++) {
                result[i] = cb(i);
            }
            return result;
        }
        ts.fill = fill;
        function cartesianProduct(arrays) {
            var result = [];
            cartesianProductWorker(arrays, result, /*outer*/ undefined, 0);
            return result;
        }
        ts.cartesianProduct = cartesianProduct;
        function cartesianProductWorker(arrays, result, outer, index) {
            for (var _i = 0, _a = arrays[index]; _i < _a.length; _i++) {
                var element = _a[_i];
                var inner = void 0;
                if (outer) {
                    inner = outer.slice();
                    inner.push(element);
                }
                else {
                    inner = [element];
                }
                if (index === arrays.length - 1) {
                    result.push(inner);
                }
                else {
                    cartesianProductWorker(arrays, result, inner, index + 1);
                }
            }
        }
        /**
         * Returns string left-padded with spaces or zeros until it reaches the given length.
         *
         * @param s String to pad.
         * @param length Final padded length. If less than or equal to 's.length', returns 's' unchanged.
         * @param padString Character to use as padding (default " ").
         */
        function padLeft(s, length, padString) {
            if (padString === void 0) {
                padString = " ";
            }
            return length <= s.length ? s : padString.repeat(length - s.length) + s;
        }
        ts.padLeft = padLeft;
        /**
         * Returns string right-padded with spaces until it reaches the given length.
         *
         * @param s String to pad.
         * @param length Final padded length. If less than or equal to 's.length', returns 's' unchanged.
         * @param padString Character to use as padding (default " ").
         */
        function padRight(s, length, padString) {
            if (padString === void 0) {
                padString = " ";
            }
            return length <= s.length ? s : s + padString.repeat(length - s.length);
        }
        ts.padRight = padRight;
        function takeWhile(array, predicate) {
            var len = array.length;
            var index = 0;
            while (index < len && predicate(array[index])) {
                index++;
            }
            return array.slice(0, index);
        }
        ts.takeWhile = takeWhile;
        /**
         * Removes the leading and trailing white space and line terminator characters from a string.
         */
        ts.trimString = !!String.prototype.trim ? (function (s) { return s.trim(); }) : function (s) { return ts.trimStringEnd(ts.trimStringStart(s)); };
        /**
         * Returns a copy with trailing whitespace removed.
         */
        ts.trimStringEnd = !!String.prototype.trimEnd ? (function (s) { return s.trimEnd(); }) : trimEndImpl;
        /**
         * Returns a copy with leading whitespace removed.
         */
        ts.trimStringStart = !!String.prototype.trimStart ? (function (s) { return s.trimStart(); }) : function (s) { return s.replace(/^\s+/g, ""); };
        /**
         * https://jsbench.me/gjkoxld4au/1
         * The simple regex for this, /\s+$/g is O(n^2) in v8.
         * The native .trimEnd method is by far best, but since that's technically ES2019,
         * we provide a (still much faster than the simple regex) fallback.
         */
        function trimEndImpl(s) {
            var end = s.length - 1;
            while (end >= 0) {
                if (!ts.isWhiteSpaceLike(s.charCodeAt(end)))
                    break;
                end--;
            }
            return s.slice(0, end + 1);
        }
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var LogLevel;
        (function (LogLevel) {
            LogLevel[LogLevel["Off"] = 0] = "Off";
            LogLevel[LogLevel["Error"] = 1] = "Error";
            LogLevel[LogLevel["Warning"] = 2] = "Warning";
            LogLevel[LogLevel["Info"] = 3] = "Info";
            LogLevel[LogLevel["Verbose"] = 4] = "Verbose";
        })(LogLevel = ts.LogLevel || (ts.LogLevel = {}));
        var Debug;
        (function (Debug) {
            var typeScriptVersion;
            /* eslint-disable prefer-const */
            var currentAssertionLevel = 0 /* None */;
            Debug.currentLogLevel = LogLevel.Warning;
            Debug.isDebugging = false;
            function getTypeScriptVersion() {
                return typeScriptVersion !== null && typeScriptVersion !== void 0 ? typeScriptVersion : (typeScriptVersion = new ts.Version(ts.version));
            }
            Debug.getTypeScriptVersion = getTypeScriptVersion;
            function shouldLog(level) {
                return Debug.currentLogLevel <= level;
            }
            Debug.shouldLog = shouldLog;
            function logMessage(level, s) {
                if (Debug.loggingHost && shouldLog(level)) {
                    Debug.loggingHost.log(level, s);
                }
            }
            function log(s) {
                logMessage(LogLevel.Info, s);
            }
            Debug.log = log;
            (function (log_1) {
                function error(s) {
                    logMessage(LogLevel.Error, s);
                }
                log_1.error = error;
                function warn(s) {
                    logMessage(LogLevel.Warning, s);
                }
                log_1.warn = warn;
                function log(s) {
                    logMessage(LogLevel.Info, s);
                }
                log_1.log = log;
                function trace(s) {
                    logMessage(LogLevel.Verbose, s);
                }
                log_1.trace = trace;
            })(log = Debug.log || (Debug.log = {}));
            var assertionCache = {};
            function getAssertionLevel() {
                return currentAssertionLevel;
            }
            Debug.getAssertionLevel = getAssertionLevel;
            function setAssertionLevel(level) {
                var prevAssertionLevel = currentAssertionLevel;
                currentAssertionLevel = level;
                if (level > prevAssertionLevel) {
                    // restore assertion functions for the current assertion level (see `shouldAssertFunction`).
                    for (var _i = 0, _a = ts.getOwnKeys(assertionCache); _i < _a.length; _i++) {
                        var key = _a[_i];
                        var cachedFunc = assertionCache[key];
                        if (cachedFunc !== undefined && Debug[key] !== cachedFunc.assertion && level >= cachedFunc.level) {
                            Debug[key] = cachedFunc;
                            assertionCache[key] = undefined;
                        }
                    }
                }
            }
            Debug.setAssertionLevel = setAssertionLevel;
            function shouldAssert(level) {
                return currentAssertionLevel >= level;
            }
            Debug.shouldAssert = shouldAssert;
            /**
             * Tests whether an assertion function should be executed. If it shouldn't, it is cached and replaced with `ts.noop`.
             * Replaced assertion functions are restored when `Debug.setAssertionLevel` is set to a high enough level.
             * @param level The minimum assertion level required.
             * @param name The name of the current assertion function.
             */
            function shouldAssertFunction(level, name) {
                if (!shouldAssert(level)) {
                    assertionCache[name] = { level: level, assertion: Debug[name] };
                    Debug[name] = ts.noop;
                    return false;
                }
                return true;
            }
            function fail(message, stackCrawlMark) {
                debugger;
                var e = new Error(message ? "Debug Failure. " + message : "Debug Failure.");
                if (Error.captureStackTrace) {
                    Error.captureStackTrace(e, stackCrawlMark || fail);
                }
                throw e;
            }
            Debug.fail = fail;
            function failBadSyntaxKind(node, message, stackCrawlMark) {
                return fail((message || "Unexpected node.") + "\r\nNode " + formatSyntaxKind(node.kind) + " was unexpected.", stackCrawlMark || failBadSyntaxKind);
            }
            Debug.failBadSyntaxKind = failBadSyntaxKind;
            function assert(expression, message, verboseDebugInfo, stackCrawlMark) {
                if (!expression) {
                    message = message ? "False expression: " + message : "False expression.";
                    if (verboseDebugInfo) {
                        message += "\r\nVerbose Debug Information: " + (typeof verboseDebugInfo === "string" ? verboseDebugInfo : verboseDebugInfo());
                    }
                    fail(message, stackCrawlMark || assert);
                }
            }
            Debug.assert = assert;
            function assertEqual(a, b, msg, msg2, stackCrawlMark) {
                if (a !== b) {
                    var message = msg ? msg2 ? msg + " " + msg2 : msg : "";
                    fail("Expected " + a + " === " + b + ". " + message, stackCrawlMark || assertEqual);
                }
            }
            Debug.assertEqual = assertEqual;
            function assertLessThan(a, b, msg, stackCrawlMark) {
                if (a >= b) {
                    fail("Expected " + a + " < " + b + ". " + (msg || ""), stackCrawlMark || assertLessThan);
                }
            }
            Debug.assertLessThan = assertLessThan;
            function assertLessThanOrEqual(a, b, stackCrawlMark) {
                if (a > b) {
                    fail("Expected " + a + " <= " + b, stackCrawlMark || assertLessThanOrEqual);
                }
            }
            Debug.assertLessThanOrEqual = assertLessThanOrEqual;
            function assertGreaterThanOrEqual(a, b, stackCrawlMark) {
                if (a < b) {
                    fail("Expected " + a + " >= " + b, stackCrawlMark || assertGreaterThanOrEqual);
                }
            }
            Debug.assertGreaterThanOrEqual = assertGreaterThanOrEqual;
            function assertIsDefined(value, message, stackCrawlMark) {
                // eslint-disable-next-line no-null/no-null
                if (value === undefined || value === null) {
                    fail(message, stackCrawlMark || assertIsDefined);
                }
            }
            Debug.assertIsDefined = assertIsDefined;
            function checkDefined(value, message, stackCrawlMark) {
                assertIsDefined(value, message, stackCrawlMark || checkDefined);
                return value;
            }
            Debug.checkDefined = checkDefined;
            /**
             * @deprecated Use `checkDefined` to check whether a value is defined inline. Use `assertIsDefined` to check whether
             * a value is defined at the statement level.
             */
            Debug.assertDefined = checkDefined;
            function assertEachIsDefined(value, message, stackCrawlMark) {
                for (var _i = 0, value_1 = value; _i < value_1.length; _i++) {
                    var v = value_1[_i];
                    assertIsDefined(v, message, stackCrawlMark || assertEachIsDefined);
                }
            }
            Debug.assertEachIsDefined = assertEachIsDefined;
            function checkEachDefined(value, message, stackCrawlMark) {
                assertEachIsDefined(value, message, stackCrawlMark || checkEachDefined);
                return value;
            }
            Debug.checkEachDefined = checkEachDefined;
            /**
             * @deprecated Use `checkEachDefined` to check whether the elements of an array are defined inline. Use `assertEachIsDefined` to check whether
             * the elements of an array are defined at the statement level.
             */
            Debug.assertEachDefined = checkEachDefined;
            function assertNever(member, message, stackCrawlMark) {
                if (message === void 0) {
                    message = "Illegal value:";
                }
                var detail = typeof member === "object" && ts.hasProperty(member, "kind") && ts.hasProperty(member, "pos") && formatSyntaxKind ? "SyntaxKind: " + formatSyntaxKind(member.kind) : JSON.stringify(member);
                return fail(message + " " + detail, stackCrawlMark || assertNever);
            }
            Debug.assertNever = assertNever;
            function assertEachNode(nodes, test, message, stackCrawlMark) {
                if (shouldAssertFunction(1 /* Normal */, "assertEachNode")) {
                    assert(test === undefined || ts.every(nodes, test), message || "Unexpected node.", function () { return "Node array did not pass test '" + getFunctionName(test) + "'."; }, stackCrawlMark || assertEachNode);
                }
            }
            Debug.assertEachNode = assertEachNode;
            function assertNode(node, test, message, stackCrawlMark) {
                if (shouldAssertFunction(1 /* Normal */, "assertNode")) {
                    assert(node !== undefined && (test === undefined || test(node)), message || "Unexpected node.", function () { return "Node " + formatSyntaxKind(node === null || node === void 0 ? void 0 : node.kind) + " did not pass test '" + getFunctionName(test) + "'."; }, stackCrawlMark || assertNode);
                }
            }
            Debug.assertNode = assertNode;
            function assertNotNode(node, test, message, stackCrawlMark) {
                if (shouldAssertFunction(1 /* Normal */, "assertNotNode")) {
                    assert(node === undefined || test === undefined || !test(node), message || "Unexpected node.", function () { return "Node " + formatSyntaxKind(node.kind) + " should not have passed test '" + getFunctionName(test) + "'."; }, stackCrawlMark || assertNotNode);
                }
            }
            Debug.assertNotNode = assertNotNode;
            function assertOptionalNode(node, test, message, stackCrawlMark) {
                if (shouldAssertFunction(1 /* Normal */, "assertOptionalNode")) {
                    assert(test === undefined || node === undefined || test(node), message || "Unexpected node.", function () { return "Node " + formatSyntaxKind(node === null || node === void 0 ? void 0 : node.kind) + " did not pass test '" + getFunctionName(test) + "'."; }, stackCrawlMark || assertOptionalNode);
                }
            }
            Debug.assertOptionalNode = assertOptionalNode;
            function assertOptionalToken(node, kind, message, stackCrawlMark) {
                if (shouldAssertFunction(1 /* Normal */, "assertOptionalToken")) {
                    assert(kind === undefined || node === undefined || node.kind === kind, message || "Unexpected node.", function () { return "Node " + formatSyntaxKind(node === null || node === void 0 ? void 0 : node.kind) + " was not a '" + formatSyntaxKind(kind) + "' token."; }, stackCrawlMark || assertOptionalToken);
                }
            }
            Debug.assertOptionalToken = assertOptionalToken;
            function assertMissingNode(node, message, stackCrawlMark) {
                if (shouldAssertFunction(1 /* Normal */, "assertMissingNode")) {
                    assert(node === undefined, message || "Unexpected node.", function () { return "Node " + formatSyntaxKind(node.kind) + " was unexpected'."; }, stackCrawlMark || assertMissingNode);
                }
            }
            Debug.assertMissingNode = assertMissingNode;
            function type(_value) { }
            Debug.type = type;
            function getFunctionName(func) {
                if (typeof func !== "function") {
                    return "";
                }
                else if (func.hasOwnProperty("name")) {
                    return func.name;
                }
                else {
                    var text = Function.prototype.toString.call(func);
                    var match = /^function\s+([\w\$]+)\s*\(/.exec(text);
                    return match ? match[1] : "";
                }
            }
            Debug.getFunctionName = getFunctionName;
            function formatSymbol(symbol) {
                return "{ name: " + ts.unescapeLeadingUnderscores(symbol.escapedName) + "; flags: " + formatSymbolFlags(symbol.flags) + "; declarations: " + ts.map(symbol.declarations, function (node) { return formatSyntaxKind(node.kind); }) + " }";
            }
            Debug.formatSymbol = formatSymbol;
            /**
             * Formats an enum value as a string for debugging and debug assertions.
             */
            function formatEnum(value, enumObject, isFlags) {
                if (value === void 0) {
                    value = 0;
                }
                var members = getEnumMembers(enumObject);
                if (value === 0) {
                    return members.length > 0 && members[0][0] === 0 ? members[0][1] : "0";
                }
                if (isFlags) {
                    var result = "";
                    var remainingFlags = value;
                    for (var _i = 0, members_1 = members; _i < members_1.length; _i++) {
                        var _a = members_1[_i], enumValue = _a[0], enumName = _a[1];
                        if (enumValue > value) {
                            break;
                        }
                        if (enumValue !== 0 && enumValue & value) {
                            result = "" + result + (result ? "|" : "") + enumName;
                            remainingFlags &= ~enumValue;
                        }
                    }
                    if (remainingFlags === 0) {
                        return result;
                    }
                }
                else {
                    for (var _b = 0, members_2 = members; _b < members_2.length; _b++) {
                        var _c = members_2[_b], enumValue = _c[0], enumName = _c[1];
                        if (enumValue === value) {
                            return enumName;
                        }
                    }
                }
                return value.toString();
            }
            Debug.formatEnum = formatEnum;
            function getEnumMembers(enumObject) {
                var result = [];
                for (var name in enumObject) {
                    var value = enumObject[name];
                    if (typeof value === "number") {
                        result.push([value, name]);
                    }
                }
                return ts.stableSort(result, function (x, y) { return ts.compareValues(x[0], y[0]); });
            }
            function formatSyntaxKind(kind) {
                return formatEnum(kind, ts.SyntaxKind, /*isFlags*/ false);
            }
            Debug.formatSyntaxKind = formatSyntaxKind;
            function formatNodeFlags(flags) {
                return formatEnum(flags, ts.NodeFlags, /*isFlags*/ true);
            }
            Debug.formatNodeFlags = formatNodeFlags;
            function formatModifierFlags(flags) {
                return formatEnum(flags, ts.ModifierFlags, /*isFlags*/ true);
            }
            Debug.formatModifierFlags = formatModifierFlags;
            function formatTransformFlags(flags) {
                return formatEnum(flags, ts.TransformFlags, /*isFlags*/ true);
            }
            Debug.formatTransformFlags = formatTransformFlags;
            function formatEmitFlags(flags) {
                return formatEnum(flags, ts.EmitFlags, /*isFlags*/ true);
            }
            Debug.formatEmitFlags = formatEmitFlags;
            function formatSymbolFlags(flags) {
                return formatEnum(flags, ts.SymbolFlags, /*isFlags*/ true);
            }
            Debug.formatSymbolFlags = formatSymbolFlags;
            function formatTypeFlags(flags) {
                return formatEnum(flags, ts.TypeFlags, /*isFlags*/ true);
            }
            Debug.formatTypeFlags = formatTypeFlags;
            function formatSignatureFlags(flags) {
                return formatEnum(flags, ts.SignatureFlags, /*isFlags*/ true);
            }
            Debug.formatSignatureFlags = formatSignatureFlags;
            function formatObjectFlags(flags) {
                return formatEnum(flags, ts.ObjectFlags, /*isFlags*/ true);
            }
            Debug.formatObjectFlags = formatObjectFlags;
            function formatFlowFlags(flags) {
                return formatEnum(flags, ts.FlowFlags, /*isFlags*/ true);
            }
            Debug.formatFlowFlags = formatFlowFlags;
            var isDebugInfoEnabled = false;
            var extendedDebugModule;
            function extendedDebug() {
                enableDebugInfo();
                if (!extendedDebugModule) {
                    throw new Error("Debugging helpers could not be loaded.");
                }
                return extendedDebugModule;
            }
            function printControlFlowGraph(flowNode) {
                return console.log(formatControlFlowGraph(flowNode));
            }
            Debug.printControlFlowGraph = printControlFlowGraph;
            function formatControlFlowGraph(flowNode) {
                return extendedDebug().formatControlFlowGraph(flowNode);
            }
            Debug.formatControlFlowGraph = formatControlFlowGraph;
            var flowNodeProto;
            function attachFlowNodeDebugInfoWorker(flowNode) {
                if (!("__debugFlowFlags" in flowNode)) { // eslint-disable-line no-in-operator
                    Object.defineProperties(flowNode, {
                        // for use with vscode-js-debug's new customDescriptionGenerator in launch.json
                        __tsDebuggerDisplay: {
                            value: function () {
                                var flowHeader = this.flags & 2 /* Start */ ? "FlowStart" :
                                    this.flags & 4 /* BranchLabel */ ? "FlowBranchLabel" :
                                        this.flags & 8 /* LoopLabel */ ? "FlowLoopLabel" :
                                            this.flags & 16 /* Assignment */ ? "FlowAssignment" :
                                                this.flags & 32 /* TrueCondition */ ? "FlowTrueCondition" :
                                                    this.flags & 64 /* FalseCondition */ ? "FlowFalseCondition" :
                                                        this.flags & 128 /* SwitchClause */ ? "FlowSwitchClause" :
                                                            this.flags & 256 /* ArrayMutation */ ? "FlowArrayMutation" :
                                                                this.flags & 512 /* Call */ ? "FlowCall" :
                                                                    this.flags & 1024 /* ReduceLabel */ ? "FlowReduceLabel" :
                                                                        this.flags & 1 /* Unreachable */ ? "FlowUnreachable" :
                                                                            "UnknownFlow";
                                var remainingFlags = this.flags & ~(2048 /* Referenced */ - 1);
                                return "" + flowHeader + (remainingFlags ? " (" + formatFlowFlags(remainingFlags) + ")" : "");
                            }
                        },
                        __debugFlowFlags: { get: function () { return formatEnum(this.flags, ts.FlowFlags, /*isFlags*/ true); } },
                        __debugToString: { value: function () { return formatControlFlowGraph(this); } }
                    });
                }
            }
            function attachFlowNodeDebugInfo(flowNode) {
                if (isDebugInfoEnabled) {
                    if (typeof Object.setPrototypeOf === "function") {
                        // if we're in es2015, attach the method to a shared prototype for `FlowNode`
                        // so the method doesn't show up in the watch window.
                        if (!flowNodeProto) {
                            flowNodeProto = Object.create(Object.prototype);
                            attachFlowNodeDebugInfoWorker(flowNodeProto);
                        }
                        Object.setPrototypeOf(flowNode, flowNodeProto);
                    }
                    else {
                        // not running in an es2015 environment, attach the method directly.
                        attachFlowNodeDebugInfoWorker(flowNode);
                    }
                }
            }
            Debug.attachFlowNodeDebugInfo = attachFlowNodeDebugInfo;
            var nodeArrayProto;
            function attachNodeArrayDebugInfoWorker(array) {
                if (!("__tsDebuggerDisplay" in array)) { // eslint-disable-line no-in-operator
                    Object.defineProperties(array, {
                        __tsDebuggerDisplay: {
                            value: function (defaultValue) {
                                // An `Array` with extra properties is rendered as `[A, B, prop1: 1, prop2: 2]`. Most of
                                // these aren't immediately useful so we trim off the `prop1: ..., prop2: ...` part from the
                                // formatted string.
                                // This regex can trigger slow backtracking because of overlapping potential captures.
                                // We don't care, this is debug code that's only enabled with a debugger attached -
                                // we're just taking note of it for anyone checking regex performance in the future.
                                defaultValue = String(defaultValue).replace(/(?:,[\s\w\d_]+:[^,]+)+\]$/, "]");
                                return "NodeArray " + defaultValue;
                            }
                        }
                    });
                }
            }
            function attachNodeArrayDebugInfo(array) {
                if (isDebugInfoEnabled) {
                    if (typeof Object.setPrototypeOf === "function") {
                        // if we're in es2015, attach the method to a shared prototype for `NodeArray`
                        // so the method doesn't show up in the watch window.
                        if (!nodeArrayProto) {
                            nodeArrayProto = Object.create(Array.prototype);
                            attachNodeArrayDebugInfoWorker(nodeArrayProto);
                        }
                        Object.setPrototypeOf(array, nodeArrayProto);
                    }
                    else {
                        // not running in an es2015 environment, attach the method directly.
                        attachNodeArrayDebugInfoWorker(array);
                    }
                }
            }
            Debug.attachNodeArrayDebugInfo = attachNodeArrayDebugInfo;
            /**
             * Injects debug information into frequently used types.
             */
            function enableDebugInfo() {
                if (isDebugInfoEnabled)
                    return;
                // avoid recomputing
                var weakTypeTextMap;
                var weakNodeTextMap;
                function getWeakTypeTextMap() {
                    if (weakTypeTextMap === undefined) {
                        if (typeof WeakMap === "function")
                            weakTypeTextMap = new WeakMap();
                    }
                    return weakTypeTextMap;
                }
                function getWeakNodeTextMap() {
                    if (weakNodeTextMap === undefined) {
                        if (typeof WeakMap === "function")
                            weakNodeTextMap = new WeakMap();
                    }
                    return weakNodeTextMap;
                }
                // Add additional properties in debug mode to assist with debugging.
                Object.defineProperties(ts.objectAllocator.getSymbolConstructor().prototype, {
                    // for use with vscode-js-debug's new customDescriptionGenerator in launch.json
                    __tsDebuggerDisplay: {
                        value: function () {
                            var symbolHeader = this.flags & 33554432 /* Transient */ ? "TransientSymbol" :
                                "Symbol";
                            var remainingSymbolFlags = this.flags & ~33554432 /* Transient */;
                            return symbolHeader + " '" + ts.symbolName(this) + "'" + (remainingSymbolFlags ? " (" + formatSymbolFlags(remainingSymbolFlags) + ")" : "");
                        }
                    },
                    __debugFlags: { get: function () { return formatSymbolFlags(this.flags); } }
                });
                Object.defineProperties(ts.objectAllocator.getTypeConstructor().prototype, {
                    // for use with vscode-js-debug's new customDescriptionGenerator in launch.json
                    __tsDebuggerDisplay: {
                        value: function () {
                            var typeHeader = this.flags & 98304 /* Nullable */ ? "NullableType" :
                                this.flags & 384 /* StringOrNumberLiteral */ ? "LiteralType " + JSON.stringify(this.value) :
                                    this.flags & 2048 /* BigIntLiteral */ ? "LiteralType " + (this.value.negative ? "-" : "") + this.value.base10Value + "n" :
                                        this.flags & 8192 /* UniqueESSymbol */ ? "UniqueESSymbolType" :
                                            this.flags & 32 /* Enum */ ? "EnumType" :
                                                this.flags & 67359327 /* Intrinsic */ ? "IntrinsicType " + this.intrinsicName :
                                                    this.flags & 1048576 /* Union */ ? "UnionType" :
                                                        this.flags & 2097152 /* Intersection */ ? "IntersectionType" :
                                                            this.flags & 4194304 /* Index */ ? "IndexType" :
                                                                this.flags & 8388608 /* IndexedAccess */ ? "IndexedAccessType" :
                                                                    this.flags & 16777216 /* Conditional */ ? "ConditionalType" :
                                                                        this.flags & 33554432 /* Substitution */ ? "SubstitutionType" :
                                                                            this.flags & 262144 /* TypeParameter */ ? "TypeParameter" :
                                                                                this.flags & 524288 /* Object */ ?
                                                                                    this.objectFlags & 3 /* ClassOrInterface */ ? "InterfaceType" :
                                                                                        this.objectFlags & 4 /* Reference */ ? "TypeReference" :
                                                                                            this.objectFlags & 8 /* Tuple */ ? "TupleType" :
                                                                                                this.objectFlags & 16 /* Anonymous */ ? "AnonymousType" :
                                                                                                    this.objectFlags & 32 /* Mapped */ ? "MappedType" :
                                                                                                        this.objectFlags & 1024 /* ReverseMapped */ ? "ReverseMappedType" :
                                                                                                            this.objectFlags & 256 /* EvolvingArray */ ? "EvolvingArrayType" :
                                                                                                                "ObjectType" :
                                                                                    "Type";
                            var remainingObjectFlags = this.flags & 524288 /* Object */ ? this.objectFlags & ~1343 /* ObjectTypeKindMask */ : 0;
                            return "" + typeHeader + (this.symbol ? " '" + ts.symbolName(this.symbol) + "'" : "") + (remainingObjectFlags ? " (" + formatObjectFlags(remainingObjectFlags) + ")" : "");
                        }
                    },
                    __debugFlags: { get: function () { return formatTypeFlags(this.flags); } },
                    __debugObjectFlags: { get: function () { return this.flags & 524288 /* Object */ ? formatObjectFlags(this.objectFlags) : ""; } },
                    __debugTypeToString: {
                        value: function () {
                            // avoid recomputing
                            var map = getWeakTypeTextMap();
                            var text = map === null || map === void 0 ? void 0 : map.get(this);
                            if (text === undefined) {
                                text = this.checker.typeToString(this);
                                map === null || map === void 0 ? void 0 : map.set(this, text);
                            }
                            return text;
                        }
                    },
                });
                Object.defineProperties(ts.objectAllocator.getSignatureConstructor().prototype, {
                    __debugFlags: { get: function () { return formatSignatureFlags(this.flags); } },
                    __debugSignatureToString: { value: function () { var _a; return (_a = this.checker) === null || _a === void 0 ? void 0 : _a.signatureToString(this); } }
                });
                var nodeConstructors = [
                    ts.objectAllocator.getNodeConstructor(),
                    ts.objectAllocator.getIdentifierConstructor(),
                    ts.objectAllocator.getTokenConstructor(),
                    ts.objectAllocator.getSourceFileConstructor()
                ];
                for (var _i = 0, nodeConstructors_1 = nodeConstructors; _i < nodeConstructors_1.length; _i++) {
                    var ctor = nodeConstructors_1[_i];
                    if (!ctor.prototype.hasOwnProperty("__debugKind")) {
                        Object.defineProperties(ctor.prototype, {
                            // for use with vscode-js-debug's new customDescriptionGenerator in launch.json
                            __tsDebuggerDisplay: {
                                value: function () {
                                    var nodeHeader = ts.isGeneratedIdentifier(this) ? "GeneratedIdentifier" :
                                        ts.isIdentifier(this) ? "Identifier '" + ts.idText(this) + "'" :
                                            ts.isPrivateIdentifier(this) ? "PrivateIdentifier '" + ts.idText(this) + "'" :
                                                ts.isStringLiteral(this) ? "StringLiteral " + JSON.stringify(this.text.length < 10 ? this.text : this.text.slice(10) + "...") :
                                                    ts.isNumericLiteral(this) ? "NumericLiteral " + this.text :
                                                        ts.isBigIntLiteral(this) ? "BigIntLiteral " + this.text + "n" :
                                                            ts.isTypeParameterDeclaration(this) ? "TypeParameterDeclaration" :
                                                                ts.isParameter(this) ? "ParameterDeclaration" :
                                                                    ts.isConstructorDeclaration(this) ? "ConstructorDeclaration" :
                                                                        ts.isGetAccessorDeclaration(this) ? "GetAccessorDeclaration" :
                                                                            ts.isSetAccessorDeclaration(this) ? "SetAccessorDeclaration" :
                                                                                ts.isCallSignatureDeclaration(this) ? "CallSignatureDeclaration" :
                                                                                    ts.isConstructSignatureDeclaration(this) ? "ConstructSignatureDeclaration" :
                                                                                        ts.isIndexSignatureDeclaration(this) ? "IndexSignatureDeclaration" :
                                                                                            ts.isTypePredicateNode(this) ? "TypePredicateNode" :
                                                                                                ts.isTypeReferenceNode(this) ? "TypeReferenceNode" :
                                                                                                    ts.isFunctionTypeNode(this) ? "FunctionTypeNode" :
                                                                                                        ts.isConstructorTypeNode(this) ? "ConstructorTypeNode" :
                                                                                                            ts.isTypeQueryNode(this) ? "TypeQueryNode" :
                                                                                                                ts.isTypeLiteralNode(this) ? "TypeLiteralNode" :
                                                                                                                    ts.isArrayTypeNode(this) ? "ArrayTypeNode" :
                                                                                                                        ts.isTupleTypeNode(this) ? "TupleTypeNode" :
                                                                                                                            ts.isOptionalTypeNode(this) ? "OptionalTypeNode" :
                                                                                                                                ts.isRestTypeNode(this) ? "RestTypeNode" :
                                                                                                                                    ts.isUnionTypeNode(this) ? "UnionTypeNode" :
                                                                                                                                        ts.isIntersectionTypeNode(this) ? "IntersectionTypeNode" :
                                                                                                                                            ts.isConditionalTypeNode(this) ? "ConditionalTypeNode" :
                                                                                                                                                ts.isInferTypeNode(this) ? "InferTypeNode" :
                                                                                                                                                    ts.isParenthesizedTypeNode(this) ? "ParenthesizedTypeNode" :
                                                                                                                                                        ts.isThisTypeNode(this) ? "ThisTypeNode" :
                                                                                                                                                            ts.isTypeOperatorNode(this) ? "TypeOperatorNode" :
                                                                                                                                                                ts.isIndexedAccessTypeNode(this) ? "IndexedAccessTypeNode" :
                                                                                                                                                                    ts.isMappedTypeNode(this) ? "MappedTypeNode" :
                                                                                                                                                                        ts.isLiteralTypeNode(this) ? "LiteralTypeNode" :
                                                                                                                                                                            ts.isNamedTupleMember(this) ? "NamedTupleMember" :
                                                                                                                                                                                ts.isImportTypeNode(this) ? "ImportTypeNode" :
                                                                                                                                                                                    formatSyntaxKind(this.kind);
                                    return "" + nodeHeader + (this.flags ? " (" + formatNodeFlags(this.flags) + ")" : "");
                                }
                            },
                            __debugKind: { get: function () { return formatSyntaxKind(this.kind); } },
                            __debugNodeFlags: { get: function () { return formatNodeFlags(this.flags); } },
                            __debugModifierFlags: { get: function () { return formatModifierFlags(ts.getEffectiveModifierFlagsNoCache(this)); } },
                            __debugTransformFlags: { get: function () { return formatTransformFlags(this.transformFlags); } },
                            __debugIsParseTreeNode: { get: function () { return ts.isParseTreeNode(this); } },
                            __debugEmitFlags: { get: function () { return formatEmitFlags(ts.getEmitFlags(this)); } },
                            __debugGetText: {
                                value: function (includeTrivia) {
                                    if (ts.nodeIsSynthesized(this))
                                        return "";
                                    // avoid recomputing
                                    var map = getWeakNodeTextMap();
                                    var text = map === null || map === void 0 ? void 0 : map.get(this);
                                    if (text === undefined) {
                                        var parseNode = ts.getParseTreeNode(this);
                                        var sourceFile = parseNode && ts.getSourceFileOfNode(parseNode);
                                        text = sourceFile ? ts.getSourceTextOfNodeFromSourceFile(sourceFile, parseNode, includeTrivia) : "";
                                        map === null || map === void 0 ? void 0 : map.set(this, text);
                                    }
                                    return text;
                                }
                            }
                        });
                    }
                }
                // attempt to load extended debugging information
                try {
                    if (ts.sys && ts.sys.require) {
                        var basePath = ts.getDirectoryPath(ts.resolvePath(ts.sys.getExecutingFilePath()));
                        var result = ts.sys.require(basePath, "./compiler-debug");
                        if (!result.error) {
                            result.module.init(ts);
                            extendedDebugModule = result.module;
                        }
                    }
                }
                catch (_a) {
                    // do nothing
                }
                isDebugInfoEnabled = true;
            }
            Debug.enableDebugInfo = enableDebugInfo;
            function formatDeprecationMessage(name, error, errorAfter, since, message) {
                var deprecationMessage = error ? "DeprecationError: " : "DeprecationWarning: ";
                deprecationMessage += "'" + name + "' ";
                deprecationMessage += since ? "has been deprecated since v" + since : "is deprecated";
                deprecationMessage += error ? " and can no longer be used." : errorAfter ? " and will no longer be usable after v" + errorAfter + "." : ".";
                deprecationMessage += message ? " " + ts.formatStringFromArgs(message, [name], 0) : "";
                return deprecationMessage;
            }
            function createErrorDeprecation(name, errorAfter, since, message) {
                var deprecationMessage = formatDeprecationMessage(name, /*error*/ true, errorAfter, since, message);
                return function () {
                    throw new TypeError(deprecationMessage);
                };
            }
            function createWarningDeprecation(name, errorAfter, since, message) {
                var hasWrittenDeprecation = false;
                return function () {
                    if (!hasWrittenDeprecation) {
                        log.warn(formatDeprecationMessage(name, /*error*/ false, errorAfter, since, message));
                        hasWrittenDeprecation = true;
                    }
                };
            }
            function createDeprecation(name, options) {
                var _a, _b;
                if (options === void 0) {
                    options = {};
                }
                var version = typeof options.typeScriptVersion === "string" ? new ts.Version(options.typeScriptVersion) : (_a = options.typeScriptVersion) !== null && _a !== void 0 ? _a : getTypeScriptVersion();
                var errorAfter = typeof options.errorAfter === "string" ? new ts.Version(options.errorAfter) : options.errorAfter;
                var warnAfter = typeof options.warnAfter === "string" ? new ts.Version(options.warnAfter) : options.warnAfter;
                var since = typeof options.since === "string" ? new ts.Version(options.since) : (_b = options.since) !== null && _b !== void 0 ? _b : warnAfter;
                var error = options.error || errorAfter && version.compareTo(errorAfter) <= 0;
                var warn = !warnAfter || version.compareTo(warnAfter) >= 0;
                return error ? createErrorDeprecation(name, errorAfter, since, options.message) :
                    warn ? createWarningDeprecation(name, errorAfter, since, options.message) :
                        ts.noop;
            }
            function wrapFunction(deprecation, func) {
                return function () {
                    deprecation();
                    return func.apply(this, arguments);
                };
            }
            function deprecate(func, options) {
                var deprecation = createDeprecation(getFunctionName(func), options);
                return wrapFunction(deprecation, func);
            }
            Debug.deprecate = deprecate;
        })(Debug = ts.Debug || (ts.Debug = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        // https://semver.org/#spec-item-2
        // > A normal version number MUST take the form X.Y.Z where X, Y, and Z are non-negative
        // > integers, and MUST NOT contain leading zeroes. X is the major version, Y is the minor
        // > version, and Z is the patch version. Each element MUST increase numerically.
        //
        // NOTE: We differ here in that we allow X and X.Y, with missing parts having the default
        // value of `0`.
        var versionRegExp = /^(0|[1-9]\d*)(?:\.(0|[1-9]\d*)(?:\.(0|[1-9]\d*)(?:\-([a-z0-9-.]+))?(?:\+([a-z0-9-.]+))?)?)?$/i;
        // https://semver.org/#spec-item-9
        // > A pre-release version MAY be denoted by appending a hyphen and a series of dot separated
        // > identifiers immediately following the patch version. Identifiers MUST comprise only ASCII
        // > alphanumerics and hyphen [0-9A-Za-z-]. Identifiers MUST NOT be empty. Numeric identifiers
        // > MUST NOT include leading zeroes.
        var prereleaseRegExp = /^(?:0|[1-9]\d*|[a-z-][a-z0-9-]*)(?:\.(?:0|[1-9]\d*|[a-z-][a-z0-9-]*))*$/i;
        // https://semver.org/#spec-item-10
        // > Build metadata MAY be denoted by appending a plus sign and a series of dot separated
        // > identifiers immediately following the patch or pre-release version. Identifiers MUST
        // > comprise only ASCII alphanumerics and hyphen [0-9A-Za-z-]. Identifiers MUST NOT be empty.
        var buildRegExp = /^[a-z0-9-]+(?:\.[a-z0-9-]+)*$/i;
        // https://semver.org/#spec-item-9
        // > Numeric identifiers MUST NOT include leading zeroes.
        var numericIdentifierRegExp = /^(0|[1-9]\d*)$/;
        /**
         * Describes a precise semantic version number, https://semver.org
         */
        var Version = /** @class */ (function () {
            function Version(major, minor, patch, prerelease, build) {
                if (minor === void 0) {
                    minor = 0;
                }
                if (patch === void 0) {
                    patch = 0;
                }
                if (prerelease === void 0) {
                    prerelease = "";
                }
                if (build === void 0) {
                    build = "";
                }
                if (typeof major === "string") {
                    var result = ts.Debug.checkDefined(tryParseComponents(major), "Invalid version");
                    (major = result.major, minor = result.minor, patch = result.patch, prerelease = result.prerelease, build = result.build);
                }
                ts.Debug.assert(major >= 0, "Invalid argument: major");
                ts.Debug.assert(minor >= 0, "Invalid argument: minor");
                ts.Debug.assert(patch >= 0, "Invalid argument: patch");
                ts.Debug.assert(!prerelease || prereleaseRegExp.test(prerelease), "Invalid argument: prerelease");
                ts.Debug.assert(!build || buildRegExp.test(build), "Invalid argument: build");
                this.major = major;
                this.minor = minor;
                this.patch = patch;
                this.prerelease = prerelease ? prerelease.split(".") : ts.emptyArray;
                this.build = build ? build.split(".") : ts.emptyArray;
            }
            Version.tryParse = function (text) {
                var result = tryParseComponents(text);
                if (!result)
                    return undefined;
                var major = result.major, minor = result.minor, patch = result.patch, prerelease = result.prerelease, build = result.build;
                return new Version(major, minor, patch, prerelease, build);
            };
            Version.prototype.compareTo = function (other) {
                // https://semver.org/#spec-item-11
                // > Precedence is determined by the first difference when comparing each of these
                // > identifiers from left to right as follows: Major, minor, and patch versions are
                // > always compared numerically.
                //
                // https://semver.org/#spec-item-11
                // > Precedence for two pre-release versions with the same major, minor, and patch version
                // > MUST be determined by comparing each dot separated identifier from left to right until
                // > a difference is found [...]
                //
                // https://semver.org/#spec-item-11
                // > Build metadata does not figure into precedence
                if (this === other)
                    return 0 /* EqualTo */;
                if (other === undefined)
                    return 1 /* GreaterThan */;
                return ts.compareValues(this.major, other.major)
                    || ts.compareValues(this.minor, other.minor)
                    || ts.compareValues(this.patch, other.patch)
                    || comparePrereleaseIdentifiers(this.prerelease, other.prerelease);
            };
            Version.prototype.increment = function (field) {
                switch (field) {
                    case "major": return new Version(this.major + 1, 0, 0);
                    case "minor": return new Version(this.major, this.minor + 1, 0);
                    case "patch": return new Version(this.major, this.minor, this.patch + 1);
                    default: return ts.Debug.assertNever(field);
                }
            };
            Version.prototype.toString = function () {
                var result = this.major + "." + this.minor + "." + this.patch;
                if (ts.some(this.prerelease))
                    result += "-" + this.prerelease.join(".");
                if (ts.some(this.build))
                    result += "+" + this.build.join(".");
                return result;
            };
            Version.zero = new Version(0, 0, 0);
            return Version;
        }());
        ts.Version = Version;
        function tryParseComponents(text) {
            var match = versionRegExp.exec(text);
            if (!match)
                return undefined;
            var major = match[1], _a = match[2], minor = _a === void 0 ? "0" : _a, _b = match[3], patch = _b === void 0 ? "0" : _b, _c = match[4], prerelease = _c === void 0 ? "" : _c, _d = match[5], build = _d === void 0 ? "" : _d;
            if (prerelease && !prereleaseRegExp.test(prerelease))
                return undefined;
            if (build && !buildRegExp.test(build))
                return undefined;
            return {
                major: parseInt(major, 10),
                minor: parseInt(minor, 10),
                patch: parseInt(patch, 10),
                prerelease: prerelease,
                build: build
            };
        }
        function comparePrereleaseIdentifiers(left, right) {
            // https://semver.org/#spec-item-11
            // > When major, minor, and patch are equal, a pre-release version has lower precedence
            // > than a normal version.
            if (left === right)
                return 0 /* EqualTo */;
            if (left.length === 0)
                return right.length === 0 ? 0 /* EqualTo */ : 1 /* GreaterThan */;
            if (right.length === 0)
                return -1 /* LessThan */;
            // https://semver.org/#spec-item-11
            // > Precedence for two pre-release versions with the same major, minor, and patch version
            // > MUST be determined by comparing each dot separated identifier from left to right until
            // > a difference is found [...]
            var length = Math.min(left.length, right.length);
            for (var i = 0; i < length; i++) {
                var leftIdentifier = left[i];
                var rightIdentifier = right[i];
                if (leftIdentifier === rightIdentifier)
                    continue;
                var leftIsNumeric = numericIdentifierRegExp.test(leftIdentifier);
                var rightIsNumeric = numericIdentifierRegExp.test(rightIdentifier);
                if (leftIsNumeric || rightIsNumeric) {
                    // https://semver.org/#spec-item-11
                    // > Numeric identifiers always have lower precedence than non-numeric identifiers.
                    if (leftIsNumeric !== rightIsNumeric)
                        return leftIsNumeric ? -1 /* LessThan */ : 1 /* GreaterThan */;
                    // https://semver.org/#spec-item-11
                    // > identifiers consisting of only digits are compared numerically
                    var result = ts.compareValues(+leftIdentifier, +rightIdentifier);
                    if (result)
                        return result;
                }
                else {
                    // https://semver.org/#spec-item-11
                    // > identifiers with letters or hyphens are compared lexically in ASCII sort order.
                    var result = ts.compareStringsCaseSensitive(leftIdentifier, rightIdentifier);
                    if (result)
                        return result;
                }
            }
            // https://semver.org/#spec-item-11
            // > A larger set of pre-release fields has a higher precedence than a smaller set, if all
            // > of the preceding identifiers are equal.
            return ts.compareValues(left.length, right.length);
        }
        /**
         * Describes a semantic version range, per https://github.com/npm/node-semver#ranges
         */
        var VersionRange = /** @class */ (function () {
            function VersionRange(spec) {
                this._alternatives = spec ? ts.Debug.checkDefined(parseRange(spec), "Invalid range spec.") : ts.emptyArray;
            }
            VersionRange.tryParse = function (text) {
                var sets = parseRange(text);
                if (sets) {
                    var range = new VersionRange("");
                    range._alternatives = sets;
                    return range;
                }
                return undefined;
            };
            VersionRange.prototype.test = function (version) {
                if (typeof version === "string")
                    version = new Version(version);
                return testDisjunction(version, this._alternatives);
            };
            VersionRange.prototype.toString = function () {
                return formatDisjunction(this._alternatives);
            };
            return VersionRange;
        }());
        ts.VersionRange = VersionRange;
        // https://github.com/npm/node-semver#range-grammar
        //
        // range-set    ::= range ( logical-or range ) *
        // range        ::= hyphen | simple ( ' ' simple ) * | ''
        // logical-or   ::= ( ' ' ) * '||' ( ' ' ) *
        var logicalOrRegExp = /\|\|/g;
        var whitespaceRegExp = /\s+/g;
        // https://github.com/npm/node-semver#range-grammar
        //
        // partial      ::= xr ( '.' xr ( '.' xr qualifier ? )? )?
        // xr           ::= 'x' | 'X' | '*' | nr
        // nr           ::= '0' | ['1'-'9'] ( ['0'-'9'] ) *
        // qualifier    ::= ( '-' pre )? ( '+' build )?
        // pre          ::= parts
        // build        ::= parts
        // parts        ::= part ( '.' part ) *
        // part         ::= nr | [-0-9A-Za-z]+
        var partialRegExp = /^([xX*0]|[1-9]\d*)(?:\.([xX*0]|[1-9]\d*)(?:\.([xX*0]|[1-9]\d*)(?:-([a-z0-9-.]+))?(?:\+([a-z0-9-.]+))?)?)?$/i;
        // https://github.com/npm/node-semver#range-grammar
        //
        // hyphen       ::= partial ' - ' partial
        var hyphenRegExp = /^\s*([a-z0-9-+.*]+)\s+-\s+([a-z0-9-+.*]+)\s*$/i;
        // https://github.com/npm/node-semver#range-grammar
        //
        // simple       ::= primitive | partial | tilde | caret
        // primitive    ::= ( '<' | '>' | '>=' | '<=' | '=' ) partial
        // tilde        ::= '~' partial
        // caret        ::= '^' partial
        var rangeRegExp = /^(~|\^|<|<=|>|>=|=)?\s*([a-z0-9-+.*]+)$/i;
        function parseRange(text) {
            var alternatives = [];
            for (var _i = 0, _a = ts.trimString(text).split(logicalOrRegExp); _i < _a.length; _i++) {
                var range = _a[_i];
                if (!range)
                    continue;
                var comparators = [];
                range = ts.trimString(range);
                var match = hyphenRegExp.exec(range);
                if (match) {
                    if (!parseHyphen(match[1], match[2], comparators))
                        return undefined;
                }
                else {
                    for (var _b = 0, _c = range.split(whitespaceRegExp); _b < _c.length; _b++) {
                        var simple = _c[_b];
                        var match_1 = rangeRegExp.exec(ts.trimString(simple));
                        if (!match_1 || !parseComparator(match_1[1], match_1[2], comparators))
                            return undefined;
                    }
                }
                alternatives.push(comparators);
            }
            return alternatives;
        }
        function parsePartial(text) {
            var match = partialRegExp.exec(text);
            if (!match)
                return undefined;
            var major = match[1], _a = match[2], minor = _a === void 0 ? "*" : _a, _b = match[3], patch = _b === void 0 ? "*" : _b, prerelease = match[4], build = match[5];
            var version = new Version(isWildcard(major) ? 0 : parseInt(major, 10), isWildcard(major) || isWildcard(minor) ? 0 : parseInt(minor, 10), isWildcard(major) || isWildcard(minor) || isWildcard(patch) ? 0 : parseInt(patch, 10), prerelease, build);
            return { version: version, major: major, minor: minor, patch: patch };
        }
        function parseHyphen(left, right, comparators) {
            var leftResult = parsePartial(left);
            if (!leftResult)
                return false;
            var rightResult = parsePartial(right);
            if (!rightResult)
                return false;
            if (!isWildcard(leftResult.major)) {
                comparators.push(createComparator(">=", leftResult.version));
            }
            if (!isWildcard(rightResult.major)) {
                comparators.push(isWildcard(rightResult.minor) ? createComparator("<", rightResult.version.increment("major")) :
                    isWildcard(rightResult.patch) ? createComparator("<", rightResult.version.increment("minor")) :
                        createComparator("<=", rightResult.version));
            }
            return true;
        }
        function parseComparator(operator, text, comparators) {
            var result = parsePartial(text);
            if (!result)
                return false;
            var version = result.version, major = result.major, minor = result.minor, patch = result.patch;
            if (!isWildcard(major)) {
                switch (operator) {
                    case "~":
                        comparators.push(createComparator(">=", version));
                        comparators.push(createComparator("<", version.increment(isWildcard(minor) ? "major" :
                            "minor")));
                        break;
                    case "^":
                        comparators.push(createComparator(">=", version));
                        comparators.push(createComparator("<", version.increment(version.major > 0 || isWildcard(minor) ? "major" :
                            version.minor > 0 || isWildcard(patch) ? "minor" :
                                "patch")));
                        break;
                    case "<":
                    case ">=":
                        comparators.push(createComparator(operator, version));
                        break;
                    case "<=":
                    case ">":
                        comparators.push(isWildcard(minor) ? createComparator(operator === "<=" ? "<" : ">=", version.increment("major")) :
                            isWildcard(patch) ? createComparator(operator === "<=" ? "<" : ">=", version.increment("minor")) :
                                createComparator(operator, version));
                        break;
                    case "=":
                    case undefined:
                        if (isWildcard(minor) || isWildcard(patch)) {
                            comparators.push(createComparator(">=", version));
                            comparators.push(createComparator("<", version.increment(isWildcard(minor) ? "major" : "minor")));
                        }
                        else {
                            comparators.push(createComparator("=", version));
                        }
                        break;
                    default:
                        // unrecognized
                        return false;
                }
            }
            else if (operator === "<" || operator === ">") {
                comparators.push(createComparator("<", Version.zero));
            }
            return true;
        }
        function isWildcard(part) {
            return part === "*" || part === "x" || part === "X";
        }
        function createComparator(operator, operand) {
            return { operator: operator, operand: operand };
        }
        function testDisjunction(version, alternatives) {
            // an empty disjunction is treated as "*" (all versions)
            if (alternatives.length === 0)
                return true;
            for (var _i = 0, alternatives_1 = alternatives; _i < alternatives_1.length; _i++) {
                var alternative = alternatives_1[_i];
                if (testAlternative(version, alternative))
                    return true;
            }
            return false;
        }
        function testAlternative(version, comparators) {
            for (var _i = 0, comparators_1 = comparators; _i < comparators_1.length; _i++) {
                var comparator = comparators_1[_i];
                if (!testComparator(version, comparator.operator, comparator.operand))
                    return false;
            }
            return true;
        }
        function testComparator(version, operator, operand) {
            var cmp = version.compareTo(operand);
            switch (operator) {
                case "<": return cmp < 0;
                case "<=": return cmp <= 0;
                case ">": return cmp > 0;
                case ">=": return cmp >= 0;
                case "=": return cmp === 0;
                default: return ts.Debug.assertNever(operator);
            }
        }
        function formatDisjunction(alternatives) {
            return ts.map(alternatives, formatAlternative).join(" || ") || "*";
        }
        function formatAlternative(comparators) {
            return ts.map(comparators, formatComparator).join(" ");
        }
        function formatComparator(comparator) {
            return "" + comparator.operator + comparator.operand;
        }
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        // The following definitions provide the minimum compatible support for the Web Performance User Timings API
        // between browsers and NodeJS:
        // eslint-disable-next-line @typescript-eslint/naming-convention
        function hasRequiredAPI(performance, PerformanceObserver) {
            return typeof performance === "object" &&
                typeof performance.timeOrigin === "number" &&
                typeof performance.mark === "function" &&
                typeof performance.measure === "function" &&
                typeof performance.now === "function" &&
                typeof PerformanceObserver === "function";
        }
        function tryGetWebPerformanceHooks() {
            if (typeof performance === "object" &&
                typeof PerformanceObserver === "function" &&
                hasRequiredAPI(performance, PerformanceObserver)) {
                return {
                    // For now we always write native performance events when running in the browser. We may
                    // make this conditional in the future if we find that native web performance hooks
                    // in the browser also slow down compilation.
                    shouldWriteNativeEvents: true,
                    performance: performance,
                    PerformanceObserver: PerformanceObserver
                };
            }
        }
        function tryGetNodePerformanceHooks() {
            if (typeof process !== "undefined" && process.nextTick && !process.browser && typeof module === "object" && typeof require === "function") {
                try {
                    var performance_1;
                    var _a = require("perf_hooks"), nodePerformance_1 = _a.performance, PerformanceObserver_1 = _a.PerformanceObserver;
                    if (hasRequiredAPI(nodePerformance_1, PerformanceObserver_1)) {
                        performance_1 = nodePerformance_1;
                        // There is a bug in Node's performance.measure prior to 12.16.3/13.13.0 that does not
                        // match the Web Performance API specification. Node's implementation did not allow
                        // optional `start` and `end` arguments for `performance.measure`.
                        // See https://github.com/nodejs/node/pull/32651 for more information.
                        var version_1 = new ts.Version(process.versions.node);
                        var range = new ts.VersionRange("<12.16.3 || 13 <13.13");
                        if (range.test(version_1)) {
                            performance_1 = {
                                get timeOrigin() { return nodePerformance_1.timeOrigin; },
                                now: function () { return nodePerformance_1.now(); },
                                mark: function (name) { return nodePerformance_1.mark(name); },
                                measure: function (name, start, end) {
                                    if (start === void 0) {
                                        start = "nodeStart";
                                    }
                                    if (end === undefined) {
                                        end = "__performance.measure-fix__";
                                        nodePerformance_1.mark(end);
                                    }
                                    nodePerformance_1.measure(name, start, end);
                                    if (end === "__performance.measure-fix__") {
                                        nodePerformance_1.clearMarks("__performance.measure-fix__");
                                    }
                                }
                            };
                        }
                        return {
                            // By default, only write native events when generating a cpu profile or using the v8 profiler.
                            shouldWriteNativeEvents: false,
                            performance: performance_1,
                            PerformanceObserver: PerformanceObserver_1
                        };
                    }
                }
                catch (_b) {
                    // ignore errors
                }
            }
        }
        // Unlike with the native Map/Set 'tryGet' functions in corePublic.ts, we eagerly evaluate these
        // since we will need them for `timestamp`, below.
        var nativePerformanceHooks = tryGetWebPerformanceHooks() || tryGetNodePerformanceHooks();
        var nativePerformance = nativePerformanceHooks === null || nativePerformanceHooks === void 0 ? void 0 : nativePerformanceHooks.performance;
        function tryGetNativePerformanceHooks() {
            return nativePerformanceHooks;
        }
        ts.tryGetNativePerformanceHooks = tryGetNativePerformanceHooks;
        /** Gets a timestamp with (at least) ms resolution */
        ts.timestamp = nativePerformance ? function () { return nativePerformance.now(); } :
            Date.now ? Date.now :
                function () { return +(new Date()); };
    })(ts || (ts = {}));
    /*@internal*/
    /** Performance measurements for the compiler. */
    var ts;
    (function (ts) {
        var performance;
        (function (performance) {
            var perfHooks;
            // when set, indicates the implementation of `Performance` to use for user timing.
            // when unset, indicates user timing is unavailable or disabled.
            var performanceImpl;
            function createTimerIf(condition, measureName, startMarkName, endMarkName) {
                return condition ? createTimer(measureName, startMarkName, endMarkName) : performance.nullTimer;
            }
            performance.createTimerIf = createTimerIf;
            function createTimer(measureName, startMarkName, endMarkName) {
                var enterCount = 0;
                return {
                    enter: enter,
                    exit: exit
                };
                function enter() {
                    if (++enterCount === 1) {
                        mark(startMarkName);
                    }
                }
                function exit() {
                    if (--enterCount === 0) {
                        mark(endMarkName);
                        measure(measureName, startMarkName, endMarkName);
                    }
                    else if (enterCount < 0) {
                        ts.Debug.fail("enter/exit count does not match.");
                    }
                }
            }
            performance.createTimer = createTimer;
            performance.nullTimer = { enter: ts.noop, exit: ts.noop };
            var enabled = false;
            var timeorigin = ts.timestamp();
            var marks = new ts.Map();
            var counts = new ts.Map();
            var durations = new ts.Map();
            /**
             * Marks a performance event.
             *
             * @param markName The name of the mark.
             */
            function mark(markName) {
                var _a;
                if (enabled) {
                    var count = (_a = counts.get(markName)) !== null && _a !== void 0 ? _a : 0;
                    counts.set(markName, count + 1);
                    marks.set(markName, ts.timestamp());
                    performanceImpl === null || performanceImpl === void 0 ? void 0 : performanceImpl.mark(markName);
                }
            }
            performance.mark = mark;
            /**
             * Adds a performance measurement with the specified name.
             *
             * @param measureName The name of the performance measurement.
             * @param startMarkName The name of the starting mark. If not supplied, the point at which the
             *      profiler was enabled is used.
             * @param endMarkName The name of the ending mark. If not supplied, the current timestamp is
             *      used.
             */
            function measure(measureName, startMarkName, endMarkName) {
                var _a, _b;
                if (enabled) {
                    var end = (_a = (endMarkName !== undefined ? marks.get(endMarkName) : undefined)) !== null && _a !== void 0 ? _a : ts.timestamp();
                    var start = (_b = (startMarkName !== undefined ? marks.get(startMarkName) : undefined)) !== null && _b !== void 0 ? _b : timeorigin;
                    var previousDuration = durations.get(measureName) || 0;
                    durations.set(measureName, previousDuration + (end - start));
                    performanceImpl === null || performanceImpl === void 0 ? void 0 : performanceImpl.measure(measureName, startMarkName, endMarkName);
                }
            }
            performance.measure = measure;
            /**
             * Gets the number of times a marker was encountered.
             *
             * @param markName The name of the mark.
             */
            function getCount(markName) {
                return counts.get(markName) || 0;
            }
            performance.getCount = getCount;
            /**
             * Gets the total duration of all measurements with the supplied name.
             *
             * @param measureName The name of the measure whose durations should be accumulated.
             */
            function getDuration(measureName) {
                return durations.get(measureName) || 0;
            }
            performance.getDuration = getDuration;
            /**
             * Iterate over each measure, performing some action
             *
             * @param cb The action to perform for each measure
             */
            function forEachMeasure(cb) {
                durations.forEach(function (duration, measureName) { return cb(measureName, duration); });
            }
            performance.forEachMeasure = forEachMeasure;
            /**
             * Indicates whether the performance API is enabled.
             */
            function isEnabled() {
                return enabled;
            }
            performance.isEnabled = isEnabled;
            /** Enables (and resets) performance measurements for the compiler. */
            function enable(system) {
                var _a;
                if (system === void 0) {
                    system = ts.sys;
                }
                if (!enabled) {
                    enabled = true;
                    perfHooks || (perfHooks = ts.tryGetNativePerformanceHooks());
                    if (perfHooks) {
                        timeorigin = perfHooks.performance.timeOrigin;
                        // NodeJS's Web Performance API is currently slower than expected, but we'd still like
                        // to be able to leverage native trace events when node is run with either `--cpu-prof`
                        // or `--prof`, if we're running with our own `--generateCpuProfile` flag, or when
                        // running in debug mode (since its possible to generate a cpu profile while debugging).
                        if (perfHooks.shouldWriteNativeEvents || ((_a = system === null || system === void 0 ? void 0 : system.cpuProfilingEnabled) === null || _a === void 0 ? void 0 : _a.call(system)) || (system === null || system === void 0 ? void 0 : system.debugMode)) {
                            performanceImpl = perfHooks.performance;
                        }
                    }
                }
                return true;
            }
            performance.enable = enable;
            /** Disables performance measurements for the compiler. */
            function disable() {
                if (enabled) {
                    marks.clear();
                    counts.clear();
                    durations.clear();
                    performanceImpl = undefined;
                    enabled = false;
                }
            }
            performance.disable = disable;
        })(performance = ts.performance || (ts.performance = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var _a;
        var nullLogger = {
            logEvent: ts.noop,
            logErrEvent: ts.noop,
            logPerfEvent: ts.noop,
            logInfoEvent: ts.noop,
            logStartCommand: ts.noop,
            logStopCommand: ts.noop,
            logStartUpdateProgram: ts.noop,
            logStopUpdateProgram: ts.noop,
            logStartUpdateGraph: ts.noop,
            logStopUpdateGraph: ts.noop,
            logStartResolveModule: ts.noop,
            logStopResolveModule: ts.noop,
            logStartParseSourceFile: ts.noop,
            logStopParseSourceFile: ts.noop,
            logStartReadFile: ts.noop,
            logStopReadFile: ts.noop,
            logStartBindFile: ts.noop,
            logStopBindFile: ts.noop,
            logStartScheduledOperation: ts.noop,
            logStopScheduledOperation: ts.noop,
        };
        // Load optional module to enable Event Tracing for Windows
        // See https://github.com/microsoft/typescript-etw for more information
        var etwModule;
        try {
            var etwModulePath = (_a = process.env.TS_ETW_MODULE_PATH) !== null && _a !== void 0 ? _a : "./node_modules/@microsoft/typescript-etw";
            // require() will throw an exception if the module is not found
            // It may also return undefined if not installed properly
            etwModule = require(etwModulePath);
        }
        catch (e) {
            etwModule = undefined;
        }
        /** Performance logger that will generate ETW events if possible - check for `logEvent` member, as `etwModule` will be `{}` when browserified */
        ts.perfLogger = etwModule && etwModule.logEvent ? etwModule : nullLogger;
    })(ts || (ts = {}));
    /* Tracing events for the compiler. */
    /*@internal*/
    var ts;
    (function (ts) {
        // enable the above using startTracing()
        // `tracingEnabled` should never be used directly, only through the above
        var tracingEnabled;
        (function (tracingEnabled) {
            var fs;
            var traceCount = 0;
            var traceFd = 0;
            var mode;
            var typeCatalog = []; // NB: id is index + 1
            var legendPath;
            var legend = [];
            ;
            /** Starts tracing for the given project. */
            function startTracing(tracingMode, traceDir, configFilePath) {
                ts.Debug.assert(!ts.tracing, "Tracing already started");
                if (fs === undefined) {
                    try {
                        fs = require("fs");
                    }
                    catch (e) {
                        throw new Error("tracing requires having fs\n(original error: " + (e.message || e) + ")");
                    }
                }
                mode = tracingMode;
                typeCatalog.length = 0;
                if (legendPath === undefined) {
                    legendPath = ts.combinePaths(traceDir, "legend.json");
                }
                // Note that writing will fail later on if it exists and is not a directory
                if (!fs.existsSync(traceDir)) {
                    fs.mkdirSync(traceDir, { recursive: true });
                }
                var countPart = mode === "build" ? "." + process.pid + "-" + ++traceCount
                    : mode === "server" ? "." + process.pid
                        : "";
                var tracePath = ts.combinePaths(traceDir, "trace" + countPart + ".json");
                var typesPath = ts.combinePaths(traceDir, "types" + countPart + ".json");
                legend.push({
                    configFilePath: configFilePath,
                    tracePath: tracePath,
                    typesPath: typesPath,
                });
                traceFd = fs.openSync(tracePath, "w");
                ts.tracing = tracingEnabled; // only when traceFd is properly set
                // Start with a prefix that contains some metadata that the devtools profiler expects (also avoids a warning on import)
                var meta = { cat: "__metadata", ph: "M", ts: 1000 * ts.timestamp(), pid: 1, tid: 1 };
                fs.writeSync(traceFd, "[\n"
                    + [__assign({ name: "process_name", args: { name: "tsc" } }, meta), __assign({ name: "thread_name", args: { name: "Main" } }, meta), __assign(__assign({ name: "TracingStartedInBrowser" }, meta), { cat: "disabled-by-default-devtools.timeline" })]
                        .map(function (v) { return JSON.stringify(v); }).join(",\n"));
            }
            tracingEnabled.startTracing = startTracing;
            /** Stops tracing for the in-progress project and dumps the type catalog. */
            function stopTracing() {
                ts.Debug.assert(ts.tracing, "Tracing is not in progress");
                ts.Debug.assert(!!typeCatalog.length === (mode !== "server")); // Have a type catalog iff not in server mode
                fs.writeSync(traceFd, "\n]\n");
                fs.closeSync(traceFd);
                ts.tracing = undefined;
                if (typeCatalog.length) {
                    dumpTypes(typeCatalog);
                }
                else {
                    // We pre-computed this path for convenience, but clear it
                    // now that the file won't be created.
                    legend[legend.length - 1].typesPath = undefined;
                }
            }
            tracingEnabled.stopTracing = stopTracing;
            function recordType(type) {
                if (mode !== "server") {
                    typeCatalog.push(type);
                }
            }
            tracingEnabled.recordType = recordType;
            var Phase;
            (function (Phase) {
                Phase["Parse"] = "parse";
                Phase["Program"] = "program";
                Phase["Bind"] = "bind";
                Phase["Check"] = "check";
                Phase["CheckTypes"] = "checkTypes";
                Phase["Emit"] = "emit";
                Phase["Session"] = "session";
            })(Phase = tracingEnabled.Phase || (tracingEnabled.Phase = {}));
            function instant(phase, name, args) {
                writeEvent("I", phase, name, args, "\"s\":\"g\"");
            }
            tracingEnabled.instant = instant;
            var eventStack = [];
            /**
             * @param separateBeginAndEnd - used for special cases where we need the trace point even if the event
             * never terminates (typically for reducing a scenario too big to trace to one that can be completed).
             * In the future we might implement an exit handler to dump unfinished events which would deprecate
             * these operations.
             */
            function push(phase, name, args, separateBeginAndEnd) {
                if (separateBeginAndEnd === void 0) {
                    separateBeginAndEnd = false;
                }
                if (separateBeginAndEnd) {
                    writeEvent("B", phase, name, args);
                }
                eventStack.push({ phase: phase, name: name, args: args, time: 1000 * ts.timestamp(), separateBeginAndEnd: separateBeginAndEnd });
            }
            tracingEnabled.push = push;
            function pop() {
                ts.Debug.assert(eventStack.length > 0);
                writeStackEvent(eventStack.length - 1, 1000 * ts.timestamp());
                eventStack.length--;
            }
            tracingEnabled.pop = pop;
            function popAll() {
                var endTime = 1000 * ts.timestamp();
                for (var i = eventStack.length - 1; i >= 0; i--) {
                    writeStackEvent(i, endTime);
                }
                eventStack.length = 0;
            }
            tracingEnabled.popAll = popAll;
            // sample every 10ms
            var sampleInterval = 1000 * 10;
            function writeStackEvent(index, endTime) {
                var _a = eventStack[index], phase = _a.phase, name = _a.name, args = _a.args, time = _a.time, separateBeginAndEnd = _a.separateBeginAndEnd;
                if (separateBeginAndEnd) {
                    writeEvent("E", phase, name, args, /*extras*/ undefined, endTime);
                }
                // test if [time,endTime) straddles a sampling point
                else if (sampleInterval - (time % sampleInterval) <= endTime - time) {
                    writeEvent("X", phase, name, args, "\"dur\":" + (endTime - time), time);
                }
            }
            function writeEvent(eventType, phase, name, args, extras, time) {
                if (time === void 0) {
                    time = 1000 * ts.timestamp();
                }
                // In server mode, there's no easy way to dump type information, so we drop events that would require it.
                if (mode === "server" && phase === "checkTypes" /* CheckTypes */)
                    return;
                ts.performance.mark("beginTracing");
                fs.writeSync(traceFd, ",\n{\"pid\":1,\"tid\":1,\"ph\":\"" + eventType + "\",\"cat\":\"" + phase + "\",\"ts\":" + time + ",\"name\":\"" + name + "\"");
                if (extras)
                    fs.writeSync(traceFd, "," + extras);
                if (args)
                    fs.writeSync(traceFd, ",\"args\":" + JSON.stringify(args));
                fs.writeSync(traceFd, "}");
                ts.performance.mark("endTracing");
                ts.performance.measure("Tracing", "beginTracing", "endTracing");
            }
            function getLocation(node) {
                var file = ts.getSourceFileOfNode(node);
                return !file
                    ? undefined
                    : {
                        path: file.path,
                        start: indexFromOne(ts.getLineAndCharacterOfPosition(file, node.pos)),
                        end: indexFromOne(ts.getLineAndCharacterOfPosition(file, node.end)),
                    };
                function indexFromOne(lc) {
                    return {
                        line: lc.line + 1,
                        character: lc.character + 1,
                    };
                }
            }
            function dumpTypes(types) {
                var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s, _t, _u, _v, _w, _x;
                ts.performance.mark("beginDumpTypes");
                var typesPath = legend[legend.length - 1].typesPath;
                var typesFd = fs.openSync(typesPath, "w");
                var recursionIdentityMap = new ts.Map();
                // Cleverness: no line break here so that the type ID will match the line number
                fs.writeSync(typesFd, "[");
                var numTypes = types.length;
                for (var i = 0; i < numTypes; i++) {
                    var type = types[i];
                    var objectFlags = type.objectFlags;
                    var symbol = (_a = type.aliasSymbol) !== null && _a !== void 0 ? _a : type.symbol;
                    // It's slow to compute the display text, so skip it unless it's really valuable (or cheap)
                    var display = void 0;
                    if ((objectFlags & 16 /* Anonymous */) | (type.flags & 2944 /* Literal */)) {
                        try {
                            display = (_b = type.checker) === null || _b === void 0 ? void 0 : _b.typeToString(type);
                        }
                        catch (_y) {
                            display = undefined;
                        }
                    }
                    var indexedAccessProperties = {};
                    if (type.flags & 8388608 /* IndexedAccess */) {
                        var indexedAccessType = type;
                        indexedAccessProperties = {
                            indexedAccessObjectType: (_c = indexedAccessType.objectType) === null || _c === void 0 ? void 0 : _c.id,
                            indexedAccessIndexType: (_d = indexedAccessType.indexType) === null || _d === void 0 ? void 0 : _d.id,
                        };
                    }
                    var referenceProperties = {};
                    if (objectFlags & 4 /* Reference */) {
                        var referenceType = type;
                        referenceProperties = {
                            instantiatedType: (_e = referenceType.target) === null || _e === void 0 ? void 0 : _e.id,
                            typeArguments: (_f = referenceType.resolvedTypeArguments) === null || _f === void 0 ? void 0 : _f.map(function (t) { return t.id; }),
                            referenceLocation: getLocation(referenceType.node),
                        };
                    }
                    var conditionalProperties = {};
                    if (type.flags & 16777216 /* Conditional */) {
                        var conditionalType = type;
                        conditionalProperties = {
                            conditionalCheckType: (_g = conditionalType.checkType) === null || _g === void 0 ? void 0 : _g.id,
                            conditionalExtendsType: (_h = conditionalType.extendsType) === null || _h === void 0 ? void 0 : _h.id,
                            conditionalTrueType: (_k = (_j = conditionalType.resolvedTrueType) === null || _j === void 0 ? void 0 : _j.id) !== null && _k !== void 0 ? _k : -1,
                            conditionalFalseType: (_m = (_l = conditionalType.resolvedFalseType) === null || _l === void 0 ? void 0 : _l.id) !== null && _m !== void 0 ? _m : -1,
                        };
                    }
                    var substitutionProperties = {};
                    if (type.flags & 33554432 /* Substitution */) {
                        var substitutionType = type;
                        substitutionProperties = {
                            substitutionBaseType: (_o = substitutionType.baseType) === null || _o === void 0 ? void 0 : _o.id,
                            substituteType: (_p = substitutionType.substitute) === null || _p === void 0 ? void 0 : _p.id,
                        };
                    }
                    var reverseMappedProperties = {};
                    if (objectFlags & 1024 /* ReverseMapped */) {
                        var reverseMappedType = type;
                        reverseMappedProperties = {
                            reverseMappedSourceType: (_q = reverseMappedType.source) === null || _q === void 0 ? void 0 : _q.id,
                            reverseMappedMappedType: (_r = reverseMappedType.mappedType) === null || _r === void 0 ? void 0 : _r.id,
                            reverseMappedConstraintType: (_s = reverseMappedType.constraintType) === null || _s === void 0 ? void 0 : _s.id,
                        };
                    }
                    var evolvingArrayProperties = {};
                    if (objectFlags & 256 /* EvolvingArray */) {
                        var evolvingArrayType = type;
                        evolvingArrayProperties = {
                            evolvingArrayElementType: evolvingArrayType.elementType.id,
                            evolvingArrayFinalType: (_t = evolvingArrayType.finalArrayType) === null || _t === void 0 ? void 0 : _t.id,
                        };
                    }
                    // We can't print out an arbitrary object, so just assign each one a unique number.
                    // Don't call it an "id" so people don't treat it as a type id.
                    var recursionToken = void 0;
                    var recursionIdentity = type.checker.getRecursionIdentity(type);
                    if (recursionIdentity) {
                        recursionToken = recursionIdentityMap.get(recursionIdentity);
                        if (!recursionToken) {
                            recursionToken = recursionIdentityMap.size;
                            recursionIdentityMap.set(recursionIdentity, recursionToken);
                        }
                    }
                    var descriptor = __assign(__assign(__assign(__assign(__assign(__assign(__assign({ id: type.id, intrinsicName: type.intrinsicName, symbolName: (symbol === null || symbol === void 0 ? void 0 : symbol.escapedName) && ts.unescapeLeadingUnderscores(symbol.escapedName), recursionId: recursionToken, isTuple: objectFlags & 8 /* Tuple */ ? true : undefined, unionTypes: (type.flags & 1048576 /* Union */) ? (_u = type.types) === null || _u === void 0 ? void 0 : _u.map(function (t) { return t.id; }) : undefined, intersectionTypes: (type.flags & 2097152 /* Intersection */) ? type.types.map(function (t) { return t.id; }) : undefined, aliasTypeArguments: (_v = type.aliasTypeArguments) === null || _v === void 0 ? void 0 : _v.map(function (t) { return t.id; }), keyofType: (type.flags & 4194304 /* Index */) ? (_w = type.type) === null || _w === void 0 ? void 0 : _w.id : undefined }, indexedAccessProperties), referenceProperties), conditionalProperties), substitutionProperties), reverseMappedProperties), evolvingArrayProperties), { destructuringPattern: getLocation(type.pattern), firstDeclaration: getLocation((_x = symbol === null || symbol === void 0 ? void 0 : symbol.declarations) === null || _x === void 0 ? void 0 : _x[0]), flags: ts.Debug.formatTypeFlags(type.flags).split("|"), display: display });
                    fs.writeSync(typesFd, JSON.stringify(descriptor));
                    if (i < numTypes - 1) {
                        fs.writeSync(typesFd, ",\n");
                    }
                }
                fs.writeSync(typesFd, "]\n");
                fs.closeSync(typesFd);
                ts.performance.mark("endDumpTypes");
                ts.performance.measure("Dump types", "beginDumpTypes", "endDumpTypes");
            }
            function dumpLegend() {
                if (!legendPath) {
                    return;
                }
                fs.writeFileSync(legendPath, JSON.stringify(legend));
            }
            tracingEnabled.dumpLegend = dumpLegend;
        })(tracingEnabled || (tracingEnabled = {}));
        // define after tracingEnabled is initialized
        ts.startTracing = tracingEnabled.startTracing;
        ts.dumpTracingLegend = tracingEnabled.dumpLegend;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        // token > SyntaxKind.Identifier => token is a keyword
        // Also, If you add a new SyntaxKind be sure to keep the `Markers` section at the bottom in sync
        var SyntaxKind;
        (function (SyntaxKind) {
            SyntaxKind[SyntaxKind["Unknown"] = 0] = "Unknown";
            SyntaxKind[SyntaxKind["EndOfFileToken"] = 1] = "EndOfFileToken";
            SyntaxKind[SyntaxKind["SingleLineCommentTrivia"] = 2] = "SingleLineCommentTrivia";
            SyntaxKind[SyntaxKind["MultiLineCommentTrivia"] = 3] = "MultiLineCommentTrivia";
            SyntaxKind[SyntaxKind["NewLineTrivia"] = 4] = "NewLineTrivia";
            SyntaxKind[SyntaxKind["WhitespaceTrivia"] = 5] = "WhitespaceTrivia";
            // We detect and preserve #! on the first line
            SyntaxKind[SyntaxKind["ShebangTrivia"] = 6] = "ShebangTrivia";
            // We detect and provide better error recovery when we encounter a git merge marker.  This
            // allows us to edit files with git-conflict markers in them in a much more pleasant manner.
            SyntaxKind[SyntaxKind["ConflictMarkerTrivia"] = 7] = "ConflictMarkerTrivia";
            // Literals
            SyntaxKind[SyntaxKind["NumericLiteral"] = 8] = "NumericLiteral";
            SyntaxKind[SyntaxKind["BigIntLiteral"] = 9] = "BigIntLiteral";
            SyntaxKind[SyntaxKind["StringLiteral"] = 10] = "StringLiteral";
            SyntaxKind[SyntaxKind["JsxText"] = 11] = "JsxText";
            SyntaxKind[SyntaxKind["JsxTextAllWhiteSpaces"] = 12] = "JsxTextAllWhiteSpaces";
            SyntaxKind[SyntaxKind["RegularExpressionLiteral"] = 13] = "RegularExpressionLiteral";
            SyntaxKind[SyntaxKind["NoSubstitutionTemplateLiteral"] = 14] = "NoSubstitutionTemplateLiteral";
            // Pseudo-literals
            SyntaxKind[SyntaxKind["TemplateHead"] = 15] = "TemplateHead";
            SyntaxKind[SyntaxKind["TemplateMiddle"] = 16] = "TemplateMiddle";
            SyntaxKind[SyntaxKind["TemplateTail"] = 17] = "TemplateTail";
            // Punctuation
            SyntaxKind[SyntaxKind["OpenBraceToken"] = 18] = "OpenBraceToken";
            SyntaxKind[SyntaxKind["CloseBraceToken"] = 19] = "CloseBraceToken";
            SyntaxKind[SyntaxKind["OpenParenToken"] = 20] = "OpenParenToken";
            SyntaxKind[SyntaxKind["CloseParenToken"] = 21] = "CloseParenToken";
            SyntaxKind[SyntaxKind["OpenBracketToken"] = 22] = "OpenBracketToken";
            SyntaxKind[SyntaxKind["CloseBracketToken"] = 23] = "CloseBracketToken";
            SyntaxKind[SyntaxKind["DotToken"] = 24] = "DotToken";
            SyntaxKind[SyntaxKind["DotDotDotToken"] = 25] = "DotDotDotToken";
            SyntaxKind[SyntaxKind["SemicolonToken"] = 26] = "SemicolonToken";
            SyntaxKind[SyntaxKind["CommaToken"] = 27] = "CommaToken";
            SyntaxKind[SyntaxKind["QuestionDotToken"] = 28] = "QuestionDotToken";
            SyntaxKind[SyntaxKind["LessThanToken"] = 29] = "LessThanToken";
            SyntaxKind[SyntaxKind["LessThanSlashToken"] = 30] = "LessThanSlashToken";
            SyntaxKind[SyntaxKind["GreaterThanToken"] = 31] = "GreaterThanToken";
            SyntaxKind[SyntaxKind["LessThanEqualsToken"] = 32] = "LessThanEqualsToken";
            SyntaxKind[SyntaxKind["GreaterThanEqualsToken"] = 33] = "GreaterThanEqualsToken";
            SyntaxKind[SyntaxKind["EqualsEqualsToken"] = 34] = "EqualsEqualsToken";
            SyntaxKind[SyntaxKind["ExclamationEqualsToken"] = 35] = "ExclamationEqualsToken";
            SyntaxKind[SyntaxKind["EqualsEqualsEqualsToken"] = 36] = "EqualsEqualsEqualsToken";
            SyntaxKind[SyntaxKind["ExclamationEqualsEqualsToken"] = 37] = "ExclamationEqualsEqualsToken";
            SyntaxKind[SyntaxKind["EqualsGreaterThanToken"] = 38] = "EqualsGreaterThanToken";
            SyntaxKind[SyntaxKind["PlusToken"] = 39] = "PlusToken";
            SyntaxKind[SyntaxKind["MinusToken"] = 40] = "MinusToken";
            SyntaxKind[SyntaxKind["AsteriskToken"] = 41] = "AsteriskToken";
            SyntaxKind[SyntaxKind["AsteriskAsteriskToken"] = 42] = "AsteriskAsteriskToken";
            SyntaxKind[SyntaxKind["SlashToken"] = 43] = "SlashToken";
            SyntaxKind[SyntaxKind["PercentToken"] = 44] = "PercentToken";
            SyntaxKind[SyntaxKind["PlusPlusToken"] = 45] = "PlusPlusToken";
            SyntaxKind[SyntaxKind["MinusMinusToken"] = 46] = "MinusMinusToken";
            SyntaxKind[SyntaxKind["LessThanLessThanToken"] = 47] = "LessThanLessThanToken";
            SyntaxKind[SyntaxKind["GreaterThanGreaterThanToken"] = 48] = "GreaterThanGreaterThanToken";
            SyntaxKind[SyntaxKind["GreaterThanGreaterThanGreaterThanToken"] = 49] = "GreaterThanGreaterThanGreaterThanToken";
            SyntaxKind[SyntaxKind["AmpersandToken"] = 50] = "AmpersandToken";
            SyntaxKind[SyntaxKind["BarToken"] = 51] = "BarToken";
            SyntaxKind[SyntaxKind["CaretToken"] = 52] = "CaretToken";
            SyntaxKind[SyntaxKind["ExclamationToken"] = 53] = "ExclamationToken";
            SyntaxKind[SyntaxKind["TildeToken"] = 54] = "TildeToken";
            SyntaxKind[SyntaxKind["AmpersandAmpersandToken"] = 55] = "AmpersandAmpersandToken";
            SyntaxKind[SyntaxKind["BarBarToken"] = 56] = "BarBarToken";
            SyntaxKind[SyntaxKind["QuestionToken"] = 57] = "QuestionToken";
            SyntaxKind[SyntaxKind["ColonToken"] = 58] = "ColonToken";
            SyntaxKind[SyntaxKind["AtToken"] = 59] = "AtToken";
            SyntaxKind[SyntaxKind["QuestionQuestionToken"] = 60] = "QuestionQuestionToken";
            /** Only the JSDoc scanner produces BacktickToken. The normal scanner produces NoSubstitutionTemplateLiteral and related kinds. */
            SyntaxKind[SyntaxKind["BacktickToken"] = 61] = "BacktickToken";
            /** Only the JSDoc scanner produces HashToken. The normal scanner produces PrivateIdentifier. */
            SyntaxKind[SyntaxKind["HashToken"] = 62] = "HashToken";
            // Assignments
            SyntaxKind[SyntaxKind["EqualsToken"] = 63] = "EqualsToken";
            SyntaxKind[SyntaxKind["PlusEqualsToken"] = 64] = "PlusEqualsToken";
            SyntaxKind[SyntaxKind["MinusEqualsToken"] = 65] = "MinusEqualsToken";
            SyntaxKind[SyntaxKind["AsteriskEqualsToken"] = 66] = "AsteriskEqualsToken";
            SyntaxKind[SyntaxKind["AsteriskAsteriskEqualsToken"] = 67] = "AsteriskAsteriskEqualsToken";
            SyntaxKind[SyntaxKind["SlashEqualsToken"] = 68] = "SlashEqualsToken";
            SyntaxKind[SyntaxKind["PercentEqualsToken"] = 69] = "PercentEqualsToken";
            SyntaxKind[SyntaxKind["LessThanLessThanEqualsToken"] = 70] = "LessThanLessThanEqualsToken";
            SyntaxKind[SyntaxKind["GreaterThanGreaterThanEqualsToken"] = 71] = "GreaterThanGreaterThanEqualsToken";
            SyntaxKind[SyntaxKind["GreaterThanGreaterThanGreaterThanEqualsToken"] = 72] = "GreaterThanGreaterThanGreaterThanEqualsToken";
            SyntaxKind[SyntaxKind["AmpersandEqualsToken"] = 73] = "AmpersandEqualsToken";
            SyntaxKind[SyntaxKind["BarEqualsToken"] = 74] = "BarEqualsToken";
            SyntaxKind[SyntaxKind["BarBarEqualsToken"] = 75] = "BarBarEqualsToken";
            SyntaxKind[SyntaxKind["AmpersandAmpersandEqualsToken"] = 76] = "AmpersandAmpersandEqualsToken";
            SyntaxKind[SyntaxKind["QuestionQuestionEqualsToken"] = 77] = "QuestionQuestionEqualsToken";
            SyntaxKind[SyntaxKind["CaretEqualsToken"] = 78] = "CaretEqualsToken";
            // Identifiers and PrivateIdentifiers
            SyntaxKind[SyntaxKind["Identifier"] = 79] = "Identifier";
            SyntaxKind[SyntaxKind["PrivateIdentifier"] = 80] = "PrivateIdentifier";
            // Reserved words
            SyntaxKind[SyntaxKind["BreakKeyword"] = 81] = "BreakKeyword";
            SyntaxKind[SyntaxKind["CaseKeyword"] = 82] = "CaseKeyword";
            SyntaxKind[SyntaxKind["CatchKeyword"] = 83] = "CatchKeyword";
            SyntaxKind[SyntaxKind["ClassKeyword"] = 84] = "ClassKeyword";
            SyntaxKind[SyntaxKind["ConstKeyword"] = 85] = "ConstKeyword";
            SyntaxKind[SyntaxKind["ContinueKeyword"] = 86] = "ContinueKeyword";
            SyntaxKind[SyntaxKind["DebuggerKeyword"] = 87] = "DebuggerKeyword";
            SyntaxKind[SyntaxKind["DefaultKeyword"] = 88] = "DefaultKeyword";
            SyntaxKind[SyntaxKind["DeleteKeyword"] = 89] = "DeleteKeyword";
            SyntaxKind[SyntaxKind["DoKeyword"] = 90] = "DoKeyword";
            SyntaxKind[SyntaxKind["ElseKeyword"] = 91] = "ElseKeyword";
            SyntaxKind[SyntaxKind["EnumKeyword"] = 92] = "EnumKeyword";
            SyntaxKind[SyntaxKind["ExportKeyword"] = 93] = "ExportKeyword";
            SyntaxKind[SyntaxKind["ExtendsKeyword"] = 94] = "ExtendsKeyword";
            SyntaxKind[SyntaxKind["FalseKeyword"] = 95] = "FalseKeyword";
            SyntaxKind[SyntaxKind["FinallyKeyword"] = 96] = "FinallyKeyword";
            SyntaxKind[SyntaxKind["ForKeyword"] = 97] = "ForKeyword";
            SyntaxKind[SyntaxKind["FunctionKeyword"] = 98] = "FunctionKeyword";
            SyntaxKind[SyntaxKind["IfKeyword"] = 99] = "IfKeyword";
            SyntaxKind[SyntaxKind["ImportKeyword"] = 100] = "ImportKeyword";
            SyntaxKind[SyntaxKind["InKeyword"] = 101] = "InKeyword";
            SyntaxKind[SyntaxKind["InstanceOfKeyword"] = 102] = "InstanceOfKeyword";
            SyntaxKind[SyntaxKind["NewKeyword"] = 103] = "NewKeyword";
            SyntaxKind[SyntaxKind["NullKeyword"] = 104] = "NullKeyword";
            SyntaxKind[SyntaxKind["ReturnKeyword"] = 105] = "ReturnKeyword";
            SyntaxKind[SyntaxKind["SuperKeyword"] = 106] = "SuperKeyword";
            SyntaxKind[SyntaxKind["SwitchKeyword"] = 107] = "SwitchKeyword";
            SyntaxKind[SyntaxKind["ThisKeyword"] = 108] = "ThisKeyword";
            SyntaxKind[SyntaxKind["ThrowKeyword"] = 109] = "ThrowKeyword";
            SyntaxKind[SyntaxKind["TrueKeyword"] = 110] = "TrueKeyword";
            SyntaxKind[SyntaxKind["TryKeyword"] = 111] = "TryKeyword";
            SyntaxKind[SyntaxKind["TypeOfKeyword"] = 112] = "TypeOfKeyword";
            SyntaxKind[SyntaxKind["VarKeyword"] = 113] = "VarKeyword";
            SyntaxKind[SyntaxKind["VoidKeyword"] = 114] = "VoidKeyword";
            SyntaxKind[SyntaxKind["WhileKeyword"] = 115] = "WhileKeyword";
            SyntaxKind[SyntaxKind["WithKeyword"] = 116] = "WithKeyword";
            // Strict mode reserved words
            SyntaxKind[SyntaxKind["ImplementsKeyword"] = 117] = "ImplementsKeyword";
            SyntaxKind[SyntaxKind["InterfaceKeyword"] = 118] = "InterfaceKeyword";
            SyntaxKind[SyntaxKind["LetKeyword"] = 119] = "LetKeyword";
            SyntaxKind[SyntaxKind["PackageKeyword"] = 120] = "PackageKeyword";
            SyntaxKind[SyntaxKind["PrivateKeyword"] = 121] = "PrivateKeyword";
            SyntaxKind[SyntaxKind["ProtectedKeyword"] = 122] = "ProtectedKeyword";
            SyntaxKind[SyntaxKind["PublicKeyword"] = 123] = "PublicKeyword";
            SyntaxKind[SyntaxKind["StaticKeyword"] = 124] = "StaticKeyword";
            SyntaxKind[SyntaxKind["YieldKeyword"] = 125] = "YieldKeyword";
            // Contextual keywords
            SyntaxKind[SyntaxKind["AbstractKeyword"] = 126] = "AbstractKeyword";
            SyntaxKind[SyntaxKind["AsKeyword"] = 127] = "AsKeyword";
            SyntaxKind[SyntaxKind["AssertsKeyword"] = 128] = "AssertsKeyword";
            SyntaxKind[SyntaxKind["AnyKeyword"] = 129] = "AnyKeyword";
            SyntaxKind[SyntaxKind["AsyncKeyword"] = 130] = "AsyncKeyword";
            SyntaxKind[SyntaxKind["AwaitKeyword"] = 131] = "AwaitKeyword";
            SyntaxKind[SyntaxKind["BooleanKeyword"] = 132] = "BooleanKeyword";
            SyntaxKind[SyntaxKind["ConstructorKeyword"] = 133] = "ConstructorKeyword";
            SyntaxKind[SyntaxKind["DeclareKeyword"] = 134] = "DeclareKeyword";
            SyntaxKind[SyntaxKind["GetKeyword"] = 135] = "GetKeyword";
            SyntaxKind[SyntaxKind["InferKeyword"] = 136] = "InferKeyword";
            SyntaxKind[SyntaxKind["IntrinsicKeyword"] = 137] = "IntrinsicKeyword";
            SyntaxKind[SyntaxKind["IsKeyword"] = 138] = "IsKeyword";
            SyntaxKind[SyntaxKind["KeyOfKeyword"] = 139] = "KeyOfKeyword";
            SyntaxKind[SyntaxKind["ModuleKeyword"] = 140] = "ModuleKeyword";
            SyntaxKind[SyntaxKind["NamespaceKeyword"] = 141] = "NamespaceKeyword";
            SyntaxKind[SyntaxKind["NeverKeyword"] = 142] = "NeverKeyword";
            SyntaxKind[SyntaxKind["ReadonlyKeyword"] = 143] = "ReadonlyKeyword";
            SyntaxKind[SyntaxKind["RequireKeyword"] = 144] = "RequireKeyword";
            SyntaxKind[SyntaxKind["NumberKeyword"] = 145] = "NumberKeyword";
            SyntaxKind[SyntaxKind["ObjectKeyword"] = 146] = "ObjectKeyword";
            SyntaxKind[SyntaxKind["SetKeyword"] = 147] = "SetKeyword";
            SyntaxKind[SyntaxKind["StringKeyword"] = 148] = "StringKeyword";
            SyntaxKind[SyntaxKind["SymbolKeyword"] = 149] = "SymbolKeyword";
            SyntaxKind[SyntaxKind["TypeKeyword"] = 150] = "TypeKeyword";
            SyntaxKind[SyntaxKind["UndefinedKeyword"] = 151] = "UndefinedKeyword";
            SyntaxKind[SyntaxKind["UniqueKeyword"] = 152] = "UniqueKeyword";
            SyntaxKind[SyntaxKind["UnknownKeyword"] = 153] = "UnknownKeyword";
            SyntaxKind[SyntaxKind["FromKeyword"] = 154] = "FromKeyword";
            SyntaxKind[SyntaxKind["GlobalKeyword"] = 155] = "GlobalKeyword";
            SyntaxKind[SyntaxKind["BigIntKeyword"] = 156] = "BigIntKeyword";
            SyntaxKind[SyntaxKind["OverrideKeyword"] = 157] = "OverrideKeyword";
            SyntaxKind[SyntaxKind["OfKeyword"] = 158] = "OfKeyword";
            // Parse tree nodes
            // Names
            SyntaxKind[SyntaxKind["QualifiedName"] = 159] = "QualifiedName";
            SyntaxKind[SyntaxKind["ComputedPropertyName"] = 160] = "ComputedPropertyName";
            // Signature elements
            SyntaxKind[SyntaxKind["TypeParameter"] = 161] = "TypeParameter";
            SyntaxKind[SyntaxKind["Parameter"] = 162] = "Parameter";
            SyntaxKind[SyntaxKind["Decorator"] = 163] = "Decorator";
            // TypeMember
            SyntaxKind[SyntaxKind["PropertySignature"] = 164] = "PropertySignature";
            SyntaxKind[SyntaxKind["PropertyDeclaration"] = 165] = "PropertyDeclaration";
            SyntaxKind[SyntaxKind["MethodSignature"] = 166] = "MethodSignature";
            SyntaxKind[SyntaxKind["MethodDeclaration"] = 167] = "MethodDeclaration";
            SyntaxKind[SyntaxKind["ClassStaticBlockDeclaration"] = 168] = "ClassStaticBlockDeclaration";
            SyntaxKind[SyntaxKind["Constructor"] = 169] = "Constructor";
            SyntaxKind[SyntaxKind["GetAccessor"] = 170] = "GetAccessor";
            SyntaxKind[SyntaxKind["SetAccessor"] = 171] = "SetAccessor";
            SyntaxKind[SyntaxKind["CallSignature"] = 172] = "CallSignature";
            SyntaxKind[SyntaxKind["ConstructSignature"] = 173] = "ConstructSignature";
            SyntaxKind[SyntaxKind["IndexSignature"] = 174] = "IndexSignature";
            // Type
            SyntaxKind[SyntaxKind["TypePredicate"] = 175] = "TypePredicate";
            SyntaxKind[SyntaxKind["TypeReference"] = 176] = "TypeReference";
            SyntaxKind[SyntaxKind["FunctionType"] = 177] = "FunctionType";
            SyntaxKind[SyntaxKind["ConstructorType"] = 178] = "ConstructorType";
            SyntaxKind[SyntaxKind["TypeQuery"] = 179] = "TypeQuery";
            SyntaxKind[SyntaxKind["TypeLiteral"] = 180] = "TypeLiteral";
            SyntaxKind[SyntaxKind["ArrayType"] = 181] = "ArrayType";
            SyntaxKind[SyntaxKind["TupleType"] = 182] = "TupleType";
            SyntaxKind[SyntaxKind["OptionalType"] = 183] = "OptionalType";
            SyntaxKind[SyntaxKind["RestType"] = 184] = "RestType";
            SyntaxKind[SyntaxKind["UnionType"] = 185] = "UnionType";
            SyntaxKind[SyntaxKind["IntersectionType"] = 186] = "IntersectionType";
            SyntaxKind[SyntaxKind["ConditionalType"] = 187] = "ConditionalType";
            SyntaxKind[SyntaxKind["InferType"] = 188] = "InferType";
            SyntaxKind[SyntaxKind["ParenthesizedType"] = 189] = "ParenthesizedType";
            SyntaxKind[SyntaxKind["ThisType"] = 190] = "ThisType";
            SyntaxKind[SyntaxKind["TypeOperator"] = 191] = "TypeOperator";
            SyntaxKind[SyntaxKind["IndexedAccessType"] = 192] = "IndexedAccessType";
            SyntaxKind[SyntaxKind["MappedType"] = 193] = "MappedType";
            SyntaxKind[SyntaxKind["LiteralType"] = 194] = "LiteralType";
            SyntaxKind[SyntaxKind["NamedTupleMember"] = 195] = "NamedTupleMember";
            SyntaxKind[SyntaxKind["TemplateLiteralType"] = 196] = "TemplateLiteralType";
            SyntaxKind[SyntaxKind["TemplateLiteralTypeSpan"] = 197] = "TemplateLiteralTypeSpan";
            SyntaxKind[SyntaxKind["ImportType"] = 198] = "ImportType";
            // Binding patterns
            SyntaxKind[SyntaxKind["ObjectBindingPattern"] = 199] = "ObjectBindingPattern";
            SyntaxKind[SyntaxKind["ArrayBindingPattern"] = 200] = "ArrayBindingPattern";
            SyntaxKind[SyntaxKind["BindingElement"] = 201] = "BindingElement";
            // Expression
            SyntaxKind[SyntaxKind["ArrayLiteralExpression"] = 202] = "ArrayLiteralExpression";
            SyntaxKind[SyntaxKind["ObjectLiteralExpression"] = 203] = "ObjectLiteralExpression";
            SyntaxKind[SyntaxKind["PropertyAccessExpression"] = 204] = "PropertyAccessExpression";
            SyntaxKind[SyntaxKind["ElementAccessExpression"] = 205] = "ElementAccessExpression";
            SyntaxKind[SyntaxKind["CallExpression"] = 206] = "CallExpression";
            SyntaxKind[SyntaxKind["NewExpression"] = 207] = "NewExpression";
            SyntaxKind[SyntaxKind["TaggedTemplateExpression"] = 208] = "TaggedTemplateExpression";
            SyntaxKind[SyntaxKind["TypeAssertionExpression"] = 209] = "TypeAssertionExpression";
            SyntaxKind[SyntaxKind["ParenthesizedExpression"] = 210] = "ParenthesizedExpression";
            SyntaxKind[SyntaxKind["FunctionExpression"] = 211] = "FunctionExpression";
            SyntaxKind[SyntaxKind["ArrowFunction"] = 212] = "ArrowFunction";
            SyntaxKind[SyntaxKind["DeleteExpression"] = 213] = "DeleteExpression";
            SyntaxKind[SyntaxKind["TypeOfExpression"] = 214] = "TypeOfExpression";
            SyntaxKind[SyntaxKind["VoidExpression"] = 215] = "VoidExpression";
            SyntaxKind[SyntaxKind["AwaitExpression"] = 216] = "AwaitExpression";
            SyntaxKind[SyntaxKind["PrefixUnaryExpression"] = 217] = "PrefixUnaryExpression";
            SyntaxKind[SyntaxKind["PostfixUnaryExpression"] = 218] = "PostfixUnaryExpression";
            SyntaxKind[SyntaxKind["BinaryExpression"] = 219] = "BinaryExpression";
            SyntaxKind[SyntaxKind["ConditionalExpression"] = 220] = "ConditionalExpression";
            SyntaxKind[SyntaxKind["TemplateExpression"] = 221] = "TemplateExpression";
            SyntaxKind[SyntaxKind["YieldExpression"] = 222] = "YieldExpression";
            SyntaxKind[SyntaxKind["SpreadElement"] = 223] = "SpreadElement";
            SyntaxKind[SyntaxKind["ClassExpression"] = 224] = "ClassExpression";
            SyntaxKind[SyntaxKind["OmittedExpression"] = 225] = "OmittedExpression";
            SyntaxKind[SyntaxKind["ExpressionWithTypeArguments"] = 226] = "ExpressionWithTypeArguments";
            SyntaxKind[SyntaxKind["AsExpression"] = 227] = "AsExpression";
            SyntaxKind[SyntaxKind["NonNullExpression"] = 228] = "NonNullExpression";
            SyntaxKind[SyntaxKind["MetaProperty"] = 229] = "MetaProperty";
            SyntaxKind[SyntaxKind["SyntheticExpression"] = 230] = "SyntheticExpression";
            // Misc
            SyntaxKind[SyntaxKind["TemplateSpan"] = 231] = "TemplateSpan";
            SyntaxKind[SyntaxKind["SemicolonClassElement"] = 232] = "SemicolonClassElement";
            // Element
            SyntaxKind[SyntaxKind["Block"] = 233] = "Block";
            SyntaxKind[SyntaxKind["EmptyStatement"] = 234] = "EmptyStatement";
            SyntaxKind[SyntaxKind["VariableStatement"] = 235] = "VariableStatement";
            SyntaxKind[SyntaxKind["ExpressionStatement"] = 236] = "ExpressionStatement";
            SyntaxKind[SyntaxKind["IfStatement"] = 237] = "IfStatement";
            SyntaxKind[SyntaxKind["DoStatement"] = 238] = "DoStatement";
            SyntaxKind[SyntaxKind["WhileStatement"] = 239] = "WhileStatement";
            SyntaxKind[SyntaxKind["ForStatement"] = 240] = "ForStatement";
            SyntaxKind[SyntaxKind["ForInStatement"] = 241] = "ForInStatement";
            SyntaxKind[SyntaxKind["ForOfStatement"] = 242] = "ForOfStatement";
            SyntaxKind[SyntaxKind["ContinueStatement"] = 243] = "ContinueStatement";
            SyntaxKind[SyntaxKind["BreakStatement"] = 244] = "BreakStatement";
            SyntaxKind[SyntaxKind["ReturnStatement"] = 245] = "ReturnStatement";
            SyntaxKind[SyntaxKind["WithStatement"] = 246] = "WithStatement";
            SyntaxKind[SyntaxKind["SwitchStatement"] = 247] = "SwitchStatement";
            SyntaxKind[SyntaxKind["LabeledStatement"] = 248] = "LabeledStatement";
            SyntaxKind[SyntaxKind["ThrowStatement"] = 249] = "ThrowStatement";
            SyntaxKind[SyntaxKind["TryStatement"] = 250] = "TryStatement";
            SyntaxKind[SyntaxKind["DebuggerStatement"] = 251] = "DebuggerStatement";
            SyntaxKind[SyntaxKind["VariableDeclaration"] = 252] = "VariableDeclaration";
            SyntaxKind[SyntaxKind["VariableDeclarationList"] = 253] = "VariableDeclarationList";
            SyntaxKind[SyntaxKind["FunctionDeclaration"] = 254] = "FunctionDeclaration";
            SyntaxKind[SyntaxKind["ClassDeclaration"] = 255] = "ClassDeclaration";
            SyntaxKind[SyntaxKind["InterfaceDeclaration"] = 256] = "InterfaceDeclaration";
            SyntaxKind[SyntaxKind["TypeAliasDeclaration"] = 257] = "TypeAliasDeclaration";
            SyntaxKind[SyntaxKind["EnumDeclaration"] = 258] = "EnumDeclaration";
            SyntaxKind[SyntaxKind["ModuleDeclaration"] = 259] = "ModuleDeclaration";
            SyntaxKind[SyntaxKind["ModuleBlock"] = 260] = "ModuleBlock";
            SyntaxKind[SyntaxKind["CaseBlock"] = 261] = "CaseBlock";
            SyntaxKind[SyntaxKind["NamespaceExportDeclaration"] = 262] = "NamespaceExportDeclaration";
            SyntaxKind[SyntaxKind["ImportEqualsDeclaration"] = 263] = "ImportEqualsDeclaration";
            SyntaxKind[SyntaxKind["ImportDeclaration"] = 264] = "ImportDeclaration";
            SyntaxKind[SyntaxKind["ImportClause"] = 265] = "ImportClause";
            SyntaxKind[SyntaxKind["NamespaceImport"] = 266] = "NamespaceImport";
            SyntaxKind[SyntaxKind["NamedImports"] = 267] = "NamedImports";
            SyntaxKind[SyntaxKind["ImportSpecifier"] = 268] = "ImportSpecifier";
            SyntaxKind[SyntaxKind["ExportAssignment"] = 269] = "ExportAssignment";
            SyntaxKind[SyntaxKind["ExportDeclaration"] = 270] = "ExportDeclaration";
            SyntaxKind[SyntaxKind["NamedExports"] = 271] = "NamedExports";
            SyntaxKind[SyntaxKind["NamespaceExport"] = 272] = "NamespaceExport";
            SyntaxKind[SyntaxKind["ExportSpecifier"] = 273] = "ExportSpecifier";
            SyntaxKind[SyntaxKind["MissingDeclaration"] = 274] = "MissingDeclaration";
            // Module references
            SyntaxKind[SyntaxKind["ExternalModuleReference"] = 275] = "ExternalModuleReference";
            // JSX
            SyntaxKind[SyntaxKind["JsxElement"] = 276] = "JsxElement";
            SyntaxKind[SyntaxKind["JsxSelfClosingElement"] = 277] = "JsxSelfClosingElement";
            SyntaxKind[SyntaxKind["JsxOpeningElement"] = 278] = "JsxOpeningElement";
            SyntaxKind[SyntaxKind["JsxClosingElement"] = 279] = "JsxClosingElement";
            SyntaxKind[SyntaxKind["JsxFragment"] = 280] = "JsxFragment";
            SyntaxKind[SyntaxKind["JsxOpeningFragment"] = 281] = "JsxOpeningFragment";
            SyntaxKind[SyntaxKind["JsxClosingFragment"] = 282] = "JsxClosingFragment";
            SyntaxKind[SyntaxKind["JsxAttribute"] = 283] = "JsxAttribute";
            SyntaxKind[SyntaxKind["JsxAttributes"] = 284] = "JsxAttributes";
            SyntaxKind[SyntaxKind["JsxSpreadAttribute"] = 285] = "JsxSpreadAttribute";
            SyntaxKind[SyntaxKind["JsxExpression"] = 286] = "JsxExpression";
            // Clauses
            SyntaxKind[SyntaxKind["CaseClause"] = 287] = "CaseClause";
            SyntaxKind[SyntaxKind["DefaultClause"] = 288] = "DefaultClause";
            SyntaxKind[SyntaxKind["HeritageClause"] = 289] = "HeritageClause";
            SyntaxKind[SyntaxKind["CatchClause"] = 290] = "CatchClause";
            // Property assignments
            SyntaxKind[SyntaxKind["PropertyAssignment"] = 291] = "PropertyAssignment";
            SyntaxKind[SyntaxKind["ShorthandPropertyAssignment"] = 292] = "ShorthandPropertyAssignment";
            SyntaxKind[SyntaxKind["SpreadAssignment"] = 293] = "SpreadAssignment";
            // Enum
            SyntaxKind[SyntaxKind["EnumMember"] = 294] = "EnumMember";
            // Unparsed
            SyntaxKind[SyntaxKind["UnparsedPrologue"] = 295] = "UnparsedPrologue";
            SyntaxKind[SyntaxKind["UnparsedPrepend"] = 296] = "UnparsedPrepend";
            SyntaxKind[SyntaxKind["UnparsedText"] = 297] = "UnparsedText";
            SyntaxKind[SyntaxKind["UnparsedInternalText"] = 298] = "UnparsedInternalText";
            SyntaxKind[SyntaxKind["UnparsedSyntheticReference"] = 299] = "UnparsedSyntheticReference";
            // Top-level nodes
            SyntaxKind[SyntaxKind["SourceFile"] = 300] = "SourceFile";
            SyntaxKind[SyntaxKind["Bundle"] = 301] = "Bundle";
            SyntaxKind[SyntaxKind["UnparsedSource"] = 302] = "UnparsedSource";
            SyntaxKind[SyntaxKind["InputFiles"] = 303] = "InputFiles";
            // JSDoc nodes
            SyntaxKind[SyntaxKind["JSDocTypeExpression"] = 304] = "JSDocTypeExpression";
            SyntaxKind[SyntaxKind["JSDocNameReference"] = 305] = "JSDocNameReference";
            SyntaxKind[SyntaxKind["JSDocMemberName"] = 306] = "JSDocMemberName";
            SyntaxKind[SyntaxKind["JSDocAllType"] = 307] = "JSDocAllType";
            SyntaxKind[SyntaxKind["JSDocUnknownType"] = 308] = "JSDocUnknownType";
            SyntaxKind[SyntaxKind["JSDocNullableType"] = 309] = "JSDocNullableType";
            SyntaxKind[SyntaxKind["JSDocNonNullableType"] = 310] = "JSDocNonNullableType";
            SyntaxKind[SyntaxKind["JSDocOptionalType"] = 311] = "JSDocOptionalType";
            SyntaxKind[SyntaxKind["JSDocFunctionType"] = 312] = "JSDocFunctionType";
            SyntaxKind[SyntaxKind["JSDocVariadicType"] = 313] = "JSDocVariadicType";
            SyntaxKind[SyntaxKind["JSDocNamepathType"] = 314] = "JSDocNamepathType";
            SyntaxKind[SyntaxKind["JSDocComment"] = 315] = "JSDocComment";
            SyntaxKind[SyntaxKind["JSDocText"] = 316] = "JSDocText";
            SyntaxKind[SyntaxKind["JSDocTypeLiteral"] = 317] = "JSDocTypeLiteral";
            SyntaxKind[SyntaxKind["JSDocSignature"] = 318] = "JSDocSignature";
            SyntaxKind[SyntaxKind["JSDocLink"] = 319] = "JSDocLink";
            SyntaxKind[SyntaxKind["JSDocLinkCode"] = 320] = "JSDocLinkCode";
            SyntaxKind[SyntaxKind["JSDocLinkPlain"] = 321] = "JSDocLinkPlain";
            SyntaxKind[SyntaxKind["JSDocTag"] = 322] = "JSDocTag";
            SyntaxKind[SyntaxKind["JSDocAugmentsTag"] = 323] = "JSDocAugmentsTag";
            SyntaxKind[SyntaxKind["JSDocImplementsTag"] = 324] = "JSDocImplementsTag";
            SyntaxKind[SyntaxKind["JSDocAuthorTag"] = 325] = "JSDocAuthorTag";
            SyntaxKind[SyntaxKind["JSDocDeprecatedTag"] = 326] = "JSDocDeprecatedTag";
            SyntaxKind[SyntaxKind["JSDocClassTag"] = 327] = "JSDocClassTag";
            SyntaxKind[SyntaxKind["JSDocPublicTag"] = 328] = "JSDocPublicTag";
            SyntaxKind[SyntaxKind["JSDocPrivateTag"] = 329] = "JSDocPrivateTag";
            SyntaxKind[SyntaxKind["JSDocProtectedTag"] = 330] = "JSDocProtectedTag";
            SyntaxKind[SyntaxKind["JSDocReadonlyTag"] = 331] = "JSDocReadonlyTag";
            SyntaxKind[SyntaxKind["JSDocOverrideTag"] = 332] = "JSDocOverrideTag";
            SyntaxKind[SyntaxKind["JSDocCallbackTag"] = 333] = "JSDocCallbackTag";
            SyntaxKind[SyntaxKind["JSDocEnumTag"] = 334] = "JSDocEnumTag";
            SyntaxKind[SyntaxKind["JSDocParameterTag"] = 335] = "JSDocParameterTag";
            SyntaxKind[SyntaxKind["JSDocReturnTag"] = 336] = "JSDocReturnTag";
            SyntaxKind[SyntaxKind["JSDocThisTag"] = 337] = "JSDocThisTag";
            SyntaxKind[SyntaxKind["JSDocTypeTag"] = 338] = "JSDocTypeTag";
            SyntaxKind[SyntaxKind["JSDocTemplateTag"] = 339] = "JSDocTemplateTag";
            SyntaxKind[SyntaxKind["JSDocTypedefTag"] = 340] = "JSDocTypedefTag";
            SyntaxKind[SyntaxKind["JSDocSeeTag"] = 341] = "JSDocSeeTag";
            SyntaxKind[SyntaxKind["JSDocPropertyTag"] = 342] = "JSDocPropertyTag";
            // Synthesized list
            SyntaxKind[SyntaxKind["SyntaxList"] = 343] = "SyntaxList";
            // Transformation nodes
            SyntaxKind[SyntaxKind["NotEmittedStatement"] = 344] = "NotEmittedStatement";
            SyntaxKind[SyntaxKind["PartiallyEmittedExpression"] = 345] = "PartiallyEmittedExpression";
            SyntaxKind[SyntaxKind["CommaListExpression"] = 346] = "CommaListExpression";
            SyntaxKind[SyntaxKind["MergeDeclarationMarker"] = 347] = "MergeDeclarationMarker";
            SyntaxKind[SyntaxKind["EndOfDeclarationMarker"] = 348] = "EndOfDeclarationMarker";
            SyntaxKind[SyntaxKind["SyntheticReferenceExpression"] = 349] = "SyntheticReferenceExpression";
            // Enum value count
            SyntaxKind[SyntaxKind["Count"] = 350] = "Count";
            // Markers
            SyntaxKind[SyntaxKind["FirstAssignment"] = 63] = "FirstAssignment";
            SyntaxKind[SyntaxKind["LastAssignment"] = 78] = "LastAssignment";
            SyntaxKind[SyntaxKind["FirstCompoundAssignment"] = 64] = "FirstCompoundAssignment";
            SyntaxKind[SyntaxKind["LastCompoundAssignment"] = 78] = "LastCompoundAssignment";
            SyntaxKind[SyntaxKind["FirstReservedWord"] = 81] = "FirstReservedWord";
            SyntaxKind[SyntaxKind["LastReservedWord"] = 116] = "LastReservedWord";
            SyntaxKind[SyntaxKind["FirstKeyword"] = 81] = "FirstKeyword";
            SyntaxKind[SyntaxKind["LastKeyword"] = 158] = "LastKeyword";
            SyntaxKind[SyntaxKind["FirstFutureReservedWord"] = 117] = "FirstFutureReservedWord";
            SyntaxKind[SyntaxKind["LastFutureReservedWord"] = 125] = "LastFutureReservedWord";
            SyntaxKind[SyntaxKind["FirstTypeNode"] = 175] = "FirstTypeNode";
            SyntaxKind[SyntaxKind["LastTypeNode"] = 198] = "LastTypeNode";
            SyntaxKind[SyntaxKind["FirstPunctuation"] = 18] = "FirstPunctuation";
            SyntaxKind[SyntaxKind["LastPunctuation"] = 78] = "LastPunctuation";
            SyntaxKind[SyntaxKind["FirstToken"] = 0] = "FirstToken";
            SyntaxKind[SyntaxKind["LastToken"] = 158] = "LastToken";
            SyntaxKind[SyntaxKind["FirstTriviaToken"] = 2] = "FirstTriviaToken";
            SyntaxKind[SyntaxKind["LastTriviaToken"] = 7] = "LastTriviaToken";
            SyntaxKind[SyntaxKind["FirstLiteralToken"] = 8] = "FirstLiteralToken";
            SyntaxKind[SyntaxKind["LastLiteralToken"] = 14] = "LastLiteralToken";
            SyntaxKind[SyntaxKind["FirstTemplateToken"] = 14] = "FirstTemplateToken";
            SyntaxKind[SyntaxKind["LastTemplateToken"] = 17] = "LastTemplateToken";
            SyntaxKind[SyntaxKind["FirstBinaryOperator"] = 29] = "FirstBinaryOperator";
            SyntaxKind[SyntaxKind["LastBinaryOperator"] = 78] = "LastBinaryOperator";
            SyntaxKind[SyntaxKind["FirstStatement"] = 235] = "FirstStatement";
            SyntaxKind[SyntaxKind["LastStatement"] = 251] = "LastStatement";
            SyntaxKind[SyntaxKind["FirstNode"] = 159] = "FirstNode";
            SyntaxKind[SyntaxKind["FirstJSDocNode"] = 304] = "FirstJSDocNode";
            SyntaxKind[SyntaxKind["LastJSDocNode"] = 342] = "LastJSDocNode";
            SyntaxKind[SyntaxKind["FirstJSDocTagNode"] = 322] = "FirstJSDocTagNode";
            SyntaxKind[SyntaxKind["LastJSDocTagNode"] = 342] = "LastJSDocTagNode";
            /* @internal */ SyntaxKind[SyntaxKind["FirstContextualKeyword"] = 126] = "FirstContextualKeyword";
            /* @internal */ SyntaxKind[SyntaxKind["LastContextualKeyword"] = 158] = "LastContextualKeyword";
        })(SyntaxKind = ts.SyntaxKind || (ts.SyntaxKind = {}));
        var NodeFlags;
        (function (NodeFlags) {
            NodeFlags[NodeFlags["None"] = 0] = "None";
            NodeFlags[NodeFlags["Let"] = 1] = "Let";
            NodeFlags[NodeFlags["Const"] = 2] = "Const";
            NodeFlags[NodeFlags["NestedNamespace"] = 4] = "NestedNamespace";
            NodeFlags[NodeFlags["Synthesized"] = 8] = "Synthesized";
            NodeFlags[NodeFlags["Namespace"] = 16] = "Namespace";
            NodeFlags[NodeFlags["OptionalChain"] = 32] = "OptionalChain";
            NodeFlags[NodeFlags["ExportContext"] = 64] = "ExportContext";
            NodeFlags[NodeFlags["ContainsThis"] = 128] = "ContainsThis";
            NodeFlags[NodeFlags["HasImplicitReturn"] = 256] = "HasImplicitReturn";
            NodeFlags[NodeFlags["HasExplicitReturn"] = 512] = "HasExplicitReturn";
            NodeFlags[NodeFlags["GlobalAugmentation"] = 1024] = "GlobalAugmentation";
            NodeFlags[NodeFlags["HasAsyncFunctions"] = 2048] = "HasAsyncFunctions";
            NodeFlags[NodeFlags["DisallowInContext"] = 4096] = "DisallowInContext";
            NodeFlags[NodeFlags["YieldContext"] = 8192] = "YieldContext";
            NodeFlags[NodeFlags["DecoratorContext"] = 16384] = "DecoratorContext";
            NodeFlags[NodeFlags["AwaitContext"] = 32768] = "AwaitContext";
            NodeFlags[NodeFlags["ThisNodeHasError"] = 65536] = "ThisNodeHasError";
            NodeFlags[NodeFlags["JavaScriptFile"] = 131072] = "JavaScriptFile";
            NodeFlags[NodeFlags["ThisNodeOrAnySubNodesHasError"] = 262144] = "ThisNodeOrAnySubNodesHasError";
            NodeFlags[NodeFlags["HasAggregatedChildData"] = 524288] = "HasAggregatedChildData";
            // These flags will be set when the parser encounters a dynamic import expression or 'import.meta' to avoid
            // walking the tree if the flags are not set. However, these flags are just a approximation
            // (hence why it's named "PossiblyContainsDynamicImport") because once set, the flags never get cleared.
            // During editing, if a dynamic import is removed, incremental parsing will *NOT* clear this flag.
            // This means that the tree will always be traversed during module resolution, or when looking for external module indicators.
            // However, the removal operation should not occur often and in the case of the
            // removal, it is likely that users will add the import anyway.
            // The advantage of this approach is its simplicity. For the case of batch compilation,
            // we guarantee that users won't have to pay the price of walking the tree if a dynamic import isn't used.
            /* @internal */ NodeFlags[NodeFlags["PossiblyContainsDynamicImport"] = 1048576] = "PossiblyContainsDynamicImport";
            /* @internal */ NodeFlags[NodeFlags["PossiblyContainsImportMeta"] = 2097152] = "PossiblyContainsImportMeta";
            NodeFlags[NodeFlags["JSDoc"] = 4194304] = "JSDoc";
            /* @internal */ NodeFlags[NodeFlags["Ambient"] = 8388608] = "Ambient";
            /* @internal */ NodeFlags[NodeFlags["InWithStatement"] = 16777216] = "InWithStatement";
            NodeFlags[NodeFlags["JsonFile"] = 33554432] = "JsonFile";
            /* @internal */ NodeFlags[NodeFlags["TypeCached"] = 67108864] = "TypeCached";
            /* @internal */ NodeFlags[NodeFlags["Deprecated"] = 134217728] = "Deprecated";
            NodeFlags[NodeFlags["BlockScoped"] = 3] = "BlockScoped";
            NodeFlags[NodeFlags["ReachabilityCheckFlags"] = 768] = "ReachabilityCheckFlags";
            NodeFlags[NodeFlags["ReachabilityAndEmitFlags"] = 2816] = "ReachabilityAndEmitFlags";
            // Parsing context flags
            NodeFlags[NodeFlags["ContextFlags"] = 25358336] = "ContextFlags";
            // Exclude these flags when parsing a Type
            NodeFlags[NodeFlags["TypeExcludesFlags"] = 40960] = "TypeExcludesFlags";
            // Represents all flags that are potentially set once and
            // never cleared on SourceFiles which get re-used in between incremental parses.
            // See the comment above on `PossiblyContainsDynamicImport` and `PossiblyContainsImportMeta`.
            /* @internal */ NodeFlags[NodeFlags["PermanentlySetIncrementalFlags"] = 3145728] = "PermanentlySetIncrementalFlags";
        })(NodeFlags = ts.NodeFlags || (ts.NodeFlags = {}));
        var ModifierFlags;
        (function (ModifierFlags) {
            ModifierFlags[ModifierFlags["None"] = 0] = "None";
            ModifierFlags[ModifierFlags["Export"] = 1] = "Export";
            ModifierFlags[ModifierFlags["Ambient"] = 2] = "Ambient";
            ModifierFlags[ModifierFlags["Public"] = 4] = "Public";
            ModifierFlags[ModifierFlags["Private"] = 8] = "Private";
            ModifierFlags[ModifierFlags["Protected"] = 16] = "Protected";
            ModifierFlags[ModifierFlags["Static"] = 32] = "Static";
            ModifierFlags[ModifierFlags["Readonly"] = 64] = "Readonly";
            ModifierFlags[ModifierFlags["Abstract"] = 128] = "Abstract";
            ModifierFlags[ModifierFlags["Async"] = 256] = "Async";
            ModifierFlags[ModifierFlags["Default"] = 512] = "Default";
            ModifierFlags[ModifierFlags["Const"] = 2048] = "Const";
            ModifierFlags[ModifierFlags["HasComputedJSDocModifiers"] = 4096] = "HasComputedJSDocModifiers";
            ModifierFlags[ModifierFlags["Deprecated"] = 8192] = "Deprecated";
            ModifierFlags[ModifierFlags["Override"] = 16384] = "Override";
            ModifierFlags[ModifierFlags["HasComputedFlags"] = 536870912] = "HasComputedFlags";
            ModifierFlags[ModifierFlags["AccessibilityModifier"] = 28] = "AccessibilityModifier";
            // Accessibility modifiers and 'readonly' can be attached to a parameter in a constructor to make it a property.
            ModifierFlags[ModifierFlags["ParameterPropertyModifier"] = 16476] = "ParameterPropertyModifier";
            ModifierFlags[ModifierFlags["NonPublicAccessibilityModifier"] = 24] = "NonPublicAccessibilityModifier";
            ModifierFlags[ModifierFlags["TypeScriptModifier"] = 18654] = "TypeScriptModifier";
            ModifierFlags[ModifierFlags["ExportDefault"] = 513] = "ExportDefault";
            ModifierFlags[ModifierFlags["All"] = 27647] = "All";
        })(ModifierFlags = ts.ModifierFlags || (ts.ModifierFlags = {}));
        var JsxFlags;
        (function (JsxFlags) {
            JsxFlags[JsxFlags["None"] = 0] = "None";
            /** An element from a named property of the JSX.IntrinsicElements interface */
            JsxFlags[JsxFlags["IntrinsicNamedElement"] = 1] = "IntrinsicNamedElement";
            /** An element inferred from the string index signature of the JSX.IntrinsicElements interface */
            JsxFlags[JsxFlags["IntrinsicIndexedElement"] = 2] = "IntrinsicIndexedElement";
            JsxFlags[JsxFlags["IntrinsicElement"] = 3] = "IntrinsicElement";
        })(JsxFlags = ts.JsxFlags || (ts.JsxFlags = {}));
        /* @internal */
        var RelationComparisonResult;
        (function (RelationComparisonResult) {
            RelationComparisonResult[RelationComparisonResult["Succeeded"] = 1] = "Succeeded";
            RelationComparisonResult[RelationComparisonResult["Failed"] = 2] = "Failed";
            RelationComparisonResult[RelationComparisonResult["Reported"] = 4] = "Reported";
            RelationComparisonResult[RelationComparisonResult["ReportsUnmeasurable"] = 8] = "ReportsUnmeasurable";
            RelationComparisonResult[RelationComparisonResult["ReportsUnreliable"] = 16] = "ReportsUnreliable";
            RelationComparisonResult[RelationComparisonResult["ReportsMask"] = 24] = "ReportsMask";
        })(RelationComparisonResult = ts.RelationComparisonResult || (ts.RelationComparisonResult = {}));
        var GeneratedIdentifierFlags;
        (function (GeneratedIdentifierFlags) {
            // Kinds
            GeneratedIdentifierFlags[GeneratedIdentifierFlags["None"] = 0] = "None";
            /*@internal*/ GeneratedIdentifierFlags[GeneratedIdentifierFlags["Auto"] = 1] = "Auto";
            /*@internal*/ GeneratedIdentifierFlags[GeneratedIdentifierFlags["Loop"] = 2] = "Loop";
            /*@internal*/ GeneratedIdentifierFlags[GeneratedIdentifierFlags["Unique"] = 3] = "Unique";
            /*@internal*/ GeneratedIdentifierFlags[GeneratedIdentifierFlags["Node"] = 4] = "Node";
            /*@internal*/ GeneratedIdentifierFlags[GeneratedIdentifierFlags["KindMask"] = 7] = "KindMask";
            // Flags
            GeneratedIdentifierFlags[GeneratedIdentifierFlags["ReservedInNestedScopes"] = 8] = "ReservedInNestedScopes";
            GeneratedIdentifierFlags[GeneratedIdentifierFlags["Optimistic"] = 16] = "Optimistic";
            GeneratedIdentifierFlags[GeneratedIdentifierFlags["FileLevel"] = 32] = "FileLevel";
            GeneratedIdentifierFlags[GeneratedIdentifierFlags["AllowNameSubstitution"] = 64] = "AllowNameSubstitution";
        })(GeneratedIdentifierFlags = ts.GeneratedIdentifierFlags || (ts.GeneratedIdentifierFlags = {}));
        var TokenFlags;
        (function (TokenFlags) {
            TokenFlags[TokenFlags["None"] = 0] = "None";
            /* @internal */
            TokenFlags[TokenFlags["PrecedingLineBreak"] = 1] = "PrecedingLineBreak";
            /* @internal */
            TokenFlags[TokenFlags["PrecedingJSDocComment"] = 2] = "PrecedingJSDocComment";
            /* @internal */
            TokenFlags[TokenFlags["Unterminated"] = 4] = "Unterminated";
            /* @internal */
            TokenFlags[TokenFlags["ExtendedUnicodeEscape"] = 8] = "ExtendedUnicodeEscape";
            TokenFlags[TokenFlags["Scientific"] = 16] = "Scientific";
            TokenFlags[TokenFlags["Octal"] = 32] = "Octal";
            TokenFlags[TokenFlags["HexSpecifier"] = 64] = "HexSpecifier";
            TokenFlags[TokenFlags["BinarySpecifier"] = 128] = "BinarySpecifier";
            TokenFlags[TokenFlags["OctalSpecifier"] = 256] = "OctalSpecifier";
            /* @internal */
            TokenFlags[TokenFlags["ContainsSeparator"] = 512] = "ContainsSeparator";
            /* @internal */
            TokenFlags[TokenFlags["UnicodeEscape"] = 1024] = "UnicodeEscape";
            /* @internal */
            TokenFlags[TokenFlags["ContainsInvalidEscape"] = 2048] = "ContainsInvalidEscape";
            /* @internal */
            TokenFlags[TokenFlags["BinaryOrOctalSpecifier"] = 384] = "BinaryOrOctalSpecifier";
            /* @internal */
            TokenFlags[TokenFlags["NumericLiteralFlags"] = 1008] = "NumericLiteralFlags";
            /* @internal */
            TokenFlags[TokenFlags["TemplateLiteralLikeFlags"] = 2048] = "TemplateLiteralLikeFlags";
        })(TokenFlags = ts.TokenFlags || (ts.TokenFlags = {}));
        // NOTE: Ensure this is up-to-date with src/debug/debug.ts
        var FlowFlags;
        (function (FlowFlags) {
            FlowFlags[FlowFlags["Unreachable"] = 1] = "Unreachable";
            FlowFlags[FlowFlags["Start"] = 2] = "Start";
            FlowFlags[FlowFlags["BranchLabel"] = 4] = "BranchLabel";
            FlowFlags[FlowFlags["LoopLabel"] = 8] = "LoopLabel";
            FlowFlags[FlowFlags["Assignment"] = 16] = "Assignment";
            FlowFlags[FlowFlags["TrueCondition"] = 32] = "TrueCondition";
            FlowFlags[FlowFlags["FalseCondition"] = 64] = "FalseCondition";
            FlowFlags[FlowFlags["SwitchClause"] = 128] = "SwitchClause";
            FlowFlags[FlowFlags["ArrayMutation"] = 256] = "ArrayMutation";
            FlowFlags[FlowFlags["Call"] = 512] = "Call";
            FlowFlags[FlowFlags["ReduceLabel"] = 1024] = "ReduceLabel";
            FlowFlags[FlowFlags["Referenced"] = 2048] = "Referenced";
            FlowFlags[FlowFlags["Shared"] = 4096] = "Shared";
            FlowFlags[FlowFlags["Label"] = 12] = "Label";
            FlowFlags[FlowFlags["Condition"] = 96] = "Condition";
        })(FlowFlags = ts.FlowFlags || (ts.FlowFlags = {}));
        /* @internal */
        var CommentDirectiveType;
        (function (CommentDirectiveType) {
            CommentDirectiveType[CommentDirectiveType["ExpectError"] = 0] = "ExpectError";
            CommentDirectiveType[CommentDirectiveType["Ignore"] = 1] = "Ignore";
        })(CommentDirectiveType = ts.CommentDirectiveType || (ts.CommentDirectiveType = {}));
        var OperationCanceledException = /** @class */ (function () {
            function OperationCanceledException() {
            }
            return OperationCanceledException;
        }());
        ts.OperationCanceledException = OperationCanceledException;
        /*@internal*/
        var FileIncludeKind;
        (function (FileIncludeKind) {
            FileIncludeKind[FileIncludeKind["RootFile"] = 0] = "RootFile";
            FileIncludeKind[FileIncludeKind["SourceFromProjectReference"] = 1] = "SourceFromProjectReference";
            FileIncludeKind[FileIncludeKind["OutputFromProjectReference"] = 2] = "OutputFromProjectReference";
            FileIncludeKind[FileIncludeKind["Import"] = 3] = "Import";
            FileIncludeKind[FileIncludeKind["ReferenceFile"] = 4] = "ReferenceFile";
            FileIncludeKind[FileIncludeKind["TypeReferenceDirective"] = 5] = "TypeReferenceDirective";
            FileIncludeKind[FileIncludeKind["LibFile"] = 6] = "LibFile";
            FileIncludeKind[FileIncludeKind["LibReferenceDirective"] = 7] = "LibReferenceDirective";
            FileIncludeKind[FileIncludeKind["AutomaticTypeDirectiveFile"] = 8] = "AutomaticTypeDirectiveFile";
        })(FileIncludeKind = ts.FileIncludeKind || (ts.FileIncludeKind = {}));
        /*@internal*/
        var FilePreprocessingDiagnosticsKind;
        (function (FilePreprocessingDiagnosticsKind) {
            FilePreprocessingDiagnosticsKind[FilePreprocessingDiagnosticsKind["FilePreprocessingReferencedDiagnostic"] = 0] = "FilePreprocessingReferencedDiagnostic";
            FilePreprocessingDiagnosticsKind[FilePreprocessingDiagnosticsKind["FilePreprocessingFileExplainingDiagnostic"] = 1] = "FilePreprocessingFileExplainingDiagnostic";
        })(FilePreprocessingDiagnosticsKind = ts.FilePreprocessingDiagnosticsKind || (ts.FilePreprocessingDiagnosticsKind = {}));
        /* @internal */
        var StructureIsReused;
        (function (StructureIsReused) {
            StructureIsReused[StructureIsReused["Not"] = 0] = "Not";
            StructureIsReused[StructureIsReused["SafeModules"] = 1] = "SafeModules";
            StructureIsReused[StructureIsReused["Completely"] = 2] = "Completely";
        })(StructureIsReused = ts.StructureIsReused || (ts.StructureIsReused = {}));
        /** Return code used by getEmitOutput function to indicate status of the function */
        var ExitStatus;
        (function (ExitStatus) {
            // Compiler ran successfully.  Either this was a simple do-nothing compilation (for example,
            // when -version or -help was provided, or this was a normal compilation, no diagnostics
            // were produced, and all outputs were generated successfully.
            ExitStatus[ExitStatus["Success"] = 0] = "Success";
            // Diagnostics were produced and because of them no code was generated.
            ExitStatus[ExitStatus["DiagnosticsPresent_OutputsSkipped"] = 1] = "DiagnosticsPresent_OutputsSkipped";
            // Diagnostics were produced and outputs were generated in spite of them.
            ExitStatus[ExitStatus["DiagnosticsPresent_OutputsGenerated"] = 2] = "DiagnosticsPresent_OutputsGenerated";
            // When build skipped because passed in project is invalid
            ExitStatus[ExitStatus["InvalidProject_OutputsSkipped"] = 3] = "InvalidProject_OutputsSkipped";
            // When build is skipped because project references form cycle
            ExitStatus[ExitStatus["ProjectReferenceCycle_OutputsSkipped"] = 4] = "ProjectReferenceCycle_OutputsSkipped";
            /** @deprecated Use ProjectReferenceCycle_OutputsSkipped instead. */
            ExitStatus[ExitStatus["ProjectReferenceCycle_OutputsSkupped"] = 4] = "ProjectReferenceCycle_OutputsSkupped";
        })(ExitStatus = ts.ExitStatus || (ts.ExitStatus = {}));
        /* @internal */
        var UnionReduction;
        (function (UnionReduction) {
            UnionReduction[UnionReduction["None"] = 0] = "None";
            UnionReduction[UnionReduction["Literal"] = 1] = "Literal";
            UnionReduction[UnionReduction["Subtype"] = 2] = "Subtype";
        })(UnionReduction = ts.UnionReduction || (ts.UnionReduction = {}));
        /* @internal */
        var ContextFlags;
        (function (ContextFlags) {
            ContextFlags[ContextFlags["None"] = 0] = "None";
            ContextFlags[ContextFlags["Signature"] = 1] = "Signature";
            ContextFlags[ContextFlags["NoConstraints"] = 2] = "NoConstraints";
            ContextFlags[ContextFlags["Completions"] = 4] = "Completions";
            ContextFlags[ContextFlags["SkipBindingPatterns"] = 8] = "SkipBindingPatterns";
        })(ContextFlags = ts.ContextFlags || (ts.ContextFlags = {}));
        // NOTE: If modifying this enum, must modify `TypeFormatFlags` too!
        var NodeBuilderFlags;
        (function (NodeBuilderFlags) {
            NodeBuilderFlags[NodeBuilderFlags["None"] = 0] = "None";
            // Options
            NodeBuilderFlags[NodeBuilderFlags["NoTruncation"] = 1] = "NoTruncation";
            NodeBuilderFlags[NodeBuilderFlags["WriteArrayAsGenericType"] = 2] = "WriteArrayAsGenericType";
            NodeBuilderFlags[NodeBuilderFlags["GenerateNamesForShadowedTypeParams"] = 4] = "GenerateNamesForShadowedTypeParams";
            NodeBuilderFlags[NodeBuilderFlags["UseStructuralFallback"] = 8] = "UseStructuralFallback";
            NodeBuilderFlags[NodeBuilderFlags["ForbidIndexedAccessSymbolReferences"] = 16] = "ForbidIndexedAccessSymbolReferences";
            NodeBuilderFlags[NodeBuilderFlags["WriteTypeArgumentsOfSignature"] = 32] = "WriteTypeArgumentsOfSignature";
            NodeBuilderFlags[NodeBuilderFlags["UseFullyQualifiedType"] = 64] = "UseFullyQualifiedType";
            NodeBuilderFlags[NodeBuilderFlags["UseOnlyExternalAliasing"] = 128] = "UseOnlyExternalAliasing";
            NodeBuilderFlags[NodeBuilderFlags["SuppressAnyReturnType"] = 256] = "SuppressAnyReturnType";
            NodeBuilderFlags[NodeBuilderFlags["WriteTypeParametersInQualifiedName"] = 512] = "WriteTypeParametersInQualifiedName";
            NodeBuilderFlags[NodeBuilderFlags["MultilineObjectLiterals"] = 1024] = "MultilineObjectLiterals";
            NodeBuilderFlags[NodeBuilderFlags["WriteClassExpressionAsTypeLiteral"] = 2048] = "WriteClassExpressionAsTypeLiteral";
            NodeBuilderFlags[NodeBuilderFlags["UseTypeOfFunction"] = 4096] = "UseTypeOfFunction";
            NodeBuilderFlags[NodeBuilderFlags["OmitParameterModifiers"] = 8192] = "OmitParameterModifiers";
            NodeBuilderFlags[NodeBuilderFlags["UseAliasDefinedOutsideCurrentScope"] = 16384] = "UseAliasDefinedOutsideCurrentScope";
            NodeBuilderFlags[NodeBuilderFlags["UseSingleQuotesForStringLiteralType"] = 268435456] = "UseSingleQuotesForStringLiteralType";
            NodeBuilderFlags[NodeBuilderFlags["NoTypeReduction"] = 536870912] = "NoTypeReduction";
            NodeBuilderFlags[NodeBuilderFlags["NoUndefinedOptionalParameterType"] = 1073741824] = "NoUndefinedOptionalParameterType";
            // Error handling
            NodeBuilderFlags[NodeBuilderFlags["AllowThisInObjectLiteral"] = 32768] = "AllowThisInObjectLiteral";
            NodeBuilderFlags[NodeBuilderFlags["AllowQualifiedNameInPlaceOfIdentifier"] = 65536] = "AllowQualifiedNameInPlaceOfIdentifier";
            /** @deprecated AllowQualifedNameInPlaceOfIdentifier. Use AllowQualifiedNameInPlaceOfIdentifier instead. */
            NodeBuilderFlags[NodeBuilderFlags["AllowQualifedNameInPlaceOfIdentifier"] = 65536] = "AllowQualifedNameInPlaceOfIdentifier";
            NodeBuilderFlags[NodeBuilderFlags["AllowAnonymousIdentifier"] = 131072] = "AllowAnonymousIdentifier";
            NodeBuilderFlags[NodeBuilderFlags["AllowEmptyUnionOrIntersection"] = 262144] = "AllowEmptyUnionOrIntersection";
            NodeBuilderFlags[NodeBuilderFlags["AllowEmptyTuple"] = 524288] = "AllowEmptyTuple";
            NodeBuilderFlags[NodeBuilderFlags["AllowUniqueESSymbolType"] = 1048576] = "AllowUniqueESSymbolType";
            NodeBuilderFlags[NodeBuilderFlags["AllowEmptyIndexInfoType"] = 2097152] = "AllowEmptyIndexInfoType";
            // Errors (cont.)
            NodeBuilderFlags[NodeBuilderFlags["AllowNodeModulesRelativePaths"] = 67108864] = "AllowNodeModulesRelativePaths";
            /* @internal */ NodeBuilderFlags[NodeBuilderFlags["DoNotIncludeSymbolChain"] = 134217728] = "DoNotIncludeSymbolChain";
            NodeBuilderFlags[NodeBuilderFlags["IgnoreErrors"] = 70221824] = "IgnoreErrors";
            // State
            NodeBuilderFlags[NodeBuilderFlags["InObjectTypeLiteral"] = 4194304] = "InObjectTypeLiteral";
            NodeBuilderFlags[NodeBuilderFlags["InTypeAlias"] = 8388608] = "InTypeAlias";
            NodeBuilderFlags[NodeBuilderFlags["InInitialEntityName"] = 16777216] = "InInitialEntityName";
        })(NodeBuilderFlags = ts.NodeBuilderFlags || (ts.NodeBuilderFlags = {}));
        // Ensure the shared flags between this and `NodeBuilderFlags` stay in alignment
        var TypeFormatFlags;
        (function (TypeFormatFlags) {
            TypeFormatFlags[TypeFormatFlags["None"] = 0] = "None";
            TypeFormatFlags[TypeFormatFlags["NoTruncation"] = 1] = "NoTruncation";
            TypeFormatFlags[TypeFormatFlags["WriteArrayAsGenericType"] = 2] = "WriteArrayAsGenericType";
            // hole because there's a hole in node builder flags
            TypeFormatFlags[TypeFormatFlags["UseStructuralFallback"] = 8] = "UseStructuralFallback";
            // hole because there's a hole in node builder flags
            TypeFormatFlags[TypeFormatFlags["WriteTypeArgumentsOfSignature"] = 32] = "WriteTypeArgumentsOfSignature";
            TypeFormatFlags[TypeFormatFlags["UseFullyQualifiedType"] = 64] = "UseFullyQualifiedType";
            // hole because `UseOnlyExternalAliasing` is here in node builder flags, but functions which take old flags use `SymbolFormatFlags` instead
            TypeFormatFlags[TypeFormatFlags["SuppressAnyReturnType"] = 256] = "SuppressAnyReturnType";
            // hole because `WriteTypeParametersInQualifiedName` is here in node builder flags, but functions which take old flags use `SymbolFormatFlags` for this instead
            TypeFormatFlags[TypeFormatFlags["MultilineObjectLiterals"] = 1024] = "MultilineObjectLiterals";
            TypeFormatFlags[TypeFormatFlags["WriteClassExpressionAsTypeLiteral"] = 2048] = "WriteClassExpressionAsTypeLiteral";
            TypeFormatFlags[TypeFormatFlags["UseTypeOfFunction"] = 4096] = "UseTypeOfFunction";
            TypeFormatFlags[TypeFormatFlags["OmitParameterModifiers"] = 8192] = "OmitParameterModifiers";
            TypeFormatFlags[TypeFormatFlags["UseAliasDefinedOutsideCurrentScope"] = 16384] = "UseAliasDefinedOutsideCurrentScope";
            TypeFormatFlags[TypeFormatFlags["UseSingleQuotesForStringLiteralType"] = 268435456] = "UseSingleQuotesForStringLiteralType";
            TypeFormatFlags[TypeFormatFlags["NoTypeReduction"] = 536870912] = "NoTypeReduction";
            // Error Handling
            TypeFormatFlags[TypeFormatFlags["AllowUniqueESSymbolType"] = 1048576] = "AllowUniqueESSymbolType";
            // TypeFormatFlags exclusive
            TypeFormatFlags[TypeFormatFlags["AddUndefined"] = 131072] = "AddUndefined";
            TypeFormatFlags[TypeFormatFlags["WriteArrowStyleSignature"] = 262144] = "WriteArrowStyleSignature";
            // State
            TypeFormatFlags[TypeFormatFlags["InArrayType"] = 524288] = "InArrayType";
            TypeFormatFlags[TypeFormatFlags["InElementType"] = 2097152] = "InElementType";
            TypeFormatFlags[TypeFormatFlags["InFirstTypeArgument"] = 4194304] = "InFirstTypeArgument";
            TypeFormatFlags[TypeFormatFlags["InTypeAlias"] = 8388608] = "InTypeAlias";
            /** @deprecated */ TypeFormatFlags[TypeFormatFlags["WriteOwnNameForAnyLike"] = 0] = "WriteOwnNameForAnyLike";
            TypeFormatFlags[TypeFormatFlags["NodeBuilderFlagsMask"] = 814775659] = "NodeBuilderFlagsMask";
        })(TypeFormatFlags = ts.TypeFormatFlags || (ts.TypeFormatFlags = {}));
        var SymbolFormatFlags;
        (function (SymbolFormatFlags) {
            SymbolFormatFlags[SymbolFormatFlags["None"] = 0] = "None";
            // Write symbols's type argument if it is instantiated symbol
            // eg. class C<T> { p: T }   <-- Show p as C<T>.p here
            //     var a: C<number>;
            //     var p = a.p; <--- Here p is property of C<number> so show it as C<number>.p instead of just C.p
            SymbolFormatFlags[SymbolFormatFlags["WriteTypeParametersOrArguments"] = 1] = "WriteTypeParametersOrArguments";
            // Use only external alias information to get the symbol name in the given context
            // eg.  module m { export class c { } } import x = m.c;
            // When this flag is specified m.c will be used to refer to the class instead of alias symbol x
            SymbolFormatFlags[SymbolFormatFlags["UseOnlyExternalAliasing"] = 2] = "UseOnlyExternalAliasing";
            // Build symbol name using any nodes needed, instead of just components of an entity name
            SymbolFormatFlags[SymbolFormatFlags["AllowAnyNodeKind"] = 4] = "AllowAnyNodeKind";
            // Prefer aliases which are not directly visible
            SymbolFormatFlags[SymbolFormatFlags["UseAliasDefinedOutsideCurrentScope"] = 8] = "UseAliasDefinedOutsideCurrentScope";
            // Skip building an accessible symbol chain
            /* @internal */ SymbolFormatFlags[SymbolFormatFlags["DoNotIncludeSymbolChain"] = 16] = "DoNotIncludeSymbolChain";
        })(SymbolFormatFlags = ts.SymbolFormatFlags || (ts.SymbolFormatFlags = {}));
        /* @internal */
        var SymbolAccessibility;
        (function (SymbolAccessibility) {
            SymbolAccessibility[SymbolAccessibility["Accessible"] = 0] = "Accessible";
            SymbolAccessibility[SymbolAccessibility["NotAccessible"] = 1] = "NotAccessible";
            SymbolAccessibility[SymbolAccessibility["CannotBeNamed"] = 2] = "CannotBeNamed";
        })(SymbolAccessibility = ts.SymbolAccessibility || (ts.SymbolAccessibility = {}));
        /* @internal */
        var SyntheticSymbolKind;
        (function (SyntheticSymbolKind) {
            SyntheticSymbolKind[SyntheticSymbolKind["UnionOrIntersection"] = 0] = "UnionOrIntersection";
            SyntheticSymbolKind[SyntheticSymbolKind["Spread"] = 1] = "Spread";
        })(SyntheticSymbolKind = ts.SyntheticSymbolKind || (ts.SyntheticSymbolKind = {}));
        var TypePredicateKind;
        (function (TypePredicateKind) {
            TypePredicateKind[TypePredicateKind["This"] = 0] = "This";
            TypePredicateKind[TypePredicateKind["Identifier"] = 1] = "Identifier";
            TypePredicateKind[TypePredicateKind["AssertsThis"] = 2] = "AssertsThis";
            TypePredicateKind[TypePredicateKind["AssertsIdentifier"] = 3] = "AssertsIdentifier";
        })(TypePredicateKind = ts.TypePredicateKind || (ts.TypePredicateKind = {}));
        /** Indicates how to serialize the name for a TypeReferenceNode when emitting decorator metadata */
        /* @internal */
        var TypeReferenceSerializationKind;
        (function (TypeReferenceSerializationKind) {
            // The TypeReferenceNode could not be resolved.
            // The type name should be emitted using a safe fallback.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["Unknown"] = 0] = "Unknown";
            // The TypeReferenceNode resolves to a type with a constructor
            // function that can be reached at runtime (e.g. a `class`
            // declaration or a `var` declaration for the static side
            // of a type, such as the global `Promise` type in lib.d.ts).
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["TypeWithConstructSignatureAndValue"] = 1] = "TypeWithConstructSignatureAndValue";
            // The TypeReferenceNode resolves to a Void-like, Nullable, or Never type.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["VoidNullableOrNeverType"] = 2] = "VoidNullableOrNeverType";
            // The TypeReferenceNode resolves to a Number-like type.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["NumberLikeType"] = 3] = "NumberLikeType";
            // The TypeReferenceNode resolves to a BigInt-like type.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["BigIntLikeType"] = 4] = "BigIntLikeType";
            // The TypeReferenceNode resolves to a String-like type.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["StringLikeType"] = 5] = "StringLikeType";
            // The TypeReferenceNode resolves to a Boolean-like type.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["BooleanType"] = 6] = "BooleanType";
            // The TypeReferenceNode resolves to an Array-like type.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["ArrayLikeType"] = 7] = "ArrayLikeType";
            // The TypeReferenceNode resolves to the ESSymbol type.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["ESSymbolType"] = 8] = "ESSymbolType";
            // The TypeReferenceNode resolved to the global Promise constructor symbol.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["Promise"] = 9] = "Promise";
            // The TypeReferenceNode resolves to a Function type or a type with call signatures.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["TypeWithCallSignature"] = 10] = "TypeWithCallSignature";
            // The TypeReferenceNode resolves to any other type.
            TypeReferenceSerializationKind[TypeReferenceSerializationKind["ObjectType"] = 11] = "ObjectType";
        })(TypeReferenceSerializationKind = ts.TypeReferenceSerializationKind || (ts.TypeReferenceSerializationKind = {}));
        var SymbolFlags;
        (function (SymbolFlags) {
            SymbolFlags[SymbolFlags["None"] = 0] = "None";
            SymbolFlags[SymbolFlags["FunctionScopedVariable"] = 1] = "FunctionScopedVariable";
            SymbolFlags[SymbolFlags["BlockScopedVariable"] = 2] = "BlockScopedVariable";
            SymbolFlags[SymbolFlags["Property"] = 4] = "Property";
            SymbolFlags[SymbolFlags["EnumMember"] = 8] = "EnumMember";
            SymbolFlags[SymbolFlags["Function"] = 16] = "Function";
            SymbolFlags[SymbolFlags["Class"] = 32] = "Class";
            SymbolFlags[SymbolFlags["Interface"] = 64] = "Interface";
            SymbolFlags[SymbolFlags["ConstEnum"] = 128] = "ConstEnum";
            SymbolFlags[SymbolFlags["RegularEnum"] = 256] = "RegularEnum";
            SymbolFlags[SymbolFlags["ValueModule"] = 512] = "ValueModule";
            SymbolFlags[SymbolFlags["NamespaceModule"] = 1024] = "NamespaceModule";
            SymbolFlags[SymbolFlags["TypeLiteral"] = 2048] = "TypeLiteral";
            SymbolFlags[SymbolFlags["ObjectLiteral"] = 4096] = "ObjectLiteral";
            SymbolFlags[SymbolFlags["Method"] = 8192] = "Method";
            SymbolFlags[SymbolFlags["Constructor"] = 16384] = "Constructor";
            SymbolFlags[SymbolFlags["GetAccessor"] = 32768] = "GetAccessor";
            SymbolFlags[SymbolFlags["SetAccessor"] = 65536] = "SetAccessor";
            SymbolFlags[SymbolFlags["Signature"] = 131072] = "Signature";
            SymbolFlags[SymbolFlags["TypeParameter"] = 262144] = "TypeParameter";
            SymbolFlags[SymbolFlags["TypeAlias"] = 524288] = "TypeAlias";
            SymbolFlags[SymbolFlags["ExportValue"] = 1048576] = "ExportValue";
            SymbolFlags[SymbolFlags["Alias"] = 2097152] = "Alias";
            SymbolFlags[SymbolFlags["Prototype"] = 4194304] = "Prototype";
            SymbolFlags[SymbolFlags["ExportStar"] = 8388608] = "ExportStar";
            SymbolFlags[SymbolFlags["Optional"] = 16777216] = "Optional";
            SymbolFlags[SymbolFlags["Transient"] = 33554432] = "Transient";
            SymbolFlags[SymbolFlags["Assignment"] = 67108864] = "Assignment";
            SymbolFlags[SymbolFlags["ModuleExports"] = 134217728] = "ModuleExports";
            /* @internal */
            SymbolFlags[SymbolFlags["All"] = 67108863] = "All";
            SymbolFlags[SymbolFlags["Enum"] = 384] = "Enum";
            SymbolFlags[SymbolFlags["Variable"] = 3] = "Variable";
            SymbolFlags[SymbolFlags["Value"] = 111551] = "Value";
            SymbolFlags[SymbolFlags["Type"] = 788968] = "Type";
            SymbolFlags[SymbolFlags["Namespace"] = 1920] = "Namespace";
            SymbolFlags[SymbolFlags["Module"] = 1536] = "Module";
            SymbolFlags[SymbolFlags["Accessor"] = 98304] = "Accessor";
            // Variables can be redeclared, but can not redeclare a block-scoped declaration with the
            // same name, or any other value that is not a variable, e.g. ValueModule or Class
            SymbolFlags[SymbolFlags["FunctionScopedVariableExcludes"] = 111550] = "FunctionScopedVariableExcludes";
            // Block-scoped declarations are not allowed to be re-declared
            // they can not merge with anything in the value space
            SymbolFlags[SymbolFlags["BlockScopedVariableExcludes"] = 111551] = "BlockScopedVariableExcludes";
            SymbolFlags[SymbolFlags["ParameterExcludes"] = 111551] = "ParameterExcludes";
            SymbolFlags[SymbolFlags["PropertyExcludes"] = 0] = "PropertyExcludes";
            SymbolFlags[SymbolFlags["EnumMemberExcludes"] = 900095] = "EnumMemberExcludes";
            SymbolFlags[SymbolFlags["FunctionExcludes"] = 110991] = "FunctionExcludes";
            SymbolFlags[SymbolFlags["ClassExcludes"] = 899503] = "ClassExcludes";
            SymbolFlags[SymbolFlags["InterfaceExcludes"] = 788872] = "InterfaceExcludes";
            SymbolFlags[SymbolFlags["RegularEnumExcludes"] = 899327] = "RegularEnumExcludes";
            SymbolFlags[SymbolFlags["ConstEnumExcludes"] = 899967] = "ConstEnumExcludes";
            SymbolFlags[SymbolFlags["ValueModuleExcludes"] = 110735] = "ValueModuleExcludes";
            SymbolFlags[SymbolFlags["NamespaceModuleExcludes"] = 0] = "NamespaceModuleExcludes";
            SymbolFlags[SymbolFlags["MethodExcludes"] = 103359] = "MethodExcludes";
            SymbolFlags[SymbolFlags["GetAccessorExcludes"] = 46015] = "GetAccessorExcludes";
            SymbolFlags[SymbolFlags["SetAccessorExcludes"] = 78783] = "SetAccessorExcludes";
            SymbolFlags[SymbolFlags["TypeParameterExcludes"] = 526824] = "TypeParameterExcludes";
            SymbolFlags[SymbolFlags["TypeAliasExcludes"] = 788968] = "TypeAliasExcludes";
            SymbolFlags[SymbolFlags["AliasExcludes"] = 2097152] = "AliasExcludes";
            SymbolFlags[SymbolFlags["ModuleMember"] = 2623475] = "ModuleMember";
            SymbolFlags[SymbolFlags["ExportHasLocal"] = 944] = "ExportHasLocal";
            SymbolFlags[SymbolFlags["BlockScoped"] = 418] = "BlockScoped";
            SymbolFlags[SymbolFlags["PropertyOrAccessor"] = 98308] = "PropertyOrAccessor";
            SymbolFlags[SymbolFlags["ClassMember"] = 106500] = "ClassMember";
            /* @internal */
            SymbolFlags[SymbolFlags["ExportSupportsDefaultModifier"] = 112] = "ExportSupportsDefaultModifier";
            /* @internal */
            SymbolFlags[SymbolFlags["ExportDoesNotSupportDefaultModifier"] = -113] = "ExportDoesNotSupportDefaultModifier";
            /* @internal */
            // The set of things we consider semantically classifiable.  Used to speed up the LS during
            // classification.
            SymbolFlags[SymbolFlags["Classifiable"] = 2885600] = "Classifiable";
            /* @internal */
            SymbolFlags[SymbolFlags["LateBindingContainer"] = 6256] = "LateBindingContainer";
        })(SymbolFlags = ts.SymbolFlags || (ts.SymbolFlags = {}));
        /* @internal */
        var EnumKind;
        (function (EnumKind) {
            EnumKind[EnumKind["Numeric"] = 0] = "Numeric";
            EnumKind[EnumKind["Literal"] = 1] = "Literal"; // Literal enum (each member has a TypeFlags.EnumLiteral type)
        })(EnumKind = ts.EnumKind || (ts.EnumKind = {}));
        /* @internal */
        var CheckFlags;
        (function (CheckFlags) {
            CheckFlags[CheckFlags["Instantiated"] = 1] = "Instantiated";
            CheckFlags[CheckFlags["SyntheticProperty"] = 2] = "SyntheticProperty";
            CheckFlags[CheckFlags["SyntheticMethod"] = 4] = "SyntheticMethod";
            CheckFlags[CheckFlags["Readonly"] = 8] = "Readonly";
            CheckFlags[CheckFlags["ReadPartial"] = 16] = "ReadPartial";
            CheckFlags[CheckFlags["WritePartial"] = 32] = "WritePartial";
            CheckFlags[CheckFlags["HasNonUniformType"] = 64] = "HasNonUniformType";
            CheckFlags[CheckFlags["HasLiteralType"] = 128] = "HasLiteralType";
            CheckFlags[CheckFlags["ContainsPublic"] = 256] = "ContainsPublic";
            CheckFlags[CheckFlags["ContainsProtected"] = 512] = "ContainsProtected";
            CheckFlags[CheckFlags["ContainsPrivate"] = 1024] = "ContainsPrivate";
            CheckFlags[CheckFlags["ContainsStatic"] = 2048] = "ContainsStatic";
            CheckFlags[CheckFlags["Late"] = 4096] = "Late";
            CheckFlags[CheckFlags["ReverseMapped"] = 8192] = "ReverseMapped";
            CheckFlags[CheckFlags["OptionalParameter"] = 16384] = "OptionalParameter";
            CheckFlags[CheckFlags["RestParameter"] = 32768] = "RestParameter";
            CheckFlags[CheckFlags["DeferredType"] = 65536] = "DeferredType";
            CheckFlags[CheckFlags["HasNeverType"] = 131072] = "HasNeverType";
            CheckFlags[CheckFlags["Mapped"] = 262144] = "Mapped";
            CheckFlags[CheckFlags["StripOptional"] = 524288] = "StripOptional";
            CheckFlags[CheckFlags["Synthetic"] = 6] = "Synthetic";
            CheckFlags[CheckFlags["Discriminant"] = 192] = "Discriminant";
            CheckFlags[CheckFlags["Partial"] = 48] = "Partial";
        })(CheckFlags = ts.CheckFlags || (ts.CheckFlags = {}));
        var InternalSymbolName;
        (function (InternalSymbolName) {
            InternalSymbolName["Call"] = "__call";
            InternalSymbolName["Constructor"] = "__constructor";
            InternalSymbolName["New"] = "__new";
            InternalSymbolName["Index"] = "__index";
            InternalSymbolName["ExportStar"] = "__export";
            InternalSymbolName["Global"] = "__global";
            InternalSymbolName["Missing"] = "__missing";
            InternalSymbolName["Type"] = "__type";
            InternalSymbolName["Object"] = "__object";
            InternalSymbolName["JSXAttributes"] = "__jsxAttributes";
            InternalSymbolName["Class"] = "__class";
            InternalSymbolName["Function"] = "__function";
            InternalSymbolName["Computed"] = "__computed";
            InternalSymbolName["Resolving"] = "__resolving__";
            InternalSymbolName["ExportEquals"] = "export=";
            InternalSymbolName["Default"] = "default";
            InternalSymbolName["This"] = "this";
        })(InternalSymbolName = ts.InternalSymbolName || (ts.InternalSymbolName = {}));
        /* @internal */
        var NodeCheckFlags;
        (function (NodeCheckFlags) {
            NodeCheckFlags[NodeCheckFlags["TypeChecked"] = 1] = "TypeChecked";
            NodeCheckFlags[NodeCheckFlags["LexicalThis"] = 2] = "LexicalThis";
            NodeCheckFlags[NodeCheckFlags["CaptureThis"] = 4] = "CaptureThis";
            NodeCheckFlags[NodeCheckFlags["CaptureNewTarget"] = 8] = "CaptureNewTarget";
            NodeCheckFlags[NodeCheckFlags["SuperInstance"] = 256] = "SuperInstance";
            NodeCheckFlags[NodeCheckFlags["SuperStatic"] = 512] = "SuperStatic";
            NodeCheckFlags[NodeCheckFlags["ContextChecked"] = 1024] = "ContextChecked";
            NodeCheckFlags[NodeCheckFlags["AsyncMethodWithSuper"] = 2048] = "AsyncMethodWithSuper";
            NodeCheckFlags[NodeCheckFlags["AsyncMethodWithSuperBinding"] = 4096] = "AsyncMethodWithSuperBinding";
            NodeCheckFlags[NodeCheckFlags["CaptureArguments"] = 8192] = "CaptureArguments";
            NodeCheckFlags[NodeCheckFlags["EnumValuesComputed"] = 16384] = "EnumValuesComputed";
            NodeCheckFlags[NodeCheckFlags["LexicalModuleMergesWithClass"] = 32768] = "LexicalModuleMergesWithClass";
            NodeCheckFlags[NodeCheckFlags["LoopWithCapturedBlockScopedBinding"] = 65536] = "LoopWithCapturedBlockScopedBinding";
            NodeCheckFlags[NodeCheckFlags["ContainsCapturedBlockScopeBinding"] = 131072] = "ContainsCapturedBlockScopeBinding";
            NodeCheckFlags[NodeCheckFlags["CapturedBlockScopedBinding"] = 262144] = "CapturedBlockScopedBinding";
            NodeCheckFlags[NodeCheckFlags["BlockScopedBindingInLoop"] = 524288] = "BlockScopedBindingInLoop";
            NodeCheckFlags[NodeCheckFlags["ClassWithBodyScopedClassBinding"] = 1048576] = "ClassWithBodyScopedClassBinding";
            NodeCheckFlags[NodeCheckFlags["BodyScopedClassBinding"] = 2097152] = "BodyScopedClassBinding";
            NodeCheckFlags[NodeCheckFlags["NeedsLoopOutParameter"] = 4194304] = "NeedsLoopOutParameter";
            NodeCheckFlags[NodeCheckFlags["AssignmentsMarked"] = 8388608] = "AssignmentsMarked";
            NodeCheckFlags[NodeCheckFlags["ClassWithConstructorReference"] = 16777216] = "ClassWithConstructorReference";
            NodeCheckFlags[NodeCheckFlags["ConstructorReferenceInClass"] = 33554432] = "ConstructorReferenceInClass";
            NodeCheckFlags[NodeCheckFlags["ContainsClassWithPrivateIdentifiers"] = 67108864] = "ContainsClassWithPrivateIdentifiers";
            NodeCheckFlags[NodeCheckFlags["ContainsSuperPropertyInStaticInitializer"] = 134217728] = "ContainsSuperPropertyInStaticInitializer";
        })(NodeCheckFlags = ts.NodeCheckFlags || (ts.NodeCheckFlags = {}));
        var TypeFlags;
        (function (TypeFlags) {
            TypeFlags[TypeFlags["Any"] = 1] = "Any";
            TypeFlags[TypeFlags["Unknown"] = 2] = "Unknown";
            TypeFlags[TypeFlags["String"] = 4] = "String";
            TypeFlags[TypeFlags["Number"] = 8] = "Number";
            TypeFlags[TypeFlags["Boolean"] = 16] = "Boolean";
            TypeFlags[TypeFlags["Enum"] = 32] = "Enum";
            TypeFlags[TypeFlags["BigInt"] = 64] = "BigInt";
            TypeFlags[TypeFlags["StringLiteral"] = 128] = "StringLiteral";
            TypeFlags[TypeFlags["NumberLiteral"] = 256] = "NumberLiteral";
            TypeFlags[TypeFlags["BooleanLiteral"] = 512] = "BooleanLiteral";
            TypeFlags[TypeFlags["EnumLiteral"] = 1024] = "EnumLiteral";
            TypeFlags[TypeFlags["BigIntLiteral"] = 2048] = "BigIntLiteral";
            TypeFlags[TypeFlags["ESSymbol"] = 4096] = "ESSymbol";
            TypeFlags[TypeFlags["UniqueESSymbol"] = 8192] = "UniqueESSymbol";
            TypeFlags[TypeFlags["Void"] = 16384] = "Void";
            TypeFlags[TypeFlags["Undefined"] = 32768] = "Undefined";
            TypeFlags[TypeFlags["Null"] = 65536] = "Null";
            TypeFlags[TypeFlags["Never"] = 131072] = "Never";
            TypeFlags[TypeFlags["TypeParameter"] = 262144] = "TypeParameter";
            TypeFlags[TypeFlags["Object"] = 524288] = "Object";
            TypeFlags[TypeFlags["Union"] = 1048576] = "Union";
            TypeFlags[TypeFlags["Intersection"] = 2097152] = "Intersection";
            TypeFlags[TypeFlags["Index"] = 4194304] = "Index";
            TypeFlags[TypeFlags["IndexedAccess"] = 8388608] = "IndexedAccess";
            TypeFlags[TypeFlags["Conditional"] = 16777216] = "Conditional";
            TypeFlags[TypeFlags["Substitution"] = 33554432] = "Substitution";
            TypeFlags[TypeFlags["NonPrimitive"] = 67108864] = "NonPrimitive";
            TypeFlags[TypeFlags["TemplateLiteral"] = 134217728] = "TemplateLiteral";
            TypeFlags[TypeFlags["StringMapping"] = 268435456] = "StringMapping";
            /* @internal */
            TypeFlags[TypeFlags["AnyOrUnknown"] = 3] = "AnyOrUnknown";
            /* @internal */
            TypeFlags[TypeFlags["Nullable"] = 98304] = "Nullable";
            TypeFlags[TypeFlags["Literal"] = 2944] = "Literal";
            TypeFlags[TypeFlags["Unit"] = 109440] = "Unit";
            TypeFlags[TypeFlags["StringOrNumberLiteral"] = 384] = "StringOrNumberLiteral";
            /* @internal */
            TypeFlags[TypeFlags["StringOrNumberLiteralOrUnique"] = 8576] = "StringOrNumberLiteralOrUnique";
            /* @internal */
            TypeFlags[TypeFlags["DefinitelyFalsy"] = 117632] = "DefinitelyFalsy";
            TypeFlags[TypeFlags["PossiblyFalsy"] = 117724] = "PossiblyFalsy";
            /* @internal */
            TypeFlags[TypeFlags["Intrinsic"] = 67359327] = "Intrinsic";
            /* @internal */
            TypeFlags[TypeFlags["Primitive"] = 131068] = "Primitive";
            TypeFlags[TypeFlags["StringLike"] = 402653316] = "StringLike";
            TypeFlags[TypeFlags["NumberLike"] = 296] = "NumberLike";
            TypeFlags[TypeFlags["BigIntLike"] = 2112] = "BigIntLike";
            TypeFlags[TypeFlags["BooleanLike"] = 528] = "BooleanLike";
            TypeFlags[TypeFlags["EnumLike"] = 1056] = "EnumLike";
            TypeFlags[TypeFlags["ESSymbolLike"] = 12288] = "ESSymbolLike";
            TypeFlags[TypeFlags["VoidLike"] = 49152] = "VoidLike";
            /* @internal */
            TypeFlags[TypeFlags["DisjointDomains"] = 469892092] = "DisjointDomains";
            TypeFlags[TypeFlags["UnionOrIntersection"] = 3145728] = "UnionOrIntersection";
            TypeFlags[TypeFlags["StructuredType"] = 3670016] = "StructuredType";
            TypeFlags[TypeFlags["TypeVariable"] = 8650752] = "TypeVariable";
            TypeFlags[TypeFlags["InstantiableNonPrimitive"] = 58982400] = "InstantiableNonPrimitive";
            TypeFlags[TypeFlags["InstantiablePrimitive"] = 406847488] = "InstantiablePrimitive";
            TypeFlags[TypeFlags["Instantiable"] = 465829888] = "Instantiable";
            TypeFlags[TypeFlags["StructuredOrInstantiable"] = 469499904] = "StructuredOrInstantiable";
            /* @internal */
            TypeFlags[TypeFlags["ObjectFlagsType"] = 3899393] = "ObjectFlagsType";
            /* @internal */
            TypeFlags[TypeFlags["Simplifiable"] = 25165824] = "Simplifiable";
            /* @internal */
            TypeFlags[TypeFlags["Singleton"] = 67358815] = "Singleton";
            // 'Narrowable' types are types where narrowing actually narrows.
            // This *should* be every type other than null, undefined, void, and never
            TypeFlags[TypeFlags["Narrowable"] = 536624127] = "Narrowable";
            /* @internal */
            TypeFlags[TypeFlags["NotPrimitiveUnion"] = 468598819] = "NotPrimitiveUnion";
            // The following flags are aggregated during union and intersection type construction
            /* @internal */
            TypeFlags[TypeFlags["IncludesMask"] = 205258751] = "IncludesMask";
            // The following flags are used for different purposes during union and intersection type construction
            /* @internal */
            TypeFlags[TypeFlags["IncludesStructuredOrInstantiable"] = 262144] = "IncludesStructuredOrInstantiable";
            /* @internal */
            TypeFlags[TypeFlags["IncludesNonWideningType"] = 4194304] = "IncludesNonWideningType";
            /* @internal */
            TypeFlags[TypeFlags["IncludesWildcard"] = 8388608] = "IncludesWildcard";
            /* @internal */
            TypeFlags[TypeFlags["IncludesEmptyObject"] = 16777216] = "IncludesEmptyObject";
        })(TypeFlags = ts.TypeFlags || (ts.TypeFlags = {}));
        // Types included in TypeFlags.ObjectFlagsType have an objectFlags property. Some ObjectFlags
        // are specific to certain types and reuse the same bit position. Those ObjectFlags require a check
        // for a certain TypeFlags value to determine their meaning.
        var ObjectFlags;
        (function (ObjectFlags) {
            ObjectFlags[ObjectFlags["Class"] = 1] = "Class";
            ObjectFlags[ObjectFlags["Interface"] = 2] = "Interface";
            ObjectFlags[ObjectFlags["Reference"] = 4] = "Reference";
            ObjectFlags[ObjectFlags["Tuple"] = 8] = "Tuple";
            ObjectFlags[ObjectFlags["Anonymous"] = 16] = "Anonymous";
            ObjectFlags[ObjectFlags["Mapped"] = 32] = "Mapped";
            ObjectFlags[ObjectFlags["Instantiated"] = 64] = "Instantiated";
            ObjectFlags[ObjectFlags["ObjectLiteral"] = 128] = "ObjectLiteral";
            ObjectFlags[ObjectFlags["EvolvingArray"] = 256] = "EvolvingArray";
            ObjectFlags[ObjectFlags["ObjectLiteralPatternWithComputedProperties"] = 512] = "ObjectLiteralPatternWithComputedProperties";
            ObjectFlags[ObjectFlags["ReverseMapped"] = 1024] = "ReverseMapped";
            ObjectFlags[ObjectFlags["JsxAttributes"] = 2048] = "JsxAttributes";
            ObjectFlags[ObjectFlags["MarkerType"] = 4096] = "MarkerType";
            ObjectFlags[ObjectFlags["JSLiteral"] = 8192] = "JSLiteral";
            ObjectFlags[ObjectFlags["FreshLiteral"] = 16384] = "FreshLiteral";
            ObjectFlags[ObjectFlags["ArrayLiteral"] = 32768] = "ArrayLiteral";
            /* @internal */
            ObjectFlags[ObjectFlags["PrimitiveUnion"] = 65536] = "PrimitiveUnion";
            /* @internal */
            ObjectFlags[ObjectFlags["ContainsWideningType"] = 131072] = "ContainsWideningType";
            /* @internal */
            ObjectFlags[ObjectFlags["ContainsObjectOrArrayLiteral"] = 262144] = "ContainsObjectOrArrayLiteral";
            /* @internal */
            ObjectFlags[ObjectFlags["NonInferrableType"] = 524288] = "NonInferrableType";
            /* @internal */
            ObjectFlags[ObjectFlags["CouldContainTypeVariablesComputed"] = 1048576] = "CouldContainTypeVariablesComputed";
            /* @internal */
            ObjectFlags[ObjectFlags["CouldContainTypeVariables"] = 2097152] = "CouldContainTypeVariables";
            ObjectFlags[ObjectFlags["ClassOrInterface"] = 3] = "ClassOrInterface";
            /* @internal */
            ObjectFlags[ObjectFlags["RequiresWidening"] = 393216] = "RequiresWidening";
            /* @internal */
            ObjectFlags[ObjectFlags["PropagatingFlags"] = 917504] = "PropagatingFlags";
            // Object flags that uniquely identify the kind of ObjectType
            /* @internal */
            ObjectFlags[ObjectFlags["ObjectTypeKindMask"] = 1343] = "ObjectTypeKindMask";
            // Flags that require TypeFlags.Object
            ObjectFlags[ObjectFlags["ContainsSpread"] = 4194304] = "ContainsSpread";
            ObjectFlags[ObjectFlags["ObjectRestType"] = 8388608] = "ObjectRestType";
            /* @internal */
            ObjectFlags[ObjectFlags["IsClassInstanceClone"] = 16777216] = "IsClassInstanceClone";
            // Flags that require TypeFlags.Object and ObjectFlags.Reference
            /* @internal */
            ObjectFlags[ObjectFlags["IdenticalBaseTypeCalculated"] = 33554432] = "IdenticalBaseTypeCalculated";
            /* @internal */
            ObjectFlags[ObjectFlags["IdenticalBaseTypeExists"] = 67108864] = "IdenticalBaseTypeExists";
            // Flags that require TypeFlags.UnionOrIntersection or TypeFlags.Substitution
            /* @internal */
            ObjectFlags[ObjectFlags["IsGenericTypeComputed"] = 4194304] = "IsGenericTypeComputed";
            /* @internal */
            ObjectFlags[ObjectFlags["IsGenericObjectType"] = 8388608] = "IsGenericObjectType";
            /* @internal */
            ObjectFlags[ObjectFlags["IsGenericIndexType"] = 16777216] = "IsGenericIndexType";
            /* @internal */
            ObjectFlags[ObjectFlags["IsGenericType"] = 25165824] = "IsGenericType";
            // Flags that require TypeFlags.Union
            /* @internal */
            ObjectFlags[ObjectFlags["ContainsIntersections"] = 33554432] = "ContainsIntersections";
            // Flags that require TypeFlags.Intersection
            /* @internal */
            ObjectFlags[ObjectFlags["IsNeverIntersectionComputed"] = 33554432] = "IsNeverIntersectionComputed";
            /* @internal */
            ObjectFlags[ObjectFlags["IsNeverIntersection"] = 67108864] = "IsNeverIntersection";
        })(ObjectFlags = ts.ObjectFlags || (ts.ObjectFlags = {}));
        /* @internal */
        var VarianceFlags;
        (function (VarianceFlags) {
            VarianceFlags[VarianceFlags["Invariant"] = 0] = "Invariant";
            VarianceFlags[VarianceFlags["Covariant"] = 1] = "Covariant";
            VarianceFlags[VarianceFlags["Contravariant"] = 2] = "Contravariant";
            VarianceFlags[VarianceFlags["Bivariant"] = 3] = "Bivariant";
            VarianceFlags[VarianceFlags["Independent"] = 4] = "Independent";
            VarianceFlags[VarianceFlags["VarianceMask"] = 7] = "VarianceMask";
            VarianceFlags[VarianceFlags["Unmeasurable"] = 8] = "Unmeasurable";
            VarianceFlags[VarianceFlags["Unreliable"] = 16] = "Unreliable";
            VarianceFlags[VarianceFlags["AllowsStructuralFallback"] = 24] = "AllowsStructuralFallback";
        })(VarianceFlags = ts.VarianceFlags || (ts.VarianceFlags = {}));
        var ElementFlags;
        (function (ElementFlags) {
            ElementFlags[ElementFlags["Required"] = 1] = "Required";
            ElementFlags[ElementFlags["Optional"] = 2] = "Optional";
            ElementFlags[ElementFlags["Rest"] = 4] = "Rest";
            ElementFlags[ElementFlags["Variadic"] = 8] = "Variadic";
            ElementFlags[ElementFlags["Fixed"] = 3] = "Fixed";
            ElementFlags[ElementFlags["Variable"] = 12] = "Variable";
            ElementFlags[ElementFlags["NonRequired"] = 14] = "NonRequired";
            ElementFlags[ElementFlags["NonRest"] = 11] = "NonRest";
        })(ElementFlags = ts.ElementFlags || (ts.ElementFlags = {}));
        /* @internal */
        var AccessFlags;
        (function (AccessFlags) {
            AccessFlags[AccessFlags["None"] = 0] = "None";
            AccessFlags[AccessFlags["IncludeUndefined"] = 1] = "IncludeUndefined";
            AccessFlags[AccessFlags["NoIndexSignatures"] = 2] = "NoIndexSignatures";
            AccessFlags[AccessFlags["Writing"] = 4] = "Writing";
            AccessFlags[AccessFlags["CacheSymbol"] = 8] = "CacheSymbol";
            AccessFlags[AccessFlags["NoTupleBoundsCheck"] = 16] = "NoTupleBoundsCheck";
            AccessFlags[AccessFlags["ExpressionPosition"] = 32] = "ExpressionPosition";
            AccessFlags[AccessFlags["ReportDeprecated"] = 64] = "ReportDeprecated";
            AccessFlags[AccessFlags["SuppressNoImplicitAnyError"] = 128] = "SuppressNoImplicitAnyError";
            AccessFlags[AccessFlags["Contextual"] = 256] = "Contextual";
            AccessFlags[AccessFlags["Persistent"] = 1] = "Persistent";
        })(AccessFlags = ts.AccessFlags || (ts.AccessFlags = {}));
        /* @internal */
        var JsxReferenceKind;
        (function (JsxReferenceKind) {
            JsxReferenceKind[JsxReferenceKind["Component"] = 0] = "Component";
            JsxReferenceKind[JsxReferenceKind["Function"] = 1] = "Function";
            JsxReferenceKind[JsxReferenceKind["Mixed"] = 2] = "Mixed";
        })(JsxReferenceKind = ts.JsxReferenceKind || (ts.JsxReferenceKind = {}));
        var SignatureKind;
        (function (SignatureKind) {
            SignatureKind[SignatureKind["Call"] = 0] = "Call";
            SignatureKind[SignatureKind["Construct"] = 1] = "Construct";
        })(SignatureKind = ts.SignatureKind || (ts.SignatureKind = {}));
        /* @internal */
        var SignatureFlags;
        (function (SignatureFlags) {
            SignatureFlags[SignatureFlags["None"] = 0] = "None";
            // Propagating flags
            SignatureFlags[SignatureFlags["HasRestParameter"] = 1] = "HasRestParameter";
            SignatureFlags[SignatureFlags["HasLiteralTypes"] = 2] = "HasLiteralTypes";
            SignatureFlags[SignatureFlags["Abstract"] = 4] = "Abstract";
            // Non-propagating flags
            SignatureFlags[SignatureFlags["IsInnerCallChain"] = 8] = "IsInnerCallChain";
            SignatureFlags[SignatureFlags["IsOuterCallChain"] = 16] = "IsOuterCallChain";
            SignatureFlags[SignatureFlags["IsUntypedSignatureInJSFile"] = 32] = "IsUntypedSignatureInJSFile";
            // We do not propagate `IsInnerCallChain` or `IsOuterCallChain` to instantiated signatures, as that would result in us
            // attempting to add `| undefined` on each recursive call to `getReturnTypeOfSignature` when
            // instantiating the return type.
            SignatureFlags[SignatureFlags["PropagatingFlags"] = 39] = "PropagatingFlags";
            SignatureFlags[SignatureFlags["CallChainFlags"] = 24] = "CallChainFlags";
        })(SignatureFlags = ts.SignatureFlags || (ts.SignatureFlags = {}));
        var IndexKind;
        (function (IndexKind) {
            IndexKind[IndexKind["String"] = 0] = "String";
            IndexKind[IndexKind["Number"] = 1] = "Number";
        })(IndexKind = ts.IndexKind || (ts.IndexKind = {}));
        /* @internal */
        var TypeMapKind;
        (function (TypeMapKind) {
            TypeMapKind[TypeMapKind["Simple"] = 0] = "Simple";
            TypeMapKind[TypeMapKind["Array"] = 1] = "Array";
            TypeMapKind[TypeMapKind["Function"] = 2] = "Function";
            TypeMapKind[TypeMapKind["Composite"] = 3] = "Composite";
            TypeMapKind[TypeMapKind["Merged"] = 4] = "Merged";
        })(TypeMapKind = ts.TypeMapKind || (ts.TypeMapKind = {}));
        var InferencePriority;
        (function (InferencePriority) {
            InferencePriority[InferencePriority["NakedTypeVariable"] = 1] = "NakedTypeVariable";
            InferencePriority[InferencePriority["SpeculativeTuple"] = 2] = "SpeculativeTuple";
            InferencePriority[InferencePriority["SubstituteSource"] = 4] = "SubstituteSource";
            InferencePriority[InferencePriority["HomomorphicMappedType"] = 8] = "HomomorphicMappedType";
            InferencePriority[InferencePriority["PartialHomomorphicMappedType"] = 16] = "PartialHomomorphicMappedType";
            InferencePriority[InferencePriority["MappedTypeConstraint"] = 32] = "MappedTypeConstraint";
            InferencePriority[InferencePriority["ContravariantConditional"] = 64] = "ContravariantConditional";
            InferencePriority[InferencePriority["ReturnType"] = 128] = "ReturnType";
            InferencePriority[InferencePriority["LiteralKeyof"] = 256] = "LiteralKeyof";
            InferencePriority[InferencePriority["NoConstraints"] = 512] = "NoConstraints";
            InferencePriority[InferencePriority["AlwaysStrict"] = 1024] = "AlwaysStrict";
            InferencePriority[InferencePriority["MaxValue"] = 2048] = "MaxValue";
            InferencePriority[InferencePriority["PriorityImpliesCombination"] = 416] = "PriorityImpliesCombination";
            InferencePriority[InferencePriority["Circularity"] = -1] = "Circularity";
        })(InferencePriority = ts.InferencePriority || (ts.InferencePriority = {}));
        /* @internal */
        var InferenceFlags;
        (function (InferenceFlags) {
            InferenceFlags[InferenceFlags["None"] = 0] = "None";
            InferenceFlags[InferenceFlags["NoDefault"] = 1] = "NoDefault";
            InferenceFlags[InferenceFlags["AnyDefault"] = 2] = "AnyDefault";
            InferenceFlags[InferenceFlags["SkippedGenericFunction"] = 4] = "SkippedGenericFunction";
        })(InferenceFlags = ts.InferenceFlags || (ts.InferenceFlags = {}));
        /**
         * Ternary values are defined such that
         * x & y picks the lesser in the order False < Unknown < Maybe < True, and
         * x | y picks the greater in the order False < Unknown < Maybe < True.
         * Generally, Ternary.Maybe is used as the result of a relation that depends on itself, and
         * Ternary.Unknown is used as the result of a variance check that depends on itself. We make
         * a distinction because we don't want to cache circular variance check results.
         */
        /* @internal */
        var Ternary;
        (function (Ternary) {
            Ternary[Ternary["False"] = 0] = "False";
            Ternary[Ternary["Unknown"] = 1] = "Unknown";
            Ternary[Ternary["Maybe"] = 3] = "Maybe";
            Ternary[Ternary["True"] = -1] = "True";
        })(Ternary = ts.Ternary || (ts.Ternary = {}));
        /* @internal */
        var AssignmentDeclarationKind;
        (function (AssignmentDeclarationKind) {
            AssignmentDeclarationKind[AssignmentDeclarationKind["None"] = 0] = "None";
            /// exports.name = expr
            /// module.exports.name = expr
            AssignmentDeclarationKind[AssignmentDeclarationKind["ExportsProperty"] = 1] = "ExportsProperty";
            /// module.exports = expr
            AssignmentDeclarationKind[AssignmentDeclarationKind["ModuleExports"] = 2] = "ModuleExports";
            /// className.prototype.name = expr
            AssignmentDeclarationKind[AssignmentDeclarationKind["PrototypeProperty"] = 3] = "PrototypeProperty";
            /// this.name = expr
            AssignmentDeclarationKind[AssignmentDeclarationKind["ThisProperty"] = 4] = "ThisProperty";
            // F.name = expr
            AssignmentDeclarationKind[AssignmentDeclarationKind["Property"] = 5] = "Property";
            // F.prototype = { ... }
            AssignmentDeclarationKind[AssignmentDeclarationKind["Prototype"] = 6] = "Prototype";
            // Object.defineProperty(x, 'name', { value: any, writable?: boolean (false by default) });
            // Object.defineProperty(x, 'name', { get: Function, set: Function });
            // Object.defineProperty(x, 'name', { get: Function });
            // Object.defineProperty(x, 'name', { set: Function });
            AssignmentDeclarationKind[AssignmentDeclarationKind["ObjectDefinePropertyValue"] = 7] = "ObjectDefinePropertyValue";
            // Object.defineProperty(exports || module.exports, 'name', ...);
            AssignmentDeclarationKind[AssignmentDeclarationKind["ObjectDefinePropertyExports"] = 8] = "ObjectDefinePropertyExports";
            // Object.defineProperty(Foo.prototype, 'name', ...);
            AssignmentDeclarationKind[AssignmentDeclarationKind["ObjectDefinePrototypeProperty"] = 9] = "ObjectDefinePrototypeProperty";
        })(AssignmentDeclarationKind = ts.AssignmentDeclarationKind || (ts.AssignmentDeclarationKind = {}));
        var DiagnosticCategory;
        (function (DiagnosticCategory) {
            DiagnosticCategory[DiagnosticCategory["Warning"] = 0] = "Warning";
            DiagnosticCategory[DiagnosticCategory["Error"] = 1] = "Error";
            DiagnosticCategory[DiagnosticCategory["Suggestion"] = 2] = "Suggestion";
            DiagnosticCategory[DiagnosticCategory["Message"] = 3] = "Message";
        })(DiagnosticCategory = ts.DiagnosticCategory || (ts.DiagnosticCategory = {}));
        /* @internal */
        function diagnosticCategoryName(d, lowerCase) {
            if (lowerCase === void 0) {
                lowerCase = true;
            }
            var name = DiagnosticCategory[d.category];
            return lowerCase ? name.toLowerCase() : name;
        }
        ts.diagnosticCategoryName = diagnosticCategoryName;
        var ModuleResolutionKind;
        (function (ModuleResolutionKind) {
            ModuleResolutionKind[ModuleResolutionKind["Classic"] = 1] = "Classic";
            ModuleResolutionKind[ModuleResolutionKind["NodeJs"] = 2] = "NodeJs";
        })(ModuleResolutionKind = ts.ModuleResolutionKind || (ts.ModuleResolutionKind = {}));
        var WatchFileKind;
        (function (WatchFileKind) {
            WatchFileKind[WatchFileKind["FixedPollingInterval"] = 0] = "FixedPollingInterval";
            WatchFileKind[WatchFileKind["PriorityPollingInterval"] = 1] = "PriorityPollingInterval";
            WatchFileKind[WatchFileKind["DynamicPriorityPolling"] = 2] = "DynamicPriorityPolling";
            WatchFileKind[WatchFileKind["FixedChunkSizePolling"] = 3] = "FixedChunkSizePolling";
            WatchFileKind[WatchFileKind["UseFsEvents"] = 4] = "UseFsEvents";
            WatchFileKind[WatchFileKind["UseFsEventsOnParentDirectory"] = 5] = "UseFsEventsOnParentDirectory";
        })(WatchFileKind = ts.WatchFileKind || (ts.WatchFileKind = {}));
        var WatchDirectoryKind;
        (function (WatchDirectoryKind) {
            WatchDirectoryKind[WatchDirectoryKind["UseFsEvents"] = 0] = "UseFsEvents";
            WatchDirectoryKind[WatchDirectoryKind["FixedPollingInterval"] = 1] = "FixedPollingInterval";
            WatchDirectoryKind[WatchDirectoryKind["DynamicPriorityPolling"] = 2] = "DynamicPriorityPolling";
            WatchDirectoryKind[WatchDirectoryKind["FixedChunkSizePolling"] = 3] = "FixedChunkSizePolling";
        })(WatchDirectoryKind = ts.WatchDirectoryKind || (ts.WatchDirectoryKind = {}));
        var PollingWatchKind;
        (function (PollingWatchKind) {
            PollingWatchKind[PollingWatchKind["FixedInterval"] = 0] = "FixedInterval";
            PollingWatchKind[PollingWatchKind["PriorityInterval"] = 1] = "PriorityInterval";
            PollingWatchKind[PollingWatchKind["DynamicPriority"] = 2] = "DynamicPriority";
            PollingWatchKind[PollingWatchKind["FixedChunkSize"] = 3] = "FixedChunkSize";
        })(PollingWatchKind = ts.PollingWatchKind || (ts.PollingWatchKind = {}));
        var ModuleKind;
        (function (ModuleKind) {
            ModuleKind[ModuleKind["None"] = 0] = "None";
            ModuleKind[ModuleKind["CommonJS"] = 1] = "CommonJS";
            ModuleKind[ModuleKind["AMD"] = 2] = "AMD";
            ModuleKind[ModuleKind["UMD"] = 3] = "UMD";
            ModuleKind[ModuleKind["System"] = 4] = "System";
            // NOTE: ES module kinds should be contiguous to more easily check whether a module kind is *any* ES module kind.
            //       Non-ES module kinds should not come between ES2015 (the earliest ES module kind) and ESNext (the last ES
            //       module kind).
            ModuleKind[ModuleKind["ES2015"] = 5] = "ES2015";
            ModuleKind[ModuleKind["ES2020"] = 6] = "ES2020";
            ModuleKind[ModuleKind["ESNext"] = 99] = "ESNext";
        })(ModuleKind = ts.ModuleKind || (ts.ModuleKind = {}));
        var JsxEmit;
        (function (JsxEmit) {
            JsxEmit[JsxEmit["None"] = 0] = "None";
            JsxEmit[JsxEmit["Preserve"] = 1] = "Preserve";
            JsxEmit[JsxEmit["React"] = 2] = "React";
            JsxEmit[JsxEmit["ReactNative"] = 3] = "ReactNative";
            JsxEmit[JsxEmit["ReactJSX"] = 4] = "ReactJSX";
            JsxEmit[JsxEmit["ReactJSXDev"] = 5] = "ReactJSXDev";
        })(JsxEmit = ts.JsxEmit || (ts.JsxEmit = {}));
        var ImportsNotUsedAsValues;
        (function (ImportsNotUsedAsValues) {
            ImportsNotUsedAsValues[ImportsNotUsedAsValues["Remove"] = 0] = "Remove";
            ImportsNotUsedAsValues[ImportsNotUsedAsValues["Preserve"] = 1] = "Preserve";
            ImportsNotUsedAsValues[ImportsNotUsedAsValues["Error"] = 2] = "Error";
        })(ImportsNotUsedAsValues = ts.ImportsNotUsedAsValues || (ts.ImportsNotUsedAsValues = {}));
        var NewLineKind;
        (function (NewLineKind) {
            NewLineKind[NewLineKind["CarriageReturnLineFeed"] = 0] = "CarriageReturnLineFeed";
            NewLineKind[NewLineKind["LineFeed"] = 1] = "LineFeed";
        })(NewLineKind = ts.NewLineKind || (ts.NewLineKind = {}));
        var ScriptKind;
        (function (ScriptKind) {
            ScriptKind[ScriptKind["Unknown"] = 0] = "Unknown";
            ScriptKind[ScriptKind["JS"] = 1] = "JS";
            ScriptKind[ScriptKind["JSX"] = 2] = "JSX";
            ScriptKind[ScriptKind["TS"] = 3] = "TS";
            ScriptKind[ScriptKind["TSX"] = 4] = "TSX";
            ScriptKind[ScriptKind["External"] = 5] = "External";
            ScriptKind[ScriptKind["JSON"] = 6] = "JSON";
            /**
             * Used on extensions that doesn't define the ScriptKind but the content defines it.
             * Deferred extensions are going to be included in all project contexts.
             */
            ScriptKind[ScriptKind["Deferred"] = 7] = "Deferred";
        })(ScriptKind = ts.ScriptKind || (ts.ScriptKind = {}));
        var ScriptTarget;
        (function (ScriptTarget) {
            ScriptTarget[ScriptTarget["ES3"] = 0] = "ES3";
            ScriptTarget[ScriptTarget["ES5"] = 1] = "ES5";
            ScriptTarget[ScriptTarget["ES2015"] = 2] = "ES2015";
            ScriptTarget[ScriptTarget["ES2016"] = 3] = "ES2016";
            ScriptTarget[ScriptTarget["ES2017"] = 4] = "ES2017";
            ScriptTarget[ScriptTarget["ES2018"] = 5] = "ES2018";
            ScriptTarget[ScriptTarget["ES2019"] = 6] = "ES2019";
            ScriptTarget[ScriptTarget["ES2020"] = 7] = "ES2020";
            ScriptTarget[ScriptTarget["ES2021"] = 8] = "ES2021";
            ScriptTarget[ScriptTarget["ESNext"] = 99] = "ESNext";
            ScriptTarget[ScriptTarget["JSON"] = 100] = "JSON";
            ScriptTarget[ScriptTarget["Latest"] = 99] = "Latest";
        })(ScriptTarget = ts.ScriptTarget || (ts.ScriptTarget = {}));
        var LanguageVariant;
        (function (LanguageVariant) {
            LanguageVariant[LanguageVariant["Standard"] = 0] = "Standard";
            LanguageVariant[LanguageVariant["JSX"] = 1] = "JSX";
        })(LanguageVariant = ts.LanguageVariant || (ts.LanguageVariant = {}));
        var WatchDirectoryFlags;
        (function (WatchDirectoryFlags) {
            WatchDirectoryFlags[WatchDirectoryFlags["None"] = 0] = "None";
            WatchDirectoryFlags[WatchDirectoryFlags["Recursive"] = 1] = "Recursive";
        })(WatchDirectoryFlags = ts.WatchDirectoryFlags || (ts.WatchDirectoryFlags = {}));
        /* @internal */
        var CharacterCodes;
        (function (CharacterCodes) {
            CharacterCodes[CharacterCodes["nullCharacter"] = 0] = "nullCharacter";
            CharacterCodes[CharacterCodes["maxAsciiCharacter"] = 127] = "maxAsciiCharacter";
            CharacterCodes[CharacterCodes["lineFeed"] = 10] = "lineFeed";
            CharacterCodes[CharacterCodes["carriageReturn"] = 13] = "carriageReturn";
            CharacterCodes[CharacterCodes["lineSeparator"] = 8232] = "lineSeparator";
            CharacterCodes[CharacterCodes["paragraphSeparator"] = 8233] = "paragraphSeparator";
            CharacterCodes[CharacterCodes["nextLine"] = 133] = "nextLine";
            // Unicode 3.0 space characters
            CharacterCodes[CharacterCodes["space"] = 32] = "space";
            CharacterCodes[CharacterCodes["nonBreakingSpace"] = 160] = "nonBreakingSpace";
            CharacterCodes[CharacterCodes["enQuad"] = 8192] = "enQuad";
            CharacterCodes[CharacterCodes["emQuad"] = 8193] = "emQuad";
            CharacterCodes[CharacterCodes["enSpace"] = 8194] = "enSpace";
            CharacterCodes[CharacterCodes["emSpace"] = 8195] = "emSpace";
            CharacterCodes[CharacterCodes["threePerEmSpace"] = 8196] = "threePerEmSpace";
            CharacterCodes[CharacterCodes["fourPerEmSpace"] = 8197] = "fourPerEmSpace";
            CharacterCodes[CharacterCodes["sixPerEmSpace"] = 8198] = "sixPerEmSpace";
            CharacterCodes[CharacterCodes["figureSpace"] = 8199] = "figureSpace";
            CharacterCodes[CharacterCodes["punctuationSpace"] = 8200] = "punctuationSpace";
            CharacterCodes[CharacterCodes["thinSpace"] = 8201] = "thinSpace";
            CharacterCodes[CharacterCodes["hairSpace"] = 8202] = "hairSpace";
            CharacterCodes[CharacterCodes["zeroWidthSpace"] = 8203] = "zeroWidthSpace";
            CharacterCodes[CharacterCodes["narrowNoBreakSpace"] = 8239] = "narrowNoBreakSpace";
            CharacterCodes[CharacterCodes["ideographicSpace"] = 12288] = "ideographicSpace";
            CharacterCodes[CharacterCodes["mathematicalSpace"] = 8287] = "mathematicalSpace";
            CharacterCodes[CharacterCodes["ogham"] = 5760] = "ogham";
            CharacterCodes[CharacterCodes["_"] = 95] = "_";
            CharacterCodes[CharacterCodes["$"] = 36] = "$";
            CharacterCodes[CharacterCodes["_0"] = 48] = "_0";
            CharacterCodes[CharacterCodes["_1"] = 49] = "_1";
            CharacterCodes[CharacterCodes["_2"] = 50] = "_2";
            CharacterCodes[CharacterCodes["_3"] = 51] = "_3";
            CharacterCodes[CharacterCodes["_4"] = 52] = "_4";
            CharacterCodes[CharacterCodes["_5"] = 53] = "_5";
            CharacterCodes[CharacterCodes["_6"] = 54] = "_6";
            CharacterCodes[CharacterCodes["_7"] = 55] = "_7";
            CharacterCodes[CharacterCodes["_8"] = 56] = "_8";
            CharacterCodes[CharacterCodes["_9"] = 57] = "_9";
            CharacterCodes[CharacterCodes["a"] = 97] = "a";
            CharacterCodes[CharacterCodes["b"] = 98] = "b";
            CharacterCodes[CharacterCodes["c"] = 99] = "c";
            CharacterCodes[CharacterCodes["d"] = 100] = "d";
            CharacterCodes[CharacterCodes["e"] = 101] = "e";
            CharacterCodes[CharacterCodes["f"] = 102] = "f";
            CharacterCodes[CharacterCodes["g"] = 103] = "g";
            CharacterCodes[CharacterCodes["h"] = 104] = "h";
            CharacterCodes[CharacterCodes["i"] = 105] = "i";
            CharacterCodes[CharacterCodes["j"] = 106] = "j";
            CharacterCodes[CharacterCodes["k"] = 107] = "k";
            CharacterCodes[CharacterCodes["l"] = 108] = "l";
            CharacterCodes[CharacterCodes["m"] = 109] = "m";
            CharacterCodes[CharacterCodes["n"] = 110] = "n";
            CharacterCodes[CharacterCodes["o"] = 111] = "o";
            CharacterCodes[CharacterCodes["p"] = 112] = "p";
            CharacterCodes[CharacterCodes["q"] = 113] = "q";
            CharacterCodes[CharacterCodes["r"] = 114] = "r";
            CharacterCodes[CharacterCodes["s"] = 115] = "s";
            CharacterCodes[CharacterCodes["t"] = 116] = "t";
            CharacterCodes[CharacterCodes["u"] = 117] = "u";
            CharacterCodes[CharacterCodes["v"] = 118] = "v";
            CharacterCodes[CharacterCodes["w"] = 119] = "w";
            CharacterCodes[CharacterCodes["x"] = 120] = "x";
            CharacterCodes[CharacterCodes["y"] = 121] = "y";
            CharacterCodes[CharacterCodes["z"] = 122] = "z";
            CharacterCodes[CharacterCodes["A"] = 65] = "A";
            CharacterCodes[CharacterCodes["B"] = 66] = "B";
            CharacterCodes[CharacterCodes["C"] = 67] = "C";
            CharacterCodes[CharacterCodes["D"] = 68] = "D";
            CharacterCodes[CharacterCodes["E"] = 69] = "E";
            CharacterCodes[CharacterCodes["F"] = 70] = "F";
            CharacterCodes[CharacterCodes["G"] = 71] = "G";
            CharacterCodes[CharacterCodes["H"] = 72] = "H";
            CharacterCodes[CharacterCodes["I"] = 73] = "I";
            CharacterCodes[CharacterCodes["J"] = 74] = "J";
            CharacterCodes[CharacterCodes["K"] = 75] = "K";
            CharacterCodes[CharacterCodes["L"] = 76] = "L";
            CharacterCodes[CharacterCodes["M"] = 77] = "M";
            CharacterCodes[CharacterCodes["N"] = 78] = "N";
            CharacterCodes[CharacterCodes["O"] = 79] = "O";
            CharacterCodes[CharacterCodes["P"] = 80] = "P";
            CharacterCodes[CharacterCodes["Q"] = 81] = "Q";
            CharacterCodes[CharacterCodes["R"] = 82] = "R";
            CharacterCodes[CharacterCodes["S"] = 83] = "S";
            CharacterCodes[CharacterCodes["T"] = 84] = "T";
            CharacterCodes[CharacterCodes["U"] = 85] = "U";
            CharacterCodes[CharacterCodes["V"] = 86] = "V";
            CharacterCodes[CharacterCodes["W"] = 87] = "W";
            CharacterCodes[CharacterCodes["X"] = 88] = "X";
            CharacterCodes[CharacterCodes["Y"] = 89] = "Y";
            CharacterCodes[CharacterCodes["Z"] = 90] = "Z";
            CharacterCodes[CharacterCodes["ampersand"] = 38] = "ampersand";
            CharacterCodes[CharacterCodes["asterisk"] = 42] = "asterisk";
            CharacterCodes[CharacterCodes["at"] = 64] = "at";
            CharacterCodes[CharacterCodes["backslash"] = 92] = "backslash";
            CharacterCodes[CharacterCodes["backtick"] = 96] = "backtick";
            CharacterCodes[CharacterCodes["bar"] = 124] = "bar";
            CharacterCodes[CharacterCodes["caret"] = 94] = "caret";
            CharacterCodes[CharacterCodes["closeBrace"] = 125] = "closeBrace";
            CharacterCodes[CharacterCodes["closeBracket"] = 93] = "closeBracket";
            CharacterCodes[CharacterCodes["closeParen"] = 41] = "closeParen";
            CharacterCodes[CharacterCodes["colon"] = 58] = "colon";
            CharacterCodes[CharacterCodes["comma"] = 44] = "comma";
            CharacterCodes[CharacterCodes["dot"] = 46] = "dot";
            CharacterCodes[CharacterCodes["doubleQuote"] = 34] = "doubleQuote";
            CharacterCodes[CharacterCodes["equals"] = 61] = "equals";
            CharacterCodes[CharacterCodes["exclamation"] = 33] = "exclamation";
            CharacterCodes[CharacterCodes["greaterThan"] = 62] = "greaterThan";
            CharacterCodes[CharacterCodes["hash"] = 35] = "hash";
            CharacterCodes[CharacterCodes["lessThan"] = 60] = "lessThan";
            CharacterCodes[CharacterCodes["minus"] = 45] = "minus";
            CharacterCodes[CharacterCodes["openBrace"] = 123] = "openBrace";
            CharacterCodes[CharacterCodes["openBracket"] = 91] = "openBracket";
            CharacterCodes[CharacterCodes["openParen"] = 40] = "openParen";
            CharacterCodes[CharacterCodes["percent"] = 37] = "percent";
            CharacterCodes[CharacterCodes["plus"] = 43] = "plus";
            CharacterCodes[CharacterCodes["question"] = 63] = "question";
            CharacterCodes[CharacterCodes["semicolon"] = 59] = "semicolon";
            CharacterCodes[CharacterCodes["singleQuote"] = 39] = "singleQuote";
            CharacterCodes[CharacterCodes["slash"] = 47] = "slash";
            CharacterCodes[CharacterCodes["tilde"] = 126] = "tilde";
            CharacterCodes[CharacterCodes["backspace"] = 8] = "backspace";
            CharacterCodes[CharacterCodes["formFeed"] = 12] = "formFeed";
            CharacterCodes[CharacterCodes["byteOrderMark"] = 65279] = "byteOrderMark";
            CharacterCodes[CharacterCodes["tab"] = 9] = "tab";
            CharacterCodes[CharacterCodes["verticalTab"] = 11] = "verticalTab";
        })(CharacterCodes = ts.CharacterCodes || (ts.CharacterCodes = {}));
        var Extension;
        (function (Extension) {
            Extension["Ts"] = ".ts";
            Extension["Tsx"] = ".tsx";
            Extension["Dts"] = ".d.ts";
            Extension["Js"] = ".js";
            Extension["Jsx"] = ".jsx";
            Extension["Json"] = ".json";
            Extension["TsBuildInfo"] = ".tsbuildinfo";
        })(Extension = ts.Extension || (ts.Extension = {}));
        /* @internal */
        var TransformFlags;
        (function (TransformFlags) {
            TransformFlags[TransformFlags["None"] = 0] = "None";
            // Facts
            // - Flags used to indicate that a node or subtree contains syntax that requires transformation.
            TransformFlags[TransformFlags["ContainsTypeScript"] = 1] = "ContainsTypeScript";
            TransformFlags[TransformFlags["ContainsJsx"] = 2] = "ContainsJsx";
            TransformFlags[TransformFlags["ContainsESNext"] = 4] = "ContainsESNext";
            TransformFlags[TransformFlags["ContainsES2021"] = 8] = "ContainsES2021";
            TransformFlags[TransformFlags["ContainsES2020"] = 16] = "ContainsES2020";
            TransformFlags[TransformFlags["ContainsES2019"] = 32] = "ContainsES2019";
            TransformFlags[TransformFlags["ContainsES2018"] = 64] = "ContainsES2018";
            TransformFlags[TransformFlags["ContainsES2017"] = 128] = "ContainsES2017";
            TransformFlags[TransformFlags["ContainsES2016"] = 256] = "ContainsES2016";
            TransformFlags[TransformFlags["ContainsES2015"] = 512] = "ContainsES2015";
            TransformFlags[TransformFlags["ContainsGenerator"] = 1024] = "ContainsGenerator";
            TransformFlags[TransformFlags["ContainsDestructuringAssignment"] = 2048] = "ContainsDestructuringAssignment";
            // Markers
            // - Flags used to indicate that a subtree contains a specific transformation.
            TransformFlags[TransformFlags["ContainsTypeScriptClassSyntax"] = 4096] = "ContainsTypeScriptClassSyntax";
            TransformFlags[TransformFlags["ContainsLexicalThis"] = 8192] = "ContainsLexicalThis";
            TransformFlags[TransformFlags["ContainsRestOrSpread"] = 16384] = "ContainsRestOrSpread";
            TransformFlags[TransformFlags["ContainsObjectRestOrSpread"] = 32768] = "ContainsObjectRestOrSpread";
            TransformFlags[TransformFlags["ContainsComputedPropertyName"] = 65536] = "ContainsComputedPropertyName";
            TransformFlags[TransformFlags["ContainsBlockScopedBinding"] = 131072] = "ContainsBlockScopedBinding";
            TransformFlags[TransformFlags["ContainsBindingPattern"] = 262144] = "ContainsBindingPattern";
            TransformFlags[TransformFlags["ContainsYield"] = 524288] = "ContainsYield";
            TransformFlags[TransformFlags["ContainsAwait"] = 1048576] = "ContainsAwait";
            TransformFlags[TransformFlags["ContainsHoistedDeclarationOrCompletion"] = 2097152] = "ContainsHoistedDeclarationOrCompletion";
            TransformFlags[TransformFlags["ContainsDynamicImport"] = 4194304] = "ContainsDynamicImport";
            TransformFlags[TransformFlags["ContainsClassFields"] = 8388608] = "ContainsClassFields";
            TransformFlags[TransformFlags["ContainsPossibleTopLevelAwait"] = 16777216] = "ContainsPossibleTopLevelAwait";
            TransformFlags[TransformFlags["ContainsLexicalSuper"] = 33554432] = "ContainsLexicalSuper";
            TransformFlags[TransformFlags["ContainsUpdateExpressionForIdentifier"] = 67108864] = "ContainsUpdateExpressionForIdentifier";
            // Please leave this as 1 << 29.
            // It is the maximum bit we can set before we outgrow the size of a v8 small integer (SMI) on an x86 system.
            // It is a good reminder of how much room we have left
            TransformFlags[TransformFlags["HasComputedFlags"] = 536870912] = "HasComputedFlags";
            // Assertions
            // - Bitmasks that are used to assert facts about the syntax of a node and its subtree.
            TransformFlags[TransformFlags["AssertTypeScript"] = 1] = "AssertTypeScript";
            TransformFlags[TransformFlags["AssertJsx"] = 2] = "AssertJsx";
            TransformFlags[TransformFlags["AssertESNext"] = 4] = "AssertESNext";
            TransformFlags[TransformFlags["AssertES2021"] = 8] = "AssertES2021";
            TransformFlags[TransformFlags["AssertES2020"] = 16] = "AssertES2020";
            TransformFlags[TransformFlags["AssertES2019"] = 32] = "AssertES2019";
            TransformFlags[TransformFlags["AssertES2018"] = 64] = "AssertES2018";
            TransformFlags[TransformFlags["AssertES2017"] = 128] = "AssertES2017";
            TransformFlags[TransformFlags["AssertES2016"] = 256] = "AssertES2016";
            TransformFlags[TransformFlags["AssertES2015"] = 512] = "AssertES2015";
            TransformFlags[TransformFlags["AssertGenerator"] = 1024] = "AssertGenerator";
            TransformFlags[TransformFlags["AssertDestructuringAssignment"] = 2048] = "AssertDestructuringAssignment";
            // Scope Exclusions
            // - Bitmasks that exclude flags from propagating out of a specific context
            //   into the subtree flags of their container.
            TransformFlags[TransformFlags["OuterExpressionExcludes"] = 536870912] = "OuterExpressionExcludes";
            TransformFlags[TransformFlags["PropertyAccessExcludes"] = 536870912] = "PropertyAccessExcludes";
            TransformFlags[TransformFlags["NodeExcludes"] = 536870912] = "NodeExcludes";
            TransformFlags[TransformFlags["ArrowFunctionExcludes"] = 557748224] = "ArrowFunctionExcludes";
            TransformFlags[TransformFlags["FunctionExcludes"] = 591310848] = "FunctionExcludes";
            TransformFlags[TransformFlags["ConstructorExcludes"] = 591306752] = "ConstructorExcludes";
            TransformFlags[TransformFlags["MethodOrAccessorExcludes"] = 574529536] = "MethodOrAccessorExcludes";
            TransformFlags[TransformFlags["PropertyExcludes"] = 570433536] = "PropertyExcludes";
            TransformFlags[TransformFlags["ClassExcludes"] = 536940544] = "ClassExcludes";
            TransformFlags[TransformFlags["ModuleExcludes"] = 589443072] = "ModuleExcludes";
            TransformFlags[TransformFlags["TypeExcludes"] = -2] = "TypeExcludes";
            TransformFlags[TransformFlags["ObjectLiteralExcludes"] = 536973312] = "ObjectLiteralExcludes";
            TransformFlags[TransformFlags["ArrayLiteralOrCallOrNewExcludes"] = 536887296] = "ArrayLiteralOrCallOrNewExcludes";
            TransformFlags[TransformFlags["VariableDeclarationListExcludes"] = 537165824] = "VariableDeclarationListExcludes";
            TransformFlags[TransformFlags["ParameterExcludes"] = 536870912] = "ParameterExcludes";
            TransformFlags[TransformFlags["CatchClauseExcludes"] = 536903680] = "CatchClauseExcludes";
            TransformFlags[TransformFlags["BindingPatternExcludes"] = 536887296] = "BindingPatternExcludes";
            TransformFlags[TransformFlags["ContainsLexicalThisOrSuper"] = 33562624] = "ContainsLexicalThisOrSuper";
            // Propagating flags
            // - Bitmasks for flags that should propagate from a child
            TransformFlags[TransformFlags["PropertyNamePropagatingFlags"] = 33562624] = "PropertyNamePropagatingFlags";
            // Masks
            // - Additional bitmasks
        })(TransformFlags = ts.TransformFlags || (ts.TransformFlags = {}));
        var EmitFlags;
        (function (EmitFlags) {
            EmitFlags[EmitFlags["None"] = 0] = "None";
            EmitFlags[EmitFlags["SingleLine"] = 1] = "SingleLine";
            EmitFlags[EmitFlags["AdviseOnEmitNode"] = 2] = "AdviseOnEmitNode";
            EmitFlags[EmitFlags["NoSubstitution"] = 4] = "NoSubstitution";
            EmitFlags[EmitFlags["CapturesThis"] = 8] = "CapturesThis";
            EmitFlags[EmitFlags["NoLeadingSourceMap"] = 16] = "NoLeadingSourceMap";
            EmitFlags[EmitFlags["NoTrailingSourceMap"] = 32] = "NoTrailingSourceMap";
            EmitFlags[EmitFlags["NoSourceMap"] = 48] = "NoSourceMap";
            EmitFlags[EmitFlags["NoNestedSourceMaps"] = 64] = "NoNestedSourceMaps";
            EmitFlags[EmitFlags["NoTokenLeadingSourceMaps"] = 128] = "NoTokenLeadingSourceMaps";
            EmitFlags[EmitFlags["NoTokenTrailingSourceMaps"] = 256] = "NoTokenTrailingSourceMaps";
            EmitFlags[EmitFlags["NoTokenSourceMaps"] = 384] = "NoTokenSourceMaps";
            EmitFlags[EmitFlags["NoLeadingComments"] = 512] = "NoLeadingComments";
            EmitFlags[EmitFlags["NoTrailingComments"] = 1024] = "NoTrailingComments";
            EmitFlags[EmitFlags["NoComments"] = 1536] = "NoComments";
            EmitFlags[EmitFlags["NoNestedComments"] = 2048] = "NoNestedComments";
            EmitFlags[EmitFlags["HelperName"] = 4096] = "HelperName";
            EmitFlags[EmitFlags["ExportName"] = 8192] = "ExportName";
            EmitFlags[EmitFlags["LocalName"] = 16384] = "LocalName";
            EmitFlags[EmitFlags["InternalName"] = 32768] = "InternalName";
            EmitFlags[EmitFlags["Indented"] = 65536] = "Indented";
            EmitFlags[EmitFlags["NoIndentation"] = 131072] = "NoIndentation";
            EmitFlags[EmitFlags["AsyncFunctionBody"] = 262144] = "AsyncFunctionBody";
            EmitFlags[EmitFlags["ReuseTempVariableScope"] = 524288] = "ReuseTempVariableScope";
            EmitFlags[EmitFlags["CustomPrologue"] = 1048576] = "CustomPrologue";
            EmitFlags[EmitFlags["NoHoisting"] = 2097152] = "NoHoisting";
            EmitFlags[EmitFlags["HasEndOfDeclarationMarker"] = 4194304] = "HasEndOfDeclarationMarker";
            EmitFlags[EmitFlags["Iterator"] = 8388608] = "Iterator";
            EmitFlags[EmitFlags["NoAsciiEscaping"] = 16777216] = "NoAsciiEscaping";
            /*@internal*/ EmitFlags[EmitFlags["TypeScriptClassWrapper"] = 33554432] = "TypeScriptClassWrapper";
            /*@internal*/ EmitFlags[EmitFlags["NeverApplyImportHelper"] = 67108864] = "NeverApplyImportHelper";
            /*@internal*/ EmitFlags[EmitFlags["IgnoreSourceNewlines"] = 134217728] = "IgnoreSourceNewlines";
            /*@internal*/ EmitFlags[EmitFlags["Immutable"] = 268435456] = "Immutable";
            /*@internal*/ EmitFlags[EmitFlags["IndirectCall"] = 536870912] = "IndirectCall";
        })(EmitFlags = ts.EmitFlags || (ts.EmitFlags = {}));
        /**
         * Used by the checker, this enum keeps track of external emit helpers that should be type
         * checked.
         */
        /* @internal */
        var ExternalEmitHelpers;
        (function (ExternalEmitHelpers) {
            ExternalEmitHelpers[ExternalEmitHelpers["Extends"] = 1] = "Extends";
            ExternalEmitHelpers[ExternalEmitHelpers["Assign"] = 2] = "Assign";
            ExternalEmitHelpers[ExternalEmitHelpers["Rest"] = 4] = "Rest";
            ExternalEmitHelpers[ExternalEmitHelpers["Decorate"] = 8] = "Decorate";
            ExternalEmitHelpers[ExternalEmitHelpers["Metadata"] = 16] = "Metadata";
            ExternalEmitHelpers[ExternalEmitHelpers["Param"] = 32] = "Param";
            ExternalEmitHelpers[ExternalEmitHelpers["Awaiter"] = 64] = "Awaiter";
            ExternalEmitHelpers[ExternalEmitHelpers["Generator"] = 128] = "Generator";
            ExternalEmitHelpers[ExternalEmitHelpers["Values"] = 256] = "Values";
            ExternalEmitHelpers[ExternalEmitHelpers["Read"] = 512] = "Read";
            ExternalEmitHelpers[ExternalEmitHelpers["SpreadArray"] = 1024] = "SpreadArray";
            ExternalEmitHelpers[ExternalEmitHelpers["Await"] = 2048] = "Await";
            ExternalEmitHelpers[ExternalEmitHelpers["AsyncGenerator"] = 4096] = "AsyncGenerator";
            ExternalEmitHelpers[ExternalEmitHelpers["AsyncDelegator"] = 8192] = "AsyncDelegator";
            ExternalEmitHelpers[ExternalEmitHelpers["AsyncValues"] = 16384] = "AsyncValues";
            ExternalEmitHelpers[ExternalEmitHelpers["ExportStar"] = 32768] = "ExportStar";
            ExternalEmitHelpers[ExternalEmitHelpers["ImportStar"] = 65536] = "ImportStar";
            ExternalEmitHelpers[ExternalEmitHelpers["ImportDefault"] = 131072] = "ImportDefault";
            ExternalEmitHelpers[ExternalEmitHelpers["MakeTemplateObject"] = 262144] = "MakeTemplateObject";
            ExternalEmitHelpers[ExternalEmitHelpers["ClassPrivateFieldGet"] = 524288] = "ClassPrivateFieldGet";
            ExternalEmitHelpers[ExternalEmitHelpers["ClassPrivateFieldSet"] = 1048576] = "ClassPrivateFieldSet";
            ExternalEmitHelpers[ExternalEmitHelpers["CreateBinding"] = 2097152] = "CreateBinding";
            ExternalEmitHelpers[ExternalEmitHelpers["FirstEmitHelper"] = 1] = "FirstEmitHelper";
            ExternalEmitHelpers[ExternalEmitHelpers["LastEmitHelper"] = 2097152] = "LastEmitHelper";
            // Helpers included by ES2015 for..of
            ExternalEmitHelpers[ExternalEmitHelpers["ForOfIncludes"] = 256] = "ForOfIncludes";
            // Helpers included by ES2017 for..await..of
            ExternalEmitHelpers[ExternalEmitHelpers["ForAwaitOfIncludes"] = 16384] = "ForAwaitOfIncludes";
            // Helpers included by ES2017 async generators
            ExternalEmitHelpers[ExternalEmitHelpers["AsyncGeneratorIncludes"] = 6144] = "AsyncGeneratorIncludes";
            // Helpers included by yield* in ES2017 async generators
            ExternalEmitHelpers[ExternalEmitHelpers["AsyncDelegatorIncludes"] = 26624] = "AsyncDelegatorIncludes";
            // Helpers included by ES2015 spread
            ExternalEmitHelpers[ExternalEmitHelpers["SpreadIncludes"] = 1536] = "SpreadIncludes";
        })(ExternalEmitHelpers = ts.ExternalEmitHelpers || (ts.ExternalEmitHelpers = {}));
        var EmitHint;
        (function (EmitHint) {
            EmitHint[EmitHint["SourceFile"] = 0] = "SourceFile";
            EmitHint[EmitHint["Expression"] = 1] = "Expression";
            EmitHint[EmitHint["IdentifierName"] = 2] = "IdentifierName";
            EmitHint[EmitHint["MappedTypeParameter"] = 3] = "MappedTypeParameter";
            EmitHint[EmitHint["Unspecified"] = 4] = "Unspecified";
            EmitHint[EmitHint["EmbeddedStatement"] = 5] = "EmbeddedStatement";
            EmitHint[EmitHint["JsxAttributeValue"] = 6] = "JsxAttributeValue";
        })(EmitHint = ts.EmitHint || (ts.EmitHint = {}));
        var OuterExpressionKinds;
        (function (OuterExpressionKinds) {
            OuterExpressionKinds[OuterExpressionKinds["Parentheses"] = 1] = "Parentheses";
            OuterExpressionKinds[OuterExpressionKinds["TypeAssertions"] = 2] = "TypeAssertions";
            OuterExpressionKinds[OuterExpressionKinds["NonNullAssertions"] = 4] = "NonNullAssertions";
            OuterExpressionKinds[OuterExpressionKinds["PartiallyEmittedExpressions"] = 8] = "PartiallyEmittedExpressions";
            OuterExpressionKinds[OuterExpressionKinds["Assertions"] = 6] = "Assertions";
            OuterExpressionKinds[OuterExpressionKinds["All"] = 15] = "All";
        })(OuterExpressionKinds = ts.OuterExpressionKinds || (ts.OuterExpressionKinds = {}));
        /* @internal */
        var LexicalEnvironmentFlags;
        (function (LexicalEnvironmentFlags) {
            LexicalEnvironmentFlags[LexicalEnvironmentFlags["None"] = 0] = "None";
            LexicalEnvironmentFlags[LexicalEnvironmentFlags["InParameters"] = 1] = "InParameters";
            LexicalEnvironmentFlags[LexicalEnvironmentFlags["VariablesHoistedInParameters"] = 2] = "VariablesHoistedInParameters"; // a temp variable was hoisted while visiting a parameter list
        })(LexicalEnvironmentFlags = ts.LexicalEnvironmentFlags || (ts.LexicalEnvironmentFlags = {}));
        /*@internal*/
        var BundleFileSectionKind;
        (function (BundleFileSectionKind) {
            BundleFileSectionKind["Prologue"] = "prologue";
            BundleFileSectionKind["EmitHelpers"] = "emitHelpers";
            BundleFileSectionKind["NoDefaultLib"] = "no-default-lib";
            BundleFileSectionKind["Reference"] = "reference";
            BundleFileSectionKind["Type"] = "type";
            BundleFileSectionKind["Lib"] = "lib";
            BundleFileSectionKind["Prepend"] = "prepend";
            BundleFileSectionKind["Text"] = "text";
            BundleFileSectionKind["Internal"] = "internal";
            // comments?
        })(BundleFileSectionKind = ts.BundleFileSectionKind || (ts.BundleFileSectionKind = {}));
        var ListFormat;
        (function (ListFormat) {
            ListFormat[ListFormat["None"] = 0] = "None";
            // Line separators
            ListFormat[ListFormat["SingleLine"] = 0] = "SingleLine";
            ListFormat[ListFormat["MultiLine"] = 1] = "MultiLine";
            ListFormat[ListFormat["PreserveLines"] = 2] = "PreserveLines";
            ListFormat[ListFormat["LinesMask"] = 3] = "LinesMask";
            // Delimiters
            ListFormat[ListFormat["NotDelimited"] = 0] = "NotDelimited";
            ListFormat[ListFormat["BarDelimited"] = 4] = "BarDelimited";
            ListFormat[ListFormat["AmpersandDelimited"] = 8] = "AmpersandDelimited";
            ListFormat[ListFormat["CommaDelimited"] = 16] = "CommaDelimited";
            ListFormat[ListFormat["AsteriskDelimited"] = 32] = "AsteriskDelimited";
            ListFormat[ListFormat["DelimitersMask"] = 60] = "DelimitersMask";
            ListFormat[ListFormat["AllowTrailingComma"] = 64] = "AllowTrailingComma";
            // Whitespace
            ListFormat[ListFormat["Indented"] = 128] = "Indented";
            ListFormat[ListFormat["SpaceBetweenBraces"] = 256] = "SpaceBetweenBraces";
            ListFormat[ListFormat["SpaceBetweenSiblings"] = 512] = "SpaceBetweenSiblings";
            // Brackets/Braces
            ListFormat[ListFormat["Braces"] = 1024] = "Braces";
            ListFormat[ListFormat["Parenthesis"] = 2048] = "Parenthesis";
            ListFormat[ListFormat["AngleBrackets"] = 4096] = "AngleBrackets";
            ListFormat[ListFormat["SquareBrackets"] = 8192] = "SquareBrackets";
            ListFormat[ListFormat["BracketsMask"] = 15360] = "BracketsMask";
            ListFormat[ListFormat["OptionalIfUndefined"] = 16384] = "OptionalIfUndefined";
            ListFormat[ListFormat["OptionalIfEmpty"] = 32768] = "OptionalIfEmpty";
            ListFormat[ListFormat["Optional"] = 49152] = "Optional";
            // Other
            ListFormat[ListFormat["PreferNewLine"] = 65536] = "PreferNewLine";
            ListFormat[ListFormat["NoTrailingNewLine"] = 131072] = "NoTrailingNewLine";
            ListFormat[ListFormat["NoInterveningComments"] = 262144] = "NoInterveningComments";
            ListFormat[ListFormat["NoSpaceIfEmpty"] = 524288] = "NoSpaceIfEmpty";
            ListFormat[ListFormat["SingleElement"] = 1048576] = "SingleElement";
            ListFormat[ListFormat["SpaceAfterList"] = 2097152] = "SpaceAfterList";
            // Precomputed Formats
            ListFormat[ListFormat["Modifiers"] = 262656] = "Modifiers";
            ListFormat[ListFormat["HeritageClauses"] = 512] = "HeritageClauses";
            ListFormat[ListFormat["SingleLineTypeLiteralMembers"] = 768] = "SingleLineTypeLiteralMembers";
            ListFormat[ListFormat["MultiLineTypeLiteralMembers"] = 32897] = "MultiLineTypeLiteralMembers";
            ListFormat[ListFormat["SingleLineTupleTypeElements"] = 528] = "SingleLineTupleTypeElements";
            ListFormat[ListFormat["MultiLineTupleTypeElements"] = 657] = "MultiLineTupleTypeElements";
            ListFormat[ListFormat["UnionTypeConstituents"] = 516] = "UnionTypeConstituents";
            ListFormat[ListFormat["IntersectionTypeConstituents"] = 520] = "IntersectionTypeConstituents";
            ListFormat[ListFormat["ObjectBindingPatternElements"] = 525136] = "ObjectBindingPatternElements";
            ListFormat[ListFormat["ArrayBindingPatternElements"] = 524880] = "ArrayBindingPatternElements";
            ListFormat[ListFormat["ObjectLiteralExpressionProperties"] = 526226] = "ObjectLiteralExpressionProperties";
            ListFormat[ListFormat["ArrayLiteralExpressionElements"] = 8914] = "ArrayLiteralExpressionElements";
            ListFormat[ListFormat["CommaListElements"] = 528] = "CommaListElements";
            ListFormat[ListFormat["CallExpressionArguments"] = 2576] = "CallExpressionArguments";
            ListFormat[ListFormat["NewExpressionArguments"] = 18960] = "NewExpressionArguments";
            ListFormat[ListFormat["TemplateExpressionSpans"] = 262144] = "TemplateExpressionSpans";
            ListFormat[ListFormat["SingleLineBlockStatements"] = 768] = "SingleLineBlockStatements";
            ListFormat[ListFormat["MultiLineBlockStatements"] = 129] = "MultiLineBlockStatements";
            ListFormat[ListFormat["VariableDeclarationList"] = 528] = "VariableDeclarationList";
            ListFormat[ListFormat["SingleLineFunctionBodyStatements"] = 768] = "SingleLineFunctionBodyStatements";
            ListFormat[ListFormat["MultiLineFunctionBodyStatements"] = 1] = "MultiLineFunctionBodyStatements";
            ListFormat[ListFormat["ClassHeritageClauses"] = 0] = "ClassHeritageClauses";
            ListFormat[ListFormat["ClassMembers"] = 129] = "ClassMembers";
            ListFormat[ListFormat["InterfaceMembers"] = 129] = "InterfaceMembers";
            ListFormat[ListFormat["EnumMembers"] = 145] = "EnumMembers";
            ListFormat[ListFormat["CaseBlockClauses"] = 129] = "CaseBlockClauses";
            ListFormat[ListFormat["NamedImportsOrExportsElements"] = 525136] = "NamedImportsOrExportsElements";
            ListFormat[ListFormat["JsxElementOrFragmentChildren"] = 262144] = "JsxElementOrFragmentChildren";
            ListFormat[ListFormat["JsxElementAttributes"] = 262656] = "JsxElementAttributes";
            ListFormat[ListFormat["CaseOrDefaultClauseStatements"] = 163969] = "CaseOrDefaultClauseStatements";
            ListFormat[ListFormat["HeritageClauseTypes"] = 528] = "HeritageClauseTypes";
            ListFormat[ListFormat["SourceFileStatements"] = 131073] = "SourceFileStatements";
            ListFormat[ListFormat["Decorators"] = 2146305] = "Decorators";
            ListFormat[ListFormat["TypeArguments"] = 53776] = "TypeArguments";
            ListFormat[ListFormat["TypeParameters"] = 53776] = "TypeParameters";
            ListFormat[ListFormat["Parameters"] = 2576] = "Parameters";
            ListFormat[ListFormat["IndexSignatureParameters"] = 8848] = "IndexSignatureParameters";
            ListFormat[ListFormat["JSDocComment"] = 33] = "JSDocComment";
        })(ListFormat = ts.ListFormat || (ts.ListFormat = {}));
        /* @internal */
        var PragmaKindFlags;
        (function (PragmaKindFlags) {
            PragmaKindFlags[PragmaKindFlags["None"] = 0] = "None";
            /**
             * Triple slash comment of the form
             * /// <pragma-name argname="value" />
             */
            PragmaKindFlags[PragmaKindFlags["TripleSlashXML"] = 1] = "TripleSlashXML";
            /**
             * Single line comment of the form
             * // @pragma-name argval1 argval2
             * or
             * /// @pragma-name argval1 argval2
             */
            PragmaKindFlags[PragmaKindFlags["SingleLine"] = 2] = "SingleLine";
            /**
             * Multiline non-jsdoc pragma of the form
             * /* @pragma-name argval1 argval2 * /
             */
            PragmaKindFlags[PragmaKindFlags["MultiLine"] = 4] = "MultiLine";
            PragmaKindFlags[PragmaKindFlags["All"] = 7] = "All";
            PragmaKindFlags[PragmaKindFlags["Default"] = 7] = "Default";
        })(PragmaKindFlags = ts.PragmaKindFlags || (ts.PragmaKindFlags = {}));
        // While not strictly a type, this is here because `PragmaMap` needs to be here to be used with `SourceFile`, and we don't
        //  fancy effectively defining it twice, once in value-space and once in type-space
        /* @internal */
        ts.commentPragmas = {
            "reference": {
                args: [
                    { name: "types", optional: true, captureSpan: true },
                    { name: "lib", optional: true, captureSpan: true },
                    { name: "path", optional: true, captureSpan: true },
                    { name: "no-default-lib", optional: true }
                ],
                kind: 1 /* TripleSlashXML */
            },
            "amd-dependency": {
                args: [{ name: "path" }, { name: "name", optional: true }],
                kind: 1 /* TripleSlashXML */
            },
            "amd-module": {
                args: [{ name: "name" }],
                kind: 1 /* TripleSlashXML */
            },
            "ts-check": {
                kind: 2 /* SingleLine */
            },
            "ts-nocheck": {
                kind: 2 /* SingleLine */
            },
            "jsx": {
                args: [{ name: "factory" }],
                kind: 4 /* MultiLine */
            },
            "jsxfrag": {
                args: [{ name: "factory" }],
                kind: 4 /* MultiLine */
            },
            "jsximportsource": {
                args: [{ name: "factory" }],
                kind: 4 /* MultiLine */
            },
            "jsxruntime": {
                args: [{ name: "factory" }],
                kind: 4 /* MultiLine */
            },
        };
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        /**
         * Internally, we represent paths as strings with '/' as the directory separator.
         * When we make system calls (eg: LanguageServiceHost.getDirectory()),
         * we expect the host to correctly handle paths in our specified format.
         */
        ts.directorySeparator = "/";
        ts.altDirectorySeparator = "\\";
        var urlSchemeSeparator = "://";
        var backslashRegExp = /\\/g;
        //// Path Tests
        /**
         * Determines whether a charCode corresponds to `/` or `\`.
         */
        function isAnyDirectorySeparator(charCode) {
            return charCode === 47 /* slash */ || charCode === 92 /* backslash */;
        }
        ts.isAnyDirectorySeparator = isAnyDirectorySeparator;
        /**
         * Determines whether a path starts with a URL scheme (e.g. starts with `http://`, `ftp://`, `file://`, etc.).
         */
        function isUrl(path) {
            return getEncodedRootLength(path) < 0;
        }
        ts.isUrl = isUrl;
        /**
         * Determines whether a path is an absolute disk path (e.g. starts with `/`, or a dos path
         * like `c:`, `c:\` or `c:/`).
         */
        function isRootedDiskPath(path) {
            return getEncodedRootLength(path) > 0;
        }
        ts.isRootedDiskPath = isRootedDiskPath;
        /**
         * Determines whether a path consists only of a path root.
         */
        function isDiskPathRoot(path) {
            var rootLength = getEncodedRootLength(path);
            return rootLength > 0 && rootLength === path.length;
        }
        ts.isDiskPathRoot = isDiskPathRoot;
        /**
         * Determines whether a path starts with an absolute path component (i.e. `/`, `c:/`, `file://`, etc.).
         *
         * ```ts
         * // POSIX
         * pathIsAbsolute("/path/to/file.ext") === true
         * // DOS
         * pathIsAbsolute("c:/path/to/file.ext") === true
         * // URL
         * pathIsAbsolute("file:///path/to/file.ext") === true
         * // Non-absolute
         * pathIsAbsolute("path/to/file.ext") === false
         * pathIsAbsolute("./path/to/file.ext") === false
         * ```
         */
        function pathIsAbsolute(path) {
            return getEncodedRootLength(path) !== 0;
        }
        ts.pathIsAbsolute = pathIsAbsolute;
        /**
         * Determines whether a path starts with a relative path component (i.e. `.` or `..`).
         */
        function pathIsRelative(path) {
            return /^\.\.?($|[\\/])/.test(path);
        }
        ts.pathIsRelative = pathIsRelative;
        /**
         * Determines whether a path is neither relative nor absolute, e.g. "path/to/file".
         * Also known misleadingly as "non-relative".
         */
        function pathIsBareSpecifier(path) {
            return !pathIsAbsolute(path) && !pathIsRelative(path);
        }
        ts.pathIsBareSpecifier = pathIsBareSpecifier;
        function hasExtension(fileName) {
            return ts.stringContains(getBaseFileName(fileName), ".");
        }
        ts.hasExtension = hasExtension;
        function fileExtensionIs(path, extension) {
            return path.length > extension.length && ts.endsWith(path, extension);
        }
        ts.fileExtensionIs = fileExtensionIs;
        function fileExtensionIsOneOf(path, extensions) {
            for (var _i = 0, extensions_1 = extensions; _i < extensions_1.length; _i++) {
                var extension = extensions_1[_i];
                if (fileExtensionIs(path, extension)) {
                    return true;
                }
            }
            return false;
        }
        ts.fileExtensionIsOneOf = fileExtensionIsOneOf;
        /**
         * Determines whether a path has a trailing separator (`/` or `\\`).
         */
        function hasTrailingDirectorySeparator(path) {
            return path.length > 0 && isAnyDirectorySeparator(path.charCodeAt(path.length - 1));
        }
        ts.hasTrailingDirectorySeparator = hasTrailingDirectorySeparator;
        //// Path Parsing
        function isVolumeCharacter(charCode) {
            return (charCode >= 97 /* a */ && charCode <= 122 /* z */) ||
                (charCode >= 65 /* A */ && charCode <= 90 /* Z */);
        }
        function getFileUrlVolumeSeparatorEnd(url, start) {
            var ch0 = url.charCodeAt(start);
            if (ch0 === 58 /* colon */)
                return start + 1;
            if (ch0 === 37 /* percent */ && url.charCodeAt(start + 1) === 51 /* _3 */) {
                var ch2 = url.charCodeAt(start + 2);
                if (ch2 === 97 /* a */ || ch2 === 65 /* A */)
                    return start + 3;
            }
            return -1;
        }
        /**
         * Returns length of the root part of a path or URL (i.e. length of "/", "x:/", "//server/share/, file:///user/files").
         * If the root is part of a URL, the twos-complement of the root length is returned.
         */
        function getEncodedRootLength(path) {
            if (!path)
                return 0;
            var ch0 = path.charCodeAt(0);
            // POSIX or UNC
            if (ch0 === 47 /* slash */ || ch0 === 92 /* backslash */) {
                if (path.charCodeAt(1) !== ch0)
                    return 1; // POSIX: "/" (or non-normalized "\")
                var p1 = path.indexOf(ch0 === 47 /* slash */ ? ts.directorySeparator : ts.altDirectorySeparator, 2);
                if (p1 < 0)
                    return path.length; // UNC: "//server" or "\\server"
                return p1 + 1; // UNC: "//server/" or "\\server\"
            }
            // DOS
            if (isVolumeCharacter(ch0) && path.charCodeAt(1) === 58 /* colon */) {
                var ch2 = path.charCodeAt(2);
                if (ch2 === 47 /* slash */ || ch2 === 92 /* backslash */)
                    return 3; // DOS: "c:/" or "c:\"
                if (path.length === 2)
                    return 2; // DOS: "c:" (but not "c:d")
            }
            // URL
            var schemeEnd = path.indexOf(urlSchemeSeparator);
            if (schemeEnd !== -1) {
                var authorityStart = schemeEnd + urlSchemeSeparator.length;
                var authorityEnd = path.indexOf(ts.directorySeparator, authorityStart);
                if (authorityEnd !== -1) { // URL: "file:///", "file://server/", "file://server/path"
                    // For local "file" URLs, include the leading DOS volume (if present).
                    // Per https://www.ietf.org/rfc/rfc1738.txt, a host of "" or "localhost" is a
                    // special case interpreted as "the machine from which the URL is being interpreted".
                    var scheme = path.slice(0, schemeEnd);
                    var authority = path.slice(authorityStart, authorityEnd);
                    if (scheme === "file" && (authority === "" || authority === "localhost") &&
                        isVolumeCharacter(path.charCodeAt(authorityEnd + 1))) {
                        var volumeSeparatorEnd = getFileUrlVolumeSeparatorEnd(path, authorityEnd + 2);
                        if (volumeSeparatorEnd !== -1) {
                            if (path.charCodeAt(volumeSeparatorEnd) === 47 /* slash */) {
                                // URL: "file:///c:/", "file://localhost/c:/", "file:///c%3a/", "file://localhost/c%3a/"
                                return ~(volumeSeparatorEnd + 1);
                            }
                            if (volumeSeparatorEnd === path.length) {
                                // URL: "file:///c:", "file://localhost/c:", "file:///c$3a", "file://localhost/c%3a"
                                // but not "file:///c:d" or "file:///c%3ad"
                                return ~volumeSeparatorEnd;
                            }
                        }
                    }
                    return ~(authorityEnd + 1); // URL: "file://server/", "http://server/"
                }
                return ~path.length; // URL: "file://server", "http://server"
            }
            // relative
            return 0;
        }
        /**
         * Returns length of the root part of a path or URL (i.e. length of "/", "x:/", "//server/share/, file:///user/files").
         *
         * For example:
         * ```ts
         * getRootLength("a") === 0                   // ""
         * getRootLength("/") === 1                   // "/"
         * getRootLength("c:") === 2                  // "c:"
         * getRootLength("c:d") === 0                 // ""
         * getRootLength("c:/") === 3                 // "c:/"
         * getRootLength("c:\\") === 3                // "c:\\"
         * getRootLength("//server") === 7            // "//server"
         * getRootLength("//server/share") === 8      // "//server/"
         * getRootLength("\\\\server") === 7          // "\\\\server"
         * getRootLength("\\\\server\\share") === 8   // "\\\\server\\"
         * getRootLength("file:///path") === 8        // "file:///"
         * getRootLength("file:///c:") === 10         // "file:///c:"
         * getRootLength("file:///c:d") === 8         // "file:///"
         * getRootLength("file:///c:/path") === 11    // "file:///c:/"
         * getRootLength("file://server") === 13      // "file://server"
         * getRootLength("file://server/path") === 14 // "file://server/"
         * getRootLength("http://server") === 13      // "http://server"
         * getRootLength("http://server/path") === 14 // "http://server/"
         * ```
         */
        function getRootLength(path) {
            var rootLength = getEncodedRootLength(path);
            return rootLength < 0 ? ~rootLength : rootLength;
        }
        ts.getRootLength = getRootLength;
        function getDirectoryPath(path) {
            path = normalizeSlashes(path);
            // If the path provided is itself the root, then return it.
            var rootLength = getRootLength(path);
            if (rootLength === path.length)
                return path;
            // return the leading portion of the path up to the last (non-terminal) directory separator
            // but not including any trailing directory separator.
            path = removeTrailingDirectorySeparator(path);
            return path.slice(0, Math.max(rootLength, path.lastIndexOf(ts.directorySeparator)));
        }
        ts.getDirectoryPath = getDirectoryPath;
        function getBaseFileName(path, extensions, ignoreCase) {
            path = normalizeSlashes(path);
            // if the path provided is itself the root, then it has not file name.
            var rootLength = getRootLength(path);
            if (rootLength === path.length)
                return "";
            // return the trailing portion of the path starting after the last (non-terminal) directory
            // separator but not including any trailing directory separator.
            path = removeTrailingDirectorySeparator(path);
            var name = path.slice(Math.max(getRootLength(path), path.lastIndexOf(ts.directorySeparator) + 1));
            var extension = extensions !== undefined && ignoreCase !== undefined ? getAnyExtensionFromPath(name, extensions, ignoreCase) : undefined;
            return extension ? name.slice(0, name.length - extension.length) : name;
        }
        ts.getBaseFileName = getBaseFileName;
        function tryGetExtensionFromPath(path, extension, stringEqualityComparer) {
            if (!ts.startsWith(extension, "."))
                extension = "." + extension;
            if (path.length >= extension.length && path.charCodeAt(path.length - extension.length) === 46 /* dot */) {
                var pathExtension = path.slice(path.length - extension.length);
                if (stringEqualityComparer(pathExtension, extension)) {
                    return pathExtension;
                }
            }
        }
        function getAnyExtensionFromPathWorker(path, extensions, stringEqualityComparer) {
            if (typeof extensions === "string") {
                return tryGetExtensionFromPath(path, extensions, stringEqualityComparer) || "";
            }
            for (var _i = 0, extensions_2 = extensions; _i < extensions_2.length; _i++) {
                var extension = extensions_2[_i];
                var result = tryGetExtensionFromPath(path, extension, stringEqualityComparer);
                if (result)
                    return result;
            }
            return "";
        }
        function getAnyExtensionFromPath(path, extensions, ignoreCase) {
            // Retrieves any string from the final "." onwards from a base file name.
            // Unlike extensionFromPath, which throws an exception on unrecognized extensions.
            if (extensions) {
                return getAnyExtensionFromPathWorker(removeTrailingDirectorySeparator(path), extensions, ignoreCase ? ts.equateStringsCaseInsensitive : ts.equateStringsCaseSensitive);
            }
            var baseFileName = getBaseFileName(path);
            var extensionIndex = baseFileName.lastIndexOf(".");
            if (extensionIndex >= 0) {
                return baseFileName.substring(extensionIndex);
            }
            return "";
        }
        ts.getAnyExtensionFromPath = getAnyExtensionFromPath;
        function pathComponents(path, rootLength) {
            var root = path.substring(0, rootLength);
            var rest = path.substring(rootLength).split(ts.directorySeparator);
            if (rest.length && !ts.lastOrUndefined(rest))
                rest.pop();
            return __spreadArray([root], rest, true);
        }
        /**
         * Parse a path into an array containing a root component (at index 0) and zero or more path
         * components (at indices > 0). The result is not normalized.
         * If the path is relative, the root component is `""`.
         * If the path is absolute, the root component includes the first path separator (`/`).
         *
         * ```ts
         * // POSIX
         * getPathComponents("/path/to/file.ext") === ["/", "path", "to", "file.ext"]
         * getPathComponents("/path/to/") === ["/", "path", "to"]
         * getPathComponents("/") === ["/"]
         * // DOS
         * getPathComponents("c:/path/to/file.ext") === ["c:/", "path", "to", "file.ext"]
         * getPathComponents("c:/path/to/") === ["c:/", "path", "to"]
         * getPathComponents("c:/") === ["c:/"]
         * getPathComponents("c:") === ["c:"]
         * // URL
         * getPathComponents("http://typescriptlang.org/path/to/file.ext") === ["http://typescriptlang.org/", "path", "to", "file.ext"]
         * getPathComponents("http://typescriptlang.org/path/to/") === ["http://typescriptlang.org/", "path", "to"]
         * getPathComponents("http://typescriptlang.org/") === ["http://typescriptlang.org/"]
         * getPathComponents("http://typescriptlang.org") === ["http://typescriptlang.org"]
         * getPathComponents("file://server/path/to/file.ext") === ["file://server/", "path", "to", "file.ext"]
         * getPathComponents("file://server/path/to/") === ["file://server/", "path", "to"]
         * getPathComponents("file://server/") === ["file://server/"]
         * getPathComponents("file://server") === ["file://server"]
         * getPathComponents("file:///path/to/file.ext") === ["file:///", "path", "to", "file.ext"]
         * getPathComponents("file:///path/to/") === ["file:///", "path", "to"]
         * getPathComponents("file:///") === ["file:///"]
         * getPathComponents("file://") === ["file://"]
         */
        function getPathComponents(path, currentDirectory) {
            if (currentDirectory === void 0) {
                currentDirectory = "";
            }
            path = combinePaths(currentDirectory, path);
            return pathComponents(path, getRootLength(path));
        }
        ts.getPathComponents = getPathComponents;
        //// Path Formatting
        /**
         * Formats a parsed path consisting of a root component (at index 0) and zero or more path
         * segments (at indices > 0).
         *
         * ```ts
         * getPathFromPathComponents(["/", "path", "to", "file.ext"]) === "/path/to/file.ext"
         * ```
         */
        function getPathFromPathComponents(pathComponents) {
            if (pathComponents.length === 0)
                return "";
            var root = pathComponents[0] && ensureTrailingDirectorySeparator(pathComponents[0]);
            return root + pathComponents.slice(1).join(ts.directorySeparator);
        }
        ts.getPathFromPathComponents = getPathFromPathComponents;
        //// Path Normalization
        /**
         * Normalize path separators, converting `\` into `/`.
         */
        function normalizeSlashes(path) {
            var index = path.indexOf("\\");
            if (index === -1) {
                return path;
            }
            backslashRegExp.lastIndex = index; // prime regex with known position
            return path.replace(backslashRegExp, ts.directorySeparator);
        }
        ts.normalizeSlashes = normalizeSlashes;
        /**
         * Reduce an array of path components to a more simplified path by navigating any
         * `"."` or `".."` entries in the path.
         */
        function reducePathComponents(components) {
            if (!ts.some(components))
                return [];
            var reduced = [components[0]];
            for (var i = 1; i < components.length; i++) {
                var component = components[i];
                if (!component)
                    continue;
                if (component === ".")
                    continue;
                if (component === "..") {
                    if (reduced.length > 1) {
                        if (reduced[reduced.length - 1] !== "..") {
                            reduced.pop();
                            continue;
                        }
                    }
                    else if (reduced[0])
                        continue;
                }
                reduced.push(component);
            }
            return reduced;
        }
        ts.reducePathComponents = reducePathComponents;
        /**
         * Combines paths. If a path is absolute, it replaces any previous path. Relative paths are not simplified.
         *
         * ```ts
         * // Non-rooted
         * combinePaths("path", "to", "file.ext") === "path/to/file.ext"
         * combinePaths("path", "dir", "..", "to", "file.ext") === "path/dir/../to/file.ext"
         * // POSIX
         * combinePaths("/path", "to", "file.ext") === "/path/to/file.ext"
         * combinePaths("/path", "/to", "file.ext") === "/to/file.ext"
         * // DOS
         * combinePaths("c:/path", "to", "file.ext") === "c:/path/to/file.ext"
         * combinePaths("c:/path", "c:/to", "file.ext") === "c:/to/file.ext"
         * // URL
         * combinePaths("file:///path", "to", "file.ext") === "file:///path/to/file.ext"
         * combinePaths("file:///path", "file:///to", "file.ext") === "file:///to/file.ext"
         * ```
         */
        function combinePaths(path) {
            var paths = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                paths[_i - 1] = arguments[_i];
            }
            if (path)
                path = normalizeSlashes(path);
            for (var _a = 0, paths_1 = paths; _a < paths_1.length; _a++) {
                var relativePath = paths_1[_a];
                if (!relativePath)
                    continue;
                relativePath = normalizeSlashes(relativePath);
                if (!path || getRootLength(relativePath) !== 0) {
                    path = relativePath;
                }
                else {
                    path = ensureTrailingDirectorySeparator(path) + relativePath;
                }
            }
            return path;
        }
        ts.combinePaths = combinePaths;
        /**
         * Combines and resolves paths. If a path is absolute, it replaces any previous path. Any
         * `.` and `..` path components are resolved. Trailing directory separators are preserved.
         *
         * ```ts
         * resolvePath("/path", "to", "file.ext") === "path/to/file.ext"
         * resolvePath("/path", "to", "file.ext/") === "path/to/file.ext/"
         * resolvePath("/path", "dir", "..", "to", "file.ext") === "path/to/file.ext"
         * ```
         */
        function resolvePath(path) {
            var paths = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                paths[_i - 1] = arguments[_i];
            }
            return normalizePath(ts.some(paths) ? combinePaths.apply(void 0, __spreadArray([path], paths, false)) : normalizeSlashes(path));
        }
        ts.resolvePath = resolvePath;
        /**
         * Parse a path into an array containing a root component (at index 0) and zero or more path
         * components (at indices > 0). The result is normalized.
         * If the path is relative, the root component is `""`.
         * If the path is absolute, the root component includes the first path separator (`/`).
         *
         * ```ts
         * getNormalizedPathComponents("to/dir/../file.ext", "/path/") === ["/", "path", "to", "file.ext"]
         * ```
         */
        function getNormalizedPathComponents(path, currentDirectory) {
            return reducePathComponents(getPathComponents(path, currentDirectory));
        }
        ts.getNormalizedPathComponents = getNormalizedPathComponents;
        function getNormalizedAbsolutePath(fileName, currentDirectory) {
            return getPathFromPathComponents(getNormalizedPathComponents(fileName, currentDirectory));
        }
        ts.getNormalizedAbsolutePath = getNormalizedAbsolutePath;
        function normalizePath(path) {
            path = normalizeSlashes(path);
            // Most paths don't require normalization
            if (!relativePathSegmentRegExp.test(path)) {
                return path;
            }
            // Some paths only require cleanup of `/./` or leading `./`
            var simplified = path.replace(/\/\.\//g, "/").replace(/^\.\//, "");
            if (simplified !== path) {
                path = simplified;
                if (!relativePathSegmentRegExp.test(path)) {
                    return path;
                }
            }
            // Other paths require full normalization
            var normalized = getPathFromPathComponents(reducePathComponents(getPathComponents(path)));
            return normalized && hasTrailingDirectorySeparator(path) ? ensureTrailingDirectorySeparator(normalized) : normalized;
        }
        ts.normalizePath = normalizePath;
        function getPathWithoutRoot(pathComponents) {
            if (pathComponents.length === 0)
                return "";
            return pathComponents.slice(1).join(ts.directorySeparator);
        }
        function getNormalizedAbsolutePathWithoutRoot(fileName, currentDirectory) {
            return getPathWithoutRoot(getNormalizedPathComponents(fileName, currentDirectory));
        }
        ts.getNormalizedAbsolutePathWithoutRoot = getNormalizedAbsolutePathWithoutRoot;
        function toPath(fileName, basePath, getCanonicalFileName) {
            var nonCanonicalizedPath = isRootedDiskPath(fileName)
                ? normalizePath(fileName)
                : getNormalizedAbsolutePath(fileName, basePath);
            return getCanonicalFileName(nonCanonicalizedPath);
        }
        ts.toPath = toPath;
        function normalizePathAndParts(path) {
            path = normalizeSlashes(path);
            var _a = reducePathComponents(getPathComponents(path)), root = _a[0], parts = _a.slice(1);
            if (parts.length) {
                var joinedParts = root + parts.join(ts.directorySeparator);
                return { path: hasTrailingDirectorySeparator(path) ? ensureTrailingDirectorySeparator(joinedParts) : joinedParts, parts: parts };
            }
            else {
                return { path: root, parts: parts };
            }
        }
        ts.normalizePathAndParts = normalizePathAndParts;
        function removeTrailingDirectorySeparator(path) {
            if (hasTrailingDirectorySeparator(path)) {
                return path.substr(0, path.length - 1);
            }
            return path;
        }
        ts.removeTrailingDirectorySeparator = removeTrailingDirectorySeparator;
        function ensureTrailingDirectorySeparator(path) {
            if (!hasTrailingDirectorySeparator(path)) {
                return path + ts.directorySeparator;
            }
            return path;
        }
        ts.ensureTrailingDirectorySeparator = ensureTrailingDirectorySeparator;
        /**
         * Ensures a path is either absolute (prefixed with `/` or `c:`) or dot-relative (prefixed
         * with `./` or `../`) so as not to be confused with an unprefixed module name.
         *
         * ```ts
         * ensurePathIsNonModuleName("/path/to/file.ext") === "/path/to/file.ext"
         * ensurePathIsNonModuleName("./path/to/file.ext") === "./path/to/file.ext"
         * ensurePathIsNonModuleName("../path/to/file.ext") === "../path/to/file.ext"
         * ensurePathIsNonModuleName("path/to/file.ext") === "./path/to/file.ext"
         * ```
         */
        function ensurePathIsNonModuleName(path) {
            return !pathIsAbsolute(path) && !pathIsRelative(path) ? "./" + path : path;
        }
        ts.ensurePathIsNonModuleName = ensurePathIsNonModuleName;
        function changeAnyExtension(path, ext, extensions, ignoreCase) {
            var pathext = extensions !== undefined && ignoreCase !== undefined ? getAnyExtensionFromPath(path, extensions, ignoreCase) : getAnyExtensionFromPath(path);
            return pathext ? path.slice(0, path.length - pathext.length) + (ts.startsWith(ext, ".") ? ext : "." + ext) : path;
        }
        ts.changeAnyExtension = changeAnyExtension;
        //// Path Comparisons
        // check path for these segments: '', '.'. '..'
        var relativePathSegmentRegExp = /(?:\/\/)|(?:^|\/)\.\.?(?:$|\/)/;
        function comparePathsWorker(a, b, componentComparer) {
            if (a === b)
                return 0 /* EqualTo */;
            if (a === undefined)
                return -1 /* LessThan */;
            if (b === undefined)
                return 1 /* GreaterThan */;
            // NOTE: Performance optimization - shortcut if the root segments differ as there would be no
            //       need to perform path reduction.
            var aRoot = a.substring(0, getRootLength(a));
            var bRoot = b.substring(0, getRootLength(b));
            var result = ts.compareStringsCaseInsensitive(aRoot, bRoot);
            if (result !== 0 /* EqualTo */) {
                return result;
            }
            // NOTE: Performance optimization - shortcut if there are no relative path segments in
            //       the non-root portion of the path
            var aRest = a.substring(aRoot.length);
            var bRest = b.substring(bRoot.length);
            if (!relativePathSegmentRegExp.test(aRest) && !relativePathSegmentRegExp.test(bRest)) {
                return componentComparer(aRest, bRest);
            }
            // The path contains a relative path segment. Normalize the paths and perform a slower component
            // by component comparison.
            var aComponents = reducePathComponents(getPathComponents(a));
            var bComponents = reducePathComponents(getPathComponents(b));
            var sharedLength = Math.min(aComponents.length, bComponents.length);
            for (var i = 1; i < sharedLength; i++) {
                var result_2 = componentComparer(aComponents[i], bComponents[i]);
                if (result_2 !== 0 /* EqualTo */) {
                    return result_2;
                }
            }
            return ts.compareValues(aComponents.length, bComponents.length);
        }
        /**
         * Performs a case-sensitive comparison of two paths. Path roots are always compared case-insensitively.
         */
        function comparePathsCaseSensitive(a, b) {
            return comparePathsWorker(a, b, ts.compareStringsCaseSensitive);
        }
        ts.comparePathsCaseSensitive = comparePathsCaseSensitive;
        /**
         * Performs a case-insensitive comparison of two paths.
         */
        function comparePathsCaseInsensitive(a, b) {
            return comparePathsWorker(a, b, ts.compareStringsCaseInsensitive);
        }
        ts.comparePathsCaseInsensitive = comparePathsCaseInsensitive;
        function comparePaths(a, b, currentDirectory, ignoreCase) {
            if (typeof currentDirectory === "string") {
                a = combinePaths(currentDirectory, a);
                b = combinePaths(currentDirectory, b);
            }
            else if (typeof currentDirectory === "boolean") {
                ignoreCase = currentDirectory;
            }
            return comparePathsWorker(a, b, ts.getStringComparer(ignoreCase));
        }
        ts.comparePaths = comparePaths;
        function containsPath(parent, child, currentDirectory, ignoreCase) {
            if (typeof currentDirectory === "string") {
                parent = combinePaths(currentDirectory, parent);
                child = combinePaths(currentDirectory, child);
            }
            else if (typeof currentDirectory === "boolean") {
                ignoreCase = currentDirectory;
            }
            if (parent === undefined || child === undefined)
                return false;
            if (parent === child)
                return true;
            var parentComponents = reducePathComponents(getPathComponents(parent));
            var childComponents = reducePathComponents(getPathComponents(child));
            if (childComponents.length < parentComponents.length) {
                return false;
            }
            var componentEqualityComparer = ignoreCase ? ts.equateStringsCaseInsensitive : ts.equateStringsCaseSensitive;
            for (var i = 0; i < parentComponents.length; i++) {
                var equalityComparer = i === 0 ? ts.equateStringsCaseInsensitive : componentEqualityComparer;
                if (!equalityComparer(parentComponents[i], childComponents[i])) {
                    return false;
                }
            }
            return true;
        }
        ts.containsPath = containsPath;
        /**
         * Determines whether `fileName` starts with the specified `directoryName` using the provided path canonicalization callback.
         * Comparison is case-sensitive between the canonical paths.
         *
         * Use `containsPath` if file names are not already reduced and absolute.
         */
        function startsWithDirectory(fileName, directoryName, getCanonicalFileName) {
            var canonicalFileName = getCanonicalFileName(fileName);
            var canonicalDirectoryName = getCanonicalFileName(directoryName);
            return ts.startsWith(canonicalFileName, canonicalDirectoryName + "/") || ts.startsWith(canonicalFileName, canonicalDirectoryName + "\\");
        }
        ts.startsWithDirectory = startsWithDirectory;
        //// Relative Paths
        function getPathComponentsRelativeTo(from, to, stringEqualityComparer, getCanonicalFileName) {
            var fromComponents = reducePathComponents(getPathComponents(from));
            var toComponents = reducePathComponents(getPathComponents(to));
            var start;
            for (start = 0; start < fromComponents.length && start < toComponents.length; start++) {
                var fromComponent = getCanonicalFileName(fromComponents[start]);
                var toComponent = getCanonicalFileName(toComponents[start]);
                var comparer = start === 0 ? ts.equateStringsCaseInsensitive : stringEqualityComparer;
                if (!comparer(fromComponent, toComponent))
                    break;
            }
            if (start === 0) {
                return toComponents;
            }
            var components = toComponents.slice(start);
            var relative = [];
            for (; start < fromComponents.length; start++) {
                relative.push("..");
            }
            return __spreadArray(__spreadArray([""], relative, true), components, true);
        }
        ts.getPathComponentsRelativeTo = getPathComponentsRelativeTo;
        function getRelativePathFromDirectory(fromDirectory, to, getCanonicalFileNameOrIgnoreCase) {
            ts.Debug.assert((getRootLength(fromDirectory) > 0) === (getRootLength(to) > 0), "Paths must either both be absolute or both be relative");
            var getCanonicalFileName = typeof getCanonicalFileNameOrIgnoreCase === "function" ? getCanonicalFileNameOrIgnoreCase : ts.identity;
            var ignoreCase = typeof getCanonicalFileNameOrIgnoreCase === "boolean" ? getCanonicalFileNameOrIgnoreCase : false;
            var pathComponents = getPathComponentsRelativeTo(fromDirectory, to, ignoreCase ? ts.equateStringsCaseInsensitive : ts.equateStringsCaseSensitive, getCanonicalFileName);
            return getPathFromPathComponents(pathComponents);
        }
        ts.getRelativePathFromDirectory = getRelativePathFromDirectory;
        function convertToRelativePath(absoluteOrRelativePath, basePath, getCanonicalFileName) {
            return !isRootedDiskPath(absoluteOrRelativePath)
                ? absoluteOrRelativePath
                : getRelativePathToDirectoryOrUrl(basePath, absoluteOrRelativePath, basePath, getCanonicalFileName, /*isAbsolutePathAnUrl*/ false);
        }
        ts.convertToRelativePath = convertToRelativePath;
        function getRelativePathFromFile(from, to, getCanonicalFileName) {
            return ensurePathIsNonModuleName(getRelativePathFromDirectory(getDirectoryPath(from), to, getCanonicalFileName));
        }
        ts.getRelativePathFromFile = getRelativePathFromFile;
        function getRelativePathToDirectoryOrUrl(directoryPathOrUrl, relativeOrAbsolutePath, currentDirectory, getCanonicalFileName, isAbsolutePathAnUrl) {
            var pathComponents = getPathComponentsRelativeTo(resolvePath(currentDirectory, directoryPathOrUrl), resolvePath(currentDirectory, relativeOrAbsolutePath), ts.equateStringsCaseSensitive, getCanonicalFileName);
            var firstComponent = pathComponents[0];
            if (isAbsolutePathAnUrl && isRootedDiskPath(firstComponent)) {
                var prefix = firstComponent.charAt(0) === ts.directorySeparator ? "file://" : "file:///";
                pathComponents[0] = prefix + firstComponent;
            }
            return getPathFromPathComponents(pathComponents);
        }
        ts.getRelativePathToDirectoryOrUrl = getRelativePathToDirectoryOrUrl;
        function forEachAncestorDirectory(directory, callback) {
            while (true) {
                var result = callback(directory);
                if (result !== undefined) {
                    return result;
                }
                var parentPath = getDirectoryPath(directory);
                if (parentPath === directory) {
                    return undefined;
                }
                directory = parentPath;
            }
        }
        ts.forEachAncestorDirectory = forEachAncestorDirectory;
        function isNodeModulesDirectory(dirPath) {
            return ts.endsWith(dirPath, "/node_modules");
        }
        ts.isNodeModulesDirectory = isNodeModulesDirectory;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        /**
         * djb2 hashing algorithm
         * http://www.cse.yorku.ca/~oz/hash.html
         */
        /* @internal */
        function generateDjb2Hash(data) {
            var acc = 5381;
            for (var i = 0; i < data.length; i++) {
                acc = ((acc << 5) + acc) + data.charCodeAt(i);
            }
            return acc.toString();
        }
        ts.generateDjb2Hash = generateDjb2Hash;
        /**
         * Set a high stack trace limit to provide more information in case of an error.
         * Called for command-line and server use cases.
         * Not called if TypeScript is used as a library.
         */
        /* @internal */
        function setStackTraceLimit() {
            if (Error.stackTraceLimit < 100) { // Also tests that we won't set the property if it doesn't exist.
                Error.stackTraceLimit = 100;
            }
        }
        ts.setStackTraceLimit = setStackTraceLimit;
        var FileWatcherEventKind;
        (function (FileWatcherEventKind) {
            FileWatcherEventKind[FileWatcherEventKind["Created"] = 0] = "Created";
            FileWatcherEventKind[FileWatcherEventKind["Changed"] = 1] = "Changed";
            FileWatcherEventKind[FileWatcherEventKind["Deleted"] = 2] = "Deleted";
        })(FileWatcherEventKind = ts.FileWatcherEventKind || (ts.FileWatcherEventKind = {}));
        /* @internal */
        var PollingInterval;
        (function (PollingInterval) {
            PollingInterval[PollingInterval["High"] = 2000] = "High";
            PollingInterval[PollingInterval["Medium"] = 500] = "Medium";
            PollingInterval[PollingInterval["Low"] = 250] = "Low";
        })(PollingInterval = ts.PollingInterval || (ts.PollingInterval = {}));
        /* @internal */
        ts.missingFileModifiedTime = new Date(0); // Any subsequent modification will occur after this time
        /* @internal */
        function getModifiedTime(host, fileName) {
            return host.getModifiedTime(fileName) || ts.missingFileModifiedTime;
        }
        ts.getModifiedTime = getModifiedTime;
        function createPollingIntervalBasedLevels(levels) {
            var _a;
            return _a = {},
                _a[PollingInterval.Low] = levels.Low,
                _a[PollingInterval.Medium] = levels.Medium,
                _a[PollingInterval.High] = levels.High,
                _a;
        }
        var defaultChunkLevels = { Low: 32, Medium: 64, High: 256 };
        var pollingChunkSize = createPollingIntervalBasedLevels(defaultChunkLevels);
        /* @internal */
        ts.unchangedPollThresholds = createPollingIntervalBasedLevels(defaultChunkLevels);
        /* @internal */
        function setCustomPollingValues(system) {
            if (!system.getEnvironmentVariable) {
                return;
            }
            var pollingIntervalChanged = setCustomLevels("TSC_WATCH_POLLINGINTERVAL", PollingInterval);
            pollingChunkSize = getCustomPollingBasedLevels("TSC_WATCH_POLLINGCHUNKSIZE", defaultChunkLevels) || pollingChunkSize;
            ts.unchangedPollThresholds = getCustomPollingBasedLevels("TSC_WATCH_UNCHANGEDPOLLTHRESHOLDS", defaultChunkLevels) || ts.unchangedPollThresholds;
            function getLevel(envVar, level) {
                return system.getEnvironmentVariable(envVar + "_" + level.toUpperCase());
            }
            function getCustomLevels(baseVariable) {
                var customLevels;
                setCustomLevel("Low");
                setCustomLevel("Medium");
                setCustomLevel("High");
                return customLevels;
                function setCustomLevel(level) {
                    var customLevel = getLevel(baseVariable, level);
                    if (customLevel) {
                        (customLevels || (customLevels = {}))[level] = Number(customLevel);
                    }
                }
            }
            function setCustomLevels(baseVariable, levels) {
                var customLevels = getCustomLevels(baseVariable);
                if (customLevels) {
                    setLevel("Low");
                    setLevel("Medium");
                    setLevel("High");
                    return true;
                }
                return false;
                function setLevel(level) {
                    levels[level] = customLevels[level] || levels[level];
                }
            }
            function getCustomPollingBasedLevels(baseVariable, defaultLevels) {
                var customLevels = getCustomLevels(baseVariable);
                return (pollingIntervalChanged || customLevels) &&
                    createPollingIntervalBasedLevels(customLevels ? __assign(__assign({}, defaultLevels), customLevels) : defaultLevels);
            }
        }
        ts.setCustomPollingValues = setCustomPollingValues;
        function pollWatchedFileQueue(host, queue, pollIndex, chunkSize, callbackOnWatchFileStat) {
            var definedValueCopyToIndex = pollIndex;
            // Max visit would be all elements of the queue
            for (var canVisit = queue.length; chunkSize && canVisit; nextPollIndex(), canVisit--) {
                var watchedFile = queue[pollIndex];
                if (!watchedFile) {
                    continue;
                }
                else if (watchedFile.isClosed) {
                    queue[pollIndex] = undefined;
                    continue;
                }
                // Only files polled count towards chunkSize
                chunkSize--;
                var fileChanged = onWatchedFileStat(watchedFile, getModifiedTime(host, watchedFile.fileName));
                if (watchedFile.isClosed) {
                    // Closed watcher as part of callback
                    queue[pollIndex] = undefined;
                    continue;
                }
                callbackOnWatchFileStat === null || callbackOnWatchFileStat === void 0 ? void 0 : callbackOnWatchFileStat(watchedFile, pollIndex, fileChanged);
                // Defragment the queue while we are at it
                if (queue[pollIndex]) {
                    // Copy this file to the non hole location
                    if (definedValueCopyToIndex < pollIndex) {
                        queue[definedValueCopyToIndex] = watchedFile;
                        queue[pollIndex] = undefined;
                    }
                    definedValueCopyToIndex++;
                }
            }
            // Return next poll index
            return pollIndex;
            function nextPollIndex() {
                pollIndex++;
                if (pollIndex === queue.length) {
                    if (definedValueCopyToIndex < pollIndex) {
                        // There are holes from definedValueCopyToIndex to end of queue, change queue size
                        queue.length = definedValueCopyToIndex;
                    }
                    pollIndex = 0;
                    definedValueCopyToIndex = 0;
                }
            }
        }
        /* @internal */
        function createDynamicPriorityPollingWatchFile(host) {
            var watchedFiles = [];
            var changedFilesInLastPoll = [];
            var lowPollingIntervalQueue = createPollingIntervalQueue(PollingInterval.Low);
            var mediumPollingIntervalQueue = createPollingIntervalQueue(PollingInterval.Medium);
            var highPollingIntervalQueue = createPollingIntervalQueue(PollingInterval.High);
            return watchFile;
            function watchFile(fileName, callback, defaultPollingInterval) {
                var file = {
                    fileName: fileName,
                    callback: callback,
                    unchangedPolls: 0,
                    mtime: getModifiedTime(host, fileName)
                };
                watchedFiles.push(file);
                addToPollingIntervalQueue(file, defaultPollingInterval);
                return {
                    close: function () {
                        file.isClosed = true;
                        // Remove from watchedFiles
                        ts.unorderedRemoveItem(watchedFiles, file);
                        // Do not update polling interval queue since that will happen as part of polling
                    }
                };
            }
            function createPollingIntervalQueue(pollingInterval) {
                var queue = [];
                queue.pollingInterval = pollingInterval;
                queue.pollIndex = 0;
                queue.pollScheduled = false;
                return queue;
            }
            function pollPollingIntervalQueue(queue) {
                queue.pollIndex = pollQueue(queue, queue.pollingInterval, queue.pollIndex, pollingChunkSize[queue.pollingInterval]);
                // Set the next polling index and timeout
                if (queue.length) {
                    scheduleNextPoll(queue.pollingInterval);
                }
                else {
                    ts.Debug.assert(queue.pollIndex === 0);
                    queue.pollScheduled = false;
                }
            }
            function pollLowPollingIntervalQueue(queue) {
                // Always poll complete list of changedFilesInLastPoll
                pollQueue(changedFilesInLastPoll, PollingInterval.Low, /*pollIndex*/ 0, changedFilesInLastPoll.length);
                // Finally do the actual polling of the queue
                pollPollingIntervalQueue(queue);
                // Schedule poll if there are files in changedFilesInLastPoll but no files in the actual queue
                // as pollPollingIntervalQueue wont schedule for next poll
                if (!queue.pollScheduled && changedFilesInLastPoll.length) {
                    scheduleNextPoll(PollingInterval.Low);
                }
            }
            function pollQueue(queue, pollingInterval, pollIndex, chunkSize) {
                return pollWatchedFileQueue(host, queue, pollIndex, chunkSize, onWatchFileStat);
                function onWatchFileStat(watchedFile, pollIndex, fileChanged) {
                    if (fileChanged) {
                        watchedFile.unchangedPolls = 0;
                        // Changed files go to changedFilesInLastPoll queue
                        if (queue !== changedFilesInLastPoll) {
                            queue[pollIndex] = undefined;
                            addChangedFileToLowPollingIntervalQueue(watchedFile);
                        }
                    }
                    else if (watchedFile.unchangedPolls !== ts.unchangedPollThresholds[pollingInterval]) {
                        watchedFile.unchangedPolls++;
                    }
                    else if (queue === changedFilesInLastPoll) {
                        // Restart unchangedPollCount for unchanged file and move to low polling interval queue
                        watchedFile.unchangedPolls = 1;
                        queue[pollIndex] = undefined;
                        addToPollingIntervalQueue(watchedFile, PollingInterval.Low);
                    }
                    else if (pollingInterval !== PollingInterval.High) {
                        watchedFile.unchangedPolls++;
                        queue[pollIndex] = undefined;
                        addToPollingIntervalQueue(watchedFile, pollingInterval === PollingInterval.Low ? PollingInterval.Medium : PollingInterval.High);
                    }
                }
            }
            function pollingIntervalQueue(pollingInterval) {
                switch (pollingInterval) {
                    case PollingInterval.Low:
                        return lowPollingIntervalQueue;
                    case PollingInterval.Medium:
                        return mediumPollingIntervalQueue;
                    case PollingInterval.High:
                        return highPollingIntervalQueue;
                }
            }
            function addToPollingIntervalQueue(file, pollingInterval) {
                pollingIntervalQueue(pollingInterval).push(file);
                scheduleNextPollIfNotAlreadyScheduled(pollingInterval);
            }
            function addChangedFileToLowPollingIntervalQueue(file) {
                changedFilesInLastPoll.push(file);
                scheduleNextPollIfNotAlreadyScheduled(PollingInterval.Low);
            }
            function scheduleNextPollIfNotAlreadyScheduled(pollingInterval) {
                if (!pollingIntervalQueue(pollingInterval).pollScheduled) {
                    scheduleNextPoll(pollingInterval);
                }
            }
            function scheduleNextPoll(pollingInterval) {
                pollingIntervalQueue(pollingInterval).pollScheduled = host.setTimeout(pollingInterval === PollingInterval.Low ? pollLowPollingIntervalQueue : pollPollingIntervalQueue, pollingInterval, pollingIntervalQueue(pollingInterval));
            }
        }
        ts.createDynamicPriorityPollingWatchFile = createDynamicPriorityPollingWatchFile;
        function createUseFsEventsOnParentDirectoryWatchFile(fsWatch, useCaseSensitiveFileNames) {
            // One file can have multiple watchers
            var fileWatcherCallbacks = ts.createMultiMap();
            var dirWatchers = new ts.Map();
            var toCanonicalName = ts.createGetCanonicalFileName(useCaseSensitiveFileNames);
            return nonPollingWatchFile;
            function nonPollingWatchFile(fileName, callback, _pollingInterval, fallbackOptions) {
                var filePath = toCanonicalName(fileName);
                fileWatcherCallbacks.add(filePath, callback);
                var dirPath = ts.getDirectoryPath(filePath) || ".";
                var watcher = dirWatchers.get(dirPath) ||
                    createDirectoryWatcher(ts.getDirectoryPath(fileName) || ".", dirPath, fallbackOptions);
                watcher.referenceCount++;
                return {
                    close: function () {
                        if (watcher.referenceCount === 1) {
                            watcher.close();
                            dirWatchers.delete(dirPath);
                        }
                        else {
                            watcher.referenceCount--;
                        }
                        fileWatcherCallbacks.remove(filePath, callback);
                    }
                };
            }
            function createDirectoryWatcher(dirName, dirPath, fallbackOptions) {
                var watcher = fsWatch(dirName, 1 /* Directory */, function (_eventName, relativeFileName) {
                    // When files are deleted from disk, the triggered "rename" event would have a relativefileName of "undefined"
                    if (!ts.isString(relativeFileName)) {
                        return;
                    }
                    var fileName = ts.getNormalizedAbsolutePath(relativeFileName, dirName);
                    // Some applications save a working file via rename operations
                    var callbacks = fileName && fileWatcherCallbacks.get(toCanonicalName(fileName));
                    if (callbacks) {
                        for (var _i = 0, callbacks_1 = callbacks; _i < callbacks_1.length; _i++) {
                            var fileCallback = callbacks_1[_i];
                            fileCallback(fileName, FileWatcherEventKind.Changed);
                        }
                    }
                }, 
                /*recursive*/ false, PollingInterval.Medium, fallbackOptions);
                watcher.referenceCount = 0;
                dirWatchers.set(dirPath, watcher);
                return watcher;
            }
        }
        function createFixedChunkSizePollingWatchFile(host) {
            var watchedFiles = [];
            var pollIndex = 0;
            var pollScheduled;
            return watchFile;
            function watchFile(fileName, callback) {
                var file = {
                    fileName: fileName,
                    callback: callback,
                    mtime: getModifiedTime(host, fileName)
                };
                watchedFiles.push(file);
                scheduleNextPoll();
                return {
                    close: function () {
                        file.isClosed = true;
                        ts.unorderedRemoveItem(watchedFiles, file);
                    }
                };
            }
            function pollQueue() {
                pollScheduled = undefined;
                pollIndex = pollWatchedFileQueue(host, watchedFiles, pollIndex, pollingChunkSize[PollingInterval.Low]);
                scheduleNextPoll();
            }
            function scheduleNextPoll() {
                if (!watchedFiles.length || pollScheduled)
                    return;
                pollScheduled = host.setTimeout(pollQueue, PollingInterval.High);
            }
        }
        /* @internal */
        function createSingleFileWatcherPerName(watchFile, useCaseSensitiveFileNames) {
            var cache = new ts.Map();
            var callbacksCache = ts.createMultiMap();
            var toCanonicalFileName = ts.createGetCanonicalFileName(useCaseSensitiveFileNames);
            return function (fileName, callback, pollingInterval, options) {
                var path = toCanonicalFileName(fileName);
                var existing = cache.get(path);
                if (existing) {
                    existing.refCount++;
                }
                else {
                    cache.set(path, {
                        watcher: watchFile(fileName, function (fileName, eventKind) { return ts.forEach(callbacksCache.get(path), function (cb) { return cb(fileName, eventKind); }); }, pollingInterval, options),
                        refCount: 1
                    });
                }
                callbacksCache.add(path, callback);
                return {
                    close: function () {
                        var watcher = ts.Debug.checkDefined(cache.get(path));
                        callbacksCache.remove(path, callback);
                        watcher.refCount--;
                        if (watcher.refCount)
                            return;
                        cache.delete(path);
                        ts.closeFileWatcherOf(watcher);
                    }
                };
            };
        }
        ts.createSingleFileWatcherPerName = createSingleFileWatcherPerName;
        /**
         * Returns true if file status changed
         */
        /*@internal*/
        function onWatchedFileStat(watchedFile, modifiedTime) {
            var oldTime = watchedFile.mtime.getTime();
            var newTime = modifiedTime.getTime();
            if (oldTime !== newTime) {
                watchedFile.mtime = modifiedTime;
                watchedFile.callback(watchedFile.fileName, getFileWatcherEventKind(oldTime, newTime));
                return true;
            }
            return false;
        }
        ts.onWatchedFileStat = onWatchedFileStat;
        /*@internal*/
        function getFileWatcherEventKind(oldTime, newTime) {
            return oldTime === 0
                ? FileWatcherEventKind.Created
                : newTime === 0
                    ? FileWatcherEventKind.Deleted
                    : FileWatcherEventKind.Changed;
        }
        ts.getFileWatcherEventKind = getFileWatcherEventKind;
        /*@internal*/
        ts.ignoredPaths = ["/node_modules/.", "/.git", "/.#"];
        /*@internal*/
        ts.sysLog = ts.noop; // eslint-disable-line prefer-const
        /*@internal*/
        function setSysLog(logger) {
            ts.sysLog = logger;
        }
        ts.setSysLog = setSysLog;
        /**
         * Watch the directory recursively using host provided method to watch child directories
         * that means if this is recursive watcher, watch the children directories as well
         * (eg on OS that dont support recursive watch using fs.watch use fs.watchFile)
         */
        /*@internal*/
        function createDirectoryWatcherSupportingRecursive(_a) {
            var watchDirectory = _a.watchDirectory, useCaseSensitiveFileNames = _a.useCaseSensitiveFileNames, getCurrentDirectory = _a.getCurrentDirectory, getAccessibleSortedChildDirectories = _a.getAccessibleSortedChildDirectories, directoryExists = _a.directoryExists, realpath = _a.realpath, setTimeout = _a.setTimeout, clearTimeout = _a.clearTimeout;
            var cache = new ts.Map();
            var callbackCache = ts.createMultiMap();
            var cacheToUpdateChildWatches = new ts.Map();
            var timerToUpdateChildWatches;
            var filePathComparer = ts.getStringComparer(!useCaseSensitiveFileNames);
            var toCanonicalFilePath = ts.createGetCanonicalFileName(useCaseSensitiveFileNames);
            return function (dirName, callback, recursive, options) {
                return recursive ?
                    createDirectoryWatcher(dirName, options, callback) :
                    watchDirectory(dirName, callback, recursive, options);
            };
            /**
             * Create the directory watcher for the dirPath.
             */
            function createDirectoryWatcher(dirName, options, callback) {
                var dirPath = toCanonicalFilePath(dirName);
                var directoryWatcher = cache.get(dirPath);
                if (directoryWatcher) {
                    directoryWatcher.refCount++;
                }
                else {
                    directoryWatcher = {
                        watcher: watchDirectory(dirName, function (fileName) {
                            if (isIgnoredPath(fileName, options))
                                return;
                            if (options === null || options === void 0 ? void 0 : options.synchronousWatchDirectory) {
                                // Call the actual callback
                                invokeCallbacks(dirPath, fileName);
                                // Iterate through existing children and update the watches if needed
                                updateChildWatches(dirName, dirPath, options);
                            }
                            else {
                                nonSyncUpdateChildWatches(dirName, dirPath, fileName, options);
                            }
                        }, /*recursive*/ false, options),
                        refCount: 1,
                        childWatches: ts.emptyArray
                    };
                    cache.set(dirPath, directoryWatcher);
                    updateChildWatches(dirName, dirPath, options);
                }
                var callbackToAdd = callback && { dirName: dirName, callback: callback };
                if (callbackToAdd) {
                    callbackCache.add(dirPath, callbackToAdd);
                }
                return {
                    dirName: dirName,
                    close: function () {
                        var directoryWatcher = ts.Debug.checkDefined(cache.get(dirPath));
                        if (callbackToAdd)
                            callbackCache.remove(dirPath, callbackToAdd);
                        directoryWatcher.refCount--;
                        if (directoryWatcher.refCount)
                            return;
                        cache.delete(dirPath);
                        ts.closeFileWatcherOf(directoryWatcher);
                        directoryWatcher.childWatches.forEach(ts.closeFileWatcher);
                    }
                };
            }
            function invokeCallbacks(dirPath, fileNameOrInvokeMap, fileNames) {
                var fileName;
                var invokeMap;
                if (ts.isString(fileNameOrInvokeMap)) {
                    fileName = fileNameOrInvokeMap;
                }
                else {
                    invokeMap = fileNameOrInvokeMap;
                }
                // Call the actual callback
                callbackCache.forEach(function (callbacks, rootDirName) {
                    var _a;
                    if (invokeMap && invokeMap.get(rootDirName) === true)
                        return;
                    if (rootDirName === dirPath || (ts.startsWith(dirPath, rootDirName) && dirPath[rootDirName.length] === ts.directorySeparator)) {
                        if (invokeMap) {
                            if (fileNames) {
                                var existing = invokeMap.get(rootDirName);
                                if (existing) {
                                    (_a = existing).push.apply(_a, fileNames);
                                }
                                else {
                                    invokeMap.set(rootDirName, fileNames.slice());
                                }
                            }
                            else {
                                invokeMap.set(rootDirName, true);
                            }
                        }
                        else {
                            callbacks.forEach(function (_a) {
                                var callback = _a.callback;
                                return callback(fileName);
                            });
                        }
                    }
                });
            }
            function nonSyncUpdateChildWatches(dirName, dirPath, fileName, options) {
                // Iterate through existing children and update the watches if needed
                var parentWatcher = cache.get(dirPath);
                if (parentWatcher && directoryExists(dirName)) {
                    // Schedule the update and postpone invoke for callbacks
                    scheduleUpdateChildWatches(dirName, dirPath, fileName, options);
                    return;
                }
                // Call the actual callbacks and remove child watches
                invokeCallbacks(dirPath, fileName);
                removeChildWatches(parentWatcher);
            }
            function scheduleUpdateChildWatches(dirName, dirPath, fileName, options) {
                var existing = cacheToUpdateChildWatches.get(dirPath);
                if (existing) {
                    existing.fileNames.push(fileName);
                }
                else {
                    cacheToUpdateChildWatches.set(dirPath, { dirName: dirName, options: options, fileNames: [fileName] });
                }
                if (timerToUpdateChildWatches) {
                    clearTimeout(timerToUpdateChildWatches);
                    timerToUpdateChildWatches = undefined;
                }
                timerToUpdateChildWatches = setTimeout(onTimerToUpdateChildWatches, 1000);
            }
            function onTimerToUpdateChildWatches() {
                timerToUpdateChildWatches = undefined;
                ts.sysLog("sysLog:: onTimerToUpdateChildWatches:: " + cacheToUpdateChildWatches.size);
                var start = ts.timestamp();
                var invokeMap = new ts.Map();
                while (!timerToUpdateChildWatches && cacheToUpdateChildWatches.size) {
                    var result = cacheToUpdateChildWatches.entries().next();
                    ts.Debug.assert(!result.done);
                    var _a = result.value, dirPath = _a[0], _b = _a[1], dirName = _b.dirName, options = _b.options, fileNames = _b.fileNames;
                    cacheToUpdateChildWatches.delete(dirPath);
                    // Because the child refresh is fresh, we would need to invalidate whole root directory being watched
                    // to ensure that all the changes are reflected at this time
                    var hasChanges = updateChildWatches(dirName, dirPath, options);
                    invokeCallbacks(dirPath, invokeMap, hasChanges ? undefined : fileNames);
                }
                ts.sysLog("sysLog:: invokingWatchers:: Elapsed:: " + (ts.timestamp() - start) + "ms:: " + cacheToUpdateChildWatches.size);
                callbackCache.forEach(function (callbacks, rootDirName) {
                    var existing = invokeMap.get(rootDirName);
                    if (existing) {
                        callbacks.forEach(function (_a) {
                            var callback = _a.callback, dirName = _a.dirName;
                            if (ts.isArray(existing)) {
                                existing.forEach(callback);
                            }
                            else {
                                callback(dirName);
                            }
                        });
                    }
                });
                var elapsed = ts.timestamp() - start;
                ts.sysLog("sysLog:: Elapsed:: " + elapsed + "ms:: onTimerToUpdateChildWatches:: " + cacheToUpdateChildWatches.size + " " + timerToUpdateChildWatches);
            }
            function removeChildWatches(parentWatcher) {
                if (!parentWatcher)
                    return;
                var existingChildWatches = parentWatcher.childWatches;
                parentWatcher.childWatches = ts.emptyArray;
                for (var _i = 0, existingChildWatches_1 = existingChildWatches; _i < existingChildWatches_1.length; _i++) {
                    var childWatcher = existingChildWatches_1[_i];
                    childWatcher.close();
                    removeChildWatches(cache.get(toCanonicalFilePath(childWatcher.dirName)));
                }
            }
            function updateChildWatches(parentDir, parentDirPath, options) {
                // Iterate through existing children and update the watches if needed
                var parentWatcher = cache.get(parentDirPath);
                if (!parentWatcher)
                    return false;
                var newChildWatches;
                var hasChanges = ts.enumerateInsertsAndDeletes(directoryExists(parentDir) ? ts.mapDefined(getAccessibleSortedChildDirectories(parentDir), function (child) {
                    var childFullName = ts.getNormalizedAbsolutePath(child, parentDir);
                    // Filter our the symbolic link directories since those arent included in recursive watch
                    // which is same behaviour when recursive: true is passed to fs.watch
                    return !isIgnoredPath(childFullName, options) && filePathComparer(childFullName, ts.normalizePath(realpath(childFullName))) === 0 /* EqualTo */ ? childFullName : undefined;
                }) : ts.emptyArray, parentWatcher.childWatches, function (child, childWatcher) { return filePathComparer(child, childWatcher.dirName); }, createAndAddChildDirectoryWatcher, ts.closeFileWatcher, addChildDirectoryWatcher);
                parentWatcher.childWatches = newChildWatches || ts.emptyArray;
                return hasChanges;
                /**
                 * Create new childDirectoryWatcher and add it to the new ChildDirectoryWatcher list
                 */
                function createAndAddChildDirectoryWatcher(childName) {
                    var result = createDirectoryWatcher(childName, options);
                    addChildDirectoryWatcher(result);
                }
                /**
                 * Add child directory watcher to the new ChildDirectoryWatcher list
                 */
                function addChildDirectoryWatcher(childWatcher) {
                    (newChildWatches || (newChildWatches = [])).push(childWatcher);
                }
            }
            function isIgnoredPath(path, options) {
                return ts.some(ts.ignoredPaths, function (searchPath) { return isInPath(path, searchPath); }) ||
                    isIgnoredByWatchOptions(path, options, useCaseSensitiveFileNames, getCurrentDirectory);
            }
            function isInPath(path, searchPath) {
                if (ts.stringContains(path, searchPath))
                    return true;
                if (useCaseSensitiveFileNames)
                    return false;
                return ts.stringContains(toCanonicalFilePath(path), searchPath);
            }
        }
        ts.createDirectoryWatcherSupportingRecursive = createDirectoryWatcherSupportingRecursive;
        /*@internal*/
        var FileSystemEntryKind;
        (function (FileSystemEntryKind) {
            FileSystemEntryKind[FileSystemEntryKind["File"] = 0] = "File";
            FileSystemEntryKind[FileSystemEntryKind["Directory"] = 1] = "Directory";
        })(FileSystemEntryKind = ts.FileSystemEntryKind || (ts.FileSystemEntryKind = {}));
        /*@internal*/
        function createFileWatcherCallback(callback) {
            return function (_fileName, eventKind) { return callback(eventKind === FileWatcherEventKind.Changed ? "change" : "rename", ""); };
        }
        ts.createFileWatcherCallback = createFileWatcherCallback;
        function createFsWatchCallbackForFileWatcherCallback(fileName, callback, fileExists) {
            return function (eventName) {
                if (eventName === "rename") {
                    callback(fileName, fileExists(fileName) ? FileWatcherEventKind.Created : FileWatcherEventKind.Deleted);
                }
                else {
                    // Change
                    callback(fileName, FileWatcherEventKind.Changed);
                }
            };
        }
        function isIgnoredByWatchOptions(pathToCheck, options, useCaseSensitiveFileNames, getCurrentDirectory) {
            return ((options === null || options === void 0 ? void 0 : options.excludeDirectories) || (options === null || options === void 0 ? void 0 : options.excludeFiles)) && (ts.matchesExclude(pathToCheck, options === null || options === void 0 ? void 0 : options.excludeFiles, useCaseSensitiveFileNames, getCurrentDirectory()) ||
                ts.matchesExclude(pathToCheck, options === null || options === void 0 ? void 0 : options.excludeDirectories, useCaseSensitiveFileNames, getCurrentDirectory()));
        }
        function createFsWatchCallbackForDirectoryWatcherCallback(directoryName, callback, options, useCaseSensitiveFileNames, getCurrentDirectory) {
            return function (eventName, relativeFileName) {
                // In watchDirectory we only care about adding and removing files (when event name is
                // "rename"); changes made within files are handled by corresponding fileWatchers (when
                // event name is "change")
                if (eventName === "rename") {
                    // When deleting a file, the passed baseFileName is null
                    var fileName = !relativeFileName ? directoryName : ts.normalizePath(ts.combinePaths(directoryName, relativeFileName));
                    if (!relativeFileName || !isIgnoredByWatchOptions(fileName, options, useCaseSensitiveFileNames, getCurrentDirectory)) {
                        callback(fileName);
                    }
                }
            };
        }
        /*@internal*/
        function createSystemWatchFunctions(_a) {
            var pollingWatchFile = _a.pollingWatchFile, getModifiedTime = _a.getModifiedTime, setTimeout = _a.setTimeout, clearTimeout = _a.clearTimeout, fsWatch = _a.fsWatch, fileExists = _a.fileExists, useCaseSensitiveFileNames = _a.useCaseSensitiveFileNames, getCurrentDirectory = _a.getCurrentDirectory, fsSupportsRecursiveFsWatch = _a.fsSupportsRecursiveFsWatch, directoryExists = _a.directoryExists, getAccessibleSortedChildDirectories = _a.getAccessibleSortedChildDirectories, realpath = _a.realpath, tscWatchFile = _a.tscWatchFile, useNonPollingWatchers = _a.useNonPollingWatchers, tscWatchDirectory = _a.tscWatchDirectory, defaultWatchFileKind = _a.defaultWatchFileKind;
            var dynamicPollingWatchFile;
            var fixedChunkSizePollingWatchFile;
            var nonPollingWatchFile;
            var hostRecursiveDirectoryWatcher;
            return {
                watchFile: watchFile,
                watchDirectory: watchDirectory
            };
            function watchFile(fileName, callback, pollingInterval, options) {
                options = updateOptionsForWatchFile(options, useNonPollingWatchers);
                var watchFileKind = ts.Debug.checkDefined(options.watchFile);
                switch (watchFileKind) {
                    case ts.WatchFileKind.FixedPollingInterval:
                        return pollingWatchFile(fileName, callback, PollingInterval.Low, /*options*/ undefined);
                    case ts.WatchFileKind.PriorityPollingInterval:
                        return pollingWatchFile(fileName, callback, pollingInterval, /*options*/ undefined);
                    case ts.WatchFileKind.DynamicPriorityPolling:
                        return ensureDynamicPollingWatchFile()(fileName, callback, pollingInterval, /*options*/ undefined);
                    case ts.WatchFileKind.FixedChunkSizePolling:
                        return ensureFixedChunkSizePollingWatchFile()(fileName, callback, /* pollingInterval */ undefined, /*options*/ undefined);
                    case ts.WatchFileKind.UseFsEvents:
                        return fsWatch(fileName, 0 /* File */, createFsWatchCallbackForFileWatcherCallback(fileName, callback, fileExists), 
                        /*recursive*/ false, pollingInterval, ts.getFallbackOptions(options));
                    case ts.WatchFileKind.UseFsEventsOnParentDirectory:
                        if (!nonPollingWatchFile) {
                            nonPollingWatchFile = createUseFsEventsOnParentDirectoryWatchFile(fsWatch, useCaseSensitiveFileNames);
                        }
                        return nonPollingWatchFile(fileName, callback, pollingInterval, ts.getFallbackOptions(options));
                    default:
                        ts.Debug.assertNever(watchFileKind);
                }
            }
            function ensureDynamicPollingWatchFile() {
                return dynamicPollingWatchFile || (dynamicPollingWatchFile = createDynamicPriorityPollingWatchFile({ getModifiedTime: getModifiedTime, setTimeout: setTimeout }));
            }
            function ensureFixedChunkSizePollingWatchFile() {
                return fixedChunkSizePollingWatchFile || (fixedChunkSizePollingWatchFile = createFixedChunkSizePollingWatchFile({ getModifiedTime: getModifiedTime, setTimeout: setTimeout }));
            }
            function updateOptionsForWatchFile(options, useNonPollingWatchers) {
                if (options && options.watchFile !== undefined)
                    return options;
                switch (tscWatchFile) {
                    case "PriorityPollingInterval":
                        // Use polling interval based on priority when create watch using host.watchFile
                        return { watchFile: ts.WatchFileKind.PriorityPollingInterval };
                    case "DynamicPriorityPolling":
                        // Use polling interval but change the interval depending on file changes and their default polling interval
                        return { watchFile: ts.WatchFileKind.DynamicPriorityPolling };
                    case "UseFsEvents":
                        // Use notifications from FS to watch with falling back to fs.watchFile
                        return generateWatchFileOptions(ts.WatchFileKind.UseFsEvents, ts.PollingWatchKind.PriorityInterval, options);
                    case "UseFsEventsWithFallbackDynamicPolling":
                        // Use notifications from FS to watch with falling back to dynamic watch file
                        return generateWatchFileOptions(ts.WatchFileKind.UseFsEvents, ts.PollingWatchKind.DynamicPriority, options);
                    case "UseFsEventsOnParentDirectory":
                        useNonPollingWatchers = true;
                    // fall through
                    default:
                        return useNonPollingWatchers ?
                            // Use notifications from FS to watch with falling back to fs.watchFile
                            generateWatchFileOptions(ts.WatchFileKind.UseFsEventsOnParentDirectory, ts.PollingWatchKind.PriorityInterval, options) :
                            // Default to do not use fixed polling interval
                            { watchFile: (defaultWatchFileKind === null || defaultWatchFileKind === void 0 ? void 0 : defaultWatchFileKind()) || ts.WatchFileKind.FixedPollingInterval };
                }
            }
            function generateWatchFileOptions(watchFile, fallbackPolling, options) {
                var defaultFallbackPolling = options === null || options === void 0 ? void 0 : options.fallbackPolling;
                return {
                    watchFile: watchFile,
                    fallbackPolling: defaultFallbackPolling === undefined ?
                        fallbackPolling :
                        defaultFallbackPolling
                };
            }
            function watchDirectory(directoryName, callback, recursive, options) {
                if (fsSupportsRecursiveFsWatch) {
                    return fsWatch(directoryName, 1 /* Directory */, createFsWatchCallbackForDirectoryWatcherCallback(directoryName, callback, options, useCaseSensitiveFileNames, getCurrentDirectory), recursive, PollingInterval.Medium, ts.getFallbackOptions(options));
                }
                if (!hostRecursiveDirectoryWatcher) {
                    hostRecursiveDirectoryWatcher = createDirectoryWatcherSupportingRecursive({
                        useCaseSensitiveFileNames: useCaseSensitiveFileNames,
                        getCurrentDirectory: getCurrentDirectory,
                        directoryExists: directoryExists,
                        getAccessibleSortedChildDirectories: getAccessibleSortedChildDirectories,
                        watchDirectory: nonRecursiveWatchDirectory,
                        realpath: realpath,
                        setTimeout: setTimeout,
                        clearTimeout: clearTimeout
                    });
                }
                return hostRecursiveDirectoryWatcher(directoryName, callback, recursive, options);
            }
            function nonRecursiveWatchDirectory(directoryName, callback, recursive, options) {
                ts.Debug.assert(!recursive);
                var watchDirectoryOptions = updateOptionsForWatchDirectory(options);
                var watchDirectoryKind = ts.Debug.checkDefined(watchDirectoryOptions.watchDirectory);
                switch (watchDirectoryKind) {
                    case ts.WatchDirectoryKind.FixedPollingInterval:
                        return pollingWatchFile(directoryName, function () { return callback(directoryName); }, PollingInterval.Medium, 
                        /*options*/ undefined);
                    case ts.WatchDirectoryKind.DynamicPriorityPolling:
                        return ensureDynamicPollingWatchFile()(directoryName, function () { return callback(directoryName); }, PollingInterval.Medium, 
                        /*options*/ undefined);
                    case ts.WatchDirectoryKind.FixedChunkSizePolling:
                        return ensureFixedChunkSizePollingWatchFile()(directoryName, function () { return callback(directoryName); }, 
                        /* pollingInterval */ undefined, 
                        /*options*/ undefined);
                    case ts.WatchDirectoryKind.UseFsEvents:
                        return fsWatch(directoryName, 1 /* Directory */, createFsWatchCallbackForDirectoryWatcherCallback(directoryName, callback, options, useCaseSensitiveFileNames, getCurrentDirectory), recursive, PollingInterval.Medium, ts.getFallbackOptions(watchDirectoryOptions));
                    default:
                        ts.Debug.assertNever(watchDirectoryKind);
                }
            }
            function updateOptionsForWatchDirectory(options) {
                if (options && options.watchDirectory !== undefined)
                    return options;
                switch (tscWatchDirectory) {
                    case "RecursiveDirectoryUsingFsWatchFile":
                        // Use polling interval based on priority when create watch using host.watchFile
                        return { watchDirectory: ts.WatchDirectoryKind.FixedPollingInterval };
                    case "RecursiveDirectoryUsingDynamicPriorityPolling":
                        // Use polling interval but change the interval depending on file changes and their default polling interval
                        return { watchDirectory: ts.WatchDirectoryKind.DynamicPriorityPolling };
                    default:
                        var defaultFallbackPolling = options === null || options === void 0 ? void 0 : options.fallbackPolling;
                        return {
                            watchDirectory: ts.WatchDirectoryKind.UseFsEvents,
                            fallbackPolling: defaultFallbackPolling !== undefined ?
                                defaultFallbackPolling :
                                undefined
                        };
                }
            }
        }
        ts.createSystemWatchFunctions = createSystemWatchFunctions;
        /**
         * patch writefile to create folder before writing the file
         */
        /*@internal*/
        function patchWriteFileEnsuringDirectory(sys) {
            // patch writefile to create folder before writing the file
            var originalWriteFile = sys.writeFile;
            sys.writeFile = function (path, data, writeBom) {
                return ts.writeFileEnsuringDirectories(path, data, !!writeBom, function (path, data, writeByteOrderMark) { return originalWriteFile.call(sys, path, data, writeByteOrderMark); }, function (path) { return sys.createDirectory(path); }, function (path) { return sys.directoryExists(path); });
            };
        }
        ts.patchWriteFileEnsuringDirectory = patchWriteFileEnsuringDirectory;
        function getNodeMajorVersion() {
            if (typeof process === "undefined") {
                return undefined;
            }
            var version = process.version;
            if (!version) {
                return undefined;
            }
            var dot = version.indexOf(".");
            if (dot === -1) {
                return undefined;
            }
            return parseInt(version.substring(1, dot));
        }
        ts.getNodeMajorVersion = getNodeMajorVersion;
        // TODO: GH#18217 this is used as if it's certainly defined in many places.
        // eslint-disable-next-line prefer-const
        ts.sys = (function () {
            // NodeJS detects "\uFEFF" at the start of the string and *replaces* it with the actual
            // byte order mark from the specified encoding. Using any other byte order mark does
            // not actually work.
            var byteOrderMarkIndicator = "\uFEFF";
            function getNodeSystem() {
                var _a;
                var nativePattern = /^native |^\([^)]+\)$|^(internal[\\/]|[a-zA-Z0-9_\s]+(\.js)?$)/;
                var _fs = require("fs");
                var _path = require("path");
                var _os = require("os");
                // crypto can be absent on reduced node installations
                var _crypto;
                try {
                    _crypto = require("crypto");
                }
                catch (_b) {
                    _crypto = undefined;
                }
                var activeSession;
                var profilePath = "./profile.cpuprofile";
                var hitSystemWatcherLimit = false;
                var Buffer = require("buffer").Buffer;
                var nodeVersion = getNodeMajorVersion();
                var isNode4OrLater = nodeVersion >= 4;
                var isLinuxOrMacOs = process.platform === "linux" || process.platform === "darwin";
                var platform = _os.platform();
                var useCaseSensitiveFileNames = isFileSystemCaseSensitive();
                var realpathSync = useCaseSensitiveFileNames ? ((_a = _fs.realpathSync.native) !== null && _a !== void 0 ? _a : _fs.realpathSync) : _fs.realpathSync;
                var fsSupportsRecursiveFsWatch = isNode4OrLater && (process.platform === "win32" || process.platform === "darwin");
                var getCurrentDirectory = ts.memoize(function () { return process.cwd(); });
                var _c = createSystemWatchFunctions({
                    pollingWatchFile: createSingleFileWatcherPerName(fsWatchFileWorker, useCaseSensitiveFileNames),
                    getModifiedTime: getModifiedTime,
                    setTimeout: setTimeout,
                    clearTimeout: clearTimeout,
                    fsWatch: fsWatch,
                    useCaseSensitiveFileNames: useCaseSensitiveFileNames,
                    getCurrentDirectory: getCurrentDirectory,
                    fileExists: fileExists,
                    // Node 4.0 `fs.watch` function supports the "recursive" option on both OSX and Windows
                    // (ref: https://github.com/nodejs/node/pull/2649 and https://github.com/Microsoft/TypeScript/issues/4643)
                    fsSupportsRecursiveFsWatch: fsSupportsRecursiveFsWatch,
                    directoryExists: directoryExists,
                    getAccessibleSortedChildDirectories: function (path) { return getAccessibleFileSystemEntries(path).directories; },
                    realpath: realpath,
                    tscWatchFile: process.env.TSC_WATCHFILE,
                    useNonPollingWatchers: process.env.TSC_NONPOLLING_WATCHER,
                    tscWatchDirectory: process.env.TSC_WATCHDIRECTORY,
                    defaultWatchFileKind: function () { var _a, _b; return (_b = (_a = sys).defaultWatchFileKind) === null || _b === void 0 ? void 0 : _b.call(_a); },
                }), watchFile = _c.watchFile, watchDirectory = _c.watchDirectory;
                var nodeSystem = {
                    args: process.argv.slice(2),
                    newLine: _os.EOL,
                    useCaseSensitiveFileNames: useCaseSensitiveFileNames,
                    write: function (s) {
                        process.stdout.write(s);
                    },
                    getWidthOfTerminal: function () {
                        return process.stdout.columns;
                    },
                    writeOutputIsTTY: function () {
                        return process.stdout.isTTY;
                    },
                    readFile: readFile,
                    writeFile: writeFile,
                    watchFile: watchFile,
                    watchDirectory: watchDirectory,
                    resolvePath: function (path) { return _path.resolve(path); },
                    fileExists: fileExists,
                    directoryExists: directoryExists,
                    createDirectory: function (directoryName) {
                        if (!nodeSystem.directoryExists(directoryName)) {
                            // Wrapped in a try-catch to prevent crashing if we are in a race
                            // with another copy of ourselves to create the same directory
                            try {
                                _fs.mkdirSync(directoryName);
                            }
                            catch (e) {
                                if (e.code !== "EEXIST") {
                                    // Failed for some other reason (access denied?); still throw
                                    throw e;
                                }
                            }
                        }
                    },
                    getExecutingFilePath: function () {
                        return __filename;
                    },
                    getCurrentDirectory: getCurrentDirectory,
                    getDirectories: getDirectories,
                    getEnvironmentVariable: function (name) {
                        return process.env[name] || "";
                    },
                    readDirectory: readDirectory,
                    getModifiedTime: getModifiedTime,
                    setModifiedTime: setModifiedTime,
                    deleteFile: deleteFile,
                    createHash: _crypto ? createSHA256Hash : generateDjb2Hash,
                    createSHA256Hash: _crypto ? createSHA256Hash : undefined,
                    getMemoryUsage: function () {
                        if (global.gc) {
                            global.gc();
                        }
                        return process.memoryUsage().heapUsed;
                    },
                    getFileSize: function (path) {
                        try {
                            var stat = statSync(path);
                            if (stat === null || stat === void 0 ? void 0 : stat.isFile()) {
                                return stat.size;
                            }
                        }
                        catch ( /*ignore*/_a) { /*ignore*/ }
                        return 0;
                    },
                    exit: function (exitCode) {
                        disableCPUProfiler(function () { return process.exit(exitCode); });
                    },
                    enableCPUProfiler: enableCPUProfiler,
                    disableCPUProfiler: disableCPUProfiler,
                    cpuProfilingEnabled: function () { return !!activeSession || ts.contains(process.execArgv, "--cpu-prof") || ts.contains(process.execArgv, "--prof"); },
                    realpath: realpath,
                    debugMode: !!process.env.NODE_INSPECTOR_IPC || !!process.env.VSCODE_INSPECTOR_OPTIONS || ts.some(process.execArgv, function (arg) { return /^--(inspect|debug)(-brk)?(=\d+)?$/i.test(arg); }),
                    tryEnableSourceMapsForHost: function () {
                        try {
                            require(22) /* source-map-support */.install();
                        }
                        catch (_a) {
                            // Could not enable source maps.
                        }
                    },
                    setTimeout: setTimeout,
                    clearTimeout: clearTimeout,
                    clearScreen: function () {
                        process.stdout.write("\x1Bc");
                    },
                    setBlocking: function () {
                        if (process.stdout && process.stdout._handle && process.stdout._handle.setBlocking) {
                            process.stdout._handle.setBlocking(true);
                        }
                    },
                    bufferFrom: bufferFrom,
                    base64decode: function (input) { return bufferFrom(input, "base64").toString("utf8"); },
                    base64encode: function (input) { return bufferFrom(input).toString("base64"); },
                    require: function (baseDir, moduleName) {
                        try {
                            var modulePath = ts.resolveJSModule(moduleName, baseDir, nodeSystem);
                            return { module: require(modulePath), modulePath: modulePath, error: undefined };
                        }
                        catch (error) {
                            return { module: undefined, modulePath: undefined, error: error };
                        }
                    }
                };
                return nodeSystem;
                /**
                 * `throwIfNoEntry` was added so recently that it's not in the node types.
                 * This helper encapsulates the mitigating usage of `any`.
                 * See https://github.com/nodejs/node/pull/33716
                 */
                function statSync(path) {
                    // throwIfNoEntry will be ignored by older versions of node
                    return _fs.statSync(path, { throwIfNoEntry: false });
                }
                /**
                 * Uses the builtin inspector APIs to capture a CPU profile
                 * See https://nodejs.org/api/inspector.html#inspector_example_usage for details
                 */
                function enableCPUProfiler(path, cb) {
                    if (activeSession) {
                        cb();
                        return false;
                    }
                    var inspector = require("inspector");
                    if (!inspector || !inspector.Session) {
                        cb();
                        return false;
                    }
                    var session = new inspector.Session();
                    session.connect();
                    session.post("Profiler.enable", function () {
                        session.post("Profiler.start", function () {
                            activeSession = session;
                            profilePath = path;
                            cb();
                        });
                    });
                    return true;
                }
                /**
                 * Strips non-TS paths from the profile, so users with private projects shouldn't
                 * need to worry about leaking paths by submitting a cpu profile to us
                 */
                function cleanupPaths(profile) {
                    var externalFileCounter = 0;
                    var remappedPaths = new ts.Map();
                    var normalizedDir = ts.normalizeSlashes(__dirname);
                    // Windows rooted dir names need an extra `/` prepended to be valid file:/// urls
                    var fileUrlRoot = "file://" + (ts.getRootLength(normalizedDir) === 1 ? "" : "/") + normalizedDir;
                    for (var _i = 0, _a = profile.nodes; _i < _a.length; _i++) {
                        var node = _a[_i];
                        if (node.callFrame.url) {
                            var url = ts.normalizeSlashes(node.callFrame.url);
                            if (ts.containsPath(fileUrlRoot, url, useCaseSensitiveFileNames)) {
                                node.callFrame.url = ts.getRelativePathToDirectoryOrUrl(fileUrlRoot, url, fileUrlRoot, ts.createGetCanonicalFileName(useCaseSensitiveFileNames), /*isAbsolutePathAnUrl*/ true);
                            }
                            else if (!nativePattern.test(url)) {
                                node.callFrame.url = (remappedPaths.has(url) ? remappedPaths : remappedPaths.set(url, "external" + externalFileCounter + ".js")).get(url);
                                externalFileCounter++;
                            }
                        }
                    }
                    return profile;
                }
                function disableCPUProfiler(cb) {
                    if (activeSession && activeSession !== "stopping") {
                        var s_1 = activeSession;
                        activeSession.post("Profiler.stop", function (err, _a) {
                            var _b;
                            var profile = _a.profile;
                            if (!err) {
                                try {
                                    if ((_b = statSync(profilePath)) === null || _b === void 0 ? void 0 : _b.isDirectory()) {
                                        profilePath = _path.join(profilePath, (new Date()).toISOString().replace(/:/g, "-") + "+P" + process.pid + ".cpuprofile");
                                    }
                                }
                                catch (_c) {
                                    // do nothing and ignore fallible fs operation
                                }
                                try {
                                    _fs.mkdirSync(_path.dirname(profilePath), { recursive: true });
                                }
                                catch (_d) {
                                    // do nothing and ignore fallible fs operation
                                }
                                _fs.writeFileSync(profilePath, JSON.stringify(cleanupPaths(profile)));
                            }
                            activeSession = undefined;
                            s_1.disconnect();
                            cb();
                        });
                        activeSession = "stopping";
                        return true;
                    }
                    else {
                        cb();
                        return false;
                    }
                }
                function bufferFrom(input, encoding) {
                    // See https://github.com/Microsoft/TypeScript/issues/25652
                    return Buffer.from && Buffer.from !== Int8Array.from
                        ? Buffer.from(input, encoding)
                        : new Buffer(input, encoding);
                }
                function isFileSystemCaseSensitive() {
                    // win32\win64 are case insensitive platforms
                    if (platform === "win32" || platform === "win64") {
                        return false;
                    }
                    // If this file exists under a different case, we must be case-insensitve.
                    return !fileExists(swapCase(__filename));
                }
                /** Convert all lowercase chars to uppercase, and vice-versa */
                function swapCase(s) {
                    return s.replace(/\w/g, function (ch) {
                        var up = ch.toUpperCase();
                        return ch === up ? ch.toLowerCase() : up;
                    });
                }
                function fsWatchFileWorker(fileName, callback, pollingInterval) {
                    _fs.watchFile(fileName, { persistent: true, interval: pollingInterval }, fileChanged);
                    var eventKind;
                    return {
                        close: function () { return _fs.unwatchFile(fileName, fileChanged); }
                    };
                    function fileChanged(curr, prev) {
                        // previous event kind check is to ensure we recongnize the file as previously also missing when it is restored or renamed twice (that is it disappears and reappears)
                        // In such case, prevTime returned is same as prev time of event when file was deleted as per node documentation
                        var isPreviouslyDeleted = +prev.mtime === 0 || eventKind === FileWatcherEventKind.Deleted;
                        if (+curr.mtime === 0) {
                            if (isPreviouslyDeleted) {
                                // Already deleted file, no need to callback again
                                return;
                            }
                            eventKind = FileWatcherEventKind.Deleted;
                        }
                        else if (isPreviouslyDeleted) {
                            eventKind = FileWatcherEventKind.Created;
                        }
                        // If there is no change in modified time, ignore the event
                        else if (+curr.mtime === +prev.mtime) {
                            return;
                        }
                        else {
                            // File changed
                            eventKind = FileWatcherEventKind.Changed;
                        }
                        callback(fileName, eventKind);
                    }
                }
                function fsWatch(fileOrDirectory, entryKind, callback, recursive, fallbackPollingInterval, fallbackOptions) {
                    var options;
                    var lastDirectoryPartWithDirectorySeparator;
                    var lastDirectoryPart;
                    if (isLinuxOrMacOs) {
                        lastDirectoryPartWithDirectorySeparator = fileOrDirectory.substr(fileOrDirectory.lastIndexOf(ts.directorySeparator));
                        lastDirectoryPart = lastDirectoryPartWithDirectorySeparator.slice(ts.directorySeparator.length);
                    }
                    /** Watcher for the file system entry depending on whether it is missing or present */
                    var watcher = !fileSystemEntryExists(fileOrDirectory, entryKind) ?
                        watchMissingFileSystemEntry() :
                        watchPresentFileSystemEntry();
                    return {
                        close: function () {
                            // Close the watcher (either existing file system entry watcher or missing file system entry watcher)
                            watcher.close();
                            watcher = undefined;
                        }
                    };
                    /**
                     * Invoke the callback with rename and update the watcher if not closed
                     * @param createWatcher
                     */
                    function invokeCallbackAndUpdateWatcher(createWatcher) {
                        ts.sysLog("sysLog:: " + fileOrDirectory + ":: Changing watcher to " + (createWatcher === watchPresentFileSystemEntry ? "Present" : "Missing") + "FileSystemEntryWatcher");
                        // Call the callback for current directory
                        callback("rename", "");
                        // If watcher is not closed, update it
                        if (watcher) {
                            watcher.close();
                            watcher = createWatcher();
                        }
                    }
                    /**
                     * Watch the file or directory that is currently present
                     * and when the watched file or directory is deleted, switch to missing file system entry watcher
                     */
                    function watchPresentFileSystemEntry() {
                        // Node 4.0 `fs.watch` function supports the "recursive" option on both OSX and Windows
                        // (ref: https://github.com/nodejs/node/pull/2649 and https://github.com/Microsoft/TypeScript/issues/4643)
                        if (options === undefined) {
                            if (fsSupportsRecursiveFsWatch) {
                                options = { persistent: true, recursive: !!recursive };
                            }
                            else {
                                options = { persistent: true };
                            }
                        }
                        if (hitSystemWatcherLimit) {
                            ts.sysLog("sysLog:: " + fileOrDirectory + ":: Defaulting to fsWatchFile");
                            return watchPresentFileSystemEntryWithFsWatchFile();
                        }
                        try {
                            var presentWatcher = _fs.watch(fileOrDirectory, options, isLinuxOrMacOs ?
                                callbackChangingToMissingFileSystemEntry :
                                callback);
                            // Watch the missing file or directory or error
                            presentWatcher.on("error", function () { return invokeCallbackAndUpdateWatcher(watchMissingFileSystemEntry); });
                            return presentWatcher;
                        }
                        catch (e) {
                            // Catch the exception and use polling instead
                            // Eg. on linux the number of watches are limited and one could easily exhaust watches and the exception ENOSPC is thrown when creating watcher at that point
                            // so instead of throwing error, use fs.watchFile
                            hitSystemWatcherLimit || (hitSystemWatcherLimit = e.code === "ENOSPC");
                            ts.sysLog("sysLog:: " + fileOrDirectory + ":: Changing to fsWatchFile");
                            return watchPresentFileSystemEntryWithFsWatchFile();
                        }
                    }
                    function callbackChangingToMissingFileSystemEntry(event, relativeName) {
                        // because relativeName is not guaranteed to be correct we need to check on each rename with few combinations
                        // Eg on ubuntu while watching app/node_modules the relativeName is "node_modules" which is neither relative nor full path
                        return event === "rename" &&
                            (!relativeName ||
                                relativeName === lastDirectoryPart ||
                                (relativeName.lastIndexOf(lastDirectoryPartWithDirectorySeparator) !== -1 && relativeName.lastIndexOf(lastDirectoryPartWithDirectorySeparator) === relativeName.length - lastDirectoryPartWithDirectorySeparator.length)) &&
                            !fileSystemEntryExists(fileOrDirectory, entryKind) ?
                            invokeCallbackAndUpdateWatcher(watchMissingFileSystemEntry) :
                            callback(event, relativeName);
                    }
                    /**
                     * Watch the file or directory using fs.watchFile since fs.watch threw exception
                     * Eg. on linux the number of watches are limited and one could easily exhaust watches and the exception ENOSPC is thrown when creating watcher at that point
                     */
                    function watchPresentFileSystemEntryWithFsWatchFile() {
                        return watchFile(fileOrDirectory, createFileWatcherCallback(callback), fallbackPollingInterval, fallbackOptions);
                    }
                    /**
                     * Watch the file or directory that is missing
                     * and switch to existing file or directory when the missing filesystem entry is created
                     */
                    function watchMissingFileSystemEntry() {
                        return watchFile(fileOrDirectory, function (_fileName, eventKind) {
                            if (eventKind === FileWatcherEventKind.Created && fileSystemEntryExists(fileOrDirectory, entryKind)) {
                                // Call the callback for current file or directory
                                // For now it could be callback for the inner directory creation,
                                // but just return current directory, better than current no-op
                                invokeCallbackAndUpdateWatcher(watchPresentFileSystemEntry);
                            }
                        }, fallbackPollingInterval, fallbackOptions);
                    }
                }
                function readFileWorker(fileName, _encoding) {
                    var buffer;
                    try {
                        buffer = _fs.readFileSync(fileName);
                    }
                    catch (e) {
                        return undefined;
                    }
                    var len = buffer.length;
                    if (len >= 2 && buffer[0] === 0xFE && buffer[1] === 0xFF) {
                        // Big endian UTF-16 byte order mark detected. Since big endian is not supported by node.js,
                        // flip all byte pairs and treat as little endian.
                        len &= ~1; // Round down to a multiple of 2
                        for (var i = 0; i < len; i += 2) {
                            var temp = buffer[i];
                            buffer[i] = buffer[i + 1];
                            buffer[i + 1] = temp;
                        }
                        return buffer.toString("utf16le", 2);
                    }
                    if (len >= 2 && buffer[0] === 0xFF && buffer[1] === 0xFE) {
                        // Little endian UTF-16 byte order mark detected
                        return buffer.toString("utf16le", 2);
                    }
                    if (len >= 3 && buffer[0] === 0xEF && buffer[1] === 0xBB && buffer[2] === 0xBF) {
                        // UTF-8 byte order mark detected
                        return buffer.toString("utf8", 3);
                    }
                    // Default is UTF-8 with no byte order mark
                    return buffer.toString("utf8");
                }
                function readFile(fileName, _encoding) {
                    ts.perfLogger.logStartReadFile(fileName);
                    var file = readFileWorker(fileName, _encoding);
                    ts.perfLogger.logStopReadFile();
                    return file;
                }
                function writeFile(fileName, data, writeByteOrderMark) {
                    ts.perfLogger.logEvent("WriteFile: " + fileName);
                    // If a BOM is required, emit one
                    if (writeByteOrderMark) {
                        data = byteOrderMarkIndicator + data;
                    }
                    var fd;
                    try {
                        fd = _fs.openSync(fileName, "w");
                        _fs.writeSync(fd, data, /*position*/ undefined, "utf8");
                    }
                    finally {
                        if (fd !== undefined) {
                            _fs.closeSync(fd);
                        }
                    }
                }
                function getAccessibleFileSystemEntries(path) {
                    ts.perfLogger.logEvent("ReadDir: " + (path || "."));
                    try {
                        var entries = _fs.readdirSync(path || ".", { withFileTypes: true });
                        var files = [];
                        var directories = [];
                        for (var _i = 0, entries_1 = entries; _i < entries_1.length; _i++) {
                            var dirent = entries_1[_i];
                            // withFileTypes is not supported before Node 10.10.
                            var entry = typeof dirent === "string" ? dirent : dirent.name;
                            // This is necessary because on some file system node fails to exclude
                            // "." and "..". See https://github.com/nodejs/node/issues/4002
                            if (entry === "." || entry === "..") {
                                continue;
                            }
                            var stat = void 0;
                            if (typeof dirent === "string" || dirent.isSymbolicLink()) {
                                var name = ts.combinePaths(path, entry);
                                try {
                                    stat = statSync(name);
                                    if (!stat) {
                                        continue;
                                    }
                                }
                                catch (e) {
                                    continue;
                                }
                            }
                            else {
                                stat = dirent;
                            }
                            if (stat.isFile()) {
                                files.push(entry);
                            }
                            else if (stat.isDirectory()) {
                                directories.push(entry);
                            }
                        }
                        files.sort();
                        directories.sort();
                        return { files: files, directories: directories };
                    }
                    catch (e) {
                        return ts.emptyFileSystemEntries;
                    }
                }
                function readDirectory(path, extensions, excludes, includes, depth) {
                    return ts.matchFiles(path, extensions, excludes, includes, useCaseSensitiveFileNames, process.cwd(), depth, getAccessibleFileSystemEntries, realpath, directoryExists);
                }
                function fileSystemEntryExists(path, entryKind) {
                    // Since the error thrown by fs.statSync isn't used, we can avoid collecting a stack trace to improve
                    // the CPU time performance.
                    var originalStackTraceLimit = Error.stackTraceLimit;
                    Error.stackTraceLimit = 0;
                    try {
                        var stat = statSync(path);
                        if (!stat) {
                            return false;
                        }
                        switch (entryKind) {
                            case 0 /* File */: return stat.isFile();
                            case 1 /* Directory */: return stat.isDirectory();
                            default: return false;
                        }
                    }
                    catch (e) {
                        return false;
                    }
                    finally {
                        Error.stackTraceLimit = originalStackTraceLimit;
                    }
                }
                function fileExists(path) {
                    return fileSystemEntryExists(path, 0 /* File */);
                }
                function directoryExists(path) {
                    return fileSystemEntryExists(path, 1 /* Directory */);
                }
                function getDirectories(path) {
                    return getAccessibleFileSystemEntries(path).directories.slice();
                }
                function realpath(path) {
                    try {
                        return realpathSync(path);
                    }
                    catch (_a) {
                        return path;
                    }
                }
                function getModifiedTime(path) {
                    var _a;
                    try {
                        return (_a = statSync(path)) === null || _a === void 0 ? void 0 : _a.mtime;
                    }
                    catch (e) {
                        return undefined;
                    }
                }
                function setModifiedTime(path, time) {
                    try {
                        _fs.utimesSync(path, time, time);
                    }
                    catch (e) {
                        return;
                    }
                }
                function deleteFile(path) {
                    try {
                        return _fs.unlinkSync(path);
                    }
                    catch (e) {
                        return;
                    }
                }
                function createSHA256Hash(data) {
                    var hash = _crypto.createHash("sha256");
                    hash.update(data);
                    return hash.digest("hex");
                }
            }
            var sys;
            if (typeof process !== "undefined" && process.nextTick && !process.browser && typeof require !== "undefined") {
                // process and process.nextTick checks if current environment is node-like
                // process.browser check excludes webpack and browserify
                sys = getNodeSystem();
            }
            if (sys) {
                // patch writefile to create folder before writing the file
                patchWriteFileEnsuringDirectory(sys);
            }
            return sys;
        })();
        if (ts.sys && ts.sys.getEnvironmentVariable) {
            setCustomPollingValues(ts.sys);
            ts.Debug.setAssertionLevel(/^development$/i.test(ts.sys.getEnvironmentVariable("NODE_ENV"))
                ? 1 /* Normal */
                : 0 /* None */);
        }
        if (ts.sys && ts.sys.debugMode) {
            ts.Debug.isDebugging = true;
        }
    })(ts || (ts = {}));
    // <auto-generated />
    // generated from './diagnosticInformationMap.generated.ts' by 'src/compiler'
    /* @internal */
    var ts;
    (function (ts) {
        function diag(code, category, key, message, reportsUnnecessary, elidedInCompatabilityPyramid, reportsDeprecated) {
            return { code: code, category: category, key: key, message: message, reportsUnnecessary: reportsUnnecessary, elidedInCompatabilityPyramid: elidedInCompatabilityPyramid, reportsDeprecated: reportsDeprecated };
        }
        ts.Diagnostics = {
            Unterminated_string_literal: diag(1002, ts.DiagnosticCategory.Error, "Unterminated_string_literal_1002", "Unterminated string literal."),
            Identifier_expected: diag(1003, ts.DiagnosticCategory.Error, "Identifier_expected_1003", "Identifier expected."),
            _0_expected: diag(1005, ts.DiagnosticCategory.Error, "_0_expected_1005", "'{0}' expected."),
            A_file_cannot_have_a_reference_to_itself: diag(1006, ts.DiagnosticCategory.Error, "A_file_cannot_have_a_reference_to_itself_1006", "A file cannot have a reference to itself."),
            The_parser_expected_to_find_a_to_match_the_token_here: diag(1007, ts.DiagnosticCategory.Error, "The_parser_expected_to_find_a_to_match_the_token_here_1007", "The parser expected to find a '}' to match the '{' token here."),
            Trailing_comma_not_allowed: diag(1009, ts.DiagnosticCategory.Error, "Trailing_comma_not_allowed_1009", "Trailing comma not allowed."),
            Asterisk_Slash_expected: diag(1010, ts.DiagnosticCategory.Error, "Asterisk_Slash_expected_1010", "'*/' expected."),
            An_element_access_expression_should_take_an_argument: diag(1011, ts.DiagnosticCategory.Error, "An_element_access_expression_should_take_an_argument_1011", "An element access expression should take an argument."),
            Unexpected_token: diag(1012, ts.DiagnosticCategory.Error, "Unexpected_token_1012", "Unexpected token."),
            A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma: diag(1013, ts.DiagnosticCategory.Error, "A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma_1013", "A rest parameter or binding pattern may not have a trailing comma."),
            A_rest_parameter_must_be_last_in_a_parameter_list: diag(1014, ts.DiagnosticCategory.Error, "A_rest_parameter_must_be_last_in_a_parameter_list_1014", "A rest parameter must be last in a parameter list."),
            Parameter_cannot_have_question_mark_and_initializer: diag(1015, ts.DiagnosticCategory.Error, "Parameter_cannot_have_question_mark_and_initializer_1015", "Parameter cannot have question mark and initializer."),
            A_required_parameter_cannot_follow_an_optional_parameter: diag(1016, ts.DiagnosticCategory.Error, "A_required_parameter_cannot_follow_an_optional_parameter_1016", "A required parameter cannot follow an optional parameter."),
            An_index_signature_cannot_have_a_rest_parameter: diag(1017, ts.DiagnosticCategory.Error, "An_index_signature_cannot_have_a_rest_parameter_1017", "An index signature cannot have a rest parameter."),
            An_index_signature_parameter_cannot_have_an_accessibility_modifier: diag(1018, ts.DiagnosticCategory.Error, "An_index_signature_parameter_cannot_have_an_accessibility_modifier_1018", "An index signature parameter cannot have an accessibility modifier."),
            An_index_signature_parameter_cannot_have_a_question_mark: diag(1019, ts.DiagnosticCategory.Error, "An_index_signature_parameter_cannot_have_a_question_mark_1019", "An index signature parameter cannot have a question mark."),
            An_index_signature_parameter_cannot_have_an_initializer: diag(1020, ts.DiagnosticCategory.Error, "An_index_signature_parameter_cannot_have_an_initializer_1020", "An index signature parameter cannot have an initializer."),
            An_index_signature_must_have_a_type_annotation: diag(1021, ts.DiagnosticCategory.Error, "An_index_signature_must_have_a_type_annotation_1021", "An index signature must have a type annotation."),
            An_index_signature_parameter_must_have_a_type_annotation: diag(1022, ts.DiagnosticCategory.Error, "An_index_signature_parameter_must_have_a_type_annotation_1022", "An index signature parameter must have a type annotation."),
            readonly_modifier_can_only_appear_on_a_property_declaration_or_index_signature: diag(1024, ts.DiagnosticCategory.Error, "readonly_modifier_can_only_appear_on_a_property_declaration_or_index_signature_1024", "'readonly' modifier can only appear on a property declaration or index signature."),
            An_index_signature_cannot_have_a_trailing_comma: diag(1025, ts.DiagnosticCategory.Error, "An_index_signature_cannot_have_a_trailing_comma_1025", "An index signature cannot have a trailing comma."),
            Accessibility_modifier_already_seen: diag(1028, ts.DiagnosticCategory.Error, "Accessibility_modifier_already_seen_1028", "Accessibility modifier already seen."),
            _0_modifier_must_precede_1_modifier: diag(1029, ts.DiagnosticCategory.Error, "_0_modifier_must_precede_1_modifier_1029", "'{0}' modifier must precede '{1}' modifier."),
            _0_modifier_already_seen: diag(1030, ts.DiagnosticCategory.Error, "_0_modifier_already_seen_1030", "'{0}' modifier already seen."),
            _0_modifier_cannot_appear_on_class_elements_of_this_kind: diag(1031, ts.DiagnosticCategory.Error, "_0_modifier_cannot_appear_on_class_elements_of_this_kind_1031", "'{0}' modifier cannot appear on class elements of this kind."),
            super_must_be_followed_by_an_argument_list_or_member_access: diag(1034, ts.DiagnosticCategory.Error, "super_must_be_followed_by_an_argument_list_or_member_access_1034", "'super' must be followed by an argument list or member access."),
            Only_ambient_modules_can_use_quoted_names: diag(1035, ts.DiagnosticCategory.Error, "Only_ambient_modules_can_use_quoted_names_1035", "Only ambient modules can use quoted names."),
            Statements_are_not_allowed_in_ambient_contexts: diag(1036, ts.DiagnosticCategory.Error, "Statements_are_not_allowed_in_ambient_contexts_1036", "Statements are not allowed in ambient contexts."),
            A_declare_modifier_cannot_be_used_in_an_already_ambient_context: diag(1038, ts.DiagnosticCategory.Error, "A_declare_modifier_cannot_be_used_in_an_already_ambient_context_1038", "A 'declare' modifier cannot be used in an already ambient context."),
            Initializers_are_not_allowed_in_ambient_contexts: diag(1039, ts.DiagnosticCategory.Error, "Initializers_are_not_allowed_in_ambient_contexts_1039", "Initializers are not allowed in ambient contexts."),
            _0_modifier_cannot_be_used_in_an_ambient_context: diag(1040, ts.DiagnosticCategory.Error, "_0_modifier_cannot_be_used_in_an_ambient_context_1040", "'{0}' modifier cannot be used in an ambient context."),
            _0_modifier_cannot_be_used_with_a_class_declaration: diag(1041, ts.DiagnosticCategory.Error, "_0_modifier_cannot_be_used_with_a_class_declaration_1041", "'{0}' modifier cannot be used with a class declaration."),
            _0_modifier_cannot_be_used_here: diag(1042, ts.DiagnosticCategory.Error, "_0_modifier_cannot_be_used_here_1042", "'{0}' modifier cannot be used here."),
            _0_modifier_cannot_appear_on_a_data_property: diag(1043, ts.DiagnosticCategory.Error, "_0_modifier_cannot_appear_on_a_data_property_1043", "'{0}' modifier cannot appear on a data property."),
            _0_modifier_cannot_appear_on_a_module_or_namespace_element: diag(1044, ts.DiagnosticCategory.Error, "_0_modifier_cannot_appear_on_a_module_or_namespace_element_1044", "'{0}' modifier cannot appear on a module or namespace element."),
            A_0_modifier_cannot_be_used_with_an_interface_declaration: diag(1045, ts.DiagnosticCategory.Error, "A_0_modifier_cannot_be_used_with_an_interface_declaration_1045", "A '{0}' modifier cannot be used with an interface declaration."),
            Top_level_declarations_in_d_ts_files_must_start_with_either_a_declare_or_export_modifier: diag(1046, ts.DiagnosticCategory.Error, "Top_level_declarations_in_d_ts_files_must_start_with_either_a_declare_or_export_modifier_1046", "Top-level declarations in .d.ts files must start with either a 'declare' or 'export' modifier."),
            A_rest_parameter_cannot_be_optional: diag(1047, ts.DiagnosticCategory.Error, "A_rest_parameter_cannot_be_optional_1047", "A rest parameter cannot be optional."),
            A_rest_parameter_cannot_have_an_initializer: diag(1048, ts.DiagnosticCategory.Error, "A_rest_parameter_cannot_have_an_initializer_1048", "A rest parameter cannot have an initializer."),
            A_set_accessor_must_have_exactly_one_parameter: diag(1049, ts.DiagnosticCategory.Error, "A_set_accessor_must_have_exactly_one_parameter_1049", "A 'set' accessor must have exactly one parameter."),
            A_set_accessor_cannot_have_an_optional_parameter: diag(1051, ts.DiagnosticCategory.Error, "A_set_accessor_cannot_have_an_optional_parameter_1051", "A 'set' accessor cannot have an optional parameter."),
            A_set_accessor_parameter_cannot_have_an_initializer: diag(1052, ts.DiagnosticCategory.Error, "A_set_accessor_parameter_cannot_have_an_initializer_1052", "A 'set' accessor parameter cannot have an initializer."),
            A_set_accessor_cannot_have_rest_parameter: diag(1053, ts.DiagnosticCategory.Error, "A_set_accessor_cannot_have_rest_parameter_1053", "A 'set' accessor cannot have rest parameter."),
            A_get_accessor_cannot_have_parameters: diag(1054, ts.DiagnosticCategory.Error, "A_get_accessor_cannot_have_parameters_1054", "A 'get' accessor cannot have parameters."),
            Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value: diag(1055, ts.DiagnosticCategory.Error, "Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Prom_1055", "Type '{0}' is not a valid async function return type in ES5/ES3 because it does not refer to a Promise-compatible constructor value."),
            Accessors_are_only_available_when_targeting_ECMAScript_5_and_higher: diag(1056, ts.DiagnosticCategory.Error, "Accessors_are_only_available_when_targeting_ECMAScript_5_and_higher_1056", "Accessors are only available when targeting ECMAScript 5 and higher."),
            An_async_function_or_method_must_have_a_valid_awaitable_return_type: diag(1057, ts.DiagnosticCategory.Error, "An_async_function_or_method_must_have_a_valid_awaitable_return_type_1057", "An async function or method must have a valid awaitable return type."),
            The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member: diag(1058, ts.DiagnosticCategory.Error, "The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_t_1058", "The return type of an async function must either be a valid promise or must not contain a callable 'then' member."),
            A_promise_must_have_a_then_method: diag(1059, ts.DiagnosticCategory.Error, "A_promise_must_have_a_then_method_1059", "A promise must have a 'then' method."),
            The_first_parameter_of_the_then_method_of_a_promise_must_be_a_callback: diag(1060, ts.DiagnosticCategory.Error, "The_first_parameter_of_the_then_method_of_a_promise_must_be_a_callback_1060", "The first parameter of the 'then' method of a promise must be a callback."),
            Enum_member_must_have_initializer: diag(1061, ts.DiagnosticCategory.Error, "Enum_member_must_have_initializer_1061", "Enum member must have initializer."),
            Type_is_referenced_directly_or_indirectly_in_the_fulfillment_callback_of_its_own_then_method: diag(1062, ts.DiagnosticCategory.Error, "Type_is_referenced_directly_or_indirectly_in_the_fulfillment_callback_of_its_own_then_method_1062", "Type is referenced directly or indirectly in the fulfillment callback of its own 'then' method."),
            An_export_assignment_cannot_be_used_in_a_namespace: diag(1063, ts.DiagnosticCategory.Error, "An_export_assignment_cannot_be_used_in_a_namespace_1063", "An export assignment cannot be used in a namespace."),
            The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type_Did_you_mean_to_write_Promise_0: diag(1064, ts.DiagnosticCategory.Error, "The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type_Did_you_mean_to_wri_1064", "The return type of an async function or method must be the global Promise<T> type. Did you mean to write 'Promise<{0}>'?"),
            In_ambient_enum_declarations_member_initializer_must_be_constant_expression: diag(1066, ts.DiagnosticCategory.Error, "In_ambient_enum_declarations_member_initializer_must_be_constant_expression_1066", "In ambient enum declarations member initializer must be constant expression."),
            Unexpected_token_A_constructor_method_accessor_or_property_was_expected: diag(1068, ts.DiagnosticCategory.Error, "Unexpected_token_A_constructor_method_accessor_or_property_was_expected_1068", "Unexpected token. A constructor, method, accessor, or property was expected."),
            Unexpected_token_A_type_parameter_name_was_expected_without_curly_braces: diag(1069, ts.DiagnosticCategory.Error, "Unexpected_token_A_type_parameter_name_was_expected_without_curly_braces_1069", "Unexpected token. A type parameter name was expected without curly braces."),
            _0_modifier_cannot_appear_on_a_type_member: diag(1070, ts.DiagnosticCategory.Error, "_0_modifier_cannot_appear_on_a_type_member_1070", "'{0}' modifier cannot appear on a type member."),
            _0_modifier_cannot_appear_on_an_index_signature: diag(1071, ts.DiagnosticCategory.Error, "_0_modifier_cannot_appear_on_an_index_signature_1071", "'{0}' modifier cannot appear on an index signature."),
            A_0_modifier_cannot_be_used_with_an_import_declaration: diag(1079, ts.DiagnosticCategory.Error, "A_0_modifier_cannot_be_used_with_an_import_declaration_1079", "A '{0}' modifier cannot be used with an import declaration."),
            Invalid_reference_directive_syntax: diag(1084, ts.DiagnosticCategory.Error, "Invalid_reference_directive_syntax_1084", "Invalid 'reference' directive syntax."),
            Octal_literals_are_not_available_when_targeting_ECMAScript_5_and_higher_Use_the_syntax_0: diag(1085, ts.DiagnosticCategory.Error, "Octal_literals_are_not_available_when_targeting_ECMAScript_5_and_higher_Use_the_syntax_0_1085", "Octal literals are not available when targeting ECMAScript 5 and higher. Use the syntax '{0}'."),
            _0_modifier_cannot_appear_on_a_constructor_declaration: diag(1089, ts.DiagnosticCategory.Error, "_0_modifier_cannot_appear_on_a_constructor_declaration_1089", "'{0}' modifier cannot appear on a constructor declaration."),
            _0_modifier_cannot_appear_on_a_parameter: diag(1090, ts.DiagnosticCategory.Error, "_0_modifier_cannot_appear_on_a_parameter_1090", "'{0}' modifier cannot appear on a parameter."),
            Only_a_single_variable_declaration_is_allowed_in_a_for_in_statement: diag(1091, ts.DiagnosticCategory.Error, "Only_a_single_variable_declaration_is_allowed_in_a_for_in_statement_1091", "Only a single variable declaration is allowed in a 'for...in' statement."),
            Type_parameters_cannot_appear_on_a_constructor_declaration: diag(1092, ts.DiagnosticCategory.Error, "Type_parameters_cannot_appear_on_a_constructor_declaration_1092", "Type parameters cannot appear on a constructor declaration."),
            Type_annotation_cannot_appear_on_a_constructor_declaration: diag(1093, ts.DiagnosticCategory.Error, "Type_annotation_cannot_appear_on_a_constructor_declaration_1093", "Type annotation cannot appear on a constructor declaration."),
            An_accessor_cannot_have_type_parameters: diag(1094, ts.DiagnosticCategory.Error, "An_accessor_cannot_have_type_parameters_1094", "An accessor cannot have type parameters."),
            A_set_accessor_cannot_have_a_return_type_annotation: diag(1095, ts.DiagnosticCategory.Error, "A_set_accessor_cannot_have_a_return_type_annotation_1095", "A 'set' accessor cannot have a return type annotation."),
            An_index_signature_must_have_exactly_one_parameter: diag(1096, ts.DiagnosticCategory.Error, "An_index_signature_must_have_exactly_one_parameter_1096", "An index signature must have exactly one parameter."),
            _0_list_cannot_be_empty: diag(1097, ts.DiagnosticCategory.Error, "_0_list_cannot_be_empty_1097", "'{0}' list cannot be empty."),
            Type_parameter_list_cannot_be_empty: diag(1098, ts.DiagnosticCategory.Error, "Type_parameter_list_cannot_be_empty_1098", "Type parameter list cannot be empty."),
            Type_argument_list_cannot_be_empty: diag(1099, ts.DiagnosticCategory.Error, "Type_argument_list_cannot_be_empty_1099", "Type argument list cannot be empty."),
            Invalid_use_of_0_in_strict_mode: diag(1100, ts.DiagnosticCategory.Error, "Invalid_use_of_0_in_strict_mode_1100", "Invalid use of '{0}' in strict mode."),
            with_statements_are_not_allowed_in_strict_mode: diag(1101, ts.DiagnosticCategory.Error, "with_statements_are_not_allowed_in_strict_mode_1101", "'with' statements are not allowed in strict mode."),
            delete_cannot_be_called_on_an_identifier_in_strict_mode: diag(1102, ts.DiagnosticCategory.Error, "delete_cannot_be_called_on_an_identifier_in_strict_mode_1102", "'delete' cannot be called on an identifier in strict mode."),
            for_await_loops_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules: diag(1103, ts.DiagnosticCategory.Error, "for_await_loops_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules_1103", "'for await' loops are only allowed within async functions and at the top levels of modules."),
            A_continue_statement_can_only_be_used_within_an_enclosing_iteration_statement: diag(1104, ts.DiagnosticCategory.Error, "A_continue_statement_can_only_be_used_within_an_enclosing_iteration_statement_1104", "A 'continue' statement can only be used within an enclosing iteration statement."),
            A_break_statement_can_only_be_used_within_an_enclosing_iteration_or_switch_statement: diag(1105, ts.DiagnosticCategory.Error, "A_break_statement_can_only_be_used_within_an_enclosing_iteration_or_switch_statement_1105", "A 'break' statement can only be used within an enclosing iteration or switch statement."),
            The_left_hand_side_of_a_for_of_statement_may_not_be_async: diag(1106, ts.DiagnosticCategory.Error, "The_left_hand_side_of_a_for_of_statement_may_not_be_async_1106", "The left-hand side of a 'for...of' statement may not be 'async'."),
            Jump_target_cannot_cross_function_boundary: diag(1107, ts.DiagnosticCategory.Error, "Jump_target_cannot_cross_function_boundary_1107", "Jump target cannot cross function boundary."),
            A_return_statement_can_only_be_used_within_a_function_body: diag(1108, ts.DiagnosticCategory.Error, "A_return_statement_can_only_be_used_within_a_function_body_1108", "A 'return' statement can only be used within a function body."),
            Expression_expected: diag(1109, ts.DiagnosticCategory.Error, "Expression_expected_1109", "Expression expected."),
            Type_expected: diag(1110, ts.DiagnosticCategory.Error, "Type_expected_1110", "Type expected."),
            A_default_clause_cannot_appear_more_than_once_in_a_switch_statement: diag(1113, ts.DiagnosticCategory.Error, "A_default_clause_cannot_appear_more_than_once_in_a_switch_statement_1113", "A 'default' clause cannot appear more than once in a 'switch' statement."),
            Duplicate_label_0: diag(1114, ts.DiagnosticCategory.Error, "Duplicate_label_0_1114", "Duplicate label '{0}'."),
            A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement: diag(1115, ts.DiagnosticCategory.Error, "A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement_1115", "A 'continue' statement can only jump to a label of an enclosing iteration statement."),
            A_break_statement_can_only_jump_to_a_label_of_an_enclosing_statement: diag(1116, ts.DiagnosticCategory.Error, "A_break_statement_can_only_jump_to_a_label_of_an_enclosing_statement_1116", "A 'break' statement can only jump to a label of an enclosing statement."),
            An_object_literal_cannot_have_multiple_properties_with_the_same_name_in_strict_mode: diag(1117, ts.DiagnosticCategory.Error, "An_object_literal_cannot_have_multiple_properties_with_the_same_name_in_strict_mode_1117", "An object literal cannot have multiple properties with the same name in strict mode."),
            An_object_literal_cannot_have_multiple_get_Slashset_accessors_with_the_same_name: diag(1118, ts.DiagnosticCategory.Error, "An_object_literal_cannot_have_multiple_get_Slashset_accessors_with_the_same_name_1118", "An object literal cannot have multiple get/set accessors with the same name."),
            An_object_literal_cannot_have_property_and_accessor_with_the_same_name: diag(1119, ts.DiagnosticCategory.Error, "An_object_literal_cannot_have_property_and_accessor_with_the_same_name_1119", "An object literal cannot have property and accessor with the same name."),
            An_export_assignment_cannot_have_modifiers: diag(1120, ts.DiagnosticCategory.Error, "An_export_assignment_cannot_have_modifiers_1120", "An export assignment cannot have modifiers."),
            Octal_literals_are_not_allowed_in_strict_mode: diag(1121, ts.DiagnosticCategory.Error, "Octal_literals_are_not_allowed_in_strict_mode_1121", "Octal literals are not allowed in strict mode."),
            Variable_declaration_list_cannot_be_empty: diag(1123, ts.DiagnosticCategory.Error, "Variable_declaration_list_cannot_be_empty_1123", "Variable declaration list cannot be empty."),
            Digit_expected: diag(1124, ts.DiagnosticCategory.Error, "Digit_expected_1124", "Digit expected."),
            Hexadecimal_digit_expected: diag(1125, ts.DiagnosticCategory.Error, "Hexadecimal_digit_expected_1125", "Hexadecimal digit expected."),
            Unexpected_end_of_text: diag(1126, ts.DiagnosticCategory.Error, "Unexpected_end_of_text_1126", "Unexpected end of text."),
            Invalid_character: diag(1127, ts.DiagnosticCategory.Error, "Invalid_character_1127", "Invalid character."),
            Declaration_or_statement_expected: diag(1128, ts.DiagnosticCategory.Error, "Declaration_or_statement_expected_1128", "Declaration or statement expected."),
            Statement_expected: diag(1129, ts.DiagnosticCategory.Error, "Statement_expected_1129", "Statement expected."),
            case_or_default_expected: diag(1130, ts.DiagnosticCategory.Error, "case_or_default_expected_1130", "'case' or 'default' expected."),
            Property_or_signature_expected: diag(1131, ts.DiagnosticCategory.Error, "Property_or_signature_expected_1131", "Property or signature expected."),
            Enum_member_expected: diag(1132, ts.DiagnosticCategory.Error, "Enum_member_expected_1132", "Enum member expected."),
            Variable_declaration_expected: diag(1134, ts.DiagnosticCategory.Error, "Variable_declaration_expected_1134", "Variable declaration expected."),
            Argument_expression_expected: diag(1135, ts.DiagnosticCategory.Error, "Argument_expression_expected_1135", "Argument expression expected."),
            Property_assignment_expected: diag(1136, ts.DiagnosticCategory.Error, "Property_assignment_expected_1136", "Property assignment expected."),
            Expression_or_comma_expected: diag(1137, ts.DiagnosticCategory.Error, "Expression_or_comma_expected_1137", "Expression or comma expected."),
            Parameter_declaration_expected: diag(1138, ts.DiagnosticCategory.Error, "Parameter_declaration_expected_1138", "Parameter declaration expected."),
            Type_parameter_declaration_expected: diag(1139, ts.DiagnosticCategory.Error, "Type_parameter_declaration_expected_1139", "Type parameter declaration expected."),
            Type_argument_expected: diag(1140, ts.DiagnosticCategory.Error, "Type_argument_expected_1140", "Type argument expected."),
            String_literal_expected: diag(1141, ts.DiagnosticCategory.Error, "String_literal_expected_1141", "String literal expected."),
            Line_break_not_permitted_here: diag(1142, ts.DiagnosticCategory.Error, "Line_break_not_permitted_here_1142", "Line break not permitted here."),
            or_expected: diag(1144, ts.DiagnosticCategory.Error, "or_expected_1144", "'{' or ';' expected."),
            Declaration_expected: diag(1146, ts.DiagnosticCategory.Error, "Declaration_expected_1146", "Declaration expected."),
            Import_declarations_in_a_namespace_cannot_reference_a_module: diag(1147, ts.DiagnosticCategory.Error, "Import_declarations_in_a_namespace_cannot_reference_a_module_1147", "Import declarations in a namespace cannot reference a module."),
            Cannot_use_imports_exports_or_module_augmentations_when_module_is_none: diag(1148, ts.DiagnosticCategory.Error, "Cannot_use_imports_exports_or_module_augmentations_when_module_is_none_1148", "Cannot use imports, exports, or module augmentations when '--module' is 'none'."),
            File_name_0_differs_from_already_included_file_name_1_only_in_casing: diag(1149, ts.DiagnosticCategory.Error, "File_name_0_differs_from_already_included_file_name_1_only_in_casing_1149", "File name '{0}' differs from already included file name '{1}' only in casing."),
            const_declarations_must_be_initialized: diag(1155, ts.DiagnosticCategory.Error, "const_declarations_must_be_initialized_1155", "'const' declarations must be initialized."),
            const_declarations_can_only_be_declared_inside_a_block: diag(1156, ts.DiagnosticCategory.Error, "const_declarations_can_only_be_declared_inside_a_block_1156", "'const' declarations can only be declared inside a block."),
            let_declarations_can_only_be_declared_inside_a_block: diag(1157, ts.DiagnosticCategory.Error, "let_declarations_can_only_be_declared_inside_a_block_1157", "'let' declarations can only be declared inside a block."),
            Unterminated_template_literal: diag(1160, ts.DiagnosticCategory.Error, "Unterminated_template_literal_1160", "Unterminated template literal."),
            Unterminated_regular_expression_literal: diag(1161, ts.DiagnosticCategory.Error, "Unterminated_regular_expression_literal_1161", "Unterminated regular expression literal."),
            An_object_member_cannot_be_declared_optional: diag(1162, ts.DiagnosticCategory.Error, "An_object_member_cannot_be_declared_optional_1162", "An object member cannot be declared optional."),
            A_yield_expression_is_only_allowed_in_a_generator_body: diag(1163, ts.DiagnosticCategory.Error, "A_yield_expression_is_only_allowed_in_a_generator_body_1163", "A 'yield' expression is only allowed in a generator body."),
            Computed_property_names_are_not_allowed_in_enums: diag(1164, ts.DiagnosticCategory.Error, "Computed_property_names_are_not_allowed_in_enums_1164", "Computed property names are not allowed in enums."),
            A_computed_property_name_in_an_ambient_context_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type: diag(1165, ts.DiagnosticCategory.Error, "A_computed_property_name_in_an_ambient_context_must_refer_to_an_expression_whose_type_is_a_literal_t_1165", "A computed property name in an ambient context must refer to an expression whose type is a literal type or a 'unique symbol' type."),
            A_computed_property_name_in_a_class_property_declaration_must_have_a_simple_literal_type_or_a_unique_symbol_type: diag(1166, ts.DiagnosticCategory.Error, "A_computed_property_name_in_a_class_property_declaration_must_have_a_simple_literal_type_or_a_unique_1166", "A computed property name in a class property declaration must have a simple literal type or a 'unique symbol' type."),
            A_computed_property_name_in_a_method_overload_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type: diag(1168, ts.DiagnosticCategory.Error, "A_computed_property_name_in_a_method_overload_must_refer_to_an_expression_whose_type_is_a_literal_ty_1168", "A computed property name in a method overload must refer to an expression whose type is a literal type or a 'unique symbol' type."),
            A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type: diag(1169, ts.DiagnosticCategory.Error, "A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_1169", "A computed property name in an interface must refer to an expression whose type is a literal type or a 'unique symbol' type."),
            A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type: diag(1170, ts.DiagnosticCategory.Error, "A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type__1170", "A computed property name in a type literal must refer to an expression whose type is a literal type or a 'unique symbol' type."),
            A_comma_expression_is_not_allowed_in_a_computed_property_name: diag(1171, ts.DiagnosticCategory.Error, "A_comma_expression_is_not_allowed_in_a_computed_property_name_1171", "A comma expression is not allowed in a computed property name."),
            extends_clause_already_seen: diag(1172, ts.DiagnosticCategory.Error, "extends_clause_already_seen_1172", "'extends' clause already seen."),
            extends_clause_must_precede_implements_clause: diag(1173, ts.DiagnosticCategory.Error, "extends_clause_must_precede_implements_clause_1173", "'extends' clause must precede 'implements' clause."),
            Classes_can_only_extend_a_single_class: diag(1174, ts.DiagnosticCategory.Error, "Classes_can_only_extend_a_single_class_1174", "Classes can only extend a single class."),
            implements_clause_already_seen: diag(1175, ts.DiagnosticCategory.Error, "implements_clause_already_seen_1175", "'implements' clause already seen."),
            Interface_declaration_cannot_have_implements_clause: diag(1176, ts.DiagnosticCategory.Error, "Interface_declaration_cannot_have_implements_clause_1176", "Interface declaration cannot have 'implements' clause."),
            Binary_digit_expected: diag(1177, ts.DiagnosticCategory.Error, "Binary_digit_expected_1177", "Binary digit expected."),
            Octal_digit_expected: diag(1178, ts.DiagnosticCategory.Error, "Octal_digit_expected_1178", "Octal digit expected."),
            Unexpected_token_expected: diag(1179, ts.DiagnosticCategory.Error, "Unexpected_token_expected_1179", "Unexpected token. '{' expected."),
            Property_destructuring_pattern_expected: diag(1180, ts.DiagnosticCategory.Error, "Property_destructuring_pattern_expected_1180", "Property destructuring pattern expected."),
            Array_element_destructuring_pattern_expected: diag(1181, ts.DiagnosticCategory.Error, "Array_element_destructuring_pattern_expected_1181", "Array element destructuring pattern expected."),
            A_destructuring_declaration_must_have_an_initializer: diag(1182, ts.DiagnosticCategory.Error, "A_destructuring_declaration_must_have_an_initializer_1182", "A destructuring declaration must have an initializer."),
            An_implementation_cannot_be_declared_in_ambient_contexts: diag(1183, ts.DiagnosticCategory.Error, "An_implementation_cannot_be_declared_in_ambient_contexts_1183", "An implementation cannot be declared in ambient contexts."),
            Modifiers_cannot_appear_here: diag(1184, ts.DiagnosticCategory.Error, "Modifiers_cannot_appear_here_1184", "Modifiers cannot appear here."),
            Merge_conflict_marker_encountered: diag(1185, ts.DiagnosticCategory.Error, "Merge_conflict_marker_encountered_1185", "Merge conflict marker encountered."),
            A_rest_element_cannot_have_an_initializer: diag(1186, ts.DiagnosticCategory.Error, "A_rest_element_cannot_have_an_initializer_1186", "A rest element cannot have an initializer."),
            A_parameter_property_may_not_be_declared_using_a_binding_pattern: diag(1187, ts.DiagnosticCategory.Error, "A_parameter_property_may_not_be_declared_using_a_binding_pattern_1187", "A parameter property may not be declared using a binding pattern."),
            Only_a_single_variable_declaration_is_allowed_in_a_for_of_statement: diag(1188, ts.DiagnosticCategory.Error, "Only_a_single_variable_declaration_is_allowed_in_a_for_of_statement_1188", "Only a single variable declaration is allowed in a 'for...of' statement."),
            The_variable_declaration_of_a_for_in_statement_cannot_have_an_initializer: diag(1189, ts.DiagnosticCategory.Error, "The_variable_declaration_of_a_for_in_statement_cannot_have_an_initializer_1189", "The variable declaration of a 'for...in' statement cannot have an initializer."),
            The_variable_declaration_of_a_for_of_statement_cannot_have_an_initializer: diag(1190, ts.DiagnosticCategory.Error, "The_variable_declaration_of_a_for_of_statement_cannot_have_an_initializer_1190", "The variable declaration of a 'for...of' statement cannot have an initializer."),
            An_import_declaration_cannot_have_modifiers: diag(1191, ts.DiagnosticCategory.Error, "An_import_declaration_cannot_have_modifiers_1191", "An import declaration cannot have modifiers."),
            Module_0_has_no_default_export: diag(1192, ts.DiagnosticCategory.Error, "Module_0_has_no_default_export_1192", "Module '{0}' has no default export."),
            An_export_declaration_cannot_have_modifiers: diag(1193, ts.DiagnosticCategory.Error, "An_export_declaration_cannot_have_modifiers_1193", "An export declaration cannot have modifiers."),
            Export_declarations_are_not_permitted_in_a_namespace: diag(1194, ts.DiagnosticCategory.Error, "Export_declarations_are_not_permitted_in_a_namespace_1194", "Export declarations are not permitted in a namespace."),
            export_Asterisk_does_not_re_export_a_default: diag(1195, ts.DiagnosticCategory.Error, "export_Asterisk_does_not_re_export_a_default_1195", "'export *' does not re-export a default."),
            Catch_clause_variable_type_annotation_must_be_any_or_unknown_if_specified: diag(1196, ts.DiagnosticCategory.Error, "Catch_clause_variable_type_annotation_must_be_any_or_unknown_if_specified_1196", "Catch clause variable type annotation must be 'any' or 'unknown' if specified."),
            Catch_clause_variable_cannot_have_an_initializer: diag(1197, ts.DiagnosticCategory.Error, "Catch_clause_variable_cannot_have_an_initializer_1197", "Catch clause variable cannot have an initializer."),
            An_extended_Unicode_escape_value_must_be_between_0x0_and_0x10FFFF_inclusive: diag(1198, ts.DiagnosticCategory.Error, "An_extended_Unicode_escape_value_must_be_between_0x0_and_0x10FFFF_inclusive_1198", "An extended Unicode escape value must be between 0x0 and 0x10FFFF inclusive."),
            Unterminated_Unicode_escape_sequence: diag(1199, ts.DiagnosticCategory.Error, "Unterminated_Unicode_escape_sequence_1199", "Unterminated Unicode escape sequence."),
            Line_terminator_not_permitted_before_arrow: diag(1200, ts.DiagnosticCategory.Error, "Line_terminator_not_permitted_before_arrow_1200", "Line terminator not permitted before arrow."),
            Import_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_import_Asterisk_as_ns_from_mod_import_a_from_mod_import_d_from_mod_or_another_module_format_instead: diag(1202, ts.DiagnosticCategory.Error, "Import_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_import_Asterisk_as_1202", "Import assignment cannot be used when targeting ECMAScript modules. Consider using 'import * as ns from \"mod\"', 'import {a} from \"mod\"', 'import d from \"mod\"', or another module format instead."),
            Export_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_export_default_or_another_module_format_instead: diag(1203, ts.DiagnosticCategory.Error, "Export_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_export_default_or__1203", "Export assignment cannot be used when targeting ECMAScript modules. Consider using 'export default' or another module format instead."),
            Re_exporting_a_type_when_the_isolatedModules_flag_is_provided_requires_using_export_type: diag(1205, ts.DiagnosticCategory.Error, "Re_exporting_a_type_when_the_isolatedModules_flag_is_provided_requires_using_export_type_1205", "Re-exporting a type when the '--isolatedModules' flag is provided requires using 'export type'."),
            Decorators_are_not_valid_here: diag(1206, ts.DiagnosticCategory.Error, "Decorators_are_not_valid_here_1206", "Decorators are not valid here."),
            Decorators_cannot_be_applied_to_multiple_get_Slashset_accessors_of_the_same_name: diag(1207, ts.DiagnosticCategory.Error, "Decorators_cannot_be_applied_to_multiple_get_Slashset_accessors_of_the_same_name_1207", "Decorators cannot be applied to multiple get/set accessors of the same name."),
            _0_cannot_be_compiled_under_isolatedModules_because_it_is_considered_a_global_script_file_Add_an_import_export_or_an_empty_export_statement_to_make_it_a_module: diag(1208, ts.DiagnosticCategory.Error, "_0_cannot_be_compiled_under_isolatedModules_because_it_is_considered_a_global_script_file_Add_an_imp_1208", "'{0}' cannot be compiled under '--isolatedModules' because it is considered a global script file. Add an import, export, or an empty 'export {}' statement to make it a module."),
            Code_contained_in_a_class_is_evaluated_in_JavaScript_s_strict_mode_which_does_not_allow_this_use_of_0_For_more_information_see_https_Colon_Slash_Slashdeveloper_mozilla_org_Slashen_US_Slashdocs_SlashWeb_SlashJavaScript_SlashReference_SlashStrict_mode: diag(1210, ts.DiagnosticCategory.Error, "Code_contained_in_a_class_is_evaluated_in_JavaScript_s_strict_mode_which_does_not_allow_this_use_of__1210", "Code contained in a class is evaluated in JavaScript's strict mode which does not allow this use of '{0}'. For more information, see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Strict_mode."),
            A_class_declaration_without_the_default_modifier_must_have_a_name: diag(1211, ts.DiagnosticCategory.Error, "A_class_declaration_without_the_default_modifier_must_have_a_name_1211", "A class declaration without the 'default' modifier must have a name."),
            Identifier_expected_0_is_a_reserved_word_in_strict_mode: diag(1212, ts.DiagnosticCategory.Error, "Identifier_expected_0_is_a_reserved_word_in_strict_mode_1212", "Identifier expected. '{0}' is a reserved word in strict mode."),
            Identifier_expected_0_is_a_reserved_word_in_strict_mode_Class_definitions_are_automatically_in_strict_mode: diag(1213, ts.DiagnosticCategory.Error, "Identifier_expected_0_is_a_reserved_word_in_strict_mode_Class_definitions_are_automatically_in_stric_1213", "Identifier expected. '{0}' is a reserved word in strict mode. Class definitions are automatically in strict mode."),
            Identifier_expected_0_is_a_reserved_word_in_strict_mode_Modules_are_automatically_in_strict_mode: diag(1214, ts.DiagnosticCategory.Error, "Identifier_expected_0_is_a_reserved_word_in_strict_mode_Modules_are_automatically_in_strict_mode_1214", "Identifier expected. '{0}' is a reserved word in strict mode. Modules are automatically in strict mode."),
            Invalid_use_of_0_Modules_are_automatically_in_strict_mode: diag(1215, ts.DiagnosticCategory.Error, "Invalid_use_of_0_Modules_are_automatically_in_strict_mode_1215", "Invalid use of '{0}'. Modules are automatically in strict mode."),
            Identifier_expected_esModule_is_reserved_as_an_exported_marker_when_transforming_ECMAScript_modules: diag(1216, ts.DiagnosticCategory.Error, "Identifier_expected_esModule_is_reserved_as_an_exported_marker_when_transforming_ECMAScript_modules_1216", "Identifier expected. '__esModule' is reserved as an exported marker when transforming ECMAScript modules."),
            Export_assignment_is_not_supported_when_module_flag_is_system: diag(1218, ts.DiagnosticCategory.Error, "Export_assignment_is_not_supported_when_module_flag_is_system_1218", "Export assignment is not supported when '--module' flag is 'system'."),
            Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_in_your_tsconfig_or_jsconfig_to_remove_this_warning: diag(1219, ts.DiagnosticCategory.Error, "Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_t_1219", "Experimental support for decorators is a feature that is subject to change in a future release. Set the 'experimentalDecorators' option in your 'tsconfig' or 'jsconfig' to remove this warning."),
            Generators_are_only_available_when_targeting_ECMAScript_2015_or_higher: diag(1220, ts.DiagnosticCategory.Error, "Generators_are_only_available_when_targeting_ECMAScript_2015_or_higher_1220", "Generators are only available when targeting ECMAScript 2015 or higher."),
            Generators_are_not_allowed_in_an_ambient_context: diag(1221, ts.DiagnosticCategory.Error, "Generators_are_not_allowed_in_an_ambient_context_1221", "Generators are not allowed in an ambient context."),
            An_overload_signature_cannot_be_declared_as_a_generator: diag(1222, ts.DiagnosticCategory.Error, "An_overload_signature_cannot_be_declared_as_a_generator_1222", "An overload signature cannot be declared as a generator."),
            _0_tag_already_specified: diag(1223, ts.DiagnosticCategory.Error, "_0_tag_already_specified_1223", "'{0}' tag already specified."),
            Signature_0_must_be_a_type_predicate: diag(1224, ts.DiagnosticCategory.Error, "Signature_0_must_be_a_type_predicate_1224", "Signature '{0}' must be a type predicate."),
            Cannot_find_parameter_0: diag(1225, ts.DiagnosticCategory.Error, "Cannot_find_parameter_0_1225", "Cannot find parameter '{0}'."),
            Type_predicate_0_is_not_assignable_to_1: diag(1226, ts.DiagnosticCategory.Error, "Type_predicate_0_is_not_assignable_to_1_1226", "Type predicate '{0}' is not assignable to '{1}'."),
            Parameter_0_is_not_in_the_same_position_as_parameter_1: diag(1227, ts.DiagnosticCategory.Error, "Parameter_0_is_not_in_the_same_position_as_parameter_1_1227", "Parameter '{0}' is not in the same position as parameter '{1}'."),
            A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods: diag(1228, ts.DiagnosticCategory.Error, "A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods_1228", "A type predicate is only allowed in return type position for functions and methods."),
            A_type_predicate_cannot_reference_a_rest_parameter: diag(1229, ts.DiagnosticCategory.Error, "A_type_predicate_cannot_reference_a_rest_parameter_1229", "A type predicate cannot reference a rest parameter."),
            A_type_predicate_cannot_reference_element_0_in_a_binding_pattern: diag(1230, ts.DiagnosticCategory.Error, "A_type_predicate_cannot_reference_element_0_in_a_binding_pattern_1230", "A type predicate cannot reference element '{0}' in a binding pattern."),
            An_export_assignment_must_be_at_the_top_level_of_a_file_or_module_declaration: diag(1231, ts.DiagnosticCategory.Error, "An_export_assignment_must_be_at_the_top_level_of_a_file_or_module_declaration_1231", "An export assignment must be at the top level of a file or module declaration."),
            An_import_declaration_can_only_be_used_in_a_namespace_or_module: diag(1232, ts.DiagnosticCategory.Error, "An_import_declaration_can_only_be_used_in_a_namespace_or_module_1232", "An import declaration can only be used in a namespace or module."),
            An_export_declaration_can_only_be_used_in_a_module: diag(1233, ts.DiagnosticCategory.Error, "An_export_declaration_can_only_be_used_in_a_module_1233", "An export declaration can only be used in a module."),
            An_ambient_module_declaration_is_only_allowed_at_the_top_level_in_a_file: diag(1234, ts.DiagnosticCategory.Error, "An_ambient_module_declaration_is_only_allowed_at_the_top_level_in_a_file_1234", "An ambient module declaration is only allowed at the top level in a file."),
            A_namespace_declaration_is_only_allowed_in_a_namespace_or_module: diag(1235, ts.DiagnosticCategory.Error, "A_namespace_declaration_is_only_allowed_in_a_namespace_or_module_1235", "A namespace declaration is only allowed in a namespace or module."),
            The_return_type_of_a_property_decorator_function_must_be_either_void_or_any: diag(1236, ts.DiagnosticCategory.Error, "The_return_type_of_a_property_decorator_function_must_be_either_void_or_any_1236", "The return type of a property decorator function must be either 'void' or 'any'."),
            The_return_type_of_a_parameter_decorator_function_must_be_either_void_or_any: diag(1237, ts.DiagnosticCategory.Error, "The_return_type_of_a_parameter_decorator_function_must_be_either_void_or_any_1237", "The return type of a parameter decorator function must be either 'void' or 'any'."),
            Unable_to_resolve_signature_of_class_decorator_when_called_as_an_expression: diag(1238, ts.DiagnosticCategory.Error, "Unable_to_resolve_signature_of_class_decorator_when_called_as_an_expression_1238", "Unable to resolve signature of class decorator when called as an expression."),
            Unable_to_resolve_signature_of_parameter_decorator_when_called_as_an_expression: diag(1239, ts.DiagnosticCategory.Error, "Unable_to_resolve_signature_of_parameter_decorator_when_called_as_an_expression_1239", "Unable to resolve signature of parameter decorator when called as an expression."),
            Unable_to_resolve_signature_of_property_decorator_when_called_as_an_expression: diag(1240, ts.DiagnosticCategory.Error, "Unable_to_resolve_signature_of_property_decorator_when_called_as_an_expression_1240", "Unable to resolve signature of property decorator when called as an expression."),
            Unable_to_resolve_signature_of_method_decorator_when_called_as_an_expression: diag(1241, ts.DiagnosticCategory.Error, "Unable_to_resolve_signature_of_method_decorator_when_called_as_an_expression_1241", "Unable to resolve signature of method decorator when called as an expression."),
            abstract_modifier_can_only_appear_on_a_class_method_or_property_declaration: diag(1242, ts.DiagnosticCategory.Error, "abstract_modifier_can_only_appear_on_a_class_method_or_property_declaration_1242", "'abstract' modifier can only appear on a class, method, or property declaration."),
            _0_modifier_cannot_be_used_with_1_modifier: diag(1243, ts.DiagnosticCategory.Error, "_0_modifier_cannot_be_used_with_1_modifier_1243", "'{0}' modifier cannot be used with '{1}' modifier."),
            Abstract_methods_can_only_appear_within_an_abstract_class: diag(1244, ts.DiagnosticCategory.Error, "Abstract_methods_can_only_appear_within_an_abstract_class_1244", "Abstract methods can only appear within an abstract class."),
            Method_0_cannot_have_an_implementation_because_it_is_marked_abstract: diag(1245, ts.DiagnosticCategory.Error, "Method_0_cannot_have_an_implementation_because_it_is_marked_abstract_1245", "Method '{0}' cannot have an implementation because it is marked abstract."),
            An_interface_property_cannot_have_an_initializer: diag(1246, ts.DiagnosticCategory.Error, "An_interface_property_cannot_have_an_initializer_1246", "An interface property cannot have an initializer."),
            A_type_literal_property_cannot_have_an_initializer: diag(1247, ts.DiagnosticCategory.Error, "A_type_literal_property_cannot_have_an_initializer_1247", "A type literal property cannot have an initializer."),
            A_class_member_cannot_have_the_0_keyword: diag(1248, ts.DiagnosticCategory.Error, "A_class_member_cannot_have_the_0_keyword_1248", "A class member cannot have the '{0}' keyword."),
            A_decorator_can_only_decorate_a_method_implementation_not_an_overload: diag(1249, ts.DiagnosticCategory.Error, "A_decorator_can_only_decorate_a_method_implementation_not_an_overload_1249", "A decorator can only decorate a method implementation, not an overload."),
            Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5: diag(1250, ts.DiagnosticCategory.Error, "Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_1250", "Function declarations are not allowed inside blocks in strict mode when targeting 'ES3' or 'ES5'."),
            Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Class_definitions_are_automatically_in_strict_mode: diag(1251, ts.DiagnosticCategory.Error, "Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Class_d_1251", "Function declarations are not allowed inside blocks in strict mode when targeting 'ES3' or 'ES5'. Class definitions are automatically in strict mode."),
            Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Modules_are_automatically_in_strict_mode: diag(1252, ts.DiagnosticCategory.Error, "Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Modules_1252", "Function declarations are not allowed inside blocks in strict mode when targeting 'ES3' or 'ES5'. Modules are automatically in strict mode."),
            _0_tag_cannot_be_used_independently_as_a_top_level_JSDoc_tag: diag(1253, ts.DiagnosticCategory.Error, "_0_tag_cannot_be_used_independently_as_a_top_level_JSDoc_tag_1253", "'{0}' tag cannot be used independently as a top level JSDoc tag."),
            A_const_initializer_in_an_ambient_context_must_be_a_string_or_numeric_literal_or_literal_enum_reference: diag(1254, ts.DiagnosticCategory.Error, "A_const_initializer_in_an_ambient_context_must_be_a_string_or_numeric_literal_or_literal_enum_refere_1254", "A 'const' initializer in an ambient context must be a string or numeric literal or literal enum reference."),
            A_definite_assignment_assertion_is_not_permitted_in_this_context: diag(1255, ts.DiagnosticCategory.Error, "A_definite_assignment_assertion_is_not_permitted_in_this_context_1255", "A definite assignment assertion '!' is not permitted in this context."),
            A_required_element_cannot_follow_an_optional_element: diag(1257, ts.DiagnosticCategory.Error, "A_required_element_cannot_follow_an_optional_element_1257", "A required element cannot follow an optional element."),
            A_default_export_must_be_at_the_top_level_of_a_file_or_module_declaration: diag(1258, ts.DiagnosticCategory.Error, "A_default_export_must_be_at_the_top_level_of_a_file_or_module_declaration_1258", "A default export must be at the top level of a file or module declaration."),
            Module_0_can_only_be_default_imported_using_the_1_flag: diag(1259, ts.DiagnosticCategory.Error, "Module_0_can_only_be_default_imported_using_the_1_flag_1259", "Module '{0}' can only be default-imported using the '{1}' flag"),
            Keywords_cannot_contain_escape_characters: diag(1260, ts.DiagnosticCategory.Error, "Keywords_cannot_contain_escape_characters_1260", "Keywords cannot contain escape characters."),
            Already_included_file_name_0_differs_from_file_name_1_only_in_casing: diag(1261, ts.DiagnosticCategory.Error, "Already_included_file_name_0_differs_from_file_name_1_only_in_casing_1261", "Already included file name '{0}' differs from file name '{1}' only in casing."),
            Identifier_expected_0_is_a_reserved_word_at_the_top_level_of_a_module: diag(1262, ts.DiagnosticCategory.Error, "Identifier_expected_0_is_a_reserved_word_at_the_top_level_of_a_module_1262", "Identifier expected. '{0}' is a reserved word at the top-level of a module."),
            Declarations_with_initializers_cannot_also_have_definite_assignment_assertions: diag(1263, ts.DiagnosticCategory.Error, "Declarations_with_initializers_cannot_also_have_definite_assignment_assertions_1263", "Declarations with initializers cannot also have definite assignment assertions."),
            Declarations_with_definite_assignment_assertions_must_also_have_type_annotations: diag(1264, ts.DiagnosticCategory.Error, "Declarations_with_definite_assignment_assertions_must_also_have_type_annotations_1264", "Declarations with definite assignment assertions must also have type annotations."),
            A_rest_element_cannot_follow_another_rest_element: diag(1265, ts.DiagnosticCategory.Error, "A_rest_element_cannot_follow_another_rest_element_1265", "A rest element cannot follow another rest element."),
            An_optional_element_cannot_follow_a_rest_element: diag(1266, ts.DiagnosticCategory.Error, "An_optional_element_cannot_follow_a_rest_element_1266", "An optional element cannot follow a rest element."),
            Property_0_cannot_have_an_initializer_because_it_is_marked_abstract: diag(1267, ts.DiagnosticCategory.Error, "Property_0_cannot_have_an_initializer_because_it_is_marked_abstract_1267", "Property '{0}' cannot have an initializer because it is marked abstract."),
            An_index_signature_parameter_type_must_be_string_number_symbol_or_a_template_literal_type: diag(1268, ts.DiagnosticCategory.Error, "An_index_signature_parameter_type_must_be_string_number_symbol_or_a_template_literal_type_1268", "An index signature parameter type must be 'string', 'number', 'symbol', or a template literal type."),
            with_statements_are_not_allowed_in_an_async_function_block: diag(1300, ts.DiagnosticCategory.Error, "with_statements_are_not_allowed_in_an_async_function_block_1300", "'with' statements are not allowed in an async function block."),
            await_expressions_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules: diag(1308, ts.DiagnosticCategory.Error, "await_expressions_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules_1308", "'await' expressions are only allowed within async functions and at the top levels of modules."),
            Did_you_mean_to_use_a_Colon_An_can_only_follow_a_property_name_when_the_containing_object_literal_is_part_of_a_destructuring_pattern: diag(1312, ts.DiagnosticCategory.Error, "Did_you_mean_to_use_a_Colon_An_can_only_follow_a_property_name_when_the_containing_object_literal_is_1312", "Did you mean to use a ':'? An '=' can only follow a property name when the containing object literal is part of a destructuring pattern."),
            The_body_of_an_if_statement_cannot_be_the_empty_statement: diag(1313, ts.DiagnosticCategory.Error, "The_body_of_an_if_statement_cannot_be_the_empty_statement_1313", "The body of an 'if' statement cannot be the empty statement."),
            Global_module_exports_may_only_appear_in_module_files: diag(1314, ts.DiagnosticCategory.Error, "Global_module_exports_may_only_appear_in_module_files_1314", "Global module exports may only appear in module files."),
            Global_module_exports_may_only_appear_in_declaration_files: diag(1315, ts.DiagnosticCategory.Error, "Global_module_exports_may_only_appear_in_declaration_files_1315", "Global module exports may only appear in declaration files."),
            Global_module_exports_may_only_appear_at_top_level: diag(1316, ts.DiagnosticCategory.Error, "Global_module_exports_may_only_appear_at_top_level_1316", "Global module exports may only appear at top level."),
            A_parameter_property_cannot_be_declared_using_a_rest_parameter: diag(1317, ts.DiagnosticCategory.Error, "A_parameter_property_cannot_be_declared_using_a_rest_parameter_1317", "A parameter property cannot be declared using a rest parameter."),
            An_abstract_accessor_cannot_have_an_implementation: diag(1318, ts.DiagnosticCategory.Error, "An_abstract_accessor_cannot_have_an_implementation_1318", "An abstract accessor cannot have an implementation."),
            A_default_export_can_only_be_used_in_an_ECMAScript_style_module: diag(1319, ts.DiagnosticCategory.Error, "A_default_export_can_only_be_used_in_an_ECMAScript_style_module_1319", "A default export can only be used in an ECMAScript-style module."),
            Type_of_await_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member: diag(1320, ts.DiagnosticCategory.Error, "Type_of_await_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member_1320", "Type of 'await' operand must either be a valid promise or must not contain a callable 'then' member."),
            Type_of_yield_operand_in_an_async_generator_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member: diag(1321, ts.DiagnosticCategory.Error, "Type_of_yield_operand_in_an_async_generator_must_either_be_a_valid_promise_or_must_not_contain_a_cal_1321", "Type of 'yield' operand in an async generator must either be a valid promise or must not contain a callable 'then' member."),
            Type_of_iterated_elements_of_a_yield_Asterisk_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member: diag(1322, ts.DiagnosticCategory.Error, "Type_of_iterated_elements_of_a_yield_Asterisk_operand_must_either_be_a_valid_promise_or_must_not_con_1322", "Type of iterated elements of a 'yield*' operand must either be a valid promise or must not contain a callable 'then' member."),
            Dynamic_imports_are_only_supported_when_the_module_flag_is_set_to_es2020_esnext_commonjs_amd_system_or_umd: diag(1323, ts.DiagnosticCategory.Error, "Dynamic_imports_are_only_supported_when_the_module_flag_is_set_to_es2020_esnext_commonjs_amd_system__1323", "Dynamic imports are only supported when the '--module' flag is set to 'es2020', 'esnext', 'commonjs', 'amd', 'system', or 'umd'."),
            Dynamic_import_must_have_one_specifier_as_an_argument: diag(1324, ts.DiagnosticCategory.Error, "Dynamic_import_must_have_one_specifier_as_an_argument_1324", "Dynamic import must have one specifier as an argument."),
            Specifier_of_dynamic_import_cannot_be_spread_element: diag(1325, ts.DiagnosticCategory.Error, "Specifier_of_dynamic_import_cannot_be_spread_element_1325", "Specifier of dynamic import cannot be spread element."),
            Dynamic_import_cannot_have_type_arguments: diag(1326, ts.DiagnosticCategory.Error, "Dynamic_import_cannot_have_type_arguments_1326", "Dynamic import cannot have type arguments."),
            String_literal_with_double_quotes_expected: diag(1327, ts.DiagnosticCategory.Error, "String_literal_with_double_quotes_expected_1327", "String literal with double quotes expected."),
            Property_value_can_only_be_string_literal_numeric_literal_true_false_null_object_literal_or_array_literal: diag(1328, ts.DiagnosticCategory.Error, "Property_value_can_only_be_string_literal_numeric_literal_true_false_null_object_literal_or_array_li_1328", "Property value can only be string literal, numeric literal, 'true', 'false', 'null', object literal or array literal."),
            _0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write_0: diag(1329, ts.DiagnosticCategory.Error, "_0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write__1329", "'{0}' accepts too few arguments to be used as a decorator here. Did you mean to call it first and write '@{0}()'?"),
            A_property_of_an_interface_or_type_literal_whose_type_is_a_unique_symbol_type_must_be_readonly: diag(1330, ts.DiagnosticCategory.Error, "A_property_of_an_interface_or_type_literal_whose_type_is_a_unique_symbol_type_must_be_readonly_1330", "A property of an interface or type literal whose type is a 'unique symbol' type must be 'readonly'."),
            A_property_of_a_class_whose_type_is_a_unique_symbol_type_must_be_both_static_and_readonly: diag(1331, ts.DiagnosticCategory.Error, "A_property_of_a_class_whose_type_is_a_unique_symbol_type_must_be_both_static_and_readonly_1331", "A property of a class whose type is a 'unique symbol' type must be both 'static' and 'readonly'."),
            A_variable_whose_type_is_a_unique_symbol_type_must_be_const: diag(1332, ts.DiagnosticCategory.Error, "A_variable_whose_type_is_a_unique_symbol_type_must_be_const_1332", "A variable whose type is a 'unique symbol' type must be 'const'."),
            unique_symbol_types_may_not_be_used_on_a_variable_declaration_with_a_binding_name: diag(1333, ts.DiagnosticCategory.Error, "unique_symbol_types_may_not_be_used_on_a_variable_declaration_with_a_binding_name_1333", "'unique symbol' types may not be used on a variable declaration with a binding name."),
            unique_symbol_types_are_only_allowed_on_variables_in_a_variable_statement: diag(1334, ts.DiagnosticCategory.Error, "unique_symbol_types_are_only_allowed_on_variables_in_a_variable_statement_1334", "'unique symbol' types are only allowed on variables in a variable statement."),
            unique_symbol_types_are_not_allowed_here: diag(1335, ts.DiagnosticCategory.Error, "unique_symbol_types_are_not_allowed_here_1335", "'unique symbol' types are not allowed here."),
            An_index_signature_parameter_type_cannot_be_a_literal_type_or_generic_type_Consider_using_a_mapped_object_type_instead: diag(1337, ts.DiagnosticCategory.Error, "An_index_signature_parameter_type_cannot_be_a_literal_type_or_generic_type_Consider_using_a_mapped_o_1337", "An index signature parameter type cannot be a literal type or generic type. Consider using a mapped object type instead."),
            infer_declarations_are_only_permitted_in_the_extends_clause_of_a_conditional_type: diag(1338, ts.DiagnosticCategory.Error, "infer_declarations_are_only_permitted_in_the_extends_clause_of_a_conditional_type_1338", "'infer' declarations are only permitted in the 'extends' clause of a conditional type."),
            Module_0_does_not_refer_to_a_value_but_is_used_as_a_value_here: diag(1339, ts.DiagnosticCategory.Error, "Module_0_does_not_refer_to_a_value_but_is_used_as_a_value_here_1339", "Module '{0}' does not refer to a value, but is used as a value here."),
            Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0: diag(1340, ts.DiagnosticCategory.Error, "Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0_1340", "Module '{0}' does not refer to a type, but is used as a type here. Did you mean 'typeof import('{0}')'?"),
            Type_arguments_cannot_be_used_here: diag(1342, ts.DiagnosticCategory.Error, "Type_arguments_cannot_be_used_here_1342", "Type arguments cannot be used here."),
            The_import_meta_meta_property_is_only_allowed_when_the_module_option_is_es2020_esnext_or_system: diag(1343, ts.DiagnosticCategory.Error, "The_import_meta_meta_property_is_only_allowed_when_the_module_option_is_es2020_esnext_or_system_1343", "The 'import.meta' meta-property is only allowed when the '--module' option is 'es2020', 'esnext', or 'system'."),
            A_label_is_not_allowed_here: diag(1344, ts.DiagnosticCategory.Error, "A_label_is_not_allowed_here_1344", "'A label is not allowed here."),
            An_expression_of_type_void_cannot_be_tested_for_truthiness: diag(1345, ts.DiagnosticCategory.Error, "An_expression_of_type_void_cannot_be_tested_for_truthiness_1345", "An expression of type 'void' cannot be tested for truthiness."),
            This_parameter_is_not_allowed_with_use_strict_directive: diag(1346, ts.DiagnosticCategory.Error, "This_parameter_is_not_allowed_with_use_strict_directive_1346", "This parameter is not allowed with 'use strict' directive."),
            use_strict_directive_cannot_be_used_with_non_simple_parameter_list: diag(1347, ts.DiagnosticCategory.Error, "use_strict_directive_cannot_be_used_with_non_simple_parameter_list_1347", "'use strict' directive cannot be used with non-simple parameter list."),
            Non_simple_parameter_declared_here: diag(1348, ts.DiagnosticCategory.Error, "Non_simple_parameter_declared_here_1348", "Non-simple parameter declared here."),
            use_strict_directive_used_here: diag(1349, ts.DiagnosticCategory.Error, "use_strict_directive_used_here_1349", "'use strict' directive used here."),
            Print_the_final_configuration_instead_of_building: diag(1350, ts.DiagnosticCategory.Message, "Print_the_final_configuration_instead_of_building_1350", "Print the final configuration instead of building."),
            An_identifier_or_keyword_cannot_immediately_follow_a_numeric_literal: diag(1351, ts.DiagnosticCategory.Error, "An_identifier_or_keyword_cannot_immediately_follow_a_numeric_literal_1351", "An identifier or keyword cannot immediately follow a numeric literal."),
            A_bigint_literal_cannot_use_exponential_notation: diag(1352, ts.DiagnosticCategory.Error, "A_bigint_literal_cannot_use_exponential_notation_1352", "A bigint literal cannot use exponential notation."),
            A_bigint_literal_must_be_an_integer: diag(1353, ts.DiagnosticCategory.Error, "A_bigint_literal_must_be_an_integer_1353", "A bigint literal must be an integer."),
            readonly_type_modifier_is_only_permitted_on_array_and_tuple_literal_types: diag(1354, ts.DiagnosticCategory.Error, "readonly_type_modifier_is_only_permitted_on_array_and_tuple_literal_types_1354", "'readonly' type modifier is only permitted on array and tuple literal types."),
            A_const_assertions_can_only_be_applied_to_references_to_enum_members_or_string_number_boolean_array_or_object_literals: diag(1355, ts.DiagnosticCategory.Error, "A_const_assertions_can_only_be_applied_to_references_to_enum_members_or_string_number_boolean_array__1355", "A 'const' assertions can only be applied to references to enum members, or string, number, boolean, array, or object literals."),
            Did_you_mean_to_mark_this_function_as_async: diag(1356, ts.DiagnosticCategory.Error, "Did_you_mean_to_mark_this_function_as_async_1356", "Did you mean to mark this function as 'async'?"),
            An_enum_member_name_must_be_followed_by_a_or: diag(1357, ts.DiagnosticCategory.Error, "An_enum_member_name_must_be_followed_by_a_or_1357", "An enum member name must be followed by a ',', '=', or '}'."),
            Tagged_template_expressions_are_not_permitted_in_an_optional_chain: diag(1358, ts.DiagnosticCategory.Error, "Tagged_template_expressions_are_not_permitted_in_an_optional_chain_1358", "Tagged template expressions are not permitted in an optional chain."),
            Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here: diag(1359, ts.DiagnosticCategory.Error, "Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here_1359", "Identifier expected. '{0}' is a reserved word that cannot be used here."),
            Did_you_mean_to_parenthesize_this_function_type: diag(1360, ts.DiagnosticCategory.Error, "Did_you_mean_to_parenthesize_this_function_type_1360", "Did you mean to parenthesize this function type?"),
            _0_cannot_be_used_as_a_value_because_it_was_imported_using_import_type: diag(1361, ts.DiagnosticCategory.Error, "_0_cannot_be_used_as_a_value_because_it_was_imported_using_import_type_1361", "'{0}' cannot be used as a value because it was imported using 'import type'."),
            _0_cannot_be_used_as_a_value_because_it_was_exported_using_export_type: diag(1362, ts.DiagnosticCategory.Error, "_0_cannot_be_used_as_a_value_because_it_was_exported_using_export_type_1362", "'{0}' cannot be used as a value because it was exported using 'export type'."),
            A_type_only_import_can_specify_a_default_import_or_named_bindings_but_not_both: diag(1363, ts.DiagnosticCategory.Error, "A_type_only_import_can_specify_a_default_import_or_named_bindings_but_not_both_1363", "A type-only import can specify a default import or named bindings, but not both."),
            Convert_to_type_only_export: diag(1364, ts.DiagnosticCategory.Message, "Convert_to_type_only_export_1364", "Convert to type-only export"),
            Convert_all_re_exported_types_to_type_only_exports: diag(1365, ts.DiagnosticCategory.Message, "Convert_all_re_exported_types_to_type_only_exports_1365", "Convert all re-exported types to type-only exports"),
            Split_into_two_separate_import_declarations: diag(1366, ts.DiagnosticCategory.Message, "Split_into_two_separate_import_declarations_1366", "Split into two separate import declarations"),
            Split_all_invalid_type_only_imports: diag(1367, ts.DiagnosticCategory.Message, "Split_all_invalid_type_only_imports_1367", "Split all invalid type-only imports"),
            Did_you_mean_0: diag(1369, ts.DiagnosticCategory.Message, "Did_you_mean_0_1369", "Did you mean '{0}'?"),
            This_import_is_never_used_as_a_value_and_must_use_import_type_because_importsNotUsedAsValues_is_set_to_error: diag(1371, ts.DiagnosticCategory.Error, "This_import_is_never_used_as_a_value_and_must_use_import_type_because_importsNotUsedAsValues_is_set__1371", "This import is never used as a value and must use 'import type' because 'importsNotUsedAsValues' is set to 'error'."),
            Convert_to_type_only_import: diag(1373, ts.DiagnosticCategory.Message, "Convert_to_type_only_import_1373", "Convert to type-only import"),
            Convert_all_imports_not_used_as_a_value_to_type_only_imports: diag(1374, ts.DiagnosticCategory.Message, "Convert_all_imports_not_used_as_a_value_to_type_only_imports_1374", "Convert all imports not used as a value to type-only imports"),
            await_expressions_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_has_no_imports_or_exports_Consider_adding_an_empty_export_to_make_this_file_a_module: diag(1375, ts.DiagnosticCategory.Error, "await_expressions_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_fi_1375", "'await' expressions are only allowed at the top level of a file when that file is a module, but this file has no imports or exports. Consider adding an empty 'export {}' to make this file a module."),
            _0_was_imported_here: diag(1376, ts.DiagnosticCategory.Message, "_0_was_imported_here_1376", "'{0}' was imported here."),
            _0_was_exported_here: diag(1377, ts.DiagnosticCategory.Message, "_0_was_exported_here_1377", "'{0}' was exported here."),
            Top_level_await_expressions_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_target_option_is_set_to_es2017_or_higher: diag(1378, ts.DiagnosticCategory.Error, "Top_level_await_expressions_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_t_1378", "Top-level 'await' expressions are only allowed when the 'module' option is set to 'esnext' or 'system', and the 'target' option is set to 'es2017' or higher."),
            An_import_alias_cannot_reference_a_declaration_that_was_exported_using_export_type: diag(1379, ts.DiagnosticCategory.Error, "An_import_alias_cannot_reference_a_declaration_that_was_exported_using_export_type_1379", "An import alias cannot reference a declaration that was exported using 'export type'."),
            An_import_alias_cannot_reference_a_declaration_that_was_imported_using_import_type: diag(1380, ts.DiagnosticCategory.Error, "An_import_alias_cannot_reference_a_declaration_that_was_imported_using_import_type_1380", "An import alias cannot reference a declaration that was imported using 'import type'."),
            Unexpected_token_Did_you_mean_or_rbrace: diag(1381, ts.DiagnosticCategory.Error, "Unexpected_token_Did_you_mean_or_rbrace_1381", "Unexpected token. Did you mean `{'}'}` or `&rbrace;`?"),
            Unexpected_token_Did_you_mean_or_gt: diag(1382, ts.DiagnosticCategory.Error, "Unexpected_token_Did_you_mean_or_gt_1382", "Unexpected token. Did you mean `{'>'}` or `&gt;`?"),
            Only_named_exports_may_use_export_type: diag(1383, ts.DiagnosticCategory.Error, "Only_named_exports_may_use_export_type_1383", "Only named exports may use 'export type'."),
            A_new_expression_with_type_arguments_must_always_be_followed_by_a_parenthesized_argument_list: diag(1384, ts.DiagnosticCategory.Error, "A_new_expression_with_type_arguments_must_always_be_followed_by_a_parenthesized_argument_list_1384", "A 'new' expression with type arguments must always be followed by a parenthesized argument list."),
            Function_type_notation_must_be_parenthesized_when_used_in_a_union_type: diag(1385, ts.DiagnosticCategory.Error, "Function_type_notation_must_be_parenthesized_when_used_in_a_union_type_1385", "Function type notation must be parenthesized when used in a union type."),
            Constructor_type_notation_must_be_parenthesized_when_used_in_a_union_type: diag(1386, ts.DiagnosticCategory.Error, "Constructor_type_notation_must_be_parenthesized_when_used_in_a_union_type_1386", "Constructor type notation must be parenthesized when used in a union type."),
            Function_type_notation_must_be_parenthesized_when_used_in_an_intersection_type: diag(1387, ts.DiagnosticCategory.Error, "Function_type_notation_must_be_parenthesized_when_used_in_an_intersection_type_1387", "Function type notation must be parenthesized when used in an intersection type."),
            Constructor_type_notation_must_be_parenthesized_when_used_in_an_intersection_type: diag(1388, ts.DiagnosticCategory.Error, "Constructor_type_notation_must_be_parenthesized_when_used_in_an_intersection_type_1388", "Constructor type notation must be parenthesized when used in an intersection type."),
            _0_is_not_allowed_as_a_variable_declaration_name: diag(1389, ts.DiagnosticCategory.Error, "_0_is_not_allowed_as_a_variable_declaration_name_1389", "'{0}' is not allowed as a variable declaration name."),
            Provides_a_root_package_name_when_using_outFile_with_declarations: diag(1390, ts.DiagnosticCategory.Message, "Provides_a_root_package_name_when_using_outFile_with_declarations_1390", "Provides a root package name when using outFile with declarations."),
            The_bundledPackageName_option_must_be_provided_when_using_outFile_and_node_module_resolution_with_declaration_emit: diag(1391, ts.DiagnosticCategory.Error, "The_bundledPackageName_option_must_be_provided_when_using_outFile_and_node_module_resolution_with_de_1391", "The 'bundledPackageName' option must be provided when using outFile and node module resolution with declaration emit."),
            An_import_alias_cannot_use_import_type: diag(1392, ts.DiagnosticCategory.Error, "An_import_alias_cannot_use_import_type_1392", "An import alias cannot use 'import type'"),
            Imported_via_0_from_file_1: diag(1393, ts.DiagnosticCategory.Message, "Imported_via_0_from_file_1_1393", "Imported via {0} from file '{1}'"),
            Imported_via_0_from_file_1_with_packageId_2: diag(1394, ts.DiagnosticCategory.Message, "Imported_via_0_from_file_1_with_packageId_2_1394", "Imported via {0} from file '{1}' with packageId '{2}'"),
            Imported_via_0_from_file_1_to_import_importHelpers_as_specified_in_compilerOptions: diag(1395, ts.DiagnosticCategory.Message, "Imported_via_0_from_file_1_to_import_importHelpers_as_specified_in_compilerOptions_1395", "Imported via {0} from file '{1}' to import 'importHelpers' as specified in compilerOptions"),
            Imported_via_0_from_file_1_with_packageId_2_to_import_importHelpers_as_specified_in_compilerOptions: diag(1396, ts.DiagnosticCategory.Message, "Imported_via_0_from_file_1_with_packageId_2_to_import_importHelpers_as_specified_in_compilerOptions_1396", "Imported via {0} from file '{1}' with packageId '{2}' to import 'importHelpers' as specified in compilerOptions"),
            Imported_via_0_from_file_1_to_import_jsx_and_jsxs_factory_functions: diag(1397, ts.DiagnosticCategory.Message, "Imported_via_0_from_file_1_to_import_jsx_and_jsxs_factory_functions_1397", "Imported via {0} from file '{1}' to import 'jsx' and 'jsxs' factory functions"),
            Imported_via_0_from_file_1_with_packageId_2_to_import_jsx_and_jsxs_factory_functions: diag(1398, ts.DiagnosticCategory.Message, "Imported_via_0_from_file_1_with_packageId_2_to_import_jsx_and_jsxs_factory_functions_1398", "Imported via {0} from file '{1}' with packageId '{2}' to import 'jsx' and 'jsxs' factory functions"),
            File_is_included_via_import_here: diag(1399, ts.DiagnosticCategory.Message, "File_is_included_via_import_here_1399", "File is included via import here."),
            Referenced_via_0_from_file_1: diag(1400, ts.DiagnosticCategory.Message, "Referenced_via_0_from_file_1_1400", "Referenced via '{0}' from file '{1}'"),
            File_is_included_via_reference_here: diag(1401, ts.DiagnosticCategory.Message, "File_is_included_via_reference_here_1401", "File is included via reference here."),
            Type_library_referenced_via_0_from_file_1: diag(1402, ts.DiagnosticCategory.Message, "Type_library_referenced_via_0_from_file_1_1402", "Type library referenced via '{0}' from file '{1}'"),
            Type_library_referenced_via_0_from_file_1_with_packageId_2: diag(1403, ts.DiagnosticCategory.Message, "Type_library_referenced_via_0_from_file_1_with_packageId_2_1403", "Type library referenced via '{0}' from file '{1}' with packageId '{2}'"),
            File_is_included_via_type_library_reference_here: diag(1404, ts.DiagnosticCategory.Message, "File_is_included_via_type_library_reference_here_1404", "File is included via type library reference here."),
            Library_referenced_via_0_from_file_1: diag(1405, ts.DiagnosticCategory.Message, "Library_referenced_via_0_from_file_1_1405", "Library referenced via '{0}' from file '{1}'"),
            File_is_included_via_library_reference_here: diag(1406, ts.DiagnosticCategory.Message, "File_is_included_via_library_reference_here_1406", "File is included via library reference here."),
            Matched_by_include_pattern_0_in_1: diag(1407, ts.DiagnosticCategory.Message, "Matched_by_include_pattern_0_in_1_1407", "Matched by include pattern '{0}' in '{1}'"),
            File_is_matched_by_include_pattern_specified_here: diag(1408, ts.DiagnosticCategory.Message, "File_is_matched_by_include_pattern_specified_here_1408", "File is matched by include pattern specified here."),
            Part_of_files_list_in_tsconfig_json: diag(1409, ts.DiagnosticCategory.Message, "Part_of_files_list_in_tsconfig_json_1409", "Part of 'files' list in tsconfig.json"),
            File_is_matched_by_files_list_specified_here: diag(1410, ts.DiagnosticCategory.Message, "File_is_matched_by_files_list_specified_here_1410", "File is matched by 'files' list specified here."),
            Output_from_referenced_project_0_included_because_1_specified: diag(1411, ts.DiagnosticCategory.Message, "Output_from_referenced_project_0_included_because_1_specified_1411", "Output from referenced project '{0}' included because '{1}' specified"),
            Output_from_referenced_project_0_included_because_module_is_specified_as_none: diag(1412, ts.DiagnosticCategory.Message, "Output_from_referenced_project_0_included_because_module_is_specified_as_none_1412", "Output from referenced project '{0}' included because '--module' is specified as 'none'"),
            File_is_output_from_referenced_project_specified_here: diag(1413, ts.DiagnosticCategory.Message, "File_is_output_from_referenced_project_specified_here_1413", "File is output from referenced project specified here."),
            Source_from_referenced_project_0_included_because_1_specified: diag(1414, ts.DiagnosticCategory.Message, "Source_from_referenced_project_0_included_because_1_specified_1414", "Source from referenced project '{0}' included because '{1}' specified"),
            Source_from_referenced_project_0_included_because_module_is_specified_as_none: diag(1415, ts.DiagnosticCategory.Message, "Source_from_referenced_project_0_included_because_module_is_specified_as_none_1415", "Source from referenced project '{0}' included because '--module' is specified as 'none'"),
            File_is_source_from_referenced_project_specified_here: diag(1416, ts.DiagnosticCategory.Message, "File_is_source_from_referenced_project_specified_here_1416", "File is source from referenced project specified here."),
            Entry_point_of_type_library_0_specified_in_compilerOptions: diag(1417, ts.DiagnosticCategory.Message, "Entry_point_of_type_library_0_specified_in_compilerOptions_1417", "Entry point of type library '{0}' specified in compilerOptions"),
            Entry_point_of_type_library_0_specified_in_compilerOptions_with_packageId_1: diag(1418, ts.DiagnosticCategory.Message, "Entry_point_of_type_library_0_specified_in_compilerOptions_with_packageId_1_1418", "Entry point of type library '{0}' specified in compilerOptions with packageId '{1}'"),
            File_is_entry_point_of_type_library_specified_here: diag(1419, ts.DiagnosticCategory.Message, "File_is_entry_point_of_type_library_specified_here_1419", "File is entry point of type library specified here."),
            Entry_point_for_implicit_type_library_0: diag(1420, ts.DiagnosticCategory.Message, "Entry_point_for_implicit_type_library_0_1420", "Entry point for implicit type library '{0}'"),
            Entry_point_for_implicit_type_library_0_with_packageId_1: diag(1421, ts.DiagnosticCategory.Message, "Entry_point_for_implicit_type_library_0_with_packageId_1_1421", "Entry point for implicit type library '{0}' with packageId '{1}'"),
            Library_0_specified_in_compilerOptions: diag(1422, ts.DiagnosticCategory.Message, "Library_0_specified_in_compilerOptions_1422", "Library '{0}' specified in compilerOptions"),
            File_is_library_specified_here: diag(1423, ts.DiagnosticCategory.Message, "File_is_library_specified_here_1423", "File is library specified here."),
            Default_library: diag(1424, ts.DiagnosticCategory.Message, "Default_library_1424", "Default library"),
            Default_library_for_target_0: diag(1425, ts.DiagnosticCategory.Message, "Default_library_for_target_0_1425", "Default library for target '{0}'"),
            File_is_default_library_for_target_specified_here: diag(1426, ts.DiagnosticCategory.Message, "File_is_default_library_for_target_specified_here_1426", "File is default library for target specified here."),
            Root_file_specified_for_compilation: diag(1427, ts.DiagnosticCategory.Message, "Root_file_specified_for_compilation_1427", "Root file specified for compilation"),
            File_is_output_of_project_reference_source_0: diag(1428, ts.DiagnosticCategory.Message, "File_is_output_of_project_reference_source_0_1428", "File is output of project reference source '{0}'"),
            File_redirects_to_file_0: diag(1429, ts.DiagnosticCategory.Message, "File_redirects_to_file_0_1429", "File redirects to file '{0}'"),
            The_file_is_in_the_program_because_Colon: diag(1430, ts.DiagnosticCategory.Message, "The_file_is_in_the_program_because_Colon_1430", "The file is in the program because:"),
            for_await_loops_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_has_no_imports_or_exports_Consider_adding_an_empty_export_to_make_this_file_a_module: diag(1431, ts.DiagnosticCategory.Error, "for_await_loops_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_1431", "'for await' loops are only allowed at the top level of a file when that file is a module, but this file has no imports or exports. Consider adding an empty 'export {}' to make this file a module."),
            Top_level_for_await_loops_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_target_option_is_set_to_es2017_or_higher: diag(1432, ts.DiagnosticCategory.Error, "Top_level_for_await_loops_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_1432", "Top-level 'for await' loops are only allowed when the 'module' option is set to 'esnext' or 'system', and the 'target' option is set to 'es2017' or higher."),
            Decorators_may_not_be_applied_to_this_parameters: diag(1433, ts.DiagnosticCategory.Error, "Decorators_may_not_be_applied_to_this_parameters_1433", "Decorators may not be applied to 'this' parameters."),
            Unexpected_keyword_or_identifier: diag(1434, ts.DiagnosticCategory.Error, "Unexpected_keyword_or_identifier_1434", "Unexpected keyword or identifier."),
            Unknown_keyword_or_identifier_Did_you_mean_0: diag(1435, ts.DiagnosticCategory.Error, "Unknown_keyword_or_identifier_Did_you_mean_0_1435", "Unknown keyword or identifier. Did you mean '{0}'?"),
            Decorators_must_precede_the_name_and_all_keywords_of_property_declarations: diag(1436, ts.DiagnosticCategory.Error, "Decorators_must_precede_the_name_and_all_keywords_of_property_declarations_1436", "Decorators must precede the name and all keywords of property declarations."),
            Namespace_must_be_given_a_name: diag(1437, ts.DiagnosticCategory.Error, "Namespace_must_be_given_a_name_1437", "Namespace must be given a name."),
            Interface_must_be_given_a_name: diag(1438, ts.DiagnosticCategory.Error, "Interface_must_be_given_a_name_1438", "Interface must be given a name."),
            Type_alias_must_be_given_a_name: diag(1439, ts.DiagnosticCategory.Error, "Type_alias_must_be_given_a_name_1439", "Type alias must be given a name."),
            Variable_declaration_not_allowed_at_this_location: diag(1440, ts.DiagnosticCategory.Error, "Variable_declaration_not_allowed_at_this_location_1440", "Variable declaration not allowed at this location."),
            Cannot_start_a_function_call_in_a_type_annotation: diag(1441, ts.DiagnosticCategory.Error, "Cannot_start_a_function_call_in_a_type_annotation_1441", "Cannot start a function call in a type annotation."),
            Expected_for_property_initializer: diag(1442, ts.DiagnosticCategory.Error, "Expected_for_property_initializer_1442", "Expected '=' for property initializer."),
            Module_declaration_names_may_only_use_or_quoted_strings: diag(1443, ts.DiagnosticCategory.Error, "Module_declaration_names_may_only_use_or_quoted_strings_1443", "Module declaration names may only use ' or \" quoted strings."),
            The_types_of_0_are_incompatible_between_these_types: diag(2200, ts.DiagnosticCategory.Error, "The_types_of_0_are_incompatible_between_these_types_2200", "The types of '{0}' are incompatible between these types."),
            The_types_returned_by_0_are_incompatible_between_these_types: diag(2201, ts.DiagnosticCategory.Error, "The_types_returned_by_0_are_incompatible_between_these_types_2201", "The types returned by '{0}' are incompatible between these types."),
            Call_signature_return_types_0_and_1_are_incompatible: diag(2202, ts.DiagnosticCategory.Error, "Call_signature_return_types_0_and_1_are_incompatible_2202", "Call signature return types '{0}' and '{1}' are incompatible.", /*reportsUnnecessary*/ undefined, /*elidedInCompatabilityPyramid*/ true),
            Construct_signature_return_types_0_and_1_are_incompatible: diag(2203, ts.DiagnosticCategory.Error, "Construct_signature_return_types_0_and_1_are_incompatible_2203", "Construct signature return types '{0}' and '{1}' are incompatible.", /*reportsUnnecessary*/ undefined, /*elidedInCompatabilityPyramid*/ true),
            Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1: diag(2204, ts.DiagnosticCategory.Error, "Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1_2204", "Call signatures with no arguments have incompatible return types '{0}' and '{1}'.", /*reportsUnnecessary*/ undefined, /*elidedInCompatabilityPyramid*/ true),
            Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1: diag(2205, ts.DiagnosticCategory.Error, "Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1_2205", "Construct signatures with no arguments have incompatible return types '{0}' and '{1}'.", /*reportsUnnecessary*/ undefined, /*elidedInCompatabilityPyramid*/ true),
            Duplicate_identifier_0: diag(2300, ts.DiagnosticCategory.Error, "Duplicate_identifier_0_2300", "Duplicate identifier '{0}'."),
            Initializer_of_instance_member_variable_0_cannot_reference_identifier_1_declared_in_the_constructor: diag(2301, ts.DiagnosticCategory.Error, "Initializer_of_instance_member_variable_0_cannot_reference_identifier_1_declared_in_the_constructor_2301", "Initializer of instance member variable '{0}' cannot reference identifier '{1}' declared in the constructor."),
            Static_members_cannot_reference_class_type_parameters: diag(2302, ts.DiagnosticCategory.Error, "Static_members_cannot_reference_class_type_parameters_2302", "Static members cannot reference class type parameters."),
            Circular_definition_of_import_alias_0: diag(2303, ts.DiagnosticCategory.Error, "Circular_definition_of_import_alias_0_2303", "Circular definition of import alias '{0}'."),
            Cannot_find_name_0: diag(2304, ts.DiagnosticCategory.Error, "Cannot_find_name_0_2304", "Cannot find name '{0}'."),
            Module_0_has_no_exported_member_1: diag(2305, ts.DiagnosticCategory.Error, "Module_0_has_no_exported_member_1_2305", "Module '{0}' has no exported member '{1}'."),
            File_0_is_not_a_module: diag(2306, ts.DiagnosticCategory.Error, "File_0_is_not_a_module_2306", "File '{0}' is not a module."),
            Cannot_find_module_0_or_its_corresponding_type_declarations: diag(2307, ts.DiagnosticCategory.Error, "Cannot_find_module_0_or_its_corresponding_type_declarations_2307", "Cannot find module '{0}' or its corresponding type declarations."),
            Module_0_has_already_exported_a_member_named_1_Consider_explicitly_re_exporting_to_resolve_the_ambiguity: diag(2308, ts.DiagnosticCategory.Error, "Module_0_has_already_exported_a_member_named_1_Consider_explicitly_re_exporting_to_resolve_the_ambig_2308", "Module {0} has already exported a member named '{1}'. Consider explicitly re-exporting to resolve the ambiguity."),
            An_export_assignment_cannot_be_used_in_a_module_with_other_exported_elements: diag(2309, ts.DiagnosticCategory.Error, "An_export_assignment_cannot_be_used_in_a_module_with_other_exported_elements_2309", "An export assignment cannot be used in a module with other exported elements."),
            Type_0_recursively_references_itself_as_a_base_type: diag(2310, ts.DiagnosticCategory.Error, "Type_0_recursively_references_itself_as_a_base_type_2310", "Type '{0}' recursively references itself as a base type."),
            A_class_may_only_extend_another_class: diag(2311, ts.DiagnosticCategory.Error, "A_class_may_only_extend_another_class_2311", "A class may only extend another class."),
            An_interface_can_only_extend_an_object_type_or_intersection_of_object_types_with_statically_known_members: diag(2312, ts.DiagnosticCategory.Error, "An_interface_can_only_extend_an_object_type_or_intersection_of_object_types_with_statically_known_me_2312", "An interface can only extend an object type or intersection of object types with statically known members."),
            Type_parameter_0_has_a_circular_constraint: diag(2313, ts.DiagnosticCategory.Error, "Type_parameter_0_has_a_circular_constraint_2313", "Type parameter '{0}' has a circular constraint."),
            Generic_type_0_requires_1_type_argument_s: diag(2314, ts.DiagnosticCategory.Error, "Generic_type_0_requires_1_type_argument_s_2314", "Generic type '{0}' requires {1} type argument(s)."),
            Type_0_is_not_generic: diag(2315, ts.DiagnosticCategory.Error, "Type_0_is_not_generic_2315", "Type '{0}' is not generic."),
            Global_type_0_must_be_a_class_or_interface_type: diag(2316, ts.DiagnosticCategory.Error, "Global_type_0_must_be_a_class_or_interface_type_2316", "Global type '{0}' must be a class or interface type."),
            Global_type_0_must_have_1_type_parameter_s: diag(2317, ts.DiagnosticCategory.Error, "Global_type_0_must_have_1_type_parameter_s_2317", "Global type '{0}' must have {1} type parameter(s)."),
            Cannot_find_global_type_0: diag(2318, ts.DiagnosticCategory.Error, "Cannot_find_global_type_0_2318", "Cannot find global type '{0}'."),
            Named_property_0_of_types_1_and_2_are_not_identical: diag(2319, ts.DiagnosticCategory.Error, "Named_property_0_of_types_1_and_2_are_not_identical_2319", "Named property '{0}' of types '{1}' and '{2}' are not identical."),
            Interface_0_cannot_simultaneously_extend_types_1_and_2: diag(2320, ts.DiagnosticCategory.Error, "Interface_0_cannot_simultaneously_extend_types_1_and_2_2320", "Interface '{0}' cannot simultaneously extend types '{1}' and '{2}'."),
            Excessive_stack_depth_comparing_types_0_and_1: diag(2321, ts.DiagnosticCategory.Error, "Excessive_stack_depth_comparing_types_0_and_1_2321", "Excessive stack depth comparing types '{0}' and '{1}'."),
            Type_0_is_not_assignable_to_type_1: diag(2322, ts.DiagnosticCategory.Error, "Type_0_is_not_assignable_to_type_1_2322", "Type '{0}' is not assignable to type '{1}'."),
            Cannot_redeclare_exported_variable_0: diag(2323, ts.DiagnosticCategory.Error, "Cannot_redeclare_exported_variable_0_2323", "Cannot redeclare exported variable '{0}'."),
            Property_0_is_missing_in_type_1: diag(2324, ts.DiagnosticCategory.Error, "Property_0_is_missing_in_type_1_2324", "Property '{0}' is missing in type '{1}'."),
            Property_0_is_private_in_type_1_but_not_in_type_2: diag(2325, ts.DiagnosticCategory.Error, "Property_0_is_private_in_type_1_but_not_in_type_2_2325", "Property '{0}' is private in type '{1}' but not in type '{2}'."),
            Types_of_property_0_are_incompatible: diag(2326, ts.DiagnosticCategory.Error, "Types_of_property_0_are_incompatible_2326", "Types of property '{0}' are incompatible."),
            Property_0_is_optional_in_type_1_but_required_in_type_2: diag(2327, ts.DiagnosticCategory.Error, "Property_0_is_optional_in_type_1_but_required_in_type_2_2327", "Property '{0}' is optional in type '{1}' but required in type '{2}'."),
            Types_of_parameters_0_and_1_are_incompatible: diag(2328, ts.DiagnosticCategory.Error, "Types_of_parameters_0_and_1_are_incompatible_2328", "Types of parameters '{0}' and '{1}' are incompatible."),
            Index_signature_for_type_0_is_missing_in_type_1: diag(2329, ts.DiagnosticCategory.Error, "Index_signature_for_type_0_is_missing_in_type_1_2329", "Index signature for type '{0}' is missing in type '{1}'."),
            _0_and_1_index_signatures_are_incompatible: diag(2330, ts.DiagnosticCategory.Error, "_0_and_1_index_signatures_are_incompatible_2330", "'{0}' and '{1}' index signatures are incompatible."),
            this_cannot_be_referenced_in_a_module_or_namespace_body: diag(2331, ts.DiagnosticCategory.Error, "this_cannot_be_referenced_in_a_module_or_namespace_body_2331", "'this' cannot be referenced in a module or namespace body."),
            this_cannot_be_referenced_in_current_location: diag(2332, ts.DiagnosticCategory.Error, "this_cannot_be_referenced_in_current_location_2332", "'this' cannot be referenced in current location."),
            this_cannot_be_referenced_in_constructor_arguments: diag(2333, ts.DiagnosticCategory.Error, "this_cannot_be_referenced_in_constructor_arguments_2333", "'this' cannot be referenced in constructor arguments."),
            this_cannot_be_referenced_in_a_static_property_initializer: diag(2334, ts.DiagnosticCategory.Error, "this_cannot_be_referenced_in_a_static_property_initializer_2334", "'this' cannot be referenced in a static property initializer."),
            super_can_only_be_referenced_in_a_derived_class: diag(2335, ts.DiagnosticCategory.Error, "super_can_only_be_referenced_in_a_derived_class_2335", "'super' can only be referenced in a derived class."),
            super_cannot_be_referenced_in_constructor_arguments: diag(2336, ts.DiagnosticCategory.Error, "super_cannot_be_referenced_in_constructor_arguments_2336", "'super' cannot be referenced in constructor arguments."),
            Super_calls_are_not_permitted_outside_constructors_or_in_nested_functions_inside_constructors: diag(2337, ts.DiagnosticCategory.Error, "Super_calls_are_not_permitted_outside_constructors_or_in_nested_functions_inside_constructors_2337", "Super calls are not permitted outside constructors or in nested functions inside constructors."),
            super_property_access_is_permitted_only_in_a_constructor_member_function_or_member_accessor_of_a_derived_class: diag(2338, ts.DiagnosticCategory.Error, "super_property_access_is_permitted_only_in_a_constructor_member_function_or_member_accessor_of_a_der_2338", "'super' property access is permitted only in a constructor, member function, or member accessor of a derived class."),
            Property_0_does_not_exist_on_type_1: diag(2339, ts.DiagnosticCategory.Error, "Property_0_does_not_exist_on_type_1_2339", "Property '{0}' does not exist on type '{1}'."),
            Only_public_and_protected_methods_of_the_base_class_are_accessible_via_the_super_keyword: diag(2340, ts.DiagnosticCategory.Error, "Only_public_and_protected_methods_of_the_base_class_are_accessible_via_the_super_keyword_2340", "Only public and protected methods of the base class are accessible via the 'super' keyword."),
            Property_0_is_private_and_only_accessible_within_class_1: diag(2341, ts.DiagnosticCategory.Error, "Property_0_is_private_and_only_accessible_within_class_1_2341", "Property '{0}' is private and only accessible within class '{1}'."),
            An_index_expression_argument_must_be_of_type_string_number_symbol_or_any: diag(2342, ts.DiagnosticCategory.Error, "An_index_expression_argument_must_be_of_type_string_number_symbol_or_any_2342", "An index expression argument must be of type 'string', 'number', 'symbol', or 'any'."),
            This_syntax_requires_an_imported_helper_named_1_which_does_not_exist_in_0_Consider_upgrading_your_version_of_0: diag(2343, ts.DiagnosticCategory.Error, "This_syntax_requires_an_imported_helper_named_1_which_does_not_exist_in_0_Consider_upgrading_your_ve_2343", "This syntax requires an imported helper named '{1}' which does not exist in '{0}'. Consider upgrading your version of '{0}'."),
            Type_0_does_not_satisfy_the_constraint_1: diag(2344, ts.DiagnosticCategory.Error, "Type_0_does_not_satisfy_the_constraint_1_2344", "Type '{0}' does not satisfy the constraint '{1}'."),
            Argument_of_type_0_is_not_assignable_to_parameter_of_type_1: diag(2345, ts.DiagnosticCategory.Error, "Argument_of_type_0_is_not_assignable_to_parameter_of_type_1_2345", "Argument of type '{0}' is not assignable to parameter of type '{1}'."),
            Call_target_does_not_contain_any_signatures: diag(2346, ts.DiagnosticCategory.Error, "Call_target_does_not_contain_any_signatures_2346", "Call target does not contain any signatures."),
            Untyped_function_calls_may_not_accept_type_arguments: diag(2347, ts.DiagnosticCategory.Error, "Untyped_function_calls_may_not_accept_type_arguments_2347", "Untyped function calls may not accept type arguments."),
            Value_of_type_0_is_not_callable_Did_you_mean_to_include_new: diag(2348, ts.DiagnosticCategory.Error, "Value_of_type_0_is_not_callable_Did_you_mean_to_include_new_2348", "Value of type '{0}' is not callable. Did you mean to include 'new'?"),
            This_expression_is_not_callable: diag(2349, ts.DiagnosticCategory.Error, "This_expression_is_not_callable_2349", "This expression is not callable."),
            Only_a_void_function_can_be_called_with_the_new_keyword: diag(2350, ts.DiagnosticCategory.Error, "Only_a_void_function_can_be_called_with_the_new_keyword_2350", "Only a void function can be called with the 'new' keyword."),
            This_expression_is_not_constructable: diag(2351, ts.DiagnosticCategory.Error, "This_expression_is_not_constructable_2351", "This expression is not constructable."),
            Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the_other_If_this_was_intentional_convert_the_expression_to_unknown_first: diag(2352, ts.DiagnosticCategory.Error, "Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the__2352", "Conversion of type '{0}' to type '{1}' may be a mistake because neither type sufficiently overlaps with the other. If this was intentional, convert the expression to 'unknown' first."),
            Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1: diag(2353, ts.DiagnosticCategory.Error, "Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1_2353", "Object literal may only specify known properties, and '{0}' does not exist in type '{1}'."),
            This_syntax_requires_an_imported_helper_but_module_0_cannot_be_found: diag(2354, ts.DiagnosticCategory.Error, "This_syntax_requires_an_imported_helper_but_module_0_cannot_be_found_2354", "This syntax requires an imported helper but module '{0}' cannot be found."),
            A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value: diag(2355, ts.DiagnosticCategory.Error, "A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value_2355", "A function whose declared type is neither 'void' nor 'any' must return a value."),
            An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type: diag(2356, ts.DiagnosticCategory.Error, "An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type_2356", "An arithmetic operand must be of type 'any', 'number', 'bigint' or an enum type."),
            The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access: diag(2357, ts.DiagnosticCategory.Error, "The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access_2357", "The operand of an increment or decrement operator must be a variable or a property access."),
            The_left_hand_side_of_an_instanceof_expression_must_be_of_type_any_an_object_type_or_a_type_parameter: diag(2358, ts.DiagnosticCategory.Error, "The_left_hand_side_of_an_instanceof_expression_must_be_of_type_any_an_object_type_or_a_type_paramete_2358", "The left-hand side of an 'instanceof' expression must be of type 'any', an object type or a type parameter."),
            The_right_hand_side_of_an_instanceof_expression_must_be_of_type_any_or_of_a_type_assignable_to_the_Function_interface_type: diag(2359, ts.DiagnosticCategory.Error, "The_right_hand_side_of_an_instanceof_expression_must_be_of_type_any_or_of_a_type_assignable_to_the_F_2359", "The right-hand side of an 'instanceof' expression must be of type 'any' or of a type assignable to the 'Function' interface type."),
            The_left_hand_side_of_an_in_expression_must_be_of_type_any_string_number_or_symbol: diag(2360, ts.DiagnosticCategory.Error, "The_left_hand_side_of_an_in_expression_must_be_of_type_any_string_number_or_symbol_2360", "The left-hand side of an 'in' expression must be of type 'any', 'string', 'number', or 'symbol'."),
            The_right_hand_side_of_an_in_expression_must_not_be_a_primitive: diag(2361, ts.DiagnosticCategory.Error, "The_right_hand_side_of_an_in_expression_must_not_be_a_primitive_2361", "The right-hand side of an 'in' expression must not be a primitive."),
            The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type: diag(2362, ts.DiagnosticCategory.Error, "The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type_2362", "The left-hand side of an arithmetic operation must be of type 'any', 'number', 'bigint' or an enum type."),
            The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type: diag(2363, ts.DiagnosticCategory.Error, "The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type_2363", "The right-hand side of an arithmetic operation must be of type 'any', 'number', 'bigint' or an enum type."),
            The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access: diag(2364, ts.DiagnosticCategory.Error, "The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access_2364", "The left-hand side of an assignment expression must be a variable or a property access."),
            Operator_0_cannot_be_applied_to_types_1_and_2: diag(2365, ts.DiagnosticCategory.Error, "Operator_0_cannot_be_applied_to_types_1_and_2_2365", "Operator '{0}' cannot be applied to types '{1}' and '{2}'."),
            Function_lacks_ending_return_statement_and_return_type_does_not_include_undefined: diag(2366, ts.DiagnosticCategory.Error, "Function_lacks_ending_return_statement_and_return_type_does_not_include_undefined_2366", "Function lacks ending return statement and return type does not include 'undefined'."),
            This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap: diag(2367, ts.DiagnosticCategory.Error, "This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap_2367", "This condition will always return '{0}' since the types '{1}' and '{2}' have no overlap."),
            Type_parameter_name_cannot_be_0: diag(2368, ts.DiagnosticCategory.Error, "Type_parameter_name_cannot_be_0_2368", "Type parameter name cannot be '{0}'."),
            A_parameter_property_is_only_allowed_in_a_constructor_implementation: diag(2369, ts.DiagnosticCategory.Error, "A_parameter_property_is_only_allowed_in_a_constructor_implementation_2369", "A parameter property is only allowed in a constructor implementation."),
            A_rest_parameter_must_be_of_an_array_type: diag(2370, ts.DiagnosticCategory.Error, "A_rest_parameter_must_be_of_an_array_type_2370", "A rest parameter must be of an array type."),
            A_parameter_initializer_is_only_allowed_in_a_function_or_constructor_implementation: diag(2371, ts.DiagnosticCategory.Error, "A_parameter_initializer_is_only_allowed_in_a_function_or_constructor_implementation_2371", "A parameter initializer is only allowed in a function or constructor implementation."),
            Parameter_0_cannot_reference_itself: diag(2372, ts.DiagnosticCategory.Error, "Parameter_0_cannot_reference_itself_2372", "Parameter '{0}' cannot reference itself."),
            Parameter_0_cannot_reference_identifier_1_declared_after_it: diag(2373, ts.DiagnosticCategory.Error, "Parameter_0_cannot_reference_identifier_1_declared_after_it_2373", "Parameter '{0}' cannot reference identifier '{1}' declared after it."),
            Duplicate_index_signature_for_type_0: diag(2374, ts.DiagnosticCategory.Error, "Duplicate_index_signature_for_type_0_2374", "Duplicate index signature for type '{0}'."),
            A_super_call_must_be_the_first_statement_in_the_constructor_when_a_class_contains_initialized_properties_parameter_properties_or_private_identifiers: diag(2376, ts.DiagnosticCategory.Error, "A_super_call_must_be_the_first_statement_in_the_constructor_when_a_class_contains_initialized_proper_2376", "A 'super' call must be the first statement in the constructor when a class contains initialized properties, parameter properties, or private identifiers."),
            Constructors_for_derived_classes_must_contain_a_super_call: diag(2377, ts.DiagnosticCategory.Error, "Constructors_for_derived_classes_must_contain_a_super_call_2377", "Constructors for derived classes must contain a 'super' call."),
            A_get_accessor_must_return_a_value: diag(2378, ts.DiagnosticCategory.Error, "A_get_accessor_must_return_a_value_2378", "A 'get' accessor must return a value."),
            The_return_type_of_a_get_accessor_must_be_assignable_to_its_set_accessor_type: diag(2380, ts.DiagnosticCategory.Error, "The_return_type_of_a_get_accessor_must_be_assignable_to_its_set_accessor_type_2380", "The return type of a 'get' accessor must be assignable to its 'set' accessor type"),
            A_signature_with_an_implementation_cannot_use_a_string_literal_type: diag(2381, ts.DiagnosticCategory.Error, "A_signature_with_an_implementation_cannot_use_a_string_literal_type_2381", "A signature with an implementation cannot use a string literal type."),
            Specialized_overload_signature_is_not_assignable_to_any_non_specialized_signature: diag(2382, ts.DiagnosticCategory.Error, "Specialized_overload_signature_is_not_assignable_to_any_non_specialized_signature_2382", "Specialized overload signature is not assignable to any non-specialized signature."),
            Overload_signatures_must_all_be_exported_or_non_exported: diag(2383, ts.DiagnosticCategory.Error, "Overload_signatures_must_all_be_exported_or_non_exported_2383", "Overload signatures must all be exported or non-exported."),
            Overload_signatures_must_all_be_ambient_or_non_ambient: diag(2384, ts.DiagnosticCategory.Error, "Overload_signatures_must_all_be_ambient_or_non_ambient_2384", "Overload signatures must all be ambient or non-ambient."),
            Overload_signatures_must_all_be_public_private_or_protected: diag(2385, ts.DiagnosticCategory.Error, "Overload_signatures_must_all_be_public_private_or_protected_2385", "Overload signatures must all be public, private or protected."),
            Overload_signatures_must_all_be_optional_or_required: diag(2386, ts.DiagnosticCategory.Error, "Overload_signatures_must_all_be_optional_or_required_2386", "Overload signatures must all be optional or required."),
            Function_overload_must_be_static: diag(2387, ts.DiagnosticCategory.Error, "Function_overload_must_be_static_2387", "Function overload must be static."),
            Function_overload_must_not_be_static: diag(2388, ts.DiagnosticCategory.Error, "Function_overload_must_not_be_static_2388", "Function overload must not be static."),
            Function_implementation_name_must_be_0: diag(2389, ts.DiagnosticCategory.Error, "Function_implementation_name_must_be_0_2389", "Function implementation name must be '{0}'."),
            Constructor_implementation_is_missing: diag(2390, ts.DiagnosticCategory.Error, "Constructor_implementation_is_missing_2390", "Constructor implementation is missing."),
            Function_implementation_is_missing_or_not_immediately_following_the_declaration: diag(2391, ts.DiagnosticCategory.Error, "Function_implementation_is_missing_or_not_immediately_following_the_declaration_2391", "Function implementation is missing or not immediately following the declaration."),
            Multiple_constructor_implementations_are_not_allowed: diag(2392, ts.DiagnosticCategory.Error, "Multiple_constructor_implementations_are_not_allowed_2392", "Multiple constructor implementations are not allowed."),
            Duplicate_function_implementation: diag(2393, ts.DiagnosticCategory.Error, "Duplicate_function_implementation_2393", "Duplicate function implementation."),
            This_overload_signature_is_not_compatible_with_its_implementation_signature: diag(2394, ts.DiagnosticCategory.Error, "This_overload_signature_is_not_compatible_with_its_implementation_signature_2394", "This overload signature is not compatible with its implementation signature."),
            Individual_declarations_in_merged_declaration_0_must_be_all_exported_or_all_local: diag(2395, ts.DiagnosticCategory.Error, "Individual_declarations_in_merged_declaration_0_must_be_all_exported_or_all_local_2395", "Individual declarations in merged declaration '{0}' must be all exported or all local."),
            Duplicate_identifier_arguments_Compiler_uses_arguments_to_initialize_rest_parameters: diag(2396, ts.DiagnosticCategory.Error, "Duplicate_identifier_arguments_Compiler_uses_arguments_to_initialize_rest_parameters_2396", "Duplicate identifier 'arguments'. Compiler uses 'arguments' to initialize rest parameters."),
            Declaration_name_conflicts_with_built_in_global_identifier_0: diag(2397, ts.DiagnosticCategory.Error, "Declaration_name_conflicts_with_built_in_global_identifier_0_2397", "Declaration name conflicts with built-in global identifier '{0}'."),
            constructor_cannot_be_used_as_a_parameter_property_name: diag(2398, ts.DiagnosticCategory.Error, "constructor_cannot_be_used_as_a_parameter_property_name_2398", "'constructor' cannot be used as a parameter property name."),
            Duplicate_identifier_this_Compiler_uses_variable_declaration_this_to_capture_this_reference: diag(2399, ts.DiagnosticCategory.Error, "Duplicate_identifier_this_Compiler_uses_variable_declaration_this_to_capture_this_reference_2399", "Duplicate identifier '_this'. Compiler uses variable declaration '_this' to capture 'this' reference."),
            Expression_resolves_to_variable_declaration_this_that_compiler_uses_to_capture_this_reference: diag(2400, ts.DiagnosticCategory.Error, "Expression_resolves_to_variable_declaration_this_that_compiler_uses_to_capture_this_reference_2400", "Expression resolves to variable declaration '_this' that compiler uses to capture 'this' reference."),
            Duplicate_identifier_super_Compiler_uses_super_to_capture_base_class_reference: diag(2401, ts.DiagnosticCategory.Error, "Duplicate_identifier_super_Compiler_uses_super_to_capture_base_class_reference_2401", "Duplicate identifier '_super'. Compiler uses '_super' to capture base class reference."),
            Expression_resolves_to_super_that_compiler_uses_to_capture_base_class_reference: diag(2402, ts.DiagnosticCategory.Error, "Expression_resolves_to_super_that_compiler_uses_to_capture_base_class_reference_2402", "Expression resolves to '_super' that compiler uses to capture base class reference."),
            Subsequent_variable_declarations_must_have_the_same_type_Variable_0_must_be_of_type_1_but_here_has_type_2: diag(2403, ts.DiagnosticCategory.Error, "Subsequent_variable_declarations_must_have_the_same_type_Variable_0_must_be_of_type_1_but_here_has_t_2403", "Subsequent variable declarations must have the same type.  Variable '{0}' must be of type '{1}', but here has type '{2}'."),
            The_left_hand_side_of_a_for_in_statement_cannot_use_a_type_annotation: diag(2404, ts.DiagnosticCategory.Error, "The_left_hand_side_of_a_for_in_statement_cannot_use_a_type_annotation_2404", "The left-hand side of a 'for...in' statement cannot use a type annotation."),
            The_left_hand_side_of_a_for_in_statement_must_be_of_type_string_or_any: diag(2405, ts.DiagnosticCategory.Error, "The_left_hand_side_of_a_for_in_statement_must_be_of_type_string_or_any_2405", "The left-hand side of a 'for...in' statement must be of type 'string' or 'any'."),
            The_left_hand_side_of_a_for_in_statement_must_be_a_variable_or_a_property_access: diag(2406, ts.DiagnosticCategory.Error, "The_left_hand_side_of_a_for_in_statement_must_be_a_variable_or_a_property_access_2406", "The left-hand side of a 'for...in' statement must be a variable or a property access."),
            The_right_hand_side_of_a_for_in_statement_must_be_of_type_any_an_object_type_or_a_type_parameter_but_here_has_type_0: diag(2407, ts.DiagnosticCategory.Error, "The_right_hand_side_of_a_for_in_statement_must_be_of_type_any_an_object_type_or_a_type_parameter_but_2407", "The right-hand side of a 'for...in' statement must be of type 'any', an object type or a type parameter, but here has type '{0}'."),
            Setters_cannot_return_a_value: diag(2408, ts.DiagnosticCategory.Error, "Setters_cannot_return_a_value_2408", "Setters cannot return a value."),
            Return_type_of_constructor_signature_must_be_assignable_to_the_instance_type_of_the_class: diag(2409, ts.DiagnosticCategory.Error, "Return_type_of_constructor_signature_must_be_assignable_to_the_instance_type_of_the_class_2409", "Return type of constructor signature must be assignable to the instance type of the class."),
            The_with_statement_is_not_supported_All_symbols_in_a_with_block_will_have_type_any: diag(2410, ts.DiagnosticCategory.Error, "The_with_statement_is_not_supported_All_symbols_in_a_with_block_will_have_type_any_2410", "The 'with' statement is not supported. All symbols in a 'with' block will have type 'any'."),
            Property_0_of_type_1_is_not_assignable_to_2_index_type_3: diag(2411, ts.DiagnosticCategory.Error, "Property_0_of_type_1_is_not_assignable_to_2_index_type_3_2411", "Property '{0}' of type '{1}' is not assignable to '{2}' index type '{3}'."),
            _0_index_type_1_is_not_assignable_to_2_index_type_3: diag(2413, ts.DiagnosticCategory.Error, "_0_index_type_1_is_not_assignable_to_2_index_type_3_2413", "'{0}' index type '{1}' is not assignable to '{2}' index type '{3}'."),
            Class_name_cannot_be_0: diag(2414, ts.DiagnosticCategory.Error, "Class_name_cannot_be_0_2414", "Class name cannot be '{0}'."),
            Class_0_incorrectly_extends_base_class_1: diag(2415, ts.DiagnosticCategory.Error, "Class_0_incorrectly_extends_base_class_1_2415", "Class '{0}' incorrectly extends base class '{1}'."),
            Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2: diag(2416, ts.DiagnosticCategory.Error, "Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2_2416", "Property '{0}' in type '{1}' is not assignable to the same property in base type '{2}'."),
            Class_static_side_0_incorrectly_extends_base_class_static_side_1: diag(2417, ts.DiagnosticCategory.Error, "Class_static_side_0_incorrectly_extends_base_class_static_side_1_2417", "Class static side '{0}' incorrectly extends base class static side '{1}'."),
            Type_of_computed_property_s_value_is_0_which_is_not_assignable_to_type_1: diag(2418, ts.DiagnosticCategory.Error, "Type_of_computed_property_s_value_is_0_which_is_not_assignable_to_type_1_2418", "Type of computed property's value is '{0}', which is not assignable to type '{1}'."),
            Types_of_construct_signatures_are_incompatible: diag(2419, ts.DiagnosticCategory.Error, "Types_of_construct_signatures_are_incompatible_2419", "Types of construct signatures are incompatible."),
            Class_0_incorrectly_implements_interface_1: diag(2420, ts.DiagnosticCategory.Error, "Class_0_incorrectly_implements_interface_1_2420", "Class '{0}' incorrectly implements interface '{1}'."),
            A_class_can_only_implement_an_object_type_or_intersection_of_object_types_with_statically_known_members: diag(2422, ts.DiagnosticCategory.Error, "A_class_can_only_implement_an_object_type_or_intersection_of_object_types_with_statically_known_memb_2422", "A class can only implement an object type or intersection of object types with statically known members."),
            Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_accessor: diag(2423, ts.DiagnosticCategory.Error, "Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_access_2423", "Class '{0}' defines instance member function '{1}', but extended class '{2}' defines it as instance member accessor."),
            Class_0_defines_instance_member_property_1_but_extended_class_2_defines_it_as_instance_member_function: diag(2425, ts.DiagnosticCategory.Error, "Class_0_defines_instance_member_property_1_but_extended_class_2_defines_it_as_instance_member_functi_2425", "Class '{0}' defines instance member property '{1}', but extended class '{2}' defines it as instance member function."),
            Class_0_defines_instance_member_accessor_1_but_extended_class_2_defines_it_as_instance_member_function: diag(2426, ts.DiagnosticCategory.Error, "Class_0_defines_instance_member_accessor_1_but_extended_class_2_defines_it_as_instance_member_functi_2426", "Class '{0}' defines instance member accessor '{1}', but extended class '{2}' defines it as instance member function."),
            Interface_name_cannot_be_0: diag(2427, ts.DiagnosticCategory.Error, "Interface_name_cannot_be_0_2427", "Interface name cannot be '{0}'."),
            All_declarations_of_0_must_have_identical_type_parameters: diag(2428, ts.DiagnosticCategory.Error, "All_declarations_of_0_must_have_identical_type_parameters_2428", "All declarations of '{0}' must have identical type parameters."),
            Interface_0_incorrectly_extends_interface_1: diag(2430, ts.DiagnosticCategory.Error, "Interface_0_incorrectly_extends_interface_1_2430", "Interface '{0}' incorrectly extends interface '{1}'."),
            Enum_name_cannot_be_0: diag(2431, ts.DiagnosticCategory.Error, "Enum_name_cannot_be_0_2431", "Enum name cannot be '{0}'."),
            In_an_enum_with_multiple_declarations_only_one_declaration_can_omit_an_initializer_for_its_first_enum_element: diag(2432, ts.DiagnosticCategory.Error, "In_an_enum_with_multiple_declarations_only_one_declaration_can_omit_an_initializer_for_its_first_enu_2432", "In an enum with multiple declarations, only one declaration can omit an initializer for its first enum element."),
            A_namespace_declaration_cannot_be_in_a_different_file_from_a_class_or_function_with_which_it_is_merged: diag(2433, ts.DiagnosticCategory.Error, "A_namespace_declaration_cannot_be_in_a_different_file_from_a_class_or_function_with_which_it_is_merg_2433", "A namespace declaration cannot be in a different file from a class or function with which it is merged."),
            A_namespace_declaration_cannot_be_located_prior_to_a_class_or_function_with_which_it_is_merged: diag(2434, ts.DiagnosticCategory.Error, "A_namespace_declaration_cannot_be_located_prior_to_a_class_or_function_with_which_it_is_merged_2434", "A namespace declaration cannot be located prior to a class or function with which it is merged."),
            Ambient_modules_cannot_be_nested_in_other_modules_or_namespaces: diag(2435, ts.DiagnosticCategory.Error, "Ambient_modules_cannot_be_nested_in_other_modules_or_namespaces_2435", "Ambient modules cannot be nested in other modules or namespaces."),
            Ambient_module_declaration_cannot_specify_relative_module_name: diag(2436, ts.DiagnosticCategory.Error, "Ambient_module_declaration_cannot_specify_relative_module_name_2436", "Ambient module declaration cannot specify relative module name."),
            Module_0_is_hidden_by_a_local_declaration_with_the_same_name: diag(2437, ts.DiagnosticCategory.Error, "Module_0_is_hidden_by_a_local_declaration_with_the_same_name_2437", "Module '{0}' is hidden by a local declaration with the same name."),
            Import_name_cannot_be_0: diag(2438, ts.DiagnosticCategory.Error, "Import_name_cannot_be_0_2438", "Import name cannot be '{0}'."),
            Import_or_export_declaration_in_an_ambient_module_declaration_cannot_reference_module_through_relative_module_name: diag(2439, ts.DiagnosticCategory.Error, "Import_or_export_declaration_in_an_ambient_module_declaration_cannot_reference_module_through_relati_2439", "Import or export declaration in an ambient module declaration cannot reference module through relative module name."),
            Import_declaration_conflicts_with_local_declaration_of_0: diag(2440, ts.DiagnosticCategory.Error, "Import_declaration_conflicts_with_local_declaration_of_0_2440", "Import declaration conflicts with local declaration of '{0}'."),
            Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module: diag(2441, ts.DiagnosticCategory.Error, "Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_2441", "Duplicate identifier '{0}'. Compiler reserves name '{1}' in top level scope of a module."),
            Types_have_separate_declarations_of_a_private_property_0: diag(2442, ts.DiagnosticCategory.Error, "Types_have_separate_declarations_of_a_private_property_0_2442", "Types have separate declarations of a private property '{0}'."),
            Property_0_is_protected_but_type_1_is_not_a_class_derived_from_2: diag(2443, ts.DiagnosticCategory.Error, "Property_0_is_protected_but_type_1_is_not_a_class_derived_from_2_2443", "Property '{0}' is protected but type '{1}' is not a class derived from '{2}'."),
            Property_0_is_protected_in_type_1_but_public_in_type_2: diag(2444, ts.DiagnosticCategory.Error, "Property_0_is_protected_in_type_1_but_public_in_type_2_2444", "Property '{0}' is protected in type '{1}' but public in type '{2}'."),
            Property_0_is_protected_and_only_accessible_within_class_1_and_its_subclasses: diag(2445, ts.DiagnosticCategory.Error, "Property_0_is_protected_and_only_accessible_within_class_1_and_its_subclasses_2445", "Property '{0}' is protected and only accessible within class '{1}' and its subclasses."),
            Property_0_is_protected_and_only_accessible_through_an_instance_of_class_1_This_is_an_instance_of_class_2: diag(2446, ts.DiagnosticCategory.Error, "Property_0_is_protected_and_only_accessible_through_an_instance_of_class_1_This_is_an_instance_of_cl_2446", "Property '{0}' is protected and only accessible through an instance of class '{1}'. This is an instance of class '{2}'."),
            The_0_operator_is_not_allowed_for_boolean_types_Consider_using_1_instead: diag(2447, ts.DiagnosticCategory.Error, "The_0_operator_is_not_allowed_for_boolean_types_Consider_using_1_instead_2447", "The '{0}' operator is not allowed for boolean types. Consider using '{1}' instead."),
            Block_scoped_variable_0_used_before_its_declaration: diag(2448, ts.DiagnosticCategory.Error, "Block_scoped_variable_0_used_before_its_declaration_2448", "Block-scoped variable '{0}' used before its declaration."),
            Class_0_used_before_its_declaration: diag(2449, ts.DiagnosticCategory.Error, "Class_0_used_before_its_declaration_2449", "Class '{0}' used before its declaration."),
            Enum_0_used_before_its_declaration: diag(2450, ts.DiagnosticCategory.Error, "Enum_0_used_before_its_declaration_2450", "Enum '{0}' used before its declaration."),
            Cannot_redeclare_block_scoped_variable_0: diag(2451, ts.DiagnosticCategory.Error, "Cannot_redeclare_block_scoped_variable_0_2451", "Cannot redeclare block-scoped variable '{0}'."),
            An_enum_member_cannot_have_a_numeric_name: diag(2452, ts.DiagnosticCategory.Error, "An_enum_member_cannot_have_a_numeric_name_2452", "An enum member cannot have a numeric name."),
            The_type_argument_for_type_parameter_0_cannot_be_inferred_from_the_usage_Consider_specifying_the_type_arguments_explicitly: diag(2453, ts.DiagnosticCategory.Error, "The_type_argument_for_type_parameter_0_cannot_be_inferred_from_the_usage_Consider_specifying_the_typ_2453", "The type argument for type parameter '{0}' cannot be inferred from the usage. Consider specifying the type arguments explicitly."),
            Variable_0_is_used_before_being_assigned: diag(2454, ts.DiagnosticCategory.Error, "Variable_0_is_used_before_being_assigned_2454", "Variable '{0}' is used before being assigned."),
            Type_argument_candidate_1_is_not_a_valid_type_argument_because_it_is_not_a_supertype_of_candidate_0: diag(2455, ts.DiagnosticCategory.Error, "Type_argument_candidate_1_is_not_a_valid_type_argument_because_it_is_not_a_supertype_of_candidate_0_2455", "Type argument candidate '{1}' is not a valid type argument because it is not a supertype of candidate '{0}'."),
            Type_alias_0_circularly_references_itself: diag(2456, ts.DiagnosticCategory.Error, "Type_alias_0_circularly_references_itself_2456", "Type alias '{0}' circularly references itself."),
            Type_alias_name_cannot_be_0: diag(2457, ts.DiagnosticCategory.Error, "Type_alias_name_cannot_be_0_2457", "Type alias name cannot be '{0}'."),
            An_AMD_module_cannot_have_multiple_name_assignments: diag(2458, ts.DiagnosticCategory.Error, "An_AMD_module_cannot_have_multiple_name_assignments_2458", "An AMD module cannot have multiple name assignments."),
            Module_0_declares_1_locally_but_it_is_not_exported: diag(2459, ts.DiagnosticCategory.Error, "Module_0_declares_1_locally_but_it_is_not_exported_2459", "Module '{0}' declares '{1}' locally, but it is not exported."),
            Module_0_declares_1_locally_but_it_is_exported_as_2: diag(2460, ts.DiagnosticCategory.Error, "Module_0_declares_1_locally_but_it_is_exported_as_2_2460", "Module '{0}' declares '{1}' locally, but it is exported as '{2}'."),
            Type_0_is_not_an_array_type: diag(2461, ts.DiagnosticCategory.Error, "Type_0_is_not_an_array_type_2461", "Type '{0}' is not an array type."),
            A_rest_element_must_be_last_in_a_destructuring_pattern: diag(2462, ts.DiagnosticCategory.Error, "A_rest_element_must_be_last_in_a_destructuring_pattern_2462", "A rest element must be last in a destructuring pattern."),
            A_binding_pattern_parameter_cannot_be_optional_in_an_implementation_signature: diag(2463, ts.DiagnosticCategory.Error, "A_binding_pattern_parameter_cannot_be_optional_in_an_implementation_signature_2463", "A binding pattern parameter cannot be optional in an implementation signature."),
            A_computed_property_name_must_be_of_type_string_number_symbol_or_any: diag(2464, ts.DiagnosticCategory.Error, "A_computed_property_name_must_be_of_type_string_number_symbol_or_any_2464", "A computed property name must be of type 'string', 'number', 'symbol', or 'any'."),
            this_cannot_be_referenced_in_a_computed_property_name: diag(2465, ts.DiagnosticCategory.Error, "this_cannot_be_referenced_in_a_computed_property_name_2465", "'this' cannot be referenced in a computed property name."),
            super_cannot_be_referenced_in_a_computed_property_name: diag(2466, ts.DiagnosticCategory.Error, "super_cannot_be_referenced_in_a_computed_property_name_2466", "'super' cannot be referenced in a computed property name."),
            A_computed_property_name_cannot_reference_a_type_parameter_from_its_containing_type: diag(2467, ts.DiagnosticCategory.Error, "A_computed_property_name_cannot_reference_a_type_parameter_from_its_containing_type_2467", "A computed property name cannot reference a type parameter from its containing type."),
            Cannot_find_global_value_0: diag(2468, ts.DiagnosticCategory.Error, "Cannot_find_global_value_0_2468", "Cannot find global value '{0}'."),
            The_0_operator_cannot_be_applied_to_type_symbol: diag(2469, ts.DiagnosticCategory.Error, "The_0_operator_cannot_be_applied_to_type_symbol_2469", "The '{0}' operator cannot be applied to type 'symbol'."),
            Symbol_reference_does_not_refer_to_the_global_Symbol_constructor_object: diag(2470, ts.DiagnosticCategory.Error, "Symbol_reference_does_not_refer_to_the_global_Symbol_constructor_object_2470", "'Symbol' reference does not refer to the global Symbol constructor object."),
            A_computed_property_name_of_the_form_0_must_be_of_type_symbol: diag(2471, ts.DiagnosticCategory.Error, "A_computed_property_name_of_the_form_0_must_be_of_type_symbol_2471", "A computed property name of the form '{0}' must be of type 'symbol'."),
            Spread_operator_in_new_expressions_is_only_available_when_targeting_ECMAScript_5_and_higher: diag(2472, ts.DiagnosticCategory.Error, "Spread_operator_in_new_expressions_is_only_available_when_targeting_ECMAScript_5_and_higher_2472", "Spread operator in 'new' expressions is only available when targeting ECMAScript 5 and higher."),
            Enum_declarations_must_all_be_const_or_non_const: diag(2473, ts.DiagnosticCategory.Error, "Enum_declarations_must_all_be_const_or_non_const_2473", "Enum declarations must all be const or non-const."),
            const_enum_member_initializers_can_only_contain_literal_values_and_other_computed_enum_values: diag(2474, ts.DiagnosticCategory.Error, "const_enum_member_initializers_can_only_contain_literal_values_and_other_computed_enum_values_2474", "const enum member initializers can only contain literal values and other computed enum values."),
            const_enums_can_only_be_used_in_property_or_index_access_expressions_or_the_right_hand_side_of_an_import_declaration_or_export_assignment_or_type_query: diag(2475, ts.DiagnosticCategory.Error, "const_enums_can_only_be_used_in_property_or_index_access_expressions_or_the_right_hand_side_of_an_im_2475", "'const' enums can only be used in property or index access expressions or the right hand side of an import declaration or export assignment or type query."),
            A_const_enum_member_can_only_be_accessed_using_a_string_literal: diag(2476, ts.DiagnosticCategory.Error, "A_const_enum_member_can_only_be_accessed_using_a_string_literal_2476", "A const enum member can only be accessed using a string literal."),
            const_enum_member_initializer_was_evaluated_to_a_non_finite_value: diag(2477, ts.DiagnosticCategory.Error, "const_enum_member_initializer_was_evaluated_to_a_non_finite_value_2477", "'const' enum member initializer was evaluated to a non-finite value."),
            const_enum_member_initializer_was_evaluated_to_disallowed_value_NaN: diag(2478, ts.DiagnosticCategory.Error, "const_enum_member_initializer_was_evaluated_to_disallowed_value_NaN_2478", "'const' enum member initializer was evaluated to disallowed value 'NaN'."),
            Property_0_does_not_exist_on_const_enum_1: diag(2479, ts.DiagnosticCategory.Error, "Property_0_does_not_exist_on_const_enum_1_2479", "Property '{0}' does not exist on 'const' enum '{1}'."),
            let_is_not_allowed_to_be_used_as_a_name_in_let_or_const_declarations: diag(2480, ts.DiagnosticCategory.Error, "let_is_not_allowed_to_be_used_as_a_name_in_let_or_const_declarations_2480", "'let' is not allowed to be used as a name in 'let' or 'const' declarations."),
            Cannot_initialize_outer_scoped_variable_0_in_the_same_scope_as_block_scoped_declaration_1: diag(2481, ts.DiagnosticCategory.Error, "Cannot_initialize_outer_scoped_variable_0_in_the_same_scope_as_block_scoped_declaration_1_2481", "Cannot initialize outer scoped variable '{0}' in the same scope as block scoped declaration '{1}'."),
            The_left_hand_side_of_a_for_of_statement_cannot_use_a_type_annotation: diag(2483, ts.DiagnosticCategory.Error, "The_left_hand_side_of_a_for_of_statement_cannot_use_a_type_annotation_2483", "The left-hand side of a 'for...of' statement cannot use a type annotation."),
            Export_declaration_conflicts_with_exported_declaration_of_0: diag(2484, ts.DiagnosticCategory.Error, "Export_declaration_conflicts_with_exported_declaration_of_0_2484", "Export declaration conflicts with exported declaration of '{0}'."),
            The_left_hand_side_of_a_for_of_statement_must_be_a_variable_or_a_property_access: diag(2487, ts.DiagnosticCategory.Error, "The_left_hand_side_of_a_for_of_statement_must_be_a_variable_or_a_property_access_2487", "The left-hand side of a 'for...of' statement must be a variable or a property access."),
            Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator: diag(2488, ts.DiagnosticCategory.Error, "Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator_2488", "Type '{0}' must have a '[Symbol.iterator]()' method that returns an iterator."),
            An_iterator_must_have_a_next_method: diag(2489, ts.DiagnosticCategory.Error, "An_iterator_must_have_a_next_method_2489", "An iterator must have a 'next()' method."),
            The_type_returned_by_the_0_method_of_an_iterator_must_have_a_value_property: diag(2490, ts.DiagnosticCategory.Error, "The_type_returned_by_the_0_method_of_an_iterator_must_have_a_value_property_2490", "The type returned by the '{0}()' method of an iterator must have a 'value' property."),
            The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern: diag(2491, ts.DiagnosticCategory.Error, "The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern_2491", "The left-hand side of a 'for...in' statement cannot be a destructuring pattern."),
            Cannot_redeclare_identifier_0_in_catch_clause: diag(2492, ts.DiagnosticCategory.Error, "Cannot_redeclare_identifier_0_in_catch_clause_2492", "Cannot redeclare identifier '{0}' in catch clause."),
            Tuple_type_0_of_length_1_has_no_element_at_index_2: diag(2493, ts.DiagnosticCategory.Error, "Tuple_type_0_of_length_1_has_no_element_at_index_2_2493", "Tuple type '{0}' of length '{1}' has no element at index '{2}'."),
            Using_a_string_in_a_for_of_statement_is_only_supported_in_ECMAScript_5_and_higher: diag(2494, ts.DiagnosticCategory.Error, "Using_a_string_in_a_for_of_statement_is_only_supported_in_ECMAScript_5_and_higher_2494", "Using a string in a 'for...of' statement is only supported in ECMAScript 5 and higher."),
            Type_0_is_not_an_array_type_or_a_string_type: diag(2495, ts.DiagnosticCategory.Error, "Type_0_is_not_an_array_type_or_a_string_type_2495", "Type '{0}' is not an array type or a string type."),
            The_arguments_object_cannot_be_referenced_in_an_arrow_function_in_ES3_and_ES5_Consider_using_a_standard_function_expression: diag(2496, ts.DiagnosticCategory.Error, "The_arguments_object_cannot_be_referenced_in_an_arrow_function_in_ES3_and_ES5_Consider_using_a_stand_2496", "The 'arguments' object cannot be referenced in an arrow function in ES3 and ES5. Consider using a standard function expression."),
            This_module_can_only_be_referenced_with_ECMAScript_imports_Slashexports_by_turning_on_the_0_flag_and_referencing_its_default_export: diag(2497, ts.DiagnosticCategory.Error, "This_module_can_only_be_referenced_with_ECMAScript_imports_Slashexports_by_turning_on_the_0_flag_and_2497", "This module can only be referenced with ECMAScript imports/exports by turning on the '{0}' flag and referencing its default export."),
            Module_0_uses_export_and_cannot_be_used_with_export_Asterisk: diag(2498, ts.DiagnosticCategory.Error, "Module_0_uses_export_and_cannot_be_used_with_export_Asterisk_2498", "Module '{0}' uses 'export =' and cannot be used with 'export *'."),
            An_interface_can_only_extend_an_identifier_Slashqualified_name_with_optional_type_arguments: diag(2499, ts.DiagnosticCategory.Error, "An_interface_can_only_extend_an_identifier_Slashqualified_name_with_optional_type_arguments_2499", "An interface can only extend an identifier/qualified-name with optional type arguments."),
            A_class_can_only_implement_an_identifier_Slashqualified_name_with_optional_type_arguments: diag(2500, ts.DiagnosticCategory.Error, "A_class_can_only_implement_an_identifier_Slashqualified_name_with_optional_type_arguments_2500", "A class can only implement an identifier/qualified-name with optional type arguments."),
            A_rest_element_cannot_contain_a_binding_pattern: diag(2501, ts.DiagnosticCategory.Error, "A_rest_element_cannot_contain_a_binding_pattern_2501", "A rest element cannot contain a binding pattern."),
            _0_is_referenced_directly_or_indirectly_in_its_own_type_annotation: diag(2502, ts.DiagnosticCategory.Error, "_0_is_referenced_directly_or_indirectly_in_its_own_type_annotation_2502", "'{0}' is referenced directly or indirectly in its own type annotation."),
            Cannot_find_namespace_0: diag(2503, ts.DiagnosticCategory.Error, "Cannot_find_namespace_0_2503", "Cannot find namespace '{0}'."),
            Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator: diag(2504, ts.DiagnosticCategory.Error, "Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator_2504", "Type '{0}' must have a '[Symbol.asyncIterator]()' method that returns an async iterator."),
            A_generator_cannot_have_a_void_type_annotation: diag(2505, ts.DiagnosticCategory.Error, "A_generator_cannot_have_a_void_type_annotation_2505", "A generator cannot have a 'void' type annotation."),
            _0_is_referenced_directly_or_indirectly_in_its_own_base_expression: diag(2506, ts.DiagnosticCategory.Error, "_0_is_referenced_directly_or_indirectly_in_its_own_base_expression_2506", "'{0}' is referenced directly or indirectly in its own base expression."),
            Type_0_is_not_a_constructor_function_type: diag(2507, ts.DiagnosticCategory.Error, "Type_0_is_not_a_constructor_function_type_2507", "Type '{0}' is not a constructor function type."),
            No_base_constructor_has_the_specified_number_of_type_arguments: diag(2508, ts.DiagnosticCategory.Error, "No_base_constructor_has_the_specified_number_of_type_arguments_2508", "No base constructor has the specified number of type arguments."),
            Base_constructor_return_type_0_is_not_an_object_type_or_intersection_of_object_types_with_statically_known_members: diag(2509, ts.DiagnosticCategory.Error, "Base_constructor_return_type_0_is_not_an_object_type_or_intersection_of_object_types_with_statically_2509", "Base constructor return type '{0}' is not an object type or intersection of object types with statically known members."),
            Base_constructors_must_all_have_the_same_return_type: diag(2510, ts.DiagnosticCategory.Error, "Base_constructors_must_all_have_the_same_return_type_2510", "Base constructors must all have the same return type."),
            Cannot_create_an_instance_of_an_abstract_class: diag(2511, ts.DiagnosticCategory.Error, "Cannot_create_an_instance_of_an_abstract_class_2511", "Cannot create an instance of an abstract class."),
            Overload_signatures_must_all_be_abstract_or_non_abstract: diag(2512, ts.DiagnosticCategory.Error, "Overload_signatures_must_all_be_abstract_or_non_abstract_2512", "Overload signatures must all be abstract or non-abstract."),
            Abstract_method_0_in_class_1_cannot_be_accessed_via_super_expression: diag(2513, ts.DiagnosticCategory.Error, "Abstract_method_0_in_class_1_cannot_be_accessed_via_super_expression_2513", "Abstract method '{0}' in class '{1}' cannot be accessed via super expression."),
            Classes_containing_abstract_methods_must_be_marked_abstract: diag(2514, ts.DiagnosticCategory.Error, "Classes_containing_abstract_methods_must_be_marked_abstract_2514", "Classes containing abstract methods must be marked abstract."),
            Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2: diag(2515, ts.DiagnosticCategory.Error, "Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2_2515", "Non-abstract class '{0}' does not implement inherited abstract member '{1}' from class '{2}'."),
            All_declarations_of_an_abstract_method_must_be_consecutive: diag(2516, ts.DiagnosticCategory.Error, "All_declarations_of_an_abstract_method_must_be_consecutive_2516", "All declarations of an abstract method must be consecutive."),
            Cannot_assign_an_abstract_constructor_type_to_a_non_abstract_constructor_type: diag(2517, ts.DiagnosticCategory.Error, "Cannot_assign_an_abstract_constructor_type_to_a_non_abstract_constructor_type_2517", "Cannot assign an abstract constructor type to a non-abstract constructor type."),
            A_this_based_type_guard_is_not_compatible_with_a_parameter_based_type_guard: diag(2518, ts.DiagnosticCategory.Error, "A_this_based_type_guard_is_not_compatible_with_a_parameter_based_type_guard_2518", "A 'this'-based type guard is not compatible with a parameter-based type guard."),
            An_async_iterator_must_have_a_next_method: diag(2519, ts.DiagnosticCategory.Error, "An_async_iterator_must_have_a_next_method_2519", "An async iterator must have a 'next()' method."),
            Duplicate_identifier_0_Compiler_uses_declaration_1_to_support_async_functions: diag(2520, ts.DiagnosticCategory.Error, "Duplicate_identifier_0_Compiler_uses_declaration_1_to_support_async_functions_2520", "Duplicate identifier '{0}'. Compiler uses declaration '{1}' to support async functions."),
            Expression_resolves_to_variable_declaration_0_that_compiler_uses_to_support_async_functions: diag(2521, ts.DiagnosticCategory.Error, "Expression_resolves_to_variable_declaration_0_that_compiler_uses_to_support_async_functions_2521", "Expression resolves to variable declaration '{0}' that compiler uses to support async functions."),
            The_arguments_object_cannot_be_referenced_in_an_async_function_or_method_in_ES3_and_ES5_Consider_using_a_standard_function_or_method: diag(2522, ts.DiagnosticCategory.Error, "The_arguments_object_cannot_be_referenced_in_an_async_function_or_method_in_ES3_and_ES5_Consider_usi_2522", "The 'arguments' object cannot be referenced in an async function or method in ES3 and ES5. Consider using a standard function or method."),
            yield_expressions_cannot_be_used_in_a_parameter_initializer: diag(2523, ts.DiagnosticCategory.Error, "yield_expressions_cannot_be_used_in_a_parameter_initializer_2523", "'yield' expressions cannot be used in a parameter initializer."),
            await_expressions_cannot_be_used_in_a_parameter_initializer: diag(2524, ts.DiagnosticCategory.Error, "await_expressions_cannot_be_used_in_a_parameter_initializer_2524", "'await' expressions cannot be used in a parameter initializer."),
            Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value: diag(2525, ts.DiagnosticCategory.Error, "Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value_2525", "Initializer provides no value for this binding element and the binding element has no default value."),
            A_this_type_is_available_only_in_a_non_static_member_of_a_class_or_interface: diag(2526, ts.DiagnosticCategory.Error, "A_this_type_is_available_only_in_a_non_static_member_of_a_class_or_interface_2526", "A 'this' type is available only in a non-static member of a class or interface."),
            The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary: diag(2527, ts.DiagnosticCategory.Error, "The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary_2527", "The inferred type of '{0}' references an inaccessible '{1}' type. A type annotation is necessary."),
            A_module_cannot_have_multiple_default_exports: diag(2528, ts.DiagnosticCategory.Error, "A_module_cannot_have_multiple_default_exports_2528", "A module cannot have multiple default exports."),
            Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_containing_async_functions: diag(2529, ts.DiagnosticCategory.Error, "Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_containing_async_func_2529", "Duplicate identifier '{0}'. Compiler reserves name '{1}' in top level scope of a module containing async functions."),
            Property_0_is_incompatible_with_index_signature: diag(2530, ts.DiagnosticCategory.Error, "Property_0_is_incompatible_with_index_signature_2530", "Property '{0}' is incompatible with index signature."),
            Object_is_possibly_null: diag(2531, ts.DiagnosticCategory.Error, "Object_is_possibly_null_2531", "Object is possibly 'null'."),
            Object_is_possibly_undefined: diag(2532, ts.DiagnosticCategory.Error, "Object_is_possibly_undefined_2532", "Object is possibly 'undefined'."),
            Object_is_possibly_null_or_undefined: diag(2533, ts.DiagnosticCategory.Error, "Object_is_possibly_null_or_undefined_2533", "Object is possibly 'null' or 'undefined'."),
            A_function_returning_never_cannot_have_a_reachable_end_point: diag(2534, ts.DiagnosticCategory.Error, "A_function_returning_never_cannot_have_a_reachable_end_point_2534", "A function returning 'never' cannot have a reachable end point."),
            Enum_type_0_has_members_with_initializers_that_are_not_literals: diag(2535, ts.DiagnosticCategory.Error, "Enum_type_0_has_members_with_initializers_that_are_not_literals_2535", "Enum type '{0}' has members with initializers that are not literals."),
            Type_0_cannot_be_used_to_index_type_1: diag(2536, ts.DiagnosticCategory.Error, "Type_0_cannot_be_used_to_index_type_1_2536", "Type '{0}' cannot be used to index type '{1}'."),
            Type_0_has_no_matching_index_signature_for_type_1: diag(2537, ts.DiagnosticCategory.Error, "Type_0_has_no_matching_index_signature_for_type_1_2537", "Type '{0}' has no matching index signature for type '{1}'."),
            Type_0_cannot_be_used_as_an_index_type: diag(2538, ts.DiagnosticCategory.Error, "Type_0_cannot_be_used_as_an_index_type_2538", "Type '{0}' cannot be used as an index type."),
            Cannot_assign_to_0_because_it_is_not_a_variable: diag(2539, ts.DiagnosticCategory.Error, "Cannot_assign_to_0_because_it_is_not_a_variable_2539", "Cannot assign to '{0}' because it is not a variable."),
            Cannot_assign_to_0_because_it_is_a_read_only_property: diag(2540, ts.DiagnosticCategory.Error, "Cannot_assign_to_0_because_it_is_a_read_only_property_2540", "Cannot assign to '{0}' because it is a read-only property."),
            The_target_of_an_assignment_must_be_a_variable_or_a_property_access: diag(2541, ts.DiagnosticCategory.Error, "The_target_of_an_assignment_must_be_a_variable_or_a_property_access_2541", "The target of an assignment must be a variable or a property access."),
            Index_signature_in_type_0_only_permits_reading: diag(2542, ts.DiagnosticCategory.Error, "Index_signature_in_type_0_only_permits_reading_2542", "Index signature in type '{0}' only permits reading."),
            Duplicate_identifier_newTarget_Compiler_uses_variable_declaration_newTarget_to_capture_new_target_meta_property_reference: diag(2543, ts.DiagnosticCategory.Error, "Duplicate_identifier_newTarget_Compiler_uses_variable_declaration_newTarget_to_capture_new_target_me_2543", "Duplicate identifier '_newTarget'. Compiler uses variable declaration '_newTarget' to capture 'new.target' meta-property reference."),
            Expression_resolves_to_variable_declaration_newTarget_that_compiler_uses_to_capture_new_target_meta_property_reference: diag(2544, ts.DiagnosticCategory.Error, "Expression_resolves_to_variable_declaration_newTarget_that_compiler_uses_to_capture_new_target_meta__2544", "Expression resolves to variable declaration '_newTarget' that compiler uses to capture 'new.target' meta-property reference."),
            A_mixin_class_must_have_a_constructor_with_a_single_rest_parameter_of_type_any: diag(2545, ts.DiagnosticCategory.Error, "A_mixin_class_must_have_a_constructor_with_a_single_rest_parameter_of_type_any_2545", "A mixin class must have a constructor with a single rest parameter of type 'any[]'."),
            The_type_returned_by_the_0_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value_property: diag(2547, ts.DiagnosticCategory.Error, "The_type_returned_by_the_0_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value_pro_2547", "The type returned by the '{0}()' method of an async iterator must be a promise for a type with a 'value' property."),
            Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator: diag(2548, ts.DiagnosticCategory.Error, "Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator_2548", "Type '{0}' is not an array type or does not have a '[Symbol.iterator]()' method that returns an iterator."),
            Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator: diag(2549, ts.DiagnosticCategory.Error, "Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns__2549", "Type '{0}' is not an array type or a string type or does not have a '[Symbol.iterator]()' method that returns an iterator."),
            Property_0_does_not_exist_on_type_1_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2_or_later: diag(2550, ts.DiagnosticCategory.Error, "Property_0_does_not_exist_on_type_1_Do_you_need_to_change_your_target_library_Try_changing_the_lib_c_2550", "Property '{0}' does not exist on type '{1}'. Do you need to change your target library? Try changing the 'lib' compiler option to '{2}' or later."),
            Property_0_does_not_exist_on_type_1_Did_you_mean_2: diag(2551, ts.DiagnosticCategory.Error, "Property_0_does_not_exist_on_type_1_Did_you_mean_2_2551", "Property '{0}' does not exist on type '{1}'. Did you mean '{2}'?"),
            Cannot_find_name_0_Did_you_mean_1: diag(2552, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Did_you_mean_1_2552", "Cannot find name '{0}'. Did you mean '{1}'?"),
            Computed_values_are_not_permitted_in_an_enum_with_string_valued_members: diag(2553, ts.DiagnosticCategory.Error, "Computed_values_are_not_permitted_in_an_enum_with_string_valued_members_2553", "Computed values are not permitted in an enum with string valued members."),
            Expected_0_arguments_but_got_1: diag(2554, ts.DiagnosticCategory.Error, "Expected_0_arguments_but_got_1_2554", "Expected {0} arguments, but got {1}."),
            Expected_at_least_0_arguments_but_got_1: diag(2555, ts.DiagnosticCategory.Error, "Expected_at_least_0_arguments_but_got_1_2555", "Expected at least {0} arguments, but got {1}."),
            A_spread_argument_must_either_have_a_tuple_type_or_be_passed_to_a_rest_parameter: diag(2556, ts.DiagnosticCategory.Error, "A_spread_argument_must_either_have_a_tuple_type_or_be_passed_to_a_rest_parameter_2556", "A spread argument must either have a tuple type or be passed to a rest parameter."),
            Expected_0_type_arguments_but_got_1: diag(2558, ts.DiagnosticCategory.Error, "Expected_0_type_arguments_but_got_1_2558", "Expected {0} type arguments, but got {1}."),
            Type_0_has_no_properties_in_common_with_type_1: diag(2559, ts.DiagnosticCategory.Error, "Type_0_has_no_properties_in_common_with_type_1_2559", "Type '{0}' has no properties in common with type '{1}'."),
            Value_of_type_0_has_no_properties_in_common_with_type_1_Did_you_mean_to_call_it: diag(2560, ts.DiagnosticCategory.Error, "Value_of_type_0_has_no_properties_in_common_with_type_1_Did_you_mean_to_call_it_2560", "Value of type '{0}' has no properties in common with type '{1}'. Did you mean to call it?"),
            Object_literal_may_only_specify_known_properties_but_0_does_not_exist_in_type_1_Did_you_mean_to_write_2: diag(2561, ts.DiagnosticCategory.Error, "Object_literal_may_only_specify_known_properties_but_0_does_not_exist_in_type_1_Did_you_mean_to_writ_2561", "Object literal may only specify known properties, but '{0}' does not exist in type '{1}'. Did you mean to write '{2}'?"),
            Base_class_expressions_cannot_reference_class_type_parameters: diag(2562, ts.DiagnosticCategory.Error, "Base_class_expressions_cannot_reference_class_type_parameters_2562", "Base class expressions cannot reference class type parameters."),
            The_containing_function_or_module_body_is_too_large_for_control_flow_analysis: diag(2563, ts.DiagnosticCategory.Error, "The_containing_function_or_module_body_is_too_large_for_control_flow_analysis_2563", "The containing function or module body is too large for control flow analysis."),
            Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor: diag(2564, ts.DiagnosticCategory.Error, "Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor_2564", "Property '{0}' has no initializer and is not definitely assigned in the constructor."),
            Property_0_is_used_before_being_assigned: diag(2565, ts.DiagnosticCategory.Error, "Property_0_is_used_before_being_assigned_2565", "Property '{0}' is used before being assigned."),
            A_rest_element_cannot_have_a_property_name: diag(2566, ts.DiagnosticCategory.Error, "A_rest_element_cannot_have_a_property_name_2566", "A rest element cannot have a property name."),
            Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations: diag(2567, ts.DiagnosticCategory.Error, "Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations_2567", "Enum declarations can only merge with namespace or other enum declarations."),
            Property_0_may_not_exist_on_type_1_Did_you_mean_2: diag(2568, ts.DiagnosticCategory.Error, "Property_0_may_not_exist_on_type_1_Did_you_mean_2_2568", "Property '{0}' may not exist on type '{1}'. Did you mean '{2}'?"),
            Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators: diag(2569, ts.DiagnosticCategory.Error, "Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterati_2569", "Type '{0}' is not an array type or a string type. Use compiler option '--downlevelIteration' to allow iterating of iterators."),
            Could_not_find_name_0_Did_you_mean_1: diag(2570, ts.DiagnosticCategory.Error, "Could_not_find_name_0_Did_you_mean_1_2570", "Could not find name '{0}'. Did you mean '{1}'?"),
            Object_is_of_type_unknown: diag(2571, ts.DiagnosticCategory.Error, "Object_is_of_type_unknown_2571", "Object is of type 'unknown'."),
            Rest_signatures_are_incompatible: diag(2572, ts.DiagnosticCategory.Error, "Rest_signatures_are_incompatible_2572", "Rest signatures are incompatible."),
            Property_0_is_incompatible_with_rest_element_type: diag(2573, ts.DiagnosticCategory.Error, "Property_0_is_incompatible_with_rest_element_type_2573", "Property '{0}' is incompatible with rest element type."),
            A_rest_element_type_must_be_an_array_type: diag(2574, ts.DiagnosticCategory.Error, "A_rest_element_type_must_be_an_array_type_2574", "A rest element type must be an array type."),
            No_overload_expects_0_arguments_but_overloads_do_exist_that_expect_either_1_or_2_arguments: diag(2575, ts.DiagnosticCategory.Error, "No_overload_expects_0_arguments_but_overloads_do_exist_that_expect_either_1_or_2_arguments_2575", "No overload expects {0} arguments, but overloads do exist that expect either {1} or {2} arguments."),
            Property_0_does_not_exist_on_type_1_Did_you_mean_to_access_the_static_member_2_instead: diag(2576, ts.DiagnosticCategory.Error, "Property_0_does_not_exist_on_type_1_Did_you_mean_to_access_the_static_member_2_instead_2576", "Property '{0}' does not exist on type '{1}'. Did you mean to access the static member '{2}' instead?"),
            Return_type_annotation_circularly_references_itself: diag(2577, ts.DiagnosticCategory.Error, "Return_type_annotation_circularly_references_itself_2577", "Return type annotation circularly references itself."),
            Unused_ts_expect_error_directive: diag(2578, ts.DiagnosticCategory.Error, "Unused_ts_expect_error_directive_2578", "Unused '@ts-expect-error' directive."),
            Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashnode: diag(2580, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashno_2580", "Cannot find name '{0}'. Do you need to install type definitions for node? Try `npm i --save-dev @types/node`."),
            Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slashjquery: diag(2581, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slash_2581", "Cannot find name '{0}'. Do you need to install type definitions for jQuery? Try `npm i --save-dev @types/jquery`."),
            Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_types_Slashjest_or_npm_i_save_dev_types_Slashmocha: diag(2582, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_type_2582", "Cannot find name '{0}'. Do you need to install type definitions for a test runner? Try `npm i --save-dev @types/jest` or `npm i --save-dev @types/mocha`."),
            Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_1_or_later: diag(2583, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2583", "Cannot find name '{0}'. Do you need to change your target library? Try changing the 'lib' compiler option to '{1}' or later."),
            Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_include_dom: diag(2584, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2584", "Cannot find name '{0}'. Do you need to change your target library? Try changing the 'lib' compiler option to include 'dom'."),
            _0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_es2015_or_later: diag(2585, ts.DiagnosticCategory.Error, "_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Do_you_need_to_change_your_target_library_2585", "'{0}' only refers to a type, but is being used as a value here. Do you need to change your target library? Try changing the 'lib' compiler option to es2015 or later."),
            Enum_type_0_circularly_references_itself: diag(2586, ts.DiagnosticCategory.Error, "Enum_type_0_circularly_references_itself_2586", "Enum type '{0}' circularly references itself."),
            JSDoc_type_0_circularly_references_itself: diag(2587, ts.DiagnosticCategory.Error, "JSDoc_type_0_circularly_references_itself_2587", "JSDoc type '{0}' circularly references itself."),
            Cannot_assign_to_0_because_it_is_a_constant: diag(2588, ts.DiagnosticCategory.Error, "Cannot_assign_to_0_because_it_is_a_constant_2588", "Cannot assign to '{0}' because it is a constant."),
            Type_instantiation_is_excessively_deep_and_possibly_infinite: diag(2589, ts.DiagnosticCategory.Error, "Type_instantiation_is_excessively_deep_and_possibly_infinite_2589", "Type instantiation is excessively deep and possibly infinite."),
            Expression_produces_a_union_type_that_is_too_complex_to_represent: diag(2590, ts.DiagnosticCategory.Error, "Expression_produces_a_union_type_that_is_too_complex_to_represent_2590", "Expression produces a union type that is too complex to represent."),
            Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashnode_and_then_add_node_to_the_types_field_in_your_tsconfig: diag(2591, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashno_2591", "Cannot find name '{0}'. Do you need to install type definitions for node? Try `npm i --save-dev @types/node` and then add 'node' to the types field in your tsconfig."),
            Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slashjquery_and_then_add_jquery_to_the_types_field_in_your_tsconfig: diag(2592, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slash_2592", "Cannot find name '{0}'. Do you need to install type definitions for jQuery? Try `npm i --save-dev @types/jquery` and then add 'jquery' to the types field in your tsconfig."),
            Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_types_Slashjest_or_npm_i_save_dev_types_Slashmocha_and_then_add_jest_or_mocha_to_the_types_field_in_your_tsconfig: diag(2593, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_type_2593", "Cannot find name '{0}'. Do you need to install type definitions for a test runner? Try `npm i --save-dev @types/jest` or `npm i --save-dev @types/mocha` and then add 'jest' or 'mocha' to the types field in your tsconfig."),
            This_module_is_declared_with_using_export_and_can_only_be_used_with_a_default_import_when_using_the_0_flag: diag(2594, ts.DiagnosticCategory.Error, "This_module_is_declared_with_using_export_and_can_only_be_used_with_a_default_import_when_using_the__2594", "This module is declared with using 'export =', and can only be used with a default import when using the '{0}' flag."),
            _0_can_only_be_imported_by_using_a_default_import: diag(2595, ts.DiagnosticCategory.Error, "_0_can_only_be_imported_by_using_a_default_import_2595", "'{0}' can only be imported by using a default import."),
            _0_can_only_be_imported_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import: diag(2596, ts.DiagnosticCategory.Error, "_0_can_only_be_imported_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import_2596", "'{0}' can only be imported by turning on the 'esModuleInterop' flag and using a default import."),
            _0_can_only_be_imported_by_using_a_require_call_or_by_using_a_default_import: diag(2597, ts.DiagnosticCategory.Error, "_0_can_only_be_imported_by_using_a_require_call_or_by_using_a_default_import_2597", "'{0}' can only be imported by using a 'require' call or by using a default import."),
            _0_can_only_be_imported_by_using_a_require_call_or_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import: diag(2598, ts.DiagnosticCategory.Error, "_0_can_only_be_imported_by_using_a_require_call_or_by_turning_on_the_esModuleInterop_flag_and_using__2598", "'{0}' can only be imported by using a 'require' call or by turning on the 'esModuleInterop' flag and using a default import."),
            JSX_element_attributes_type_0_may_not_be_a_union_type: diag(2600, ts.DiagnosticCategory.Error, "JSX_element_attributes_type_0_may_not_be_a_union_type_2600", "JSX element attributes type '{0}' may not be a union type."),
            The_return_type_of_a_JSX_element_constructor_must_return_an_object_type: diag(2601, ts.DiagnosticCategory.Error, "The_return_type_of_a_JSX_element_constructor_must_return_an_object_type_2601", "The return type of a JSX element constructor must return an object type."),
            JSX_element_implicitly_has_type_any_because_the_global_type_JSX_Element_does_not_exist: diag(2602, ts.DiagnosticCategory.Error, "JSX_element_implicitly_has_type_any_because_the_global_type_JSX_Element_does_not_exist_2602", "JSX element implicitly has type 'any' because the global type 'JSX.Element' does not exist."),
            Property_0_in_type_1_is_not_assignable_to_type_2: diag(2603, ts.DiagnosticCategory.Error, "Property_0_in_type_1_is_not_assignable_to_type_2_2603", "Property '{0}' in type '{1}' is not assignable to type '{2}'."),
            JSX_element_type_0_does_not_have_any_construct_or_call_signatures: diag(2604, ts.DiagnosticCategory.Error, "JSX_element_type_0_does_not_have_any_construct_or_call_signatures_2604", "JSX element type '{0}' does not have any construct or call signatures."),
            JSX_element_type_0_is_not_a_constructor_function_for_JSX_elements: diag(2605, ts.DiagnosticCategory.Error, "JSX_element_type_0_is_not_a_constructor_function_for_JSX_elements_2605", "JSX element type '{0}' is not a constructor function for JSX elements."),
            Property_0_of_JSX_spread_attribute_is_not_assignable_to_target_property: diag(2606, ts.DiagnosticCategory.Error, "Property_0_of_JSX_spread_attribute_is_not_assignable_to_target_property_2606", "Property '{0}' of JSX spread attribute is not assignable to target property."),
            JSX_element_class_does_not_support_attributes_because_it_does_not_have_a_0_property: diag(2607, ts.DiagnosticCategory.Error, "JSX_element_class_does_not_support_attributes_because_it_does_not_have_a_0_property_2607", "JSX element class does not support attributes because it does not have a '{0}' property."),
            The_global_type_JSX_0_may_not_have_more_than_one_property: diag(2608, ts.DiagnosticCategory.Error, "The_global_type_JSX_0_may_not_have_more_than_one_property_2608", "The global type 'JSX.{0}' may not have more than one property."),
            JSX_spread_child_must_be_an_array_type: diag(2609, ts.DiagnosticCategory.Error, "JSX_spread_child_must_be_an_array_type_2609", "JSX spread child must be an array type."),
            _0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property: diag(2610, ts.DiagnosticCategory.Error, "_0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property_2610", "'{0}' is defined as an accessor in class '{1}', but is overridden here in '{2}' as an instance property."),
            _0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor: diag(2611, ts.DiagnosticCategory.Error, "_0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor_2611", "'{0}' is defined as a property in class '{1}', but is overridden here in '{2}' as an accessor."),
            Property_0_will_overwrite_the_base_property_in_1_If_this_is_intentional_add_an_initializer_Otherwise_add_a_declare_modifier_or_remove_the_redundant_declaration: diag(2612, ts.DiagnosticCategory.Error, "Property_0_will_overwrite_the_base_property_in_1_If_this_is_intentional_add_an_initializer_Otherwise_2612", "Property '{0}' will overwrite the base property in '{1}'. If this is intentional, add an initializer. Otherwise, add a 'declare' modifier or remove the redundant declaration."),
            Module_0_has_no_default_export_Did_you_mean_to_use_import_1_from_0_instead: diag(2613, ts.DiagnosticCategory.Error, "Module_0_has_no_default_export_Did_you_mean_to_use_import_1_from_0_instead_2613", "Module '{0}' has no default export. Did you mean to use 'import { {1} } from {0}' instead?"),
            Module_0_has_no_exported_member_1_Did_you_mean_to_use_import_1_from_0_instead: diag(2614, ts.DiagnosticCategory.Error, "Module_0_has_no_exported_member_1_Did_you_mean_to_use_import_1_from_0_instead_2614", "Module '{0}' has no exported member '{1}'. Did you mean to use 'import {1} from {0}' instead?"),
            Type_of_property_0_circularly_references_itself_in_mapped_type_1: diag(2615, ts.DiagnosticCategory.Error, "Type_of_property_0_circularly_references_itself_in_mapped_type_1_2615", "Type of property '{0}' circularly references itself in mapped type '{1}'."),
            _0_can_only_be_imported_by_using_import_1_require_2_or_a_default_import: diag(2616, ts.DiagnosticCategory.Error, "_0_can_only_be_imported_by_using_import_1_require_2_or_a_default_import_2616", "'{0}' can only be imported by using 'import {1} = require({2})' or a default import."),
            _0_can_only_be_imported_by_using_import_1_require_2_or_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import: diag(2617, ts.DiagnosticCategory.Error, "_0_can_only_be_imported_by_using_import_1_require_2_or_by_turning_on_the_esModuleInterop_flag_and_us_2617", "'{0}' can only be imported by using 'import {1} = require({2})' or by turning on the 'esModuleInterop' flag and using a default import."),
            Source_has_0_element_s_but_target_requires_1: diag(2618, ts.DiagnosticCategory.Error, "Source_has_0_element_s_but_target_requires_1_2618", "Source has {0} element(s) but target requires {1}."),
            Source_has_0_element_s_but_target_allows_only_1: diag(2619, ts.DiagnosticCategory.Error, "Source_has_0_element_s_but_target_allows_only_1_2619", "Source has {0} element(s) but target allows only {1}."),
            Target_requires_0_element_s_but_source_may_have_fewer: diag(2620, ts.DiagnosticCategory.Error, "Target_requires_0_element_s_but_source_may_have_fewer_2620", "Target requires {0} element(s) but source may have fewer."),
            Target_allows_only_0_element_s_but_source_may_have_more: diag(2621, ts.DiagnosticCategory.Error, "Target_allows_only_0_element_s_but_source_may_have_more_2621", "Target allows only {0} element(s) but source may have more."),
            Source_provides_no_match_for_required_element_at_position_0_in_target: diag(2623, ts.DiagnosticCategory.Error, "Source_provides_no_match_for_required_element_at_position_0_in_target_2623", "Source provides no match for required element at position {0} in target."),
            Source_provides_no_match_for_variadic_element_at_position_0_in_target: diag(2624, ts.DiagnosticCategory.Error, "Source_provides_no_match_for_variadic_element_at_position_0_in_target_2624", "Source provides no match for variadic element at position {0} in target."),
            Variadic_element_at_position_0_in_source_does_not_match_element_at_position_1_in_target: diag(2625, ts.DiagnosticCategory.Error, "Variadic_element_at_position_0_in_source_does_not_match_element_at_position_1_in_target_2625", "Variadic element at position {0} in source does not match element at position {1} in target."),
            Type_at_position_0_in_source_is_not_compatible_with_type_at_position_1_in_target: diag(2626, ts.DiagnosticCategory.Error, "Type_at_position_0_in_source_is_not_compatible_with_type_at_position_1_in_target_2626", "Type at position {0} in source is not compatible with type at position {1} in target."),
            Type_at_positions_0_through_1_in_source_is_not_compatible_with_type_at_position_2_in_target: diag(2627, ts.DiagnosticCategory.Error, "Type_at_positions_0_through_1_in_source_is_not_compatible_with_type_at_position_2_in_target_2627", "Type at positions {0} through {1} in source is not compatible with type at position {2} in target."),
            Cannot_assign_to_0_because_it_is_an_enum: diag(2628, ts.DiagnosticCategory.Error, "Cannot_assign_to_0_because_it_is_an_enum_2628", "Cannot assign to '{0}' because it is an enum."),
            Cannot_assign_to_0_because_it_is_a_class: diag(2629, ts.DiagnosticCategory.Error, "Cannot_assign_to_0_because_it_is_a_class_2629", "Cannot assign to '{0}' because it is a class."),
            Cannot_assign_to_0_because_it_is_a_function: diag(2630, ts.DiagnosticCategory.Error, "Cannot_assign_to_0_because_it_is_a_function_2630", "Cannot assign to '{0}' because it is a function."),
            Cannot_assign_to_0_because_it_is_a_namespace: diag(2631, ts.DiagnosticCategory.Error, "Cannot_assign_to_0_because_it_is_a_namespace_2631", "Cannot assign to '{0}' because it is a namespace."),
            Cannot_assign_to_0_because_it_is_an_import: diag(2632, ts.DiagnosticCategory.Error, "Cannot_assign_to_0_because_it_is_an_import_2632", "Cannot assign to '{0}' because it is an import."),
            JSX_property_access_expressions_cannot_include_JSX_namespace_names: diag(2633, ts.DiagnosticCategory.Error, "JSX_property_access_expressions_cannot_include_JSX_namespace_names_2633", "JSX property access expressions cannot include JSX namespace names"),
            _0_index_signatures_are_incompatible: diag(2634, ts.DiagnosticCategory.Error, "_0_index_signatures_are_incompatible_2634", "'{0}' index signatures are incompatible."),
            Cannot_augment_module_0_with_value_exports_because_it_resolves_to_a_non_module_entity: diag(2649, ts.DiagnosticCategory.Error, "Cannot_augment_module_0_with_value_exports_because_it_resolves_to_a_non_module_entity_2649", "Cannot augment module '{0}' with value exports because it resolves to a non-module entity."),
            A_member_initializer_in_a_enum_declaration_cannot_reference_members_declared_after_it_including_members_defined_in_other_enums: diag(2651, ts.DiagnosticCategory.Error, "A_member_initializer_in_a_enum_declaration_cannot_reference_members_declared_after_it_including_memb_2651", "A member initializer in a enum declaration cannot reference members declared after it, including members defined in other enums."),
            Merged_declaration_0_cannot_include_a_default_export_declaration_Consider_adding_a_separate_export_default_0_declaration_instead: diag(2652, ts.DiagnosticCategory.Error, "Merged_declaration_0_cannot_include_a_default_export_declaration_Consider_adding_a_separate_export_d_2652", "Merged declaration '{0}' cannot include a default export declaration. Consider adding a separate 'export default {0}' declaration instead."),
            Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1: diag(2653, ts.DiagnosticCategory.Error, "Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1_2653", "Non-abstract class expression does not implement inherited abstract member '{0}' from class '{1}'."),
            Exported_external_package_typings_file_cannot_contain_tripleslash_references_Please_contact_the_package_author_to_update_the_package_definition: diag(2654, ts.DiagnosticCategory.Error, "Exported_external_package_typings_file_cannot_contain_tripleslash_references_Please_contact_the_pack_2654", "Exported external package typings file cannot contain tripleslash references. Please contact the package author to update the package definition."),
            Exported_external_package_typings_file_0_is_not_a_module_Please_contact_the_package_author_to_update_the_package_definition: diag(2656, ts.DiagnosticCategory.Error, "Exported_external_package_typings_file_0_is_not_a_module_Please_contact_the_package_author_to_update_2656", "Exported external package typings file '{0}' is not a module. Please contact the package author to update the package definition."),
            JSX_expressions_must_have_one_parent_element: diag(2657, ts.DiagnosticCategory.Error, "JSX_expressions_must_have_one_parent_element_2657", "JSX expressions must have one parent element."),
            Type_0_provides_no_match_for_the_signature_1: diag(2658, ts.DiagnosticCategory.Error, "Type_0_provides_no_match_for_the_signature_1_2658", "Type '{0}' provides no match for the signature '{1}'."),
            super_is_only_allowed_in_members_of_object_literal_expressions_when_option_target_is_ES2015_or_higher: diag(2659, ts.DiagnosticCategory.Error, "super_is_only_allowed_in_members_of_object_literal_expressions_when_option_target_is_ES2015_or_highe_2659", "'super' is only allowed in members of object literal expressions when option 'target' is 'ES2015' or higher."),
            super_can_only_be_referenced_in_members_of_derived_classes_or_object_literal_expressions: diag(2660, ts.DiagnosticCategory.Error, "super_can_only_be_referenced_in_members_of_derived_classes_or_object_literal_expressions_2660", "'super' can only be referenced in members of derived classes or object literal expressions."),
            Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module: diag(2661, ts.DiagnosticCategory.Error, "Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module_2661", "Cannot export '{0}'. Only local declarations can be exported from a module."),
            Cannot_find_name_0_Did_you_mean_the_static_member_1_0: diag(2662, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Did_you_mean_the_static_member_1_0_2662", "Cannot find name '{0}'. Did you mean the static member '{1}.{0}'?"),
            Cannot_find_name_0_Did_you_mean_the_instance_member_this_0: diag(2663, ts.DiagnosticCategory.Error, "Cannot_find_name_0_Did_you_mean_the_instance_member_this_0_2663", "Cannot find name '{0}'. Did you mean the instance member 'this.{0}'?"),
            Invalid_module_name_in_augmentation_module_0_cannot_be_found: diag(2664, ts.DiagnosticCategory.Error, "Invalid_module_name_in_augmentation_module_0_cannot_be_found_2664", "Invalid module name in augmentation, module '{0}' cannot be found."),
            Invalid_module_name_in_augmentation_Module_0_resolves_to_an_untyped_module_at_1_which_cannot_be_augmented: diag(2665, ts.DiagnosticCategory.Error, "Invalid_module_name_in_augmentation_Module_0_resolves_to_an_untyped_module_at_1_which_cannot_be_augm_2665", "Invalid module name in augmentation. Module '{0}' resolves to an untyped module at '{1}', which cannot be augmented."),
            Exports_and_export_assignments_are_not_permitted_in_module_augmentations: diag(2666, ts.DiagnosticCategory.Error, "Exports_and_export_assignments_are_not_permitted_in_module_augmentations_2666", "Exports and export assignments are not permitted in module augmentations."),
            Imports_are_not_permitted_in_module_augmentations_Consider_moving_them_to_the_enclosing_external_module: diag(2667, ts.DiagnosticCategory.Error, "Imports_are_not_permitted_in_module_augmentations_Consider_moving_them_to_the_enclosing_external_mod_2667", "Imports are not permitted in module augmentations. Consider moving them to the enclosing external module."),
            export_modifier_cannot_be_applied_to_ambient_modules_and_module_augmentations_since_they_are_always_visible: diag(2668, ts.DiagnosticCategory.Error, "export_modifier_cannot_be_applied_to_ambient_modules_and_module_augmentations_since_they_are_always__2668", "'export' modifier cannot be applied to ambient modules and module augmentations since they are always visible."),
            Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_declarations: diag(2669, ts.DiagnosticCategory.Error, "Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_2669", "Augmentations for the global scope can only be directly nested in external modules or ambient module declarations."),
            Augmentations_for_the_global_scope_should_have_declare_modifier_unless_they_appear_in_already_ambient_context: diag(2670, ts.DiagnosticCategory.Error, "Augmentations_for_the_global_scope_should_have_declare_modifier_unless_they_appear_in_already_ambien_2670", "Augmentations for the global scope should have 'declare' modifier unless they appear in already ambient context."),
            Cannot_augment_module_0_because_it_resolves_to_a_non_module_entity: diag(2671, ts.DiagnosticCategory.Error, "Cannot_augment_module_0_because_it_resolves_to_a_non_module_entity_2671", "Cannot augment module '{0}' because it resolves to a non-module entity."),
            Cannot_assign_a_0_constructor_type_to_a_1_constructor_type: diag(2672, ts.DiagnosticCategory.Error, "Cannot_assign_a_0_constructor_type_to_a_1_constructor_type_2672", "Cannot assign a '{0}' constructor type to a '{1}' constructor type."),
            Constructor_of_class_0_is_private_and_only_accessible_within_the_class_declaration: diag(2673, ts.DiagnosticCategory.Error, "Constructor_of_class_0_is_private_and_only_accessible_within_the_class_declaration_2673", "Constructor of class '{0}' is private and only accessible within the class declaration."),
            Constructor_of_class_0_is_protected_and_only_accessible_within_the_class_declaration: diag(2674, ts.DiagnosticCategory.Error, "Constructor_of_class_0_is_protected_and_only_accessible_within_the_class_declaration_2674", "Constructor of class '{0}' is protected and only accessible within the class declaration."),
            Cannot_extend_a_class_0_Class_constructor_is_marked_as_private: diag(2675, ts.DiagnosticCategory.Error, "Cannot_extend_a_class_0_Class_constructor_is_marked_as_private_2675", "Cannot extend a class '{0}'. Class constructor is marked as private."),
            Accessors_must_both_be_abstract_or_non_abstract: diag(2676, ts.DiagnosticCategory.Error, "Accessors_must_both_be_abstract_or_non_abstract_2676", "Accessors must both be abstract or non-abstract."),
            A_type_predicate_s_type_must_be_assignable_to_its_parameter_s_type: diag(2677, ts.DiagnosticCategory.Error, "A_type_predicate_s_type_must_be_assignable_to_its_parameter_s_type_2677", "A type predicate's type must be assignable to its parameter's type."),
            Type_0_is_not_comparable_to_type_1: diag(2678, ts.DiagnosticCategory.Error, "Type_0_is_not_comparable_to_type_1_2678", "Type '{0}' is not comparable to type '{1}'."),
            A_function_that_is_called_with_the_new_keyword_cannot_have_a_this_type_that_is_void: diag(2679, ts.DiagnosticCategory.Error, "A_function_that_is_called_with_the_new_keyword_cannot_have_a_this_type_that_is_void_2679", "A function that is called with the 'new' keyword cannot have a 'this' type that is 'void'."),
            A_0_parameter_must_be_the_first_parameter: diag(2680, ts.DiagnosticCategory.Error, "A_0_parameter_must_be_the_first_parameter_2680", "A '{0}' parameter must be the first parameter."),
            A_constructor_cannot_have_a_this_parameter: diag(2681, ts.DiagnosticCategory.Error, "A_constructor_cannot_have_a_this_parameter_2681", "A constructor cannot have a 'this' parameter."),
            get_and_set_accessor_must_have_the_same_this_type: diag(2682, ts.DiagnosticCategory.Error, "get_and_set_accessor_must_have_the_same_this_type_2682", "'get' and 'set' accessor must have the same 'this' type."),
            this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation: diag(2683, ts.DiagnosticCategory.Error, "this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_2683", "'this' implicitly has type 'any' because it does not have a type annotation."),
            The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1: diag(2684, ts.DiagnosticCategory.Error, "The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1_2684", "The 'this' context of type '{0}' is not assignable to method's 'this' of type '{1}'."),
            The_this_types_of_each_signature_are_incompatible: diag(2685, ts.DiagnosticCategory.Error, "The_this_types_of_each_signature_are_incompatible_2685", "The 'this' types of each signature are incompatible."),
            _0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead: diag(2686, ts.DiagnosticCategory.Error, "_0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead_2686", "'{0}' refers to a UMD global, but the current file is a module. Consider adding an import instead."),
            All_declarations_of_0_must_have_identical_modifiers: diag(2687, ts.DiagnosticCategory.Error, "All_declarations_of_0_must_have_identical_modifiers_2687", "All declarations of '{0}' must have identical modifiers."),
            Cannot_find_type_definition_file_for_0: diag(2688, ts.DiagnosticCategory.Error, "Cannot_find_type_definition_file_for_0_2688", "Cannot find type definition file for '{0}'."),
            Cannot_extend_an_interface_0_Did_you_mean_implements: diag(2689, ts.DiagnosticCategory.Error, "Cannot_extend_an_interface_0_Did_you_mean_implements_2689", "Cannot extend an interface '{0}'. Did you mean 'implements'?"),
            _0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Did_you_mean_to_use_1_in_0: diag(2690, ts.DiagnosticCategory.Error, "_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Did_you_mean_to_use_1_in_0_2690", "'{0}' only refers to a type, but is being used as a value here. Did you mean to use '{1} in {0}'?"),
            An_import_path_cannot_end_with_a_0_extension_Consider_importing_1_instead: diag(2691, ts.DiagnosticCategory.Error, "An_import_path_cannot_end_with_a_0_extension_Consider_importing_1_instead_2691", "An import path cannot end with a '{0}' extension. Consider importing '{1}' instead."),
            _0_is_a_primitive_but_1_is_a_wrapper_object_Prefer_using_0_when_possible: diag(2692, ts.DiagnosticCategory.Error, "_0_is_a_primitive_but_1_is_a_wrapper_object_Prefer_using_0_when_possible_2692", "'{0}' is a primitive, but '{1}' is a wrapper object. Prefer using '{0}' when possible."),
            _0_only_refers_to_a_type_but_is_being_used_as_a_value_here: diag(2693, ts.DiagnosticCategory.Error, "_0_only_refers_to_a_type_but_is_being_used_as_a_value_here_2693", "'{0}' only refers to a type, but is being used as a value here."),
            Namespace_0_has_no_exported_member_1: diag(2694, ts.DiagnosticCategory.Error, "Namespace_0_has_no_exported_member_1_2694", "Namespace '{0}' has no exported member '{1}'."),
            Left_side_of_comma_operator_is_unused_and_has_no_side_effects: diag(2695, ts.DiagnosticCategory.Error, "Left_side_of_comma_operator_is_unused_and_has_no_side_effects_2695", "Left side of comma operator is unused and has no side effects.", /*reportsUnnecessary*/ true),
            The_Object_type_is_assignable_to_very_few_other_types_Did_you_mean_to_use_the_any_type_instead: diag(2696, ts.DiagnosticCategory.Error, "The_Object_type_is_assignable_to_very_few_other_types_Did_you_mean_to_use_the_any_type_instead_2696", "The 'Object' type is assignable to very few other types. Did you mean to use the 'any' type instead?"),
            An_async_function_or_method_must_return_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option: diag(2697, ts.DiagnosticCategory.Error, "An_async_function_or_method_must_return_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_in_2697", "An async function or method must return a 'Promise'. Make sure you have a declaration for 'Promise' or include 'ES2015' in your '--lib' option."),
            Spread_types_may_only_be_created_from_object_types: diag(2698, ts.DiagnosticCategory.Error, "Spread_types_may_only_be_created_from_object_types_2698", "Spread types may only be created from object types."),
            Static_property_0_conflicts_with_built_in_property_Function_0_of_constructor_function_1: diag(2699, ts.DiagnosticCategory.Error, "Static_property_0_conflicts_with_built_in_property_Function_0_of_constructor_function_1_2699", "Static property '{0}' conflicts with built-in property 'Function.{0}' of constructor function '{1}'."),
            Rest_types_may_only_be_created_from_object_types: diag(2700, ts.DiagnosticCategory.Error, "Rest_types_may_only_be_created_from_object_types_2700", "Rest types may only be created from object types."),
            The_target_of_an_object_rest_assignment_must_be_a_variable_or_a_property_access: diag(2701, ts.DiagnosticCategory.Error, "The_target_of_an_object_rest_assignment_must_be_a_variable_or_a_property_access_2701", "The target of an object rest assignment must be a variable or a property access."),
            _0_only_refers_to_a_type_but_is_being_used_as_a_namespace_here: diag(2702, ts.DiagnosticCategory.Error, "_0_only_refers_to_a_type_but_is_being_used_as_a_namespace_here_2702", "'{0}' only refers to a type, but is being used as a namespace here."),
            The_operand_of_a_delete_operator_must_be_a_property_reference: diag(2703, ts.DiagnosticCategory.Error, "The_operand_of_a_delete_operator_must_be_a_property_reference_2703", "The operand of a 'delete' operator must be a property reference."),
            The_operand_of_a_delete_operator_cannot_be_a_read_only_property: diag(2704, ts.DiagnosticCategory.Error, "The_operand_of_a_delete_operator_cannot_be_a_read_only_property_2704", "The operand of a 'delete' operator cannot be a read-only property."),
            An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option: diag(2705, ts.DiagnosticCategory.Error, "An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_de_2705", "An async function or method in ES5/ES3 requires the 'Promise' constructor.  Make sure you have a declaration for the 'Promise' constructor or include 'ES2015' in your '--lib' option."),
            Required_type_parameters_may_not_follow_optional_type_parameters: diag(2706, ts.DiagnosticCategory.Error, "Required_type_parameters_may_not_follow_optional_type_parameters_2706", "Required type parameters may not follow optional type parameters."),
            Generic_type_0_requires_between_1_and_2_type_arguments: diag(2707, ts.DiagnosticCategory.Error, "Generic_type_0_requires_between_1_and_2_type_arguments_2707", "Generic type '{0}' requires between {1} and {2} type arguments."),
            Cannot_use_namespace_0_as_a_value: diag(2708, ts.DiagnosticCategory.Error, "Cannot_use_namespace_0_as_a_value_2708", "Cannot use namespace '{0}' as a value."),
            Cannot_use_namespace_0_as_a_type: diag(2709, ts.DiagnosticCategory.Error, "Cannot_use_namespace_0_as_a_type_2709", "Cannot use namespace '{0}' as a type."),
            _0_are_specified_twice_The_attribute_named_0_will_be_overwritten: diag(2710, ts.DiagnosticCategory.Error, "_0_are_specified_twice_The_attribute_named_0_will_be_overwritten_2710", "'{0}' are specified twice. The attribute named '{0}' will be overwritten."),
            A_dynamic_import_call_returns_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option: diag(2711, ts.DiagnosticCategory.Error, "A_dynamic_import_call_returns_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES20_2711", "A dynamic import call returns a 'Promise'. Make sure you have a declaration for 'Promise' or include 'ES2015' in your '--lib' option."),
            A_dynamic_import_call_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option: diag(2712, ts.DiagnosticCategory.Error, "A_dynamic_import_call_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declarat_2712", "A dynamic import call in ES5/ES3 requires the 'Promise' constructor.  Make sure you have a declaration for the 'Promise' constructor or include 'ES2015' in your '--lib' option."),
            Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_property_1_in_0_with_0_1: diag(2713, ts.DiagnosticCategory.Error, "Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_p_2713", "Cannot access '{0}.{1}' because '{0}' is a type, but not a namespace. Did you mean to retrieve the type of the property '{1}' in '{0}' with '{0}[\"{1}\"]'?"),
            The_expression_of_an_export_assignment_must_be_an_identifier_or_qualified_name_in_an_ambient_context: diag(2714, ts.DiagnosticCategory.Error, "The_expression_of_an_export_assignment_must_be_an_identifier_or_qualified_name_in_an_ambient_context_2714", "The expression of an export assignment must be an identifier or qualified name in an ambient context."),
            Abstract_property_0_in_class_1_cannot_be_accessed_in_the_constructor: diag(2715, ts.DiagnosticCategory.Error, "Abstract_property_0_in_class_1_cannot_be_accessed_in_the_constructor_2715", "Abstract property '{0}' in class '{1}' cannot be accessed in the constructor."),
            Type_parameter_0_has_a_circular_default: diag(2716, ts.DiagnosticCategory.Error, "Type_parameter_0_has_a_circular_default_2716", "Type parameter '{0}' has a circular default."),
            Subsequent_property_declarations_must_have_the_same_type_Property_0_must_be_of_type_1_but_here_has_type_2: diag(2717, ts.DiagnosticCategory.Error, "Subsequent_property_declarations_must_have_the_same_type_Property_0_must_be_of_type_1_but_here_has_t_2717", "Subsequent property declarations must have the same type.  Property '{0}' must be of type '{1}', but here has type '{2}'."),
            Duplicate_property_0: diag(2718, ts.DiagnosticCategory.Error, "Duplicate_property_0_2718", "Duplicate property '{0}'."),
            Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated: diag(2719, ts.DiagnosticCategory.Error, "Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated_2719", "Type '{0}' is not assignable to type '{1}'. Two different types with this name exist, but they are unrelated."),
            Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass: diag(2720, ts.DiagnosticCategory.Error, "Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclas_2720", "Class '{0}' incorrectly implements class '{1}'. Did you mean to extend '{1}' and inherit its members as a subclass?"),
            Cannot_invoke_an_object_which_is_possibly_null: diag(2721, ts.DiagnosticCategory.Error, "Cannot_invoke_an_object_which_is_possibly_null_2721", "Cannot invoke an object which is possibly 'null'."),
            Cannot_invoke_an_object_which_is_possibly_undefined: diag(2722, ts.DiagnosticCategory.Error, "Cannot_invoke_an_object_which_is_possibly_undefined_2722", "Cannot invoke an object which is possibly 'undefined'."),
            Cannot_invoke_an_object_which_is_possibly_null_or_undefined: diag(2723, ts.DiagnosticCategory.Error, "Cannot_invoke_an_object_which_is_possibly_null_or_undefined_2723", "Cannot invoke an object which is possibly 'null' or 'undefined'."),
            _0_has_no_exported_member_named_1_Did_you_mean_2: diag(2724, ts.DiagnosticCategory.Error, "_0_has_no_exported_member_named_1_Did_you_mean_2_2724", "'{0}' has no exported member named '{1}'. Did you mean '{2}'?"),
            Class_name_cannot_be_Object_when_targeting_ES5_with_module_0: diag(2725, ts.DiagnosticCategory.Error, "Class_name_cannot_be_Object_when_targeting_ES5_with_module_0_2725", "Class name cannot be 'Object' when targeting ES5 with module {0}."),
            Cannot_find_lib_definition_for_0: diag(2726, ts.DiagnosticCategory.Error, "Cannot_find_lib_definition_for_0_2726", "Cannot find lib definition for '{0}'."),
            Cannot_find_lib_definition_for_0_Did_you_mean_1: diag(2727, ts.DiagnosticCategory.Error, "Cannot_find_lib_definition_for_0_Did_you_mean_1_2727", "Cannot find lib definition for '{0}'. Did you mean '{1}'?"),
            _0_is_declared_here: diag(2728, ts.DiagnosticCategory.Message, "_0_is_declared_here_2728", "'{0}' is declared here."),
            Property_0_is_used_before_its_initialization: diag(2729, ts.DiagnosticCategory.Error, "Property_0_is_used_before_its_initialization_2729", "Property '{0}' is used before its initialization."),
            An_arrow_function_cannot_have_a_this_parameter: diag(2730, ts.DiagnosticCategory.Error, "An_arrow_function_cannot_have_a_this_parameter_2730", "An arrow function cannot have a 'this' parameter."),
            Implicit_conversion_of_a_symbol_to_a_string_will_fail_at_runtime_Consider_wrapping_this_expression_in_String: diag(2731, ts.DiagnosticCategory.Error, "Implicit_conversion_of_a_symbol_to_a_string_will_fail_at_runtime_Consider_wrapping_this_expression_i_2731", "Implicit conversion of a 'symbol' to a 'string' will fail at runtime. Consider wrapping this expression in 'String(...)'."),
            Cannot_find_module_0_Consider_using_resolveJsonModule_to_import_module_with_json_extension: diag(2732, ts.DiagnosticCategory.Error, "Cannot_find_module_0_Consider_using_resolveJsonModule_to_import_module_with_json_extension_2732", "Cannot find module '{0}'. Consider using '--resolveJsonModule' to import module with '.json' extension."),
            Property_0_was_also_declared_here: diag(2733, ts.DiagnosticCategory.Error, "Property_0_was_also_declared_here_2733", "Property '{0}' was also declared here."),
            Are_you_missing_a_semicolon: diag(2734, ts.DiagnosticCategory.Error, "Are_you_missing_a_semicolon_2734", "Are you missing a semicolon?"),
            Did_you_mean_for_0_to_be_constrained_to_type_new_args_Colon_any_1: diag(2735, ts.DiagnosticCategory.Error, "Did_you_mean_for_0_to_be_constrained_to_type_new_args_Colon_any_1_2735", "Did you mean for '{0}' to be constrained to type 'new (...args: any[]) => {1}'?"),
            Operator_0_cannot_be_applied_to_type_1: diag(2736, ts.DiagnosticCategory.Error, "Operator_0_cannot_be_applied_to_type_1_2736", "Operator '{0}' cannot be applied to type '{1}'."),
            BigInt_literals_are_not_available_when_targeting_lower_than_ES2020: diag(2737, ts.DiagnosticCategory.Error, "BigInt_literals_are_not_available_when_targeting_lower_than_ES2020_2737", "BigInt literals are not available when targeting lower than ES2020."),
            An_outer_value_of_this_is_shadowed_by_this_container: diag(2738, ts.DiagnosticCategory.Message, "An_outer_value_of_this_is_shadowed_by_this_container_2738", "An outer value of 'this' is shadowed by this container."),
            Type_0_is_missing_the_following_properties_from_type_1_Colon_2: diag(2739, ts.DiagnosticCategory.Error, "Type_0_is_missing_the_following_properties_from_type_1_Colon_2_2739", "Type '{0}' is missing the following properties from type '{1}': {2}"),
            Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more: diag(2740, ts.DiagnosticCategory.Error, "Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more_2740", "Type '{0}' is missing the following properties from type '{1}': {2}, and {3} more."),
            Property_0_is_missing_in_type_1_but_required_in_type_2: diag(2741, ts.DiagnosticCategory.Error, "Property_0_is_missing_in_type_1_but_required_in_type_2_2741", "Property '{0}' is missing in type '{1}' but required in type '{2}'."),
            The_inferred_type_of_0_cannot_be_named_without_a_reference_to_1_This_is_likely_not_portable_A_type_annotation_is_necessary: diag(2742, ts.DiagnosticCategory.Error, "The_inferred_type_of_0_cannot_be_named_without_a_reference_to_1_This_is_likely_not_portable_A_type_a_2742", "The inferred type of '{0}' cannot be named without a reference to '{1}'. This is likely not portable. A type annotation is necessary."),
            No_overload_expects_0_type_arguments_but_overloads_do_exist_that_expect_either_1_or_2_type_arguments: diag(2743, ts.DiagnosticCategory.Error, "No_overload_expects_0_type_arguments_but_overloads_do_exist_that_expect_either_1_or_2_type_arguments_2743", "No overload expects {0} type arguments, but overloads do exist that expect either {1} or {2} type arguments."),
            Type_parameter_defaults_can_only_reference_previously_declared_type_parameters: diag(2744, ts.DiagnosticCategory.Error, "Type_parameter_defaults_can_only_reference_previously_declared_type_parameters_2744", "Type parameter defaults can only reference previously declared type parameters."),
            This_JSX_tag_s_0_prop_expects_type_1_which_requires_multiple_children_but_only_a_single_child_was_provided: diag(2745, ts.DiagnosticCategory.Error, "This_JSX_tag_s_0_prop_expects_type_1_which_requires_multiple_children_but_only_a_single_child_was_pr_2745", "This JSX tag's '{0}' prop expects type '{1}' which requires multiple children, but only a single child was provided."),
            This_JSX_tag_s_0_prop_expects_a_single_child_of_type_1_but_multiple_children_were_provided: diag(2746, ts.DiagnosticCategory.Error, "This_JSX_tag_s_0_prop_expects_a_single_child_of_type_1_but_multiple_children_were_provided_2746", "This JSX tag's '{0}' prop expects a single child of type '{1}', but multiple children were provided."),
            _0_components_don_t_accept_text_as_child_elements_Text_in_JSX_has_the_type_string_but_the_expected_type_of_1_is_2: diag(2747, ts.DiagnosticCategory.Error, "_0_components_don_t_accept_text_as_child_elements_Text_in_JSX_has_the_type_string_but_the_expected_t_2747", "'{0}' components don't accept text as child elements. Text in JSX has the type 'string', but the expected type of '{1}' is '{2}'."),
            Cannot_access_ambient_const_enums_when_the_isolatedModules_flag_is_provided: diag(2748, ts.DiagnosticCategory.Error, "Cannot_access_ambient_const_enums_when_the_isolatedModules_flag_is_provided_2748", "Cannot access ambient const enums when the '--isolatedModules' flag is provided."),
            _0_refers_to_a_value_but_is_being_used_as_a_type_here_Did_you_mean_typeof_0: diag(2749, ts.DiagnosticCategory.Error, "_0_refers_to_a_value_but_is_being_used_as_a_type_here_Did_you_mean_typeof_0_2749", "'{0}' refers to a value, but is being used as a type here. Did you mean 'typeof {0}'?"),
            The_implementation_signature_is_declared_here: diag(2750, ts.DiagnosticCategory.Error, "The_implementation_signature_is_declared_here_2750", "The implementation signature is declared here."),
            Circularity_originates_in_type_at_this_location: diag(2751, ts.DiagnosticCategory.Error, "Circularity_originates_in_type_at_this_location_2751", "Circularity originates in type at this location."),
            The_first_export_default_is_here: diag(2752, ts.DiagnosticCategory.Error, "The_first_export_default_is_here_2752", "The first export default is here."),
            Another_export_default_is_here: diag(2753, ts.DiagnosticCategory.Error, "Another_export_default_is_here_2753", "Another export default is here."),
            super_may_not_use_type_arguments: diag(2754, ts.DiagnosticCategory.Error, "super_may_not_use_type_arguments_2754", "'super' may not use type arguments."),
            No_constituent_of_type_0_is_callable: diag(2755, ts.DiagnosticCategory.Error, "No_constituent_of_type_0_is_callable_2755", "No constituent of type '{0}' is callable."),
            Not_all_constituents_of_type_0_are_callable: diag(2756, ts.DiagnosticCategory.Error, "Not_all_constituents_of_type_0_are_callable_2756", "Not all constituents of type '{0}' are callable."),
            Type_0_has_no_call_signatures: diag(2757, ts.DiagnosticCategory.Error, "Type_0_has_no_call_signatures_2757", "Type '{0}' has no call signatures."),
            Each_member_of_the_union_type_0_has_signatures_but_none_of_those_signatures_are_compatible_with_each_other: diag(2758, ts.DiagnosticCategory.Error, "Each_member_of_the_union_type_0_has_signatures_but_none_of_those_signatures_are_compatible_with_each_2758", "Each member of the union type '{0}' has signatures, but none of those signatures are compatible with each other."),
            No_constituent_of_type_0_is_constructable: diag(2759, ts.DiagnosticCategory.Error, "No_constituent_of_type_0_is_constructable_2759", "No constituent of type '{0}' is constructable."),
            Not_all_constituents_of_type_0_are_constructable: diag(2760, ts.DiagnosticCategory.Error, "Not_all_constituents_of_type_0_are_constructable_2760", "Not all constituents of type '{0}' are constructable."),
            Type_0_has_no_construct_signatures: diag(2761, ts.DiagnosticCategory.Error, "Type_0_has_no_construct_signatures_2761", "Type '{0}' has no construct signatures."),
            Each_member_of_the_union_type_0_has_construct_signatures_but_none_of_those_signatures_are_compatible_with_each_other: diag(2762, ts.DiagnosticCategory.Error, "Each_member_of_the_union_type_0_has_construct_signatures_but_none_of_those_signatures_are_compatible_2762", "Each member of the union type '{0}' has construct signatures, but none of those signatures are compatible with each other."),
            Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_for_of_will_always_send_0: diag(2763, ts.DiagnosticCategory.Error, "Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_for_of_will_always_s_2763", "Cannot iterate value because the 'next' method of its iterator expects type '{1}', but for-of will always send '{0}'."),
            Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_spread_will_always_send_0: diag(2764, ts.DiagnosticCategory.Error, "Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_spread_will_al_2764", "Cannot iterate value because the 'next' method of its iterator expects type '{1}', but array spread will always send '{0}'."),
            Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_destructuring_will_always_send_0: diag(2765, ts.DiagnosticCategory.Error, "Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_destructuring__2765", "Cannot iterate value because the 'next' method of its iterator expects type '{1}', but array destructuring will always send '{0}'."),
            Cannot_delegate_iteration_to_value_because_the_next_method_of_its_iterator_expects_type_1_but_the_containing_generator_will_always_send_0: diag(2766, ts.DiagnosticCategory.Error, "Cannot_delegate_iteration_to_value_because_the_next_method_of_its_iterator_expects_type_1_but_the_co_2766", "Cannot delegate iteration to value because the 'next' method of its iterator expects type '{1}', but the containing generator will always send '{0}'."),
            The_0_property_of_an_iterator_must_be_a_method: diag(2767, ts.DiagnosticCategory.Error, "The_0_property_of_an_iterator_must_be_a_method_2767", "The '{0}' property of an iterator must be a method."),
            The_0_property_of_an_async_iterator_must_be_a_method: diag(2768, ts.DiagnosticCategory.Error, "The_0_property_of_an_async_iterator_must_be_a_method_2768", "The '{0}' property of an async iterator must be a method."),
            No_overload_matches_this_call: diag(2769, ts.DiagnosticCategory.Error, "No_overload_matches_this_call_2769", "No overload matches this call."),
            The_last_overload_gave_the_following_error: diag(2770, ts.DiagnosticCategory.Error, "The_last_overload_gave_the_following_error_2770", "The last overload gave the following error."),
            The_last_overload_is_declared_here: diag(2771, ts.DiagnosticCategory.Error, "The_last_overload_is_declared_here_2771", "The last overload is declared here."),
            Overload_0_of_1_2_gave_the_following_error: diag(2772, ts.DiagnosticCategory.Error, "Overload_0_of_1_2_gave_the_following_error_2772", "Overload {0} of {1}, '{2}', gave the following error."),
            Did_you_forget_to_use_await: diag(2773, ts.DiagnosticCategory.Error, "Did_you_forget_to_use_await_2773", "Did you forget to use 'await'?"),
            This_condition_will_always_return_true_since_this_function_is_always_defined_Did_you_mean_to_call_it_instead: diag(2774, ts.DiagnosticCategory.Error, "This_condition_will_always_return_true_since_this_function_is_always_defined_Did_you_mean_to_call_it_2774", "This condition will always return true since this function is always defined. Did you mean to call it instead?"),
            Assertions_require_every_name_in_the_call_target_to_be_declared_with_an_explicit_type_annotation: diag(2775, ts.DiagnosticCategory.Error, "Assertions_require_every_name_in_the_call_target_to_be_declared_with_an_explicit_type_annotation_2775", "Assertions require every name in the call target to be declared with an explicit type annotation."),
            Assertions_require_the_call_target_to_be_an_identifier_or_qualified_name: diag(2776, ts.DiagnosticCategory.Error, "Assertions_require_the_call_target_to_be_an_identifier_or_qualified_name_2776", "Assertions require the call target to be an identifier or qualified name."),
            The_operand_of_an_increment_or_decrement_operator_may_not_be_an_optional_property_access: diag(2777, ts.DiagnosticCategory.Error, "The_operand_of_an_increment_or_decrement_operator_may_not_be_an_optional_property_access_2777", "The operand of an increment or decrement operator may not be an optional property access."),
            The_target_of_an_object_rest_assignment_may_not_be_an_optional_property_access: diag(2778, ts.DiagnosticCategory.Error, "The_target_of_an_object_rest_assignment_may_not_be_an_optional_property_access_2778", "The target of an object rest assignment may not be an optional property access."),
            The_left_hand_side_of_an_assignment_expression_may_not_be_an_optional_property_access: diag(2779, ts.DiagnosticCategory.Error, "The_left_hand_side_of_an_assignment_expression_may_not_be_an_optional_property_access_2779", "The left-hand side of an assignment expression may not be an optional property access."),
            The_left_hand_side_of_a_for_in_statement_may_not_be_an_optional_property_access: diag(2780, ts.DiagnosticCategory.Error, "The_left_hand_side_of_a_for_in_statement_may_not_be_an_optional_property_access_2780", "The left-hand side of a 'for...in' statement may not be an optional property access."),
            The_left_hand_side_of_a_for_of_statement_may_not_be_an_optional_property_access: diag(2781, ts.DiagnosticCategory.Error, "The_left_hand_side_of_a_for_of_statement_may_not_be_an_optional_property_access_2781", "The left-hand side of a 'for...of' statement may not be an optional property access."),
            _0_needs_an_explicit_type_annotation: diag(2782, ts.DiagnosticCategory.Message, "_0_needs_an_explicit_type_annotation_2782", "'{0}' needs an explicit type annotation."),
            _0_is_specified_more_than_once_so_this_usage_will_be_overwritten: diag(2783, ts.DiagnosticCategory.Error, "_0_is_specified_more_than_once_so_this_usage_will_be_overwritten_2783", "'{0}' is specified more than once, so this usage will be overwritten."),
            get_and_set_accessors_cannot_declare_this_parameters: diag(2784, ts.DiagnosticCategory.Error, "get_and_set_accessors_cannot_declare_this_parameters_2784", "'get' and 'set' accessors cannot declare 'this' parameters."),
            This_spread_always_overwrites_this_property: diag(2785, ts.DiagnosticCategory.Error, "This_spread_always_overwrites_this_property_2785", "This spread always overwrites this property."),
            _0_cannot_be_used_as_a_JSX_component: diag(2786, ts.DiagnosticCategory.Error, "_0_cannot_be_used_as_a_JSX_component_2786", "'{0}' cannot be used as a JSX component."),
            Its_return_type_0_is_not_a_valid_JSX_element: diag(2787, ts.DiagnosticCategory.Error, "Its_return_type_0_is_not_a_valid_JSX_element_2787", "Its return type '{0}' is not a valid JSX element."),
            Its_instance_type_0_is_not_a_valid_JSX_element: diag(2788, ts.DiagnosticCategory.Error, "Its_instance_type_0_is_not_a_valid_JSX_element_2788", "Its instance type '{0}' is not a valid JSX element."),
            Its_element_type_0_is_not_a_valid_JSX_element: diag(2789, ts.DiagnosticCategory.Error, "Its_element_type_0_is_not_a_valid_JSX_element_2789", "Its element type '{0}' is not a valid JSX element."),
            The_operand_of_a_delete_operator_must_be_optional: diag(2790, ts.DiagnosticCategory.Error, "The_operand_of_a_delete_operator_must_be_optional_2790", "The operand of a 'delete' operator must be optional."),
            Exponentiation_cannot_be_performed_on_bigint_values_unless_the_target_option_is_set_to_es2016_or_later: diag(2791, ts.DiagnosticCategory.Error, "Exponentiation_cannot_be_performed_on_bigint_values_unless_the_target_option_is_set_to_es2016_or_lat_2791", "Exponentiation cannot be performed on 'bigint' values unless the 'target' option is set to 'es2016' or later."),
            Cannot_find_module_0_Did_you_mean_to_set_the_moduleResolution_option_to_node_or_to_add_aliases_to_the_paths_option: diag(2792, ts.DiagnosticCategory.Error, "Cannot_find_module_0_Did_you_mean_to_set_the_moduleResolution_option_to_node_or_to_add_aliases_to_th_2792", "Cannot find module '{0}'. Did you mean to set the 'moduleResolution' option to 'node', or to add aliases to the 'paths' option?"),
            The_call_would_have_succeeded_against_this_implementation_but_implementation_signatures_of_overloads_are_not_externally_visible: diag(2793, ts.DiagnosticCategory.Error, "The_call_would_have_succeeded_against_this_implementation_but_implementation_signatures_of_overloads_2793", "The call would have succeeded against this implementation, but implementation signatures of overloads are not externally visible."),
            Expected_0_arguments_but_got_1_Did_you_forget_to_include_void_in_your_type_argument_to_Promise: diag(2794, ts.DiagnosticCategory.Error, "Expected_0_arguments_but_got_1_Did_you_forget_to_include_void_in_your_type_argument_to_Promise_2794", "Expected {0} arguments, but got {1}. Did you forget to include 'void' in your type argument to 'Promise'?"),
            The_intrinsic_keyword_can_only_be_used_to_declare_compiler_provided_intrinsic_types: diag(2795, ts.DiagnosticCategory.Error, "The_intrinsic_keyword_can_only_be_used_to_declare_compiler_provided_intrinsic_types_2795", "The 'intrinsic' keyword can only be used to declare compiler provided intrinsic types."),
            It_is_likely_that_you_are_missing_a_comma_to_separate_these_two_template_expressions_They_form_a_tagged_template_expression_which_cannot_be_invoked: diag(2796, ts.DiagnosticCategory.Error, "It_is_likely_that_you_are_missing_a_comma_to_separate_these_two_template_expressions_They_form_a_tag_2796", "It is likely that you are missing a comma to separate these two template expressions. They form a tagged template expression which cannot be invoked."),
            A_mixin_class_that_extends_from_a_type_variable_containing_an_abstract_construct_signature_must_also_be_declared_abstract: diag(2797, ts.DiagnosticCategory.Error, "A_mixin_class_that_extends_from_a_type_variable_containing_an_abstract_construct_signature_must_also_2797", "A mixin class that extends from a type variable containing an abstract construct signature must also be declared 'abstract'."),
            The_declaration_was_marked_as_deprecated_here: diag(2798, ts.DiagnosticCategory.Error, "The_declaration_was_marked_as_deprecated_here_2798", "The declaration was marked as deprecated here."),
            Type_produces_a_tuple_type_that_is_too_large_to_represent: diag(2799, ts.DiagnosticCategory.Error, "Type_produces_a_tuple_type_that_is_too_large_to_represent_2799", "Type produces a tuple type that is too large to represent."),
            Expression_produces_a_tuple_type_that_is_too_large_to_represent: diag(2800, ts.DiagnosticCategory.Error, "Expression_produces_a_tuple_type_that_is_too_large_to_represent_2800", "Expression produces a tuple type that is too large to represent."),
            This_condition_will_always_return_true_since_this_0_is_always_defined: diag(2801, ts.DiagnosticCategory.Error, "This_condition_will_always_return_true_since_this_0_is_always_defined_2801", "This condition will always return true since this '{0}' is always defined."),
            Type_0_can_only_be_iterated_through_when_using_the_downlevelIteration_flag_or_with_a_target_of_es2015_or_higher: diag(2802, ts.DiagnosticCategory.Error, "Type_0_can_only_be_iterated_through_when_using_the_downlevelIteration_flag_or_with_a_target_of_es201_2802", "Type '{0}' can only be iterated through when using the '--downlevelIteration' flag or with a '--target' of 'es2015' or higher."),
            Cannot_assign_to_private_method_0_Private_methods_are_not_writable: diag(2803, ts.DiagnosticCategory.Error, "Cannot_assign_to_private_method_0_Private_methods_are_not_writable_2803", "Cannot assign to private method '{0}'. Private methods are not writable."),
            Duplicate_identifier_0_Static_and_instance_elements_cannot_share_the_same_private_name: diag(2804, ts.DiagnosticCategory.Error, "Duplicate_identifier_0_Static_and_instance_elements_cannot_share_the_same_private_name_2804", "Duplicate identifier '{0}'. Static and instance elements cannot share the same private name."),
            Static_fields_with_private_names_can_t_have_initializers_when_the_useDefineForClassFields_flag_is_not_specified_with_a_target_of_esnext_Consider_adding_the_useDefineForClassFields_flag: diag(2805, ts.DiagnosticCategory.Error, "Static_fields_with_private_names_can_t_have_initializers_when_the_useDefineForClassFields_flag_is_no_2805", "Static fields with private names can't have initializers when the '--useDefineForClassFields' flag is not specified with a '--target' of 'esnext'. Consider adding the '--useDefineForClassFields' flag."),
            Private_accessor_was_defined_without_a_getter: diag(2806, ts.DiagnosticCategory.Error, "Private_accessor_was_defined_without_a_getter_2806", "Private accessor was defined without a getter."),
            This_syntax_requires_an_imported_helper_named_1_with_2_parameters_which_is_not_compatible_with_the_one_in_0_Consider_upgrading_your_version_of_0: diag(2807, ts.DiagnosticCategory.Error, "This_syntax_requires_an_imported_helper_named_1_with_2_parameters_which_is_not_compatible_with_the_o_2807", "This syntax requires an imported helper named '{1}' with {2} parameters, which is not compatible with the one in '{0}'. Consider upgrading your version of '{0}'."),
            A_get_accessor_must_be_at_least_as_accessible_as_the_setter: diag(2808, ts.DiagnosticCategory.Error, "A_get_accessor_must_be_at_least_as_accessible_as_the_setter_2808", "A get accessor must be at least as accessible as the setter"),
            Declaration_or_statement_expected_This_follows_a_block_of_statements_so_if_you_intended_to_write_a_destructuring_assignment_you_might_need_to_wrap_the_the_whole_assignment_in_parentheses: diag(2809, ts.DiagnosticCategory.Error, "Declaration_or_statement_expected_This_follows_a_block_of_statements_so_if_you_intended_to_write_a_d_2809", "Declaration or statement expected. This '=' follows a block of statements, so if you intended to write a destructuring assignment, you might need to wrap the the whole assignment in parentheses."),
            Property_0_may_not_be_used_in_a_static_property_s_initializer_in_the_same_class_when_target_is_esnext_and_useDefineForClassFields_is_false: diag(2810, ts.DiagnosticCategory.Error, "Property_0_may_not_be_used_in_a_static_property_s_initializer_in_the_same_class_when_target_is_esnex_2810", "Property '{0}' may not be used in a static property's initializer in the same class when 'target' is 'esnext' and 'useDefineForClassFields' is 'false'."),
            Initializer_for_property_0: diag(2811, ts.DiagnosticCategory.Error, "Initializer_for_property_0_2811", "Initializer for property '{0}'"),
            Property_0_does_not_exist_on_type_1_Try_changing_the_lib_compiler_option_to_include_dom: diag(2812, ts.DiagnosticCategory.Error, "Property_0_does_not_exist_on_type_1_Try_changing_the_lib_compiler_option_to_include_dom_2812", "Property '{0}' does not exist on type '{1}'. Try changing the 'lib' compiler option to include 'dom'."),
            Class_declaration_cannot_implement_overload_list_for_0: diag(2813, ts.DiagnosticCategory.Error, "Class_declaration_cannot_implement_overload_list_for_0_2813", "Class declaration cannot implement overload list for '{0}'."),
            Function_with_bodies_can_only_merge_with_classes_that_are_ambient: diag(2814, ts.DiagnosticCategory.Error, "Function_with_bodies_can_only_merge_with_classes_that_are_ambient_2814", "Function with bodies can only merge with classes that are ambient."),
            arguments_cannot_be_referenced_in_property_initializers: diag(2815, ts.DiagnosticCategory.Error, "arguments_cannot_be_referenced_in_property_initializers_2815", "'arguments' cannot be referenced in property initializers."),
            Cannot_use_this_in_a_static_property_initializer_of_a_decorated_class: diag(2816, ts.DiagnosticCategory.Error, "Cannot_use_this_in_a_static_property_initializer_of_a_decorated_class_2816", "Cannot use 'this' in a static property initializer of a decorated class."),
            Property_0_has_no_initializer_and_is_not_definitely_assigned_in_a_class_static_block: diag(2817, ts.DiagnosticCategory.Error, "Property_0_has_no_initializer_and_is_not_definitely_assigned_in_a_class_static_block_2817", "Property '{0}' has no initializer and is not definitely assigned in a class static block."),
            Duplicate_identifier_0_Compiler_reserves_name_1_when_emitting_super_references_in_static_initializers: diag(2818, ts.DiagnosticCategory.Error, "Duplicate_identifier_0_Compiler_reserves_name_1_when_emitting_super_references_in_static_initializer_2818", "Duplicate identifier '{0}'. Compiler reserves name '{1}' when emitting 'super' references in static initializers."),
            Namespace_name_cannot_be_0: diag(2819, ts.DiagnosticCategory.Error, "Namespace_name_cannot_be_0_2819", "Namespace name cannot be '{0}'."),
            Import_declaration_0_is_using_private_name_1: diag(4000, ts.DiagnosticCategory.Error, "Import_declaration_0_is_using_private_name_1_4000", "Import declaration '{0}' is using private name '{1}'."),
            Type_parameter_0_of_exported_class_has_or_is_using_private_name_1: diag(4002, ts.DiagnosticCategory.Error, "Type_parameter_0_of_exported_class_has_or_is_using_private_name_1_4002", "Type parameter '{0}' of exported class has or is using private name '{1}'."),
            Type_parameter_0_of_exported_interface_has_or_is_using_private_name_1: diag(4004, ts.DiagnosticCategory.Error, "Type_parameter_0_of_exported_interface_has_or_is_using_private_name_1_4004", "Type parameter '{0}' of exported interface has or is using private name '{1}'."),
            Type_parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1: diag(4006, ts.DiagnosticCategory.Error, "Type_parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1_4006", "Type parameter '{0}' of constructor signature from exported interface has or is using private name '{1}'."),
            Type_parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1: diag(4008, ts.DiagnosticCategory.Error, "Type_parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1_4008", "Type parameter '{0}' of call signature from exported interface has or is using private name '{1}'."),
            Type_parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1: diag(4010, ts.DiagnosticCategory.Error, "Type_parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1_4010", "Type parameter '{0}' of public static method from exported class has or is using private name '{1}'."),
            Type_parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1: diag(4012, ts.DiagnosticCategory.Error, "Type_parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1_4012", "Type parameter '{0}' of public method from exported class has or is using private name '{1}'."),
            Type_parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1: diag(4014, ts.DiagnosticCategory.Error, "Type_parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1_4014", "Type parameter '{0}' of method from exported interface has or is using private name '{1}'."),
            Type_parameter_0_of_exported_function_has_or_is_using_private_name_1: diag(4016, ts.DiagnosticCategory.Error, "Type_parameter_0_of_exported_function_has_or_is_using_private_name_1_4016", "Type parameter '{0}' of exported function has or is using private name '{1}'."),
            Implements_clause_of_exported_class_0_has_or_is_using_private_name_1: diag(4019, ts.DiagnosticCategory.Error, "Implements_clause_of_exported_class_0_has_or_is_using_private_name_1_4019", "Implements clause of exported class '{0}' has or is using private name '{1}'."),
            extends_clause_of_exported_class_0_has_or_is_using_private_name_1: diag(4020, ts.DiagnosticCategory.Error, "extends_clause_of_exported_class_0_has_or_is_using_private_name_1_4020", "'extends' clause of exported class '{0}' has or is using private name '{1}'."),
            extends_clause_of_exported_class_has_or_is_using_private_name_0: diag(4021, ts.DiagnosticCategory.Error, "extends_clause_of_exported_class_has_or_is_using_private_name_0_4021", "'extends' clause of exported class has or is using private name '{0}'."),
            extends_clause_of_exported_interface_0_has_or_is_using_private_name_1: diag(4022, ts.DiagnosticCategory.Error, "extends_clause_of_exported_interface_0_has_or_is_using_private_name_1_4022", "'extends' clause of exported interface '{0}' has or is using private name '{1}'."),
            Exported_variable_0_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4023, ts.DiagnosticCategory.Error, "Exported_variable_0_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4023", "Exported variable '{0}' has or is using name '{1}' from external module {2} but cannot be named."),
            Exported_variable_0_has_or_is_using_name_1_from_private_module_2: diag(4024, ts.DiagnosticCategory.Error, "Exported_variable_0_has_or_is_using_name_1_from_private_module_2_4024", "Exported variable '{0}' has or is using name '{1}' from private module '{2}'."),
            Exported_variable_0_has_or_is_using_private_name_1: diag(4025, ts.DiagnosticCategory.Error, "Exported_variable_0_has_or_is_using_private_name_1_4025", "Exported variable '{0}' has or is using private name '{1}'."),
            Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4026, ts.DiagnosticCategory.Error, "Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot__4026", "Public static property '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),
            Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4027, ts.DiagnosticCategory.Error, "Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4027", "Public static property '{0}' of exported class has or is using name '{1}' from private module '{2}'."),
            Public_static_property_0_of_exported_class_has_or_is_using_private_name_1: diag(4028, ts.DiagnosticCategory.Error, "Public_static_property_0_of_exported_class_has_or_is_using_private_name_1_4028", "Public static property '{0}' of exported class has or is using private name '{1}'."),
            Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4029, ts.DiagnosticCategory.Error, "Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_name_4029", "Public property '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),
            Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4030, ts.DiagnosticCategory.Error, "Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4030", "Public property '{0}' of exported class has or is using name '{1}' from private module '{2}'."),
            Public_property_0_of_exported_class_has_or_is_using_private_name_1: diag(4031, ts.DiagnosticCategory.Error, "Public_property_0_of_exported_class_has_or_is_using_private_name_1_4031", "Public property '{0}' of exported class has or is using private name '{1}'."),
            Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2: diag(4032, ts.DiagnosticCategory.Error, "Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2_4032", "Property '{0}' of exported interface has or is using name '{1}' from private module '{2}'."),
            Property_0_of_exported_interface_has_or_is_using_private_name_1: diag(4033, ts.DiagnosticCategory.Error, "Property_0_of_exported_interface_has_or_is_using_private_name_1_4033", "Property '{0}' of exported interface has or is using private name '{1}'."),
            Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4034, ts.DiagnosticCategory.Error, "Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_name_1_from_private_mod_4034", "Parameter type of public static setter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),
            Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_private_name_1: diag(4035, ts.DiagnosticCategory.Error, "Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_private_name_1_4035", "Parameter type of public static setter '{0}' from exported class has or is using private name '{1}'."),
            Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4036, ts.DiagnosticCategory.Error, "Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2_4036", "Parameter type of public setter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),
            Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_private_name_1: diag(4037, ts.DiagnosticCategory.Error, "Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_private_name_1_4037", "Parameter type of public setter '{0}' from exported class has or is using private name '{1}'."),
            Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4038, ts.DiagnosticCategory.Error, "Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_external_modul_4038", "Return type of public static getter '{0}' from exported class has or is using name '{1}' from external module {2} but cannot be named."),
            Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4039, ts.DiagnosticCategory.Error, "Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_4039", "Return type of public static getter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),
            Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_private_name_1: diag(4040, ts.DiagnosticCategory.Error, "Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_private_name_1_4040", "Return type of public static getter '{0}' from exported class has or is using private name '{1}'."),
            Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4041, ts.DiagnosticCategory.Error, "Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_4041", "Return type of public getter '{0}' from exported class has or is using name '{1}' from external module {2} but cannot be named."),
            Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4042, ts.DiagnosticCategory.Error, "Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2_4042", "Return type of public getter '{0}' from exported class has or is using name '{1}' from private module '{2}'."),
            Return_type_of_public_getter_0_from_exported_class_has_or_is_using_private_name_1: diag(4043, ts.DiagnosticCategory.Error, "Return_type_of_public_getter_0_from_exported_class_has_or_is_using_private_name_1_4043", "Return type of public getter '{0}' from exported class has or is using private name '{1}'."),
            Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1: diag(4044, ts.DiagnosticCategory.Error, "Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_name_0_from_private_mod_4044", "Return type of constructor signature from exported interface has or is using name '{0}' from private module '{1}'."),
            Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_0: diag(4045, ts.DiagnosticCategory.Error, "Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_0_4045", "Return type of constructor signature from exported interface has or is using private name '{0}'."),
            Return_type_of_call_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1: diag(4046, ts.DiagnosticCategory.Error, "Return_type_of_call_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1_4046", "Return type of call signature from exported interface has or is using name '{0}' from private module '{1}'."),
            Return_type_of_call_signature_from_exported_interface_has_or_is_using_private_name_0: diag(4047, ts.DiagnosticCategory.Error, "Return_type_of_call_signature_from_exported_interface_has_or_is_using_private_name_0_4047", "Return type of call signature from exported interface has or is using private name '{0}'."),
            Return_type_of_index_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1: diag(4048, ts.DiagnosticCategory.Error, "Return_type_of_index_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1_4048", "Return type of index signature from exported interface has or is using name '{0}' from private module '{1}'."),
            Return_type_of_index_signature_from_exported_interface_has_or_is_using_private_name_0: diag(4049, ts.DiagnosticCategory.Error, "Return_type_of_index_signature_from_exported_interface_has_or_is_using_private_name_0_4049", "Return type of index signature from exported interface has or is using private name '{0}'."),
            Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named: diag(4050, ts.DiagnosticCategory.Error, "Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_external_module__4050", "Return type of public static method from exported class has or is using name '{0}' from external module {1} but cannot be named."),
            Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_private_module_1: diag(4051, ts.DiagnosticCategory.Error, "Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_private_module_1_4051", "Return type of public static method from exported class has or is using name '{0}' from private module '{1}'."),
            Return_type_of_public_static_method_from_exported_class_has_or_is_using_private_name_0: diag(4052, ts.DiagnosticCategory.Error, "Return_type_of_public_static_method_from_exported_class_has_or_is_using_private_name_0_4052", "Return type of public static method from exported class has or is using private name '{0}'."),
            Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named: diag(4053, ts.DiagnosticCategory.Error, "Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_c_4053", "Return type of public method from exported class has or is using name '{0}' from external module {1} but cannot be named."),
            Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_private_module_1: diag(4054, ts.DiagnosticCategory.Error, "Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_private_module_1_4054", "Return type of public method from exported class has or is using name '{0}' from private module '{1}'."),
            Return_type_of_public_method_from_exported_class_has_or_is_using_private_name_0: diag(4055, ts.DiagnosticCategory.Error, "Return_type_of_public_method_from_exported_class_has_or_is_using_private_name_0_4055", "Return type of public method from exported class has or is using private name '{0}'."),
            Return_type_of_method_from_exported_interface_has_or_is_using_name_0_from_private_module_1: diag(4056, ts.DiagnosticCategory.Error, "Return_type_of_method_from_exported_interface_has_or_is_using_name_0_from_private_module_1_4056", "Return type of method from exported interface has or is using name '{0}' from private module '{1}'."),
            Return_type_of_method_from_exported_interface_has_or_is_using_private_name_0: diag(4057, ts.DiagnosticCategory.Error, "Return_type_of_method_from_exported_interface_has_or_is_using_private_name_0_4057", "Return type of method from exported interface has or is using private name '{0}'."),
            Return_type_of_exported_function_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named: diag(4058, ts.DiagnosticCategory.Error, "Return_type_of_exported_function_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named_4058", "Return type of exported function has or is using name '{0}' from external module {1} but cannot be named."),
            Return_type_of_exported_function_has_or_is_using_name_0_from_private_module_1: diag(4059, ts.DiagnosticCategory.Error, "Return_type_of_exported_function_has_or_is_using_name_0_from_private_module_1_4059", "Return type of exported function has or is using name '{0}' from private module '{1}'."),
            Return_type_of_exported_function_has_or_is_using_private_name_0: diag(4060, ts.DiagnosticCategory.Error, "Return_type_of_exported_function_has_or_is_using_private_name_0_4060", "Return type of exported function has or is using private name '{0}'."),
            Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4061, ts.DiagnosticCategory.Error, "Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_can_4061", "Parameter '{0}' of constructor from exported class has or is using name '{1}' from external module {2} but cannot be named."),
            Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4062, ts.DiagnosticCategory.Error, "Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_private_module_2_4062", "Parameter '{0}' of constructor from exported class has or is using name '{1}' from private module '{2}'."),
            Parameter_0_of_constructor_from_exported_class_has_or_is_using_private_name_1: diag(4063, ts.DiagnosticCategory.Error, "Parameter_0_of_constructor_from_exported_class_has_or_is_using_private_name_1_4063", "Parameter '{0}' of constructor from exported class has or is using private name '{1}'."),
            Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2: diag(4064, ts.DiagnosticCategory.Error, "Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_name_1_from_private_mod_4064", "Parameter '{0}' of constructor signature from exported interface has or is using name '{1}' from private module '{2}'."),
            Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1: diag(4065, ts.DiagnosticCategory.Error, "Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1_4065", "Parameter '{0}' of constructor signature from exported interface has or is using private name '{1}'."),
            Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2: diag(4066, ts.DiagnosticCategory.Error, "Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2_4066", "Parameter '{0}' of call signature from exported interface has or is using name '{1}' from private module '{2}'."),
            Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1: diag(4067, ts.DiagnosticCategory.Error, "Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1_4067", "Parameter '{0}' of call signature from exported interface has or is using private name '{1}'."),
            Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4068, ts.DiagnosticCategory.Error, "Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_external_module__4068", "Parameter '{0}' of public static method from exported class has or is using name '{1}' from external module {2} but cannot be named."),
            Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4069, ts.DiagnosticCategory.Error, "Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_private_module_2_4069", "Parameter '{0}' of public static method from exported class has or is using name '{1}' from private module '{2}'."),
            Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1: diag(4070, ts.DiagnosticCategory.Error, "Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1_4070", "Parameter '{0}' of public static method from exported class has or is using private name '{1}'."),
            Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4071, ts.DiagnosticCategory.Error, "Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_c_4071", "Parameter '{0}' of public method from exported class has or is using name '{1}' from external module {2} but cannot be named."),
            Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4072, ts.DiagnosticCategory.Error, "Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_private_module_2_4072", "Parameter '{0}' of public method from exported class has or is using name '{1}' from private module '{2}'."),
            Parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1: diag(4073, ts.DiagnosticCategory.Error, "Parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1_4073", "Parameter '{0}' of public method from exported class has or is using private name '{1}'."),
            Parameter_0_of_method_from_exported_interface_has_or_is_using_name_1_from_private_module_2: diag(4074, ts.DiagnosticCategory.Error, "Parameter_0_of_method_from_exported_interface_has_or_is_using_name_1_from_private_module_2_4074", "Parameter '{0}' of method from exported interface has or is using name '{1}' from private module '{2}'."),
            Parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1: diag(4075, ts.DiagnosticCategory.Error, "Parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1_4075", "Parameter '{0}' of method from exported interface has or is using private name '{1}'."),
            Parameter_0_of_exported_function_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4076, ts.DiagnosticCategory.Error, "Parameter_0_of_exported_function_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4076", "Parameter '{0}' of exported function has or is using name '{1}' from external module {2} but cannot be named."),
            Parameter_0_of_exported_function_has_or_is_using_name_1_from_private_module_2: diag(4077, ts.DiagnosticCategory.Error, "Parameter_0_of_exported_function_has_or_is_using_name_1_from_private_module_2_4077", "Parameter '{0}' of exported function has or is using name '{1}' from private module '{2}'."),
            Parameter_0_of_exported_function_has_or_is_using_private_name_1: diag(4078, ts.DiagnosticCategory.Error, "Parameter_0_of_exported_function_has_or_is_using_private_name_1_4078", "Parameter '{0}' of exported function has or is using private name '{1}'."),
            Exported_type_alias_0_has_or_is_using_private_name_1: diag(4081, ts.DiagnosticCategory.Error, "Exported_type_alias_0_has_or_is_using_private_name_1_4081", "Exported type alias '{0}' has or is using private name '{1}'."),
            Default_export_of_the_module_has_or_is_using_private_name_0: diag(4082, ts.DiagnosticCategory.Error, "Default_export_of_the_module_has_or_is_using_private_name_0_4082", "Default export of the module has or is using private name '{0}'."),
            Type_parameter_0_of_exported_type_alias_has_or_is_using_private_name_1: diag(4083, ts.DiagnosticCategory.Error, "Type_parameter_0_of_exported_type_alias_has_or_is_using_private_name_1_4083", "Type parameter '{0}' of exported type alias has or is using private name '{1}'."),
            Exported_type_alias_0_has_or_is_using_private_name_1_from_module_2: diag(4084, ts.DiagnosticCategory.Error, "Exported_type_alias_0_has_or_is_using_private_name_1_from_module_2_4084", "Exported type alias '{0}' has or is using private name '{1}' from module {2}."),
            Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_library_to_resolve_the_conflict: diag(4090, ts.DiagnosticCategory.Error, "Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_librar_4090", "Conflicting definitions for '{0}' found at '{1}' and '{2}'. Consider installing a specific version of this library to resolve the conflict."),
            Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2: diag(4091, ts.DiagnosticCategory.Error, "Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2_4091", "Parameter '{0}' of index signature from exported interface has or is using name '{1}' from private module '{2}'."),
            Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_private_name_1: diag(4092, ts.DiagnosticCategory.Error, "Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_private_name_1_4092", "Parameter '{0}' of index signature from exported interface has or is using private name '{1}'."),
            Property_0_of_exported_class_expression_may_not_be_private_or_protected: diag(4094, ts.DiagnosticCategory.Error, "Property_0_of_exported_class_expression_may_not_be_private_or_protected_4094", "Property '{0}' of exported class expression may not be private or protected."),
            Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4095, ts.DiagnosticCategory.Error, "Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_4095", "Public static method '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),
            Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4096, ts.DiagnosticCategory.Error, "Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4096", "Public static method '{0}' of exported class has or is using name '{1}' from private module '{2}'."),
            Public_static_method_0_of_exported_class_has_or_is_using_private_name_1: diag(4097, ts.DiagnosticCategory.Error, "Public_static_method_0_of_exported_class_has_or_is_using_private_name_1_4097", "Public static method '{0}' of exported class has or is using private name '{1}'."),
            Public_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4098, ts.DiagnosticCategory.Error, "Public_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4098", "Public method '{0}' of exported class has or is using name '{1}' from external module {2} but cannot be named."),
            Public_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2: diag(4099, ts.DiagnosticCategory.Error, "Public_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2_4099", "Public method '{0}' of exported class has or is using name '{1}' from private module '{2}'."),
            Public_method_0_of_exported_class_has_or_is_using_private_name_1: diag(4100, ts.DiagnosticCategory.Error, "Public_method_0_of_exported_class_has_or_is_using_private_name_1_4100", "Public method '{0}' of exported class has or is using private name '{1}'."),
            Method_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2: diag(4101, ts.DiagnosticCategory.Error, "Method_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2_4101", "Method '{0}' of exported interface has or is using name '{1}' from private module '{2}'."),
            Method_0_of_exported_interface_has_or_is_using_private_name_1: diag(4102, ts.DiagnosticCategory.Error, "Method_0_of_exported_interface_has_or_is_using_private_name_1_4102", "Method '{0}' of exported interface has or is using private name '{1}'."),
            Type_parameter_0_of_exported_mapped_object_type_is_using_private_name_1: diag(4103, ts.DiagnosticCategory.Error, "Type_parameter_0_of_exported_mapped_object_type_is_using_private_name_1_4103", "Type parameter '{0}' of exported mapped object type is using private name '{1}'."),
            The_type_0_is_readonly_and_cannot_be_assigned_to_the_mutable_type_1: diag(4104, ts.DiagnosticCategory.Error, "The_type_0_is_readonly_and_cannot_be_assigned_to_the_mutable_type_1_4104", "The type '{0}' is 'readonly' and cannot be assigned to the mutable type '{1}'."),
            Private_or_protected_member_0_cannot_be_accessed_on_a_type_parameter: diag(4105, ts.DiagnosticCategory.Error, "Private_or_protected_member_0_cannot_be_accessed_on_a_type_parameter_4105", "Private or protected member '{0}' cannot be accessed on a type parameter."),
            Parameter_0_of_accessor_has_or_is_using_private_name_1: diag(4106, ts.DiagnosticCategory.Error, "Parameter_0_of_accessor_has_or_is_using_private_name_1_4106", "Parameter '{0}' of accessor has or is using private name '{1}'."),
            Parameter_0_of_accessor_has_or_is_using_name_1_from_private_module_2: diag(4107, ts.DiagnosticCategory.Error, "Parameter_0_of_accessor_has_or_is_using_name_1_from_private_module_2_4107", "Parameter '{0}' of accessor has or is using name '{1}' from private module '{2}'."),
            Parameter_0_of_accessor_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named: diag(4108, ts.DiagnosticCategory.Error, "Parameter_0_of_accessor_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named_4108", "Parameter '{0}' of accessor has or is using name '{1}' from external module '{2}' but cannot be named."),
            Type_arguments_for_0_circularly_reference_themselves: diag(4109, ts.DiagnosticCategory.Error, "Type_arguments_for_0_circularly_reference_themselves_4109", "Type arguments for '{0}' circularly reference themselves."),
            Tuple_type_arguments_circularly_reference_themselves: diag(4110, ts.DiagnosticCategory.Error, "Tuple_type_arguments_circularly_reference_themselves_4110", "Tuple type arguments circularly reference themselves."),
            Property_0_comes_from_an_index_signature_so_it_must_be_accessed_with_0: diag(4111, ts.DiagnosticCategory.Error, "Property_0_comes_from_an_index_signature_so_it_must_be_accessed_with_0_4111", "Property '{0}' comes from an index signature, so it must be accessed with ['{0}']."),
            This_member_cannot_have_an_override_modifier_because_its_containing_class_0_does_not_extend_another_class: diag(4112, ts.DiagnosticCategory.Error, "This_member_cannot_have_an_override_modifier_because_its_containing_class_0_does_not_extend_another__4112", "This member cannot have an 'override' modifier because its containing class '{0}' does not extend another class."),
            This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0: diag(4113, ts.DiagnosticCategory.Error, "This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0_4113", "This member cannot have an 'override' modifier because it is not declared in the base class '{0}'."),
            This_member_must_have_an_override_modifier_because_it_overrides_a_member_in_the_base_class_0: diag(4114, ts.DiagnosticCategory.Error, "This_member_must_have_an_override_modifier_because_it_overrides_a_member_in_the_base_class_0_4114", "This member must have an 'override' modifier because it overrides a member in the base class '{0}'."),
            This_parameter_property_must_have_an_override_modifier_because_it_overrides_a_member_in_base_class_0: diag(4115, ts.DiagnosticCategory.Error, "This_parameter_property_must_have_an_override_modifier_because_it_overrides_a_member_in_base_class_0_4115", "This parameter property must have an 'override' modifier because it overrides a member in base class '{0}'."),
            This_member_must_have_an_override_modifier_because_it_overrides_an_abstract_method_that_is_declared_in_the_base_class_0: diag(4116, ts.DiagnosticCategory.Error, "This_member_must_have_an_override_modifier_because_it_overrides_an_abstract_method_that_is_declared__4116", "This member must have an 'override' modifier because it overrides an abstract method that is declared in the base class '{0}'."),
            This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0_Did_you_mean_1: diag(4117, ts.DiagnosticCategory.Error, "This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0_Did_you__4117", "This member cannot have an 'override' modifier because it is not declared in the base class '{0}'. Did you mean '{1}'?"),
            The_type_of_this_node_cannot_be_serialized_because_its_property_0_cannot_be_serialized: diag(4118, ts.DiagnosticCategory.Error, "The_type_of_this_node_cannot_be_serialized_because_its_property_0_cannot_be_serialized_4118", "The type of this node cannot be serialized because its property '{0}' cannot be serialized."),
            The_current_host_does_not_support_the_0_option: diag(5001, ts.DiagnosticCategory.Error, "The_current_host_does_not_support_the_0_option_5001", "The current host does not support the '{0}' option."),
            Cannot_find_the_common_subdirectory_path_for_the_input_files: diag(5009, ts.DiagnosticCategory.Error, "Cannot_find_the_common_subdirectory_path_for_the_input_files_5009", "Cannot find the common subdirectory path for the input files."),
            File_specification_cannot_end_in_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0: diag(5010, ts.DiagnosticCategory.Error, "File_specification_cannot_end_in_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0_5010", "File specification cannot end in a recursive directory wildcard ('**'): '{0}'."),
            Cannot_read_file_0_Colon_1: diag(5012, ts.DiagnosticCategory.Error, "Cannot_read_file_0_Colon_1_5012", "Cannot read file '{0}': {1}."),
            Failed_to_parse_file_0_Colon_1: diag(5014, ts.DiagnosticCategory.Error, "Failed_to_parse_file_0_Colon_1_5014", "Failed to parse file '{0}': {1}."),
            Unknown_compiler_option_0: diag(5023, ts.DiagnosticCategory.Error, "Unknown_compiler_option_0_5023", "Unknown compiler option '{0}'."),
            Compiler_option_0_requires_a_value_of_type_1: diag(5024, ts.DiagnosticCategory.Error, "Compiler_option_0_requires_a_value_of_type_1_5024", "Compiler option '{0}' requires a value of type {1}."),
            Unknown_compiler_option_0_Did_you_mean_1: diag(5025, ts.DiagnosticCategory.Error, "Unknown_compiler_option_0_Did_you_mean_1_5025", "Unknown compiler option '{0}'. Did you mean '{1}'?"),
            Could_not_write_file_0_Colon_1: diag(5033, ts.DiagnosticCategory.Error, "Could_not_write_file_0_Colon_1_5033", "Could not write file '{0}': {1}."),
            Option_project_cannot_be_mixed_with_source_files_on_a_command_line: diag(5042, ts.DiagnosticCategory.Error, "Option_project_cannot_be_mixed_with_source_files_on_a_command_line_5042", "Option 'project' cannot be mixed with source files on a command line."),
            Option_isolatedModules_can_only_be_used_when_either_option_module_is_provided_or_option_target_is_ES2015_or_higher: diag(5047, ts.DiagnosticCategory.Error, "Option_isolatedModules_can_only_be_used_when_either_option_module_is_provided_or_option_target_is_ES_5047", "Option 'isolatedModules' can only be used when either option '--module' is provided or option 'target' is 'ES2015' or higher."),
            Option_0_cannot_be_specified_when_option_target_is_ES3: diag(5048, ts.DiagnosticCategory.Error, "Option_0_cannot_be_specified_when_option_target_is_ES3_5048", "Option '{0}' cannot be specified when option 'target' is 'ES3'."),
            Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided: diag(5051, ts.DiagnosticCategory.Error, "Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided_5051", "Option '{0} can only be used when either option '--inlineSourceMap' or option '--sourceMap' is provided."),
            Option_0_cannot_be_specified_without_specifying_option_1: diag(5052, ts.DiagnosticCategory.Error, "Option_0_cannot_be_specified_without_specifying_option_1_5052", "Option '{0}' cannot be specified without specifying option '{1}'."),
            Option_0_cannot_be_specified_with_option_1: diag(5053, ts.DiagnosticCategory.Error, "Option_0_cannot_be_specified_with_option_1_5053", "Option '{0}' cannot be specified with option '{1}'."),
            A_tsconfig_json_file_is_already_defined_at_Colon_0: diag(5054, ts.DiagnosticCategory.Error, "A_tsconfig_json_file_is_already_defined_at_Colon_0_5054", "A 'tsconfig.json' file is already defined at: '{0}'."),
            Cannot_write_file_0_because_it_would_overwrite_input_file: diag(5055, ts.DiagnosticCategory.Error, "Cannot_write_file_0_because_it_would_overwrite_input_file_5055", "Cannot write file '{0}' because it would overwrite input file."),
            Cannot_write_file_0_because_it_would_be_overwritten_by_multiple_input_files: diag(5056, ts.DiagnosticCategory.Error, "Cannot_write_file_0_because_it_would_be_overwritten_by_multiple_input_files_5056", "Cannot write file '{0}' because it would be overwritten by multiple input files."),
            Cannot_find_a_tsconfig_json_file_at_the_specified_directory_Colon_0: diag(5057, ts.DiagnosticCategory.Error, "Cannot_find_a_tsconfig_json_file_at_the_specified_directory_Colon_0_5057", "Cannot find a tsconfig.json file at the specified directory: '{0}'."),
            The_specified_path_does_not_exist_Colon_0: diag(5058, ts.DiagnosticCategory.Error, "The_specified_path_does_not_exist_Colon_0_5058", "The specified path does not exist: '{0}'."),
            Invalid_value_for_reactNamespace_0_is_not_a_valid_identifier: diag(5059, ts.DiagnosticCategory.Error, "Invalid_value_for_reactNamespace_0_is_not_a_valid_identifier_5059", "Invalid value for '--reactNamespace'. '{0}' is not a valid identifier."),
            Pattern_0_can_have_at_most_one_Asterisk_character: diag(5061, ts.DiagnosticCategory.Error, "Pattern_0_can_have_at_most_one_Asterisk_character_5061", "Pattern '{0}' can have at most one '*' character."),
            Substitution_0_in_pattern_1_can_have_at_most_one_Asterisk_character: diag(5062, ts.DiagnosticCategory.Error, "Substitution_0_in_pattern_1_can_have_at_most_one_Asterisk_character_5062", "Substitution '{0}' in pattern '{1}' can have at most one '*' character."),
            Substitutions_for_pattern_0_should_be_an_array: diag(5063, ts.DiagnosticCategory.Error, "Substitutions_for_pattern_0_should_be_an_array_5063", "Substitutions for pattern '{0}' should be an array."),
            Substitution_0_for_pattern_1_has_incorrect_type_expected_string_got_2: diag(5064, ts.DiagnosticCategory.Error, "Substitution_0_for_pattern_1_has_incorrect_type_expected_string_got_2_5064", "Substitution '{0}' for pattern '{1}' has incorrect type, expected 'string', got '{2}'."),
            File_specification_cannot_contain_a_parent_directory_that_appears_after_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0: diag(5065, ts.DiagnosticCategory.Error, "File_specification_cannot_contain_a_parent_directory_that_appears_after_a_recursive_directory_wildca_5065", "File specification cannot contain a parent directory ('..') that appears after a recursive directory wildcard ('**'): '{0}'."),
            Substitutions_for_pattern_0_shouldn_t_be_an_empty_array: diag(5066, ts.DiagnosticCategory.Error, "Substitutions_for_pattern_0_shouldn_t_be_an_empty_array_5066", "Substitutions for pattern '{0}' shouldn't be an empty array."),
            Invalid_value_for_jsxFactory_0_is_not_a_valid_identifier_or_qualified_name: diag(5067, ts.DiagnosticCategory.Error, "Invalid_value_for_jsxFactory_0_is_not_a_valid_identifier_or_qualified_name_5067", "Invalid value for 'jsxFactory'. '{0}' is not a valid identifier or qualified-name."),
            Adding_a_tsconfig_json_file_will_help_organize_projects_that_contain_both_TypeScript_and_JavaScript_files_Learn_more_at_https_Colon_Slash_Slashaka_ms_Slashtsconfig: diag(5068, ts.DiagnosticCategory.Error, "Adding_a_tsconfig_json_file_will_help_organize_projects_that_contain_both_TypeScript_and_JavaScript__5068", "Adding a tsconfig.json file will help organize projects that contain both TypeScript and JavaScript files. Learn more at https://aka.ms/tsconfig."),
            Option_0_cannot_be_specified_without_specifying_option_1_or_option_2: diag(5069, ts.DiagnosticCategory.Error, "Option_0_cannot_be_specified_without_specifying_option_1_or_option_2_5069", "Option '{0}' cannot be specified without specifying option '{1}' or option '{2}'."),
            Option_resolveJsonModule_cannot_be_specified_without_node_module_resolution_strategy: diag(5070, ts.DiagnosticCategory.Error, "Option_resolveJsonModule_cannot_be_specified_without_node_module_resolution_strategy_5070", "Option '--resolveJsonModule' cannot be specified without 'node' module resolution strategy."),
            Option_resolveJsonModule_can_only_be_specified_when_module_code_generation_is_commonjs_amd_es2015_or_esNext: diag(5071, ts.DiagnosticCategory.Error, "Option_resolveJsonModule_can_only_be_specified_when_module_code_generation_is_commonjs_amd_es2015_or_5071", "Option '--resolveJsonModule' can only be specified when module code generation is 'commonjs', 'amd', 'es2015' or 'esNext'."),
            Unknown_build_option_0: diag(5072, ts.DiagnosticCategory.Error, "Unknown_build_option_0_5072", "Unknown build option '{0}'."),
            Build_option_0_requires_a_value_of_type_1: diag(5073, ts.DiagnosticCategory.Error, "Build_option_0_requires_a_value_of_type_1_5073", "Build option '{0}' requires a value of type {1}."),
            Option_incremental_can_only_be_specified_using_tsconfig_emitting_to_single_file_or_when_option_tsBuildInfoFile_is_specified: diag(5074, ts.DiagnosticCategory.Error, "Option_incremental_can_only_be_specified_using_tsconfig_emitting_to_single_file_or_when_option_tsBui_5074", "Option '--incremental' can only be specified using tsconfig, emitting to single file or when option '--tsBuildInfoFile' is specified."),
            _0_is_assignable_to_the_constraint_of_type_1_but_1_could_be_instantiated_with_a_different_subtype_of_constraint_2: diag(5075, ts.DiagnosticCategory.Error, "_0_is_assignable_to_the_constraint_of_type_1_but_1_could_be_instantiated_with_a_different_subtype_of_5075", "'{0}' is assignable to the constraint of type '{1}', but '{1}' could be instantiated with a different subtype of constraint '{2}'."),
            _0_and_1_operations_cannot_be_mixed_without_parentheses: diag(5076, ts.DiagnosticCategory.Error, "_0_and_1_operations_cannot_be_mixed_without_parentheses_5076", "'{0}' and '{1}' operations cannot be mixed without parentheses."),
            Unknown_build_option_0_Did_you_mean_1: diag(5077, ts.DiagnosticCategory.Error, "Unknown_build_option_0_Did_you_mean_1_5077", "Unknown build option '{0}'. Did you mean '{1}'?"),
            Unknown_watch_option_0: diag(5078, ts.DiagnosticCategory.Error, "Unknown_watch_option_0_5078", "Unknown watch option '{0}'."),
            Unknown_watch_option_0_Did_you_mean_1: diag(5079, ts.DiagnosticCategory.Error, "Unknown_watch_option_0_Did_you_mean_1_5079", "Unknown watch option '{0}'. Did you mean '{1}'?"),
            Watch_option_0_requires_a_value_of_type_1: diag(5080, ts.DiagnosticCategory.Error, "Watch_option_0_requires_a_value_of_type_1_5080", "Watch option '{0}' requires a value of type {1}."),
            Cannot_find_a_tsconfig_json_file_at_the_current_directory_Colon_0: diag(5081, ts.DiagnosticCategory.Error, "Cannot_find_a_tsconfig_json_file_at_the_current_directory_Colon_0_5081", "Cannot find a tsconfig.json file at the current directory: {0}."),
            _0_could_be_instantiated_with_an_arbitrary_type_which_could_be_unrelated_to_1: diag(5082, ts.DiagnosticCategory.Error, "_0_could_be_instantiated_with_an_arbitrary_type_which_could_be_unrelated_to_1_5082", "'{0}' could be instantiated with an arbitrary type which could be unrelated to '{1}'."),
            Cannot_read_file_0: diag(5083, ts.DiagnosticCategory.Error, "Cannot_read_file_0_5083", "Cannot read file '{0}'."),
            Tuple_members_must_all_have_names_or_all_not_have_names: diag(5084, ts.DiagnosticCategory.Error, "Tuple_members_must_all_have_names_or_all_not_have_names_5084", "Tuple members must all have names or all not have names."),
            A_tuple_member_cannot_be_both_optional_and_rest: diag(5085, ts.DiagnosticCategory.Error, "A_tuple_member_cannot_be_both_optional_and_rest_5085", "A tuple member cannot be both optional and rest."),
            A_labeled_tuple_element_is_declared_as_optional_with_a_question_mark_after_the_name_and_before_the_colon_rather_than_after_the_type: diag(5086, ts.DiagnosticCategory.Error, "A_labeled_tuple_element_is_declared_as_optional_with_a_question_mark_after_the_name_and_before_the_c_5086", "A labeled tuple element is declared as optional with a question mark after the name and before the colon, rather than after the type."),
            A_labeled_tuple_element_is_declared_as_rest_with_a_before_the_name_rather_than_before_the_type: diag(5087, ts.DiagnosticCategory.Error, "A_labeled_tuple_element_is_declared_as_rest_with_a_before_the_name_rather_than_before_the_type_5087", "A labeled tuple element is declared as rest with a '...' before the name, rather than before the type."),
            The_inferred_type_of_0_references_a_type_with_a_cyclic_structure_which_cannot_be_trivially_serialized_A_type_annotation_is_necessary: diag(5088, ts.DiagnosticCategory.Error, "The_inferred_type_of_0_references_a_type_with_a_cyclic_structure_which_cannot_be_trivially_serialize_5088", "The inferred type of '{0}' references a type with a cyclic structure which cannot be trivially serialized. A type annotation is necessary."),
            Option_0_cannot_be_specified_when_option_jsx_is_1: diag(5089, ts.DiagnosticCategory.Error, "Option_0_cannot_be_specified_when_option_jsx_is_1_5089", "Option '{0}' cannot be specified when option 'jsx' is '{1}'."),
            Non_relative_paths_are_not_allowed_when_baseUrl_is_not_set_Did_you_forget_a_leading_Slash: diag(5090, ts.DiagnosticCategory.Error, "Non_relative_paths_are_not_allowed_when_baseUrl_is_not_set_Did_you_forget_a_leading_Slash_5090", "Non-relative paths are not allowed when 'baseUrl' is not set. Did you forget a leading './'?"),
            Option_preserveConstEnums_cannot_be_disabled_when_isolatedModules_is_enabled: diag(5091, ts.DiagnosticCategory.Error, "Option_preserveConstEnums_cannot_be_disabled_when_isolatedModules_is_enabled_5091", "Option 'preserveConstEnums' cannot be disabled when 'isolatedModules' is enabled."),
            The_root_value_of_a_0_file_must_be_an_object: diag(5092, ts.DiagnosticCategory.Error, "The_root_value_of_a_0_file_must_be_an_object_5092", "The root value of a '{0}' file must be an object."),
            Compiler_option_0_may_only_be_used_with_build: diag(5093, ts.DiagnosticCategory.Error, "Compiler_option_0_may_only_be_used_with_build_5093", "Compiler option '--{0}' may only be used with '--build'."),
            Compiler_option_0_may_not_be_used_with_build: diag(5094, ts.DiagnosticCategory.Error, "Compiler_option_0_may_not_be_used_with_build_5094", "Compiler option '--{0}' may not be used with '--build'."),
            Generates_a_sourcemap_for_each_corresponding_d_ts_file: diag(6000, ts.DiagnosticCategory.Message, "Generates_a_sourcemap_for_each_corresponding_d_ts_file_6000", "Generates a sourcemap for each corresponding '.d.ts' file."),
            Concatenate_and_emit_output_to_single_file: diag(6001, ts.DiagnosticCategory.Message, "Concatenate_and_emit_output_to_single_file_6001", "Concatenate and emit output to single file."),
            Generates_corresponding_d_ts_file: diag(6002, ts.DiagnosticCategory.Message, "Generates_corresponding_d_ts_file_6002", "Generates corresponding '.d.ts' file."),
            Specify_the_location_where_debugger_should_locate_map_files_instead_of_generated_locations: diag(6655, ts.DiagnosticCategory.Message, "Specify_the_location_where_debugger_should_locate_map_files_instead_of_generated_locations_6655", "Specify the location where debugger should locate map files instead of generated locations."),
            Specify_the_location_where_debugger_should_locate_TypeScript_files_instead_of_source_locations: diag(6004, ts.DiagnosticCategory.Message, "Specify_the_location_where_debugger_should_locate_TypeScript_files_instead_of_source_locations_6004", "Specify the location where debugger should locate TypeScript files instead of source locations."),
            Watch_input_files: diag(6005, ts.DiagnosticCategory.Message, "Watch_input_files_6005", "Watch input files."),
            Redirect_output_structure_to_the_directory: diag(6006, ts.DiagnosticCategory.Message, "Redirect_output_structure_to_the_directory_6006", "Redirect output structure to the directory."),
            Do_not_erase_const_enum_declarations_in_generated_code: diag(6007, ts.DiagnosticCategory.Message, "Do_not_erase_const_enum_declarations_in_generated_code_6007", "Do not erase const enum declarations in generated code."),
            Do_not_emit_outputs_if_any_errors_were_reported: diag(6008, ts.DiagnosticCategory.Message, "Do_not_emit_outputs_if_any_errors_were_reported_6008", "Do not emit outputs if any errors were reported."),
            Do_not_emit_comments_to_output: diag(6009, ts.DiagnosticCategory.Message, "Do_not_emit_comments_to_output_6009", "Do not emit comments to output."),
            Do_not_emit_outputs: diag(6010, ts.DiagnosticCategory.Message, "Do_not_emit_outputs_6010", "Do not emit outputs."),
            Allow_default_imports_from_modules_with_no_default_export_This_does_not_affect_code_emit_just_typechecking: diag(6011, ts.DiagnosticCategory.Message, "Allow_default_imports_from_modules_with_no_default_export_This_does_not_affect_code_emit_just_typech_6011", "Allow default imports from modules with no default export. This does not affect code emit, just typechecking."),
            Skip_type_checking_of_declaration_files: diag(6012, ts.DiagnosticCategory.Message, "Skip_type_checking_of_declaration_files_6012", "Skip type checking of declaration files."),
            Do_not_resolve_the_real_path_of_symlinks: diag(6013, ts.DiagnosticCategory.Message, "Do_not_resolve_the_real_path_of_symlinks_6013", "Do not resolve the real path of symlinks."),
            Only_emit_d_ts_declaration_files: diag(6014, ts.DiagnosticCategory.Message, "Only_emit_d_ts_declaration_files_6014", "Only emit '.d.ts' declaration files."),
            Specify_ECMAScript_target_version: diag(6015, ts.DiagnosticCategory.Message, "Specify_ECMAScript_target_version_6015", "Specify ECMAScript target version."),
            Specify_module_code_generation: diag(6016, ts.DiagnosticCategory.Message, "Specify_module_code_generation_6016", "Specify module code generation."),
            Print_this_message: diag(6017, ts.DiagnosticCategory.Message, "Print_this_message_6017", "Print this message."),
            Print_the_compiler_s_version: diag(6019, ts.DiagnosticCategory.Message, "Print_the_compiler_s_version_6019", "Print the compiler's version."),
            Compile_the_project_given_the_path_to_its_configuration_file_or_to_a_folder_with_a_tsconfig_json: diag(6020, ts.DiagnosticCategory.Message, "Compile_the_project_given_the_path_to_its_configuration_file_or_to_a_folder_with_a_tsconfig_json_6020", "Compile the project given the path to its configuration file, or to a folder with a 'tsconfig.json'."),
            Syntax_Colon_0: diag(6023, ts.DiagnosticCategory.Message, "Syntax_Colon_0_6023", "Syntax: {0}"),
            options: diag(6024, ts.DiagnosticCategory.Message, "options_6024", "options"),
            file: diag(6025, ts.DiagnosticCategory.Message, "file_6025", "file"),
            Examples_Colon_0: diag(6026, ts.DiagnosticCategory.Message, "Examples_Colon_0_6026", "Examples: {0}"),
            Options_Colon: diag(6027, ts.DiagnosticCategory.Message, "Options_Colon_6027", "Options:"),
            Version_0: diag(6029, ts.DiagnosticCategory.Message, "Version_0_6029", "Version {0}"),
            Insert_command_line_options_and_files_from_a_file: diag(6030, ts.DiagnosticCategory.Message, "Insert_command_line_options_and_files_from_a_file_6030", "Insert command line options and files from a file."),
            Starting_compilation_in_watch_mode: diag(6031, ts.DiagnosticCategory.Message, "Starting_compilation_in_watch_mode_6031", "Starting compilation in watch mode..."),
            File_change_detected_Starting_incremental_compilation: diag(6032, ts.DiagnosticCategory.Message, "File_change_detected_Starting_incremental_compilation_6032", "File change detected. Starting incremental compilation..."),
            KIND: diag(6034, ts.DiagnosticCategory.Message, "KIND_6034", "KIND"),
            FILE: diag(6035, ts.DiagnosticCategory.Message, "FILE_6035", "FILE"),
            VERSION: diag(6036, ts.DiagnosticCategory.Message, "VERSION_6036", "VERSION"),
            LOCATION: diag(6037, ts.DiagnosticCategory.Message, "LOCATION_6037", "LOCATION"),
            DIRECTORY: diag(6038, ts.DiagnosticCategory.Message, "DIRECTORY_6038", "DIRECTORY"),
            STRATEGY: diag(6039, ts.DiagnosticCategory.Message, "STRATEGY_6039", "STRATEGY"),
            FILE_OR_DIRECTORY: diag(6040, ts.DiagnosticCategory.Message, "FILE_OR_DIRECTORY_6040", "FILE OR DIRECTORY"),
            Generates_corresponding_map_file: diag(6043, ts.DiagnosticCategory.Message, "Generates_corresponding_map_file_6043", "Generates corresponding '.map' file."),
            Compiler_option_0_expects_an_argument: diag(6044, ts.DiagnosticCategory.Error, "Compiler_option_0_expects_an_argument_6044", "Compiler option '{0}' expects an argument."),
            Unterminated_quoted_string_in_response_file_0: diag(6045, ts.DiagnosticCategory.Error, "Unterminated_quoted_string_in_response_file_0_6045", "Unterminated quoted string in response file '{0}'."),
            Argument_for_0_option_must_be_Colon_1: diag(6046, ts.DiagnosticCategory.Error, "Argument_for_0_option_must_be_Colon_1_6046", "Argument for '{0}' option must be: {1}."),
            Locale_must_be_of_the_form_language_or_language_territory_For_example_0_or_1: diag(6048, ts.DiagnosticCategory.Error, "Locale_must_be_of_the_form_language_or_language_territory_For_example_0_or_1_6048", "Locale must be of the form <language> or <language>-<territory>. For example '{0}' or '{1}'."),
            Unsupported_locale_0: diag(6049, ts.DiagnosticCategory.Error, "Unsupported_locale_0_6049", "Unsupported locale '{0}'."),
            Unable_to_open_file_0: diag(6050, ts.DiagnosticCategory.Error, "Unable_to_open_file_0_6050", "Unable to open file '{0}'."),
            Corrupted_locale_file_0: diag(6051, ts.DiagnosticCategory.Error, "Corrupted_locale_file_0_6051", "Corrupted locale file {0}."),
            Raise_error_on_expressions_and_declarations_with_an_implied_any_type: diag(6052, ts.DiagnosticCategory.Message, "Raise_error_on_expressions_and_declarations_with_an_implied_any_type_6052", "Raise error on expressions and declarations with an implied 'any' type."),
            File_0_not_found: diag(6053, ts.DiagnosticCategory.Error, "File_0_not_found_6053", "File '{0}' not found."),
            File_0_has_an_unsupported_extension_The_only_supported_extensions_are_1: diag(6054, ts.DiagnosticCategory.Error, "File_0_has_an_unsupported_extension_The_only_supported_extensions_are_1_6054", "File '{0}' has an unsupported extension. The only supported extensions are {1}."),
            Suppress_noImplicitAny_errors_for_indexing_objects_lacking_index_signatures: diag(6055, ts.DiagnosticCategory.Message, "Suppress_noImplicitAny_errors_for_indexing_objects_lacking_index_signatures_6055", "Suppress noImplicitAny errors for indexing objects lacking index signatures."),
            Do_not_emit_declarations_for_code_that_has_an_internal_annotation: diag(6056, ts.DiagnosticCategory.Message, "Do_not_emit_declarations_for_code_that_has_an_internal_annotation_6056", "Do not emit declarations for code that has an '@internal' annotation."),
            Specify_the_root_directory_of_input_files_Use_to_control_the_output_directory_structure_with_outDir: diag(6058, ts.DiagnosticCategory.Message, "Specify_the_root_directory_of_input_files_Use_to_control_the_output_directory_structure_with_outDir_6058", "Specify the root directory of input files. Use to control the output directory structure with --outDir."),
            File_0_is_not_under_rootDir_1_rootDir_is_expected_to_contain_all_source_files: diag(6059, ts.DiagnosticCategory.Error, "File_0_is_not_under_rootDir_1_rootDir_is_expected_to_contain_all_source_files_6059", "File '{0}' is not under 'rootDir' '{1}'. 'rootDir' is expected to contain all source files."),
            Specify_the_end_of_line_sequence_to_be_used_when_emitting_files_Colon_CRLF_dos_or_LF_unix: diag(6060, ts.DiagnosticCategory.Message, "Specify_the_end_of_line_sequence_to_be_used_when_emitting_files_Colon_CRLF_dos_or_LF_unix_6060", "Specify the end of line sequence to be used when emitting files: 'CRLF' (dos) or 'LF' (unix)."),
            NEWLINE: diag(6061, ts.DiagnosticCategory.Message, "NEWLINE_6061", "NEWLINE"),
            Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_null_on_command_line: diag(6064, ts.DiagnosticCategory.Error, "Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_null_on_command_line_6064", "Option '{0}' can only be specified in 'tsconfig.json' file or set to 'null' on command line."),
            Enables_experimental_support_for_ES7_decorators: diag(6065, ts.DiagnosticCategory.Message, "Enables_experimental_support_for_ES7_decorators_6065", "Enables experimental support for ES7 decorators."),
            Enables_experimental_support_for_emitting_type_metadata_for_decorators: diag(6066, ts.DiagnosticCategory.Message, "Enables_experimental_support_for_emitting_type_metadata_for_decorators_6066", "Enables experimental support for emitting type metadata for decorators."),
            Enables_experimental_support_for_ES7_async_functions: diag(6068, ts.DiagnosticCategory.Message, "Enables_experimental_support_for_ES7_async_functions_6068", "Enables experimental support for ES7 async functions."),
            Specify_module_resolution_strategy_Colon_node_Node_js_or_classic_TypeScript_pre_1_6: diag(6069, ts.DiagnosticCategory.Message, "Specify_module_resolution_strategy_Colon_node_Node_js_or_classic_TypeScript_pre_1_6_6069", "Specify module resolution strategy: 'node' (Node.js) or 'classic' (TypeScript pre-1.6)."),
            Initializes_a_TypeScript_project_and_creates_a_tsconfig_json_file: diag(6070, ts.DiagnosticCategory.Message, "Initializes_a_TypeScript_project_and_creates_a_tsconfig_json_file_6070", "Initializes a TypeScript project and creates a tsconfig.json file."),
            Successfully_created_a_tsconfig_json_file: diag(6071, ts.DiagnosticCategory.Message, "Successfully_created_a_tsconfig_json_file_6071", "Successfully created a tsconfig.json file."),
            Suppress_excess_property_checks_for_object_literals: diag(6072, ts.DiagnosticCategory.Message, "Suppress_excess_property_checks_for_object_literals_6072", "Suppress excess property checks for object literals."),
            Stylize_errors_and_messages_using_color_and_context_experimental: diag(6073, ts.DiagnosticCategory.Message, "Stylize_errors_and_messages_using_color_and_context_experimental_6073", "Stylize errors and messages using color and context (experimental)."),
            Do_not_report_errors_on_unused_labels: diag(6074, ts.DiagnosticCategory.Message, "Do_not_report_errors_on_unused_labels_6074", "Do not report errors on unused labels."),
            Report_error_when_not_all_code_paths_in_function_return_a_value: diag(6075, ts.DiagnosticCategory.Message, "Report_error_when_not_all_code_paths_in_function_return_a_value_6075", "Report error when not all code paths in function return a value."),
            Report_errors_for_fallthrough_cases_in_switch_statement: diag(6076, ts.DiagnosticCategory.Message, "Report_errors_for_fallthrough_cases_in_switch_statement_6076", "Report errors for fallthrough cases in switch statement."),
            Do_not_report_errors_on_unreachable_code: diag(6077, ts.DiagnosticCategory.Message, "Do_not_report_errors_on_unreachable_code_6077", "Do not report errors on unreachable code."),
            Disallow_inconsistently_cased_references_to_the_same_file: diag(6078, ts.DiagnosticCategory.Message, "Disallow_inconsistently_cased_references_to_the_same_file_6078", "Disallow inconsistently-cased references to the same file."),
            Specify_library_files_to_be_included_in_the_compilation: diag(6079, ts.DiagnosticCategory.Message, "Specify_library_files_to_be_included_in_the_compilation_6079", "Specify library files to be included in the compilation."),
            Specify_JSX_code_generation: diag(6080, ts.DiagnosticCategory.Message, "Specify_JSX_code_generation_6080", "Specify JSX code generation."),
            File_0_has_an_unsupported_extension_so_skipping_it: diag(6081, ts.DiagnosticCategory.Message, "File_0_has_an_unsupported_extension_so_skipping_it_6081", "File '{0}' has an unsupported extension, so skipping it."),
            Only_amd_and_system_modules_are_supported_alongside_0: diag(6082, ts.DiagnosticCategory.Error, "Only_amd_and_system_modules_are_supported_alongside_0_6082", "Only 'amd' and 'system' modules are supported alongside --{0}."),
            Base_directory_to_resolve_non_absolute_module_names: diag(6083, ts.DiagnosticCategory.Message, "Base_directory_to_resolve_non_absolute_module_names_6083", "Base directory to resolve non-absolute module names."),
            Deprecated_Use_jsxFactory_instead_Specify_the_object_invoked_for_createElement_when_targeting_react_JSX_emit: diag(6084, ts.DiagnosticCategory.Message, "Deprecated_Use_jsxFactory_instead_Specify_the_object_invoked_for_createElement_when_targeting_react__6084", "[Deprecated] Use '--jsxFactory' instead. Specify the object invoked for createElement when targeting 'react' JSX emit"),
            Enable_tracing_of_the_name_resolution_process: diag(6085, ts.DiagnosticCategory.Message, "Enable_tracing_of_the_name_resolution_process_6085", "Enable tracing of the name resolution process."),
            Resolving_module_0_from_1: diag(6086, ts.DiagnosticCategory.Message, "Resolving_module_0_from_1_6086", "======== Resolving module '{0}' from '{1}'. ========"),
            Explicitly_specified_module_resolution_kind_Colon_0: diag(6087, ts.DiagnosticCategory.Message, "Explicitly_specified_module_resolution_kind_Colon_0_6087", "Explicitly specified module resolution kind: '{0}'."),
            Module_resolution_kind_is_not_specified_using_0: diag(6088, ts.DiagnosticCategory.Message, "Module_resolution_kind_is_not_specified_using_0_6088", "Module resolution kind is not specified, using '{0}'."),
            Module_name_0_was_successfully_resolved_to_1: diag(6089, ts.DiagnosticCategory.Message, "Module_name_0_was_successfully_resolved_to_1_6089", "======== Module name '{0}' was successfully resolved to '{1}'. ========"),
            Module_name_0_was_not_resolved: diag(6090, ts.DiagnosticCategory.Message, "Module_name_0_was_not_resolved_6090", "======== Module name '{0}' was not resolved. ========"),
            paths_option_is_specified_looking_for_a_pattern_to_match_module_name_0: diag(6091, ts.DiagnosticCategory.Message, "paths_option_is_specified_looking_for_a_pattern_to_match_module_name_0_6091", "'paths' option is specified, looking for a pattern to match module name '{0}'."),
            Module_name_0_matched_pattern_1: diag(6092, ts.DiagnosticCategory.Message, "Module_name_0_matched_pattern_1_6092", "Module name '{0}', matched pattern '{1}'."),
            Trying_substitution_0_candidate_module_location_Colon_1: diag(6093, ts.DiagnosticCategory.Message, "Trying_substitution_0_candidate_module_location_Colon_1_6093", "Trying substitution '{0}', candidate module location: '{1}'."),
            Resolving_module_name_0_relative_to_base_url_1_2: diag(6094, ts.DiagnosticCategory.Message, "Resolving_module_name_0_relative_to_base_url_1_2_6094", "Resolving module name '{0}' relative to base url '{1}' - '{2}'."),
            Loading_module_as_file_Slash_folder_candidate_module_location_0_target_file_type_1: diag(6095, ts.DiagnosticCategory.Message, "Loading_module_as_file_Slash_folder_candidate_module_location_0_target_file_type_1_6095", "Loading module as file / folder, candidate module location '{0}', target file type '{1}'."),
            File_0_does_not_exist: diag(6096, ts.DiagnosticCategory.Message, "File_0_does_not_exist_6096", "File '{0}' does not exist."),
            File_0_exist_use_it_as_a_name_resolution_result: diag(6097, ts.DiagnosticCategory.Message, "File_0_exist_use_it_as_a_name_resolution_result_6097", "File '{0}' exist - use it as a name resolution result."),
            Loading_module_0_from_node_modules_folder_target_file_type_1: diag(6098, ts.DiagnosticCategory.Message, "Loading_module_0_from_node_modules_folder_target_file_type_1_6098", "Loading module '{0}' from 'node_modules' folder, target file type '{1}'."),
            Found_package_json_at_0: diag(6099, ts.DiagnosticCategory.Message, "Found_package_json_at_0_6099", "Found 'package.json' at '{0}'."),
            package_json_does_not_have_a_0_field: diag(6100, ts.DiagnosticCategory.Message, "package_json_does_not_have_a_0_field_6100", "'package.json' does not have a '{0}' field."),
            package_json_has_0_field_1_that_references_2: diag(6101, ts.DiagnosticCategory.Message, "package_json_has_0_field_1_that_references_2_6101", "'package.json' has '{0}' field '{1}' that references '{2}'."),
            Allow_javascript_files_to_be_compiled: diag(6102, ts.DiagnosticCategory.Message, "Allow_javascript_files_to_be_compiled_6102", "Allow javascript files to be compiled."),
            Option_0_should_have_array_of_strings_as_a_value: diag(6103, ts.DiagnosticCategory.Error, "Option_0_should_have_array_of_strings_as_a_value_6103", "Option '{0}' should have array of strings as a value."),
            Checking_if_0_is_the_longest_matching_prefix_for_1_2: diag(6104, ts.DiagnosticCategory.Message, "Checking_if_0_is_the_longest_matching_prefix_for_1_2_6104", "Checking if '{0}' is the longest matching prefix for '{1}' - '{2}'."),
            Expected_type_of_0_field_in_package_json_to_be_1_got_2: diag(6105, ts.DiagnosticCategory.Message, "Expected_type_of_0_field_in_package_json_to_be_1_got_2_6105", "Expected type of '{0}' field in 'package.json' to be '{1}', got '{2}'."),
            baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1: diag(6106, ts.DiagnosticCategory.Message, "baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1_6106", "'baseUrl' option is set to '{0}', using this value to resolve non-relative module name '{1}'."),
            rootDirs_option_is_set_using_it_to_resolve_relative_module_name_0: diag(6107, ts.DiagnosticCategory.Message, "rootDirs_option_is_set_using_it_to_resolve_relative_module_name_0_6107", "'rootDirs' option is set, using it to resolve relative module name '{0}'."),
            Longest_matching_prefix_for_0_is_1: diag(6108, ts.DiagnosticCategory.Message, "Longest_matching_prefix_for_0_is_1_6108", "Longest matching prefix for '{0}' is '{1}'."),
            Loading_0_from_the_root_dir_1_candidate_location_2: diag(6109, ts.DiagnosticCategory.Message, "Loading_0_from_the_root_dir_1_candidate_location_2_6109", "Loading '{0}' from the root dir '{1}', candidate location '{2}'."),
            Trying_other_entries_in_rootDirs: diag(6110, ts.DiagnosticCategory.Message, "Trying_other_entries_in_rootDirs_6110", "Trying other entries in 'rootDirs'."),
            Module_resolution_using_rootDirs_has_failed: diag(6111, ts.DiagnosticCategory.Message, "Module_resolution_using_rootDirs_has_failed_6111", "Module resolution using 'rootDirs' has failed."),
            Do_not_emit_use_strict_directives_in_module_output: diag(6112, ts.DiagnosticCategory.Message, "Do_not_emit_use_strict_directives_in_module_output_6112", "Do not emit 'use strict' directives in module output."),
            Enable_strict_null_checks: diag(6113, ts.DiagnosticCategory.Message, "Enable_strict_null_checks_6113", "Enable strict null checks."),
            Unknown_option_excludes_Did_you_mean_exclude: diag(6114, ts.DiagnosticCategory.Error, "Unknown_option_excludes_Did_you_mean_exclude_6114", "Unknown option 'excludes'. Did you mean 'exclude'?"),
            Raise_error_on_this_expressions_with_an_implied_any_type: diag(6115, ts.DiagnosticCategory.Message, "Raise_error_on_this_expressions_with_an_implied_any_type_6115", "Raise error on 'this' expressions with an implied 'any' type."),
            Resolving_type_reference_directive_0_containing_file_1_root_directory_2: diag(6116, ts.DiagnosticCategory.Message, "Resolving_type_reference_directive_0_containing_file_1_root_directory_2_6116", "======== Resolving type reference directive '{0}', containing file '{1}', root directory '{2}'. ========"),
            Resolving_using_primary_search_paths: diag(6117, ts.DiagnosticCategory.Message, "Resolving_using_primary_search_paths_6117", "Resolving using primary search paths..."),
            Resolving_from_node_modules_folder: diag(6118, ts.DiagnosticCategory.Message, "Resolving_from_node_modules_folder_6118", "Resolving from node_modules folder..."),
            Type_reference_directive_0_was_successfully_resolved_to_1_primary_Colon_2: diag(6119, ts.DiagnosticCategory.Message, "Type_reference_directive_0_was_successfully_resolved_to_1_primary_Colon_2_6119", "======== Type reference directive '{0}' was successfully resolved to '{1}', primary: {2}. ========"),
            Type_reference_directive_0_was_not_resolved: diag(6120, ts.DiagnosticCategory.Message, "Type_reference_directive_0_was_not_resolved_6120", "======== Type reference directive '{0}' was not resolved. ========"),
            Resolving_with_primary_search_path_0: diag(6121, ts.DiagnosticCategory.Message, "Resolving_with_primary_search_path_0_6121", "Resolving with primary search path '{0}'."),
            Root_directory_cannot_be_determined_skipping_primary_search_paths: diag(6122, ts.DiagnosticCategory.Message, "Root_directory_cannot_be_determined_skipping_primary_search_paths_6122", "Root directory cannot be determined, skipping primary search paths."),
            Resolving_type_reference_directive_0_containing_file_1_root_directory_not_set: diag(6123, ts.DiagnosticCategory.Message, "Resolving_type_reference_directive_0_containing_file_1_root_directory_not_set_6123", "======== Resolving type reference directive '{0}', containing file '{1}', root directory not set. ========"),
            Type_declaration_files_to_be_included_in_compilation: diag(6124, ts.DiagnosticCategory.Message, "Type_declaration_files_to_be_included_in_compilation_6124", "Type declaration files to be included in compilation."),
            Looking_up_in_node_modules_folder_initial_location_0: diag(6125, ts.DiagnosticCategory.Message, "Looking_up_in_node_modules_folder_initial_location_0_6125", "Looking up in 'node_modules' folder, initial location '{0}'."),
            Containing_file_is_not_specified_and_root_directory_cannot_be_determined_skipping_lookup_in_node_modules_folder: diag(6126, ts.DiagnosticCategory.Message, "Containing_file_is_not_specified_and_root_directory_cannot_be_determined_skipping_lookup_in_node_mod_6126", "Containing file is not specified and root directory cannot be determined, skipping lookup in 'node_modules' folder."),
            Resolving_type_reference_directive_0_containing_file_not_set_root_directory_1: diag(6127, ts.DiagnosticCategory.Message, "Resolving_type_reference_directive_0_containing_file_not_set_root_directory_1_6127", "======== Resolving type reference directive '{0}', containing file not set, root directory '{1}'. ========"),
            Resolving_type_reference_directive_0_containing_file_not_set_root_directory_not_set: diag(6128, ts.DiagnosticCategory.Message, "Resolving_type_reference_directive_0_containing_file_not_set_root_directory_not_set_6128", "======== Resolving type reference directive '{0}', containing file not set, root directory not set. ========"),
            Resolving_real_path_for_0_result_1: diag(6130, ts.DiagnosticCategory.Message, "Resolving_real_path_for_0_result_1_6130", "Resolving real path for '{0}', result '{1}'."),
            Cannot_compile_modules_using_option_0_unless_the_module_flag_is_amd_or_system: diag(6131, ts.DiagnosticCategory.Error, "Cannot_compile_modules_using_option_0_unless_the_module_flag_is_amd_or_system_6131", "Cannot compile modules using option '{0}' unless the '--module' flag is 'amd' or 'system'."),
            File_name_0_has_a_1_extension_stripping_it: diag(6132, ts.DiagnosticCategory.Message, "File_name_0_has_a_1_extension_stripping_it_6132", "File name '{0}' has a '{1}' extension - stripping it."),
            _0_is_declared_but_its_value_is_never_read: diag(6133, ts.DiagnosticCategory.Error, "_0_is_declared_but_its_value_is_never_read_6133", "'{0}' is declared but its value is never read.", /*reportsUnnecessary*/ true),
            Report_errors_on_unused_locals: diag(6134, ts.DiagnosticCategory.Message, "Report_errors_on_unused_locals_6134", "Report errors on unused locals."),
            Report_errors_on_unused_parameters: diag(6135, ts.DiagnosticCategory.Message, "Report_errors_on_unused_parameters_6135", "Report errors on unused parameters."),
            The_maximum_dependency_depth_to_search_under_node_modules_and_load_JavaScript_files: diag(6136, ts.DiagnosticCategory.Message, "The_maximum_dependency_depth_to_search_under_node_modules_and_load_JavaScript_files_6136", "The maximum dependency depth to search under node_modules and load JavaScript files."),
            Cannot_import_type_declaration_files_Consider_importing_0_instead_of_1: diag(6137, ts.DiagnosticCategory.Error, "Cannot_import_type_declaration_files_Consider_importing_0_instead_of_1_6137", "Cannot import type declaration files. Consider importing '{0}' instead of '{1}'."),
            Property_0_is_declared_but_its_value_is_never_read: diag(6138, ts.DiagnosticCategory.Error, "Property_0_is_declared_but_its_value_is_never_read_6138", "Property '{0}' is declared but its value is never read.", /*reportsUnnecessary*/ true),
            Import_emit_helpers_from_tslib: diag(6139, ts.DiagnosticCategory.Message, "Import_emit_helpers_from_tslib_6139", "Import emit helpers from 'tslib'."),
            Auto_discovery_for_typings_is_enabled_in_project_0_Running_extra_resolution_pass_for_module_1_using_cache_location_2: diag(6140, ts.DiagnosticCategory.Error, "Auto_discovery_for_typings_is_enabled_in_project_0_Running_extra_resolution_pass_for_module_1_using__6140", "Auto discovery for typings is enabled in project '{0}'. Running extra resolution pass for module '{1}' using cache location '{2}'."),
            Parse_in_strict_mode_and_emit_use_strict_for_each_source_file: diag(6141, ts.DiagnosticCategory.Message, "Parse_in_strict_mode_and_emit_use_strict_for_each_source_file_6141", "Parse in strict mode and emit \"use strict\" for each source file."),
            Module_0_was_resolved_to_1_but_jsx_is_not_set: diag(6142, ts.DiagnosticCategory.Error, "Module_0_was_resolved_to_1_but_jsx_is_not_set_6142", "Module '{0}' was resolved to '{1}', but '--jsx' is not set."),
            Module_0_was_resolved_as_locally_declared_ambient_module_in_file_1: diag(6144, ts.DiagnosticCategory.Message, "Module_0_was_resolved_as_locally_declared_ambient_module_in_file_1_6144", "Module '{0}' was resolved as locally declared ambient module in file '{1}'."),
            Module_0_was_resolved_as_ambient_module_declared_in_1_since_this_file_was_not_modified: diag(6145, ts.DiagnosticCategory.Message, "Module_0_was_resolved_as_ambient_module_declared_in_1_since_this_file_was_not_modified_6145", "Module '{0}' was resolved as ambient module declared in '{1}' since this file was not modified."),
            Specify_the_JSX_factory_function_to_use_when_targeting_react_JSX_emit_e_g_React_createElement_or_h: diag(6146, ts.DiagnosticCategory.Message, "Specify_the_JSX_factory_function_to_use_when_targeting_react_JSX_emit_e_g_React_createElement_or_h_6146", "Specify the JSX factory function to use when targeting 'react' JSX emit, e.g. 'React.createElement' or 'h'."),
            Resolution_for_module_0_was_found_in_cache_from_location_1: diag(6147, ts.DiagnosticCategory.Message, "Resolution_for_module_0_was_found_in_cache_from_location_1_6147", "Resolution for module '{0}' was found in cache from location '{1}'."),
            Directory_0_does_not_exist_skipping_all_lookups_in_it: diag(6148, ts.DiagnosticCategory.Message, "Directory_0_does_not_exist_skipping_all_lookups_in_it_6148", "Directory '{0}' does not exist, skipping all lookups in it."),
            Show_diagnostic_information: diag(6149, ts.DiagnosticCategory.Message, "Show_diagnostic_information_6149", "Show diagnostic information."),
            Show_verbose_diagnostic_information: diag(6150, ts.DiagnosticCategory.Message, "Show_verbose_diagnostic_information_6150", "Show verbose diagnostic information."),
            Emit_a_single_file_with_source_maps_instead_of_having_a_separate_file: diag(6151, ts.DiagnosticCategory.Message, "Emit_a_single_file_with_source_maps_instead_of_having_a_separate_file_6151", "Emit a single file with source maps instead of having a separate file."),
            Emit_the_source_alongside_the_sourcemaps_within_a_single_file_requires_inlineSourceMap_or_sourceMap_to_be_set: diag(6152, ts.DiagnosticCategory.Message, "Emit_the_source_alongside_the_sourcemaps_within_a_single_file_requires_inlineSourceMap_or_sourceMap__6152", "Emit the source alongside the sourcemaps within a single file; requires '--inlineSourceMap' or '--sourceMap' to be set."),
            Transpile_each_file_as_a_separate_module_similar_to_ts_transpileModule: diag(6153, ts.DiagnosticCategory.Message, "Transpile_each_file_as_a_separate_module_similar_to_ts_transpileModule_6153", "Transpile each file as a separate module (similar to 'ts.transpileModule')."),
            Print_names_of_generated_files_part_of_the_compilation: diag(6154, ts.DiagnosticCategory.Message, "Print_names_of_generated_files_part_of_the_compilation_6154", "Print names of generated files part of the compilation."),
            Print_names_of_files_part_of_the_compilation: diag(6155, ts.DiagnosticCategory.Message, "Print_names_of_files_part_of_the_compilation_6155", "Print names of files part of the compilation."),
            The_locale_used_when_displaying_messages_to_the_user_e_g_en_us: diag(6156, ts.DiagnosticCategory.Message, "The_locale_used_when_displaying_messages_to_the_user_e_g_en_us_6156", "The locale used when displaying messages to the user (e.g. 'en-us')"),
            Do_not_generate_custom_helper_functions_like_extends_in_compiled_output: diag(6157, ts.DiagnosticCategory.Message, "Do_not_generate_custom_helper_functions_like_extends_in_compiled_output_6157", "Do not generate custom helper functions like '__extends' in compiled output."),
            Do_not_include_the_default_library_file_lib_d_ts: diag(6158, ts.DiagnosticCategory.Message, "Do_not_include_the_default_library_file_lib_d_ts_6158", "Do not include the default library file (lib.d.ts)."),
            Do_not_add_triple_slash_references_or_imported_modules_to_the_list_of_compiled_files: diag(6159, ts.DiagnosticCategory.Message, "Do_not_add_triple_slash_references_or_imported_modules_to_the_list_of_compiled_files_6159", "Do not add triple-slash references or imported modules to the list of compiled files."),
            Deprecated_Use_skipLibCheck_instead_Skip_type_checking_of_default_library_declaration_files: diag(6160, ts.DiagnosticCategory.Message, "Deprecated_Use_skipLibCheck_instead_Skip_type_checking_of_default_library_declaration_files_6160", "[Deprecated] Use '--skipLibCheck' instead. Skip type checking of default library declaration files."),
            List_of_folders_to_include_type_definitions_from: diag(6161, ts.DiagnosticCategory.Message, "List_of_folders_to_include_type_definitions_from_6161", "List of folders to include type definitions from."),
            Disable_size_limitations_on_JavaScript_projects: diag(6162, ts.DiagnosticCategory.Message, "Disable_size_limitations_on_JavaScript_projects_6162", "Disable size limitations on JavaScript projects."),
            The_character_set_of_the_input_files: diag(6163, ts.DiagnosticCategory.Message, "The_character_set_of_the_input_files_6163", "The character set of the input files."),
            Emit_a_UTF_8_Byte_Order_Mark_BOM_in_the_beginning_of_output_files: diag(6622, ts.DiagnosticCategory.Message, "Emit_a_UTF_8_Byte_Order_Mark_BOM_in_the_beginning_of_output_files_6622", "Emit a UTF-8 Byte Order Mark (BOM) in the beginning of output files."),
            Do_not_truncate_error_messages: diag(6165, ts.DiagnosticCategory.Message, "Do_not_truncate_error_messages_6165", "Do not truncate error messages."),
            Output_directory_for_generated_declaration_files: diag(6166, ts.DiagnosticCategory.Message, "Output_directory_for_generated_declaration_files_6166", "Output directory for generated declaration files."),
            A_series_of_entries_which_re_map_imports_to_lookup_locations_relative_to_the_baseUrl: diag(6167, ts.DiagnosticCategory.Message, "A_series_of_entries_which_re_map_imports_to_lookup_locations_relative_to_the_baseUrl_6167", "A series of entries which re-map imports to lookup locations relative to the 'baseUrl'."),
            List_of_root_folders_whose_combined_content_represents_the_structure_of_the_project_at_runtime: diag(6168, ts.DiagnosticCategory.Message, "List_of_root_folders_whose_combined_content_represents_the_structure_of_the_project_at_runtime_6168", "List of root folders whose combined content represents the structure of the project at runtime."),
            Show_all_compiler_options: diag(6169, ts.DiagnosticCategory.Message, "Show_all_compiler_options_6169", "Show all compiler options."),
            Deprecated_Use_outFile_instead_Concatenate_and_emit_output_to_single_file: diag(6170, ts.DiagnosticCategory.Message, "Deprecated_Use_outFile_instead_Concatenate_and_emit_output_to_single_file_6170", "[Deprecated] Use '--outFile' instead. Concatenate and emit output to single file"),
            Command_line_Options: diag(6171, ts.DiagnosticCategory.Message, "Command_line_Options_6171", "Command-line Options"),
            Provide_full_support_for_iterables_in_for_of_spread_and_destructuring_when_targeting_ES5_or_ES3: diag(6179, ts.DiagnosticCategory.Message, "Provide_full_support_for_iterables_in_for_of_spread_and_destructuring_when_targeting_ES5_or_ES3_6179", "Provide full support for iterables in 'for-of', spread, and destructuring when targeting 'ES5' or 'ES3'."),
            Enable_all_strict_type_checking_options: diag(6180, ts.DiagnosticCategory.Message, "Enable_all_strict_type_checking_options_6180", "Enable all strict type-checking options."),
            List_of_language_service_plugins: diag(6181, ts.DiagnosticCategory.Message, "List_of_language_service_plugins_6181", "List of language service plugins."),
            Scoped_package_detected_looking_in_0: diag(6182, ts.DiagnosticCategory.Message, "Scoped_package_detected_looking_in_0_6182", "Scoped package detected, looking in '{0}'"),
            Reusing_resolution_of_module_0_from_1_of_old_program_it_was_successfully_resolved_to_2: diag(6183, ts.DiagnosticCategory.Message, "Reusing_resolution_of_module_0_from_1_of_old_program_it_was_successfully_resolved_to_2_6183", "Reusing resolution of module '{0}' from '{1}' of old program, it was successfully resolved to '{2}'."),
            Reusing_resolution_of_module_0_from_1_of_old_program_it_was_successfully_resolved_to_2_with_Package_ID_3: diag(6184, ts.DiagnosticCategory.Message, "Reusing_resolution_of_module_0_from_1_of_old_program_it_was_successfully_resolved_to_2_with_Package__6184", "Reusing resolution of module '{0}' from '{1}' of old program, it was successfully resolved to '{2}' with Package ID '{3}'."),
            Enable_strict_checking_of_function_types: diag(6186, ts.DiagnosticCategory.Message, "Enable_strict_checking_of_function_types_6186", "Enable strict checking of function types."),
            Enable_strict_checking_of_property_initialization_in_classes: diag(6187, ts.DiagnosticCategory.Message, "Enable_strict_checking_of_property_initialization_in_classes_6187", "Enable strict checking of property initialization in classes."),
            Numeric_separators_are_not_allowed_here: diag(6188, ts.DiagnosticCategory.Error, "Numeric_separators_are_not_allowed_here_6188", "Numeric separators are not allowed here."),
            Multiple_consecutive_numeric_separators_are_not_permitted: diag(6189, ts.DiagnosticCategory.Error, "Multiple_consecutive_numeric_separators_are_not_permitted_6189", "Multiple consecutive numeric separators are not permitted."),
            Whether_to_keep_outdated_console_output_in_watch_mode_instead_of_clearing_the_screen: diag(6191, ts.DiagnosticCategory.Message, "Whether_to_keep_outdated_console_output_in_watch_mode_instead_of_clearing_the_screen_6191", "Whether to keep outdated console output in watch mode instead of clearing the screen."),
            All_imports_in_import_declaration_are_unused: diag(6192, ts.DiagnosticCategory.Error, "All_imports_in_import_declaration_are_unused_6192", "All imports in import declaration are unused.", /*reportsUnnecessary*/ true),
            Found_1_error_Watching_for_file_changes: diag(6193, ts.DiagnosticCategory.Message, "Found_1_error_Watching_for_file_changes_6193", "Found 1 error. Watching for file changes."),
            Found_0_errors_Watching_for_file_changes: diag(6194, ts.DiagnosticCategory.Message, "Found_0_errors_Watching_for_file_changes_6194", "Found {0} errors. Watching for file changes."),
            Resolve_keyof_to_string_valued_property_names_only_no_numbers_or_symbols: diag(6195, ts.DiagnosticCategory.Message, "Resolve_keyof_to_string_valued_property_names_only_no_numbers_or_symbols_6195", "Resolve 'keyof' to string valued property names only (no numbers or symbols)."),
            _0_is_declared_but_never_used: diag(6196, ts.DiagnosticCategory.Error, "_0_is_declared_but_never_used_6196", "'{0}' is declared but never used.", /*reportsUnnecessary*/ true),
            Include_modules_imported_with_json_extension: diag(6197, ts.DiagnosticCategory.Message, "Include_modules_imported_with_json_extension_6197", "Include modules imported with '.json' extension"),
            All_destructured_elements_are_unused: diag(6198, ts.DiagnosticCategory.Error, "All_destructured_elements_are_unused_6198", "All destructured elements are unused.", /*reportsUnnecessary*/ true),
            All_variables_are_unused: diag(6199, ts.DiagnosticCategory.Error, "All_variables_are_unused_6199", "All variables are unused.", /*reportsUnnecessary*/ true),
            Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0: diag(6200, ts.DiagnosticCategory.Error, "Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0_6200", "Definitions of the following identifiers conflict with those in another file: {0}"),
            Conflicts_are_in_this_file: diag(6201, ts.DiagnosticCategory.Message, "Conflicts_are_in_this_file_6201", "Conflicts are in this file."),
            Project_references_may_not_form_a_circular_graph_Cycle_detected_Colon_0: diag(6202, ts.DiagnosticCategory.Error, "Project_references_may_not_form_a_circular_graph_Cycle_detected_Colon_0_6202", "Project references may not form a circular graph. Cycle detected: {0}"),
            _0_was_also_declared_here: diag(6203, ts.DiagnosticCategory.Message, "_0_was_also_declared_here_6203", "'{0}' was also declared here."),
            and_here: diag(6204, ts.DiagnosticCategory.Message, "and_here_6204", "and here."),
            All_type_parameters_are_unused: diag(6205, ts.DiagnosticCategory.Error, "All_type_parameters_are_unused_6205", "All type parameters are unused."),
            package_json_has_a_typesVersions_field_with_version_specific_path_mappings: diag(6206, ts.DiagnosticCategory.Message, "package_json_has_a_typesVersions_field_with_version_specific_path_mappings_6206", "'package.json' has a 'typesVersions' field with version-specific path mappings."),
            package_json_does_not_have_a_typesVersions_entry_that_matches_version_0: diag(6207, ts.DiagnosticCategory.Message, "package_json_does_not_have_a_typesVersions_entry_that_matches_version_0_6207", "'package.json' does not have a 'typesVersions' entry that matches version '{0}'."),
            package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_match_module_name_2: diag(6208, ts.DiagnosticCategory.Message, "package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_ma_6208", "'package.json' has a 'typesVersions' entry '{0}' that matches compiler version '{1}', looking for a pattern to match module name '{2}'."),
            package_json_has_a_typesVersions_entry_0_that_is_not_a_valid_semver_range: diag(6209, ts.DiagnosticCategory.Message, "package_json_has_a_typesVersions_entry_0_that_is_not_a_valid_semver_range_6209", "'package.json' has a 'typesVersions' entry '{0}' that is not a valid semver range."),
            An_argument_for_0_was_not_provided: diag(6210, ts.DiagnosticCategory.Message, "An_argument_for_0_was_not_provided_6210", "An argument for '{0}' was not provided."),
            An_argument_matching_this_binding_pattern_was_not_provided: diag(6211, ts.DiagnosticCategory.Message, "An_argument_matching_this_binding_pattern_was_not_provided_6211", "An argument matching this binding pattern was not provided."),
            Did_you_mean_to_call_this_expression: diag(6212, ts.DiagnosticCategory.Message, "Did_you_mean_to_call_this_expression_6212", "Did you mean to call this expression?"),
            Did_you_mean_to_use_new_with_this_expression: diag(6213, ts.DiagnosticCategory.Message, "Did_you_mean_to_use_new_with_this_expression_6213", "Did you mean to use 'new' with this expression?"),
            Enable_strict_bind_call_and_apply_methods_on_functions: diag(6214, ts.DiagnosticCategory.Message, "Enable_strict_bind_call_and_apply_methods_on_functions_6214", "Enable strict 'bind', 'call', and 'apply' methods on functions."),
            Using_compiler_options_of_project_reference_redirect_0: diag(6215, ts.DiagnosticCategory.Message, "Using_compiler_options_of_project_reference_redirect_0_6215", "Using compiler options of project reference redirect '{0}'."),
            Found_1_error: diag(6216, ts.DiagnosticCategory.Message, "Found_1_error_6216", "Found 1 error."),
            Found_0_errors: diag(6217, ts.DiagnosticCategory.Message, "Found_0_errors_6217", "Found {0} errors."),
            Module_name_0_was_successfully_resolved_to_1_with_Package_ID_2: diag(6218, ts.DiagnosticCategory.Message, "Module_name_0_was_successfully_resolved_to_1_with_Package_ID_2_6218", "======== Module name '{0}' was successfully resolved to '{1}' with Package ID '{2}'. ========"),
            Type_reference_directive_0_was_successfully_resolved_to_1_with_Package_ID_2_primary_Colon_3: diag(6219, ts.DiagnosticCategory.Message, "Type_reference_directive_0_was_successfully_resolved_to_1_with_Package_ID_2_primary_Colon_3_6219", "======== Type reference directive '{0}' was successfully resolved to '{1}' with Package ID '{2}', primary: {3}. ========"),
            package_json_had_a_falsy_0_field: diag(6220, ts.DiagnosticCategory.Message, "package_json_had_a_falsy_0_field_6220", "'package.json' had a falsy '{0}' field."),
            Disable_use_of_source_files_instead_of_declaration_files_from_referenced_projects: diag(6221, ts.DiagnosticCategory.Message, "Disable_use_of_source_files_instead_of_declaration_files_from_referenced_projects_6221", "Disable use of source files instead of declaration files from referenced projects."),
            Emit_class_fields_with_Define_instead_of_Set: diag(6222, ts.DiagnosticCategory.Message, "Emit_class_fields_with_Define_instead_of_Set_6222", "Emit class fields with Define instead of Set."),
            Generates_a_CPU_profile: diag(6223, ts.DiagnosticCategory.Message, "Generates_a_CPU_profile_6223", "Generates a CPU profile."),
            Disable_solution_searching_for_this_project: diag(6224, ts.DiagnosticCategory.Message, "Disable_solution_searching_for_this_project_6224", "Disable solution searching for this project."),
            Specify_strategy_for_watching_file_Colon_FixedPollingInterval_default_PriorityPollingInterval_DynamicPriorityPolling_FixedChunkSizePolling_UseFsEvents_UseFsEventsOnParentDirectory: diag(6225, ts.DiagnosticCategory.Message, "Specify_strategy_for_watching_file_Colon_FixedPollingInterval_default_PriorityPollingInterval_Dynami_6225", "Specify strategy for watching file: 'FixedPollingInterval' (default), 'PriorityPollingInterval', 'DynamicPriorityPolling', 'FixedChunkSizePolling', 'UseFsEvents', 'UseFsEventsOnParentDirectory'."),
            Specify_strategy_for_watching_directory_on_platforms_that_don_t_support_recursive_watching_natively_Colon_UseFsEvents_default_FixedPollingInterval_DynamicPriorityPolling_FixedChunkSizePolling: diag(6226, ts.DiagnosticCategory.Message, "Specify_strategy_for_watching_directory_on_platforms_that_don_t_support_recursive_watching_natively__6226", "Specify strategy for watching directory on platforms that don't support recursive watching natively: 'UseFsEvents' (default), 'FixedPollingInterval', 'DynamicPriorityPolling', 'FixedChunkSizePolling'."),
            Specify_strategy_for_creating_a_polling_watch_when_it_fails_to_create_using_file_system_events_Colon_FixedInterval_default_PriorityInterval_DynamicPriority_FixedChunkSize: diag(6227, ts.DiagnosticCategory.Message, "Specify_strategy_for_creating_a_polling_watch_when_it_fails_to_create_using_file_system_events_Colon_6227", "Specify strategy for creating a polling watch when it fails to create using file system events: 'FixedInterval' (default), 'PriorityInterval', 'DynamicPriority', 'FixedChunkSize'."),
            Tag_0_expects_at_least_1_arguments_but_the_JSX_factory_2_provides_at_most_3: diag(6229, ts.DiagnosticCategory.Error, "Tag_0_expects_at_least_1_arguments_but_the_JSX_factory_2_provides_at_most_3_6229", "Tag '{0}' expects at least '{1}' arguments, but the JSX factory '{2}' provides at most '{3}'."),
            Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_false_or_null_on_command_line: diag(6230, ts.DiagnosticCategory.Error, "Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_false_or_null_on_command_line_6230", "Option '{0}' can only be specified in 'tsconfig.json' file or set to 'false' or 'null' on command line."),
            Could_not_resolve_the_path_0_with_the_extensions_Colon_1: diag(6231, ts.DiagnosticCategory.Error, "Could_not_resolve_the_path_0_with_the_extensions_Colon_1_6231", "Could not resolve the path '{0}' with the extensions: {1}."),
            Declaration_augments_declaration_in_another_file_This_cannot_be_serialized: diag(6232, ts.DiagnosticCategory.Error, "Declaration_augments_declaration_in_another_file_This_cannot_be_serialized_6232", "Declaration augments declaration in another file. This cannot be serialized."),
            This_is_the_declaration_being_augmented_Consider_moving_the_augmenting_declaration_into_the_same_file: diag(6233, ts.DiagnosticCategory.Error, "This_is_the_declaration_being_augmented_Consider_moving_the_augmenting_declaration_into_the_same_fil_6233", "This is the declaration being augmented. Consider moving the augmenting declaration into the same file."),
            This_expression_is_not_callable_because_it_is_a_get_accessor_Did_you_mean_to_use_it_without: diag(6234, ts.DiagnosticCategory.Error, "This_expression_is_not_callable_because_it_is_a_get_accessor_Did_you_mean_to_use_it_without_6234", "This expression is not callable because it is a 'get' accessor. Did you mean to use it without '()'?"),
            Disable_loading_referenced_projects: diag(6235, ts.DiagnosticCategory.Message, "Disable_loading_referenced_projects_6235", "Disable loading referenced projects."),
            Arguments_for_the_rest_parameter_0_were_not_provided: diag(6236, ts.DiagnosticCategory.Error, "Arguments_for_the_rest_parameter_0_were_not_provided_6236", "Arguments for the rest parameter '{0}' were not provided."),
            Generates_an_event_trace_and_a_list_of_types: diag(6237, ts.DiagnosticCategory.Message, "Generates_an_event_trace_and_a_list_of_types_6237", "Generates an event trace and a list of types."),
            Specify_the_module_specifier_to_be_used_to_import_the_jsx_and_jsxs_factory_functions_from_eg_react: diag(6238, ts.DiagnosticCategory.Error, "Specify_the_module_specifier_to_be_used_to_import_the_jsx_and_jsxs_factory_functions_from_eg_react_6238", "Specify the module specifier to be used to import the 'jsx' and 'jsxs' factory functions from. eg, react"),
            File_0_exists_according_to_earlier_cached_lookups: diag(6239, ts.DiagnosticCategory.Message, "File_0_exists_according_to_earlier_cached_lookups_6239", "File '{0}' exists according to earlier cached lookups."),
            File_0_does_not_exist_according_to_earlier_cached_lookups: diag(6240, ts.DiagnosticCategory.Message, "File_0_does_not_exist_according_to_earlier_cached_lookups_6240", "File '{0}' does not exist according to earlier cached lookups."),
            Resolution_for_type_reference_directive_0_was_found_in_cache_from_location_1: diag(6241, ts.DiagnosticCategory.Message, "Resolution_for_type_reference_directive_0_was_found_in_cache_from_location_1_6241", "Resolution for type reference directive '{0}' was found in cache from location '{1}'."),
            Resolving_type_reference_directive_0_containing_file_1: diag(6242, ts.DiagnosticCategory.Message, "Resolving_type_reference_directive_0_containing_file_1_6242", "======== Resolving type reference directive '{0}', containing file '{1}'. ========"),
            Interpret_optional_property_types_as_written_rather_than_adding_undefined: diag(6243, ts.DiagnosticCategory.Message, "Interpret_optional_property_types_as_written_rather_than_adding_undefined_6243", "Interpret optional property types as written, rather than adding 'undefined'."),
            Modules: diag(6244, ts.DiagnosticCategory.Message, "Modules_6244", "Modules"),
            File_Management: diag(6245, ts.DiagnosticCategory.Message, "File_Management_6245", "File Management"),
            Emit: diag(6246, ts.DiagnosticCategory.Message, "Emit_6246", "Emit"),
            JavaScript_Support: diag(6247, ts.DiagnosticCategory.Message, "JavaScript_Support_6247", "JavaScript Support"),
            Type_Checking: diag(6248, ts.DiagnosticCategory.Message, "Type_Checking_6248", "Type Checking"),
            Editor_Support: diag(6249, ts.DiagnosticCategory.Message, "Editor_Support_6249", "Editor Support"),
            Watch_and_Build_Modes: diag(6250, ts.DiagnosticCategory.Message, "Watch_and_Build_Modes_6250", "Watch and Build Modes"),
            Compiler_Diagnostics: diag(6251, ts.DiagnosticCategory.Message, "Compiler_Diagnostics_6251", "Compiler Diagnostics"),
            Interop_Constraints: diag(6252, ts.DiagnosticCategory.Message, "Interop_Constraints_6252", "Interop Constraints"),
            Backwards_Compatibility: diag(6253, ts.DiagnosticCategory.Message, "Backwards_Compatibility_6253", "Backwards Compatibility"),
            Language_and_Environment: diag(6254, ts.DiagnosticCategory.Message, "Language_and_Environment_6254", "Language and Environment"),
            Projects: diag(6255, ts.DiagnosticCategory.Message, "Projects_6255", "Projects"),
            Output_Formatting: diag(6256, ts.DiagnosticCategory.Message, "Output_Formatting_6256", "Output Formatting"),
            Completeness: diag(6257, ts.DiagnosticCategory.Message, "Completeness_6257", "Completeness"),
            Projects_to_reference: diag(6300, ts.DiagnosticCategory.Message, "Projects_to_reference_6300", "Projects to reference"),
            Enable_project_compilation: diag(6302, ts.DiagnosticCategory.Message, "Enable_project_compilation_6302", "Enable project compilation"),
            Composite_projects_may_not_disable_declaration_emit: diag(6304, ts.DiagnosticCategory.Error, "Composite_projects_may_not_disable_declaration_emit_6304", "Composite projects may not disable declaration emit."),
            Output_file_0_has_not_been_built_from_source_file_1: diag(6305, ts.DiagnosticCategory.Error, "Output_file_0_has_not_been_built_from_source_file_1_6305", "Output file '{0}' has not been built from source file '{1}'."),
            Referenced_project_0_must_have_setting_composite_Colon_true: diag(6306, ts.DiagnosticCategory.Error, "Referenced_project_0_must_have_setting_composite_Colon_true_6306", "Referenced project '{0}' must have setting \"composite\": true."),
            File_0_is_not_listed_within_the_file_list_of_project_1_Projects_must_list_all_files_or_use_an_include_pattern: diag(6307, ts.DiagnosticCategory.Error, "File_0_is_not_listed_within_the_file_list_of_project_1_Projects_must_list_all_files_or_use_an_includ_6307", "File '{0}' is not listed within the file list of project '{1}'. Projects must list all files or use an 'include' pattern."),
            Cannot_prepend_project_0_because_it_does_not_have_outFile_set: diag(6308, ts.DiagnosticCategory.Error, "Cannot_prepend_project_0_because_it_does_not_have_outFile_set_6308", "Cannot prepend project '{0}' because it does not have 'outFile' set"),
            Output_file_0_from_project_1_does_not_exist: diag(6309, ts.DiagnosticCategory.Error, "Output_file_0_from_project_1_does_not_exist_6309", "Output file '{0}' from project '{1}' does not exist"),
            Referenced_project_0_may_not_disable_emit: diag(6310, ts.DiagnosticCategory.Error, "Referenced_project_0_may_not_disable_emit_6310", "Referenced project '{0}' may not disable emit."),
            Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2: diag(6350, ts.DiagnosticCategory.Message, "Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2_6350", "Project '{0}' is out of date because oldest output '{1}' is older than newest input '{2}'"),
            Project_0_is_up_to_date_because_newest_input_1_is_older_than_oldest_output_2: diag(6351, ts.DiagnosticCategory.Message, "Project_0_is_up_to_date_because_newest_input_1_is_older_than_oldest_output_2_6351", "Project '{0}' is up to date because newest input '{1}' is older than oldest output '{2}'"),
            Project_0_is_out_of_date_because_output_file_1_does_not_exist: diag(6352, ts.DiagnosticCategory.Message, "Project_0_is_out_of_date_because_output_file_1_does_not_exist_6352", "Project '{0}' is out of date because output file '{1}' does not exist"),
            Project_0_is_out_of_date_because_its_dependency_1_is_out_of_date: diag(6353, ts.DiagnosticCategory.Message, "Project_0_is_out_of_date_because_its_dependency_1_is_out_of_date_6353", "Project '{0}' is out of date because its dependency '{1}' is out of date"),
            Project_0_is_up_to_date_with_d_ts_files_from_its_dependencies: diag(6354, ts.DiagnosticCategory.Message, "Project_0_is_up_to_date_with_d_ts_files_from_its_dependencies_6354", "Project '{0}' is up to date with .d.ts files from its dependencies"),
            Projects_in_this_build_Colon_0: diag(6355, ts.DiagnosticCategory.Message, "Projects_in_this_build_Colon_0_6355", "Projects in this build: {0}"),
            A_non_dry_build_would_delete_the_following_files_Colon_0: diag(6356, ts.DiagnosticCategory.Message, "A_non_dry_build_would_delete_the_following_files_Colon_0_6356", "A non-dry build would delete the following files: {0}"),
            A_non_dry_build_would_build_project_0: diag(6357, ts.DiagnosticCategory.Message, "A_non_dry_build_would_build_project_0_6357", "A non-dry build would build project '{0}'"),
            Building_project_0: diag(6358, ts.DiagnosticCategory.Message, "Building_project_0_6358", "Building project '{0}'..."),
            Updating_output_timestamps_of_project_0: diag(6359, ts.DiagnosticCategory.Message, "Updating_output_timestamps_of_project_0_6359", "Updating output timestamps of project '{0}'..."),
            delete_this_Project_0_is_up_to_date_because_it_was_previously_built: diag(6360, ts.DiagnosticCategory.Message, "delete_this_Project_0_is_up_to_date_because_it_was_previously_built_6360", "delete this - Project '{0}' is up to date because it was previously built"),
            Project_0_is_up_to_date: diag(6361, ts.DiagnosticCategory.Message, "Project_0_is_up_to_date_6361", "Project '{0}' is up to date"),
            Skipping_build_of_project_0_because_its_dependency_1_has_errors: diag(6362, ts.DiagnosticCategory.Message, "Skipping_build_of_project_0_because_its_dependency_1_has_errors_6362", "Skipping build of project '{0}' because its dependency '{1}' has errors"),
            Project_0_can_t_be_built_because_its_dependency_1_has_errors: diag(6363, ts.DiagnosticCategory.Message, "Project_0_can_t_be_built_because_its_dependency_1_has_errors_6363", "Project '{0}' can't be built because its dependency '{1}' has errors"),
            Build_one_or_more_projects_and_their_dependencies_if_out_of_date: diag(6364, ts.DiagnosticCategory.Message, "Build_one_or_more_projects_and_their_dependencies_if_out_of_date_6364", "Build one or more projects and their dependencies, if out of date"),
            Delete_the_outputs_of_all_projects: diag(6365, ts.DiagnosticCategory.Message, "Delete_the_outputs_of_all_projects_6365", "Delete the outputs of all projects"),
            Show_what_would_be_built_or_deleted_if_specified_with_clean: diag(6367, ts.DiagnosticCategory.Message, "Show_what_would_be_built_or_deleted_if_specified_with_clean_6367", "Show what would be built (or deleted, if specified with '--clean')"),
            Option_build_must_be_the_first_command_line_argument: diag(6369, ts.DiagnosticCategory.Error, "Option_build_must_be_the_first_command_line_argument_6369", "Option '--build' must be the first command line argument."),
            Options_0_and_1_cannot_be_combined: diag(6370, ts.DiagnosticCategory.Error, "Options_0_and_1_cannot_be_combined_6370", "Options '{0}' and '{1}' cannot be combined."),
            Updating_unchanged_output_timestamps_of_project_0: diag(6371, ts.DiagnosticCategory.Message, "Updating_unchanged_output_timestamps_of_project_0_6371", "Updating unchanged output timestamps of project '{0}'..."),
            Project_0_is_out_of_date_because_output_of_its_dependency_1_has_changed: diag(6372, ts.DiagnosticCategory.Message, "Project_0_is_out_of_date_because_output_of_its_dependency_1_has_changed_6372", "Project '{0}' is out of date because output of its dependency '{1}' has changed"),
            Updating_output_of_project_0: diag(6373, ts.DiagnosticCategory.Message, "Updating_output_of_project_0_6373", "Updating output of project '{0}'..."),
            A_non_dry_build_would_update_timestamps_for_output_of_project_0: diag(6374, ts.DiagnosticCategory.Message, "A_non_dry_build_would_update_timestamps_for_output_of_project_0_6374", "A non-dry build would update timestamps for output of project '{0}'"),
            A_non_dry_build_would_update_output_of_project_0: diag(6375, ts.DiagnosticCategory.Message, "A_non_dry_build_would_update_output_of_project_0_6375", "A non-dry build would update output of project '{0}'"),
            Cannot_update_output_of_project_0_because_there_was_error_reading_file_1: diag(6376, ts.DiagnosticCategory.Message, "Cannot_update_output_of_project_0_because_there_was_error_reading_file_1_6376", "Cannot update output of project '{0}' because there was error reading file '{1}'"),
            Cannot_write_file_0_because_it_will_overwrite_tsbuildinfo_file_generated_by_referenced_project_1: diag(6377, ts.DiagnosticCategory.Error, "Cannot_write_file_0_because_it_will_overwrite_tsbuildinfo_file_generated_by_referenced_project_1_6377", "Cannot write file '{0}' because it will overwrite '.tsbuildinfo' file generated by referenced project '{1}'"),
            Enable_incremental_compilation: diag(6378, ts.DiagnosticCategory.Message, "Enable_incremental_compilation_6378", "Enable incremental compilation"),
            Composite_projects_may_not_disable_incremental_compilation: diag(6379, ts.DiagnosticCategory.Error, "Composite_projects_may_not_disable_incremental_compilation_6379", "Composite projects may not disable incremental compilation."),
            Specify_file_to_store_incremental_compilation_information: diag(6380, ts.DiagnosticCategory.Message, "Specify_file_to_store_incremental_compilation_information_6380", "Specify file to store incremental compilation information"),
            Project_0_is_out_of_date_because_output_for_it_was_generated_with_version_1_that_differs_with_current_version_2: diag(6381, ts.DiagnosticCategory.Message, "Project_0_is_out_of_date_because_output_for_it_was_generated_with_version_1_that_differs_with_curren_6381", "Project '{0}' is out of date because output for it was generated with version '{1}' that differs with current version '{2}'"),
            Skipping_build_of_project_0_because_its_dependency_1_was_not_built: diag(6382, ts.DiagnosticCategory.Message, "Skipping_build_of_project_0_because_its_dependency_1_was_not_built_6382", "Skipping build of project '{0}' because its dependency '{1}' was not built"),
            Project_0_can_t_be_built_because_its_dependency_1_was_not_built: diag(6383, ts.DiagnosticCategory.Message, "Project_0_can_t_be_built_because_its_dependency_1_was_not_built_6383", "Project '{0}' can't be built because its dependency '{1}' was not built"),
            Have_recompiles_in_incremental_and_watch_assume_that_changes_within_a_file_will_only_affect_files_directly_depending_on_it: diag(6384, ts.DiagnosticCategory.Message, "Have_recompiles_in_incremental_and_watch_assume_that_changes_within_a_file_will_only_affect_files_di_6384", "Have recompiles in '--incremental' and '--watch' assume that changes within a file will only affect files directly depending on it."),
            _0_is_deprecated: diag(6385, ts.DiagnosticCategory.Suggestion, "_0_is_deprecated_6385", "'{0}' is deprecated.", /*reportsUnnecessary*/ undefined, /*elidedInCompatabilityPyramid*/ undefined, /*reportsDeprecated*/ true),
            Performance_timings_for_diagnostics_or_extendedDiagnostics_are_not_available_in_this_session_A_native_implementation_of_the_Web_Performance_API_could_not_be_found: diag(6386, ts.DiagnosticCategory.Message, "Performance_timings_for_diagnostics_or_extendedDiagnostics_are_not_available_in_this_session_A_nativ_6386", "Performance timings for '--diagnostics' or '--extendedDiagnostics' are not available in this session. A native implementation of the Web Performance API could not be found."),
            The_signature_0_of_1_is_deprecated: diag(6387, ts.DiagnosticCategory.Suggestion, "The_signature_0_of_1_is_deprecated_6387", "The signature '{0}' of '{1}' is deprecated.", /*reportsUnnecessary*/ undefined, /*elidedInCompatabilityPyramid*/ undefined, /*reportsDeprecated*/ true),
            Project_0_is_being_forcibly_rebuilt: diag(6388, ts.DiagnosticCategory.Message, "Project_0_is_being_forcibly_rebuilt_6388", "Project '{0}' is being forcibly rebuilt"),
            Reusing_resolution_of_module_0_from_1_of_old_program_it_was_not_resolved: diag(6389, ts.DiagnosticCategory.Message, "Reusing_resolution_of_module_0_from_1_of_old_program_it_was_not_resolved_6389", "Reusing resolution of module '{0}' from '{1}' of old program, it was not resolved."),
            Reusing_resolution_of_type_reference_directive_0_from_1_of_old_program_it_was_successfully_resolved_to_2: diag(6390, ts.DiagnosticCategory.Message, "Reusing_resolution_of_type_reference_directive_0_from_1_of_old_program_it_was_successfully_resolved__6390", "Reusing resolution of type reference directive '{0}' from '{1}' of old program, it was successfully resolved to '{2}'."),
            Reusing_resolution_of_type_reference_directive_0_from_1_of_old_program_it_was_successfully_resolved_to_2_with_Package_ID_3: diag(6391, ts.DiagnosticCategory.Message, "Reusing_resolution_of_type_reference_directive_0_from_1_of_old_program_it_was_successfully_resolved__6391", "Reusing resolution of type reference directive '{0}' from '{1}' of old program, it was successfully resolved to '{2}' with Package ID '{3}'."),
            Reusing_resolution_of_type_reference_directive_0_from_1_of_old_program_it_was_not_resolved: diag(6392, ts.DiagnosticCategory.Message, "Reusing_resolution_of_type_reference_directive_0_from_1_of_old_program_it_was_not_resolved_6392", "Reusing resolution of type reference directive '{0}' from '{1}' of old program, it was not resolved."),
            Reusing_resolution_of_module_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_3: diag(6393, ts.DiagnosticCategory.Message, "Reusing_resolution_of_module_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_6393", "Reusing resolution of module '{0}' from '{1}' found in cache from location '{2}', it was successfully resolved to '{3}'."),
            Reusing_resolution_of_module_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_3_with_Package_ID_4: diag(6394, ts.DiagnosticCategory.Message, "Reusing_resolution_of_module_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_6394", "Reusing resolution of module '{0}' from '{1}' found in cache from location '{2}', it was successfully resolved to '{3}' with Package ID '{4}'."),
            Reusing_resolution_of_module_0_from_1_found_in_cache_from_location_2_it_was_not_resolved: diag(6395, ts.DiagnosticCategory.Message, "Reusing_resolution_of_module_0_from_1_found_in_cache_from_location_2_it_was_not_resolved_6395", "Reusing resolution of module '{0}' from '{1}' found in cache from location '{2}', it was not resolved."),
            Reusing_resolution_of_type_reference_directive_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_3: diag(6396, ts.DiagnosticCategory.Message, "Reusing_resolution_of_type_reference_directive_0_from_1_found_in_cache_from_location_2_it_was_succes_6396", "Reusing resolution of type reference directive '{0}' from '{1}' found in cache from location '{2}', it was successfully resolved to '{3}'."),
            Reusing_resolution_of_type_reference_directive_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_3_with_Package_ID_4: diag(6397, ts.DiagnosticCategory.Message, "Reusing_resolution_of_type_reference_directive_0_from_1_found_in_cache_from_location_2_it_was_succes_6397", "Reusing resolution of type reference directive '{0}' from '{1}' found in cache from location '{2}', it was successfully resolved to '{3}' with Package ID '{4}'."),
            Reusing_resolution_of_type_reference_directive_0_from_1_found_in_cache_from_location_2_it_was_not_resolved: diag(6398, ts.DiagnosticCategory.Message, "Reusing_resolution_of_type_reference_directive_0_from_1_found_in_cache_from_location_2_it_was_not_re_6398", "Reusing resolution of type reference directive '{0}' from '{1}' found in cache from location '{2}', it was not resolved."),
            The_expected_type_comes_from_property_0_which_is_declared_here_on_type_1: diag(6500, ts.DiagnosticCategory.Message, "The_expected_type_comes_from_property_0_which_is_declared_here_on_type_1_6500", "The expected type comes from property '{0}' which is declared here on type '{1}'"),
            The_expected_type_comes_from_this_index_signature: diag(6501, ts.DiagnosticCategory.Message, "The_expected_type_comes_from_this_index_signature_6501", "The expected type comes from this index signature."),
            The_expected_type_comes_from_the_return_type_of_this_signature: diag(6502, ts.DiagnosticCategory.Message, "The_expected_type_comes_from_the_return_type_of_this_signature_6502", "The expected type comes from the return type of this signature."),
            Print_names_of_files_that_are_part_of_the_compilation_and_then_stop_processing: diag(6503, ts.DiagnosticCategory.Message, "Print_names_of_files_that_are_part_of_the_compilation_and_then_stop_processing_6503", "Print names of files that are part of the compilation and then stop processing."),
            File_0_is_a_JavaScript_file_Did_you_mean_to_enable_the_allowJs_option: diag(6504, ts.DiagnosticCategory.Error, "File_0_is_a_JavaScript_file_Did_you_mean_to_enable_the_allowJs_option_6504", "File '{0}' is a JavaScript file. Did you mean to enable the 'allowJs' option?"),
            Print_names_of_files_and_the_reason_they_are_part_of_the_compilation: diag(6505, ts.DiagnosticCategory.Message, "Print_names_of_files_and_the_reason_they_are_part_of_the_compilation_6505", "Print names of files and the reason they are part of the compilation."),
            Consider_adding_a_declare_modifier_to_this_class: diag(6506, ts.DiagnosticCategory.Message, "Consider_adding_a_declare_modifier_to_this_class_6506", "Consider adding a 'declare' modifier to this class."),
            Allow_JavaScript_files_to_be_a_part_of_your_program_Use_the_checkJS_option_to_get_errors_from_these_files: diag(6600, ts.DiagnosticCategory.Message, "Allow_JavaScript_files_to_be_a_part_of_your_program_Use_the_checkJS_option_to_get_errors_from_these__6600", "Allow JavaScript files to be a part of your program. Use the `checkJS` option to get errors from these files."),
            Allow_import_x_from_y_when_a_module_doesn_t_have_a_default_export: diag(6601, ts.DiagnosticCategory.Message, "Allow_import_x_from_y_when_a_module_doesn_t_have_a_default_export_6601", "Allow 'import x from y' when a module doesn't have a default export."),
            Allow_accessing_UMD_globals_from_modules: diag(6602, ts.DiagnosticCategory.Message, "Allow_accessing_UMD_globals_from_modules_6602", "Allow accessing UMD globals from modules."),
            Disable_error_reporting_for_unreachable_code: diag(6603, ts.DiagnosticCategory.Message, "Disable_error_reporting_for_unreachable_code_6603", "Disable error reporting for unreachable code."),
            Disable_error_reporting_for_unused_labels: diag(6604, ts.DiagnosticCategory.Message, "Disable_error_reporting_for_unused_labels_6604", "Disable error reporting for unused labels."),
            Ensure_use_strict_is_always_emitted: diag(6605, ts.DiagnosticCategory.Message, "Ensure_use_strict_is_always_emitted_6605", "Ensure 'use strict' is always emitted."),
            Have_recompiles_in_projects_that_use_incremental_and_watch_mode_assume_that_changes_within_a_file_will_only_affect_files_directly_depending_on_it: diag(6606, ts.DiagnosticCategory.Message, "Have_recompiles_in_projects_that_use_incremental_and_watch_mode_assume_that_changes_within_a_file_wi_6606", "Have recompiles in projects that use `incremental` and `watch` mode assume that changes within a file will only affect files directly depending on it."),
            Specify_the_base_directory_to_resolve_non_relative_module_names: diag(6607, ts.DiagnosticCategory.Message, "Specify_the_base_directory_to_resolve_non_relative_module_names_6607", "Specify the base directory to resolve non-relative module names."),
            No_longer_supported_In_early_versions_manually_set_the_text_encoding_for_reading_files: diag(6608, ts.DiagnosticCategory.Message, "No_longer_supported_In_early_versions_manually_set_the_text_encoding_for_reading_files_6608", "No longer supported. In early versions, manually set the text encoding for reading files."),
            Enable_error_reporting_in_type_checked_JavaScript_files: diag(6609, ts.DiagnosticCategory.Message, "Enable_error_reporting_in_type_checked_JavaScript_files_6609", "Enable error reporting in type-checked JavaScript files."),
            Enable_constraints_that_allow_a_TypeScript_project_to_be_used_with_project_references: diag(6611, ts.DiagnosticCategory.Message, "Enable_constraints_that_allow_a_TypeScript_project_to_be_used_with_project_references_6611", "Enable constraints that allow a TypeScript project to be used with project references."),
            Generate_d_ts_files_from_TypeScript_and_JavaScript_files_in_your_project: diag(6612, ts.DiagnosticCategory.Message, "Generate_d_ts_files_from_TypeScript_and_JavaScript_files_in_your_project_6612", "Generate .d.ts files from TypeScript and JavaScript files in your project."),
            Specify_the_output_directory_for_generated_declaration_files: diag(6613, ts.DiagnosticCategory.Message, "Specify_the_output_directory_for_generated_declaration_files_6613", "Specify the output directory for generated declaration files."),
            Create_sourcemaps_for_d_ts_files: diag(6614, ts.DiagnosticCategory.Message, "Create_sourcemaps_for_d_ts_files_6614", "Create sourcemaps for d.ts files."),
            Output_compiler_performance_information_after_building: diag(6615, ts.DiagnosticCategory.Message, "Output_compiler_performance_information_after_building_6615", "Output compiler performance information after building."),
            Disables_inference_for_type_acquisition_by_looking_at_filenames_in_a_project: diag(6616, ts.DiagnosticCategory.Message, "Disables_inference_for_type_acquisition_by_looking_at_filenames_in_a_project_6616", "Disables inference for type acquisition by looking at filenames in a project."),
            Reduce_the_number_of_projects_loaded_automatically_by_TypeScript: diag(6617, ts.DiagnosticCategory.Message, "Reduce_the_number_of_projects_loaded_automatically_by_TypeScript_6617", "Reduce the number of projects loaded automatically by TypeScript."),
            Remove_the_20mb_cap_on_total_source_code_size_for_JavaScript_files_in_the_TypeScript_language_server: diag(6618, ts.DiagnosticCategory.Message, "Remove_the_20mb_cap_on_total_source_code_size_for_JavaScript_files_in_the_TypeScript_language_server_6618", "Remove the 20mb cap on total source code size for JavaScript files in the TypeScript language server."),
            Opt_a_project_out_of_multi_project_reference_checking_when_editing: diag(6619, ts.DiagnosticCategory.Message, "Opt_a_project_out_of_multi_project_reference_checking_when_editing_6619", "Opt a project out of multi-project reference checking when editing."),
            Disable_preferring_source_files_instead_of_declaration_files_when_referencing_composite_projects: diag(6620, ts.DiagnosticCategory.Message, "Disable_preferring_source_files_instead_of_declaration_files_when_referencing_composite_projects_6620", "Disable preferring source files instead of declaration files when referencing composite projects"),
            Emit_more_compliant_but_verbose_and_less_performant_JavaScript_for_iteration: diag(6621, ts.DiagnosticCategory.Message, "Emit_more_compliant_but_verbose_and_less_performant_JavaScript_for_iteration_6621", "Emit more compliant, but verbose and less performant JavaScript for iteration."),
            Only_output_d_ts_files_and_not_JavaScript_files: diag(6623, ts.DiagnosticCategory.Message, "Only_output_d_ts_files_and_not_JavaScript_files_6623", "Only output d.ts files and not JavaScript files."),
            Emit_design_type_metadata_for_decorated_declarations_in_source_files: diag(6624, ts.DiagnosticCategory.Message, "Emit_design_type_metadata_for_decorated_declarations_in_source_files_6624", "Emit design-type metadata for decorated declarations in source files."),
            Disable_the_type_acquisition_for_JavaScript_projects: diag(6625, ts.DiagnosticCategory.Message, "Disable_the_type_acquisition_for_JavaScript_projects_6625", "Disable the type acquisition for JavaScript projects"),
            Emit_additional_JavaScript_to_ease_support_for_importing_CommonJS_modules_This_enables_allowSyntheticDefaultImports_for_type_compatibility: diag(6626, ts.DiagnosticCategory.Message, "Emit_additional_JavaScript_to_ease_support_for_importing_CommonJS_modules_This_enables_allowSyntheti_6626", "Emit additional JavaScript to ease support for importing CommonJS modules. This enables `allowSyntheticDefaultImports` for type compatibility."),
            Filters_results_from_the_include_option: diag(6627, ts.DiagnosticCategory.Message, "Filters_results_from_the_include_option_6627", "Filters results from the `include` option."),
            Remove_a_list_of_directories_from_the_watch_process: diag(6628, ts.DiagnosticCategory.Message, "Remove_a_list_of_directories_from_the_watch_process_6628", "Remove a list of directories from the watch process."),
            Remove_a_list_of_files_from_the_watch_mode_s_processing: diag(6629, ts.DiagnosticCategory.Message, "Remove_a_list_of_files_from_the_watch_mode_s_processing_6629", "Remove a list of files from the watch mode's processing."),
            Enable_experimental_support_for_TC39_stage_2_draft_decorators: diag(6630, ts.DiagnosticCategory.Message, "Enable_experimental_support_for_TC39_stage_2_draft_decorators_6630", "Enable experimental support for TC39 stage 2 draft decorators."),
            Print_files_read_during_the_compilation_including_why_it_was_included: diag(6631, ts.DiagnosticCategory.Message, "Print_files_read_during_the_compilation_including_why_it_was_included_6631", "Print files read during the compilation including why it was included."),
            Output_more_detailed_compiler_performance_information_after_building: diag(6632, ts.DiagnosticCategory.Message, "Output_more_detailed_compiler_performance_information_after_building_6632", "Output more detailed compiler performance information after building."),
            Specify_one_or_more_path_or_node_module_references_to_base_configuration_files_from_which_settings_are_inherited: diag(6633, ts.DiagnosticCategory.Message, "Specify_one_or_more_path_or_node_module_references_to_base_configuration_files_from_which_settings_a_6633", "Specify one or more path or node module references to base configuration files from which settings are inherited."),
            Specify_what_approach_the_watcher_should_use_if_the_system_runs_out_of_native_file_watchers: diag(6634, ts.DiagnosticCategory.Message, "Specify_what_approach_the_watcher_should_use_if_the_system_runs_out_of_native_file_watchers_6634", "Specify what approach the watcher should use if the system runs out of native file watchers."),
            Include_a_list_of_files_This_does_not_support_glob_patterns_as_opposed_to_include: diag(6635, ts.DiagnosticCategory.Message, "Include_a_list_of_files_This_does_not_support_glob_patterns_as_opposed_to_include_6635", "Include a list of files. This does not support glob patterns, as opposed to `include`."),
            Build_all_projects_including_those_that_appear_to_be_up_to_date: diag(6636, ts.DiagnosticCategory.Message, "Build_all_projects_including_those_that_appear_to_be_up_to_date_6636", "Build all projects, including those that appear to be up to date"),
            Ensure_that_casing_is_correct_in_imports: diag(6637, ts.DiagnosticCategory.Message, "Ensure_that_casing_is_correct_in_imports_6637", "Ensure that casing is correct in imports."),
            Emit_a_v8_CPU_profile_of_the_compiler_run_for_debugging: diag(6638, ts.DiagnosticCategory.Message, "Emit_a_v8_CPU_profile_of_the_compiler_run_for_debugging_6638", "Emit a v8 CPU profile of the compiler run for debugging."),
            Allow_importing_helper_functions_from_tslib_once_per_project_instead_of_including_them_per_file: diag(6639, ts.DiagnosticCategory.Message, "Allow_importing_helper_functions_from_tslib_once_per_project_instead_of_including_them_per_file_6639", "Allow importing helper functions from tslib once per project, instead of including them per-file."),
            Specify_a_list_of_glob_patterns_that_match_files_to_be_included_in_compilation: diag(6641, ts.DiagnosticCategory.Message, "Specify_a_list_of_glob_patterns_that_match_files_to_be_included_in_compilation_6641", "Specify a list of glob patterns that match files to be included in compilation."),
            Save_tsbuildinfo_files_to_allow_for_incremental_compilation_of_projects: diag(6642, ts.DiagnosticCategory.Message, "Save_tsbuildinfo_files_to_allow_for_incremental_compilation_of_projects_6642", "Save .tsbuildinfo files to allow for incremental compilation of projects."),
            Include_sourcemap_files_inside_the_emitted_JavaScript: diag(6643, ts.DiagnosticCategory.Message, "Include_sourcemap_files_inside_the_emitted_JavaScript_6643", "Include sourcemap files inside the emitted JavaScript."),
            Include_source_code_in_the_sourcemaps_inside_the_emitted_JavaScript: diag(6644, ts.DiagnosticCategory.Message, "Include_source_code_in_the_sourcemaps_inside_the_emitted_JavaScript_6644", "Include source code in the sourcemaps inside the emitted JavaScript."),
            Ensure_that_each_file_can_be_safely_transpiled_without_relying_on_other_imports: diag(6645, ts.DiagnosticCategory.Message, "Ensure_that_each_file_can_be_safely_transpiled_without_relying_on_other_imports_6645", "Ensure that each file can be safely transpiled without relying on other imports."),
            Specify_what_JSX_code_is_generated: diag(6646, ts.DiagnosticCategory.Message, "Specify_what_JSX_code_is_generated_6646", "Specify what JSX code is generated."),
            Specify_the_JSX_factory_function_used_when_targeting_React_JSX_emit_e_g_React_createElement_or_h: diag(6647, ts.DiagnosticCategory.Message, "Specify_the_JSX_factory_function_used_when_targeting_React_JSX_emit_e_g_React_createElement_or_h_6647", "Specify the JSX factory function used when targeting React JSX emit, e.g. 'React.createElement' or 'h'"),
            Specify_the_JSX_Fragment_reference_used_for_fragments_when_targeting_React_JSX_emit_e_g_React_Fragment_or_Fragment: diag(6648, ts.DiagnosticCategory.Message, "Specify_the_JSX_Fragment_reference_used_for_fragments_when_targeting_React_JSX_emit_e_g_React_Fragme_6648", "Specify the JSX Fragment reference used for fragments when targeting React JSX emit e.g. 'React.Fragment' or 'Fragment'."),
            Specify_module_specifier_used_to_import_the_JSX_factory_functions_when_using_jsx_Colon_react_jsx_Asterisk: diag(6649, ts.DiagnosticCategory.Message, "Specify_module_specifier_used_to_import_the_JSX_factory_functions_when_using_jsx_Colon_react_jsx_Ast_6649", "Specify module specifier used to import the JSX factory functions when using `jsx: react-jsx*`.`"),
            Make_keyof_only_return_strings_instead_of_string_numbers_or_symbols_Legacy_option: diag(6650, ts.DiagnosticCategory.Message, "Make_keyof_only_return_strings_instead_of_string_numbers_or_symbols_Legacy_option_6650", "Make keyof only return strings instead of string, numbers or symbols. Legacy option."),
            Specify_a_set_of_bundled_library_declaration_files_that_describe_the_target_runtime_environment: diag(6651, ts.DiagnosticCategory.Message, "Specify_a_set_of_bundled_library_declaration_files_that_describe_the_target_runtime_environment_6651", "Specify a set of bundled library declaration files that describe the target runtime environment."),
            Print_the_names_of_emitted_files_after_a_compilation: diag(6652, ts.DiagnosticCategory.Message, "Print_the_names_of_emitted_files_after_a_compilation_6652", "Print the names of emitted files after a compilation."),
            Print_all_of_the_files_read_during_the_compilation: diag(6653, ts.DiagnosticCategory.Message, "Print_all_of_the_files_read_during_the_compilation_6653", "Print all of the files read during the compilation."),
            Set_the_language_of_the_messaging_from_TypeScript_This_does_not_affect_emit: diag(6654, ts.DiagnosticCategory.Message, "Set_the_language_of_the_messaging_from_TypeScript_This_does_not_affect_emit_6654", "Set the language of the messaging from TypeScript. This does not affect emit."),
            Specify_the_maximum_folder_depth_used_for_checking_JavaScript_files_from_node_modules_Only_applicable_with_allowJs: diag(6656, ts.DiagnosticCategory.Message, "Specify_the_maximum_folder_depth_used_for_checking_JavaScript_files_from_node_modules_Only_applicabl_6656", "Specify the maximum folder depth used for checking JavaScript files from `node_modules`. Only applicable with `allowJs`."),
            Specify_what_module_code_is_generated: diag(6657, ts.DiagnosticCategory.Message, "Specify_what_module_code_is_generated_6657", "Specify what module code is generated."),
            Specify_how_TypeScript_looks_up_a_file_from_a_given_module_specifier: diag(6658, ts.DiagnosticCategory.Message, "Specify_how_TypeScript_looks_up_a_file_from_a_given_module_specifier_6658", "Specify how TypeScript looks up a file from a given module specifier."),
            Set_the_newline_character_for_emitting_files: diag(6659, ts.DiagnosticCategory.Message, "Set_the_newline_character_for_emitting_files_6659", "Set the newline character for emitting files."),
            Disable_emitting_files_from_a_compilation: diag(6660, ts.DiagnosticCategory.Message, "Disable_emitting_files_from_a_compilation_6660", "Disable emitting files from a compilation."),
            Disable_generating_custom_helper_functions_like_extends_in_compiled_output: diag(6661, ts.DiagnosticCategory.Message, "Disable_generating_custom_helper_functions_like_extends_in_compiled_output_6661", "Disable generating custom helper functions like `__extends` in compiled output."),
            Disable_emitting_files_if_any_type_checking_errors_are_reported: diag(6662, ts.DiagnosticCategory.Message, "Disable_emitting_files_if_any_type_checking_errors_are_reported_6662", "Disable emitting files if any type checking errors are reported."),
            Disable_truncating_types_in_error_messages: diag(6663, ts.DiagnosticCategory.Message, "Disable_truncating_types_in_error_messages_6663", "Disable truncating types in error messages."),
            Enable_error_reporting_for_fallthrough_cases_in_switch_statements: diag(6664, ts.DiagnosticCategory.Message, "Enable_error_reporting_for_fallthrough_cases_in_switch_statements_6664", "Enable error reporting for fallthrough cases in switch statements."),
            Enable_error_reporting_for_expressions_and_declarations_with_an_implied_any_type: diag(6665, ts.DiagnosticCategory.Message, "Enable_error_reporting_for_expressions_and_declarations_with_an_implied_any_type_6665", "Enable error reporting for expressions and declarations with an implied `any` type.."),
            Ensure_overriding_members_in_derived_classes_are_marked_with_an_override_modifier: diag(6666, ts.DiagnosticCategory.Message, "Ensure_overriding_members_in_derived_classes_are_marked_with_an_override_modifier_6666", "Ensure overriding members in derived classes are marked with an override modifier."),
            Enable_error_reporting_for_codepaths_that_do_not_explicitly_return_in_a_function: diag(6667, ts.DiagnosticCategory.Message, "Enable_error_reporting_for_codepaths_that_do_not_explicitly_return_in_a_function_6667", "Enable error reporting for codepaths that do not explicitly return in a function."),
            Enable_error_reporting_when_this_is_given_the_type_any: diag(6668, ts.DiagnosticCategory.Message, "Enable_error_reporting_when_this_is_given_the_type_any_6668", "Enable error reporting when `this` is given the type `any`."),
            Disable_adding_use_strict_directives_in_emitted_JavaScript_files: diag(6669, ts.DiagnosticCategory.Message, "Disable_adding_use_strict_directives_in_emitted_JavaScript_files_6669", "Disable adding 'use strict' directives in emitted JavaScript files."),
            Disable_including_any_library_files_including_the_default_lib_d_ts: diag(6670, ts.DiagnosticCategory.Message, "Disable_including_any_library_files_including_the_default_lib_d_ts_6670", "Disable including any library files, including the default lib.d.ts."),
            Enforces_using_indexed_accessors_for_keys_declared_using_an_indexed_type: diag(6671, ts.DiagnosticCategory.Message, "Enforces_using_indexed_accessors_for_keys_declared_using_an_indexed_type_6671", "Enforces using indexed accessors for keys declared using an indexed type"),
            Disallow_import_s_require_s_or_reference_s_from_expanding_the_number_of_files_TypeScript_should_add_to_a_project: diag(6672, ts.DiagnosticCategory.Message, "Disallow_import_s_require_s_or_reference_s_from_expanding_the_number_of_files_TypeScript_should_add__6672", "Disallow `import`s, `require`s or `<reference>`s from expanding the number of files TypeScript should add to a project."),
            Disable_strict_checking_of_generic_signatures_in_function_types: diag(6673, ts.DiagnosticCategory.Message, "Disable_strict_checking_of_generic_signatures_in_function_types_6673", "Disable strict checking of generic signatures in function types."),
            Add_undefined_to_a_type_when_accessed_using_an_index: diag(6674, ts.DiagnosticCategory.Message, "Add_undefined_to_a_type_when_accessed_using_an_index_6674", "Add `undefined` to a type when accessed using an index."),
            Enable_error_reporting_when_a_local_variables_aren_t_read: diag(6675, ts.DiagnosticCategory.Message, "Enable_error_reporting_when_a_local_variables_aren_t_read_6675", "Enable error reporting when a local variables aren't read."),
            Raise_an_error_when_a_function_parameter_isn_t_read: diag(6676, ts.DiagnosticCategory.Message, "Raise_an_error_when_a_function_parameter_isn_t_read_6676", "Raise an error when a function parameter isn't read"),
            Deprecated_setting_Use_outFile_instead: diag(6677, ts.DiagnosticCategory.Message, "Deprecated_setting_Use_outFile_instead_6677", "Deprecated setting. Use `outFile` instead."),
            Specify_an_output_folder_for_all_emitted_files: diag(6678, ts.DiagnosticCategory.Message, "Specify_an_output_folder_for_all_emitted_files_6678", "Specify an output folder for all emitted files."),
            Specify_a_file_that_bundles_all_outputs_into_one_JavaScript_file_If_declaration_is_true_also_designates_a_file_that_bundles_all_d_ts_output: diag(6679, ts.DiagnosticCategory.Message, "Specify_a_file_that_bundles_all_outputs_into_one_JavaScript_file_If_declaration_is_true_also_designa_6679", "Specify a file that bundles all outputs into one JavaScript file. If `declaration` is true, also designates a file that bundles all .d.ts output."),
            Specify_a_set_of_entries_that_re_map_imports_to_additional_lookup_locations: diag(6680, ts.DiagnosticCategory.Message, "Specify_a_set_of_entries_that_re_map_imports_to_additional_lookup_locations_6680", "Specify a set of entries that re-map imports to additional lookup locations."),
            Specify_a_list_of_language_service_plugins_to_include: diag(6681, ts.DiagnosticCategory.Message, "Specify_a_list_of_language_service_plugins_to_include_6681", "Specify a list of language service plugins to include."),
            Disable_erasing_const_enum_declarations_in_generated_code: diag(6682, ts.DiagnosticCategory.Message, "Disable_erasing_const_enum_declarations_in_generated_code_6682", "Disable erasing `const enum` declarations in generated code."),
            Disable_resolving_symlinks_to_their_realpath_This_correlates_to_the_same_flag_in_node: diag(6683, ts.DiagnosticCategory.Message, "Disable_resolving_symlinks_to_their_realpath_This_correlates_to_the_same_flag_in_node_6683", "Disable resolving symlinks to their realpath. This correlates to the same flag in node."),
            Disable_wiping_the_console_in_watch_mode: diag(6684, ts.DiagnosticCategory.Message, "Disable_wiping_the_console_in_watch_mode_6684", "Disable wiping the console in watch mode"),
            Enable_color_and_formatting_in_TypeScript_s_output_to_make_compiler_errors_easier_to_read: diag(6685, ts.DiagnosticCategory.Message, "Enable_color_and_formatting_in_TypeScript_s_output_to_make_compiler_errors_easier_to_read_6685", "Enable color and formatting in TypeScript's output to make compiler errors easier to read"),
            Specify_the_object_invoked_for_createElement_This_only_applies_when_targeting_react_JSX_emit: diag(6686, ts.DiagnosticCategory.Message, "Specify_the_object_invoked_for_createElement_This_only_applies_when_targeting_react_JSX_emit_6686", "Specify the object invoked for `createElement`. This only applies when targeting `react` JSX emit."),
            Specify_an_array_of_objects_that_specify_paths_for_projects_Used_in_project_references: diag(6687, ts.DiagnosticCategory.Message, "Specify_an_array_of_objects_that_specify_paths_for_projects_Used_in_project_references_6687", "Specify an array of objects that specify paths for projects. Used in project references."),
            Disable_emitting_comments: diag(6688, ts.DiagnosticCategory.Message, "Disable_emitting_comments_6688", "Disable emitting comments."),
            Enable_importing_json_files: diag(6689, ts.DiagnosticCategory.Message, "Enable_importing_json_files_6689", "Enable importing .json files"),
            Specify_the_root_folder_within_your_source_files: diag(6690, ts.DiagnosticCategory.Message, "Specify_the_root_folder_within_your_source_files_6690", "Specify the root folder within your source files."),
            Allow_multiple_folders_to_be_treated_as_one_when_resolving_modules: diag(6691, ts.DiagnosticCategory.Message, "Allow_multiple_folders_to_be_treated_as_one_when_resolving_modules_6691", "Allow multiple folders to be treated as one when resolving modules."),
            Skip_type_checking_d_ts_files_that_are_included_with_TypeScript: diag(6692, ts.DiagnosticCategory.Message, "Skip_type_checking_d_ts_files_that_are_included_with_TypeScript_6692", "Skip type checking .d.ts files that are included with TypeScript."),
            Skip_type_checking_all_d_ts_files: diag(6693, ts.DiagnosticCategory.Message, "Skip_type_checking_all_d_ts_files_6693", "Skip type checking all .d.ts files."),
            Create_source_map_files_for_emitted_JavaScript_files: diag(6694, ts.DiagnosticCategory.Message, "Create_source_map_files_for_emitted_JavaScript_files_6694", "Create source map files for emitted JavaScript files."),
            Specify_the_root_path_for_debuggers_to_find_the_reference_source_code: diag(6695, ts.DiagnosticCategory.Message, "Specify_the_root_path_for_debuggers_to_find_the_reference_source_code_6695", "Specify the root path for debuggers to find the reference source code."),
            Check_that_the_arguments_for_bind_call_and_apply_methods_match_the_original_function: diag(6697, ts.DiagnosticCategory.Message, "Check_that_the_arguments_for_bind_call_and_apply_methods_match_the_original_function_6697", "Check that the arguments for `bind`, `call`, and `apply` methods match the original function."),
            When_assigning_functions_check_to_ensure_parameters_and_the_return_values_are_subtype_compatible: diag(6698, ts.DiagnosticCategory.Message, "When_assigning_functions_check_to_ensure_parameters_and_the_return_values_are_subtype_compatible_6698", "When assigning functions, check to ensure parameters and the return values are subtype-compatible."),
            When_type_checking_take_into_account_null_and_undefined: diag(6699, ts.DiagnosticCategory.Message, "When_type_checking_take_into_account_null_and_undefined_6699", "When type checking, take into account `null` and `undefined`."),
            Check_for_class_properties_that_are_declared_but_not_set_in_the_constructor: diag(6700, ts.DiagnosticCategory.Message, "Check_for_class_properties_that_are_declared_but_not_set_in_the_constructor_6700", "Check for class properties that are declared but not set in the constructor."),
            Disable_emitting_declarations_that_have_internal_in_their_JSDoc_comments: diag(6701, ts.DiagnosticCategory.Message, "Disable_emitting_declarations_that_have_internal_in_their_JSDoc_comments_6701", "Disable emitting declarations that have `@internal` in their JSDoc comments."),
            Disable_reporting_of_excess_property_errors_during_the_creation_of_object_literals: diag(6702, ts.DiagnosticCategory.Message, "Disable_reporting_of_excess_property_errors_during_the_creation_of_object_literals_6702", "Disable reporting of excess property errors during the creation of object literals."),
            Suppress_noImplicitAny_errors_when_indexing_objects_that_lack_index_signatures: diag(6703, ts.DiagnosticCategory.Message, "Suppress_noImplicitAny_errors_when_indexing_objects_that_lack_index_signatures_6703", "Suppress `noImplicitAny` errors when indexing objects that lack index signatures."),
            Synchronously_call_callbacks_and_update_the_state_of_directory_watchers_on_platforms_that_don_t_support_recursive_watching_natively: diag(6704, ts.DiagnosticCategory.Message, "Synchronously_call_callbacks_and_update_the_state_of_directory_watchers_on_platforms_that_don_t_supp_6704", "Synchronously call callbacks and update the state of directory watchers on platforms that don`t support recursive watching natively."),
            Set_the_JavaScript_language_version_for_emitted_JavaScript_and_include_compatible_library_declarations: diag(6705, ts.DiagnosticCategory.Message, "Set_the_JavaScript_language_version_for_emitted_JavaScript_and_include_compatible_library_declaratio_6705", "Set the JavaScript language version for emitted JavaScript and include compatible library declarations."),
            Log_paths_used_during_the_moduleResolution_process: diag(6706, ts.DiagnosticCategory.Message, "Log_paths_used_during_the_moduleResolution_process_6706", "Log paths used during the `moduleResolution` process."),
            Specify_the_folder_for_tsbuildinfo_incremental_compilation_files: diag(6707, ts.DiagnosticCategory.Message, "Specify_the_folder_for_tsbuildinfo_incremental_compilation_files_6707", "Specify the folder for .tsbuildinfo incremental compilation files."),
            Specify_options_for_automatic_acquisition_of_declaration_files: diag(6709, ts.DiagnosticCategory.Message, "Specify_options_for_automatic_acquisition_of_declaration_files_6709", "Specify options for automatic acquisition of declaration files."),
            Specify_multiple_folders_that_act_like_Slashnode_modules_Slash_types: diag(6710, ts.DiagnosticCategory.Message, "Specify_multiple_folders_that_act_like_Slashnode_modules_Slash_types_6710", "Specify multiple folders that act like `./node_modules/@types`."),
            Specify_type_package_names_to_be_included_without_being_referenced_in_a_source_file: diag(6711, ts.DiagnosticCategory.Message, "Specify_type_package_names_to_be_included_without_being_referenced_in_a_source_file_6711", "Specify type package names to be included without being referenced in a source file."),
            Emit_ECMAScript_standard_compliant_class_fields: diag(6712, ts.DiagnosticCategory.Message, "Emit_ECMAScript_standard_compliant_class_fields_6712", "Emit ECMAScript-standard-compliant class fields."),
            Enable_verbose_logging: diag(6713, ts.DiagnosticCategory.Message, "Enable_verbose_logging_6713", "Enable verbose logging"),
            Specify_how_directories_are_watched_on_systems_that_lack_recursive_file_watching_functionality: diag(6714, ts.DiagnosticCategory.Message, "Specify_how_directories_are_watched_on_systems_that_lack_recursive_file_watching_functionality_6714", "Specify how directories are watched on systems that lack recursive file-watching functionality."),
            Specify_how_the_TypeScript_watch_mode_works: diag(6715, ts.DiagnosticCategory.Message, "Specify_how_the_TypeScript_watch_mode_works_6715", "Specify how the TypeScript watch mode works."),
            Include_undefined_in_index_signature_results: diag(6716, ts.DiagnosticCategory.Message, "Include_undefined_in_index_signature_results_6716", "Include 'undefined' in index signature results"),
            Require_undeclared_properties_from_index_signatures_to_use_element_accesses: diag(6717, ts.DiagnosticCategory.Message, "Require_undeclared_properties_from_index_signatures_to_use_element_accesses_6717", "Require undeclared properties from index signatures to use element accesses."),
            Specify_emit_Slashchecking_behavior_for_imports_that_are_only_used_for_types: diag(6718, ts.DiagnosticCategory.Message, "Specify_emit_Slashchecking_behavior_for_imports_that_are_only_used_for_types_6718", "Specify emit/checking behavior for imports that are only used for types"),
            Type_catch_clause_variables_as_unknown_instead_of_any: diag(6803, ts.DiagnosticCategory.Message, "Type_catch_clause_variables_as_unknown_instead_of_any_6803", "Type catch clause variables as 'unknown' instead of 'any'."),
            one_of_Colon: diag(6900, ts.DiagnosticCategory.Message, "one_of_Colon_6900", "one of:"),
            one_or_more_Colon: diag(6901, ts.DiagnosticCategory.Message, "one_or_more_Colon_6901", "one or more:"),
            type_Colon: diag(6902, ts.DiagnosticCategory.Message, "type_Colon_6902", "type:"),
            default_Colon: diag(6903, ts.DiagnosticCategory.Message, "default_Colon_6903", "default:"),
            module_system_or_esModuleInterop: diag(6904, ts.DiagnosticCategory.Message, "module_system_or_esModuleInterop_6904", "module === \"system\" or esModuleInterop"),
            false_unless_strict_is_set: diag(6905, ts.DiagnosticCategory.Message, "false_unless_strict_is_set_6905", "`false`, unless `strict` is set"),
            false_unless_composite_is_set: diag(6906, ts.DiagnosticCategory.Message, "false_unless_composite_is_set_6906", "`false`, unless `composite` is set"),
            node_modules_bower_components_jspm_packages_plus_the_value_of_outDir_if_one_is_specified: diag(6907, ts.DiagnosticCategory.Message, "node_modules_bower_components_jspm_packages_plus_the_value_of_outDir_if_one_is_specified_6907", "`[\"node_modules\", \"bower_components\", \"jspm_packages\"]`, plus the value of `outDir` if one is specified."),
            if_files_is_specified_otherwise_Asterisk_Asterisk_Slash_Asterisk: diag(6908, ts.DiagnosticCategory.Message, "if_files_is_specified_otherwise_Asterisk_Asterisk_Slash_Asterisk_6908", "`[]` if `files` is specified, otherwise `[\"**/*\"]`"),
            true_if_composite_false_otherwise: diag(6909, ts.DiagnosticCategory.Message, "true_if_composite_false_otherwise_6909", "`true` if `composite`, `false` otherwise"),
            module_AMD_or_UMD_or_System_or_ES6_then_Classic_Otherwise_Node: diag(69010, ts.DiagnosticCategory.Message, "module_AMD_or_UMD_or_System_or_ES6_then_Classic_Otherwise_Node_69010", "module === `AMD` or `UMD` or `System` or `ES6`, then `Classic`, Otherwise `Node`"),
            Computed_from_the_list_of_input_files: diag(6911, ts.DiagnosticCategory.Message, "Computed_from_the_list_of_input_files_6911", "Computed from the list of input files"),
            Platform_specific: diag(6912, ts.DiagnosticCategory.Message, "Platform_specific_6912", "Platform specific"),
            You_can_learn_about_all_of_the_compiler_options_at_0: diag(6913, ts.DiagnosticCategory.Message, "You_can_learn_about_all_of_the_compiler_options_at_0_6913", "You can learn about all of the compiler options at {0}"),
            Including_watch_w_will_start_watching_the_current_project_for_the_file_changes_Once_set_you_can_config_watch_mode_with_Colon: diag(6914, ts.DiagnosticCategory.Message, "Including_watch_w_will_start_watching_the_current_project_for_the_file_changes_Once_set_you_can_conf_6914", "Including --watch, -w will start watching the current project for the file changes. Once set, you can config watch mode with:"),
            Using_build_b_will_make_tsc_behave_more_like_a_build_orchestrator_than_a_compiler_This_is_used_to_trigger_building_composite_projects_which_you_can_learn_more_about_at_0: diag(6915, ts.DiagnosticCategory.Message, "Using_build_b_will_make_tsc_behave_more_like_a_build_orchestrator_than_a_compiler_This_is_used_to_tr_6915", "Using --build, -b will make tsc behave more like a build orchestrator than a compiler. This is used to trigger building composite projects which you can learn more about at {0}"),
            COMMON_COMMANDS: diag(6916, ts.DiagnosticCategory.Message, "COMMON_COMMANDS_6916", "COMMON COMMANDS"),
            ALL_COMPILER_OPTIONS: diag(6917, ts.DiagnosticCategory.Message, "ALL_COMPILER_OPTIONS_6917", "ALL COMPILER OPTIONS"),
            WATCH_OPTIONS: diag(6918, ts.DiagnosticCategory.Message, "WATCH_OPTIONS_6918", "WATCH OPTIONS"),
            BUILD_OPTIONS: diag(6919, ts.DiagnosticCategory.Message, "BUILD_OPTIONS_6919", "BUILD OPTIONS"),
            COMMON_COMPILER_OPTIONS: diag(6920, ts.DiagnosticCategory.Message, "COMMON_COMPILER_OPTIONS_6920", "COMMON COMPILER OPTIONS"),
            COMMAND_LINE_FLAGS: diag(6921, ts.DiagnosticCategory.Message, "COMMAND_LINE_FLAGS_6921", "COMMAND LINE FLAGS"),
            tsc_Colon_The_TypeScript_Compiler: diag(6922, ts.DiagnosticCategory.Message, "tsc_Colon_The_TypeScript_Compiler_6922", "tsc: The TypeScript Compiler"),
            Compiles_the_current_project_tsconfig_json_in_the_working_directory: diag(6923, ts.DiagnosticCategory.Message, "Compiles_the_current_project_tsconfig_json_in_the_working_directory_6923", "Compiles the current project (tsconfig.json in the working directory.)"),
            Ignoring_tsconfig_json_compiles_the_specified_files_with_default_compiler_options: diag(6924, ts.DiagnosticCategory.Message, "Ignoring_tsconfig_json_compiles_the_specified_files_with_default_compiler_options_6924", "Ignoring tsconfig.json, compiles the specified files with default compiler options."),
            Build_a_composite_project_in_the_working_directory: diag(6925, ts.DiagnosticCategory.Message, "Build_a_composite_project_in_the_working_directory_6925", "Build a composite project in the working directory."),
            Creates_a_tsconfig_json_with_the_recommended_settings_in_the_working_directory: diag(6926, ts.DiagnosticCategory.Message, "Creates_a_tsconfig_json_with_the_recommended_settings_in_the_working_directory_6926", "Creates a tsconfig.json with the recommended settings in the working directory."),
            Compiles_the_TypeScript_project_located_at_the_specified_path: diag(6927, ts.DiagnosticCategory.Message, "Compiles_the_TypeScript_project_located_at_the_specified_path_6927", "Compiles the TypeScript project located at the specified path."),
            An_expanded_version_of_this_information_showing_all_possible_compiler_options: diag(6928, ts.DiagnosticCategory.Message, "An_expanded_version_of_this_information_showing_all_possible_compiler_options_6928", "An expanded version of this information, showing all possible compiler options"),
            Compiles_the_current_project_with_additional_settings: diag(6929, ts.DiagnosticCategory.Message, "Compiles_the_current_project_with_additional_settings_6929", "Compiles the current project, with additional settings."),
            Variable_0_implicitly_has_an_1_type: diag(7005, ts.DiagnosticCategory.Error, "Variable_0_implicitly_has_an_1_type_7005", "Variable '{0}' implicitly has an '{1}' type."),
            Parameter_0_implicitly_has_an_1_type: diag(7006, ts.DiagnosticCategory.Error, "Parameter_0_implicitly_has_an_1_type_7006", "Parameter '{0}' implicitly has an '{1}' type."),
            Member_0_implicitly_has_an_1_type: diag(7008, ts.DiagnosticCategory.Error, "Member_0_implicitly_has_an_1_type_7008", "Member '{0}' implicitly has an '{1}' type."),
            new_expression_whose_target_lacks_a_construct_signature_implicitly_has_an_any_type: diag(7009, ts.DiagnosticCategory.Error, "new_expression_whose_target_lacks_a_construct_signature_implicitly_has_an_any_type_7009", "'new' expression, whose target lacks a construct signature, implicitly has an 'any' type."),
            _0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type: diag(7010, ts.DiagnosticCategory.Error, "_0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type_7010", "'{0}', which lacks return-type annotation, implicitly has an '{1}' return type."),
            Function_expression_which_lacks_return_type_annotation_implicitly_has_an_0_return_type: diag(7011, ts.DiagnosticCategory.Error, "Function_expression_which_lacks_return_type_annotation_implicitly_has_an_0_return_type_7011", "Function expression, which lacks return-type annotation, implicitly has an '{0}' return type."),
            Construct_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type: diag(7013, ts.DiagnosticCategory.Error, "Construct_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type_7013", "Construct signature, which lacks return-type annotation, implicitly has an 'any' return type."),
            Function_type_which_lacks_return_type_annotation_implicitly_has_an_0_return_type: diag(7014, ts.DiagnosticCategory.Error, "Function_type_which_lacks_return_type_annotation_implicitly_has_an_0_return_type_7014", "Function type, which lacks return-type annotation, implicitly has an '{0}' return type."),
            Element_implicitly_has_an_any_type_because_index_expression_is_not_of_type_number: diag(7015, ts.DiagnosticCategory.Error, "Element_implicitly_has_an_any_type_because_index_expression_is_not_of_type_number_7015", "Element implicitly has an 'any' type because index expression is not of type 'number'."),
            Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type: diag(7016, ts.DiagnosticCategory.Error, "Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type_7016", "Could not find a declaration file for module '{0}'. '{1}' implicitly has an 'any' type."),
            Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature: diag(7017, ts.DiagnosticCategory.Error, "Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature_7017", "Element implicitly has an 'any' type because type '{0}' has no index signature."),
            Object_literal_s_property_0_implicitly_has_an_1_type: diag(7018, ts.DiagnosticCategory.Error, "Object_literal_s_property_0_implicitly_has_an_1_type_7018", "Object literal's property '{0}' implicitly has an '{1}' type."),
            Rest_parameter_0_implicitly_has_an_any_type: diag(7019, ts.DiagnosticCategory.Error, "Rest_parameter_0_implicitly_has_an_any_type_7019", "Rest parameter '{0}' implicitly has an 'any[]' type."),
            Call_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type: diag(7020, ts.DiagnosticCategory.Error, "Call_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type_7020", "Call signature, which lacks return-type annotation, implicitly has an 'any' return type."),
            _0_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_and_is_referenced_directly_or_indirectly_in_its_own_initializer: diag(7022, ts.DiagnosticCategory.Error, "_0_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_and_is_referenced_directly_or__7022", "'{0}' implicitly has type 'any' because it does not have a type annotation and is referenced directly or indirectly in its own initializer."),
            _0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions: diag(7023, ts.DiagnosticCategory.Error, "_0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_reference_7023", "'{0}' implicitly has return type 'any' because it does not have a return type annotation and is referenced directly or indirectly in one of its return expressions."),
            Function_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions: diag(7024, ts.DiagnosticCategory.Error, "Function_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_ref_7024", "Function implicitly has return type 'any' because it does not have a return type annotation and is referenced directly or indirectly in one of its return expressions."),
            Generator_implicitly_has_yield_type_0_because_it_does_not_yield_any_values_Consider_supplying_a_return_type_annotation: diag(7025, ts.DiagnosticCategory.Error, "Generator_implicitly_has_yield_type_0_because_it_does_not_yield_any_values_Consider_supplying_a_retu_7025", "Generator implicitly has yield type '{0}' because it does not yield any values. Consider supplying a return type annotation."),
            JSX_element_implicitly_has_type_any_because_no_interface_JSX_0_exists: diag(7026, ts.DiagnosticCategory.Error, "JSX_element_implicitly_has_type_any_because_no_interface_JSX_0_exists_7026", "JSX element implicitly has type 'any' because no interface 'JSX.{0}' exists."),
            Unreachable_code_detected: diag(7027, ts.DiagnosticCategory.Error, "Unreachable_code_detected_7027", "Unreachable code detected.", /*reportsUnnecessary*/ true),
            Unused_label: diag(7028, ts.DiagnosticCategory.Error, "Unused_label_7028", "Unused label.", /*reportsUnnecessary*/ true),
            Fallthrough_case_in_switch: diag(7029, ts.DiagnosticCategory.Error, "Fallthrough_case_in_switch_7029", "Fallthrough case in switch."),
            Not_all_code_paths_return_a_value: diag(7030, ts.DiagnosticCategory.Error, "Not_all_code_paths_return_a_value_7030", "Not all code paths return a value."),
            Binding_element_0_implicitly_has_an_1_type: diag(7031, ts.DiagnosticCategory.Error, "Binding_element_0_implicitly_has_an_1_type_7031", "Binding element '{0}' implicitly has an '{1}' type."),
            Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation: diag(7032, ts.DiagnosticCategory.Error, "Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation_7032", "Property '{0}' implicitly has type 'any', because its set accessor lacks a parameter type annotation."),
            Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation: diag(7033, ts.DiagnosticCategory.Error, "Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation_7033", "Property '{0}' implicitly has type 'any', because its get accessor lacks a return type annotation."),
            Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined: diag(7034, ts.DiagnosticCategory.Error, "Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined_7034", "Variable '{0}' implicitly has type '{1}' in some locations where its type cannot be determined."),
            Try_npm_i_save_dev_types_Slash_1_if_it_exists_or_add_a_new_declaration_d_ts_file_containing_declare_module_0: diag(7035, ts.DiagnosticCategory.Error, "Try_npm_i_save_dev_types_Slash_1_if_it_exists_or_add_a_new_declaration_d_ts_file_containing_declare__7035", "Try `npm i --save-dev @types/{1}` if it exists or add a new declaration (.d.ts) file containing `declare module '{0}';`"),
            Dynamic_import_s_specifier_must_be_of_type_string_but_here_has_type_0: diag(7036, ts.DiagnosticCategory.Error, "Dynamic_import_s_specifier_must_be_of_type_string_but_here_has_type_0_7036", "Dynamic import's specifier must be of type 'string', but here has type '{0}'."),
            Enables_emit_interoperability_between_CommonJS_and_ES_Modules_via_creation_of_namespace_objects_for_all_imports_Implies_allowSyntheticDefaultImports: diag(7037, ts.DiagnosticCategory.Message, "Enables_emit_interoperability_between_CommonJS_and_ES_Modules_via_creation_of_namespace_objects_for__7037", "Enables emit interoperability between CommonJS and ES Modules via creation of namespace objects for all imports. Implies 'allowSyntheticDefaultImports'."),
            Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cause_a_failure_at_runtime_Consider_using_a_default_import_or_import_require_here_instead: diag(7038, ts.DiagnosticCategory.Message, "Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cau_7038", "Type originates at this import. A namespace-style import cannot be called or constructed, and will cause a failure at runtime. Consider using a default import or import require here instead."),
            Mapped_object_type_implicitly_has_an_any_template_type: diag(7039, ts.DiagnosticCategory.Error, "Mapped_object_type_implicitly_has_an_any_template_type_7039", "Mapped object type implicitly has an 'any' template type."),
            If_the_0_package_actually_exposes_this_module_consider_sending_a_pull_request_to_amend_https_Colon_Slash_Slashgithub_com_SlashDefinitelyTyped_SlashDefinitelyTyped_Slashtree_Slashmaster_Slashtypes_Slash_1: diag(7040, ts.DiagnosticCategory.Error, "If_the_0_package_actually_exposes_this_module_consider_sending_a_pull_request_to_amend_https_Colon_S_7040", "If the '{0}' package actually exposes this module, consider sending a pull request to amend 'https://github.com/DefinitelyTyped/DefinitelyTyped/tree/master/types/{1}'"),
            The_containing_arrow_function_captures_the_global_value_of_this: diag(7041, ts.DiagnosticCategory.Error, "The_containing_arrow_function_captures_the_global_value_of_this_7041", "The containing arrow function captures the global value of 'this'."),
            Module_0_was_resolved_to_1_but_resolveJsonModule_is_not_used: diag(7042, ts.DiagnosticCategory.Error, "Module_0_was_resolved_to_1_but_resolveJsonModule_is_not_used_7042", "Module '{0}' was resolved to '{1}', but '--resolveJsonModule' is not used."),
            Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage: diag(7043, ts.DiagnosticCategory.Suggestion, "Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage_7043", "Variable '{0}' implicitly has an '{1}' type, but a better type may be inferred from usage."),
            Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage: diag(7044, ts.DiagnosticCategory.Suggestion, "Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage_7044", "Parameter '{0}' implicitly has an '{1}' type, but a better type may be inferred from usage."),
            Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage: diag(7045, ts.DiagnosticCategory.Suggestion, "Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage_7045", "Member '{0}' implicitly has an '{1}' type, but a better type may be inferred from usage."),
            Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage: diag(7046, ts.DiagnosticCategory.Suggestion, "Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage_7046", "Variable '{0}' implicitly has type '{1}' in some locations, but a better type may be inferred from usage."),
            Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage: diag(7047, ts.DiagnosticCategory.Suggestion, "Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage_7047", "Rest parameter '{0}' implicitly has an 'any[]' type, but a better type may be inferred from usage."),
            Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage: diag(7048, ts.DiagnosticCategory.Suggestion, "Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage_7048", "Property '{0}' implicitly has type 'any', but a better type for its get accessor may be inferred from usage."),
            Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage: diag(7049, ts.DiagnosticCategory.Suggestion, "Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage_7049", "Property '{0}' implicitly has type 'any', but a better type for its set accessor may be inferred from usage."),
            _0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage: diag(7050, ts.DiagnosticCategory.Suggestion, "_0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage_7050", "'{0}' implicitly has an '{1}' return type, but a better type may be inferred from usage."),
            Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1: diag(7051, ts.DiagnosticCategory.Error, "Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1_7051", "Parameter has a name but no type. Did you mean '{0}: {1}'?"),
            Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature_Did_you_mean_to_call_1: diag(7052, ts.DiagnosticCategory.Error, "Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature_Did_you_mean_to_call_1_7052", "Element implicitly has an 'any' type because type '{0}' has no index signature. Did you mean to call '{1}'?"),
            Element_implicitly_has_an_any_type_because_expression_of_type_0_can_t_be_used_to_index_type_1: diag(7053, ts.DiagnosticCategory.Error, "Element_implicitly_has_an_any_type_because_expression_of_type_0_can_t_be_used_to_index_type_1_7053", "Element implicitly has an 'any' type because expression of type '{0}' can't be used to index type '{1}'."),
            No_index_signature_with_a_parameter_of_type_0_was_found_on_type_1: diag(7054, ts.DiagnosticCategory.Error, "No_index_signature_with_a_parameter_of_type_0_was_found_on_type_1_7054", "No index signature with a parameter of type '{0}' was found on type '{1}'."),
            _0_which_lacks_return_type_annotation_implicitly_has_an_1_yield_type: diag(7055, ts.DiagnosticCategory.Error, "_0_which_lacks_return_type_annotation_implicitly_has_an_1_yield_type_7055", "'{0}', which lacks return-type annotation, implicitly has an '{1}' yield type."),
            The_inferred_type_of_this_node_exceeds_the_maximum_length_the_compiler_will_serialize_An_explicit_type_annotation_is_needed: diag(7056, ts.DiagnosticCategory.Error, "The_inferred_type_of_this_node_exceeds_the_maximum_length_the_compiler_will_serialize_An_explicit_ty_7056", "The inferred type of this node exceeds the maximum length the compiler will serialize. An explicit type annotation is needed."),
            yield_expression_implicitly_results_in_an_any_type_because_its_containing_generator_lacks_a_return_type_annotation: diag(7057, ts.DiagnosticCategory.Error, "yield_expression_implicitly_results_in_an_any_type_because_its_containing_generator_lacks_a_return_t_7057", "'yield' expression implicitly results in an 'any' type because its containing generator lacks a return-type annotation."),
            You_cannot_rename_this_element: diag(8000, ts.DiagnosticCategory.Error, "You_cannot_rename_this_element_8000", "You cannot rename this element."),
            You_cannot_rename_elements_that_are_defined_in_the_standard_TypeScript_library: diag(8001, ts.DiagnosticCategory.Error, "You_cannot_rename_elements_that_are_defined_in_the_standard_TypeScript_library_8001", "You cannot rename elements that are defined in the standard TypeScript library."),
            import_can_only_be_used_in_TypeScript_files: diag(8002, ts.DiagnosticCategory.Error, "import_can_only_be_used_in_TypeScript_files_8002", "'import ... =' can only be used in TypeScript files."),
            export_can_only_be_used_in_TypeScript_files: diag(8003, ts.DiagnosticCategory.Error, "export_can_only_be_used_in_TypeScript_files_8003", "'export =' can only be used in TypeScript files."),
            Type_parameter_declarations_can_only_be_used_in_TypeScript_files: diag(8004, ts.DiagnosticCategory.Error, "Type_parameter_declarations_can_only_be_used_in_TypeScript_files_8004", "Type parameter declarations can only be used in TypeScript files."),
            implements_clauses_can_only_be_used_in_TypeScript_files: diag(8005, ts.DiagnosticCategory.Error, "implements_clauses_can_only_be_used_in_TypeScript_files_8005", "'implements' clauses can only be used in TypeScript files."),
            _0_declarations_can_only_be_used_in_TypeScript_files: diag(8006, ts.DiagnosticCategory.Error, "_0_declarations_can_only_be_used_in_TypeScript_files_8006", "'{0}' declarations can only be used in TypeScript files."),
            Type_aliases_can_only_be_used_in_TypeScript_files: diag(8008, ts.DiagnosticCategory.Error, "Type_aliases_can_only_be_used_in_TypeScript_files_8008", "Type aliases can only be used in TypeScript files."),
            The_0_modifier_can_only_be_used_in_TypeScript_files: diag(8009, ts.DiagnosticCategory.Error, "The_0_modifier_can_only_be_used_in_TypeScript_files_8009", "The '{0}' modifier can only be used in TypeScript files."),
            Type_annotations_can_only_be_used_in_TypeScript_files: diag(8010, ts.DiagnosticCategory.Error, "Type_annotations_can_only_be_used_in_TypeScript_files_8010", "Type annotations can only be used in TypeScript files."),
            Type_arguments_can_only_be_used_in_TypeScript_files: diag(8011, ts.DiagnosticCategory.Error, "Type_arguments_can_only_be_used_in_TypeScript_files_8011", "Type arguments can only be used in TypeScript files."),
            Parameter_modifiers_can_only_be_used_in_TypeScript_files: diag(8012, ts.DiagnosticCategory.Error, "Parameter_modifiers_can_only_be_used_in_TypeScript_files_8012", "Parameter modifiers can only be used in TypeScript files."),
            Non_null_assertions_can_only_be_used_in_TypeScript_files: diag(8013, ts.DiagnosticCategory.Error, "Non_null_assertions_can_only_be_used_in_TypeScript_files_8013", "Non-null assertions can only be used in TypeScript files."),
            Type_assertion_expressions_can_only_be_used_in_TypeScript_files: diag(8016, ts.DiagnosticCategory.Error, "Type_assertion_expressions_can_only_be_used_in_TypeScript_files_8016", "Type assertion expressions can only be used in TypeScript files."),
            Octal_literal_types_must_use_ES2015_syntax_Use_the_syntax_0: diag(8017, ts.DiagnosticCategory.Error, "Octal_literal_types_must_use_ES2015_syntax_Use_the_syntax_0_8017", "Octal literal types must use ES2015 syntax. Use the syntax '{0}'."),
            Octal_literals_are_not_allowed_in_enums_members_initializer_Use_the_syntax_0: diag(8018, ts.DiagnosticCategory.Error, "Octal_literals_are_not_allowed_in_enums_members_initializer_Use_the_syntax_0_8018", "Octal literals are not allowed in enums members initializer. Use the syntax '{0}'."),
            Report_errors_in_js_files: diag(8019, ts.DiagnosticCategory.Message, "Report_errors_in_js_files_8019", "Report errors in .js files."),
            JSDoc_types_can_only_be_used_inside_documentation_comments: diag(8020, ts.DiagnosticCategory.Error, "JSDoc_types_can_only_be_used_inside_documentation_comments_8020", "JSDoc types can only be used inside documentation comments."),
            JSDoc_typedef_tag_should_either_have_a_type_annotation_or_be_followed_by_property_or_member_tags: diag(8021, ts.DiagnosticCategory.Error, "JSDoc_typedef_tag_should_either_have_a_type_annotation_or_be_followed_by_property_or_member_tags_8021", "JSDoc '@typedef' tag should either have a type annotation or be followed by '@property' or '@member' tags."),
            JSDoc_0_is_not_attached_to_a_class: diag(8022, ts.DiagnosticCategory.Error, "JSDoc_0_is_not_attached_to_a_class_8022", "JSDoc '@{0}' is not attached to a class."),
            JSDoc_0_1_does_not_match_the_extends_2_clause: diag(8023, ts.DiagnosticCategory.Error, "JSDoc_0_1_does_not_match_the_extends_2_clause_8023", "JSDoc '@{0} {1}' does not match the 'extends {2}' clause."),
            JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name: diag(8024, ts.DiagnosticCategory.Error, "JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_8024", "JSDoc '@param' tag has name '{0}', but there is no parameter with that name."),
            Class_declarations_cannot_have_more_than_one_augments_or_extends_tag: diag(8025, ts.DiagnosticCategory.Error, "Class_declarations_cannot_have_more_than_one_augments_or_extends_tag_8025", "Class declarations cannot have more than one '@augments' or '@extends' tag."),
            Expected_0_type_arguments_provide_these_with_an_extends_tag: diag(8026, ts.DiagnosticCategory.Error, "Expected_0_type_arguments_provide_these_with_an_extends_tag_8026", "Expected {0} type arguments; provide these with an '@extends' tag."),
            Expected_0_1_type_arguments_provide_these_with_an_extends_tag: diag(8027, ts.DiagnosticCategory.Error, "Expected_0_1_type_arguments_provide_these_with_an_extends_tag_8027", "Expected {0}-{1} type arguments; provide these with an '@extends' tag."),
            JSDoc_may_only_appear_in_the_last_parameter_of_a_signature: diag(8028, ts.DiagnosticCategory.Error, "JSDoc_may_only_appear_in_the_last_parameter_of_a_signature_8028", "JSDoc '...' may only appear in the last parameter of a signature."),
            JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_It_would_match_arguments_if_it_had_an_array_type: diag(8029, ts.DiagnosticCategory.Error, "JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_It_would_match_arguments_if_it_h_8029", "JSDoc '@param' tag has name '{0}', but there is no parameter with that name. It would match 'arguments' if it had an array type."),
            The_type_of_a_function_declaration_must_match_the_function_s_signature: diag(8030, ts.DiagnosticCategory.Error, "The_type_of_a_function_declaration_must_match_the_function_s_signature_8030", "The type of a function declaration must match the function's signature."),
            You_cannot_rename_a_module_via_a_global_import: diag(8031, ts.DiagnosticCategory.Error, "You_cannot_rename_a_module_via_a_global_import_8031", "You cannot rename a module via a global import."),
            Qualified_name_0_is_not_allowed_without_a_leading_param_object_1: diag(8032, ts.DiagnosticCategory.Error, "Qualified_name_0_is_not_allowed_without_a_leading_param_object_1_8032", "Qualified name '{0}' is not allowed without a leading '@param {object} {1}'."),
            A_JSDoc_typedef_comment_may_not_contain_multiple_type_tags: diag(8033, ts.DiagnosticCategory.Error, "A_JSDoc_typedef_comment_may_not_contain_multiple_type_tags_8033", "A JSDoc '@typedef' comment may not contain multiple '@type' tags."),
            The_tag_was_first_specified_here: diag(8034, ts.DiagnosticCategory.Error, "The_tag_was_first_specified_here_8034", "The tag was first specified here."),
            Only_identifiers_Slashqualified_names_with_optional_type_arguments_are_currently_supported_in_a_class_extends_clause: diag(9002, ts.DiagnosticCategory.Error, "Only_identifiers_Slashqualified_names_with_optional_type_arguments_are_currently_supported_in_a_clas_9002", "Only identifiers/qualified-names with optional type arguments are currently supported in a class 'extends' clause."),
            class_expressions_are_not_currently_supported: diag(9003, ts.DiagnosticCategory.Error, "class_expressions_are_not_currently_supported_9003", "'class' expressions are not currently supported."),
            Language_service_is_disabled: diag(9004, ts.DiagnosticCategory.Error, "Language_service_is_disabled_9004", "Language service is disabled."),
            Declaration_emit_for_this_file_requires_using_private_name_0_An_explicit_type_annotation_may_unblock_declaration_emit: diag(9005, ts.DiagnosticCategory.Error, "Declaration_emit_for_this_file_requires_using_private_name_0_An_explicit_type_annotation_may_unblock_9005", "Declaration emit for this file requires using private name '{0}'. An explicit type annotation may unblock declaration emit."),
            Declaration_emit_for_this_file_requires_using_private_name_0_from_module_1_An_explicit_type_annotation_may_unblock_declaration_emit: diag(9006, ts.DiagnosticCategory.Error, "Declaration_emit_for_this_file_requires_using_private_name_0_from_module_1_An_explicit_type_annotati_9006", "Declaration emit for this file requires using private name '{0}' from module '{1}'. An explicit type annotation may unblock declaration emit."),
            JSX_attributes_must_only_be_assigned_a_non_empty_expression: diag(17000, ts.DiagnosticCategory.Error, "JSX_attributes_must_only_be_assigned_a_non_empty_expression_17000", "JSX attributes must only be assigned a non-empty 'expression'."),
            JSX_elements_cannot_have_multiple_attributes_with_the_same_name: diag(17001, ts.DiagnosticCategory.Error, "JSX_elements_cannot_have_multiple_attributes_with_the_same_name_17001", "JSX elements cannot have multiple attributes with the same name."),
            Expected_corresponding_JSX_closing_tag_for_0: diag(17002, ts.DiagnosticCategory.Error, "Expected_corresponding_JSX_closing_tag_for_0_17002", "Expected corresponding JSX closing tag for '{0}'."),
            JSX_attribute_expected: diag(17003, ts.DiagnosticCategory.Error, "JSX_attribute_expected_17003", "JSX attribute expected."),
            Cannot_use_JSX_unless_the_jsx_flag_is_provided: diag(17004, ts.DiagnosticCategory.Error, "Cannot_use_JSX_unless_the_jsx_flag_is_provided_17004", "Cannot use JSX unless the '--jsx' flag is provided."),
            A_constructor_cannot_contain_a_super_call_when_its_class_extends_null: diag(17005, ts.DiagnosticCategory.Error, "A_constructor_cannot_contain_a_super_call_when_its_class_extends_null_17005", "A constructor cannot contain a 'super' call when its class extends 'null'."),
            An_unary_expression_with_the_0_operator_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses: diag(17006, ts.DiagnosticCategory.Error, "An_unary_expression_with_the_0_operator_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_ex_17006", "An unary expression with the '{0}' operator is not allowed in the left-hand side of an exponentiation expression. Consider enclosing the expression in parentheses."),
            A_type_assertion_expression_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses: diag(17007, ts.DiagnosticCategory.Error, "A_type_assertion_expression_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Con_17007", "A type assertion expression is not allowed in the left-hand side of an exponentiation expression. Consider enclosing the expression in parentheses."),
            JSX_element_0_has_no_corresponding_closing_tag: diag(17008, ts.DiagnosticCategory.Error, "JSX_element_0_has_no_corresponding_closing_tag_17008", "JSX element '{0}' has no corresponding closing tag."),
            super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class: diag(17009, ts.DiagnosticCategory.Error, "super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class_17009", "'super' must be called before accessing 'this' in the constructor of a derived class."),
            Unknown_type_acquisition_option_0: diag(17010, ts.DiagnosticCategory.Error, "Unknown_type_acquisition_option_0_17010", "Unknown type acquisition option '{0}'."),
            super_must_be_called_before_accessing_a_property_of_super_in_the_constructor_of_a_derived_class: diag(17011, ts.DiagnosticCategory.Error, "super_must_be_called_before_accessing_a_property_of_super_in_the_constructor_of_a_derived_class_17011", "'super' must be called before accessing a property of 'super' in the constructor of a derived class."),
            _0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2: diag(17012, ts.DiagnosticCategory.Error, "_0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2_17012", "'{0}' is not a valid meta-property for keyword '{1}'. Did you mean '{2}'?"),
            Meta_property_0_is_only_allowed_in_the_body_of_a_function_declaration_function_expression_or_constructor: diag(17013, ts.DiagnosticCategory.Error, "Meta_property_0_is_only_allowed_in_the_body_of_a_function_declaration_function_expression_or_constru_17013", "Meta-property '{0}' is only allowed in the body of a function declaration, function expression, or constructor."),
            JSX_fragment_has_no_corresponding_closing_tag: diag(17014, ts.DiagnosticCategory.Error, "JSX_fragment_has_no_corresponding_closing_tag_17014", "JSX fragment has no corresponding closing tag."),
            Expected_corresponding_closing_tag_for_JSX_fragment: diag(17015, ts.DiagnosticCategory.Error, "Expected_corresponding_closing_tag_for_JSX_fragment_17015", "Expected corresponding closing tag for JSX fragment."),
            The_jsxFragmentFactory_compiler_option_must_be_provided_to_use_JSX_fragments_with_the_jsxFactory_compiler_option: diag(17016, ts.DiagnosticCategory.Error, "The_jsxFragmentFactory_compiler_option_must_be_provided_to_use_JSX_fragments_with_the_jsxFactory_com_17016", "The 'jsxFragmentFactory' compiler option must be provided to use JSX fragments with the 'jsxFactory' compiler option."),
            An_jsxFrag_pragma_is_required_when_using_an_jsx_pragma_with_JSX_fragments: diag(17017, ts.DiagnosticCategory.Error, "An_jsxFrag_pragma_is_required_when_using_an_jsx_pragma_with_JSX_fragments_17017", "An @jsxFrag pragma is required when using an @jsx pragma with JSX fragments."),
            Unknown_type_acquisition_option_0_Did_you_mean_1: diag(17018, ts.DiagnosticCategory.Error, "Unknown_type_acquisition_option_0_Did_you_mean_1_17018", "Unknown type acquisition option '{0}'. Did you mean '{1}'?"),
            Circularity_detected_while_resolving_configuration_Colon_0: diag(18000, ts.DiagnosticCategory.Error, "Circularity_detected_while_resolving_configuration_Colon_0_18000", "Circularity detected while resolving configuration: {0}"),
            A_path_in_an_extends_option_must_be_relative_or_rooted_but_0_is_not: diag(18001, ts.DiagnosticCategory.Error, "A_path_in_an_extends_option_must_be_relative_or_rooted_but_0_is_not_18001", "A path in an 'extends' option must be relative or rooted, but '{0}' is not."),
            The_files_list_in_config_file_0_is_empty: diag(18002, ts.DiagnosticCategory.Error, "The_files_list_in_config_file_0_is_empty_18002", "The 'files' list in config file '{0}' is empty."),
            No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2: diag(18003, ts.DiagnosticCategory.Error, "No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2_18003", "No inputs were found in config file '{0}'. Specified 'include' paths were '{1}' and 'exclude' paths were '{2}'."),
            File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module: diag(80001, ts.DiagnosticCategory.Suggestion, "File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module_80001", "File is a CommonJS module; it may be converted to an ES6 module."),
            This_constructor_function_may_be_converted_to_a_class_declaration: diag(80002, ts.DiagnosticCategory.Suggestion, "This_constructor_function_may_be_converted_to_a_class_declaration_80002", "This constructor function may be converted to a class declaration."),
            Import_may_be_converted_to_a_default_import: diag(80003, ts.DiagnosticCategory.Suggestion, "Import_may_be_converted_to_a_default_import_80003", "Import may be converted to a default import."),
            JSDoc_types_may_be_moved_to_TypeScript_types: diag(80004, ts.DiagnosticCategory.Suggestion, "JSDoc_types_may_be_moved_to_TypeScript_types_80004", "JSDoc types may be moved to TypeScript types."),
            require_call_may_be_converted_to_an_import: diag(80005, ts.DiagnosticCategory.Suggestion, "require_call_may_be_converted_to_an_import_80005", "'require' call may be converted to an import."),
            This_may_be_converted_to_an_async_function: diag(80006, ts.DiagnosticCategory.Suggestion, "This_may_be_converted_to_an_async_function_80006", "This may be converted to an async function."),
            await_has_no_effect_on_the_type_of_this_expression: diag(80007, ts.DiagnosticCategory.Suggestion, "await_has_no_effect_on_the_type_of_this_expression_80007", "'await' has no effect on the type of this expression."),
            Numeric_literals_with_absolute_values_equal_to_2_53_or_greater_are_too_large_to_be_represented_accurately_as_integers: diag(80008, ts.DiagnosticCategory.Suggestion, "Numeric_literals_with_absolute_values_equal_to_2_53_or_greater_are_too_large_to_be_represented_accur_80008", "Numeric literals with absolute values equal to 2^53 or greater are too large to be represented accurately as integers."),
            Add_missing_super_call: diag(90001, ts.DiagnosticCategory.Message, "Add_missing_super_call_90001", "Add missing 'super()' call"),
            Make_super_call_the_first_statement_in_the_constructor: diag(90002, ts.DiagnosticCategory.Message, "Make_super_call_the_first_statement_in_the_constructor_90002", "Make 'super()' call the first statement in the constructor"),
            Change_extends_to_implements: diag(90003, ts.DiagnosticCategory.Message, "Change_extends_to_implements_90003", "Change 'extends' to 'implements'"),
            Remove_unused_declaration_for_Colon_0: diag(90004, ts.DiagnosticCategory.Message, "Remove_unused_declaration_for_Colon_0_90004", "Remove unused declaration for: '{0}'"),
            Remove_import_from_0: diag(90005, ts.DiagnosticCategory.Message, "Remove_import_from_0_90005", "Remove import from '{0}'"),
            Implement_interface_0: diag(90006, ts.DiagnosticCategory.Message, "Implement_interface_0_90006", "Implement interface '{0}'"),
            Implement_inherited_abstract_class: diag(90007, ts.DiagnosticCategory.Message, "Implement_inherited_abstract_class_90007", "Implement inherited abstract class"),
            Add_0_to_unresolved_variable: diag(90008, ts.DiagnosticCategory.Message, "Add_0_to_unresolved_variable_90008", "Add '{0}.' to unresolved variable"),
            Remove_variable_statement: diag(90010, ts.DiagnosticCategory.Message, "Remove_variable_statement_90010", "Remove variable statement"),
            Remove_template_tag: diag(90011, ts.DiagnosticCategory.Message, "Remove_template_tag_90011", "Remove template tag"),
            Remove_type_parameters: diag(90012, ts.DiagnosticCategory.Message, "Remove_type_parameters_90012", "Remove type parameters"),
            Import_0_from_module_1: diag(90013, ts.DiagnosticCategory.Message, "Import_0_from_module_1_90013", "Import '{0}' from module \"{1}\""),
            Change_0_to_1: diag(90014, ts.DiagnosticCategory.Message, "Change_0_to_1_90014", "Change '{0}' to '{1}'"),
            Add_0_to_existing_import_declaration_from_1: diag(90015, ts.DiagnosticCategory.Message, "Add_0_to_existing_import_declaration_from_1_90015", "Add '{0}' to existing import declaration from \"{1}\""),
            Declare_property_0: diag(90016, ts.DiagnosticCategory.Message, "Declare_property_0_90016", "Declare property '{0}'"),
            Add_index_signature_for_property_0: diag(90017, ts.DiagnosticCategory.Message, "Add_index_signature_for_property_0_90017", "Add index signature for property '{0}'"),
            Disable_checking_for_this_file: diag(90018, ts.DiagnosticCategory.Message, "Disable_checking_for_this_file_90018", "Disable checking for this file"),
            Ignore_this_error_message: diag(90019, ts.DiagnosticCategory.Message, "Ignore_this_error_message_90019", "Ignore this error message"),
            Initialize_property_0_in_the_constructor: diag(90020, ts.DiagnosticCategory.Message, "Initialize_property_0_in_the_constructor_90020", "Initialize property '{0}' in the constructor"),
            Initialize_static_property_0: diag(90021, ts.DiagnosticCategory.Message, "Initialize_static_property_0_90021", "Initialize static property '{0}'"),
            Change_spelling_to_0: diag(90022, ts.DiagnosticCategory.Message, "Change_spelling_to_0_90022", "Change spelling to '{0}'"),
            Declare_method_0: diag(90023, ts.DiagnosticCategory.Message, "Declare_method_0_90023", "Declare method '{0}'"),
            Declare_static_method_0: diag(90024, ts.DiagnosticCategory.Message, "Declare_static_method_0_90024", "Declare static method '{0}'"),
            Prefix_0_with_an_underscore: diag(90025, ts.DiagnosticCategory.Message, "Prefix_0_with_an_underscore_90025", "Prefix '{0}' with an underscore"),
            Rewrite_as_the_indexed_access_type_0: diag(90026, ts.DiagnosticCategory.Message, "Rewrite_as_the_indexed_access_type_0_90026", "Rewrite as the indexed access type '{0}'"),
            Declare_static_property_0: diag(90027, ts.DiagnosticCategory.Message, "Declare_static_property_0_90027", "Declare static property '{0}'"),
            Call_decorator_expression: diag(90028, ts.DiagnosticCategory.Message, "Call_decorator_expression_90028", "Call decorator expression"),
            Add_async_modifier_to_containing_function: diag(90029, ts.DiagnosticCategory.Message, "Add_async_modifier_to_containing_function_90029", "Add async modifier to containing function"),
            Replace_infer_0_with_unknown: diag(90030, ts.DiagnosticCategory.Message, "Replace_infer_0_with_unknown_90030", "Replace 'infer {0}' with 'unknown'"),
            Replace_all_unused_infer_with_unknown: diag(90031, ts.DiagnosticCategory.Message, "Replace_all_unused_infer_with_unknown_90031", "Replace all unused 'infer' with 'unknown'"),
            Import_default_0_from_module_1: diag(90032, ts.DiagnosticCategory.Message, "Import_default_0_from_module_1_90032", "Import default '{0}' from module \"{1}\""),
            Add_default_import_0_to_existing_import_declaration_from_1: diag(90033, ts.DiagnosticCategory.Message, "Add_default_import_0_to_existing_import_declaration_from_1_90033", "Add default import '{0}' to existing import declaration from \"{1}\""),
            Add_parameter_name: diag(90034, ts.DiagnosticCategory.Message, "Add_parameter_name_90034", "Add parameter name"),
            Declare_private_property_0: diag(90035, ts.DiagnosticCategory.Message, "Declare_private_property_0_90035", "Declare private property '{0}'"),
            Replace_0_with_Promise_1: diag(90036, ts.DiagnosticCategory.Message, "Replace_0_with_Promise_1_90036", "Replace '{0}' with 'Promise<{1}>'"),
            Fix_all_incorrect_return_type_of_an_async_functions: diag(90037, ts.DiagnosticCategory.Message, "Fix_all_incorrect_return_type_of_an_async_functions_90037", "Fix all incorrect return type of an async functions"),
            Declare_private_method_0: diag(90038, ts.DiagnosticCategory.Message, "Declare_private_method_0_90038", "Declare private method '{0}'"),
            Remove_unused_destructuring_declaration: diag(90039, ts.DiagnosticCategory.Message, "Remove_unused_destructuring_declaration_90039", "Remove unused destructuring declaration"),
            Remove_unused_declarations_for_Colon_0: diag(90041, ts.DiagnosticCategory.Message, "Remove_unused_declarations_for_Colon_0_90041", "Remove unused declarations for: '{0}'"),
            Declare_a_private_field_named_0: diag(90053, ts.DiagnosticCategory.Message, "Declare_a_private_field_named_0_90053", "Declare a private field named '{0}'."),
            Convert_function_to_an_ES2015_class: diag(95001, ts.DiagnosticCategory.Message, "Convert_function_to_an_ES2015_class_95001", "Convert function to an ES2015 class"),
            Convert_function_0_to_class: diag(95002, ts.DiagnosticCategory.Message, "Convert_function_0_to_class_95002", "Convert function '{0}' to class"),
            Convert_0_to_1_in_0: diag(95003, ts.DiagnosticCategory.Message, "Convert_0_to_1_in_0_95003", "Convert '{0}' to '{1} in {0}'"),
            Extract_to_0_in_1: diag(95004, ts.DiagnosticCategory.Message, "Extract_to_0_in_1_95004", "Extract to {0} in {1}"),
            Extract_function: diag(95005, ts.DiagnosticCategory.Message, "Extract_function_95005", "Extract function"),
            Extract_constant: diag(95006, ts.DiagnosticCategory.Message, "Extract_constant_95006", "Extract constant"),
            Extract_to_0_in_enclosing_scope: diag(95007, ts.DiagnosticCategory.Message, "Extract_to_0_in_enclosing_scope_95007", "Extract to {0} in enclosing scope"),
            Extract_to_0_in_1_scope: diag(95008, ts.DiagnosticCategory.Message, "Extract_to_0_in_1_scope_95008", "Extract to {0} in {1} scope"),
            Annotate_with_type_from_JSDoc: diag(95009, ts.DiagnosticCategory.Message, "Annotate_with_type_from_JSDoc_95009", "Annotate with type from JSDoc"),
            Annotate_with_types_from_JSDoc: diag(95010, ts.DiagnosticCategory.Message, "Annotate_with_types_from_JSDoc_95010", "Annotate with types from JSDoc"),
            Infer_type_of_0_from_usage: diag(95011, ts.DiagnosticCategory.Message, "Infer_type_of_0_from_usage_95011", "Infer type of '{0}' from usage"),
            Infer_parameter_types_from_usage: diag(95012, ts.DiagnosticCategory.Message, "Infer_parameter_types_from_usage_95012", "Infer parameter types from usage"),
            Convert_to_default_import: diag(95013, ts.DiagnosticCategory.Message, "Convert_to_default_import_95013", "Convert to default import"),
            Install_0: diag(95014, ts.DiagnosticCategory.Message, "Install_0_95014", "Install '{0}'"),
            Replace_import_with_0: diag(95015, ts.DiagnosticCategory.Message, "Replace_import_with_0_95015", "Replace import with '{0}'."),
            Use_synthetic_default_member: diag(95016, ts.DiagnosticCategory.Message, "Use_synthetic_default_member_95016", "Use synthetic 'default' member."),
            Convert_to_ES6_module: diag(95017, ts.DiagnosticCategory.Message, "Convert_to_ES6_module_95017", "Convert to ES6 module"),
            Add_undefined_type_to_property_0: diag(95018, ts.DiagnosticCategory.Message, "Add_undefined_type_to_property_0_95018", "Add 'undefined' type to property '{0}'"),
            Add_initializer_to_property_0: diag(95019, ts.DiagnosticCategory.Message, "Add_initializer_to_property_0_95019", "Add initializer to property '{0}'"),
            Add_definite_assignment_assertion_to_property_0: diag(95020, ts.DiagnosticCategory.Message, "Add_definite_assignment_assertion_to_property_0_95020", "Add definite assignment assertion to property '{0}'"),
            Convert_all_type_literals_to_mapped_type: diag(95021, ts.DiagnosticCategory.Message, "Convert_all_type_literals_to_mapped_type_95021", "Convert all type literals to mapped type"),
            Add_all_missing_members: diag(95022, ts.DiagnosticCategory.Message, "Add_all_missing_members_95022", "Add all missing members"),
            Infer_all_types_from_usage: diag(95023, ts.DiagnosticCategory.Message, "Infer_all_types_from_usage_95023", "Infer all types from usage"),
            Delete_all_unused_declarations: diag(95024, ts.DiagnosticCategory.Message, "Delete_all_unused_declarations_95024", "Delete all unused declarations"),
            Prefix_all_unused_declarations_with_where_possible: diag(95025, ts.DiagnosticCategory.Message, "Prefix_all_unused_declarations_with_where_possible_95025", "Prefix all unused declarations with '_' where possible"),
            Fix_all_detected_spelling_errors: diag(95026, ts.DiagnosticCategory.Message, "Fix_all_detected_spelling_errors_95026", "Fix all detected spelling errors"),
            Add_initializers_to_all_uninitialized_properties: diag(95027, ts.DiagnosticCategory.Message, "Add_initializers_to_all_uninitialized_properties_95027", "Add initializers to all uninitialized properties"),
            Add_definite_assignment_assertions_to_all_uninitialized_properties: diag(95028, ts.DiagnosticCategory.Message, "Add_definite_assignment_assertions_to_all_uninitialized_properties_95028", "Add definite assignment assertions to all uninitialized properties"),
            Add_undefined_type_to_all_uninitialized_properties: diag(95029, ts.DiagnosticCategory.Message, "Add_undefined_type_to_all_uninitialized_properties_95029", "Add undefined type to all uninitialized properties"),
            Change_all_jsdoc_style_types_to_TypeScript: diag(95030, ts.DiagnosticCategory.Message, "Change_all_jsdoc_style_types_to_TypeScript_95030", "Change all jsdoc-style types to TypeScript"),
            Change_all_jsdoc_style_types_to_TypeScript_and_add_undefined_to_nullable_types: diag(95031, ts.DiagnosticCategory.Message, "Change_all_jsdoc_style_types_to_TypeScript_and_add_undefined_to_nullable_types_95031", "Change all jsdoc-style types to TypeScript (and add '| undefined' to nullable types)"),
            Implement_all_unimplemented_interfaces: diag(95032, ts.DiagnosticCategory.Message, "Implement_all_unimplemented_interfaces_95032", "Implement all unimplemented interfaces"),
            Install_all_missing_types_packages: diag(95033, ts.DiagnosticCategory.Message, "Install_all_missing_types_packages_95033", "Install all missing types packages"),
            Rewrite_all_as_indexed_access_types: diag(95034, ts.DiagnosticCategory.Message, "Rewrite_all_as_indexed_access_types_95034", "Rewrite all as indexed access types"),
            Convert_all_to_default_imports: diag(95035, ts.DiagnosticCategory.Message, "Convert_all_to_default_imports_95035", "Convert all to default imports"),
            Make_all_super_calls_the_first_statement_in_their_constructor: diag(95036, ts.DiagnosticCategory.Message, "Make_all_super_calls_the_first_statement_in_their_constructor_95036", "Make all 'super()' calls the first statement in their constructor"),
            Add_qualifier_to_all_unresolved_variables_matching_a_member_name: diag(95037, ts.DiagnosticCategory.Message, "Add_qualifier_to_all_unresolved_variables_matching_a_member_name_95037", "Add qualifier to all unresolved variables matching a member name"),
            Change_all_extended_interfaces_to_implements: diag(95038, ts.DiagnosticCategory.Message, "Change_all_extended_interfaces_to_implements_95038", "Change all extended interfaces to 'implements'"),
            Add_all_missing_super_calls: diag(95039, ts.DiagnosticCategory.Message, "Add_all_missing_super_calls_95039", "Add all missing super calls"),
            Implement_all_inherited_abstract_classes: diag(95040, ts.DiagnosticCategory.Message, "Implement_all_inherited_abstract_classes_95040", "Implement all inherited abstract classes"),
            Add_all_missing_async_modifiers: diag(95041, ts.DiagnosticCategory.Message, "Add_all_missing_async_modifiers_95041", "Add all missing 'async' modifiers"),
            Add_ts_ignore_to_all_error_messages: diag(95042, ts.DiagnosticCategory.Message, "Add_ts_ignore_to_all_error_messages_95042", "Add '@ts-ignore' to all error messages"),
            Annotate_everything_with_types_from_JSDoc: diag(95043, ts.DiagnosticCategory.Message, "Annotate_everything_with_types_from_JSDoc_95043", "Annotate everything with types from JSDoc"),
            Add_to_all_uncalled_decorators: diag(95044, ts.DiagnosticCategory.Message, "Add_to_all_uncalled_decorators_95044", "Add '()' to all uncalled decorators"),
            Convert_all_constructor_functions_to_classes: diag(95045, ts.DiagnosticCategory.Message, "Convert_all_constructor_functions_to_classes_95045", "Convert all constructor functions to classes"),
            Generate_get_and_set_accessors: diag(95046, ts.DiagnosticCategory.Message, "Generate_get_and_set_accessors_95046", "Generate 'get' and 'set' accessors"),
            Convert_require_to_import: diag(95047, ts.DiagnosticCategory.Message, "Convert_require_to_import_95047", "Convert 'require' to 'import'"),
            Convert_all_require_to_import: diag(95048, ts.DiagnosticCategory.Message, "Convert_all_require_to_import_95048", "Convert all 'require' to 'import'"),
            Move_to_a_new_file: diag(95049, ts.DiagnosticCategory.Message, "Move_to_a_new_file_95049", "Move to a new file"),
            Remove_unreachable_code: diag(95050, ts.DiagnosticCategory.Message, "Remove_unreachable_code_95050", "Remove unreachable code"),
            Remove_all_unreachable_code: diag(95051, ts.DiagnosticCategory.Message, "Remove_all_unreachable_code_95051", "Remove all unreachable code"),
            Add_missing_typeof: diag(95052, ts.DiagnosticCategory.Message, "Add_missing_typeof_95052", "Add missing 'typeof'"),
            Remove_unused_label: diag(95053, ts.DiagnosticCategory.Message, "Remove_unused_label_95053", "Remove unused label"),
            Remove_all_unused_labels: diag(95054, ts.DiagnosticCategory.Message, "Remove_all_unused_labels_95054", "Remove all unused labels"),
            Convert_0_to_mapped_object_type: diag(95055, ts.DiagnosticCategory.Message, "Convert_0_to_mapped_object_type_95055", "Convert '{0}' to mapped object type"),
            Convert_namespace_import_to_named_imports: diag(95056, ts.DiagnosticCategory.Message, "Convert_namespace_import_to_named_imports_95056", "Convert namespace import to named imports"),
            Convert_named_imports_to_namespace_import: diag(95057, ts.DiagnosticCategory.Message, "Convert_named_imports_to_namespace_import_95057", "Convert named imports to namespace import"),
            Add_or_remove_braces_in_an_arrow_function: diag(95058, ts.DiagnosticCategory.Message, "Add_or_remove_braces_in_an_arrow_function_95058", "Add or remove braces in an arrow function"),
            Add_braces_to_arrow_function: diag(95059, ts.DiagnosticCategory.Message, "Add_braces_to_arrow_function_95059", "Add braces to arrow function"),
            Remove_braces_from_arrow_function: diag(95060, ts.DiagnosticCategory.Message, "Remove_braces_from_arrow_function_95060", "Remove braces from arrow function"),
            Convert_default_export_to_named_export: diag(95061, ts.DiagnosticCategory.Message, "Convert_default_export_to_named_export_95061", "Convert default export to named export"),
            Convert_named_export_to_default_export: diag(95062, ts.DiagnosticCategory.Message, "Convert_named_export_to_default_export_95062", "Convert named export to default export"),
            Add_missing_enum_member_0: diag(95063, ts.DiagnosticCategory.Message, "Add_missing_enum_member_0_95063", "Add missing enum member '{0}'"),
            Add_all_missing_imports: diag(95064, ts.DiagnosticCategory.Message, "Add_all_missing_imports_95064", "Add all missing imports"),
            Convert_to_async_function: diag(95065, ts.DiagnosticCategory.Message, "Convert_to_async_function_95065", "Convert to async function"),
            Convert_all_to_async_functions: diag(95066, ts.DiagnosticCategory.Message, "Convert_all_to_async_functions_95066", "Convert all to async functions"),
            Add_missing_call_parentheses: diag(95067, ts.DiagnosticCategory.Message, "Add_missing_call_parentheses_95067", "Add missing call parentheses"),
            Add_all_missing_call_parentheses: diag(95068, ts.DiagnosticCategory.Message, "Add_all_missing_call_parentheses_95068", "Add all missing call parentheses"),
            Add_unknown_conversion_for_non_overlapping_types: diag(95069, ts.DiagnosticCategory.Message, "Add_unknown_conversion_for_non_overlapping_types_95069", "Add 'unknown' conversion for non-overlapping types"),
            Add_unknown_to_all_conversions_of_non_overlapping_types: diag(95070, ts.DiagnosticCategory.Message, "Add_unknown_to_all_conversions_of_non_overlapping_types_95070", "Add 'unknown' to all conversions of non-overlapping types"),
            Add_missing_new_operator_to_call: diag(95071, ts.DiagnosticCategory.Message, "Add_missing_new_operator_to_call_95071", "Add missing 'new' operator to call"),
            Add_missing_new_operator_to_all_calls: diag(95072, ts.DiagnosticCategory.Message, "Add_missing_new_operator_to_all_calls_95072", "Add missing 'new' operator to all calls"),
            Add_names_to_all_parameters_without_names: diag(95073, ts.DiagnosticCategory.Message, "Add_names_to_all_parameters_without_names_95073", "Add names to all parameters without names"),
            Enable_the_experimentalDecorators_option_in_your_configuration_file: diag(95074, ts.DiagnosticCategory.Message, "Enable_the_experimentalDecorators_option_in_your_configuration_file_95074", "Enable the 'experimentalDecorators' option in your configuration file"),
            Convert_parameters_to_destructured_object: diag(95075, ts.DiagnosticCategory.Message, "Convert_parameters_to_destructured_object_95075", "Convert parameters to destructured object"),
            Extract_type: diag(95077, ts.DiagnosticCategory.Message, "Extract_type_95077", "Extract type"),
            Extract_to_type_alias: diag(95078, ts.DiagnosticCategory.Message, "Extract_to_type_alias_95078", "Extract to type alias"),
            Extract_to_typedef: diag(95079, ts.DiagnosticCategory.Message, "Extract_to_typedef_95079", "Extract to typedef"),
            Infer_this_type_of_0_from_usage: diag(95080, ts.DiagnosticCategory.Message, "Infer_this_type_of_0_from_usage_95080", "Infer 'this' type of '{0}' from usage"),
            Add_const_to_unresolved_variable: diag(95081, ts.DiagnosticCategory.Message, "Add_const_to_unresolved_variable_95081", "Add 'const' to unresolved variable"),
            Add_const_to_all_unresolved_variables: diag(95082, ts.DiagnosticCategory.Message, "Add_const_to_all_unresolved_variables_95082", "Add 'const' to all unresolved variables"),
            Add_await: diag(95083, ts.DiagnosticCategory.Message, "Add_await_95083", "Add 'await'"),
            Add_await_to_initializer_for_0: diag(95084, ts.DiagnosticCategory.Message, "Add_await_to_initializer_for_0_95084", "Add 'await' to initializer for '{0}'"),
            Fix_all_expressions_possibly_missing_await: diag(95085, ts.DiagnosticCategory.Message, "Fix_all_expressions_possibly_missing_await_95085", "Fix all expressions possibly missing 'await'"),
            Remove_unnecessary_await: diag(95086, ts.DiagnosticCategory.Message, "Remove_unnecessary_await_95086", "Remove unnecessary 'await'"),
            Remove_all_unnecessary_uses_of_await: diag(95087, ts.DiagnosticCategory.Message, "Remove_all_unnecessary_uses_of_await_95087", "Remove all unnecessary uses of 'await'"),
            Enable_the_jsx_flag_in_your_configuration_file: diag(95088, ts.DiagnosticCategory.Message, "Enable_the_jsx_flag_in_your_configuration_file_95088", "Enable the '--jsx' flag in your configuration file"),
            Add_await_to_initializers: diag(95089, ts.DiagnosticCategory.Message, "Add_await_to_initializers_95089", "Add 'await' to initializers"),
            Extract_to_interface: diag(95090, ts.DiagnosticCategory.Message, "Extract_to_interface_95090", "Extract to interface"),
            Convert_to_a_bigint_numeric_literal: diag(95091, ts.DiagnosticCategory.Message, "Convert_to_a_bigint_numeric_literal_95091", "Convert to a bigint numeric literal"),
            Convert_all_to_bigint_numeric_literals: diag(95092, ts.DiagnosticCategory.Message, "Convert_all_to_bigint_numeric_literals_95092", "Convert all to bigint numeric literals"),
            Convert_const_to_let: diag(95093, ts.DiagnosticCategory.Message, "Convert_const_to_let_95093", "Convert 'const' to 'let'"),
            Prefix_with_declare: diag(95094, ts.DiagnosticCategory.Message, "Prefix_with_declare_95094", "Prefix with 'declare'"),
            Prefix_all_incorrect_property_declarations_with_declare: diag(95095, ts.DiagnosticCategory.Message, "Prefix_all_incorrect_property_declarations_with_declare_95095", "Prefix all incorrect property declarations with 'declare'"),
            Convert_to_template_string: diag(95096, ts.DiagnosticCategory.Message, "Convert_to_template_string_95096", "Convert to template string"),
            Add_export_to_make_this_file_into_a_module: diag(95097, ts.DiagnosticCategory.Message, "Add_export_to_make_this_file_into_a_module_95097", "Add 'export {}' to make this file into a module"),
            Set_the_target_option_in_your_configuration_file_to_0: diag(95098, ts.DiagnosticCategory.Message, "Set_the_target_option_in_your_configuration_file_to_0_95098", "Set the 'target' option in your configuration file to '{0}'"),
            Set_the_module_option_in_your_configuration_file_to_0: diag(95099, ts.DiagnosticCategory.Message, "Set_the_module_option_in_your_configuration_file_to_0_95099", "Set the 'module' option in your configuration file to '{0}'"),
            Convert_invalid_character_to_its_html_entity_code: diag(95100, ts.DiagnosticCategory.Message, "Convert_invalid_character_to_its_html_entity_code_95100", "Convert invalid character to its html entity code"),
            Convert_all_invalid_characters_to_HTML_entity_code: diag(95101, ts.DiagnosticCategory.Message, "Convert_all_invalid_characters_to_HTML_entity_code_95101", "Convert all invalid characters to HTML entity code"),
            Add_class_tag: diag(95102, ts.DiagnosticCategory.Message, "Add_class_tag_95102", "Add '@class' tag"),
            Add_this_tag: diag(95103, ts.DiagnosticCategory.Message, "Add_this_tag_95103", "Add '@this' tag"),
            Add_this_parameter: diag(95104, ts.DiagnosticCategory.Message, "Add_this_parameter_95104", "Add 'this' parameter."),
            Convert_function_expression_0_to_arrow_function: diag(95105, ts.DiagnosticCategory.Message, "Convert_function_expression_0_to_arrow_function_95105", "Convert function expression '{0}' to arrow function"),
            Convert_function_declaration_0_to_arrow_function: diag(95106, ts.DiagnosticCategory.Message, "Convert_function_declaration_0_to_arrow_function_95106", "Convert function declaration '{0}' to arrow function"),
            Fix_all_implicit_this_errors: diag(95107, ts.DiagnosticCategory.Message, "Fix_all_implicit_this_errors_95107", "Fix all implicit-'this' errors"),
            Wrap_invalid_character_in_an_expression_container: diag(95108, ts.DiagnosticCategory.Message, "Wrap_invalid_character_in_an_expression_container_95108", "Wrap invalid character in an expression container"),
            Wrap_all_invalid_characters_in_an_expression_container: diag(95109, ts.DiagnosticCategory.Message, "Wrap_all_invalid_characters_in_an_expression_container_95109", "Wrap all invalid characters in an expression container"),
            Visit_https_Colon_Slash_Slashaka_ms_Slashtsconfig_json_to_read_more_about_this_file: diag(95110, ts.DiagnosticCategory.Message, "Visit_https_Colon_Slash_Slashaka_ms_Slashtsconfig_json_to_read_more_about_this_file_95110", "Visit https://aka.ms/tsconfig.json to read more about this file"),
            Add_a_return_statement: diag(95111, ts.DiagnosticCategory.Message, "Add_a_return_statement_95111", "Add a return statement"),
            Remove_braces_from_arrow_function_body: diag(95112, ts.DiagnosticCategory.Message, "Remove_braces_from_arrow_function_body_95112", "Remove braces from arrow function body"),
            Wrap_the_following_body_with_parentheses_which_should_be_an_object_literal: diag(95113, ts.DiagnosticCategory.Message, "Wrap_the_following_body_with_parentheses_which_should_be_an_object_literal_95113", "Wrap the following body with parentheses which should be an object literal"),
            Add_all_missing_return_statement: diag(95114, ts.DiagnosticCategory.Message, "Add_all_missing_return_statement_95114", "Add all missing return statement"),
            Remove_braces_from_all_arrow_function_bodies_with_relevant_issues: diag(95115, ts.DiagnosticCategory.Message, "Remove_braces_from_all_arrow_function_bodies_with_relevant_issues_95115", "Remove braces from all arrow function bodies with relevant issues"),
            Wrap_all_object_literal_with_parentheses: diag(95116, ts.DiagnosticCategory.Message, "Wrap_all_object_literal_with_parentheses_95116", "Wrap all object literal with parentheses"),
            Move_labeled_tuple_element_modifiers_to_labels: diag(95117, ts.DiagnosticCategory.Message, "Move_labeled_tuple_element_modifiers_to_labels_95117", "Move labeled tuple element modifiers to labels"),
            Convert_overload_list_to_single_signature: diag(95118, ts.DiagnosticCategory.Message, "Convert_overload_list_to_single_signature_95118", "Convert overload list to single signature"),
            Generate_get_and_set_accessors_for_all_overriding_properties: diag(95119, ts.DiagnosticCategory.Message, "Generate_get_and_set_accessors_for_all_overriding_properties_95119", "Generate 'get' and 'set' accessors for all overriding properties"),
            Wrap_in_JSX_fragment: diag(95120, ts.DiagnosticCategory.Message, "Wrap_in_JSX_fragment_95120", "Wrap in JSX fragment"),
            Wrap_all_unparented_JSX_in_JSX_fragment: diag(95121, ts.DiagnosticCategory.Message, "Wrap_all_unparented_JSX_in_JSX_fragment_95121", "Wrap all unparented JSX in JSX fragment"),
            Convert_arrow_function_or_function_expression: diag(95122, ts.DiagnosticCategory.Message, "Convert_arrow_function_or_function_expression_95122", "Convert arrow function or function expression"),
            Convert_to_anonymous_function: diag(95123, ts.DiagnosticCategory.Message, "Convert_to_anonymous_function_95123", "Convert to anonymous function"),
            Convert_to_named_function: diag(95124, ts.DiagnosticCategory.Message, "Convert_to_named_function_95124", "Convert to named function"),
            Convert_to_arrow_function: diag(95125, ts.DiagnosticCategory.Message, "Convert_to_arrow_function_95125", "Convert to arrow function"),
            Remove_parentheses: diag(95126, ts.DiagnosticCategory.Message, "Remove_parentheses_95126", "Remove parentheses"),
            Could_not_find_a_containing_arrow_function: diag(95127, ts.DiagnosticCategory.Message, "Could_not_find_a_containing_arrow_function_95127", "Could not find a containing arrow function"),
            Containing_function_is_not_an_arrow_function: diag(95128, ts.DiagnosticCategory.Message, "Containing_function_is_not_an_arrow_function_95128", "Containing function is not an arrow function"),
            Could_not_find_export_statement: diag(95129, ts.DiagnosticCategory.Message, "Could_not_find_export_statement_95129", "Could not find export statement"),
            This_file_already_has_a_default_export: diag(95130, ts.DiagnosticCategory.Message, "This_file_already_has_a_default_export_95130", "This file already has a default export"),
            Could_not_find_import_clause: diag(95131, ts.DiagnosticCategory.Message, "Could_not_find_import_clause_95131", "Could not find import clause"),
            Could_not_find_namespace_import_or_named_imports: diag(95132, ts.DiagnosticCategory.Message, "Could_not_find_namespace_import_or_named_imports_95132", "Could not find namespace import or named imports"),
            Selection_is_not_a_valid_type_node: diag(95133, ts.DiagnosticCategory.Message, "Selection_is_not_a_valid_type_node_95133", "Selection is not a valid type node"),
            No_type_could_be_extracted_from_this_type_node: diag(95134, ts.DiagnosticCategory.Message, "No_type_could_be_extracted_from_this_type_node_95134", "No type could be extracted from this type node"),
            Could_not_find_property_for_which_to_generate_accessor: diag(95135, ts.DiagnosticCategory.Message, "Could_not_find_property_for_which_to_generate_accessor_95135", "Could not find property for which to generate accessor"),
            Name_is_not_valid: diag(95136, ts.DiagnosticCategory.Message, "Name_is_not_valid_95136", "Name is not valid"),
            Can_only_convert_property_with_modifier: diag(95137, ts.DiagnosticCategory.Message, "Can_only_convert_property_with_modifier_95137", "Can only convert property with modifier"),
            Switch_each_misused_0_to_1: diag(95138, ts.DiagnosticCategory.Message, "Switch_each_misused_0_to_1_95138", "Switch each misused '{0}' to '{1}'"),
            Convert_to_optional_chain_expression: diag(95139, ts.DiagnosticCategory.Message, "Convert_to_optional_chain_expression_95139", "Convert to optional chain expression"),
            Could_not_find_convertible_access_expression: diag(95140, ts.DiagnosticCategory.Message, "Could_not_find_convertible_access_expression_95140", "Could not find convertible access expression"),
            Could_not_find_matching_access_expressions: diag(95141, ts.DiagnosticCategory.Message, "Could_not_find_matching_access_expressions_95141", "Could not find matching access expressions"),
            Can_only_convert_logical_AND_access_chains: diag(95142, ts.DiagnosticCategory.Message, "Can_only_convert_logical_AND_access_chains_95142", "Can only convert logical AND access chains"),
            Add_void_to_Promise_resolved_without_a_value: diag(95143, ts.DiagnosticCategory.Message, "Add_void_to_Promise_resolved_without_a_value_95143", "Add 'void' to Promise resolved without a value"),
            Add_void_to_all_Promises_resolved_without_a_value: diag(95144, ts.DiagnosticCategory.Message, "Add_void_to_all_Promises_resolved_without_a_value_95144", "Add 'void' to all Promises resolved without a value"),
            Use_element_access_for_0: diag(95145, ts.DiagnosticCategory.Message, "Use_element_access_for_0_95145", "Use element access for '{0}'"),
            Use_element_access_for_all_undeclared_properties: diag(95146, ts.DiagnosticCategory.Message, "Use_element_access_for_all_undeclared_properties_95146", "Use element access for all undeclared properties."),
            Delete_all_unused_imports: diag(95147, ts.DiagnosticCategory.Message, "Delete_all_unused_imports_95147", "Delete all unused imports"),
            Infer_function_return_type: diag(95148, ts.DiagnosticCategory.Message, "Infer_function_return_type_95148", "Infer function return type"),
            Return_type_must_be_inferred_from_a_function: diag(95149, ts.DiagnosticCategory.Message, "Return_type_must_be_inferred_from_a_function_95149", "Return type must be inferred from a function"),
            Could_not_determine_function_return_type: diag(95150, ts.DiagnosticCategory.Message, "Could_not_determine_function_return_type_95150", "Could not determine function return type"),
            Could_not_convert_to_arrow_function: diag(95151, ts.DiagnosticCategory.Message, "Could_not_convert_to_arrow_function_95151", "Could not convert to arrow function"),
            Could_not_convert_to_named_function: diag(95152, ts.DiagnosticCategory.Message, "Could_not_convert_to_named_function_95152", "Could not convert to named function"),
            Could_not_convert_to_anonymous_function: diag(95153, ts.DiagnosticCategory.Message, "Could_not_convert_to_anonymous_function_95153", "Could not convert to anonymous function"),
            Can_only_convert_string_concatenation: diag(95154, ts.DiagnosticCategory.Message, "Can_only_convert_string_concatenation_95154", "Can only convert string concatenation"),
            Selection_is_not_a_valid_statement_or_statements: diag(95155, ts.DiagnosticCategory.Message, "Selection_is_not_a_valid_statement_or_statements_95155", "Selection is not a valid statement or statements"),
            Add_missing_function_declaration_0: diag(95156, ts.DiagnosticCategory.Message, "Add_missing_function_declaration_0_95156", "Add missing function declaration '{0}'"),
            Add_all_missing_function_declarations: diag(95157, ts.DiagnosticCategory.Message, "Add_all_missing_function_declarations_95157", "Add all missing function declarations"),
            Method_not_implemented: diag(95158, ts.DiagnosticCategory.Message, "Method_not_implemented_95158", "Method not implemented."),
            Function_not_implemented: diag(95159, ts.DiagnosticCategory.Message, "Function_not_implemented_95159", "Function not implemented."),
            Add_override_modifier: diag(95160, ts.DiagnosticCategory.Message, "Add_override_modifier_95160", "Add 'override' modifier"),
            Remove_override_modifier: diag(95161, ts.DiagnosticCategory.Message, "Remove_override_modifier_95161", "Remove 'override' modifier"),
            Add_all_missing_override_modifiers: diag(95162, ts.DiagnosticCategory.Message, "Add_all_missing_override_modifiers_95162", "Add all missing 'override' modifiers"),
            Remove_all_unnecessary_override_modifiers: diag(95163, ts.DiagnosticCategory.Message, "Remove_all_unnecessary_override_modifiers_95163", "Remove all unnecessary 'override' modifiers"),
            Can_only_convert_named_export: diag(95164, ts.DiagnosticCategory.Message, "Can_only_convert_named_export_95164", "Can only convert named export"),
            Add_missing_properties: diag(95165, ts.DiagnosticCategory.Message, "Add_missing_properties_95165", "Add missing properties"),
            Add_all_missing_properties: diag(95166, ts.DiagnosticCategory.Message, "Add_all_missing_properties_95166", "Add all missing properties"),
            Add_missing_attributes: diag(95167, ts.DiagnosticCategory.Message, "Add_missing_attributes_95167", "Add missing attributes"),
            Add_all_missing_attributes: diag(95168, ts.DiagnosticCategory.Message, "Add_all_missing_attributes_95168", "Add all missing attributes"),
            No_value_exists_in_scope_for_the_shorthand_property_0_Either_declare_one_or_provide_an_initializer: diag(18004, ts.DiagnosticCategory.Error, "No_value_exists_in_scope_for_the_shorthand_property_0_Either_declare_one_or_provide_an_initializer_18004", "No value exists in scope for the shorthand property '{0}'. Either declare one or provide an initializer."),
            Classes_may_not_have_a_field_named_constructor: diag(18006, ts.DiagnosticCategory.Error, "Classes_may_not_have_a_field_named_constructor_18006", "Classes may not have a field named 'constructor'."),
            JSX_expressions_may_not_use_the_comma_operator_Did_you_mean_to_write_an_array: diag(18007, ts.DiagnosticCategory.Error, "JSX_expressions_may_not_use_the_comma_operator_Did_you_mean_to_write_an_array_18007", "JSX expressions may not use the comma operator. Did you mean to write an array?"),
            Private_identifiers_cannot_be_used_as_parameters: diag(18009, ts.DiagnosticCategory.Error, "Private_identifiers_cannot_be_used_as_parameters_18009", "Private identifiers cannot be used as parameters."),
            An_accessibility_modifier_cannot_be_used_with_a_private_identifier: diag(18010, ts.DiagnosticCategory.Error, "An_accessibility_modifier_cannot_be_used_with_a_private_identifier_18010", "An accessibility modifier cannot be used with a private identifier."),
            The_operand_of_a_delete_operator_cannot_be_a_private_identifier: diag(18011, ts.DiagnosticCategory.Error, "The_operand_of_a_delete_operator_cannot_be_a_private_identifier_18011", "The operand of a 'delete' operator cannot be a private identifier."),
            constructor_is_a_reserved_word: diag(18012, ts.DiagnosticCategory.Error, "constructor_is_a_reserved_word_18012", "'#constructor' is a reserved word."),
            Property_0_is_not_accessible_outside_class_1_because_it_has_a_private_identifier: diag(18013, ts.DiagnosticCategory.Error, "Property_0_is_not_accessible_outside_class_1_because_it_has_a_private_identifier_18013", "Property '{0}' is not accessible outside class '{1}' because it has a private identifier."),
            The_property_0_cannot_be_accessed_on_type_1_within_this_class_because_it_is_shadowed_by_another_private_identifier_with_the_same_spelling: diag(18014, ts.DiagnosticCategory.Error, "The_property_0_cannot_be_accessed_on_type_1_within_this_class_because_it_is_shadowed_by_another_priv_18014", "The property '{0}' cannot be accessed on type '{1}' within this class because it is shadowed by another private identifier with the same spelling."),
            Property_0_in_type_1_refers_to_a_different_member_that_cannot_be_accessed_from_within_type_2: diag(18015, ts.DiagnosticCategory.Error, "Property_0_in_type_1_refers_to_a_different_member_that_cannot_be_accessed_from_within_type_2_18015", "Property '{0}' in type '{1}' refers to a different member that cannot be accessed from within type '{2}'."),
            Private_identifiers_are_not_allowed_outside_class_bodies: diag(18016, ts.DiagnosticCategory.Error, "Private_identifiers_are_not_allowed_outside_class_bodies_18016", "Private identifiers are not allowed outside class bodies."),
            The_shadowing_declaration_of_0_is_defined_here: diag(18017, ts.DiagnosticCategory.Error, "The_shadowing_declaration_of_0_is_defined_here_18017", "The shadowing declaration of '{0}' is defined here"),
            The_declaration_of_0_that_you_probably_intended_to_use_is_defined_here: diag(18018, ts.DiagnosticCategory.Error, "The_declaration_of_0_that_you_probably_intended_to_use_is_defined_here_18018", "The declaration of '{0}' that you probably intended to use is defined here"),
            _0_modifier_cannot_be_used_with_a_private_identifier: diag(18019, ts.DiagnosticCategory.Error, "_0_modifier_cannot_be_used_with_a_private_identifier_18019", "'{0}' modifier cannot be used with a private identifier."),
            An_enum_member_cannot_be_named_with_a_private_identifier: diag(18024, ts.DiagnosticCategory.Error, "An_enum_member_cannot_be_named_with_a_private_identifier_18024", "An enum member cannot be named with a private identifier."),
            can_only_be_used_at_the_start_of_a_file: diag(18026, ts.DiagnosticCategory.Error, "can_only_be_used_at_the_start_of_a_file_18026", "'#!' can only be used at the start of a file."),
            Compiler_reserves_name_0_when_emitting_private_identifier_downlevel: diag(18027, ts.DiagnosticCategory.Error, "Compiler_reserves_name_0_when_emitting_private_identifier_downlevel_18027", "Compiler reserves name '{0}' when emitting private identifier downlevel."),
            Private_identifiers_are_only_available_when_targeting_ECMAScript_2015_and_higher: diag(18028, ts.DiagnosticCategory.Error, "Private_identifiers_are_only_available_when_targeting_ECMAScript_2015_and_higher_18028", "Private identifiers are only available when targeting ECMAScript 2015 and higher."),
            Private_identifiers_are_not_allowed_in_variable_declarations: diag(18029, ts.DiagnosticCategory.Error, "Private_identifiers_are_not_allowed_in_variable_declarations_18029", "Private identifiers are not allowed in variable declarations."),
            An_optional_chain_cannot_contain_private_identifiers: diag(18030, ts.DiagnosticCategory.Error, "An_optional_chain_cannot_contain_private_identifiers_18030", "An optional chain cannot contain private identifiers."),
            The_intersection_0_was_reduced_to_never_because_property_1_has_conflicting_types_in_some_constituents: diag(18031, ts.DiagnosticCategory.Error, "The_intersection_0_was_reduced_to_never_because_property_1_has_conflicting_types_in_some_constituent_18031", "The intersection '{0}' was reduced to 'never' because property '{1}' has conflicting types in some constituents."),
            The_intersection_0_was_reduced_to_never_because_property_1_exists_in_multiple_constituents_and_is_private_in_some: diag(18032, ts.DiagnosticCategory.Error, "The_intersection_0_was_reduced_to_never_because_property_1_exists_in_multiple_constituents_and_is_pr_18032", "The intersection '{0}' was reduced to 'never' because property '{1}' exists in multiple constituents and is private in some."),
            Only_numeric_enums_can_have_computed_members_but_this_expression_has_type_0_If_you_do_not_need_exhaustiveness_checks_consider_using_an_object_literal_instead: diag(18033, ts.DiagnosticCategory.Error, "Only_numeric_enums_can_have_computed_members_but_this_expression_has_type_0_If_you_do_not_need_exhau_18033", "Only numeric enums can have computed members, but this expression has type '{0}'. If you do not need exhaustiveness checks, consider using an object literal instead."),
            Specify_the_JSX_fragment_factory_function_to_use_when_targeting_react_JSX_emit_with_jsxFactory_compiler_option_is_specified_e_g_Fragment: diag(18034, ts.DiagnosticCategory.Message, "Specify_the_JSX_fragment_factory_function_to_use_when_targeting_react_JSX_emit_with_jsxFactory_compi_18034", "Specify the JSX fragment factory function to use when targeting 'react' JSX emit with 'jsxFactory' compiler option is specified, e.g. 'Fragment'."),
            Invalid_value_for_jsxFragmentFactory_0_is_not_a_valid_identifier_or_qualified_name: diag(18035, ts.DiagnosticCategory.Error, "Invalid_value_for_jsxFragmentFactory_0_is_not_a_valid_identifier_or_qualified_name_18035", "Invalid value for 'jsxFragmentFactory'. '{0}' is not a valid identifier or qualified-name."),
            Class_decorators_can_t_be_used_with_static_private_identifier_Consider_removing_the_experimental_decorator: diag(18036, ts.DiagnosticCategory.Error, "Class_decorators_can_t_be_used_with_static_private_identifier_Consider_removing_the_experimental_dec_18036", "Class decorators can't be used with static private identifier. Consider removing the experimental decorator."),
            Await_expression_cannot_be_used_inside_a_class_static_block: diag(18037, ts.DiagnosticCategory.Error, "Await_expression_cannot_be_used_inside_a_class_static_block_18037", "Await expression cannot be used inside a class static block."),
            For_await_loops_cannot_be_used_inside_a_class_static_block: diag(18038, ts.DiagnosticCategory.Error, "For_await_loops_cannot_be_used_inside_a_class_static_block_18038", "'For await' loops cannot be used inside a class static block."),
            Invalid_use_of_0_It_cannot_be_used_inside_a_class_static_block: diag(18039, ts.DiagnosticCategory.Error, "Invalid_use_of_0_It_cannot_be_used_inside_a_class_static_block_18039", "Invalid use of '{0}'. It cannot be used inside a class static block."),
            A_return_statement_cannot_be_used_inside_a_class_static_block: diag(18041, ts.DiagnosticCategory.Error, "A_return_statement_cannot_be_used_inside_a_class_static_block_18041", "A 'return' statement cannot be used inside a class static block."),
        };
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        var _a;
        /* @internal */
        function tokenIsIdentifierOrKeyword(token) {
            return token >= 79 /* Identifier */;
        }
        ts.tokenIsIdentifierOrKeyword = tokenIsIdentifierOrKeyword;
        /* @internal */
        function tokenIsIdentifierOrKeywordOrGreaterThan(token) {
            return token === 31 /* GreaterThanToken */ || tokenIsIdentifierOrKeyword(token);
        }
        ts.tokenIsIdentifierOrKeywordOrGreaterThan = tokenIsIdentifierOrKeywordOrGreaterThan;
        /** @internal */
        ts.textToKeywordObj = (_a = {
            abstract: 126 /* AbstractKeyword */,
            any: 129 /* AnyKeyword */,
            as: 127 /* AsKeyword */,
            asserts: 128 /* AssertsKeyword */,
            bigint: 156 /* BigIntKeyword */,
            boolean: 132 /* BooleanKeyword */,
            break: 81 /* BreakKeyword */,
            case: 82 /* CaseKeyword */,
            catch: 83 /* CatchKeyword */,
            class: 84 /* ClassKeyword */,
            continue: 86 /* ContinueKeyword */,
            const: 85 /* ConstKeyword */
        },
            _a["" + "constructor"] = 133 /* ConstructorKeyword */,
            _a.debugger = 87 /* DebuggerKeyword */,
            _a.declare = 134 /* DeclareKeyword */,
            _a.default = 88 /* DefaultKeyword */,
            _a.delete = 89 /* DeleteKeyword */,
            _a.do = 90 /* DoKeyword */,
            _a.else = 91 /* ElseKeyword */,
            _a.enum = 92 /* EnumKeyword */,
            _a.export = 93 /* ExportKeyword */,
            _a.extends = 94 /* ExtendsKeyword */,
            _a.false = 95 /* FalseKeyword */,
            _a.finally = 96 /* FinallyKeyword */,
            _a.for = 97 /* ForKeyword */,
            _a.from = 154 /* FromKeyword */,
            _a.function = 98 /* FunctionKeyword */,
            _a.get = 135 /* GetKeyword */,
            _a.if = 99 /* IfKeyword */,
            _a.implements = 117 /* ImplementsKeyword */,
            _a.import = 100 /* ImportKeyword */,
            _a.in = 101 /* InKeyword */,
            _a.infer = 136 /* InferKeyword */,
            _a.instanceof = 102 /* InstanceOfKeyword */,
            _a.interface = 118 /* InterfaceKeyword */,
            _a.intrinsic = 137 /* IntrinsicKeyword */,
            _a.is = 138 /* IsKeyword */,
            _a.keyof = 139 /* KeyOfKeyword */,
            _a.let = 119 /* LetKeyword */,
            _a.module = 140 /* ModuleKeyword */,
            _a.namespace = 141 /* NamespaceKeyword */,
            _a.never = 142 /* NeverKeyword */,
            _a.new = 103 /* NewKeyword */,
            _a.null = 104 /* NullKeyword */,
            _a.number = 145 /* NumberKeyword */,
            _a.object = 146 /* ObjectKeyword */,
            _a.package = 120 /* PackageKeyword */,
            _a.private = 121 /* PrivateKeyword */,
            _a.protected = 122 /* ProtectedKeyword */,
            _a.public = 123 /* PublicKeyword */,
            _a.override = 157 /* OverrideKeyword */,
            _a.readonly = 143 /* ReadonlyKeyword */,
            _a.require = 144 /* RequireKeyword */,
            _a.global = 155 /* GlobalKeyword */,
            _a.return = 105 /* ReturnKeyword */,
            _a.set = 147 /* SetKeyword */,
            _a.static = 124 /* StaticKeyword */,
            _a.string = 148 /* StringKeyword */,
            _a.super = 106 /* SuperKeyword */,
            _a.switch = 107 /* SwitchKeyword */,
            _a.symbol = 149 /* SymbolKeyword */,
            _a.this = 108 /* ThisKeyword */,
            _a.throw = 109 /* ThrowKeyword */,
            _a.true = 110 /* TrueKeyword */,
            _a.try = 111 /* TryKeyword */,
            _a.type = 150 /* TypeKeyword */,
            _a.typeof = 112 /* TypeOfKeyword */,
            _a.undefined = 151 /* UndefinedKeyword */,
            _a.unique = 152 /* UniqueKeyword */,
            _a.unknown = 153 /* UnknownKeyword */,
            _a.var = 113 /* VarKeyword */,
            _a.void = 114 /* VoidKeyword */,
            _a.while = 115 /* WhileKeyword */,
            _a.with = 116 /* WithKeyword */,
            _a.yield = 125 /* YieldKeyword */,
            _a.async = 130 /* AsyncKeyword */,
            _a.await = 131 /* AwaitKeyword */,
            _a.of = 158 /* OfKeyword */,
            _a);
        var textToKeyword = new ts.Map(ts.getEntries(ts.textToKeywordObj));
        var textToToken = new ts.Map(ts.getEntries(__assign(__assign({}, ts.textToKeywordObj), { "{": 18 /* OpenBraceToken */, "}": 19 /* CloseBraceToken */, "(": 20 /* OpenParenToken */, ")": 21 /* CloseParenToken */, "[": 22 /* OpenBracketToken */, "]": 23 /* CloseBracketToken */, ".": 24 /* DotToken */, "...": 25 /* DotDotDotToken */, ";": 26 /* SemicolonToken */, ",": 27 /* CommaToken */, "<": 29 /* LessThanToken */, ">": 31 /* GreaterThanToken */, "<=": 32 /* LessThanEqualsToken */, ">=": 33 /* GreaterThanEqualsToken */, "==": 34 /* EqualsEqualsToken */, "!=": 35 /* ExclamationEqualsToken */, "===": 36 /* EqualsEqualsEqualsToken */, "!==": 37 /* ExclamationEqualsEqualsToken */, "=>": 38 /* EqualsGreaterThanToken */, "+": 39 /* PlusToken */, "-": 40 /* MinusToken */, "**": 42 /* AsteriskAsteriskToken */, "*": 41 /* AsteriskToken */, "/": 43 /* SlashToken */, "%": 44 /* PercentToken */, "++": 45 /* PlusPlusToken */, "--": 46 /* MinusMinusToken */, "<<": 47 /* LessThanLessThanToken */, "</": 30 /* LessThanSlashToken */, ">>": 48 /* GreaterThanGreaterThanToken */, ">>>": 49 /* GreaterThanGreaterThanGreaterThanToken */, "&": 50 /* AmpersandToken */, "|": 51 /* BarToken */, "^": 52 /* CaretToken */, "!": 53 /* ExclamationToken */, "~": 54 /* TildeToken */, "&&": 55 /* AmpersandAmpersandToken */, "||": 56 /* BarBarToken */, "?": 57 /* QuestionToken */, "??": 60 /* QuestionQuestionToken */, "?.": 28 /* QuestionDotToken */, ":": 58 /* ColonToken */, "=": 63 /* EqualsToken */, "+=": 64 /* PlusEqualsToken */, "-=": 65 /* MinusEqualsToken */, "*=": 66 /* AsteriskEqualsToken */, "**=": 67 /* AsteriskAsteriskEqualsToken */, "/=": 68 /* SlashEqualsToken */, "%=": 69 /* PercentEqualsToken */, "<<=": 70 /* LessThanLessThanEqualsToken */, ">>=": 71 /* GreaterThanGreaterThanEqualsToken */, ">>>=": 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */, "&=": 73 /* AmpersandEqualsToken */, "|=": 74 /* BarEqualsToken */, "^=": 78 /* CaretEqualsToken */, "||=": 75 /* BarBarEqualsToken */, "&&=": 76 /* AmpersandAmpersandEqualsToken */, "??=": 77 /* QuestionQuestionEqualsToken */, "@": 59 /* AtToken */, "#": 62 /* HashToken */, "`": 61 /* BacktickToken */ })));
        /*
            As per ECMAScript Language Specification 3th Edition, Section 7.6: Identifiers
            IdentifierStart ::
                Can contain Unicode 3.0.0 categories:
                Uppercase letter (Lu),
                Lowercase letter (Ll),
                Titlecase letter (Lt),
                Modifier letter (Lm),
                Other letter (Lo), or
                Letter number (Nl).
            IdentifierPart :: =
                Can contain IdentifierStart + Unicode 3.0.0 categories:
                Non-spacing mark (Mn),
                Combining spacing mark (Mc),
                Decimal number (Nd), or
                Connector punctuation (Pc).
    
            Codepoint ranges for ES3 Identifiers are extracted from the Unicode 3.0.0 specification at:
            http://www.unicode.org/Public/3.0-Update/UnicodeData-3.0.0.txt
        */
        var unicodeES3IdentifierStart = [170, 170, 181, 181, 186, 186, 192, 214, 216, 246, 248, 543, 546, 563, 592, 685, 688, 696, 699, 705, 720, 721, 736, 740, 750, 750, 890, 890, 902, 902, 904, 906, 908, 908, 910, 929, 931, 974, 976, 983, 986, 1011, 1024, 1153, 1164, 1220, 1223, 1224, 1227, 1228, 1232, 1269, 1272, 1273, 1329, 1366, 1369, 1369, 1377, 1415, 1488, 1514, 1520, 1522, 1569, 1594, 1600, 1610, 1649, 1747, 1749, 1749, 1765, 1766, 1786, 1788, 1808, 1808, 1810, 1836, 1920, 1957, 2309, 2361, 2365, 2365, 2384, 2384, 2392, 2401, 2437, 2444, 2447, 2448, 2451, 2472, 2474, 2480, 2482, 2482, 2486, 2489, 2524, 2525, 2527, 2529, 2544, 2545, 2565, 2570, 2575, 2576, 2579, 2600, 2602, 2608, 2610, 2611, 2613, 2614, 2616, 2617, 2649, 2652, 2654, 2654, 2674, 2676, 2693, 2699, 2701, 2701, 2703, 2705, 2707, 2728, 2730, 2736, 2738, 2739, 2741, 2745, 2749, 2749, 2768, 2768, 2784, 2784, 2821, 2828, 2831, 2832, 2835, 2856, 2858, 2864, 2866, 2867, 2870, 2873, 2877, 2877, 2908, 2909, 2911, 2913, 2949, 2954, 2958, 2960, 2962, 2965, 2969, 2970, 2972, 2972, 2974, 2975, 2979, 2980, 2984, 2986, 2990, 2997, 2999, 3001, 3077, 3084, 3086, 3088, 3090, 3112, 3114, 3123, 3125, 3129, 3168, 3169, 3205, 3212, 3214, 3216, 3218, 3240, 3242, 3251, 3253, 3257, 3294, 3294, 3296, 3297, 3333, 3340, 3342, 3344, 3346, 3368, 3370, 3385, 3424, 3425, 3461, 3478, 3482, 3505, 3507, 3515, 3517, 3517, 3520, 3526, 3585, 3632, 3634, 3635, 3648, 3654, 3713, 3714, 3716, 3716, 3719, 3720, 3722, 3722, 3725, 3725, 3732, 3735, 3737, 3743, 3745, 3747, 3749, 3749, 3751, 3751, 3754, 3755, 3757, 3760, 3762, 3763, 3773, 3773, 3776, 3780, 3782, 3782, 3804, 3805, 3840, 3840, 3904, 3911, 3913, 3946, 3976, 3979, 4096, 4129, 4131, 4135, 4137, 4138, 4176, 4181, 4256, 4293, 4304, 4342, 4352, 4441, 4447, 4514, 4520, 4601, 4608, 4614, 4616, 4678, 4680, 4680, 4682, 4685, 4688, 4694, 4696, 4696, 4698, 4701, 4704, 4742, 4744, 4744, 4746, 4749, 4752, 4782, 4784, 4784, 4786, 4789, 4792, 4798, 4800, 4800, 4802, 4805, 4808, 4814, 4816, 4822, 4824, 4846, 4848, 4878, 4880, 4880, 4882, 4885, 4888, 4894, 4896, 4934, 4936, 4954, 5024, 5108, 5121, 5740, 5743, 5750, 5761, 5786, 5792, 5866, 6016, 6067, 6176, 6263, 6272, 6312, 7680, 7835, 7840, 7929, 7936, 7957, 7960, 7965, 7968, 8005, 8008, 8013, 8016, 8023, 8025, 8025, 8027, 8027, 8029, 8029, 8031, 8061, 8064, 8116, 8118, 8124, 8126, 8126, 8130, 8132, 8134, 8140, 8144, 8147, 8150, 8155, 8160, 8172, 8178, 8180, 8182, 8188, 8319, 8319, 8450, 8450, 8455, 8455, 8458, 8467, 8469, 8469, 8473, 8477, 8484, 8484, 8486, 8486, 8488, 8488, 8490, 8493, 8495, 8497, 8499, 8505, 8544, 8579, 12293, 12295, 12321, 12329, 12337, 12341, 12344, 12346, 12353, 12436, 12445, 12446, 12449, 12538, 12540, 12542, 12549, 12588, 12593, 12686, 12704, 12727, 13312, 19893, 19968, 40869, 40960, 42124, 44032, 55203, 63744, 64045, 64256, 64262, 64275, 64279, 64285, 64285, 64287, 64296, 64298, 64310, 64312, 64316, 64318, 64318, 64320, 64321, 64323, 64324, 64326, 64433, 64467, 64829, 64848, 64911, 64914, 64967, 65008, 65019, 65136, 65138, 65140, 65140, 65142, 65276, 65313, 65338, 65345, 65370, 65382, 65470, 65474, 65479, 65482, 65487, 65490, 65495, 65498, 65500,];
        var unicodeES3IdentifierPart = [170, 170, 181, 181, 186, 186, 192, 214, 216, 246, 248, 543, 546, 563, 592, 685, 688, 696, 699, 705, 720, 721, 736, 740, 750, 750, 768, 846, 864, 866, 890, 890, 902, 902, 904, 906, 908, 908, 910, 929, 931, 974, 976, 983, 986, 1011, 1024, 1153, 1155, 1158, 1164, 1220, 1223, 1224, 1227, 1228, 1232, 1269, 1272, 1273, 1329, 1366, 1369, 1369, 1377, 1415, 1425, 1441, 1443, 1465, 1467, 1469, 1471, 1471, 1473, 1474, 1476, 1476, 1488, 1514, 1520, 1522, 1569, 1594, 1600, 1621, 1632, 1641, 1648, 1747, 1749, 1756, 1759, 1768, 1770, 1773, 1776, 1788, 1808, 1836, 1840, 1866, 1920, 1968, 2305, 2307, 2309, 2361, 2364, 2381, 2384, 2388, 2392, 2403, 2406, 2415, 2433, 2435, 2437, 2444, 2447, 2448, 2451, 2472, 2474, 2480, 2482, 2482, 2486, 2489, 2492, 2492, 2494, 2500, 2503, 2504, 2507, 2509, 2519, 2519, 2524, 2525, 2527, 2531, 2534, 2545, 2562, 2562, 2565, 2570, 2575, 2576, 2579, 2600, 2602, 2608, 2610, 2611, 2613, 2614, 2616, 2617, 2620, 2620, 2622, 2626, 2631, 2632, 2635, 2637, 2649, 2652, 2654, 2654, 2662, 2676, 2689, 2691, 2693, 2699, 2701, 2701, 2703, 2705, 2707, 2728, 2730, 2736, 2738, 2739, 2741, 2745, 2748, 2757, 2759, 2761, 2763, 2765, 2768, 2768, 2784, 2784, 2790, 2799, 2817, 2819, 2821, 2828, 2831, 2832, 2835, 2856, 2858, 2864, 2866, 2867, 2870, 2873, 2876, 2883, 2887, 2888, 2891, 2893, 2902, 2903, 2908, 2909, 2911, 2913, 2918, 2927, 2946, 2947, 2949, 2954, 2958, 2960, 2962, 2965, 2969, 2970, 2972, 2972, 2974, 2975, 2979, 2980, 2984, 2986, 2990, 2997, 2999, 3001, 3006, 3010, 3014, 3016, 3018, 3021, 3031, 3031, 3047, 3055, 3073, 3075, 3077, 3084, 3086, 3088, 3090, 3112, 3114, 3123, 3125, 3129, 3134, 3140, 3142, 3144, 3146, 3149, 3157, 3158, 3168, 3169, 3174, 3183, 3202, 3203, 3205, 3212, 3214, 3216, 3218, 3240, 3242, 3251, 3253, 3257, 3262, 3268, 3270, 3272, 3274, 3277, 3285, 3286, 3294, 3294, 3296, 3297, 3302, 3311, 3330, 3331, 3333, 3340, 3342, 3344, 3346, 3368, 3370, 3385, 3390, 3395, 3398, 3400, 3402, 3405, 3415, 3415, 3424, 3425, 3430, 3439, 3458, 3459, 3461, 3478, 3482, 3505, 3507, 3515, 3517, 3517, 3520, 3526, 3530, 3530, 3535, 3540, 3542, 3542, 3544, 3551, 3570, 3571, 3585, 3642, 3648, 3662, 3664, 3673, 3713, 3714, 3716, 3716, 3719, 3720, 3722, 3722, 3725, 3725, 3732, 3735, 3737, 3743, 3745, 3747, 3749, 3749, 3751, 3751, 3754, 3755, 3757, 3769, 3771, 3773, 3776, 3780, 3782, 3782, 3784, 3789, 3792, 3801, 3804, 3805, 3840, 3840, 3864, 3865, 3872, 3881, 3893, 3893, 3895, 3895, 3897, 3897, 3902, 3911, 3913, 3946, 3953, 3972, 3974, 3979, 3984, 3991, 3993, 4028, 4038, 4038, 4096, 4129, 4131, 4135, 4137, 4138, 4140, 4146, 4150, 4153, 4160, 4169, 4176, 4185, 4256, 4293, 4304, 4342, 4352, 4441, 4447, 4514, 4520, 4601, 4608, 4614, 4616, 4678, 4680, 4680, 4682, 4685, 4688, 4694, 4696, 4696, 4698, 4701, 4704, 4742, 4744, 4744, 4746, 4749, 4752, 4782, 4784, 4784, 4786, 4789, 4792, 4798, 4800, 4800, 4802, 4805, 4808, 4814, 4816, 4822, 4824, 4846, 4848, 4878, 4880, 4880, 4882, 4885, 4888, 4894, 4896, 4934, 4936, 4954, 4969, 4977, 5024, 5108, 5121, 5740, 5743, 5750, 5761, 5786, 5792, 5866, 6016, 6099, 6112, 6121, 6160, 6169, 6176, 6263, 6272, 6313, 7680, 7835, 7840, 7929, 7936, 7957, 7960, 7965, 7968, 8005, 8008, 8013, 8016, 8023, 8025, 8025, 8027, 8027, 8029, 8029, 8031, 8061, 8064, 8116, 8118, 8124, 8126, 8126, 8130, 8132, 8134, 8140, 8144, 8147, 8150, 8155, 8160, 8172, 8178, 8180, 8182, 8188, 8255, 8256, 8319, 8319, 8400, 8412, 8417, 8417, 8450, 8450, 8455, 8455, 8458, 8467, 8469, 8469, 8473, 8477, 8484, 8484, 8486, 8486, 8488, 8488, 8490, 8493, 8495, 8497, 8499, 8505, 8544, 8579, 12293, 12295, 12321, 12335, 12337, 12341, 12344, 12346, 12353, 12436, 12441, 12442, 12445, 12446, 12449, 12542, 12549, 12588, 12593, 12686, 12704, 12727, 13312, 19893, 19968, 40869, 40960, 42124, 44032, 55203, 63744, 64045, 64256, 64262, 64275, 64279, 64285, 64296, 64298, 64310, 64312, 64316, 64318, 64318, 64320, 64321, 64323, 64324, 64326, 64433, 64467, 64829, 64848, 64911, 64914, 64967, 65008, 65019, 65056, 65059, 65075, 65076, 65101, 65103, 65136, 65138, 65140, 65140, 65142, 65276, 65296, 65305, 65313, 65338, 65343, 65343, 65345, 65370, 65381, 65470, 65474, 65479, 65482, 65487, 65490, 65495, 65498, 65500,];
        /*
            As per ECMAScript Language Specification 5th Edition, Section 7.6: ISyntaxToken Names and Identifiers
            IdentifierStart ::
                Can contain Unicode 6.2 categories:
                Uppercase letter (Lu),
                Lowercase letter (Ll),
                Titlecase letter (Lt),
                Modifier letter (Lm),
                Other letter (Lo), or
                Letter number (Nl).
            IdentifierPart ::
                Can contain IdentifierStart + Unicode 6.2 categories:
                Non-spacing mark (Mn),
                Combining spacing mark (Mc),
                Decimal number (Nd),
                Connector punctuation (Pc),
                <ZWNJ>, or
                <ZWJ>.
    
            Codepoint ranges for ES5 Identifiers are extracted from the Unicode 6.2 specification at:
            http://www.unicode.org/Public/6.2.0/ucd/UnicodeData.txt
        */
        var unicodeES5IdentifierStart = [170, 170, 181, 181, 186, 186, 192, 214, 216, 246, 248, 705, 710, 721, 736, 740, 748, 748, 750, 750, 880, 884, 886, 887, 890, 893, 902, 902, 904, 906, 908, 908, 910, 929, 931, 1013, 1015, 1153, 1162, 1319, 1329, 1366, 1369, 1369, 1377, 1415, 1488, 1514, 1520, 1522, 1568, 1610, 1646, 1647, 1649, 1747, 1749, 1749, 1765, 1766, 1774, 1775, 1786, 1788, 1791, 1791, 1808, 1808, 1810, 1839, 1869, 1957, 1969, 1969, 1994, 2026, 2036, 2037, 2042, 2042, 2048, 2069, 2074, 2074, 2084, 2084, 2088, 2088, 2112, 2136, 2208, 2208, 2210, 2220, 2308, 2361, 2365, 2365, 2384, 2384, 2392, 2401, 2417, 2423, 2425, 2431, 2437, 2444, 2447, 2448, 2451, 2472, 2474, 2480, 2482, 2482, 2486, 2489, 2493, 2493, 2510, 2510, 2524, 2525, 2527, 2529, 2544, 2545, 2565, 2570, 2575, 2576, 2579, 2600, 2602, 2608, 2610, 2611, 2613, 2614, 2616, 2617, 2649, 2652, 2654, 2654, 2674, 2676, 2693, 2701, 2703, 2705, 2707, 2728, 2730, 2736, 2738, 2739, 2741, 2745, 2749, 2749, 2768, 2768, 2784, 2785, 2821, 2828, 2831, 2832, 2835, 2856, 2858, 2864, 2866, 2867, 2869, 2873, 2877, 2877, 2908, 2909, 2911, 2913, 2929, 2929, 2947, 2947, 2949, 2954, 2958, 2960, 2962, 2965, 2969, 2970, 2972, 2972, 2974, 2975, 2979, 2980, 2984, 2986, 2990, 3001, 3024, 3024, 3077, 3084, 3086, 3088, 3090, 3112, 3114, 3123, 3125, 3129, 3133, 3133, 3160, 3161, 3168, 3169, 3205, 3212, 3214, 3216, 3218, 3240, 3242, 3251, 3253, 3257, 3261, 3261, 3294, 3294, 3296, 3297, 3313, 3314, 3333, 3340, 3342, 3344, 3346, 3386, 3389, 3389, 3406, 3406, 3424, 3425, 3450, 3455, 3461, 3478, 3482, 3505, 3507, 3515, 3517, 3517, 3520, 3526, 3585, 3632, 3634, 3635, 3648, 3654, 3713, 3714, 3716, 3716, 3719, 3720, 3722, 3722, 3725, 3725, 3732, 3735, 3737, 3743, 3745, 3747, 3749, 3749, 3751, 3751, 3754, 3755, 3757, 3760, 3762, 3763, 3773, 3773, 3776, 3780, 3782, 3782, 3804, 3807, 3840, 3840, 3904, 3911, 3913, 3948, 3976, 3980, 4096, 4138, 4159, 4159, 4176, 4181, 4186, 4189, 4193, 4193, 4197, 4198, 4206, 4208, 4213, 4225, 4238, 4238, 4256, 4293, 4295, 4295, 4301, 4301, 4304, 4346, 4348, 4680, 4682, 4685, 4688, 4694, 4696, 4696, 4698, 4701, 4704, 4744, 4746, 4749, 4752, 4784, 4786, 4789, 4792, 4798, 4800, 4800, 4802, 4805, 4808, 4822, 4824, 4880, 4882, 4885, 4888, 4954, 4992, 5007, 5024, 5108, 5121, 5740, 5743, 5759, 5761, 5786, 5792, 5866, 5870, 5872, 5888, 5900, 5902, 5905, 5920, 5937, 5952, 5969, 5984, 5996, 5998, 6000, 6016, 6067, 6103, 6103, 6108, 6108, 6176, 6263, 6272, 6312, 6314, 6314, 6320, 6389, 6400, 6428, 6480, 6509, 6512, 6516, 6528, 6571, 6593, 6599, 6656, 6678, 6688, 6740, 6823, 6823, 6917, 6963, 6981, 6987, 7043, 7072, 7086, 7087, 7098, 7141, 7168, 7203, 7245, 7247, 7258, 7293, 7401, 7404, 7406, 7409, 7413, 7414, 7424, 7615, 7680, 7957, 7960, 7965, 7968, 8005, 8008, 8013, 8016, 8023, 8025, 8025, 8027, 8027, 8029, 8029, 8031, 8061, 8064, 8116, 8118, 8124, 8126, 8126, 8130, 8132, 8134, 8140, 8144, 8147, 8150, 8155, 8160, 8172, 8178, 8180, 8182, 8188, 8305, 8305, 8319, 8319, 8336, 8348, 8450, 8450, 8455, 8455, 8458, 8467, 8469, 8469, 8473, 8477, 8484, 8484, 8486, 8486, 8488, 8488, 8490, 8493, 8495, 8505, 8508, 8511, 8517, 8521, 8526, 8526, 8544, 8584, 11264, 11310, 11312, 11358, 11360, 11492, 11499, 11502, 11506, 11507, 11520, 11557, 11559, 11559, 11565, 11565, 11568, 11623, 11631, 11631, 11648, 11670, 11680, 11686, 11688, 11694, 11696, 11702, 11704, 11710, 11712, 11718, 11720, 11726, 11728, 11734, 11736, 11742, 11823, 11823, 12293, 12295, 12321, 12329, 12337, 12341, 12344, 12348, 12353, 12438, 12445, 12447, 12449, 12538, 12540, 12543, 12549, 12589, 12593, 12686, 12704, 12730, 12784, 12799, 13312, 19893, 19968, 40908, 40960, 42124, 42192, 42237, 42240, 42508, 42512, 42527, 42538, 42539, 42560, 42606, 42623, 42647, 42656, 42735, 42775, 42783, 42786, 42888, 42891, 42894, 42896, 42899, 42912, 42922, 43000, 43009, 43011, 43013, 43015, 43018, 43020, 43042, 43072, 43123, 43138, 43187, 43250, 43255, 43259, 43259, 43274, 43301, 43312, 43334, 43360, 43388, 43396, 43442, 43471, 43471, 43520, 43560, 43584, 43586, 43588, 43595, 43616, 43638, 43642, 43642, 43648, 43695, 43697, 43697, 43701, 43702, 43705, 43709, 43712, 43712, 43714, 43714, 43739, 43741, 43744, 43754, 43762, 43764, 43777, 43782, 43785, 43790, 43793, 43798, 43808, 43814, 43816, 43822, 43968, 44002, 44032, 55203, 55216, 55238, 55243, 55291, 63744, 64109, 64112, 64217, 64256, 64262, 64275, 64279, 64285, 64285, 64287, 64296, 64298, 64310, 64312, 64316, 64318, 64318, 64320, 64321, 64323, 64324, 64326, 64433, 64467, 64829, 64848, 64911, 64914, 64967, 65008, 65019, 65136, 65140, 65142, 65276, 65313, 65338, 65345, 65370, 65382, 65470, 65474, 65479, 65482, 65487, 65490, 65495, 65498, 65500,];
        var unicodeES5IdentifierPart = [170, 170, 181, 181, 186, 186, 192, 214, 216, 246, 248, 705, 710, 721, 736, 740, 748, 748, 750, 750, 768, 884, 886, 887, 890, 893, 902, 902, 904, 906, 908, 908, 910, 929, 931, 1013, 1015, 1153, 1155, 1159, 1162, 1319, 1329, 1366, 1369, 1369, 1377, 1415, 1425, 1469, 1471, 1471, 1473, 1474, 1476, 1477, 1479, 1479, 1488, 1514, 1520, 1522, 1552, 1562, 1568, 1641, 1646, 1747, 1749, 1756, 1759, 1768, 1770, 1788, 1791, 1791, 1808, 1866, 1869, 1969, 1984, 2037, 2042, 2042, 2048, 2093, 2112, 2139, 2208, 2208, 2210, 2220, 2276, 2302, 2304, 2403, 2406, 2415, 2417, 2423, 2425, 2431, 2433, 2435, 2437, 2444, 2447, 2448, 2451, 2472, 2474, 2480, 2482, 2482, 2486, 2489, 2492, 2500, 2503, 2504, 2507, 2510, 2519, 2519, 2524, 2525, 2527, 2531, 2534, 2545, 2561, 2563, 2565, 2570, 2575, 2576, 2579, 2600, 2602, 2608, 2610, 2611, 2613, 2614, 2616, 2617, 2620, 2620, 2622, 2626, 2631, 2632, 2635, 2637, 2641, 2641, 2649, 2652, 2654, 2654, 2662, 2677, 2689, 2691, 2693, 2701, 2703, 2705, 2707, 2728, 2730, 2736, 2738, 2739, 2741, 2745, 2748, 2757, 2759, 2761, 2763, 2765, 2768, 2768, 2784, 2787, 2790, 2799, 2817, 2819, 2821, 2828, 2831, 2832, 2835, 2856, 2858, 2864, 2866, 2867, 2869, 2873, 2876, 2884, 2887, 2888, 2891, 2893, 2902, 2903, 2908, 2909, 2911, 2915, 2918, 2927, 2929, 2929, 2946, 2947, 2949, 2954, 2958, 2960, 2962, 2965, 2969, 2970, 2972, 2972, 2974, 2975, 2979, 2980, 2984, 2986, 2990, 3001, 3006, 3010, 3014, 3016, 3018, 3021, 3024, 3024, 3031, 3031, 3046, 3055, 3073, 3075, 3077, 3084, 3086, 3088, 3090, 3112, 3114, 3123, 3125, 3129, 3133, 3140, 3142, 3144, 3146, 3149, 3157, 3158, 3160, 3161, 3168, 3171, 3174, 3183, 3202, 3203, 3205, 3212, 3214, 3216, 3218, 3240, 3242, 3251, 3253, 3257, 3260, 3268, 3270, 3272, 3274, 3277, 3285, 3286, 3294, 3294, 3296, 3299, 3302, 3311, 3313, 3314, 3330, 3331, 3333, 3340, 3342, 3344, 3346, 3386, 3389, 3396, 3398, 3400, 3402, 3406, 3415, 3415, 3424, 3427, 3430, 3439, 3450, 3455, 3458, 3459, 3461, 3478, 3482, 3505, 3507, 3515, 3517, 3517, 3520, 3526, 3530, 3530, 3535, 3540, 3542, 3542, 3544, 3551, 3570, 3571, 3585, 3642, 3648, 3662, 3664, 3673, 3713, 3714, 3716, 3716, 3719, 3720, 3722, 3722, 3725, 3725, 3732, 3735, 3737, 3743, 3745, 3747, 3749, 3749, 3751, 3751, 3754, 3755, 3757, 3769, 3771, 3773, 3776, 3780, 3782, 3782, 3784, 3789, 3792, 3801, 3804, 3807, 3840, 3840, 3864, 3865, 3872, 3881, 3893, 3893, 3895, 3895, 3897, 3897, 3902, 3911, 3913, 3948, 3953, 3972, 3974, 3991, 3993, 4028, 4038, 4038, 4096, 4169, 4176, 4253, 4256, 4293, 4295, 4295, 4301, 4301, 4304, 4346, 4348, 4680, 4682, 4685, 4688, 4694, 4696, 4696, 4698, 4701, 4704, 4744, 4746, 4749, 4752, 4784, 4786, 4789, 4792, 4798, 4800, 4800, 4802, 4805, 4808, 4822, 4824, 4880, 4882, 4885, 4888, 4954, 4957, 4959, 4992, 5007, 5024, 5108, 5121, 5740, 5743, 5759, 5761, 5786, 5792, 5866, 5870, 5872, 5888, 5900, 5902, 5908, 5920, 5940, 5952, 5971, 5984, 5996, 5998, 6000, 6002, 6003, 6016, 6099, 6103, 6103, 6108, 6109, 6112, 6121, 6155, 6157, 6160, 6169, 6176, 6263, 6272, 6314, 6320, 6389, 6400, 6428, 6432, 6443, 6448, 6459, 6470, 6509, 6512, 6516, 6528, 6571, 6576, 6601, 6608, 6617, 6656, 6683, 6688, 6750, 6752, 6780, 6783, 6793, 6800, 6809, 6823, 6823, 6912, 6987, 6992, 7001, 7019, 7027, 7040, 7155, 7168, 7223, 7232, 7241, 7245, 7293, 7376, 7378, 7380, 7414, 7424, 7654, 7676, 7957, 7960, 7965, 7968, 8005, 8008, 8013, 8016, 8023, 8025, 8025, 8027, 8027, 8029, 8029, 8031, 8061, 8064, 8116, 8118, 8124, 8126, 8126, 8130, 8132, 8134, 8140, 8144, 8147, 8150, 8155, 8160, 8172, 8178, 8180, 8182, 8188, 8204, 8205, 8255, 8256, 8276, 8276, 8305, 8305, 8319, 8319, 8336, 8348, 8400, 8412, 8417, 8417, 8421, 8432, 8450, 8450, 8455, 8455, 8458, 8467, 8469, 8469, 8473, 8477, 8484, 8484, 8486, 8486, 8488, 8488, 8490, 8493, 8495, 8505, 8508, 8511, 8517, 8521, 8526, 8526, 8544, 8584, 11264, 11310, 11312, 11358, 11360, 11492, 11499, 11507, 11520, 11557, 11559, 11559, 11565, 11565, 11568, 11623, 11631, 11631, 11647, 11670, 11680, 11686, 11688, 11694, 11696, 11702, 11704, 11710, 11712, 11718, 11720, 11726, 11728, 11734, 11736, 11742, 11744, 11775, 11823, 11823, 12293, 12295, 12321, 12335, 12337, 12341, 12344, 12348, 12353, 12438, 12441, 12442, 12445, 12447, 12449, 12538, 12540, 12543, 12549, 12589, 12593, 12686, 12704, 12730, 12784, 12799, 13312, 19893, 19968, 40908, 40960, 42124, 42192, 42237, 42240, 42508, 42512, 42539, 42560, 42607, 42612, 42621, 42623, 42647, 42655, 42737, 42775, 42783, 42786, 42888, 42891, 42894, 42896, 42899, 42912, 42922, 43000, 43047, 43072, 43123, 43136, 43204, 43216, 43225, 43232, 43255, 43259, 43259, 43264, 43309, 43312, 43347, 43360, 43388, 43392, 43456, 43471, 43481, 43520, 43574, 43584, 43597, 43600, 43609, 43616, 43638, 43642, 43643, 43648, 43714, 43739, 43741, 43744, 43759, 43762, 43766, 43777, 43782, 43785, 43790, 43793, 43798, 43808, 43814, 43816, 43822, 43968, 44010, 44012, 44013, 44016, 44025, 44032, 55203, 55216, 55238, 55243, 55291, 63744, 64109, 64112, 64217, 64256, 64262, 64275, 64279, 64285, 64296, 64298, 64310, 64312, 64316, 64318, 64318, 64320, 64321, 64323, 64324, 64326, 64433, 64467, 64829, 64848, 64911, 64914, 64967, 65008, 65019, 65024, 65039, 65056, 65062, 65075, 65076, 65101, 65103, 65136, 65140, 65142, 65276, 65296, 65305, 65313, 65338, 65343, 65343, 65345, 65370, 65382, 65470, 65474, 65479, 65482, 65487, 65490, 65495, 65498, 65500,];
        /**
         * Generated by scripts/regenerate-unicode-identifier-parts.js on node v12.4.0 with unicode 12.1
         * based on http://www.unicode.org/reports/tr31/ and https://www.ecma-international.org/ecma-262/6.0/#sec-names-and-keywords
         * unicodeESNextIdentifierStart corresponds to the ID_Start and Other_ID_Start property, and
         * unicodeESNextIdentifierPart corresponds to ID_Continue, Other_ID_Continue, plus ID_Start and Other_ID_Start
         */
        var unicodeESNextIdentifierStart = [65, 90, 97, 122, 170, 170, 181, 181, 186, 186, 192, 214, 216, 246, 248, 705, 710, 721, 736, 740, 748, 748, 750, 750, 880, 884, 886, 887, 890, 893, 895, 895, 902, 902, 904, 906, 908, 908, 910, 929, 931, 1013, 1015, 1153, 1162, 1327, 1329, 1366, 1369, 1369, 1376, 1416, 1488, 1514, 1519, 1522, 1568, 1610, 1646, 1647, 1649, 1747, 1749, 1749, 1765, 1766, 1774, 1775, 1786, 1788, 1791, 1791, 1808, 1808, 1810, 1839, 1869, 1957, 1969, 1969, 1994, 2026, 2036, 2037, 2042, 2042, 2048, 2069, 2074, 2074, 2084, 2084, 2088, 2088, 2112, 2136, 2144, 2154, 2208, 2228, 2230, 2237, 2308, 2361, 2365, 2365, 2384, 2384, 2392, 2401, 2417, 2432, 2437, 2444, 2447, 2448, 2451, 2472, 2474, 2480, 2482, 2482, 2486, 2489, 2493, 2493, 2510, 2510, 2524, 2525, 2527, 2529, 2544, 2545, 2556, 2556, 2565, 2570, 2575, 2576, 2579, 2600, 2602, 2608, 2610, 2611, 2613, 2614, 2616, 2617, 2649, 2652, 2654, 2654, 2674, 2676, 2693, 2701, 2703, 2705, 2707, 2728, 2730, 2736, 2738, 2739, 2741, 2745, 2749, 2749, 2768, 2768, 2784, 2785, 2809, 2809, 2821, 2828, 2831, 2832, 2835, 2856, 2858, 2864, 2866, 2867, 2869, 2873, 2877, 2877, 2908, 2909, 2911, 2913, 2929, 2929, 2947, 2947, 2949, 2954, 2958, 2960, 2962, 2965, 2969, 2970, 2972, 2972, 2974, 2975, 2979, 2980, 2984, 2986, 2990, 3001, 3024, 3024, 3077, 3084, 3086, 3088, 3090, 3112, 3114, 3129, 3133, 3133, 3160, 3162, 3168, 3169, 3200, 3200, 3205, 3212, 3214, 3216, 3218, 3240, 3242, 3251, 3253, 3257, 3261, 3261, 3294, 3294, 3296, 3297, 3313, 3314, 3333, 3340, 3342, 3344, 3346, 3386, 3389, 3389, 3406, 3406, 3412, 3414, 3423, 3425, 3450, 3455, 3461, 3478, 3482, 3505, 3507, 3515, 3517, 3517, 3520, 3526, 3585, 3632, 3634, 3635, 3648, 3654, 3713, 3714, 3716, 3716, 3718, 3722, 3724, 3747, 3749, 3749, 3751, 3760, 3762, 3763, 3773, 3773, 3776, 3780, 3782, 3782, 3804, 3807, 3840, 3840, 3904, 3911, 3913, 3948, 3976, 3980, 4096, 4138, 4159, 4159, 4176, 4181, 4186, 4189, 4193, 4193, 4197, 4198, 4206, 4208, 4213, 4225, 4238, 4238, 4256, 4293, 4295, 4295, 4301, 4301, 4304, 4346, 4348, 4680, 4682, 4685, 4688, 4694, 4696, 4696, 4698, 4701, 4704, 4744, 4746, 4749, 4752, 4784, 4786, 4789, 4792, 4798, 4800, 4800, 4802, 4805, 4808, 4822, 4824, 4880, 4882, 4885, 4888, 4954, 4992, 5007, 5024, 5109, 5112, 5117, 5121, 5740, 5743, 5759, 5761, 5786, 5792, 5866, 5870, 5880, 5888, 5900, 5902, 5905, 5920, 5937, 5952, 5969, 5984, 5996, 5998, 6000, 6016, 6067, 6103, 6103, 6108, 6108, 6176, 6264, 6272, 6312, 6314, 6314, 6320, 6389, 6400, 6430, 6480, 6509, 6512, 6516, 6528, 6571, 6576, 6601, 6656, 6678, 6688, 6740, 6823, 6823, 6917, 6963, 6981, 6987, 7043, 7072, 7086, 7087, 7098, 7141, 7168, 7203, 7245, 7247, 7258, 7293, 7296, 7304, 7312, 7354, 7357, 7359, 7401, 7404, 7406, 7411, 7413, 7414, 7418, 7418, 7424, 7615, 7680, 7957, 7960, 7965, 7968, 8005, 8008, 8013, 8016, 8023, 8025, 8025, 8027, 8027, 8029, 8029, 8031, 8061, 8064, 8116, 8118, 8124, 8126, 8126, 8130, 8132, 8134, 8140, 8144, 8147, 8150, 8155, 8160, 8172, 8178, 8180, 8182, 8188, 8305, 8305, 8319, 8319, 8336, 8348, 8450, 8450, 8455, 8455, 8458, 8467, 8469, 8469, 8472, 8477, 8484, 8484, 8486, 8486, 8488, 8488, 8490, 8505, 8508, 8511, 8517, 8521, 8526, 8526, 8544, 8584, 11264, 11310, 11312, 11358, 11360, 11492, 11499, 11502, 11506, 11507, 11520, 11557, 11559, 11559, 11565, 11565, 11568, 11623, 11631, 11631, 11648, 11670, 11680, 11686, 11688, 11694, 11696, 11702, 11704, 11710, 11712, 11718, 11720, 11726, 11728, 11734, 11736, 11742, 12293, 12295, 12321, 12329, 12337, 12341, 12344, 12348, 12353, 12438, 12443, 12447, 12449, 12538, 12540, 12543, 12549, 12591, 12593, 12686, 12704, 12730, 12784, 12799, 13312, 19893, 19968, 40943, 40960, 42124, 42192, 42237, 42240, 42508, 42512, 42527, 42538, 42539, 42560, 42606, 42623, 42653, 42656, 42735, 42775, 42783, 42786, 42888, 42891, 42943, 42946, 42950, 42999, 43009, 43011, 43013, 43015, 43018, 43020, 43042, 43072, 43123, 43138, 43187, 43250, 43255, 43259, 43259, 43261, 43262, 43274, 43301, 43312, 43334, 43360, 43388, 43396, 43442, 43471, 43471, 43488, 43492, 43494, 43503, 43514, 43518, 43520, 43560, 43584, 43586, 43588, 43595, 43616, 43638, 43642, 43642, 43646, 43695, 43697, 43697, 43701, 43702, 43705, 43709, 43712, 43712, 43714, 43714, 43739, 43741, 43744, 43754, 43762, 43764, 43777, 43782, 43785, 43790, 43793, 43798, 43808, 43814, 43816, 43822, 43824, 43866, 43868, 43879, 43888, 44002, 44032, 55203, 55216, 55238, 55243, 55291, 63744, 64109, 64112, 64217, 64256, 64262, 64275, 64279, 64285, 64285, 64287, 64296, 64298, 64310, 64312, 64316, 64318, 64318, 64320, 64321, 64323, 64324, 64326, 64433, 64467, 64829, 64848, 64911, 64914, 64967, 65008, 65019, 65136, 65140, 65142, 65276, 65313, 65338, 65345, 65370, 65382, 65470, 65474, 65479, 65482, 65487, 65490, 65495, 65498, 65500, 65536, 65547, 65549, 65574, 65576, 65594, 65596, 65597, 65599, 65613, 65616, 65629, 65664, 65786, 65856, 65908, 66176, 66204, 66208, 66256, 66304, 66335, 66349, 66378, 66384, 66421, 66432, 66461, 66464, 66499, 66504, 66511, 66513, 66517, 66560, 66717, 66736, 66771, 66776, 66811, 66816, 66855, 66864, 66915, 67072, 67382, 67392, 67413, 67424, 67431, 67584, 67589, 67592, 67592, 67594, 67637, 67639, 67640, 67644, 67644, 67647, 67669, 67680, 67702, 67712, 67742, 67808, 67826, 67828, 67829, 67840, 67861, 67872, 67897, 67968, 68023, 68030, 68031, 68096, 68096, 68112, 68115, 68117, 68119, 68121, 68149, 68192, 68220, 68224, 68252, 68288, 68295, 68297, 68324, 68352, 68405, 68416, 68437, 68448, 68466, 68480, 68497, 68608, 68680, 68736, 68786, 68800, 68850, 68864, 68899, 69376, 69404, 69415, 69415, 69424, 69445, 69600, 69622, 69635, 69687, 69763, 69807, 69840, 69864, 69891, 69926, 69956, 69956, 69968, 70002, 70006, 70006, 70019, 70066, 70081, 70084, 70106, 70106, 70108, 70108, 70144, 70161, 70163, 70187, 70272, 70278, 70280, 70280, 70282, 70285, 70287, 70301, 70303, 70312, 70320, 70366, 70405, 70412, 70415, 70416, 70419, 70440, 70442, 70448, 70450, 70451, 70453, 70457, 70461, 70461, 70480, 70480, 70493, 70497, 70656, 70708, 70727, 70730, 70751, 70751, 70784, 70831, 70852, 70853, 70855, 70855, 71040, 71086, 71128, 71131, 71168, 71215, 71236, 71236, 71296, 71338, 71352, 71352, 71424, 71450, 71680, 71723, 71840, 71903, 71935, 71935, 72096, 72103, 72106, 72144, 72161, 72161, 72163, 72163, 72192, 72192, 72203, 72242, 72250, 72250, 72272, 72272, 72284, 72329, 72349, 72349, 72384, 72440, 72704, 72712, 72714, 72750, 72768, 72768, 72818, 72847, 72960, 72966, 72968, 72969, 72971, 73008, 73030, 73030, 73056, 73061, 73063, 73064, 73066, 73097, 73112, 73112, 73440, 73458, 73728, 74649, 74752, 74862, 74880, 75075, 77824, 78894, 82944, 83526, 92160, 92728, 92736, 92766, 92880, 92909, 92928, 92975, 92992, 92995, 93027, 93047, 93053, 93071, 93760, 93823, 93952, 94026, 94032, 94032, 94099, 94111, 94176, 94177, 94179, 94179, 94208, 100343, 100352, 101106, 110592, 110878, 110928, 110930, 110948, 110951, 110960, 111355, 113664, 113770, 113776, 113788, 113792, 113800, 113808, 113817, 119808, 119892, 119894, 119964, 119966, 119967, 119970, 119970, 119973, 119974, 119977, 119980, 119982, 119993, 119995, 119995, 119997, 120003, 120005, 120069, 120071, 120074, 120077, 120084, 120086, 120092, 120094, 120121, 120123, 120126, 120128, 120132, 120134, 120134, 120138, 120144, 120146, 120485, 120488, 120512, 120514, 120538, 120540, 120570, 120572, 120596, 120598, 120628, 120630, 120654, 120656, 120686, 120688, 120712, 120714, 120744, 120746, 120770, 120772, 120779, 123136, 123180, 123191, 123197, 123214, 123214, 123584, 123627, 124928, 125124, 125184, 125251, 125259, 125259, 126464, 126467, 126469, 126495, 126497, 126498, 126500, 126500, 126503, 126503, 126505, 126514, 126516, 126519, 126521, 126521, 126523, 126523, 126530, 126530, 126535, 126535, 126537, 126537, 126539, 126539, 126541, 126543, 126545, 126546, 126548, 126548, 126551, 126551, 126553, 126553, 126555, 126555, 126557, 126557, 126559, 126559, 126561, 126562, 126564, 126564, 126567, 126570, 126572, 126578, 126580, 126583, 126585, 126588, 126590, 126590, 126592, 126601, 126603, 126619, 126625, 126627, 126629, 126633, 126635, 126651, 131072, 173782, 173824, 177972, 177984, 178205, 178208, 183969, 183984, 191456, 194560, 195101];
        var unicodeESNextIdentifierPart = [48, 57, 65, 90, 95, 95, 97, 122, 170, 170, 181, 181, 183, 183, 186, 186, 192, 214, 216, 246, 248, 705, 710, 721, 736, 740, 748, 748, 750, 750, 768, 884, 886, 887, 890, 893, 895, 895, 902, 906, 908, 908, 910, 929, 931, 1013, 1015, 1153, 1155, 1159, 1162, 1327, 1329, 1366, 1369, 1369, 1376, 1416, 1425, 1469, 1471, 1471, 1473, 1474, 1476, 1477, 1479, 1479, 1488, 1514, 1519, 1522, 1552, 1562, 1568, 1641, 1646, 1747, 1749, 1756, 1759, 1768, 1770, 1788, 1791, 1791, 1808, 1866, 1869, 1969, 1984, 2037, 2042, 2042, 2045, 2045, 2048, 2093, 2112, 2139, 2144, 2154, 2208, 2228, 2230, 2237, 2259, 2273, 2275, 2403, 2406, 2415, 2417, 2435, 2437, 2444, 2447, 2448, 2451, 2472, 2474, 2480, 2482, 2482, 2486, 2489, 2492, 2500, 2503, 2504, 2507, 2510, 2519, 2519, 2524, 2525, 2527, 2531, 2534, 2545, 2556, 2556, 2558, 2558, 2561, 2563, 2565, 2570, 2575, 2576, 2579, 2600, 2602, 2608, 2610, 2611, 2613, 2614, 2616, 2617, 2620, 2620, 2622, 2626, 2631, 2632, 2635, 2637, 2641, 2641, 2649, 2652, 2654, 2654, 2662, 2677, 2689, 2691, 2693, 2701, 2703, 2705, 2707, 2728, 2730, 2736, 2738, 2739, 2741, 2745, 2748, 2757, 2759, 2761, 2763, 2765, 2768, 2768, 2784, 2787, 2790, 2799, 2809, 2815, 2817, 2819, 2821, 2828, 2831, 2832, 2835, 2856, 2858, 2864, 2866, 2867, 2869, 2873, 2876, 2884, 2887, 2888, 2891, 2893, 2902, 2903, 2908, 2909, 2911, 2915, 2918, 2927, 2929, 2929, 2946, 2947, 2949, 2954, 2958, 2960, 2962, 2965, 2969, 2970, 2972, 2972, 2974, 2975, 2979, 2980, 2984, 2986, 2990, 3001, 3006, 3010, 3014, 3016, 3018, 3021, 3024, 3024, 3031, 3031, 3046, 3055, 3072, 3084, 3086, 3088, 3090, 3112, 3114, 3129, 3133, 3140, 3142, 3144, 3146, 3149, 3157, 3158, 3160, 3162, 3168, 3171, 3174, 3183, 3200, 3203, 3205, 3212, 3214, 3216, 3218, 3240, 3242, 3251, 3253, 3257, 3260, 3268, 3270, 3272, 3274, 3277, 3285, 3286, 3294, 3294, 3296, 3299, 3302, 3311, 3313, 3314, 3328, 3331, 3333, 3340, 3342, 3344, 3346, 3396, 3398, 3400, 3402, 3406, 3412, 3415, 3423, 3427, 3430, 3439, 3450, 3455, 3458, 3459, 3461, 3478, 3482, 3505, 3507, 3515, 3517, 3517, 3520, 3526, 3530, 3530, 3535, 3540, 3542, 3542, 3544, 3551, 3558, 3567, 3570, 3571, 3585, 3642, 3648, 3662, 3664, 3673, 3713, 3714, 3716, 3716, 3718, 3722, 3724, 3747, 3749, 3749, 3751, 3773, 3776, 3780, 3782, 3782, 3784, 3789, 3792, 3801, 3804, 3807, 3840, 3840, 3864, 3865, 3872, 3881, 3893, 3893, 3895, 3895, 3897, 3897, 3902, 3911, 3913, 3948, 3953, 3972, 3974, 3991, 3993, 4028, 4038, 4038, 4096, 4169, 4176, 4253, 4256, 4293, 4295, 4295, 4301, 4301, 4304, 4346, 4348, 4680, 4682, 4685, 4688, 4694, 4696, 4696, 4698, 4701, 4704, 4744, 4746, 4749, 4752, 4784, 4786, 4789, 4792, 4798, 4800, 4800, 4802, 4805, 4808, 4822, 4824, 4880, 4882, 4885, 4888, 4954, 4957, 4959, 4969, 4977, 4992, 5007, 5024, 5109, 5112, 5117, 5121, 5740, 5743, 5759, 5761, 5786, 5792, 5866, 5870, 5880, 5888, 5900, 5902, 5908, 5920, 5940, 5952, 5971, 5984, 5996, 5998, 6000, 6002, 6003, 6016, 6099, 6103, 6103, 6108, 6109, 6112, 6121, 6155, 6157, 6160, 6169, 6176, 6264, 6272, 6314, 6320, 6389, 6400, 6430, 6432, 6443, 6448, 6459, 6470, 6509, 6512, 6516, 6528, 6571, 6576, 6601, 6608, 6618, 6656, 6683, 6688, 6750, 6752, 6780, 6783, 6793, 6800, 6809, 6823, 6823, 6832, 6845, 6912, 6987, 6992, 7001, 7019, 7027, 7040, 7155, 7168, 7223, 7232, 7241, 7245, 7293, 7296, 7304, 7312, 7354, 7357, 7359, 7376, 7378, 7380, 7418, 7424, 7673, 7675, 7957, 7960, 7965, 7968, 8005, 8008, 8013, 8016, 8023, 8025, 8025, 8027, 8027, 8029, 8029, 8031, 8061, 8064, 8116, 8118, 8124, 8126, 8126, 8130, 8132, 8134, 8140, 8144, 8147, 8150, 8155, 8160, 8172, 8178, 8180, 8182, 8188, 8255, 8256, 8276, 8276, 8305, 8305, 8319, 8319, 8336, 8348, 8400, 8412, 8417, 8417, 8421, 8432, 8450, 8450, 8455, 8455, 8458, 8467, 8469, 8469, 8472, 8477, 8484, 8484, 8486, 8486, 8488, 8488, 8490, 8505, 8508, 8511, 8517, 8521, 8526, 8526, 8544, 8584, 11264, 11310, 11312, 11358, 11360, 11492, 11499, 11507, 11520, 11557, 11559, 11559, 11565, 11565, 11568, 11623, 11631, 11631, 11647, 11670, 11680, 11686, 11688, 11694, 11696, 11702, 11704, 11710, 11712, 11718, 11720, 11726, 11728, 11734, 11736, 11742, 11744, 11775, 12293, 12295, 12321, 12335, 12337, 12341, 12344, 12348, 12353, 12438, 12441, 12447, 12449, 12538, 12540, 12543, 12549, 12591, 12593, 12686, 12704, 12730, 12784, 12799, 13312, 19893, 19968, 40943, 40960, 42124, 42192, 42237, 42240, 42508, 42512, 42539, 42560, 42607, 42612, 42621, 42623, 42737, 42775, 42783, 42786, 42888, 42891, 42943, 42946, 42950, 42999, 43047, 43072, 43123, 43136, 43205, 43216, 43225, 43232, 43255, 43259, 43259, 43261, 43309, 43312, 43347, 43360, 43388, 43392, 43456, 43471, 43481, 43488, 43518, 43520, 43574, 43584, 43597, 43600, 43609, 43616, 43638, 43642, 43714, 43739, 43741, 43744, 43759, 43762, 43766, 43777, 43782, 43785, 43790, 43793, 43798, 43808, 43814, 43816, 43822, 43824, 43866, 43868, 43879, 43888, 44010, 44012, 44013, 44016, 44025, 44032, 55203, 55216, 55238, 55243, 55291, 63744, 64109, 64112, 64217, 64256, 64262, 64275, 64279, 64285, 64296, 64298, 64310, 64312, 64316, 64318, 64318, 64320, 64321, 64323, 64324, 64326, 64433, 64467, 64829, 64848, 64911, 64914, 64967, 65008, 65019, 65024, 65039, 65056, 65071, 65075, 65076, 65101, 65103, 65136, 65140, 65142, 65276, 65296, 65305, 65313, 65338, 65343, 65343, 65345, 65370, 65382, 65470, 65474, 65479, 65482, 65487, 65490, 65495, 65498, 65500, 65536, 65547, 65549, 65574, 65576, 65594, 65596, 65597, 65599, 65613, 65616, 65629, 65664, 65786, 65856, 65908, 66045, 66045, 66176, 66204, 66208, 66256, 66272, 66272, 66304, 66335, 66349, 66378, 66384, 66426, 66432, 66461, 66464, 66499, 66504, 66511, 66513, 66517, 66560, 66717, 66720, 66729, 66736, 66771, 66776, 66811, 66816, 66855, 66864, 66915, 67072, 67382, 67392, 67413, 67424, 67431, 67584, 67589, 67592, 67592, 67594, 67637, 67639, 67640, 67644, 67644, 67647, 67669, 67680, 67702, 67712, 67742, 67808, 67826, 67828, 67829, 67840, 67861, 67872, 67897, 67968, 68023, 68030, 68031, 68096, 68099, 68101, 68102, 68108, 68115, 68117, 68119, 68121, 68149, 68152, 68154, 68159, 68159, 68192, 68220, 68224, 68252, 68288, 68295, 68297, 68326, 68352, 68405, 68416, 68437, 68448, 68466, 68480, 68497, 68608, 68680, 68736, 68786, 68800, 68850, 68864, 68903, 68912, 68921, 69376, 69404, 69415, 69415, 69424, 69456, 69600, 69622, 69632, 69702, 69734, 69743, 69759, 69818, 69840, 69864, 69872, 69881, 69888, 69940, 69942, 69951, 69956, 69958, 69968, 70003, 70006, 70006, 70016, 70084, 70089, 70092, 70096, 70106, 70108, 70108, 70144, 70161, 70163, 70199, 70206, 70206, 70272, 70278, 70280, 70280, 70282, 70285, 70287, 70301, 70303, 70312, 70320, 70378, 70384, 70393, 70400, 70403, 70405, 70412, 70415, 70416, 70419, 70440, 70442, 70448, 70450, 70451, 70453, 70457, 70459, 70468, 70471, 70472, 70475, 70477, 70480, 70480, 70487, 70487, 70493, 70499, 70502, 70508, 70512, 70516, 70656, 70730, 70736, 70745, 70750, 70751, 70784, 70853, 70855, 70855, 70864, 70873, 71040, 71093, 71096, 71104, 71128, 71133, 71168, 71232, 71236, 71236, 71248, 71257, 71296, 71352, 71360, 71369, 71424, 71450, 71453, 71467, 71472, 71481, 71680, 71738, 71840, 71913, 71935, 71935, 72096, 72103, 72106, 72151, 72154, 72161, 72163, 72164, 72192, 72254, 72263, 72263, 72272, 72345, 72349, 72349, 72384, 72440, 72704, 72712, 72714, 72758, 72760, 72768, 72784, 72793, 72818, 72847, 72850, 72871, 72873, 72886, 72960, 72966, 72968, 72969, 72971, 73014, 73018, 73018, 73020, 73021, 73023, 73031, 73040, 73049, 73056, 73061, 73063, 73064, 73066, 73102, 73104, 73105, 73107, 73112, 73120, 73129, 73440, 73462, 73728, 74649, 74752, 74862, 74880, 75075, 77824, 78894, 82944, 83526, 92160, 92728, 92736, 92766, 92768, 92777, 92880, 92909, 92912, 92916, 92928, 92982, 92992, 92995, 93008, 93017, 93027, 93047, 93053, 93071, 93760, 93823, 93952, 94026, 94031, 94087, 94095, 94111, 94176, 94177, 94179, 94179, 94208, 100343, 100352, 101106, 110592, 110878, 110928, 110930, 110948, 110951, 110960, 111355, 113664, 113770, 113776, 113788, 113792, 113800, 113808, 113817, 113821, 113822, 119141, 119145, 119149, 119154, 119163, 119170, 119173, 119179, 119210, 119213, 119362, 119364, 119808, 119892, 119894, 119964, 119966, 119967, 119970, 119970, 119973, 119974, 119977, 119980, 119982, 119993, 119995, 119995, 119997, 120003, 120005, 120069, 120071, 120074, 120077, 120084, 120086, 120092, 120094, 120121, 120123, 120126, 120128, 120132, 120134, 120134, 120138, 120144, 120146, 120485, 120488, 120512, 120514, 120538, 120540, 120570, 120572, 120596, 120598, 120628, 120630, 120654, 120656, 120686, 120688, 120712, 120714, 120744, 120746, 120770, 120772, 120779, 120782, 120831, 121344, 121398, 121403, 121452, 121461, 121461, 121476, 121476, 121499, 121503, 121505, 121519, 122880, 122886, 122888, 122904, 122907, 122913, 122915, 122916, 122918, 122922, 123136, 123180, 123184, 123197, 123200, 123209, 123214, 123214, 123584, 123641, 124928, 125124, 125136, 125142, 125184, 125259, 125264, 125273, 126464, 126467, 126469, 126495, 126497, 126498, 126500, 126500, 126503, 126503, 126505, 126514, 126516, 126519, 126521, 126521, 126523, 126523, 126530, 126530, 126535, 126535, 126537, 126537, 126539, 126539, 126541, 126543, 126545, 126546, 126548, 126548, 126551, 126551, 126553, 126553, 126555, 126555, 126557, 126557, 126559, 126559, 126561, 126562, 126564, 126564, 126567, 126570, 126572, 126578, 126580, 126583, 126585, 126588, 126590, 126590, 126592, 126601, 126603, 126619, 126625, 126627, 126629, 126633, 126635, 126651, 131072, 173782, 173824, 177972, 177984, 178205, 178208, 183969, 183984, 191456, 194560, 195101, 917760, 917999];
        /**
         * Test for whether a single line comment with leading whitespace trimmed's text contains a directive.
         */
        var commentDirectiveRegExSingleLine = /^\/\/\/?\s*@(ts-expect-error|ts-ignore)/;
        /**
         * Test for whether a multi-line comment with leading whitespace trimmed's last line contains a directive.
         */
        var commentDirectiveRegExMultiLine = /^(?:\/|\*)*\s*@(ts-expect-error|ts-ignore)/;
        function lookupInUnicodeMap(code, map) {
            // Bail out quickly if it couldn't possibly be in the map.
            if (code < map[0]) {
                return false;
            }
            // Perform binary search in one of the Unicode range maps
            var lo = 0;
            var hi = map.length;
            var mid;
            while (lo + 1 < hi) {
                mid = lo + (hi - lo) / 2;
                // mid has to be even to catch a range's beginning
                mid -= mid % 2;
                if (map[mid] <= code && code <= map[mid + 1]) {
                    return true;
                }
                if (code < map[mid]) {
                    hi = mid;
                }
                else {
                    lo = mid + 2;
                }
            }
            return false;
        }
        /* @internal */ function isUnicodeIdentifierStart(code, languageVersion) {
            return languageVersion >= 2 /* ES2015 */ ?
                lookupInUnicodeMap(code, unicodeESNextIdentifierStart) :
                languageVersion === 1 /* ES5 */ ? lookupInUnicodeMap(code, unicodeES5IdentifierStart) :
                    lookupInUnicodeMap(code, unicodeES3IdentifierStart);
        }
        ts.isUnicodeIdentifierStart = isUnicodeIdentifierStart;
        function isUnicodeIdentifierPart(code, languageVersion) {
            return languageVersion >= 2 /* ES2015 */ ?
                lookupInUnicodeMap(code, unicodeESNextIdentifierPart) :
                languageVersion === 1 /* ES5 */ ? lookupInUnicodeMap(code, unicodeES5IdentifierPart) :
                    lookupInUnicodeMap(code, unicodeES3IdentifierPart);
        }
        function makeReverseMap(source) {
            var result = [];
            source.forEach(function (value, name) {
                result[value] = name;
            });
            return result;
        }
        var tokenStrings = makeReverseMap(textToToken);
        function tokenToString(t) {
            return tokenStrings[t];
        }
        ts.tokenToString = tokenToString;
        /* @internal */
        function stringToToken(s) {
            return textToToken.get(s);
        }
        ts.stringToToken = stringToToken;
        /* @internal */
        function computeLineStarts(text) {
            var result = new Array();
            var pos = 0;
            var lineStart = 0;
            while (pos < text.length) {
                var ch = text.charCodeAt(pos);
                pos++;
                switch (ch) {
                    case 13 /* carriageReturn */:
                        if (text.charCodeAt(pos) === 10 /* lineFeed */) {
                            pos++;
                        }
                    // falls through
                    case 10 /* lineFeed */:
                        result.push(lineStart);
                        lineStart = pos;
                        break;
                    default:
                        if (ch > 127 /* maxAsciiCharacter */ && isLineBreak(ch)) {
                            result.push(lineStart);
                            lineStart = pos;
                        }
                        break;
                }
            }
            result.push(lineStart);
            return result;
        }
        ts.computeLineStarts = computeLineStarts;
        function getPositionOfLineAndCharacter(sourceFile, line, character, allowEdits) {
            return sourceFile.getPositionOfLineAndCharacter ?
                sourceFile.getPositionOfLineAndCharacter(line, character, allowEdits) :
                computePositionOfLineAndCharacter(getLineStarts(sourceFile), line, character, sourceFile.text, allowEdits);
        }
        ts.getPositionOfLineAndCharacter = getPositionOfLineAndCharacter;
        /* @internal */
        function computePositionOfLineAndCharacter(lineStarts, line, character, debugText, allowEdits) {
            if (line < 0 || line >= lineStarts.length) {
                if (allowEdits) {
                    // Clamp line to nearest allowable value
                    line = line < 0 ? 0 : line >= lineStarts.length ? lineStarts.length - 1 : line;
                }
                else {
                    ts.Debug.fail("Bad line number. Line: " + line + ", lineStarts.length: " + lineStarts.length + " , line map is correct? " + (debugText !== undefined ? ts.arraysEqual(lineStarts, computeLineStarts(debugText)) : "unknown"));
                }
            }
            var res = lineStarts[line] + character;
            if (allowEdits) {
                // Clamp to nearest allowable values to allow the underlying to be edited without crashing (accuracy is lost, instead)
                // TODO: Somehow track edits between file as it was during the creation of sourcemap we have and the current file and
                // apply them to the computed position to improve accuracy
                return res > lineStarts[line + 1] ? lineStarts[line + 1] : typeof debugText === "string" && res > debugText.length ? debugText.length : res;
            }
            if (line < lineStarts.length - 1) {
                ts.Debug.assert(res < lineStarts[line + 1]);
            }
            else if (debugText !== undefined) {
                ts.Debug.assert(res <= debugText.length); // Allow single character overflow for trailing newline
            }
            return res;
        }
        ts.computePositionOfLineAndCharacter = computePositionOfLineAndCharacter;
        /* @internal */
        function getLineStarts(sourceFile) {
            return sourceFile.lineMap || (sourceFile.lineMap = computeLineStarts(sourceFile.text));
        }
        ts.getLineStarts = getLineStarts;
        /* @internal */
        function computeLineAndCharacterOfPosition(lineStarts, position) {
            var lineNumber = computeLineOfPosition(lineStarts, position);
            return {
                line: lineNumber,
                character: position - lineStarts[lineNumber]
            };
        }
        ts.computeLineAndCharacterOfPosition = computeLineAndCharacterOfPosition;
        /**
         * @internal
         * We assume the first line starts at position 0 and 'position' is non-negative.
         */
        function computeLineOfPosition(lineStarts, position, lowerBound) {
            var lineNumber = ts.binarySearch(lineStarts, position, ts.identity, ts.compareValues, lowerBound);
            if (lineNumber < 0) {
                // If the actual position was not found,
                // the binary search returns the 2's-complement of the next line start
                // e.g. if the line starts at [5, 10, 23, 80] and the position requested was 20
                // then the search will return -2.
                //
                // We want the index of the previous line start, so we subtract 1.
                // Review 2's-complement if this is confusing.
                lineNumber = ~lineNumber - 1;
                ts.Debug.assert(lineNumber !== -1, "position cannot precede the beginning of the file");
            }
            return lineNumber;
        }
        ts.computeLineOfPosition = computeLineOfPosition;
        /** @internal */
        function getLinesBetweenPositions(sourceFile, pos1, pos2) {
            if (pos1 === pos2)
                return 0;
            var lineStarts = getLineStarts(sourceFile);
            var lower = Math.min(pos1, pos2);
            var isNegative = lower === pos2;
            var upper = isNegative ? pos1 : pos2;
            var lowerLine = computeLineOfPosition(lineStarts, lower);
            var upperLine = computeLineOfPosition(lineStarts, upper, lowerLine);
            return isNegative ? lowerLine - upperLine : upperLine - lowerLine;
        }
        ts.getLinesBetweenPositions = getLinesBetweenPositions;
        function getLineAndCharacterOfPosition(sourceFile, position) {
            return computeLineAndCharacterOfPosition(getLineStarts(sourceFile), position);
        }
        ts.getLineAndCharacterOfPosition = getLineAndCharacterOfPosition;
        function isWhiteSpaceLike(ch) {
            return isWhiteSpaceSingleLine(ch) || isLineBreak(ch);
        }
        ts.isWhiteSpaceLike = isWhiteSpaceLike;
        /** Does not include line breaks. For that, see isWhiteSpaceLike. */
        function isWhiteSpaceSingleLine(ch) {
            // Note: nextLine is in the Zs space, and should be considered to be a whitespace.
            // It is explicitly not a line-break as it isn't in the exact set specified by EcmaScript.
            return ch === 32 /* space */ ||
                ch === 9 /* tab */ ||
                ch === 11 /* verticalTab */ ||
                ch === 12 /* formFeed */ ||
                ch === 160 /* nonBreakingSpace */ ||
                ch === 133 /* nextLine */ ||
                ch === 5760 /* ogham */ ||
                ch >= 8192 /* enQuad */ && ch <= 8203 /* zeroWidthSpace */ ||
                ch === 8239 /* narrowNoBreakSpace */ ||
                ch === 8287 /* mathematicalSpace */ ||
                ch === 12288 /* ideographicSpace */ ||
                ch === 65279 /* byteOrderMark */;
        }
        ts.isWhiteSpaceSingleLine = isWhiteSpaceSingleLine;
        function isLineBreak(ch) {
            // ES5 7.3:
            // The ECMAScript line terminator characters are listed in Table 3.
            //     Table 3: Line Terminator Characters
            //     Code Unit Value     Name                    Formal Name
            //     \u000A              Line Feed               <LF>
            //     \u000D              Carriage Return         <CR>
            //     \u2028              Line separator          <LS>
            //     \u2029              Paragraph separator     <PS>
            // Only the characters in Table 3 are treated as line terminators. Other new line or line
            // breaking characters are treated as white space but not as line terminators.
            return ch === 10 /* lineFeed */ ||
                ch === 13 /* carriageReturn */ ||
                ch === 8232 /* lineSeparator */ ||
                ch === 8233 /* paragraphSeparator */;
        }
        ts.isLineBreak = isLineBreak;
        function isDigit(ch) {
            return ch >= 48 /* _0 */ && ch <= 57 /* _9 */;
        }
        function isHexDigit(ch) {
            return isDigit(ch) || ch >= 65 /* A */ && ch <= 70 /* F */ || ch >= 97 /* a */ && ch <= 102 /* f */;
        }
        function isCodePoint(code) {
            return code <= 0x10FFFF;
        }
        /* @internal */
        function isOctalDigit(ch) {
            return ch >= 48 /* _0 */ && ch <= 55 /* _7 */;
        }
        ts.isOctalDigit = isOctalDigit;
        function couldStartTrivia(text, pos) {
            // Keep in sync with skipTrivia
            var ch = text.charCodeAt(pos);
            switch (ch) {
                case 13 /* carriageReturn */:
                case 10 /* lineFeed */:
                case 9 /* tab */:
                case 11 /* verticalTab */:
                case 12 /* formFeed */:
                case 32 /* space */:
                case 47 /* slash */:
                // starts of normal trivia
                // falls through
                case 60 /* lessThan */:
                case 124 /* bar */:
                case 61 /* equals */:
                case 62 /* greaterThan */:
                    // Starts of conflict marker trivia
                    return true;
                case 35 /* hash */:
                    // Only if its the beginning can we have #! trivia
                    return pos === 0;
                default:
                    return ch > 127 /* maxAsciiCharacter */;
            }
        }
        ts.couldStartTrivia = couldStartTrivia;
        /* @internal */
        function skipTrivia(text, pos, stopAfterLineBreak, stopAtComments, inJSDoc) {
            if (ts.positionIsSynthesized(pos)) {
                return pos;
            }
            var canConsumeStar = false;
            // Keep in sync with couldStartTrivia
            while (true) {
                var ch = text.charCodeAt(pos);
                switch (ch) {
                    case 13 /* carriageReturn */:
                        if (text.charCodeAt(pos + 1) === 10 /* lineFeed */) {
                            pos++;
                        }
                    // falls through
                    case 10 /* lineFeed */:
                        pos++;
                        if (stopAfterLineBreak) {
                            return pos;
                        }
                        canConsumeStar = !!inJSDoc;
                        continue;
                    case 9 /* tab */:
                    case 11 /* verticalTab */:
                    case 12 /* formFeed */:
                    case 32 /* space */:
                        pos++;
                        continue;
                    case 47 /* slash */:
                        if (stopAtComments) {
                            break;
                        }
                        if (text.charCodeAt(pos + 1) === 47 /* slash */) {
                            pos += 2;
                            while (pos < text.length) {
                                if (isLineBreak(text.charCodeAt(pos))) {
                                    break;
                                }
                                pos++;
                            }
                            canConsumeStar = false;
                            continue;
                        }
                        if (text.charCodeAt(pos + 1) === 42 /* asterisk */) {
                            pos += 2;
                            while (pos < text.length) {
                                if (text.charCodeAt(pos) === 42 /* asterisk */ && text.charCodeAt(pos + 1) === 47 /* slash */) {
                                    pos += 2;
                                    break;
                                }
                                pos++;
                            }
                            canConsumeStar = false;
                            continue;
                        }
                        break;
                    case 60 /* lessThan */:
                    case 124 /* bar */:
                    case 61 /* equals */:
                    case 62 /* greaterThan */:
                        if (isConflictMarkerTrivia(text, pos)) {
                            pos = scanConflictMarkerTrivia(text, pos);
                            canConsumeStar = false;
                            continue;
                        }
                        break;
                    case 35 /* hash */:
                        if (pos === 0 && isShebangTrivia(text, pos)) {
                            pos = scanShebangTrivia(text, pos);
                            canConsumeStar = false;
                            continue;
                        }
                        break;
                    case 42 /* asterisk */:
                        if (canConsumeStar) {
                            pos++;
                            canConsumeStar = false;
                            continue;
                        }
                        break;
                    default:
                        if (ch > 127 /* maxAsciiCharacter */ && (isWhiteSpaceLike(ch))) {
                            pos++;
                            continue;
                        }
                        break;
                }
                return pos;
            }
        }
        ts.skipTrivia = skipTrivia;
        // All conflict markers consist of the same character repeated seven times.  If it is
        // a <<<<<<< or >>>>>>> marker then it is also followed by a space.
        var mergeConflictMarkerLength = "<<<<<<<".length;
        function isConflictMarkerTrivia(text, pos) {
            ts.Debug.assert(pos >= 0);
            // Conflict markers must be at the start of a line.
            if (pos === 0 || isLineBreak(text.charCodeAt(pos - 1))) {
                var ch = text.charCodeAt(pos);
                if ((pos + mergeConflictMarkerLength) < text.length) {
                    for (var i = 0; i < mergeConflictMarkerLength; i++) {
                        if (text.charCodeAt(pos + i) !== ch) {
                            return false;
                        }
                    }
                    return ch === 61 /* equals */ ||
                        text.charCodeAt(pos + mergeConflictMarkerLength) === 32 /* space */;
                }
            }
            return false;
        }
        function scanConflictMarkerTrivia(text, pos, error) {
            if (error) {
                error(ts.Diagnostics.Merge_conflict_marker_encountered, pos, mergeConflictMarkerLength);
            }
            var ch = text.charCodeAt(pos);
            var len = text.length;
            if (ch === 60 /* lessThan */ || ch === 62 /* greaterThan */) {
                while (pos < len && !isLineBreak(text.charCodeAt(pos))) {
                    pos++;
                }
            }
            else {
                ts.Debug.assert(ch === 124 /* bar */ || ch === 61 /* equals */);
                // Consume everything from the start of a ||||||| or ======= marker to the start
                // of the next ======= or >>>>>>> marker.
                while (pos < len) {
                    var currentChar = text.charCodeAt(pos);
                    if ((currentChar === 61 /* equals */ || currentChar === 62 /* greaterThan */) && currentChar !== ch && isConflictMarkerTrivia(text, pos)) {
                        break;
                    }
                    pos++;
                }
            }
            return pos;
        }
        var shebangTriviaRegex = /^#!.*/;
        /*@internal*/
        function isShebangTrivia(text, pos) {
            // Shebangs check must only be done at the start of the file
            ts.Debug.assert(pos === 0);
            return shebangTriviaRegex.test(text);
        }
        ts.isShebangTrivia = isShebangTrivia;
        /*@internal*/
        function scanShebangTrivia(text, pos) {
            var shebang = shebangTriviaRegex.exec(text)[0];
            pos = pos + shebang.length;
            return pos;
        }
        ts.scanShebangTrivia = scanShebangTrivia;
        /**
         * Invokes a callback for each comment range following the provided position.
         *
         * Single-line comment ranges include the leading double-slash characters but not the ending
         * line break. Multi-line comment ranges include the leading slash-asterisk and trailing
         * asterisk-slash characters.
         *
         * @param reduce If true, accumulates the result of calling the callback in a fashion similar
         *      to reduceLeft. If false, iteration stops when the callback returns a truthy value.
         * @param text The source text to scan.
         * @param pos The position at which to start scanning.
         * @param trailing If false, whitespace is skipped until the first line break and comments
         *      between that location and the next token are returned. If true, comments occurring
         *      between the given position and the next line break are returned.
         * @param cb The callback to execute as each comment range is encountered.
         * @param state A state value to pass to each iteration of the callback.
         * @param initial An initial value to pass when accumulating results (when "reduce" is true).
         * @returns If "reduce" is true, the accumulated value. If "reduce" is false, the first truthy
         *      return value of the callback.
         */
        function iterateCommentRanges(reduce, text, pos, trailing, cb, state, initial) {
            var pendingPos;
            var pendingEnd;
            var pendingKind;
            var pendingHasTrailingNewLine;
            var hasPendingCommentRange = false;
            var collecting = trailing;
            var accumulator = initial;
            if (pos === 0) {
                collecting = true;
                var shebang = getShebang(text);
                if (shebang) {
                    pos = shebang.length;
                }
            }
            scan: while (pos >= 0 && pos < text.length) {
                var ch = text.charCodeAt(pos);
                switch (ch) {
                    case 13 /* carriageReturn */:
                        if (text.charCodeAt(pos + 1) === 10 /* lineFeed */) {
                            pos++;
                        }
                    // falls through
                    case 10 /* lineFeed */:
                        pos++;
                        if (trailing) {
                            break scan;
                        }
                        collecting = true;
                        if (hasPendingCommentRange) {
                            pendingHasTrailingNewLine = true;
                        }
                        continue;
                    case 9 /* tab */:
                    case 11 /* verticalTab */:
                    case 12 /* formFeed */:
                    case 32 /* space */:
                        pos++;
                        continue;
                    case 47 /* slash */:
                        var nextChar = text.charCodeAt(pos + 1);
                        var hasTrailingNewLine = false;
                        if (nextChar === 47 /* slash */ || nextChar === 42 /* asterisk */) {
                            var kind = nextChar === 47 /* slash */ ? 2 /* SingleLineCommentTrivia */ : 3 /* MultiLineCommentTrivia */;
                            var startPos = pos;
                            pos += 2;
                            if (nextChar === 47 /* slash */) {
                                while (pos < text.length) {
                                    if (isLineBreak(text.charCodeAt(pos))) {
                                        hasTrailingNewLine = true;
                                        break;
                                    }
                                    pos++;
                                }
                            }
                            else {
                                while (pos < text.length) {
                                    if (text.charCodeAt(pos) === 42 /* asterisk */ && text.charCodeAt(pos + 1) === 47 /* slash */) {
                                        pos += 2;
                                        break;
                                    }
                                    pos++;
                                }
                            }
                            if (collecting) {
                                if (hasPendingCommentRange) {
                                    accumulator = cb(pendingPos, pendingEnd, pendingKind, pendingHasTrailingNewLine, state, accumulator);
                                    if (!reduce && accumulator) {
                                        // If we are not reducing and we have a truthy result, return it.
                                        return accumulator;
                                    }
                                }
                                pendingPos = startPos;
                                pendingEnd = pos;
                                pendingKind = kind;
                                pendingHasTrailingNewLine = hasTrailingNewLine;
                                hasPendingCommentRange = true;
                            }
                            continue;
                        }
                        break scan;
                    default:
                        if (ch > 127 /* maxAsciiCharacter */ && (isWhiteSpaceLike(ch))) {
                            if (hasPendingCommentRange && isLineBreak(ch)) {
                                pendingHasTrailingNewLine = true;
                            }
                            pos++;
                            continue;
                        }
                        break scan;
                }
            }
            if (hasPendingCommentRange) {
                accumulator = cb(pendingPos, pendingEnd, pendingKind, pendingHasTrailingNewLine, state, accumulator);
            }
            return accumulator;
        }
        function forEachLeadingCommentRange(text, pos, cb, state) {
            return iterateCommentRanges(/*reduce*/ false, text, pos, /*trailing*/ false, cb, state);
        }
        ts.forEachLeadingCommentRange = forEachLeadingCommentRange;
        function forEachTrailingCommentRange(text, pos, cb, state) {
            return iterateCommentRanges(/*reduce*/ false, text, pos, /*trailing*/ true, cb, state);
        }
        ts.forEachTrailingCommentRange = forEachTrailingCommentRange;
        function reduceEachLeadingCommentRange(text, pos, cb, state, initial) {
            return iterateCommentRanges(/*reduce*/ true, text, pos, /*trailing*/ false, cb, state, initial);
        }
        ts.reduceEachLeadingCommentRange = reduceEachLeadingCommentRange;
        function reduceEachTrailingCommentRange(text, pos, cb, state, initial) {
            return iterateCommentRanges(/*reduce*/ true, text, pos, /*trailing*/ true, cb, state, initial);
        }
        ts.reduceEachTrailingCommentRange = reduceEachTrailingCommentRange;
        function appendCommentRange(pos, end, kind, hasTrailingNewLine, _state, comments) {
            if (!comments) {
                comments = [];
            }
            comments.push({ kind: kind, pos: pos, end: end, hasTrailingNewLine: hasTrailingNewLine });
            return comments;
        }
        function getLeadingCommentRanges(text, pos) {
            return reduceEachLeadingCommentRange(text, pos, appendCommentRange, /*state*/ undefined, /*initial*/ undefined);
        }
        ts.getLeadingCommentRanges = getLeadingCommentRanges;
        function getTrailingCommentRanges(text, pos) {
            return reduceEachTrailingCommentRange(text, pos, appendCommentRange, /*state*/ undefined, /*initial*/ undefined);
        }
        ts.getTrailingCommentRanges = getTrailingCommentRanges;
        /** Optionally, get the shebang */
        function getShebang(text) {
            var match = shebangTriviaRegex.exec(text);
            if (match) {
                return match[0];
            }
        }
        ts.getShebang = getShebang;
        function isIdentifierStart(ch, languageVersion) {
            return ch >= 65 /* A */ && ch <= 90 /* Z */ || ch >= 97 /* a */ && ch <= 122 /* z */ ||
                ch === 36 /* $ */ || ch === 95 /* _ */ ||
                ch > 127 /* maxAsciiCharacter */ && isUnicodeIdentifierStart(ch, languageVersion);
        }
        ts.isIdentifierStart = isIdentifierStart;
        function isIdentifierPart(ch, languageVersion, identifierVariant) {
            return ch >= 65 /* A */ && ch <= 90 /* Z */ || ch >= 97 /* a */ && ch <= 122 /* z */ ||
                ch >= 48 /* _0 */ && ch <= 57 /* _9 */ || ch === 36 /* $ */ || ch === 95 /* _ */ ||
                // "-" and ":" are valid in JSX Identifiers
                (identifierVariant === 1 /* JSX */ ? (ch === 45 /* minus */ || ch === 58 /* colon */) : false) ||
                ch > 127 /* maxAsciiCharacter */ && isUnicodeIdentifierPart(ch, languageVersion);
        }
        ts.isIdentifierPart = isIdentifierPart;
        /* @internal */
        function isIdentifierText(name, languageVersion, identifierVariant) {
            var ch = codePointAt(name, 0);
            if (!isIdentifierStart(ch, languageVersion)) {
                return false;
            }
            for (var i = charSize(ch); i < name.length; i += charSize(ch)) {
                if (!isIdentifierPart(ch = codePointAt(name, i), languageVersion, identifierVariant)) {
                    return false;
                }
            }
            return true;
        }
        ts.isIdentifierText = isIdentifierText;
        // Creates a scanner over a (possibly unspecified) range of a piece of text.
        function createScanner(languageVersion, skipTrivia, languageVariant, textInitial, onError, start, length) {
            if (languageVariant === void 0) {
                languageVariant = 0 /* Standard */;
            }
            var text = textInitial;
            // Current position (end position of text of current token)
            var pos;
            // end of text
            var end;
            // Start position of whitespace before current token
            var startPos;
            // Start position of text of current token
            var tokenPos;
            var token;
            var tokenValue;
            var tokenFlags;
            var commentDirectives;
            var inJSDocType = 0;
            setText(text, start, length);
            var scanner = {
                getStartPos: function () { return startPos; },
                getTextPos: function () { return pos; },
                getToken: function () { return token; },
                getTokenPos: function () { return tokenPos; },
                getTokenText: function () { return text.substring(tokenPos, pos); },
                getTokenValue: function () { return tokenValue; },
                hasUnicodeEscape: function () { return (tokenFlags & 1024 /* UnicodeEscape */) !== 0; },
                hasExtendedUnicodeEscape: function () { return (tokenFlags & 8 /* ExtendedUnicodeEscape */) !== 0; },
                hasPrecedingLineBreak: function () { return (tokenFlags & 1 /* PrecedingLineBreak */) !== 0; },
                hasPrecedingJSDocComment: function () { return (tokenFlags & 2 /* PrecedingJSDocComment */) !== 0; },
                isIdentifier: function () { return token === 79 /* Identifier */ || token > 116 /* LastReservedWord */; },
                isReservedWord: function () { return token >= 81 /* FirstReservedWord */ && token <= 116 /* LastReservedWord */; },
                isUnterminated: function () { return (tokenFlags & 4 /* Unterminated */) !== 0; },
                getCommentDirectives: function () { return commentDirectives; },
                getNumericLiteralFlags: function () { return tokenFlags & 1008 /* NumericLiteralFlags */; },
                getTokenFlags: function () { return tokenFlags; },
                reScanGreaterToken: reScanGreaterToken,
                reScanAsteriskEqualsToken: reScanAsteriskEqualsToken,
                reScanSlashToken: reScanSlashToken,
                reScanTemplateToken: reScanTemplateToken,
                reScanTemplateHeadOrNoSubstitutionTemplate: reScanTemplateHeadOrNoSubstitutionTemplate,
                scanJsxIdentifier: scanJsxIdentifier,
                scanJsxAttributeValue: scanJsxAttributeValue,
                reScanJsxAttributeValue: reScanJsxAttributeValue,
                reScanJsxToken: reScanJsxToken,
                reScanLessThanToken: reScanLessThanToken,
                reScanHashToken: reScanHashToken,
                reScanQuestionToken: reScanQuestionToken,
                reScanInvalidIdentifier: reScanInvalidIdentifier,
                scanJsxToken: scanJsxToken,
                scanJsDocToken: scanJsDocToken,
                scan: scan,
                getText: getText,
                clearCommentDirectives: clearCommentDirectives,
                setText: setText,
                setScriptTarget: setScriptTarget,
                setLanguageVariant: setLanguageVariant,
                setOnError: setOnError,
                setTextPos: setTextPos,
                setInJSDocType: setInJSDocType,
                tryScan: tryScan,
                lookAhead: lookAhead,
                scanRange: scanRange,
            };
            if (ts.Debug.isDebugging) {
                Object.defineProperty(scanner, "__debugShowCurrentPositionInText", {
                    get: function () {
                        var text = scanner.getText();
                        return text.slice(0, scanner.getStartPos()) + "║" + text.slice(scanner.getStartPos());
                    },
                });
            }
            return scanner;
            function error(message, errPos, length) {
                if (errPos === void 0) {
                    errPos = pos;
                }
                if (onError) {
                    var oldPos = pos;
                    pos = errPos;
                    onError(message, length || 0);
                    pos = oldPos;
                }
            }
            function scanNumberFragment() {
                var start = pos;
                var allowSeparator = false;
                var isPreviousTokenSeparator = false;
                var result = "";
                while (true) {
                    var ch = text.charCodeAt(pos);
                    if (ch === 95 /* _ */) {
                        tokenFlags |= 512 /* ContainsSeparator */;
                        if (allowSeparator) {
                            allowSeparator = false;
                            isPreviousTokenSeparator = true;
                            result += text.substring(start, pos);
                        }
                        else if (isPreviousTokenSeparator) {
                            error(ts.Diagnostics.Multiple_consecutive_numeric_separators_are_not_permitted, pos, 1);
                        }
                        else {
                            error(ts.Diagnostics.Numeric_separators_are_not_allowed_here, pos, 1);
                        }
                        pos++;
                        start = pos;
                        continue;
                    }
                    if (isDigit(ch)) {
                        allowSeparator = true;
                        isPreviousTokenSeparator = false;
                        pos++;
                        continue;
                    }
                    break;
                }
                if (text.charCodeAt(pos - 1) === 95 /* _ */) {
                    error(ts.Diagnostics.Numeric_separators_are_not_allowed_here, pos - 1, 1);
                }
                return result + text.substring(start, pos);
            }
            function scanNumber() {
                var start = pos;
                var mainFragment = scanNumberFragment();
                var decimalFragment;
                var scientificFragment;
                if (text.charCodeAt(pos) === 46 /* dot */) {
                    pos++;
                    decimalFragment = scanNumberFragment();
                }
                var end = pos;
                if (text.charCodeAt(pos) === 69 /* E */ || text.charCodeAt(pos) === 101 /* e */) {
                    pos++;
                    tokenFlags |= 16 /* Scientific */;
                    if (text.charCodeAt(pos) === 43 /* plus */ || text.charCodeAt(pos) === 45 /* minus */)
                        pos++;
                    var preNumericPart = pos;
                    var finalFragment = scanNumberFragment();
                    if (!finalFragment) {
                        error(ts.Diagnostics.Digit_expected);
                    }
                    else {
                        scientificFragment = text.substring(end, preNumericPart) + finalFragment;
                        end = pos;
                    }
                }
                var result;
                if (tokenFlags & 512 /* ContainsSeparator */) {
                    result = mainFragment;
                    if (decimalFragment) {
                        result += "." + decimalFragment;
                    }
                    if (scientificFragment) {
                        result += scientificFragment;
                    }
                }
                else {
                    result = text.substring(start, end); // No need to use all the fragments; no _ removal needed
                }
                if (decimalFragment !== undefined || tokenFlags & 16 /* Scientific */) {
                    checkForIdentifierStartAfterNumericLiteral(start, decimalFragment === undefined && !!(tokenFlags & 16 /* Scientific */));
                    return {
                        type: 8 /* NumericLiteral */,
                        value: "" + +result // if value is not an integer, it can be safely coerced to a number
                    };
                }
                else {
                    tokenValue = result;
                    var type = checkBigIntSuffix(); // if value is an integer, check whether it is a bigint
                    checkForIdentifierStartAfterNumericLiteral(start);
                    return { type: type, value: tokenValue };
                }
            }
            function checkForIdentifierStartAfterNumericLiteral(numericStart, isScientific) {
                if (!isIdentifierStart(codePointAt(text, pos), languageVersion)) {
                    return;
                }
                var identifierStart = pos;
                var length = scanIdentifierParts().length;
                if (length === 1 && text[identifierStart] === "n") {
                    if (isScientific) {
                        error(ts.Diagnostics.A_bigint_literal_cannot_use_exponential_notation, numericStart, identifierStart - numericStart + 1);
                    }
                    else {
                        error(ts.Diagnostics.A_bigint_literal_must_be_an_integer, numericStart, identifierStart - numericStart + 1);
                    }
                }
                else {
                    error(ts.Diagnostics.An_identifier_or_keyword_cannot_immediately_follow_a_numeric_literal, identifierStart, length);
                    pos = identifierStart;
                }
            }
            function scanOctalDigits() {
                var start = pos;
                while (isOctalDigit(text.charCodeAt(pos))) {
                    pos++;
                }
                return +(text.substring(start, pos));
            }
            /**
             * Scans the given number of hexadecimal digits in the text,
             * returning -1 if the given number is unavailable.
             */
            function scanExactNumberOfHexDigits(count, canHaveSeparators) {
                var valueString = scanHexDigits(/*minCount*/ count, /*scanAsManyAsPossible*/ false, canHaveSeparators);
                return valueString ? parseInt(valueString, 16) : -1;
            }
            /**
             * Scans as many hexadecimal digits as are available in the text,
             * returning "" if the given number of digits was unavailable.
             */
            function scanMinimumNumberOfHexDigits(count, canHaveSeparators) {
                return scanHexDigits(/*minCount*/ count, /*scanAsManyAsPossible*/ true, canHaveSeparators);
            }
            function scanHexDigits(minCount, scanAsManyAsPossible, canHaveSeparators) {
                var valueChars = [];
                var allowSeparator = false;
                var isPreviousTokenSeparator = false;
                while (valueChars.length < minCount || scanAsManyAsPossible) {
                    var ch = text.charCodeAt(pos);
                    if (canHaveSeparators && ch === 95 /* _ */) {
                        tokenFlags |= 512 /* ContainsSeparator */;
                        if (allowSeparator) {
                            allowSeparator = false;
                            isPreviousTokenSeparator = true;
                        }
                        else if (isPreviousTokenSeparator) {
                            error(ts.Diagnostics.Multiple_consecutive_numeric_separators_are_not_permitted, pos, 1);
                        }
                        else {
                            error(ts.Diagnostics.Numeric_separators_are_not_allowed_here, pos, 1);
                        }
                        pos++;
                        continue;
                    }
                    allowSeparator = canHaveSeparators;
                    if (ch >= 65 /* A */ && ch <= 70 /* F */) {
                        ch += 97 /* a */ - 65 /* A */; // standardize hex literals to lowercase
                    }
                    else if (!((ch >= 48 /* _0 */ && ch <= 57 /* _9 */) ||
                        (ch >= 97 /* a */ && ch <= 102 /* f */))) {
                        break;
                    }
                    valueChars.push(ch);
                    pos++;
                    isPreviousTokenSeparator = false;
                }
                if (valueChars.length < minCount) {
                    valueChars = [];
                }
                if (text.charCodeAt(pos - 1) === 95 /* _ */) {
                    error(ts.Diagnostics.Numeric_separators_are_not_allowed_here, pos - 1, 1);
                }
                return String.fromCharCode.apply(String, valueChars);
            }
            function scanString(jsxAttributeString) {
                if (jsxAttributeString === void 0) {
                    jsxAttributeString = false;
                }
                var quote = text.charCodeAt(pos);
                pos++;
                var result = "";
                var start = pos;
                while (true) {
                    if (pos >= end) {
                        result += text.substring(start, pos);
                        tokenFlags |= 4 /* Unterminated */;
                        error(ts.Diagnostics.Unterminated_string_literal);
                        break;
                    }
                    var ch = text.charCodeAt(pos);
                    if (ch === quote) {
                        result += text.substring(start, pos);
                        pos++;
                        break;
                    }
                    if (ch === 92 /* backslash */ && !jsxAttributeString) {
                        result += text.substring(start, pos);
                        result += scanEscapeSequence();
                        start = pos;
                        continue;
                    }
                    if (isLineBreak(ch) && !jsxAttributeString) {
                        result += text.substring(start, pos);
                        tokenFlags |= 4 /* Unterminated */;
                        error(ts.Diagnostics.Unterminated_string_literal);
                        break;
                    }
                    pos++;
                }
                return result;
            }
            /**
             * Sets the current 'tokenValue' and returns a NoSubstitutionTemplateLiteral or
             * a literal component of a TemplateExpression.
             */
            function scanTemplateAndSetTokenValue(isTaggedTemplate) {
                var startedWithBacktick = text.charCodeAt(pos) === 96 /* backtick */;
                pos++;
                var start = pos;
                var contents = "";
                var resultingToken;
                while (true) {
                    if (pos >= end) {
                        contents += text.substring(start, pos);
                        tokenFlags |= 4 /* Unterminated */;
                        error(ts.Diagnostics.Unterminated_template_literal);
                        resultingToken = startedWithBacktick ? 14 /* NoSubstitutionTemplateLiteral */ : 17 /* TemplateTail */;
                        break;
                    }
                    var currChar = text.charCodeAt(pos);
                    // '`'
                    if (currChar === 96 /* backtick */) {
                        contents += text.substring(start, pos);
                        pos++;
                        resultingToken = startedWithBacktick ? 14 /* NoSubstitutionTemplateLiteral */ : 17 /* TemplateTail */;
                        break;
                    }
                    // '${'
                    if (currChar === 36 /* $ */ && pos + 1 < end && text.charCodeAt(pos + 1) === 123 /* openBrace */) {
                        contents += text.substring(start, pos);
                        pos += 2;
                        resultingToken = startedWithBacktick ? 15 /* TemplateHead */ : 16 /* TemplateMiddle */;
                        break;
                    }
                    // Escape character
                    if (currChar === 92 /* backslash */) {
                        contents += text.substring(start, pos);
                        contents += scanEscapeSequence(isTaggedTemplate);
                        start = pos;
                        continue;
                    }
                    // Speculated ECMAScript 6 Spec 11.8.6.1:
                    // <CR><LF> and <CR> LineTerminatorSequences are normalized to <LF> for Template Values
                    if (currChar === 13 /* carriageReturn */) {
                        contents += text.substring(start, pos);
                        pos++;
                        if (pos < end && text.charCodeAt(pos) === 10 /* lineFeed */) {
                            pos++;
                        }
                        contents += "\n";
                        start = pos;
                        continue;
                    }
                    pos++;
                }
                ts.Debug.assert(resultingToken !== undefined);
                tokenValue = contents;
                return resultingToken;
            }
            function scanEscapeSequence(isTaggedTemplate) {
                var start = pos;
                pos++;
                if (pos >= end) {
                    error(ts.Diagnostics.Unexpected_end_of_text);
                    return "";
                }
                var ch = text.charCodeAt(pos);
                pos++;
                switch (ch) {
                    case 48 /* _0 */:
                        // '\01'
                        if (isTaggedTemplate && pos < end && isDigit(text.charCodeAt(pos))) {
                            pos++;
                            tokenFlags |= 2048 /* ContainsInvalidEscape */;
                            return text.substring(start, pos);
                        }
                        return "\0";
                    case 98 /* b */:
                        return "\b";
                    case 116 /* t */:
                        return "\t";
                    case 110 /* n */:
                        return "\n";
                    case 118 /* v */:
                        return "\v";
                    case 102 /* f */:
                        return "\f";
                    case 114 /* r */:
                        return "\r";
                    case 39 /* singleQuote */:
                        return "\'";
                    case 34 /* doubleQuote */:
                        return "\"";
                    case 117 /* u */:
                        if (isTaggedTemplate) {
                            // '\u' or '\u0' or '\u00' or '\u000'
                            for (var escapePos = pos; escapePos < pos + 4; escapePos++) {
                                if (escapePos < end && !isHexDigit(text.charCodeAt(escapePos)) && text.charCodeAt(escapePos) !== 123 /* openBrace */) {
                                    pos = escapePos;
                                    tokenFlags |= 2048 /* ContainsInvalidEscape */;
                                    return text.substring(start, pos);
                                }
                            }
                        }
                        // '\u{DDDDDDDD}'
                        if (pos < end && text.charCodeAt(pos) === 123 /* openBrace */) {
                            pos++;
                            // '\u{'
                            if (isTaggedTemplate && !isHexDigit(text.charCodeAt(pos))) {
                                tokenFlags |= 2048 /* ContainsInvalidEscape */;
                                return text.substring(start, pos);
                            }
                            if (isTaggedTemplate) {
                                var savePos = pos;
                                var escapedValueString = scanMinimumNumberOfHexDigits(1, /*canHaveSeparators*/ false);
                                var escapedValue = escapedValueString ? parseInt(escapedValueString, 16) : -1;
                                // '\u{Not Code Point' or '\u{CodePoint'
                                if (!isCodePoint(escapedValue) || text.charCodeAt(pos) !== 125 /* closeBrace */) {
                                    tokenFlags |= 2048 /* ContainsInvalidEscape */;
                                    return text.substring(start, pos);
                                }
                                else {
                                    pos = savePos;
                                }
                            }
                            tokenFlags |= 8 /* ExtendedUnicodeEscape */;
                            return scanExtendedUnicodeEscape();
                        }
                        tokenFlags |= 1024 /* UnicodeEscape */;
                        // '\uDDDD'
                        return scanHexadecimalEscape(/*numDigits*/ 4);
                    case 120 /* x */:
                        if (isTaggedTemplate) {
                            if (!isHexDigit(text.charCodeAt(pos))) {
                                tokenFlags |= 2048 /* ContainsInvalidEscape */;
                                return text.substring(start, pos);
                            }
                            else if (!isHexDigit(text.charCodeAt(pos + 1))) {
                                pos++;
                                tokenFlags |= 2048 /* ContainsInvalidEscape */;
                                return text.substring(start, pos);
                            }
                        }
                        // '\xDD'
                        return scanHexadecimalEscape(/*numDigits*/ 2);
                    // when encountering a LineContinuation (i.e. a backslash and a line terminator sequence),
                    // the line terminator is interpreted to be "the empty code unit sequence".
                    case 13 /* carriageReturn */:
                        if (pos < end && text.charCodeAt(pos) === 10 /* lineFeed */) {
                            pos++;
                        }
                    // falls through
                    case 10 /* lineFeed */:
                    case 8232 /* lineSeparator */:
                    case 8233 /* paragraphSeparator */:
                        return "";
                    default:
                        return String.fromCharCode(ch);
                }
            }
            function scanHexadecimalEscape(numDigits) {
                var escapedValue = scanExactNumberOfHexDigits(numDigits, /*canHaveSeparators*/ false);
                if (escapedValue >= 0) {
                    return String.fromCharCode(escapedValue);
                }
                else {
                    error(ts.Diagnostics.Hexadecimal_digit_expected);
                    return "";
                }
            }
            function scanExtendedUnicodeEscape() {
                var escapedValueString = scanMinimumNumberOfHexDigits(1, /*canHaveSeparators*/ false);
                var escapedValue = escapedValueString ? parseInt(escapedValueString, 16) : -1;
                var isInvalidExtendedEscape = false;
                // Validate the value of the digit
                if (escapedValue < 0) {
                    error(ts.Diagnostics.Hexadecimal_digit_expected);
                    isInvalidExtendedEscape = true;
                }
                else if (escapedValue > 0x10FFFF) {
                    error(ts.Diagnostics.An_extended_Unicode_escape_value_must_be_between_0x0_and_0x10FFFF_inclusive);
                    isInvalidExtendedEscape = true;
                }
                if (pos >= end) {
                    error(ts.Diagnostics.Unexpected_end_of_text);
                    isInvalidExtendedEscape = true;
                }
                else if (text.charCodeAt(pos) === 125 /* closeBrace */) {
                    // Only swallow the following character up if it's a '}'.
                    pos++;
                }
                else {
                    error(ts.Diagnostics.Unterminated_Unicode_escape_sequence);
                    isInvalidExtendedEscape = true;
                }
                if (isInvalidExtendedEscape) {
                    return "";
                }
                return utf16EncodeAsString(escapedValue);
            }
            // Current character is known to be a backslash. Check for Unicode escape of the form '\uXXXX'
            // and return code point value if valid Unicode escape is found. Otherwise return -1.
            function peekUnicodeEscape() {
                if (pos + 5 < end && text.charCodeAt(pos + 1) === 117 /* u */) {
                    var start_1 = pos;
                    pos += 2;
                    var value = scanExactNumberOfHexDigits(4, /*canHaveSeparators*/ false);
                    pos = start_1;
                    return value;
                }
                return -1;
            }
            function peekExtendedUnicodeEscape() {
                if (languageVersion >= 2 /* ES2015 */ && codePointAt(text, pos + 1) === 117 /* u */ && codePointAt(text, pos + 2) === 123 /* openBrace */) {
                    var start_2 = pos;
                    pos += 3;
                    var escapedValueString = scanMinimumNumberOfHexDigits(1, /*canHaveSeparators*/ false);
                    var escapedValue = escapedValueString ? parseInt(escapedValueString, 16) : -1;
                    pos = start_2;
                    return escapedValue;
                }
                return -1;
            }
            function scanIdentifierParts() {
                var result = "";
                var start = pos;
                while (pos < end) {
                    var ch = codePointAt(text, pos);
                    if (isIdentifierPart(ch, languageVersion)) {
                        pos += charSize(ch);
                    }
                    else if (ch === 92 /* backslash */) {
                        ch = peekExtendedUnicodeEscape();
                        if (ch >= 0 && isIdentifierPart(ch, languageVersion)) {
                            pos += 3;
                            tokenFlags |= 8 /* ExtendedUnicodeEscape */;
                            result += scanExtendedUnicodeEscape();
                            start = pos;
                            continue;
                        }
                        ch = peekUnicodeEscape();
                        if (!(ch >= 0 && isIdentifierPart(ch, languageVersion))) {
                            break;
                        }
                        tokenFlags |= 1024 /* UnicodeEscape */;
                        result += text.substring(start, pos);
                        result += utf16EncodeAsString(ch);
                        // Valid Unicode escape is always six characters
                        pos += 6;
                        start = pos;
                    }
                    else {
                        break;
                    }
                }
                result += text.substring(start, pos);
                return result;
            }
            function getIdentifierToken() {
                // Reserved words are between 2 and 12 characters long and start with a lowercase letter
                var len = tokenValue.length;
                if (len >= 2 && len <= 12) {
                    var ch = tokenValue.charCodeAt(0);
                    if (ch >= 97 /* a */ && ch <= 122 /* z */) {
                        var keyword = textToKeyword.get(tokenValue);
                        if (keyword !== undefined) {
                            return token = keyword;
                        }
                    }
                }
                return token = 79 /* Identifier */;
            }
            function scanBinaryOrOctalDigits(base) {
                var value = "";
                // For counting number of digits; Valid binaryIntegerLiteral must have at least one binary digit following B or b.
                // Similarly valid octalIntegerLiteral must have at least one octal digit following o or O.
                var separatorAllowed = false;
                var isPreviousTokenSeparator = false;
                while (true) {
                    var ch = text.charCodeAt(pos);
                    // Numeric separators are allowed anywhere within a numeric literal, except not at the beginning, or following another separator
                    if (ch === 95 /* _ */) {
                        tokenFlags |= 512 /* ContainsSeparator */;
                        if (separatorAllowed) {
                            separatorAllowed = false;
                            isPreviousTokenSeparator = true;
                        }
                        else if (isPreviousTokenSeparator) {
                            error(ts.Diagnostics.Multiple_consecutive_numeric_separators_are_not_permitted, pos, 1);
                        }
                        else {
                            error(ts.Diagnostics.Numeric_separators_are_not_allowed_here, pos, 1);
                        }
                        pos++;
                        continue;
                    }
                    separatorAllowed = true;
                    if (!isDigit(ch) || ch - 48 /* _0 */ >= base) {
                        break;
                    }
                    value += text[pos];
                    pos++;
                    isPreviousTokenSeparator = false;
                }
                if (text.charCodeAt(pos - 1) === 95 /* _ */) {
                    // Literal ends with underscore - not allowed
                    error(ts.Diagnostics.Numeric_separators_are_not_allowed_here, pos - 1, 1);
                }
                return value;
            }
            function checkBigIntSuffix() {
                if (text.charCodeAt(pos) === 110 /* n */) {
                    tokenValue += "n";
                    // Use base 10 instead of base 2 or base 8 for shorter literals
                    if (tokenFlags & 384 /* BinaryOrOctalSpecifier */) {
                        tokenValue = ts.parsePseudoBigInt(tokenValue) + "n";
                    }
                    pos++;
                    return 9 /* BigIntLiteral */;
                }
                else { // not a bigint, so can convert to number in simplified form
                    // Number() may not support 0b or 0o, so use parseInt() instead
                    var numericValue = tokenFlags & 128 /* BinarySpecifier */
                        ? parseInt(tokenValue.slice(2), 2) // skip "0b"
                        : tokenFlags & 256 /* OctalSpecifier */
                            ? parseInt(tokenValue.slice(2), 8) // skip "0o"
                            : +tokenValue;
                    tokenValue = "" + numericValue;
                    return 8 /* NumericLiteral */;
                }
            }
            function scan() {
                var _a;
                startPos = pos;
                tokenFlags = 0 /* None */;
                var asteriskSeen = false;
                while (true) {
                    tokenPos = pos;
                    if (pos >= end) {
                        return token = 1 /* EndOfFileToken */;
                    }
                    var ch = codePointAt(text, pos);
                    // Special handling for shebang
                    if (ch === 35 /* hash */ && pos === 0 && isShebangTrivia(text, pos)) {
                        pos = scanShebangTrivia(text, pos);
                        if (skipTrivia) {
                            continue;
                        }
                        else {
                            return token = 6 /* ShebangTrivia */;
                        }
                    }
                    switch (ch) {
                        case 10 /* lineFeed */:
                        case 13 /* carriageReturn */:
                            tokenFlags |= 1 /* PrecedingLineBreak */;
                            if (skipTrivia) {
                                pos++;
                                continue;
                            }
                            else {
                                if (ch === 13 /* carriageReturn */ && pos + 1 < end && text.charCodeAt(pos + 1) === 10 /* lineFeed */) {
                                    // consume both CR and LF
                                    pos += 2;
                                }
                                else {
                                    pos++;
                                }
                                return token = 4 /* NewLineTrivia */;
                            }
                        case 9 /* tab */:
                        case 11 /* verticalTab */:
                        case 12 /* formFeed */:
                        case 32 /* space */:
                        case 160 /* nonBreakingSpace */:
                        case 5760 /* ogham */:
                        case 8192 /* enQuad */:
                        case 8193 /* emQuad */:
                        case 8194 /* enSpace */:
                        case 8195 /* emSpace */:
                        case 8196 /* threePerEmSpace */:
                        case 8197 /* fourPerEmSpace */:
                        case 8198 /* sixPerEmSpace */:
                        case 8199 /* figureSpace */:
                        case 8200 /* punctuationSpace */:
                        case 8201 /* thinSpace */:
                        case 8202 /* hairSpace */:
                        case 8203 /* zeroWidthSpace */:
                        case 8239 /* narrowNoBreakSpace */:
                        case 8287 /* mathematicalSpace */:
                        case 12288 /* ideographicSpace */:
                        case 65279 /* byteOrderMark */:
                            if (skipTrivia) {
                                pos++;
                                continue;
                            }
                            else {
                                while (pos < end && isWhiteSpaceSingleLine(text.charCodeAt(pos))) {
                                    pos++;
                                }
                                return token = 5 /* WhitespaceTrivia */;
                            }
                        case 33 /* exclamation */:
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                if (text.charCodeAt(pos + 2) === 61 /* equals */) {
                                    return pos += 3, token = 37 /* ExclamationEqualsEqualsToken */;
                                }
                                return pos += 2, token = 35 /* ExclamationEqualsToken */;
                            }
                            pos++;
                            return token = 53 /* ExclamationToken */;
                        case 34 /* doubleQuote */:
                        case 39 /* singleQuote */:
                            tokenValue = scanString();
                            return token = 10 /* StringLiteral */;
                        case 96 /* backtick */:
                            return token = scanTemplateAndSetTokenValue(/* isTaggedTemplate */ false);
                        case 37 /* percent */:
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                return pos += 2, token = 69 /* PercentEqualsToken */;
                            }
                            pos++;
                            return token = 44 /* PercentToken */;
                        case 38 /* ampersand */:
                            if (text.charCodeAt(pos + 1) === 38 /* ampersand */) {
                                if (text.charCodeAt(pos + 2) === 61 /* equals */) {
                                    return pos += 3, token = 76 /* AmpersandAmpersandEqualsToken */;
                                }
                                return pos += 2, token = 55 /* AmpersandAmpersandToken */;
                            }
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                return pos += 2, token = 73 /* AmpersandEqualsToken */;
                            }
                            pos++;
                            return token = 50 /* AmpersandToken */;
                        case 40 /* openParen */:
                            pos++;
                            return token = 20 /* OpenParenToken */;
                        case 41 /* closeParen */:
                            pos++;
                            return token = 21 /* CloseParenToken */;
                        case 42 /* asterisk */:
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                return pos += 2, token = 66 /* AsteriskEqualsToken */;
                            }
                            if (text.charCodeAt(pos + 1) === 42 /* asterisk */) {
                                if (text.charCodeAt(pos + 2) === 61 /* equals */) {
                                    return pos += 3, token = 67 /* AsteriskAsteriskEqualsToken */;
                                }
                                return pos += 2, token = 42 /* AsteriskAsteriskToken */;
                            }
                            pos++;
                            if (inJSDocType && !asteriskSeen && (tokenFlags & 1 /* PrecedingLineBreak */)) {
                                // decoration at the start of a JSDoc comment line
                                asteriskSeen = true;
                                continue;
                            }
                            return token = 41 /* AsteriskToken */;
                        case 43 /* plus */:
                            if (text.charCodeAt(pos + 1) === 43 /* plus */) {
                                return pos += 2, token = 45 /* PlusPlusToken */;
                            }
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                return pos += 2, token = 64 /* PlusEqualsToken */;
                            }
                            pos++;
                            return token = 39 /* PlusToken */;
                        case 44 /* comma */:
                            pos++;
                            return token = 27 /* CommaToken */;
                        case 45 /* minus */:
                            if (text.charCodeAt(pos + 1) === 45 /* minus */) {
                                return pos += 2, token = 46 /* MinusMinusToken */;
                            }
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                return pos += 2, token = 65 /* MinusEqualsToken */;
                            }
                            pos++;
                            return token = 40 /* MinusToken */;
                        case 46 /* dot */:
                            if (isDigit(text.charCodeAt(pos + 1))) {
                                tokenValue = scanNumber().value;
                                return token = 8 /* NumericLiteral */;
                            }
                            if (text.charCodeAt(pos + 1) === 46 /* dot */ && text.charCodeAt(pos + 2) === 46 /* dot */) {
                                return pos += 3, token = 25 /* DotDotDotToken */;
                            }
                            pos++;
                            return token = 24 /* DotToken */;
                        case 47 /* slash */:
                            // Single-line comment
                            if (text.charCodeAt(pos + 1) === 47 /* slash */) {
                                pos += 2;
                                while (pos < end) {
                                    if (isLineBreak(text.charCodeAt(pos))) {
                                        break;
                                    }
                                    pos++;
                                }
                                commentDirectives = appendIfCommentDirective(commentDirectives, text.slice(tokenPos, pos), commentDirectiveRegExSingleLine, tokenPos);
                                if (skipTrivia) {
                                    continue;
                                }
                                else {
                                    return token = 2 /* SingleLineCommentTrivia */;
                                }
                            }
                            // Multi-line comment
                            if (text.charCodeAt(pos + 1) === 42 /* asterisk */) {
                                pos += 2;
                                if (text.charCodeAt(pos) === 42 /* asterisk */ && text.charCodeAt(pos + 1) !== 47 /* slash */) {
                                    tokenFlags |= 2 /* PrecedingJSDocComment */;
                                }
                                var commentClosed = false;
                                var lastLineStart = tokenPos;
                                while (pos < end) {
                                    var ch_1 = text.charCodeAt(pos);
                                    if (ch_1 === 42 /* asterisk */ && text.charCodeAt(pos + 1) === 47 /* slash */) {
                                        pos += 2;
                                        commentClosed = true;
                                        break;
                                    }
                                    pos++;
                                    if (isLineBreak(ch_1)) {
                                        lastLineStart = pos;
                                        tokenFlags |= 1 /* PrecedingLineBreak */;
                                    }
                                }
                                commentDirectives = appendIfCommentDirective(commentDirectives, text.slice(lastLineStart, pos), commentDirectiveRegExMultiLine, lastLineStart);
                                if (!commentClosed) {
                                    error(ts.Diagnostics.Asterisk_Slash_expected);
                                }
                                if (skipTrivia) {
                                    continue;
                                }
                                else {
                                    if (!commentClosed) {
                                        tokenFlags |= 4 /* Unterminated */;
                                    }
                                    return token = 3 /* MultiLineCommentTrivia */;
                                }
                            }
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                return pos += 2, token = 68 /* SlashEqualsToken */;
                            }
                            pos++;
                            return token = 43 /* SlashToken */;
                        case 48 /* _0 */:
                            if (pos + 2 < end && (text.charCodeAt(pos + 1) === 88 /* X */ || text.charCodeAt(pos + 1) === 120 /* x */)) {
                                pos += 2;
                                tokenValue = scanMinimumNumberOfHexDigits(1, /*canHaveSeparators*/ true);
                                if (!tokenValue) {
                                    error(ts.Diagnostics.Hexadecimal_digit_expected);
                                    tokenValue = "0";
                                }
                                tokenValue = "0x" + tokenValue;
                                tokenFlags |= 64 /* HexSpecifier */;
                                return token = checkBigIntSuffix();
                            }
                            else if (pos + 2 < end && (text.charCodeAt(pos + 1) === 66 /* B */ || text.charCodeAt(pos + 1) === 98 /* b */)) {
                                pos += 2;
                                tokenValue = scanBinaryOrOctalDigits(/* base */ 2);
                                if (!tokenValue) {
                                    error(ts.Diagnostics.Binary_digit_expected);
                                    tokenValue = "0";
                                }
                                tokenValue = "0b" + tokenValue;
                                tokenFlags |= 128 /* BinarySpecifier */;
                                return token = checkBigIntSuffix();
                            }
                            else if (pos + 2 < end && (text.charCodeAt(pos + 1) === 79 /* O */ || text.charCodeAt(pos + 1) === 111 /* o */)) {
                                pos += 2;
                                tokenValue = scanBinaryOrOctalDigits(/* base */ 8);
                                if (!tokenValue) {
                                    error(ts.Diagnostics.Octal_digit_expected);
                                    tokenValue = "0";
                                }
                                tokenValue = "0o" + tokenValue;
                                tokenFlags |= 256 /* OctalSpecifier */;
                                return token = checkBigIntSuffix();
                            }
                            // Try to parse as an octal
                            if (pos + 1 < end && isOctalDigit(text.charCodeAt(pos + 1))) {
                                tokenValue = "" + scanOctalDigits();
                                tokenFlags |= 32 /* Octal */;
                                return token = 8 /* NumericLiteral */;
                            }
                        // This fall-through is a deviation from the EcmaScript grammar. The grammar says that a leading zero
                        // can only be followed by an octal digit, a dot, or the end of the number literal. However, we are being
                        // permissive and allowing decimal digits of the form 08* and 09* (which many browsers also do).
                        // falls through
                        case 49 /* _1 */:
                        case 50 /* _2 */:
                        case 51 /* _3 */:
                        case 52 /* _4 */:
                        case 53 /* _5 */:
                        case 54 /* _6 */:
                        case 55 /* _7 */:
                        case 56 /* _8 */:
                        case 57 /* _9 */:
                            (_a = scanNumber(), token = _a.type, tokenValue = _a.value);
                            return token;
                        case 58 /* colon */:
                            pos++;
                            return token = 58 /* ColonToken */;
                        case 59 /* semicolon */:
                            pos++;
                            return token = 26 /* SemicolonToken */;
                        case 60 /* lessThan */:
                            if (isConflictMarkerTrivia(text, pos)) {
                                pos = scanConflictMarkerTrivia(text, pos, error);
                                if (skipTrivia) {
                                    continue;
                                }
                                else {
                                    return token = 7 /* ConflictMarkerTrivia */;
                                }
                            }
                            if (text.charCodeAt(pos + 1) === 60 /* lessThan */) {
                                if (text.charCodeAt(pos + 2) === 61 /* equals */) {
                                    return pos += 3, token = 70 /* LessThanLessThanEqualsToken */;
                                }
                                return pos += 2, token = 47 /* LessThanLessThanToken */;
                            }
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                return pos += 2, token = 32 /* LessThanEqualsToken */;
                            }
                            if (languageVariant === 1 /* JSX */ &&
                                text.charCodeAt(pos + 1) === 47 /* slash */ &&
                                text.charCodeAt(pos + 2) !== 42 /* asterisk */) {
                                return pos += 2, token = 30 /* LessThanSlashToken */;
                            }
                            pos++;
                            return token = 29 /* LessThanToken */;
                        case 61 /* equals */:
                            if (isConflictMarkerTrivia(text, pos)) {
                                pos = scanConflictMarkerTrivia(text, pos, error);
                                if (skipTrivia) {
                                    continue;
                                }
                                else {
                                    return token = 7 /* ConflictMarkerTrivia */;
                                }
                            }
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                if (text.charCodeAt(pos + 2) === 61 /* equals */) {
                                    return pos += 3, token = 36 /* EqualsEqualsEqualsToken */;
                                }
                                return pos += 2, token = 34 /* EqualsEqualsToken */;
                            }
                            if (text.charCodeAt(pos + 1) === 62 /* greaterThan */) {
                                return pos += 2, token = 38 /* EqualsGreaterThanToken */;
                            }
                            pos++;
                            return token = 63 /* EqualsToken */;
                        case 62 /* greaterThan */:
                            if (isConflictMarkerTrivia(text, pos)) {
                                pos = scanConflictMarkerTrivia(text, pos, error);
                                if (skipTrivia) {
                                    continue;
                                }
                                else {
                                    return token = 7 /* ConflictMarkerTrivia */;
                                }
                            }
                            pos++;
                            return token = 31 /* GreaterThanToken */;
                        case 63 /* question */:
                            if (text.charCodeAt(pos + 1) === 46 /* dot */ && !isDigit(text.charCodeAt(pos + 2))) {
                                return pos += 2, token = 28 /* QuestionDotToken */;
                            }
                            if (text.charCodeAt(pos + 1) === 63 /* question */) {
                                if (text.charCodeAt(pos + 2) === 61 /* equals */) {
                                    return pos += 3, token = 77 /* QuestionQuestionEqualsToken */;
                                }
                                return pos += 2, token = 60 /* QuestionQuestionToken */;
                            }
                            pos++;
                            return token = 57 /* QuestionToken */;
                        case 91 /* openBracket */:
                            pos++;
                            return token = 22 /* OpenBracketToken */;
                        case 93 /* closeBracket */:
                            pos++;
                            return token = 23 /* CloseBracketToken */;
                        case 94 /* caret */:
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                return pos += 2, token = 78 /* CaretEqualsToken */;
                            }
                            pos++;
                            return token = 52 /* CaretToken */;
                        case 123 /* openBrace */:
                            pos++;
                            return token = 18 /* OpenBraceToken */;
                        case 124 /* bar */:
                            if (isConflictMarkerTrivia(text, pos)) {
                                pos = scanConflictMarkerTrivia(text, pos, error);
                                if (skipTrivia) {
                                    continue;
                                }
                                else {
                                    return token = 7 /* ConflictMarkerTrivia */;
                                }
                            }
                            if (text.charCodeAt(pos + 1) === 124 /* bar */) {
                                if (text.charCodeAt(pos + 2) === 61 /* equals */) {
                                    return pos += 3, token = 75 /* BarBarEqualsToken */;
                                }
                                return pos += 2, token = 56 /* BarBarToken */;
                            }
                            if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                                return pos += 2, token = 74 /* BarEqualsToken */;
                            }
                            pos++;
                            return token = 51 /* BarToken */;
                        case 125 /* closeBrace */:
                            pos++;
                            return token = 19 /* CloseBraceToken */;
                        case 126 /* tilde */:
                            pos++;
                            return token = 54 /* TildeToken */;
                        case 64 /* at */:
                            pos++;
                            return token = 59 /* AtToken */;
                        case 92 /* backslash */:
                            var extendedCookedChar = peekExtendedUnicodeEscape();
                            if (extendedCookedChar >= 0 && isIdentifierStart(extendedCookedChar, languageVersion)) {
                                pos += 3;
                                tokenFlags |= 8 /* ExtendedUnicodeEscape */;
                                tokenValue = scanExtendedUnicodeEscape() + scanIdentifierParts();
                                return token = getIdentifierToken();
                            }
                            var cookedChar = peekUnicodeEscape();
                            if (cookedChar >= 0 && isIdentifierStart(cookedChar, languageVersion)) {
                                pos += 6;
                                tokenFlags |= 1024 /* UnicodeEscape */;
                                tokenValue = String.fromCharCode(cookedChar) + scanIdentifierParts();
                                return token = getIdentifierToken();
                            }
                            error(ts.Diagnostics.Invalid_character);
                            pos++;
                            return token = 0 /* Unknown */;
                        case 35 /* hash */:
                            if (pos !== 0 && text[pos + 1] === "!") {
                                error(ts.Diagnostics.can_only_be_used_at_the_start_of_a_file);
                                pos++;
                                return token = 0 /* Unknown */;
                            }
                            if (isIdentifierStart(codePointAt(text, pos + 1), languageVersion)) {
                                pos++;
                                scanIdentifier(codePointAt(text, pos), languageVersion);
                            }
                            else {
                                tokenValue = String.fromCharCode(codePointAt(text, pos));
                                error(ts.Diagnostics.Invalid_character, pos++, charSize(ch));
                            }
                            return token = 80 /* PrivateIdentifier */;
                        default:
                            var identifierKind = scanIdentifier(ch, languageVersion);
                            if (identifierKind) {
                                return token = identifierKind;
                            }
                            else if (isWhiteSpaceSingleLine(ch)) {
                                pos += charSize(ch);
                                continue;
                            }
                            else if (isLineBreak(ch)) {
                                tokenFlags |= 1 /* PrecedingLineBreak */;
                                pos += charSize(ch);
                                continue;
                            }
                            var size = charSize(ch);
                            error(ts.Diagnostics.Invalid_character, pos, size);
                            pos += size;
                            return token = 0 /* Unknown */;
                    }
                }
            }
            function reScanInvalidIdentifier() {
                ts.Debug.assert(token === 0 /* Unknown */, "'reScanInvalidIdentifier' should only be called when the current token is 'SyntaxKind.Unknown'.");
                pos = tokenPos = startPos;
                tokenFlags = 0;
                var ch = codePointAt(text, pos);
                var identifierKind = scanIdentifier(ch, 99 /* ESNext */);
                if (identifierKind) {
                    return token = identifierKind;
                }
                pos += charSize(ch);
                return token; // Still `SyntaKind.Unknown`
            }
            function scanIdentifier(startCharacter, languageVersion) {
                var ch = startCharacter;
                if (isIdentifierStart(ch, languageVersion)) {
                    pos += charSize(ch);
                    while (pos < end && isIdentifierPart(ch = codePointAt(text, pos), languageVersion))
                        pos += charSize(ch);
                    tokenValue = text.substring(tokenPos, pos);
                    if (ch === 92 /* backslash */) {
                        tokenValue += scanIdentifierParts();
                    }
                    return getIdentifierToken();
                }
            }
            function reScanGreaterToken() {
                if (token === 31 /* GreaterThanToken */) {
                    if (text.charCodeAt(pos) === 62 /* greaterThan */) {
                        if (text.charCodeAt(pos + 1) === 62 /* greaterThan */) {
                            if (text.charCodeAt(pos + 2) === 61 /* equals */) {
                                return pos += 3, token = 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */;
                            }
                            return pos += 2, token = 49 /* GreaterThanGreaterThanGreaterThanToken */;
                        }
                        if (text.charCodeAt(pos + 1) === 61 /* equals */) {
                            return pos += 2, token = 71 /* GreaterThanGreaterThanEqualsToken */;
                        }
                        pos++;
                        return token = 48 /* GreaterThanGreaterThanToken */;
                    }
                    if (text.charCodeAt(pos) === 61 /* equals */) {
                        pos++;
                        return token = 33 /* GreaterThanEqualsToken */;
                    }
                }
                return token;
            }
            function reScanAsteriskEqualsToken() {
                ts.Debug.assert(token === 66 /* AsteriskEqualsToken */, "'reScanAsteriskEqualsToken' should only be called on a '*='");
                pos = tokenPos + 1;
                return token = 63 /* EqualsToken */;
            }
            function reScanSlashToken() {
                if (token === 43 /* SlashToken */ || token === 68 /* SlashEqualsToken */) {
                    var p = tokenPos + 1;
                    var inEscape = false;
                    var inCharacterClass = false;
                    while (true) {
                        // If we reach the end of a file, or hit a newline, then this is an unterminated
                        // regex.  Report error and return what we have so far.
                        if (p >= end) {
                            tokenFlags |= 4 /* Unterminated */;
                            error(ts.Diagnostics.Unterminated_regular_expression_literal);
                            break;
                        }
                        var ch = text.charCodeAt(p);
                        if (isLineBreak(ch)) {
                            tokenFlags |= 4 /* Unterminated */;
                            error(ts.Diagnostics.Unterminated_regular_expression_literal);
                            break;
                        }
                        if (inEscape) {
                            // Parsing an escape character;
                            // reset the flag and just advance to the next char.
                            inEscape = false;
                        }
                        else if (ch === 47 /* slash */ && !inCharacterClass) {
                            // A slash within a character class is permissible,
                            // but in general it signals the end of the regexp literal.
                            p++;
                            break;
                        }
                        else if (ch === 91 /* openBracket */) {
                            inCharacterClass = true;
                        }
                        else if (ch === 92 /* backslash */) {
                            inEscape = true;
                        }
                        else if (ch === 93 /* closeBracket */) {
                            inCharacterClass = false;
                        }
                        p++;
                    }
                    while (p < end && isIdentifierPart(text.charCodeAt(p), languageVersion)) {
                        p++;
                    }
                    pos = p;
                    tokenValue = text.substring(tokenPos, pos);
                    token = 13 /* RegularExpressionLiteral */;
                }
                return token;
            }
            function appendIfCommentDirective(commentDirectives, text, commentDirectiveRegEx, lineStart) {
                var type = getDirectiveFromComment(ts.trimStringStart(text), commentDirectiveRegEx);
                if (type === undefined) {
                    return commentDirectives;
                }
                return ts.append(commentDirectives, {
                    range: { pos: lineStart, end: pos },
                    type: type,
                });
            }
            function getDirectiveFromComment(text, commentDirectiveRegEx) {
                var match = commentDirectiveRegEx.exec(text);
                if (!match) {
                    return undefined;
                }
                switch (match[1]) {
                    case "ts-expect-error":
                        return 0 /* ExpectError */;
                    case "ts-ignore":
                        return 1 /* Ignore */;
                }
                return undefined;
            }
            /**
             * Unconditionally back up and scan a template expression portion.
             */
            function reScanTemplateToken(isTaggedTemplate) {
                ts.Debug.assert(token === 19 /* CloseBraceToken */, "'reScanTemplateToken' should only be called on a '}'");
                pos = tokenPos;
                return token = scanTemplateAndSetTokenValue(isTaggedTemplate);
            }
            function reScanTemplateHeadOrNoSubstitutionTemplate() {
                pos = tokenPos;
                return token = scanTemplateAndSetTokenValue(/* isTaggedTemplate */ true);
            }
            function reScanJsxToken(allowMultilineJsxText) {
                if (allowMultilineJsxText === void 0) {
                    allowMultilineJsxText = true;
                }
                pos = tokenPos = startPos;
                return token = scanJsxToken(allowMultilineJsxText);
            }
            function reScanLessThanToken() {
                if (token === 47 /* LessThanLessThanToken */) {
                    pos = tokenPos + 1;
                    return token = 29 /* LessThanToken */;
                }
                return token;
            }
            function reScanHashToken() {
                if (token === 80 /* PrivateIdentifier */) {
                    pos = tokenPos + 1;
                    return token = 62 /* HashToken */;
                }
                return token;
            }
            function reScanQuestionToken() {
                ts.Debug.assert(token === 60 /* QuestionQuestionToken */, "'reScanQuestionToken' should only be called on a '??'");
                pos = tokenPos + 1;
                return token = 57 /* QuestionToken */;
            }
            function scanJsxToken(allowMultilineJsxText) {
                if (allowMultilineJsxText === void 0) {
                    allowMultilineJsxText = true;
                }
                startPos = tokenPos = pos;
                if (pos >= end) {
                    return token = 1 /* EndOfFileToken */;
                }
                var char = text.charCodeAt(pos);
                if (char === 60 /* lessThan */) {
                    if (text.charCodeAt(pos + 1) === 47 /* slash */) {
                        pos += 2;
                        return token = 30 /* LessThanSlashToken */;
                    }
                    pos++;
                    return token = 29 /* LessThanToken */;
                }
                if (char === 123 /* openBrace */) {
                    pos++;
                    return token = 18 /* OpenBraceToken */;
                }
                // First non-whitespace character on this line.
                var firstNonWhitespace = 0;
                // These initial values are special because the first line is:
                // firstNonWhitespace = 0 to indicate that we want leading whitespace,
                while (pos < end) {
                    char = text.charCodeAt(pos);
                    if (char === 123 /* openBrace */) {
                        break;
                    }
                    if (char === 60 /* lessThan */) {
                        if (isConflictMarkerTrivia(text, pos)) {
                            pos = scanConflictMarkerTrivia(text, pos, error);
                            return token = 7 /* ConflictMarkerTrivia */;
                        }
                        break;
                    }
                    if (char === 62 /* greaterThan */) {
                        error(ts.Diagnostics.Unexpected_token_Did_you_mean_or_gt, pos, 1);
                    }
                    if (char === 125 /* closeBrace */) {
                        error(ts.Diagnostics.Unexpected_token_Did_you_mean_or_rbrace, pos, 1);
                    }
                    // FirstNonWhitespace is 0, then we only see whitespaces so far. If we see a linebreak, we want to ignore that whitespaces.
                    // i.e (- : whitespace)
                    //      <div>----
                    //      </div> becomes <div></div>
                    //
                    //      <div>----</div> becomes <div>----</div>
                    if (isLineBreak(char) && firstNonWhitespace === 0) {
                        firstNonWhitespace = -1;
                    }
                    else if (!allowMultilineJsxText && isLineBreak(char) && firstNonWhitespace > 0) {
                        // Stop JsxText on each line during formatting. This allows the formatter to
                        // indent each line correctly.
                        break;
                    }
                    else if (!isWhiteSpaceLike(char)) {
                        firstNonWhitespace = pos;
                    }
                    pos++;
                }
                tokenValue = text.substring(startPos, pos);
                return firstNonWhitespace === -1 ? 12 /* JsxTextAllWhiteSpaces */ : 11 /* JsxText */;
            }
            // Scans a JSX identifier; these differ from normal identifiers in that
            // they allow dashes
            function scanJsxIdentifier() {
                if (tokenIsIdentifierOrKeyword(token)) {
                    // An identifier or keyword has already been parsed - check for a `-` or a single instance of `:` and then append it and
                    // everything after it to the token
                    // Do note that this means that `scanJsxIdentifier` effectively _mutates_ the visible token without advancing to a new token
                    // Any caller should be expecting this behavior and should only read the pos or token value after calling it.
                    var namespaceSeparator = false;
                    while (pos < end) {
                        var ch = text.charCodeAt(pos);
                        if (ch === 45 /* minus */) {
                            tokenValue += "-";
                            pos++;
                            continue;
                        }
                        else if (ch === 58 /* colon */ && !namespaceSeparator) {
                            tokenValue += ":";
                            pos++;
                            namespaceSeparator = true;
                            token = 79 /* Identifier */; // swap from keyword kind to identifier kind
                            continue;
                        }
                        var oldPos = pos;
                        tokenValue += scanIdentifierParts(); // reuse `scanIdentifierParts` so unicode escapes are handled
                        if (pos === oldPos) {
                            break;
                        }
                    }
                    // Do not include a trailing namespace separator in the token, since this is against the spec.
                    if (tokenValue.slice(-1) === ":") {
                        tokenValue = tokenValue.slice(0, -1);
                        pos--;
                    }
                }
                return token;
            }
            function scanJsxAttributeValue() {
                startPos = pos;
                switch (text.charCodeAt(pos)) {
                    case 34 /* doubleQuote */:
                    case 39 /* singleQuote */:
                        tokenValue = scanString(/*jsxAttributeString*/ true);
                        return token = 10 /* StringLiteral */;
                    default:
                        // If this scans anything other than `{`, it's a parse error.
                        return scan();
                }
            }
            function reScanJsxAttributeValue() {
                pos = tokenPos = startPos;
                return scanJsxAttributeValue();
            }
            function scanJsDocToken() {
                startPos = tokenPos = pos;
                tokenFlags = 0 /* None */;
                if (pos >= end) {
                    return token = 1 /* EndOfFileToken */;
                }
                var ch = codePointAt(text, pos);
                pos += charSize(ch);
                switch (ch) {
                    case 9 /* tab */:
                    case 11 /* verticalTab */:
                    case 12 /* formFeed */:
                    case 32 /* space */:
                        while (pos < end && isWhiteSpaceSingleLine(text.charCodeAt(pos))) {
                            pos++;
                        }
                        return token = 5 /* WhitespaceTrivia */;
                    case 64 /* at */:
                        return token = 59 /* AtToken */;
                    case 13 /* carriageReturn */:
                        if (text.charCodeAt(pos) === 10 /* lineFeed */) {
                            pos++;
                        }
                    // falls through
                    case 10 /* lineFeed */:
                        tokenFlags |= 1 /* PrecedingLineBreak */;
                        return token = 4 /* NewLineTrivia */;
                    case 42 /* asterisk */:
                        return token = 41 /* AsteriskToken */;
                    case 123 /* openBrace */:
                        return token = 18 /* OpenBraceToken */;
                    case 125 /* closeBrace */:
                        return token = 19 /* CloseBraceToken */;
                    case 91 /* openBracket */:
                        return token = 22 /* OpenBracketToken */;
                    case 93 /* closeBracket */:
                        return token = 23 /* CloseBracketToken */;
                    case 60 /* lessThan */:
                        return token = 29 /* LessThanToken */;
                    case 62 /* greaterThan */:
                        return token = 31 /* GreaterThanToken */;
                    case 61 /* equals */:
                        return token = 63 /* EqualsToken */;
                    case 44 /* comma */:
                        return token = 27 /* CommaToken */;
                    case 46 /* dot */:
                        return token = 24 /* DotToken */;
                    case 96 /* backtick */:
                        return token = 61 /* BacktickToken */;
                    case 35 /* hash */:
                        return token = 62 /* HashToken */;
                    case 92 /* backslash */:
                        pos--;
                        var extendedCookedChar = peekExtendedUnicodeEscape();
                        if (extendedCookedChar >= 0 && isIdentifierStart(extendedCookedChar, languageVersion)) {
                            pos += 3;
                            tokenFlags |= 8 /* ExtendedUnicodeEscape */;
                            tokenValue = scanExtendedUnicodeEscape() + scanIdentifierParts();
                            return token = getIdentifierToken();
                        }
                        var cookedChar = peekUnicodeEscape();
                        if (cookedChar >= 0 && isIdentifierStart(cookedChar, languageVersion)) {
                            pos += 6;
                            tokenFlags |= 1024 /* UnicodeEscape */;
                            tokenValue = String.fromCharCode(cookedChar) + scanIdentifierParts();
                            return token = getIdentifierToken();
                        }
                        pos++;
                        return token = 0 /* Unknown */;
                }
                if (isIdentifierStart(ch, languageVersion)) {
                    var char = ch;
                    while (pos < end && isIdentifierPart(char = codePointAt(text, pos), languageVersion) || text.charCodeAt(pos) === 45 /* minus */)
                        pos += charSize(char);
                    tokenValue = text.substring(tokenPos, pos);
                    if (char === 92 /* backslash */) {
                        tokenValue += scanIdentifierParts();
                    }
                    return token = getIdentifierToken();
                }
                else {
                    return token = 0 /* Unknown */;
                }
            }
            function speculationHelper(callback, isLookahead) {
                var savePos = pos;
                var saveStartPos = startPos;
                var saveTokenPos = tokenPos;
                var saveToken = token;
                var saveTokenValue = tokenValue;
                var saveTokenFlags = tokenFlags;
                var result = callback();
                // If our callback returned something 'falsy' or we're just looking ahead,
                // then unconditionally restore us to where we were.
                if (!result || isLookahead) {
                    pos = savePos;
                    startPos = saveStartPos;
                    tokenPos = saveTokenPos;
                    token = saveToken;
                    tokenValue = saveTokenValue;
                    tokenFlags = saveTokenFlags;
                }
                return result;
            }
            function scanRange(start, length, callback) {
                var saveEnd = end;
                var savePos = pos;
                var saveStartPos = startPos;
                var saveTokenPos = tokenPos;
                var saveToken = token;
                var saveTokenValue = tokenValue;
                var saveTokenFlags = tokenFlags;
                var saveErrorExpectations = commentDirectives;
                setText(text, start, length);
                var result = callback();
                end = saveEnd;
                pos = savePos;
                startPos = saveStartPos;
                tokenPos = saveTokenPos;
                token = saveToken;
                tokenValue = saveTokenValue;
                tokenFlags = saveTokenFlags;
                commentDirectives = saveErrorExpectations;
                return result;
            }
            function lookAhead(callback) {
                return speculationHelper(callback, /*isLookahead*/ true);
            }
            function tryScan(callback) {
                return speculationHelper(callback, /*isLookahead*/ false);
            }
            function getText() {
                return text;
            }
            function clearCommentDirectives() {
                commentDirectives = undefined;
            }
            function setText(newText, start, length) {
                text = newText || "";
                end = length === undefined ? text.length : start + length;
                setTextPos(start || 0);
            }
            function setOnError(errorCallback) {
                onError = errorCallback;
            }
            function setScriptTarget(scriptTarget) {
                languageVersion = scriptTarget;
            }
            function setLanguageVariant(variant) {
                languageVariant = variant;
            }
            function setTextPos(textPos) {
                ts.Debug.assert(textPos >= 0);
                pos = textPos;
                startPos = textPos;
                tokenPos = textPos;
                token = 0 /* Unknown */;
                tokenValue = undefined;
                tokenFlags = 0 /* None */;
            }
            function setInJSDocType(inType) {
                inJSDocType += inType ? 1 : -1;
            }
        }
        ts.createScanner = createScanner;
        /* @internal */
        var codePointAt = String.prototype.codePointAt ? function (s, i) { return s.codePointAt(i); } : function codePointAt(str, i) {
            // from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/codePointAt
            var size = str.length;
            // Account for out-of-bounds indices:
            if (i < 0 || i >= size) {
                return undefined; // String.codePointAt returns `undefined` for OOB indexes
            }
            // Get the first code unit
            var first = str.charCodeAt(i);
            // check if it’s the start of a surrogate pair
            if (first >= 0xD800 && first <= 0xDBFF && size > i + 1) { // high surrogate and there is a next code unit
                var second = str.charCodeAt(i + 1);
                if (second >= 0xDC00 && second <= 0xDFFF) { // low surrogate
                    // https://mathiasbynens.be/notes/javascript-encoding#surrogate-formulae
                    return (first - 0xD800) * 0x400 + second - 0xDC00 + 0x10000;
                }
            }
            return first;
        };
        /* @internal */
        function charSize(ch) {
            if (ch >= 0x10000) {
                return 2;
            }
            return 1;
        }
        // Derived from the 10.1.1 UTF16Encoding of the ES6 Spec.
        function utf16EncodeAsStringFallback(codePoint) {
            ts.Debug.assert(0x0 <= codePoint && codePoint <= 0x10FFFF);
            if (codePoint <= 65535) {
                return String.fromCharCode(codePoint);
            }
            var codeUnit1 = Math.floor((codePoint - 65536) / 1024) + 0xD800;
            var codeUnit2 = ((codePoint - 65536) % 1024) + 0xDC00;
            return String.fromCharCode(codeUnit1, codeUnit2);
        }
        var utf16EncodeAsStringWorker = String.fromCodePoint ? function (codePoint) { return String.fromCodePoint(codePoint); } : utf16EncodeAsStringFallback;
        /* @internal */
        function utf16EncodeAsString(codePoint) {
            return utf16EncodeAsStringWorker(codePoint);
        }
        ts.utf16EncodeAsString = utf16EncodeAsString;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        function isExternalModuleNameRelative(moduleName) {
            // TypeScript 1.0 spec (April 2014): 11.2.1
            // An external module name is "relative" if the first term is "." or "..".
            // Update: We also consider a path like `C:\foo.ts` "relative" because we do not search for it in `node_modules` or treat it as an ambient module.
            return ts.pathIsRelative(moduleName) || ts.isRootedDiskPath(moduleName);
        }
        ts.isExternalModuleNameRelative = isExternalModuleNameRelative;
        function sortAndDeduplicateDiagnostics(diagnostics) {
            return ts.sortAndDeduplicate(diagnostics, ts.compareDiagnostics);
        }
        ts.sortAndDeduplicateDiagnostics = sortAndDeduplicateDiagnostics;
        function getDefaultLibFileName(options) {
            switch (options.target) {
                case 99 /* ESNext */:
                    return "lib.esnext.full.d.ts";
                case 8 /* ES2021 */:
                    return "lib.es2021.full.d.ts";
                case 7 /* ES2020 */:
                    return "lib.es2020.full.d.ts";
                case 6 /* ES2019 */:
                    return "lib.es2019.full.d.ts";
                case 5 /* ES2018 */:
                    return "lib.es2018.full.d.ts";
                case 4 /* ES2017 */:
                    return "lib.es2017.full.d.ts";
                case 3 /* ES2016 */:
                    return "lib.es2016.full.d.ts";
                case 2 /* ES2015 */:
                    return "lib.es6.d.ts"; // We don't use lib.es2015.full.d.ts due to breaking change.
                default:
                    return "lib.d.ts";
            }
        }
        ts.getDefaultLibFileName = getDefaultLibFileName;
        function textSpanEnd(span) {
            return span.start + span.length;
        }
        ts.textSpanEnd = textSpanEnd;
        function textSpanIsEmpty(span) {
            return span.length === 0;
        }
        ts.textSpanIsEmpty = textSpanIsEmpty;
        function textSpanContainsPosition(span, position) {
            return position >= span.start && position < textSpanEnd(span);
        }
        ts.textSpanContainsPosition = textSpanContainsPosition;
        /* @internal */
        function textRangeContainsPositionInclusive(span, position) {
            return position >= span.pos && position <= span.end;
        }
        ts.textRangeContainsPositionInclusive = textRangeContainsPositionInclusive;
        // Returns true if 'span' contains 'other'.
        function textSpanContainsTextSpan(span, other) {
            return other.start >= span.start && textSpanEnd(other) <= textSpanEnd(span);
        }
        ts.textSpanContainsTextSpan = textSpanContainsTextSpan;
        function textSpanOverlapsWith(span, other) {
            return textSpanOverlap(span, other) !== undefined;
        }
        ts.textSpanOverlapsWith = textSpanOverlapsWith;
        function textSpanOverlap(span1, span2) {
            var overlap = textSpanIntersection(span1, span2);
            return overlap && overlap.length === 0 ? undefined : overlap;
        }
        ts.textSpanOverlap = textSpanOverlap;
        function textSpanIntersectsWithTextSpan(span, other) {
            return decodedTextSpanIntersectsWith(span.start, span.length, other.start, other.length);
        }
        ts.textSpanIntersectsWithTextSpan = textSpanIntersectsWithTextSpan;
        function textSpanIntersectsWith(span, start, length) {
            return decodedTextSpanIntersectsWith(span.start, span.length, start, length);
        }
        ts.textSpanIntersectsWith = textSpanIntersectsWith;
        function decodedTextSpanIntersectsWith(start1, length1, start2, length2) {
            var end1 = start1 + length1;
            var end2 = start2 + length2;
            return start2 <= end1 && end2 >= start1;
        }
        ts.decodedTextSpanIntersectsWith = decodedTextSpanIntersectsWith;
        function textSpanIntersectsWithPosition(span, position) {
            return position <= textSpanEnd(span) && position >= span.start;
        }
        ts.textSpanIntersectsWithPosition = textSpanIntersectsWithPosition;
        function textSpanIntersection(span1, span2) {
            var start = Math.max(span1.start, span2.start);
            var end = Math.min(textSpanEnd(span1), textSpanEnd(span2));
            return start <= end ? createTextSpanFromBounds(start, end) : undefined;
        }
        ts.textSpanIntersection = textSpanIntersection;
        function createTextSpan(start, length) {
            if (start < 0) {
                throw new Error("start < 0");
            }
            if (length < 0) {
                throw new Error("length < 0");
            }
            return { start: start, length: length };
        }
        ts.createTextSpan = createTextSpan;
        function createTextSpanFromBounds(start, end) {
            return createTextSpan(start, end - start);
        }
        ts.createTextSpanFromBounds = createTextSpanFromBounds;
        function textChangeRangeNewSpan(range) {
            return createTextSpan(range.span.start, range.newLength);
        }
        ts.textChangeRangeNewSpan = textChangeRangeNewSpan;
        function textChangeRangeIsUnchanged(range) {
            return textSpanIsEmpty(range.span) && range.newLength === 0;
        }
        ts.textChangeRangeIsUnchanged = textChangeRangeIsUnchanged;
        function createTextChangeRange(span, newLength) {
            if (newLength < 0) {
                throw new Error("newLength < 0");
            }
            return { span: span, newLength: newLength };
        }
        ts.createTextChangeRange = createTextChangeRange;
        ts.unchangedTextChangeRange = createTextChangeRange(createTextSpan(0, 0), 0); // eslint-disable-line prefer-const
        /**
         * Called to merge all the changes that occurred across several versions of a script snapshot
         * into a single change.  i.e. if a user keeps making successive edits to a script we will
         * have a text change from V1 to V2, V2 to V3, ..., Vn.
         *
         * This function will then merge those changes into a single change range valid between V1 and
         * Vn.
         */
        function collapseTextChangeRangesAcrossMultipleVersions(changes) {
            if (changes.length === 0) {
                return ts.unchangedTextChangeRange;
            }
            if (changes.length === 1) {
                return changes[0];
            }
            // We change from talking about { { oldStart, oldLength }, newLength } to { oldStart, oldEnd, newEnd }
            // as it makes things much easier to reason about.
            var change0 = changes[0];
            var oldStartN = change0.span.start;
            var oldEndN = textSpanEnd(change0.span);
            var newEndN = oldStartN + change0.newLength;
            for (var i = 1; i < changes.length; i++) {
                var nextChange = changes[i];
                // Consider the following case:
                // i.e. two edits.  The first represents the text change range { { 10, 50 }, 30 }.  i.e. The span starting
                // at 10, with length 50 is reduced to length 30.  The second represents the text change range { { 30, 30 }, 40 }.
                // i.e. the span starting at 30 with length 30 is increased to length 40.
                //
                //      0         10        20        30        40        50        60        70        80        90        100
                //      -------------------------------------------------------------------------------------------------------
                //                |                                                 /
                //                |                                            /----
                //  T1            |                                       /----
                //                |                                  /----
                //                |                             /----
                //      -------------------------------------------------------------------------------------------------------
                //                                     |                            \
                //                                     |                               \
                //   T2                                |                                 \
                //                                     |                                   \
                //                                     |                                      \
                //      -------------------------------------------------------------------------------------------------------
                //
                // Merging these turns out to not be too difficult.  First, determining the new start of the change is trivial
                // it's just the min of the old and new starts.  i.e.:
                //
                //      0         10        20        30        40        50        60        70        80        90        100
                //      ------------------------------------------------------------*------------------------------------------
                //                |                                                 /
                //                |                                            /----
                //  T1            |                                       /----
                //                |                                  /----
                //                |                             /----
                //      ----------------------------------------$-------------------$------------------------------------------
                //                .                    |                            \
                //                .                    |                               \
                //   T2           .                    |                                 \
                //                .                    |                                   \
                //                .                    |                                      \
                //      ----------------------------------------------------------------------*--------------------------------
                //
                // (Note the dots represent the newly inferred start.
                // Determining the new and old end is also pretty simple.  Basically it boils down to paying attention to the
                // absolute positions at the asterisks, and the relative change between the dollar signs. Basically, we see
                // which if the two $'s precedes the other, and we move that one forward until they line up.  in this case that
                // means:
                //
                //      0         10        20        30        40        50        60        70        80        90        100
                //      --------------------------------------------------------------------------------*----------------------
                //                |                                                                     /
                //                |                                                                /----
                //  T1            |                                                           /----
                //                |                                                      /----
                //                |                                                 /----
                //      ------------------------------------------------------------$------------------------------------------
                //                .                    |                            \
                //                .                    |                               \
                //   T2           .                    |                                 \
                //                .                    |                                   \
                //                .                    |                                      \
                //      ----------------------------------------------------------------------*--------------------------------
                //
                // In other words (in this case), we're recognizing that the second edit happened after where the first edit
                // ended with a delta of 20 characters (60 - 40).  Thus, if we go back in time to where the first edit started
                // that's the same as if we started at char 80 instead of 60.
                //
                // As it so happens, the same logic applies if the second edit precedes the first edit.  In that case rather
                // than pushing the first edit forward to match the second, we'll push the second edit forward to match the
                // first.
                //
                // In this case that means we have { oldStart: 10, oldEnd: 80, newEnd: 70 } or, in TextChangeRange
                // semantics: { { start: 10, length: 70 }, newLength: 60 }
                //
                // The math then works out as follows.
                // If we have { oldStart1, oldEnd1, newEnd1 } and { oldStart2, oldEnd2, newEnd2 } then we can compute the
                // final result like so:
                //
                // {
                //      oldStart3: Min(oldStart1, oldStart2),
                //      oldEnd3: Max(oldEnd1, oldEnd1 + (oldEnd2 - newEnd1)),
                //      newEnd3: Max(newEnd2, newEnd2 + (newEnd1 - oldEnd2))
                // }
                var oldStart1 = oldStartN;
                var oldEnd1 = oldEndN;
                var newEnd1 = newEndN;
                var oldStart2 = nextChange.span.start;
                var oldEnd2 = textSpanEnd(nextChange.span);
                var newEnd2 = oldStart2 + nextChange.newLength;
                oldStartN = Math.min(oldStart1, oldStart2);
                oldEndN = Math.max(oldEnd1, oldEnd1 + (oldEnd2 - newEnd1));
                newEndN = Math.max(newEnd2, newEnd2 + (newEnd1 - oldEnd2));
            }
            return createTextChangeRange(createTextSpanFromBounds(oldStartN, oldEndN), /*newLength*/ newEndN - oldStartN);
        }
        ts.collapseTextChangeRangesAcrossMultipleVersions = collapseTextChangeRangesAcrossMultipleVersions;
        function getTypeParameterOwner(d) {
            if (d && d.kind === 161 /* TypeParameter */) {
                for (var current = d; current; current = current.parent) {
                    if (isFunctionLike(current) || isClassLike(current) || current.kind === 256 /* InterfaceDeclaration */) {
                        return current;
                    }
                }
            }
        }
        ts.getTypeParameterOwner = getTypeParameterOwner;
        function isParameterPropertyDeclaration(node, parent) {
            return ts.hasSyntacticModifier(node, 16476 /* ParameterPropertyModifier */) && parent.kind === 169 /* Constructor */;
        }
        ts.isParameterPropertyDeclaration = isParameterPropertyDeclaration;
        function isEmptyBindingPattern(node) {
            if (isBindingPattern(node)) {
                return ts.every(node.elements, isEmptyBindingElement);
            }
            return false;
        }
        ts.isEmptyBindingPattern = isEmptyBindingPattern;
        function isEmptyBindingElement(node) {
            if (ts.isOmittedExpression(node)) {
                return true;
            }
            return isEmptyBindingPattern(node.name);
        }
        ts.isEmptyBindingElement = isEmptyBindingElement;
        function walkUpBindingElementsAndPatterns(binding) {
            var node = binding.parent;
            while (ts.isBindingElement(node.parent)) {
                node = node.parent.parent;
            }
            return node.parent;
        }
        ts.walkUpBindingElementsAndPatterns = walkUpBindingElementsAndPatterns;
        function getCombinedFlags(node, getFlags) {
            if (ts.isBindingElement(node)) {
                node = walkUpBindingElementsAndPatterns(node);
            }
            var flags = getFlags(node);
            if (node.kind === 252 /* VariableDeclaration */) {
                node = node.parent;
            }
            if (node && node.kind === 253 /* VariableDeclarationList */) {
                flags |= getFlags(node);
                node = node.parent;
            }
            if (node && node.kind === 235 /* VariableStatement */) {
                flags |= getFlags(node);
            }
            return flags;
        }
        function getCombinedModifierFlags(node) {
            return getCombinedFlags(node, ts.getEffectiveModifierFlags);
        }
        ts.getCombinedModifierFlags = getCombinedModifierFlags;
        /* @internal */
        function getCombinedNodeFlagsAlwaysIncludeJSDoc(node) {
            return getCombinedFlags(node, ts.getEffectiveModifierFlagsAlwaysIncludeJSDoc);
        }
        ts.getCombinedNodeFlagsAlwaysIncludeJSDoc = getCombinedNodeFlagsAlwaysIncludeJSDoc;
        // Returns the node flags for this node and all relevant parent nodes.  This is done so that
        // nodes like variable declarations and binding elements can returned a view of their flags
        // that includes the modifiers from their container.  i.e. flags like export/declare aren't
        // stored on the variable declaration directly, but on the containing variable statement
        // (if it has one).  Similarly, flags for let/const are stored on the variable declaration
        // list.  By calling this function, all those flags are combined so that the client can treat
        // the node as if it actually had those flags.
        function getCombinedNodeFlags(node) {
            return getCombinedFlags(node, function (n) { return n.flags; });
        }
        ts.getCombinedNodeFlags = getCombinedNodeFlags;
        /* @internal */
        ts.supportedLocaleDirectories = ["cs", "de", "es", "fr", "it", "ja", "ko", "pl", "pt-br", "ru", "tr", "zh-cn", "zh-tw"];
        /**
         * Checks to see if the locale is in the appropriate format,
         * and if it is, attempts to set the appropriate language.
         */
        function validateLocaleAndSetLanguage(locale, sys, errors) {
            var lowerCaseLocale = locale.toLowerCase();
            var matchResult = /^([a-z]+)([_\-]([a-z]+))?$/.exec(lowerCaseLocale);
            if (!matchResult) {
                if (errors) {
                    errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Locale_must_be_of_the_form_language_or_language_territory_For_example_0_or_1, "en", "ja-jp"));
                }
                return;
            }
            var language = matchResult[1];
            var territory = matchResult[3];
            // First try the entire locale, then fall back to just language if that's all we have.
            // Either ways do not fail, and fallback to the English diagnostic strings.
            if (ts.contains(ts.supportedLocaleDirectories, lowerCaseLocale) && !trySetLanguageAndTerritory(language, territory, errors)) {
                trySetLanguageAndTerritory(language, /*territory*/ undefined, errors);
            }
            // Set the UI locale for string collation
            ts.setUILocale(locale);
            function trySetLanguageAndTerritory(language, territory, errors) {
                var compilerFilePath = ts.normalizePath(sys.getExecutingFilePath());
                var containingDirectoryPath = ts.getDirectoryPath(compilerFilePath);
                var filePath = ts.combinePaths(containingDirectoryPath, language);
                if (territory) {
                    filePath = filePath + "-" + territory;
                }
                filePath = sys.resolvePath(ts.combinePaths(filePath, "diagnosticMessages.generated.json"));
                if (!sys.fileExists(filePath)) {
                    return false;
                }
                // TODO: Add codePage support for readFile?
                var fileContents = "";
                try {
                    fileContents = sys.readFile(filePath);
                }
                catch (e) {
                    if (errors) {
                        errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Unable_to_open_file_0, filePath));
                    }
                    return false;
                }
                try {
                    // this is a global mutation (or live binding update)!
                    ts.setLocalizedDiagnosticMessages(JSON.parse(fileContents));
                }
                catch (_a) {
                    if (errors) {
                        errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Corrupted_locale_file_0, filePath));
                    }
                    return false;
                }
                return true;
            }
        }
        ts.validateLocaleAndSetLanguage = validateLocaleAndSetLanguage;
        function getOriginalNode(node, nodeTest) {
            if (node) {
                while (node.original !== undefined) {
                    node = node.original;
                }
            }
            return !nodeTest || nodeTest(node) ? node : undefined;
        }
        ts.getOriginalNode = getOriginalNode;
        function findAncestor(node, callback) {
            while (node) {
                var result = callback(node);
                if (result === "quit") {
                    return undefined;
                }
                else if (result) {
                    return node;
                }
                node = node.parent;
            }
            return undefined;
        }
        ts.findAncestor = findAncestor;
        /**
         * Gets a value indicating whether a node originated in the parse tree.
         *
         * @param node The node to test.
         */
        function isParseTreeNode(node) {
            return (node.flags & 8 /* Synthesized */) === 0;
        }
        ts.isParseTreeNode = isParseTreeNode;
        function getParseTreeNode(node, nodeTest) {
            if (node === undefined || isParseTreeNode(node)) {
                return node;
            }
            node = node.original;
            while (node) {
                if (isParseTreeNode(node)) {
                    return !nodeTest || nodeTest(node) ? node : undefined;
                }
                node = node.original;
            }
        }
        ts.getParseTreeNode = getParseTreeNode;
        /** Add an extra underscore to identifiers that start with two underscores to avoid issues with magic names like '__proto__' */
        function escapeLeadingUnderscores(identifier) {
            return (identifier.length >= 2 && identifier.charCodeAt(0) === 95 /* _ */ && identifier.charCodeAt(1) === 95 /* _ */ ? "_" + identifier : identifier);
        }
        ts.escapeLeadingUnderscores = escapeLeadingUnderscores;
        /**
         * Remove extra underscore from escaped identifier text content.
         *
         * @param identifier The escaped identifier text.
         * @returns The unescaped identifier text.
         */
        function unescapeLeadingUnderscores(identifier) {
            var id = identifier;
            return id.length >= 3 && id.charCodeAt(0) === 95 /* _ */ && id.charCodeAt(1) === 95 /* _ */ && id.charCodeAt(2) === 95 /* _ */ ? id.substr(1) : id;
        }
        ts.unescapeLeadingUnderscores = unescapeLeadingUnderscores;
        function idText(identifierOrPrivateName) {
            return unescapeLeadingUnderscores(identifierOrPrivateName.escapedText);
        }
        ts.idText = idText;
        function symbolName(symbol) {
            if (symbol.valueDeclaration && isPrivateIdentifierClassElementDeclaration(symbol.valueDeclaration)) {
                return idText(symbol.valueDeclaration.name);
            }
            return unescapeLeadingUnderscores(symbol.escapedName);
        }
        ts.symbolName = symbolName;
        /**
         * A JSDocTypedef tag has an _optional_ name field - if a name is not directly present, we should
         * attempt to draw the name from the node the declaration is on (as that declaration is what its' symbol
         * will be merged with)
         */
        function nameForNamelessJSDocTypedef(declaration) {
            var hostNode = declaration.parent.parent;
            if (!hostNode) {
                return undefined;
            }
            // Covers classes, functions - any named declaration host node
            if (isDeclaration(hostNode)) {
                return getDeclarationIdentifier(hostNode);
            }
            // Covers remaining cases (returning undefined if none match).
            switch (hostNode.kind) {
                case 235 /* VariableStatement */:
                    if (hostNode.declarationList && hostNode.declarationList.declarations[0]) {
                        return getDeclarationIdentifier(hostNode.declarationList.declarations[0]);
                    }
                    break;
                case 236 /* ExpressionStatement */:
                    var expr = hostNode.expression;
                    if (expr.kind === 219 /* BinaryExpression */ && expr.operatorToken.kind === 63 /* EqualsToken */) {
                        expr = expr.left;
                    }
                    switch (expr.kind) {
                        case 204 /* PropertyAccessExpression */:
                            return expr.name;
                        case 205 /* ElementAccessExpression */:
                            var arg = expr.argumentExpression;
                            if (ts.isIdentifier(arg)) {
                                return arg;
                            }
                    }
                    break;
                case 210 /* ParenthesizedExpression */: {
                    return getDeclarationIdentifier(hostNode.expression);
                }
                case 248 /* LabeledStatement */: {
                    if (isDeclaration(hostNode.statement) || isExpression(hostNode.statement)) {
                        return getDeclarationIdentifier(hostNode.statement);
                    }
                    break;
                }
            }
        }
        function getDeclarationIdentifier(node) {
            var name = getNameOfDeclaration(node);
            return name && ts.isIdentifier(name) ? name : undefined;
        }
        /** @internal */
        function nodeHasName(statement, name) {
            if (isNamedDeclaration(statement) && ts.isIdentifier(statement.name) && idText(statement.name) === idText(name)) {
                return true;
            }
            if (ts.isVariableStatement(statement) && ts.some(statement.declarationList.declarations, function (d) { return nodeHasName(d, name); })) {
                return true;
            }
            return false;
        }
        ts.nodeHasName = nodeHasName;
        function getNameOfJSDocTypedef(declaration) {
            return declaration.name || nameForNamelessJSDocTypedef(declaration);
        }
        ts.getNameOfJSDocTypedef = getNameOfJSDocTypedef;
        /** @internal */
        function isNamedDeclaration(node) {
            return !!node.name; // A 'name' property should always be a DeclarationName.
        }
        ts.isNamedDeclaration = isNamedDeclaration;
        /** @internal */
        function getNonAssignedNameOfDeclaration(declaration) {
            switch (declaration.kind) {
                case 79 /* Identifier */:
                    return declaration;
                case 342 /* JSDocPropertyTag */:
                case 335 /* JSDocParameterTag */: {
                    var name = declaration.name;
                    if (name.kind === 159 /* QualifiedName */) {
                        return name.right;
                    }
                    break;
                }
                case 206 /* CallExpression */:
                case 219 /* BinaryExpression */: {
                    var expr_1 = declaration;
                    switch (ts.getAssignmentDeclarationKind(expr_1)) {
                        case 1 /* ExportsProperty */:
                        case 4 /* ThisProperty */:
                        case 5 /* Property */:
                        case 3 /* PrototypeProperty */:
                            return ts.getElementOrPropertyAccessArgumentExpressionOrName(expr_1.left);
                        case 7 /* ObjectDefinePropertyValue */:
                        case 8 /* ObjectDefinePropertyExports */:
                        case 9 /* ObjectDefinePrototypeProperty */:
                            return expr_1.arguments[1];
                        default:
                            return undefined;
                    }
                }
                case 340 /* JSDocTypedefTag */:
                    return getNameOfJSDocTypedef(declaration);
                case 334 /* JSDocEnumTag */:
                    return nameForNamelessJSDocTypedef(declaration);
                case 269 /* ExportAssignment */: {
                    var expression = declaration.expression;
                    return ts.isIdentifier(expression) ? expression : undefined;
                }
                case 205 /* ElementAccessExpression */:
                    var expr = declaration;
                    if (ts.isBindableStaticElementAccessExpression(expr)) {
                        return expr.argumentExpression;
                    }
            }
            return declaration.name;
        }
        ts.getNonAssignedNameOfDeclaration = getNonAssignedNameOfDeclaration;
        function getNameOfDeclaration(declaration) {
            if (declaration === undefined)
                return undefined;
            return getNonAssignedNameOfDeclaration(declaration) ||
                (ts.isFunctionExpression(declaration) || ts.isArrowFunction(declaration) || ts.isClassExpression(declaration) ? getAssignedName(declaration) : undefined);
        }
        ts.getNameOfDeclaration = getNameOfDeclaration;
        /*@internal*/
        function getAssignedName(node) {
            if (!node.parent) {
                return undefined;
            }
            else if (ts.isPropertyAssignment(node.parent) || ts.isBindingElement(node.parent)) {
                return node.parent.name;
            }
            else if (ts.isBinaryExpression(node.parent) && node === node.parent.right) {
                if (ts.isIdentifier(node.parent.left)) {
                    return node.parent.left;
                }
                else if (ts.isAccessExpression(node.parent.left)) {
                    return ts.getElementOrPropertyAccessArgumentExpressionOrName(node.parent.left);
                }
            }
            else if (ts.isVariableDeclaration(node.parent) && ts.isIdentifier(node.parent.name)) {
                return node.parent.name;
            }
        }
        ts.getAssignedName = getAssignedName;
        function getJSDocParameterTagsWorker(param, noCache) {
            if (param.name) {
                if (ts.isIdentifier(param.name)) {
                    var name_1 = param.name.escapedText;
                    return getJSDocTagsWorker(param.parent, noCache).filter(function (tag) { return ts.isJSDocParameterTag(tag) && ts.isIdentifier(tag.name) && tag.name.escapedText === name_1; });
                }
                else {
                    var i = param.parent.parameters.indexOf(param);
                    ts.Debug.assert(i > -1, "Parameters should always be in their parents' parameter list");
                    var paramTags = getJSDocTagsWorker(param.parent, noCache).filter(ts.isJSDocParameterTag);
                    if (i < paramTags.length) {
                        return [paramTags[i]];
                    }
                }
            }
            // return empty array for: out-of-order binding patterns and JSDoc function syntax, which has un-named parameters
            return ts.emptyArray;
        }
        /**
         * Gets the JSDoc parameter tags for the node if present.
         *
         * @remarks Returns any JSDoc param tag whose name matches the provided
         * parameter, whether a param tag on a containing function
         * expression, or a param tag on a variable declaration whose
         * initializer is the containing function. The tags closest to the
         * node are returned first, so in the previous example, the param
         * tag on the containing function expression would be first.
         *
         * For binding patterns, parameter tags are matched by position.
         */
        function getJSDocParameterTags(param) {
            return getJSDocParameterTagsWorker(param, /*noCache*/ false);
        }
        ts.getJSDocParameterTags = getJSDocParameterTags;
        /* @internal */
        function getJSDocParameterTagsNoCache(param) {
            return getJSDocParameterTagsWorker(param, /*noCache*/ true);
        }
        ts.getJSDocParameterTagsNoCache = getJSDocParameterTagsNoCache;
        function getJSDocTypeParameterTagsWorker(param, noCache) {
            var name = param.name.escapedText;
            return getJSDocTagsWorker(param.parent, noCache).filter(function (tag) {
                return ts.isJSDocTemplateTag(tag) && tag.typeParameters.some(function (tp) { return tp.name.escapedText === name; });
            });
        }
        /**
         * Gets the JSDoc type parameter tags for the node if present.
         *
         * @remarks Returns any JSDoc template tag whose names match the provided
         * parameter, whether a template tag on a containing function
         * expression, or a template tag on a variable declaration whose
         * initializer is the containing function. The tags closest to the
         * node are returned first, so in the previous example, the template
         * tag on the containing function expression would be first.
         */
        function getJSDocTypeParameterTags(param) {
            return getJSDocTypeParameterTagsWorker(param, /*noCache*/ false);
        }
        ts.getJSDocTypeParameterTags = getJSDocTypeParameterTags;
        /* @internal */
        function getJSDocTypeParameterTagsNoCache(param) {
            return getJSDocTypeParameterTagsWorker(param, /*noCache*/ true);
        }
        ts.getJSDocTypeParameterTagsNoCache = getJSDocTypeParameterTagsNoCache;
        /**
         * Return true if the node has JSDoc parameter tags.
         *
         * @remarks Includes parameter tags that are not directly on the node,
         * for example on a variable declaration whose initializer is a function expression.
         */
        function hasJSDocParameterTags(node) {
            return !!getFirstJSDocTag(node, ts.isJSDocParameterTag);
        }
        ts.hasJSDocParameterTags = hasJSDocParameterTags;
        /** Gets the JSDoc augments tag for the node if present */
        function getJSDocAugmentsTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocAugmentsTag);
        }
        ts.getJSDocAugmentsTag = getJSDocAugmentsTag;
        /** Gets the JSDoc implements tags for the node if present */
        function getJSDocImplementsTags(node) {
            return getAllJSDocTags(node, ts.isJSDocImplementsTag);
        }
        ts.getJSDocImplementsTags = getJSDocImplementsTags;
        /** Gets the JSDoc class tag for the node if present */
        function getJSDocClassTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocClassTag);
        }
        ts.getJSDocClassTag = getJSDocClassTag;
        /** Gets the JSDoc public tag for the node if present */
        function getJSDocPublicTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocPublicTag);
        }
        ts.getJSDocPublicTag = getJSDocPublicTag;
        /*@internal*/
        function getJSDocPublicTagNoCache(node) {
            return getFirstJSDocTag(node, ts.isJSDocPublicTag, /*noCache*/ true);
        }
        ts.getJSDocPublicTagNoCache = getJSDocPublicTagNoCache;
        /** Gets the JSDoc private tag for the node if present */
        function getJSDocPrivateTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocPrivateTag);
        }
        ts.getJSDocPrivateTag = getJSDocPrivateTag;
        /*@internal*/
        function getJSDocPrivateTagNoCache(node) {
            return getFirstJSDocTag(node, ts.isJSDocPrivateTag, /*noCache*/ true);
        }
        ts.getJSDocPrivateTagNoCache = getJSDocPrivateTagNoCache;
        /** Gets the JSDoc protected tag for the node if present */
        function getJSDocProtectedTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocProtectedTag);
        }
        ts.getJSDocProtectedTag = getJSDocProtectedTag;
        /*@internal*/
        function getJSDocProtectedTagNoCache(node) {
            return getFirstJSDocTag(node, ts.isJSDocProtectedTag, /*noCache*/ true);
        }
        ts.getJSDocProtectedTagNoCache = getJSDocProtectedTagNoCache;
        /** Gets the JSDoc protected tag for the node if present */
        function getJSDocReadonlyTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocReadonlyTag);
        }
        ts.getJSDocReadonlyTag = getJSDocReadonlyTag;
        /*@internal*/
        function getJSDocReadonlyTagNoCache(node) {
            return getFirstJSDocTag(node, ts.isJSDocReadonlyTag, /*noCache*/ true);
        }
        ts.getJSDocReadonlyTagNoCache = getJSDocReadonlyTagNoCache;
        function getJSDocOverrideTagNoCache(node) {
            return getFirstJSDocTag(node, ts.isJSDocOverrideTag, /*noCache*/ true);
        }
        ts.getJSDocOverrideTagNoCache = getJSDocOverrideTagNoCache;
        /** Gets the JSDoc deprecated tag for the node if present */
        function getJSDocDeprecatedTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocDeprecatedTag);
        }
        ts.getJSDocDeprecatedTag = getJSDocDeprecatedTag;
        /*@internal */
        function getJSDocDeprecatedTagNoCache(node) {
            return getFirstJSDocTag(node, ts.isJSDocDeprecatedTag, /*noCache*/ true);
        }
        ts.getJSDocDeprecatedTagNoCache = getJSDocDeprecatedTagNoCache;
        /** Gets the JSDoc enum tag for the node if present */
        function getJSDocEnumTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocEnumTag);
        }
        ts.getJSDocEnumTag = getJSDocEnumTag;
        /** Gets the JSDoc this tag for the node if present */
        function getJSDocThisTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocThisTag);
        }
        ts.getJSDocThisTag = getJSDocThisTag;
        /** Gets the JSDoc return tag for the node if present */
        function getJSDocReturnTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocReturnTag);
        }
        ts.getJSDocReturnTag = getJSDocReturnTag;
        /** Gets the JSDoc template tag for the node if present */
        function getJSDocTemplateTag(node) {
            return getFirstJSDocTag(node, ts.isJSDocTemplateTag);
        }
        ts.getJSDocTemplateTag = getJSDocTemplateTag;
        /** Gets the JSDoc type tag for the node if present and valid */
        function getJSDocTypeTag(node) {
            // We should have already issued an error if there were multiple type jsdocs, so just use the first one.
            var tag = getFirstJSDocTag(node, ts.isJSDocTypeTag);
            if (tag && tag.typeExpression && tag.typeExpression.type) {
                return tag;
            }
            return undefined;
        }
        ts.getJSDocTypeTag = getJSDocTypeTag;
        /**
         * Gets the type node for the node if provided via JSDoc.
         *
         * @remarks The search includes any JSDoc param tag that relates
         * to the provided parameter, for example a type tag on the
         * parameter itself, or a param tag on a containing function
         * expression, or a param tag on a variable declaration whose
         * initializer is the containing function. The tags closest to the
         * node are examined first, so in the previous example, the type
         * tag directly on the node would be returned.
         */
        function getJSDocType(node) {
            var tag = getFirstJSDocTag(node, ts.isJSDocTypeTag);
            if (!tag && ts.isParameter(node)) {
                tag = ts.find(getJSDocParameterTags(node), function (tag) { return !!tag.typeExpression; });
            }
            return tag && tag.typeExpression && tag.typeExpression.type;
        }
        ts.getJSDocType = getJSDocType;
        /**
         * Gets the return type node for the node if provided via JSDoc return tag or type tag.
         *
         * @remarks `getJSDocReturnTag` just gets the whole JSDoc tag. This function
         * gets the type from inside the braces, after the fat arrow, etc.
         */
        function getJSDocReturnType(node) {
            var returnTag = getJSDocReturnTag(node);
            if (returnTag && returnTag.typeExpression) {
                return returnTag.typeExpression.type;
            }
            var typeTag = getJSDocTypeTag(node);
            if (typeTag && typeTag.typeExpression) {
                var type = typeTag.typeExpression.type;
                if (ts.isTypeLiteralNode(type)) {
                    var sig = ts.find(type.members, ts.isCallSignatureDeclaration);
                    return sig && sig.type;
                }
                if (ts.isFunctionTypeNode(type) || ts.isJSDocFunctionType(type)) {
                    return type.type;
                }
            }
        }
        ts.getJSDocReturnType = getJSDocReturnType;
        function getJSDocTagsWorker(node, noCache) {
            var tags = node.jsDocCache;
            // If cache is 'null', that means we did the work of searching for JSDoc tags and came up with nothing.
            if (tags === undefined || noCache) {
                var comments = ts.getJSDocCommentsAndTags(node, noCache);
                ts.Debug.assert(comments.length < 2 || comments[0] !== comments[1]);
                tags = ts.flatMap(comments, function (j) { return ts.isJSDoc(j) ? j.tags : j; });
                if (!noCache) {
                    node.jsDocCache = tags;
                }
            }
            return tags;
        }
        /** Get all JSDoc tags related to a node, including those on parent nodes. */
        function getJSDocTags(node) {
            return getJSDocTagsWorker(node, /*noCache*/ false);
        }
        ts.getJSDocTags = getJSDocTags;
        /* @internal */
        function getJSDocTagsNoCache(node) {
            return getJSDocTagsWorker(node, /*noCache*/ true);
        }
        ts.getJSDocTagsNoCache = getJSDocTagsNoCache;
        /** Get the first JSDoc tag of a specified kind, or undefined if not present. */
        function getFirstJSDocTag(node, predicate, noCache) {
            return ts.find(getJSDocTagsWorker(node, noCache), predicate);
        }
        /** Gets all JSDoc tags that match a specified predicate */
        function getAllJSDocTags(node, predicate) {
            return getJSDocTags(node).filter(predicate);
        }
        ts.getAllJSDocTags = getAllJSDocTags;
        /** Gets all JSDoc tags of a specified kind */
        function getAllJSDocTagsOfKind(node, kind) {
            return getJSDocTags(node).filter(function (doc) { return doc.kind === kind; });
        }
        ts.getAllJSDocTagsOfKind = getAllJSDocTagsOfKind;
        /** Gets the text of a jsdoc comment, flattening links to their text. */
        function getTextOfJSDocComment(comment) {
            return typeof comment === "string" ? comment
                : comment === null || comment === void 0 ? void 0 : comment.map(function (c) {
                    // TODO: Other kinds here
                    return c.kind === 316 /* JSDocText */ ? c.text : "{@link " + (c.name ? ts.entityNameToString(c.name) + " " : "") + c.text + "}";
                }).join("");
        }
        ts.getTextOfJSDocComment = getTextOfJSDocComment;
        /**
         * Gets the effective type parameters. If the node was parsed in a
         * JavaScript file, gets the type parameters from the `@template` tag from JSDoc.
         */
        function getEffectiveTypeParameterDeclarations(node) {
            if (ts.isJSDocSignature(node)) {
                return ts.emptyArray;
            }
            if (ts.isJSDocTypeAlias(node)) {
                ts.Debug.assert(node.parent.kind === 315 /* JSDocComment */);
                return ts.flatMap(node.parent.tags, function (tag) { return ts.isJSDocTemplateTag(tag) ? tag.typeParameters : undefined; });
            }
            if (node.typeParameters) {
                return node.typeParameters;
            }
            if (ts.isInJSFile(node)) {
                var decls = ts.getJSDocTypeParameterDeclarations(node);
                if (decls.length) {
                    return decls;
                }
                var typeTag = getJSDocType(node);
                if (typeTag && ts.isFunctionTypeNode(typeTag) && typeTag.typeParameters) {
                    return typeTag.typeParameters;
                }
            }
            return ts.emptyArray;
        }
        ts.getEffectiveTypeParameterDeclarations = getEffectiveTypeParameterDeclarations;
        function getEffectiveConstraintOfTypeParameter(node) {
            return node.constraint ? node.constraint :
                ts.isJSDocTemplateTag(node.parent) && node === node.parent.typeParameters[0] ? node.parent.constraint :
                    undefined;
        }
        ts.getEffectiveConstraintOfTypeParameter = getEffectiveConstraintOfTypeParameter;
        // #region
        function isMemberName(node) {
            return node.kind === 79 /* Identifier */ || node.kind === 80 /* PrivateIdentifier */;
        }
        ts.isMemberName = isMemberName;
        /* @internal */
        function isGetOrSetAccessorDeclaration(node) {
            return node.kind === 171 /* SetAccessor */ || node.kind === 170 /* GetAccessor */;
        }
        ts.isGetOrSetAccessorDeclaration = isGetOrSetAccessorDeclaration;
        function isPropertyAccessChain(node) {
            return ts.isPropertyAccessExpression(node) && !!(node.flags & 32 /* OptionalChain */);
        }
        ts.isPropertyAccessChain = isPropertyAccessChain;
        function isElementAccessChain(node) {
            return ts.isElementAccessExpression(node) && !!(node.flags & 32 /* OptionalChain */);
        }
        ts.isElementAccessChain = isElementAccessChain;
        function isCallChain(node) {
            return ts.isCallExpression(node) && !!(node.flags & 32 /* OptionalChain */);
        }
        ts.isCallChain = isCallChain;
        function isOptionalChain(node) {
            var kind = node.kind;
            return !!(node.flags & 32 /* OptionalChain */) &&
                (kind === 204 /* PropertyAccessExpression */
                    || kind === 205 /* ElementAccessExpression */
                    || kind === 206 /* CallExpression */
                    || kind === 228 /* NonNullExpression */);
        }
        ts.isOptionalChain = isOptionalChain;
        /* @internal */
        function isOptionalChainRoot(node) {
            return isOptionalChain(node) && !ts.isNonNullExpression(node) && !!node.questionDotToken;
        }
        ts.isOptionalChainRoot = isOptionalChainRoot;
        /**
         * Determines whether a node is the expression preceding an optional chain (i.e. `a` in `a?.b`).
         */
        /* @internal */
        function isExpressionOfOptionalChainRoot(node) {
            return isOptionalChainRoot(node.parent) && node.parent.expression === node;
        }
        ts.isExpressionOfOptionalChainRoot = isExpressionOfOptionalChainRoot;
        /**
         * Determines whether a node is the outermost `OptionalChain` in an ECMAScript `OptionalExpression`:
         *
         * 1. For `a?.b.c`, the outermost chain is `a?.b.c` (`c` is the end of the chain starting at `a?.`)
         * 2. For `a?.b!`, the outermost chain is `a?.b` (`b` is the end of the chain starting at `a?.`)
         * 3. For `(a?.b.c).d`, the outermost chain is `a?.b.c` (`c` is the end of the chain starting at `a?.` since parens end the chain)
         * 4. For `a?.b.c?.d`, both `a?.b.c` and `a?.b.c?.d` are outermost (`c` is the end of the chain starting at `a?.`, and `d` is
         *   the end of the chain starting at `c?.`)
         * 5. For `a?.(b?.c).d`, both `b?.c` and `a?.(b?.c)d` are outermost (`c` is the end of the chain starting at `b`, and `d` is
         *   the end of the chain starting at `a?.`)
         */
        /* @internal */
        function isOutermostOptionalChain(node) {
            return !isOptionalChain(node.parent) // cases 1, 2, and 3
                || isOptionalChainRoot(node.parent) // case 4
                || node !== node.parent.expression; // case 5
        }
        ts.isOutermostOptionalChain = isOutermostOptionalChain;
        function isNullishCoalesce(node) {
            return node.kind === 219 /* BinaryExpression */ && node.operatorToken.kind === 60 /* QuestionQuestionToken */;
        }
        ts.isNullishCoalesce = isNullishCoalesce;
        function isConstTypeReference(node) {
            return ts.isTypeReferenceNode(node) && ts.isIdentifier(node.typeName) &&
                node.typeName.escapedText === "const" && !node.typeArguments;
        }
        ts.isConstTypeReference = isConstTypeReference;
        function skipPartiallyEmittedExpressions(node) {
            return ts.skipOuterExpressions(node, 8 /* PartiallyEmittedExpressions */);
        }
        ts.skipPartiallyEmittedExpressions = skipPartiallyEmittedExpressions;
        function isNonNullChain(node) {
            return ts.isNonNullExpression(node) && !!(node.flags & 32 /* OptionalChain */);
        }
        ts.isNonNullChain = isNonNullChain;
        function isBreakOrContinueStatement(node) {
            return node.kind === 244 /* BreakStatement */ || node.kind === 243 /* ContinueStatement */;
        }
        ts.isBreakOrContinueStatement = isBreakOrContinueStatement;
        function isNamedExportBindings(node) {
            return node.kind === 272 /* NamespaceExport */ || node.kind === 271 /* NamedExports */;
        }
        ts.isNamedExportBindings = isNamedExportBindings;
        function isUnparsedTextLike(node) {
            switch (node.kind) {
                case 297 /* UnparsedText */:
                case 298 /* UnparsedInternalText */:
                    return true;
                default:
                    return false;
            }
        }
        ts.isUnparsedTextLike = isUnparsedTextLike;
        function isUnparsedNode(node) {
            return isUnparsedTextLike(node) ||
                node.kind === 295 /* UnparsedPrologue */ ||
                node.kind === 299 /* UnparsedSyntheticReference */;
        }
        ts.isUnparsedNode = isUnparsedNode;
        function isJSDocPropertyLikeTag(node) {
            return node.kind === 342 /* JSDocPropertyTag */ || node.kind === 335 /* JSDocParameterTag */;
        }
        ts.isJSDocPropertyLikeTag = isJSDocPropertyLikeTag;
        // #endregion
        // #region
        // Node tests
        //
        // All node tests in the following list should *not* reference parent pointers so that
        // they may be used with transformations.
        /* @internal */
        function isNode(node) {
            return isNodeKind(node.kind);
        }
        ts.isNode = isNode;
        /* @internal */
        function isNodeKind(kind) {
            return kind >= 159 /* FirstNode */;
        }
        ts.isNodeKind = isNodeKind;
        /**
         * True if kind is of some token syntax kind.
         * For example, this is true for an IfKeyword but not for an IfStatement.
         * Literals are considered tokens, except TemplateLiteral, but does include TemplateHead/Middle/Tail.
         */
        function isTokenKind(kind) {
            return kind >= 0 /* FirstToken */ && kind <= 158 /* LastToken */;
        }
        ts.isTokenKind = isTokenKind;
        /**
         * True if node is of some token syntax kind.
         * For example, this is true for an IfKeyword but not for an IfStatement.
         * Literals are considered tokens, except TemplateLiteral, but does include TemplateHead/Middle/Tail.
         */
        function isToken(n) {
            return isTokenKind(n.kind);
        }
        ts.isToken = isToken;
        // Node Arrays
        /* @internal */
        function isNodeArray(array) {
            return array.hasOwnProperty("pos") && array.hasOwnProperty("end");
        }
        ts.isNodeArray = isNodeArray;
        // Literals
        /* @internal */
        function isLiteralKind(kind) {
            return 8 /* FirstLiteralToken */ <= kind && kind <= 14 /* LastLiteralToken */;
        }
        ts.isLiteralKind = isLiteralKind;
        function isLiteralExpression(node) {
            return isLiteralKind(node.kind);
        }
        ts.isLiteralExpression = isLiteralExpression;
        // Pseudo-literals
        /* @internal */
        function isTemplateLiteralKind(kind) {
            return 14 /* FirstTemplateToken */ <= kind && kind <= 17 /* LastTemplateToken */;
        }
        ts.isTemplateLiteralKind = isTemplateLiteralKind;
        function isTemplateLiteralToken(node) {
            return isTemplateLiteralKind(node.kind);
        }
        ts.isTemplateLiteralToken = isTemplateLiteralToken;
        function isTemplateMiddleOrTemplateTail(node) {
            var kind = node.kind;
            return kind === 16 /* TemplateMiddle */
                || kind === 17 /* TemplateTail */;
        }
        ts.isTemplateMiddleOrTemplateTail = isTemplateMiddleOrTemplateTail;
        function isImportOrExportSpecifier(node) {
            return ts.isImportSpecifier(node) || ts.isExportSpecifier(node);
        }
        ts.isImportOrExportSpecifier = isImportOrExportSpecifier;
        function isTypeOnlyImportOrExportDeclaration(node) {
            switch (node.kind) {
                case 268 /* ImportSpecifier */:
                case 273 /* ExportSpecifier */:
                    return node.parent.parent.isTypeOnly;
                case 266 /* NamespaceImport */:
                    return node.parent.isTypeOnly;
                case 265 /* ImportClause */:
                case 263 /* ImportEqualsDeclaration */:
                    return node.isTypeOnly;
                default:
                    return false;
            }
        }
        ts.isTypeOnlyImportOrExportDeclaration = isTypeOnlyImportOrExportDeclaration;
        function isStringTextContainingNode(node) {
            return node.kind === 10 /* StringLiteral */ || isTemplateLiteralKind(node.kind);
        }
        ts.isStringTextContainingNode = isStringTextContainingNode;
        // Identifiers
        /* @internal */
        function isGeneratedIdentifier(node) {
            return ts.isIdentifier(node) && (node.autoGenerateFlags & 7 /* KindMask */) > 0 /* None */;
        }
        ts.isGeneratedIdentifier = isGeneratedIdentifier;
        // Private Identifiers
        /*@internal*/
        function isPrivateIdentifierClassElementDeclaration(node) {
            return (ts.isPropertyDeclaration(node) || isMethodOrAccessor(node)) && ts.isPrivateIdentifier(node.name);
        }
        ts.isPrivateIdentifierClassElementDeclaration = isPrivateIdentifierClassElementDeclaration;
        /*@internal*/
        function isPrivateIdentifierPropertyAccessExpression(node) {
            return ts.isPropertyAccessExpression(node) && ts.isPrivateIdentifier(node.name);
        }
        ts.isPrivateIdentifierPropertyAccessExpression = isPrivateIdentifierPropertyAccessExpression;
        // Keywords
        /* @internal */
        function isModifierKind(token) {
            switch (token) {
                case 126 /* AbstractKeyword */:
                case 130 /* AsyncKeyword */:
                case 85 /* ConstKeyword */:
                case 134 /* DeclareKeyword */:
                case 88 /* DefaultKeyword */:
                case 93 /* ExportKeyword */:
                case 123 /* PublicKeyword */:
                case 121 /* PrivateKeyword */:
                case 122 /* ProtectedKeyword */:
                case 143 /* ReadonlyKeyword */:
                case 124 /* StaticKeyword */:
                case 157 /* OverrideKeyword */:
                    return true;
            }
            return false;
        }
        ts.isModifierKind = isModifierKind;
        /* @internal */
        function isParameterPropertyModifier(kind) {
            return !!(ts.modifierToFlag(kind) & 16476 /* ParameterPropertyModifier */);
        }
        ts.isParameterPropertyModifier = isParameterPropertyModifier;
        /* @internal */
        function isClassMemberModifier(idToken) {
            return isParameterPropertyModifier(idToken) || idToken === 124 /* StaticKeyword */ || idToken === 157 /* OverrideKeyword */;
        }
        ts.isClassMemberModifier = isClassMemberModifier;
        function isModifier(node) {
            return isModifierKind(node.kind);
        }
        ts.isModifier = isModifier;
        function isEntityName(node) {
            var kind = node.kind;
            return kind === 159 /* QualifiedName */
                || kind === 79 /* Identifier */;
        }
        ts.isEntityName = isEntityName;
        function isPropertyName(node) {
            var kind = node.kind;
            return kind === 79 /* Identifier */
                || kind === 80 /* PrivateIdentifier */
                || kind === 10 /* StringLiteral */
                || kind === 8 /* NumericLiteral */
                || kind === 160 /* ComputedPropertyName */;
        }
        ts.isPropertyName = isPropertyName;
        function isBindingName(node) {
            var kind = node.kind;
            return kind === 79 /* Identifier */
                || kind === 199 /* ObjectBindingPattern */
                || kind === 200 /* ArrayBindingPattern */;
        }
        ts.isBindingName = isBindingName;
        // Functions
        function isFunctionLike(node) {
            return !!node && isFunctionLikeKind(node.kind);
        }
        ts.isFunctionLike = isFunctionLike;
        /* @internal */
        function isFunctionLikeOrClassStaticBlockDeclaration(node) {
            return !!node && (isFunctionLikeKind(node.kind) || ts.isClassStaticBlockDeclaration(node));
        }
        ts.isFunctionLikeOrClassStaticBlockDeclaration = isFunctionLikeOrClassStaticBlockDeclaration;
        /* @internal */
        function isFunctionLikeDeclaration(node) {
            return node && isFunctionLikeDeclarationKind(node.kind);
        }
        ts.isFunctionLikeDeclaration = isFunctionLikeDeclaration;
        /* @internal */
        function isBooleanLiteral(node) {
            return node.kind === 110 /* TrueKeyword */ || node.kind === 95 /* FalseKeyword */;
        }
        ts.isBooleanLiteral = isBooleanLiteral;
        function isFunctionLikeDeclarationKind(kind) {
            switch (kind) {
                case 254 /* FunctionDeclaration */:
                case 167 /* MethodDeclaration */:
                case 169 /* Constructor */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 211 /* FunctionExpression */:
                case 212 /* ArrowFunction */:
                    return true;
                default:
                    return false;
            }
        }
        /* @internal */
        function isFunctionLikeKind(kind) {
            switch (kind) {
                case 166 /* MethodSignature */:
                case 172 /* CallSignature */:
                case 318 /* JSDocSignature */:
                case 173 /* ConstructSignature */:
                case 174 /* IndexSignature */:
                case 177 /* FunctionType */:
                case 312 /* JSDocFunctionType */:
                case 178 /* ConstructorType */:
                    return true;
                default:
                    return isFunctionLikeDeclarationKind(kind);
            }
        }
        ts.isFunctionLikeKind = isFunctionLikeKind;
        /* @internal */
        function isFunctionOrModuleBlock(node) {
            return ts.isSourceFile(node) || ts.isModuleBlock(node) || ts.isBlock(node) && isFunctionLike(node.parent);
        }
        ts.isFunctionOrModuleBlock = isFunctionOrModuleBlock;
        // Classes
        function isClassElement(node) {
            var kind = node.kind;
            return kind === 169 /* Constructor */
                || kind === 165 /* PropertyDeclaration */
                || kind === 167 /* MethodDeclaration */
                || kind === 170 /* GetAccessor */
                || kind === 171 /* SetAccessor */
                || kind === 174 /* IndexSignature */
                || kind === 168 /* ClassStaticBlockDeclaration */
                || kind === 232 /* SemicolonClassElement */;
        }
        ts.isClassElement = isClassElement;
        function isClassLike(node) {
            return node && (node.kind === 255 /* ClassDeclaration */ || node.kind === 224 /* ClassExpression */);
        }
        ts.isClassLike = isClassLike;
        function isAccessor(node) {
            return node && (node.kind === 170 /* GetAccessor */ || node.kind === 171 /* SetAccessor */);
        }
        ts.isAccessor = isAccessor;
        /* @internal */
        function isMethodOrAccessor(node) {
            switch (node.kind) {
                case 167 /* MethodDeclaration */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                    return true;
                default:
                    return false;
            }
        }
        ts.isMethodOrAccessor = isMethodOrAccessor;
        // Type members
        function isTypeElement(node) {
            var kind = node.kind;
            return kind === 173 /* ConstructSignature */
                || kind === 172 /* CallSignature */
                || kind === 164 /* PropertySignature */
                || kind === 166 /* MethodSignature */
                || kind === 174 /* IndexSignature */;
        }
        ts.isTypeElement = isTypeElement;
        function isClassOrTypeElement(node) {
            return isTypeElement(node) || isClassElement(node);
        }
        ts.isClassOrTypeElement = isClassOrTypeElement;
        function isObjectLiteralElementLike(node) {
            var kind = node.kind;
            return kind === 291 /* PropertyAssignment */
                || kind === 292 /* ShorthandPropertyAssignment */
                || kind === 293 /* SpreadAssignment */
                || kind === 167 /* MethodDeclaration */
                || kind === 170 /* GetAccessor */
                || kind === 171 /* SetAccessor */;
        }
        ts.isObjectLiteralElementLike = isObjectLiteralElementLike;
        // Type
        /**
         * Node test that determines whether a node is a valid type node.
         * This differs from the `isPartOfTypeNode` function which determines whether a node is *part*
         * of a TypeNode.
         */
        function isTypeNode(node) {
            return ts.isTypeNodeKind(node.kind);
        }
        ts.isTypeNode = isTypeNode;
        function isFunctionOrConstructorTypeNode(node) {
            switch (node.kind) {
                case 177 /* FunctionType */:
                case 178 /* ConstructorType */:
                    return true;
            }
            return false;
        }
        ts.isFunctionOrConstructorTypeNode = isFunctionOrConstructorTypeNode;
        // Binding patterns
        /* @internal */
        function isBindingPattern(node) {
            if (node) {
                var kind = node.kind;
                return kind === 200 /* ArrayBindingPattern */
                    || kind === 199 /* ObjectBindingPattern */;
            }
            return false;
        }
        ts.isBindingPattern = isBindingPattern;
        /* @internal */
        function isAssignmentPattern(node) {
            var kind = node.kind;
            return kind === 202 /* ArrayLiteralExpression */
                || kind === 203 /* ObjectLiteralExpression */;
        }
        ts.isAssignmentPattern = isAssignmentPattern;
        /* @internal */
        function isArrayBindingElement(node) {
            var kind = node.kind;
            return kind === 201 /* BindingElement */
                || kind === 225 /* OmittedExpression */;
        }
        ts.isArrayBindingElement = isArrayBindingElement;
        /**
         * Determines whether the BindingOrAssignmentElement is a BindingElement-like declaration
         */
        /* @internal */
        function isDeclarationBindingElement(bindingElement) {
            switch (bindingElement.kind) {
                case 252 /* VariableDeclaration */:
                case 162 /* Parameter */:
                case 201 /* BindingElement */:
                    return true;
            }
            return false;
        }
        ts.isDeclarationBindingElement = isDeclarationBindingElement;
        /**
         * Determines whether a node is a BindingOrAssignmentPattern
         */
        /* @internal */
        function isBindingOrAssignmentPattern(node) {
            return isObjectBindingOrAssignmentPattern(node)
                || isArrayBindingOrAssignmentPattern(node);
        }
        ts.isBindingOrAssignmentPattern = isBindingOrAssignmentPattern;
        /**
         * Determines whether a node is an ObjectBindingOrAssignmentPattern
         */
        /* @internal */
        function isObjectBindingOrAssignmentPattern(node) {
            switch (node.kind) {
                case 199 /* ObjectBindingPattern */:
                case 203 /* ObjectLiteralExpression */:
                    return true;
            }
            return false;
        }
        ts.isObjectBindingOrAssignmentPattern = isObjectBindingOrAssignmentPattern;
        /* @internal */
        function isObjectBindingOrAssignmentElement(node) {
            switch (node.kind) {
                case 201 /* BindingElement */:
                case 291 /* PropertyAssignment */: // AssignmentProperty
                case 292 /* ShorthandPropertyAssignment */: // AssignmentProperty
                case 293 /* SpreadAssignment */: // AssignmentRestProperty
                    return true;
            }
            return false;
        }
        ts.isObjectBindingOrAssignmentElement = isObjectBindingOrAssignmentElement;
        /**
         * Determines whether a node is an ArrayBindingOrAssignmentPattern
         */
        /* @internal */
        function isArrayBindingOrAssignmentPattern(node) {
            switch (node.kind) {
                case 200 /* ArrayBindingPattern */:
                case 202 /* ArrayLiteralExpression */:
                    return true;
            }
            return false;
        }
        ts.isArrayBindingOrAssignmentPattern = isArrayBindingOrAssignmentPattern;
        /* @internal */
        function isPropertyAccessOrQualifiedNameOrImportTypeNode(node) {
            var kind = node.kind;
            return kind === 204 /* PropertyAccessExpression */
                || kind === 159 /* QualifiedName */
                || kind === 198 /* ImportType */;
        }
        ts.isPropertyAccessOrQualifiedNameOrImportTypeNode = isPropertyAccessOrQualifiedNameOrImportTypeNode;
        // Expression
        function isPropertyAccessOrQualifiedName(node) {
            var kind = node.kind;
            return kind === 204 /* PropertyAccessExpression */
                || kind === 159 /* QualifiedName */;
        }
        ts.isPropertyAccessOrQualifiedName = isPropertyAccessOrQualifiedName;
        function isCallLikeExpression(node) {
            switch (node.kind) {
                case 278 /* JsxOpeningElement */:
                case 277 /* JsxSelfClosingElement */:
                case 206 /* CallExpression */:
                case 207 /* NewExpression */:
                case 208 /* TaggedTemplateExpression */:
                case 163 /* Decorator */:
                    return true;
                default:
                    return false;
            }
        }
        ts.isCallLikeExpression = isCallLikeExpression;
        function isCallOrNewExpression(node) {
            return node.kind === 206 /* CallExpression */ || node.kind === 207 /* NewExpression */;
        }
        ts.isCallOrNewExpression = isCallOrNewExpression;
        function isTemplateLiteral(node) {
            var kind = node.kind;
            return kind === 221 /* TemplateExpression */
                || kind === 14 /* NoSubstitutionTemplateLiteral */;
        }
        ts.isTemplateLiteral = isTemplateLiteral;
        /* @internal */
        function isLeftHandSideExpression(node) {
            return isLeftHandSideExpressionKind(skipPartiallyEmittedExpressions(node).kind);
        }
        ts.isLeftHandSideExpression = isLeftHandSideExpression;
        function isLeftHandSideExpressionKind(kind) {
            switch (kind) {
                case 204 /* PropertyAccessExpression */:
                case 205 /* ElementAccessExpression */:
                case 207 /* NewExpression */:
                case 206 /* CallExpression */:
                case 276 /* JsxElement */:
                case 277 /* JsxSelfClosingElement */:
                case 280 /* JsxFragment */:
                case 208 /* TaggedTemplateExpression */:
                case 202 /* ArrayLiteralExpression */:
                case 210 /* ParenthesizedExpression */:
                case 203 /* ObjectLiteralExpression */:
                case 224 /* ClassExpression */:
                case 211 /* FunctionExpression */:
                case 79 /* Identifier */:
                case 13 /* RegularExpressionLiteral */:
                case 8 /* NumericLiteral */:
                case 9 /* BigIntLiteral */:
                case 10 /* StringLiteral */:
                case 14 /* NoSubstitutionTemplateLiteral */:
                case 221 /* TemplateExpression */:
                case 95 /* FalseKeyword */:
                case 104 /* NullKeyword */:
                case 108 /* ThisKeyword */:
                case 110 /* TrueKeyword */:
                case 106 /* SuperKeyword */:
                case 228 /* NonNullExpression */:
                case 229 /* MetaProperty */:
                case 100 /* ImportKeyword */: // technically this is only an Expression if it's in a CallExpression
                    return true;
                default:
                    return false;
            }
        }
        /* @internal */
        function isUnaryExpression(node) {
            return isUnaryExpressionKind(skipPartiallyEmittedExpressions(node).kind);
        }
        ts.isUnaryExpression = isUnaryExpression;
        function isUnaryExpressionKind(kind) {
            switch (kind) {
                case 217 /* PrefixUnaryExpression */:
                case 218 /* PostfixUnaryExpression */:
                case 213 /* DeleteExpression */:
                case 214 /* TypeOfExpression */:
                case 215 /* VoidExpression */:
                case 216 /* AwaitExpression */:
                case 209 /* TypeAssertionExpression */:
                    return true;
                default:
                    return isLeftHandSideExpressionKind(kind);
            }
        }
        /* @internal */
        function isUnaryExpressionWithWrite(expr) {
            switch (expr.kind) {
                case 218 /* PostfixUnaryExpression */:
                    return true;
                case 217 /* PrefixUnaryExpression */:
                    return expr.operator === 45 /* PlusPlusToken */ ||
                        expr.operator === 46 /* MinusMinusToken */;
                default:
                    return false;
            }
        }
        ts.isUnaryExpressionWithWrite = isUnaryExpressionWithWrite;
        /* @internal */
        /**
         * Determines whether a node is an expression based only on its kind.
         * Use `isExpressionNode` if not in transforms.
         */
        function isExpression(node) {
            return isExpressionKind(skipPartiallyEmittedExpressions(node).kind);
        }
        ts.isExpression = isExpression;
        function isExpressionKind(kind) {
            switch (kind) {
                case 220 /* ConditionalExpression */:
                case 222 /* YieldExpression */:
                case 212 /* ArrowFunction */:
                case 219 /* BinaryExpression */:
                case 223 /* SpreadElement */:
                case 227 /* AsExpression */:
                case 225 /* OmittedExpression */:
                case 346 /* CommaListExpression */:
                case 345 /* PartiallyEmittedExpression */:
                    return true;
                default:
                    return isUnaryExpressionKind(kind);
            }
        }
        function isAssertionExpression(node) {
            var kind = node.kind;
            return kind === 209 /* TypeAssertionExpression */
                || kind === 227 /* AsExpression */;
        }
        ts.isAssertionExpression = isAssertionExpression;
        /* @internal */
        function isNotEmittedOrPartiallyEmittedNode(node) {
            return ts.isNotEmittedStatement(node)
                || ts.isPartiallyEmittedExpression(node);
        }
        ts.isNotEmittedOrPartiallyEmittedNode = isNotEmittedOrPartiallyEmittedNode;
        function isIterationStatement(node, lookInLabeledStatements) {
            switch (node.kind) {
                case 240 /* ForStatement */:
                case 241 /* ForInStatement */:
                case 242 /* ForOfStatement */:
                case 238 /* DoStatement */:
                case 239 /* WhileStatement */:
                    return true;
                case 248 /* LabeledStatement */:
                    return lookInLabeledStatements && isIterationStatement(node.statement, lookInLabeledStatements);
            }
            return false;
        }
        ts.isIterationStatement = isIterationStatement;
        /* @internal */
        function isScopeMarker(node) {
            return ts.isExportAssignment(node) || ts.isExportDeclaration(node);
        }
        ts.isScopeMarker = isScopeMarker;
        /* @internal */
        function hasScopeMarker(statements) {
            return ts.some(statements, isScopeMarker);
        }
        ts.hasScopeMarker = hasScopeMarker;
        /* @internal */
        function needsScopeMarker(result) {
            return !ts.isAnyImportOrReExport(result) && !ts.isExportAssignment(result) && !ts.hasSyntacticModifier(result, 1 /* Export */) && !ts.isAmbientModule(result);
        }
        ts.needsScopeMarker = needsScopeMarker;
        /* @internal */
        function isExternalModuleIndicator(result) {
            // Exported top-level member indicates moduleness
            return ts.isAnyImportOrReExport(result) || ts.isExportAssignment(result) || ts.hasSyntacticModifier(result, 1 /* Export */);
        }
        ts.isExternalModuleIndicator = isExternalModuleIndicator;
        /* @internal */
        function isForInOrOfStatement(node) {
            return node.kind === 241 /* ForInStatement */ || node.kind === 242 /* ForOfStatement */;
        }
        ts.isForInOrOfStatement = isForInOrOfStatement;
        // Element
        /* @internal */
        function isConciseBody(node) {
            return ts.isBlock(node)
                || isExpression(node);
        }
        ts.isConciseBody = isConciseBody;
        /* @internal */
        function isFunctionBody(node) {
            return ts.isBlock(node);
        }
        ts.isFunctionBody = isFunctionBody;
        /* @internal */
        function isForInitializer(node) {
            return ts.isVariableDeclarationList(node)
                || isExpression(node);
        }
        ts.isForInitializer = isForInitializer;
        /* @internal */
        function isModuleBody(node) {
            var kind = node.kind;
            return kind === 260 /* ModuleBlock */
                || kind === 259 /* ModuleDeclaration */
                || kind === 79 /* Identifier */;
        }
        ts.isModuleBody = isModuleBody;
        /* @internal */
        function isNamespaceBody(node) {
            var kind = node.kind;
            return kind === 260 /* ModuleBlock */
                || kind === 259 /* ModuleDeclaration */;
        }
        ts.isNamespaceBody = isNamespaceBody;
        /* @internal */
        function isJSDocNamespaceBody(node) {
            var kind = node.kind;
            return kind === 79 /* Identifier */
                || kind === 259 /* ModuleDeclaration */;
        }
        ts.isJSDocNamespaceBody = isJSDocNamespaceBody;
        /* @internal */
        function isNamedImportBindings(node) {
            var kind = node.kind;
            return kind === 267 /* NamedImports */
                || kind === 266 /* NamespaceImport */;
        }
        ts.isNamedImportBindings = isNamedImportBindings;
        /* @internal */
        function isModuleOrEnumDeclaration(node) {
            return node.kind === 259 /* ModuleDeclaration */ || node.kind === 258 /* EnumDeclaration */;
        }
        ts.isModuleOrEnumDeclaration = isModuleOrEnumDeclaration;
        function isDeclarationKind(kind) {
            return kind === 212 /* ArrowFunction */
                || kind === 201 /* BindingElement */
                || kind === 255 /* ClassDeclaration */
                || kind === 224 /* ClassExpression */
                || kind === 168 /* ClassStaticBlockDeclaration */
                || kind === 169 /* Constructor */
                || kind === 258 /* EnumDeclaration */
                || kind === 294 /* EnumMember */
                || kind === 273 /* ExportSpecifier */
                || kind === 254 /* FunctionDeclaration */
                || kind === 211 /* FunctionExpression */
                || kind === 170 /* GetAccessor */
                || kind === 265 /* ImportClause */
                || kind === 263 /* ImportEqualsDeclaration */
                || kind === 268 /* ImportSpecifier */
                || kind === 256 /* InterfaceDeclaration */
                || kind === 283 /* JsxAttribute */
                || kind === 167 /* MethodDeclaration */
                || kind === 166 /* MethodSignature */
                || kind === 259 /* ModuleDeclaration */
                || kind === 262 /* NamespaceExportDeclaration */
                || kind === 266 /* NamespaceImport */
                || kind === 272 /* NamespaceExport */
                || kind === 162 /* Parameter */
                || kind === 291 /* PropertyAssignment */
                || kind === 165 /* PropertyDeclaration */
                || kind === 164 /* PropertySignature */
                || kind === 171 /* SetAccessor */
                || kind === 292 /* ShorthandPropertyAssignment */
                || kind === 257 /* TypeAliasDeclaration */
                || kind === 161 /* TypeParameter */
                || kind === 252 /* VariableDeclaration */
                || kind === 340 /* JSDocTypedefTag */
                || kind === 333 /* JSDocCallbackTag */
                || kind === 342 /* JSDocPropertyTag */;
        }
        function isDeclarationStatementKind(kind) {
            return kind === 254 /* FunctionDeclaration */
                || kind === 274 /* MissingDeclaration */
                || kind === 255 /* ClassDeclaration */
                || kind === 256 /* InterfaceDeclaration */
                || kind === 257 /* TypeAliasDeclaration */
                || kind === 258 /* EnumDeclaration */
                || kind === 259 /* ModuleDeclaration */
                || kind === 264 /* ImportDeclaration */
                || kind === 263 /* ImportEqualsDeclaration */
                || kind === 270 /* ExportDeclaration */
                || kind === 269 /* ExportAssignment */
                || kind === 262 /* NamespaceExportDeclaration */;
        }
        function isStatementKindButNotDeclarationKind(kind) {
            return kind === 244 /* BreakStatement */
                || kind === 243 /* ContinueStatement */
                || kind === 251 /* DebuggerStatement */
                || kind === 238 /* DoStatement */
                || kind === 236 /* ExpressionStatement */
                || kind === 234 /* EmptyStatement */
                || kind === 241 /* ForInStatement */
                || kind === 242 /* ForOfStatement */
                || kind === 240 /* ForStatement */
                || kind === 237 /* IfStatement */
                || kind === 248 /* LabeledStatement */
                || kind === 245 /* ReturnStatement */
                || kind === 247 /* SwitchStatement */
                || kind === 249 /* ThrowStatement */
                || kind === 250 /* TryStatement */
                || kind === 235 /* VariableStatement */
                || kind === 239 /* WhileStatement */
                || kind === 246 /* WithStatement */
                || kind === 344 /* NotEmittedStatement */
                || kind === 348 /* EndOfDeclarationMarker */
                || kind === 347 /* MergeDeclarationMarker */;
        }
        /* @internal */
        function isDeclaration(node) {
            if (node.kind === 161 /* TypeParameter */) {
                return (node.parent && node.parent.kind !== 339 /* JSDocTemplateTag */) || ts.isInJSFile(node);
            }
            return isDeclarationKind(node.kind);
        }
        ts.isDeclaration = isDeclaration;
        /* @internal */
        function isDeclarationStatement(node) {
            return isDeclarationStatementKind(node.kind);
        }
        ts.isDeclarationStatement = isDeclarationStatement;
        /**
         * Determines whether the node is a statement that is not also a declaration
         */
        /* @internal */
        function isStatementButNotDeclaration(node) {
            return isStatementKindButNotDeclarationKind(node.kind);
        }
        ts.isStatementButNotDeclaration = isStatementButNotDeclaration;
        /* @internal */
        function isStatement(node) {
            var kind = node.kind;
            return isStatementKindButNotDeclarationKind(kind)
                || isDeclarationStatementKind(kind)
                || isBlockStatement(node);
        }
        ts.isStatement = isStatement;
        function isBlockStatement(node) {
            if (node.kind !== 233 /* Block */)
                return false;
            if (node.parent !== undefined) {
                if (node.parent.kind === 250 /* TryStatement */ || node.parent.kind === 290 /* CatchClause */) {
                    return false;
                }
            }
            return !ts.isFunctionBlock(node);
        }
        /**
         * NOTE: This is similar to `isStatement` but does not access parent pointers.
         */
        /* @internal */
        function isStatementOrBlock(node) {
            var kind = node.kind;
            return isStatementKindButNotDeclarationKind(kind)
                || isDeclarationStatementKind(kind)
                || kind === 233 /* Block */;
        }
        ts.isStatementOrBlock = isStatementOrBlock;
        // Module references
        /* @internal */
        function isModuleReference(node) {
            var kind = node.kind;
            return kind === 275 /* ExternalModuleReference */
                || kind === 159 /* QualifiedName */
                || kind === 79 /* Identifier */;
        }
        ts.isModuleReference = isModuleReference;
        // JSX
        /* @internal */
        function isJsxTagNameExpression(node) {
            var kind = node.kind;
            return kind === 108 /* ThisKeyword */
                || kind === 79 /* Identifier */
                || kind === 204 /* PropertyAccessExpression */;
        }
        ts.isJsxTagNameExpression = isJsxTagNameExpression;
        /* @internal */
        function isJsxChild(node) {
            var kind = node.kind;
            return kind === 276 /* JsxElement */
                || kind === 286 /* JsxExpression */
                || kind === 277 /* JsxSelfClosingElement */
                || kind === 11 /* JsxText */
                || kind === 280 /* JsxFragment */;
        }
        ts.isJsxChild = isJsxChild;
        /* @internal */
        function isJsxAttributeLike(node) {
            var kind = node.kind;
            return kind === 283 /* JsxAttribute */
                || kind === 285 /* JsxSpreadAttribute */;
        }
        ts.isJsxAttributeLike = isJsxAttributeLike;
        /* @internal */
        function isStringLiteralOrJsxExpression(node) {
            var kind = node.kind;
            return kind === 10 /* StringLiteral */
                || kind === 286 /* JsxExpression */;
        }
        ts.isStringLiteralOrJsxExpression = isStringLiteralOrJsxExpression;
        function isJsxOpeningLikeElement(node) {
            var kind = node.kind;
            return kind === 278 /* JsxOpeningElement */
                || kind === 277 /* JsxSelfClosingElement */;
        }
        ts.isJsxOpeningLikeElement = isJsxOpeningLikeElement;
        // Clauses
        function isCaseOrDefaultClause(node) {
            var kind = node.kind;
            return kind === 287 /* CaseClause */
                || kind === 288 /* DefaultClause */;
        }
        ts.isCaseOrDefaultClause = isCaseOrDefaultClause;
        // JSDoc
        /** True if node is of some JSDoc syntax kind. */
        /* @internal */
        function isJSDocNode(node) {
            return node.kind >= 304 /* FirstJSDocNode */ && node.kind <= 342 /* LastJSDocNode */;
        }
        ts.isJSDocNode = isJSDocNode;
        /** True if node is of a kind that may contain comment text. */
        function isJSDocCommentContainingNode(node) {
            return node.kind === 315 /* JSDocComment */
                || node.kind === 314 /* JSDocNamepathType */
                || node.kind === 316 /* JSDocText */
                || isJSDocLinkLike(node)
                || isJSDocTag(node)
                || ts.isJSDocTypeLiteral(node)
                || ts.isJSDocSignature(node);
        }
        ts.isJSDocCommentContainingNode = isJSDocCommentContainingNode;
        // TODO: determine what this does before making it public.
        /* @internal */
        function isJSDocTag(node) {
            return node.kind >= 322 /* FirstJSDocTagNode */ && node.kind <= 342 /* LastJSDocTagNode */;
        }
        ts.isJSDocTag = isJSDocTag;
        function isSetAccessor(node) {
            return node.kind === 171 /* SetAccessor */;
        }
        ts.isSetAccessor = isSetAccessor;
        function isGetAccessor(node) {
            return node.kind === 170 /* GetAccessor */;
        }
        ts.isGetAccessor = isGetAccessor;
        /** True if has jsdoc nodes attached to it. */
        /* @internal */
        // TODO: GH#19856 Would like to return `node is Node & { jsDoc: JSDoc[] }` but it causes long compile times
        function hasJSDocNodes(node) {
            var jsDoc = node.jsDoc;
            return !!jsDoc && jsDoc.length > 0;
        }
        ts.hasJSDocNodes = hasJSDocNodes;
        /** True if has type node attached to it. */
        /* @internal */
        function hasType(node) {
            return !!node.type;
        }
        ts.hasType = hasType;
        /** True if has initializer node attached to it. */
        /* @internal */
        function hasInitializer(node) {
            return !!node.initializer;
        }
        ts.hasInitializer = hasInitializer;
        /** True if has initializer node attached to it. */
        function hasOnlyExpressionInitializer(node) {
            switch (node.kind) {
                case 252 /* VariableDeclaration */:
                case 162 /* Parameter */:
                case 201 /* BindingElement */:
                case 164 /* PropertySignature */:
                case 165 /* PropertyDeclaration */:
                case 291 /* PropertyAssignment */:
                case 294 /* EnumMember */:
                    return true;
                default:
                    return false;
            }
        }
        ts.hasOnlyExpressionInitializer = hasOnlyExpressionInitializer;
        function isObjectLiteralElement(node) {
            return node.kind === 283 /* JsxAttribute */ || node.kind === 285 /* JsxSpreadAttribute */ || isObjectLiteralElementLike(node);
        }
        ts.isObjectLiteralElement = isObjectLiteralElement;
        /* @internal */
        function isTypeReferenceType(node) {
            return node.kind === 176 /* TypeReference */ || node.kind === 226 /* ExpressionWithTypeArguments */;
        }
        ts.isTypeReferenceType = isTypeReferenceType;
        var MAX_SMI_X86 = 1073741823;
        /* @internal */
        function guessIndentation(lines) {
            var indentation = MAX_SMI_X86;
            for (var _i = 0, lines_1 = lines; _i < lines_1.length; _i++) {
                var line = lines_1[_i];
                if (!line.length) {
                    continue;
                }
                var i = 0;
                for (; i < line.length && i < indentation; i++) {
                    if (!ts.isWhiteSpaceLike(line.charCodeAt(i))) {
                        break;
                    }
                }
                if (i < indentation) {
                    indentation = i;
                }
                if (indentation === 0) {
                    return 0;
                }
            }
            return indentation === MAX_SMI_X86 ? undefined : indentation;
        }
        ts.guessIndentation = guessIndentation;
        function isStringLiteralLike(node) {
            return node.kind === 10 /* StringLiteral */ || node.kind === 14 /* NoSubstitutionTemplateLiteral */;
        }
        ts.isStringLiteralLike = isStringLiteralLike;
        function isJSDocLinkLike(node) {
            return node.kind === 319 /* JSDocLink */ || node.kind === 320 /* JSDocLinkCode */ || node.kind === 321 /* JSDocLinkPlain */;
        }
        ts.isJSDocLinkLike = isJSDocLinkLike;
        // #endregion
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        ts.resolvingEmptyArray = [];
        ts.externalHelpersModuleNameText = "tslib";
        ts.defaultMaximumTruncationLength = 160;
        ts.noTruncationMaximumTruncationLength = 1000000;
        function getDeclarationOfKind(symbol, kind) {
            var declarations = symbol.declarations;
            if (declarations) {
                for (var _i = 0, declarations_1 = declarations; _i < declarations_1.length; _i++) {
                    var declaration = declarations_1[_i];
                    if (declaration.kind === kind) {
                        return declaration;
                    }
                }
            }
            return undefined;
        }
        ts.getDeclarationOfKind = getDeclarationOfKind;
        /**
         * Create a new escaped identifier map.
         * @deprecated Use `new Map<__String, T>()` instead.
         */
        function createUnderscoreEscapedMap() {
            return new ts.Map();
        }
        ts.createUnderscoreEscapedMap = createUnderscoreEscapedMap;
        /**
         * @deprecated Use `!!map?.size` instead
         */
        function hasEntries(map) {
            return !!map && !!map.size;
        }
        ts.hasEntries = hasEntries;
        function createSymbolTable(symbols) {
            var result = new ts.Map();
            if (symbols) {
                for (var _i = 0, symbols_1 = symbols; _i < symbols_1.length; _i++) {
                    var symbol = symbols_1[_i];
                    result.set(symbol.escapedName, symbol);
                }
            }
            return result;
        }
        ts.createSymbolTable = createSymbolTable;
        function isTransientSymbol(symbol) {
            return (symbol.flags & 33554432 /* Transient */) !== 0;
        }
        ts.isTransientSymbol = isTransientSymbol;
        var stringWriter = createSingleLineStringWriter();
        function createSingleLineStringWriter() {
            var str = "";
            var writeText = function (text) { return str += text; };
            return {
                getText: function () { return str; },
                write: writeText,
                rawWrite: writeText,
                writeKeyword: writeText,
                writeOperator: writeText,
                writePunctuation: writeText,
                writeSpace: writeText,
                writeStringLiteral: writeText,
                writeLiteral: writeText,
                writeParameter: writeText,
                writeProperty: writeText,
                writeSymbol: function (s, _) { return writeText(s); },
                writeTrailingSemicolon: writeText,
                writeComment: writeText,
                getTextPos: function () { return str.length; },
                getLine: function () { return 0; },
                getColumn: function () { return 0; },
                getIndent: function () { return 0; },
                isAtStartOfLine: function () { return false; },
                hasTrailingComment: function () { return false; },
                hasTrailingWhitespace: function () { return !!str.length && ts.isWhiteSpaceLike(str.charCodeAt(str.length - 1)); },
                // Completely ignore indentation for string writers.  And map newlines to
                // a single space.
                writeLine: function () { return str += " "; },
                increaseIndent: ts.noop,
                decreaseIndent: ts.noop,
                clear: function () { return str = ""; },
                trackSymbol: function () { return false; },
                reportInaccessibleThisError: ts.noop,
                reportInaccessibleUniqueSymbolError: ts.noop,
                reportPrivateInBaseOfClassExpression: ts.noop,
            };
        }
        function changesAffectModuleResolution(oldOptions, newOptions) {
            return oldOptions.configFilePath !== newOptions.configFilePath ||
                optionsHaveModuleResolutionChanges(oldOptions, newOptions);
        }
        ts.changesAffectModuleResolution = changesAffectModuleResolution;
        function optionsHaveModuleResolutionChanges(oldOptions, newOptions) {
            return optionsHaveChanges(oldOptions, newOptions, ts.moduleResolutionOptionDeclarations);
        }
        ts.optionsHaveModuleResolutionChanges = optionsHaveModuleResolutionChanges;
        function changesAffectingProgramStructure(oldOptions, newOptions) {
            return optionsHaveChanges(oldOptions, newOptions, ts.optionsAffectingProgramStructure);
        }
        ts.changesAffectingProgramStructure = changesAffectingProgramStructure;
        function optionsHaveChanges(oldOptions, newOptions, optionDeclarations) {
            return oldOptions !== newOptions && optionDeclarations.some(function (o) {
                return !isJsonEqual(getCompilerOptionValue(oldOptions, o), getCompilerOptionValue(newOptions, o));
            });
        }
        ts.optionsHaveChanges = optionsHaveChanges;
        function forEachAncestor(node, callback) {
            while (true) {
                var res = callback(node);
                if (res === "quit")
                    return undefined;
                if (res !== undefined)
                    return res;
                if (ts.isSourceFile(node))
                    return undefined;
                node = node.parent;
            }
        }
        ts.forEachAncestor = forEachAncestor;
        /**
         * Calls `callback` for each entry in the map, returning the first truthy result.
         * Use `map.forEach` instead for normal iteration.
         */
        function forEachEntry(map, callback) {
            var iterator = map.entries();
            for (var iterResult = iterator.next(); !iterResult.done; iterResult = iterator.next()) {
                var _a = iterResult.value, key = _a[0], value = _a[1];
                var result = callback(value, key);
                if (result) {
                    return result;
                }
            }
            return undefined;
        }
        ts.forEachEntry = forEachEntry;
        /** `forEachEntry` for just keys. */
        function forEachKey(map, callback) {
            var iterator = map.keys();
            for (var iterResult = iterator.next(); !iterResult.done; iterResult = iterator.next()) {
                var result = callback(iterResult.value);
                if (result) {
                    return result;
                }
            }
            return undefined;
        }
        ts.forEachKey = forEachKey;
        /** Copy entries from `source` to `target`. */
        function copyEntries(source, target) {
            source.forEach(function (value, key) {
                target.set(key, value);
            });
        }
        ts.copyEntries = copyEntries;
        function usingSingleLineStringWriter(action) {
            var oldString = stringWriter.getText();
            try {
                action(stringWriter);
                return stringWriter.getText();
            }
            finally {
                stringWriter.clear();
                stringWriter.writeKeyword(oldString);
            }
        }
        ts.usingSingleLineStringWriter = usingSingleLineStringWriter;
        function getFullWidth(node) {
            return node.end - node.pos;
        }
        ts.getFullWidth = getFullWidth;
        function getResolvedModule(sourceFile, moduleNameText) {
            return sourceFile && sourceFile.resolvedModules && sourceFile.resolvedModules.get(moduleNameText);
        }
        ts.getResolvedModule = getResolvedModule;
        function setResolvedModule(sourceFile, moduleNameText, resolvedModule) {
            if (!sourceFile.resolvedModules) {
                sourceFile.resolvedModules = new ts.Map();
            }
            sourceFile.resolvedModules.set(moduleNameText, resolvedModule);
        }
        ts.setResolvedModule = setResolvedModule;
        function setResolvedTypeReferenceDirective(sourceFile, typeReferenceDirectiveName, resolvedTypeReferenceDirective) {
            if (!sourceFile.resolvedTypeReferenceDirectiveNames) {
                sourceFile.resolvedTypeReferenceDirectiveNames = new ts.Map();
            }
            sourceFile.resolvedTypeReferenceDirectiveNames.set(typeReferenceDirectiveName, resolvedTypeReferenceDirective);
        }
        ts.setResolvedTypeReferenceDirective = setResolvedTypeReferenceDirective;
        function projectReferenceIsEqualTo(oldRef, newRef) {
            return oldRef.path === newRef.path &&
                !oldRef.prepend === !newRef.prepend &&
                !oldRef.circular === !newRef.circular;
        }
        ts.projectReferenceIsEqualTo = projectReferenceIsEqualTo;
        function moduleResolutionIsEqualTo(oldResolution, newResolution) {
            return oldResolution.isExternalLibraryImport === newResolution.isExternalLibraryImport &&
                oldResolution.extension === newResolution.extension &&
                oldResolution.resolvedFileName === newResolution.resolvedFileName &&
                oldResolution.originalPath === newResolution.originalPath &&
                packageIdIsEqual(oldResolution.packageId, newResolution.packageId);
        }
        ts.moduleResolutionIsEqualTo = moduleResolutionIsEqualTo;
        function packageIdIsEqual(a, b) {
            return a === b || !!a && !!b && a.name === b.name && a.subModuleName === b.subModuleName && a.version === b.version;
        }
        function packageIdToString(_a) {
            var name = _a.name, subModuleName = _a.subModuleName, version = _a.version;
            var fullName = subModuleName ? name + "/" + subModuleName : name;
            return fullName + "@" + version;
        }
        ts.packageIdToString = packageIdToString;
        function typeDirectiveIsEqualTo(oldResolution, newResolution) {
            return oldResolution.resolvedFileName === newResolution.resolvedFileName
                && oldResolution.primary === newResolution.primary
                && oldResolution.originalPath === newResolution.originalPath;
        }
        ts.typeDirectiveIsEqualTo = typeDirectiveIsEqualTo;
        function hasChangesInResolutions(names, newResolutions, oldResolutions, comparer) {
            ts.Debug.assert(names.length === newResolutions.length);
            for (var i = 0; i < names.length; i++) {
                var newResolution = newResolutions[i];
                var oldResolution = oldResolutions && oldResolutions.get(names[i]);
                var changed = oldResolution
                    ? !newResolution || !comparer(oldResolution, newResolution)
                    : newResolution;
                if (changed) {
                    return true;
                }
            }
            return false;
        }
        ts.hasChangesInResolutions = hasChangesInResolutions;
        // Returns true if this node contains a parse error anywhere underneath it.
        function containsParseError(node) {
            aggregateChildData(node);
            return (node.flags & 262144 /* ThisNodeOrAnySubNodesHasError */) !== 0;
        }
        ts.containsParseError = containsParseError;
        function aggregateChildData(node) {
            if (!(node.flags & 524288 /* HasAggregatedChildData */)) {
                // A node is considered to contain a parse error if:
                //  a) the parser explicitly marked that it had an error
                //  b) any of it's children reported that it had an error.
                var thisNodeOrAnySubNodesHasError = ((node.flags & 65536 /* ThisNodeHasError */) !== 0) ||
                    ts.forEachChild(node, containsParseError);
                // If so, mark ourselves accordingly.
                if (thisNodeOrAnySubNodesHasError) {
                    node.flags |= 262144 /* ThisNodeOrAnySubNodesHasError */;
                }
                // Also mark that we've propagated the child information to this node.  This way we can
                // always consult the bit directly on this node without needing to check its children
                // again.
                node.flags |= 524288 /* HasAggregatedChildData */;
            }
        }
        function getSourceFileOfNode(node) {
            while (node && node.kind !== 300 /* SourceFile */) {
                node = node.parent;
            }
            return node;
        }
        ts.getSourceFileOfNode = getSourceFileOfNode;
        function getSourceFileOfModule(module) {
            return getSourceFileOfNode(module.valueDeclaration || getNonAugmentationDeclaration(module));
        }
        ts.getSourceFileOfModule = getSourceFileOfModule;
        function isStatementWithLocals(node) {
            switch (node.kind) {
                case 233 /* Block */:
                case 261 /* CaseBlock */:
                case 240 /* ForStatement */:
                case 241 /* ForInStatement */:
                case 242 /* ForOfStatement */:
                    return true;
            }
            return false;
        }
        ts.isStatementWithLocals = isStatementWithLocals;
        function getStartPositionOfLine(line, sourceFile) {
            ts.Debug.assert(line >= 0);
            return ts.getLineStarts(sourceFile)[line];
        }
        ts.getStartPositionOfLine = getStartPositionOfLine;
        // This is a useful function for debugging purposes.
        function nodePosToString(node) {
            var file = getSourceFileOfNode(node);
            var loc = ts.getLineAndCharacterOfPosition(file, node.pos);
            return file.fileName + "(" + (loc.line + 1) + "," + (loc.character + 1) + ")";
        }
        ts.nodePosToString = nodePosToString;
        function getEndLinePosition(line, sourceFile) {
            ts.Debug.assert(line >= 0);
            var lineStarts = ts.getLineStarts(sourceFile);
            var lineIndex = line;
            var sourceText = sourceFile.text;
            if (lineIndex + 1 === lineStarts.length) {
                // last line - return EOF
                return sourceText.length - 1;
            }
            else {
                // current line start
                var start = lineStarts[lineIndex];
                // take the start position of the next line - 1 = it should be some line break
                var pos = lineStarts[lineIndex + 1] - 1;
                ts.Debug.assert(ts.isLineBreak(sourceText.charCodeAt(pos)));
                // walk backwards skipping line breaks, stop the the beginning of current line.
                // i.e:
                // <some text>
                // $ <- end of line for this position should match the start position
                while (start <= pos && ts.isLineBreak(sourceText.charCodeAt(pos))) {
                    pos--;
                }
                return pos;
            }
        }
        ts.getEndLinePosition = getEndLinePosition;
        /**
         * Returns a value indicating whether a name is unique globally or within the current file.
         * Note: This does not consider whether a name appears as a free identifier or not, so at the expression `x.y` this includes both `x` and `y`.
         */
        function isFileLevelUniqueName(sourceFile, name, hasGlobalName) {
            return !(hasGlobalName && hasGlobalName(name)) && !sourceFile.identifiers.has(name);
        }
        ts.isFileLevelUniqueName = isFileLevelUniqueName;
        // Returns true if this node is missing from the actual source code. A 'missing' node is different
        // from 'undefined/defined'. When a node is undefined (which can happen for optional nodes
        // in the tree), it is definitely missing. However, a node may be defined, but still be
        // missing.  This happens whenever the parser knows it needs to parse something, but can't
        // get anything in the source code that it expects at that location. For example:
        //
        //          let a: ;
        //
        // Here, the Type in the Type-Annotation is not-optional (as there is a colon in the source
        // code). So the parser will attempt to parse out a type, and will create an actual node.
        // However, this node will be 'missing' in the sense that no actual source-code/tokens are
        // contained within it.
        function nodeIsMissing(node) {
            if (node === undefined) {
                return true;
            }
            return node.pos === node.end && node.pos >= 0 && node.kind !== 1 /* EndOfFileToken */;
        }
        ts.nodeIsMissing = nodeIsMissing;
        function nodeIsPresent(node) {
            return !nodeIsMissing(node);
        }
        ts.nodeIsPresent = nodeIsPresent;
        function insertStatementsAfterPrologue(to, from, isPrologueDirective) {
            if (from === undefined || from.length === 0)
                return to;
            var statementIndex = 0;
            // skip all prologue directives to insert at the correct position
            for (; statementIndex < to.length; ++statementIndex) {
                if (!isPrologueDirective(to[statementIndex])) {
                    break;
                }
            }
            to.splice.apply(to, __spreadArray([statementIndex, 0], from, false));
            return to;
        }
        function insertStatementAfterPrologue(to, statement, isPrologueDirective) {
            if (statement === undefined)
                return to;
            var statementIndex = 0;
            // skip all prologue directives to insert at the correct position
            for (; statementIndex < to.length; ++statementIndex) {
                if (!isPrologueDirective(to[statementIndex])) {
                    break;
                }
            }
            to.splice(statementIndex, 0, statement);
            return to;
        }
        function isAnyPrologueDirective(node) {
            return isPrologueDirective(node) || !!(getEmitFlags(node) & 1048576 /* CustomPrologue */);
        }
        /**
         * Prepends statements to an array while taking care of prologue directives.
         */
        function insertStatementsAfterStandardPrologue(to, from) {
            return insertStatementsAfterPrologue(to, from, isPrologueDirective);
        }
        ts.insertStatementsAfterStandardPrologue = insertStatementsAfterStandardPrologue;
        function insertStatementsAfterCustomPrologue(to, from) {
            return insertStatementsAfterPrologue(to, from, isAnyPrologueDirective);
        }
        ts.insertStatementsAfterCustomPrologue = insertStatementsAfterCustomPrologue;
        /**
         * Prepends statements to an array while taking care of prologue directives.
         */
        function insertStatementAfterStandardPrologue(to, statement) {
            return insertStatementAfterPrologue(to, statement, isPrologueDirective);
        }
        ts.insertStatementAfterStandardPrologue = insertStatementAfterStandardPrologue;
        function insertStatementAfterCustomPrologue(to, statement) {
            return insertStatementAfterPrologue(to, statement, isAnyPrologueDirective);
        }
        ts.insertStatementAfterCustomPrologue = insertStatementAfterCustomPrologue;
        /**
         * Determine if the given comment is a triple-slash
         *
         * @return true if the comment is a triple-slash comment else false
         */
        function isRecognizedTripleSlashComment(text, commentPos, commentEnd) {
            // Verify this is /// comment, but do the regexp match only when we first can find /// in the comment text
            // so that we don't end up computing comment string and doing match for all // comments
            if (text.charCodeAt(commentPos + 1) === 47 /* slash */ &&
                commentPos + 2 < commentEnd &&
                text.charCodeAt(commentPos + 2) === 47 /* slash */) {
                var textSubStr = text.substring(commentPos, commentEnd);
                return ts.fullTripleSlashReferencePathRegEx.test(textSubStr) ||
                    ts.fullTripleSlashAMDReferencePathRegEx.test(textSubStr) ||
                    fullTripleSlashReferenceTypeReferenceDirectiveRegEx.test(textSubStr) ||
                    defaultLibReferenceRegEx.test(textSubStr) ?
                    true : false;
            }
            return false;
        }
        ts.isRecognizedTripleSlashComment = isRecognizedTripleSlashComment;
        function isPinnedComment(text, start) {
            return text.charCodeAt(start + 1) === 42 /* asterisk */ &&
                text.charCodeAt(start + 2) === 33 /* exclamation */;
        }
        ts.isPinnedComment = isPinnedComment;
        function createCommentDirectivesMap(sourceFile, commentDirectives) {
            var directivesByLine = new ts.Map(commentDirectives.map(function (commentDirective) {
                return ([
                    "" + ts.getLineAndCharacterOfPosition(sourceFile, commentDirective.range.end).line,
                    commentDirective,
                ]);
            }));
            var usedLines = new ts.Map();
            return { getUnusedExpectations: getUnusedExpectations, markUsed: markUsed };
            function getUnusedExpectations() {
                return ts.arrayFrom(directivesByLine.entries())
                    .filter(function (_a) {
                    var line = _a[0], directive = _a[1];
                    return directive.type === 0 /* ExpectError */ && !usedLines.get(line);
                })
                    .map(function (_a) {
                    var _ = _a[0], directive = _a[1];
                    return directive;
                });
            }
            function markUsed(line) {
                if (!directivesByLine.has("" + line)) {
                    return false;
                }
                usedLines.set("" + line, true);
                return true;
            }
        }
        ts.createCommentDirectivesMap = createCommentDirectivesMap;
        function getTokenPosOfNode(node, sourceFile, includeJsDoc) {
            // With nodes that have no width (i.e. 'Missing' nodes), we actually *don't*
            // want to skip trivia because this will launch us forward to the next token.
            if (nodeIsMissing(node)) {
                return node.pos;
            }
            if (ts.isJSDocNode(node) || node.kind === 11 /* JsxText */) {
                // JsxText cannot actually contain comments, even though the scanner will think it sees comments
                return ts.skipTrivia((sourceFile || getSourceFileOfNode(node)).text, node.pos, /*stopAfterLineBreak*/ false, /*stopAtComments*/ true);
            }
            if (includeJsDoc && ts.hasJSDocNodes(node)) {
                return getTokenPosOfNode(node.jsDoc[0], sourceFile);
            }
            // For a syntax list, it is possible that one of its children has JSDocComment nodes, while
            // the syntax list itself considers them as normal trivia. Therefore if we simply skip
            // trivia for the list, we may have skipped the JSDocComment as well. So we should process its
            // first child to determine the actual position of its first token.
            if (node.kind === 343 /* SyntaxList */ && node._children.length > 0) {
                return getTokenPosOfNode(node._children[0], sourceFile, includeJsDoc);
            }
            return ts.skipTrivia((sourceFile || getSourceFileOfNode(node)).text, node.pos, 
            /*stopAfterLineBreak*/ false, 
            /*stopAtComments*/ false, isInJSDoc(node));
        }
        ts.getTokenPosOfNode = getTokenPosOfNode;
        function getNonDecoratorTokenPosOfNode(node, sourceFile) {
            if (nodeIsMissing(node) || !node.decorators) {
                return getTokenPosOfNode(node, sourceFile);
            }
            return ts.skipTrivia((sourceFile || getSourceFileOfNode(node)).text, node.decorators.end);
        }
        ts.getNonDecoratorTokenPosOfNode = getNonDecoratorTokenPosOfNode;
        function getSourceTextOfNodeFromSourceFile(sourceFile, node, includeTrivia) {
            if (includeTrivia === void 0) {
                includeTrivia = false;
            }
            return getTextOfNodeFromSourceText(sourceFile.text, node, includeTrivia);
        }
        ts.getSourceTextOfNodeFromSourceFile = getSourceTextOfNodeFromSourceFile;
        function isJSDocTypeExpressionOrChild(node) {
            return !!ts.findAncestor(node, ts.isJSDocTypeExpression);
        }
        function isExportNamespaceAsDefaultDeclaration(node) {
            return !!(ts.isExportDeclaration(node) && node.exportClause && ts.isNamespaceExport(node.exportClause) && node.exportClause.name.escapedText === "default");
        }
        ts.isExportNamespaceAsDefaultDeclaration = isExportNamespaceAsDefaultDeclaration;
        function getTextOfNodeFromSourceText(sourceText, node, includeTrivia) {
            if (includeTrivia === void 0) {
                includeTrivia = false;
            }
            if (nodeIsMissing(node)) {
                return "";
            }
            var text = sourceText.substring(includeTrivia ? node.pos : ts.skipTrivia(sourceText, node.pos), node.end);
            if (isJSDocTypeExpressionOrChild(node)) {
                // strip space + asterisk at line start
                text = text.split(/\r\n|\n|\r/).map(function (line) { return ts.trimStringStart(line.replace(/^\s*\*/, "")); }).join("\n");
            }
            return text;
        }
        ts.getTextOfNodeFromSourceText = getTextOfNodeFromSourceText;
        function getTextOfNode(node, includeTrivia) {
            if (includeTrivia === void 0) {
                includeTrivia = false;
            }
            return getSourceTextOfNodeFromSourceFile(getSourceFileOfNode(node), node, includeTrivia);
        }
        ts.getTextOfNode = getTextOfNode;
        function getPos(range) {
            return range.pos;
        }
        /**
         * Note: it is expected that the `nodeArray` and the `node` are within the same file.
         * For example, searching for a `SourceFile` in a `SourceFile[]` wouldn't work.
         */
        function indexOfNode(nodeArray, node) {
            return ts.binarySearch(nodeArray, node, getPos, ts.compareValues);
        }
        ts.indexOfNode = indexOfNode;
        /**
         * Gets flags that control emit behavior of a node.
         */
        function getEmitFlags(node) {
            var emitNode = node.emitNode;
            return emitNode && emitNode.flags || 0;
        }
        ts.getEmitFlags = getEmitFlags;
        ;
        function getScriptTargetFeatures() {
            return {
                es2015: {
                    Array: ["find", "findIndex", "fill", "copyWithin", "entries", "keys", "values"],
                    RegExp: ["flags", "sticky", "unicode"],
                    Reflect: ["apply", "construct", "defineProperty", "deleteProperty", "get", " getOwnPropertyDescriptor", "getPrototypeOf", "has", "isExtensible", "ownKeys", "preventExtensions", "set", "setPrototypeOf"],
                    ArrayConstructor: ["from", "of"],
                    ObjectConstructor: ["assign", "getOwnPropertySymbols", "keys", "is", "setPrototypeOf"],
                    NumberConstructor: ["isFinite", "isInteger", "isNaN", "isSafeInteger", "parseFloat", "parseInt"],
                    Math: ["clz32", "imul", "sign", "log10", "log2", "log1p", "expm1", "cosh", "sinh", "tanh", "acosh", "asinh", "atanh", "hypot", "trunc", "fround", "cbrt"],
                    Map: ["entries", "keys", "values"],
                    Set: ["entries", "keys", "values"],
                    Promise: ts.emptyArray,
                    PromiseConstructor: ["all", "race", "reject", "resolve"],
                    Symbol: ["for", "keyFor"],
                    WeakMap: ["entries", "keys", "values"],
                    WeakSet: ["entries", "keys", "values"],
                    Iterator: ts.emptyArray,
                    AsyncIterator: ts.emptyArray,
                    String: ["codePointAt", "includes", "endsWith", "normalize", "repeat", "startsWith", "anchor", "big", "blink", "bold", "fixed", "fontcolor", "fontsize", "italics", "link", "small", "strike", "sub", "sup"],
                    StringConstructor: ["fromCodePoint", "raw"]
                },
                es2016: {
                    Array: ["includes"]
                },
                es2017: {
                    Atomics: ts.emptyArray,
                    SharedArrayBuffer: ts.emptyArray,
                    String: ["padStart", "padEnd"],
                    ObjectConstructor: ["values", "entries", "getOwnPropertyDescriptors"],
                    DateTimeFormat: ["formatToParts"]
                },
                es2018: {
                    Promise: ["finally"],
                    RegExpMatchArray: ["groups"],
                    RegExpExecArray: ["groups"],
                    RegExp: ["dotAll"],
                    Intl: ["PluralRules"],
                    AsyncIterable: ts.emptyArray,
                    AsyncIterableIterator: ts.emptyArray,
                    AsyncGenerator: ts.emptyArray,
                    AsyncGeneratorFunction: ts.emptyArray,
                },
                es2019: {
                    Array: ["flat", "flatMap"],
                    ObjectConstructor: ["fromEntries"],
                    String: ["trimStart", "trimEnd", "trimLeft", "trimRight"],
                    Symbol: ["description"]
                },
                es2020: {
                    BigInt: ts.emptyArray,
                    BigInt64Array: ts.emptyArray,
                    BigUint64Array: ts.emptyArray,
                    PromiseConstructor: ["allSettled"],
                    SymbolConstructor: ["matchAll"],
                    String: ["matchAll"],
                    DataView: ["setBigInt64", "setBigUint64", "getBigInt64", "getBigUint64"],
                    RelativeTimeFormat: ["format", "formatToParts", "resolvedOptions"]
                },
                es2021: {
                    PromiseConstructor: ["any"],
                    String: ["replaceAll"]
                },
                esnext: {
                    NumberFormat: ["formatToParts"]
                }
            };
        }
        ts.getScriptTargetFeatures = getScriptTargetFeatures;
        var GetLiteralTextFlags;
        (function (GetLiteralTextFlags) {
            GetLiteralTextFlags[GetLiteralTextFlags["None"] = 0] = "None";
            GetLiteralTextFlags[GetLiteralTextFlags["NeverAsciiEscape"] = 1] = "NeverAsciiEscape";
            GetLiteralTextFlags[GetLiteralTextFlags["JsxAttributeEscape"] = 2] = "JsxAttributeEscape";
            GetLiteralTextFlags[GetLiteralTextFlags["TerminateUnterminatedLiterals"] = 4] = "TerminateUnterminatedLiterals";
            GetLiteralTextFlags[GetLiteralTextFlags["AllowNumericSeparator"] = 8] = "AllowNumericSeparator";
        })(GetLiteralTextFlags = ts.GetLiteralTextFlags || (ts.GetLiteralTextFlags = {}));
        function getLiteralText(node, sourceFile, flags) {
            var _a;
            // If we don't need to downlevel and we can reach the original source text using
            // the node's parent reference, then simply get the text as it was originally written.
            if (canUseOriginalText(node, flags)) {
                return getSourceTextOfNodeFromSourceFile(sourceFile, node);
            }
            // If we can't reach the original source text, use the canonical form if it's a number,
            // or a (possibly escaped) quoted form of the original text if it's string-like.
            switch (node.kind) {
                case 10 /* StringLiteral */: {
                    var escapeText = flags & 2 /* JsxAttributeEscape */ ? escapeJsxAttributeString :
                        flags & 1 /* NeverAsciiEscape */ || (getEmitFlags(node) & 16777216 /* NoAsciiEscaping */) ? escapeString :
                            escapeNonAsciiString;
                    if (node.singleQuote) {
                        return "'" + escapeText(node.text, 39 /* singleQuote */) + "'";
                    }
                    else {
                        return '"' + escapeText(node.text, 34 /* doubleQuote */) + '"';
                    }
                }
                case 14 /* NoSubstitutionTemplateLiteral */:
                case 15 /* TemplateHead */:
                case 16 /* TemplateMiddle */:
                case 17 /* TemplateTail */: {
                    // If a NoSubstitutionTemplateLiteral appears to have a substitution in it, the original text
                    // had to include a backslash: `not \${a} substitution`.
                    var escapeText = flags & 1 /* NeverAsciiEscape */ || (getEmitFlags(node) & 16777216 /* NoAsciiEscaping */) ? escapeString :
                        escapeNonAsciiString;
                    var rawText = (_a = node.rawText) !== null && _a !== void 0 ? _a : escapeTemplateSubstitution(escapeText(node.text, 96 /* backtick */));
                    switch (node.kind) {
                        case 14 /* NoSubstitutionTemplateLiteral */:
                            return "`" + rawText + "`";
                        case 15 /* TemplateHead */:
                            return "`" + rawText + "${";
                        case 16 /* TemplateMiddle */:
                            return "}" + rawText + "${";
                        case 17 /* TemplateTail */:
                            return "}" + rawText + "`";
                    }
                    break;
                }
                case 8 /* NumericLiteral */:
                case 9 /* BigIntLiteral */:
                    return node.text;
                case 13 /* RegularExpressionLiteral */:
                    if (flags & 4 /* TerminateUnterminatedLiterals */ && node.isUnterminated) {
                        return node.text + (node.text.charCodeAt(node.text.length - 1) === 92 /* backslash */ ? " /" : "/");
                    }
                    return node.text;
            }
            return ts.Debug.fail("Literal kind '" + node.kind + "' not accounted for.");
        }
        ts.getLiteralText = getLiteralText;
        function canUseOriginalText(node, flags) {
            if (nodeIsSynthesized(node) || !node.parent || (flags & 4 /* TerminateUnterminatedLiterals */ && node.isUnterminated)) {
                return false;
            }
            if (ts.isNumericLiteral(node) && node.numericLiteralFlags & 512 /* ContainsSeparator */) {
                return !!(flags & 8 /* AllowNumericSeparator */);
            }
            return !ts.isBigIntLiteral(node);
        }
        function getTextOfConstantValue(value) {
            return ts.isString(value) ? '"' + escapeNonAsciiString(value) + '"' : "" + value;
        }
        ts.getTextOfConstantValue = getTextOfConstantValue;
        // Make an identifier from an external module name by extracting the string after the last "/" and replacing
        // all non-alphanumeric characters with underscores
        function makeIdentifierFromModuleName(moduleName) {
            return ts.getBaseFileName(moduleName).replace(/^(\d)/, "_$1").replace(/\W/g, "_");
        }
        ts.makeIdentifierFromModuleName = makeIdentifierFromModuleName;
        function isBlockOrCatchScoped(declaration) {
            return (ts.getCombinedNodeFlags(declaration) & 3 /* BlockScoped */) !== 0 ||
                isCatchClauseVariableDeclarationOrBindingElement(declaration);
        }
        ts.isBlockOrCatchScoped = isBlockOrCatchScoped;
        function isCatchClauseVariableDeclarationOrBindingElement(declaration) {
            var node = getRootDeclaration(declaration);
            return node.kind === 252 /* VariableDeclaration */ && node.parent.kind === 290 /* CatchClause */;
        }
        ts.isCatchClauseVariableDeclarationOrBindingElement = isCatchClauseVariableDeclarationOrBindingElement;
        function isAmbientModule(node) {
            return ts.isModuleDeclaration(node) && (node.name.kind === 10 /* StringLiteral */ || isGlobalScopeAugmentation(node));
        }
        ts.isAmbientModule = isAmbientModule;
        function isModuleWithStringLiteralName(node) {
            return ts.isModuleDeclaration(node) && node.name.kind === 10 /* StringLiteral */;
        }
        ts.isModuleWithStringLiteralName = isModuleWithStringLiteralName;
        function isNonGlobalAmbientModule(node) {
            return ts.isModuleDeclaration(node) && ts.isStringLiteral(node.name);
        }
        ts.isNonGlobalAmbientModule = isNonGlobalAmbientModule;
        /**
         * An effective module (namespace) declaration is either
         * 1. An actual declaration: namespace X { ... }
         * 2. A Javascript declaration, which is:
         *    An identifier in a nested property access expression: Y in `X.Y.Z = { ... }`
         */
        function isEffectiveModuleDeclaration(node) {
            return ts.isModuleDeclaration(node) || ts.isIdentifier(node);
        }
        ts.isEffectiveModuleDeclaration = isEffectiveModuleDeclaration;
        /** Given a symbol for a module, checks that it is a shorthand ambient module. */
        function isShorthandAmbientModuleSymbol(moduleSymbol) {
            return isShorthandAmbientModule(moduleSymbol.valueDeclaration);
        }
        ts.isShorthandAmbientModuleSymbol = isShorthandAmbientModuleSymbol;
        function isShorthandAmbientModule(node) {
            // The only kind of module that can be missing a body is a shorthand ambient module.
            return !!node && node.kind === 259 /* ModuleDeclaration */ && (!node.body);
        }
        function isBlockScopedContainerTopLevel(node) {
            return node.kind === 300 /* SourceFile */ ||
                node.kind === 259 /* ModuleDeclaration */ ||
                ts.isFunctionLikeOrClassStaticBlockDeclaration(node);
        }
        ts.isBlockScopedContainerTopLevel = isBlockScopedContainerTopLevel;
        function isGlobalScopeAugmentation(module) {
            return !!(module.flags & 1024 /* GlobalAugmentation */);
        }
        ts.isGlobalScopeAugmentation = isGlobalScopeAugmentation;
        function isExternalModuleAugmentation(node) {
            return isAmbientModule(node) && isModuleAugmentationExternal(node);
        }
        ts.isExternalModuleAugmentation = isExternalModuleAugmentation;
        function isModuleAugmentationExternal(node) {
            // external module augmentation is a ambient module declaration that is either:
            // - defined in the top level scope and source file is an external module
            // - defined inside ambient module declaration located in the top level scope and source file not an external module
            switch (node.parent.kind) {
                case 300 /* SourceFile */:
                    return ts.isExternalModule(node.parent);
                case 260 /* ModuleBlock */:
                    return isAmbientModule(node.parent.parent) && ts.isSourceFile(node.parent.parent.parent) && !ts.isExternalModule(node.parent.parent.parent);
            }
            return false;
        }
        ts.isModuleAugmentationExternal = isModuleAugmentationExternal;
        function getNonAugmentationDeclaration(symbol) {
            var _a;
            return (_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.find(function (d) { return !isExternalModuleAugmentation(d) && !(ts.isModuleDeclaration(d) && isGlobalScopeAugmentation(d)); });
        }
        ts.getNonAugmentationDeclaration = getNonAugmentationDeclaration;
        function isEffectiveExternalModule(node, compilerOptions) {
            return ts.isExternalModule(node) || compilerOptions.isolatedModules || ((getEmitModuleKind(compilerOptions) === ts.ModuleKind.CommonJS) && !!node.commonJsModuleIndicator);
        }
        ts.isEffectiveExternalModule = isEffectiveExternalModule;
        /**
         * Returns whether the source file will be treated as if it were in strict mode at runtime.
         */
        function isEffectiveStrictModeSourceFile(node, compilerOptions) {
            // We can only verify strict mode for JS/TS files
            switch (node.scriptKind) {
                case 1 /* JS */:
                case 3 /* TS */:
                case 2 /* JSX */:
                case 4 /* TSX */:
                    break;
                default:
                    return false;
            }
            // Strict mode does not matter for declaration files.
            if (node.isDeclarationFile) {
                return false;
            }
            // If `alwaysStrict` is set, then treat the file as strict.
            if (getStrictOptionValue(compilerOptions, "alwaysStrict")) {
                return true;
            }
            // Starting with a "use strict" directive indicates the file is strict.
            if (ts.startsWithUseStrict(node.statements)) {
                return true;
            }
            if (ts.isExternalModule(node) || compilerOptions.isolatedModules) {
                // ECMAScript Modules are always strict.
                if (getEmitModuleKind(compilerOptions) >= ts.ModuleKind.ES2015) {
                    return true;
                }
                // Other modules are strict unless otherwise specified.
                return !compilerOptions.noImplicitUseStrict;
            }
            return false;
        }
        ts.isEffectiveStrictModeSourceFile = isEffectiveStrictModeSourceFile;
        function isBlockScope(node, parentNode) {
            switch (node.kind) {
                case 300 /* SourceFile */:
                case 261 /* CaseBlock */:
                case 290 /* CatchClause */:
                case 259 /* ModuleDeclaration */:
                case 240 /* ForStatement */:
                case 241 /* ForInStatement */:
                case 242 /* ForOfStatement */:
                case 169 /* Constructor */:
                case 167 /* MethodDeclaration */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 254 /* FunctionDeclaration */:
                case 211 /* FunctionExpression */:
                case 212 /* ArrowFunction */:
                case 165 /* PropertyDeclaration */:
                case 168 /* ClassStaticBlockDeclaration */:
                    return true;
                case 233 /* Block */:
                    // function block is not considered block-scope container
                    // see comment in binder.ts: bind(...), case for SyntaxKind.Block
                    return !ts.isFunctionLikeOrClassStaticBlockDeclaration(parentNode);
            }
            return false;
        }
        ts.isBlockScope = isBlockScope;
        function isDeclarationWithTypeParameters(node) {
            switch (node.kind) {
                case 333 /* JSDocCallbackTag */:
                case 340 /* JSDocTypedefTag */:
                case 318 /* JSDocSignature */:
                    return true;
                default:
                    ts.assertType(node);
                    return isDeclarationWithTypeParameterChildren(node);
            }
        }
        ts.isDeclarationWithTypeParameters = isDeclarationWithTypeParameters;
        function isDeclarationWithTypeParameterChildren(node) {
            switch (node.kind) {
                case 172 /* CallSignature */:
                case 173 /* ConstructSignature */:
                case 166 /* MethodSignature */:
                case 174 /* IndexSignature */:
                case 177 /* FunctionType */:
                case 178 /* ConstructorType */:
                case 312 /* JSDocFunctionType */:
                case 255 /* ClassDeclaration */:
                case 224 /* ClassExpression */:
                case 256 /* InterfaceDeclaration */:
                case 257 /* TypeAliasDeclaration */:
                case 339 /* JSDocTemplateTag */:
                case 254 /* FunctionDeclaration */:
                case 167 /* MethodDeclaration */:
                case 169 /* Constructor */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 211 /* FunctionExpression */:
                case 212 /* ArrowFunction */:
                    return true;
                default:
                    ts.assertType(node);
                    return false;
            }
        }
        ts.isDeclarationWithTypeParameterChildren = isDeclarationWithTypeParameterChildren;
        function isAnyImportSyntax(node) {
            switch (node.kind) {
                case 264 /* ImportDeclaration */:
                case 263 /* ImportEqualsDeclaration */:
                    return true;
                default:
                    return false;
            }
        }
        ts.isAnyImportSyntax = isAnyImportSyntax;
        function isLateVisibilityPaintedStatement(node) {
            switch (node.kind) {
                case 264 /* ImportDeclaration */:
                case 263 /* ImportEqualsDeclaration */:
                case 235 /* VariableStatement */:
                case 255 /* ClassDeclaration */:
                case 254 /* FunctionDeclaration */:
                case 259 /* ModuleDeclaration */:
                case 257 /* TypeAliasDeclaration */:
                case 256 /* InterfaceDeclaration */:
                case 258 /* EnumDeclaration */:
                    return true;
                default:
                    return false;
            }
        }
        ts.isLateVisibilityPaintedStatement = isLateVisibilityPaintedStatement;
        function hasPossibleExternalModuleReference(node) {
            return isAnyImportOrReExport(node) || ts.isModuleDeclaration(node) || ts.isImportTypeNode(node) || isImportCall(node);
        }
        ts.hasPossibleExternalModuleReference = hasPossibleExternalModuleReference;
        function isAnyImportOrReExport(node) {
            return isAnyImportSyntax(node) || ts.isExportDeclaration(node);
        }
        ts.isAnyImportOrReExport = isAnyImportOrReExport;
        // Gets the nearest enclosing block scope container that has the provided node
        // as a descendant, that is not the provided node.
        function getEnclosingBlockScopeContainer(node) {
            return ts.findAncestor(node.parent, function (current) { return isBlockScope(current, current.parent); });
        }
        ts.getEnclosingBlockScopeContainer = getEnclosingBlockScopeContainer;
        function forEachEnclosingBlockScopeContainer(node, cb) {
            var container = getEnclosingBlockScopeContainer(node);
            while (container) {
                cb(container);
                container = getEnclosingBlockScopeContainer(container);
            }
        }
        ts.forEachEnclosingBlockScopeContainer = forEachEnclosingBlockScopeContainer;
        // Return display name of an identifier
        // Computed property names will just be emitted as "[<expr>]", where <expr> is the source
        // text of the expression in the computed property.
        function declarationNameToString(name) {
            return !name || getFullWidth(name) === 0 ? "(Missing)" : getTextOfNode(name);
        }
        ts.declarationNameToString = declarationNameToString;
        function getNameFromIndexInfo(info) {
            return info.declaration ? declarationNameToString(info.declaration.parameters[0].name) : undefined;
        }
        ts.getNameFromIndexInfo = getNameFromIndexInfo;
        function isComputedNonLiteralName(name) {
            return name.kind === 160 /* ComputedPropertyName */ && !isStringOrNumericLiteralLike(name.expression);
        }
        ts.isComputedNonLiteralName = isComputedNonLiteralName;
        function getTextOfPropertyName(name) {
            switch (name.kind) {
                case 79 /* Identifier */:
                case 80 /* PrivateIdentifier */:
                    return name.escapedText;
                case 10 /* StringLiteral */:
                case 8 /* NumericLiteral */:
                case 14 /* NoSubstitutionTemplateLiteral */:
                    return ts.escapeLeadingUnderscores(name.text);
                case 160 /* ComputedPropertyName */:
                    if (isStringOrNumericLiteralLike(name.expression))
                        return ts.escapeLeadingUnderscores(name.expression.text);
                    return ts.Debug.fail("Text of property name cannot be read from non-literal-valued ComputedPropertyNames");
                default:
                    return ts.Debug.assertNever(name);
            }
        }
        ts.getTextOfPropertyName = getTextOfPropertyName;
        function entityNameToString(name) {
            switch (name.kind) {
                case 108 /* ThisKeyword */:
                    return "this";
                case 80 /* PrivateIdentifier */:
                case 79 /* Identifier */:
                    return getFullWidth(name) === 0 ? ts.idText(name) : getTextOfNode(name);
                case 159 /* QualifiedName */:
                    return entityNameToString(name.left) + "." + entityNameToString(name.right);
                case 204 /* PropertyAccessExpression */:
                    if (ts.isIdentifier(name.name) || ts.isPrivateIdentifier(name.name)) {
                        return entityNameToString(name.expression) + "." + entityNameToString(name.name);
                    }
                    else {
                        return ts.Debug.assertNever(name.name);
                    }
                case 306 /* JSDocMemberName */:
                    return entityNameToString(name.left) + entityNameToString(name.right);
                default:
                    return ts.Debug.assertNever(name);
            }
        }
        ts.entityNameToString = entityNameToString;
        function createDiagnosticForNode(node, message, arg0, arg1, arg2, arg3) {
            var sourceFile = getSourceFileOfNode(node);
            return createDiagnosticForNodeInSourceFile(sourceFile, node, message, arg0, arg1, arg2, arg3);
        }
        ts.createDiagnosticForNode = createDiagnosticForNode;
        function createDiagnosticForNodeArray(sourceFile, nodes, message, arg0, arg1, arg2, arg3) {
            var start = ts.skipTrivia(sourceFile.text, nodes.pos);
            return createFileDiagnostic(sourceFile, start, nodes.end - start, message, arg0, arg1, arg2, arg3);
        }
        ts.createDiagnosticForNodeArray = createDiagnosticForNodeArray;
        function createDiagnosticForNodeInSourceFile(sourceFile, node, message, arg0, arg1, arg2, arg3) {
            var span = getErrorSpanForNode(sourceFile, node);
            return createFileDiagnostic(sourceFile, span.start, span.length, message, arg0, arg1, arg2, arg3);
        }
        ts.createDiagnosticForNodeInSourceFile = createDiagnosticForNodeInSourceFile;
        function createDiagnosticForNodeFromMessageChain(node, messageChain, relatedInformation) {
            var sourceFile = getSourceFileOfNode(node);
            var span = getErrorSpanForNode(sourceFile, node);
            return createFileDiagnosticFromMessageChain(sourceFile, span.start, span.length, messageChain, relatedInformation);
        }
        ts.createDiagnosticForNodeFromMessageChain = createDiagnosticForNodeFromMessageChain;
        function assertDiagnosticLocation(file, start, length) {
            ts.Debug.assertGreaterThanOrEqual(start, 0);
            ts.Debug.assertGreaterThanOrEqual(length, 0);
            if (file) {
                ts.Debug.assertLessThanOrEqual(start, file.text.length);
                ts.Debug.assertLessThanOrEqual(start + length, file.text.length);
            }
        }
        function createFileDiagnosticFromMessageChain(file, start, length, messageChain, relatedInformation) {
            assertDiagnosticLocation(file, start, length);
            return {
                file: file,
                start: start,
                length: length,
                code: messageChain.code,
                category: messageChain.category,
                messageText: messageChain.next ? messageChain : messageChain.messageText,
                relatedInformation: relatedInformation
            };
        }
        ts.createFileDiagnosticFromMessageChain = createFileDiagnosticFromMessageChain;
        function createDiagnosticForFileFromMessageChain(sourceFile, messageChain, relatedInformation) {
            return {
                file: sourceFile,
                start: 0,
                length: 0,
                code: messageChain.code,
                category: messageChain.category,
                messageText: messageChain.next ? messageChain : messageChain.messageText,
                relatedInformation: relatedInformation
            };
        }
        ts.createDiagnosticForFileFromMessageChain = createDiagnosticForFileFromMessageChain;
        function createDiagnosticForRange(sourceFile, range, message) {
            return {
                file: sourceFile,
                start: range.pos,
                length: range.end - range.pos,
                code: message.code,
                category: message.category,
                messageText: message.message,
            };
        }
        ts.createDiagnosticForRange = createDiagnosticForRange;
        function getSpanOfTokenAtPosition(sourceFile, pos) {
            var scanner = ts.createScanner(sourceFile.languageVersion, /*skipTrivia*/ true, sourceFile.languageVariant, sourceFile.text, /*onError:*/ undefined, pos);
            scanner.scan();
            var start = scanner.getTokenPos();
            return ts.createTextSpanFromBounds(start, scanner.getTextPos());
        }
        ts.getSpanOfTokenAtPosition = getSpanOfTokenAtPosition;
        function getErrorSpanForArrowFunction(sourceFile, node) {
            var pos = ts.skipTrivia(sourceFile.text, node.pos);
            if (node.body && node.body.kind === 233 /* Block */) {
                var startLine = ts.getLineAndCharacterOfPosition(sourceFile, node.body.pos).line;
                var endLine = ts.getLineAndCharacterOfPosition(sourceFile, node.body.end).line;
                if (startLine < endLine) {
                    // The arrow function spans multiple lines,
                    // make the error span be the first line, inclusive.
                    return ts.createTextSpan(pos, getEndLinePosition(startLine, sourceFile) - pos + 1);
                }
            }
            return ts.createTextSpanFromBounds(pos, node.end);
        }
        function getErrorSpanForNode(sourceFile, node) {
            var errorNode = node;
            switch (node.kind) {
                case 300 /* SourceFile */:
                    var pos_1 = ts.skipTrivia(sourceFile.text, 0, /*stopAfterLineBreak*/ false);
                    if (pos_1 === sourceFile.text.length) {
                        // file is empty - return span for the beginning of the file
                        return ts.createTextSpan(0, 0);
                    }
                    return getSpanOfTokenAtPosition(sourceFile, pos_1);
                // This list is a work in progress. Add missing node kinds to improve their error
                // spans.
                case 252 /* VariableDeclaration */:
                case 201 /* BindingElement */:
                case 255 /* ClassDeclaration */:
                case 224 /* ClassExpression */:
                case 256 /* InterfaceDeclaration */:
                case 259 /* ModuleDeclaration */:
                case 258 /* EnumDeclaration */:
                case 294 /* EnumMember */:
                case 254 /* FunctionDeclaration */:
                case 211 /* FunctionExpression */:
                case 167 /* MethodDeclaration */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 257 /* TypeAliasDeclaration */:
                case 165 /* PropertyDeclaration */:
                case 164 /* PropertySignature */:
                case 266 /* NamespaceImport */:
                    errorNode = node.name;
                    break;
                case 212 /* ArrowFunction */:
                    return getErrorSpanForArrowFunction(sourceFile, node);
                case 287 /* CaseClause */:
                case 288 /* DefaultClause */:
                    var start = ts.skipTrivia(sourceFile.text, node.pos);
                    var end = node.statements.length > 0 ? node.statements[0].pos : node.end;
                    return ts.createTextSpanFromBounds(start, end);
            }
            if (errorNode === undefined) {
                // If we don't have a better node, then just set the error on the first token of
                // construct.
                return getSpanOfTokenAtPosition(sourceFile, node.pos);
            }
            ts.Debug.assert(!ts.isJSDoc(errorNode));
            var isMissing = nodeIsMissing(errorNode);
            var pos = isMissing || ts.isJsxText(node)
                ? errorNode.pos
                : ts.skipTrivia(sourceFile.text, errorNode.pos);
            // These asserts should all be satisfied for a properly constructed `errorNode`.
            if (isMissing) {
                ts.Debug.assert(pos === errorNode.pos, "This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809");
                ts.Debug.assert(pos === errorNode.end, "This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809");
            }
            else {
                ts.Debug.assert(pos >= errorNode.pos, "This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809");
                ts.Debug.assert(pos <= errorNode.end, "This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809");
            }
            return ts.createTextSpanFromBounds(pos, errorNode.end);
        }
        ts.getErrorSpanForNode = getErrorSpanForNode;
        function isExternalOrCommonJsModule(file) {
            return (file.externalModuleIndicator || file.commonJsModuleIndicator) !== undefined;
        }
        ts.isExternalOrCommonJsModule = isExternalOrCommonJsModule;
        function isJsonSourceFile(file) {
            return file.scriptKind === 6 /* JSON */;
        }
        ts.isJsonSourceFile = isJsonSourceFile;
        function isEnumConst(node) {
            return !!(ts.getCombinedModifierFlags(node) & 2048 /* Const */);
        }
        ts.isEnumConst = isEnumConst;
        function isDeclarationReadonly(declaration) {
            return !!(ts.getCombinedModifierFlags(declaration) & 64 /* Readonly */ && !ts.isParameterPropertyDeclaration(declaration, declaration.parent));
        }
        ts.isDeclarationReadonly = isDeclarationReadonly;
        function isVarConst(node) {
            return !!(ts.getCombinedNodeFlags(node) & 2 /* Const */);
        }
        ts.isVarConst = isVarConst;
        function isLet(node) {
            return !!(ts.getCombinedNodeFlags(node) & 1 /* Let */);
        }
        ts.isLet = isLet;
        function isSuperCall(n) {
            return n.kind === 206 /* CallExpression */ && n.expression.kind === 106 /* SuperKeyword */;
        }
        ts.isSuperCall = isSuperCall;
        function isImportCall(n) {
            return n.kind === 206 /* CallExpression */ && n.expression.kind === 100 /* ImportKeyword */;
        }
        ts.isImportCall = isImportCall;
        function isImportMeta(n) {
            return ts.isMetaProperty(n)
                && n.keywordToken === 100 /* ImportKeyword */
                && n.name.escapedText === "meta";
        }
        ts.isImportMeta = isImportMeta;
        function isLiteralImportTypeNode(n) {
            return ts.isImportTypeNode(n) && ts.isLiteralTypeNode(n.argument) && ts.isStringLiteral(n.argument.literal);
        }
        ts.isLiteralImportTypeNode = isLiteralImportTypeNode;
        function isPrologueDirective(node) {
            return node.kind === 236 /* ExpressionStatement */
                && node.expression.kind === 10 /* StringLiteral */;
        }
        ts.isPrologueDirective = isPrologueDirective;
        function isCustomPrologue(node) {
            return !!(getEmitFlags(node) & 1048576 /* CustomPrologue */);
        }
        ts.isCustomPrologue = isCustomPrologue;
        function isHoistedFunction(node) {
            return isCustomPrologue(node)
                && ts.isFunctionDeclaration(node);
        }
        ts.isHoistedFunction = isHoistedFunction;
        function isHoistedVariable(node) {
            return ts.isIdentifier(node.name)
                && !node.initializer;
        }
        function isHoistedVariableStatement(node) {
            return isCustomPrologue(node)
                && ts.isVariableStatement(node)
                && ts.every(node.declarationList.declarations, isHoistedVariable);
        }
        ts.isHoistedVariableStatement = isHoistedVariableStatement;
        function getLeadingCommentRangesOfNode(node, sourceFileOfNode) {
            return node.kind !== 11 /* JsxText */ ? ts.getLeadingCommentRanges(sourceFileOfNode.text, node.pos) : undefined;
        }
        ts.getLeadingCommentRangesOfNode = getLeadingCommentRangesOfNode;
        function getJSDocCommentRanges(node, text) {
            var commentRanges = (node.kind === 162 /* Parameter */ ||
                node.kind === 161 /* TypeParameter */ ||
                node.kind === 211 /* FunctionExpression */ ||
                node.kind === 212 /* ArrowFunction */ ||
                node.kind === 210 /* ParenthesizedExpression */ ||
                node.kind === 252 /* VariableDeclaration */) ?
                ts.concatenate(ts.getTrailingCommentRanges(text, node.pos), ts.getLeadingCommentRanges(text, node.pos)) :
                ts.getLeadingCommentRanges(text, node.pos);
            // True if the comment starts with '/**' but not if it is '/**/'
            return ts.filter(commentRanges, function (comment) {
                return text.charCodeAt(comment.pos + 1) === 42 /* asterisk */ &&
                    text.charCodeAt(comment.pos + 2) === 42 /* asterisk */ &&
                    text.charCodeAt(comment.pos + 3) !== 47 /* slash */;
            });
        }
        ts.getJSDocCommentRanges = getJSDocCommentRanges;
        ts.fullTripleSlashReferencePathRegEx = /^(\/\/\/\s*<reference\s+path\s*=\s*)(('[^']*')|("[^"]*")).*?\/>/;
        var fullTripleSlashReferenceTypeReferenceDirectiveRegEx = /^(\/\/\/\s*<reference\s+types\s*=\s*)(('[^']*')|("[^"]*")).*?\/>/;
        ts.fullTripleSlashAMDReferencePathRegEx = /^(\/\/\/\s*<amd-dependency\s+path\s*=\s*)(('[^']*')|("[^"]*")).*?\/>/;
        var defaultLibReferenceRegEx = /^(\/\/\/\s*<reference\s+no-default-lib\s*=\s*)(('[^']*')|("[^"]*"))\s*\/>/;
        function isPartOfTypeNode(node) {
            if (175 /* FirstTypeNode */ <= node.kind && node.kind <= 198 /* LastTypeNode */) {
                return true;
            }
            switch (node.kind) {
                case 129 /* AnyKeyword */:
                case 153 /* UnknownKeyword */:
                case 145 /* NumberKeyword */:
                case 156 /* BigIntKeyword */:
                case 148 /* StringKeyword */:
                case 132 /* BooleanKeyword */:
                case 149 /* SymbolKeyword */:
                case 146 /* ObjectKeyword */:
                case 151 /* UndefinedKeyword */:
                case 142 /* NeverKeyword */:
                    return true;
                case 114 /* VoidKeyword */:
                    return node.parent.kind !== 215 /* VoidExpression */;
                case 226 /* ExpressionWithTypeArguments */:
                    return !isExpressionWithTypeArgumentsInClassExtendsClause(node);
                case 161 /* TypeParameter */:
                    return node.parent.kind === 193 /* MappedType */ || node.parent.kind === 188 /* InferType */;
                // Identifiers and qualified names may be type nodes, depending on their context. Climb
                // above them to find the lowest container
                case 79 /* Identifier */:
                    // If the identifier is the RHS of a qualified name, then it's a type iff its parent is.
                    if (node.parent.kind === 159 /* QualifiedName */ && node.parent.right === node) {
                        node = node.parent;
                    }
                    else if (node.parent.kind === 204 /* PropertyAccessExpression */ && node.parent.name === node) {
                        node = node.parent;
                    }
                    // At this point, node is either a qualified name or an identifier
                    ts.Debug.assert(node.kind === 79 /* Identifier */ || node.kind === 159 /* QualifiedName */ || node.kind === 204 /* PropertyAccessExpression */, "'node' was expected to be a qualified name, identifier or property access in 'isPartOfTypeNode'.");
                // falls through
                case 159 /* QualifiedName */:
                case 204 /* PropertyAccessExpression */:
                case 108 /* ThisKeyword */: {
                    var parent = node.parent;
                    if (parent.kind === 179 /* TypeQuery */) {
                        return false;
                    }
                    if (parent.kind === 198 /* ImportType */) {
                        return !parent.isTypeOf;
                    }
                    // Do not recursively call isPartOfTypeNode on the parent. In the example:
                    //
                    //     let a: A.B.C;
                    //
                    // Calling isPartOfTypeNode would consider the qualified name A.B a type node.
                    // Only C and A.B.C are type nodes.
                    if (175 /* FirstTypeNode */ <= parent.kind && parent.kind <= 198 /* LastTypeNode */) {
                        return true;
                    }
                    switch (parent.kind) {
                        case 226 /* ExpressionWithTypeArguments */:
                            return !isExpressionWithTypeArgumentsInClassExtendsClause(parent);
                        case 161 /* TypeParameter */:
                            return node === parent.constraint;
                        case 339 /* JSDocTemplateTag */:
                            return node === parent.constraint;
                        case 165 /* PropertyDeclaration */:
                        case 164 /* PropertySignature */:
                        case 162 /* Parameter */:
                        case 252 /* VariableDeclaration */:
                            return node === parent.type;
                        case 254 /* FunctionDeclaration */:
                        case 211 /* FunctionExpression */:
                        case 212 /* ArrowFunction */:
                        case 169 /* Constructor */:
                        case 167 /* MethodDeclaration */:
                        case 166 /* MethodSignature */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            return node === parent.type;
                        case 172 /* CallSignature */:
                        case 173 /* ConstructSignature */:
                        case 174 /* IndexSignature */:
                            return node === parent.type;
                        case 209 /* TypeAssertionExpression */:
                            return node === parent.type;
                        case 206 /* CallExpression */:
                        case 207 /* NewExpression */:
                            return ts.contains(parent.typeArguments, node);
                        case 208 /* TaggedTemplateExpression */:
                            // TODO (drosen): TaggedTemplateExpressions may eventually support type arguments.
                            return false;
                    }
                }
            }
            return false;
        }
        ts.isPartOfTypeNode = isPartOfTypeNode;
        function isChildOfNodeWithKind(node, kind) {
            while (node) {
                if (node.kind === kind) {
                    return true;
                }
                node = node.parent;
            }
            return false;
        }
        ts.isChildOfNodeWithKind = isChildOfNodeWithKind;
        // Warning: This has the same semantics as the forEach family of functions,
        //          in that traversal terminates in the event that 'visitor' supplies a truthy value.
        function forEachReturnStatement(body, visitor) {
            return traverse(body);
            function traverse(node) {
                switch (node.kind) {
                    case 245 /* ReturnStatement */:
                        return visitor(node);
                    case 261 /* CaseBlock */:
                    case 233 /* Block */:
                    case 237 /* IfStatement */:
                    case 238 /* DoStatement */:
                    case 239 /* WhileStatement */:
                    case 240 /* ForStatement */:
                    case 241 /* ForInStatement */:
                    case 242 /* ForOfStatement */:
                    case 246 /* WithStatement */:
                    case 247 /* SwitchStatement */:
                    case 287 /* CaseClause */:
                    case 288 /* DefaultClause */:
                    case 248 /* LabeledStatement */:
                    case 250 /* TryStatement */:
                    case 290 /* CatchClause */:
                        return ts.forEachChild(node, traverse);
                }
            }
        }
        ts.forEachReturnStatement = forEachReturnStatement;
        function forEachYieldExpression(body, visitor) {
            return traverse(body);
            function traverse(node) {
                switch (node.kind) {
                    case 222 /* YieldExpression */:
                        visitor(node);
                        var operand = node.expression;
                        if (operand) {
                            traverse(operand);
                        }
                        return;
                    case 258 /* EnumDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 259 /* ModuleDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                        // These are not allowed inside a generator now, but eventually they may be allowed
                        // as local types. Regardless, skip them to avoid the work.
                        return;
                    default:
                        if (ts.isFunctionLike(node)) {
                            if (node.name && node.name.kind === 160 /* ComputedPropertyName */) {
                                // Note that we will not include methods/accessors of a class because they would require
                                // first descending into the class. This is by design.
                                traverse(node.name.expression);
                                return;
                            }
                        }
                        else if (!isPartOfTypeNode(node)) {
                            // This is the general case, which should include mostly expressions and statements.
                            // Also includes NodeArrays.
                            ts.forEachChild(node, traverse);
                        }
                }
            }
        }
        ts.forEachYieldExpression = forEachYieldExpression;
        /**
         * Gets the most likely element type for a TypeNode. This is not an exhaustive test
         * as it assumes a rest argument can only be an array type (either T[], or Array<T>).
         *
         * @param node The type node.
         */
        function getRestParameterElementType(node) {
            if (node && node.kind === 181 /* ArrayType */) {
                return node.elementType;
            }
            else if (node && node.kind === 176 /* TypeReference */) {
                return ts.singleOrUndefined(node.typeArguments);
            }
            else {
                return undefined;
            }
        }
        ts.getRestParameterElementType = getRestParameterElementType;
        function getMembersOfDeclaration(node) {
            switch (node.kind) {
                case 256 /* InterfaceDeclaration */:
                case 255 /* ClassDeclaration */:
                case 224 /* ClassExpression */:
                case 180 /* TypeLiteral */:
                    return node.members;
                case 203 /* ObjectLiteralExpression */:
                    return node.properties;
            }
        }
        ts.getMembersOfDeclaration = getMembersOfDeclaration;
        function isVariableLike(node) {
            if (node) {
                switch (node.kind) {
                    case 201 /* BindingElement */:
                    case 294 /* EnumMember */:
                    case 162 /* Parameter */:
                    case 291 /* PropertyAssignment */:
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                    case 292 /* ShorthandPropertyAssignment */:
                    case 252 /* VariableDeclaration */:
                        return true;
                }
            }
            return false;
        }
        ts.isVariableLike = isVariableLike;
        function isVariableLikeOrAccessor(node) {
            return isVariableLike(node) || ts.isAccessor(node);
        }
        ts.isVariableLikeOrAccessor = isVariableLikeOrAccessor;
        function isVariableDeclarationInVariableStatement(node) {
            return node.parent.kind === 253 /* VariableDeclarationList */
                && node.parent.parent.kind === 235 /* VariableStatement */;
        }
        ts.isVariableDeclarationInVariableStatement = isVariableDeclarationInVariableStatement;
        function isValidESSymbolDeclaration(node) {
            return ts.isVariableDeclaration(node) ? isVarConst(node) && ts.isIdentifier(node.name) && isVariableDeclarationInVariableStatement(node) :
                ts.isPropertyDeclaration(node) ? hasEffectiveReadonlyModifier(node) && hasStaticModifier(node) :
                    ts.isPropertySignature(node) && hasEffectiveReadonlyModifier(node);
        }
        ts.isValidESSymbolDeclaration = isValidESSymbolDeclaration;
        function introducesArgumentsExoticObject(node) {
            switch (node.kind) {
                case 167 /* MethodDeclaration */:
                case 166 /* MethodSignature */:
                case 169 /* Constructor */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 254 /* FunctionDeclaration */:
                case 211 /* FunctionExpression */:
                    return true;
            }
            return false;
        }
        ts.introducesArgumentsExoticObject = introducesArgumentsExoticObject;
        function unwrapInnermostStatementOfLabel(node, beforeUnwrapLabelCallback) {
            while (true) {
                if (beforeUnwrapLabelCallback) {
                    beforeUnwrapLabelCallback(node);
                }
                if (node.statement.kind !== 248 /* LabeledStatement */) {
                    return node.statement;
                }
                node = node.statement;
            }
        }
        ts.unwrapInnermostStatementOfLabel = unwrapInnermostStatementOfLabel;
        function isFunctionBlock(node) {
            return node && node.kind === 233 /* Block */ && ts.isFunctionLike(node.parent);
        }
        ts.isFunctionBlock = isFunctionBlock;
        function isObjectLiteralMethod(node) {
            return node && node.kind === 167 /* MethodDeclaration */ && node.parent.kind === 203 /* ObjectLiteralExpression */;
        }
        ts.isObjectLiteralMethod = isObjectLiteralMethod;
        function isObjectLiteralOrClassExpressionMethod(node) {
            return node.kind === 167 /* MethodDeclaration */ &&
                (node.parent.kind === 203 /* ObjectLiteralExpression */ ||
                    node.parent.kind === 224 /* ClassExpression */);
        }
        ts.isObjectLiteralOrClassExpressionMethod = isObjectLiteralOrClassExpressionMethod;
        function isIdentifierTypePredicate(predicate) {
            return predicate && predicate.kind === 1 /* Identifier */;
        }
        ts.isIdentifierTypePredicate = isIdentifierTypePredicate;
        function isThisTypePredicate(predicate) {
            return predicate && predicate.kind === 0 /* This */;
        }
        ts.isThisTypePredicate = isThisTypePredicate;
        function getPropertyAssignment(objectLiteral, key, key2) {
            return objectLiteral.properties.filter(function (property) {
                if (property.kind === 291 /* PropertyAssignment */) {
                    var propName = getTextOfPropertyName(property.name);
                    return key === propName || (!!key2 && key2 === propName);
                }
                return false;
            });
        }
        ts.getPropertyAssignment = getPropertyAssignment;
        function getPropertyArrayElementValue(objectLiteral, propKey, elementValue) {
            return ts.firstDefined(getPropertyAssignment(objectLiteral, propKey), function (property) {
                return ts.isArrayLiteralExpression(property.initializer) ?
                    ts.find(property.initializer.elements, function (element) { return ts.isStringLiteral(element) && element.text === elementValue; }) :
                    undefined;
            });
        }
        ts.getPropertyArrayElementValue = getPropertyArrayElementValue;
        function getTsConfigObjectLiteralExpression(tsConfigSourceFile) {
            if (tsConfigSourceFile && tsConfigSourceFile.statements.length) {
                var expression = tsConfigSourceFile.statements[0].expression;
                return ts.tryCast(expression, ts.isObjectLiteralExpression);
            }
        }
        ts.getTsConfigObjectLiteralExpression = getTsConfigObjectLiteralExpression;
        function getTsConfigPropArrayElementValue(tsConfigSourceFile, propKey, elementValue) {
            return ts.firstDefined(getTsConfigPropArray(tsConfigSourceFile, propKey), function (property) {
                return ts.isArrayLiteralExpression(property.initializer) ?
                    ts.find(property.initializer.elements, function (element) { return ts.isStringLiteral(element) && element.text === elementValue; }) :
                    undefined;
            });
        }
        ts.getTsConfigPropArrayElementValue = getTsConfigPropArrayElementValue;
        function getTsConfigPropArray(tsConfigSourceFile, propKey) {
            var jsonObjectLiteral = getTsConfigObjectLiteralExpression(tsConfigSourceFile);
            return jsonObjectLiteral ? getPropertyAssignment(jsonObjectLiteral, propKey) : ts.emptyArray;
        }
        ts.getTsConfigPropArray = getTsConfigPropArray;
        function getContainingFunction(node) {
            return ts.findAncestor(node.parent, ts.isFunctionLike);
        }
        ts.getContainingFunction = getContainingFunction;
        function getContainingFunctionDeclaration(node) {
            return ts.findAncestor(node.parent, ts.isFunctionLikeDeclaration);
        }
        ts.getContainingFunctionDeclaration = getContainingFunctionDeclaration;
        function getContainingClass(node) {
            return ts.findAncestor(node.parent, ts.isClassLike);
        }
        ts.getContainingClass = getContainingClass;
        function getContainingClassStaticBlock(node) {
            return ts.findAncestor(node.parent, function (n) {
                if (ts.isClassLike(n) || ts.isFunctionLike(n)) {
                    return "quit";
                }
                return ts.isClassStaticBlockDeclaration(n);
            });
        }
        ts.getContainingClassStaticBlock = getContainingClassStaticBlock;
        function getContainingFunctionOrClassStaticBlock(node) {
            return ts.findAncestor(node.parent, ts.isFunctionLikeOrClassStaticBlockDeclaration);
        }
        ts.getContainingFunctionOrClassStaticBlock = getContainingFunctionOrClassStaticBlock;
        function getThisContainer(node, includeArrowFunctions) {
            ts.Debug.assert(node.kind !== 300 /* SourceFile */);
            while (true) {
                node = node.parent;
                if (!node) {
                    return ts.Debug.fail(); // If we never pass in a SourceFile, this should be unreachable, since we'll stop when we reach that.
                }
                switch (node.kind) {
                    case 160 /* ComputedPropertyName */:
                        // If the grandparent node is an object literal (as opposed to a class),
                        // then the computed property is not a 'this' container.
                        // A computed property name in a class needs to be a this container
                        // so that we can error on it.
                        if (ts.isClassLike(node.parent.parent)) {
                            return node;
                        }
                        // If this is a computed property, then the parent should not
                        // make it a this container. The parent might be a property
                        // in an object literal, like a method or accessor. But in order for
                        // such a parent to be a this container, the reference must be in
                        // the *body* of the container.
                        node = node.parent;
                        break;
                    case 163 /* Decorator */:
                        // Decorators are always applied outside of the body of a class or method.
                        if (node.parent.kind === 162 /* Parameter */ && ts.isClassElement(node.parent.parent)) {
                            // If the decorator's parent is a Parameter, we resolve the this container from
                            // the grandparent class declaration.
                            node = node.parent.parent;
                        }
                        else if (ts.isClassElement(node.parent)) {
                            // If the decorator's parent is a class element, we resolve the 'this' container
                            // from the parent class declaration.
                            node = node.parent;
                        }
                        break;
                    case 212 /* ArrowFunction */:
                        if (!includeArrowFunctions) {
                            continue;
                        }
                    // falls through
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 259 /* ModuleDeclaration */:
                    case 168 /* ClassStaticBlockDeclaration */:
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 169 /* Constructor */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 172 /* CallSignature */:
                    case 173 /* ConstructSignature */:
                    case 174 /* IndexSignature */:
                    case 258 /* EnumDeclaration */:
                    case 300 /* SourceFile */:
                        return node;
                }
            }
        }
        ts.getThisContainer = getThisContainer;
        function isInTopLevelContext(node) {
            // The name of a class or function declaration is a BindingIdentifier in its surrounding scope.
            if (ts.isIdentifier(node) && (ts.isClassDeclaration(node.parent) || ts.isFunctionDeclaration(node.parent)) && node.parent.name === node) {
                node = node.parent;
            }
            var container = getThisContainer(node, /*includeArrowFunctions*/ true);
            return ts.isSourceFile(container);
        }
        ts.isInTopLevelContext = isInTopLevelContext;
        function getNewTargetContainer(node) {
            var container = getThisContainer(node, /*includeArrowFunctions*/ false);
            if (container) {
                switch (container.kind) {
                    case 169 /* Constructor */:
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                        return container;
                }
            }
            return undefined;
        }
        ts.getNewTargetContainer = getNewTargetContainer;
        /**
         * Given an super call/property node, returns the closest node where
         * - a super call/property access is legal in the node and not legal in the parent node the node.
         *   i.e. super call is legal in constructor but not legal in the class body.
         * - the container is an arrow function (so caller might need to call getSuperContainer again in case it needs to climb higher)
         * - a super call/property is definitely illegal in the container (but might be legal in some subnode)
         *   i.e. super property access is illegal in function declaration but can be legal in the statement list
         */
        function getSuperContainer(node, stopOnFunctions) {
            while (true) {
                node = node.parent;
                if (!node) {
                    return node;
                }
                switch (node.kind) {
                    case 160 /* ComputedPropertyName */:
                        node = node.parent;
                        break;
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        if (!stopOnFunctions) {
                            continue;
                        }
                    // falls through
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 169 /* Constructor */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 168 /* ClassStaticBlockDeclaration */:
                        return node;
                    case 163 /* Decorator */:
                        // Decorators are always applied outside of the body of a class or method.
                        if (node.parent.kind === 162 /* Parameter */ && ts.isClassElement(node.parent.parent)) {
                            // If the decorator's parent is a Parameter, we resolve the this container from
                            // the grandparent class declaration.
                            node = node.parent.parent;
                        }
                        else if (ts.isClassElement(node.parent)) {
                            // If the decorator's parent is a class element, we resolve the 'this' container
                            // from the parent class declaration.
                            node = node.parent;
                        }
                        break;
                }
            }
        }
        ts.getSuperContainer = getSuperContainer;
        function getImmediatelyInvokedFunctionExpression(func) {
            if (func.kind === 211 /* FunctionExpression */ || func.kind === 212 /* ArrowFunction */) {
                var prev = func;
                var parent = func.parent;
                while (parent.kind === 210 /* ParenthesizedExpression */) {
                    prev = parent;
                    parent = parent.parent;
                }
                if (parent.kind === 206 /* CallExpression */ && parent.expression === prev) {
                    return parent;
                }
            }
        }
        ts.getImmediatelyInvokedFunctionExpression = getImmediatelyInvokedFunctionExpression;
        function isSuperOrSuperProperty(node) {
            return node.kind === 106 /* SuperKeyword */
                || isSuperProperty(node);
        }
        ts.isSuperOrSuperProperty = isSuperOrSuperProperty;
        /**
         * Determines whether a node is a property or element access expression for `super`.
         */
        function isSuperProperty(node) {
            var kind = node.kind;
            return (kind === 204 /* PropertyAccessExpression */ || kind === 205 /* ElementAccessExpression */)
                && node.expression.kind === 106 /* SuperKeyword */;
        }
        ts.isSuperProperty = isSuperProperty;
        /**
         * Determines whether a node is a property or element access expression for `this`.
         */
        function isThisProperty(node) {
            var kind = node.kind;
            return (kind === 204 /* PropertyAccessExpression */ || kind === 205 /* ElementAccessExpression */)
                && node.expression.kind === 108 /* ThisKeyword */;
        }
        ts.isThisProperty = isThisProperty;
        function isThisInitializedDeclaration(node) {
            var _a;
            return !!node && ts.isVariableDeclaration(node) && ((_a = node.initializer) === null || _a === void 0 ? void 0 : _a.kind) === 108 /* ThisKeyword */;
        }
        ts.isThisInitializedDeclaration = isThisInitializedDeclaration;
        function isThisInitializedObjectBindingExpression(node) {
            return !!node
                && (ts.isShorthandPropertyAssignment(node) || ts.isPropertyAssignment(node))
                && ts.isBinaryExpression(node.parent.parent)
                && node.parent.parent.operatorToken.kind === 63 /* EqualsToken */
                && node.parent.parent.right.kind === 108 /* ThisKeyword */;
        }
        ts.isThisInitializedObjectBindingExpression = isThisInitializedObjectBindingExpression;
        function getEntityNameFromTypeNode(node) {
            switch (node.kind) {
                case 176 /* TypeReference */:
                    return node.typeName;
                case 226 /* ExpressionWithTypeArguments */:
                    return isEntityNameExpression(node.expression)
                        ? node.expression
                        : undefined;
                // TODO(rbuckton): These aren't valid TypeNodes, but we treat them as such because of `isPartOfTypeNode`, which returns `true` for things that aren't `TypeNode`s.
                case 79 /* Identifier */:
                case 159 /* QualifiedName */:
                    return node;
            }
            return undefined;
        }
        ts.getEntityNameFromTypeNode = getEntityNameFromTypeNode;
        function getInvokedExpression(node) {
            switch (node.kind) {
                case 208 /* TaggedTemplateExpression */:
                    return node.tag;
                case 278 /* JsxOpeningElement */:
                case 277 /* JsxSelfClosingElement */:
                    return node.tagName;
                default:
                    return node.expression;
            }
        }
        ts.getInvokedExpression = getInvokedExpression;
        function nodeCanBeDecorated(node, parent, grandparent) {
            // private names cannot be used with decorators yet
            if (ts.isNamedDeclaration(node) && ts.isPrivateIdentifier(node.name)) {
                return false;
            }
            switch (node.kind) {
                case 255 /* ClassDeclaration */:
                    // classes are valid targets
                    return true;
                case 165 /* PropertyDeclaration */:
                    // property declarations are valid if their parent is a class declaration.
                    return parent.kind === 255 /* ClassDeclaration */;
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 167 /* MethodDeclaration */:
                    // if this method has a body and its parent is a class declaration, this is a valid target.
                    return node.body !== undefined
                        && parent.kind === 255 /* ClassDeclaration */;
                case 162 /* Parameter */:
                    // if the parameter's parent has a body and its grandparent is a class declaration, this is a valid target;
                    return parent.body !== undefined
                        && (parent.kind === 169 /* Constructor */
                            || parent.kind === 167 /* MethodDeclaration */
                            || parent.kind === 171 /* SetAccessor */)
                        && grandparent.kind === 255 /* ClassDeclaration */;
            }
            return false;
        }
        ts.nodeCanBeDecorated = nodeCanBeDecorated;
        function nodeIsDecorated(node, parent, grandparent) {
            return node.decorators !== undefined
                && nodeCanBeDecorated(node, parent, grandparent); // TODO: GH#18217
        }
        ts.nodeIsDecorated = nodeIsDecorated;
        function nodeOrChildIsDecorated(node, parent, grandparent) {
            return nodeIsDecorated(node, parent, grandparent) || childIsDecorated(node, parent); // TODO: GH#18217
        }
        ts.nodeOrChildIsDecorated = nodeOrChildIsDecorated;
        function childIsDecorated(node, parent) {
            switch (node.kind) {
                case 255 /* ClassDeclaration */:
                    return ts.some(node.members, function (m) { return nodeOrChildIsDecorated(m, node, parent); }); // TODO: GH#18217
                case 167 /* MethodDeclaration */:
                case 171 /* SetAccessor */:
                case 169 /* Constructor */:
                    return ts.some(node.parameters, function (p) { return nodeIsDecorated(p, node, parent); }); // TODO: GH#18217
                default:
                    return false;
            }
        }
        ts.childIsDecorated = childIsDecorated;
        function classOrConstructorParameterIsDecorated(node) {
            if (nodeIsDecorated(node))
                return true;
            var constructor = getFirstConstructorWithBody(node);
            return !!constructor && childIsDecorated(constructor, node);
        }
        ts.classOrConstructorParameterIsDecorated = classOrConstructorParameterIsDecorated;
        function isJSXTagName(node) {
            var parent = node.parent;
            if (parent.kind === 278 /* JsxOpeningElement */ ||
                parent.kind === 277 /* JsxSelfClosingElement */ ||
                parent.kind === 279 /* JsxClosingElement */) {
                return parent.tagName === node;
            }
            return false;
        }
        ts.isJSXTagName = isJSXTagName;
        function isExpressionNode(node) {
            switch (node.kind) {
                case 106 /* SuperKeyword */:
                case 104 /* NullKeyword */:
                case 110 /* TrueKeyword */:
                case 95 /* FalseKeyword */:
                case 13 /* RegularExpressionLiteral */:
                case 202 /* ArrayLiteralExpression */:
                case 203 /* ObjectLiteralExpression */:
                case 204 /* PropertyAccessExpression */:
                case 205 /* ElementAccessExpression */:
                case 206 /* CallExpression */:
                case 207 /* NewExpression */:
                case 208 /* TaggedTemplateExpression */:
                case 227 /* AsExpression */:
                case 209 /* TypeAssertionExpression */:
                case 228 /* NonNullExpression */:
                case 210 /* ParenthesizedExpression */:
                case 211 /* FunctionExpression */:
                case 224 /* ClassExpression */:
                case 212 /* ArrowFunction */:
                case 215 /* VoidExpression */:
                case 213 /* DeleteExpression */:
                case 214 /* TypeOfExpression */:
                case 217 /* PrefixUnaryExpression */:
                case 218 /* PostfixUnaryExpression */:
                case 219 /* BinaryExpression */:
                case 220 /* ConditionalExpression */:
                case 223 /* SpreadElement */:
                case 221 /* TemplateExpression */:
                case 225 /* OmittedExpression */:
                case 276 /* JsxElement */:
                case 277 /* JsxSelfClosingElement */:
                case 280 /* JsxFragment */:
                case 222 /* YieldExpression */:
                case 216 /* AwaitExpression */:
                case 229 /* MetaProperty */:
                    return true;
                case 159 /* QualifiedName */:
                    while (node.parent.kind === 159 /* QualifiedName */) {
                        node = node.parent;
                    }
                    return node.parent.kind === 179 /* TypeQuery */ || ts.isJSDocLinkLike(node.parent) || ts.isJSDocNameReference(node.parent) || ts.isJSDocMemberName(node.parent) || isJSXTagName(node);
                case 306 /* JSDocMemberName */:
                    while (ts.isJSDocMemberName(node.parent)) {
                        node = node.parent;
                    }
                    return node.parent.kind === 179 /* TypeQuery */ || ts.isJSDocLinkLike(node.parent) || ts.isJSDocNameReference(node.parent) || ts.isJSDocMemberName(node.parent) || isJSXTagName(node);
                case 79 /* Identifier */:
                    if (node.parent.kind === 179 /* TypeQuery */ || ts.isJSDocLinkLike(node.parent) || ts.isJSDocNameReference(node.parent) || ts.isJSDocMemberName(node.parent) || isJSXTagName(node)) {
                        return true;
                    }
                // falls through
                case 8 /* NumericLiteral */:
                case 9 /* BigIntLiteral */:
                case 10 /* StringLiteral */:
                case 14 /* NoSubstitutionTemplateLiteral */:
                case 108 /* ThisKeyword */:
                    return isInExpressionContext(node);
                default:
                    return false;
            }
        }
        ts.isExpressionNode = isExpressionNode;
        function isInExpressionContext(node) {
            var parent = node.parent;
            switch (parent.kind) {
                case 252 /* VariableDeclaration */:
                case 162 /* Parameter */:
                case 165 /* PropertyDeclaration */:
                case 164 /* PropertySignature */:
                case 294 /* EnumMember */:
                case 291 /* PropertyAssignment */:
                case 201 /* BindingElement */:
                    return parent.initializer === node;
                case 236 /* ExpressionStatement */:
                case 237 /* IfStatement */:
                case 238 /* DoStatement */:
                case 239 /* WhileStatement */:
                case 245 /* ReturnStatement */:
                case 246 /* WithStatement */:
                case 247 /* SwitchStatement */:
                case 287 /* CaseClause */:
                case 249 /* ThrowStatement */:
                    return parent.expression === node;
                case 240 /* ForStatement */:
                    var forStatement = parent;
                    return (forStatement.initializer === node && forStatement.initializer.kind !== 253 /* VariableDeclarationList */) ||
                        forStatement.condition === node ||
                        forStatement.incrementor === node;
                case 241 /* ForInStatement */:
                case 242 /* ForOfStatement */:
                    var forInStatement = parent;
                    return (forInStatement.initializer === node && forInStatement.initializer.kind !== 253 /* VariableDeclarationList */) ||
                        forInStatement.expression === node;
                case 209 /* TypeAssertionExpression */:
                case 227 /* AsExpression */:
                    return node === parent.expression;
                case 231 /* TemplateSpan */:
                    return node === parent.expression;
                case 160 /* ComputedPropertyName */:
                    return node === parent.expression;
                case 163 /* Decorator */:
                case 286 /* JsxExpression */:
                case 285 /* JsxSpreadAttribute */:
                case 293 /* SpreadAssignment */:
                    return true;
                case 226 /* ExpressionWithTypeArguments */:
                    return parent.expression === node && isExpressionWithTypeArgumentsInClassExtendsClause(parent);
                case 292 /* ShorthandPropertyAssignment */:
                    return parent.objectAssignmentInitializer === node;
                default:
                    return isExpressionNode(parent);
            }
        }
        ts.isInExpressionContext = isInExpressionContext;
        function isPartOfTypeQuery(node) {
            while (node.kind === 159 /* QualifiedName */ || node.kind === 79 /* Identifier */) {
                node = node.parent;
            }
            return node.kind === 179 /* TypeQuery */;
        }
        ts.isPartOfTypeQuery = isPartOfTypeQuery;
        function isNamespaceReexportDeclaration(node) {
            return ts.isNamespaceExport(node) && !!node.parent.moduleSpecifier;
        }
        ts.isNamespaceReexportDeclaration = isNamespaceReexportDeclaration;
        function isExternalModuleImportEqualsDeclaration(node) {
            return node.kind === 263 /* ImportEqualsDeclaration */ && node.moduleReference.kind === 275 /* ExternalModuleReference */;
        }
        ts.isExternalModuleImportEqualsDeclaration = isExternalModuleImportEqualsDeclaration;
        function getExternalModuleImportEqualsDeclarationExpression(node) {
            ts.Debug.assert(isExternalModuleImportEqualsDeclaration(node));
            return node.moduleReference.expression;
        }
        ts.getExternalModuleImportEqualsDeclarationExpression = getExternalModuleImportEqualsDeclarationExpression;
        function getExternalModuleRequireArgument(node) {
            return isRequireVariableDeclaration(node) && getLeftmostAccessExpression(node.initializer).arguments[0];
        }
        ts.getExternalModuleRequireArgument = getExternalModuleRequireArgument;
        function isInternalModuleImportEqualsDeclaration(node) {
            return node.kind === 263 /* ImportEqualsDeclaration */ && node.moduleReference.kind !== 275 /* ExternalModuleReference */;
        }
        ts.isInternalModuleImportEqualsDeclaration = isInternalModuleImportEqualsDeclaration;
        function isSourceFileJS(file) {
            return isInJSFile(file);
        }
        ts.isSourceFileJS = isSourceFileJS;
        function isSourceFileNotJS(file) {
            return !isInJSFile(file);
        }
        ts.isSourceFileNotJS = isSourceFileNotJS;
        function isInJSFile(node) {
            return !!node && !!(node.flags & 131072 /* JavaScriptFile */);
        }
        ts.isInJSFile = isInJSFile;
        function isInJsonFile(node) {
            return !!node && !!(node.flags & 33554432 /* JsonFile */);
        }
        ts.isInJsonFile = isInJsonFile;
        function isSourceFileNotJson(file) {
            return !isJsonSourceFile(file);
        }
        ts.isSourceFileNotJson = isSourceFileNotJson;
        function isInJSDoc(node) {
            return !!node && !!(node.flags & 4194304 /* JSDoc */);
        }
        ts.isInJSDoc = isInJSDoc;
        function isJSDocIndexSignature(node) {
            return ts.isTypeReferenceNode(node) &&
                ts.isIdentifier(node.typeName) &&
                node.typeName.escapedText === "Object" &&
                node.typeArguments && node.typeArguments.length === 2 &&
                (node.typeArguments[0].kind === 148 /* StringKeyword */ || node.typeArguments[0].kind === 145 /* NumberKeyword */);
        }
        ts.isJSDocIndexSignature = isJSDocIndexSignature;
        function isRequireCall(callExpression, requireStringLiteralLikeArgument) {
            if (callExpression.kind !== 206 /* CallExpression */) {
                return false;
            }
            var _a = callExpression, expression = _a.expression, args = _a.arguments;
            if (expression.kind !== 79 /* Identifier */ || expression.escapedText !== "require") {
                return false;
            }
            if (args.length !== 1) {
                return false;
            }
            var arg = args[0];
            return !requireStringLiteralLikeArgument || ts.isStringLiteralLike(arg);
        }
        ts.isRequireCall = isRequireCall;
        /**
         * Returns true if the node is a VariableDeclaration initialized to a require call (see `isRequireCall`).
         * This function does not test if the node is in a JavaScript file or not.
         */
        function isRequireVariableDeclaration(node) {
            if (node.kind === 201 /* BindingElement */) {
                node = node.parent.parent;
            }
            return ts.isVariableDeclaration(node) && !!node.initializer && isRequireCall(getLeftmostAccessExpression(node.initializer), /*requireStringLiteralLikeArgument*/ true);
        }
        ts.isRequireVariableDeclaration = isRequireVariableDeclaration;
        function isRequireVariableStatement(node) {
            return ts.isVariableStatement(node)
                && node.declarationList.declarations.length > 0
                && ts.every(node.declarationList.declarations, function (decl) { return isRequireVariableDeclaration(decl); });
        }
        ts.isRequireVariableStatement = isRequireVariableStatement;
        function isSingleOrDoubleQuote(charCode) {
            return charCode === 39 /* singleQuote */ || charCode === 34 /* doubleQuote */;
        }
        ts.isSingleOrDoubleQuote = isSingleOrDoubleQuote;
        function isStringDoubleQuoted(str, sourceFile) {
            return getSourceTextOfNodeFromSourceFile(sourceFile, str).charCodeAt(0) === 34 /* doubleQuote */;
        }
        ts.isStringDoubleQuoted = isStringDoubleQuoted;
        function isAssignmentDeclaration(decl) {
            return ts.isBinaryExpression(decl) || isAccessExpression(decl) || ts.isIdentifier(decl) || ts.isCallExpression(decl);
        }
        ts.isAssignmentDeclaration = isAssignmentDeclaration;
        /** Get the initializer, taking into account defaulted Javascript initializers */
        function getEffectiveInitializer(node) {
            if (isInJSFile(node) && node.initializer &&
                ts.isBinaryExpression(node.initializer) &&
                (node.initializer.operatorToken.kind === 56 /* BarBarToken */ || node.initializer.operatorToken.kind === 60 /* QuestionQuestionToken */) &&
                node.name && isEntityNameExpression(node.name) && isSameEntityName(node.name, node.initializer.left)) {
                return node.initializer.right;
            }
            return node.initializer;
        }
        ts.getEffectiveInitializer = getEffectiveInitializer;
        /** Get the declaration initializer when it is container-like (See getExpandoInitializer). */
        function getDeclaredExpandoInitializer(node) {
            var init = getEffectiveInitializer(node);
            return init && getExpandoInitializer(init, isPrototypeAccess(node.name));
        }
        ts.getDeclaredExpandoInitializer = getDeclaredExpandoInitializer;
        function hasExpandoValueProperty(node, isPrototypeAssignment) {
            return ts.forEach(node.properties, function (p) {
                return ts.isPropertyAssignment(p) &&
                    ts.isIdentifier(p.name) &&
                    p.name.escapedText === "value" &&
                    p.initializer &&
                    getExpandoInitializer(p.initializer, isPrototypeAssignment);
            });
        }
        /**
         * Get the assignment 'initializer' -- the righthand side-- when the initializer is container-like (See getExpandoInitializer).
         * We treat the right hand side of assignments with container-like initializers as declarations.
         */
        function getAssignedExpandoInitializer(node) {
            if (node && node.parent && ts.isBinaryExpression(node.parent) && node.parent.operatorToken.kind === 63 /* EqualsToken */) {
                var isPrototypeAssignment = isPrototypeAccess(node.parent.left);
                return getExpandoInitializer(node.parent.right, isPrototypeAssignment) ||
                    getDefaultedExpandoInitializer(node.parent.left, node.parent.right, isPrototypeAssignment);
            }
            if (node && ts.isCallExpression(node) && isBindableObjectDefinePropertyCall(node)) {
                var result = hasExpandoValueProperty(node.arguments[2], node.arguments[1].text === "prototype");
                if (result) {
                    return result;
                }
            }
        }
        ts.getAssignedExpandoInitializer = getAssignedExpandoInitializer;
        /**
         * Recognized expando initializers are:
         * 1. (function() {})() -- IIFEs
         * 2. function() { } -- Function expressions
         * 3. class { } -- Class expressions
         * 4. {} -- Empty object literals
         * 5. { ... } -- Non-empty object literals, when used to initialize a prototype, like `C.prototype = { m() { } }`
         *
         * This function returns the provided initializer, or undefined if it is not valid.
         */
        function getExpandoInitializer(initializer, isPrototypeAssignment) {
            if (ts.isCallExpression(initializer)) {
                var e = skipParentheses(initializer.expression);
                return e.kind === 211 /* FunctionExpression */ || e.kind === 212 /* ArrowFunction */ ? initializer : undefined;
            }
            if (initializer.kind === 211 /* FunctionExpression */ ||
                initializer.kind === 224 /* ClassExpression */ ||
                initializer.kind === 212 /* ArrowFunction */) {
                return initializer;
            }
            if (ts.isObjectLiteralExpression(initializer) && (initializer.properties.length === 0 || isPrototypeAssignment)) {
                return initializer;
            }
        }
        ts.getExpandoInitializer = getExpandoInitializer;
        /**
         * A defaulted expando initializer matches the pattern
         * `Lhs = Lhs || ExpandoInitializer`
         * or `var Lhs = Lhs || ExpandoInitializer`
         *
         * The second Lhs is required to be the same as the first except that it may be prefixed with
         * 'window.', 'global.' or 'self.' The second Lhs is otherwise ignored by the binder and checker.
         */
        function getDefaultedExpandoInitializer(name, initializer, isPrototypeAssignment) {
            var e = ts.isBinaryExpression(initializer)
                && (initializer.operatorToken.kind === 56 /* BarBarToken */ || initializer.operatorToken.kind === 60 /* QuestionQuestionToken */)
                && getExpandoInitializer(initializer.right, isPrototypeAssignment);
            if (e && isSameEntityName(name, initializer.left)) {
                return e;
            }
        }
        function isDefaultedExpandoInitializer(node) {
            var name = ts.isVariableDeclaration(node.parent) ? node.parent.name :
                ts.isBinaryExpression(node.parent) && node.parent.operatorToken.kind === 63 /* EqualsToken */ ? node.parent.left :
                    undefined;
            return name && getExpandoInitializer(node.right, isPrototypeAccess(name)) && isEntityNameExpression(name) && isSameEntityName(name, node.left);
        }
        ts.isDefaultedExpandoInitializer = isDefaultedExpandoInitializer;
        /** Given an expando initializer, return its declaration name, or the left-hand side of the assignment if it's part of an assignment declaration. */
        function getNameOfExpando(node) {
            if (ts.isBinaryExpression(node.parent)) {
                var parent = ((node.parent.operatorToken.kind === 56 /* BarBarToken */ || node.parent.operatorToken.kind === 60 /* QuestionQuestionToken */) && ts.isBinaryExpression(node.parent.parent)) ? node.parent.parent : node.parent;
                if (parent.operatorToken.kind === 63 /* EqualsToken */ && ts.isIdentifier(parent.left)) {
                    return parent.left;
                }
            }
            else if (ts.isVariableDeclaration(node.parent)) {
                return node.parent.name;
            }
        }
        ts.getNameOfExpando = getNameOfExpando;
        /**
         * Is the 'declared' name the same as the one in the initializer?
         * @return true for identical entity names, as well as ones where the initializer is prefixed with
         * 'window', 'self' or 'global'. For example:
         *
         * var my = my || {}
         * var min = window.min || {}
         * my.app = self.my.app || class { }
         */
        function isSameEntityName(name, initializer) {
            if (isPropertyNameLiteral(name) && isPropertyNameLiteral(initializer)) {
                return getTextOfIdentifierOrLiteral(name) === getTextOfIdentifierOrLiteral(initializer);
            }
            if (ts.isIdentifier(name) && isLiteralLikeAccess(initializer) &&
                (initializer.expression.kind === 108 /* ThisKeyword */ ||
                    ts.isIdentifier(initializer.expression) &&
                        (initializer.expression.escapedText === "window" ||
                            initializer.expression.escapedText === "self" ||
                            initializer.expression.escapedText === "global"))) {
                var nameOrArgument = getNameOrArgument(initializer);
                if (ts.isPrivateIdentifier(nameOrArgument)) {
                    ts.Debug.fail("Unexpected PrivateIdentifier in name expression with literal-like access.");
                }
                return isSameEntityName(name, nameOrArgument);
            }
            if (isLiteralLikeAccess(name) && isLiteralLikeAccess(initializer)) {
                return getElementOrPropertyAccessName(name) === getElementOrPropertyAccessName(initializer)
                    && isSameEntityName(name.expression, initializer.expression);
            }
            return false;
        }
        ts.isSameEntityName = isSameEntityName;
        function getRightMostAssignedExpression(node) {
            while (isAssignmentExpression(node, /*excludeCompoundAssignments*/ true)) {
                node = node.right;
            }
            return node;
        }
        ts.getRightMostAssignedExpression = getRightMostAssignedExpression;
        function isExportsIdentifier(node) {
            return ts.isIdentifier(node) && node.escapedText === "exports";
        }
        ts.isExportsIdentifier = isExportsIdentifier;
        function isModuleIdentifier(node) {
            return ts.isIdentifier(node) && node.escapedText === "module";
        }
        ts.isModuleIdentifier = isModuleIdentifier;
        function isModuleExportsAccessExpression(node) {
            return (ts.isPropertyAccessExpression(node) || isLiteralLikeElementAccess(node))
                && isModuleIdentifier(node.expression)
                && getElementOrPropertyAccessName(node) === "exports";
        }
        ts.isModuleExportsAccessExpression = isModuleExportsAccessExpression;
        /// Given a BinaryExpression, returns SpecialPropertyAssignmentKind for the various kinds of property
        /// assignments we treat as special in the binder
        function getAssignmentDeclarationKind(expr) {
            var special = getAssignmentDeclarationKindWorker(expr);
            return special === 5 /* Property */ || isInJSFile(expr) ? special : 0 /* None */;
        }
        ts.getAssignmentDeclarationKind = getAssignmentDeclarationKind;
        function isBindableObjectDefinePropertyCall(expr) {
            return ts.length(expr.arguments) === 3 &&
                ts.isPropertyAccessExpression(expr.expression) &&
                ts.isIdentifier(expr.expression.expression) &&
                ts.idText(expr.expression.expression) === "Object" &&
                ts.idText(expr.expression.name) === "defineProperty" &&
                isStringOrNumericLiteralLike(expr.arguments[1]) &&
                isBindableStaticNameExpression(expr.arguments[0], /*excludeThisKeyword*/ true);
        }
        ts.isBindableObjectDefinePropertyCall = isBindableObjectDefinePropertyCall;
        /** x.y OR x[0] */
        function isLiteralLikeAccess(node) {
            return ts.isPropertyAccessExpression(node) || isLiteralLikeElementAccess(node);
        }
        ts.isLiteralLikeAccess = isLiteralLikeAccess;
        /** x[0] OR x['a'] OR x[Symbol.y] */
        function isLiteralLikeElementAccess(node) {
            return ts.isElementAccessExpression(node) && isStringOrNumericLiteralLike(node.argumentExpression);
        }
        ts.isLiteralLikeElementAccess = isLiteralLikeElementAccess;
        /** Any series of property and element accesses. */
        function isBindableStaticAccessExpression(node, excludeThisKeyword) {
            return ts.isPropertyAccessExpression(node) && (!excludeThisKeyword && node.expression.kind === 108 /* ThisKeyword */ || ts.isIdentifier(node.name) && isBindableStaticNameExpression(node.expression, /*excludeThisKeyword*/ true))
                || isBindableStaticElementAccessExpression(node, excludeThisKeyword);
        }
        ts.isBindableStaticAccessExpression = isBindableStaticAccessExpression;
        /** Any series of property and element accesses, ending in a literal element access */
        function isBindableStaticElementAccessExpression(node, excludeThisKeyword) {
            return isLiteralLikeElementAccess(node)
                && ((!excludeThisKeyword && node.expression.kind === 108 /* ThisKeyword */) ||
                    isEntityNameExpression(node.expression) ||
                    isBindableStaticAccessExpression(node.expression, /*excludeThisKeyword*/ true));
        }
        ts.isBindableStaticElementAccessExpression = isBindableStaticElementAccessExpression;
        function isBindableStaticNameExpression(node, excludeThisKeyword) {
            return isEntityNameExpression(node) || isBindableStaticAccessExpression(node, excludeThisKeyword);
        }
        ts.isBindableStaticNameExpression = isBindableStaticNameExpression;
        function getNameOrArgument(expr) {
            if (ts.isPropertyAccessExpression(expr)) {
                return expr.name;
            }
            return expr.argumentExpression;
        }
        ts.getNameOrArgument = getNameOrArgument;
        function getAssignmentDeclarationKindWorker(expr) {
            if (ts.isCallExpression(expr)) {
                if (!isBindableObjectDefinePropertyCall(expr)) {
                    return 0 /* None */;
                }
                var entityName = expr.arguments[0];
                if (isExportsIdentifier(entityName) || isModuleExportsAccessExpression(entityName)) {
                    return 8 /* ObjectDefinePropertyExports */;
                }
                if (isBindableStaticAccessExpression(entityName) && getElementOrPropertyAccessName(entityName) === "prototype") {
                    return 9 /* ObjectDefinePrototypeProperty */;
                }
                return 7 /* ObjectDefinePropertyValue */;
            }
            if (expr.operatorToken.kind !== 63 /* EqualsToken */ || !isAccessExpression(expr.left) || isVoidZero(getRightMostAssignedExpression(expr))) {
                return 0 /* None */;
            }
            if (isBindableStaticNameExpression(expr.left.expression, /*excludeThisKeyword*/ true) && getElementOrPropertyAccessName(expr.left) === "prototype" && ts.isObjectLiteralExpression(getInitializerOfBinaryExpression(expr))) {
                // F.prototype = { ... }
                return 6 /* Prototype */;
            }
            return getAssignmentDeclarationPropertyAccessKind(expr.left);
        }
        function isVoidZero(node) {
            return ts.isVoidExpression(node) && ts.isNumericLiteral(node.expression) && node.expression.text === "0";
        }
        /**
         * Does not handle signed numeric names like `a[+0]` - handling those would require handling prefix unary expressions
         * throughout late binding handling as well, which is awkward (but ultimately probably doable if there is demand)
         */
        /* @internal */
        function getElementOrPropertyAccessArgumentExpressionOrName(node) {
            if (ts.isPropertyAccessExpression(node)) {
                return node.name;
            }
            var arg = skipParentheses(node.argumentExpression);
            if (ts.isNumericLiteral(arg) || ts.isStringLiteralLike(arg)) {
                return arg;
            }
            return node;
        }
        ts.getElementOrPropertyAccessArgumentExpressionOrName = getElementOrPropertyAccessArgumentExpressionOrName;
        function getElementOrPropertyAccessName(node) {
            var name = getElementOrPropertyAccessArgumentExpressionOrName(node);
            if (name) {
                if (ts.isIdentifier(name)) {
                    return name.escapedText;
                }
                if (ts.isStringLiteralLike(name) || ts.isNumericLiteral(name)) {
                    return ts.escapeLeadingUnderscores(name.text);
                }
            }
            return undefined;
        }
        ts.getElementOrPropertyAccessName = getElementOrPropertyAccessName;
        function getAssignmentDeclarationPropertyAccessKind(lhs) {
            if (lhs.expression.kind === 108 /* ThisKeyword */) {
                return 4 /* ThisProperty */;
            }
            else if (isModuleExportsAccessExpression(lhs)) {
                // module.exports = expr
                return 2 /* ModuleExports */;
            }
            else if (isBindableStaticNameExpression(lhs.expression, /*excludeThisKeyword*/ true)) {
                if (isPrototypeAccess(lhs.expression)) {
                    // F.G....prototype.x = expr
                    return 3 /* PrototypeProperty */;
                }
                var nextToLast = lhs;
                while (!ts.isIdentifier(nextToLast.expression)) {
                    nextToLast = nextToLast.expression;
                }
                var id = nextToLast.expression;
                if ((id.escapedText === "exports" ||
                    id.escapedText === "module" && getElementOrPropertyAccessName(nextToLast) === "exports") &&
                    // ExportsProperty does not support binding with computed names
                    isBindableStaticAccessExpression(lhs)) {
                    // exports.name = expr OR module.exports.name = expr OR exports["name"] = expr ...
                    return 1 /* ExportsProperty */;
                }
                if (isBindableStaticNameExpression(lhs, /*excludeThisKeyword*/ true) || (ts.isElementAccessExpression(lhs) && isDynamicName(lhs))) {
                    // F.G...x = expr
                    return 5 /* Property */;
                }
            }
            return 0 /* None */;
        }
        ts.getAssignmentDeclarationPropertyAccessKind = getAssignmentDeclarationPropertyAccessKind;
        function getInitializerOfBinaryExpression(expr) {
            while (ts.isBinaryExpression(expr.right)) {
                expr = expr.right;
            }
            return expr.right;
        }
        ts.getInitializerOfBinaryExpression = getInitializerOfBinaryExpression;
        function isPrototypePropertyAssignment(node) {
            return ts.isBinaryExpression(node) && getAssignmentDeclarationKind(node) === 3 /* PrototypeProperty */;
        }
        ts.isPrototypePropertyAssignment = isPrototypePropertyAssignment;
        function isSpecialPropertyDeclaration(expr) {
            return isInJSFile(expr) &&
                expr.parent && expr.parent.kind === 236 /* ExpressionStatement */ &&
                (!ts.isElementAccessExpression(expr) || isLiteralLikeElementAccess(expr)) &&
                !!ts.getJSDocTypeTag(expr.parent);
        }
        ts.isSpecialPropertyDeclaration = isSpecialPropertyDeclaration;
        function setValueDeclaration(symbol, node) {
            var valueDeclaration = symbol.valueDeclaration;
            if (!valueDeclaration ||
                !(node.flags & 8388608 /* Ambient */ && !(valueDeclaration.flags & 8388608 /* Ambient */)) &&
                    (isAssignmentDeclaration(valueDeclaration) && !isAssignmentDeclaration(node)) ||
                (valueDeclaration.kind !== node.kind && isEffectiveModuleDeclaration(valueDeclaration))) {
                // other kinds of value declarations take precedence over modules and assignment declarations
                symbol.valueDeclaration = node;
            }
        }
        ts.setValueDeclaration = setValueDeclaration;
        function isFunctionSymbol(symbol) {
            if (!symbol || !symbol.valueDeclaration) {
                return false;
            }
            var decl = symbol.valueDeclaration;
            return decl.kind === 254 /* FunctionDeclaration */ || ts.isVariableDeclaration(decl) && decl.initializer && ts.isFunctionLike(decl.initializer);
        }
        ts.isFunctionSymbol = isFunctionSymbol;
        function tryGetModuleSpecifierFromDeclaration(node) {
            var _a, _b, _c;
            switch (node.kind) {
                case 252 /* VariableDeclaration */:
                    return node.initializer.arguments[0].text;
                case 264 /* ImportDeclaration */:
                    return (_a = ts.tryCast(node.moduleSpecifier, ts.isStringLiteralLike)) === null || _a === void 0 ? void 0 : _a.text;
                case 263 /* ImportEqualsDeclaration */:
                    return (_c = ts.tryCast((_b = ts.tryCast(node.moduleReference, ts.isExternalModuleReference)) === null || _b === void 0 ? void 0 : _b.expression, ts.isStringLiteralLike)) === null || _c === void 0 ? void 0 : _c.text;
                default:
                    ts.Debug.assertNever(node);
            }
        }
        ts.tryGetModuleSpecifierFromDeclaration = tryGetModuleSpecifierFromDeclaration;
        function importFromModuleSpecifier(node) {
            return tryGetImportFromModuleSpecifier(node) || ts.Debug.failBadSyntaxKind(node.parent);
        }
        ts.importFromModuleSpecifier = importFromModuleSpecifier;
        function tryGetImportFromModuleSpecifier(node) {
            switch (node.parent.kind) {
                case 264 /* ImportDeclaration */:
                case 270 /* ExportDeclaration */:
                    return node.parent;
                case 275 /* ExternalModuleReference */:
                    return node.parent.parent;
                case 206 /* CallExpression */:
                    return isImportCall(node.parent) || isRequireCall(node.parent, /*checkArg*/ false) ? node.parent : undefined;
                case 194 /* LiteralType */:
                    ts.Debug.assert(ts.isStringLiteral(node));
                    return ts.tryCast(node.parent.parent, ts.isImportTypeNode);
                default:
                    return undefined;
            }
        }
        ts.tryGetImportFromModuleSpecifier = tryGetImportFromModuleSpecifier;
        function getExternalModuleName(node) {
            switch (node.kind) {
                case 264 /* ImportDeclaration */:
                case 270 /* ExportDeclaration */:
                    return node.moduleSpecifier;
                case 263 /* ImportEqualsDeclaration */:
                    return node.moduleReference.kind === 275 /* ExternalModuleReference */ ? node.moduleReference.expression : undefined;
                case 198 /* ImportType */:
                    return isLiteralImportTypeNode(node) ? node.argument.literal : undefined;
                case 206 /* CallExpression */:
                    return node.arguments[0];
                case 259 /* ModuleDeclaration */:
                    return node.name.kind === 10 /* StringLiteral */ ? node.name : undefined;
                default:
                    return ts.Debug.assertNever(node);
            }
        }
        ts.getExternalModuleName = getExternalModuleName;
        function getNamespaceDeclarationNode(node) {
            switch (node.kind) {
                case 264 /* ImportDeclaration */:
                    return node.importClause && ts.tryCast(node.importClause.namedBindings, ts.isNamespaceImport);
                case 263 /* ImportEqualsDeclaration */:
                    return node;
                case 270 /* ExportDeclaration */:
                    return node.exportClause && ts.tryCast(node.exportClause, ts.isNamespaceExport);
                default:
                    return ts.Debug.assertNever(node);
            }
        }
        ts.getNamespaceDeclarationNode = getNamespaceDeclarationNode;
        function isDefaultImport(node) {
            return node.kind === 264 /* ImportDeclaration */ && !!node.importClause && !!node.importClause.name;
        }
        ts.isDefaultImport = isDefaultImport;
        function forEachImportClauseDeclaration(node, action) {
            if (node.name) {
                var result = action(node);
                if (result)
                    return result;
            }
            if (node.namedBindings) {
                var result = ts.isNamespaceImport(node.namedBindings)
                    ? action(node.namedBindings)
                    : ts.forEach(node.namedBindings.elements, action);
                if (result)
                    return result;
            }
        }
        ts.forEachImportClauseDeclaration = forEachImportClauseDeclaration;
        function hasQuestionToken(node) {
            if (node) {
                switch (node.kind) {
                    case 162 /* Parameter */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 292 /* ShorthandPropertyAssignment */:
                    case 291 /* PropertyAssignment */:
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                        return node.questionToken !== undefined;
                }
            }
            return false;
        }
        ts.hasQuestionToken = hasQuestionToken;
        function isJSDocConstructSignature(node) {
            var param = ts.isJSDocFunctionType(node) ? ts.firstOrUndefined(node.parameters) : undefined;
            var name = ts.tryCast(param && param.name, ts.isIdentifier);
            return !!name && name.escapedText === "new";
        }
        ts.isJSDocConstructSignature = isJSDocConstructSignature;
        function isJSDocTypeAlias(node) {
            return node.kind === 340 /* JSDocTypedefTag */ || node.kind === 333 /* JSDocCallbackTag */ || node.kind === 334 /* JSDocEnumTag */;
        }
        ts.isJSDocTypeAlias = isJSDocTypeAlias;
        function isTypeAlias(node) {
            return isJSDocTypeAlias(node) || ts.isTypeAliasDeclaration(node);
        }
        ts.isTypeAlias = isTypeAlias;
        function getSourceOfAssignment(node) {
            return ts.isExpressionStatement(node) &&
                ts.isBinaryExpression(node.expression) &&
                node.expression.operatorToken.kind === 63 /* EqualsToken */
                ? getRightMostAssignedExpression(node.expression)
                : undefined;
        }
        function getSourceOfDefaultedAssignment(node) {
            return ts.isExpressionStatement(node) &&
                ts.isBinaryExpression(node.expression) &&
                getAssignmentDeclarationKind(node.expression) !== 0 /* None */ &&
                ts.isBinaryExpression(node.expression.right) &&
                (node.expression.right.operatorToken.kind === 56 /* BarBarToken */ || node.expression.right.operatorToken.kind === 60 /* QuestionQuestionToken */)
                ? node.expression.right.right
                : undefined;
        }
        function getSingleInitializerOfVariableStatementOrPropertyDeclaration(node) {
            switch (node.kind) {
                case 235 /* VariableStatement */:
                    var v = getSingleVariableOfVariableStatement(node);
                    return v && v.initializer;
                case 165 /* PropertyDeclaration */:
                    return node.initializer;
                case 291 /* PropertyAssignment */:
                    return node.initializer;
            }
        }
        ts.getSingleInitializerOfVariableStatementOrPropertyDeclaration = getSingleInitializerOfVariableStatementOrPropertyDeclaration;
        function getSingleVariableOfVariableStatement(node) {
            return ts.isVariableStatement(node) ? ts.firstOrUndefined(node.declarationList.declarations) : undefined;
        }
        ts.getSingleVariableOfVariableStatement = getSingleVariableOfVariableStatement;
        function getNestedModuleDeclaration(node) {
            return ts.isModuleDeclaration(node) &&
                node.body &&
                node.body.kind === 259 /* ModuleDeclaration */
                ? node.body
                : undefined;
        }
        function getJSDocCommentsAndTags(hostNode, noCache) {
            var result;
            // Pull parameter comments from declaring function as well
            if (isVariableLike(hostNode) && ts.hasInitializer(hostNode) && ts.hasJSDocNodes(hostNode.initializer)) {
                result = ts.append(result, ts.last(hostNode.initializer.jsDoc));
            }
            var node = hostNode;
            while (node && node.parent) {
                if (ts.hasJSDocNodes(node)) {
                    result = ts.append(result, ts.last(node.jsDoc));
                }
                if (node.kind === 162 /* Parameter */) {
                    result = ts.addRange(result, (noCache ? ts.getJSDocParameterTagsNoCache : ts.getJSDocParameterTags)(node));
                    break;
                }
                if (node.kind === 161 /* TypeParameter */) {
                    result = ts.addRange(result, (noCache ? ts.getJSDocTypeParameterTagsNoCache : ts.getJSDocTypeParameterTags)(node));
                    break;
                }
                node = getNextJSDocCommentLocation(node);
            }
            return result || ts.emptyArray;
        }
        ts.getJSDocCommentsAndTags = getJSDocCommentsAndTags;
        function getNextJSDocCommentLocation(node) {
            var parent = node.parent;
            if (parent.kind === 291 /* PropertyAssignment */ ||
                parent.kind === 269 /* ExportAssignment */ ||
                parent.kind === 165 /* PropertyDeclaration */ ||
                parent.kind === 236 /* ExpressionStatement */ && node.kind === 204 /* PropertyAccessExpression */ ||
                parent.kind === 245 /* ReturnStatement */ ||
                getNestedModuleDeclaration(parent) ||
                ts.isBinaryExpression(node) && node.operatorToken.kind === 63 /* EqualsToken */) {
                return parent;
            }
            // Try to recognize this pattern when node is initializer of variable declaration and JSDoc comments are on containing variable statement.
            // /**
            //   * @param {number} name
            //   * @returns {number}
            //   */
            // var x = function(name) { return name.length; }
            else if (parent.parent &&
                (getSingleVariableOfVariableStatement(parent.parent) === node ||
                    ts.isBinaryExpression(parent) && parent.operatorToken.kind === 63 /* EqualsToken */)) {
                return parent.parent;
            }
            else if (parent.parent && parent.parent.parent &&
                (getSingleVariableOfVariableStatement(parent.parent.parent) ||
                    getSingleInitializerOfVariableStatementOrPropertyDeclaration(parent.parent.parent) === node ||
                    getSourceOfDefaultedAssignment(parent.parent.parent))) {
                return parent.parent.parent;
            }
        }
        ts.getNextJSDocCommentLocation = getNextJSDocCommentLocation;
        /** Does the opposite of `getJSDocParameterTags`: given a JSDoc parameter, finds the parameter corresponding to it. */
        function getParameterSymbolFromJSDoc(node) {
            if (node.symbol) {
                return node.symbol;
            }
            if (!ts.isIdentifier(node.name)) {
                return undefined;
            }
            var name = node.name.escapedText;
            var decl = getHostSignatureFromJSDoc(node);
            if (!decl) {
                return undefined;
            }
            var parameter = ts.find(decl.parameters, function (p) { return p.name.kind === 79 /* Identifier */ && p.name.escapedText === name; });
            return parameter && parameter.symbol;
        }
        ts.getParameterSymbolFromJSDoc = getParameterSymbolFromJSDoc;
        function getHostSignatureFromJSDoc(node) {
            var host = getEffectiveJSDocHost(node);
            return host && ts.isFunctionLike(host) ? host : undefined;
        }
        ts.getHostSignatureFromJSDoc = getHostSignatureFromJSDoc;
        function getEffectiveJSDocHost(node) {
            var host = getJSDocHost(node);
            if (host) {
                return getSourceOfDefaultedAssignment(host)
                    || getSourceOfAssignment(host)
                    || getSingleInitializerOfVariableStatementOrPropertyDeclaration(host)
                    || getSingleVariableOfVariableStatement(host)
                    || getNestedModuleDeclaration(host)
                    || host;
            }
        }
        ts.getEffectiveJSDocHost = getEffectiveJSDocHost;
        /** Use getEffectiveJSDocHost if you additionally need to look for jsdoc on parent nodes, like assignments. */
        function getJSDocHost(node) {
            var jsDoc = getJSDocRoot(node);
            if (!jsDoc) {
                return undefined;
            }
            var host = jsDoc.parent;
            if (host && host.jsDoc && jsDoc === ts.lastOrUndefined(host.jsDoc)) {
                return host;
            }
        }
        ts.getJSDocHost = getJSDocHost;
        function getJSDocRoot(node) {
            return ts.findAncestor(node.parent, ts.isJSDoc);
        }
        ts.getJSDocRoot = getJSDocRoot;
        function getTypeParameterFromJsDoc(node) {
            var name = node.name.escapedText;
            var typeParameters = node.parent.parent.parent.typeParameters;
            return typeParameters && ts.find(typeParameters, function (p) { return p.name.escapedText === name; });
        }
        ts.getTypeParameterFromJsDoc = getTypeParameterFromJsDoc;
        function hasRestParameter(s) {
            var last = ts.lastOrUndefined(s.parameters);
            return !!last && isRestParameter(last);
        }
        ts.hasRestParameter = hasRestParameter;
        function isRestParameter(node) {
            var type = ts.isJSDocParameterTag(node) ? (node.typeExpression && node.typeExpression.type) : node.type;
            return node.dotDotDotToken !== undefined || !!type && type.kind === 313 /* JSDocVariadicType */;
        }
        ts.isRestParameter = isRestParameter;
        function hasTypeArguments(node) {
            return !!node.typeArguments;
        }
        ts.hasTypeArguments = hasTypeArguments;
        var AssignmentKind;
        (function (AssignmentKind) {
            AssignmentKind[AssignmentKind["None"] = 0] = "None";
            AssignmentKind[AssignmentKind["Definite"] = 1] = "Definite";
            AssignmentKind[AssignmentKind["Compound"] = 2] = "Compound";
        })(AssignmentKind = ts.AssignmentKind || (ts.AssignmentKind = {}));
        function getAssignmentTargetKind(node) {
            var parent = node.parent;
            while (true) {
                switch (parent.kind) {
                    case 219 /* BinaryExpression */:
                        var binaryOperator = parent.operatorToken.kind;
                        return isAssignmentOperator(binaryOperator) && parent.left === node ?
                            binaryOperator === 63 /* EqualsToken */ || isLogicalOrCoalescingAssignmentOperator(binaryOperator) ? 1 /* Definite */ : 2 /* Compound */ :
                            0 /* None */;
                    case 217 /* PrefixUnaryExpression */:
                    case 218 /* PostfixUnaryExpression */:
                        var unaryOperator = parent.operator;
                        return unaryOperator === 45 /* PlusPlusToken */ || unaryOperator === 46 /* MinusMinusToken */ ? 2 /* Compound */ : 0 /* None */;
                    case 241 /* ForInStatement */:
                    case 242 /* ForOfStatement */:
                        return parent.initializer === node ? 1 /* Definite */ : 0 /* None */;
                    case 210 /* ParenthesizedExpression */:
                    case 202 /* ArrayLiteralExpression */:
                    case 223 /* SpreadElement */:
                    case 228 /* NonNullExpression */:
                        node = parent;
                        break;
                    case 293 /* SpreadAssignment */:
                        node = parent.parent;
                        break;
                    case 292 /* ShorthandPropertyAssignment */:
                        if (parent.name !== node) {
                            return 0 /* None */;
                        }
                        node = parent.parent;
                        break;
                    case 291 /* PropertyAssignment */:
                        if (parent.name === node) {
                            return 0 /* None */;
                        }
                        node = parent.parent;
                        break;
                    default:
                        return 0 /* None */;
                }
                parent = node.parent;
            }
        }
        ts.getAssignmentTargetKind = getAssignmentTargetKind;
        // A node is an assignment target if it is on the left hand side of an '=' token, if it is parented by a property
        // assignment in an object literal that is an assignment target, or if it is parented by an array literal that is
        // an assignment target. Examples include 'a = xxx', '{ p: a } = xxx', '[{ a }] = xxx'.
        // (Note that `p` is not a target in the above examples, only `a`.)
        function isAssignmentTarget(node) {
            return getAssignmentTargetKind(node) !== 0 /* None */;
        }
        ts.isAssignmentTarget = isAssignmentTarget;
        /**
         * Indicates whether a node could contain a `var` VariableDeclarationList that contributes to
         * the same `var` declaration scope as the node's parent.
         */
        function isNodeWithPossibleHoistedDeclaration(node) {
            switch (node.kind) {
                case 233 /* Block */:
                case 235 /* VariableStatement */:
                case 246 /* WithStatement */:
                case 237 /* IfStatement */:
                case 247 /* SwitchStatement */:
                case 261 /* CaseBlock */:
                case 287 /* CaseClause */:
                case 288 /* DefaultClause */:
                case 248 /* LabeledStatement */:
                case 240 /* ForStatement */:
                case 241 /* ForInStatement */:
                case 242 /* ForOfStatement */:
                case 238 /* DoStatement */:
                case 239 /* WhileStatement */:
                case 250 /* TryStatement */:
                case 290 /* CatchClause */:
                    return true;
            }
            return false;
        }
        ts.isNodeWithPossibleHoistedDeclaration = isNodeWithPossibleHoistedDeclaration;
        function isValueSignatureDeclaration(node) {
            return ts.isFunctionExpression(node) || ts.isArrowFunction(node) || ts.isMethodOrAccessor(node) || ts.isFunctionDeclaration(node) || ts.isConstructorDeclaration(node);
        }
        ts.isValueSignatureDeclaration = isValueSignatureDeclaration;
        function walkUp(node, kind) {
            while (node && node.kind === kind) {
                node = node.parent;
            }
            return node;
        }
        function walkUpParenthesizedTypes(node) {
            return walkUp(node, 189 /* ParenthesizedType */);
        }
        ts.walkUpParenthesizedTypes = walkUpParenthesizedTypes;
        function walkUpParenthesizedExpressions(node) {
            return walkUp(node, 210 /* ParenthesizedExpression */);
        }
        ts.walkUpParenthesizedExpressions = walkUpParenthesizedExpressions;
        /**
         * Walks up parenthesized types.
         * It returns both the outermost parenthesized type and its parent.
         * If given node is not a parenthesiezd type, undefined is return as the former.
         */
        function walkUpParenthesizedTypesAndGetParentAndChild(node) {
            var child;
            while (node && node.kind === 189 /* ParenthesizedType */) {
                child = node;
                node = node.parent;
            }
            return [child, node];
        }
        ts.walkUpParenthesizedTypesAndGetParentAndChild = walkUpParenthesizedTypesAndGetParentAndChild;
        function skipParentheses(node) {
            return ts.skipOuterExpressions(node, 1 /* Parentheses */);
        }
        ts.skipParentheses = skipParentheses;
        // a node is delete target iff. it is PropertyAccessExpression/ElementAccessExpression with parentheses skipped
        function isDeleteTarget(node) {
            if (node.kind !== 204 /* PropertyAccessExpression */ && node.kind !== 205 /* ElementAccessExpression */) {
                return false;
            }
            node = walkUpParenthesizedExpressions(node.parent);
            return node && node.kind === 213 /* DeleteExpression */;
        }
        ts.isDeleteTarget = isDeleteTarget;
        function isNodeDescendantOf(node, ancestor) {
            while (node) {
                if (node === ancestor)
                    return true;
                node = node.parent;
            }
            return false;
        }
        ts.isNodeDescendantOf = isNodeDescendantOf;
        // True if `name` is the name of a declaration node
        function isDeclarationName(name) {
            return !ts.isSourceFile(name) && !ts.isBindingPattern(name) && ts.isDeclaration(name.parent) && name.parent.name === name;
        }
        ts.isDeclarationName = isDeclarationName;
        // See GH#16030
        function getDeclarationFromName(name) {
            var parent = name.parent;
            switch (name.kind) {
                case 10 /* StringLiteral */:
                case 14 /* NoSubstitutionTemplateLiteral */:
                case 8 /* NumericLiteral */:
                    if (ts.isComputedPropertyName(parent))
                        return parent.parent;
                // falls through
                case 79 /* Identifier */:
                    if (ts.isDeclaration(parent)) {
                        return parent.name === name ? parent : undefined;
                    }
                    else if (ts.isQualifiedName(parent)) {
                        var tag = parent.parent;
                        return ts.isJSDocParameterTag(tag) && tag.name === parent ? tag : undefined;
                    }
                    else {
                        var binExp = parent.parent;
                        return ts.isBinaryExpression(binExp) &&
                            getAssignmentDeclarationKind(binExp) !== 0 /* None */ &&
                            (binExp.left.symbol || binExp.symbol) &&
                            ts.getNameOfDeclaration(binExp) === name
                            ? binExp
                            : undefined;
                    }
                case 80 /* PrivateIdentifier */:
                    return ts.isDeclaration(parent) && parent.name === name ? parent : undefined;
                default:
                    return undefined;
            }
        }
        ts.getDeclarationFromName = getDeclarationFromName;
        function isLiteralComputedPropertyDeclarationName(node) {
            return isStringOrNumericLiteralLike(node) &&
                node.parent.kind === 160 /* ComputedPropertyName */ &&
                ts.isDeclaration(node.parent.parent);
        }
        ts.isLiteralComputedPropertyDeclarationName = isLiteralComputedPropertyDeclarationName;
        // Return true if the given identifier is classified as an IdentifierName
        function isIdentifierName(node) {
            var parent = node.parent;
            switch (parent.kind) {
                case 165 /* PropertyDeclaration */:
                case 164 /* PropertySignature */:
                case 167 /* MethodDeclaration */:
                case 166 /* MethodSignature */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 294 /* EnumMember */:
                case 291 /* PropertyAssignment */:
                case 204 /* PropertyAccessExpression */:
                    // Name in member declaration or property name in property access
                    return parent.name === node;
                case 159 /* QualifiedName */:
                    // Name on right hand side of dot in a type query or type reference
                    return parent.right === node;
                case 201 /* BindingElement */:
                case 268 /* ImportSpecifier */:
                    // Property name in binding element or import specifier
                    return parent.propertyName === node;
                case 273 /* ExportSpecifier */:
                case 283 /* JsxAttribute */:
                    // Any name in an export specifier or JSX Attribute
                    return true;
            }
            return false;
        }
        ts.isIdentifierName = isIdentifierName;
        // An alias symbol is created by one of the following declarations:
        // import <symbol> = ...
        // import <symbol> from ...
        // import * as <symbol> from ...
        // import { x as <symbol> } from ...
        // export { x as <symbol> } from ...
        // export * as ns <symbol> from ...
        // export = <EntityNameExpression>
        // export default <EntityNameExpression>
        // module.exports = <EntityNameExpression>
        // {<Identifier>}
        // {name: <EntityNameExpression>}
        function isAliasSymbolDeclaration(node) {
            return node.kind === 263 /* ImportEqualsDeclaration */ ||
                node.kind === 262 /* NamespaceExportDeclaration */ ||
                node.kind === 265 /* ImportClause */ && !!node.name ||
                node.kind === 266 /* NamespaceImport */ ||
                node.kind === 272 /* NamespaceExport */ ||
                node.kind === 268 /* ImportSpecifier */ ||
                node.kind === 273 /* ExportSpecifier */ ||
                node.kind === 269 /* ExportAssignment */ && exportAssignmentIsAlias(node) ||
                ts.isBinaryExpression(node) && getAssignmentDeclarationKind(node) === 2 /* ModuleExports */ && exportAssignmentIsAlias(node) ||
                ts.isPropertyAccessExpression(node) && ts.isBinaryExpression(node.parent) && node.parent.left === node && node.parent.operatorToken.kind === 63 /* EqualsToken */ && isAliasableExpression(node.parent.right) ||
                node.kind === 292 /* ShorthandPropertyAssignment */ ||
                node.kind === 291 /* PropertyAssignment */ && isAliasableExpression(node.initializer);
        }
        ts.isAliasSymbolDeclaration = isAliasSymbolDeclaration;
        function getAliasDeclarationFromName(node) {
            switch (node.parent.kind) {
                case 265 /* ImportClause */:
                case 268 /* ImportSpecifier */:
                case 266 /* NamespaceImport */:
                case 273 /* ExportSpecifier */:
                case 269 /* ExportAssignment */:
                case 263 /* ImportEqualsDeclaration */:
                    return node.parent;
                case 159 /* QualifiedName */:
                    do {
                        node = node.parent;
                    } while (node.parent.kind === 159 /* QualifiedName */);
                    return getAliasDeclarationFromName(node);
            }
        }
        ts.getAliasDeclarationFromName = getAliasDeclarationFromName;
        function isAliasableExpression(e) {
            return isEntityNameExpression(e) || ts.isClassExpression(e);
        }
        ts.isAliasableExpression = isAliasableExpression;
        function exportAssignmentIsAlias(node) {
            var e = getExportAssignmentExpression(node);
            return isAliasableExpression(e);
        }
        ts.exportAssignmentIsAlias = exportAssignmentIsAlias;
        function getExportAssignmentExpression(node) {
            return ts.isExportAssignment(node) ? node.expression : node.right;
        }
        ts.getExportAssignmentExpression = getExportAssignmentExpression;
        function getPropertyAssignmentAliasLikeExpression(node) {
            return node.kind === 292 /* ShorthandPropertyAssignment */ ? node.name : node.kind === 291 /* PropertyAssignment */ ? node.initializer :
                node.parent.right;
        }
        ts.getPropertyAssignmentAliasLikeExpression = getPropertyAssignmentAliasLikeExpression;
        function getEffectiveBaseTypeNode(node) {
            var baseType = getClassExtendsHeritageElement(node);
            if (baseType && isInJSFile(node)) {
                // Prefer an @augments tag because it may have type parameters.
                var tag = ts.getJSDocAugmentsTag(node);
                if (tag) {
                    return tag.class;
                }
            }
            return baseType;
        }
        ts.getEffectiveBaseTypeNode = getEffectiveBaseTypeNode;
        function getClassExtendsHeritageElement(node) {
            var heritageClause = getHeritageClause(node.heritageClauses, 94 /* ExtendsKeyword */);
            return heritageClause && heritageClause.types.length > 0 ? heritageClause.types[0] : undefined;
        }
        ts.getClassExtendsHeritageElement = getClassExtendsHeritageElement;
        function getEffectiveImplementsTypeNodes(node) {
            if (isInJSFile(node)) {
                return ts.getJSDocImplementsTags(node).map(function (n) { return n.class; });
            }
            else {
                var heritageClause = getHeritageClause(node.heritageClauses, 117 /* ImplementsKeyword */);
                return heritageClause === null || heritageClause === void 0 ? void 0 : heritageClause.types;
            }
        }
        ts.getEffectiveImplementsTypeNodes = getEffectiveImplementsTypeNodes;
        /** Returns the node in an `extends` or `implements` clause of a class or interface. */
        function getAllSuperTypeNodes(node) {
            return ts.isInterfaceDeclaration(node) ? getInterfaceBaseTypeNodes(node) || ts.emptyArray :
                ts.isClassLike(node) ? ts.concatenate(ts.singleElementArray(getEffectiveBaseTypeNode(node)), getEffectiveImplementsTypeNodes(node)) || ts.emptyArray :
                    ts.emptyArray;
        }
        ts.getAllSuperTypeNodes = getAllSuperTypeNodes;
        function getInterfaceBaseTypeNodes(node) {
            var heritageClause = getHeritageClause(node.heritageClauses, 94 /* ExtendsKeyword */);
            return heritageClause ? heritageClause.types : undefined;
        }
        ts.getInterfaceBaseTypeNodes = getInterfaceBaseTypeNodes;
        function getHeritageClause(clauses, kind) {
            if (clauses) {
                for (var _i = 0, clauses_1 = clauses; _i < clauses_1.length; _i++) {
                    var clause = clauses_1[_i];
                    if (clause.token === kind) {
                        return clause;
                    }
                }
            }
            return undefined;
        }
        ts.getHeritageClause = getHeritageClause;
        function getAncestor(node, kind) {
            while (node) {
                if (node.kind === kind) {
                    return node;
                }
                node = node.parent;
            }
            return undefined;
        }
        ts.getAncestor = getAncestor;
        function isKeyword(token) {
            return 81 /* FirstKeyword */ <= token && token <= 158 /* LastKeyword */;
        }
        ts.isKeyword = isKeyword;
        function isContextualKeyword(token) {
            return 126 /* FirstContextualKeyword */ <= token && token <= 158 /* LastContextualKeyword */;
        }
        ts.isContextualKeyword = isContextualKeyword;
        function isNonContextualKeyword(token) {
            return isKeyword(token) && !isContextualKeyword(token);
        }
        ts.isNonContextualKeyword = isNonContextualKeyword;
        function isFutureReservedKeyword(token) {
            return 117 /* FirstFutureReservedWord */ <= token && token <= 125 /* LastFutureReservedWord */;
        }
        ts.isFutureReservedKeyword = isFutureReservedKeyword;
        function isStringANonContextualKeyword(name) {
            var token = ts.stringToToken(name);
            return token !== undefined && isNonContextualKeyword(token);
        }
        ts.isStringANonContextualKeyword = isStringANonContextualKeyword;
        function isStringAKeyword(name) {
            var token = ts.stringToToken(name);
            return token !== undefined && isKeyword(token);
        }
        ts.isStringAKeyword = isStringAKeyword;
        function isIdentifierANonContextualKeyword(_a) {
            var originalKeywordKind = _a.originalKeywordKind;
            return !!originalKeywordKind && !isContextualKeyword(originalKeywordKind);
        }
        ts.isIdentifierANonContextualKeyword = isIdentifierANonContextualKeyword;
        function isTrivia(token) {
            return 2 /* FirstTriviaToken */ <= token && token <= 7 /* LastTriviaToken */;
        }
        ts.isTrivia = isTrivia;
        var FunctionFlags;
        (function (FunctionFlags) {
            FunctionFlags[FunctionFlags["Normal"] = 0] = "Normal";
            FunctionFlags[FunctionFlags["Generator"] = 1] = "Generator";
            FunctionFlags[FunctionFlags["Async"] = 2] = "Async";
            FunctionFlags[FunctionFlags["Invalid"] = 4] = "Invalid";
            FunctionFlags[FunctionFlags["AsyncGenerator"] = 3] = "AsyncGenerator";
        })(FunctionFlags = ts.FunctionFlags || (ts.FunctionFlags = {}));
        function getFunctionFlags(node) {
            if (!node) {
                return 4 /* Invalid */;
            }
            var flags = 0 /* Normal */;
            switch (node.kind) {
                case 254 /* FunctionDeclaration */:
                case 211 /* FunctionExpression */:
                case 167 /* MethodDeclaration */:
                    if (node.asteriskToken) {
                        flags |= 1 /* Generator */;
                    }
                // falls through
                case 212 /* ArrowFunction */:
                    if (hasSyntacticModifier(node, 256 /* Async */)) {
                        flags |= 2 /* Async */;
                    }
                    break;
            }
            if (!node.body) {
                flags |= 4 /* Invalid */;
            }
            return flags;
        }
        ts.getFunctionFlags = getFunctionFlags;
        function isAsyncFunction(node) {
            switch (node.kind) {
                case 254 /* FunctionDeclaration */:
                case 211 /* FunctionExpression */:
                case 212 /* ArrowFunction */:
                case 167 /* MethodDeclaration */:
                    return node.body !== undefined
                        && node.asteriskToken === undefined
                        && hasSyntacticModifier(node, 256 /* Async */);
            }
            return false;
        }
        ts.isAsyncFunction = isAsyncFunction;
        function isStringOrNumericLiteralLike(node) {
            return ts.isStringLiteralLike(node) || ts.isNumericLiteral(node);
        }
        ts.isStringOrNumericLiteralLike = isStringOrNumericLiteralLike;
        function isSignedNumericLiteral(node) {
            return ts.isPrefixUnaryExpression(node) && (node.operator === 39 /* PlusToken */ || node.operator === 40 /* MinusToken */) && ts.isNumericLiteral(node.operand);
        }
        ts.isSignedNumericLiteral = isSignedNumericLiteral;
        /**
         * A declaration has a dynamic name if all of the following are true:
         *   1. The declaration has a computed property name.
         *   2. The computed name is *not* expressed as a StringLiteral.
         *   3. The computed name is *not* expressed as a NumericLiteral.
         *   4. The computed name is *not* expressed as a PlusToken or MinusToken
         *      immediately followed by a NumericLiteral.
         */
        function hasDynamicName(declaration) {
            var name = ts.getNameOfDeclaration(declaration);
            return !!name && isDynamicName(name);
        }
        ts.hasDynamicName = hasDynamicName;
        function isDynamicName(name) {
            if (!(name.kind === 160 /* ComputedPropertyName */ || name.kind === 205 /* ElementAccessExpression */)) {
                return false;
            }
            var expr = ts.isElementAccessExpression(name) ? skipParentheses(name.argumentExpression) : name.expression;
            return !isStringOrNumericLiteralLike(expr) &&
                !isSignedNumericLiteral(expr);
        }
        ts.isDynamicName = isDynamicName;
        function getPropertyNameForPropertyNameNode(name) {
            switch (name.kind) {
                case 79 /* Identifier */:
                case 80 /* PrivateIdentifier */:
                    return name.escapedText;
                case 10 /* StringLiteral */:
                case 8 /* NumericLiteral */:
                    return ts.escapeLeadingUnderscores(name.text);
                case 160 /* ComputedPropertyName */:
                    var nameExpression = name.expression;
                    if (isStringOrNumericLiteralLike(nameExpression)) {
                        return ts.escapeLeadingUnderscores(nameExpression.text);
                    }
                    else if (isSignedNumericLiteral(nameExpression)) {
                        if (nameExpression.operator === 40 /* MinusToken */) {
                            return ts.tokenToString(nameExpression.operator) + nameExpression.operand.text;
                        }
                        return nameExpression.operand.text;
                    }
                    return undefined;
                default:
                    return ts.Debug.assertNever(name);
            }
        }
        ts.getPropertyNameForPropertyNameNode = getPropertyNameForPropertyNameNode;
        function isPropertyNameLiteral(node) {
            switch (node.kind) {
                case 79 /* Identifier */:
                case 10 /* StringLiteral */:
                case 14 /* NoSubstitutionTemplateLiteral */:
                case 8 /* NumericLiteral */:
                    return true;
                default:
                    return false;
            }
        }
        ts.isPropertyNameLiteral = isPropertyNameLiteral;
        function getTextOfIdentifierOrLiteral(node) {
            return ts.isMemberName(node) ? ts.idText(node) : node.text;
        }
        ts.getTextOfIdentifierOrLiteral = getTextOfIdentifierOrLiteral;
        function getEscapedTextOfIdentifierOrLiteral(node) {
            return ts.isMemberName(node) ? node.escapedText : ts.escapeLeadingUnderscores(node.text);
        }
        ts.getEscapedTextOfIdentifierOrLiteral = getEscapedTextOfIdentifierOrLiteral;
        function getPropertyNameForUniqueESSymbol(symbol) {
            return "__@" + ts.getSymbolId(symbol) + "@" + symbol.escapedName;
        }
        ts.getPropertyNameForUniqueESSymbol = getPropertyNameForUniqueESSymbol;
        function getSymbolNameForPrivateIdentifier(containingClassSymbol, description) {
            return "__#" + ts.getSymbolId(containingClassSymbol) + "@" + description;
        }
        ts.getSymbolNameForPrivateIdentifier = getSymbolNameForPrivateIdentifier;
        function isKnownSymbol(symbol) {
            return ts.startsWith(symbol.escapedName, "__@");
        }
        ts.isKnownSymbol = isKnownSymbol;
        function isPrivateIdentifierSymbol(symbol) {
            return ts.startsWith(symbol.escapedName, "__#");
        }
        ts.isPrivateIdentifierSymbol = isPrivateIdentifierSymbol;
        /**
         * Includes the word "Symbol" with unicode escapes
         */
        function isESSymbolIdentifier(node) {
            return node.kind === 79 /* Identifier */ && node.escapedText === "Symbol";
        }
        ts.isESSymbolIdentifier = isESSymbolIdentifier;
        function isPushOrUnshiftIdentifier(node) {
            return node.escapedText === "push" || node.escapedText === "unshift";
        }
        ts.isPushOrUnshiftIdentifier = isPushOrUnshiftIdentifier;
        function isParameterDeclaration(node) {
            var root = getRootDeclaration(node);
            return root.kind === 162 /* Parameter */;
        }
        ts.isParameterDeclaration = isParameterDeclaration;
        function getRootDeclaration(node) {
            while (node.kind === 201 /* BindingElement */) {
                node = node.parent.parent;
            }
            return node;
        }
        ts.getRootDeclaration = getRootDeclaration;
        function nodeStartsNewLexicalEnvironment(node) {
            var kind = node.kind;
            return kind === 169 /* Constructor */
                || kind === 211 /* FunctionExpression */
                || kind === 254 /* FunctionDeclaration */
                || kind === 212 /* ArrowFunction */
                || kind === 167 /* MethodDeclaration */
                || kind === 170 /* GetAccessor */
                || kind === 171 /* SetAccessor */
                || kind === 259 /* ModuleDeclaration */
                || kind === 300 /* SourceFile */;
        }
        ts.nodeStartsNewLexicalEnvironment = nodeStartsNewLexicalEnvironment;
        function nodeIsSynthesized(range) {
            return positionIsSynthesized(range.pos)
                || positionIsSynthesized(range.end);
        }
        ts.nodeIsSynthesized = nodeIsSynthesized;
        function getOriginalSourceFile(sourceFile) {
            return ts.getParseTreeNode(sourceFile, ts.isSourceFile) || sourceFile;
        }
        ts.getOriginalSourceFile = getOriginalSourceFile;
        var Associativity;
        (function (Associativity) {
            Associativity[Associativity["Left"] = 0] = "Left";
            Associativity[Associativity["Right"] = 1] = "Right";
        })(Associativity = ts.Associativity || (ts.Associativity = {}));
        function getExpressionAssociativity(expression) {
            var operator = getOperator(expression);
            var hasArguments = expression.kind === 207 /* NewExpression */ && expression.arguments !== undefined;
            return getOperatorAssociativity(expression.kind, operator, hasArguments);
        }
        ts.getExpressionAssociativity = getExpressionAssociativity;
        function getOperatorAssociativity(kind, operator, hasArguments) {
            switch (kind) {
                case 207 /* NewExpression */:
                    return hasArguments ? 0 /* Left */ : 1 /* Right */;
                case 217 /* PrefixUnaryExpression */:
                case 214 /* TypeOfExpression */:
                case 215 /* VoidExpression */:
                case 213 /* DeleteExpression */:
                case 216 /* AwaitExpression */:
                case 220 /* ConditionalExpression */:
                case 222 /* YieldExpression */:
                    return 1 /* Right */;
                case 219 /* BinaryExpression */:
                    switch (operator) {
                        case 42 /* AsteriskAsteriskToken */:
                        case 63 /* EqualsToken */:
                        case 64 /* PlusEqualsToken */:
                        case 65 /* MinusEqualsToken */:
                        case 67 /* AsteriskAsteriskEqualsToken */:
                        case 66 /* AsteriskEqualsToken */:
                        case 68 /* SlashEqualsToken */:
                        case 69 /* PercentEqualsToken */:
                        case 70 /* LessThanLessThanEqualsToken */:
                        case 71 /* GreaterThanGreaterThanEqualsToken */:
                        case 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */:
                        case 73 /* AmpersandEqualsToken */:
                        case 78 /* CaretEqualsToken */:
                        case 74 /* BarEqualsToken */:
                        case 75 /* BarBarEqualsToken */:
                        case 76 /* AmpersandAmpersandEqualsToken */:
                        case 77 /* QuestionQuestionEqualsToken */:
                            return 1 /* Right */;
                    }
            }
            return 0 /* Left */;
        }
        ts.getOperatorAssociativity = getOperatorAssociativity;
        function getExpressionPrecedence(expression) {
            var operator = getOperator(expression);
            var hasArguments = expression.kind === 207 /* NewExpression */ && expression.arguments !== undefined;
            return getOperatorPrecedence(expression.kind, operator, hasArguments);
        }
        ts.getExpressionPrecedence = getExpressionPrecedence;
        function getOperator(expression) {
            if (expression.kind === 219 /* BinaryExpression */) {
                return expression.operatorToken.kind;
            }
            else if (expression.kind === 217 /* PrefixUnaryExpression */ || expression.kind === 218 /* PostfixUnaryExpression */) {
                return expression.operator;
            }
            else {
                return expression.kind;
            }
        }
        ts.getOperator = getOperator;
        var OperatorPrecedence;
        (function (OperatorPrecedence) {
            // Expression:
            //     AssignmentExpression
            //     Expression `,` AssignmentExpression
            OperatorPrecedence[OperatorPrecedence["Comma"] = 0] = "Comma";
            // NOTE: `Spread` is higher than `Comma` due to how it is parsed in |ElementList|
            // SpreadElement:
            //     `...` AssignmentExpression
            OperatorPrecedence[OperatorPrecedence["Spread"] = 1] = "Spread";
            // AssignmentExpression:
            //     ConditionalExpression
            //     YieldExpression
            //     ArrowFunction
            //     AsyncArrowFunction
            //     LeftHandSideExpression `=` AssignmentExpression
            //     LeftHandSideExpression AssignmentOperator AssignmentExpression
            //
            // NOTE: AssignmentExpression is broken down into several precedences due to the requirements
            //       of the parenthesizer rules.
            // AssignmentExpression: YieldExpression
            // YieldExpression:
            //     `yield`
            //     `yield` AssignmentExpression
            //     `yield` `*` AssignmentExpression
            OperatorPrecedence[OperatorPrecedence["Yield"] = 2] = "Yield";
            // AssignmentExpression: LeftHandSideExpression `=` AssignmentExpression
            // AssignmentExpression: LeftHandSideExpression AssignmentOperator AssignmentExpression
            // AssignmentOperator: one of
            //     `*=` `/=` `%=` `+=` `-=` `<<=` `>>=` `>>>=` `&=` `^=` `|=` `**=`
            OperatorPrecedence[OperatorPrecedence["Assignment"] = 3] = "Assignment";
            // NOTE: `Conditional` is considered higher than `Assignment` here, but in reality they have
            //       the same precedence.
            // AssignmentExpression: ConditionalExpression
            // ConditionalExpression:
            //     ShortCircuitExpression
            //     ShortCircuitExpression `?` AssignmentExpression `:` AssignmentExpression
            // ShortCircuitExpression:
            //     LogicalORExpression
            //     CoalesceExpression
            OperatorPrecedence[OperatorPrecedence["Conditional"] = 4] = "Conditional";
            // CoalesceExpression:
            //     CoalesceExpressionHead `??` BitwiseORExpression
            // CoalesceExpressionHead:
            //     CoalesceExpression
            //     BitwiseORExpression
            OperatorPrecedence[OperatorPrecedence["Coalesce"] = 4] = "Coalesce";
            // LogicalORExpression:
            //     LogicalANDExpression
            //     LogicalORExpression `||` LogicalANDExpression
            OperatorPrecedence[OperatorPrecedence["LogicalOR"] = 5] = "LogicalOR";
            // LogicalANDExpression:
            //     BitwiseORExpression
            //     LogicalANDExprerssion `&&` BitwiseORExpression
            OperatorPrecedence[OperatorPrecedence["LogicalAND"] = 6] = "LogicalAND";
            // BitwiseORExpression:
            //     BitwiseXORExpression
            //     BitwiseORExpression `^` BitwiseXORExpression
            OperatorPrecedence[OperatorPrecedence["BitwiseOR"] = 7] = "BitwiseOR";
            // BitwiseXORExpression:
            //     BitwiseANDExpression
            //     BitwiseXORExpression `^` BitwiseANDExpression
            OperatorPrecedence[OperatorPrecedence["BitwiseXOR"] = 8] = "BitwiseXOR";
            // BitwiseANDExpression:
            //     EqualityExpression
            //     BitwiseANDExpression `^` EqualityExpression
            OperatorPrecedence[OperatorPrecedence["BitwiseAND"] = 9] = "BitwiseAND";
            // EqualityExpression:
            //     RelationalExpression
            //     EqualityExpression `==` RelationalExpression
            //     EqualityExpression `!=` RelationalExpression
            //     EqualityExpression `===` RelationalExpression
            //     EqualityExpression `!==` RelationalExpression
            OperatorPrecedence[OperatorPrecedence["Equality"] = 10] = "Equality";
            // RelationalExpression:
            //     ShiftExpression
            //     RelationalExpression `<` ShiftExpression
            //     RelationalExpression `>` ShiftExpression
            //     RelationalExpression `<=` ShiftExpression
            //     RelationalExpression `>=` ShiftExpression
            //     RelationalExpression `instanceof` ShiftExpression
            //     RelationalExpression `in` ShiftExpression
            //     [+TypeScript] RelationalExpression `as` Type
            OperatorPrecedence[OperatorPrecedence["Relational"] = 11] = "Relational";
            // ShiftExpression:
            //     AdditiveExpression
            //     ShiftExpression `<<` AdditiveExpression
            //     ShiftExpression `>>` AdditiveExpression
            //     ShiftExpression `>>>` AdditiveExpression
            OperatorPrecedence[OperatorPrecedence["Shift"] = 12] = "Shift";
            // AdditiveExpression:
            //     MultiplicativeExpression
            //     AdditiveExpression `+` MultiplicativeExpression
            //     AdditiveExpression `-` MultiplicativeExpression
            OperatorPrecedence[OperatorPrecedence["Additive"] = 13] = "Additive";
            // MultiplicativeExpression:
            //     ExponentiationExpression
            //     MultiplicativeExpression MultiplicativeOperator ExponentiationExpression
            // MultiplicativeOperator: one of `*`, `/`, `%`
            OperatorPrecedence[OperatorPrecedence["Multiplicative"] = 14] = "Multiplicative";
            // ExponentiationExpression:
            //     UnaryExpression
            //     UpdateExpression `**` ExponentiationExpression
            OperatorPrecedence[OperatorPrecedence["Exponentiation"] = 15] = "Exponentiation";
            // UnaryExpression:
            //     UpdateExpression
            //     `delete` UnaryExpression
            //     `void` UnaryExpression
            //     `typeof` UnaryExpression
            //     `+` UnaryExpression
            //     `-` UnaryExpression
            //     `~` UnaryExpression
            //     `!` UnaryExpression
            //     AwaitExpression
            // UpdateExpression:            // TODO: Do we need to investigate the precedence here?
            //     `++` UnaryExpression
            //     `--` UnaryExpression
            OperatorPrecedence[OperatorPrecedence["Unary"] = 16] = "Unary";
            // UpdateExpression:
            //     LeftHandSideExpression
            //     LeftHandSideExpression `++`
            //     LeftHandSideExpression `--`
            OperatorPrecedence[OperatorPrecedence["Update"] = 17] = "Update";
            // LeftHandSideExpression:
            //     NewExpression
            //     CallExpression
            // NewExpression:
            //     MemberExpression
            //     `new` NewExpression
            OperatorPrecedence[OperatorPrecedence["LeftHandSide"] = 18] = "LeftHandSide";
            // CallExpression:
            //     CoverCallExpressionAndAsyncArrowHead
            //     SuperCall
            //     ImportCall
            //     CallExpression Arguments
            //     CallExpression `[` Expression `]`
            //     CallExpression `.` IdentifierName
            //     CallExpression TemplateLiteral
            // MemberExpression:
            //     PrimaryExpression
            //     MemberExpression `[` Expression `]`
            //     MemberExpression `.` IdentifierName
            //     MemberExpression TemplateLiteral
            //     SuperProperty
            //     MetaProperty
            //     `new` MemberExpression Arguments
            OperatorPrecedence[OperatorPrecedence["Member"] = 19] = "Member";
            // TODO: JSXElement?
            // PrimaryExpression:
            //     `this`
            //     IdentifierReference
            //     Literal
            //     ArrayLiteral
            //     ObjectLiteral
            //     FunctionExpression
            //     ClassExpression
            //     GeneratorExpression
            //     AsyncFunctionExpression
            //     AsyncGeneratorExpression
            //     RegularExpressionLiteral
            //     TemplateLiteral
            //     CoverParenthesizedExpressionAndArrowParameterList
            OperatorPrecedence[OperatorPrecedence["Primary"] = 20] = "Primary";
            OperatorPrecedence[OperatorPrecedence["Highest"] = 20] = "Highest";
            OperatorPrecedence[OperatorPrecedence["Lowest"] = 0] = "Lowest";
            // -1 is lower than all other precedences. Returning it will cause binary expression
            // parsing to stop.
            OperatorPrecedence[OperatorPrecedence["Invalid"] = -1] = "Invalid";
        })(OperatorPrecedence = ts.OperatorPrecedence || (ts.OperatorPrecedence = {}));
        function getOperatorPrecedence(nodeKind, operatorKind, hasArguments) {
            switch (nodeKind) {
                case 346 /* CommaListExpression */:
                    return 0 /* Comma */;
                case 223 /* SpreadElement */:
                    return 1 /* Spread */;
                case 222 /* YieldExpression */:
                    return 2 /* Yield */;
                case 220 /* ConditionalExpression */:
                    return 4 /* Conditional */;
                case 219 /* BinaryExpression */:
                    switch (operatorKind) {
                        case 27 /* CommaToken */:
                            return 0 /* Comma */;
                        case 63 /* EqualsToken */:
                        case 64 /* PlusEqualsToken */:
                        case 65 /* MinusEqualsToken */:
                        case 67 /* AsteriskAsteriskEqualsToken */:
                        case 66 /* AsteriskEqualsToken */:
                        case 68 /* SlashEqualsToken */:
                        case 69 /* PercentEqualsToken */:
                        case 70 /* LessThanLessThanEqualsToken */:
                        case 71 /* GreaterThanGreaterThanEqualsToken */:
                        case 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */:
                        case 73 /* AmpersandEqualsToken */:
                        case 78 /* CaretEqualsToken */:
                        case 74 /* BarEqualsToken */:
                        case 75 /* BarBarEqualsToken */:
                        case 76 /* AmpersandAmpersandEqualsToken */:
                        case 77 /* QuestionQuestionEqualsToken */:
                            return 3 /* Assignment */;
                        default:
                            return getBinaryOperatorPrecedence(operatorKind);
                    }
                // TODO: Should prefix `++` and `--` be moved to the `Update` precedence?
                case 209 /* TypeAssertionExpression */:
                case 228 /* NonNullExpression */:
                case 217 /* PrefixUnaryExpression */:
                case 214 /* TypeOfExpression */:
                case 215 /* VoidExpression */:
                case 213 /* DeleteExpression */:
                case 216 /* AwaitExpression */:
                    return 16 /* Unary */;
                case 218 /* PostfixUnaryExpression */:
                    return 17 /* Update */;
                case 206 /* CallExpression */:
                    return 18 /* LeftHandSide */;
                case 207 /* NewExpression */:
                    return hasArguments ? 19 /* Member */ : 18 /* LeftHandSide */;
                case 208 /* TaggedTemplateExpression */:
                case 204 /* PropertyAccessExpression */:
                case 205 /* ElementAccessExpression */:
                case 229 /* MetaProperty */:
                    return 19 /* Member */;
                case 227 /* AsExpression */:
                    return 11 /* Relational */;
                case 108 /* ThisKeyword */:
                case 106 /* SuperKeyword */:
                case 79 /* Identifier */:
                case 104 /* NullKeyword */:
                case 110 /* TrueKeyword */:
                case 95 /* FalseKeyword */:
                case 8 /* NumericLiteral */:
                case 9 /* BigIntLiteral */:
                case 10 /* StringLiteral */:
                case 202 /* ArrayLiteralExpression */:
                case 203 /* ObjectLiteralExpression */:
                case 211 /* FunctionExpression */:
                case 212 /* ArrowFunction */:
                case 224 /* ClassExpression */:
                case 13 /* RegularExpressionLiteral */:
                case 14 /* NoSubstitutionTemplateLiteral */:
                case 221 /* TemplateExpression */:
                case 210 /* ParenthesizedExpression */:
                case 225 /* OmittedExpression */:
                case 276 /* JsxElement */:
                case 277 /* JsxSelfClosingElement */:
                case 280 /* JsxFragment */:
                    return 20 /* Primary */;
                default:
                    return -1 /* Invalid */;
            }
        }
        ts.getOperatorPrecedence = getOperatorPrecedence;
        function getBinaryOperatorPrecedence(kind) {
            switch (kind) {
                case 60 /* QuestionQuestionToken */:
                    return 4 /* Coalesce */;
                case 56 /* BarBarToken */:
                    return 5 /* LogicalOR */;
                case 55 /* AmpersandAmpersandToken */:
                    return 6 /* LogicalAND */;
                case 51 /* BarToken */:
                    return 7 /* BitwiseOR */;
                case 52 /* CaretToken */:
                    return 8 /* BitwiseXOR */;
                case 50 /* AmpersandToken */:
                    return 9 /* BitwiseAND */;
                case 34 /* EqualsEqualsToken */:
                case 35 /* ExclamationEqualsToken */:
                case 36 /* EqualsEqualsEqualsToken */:
                case 37 /* ExclamationEqualsEqualsToken */:
                    return 10 /* Equality */;
                case 29 /* LessThanToken */:
                case 31 /* GreaterThanToken */:
                case 32 /* LessThanEqualsToken */:
                case 33 /* GreaterThanEqualsToken */:
                case 102 /* InstanceOfKeyword */:
                case 101 /* InKeyword */:
                case 127 /* AsKeyword */:
                    return 11 /* Relational */;
                case 47 /* LessThanLessThanToken */:
                case 48 /* GreaterThanGreaterThanToken */:
                case 49 /* GreaterThanGreaterThanGreaterThanToken */:
                    return 12 /* Shift */;
                case 39 /* PlusToken */:
                case 40 /* MinusToken */:
                    return 13 /* Additive */;
                case 41 /* AsteriskToken */:
                case 43 /* SlashToken */:
                case 44 /* PercentToken */:
                    return 14 /* Multiplicative */;
                case 42 /* AsteriskAsteriskToken */:
                    return 15 /* Exponentiation */;
            }
            // -1 is lower than all other precedences.  Returning it will cause binary expression
            // parsing to stop.
            return -1;
        }
        ts.getBinaryOperatorPrecedence = getBinaryOperatorPrecedence;
        function getSemanticJsxChildren(children) {
            return ts.filter(children, function (i) {
                switch (i.kind) {
                    case 286 /* JsxExpression */:
                        return !!i.expression;
                    case 11 /* JsxText */:
                        return !i.containsOnlyTriviaWhiteSpaces;
                    default:
                        return true;
                }
            });
        }
        ts.getSemanticJsxChildren = getSemanticJsxChildren;
        function createDiagnosticCollection() {
            var nonFileDiagnostics = []; // See GH#19873
            var filesWithDiagnostics = [];
            var fileDiagnostics = new ts.Map();
            var hasReadNonFileDiagnostics = false;
            return {
                add: add,
                lookup: lookup,
                getGlobalDiagnostics: getGlobalDiagnostics,
                getDiagnostics: getDiagnostics,
            };
            function lookup(diagnostic) {
                var diagnostics;
                if (diagnostic.file) {
                    diagnostics = fileDiagnostics.get(diagnostic.file.fileName);
                }
                else {
                    diagnostics = nonFileDiagnostics;
                }
                if (!diagnostics) {
                    return undefined;
                }
                var result = ts.binarySearch(diagnostics, diagnostic, ts.identity, compareDiagnosticsSkipRelatedInformation);
                if (result >= 0) {
                    return diagnostics[result];
                }
                return undefined;
            }
            function add(diagnostic) {
                var diagnostics;
                if (diagnostic.file) {
                    diagnostics = fileDiagnostics.get(diagnostic.file.fileName);
                    if (!diagnostics) {
                        diagnostics = []; // See GH#19873
                        fileDiagnostics.set(diagnostic.file.fileName, diagnostics);
                        ts.insertSorted(filesWithDiagnostics, diagnostic.file.fileName, ts.compareStringsCaseSensitive);
                    }
                }
                else {
                    // If we've already read the non-file diagnostics, do not modify the existing array.
                    if (hasReadNonFileDiagnostics) {
                        hasReadNonFileDiagnostics = false;
                        nonFileDiagnostics = nonFileDiagnostics.slice();
                    }
                    diagnostics = nonFileDiagnostics;
                }
                ts.insertSorted(diagnostics, diagnostic, compareDiagnostics);
            }
            function getGlobalDiagnostics() {
                hasReadNonFileDiagnostics = true;
                return nonFileDiagnostics;
            }
            function getDiagnostics(fileName) {
                if (fileName) {
                    return fileDiagnostics.get(fileName) || [];
                }
                var fileDiags = ts.flatMapToMutable(filesWithDiagnostics, function (f) { return fileDiagnostics.get(f); });
                if (!nonFileDiagnostics.length) {
                    return fileDiags;
                }
                fileDiags.unshift.apply(fileDiags, nonFileDiagnostics);
                return fileDiags;
            }
        }
        ts.createDiagnosticCollection = createDiagnosticCollection;
        var templateSubstitutionRegExp = /\$\{/g;
        function escapeTemplateSubstitution(str) {
            return str.replace(templateSubstitutionRegExp, "\\${");
        }
        /** @internal */
        function hasInvalidEscape(template) {
            return template && !!(ts.isNoSubstitutionTemplateLiteral(template)
                ? template.templateFlags
                : (template.head.templateFlags || ts.some(template.templateSpans, function (span) { return !!span.literal.templateFlags; })));
        }
        ts.hasInvalidEscape = hasInvalidEscape;
        // This consists of the first 19 unprintable ASCII characters, canonical escapes, lineSeparator,
        // paragraphSeparator, and nextLine. The latter three are just desirable to suppress new lines in
        // the language service. These characters should be escaped when printing, and if any characters are added,
        // the map below must be updated. Note that this regexp *does not* include the 'delete' character.
        // There is no reason for this other than that JSON.stringify does not handle it either.
        var doubleQuoteEscapedCharsRegExp = /[\\\"\u0000-\u001f\t\v\f\b\r\n\u2028\u2029\u0085]/g;
        var singleQuoteEscapedCharsRegExp = /[\\\'\u0000-\u001f\t\v\f\b\r\n\u2028\u2029\u0085]/g;
        // Template strings preserve simple LF newlines, still encode CRLF (or CR)
        var backtickQuoteEscapedCharsRegExp = /\r\n|[\\\`\u0000-\u001f\t\v\f\b\r\u2028\u2029\u0085]/g;
        var escapedCharsMap = new ts.Map(ts.getEntries({
            "\t": "\\t",
            "\v": "\\v",
            "\f": "\\f",
            "\b": "\\b",
            "\r": "\\r",
            "\n": "\\n",
            "\\": "\\\\",
            "\"": "\\\"",
            "\'": "\\\'",
            "\`": "\\\`",
            "\u2028": "\\u2028",
            "\u2029": "\\u2029",
            "\u0085": "\\u0085",
            "\r\n": "\\r\\n", // special case for CRLFs in backticks
        }));
        function encodeUtf16EscapeSequence(charCode) {
            var hexCharCode = charCode.toString(16).toUpperCase();
            var paddedHexCode = ("0000" + hexCharCode).slice(-4);
            return "\\u" + paddedHexCode;
        }
        function getReplacement(c, offset, input) {
            if (c.charCodeAt(0) === 0 /* nullCharacter */) {
                var lookAhead = input.charCodeAt(offset + c.length);
                if (lookAhead >= 48 /* _0 */ && lookAhead <= 57 /* _9 */) {
                    // If the null character is followed by digits, print as a hex escape to prevent the result from parsing as an octal (which is forbidden in strict mode)
                    return "\\x00";
                }
                // Otherwise, keep printing a literal \0 for the null character
                return "\\0";
            }
            return escapedCharsMap.get(c) || encodeUtf16EscapeSequence(c.charCodeAt(0));
        }
        /**
         * Based heavily on the abstract 'Quote'/'QuoteJSONString' operation from ECMA-262 (24.3.2.2),
         * but augmented for a few select characters (e.g. lineSeparator, paragraphSeparator, nextLine)
         * Note that this doesn't actually wrap the input in double quotes.
         */
        function escapeString(s, quoteChar) {
            var escapedCharsRegExp = quoteChar === 96 /* backtick */ ? backtickQuoteEscapedCharsRegExp :
                quoteChar === 39 /* singleQuote */ ? singleQuoteEscapedCharsRegExp :
                    doubleQuoteEscapedCharsRegExp;
            return s.replace(escapedCharsRegExp, getReplacement);
        }
        ts.escapeString = escapeString;
        var nonAsciiCharacters = /[^\u0000-\u007F]/g;
        function escapeNonAsciiString(s, quoteChar) {
            s = escapeString(s, quoteChar);
            // Replace non-ASCII characters with '\uNNNN' escapes if any exist.
            // Otherwise just return the original string.
            return nonAsciiCharacters.test(s) ?
                s.replace(nonAsciiCharacters, function (c) { return encodeUtf16EscapeSequence(c.charCodeAt(0)); }) :
                s;
        }
        ts.escapeNonAsciiString = escapeNonAsciiString;
        // This consists of the first 19 unprintable ASCII characters, JSX canonical escapes, lineSeparator,
        // paragraphSeparator, and nextLine. The latter three are just desirable to suppress new lines in
        // the language service. These characters should be escaped when printing, and if any characters are added,
        // the map below must be updated.
        var jsxDoubleQuoteEscapedCharsRegExp = /[\"\u0000-\u001f\u2028\u2029\u0085]/g;
        var jsxSingleQuoteEscapedCharsRegExp = /[\'\u0000-\u001f\u2028\u2029\u0085]/g;
        var jsxEscapedCharsMap = new ts.Map(ts.getEntries({
            "\"": "&quot;",
            "\'": "&apos;"
        }));
        function encodeJsxCharacterEntity(charCode) {
            var hexCharCode = charCode.toString(16).toUpperCase();
            return "&#x" + hexCharCode + ";";
        }
        function getJsxAttributeStringReplacement(c) {
            if (c.charCodeAt(0) === 0 /* nullCharacter */) {
                return "&#0;";
            }
            return jsxEscapedCharsMap.get(c) || encodeJsxCharacterEntity(c.charCodeAt(0));
        }
        function escapeJsxAttributeString(s, quoteChar) {
            var escapedCharsRegExp = quoteChar === 39 /* singleQuote */ ? jsxSingleQuoteEscapedCharsRegExp :
                jsxDoubleQuoteEscapedCharsRegExp;
            return s.replace(escapedCharsRegExp, getJsxAttributeStringReplacement);
        }
        ts.escapeJsxAttributeString = escapeJsxAttributeString;
        /**
         * Strip off existed surrounding single quotes, double quotes, or backticks from a given string
         *
         * @return non-quoted string
         */
        function stripQuotes(name) {
            var length = name.length;
            if (length >= 2 && name.charCodeAt(0) === name.charCodeAt(length - 1) && isQuoteOrBacktick(name.charCodeAt(0))) {
                return name.substring(1, length - 1);
            }
            return name;
        }
        ts.stripQuotes = stripQuotes;
        function isQuoteOrBacktick(charCode) {
            return charCode === 39 /* singleQuote */ ||
                charCode === 34 /* doubleQuote */ ||
                charCode === 96 /* backtick */;
        }
        function isIntrinsicJsxName(name) {
            var ch = name.charCodeAt(0);
            return (ch >= 97 /* a */ && ch <= 122 /* z */) || ts.stringContains(name, "-") || ts.stringContains(name, ":");
        }
        ts.isIntrinsicJsxName = isIntrinsicJsxName;
        var indentStrings = ["", "    "];
        function getIndentString(level) {
            // prepopulate cache
            var singleLevel = indentStrings[1];
            for (var current = indentStrings.length; current <= level; current++) {
                indentStrings.push(indentStrings[current - 1] + singleLevel);
            }
            return indentStrings[level];
        }
        ts.getIndentString = getIndentString;
        function getIndentSize() {
            return indentStrings[1].length;
        }
        ts.getIndentSize = getIndentSize;
        function createTextWriter(newLine) {
            var output;
            var indent;
            var lineStart;
            var lineCount;
            var linePos;
            var hasTrailingComment = false;
            function updateLineCountAndPosFor(s) {
                var lineStartsOfS = ts.computeLineStarts(s);
                if (lineStartsOfS.length > 1) {
                    lineCount = lineCount + lineStartsOfS.length - 1;
                    linePos = output.length - s.length + ts.last(lineStartsOfS);
                    lineStart = (linePos - output.length) === 0;
                }
                else {
                    lineStart = false;
                }
            }
            function writeText(s) {
                if (s && s.length) {
                    if (lineStart) {
                        s = getIndentString(indent) + s;
                        lineStart = false;
                    }
                    output += s;
                    updateLineCountAndPosFor(s);
                }
            }
            function write(s) {
                if (s)
                    hasTrailingComment = false;
                writeText(s);
            }
            function writeComment(s) {
                if (s)
                    hasTrailingComment = true;
                writeText(s);
            }
            function reset() {
                output = "";
                indent = 0;
                lineStart = true;
                lineCount = 0;
                linePos = 0;
                hasTrailingComment = false;
            }
            function rawWrite(s) {
                if (s !== undefined) {
                    output += s;
                    updateLineCountAndPosFor(s);
                    hasTrailingComment = false;
                }
            }
            function writeLiteral(s) {
                if (s && s.length) {
                    write(s);
                }
            }
            function writeLine(force) {
                if (!lineStart || force) {
                    output += newLine;
                    lineCount++;
                    linePos = output.length;
                    lineStart = true;
                    hasTrailingComment = false;
                }
            }
            function getTextPosWithWriteLine() {
                return lineStart ? output.length : (output.length + newLine.length);
            }
            reset();
            return {
                write: write,
                rawWrite: rawWrite,
                writeLiteral: writeLiteral,
                writeLine: writeLine,
                increaseIndent: function () { indent++; },
                decreaseIndent: function () { indent--; },
                getIndent: function () { return indent; },
                getTextPos: function () { return output.length; },
                getLine: function () { return lineCount; },
                getColumn: function () { return lineStart ? indent * getIndentSize() : output.length - linePos; },
                getText: function () { return output; },
                isAtStartOfLine: function () { return lineStart; },
                hasTrailingComment: function () { return hasTrailingComment; },
                hasTrailingWhitespace: function () { return !!output.length && ts.isWhiteSpaceLike(output.charCodeAt(output.length - 1)); },
                clear: reset,
                reportInaccessibleThisError: ts.noop,
                reportPrivateInBaseOfClassExpression: ts.noop,
                reportInaccessibleUniqueSymbolError: ts.noop,
                trackSymbol: function () { return false; },
                writeKeyword: write,
                writeOperator: write,
                writeParameter: write,
                writeProperty: write,
                writePunctuation: write,
                writeSpace: write,
                writeStringLiteral: write,
                writeSymbol: function (s, _) { return write(s); },
                writeTrailingSemicolon: write,
                writeComment: writeComment,
                getTextPosWithWriteLine: getTextPosWithWriteLine
            };
        }
        ts.createTextWriter = createTextWriter;
        function getTrailingSemicolonDeferringWriter(writer) {
            var pendingTrailingSemicolon = false;
            function commitPendingTrailingSemicolon() {
                if (pendingTrailingSemicolon) {
                    writer.writeTrailingSemicolon(";");
                    pendingTrailingSemicolon = false;
                }
            }
            return __assign(__assign({}, writer), { writeTrailingSemicolon: function () {
                    pendingTrailingSemicolon = true;
                }, writeLiteral: function (s) {
                    commitPendingTrailingSemicolon();
                    writer.writeLiteral(s);
                }, writeStringLiteral: function (s) {
                    commitPendingTrailingSemicolon();
                    writer.writeStringLiteral(s);
                }, writeSymbol: function (s, sym) {
                    commitPendingTrailingSemicolon();
                    writer.writeSymbol(s, sym);
                }, writePunctuation: function (s) {
                    commitPendingTrailingSemicolon();
                    writer.writePunctuation(s);
                }, writeKeyword: function (s) {
                    commitPendingTrailingSemicolon();
                    writer.writeKeyword(s);
                }, writeOperator: function (s) {
                    commitPendingTrailingSemicolon();
                    writer.writeOperator(s);
                }, writeParameter: function (s) {
                    commitPendingTrailingSemicolon();
                    writer.writeParameter(s);
                }, writeSpace: function (s) {
                    commitPendingTrailingSemicolon();
                    writer.writeSpace(s);
                }, writeProperty: function (s) {
                    commitPendingTrailingSemicolon();
                    writer.writeProperty(s);
                }, writeComment: function (s) {
                    commitPendingTrailingSemicolon();
                    writer.writeComment(s);
                }, writeLine: function () {
                    commitPendingTrailingSemicolon();
                    writer.writeLine();
                }, increaseIndent: function () {
                    commitPendingTrailingSemicolon();
                    writer.increaseIndent();
                }, decreaseIndent: function () {
                    commitPendingTrailingSemicolon();
                    writer.decreaseIndent();
                } });
        }
        ts.getTrailingSemicolonDeferringWriter = getTrailingSemicolonDeferringWriter;
        function hostUsesCaseSensitiveFileNames(host) {
            return host.useCaseSensitiveFileNames ? host.useCaseSensitiveFileNames() : false;
        }
        ts.hostUsesCaseSensitiveFileNames = hostUsesCaseSensitiveFileNames;
        function hostGetCanonicalFileName(host) {
            return ts.createGetCanonicalFileName(hostUsesCaseSensitiveFileNames(host));
        }
        ts.hostGetCanonicalFileName = hostGetCanonicalFileName;
        function getResolvedExternalModuleName(host, file, referenceFile) {
            return file.moduleName || getExternalModuleNameFromPath(host, file.fileName, referenceFile && referenceFile.fileName);
        }
        ts.getResolvedExternalModuleName = getResolvedExternalModuleName;
        function getCanonicalAbsolutePath(host, path) {
            return host.getCanonicalFileName(ts.getNormalizedAbsolutePath(path, host.getCurrentDirectory()));
        }
        function getExternalModuleNameFromDeclaration(host, resolver, declaration) {
            var file = resolver.getExternalModuleFileFromDeclaration(declaration);
            if (!file || file.isDeclarationFile) {
                return undefined;
            }
            // If the declaration already uses a non-relative name, and is outside the common source directory, continue to use it
            var specifier = getExternalModuleName(declaration);
            if (specifier && ts.isStringLiteralLike(specifier) && !ts.pathIsRelative(specifier.text) &&
                getCanonicalAbsolutePath(host, file.path).indexOf(getCanonicalAbsolutePath(host, ts.ensureTrailingDirectorySeparator(host.getCommonSourceDirectory()))) === -1) {
                return undefined;
            }
            return getResolvedExternalModuleName(host, file);
        }
        ts.getExternalModuleNameFromDeclaration = getExternalModuleNameFromDeclaration;
        /**
         * Resolves a local path to a path which is absolute to the base of the emit
         */
        function getExternalModuleNameFromPath(host, fileName, referencePath) {
            var getCanonicalFileName = function (f) { return host.getCanonicalFileName(f); };
            var dir = ts.toPath(referencePath ? ts.getDirectoryPath(referencePath) : host.getCommonSourceDirectory(), host.getCurrentDirectory(), getCanonicalFileName);
            var filePath = ts.getNormalizedAbsolutePath(fileName, host.getCurrentDirectory());
            var relativePath = ts.getRelativePathToDirectoryOrUrl(dir, filePath, dir, getCanonicalFileName, /*isAbsolutePathAnUrl*/ false);
            var extensionless = removeFileExtension(relativePath);
            return referencePath ? ts.ensurePathIsNonModuleName(extensionless) : extensionless;
        }
        ts.getExternalModuleNameFromPath = getExternalModuleNameFromPath;
        function getOwnEmitOutputFilePath(fileName, host, extension) {
            var compilerOptions = host.getCompilerOptions();
            var emitOutputFilePathWithoutExtension;
            if (compilerOptions.outDir) {
                emitOutputFilePathWithoutExtension = removeFileExtension(getSourceFilePathInNewDir(fileName, host, compilerOptions.outDir));
            }
            else {
                emitOutputFilePathWithoutExtension = removeFileExtension(fileName);
            }
            return emitOutputFilePathWithoutExtension + extension;
        }
        ts.getOwnEmitOutputFilePath = getOwnEmitOutputFilePath;
        function getDeclarationEmitOutputFilePath(fileName, host) {
            return getDeclarationEmitOutputFilePathWorker(fileName, host.getCompilerOptions(), host.getCurrentDirectory(), host.getCommonSourceDirectory(), function (f) { return host.getCanonicalFileName(f); });
        }
        ts.getDeclarationEmitOutputFilePath = getDeclarationEmitOutputFilePath;
        function getDeclarationEmitOutputFilePathWorker(fileName, options, currentDirectory, commonSourceDirectory, getCanonicalFileName) {
            var outputDir = options.declarationDir || options.outDir; // Prefer declaration folder if specified
            var path = outputDir
                ? getSourceFilePathInNewDirWorker(fileName, outputDir, currentDirectory, commonSourceDirectory, getCanonicalFileName)
                : fileName;
            return removeFileExtension(path) + ".d.ts" /* Dts */;
        }
        ts.getDeclarationEmitOutputFilePathWorker = getDeclarationEmitOutputFilePathWorker;
        function outFile(options) {
            return options.outFile || options.out;
        }
        ts.outFile = outFile;
        /** Returns 'undefined' if and only if 'options.paths' is undefined. */
        function getPathsBasePath(options, host) {
            var _a, _b;
            if (!options.paths)
                return undefined;
            return (_a = options.baseUrl) !== null && _a !== void 0 ? _a : ts.Debug.checkDefined(options.pathsBasePath || ((_b = host.getCurrentDirectory) === null || _b === void 0 ? void 0 : _b.call(host)), "Encountered 'paths' without a 'baseUrl', config file, or host 'getCurrentDirectory'.");
        }
        ts.getPathsBasePath = getPathsBasePath;
        /**
         * Gets the source files that are expected to have an emit output.
         *
         * Originally part of `forEachExpectedEmitFile`, this functionality was extracted to support
         * transformations.
         *
         * @param host An EmitHost.
         * @param targetSourceFile An optional target source file to emit.
         */
        function getSourceFilesToEmit(host, targetSourceFile, forceDtsEmit) {
            var options = host.getCompilerOptions();
            if (outFile(options)) {
                var moduleKind = getEmitModuleKind(options);
                var moduleEmitEnabled_1 = options.emitDeclarationOnly || moduleKind === ts.ModuleKind.AMD || moduleKind === ts.ModuleKind.System;
                // Can emit only sources that are not declaration file and are either non module code or module with --module or --target es6 specified
                return ts.filter(host.getSourceFiles(), function (sourceFile) {
                    return (moduleEmitEnabled_1 || !ts.isExternalModule(sourceFile)) &&
                        sourceFileMayBeEmitted(sourceFile, host, forceDtsEmit);
                });
            }
            else {
                var sourceFiles = targetSourceFile === undefined ? host.getSourceFiles() : [targetSourceFile];
                return ts.filter(sourceFiles, function (sourceFile) { return sourceFileMayBeEmitted(sourceFile, host, forceDtsEmit); });
            }
        }
        ts.getSourceFilesToEmit = getSourceFilesToEmit;
        /** Don't call this for `--outFile`, just for `--outDir` or plain emit. `--outFile` needs additional checks. */
        function sourceFileMayBeEmitted(sourceFile, host, forceDtsEmit) {
            var options = host.getCompilerOptions();
            return !(options.noEmitForJsFiles && isSourceFileJS(sourceFile)) &&
                !sourceFile.isDeclarationFile &&
                !host.isSourceFileFromExternalLibrary(sourceFile) &&
                (forceDtsEmit || (!(isJsonSourceFile(sourceFile) && host.getResolvedProjectReferenceToRedirect(sourceFile.fileName)) &&
                    !host.isSourceOfProjectReferenceRedirect(sourceFile.fileName)));
        }
        ts.sourceFileMayBeEmitted = sourceFileMayBeEmitted;
        function getSourceFilePathInNewDir(fileName, host, newDirPath) {
            return getSourceFilePathInNewDirWorker(fileName, newDirPath, host.getCurrentDirectory(), host.getCommonSourceDirectory(), function (f) { return host.getCanonicalFileName(f); });
        }
        ts.getSourceFilePathInNewDir = getSourceFilePathInNewDir;
        function getSourceFilePathInNewDirWorker(fileName, newDirPath, currentDirectory, commonSourceDirectory, getCanonicalFileName) {
            var sourceFilePath = ts.getNormalizedAbsolutePath(fileName, currentDirectory);
            var isSourceFileInCommonSourceDirectory = getCanonicalFileName(sourceFilePath).indexOf(getCanonicalFileName(commonSourceDirectory)) === 0;
            sourceFilePath = isSourceFileInCommonSourceDirectory ? sourceFilePath.substring(commonSourceDirectory.length) : sourceFilePath;
            return ts.combinePaths(newDirPath, sourceFilePath);
        }
        ts.getSourceFilePathInNewDirWorker = getSourceFilePathInNewDirWorker;
        function writeFile(host, diagnostics, fileName, data, writeByteOrderMark, sourceFiles) {
            host.writeFile(fileName, data, writeByteOrderMark, function (hostErrorMessage) {
                diagnostics.add(createCompilerDiagnostic(ts.Diagnostics.Could_not_write_file_0_Colon_1, fileName, hostErrorMessage));
            }, sourceFiles);
        }
        ts.writeFile = writeFile;
        function ensureDirectoriesExist(directoryPath, createDirectory, directoryExists) {
            if (directoryPath.length > ts.getRootLength(directoryPath) && !directoryExists(directoryPath)) {
                var parentDirectory = ts.getDirectoryPath(directoryPath);
                ensureDirectoriesExist(parentDirectory, createDirectory, directoryExists);
                createDirectory(directoryPath);
            }
        }
        function writeFileEnsuringDirectories(path, data, writeByteOrderMark, writeFile, createDirectory, directoryExists) {
            // PERF: Checking for directory existence is expensive.  Instead, assume the directory exists
            // and fall back to creating it if the file write fails.
            try {
                writeFile(path, data, writeByteOrderMark);
            }
            catch (_a) {
                ensureDirectoriesExist(ts.getDirectoryPath(ts.normalizePath(path)), createDirectory, directoryExists);
                writeFile(path, data, writeByteOrderMark);
            }
        }
        ts.writeFileEnsuringDirectories = writeFileEnsuringDirectories;
        function getLineOfLocalPosition(sourceFile, pos) {
            var lineStarts = ts.getLineStarts(sourceFile);
            return ts.computeLineOfPosition(lineStarts, pos);
        }
        ts.getLineOfLocalPosition = getLineOfLocalPosition;
        function getLineOfLocalPositionFromLineMap(lineMap, pos) {
            return ts.computeLineOfPosition(lineMap, pos);
        }
        ts.getLineOfLocalPositionFromLineMap = getLineOfLocalPositionFromLineMap;
        function getFirstConstructorWithBody(node) {
            return ts.find(node.members, function (member) { return ts.isConstructorDeclaration(member) && nodeIsPresent(member.body); });
        }
        ts.getFirstConstructorWithBody = getFirstConstructorWithBody;
        function getSetAccessorValueParameter(accessor) {
            if (accessor && accessor.parameters.length > 0) {
                var hasThis = accessor.parameters.length === 2 && parameterIsThisKeyword(accessor.parameters[0]);
                return accessor.parameters[hasThis ? 1 : 0];
            }
        }
        ts.getSetAccessorValueParameter = getSetAccessorValueParameter;
        /** Get the type annotation for the value parameter. */
        function getSetAccessorTypeAnnotationNode(accessor) {
            var parameter = getSetAccessorValueParameter(accessor);
            return parameter && parameter.type;
        }
        ts.getSetAccessorTypeAnnotationNode = getSetAccessorTypeAnnotationNode;
        function getThisParameter(signature) {
            // callback tags do not currently support this parameters
            if (signature.parameters.length && !ts.isJSDocSignature(signature)) {
                var thisParameter = signature.parameters[0];
                if (parameterIsThisKeyword(thisParameter)) {
                    return thisParameter;
                }
            }
        }
        ts.getThisParameter = getThisParameter;
        function parameterIsThisKeyword(parameter) {
            return isThisIdentifier(parameter.name);
        }
        ts.parameterIsThisKeyword = parameterIsThisKeyword;
        function isThisIdentifier(node) {
            return !!node && node.kind === 79 /* Identifier */ && identifierIsThisKeyword(node);
        }
        ts.isThisIdentifier = isThisIdentifier;
        function isThisInTypeQuery(node) {
            if (!isThisIdentifier(node)) {
                return false;
            }
            while (ts.isQualifiedName(node.parent) && node.parent.left === node) {
                node = node.parent;
            }
            return node.parent.kind === 179 /* TypeQuery */;
        }
        ts.isThisInTypeQuery = isThisInTypeQuery;
        function identifierIsThisKeyword(id) {
            return id.originalKeywordKind === 108 /* ThisKeyword */;
        }
        ts.identifierIsThisKeyword = identifierIsThisKeyword;
        function getAllAccessorDeclarations(declarations, accessor) {
            // TODO: GH#18217
            var firstAccessor;
            var secondAccessor;
            var getAccessor;
            var setAccessor;
            if (hasDynamicName(accessor)) {
                firstAccessor = accessor;
                if (accessor.kind === 170 /* GetAccessor */) {
                    getAccessor = accessor;
                }
                else if (accessor.kind === 171 /* SetAccessor */) {
                    setAccessor = accessor;
                }
                else {
                    ts.Debug.fail("Accessor has wrong kind");
                }
            }
            else {
                ts.forEach(declarations, function (member) {
                    if (ts.isAccessor(member)
                        && isStatic(member) === isStatic(accessor)) {
                        var memberName = getPropertyNameForPropertyNameNode(member.name);
                        var accessorName = getPropertyNameForPropertyNameNode(accessor.name);
                        if (memberName === accessorName) {
                            if (!firstAccessor) {
                                firstAccessor = member;
                            }
                            else if (!secondAccessor) {
                                secondAccessor = member;
                            }
                            if (member.kind === 170 /* GetAccessor */ && !getAccessor) {
                                getAccessor = member;
                            }
                            if (member.kind === 171 /* SetAccessor */ && !setAccessor) {
                                setAccessor = member;
                            }
                        }
                    }
                });
            }
            return {
                firstAccessor: firstAccessor,
                secondAccessor: secondAccessor,
                getAccessor: getAccessor,
                setAccessor: setAccessor
            };
        }
        ts.getAllAccessorDeclarations = getAllAccessorDeclarations;
        /**
         * Gets the effective type annotation of a variable, parameter, or property. If the node was
         * parsed in a JavaScript file, gets the type annotation from JSDoc.  Also gets the type of
         * functions only the JSDoc case.
         */
        function getEffectiveTypeAnnotationNode(node) {
            if (!isInJSFile(node) && ts.isFunctionDeclaration(node))
                return undefined;
            var type = node.type;
            if (type || !isInJSFile(node))
                return type;
            return ts.isJSDocPropertyLikeTag(node) ? node.typeExpression && node.typeExpression.type : ts.getJSDocType(node);
        }
        ts.getEffectiveTypeAnnotationNode = getEffectiveTypeAnnotationNode;
        function getTypeAnnotationNode(node) {
            return node.type;
        }
        ts.getTypeAnnotationNode = getTypeAnnotationNode;
        /**
         * Gets the effective return type annotation of a signature. If the node was parsed in a
         * JavaScript file, gets the return type annotation from JSDoc.
         */
        function getEffectiveReturnTypeNode(node) {
            return ts.isJSDocSignature(node) ?
                node.type && node.type.typeExpression && node.type.typeExpression.type :
                node.type || (isInJSFile(node) ? ts.getJSDocReturnType(node) : undefined);
        }
        ts.getEffectiveReturnTypeNode = getEffectiveReturnTypeNode;
        function getJSDocTypeParameterDeclarations(node) {
            return ts.flatMap(ts.getJSDocTags(node), function (tag) { return isNonTypeAliasTemplate(tag) ? tag.typeParameters : undefined; });
        }
        ts.getJSDocTypeParameterDeclarations = getJSDocTypeParameterDeclarations;
        /** template tags are only available when a typedef isn't already using them */
        function isNonTypeAliasTemplate(tag) {
            return ts.isJSDocTemplateTag(tag) && !(tag.parent.kind === 315 /* JSDocComment */ && tag.parent.tags.some(isJSDocTypeAlias));
        }
        /**
         * Gets the effective type annotation of the value parameter of a set accessor. If the node
         * was parsed in a JavaScript file, gets the type annotation from JSDoc.
         */
        function getEffectiveSetAccessorTypeAnnotationNode(node) {
            var parameter = getSetAccessorValueParameter(node);
            return parameter && getEffectiveTypeAnnotationNode(parameter);
        }
        ts.getEffectiveSetAccessorTypeAnnotationNode = getEffectiveSetAccessorTypeAnnotationNode;
        function emitNewLineBeforeLeadingComments(lineMap, writer, node, leadingComments) {
            emitNewLineBeforeLeadingCommentsOfPosition(lineMap, writer, node.pos, leadingComments);
        }
        ts.emitNewLineBeforeLeadingComments = emitNewLineBeforeLeadingComments;
        function emitNewLineBeforeLeadingCommentsOfPosition(lineMap, writer, pos, leadingComments) {
            // If the leading comments start on different line than the start of node, write new line
            if (leadingComments && leadingComments.length && pos !== leadingComments[0].pos &&
                getLineOfLocalPositionFromLineMap(lineMap, pos) !== getLineOfLocalPositionFromLineMap(lineMap, leadingComments[0].pos)) {
                writer.writeLine();
            }
        }
        ts.emitNewLineBeforeLeadingCommentsOfPosition = emitNewLineBeforeLeadingCommentsOfPosition;
        function emitNewLineBeforeLeadingCommentOfPosition(lineMap, writer, pos, commentPos) {
            // If the leading comments start on different line than the start of node, write new line
            if (pos !== commentPos &&
                getLineOfLocalPositionFromLineMap(lineMap, pos) !== getLineOfLocalPositionFromLineMap(lineMap, commentPos)) {
                writer.writeLine();
            }
        }
        ts.emitNewLineBeforeLeadingCommentOfPosition = emitNewLineBeforeLeadingCommentOfPosition;
        function emitComments(text, lineMap, writer, comments, leadingSeparator, trailingSeparator, newLine, writeComment) {
            if (comments && comments.length > 0) {
                if (leadingSeparator) {
                    writer.writeSpace(" ");
                }
                var emitInterveningSeparator = false;
                for (var _i = 0, comments_1 = comments; _i < comments_1.length; _i++) {
                    var comment = comments_1[_i];
                    if (emitInterveningSeparator) {
                        writer.writeSpace(" ");
                        emitInterveningSeparator = false;
                    }
                    writeComment(text, lineMap, writer, comment.pos, comment.end, newLine);
                    if (comment.hasTrailingNewLine) {
                        writer.writeLine();
                    }
                    else {
                        emitInterveningSeparator = true;
                    }
                }
                if (emitInterveningSeparator && trailingSeparator) {
                    writer.writeSpace(" ");
                }
            }
        }
        ts.emitComments = emitComments;
        /**
         * Detached comment is a comment at the top of file or function body that is separated from
         * the next statement by space.
         */
        function emitDetachedComments(text, lineMap, writer, writeComment, node, newLine, removeComments) {
            var leadingComments;
            var currentDetachedCommentInfo;
            if (removeComments) {
                // removeComments is true, only reserve pinned comment at the top of file
                // For example:
                //      /*! Pinned Comment */
                //
                //      var x = 10;
                if (node.pos === 0) {
                    leadingComments = ts.filter(ts.getLeadingCommentRanges(text, node.pos), isPinnedCommentLocal);
                }
            }
            else {
                // removeComments is false, just get detached as normal and bypass the process to filter comment
                leadingComments = ts.getLeadingCommentRanges(text, node.pos);
            }
            if (leadingComments) {
                var detachedComments = [];
                var lastComment = void 0;
                for (var _i = 0, leadingComments_1 = leadingComments; _i < leadingComments_1.length; _i++) {
                    var comment = leadingComments_1[_i];
                    if (lastComment) {
                        var lastCommentLine = getLineOfLocalPositionFromLineMap(lineMap, lastComment.end);
                        var commentLine = getLineOfLocalPositionFromLineMap(lineMap, comment.pos);
                        if (commentLine >= lastCommentLine + 2) {
                            // There was a blank line between the last comment and this comment.  This
                            // comment is not part of the copyright comments.  Return what we have so
                            // far.
                            break;
                        }
                    }
                    detachedComments.push(comment);
                    lastComment = comment;
                }
                if (detachedComments.length) {
                    // All comments look like they could have been part of the copyright header.  Make
                    // sure there is at least one blank line between it and the node.  If not, it's not
                    // a copyright header.
                    var lastCommentLine = getLineOfLocalPositionFromLineMap(lineMap, ts.last(detachedComments).end);
                    var nodeLine = getLineOfLocalPositionFromLineMap(lineMap, ts.skipTrivia(text, node.pos));
                    if (nodeLine >= lastCommentLine + 2) {
                        // Valid detachedComments
                        emitNewLineBeforeLeadingComments(lineMap, writer, node, leadingComments);
                        emitComments(text, lineMap, writer, detachedComments, /*leadingSeparator*/ false, /*trailingSeparator*/ true, newLine, writeComment);
                        currentDetachedCommentInfo = { nodePos: node.pos, detachedCommentEndPos: ts.last(detachedComments).end };
                    }
                }
            }
            return currentDetachedCommentInfo;
            function isPinnedCommentLocal(comment) {
                return isPinnedComment(text, comment.pos);
            }
        }
        ts.emitDetachedComments = emitDetachedComments;
        function writeCommentRange(text, lineMap, writer, commentPos, commentEnd, newLine) {
            if (text.charCodeAt(commentPos + 1) === 42 /* asterisk */) {
                var firstCommentLineAndCharacter = ts.computeLineAndCharacterOfPosition(lineMap, commentPos);
                var lineCount = lineMap.length;
                var firstCommentLineIndent = void 0;
                for (var pos = commentPos, currentLine = firstCommentLineAndCharacter.line; pos < commentEnd; currentLine++) {
                    var nextLineStart = (currentLine + 1) === lineCount
                        ? text.length + 1
                        : lineMap[currentLine + 1];
                    if (pos !== commentPos) {
                        // If we are not emitting first line, we need to write the spaces to adjust the alignment
                        if (firstCommentLineIndent === undefined) {
                            firstCommentLineIndent = calculateIndent(text, lineMap[firstCommentLineAndCharacter.line], commentPos);
                        }
                        // These are number of spaces writer is going to write at current indent
                        var currentWriterIndentSpacing = writer.getIndent() * getIndentSize();
                        // Number of spaces we want to be writing
                        // eg: Assume writer indent
                        // module m {
                        //         /* starts at character 9 this is line 1
                        //    * starts at character pos 4 line                        --1  = 8 - 8 + 3
                        //   More left indented comment */                            --2  = 8 - 8 + 2
                        //     class c { }
                        // }
                        // module m {
                        //     /* this is line 1 -- Assume current writer indent 8
                        //      * line                                                --3 = 8 - 4 + 5
                        //            More right indented comment */                  --4 = 8 - 4 + 11
                        //     class c { }
                        // }
                        var spacesToEmit = currentWriterIndentSpacing - firstCommentLineIndent + calculateIndent(text, pos, nextLineStart);
                        if (spacesToEmit > 0) {
                            var numberOfSingleSpacesToEmit = spacesToEmit % getIndentSize();
                            var indentSizeSpaceString = getIndentString((spacesToEmit - numberOfSingleSpacesToEmit) / getIndentSize());
                            // Write indent size string ( in eg 1: = "", 2: "" , 3: string with 8 spaces 4: string with 12 spaces
                            writer.rawWrite(indentSizeSpaceString);
                            // Emit the single spaces (in eg: 1: 3 spaces, 2: 2 spaces, 3: 1 space, 4: 3 spaces)
                            while (numberOfSingleSpacesToEmit) {
                                writer.rawWrite(" ");
                                numberOfSingleSpacesToEmit--;
                            }
                        }
                        else {
                            // No spaces to emit write empty string
                            writer.rawWrite("");
                        }
                    }
                    // Write the comment line text
                    writeTrimmedCurrentLine(text, commentEnd, writer, newLine, pos, nextLineStart);
                    pos = nextLineStart;
                }
            }
            else {
                // Single line comment of style //....
                writer.writeComment(text.substring(commentPos, commentEnd));
            }
        }
        ts.writeCommentRange = writeCommentRange;
        function writeTrimmedCurrentLine(text, commentEnd, writer, newLine, pos, nextLineStart) {
            var end = Math.min(commentEnd, nextLineStart - 1);
            var currentLineText = ts.trimString(text.substring(pos, end));
            if (currentLineText) {
                // trimmed forward and ending spaces text
                writer.writeComment(currentLineText);
                if (end !== commentEnd) {
                    writer.writeLine();
                }
            }
            else {
                // Empty string - make sure we write empty line
                writer.rawWrite(newLine);
            }
        }
        function calculateIndent(text, pos, end) {
            var currentLineIndent = 0;
            for (; pos < end && ts.isWhiteSpaceSingleLine(text.charCodeAt(pos)); pos++) {
                if (text.charCodeAt(pos) === 9 /* tab */) {
                    // Tabs = TabSize = indent size and go to next tabStop
                    currentLineIndent += getIndentSize() - (currentLineIndent % getIndentSize());
                }
                else {
                    // Single space
                    currentLineIndent++;
                }
            }
            return currentLineIndent;
        }
        function hasEffectiveModifiers(node) {
            return getEffectiveModifierFlags(node) !== 0 /* None */;
        }
        ts.hasEffectiveModifiers = hasEffectiveModifiers;
        function hasSyntacticModifiers(node) {
            return getSyntacticModifierFlags(node) !== 0 /* None */;
        }
        ts.hasSyntacticModifiers = hasSyntacticModifiers;
        function hasEffectiveModifier(node, flags) {
            return !!getSelectedEffectiveModifierFlags(node, flags);
        }
        ts.hasEffectiveModifier = hasEffectiveModifier;
        function hasSyntacticModifier(node, flags) {
            return !!getSelectedSyntacticModifierFlags(node, flags);
        }
        ts.hasSyntacticModifier = hasSyntacticModifier;
        function isStatic(node) {
            // https://tc39.es/ecma262/#sec-static-semantics-isstatic
            return ts.isClassElement(node) && hasStaticModifier(node) || ts.isClassStaticBlockDeclaration(node);
        }
        ts.isStatic = isStatic;
        function hasStaticModifier(node) {
            return hasSyntacticModifier(node, 32 /* Static */);
        }
        ts.hasStaticModifier = hasStaticModifier;
        function hasOverrideModifier(node) {
            return hasEffectiveModifier(node, 16384 /* Override */);
        }
        ts.hasOverrideModifier = hasOverrideModifier;
        function hasAbstractModifier(node) {
            return hasSyntacticModifier(node, 128 /* Abstract */);
        }
        ts.hasAbstractModifier = hasAbstractModifier;
        function hasAmbientModifier(node) {
            return hasSyntacticModifier(node, 2 /* Ambient */);
        }
        ts.hasAmbientModifier = hasAmbientModifier;
        function hasEffectiveReadonlyModifier(node) {
            return hasEffectiveModifier(node, 64 /* Readonly */);
        }
        ts.hasEffectiveReadonlyModifier = hasEffectiveReadonlyModifier;
        function getSelectedEffectiveModifierFlags(node, flags) {
            return getEffectiveModifierFlags(node) & flags;
        }
        ts.getSelectedEffectiveModifierFlags = getSelectedEffectiveModifierFlags;
        function getSelectedSyntacticModifierFlags(node, flags) {
            return getSyntacticModifierFlags(node) & flags;
        }
        ts.getSelectedSyntacticModifierFlags = getSelectedSyntacticModifierFlags;
        function getModifierFlagsWorker(node, includeJSDoc, alwaysIncludeJSDoc) {
            if (node.kind >= 0 /* FirstToken */ && node.kind <= 158 /* LastToken */) {
                return 0 /* None */;
            }
            if (!(node.modifierFlagsCache & 536870912 /* HasComputedFlags */)) {
                node.modifierFlagsCache = getSyntacticModifierFlagsNoCache(node) | 536870912 /* HasComputedFlags */;
            }
            if (includeJSDoc && !(node.modifierFlagsCache & 4096 /* HasComputedJSDocModifiers */) && (alwaysIncludeJSDoc || isInJSFile(node)) && node.parent) {
                node.modifierFlagsCache |= getJSDocModifierFlagsNoCache(node) | 4096 /* HasComputedJSDocModifiers */;
            }
            return node.modifierFlagsCache & ~(536870912 /* HasComputedFlags */ | 4096 /* HasComputedJSDocModifiers */);
        }
        /**
         * Gets the effective ModifierFlags for the provided node, including JSDoc modifiers. The modifiers will be cached on the node to improve performance.
         *
         * NOTE: This function may use `parent` pointers.
         */
        function getEffectiveModifierFlags(node) {
            return getModifierFlagsWorker(node, /*includeJSDoc*/ true);
        }
        ts.getEffectiveModifierFlags = getEffectiveModifierFlags;
        function getEffectiveModifierFlagsAlwaysIncludeJSDoc(node) {
            return getModifierFlagsWorker(node, /*includeJSDOc*/ true, /*alwaysIncludeJSDOc*/ true);
        }
        ts.getEffectiveModifierFlagsAlwaysIncludeJSDoc = getEffectiveModifierFlagsAlwaysIncludeJSDoc;
        /**
         * Gets the ModifierFlags for syntactic modifiers on the provided node. The modifiers will be cached on the node to improve performance.
         *
         * NOTE: This function does not use `parent` pointers and will not include modifiers from JSDoc.
         */
        function getSyntacticModifierFlags(node) {
            return getModifierFlagsWorker(node, /*includeJSDoc*/ false);
        }
        ts.getSyntacticModifierFlags = getSyntacticModifierFlags;
        function getJSDocModifierFlagsNoCache(node) {
            var flags = 0 /* None */;
            if (!!node.parent && !ts.isParameter(node)) {
                if (isInJSFile(node)) {
                    if (ts.getJSDocPublicTagNoCache(node))
                        flags |= 4 /* Public */;
                    if (ts.getJSDocPrivateTagNoCache(node))
                        flags |= 8 /* Private */;
                    if (ts.getJSDocProtectedTagNoCache(node))
                        flags |= 16 /* Protected */;
                    if (ts.getJSDocReadonlyTagNoCache(node))
                        flags |= 64 /* Readonly */;
                    if (ts.getJSDocOverrideTagNoCache(node))
                        flags |= 16384 /* Override */;
                }
                if (ts.getJSDocDeprecatedTagNoCache(node))
                    flags |= 8192 /* Deprecated */;
            }
            return flags;
        }
        /**
         * Gets the effective ModifierFlags for the provided node, including JSDoc modifiers. The modifier flags cache on the node is ignored.
         *
         * NOTE: This function may use `parent` pointers.
         */
        function getEffectiveModifierFlagsNoCache(node) {
            return getSyntacticModifierFlagsNoCache(node) | getJSDocModifierFlagsNoCache(node);
        }
        ts.getEffectiveModifierFlagsNoCache = getEffectiveModifierFlagsNoCache;
        /**
         * Gets the ModifierFlags for syntactic modifiers on the provided node. The modifier flags cache on the node is ignored.
         *
         * NOTE: This function does not use `parent` pointers and will not include modifiers from JSDoc.
         */
        function getSyntacticModifierFlagsNoCache(node) {
            var flags = modifiersToFlags(node.modifiers);
            if (node.flags & 4 /* NestedNamespace */ || (node.kind === 79 /* Identifier */ && node.isInJSDocNamespace)) {
                flags |= 1 /* Export */;
            }
            return flags;
        }
        ts.getSyntacticModifierFlagsNoCache = getSyntacticModifierFlagsNoCache;
        function modifiersToFlags(modifiers) {
            var flags = 0 /* None */;
            if (modifiers) {
                for (var _i = 0, modifiers_1 = modifiers; _i < modifiers_1.length; _i++) {
                    var modifier = modifiers_1[_i];
                    flags |= modifierToFlag(modifier.kind);
                }
            }
            return flags;
        }
        ts.modifiersToFlags = modifiersToFlags;
        function modifierToFlag(token) {
            switch (token) {
                case 124 /* StaticKeyword */: return 32 /* Static */;
                case 123 /* PublicKeyword */: return 4 /* Public */;
                case 122 /* ProtectedKeyword */: return 16 /* Protected */;
                case 121 /* PrivateKeyword */: return 8 /* Private */;
                case 126 /* AbstractKeyword */: return 128 /* Abstract */;
                case 93 /* ExportKeyword */: return 1 /* Export */;
                case 134 /* DeclareKeyword */: return 2 /* Ambient */;
                case 85 /* ConstKeyword */: return 2048 /* Const */;
                case 88 /* DefaultKeyword */: return 512 /* Default */;
                case 130 /* AsyncKeyword */: return 256 /* Async */;
                case 143 /* ReadonlyKeyword */: return 64 /* Readonly */;
                case 157 /* OverrideKeyword */: return 16384 /* Override */;
            }
            return 0 /* None */;
        }
        ts.modifierToFlag = modifierToFlag;
        function createModifiers(modifierFlags) {
            return modifierFlags ? ts.factory.createNodeArray(ts.factory.createModifiersFromModifierFlags(modifierFlags)) : undefined;
        }
        ts.createModifiers = createModifiers;
        function isLogicalOperator(token) {
            return token === 56 /* BarBarToken */
                || token === 55 /* AmpersandAmpersandToken */
                || token === 53 /* ExclamationToken */;
        }
        ts.isLogicalOperator = isLogicalOperator;
        function isLogicalOrCoalescingAssignmentOperator(token) {
            return token === 75 /* BarBarEqualsToken */
                || token === 76 /* AmpersandAmpersandEqualsToken */
                || token === 77 /* QuestionQuestionEqualsToken */;
        }
        ts.isLogicalOrCoalescingAssignmentOperator = isLogicalOrCoalescingAssignmentOperator;
        function isLogicalOrCoalescingAssignmentExpression(expr) {
            return isLogicalOrCoalescingAssignmentOperator(expr.operatorToken.kind);
        }
        ts.isLogicalOrCoalescingAssignmentExpression = isLogicalOrCoalescingAssignmentExpression;
        function isAssignmentOperator(token) {
            return token >= 63 /* FirstAssignment */ && token <= 78 /* LastAssignment */;
        }
        ts.isAssignmentOperator = isAssignmentOperator;
        /** Get `C` given `N` if `N` is in the position `class C extends N` where `N` is an ExpressionWithTypeArguments. */
        function tryGetClassExtendingExpressionWithTypeArguments(node) {
            var cls = tryGetClassImplementingOrExtendingExpressionWithTypeArguments(node);
            return cls && !cls.isImplements ? cls.class : undefined;
        }
        ts.tryGetClassExtendingExpressionWithTypeArguments = tryGetClassExtendingExpressionWithTypeArguments;
        function tryGetClassImplementingOrExtendingExpressionWithTypeArguments(node) {
            return ts.isExpressionWithTypeArguments(node)
                && ts.isHeritageClause(node.parent)
                && ts.isClassLike(node.parent.parent)
                ? { class: node.parent.parent, isImplements: node.parent.token === 117 /* ImplementsKeyword */ }
                : undefined;
        }
        ts.tryGetClassImplementingOrExtendingExpressionWithTypeArguments = tryGetClassImplementingOrExtendingExpressionWithTypeArguments;
        function isAssignmentExpression(node, excludeCompoundAssignment) {
            return ts.isBinaryExpression(node)
                && (excludeCompoundAssignment
                    ? node.operatorToken.kind === 63 /* EqualsToken */
                    : isAssignmentOperator(node.operatorToken.kind))
                && ts.isLeftHandSideExpression(node.left);
        }
        ts.isAssignmentExpression = isAssignmentExpression;
        function isLeftHandSideOfAssignment(node) {
            return isAssignmentExpression(node.parent) && node.parent.left === node;
        }
        ts.isLeftHandSideOfAssignment = isLeftHandSideOfAssignment;
        function isDestructuringAssignment(node) {
            if (isAssignmentExpression(node, /*excludeCompoundAssignment*/ true)) {
                var kind = node.left.kind;
                return kind === 203 /* ObjectLiteralExpression */
                    || kind === 202 /* ArrayLiteralExpression */;
            }
            return false;
        }
        ts.isDestructuringAssignment = isDestructuringAssignment;
        function isExpressionWithTypeArgumentsInClassExtendsClause(node) {
            return tryGetClassExtendingExpressionWithTypeArguments(node) !== undefined;
        }
        ts.isExpressionWithTypeArgumentsInClassExtendsClause = isExpressionWithTypeArgumentsInClassExtendsClause;
        function isEntityNameExpression(node) {
            return node.kind === 79 /* Identifier */ || isPropertyAccessEntityNameExpression(node);
        }
        ts.isEntityNameExpression = isEntityNameExpression;
        function getFirstIdentifier(node) {
            switch (node.kind) {
                case 79 /* Identifier */:
                    return node;
                case 159 /* QualifiedName */:
                    do {
                        node = node.left;
                    } while (node.kind !== 79 /* Identifier */);
                    return node;
                case 204 /* PropertyAccessExpression */:
                    do {
                        node = node.expression;
                    } while (node.kind !== 79 /* Identifier */);
                    return node;
            }
        }
        ts.getFirstIdentifier = getFirstIdentifier;
        function isDottedName(node) {
            return node.kind === 79 /* Identifier */
                || node.kind === 108 /* ThisKeyword */
                || node.kind === 106 /* SuperKeyword */
                || node.kind === 229 /* MetaProperty */
                || node.kind === 204 /* PropertyAccessExpression */ && isDottedName(node.expression)
                || node.kind === 210 /* ParenthesizedExpression */ && isDottedName(node.expression);
        }
        ts.isDottedName = isDottedName;
        function isPropertyAccessEntityNameExpression(node) {
            return ts.isPropertyAccessExpression(node) && ts.isIdentifier(node.name) && isEntityNameExpression(node.expression);
        }
        ts.isPropertyAccessEntityNameExpression = isPropertyAccessEntityNameExpression;
        function tryGetPropertyAccessOrIdentifierToString(expr) {
            if (ts.isPropertyAccessExpression(expr)) {
                var baseStr = tryGetPropertyAccessOrIdentifierToString(expr.expression);
                if (baseStr !== undefined) {
                    return baseStr + "." + entityNameToString(expr.name);
                }
            }
            else if (ts.isElementAccessExpression(expr)) {
                var baseStr = tryGetPropertyAccessOrIdentifierToString(expr.expression);
                if (baseStr !== undefined && ts.isPropertyName(expr.argumentExpression)) {
                    return baseStr + "." + getPropertyNameForPropertyNameNode(expr.argumentExpression);
                }
            }
            else if (ts.isIdentifier(expr)) {
                return ts.unescapeLeadingUnderscores(expr.escapedText);
            }
            return undefined;
        }
        ts.tryGetPropertyAccessOrIdentifierToString = tryGetPropertyAccessOrIdentifierToString;
        function isPrototypeAccess(node) {
            return isBindableStaticAccessExpression(node) && getElementOrPropertyAccessName(node) === "prototype";
        }
        ts.isPrototypeAccess = isPrototypeAccess;
        function isRightSideOfQualifiedNameOrPropertyAccess(node) {
            return (node.parent.kind === 159 /* QualifiedName */ && node.parent.right === node) ||
                (node.parent.kind === 204 /* PropertyAccessExpression */ && node.parent.name === node);
        }
        ts.isRightSideOfQualifiedNameOrPropertyAccess = isRightSideOfQualifiedNameOrPropertyAccess;
        function isRightSideOfQualifiedNameOrPropertyAccessOrJSDocMemberName(node) {
            return ts.isQualifiedName(node.parent) && node.parent.right === node
                || ts.isPropertyAccessExpression(node.parent) && node.parent.name === node
                || ts.isJSDocMemberName(node.parent) && node.parent.right === node;
        }
        ts.isRightSideOfQualifiedNameOrPropertyAccessOrJSDocMemberName = isRightSideOfQualifiedNameOrPropertyAccessOrJSDocMemberName;
        function isEmptyObjectLiteral(expression) {
            return expression.kind === 203 /* ObjectLiteralExpression */ &&
                expression.properties.length === 0;
        }
        ts.isEmptyObjectLiteral = isEmptyObjectLiteral;
        function isEmptyArrayLiteral(expression) {
            return expression.kind === 202 /* ArrayLiteralExpression */ &&
                expression.elements.length === 0;
        }
        ts.isEmptyArrayLiteral = isEmptyArrayLiteral;
        function getLocalSymbolForExportDefault(symbol) {
            if (!isExportDefaultSymbol(symbol) || !symbol.declarations)
                return undefined;
            for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                var decl = _a[_i];
                if (decl.localSymbol)
                    return decl.localSymbol;
            }
            return undefined;
        }
        ts.getLocalSymbolForExportDefault = getLocalSymbolForExportDefault;
        function isExportDefaultSymbol(symbol) {
            return symbol && ts.length(symbol.declarations) > 0 && hasSyntacticModifier(symbol.declarations[0], 512 /* Default */);
        }
        /** Return ".ts", ".d.ts", or ".tsx", if that is the extension. */
        function tryExtractTSExtension(fileName) {
            return ts.find(ts.supportedTSExtensionsForExtractExtension, function (extension) { return ts.fileExtensionIs(fileName, extension); });
        }
        ts.tryExtractTSExtension = tryExtractTSExtension;
        /**
         * Replace each instance of non-ascii characters by one, two, three, or four escape sequences
         * representing the UTF-8 encoding of the character, and return the expanded char code list.
         */
        function getExpandedCharCodes(input) {
            var output = [];
            var length = input.length;
            for (var i = 0; i < length; i++) {
                var charCode = input.charCodeAt(i);
                // handle utf8
                if (charCode < 0x80) {
                    output.push(charCode);
                }
                else if (charCode < 0x800) {
                    output.push((charCode >> 6) | 192);
                    output.push((charCode & 63) | 128);
                }
                else if (charCode < 0x10000) {
                    output.push((charCode >> 12) | 224);
                    output.push(((charCode >> 6) & 63) | 128);
                    output.push((charCode & 63) | 128);
                }
                else if (charCode < 0x20000) {
                    output.push((charCode >> 18) | 240);
                    output.push(((charCode >> 12) & 63) | 128);
                    output.push(((charCode >> 6) & 63) | 128);
                    output.push((charCode & 63) | 128);
                }
                else {
                    ts.Debug.assert(false, "Unexpected code point");
                }
            }
            return output;
        }
        var base64Digits = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
        /**
         * Converts a string to a base-64 encoded ASCII string.
         */
        function convertToBase64(input) {
            var result = "";
            var charCodes = getExpandedCharCodes(input);
            var i = 0;
            var length = charCodes.length;
            var byte1, byte2, byte3, byte4;
            while (i < length) {
                // Convert every 6-bits in the input 3 character points
                // into a base64 digit
                byte1 = charCodes[i] >> 2;
                byte2 = (charCodes[i] & 3) << 4 | charCodes[i + 1] >> 4;
                byte3 = (charCodes[i + 1] & 15) << 2 | charCodes[i + 2] >> 6;
                byte4 = charCodes[i + 2] & 63;
                // We are out of characters in the input, set the extra
                // digits to 64 (padding character).
                if (i + 1 >= length) {
                    byte3 = byte4 = 64;
                }
                else if (i + 2 >= length) {
                    byte4 = 64;
                }
                // Write to the output
                result += base64Digits.charAt(byte1) + base64Digits.charAt(byte2) + base64Digits.charAt(byte3) + base64Digits.charAt(byte4);
                i += 3;
            }
            return result;
        }
        ts.convertToBase64 = convertToBase64;
        function getStringFromExpandedCharCodes(codes) {
            var output = "";
            var i = 0;
            var length = codes.length;
            while (i < length) {
                var charCode = codes[i];
                if (charCode < 0x80) {
                    output += String.fromCharCode(charCode);
                    i++;
                }
                else if ((charCode & 192) === 192) {
                    var value = charCode & 63;
                    i++;
                    var nextCode = codes[i];
                    while ((nextCode & 192) === 128) {
                        value = (value << 6) | (nextCode & 63);
                        i++;
                        nextCode = codes[i];
                    }
                    // `value` may be greater than 10FFFF (the maximum unicode codepoint) - JS will just make this into an invalid character for us
                    output += String.fromCharCode(value);
                }
                else {
                    // We don't want to kill the process when decoding fails (due to a following char byte not
                    // following a leading char), so we just print the (bad) value
                    output += String.fromCharCode(charCode);
                    i++;
                }
            }
            return output;
        }
        function base64encode(host, input) {
            if (host && host.base64encode) {
                return host.base64encode(input);
            }
            return convertToBase64(input);
        }
        ts.base64encode = base64encode;
        function base64decode(host, input) {
            if (host && host.base64decode) {
                return host.base64decode(input);
            }
            var length = input.length;
            var expandedCharCodes = [];
            var i = 0;
            while (i < length) {
                // Stop decoding once padding characters are present
                if (input.charCodeAt(i) === base64Digits.charCodeAt(64)) {
                    break;
                }
                // convert 4 input digits into three characters, ignoring padding characters at the end
                var ch1 = base64Digits.indexOf(input[i]);
                var ch2 = base64Digits.indexOf(input[i + 1]);
                var ch3 = base64Digits.indexOf(input[i + 2]);
                var ch4 = base64Digits.indexOf(input[i + 3]);
                var code1 = ((ch1 & 63) << 2) | ((ch2 >> 4) & 3);
                var code2 = ((ch2 & 15) << 4) | ((ch3 >> 2) & 15);
                var code3 = ((ch3 & 3) << 6) | (ch4 & 63);
                if (code2 === 0 && ch3 !== 0) { // code2 decoded to zero, but ch3 was padding - elide code2 and code3
                    expandedCharCodes.push(code1);
                }
                else if (code3 === 0 && ch4 !== 0) { // code3 decoded to zero, but ch4 was padding, elide code3
                    expandedCharCodes.push(code1, code2);
                }
                else {
                    expandedCharCodes.push(code1, code2, code3);
                }
                i += 4;
            }
            return getStringFromExpandedCharCodes(expandedCharCodes);
        }
        ts.base64decode = base64decode;
        function readJson(path, host) {
            try {
                var jsonText = host.readFile(path);
                if (!jsonText)
                    return {};
                var result = ts.parseConfigFileTextToJson(path, jsonText);
                if (result.error) {
                    return {};
                }
                return result.config;
            }
            catch (e) {
                // gracefully handle if readFile fails or returns not JSON
                return {};
            }
        }
        ts.readJson = readJson;
        function directoryProbablyExists(directoryName, host) {
            // if host does not support 'directoryExists' assume that directory will exist
            return !host.directoryExists || host.directoryExists(directoryName);
        }
        ts.directoryProbablyExists = directoryProbablyExists;
        var carriageReturnLineFeed = "\r\n";
        var lineFeed = "\n";
        function getNewLineCharacter(options, getNewLine) {
            switch (options.newLine) {
                case 0 /* CarriageReturnLineFeed */:
                    return carriageReturnLineFeed;
                case 1 /* LineFeed */:
                    return lineFeed;
            }
            return getNewLine ? getNewLine() : ts.sys ? ts.sys.newLine : carriageReturnLineFeed;
        }
        ts.getNewLineCharacter = getNewLineCharacter;
        /**
         * Creates a new TextRange from the provided pos and end.
         *
         * @param pos The start position.
         * @param end The end position.
         */
        function createRange(pos, end) {
            if (end === void 0) {
                end = pos;
            }
            ts.Debug.assert(end >= pos || end === -1);
            return { pos: pos, end: end };
        }
        ts.createRange = createRange;
        /**
         * Creates a new TextRange from a provided range with a new end position.
         *
         * @param range A TextRange.
         * @param end The new end position.
         */
        function moveRangeEnd(range, end) {
            return createRange(range.pos, end);
        }
        ts.moveRangeEnd = moveRangeEnd;
        /**
         * Creates a new TextRange from a provided range with a new start position.
         *
         * @param range A TextRange.
         * @param pos The new Start position.
         */
        function moveRangePos(range, pos) {
            return createRange(pos, range.end);
        }
        ts.moveRangePos = moveRangePos;
        /**
         * Moves the start position of a range past any decorators.
         */
        function moveRangePastDecorators(node) {
            return node.decorators && node.decorators.length > 0
                ? moveRangePos(node, node.decorators.end)
                : node;
        }
        ts.moveRangePastDecorators = moveRangePastDecorators;
        /**
         * Moves the start position of a range past any decorators or modifiers.
         */
        function moveRangePastModifiers(node) {
            return node.modifiers && node.modifiers.length > 0
                ? moveRangePos(node, node.modifiers.end)
                : moveRangePastDecorators(node);
        }
        ts.moveRangePastModifiers = moveRangePastModifiers;
        /**
         * Determines whether a TextRange has the same start and end positions.
         *
         * @param range A TextRange.
         */
        function isCollapsedRange(range) {
            return range.pos === range.end;
        }
        ts.isCollapsedRange = isCollapsedRange;
        /**
         * Creates a new TextRange for a token at the provides start position.
         *
         * @param pos The start position.
         * @param token The token.
         */
        function createTokenRange(pos, token) {
            return createRange(pos, pos + ts.tokenToString(token).length);
        }
        ts.createTokenRange = createTokenRange;
        function rangeIsOnSingleLine(range, sourceFile) {
            return rangeStartIsOnSameLineAsRangeEnd(range, range, sourceFile);
        }
        ts.rangeIsOnSingleLine = rangeIsOnSingleLine;
        function rangeStartPositionsAreOnSameLine(range1, range2, sourceFile) {
            return positionsAreOnSameLine(getStartPositionOfRange(range1, sourceFile, /*includeComments*/ false), getStartPositionOfRange(range2, sourceFile, /*includeComments*/ false), sourceFile);
        }
        ts.rangeStartPositionsAreOnSameLine = rangeStartPositionsAreOnSameLine;
        function rangeEndPositionsAreOnSameLine(range1, range2, sourceFile) {
            return positionsAreOnSameLine(range1.end, range2.end, sourceFile);
        }
        ts.rangeEndPositionsAreOnSameLine = rangeEndPositionsAreOnSameLine;
        function rangeStartIsOnSameLineAsRangeEnd(range1, range2, sourceFile) {
            return positionsAreOnSameLine(getStartPositionOfRange(range1, sourceFile, /*includeComments*/ false), range2.end, sourceFile);
        }
        ts.rangeStartIsOnSameLineAsRangeEnd = rangeStartIsOnSameLineAsRangeEnd;
        function rangeEndIsOnSameLineAsRangeStart(range1, range2, sourceFile) {
            return positionsAreOnSameLine(range1.end, getStartPositionOfRange(range2, sourceFile, /*includeComments*/ false), sourceFile);
        }
        ts.rangeEndIsOnSameLineAsRangeStart = rangeEndIsOnSameLineAsRangeStart;
        function getLinesBetweenRangeEndAndRangeStart(range1, range2, sourceFile, includeSecondRangeComments) {
            var range2Start = getStartPositionOfRange(range2, sourceFile, includeSecondRangeComments);
            return ts.getLinesBetweenPositions(sourceFile, range1.end, range2Start);
        }
        ts.getLinesBetweenRangeEndAndRangeStart = getLinesBetweenRangeEndAndRangeStart;
        function getLinesBetweenRangeEndPositions(range1, range2, sourceFile) {
            return ts.getLinesBetweenPositions(sourceFile, range1.end, range2.end);
        }
        ts.getLinesBetweenRangeEndPositions = getLinesBetweenRangeEndPositions;
        function isNodeArrayMultiLine(list, sourceFile) {
            return !positionsAreOnSameLine(list.pos, list.end, sourceFile);
        }
        ts.isNodeArrayMultiLine = isNodeArrayMultiLine;
        function positionsAreOnSameLine(pos1, pos2, sourceFile) {
            return ts.getLinesBetweenPositions(sourceFile, pos1, pos2) === 0;
        }
        ts.positionsAreOnSameLine = positionsAreOnSameLine;
        function getStartPositionOfRange(range, sourceFile, includeComments) {
            return positionIsSynthesized(range.pos) ? -1 : ts.skipTrivia(sourceFile.text, range.pos, /*stopAfterLineBreak*/ false, includeComments);
        }
        ts.getStartPositionOfRange = getStartPositionOfRange;
        function getLinesBetweenPositionAndPrecedingNonWhitespaceCharacter(pos, stopPos, sourceFile, includeComments) {
            var startPos = ts.skipTrivia(sourceFile.text, pos, /*stopAfterLineBreak*/ false, includeComments);
            var prevPos = getPreviousNonWhitespacePosition(startPos, stopPos, sourceFile);
            return ts.getLinesBetweenPositions(sourceFile, prevPos !== null && prevPos !== void 0 ? prevPos : stopPos, startPos);
        }
        ts.getLinesBetweenPositionAndPrecedingNonWhitespaceCharacter = getLinesBetweenPositionAndPrecedingNonWhitespaceCharacter;
        function getLinesBetweenPositionAndNextNonWhitespaceCharacter(pos, stopPos, sourceFile, includeComments) {
            var nextPos = ts.skipTrivia(sourceFile.text, pos, /*stopAfterLineBreak*/ false, includeComments);
            return ts.getLinesBetweenPositions(sourceFile, pos, Math.min(stopPos, nextPos));
        }
        ts.getLinesBetweenPositionAndNextNonWhitespaceCharacter = getLinesBetweenPositionAndNextNonWhitespaceCharacter;
        function getPreviousNonWhitespacePosition(pos, stopPos, sourceFile) {
            if (stopPos === void 0) {
                stopPos = 0;
            }
            while (pos-- > stopPos) {
                if (!ts.isWhiteSpaceLike(sourceFile.text.charCodeAt(pos))) {
                    return pos;
                }
            }
        }
        /**
         * Determines whether a name was originally the declaration name of an enum or namespace
         * declaration.
         */
        function isDeclarationNameOfEnumOrNamespace(node) {
            var parseNode = ts.getParseTreeNode(node);
            if (parseNode) {
                switch (parseNode.parent.kind) {
                    case 258 /* EnumDeclaration */:
                    case 259 /* ModuleDeclaration */:
                        return parseNode === parseNode.parent.name;
                }
            }
            return false;
        }
        ts.isDeclarationNameOfEnumOrNamespace = isDeclarationNameOfEnumOrNamespace;
        function getInitializedVariables(node) {
            return ts.filter(node.declarations, isInitializedVariable);
        }
        ts.getInitializedVariables = getInitializedVariables;
        function isInitializedVariable(node) {
            return node.initializer !== undefined;
        }
        function isWatchSet(options) {
            // Firefox has Object.prototype.watch
            return options.watch && options.hasOwnProperty("watch");
        }
        ts.isWatchSet = isWatchSet;
        function closeFileWatcher(watcher) {
            watcher.close();
        }
        ts.closeFileWatcher = closeFileWatcher;
        function getCheckFlags(symbol) {
            return symbol.flags & 33554432 /* Transient */ ? symbol.checkFlags : 0;
        }
        ts.getCheckFlags = getCheckFlags;
        function getDeclarationModifierFlagsFromSymbol(s, isWrite) {
            if (isWrite === void 0) {
                isWrite = false;
            }
            if (s.valueDeclaration) {
                var declaration = (isWrite && s.declarations && ts.find(s.declarations, function (d) { return d.kind === 171 /* SetAccessor */; })) || s.valueDeclaration;
                var flags = ts.getCombinedModifierFlags(declaration);
                return s.parent && s.parent.flags & 32 /* Class */ ? flags : flags & ~28 /* AccessibilityModifier */;
            }
            if (getCheckFlags(s) & 6 /* Synthetic */) {
                var checkFlags = s.checkFlags;
                var accessModifier = checkFlags & 1024 /* ContainsPrivate */ ? 8 /* Private */ :
                    checkFlags & 256 /* ContainsPublic */ ? 4 /* Public */ :
                        16 /* Protected */;
                var staticModifier = checkFlags & 2048 /* ContainsStatic */ ? 32 /* Static */ : 0;
                return accessModifier | staticModifier;
            }
            if (s.flags & 4194304 /* Prototype */) {
                return 4 /* Public */ | 32 /* Static */;
            }
            return 0;
        }
        ts.getDeclarationModifierFlagsFromSymbol = getDeclarationModifierFlagsFromSymbol;
        function skipAlias(symbol, checker) {
            return symbol.flags & 2097152 /* Alias */ ? checker.getAliasedSymbol(symbol) : symbol;
        }
        ts.skipAlias = skipAlias;
        /** See comment on `declareModuleMember` in `binder.ts`. */
        function getCombinedLocalAndExportSymbolFlags(symbol) {
            return symbol.exportSymbol ? symbol.exportSymbol.flags | symbol.flags : symbol.flags;
        }
        ts.getCombinedLocalAndExportSymbolFlags = getCombinedLocalAndExportSymbolFlags;
        function isWriteOnlyAccess(node) {
            return accessKind(node) === 1 /* Write */;
        }
        ts.isWriteOnlyAccess = isWriteOnlyAccess;
        function isWriteAccess(node) {
            return accessKind(node) !== 0 /* Read */;
        }
        ts.isWriteAccess = isWriteAccess;
        var AccessKind;
        (function (AccessKind) {
            /** Only reads from a variable. */
            AccessKind[AccessKind["Read"] = 0] = "Read";
            /** Only writes to a variable without using the result. E.g.: `x++;`. */
            AccessKind[AccessKind["Write"] = 1] = "Write";
            /** Writes to a variable and uses the result as an expression. E.g.: `f(x++);`. */
            AccessKind[AccessKind["ReadWrite"] = 2] = "ReadWrite";
        })(AccessKind || (AccessKind = {}));
        function accessKind(node) {
            var parent = node.parent;
            if (!parent)
                return 0 /* Read */;
            switch (parent.kind) {
                case 210 /* ParenthesizedExpression */:
                    return accessKind(parent);
                case 218 /* PostfixUnaryExpression */:
                case 217 /* PrefixUnaryExpression */:
                    var operator = parent.operator;
                    return operator === 45 /* PlusPlusToken */ || operator === 46 /* MinusMinusToken */ ? writeOrReadWrite() : 0 /* Read */;
                case 219 /* BinaryExpression */:
                    var _a = parent, left = _a.left, operatorToken = _a.operatorToken;
                    return left === node && isAssignmentOperator(operatorToken.kind) ?
                        operatorToken.kind === 63 /* EqualsToken */ ? 1 /* Write */ : writeOrReadWrite()
                        : 0 /* Read */;
                case 204 /* PropertyAccessExpression */:
                    return parent.name !== node ? 0 /* Read */ : accessKind(parent);
                case 291 /* PropertyAssignment */: {
                    var parentAccess = accessKind(parent.parent);
                    // In `({ x: varname }) = { x: 1 }`, the left `x` is a read, the right `x` is a write.
                    return node === parent.name ? reverseAccessKind(parentAccess) : parentAccess;
                }
                case 292 /* ShorthandPropertyAssignment */:
                    // Assume it's the local variable being accessed, since we don't check public properties for --noUnusedLocals.
                    return node === parent.objectAssignmentInitializer ? 0 /* Read */ : accessKind(parent.parent);
                case 202 /* ArrayLiteralExpression */:
                    return accessKind(parent);
                default:
                    return 0 /* Read */;
            }
            function writeOrReadWrite() {
                // If grandparent is not an ExpressionStatement, this is used as an expression in addition to having a side effect.
                return parent.parent && walkUpParenthesizedExpressions(parent.parent).kind === 236 /* ExpressionStatement */ ? 1 /* Write */ : 2 /* ReadWrite */;
            }
        }
        function reverseAccessKind(a) {
            switch (a) {
                case 0 /* Read */:
                    return 1 /* Write */;
                case 1 /* Write */:
                    return 0 /* Read */;
                case 2 /* ReadWrite */:
                    return 2 /* ReadWrite */;
                default:
                    return ts.Debug.assertNever(a);
            }
        }
        function compareDataObjects(dst, src) {
            if (!dst || !src || Object.keys(dst).length !== Object.keys(src).length) {
                return false;
            }
            for (var e in dst) {
                if (typeof dst[e] === "object") {
                    if (!compareDataObjects(dst[e], src[e])) {
                        return false;
                    }
                }
                else if (typeof dst[e] !== "function") {
                    if (dst[e] !== src[e]) {
                        return false;
                    }
                }
            }
            return true;
        }
        ts.compareDataObjects = compareDataObjects;
        /**
         * clears already present map by calling onDeleteExistingValue callback before deleting that key/value
         */
        function clearMap(map, onDeleteValue) {
            // Remove all
            map.forEach(onDeleteValue);
            map.clear();
        }
        ts.clearMap = clearMap;
        /**
         * Mutates the map with newMap such that keys in map will be same as newMap.
         */
        function mutateMapSkippingNewValues(map, newMap, options) {
            var onDeleteValue = options.onDeleteValue, onExistingValue = options.onExistingValue;
            // Needs update
            map.forEach(function (existingValue, key) {
                var valueInNewMap = newMap.get(key);
                // Not present any more in new map, remove it
                if (valueInNewMap === undefined) {
                    map.delete(key);
                    onDeleteValue(existingValue, key);
                }
                // If present notify about existing values
                else if (onExistingValue) {
                    onExistingValue(existingValue, valueInNewMap, key);
                }
            });
        }
        ts.mutateMapSkippingNewValues = mutateMapSkippingNewValues;
        /**
         * Mutates the map with newMap such that keys in map will be same as newMap.
         */
        function mutateMap(map, newMap, options) {
            // Needs update
            mutateMapSkippingNewValues(map, newMap, options);
            var createNewValue = options.createNewValue;
            // Add new values that are not already present
            newMap.forEach(function (valueInNewMap, key) {
                if (!map.has(key)) {
                    // New values
                    map.set(key, createNewValue(key, valueInNewMap));
                }
            });
        }
        ts.mutateMap = mutateMap;
        function isAbstractConstructorSymbol(symbol) {
            if (symbol.flags & 32 /* Class */) {
                var declaration = getClassLikeDeclarationOfSymbol(symbol);
                return !!declaration && hasSyntacticModifier(declaration, 128 /* Abstract */);
            }
            return false;
        }
        ts.isAbstractConstructorSymbol = isAbstractConstructorSymbol;
        function getClassLikeDeclarationOfSymbol(symbol) {
            var _a;
            return (_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.isClassLike);
        }
        ts.getClassLikeDeclarationOfSymbol = getClassLikeDeclarationOfSymbol;
        function getObjectFlags(type) {
            return type.flags & 3899393 /* ObjectFlagsType */ ? type.objectFlags : 0;
        }
        ts.getObjectFlags = getObjectFlags;
        function typeHasCallOrConstructSignatures(type, checker) {
            return checker.getSignaturesOfType(type, 0 /* Call */).length !== 0 || checker.getSignaturesOfType(type, 1 /* Construct */).length !== 0;
        }
        ts.typeHasCallOrConstructSignatures = typeHasCallOrConstructSignatures;
        function forSomeAncestorDirectory(directory, callback) {
            return !!ts.forEachAncestorDirectory(directory, function (d) { return callback(d) ? true : undefined; });
        }
        ts.forSomeAncestorDirectory = forSomeAncestorDirectory;
        function isUMDExportSymbol(symbol) {
            return !!symbol && !!symbol.declarations && !!symbol.declarations[0] && ts.isNamespaceExportDeclaration(symbol.declarations[0]);
        }
        ts.isUMDExportSymbol = isUMDExportSymbol;
        function showModuleSpecifier(_a) {
            var moduleSpecifier = _a.moduleSpecifier;
            return ts.isStringLiteral(moduleSpecifier) ? moduleSpecifier.text : getTextOfNode(moduleSpecifier);
        }
        ts.showModuleSpecifier = showModuleSpecifier;
        function getLastChild(node) {
            var lastChild;
            ts.forEachChild(node, function (child) {
                if (nodeIsPresent(child))
                    lastChild = child;
            }, function (children) {
                // As an optimization, jump straight to the end of the list.
                for (var i = children.length - 1; i >= 0; i--) {
                    if (nodeIsPresent(children[i])) {
                        lastChild = children[i];
                        break;
                    }
                }
            });
            return lastChild;
        }
        ts.getLastChild = getLastChild;
        function addToSeen(seen, key, value) {
            if (value === void 0) {
                value = true;
            }
            if (seen.has(key)) {
                return false;
            }
            seen.set(key, value);
            return true;
        }
        ts.addToSeen = addToSeen;
        function isObjectTypeDeclaration(node) {
            return ts.isClassLike(node) || ts.isInterfaceDeclaration(node) || ts.isTypeLiteralNode(node);
        }
        ts.isObjectTypeDeclaration = isObjectTypeDeclaration;
        function isTypeNodeKind(kind) {
            return (kind >= 175 /* FirstTypeNode */ && kind <= 198 /* LastTypeNode */)
                || kind === 129 /* AnyKeyword */
                || kind === 153 /* UnknownKeyword */
                || kind === 145 /* NumberKeyword */
                || kind === 156 /* BigIntKeyword */
                || kind === 146 /* ObjectKeyword */
                || kind === 132 /* BooleanKeyword */
                || kind === 148 /* StringKeyword */
                || kind === 149 /* SymbolKeyword */
                || kind === 114 /* VoidKeyword */
                || kind === 151 /* UndefinedKeyword */
                || kind === 142 /* NeverKeyword */
                || kind === 226 /* ExpressionWithTypeArguments */
                || kind === 307 /* JSDocAllType */
                || kind === 308 /* JSDocUnknownType */
                || kind === 309 /* JSDocNullableType */
                || kind === 310 /* JSDocNonNullableType */
                || kind === 311 /* JSDocOptionalType */
                || kind === 312 /* JSDocFunctionType */
                || kind === 313 /* JSDocVariadicType */;
        }
        ts.isTypeNodeKind = isTypeNodeKind;
        function isAccessExpression(node) {
            return node.kind === 204 /* PropertyAccessExpression */ || node.kind === 205 /* ElementAccessExpression */;
        }
        ts.isAccessExpression = isAccessExpression;
        function getNameOfAccessExpression(node) {
            if (node.kind === 204 /* PropertyAccessExpression */) {
                return node.name;
            }
            ts.Debug.assert(node.kind === 205 /* ElementAccessExpression */);
            return node.argumentExpression;
        }
        ts.getNameOfAccessExpression = getNameOfAccessExpression;
        function isBundleFileTextLike(section) {
            switch (section.kind) {
                case "text" /* Text */:
                case "internal" /* Internal */:
                    return true;
                default:
                    return false;
            }
        }
        ts.isBundleFileTextLike = isBundleFileTextLike;
        function isNamedImportsOrExports(node) {
            return node.kind === 267 /* NamedImports */ || node.kind === 271 /* NamedExports */;
        }
        ts.isNamedImportsOrExports = isNamedImportsOrExports;
        function getLeftmostAccessExpression(expr) {
            while (isAccessExpression(expr)) {
                expr = expr.expression;
            }
            return expr;
        }
        ts.getLeftmostAccessExpression = getLeftmostAccessExpression;
        function getLeftmostExpression(node, stopAtCallExpressions) {
            while (true) {
                switch (node.kind) {
                    case 218 /* PostfixUnaryExpression */:
                        node = node.operand;
                        continue;
                    case 219 /* BinaryExpression */:
                        node = node.left;
                        continue;
                    case 220 /* ConditionalExpression */:
                        node = node.condition;
                        continue;
                    case 208 /* TaggedTemplateExpression */:
                        node = node.tag;
                        continue;
                    case 206 /* CallExpression */:
                        if (stopAtCallExpressions) {
                            return node;
                        }
                    // falls through
                    case 227 /* AsExpression */:
                    case 205 /* ElementAccessExpression */:
                    case 204 /* PropertyAccessExpression */:
                    case 228 /* NonNullExpression */:
                    case 345 /* PartiallyEmittedExpression */:
                        node = node.expression;
                        continue;
                }
                return node;
            }
        }
        ts.getLeftmostExpression = getLeftmostExpression;
        function Symbol(flags, name) {
            this.flags = flags;
            this.escapedName = name;
            this.declarations = undefined;
            this.valueDeclaration = undefined;
            this.id = undefined;
            this.mergeId = undefined;
            this.parent = undefined;
        }
        function Type(checker, flags) {
            this.flags = flags;
            if (ts.Debug.isDebugging || ts.tracing) {
                this.checker = checker;
            }
        }
        function Signature(checker, flags) {
            this.flags = flags;
            if (ts.Debug.isDebugging) {
                this.checker = checker;
            }
        }
        function Node(kind, pos, end) {
            this.pos = pos;
            this.end = end;
            this.kind = kind;
            this.id = 0;
            this.flags = 0 /* None */;
            this.modifierFlagsCache = 0 /* None */;
            this.transformFlags = 0 /* None */;
            this.parent = undefined;
            this.original = undefined;
        }
        function Token(kind, pos, end) {
            this.pos = pos;
            this.end = end;
            this.kind = kind;
            this.id = 0;
            this.flags = 0 /* None */;
            this.transformFlags = 0 /* None */;
            this.parent = undefined;
        }
        function Identifier(kind, pos, end) {
            this.pos = pos;
            this.end = end;
            this.kind = kind;
            this.id = 0;
            this.flags = 0 /* None */;
            this.transformFlags = 0 /* None */;
            this.parent = undefined;
            this.original = undefined;
            this.flowNode = undefined;
        }
        function SourceMapSource(fileName, text, skipTrivia) {
            this.fileName = fileName;
            this.text = text;
            this.skipTrivia = skipTrivia || (function (pos) { return pos; });
        }
        // eslint-disable-next-line prefer-const
        ts.objectAllocator = {
            getNodeConstructor: function () { return Node; },
            getTokenConstructor: function () { return Token; },
            getIdentifierConstructor: function () { return Identifier; },
            getPrivateIdentifierConstructor: function () { return Node; },
            getSourceFileConstructor: function () { return Node; },
            getSymbolConstructor: function () { return Symbol; },
            getTypeConstructor: function () { return Type; },
            getSignatureConstructor: function () { return Signature; },
            getSourceMapSourceConstructor: function () { return SourceMapSource; },
        };
        function setObjectAllocator(alloc) {
            ts.objectAllocator = alloc;
        }
        ts.setObjectAllocator = setObjectAllocator;
        function formatStringFromArgs(text, args, baseIndex) {
            if (baseIndex === void 0) {
                baseIndex = 0;
            }
            return text.replace(/{(\d+)}/g, function (_match, index) { return "" + ts.Debug.checkDefined(args[+index + baseIndex]); });
        }
        ts.formatStringFromArgs = formatStringFromArgs;
        /* @internal */
        function setLocalizedDiagnosticMessages(messages) {
            ts.localizedDiagnosticMessages = messages;
        }
        ts.setLocalizedDiagnosticMessages = setLocalizedDiagnosticMessages;
        function getLocaleSpecificMessage(message) {
            return ts.localizedDiagnosticMessages && ts.localizedDiagnosticMessages[message.key] || message.message;
        }
        ts.getLocaleSpecificMessage = getLocaleSpecificMessage;
        function createDetachedDiagnostic(fileName, start, length, message) {
            assertDiagnosticLocation(/*file*/ undefined, start, length);
            var text = getLocaleSpecificMessage(message);
            if (arguments.length > 4) {
                text = formatStringFromArgs(text, arguments, 4);
            }
            return {
                file: undefined,
                start: start,
                length: length,
                messageText: text,
                category: message.category,
                code: message.code,
                reportsUnnecessary: message.reportsUnnecessary,
                fileName: fileName,
            };
        }
        ts.createDetachedDiagnostic = createDetachedDiagnostic;
        function isDiagnosticWithDetachedLocation(diagnostic) {
            return diagnostic.file === undefined
                && diagnostic.start !== undefined
                && diagnostic.length !== undefined
                && typeof diagnostic.fileName === "string";
        }
        function attachFileToDiagnostic(diagnostic, file) {
            var fileName = file.fileName || "";
            var length = file.text.length;
            ts.Debug.assertEqual(diagnostic.fileName, fileName);
            ts.Debug.assertLessThanOrEqual(diagnostic.start, length);
            ts.Debug.assertLessThanOrEqual(diagnostic.start + diagnostic.length, length);
            var diagnosticWithLocation = {
                file: file,
                start: diagnostic.start,
                length: diagnostic.length,
                messageText: diagnostic.messageText,
                category: diagnostic.category,
                code: diagnostic.code,
                reportsUnnecessary: diagnostic.reportsUnnecessary
            };
            if (diagnostic.relatedInformation) {
                diagnosticWithLocation.relatedInformation = [];
                for (var _i = 0, _a = diagnostic.relatedInformation; _i < _a.length; _i++) {
                    var related = _a[_i];
                    if (isDiagnosticWithDetachedLocation(related) && related.fileName === fileName) {
                        ts.Debug.assertLessThanOrEqual(related.start, length);
                        ts.Debug.assertLessThanOrEqual(related.start + related.length, length);
                        diagnosticWithLocation.relatedInformation.push(attachFileToDiagnostic(related, file));
                    }
                    else {
                        diagnosticWithLocation.relatedInformation.push(related);
                    }
                }
            }
            return diagnosticWithLocation;
        }
        function attachFileToDiagnostics(diagnostics, file) {
            var diagnosticsWithLocation = [];
            for (var _i = 0, diagnostics_1 = diagnostics; _i < diagnostics_1.length; _i++) {
                var diagnostic = diagnostics_1[_i];
                diagnosticsWithLocation.push(attachFileToDiagnostic(diagnostic, file));
            }
            return diagnosticsWithLocation;
        }
        ts.attachFileToDiagnostics = attachFileToDiagnostics;
        function createFileDiagnostic(file, start, length, message) {
            assertDiagnosticLocation(file, start, length);
            var text = getLocaleSpecificMessage(message);
            if (arguments.length > 4) {
                text = formatStringFromArgs(text, arguments, 4);
            }
            return {
                file: file,
                start: start,
                length: length,
                messageText: text,
                category: message.category,
                code: message.code,
                reportsUnnecessary: message.reportsUnnecessary,
                reportsDeprecated: message.reportsDeprecated
            };
        }
        ts.createFileDiagnostic = createFileDiagnostic;
        function formatMessage(_dummy, message) {
            var text = getLocaleSpecificMessage(message);
            if (arguments.length > 2) {
                text = formatStringFromArgs(text, arguments, 2);
            }
            return text;
        }
        ts.formatMessage = formatMessage;
        function createCompilerDiagnostic(message) {
            var text = getLocaleSpecificMessage(message);
            if (arguments.length > 1) {
                text = formatStringFromArgs(text, arguments, 1);
            }
            return {
                file: undefined,
                start: undefined,
                length: undefined,
                messageText: text,
                category: message.category,
                code: message.code,
                reportsUnnecessary: message.reportsUnnecessary,
                reportsDeprecated: message.reportsDeprecated
            };
        }
        ts.createCompilerDiagnostic = createCompilerDiagnostic;
        function createCompilerDiagnosticFromMessageChain(chain, relatedInformation) {
            return {
                file: undefined,
                start: undefined,
                length: undefined,
                code: chain.code,
                category: chain.category,
                messageText: chain.next ? chain : chain.messageText,
                relatedInformation: relatedInformation
            };
        }
        ts.createCompilerDiagnosticFromMessageChain = createCompilerDiagnosticFromMessageChain;
        function chainDiagnosticMessages(details, message) {
            var text = getLocaleSpecificMessage(message);
            if (arguments.length > 2) {
                text = formatStringFromArgs(text, arguments, 2);
            }
            return {
                messageText: text,
                category: message.category,
                code: message.code,
                next: details === undefined || Array.isArray(details) ? details : [details]
            };
        }
        ts.chainDiagnosticMessages = chainDiagnosticMessages;
        function concatenateDiagnosticMessageChains(headChain, tailChain) {
            var lastChain = headChain;
            while (lastChain.next) {
                lastChain = lastChain.next[0];
            }
            lastChain.next = [tailChain];
        }
        ts.concatenateDiagnosticMessageChains = concatenateDiagnosticMessageChains;
        function getDiagnosticFilePath(diagnostic) {
            return diagnostic.file ? diagnostic.file.path : undefined;
        }
        function compareDiagnostics(d1, d2) {
            return compareDiagnosticsSkipRelatedInformation(d1, d2) ||
                compareRelatedInformation(d1, d2) ||
                0 /* EqualTo */;
        }
        ts.compareDiagnostics = compareDiagnostics;
        function compareDiagnosticsSkipRelatedInformation(d1, d2) {
            return ts.compareStringsCaseSensitive(getDiagnosticFilePath(d1), getDiagnosticFilePath(d2)) ||
                ts.compareValues(d1.start, d2.start) ||
                ts.compareValues(d1.length, d2.length) ||
                ts.compareValues(d1.code, d2.code) ||
                compareMessageText(d1.messageText, d2.messageText) ||
                0 /* EqualTo */;
        }
        ts.compareDiagnosticsSkipRelatedInformation = compareDiagnosticsSkipRelatedInformation;
        function compareRelatedInformation(d1, d2) {
            if (!d1.relatedInformation && !d2.relatedInformation) {
                return 0 /* EqualTo */;
            }
            if (d1.relatedInformation && d2.relatedInformation) {
                return ts.compareValues(d1.relatedInformation.length, d2.relatedInformation.length) || ts.forEach(d1.relatedInformation, function (d1i, index) {
                    var d2i = d2.relatedInformation[index];
                    return compareDiagnostics(d1i, d2i); // EqualTo is 0, so falsy, and will cause the next item to be compared
                }) || 0 /* EqualTo */;
            }
            return d1.relatedInformation ? -1 /* LessThan */ : 1 /* GreaterThan */;
        }
        function compareMessageText(t1, t2) {
            if (typeof t1 === "string" && typeof t2 === "string") {
                return ts.compareStringsCaseSensitive(t1, t2);
            }
            else if (typeof t1 === "string") {
                return -1 /* LessThan */;
            }
            else if (typeof t2 === "string") {
                return 1 /* GreaterThan */;
            }
            var res = ts.compareStringsCaseSensitive(t1.messageText, t2.messageText);
            if (res) {
                return res;
            }
            if (!t1.next && !t2.next) {
                return 0 /* EqualTo */;
            }
            if (!t1.next) {
                return -1 /* LessThan */;
            }
            if (!t2.next) {
                return 1 /* GreaterThan */;
            }
            var len = Math.min(t1.next.length, t2.next.length);
            for (var i = 0; i < len; i++) {
                res = compareMessageText(t1.next[i], t2.next[i]);
                if (res) {
                    return res;
                }
            }
            if (t1.next.length < t2.next.length) {
                return -1 /* LessThan */;
            }
            else if (t1.next.length > t2.next.length) {
                return 1 /* GreaterThan */;
            }
            return 0 /* EqualTo */;
        }
        function getLanguageVariant(scriptKind) {
            // .tsx and .jsx files are treated as jsx language variant.
            return scriptKind === 4 /* TSX */ || scriptKind === 2 /* JSX */ || scriptKind === 1 /* JS */ || scriptKind === 6 /* JSON */ ? 1 /* JSX */ : 0 /* Standard */;
        }
        ts.getLanguageVariant = getLanguageVariant;
        function getEmitScriptTarget(compilerOptions) {
            return compilerOptions.target || 0 /* ES3 */;
        }
        ts.getEmitScriptTarget = getEmitScriptTarget;
        function getEmitModuleKind(compilerOptions) {
            return typeof compilerOptions.module === "number" ?
                compilerOptions.module :
                getEmitScriptTarget(compilerOptions) >= 2 /* ES2015 */ ? ts.ModuleKind.ES2015 : ts.ModuleKind.CommonJS;
        }
        ts.getEmitModuleKind = getEmitModuleKind;
        function getEmitModuleResolutionKind(compilerOptions) {
            var moduleResolution = compilerOptions.moduleResolution;
            if (moduleResolution === undefined) {
                moduleResolution = getEmitModuleKind(compilerOptions) === ts.ModuleKind.CommonJS ? ts.ModuleResolutionKind.NodeJs : ts.ModuleResolutionKind.Classic;
            }
            return moduleResolution;
        }
        ts.getEmitModuleResolutionKind = getEmitModuleResolutionKind;
        function hasJsonModuleEmitEnabled(options) {
            switch (getEmitModuleKind(options)) {
                case ts.ModuleKind.CommonJS:
                case ts.ModuleKind.AMD:
                case ts.ModuleKind.ES2015:
                case ts.ModuleKind.ES2020:
                case ts.ModuleKind.ESNext:
                    return true;
                default:
                    return false;
            }
        }
        ts.hasJsonModuleEmitEnabled = hasJsonModuleEmitEnabled;
        function unreachableCodeIsError(options) {
            return options.allowUnreachableCode === false;
        }
        ts.unreachableCodeIsError = unreachableCodeIsError;
        function unusedLabelIsError(options) {
            return options.allowUnusedLabels === false;
        }
        ts.unusedLabelIsError = unusedLabelIsError;
        function getAreDeclarationMapsEnabled(options) {
            return !!(getEmitDeclarations(options) && options.declarationMap);
        }
        ts.getAreDeclarationMapsEnabled = getAreDeclarationMapsEnabled;
        function getAllowSyntheticDefaultImports(compilerOptions) {
            var moduleKind = getEmitModuleKind(compilerOptions);
            return compilerOptions.allowSyntheticDefaultImports !== undefined
                ? compilerOptions.allowSyntheticDefaultImports
                : compilerOptions.esModuleInterop ||
                    moduleKind === ts.ModuleKind.System;
        }
        ts.getAllowSyntheticDefaultImports = getAllowSyntheticDefaultImports;
        function getEmitDeclarations(compilerOptions) {
            return !!(compilerOptions.declaration || compilerOptions.composite);
        }
        ts.getEmitDeclarations = getEmitDeclarations;
        function shouldPreserveConstEnums(compilerOptions) {
            return !!(compilerOptions.preserveConstEnums || compilerOptions.isolatedModules);
        }
        ts.shouldPreserveConstEnums = shouldPreserveConstEnums;
        function isIncrementalCompilation(options) {
            return !!(options.incremental || options.composite);
        }
        ts.isIncrementalCompilation = isIncrementalCompilation;
        function getStrictOptionValue(compilerOptions, flag) {
            return compilerOptions[flag] === undefined ? !!compilerOptions.strict : !!compilerOptions[flag];
        }
        ts.getStrictOptionValue = getStrictOptionValue;
        function getAllowJSCompilerOption(compilerOptions) {
            return compilerOptions.allowJs === undefined ? !!compilerOptions.checkJs : compilerOptions.allowJs;
        }
        ts.getAllowJSCompilerOption = getAllowJSCompilerOption;
        function getUseDefineForClassFields(compilerOptions) {
            return compilerOptions.useDefineForClassFields === undefined ? compilerOptions.target === 99 /* ESNext */ : compilerOptions.useDefineForClassFields;
        }
        ts.getUseDefineForClassFields = getUseDefineForClassFields;
        function compilerOptionsAffectSemanticDiagnostics(newOptions, oldOptions) {
            return optionsHaveChanges(oldOptions, newOptions, ts.semanticDiagnosticsOptionDeclarations);
        }
        ts.compilerOptionsAffectSemanticDiagnostics = compilerOptionsAffectSemanticDiagnostics;
        function compilerOptionsAffectEmit(newOptions, oldOptions) {
            return optionsHaveChanges(oldOptions, newOptions, ts.affectsEmitOptionDeclarations);
        }
        ts.compilerOptionsAffectEmit = compilerOptionsAffectEmit;
        function getCompilerOptionValue(options, option) {
            return option.strictFlag ? getStrictOptionValue(options, option.name) : options[option.name];
        }
        ts.getCompilerOptionValue = getCompilerOptionValue;
        function getJSXTransformEnabled(options) {
            var jsx = options.jsx;
            return jsx === 2 /* React */ || jsx === 4 /* ReactJSX */ || jsx === 5 /* ReactJSXDev */;
        }
        ts.getJSXTransformEnabled = getJSXTransformEnabled;
        function getJSXImplicitImportBase(compilerOptions, file) {
            var jsxImportSourcePragmas = file === null || file === void 0 ? void 0 : file.pragmas.get("jsximportsource");
            var jsxImportSourcePragma = ts.isArray(jsxImportSourcePragmas) ? jsxImportSourcePragmas[jsxImportSourcePragmas.length - 1] : jsxImportSourcePragmas;
            return compilerOptions.jsx === 4 /* ReactJSX */ ||
                compilerOptions.jsx === 5 /* ReactJSXDev */ ||
                compilerOptions.jsxImportSource ||
                jsxImportSourcePragma ?
                (jsxImportSourcePragma === null || jsxImportSourcePragma === void 0 ? void 0 : jsxImportSourcePragma.arguments.factory) || compilerOptions.jsxImportSource || "react" :
                undefined;
        }
        ts.getJSXImplicitImportBase = getJSXImplicitImportBase;
        function getJSXRuntimeImport(base, options) {
            return base ? base + "/" + (options.jsx === 5 /* ReactJSXDev */ ? "jsx-dev-runtime" : "jsx-runtime") : undefined;
        }
        ts.getJSXRuntimeImport = getJSXRuntimeImport;
        function hasZeroOrOneAsteriskCharacter(str) {
            var seenAsterisk = false;
            for (var i = 0; i < str.length; i++) {
                if (str.charCodeAt(i) === 42 /* asterisk */) {
                    if (!seenAsterisk) {
                        seenAsterisk = true;
                    }
                    else {
                        // have already seen asterisk
                        return false;
                    }
                }
            }
            return true;
        }
        ts.hasZeroOrOneAsteriskCharacter = hasZeroOrOneAsteriskCharacter;
        function createSymlinkCache(cwd, getCanonicalFileName) {
            var symlinkedDirectories;
            var symlinkedDirectoriesByRealpath;
            var symlinkedFiles;
            var hasProcessedResolutions = false;
            return {
                getSymlinkedFiles: function () { return symlinkedFiles; },
                getSymlinkedDirectories: function () { return symlinkedDirectories; },
                getSymlinkedDirectoriesByRealpath: function () { return symlinkedDirectoriesByRealpath; },
                setSymlinkedFile: function (path, real) { return (symlinkedFiles || (symlinkedFiles = new ts.Map())).set(path, real); },
                setSymlinkedDirectory: function (symlink, real) {
                    // Large, interconnected dependency graphs in pnpm will have a huge number of symlinks
                    // where both the realpath and the symlink path are inside node_modules/.pnpm. Since
                    // this path is never a candidate for a module specifier, we can ignore it entirely.
                    var symlinkPath = ts.toPath(symlink, cwd, getCanonicalFileName);
                    if (!containsIgnoredPath(symlinkPath)) {
                        symlinkPath = ts.ensureTrailingDirectorySeparator(symlinkPath);
                        if (real !== false && !(symlinkedDirectories === null || symlinkedDirectories === void 0 ? void 0 : symlinkedDirectories.has(symlinkPath))) {
                            (symlinkedDirectoriesByRealpath || (symlinkedDirectoriesByRealpath = ts.createMultiMap())).add(ts.ensureTrailingDirectorySeparator(real.realPath), symlink);
                        }
                        (symlinkedDirectories || (symlinkedDirectories = new ts.Map())).set(symlinkPath, real);
                    }
                },
                setSymlinkedDirectoryFromSymlinkedFile: function (symlink, real) {
                    this.setSymlinkedFile(ts.toPath(symlink, cwd, getCanonicalFileName), real);
                    var _a = guessDirectorySymlink(real, symlink, cwd, getCanonicalFileName) || ts.emptyArray, commonResolved = _a[0], commonOriginal = _a[1];
                    if (commonResolved && commonOriginal) {
                        this.setSymlinkedDirectory(commonOriginal, {
                            real: commonResolved,
                            realPath: ts.toPath(commonResolved, cwd, getCanonicalFileName),
                        });
                    }
                },
                setSymlinksFromResolutions: function (files, typeReferenceDirectives) {
                    var _this = this;
                    var _a;
                    ts.Debug.assert(!hasProcessedResolutions);
                    hasProcessedResolutions = true;
                    for (var _i = 0, files_1 = files; _i < files_1.length; _i++) {
                        var file = files_1[_i];
                        (_a = file.resolvedModules) === null || _a === void 0 ? void 0 : _a.forEach(function (resolution) { return processResolution(_this, resolution); });
                    }
                    typeReferenceDirectives === null || typeReferenceDirectives === void 0 ? void 0 : typeReferenceDirectives.forEach(function (resolution) { return processResolution(_this, resolution); });
                },
                hasProcessedResolutions: function () { return hasProcessedResolutions; },
            };
            function processResolution(cache, resolution) {
                if (!resolution || !resolution.originalPath || !resolution.resolvedFileName)
                    return;
                var resolvedFileName = resolution.resolvedFileName, originalPath = resolution.originalPath;
                cache.setSymlinkedFile(ts.toPath(originalPath, cwd, getCanonicalFileName), resolvedFileName);
                var _a = guessDirectorySymlink(resolvedFileName, originalPath, cwd, getCanonicalFileName) || ts.emptyArray, commonResolved = _a[0], commonOriginal = _a[1];
                if (commonResolved && commonOriginal) {
                    cache.setSymlinkedDirectory(commonOriginal, { real: commonResolved, realPath: ts.toPath(commonResolved, cwd, getCanonicalFileName) });
                }
            }
        }
        ts.createSymlinkCache = createSymlinkCache;
        function guessDirectorySymlink(a, b, cwd, getCanonicalFileName) {
            var aParts = ts.getPathComponents(ts.getNormalizedAbsolutePath(a, cwd));
            var bParts = ts.getPathComponents(ts.getNormalizedAbsolutePath(b, cwd));
            var isDirectory = false;
            while (!isNodeModulesOrScopedPackageDirectory(aParts[aParts.length - 2], getCanonicalFileName) &&
                !isNodeModulesOrScopedPackageDirectory(bParts[bParts.length - 2], getCanonicalFileName) &&
                getCanonicalFileName(aParts[aParts.length - 1]) === getCanonicalFileName(bParts[bParts.length - 1])) {
                aParts.pop();
                bParts.pop();
                isDirectory = true;
            }
            return isDirectory ? [ts.getPathFromPathComponents(aParts), ts.getPathFromPathComponents(bParts)] : undefined;
        }
        // KLUDGE: Don't assume one 'node_modules' links to another. More likely a single directory inside the node_modules is the symlink.
        // ALso, don't assume that an `@foo` directory is linked. More likely the contents of that are linked.
        function isNodeModulesOrScopedPackageDirectory(s, getCanonicalFileName) {
            return getCanonicalFileName(s) === "node_modules" || ts.startsWith(s, "@");
        }
        function stripLeadingDirectorySeparator(s) {
            return ts.isAnyDirectorySeparator(s.charCodeAt(0)) ? s.slice(1) : undefined;
        }
        function tryRemoveDirectoryPrefix(path, dirPath, getCanonicalFileName) {
            var withoutPrefix = ts.tryRemovePrefix(path, dirPath, getCanonicalFileName);
            return withoutPrefix === undefined ? undefined : stripLeadingDirectorySeparator(withoutPrefix);
        }
        ts.tryRemoveDirectoryPrefix = tryRemoveDirectoryPrefix;
        // Reserved characters, forces escaping of any non-word (or digit), non-whitespace character.
        // It may be inefficient (we could just match (/[-[\]{}()*+?.,\\^$|#\s]/g), but this is future
        // proof.
        var reservedCharacterPattern = /[^\w\s\/]/g;
        function regExpEscape(text) {
            return text.replace(reservedCharacterPattern, escapeRegExpCharacter);
        }
        ts.regExpEscape = regExpEscape;
        function escapeRegExpCharacter(match) {
            return "\\" + match;
        }
        var wildcardCharCodes = [42 /* asterisk */, 63 /* question */];
        ts.commonPackageFolders = ["node_modules", "bower_components", "jspm_packages"];
        var implicitExcludePathRegexPattern = "(?!(" + ts.commonPackageFolders.join("|") + ")(/|$))";
        var filesMatcher = {
            /**
             * Matches any single directory segment unless it is the last segment and a .min.js file
             * Breakdown:
             *  [^./]                   # matches everything up to the first . character (excluding directory separators)
             *  (\\.(?!min\\.js$))?     # matches . characters but not if they are part of the .min.js file extension
             */
            singleAsteriskRegexFragment: "([^./]|(\\.(?!min\\.js$))?)*",
            /**
             * Regex for the ** wildcard. Matches any number of subdirectories. When used for including
             * files or directories, does not match subdirectories that start with a . character
             */
            doubleAsteriskRegexFragment: "(/" + implicitExcludePathRegexPattern + "[^/.][^/]*)*?",
            replaceWildcardCharacter: function (match) { return replaceWildcardCharacter(match, filesMatcher.singleAsteriskRegexFragment); }
        };
        var directoriesMatcher = {
            singleAsteriskRegexFragment: "[^/]*",
            /**
             * Regex for the ** wildcard. Matches any number of subdirectories. When used for including
             * files or directories, does not match subdirectories that start with a . character
             */
            doubleAsteriskRegexFragment: "(/" + implicitExcludePathRegexPattern + "[^/.][^/]*)*?",
            replaceWildcardCharacter: function (match) { return replaceWildcardCharacter(match, directoriesMatcher.singleAsteriskRegexFragment); }
        };
        var excludeMatcher = {
            singleAsteriskRegexFragment: "[^/]*",
            doubleAsteriskRegexFragment: "(/.+?)?",
            replaceWildcardCharacter: function (match) { return replaceWildcardCharacter(match, excludeMatcher.singleAsteriskRegexFragment); }
        };
        var wildcardMatchers = {
            files: filesMatcher,
            directories: directoriesMatcher,
            exclude: excludeMatcher
        };
        function getRegularExpressionForWildcard(specs, basePath, usage) {
            var patterns = getRegularExpressionsForWildcards(specs, basePath, usage);
            if (!patterns || !patterns.length) {
                return undefined;
            }
            var pattern = patterns.map(function (pattern) { return "(" + pattern + ")"; }).join("|");
            // If excluding, match "foo/bar/baz...", but if including, only allow "foo".
            var terminator = usage === "exclude" ? "($|/)" : "$";
            return "^(" + pattern + ")" + terminator;
        }
        ts.getRegularExpressionForWildcard = getRegularExpressionForWildcard;
        function getRegularExpressionsForWildcards(specs, basePath, usage) {
            if (specs === undefined || specs.length === 0) {
                return undefined;
            }
            return ts.flatMap(specs, function (spec) {
                return spec && getSubPatternFromSpec(spec, basePath, usage, wildcardMatchers[usage]);
            });
        }
        ts.getRegularExpressionsForWildcards = getRegularExpressionsForWildcards;
        /**
         * An "includes" path "foo" is implicitly a glob "foo/** /*" (without the space) if its last component has no extension,
         * and does not contain any glob characters itself.
         */
        function isImplicitGlob(lastPathComponent) {
            return !/[.*?]/.test(lastPathComponent);
        }
        ts.isImplicitGlob = isImplicitGlob;
        function getPatternFromSpec(spec, basePath, usage) {
            var pattern = spec && getSubPatternFromSpec(spec, basePath, usage, wildcardMatchers[usage]);
            return pattern && "^(" + pattern + ")" + (usage === "exclude" ? "($|/)" : "$");
        }
        ts.getPatternFromSpec = getPatternFromSpec;
        function getSubPatternFromSpec(spec, basePath, usage, _a) {
            var singleAsteriskRegexFragment = _a.singleAsteriskRegexFragment, doubleAsteriskRegexFragment = _a.doubleAsteriskRegexFragment, replaceWildcardCharacter = _a.replaceWildcardCharacter;
            var subpattern = "";
            var hasWrittenComponent = false;
            var components = ts.getNormalizedPathComponents(spec, basePath);
            var lastComponent = ts.last(components);
            if (usage !== "exclude" && lastComponent === "**") {
                return undefined;
            }
            // getNormalizedPathComponents includes the separator for the root component.
            // We need to remove to create our regex correctly.
            components[0] = ts.removeTrailingDirectorySeparator(components[0]);
            if (isImplicitGlob(lastComponent)) {
                components.push("**", "*");
            }
            var optionalCount = 0;
            for (var _i = 0, components_1 = components; _i < components_1.length; _i++) {
                var component = components_1[_i];
                if (component === "**") {
                    subpattern += doubleAsteriskRegexFragment;
                }
                else {
                    if (usage === "directories") {
                        subpattern += "(";
                        optionalCount++;
                    }
                    if (hasWrittenComponent) {
                        subpattern += ts.directorySeparator;
                    }
                    if (usage !== "exclude") {
                        var componentPattern = "";
                        // The * and ? wildcards should not match directories or files that start with . if they
                        // appear first in a component. Dotted directories and files can be included explicitly
                        // like so: **/.*/.*
                        if (component.charCodeAt(0) === 42 /* asterisk */) {
                            componentPattern += "([^./]" + singleAsteriskRegexFragment + ")?";
                            component = component.substr(1);
                        }
                        else if (component.charCodeAt(0) === 63 /* question */) {
                            componentPattern += "[^./]";
                            component = component.substr(1);
                        }
                        componentPattern += component.replace(reservedCharacterPattern, replaceWildcardCharacter);
                        // Patterns should not include subfolders like node_modules unless they are
                        // explicitly included as part of the path.
                        //
                        // As an optimization, if the component pattern is the same as the component,
                        // then there definitely were no wildcard characters and we do not need to
                        // add the exclusion pattern.
                        if (componentPattern !== component) {
                            subpattern += implicitExcludePathRegexPattern;
                        }
                        subpattern += componentPattern;
                    }
                    else {
                        subpattern += component.replace(reservedCharacterPattern, replaceWildcardCharacter);
                    }
                }
                hasWrittenComponent = true;
            }
            while (optionalCount > 0) {
                subpattern += ")?";
                optionalCount--;
            }
            return subpattern;
        }
        function replaceWildcardCharacter(match, singleAsteriskRegexFragment) {
            return match === "*" ? singleAsteriskRegexFragment : match === "?" ? "[^/]" : "\\" + match;
        }
        /** @param path directory of the tsconfig.json */
        function getFileMatcherPatterns(path, excludes, includes, useCaseSensitiveFileNames, currentDirectory) {
            path = ts.normalizePath(path);
            currentDirectory = ts.normalizePath(currentDirectory);
            var absolutePath = ts.combinePaths(currentDirectory, path);
            return {
                includeFilePatterns: ts.map(getRegularExpressionsForWildcards(includes, absolutePath, "files"), function (pattern) { return "^" + pattern + "$"; }),
                includeFilePattern: getRegularExpressionForWildcard(includes, absolutePath, "files"),
                includeDirectoryPattern: getRegularExpressionForWildcard(includes, absolutePath, "directories"),
                excludePattern: getRegularExpressionForWildcard(excludes, absolutePath, "exclude"),
                basePaths: getBasePaths(path, includes, useCaseSensitiveFileNames)
            };
        }
        ts.getFileMatcherPatterns = getFileMatcherPatterns;
        function getRegexFromPattern(pattern, useCaseSensitiveFileNames) {
            return new RegExp(pattern, useCaseSensitiveFileNames ? "" : "i");
        }
        ts.getRegexFromPattern = getRegexFromPattern;
        /** @param path directory of the tsconfig.json */
        function matchFiles(path, extensions, excludes, includes, useCaseSensitiveFileNames, currentDirectory, depth, getFileSystemEntries, realpath, directoryExists) {
            path = ts.normalizePath(path);
            currentDirectory = ts.normalizePath(currentDirectory);
            var patterns = getFileMatcherPatterns(path, excludes, includes, useCaseSensitiveFileNames, currentDirectory);
            var includeFileRegexes = patterns.includeFilePatterns && patterns.includeFilePatterns.map(function (pattern) { return getRegexFromPattern(pattern, useCaseSensitiveFileNames); });
            var includeDirectoryRegex = patterns.includeDirectoryPattern && getRegexFromPattern(patterns.includeDirectoryPattern, useCaseSensitiveFileNames);
            var excludeRegex = patterns.excludePattern && getRegexFromPattern(patterns.excludePattern, useCaseSensitiveFileNames);
            // Associate an array of results with each include regex. This keeps results in order of the "include" order.
            // If there are no "includes", then just put everything in results[0].
            var results = includeFileRegexes ? includeFileRegexes.map(function () { return []; }) : [[]];
            var visited = new ts.Map();
            var toCanonical = ts.createGetCanonicalFileName(useCaseSensitiveFileNames);
            for (var _i = 0, _a = patterns.basePaths; _i < _a.length; _i++) {
                var basePath = _a[_i];
                if (directoryExists(basePath)) {
                    visitDirectory(basePath, ts.combinePaths(currentDirectory, basePath), depth);
                }
            }
            return ts.flatten(results);
            function visitDirectory(path, absolutePath, depth) {
                var canonicalPath = toCanonical(realpath(absolutePath));
                if (visited.has(canonicalPath))
                    return;
                visited.set(canonicalPath, true);
                var _a = getFileSystemEntries(path), files = _a.files, directories = _a.directories;
                var _loop_1 = function (current) {
                    var name = ts.combinePaths(path, current);
                    var absoluteName = ts.combinePaths(absolutePath, current);
                    if (extensions && !ts.fileExtensionIsOneOf(name, extensions))
                        return "continue";
                    if (excludeRegex && excludeRegex.test(absoluteName))
                        return "continue";
                    if (!includeFileRegexes) {
                        results[0].push(name);
                    }
                    else {
                        var includeIndex = ts.findIndex(includeFileRegexes, function (re) { return re.test(absoluteName); });
                        if (includeIndex !== -1) {
                            results[includeIndex].push(name);
                        }
                    }
                };
                for (var _i = 0, _b = ts.sort(files, ts.compareStringsCaseSensitive); _i < _b.length; _i++) {
                    var current = _b[_i];
                    _loop_1(current);
                }
                if (depth !== undefined) {
                    depth--;
                    if (depth === 0) {
                        return;
                    }
                }
                for (var _c = 0, _d = ts.sort(directories, ts.compareStringsCaseSensitive); _c < _d.length; _c++) {
                    var current = _d[_c];
                    var name = ts.combinePaths(path, current);
                    var absoluteName = ts.combinePaths(absolutePath, current);
                    if ((!includeDirectoryRegex || includeDirectoryRegex.test(absoluteName)) &&
                        (!excludeRegex || !excludeRegex.test(absoluteName))) {
                        visitDirectory(name, absoluteName, depth);
                    }
                }
            }
        }
        ts.matchFiles = matchFiles;
        /**
         * Computes the unique non-wildcard base paths amongst the provided include patterns.
         */
        function getBasePaths(path, includes, useCaseSensitiveFileNames) {
            // Storage for our results in the form of literal paths (e.g. the paths as written by the user).
            var basePaths = [path];
            if (includes) {
                // Storage for literal base paths amongst the include patterns.
                var includeBasePaths = [];
                for (var _i = 0, includes_1 = includes; _i < includes_1.length; _i++) {
                    var include = includes_1[_i];
                    // We also need to check the relative paths by converting them to absolute and normalizing
                    // in case they escape the base path (e.g "..\somedirectory")
                    var absolute = ts.isRootedDiskPath(include) ? include : ts.normalizePath(ts.combinePaths(path, include));
                    // Append the literal and canonical candidate base paths.
                    includeBasePaths.push(getIncludeBasePath(absolute));
                }
                // Sort the offsets array using either the literal or canonical path representations.
                includeBasePaths.sort(ts.getStringComparer(!useCaseSensitiveFileNames));
                var _loop_2 = function (includeBasePath) {
                    if (ts.every(basePaths, function (basePath) { return !ts.containsPath(basePath, includeBasePath, path, !useCaseSensitiveFileNames); })) {
                        basePaths.push(includeBasePath);
                    }
                };
                // Iterate over each include base path and include unique base paths that are not a
                // subpath of an existing base path
                for (var _a = 0, includeBasePaths_1 = includeBasePaths; _a < includeBasePaths_1.length; _a++) {
                    var includeBasePath = includeBasePaths_1[_a];
                    _loop_2(includeBasePath);
                }
            }
            return basePaths;
        }
        function getIncludeBasePath(absolute) {
            var wildcardOffset = ts.indexOfAnyCharCode(absolute, wildcardCharCodes);
            if (wildcardOffset < 0) {
                // No "*" or "?" in the path
                return !ts.hasExtension(absolute)
                    ? absolute
                    : ts.removeTrailingDirectorySeparator(ts.getDirectoryPath(absolute));
            }
            return absolute.substring(0, absolute.lastIndexOf(ts.directorySeparator, wildcardOffset));
        }
        function ensureScriptKind(fileName, scriptKind) {
            // Using scriptKind as a condition handles both:
            // - 'scriptKind' is unspecified and thus it is `undefined`
            // - 'scriptKind' is set and it is `Unknown` (0)
            // If the 'scriptKind' is 'undefined' or 'Unknown' then we attempt
            // to get the ScriptKind from the file name. If it cannot be resolved
            // from the file name then the default 'TS' script kind is returned.
            return scriptKind || getScriptKindFromFileName(fileName) || 3 /* TS */;
        }
        ts.ensureScriptKind = ensureScriptKind;
        function getScriptKindFromFileName(fileName) {
            var ext = fileName.substr(fileName.lastIndexOf("."));
            switch (ext.toLowerCase()) {
                case ".js" /* Js */:
                    return 1 /* JS */;
                case ".jsx" /* Jsx */:
                    return 2 /* JSX */;
                case ".ts" /* Ts */:
                    return 3 /* TS */;
                case ".tsx" /* Tsx */:
                    return 4 /* TSX */;
                case ".json" /* Json */:
                    return 6 /* JSON */;
                default:
                    return 0 /* Unknown */;
            }
        }
        ts.getScriptKindFromFileName = getScriptKindFromFileName;
        /**
         *  List of supported extensions in order of file resolution precedence.
         */
        ts.supportedTSExtensions = [".ts" /* Ts */, ".tsx" /* Tsx */, ".d.ts" /* Dts */];
        ts.supportedTSExtensionsWithJson = [".ts" /* Ts */, ".tsx" /* Tsx */, ".d.ts" /* Dts */, ".json" /* Json */];
        /** Must have ".d.ts" first because if ".ts" goes first, that will be detected as the extension instead of ".d.ts". */
        ts.supportedTSExtensionsForExtractExtension = [".d.ts" /* Dts */, ".ts" /* Ts */, ".tsx" /* Tsx */];
        ts.supportedJSExtensions = [".js" /* Js */, ".jsx" /* Jsx */];
        ts.supportedJSAndJsonExtensions = [".js" /* Js */, ".jsx" /* Jsx */, ".json" /* Json */];
        var allSupportedExtensions = __spreadArray(__spreadArray([], ts.supportedTSExtensions, true), ts.supportedJSExtensions, true);
        var allSupportedExtensionsWithJson = __spreadArray(__spreadArray(__spreadArray([], ts.supportedTSExtensions, true), ts.supportedJSExtensions, true), [".json" /* Json */], false);
        function getSupportedExtensions(options, extraFileExtensions) {
            var needJsExtensions = options && getAllowJSCompilerOption(options);
            if (!extraFileExtensions || extraFileExtensions.length === 0) {
                return needJsExtensions ? allSupportedExtensions : ts.supportedTSExtensions;
            }
            var extensions = __spreadArray(__spreadArray([], needJsExtensions ? allSupportedExtensions : ts.supportedTSExtensions, true), ts.mapDefined(extraFileExtensions, function (x) { return x.scriptKind === 7 /* Deferred */ || needJsExtensions && isJSLike(x.scriptKind) ? x.extension : undefined; }), true);
            return ts.deduplicate(extensions, ts.equateStringsCaseSensitive, ts.compareStringsCaseSensitive);
        }
        ts.getSupportedExtensions = getSupportedExtensions;
        function getSuppoertedExtensionsWithJsonIfResolveJsonModule(options, supportedExtensions) {
            if (!options || !options.resolveJsonModule) {
                return supportedExtensions;
            }
            if (supportedExtensions === allSupportedExtensions) {
                return allSupportedExtensionsWithJson;
            }
            if (supportedExtensions === ts.supportedTSExtensions) {
                return ts.supportedTSExtensionsWithJson;
            }
            return __spreadArray(__spreadArray([], supportedExtensions, true), [".json" /* Json */], false);
        }
        ts.getSuppoertedExtensionsWithJsonIfResolveJsonModule = getSuppoertedExtensionsWithJsonIfResolveJsonModule;
        function isJSLike(scriptKind) {
            return scriptKind === 1 /* JS */ || scriptKind === 2 /* JSX */;
        }
        function hasJSFileExtension(fileName) {
            return ts.some(ts.supportedJSExtensions, function (extension) { return ts.fileExtensionIs(fileName, extension); });
        }
        ts.hasJSFileExtension = hasJSFileExtension;
        function hasTSFileExtension(fileName) {
            return ts.some(ts.supportedTSExtensions, function (extension) { return ts.fileExtensionIs(fileName, extension); });
        }
        ts.hasTSFileExtension = hasTSFileExtension;
        function isSupportedSourceFileName(fileName, compilerOptions, extraFileExtensions) {
            if (!fileName) {
                return false;
            }
            var supportedExtensions = getSupportedExtensions(compilerOptions, extraFileExtensions);
            for (var _i = 0, _a = getSuppoertedExtensionsWithJsonIfResolveJsonModule(compilerOptions, supportedExtensions); _i < _a.length; _i++) {
                var extension = _a[_i];
                if (ts.fileExtensionIs(fileName, extension)) {
                    return true;
                }
            }
            return false;
        }
        ts.isSupportedSourceFileName = isSupportedSourceFileName;
        function numberOfDirectorySeparators(str) {
            var match = str.match(/\//g);
            return match ? match.length : 0;
        }
        function compareNumberOfDirectorySeparators(path1, path2) {
            return ts.compareValues(numberOfDirectorySeparators(path1), numberOfDirectorySeparators(path2));
        }
        ts.compareNumberOfDirectorySeparators = compareNumberOfDirectorySeparators;
        /**
         * Extension boundaries by priority. Lower numbers indicate higher priorities, and are
         * aligned to the offset of the highest priority extension in the
         * allSupportedExtensions array.
         */
        var ExtensionPriority;
        (function (ExtensionPriority) {
            ExtensionPriority[ExtensionPriority["TypeScriptFiles"] = 0] = "TypeScriptFiles";
            ExtensionPriority[ExtensionPriority["DeclarationAndJavaScriptFiles"] = 2] = "DeclarationAndJavaScriptFiles";
            ExtensionPriority[ExtensionPriority["Highest"] = 0] = "Highest";
            ExtensionPriority[ExtensionPriority["Lowest"] = 2] = "Lowest";
        })(ExtensionPriority = ts.ExtensionPriority || (ts.ExtensionPriority = {}));
        function getExtensionPriority(path, supportedExtensions) {
            for (var i = supportedExtensions.length - 1; i >= 0; i--) {
                if (ts.fileExtensionIs(path, supportedExtensions[i])) {
                    return adjustExtensionPriority(i, supportedExtensions);
                }
            }
            // If its not in the list of supported extensions, this is likely a
            // TypeScript file with a non-ts extension
            return 0 /* Highest */;
        }
        ts.getExtensionPriority = getExtensionPriority;
        /**
         * Adjusts an extension priority to be the highest priority within the same range.
         */
        function adjustExtensionPriority(extensionPriority, supportedExtensions) {
            if (extensionPriority < 2 /* DeclarationAndJavaScriptFiles */) {
                return 0 /* TypeScriptFiles */;
            }
            else if (extensionPriority < supportedExtensions.length) {
                return 2 /* DeclarationAndJavaScriptFiles */;
            }
            else {
                return supportedExtensions.length;
            }
        }
        ts.adjustExtensionPriority = adjustExtensionPriority;
        /**
         * Gets the next lowest extension priority for a given priority.
         */
        function getNextLowestExtensionPriority(extensionPriority, supportedExtensions) {
            if (extensionPriority < 2 /* DeclarationAndJavaScriptFiles */) {
                return 2 /* DeclarationAndJavaScriptFiles */;
            }
            else {
                return supportedExtensions.length;
            }
        }
        ts.getNextLowestExtensionPriority = getNextLowestExtensionPriority;
        var extensionsToRemove = [".d.ts" /* Dts */, ".ts" /* Ts */, ".js" /* Js */, ".tsx" /* Tsx */, ".jsx" /* Jsx */, ".json" /* Json */];
        function removeFileExtension(path) {
            for (var _i = 0, extensionsToRemove_1 = extensionsToRemove; _i < extensionsToRemove_1.length; _i++) {
                var ext = extensionsToRemove_1[_i];
                var extensionless = tryRemoveExtension(path, ext);
                if (extensionless !== undefined) {
                    return extensionless;
                }
            }
            return path;
        }
        ts.removeFileExtension = removeFileExtension;
        function tryRemoveExtension(path, extension) {
            return ts.fileExtensionIs(path, extension) ? removeExtension(path, extension) : undefined;
        }
        ts.tryRemoveExtension = tryRemoveExtension;
        function removeExtension(path, extension) {
            return path.substring(0, path.length - extension.length);
        }
        ts.removeExtension = removeExtension;
        function changeExtension(path, newExtension) {
            return ts.changeAnyExtension(path, newExtension, extensionsToRemove, /*ignoreCase*/ false);
        }
        ts.changeExtension = changeExtension;
        /**
         * Returns the input if there are no stars, a pattern if there is exactly one,
         * and undefined if there are more.
         */
        function tryParsePattern(pattern) {
            var indexOfStar = pattern.indexOf("*");
            if (indexOfStar === -1) {
                return pattern;
            }
            return pattern.indexOf("*", indexOfStar + 1) !== -1
                ? undefined
                : {
                    prefix: pattern.substr(0, indexOfStar),
                    suffix: pattern.substr(indexOfStar + 1)
                };
        }
        ts.tryParsePattern = tryParsePattern;
        function tryParsePatterns(paths) {
            return ts.mapDefined(ts.getOwnKeys(paths), function (path) { return tryParsePattern(path); });
        }
        ts.tryParsePatterns = tryParsePatterns;
        function positionIsSynthesized(pos) {
            // This is a fast way of testing the following conditions:
            //  pos === undefined || pos === null || isNaN(pos) || pos < 0;
            return !(pos >= 0);
        }
        ts.positionIsSynthesized = positionIsSynthesized;
        /** True if an extension is one of the supported TypeScript extensions. */
        function extensionIsTS(ext) {
            return ext === ".ts" /* Ts */ || ext === ".tsx" /* Tsx */ || ext === ".d.ts" /* Dts */;
        }
        ts.extensionIsTS = extensionIsTS;
        function resolutionExtensionIsTSOrJson(ext) {
            return extensionIsTS(ext) || ext === ".json" /* Json */;
        }
        ts.resolutionExtensionIsTSOrJson = resolutionExtensionIsTSOrJson;
        /**
         * Gets the extension from a path.
         * Path must have a valid extension.
         */
        function extensionFromPath(path) {
            var ext = tryGetExtensionFromPath(path);
            return ext !== undefined ? ext : ts.Debug.fail("File " + path + " has unknown extension.");
        }
        ts.extensionFromPath = extensionFromPath;
        function isAnySupportedFileExtension(path) {
            return tryGetExtensionFromPath(path) !== undefined;
        }
        ts.isAnySupportedFileExtension = isAnySupportedFileExtension;
        function tryGetExtensionFromPath(path) {
            return ts.find(extensionsToRemove, function (e) { return ts.fileExtensionIs(path, e); });
        }
        ts.tryGetExtensionFromPath = tryGetExtensionFromPath;
        function isCheckJsEnabledForFile(sourceFile, compilerOptions) {
            return sourceFile.checkJsDirective ? sourceFile.checkJsDirective.enabled : compilerOptions.checkJs;
        }
        ts.isCheckJsEnabledForFile = isCheckJsEnabledForFile;
        ts.emptyFileSystemEntries = {
            files: ts.emptyArray,
            directories: ts.emptyArray
        };
        /**
         * patternOrStrings contains both patterns (containing "*") and regular strings.
         * Return an exact match if possible, or a pattern match, or undefined.
         * (These are verified by verifyCompilerOptions to have 0 or 1 "*" characters.)
         */
        function matchPatternOrExact(patternOrStrings, candidate) {
            var patterns = [];
            for (var _i = 0, patternOrStrings_1 = patternOrStrings; _i < patternOrStrings_1.length; _i++) {
                var patternOrString = patternOrStrings_1[_i];
                if (patternOrString === candidate) {
                    return candidate;
                }
                if (!ts.isString(patternOrString)) {
                    patterns.push(patternOrString);
                }
            }
            return ts.findBestPatternMatch(patterns, function (_) { return _; }, candidate);
        }
        ts.matchPatternOrExact = matchPatternOrExact;
        function sliceAfter(arr, value) {
            var index = arr.indexOf(value);
            ts.Debug.assert(index !== -1);
            return arr.slice(index);
        }
        ts.sliceAfter = sliceAfter;
        function addRelatedInfo(diagnostic) {
            var _a;
            var relatedInformation = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                relatedInformation[_i - 1] = arguments[_i];
            }
            if (!relatedInformation.length) {
                return diagnostic;
            }
            if (!diagnostic.relatedInformation) {
                diagnostic.relatedInformation = [];
            }
            ts.Debug.assert(diagnostic.relatedInformation !== ts.emptyArray, "Diagnostic had empty array singleton for related info, but is still being constructed!");
            (_a = diagnostic.relatedInformation).push.apply(_a, relatedInformation);
            return diagnostic;
        }
        ts.addRelatedInfo = addRelatedInfo;
        function minAndMax(arr, getValue) {
            ts.Debug.assert(arr.length !== 0);
            var min = getValue(arr[0]);
            var max = min;
            for (var i = 1; i < arr.length; i++) {
                var value = getValue(arr[i]);
                if (value < min) {
                    min = value;
                }
                else if (value > max) {
                    max = value;
                }
            }
            return { min: min, max: max };
        }
        ts.minAndMax = minAndMax;
        function rangeOfNode(node) {
            return { pos: getTokenPosOfNode(node), end: node.end };
        }
        ts.rangeOfNode = rangeOfNode;
        function rangeOfTypeParameters(sourceFile, typeParameters) {
            // Include the `<>`
            var pos = typeParameters.pos - 1;
            var end = ts.skipTrivia(sourceFile.text, typeParameters.end) + 1;
            return { pos: pos, end: end };
        }
        ts.rangeOfTypeParameters = rangeOfTypeParameters;
        function skipTypeChecking(sourceFile, options, host) {
            // If skipLibCheck is enabled, skip reporting errors if file is a declaration file.
            // If skipDefaultLibCheck is enabled, skip reporting errors if file contains a
            // '/// <reference no-default-lib="true"/>' directive.
            return (options.skipLibCheck && sourceFile.isDeclarationFile ||
                options.skipDefaultLibCheck && sourceFile.hasNoDefaultLib) ||
                host.isSourceOfProjectReferenceRedirect(sourceFile.fileName);
        }
        ts.skipTypeChecking = skipTypeChecking;
        function isJsonEqual(a, b) {
            // eslint-disable-next-line no-null/no-null
            return a === b || typeof a === "object" && a !== null && typeof b === "object" && b !== null && ts.equalOwnProperties(a, b, isJsonEqual);
        }
        ts.isJsonEqual = isJsonEqual;
        /**
         * Converts a bigint literal string, e.g. `0x1234n`,
         * to its decimal string representation, e.g. `4660`.
         */
        function parsePseudoBigInt(stringValue) {
            var log2Base;
            switch (stringValue.charCodeAt(1)) { // "x" in "0x123"
                case 98 /* b */:
                case 66 /* B */: // 0b or 0B
                    log2Base = 1;
                    break;
                case 111 /* o */:
                case 79 /* O */: // 0o or 0O
                    log2Base = 3;
                    break;
                case 120 /* x */:
                case 88 /* X */: // 0x or 0X
                    log2Base = 4;
                    break;
                default: // already in decimal; omit trailing "n"
                    var nIndex = stringValue.length - 1;
                    // Skip leading 0s
                    var nonZeroStart = 0;
                    while (stringValue.charCodeAt(nonZeroStart) === 48 /* _0 */) {
                        nonZeroStart++;
                    }
                    return stringValue.slice(nonZeroStart, nIndex) || "0";
            }
            // Omit leading "0b", "0o", or "0x", and trailing "n"
            var startIndex = 2, endIndex = stringValue.length - 1;
            var bitsNeeded = (endIndex - startIndex) * log2Base;
            // Stores the value specified by the string as a LE array of 16-bit integers
            // using Uint16 instead of Uint32 so combining steps can use bitwise operators
            var segments = new Uint16Array((bitsNeeded >>> 4) + (bitsNeeded & 15 ? 1 : 0));
            // Add the digits, one at a time
            for (var i = endIndex - 1, bitOffset = 0; i >= startIndex; i--, bitOffset += log2Base) {
                var segment = bitOffset >>> 4;
                var digitChar = stringValue.charCodeAt(i);
                // Find character range: 0-9 < A-F < a-f
                var digit = digitChar <= 57 /* _9 */
                    ? digitChar - 48 /* _0 */
                    : 10 + digitChar -
                        (digitChar <= 70 /* F */ ? 65 /* A */ : 97 /* a */);
                var shiftedDigit = digit << (bitOffset & 15);
                segments[segment] |= shiftedDigit;
                var residual = shiftedDigit >>> 16;
                if (residual)
                    segments[segment + 1] |= residual; // overflows segment
            }
            // Repeatedly divide segments by 10 and add remainder to base10Value
            var base10Value = "";
            var firstNonzeroSegment = segments.length - 1;
            var segmentsRemaining = true;
            while (segmentsRemaining) {
                var mod10 = 0;
                segmentsRemaining = false;
                for (var segment = firstNonzeroSegment; segment >= 0; segment--) {
                    var newSegment = mod10 << 16 | segments[segment];
                    var segmentValue = (newSegment / 10) | 0;
                    segments[segment] = segmentValue;
                    mod10 = newSegment - segmentValue * 10;
                    if (segmentValue && !segmentsRemaining) {
                        firstNonzeroSegment = segment;
                        segmentsRemaining = true;
                    }
                }
                base10Value = mod10 + base10Value;
            }
            return base10Value;
        }
        ts.parsePseudoBigInt = parsePseudoBigInt;
        function pseudoBigIntToString(_a) {
            var negative = _a.negative, base10Value = _a.base10Value;
            return (negative && base10Value !== "0" ? "-" : "") + base10Value;
        }
        ts.pseudoBigIntToString = pseudoBigIntToString;
        function isValidTypeOnlyAliasUseSite(useSite) {
            return !!(useSite.flags & 8388608 /* Ambient */)
                || isPartOfTypeQuery(useSite)
                || isIdentifierInNonEmittingHeritageClause(useSite)
                || isPartOfPossiblyValidTypeOrAbstractComputedPropertyName(useSite)
                || !(isExpressionNode(useSite) || isShorthandPropertyNameUseSite(useSite));
        }
        ts.isValidTypeOnlyAliasUseSite = isValidTypeOnlyAliasUseSite;
        function typeOnlyDeclarationIsExport(typeOnlyDeclaration) {
            return typeOnlyDeclaration.kind === 273 /* ExportSpecifier */;
        }
        ts.typeOnlyDeclarationIsExport = typeOnlyDeclarationIsExport;
        function isShorthandPropertyNameUseSite(useSite) {
            return ts.isIdentifier(useSite) && ts.isShorthandPropertyAssignment(useSite.parent) && useSite.parent.name === useSite;
        }
        function isPartOfPossiblyValidTypeOrAbstractComputedPropertyName(node) {
            while (node.kind === 79 /* Identifier */ || node.kind === 204 /* PropertyAccessExpression */) {
                node = node.parent;
            }
            if (node.kind !== 160 /* ComputedPropertyName */) {
                return false;
            }
            if (hasSyntacticModifier(node.parent, 128 /* Abstract */)) {
                return true;
            }
            var containerKind = node.parent.parent.kind;
            return containerKind === 256 /* InterfaceDeclaration */ || containerKind === 180 /* TypeLiteral */;
        }
        /** Returns true for an identifier in 1) an `implements` clause, and 2) an `extends` clause of an interface. */
        function isIdentifierInNonEmittingHeritageClause(node) {
            if (node.kind !== 79 /* Identifier */)
                return false;
            var heritageClause = ts.findAncestor(node.parent, function (parent) {
                switch (parent.kind) {
                    case 289 /* HeritageClause */:
                        return true;
                    case 204 /* PropertyAccessExpression */:
                    case 226 /* ExpressionWithTypeArguments */:
                        return false;
                    default:
                        return "quit";
                }
            });
            return (heritageClause === null || heritageClause === void 0 ? void 0 : heritageClause.token) === 117 /* ImplementsKeyword */ || (heritageClause === null || heritageClause === void 0 ? void 0 : heritageClause.parent.kind) === 256 /* InterfaceDeclaration */;
        }
        function isIdentifierTypeReference(node) {
            return ts.isTypeReferenceNode(node) && ts.isIdentifier(node.typeName);
        }
        ts.isIdentifierTypeReference = isIdentifierTypeReference;
        function arrayIsHomogeneous(array, comparer) {
            if (comparer === void 0) {
                comparer = ts.equateValues;
            }
            if (array.length < 2)
                return true;
            var first = array[0];
            for (var i = 1, length_1 = array.length; i < length_1; i++) {
                var target = array[i];
                if (!comparer(first, target))
                    return false;
            }
            return true;
        }
        ts.arrayIsHomogeneous = arrayIsHomogeneous;
        /**
         * Bypasses immutability and directly sets the `pos` property of a `TextRange` or `Node`.
         */
        /* @internal */
        function setTextRangePos(range, pos) {
            range.pos = pos;
            return range;
        }
        ts.setTextRangePos = setTextRangePos;
        /**
         * Bypasses immutability and directly sets the `end` property of a `TextRange` or `Node`.
         */
        /* @internal */
        function setTextRangeEnd(range, end) {
            range.end = end;
            return range;
        }
        ts.setTextRangeEnd = setTextRangeEnd;
        /**
         * Bypasses immutability and directly sets the `pos` and `end` properties of a `TextRange` or `Node`.
         */
        /* @internal */
        function setTextRangePosEnd(range, pos, end) {
            return setTextRangeEnd(setTextRangePos(range, pos), end);
        }
        ts.setTextRangePosEnd = setTextRangePosEnd;
        /**
         * Bypasses immutability and directly sets the `pos` and `end` properties of a `TextRange` or `Node` from the
         * provided position and width.
         */
        /* @internal */
        function setTextRangePosWidth(range, pos, width) {
            return setTextRangePosEnd(range, pos, pos + width);
        }
        ts.setTextRangePosWidth = setTextRangePosWidth;
        function setNodeFlags(node, newFlags) {
            if (node) {
                node.flags = newFlags;
            }
            return node;
        }
        ts.setNodeFlags = setNodeFlags;
        function setParent(child, parent) {
            if (child && parent) {
                child.parent = parent;
            }
            return child;
        }
        ts.setParent = setParent;
        function setEachParent(children, parent) {
            if (children) {
                for (var _i = 0, children_1 = children; _i < children_1.length; _i++) {
                    var child = children_1[_i];
                    setParent(child, parent);
                }
            }
            return children;
        }
        ts.setEachParent = setEachParent;
        function setParentRecursive(rootNode, incremental) {
            if (!rootNode)
                return rootNode;
            ts.forEachChildRecursively(rootNode, ts.isJSDocNode(rootNode) ? bindParentToChildIgnoringJSDoc : bindParentToChild);
            return rootNode;
            function bindParentToChildIgnoringJSDoc(child, parent) {
                if (incremental && child.parent === parent) {
                    return "skip";
                }
                setParent(child, parent);
            }
            function bindJSDoc(child) {
                if (ts.hasJSDocNodes(child)) {
                    for (var _i = 0, _a = child.jsDoc; _i < _a.length; _i++) {
                        var doc = _a[_i];
                        bindParentToChildIgnoringJSDoc(doc, child);
                        ts.forEachChildRecursively(doc, bindParentToChildIgnoringJSDoc);
                    }
                }
            }
            function bindParentToChild(child, parent) {
                return bindParentToChildIgnoringJSDoc(child, parent) || bindJSDoc(child);
            }
        }
        ts.setParentRecursive = setParentRecursive;
        function isPackedElement(node) {
            return !ts.isOmittedExpression(node);
        }
        /**
         * Determines whether the provided node is an ArrayLiteralExpression that contains no missing elements.
         */
        function isPackedArrayLiteral(node) {
            return ts.isArrayLiteralExpression(node) && ts.every(node.elements, isPackedElement);
        }
        ts.isPackedArrayLiteral = isPackedArrayLiteral;
        /**
         * Indicates whether the result of an `Expression` will be unused.
         *
         * NOTE: This requires a node with a valid `parent` pointer.
         */
        function expressionResultIsUnused(node) {
            ts.Debug.assertIsDefined(node.parent);
            while (true) {
                var parent = node.parent;
                // walk up parenthesized expressions, but keep a pointer to the top-most parenthesized expression
                if (ts.isParenthesizedExpression(parent)) {
                    node = parent;
                    continue;
                }
                // result is unused in an expression statement, `void` expression, or the initializer or incrementer of a `for` loop
                if (ts.isExpressionStatement(parent) ||
                    ts.isVoidExpression(parent) ||
                    ts.isForStatement(parent) && (parent.initializer === node || parent.incrementor === node)) {
                    return true;
                }
                if (ts.isCommaListExpression(parent)) {
                    // left side of comma is always unused
                    if (node !== ts.last(parent.elements))
                        return true;
                    // right side of comma is unused if parent is unused
                    node = parent;
                    continue;
                }
                if (ts.isBinaryExpression(parent) && parent.operatorToken.kind === 27 /* CommaToken */) {
                    // left side of comma is always unused
                    if (node === parent.left)
                        return true;
                    // right side of comma is unused if parent is unused
                    node = parent;
                    continue;
                }
                return false;
            }
        }
        ts.expressionResultIsUnused = expressionResultIsUnused;
        function containsIgnoredPath(path) {
            return ts.some(ts.ignoredPaths, function (p) { return ts.stringContains(path, p); });
        }
        ts.containsIgnoredPath = containsIgnoredPath;
        function getContainingNodeArray(node) {
            if (!node.parent)
                return undefined;
            switch (node.kind) {
                case 161 /* TypeParameter */:
                    var parent_1 = node.parent;
                    return parent_1.kind === 188 /* InferType */ ? undefined : parent_1.typeParameters;
                case 162 /* Parameter */:
                    return node.parent.parameters;
                case 197 /* TemplateLiteralTypeSpan */:
                    return node.parent.templateSpans;
                case 231 /* TemplateSpan */:
                    return node.parent.templateSpans;
                case 163 /* Decorator */:
                    return node.parent.decorators;
                case 289 /* HeritageClause */:
                    return node.parent.heritageClauses;
            }
            var parent = node.parent;
            if (ts.isJSDocTag(node)) {
                return ts.isJSDocTypeLiteral(node.parent) ? undefined : node.parent.tags;
            }
            switch (parent.kind) {
                case 180 /* TypeLiteral */:
                case 256 /* InterfaceDeclaration */:
                    return ts.isTypeElement(node) ? parent.members : undefined;
                case 185 /* UnionType */:
                case 186 /* IntersectionType */:
                    return parent.types;
                case 182 /* TupleType */:
                case 202 /* ArrayLiteralExpression */:
                case 346 /* CommaListExpression */:
                case 267 /* NamedImports */:
                case 271 /* NamedExports */:
                    return parent.elements;
                case 203 /* ObjectLiteralExpression */:
                case 284 /* JsxAttributes */:
                    return parent.properties;
                case 206 /* CallExpression */:
                case 207 /* NewExpression */:
                    return ts.isTypeNode(node) ? parent.typeArguments :
                        parent.expression === node ? undefined :
                            parent.arguments;
                case 276 /* JsxElement */:
                case 280 /* JsxFragment */:
                    return ts.isJsxChild(node) ? parent.children : undefined;
                case 278 /* JsxOpeningElement */:
                case 277 /* JsxSelfClosingElement */:
                    return ts.isTypeNode(node) ? parent.typeArguments : undefined;
                case 233 /* Block */:
                case 287 /* CaseClause */:
                case 288 /* DefaultClause */:
                case 260 /* ModuleBlock */:
                    return parent.statements;
                case 261 /* CaseBlock */:
                    return parent.clauses;
                case 255 /* ClassDeclaration */:
                case 224 /* ClassExpression */:
                    return ts.isClassElement(node) ? parent.members : undefined;
                case 258 /* EnumDeclaration */:
                    return ts.isEnumMember(node) ? parent.members : undefined;
                case 300 /* SourceFile */:
                    return parent.statements;
            }
        }
        ts.getContainingNodeArray = getContainingNodeArray;
        function hasContextSensitiveParameters(node) {
            // Functions with type parameters are not context sensitive.
            if (!node.typeParameters) {
                // Functions with any parameters that lack type annotations are context sensitive.
                if (ts.some(node.parameters, function (p) { return !getEffectiveTypeAnnotationNode(p); })) {
                    return true;
                }
                if (node.kind !== 212 /* ArrowFunction */) {
                    // If the first parameter is not an explicit 'this' parameter, then the function has
                    // an implicit 'this' parameter which is subject to contextual typing.
                    var parameter = ts.firstOrUndefined(node.parameters);
                    if (!(parameter && parameterIsThisKeyword(parameter))) {
                        return true;
                    }
                }
            }
            return false;
        }
        ts.hasContextSensitiveParameters = hasContextSensitiveParameters;
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        /**
         * Creates a `BaseNodeFactory` which can be used to create `Node` instances from the constructors provided by the object allocator.
         */
        function createBaseNodeFactory() {
            // tslint:disable variable-name
            var NodeConstructor;
            var TokenConstructor;
            var IdentifierConstructor;
            var PrivateIdentifierConstructor;
            var SourceFileConstructor;
            // tslint:enable variable-name
            return {
                createBaseSourceFileNode: createBaseSourceFileNode,
                createBaseIdentifierNode: createBaseIdentifierNode,
                createBasePrivateIdentifierNode: createBasePrivateIdentifierNode,
                createBaseTokenNode: createBaseTokenNode,
                createBaseNode: createBaseNode
            };
            function createBaseSourceFileNode(kind) {
                return new (SourceFileConstructor || (SourceFileConstructor = ts.objectAllocator.getSourceFileConstructor()))(kind, /*pos*/ -1, /*end*/ -1);
            }
            function createBaseIdentifierNode(kind) {
                return new (IdentifierConstructor || (IdentifierConstructor = ts.objectAllocator.getIdentifierConstructor()))(kind, /*pos*/ -1, /*end*/ -1);
            }
            function createBasePrivateIdentifierNode(kind) {
                return new (PrivateIdentifierConstructor || (PrivateIdentifierConstructor = ts.objectAllocator.getPrivateIdentifierConstructor()))(kind, /*pos*/ -1, /*end*/ -1);
            }
            function createBaseTokenNode(kind) {
                return new (TokenConstructor || (TokenConstructor = ts.objectAllocator.getTokenConstructor()))(kind, /*pos*/ -1, /*end*/ -1);
            }
            function createBaseNode(kind) {
                return new (NodeConstructor || (NodeConstructor = ts.objectAllocator.getNodeConstructor()))(kind, /*pos*/ -1, /*end*/ -1);
            }
        }
        ts.createBaseNodeFactory = createBaseNodeFactory;
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function createParenthesizerRules(factory) {
            var binaryLeftOperandParenthesizerCache;
            var binaryRightOperandParenthesizerCache;
            return {
                getParenthesizeLeftSideOfBinaryForOperator: getParenthesizeLeftSideOfBinaryForOperator,
                getParenthesizeRightSideOfBinaryForOperator: getParenthesizeRightSideOfBinaryForOperator,
                parenthesizeLeftSideOfBinary: parenthesizeLeftSideOfBinary,
                parenthesizeRightSideOfBinary: parenthesizeRightSideOfBinary,
                parenthesizeExpressionOfComputedPropertyName: parenthesizeExpressionOfComputedPropertyName,
                parenthesizeConditionOfConditionalExpression: parenthesizeConditionOfConditionalExpression,
                parenthesizeBranchOfConditionalExpression: parenthesizeBranchOfConditionalExpression,
                parenthesizeExpressionOfExportDefault: parenthesizeExpressionOfExportDefault,
                parenthesizeExpressionOfNew: parenthesizeExpressionOfNew,
                parenthesizeLeftSideOfAccess: parenthesizeLeftSideOfAccess,
                parenthesizeOperandOfPostfixUnary: parenthesizeOperandOfPostfixUnary,
                parenthesizeOperandOfPrefixUnary: parenthesizeOperandOfPrefixUnary,
                parenthesizeExpressionsOfCommaDelimitedList: parenthesizeExpressionsOfCommaDelimitedList,
                parenthesizeExpressionForDisallowedComma: parenthesizeExpressionForDisallowedComma,
                parenthesizeExpressionOfExpressionStatement: parenthesizeExpressionOfExpressionStatement,
                parenthesizeConciseBodyOfArrowFunction: parenthesizeConciseBodyOfArrowFunction,
                parenthesizeMemberOfConditionalType: parenthesizeMemberOfConditionalType,
                parenthesizeMemberOfElementType: parenthesizeMemberOfElementType,
                parenthesizeElementTypeOfArrayType: parenthesizeElementTypeOfArrayType,
                parenthesizeConstituentTypesOfUnionOrIntersectionType: parenthesizeConstituentTypesOfUnionOrIntersectionType,
                parenthesizeTypeArguments: parenthesizeTypeArguments,
            };
            function getParenthesizeLeftSideOfBinaryForOperator(operatorKind) {
                binaryLeftOperandParenthesizerCache || (binaryLeftOperandParenthesizerCache = new ts.Map());
                var parenthesizerRule = binaryLeftOperandParenthesizerCache.get(operatorKind);
                if (!parenthesizerRule) {
                    parenthesizerRule = function (node) { return parenthesizeLeftSideOfBinary(operatorKind, node); };
                    binaryLeftOperandParenthesizerCache.set(operatorKind, parenthesizerRule);
                }
                return parenthesizerRule;
            }
            function getParenthesizeRightSideOfBinaryForOperator(operatorKind) {
                binaryRightOperandParenthesizerCache || (binaryRightOperandParenthesizerCache = new ts.Map());
                var parenthesizerRule = binaryRightOperandParenthesizerCache.get(operatorKind);
                if (!parenthesizerRule) {
                    parenthesizerRule = function (node) { return parenthesizeRightSideOfBinary(operatorKind, /*leftSide*/ undefined, node); };
                    binaryRightOperandParenthesizerCache.set(operatorKind, parenthesizerRule);
                }
                return parenthesizerRule;
            }
            /**
             * Determines whether the operand to a BinaryExpression needs to be parenthesized.
             *
             * @param binaryOperator The operator for the BinaryExpression.
             * @param operand The operand for the BinaryExpression.
             * @param isLeftSideOfBinary A value indicating whether the operand is the left side of the
             *                           BinaryExpression.
             */
            function binaryOperandNeedsParentheses(binaryOperator, operand, isLeftSideOfBinary, leftOperand) {
                // If the operand has lower precedence, then it needs to be parenthesized to preserve the
                // intent of the expression. For example, if the operand is `a + b` and the operator is
                // `*`, then we need to parenthesize the operand to preserve the intended order of
                // operations: `(a + b) * x`.
                //
                // If the operand has higher precedence, then it does not need to be parenthesized. For
                // example, if the operand is `a * b` and the operator is `+`, then we do not need to
                // parenthesize to preserve the intended order of operations: `a * b + x`.
                //
                // If the operand has the same precedence, then we need to check the associativity of
                // the operator based on whether this is the left or right operand of the expression.
                //
                // For example, if `a / d` is on the right of operator `*`, we need to parenthesize
                // to preserve the intended order of operations: `x * (a / d)`
                //
                // If `a ** d` is on the left of operator `**`, we need to parenthesize to preserve
                // the intended order of operations: `(a ** b) ** c`
                var binaryOperatorPrecedence = ts.getOperatorPrecedence(219 /* BinaryExpression */, binaryOperator);
                var binaryOperatorAssociativity = ts.getOperatorAssociativity(219 /* BinaryExpression */, binaryOperator);
                var emittedOperand = ts.skipPartiallyEmittedExpressions(operand);
                if (!isLeftSideOfBinary && operand.kind === 212 /* ArrowFunction */ && binaryOperatorPrecedence > 3 /* Assignment */) {
                    // We need to parenthesize arrow functions on the right side to avoid it being
                    // parsed as parenthesized expression: `a && (() => {})`
                    return true;
                }
                var operandPrecedence = ts.getExpressionPrecedence(emittedOperand);
                switch (ts.compareValues(operandPrecedence, binaryOperatorPrecedence)) {
                    case -1 /* LessThan */:
                        // If the operand is the right side of a right-associative binary operation
                        // and is a yield expression, then we do not need parentheses.
                        if (!isLeftSideOfBinary
                            && binaryOperatorAssociativity === 1 /* Right */
                            && operand.kind === 222 /* YieldExpression */) {
                            return false;
                        }
                        return true;
                    case 1 /* GreaterThan */:
                        return false;
                    case 0 /* EqualTo */:
                        if (isLeftSideOfBinary) {
                            // No need to parenthesize the left operand when the binary operator is
                            // left associative:
                            //  (a*b)/x    -> a*b/x
                            //  (a**b)/x   -> a**b/x
                            //
                            // Parentheses are needed for the left operand when the binary operator is
                            // right associative:
                            //  (a/b)**x   -> (a/b)**x
                            //  (a**b)**x  -> (a**b)**x
                            return binaryOperatorAssociativity === 1 /* Right */;
                        }
                        else {
                            if (ts.isBinaryExpression(emittedOperand)
                                && emittedOperand.operatorToken.kind === binaryOperator) {
                                // No need to parenthesize the right operand when the binary operator and
                                // operand are the same and one of the following:
                                //  x*(a*b)     => x*a*b
                                //  x|(a|b)     => x|a|b
                                //  x&(a&b)     => x&a&b
                                //  x^(a^b)     => x^a^b
                                if (operatorHasAssociativeProperty(binaryOperator)) {
                                    return false;
                                }
                                // No need to parenthesize the right operand when the binary operator
                                // is plus (+) if both the left and right operands consist solely of either
                                // literals of the same kind or binary plus (+) expressions for literals of
                                // the same kind (recursively).
                                //  "a"+(1+2)       => "a"+(1+2)
                                //  "a"+("b"+"c")   => "a"+"b"+"c"
                                if (binaryOperator === 39 /* PlusToken */) {
                                    var leftKind = leftOperand ? getLiteralKindOfBinaryPlusOperand(leftOperand) : 0 /* Unknown */;
                                    if (ts.isLiteralKind(leftKind) && leftKind === getLiteralKindOfBinaryPlusOperand(emittedOperand)) {
                                        return false;
                                    }
                                }
                            }
                            // No need to parenthesize the right operand when the operand is right
                            // associative:
                            //  x/(a**b)    -> x/a**b
                            //  x**(a**b)   -> x**a**b
                            //
                            // Parentheses are needed for the right operand when the operand is left
                            // associative:
                            //  x/(a*b)     -> x/(a*b)
                            //  x**(a/b)    -> x**(a/b)
                            var operandAssociativity = ts.getExpressionAssociativity(emittedOperand);
                            return operandAssociativity === 0 /* Left */;
                        }
                }
            }
            /**
             * Determines whether a binary operator is mathematically associative.
             *
             * @param binaryOperator The binary operator.
             */
            function operatorHasAssociativeProperty(binaryOperator) {
                // The following operators are associative in JavaScript:
                //  (a*b)*c     -> a*(b*c)  -> a*b*c
                //  (a|b)|c     -> a|(b|c)  -> a|b|c
                //  (a&b)&c     -> a&(b&c)  -> a&b&c
                //  (a^b)^c     -> a^(b^c)  -> a^b^c
                //
                // While addition is associative in mathematics, JavaScript's `+` is not
                // guaranteed to be associative as it is overloaded with string concatenation.
                return binaryOperator === 41 /* AsteriskToken */
                    || binaryOperator === 51 /* BarToken */
                    || binaryOperator === 50 /* AmpersandToken */
                    || binaryOperator === 52 /* CaretToken */;
            }
            /**
             * This function determines whether an expression consists of a homogeneous set of
             * literal expressions or binary plus expressions that all share the same literal kind.
             * It is used to determine whether the right-hand operand of a binary plus expression can be
             * emitted without parentheses.
             */
            function getLiteralKindOfBinaryPlusOperand(node) {
                node = ts.skipPartiallyEmittedExpressions(node);
                if (ts.isLiteralKind(node.kind)) {
                    return node.kind;
                }
                if (node.kind === 219 /* BinaryExpression */ && node.operatorToken.kind === 39 /* PlusToken */) {
                    if (node.cachedLiteralKind !== undefined) {
                        return node.cachedLiteralKind;
                    }
                    var leftKind = getLiteralKindOfBinaryPlusOperand(node.left);
                    var literalKind = ts.isLiteralKind(leftKind)
                        && leftKind === getLiteralKindOfBinaryPlusOperand(node.right)
                        ? leftKind
                        : 0 /* Unknown */;
                    node.cachedLiteralKind = literalKind;
                    return literalKind;
                }
                return 0 /* Unknown */;
            }
            /**
             * Wraps the operand to a BinaryExpression in parentheses if they are needed to preserve the intended
             * order of operations.
             *
             * @param binaryOperator The operator for the BinaryExpression.
             * @param operand The operand for the BinaryExpression.
             * @param isLeftSideOfBinary A value indicating whether the operand is the left side of the
             *                           BinaryExpression.
             */
            function parenthesizeBinaryOperand(binaryOperator, operand, isLeftSideOfBinary, leftOperand) {
                var skipped = ts.skipPartiallyEmittedExpressions(operand);
                // If the resulting expression is already parenthesized, we do not need to do any further processing.
                if (skipped.kind === 210 /* ParenthesizedExpression */) {
                    return operand;
                }
                return binaryOperandNeedsParentheses(binaryOperator, operand, isLeftSideOfBinary, leftOperand)
                    ? factory.createParenthesizedExpression(operand)
                    : operand;
            }
            function parenthesizeLeftSideOfBinary(binaryOperator, leftSide) {
                return parenthesizeBinaryOperand(binaryOperator, leftSide, /*isLeftSideOfBinary*/ true);
            }
            function parenthesizeRightSideOfBinary(binaryOperator, leftSide, rightSide) {
                return parenthesizeBinaryOperand(binaryOperator, rightSide, /*isLeftSideOfBinary*/ false, leftSide);
            }
            function parenthesizeExpressionOfComputedPropertyName(expression) {
                return ts.isCommaSequence(expression) ? factory.createParenthesizedExpression(expression) : expression;
            }
            function parenthesizeConditionOfConditionalExpression(condition) {
                var conditionalPrecedence = ts.getOperatorPrecedence(220 /* ConditionalExpression */, 57 /* QuestionToken */);
                var emittedCondition = ts.skipPartiallyEmittedExpressions(condition);
                var conditionPrecedence = ts.getExpressionPrecedence(emittedCondition);
                if (ts.compareValues(conditionPrecedence, conditionalPrecedence) !== 1 /* GreaterThan */) {
                    return factory.createParenthesizedExpression(condition);
                }
                return condition;
            }
            function parenthesizeBranchOfConditionalExpression(branch) {
                // per ES grammar both 'whenTrue' and 'whenFalse' parts of conditional expression are assignment expressions
                // so in case when comma expression is introduced as a part of previous transformations
                // if should be wrapped in parens since comma operator has the lowest precedence
                var emittedExpression = ts.skipPartiallyEmittedExpressions(branch);
                return ts.isCommaSequence(emittedExpression)
                    ? factory.createParenthesizedExpression(branch)
                    : branch;
            }
            /**
             *  [Per the spec](https://tc39.github.io/ecma262/#prod-ExportDeclaration), `export default` accepts _AssigmentExpression_ but
             *  has a lookahead restriction for `function`, `async function`, and `class`.
             *
             * Basically, that means we need to parenthesize in the following cases:
             *
             * - BinaryExpression of CommaToken
             * - CommaList (synthetic list of multiple comma expressions)
             * - FunctionExpression
             * - ClassExpression
             */
            function parenthesizeExpressionOfExportDefault(expression) {
                var check = ts.skipPartiallyEmittedExpressions(expression);
                var needsParens = ts.isCommaSequence(check);
                if (!needsParens) {
                    switch (ts.getLeftmostExpression(check, /*stopAtCallExpression*/ false).kind) {
                        case 224 /* ClassExpression */:
                        case 211 /* FunctionExpression */:
                            needsParens = true;
                    }
                }
                return needsParens ? factory.createParenthesizedExpression(expression) : expression;
            }
            /**
             * Wraps an expression in parentheses if it is needed in order to use the expression
             * as the expression of a `NewExpression` node.
             */
            function parenthesizeExpressionOfNew(expression) {
                var leftmostExpr = ts.getLeftmostExpression(expression, /*stopAtCallExpressions*/ true);
                switch (leftmostExpr.kind) {
                    case 206 /* CallExpression */:
                        return factory.createParenthesizedExpression(expression);
                    case 207 /* NewExpression */:
                        return !leftmostExpr.arguments
                            ? factory.createParenthesizedExpression(expression)
                            : expression; // TODO(rbuckton): Verify this assertion holds
                }
                return parenthesizeLeftSideOfAccess(expression);
            }
            /**
             * Wraps an expression in parentheses if it is needed in order to use the expression for
             * property or element access.
             */
            function parenthesizeLeftSideOfAccess(expression) {
                // isLeftHandSideExpression is almost the correct criterion for when it is not necessary
                // to parenthesize the expression before a dot. The known exception is:
                //
                //    NewExpression:
                //       new C.x        -> not the same as (new C).x
                //
                var emittedExpression = ts.skipPartiallyEmittedExpressions(expression);
                if (ts.isLeftHandSideExpression(emittedExpression)
                    && (emittedExpression.kind !== 207 /* NewExpression */ || emittedExpression.arguments)) {
                    // TODO(rbuckton): Verify whether this assertion holds.
                    return expression;
                }
                // TODO(rbuckton): Verifiy whether `setTextRange` is needed.
                return ts.setTextRange(factory.createParenthesizedExpression(expression), expression);
            }
            function parenthesizeOperandOfPostfixUnary(operand) {
                // TODO(rbuckton): Verifiy whether `setTextRange` is needed.
                return ts.isLeftHandSideExpression(operand) ? operand : ts.setTextRange(factory.createParenthesizedExpression(operand), operand);
            }
            function parenthesizeOperandOfPrefixUnary(operand) {
                // TODO(rbuckton): Verifiy whether `setTextRange` is needed.
                return ts.isUnaryExpression(operand) ? operand : ts.setTextRange(factory.createParenthesizedExpression(operand), operand);
            }
            function parenthesizeExpressionsOfCommaDelimitedList(elements) {
                var result = ts.sameMap(elements, parenthesizeExpressionForDisallowedComma);
                return ts.setTextRange(factory.createNodeArray(result, elements.hasTrailingComma), elements);
            }
            function parenthesizeExpressionForDisallowedComma(expression) {
                var emittedExpression = ts.skipPartiallyEmittedExpressions(expression);
                var expressionPrecedence = ts.getExpressionPrecedence(emittedExpression);
                var commaPrecedence = ts.getOperatorPrecedence(219 /* BinaryExpression */, 27 /* CommaToken */);
                // TODO(rbuckton): Verifiy whether `setTextRange` is needed.
                return expressionPrecedence > commaPrecedence ? expression : ts.setTextRange(factory.createParenthesizedExpression(expression), expression);
            }
            function parenthesizeExpressionOfExpressionStatement(expression) {
                var emittedExpression = ts.skipPartiallyEmittedExpressions(expression);
                if (ts.isCallExpression(emittedExpression)) {
                    var callee = emittedExpression.expression;
                    var kind = ts.skipPartiallyEmittedExpressions(callee).kind;
                    if (kind === 211 /* FunctionExpression */ || kind === 212 /* ArrowFunction */) {
                        // TODO(rbuckton): Verifiy whether `setTextRange` is needed.
                        var updated = factory.updateCallExpression(emittedExpression, ts.setTextRange(factory.createParenthesizedExpression(callee), callee), emittedExpression.typeArguments, emittedExpression.arguments);
                        return factory.restoreOuterExpressions(expression, updated, 8 /* PartiallyEmittedExpressions */);
                    }
                }
                var leftmostExpressionKind = ts.getLeftmostExpression(emittedExpression, /*stopAtCallExpressions*/ false).kind;
                if (leftmostExpressionKind === 203 /* ObjectLiteralExpression */ || leftmostExpressionKind === 211 /* FunctionExpression */) {
                    // TODO(rbuckton): Verifiy whether `setTextRange` is needed.
                    return ts.setTextRange(factory.createParenthesizedExpression(expression), expression);
                }
                return expression;
            }
            function parenthesizeConciseBodyOfArrowFunction(body) {
                if (!ts.isBlock(body) && (ts.isCommaSequence(body) || ts.getLeftmostExpression(body, /*stopAtCallExpressions*/ false).kind === 203 /* ObjectLiteralExpression */)) {
                    // TODO(rbuckton): Verifiy whether `setTextRange` is needed.
                    return ts.setTextRange(factory.createParenthesizedExpression(body), body);
                }
                return body;
            }
            function parenthesizeMemberOfConditionalType(member) {
                return member.kind === 187 /* ConditionalType */ ? factory.createParenthesizedType(member) : member;
            }
            function parenthesizeMemberOfElementType(member) {
                switch (member.kind) {
                    case 185 /* UnionType */:
                    case 186 /* IntersectionType */:
                    case 177 /* FunctionType */:
                    case 178 /* ConstructorType */:
                        return factory.createParenthesizedType(member);
                }
                return parenthesizeMemberOfConditionalType(member);
            }
            function parenthesizeElementTypeOfArrayType(member) {
                switch (member.kind) {
                    case 179 /* TypeQuery */:
                    case 191 /* TypeOperator */:
                    case 188 /* InferType */:
                        return factory.createParenthesizedType(member);
                }
                return parenthesizeMemberOfElementType(member);
            }
            function parenthesizeConstituentTypesOfUnionOrIntersectionType(members) {
                return factory.createNodeArray(ts.sameMap(members, parenthesizeMemberOfElementType));
            }
            function parenthesizeOrdinalTypeArgument(node, i) {
                return i === 0 && ts.isFunctionOrConstructorTypeNode(node) && node.typeParameters ? factory.createParenthesizedType(node) : node;
            }
            function parenthesizeTypeArguments(typeArguments) {
                if (ts.some(typeArguments)) {
                    return factory.createNodeArray(ts.sameMap(typeArguments, parenthesizeOrdinalTypeArgument));
                }
            }
        }
        ts.createParenthesizerRules = createParenthesizerRules;
        ts.nullParenthesizerRules = {
            getParenthesizeLeftSideOfBinaryForOperator: function (_) { return ts.identity; },
            getParenthesizeRightSideOfBinaryForOperator: function (_) { return ts.identity; },
            parenthesizeLeftSideOfBinary: function (_binaryOperator, leftSide) { return leftSide; },
            parenthesizeRightSideOfBinary: function (_binaryOperator, _leftSide, rightSide) { return rightSide; },
            parenthesizeExpressionOfComputedPropertyName: ts.identity,
            parenthesizeConditionOfConditionalExpression: ts.identity,
            parenthesizeBranchOfConditionalExpression: ts.identity,
            parenthesizeExpressionOfExportDefault: ts.identity,
            parenthesizeExpressionOfNew: function (expression) { return ts.cast(expression, ts.isLeftHandSideExpression); },
            parenthesizeLeftSideOfAccess: function (expression) { return ts.cast(expression, ts.isLeftHandSideExpression); },
            parenthesizeOperandOfPostfixUnary: function (operand) { return ts.cast(operand, ts.isLeftHandSideExpression); },
            parenthesizeOperandOfPrefixUnary: function (operand) { return ts.cast(operand, ts.isUnaryExpression); },
            parenthesizeExpressionsOfCommaDelimitedList: function (nodes) { return ts.cast(nodes, ts.isNodeArray); },
            parenthesizeExpressionForDisallowedComma: ts.identity,
            parenthesizeExpressionOfExpressionStatement: ts.identity,
            parenthesizeConciseBodyOfArrowFunction: ts.identity,
            parenthesizeMemberOfConditionalType: ts.identity,
            parenthesizeMemberOfElementType: ts.identity,
            parenthesizeElementTypeOfArrayType: ts.identity,
            parenthesizeConstituentTypesOfUnionOrIntersectionType: function (nodes) { return ts.cast(nodes, ts.isNodeArray); },
            parenthesizeTypeArguments: function (nodes) { return nodes && ts.cast(nodes, ts.isNodeArray); },
        };
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function createNodeConverters(factory) {
            return {
                convertToFunctionBlock: convertToFunctionBlock,
                convertToFunctionExpression: convertToFunctionExpression,
                convertToArrayAssignmentElement: convertToArrayAssignmentElement,
                convertToObjectAssignmentElement: convertToObjectAssignmentElement,
                convertToAssignmentPattern: convertToAssignmentPattern,
                convertToObjectAssignmentPattern: convertToObjectAssignmentPattern,
                convertToArrayAssignmentPattern: convertToArrayAssignmentPattern,
                convertToAssignmentElementTarget: convertToAssignmentElementTarget,
            };
            function convertToFunctionBlock(node, multiLine) {
                if (ts.isBlock(node))
                    return node;
                var returnStatement = factory.createReturnStatement(node);
                ts.setTextRange(returnStatement, node);
                var body = factory.createBlock([returnStatement], multiLine);
                ts.setTextRange(body, node);
                return body;
            }
            function convertToFunctionExpression(node) {
                if (!node.body)
                    return ts.Debug.fail("Cannot convert a FunctionDeclaration without a body");
                var updated = factory.createFunctionExpression(node.modifiers, node.asteriskToken, node.name, node.typeParameters, node.parameters, node.type, node.body);
                ts.setOriginalNode(updated, node);
                ts.setTextRange(updated, node);
                if (ts.getStartsOnNewLine(node)) {
                    ts.setStartsOnNewLine(updated, /*newLine*/ true);
                }
                return updated;
            }
            function convertToArrayAssignmentElement(element) {
                if (ts.isBindingElement(element)) {
                    if (element.dotDotDotToken) {
                        ts.Debug.assertNode(element.name, ts.isIdentifier);
                        return ts.setOriginalNode(ts.setTextRange(factory.createSpreadElement(element.name), element), element);
                    }
                    var expression = convertToAssignmentElementTarget(element.name);
                    return element.initializer
                        ? ts.setOriginalNode(ts.setTextRange(factory.createAssignment(expression, element.initializer), element), element)
                        : expression;
                }
                return ts.cast(element, ts.isExpression);
            }
            function convertToObjectAssignmentElement(element) {
                if (ts.isBindingElement(element)) {
                    if (element.dotDotDotToken) {
                        ts.Debug.assertNode(element.name, ts.isIdentifier);
                        return ts.setOriginalNode(ts.setTextRange(factory.createSpreadAssignment(element.name), element), element);
                    }
                    if (element.propertyName) {
                        var expression = convertToAssignmentElementTarget(element.name);
                        return ts.setOriginalNode(ts.setTextRange(factory.createPropertyAssignment(element.propertyName, element.initializer ? factory.createAssignment(expression, element.initializer) : expression), element), element);
                    }
                    ts.Debug.assertNode(element.name, ts.isIdentifier);
                    return ts.setOriginalNode(ts.setTextRange(factory.createShorthandPropertyAssignment(element.name, element.initializer), element), element);
                }
                return ts.cast(element, ts.isObjectLiteralElementLike);
            }
            function convertToAssignmentPattern(node) {
                switch (node.kind) {
                    case 200 /* ArrayBindingPattern */:
                    case 202 /* ArrayLiteralExpression */:
                        return convertToArrayAssignmentPattern(node);
                    case 199 /* ObjectBindingPattern */:
                    case 203 /* ObjectLiteralExpression */:
                        return convertToObjectAssignmentPattern(node);
                }
            }
            function convertToObjectAssignmentPattern(node) {
                if (ts.isObjectBindingPattern(node)) {
                    return ts.setOriginalNode(ts.setTextRange(factory.createObjectLiteralExpression(ts.map(node.elements, convertToObjectAssignmentElement)), node), node);
                }
                return ts.cast(node, ts.isObjectLiteralExpression);
            }
            function convertToArrayAssignmentPattern(node) {
                if (ts.isArrayBindingPattern(node)) {
                    return ts.setOriginalNode(ts.setTextRange(factory.createArrayLiteralExpression(ts.map(node.elements, convertToArrayAssignmentElement)), node), node);
                }
                return ts.cast(node, ts.isArrayLiteralExpression);
            }
            function convertToAssignmentElementTarget(node) {
                if (ts.isBindingPattern(node)) {
                    return convertToAssignmentPattern(node);
                }
                return ts.cast(node, ts.isExpression);
            }
        }
        ts.createNodeConverters = createNodeConverters;
        ts.nullNodeConverters = {
            convertToFunctionBlock: ts.notImplemented,
            convertToFunctionExpression: ts.notImplemented,
            convertToArrayAssignmentElement: ts.notImplemented,
            convertToObjectAssignmentElement: ts.notImplemented,
            convertToAssignmentPattern: ts.notImplemented,
            convertToObjectAssignmentPattern: ts.notImplemented,
            convertToArrayAssignmentPattern: ts.notImplemented,
            convertToAssignmentElementTarget: ts.notImplemented,
        };
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        var nextAutoGenerateId = 0;
        /* @internal */
        var NodeFactoryFlags;
        (function (NodeFactoryFlags) {
            NodeFactoryFlags[NodeFactoryFlags["None"] = 0] = "None";
            // Disables the parenthesizer rules for the factory.
            NodeFactoryFlags[NodeFactoryFlags["NoParenthesizerRules"] = 1] = "NoParenthesizerRules";
            // Disables the node converters for the factory.
            NodeFactoryFlags[NodeFactoryFlags["NoNodeConverters"] = 2] = "NoNodeConverters";
            // Ensures new `PropertyAccessExpression` nodes are created with the `NoIndentation` emit flag set.
            NodeFactoryFlags[NodeFactoryFlags["NoIndentationOnFreshPropertyAccess"] = 4] = "NoIndentationOnFreshPropertyAccess";
            // Do not set an `original` pointer when updating a node.
            NodeFactoryFlags[NodeFactoryFlags["NoOriginalNode"] = 8] = "NoOriginalNode";
        })(NodeFactoryFlags = ts.NodeFactoryFlags || (ts.NodeFactoryFlags = {}));
        /**
         * Creates a `NodeFactory` that can be used to create and update a syntax tree.
         * @param flags Flags that control factory behavior.
         * @param baseFactory A `BaseNodeFactory` used to create the base `Node` objects.
         */
        /* @internal */
        function createNodeFactory(flags, baseFactory) {
            var update = flags & 8 /* NoOriginalNode */ ? updateWithoutOriginal : updateWithOriginal;
            // Lazily load the parenthesizer, node converters, and some factory methods until they are used.
            var parenthesizerRules = ts.memoize(function () { return flags & 1 /* NoParenthesizerRules */ ? ts.nullParenthesizerRules : ts.createParenthesizerRules(factory); });
            var converters = ts.memoize(function () { return flags & 2 /* NoNodeConverters */ ? ts.nullNodeConverters : ts.createNodeConverters(factory); });
            // lazy initializaton of common operator factories
            var getBinaryCreateFunction = ts.memoizeOne(function (operator) { return function (left, right) { return createBinaryExpression(left, operator, right); }; });
            var getPrefixUnaryCreateFunction = ts.memoizeOne(function (operator) { return function (operand) { return createPrefixUnaryExpression(operator, operand); }; });
            var getPostfixUnaryCreateFunction = ts.memoizeOne(function (operator) { return function (operand) { return createPostfixUnaryExpression(operand, operator); }; });
            var getJSDocPrimaryTypeCreateFunction = ts.memoizeOne(function (kind) { return function () { return createJSDocPrimaryTypeWorker(kind); }; });
            var getJSDocUnaryTypeCreateFunction = ts.memoizeOne(function (kind) { return function (type) { return createJSDocUnaryTypeWorker(kind, type); }; });
            var getJSDocUnaryTypeUpdateFunction = ts.memoizeOne(function (kind) { return function (node, type) { return updateJSDocUnaryTypeWorker(kind, node, type); }; });
            var getJSDocSimpleTagCreateFunction = ts.memoizeOne(function (kind) { return function (tagName, comment) { return createJSDocSimpleTagWorker(kind, tagName, comment); }; });
            var getJSDocSimpleTagUpdateFunction = ts.memoizeOne(function (kind) { return function (node, tagName, comment) { return updateJSDocSimpleTagWorker(kind, node, tagName, comment); }; });
            var getJSDocTypeLikeTagCreateFunction = ts.memoizeOne(function (kind) { return function (tagName, typeExpression, comment) { return createJSDocTypeLikeTagWorker(kind, tagName, typeExpression, comment); }; });
            var getJSDocTypeLikeTagUpdateFunction = ts.memoizeOne(function (kind) { return function (node, tagName, typeExpression, comment) { return updateJSDocTypeLikeTagWorker(kind, node, tagName, typeExpression, comment); }; });
            var factory = {
                get parenthesizer() { return parenthesizerRules(); },
                get converters() { return converters(); },
                createNodeArray: createNodeArray,
                createNumericLiteral: createNumericLiteral,
                createBigIntLiteral: createBigIntLiteral,
                createStringLiteral: createStringLiteral,
                createStringLiteralFromNode: createStringLiteralFromNode,
                createRegularExpressionLiteral: createRegularExpressionLiteral,
                createLiteralLikeNode: createLiteralLikeNode,
                createIdentifier: createIdentifier,
                updateIdentifier: updateIdentifier,
                createTempVariable: createTempVariable,
                createLoopVariable: createLoopVariable,
                createUniqueName: createUniqueName,
                getGeneratedNameForNode: getGeneratedNameForNode,
                createPrivateIdentifier: createPrivateIdentifier,
                createToken: createToken,
                createSuper: createSuper,
                createThis: createThis,
                createNull: createNull,
                createTrue: createTrue,
                createFalse: createFalse,
                createModifier: createModifier,
                createModifiersFromModifierFlags: createModifiersFromModifierFlags,
                createQualifiedName: createQualifiedName,
                updateQualifiedName: updateQualifiedName,
                createComputedPropertyName: createComputedPropertyName,
                updateComputedPropertyName: updateComputedPropertyName,
                createTypeParameterDeclaration: createTypeParameterDeclaration,
                updateTypeParameterDeclaration: updateTypeParameterDeclaration,
                createParameterDeclaration: createParameterDeclaration,
                updateParameterDeclaration: updateParameterDeclaration,
                createDecorator: createDecorator,
                updateDecorator: updateDecorator,
                createPropertySignature: createPropertySignature,
                updatePropertySignature: updatePropertySignature,
                createPropertyDeclaration: createPropertyDeclaration,
                updatePropertyDeclaration: updatePropertyDeclaration,
                createMethodSignature: createMethodSignature,
                updateMethodSignature: updateMethodSignature,
                createMethodDeclaration: createMethodDeclaration,
                updateMethodDeclaration: updateMethodDeclaration,
                createConstructorDeclaration: createConstructorDeclaration,
                updateConstructorDeclaration: updateConstructorDeclaration,
                createGetAccessorDeclaration: createGetAccessorDeclaration,
                updateGetAccessorDeclaration: updateGetAccessorDeclaration,
                createSetAccessorDeclaration: createSetAccessorDeclaration,
                updateSetAccessorDeclaration: updateSetAccessorDeclaration,
                createCallSignature: createCallSignature,
                updateCallSignature: updateCallSignature,
                createConstructSignature: createConstructSignature,
                updateConstructSignature: updateConstructSignature,
                createIndexSignature: createIndexSignature,
                updateIndexSignature: updateIndexSignature,
                createClassStaticBlockDeclaration: createClassStaticBlockDeclaration,
                updateClassStaticBlockDeclaration: updateClassStaticBlockDeclaration,
                createTemplateLiteralTypeSpan: createTemplateLiteralTypeSpan,
                updateTemplateLiteralTypeSpan: updateTemplateLiteralTypeSpan,
                createKeywordTypeNode: createKeywordTypeNode,
                createTypePredicateNode: createTypePredicateNode,
                updateTypePredicateNode: updateTypePredicateNode,
                createTypeReferenceNode: createTypeReferenceNode,
                updateTypeReferenceNode: updateTypeReferenceNode,
                createFunctionTypeNode: createFunctionTypeNode,
                updateFunctionTypeNode: updateFunctionTypeNode,
                createConstructorTypeNode: createConstructorTypeNode,
                updateConstructorTypeNode: updateConstructorTypeNode,
                createTypeQueryNode: createTypeQueryNode,
                updateTypeQueryNode: updateTypeQueryNode,
                createTypeLiteralNode: createTypeLiteralNode,
                updateTypeLiteralNode: updateTypeLiteralNode,
                createArrayTypeNode: createArrayTypeNode,
                updateArrayTypeNode: updateArrayTypeNode,
                createTupleTypeNode: createTupleTypeNode,
                updateTupleTypeNode: updateTupleTypeNode,
                createNamedTupleMember: createNamedTupleMember,
                updateNamedTupleMember: updateNamedTupleMember,
                createOptionalTypeNode: createOptionalTypeNode,
                updateOptionalTypeNode: updateOptionalTypeNode,
                createRestTypeNode: createRestTypeNode,
                updateRestTypeNode: updateRestTypeNode,
                createUnionTypeNode: createUnionTypeNode,
                updateUnionTypeNode: updateUnionTypeNode,
                createIntersectionTypeNode: createIntersectionTypeNode,
                updateIntersectionTypeNode: updateIntersectionTypeNode,
                createConditionalTypeNode: createConditionalTypeNode,
                updateConditionalTypeNode: updateConditionalTypeNode,
                createInferTypeNode: createInferTypeNode,
                updateInferTypeNode: updateInferTypeNode,
                createImportTypeNode: createImportTypeNode,
                updateImportTypeNode: updateImportTypeNode,
                createParenthesizedType: createParenthesizedType,
                updateParenthesizedType: updateParenthesizedType,
                createThisTypeNode: createThisTypeNode,
                createTypeOperatorNode: createTypeOperatorNode,
                updateTypeOperatorNode: updateTypeOperatorNode,
                createIndexedAccessTypeNode: createIndexedAccessTypeNode,
                updateIndexedAccessTypeNode: updateIndexedAccessTypeNode,
                createMappedTypeNode: createMappedTypeNode,
                updateMappedTypeNode: updateMappedTypeNode,
                createLiteralTypeNode: createLiteralTypeNode,
                updateLiteralTypeNode: updateLiteralTypeNode,
                createTemplateLiteralType: createTemplateLiteralType,
                updateTemplateLiteralType: updateTemplateLiteralType,
                createObjectBindingPattern: createObjectBindingPattern,
                updateObjectBindingPattern: updateObjectBindingPattern,
                createArrayBindingPattern: createArrayBindingPattern,
                updateArrayBindingPattern: updateArrayBindingPattern,
                createBindingElement: createBindingElement,
                updateBindingElement: updateBindingElement,
                createArrayLiteralExpression: createArrayLiteralExpression,
                updateArrayLiteralExpression: updateArrayLiteralExpression,
                createObjectLiteralExpression: createObjectLiteralExpression,
                updateObjectLiteralExpression: updateObjectLiteralExpression,
                createPropertyAccessExpression: flags & 4 /* NoIndentationOnFreshPropertyAccess */ ?
                    function (expression, name) { return ts.setEmitFlags(createPropertyAccessExpression(expression, name), 131072 /* NoIndentation */); } :
                    createPropertyAccessExpression,
                updatePropertyAccessExpression: updatePropertyAccessExpression,
                createPropertyAccessChain: flags & 4 /* NoIndentationOnFreshPropertyAccess */ ?
                    function (expression, questionDotToken, name) { return ts.setEmitFlags(createPropertyAccessChain(expression, questionDotToken, name), 131072 /* NoIndentation */); } :
                    createPropertyAccessChain,
                updatePropertyAccessChain: updatePropertyAccessChain,
                createElementAccessExpression: createElementAccessExpression,
                updateElementAccessExpression: updateElementAccessExpression,
                createElementAccessChain: createElementAccessChain,
                updateElementAccessChain: updateElementAccessChain,
                createCallExpression: createCallExpression,
                updateCallExpression: updateCallExpression,
                createCallChain: createCallChain,
                updateCallChain: updateCallChain,
                createNewExpression: createNewExpression,
                updateNewExpression: updateNewExpression,
                createTaggedTemplateExpression: createTaggedTemplateExpression,
                updateTaggedTemplateExpression: updateTaggedTemplateExpression,
                createTypeAssertion: createTypeAssertion,
                updateTypeAssertion: updateTypeAssertion,
                createParenthesizedExpression: createParenthesizedExpression,
                updateParenthesizedExpression: updateParenthesizedExpression,
                createFunctionExpression: createFunctionExpression,
                updateFunctionExpression: updateFunctionExpression,
                createArrowFunction: createArrowFunction,
                updateArrowFunction: updateArrowFunction,
                createDeleteExpression: createDeleteExpression,
                updateDeleteExpression: updateDeleteExpression,
                createTypeOfExpression: createTypeOfExpression,
                updateTypeOfExpression: updateTypeOfExpression,
                createVoidExpression: createVoidExpression,
                updateVoidExpression: updateVoidExpression,
                createAwaitExpression: createAwaitExpression,
                updateAwaitExpression: updateAwaitExpression,
                createPrefixUnaryExpression: createPrefixUnaryExpression,
                updatePrefixUnaryExpression: updatePrefixUnaryExpression,
                createPostfixUnaryExpression: createPostfixUnaryExpression,
                updatePostfixUnaryExpression: updatePostfixUnaryExpression,
                createBinaryExpression: createBinaryExpression,
                updateBinaryExpression: updateBinaryExpression,
                createConditionalExpression: createConditionalExpression,
                updateConditionalExpression: updateConditionalExpression,
                createTemplateExpression: createTemplateExpression,
                updateTemplateExpression: updateTemplateExpression,
                createTemplateHead: createTemplateHead,
                createTemplateMiddle: createTemplateMiddle,
                createTemplateTail: createTemplateTail,
                createNoSubstitutionTemplateLiteral: createNoSubstitutionTemplateLiteral,
                createTemplateLiteralLikeNode: createTemplateLiteralLikeNode,
                createYieldExpression: createYieldExpression,
                updateYieldExpression: updateYieldExpression,
                createSpreadElement: createSpreadElement,
                updateSpreadElement: updateSpreadElement,
                createClassExpression: createClassExpression,
                updateClassExpression: updateClassExpression,
                createOmittedExpression: createOmittedExpression,
                createExpressionWithTypeArguments: createExpressionWithTypeArguments,
                updateExpressionWithTypeArguments: updateExpressionWithTypeArguments,
                createAsExpression: createAsExpression,
                updateAsExpression: updateAsExpression,
                createNonNullExpression: createNonNullExpression,
                updateNonNullExpression: updateNonNullExpression,
                createNonNullChain: createNonNullChain,
                updateNonNullChain: updateNonNullChain,
                createMetaProperty: createMetaProperty,
                updateMetaProperty: updateMetaProperty,
                createTemplateSpan: createTemplateSpan,
                updateTemplateSpan: updateTemplateSpan,
                createSemicolonClassElement: createSemicolonClassElement,
                createBlock: createBlock,
                updateBlock: updateBlock,
                createVariableStatement: createVariableStatement,
                updateVariableStatement: updateVariableStatement,
                createEmptyStatement: createEmptyStatement,
                createExpressionStatement: createExpressionStatement,
                updateExpressionStatement: updateExpressionStatement,
                createIfStatement: createIfStatement,
                updateIfStatement: updateIfStatement,
                createDoStatement: createDoStatement,
                updateDoStatement: updateDoStatement,
                createWhileStatement: createWhileStatement,
                updateWhileStatement: updateWhileStatement,
                createForStatement: createForStatement,
                updateForStatement: updateForStatement,
                createForInStatement: createForInStatement,
                updateForInStatement: updateForInStatement,
                createForOfStatement: createForOfStatement,
                updateForOfStatement: updateForOfStatement,
                createContinueStatement: createContinueStatement,
                updateContinueStatement: updateContinueStatement,
                createBreakStatement: createBreakStatement,
                updateBreakStatement: updateBreakStatement,
                createReturnStatement: createReturnStatement,
                updateReturnStatement: updateReturnStatement,
                createWithStatement: createWithStatement,
                updateWithStatement: updateWithStatement,
                createSwitchStatement: createSwitchStatement,
                updateSwitchStatement: updateSwitchStatement,
                createLabeledStatement: createLabeledStatement,
                updateLabeledStatement: updateLabeledStatement,
                createThrowStatement: createThrowStatement,
                updateThrowStatement: updateThrowStatement,
                createTryStatement: createTryStatement,
                updateTryStatement: updateTryStatement,
                createDebuggerStatement: createDebuggerStatement,
                createVariableDeclaration: createVariableDeclaration,
                updateVariableDeclaration: updateVariableDeclaration,
                createVariableDeclarationList: createVariableDeclarationList,
                updateVariableDeclarationList: updateVariableDeclarationList,
                createFunctionDeclaration: createFunctionDeclaration,
                updateFunctionDeclaration: updateFunctionDeclaration,
                createClassDeclaration: createClassDeclaration,
                updateClassDeclaration: updateClassDeclaration,
                createInterfaceDeclaration: createInterfaceDeclaration,
                updateInterfaceDeclaration: updateInterfaceDeclaration,
                createTypeAliasDeclaration: createTypeAliasDeclaration,
                updateTypeAliasDeclaration: updateTypeAliasDeclaration,
                createEnumDeclaration: createEnumDeclaration,
                updateEnumDeclaration: updateEnumDeclaration,
                createModuleDeclaration: createModuleDeclaration,
                updateModuleDeclaration: updateModuleDeclaration,
                createModuleBlock: createModuleBlock,
                updateModuleBlock: updateModuleBlock,
                createCaseBlock: createCaseBlock,
                updateCaseBlock: updateCaseBlock,
                createNamespaceExportDeclaration: createNamespaceExportDeclaration,
                updateNamespaceExportDeclaration: updateNamespaceExportDeclaration,
                createImportEqualsDeclaration: createImportEqualsDeclaration,
                updateImportEqualsDeclaration: updateImportEqualsDeclaration,
                createImportDeclaration: createImportDeclaration,
                updateImportDeclaration: updateImportDeclaration,
                createImportClause: createImportClause,
                updateImportClause: updateImportClause,
                createNamespaceImport: createNamespaceImport,
                updateNamespaceImport: updateNamespaceImport,
                createNamespaceExport: createNamespaceExport,
                updateNamespaceExport: updateNamespaceExport,
                createNamedImports: createNamedImports,
                updateNamedImports: updateNamedImports,
                createImportSpecifier: createImportSpecifier,
                updateImportSpecifier: updateImportSpecifier,
                createExportAssignment: createExportAssignment,
                updateExportAssignment: updateExportAssignment,
                createExportDeclaration: createExportDeclaration,
                updateExportDeclaration: updateExportDeclaration,
                createNamedExports: createNamedExports,
                updateNamedExports: updateNamedExports,
                createExportSpecifier: createExportSpecifier,
                updateExportSpecifier: updateExportSpecifier,
                createMissingDeclaration: createMissingDeclaration,
                createExternalModuleReference: createExternalModuleReference,
                updateExternalModuleReference: updateExternalModuleReference,
                // lazily load factory members for JSDoc types with similar structure
                get createJSDocAllType() { return getJSDocPrimaryTypeCreateFunction(307 /* JSDocAllType */); },
                get createJSDocUnknownType() { return getJSDocPrimaryTypeCreateFunction(308 /* JSDocUnknownType */); },
                get createJSDocNonNullableType() { return getJSDocUnaryTypeCreateFunction(310 /* JSDocNonNullableType */); },
                get updateJSDocNonNullableType() { return getJSDocUnaryTypeUpdateFunction(310 /* JSDocNonNullableType */); },
                get createJSDocNullableType() { return getJSDocUnaryTypeCreateFunction(309 /* JSDocNullableType */); },
                get updateJSDocNullableType() { return getJSDocUnaryTypeUpdateFunction(309 /* JSDocNullableType */); },
                get createJSDocOptionalType() { return getJSDocUnaryTypeCreateFunction(311 /* JSDocOptionalType */); },
                get updateJSDocOptionalType() { return getJSDocUnaryTypeUpdateFunction(311 /* JSDocOptionalType */); },
                get createJSDocVariadicType() { return getJSDocUnaryTypeCreateFunction(313 /* JSDocVariadicType */); },
                get updateJSDocVariadicType() { return getJSDocUnaryTypeUpdateFunction(313 /* JSDocVariadicType */); },
                get createJSDocNamepathType() { return getJSDocUnaryTypeCreateFunction(314 /* JSDocNamepathType */); },
                get updateJSDocNamepathType() { return getJSDocUnaryTypeUpdateFunction(314 /* JSDocNamepathType */); },
                createJSDocFunctionType: createJSDocFunctionType,
                updateJSDocFunctionType: updateJSDocFunctionType,
                createJSDocTypeLiteral: createJSDocTypeLiteral,
                updateJSDocTypeLiteral: updateJSDocTypeLiteral,
                createJSDocTypeExpression: createJSDocTypeExpression,
                updateJSDocTypeExpression: updateJSDocTypeExpression,
                createJSDocSignature: createJSDocSignature,
                updateJSDocSignature: updateJSDocSignature,
                createJSDocTemplateTag: createJSDocTemplateTag,
                updateJSDocTemplateTag: updateJSDocTemplateTag,
                createJSDocTypedefTag: createJSDocTypedefTag,
                updateJSDocTypedefTag: updateJSDocTypedefTag,
                createJSDocParameterTag: createJSDocParameterTag,
                updateJSDocParameterTag: updateJSDocParameterTag,
                createJSDocPropertyTag: createJSDocPropertyTag,
                updateJSDocPropertyTag: updateJSDocPropertyTag,
                createJSDocCallbackTag: createJSDocCallbackTag,
                updateJSDocCallbackTag: updateJSDocCallbackTag,
                createJSDocAugmentsTag: createJSDocAugmentsTag,
                updateJSDocAugmentsTag: updateJSDocAugmentsTag,
                createJSDocImplementsTag: createJSDocImplementsTag,
                updateJSDocImplementsTag: updateJSDocImplementsTag,
                createJSDocSeeTag: createJSDocSeeTag,
                updateJSDocSeeTag: updateJSDocSeeTag,
                createJSDocNameReference: createJSDocNameReference,
                updateJSDocNameReference: updateJSDocNameReference,
                createJSDocMemberName: createJSDocMemberName,
                updateJSDocMemberName: updateJSDocMemberName,
                createJSDocLink: createJSDocLink,
                updateJSDocLink: updateJSDocLink,
                createJSDocLinkCode: createJSDocLinkCode,
                updateJSDocLinkCode: updateJSDocLinkCode,
                createJSDocLinkPlain: createJSDocLinkPlain,
                updateJSDocLinkPlain: updateJSDocLinkPlain,
                // lazily load factory members for JSDoc tags with similar structure
                get createJSDocTypeTag() { return getJSDocTypeLikeTagCreateFunction(338 /* JSDocTypeTag */); },
                get updateJSDocTypeTag() { return getJSDocTypeLikeTagUpdateFunction(338 /* JSDocTypeTag */); },
                get createJSDocReturnTag() { return getJSDocTypeLikeTagCreateFunction(336 /* JSDocReturnTag */); },
                get updateJSDocReturnTag() { return getJSDocTypeLikeTagUpdateFunction(336 /* JSDocReturnTag */); },
                get createJSDocThisTag() { return getJSDocTypeLikeTagCreateFunction(337 /* JSDocThisTag */); },
                get updateJSDocThisTag() { return getJSDocTypeLikeTagUpdateFunction(337 /* JSDocThisTag */); },
                get createJSDocEnumTag() { return getJSDocTypeLikeTagCreateFunction(334 /* JSDocEnumTag */); },
                get updateJSDocEnumTag() { return getJSDocTypeLikeTagUpdateFunction(334 /* JSDocEnumTag */); },
                get createJSDocAuthorTag() { return getJSDocSimpleTagCreateFunction(325 /* JSDocAuthorTag */); },
                get updateJSDocAuthorTag() { return getJSDocSimpleTagUpdateFunction(325 /* JSDocAuthorTag */); },
                get createJSDocClassTag() { return getJSDocSimpleTagCreateFunction(327 /* JSDocClassTag */); },
                get updateJSDocClassTag() { return getJSDocSimpleTagUpdateFunction(327 /* JSDocClassTag */); },
                get createJSDocPublicTag() { return getJSDocSimpleTagCreateFunction(328 /* JSDocPublicTag */); },
                get updateJSDocPublicTag() { return getJSDocSimpleTagUpdateFunction(328 /* JSDocPublicTag */); },
                get createJSDocPrivateTag() { return getJSDocSimpleTagCreateFunction(329 /* JSDocPrivateTag */); },
                get updateJSDocPrivateTag() { return getJSDocSimpleTagUpdateFunction(329 /* JSDocPrivateTag */); },
                get createJSDocProtectedTag() { return getJSDocSimpleTagCreateFunction(330 /* JSDocProtectedTag */); },
                get updateJSDocProtectedTag() { return getJSDocSimpleTagUpdateFunction(330 /* JSDocProtectedTag */); },
                get createJSDocReadonlyTag() { return getJSDocSimpleTagCreateFunction(331 /* JSDocReadonlyTag */); },
                get updateJSDocReadonlyTag() { return getJSDocSimpleTagUpdateFunction(331 /* JSDocReadonlyTag */); },
                get createJSDocOverrideTag() { return getJSDocSimpleTagCreateFunction(332 /* JSDocOverrideTag */); },
                get updateJSDocOverrideTag() { return getJSDocSimpleTagUpdateFunction(332 /* JSDocOverrideTag */); },
                get createJSDocDeprecatedTag() { return getJSDocSimpleTagCreateFunction(326 /* JSDocDeprecatedTag */); },
                get updateJSDocDeprecatedTag() { return getJSDocSimpleTagUpdateFunction(326 /* JSDocDeprecatedTag */); },
                createJSDocUnknownTag: createJSDocUnknownTag,
                updateJSDocUnknownTag: updateJSDocUnknownTag,
                createJSDocText: createJSDocText,
                updateJSDocText: updateJSDocText,
                createJSDocComment: createJSDocComment,
                updateJSDocComment: updateJSDocComment,
                createJsxElement: createJsxElement,
                updateJsxElement: updateJsxElement,
                createJsxSelfClosingElement: createJsxSelfClosingElement,
                updateJsxSelfClosingElement: updateJsxSelfClosingElement,
                createJsxOpeningElement: createJsxOpeningElement,
                updateJsxOpeningElement: updateJsxOpeningElement,
                createJsxClosingElement: createJsxClosingElement,
                updateJsxClosingElement: updateJsxClosingElement,
                createJsxFragment: createJsxFragment,
                createJsxText: createJsxText,
                updateJsxText: updateJsxText,
                createJsxOpeningFragment: createJsxOpeningFragment,
                createJsxJsxClosingFragment: createJsxJsxClosingFragment,
                updateJsxFragment: updateJsxFragment,
                createJsxAttribute: createJsxAttribute,
                updateJsxAttribute: updateJsxAttribute,
                createJsxAttributes: createJsxAttributes,
                updateJsxAttributes: updateJsxAttributes,
                createJsxSpreadAttribute: createJsxSpreadAttribute,
                updateJsxSpreadAttribute: updateJsxSpreadAttribute,
                createJsxExpression: createJsxExpression,
                updateJsxExpression: updateJsxExpression,
                createCaseClause: createCaseClause,
                updateCaseClause: updateCaseClause,
                createDefaultClause: createDefaultClause,
                updateDefaultClause: updateDefaultClause,
                createHeritageClause: createHeritageClause,
                updateHeritageClause: updateHeritageClause,
                createCatchClause: createCatchClause,
                updateCatchClause: updateCatchClause,
                createPropertyAssignment: createPropertyAssignment,
                updatePropertyAssignment: updatePropertyAssignment,
                createShorthandPropertyAssignment: createShorthandPropertyAssignment,
                updateShorthandPropertyAssignment: updateShorthandPropertyAssignment,
                createSpreadAssignment: createSpreadAssignment,
                updateSpreadAssignment: updateSpreadAssignment,
                createEnumMember: createEnumMember,
                updateEnumMember: updateEnumMember,
                createSourceFile: createSourceFile,
                updateSourceFile: updateSourceFile,
                createBundle: createBundle,
                updateBundle: updateBundle,
                createUnparsedSource: createUnparsedSource,
                createUnparsedPrologue: createUnparsedPrologue,
                createUnparsedPrepend: createUnparsedPrepend,
                createUnparsedTextLike: createUnparsedTextLike,
                createUnparsedSyntheticReference: createUnparsedSyntheticReference,
                createInputFiles: createInputFiles,
                createSyntheticExpression: createSyntheticExpression,
                createSyntaxList: createSyntaxList,
                createNotEmittedStatement: createNotEmittedStatement,
                createPartiallyEmittedExpression: createPartiallyEmittedExpression,
                updatePartiallyEmittedExpression: updatePartiallyEmittedExpression,
                createCommaListExpression: createCommaListExpression,
                updateCommaListExpression: updateCommaListExpression,
                createEndOfDeclarationMarker: createEndOfDeclarationMarker,
                createMergeDeclarationMarker: createMergeDeclarationMarker,
                createSyntheticReferenceExpression: createSyntheticReferenceExpression,
                updateSyntheticReferenceExpression: updateSyntheticReferenceExpression,
                cloneNode: cloneNode,
                // Lazily load factory methods for common operator factories and utilities
                get createComma() { return getBinaryCreateFunction(27 /* CommaToken */); },
                get createAssignment() { return getBinaryCreateFunction(63 /* EqualsToken */); },
                get createLogicalOr() { return getBinaryCreateFunction(56 /* BarBarToken */); },
                get createLogicalAnd() { return getBinaryCreateFunction(55 /* AmpersandAmpersandToken */); },
                get createBitwiseOr() { return getBinaryCreateFunction(51 /* BarToken */); },
                get createBitwiseXor() { return getBinaryCreateFunction(52 /* CaretToken */); },
                get createBitwiseAnd() { return getBinaryCreateFunction(50 /* AmpersandToken */); },
                get createStrictEquality() { return getBinaryCreateFunction(36 /* EqualsEqualsEqualsToken */); },
                get createStrictInequality() { return getBinaryCreateFunction(37 /* ExclamationEqualsEqualsToken */); },
                get createEquality() { return getBinaryCreateFunction(34 /* EqualsEqualsToken */); },
                get createInequality() { return getBinaryCreateFunction(35 /* ExclamationEqualsToken */); },
                get createLessThan() { return getBinaryCreateFunction(29 /* LessThanToken */); },
                get createLessThanEquals() { return getBinaryCreateFunction(32 /* LessThanEqualsToken */); },
                get createGreaterThan() { return getBinaryCreateFunction(31 /* GreaterThanToken */); },
                get createGreaterThanEquals() { return getBinaryCreateFunction(33 /* GreaterThanEqualsToken */); },
                get createLeftShift() { return getBinaryCreateFunction(47 /* LessThanLessThanToken */); },
                get createRightShift() { return getBinaryCreateFunction(48 /* GreaterThanGreaterThanToken */); },
                get createUnsignedRightShift() { return getBinaryCreateFunction(49 /* GreaterThanGreaterThanGreaterThanToken */); },
                get createAdd() { return getBinaryCreateFunction(39 /* PlusToken */); },
                get createSubtract() { return getBinaryCreateFunction(40 /* MinusToken */); },
                get createMultiply() { return getBinaryCreateFunction(41 /* AsteriskToken */); },
                get createDivide() { return getBinaryCreateFunction(43 /* SlashToken */); },
                get createModulo() { return getBinaryCreateFunction(44 /* PercentToken */); },
                get createExponent() { return getBinaryCreateFunction(42 /* AsteriskAsteriskToken */); },
                get createPrefixPlus() { return getPrefixUnaryCreateFunction(39 /* PlusToken */); },
                get createPrefixMinus() { return getPrefixUnaryCreateFunction(40 /* MinusToken */); },
                get createPrefixIncrement() { return getPrefixUnaryCreateFunction(45 /* PlusPlusToken */); },
                get createPrefixDecrement() { return getPrefixUnaryCreateFunction(46 /* MinusMinusToken */); },
                get createBitwiseNot() { return getPrefixUnaryCreateFunction(54 /* TildeToken */); },
                get createLogicalNot() { return getPrefixUnaryCreateFunction(53 /* ExclamationToken */); },
                get createPostfixIncrement() { return getPostfixUnaryCreateFunction(45 /* PlusPlusToken */); },
                get createPostfixDecrement() { return getPostfixUnaryCreateFunction(46 /* MinusMinusToken */); },
                // Compound nodes
                createImmediatelyInvokedFunctionExpression: createImmediatelyInvokedFunctionExpression,
                createImmediatelyInvokedArrowFunction: createImmediatelyInvokedArrowFunction,
                createVoidZero: createVoidZero,
                createExportDefault: createExportDefault,
                createExternalModuleExport: createExternalModuleExport,
                createTypeCheck: createTypeCheck,
                createMethodCall: createMethodCall,
                createGlobalMethodCall: createGlobalMethodCall,
                createFunctionBindCall: createFunctionBindCall,
                createFunctionCallCall: createFunctionCallCall,
                createFunctionApplyCall: createFunctionApplyCall,
                createArraySliceCall: createArraySliceCall,
                createArrayConcatCall: createArrayConcatCall,
                createObjectDefinePropertyCall: createObjectDefinePropertyCall,
                createReflectGetCall: createReflectGetCall,
                createReflectSetCall: createReflectSetCall,
                createPropertyDescriptor: createPropertyDescriptor,
                createCallBinding: createCallBinding,
                createAssignmentTargetWrapper: createAssignmentTargetWrapper,
                // Utilities
                inlineExpressions: inlineExpressions,
                getInternalName: getInternalName,
                getLocalName: getLocalName,
                getExportName: getExportName,
                getDeclarationName: getDeclarationName,
                getNamespaceMemberName: getNamespaceMemberName,
                getExternalModuleOrNamespaceExportName: getExternalModuleOrNamespaceExportName,
                restoreOuterExpressions: restoreOuterExpressions,
                restoreEnclosingLabel: restoreEnclosingLabel,
                createUseStrictPrologue: createUseStrictPrologue,
                copyPrologue: copyPrologue,
                copyStandardPrologue: copyStandardPrologue,
                copyCustomPrologue: copyCustomPrologue,
                ensureUseStrict: ensureUseStrict,
                liftToBlock: liftToBlock,
                mergeLexicalEnvironment: mergeLexicalEnvironment,
                updateModifiers: updateModifiers,
            };
            return factory;
            // @api
            function createNodeArray(elements, hasTrailingComma) {
                if (elements === undefined || elements === ts.emptyArray) {
                    elements = [];
                }
                else if (ts.isNodeArray(elements)) {
                    if (hasTrailingComma === undefined || elements.hasTrailingComma === hasTrailingComma) {
                        // Ensure the transform flags have been aggregated for this NodeArray
                        if (elements.transformFlags === undefined) {
                            aggregateChildrenFlags(elements);
                        }
                        ts.Debug.attachNodeArrayDebugInfo(elements);
                        return elements;
                    }
                    // This *was* a `NodeArray`, but the `hasTrailingComma` option differs. Recreate the
                    // array with the same elements, text range, and transform flags but with the updated
                    // value for `hasTrailingComma`
                    var array_8 = elements.slice();
                    array_8.pos = elements.pos;
                    array_8.end = elements.end;
                    array_8.hasTrailingComma = hasTrailingComma;
                    array_8.transformFlags = elements.transformFlags;
                    ts.Debug.attachNodeArrayDebugInfo(array_8);
                    return array_8;
                }
                // Since the element list of a node array is typically created by starting with an empty array and
                // repeatedly calling push(), the list may not have the optimal memory layout. We invoke slice() for
                // small arrays (1 to 4 elements) to give the VM a chance to allocate an optimal representation.
                var length = elements.length;
                var array = (length >= 1 && length <= 4 ? elements.slice() : elements);
                ts.setTextRangePosEnd(array, -1, -1);
                array.hasTrailingComma = !!hasTrailingComma;
                aggregateChildrenFlags(array);
                ts.Debug.attachNodeArrayDebugInfo(array);
                return array;
            }
            function createBaseNode(kind) {
                return baseFactory.createBaseNode(kind);
            }
            function createBaseDeclaration(kind, decorators, modifiers) {
                var node = createBaseNode(kind);
                node.decorators = asNodeArray(decorators);
                node.modifiers = asNodeArray(modifiers);
                node.transformFlags |=
                    propagateChildrenFlags(node.decorators) |
                        propagateChildrenFlags(node.modifiers);
                // NOTE: The following properties are commonly set by the binder and are added here to
                // ensure declarations have a stable shape.
                node.symbol = undefined; // initialized by binder
                node.localSymbol = undefined; // initialized by binder
                node.locals = undefined; // initialized by binder
                node.nextContainer = undefined; // initialized by binder
                return node;
            }
            function createBaseNamedDeclaration(kind, decorators, modifiers, name) {
                var node = createBaseDeclaration(kind, decorators, modifiers);
                name = asName(name);
                node.name = name;
                // The PropertyName of a member is allowed to be `await`.
                // We don't need to exclude `await` for type signatures since types
                // don't propagate child flags.
                if (name) {
                    switch (node.kind) {
                        case 167 /* MethodDeclaration */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                        case 165 /* PropertyDeclaration */:
                        case 291 /* PropertyAssignment */:
                            if (ts.isIdentifier(name)) {
                                node.transformFlags |= propagateIdentifierNameFlags(name);
                                break;
                            }
                        // fall through
                        default:
                            node.transformFlags |= propagateChildFlags(name);
                            break;
                    }
                }
                return node;
            }
            function createBaseGenericNamedDeclaration(kind, decorators, modifiers, name, typeParameters) {
                var node = createBaseNamedDeclaration(kind, decorators, modifiers, name);
                node.typeParameters = asNodeArray(typeParameters);
                node.transformFlags |= propagateChildrenFlags(node.typeParameters);
                if (typeParameters)
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                return node;
            }
            function createBaseSignatureDeclaration(kind, decorators, modifiers, name, typeParameters, parameters, type) {
                var node = createBaseGenericNamedDeclaration(kind, decorators, modifiers, name, typeParameters);
                node.parameters = createNodeArray(parameters);
                node.type = type;
                node.transformFlags |=
                    propagateChildrenFlags(node.parameters) |
                        propagateChildFlags(node.type);
                if (type)
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                return node;
            }
            function updateBaseSignatureDeclaration(updated, original) {
                // copy children used only for error reporting
                if (original.typeArguments)
                    updated.typeArguments = original.typeArguments;
                return update(updated, original);
            }
            function createBaseFunctionLikeDeclaration(kind, decorators, modifiers, name, typeParameters, parameters, type, body) {
                var node = createBaseSignatureDeclaration(kind, decorators, modifiers, name, typeParameters, parameters, type);
                node.body = body;
                node.transformFlags |= propagateChildFlags(node.body) & ~16777216 /* ContainsPossibleTopLevelAwait */;
                if (!body)
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                return node;
            }
            function updateBaseFunctionLikeDeclaration(updated, original) {
                // copy children used only for error reporting
                if (original.exclamationToken)
                    updated.exclamationToken = original.exclamationToken;
                if (original.typeArguments)
                    updated.typeArguments = original.typeArguments;
                return updateBaseSignatureDeclaration(updated, original);
            }
            function createBaseInterfaceOrClassLikeDeclaration(kind, decorators, modifiers, name, typeParameters, heritageClauses) {
                var node = createBaseGenericNamedDeclaration(kind, decorators, modifiers, name, typeParameters);
                node.heritageClauses = asNodeArray(heritageClauses);
                node.transformFlags |= propagateChildrenFlags(node.heritageClauses);
                return node;
            }
            function createBaseClassLikeDeclaration(kind, decorators, modifiers, name, typeParameters, heritageClauses, members) {
                var node = createBaseInterfaceOrClassLikeDeclaration(kind, decorators, modifiers, name, typeParameters, heritageClauses);
                node.members = createNodeArray(members);
                node.transformFlags |= propagateChildrenFlags(node.members);
                return node;
            }
            function createBaseBindingLikeDeclaration(kind, decorators, modifiers, name, initializer) {
                var node = createBaseNamedDeclaration(kind, decorators, modifiers, name);
                node.initializer = initializer;
                node.transformFlags |= propagateChildFlags(node.initializer);
                return node;
            }
            function createBaseVariableLikeDeclaration(kind, decorators, modifiers, name, type, initializer) {
                var node = createBaseBindingLikeDeclaration(kind, decorators, modifiers, name, initializer);
                node.type = type;
                node.transformFlags |= propagateChildFlags(type);
                if (type)
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                return node;
            }
            //
            // Literals
            //
            function createBaseLiteral(kind, text) {
                var node = createBaseToken(kind);
                node.text = text;
                return node;
            }
            // @api
            function createNumericLiteral(value, numericLiteralFlags) {
                if (numericLiteralFlags === void 0) {
                    numericLiteralFlags = 0 /* None */;
                }
                var node = createBaseLiteral(8 /* NumericLiteral */, typeof value === "number" ? value + "" : value);
                node.numericLiteralFlags = numericLiteralFlags;
                if (numericLiteralFlags & 384 /* BinaryOrOctalSpecifier */)
                    node.transformFlags |= 512 /* ContainsES2015 */;
                return node;
            }
            // @api
            function createBigIntLiteral(value) {
                var node = createBaseLiteral(9 /* BigIntLiteral */, typeof value === "string" ? value : ts.pseudoBigIntToString(value) + "n");
                node.transformFlags |= 4 /* ContainsESNext */;
                return node;
            }
            function createBaseStringLiteral(text, isSingleQuote) {
                var node = createBaseLiteral(10 /* StringLiteral */, text);
                node.singleQuote = isSingleQuote;
                return node;
            }
            // @api
            function createStringLiteral(text, isSingleQuote, hasExtendedUnicodeEscape) {
                var node = createBaseStringLiteral(text, isSingleQuote);
                node.hasExtendedUnicodeEscape = hasExtendedUnicodeEscape;
                if (hasExtendedUnicodeEscape)
                    node.transformFlags |= 512 /* ContainsES2015 */;
                return node;
            }
            // @api
            function createStringLiteralFromNode(sourceNode) {
                var node = createBaseStringLiteral(ts.getTextOfIdentifierOrLiteral(sourceNode), /*isSingleQuote*/ undefined);
                node.textSourceNode = sourceNode;
                return node;
            }
            // @api
            function createRegularExpressionLiteral(text) {
                var node = createBaseLiteral(13 /* RegularExpressionLiteral */, text);
                return node;
            }
            // @api
            function createLiteralLikeNode(kind, text) {
                switch (kind) {
                    case 8 /* NumericLiteral */: return createNumericLiteral(text, /*numericLiteralFlags*/ 0);
                    case 9 /* BigIntLiteral */: return createBigIntLiteral(text);
                    case 10 /* StringLiteral */: return createStringLiteral(text, /*isSingleQuote*/ undefined);
                    case 11 /* JsxText */: return createJsxText(text, /*containsOnlyTriviaWhiteSpaces*/ false);
                    case 12 /* JsxTextAllWhiteSpaces */: return createJsxText(text, /*containsOnlyTriviaWhiteSpaces*/ true);
                    case 13 /* RegularExpressionLiteral */: return createRegularExpressionLiteral(text);
                    case 14 /* NoSubstitutionTemplateLiteral */: return createTemplateLiteralLikeNode(kind, text, /*rawText*/ undefined, /*templateFlags*/ 0);
                }
            }
            //
            // Identifiers
            //
            function createBaseIdentifier(text, originalKeywordKind) {
                if (originalKeywordKind === undefined && text) {
                    originalKeywordKind = ts.stringToToken(text);
                }
                if (originalKeywordKind === 79 /* Identifier */) {
                    originalKeywordKind = undefined;
                }
                var node = baseFactory.createBaseIdentifierNode(79 /* Identifier */);
                node.originalKeywordKind = originalKeywordKind;
                node.escapedText = ts.escapeLeadingUnderscores(text);
                return node;
            }
            function createBaseGeneratedIdentifier(text, autoGenerateFlags) {
                var node = createBaseIdentifier(text, /*originalKeywordKind*/ undefined);
                node.autoGenerateFlags = autoGenerateFlags;
                node.autoGenerateId = nextAutoGenerateId;
                nextAutoGenerateId++;
                return node;
            }
            // @api
            function createIdentifier(text, typeArguments, originalKeywordKind) {
                var node = createBaseIdentifier(text, originalKeywordKind);
                if (typeArguments) {
                    // NOTE: we do not use `setChildren` here because typeArguments in an identifier do not contribute to transformations
                    node.typeArguments = createNodeArray(typeArguments);
                }
                if (node.originalKeywordKind === 131 /* AwaitKeyword */) {
                    node.transformFlags |= 16777216 /* ContainsPossibleTopLevelAwait */;
                }
                return node;
            }
            // @api
            function updateIdentifier(node, typeArguments) {
                return node.typeArguments !== typeArguments
                    ? update(createIdentifier(ts.idText(node), typeArguments), node)
                    : node;
            }
            // @api
            function createTempVariable(recordTempVariable, reservedInNestedScopes) {
                var flags = 1 /* Auto */;
                if (reservedInNestedScopes)
                    flags |= 8 /* ReservedInNestedScopes */;
                var name = createBaseGeneratedIdentifier("", flags);
                if (recordTempVariable) {
                    recordTempVariable(name);
                }
                return name;
            }
            /** Create a unique temporary variable for use in a loop. */
            // @api
            function createLoopVariable(reservedInNestedScopes) {
                var flags = 2 /* Loop */;
                if (reservedInNestedScopes)
                    flags |= 8 /* ReservedInNestedScopes */;
                return createBaseGeneratedIdentifier("", flags);
            }
            /** Create a unique name based on the supplied text. */
            // @api
            function createUniqueName(text, flags) {
                if (flags === void 0) {
                    flags = 0 /* None */;
                }
                ts.Debug.assert(!(flags & 7 /* KindMask */), "Argument out of range: flags");
                ts.Debug.assert((flags & (16 /* Optimistic */ | 32 /* FileLevel */)) !== 32 /* FileLevel */, "GeneratedIdentifierFlags.FileLevel cannot be set without also setting GeneratedIdentifierFlags.Optimistic");
                return createBaseGeneratedIdentifier(text, 3 /* Unique */ | flags);
            }
            /** Create a unique name generated for a node. */
            // @api
            function getGeneratedNameForNode(node, flags) {
                if (flags === void 0) {
                    flags = 0;
                }
                ts.Debug.assert(!(flags & 7 /* KindMask */), "Argument out of range: flags");
                var name = createBaseGeneratedIdentifier(node && ts.isIdentifier(node) ? ts.idText(node) : "", 4 /* Node */ | flags);
                name.original = node;
                return name;
            }
            // @api
            function createPrivateIdentifier(text) {
                if (!ts.startsWith(text, "#"))
                    ts.Debug.fail("First character of private identifier must be #: " + text);
                var node = baseFactory.createBasePrivateIdentifierNode(80 /* PrivateIdentifier */);
                node.escapedText = ts.escapeLeadingUnderscores(text);
                node.transformFlags |= 8388608 /* ContainsClassFields */;
                return node;
            }
            //
            // Punctuation
            //
            function createBaseToken(kind) {
                return baseFactory.createBaseTokenNode(kind);
            }
            function createToken(token) {
                ts.Debug.assert(token >= 0 /* FirstToken */ && token <= 158 /* LastToken */, "Invalid token");
                ts.Debug.assert(token <= 14 /* FirstTemplateToken */ || token >= 17 /* LastTemplateToken */, "Invalid token. Use 'createTemplateLiteralLikeNode' to create template literals.");
                ts.Debug.assert(token <= 8 /* FirstLiteralToken */ || token >= 14 /* LastLiteralToken */, "Invalid token. Use 'createLiteralLikeNode' to create literals.");
                ts.Debug.assert(token !== 79 /* Identifier */, "Invalid token. Use 'createIdentifier' to create identifiers");
                var node = createBaseToken(token);
                var transformFlags = 0 /* None */;
                switch (token) {
                    case 130 /* AsyncKeyword */:
                        // 'async' modifier is ES2017 (async functions) or ES2018 (async generators)
                        transformFlags =
                            128 /* ContainsES2017 */ |
                                64 /* ContainsES2018 */;
                        break;
                    case 123 /* PublicKeyword */:
                    case 121 /* PrivateKeyword */:
                    case 122 /* ProtectedKeyword */:
                    case 143 /* ReadonlyKeyword */:
                    case 126 /* AbstractKeyword */:
                    case 134 /* DeclareKeyword */:
                    case 85 /* ConstKeyword */:
                    case 129 /* AnyKeyword */:
                    case 145 /* NumberKeyword */:
                    case 156 /* BigIntKeyword */:
                    case 142 /* NeverKeyword */:
                    case 146 /* ObjectKeyword */:
                    case 157 /* OverrideKeyword */:
                    case 148 /* StringKeyword */:
                    case 132 /* BooleanKeyword */:
                    case 149 /* SymbolKeyword */:
                    case 114 /* VoidKeyword */:
                    case 153 /* UnknownKeyword */:
                    case 151 /* UndefinedKeyword */: // `undefined` is an Identifier in the expression case.
                        transformFlags = 1 /* ContainsTypeScript */;
                        break;
                    case 106 /* SuperKeyword */:
                        transformFlags = 512 /* ContainsES2015 */ | 33554432 /* ContainsLexicalSuper */;
                        break;
                    case 124 /* StaticKeyword */:
                        transformFlags = 512 /* ContainsES2015 */;
                        break;
                    case 108 /* ThisKeyword */:
                        // 'this' indicates a lexical 'this'
                        transformFlags = 8192 /* ContainsLexicalThis */;
                        break;
                }
                if (transformFlags) {
                    node.transformFlags |= transformFlags;
                }
                return node;
            }
            //
            // Reserved words
            //
            // @api
            function createSuper() {
                return createToken(106 /* SuperKeyword */);
            }
            // @api
            function createThis() {
                return createToken(108 /* ThisKeyword */);
            }
            // @api
            function createNull() {
                return createToken(104 /* NullKeyword */);
            }
            // @api
            function createTrue() {
                return createToken(110 /* TrueKeyword */);
            }
            // @api
            function createFalse() {
                return createToken(95 /* FalseKeyword */);
            }
            //
            // Modifiers
            //
            // @api
            function createModifier(kind) {
                return createToken(kind);
            }
            // @api
            function createModifiersFromModifierFlags(flags) {
                var result = [];
                if (flags & 1 /* Export */) {
                    result.push(createModifier(93 /* ExportKeyword */));
                }
                if (flags & 2 /* Ambient */) {
                    result.push(createModifier(134 /* DeclareKeyword */));
                }
                if (flags & 512 /* Default */) {
                    result.push(createModifier(88 /* DefaultKeyword */));
                }
                if (flags & 2048 /* Const */) {
                    result.push(createModifier(85 /* ConstKeyword */));
                }
                if (flags & 4 /* Public */) {
                    result.push(createModifier(123 /* PublicKeyword */));
                }
                if (flags & 8 /* Private */) {
                    result.push(createModifier(121 /* PrivateKeyword */));
                }
                if (flags & 16 /* Protected */) {
                    result.push(createModifier(122 /* ProtectedKeyword */));
                }
                if (flags & 128 /* Abstract */) {
                    result.push(createModifier(126 /* AbstractKeyword */));
                }
                if (flags & 32 /* Static */) {
                    result.push(createModifier(124 /* StaticKeyword */));
                }
                if (flags & 16384 /* Override */) {
                    result.push(createModifier(157 /* OverrideKeyword */));
                }
                if (flags & 64 /* Readonly */) {
                    result.push(createModifier(143 /* ReadonlyKeyword */));
                }
                if (flags & 256 /* Async */) {
                    result.push(createModifier(130 /* AsyncKeyword */));
                }
                return result;
            }
            //
            // Names
            //
            // @api
            function createQualifiedName(left, right) {
                var node = createBaseNode(159 /* QualifiedName */);
                node.left = left;
                node.right = asName(right);
                node.transformFlags |=
                    propagateChildFlags(node.left) |
                        propagateIdentifierNameFlags(node.right);
                return node;
            }
            // @api
            function updateQualifiedName(node, left, right) {
                return node.left !== left
                    || node.right !== right
                    ? update(createQualifiedName(left, right), node)
                    : node;
            }
            // @api
            function createComputedPropertyName(expression) {
                var node = createBaseNode(160 /* ComputedPropertyName */);
                node.expression = parenthesizerRules().parenthesizeExpressionOfComputedPropertyName(expression);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        512 /* ContainsES2015 */ |
                        65536 /* ContainsComputedPropertyName */;
                return node;
            }
            // @api
            function updateComputedPropertyName(node, expression) {
                return node.expression !== expression
                    ? update(createComputedPropertyName(expression), node)
                    : node;
            }
            //
            // Signature elements
            //
            // @api
            function createTypeParameterDeclaration(name, constraint, defaultType) {
                var node = createBaseNamedDeclaration(161 /* TypeParameter */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, name);
                node.constraint = constraint;
                node.default = defaultType;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTypeParameterDeclaration(node, name, constraint, defaultType) {
                return node.name !== name
                    || node.constraint !== constraint
                    || node.default !== defaultType
                    ? update(createTypeParameterDeclaration(name, constraint, defaultType), node)
                    : node;
            }
            // @api
            function createParameterDeclaration(decorators, modifiers, dotDotDotToken, name, questionToken, type, initializer) {
                var node = createBaseVariableLikeDeclaration(162 /* Parameter */, decorators, modifiers, name, type, initializer && parenthesizerRules().parenthesizeExpressionForDisallowedComma(initializer));
                node.dotDotDotToken = dotDotDotToken;
                node.questionToken = questionToken;
                if (ts.isThisIdentifier(node.name)) {
                    node.transformFlags = 1 /* ContainsTypeScript */;
                }
                else {
                    node.transformFlags |=
                        propagateChildFlags(node.dotDotDotToken) |
                            propagateChildFlags(node.questionToken);
                    if (questionToken)
                        node.transformFlags |= 1 /* ContainsTypeScript */;
                    if (ts.modifiersToFlags(node.modifiers) & 16476 /* ParameterPropertyModifier */)
                        node.transformFlags |= 4096 /* ContainsTypeScriptClassSyntax */;
                    if (initializer || dotDotDotToken)
                        node.transformFlags |= 512 /* ContainsES2015 */;
                }
                return node;
            }
            // @api
            function updateParameterDeclaration(node, decorators, modifiers, dotDotDotToken, name, questionToken, type, initializer) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.dotDotDotToken !== dotDotDotToken
                    || node.name !== name
                    || node.questionToken !== questionToken
                    || node.type !== type
                    || node.initializer !== initializer
                    ? update(createParameterDeclaration(decorators, modifiers, dotDotDotToken, name, questionToken, type, initializer), node)
                    : node;
            }
            // @api
            function createDecorator(expression) {
                var node = createBaseNode(163 /* Decorator */);
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        1 /* ContainsTypeScript */ |
                        4096 /* ContainsTypeScriptClassSyntax */;
                return node;
            }
            // @api
            function updateDecorator(node, expression) {
                return node.expression !== expression
                    ? update(createDecorator(expression), node)
                    : node;
            }
            //
            // Type Elements
            //
            // @api
            function createPropertySignature(modifiers, name, questionToken, type) {
                var node = createBaseNamedDeclaration(164 /* PropertySignature */, 
                /*decorators*/ undefined, modifiers, name);
                node.type = type;
                node.questionToken = questionToken;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updatePropertySignature(node, modifiers, name, questionToken, type) {
                return node.modifiers !== modifiers
                    || node.name !== name
                    || node.questionToken !== questionToken
                    || node.type !== type
                    ? update(createPropertySignature(modifiers, name, questionToken, type), node)
                    : node;
            }
            // @api
            function createPropertyDeclaration(decorators, modifiers, name, questionOrExclamationToken, type, initializer) {
                var node = createBaseVariableLikeDeclaration(165 /* PropertyDeclaration */, decorators, modifiers, name, type, initializer);
                node.questionToken = questionOrExclamationToken && ts.isQuestionToken(questionOrExclamationToken) ? questionOrExclamationToken : undefined;
                node.exclamationToken = questionOrExclamationToken && ts.isExclamationToken(questionOrExclamationToken) ? questionOrExclamationToken : undefined;
                node.transformFlags |=
                    propagateChildFlags(node.questionToken) |
                        propagateChildFlags(node.exclamationToken) |
                        8388608 /* ContainsClassFields */;
                if (ts.isComputedPropertyName(node.name) || (ts.hasStaticModifier(node) && node.initializer)) {
                    node.transformFlags |= 4096 /* ContainsTypeScriptClassSyntax */;
                }
                if (questionOrExclamationToken || ts.modifiersToFlags(node.modifiers) & 2 /* Ambient */) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                return node;
            }
            // @api
            function updatePropertyDeclaration(node, decorators, modifiers, name, questionOrExclamationToken, type, initializer) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.name !== name
                    || node.questionToken !== (questionOrExclamationToken !== undefined && ts.isQuestionToken(questionOrExclamationToken) ? questionOrExclamationToken : undefined)
                    || node.exclamationToken !== (questionOrExclamationToken !== undefined && ts.isExclamationToken(questionOrExclamationToken) ? questionOrExclamationToken : undefined)
                    || node.type !== type
                    || node.initializer !== initializer
                    ? update(createPropertyDeclaration(decorators, modifiers, name, questionOrExclamationToken, type, initializer), node)
                    : node;
            }
            // @api
            function createMethodSignature(modifiers, name, questionToken, typeParameters, parameters, type) {
                var node = createBaseSignatureDeclaration(166 /* MethodSignature */, 
                /*decorators*/ undefined, modifiers, name, typeParameters, parameters, type);
                node.questionToken = questionToken;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateMethodSignature(node, modifiers, name, questionToken, typeParameters, parameters, type) {
                return node.modifiers !== modifiers
                    || node.name !== name
                    || node.questionToken !== questionToken
                    || node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    ? updateBaseSignatureDeclaration(createMethodSignature(modifiers, name, questionToken, typeParameters, parameters, type), node)
                    : node;
            }
            // @api
            function createMethodDeclaration(decorators, modifiers, asteriskToken, name, questionToken, typeParameters, parameters, type, body) {
                var node = createBaseFunctionLikeDeclaration(167 /* MethodDeclaration */, decorators, modifiers, name, typeParameters, parameters, type, body);
                node.asteriskToken = asteriskToken;
                node.questionToken = questionToken;
                node.transformFlags |=
                    propagateChildFlags(node.asteriskToken) |
                        propagateChildFlags(node.questionToken) |
                        512 /* ContainsES2015 */;
                if (questionToken) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                if (ts.modifiersToFlags(node.modifiers) & 256 /* Async */) {
                    if (asteriskToken) {
                        node.transformFlags |= 64 /* ContainsES2018 */;
                    }
                    else {
                        node.transformFlags |= 128 /* ContainsES2017 */;
                    }
                }
                else if (asteriskToken) {
                    node.transformFlags |= 1024 /* ContainsGenerator */;
                }
                return node;
            }
            // @api
            function updateMethodDeclaration(node, decorators, modifiers, asteriskToken, name, questionToken, typeParameters, parameters, type, body) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.asteriskToken !== asteriskToken
                    || node.name !== name
                    || node.questionToken !== questionToken
                    || node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    || node.body !== body
                    ? updateBaseFunctionLikeDeclaration(createMethodDeclaration(decorators, modifiers, asteriskToken, name, questionToken, typeParameters, parameters, type, body), node)
                    : node;
            }
            // @api
            function createClassStaticBlockDeclaration(decorators, modifiers, body) {
                var node = createBaseGenericNamedDeclaration(168 /* ClassStaticBlockDeclaration */, decorators, modifiers, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined);
                node.body = body;
                node.transformFlags = propagateChildFlags(body) | 8388608 /* ContainsClassFields */;
                return node;
            }
            // @api
            function updateClassStaticBlockDeclaration(node, decorators, modifiers, body) {
                return node.decorators !== decorators
                    || node.modifier !== modifiers
                    || node.body !== body
                    ? update(createClassStaticBlockDeclaration(decorators, modifiers, body), node)
                    : node;
            }
            // @api
            function createConstructorDeclaration(decorators, modifiers, parameters, body) {
                var node = createBaseFunctionLikeDeclaration(169 /* Constructor */, decorators, modifiers, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, parameters, 
                /*type*/ undefined, body);
                node.transformFlags |= 512 /* ContainsES2015 */;
                return node;
            }
            // @api
            function updateConstructorDeclaration(node, decorators, modifiers, parameters, body) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.parameters !== parameters
                    || node.body !== body
                    ? updateBaseFunctionLikeDeclaration(createConstructorDeclaration(decorators, modifiers, parameters, body), node)
                    : node;
            }
            // @api
            function createGetAccessorDeclaration(decorators, modifiers, name, parameters, type, body) {
                return createBaseFunctionLikeDeclaration(170 /* GetAccessor */, decorators, modifiers, name, 
                /*typeParameters*/ undefined, parameters, type, body);
            }
            // @api
            function updateGetAccessorDeclaration(node, decorators, modifiers, name, parameters, type, body) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.name !== name
                    || node.parameters !== parameters
                    || node.type !== type
                    || node.body !== body
                    ? updateBaseFunctionLikeDeclaration(createGetAccessorDeclaration(decorators, modifiers, name, parameters, type, body), node)
                    : node;
            }
            // @api
            function createSetAccessorDeclaration(decorators, modifiers, name, parameters, body) {
                return createBaseFunctionLikeDeclaration(171 /* SetAccessor */, decorators, modifiers, name, 
                /*typeParameters*/ undefined, parameters, 
                /*type*/ undefined, body);
            }
            // @api
            function updateSetAccessorDeclaration(node, decorators, modifiers, name, parameters, body) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.name !== name
                    || node.parameters !== parameters
                    || node.body !== body
                    ? updateBaseFunctionLikeDeclaration(createSetAccessorDeclaration(decorators, modifiers, name, parameters, body), node)
                    : node;
            }
            // @api
            function createCallSignature(typeParameters, parameters, type) {
                var node = createBaseSignatureDeclaration(172 /* CallSignature */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*name*/ undefined, typeParameters, parameters, type);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateCallSignature(node, typeParameters, parameters, type) {
                return node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    ? updateBaseSignatureDeclaration(createCallSignature(typeParameters, parameters, type), node)
                    : node;
            }
            // @api
            function createConstructSignature(typeParameters, parameters, type) {
                var node = createBaseSignatureDeclaration(173 /* ConstructSignature */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*name*/ undefined, typeParameters, parameters, type);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateConstructSignature(node, typeParameters, parameters, type) {
                return node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    ? updateBaseSignatureDeclaration(createConstructSignature(typeParameters, parameters, type), node)
                    : node;
            }
            // @api
            function createIndexSignature(decorators, modifiers, parameters, type) {
                var node = createBaseSignatureDeclaration(174 /* IndexSignature */, decorators, modifiers, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, parameters, type);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateIndexSignature(node, decorators, modifiers, parameters, type) {
                return node.parameters !== parameters
                    || node.type !== type
                    || node.decorators !== decorators
                    || node.modifiers !== modifiers
                    ? updateBaseSignatureDeclaration(createIndexSignature(decorators, modifiers, parameters, type), node)
                    : node;
            }
            // @api
            function createTemplateLiteralTypeSpan(type, literal) {
                var node = createBaseNode(197 /* TemplateLiteralTypeSpan */);
                node.type = type;
                node.literal = literal;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTemplateLiteralTypeSpan(node, type, literal) {
                return node.type !== type
                    || node.literal !== literal
                    ? update(createTemplateLiteralTypeSpan(type, literal), node)
                    : node;
            }
            //
            // Types
            //
            // @api
            function createKeywordTypeNode(kind) {
                return createToken(kind);
            }
            // @api
            function createTypePredicateNode(assertsModifier, parameterName, type) {
                var node = createBaseNode(175 /* TypePredicate */);
                node.assertsModifier = assertsModifier;
                node.parameterName = asName(parameterName);
                node.type = type;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTypePredicateNode(node, assertsModifier, parameterName, type) {
                return node.assertsModifier !== assertsModifier
                    || node.parameterName !== parameterName
                    || node.type !== type
                    ? update(createTypePredicateNode(assertsModifier, parameterName, type), node)
                    : node;
            }
            // @api
            function createTypeReferenceNode(typeName, typeArguments) {
                var node = createBaseNode(176 /* TypeReference */);
                node.typeName = asName(typeName);
                node.typeArguments = typeArguments && parenthesizerRules().parenthesizeTypeArguments(createNodeArray(typeArguments));
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTypeReferenceNode(node, typeName, typeArguments) {
                return node.typeName !== typeName
                    || node.typeArguments !== typeArguments
                    ? update(createTypeReferenceNode(typeName, typeArguments), node)
                    : node;
            }
            // @api
            function createFunctionTypeNode(typeParameters, parameters, type) {
                var node = createBaseSignatureDeclaration(177 /* FunctionType */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*name*/ undefined, typeParameters, parameters, type);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateFunctionTypeNode(node, typeParameters, parameters, type) {
                return node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    ? updateBaseSignatureDeclaration(createFunctionTypeNode(typeParameters, parameters, type), node)
                    : node;
            }
            // @api
            function createConstructorTypeNode() {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return args.length === 4 ? createConstructorTypeNode1.apply(void 0, args) :
                    args.length === 3 ? createConstructorTypeNode2.apply(void 0, args) :
                        ts.Debug.fail("Incorrect number of arguments specified.");
            }
            function createConstructorTypeNode1(modifiers, typeParameters, parameters, type) {
                var node = createBaseSignatureDeclaration(178 /* ConstructorType */, 
                /*decorators*/ undefined, modifiers, 
                /*name*/ undefined, typeParameters, parameters, type);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            /** @deprecated */
            function createConstructorTypeNode2(typeParameters, parameters, type) {
                return createConstructorTypeNode1(/*modifiers*/ undefined, typeParameters, parameters, type);
            }
            // @api
            function updateConstructorTypeNode() {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return args.length === 5 ? updateConstructorTypeNode1.apply(void 0, args) :
                    args.length === 4 ? updateConstructorTypeNode2.apply(void 0, args) :
                        ts.Debug.fail("Incorrect number of arguments specified.");
            }
            function updateConstructorTypeNode1(node, modifiers, typeParameters, parameters, type) {
                return node.modifiers !== modifiers
                    || node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    ? updateBaseSignatureDeclaration(createConstructorTypeNode(modifiers, typeParameters, parameters, type), node)
                    : node;
            }
            /** @deprecated */
            function updateConstructorTypeNode2(node, typeParameters, parameters, type) {
                return updateConstructorTypeNode1(node, node.modifiers, typeParameters, parameters, type);
            }
            // @api
            function createTypeQueryNode(exprName) {
                var node = createBaseNode(179 /* TypeQuery */);
                node.exprName = exprName;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTypeQueryNode(node, exprName) {
                return node.exprName !== exprName
                    ? update(createTypeQueryNode(exprName), node)
                    : node;
            }
            // @api
            function createTypeLiteralNode(members) {
                var node = createBaseNode(180 /* TypeLiteral */);
                node.members = createNodeArray(members);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTypeLiteralNode(node, members) {
                return node.members !== members
                    ? update(createTypeLiteralNode(members), node)
                    : node;
            }
            // @api
            function createArrayTypeNode(elementType) {
                var node = createBaseNode(181 /* ArrayType */);
                node.elementType = parenthesizerRules().parenthesizeElementTypeOfArrayType(elementType);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateArrayTypeNode(node, elementType) {
                return node.elementType !== elementType
                    ? update(createArrayTypeNode(elementType), node)
                    : node;
            }
            // @api
            function createTupleTypeNode(elements) {
                var node = createBaseNode(182 /* TupleType */);
                node.elements = createNodeArray(elements);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTupleTypeNode(node, elements) {
                return node.elements !== elements
                    ? update(createTupleTypeNode(elements), node)
                    : node;
            }
            // @api
            function createNamedTupleMember(dotDotDotToken, name, questionToken, type) {
                var node = createBaseNode(195 /* NamedTupleMember */);
                node.dotDotDotToken = dotDotDotToken;
                node.name = name;
                node.questionToken = questionToken;
                node.type = type;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateNamedTupleMember(node, dotDotDotToken, name, questionToken, type) {
                return node.dotDotDotToken !== dotDotDotToken
                    || node.name !== name
                    || node.questionToken !== questionToken
                    || node.type !== type
                    ? update(createNamedTupleMember(dotDotDotToken, name, questionToken, type), node)
                    : node;
            }
            // @api
            function createOptionalTypeNode(type) {
                var node = createBaseNode(183 /* OptionalType */);
                node.type = parenthesizerRules().parenthesizeElementTypeOfArrayType(type);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateOptionalTypeNode(node, type) {
                return node.type !== type
                    ? update(createOptionalTypeNode(type), node)
                    : node;
            }
            // @api
            function createRestTypeNode(type) {
                var node = createBaseNode(184 /* RestType */);
                node.type = type;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateRestTypeNode(node, type) {
                return node.type !== type
                    ? update(createRestTypeNode(type), node)
                    : node;
            }
            function createUnionOrIntersectionTypeNode(kind, types) {
                var node = createBaseNode(kind);
                node.types = parenthesizerRules().parenthesizeConstituentTypesOfUnionOrIntersectionType(types);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            function updateUnionOrIntersectionTypeNode(node, types) {
                return node.types !== types
                    ? update(createUnionOrIntersectionTypeNode(node.kind, types), node)
                    : node;
            }
            // @api
            function createUnionTypeNode(types) {
                return createUnionOrIntersectionTypeNode(185 /* UnionType */, types);
            }
            // @api
            function updateUnionTypeNode(node, types) {
                return updateUnionOrIntersectionTypeNode(node, types);
            }
            // @api
            function createIntersectionTypeNode(types) {
                return createUnionOrIntersectionTypeNode(186 /* IntersectionType */, types);
            }
            // @api
            function updateIntersectionTypeNode(node, types) {
                return updateUnionOrIntersectionTypeNode(node, types);
            }
            // @api
            function createConditionalTypeNode(checkType, extendsType, trueType, falseType) {
                var node = createBaseNode(187 /* ConditionalType */);
                node.checkType = parenthesizerRules().parenthesizeMemberOfConditionalType(checkType);
                node.extendsType = parenthesizerRules().parenthesizeMemberOfConditionalType(extendsType);
                node.trueType = trueType;
                node.falseType = falseType;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateConditionalTypeNode(node, checkType, extendsType, trueType, falseType) {
                return node.checkType !== checkType
                    || node.extendsType !== extendsType
                    || node.trueType !== trueType
                    || node.falseType !== falseType
                    ? update(createConditionalTypeNode(checkType, extendsType, trueType, falseType), node)
                    : node;
            }
            // @api
            function createInferTypeNode(typeParameter) {
                var node = createBaseNode(188 /* InferType */);
                node.typeParameter = typeParameter;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateInferTypeNode(node, typeParameter) {
                return node.typeParameter !== typeParameter
                    ? update(createInferTypeNode(typeParameter), node)
                    : node;
            }
            // @api
            function createTemplateLiteralType(head, templateSpans) {
                var node = createBaseNode(196 /* TemplateLiteralType */);
                node.head = head;
                node.templateSpans = createNodeArray(templateSpans);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTemplateLiteralType(node, head, templateSpans) {
                return node.head !== head
                    || node.templateSpans !== templateSpans
                    ? update(createTemplateLiteralType(head, templateSpans), node)
                    : node;
            }
            // @api
            function createImportTypeNode(argument, qualifier, typeArguments, isTypeOf) {
                if (isTypeOf === void 0) {
                    isTypeOf = false;
                }
                var node = createBaseNode(198 /* ImportType */);
                node.argument = argument;
                node.qualifier = qualifier;
                node.typeArguments = typeArguments && parenthesizerRules().parenthesizeTypeArguments(typeArguments);
                node.isTypeOf = isTypeOf;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateImportTypeNode(node, argument, qualifier, typeArguments, isTypeOf) {
                if (isTypeOf === void 0) {
                    isTypeOf = node.isTypeOf;
                }
                return node.argument !== argument
                    || node.qualifier !== qualifier
                    || node.typeArguments !== typeArguments
                    || node.isTypeOf !== isTypeOf
                    ? update(createImportTypeNode(argument, qualifier, typeArguments, isTypeOf), node)
                    : node;
            }
            // @api
            function createParenthesizedType(type) {
                var node = createBaseNode(189 /* ParenthesizedType */);
                node.type = type;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateParenthesizedType(node, type) {
                return node.type !== type
                    ? update(createParenthesizedType(type), node)
                    : node;
            }
            // @api
            function createThisTypeNode() {
                var node = createBaseNode(190 /* ThisType */);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function createTypeOperatorNode(operator, type) {
                var node = createBaseNode(191 /* TypeOperator */);
                node.operator = operator;
                node.type = parenthesizerRules().parenthesizeMemberOfElementType(type);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTypeOperatorNode(node, type) {
                return node.type !== type
                    ? update(createTypeOperatorNode(node.operator, type), node)
                    : node;
            }
            // @api
            function createIndexedAccessTypeNode(objectType, indexType) {
                var node = createBaseNode(192 /* IndexedAccessType */);
                node.objectType = parenthesizerRules().parenthesizeMemberOfElementType(objectType);
                node.indexType = indexType;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateIndexedAccessTypeNode(node, objectType, indexType) {
                return node.objectType !== objectType
                    || node.indexType !== indexType
                    ? update(createIndexedAccessTypeNode(objectType, indexType), node)
                    : node;
            }
            // @api
            function createMappedTypeNode(readonlyToken, typeParameter, nameType, questionToken, type) {
                var node = createBaseNode(193 /* MappedType */);
                node.readonlyToken = readonlyToken;
                node.typeParameter = typeParameter;
                node.nameType = nameType;
                node.questionToken = questionToken;
                node.type = type;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateMappedTypeNode(node, readonlyToken, typeParameter, nameType, questionToken, type) {
                return node.readonlyToken !== readonlyToken
                    || node.typeParameter !== typeParameter
                    || node.nameType !== nameType
                    || node.questionToken !== questionToken
                    || node.type !== type
                    ? update(createMappedTypeNode(readonlyToken, typeParameter, nameType, questionToken, type), node)
                    : node;
            }
            // @api
            function createLiteralTypeNode(literal) {
                var node = createBaseNode(194 /* LiteralType */);
                node.literal = literal;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateLiteralTypeNode(node, literal) {
                return node.literal !== literal
                    ? update(createLiteralTypeNode(literal), node)
                    : node;
            }
            //
            // Binding Patterns
            //
            // @api
            function createObjectBindingPattern(elements) {
                var node = createBaseNode(199 /* ObjectBindingPattern */);
                node.elements = createNodeArray(elements);
                node.transformFlags |=
                    propagateChildrenFlags(node.elements) |
                        512 /* ContainsES2015 */ |
                        262144 /* ContainsBindingPattern */;
                if (node.transformFlags & 16384 /* ContainsRestOrSpread */) {
                    node.transformFlags |=
                        64 /* ContainsES2018 */ |
                            32768 /* ContainsObjectRestOrSpread */;
                }
                return node;
            }
            // @api
            function updateObjectBindingPattern(node, elements) {
                return node.elements !== elements
                    ? update(createObjectBindingPattern(elements), node)
                    : node;
            }
            // @api
            function createArrayBindingPattern(elements) {
                var node = createBaseNode(200 /* ArrayBindingPattern */);
                node.elements = createNodeArray(elements);
                node.transformFlags |=
                    propagateChildrenFlags(node.elements) |
                        512 /* ContainsES2015 */ |
                        262144 /* ContainsBindingPattern */;
                return node;
            }
            // @api
            function updateArrayBindingPattern(node, elements) {
                return node.elements !== elements
                    ? update(createArrayBindingPattern(elements), node)
                    : node;
            }
            // @api
            function createBindingElement(dotDotDotToken, propertyName, name, initializer) {
                var node = createBaseBindingLikeDeclaration(201 /* BindingElement */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, name, initializer && parenthesizerRules().parenthesizeExpressionForDisallowedComma(initializer));
                node.propertyName = asName(propertyName);
                node.dotDotDotToken = dotDotDotToken;
                node.transformFlags |=
                    propagateChildFlags(node.dotDotDotToken) |
                        512 /* ContainsES2015 */;
                if (node.propertyName) {
                    node.transformFlags |= ts.isIdentifier(node.propertyName) ?
                        propagateIdentifierNameFlags(node.propertyName) :
                        propagateChildFlags(node.propertyName);
                }
                if (dotDotDotToken)
                    node.transformFlags |= 16384 /* ContainsRestOrSpread */;
                return node;
            }
            // @api
            function updateBindingElement(node, dotDotDotToken, propertyName, name, initializer) {
                return node.propertyName !== propertyName
                    || node.dotDotDotToken !== dotDotDotToken
                    || node.name !== name
                    || node.initializer !== initializer
                    ? update(createBindingElement(dotDotDotToken, propertyName, name, initializer), node)
                    : node;
            }
            //
            // Expression
            //
            function createBaseExpression(kind) {
                var node = createBaseNode(kind);
                // the following properties are commonly set by the checker/binder
                return node;
            }
            // @api
            function createArrayLiteralExpression(elements, multiLine) {
                var node = createBaseExpression(202 /* ArrayLiteralExpression */);
                // Ensure we add a trailing comma for something like `[NumericLiteral(1), NumericLiteral(2), OmittedExpresion]` so that
                // we end up with `[1, 2, ,]` instead of `[1, 2, ]` otherwise the `OmittedExpression` will just end up being treated like
                // a trailing comma.
                var lastElement = elements && ts.lastOrUndefined(elements);
                var elementsArray = createNodeArray(elements, lastElement && ts.isOmittedExpression(lastElement) ? true : undefined);
                node.elements = parenthesizerRules().parenthesizeExpressionsOfCommaDelimitedList(elementsArray);
                node.multiLine = multiLine;
                node.transformFlags |= propagateChildrenFlags(node.elements);
                return node;
            }
            // @api
            function updateArrayLiteralExpression(node, elements) {
                return node.elements !== elements
                    ? update(createArrayLiteralExpression(elements, node.multiLine), node)
                    : node;
            }
            // @api
            function createObjectLiteralExpression(properties, multiLine) {
                var node = createBaseExpression(203 /* ObjectLiteralExpression */);
                node.properties = createNodeArray(properties);
                node.multiLine = multiLine;
                node.transformFlags |= propagateChildrenFlags(node.properties);
                return node;
            }
            // @api
            function updateObjectLiteralExpression(node, properties) {
                return node.properties !== properties
                    ? update(createObjectLiteralExpression(properties, node.multiLine), node)
                    : node;
            }
            // @api
            function createPropertyAccessExpression(expression, name) {
                var node = createBaseExpression(204 /* PropertyAccessExpression */);
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.name = asName(name);
                node.transformFlags =
                    propagateChildFlags(node.expression) |
                        (ts.isIdentifier(node.name) ?
                            propagateIdentifierNameFlags(node.name) :
                            propagateChildFlags(node.name));
                if (ts.isSuperKeyword(expression)) {
                    // super method calls require a lexical 'this'
                    // super method calls require 'super' hoisting in ES2017 and ES2018 async functions and async generators
                    node.transformFlags |=
                        128 /* ContainsES2017 */ |
                            64 /* ContainsES2018 */;
                }
                return node;
            }
            // @api
            function updatePropertyAccessExpression(node, expression, name) {
                if (ts.isPropertyAccessChain(node)) {
                    return updatePropertyAccessChain(node, expression, node.questionDotToken, ts.cast(name, ts.isIdentifier));
                }
                return node.expression !== expression
                    || node.name !== name
                    ? update(createPropertyAccessExpression(expression, name), node)
                    : node;
            }
            // @api
            function createPropertyAccessChain(expression, questionDotToken, name) {
                var node = createBaseExpression(204 /* PropertyAccessExpression */);
                node.flags |= 32 /* OptionalChain */;
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.questionDotToken = questionDotToken;
                node.name = asName(name);
                node.transformFlags |=
                    16 /* ContainsES2020 */ |
                        propagateChildFlags(node.expression) |
                        propagateChildFlags(node.questionDotToken) |
                        (ts.isIdentifier(node.name) ?
                            propagateIdentifierNameFlags(node.name) :
                            propagateChildFlags(node.name));
                return node;
            }
            // @api
            function updatePropertyAccessChain(node, expression, questionDotToken, name) {
                ts.Debug.assert(!!(node.flags & 32 /* OptionalChain */), "Cannot update a PropertyAccessExpression using updatePropertyAccessChain. Use updatePropertyAccess instead.");
                // Because we are updating an existing PropertyAccessChain we want to inherit its emitFlags
                // instead of using the default from createPropertyAccess
                return node.expression !== expression
                    || node.questionDotToken !== questionDotToken
                    || node.name !== name
                    ? update(createPropertyAccessChain(expression, questionDotToken, name), node)
                    : node;
            }
            // @api
            function createElementAccessExpression(expression, index) {
                var node = createBaseExpression(205 /* ElementAccessExpression */);
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.argumentExpression = asExpression(index);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.argumentExpression);
                if (ts.isSuperKeyword(expression)) {
                    // super method calls require a lexical 'this'
                    // super method calls require 'super' hoisting in ES2017 and ES2018 async functions and async generators
                    node.transformFlags |=
                        128 /* ContainsES2017 */ |
                            64 /* ContainsES2018 */;
                }
                return node;
            }
            // @api
            function updateElementAccessExpression(node, expression, argumentExpression) {
                if (ts.isElementAccessChain(node)) {
                    return updateElementAccessChain(node, expression, node.questionDotToken, argumentExpression);
                }
                return node.expression !== expression
                    || node.argumentExpression !== argumentExpression
                    ? update(createElementAccessExpression(expression, argumentExpression), node)
                    : node;
            }
            // @api
            function createElementAccessChain(expression, questionDotToken, index) {
                var node = createBaseExpression(205 /* ElementAccessExpression */);
                node.flags |= 32 /* OptionalChain */;
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.questionDotToken = questionDotToken;
                node.argumentExpression = asExpression(index);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.questionDotToken) |
                        propagateChildFlags(node.argumentExpression) |
                        16 /* ContainsES2020 */;
                return node;
            }
            // @api
            function updateElementAccessChain(node, expression, questionDotToken, argumentExpression) {
                ts.Debug.assert(!!(node.flags & 32 /* OptionalChain */), "Cannot update a ElementAccessExpression using updateElementAccessChain. Use updateElementAccess instead.");
                // Because we are updating an existing ElementAccessChain we want to inherit its emitFlags
                // instead of using the default from createElementAccess
                return node.expression !== expression
                    || node.questionDotToken !== questionDotToken
                    || node.argumentExpression !== argumentExpression
                    ? update(createElementAccessChain(expression, questionDotToken, argumentExpression), node)
                    : node;
            }
            // @api
            function createCallExpression(expression, typeArguments, argumentsArray) {
                var node = createBaseExpression(206 /* CallExpression */);
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.typeArguments = asNodeArray(typeArguments);
                node.arguments = parenthesizerRules().parenthesizeExpressionsOfCommaDelimitedList(createNodeArray(argumentsArray));
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildrenFlags(node.typeArguments) |
                        propagateChildrenFlags(node.arguments);
                if (node.typeArguments) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                if (ts.isImportKeyword(node.expression)) {
                    node.transformFlags |= 4194304 /* ContainsDynamicImport */;
                }
                else if (ts.isSuperProperty(node.expression)) {
                    node.transformFlags |= 8192 /* ContainsLexicalThis */;
                }
                return node;
            }
            // @api
            function updateCallExpression(node, expression, typeArguments, argumentsArray) {
                if (ts.isCallChain(node)) {
                    return updateCallChain(node, expression, node.questionDotToken, typeArguments, argumentsArray);
                }
                return node.expression !== expression
                    || node.typeArguments !== typeArguments
                    || node.arguments !== argumentsArray
                    ? update(createCallExpression(expression, typeArguments, argumentsArray), node)
                    : node;
            }
            // @api
            function createCallChain(expression, questionDotToken, typeArguments, argumentsArray) {
                var node = createBaseExpression(206 /* CallExpression */);
                node.flags |= 32 /* OptionalChain */;
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.questionDotToken = questionDotToken;
                node.typeArguments = asNodeArray(typeArguments);
                node.arguments = parenthesizerRules().parenthesizeExpressionsOfCommaDelimitedList(createNodeArray(argumentsArray));
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.questionDotToken) |
                        propagateChildrenFlags(node.typeArguments) |
                        propagateChildrenFlags(node.arguments) |
                        16 /* ContainsES2020 */;
                if (node.typeArguments) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                if (ts.isSuperProperty(node.expression)) {
                    node.transformFlags |= 8192 /* ContainsLexicalThis */;
                }
                return node;
            }
            // @api
            function updateCallChain(node, expression, questionDotToken, typeArguments, argumentsArray) {
                ts.Debug.assert(!!(node.flags & 32 /* OptionalChain */), "Cannot update a CallExpression using updateCallChain. Use updateCall instead.");
                return node.expression !== expression
                    || node.questionDotToken !== questionDotToken
                    || node.typeArguments !== typeArguments
                    || node.arguments !== argumentsArray
                    ? update(createCallChain(expression, questionDotToken, typeArguments, argumentsArray), node)
                    : node;
            }
            // @api
            function createNewExpression(expression, typeArguments, argumentsArray) {
                var node = createBaseExpression(207 /* NewExpression */);
                node.expression = parenthesizerRules().parenthesizeExpressionOfNew(expression);
                node.typeArguments = asNodeArray(typeArguments);
                node.arguments = argumentsArray ? parenthesizerRules().parenthesizeExpressionsOfCommaDelimitedList(argumentsArray) : undefined;
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildrenFlags(node.typeArguments) |
                        propagateChildrenFlags(node.arguments) |
                        16 /* ContainsES2020 */;
                if (node.typeArguments) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                return node;
            }
            // @api
            function updateNewExpression(node, expression, typeArguments, argumentsArray) {
                return node.expression !== expression
                    || node.typeArguments !== typeArguments
                    || node.arguments !== argumentsArray
                    ? update(createNewExpression(expression, typeArguments, argumentsArray), node)
                    : node;
            }
            // @api
            function createTaggedTemplateExpression(tag, typeArguments, template) {
                var node = createBaseExpression(208 /* TaggedTemplateExpression */);
                node.tag = parenthesizerRules().parenthesizeLeftSideOfAccess(tag);
                node.typeArguments = asNodeArray(typeArguments);
                node.template = template;
                node.transformFlags |=
                    propagateChildFlags(node.tag) |
                        propagateChildrenFlags(node.typeArguments) |
                        propagateChildFlags(node.template) |
                        512 /* ContainsES2015 */;
                if (node.typeArguments) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                if (ts.hasInvalidEscape(node.template)) {
                    node.transformFlags |= 64 /* ContainsES2018 */;
                }
                return node;
            }
            // @api
            function updateTaggedTemplateExpression(node, tag, typeArguments, template) {
                return node.tag !== tag
                    || node.typeArguments !== typeArguments
                    || node.template !== template
                    ? update(createTaggedTemplateExpression(tag, typeArguments, template), node)
                    : node;
            }
            // @api
            function createTypeAssertion(type, expression) {
                var node = createBaseExpression(209 /* TypeAssertionExpression */);
                node.expression = parenthesizerRules().parenthesizeOperandOfPrefixUnary(expression);
                node.type = type;
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.type) |
                        1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTypeAssertion(node, type, expression) {
                return node.type !== type
                    || node.expression !== expression
                    ? update(createTypeAssertion(type, expression), node)
                    : node;
            }
            // @api
            function createParenthesizedExpression(expression) {
                var node = createBaseExpression(210 /* ParenthesizedExpression */);
                node.expression = expression;
                node.transformFlags = propagateChildFlags(node.expression);
                return node;
            }
            // @api
            function updateParenthesizedExpression(node, expression) {
                return node.expression !== expression
                    ? update(createParenthesizedExpression(expression), node)
                    : node;
            }
            // @api
            function createFunctionExpression(modifiers, asteriskToken, name, typeParameters, parameters, type, body) {
                var node = createBaseFunctionLikeDeclaration(211 /* FunctionExpression */, 
                /*decorators*/ undefined, modifiers, name, typeParameters, parameters, type, body);
                node.asteriskToken = asteriskToken;
                node.transformFlags |= propagateChildFlags(node.asteriskToken);
                if (node.typeParameters) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                if (ts.modifiersToFlags(node.modifiers) & 256 /* Async */) {
                    if (node.asteriskToken) {
                        node.transformFlags |= 64 /* ContainsES2018 */;
                    }
                    else {
                        node.transformFlags |= 128 /* ContainsES2017 */;
                    }
                }
                else if (node.asteriskToken) {
                    node.transformFlags |= 1024 /* ContainsGenerator */;
                }
                return node;
            }
            // @api
            function updateFunctionExpression(node, modifiers, asteriskToken, name, typeParameters, parameters, type, body) {
                return node.name !== name
                    || node.modifiers !== modifiers
                    || node.asteriskToken !== asteriskToken
                    || node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    || node.body !== body
                    ? updateBaseFunctionLikeDeclaration(createFunctionExpression(modifiers, asteriskToken, name, typeParameters, parameters, type, body), node)
                    : node;
            }
            // @api
            function createArrowFunction(modifiers, typeParameters, parameters, type, equalsGreaterThanToken, body) {
                var node = createBaseFunctionLikeDeclaration(212 /* ArrowFunction */, 
                /*decorators*/ undefined, modifiers, 
                /*name*/ undefined, typeParameters, parameters, type, parenthesizerRules().parenthesizeConciseBodyOfArrowFunction(body));
                node.equalsGreaterThanToken = equalsGreaterThanToken !== null && equalsGreaterThanToken !== void 0 ? equalsGreaterThanToken : createToken(38 /* EqualsGreaterThanToken */);
                node.transformFlags |=
                    propagateChildFlags(node.equalsGreaterThanToken) |
                        512 /* ContainsES2015 */;
                if (ts.modifiersToFlags(node.modifiers) & 256 /* Async */) {
                    node.transformFlags |= 128 /* ContainsES2017 */ | 8192 /* ContainsLexicalThis */;
                }
                return node;
            }
            // @api
            function updateArrowFunction(node, modifiers, typeParameters, parameters, type, equalsGreaterThanToken, body) {
                return node.modifiers !== modifiers
                    || node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    || node.equalsGreaterThanToken !== equalsGreaterThanToken
                    || node.body !== body
                    ? updateBaseFunctionLikeDeclaration(createArrowFunction(modifiers, typeParameters, parameters, type, equalsGreaterThanToken, body), node)
                    : node;
            }
            // @api
            function createDeleteExpression(expression) {
                var node = createBaseExpression(213 /* DeleteExpression */);
                node.expression = parenthesizerRules().parenthesizeOperandOfPrefixUnary(expression);
                node.transformFlags |= propagateChildFlags(node.expression);
                return node;
            }
            // @api
            function updateDeleteExpression(node, expression) {
                return node.expression !== expression
                    ? update(createDeleteExpression(expression), node)
                    : node;
            }
            // @api
            function createTypeOfExpression(expression) {
                var node = createBaseExpression(214 /* TypeOfExpression */);
                node.expression = parenthesizerRules().parenthesizeOperandOfPrefixUnary(expression);
                node.transformFlags |= propagateChildFlags(node.expression);
                return node;
            }
            // @api
            function updateTypeOfExpression(node, expression) {
                return node.expression !== expression
                    ? update(createTypeOfExpression(expression), node)
                    : node;
            }
            // @api
            function createVoidExpression(expression) {
                var node = createBaseExpression(215 /* VoidExpression */);
                node.expression = parenthesizerRules().parenthesizeOperandOfPrefixUnary(expression);
                node.transformFlags |= propagateChildFlags(node.expression);
                return node;
            }
            // @api
            function updateVoidExpression(node, expression) {
                return node.expression !== expression
                    ? update(createVoidExpression(expression), node)
                    : node;
            }
            // @api
            function createAwaitExpression(expression) {
                var node = createBaseExpression(216 /* AwaitExpression */);
                node.expression = parenthesizerRules().parenthesizeOperandOfPrefixUnary(expression);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        128 /* ContainsES2017 */ |
                        64 /* ContainsES2018 */ |
                        1048576 /* ContainsAwait */;
                return node;
            }
            // @api
            function updateAwaitExpression(node, expression) {
                return node.expression !== expression
                    ? update(createAwaitExpression(expression), node)
                    : node;
            }
            // @api
            function createPrefixUnaryExpression(operator, operand) {
                var node = createBaseExpression(217 /* PrefixUnaryExpression */);
                node.operator = operator;
                node.operand = parenthesizerRules().parenthesizeOperandOfPrefixUnary(operand);
                node.transformFlags |= propagateChildFlags(node.operand);
                // Only set this flag for non-generated identifiers and non-"local" names. See the
                // comment in `visitPreOrPostfixUnaryExpression` in module.ts
                if ((operator === 45 /* PlusPlusToken */ || operator === 46 /* MinusMinusToken */) &&
                    ts.isIdentifier(node.operand) &&
                    !ts.isGeneratedIdentifier(node.operand) &&
                    !ts.isLocalName(node.operand)) {
                    node.transformFlags |= 67108864 /* ContainsUpdateExpressionForIdentifier */;
                }
                return node;
            }
            // @api
            function updatePrefixUnaryExpression(node, operand) {
                return node.operand !== operand
                    ? update(createPrefixUnaryExpression(node.operator, operand), node)
                    : node;
            }
            // @api
            function createPostfixUnaryExpression(operand, operator) {
                var node = createBaseExpression(218 /* PostfixUnaryExpression */);
                node.operator = operator;
                node.operand = parenthesizerRules().parenthesizeOperandOfPostfixUnary(operand);
                node.transformFlags |= propagateChildFlags(node.operand);
                // Only set this flag for non-generated identifiers and non-"local" names. See the
                // comment in `visitPreOrPostfixUnaryExpression` in module.ts
                if (ts.isIdentifier(node.operand) &&
                    !ts.isGeneratedIdentifier(node.operand) &&
                    !ts.isLocalName(node.operand)) {
                    node.transformFlags |= 67108864 /* ContainsUpdateExpressionForIdentifier */;
                }
                return node;
            }
            // @api
            function updatePostfixUnaryExpression(node, operand) {
                return node.operand !== operand
                    ? update(createPostfixUnaryExpression(operand, node.operator), node)
                    : node;
            }
            // @api
            function createBinaryExpression(left, operator, right) {
                var node = createBaseExpression(219 /* BinaryExpression */);
                var operatorToken = asToken(operator);
                var operatorKind = operatorToken.kind;
                node.left = parenthesizerRules().parenthesizeLeftSideOfBinary(operatorKind, left);
                node.operatorToken = operatorToken;
                node.right = parenthesizerRules().parenthesizeRightSideOfBinary(operatorKind, node.left, right);
                node.transformFlags |=
                    propagateChildFlags(node.left) |
                        propagateChildFlags(node.operatorToken) |
                        propagateChildFlags(node.right);
                if (operatorKind === 60 /* QuestionQuestionToken */) {
                    node.transformFlags |= 16 /* ContainsES2020 */;
                }
                else if (operatorKind === 63 /* EqualsToken */) {
                    if (ts.isObjectLiteralExpression(node.left)) {
                        node.transformFlags |=
                            512 /* ContainsES2015 */ |
                                64 /* ContainsES2018 */ |
                                2048 /* ContainsDestructuringAssignment */ |
                                propagateAssignmentPatternFlags(node.left);
                    }
                    else if (ts.isArrayLiteralExpression(node.left)) {
                        node.transformFlags |=
                            512 /* ContainsES2015 */ |
                                2048 /* ContainsDestructuringAssignment */ |
                                propagateAssignmentPatternFlags(node.left);
                    }
                }
                else if (operatorKind === 42 /* AsteriskAsteriskToken */ || operatorKind === 67 /* AsteriskAsteriskEqualsToken */) {
                    node.transformFlags |= 256 /* ContainsES2016 */;
                }
                else if (ts.isLogicalOrCoalescingAssignmentOperator(operatorKind)) {
                    node.transformFlags |= 8 /* ContainsES2021 */;
                }
                return node;
            }
            function propagateAssignmentPatternFlags(node) {
                if (node.transformFlags & 32768 /* ContainsObjectRestOrSpread */)
                    return 32768 /* ContainsObjectRestOrSpread */;
                if (node.transformFlags & 64 /* ContainsES2018 */) {
                    // check for nested spread assignments, otherwise '{ x: { a, ...b } = foo } = c'
                    // will not be correctly interpreted by the ES2018 transformer
                    for (var _i = 0, _a = ts.getElementsOfBindingOrAssignmentPattern(node); _i < _a.length; _i++) {
                        var element = _a[_i];
                        var target = ts.getTargetOfBindingOrAssignmentElement(element);
                        if (target && ts.isAssignmentPattern(target)) {
                            if (target.transformFlags & 32768 /* ContainsObjectRestOrSpread */) {
                                return 32768 /* ContainsObjectRestOrSpread */;
                            }
                            if (target.transformFlags & 64 /* ContainsES2018 */) {
                                var flags_1 = propagateAssignmentPatternFlags(target);
                                if (flags_1)
                                    return flags_1;
                            }
                        }
                    }
                }
                return 0 /* None */;
            }
            // @api
            function updateBinaryExpression(node, left, operator, right) {
                return node.left !== left
                    || node.operatorToken !== operator
                    || node.right !== right
                    ? update(createBinaryExpression(left, operator, right), node)
                    : node;
            }
            // @api
            function createConditionalExpression(condition, questionToken, whenTrue, colonToken, whenFalse) {
                var node = createBaseExpression(220 /* ConditionalExpression */);
                node.condition = parenthesizerRules().parenthesizeConditionOfConditionalExpression(condition);
                node.questionToken = questionToken !== null && questionToken !== void 0 ? questionToken : createToken(57 /* QuestionToken */);
                node.whenTrue = parenthesizerRules().parenthesizeBranchOfConditionalExpression(whenTrue);
                node.colonToken = colonToken !== null && colonToken !== void 0 ? colonToken : createToken(58 /* ColonToken */);
                node.whenFalse = parenthesizerRules().parenthesizeBranchOfConditionalExpression(whenFalse);
                node.transformFlags |=
                    propagateChildFlags(node.condition) |
                        propagateChildFlags(node.questionToken) |
                        propagateChildFlags(node.whenTrue) |
                        propagateChildFlags(node.colonToken) |
                        propagateChildFlags(node.whenFalse);
                return node;
            }
            // @api
            function updateConditionalExpression(node, condition, questionToken, whenTrue, colonToken, whenFalse) {
                return node.condition !== condition
                    || node.questionToken !== questionToken
                    || node.whenTrue !== whenTrue
                    || node.colonToken !== colonToken
                    || node.whenFalse !== whenFalse
                    ? update(createConditionalExpression(condition, questionToken, whenTrue, colonToken, whenFalse), node)
                    : node;
            }
            // @api
            function createTemplateExpression(head, templateSpans) {
                var node = createBaseExpression(221 /* TemplateExpression */);
                node.head = head;
                node.templateSpans = createNodeArray(templateSpans);
                node.transformFlags |=
                    propagateChildFlags(node.head) |
                        propagateChildrenFlags(node.templateSpans) |
                        512 /* ContainsES2015 */;
                return node;
            }
            // @api
            function updateTemplateExpression(node, head, templateSpans) {
                return node.head !== head
                    || node.templateSpans !== templateSpans
                    ? update(createTemplateExpression(head, templateSpans), node)
                    : node;
            }
            function createTemplateLiteralLikeNodeChecked(kind, text, rawText, templateFlags) {
                if (templateFlags === void 0) {
                    templateFlags = 0 /* None */;
                }
                ts.Debug.assert(!(templateFlags & ~2048 /* TemplateLiteralLikeFlags */), "Unsupported template flags.");
                // NOTE: without the assignment to `undefined`, we don't narrow the initial type of `cooked`.
                // eslint-disable-next-line no-undef-init
                var cooked = undefined;
                if (rawText !== undefined && rawText !== text) {
                    cooked = getCookedText(kind, rawText);
                    if (typeof cooked === "object") {
                        return ts.Debug.fail("Invalid raw text");
                    }
                }
                if (text === undefined) {
                    if (cooked === undefined) {
                        return ts.Debug.fail("Arguments 'text' and 'rawText' may not both be undefined.");
                    }
                    text = cooked;
                }
                else if (cooked !== undefined) {
                    ts.Debug.assert(text === cooked, "Expected argument 'text' to be the normalized (i.e. 'cooked') version of argument 'rawText'.");
                }
                return createTemplateLiteralLikeNode(kind, text, rawText, templateFlags);
            }
            // @api
            function createTemplateLiteralLikeNode(kind, text, rawText, templateFlags) {
                var node = createBaseToken(kind);
                node.text = text;
                node.rawText = rawText;
                node.templateFlags = templateFlags & 2048 /* TemplateLiteralLikeFlags */;
                node.transformFlags |= 512 /* ContainsES2015 */;
                if (node.templateFlags) {
                    node.transformFlags |= 64 /* ContainsES2018 */;
                }
                return node;
            }
            // @api
            function createTemplateHead(text, rawText, templateFlags) {
                return createTemplateLiteralLikeNodeChecked(15 /* TemplateHead */, text, rawText, templateFlags);
            }
            // @api
            function createTemplateMiddle(text, rawText, templateFlags) {
                return createTemplateLiteralLikeNodeChecked(16 /* TemplateMiddle */, text, rawText, templateFlags);
            }
            // @api
            function createTemplateTail(text, rawText, templateFlags) {
                return createTemplateLiteralLikeNodeChecked(17 /* TemplateTail */, text, rawText, templateFlags);
            }
            // @api
            function createNoSubstitutionTemplateLiteral(text, rawText, templateFlags) {
                return createTemplateLiteralLikeNodeChecked(14 /* NoSubstitutionTemplateLiteral */, text, rawText, templateFlags);
            }
            // @api
            function createYieldExpression(asteriskToken, expression) {
                ts.Debug.assert(!asteriskToken || !!expression, "A `YieldExpression` with an asteriskToken must have an expression.");
                var node = createBaseExpression(222 /* YieldExpression */);
                node.expression = expression && parenthesizerRules().parenthesizeExpressionForDisallowedComma(expression);
                node.asteriskToken = asteriskToken;
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.asteriskToken) |
                        512 /* ContainsES2015 */ |
                        64 /* ContainsES2018 */ |
                        524288 /* ContainsYield */;
                return node;
            }
            // @api
            function updateYieldExpression(node, asteriskToken, expression) {
                return node.expression !== expression
                    || node.asteriskToken !== asteriskToken
                    ? update(createYieldExpression(asteriskToken, expression), node)
                    : node;
            }
            // @api
            function createSpreadElement(expression) {
                var node = createBaseExpression(223 /* SpreadElement */);
                node.expression = parenthesizerRules().parenthesizeExpressionForDisallowedComma(expression);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        512 /* ContainsES2015 */ |
                        16384 /* ContainsRestOrSpread */;
                return node;
            }
            // @api
            function updateSpreadElement(node, expression) {
                return node.expression !== expression
                    ? update(createSpreadElement(expression), node)
                    : node;
            }
            // @api
            function createClassExpression(decorators, modifiers, name, typeParameters, heritageClauses, members) {
                var node = createBaseClassLikeDeclaration(224 /* ClassExpression */, decorators, modifiers, name, typeParameters, heritageClauses, members);
                node.transformFlags |= 512 /* ContainsES2015 */;
                return node;
            }
            // @api
            function updateClassExpression(node, decorators, modifiers, name, typeParameters, heritageClauses, members) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.name !== name
                    || node.typeParameters !== typeParameters
                    || node.heritageClauses !== heritageClauses
                    || node.members !== members
                    ? update(createClassExpression(decorators, modifiers, name, typeParameters, heritageClauses, members), node)
                    : node;
            }
            // @api
            function createOmittedExpression() {
                return createBaseExpression(225 /* OmittedExpression */);
            }
            // @api
            function createExpressionWithTypeArguments(expression, typeArguments) {
                var node = createBaseNode(226 /* ExpressionWithTypeArguments */);
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.typeArguments = typeArguments && parenthesizerRules().parenthesizeTypeArguments(typeArguments);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildrenFlags(node.typeArguments) |
                        512 /* ContainsES2015 */;
                return node;
            }
            // @api
            function updateExpressionWithTypeArguments(node, expression, typeArguments) {
                return node.expression !== expression
                    || node.typeArguments !== typeArguments
                    ? update(createExpressionWithTypeArguments(expression, typeArguments), node)
                    : node;
            }
            // @api
            function createAsExpression(expression, type) {
                var node = createBaseExpression(227 /* AsExpression */);
                node.expression = expression;
                node.type = type;
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.type) |
                        1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateAsExpression(node, expression, type) {
                return node.expression !== expression
                    || node.type !== type
                    ? update(createAsExpression(expression, type), node)
                    : node;
            }
            // @api
            function createNonNullExpression(expression) {
                var node = createBaseExpression(228 /* NonNullExpression */);
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateNonNullExpression(node, expression) {
                if (ts.isNonNullChain(node)) {
                    return updateNonNullChain(node, expression);
                }
                return node.expression !== expression
                    ? update(createNonNullExpression(expression), node)
                    : node;
            }
            // @api
            function createNonNullChain(expression) {
                var node = createBaseExpression(228 /* NonNullExpression */);
                node.flags |= 32 /* OptionalChain */;
                node.expression = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateNonNullChain(node, expression) {
                ts.Debug.assert(!!(node.flags & 32 /* OptionalChain */), "Cannot update a NonNullExpression using updateNonNullChain. Use updateNonNullExpression instead.");
                return node.expression !== expression
                    ? update(createNonNullChain(expression), node)
                    : node;
            }
            // @api
            function createMetaProperty(keywordToken, name) {
                var node = createBaseExpression(229 /* MetaProperty */);
                node.keywordToken = keywordToken;
                node.name = name;
                node.transformFlags |= propagateChildFlags(node.name);
                switch (keywordToken) {
                    case 103 /* NewKeyword */:
                        node.transformFlags |= 512 /* ContainsES2015 */;
                        break;
                    case 100 /* ImportKeyword */:
                        node.transformFlags |= 4 /* ContainsESNext */;
                        break;
                    default:
                        return ts.Debug.assertNever(keywordToken);
                }
                return node;
            }
            // @api
            function updateMetaProperty(node, name) {
                return node.name !== name
                    ? update(createMetaProperty(node.keywordToken, name), node)
                    : node;
            }
            //
            // Misc
            //
            // @api
            function createTemplateSpan(expression, literal) {
                var node = createBaseNode(231 /* TemplateSpan */);
                node.expression = expression;
                node.literal = literal;
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.literal) |
                        512 /* ContainsES2015 */;
                return node;
            }
            // @api
            function updateTemplateSpan(node, expression, literal) {
                return node.expression !== expression
                    || node.literal !== literal
                    ? update(createTemplateSpan(expression, literal), node)
                    : node;
            }
            // @api
            function createSemicolonClassElement() {
                var node = createBaseNode(232 /* SemicolonClassElement */);
                node.transformFlags |= 512 /* ContainsES2015 */;
                return node;
            }
            //
            // Element
            //
            // @api
            function createBlock(statements, multiLine) {
                var node = createBaseNode(233 /* Block */);
                node.statements = createNodeArray(statements);
                node.multiLine = multiLine;
                node.transformFlags |= propagateChildrenFlags(node.statements);
                return node;
            }
            // @api
            function updateBlock(node, statements) {
                return node.statements !== statements
                    ? update(createBlock(statements, node.multiLine), node)
                    : node;
            }
            // @api
            function createVariableStatement(modifiers, declarationList) {
                var node = createBaseDeclaration(235 /* VariableStatement */, /*decorators*/ undefined, modifiers);
                node.declarationList = ts.isArray(declarationList) ? createVariableDeclarationList(declarationList) : declarationList;
                node.transformFlags |=
                    propagateChildFlags(node.declarationList);
                if (ts.modifiersToFlags(node.modifiers) & 2 /* Ambient */) {
                    node.transformFlags = 1 /* ContainsTypeScript */;
                }
                return node;
            }
            // @api
            function updateVariableStatement(node, modifiers, declarationList) {
                return node.modifiers !== modifiers
                    || node.declarationList !== declarationList
                    ? update(createVariableStatement(modifiers, declarationList), node)
                    : node;
            }
            // @api
            function createEmptyStatement() {
                return createBaseNode(234 /* EmptyStatement */);
            }
            // @api
            function createExpressionStatement(expression) {
                var node = createBaseNode(236 /* ExpressionStatement */);
                node.expression = parenthesizerRules().parenthesizeExpressionOfExpressionStatement(expression);
                node.transformFlags |= propagateChildFlags(node.expression);
                return node;
            }
            // @api
            function updateExpressionStatement(node, expression) {
                return node.expression !== expression
                    ? update(createExpressionStatement(expression), node)
                    : node;
            }
            // @api
            function createIfStatement(expression, thenStatement, elseStatement) {
                var node = createBaseNode(237 /* IfStatement */);
                node.expression = expression;
                node.thenStatement = asEmbeddedStatement(thenStatement);
                node.elseStatement = asEmbeddedStatement(elseStatement);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.thenStatement) |
                        propagateChildFlags(node.elseStatement);
                return node;
            }
            // @api
            function updateIfStatement(node, expression, thenStatement, elseStatement) {
                return node.expression !== expression
                    || node.thenStatement !== thenStatement
                    || node.elseStatement !== elseStatement
                    ? update(createIfStatement(expression, thenStatement, elseStatement), node)
                    : node;
            }
            // @api
            function createDoStatement(statement, expression) {
                var node = createBaseNode(238 /* DoStatement */);
                node.statement = asEmbeddedStatement(statement);
                node.expression = expression;
                node.transformFlags |=
                    propagateChildFlags(node.statement) |
                        propagateChildFlags(node.expression);
                return node;
            }
            // @api
            function updateDoStatement(node, statement, expression) {
                return node.statement !== statement
                    || node.expression !== expression
                    ? update(createDoStatement(statement, expression), node)
                    : node;
            }
            // @api
            function createWhileStatement(expression, statement) {
                var node = createBaseNode(239 /* WhileStatement */);
                node.expression = expression;
                node.statement = asEmbeddedStatement(statement);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.statement);
                return node;
            }
            // @api
            function updateWhileStatement(node, expression, statement) {
                return node.expression !== expression
                    || node.statement !== statement
                    ? update(createWhileStatement(expression, statement), node)
                    : node;
            }
            // @api
            function createForStatement(initializer, condition, incrementor, statement) {
                var node = createBaseNode(240 /* ForStatement */);
                node.initializer = initializer;
                node.condition = condition;
                node.incrementor = incrementor;
                node.statement = asEmbeddedStatement(statement);
                node.transformFlags |=
                    propagateChildFlags(node.initializer) |
                        propagateChildFlags(node.condition) |
                        propagateChildFlags(node.incrementor) |
                        propagateChildFlags(node.statement);
                return node;
            }
            // @api
            function updateForStatement(node, initializer, condition, incrementor, statement) {
                return node.initializer !== initializer
                    || node.condition !== condition
                    || node.incrementor !== incrementor
                    || node.statement !== statement
                    ? update(createForStatement(initializer, condition, incrementor, statement), node)
                    : node;
            }
            // @api
            function createForInStatement(initializer, expression, statement) {
                var node = createBaseNode(241 /* ForInStatement */);
                node.initializer = initializer;
                node.expression = expression;
                node.statement = asEmbeddedStatement(statement);
                node.transformFlags |=
                    propagateChildFlags(node.initializer) |
                        propagateChildFlags(node.expression) |
                        propagateChildFlags(node.statement);
                return node;
            }
            // @api
            function updateForInStatement(node, initializer, expression, statement) {
                return node.initializer !== initializer
                    || node.expression !== expression
                    || node.statement !== statement
                    ? update(createForInStatement(initializer, expression, statement), node)
                    : node;
            }
            // @api
            function createForOfStatement(awaitModifier, initializer, expression, statement) {
                var node = createBaseNode(242 /* ForOfStatement */);
                node.awaitModifier = awaitModifier;
                node.initializer = initializer;
                node.expression = parenthesizerRules().parenthesizeExpressionForDisallowedComma(expression);
                node.statement = asEmbeddedStatement(statement);
                node.transformFlags |=
                    propagateChildFlags(node.awaitModifier) |
                        propagateChildFlags(node.initializer) |
                        propagateChildFlags(node.expression) |
                        propagateChildFlags(node.statement) |
                        512 /* ContainsES2015 */;
                if (awaitModifier)
                    node.transformFlags |= 64 /* ContainsES2018 */;
                return node;
            }
            // @api
            function updateForOfStatement(node, awaitModifier, initializer, expression, statement) {
                return node.awaitModifier !== awaitModifier
                    || node.initializer !== initializer
                    || node.expression !== expression
                    || node.statement !== statement
                    ? update(createForOfStatement(awaitModifier, initializer, expression, statement), node)
                    : node;
            }
            // @api
            function createContinueStatement(label) {
                var node = createBaseNode(243 /* ContinueStatement */);
                node.label = asName(label);
                node.transformFlags |=
                    propagateChildFlags(node.label) |
                        2097152 /* ContainsHoistedDeclarationOrCompletion */;
                return node;
            }
            // @api
            function updateContinueStatement(node, label) {
                return node.label !== label
                    ? update(createContinueStatement(label), node)
                    : node;
            }
            // @api
            function createBreakStatement(label) {
                var node = createBaseNode(244 /* BreakStatement */);
                node.label = asName(label);
                node.transformFlags |=
                    propagateChildFlags(node.label) |
                        2097152 /* ContainsHoistedDeclarationOrCompletion */;
                return node;
            }
            // @api
            function updateBreakStatement(node, label) {
                return node.label !== label
                    ? update(createBreakStatement(label), node)
                    : node;
            }
            // @api
            function createReturnStatement(expression) {
                var node = createBaseNode(245 /* ReturnStatement */);
                node.expression = expression;
                // return in an ES2018 async generator must be awaited
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        64 /* ContainsES2018 */ |
                        2097152 /* ContainsHoistedDeclarationOrCompletion */;
                return node;
            }
            // @api
            function updateReturnStatement(node, expression) {
                return node.expression !== expression
                    ? update(createReturnStatement(expression), node)
                    : node;
            }
            // @api
            function createWithStatement(expression, statement) {
                var node = createBaseNode(246 /* WithStatement */);
                node.expression = expression;
                node.statement = asEmbeddedStatement(statement);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.statement);
                return node;
            }
            // @api
            function updateWithStatement(node, expression, statement) {
                return node.expression !== expression
                    || node.statement !== statement
                    ? update(createWithStatement(expression, statement), node)
                    : node;
            }
            // @api
            function createSwitchStatement(expression, caseBlock) {
                var node = createBaseNode(247 /* SwitchStatement */);
                node.expression = parenthesizerRules().parenthesizeExpressionForDisallowedComma(expression);
                node.caseBlock = caseBlock;
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.caseBlock);
                return node;
            }
            // @api
            function updateSwitchStatement(node, expression, caseBlock) {
                return node.expression !== expression
                    || node.caseBlock !== caseBlock
                    ? update(createSwitchStatement(expression, caseBlock), node)
                    : node;
            }
            // @api
            function createLabeledStatement(label, statement) {
                var node = createBaseNode(248 /* LabeledStatement */);
                node.label = asName(label);
                node.statement = asEmbeddedStatement(statement);
                node.transformFlags |=
                    propagateChildFlags(node.label) |
                        propagateChildFlags(node.statement);
                return node;
            }
            // @api
            function updateLabeledStatement(node, label, statement) {
                return node.label !== label
                    || node.statement !== statement
                    ? update(createLabeledStatement(label, statement), node)
                    : node;
            }
            // @api
            function createThrowStatement(expression) {
                var node = createBaseNode(249 /* ThrowStatement */);
                node.expression = expression;
                node.transformFlags |= propagateChildFlags(node.expression);
                return node;
            }
            // @api
            function updateThrowStatement(node, expression) {
                return node.expression !== expression
                    ? update(createThrowStatement(expression), node)
                    : node;
            }
            // @api
            function createTryStatement(tryBlock, catchClause, finallyBlock) {
                var node = createBaseNode(250 /* TryStatement */);
                node.tryBlock = tryBlock;
                node.catchClause = catchClause;
                node.finallyBlock = finallyBlock;
                node.transformFlags |=
                    propagateChildFlags(node.tryBlock) |
                        propagateChildFlags(node.catchClause) |
                        propagateChildFlags(node.finallyBlock);
                return node;
            }
            // @api
            function updateTryStatement(node, tryBlock, catchClause, finallyBlock) {
                return node.tryBlock !== tryBlock
                    || node.catchClause !== catchClause
                    || node.finallyBlock !== finallyBlock
                    ? update(createTryStatement(tryBlock, catchClause, finallyBlock), node)
                    : node;
            }
            // @api
            function createDebuggerStatement() {
                return createBaseNode(251 /* DebuggerStatement */);
            }
            // @api
            function createVariableDeclaration(name, exclamationToken, type, initializer) {
                var node = createBaseVariableLikeDeclaration(252 /* VariableDeclaration */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, name, type, initializer && parenthesizerRules().parenthesizeExpressionForDisallowedComma(initializer));
                node.exclamationToken = exclamationToken;
                node.transformFlags |= propagateChildFlags(node.exclamationToken);
                if (exclamationToken) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                return node;
            }
            // @api
            function updateVariableDeclaration(node, name, exclamationToken, type, initializer) {
                return node.name !== name
                    || node.type !== type
                    || node.exclamationToken !== exclamationToken
                    || node.initializer !== initializer
                    ? update(createVariableDeclaration(name, exclamationToken, type, initializer), node)
                    : node;
            }
            // @api
            function createVariableDeclarationList(declarations, flags) {
                if (flags === void 0) {
                    flags = 0 /* None */;
                }
                var node = createBaseNode(253 /* VariableDeclarationList */);
                node.flags |= flags & 3 /* BlockScoped */;
                node.declarations = createNodeArray(declarations);
                node.transformFlags |=
                    propagateChildrenFlags(node.declarations) |
                        2097152 /* ContainsHoistedDeclarationOrCompletion */;
                if (flags & 3 /* BlockScoped */) {
                    node.transformFlags |=
                        512 /* ContainsES2015 */ |
                            131072 /* ContainsBlockScopedBinding */;
                }
                return node;
            }
            // @api
            function updateVariableDeclarationList(node, declarations) {
                return node.declarations !== declarations
                    ? update(createVariableDeclarationList(declarations, node.flags), node)
                    : node;
            }
            // @api
            function createFunctionDeclaration(decorators, modifiers, asteriskToken, name, typeParameters, parameters, type, body) {
                var node = createBaseFunctionLikeDeclaration(254 /* FunctionDeclaration */, decorators, modifiers, name, typeParameters, parameters, type, body);
                node.asteriskToken = asteriskToken;
                if (!node.body || ts.modifiersToFlags(node.modifiers) & 2 /* Ambient */) {
                    node.transformFlags = 1 /* ContainsTypeScript */;
                }
                else {
                    node.transformFlags |=
                        propagateChildFlags(node.asteriskToken) |
                            2097152 /* ContainsHoistedDeclarationOrCompletion */;
                    if (ts.modifiersToFlags(node.modifiers) & 256 /* Async */) {
                        if (node.asteriskToken) {
                            node.transformFlags |= 64 /* ContainsES2018 */;
                        }
                        else {
                            node.transformFlags |= 128 /* ContainsES2017 */;
                        }
                    }
                    else if (node.asteriskToken) {
                        node.transformFlags |= 1024 /* ContainsGenerator */;
                    }
                }
                return node;
            }
            // @api
            function updateFunctionDeclaration(node, decorators, modifiers, asteriskToken, name, typeParameters, parameters, type, body) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.asteriskToken !== asteriskToken
                    || node.name !== name
                    || node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    || node.body !== body
                    ? updateBaseFunctionLikeDeclaration(createFunctionDeclaration(decorators, modifiers, asteriskToken, name, typeParameters, parameters, type, body), node)
                    : node;
            }
            // @api
            function createClassDeclaration(decorators, modifiers, name, typeParameters, heritageClauses, members) {
                var node = createBaseClassLikeDeclaration(255 /* ClassDeclaration */, decorators, modifiers, name, typeParameters, heritageClauses, members);
                if (ts.modifiersToFlags(node.modifiers) & 2 /* Ambient */) {
                    node.transformFlags = 1 /* ContainsTypeScript */;
                }
                else {
                    node.transformFlags |= 512 /* ContainsES2015 */;
                    if (node.transformFlags & 4096 /* ContainsTypeScriptClassSyntax */) {
                        node.transformFlags |= 1 /* ContainsTypeScript */;
                    }
                }
                return node;
            }
            // @api
            function updateClassDeclaration(node, decorators, modifiers, name, typeParameters, heritageClauses, members) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.name !== name
                    || node.typeParameters !== typeParameters
                    || node.heritageClauses !== heritageClauses
                    || node.members !== members
                    ? update(createClassDeclaration(decorators, modifiers, name, typeParameters, heritageClauses, members), node)
                    : node;
            }
            // @api
            function createInterfaceDeclaration(decorators, modifiers, name, typeParameters, heritageClauses, members) {
                var node = createBaseInterfaceOrClassLikeDeclaration(256 /* InterfaceDeclaration */, decorators, modifiers, name, typeParameters, heritageClauses);
                node.members = createNodeArray(members);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateInterfaceDeclaration(node, decorators, modifiers, name, typeParameters, heritageClauses, members) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.name !== name
                    || node.typeParameters !== typeParameters
                    || node.heritageClauses !== heritageClauses
                    || node.members !== members
                    ? update(createInterfaceDeclaration(decorators, modifiers, name, typeParameters, heritageClauses, members), node)
                    : node;
            }
            // @api
            function createTypeAliasDeclaration(decorators, modifiers, name, typeParameters, type) {
                var node = createBaseGenericNamedDeclaration(257 /* TypeAliasDeclaration */, decorators, modifiers, name, typeParameters);
                node.type = type;
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateTypeAliasDeclaration(node, decorators, modifiers, name, typeParameters, type) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.name !== name
                    || node.typeParameters !== typeParameters
                    || node.type !== type
                    ? update(createTypeAliasDeclaration(decorators, modifiers, name, typeParameters, type), node)
                    : node;
            }
            // @api
            function createEnumDeclaration(decorators, modifiers, name, members) {
                var node = createBaseNamedDeclaration(258 /* EnumDeclaration */, decorators, modifiers, name);
                node.members = createNodeArray(members);
                node.transformFlags |=
                    propagateChildrenFlags(node.members) |
                        1 /* ContainsTypeScript */;
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // Enum declarations cannot contain `await`
                return node;
            }
            // @api
            function updateEnumDeclaration(node, decorators, modifiers, name, members) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.name !== name
                    || node.members !== members
                    ? update(createEnumDeclaration(decorators, modifiers, name, members), node)
                    : node;
            }
            // @api
            function createModuleDeclaration(decorators, modifiers, name, body, flags) {
                if (flags === void 0) {
                    flags = 0 /* None */;
                }
                var node = createBaseDeclaration(259 /* ModuleDeclaration */, decorators, modifiers);
                node.flags |= flags & (16 /* Namespace */ | 4 /* NestedNamespace */ | 1024 /* GlobalAugmentation */);
                node.name = name;
                node.body = body;
                if (ts.modifiersToFlags(node.modifiers) & 2 /* Ambient */) {
                    node.transformFlags = 1 /* ContainsTypeScript */;
                }
                else {
                    node.transformFlags |=
                        propagateChildFlags(node.name) |
                            propagateChildFlags(node.body) |
                            1 /* ContainsTypeScript */;
                }
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // Module declarations cannot contain `await`.
                return node;
            }
            // @api
            function updateModuleDeclaration(node, decorators, modifiers, name, body) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.name !== name
                    || node.body !== body
                    ? update(createModuleDeclaration(decorators, modifiers, name, body, node.flags), node)
                    : node;
            }
            // @api
            function createModuleBlock(statements) {
                var node = createBaseNode(260 /* ModuleBlock */);
                node.statements = createNodeArray(statements);
                node.transformFlags |= propagateChildrenFlags(node.statements);
                return node;
            }
            // @api
            function updateModuleBlock(node, statements) {
                return node.statements !== statements
                    ? update(createModuleBlock(statements), node)
                    : node;
            }
            // @api
            function createCaseBlock(clauses) {
                var node = createBaseNode(261 /* CaseBlock */);
                node.clauses = createNodeArray(clauses);
                node.transformFlags |= propagateChildrenFlags(node.clauses);
                return node;
            }
            // @api
            function updateCaseBlock(node, clauses) {
                return node.clauses !== clauses
                    ? update(createCaseBlock(clauses), node)
                    : node;
            }
            // @api
            function createNamespaceExportDeclaration(name) {
                var node = createBaseNamedDeclaration(262 /* NamespaceExportDeclaration */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, name);
                node.transformFlags = 1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateNamespaceExportDeclaration(node, name) {
                return node.name !== name
                    ? update(createNamespaceExportDeclaration(name), node)
                    : node;
            }
            // @api
            function createImportEqualsDeclaration(decorators, modifiers, isTypeOnly, name, moduleReference) {
                var node = createBaseNamedDeclaration(263 /* ImportEqualsDeclaration */, decorators, modifiers, name);
                node.isTypeOnly = isTypeOnly;
                node.moduleReference = moduleReference;
                node.transformFlags |= propagateChildFlags(node.moduleReference);
                if (!ts.isExternalModuleReference(node.moduleReference))
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // Import= declaration is always parsed in an Await context
                return node;
            }
            // @api
            function updateImportEqualsDeclaration(node, decorators, modifiers, isTypeOnly, name, moduleReference) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.isTypeOnly !== isTypeOnly
                    || node.name !== name
                    || node.moduleReference !== moduleReference
                    ? update(createImportEqualsDeclaration(decorators, modifiers, isTypeOnly, name, moduleReference), node)
                    : node;
            }
            // @api
            function createImportDeclaration(decorators, modifiers, importClause, moduleSpecifier) {
                var node = createBaseDeclaration(264 /* ImportDeclaration */, decorators, modifiers);
                node.importClause = importClause;
                node.moduleSpecifier = moduleSpecifier;
                node.transformFlags |=
                    propagateChildFlags(node.importClause) |
                        propagateChildFlags(node.moduleSpecifier);
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateImportDeclaration(node, decorators, modifiers, importClause, moduleSpecifier) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.importClause !== importClause
                    || node.moduleSpecifier !== moduleSpecifier
                    ? update(createImportDeclaration(decorators, modifiers, importClause, moduleSpecifier), node)
                    : node;
            }
            // @api
            function createImportClause(isTypeOnly, name, namedBindings) {
                var node = createBaseNode(265 /* ImportClause */);
                node.isTypeOnly = isTypeOnly;
                node.name = name;
                node.namedBindings = namedBindings;
                node.transformFlags |=
                    propagateChildFlags(node.name) |
                        propagateChildFlags(node.namedBindings);
                if (isTypeOnly) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateImportClause(node, isTypeOnly, name, namedBindings) {
                return node.isTypeOnly !== isTypeOnly
                    || node.name !== name
                    || node.namedBindings !== namedBindings
                    ? update(createImportClause(isTypeOnly, name, namedBindings), node)
                    : node;
            }
            // @api
            function createNamespaceImport(name) {
                var node = createBaseNode(266 /* NamespaceImport */);
                node.name = name;
                node.transformFlags |= propagateChildFlags(node.name);
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateNamespaceImport(node, name) {
                return node.name !== name
                    ? update(createNamespaceImport(name), node)
                    : node;
            }
            // @api
            function createNamespaceExport(name) {
                var node = createBaseNode(272 /* NamespaceExport */);
                node.name = name;
                node.transformFlags |=
                    propagateChildFlags(node.name) |
                        4 /* ContainsESNext */;
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateNamespaceExport(node, name) {
                return node.name !== name
                    ? update(createNamespaceExport(name), node)
                    : node;
            }
            // @api
            function createNamedImports(elements) {
                var node = createBaseNode(267 /* NamedImports */);
                node.elements = createNodeArray(elements);
                node.transformFlags |= propagateChildrenFlags(node.elements);
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateNamedImports(node, elements) {
                return node.elements !== elements
                    ? update(createNamedImports(elements), node)
                    : node;
            }
            // @api
            function createImportSpecifier(propertyName, name) {
                var node = createBaseNode(268 /* ImportSpecifier */);
                node.propertyName = propertyName;
                node.name = name;
                node.transformFlags |=
                    propagateChildFlags(node.propertyName) |
                        propagateChildFlags(node.name);
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateImportSpecifier(node, propertyName, name) {
                return node.propertyName !== propertyName
                    || node.name !== name
                    ? update(createImportSpecifier(propertyName, name), node)
                    : node;
            }
            // @api
            function createExportAssignment(decorators, modifiers, isExportEquals, expression) {
                var node = createBaseDeclaration(269 /* ExportAssignment */, decorators, modifiers);
                node.isExportEquals = isExportEquals;
                node.expression = isExportEquals
                    ? parenthesizerRules().parenthesizeRightSideOfBinary(63 /* EqualsToken */, /*leftSide*/ undefined, expression)
                    : parenthesizerRules().parenthesizeExpressionOfExportDefault(expression);
                node.transformFlags |= propagateChildFlags(node.expression);
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateExportAssignment(node, decorators, modifiers, expression) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.expression !== expression
                    ? update(createExportAssignment(decorators, modifiers, node.isExportEquals, expression), node)
                    : node;
            }
            // @api
            function createExportDeclaration(decorators, modifiers, isTypeOnly, exportClause, moduleSpecifier) {
                var node = createBaseDeclaration(270 /* ExportDeclaration */, decorators, modifiers);
                node.isTypeOnly = isTypeOnly;
                node.exportClause = exportClause;
                node.moduleSpecifier = moduleSpecifier;
                node.transformFlags |=
                    propagateChildFlags(node.exportClause) |
                        propagateChildFlags(node.moduleSpecifier);
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateExportDeclaration(node, decorators, modifiers, isTypeOnly, exportClause, moduleSpecifier) {
                return node.decorators !== decorators
                    || node.modifiers !== modifiers
                    || node.isTypeOnly !== isTypeOnly
                    || node.exportClause !== exportClause
                    || node.moduleSpecifier !== moduleSpecifier
                    ? update(createExportDeclaration(decorators, modifiers, isTypeOnly, exportClause, moduleSpecifier), node)
                    : node;
            }
            // @api
            function createNamedExports(elements) {
                var node = createBaseNode(271 /* NamedExports */);
                node.elements = createNodeArray(elements);
                node.transformFlags |= propagateChildrenFlags(node.elements);
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateNamedExports(node, elements) {
                return node.elements !== elements
                    ? update(createNamedExports(elements), node)
                    : node;
            }
            // @api
            function createExportSpecifier(propertyName, name) {
                var node = createBaseNode(273 /* ExportSpecifier */);
                node.propertyName = asName(propertyName);
                node.name = asName(name);
                node.transformFlags |=
                    propagateChildFlags(node.propertyName) |
                        propagateChildFlags(node.name);
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateExportSpecifier(node, propertyName, name) {
                return node.propertyName !== propertyName
                    || node.name !== name
                    ? update(createExportSpecifier(propertyName, name), node)
                    : node;
            }
            // @api
            function createMissingDeclaration() {
                var node = createBaseDeclaration(274 /* MissingDeclaration */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined);
                return node;
            }
            //
            // Module references
            //
            // @api
            function createExternalModuleReference(expression) {
                var node = createBaseNode(275 /* ExternalModuleReference */);
                node.expression = expression;
                node.transformFlags |= propagateChildFlags(node.expression);
                node.transformFlags &= ~16777216 /* ContainsPossibleTopLevelAwait */; // always parsed in an Await context
                return node;
            }
            // @api
            function updateExternalModuleReference(node, expression) {
                return node.expression !== expression
                    ? update(createExternalModuleReference(expression), node)
                    : node;
            }
            //
            // JSDoc
            //
            // @api
            // createJSDocAllType
            // createJSDocUnknownType
            function createJSDocPrimaryTypeWorker(kind) {
                return createBaseNode(kind);
            }
            // @api
            // createJSDocNonNullableType
            // createJSDocNullableType
            // createJSDocOptionalType
            // createJSDocVariadicType
            // createJSDocNamepathType
            function createJSDocUnaryTypeWorker(kind, type) {
                var node = createBaseNode(kind);
                node.type = type;
                return node;
            }
            // @api
            // updateJSDocNonNullableType
            // updateJSDocNullableType
            // updateJSDocOptionalType
            // updateJSDocVariadicType
            // updateJSDocNamepathType
            function updateJSDocUnaryTypeWorker(kind, node, type) {
                return node.type !== type
                    ? update(createJSDocUnaryTypeWorker(kind, type), node)
                    : node;
            }
            // @api
            function createJSDocFunctionType(parameters, type) {
                var node = createBaseSignatureDeclaration(312 /* JSDocFunctionType */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, parameters, type);
                return node;
            }
            // @api
            function updateJSDocFunctionType(node, parameters, type) {
                return node.parameters !== parameters
                    || node.type !== type
                    ? update(createJSDocFunctionType(parameters, type), node)
                    : node;
            }
            // @api
            function createJSDocTypeLiteral(propertyTags, isArrayType) {
                if (isArrayType === void 0) {
                    isArrayType = false;
                }
                var node = createBaseNode(317 /* JSDocTypeLiteral */);
                node.jsDocPropertyTags = asNodeArray(propertyTags);
                node.isArrayType = isArrayType;
                return node;
            }
            // @api
            function updateJSDocTypeLiteral(node, propertyTags, isArrayType) {
                return node.jsDocPropertyTags !== propertyTags
                    || node.isArrayType !== isArrayType
                    ? update(createJSDocTypeLiteral(propertyTags, isArrayType), node)
                    : node;
            }
            // @api
            function createJSDocTypeExpression(type) {
                var node = createBaseNode(304 /* JSDocTypeExpression */);
                node.type = type;
                return node;
            }
            // @api
            function updateJSDocTypeExpression(node, type) {
                return node.type !== type
                    ? update(createJSDocTypeExpression(type), node)
                    : node;
            }
            // @api
            function createJSDocSignature(typeParameters, parameters, type) {
                var node = createBaseNode(318 /* JSDocSignature */);
                node.typeParameters = asNodeArray(typeParameters);
                node.parameters = createNodeArray(parameters);
                node.type = type;
                return node;
            }
            // @api
            function updateJSDocSignature(node, typeParameters, parameters, type) {
                return node.typeParameters !== typeParameters
                    || node.parameters !== parameters
                    || node.type !== type
                    ? update(createJSDocSignature(typeParameters, parameters, type), node)
                    : node;
            }
            function getDefaultTagName(node) {
                var defaultTagName = getDefaultTagNameForKind(node.kind);
                return node.tagName.escapedText === ts.escapeLeadingUnderscores(defaultTagName)
                    ? node.tagName
                    : createIdentifier(defaultTagName);
            }
            // @api
            function createBaseJSDocTag(kind, tagName, comment) {
                var node = createBaseNode(kind);
                node.tagName = tagName;
                node.comment = comment;
                return node;
            }
            // @api
            function createJSDocTemplateTag(tagName, constraint, typeParameters, comment) {
                var node = createBaseJSDocTag(339 /* JSDocTemplateTag */, tagName !== null && tagName !== void 0 ? tagName : createIdentifier("template"), comment);
                node.constraint = constraint;
                node.typeParameters = createNodeArray(typeParameters);
                return node;
            }
            // @api
            function updateJSDocTemplateTag(node, tagName, constraint, typeParameters, comment) {
                if (tagName === void 0) {
                    tagName = getDefaultTagName(node);
                }
                return node.tagName !== tagName
                    || node.constraint !== constraint
                    || node.typeParameters !== typeParameters
                    || node.comment !== comment
                    ? update(createJSDocTemplateTag(tagName, constraint, typeParameters, comment), node)
                    : node;
            }
            // @api
            function createJSDocTypedefTag(tagName, typeExpression, fullName, comment) {
                var node = createBaseJSDocTag(340 /* JSDocTypedefTag */, tagName !== null && tagName !== void 0 ? tagName : createIdentifier("typedef"), comment);
                node.typeExpression = typeExpression;
                node.fullName = fullName;
                node.name = ts.getJSDocTypeAliasName(fullName);
                return node;
            }
            // @api
            function updateJSDocTypedefTag(node, tagName, typeExpression, fullName, comment) {
                if (tagName === void 0) {
                    tagName = getDefaultTagName(node);
                }
                return node.tagName !== tagName
                    || node.typeExpression !== typeExpression
                    || node.fullName !== fullName
                    || node.comment !== comment
                    ? update(createJSDocTypedefTag(tagName, typeExpression, fullName, comment), node)
                    : node;
            }
            // @api
            function createJSDocParameterTag(tagName, name, isBracketed, typeExpression, isNameFirst, comment) {
                var node = createBaseJSDocTag(335 /* JSDocParameterTag */, tagName !== null && tagName !== void 0 ? tagName : createIdentifier("param"), comment);
                node.typeExpression = typeExpression;
                node.name = name;
                node.isNameFirst = !!isNameFirst;
                node.isBracketed = isBracketed;
                return node;
            }
            // @api
            function updateJSDocParameterTag(node, tagName, name, isBracketed, typeExpression, isNameFirst, comment) {
                if (tagName === void 0) {
                    tagName = getDefaultTagName(node);
                }
                return node.tagName !== tagName
                    || node.name !== name
                    || node.isBracketed !== isBracketed
                    || node.typeExpression !== typeExpression
                    || node.isNameFirst !== isNameFirst
                    || node.comment !== comment
                    ? update(createJSDocParameterTag(tagName, name, isBracketed, typeExpression, isNameFirst, comment), node)
                    : node;
            }
            // @api
            function createJSDocPropertyTag(tagName, name, isBracketed, typeExpression, isNameFirst, comment) {
                var node = createBaseJSDocTag(342 /* JSDocPropertyTag */, tagName !== null && tagName !== void 0 ? tagName : createIdentifier("prop"), comment);
                node.typeExpression = typeExpression;
                node.name = name;
                node.isNameFirst = !!isNameFirst;
                node.isBracketed = isBracketed;
                return node;
            }
            // @api
            function updateJSDocPropertyTag(node, tagName, name, isBracketed, typeExpression, isNameFirst, comment) {
                if (tagName === void 0) {
                    tagName = getDefaultTagName(node);
                }
                return node.tagName !== tagName
                    || node.name !== name
                    || node.isBracketed !== isBracketed
                    || node.typeExpression !== typeExpression
                    || node.isNameFirst !== isNameFirst
                    || node.comment !== comment
                    ? update(createJSDocPropertyTag(tagName, name, isBracketed, typeExpression, isNameFirst, comment), node)
                    : node;
            }
            // @api
            function createJSDocCallbackTag(tagName, typeExpression, fullName, comment) {
                var node = createBaseJSDocTag(333 /* JSDocCallbackTag */, tagName !== null && tagName !== void 0 ? tagName : createIdentifier("callback"), comment);
                node.typeExpression = typeExpression;
                node.fullName = fullName;
                node.name = ts.getJSDocTypeAliasName(fullName);
                return node;
            }
            // @api
            function updateJSDocCallbackTag(node, tagName, typeExpression, fullName, comment) {
                if (tagName === void 0) {
                    tagName = getDefaultTagName(node);
                }
                return node.tagName !== tagName
                    || node.typeExpression !== typeExpression
                    || node.fullName !== fullName
                    || node.comment !== comment
                    ? update(createJSDocCallbackTag(tagName, typeExpression, fullName, comment), node)
                    : node;
            }
            // @api
            function createJSDocAugmentsTag(tagName, className, comment) {
                var node = createBaseJSDocTag(323 /* JSDocAugmentsTag */, tagName !== null && tagName !== void 0 ? tagName : createIdentifier("augments"), comment);
                node.class = className;
                return node;
            }
            // @api
            function updateJSDocAugmentsTag(node, tagName, className, comment) {
                if (tagName === void 0) {
                    tagName = getDefaultTagName(node);
                }
                return node.tagName !== tagName
                    || node.class !== className
                    || node.comment !== comment
                    ? update(createJSDocAugmentsTag(tagName, className, comment), node)
                    : node;
            }
            // @api
            function createJSDocImplementsTag(tagName, className, comment) {
                var node = createBaseJSDocTag(324 /* JSDocImplementsTag */, tagName !== null && tagName !== void 0 ? tagName : createIdentifier("implements"), comment);
                node.class = className;
                return node;
            }
            // @api
            function createJSDocSeeTag(tagName, name, comment) {
                var node = createBaseJSDocTag(341 /* JSDocSeeTag */, tagName !== null && tagName !== void 0 ? tagName : createIdentifier("see"), comment);
                node.name = name;
                return node;
            }
            // @api
            function updateJSDocSeeTag(node, tagName, name, comment) {
                return node.tagName !== tagName
                    || node.name !== name
                    || node.comment !== comment
                    ? update(createJSDocSeeTag(tagName, name, comment), node)
                    : node;
            }
            // @api
            function createJSDocNameReference(name) {
                var node = createBaseNode(305 /* JSDocNameReference */);
                node.name = name;
                return node;
            }
            // @api
            function updateJSDocNameReference(node, name) {
                return node.name !== name
                    ? update(createJSDocNameReference(name), node)
                    : node;
            }
            // @api
            function createJSDocMemberName(left, right) {
                var node = createBaseNode(306 /* JSDocMemberName */);
                node.left = left;
                node.right = right;
                node.transformFlags |=
                    propagateChildFlags(node.left) |
                        propagateChildFlags(node.right);
                return node;
            }
            // @api
            function updateJSDocMemberName(node, left, right) {
                return node.left !== left
                    || node.right !== right
                    ? update(createJSDocMemberName(left, right), node)
                    : node;
            }
            // @api
            function createJSDocLink(name, text) {
                var node = createBaseNode(319 /* JSDocLink */);
                node.name = name;
                node.text = text;
                return node;
            }
            // @api
            function updateJSDocLink(node, name, text) {
                return node.name !== name
                    ? update(createJSDocLink(name, text), node)
                    : node;
            }
            // @api
            function createJSDocLinkCode(name, text) {
                var node = createBaseNode(320 /* JSDocLinkCode */);
                node.name = name;
                node.text = text;
                return node;
            }
            // @api
            function updateJSDocLinkCode(node, name, text) {
                return node.name !== name
                    ? update(createJSDocLinkCode(name, text), node)
                    : node;
            }
            // @api
            function createJSDocLinkPlain(name, text) {
                var node = createBaseNode(321 /* JSDocLinkPlain */);
                node.name = name;
                node.text = text;
                return node;
            }
            // @api
            function updateJSDocLinkPlain(node, name, text) {
                return node.name !== name
                    ? update(createJSDocLinkPlain(name, text), node)
                    : node;
            }
            // @api
            function updateJSDocImplementsTag(node, tagName, className, comment) {
                if (tagName === void 0) {
                    tagName = getDefaultTagName(node);
                }
                return node.tagName !== tagName
                    || node.class !== className
                    || node.comment !== comment
                    ? update(createJSDocImplementsTag(tagName, className, comment), node)
                    : node;
            }
            // @api
            // createJSDocAuthorTag
            // createJSDocClassTag
            // createJSDocPublicTag
            // createJSDocPrivateTag
            // createJSDocProtectedTag
            // createJSDocReadonlyTag
            // createJSDocDeprecatedTag
            function createJSDocSimpleTagWorker(kind, tagName, comment) {
                var node = createBaseJSDocTag(kind, tagName !== null && tagName !== void 0 ? tagName : createIdentifier(getDefaultTagNameForKind(kind)), comment);
                return node;
            }
            // @api
            // updateJSDocAuthorTag
            // updateJSDocClassTag
            // updateJSDocPublicTag
            // updateJSDocPrivateTag
            // updateJSDocProtectedTag
            // updateJSDocReadonlyTag
            // updateJSDocDeprecatedTag
            function updateJSDocSimpleTagWorker(kind, node, tagName, comment) {
                if (tagName === void 0) {
                    tagName = getDefaultTagName(node);
                }
                return node.tagName !== tagName
                    || node.comment !== comment
                    ? update(createJSDocSimpleTagWorker(kind, tagName, comment), node) :
                    node;
            }
            // @api
            // createJSDocTypeTag
            // createJSDocReturnTag
            // createJSDocThisTag
            // createJSDocEnumTag
            function createJSDocTypeLikeTagWorker(kind, tagName, typeExpression, comment) {
                var node = createBaseJSDocTag(kind, tagName !== null && tagName !== void 0 ? tagName : createIdentifier(getDefaultTagNameForKind(kind)), comment);
                node.typeExpression = typeExpression;
                return node;
            }
            // @api
            // updateJSDocTypeTag
            // updateJSDocReturnTag
            // updateJSDocThisTag
            // updateJSDocEnumTag
            function updateJSDocTypeLikeTagWorker(kind, node, tagName, typeExpression, comment) {
                if (tagName === void 0) {
                    tagName = getDefaultTagName(node);
                }
                return node.tagName !== tagName
                    || node.typeExpression !== typeExpression
                    || node.comment !== comment
                    ? update(createJSDocTypeLikeTagWorker(kind, tagName, typeExpression, comment), node)
                    : node;
            }
            // @api
            function createJSDocUnknownTag(tagName, comment) {
                var node = createBaseJSDocTag(322 /* JSDocTag */, tagName, comment);
                return node;
            }
            // @api
            function updateJSDocUnknownTag(node, tagName, comment) {
                return node.tagName !== tagName
                    || node.comment !== comment
                    ? update(createJSDocUnknownTag(tagName, comment), node)
                    : node;
            }
            // @api
            function createJSDocText(text) {
                var node = createBaseNode(316 /* JSDocText */);
                node.text = text;
                return node;
            }
            // @api
            function updateJSDocText(node, text) {
                return node.text !== text
                    ? update(createJSDocText(text), node)
                    : node;
            }
            // @api
            function createJSDocComment(comment, tags) {
                var node = createBaseNode(315 /* JSDocComment */);
                node.comment = comment;
                node.tags = asNodeArray(tags);
                return node;
            }
            // @api
            function updateJSDocComment(node, comment, tags) {
                return node.comment !== comment
                    || node.tags !== tags
                    ? update(createJSDocComment(comment, tags), node)
                    : node;
            }
            //
            // JSX
            //
            // @api
            function createJsxElement(openingElement, children, closingElement) {
                var node = createBaseNode(276 /* JsxElement */);
                node.openingElement = openingElement;
                node.children = createNodeArray(children);
                node.closingElement = closingElement;
                node.transformFlags |=
                    propagateChildFlags(node.openingElement) |
                        propagateChildrenFlags(node.children) |
                        propagateChildFlags(node.closingElement) |
                        2 /* ContainsJsx */;
                return node;
            }
            // @api
            function updateJsxElement(node, openingElement, children, closingElement) {
                return node.openingElement !== openingElement
                    || node.children !== children
                    || node.closingElement !== closingElement
                    ? update(createJsxElement(openingElement, children, closingElement), node)
                    : node;
            }
            // @api
            function createJsxSelfClosingElement(tagName, typeArguments, attributes) {
                var node = createBaseNode(277 /* JsxSelfClosingElement */);
                node.tagName = tagName;
                node.typeArguments = asNodeArray(typeArguments);
                node.attributes = attributes;
                node.transformFlags |=
                    propagateChildFlags(node.tagName) |
                        propagateChildrenFlags(node.typeArguments) |
                        propagateChildFlags(node.attributes) |
                        2 /* ContainsJsx */;
                if (node.typeArguments) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                return node;
            }
            // @api
            function updateJsxSelfClosingElement(node, tagName, typeArguments, attributes) {
                return node.tagName !== tagName
                    || node.typeArguments !== typeArguments
                    || node.attributes !== attributes
                    ? update(createJsxSelfClosingElement(tagName, typeArguments, attributes), node)
                    : node;
            }
            // @api
            function createJsxOpeningElement(tagName, typeArguments, attributes) {
                var node = createBaseNode(278 /* JsxOpeningElement */);
                node.tagName = tagName;
                node.typeArguments = asNodeArray(typeArguments);
                node.attributes = attributes;
                node.transformFlags |=
                    propagateChildFlags(node.tagName) |
                        propagateChildrenFlags(node.typeArguments) |
                        propagateChildFlags(node.attributes) |
                        2 /* ContainsJsx */;
                if (typeArguments) {
                    node.transformFlags |= 1 /* ContainsTypeScript */;
                }
                return node;
            }
            // @api
            function updateJsxOpeningElement(node, tagName, typeArguments, attributes) {
                return node.tagName !== tagName
                    || node.typeArguments !== typeArguments
                    || node.attributes !== attributes
                    ? update(createJsxOpeningElement(tagName, typeArguments, attributes), node)
                    : node;
            }
            // @api
            function createJsxClosingElement(tagName) {
                var node = createBaseNode(279 /* JsxClosingElement */);
                node.tagName = tagName;
                node.transformFlags |=
                    propagateChildFlags(node.tagName) |
                        2 /* ContainsJsx */;
                return node;
            }
            // @api
            function updateJsxClosingElement(node, tagName) {
                return node.tagName !== tagName
                    ? update(createJsxClosingElement(tagName), node)
                    : node;
            }
            // @api
            function createJsxFragment(openingFragment, children, closingFragment) {
                var node = createBaseNode(280 /* JsxFragment */);
                node.openingFragment = openingFragment;
                node.children = createNodeArray(children);
                node.closingFragment = closingFragment;
                node.transformFlags |=
                    propagateChildFlags(node.openingFragment) |
                        propagateChildrenFlags(node.children) |
                        propagateChildFlags(node.closingFragment) |
                        2 /* ContainsJsx */;
                return node;
            }
            // @api
            function updateJsxFragment(node, openingFragment, children, closingFragment) {
                return node.openingFragment !== openingFragment
                    || node.children !== children
                    || node.closingFragment !== closingFragment
                    ? update(createJsxFragment(openingFragment, children, closingFragment), node)
                    : node;
            }
            // @api
            function createJsxText(text, containsOnlyTriviaWhiteSpaces) {
                var node = createBaseNode(11 /* JsxText */);
                node.text = text;
                node.containsOnlyTriviaWhiteSpaces = !!containsOnlyTriviaWhiteSpaces;
                node.transformFlags |= 2 /* ContainsJsx */;
                return node;
            }
            // @api
            function updateJsxText(node, text, containsOnlyTriviaWhiteSpaces) {
                return node.text !== text
                    || node.containsOnlyTriviaWhiteSpaces !== containsOnlyTriviaWhiteSpaces
                    ? update(createJsxText(text, containsOnlyTriviaWhiteSpaces), node)
                    : node;
            }
            // @api
            function createJsxOpeningFragment() {
                var node = createBaseNode(281 /* JsxOpeningFragment */);
                node.transformFlags |= 2 /* ContainsJsx */;
                return node;
            }
            // @api
            function createJsxJsxClosingFragment() {
                var node = createBaseNode(282 /* JsxClosingFragment */);
                node.transformFlags |= 2 /* ContainsJsx */;
                return node;
            }
            // @api
            function createJsxAttribute(name, initializer) {
                var node = createBaseNode(283 /* JsxAttribute */);
                node.name = name;
                node.initializer = initializer;
                node.transformFlags |=
                    propagateChildFlags(node.name) |
                        propagateChildFlags(node.initializer) |
                        2 /* ContainsJsx */;
                return node;
            }
            // @api
            function updateJsxAttribute(node, name, initializer) {
                return node.name !== name
                    || node.initializer !== initializer
                    ? update(createJsxAttribute(name, initializer), node)
                    : node;
            }
            // @api
            function createJsxAttributes(properties) {
                var node = createBaseNode(284 /* JsxAttributes */);
                node.properties = createNodeArray(properties);
                node.transformFlags |=
                    propagateChildrenFlags(node.properties) |
                        2 /* ContainsJsx */;
                return node;
            }
            // @api
            function updateJsxAttributes(node, properties) {
                return node.properties !== properties
                    ? update(createJsxAttributes(properties), node)
                    : node;
            }
            // @api
            function createJsxSpreadAttribute(expression) {
                var node = createBaseNode(285 /* JsxSpreadAttribute */);
                node.expression = expression;
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        2 /* ContainsJsx */;
                return node;
            }
            // @api
            function updateJsxSpreadAttribute(node, expression) {
                return node.expression !== expression
                    ? update(createJsxSpreadAttribute(expression), node)
                    : node;
            }
            // @api
            function createJsxExpression(dotDotDotToken, expression) {
                var node = createBaseNode(286 /* JsxExpression */);
                node.dotDotDotToken = dotDotDotToken;
                node.expression = expression;
                node.transformFlags |=
                    propagateChildFlags(node.dotDotDotToken) |
                        propagateChildFlags(node.expression) |
                        2 /* ContainsJsx */;
                return node;
            }
            // @api
            function updateJsxExpression(node, expression) {
                return node.expression !== expression
                    ? update(createJsxExpression(node.dotDotDotToken, expression), node)
                    : node;
            }
            //
            // Clauses
            //
            // @api
            function createCaseClause(expression, statements) {
                var node = createBaseNode(287 /* CaseClause */);
                node.expression = parenthesizerRules().parenthesizeExpressionForDisallowedComma(expression);
                node.statements = createNodeArray(statements);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildrenFlags(node.statements);
                return node;
            }
            // @api
            function updateCaseClause(node, expression, statements) {
                return node.expression !== expression
                    || node.statements !== statements
                    ? update(createCaseClause(expression, statements), node)
                    : node;
            }
            // @api
            function createDefaultClause(statements) {
                var node = createBaseNode(288 /* DefaultClause */);
                node.statements = createNodeArray(statements);
                node.transformFlags = propagateChildrenFlags(node.statements);
                return node;
            }
            // @api
            function updateDefaultClause(node, statements) {
                return node.statements !== statements
                    ? update(createDefaultClause(statements), node)
                    : node;
            }
            // @api
            function createHeritageClause(token, types) {
                var node = createBaseNode(289 /* HeritageClause */);
                node.token = token;
                node.types = createNodeArray(types);
                node.transformFlags |= propagateChildrenFlags(node.types);
                switch (token) {
                    case 94 /* ExtendsKeyword */:
                        node.transformFlags |= 512 /* ContainsES2015 */;
                        break;
                    case 117 /* ImplementsKeyword */:
                        node.transformFlags |= 1 /* ContainsTypeScript */;
                        break;
                    default:
                        return ts.Debug.assertNever(token);
                }
                return node;
            }
            // @api
            function updateHeritageClause(node, types) {
                return node.types !== types
                    ? update(createHeritageClause(node.token, types), node)
                    : node;
            }
            // @api
            function createCatchClause(variableDeclaration, block) {
                var node = createBaseNode(290 /* CatchClause */);
                variableDeclaration = !ts.isString(variableDeclaration) ? variableDeclaration : createVariableDeclaration(variableDeclaration, 
                /*exclamationToken*/ undefined, 
                /*type*/ undefined, 
                /*initializer*/ undefined);
                node.variableDeclaration = variableDeclaration;
                node.block = block;
                node.transformFlags |=
                    propagateChildFlags(node.variableDeclaration) |
                        propagateChildFlags(node.block);
                if (!variableDeclaration)
                    node.transformFlags |= 32 /* ContainsES2019 */;
                return node;
            }
            // @api
            function updateCatchClause(node, variableDeclaration, block) {
                return node.variableDeclaration !== variableDeclaration
                    || node.block !== block
                    ? update(createCatchClause(variableDeclaration, block), node)
                    : node;
            }
            //
            // Property assignments
            //
            // @api
            function createPropertyAssignment(name, initializer) {
                var node = createBaseNamedDeclaration(291 /* PropertyAssignment */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, name);
                node.initializer = parenthesizerRules().parenthesizeExpressionForDisallowedComma(initializer);
                node.transformFlags |=
                    propagateChildFlags(node.name) |
                        propagateChildFlags(node.initializer);
                return node;
            }
            function finishUpdatePropertyAssignment(updated, original) {
                // copy children used only for error reporting
                if (original.decorators)
                    updated.decorators = original.decorators;
                if (original.modifiers)
                    updated.modifiers = original.modifiers;
                if (original.questionToken)
                    updated.questionToken = original.questionToken;
                if (original.exclamationToken)
                    updated.exclamationToken = original.exclamationToken;
                return update(updated, original);
            }
            // @api
            function updatePropertyAssignment(node, name, initializer) {
                return node.name !== name
                    || node.initializer !== initializer
                    ? finishUpdatePropertyAssignment(createPropertyAssignment(name, initializer), node)
                    : node;
            }
            // @api
            function createShorthandPropertyAssignment(name, objectAssignmentInitializer) {
                var node = createBaseNamedDeclaration(292 /* ShorthandPropertyAssignment */, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, name);
                node.objectAssignmentInitializer = objectAssignmentInitializer && parenthesizerRules().parenthesizeExpressionForDisallowedComma(objectAssignmentInitializer);
                node.transformFlags |=
                    propagateChildFlags(node.objectAssignmentInitializer) |
                        512 /* ContainsES2015 */;
                return node;
            }
            function finishUpdateShorthandPropertyAssignment(updated, original) {
                // copy children used only for error reporting
                if (original.decorators)
                    updated.decorators = original.decorators;
                if (original.modifiers)
                    updated.modifiers = original.modifiers;
                if (original.equalsToken)
                    updated.equalsToken = original.equalsToken;
                if (original.questionToken)
                    updated.questionToken = original.questionToken;
                if (original.exclamationToken)
                    updated.exclamationToken = original.exclamationToken;
                return update(updated, original);
            }
            // @api
            function updateShorthandPropertyAssignment(node, name, objectAssignmentInitializer) {
                return node.name !== name
                    || node.objectAssignmentInitializer !== objectAssignmentInitializer
                    ? finishUpdateShorthandPropertyAssignment(createShorthandPropertyAssignment(name, objectAssignmentInitializer), node)
                    : node;
            }
            // @api
            function createSpreadAssignment(expression) {
                var node = createBaseNode(293 /* SpreadAssignment */);
                node.expression = parenthesizerRules().parenthesizeExpressionForDisallowedComma(expression);
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        64 /* ContainsES2018 */ |
                        32768 /* ContainsObjectRestOrSpread */;
                return node;
            }
            // @api
            function updateSpreadAssignment(node, expression) {
                return node.expression !== expression
                    ? update(createSpreadAssignment(expression), node)
                    : node;
            }
            //
            // Enum
            //
            // @api
            function createEnumMember(name, initializer) {
                var node = createBaseNode(294 /* EnumMember */);
                node.name = asName(name);
                node.initializer = initializer && parenthesizerRules().parenthesizeExpressionForDisallowedComma(initializer);
                node.transformFlags |=
                    propagateChildFlags(node.name) |
                        propagateChildFlags(node.initializer) |
                        1 /* ContainsTypeScript */;
                return node;
            }
            // @api
            function updateEnumMember(node, name, initializer) {
                return node.name !== name
                    || node.initializer !== initializer
                    ? update(createEnumMember(name, initializer), node)
                    : node;
            }
            //
            // Top-level nodes
            //
            // @api
            function createSourceFile(statements, endOfFileToken, flags) {
                var node = baseFactory.createBaseSourceFileNode(300 /* SourceFile */);
                node.statements = createNodeArray(statements);
                node.endOfFileToken = endOfFileToken;
                node.flags |= flags;
                node.fileName = "";
                node.text = "";
                node.languageVersion = 0;
                node.languageVariant = 0;
                node.scriptKind = 0;
                node.isDeclarationFile = false;
                node.hasNoDefaultLib = false;
                node.transformFlags |=
                    propagateChildrenFlags(node.statements) |
                        propagateChildFlags(node.endOfFileToken);
                return node;
            }
            function cloneSourceFileWithChanges(source, statements, isDeclarationFile, referencedFiles, typeReferences, hasNoDefaultLib, libReferences) {
                var node = baseFactory.createBaseSourceFileNode(300 /* SourceFile */);
                for (var p in source) {
                    if (p === "emitNode" || ts.hasProperty(node, p) || !ts.hasProperty(source, p))
                        continue;
                    node[p] = source[p];
                }
                node.flags |= source.flags;
                node.statements = createNodeArray(statements);
                node.endOfFileToken = source.endOfFileToken;
                node.isDeclarationFile = isDeclarationFile;
                node.referencedFiles = referencedFiles;
                node.typeReferenceDirectives = typeReferences;
                node.hasNoDefaultLib = hasNoDefaultLib;
                node.libReferenceDirectives = libReferences;
                node.transformFlags =
                    propagateChildrenFlags(node.statements) |
                        propagateChildFlags(node.endOfFileToken);
                return node;
            }
            // @api
            function updateSourceFile(node, statements, isDeclarationFile, referencedFiles, typeReferenceDirectives, hasNoDefaultLib, libReferenceDirectives) {
                if (isDeclarationFile === void 0) {
                    isDeclarationFile = node.isDeclarationFile;
                }
                if (referencedFiles === void 0) {
                    referencedFiles = node.referencedFiles;
                }
                if (typeReferenceDirectives === void 0) {
                    typeReferenceDirectives = node.typeReferenceDirectives;
                }
                if (hasNoDefaultLib === void 0) {
                    hasNoDefaultLib = node.hasNoDefaultLib;
                }
                if (libReferenceDirectives === void 0) {
                    libReferenceDirectives = node.libReferenceDirectives;
                }
                return node.statements !== statements
                    || node.isDeclarationFile !== isDeclarationFile
                    || node.referencedFiles !== referencedFiles
                    || node.typeReferenceDirectives !== typeReferenceDirectives
                    || node.hasNoDefaultLib !== hasNoDefaultLib
                    || node.libReferenceDirectives !== libReferenceDirectives
                    ? update(cloneSourceFileWithChanges(node, statements, isDeclarationFile, referencedFiles, typeReferenceDirectives, hasNoDefaultLib, libReferenceDirectives), node)
                    : node;
            }
            // @api
            function createBundle(sourceFiles, prepends) {
                if (prepends === void 0) {
                    prepends = ts.emptyArray;
                }
                var node = createBaseNode(301 /* Bundle */);
                node.prepends = prepends;
                node.sourceFiles = sourceFiles;
                return node;
            }
            // @api
            function updateBundle(node, sourceFiles, prepends) {
                if (prepends === void 0) {
                    prepends = ts.emptyArray;
                }
                return node.sourceFiles !== sourceFiles
                    || node.prepends !== prepends
                    ? update(createBundle(sourceFiles, prepends), node)
                    : node;
            }
            // @api
            function createUnparsedSource(prologues, syntheticReferences, texts) {
                var node = createBaseNode(302 /* UnparsedSource */);
                node.prologues = prologues;
                node.syntheticReferences = syntheticReferences;
                node.texts = texts;
                node.fileName = "";
                node.text = "";
                node.referencedFiles = ts.emptyArray;
                node.libReferenceDirectives = ts.emptyArray;
                node.getLineAndCharacterOfPosition = function (pos) { return ts.getLineAndCharacterOfPosition(node, pos); };
                return node;
            }
            function createBaseUnparsedNode(kind, data) {
                var node = createBaseNode(kind);
                node.data = data;
                return node;
            }
            // @api
            function createUnparsedPrologue(data) {
                return createBaseUnparsedNode(295 /* UnparsedPrologue */, data);
            }
            // @api
            function createUnparsedPrepend(data, texts) {
                var node = createBaseUnparsedNode(296 /* UnparsedPrepend */, data);
                node.texts = texts;
                return node;
            }
            // @api
            function createUnparsedTextLike(data, internal) {
                return createBaseUnparsedNode(internal ? 298 /* UnparsedInternalText */ : 297 /* UnparsedText */, data);
            }
            // @api
            function createUnparsedSyntheticReference(section) {
                var node = createBaseNode(299 /* UnparsedSyntheticReference */);
                node.data = section.data;
                node.section = section;
                return node;
            }
            // @api
            function createInputFiles() {
                var node = createBaseNode(303 /* InputFiles */);
                node.javascriptText = "";
                node.declarationText = "";
                return node;
            }
            //
            // Synthetic Nodes (used by checker)
            //
            // @api
            function createSyntheticExpression(type, isSpread, tupleNameSource) {
                if (isSpread === void 0) {
                    isSpread = false;
                }
                var node = createBaseNode(230 /* SyntheticExpression */);
                node.type = type;
                node.isSpread = isSpread;
                node.tupleNameSource = tupleNameSource;
                return node;
            }
            // @api
            function createSyntaxList(children) {
                var node = createBaseNode(343 /* SyntaxList */);
                node._children = children;
                return node;
            }
            //
            // Transformation nodes
            //
            /**
             * Creates a synthetic statement to act as a placeholder for a not-emitted statement in
             * order to preserve comments.
             *
             * @param original The original statement.
             */
            // @api
            function createNotEmittedStatement(original) {
                var node = createBaseNode(344 /* NotEmittedStatement */);
                node.original = original;
                ts.setTextRange(node, original);
                return node;
            }
            /**
             * Creates a synthetic expression to act as a placeholder for a not-emitted expression in
             * order to preserve comments or sourcemap positions.
             *
             * @param expression The inner expression to emit.
             * @param original The original outer expression.
             */
            // @api
            function createPartiallyEmittedExpression(expression, original) {
                var node = createBaseNode(345 /* PartiallyEmittedExpression */);
                node.expression = expression;
                node.original = original;
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        1 /* ContainsTypeScript */;
                ts.setTextRange(node, original);
                return node;
            }
            // @api
            function updatePartiallyEmittedExpression(node, expression) {
                return node.expression !== expression
                    ? update(createPartiallyEmittedExpression(expression, node.original), node)
                    : node;
            }
            function flattenCommaElements(node) {
                if (ts.nodeIsSynthesized(node) && !ts.isParseTreeNode(node) && !node.original && !node.emitNode && !node.id) {
                    if (ts.isCommaListExpression(node)) {
                        return node.elements;
                    }
                    if (ts.isBinaryExpression(node) && ts.isCommaToken(node.operatorToken)) {
                        return [node.left, node.right];
                    }
                }
                return node;
            }
            // @api
            function createCommaListExpression(elements) {
                var node = createBaseNode(346 /* CommaListExpression */);
                node.elements = createNodeArray(ts.sameFlatMap(elements, flattenCommaElements));
                node.transformFlags |= propagateChildrenFlags(node.elements);
                return node;
            }
            // @api
            function updateCommaListExpression(node, elements) {
                return node.elements !== elements
                    ? update(createCommaListExpression(elements), node)
                    : node;
            }
            /**
             * Creates a synthetic element to act as a placeholder for the end of an emitted declaration in
             * order to properly emit exports.
             */
            // @api
            function createEndOfDeclarationMarker(original) {
                var node = createBaseNode(348 /* EndOfDeclarationMarker */);
                node.emitNode = {};
                node.original = original;
                return node;
            }
            /**
             * Creates a synthetic element to act as a placeholder for the beginning of a merged declaration in
             * order to properly emit exports.
             */
            // @api
            function createMergeDeclarationMarker(original) {
                var node = createBaseNode(347 /* MergeDeclarationMarker */);
                node.emitNode = {};
                node.original = original;
                return node;
            }
            // @api
            function createSyntheticReferenceExpression(expression, thisArg) {
                var node = createBaseNode(349 /* SyntheticReferenceExpression */);
                node.expression = expression;
                node.thisArg = thisArg;
                node.transformFlags |=
                    propagateChildFlags(node.expression) |
                        propagateChildFlags(node.thisArg);
                return node;
            }
            // @api
            function updateSyntheticReferenceExpression(node, expression, thisArg) {
                return node.expression !== expression
                    || node.thisArg !== thisArg
                    ? update(createSyntheticReferenceExpression(expression, thisArg), node)
                    : node;
            }
            function cloneNode(node) {
                // We don't use "clone" from core.ts here, as we need to preserve the prototype chain of
                // the original node. We also need to exclude specific properties and only include own-
                // properties (to skip members already defined on the shared prototype).
                if (node === undefined) {
                    return node;
                }
                var clone = ts.isSourceFile(node) ? baseFactory.createBaseSourceFileNode(300 /* SourceFile */) :
                    ts.isIdentifier(node) ? baseFactory.createBaseIdentifierNode(79 /* Identifier */) :
                        ts.isPrivateIdentifier(node) ? baseFactory.createBasePrivateIdentifierNode(80 /* PrivateIdentifier */) :
                            !ts.isNodeKind(node.kind) ? baseFactory.createBaseTokenNode(node.kind) :
                                baseFactory.createBaseNode(node.kind);
                clone.flags |= (node.flags & ~8 /* Synthesized */);
                clone.transformFlags = node.transformFlags;
                setOriginalNode(clone, node);
                for (var key in node) {
                    if (clone.hasOwnProperty(key) || !node.hasOwnProperty(key)) {
                        continue;
                    }
                    clone[key] = node[key];
                }
                return clone;
            }
            function createImmediatelyInvokedFunctionExpression(statements, param, paramValue) {
                return createCallExpression(createFunctionExpression(
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, 
                /*parameters*/ param ? [param] : [], 
                /*type*/ undefined, createBlock(statements, /*multiLine*/ true)), 
                /*typeArguments*/ undefined, 
                /*argumentsArray*/ paramValue ? [paramValue] : []);
            }
            function createImmediatelyInvokedArrowFunction(statements, param, paramValue) {
                return createCallExpression(createArrowFunction(
                /*modifiers*/ undefined, 
                /*typeParameters*/ undefined, 
                /*parameters*/ param ? [param] : [], 
                /*type*/ undefined, 
                /*equalsGreaterThanToken*/ undefined, createBlock(statements, /*multiLine*/ true)), 
                /*typeArguments*/ undefined, 
                /*argumentsArray*/ paramValue ? [paramValue] : []);
            }
            function createVoidZero() {
                return createVoidExpression(createNumericLiteral("0"));
            }
            function createExportDefault(expression) {
                return createExportAssignment(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*isExportEquals*/ false, expression);
            }
            function createExternalModuleExport(exportName) {
                return createExportDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*isTypeOnly*/ false, createNamedExports([
                    createExportSpecifier(/*propertyName*/ undefined, exportName)
                ]));
            }
            //
            // Utilities
            //
            function createTypeCheck(value, tag) {
                return tag === "undefined"
                    ? factory.createStrictEquality(value, createVoidZero())
                    : factory.createStrictEquality(createTypeOfExpression(value), createStringLiteral(tag));
            }
            function createMethodCall(object, methodName, argumentsList) {
                // Preserve the optionality of `object`.
                if (ts.isCallChain(object)) {
                    return createCallChain(createPropertyAccessChain(object, /*questionDotToken*/ undefined, methodName), 
                    /*questionDotToken*/ undefined, 
                    /*typeArguments*/ undefined, argumentsList);
                }
                return createCallExpression(createPropertyAccessExpression(object, methodName), 
                /*typeArguments*/ undefined, argumentsList);
            }
            function createFunctionBindCall(target, thisArg, argumentsList) {
                return createMethodCall(target, "bind", __spreadArray([thisArg], argumentsList, true));
            }
            function createFunctionCallCall(target, thisArg, argumentsList) {
                return createMethodCall(target, "call", __spreadArray([thisArg], argumentsList, true));
            }
            function createFunctionApplyCall(target, thisArg, argumentsExpression) {
                return createMethodCall(target, "apply", [thisArg, argumentsExpression]);
            }
            function createGlobalMethodCall(globalObjectName, methodName, argumentsList) {
                return createMethodCall(createIdentifier(globalObjectName), methodName, argumentsList);
            }
            function createArraySliceCall(array, start) {
                return createMethodCall(array, "slice", start === undefined ? [] : [asExpression(start)]);
            }
            function createArrayConcatCall(array, argumentsList) {
                return createMethodCall(array, "concat", argumentsList);
            }
            function createObjectDefinePropertyCall(target, propertyName, attributes) {
                return createGlobalMethodCall("Object", "defineProperty", [target, asExpression(propertyName), attributes]);
            }
            function createReflectGetCall(target, propertyKey, receiver) {
                return createGlobalMethodCall("Reflect", "get", receiver ? [target, propertyKey, receiver] : [target, propertyKey]);
            }
            function createReflectSetCall(target, propertyKey, value, receiver) {
                return createGlobalMethodCall("Reflect", "set", receiver ? [target, propertyKey, value, receiver] : [target, propertyKey, value]);
            }
            function tryAddPropertyAssignment(properties, propertyName, expression) {
                if (expression) {
                    properties.push(createPropertyAssignment(propertyName, expression));
                    return true;
                }
                return false;
            }
            function createPropertyDescriptor(attributes, singleLine) {
                var properties = [];
                tryAddPropertyAssignment(properties, "enumerable", asExpression(attributes.enumerable));
                tryAddPropertyAssignment(properties, "configurable", asExpression(attributes.configurable));
                var isData = tryAddPropertyAssignment(properties, "writable", asExpression(attributes.writable));
                isData = tryAddPropertyAssignment(properties, "value", attributes.value) || isData;
                var isAccessor = tryAddPropertyAssignment(properties, "get", attributes.get);
                isAccessor = tryAddPropertyAssignment(properties, "set", attributes.set) || isAccessor;
                ts.Debug.assert(!(isData && isAccessor), "A PropertyDescriptor may not be both an accessor descriptor and a data descriptor.");
                return createObjectLiteralExpression(properties, !singleLine);
            }
            function updateOuterExpression(outerExpression, expression) {
                switch (outerExpression.kind) {
                    case 210 /* ParenthesizedExpression */: return updateParenthesizedExpression(outerExpression, expression);
                    case 209 /* TypeAssertionExpression */: return updateTypeAssertion(outerExpression, outerExpression.type, expression);
                    case 227 /* AsExpression */: return updateAsExpression(outerExpression, expression, outerExpression.type);
                    case 228 /* NonNullExpression */: return updateNonNullExpression(outerExpression, expression);
                    case 345 /* PartiallyEmittedExpression */: return updatePartiallyEmittedExpression(outerExpression, expression);
                }
            }
            /**
             * Determines whether a node is a parenthesized expression that can be ignored when recreating outer expressions.
             *
             * A parenthesized expression can be ignored when all of the following are true:
             *
             * - It's `pos` and `end` are not -1
             * - It does not have a custom source map range
             * - It does not have a custom comment range
             * - It does not have synthetic leading or trailing comments
             *
             * If an outermost parenthesized expression is ignored, but the containing expression requires a parentheses around
             * the expression to maintain precedence, a new parenthesized expression should be created automatically when
             * the containing expression is created/updated.
             */
            function isIgnorableParen(node) {
                return ts.isParenthesizedExpression(node)
                    && ts.nodeIsSynthesized(node)
                    && ts.nodeIsSynthesized(ts.getSourceMapRange(node))
                    && ts.nodeIsSynthesized(ts.getCommentRange(node))
                    && !ts.some(ts.getSyntheticLeadingComments(node))
                    && !ts.some(ts.getSyntheticTrailingComments(node));
            }
            function restoreOuterExpressions(outerExpression, innerExpression, kinds) {
                if (kinds === void 0) {
                    kinds = 15 /* All */;
                }
                if (outerExpression && ts.isOuterExpression(outerExpression, kinds) && !isIgnorableParen(outerExpression)) {
                    return updateOuterExpression(outerExpression, restoreOuterExpressions(outerExpression.expression, innerExpression));
                }
                return innerExpression;
            }
            function restoreEnclosingLabel(node, outermostLabeledStatement, afterRestoreLabelCallback) {
                if (!outermostLabeledStatement) {
                    return node;
                }
                var updated = updateLabeledStatement(outermostLabeledStatement, outermostLabeledStatement.label, ts.isLabeledStatement(outermostLabeledStatement.statement)
                    ? restoreEnclosingLabel(node, outermostLabeledStatement.statement)
                    : node);
                if (afterRestoreLabelCallback) {
                    afterRestoreLabelCallback(outermostLabeledStatement);
                }
                return updated;
            }
            function shouldBeCapturedInTempVariable(node, cacheIdentifiers) {
                var target = ts.skipParentheses(node);
                switch (target.kind) {
                    case 79 /* Identifier */:
                        return cacheIdentifiers;
                    case 108 /* ThisKeyword */:
                    case 8 /* NumericLiteral */:
                    case 9 /* BigIntLiteral */:
                    case 10 /* StringLiteral */:
                        return false;
                    case 202 /* ArrayLiteralExpression */:
                        var elements = target.elements;
                        if (elements.length === 0) {
                            return false;
                        }
                        return true;
                    case 203 /* ObjectLiteralExpression */:
                        return target.properties.length > 0;
                    default:
                        return true;
                }
            }
            function createCallBinding(expression, recordTempVariable, languageVersion, cacheIdentifiers) {
                if (cacheIdentifiers === void 0) {
                    cacheIdentifiers = false;
                }
                var callee = ts.skipOuterExpressions(expression, 15 /* All */);
                var thisArg;
                var target;
                if (ts.isSuperProperty(callee)) {
                    thisArg = createThis();
                    target = callee;
                }
                else if (ts.isSuperKeyword(callee)) {
                    thisArg = createThis();
                    target = languageVersion !== undefined && languageVersion < 2 /* ES2015 */
                        ? ts.setTextRange(createIdentifier("_super"), callee)
                        : callee;
                }
                else if (ts.getEmitFlags(callee) & 4096 /* HelperName */) {
                    thisArg = createVoidZero();
                    target = parenthesizerRules().parenthesizeLeftSideOfAccess(callee);
                }
                else if (ts.isPropertyAccessExpression(callee)) {
                    if (shouldBeCapturedInTempVariable(callee.expression, cacheIdentifiers)) {
                        // for `a.b()` target is `(_a = a).b` and thisArg is `_a`
                        thisArg = createTempVariable(recordTempVariable);
                        target = createPropertyAccessExpression(ts.setTextRange(factory.createAssignment(thisArg, callee.expression), callee.expression), callee.name);
                        ts.setTextRange(target, callee);
                    }
                    else {
                        thisArg = callee.expression;
                        target = callee;
                    }
                }
                else if (ts.isElementAccessExpression(callee)) {
                    if (shouldBeCapturedInTempVariable(callee.expression, cacheIdentifiers)) {
                        // for `a[b]()` target is `(_a = a)[b]` and thisArg is `_a`
                        thisArg = createTempVariable(recordTempVariable);
                        target = createElementAccessExpression(ts.setTextRange(factory.createAssignment(thisArg, callee.expression), callee.expression), callee.argumentExpression);
                        ts.setTextRange(target, callee);
                    }
                    else {
                        thisArg = callee.expression;
                        target = callee;
                    }
                }
                else {
                    // for `a()` target is `a` and thisArg is `void 0`
                    thisArg = createVoidZero();
                    target = parenthesizerRules().parenthesizeLeftSideOfAccess(expression);
                }
                return { target: target, thisArg: thisArg };
            }
            function createAssignmentTargetWrapper(paramName, expression) {
                return createPropertyAccessExpression(
                // Explicit parens required because of v8 regression (https://bugs.chromium.org/p/v8/issues/detail?id=9560)
                createParenthesizedExpression(createObjectLiteralExpression([
                    createSetAccessorDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, "value", [createParameterDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, 
                        /*dotDotDotToken*/ undefined, paramName, 
                        /*questionToken*/ undefined, 
                        /*type*/ undefined, 
                        /*initializer*/ undefined)], createBlock([
                        createExpressionStatement(expression)
                    ]))
                ])), "value");
            }
            function inlineExpressions(expressions) {
                // Avoid deeply nested comma expressions as traversing them during emit can result in "Maximum call
                // stack size exceeded" errors.
                return expressions.length > 10
                    ? createCommaListExpression(expressions)
                    : ts.reduceLeft(expressions, factory.createComma);
            }
            function getName(node, allowComments, allowSourceMaps, emitFlags) {
                if (emitFlags === void 0) {
                    emitFlags = 0;
                }
                var nodeName = ts.getNameOfDeclaration(node);
                if (nodeName && ts.isIdentifier(nodeName) && !ts.isGeneratedIdentifier(nodeName)) {
                    // TODO(rbuckton): Does this need to be parented?
                    var name = ts.setParent(ts.setTextRange(cloneNode(nodeName), nodeName), nodeName.parent);
                    emitFlags |= ts.getEmitFlags(nodeName);
                    if (!allowSourceMaps)
                        emitFlags |= 48 /* NoSourceMap */;
                    if (!allowComments)
                        emitFlags |= 1536 /* NoComments */;
                    if (emitFlags)
                        ts.setEmitFlags(name, emitFlags);
                    return name;
                }
                return getGeneratedNameForNode(node);
            }
            /**
             * Gets the internal name of a declaration. This is primarily used for declarations that can be
             * referred to by name in the body of an ES5 class function body. An internal name will *never*
             * be prefixed with an module or namespace export modifier like "exports." when emitted as an
             * expression. An internal name will also *never* be renamed due to a collision with a block
             * scoped variable.
             *
             * @param node The declaration.
             * @param allowComments A value indicating whether comments may be emitted for the name.
             * @param allowSourceMaps A value indicating whether source maps may be emitted for the name.
             */
            function getInternalName(node, allowComments, allowSourceMaps) {
                return getName(node, allowComments, allowSourceMaps, 16384 /* LocalName */ | 32768 /* InternalName */);
            }
            /**
             * Gets the local name of a declaration. This is primarily used for declarations that can be
             * referred to by name in the declaration's immediate scope (classes, enums, namespaces). A
             * local name will *never* be prefixed with an module or namespace export modifier like
             * "exports." when emitted as an expression.
             *
             * @param node The declaration.
             * @param allowComments A value indicating whether comments may be emitted for the name.
             * @param allowSourceMaps A value indicating whether source maps may be emitted for the name.
             */
            function getLocalName(node, allowComments, allowSourceMaps) {
                return getName(node, allowComments, allowSourceMaps, 16384 /* LocalName */);
            }
            /**
             * Gets the export name of a declaration. This is primarily used for declarations that can be
             * referred to by name in the declaration's immediate scope (classes, enums, namespaces). An
             * export name will *always* be prefixed with an module or namespace export modifier like
             * `"exports."` when emitted as an expression if the name points to an exported symbol.
             *
             * @param node The declaration.
             * @param allowComments A value indicating whether comments may be emitted for the name.
             * @param allowSourceMaps A value indicating whether source maps may be emitted for the name.
             */
            function getExportName(node, allowComments, allowSourceMaps) {
                return getName(node, allowComments, allowSourceMaps, 8192 /* ExportName */);
            }
            /**
             * Gets the name of a declaration for use in declarations.
             *
             * @param node The declaration.
             * @param allowComments A value indicating whether comments may be emitted for the name.
             * @param allowSourceMaps A value indicating whether source maps may be emitted for the name.
             */
            function getDeclarationName(node, allowComments, allowSourceMaps) {
                return getName(node, allowComments, allowSourceMaps);
            }
            /**
             * Gets a namespace-qualified name for use in expressions.
             *
             * @param ns The namespace identifier.
             * @param name The name.
             * @param allowComments A value indicating whether comments may be emitted for the name.
             * @param allowSourceMaps A value indicating whether source maps may be emitted for the name.
             */
            function getNamespaceMemberName(ns, name, allowComments, allowSourceMaps) {
                var qualifiedName = createPropertyAccessExpression(ns, ts.nodeIsSynthesized(name) ? name : cloneNode(name));
                ts.setTextRange(qualifiedName, name);
                var emitFlags = 0;
                if (!allowSourceMaps)
                    emitFlags |= 48 /* NoSourceMap */;
                if (!allowComments)
                    emitFlags |= 1536 /* NoComments */;
                if (emitFlags)
                    ts.setEmitFlags(qualifiedName, emitFlags);
                return qualifiedName;
            }
            /**
             * Gets the exported name of a declaration for use in expressions.
             *
             * An exported name will *always* be prefixed with an module or namespace export modifier like
             * "exports." if the name points to an exported symbol.
             *
             * @param ns The namespace identifier.
             * @param node The declaration.
             * @param allowComments A value indicating whether comments may be emitted for the name.
             * @param allowSourceMaps A value indicating whether source maps may be emitted for the name.
             */
            function getExternalModuleOrNamespaceExportName(ns, node, allowComments, allowSourceMaps) {
                if (ns && ts.hasSyntacticModifier(node, 1 /* Export */)) {
                    return getNamespaceMemberName(ns, getName(node), allowComments, allowSourceMaps);
                }
                return getExportName(node, allowComments, allowSourceMaps);
            }
            /**
             * Copies any necessary standard and custom prologue-directives into target array.
             * @param source origin statements array
             * @param target result statements array
             * @param ensureUseStrict boolean determining whether the function need to add prologue-directives
             * @param visitor Optional callback used to visit any custom prologue directives.
             */
            function copyPrologue(source, target, ensureUseStrict, visitor) {
                var offset = copyStandardPrologue(source, target, ensureUseStrict);
                return copyCustomPrologue(source, target, offset, visitor);
            }
            function isUseStrictPrologue(node) {
                return ts.isStringLiteral(node.expression) && node.expression.text === "use strict";
            }
            function createUseStrictPrologue() {
                return ts.startOnNewLine(createExpressionStatement(createStringLiteral("use strict")));
            }
            /**
             * Copies only the standard (string-expression) prologue-directives into the target statement-array.
             * @param source origin statements array
             * @param target result statements array
             * @param ensureUseStrict boolean determining whether the function need to add prologue-directives
             */
            function copyStandardPrologue(source, target, ensureUseStrict) {
                ts.Debug.assert(target.length === 0, "Prologue directives should be at the first statement in the target statements array");
                var foundUseStrict = false;
                var statementOffset = 0;
                var numStatements = source.length;
                while (statementOffset < numStatements) {
                    var statement = source[statementOffset];
                    if (ts.isPrologueDirective(statement)) {
                        if (isUseStrictPrologue(statement)) {
                            foundUseStrict = true;
                        }
                        target.push(statement);
                    }
                    else {
                        break;
                    }
                    statementOffset++;
                }
                if (ensureUseStrict && !foundUseStrict) {
                    target.push(createUseStrictPrologue());
                }
                return statementOffset;
            }
            function copyCustomPrologue(source, target, statementOffset, visitor, filter) {
                if (filter === void 0) {
                    filter = ts.returnTrue;
                }
                var numStatements = source.length;
                while (statementOffset !== undefined && statementOffset < numStatements) {
                    var statement = source[statementOffset];
                    if (ts.getEmitFlags(statement) & 1048576 /* CustomPrologue */ && filter(statement)) {
                        ts.append(target, visitor ? ts.visitNode(statement, visitor, ts.isStatement) : statement);
                    }
                    else {
                        break;
                    }
                    statementOffset++;
                }
                return statementOffset;
            }
            /**
             * Ensures "use strict" directive is added
             *
             * @param statements An array of statements
             */
            function ensureUseStrict(statements) {
                var foundUseStrict = ts.findUseStrictPrologue(statements);
                if (!foundUseStrict) {
                    return ts.setTextRange(createNodeArray(__spreadArray([createUseStrictPrologue()], statements, true)), statements);
                }
                return statements;
            }
            /**
             * Lifts a NodeArray containing only Statement nodes to a block.
             *
             * @param nodes The NodeArray.
             */
            function liftToBlock(nodes) {
                ts.Debug.assert(ts.every(nodes, ts.isStatementOrBlock), "Cannot lift nodes to a Block.");
                return ts.singleOrUndefined(nodes) || createBlock(nodes);
            }
            function findSpanEnd(array, test, start) {
                var i = start;
                while (i < array.length && test(array[i])) {
                    i++;
                }
                return i;
            }
            function mergeLexicalEnvironment(statements, declarations) {
                if (!ts.some(declarations)) {
                    return statements;
                }
                // When we merge new lexical statements into an existing statement list, we merge them in the following manner:
                //
                // Given:
                //
                // | Left                               | Right                               |
                // |------------------------------------|-------------------------------------|
                // | [standard prologues (left)]        | [standard prologues (right)]        |
                // | [hoisted functions (left)]         | [hoisted functions (right)]         |
                // | [hoisted variables (left)]         | [hoisted variables (right)]         |
                // | [lexical init statements (left)]   | [lexical init statements (right)]   |
                // | [other statements (left)]          |                                     |
                //
                // The resulting statement list will be:
                //
                // | Result                              |
                // |-------------------------------------|
                // | [standard prologues (right)]        |
                // | [standard prologues (left)]         |
                // | [hoisted functions (right)]         |
                // | [hoisted functions (left)]          |
                // | [hoisted variables (right)]         |
                // | [hoisted variables (left)]          |
                // | [lexical init statements (right)]   |
                // | [lexical init statements (left)]    |
                // | [other statements (left)]           |
                //
                // NOTE: It is expected that new lexical init statements must be evaluated before existing lexical init statements,
                // as the prior transformation may depend on the evaluation of the lexical init statements to be in the correct state.
                // find standard prologues on left in the following order: standard directives, hoisted functions, hoisted variables, other custom
                var leftStandardPrologueEnd = findSpanEnd(statements, ts.isPrologueDirective, 0);
                var leftHoistedFunctionsEnd = findSpanEnd(statements, ts.isHoistedFunction, leftStandardPrologueEnd);
                var leftHoistedVariablesEnd = findSpanEnd(statements, ts.isHoistedVariableStatement, leftHoistedFunctionsEnd);
                // find standard prologues on right in the following order: standard directives, hoisted functions, hoisted variables, other custom
                var rightStandardPrologueEnd = findSpanEnd(declarations, ts.isPrologueDirective, 0);
                var rightHoistedFunctionsEnd = findSpanEnd(declarations, ts.isHoistedFunction, rightStandardPrologueEnd);
                var rightHoistedVariablesEnd = findSpanEnd(declarations, ts.isHoistedVariableStatement, rightHoistedFunctionsEnd);
                var rightCustomPrologueEnd = findSpanEnd(declarations, ts.isCustomPrologue, rightHoistedVariablesEnd);
                ts.Debug.assert(rightCustomPrologueEnd === declarations.length, "Expected declarations to be valid standard or custom prologues");
                // splice prologues from the right into the left. We do this in reverse order
                // so that we don't need to recompute the index on the left when we insert items.
                var left = ts.isNodeArray(statements) ? statements.slice() : statements;
                // splice other custom prologues from right into left
                if (rightCustomPrologueEnd > rightHoistedVariablesEnd) {
                    left.splice.apply(left, __spreadArray([leftHoistedVariablesEnd, 0], declarations.slice(rightHoistedVariablesEnd, rightCustomPrologueEnd), false));
                }
                // splice hoisted variables from right into left
                if (rightHoistedVariablesEnd > rightHoistedFunctionsEnd) {
                    left.splice.apply(left, __spreadArray([leftHoistedFunctionsEnd, 0], declarations.slice(rightHoistedFunctionsEnd, rightHoistedVariablesEnd), false));
                }
                // splice hoisted functions from right into left
                if (rightHoistedFunctionsEnd > rightStandardPrologueEnd) {
                    left.splice.apply(left, __spreadArray([leftStandardPrologueEnd, 0], declarations.slice(rightStandardPrologueEnd, rightHoistedFunctionsEnd), false));
                }
                // splice standard prologues from right into left (that are not already in left)
                if (rightStandardPrologueEnd > 0) {
                    if (leftStandardPrologueEnd === 0) {
                        left.splice.apply(left, __spreadArray([0, 0], declarations.slice(0, rightStandardPrologueEnd), false));
                    }
                    else {
                        var leftPrologues = new ts.Map();
                        for (var i = 0; i < leftStandardPrologueEnd; i++) {
                            var leftPrologue = statements[i];
                            leftPrologues.set(leftPrologue.expression.text, true);
                        }
                        for (var i = rightStandardPrologueEnd - 1; i >= 0; i--) {
                            var rightPrologue = declarations[i];
                            if (!leftPrologues.has(rightPrologue.expression.text)) {
                                left.unshift(rightPrologue);
                            }
                        }
                    }
                }
                if (ts.isNodeArray(statements)) {
                    return ts.setTextRange(createNodeArray(left, statements.hasTrailingComma), statements);
                }
                return statements;
            }
            function updateModifiers(node, modifiers) {
                var _a;
                if (typeof modifiers === "number") {
                    modifiers = createModifiersFromModifierFlags(modifiers);
                }
                return ts.isParameter(node) ? updateParameterDeclaration(node, node.decorators, modifiers, node.dotDotDotToken, node.name, node.questionToken, node.type, node.initializer) :
                    ts.isPropertySignature(node) ? updatePropertySignature(node, modifiers, node.name, node.questionToken, node.type) :
                        ts.isPropertyDeclaration(node) ? updatePropertyDeclaration(node, node.decorators, modifiers, node.name, (_a = node.questionToken) !== null && _a !== void 0 ? _a : node.exclamationToken, node.type, node.initializer) :
                            ts.isMethodSignature(node) ? updateMethodSignature(node, modifiers, node.name, node.questionToken, node.typeParameters, node.parameters, node.type) :
                                ts.isMethodDeclaration(node) ? updateMethodDeclaration(node, node.decorators, modifiers, node.asteriskToken, node.name, node.questionToken, node.typeParameters, node.parameters, node.type, node.body) :
                                    ts.isConstructorDeclaration(node) ? updateConstructorDeclaration(node, node.decorators, modifiers, node.parameters, node.body) :
                                        ts.isGetAccessorDeclaration(node) ? updateGetAccessorDeclaration(node, node.decorators, modifiers, node.name, node.parameters, node.type, node.body) :
                                            ts.isSetAccessorDeclaration(node) ? updateSetAccessorDeclaration(node, node.decorators, modifiers, node.name, node.parameters, node.body) :
                                                ts.isIndexSignatureDeclaration(node) ? updateIndexSignature(node, node.decorators, modifiers, node.parameters, node.type) :
                                                    ts.isFunctionExpression(node) ? updateFunctionExpression(node, modifiers, node.asteriskToken, node.name, node.typeParameters, node.parameters, node.type, node.body) :
                                                        ts.isArrowFunction(node) ? updateArrowFunction(node, modifiers, node.typeParameters, node.parameters, node.type, node.equalsGreaterThanToken, node.body) :
                                                            ts.isClassExpression(node) ? updateClassExpression(node, node.decorators, modifiers, node.name, node.typeParameters, node.heritageClauses, node.members) :
                                                                ts.isVariableStatement(node) ? updateVariableStatement(node, modifiers, node.declarationList) :
                                                                    ts.isFunctionDeclaration(node) ? updateFunctionDeclaration(node, node.decorators, modifiers, node.asteriskToken, node.name, node.typeParameters, node.parameters, node.type, node.body) :
                                                                        ts.isClassDeclaration(node) ? updateClassDeclaration(node, node.decorators, modifiers, node.name, node.typeParameters, node.heritageClauses, node.members) :
                                                                            ts.isInterfaceDeclaration(node) ? updateInterfaceDeclaration(node, node.decorators, modifiers, node.name, node.typeParameters, node.heritageClauses, node.members) :
                                                                                ts.isTypeAliasDeclaration(node) ? updateTypeAliasDeclaration(node, node.decorators, modifiers, node.name, node.typeParameters, node.type) :
                                                                                    ts.isEnumDeclaration(node) ? updateEnumDeclaration(node, node.decorators, modifiers, node.name, node.members) :
                                                                                        ts.isModuleDeclaration(node) ? updateModuleDeclaration(node, node.decorators, modifiers, node.name, node.body) :
                                                                                            ts.isImportEqualsDeclaration(node) ? updateImportEqualsDeclaration(node, node.decorators, modifiers, node.isTypeOnly, node.name, node.moduleReference) :
                                                                                                ts.isImportDeclaration(node) ? updateImportDeclaration(node, node.decorators, modifiers, node.importClause, node.moduleSpecifier) :
                                                                                                    ts.isExportAssignment(node) ? updateExportAssignment(node, node.decorators, modifiers, node.expression) :
                                                                                                        ts.isExportDeclaration(node) ? updateExportDeclaration(node, node.decorators, modifiers, node.isTypeOnly, node.exportClause, node.moduleSpecifier) :
                                                                                                            ts.Debug.assertNever(node);
            }
            function asNodeArray(array) {
                return array ? createNodeArray(array) : undefined;
            }
            function asName(name) {
                return typeof name === "string" ? createIdentifier(name) :
                    name;
            }
            function asExpression(value) {
                return typeof value === "string" ? createStringLiteral(value) :
                    typeof value === "number" ? createNumericLiteral(value) :
                        typeof value === "boolean" ? value ? createTrue() : createFalse() :
                            value;
            }
            function asToken(value) {
                return typeof value === "number" ? createToken(value) : value;
            }
            function asEmbeddedStatement(statement) {
                return statement && ts.isNotEmittedStatement(statement) ? ts.setTextRange(setOriginalNode(createEmptyStatement(), statement), statement) : statement;
            }
        }
        ts.createNodeFactory = createNodeFactory;
        function updateWithoutOriginal(updated, original) {
            if (updated !== original) {
                ts.setTextRange(updated, original);
            }
            return updated;
        }
        function updateWithOriginal(updated, original) {
            if (updated !== original) {
                setOriginalNode(updated, original);
                ts.setTextRange(updated, original);
            }
            return updated;
        }
        function getDefaultTagNameForKind(kind) {
            switch (kind) {
                case 338 /* JSDocTypeTag */: return "type";
                case 336 /* JSDocReturnTag */: return "returns";
                case 337 /* JSDocThisTag */: return "this";
                case 334 /* JSDocEnumTag */: return "enum";
                case 325 /* JSDocAuthorTag */: return "author";
                case 327 /* JSDocClassTag */: return "class";
                case 328 /* JSDocPublicTag */: return "public";
                case 329 /* JSDocPrivateTag */: return "private";
                case 330 /* JSDocProtectedTag */: return "protected";
                case 331 /* JSDocReadonlyTag */: return "readonly";
                case 332 /* JSDocOverrideTag */: return "override";
                case 339 /* JSDocTemplateTag */: return "template";
                case 340 /* JSDocTypedefTag */: return "typedef";
                case 335 /* JSDocParameterTag */: return "param";
                case 342 /* JSDocPropertyTag */: return "prop";
                case 333 /* JSDocCallbackTag */: return "callback";
                case 323 /* JSDocAugmentsTag */: return "augments";
                case 324 /* JSDocImplementsTag */: return "implements";
                default:
                    return ts.Debug.fail("Unsupported kind: " + ts.Debug.formatSyntaxKind(kind));
            }
        }
        var rawTextScanner;
        var invalidValueSentinel = {};
        function getCookedText(kind, rawText) {
            if (!rawTextScanner) {
                rawTextScanner = ts.createScanner(99 /* Latest */, /*skipTrivia*/ false, 0 /* Standard */);
            }
            switch (kind) {
                case 14 /* NoSubstitutionTemplateLiteral */:
                    rawTextScanner.setText("`" + rawText + "`");
                    break;
                case 15 /* TemplateHead */:
                    // tslint:disable-next-line no-invalid-template-strings
                    rawTextScanner.setText("`" + rawText + "${");
                    break;
                case 16 /* TemplateMiddle */:
                    // tslint:disable-next-line no-invalid-template-strings
                    rawTextScanner.setText("}" + rawText + "${");
                    break;
                case 17 /* TemplateTail */:
                    rawTextScanner.setText("}" + rawText + "`");
                    break;
            }
            var token = rawTextScanner.scan();
            if (token === 19 /* CloseBraceToken */) {
                token = rawTextScanner.reScanTemplateToken(/*isTaggedTemplate*/ false);
            }
            if (rawTextScanner.isUnterminated()) {
                rawTextScanner.setText(undefined);
                return invalidValueSentinel;
            }
            var tokenValue;
            switch (token) {
                case 14 /* NoSubstitutionTemplateLiteral */:
                case 15 /* TemplateHead */:
                case 16 /* TemplateMiddle */:
                case 17 /* TemplateTail */:
                    tokenValue = rawTextScanner.getTokenValue();
                    break;
            }
            if (tokenValue === undefined || rawTextScanner.scan() !== 1 /* EndOfFileToken */) {
                rawTextScanner.setText(undefined);
                return invalidValueSentinel;
            }
            rawTextScanner.setText(undefined);
            return tokenValue;
        }
        function propagateIdentifierNameFlags(node) {
            // An IdentifierName is allowed to be `await`
            return propagateChildFlags(node) & ~16777216 /* ContainsPossibleTopLevelAwait */;
        }
        function propagatePropertyNameFlagsOfChild(node, transformFlags) {
            return transformFlags | (node.transformFlags & 33562624 /* PropertyNamePropagatingFlags */);
        }
        function propagateChildFlags(child) {
            if (!child)
                return 0 /* None */;
            var childFlags = child.transformFlags & ~getTransformFlagsSubtreeExclusions(child.kind);
            return ts.isNamedDeclaration(child) && ts.isPropertyName(child.name) ? propagatePropertyNameFlagsOfChild(child.name, childFlags) : childFlags;
        }
        function propagateChildrenFlags(children) {
            return children ? children.transformFlags : 0 /* None */;
        }
        function aggregateChildrenFlags(children) {
            var subtreeFlags = 0 /* None */;
            for (var _i = 0, children_2 = children; _i < children_2.length; _i++) {
                var child = children_2[_i];
                subtreeFlags |= propagateChildFlags(child);
            }
            children.transformFlags = subtreeFlags;
        }
        /**
         * Gets the transform flags to exclude when unioning the transform flags of a subtree.
         */
        /* @internal */
        function getTransformFlagsSubtreeExclusions(kind) {
            if (kind >= 175 /* FirstTypeNode */ && kind <= 198 /* LastTypeNode */) {
                return -2 /* TypeExcludes */;
            }
            switch (kind) {
                case 206 /* CallExpression */:
                case 207 /* NewExpression */:
                case 202 /* ArrayLiteralExpression */:
                    return 536887296 /* ArrayLiteralOrCallOrNewExcludes */;
                case 259 /* ModuleDeclaration */:
                    return 589443072 /* ModuleExcludes */;
                case 162 /* Parameter */:
                    return 536870912 /* ParameterExcludes */;
                case 212 /* ArrowFunction */:
                    return 557748224 /* ArrowFunctionExcludes */;
                case 211 /* FunctionExpression */:
                case 254 /* FunctionDeclaration */:
                    return 591310848 /* FunctionExcludes */;
                case 253 /* VariableDeclarationList */:
                    return 537165824 /* VariableDeclarationListExcludes */;
                case 255 /* ClassDeclaration */:
                case 224 /* ClassExpression */:
                    return 536940544 /* ClassExcludes */;
                case 169 /* Constructor */:
                    return 591306752 /* ConstructorExcludes */;
                case 165 /* PropertyDeclaration */:
                    return 570433536 /* PropertyExcludes */;
                case 167 /* MethodDeclaration */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                    return 574529536 /* MethodOrAccessorExcludes */;
                case 129 /* AnyKeyword */:
                case 145 /* NumberKeyword */:
                case 156 /* BigIntKeyword */:
                case 142 /* NeverKeyword */:
                case 148 /* StringKeyword */:
                case 146 /* ObjectKeyword */:
                case 132 /* BooleanKeyword */:
                case 149 /* SymbolKeyword */:
                case 114 /* VoidKeyword */:
                case 161 /* TypeParameter */:
                case 164 /* PropertySignature */:
                case 166 /* MethodSignature */:
                case 172 /* CallSignature */:
                case 173 /* ConstructSignature */:
                case 174 /* IndexSignature */:
                case 256 /* InterfaceDeclaration */:
                case 257 /* TypeAliasDeclaration */:
                    return -2 /* TypeExcludes */;
                case 203 /* ObjectLiteralExpression */:
                    return 536973312 /* ObjectLiteralExcludes */;
                case 290 /* CatchClause */:
                    return 536903680 /* CatchClauseExcludes */;
                case 199 /* ObjectBindingPattern */:
                case 200 /* ArrayBindingPattern */:
                    return 536887296 /* BindingPatternExcludes */;
                case 209 /* TypeAssertionExpression */:
                case 227 /* AsExpression */:
                case 345 /* PartiallyEmittedExpression */:
                case 210 /* ParenthesizedExpression */:
                case 106 /* SuperKeyword */:
                    return 536870912 /* OuterExpressionExcludes */;
                case 204 /* PropertyAccessExpression */:
                case 205 /* ElementAccessExpression */:
                    return 536870912 /* PropertyAccessExcludes */;
                default:
                    return 536870912 /* NodeExcludes */;
            }
        }
        ts.getTransformFlagsSubtreeExclusions = getTransformFlagsSubtreeExclusions;
        var baseFactory = ts.createBaseNodeFactory();
        function makeSynthetic(node) {
            node.flags |= 8 /* Synthesized */;
            return node;
        }
        var syntheticFactory = {
            createBaseSourceFileNode: function (kind) { return makeSynthetic(baseFactory.createBaseSourceFileNode(kind)); },
            createBaseIdentifierNode: function (kind) { return makeSynthetic(baseFactory.createBaseIdentifierNode(kind)); },
            createBasePrivateIdentifierNode: function (kind) { return makeSynthetic(baseFactory.createBasePrivateIdentifierNode(kind)); },
            createBaseTokenNode: function (kind) { return makeSynthetic(baseFactory.createBaseTokenNode(kind)); },
            createBaseNode: function (kind) { return makeSynthetic(baseFactory.createBaseNode(kind)); },
        };
        ts.factory = createNodeFactory(4 /* NoIndentationOnFreshPropertyAccess */, syntheticFactory);
        function createUnparsedSourceFile(textOrInputFiles, mapPathOrType, mapTextOrStripInternal) {
            var stripInternal;
            var bundleFileInfo;
            var fileName;
            var text;
            var length;
            var sourceMapPath;
            var sourceMapText;
            var getText;
            var getSourceMapText;
            var oldFileOfCurrentEmit;
            if (!ts.isString(textOrInputFiles)) {
                ts.Debug.assert(mapPathOrType === "js" || mapPathOrType === "dts");
                fileName = (mapPathOrType === "js" ? textOrInputFiles.javascriptPath : textOrInputFiles.declarationPath) || "";
                sourceMapPath = mapPathOrType === "js" ? textOrInputFiles.javascriptMapPath : textOrInputFiles.declarationMapPath;
                getText = function () { return mapPathOrType === "js" ? textOrInputFiles.javascriptText : textOrInputFiles.declarationText; };
                getSourceMapText = function () { return mapPathOrType === "js" ? textOrInputFiles.javascriptMapText : textOrInputFiles.declarationMapText; };
                length = function () { return getText().length; };
                if (textOrInputFiles.buildInfo && textOrInputFiles.buildInfo.bundle) {
                    ts.Debug.assert(mapTextOrStripInternal === undefined || typeof mapTextOrStripInternal === "boolean");
                    stripInternal = mapTextOrStripInternal;
                    bundleFileInfo = mapPathOrType === "js" ? textOrInputFiles.buildInfo.bundle.js : textOrInputFiles.buildInfo.bundle.dts;
                    oldFileOfCurrentEmit = textOrInputFiles.oldFileOfCurrentEmit;
                }
            }
            else {
                fileName = "";
                text = textOrInputFiles;
                length = textOrInputFiles.length;
                sourceMapPath = mapPathOrType;
                sourceMapText = mapTextOrStripInternal;
            }
            var node = oldFileOfCurrentEmit ?
                parseOldFileOfCurrentEmit(ts.Debug.assertDefined(bundleFileInfo)) :
                parseUnparsedSourceFile(bundleFileInfo, stripInternal, length);
            node.fileName = fileName;
            node.sourceMapPath = sourceMapPath;
            node.oldFileOfCurrentEmit = oldFileOfCurrentEmit;
            if (getText && getSourceMapText) {
                Object.defineProperty(node, "text", { get: getText });
                Object.defineProperty(node, "sourceMapText", { get: getSourceMapText });
            }
            else {
                ts.Debug.assert(!oldFileOfCurrentEmit);
                node.text = text !== null && text !== void 0 ? text : "";
                node.sourceMapText = sourceMapText;
            }
            return node;
        }
        ts.createUnparsedSourceFile = createUnparsedSourceFile;
        function parseUnparsedSourceFile(bundleFileInfo, stripInternal, length) {
            var prologues;
            var helpers;
            var referencedFiles;
            var typeReferenceDirectives;
            var libReferenceDirectives;
            var prependChildren;
            var texts;
            var hasNoDefaultLib;
            for (var _i = 0, _a = bundleFileInfo ? bundleFileInfo.sections : ts.emptyArray; _i < _a.length; _i++) {
                var section = _a[_i];
                switch (section.kind) {
                    case "prologue" /* Prologue */:
                        prologues = ts.append(prologues, ts.setTextRange(ts.factory.createUnparsedPrologue(section.data), section));
                        break;
                    case "emitHelpers" /* EmitHelpers */:
                        helpers = ts.append(helpers, ts.getAllUnscopedEmitHelpers().get(section.data));
                        break;
                    case "no-default-lib" /* NoDefaultLib */:
                        hasNoDefaultLib = true;
                        break;
                    case "reference" /* Reference */:
                        referencedFiles = ts.append(referencedFiles, { pos: -1, end: -1, fileName: section.data });
                        break;
                    case "type" /* Type */:
                        typeReferenceDirectives = ts.append(typeReferenceDirectives, section.data);
                        break;
                    case "lib" /* Lib */:
                        libReferenceDirectives = ts.append(libReferenceDirectives, { pos: -1, end: -1, fileName: section.data });
                        break;
                    case "prepend" /* Prepend */:
                        var prependTexts = void 0;
                        for (var _b = 0, _c = section.texts; _b < _c.length; _b++) {
                            var text = _c[_b];
                            if (!stripInternal || text.kind !== "internal" /* Internal */) {
                                prependTexts = ts.append(prependTexts, ts.setTextRange(ts.factory.createUnparsedTextLike(text.data, text.kind === "internal" /* Internal */), text));
                            }
                        }
                        prependChildren = ts.addRange(prependChildren, prependTexts);
                        texts = ts.append(texts, ts.factory.createUnparsedPrepend(section.data, prependTexts !== null && prependTexts !== void 0 ? prependTexts : ts.emptyArray));
                        break;
                    case "internal" /* Internal */:
                        if (stripInternal) {
                            if (!texts)
                                texts = [];
                            break;
                        }
                    // falls through
                    case "text" /* Text */:
                        texts = ts.append(texts, ts.setTextRange(ts.factory.createUnparsedTextLike(section.data, section.kind === "internal" /* Internal */), section));
                        break;
                    default:
                        ts.Debug.assertNever(section);
                }
            }
            if (!texts) {
                var textNode = ts.factory.createUnparsedTextLike(/*data*/ undefined, /*internal*/ false);
                ts.setTextRangePosWidth(textNode, 0, typeof length === "function" ? length() : length);
                texts = [textNode];
            }
            var node = ts.parseNodeFactory.createUnparsedSource(prologues !== null && prologues !== void 0 ? prologues : ts.emptyArray, /*syntheticReferences*/ undefined, texts);
            ts.setEachParent(prologues, node);
            ts.setEachParent(texts, node);
            ts.setEachParent(prependChildren, node);
            node.hasNoDefaultLib = hasNoDefaultLib;
            node.helpers = helpers;
            node.referencedFiles = referencedFiles || ts.emptyArray;
            node.typeReferenceDirectives = typeReferenceDirectives;
            node.libReferenceDirectives = libReferenceDirectives || ts.emptyArray;
            return node;
        }
        function parseOldFileOfCurrentEmit(bundleFileInfo) {
            var texts;
            var syntheticReferences;
            for (var _i = 0, _a = bundleFileInfo.sections; _i < _a.length; _i++) {
                var section = _a[_i];
                switch (section.kind) {
                    case "internal" /* Internal */:
                    case "text" /* Text */:
                        texts = ts.append(texts, ts.setTextRange(ts.factory.createUnparsedTextLike(section.data, section.kind === "internal" /* Internal */), section));
                        break;
                    case "no-default-lib" /* NoDefaultLib */:
                    case "reference" /* Reference */:
                    case "type" /* Type */:
                    case "lib" /* Lib */:
                        syntheticReferences = ts.append(syntheticReferences, ts.setTextRange(ts.factory.createUnparsedSyntheticReference(section), section));
                        break;
                    // Ignore
                    case "prologue" /* Prologue */:
                    case "emitHelpers" /* EmitHelpers */:
                    case "prepend" /* Prepend */:
                        break;
                    default:
                        ts.Debug.assertNever(section);
                }
            }
            var node = ts.factory.createUnparsedSource(ts.emptyArray, syntheticReferences, texts !== null && texts !== void 0 ? texts : ts.emptyArray);
            ts.setEachParent(syntheticReferences, node);
            ts.setEachParent(texts, node);
            node.helpers = ts.map(bundleFileInfo.sources && bundleFileInfo.sources.helpers, function (name) { return ts.getAllUnscopedEmitHelpers().get(name); });
            return node;
        }
        function createInputFiles(javascriptTextOrReadFileText, declarationTextOrJavascriptPath, javascriptMapPath, javascriptMapTextOrDeclarationPath, declarationMapPath, declarationMapTextOrBuildInfoPath, javascriptPath, declarationPath, buildInfoPath, buildInfo, oldFileOfCurrentEmit) {
            var node = ts.parseNodeFactory.createInputFiles();
            if (!ts.isString(javascriptTextOrReadFileText)) {
                var cache_1 = new ts.Map();
                var textGetter_1 = function (path) {
                    if (path === undefined)
                        return undefined;
                    var value = cache_1.get(path);
                    if (value === undefined) {
                        value = javascriptTextOrReadFileText(path);
                        cache_1.set(path, value !== undefined ? value : false);
                    }
                    return value !== false ? value : undefined;
                };
                var definedTextGetter_1 = function (path) {
                    var result = textGetter_1(path);
                    return result !== undefined ? result : "/* Input file " + path + " was missing */\r\n";
                };
                var buildInfo_1;
                var getAndCacheBuildInfo_1 = function (getText) {
                    if (buildInfo_1 === undefined) {
                        var result = getText();
                        buildInfo_1 = result !== undefined ? ts.getBuildInfo(result) : false;
                    }
                    return buildInfo_1 || undefined;
                };
                node.javascriptPath = declarationTextOrJavascriptPath;
                node.javascriptMapPath = javascriptMapPath;
                node.declarationPath = ts.Debug.assertDefined(javascriptMapTextOrDeclarationPath);
                node.declarationMapPath = declarationMapPath;
                node.buildInfoPath = declarationMapTextOrBuildInfoPath;
                Object.defineProperties(node, {
                    javascriptText: { get: function () { return definedTextGetter_1(declarationTextOrJavascriptPath); } },
                    javascriptMapText: { get: function () { return textGetter_1(javascriptMapPath); } },
                    declarationText: { get: function () { return definedTextGetter_1(ts.Debug.assertDefined(javascriptMapTextOrDeclarationPath)); } },
                    declarationMapText: { get: function () { return textGetter_1(declarationMapPath); } },
                    buildInfo: { get: function () { return getAndCacheBuildInfo_1(function () { return textGetter_1(declarationMapTextOrBuildInfoPath); }); } }
                });
            }
            else {
                node.javascriptText = javascriptTextOrReadFileText;
                node.javascriptMapPath = javascriptMapPath;
                node.javascriptMapText = javascriptMapTextOrDeclarationPath;
                node.declarationText = declarationTextOrJavascriptPath;
                node.declarationMapPath = declarationMapPath;
                node.declarationMapText = declarationMapTextOrBuildInfoPath;
                node.javascriptPath = javascriptPath;
                node.declarationPath = declarationPath;
                node.buildInfoPath = buildInfoPath;
                node.buildInfo = buildInfo;
                node.oldFileOfCurrentEmit = oldFileOfCurrentEmit;
            }
            return node;
        }
        ts.createInputFiles = createInputFiles;
        // tslint:disable-next-line variable-name
        var SourceMapSource;
        /**
         * Create an external source map source file reference
         */
        function createSourceMapSource(fileName, text, skipTrivia) {
            return new (SourceMapSource || (SourceMapSource = ts.objectAllocator.getSourceMapSourceConstructor()))(fileName, text, skipTrivia);
        }
        ts.createSourceMapSource = createSourceMapSource;
        // Utilities
        function setOriginalNode(node, original) {
            node.original = original;
            if (original) {
                var emitNode = original.emitNode;
                if (emitNode)
                    node.emitNode = mergeEmitNode(emitNode, node.emitNode);
            }
            return node;
        }
        ts.setOriginalNode = setOriginalNode;
        function mergeEmitNode(sourceEmitNode, destEmitNode) {
            var flags = sourceEmitNode.flags, leadingComments = sourceEmitNode.leadingComments, trailingComments = sourceEmitNode.trailingComments, commentRange = sourceEmitNode.commentRange, sourceMapRange = sourceEmitNode.sourceMapRange, tokenSourceMapRanges = sourceEmitNode.tokenSourceMapRanges, constantValue = sourceEmitNode.constantValue, helpers = sourceEmitNode.helpers, startsOnNewLine = sourceEmitNode.startsOnNewLine;
            if (!destEmitNode)
                destEmitNode = {};
            // We are using `.slice()` here in case `destEmitNode.leadingComments` is pushed to later.
            if (leadingComments)
                destEmitNode.leadingComments = ts.addRange(leadingComments.slice(), destEmitNode.leadingComments);
            if (trailingComments)
                destEmitNode.trailingComments = ts.addRange(trailingComments.slice(), destEmitNode.trailingComments);
            if (flags)
                destEmitNode.flags = flags & ~268435456 /* Immutable */;
            if (commentRange)
                destEmitNode.commentRange = commentRange;
            if (sourceMapRange)
                destEmitNode.sourceMapRange = sourceMapRange;
            if (tokenSourceMapRanges)
                destEmitNode.tokenSourceMapRanges = mergeTokenSourceMapRanges(tokenSourceMapRanges, destEmitNode.tokenSourceMapRanges);
            if (constantValue !== undefined)
                destEmitNode.constantValue = constantValue;
            if (helpers) {
                for (var _i = 0, helpers_1 = helpers; _i < helpers_1.length; _i++) {
                    var helper = helpers_1[_i];
                    destEmitNode.helpers = ts.appendIfUnique(destEmitNode.helpers, helper);
                }
            }
            if (startsOnNewLine !== undefined)
                destEmitNode.startsOnNewLine = startsOnNewLine;
            return destEmitNode;
        }
        function mergeTokenSourceMapRanges(sourceRanges, destRanges) {
            if (!destRanges)
                destRanges = [];
            for (var key in sourceRanges) {
                destRanges[key] = sourceRanges[key];
            }
            return destRanges;
        }
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        /**
         * Associates a node with the current transformation, initializing
         * various transient transformation properties.
         * @internal
         */
        function getOrCreateEmitNode(node) {
            var _a;
            if (!node.emitNode) {
                if (ts.isParseTreeNode(node)) {
                    // To avoid holding onto transformation artifacts, we keep track of any
                    // parse tree node we are annotating. This allows us to clean them up after
                    // all transformations have completed.
                    if (node.kind === 300 /* SourceFile */) {
                        return node.emitNode = { annotatedNodes: [node] };
                    }
                    var sourceFile = (_a = ts.getSourceFileOfNode(ts.getParseTreeNode(ts.getSourceFileOfNode(node)))) !== null && _a !== void 0 ? _a : ts.Debug.fail("Could not determine parsed source file.");
                    getOrCreateEmitNode(sourceFile).annotatedNodes.push(node);
                }
                node.emitNode = {};
            }
            else {
                ts.Debug.assert(!(node.emitNode.flags & 268435456 /* Immutable */), "Invalid attempt to mutate an immutable node.");
            }
            return node.emitNode;
        }
        ts.getOrCreateEmitNode = getOrCreateEmitNode;
        /**
         * Clears any `EmitNode` entries from parse-tree nodes.
         * @param sourceFile A source file.
         */
        function disposeEmitNodes(sourceFile) {
            var _a, _b;
            // During transformation we may need to annotate a parse tree node with transient
            // transformation properties. As parse tree nodes live longer than transformation
            // nodes, we need to make sure we reclaim any memory allocated for custom ranges
            // from these nodes to ensure we do not hold onto entire subtrees just for position
            // information. We also need to reset these nodes to a pre-transformation state
            // for incremental parsing scenarios so that we do not impact later emit.
            var annotatedNodes = (_b = (_a = ts.getSourceFileOfNode(ts.getParseTreeNode(sourceFile))) === null || _a === void 0 ? void 0 : _a.emitNode) === null || _b === void 0 ? void 0 : _b.annotatedNodes;
            if (annotatedNodes) {
                for (var _i = 0, annotatedNodes_1 = annotatedNodes; _i < annotatedNodes_1.length; _i++) {
                    var node = annotatedNodes_1[_i];
                    node.emitNode = undefined;
                }
            }
        }
        ts.disposeEmitNodes = disposeEmitNodes;
        /**
         * Sets `EmitFlags.NoComments` on a node and removes any leading and trailing synthetic comments.
         * @internal
         */
        function removeAllComments(node) {
            var emitNode = getOrCreateEmitNode(node);
            emitNode.flags |= 1536 /* NoComments */;
            emitNode.leadingComments = undefined;
            emitNode.trailingComments = undefined;
            return node;
        }
        ts.removeAllComments = removeAllComments;
        /**
         * Sets flags that control emit behavior of a node.
         */
        function setEmitFlags(node, emitFlags) {
            getOrCreateEmitNode(node).flags = emitFlags;
            return node;
        }
        ts.setEmitFlags = setEmitFlags;
        /**
         * Sets flags that control emit behavior of a node.
         */
        /* @internal */
        function addEmitFlags(node, emitFlags) {
            var emitNode = getOrCreateEmitNode(node);
            emitNode.flags = emitNode.flags | emitFlags;
            return node;
        }
        ts.addEmitFlags = addEmitFlags;
        /**
         * Gets a custom text range to use when emitting source maps.
         */
        function getSourceMapRange(node) {
            var _a, _b;
            return (_b = (_a = node.emitNode) === null || _a === void 0 ? void 0 : _a.sourceMapRange) !== null && _b !== void 0 ? _b : node;
        }
        ts.getSourceMapRange = getSourceMapRange;
        /**
         * Sets a custom text range to use when emitting source maps.
         */
        function setSourceMapRange(node, range) {
            getOrCreateEmitNode(node).sourceMapRange = range;
            return node;
        }
        ts.setSourceMapRange = setSourceMapRange;
        /**
         * Gets the TextRange to use for source maps for a token of a node.
         */
        function getTokenSourceMapRange(node, token) {
            var _a, _b;
            return (_b = (_a = node.emitNode) === null || _a === void 0 ? void 0 : _a.tokenSourceMapRanges) === null || _b === void 0 ? void 0 : _b[token];
        }
        ts.getTokenSourceMapRange = getTokenSourceMapRange;
        /**
         * Sets the TextRange to use for source maps for a token of a node.
         */
        function setTokenSourceMapRange(node, token, range) {
            var _a;
            var emitNode = getOrCreateEmitNode(node);
            var tokenSourceMapRanges = (_a = emitNode.tokenSourceMapRanges) !== null && _a !== void 0 ? _a : (emitNode.tokenSourceMapRanges = []);
            tokenSourceMapRanges[token] = range;
            return node;
        }
        ts.setTokenSourceMapRange = setTokenSourceMapRange;
        /**
         * Gets a custom text range to use when emitting comments.
         */
        /*@internal*/
        function getStartsOnNewLine(node) {
            var _a;
            return (_a = node.emitNode) === null || _a === void 0 ? void 0 : _a.startsOnNewLine;
        }
        ts.getStartsOnNewLine = getStartsOnNewLine;
        /**
         * Sets a custom text range to use when emitting comments.
         */
        /*@internal*/
        function setStartsOnNewLine(node, newLine) {
            getOrCreateEmitNode(node).startsOnNewLine = newLine;
            return node;
        }
        ts.setStartsOnNewLine = setStartsOnNewLine;
        /**
         * Gets a custom text range to use when emitting comments.
         */
        function getCommentRange(node) {
            var _a, _b;
            return (_b = (_a = node.emitNode) === null || _a === void 0 ? void 0 : _a.commentRange) !== null && _b !== void 0 ? _b : node;
        }
        ts.getCommentRange = getCommentRange;
        /**
         * Sets a custom text range to use when emitting comments.
         */
        function setCommentRange(node, range) {
            getOrCreateEmitNode(node).commentRange = range;
            return node;
        }
        ts.setCommentRange = setCommentRange;
        function getSyntheticLeadingComments(node) {
            var _a;
            return (_a = node.emitNode) === null || _a === void 0 ? void 0 : _a.leadingComments;
        }
        ts.getSyntheticLeadingComments = getSyntheticLeadingComments;
        function setSyntheticLeadingComments(node, comments) {
            getOrCreateEmitNode(node).leadingComments = comments;
            return node;
        }
        ts.setSyntheticLeadingComments = setSyntheticLeadingComments;
        function addSyntheticLeadingComment(node, kind, text, hasTrailingNewLine) {
            return setSyntheticLeadingComments(node, ts.append(getSyntheticLeadingComments(node), { kind: kind, pos: -1, end: -1, hasTrailingNewLine: hasTrailingNewLine, text: text }));
        }
        ts.addSyntheticLeadingComment = addSyntheticLeadingComment;
        function getSyntheticTrailingComments(node) {
            var _a;
            return (_a = node.emitNode) === null || _a === void 0 ? void 0 : _a.trailingComments;
        }
        ts.getSyntheticTrailingComments = getSyntheticTrailingComments;
        function setSyntheticTrailingComments(node, comments) {
            getOrCreateEmitNode(node).trailingComments = comments;
            return node;
        }
        ts.setSyntheticTrailingComments = setSyntheticTrailingComments;
        function addSyntheticTrailingComment(node, kind, text, hasTrailingNewLine) {
            return setSyntheticTrailingComments(node, ts.append(getSyntheticTrailingComments(node), { kind: kind, pos: -1, end: -1, hasTrailingNewLine: hasTrailingNewLine, text: text }));
        }
        ts.addSyntheticTrailingComment = addSyntheticTrailingComment;
        function moveSyntheticComments(node, original) {
            setSyntheticLeadingComments(node, getSyntheticLeadingComments(original));
            setSyntheticTrailingComments(node, getSyntheticTrailingComments(original));
            var emit = getOrCreateEmitNode(original);
            emit.leadingComments = undefined;
            emit.trailingComments = undefined;
            return node;
        }
        ts.moveSyntheticComments = moveSyntheticComments;
        /**
         * Gets the constant value to emit for an expression representing an enum.
         */
        function getConstantValue(node) {
            var _a;
            return (_a = node.emitNode) === null || _a === void 0 ? void 0 : _a.constantValue;
        }
        ts.getConstantValue = getConstantValue;
        /**
         * Sets the constant value to emit for an expression.
         */
        function setConstantValue(node, value) {
            var emitNode = getOrCreateEmitNode(node);
            emitNode.constantValue = value;
            return node;
        }
        ts.setConstantValue = setConstantValue;
        /**
         * Adds an EmitHelper to a node.
         */
        function addEmitHelper(node, helper) {
            var emitNode = getOrCreateEmitNode(node);
            emitNode.helpers = ts.append(emitNode.helpers, helper);
            return node;
        }
        ts.addEmitHelper = addEmitHelper;
        /**
         * Add EmitHelpers to a node.
         */
        function addEmitHelpers(node, helpers) {
            if (ts.some(helpers)) {
                var emitNode = getOrCreateEmitNode(node);
                for (var _i = 0, helpers_2 = helpers; _i < helpers_2.length; _i++) {
                    var helper = helpers_2[_i];
                    emitNode.helpers = ts.appendIfUnique(emitNode.helpers, helper);
                }
            }
            return node;
        }
        ts.addEmitHelpers = addEmitHelpers;
        /**
         * Removes an EmitHelper from a node.
         */
        function removeEmitHelper(node, helper) {
            var _a;
            var helpers = (_a = node.emitNode) === null || _a === void 0 ? void 0 : _a.helpers;
            if (helpers) {
                return ts.orderedRemoveItem(helpers, helper);
            }
            return false;
        }
        ts.removeEmitHelper = removeEmitHelper;
        /**
         * Gets the EmitHelpers of a node.
         */
        function getEmitHelpers(node) {
            var _a;
            return (_a = node.emitNode) === null || _a === void 0 ? void 0 : _a.helpers;
        }
        ts.getEmitHelpers = getEmitHelpers;
        /**
         * Moves matching emit helpers from a source node to a target node.
         */
        function moveEmitHelpers(source, target, predicate) {
            var sourceEmitNode = source.emitNode;
            var sourceEmitHelpers = sourceEmitNode && sourceEmitNode.helpers;
            if (!ts.some(sourceEmitHelpers))
                return;
            var targetEmitNode = getOrCreateEmitNode(target);
            var helpersRemoved = 0;
            for (var i = 0; i < sourceEmitHelpers.length; i++) {
                var helper = sourceEmitHelpers[i];
                if (predicate(helper)) {
                    helpersRemoved++;
                    targetEmitNode.helpers = ts.appendIfUnique(targetEmitNode.helpers, helper);
                }
                else if (helpersRemoved > 0) {
                    sourceEmitHelpers[i - helpersRemoved] = helper;
                }
            }
            if (helpersRemoved > 0) {
                sourceEmitHelpers.length -= helpersRemoved;
            }
        }
        ts.moveEmitHelpers = moveEmitHelpers;
        /* @internal */
        function ignoreSourceNewlines(node) {
            getOrCreateEmitNode(node).flags |= 134217728 /* IgnoreSourceNewlines */;
            return node;
        }
        ts.ignoreSourceNewlines = ignoreSourceNewlines;
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function createEmitHelperFactory(context) {
            var factory = context.factory;
            var immutableTrue = ts.memoize(function () { return ts.setEmitFlags(factory.createTrue(), 268435456 /* Immutable */); });
            var immutableFalse = ts.memoize(function () { return ts.setEmitFlags(factory.createFalse(), 268435456 /* Immutable */); });
            return {
                getUnscopedHelperName: getUnscopedHelperName,
                // TypeScript Helpers
                createDecorateHelper: createDecorateHelper,
                createMetadataHelper: createMetadataHelper,
                createParamHelper: createParamHelper,
                // ES2018 Helpers
                createAssignHelper: createAssignHelper,
                createAwaitHelper: createAwaitHelper,
                createAsyncGeneratorHelper: createAsyncGeneratorHelper,
                createAsyncDelegatorHelper: createAsyncDelegatorHelper,
                createAsyncValuesHelper: createAsyncValuesHelper,
                // ES2018 Destructuring Helpers
                createRestHelper: createRestHelper,
                // ES2017 Helpers
                createAwaiterHelper: createAwaiterHelper,
                // ES2015 Helpers
                createExtendsHelper: createExtendsHelper,
                createTemplateObjectHelper: createTemplateObjectHelper,
                createSpreadArrayHelper: createSpreadArrayHelper,
                // ES2015 Destructuring Helpers
                createValuesHelper: createValuesHelper,
                createReadHelper: createReadHelper,
                // ES2015 Generator Helpers
                createGeneratorHelper: createGeneratorHelper,
                // ES Module Helpers
                createCreateBindingHelper: createCreateBindingHelper,
                createImportStarHelper: createImportStarHelper,
                createImportStarCallbackHelper: createImportStarCallbackHelper,
                createImportDefaultHelper: createImportDefaultHelper,
                createExportStarHelper: createExportStarHelper,
                // Class Fields Helpers
                createClassPrivateFieldGetHelper: createClassPrivateFieldGetHelper,
                createClassPrivateFieldSetHelper: createClassPrivateFieldSetHelper,
            };
            /**
             * Gets an identifier for the name of an *unscoped* emit helper.
             */
            function getUnscopedHelperName(name) {
                return ts.setEmitFlags(factory.createIdentifier(name), 4096 /* HelperName */ | 2 /* AdviseOnEmitNode */);
            }
            // TypeScript Helpers
            function createDecorateHelper(decoratorExpressions, target, memberName, descriptor) {
                context.requestEmitHelper(ts.decorateHelper);
                var argumentsArray = [];
                argumentsArray.push(factory.createArrayLiteralExpression(decoratorExpressions, /*multiLine*/ true));
                argumentsArray.push(target);
                if (memberName) {
                    argumentsArray.push(memberName);
                    if (descriptor) {
                        argumentsArray.push(descriptor);
                    }
                }
                return factory.createCallExpression(getUnscopedHelperName("__decorate"), 
                /*typeArguments*/ undefined, argumentsArray);
            }
            function createMetadataHelper(metadataKey, metadataValue) {
                context.requestEmitHelper(ts.metadataHelper);
                return factory.createCallExpression(getUnscopedHelperName("__metadata"), 
                /*typeArguments*/ undefined, [
                    factory.createStringLiteral(metadataKey),
                    metadataValue
                ]);
            }
            function createParamHelper(expression, parameterOffset, location) {
                context.requestEmitHelper(ts.paramHelper);
                return ts.setTextRange(factory.createCallExpression(getUnscopedHelperName("__param"), 
                /*typeArguments*/ undefined, [
                    factory.createNumericLiteral(parameterOffset + ""),
                    expression
                ]), location);
            }
            // ES2018 Helpers
            function createAssignHelper(attributesSegments) {
                if (context.getCompilerOptions().target >= 2 /* ES2015 */) {
                    return factory.createCallExpression(factory.createPropertyAccessExpression(factory.createIdentifier("Object"), "assign"), 
                    /*typeArguments*/ undefined, attributesSegments);
                }
                context.requestEmitHelper(ts.assignHelper);
                return factory.createCallExpression(getUnscopedHelperName("__assign"), 
                /*typeArguments*/ undefined, attributesSegments);
            }
            function createAwaitHelper(expression) {
                context.requestEmitHelper(ts.awaitHelper);
                return factory.createCallExpression(getUnscopedHelperName("__await"), /*typeArguments*/ undefined, [expression]);
            }
            function createAsyncGeneratorHelper(generatorFunc, hasLexicalThis) {
                context.requestEmitHelper(ts.awaitHelper);
                context.requestEmitHelper(ts.asyncGeneratorHelper);
                // Mark this node as originally an async function
                (generatorFunc.emitNode || (generatorFunc.emitNode = {})).flags |= 262144 /* AsyncFunctionBody */ | 524288 /* ReuseTempVariableScope */;
                return factory.createCallExpression(getUnscopedHelperName("__asyncGenerator"), 
                /*typeArguments*/ undefined, [
                    hasLexicalThis ? factory.createThis() : factory.createVoidZero(),
                    factory.createIdentifier("arguments"),
                    generatorFunc
                ]);
            }
            function createAsyncDelegatorHelper(expression) {
                context.requestEmitHelper(ts.awaitHelper);
                context.requestEmitHelper(ts.asyncDelegator);
                return factory.createCallExpression(getUnscopedHelperName("__asyncDelegator"), 
                /*typeArguments*/ undefined, [expression]);
            }
            function createAsyncValuesHelper(expression) {
                context.requestEmitHelper(ts.asyncValues);
                return factory.createCallExpression(getUnscopedHelperName("__asyncValues"), 
                /*typeArguments*/ undefined, [expression]);
            }
            // ES2018 Destructuring Helpers
            /** Given value: o, propName: p, pattern: { a, b, ...p } from the original statement
             * `{ a, b, ...p } = o`, create `p = __rest(o, ["a", "b"]);`
             */
            function createRestHelper(value, elements, computedTempVariables, location) {
                context.requestEmitHelper(ts.restHelper);
                var propertyNames = [];
                var computedTempVariableOffset = 0;
                for (var i = 0; i < elements.length - 1; i++) {
                    var propertyName = ts.getPropertyNameOfBindingOrAssignmentElement(elements[i]);
                    if (propertyName) {
                        if (ts.isComputedPropertyName(propertyName)) {
                            ts.Debug.assertIsDefined(computedTempVariables, "Encountered computed property name but 'computedTempVariables' argument was not provided.");
                            var temp = computedTempVariables[computedTempVariableOffset];
                            computedTempVariableOffset++;
                            // typeof _tmp === "symbol" ? _tmp : _tmp + ""
                            propertyNames.push(factory.createConditionalExpression(factory.createTypeCheck(temp, "symbol"), 
                            /*questionToken*/ undefined, temp, 
                            /*colonToken*/ undefined, factory.createAdd(temp, factory.createStringLiteral(""))));
                        }
                        else {
                            propertyNames.push(factory.createStringLiteralFromNode(propertyName));
                        }
                    }
                }
                return factory.createCallExpression(getUnscopedHelperName("__rest"), 
                /*typeArguments*/ undefined, [
                    value,
                    ts.setTextRange(factory.createArrayLiteralExpression(propertyNames), location)
                ]);
            }
            // ES2017 Helpers
            function createAwaiterHelper(hasLexicalThis, hasLexicalArguments, promiseConstructor, body) {
                context.requestEmitHelper(ts.awaiterHelper);
                var generatorFunc = factory.createFunctionExpression(
                /*modifiers*/ undefined, factory.createToken(41 /* AsteriskToken */), 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, 
                /*parameters*/ [], 
                /*type*/ undefined, body);
                // Mark this node as originally an async function
                (generatorFunc.emitNode || (generatorFunc.emitNode = {})).flags |= 262144 /* AsyncFunctionBody */ | 524288 /* ReuseTempVariableScope */;
                return factory.createCallExpression(getUnscopedHelperName("__awaiter"), 
                /*typeArguments*/ undefined, [
                    hasLexicalThis ? factory.createThis() : factory.createVoidZero(),
                    hasLexicalArguments ? factory.createIdentifier("arguments") : factory.createVoidZero(),
                    promiseConstructor ? ts.createExpressionFromEntityName(factory, promiseConstructor) : factory.createVoidZero(),
                    generatorFunc
                ]);
            }
            // ES2015 Helpers
            function createExtendsHelper(name) {
                context.requestEmitHelper(ts.extendsHelper);
                return factory.createCallExpression(getUnscopedHelperName("__extends"), 
                /*typeArguments*/ undefined, [name, factory.createUniqueName("_super", 16 /* Optimistic */ | 32 /* FileLevel */)]);
            }
            function createTemplateObjectHelper(cooked, raw) {
                context.requestEmitHelper(ts.templateObjectHelper);
                return factory.createCallExpression(getUnscopedHelperName("__makeTemplateObject"), 
                /*typeArguments*/ undefined, [cooked, raw]);
            }
            function createSpreadArrayHelper(to, from, packFrom) {
                context.requestEmitHelper(ts.spreadArrayHelper);
                return factory.createCallExpression(getUnscopedHelperName("__spreadArray"), 
                /*typeArguments*/ undefined, [to, from, packFrom ? immutableTrue() : immutableFalse()]);
            }
            // ES2015 Destructuring Helpers
            function createValuesHelper(expression) {
                context.requestEmitHelper(ts.valuesHelper);
                return factory.createCallExpression(getUnscopedHelperName("__values"), 
                /*typeArguments*/ undefined, [expression]);
            }
            function createReadHelper(iteratorRecord, count) {
                context.requestEmitHelper(ts.readHelper);
                return factory.createCallExpression(getUnscopedHelperName("__read"), 
                /*typeArguments*/ undefined, count !== undefined
                    ? [iteratorRecord, factory.createNumericLiteral(count + "")]
                    : [iteratorRecord]);
            }
            // ES2015 Generator Helpers
            function createGeneratorHelper(body) {
                context.requestEmitHelper(ts.generatorHelper);
                return factory.createCallExpression(getUnscopedHelperName("__generator"), 
                /*typeArguments*/ undefined, [factory.createThis(), body]);
            }
            // ES Module Helpers
            function createCreateBindingHelper(module, inputName, outputName) {
                context.requestEmitHelper(ts.createBindingHelper);
                return factory.createCallExpression(getUnscopedHelperName("__createBinding"), 
                /*typeArguments*/ undefined, __spreadArray([factory.createIdentifier("exports"), module, inputName], (outputName ? [outputName] : []), true));
            }
            function createImportStarHelper(expression) {
                context.requestEmitHelper(ts.importStarHelper);
                return factory.createCallExpression(getUnscopedHelperName("__importStar"), 
                /*typeArguments*/ undefined, [expression]);
            }
            function createImportStarCallbackHelper() {
                context.requestEmitHelper(ts.importStarHelper);
                return getUnscopedHelperName("__importStar");
            }
            function createImportDefaultHelper(expression) {
                context.requestEmitHelper(ts.importDefaultHelper);
                return factory.createCallExpression(getUnscopedHelperName("__importDefault"), 
                /*typeArguments*/ undefined, [expression]);
            }
            function createExportStarHelper(moduleExpression, exportsExpression) {
                if (exportsExpression === void 0) {
                    exportsExpression = factory.createIdentifier("exports");
                }
                context.requestEmitHelper(ts.exportStarHelper);
                context.requestEmitHelper(ts.createBindingHelper);
                return factory.createCallExpression(getUnscopedHelperName("__exportStar"), 
                /*typeArguments*/ undefined, [moduleExpression, exportsExpression]);
            }
            // Class Fields Helpers
            function createClassPrivateFieldGetHelper(receiver, state, kind, f) {
                context.requestEmitHelper(ts.classPrivateFieldGetHelper);
                var args;
                if (!f) {
                    args = [receiver, state, factory.createStringLiteral(kind)];
                }
                else {
                    args = [receiver, state, factory.createStringLiteral(kind), f];
                }
                return factory.createCallExpression(getUnscopedHelperName("__classPrivateFieldGet"), /*typeArguments*/ undefined, args);
            }
            function createClassPrivateFieldSetHelper(receiver, state, value, kind, f) {
                context.requestEmitHelper(ts.classPrivateFieldSetHelper);
                var args;
                if (!f) {
                    args = [receiver, state, value, factory.createStringLiteral(kind)];
                }
                else {
                    args = [receiver, state, value, factory.createStringLiteral(kind), f];
                }
                return factory.createCallExpression(getUnscopedHelperName("__classPrivateFieldSet"), /*typeArguments*/ undefined, args);
            }
        }
        ts.createEmitHelperFactory = createEmitHelperFactory;
        /* @internal */
        function compareEmitHelpers(x, y) {
            if (x === y)
                return 0 /* EqualTo */;
            if (x.priority === y.priority)
                return 0 /* EqualTo */;
            if (x.priority === undefined)
                return 1 /* GreaterThan */;
            if (y.priority === undefined)
                return -1 /* LessThan */;
            return ts.compareValues(x.priority, y.priority);
        }
        ts.compareEmitHelpers = compareEmitHelpers;
        /**
         * @param input Template string input strings
         * @param args Names which need to be made file-level unique
         */
        function helperString(input) {
            var args = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                args[_i - 1] = arguments[_i];
            }
            return function (uniqueName) {
                var result = "";
                for (var i = 0; i < args.length; i++) {
                    result += input[i];
                    result += uniqueName(args[i]);
                }
                result += input[input.length - 1];
                return result;
            };
        }
        ts.helperString = helperString;
        // TypeScript Helpers
        ts.decorateHelper = {
            name: "typescript:decorate",
            importName: "__decorate",
            scoped: false,
            priority: 2,
            text: "\n            var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {\n                var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;\n                if (typeof Reflect === \"object\" && typeof Reflect.decorate === \"function\") r = Reflect.decorate(decorators, target, key, desc);\n                else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;\n                return c > 3 && r && Object.defineProperty(target, key, r), r;\n            };"
        };
        ts.metadataHelper = {
            name: "typescript:metadata",
            importName: "__metadata",
            scoped: false,
            priority: 3,
            text: "\n            var __metadata = (this && this.__metadata) || function (k, v) {\n                if (typeof Reflect === \"object\" && typeof Reflect.metadata === \"function\") return Reflect.metadata(k, v);\n            };"
        };
        ts.paramHelper = {
            name: "typescript:param",
            importName: "__param",
            scoped: false,
            priority: 4,
            text: "\n            var __param = (this && this.__param) || function (paramIndex, decorator) {\n                return function (target, key) { decorator(target, key, paramIndex); }\n            };"
        };
        // ES2018 Helpers
        ts.assignHelper = {
            name: "typescript:assign",
            importName: "__assign",
            scoped: false,
            priority: 1,
            text: "\n            var __assign = (this && this.__assign) || function () {\n                __assign = Object.assign || function(t) {\n                    for (var s, i = 1, n = arguments.length; i < n; i++) {\n                        s = arguments[i];\n                        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))\n                            t[p] = s[p];\n                    }\n                    return t;\n                };\n                return __assign.apply(this, arguments);\n            };"
        };
        ts.awaitHelper = {
            name: "typescript:await",
            importName: "__await",
            scoped: false,
            text: "\n            var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }"
        };
        ts.asyncGeneratorHelper = {
            name: "typescript:asyncGenerator",
            importName: "__asyncGenerator",
            scoped: false,
            dependencies: [ts.awaitHelper],
            text: "\n            var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {\n                if (!Symbol.asyncIterator) throw new TypeError(\"Symbol.asyncIterator is not defined.\");\n                var g = generator.apply(thisArg, _arguments || []), i, q = [];\n                return i = {}, verb(\"next\"), verb(\"throw\"), verb(\"return\"), i[Symbol.asyncIterator] = function () { return this; }, i;\n                function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }\n                function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }\n                function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }\n                function fulfill(value) { resume(\"next\", value); }\n                function reject(value) { resume(\"throw\", value); }\n                function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }\n            };"
        };
        ts.asyncDelegator = {
            name: "typescript:asyncDelegator",
            importName: "__asyncDelegator",
            scoped: false,
            dependencies: [ts.awaitHelper],
            text: "\n            var __asyncDelegator = (this && this.__asyncDelegator) || function (o) {\n                var i, p;\n                return i = {}, verb(\"next\"), verb(\"throw\", function (e) { throw e; }), verb(\"return\"), i[Symbol.iterator] = function () { return this; }, i;\n                function verb(n, f) { i[n] = o[n] ? function (v) { return (p = !p) ? { value: __await(o[n](v)), done: n === \"return\" } : f ? f(v) : v; } : f; }\n            };"
        };
        ts.asyncValues = {
            name: "typescript:asyncValues",
            importName: "__asyncValues",
            scoped: false,
            text: "\n            var __asyncValues = (this && this.__asyncValues) || function (o) {\n                if (!Symbol.asyncIterator) throw new TypeError(\"Symbol.asyncIterator is not defined.\");\n                var m = o[Symbol.asyncIterator], i;\n                return m ? m.call(o) : (o = typeof __values === \"function\" ? __values(o) : o[Symbol.iterator](), i = {}, verb(\"next\"), verb(\"throw\"), verb(\"return\"), i[Symbol.asyncIterator] = function () { return this; }, i);\n                function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }\n                function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }\n            };"
        };
        // ES2018 Destructuring Helpers
        ts.restHelper = {
            name: "typescript:rest",
            importName: "__rest",
            scoped: false,
            text: "\n            var __rest = (this && this.__rest) || function (s, e) {\n                var t = {};\n                for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)\n                    t[p] = s[p];\n                if (s != null && typeof Object.getOwnPropertySymbols === \"function\")\n                    for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {\n                        if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))\n                            t[p[i]] = s[p[i]];\n                    }\n                return t;\n            };"
        };
        // ES2017 Helpers
        ts.awaiterHelper = {
            name: "typescript:awaiter",
            importName: "__awaiter",
            scoped: false,
            priority: 5,
            text: "\n            var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {\n                function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }\n                return new (P || (P = Promise))(function (resolve, reject) {\n                    function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }\n                    function rejected(value) { try { step(generator[\"throw\"](value)); } catch (e) { reject(e); } }\n                    function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }\n                    step((generator = generator.apply(thisArg, _arguments || [])).next());\n                });\n            };"
        };
        // ES2015 Helpers
        ts.extendsHelper = {
            name: "typescript:extends",
            importName: "__extends",
            scoped: false,
            priority: 0,
            text: "\n            var __extends = (this && this.__extends) || (function () {\n                var extendStatics = function (d, b) {\n                    extendStatics = Object.setPrototypeOf ||\n                        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||\n                        function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };\n                    return extendStatics(d, b);\n                };\n\n                return function (d, b) {\n                    if (typeof b !== \"function\" && b !== null)\n                        throw new TypeError(\"Class extends value \" + String(b) + \" is not a constructor or null\");\n                    extendStatics(d, b);\n                    function __() { this.constructor = d; }\n                    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());\n                };\n            })();"
        };
        ts.templateObjectHelper = {
            name: "typescript:makeTemplateObject",
            importName: "__makeTemplateObject",
            scoped: false,
            priority: 0,
            text: "\n            var __makeTemplateObject = (this && this.__makeTemplateObject) || function (cooked, raw) {\n                if (Object.defineProperty) { Object.defineProperty(cooked, \"raw\", { value: raw }); } else { cooked.raw = raw; }\n                return cooked;\n            };"
        };
        ts.readHelper = {
            name: "typescript:read",
            importName: "__read",
            scoped: false,
            text: "\n            var __read = (this && this.__read) || function (o, n) {\n                var m = typeof Symbol === \"function\" && o[Symbol.iterator];\n                if (!m) return o;\n                var i = m.call(o), r, ar = [], e;\n                try {\n                    while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);\n                }\n                catch (error) { e = { error: error }; }\n                finally {\n                    try {\n                        if (r && !r.done && (m = i[\"return\"])) m.call(i);\n                    }\n                    finally { if (e) throw e.error; }\n                }\n                return ar;\n            };"
        };
        ts.spreadArrayHelper = {
            name: "typescript:spreadArray",
            importName: "__spreadArray",
            scoped: false,
            text: "\n            var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {\n                if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {\n                    if (ar || !(i in from)) {\n                        if (!ar) ar = Array.prototype.slice.call(from, 0, i);\n                        ar[i] = from[i];\n                    }\n                }\n                return to.concat(ar || Array.prototype.slice.call(from));\n            };"
        };
        // ES2015 Destructuring Helpers
        ts.valuesHelper = {
            name: "typescript:values",
            importName: "__values",
            scoped: false,
            text: "\n            var __values = (this && this.__values) || function(o) {\n                var s = typeof Symbol === \"function\" && Symbol.iterator, m = s && o[s], i = 0;\n                if (m) return m.call(o);\n                if (o && typeof o.length === \"number\") return {\n                    next: function () {\n                        if (o && i >= o.length) o = void 0;\n                        return { value: o && o[i++], done: !o };\n                    }\n                };\n                throw new TypeError(s ? \"Object is not iterable.\" : \"Symbol.iterator is not defined.\");\n            };"
        };
        // ES2015 Generator Helpers
        // The __generator helper is used by down-level transformations to emulate the runtime
        // semantics of an ES2015 generator function. When called, this helper returns an
        // object that implements the Iterator protocol, in that it has `next`, `return`, and
        // `throw` methods that step through the generator when invoked.
        //
        // parameters:
        //  @param thisArg  The value to use as the `this` binding for the transformed generator body.
        //  @param body     A function that acts as the transformed generator body.
        //
        // variables:
        //  _       Persistent state for the generator that is shared between the helper and the
        //          generator body. The state object has the following members:
        //            sent() - A method that returns or throws the current completion value.
        //            label  - The next point at which to resume evaluation of the generator body.
        //            trys   - A stack of protected regions (try/catch/finally blocks).
        //            ops    - A stack of pending instructions when inside of a finally block.
        //  f       A value indicating whether the generator is executing.
        //  y       An iterator to delegate for a yield*.
        //  t       A temporary variable that holds one of the following values (note that these
        //          cases do not overlap):
        //          - The completion value when resuming from a `yield` or `yield*`.
        //          - The error value for a catch block.
        //          - The current protected region (array of try/catch/finally/end labels).
        //          - The verb (`next`, `throw`, or `return` method) to delegate to the expression
        //            of a `yield*`.
        //          - The result of evaluating the verb delegated to the expression of a `yield*`.
        //
        // functions:
        //  verb(n)     Creates a bound callback to the `step` function for opcode `n`.
        //  step(op)    Evaluates opcodes in a generator body until execution is suspended or
        //              completed.
        //
        // The __generator helper understands a limited set of instructions:
        //  0: next(value?)     - Start or resume the generator with the specified value.
        //  1: throw(error)     - Resume the generator with an exception. If the generator is
        //                        suspended inside of one or more protected regions, evaluates
        //                        any intervening finally blocks between the current label and
        //                        the nearest catch block or function boundary. If uncaught, the
        //                        exception is thrown to the caller.
        //  2: return(value?)   - Resume the generator as if with a return. If the generator is
        //                        suspended inside of one or more protected regions, evaluates any
        //                        intervening finally blocks.
        //  3: break(label)     - Jump to the specified label. If the label is outside of the
        //                        current protected region, evaluates any intervening finally
        //                        blocks.
        //  4: yield(value?)    - Yield execution to the caller with an optional value. When
        //                        resumed, the generator will continue at the next label.
        //  5: yield*(value)    - Delegates evaluation to the supplied iterator. When
        //                        delegation completes, the generator will continue at the next
        //                        label.
        //  6: catch(error)     - Handles an exception thrown from within the generator body. If
        //                        the current label is inside of one or more protected regions,
        //                        evaluates any intervening finally blocks between the current
        //                        label and the nearest catch block or function boundary. If
        //                        uncaught, the exception is thrown to the caller.
        //  7: endfinally       - Ends a finally block, resuming the last instruction prior to
        //                        entering a finally block.
        //
        // For examples of how these are used, see the comments in ./transformers/generators.ts
        ts.generatorHelper = {
            name: "typescript:generator",
            importName: "__generator",
            scoped: false,
            priority: 6,
            text: "\n            var __generator = (this && this.__generator) || function (thisArg, body) {\n                var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;\n                return g = { next: verb(0), \"throw\": verb(1), \"return\": verb(2) }, typeof Symbol === \"function\" && (g[Symbol.iterator] = function() { return this; }), g;\n                function verb(n) { return function (v) { return step([n, v]); }; }\n                function step(op) {\n                    if (f) throw new TypeError(\"Generator is already executing.\");\n                    while (_) try {\n                        if (f = 1, y && (t = op[0] & 2 ? y[\"return\"] : op[0] ? y[\"throw\"] || ((t = y[\"return\"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;\n                        if (y = 0, t) op = [op[0] & 2, t.value];\n                        switch (op[0]) {\n                            case 0: case 1: t = op; break;\n                            case 4: _.label++; return { value: op[1], done: false };\n                            case 5: _.label++; y = op[1]; op = [0]; continue;\n                            case 7: op = _.ops.pop(); _.trys.pop(); continue;\n                            default:\n                                if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }\n                                if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }\n                                if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }\n                                if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }\n                                if (t[2]) _.ops.pop();\n                                _.trys.pop(); continue;\n                        }\n                        op = body.call(thisArg, _);\n                    } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }\n                    if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };\n                }\n            };"
        };
        // ES Module Helpers
        ts.createBindingHelper = {
            name: "typescript:commonjscreatebinding",
            importName: "__createBinding",
            scoped: false,
            priority: 1,
            text: "\n            var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {\n                if (k2 === undefined) k2 = k;\n                Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });\n            }) : (function(o, m, k, k2) {\n                if (k2 === undefined) k2 = k;\n                o[k2] = m[k];\n            }));"
        };
        ts.setModuleDefaultHelper = {
            name: "typescript:commonjscreatevalue",
            importName: "__setModuleDefault",
            scoped: false,
            priority: 1,
            text: "\n            var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {\n                Object.defineProperty(o, \"default\", { enumerable: true, value: v });\n            }) : function(o, v) {\n                o[\"default\"] = v;\n            });"
        };
        // emit helper for `import * as Name from "foo"`
        ts.importStarHelper = {
            name: "typescript:commonjsimportstar",
            importName: "__importStar",
            scoped: false,
            dependencies: [ts.createBindingHelper, ts.setModuleDefaultHelper],
            priority: 2,
            text: "\n            var __importStar = (this && this.__importStar) || function (mod) {\n                if (mod && mod.__esModule) return mod;\n                var result = {};\n                if (mod != null) for (var k in mod) if (k !== \"default\" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);\n                __setModuleDefault(result, mod);\n                return result;\n            };"
        };
        // emit helper for `import Name from "foo"`
        ts.importDefaultHelper = {
            name: "typescript:commonjsimportdefault",
            importName: "__importDefault",
            scoped: false,
            text: "\n            var __importDefault = (this && this.__importDefault) || function (mod) {\n                return (mod && mod.__esModule) ? mod : { \"default\": mod };\n            };"
        };
        ts.exportStarHelper = {
            name: "typescript:export-star",
            importName: "__exportStar",
            scoped: false,
            dependencies: [ts.createBindingHelper],
            priority: 2,
            text: "\n            var __exportStar = (this && this.__exportStar) || function(m, exports) {\n                for (var p in m) if (p !== \"default\" && !Object.prototype.hasOwnProperty.call(exports, p)) __createBinding(exports, m, p);\n            };"
        };
        /**
         * Parameters:
         *  @param receiver — The object from which the private member will be read.
         *  @param state — One of the following:
         *      - A WeakMap used to read a private instance field.
         *      - A WeakSet used as an instance brand for private instance methods and accessors.
         *      - A function value that should be the undecorated class constructor used to brand check private static fields, methods, and accessors.
         *  @param kind — (optional pre TS 4.3, required for TS 4.3+) One of the following values:
         *      - undefined — Indicates a private instance field (pre TS 4.3).
         *      - "f" — Indicates a private field (instance or static).
         *      - "m" — Indicates a private method (instance or static).
         *      - "a" — Indicates a private accessor (instance or static).
         *  @param f — (optional pre TS 4.3) Depends on the arguments for state and kind:
         *      - If kind is "m", this should be the function corresponding to the static or instance method.
         *      - If kind is "a", this should be the function corresponding to the getter method, or undefined if the getter was not defined.
         *      - If kind is "f" and state is a function, this should be an object holding the value of a static field, or undefined if the static field declaration has not yet been evaluated.
         * Usage:
         * This helper will only ever be used by the compiler in the following ways:
         *
         * Reading from a private instance field (pre TS 4.3):
         *      __classPrivateFieldGet(<any>, <WeakMap>)
         *
         * Reading from a private instance field (TS 4.3+):
         *      __classPrivateFieldGet(<any>, <WeakMap>, "f")
         *
         * Reading from a private instance get accessor (when defined, TS 4.3+):
         *      __classPrivateFieldGet(<any>, <WeakSet>, "a", <function>)
         *
         * Reading from a private instance get accessor (when not defined, TS 4.3+):
         *      __classPrivateFieldGet(<any>, <WeakSet>, "a", void 0)
         *      NOTE: This always results in a runtime error.
         *
         * Reading from a private instance method (TS 4.3+):
         *      __classPrivateFieldGet(<any>, <WeakSet>, "m", <function>)
         *
         * Reading from a private static field (TS 4.3+):
         *      __classPrivateFieldGet(<any>, <constructor>, "f", <{ value: any }>)
         *
         * Reading from a private static get accessor (when defined, TS 4.3+):
         *      __classPrivateFieldGet(<any>, <constructor>, "a", <function>)
         *
         * Reading from a private static get accessor (when not defined, TS 4.3+):
         *      __classPrivateFieldGet(<any>, <constructor>, "a", void 0)
         *      NOTE: This always results in a runtime error.
         *
         * Reading from a private static method (TS 4.3+):
         *      __classPrivateFieldGet(<any>, <constructor>, "m", <function>)
         */
        ts.classPrivateFieldGetHelper = {
            name: "typescript:classPrivateFieldGet",
            importName: "__classPrivateFieldGet",
            scoped: false,
            text: "\n            var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {\n                if (kind === \"a\" && !f) throw new TypeError(\"Private accessor was defined without a getter\");\n                if (typeof state === \"function\" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError(\"Cannot read private member from an object whose class did not declare it\");\n                return kind === \"m\" ? f : kind === \"a\" ? f.call(receiver) : f ? f.value : state.get(receiver);\n            };"
        };
        /**
         * Parameters:
         *  @param receiver — The object on which the private member will be set.
         *  @param state — One of the following:
         *      - A WeakMap used to store a private instance field.
         *      - A WeakSet used as an instance brand for private instance methods and accessors.
         *      - A function value that should be the undecorated class constructor used to brand check private static fields, methods, and accessors.
         *  @param value — The value to set.
         *  @param kind — (optional pre TS 4.3, required for TS 4.3+) One of the following values:
         *       - undefined — Indicates a private instance field (pre TS 4.3).
         *       - "f" — Indicates a private field (instance or static).
         *       - "m" — Indicates a private method (instance or static).
         *       - "a" — Indicates a private accessor (instance or static).
         *   @param f — (optional pre TS 4.3) Depends on the arguments for state and kind:
         *       - If kind is "m", this should be the function corresponding to the static or instance method.
         *       - If kind is "a", this should be the function corresponding to the setter method, or undefined if the setter was not defined.
         *       - If kind is "f" and state is a function, this should be an object holding the value of a static field, or undefined if the static field declaration has not yet been evaluated.
         * Usage:
         * This helper will only ever be used by the compiler in the following ways:
         *
         * Writing to a private instance field (pre TS 4.3):
         *      __classPrivateFieldSet(<any>, <WeakMap>, <any>)
         *
         * Writing to a private instance field (TS 4.3+):
         *      __classPrivateFieldSet(<any>, <WeakMap>, <any>, "f")
         *
         * Writing to a private instance set accessor (when defined, TS 4.3+):
         *      __classPrivateFieldSet(<any>, <WeakSet>, <any>, "a", <function>)
         *
         * Writing to a private instance set accessor (when not defined, TS 4.3+):
         *      __classPrivateFieldSet(<any>, <WeakSet>, <any>, "a", void 0)
         *      NOTE: This always results in a runtime error.
         *
         * Writing to a private instance method (TS 4.3+):
         *      __classPrivateFieldSet(<any>, <WeakSet>, <any>, "m", <function>)
         *      NOTE: This always results in a runtime error.
         *
         * Writing to a private static field (TS 4.3+):
         *      __classPrivateFieldSet(<any>, <constructor>, <any>, "f", <{ value: any }>)
         *
         * Writing to a private static set accessor (when defined, TS 4.3+):
         *      __classPrivateFieldSet(<any>, <constructor>, <any>, "a", <function>)
         *
         * Writing to a private static set accessor (when not defined, TS 4.3+):
         *      __classPrivateFieldSet(<any>, <constructor>, <any>, "a", void 0)
         *      NOTE: This always results in a runtime error.
         *
         * Writing to a private static method (TS 4.3+):
         *      __classPrivateFieldSet(<any>, <constructor>, <any>, "m", <function>)
         *      NOTE: This always results in a runtime error.
         */
        ts.classPrivateFieldSetHelper = {
            name: "typescript:classPrivateFieldSet",
            importName: "__classPrivateFieldSet",
            scoped: false,
            text: "\n            var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {\n                if (kind === \"m\") throw new TypeError(\"Private method is not writable\");\n                if (kind === \"a\" && !f) throw new TypeError(\"Private accessor was defined without a setter\");\n                if (typeof state === \"function\" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError(\"Cannot write private member to an object whose class did not declare it\");\n                return (kind === \"a\" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;\n            };"
        };
        var allUnscopedEmitHelpers;
        function getAllUnscopedEmitHelpers() {
            return allUnscopedEmitHelpers || (allUnscopedEmitHelpers = ts.arrayToMap([
                ts.decorateHelper,
                ts.metadataHelper,
                ts.paramHelper,
                ts.assignHelper,
                ts.awaitHelper,
                ts.asyncGeneratorHelper,
                ts.asyncDelegator,
                ts.asyncValues,
                ts.restHelper,
                ts.awaiterHelper,
                ts.extendsHelper,
                ts.templateObjectHelper,
                ts.spreadArrayHelper,
                ts.valuesHelper,
                ts.readHelper,
                ts.generatorHelper,
                ts.importStarHelper,
                ts.importDefaultHelper,
                ts.exportStarHelper,
                ts.classPrivateFieldGetHelper,
                ts.classPrivateFieldSetHelper,
                ts.createBindingHelper,
                ts.setModuleDefaultHelper
            ], function (helper) { return helper.name; }));
        }
        ts.getAllUnscopedEmitHelpers = getAllUnscopedEmitHelpers;
        ts.asyncSuperHelper = {
            name: "typescript:async-super",
            scoped: true,
            text: helperString(__makeTemplateObject(["\n            const ", " = name => super[name];"], ["\n            const ", " = name => super[name];"]), "_superIndex")
        };
        ts.advancedAsyncSuperHelper = {
            name: "typescript:advanced-async-super",
            scoped: true,
            text: helperString(__makeTemplateObject(["\n            const ", " = (function (geti, seti) {\n                const cache = Object.create(null);\n                return name => cache[name] || (cache[name] = { get value() { return geti(name); }, set value(v) { seti(name, v); } });\n            })(name => super[name], (name, value) => super[name] = value);"], ["\n            const ", " = (function (geti, seti) {\n                const cache = Object.create(null);\n                return name => cache[name] || (cache[name] = { get value() { return geti(name); }, set value(v) { seti(name, v); } });\n            })(name => super[name], (name, value) => super[name] = value);"]), "_superIndex")
        };
        function isCallToHelper(firstSegment, helperName) {
            return ts.isCallExpression(firstSegment)
                && ts.isIdentifier(firstSegment.expression)
                && (ts.getEmitFlags(firstSegment.expression) & 4096 /* HelperName */) !== 0
                && firstSegment.expression.escapedText === helperName;
        }
        ts.isCallToHelper = isCallToHelper;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        // Literals
        function isNumericLiteral(node) {
            return node.kind === 8 /* NumericLiteral */;
        }
        ts.isNumericLiteral = isNumericLiteral;
        function isBigIntLiteral(node) {
            return node.kind === 9 /* BigIntLiteral */;
        }
        ts.isBigIntLiteral = isBigIntLiteral;
        function isStringLiteral(node) {
            return node.kind === 10 /* StringLiteral */;
        }
        ts.isStringLiteral = isStringLiteral;
        function isJsxText(node) {
            return node.kind === 11 /* JsxText */;
        }
        ts.isJsxText = isJsxText;
        function isRegularExpressionLiteral(node) {
            return node.kind === 13 /* RegularExpressionLiteral */;
        }
        ts.isRegularExpressionLiteral = isRegularExpressionLiteral;
        function isNoSubstitutionTemplateLiteral(node) {
            return node.kind === 14 /* NoSubstitutionTemplateLiteral */;
        }
        ts.isNoSubstitutionTemplateLiteral = isNoSubstitutionTemplateLiteral;
        // Pseudo-literals
        function isTemplateHead(node) {
            return node.kind === 15 /* TemplateHead */;
        }
        ts.isTemplateHead = isTemplateHead;
        function isTemplateMiddle(node) {
            return node.kind === 16 /* TemplateMiddle */;
        }
        ts.isTemplateMiddle = isTemplateMiddle;
        function isTemplateTail(node) {
            return node.kind === 17 /* TemplateTail */;
        }
        ts.isTemplateTail = isTemplateTail;
        // Punctuation
        function isDotDotDotToken(node) {
            return node.kind === 25 /* DotDotDotToken */;
        }
        ts.isDotDotDotToken = isDotDotDotToken;
        /*@internal*/
        function isCommaToken(node) {
            return node.kind === 27 /* CommaToken */;
        }
        ts.isCommaToken = isCommaToken;
        function isPlusToken(node) {
            return node.kind === 39 /* PlusToken */;
        }
        ts.isPlusToken = isPlusToken;
        function isMinusToken(node) {
            return node.kind === 40 /* MinusToken */;
        }
        ts.isMinusToken = isMinusToken;
        function isAsteriskToken(node) {
            return node.kind === 41 /* AsteriskToken */;
        }
        ts.isAsteriskToken = isAsteriskToken;
        /*@internal*/
        function isExclamationToken(node) {
            return node.kind === 53 /* ExclamationToken */;
        }
        ts.isExclamationToken = isExclamationToken;
        /*@internal*/
        function isQuestionToken(node) {
            return node.kind === 57 /* QuestionToken */;
        }
        ts.isQuestionToken = isQuestionToken;
        /*@internal*/
        function isColonToken(node) {
            return node.kind === 58 /* ColonToken */;
        }
        ts.isColonToken = isColonToken;
        /*@internal*/
        function isQuestionDotToken(node) {
            return node.kind === 28 /* QuestionDotToken */;
        }
        ts.isQuestionDotToken = isQuestionDotToken;
        /*@internal*/
        function isEqualsGreaterThanToken(node) {
            return node.kind === 38 /* EqualsGreaterThanToken */;
        }
        ts.isEqualsGreaterThanToken = isEqualsGreaterThanToken;
        // Identifiers
        function isIdentifier(node) {
            return node.kind === 79 /* Identifier */;
        }
        ts.isIdentifier = isIdentifier;
        function isPrivateIdentifier(node) {
            return node.kind === 80 /* PrivateIdentifier */;
        }
        ts.isPrivateIdentifier = isPrivateIdentifier;
        // Reserved Words
        /* @internal */
        function isExportModifier(node) {
            return node.kind === 93 /* ExportKeyword */;
        }
        ts.isExportModifier = isExportModifier;
        /* @internal */
        function isAsyncModifier(node) {
            return node.kind === 130 /* AsyncKeyword */;
        }
        ts.isAsyncModifier = isAsyncModifier;
        /* @internal */
        function isAssertsKeyword(node) {
            return node.kind === 128 /* AssertsKeyword */;
        }
        ts.isAssertsKeyword = isAssertsKeyword;
        /* @internal */
        function isAwaitKeyword(node) {
            return node.kind === 131 /* AwaitKeyword */;
        }
        ts.isAwaitKeyword = isAwaitKeyword;
        /* @internal */
        function isReadonlyKeyword(node) {
            return node.kind === 143 /* ReadonlyKeyword */;
        }
        ts.isReadonlyKeyword = isReadonlyKeyword;
        /* @internal */
        function isStaticModifier(node) {
            return node.kind === 124 /* StaticKeyword */;
        }
        ts.isStaticModifier = isStaticModifier;
        /* @internal */
        function isAbstractModifier(node) {
            return node.kind === 126 /* AbstractKeyword */;
        }
        ts.isAbstractModifier = isAbstractModifier;
        /*@internal*/
        function isSuperKeyword(node) {
            return node.kind === 106 /* SuperKeyword */;
        }
        ts.isSuperKeyword = isSuperKeyword;
        /*@internal*/
        function isImportKeyword(node) {
            return node.kind === 100 /* ImportKeyword */;
        }
        ts.isImportKeyword = isImportKeyword;
        // Names
        function isQualifiedName(node) {
            return node.kind === 159 /* QualifiedName */;
        }
        ts.isQualifiedName = isQualifiedName;
        function isComputedPropertyName(node) {
            return node.kind === 160 /* ComputedPropertyName */;
        }
        ts.isComputedPropertyName = isComputedPropertyName;
        // Signature elements
        function isTypeParameterDeclaration(node) {
            return node.kind === 161 /* TypeParameter */;
        }
        ts.isTypeParameterDeclaration = isTypeParameterDeclaration;
        // TODO(rbuckton): Rename to 'isParameterDeclaration'
        function isParameter(node) {
            return node.kind === 162 /* Parameter */;
        }
        ts.isParameter = isParameter;
        function isDecorator(node) {
            return node.kind === 163 /* Decorator */;
        }
        ts.isDecorator = isDecorator;
        // TypeMember
        function isPropertySignature(node) {
            return node.kind === 164 /* PropertySignature */;
        }
        ts.isPropertySignature = isPropertySignature;
        function isPropertyDeclaration(node) {
            return node.kind === 165 /* PropertyDeclaration */;
        }
        ts.isPropertyDeclaration = isPropertyDeclaration;
        function isMethodSignature(node) {
            return node.kind === 166 /* MethodSignature */;
        }
        ts.isMethodSignature = isMethodSignature;
        function isMethodDeclaration(node) {
            return node.kind === 167 /* MethodDeclaration */;
        }
        ts.isMethodDeclaration = isMethodDeclaration;
        function isClassStaticBlockDeclaration(node) {
            return node.kind === 168 /* ClassStaticBlockDeclaration */;
        }
        ts.isClassStaticBlockDeclaration = isClassStaticBlockDeclaration;
        function isConstructorDeclaration(node) {
            return node.kind === 169 /* Constructor */;
        }
        ts.isConstructorDeclaration = isConstructorDeclaration;
        function isGetAccessorDeclaration(node) {
            return node.kind === 170 /* GetAccessor */;
        }
        ts.isGetAccessorDeclaration = isGetAccessorDeclaration;
        function isSetAccessorDeclaration(node) {
            return node.kind === 171 /* SetAccessor */;
        }
        ts.isSetAccessorDeclaration = isSetAccessorDeclaration;
        function isCallSignatureDeclaration(node) {
            return node.kind === 172 /* CallSignature */;
        }
        ts.isCallSignatureDeclaration = isCallSignatureDeclaration;
        function isConstructSignatureDeclaration(node) {
            return node.kind === 173 /* ConstructSignature */;
        }
        ts.isConstructSignatureDeclaration = isConstructSignatureDeclaration;
        function isIndexSignatureDeclaration(node) {
            return node.kind === 174 /* IndexSignature */;
        }
        ts.isIndexSignatureDeclaration = isIndexSignatureDeclaration;
        // Type
        function isTypePredicateNode(node) {
            return node.kind === 175 /* TypePredicate */;
        }
        ts.isTypePredicateNode = isTypePredicateNode;
        function isTypeReferenceNode(node) {
            return node.kind === 176 /* TypeReference */;
        }
        ts.isTypeReferenceNode = isTypeReferenceNode;
        function isFunctionTypeNode(node) {
            return node.kind === 177 /* FunctionType */;
        }
        ts.isFunctionTypeNode = isFunctionTypeNode;
        function isConstructorTypeNode(node) {
            return node.kind === 178 /* ConstructorType */;
        }
        ts.isConstructorTypeNode = isConstructorTypeNode;
        function isTypeQueryNode(node) {
            return node.kind === 179 /* TypeQuery */;
        }
        ts.isTypeQueryNode = isTypeQueryNode;
        function isTypeLiteralNode(node) {
            return node.kind === 180 /* TypeLiteral */;
        }
        ts.isTypeLiteralNode = isTypeLiteralNode;
        function isArrayTypeNode(node) {
            return node.kind === 181 /* ArrayType */;
        }
        ts.isArrayTypeNode = isArrayTypeNode;
        function isTupleTypeNode(node) {
            return node.kind === 182 /* TupleType */;
        }
        ts.isTupleTypeNode = isTupleTypeNode;
        function isNamedTupleMember(node) {
            return node.kind === 195 /* NamedTupleMember */;
        }
        ts.isNamedTupleMember = isNamedTupleMember;
        function isOptionalTypeNode(node) {
            return node.kind === 183 /* OptionalType */;
        }
        ts.isOptionalTypeNode = isOptionalTypeNode;
        function isRestTypeNode(node) {
            return node.kind === 184 /* RestType */;
        }
        ts.isRestTypeNode = isRestTypeNode;
        function isUnionTypeNode(node) {
            return node.kind === 185 /* UnionType */;
        }
        ts.isUnionTypeNode = isUnionTypeNode;
        function isIntersectionTypeNode(node) {
            return node.kind === 186 /* IntersectionType */;
        }
        ts.isIntersectionTypeNode = isIntersectionTypeNode;
        function isConditionalTypeNode(node) {
            return node.kind === 187 /* ConditionalType */;
        }
        ts.isConditionalTypeNode = isConditionalTypeNode;
        function isInferTypeNode(node) {
            return node.kind === 188 /* InferType */;
        }
        ts.isInferTypeNode = isInferTypeNode;
        function isParenthesizedTypeNode(node) {
            return node.kind === 189 /* ParenthesizedType */;
        }
        ts.isParenthesizedTypeNode = isParenthesizedTypeNode;
        function isThisTypeNode(node) {
            return node.kind === 190 /* ThisType */;
        }
        ts.isThisTypeNode = isThisTypeNode;
        function isTypeOperatorNode(node) {
            return node.kind === 191 /* TypeOperator */;
        }
        ts.isTypeOperatorNode = isTypeOperatorNode;
        function isIndexedAccessTypeNode(node) {
            return node.kind === 192 /* IndexedAccessType */;
        }
        ts.isIndexedAccessTypeNode = isIndexedAccessTypeNode;
        function isMappedTypeNode(node) {
            return node.kind === 193 /* MappedType */;
        }
        ts.isMappedTypeNode = isMappedTypeNode;
        function isLiteralTypeNode(node) {
            return node.kind === 194 /* LiteralType */;
        }
        ts.isLiteralTypeNode = isLiteralTypeNode;
        function isImportTypeNode(node) {
            return node.kind === 198 /* ImportType */;
        }
        ts.isImportTypeNode = isImportTypeNode;
        function isTemplateLiteralTypeSpan(node) {
            return node.kind === 197 /* TemplateLiteralTypeSpan */;
        }
        ts.isTemplateLiteralTypeSpan = isTemplateLiteralTypeSpan;
        function isTemplateLiteralTypeNode(node) {
            return node.kind === 196 /* TemplateLiteralType */;
        }
        ts.isTemplateLiteralTypeNode = isTemplateLiteralTypeNode;
        // Binding patterns
        function isObjectBindingPattern(node) {
            return node.kind === 199 /* ObjectBindingPattern */;
        }
        ts.isObjectBindingPattern = isObjectBindingPattern;
        function isArrayBindingPattern(node) {
            return node.kind === 200 /* ArrayBindingPattern */;
        }
        ts.isArrayBindingPattern = isArrayBindingPattern;
        function isBindingElement(node) {
            return node.kind === 201 /* BindingElement */;
        }
        ts.isBindingElement = isBindingElement;
        // Expression
        function isArrayLiteralExpression(node) {
            return node.kind === 202 /* ArrayLiteralExpression */;
        }
        ts.isArrayLiteralExpression = isArrayLiteralExpression;
        function isObjectLiteralExpression(node) {
            return node.kind === 203 /* ObjectLiteralExpression */;
        }
        ts.isObjectLiteralExpression = isObjectLiteralExpression;
        function isPropertyAccessExpression(node) {
            return node.kind === 204 /* PropertyAccessExpression */;
        }
        ts.isPropertyAccessExpression = isPropertyAccessExpression;
        function isElementAccessExpression(node) {
            return node.kind === 205 /* ElementAccessExpression */;
        }
        ts.isElementAccessExpression = isElementAccessExpression;
        function isCallExpression(node) {
            return node.kind === 206 /* CallExpression */;
        }
        ts.isCallExpression = isCallExpression;
        function isNewExpression(node) {
            return node.kind === 207 /* NewExpression */;
        }
        ts.isNewExpression = isNewExpression;
        function isTaggedTemplateExpression(node) {
            return node.kind === 208 /* TaggedTemplateExpression */;
        }
        ts.isTaggedTemplateExpression = isTaggedTemplateExpression;
        function isTypeAssertionExpression(node) {
            return node.kind === 209 /* TypeAssertionExpression */;
        }
        ts.isTypeAssertionExpression = isTypeAssertionExpression;
        function isParenthesizedExpression(node) {
            return node.kind === 210 /* ParenthesizedExpression */;
        }
        ts.isParenthesizedExpression = isParenthesizedExpression;
        function isFunctionExpression(node) {
            return node.kind === 211 /* FunctionExpression */;
        }
        ts.isFunctionExpression = isFunctionExpression;
        function isArrowFunction(node) {
            return node.kind === 212 /* ArrowFunction */;
        }
        ts.isArrowFunction = isArrowFunction;
        function isDeleteExpression(node) {
            return node.kind === 213 /* DeleteExpression */;
        }
        ts.isDeleteExpression = isDeleteExpression;
        function isTypeOfExpression(node) {
            return node.kind === 214 /* TypeOfExpression */;
        }
        ts.isTypeOfExpression = isTypeOfExpression;
        function isVoidExpression(node) {
            return node.kind === 215 /* VoidExpression */;
        }
        ts.isVoidExpression = isVoidExpression;
        function isAwaitExpression(node) {
            return node.kind === 216 /* AwaitExpression */;
        }
        ts.isAwaitExpression = isAwaitExpression;
        function isPrefixUnaryExpression(node) {
            return node.kind === 217 /* PrefixUnaryExpression */;
        }
        ts.isPrefixUnaryExpression = isPrefixUnaryExpression;
        function isPostfixUnaryExpression(node) {
            return node.kind === 218 /* PostfixUnaryExpression */;
        }
        ts.isPostfixUnaryExpression = isPostfixUnaryExpression;
        function isBinaryExpression(node) {
            return node.kind === 219 /* BinaryExpression */;
        }
        ts.isBinaryExpression = isBinaryExpression;
        function isConditionalExpression(node) {
            return node.kind === 220 /* ConditionalExpression */;
        }
        ts.isConditionalExpression = isConditionalExpression;
        function isTemplateExpression(node) {
            return node.kind === 221 /* TemplateExpression */;
        }
        ts.isTemplateExpression = isTemplateExpression;
        function isYieldExpression(node) {
            return node.kind === 222 /* YieldExpression */;
        }
        ts.isYieldExpression = isYieldExpression;
        function isSpreadElement(node) {
            return node.kind === 223 /* SpreadElement */;
        }
        ts.isSpreadElement = isSpreadElement;
        function isClassExpression(node) {
            return node.kind === 224 /* ClassExpression */;
        }
        ts.isClassExpression = isClassExpression;
        function isOmittedExpression(node) {
            return node.kind === 225 /* OmittedExpression */;
        }
        ts.isOmittedExpression = isOmittedExpression;
        function isExpressionWithTypeArguments(node) {
            return node.kind === 226 /* ExpressionWithTypeArguments */;
        }
        ts.isExpressionWithTypeArguments = isExpressionWithTypeArguments;
        function isAsExpression(node) {
            return node.kind === 227 /* AsExpression */;
        }
        ts.isAsExpression = isAsExpression;
        function isNonNullExpression(node) {
            return node.kind === 228 /* NonNullExpression */;
        }
        ts.isNonNullExpression = isNonNullExpression;
        function isMetaProperty(node) {
            return node.kind === 229 /* MetaProperty */;
        }
        ts.isMetaProperty = isMetaProperty;
        function isSyntheticExpression(node) {
            return node.kind === 230 /* SyntheticExpression */;
        }
        ts.isSyntheticExpression = isSyntheticExpression;
        function isPartiallyEmittedExpression(node) {
            return node.kind === 345 /* PartiallyEmittedExpression */;
        }
        ts.isPartiallyEmittedExpression = isPartiallyEmittedExpression;
        function isCommaListExpression(node) {
            return node.kind === 346 /* CommaListExpression */;
        }
        ts.isCommaListExpression = isCommaListExpression;
        // Misc
        function isTemplateSpan(node) {
            return node.kind === 231 /* TemplateSpan */;
        }
        ts.isTemplateSpan = isTemplateSpan;
        function isSemicolonClassElement(node) {
            return node.kind === 232 /* SemicolonClassElement */;
        }
        ts.isSemicolonClassElement = isSemicolonClassElement;
        // Elements
        function isBlock(node) {
            return node.kind === 233 /* Block */;
        }
        ts.isBlock = isBlock;
        function isVariableStatement(node) {
            return node.kind === 235 /* VariableStatement */;
        }
        ts.isVariableStatement = isVariableStatement;
        function isEmptyStatement(node) {
            return node.kind === 234 /* EmptyStatement */;
        }
        ts.isEmptyStatement = isEmptyStatement;
        function isExpressionStatement(node) {
            return node.kind === 236 /* ExpressionStatement */;
        }
        ts.isExpressionStatement = isExpressionStatement;
        function isIfStatement(node) {
            return node.kind === 237 /* IfStatement */;
        }
        ts.isIfStatement = isIfStatement;
        function isDoStatement(node) {
            return node.kind === 238 /* DoStatement */;
        }
        ts.isDoStatement = isDoStatement;
        function isWhileStatement(node) {
            return node.kind === 239 /* WhileStatement */;
        }
        ts.isWhileStatement = isWhileStatement;
        function isForStatement(node) {
            return node.kind === 240 /* ForStatement */;
        }
        ts.isForStatement = isForStatement;
        function isForInStatement(node) {
            return node.kind === 241 /* ForInStatement */;
        }
        ts.isForInStatement = isForInStatement;
        function isForOfStatement(node) {
            return node.kind === 242 /* ForOfStatement */;
        }
        ts.isForOfStatement = isForOfStatement;
        function isContinueStatement(node) {
            return node.kind === 243 /* ContinueStatement */;
        }
        ts.isContinueStatement = isContinueStatement;
        function isBreakStatement(node) {
            return node.kind === 244 /* BreakStatement */;
        }
        ts.isBreakStatement = isBreakStatement;
        function isReturnStatement(node) {
            return node.kind === 245 /* ReturnStatement */;
        }
        ts.isReturnStatement = isReturnStatement;
        function isWithStatement(node) {
            return node.kind === 246 /* WithStatement */;
        }
        ts.isWithStatement = isWithStatement;
        function isSwitchStatement(node) {
            return node.kind === 247 /* SwitchStatement */;
        }
        ts.isSwitchStatement = isSwitchStatement;
        function isLabeledStatement(node) {
            return node.kind === 248 /* LabeledStatement */;
        }
        ts.isLabeledStatement = isLabeledStatement;
        function isThrowStatement(node) {
            return node.kind === 249 /* ThrowStatement */;
        }
        ts.isThrowStatement = isThrowStatement;
        function isTryStatement(node) {
            return node.kind === 250 /* TryStatement */;
        }
        ts.isTryStatement = isTryStatement;
        function isDebuggerStatement(node) {
            return node.kind === 251 /* DebuggerStatement */;
        }
        ts.isDebuggerStatement = isDebuggerStatement;
        function isVariableDeclaration(node) {
            return node.kind === 252 /* VariableDeclaration */;
        }
        ts.isVariableDeclaration = isVariableDeclaration;
        function isVariableDeclarationList(node) {
            return node.kind === 253 /* VariableDeclarationList */;
        }
        ts.isVariableDeclarationList = isVariableDeclarationList;
        function isFunctionDeclaration(node) {
            return node.kind === 254 /* FunctionDeclaration */;
        }
        ts.isFunctionDeclaration = isFunctionDeclaration;
        function isClassDeclaration(node) {
            return node.kind === 255 /* ClassDeclaration */;
        }
        ts.isClassDeclaration = isClassDeclaration;
        function isInterfaceDeclaration(node) {
            return node.kind === 256 /* InterfaceDeclaration */;
        }
        ts.isInterfaceDeclaration = isInterfaceDeclaration;
        function isTypeAliasDeclaration(node) {
            return node.kind === 257 /* TypeAliasDeclaration */;
        }
        ts.isTypeAliasDeclaration = isTypeAliasDeclaration;
        function isEnumDeclaration(node) {
            return node.kind === 258 /* EnumDeclaration */;
        }
        ts.isEnumDeclaration = isEnumDeclaration;
        function isModuleDeclaration(node) {
            return node.kind === 259 /* ModuleDeclaration */;
        }
        ts.isModuleDeclaration = isModuleDeclaration;
        function isModuleBlock(node) {
            return node.kind === 260 /* ModuleBlock */;
        }
        ts.isModuleBlock = isModuleBlock;
        function isCaseBlock(node) {
            return node.kind === 261 /* CaseBlock */;
        }
        ts.isCaseBlock = isCaseBlock;
        function isNamespaceExportDeclaration(node) {
            return node.kind === 262 /* NamespaceExportDeclaration */;
        }
        ts.isNamespaceExportDeclaration = isNamespaceExportDeclaration;
        function isImportEqualsDeclaration(node) {
            return node.kind === 263 /* ImportEqualsDeclaration */;
        }
        ts.isImportEqualsDeclaration = isImportEqualsDeclaration;
        function isImportDeclaration(node) {
            return node.kind === 264 /* ImportDeclaration */;
        }
        ts.isImportDeclaration = isImportDeclaration;
        function isImportClause(node) {
            return node.kind === 265 /* ImportClause */;
        }
        ts.isImportClause = isImportClause;
        function isNamespaceImport(node) {
            return node.kind === 266 /* NamespaceImport */;
        }
        ts.isNamespaceImport = isNamespaceImport;
        function isNamespaceExport(node) {
            return node.kind === 272 /* NamespaceExport */;
        }
        ts.isNamespaceExport = isNamespaceExport;
        function isNamedImports(node) {
            return node.kind === 267 /* NamedImports */;
        }
        ts.isNamedImports = isNamedImports;
        function isImportSpecifier(node) {
            return node.kind === 268 /* ImportSpecifier */;
        }
        ts.isImportSpecifier = isImportSpecifier;
        function isExportAssignment(node) {
            return node.kind === 269 /* ExportAssignment */;
        }
        ts.isExportAssignment = isExportAssignment;
        function isExportDeclaration(node) {
            return node.kind === 270 /* ExportDeclaration */;
        }
        ts.isExportDeclaration = isExportDeclaration;
        function isNamedExports(node) {
            return node.kind === 271 /* NamedExports */;
        }
        ts.isNamedExports = isNamedExports;
        function isExportSpecifier(node) {
            return node.kind === 273 /* ExportSpecifier */;
        }
        ts.isExportSpecifier = isExportSpecifier;
        function isMissingDeclaration(node) {
            return node.kind === 274 /* MissingDeclaration */;
        }
        ts.isMissingDeclaration = isMissingDeclaration;
        function isNotEmittedStatement(node) {
            return node.kind === 344 /* NotEmittedStatement */;
        }
        ts.isNotEmittedStatement = isNotEmittedStatement;
        /* @internal */
        function isSyntheticReference(node) {
            return node.kind === 349 /* SyntheticReferenceExpression */;
        }
        ts.isSyntheticReference = isSyntheticReference;
        /* @internal */
        function isMergeDeclarationMarker(node) {
            return node.kind === 347 /* MergeDeclarationMarker */;
        }
        ts.isMergeDeclarationMarker = isMergeDeclarationMarker;
        /* @internal */
        function isEndOfDeclarationMarker(node) {
            return node.kind === 348 /* EndOfDeclarationMarker */;
        }
        ts.isEndOfDeclarationMarker = isEndOfDeclarationMarker;
        // Module References
        function isExternalModuleReference(node) {
            return node.kind === 275 /* ExternalModuleReference */;
        }
        ts.isExternalModuleReference = isExternalModuleReference;
        // JSX
        function isJsxElement(node) {
            return node.kind === 276 /* JsxElement */;
        }
        ts.isJsxElement = isJsxElement;
        function isJsxSelfClosingElement(node) {
            return node.kind === 277 /* JsxSelfClosingElement */;
        }
        ts.isJsxSelfClosingElement = isJsxSelfClosingElement;
        function isJsxOpeningElement(node) {
            return node.kind === 278 /* JsxOpeningElement */;
        }
        ts.isJsxOpeningElement = isJsxOpeningElement;
        function isJsxClosingElement(node) {
            return node.kind === 279 /* JsxClosingElement */;
        }
        ts.isJsxClosingElement = isJsxClosingElement;
        function isJsxFragment(node) {
            return node.kind === 280 /* JsxFragment */;
        }
        ts.isJsxFragment = isJsxFragment;
        function isJsxOpeningFragment(node) {
            return node.kind === 281 /* JsxOpeningFragment */;
        }
        ts.isJsxOpeningFragment = isJsxOpeningFragment;
        function isJsxClosingFragment(node) {
            return node.kind === 282 /* JsxClosingFragment */;
        }
        ts.isJsxClosingFragment = isJsxClosingFragment;
        function isJsxAttribute(node) {
            return node.kind === 283 /* JsxAttribute */;
        }
        ts.isJsxAttribute = isJsxAttribute;
        function isJsxAttributes(node) {
            return node.kind === 284 /* JsxAttributes */;
        }
        ts.isJsxAttributes = isJsxAttributes;
        function isJsxSpreadAttribute(node) {
            return node.kind === 285 /* JsxSpreadAttribute */;
        }
        ts.isJsxSpreadAttribute = isJsxSpreadAttribute;
        function isJsxExpression(node) {
            return node.kind === 286 /* JsxExpression */;
        }
        ts.isJsxExpression = isJsxExpression;
        // Clauses
        function isCaseClause(node) {
            return node.kind === 287 /* CaseClause */;
        }
        ts.isCaseClause = isCaseClause;
        function isDefaultClause(node) {
            return node.kind === 288 /* DefaultClause */;
        }
        ts.isDefaultClause = isDefaultClause;
        function isHeritageClause(node) {
            return node.kind === 289 /* HeritageClause */;
        }
        ts.isHeritageClause = isHeritageClause;
        function isCatchClause(node) {
            return node.kind === 290 /* CatchClause */;
        }
        ts.isCatchClause = isCatchClause;
        // Property assignments
        function isPropertyAssignment(node) {
            return node.kind === 291 /* PropertyAssignment */;
        }
        ts.isPropertyAssignment = isPropertyAssignment;
        function isShorthandPropertyAssignment(node) {
            return node.kind === 292 /* ShorthandPropertyAssignment */;
        }
        ts.isShorthandPropertyAssignment = isShorthandPropertyAssignment;
        function isSpreadAssignment(node) {
            return node.kind === 293 /* SpreadAssignment */;
        }
        ts.isSpreadAssignment = isSpreadAssignment;
        // Enum
        function isEnumMember(node) {
            return node.kind === 294 /* EnumMember */;
        }
        ts.isEnumMember = isEnumMember;
        // Unparsed
        // TODO(rbuckton): isUnparsedPrologue
        function isUnparsedPrepend(node) {
            return node.kind === 296 /* UnparsedPrepend */;
        }
        ts.isUnparsedPrepend = isUnparsedPrepend;
        // TODO(rbuckton): isUnparsedText
        // TODO(rbuckton): isUnparsedInternalText
        // TODO(rbuckton): isUnparsedSyntheticReference
        // Top-level nodes
        function isSourceFile(node) {
            return node.kind === 300 /* SourceFile */;
        }
        ts.isSourceFile = isSourceFile;
        function isBundle(node) {
            return node.kind === 301 /* Bundle */;
        }
        ts.isBundle = isBundle;
        function isUnparsedSource(node) {
            return node.kind === 302 /* UnparsedSource */;
        }
        ts.isUnparsedSource = isUnparsedSource;
        // TODO(rbuckton): isInputFiles
        // JSDoc Elements
        function isJSDocTypeExpression(node) {
            return node.kind === 304 /* JSDocTypeExpression */;
        }
        ts.isJSDocTypeExpression = isJSDocTypeExpression;
        function isJSDocNameReference(node) {
            return node.kind === 305 /* JSDocNameReference */;
        }
        ts.isJSDocNameReference = isJSDocNameReference;
        function isJSDocMemberName(node) {
            return node.kind === 306 /* JSDocMemberName */;
        }
        ts.isJSDocMemberName = isJSDocMemberName;
        function isJSDocLink(node) {
            return node.kind === 319 /* JSDocLink */;
        }
        ts.isJSDocLink = isJSDocLink;
        function isJSDocLinkCode(node) {
            return node.kind === 320 /* JSDocLinkCode */;
        }
        ts.isJSDocLinkCode = isJSDocLinkCode;
        function isJSDocLinkPlain(node) {
            return node.kind === 321 /* JSDocLinkPlain */;
        }
        ts.isJSDocLinkPlain = isJSDocLinkPlain;
        function isJSDocAllType(node) {
            return node.kind === 307 /* JSDocAllType */;
        }
        ts.isJSDocAllType = isJSDocAllType;
        function isJSDocUnknownType(node) {
            return node.kind === 308 /* JSDocUnknownType */;
        }
        ts.isJSDocUnknownType = isJSDocUnknownType;
        function isJSDocNullableType(node) {
            return node.kind === 309 /* JSDocNullableType */;
        }
        ts.isJSDocNullableType = isJSDocNullableType;
        function isJSDocNonNullableType(node) {
            return node.kind === 310 /* JSDocNonNullableType */;
        }
        ts.isJSDocNonNullableType = isJSDocNonNullableType;
        function isJSDocOptionalType(node) {
            return node.kind === 311 /* JSDocOptionalType */;
        }
        ts.isJSDocOptionalType = isJSDocOptionalType;
        function isJSDocFunctionType(node) {
            return node.kind === 312 /* JSDocFunctionType */;
        }
        ts.isJSDocFunctionType = isJSDocFunctionType;
        function isJSDocVariadicType(node) {
            return node.kind === 313 /* JSDocVariadicType */;
        }
        ts.isJSDocVariadicType = isJSDocVariadicType;
        function isJSDocNamepathType(node) {
            return node.kind === 314 /* JSDocNamepathType */;
        }
        ts.isJSDocNamepathType = isJSDocNamepathType;
        function isJSDoc(node) {
            return node.kind === 315 /* JSDocComment */;
        }
        ts.isJSDoc = isJSDoc;
        function isJSDocTypeLiteral(node) {
            return node.kind === 317 /* JSDocTypeLiteral */;
        }
        ts.isJSDocTypeLiteral = isJSDocTypeLiteral;
        function isJSDocSignature(node) {
            return node.kind === 318 /* JSDocSignature */;
        }
        ts.isJSDocSignature = isJSDocSignature;
        // JSDoc Tags
        function isJSDocAugmentsTag(node) {
            return node.kind === 323 /* JSDocAugmentsTag */;
        }
        ts.isJSDocAugmentsTag = isJSDocAugmentsTag;
        function isJSDocAuthorTag(node) {
            return node.kind === 325 /* JSDocAuthorTag */;
        }
        ts.isJSDocAuthorTag = isJSDocAuthorTag;
        function isJSDocClassTag(node) {
            return node.kind === 327 /* JSDocClassTag */;
        }
        ts.isJSDocClassTag = isJSDocClassTag;
        function isJSDocCallbackTag(node) {
            return node.kind === 333 /* JSDocCallbackTag */;
        }
        ts.isJSDocCallbackTag = isJSDocCallbackTag;
        function isJSDocPublicTag(node) {
            return node.kind === 328 /* JSDocPublicTag */;
        }
        ts.isJSDocPublicTag = isJSDocPublicTag;
        function isJSDocPrivateTag(node) {
            return node.kind === 329 /* JSDocPrivateTag */;
        }
        ts.isJSDocPrivateTag = isJSDocPrivateTag;
        function isJSDocProtectedTag(node) {
            return node.kind === 330 /* JSDocProtectedTag */;
        }
        ts.isJSDocProtectedTag = isJSDocProtectedTag;
        function isJSDocReadonlyTag(node) {
            return node.kind === 331 /* JSDocReadonlyTag */;
        }
        ts.isJSDocReadonlyTag = isJSDocReadonlyTag;
        function isJSDocOverrideTag(node) {
            return node.kind === 332 /* JSDocOverrideTag */;
        }
        ts.isJSDocOverrideTag = isJSDocOverrideTag;
        function isJSDocDeprecatedTag(node) {
            return node.kind === 326 /* JSDocDeprecatedTag */;
        }
        ts.isJSDocDeprecatedTag = isJSDocDeprecatedTag;
        function isJSDocSeeTag(node) {
            return node.kind === 341 /* JSDocSeeTag */;
        }
        ts.isJSDocSeeTag = isJSDocSeeTag;
        function isJSDocEnumTag(node) {
            return node.kind === 334 /* JSDocEnumTag */;
        }
        ts.isJSDocEnumTag = isJSDocEnumTag;
        function isJSDocParameterTag(node) {
            return node.kind === 335 /* JSDocParameterTag */;
        }
        ts.isJSDocParameterTag = isJSDocParameterTag;
        function isJSDocReturnTag(node) {
            return node.kind === 336 /* JSDocReturnTag */;
        }
        ts.isJSDocReturnTag = isJSDocReturnTag;
        function isJSDocThisTag(node) {
            return node.kind === 337 /* JSDocThisTag */;
        }
        ts.isJSDocThisTag = isJSDocThisTag;
        function isJSDocTypeTag(node) {
            return node.kind === 338 /* JSDocTypeTag */;
        }
        ts.isJSDocTypeTag = isJSDocTypeTag;
        function isJSDocTemplateTag(node) {
            return node.kind === 339 /* JSDocTemplateTag */;
        }
        ts.isJSDocTemplateTag = isJSDocTemplateTag;
        function isJSDocTypedefTag(node) {
            return node.kind === 340 /* JSDocTypedefTag */;
        }
        ts.isJSDocTypedefTag = isJSDocTypedefTag;
        function isJSDocUnknownTag(node) {
            return node.kind === 322 /* JSDocTag */;
        }
        ts.isJSDocUnknownTag = isJSDocUnknownTag;
        function isJSDocPropertyTag(node) {
            return node.kind === 342 /* JSDocPropertyTag */;
        }
        ts.isJSDocPropertyTag = isJSDocPropertyTag;
        function isJSDocImplementsTag(node) {
            return node.kind === 324 /* JSDocImplementsTag */;
        }
        ts.isJSDocImplementsTag = isJSDocImplementsTag;
        // Synthesized list
        /* @internal */
        function isSyntaxList(n) {
            return n.kind === 343 /* SyntaxList */;
        }
        ts.isSyntaxList = isSyntaxList;
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        // Compound nodes
        function createEmptyExports(factory) {
            return factory.createExportDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*isTypeOnly*/ false, factory.createNamedExports([]), /*moduleSpecifier*/ undefined);
        }
        ts.createEmptyExports = createEmptyExports;
        function createMemberAccessForPropertyName(factory, target, memberName, location) {
            if (ts.isComputedPropertyName(memberName)) {
                return ts.setTextRange(factory.createElementAccessExpression(target, memberName.expression), location);
            }
            else {
                var expression = ts.setTextRange(ts.isMemberName(memberName)
                    ? factory.createPropertyAccessExpression(target, memberName)
                    : factory.createElementAccessExpression(target, memberName), memberName);
                ts.getOrCreateEmitNode(expression).flags |= 64 /* NoNestedSourceMaps */;
                return expression;
            }
        }
        ts.createMemberAccessForPropertyName = createMemberAccessForPropertyName;
        function createReactNamespace(reactNamespace, parent) {
            // To ensure the emit resolver can properly resolve the namespace, we need to
            // treat this identifier as if it were a source tree node by clearing the `Synthesized`
            // flag and setting a parent node.
            var react = ts.parseNodeFactory.createIdentifier(reactNamespace || "React");
            // Set the parent that is in parse tree
            // this makes sure that parent chain is intact for checker to traverse complete scope tree
            ts.setParent(react, ts.getParseTreeNode(parent));
            return react;
        }
        function createJsxFactoryExpressionFromEntityName(factory, jsxFactory, parent) {
            if (ts.isQualifiedName(jsxFactory)) {
                var left = createJsxFactoryExpressionFromEntityName(factory, jsxFactory.left, parent);
                var right = factory.createIdentifier(ts.idText(jsxFactory.right));
                right.escapedText = jsxFactory.right.escapedText;
                return factory.createPropertyAccessExpression(left, right);
            }
            else {
                return createReactNamespace(ts.idText(jsxFactory), parent);
            }
        }
        function createJsxFactoryExpression(factory, jsxFactoryEntity, reactNamespace, parent) {
            return jsxFactoryEntity ?
                createJsxFactoryExpressionFromEntityName(factory, jsxFactoryEntity, parent) :
                factory.createPropertyAccessExpression(createReactNamespace(reactNamespace, parent), "createElement");
        }
        ts.createJsxFactoryExpression = createJsxFactoryExpression;
        function createJsxFragmentFactoryExpression(factory, jsxFragmentFactoryEntity, reactNamespace, parent) {
            return jsxFragmentFactoryEntity ?
                createJsxFactoryExpressionFromEntityName(factory, jsxFragmentFactoryEntity, parent) :
                factory.createPropertyAccessExpression(createReactNamespace(reactNamespace, parent), "Fragment");
        }
        function createExpressionForJsxElement(factory, callee, tagName, props, children, location) {
            var argumentsList = [tagName];
            if (props) {
                argumentsList.push(props);
            }
            if (children && children.length > 0) {
                if (!props) {
                    argumentsList.push(factory.createNull());
                }
                if (children.length > 1) {
                    for (var _i = 0, children_3 = children; _i < children_3.length; _i++) {
                        var child = children_3[_i];
                        startOnNewLine(child);
                        argumentsList.push(child);
                    }
                }
                else {
                    argumentsList.push(children[0]);
                }
            }
            return ts.setTextRange(factory.createCallExpression(callee, 
            /*typeArguments*/ undefined, argumentsList), location);
        }
        ts.createExpressionForJsxElement = createExpressionForJsxElement;
        function createExpressionForJsxFragment(factory, jsxFactoryEntity, jsxFragmentFactoryEntity, reactNamespace, children, parentElement, location) {
            var tagName = createJsxFragmentFactoryExpression(factory, jsxFragmentFactoryEntity, reactNamespace, parentElement);
            var argumentsList = [tagName, factory.createNull()];
            if (children && children.length > 0) {
                if (children.length > 1) {
                    for (var _i = 0, children_4 = children; _i < children_4.length; _i++) {
                        var child = children_4[_i];
                        startOnNewLine(child);
                        argumentsList.push(child);
                    }
                }
                else {
                    argumentsList.push(children[0]);
                }
            }
            return ts.setTextRange(factory.createCallExpression(createJsxFactoryExpression(factory, jsxFactoryEntity, reactNamespace, parentElement), 
            /*typeArguments*/ undefined, argumentsList), location);
        }
        ts.createExpressionForJsxFragment = createExpressionForJsxFragment;
        // Utilities
        function createForOfBindingStatement(factory, node, boundValue) {
            if (ts.isVariableDeclarationList(node)) {
                var firstDeclaration = ts.first(node.declarations);
                var updatedDeclaration = factory.updateVariableDeclaration(firstDeclaration, firstDeclaration.name, 
                /*exclamationToken*/ undefined, 
                /*type*/ undefined, boundValue);
                return ts.setTextRange(factory.createVariableStatement(
                /*modifiers*/ undefined, factory.updateVariableDeclarationList(node, [updatedDeclaration])), 
                /*location*/ node);
            }
            else {
                var updatedExpression = ts.setTextRange(factory.createAssignment(node, boundValue), /*location*/ node);
                return ts.setTextRange(factory.createExpressionStatement(updatedExpression), /*location*/ node);
            }
        }
        ts.createForOfBindingStatement = createForOfBindingStatement;
        function insertLeadingStatement(factory, dest, source) {
            if (ts.isBlock(dest)) {
                return factory.updateBlock(dest, ts.setTextRange(factory.createNodeArray(__spreadArray([source], dest.statements, true)), dest.statements));
            }
            else {
                return factory.createBlock(factory.createNodeArray([dest, source]), /*multiLine*/ true);
            }
        }
        ts.insertLeadingStatement = insertLeadingStatement;
        function createExpressionFromEntityName(factory, node) {
            if (ts.isQualifiedName(node)) {
                var left = createExpressionFromEntityName(factory, node.left);
                // TODO(rbuckton): Does this need to be parented?
                var right = ts.setParent(ts.setTextRange(factory.cloneNode(node.right), node.right), node.right.parent);
                return ts.setTextRange(factory.createPropertyAccessExpression(left, right), node);
            }
            else {
                // TODO(rbuckton): Does this need to be parented?
                return ts.setParent(ts.setTextRange(factory.cloneNode(node), node), node.parent);
            }
        }
        ts.createExpressionFromEntityName = createExpressionFromEntityName;
        function createExpressionForPropertyName(factory, memberName) {
            if (ts.isIdentifier(memberName)) {
                return factory.createStringLiteralFromNode(memberName);
            }
            else if (ts.isComputedPropertyName(memberName)) {
                // TODO(rbuckton): Does this need to be parented?
                return ts.setParent(ts.setTextRange(factory.cloneNode(memberName.expression), memberName.expression), memberName.expression.parent);
            }
            else {
                // TODO(rbuckton): Does this need to be parented?
                return ts.setParent(ts.setTextRange(factory.cloneNode(memberName), memberName), memberName.parent);
            }
        }
        ts.createExpressionForPropertyName = createExpressionForPropertyName;
        function createExpressionForAccessorDeclaration(factory, properties, property, receiver, multiLine) {
            var _a = ts.getAllAccessorDeclarations(properties, property), firstAccessor = _a.firstAccessor, getAccessor = _a.getAccessor, setAccessor = _a.setAccessor;
            if (property === firstAccessor) {
                return ts.setTextRange(factory.createObjectDefinePropertyCall(receiver, createExpressionForPropertyName(factory, property.name), factory.createPropertyDescriptor({
                    enumerable: factory.createFalse(),
                    configurable: true,
                    get: getAccessor && ts.setTextRange(ts.setOriginalNode(factory.createFunctionExpression(getAccessor.modifiers, 
                    /*asteriskToken*/ undefined, 
                    /*name*/ undefined, 
                    /*typeParameters*/ undefined, getAccessor.parameters, 
                    /*type*/ undefined, getAccessor.body // TODO: GH#18217
                    ), getAccessor), getAccessor),
                    set: setAccessor && ts.setTextRange(ts.setOriginalNode(factory.createFunctionExpression(setAccessor.modifiers, 
                    /*asteriskToken*/ undefined, 
                    /*name*/ undefined, 
                    /*typeParameters*/ undefined, setAccessor.parameters, 
                    /*type*/ undefined, setAccessor.body // TODO: GH#18217
                    ), setAccessor), setAccessor)
                }, !multiLine)), firstAccessor);
            }
            return undefined;
        }
        function createExpressionForPropertyAssignment(factory, property, receiver) {
            return ts.setOriginalNode(ts.setTextRange(factory.createAssignment(createMemberAccessForPropertyName(factory, receiver, property.name, /*location*/ property.name), property.initializer), property), property);
        }
        function createExpressionForShorthandPropertyAssignment(factory, property, receiver) {
            return ts.setOriginalNode(ts.setTextRange(factory.createAssignment(createMemberAccessForPropertyName(factory, receiver, property.name, /*location*/ property.name), factory.cloneNode(property.name)), 
            /*location*/ property), 
            /*original*/ property);
        }
        function createExpressionForMethodDeclaration(factory, method, receiver) {
            return ts.setOriginalNode(ts.setTextRange(factory.createAssignment(createMemberAccessForPropertyName(factory, receiver, method.name, /*location*/ method.name), ts.setOriginalNode(ts.setTextRange(factory.createFunctionExpression(method.modifiers, method.asteriskToken, 
            /*name*/ undefined, 
            /*typeParameters*/ undefined, method.parameters, 
            /*type*/ undefined, method.body // TODO: GH#18217
            ), 
            /*location*/ method), 
            /*original*/ method)), 
            /*location*/ method), 
            /*original*/ method);
        }
        function createExpressionForObjectLiteralElementLike(factory, node, property, receiver) {
            if (property.name && ts.isPrivateIdentifier(property.name)) {
                ts.Debug.failBadSyntaxKind(property.name, "Private identifiers are not allowed in object literals.");
            }
            switch (property.kind) {
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                    return createExpressionForAccessorDeclaration(factory, node.properties, property, receiver, !!node.multiLine);
                case 291 /* PropertyAssignment */:
                    return createExpressionForPropertyAssignment(factory, property, receiver);
                case 292 /* ShorthandPropertyAssignment */:
                    return createExpressionForShorthandPropertyAssignment(factory, property, receiver);
                case 167 /* MethodDeclaration */:
                    return createExpressionForMethodDeclaration(factory, property, receiver);
            }
        }
        ts.createExpressionForObjectLiteralElementLike = createExpressionForObjectLiteralElementLike;
        /**
         * Expand the read and increment/decrement operations a pre- or post-increment or pre- or post-decrement expression.
         *
         * ```ts
         * // input
         * <expression>++
         * // output (if result is not discarded)
         * var <temp>;
         * (<temp> = <expression>, <resultVariable> = <temp>++, <temp>)
         * // output (if result is discarded)
         * var <temp>;
         * (<temp> = <expression>, <temp>++, <temp>)
         *
         * // input
         * ++<expression>
         * // output (if result is not discarded)
         * var <temp>;
         * (<temp> = <expression>, <resultVariable> = ++<temp>)
         * // output (if result is discarded)
         * var <temp>;
         * (<temp> = <expression>, ++<temp>)
         * ```
         *
         * It is up to the caller to supply a temporary variable for `<resultVariable>` if one is needed.
         * The temporary variable `<temp>` is injected so that `++` and `--` work uniformly with `number` and `bigint`.
         * The result of the expression is always the final result of incrementing or decrementing the expression, so that it can be used for storage.
         *
         * @param factory {@link NodeFactory} used to create the expanded representation.
         * @param node The original prefix or postfix unary node.
         * @param expression The expression to use as the value to increment or decrement
         * @param resultVariable A temporary variable in which to store the result. Pass `undefined` if the result is discarded, or if the value of `<temp>` is the expected result.
         */
        function expandPreOrPostfixIncrementOrDecrementExpression(factory, node, expression, recordTempVariable, resultVariable) {
            var operator = node.operator;
            ts.Debug.assert(operator === 45 /* PlusPlusToken */ || operator === 46 /* MinusMinusToken */, "Expected 'node' to be a pre- or post-increment or pre- or post-decrement expression");
            var temp = factory.createTempVariable(recordTempVariable);
            expression = factory.createAssignment(temp, expression);
            ts.setTextRange(expression, node.operand);
            var operation = ts.isPrefixUnaryExpression(node) ?
                factory.createPrefixUnaryExpression(operator, temp) :
                factory.createPostfixUnaryExpression(temp, operator);
            ts.setTextRange(operation, node);
            if (resultVariable) {
                operation = factory.createAssignment(resultVariable, operation);
                ts.setTextRange(operation, node);
            }
            expression = factory.createComma(expression, operation);
            ts.setTextRange(expression, node);
            if (ts.isPostfixUnaryExpression(node)) {
                expression = factory.createComma(expression, temp);
                ts.setTextRange(expression, node);
            }
            return expression;
        }
        ts.expandPreOrPostfixIncrementOrDecrementExpression = expandPreOrPostfixIncrementOrDecrementExpression;
        /**
         * Gets whether an identifier should only be referred to by its internal name.
         */
        function isInternalName(node) {
            return (ts.getEmitFlags(node) & 32768 /* InternalName */) !== 0;
        }
        ts.isInternalName = isInternalName;
        /**
         * Gets whether an identifier should only be referred to by its local name.
         */
        function isLocalName(node) {
            return (ts.getEmitFlags(node) & 16384 /* LocalName */) !== 0;
        }
        ts.isLocalName = isLocalName;
        /**
         * Gets whether an identifier should only be referred to by its export representation if the
         * name points to an exported symbol.
         */
        function isExportName(node) {
            return (ts.getEmitFlags(node) & 8192 /* ExportName */) !== 0;
        }
        ts.isExportName = isExportName;
        function isUseStrictPrologue(node) {
            return ts.isStringLiteral(node.expression) && node.expression.text === "use strict";
        }
        function findUseStrictPrologue(statements) {
            for (var _i = 0, statements_1 = statements; _i < statements_1.length; _i++) {
                var statement = statements_1[_i];
                if (ts.isPrologueDirective(statement)) {
                    if (isUseStrictPrologue(statement)) {
                        return statement;
                    }
                }
                else {
                    break;
                }
            }
            return undefined;
        }
        ts.findUseStrictPrologue = findUseStrictPrologue;
        function startsWithUseStrict(statements) {
            var firstStatement = ts.firstOrUndefined(statements);
            return firstStatement !== undefined
                && ts.isPrologueDirective(firstStatement)
                && isUseStrictPrologue(firstStatement);
        }
        ts.startsWithUseStrict = startsWithUseStrict;
        function isCommaSequence(node) {
            return node.kind === 219 /* BinaryExpression */ && node.operatorToken.kind === 27 /* CommaToken */ ||
                node.kind === 346 /* CommaListExpression */;
        }
        ts.isCommaSequence = isCommaSequence;
        function isOuterExpression(node, kinds) {
            if (kinds === void 0) {
                kinds = 15 /* All */;
            }
            switch (node.kind) {
                case 210 /* ParenthesizedExpression */:
                    return (kinds & 1 /* Parentheses */) !== 0;
                case 209 /* TypeAssertionExpression */:
                case 227 /* AsExpression */:
                    return (kinds & 2 /* TypeAssertions */) !== 0;
                case 228 /* NonNullExpression */:
                    return (kinds & 4 /* NonNullAssertions */) !== 0;
                case 345 /* PartiallyEmittedExpression */:
                    return (kinds & 8 /* PartiallyEmittedExpressions */) !== 0;
            }
            return false;
        }
        ts.isOuterExpression = isOuterExpression;
        function skipOuterExpressions(node, kinds) {
            if (kinds === void 0) {
                kinds = 15 /* All */;
            }
            while (isOuterExpression(node, kinds)) {
                node = node.expression;
            }
            return node;
        }
        ts.skipOuterExpressions = skipOuterExpressions;
        function skipAssertions(node) {
            return skipOuterExpressions(node, 6 /* Assertions */);
        }
        ts.skipAssertions = skipAssertions;
        function startOnNewLine(node) {
            return ts.setStartsOnNewLine(node, /*newLine*/ true);
        }
        ts.startOnNewLine = startOnNewLine;
        function getExternalHelpersModuleName(node) {
            var parseNode = ts.getOriginalNode(node, ts.isSourceFile);
            var emitNode = parseNode && parseNode.emitNode;
            return emitNode && emitNode.externalHelpersModuleName;
        }
        ts.getExternalHelpersModuleName = getExternalHelpersModuleName;
        function hasRecordedExternalHelpers(sourceFile) {
            var parseNode = ts.getOriginalNode(sourceFile, ts.isSourceFile);
            var emitNode = parseNode && parseNode.emitNode;
            return !!emitNode && (!!emitNode.externalHelpersModuleName || !!emitNode.externalHelpers);
        }
        ts.hasRecordedExternalHelpers = hasRecordedExternalHelpers;
        function createExternalHelpersImportDeclarationIfNeeded(nodeFactory, helperFactory, sourceFile, compilerOptions, hasExportStarsToExportValues, hasImportStar, hasImportDefault) {
            if (compilerOptions.importHelpers && ts.isEffectiveExternalModule(sourceFile, compilerOptions)) {
                var namedBindings = void 0;
                var moduleKind = ts.getEmitModuleKind(compilerOptions);
                if (moduleKind >= ts.ModuleKind.ES2015 && moduleKind <= ts.ModuleKind.ESNext) {
                    // use named imports
                    var helpers = ts.getEmitHelpers(sourceFile);
                    if (helpers) {
                        var helperNames = [];
                        for (var _i = 0, helpers_3 = helpers; _i < helpers_3.length; _i++) {
                            var helper = helpers_3[_i];
                            if (!helper.scoped) {
                                var importName = helper.importName;
                                if (importName) {
                                    ts.pushIfUnique(helperNames, importName);
                                }
                            }
                        }
                        if (ts.some(helperNames)) {
                            helperNames.sort(ts.compareStringsCaseSensitive);
                            // Alias the imports if the names are used somewhere in the file.
                            // NOTE: We don't need to care about global import collisions as this is a module.
                            namedBindings = nodeFactory.createNamedImports(ts.map(helperNames, function (name) {
                                return ts.isFileLevelUniqueName(sourceFile, name)
                                    ? nodeFactory.createImportSpecifier(/*propertyName*/ undefined, nodeFactory.createIdentifier(name))
                                    : nodeFactory.createImportSpecifier(nodeFactory.createIdentifier(name), helperFactory.getUnscopedHelperName(name));
                            }));
                            var parseNode = ts.getOriginalNode(sourceFile, ts.isSourceFile);
                            var emitNode = ts.getOrCreateEmitNode(parseNode);
                            emitNode.externalHelpers = true;
                        }
                    }
                }
                else {
                    // use a namespace import
                    var externalHelpersModuleName = getOrCreateExternalHelpersModuleNameIfNeeded(nodeFactory, sourceFile, compilerOptions, hasExportStarsToExportValues, hasImportStar || hasImportDefault);
                    if (externalHelpersModuleName) {
                        namedBindings = nodeFactory.createNamespaceImport(externalHelpersModuleName);
                    }
                }
                if (namedBindings) {
                    var externalHelpersImportDeclaration = nodeFactory.createImportDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, nodeFactory.createImportClause(/*isTypeOnly*/ false, /*name*/ undefined, namedBindings), nodeFactory.createStringLiteral(ts.externalHelpersModuleNameText));
                    ts.addEmitFlags(externalHelpersImportDeclaration, 67108864 /* NeverApplyImportHelper */);
                    return externalHelpersImportDeclaration;
                }
            }
        }
        ts.createExternalHelpersImportDeclarationIfNeeded = createExternalHelpersImportDeclarationIfNeeded;
        function getOrCreateExternalHelpersModuleNameIfNeeded(factory, node, compilerOptions, hasExportStarsToExportValues, hasImportStarOrImportDefault) {
            if (compilerOptions.importHelpers && ts.isEffectiveExternalModule(node, compilerOptions)) {
                var externalHelpersModuleName = getExternalHelpersModuleName(node);
                if (externalHelpersModuleName) {
                    return externalHelpersModuleName;
                }
                var moduleKind = ts.getEmitModuleKind(compilerOptions);
                var create = (hasExportStarsToExportValues || (compilerOptions.esModuleInterop && hasImportStarOrImportDefault))
                    && moduleKind !== ts.ModuleKind.System
                    && moduleKind < ts.ModuleKind.ES2015;
                if (!create) {
                    var helpers = ts.getEmitHelpers(node);
                    if (helpers) {
                        for (var _i = 0, helpers_4 = helpers; _i < helpers_4.length; _i++) {
                            var helper = helpers_4[_i];
                            if (!helper.scoped) {
                                create = true;
                                break;
                            }
                        }
                    }
                }
                if (create) {
                    var parseNode = ts.getOriginalNode(node, ts.isSourceFile);
                    var emitNode = ts.getOrCreateEmitNode(parseNode);
                    return emitNode.externalHelpersModuleName || (emitNode.externalHelpersModuleName = factory.createUniqueName(ts.externalHelpersModuleNameText));
                }
            }
        }
        ts.getOrCreateExternalHelpersModuleNameIfNeeded = getOrCreateExternalHelpersModuleNameIfNeeded;
        /**
         * Get the name of that target module from an import or export declaration
         */
        function getLocalNameForExternalImport(factory, node, sourceFile) {
            var namespaceDeclaration = ts.getNamespaceDeclarationNode(node);
            if (namespaceDeclaration && !ts.isDefaultImport(node) && !ts.isExportNamespaceAsDefaultDeclaration(node)) {
                var name = namespaceDeclaration.name;
                return ts.isGeneratedIdentifier(name) ? name : factory.createIdentifier(ts.getSourceTextOfNodeFromSourceFile(sourceFile, name) || ts.idText(name));
            }
            if (node.kind === 264 /* ImportDeclaration */ && node.importClause) {
                return factory.getGeneratedNameForNode(node);
            }
            if (node.kind === 270 /* ExportDeclaration */ && node.moduleSpecifier) {
                return factory.getGeneratedNameForNode(node);
            }
            return undefined;
        }
        ts.getLocalNameForExternalImport = getLocalNameForExternalImport;
        /**
         * Get the name of a target module from an import/export declaration as should be written in the emitted output.
         * The emitted output name can be different from the input if:
         *  1. The module has a /// <amd-module name="<new name>" />
         *  2. --out or --outFile is used, making the name relative to the rootDir
         *  3- The containing SourceFile has an entry in renamedDependencies for the import as requested by some module loaders (e.g. System).
         * Otherwise, a new StringLiteral node representing the module name will be returned.
         */
        function getExternalModuleNameLiteral(factory, importNode, sourceFile, host, resolver, compilerOptions) {
            var moduleName = ts.getExternalModuleName(importNode);
            if (moduleName && ts.isStringLiteral(moduleName)) {
                return tryGetModuleNameFromDeclaration(importNode, host, factory, resolver, compilerOptions)
                    || tryRenameExternalModule(factory, moduleName, sourceFile)
                    || factory.cloneNode(moduleName);
            }
            return undefined;
        }
        ts.getExternalModuleNameLiteral = getExternalModuleNameLiteral;
        /**
         * Some bundlers (SystemJS builder) sometimes want to rename dependencies.
         * Here we check if alternative name was provided for a given moduleName and return it if possible.
         */
        function tryRenameExternalModule(factory, moduleName, sourceFile) {
            var rename = sourceFile.renamedDependencies && sourceFile.renamedDependencies.get(moduleName.text);
            return rename ? factory.createStringLiteral(rename) : undefined;
        }
        /**
         * Get the name of a module as should be written in the emitted output.
         * The emitted output name can be different from the input if:
         *  1. The module has a /// <amd-module name="<new name>" />
         *  2. --out or --outFile is used, making the name relative to the rootDir
         * Otherwise, a new StringLiteral node representing the module name will be returned.
         */
        function tryGetModuleNameFromFile(factory, file, host, options) {
            if (!file) {
                return undefined;
            }
            if (file.moduleName) {
                return factory.createStringLiteral(file.moduleName);
            }
            if (!file.isDeclarationFile && ts.outFile(options)) {
                return factory.createStringLiteral(ts.getExternalModuleNameFromPath(host, file.fileName));
            }
            return undefined;
        }
        ts.tryGetModuleNameFromFile = tryGetModuleNameFromFile;
        function tryGetModuleNameFromDeclaration(declaration, host, factory, resolver, compilerOptions) {
            return tryGetModuleNameFromFile(factory, resolver.getExternalModuleFileFromDeclaration(declaration), host, compilerOptions);
        }
        /**
         * Gets the initializer of an BindingOrAssignmentElement.
         */
        function getInitializerOfBindingOrAssignmentElement(bindingElement) {
            if (ts.isDeclarationBindingElement(bindingElement)) {
                // `1` in `let { a = 1 } = ...`
                // `1` in `let { a: b = 1 } = ...`
                // `1` in `let { a: {b} = 1 } = ...`
                // `1` in `let { a: [b] = 1 } = ...`
                // `1` in `let [a = 1] = ...`
                // `1` in `let [{a} = 1] = ...`
                // `1` in `let [[a] = 1] = ...`
                return bindingElement.initializer;
            }
            if (ts.isPropertyAssignment(bindingElement)) {
                // `1` in `({ a: b = 1 } = ...)`
                // `1` in `({ a: {b} = 1 } = ...)`
                // `1` in `({ a: [b] = 1 } = ...)`
                var initializer = bindingElement.initializer;
                return ts.isAssignmentExpression(initializer, /*excludeCompoundAssignment*/ true)
                    ? initializer.right
                    : undefined;
            }
            if (ts.isShorthandPropertyAssignment(bindingElement)) {
                // `1` in `({ a = 1 } = ...)`
                return bindingElement.objectAssignmentInitializer;
            }
            if (ts.isAssignmentExpression(bindingElement, /*excludeCompoundAssignment*/ true)) {
                // `1` in `[a = 1] = ...`
                // `1` in `[{a} = 1] = ...`
                // `1` in `[[a] = 1] = ...`
                return bindingElement.right;
            }
            if (ts.isSpreadElement(bindingElement)) {
                // Recovery consistent with existing emit.
                return getInitializerOfBindingOrAssignmentElement(bindingElement.expression);
            }
        }
        ts.getInitializerOfBindingOrAssignmentElement = getInitializerOfBindingOrAssignmentElement;
        /**
         * Gets the name of an BindingOrAssignmentElement.
         */
        function getTargetOfBindingOrAssignmentElement(bindingElement) {
            if (ts.isDeclarationBindingElement(bindingElement)) {
                // `a` in `let { a } = ...`
                // `a` in `let { a = 1 } = ...`
                // `b` in `let { a: b } = ...`
                // `b` in `let { a: b = 1 } = ...`
                // `a` in `let { ...a } = ...`
                // `{b}` in `let { a: {b} } = ...`
                // `{b}` in `let { a: {b} = 1 } = ...`
                // `[b]` in `let { a: [b] } = ...`
                // `[b]` in `let { a: [b] = 1 } = ...`
                // `a` in `let [a] = ...`
                // `a` in `let [a = 1] = ...`
                // `a` in `let [...a] = ...`
                // `{a}` in `let [{a}] = ...`
                // `{a}` in `let [{a} = 1] = ...`
                // `[a]` in `let [[a]] = ...`
                // `[a]` in `let [[a] = 1] = ...`
                return bindingElement.name;
            }
            if (ts.isObjectLiteralElementLike(bindingElement)) {
                switch (bindingElement.kind) {
                    case 291 /* PropertyAssignment */:
                        // `b` in `({ a: b } = ...)`
                        // `b` in `({ a: b = 1 } = ...)`
                        // `{b}` in `({ a: {b} } = ...)`
                        // `{b}` in `({ a: {b} = 1 } = ...)`
                        // `[b]` in `({ a: [b] } = ...)`
                        // `[b]` in `({ a: [b] = 1 } = ...)`
                        // `b.c` in `({ a: b.c } = ...)`
                        // `b.c` in `({ a: b.c = 1 } = ...)`
                        // `b[0]` in `({ a: b[0] } = ...)`
                        // `b[0]` in `({ a: b[0] = 1 } = ...)`
                        return getTargetOfBindingOrAssignmentElement(bindingElement.initializer);
                    case 292 /* ShorthandPropertyAssignment */:
                        // `a` in `({ a } = ...)`
                        // `a` in `({ a = 1 } = ...)`
                        return bindingElement.name;
                    case 293 /* SpreadAssignment */:
                        // `a` in `({ ...a } = ...)`
                        return getTargetOfBindingOrAssignmentElement(bindingElement.expression);
                }
                // no target
                return undefined;
            }
            if (ts.isAssignmentExpression(bindingElement, /*excludeCompoundAssignment*/ true)) {
                // `a` in `[a = 1] = ...`
                // `{a}` in `[{a} = 1] = ...`
                // `[a]` in `[[a] = 1] = ...`
                // `a.b` in `[a.b = 1] = ...`
                // `a[0]` in `[a[0] = 1] = ...`
                return getTargetOfBindingOrAssignmentElement(bindingElement.left);
            }
            if (ts.isSpreadElement(bindingElement)) {
                // `a` in `[...a] = ...`
                return getTargetOfBindingOrAssignmentElement(bindingElement.expression);
            }
            // `a` in `[a] = ...`
            // `{a}` in `[{a}] = ...`
            // `[a]` in `[[a]] = ...`
            // `a.b` in `[a.b] = ...`
            // `a[0]` in `[a[0]] = ...`
            return bindingElement;
        }
        ts.getTargetOfBindingOrAssignmentElement = getTargetOfBindingOrAssignmentElement;
        /**
         * Determines whether an BindingOrAssignmentElement is a rest element.
         */
        function getRestIndicatorOfBindingOrAssignmentElement(bindingElement) {
            switch (bindingElement.kind) {
                case 162 /* Parameter */:
                case 201 /* BindingElement */:
                    // `...` in `let [...a] = ...`
                    return bindingElement.dotDotDotToken;
                case 223 /* SpreadElement */:
                case 293 /* SpreadAssignment */:
                    // `...` in `[...a] = ...`
                    return bindingElement;
            }
            return undefined;
        }
        ts.getRestIndicatorOfBindingOrAssignmentElement = getRestIndicatorOfBindingOrAssignmentElement;
        /**
         * Gets the property name of a BindingOrAssignmentElement
         */
        function getPropertyNameOfBindingOrAssignmentElement(bindingElement) {
            var propertyName = tryGetPropertyNameOfBindingOrAssignmentElement(bindingElement);
            ts.Debug.assert(!!propertyName || ts.isSpreadAssignment(bindingElement), "Invalid property name for binding element.");
            return propertyName;
        }
        ts.getPropertyNameOfBindingOrAssignmentElement = getPropertyNameOfBindingOrAssignmentElement;
        function tryGetPropertyNameOfBindingOrAssignmentElement(bindingElement) {
            switch (bindingElement.kind) {
                case 201 /* BindingElement */:
                    // `a` in `let { a: b } = ...`
                    // `[a]` in `let { [a]: b } = ...`
                    // `"a"` in `let { "a": b } = ...`
                    // `1` in `let { 1: b } = ...`
                    if (bindingElement.propertyName) {
                        var propertyName = bindingElement.propertyName;
                        if (ts.isPrivateIdentifier(propertyName)) {
                            return ts.Debug.failBadSyntaxKind(propertyName);
                        }
                        return ts.isComputedPropertyName(propertyName) && isStringOrNumericLiteral(propertyName.expression)
                            ? propertyName.expression
                            : propertyName;
                    }
                    break;
                case 291 /* PropertyAssignment */:
                    // `a` in `({ a: b } = ...)`
                    // `[a]` in `({ [a]: b } = ...)`
                    // `"a"` in `({ "a": b } = ...)`
                    // `1` in `({ 1: b } = ...)`
                    if (bindingElement.name) {
                        var propertyName = bindingElement.name;
                        if (ts.isPrivateIdentifier(propertyName)) {
                            return ts.Debug.failBadSyntaxKind(propertyName);
                        }
                        return ts.isComputedPropertyName(propertyName) && isStringOrNumericLiteral(propertyName.expression)
                            ? propertyName.expression
                            : propertyName;
                    }
                    break;
                case 293 /* SpreadAssignment */:
                    // `a` in `({ ...a } = ...)`
                    if (bindingElement.name && ts.isPrivateIdentifier(bindingElement.name)) {
                        return ts.Debug.failBadSyntaxKind(bindingElement.name);
                    }
                    return bindingElement.name;
            }
            var target = getTargetOfBindingOrAssignmentElement(bindingElement);
            if (target && ts.isPropertyName(target)) {
                return target;
            }
        }
        ts.tryGetPropertyNameOfBindingOrAssignmentElement = tryGetPropertyNameOfBindingOrAssignmentElement;
        function isStringOrNumericLiteral(node) {
            var kind = node.kind;
            return kind === 10 /* StringLiteral */
                || kind === 8 /* NumericLiteral */;
        }
        /**
         * Gets the elements of a BindingOrAssignmentPattern
         */
        function getElementsOfBindingOrAssignmentPattern(name) {
            switch (name.kind) {
                case 199 /* ObjectBindingPattern */:
                case 200 /* ArrayBindingPattern */:
                case 202 /* ArrayLiteralExpression */:
                    // `a` in `{a}`
                    // `a` in `[a]`
                    return name.elements;
                case 203 /* ObjectLiteralExpression */:
                    // `a` in `{a}`
                    return name.properties;
            }
        }
        ts.getElementsOfBindingOrAssignmentPattern = getElementsOfBindingOrAssignmentPattern;
        /* @internal */
        function getJSDocTypeAliasName(fullName) {
            if (fullName) {
                var rightNode = fullName;
                while (true) {
                    if (ts.isIdentifier(rightNode) || !rightNode.body) {
                        return ts.isIdentifier(rightNode) ? rightNode : rightNode.name;
                    }
                    rightNode = rightNode.body;
                }
            }
        }
        ts.getJSDocTypeAliasName = getJSDocTypeAliasName;
        function canHaveModifiers(node) {
            var kind = node.kind;
            return kind === 162 /* Parameter */
                || kind === 164 /* PropertySignature */
                || kind === 165 /* PropertyDeclaration */
                || kind === 166 /* MethodSignature */
                || kind === 167 /* MethodDeclaration */
                || kind === 169 /* Constructor */
                || kind === 170 /* GetAccessor */
                || kind === 171 /* SetAccessor */
                || kind === 174 /* IndexSignature */
                || kind === 211 /* FunctionExpression */
                || kind === 212 /* ArrowFunction */
                || kind === 224 /* ClassExpression */
                || kind === 235 /* VariableStatement */
                || kind === 254 /* FunctionDeclaration */
                || kind === 255 /* ClassDeclaration */
                || kind === 256 /* InterfaceDeclaration */
                || kind === 257 /* TypeAliasDeclaration */
                || kind === 258 /* EnumDeclaration */
                || kind === 259 /* ModuleDeclaration */
                || kind === 263 /* ImportEqualsDeclaration */
                || kind === 264 /* ImportDeclaration */
                || kind === 269 /* ExportAssignment */
                || kind === 270 /* ExportDeclaration */;
        }
        ts.canHaveModifiers = canHaveModifiers;
        ts.isTypeNodeOrTypeParameterDeclaration = ts.or(ts.isTypeNode, ts.isTypeParameterDeclaration);
        ts.isQuestionOrExclamationToken = ts.or(ts.isQuestionToken, ts.isExclamationToken);
        ts.isIdentifierOrThisTypeNode = ts.or(ts.isIdentifier, ts.isThisTypeNode);
        ts.isReadonlyKeywordOrPlusOrMinusToken = ts.or(ts.isReadonlyKeyword, ts.isPlusToken, ts.isMinusToken);
        ts.isQuestionOrPlusOrMinusToken = ts.or(ts.isQuestionToken, ts.isPlusToken, ts.isMinusToken);
        ts.isModuleName = ts.or(ts.isIdentifier, ts.isStringLiteral);
        function isLiteralTypeLikeExpression(node) {
            var kind = node.kind;
            return kind === 104 /* NullKeyword */
                || kind === 110 /* TrueKeyword */
                || kind === 95 /* FalseKeyword */
                || ts.isLiteralExpression(node)
                || ts.isPrefixUnaryExpression(node);
        }
        ts.isLiteralTypeLikeExpression = isLiteralTypeLikeExpression;
        function isExponentiationOperator(kind) {
            return kind === 42 /* AsteriskAsteriskToken */;
        }
        function isMultiplicativeOperator(kind) {
            return kind === 41 /* AsteriskToken */
                || kind === 43 /* SlashToken */
                || kind === 44 /* PercentToken */;
        }
        function isMultiplicativeOperatorOrHigher(kind) {
            return isExponentiationOperator(kind)
                || isMultiplicativeOperator(kind);
        }
        function isAdditiveOperator(kind) {
            return kind === 39 /* PlusToken */
                || kind === 40 /* MinusToken */;
        }
        function isAdditiveOperatorOrHigher(kind) {
            return isAdditiveOperator(kind)
                || isMultiplicativeOperatorOrHigher(kind);
        }
        function isShiftOperator(kind) {
            return kind === 47 /* LessThanLessThanToken */
                || kind === 48 /* GreaterThanGreaterThanToken */
                || kind === 49 /* GreaterThanGreaterThanGreaterThanToken */;
        }
        function isShiftOperatorOrHigher(kind) {
            return isShiftOperator(kind)
                || isAdditiveOperatorOrHigher(kind);
        }
        function isRelationalOperator(kind) {
            return kind === 29 /* LessThanToken */
                || kind === 32 /* LessThanEqualsToken */
                || kind === 31 /* GreaterThanToken */
                || kind === 33 /* GreaterThanEqualsToken */
                || kind === 102 /* InstanceOfKeyword */
                || kind === 101 /* InKeyword */;
        }
        function isRelationalOperatorOrHigher(kind) {
            return isRelationalOperator(kind)
                || isShiftOperatorOrHigher(kind);
        }
        function isEqualityOperator(kind) {
            return kind === 34 /* EqualsEqualsToken */
                || kind === 36 /* EqualsEqualsEqualsToken */
                || kind === 35 /* ExclamationEqualsToken */
                || kind === 37 /* ExclamationEqualsEqualsToken */;
        }
        function isEqualityOperatorOrHigher(kind) {
            return isEqualityOperator(kind)
                || isRelationalOperatorOrHigher(kind);
        }
        function isBitwiseOperator(kind) {
            return kind === 50 /* AmpersandToken */
                || kind === 51 /* BarToken */
                || kind === 52 /* CaretToken */;
        }
        function isBitwiseOperatorOrHigher(kind) {
            return isBitwiseOperator(kind)
                || isEqualityOperatorOrHigher(kind);
        }
        // NOTE: The version in utilities includes ExclamationToken, which is not a binary operator.
        function isLogicalOperator(kind) {
            return kind === 55 /* AmpersandAmpersandToken */
                || kind === 56 /* BarBarToken */;
        }
        function isLogicalOperatorOrHigher(kind) {
            return isLogicalOperator(kind)
                || isBitwiseOperatorOrHigher(kind);
        }
        function isAssignmentOperatorOrHigher(kind) {
            return kind === 60 /* QuestionQuestionToken */
                || isLogicalOperatorOrHigher(kind)
                || ts.isAssignmentOperator(kind);
        }
        function isBinaryOperator(kind) {
            return isAssignmentOperatorOrHigher(kind)
                || kind === 27 /* CommaToken */;
        }
        function isBinaryOperatorToken(node) {
            return isBinaryOperator(node.kind);
        }
        ts.isBinaryOperatorToken = isBinaryOperatorToken;
        var BinaryExpressionState;
        (function (BinaryExpressionState) {
            /**
             * Handles walking into a `BinaryExpression`.
             * @param machine State machine handler functions
             * @param frame The current frame
             * @returns The new frame
             */
            function enter(machine, stackIndex, stateStack, nodeStack, userStateStack, _resultHolder, outerState) {
                var prevUserState = stackIndex > 0 ? userStateStack[stackIndex - 1] : undefined;
                ts.Debug.assertEqual(stateStack[stackIndex], enter);
                userStateStack[stackIndex] = machine.onEnter(nodeStack[stackIndex], prevUserState, outerState);
                stateStack[stackIndex] = nextState(machine, enter);
                return stackIndex;
            }
            BinaryExpressionState.enter = enter;
            /**
             * Handles walking the `left` side of a `BinaryExpression`.
             * @param machine State machine handler functions
             * @param frame The current frame
             * @returns The new frame
             */
            function left(machine, stackIndex, stateStack, nodeStack, userStateStack, _resultHolder, _outerState) {
                ts.Debug.assertEqual(stateStack[stackIndex], left);
                ts.Debug.assertIsDefined(machine.onLeft);
                stateStack[stackIndex] = nextState(machine, left);
                var nextNode = machine.onLeft(nodeStack[stackIndex].left, userStateStack[stackIndex], nodeStack[stackIndex]);
                if (nextNode) {
                    checkCircularity(stackIndex, nodeStack, nextNode);
                    return pushStack(stackIndex, stateStack, nodeStack, userStateStack, nextNode);
                }
                return stackIndex;
            }
            BinaryExpressionState.left = left;
            /**
             * Handles walking the `operatorToken` of a `BinaryExpression`.
             * @param machine State machine handler functions
             * @param frame The current frame
             * @returns The new frame
             */
            function operator(machine, stackIndex, stateStack, nodeStack, userStateStack, _resultHolder, _outerState) {
                ts.Debug.assertEqual(stateStack[stackIndex], operator);
                ts.Debug.assertIsDefined(machine.onOperator);
                stateStack[stackIndex] = nextState(machine, operator);
                machine.onOperator(nodeStack[stackIndex].operatorToken, userStateStack[stackIndex], nodeStack[stackIndex]);
                return stackIndex;
            }
            BinaryExpressionState.operator = operator;
            /**
             * Handles walking the `right` side of a `BinaryExpression`.
             * @param machine State machine handler functions
             * @param frame The current frame
             * @returns The new frame
             */
            function right(machine, stackIndex, stateStack, nodeStack, userStateStack, _resultHolder, _outerState) {
                ts.Debug.assertEqual(stateStack[stackIndex], right);
                ts.Debug.assertIsDefined(machine.onRight);
                stateStack[stackIndex] = nextState(machine, right);
                var nextNode = machine.onRight(nodeStack[stackIndex].right, userStateStack[stackIndex], nodeStack[stackIndex]);
                if (nextNode) {
                    checkCircularity(stackIndex, nodeStack, nextNode);
                    return pushStack(stackIndex, stateStack, nodeStack, userStateStack, nextNode);
                }
                return stackIndex;
            }
            BinaryExpressionState.right = right;
            /**
             * Handles walking out of a `BinaryExpression`.
             * @param machine State machine handler functions
             * @param frame The current frame
             * @returns The new frame
             */
            function exit(machine, stackIndex, stateStack, nodeStack, userStateStack, resultHolder, _outerState) {
                ts.Debug.assertEqual(stateStack[stackIndex], exit);
                stateStack[stackIndex] = nextState(machine, exit);
                var result = machine.onExit(nodeStack[stackIndex], userStateStack[stackIndex]);
                if (stackIndex > 0) {
                    stackIndex--;
                    if (machine.foldState) {
                        var side = stateStack[stackIndex] === exit ? "right" : "left";
                        userStateStack[stackIndex] = machine.foldState(userStateStack[stackIndex], result, side);
                    }
                }
                else {
                    resultHolder.value = result;
                }
                return stackIndex;
            }
            BinaryExpressionState.exit = exit;
            /**
             * Handles a frame that is already done.
             * @returns The `done` state.
             */
            function done(_machine, stackIndex, stateStack, _nodeStack, _userStateStack, _resultHolder, _outerState) {
                ts.Debug.assertEqual(stateStack[stackIndex], done);
                return stackIndex;
            }
            BinaryExpressionState.done = done;
            function nextState(machine, currentState) {
                switch (currentState) {
                    case enter:
                        if (machine.onLeft)
                            return left;
                    // falls through
                    case left:
                        if (machine.onOperator)
                            return operator;
                    // falls through
                    case operator:
                        if (machine.onRight)
                            return right;
                    // falls through
                    case right: return exit;
                    case exit: return done;
                    case done: return done;
                    default: ts.Debug.fail("Invalid state");
                }
            }
            BinaryExpressionState.nextState = nextState;
            function pushStack(stackIndex, stateStack, nodeStack, userStateStack, node) {
                stackIndex++;
                stateStack[stackIndex] = enter;
                nodeStack[stackIndex] = node;
                userStateStack[stackIndex] = undefined;
                return stackIndex;
            }
            function checkCircularity(stackIndex, nodeStack, node) {
                if (ts.Debug.shouldAssert(2 /* Aggressive */)) {
                    while (stackIndex >= 0) {
                        ts.Debug.assert(nodeStack[stackIndex] !== node, "Circular traversal detected.");
                        stackIndex--;
                    }
                }
            }
        })(BinaryExpressionState || (BinaryExpressionState = {}));
        /**
         * Holds state machine handler functions
         */
        var BinaryExpressionStateMachine = /** @class */ (function () {
            function BinaryExpressionStateMachine(onEnter, onLeft, onOperator, onRight, onExit, foldState) {
                this.onEnter = onEnter;
                this.onLeft = onLeft;
                this.onOperator = onOperator;
                this.onRight = onRight;
                this.onExit = onExit;
                this.foldState = foldState;
            }
            return BinaryExpressionStateMachine;
        }());
        function createBinaryExpressionTrampoline(onEnter, onLeft, onOperator, onRight, onExit, foldState) {
            var machine = new BinaryExpressionStateMachine(onEnter, onLeft, onOperator, onRight, onExit, foldState);
            return trampoline;
            function trampoline(node, outerState) {
                var resultHolder = { value: undefined };
                var stateStack = [BinaryExpressionState.enter];
                var nodeStack = [node];
                var userStateStack = [undefined];
                var stackIndex = 0;
                while (stateStack[stackIndex] !== BinaryExpressionState.done) {
                    stackIndex = stateStack[stackIndex](machine, stackIndex, stateStack, nodeStack, userStateStack, resultHolder, outerState);
                }
                ts.Debug.assertEqual(stackIndex, 0);
                return resultHolder.value;
            }
        }
        ts.createBinaryExpressionTrampoline = createBinaryExpressionTrampoline;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        function setTextRange(range, location) {
            return location ? ts.setTextRangePosEnd(range, location.pos, location.end) : range;
        }
        ts.setTextRange = setTextRange;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        var SignatureFlags;
        (function (SignatureFlags) {
            SignatureFlags[SignatureFlags["None"] = 0] = "None";
            SignatureFlags[SignatureFlags["Yield"] = 1] = "Yield";
            SignatureFlags[SignatureFlags["Await"] = 2] = "Await";
            SignatureFlags[SignatureFlags["Type"] = 4] = "Type";
            SignatureFlags[SignatureFlags["IgnoreMissingOpenBrace"] = 16] = "IgnoreMissingOpenBrace";
            SignatureFlags[SignatureFlags["JSDoc"] = 32] = "JSDoc";
        })(SignatureFlags || (SignatureFlags = {}));
        var SpeculationKind;
        (function (SpeculationKind) {
            SpeculationKind[SpeculationKind["TryParse"] = 0] = "TryParse";
            SpeculationKind[SpeculationKind["Lookahead"] = 1] = "Lookahead";
            SpeculationKind[SpeculationKind["Reparse"] = 2] = "Reparse";
        })(SpeculationKind || (SpeculationKind = {}));
        var NodeConstructor;
        var TokenConstructor;
        var IdentifierConstructor;
        var PrivateIdentifierConstructor;
        var SourceFileConstructor;
        /**
         * NOTE: You should not use this, it is only exported to support `createNode` in `~/src/deprecatedCompat/deprecations.ts`.
         */
        /* @internal */
        ts.parseBaseNodeFactory = {
            createBaseSourceFileNode: function (kind) { return new (SourceFileConstructor || (SourceFileConstructor = ts.objectAllocator.getSourceFileConstructor()))(kind, -1, -1); },
            createBaseIdentifierNode: function (kind) { return new (IdentifierConstructor || (IdentifierConstructor = ts.objectAllocator.getIdentifierConstructor()))(kind, -1, -1); },
            createBasePrivateIdentifierNode: function (kind) { return new (PrivateIdentifierConstructor || (PrivateIdentifierConstructor = ts.objectAllocator.getPrivateIdentifierConstructor()))(kind, -1, -1); },
            createBaseTokenNode: function (kind) { return new (TokenConstructor || (TokenConstructor = ts.objectAllocator.getTokenConstructor()))(kind, -1, -1); },
            createBaseNode: function (kind) { return new (NodeConstructor || (NodeConstructor = ts.objectAllocator.getNodeConstructor()))(kind, -1, -1); },
        };
        /* @internal */
        ts.parseNodeFactory = ts.createNodeFactory(1 /* NoParenthesizerRules */, ts.parseBaseNodeFactory);
        function visitNode(cbNode, node) {
            return node && cbNode(node);
        }
        function visitNodes(cbNode, cbNodes, nodes) {
            if (nodes) {
                if (cbNodes) {
                    return cbNodes(nodes);
                }
                for (var _i = 0, nodes_1 = nodes; _i < nodes_1.length; _i++) {
                    var node = nodes_1[_i];
                    var result = cbNode(node);
                    if (result) {
                        return result;
                    }
                }
            }
        }
        /*@internal*/
        function isJSDocLikeText(text, start) {
            return text.charCodeAt(start + 1) === 42 /* asterisk */ &&
                text.charCodeAt(start + 2) === 42 /* asterisk */ &&
                text.charCodeAt(start + 3) !== 47 /* slash */;
        }
        ts.isJSDocLikeText = isJSDocLikeText;
        /**
         * Invokes a callback for each child of the given node. The 'cbNode' callback is invoked for all child nodes
         * stored in properties. If a 'cbNodes' callback is specified, it is invoked for embedded arrays; otherwise,
         * embedded arrays are flattened and the 'cbNode' callback is invoked for each element. If a callback returns
         * a truthy value, iteration stops and that value is returned. Otherwise, undefined is returned.
         *
         * @param node a given node to visit its children
         * @param cbNode a callback to be invoked for all child nodes
         * @param cbNodes a callback to be invoked for embedded array
         *
         * @remarks `forEachChild` must visit the children of a node in the order
         * that they appear in the source code. The language service depends on this property to locate nodes by position.
         */
        function forEachChild(node, cbNode, cbNodes) {
            if (!node || node.kind <= 158 /* LastToken */) {
                return;
            }
            switch (node.kind) {
                case 159 /* QualifiedName */:
                    return visitNode(cbNode, node.left) ||
                        visitNode(cbNode, node.right);
                case 161 /* TypeParameter */:
                    return visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.constraint) ||
                        visitNode(cbNode, node.default) ||
                        visitNode(cbNode, node.expression);
                case 292 /* ShorthandPropertyAssignment */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.questionToken) ||
                        visitNode(cbNode, node.exclamationToken) ||
                        visitNode(cbNode, node.equalsToken) ||
                        visitNode(cbNode, node.objectAssignmentInitializer);
                case 293 /* SpreadAssignment */:
                    return visitNode(cbNode, node.expression);
                case 162 /* Parameter */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.dotDotDotToken) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.questionToken) ||
                        visitNode(cbNode, node.type) ||
                        visitNode(cbNode, node.initializer);
                case 165 /* PropertyDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.questionToken) ||
                        visitNode(cbNode, node.exclamationToken) ||
                        visitNode(cbNode, node.type) ||
                        visitNode(cbNode, node.initializer);
                case 164 /* PropertySignature */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.questionToken) ||
                        visitNode(cbNode, node.type) ||
                        visitNode(cbNode, node.initializer);
                case 291 /* PropertyAssignment */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.questionToken) ||
                        visitNode(cbNode, node.initializer);
                case 252 /* VariableDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.exclamationToken) ||
                        visitNode(cbNode, node.type) ||
                        visitNode(cbNode, node.initializer);
                case 201 /* BindingElement */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.dotDotDotToken) ||
                        visitNode(cbNode, node.propertyName) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.initializer);
                case 177 /* FunctionType */:
                case 178 /* ConstructorType */:
                case 172 /* CallSignature */:
                case 173 /* ConstructSignature */:
                case 174 /* IndexSignature */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNodes(cbNode, cbNodes, node.typeParameters) ||
                        visitNodes(cbNode, cbNodes, node.parameters) ||
                        visitNode(cbNode, node.type);
                case 167 /* MethodDeclaration */:
                case 166 /* MethodSignature */:
                case 169 /* Constructor */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 211 /* FunctionExpression */:
                case 254 /* FunctionDeclaration */:
                case 212 /* ArrowFunction */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.asteriskToken) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.questionToken) ||
                        visitNode(cbNode, node.exclamationToken) ||
                        visitNodes(cbNode, cbNodes, node.typeParameters) ||
                        visitNodes(cbNode, cbNodes, node.parameters) ||
                        visitNode(cbNode, node.type) ||
                        visitNode(cbNode, node.equalsGreaterThanToken) ||
                        visitNode(cbNode, node.body);
                case 168 /* ClassStaticBlockDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.body);
                case 176 /* TypeReference */:
                    return visitNode(cbNode, node.typeName) ||
                        visitNodes(cbNode, cbNodes, node.typeArguments);
                case 175 /* TypePredicate */:
                    return visitNode(cbNode, node.assertsModifier) ||
                        visitNode(cbNode, node.parameterName) ||
                        visitNode(cbNode, node.type);
                case 179 /* TypeQuery */:
                    return visitNode(cbNode, node.exprName);
                case 180 /* TypeLiteral */:
                    return visitNodes(cbNode, cbNodes, node.members);
                case 181 /* ArrayType */:
                    return visitNode(cbNode, node.elementType);
                case 182 /* TupleType */:
                    return visitNodes(cbNode, cbNodes, node.elements);
                case 185 /* UnionType */:
                case 186 /* IntersectionType */:
                    return visitNodes(cbNode, cbNodes, node.types);
                case 187 /* ConditionalType */:
                    return visitNode(cbNode, node.checkType) ||
                        visitNode(cbNode, node.extendsType) ||
                        visitNode(cbNode, node.trueType) ||
                        visitNode(cbNode, node.falseType);
                case 188 /* InferType */:
                    return visitNode(cbNode, node.typeParameter);
                case 198 /* ImportType */:
                    return visitNode(cbNode, node.argument) ||
                        visitNode(cbNode, node.qualifier) ||
                        visitNodes(cbNode, cbNodes, node.typeArguments);
                case 189 /* ParenthesizedType */:
                case 191 /* TypeOperator */:
                    return visitNode(cbNode, node.type);
                case 192 /* IndexedAccessType */:
                    return visitNode(cbNode, node.objectType) ||
                        visitNode(cbNode, node.indexType);
                case 193 /* MappedType */:
                    return visitNode(cbNode, node.readonlyToken) ||
                        visitNode(cbNode, node.typeParameter) ||
                        visitNode(cbNode, node.nameType) ||
                        visitNode(cbNode, node.questionToken) ||
                        visitNode(cbNode, node.type);
                case 194 /* LiteralType */:
                    return visitNode(cbNode, node.literal);
                case 195 /* NamedTupleMember */:
                    return visitNode(cbNode, node.dotDotDotToken) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.questionToken) ||
                        visitNode(cbNode, node.type);
                case 199 /* ObjectBindingPattern */:
                case 200 /* ArrayBindingPattern */:
                    return visitNodes(cbNode, cbNodes, node.elements);
                case 202 /* ArrayLiteralExpression */:
                    return visitNodes(cbNode, cbNodes, node.elements);
                case 203 /* ObjectLiteralExpression */:
                    return visitNodes(cbNode, cbNodes, node.properties);
                case 204 /* PropertyAccessExpression */:
                    return visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.questionDotToken) ||
                        visitNode(cbNode, node.name);
                case 205 /* ElementAccessExpression */:
                    return visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.questionDotToken) ||
                        visitNode(cbNode, node.argumentExpression);
                case 206 /* CallExpression */:
                case 207 /* NewExpression */:
                    return visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.questionDotToken) ||
                        visitNodes(cbNode, cbNodes, node.typeArguments) ||
                        visitNodes(cbNode, cbNodes, node.arguments);
                case 208 /* TaggedTemplateExpression */:
                    return visitNode(cbNode, node.tag) ||
                        visitNode(cbNode, node.questionDotToken) ||
                        visitNodes(cbNode, cbNodes, node.typeArguments) ||
                        visitNode(cbNode, node.template);
                case 209 /* TypeAssertionExpression */:
                    return visitNode(cbNode, node.type) ||
                        visitNode(cbNode, node.expression);
                case 210 /* ParenthesizedExpression */:
                    return visitNode(cbNode, node.expression);
                case 213 /* DeleteExpression */:
                    return visitNode(cbNode, node.expression);
                case 214 /* TypeOfExpression */:
                    return visitNode(cbNode, node.expression);
                case 215 /* VoidExpression */:
                    return visitNode(cbNode, node.expression);
                case 217 /* PrefixUnaryExpression */:
                    return visitNode(cbNode, node.operand);
                case 222 /* YieldExpression */:
                    return visitNode(cbNode, node.asteriskToken) ||
                        visitNode(cbNode, node.expression);
                case 216 /* AwaitExpression */:
                    return visitNode(cbNode, node.expression);
                case 218 /* PostfixUnaryExpression */:
                    return visitNode(cbNode, node.operand);
                case 219 /* BinaryExpression */:
                    return visitNode(cbNode, node.left) ||
                        visitNode(cbNode, node.operatorToken) ||
                        visitNode(cbNode, node.right);
                case 227 /* AsExpression */:
                    return visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.type);
                case 228 /* NonNullExpression */:
                    return visitNode(cbNode, node.expression);
                case 229 /* MetaProperty */:
                    return visitNode(cbNode, node.name);
                case 220 /* ConditionalExpression */:
                    return visitNode(cbNode, node.condition) ||
                        visitNode(cbNode, node.questionToken) ||
                        visitNode(cbNode, node.whenTrue) ||
                        visitNode(cbNode, node.colonToken) ||
                        visitNode(cbNode, node.whenFalse);
                case 223 /* SpreadElement */:
                    return visitNode(cbNode, node.expression);
                case 233 /* Block */:
                case 260 /* ModuleBlock */:
                    return visitNodes(cbNode, cbNodes, node.statements);
                case 300 /* SourceFile */:
                    return visitNodes(cbNode, cbNodes, node.statements) ||
                        visitNode(cbNode, node.endOfFileToken);
                case 235 /* VariableStatement */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.declarationList);
                case 253 /* VariableDeclarationList */:
                    return visitNodes(cbNode, cbNodes, node.declarations);
                case 236 /* ExpressionStatement */:
                    return visitNode(cbNode, node.expression);
                case 237 /* IfStatement */:
                    return visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.thenStatement) ||
                        visitNode(cbNode, node.elseStatement);
                case 238 /* DoStatement */:
                    return visitNode(cbNode, node.statement) ||
                        visitNode(cbNode, node.expression);
                case 239 /* WhileStatement */:
                    return visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.statement);
                case 240 /* ForStatement */:
                    return visitNode(cbNode, node.initializer) ||
                        visitNode(cbNode, node.condition) ||
                        visitNode(cbNode, node.incrementor) ||
                        visitNode(cbNode, node.statement);
                case 241 /* ForInStatement */:
                    return visitNode(cbNode, node.initializer) ||
                        visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.statement);
                case 242 /* ForOfStatement */:
                    return visitNode(cbNode, node.awaitModifier) ||
                        visitNode(cbNode, node.initializer) ||
                        visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.statement);
                case 243 /* ContinueStatement */:
                case 244 /* BreakStatement */:
                    return visitNode(cbNode, node.label);
                case 245 /* ReturnStatement */:
                    return visitNode(cbNode, node.expression);
                case 246 /* WithStatement */:
                    return visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.statement);
                case 247 /* SwitchStatement */:
                    return visitNode(cbNode, node.expression) ||
                        visitNode(cbNode, node.caseBlock);
                case 261 /* CaseBlock */:
                    return visitNodes(cbNode, cbNodes, node.clauses);
                case 287 /* CaseClause */:
                    return visitNode(cbNode, node.expression) ||
                        visitNodes(cbNode, cbNodes, node.statements);
                case 288 /* DefaultClause */:
                    return visitNodes(cbNode, cbNodes, node.statements);
                case 248 /* LabeledStatement */:
                    return visitNode(cbNode, node.label) ||
                        visitNode(cbNode, node.statement);
                case 249 /* ThrowStatement */:
                    return visitNode(cbNode, node.expression);
                case 250 /* TryStatement */:
                    return visitNode(cbNode, node.tryBlock) ||
                        visitNode(cbNode, node.catchClause) ||
                        visitNode(cbNode, node.finallyBlock);
                case 290 /* CatchClause */:
                    return visitNode(cbNode, node.variableDeclaration) ||
                        visitNode(cbNode, node.block);
                case 163 /* Decorator */:
                    return visitNode(cbNode, node.expression);
                case 255 /* ClassDeclaration */:
                case 224 /* ClassExpression */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNodes(cbNode, cbNodes, node.typeParameters) ||
                        visitNodes(cbNode, cbNodes, node.heritageClauses) ||
                        visitNodes(cbNode, cbNodes, node.members);
                case 256 /* InterfaceDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNodes(cbNode, cbNodes, node.typeParameters) ||
                        visitNodes(cbNode, cbNodes, node.heritageClauses) ||
                        visitNodes(cbNode, cbNodes, node.members);
                case 257 /* TypeAliasDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNodes(cbNode, cbNodes, node.typeParameters) ||
                        visitNode(cbNode, node.type);
                case 258 /* EnumDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNodes(cbNode, cbNodes, node.members);
                case 294 /* EnumMember */:
                    return visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.initializer);
                case 259 /* ModuleDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.body);
                case 263 /* ImportEqualsDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.moduleReference);
                case 264 /* ImportDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.importClause) ||
                        visitNode(cbNode, node.moduleSpecifier);
                case 265 /* ImportClause */:
                    return visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.namedBindings);
                case 262 /* NamespaceExportDeclaration */:
                    return visitNode(cbNode, node.name);
                case 266 /* NamespaceImport */:
                    return visitNode(cbNode, node.name);
                case 272 /* NamespaceExport */:
                    return visitNode(cbNode, node.name);
                case 267 /* NamedImports */:
                case 271 /* NamedExports */:
                    return visitNodes(cbNode, cbNodes, node.elements);
                case 270 /* ExportDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.exportClause) ||
                        visitNode(cbNode, node.moduleSpecifier);
                case 268 /* ImportSpecifier */:
                case 273 /* ExportSpecifier */:
                    return visitNode(cbNode, node.propertyName) ||
                        visitNode(cbNode, node.name);
                case 269 /* ExportAssignment */:
                    return visitNodes(cbNode, cbNodes, node.decorators) ||
                        visitNodes(cbNode, cbNodes, node.modifiers) ||
                        visitNode(cbNode, node.expression);
                case 221 /* TemplateExpression */:
                    return visitNode(cbNode, node.head) || visitNodes(cbNode, cbNodes, node.templateSpans);
                case 231 /* TemplateSpan */:
                    return visitNode(cbNode, node.expression) || visitNode(cbNode, node.literal);
                case 196 /* TemplateLiteralType */:
                    return visitNode(cbNode, node.head) || visitNodes(cbNode, cbNodes, node.templateSpans);
                case 197 /* TemplateLiteralTypeSpan */:
                    return visitNode(cbNode, node.type) || visitNode(cbNode, node.literal);
                case 160 /* ComputedPropertyName */:
                    return visitNode(cbNode, node.expression);
                case 289 /* HeritageClause */:
                    return visitNodes(cbNode, cbNodes, node.types);
                case 226 /* ExpressionWithTypeArguments */:
                    return visitNode(cbNode, node.expression) ||
                        visitNodes(cbNode, cbNodes, node.typeArguments);
                case 275 /* ExternalModuleReference */:
                    return visitNode(cbNode, node.expression);
                case 274 /* MissingDeclaration */:
                    return visitNodes(cbNode, cbNodes, node.decorators);
                case 346 /* CommaListExpression */:
                    return visitNodes(cbNode, cbNodes, node.elements);
                case 276 /* JsxElement */:
                    return visitNode(cbNode, node.openingElement) ||
                        visitNodes(cbNode, cbNodes, node.children) ||
                        visitNode(cbNode, node.closingElement);
                case 280 /* JsxFragment */:
                    return visitNode(cbNode, node.openingFragment) ||
                        visitNodes(cbNode, cbNodes, node.children) ||
                        visitNode(cbNode, node.closingFragment);
                case 277 /* JsxSelfClosingElement */:
                case 278 /* JsxOpeningElement */:
                    return visitNode(cbNode, node.tagName) ||
                        visitNodes(cbNode, cbNodes, node.typeArguments) ||
                        visitNode(cbNode, node.attributes);
                case 284 /* JsxAttributes */:
                    return visitNodes(cbNode, cbNodes, node.properties);
                case 283 /* JsxAttribute */:
                    return visitNode(cbNode, node.name) ||
                        visitNode(cbNode, node.initializer);
                case 285 /* JsxSpreadAttribute */:
                    return visitNode(cbNode, node.expression);
                case 286 /* JsxExpression */:
                    return visitNode(cbNode, node.dotDotDotToken) ||
                        visitNode(cbNode, node.expression);
                case 279 /* JsxClosingElement */:
                    return visitNode(cbNode, node.tagName);
                case 183 /* OptionalType */:
                case 184 /* RestType */:
                case 304 /* JSDocTypeExpression */:
                case 310 /* JSDocNonNullableType */:
                case 309 /* JSDocNullableType */:
                case 311 /* JSDocOptionalType */:
                case 313 /* JSDocVariadicType */:
                    return visitNode(cbNode, node.type);
                case 312 /* JSDocFunctionType */:
                    return visitNodes(cbNode, cbNodes, node.parameters) ||
                        visitNode(cbNode, node.type);
                case 315 /* JSDocComment */:
                    return (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment))
                        || visitNodes(cbNode, cbNodes, node.tags);
                case 341 /* JSDocSeeTag */:
                    return visitNode(cbNode, node.tagName) ||
                        visitNode(cbNode, node.name) ||
                        (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment));
                case 305 /* JSDocNameReference */:
                    return visitNode(cbNode, node.name);
                case 306 /* JSDocMemberName */:
                    return visitNode(cbNode, node.left) ||
                        visitNode(cbNode, node.right);
                case 335 /* JSDocParameterTag */:
                case 342 /* JSDocPropertyTag */:
                    return visitNode(cbNode, node.tagName) ||
                        (node.isNameFirst
                            ? visitNode(cbNode, node.name) ||
                                visitNode(cbNode, node.typeExpression) ||
                                (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment))
                            : visitNode(cbNode, node.typeExpression) ||
                                visitNode(cbNode, node.name) ||
                                (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment)));
                case 325 /* JSDocAuthorTag */:
                    return visitNode(cbNode, node.tagName) ||
                        (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment));
                case 324 /* JSDocImplementsTag */:
                    return visitNode(cbNode, node.tagName) ||
                        visitNode(cbNode, node.class) ||
                        (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment));
                case 323 /* JSDocAugmentsTag */:
                    return visitNode(cbNode, node.tagName) ||
                        visitNode(cbNode, node.class) ||
                        (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment));
                case 339 /* JSDocTemplateTag */:
                    return visitNode(cbNode, node.tagName) ||
                        visitNode(cbNode, node.constraint) ||
                        visitNodes(cbNode, cbNodes, node.typeParameters) ||
                        (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment));
                case 340 /* JSDocTypedefTag */:
                    return visitNode(cbNode, node.tagName) ||
                        (node.typeExpression &&
                            node.typeExpression.kind === 304 /* JSDocTypeExpression */
                            ? visitNode(cbNode, node.typeExpression) ||
                                visitNode(cbNode, node.fullName) ||
                                (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment))
                            : visitNode(cbNode, node.fullName) ||
                                visitNode(cbNode, node.typeExpression) ||
                                (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment)));
                case 333 /* JSDocCallbackTag */:
                    return visitNode(cbNode, node.tagName) ||
                        visitNode(cbNode, node.fullName) ||
                        visitNode(cbNode, node.typeExpression) ||
                        (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment));
                case 336 /* JSDocReturnTag */:
                case 338 /* JSDocTypeTag */:
                case 337 /* JSDocThisTag */:
                case 334 /* JSDocEnumTag */:
                    return visitNode(cbNode, node.tagName) ||
                        visitNode(cbNode, node.typeExpression) ||
                        (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment));
                case 318 /* JSDocSignature */:
                    return ts.forEach(node.typeParameters, cbNode) ||
                        ts.forEach(node.parameters, cbNode) ||
                        visitNode(cbNode, node.type);
                case 319 /* JSDocLink */:
                case 320 /* JSDocLinkCode */:
                case 321 /* JSDocLinkPlain */:
                    return visitNode(cbNode, node.name);
                case 317 /* JSDocTypeLiteral */:
                    return ts.forEach(node.jsDocPropertyTags, cbNode);
                case 322 /* JSDocTag */:
                case 327 /* JSDocClassTag */:
                case 328 /* JSDocPublicTag */:
                case 329 /* JSDocPrivateTag */:
                case 330 /* JSDocProtectedTag */:
                case 331 /* JSDocReadonlyTag */:
                case 326 /* JSDocDeprecatedTag */:
                    return visitNode(cbNode, node.tagName)
                        || (typeof node.comment === "string" ? undefined : visitNodes(cbNode, cbNodes, node.comment));
                case 345 /* PartiallyEmittedExpression */:
                    return visitNode(cbNode, node.expression);
            }
        }
        ts.forEachChild = forEachChild;
        /** @internal */
        /**
         * Invokes a callback for each child of the given node. The 'cbNode' callback is invoked for all child nodes
         * stored in properties. If a 'cbNodes' callback is specified, it is invoked for embedded arrays; additionally,
         * unlike `forEachChild`, embedded arrays are flattened and the 'cbNode' callback is invoked for each element.
         *  If a callback returns a truthy value, iteration stops and that value is returned. Otherwise, undefined is returned.
         *
         * @param node a given node to visit its children
         * @param cbNode a callback to be invoked for all child nodes
         * @param cbNodes a callback to be invoked for embedded array
         *
         * @remarks Unlike `forEachChild`, `forEachChildRecursively` handles recursively invoking the traversal on each child node found,
         * and while doing so, handles traversing the structure without relying on the callstack to encode the tree structure.
         */
        function forEachChildRecursively(rootNode, cbNode, cbNodes) {
            var queue = gatherPossibleChildren(rootNode);
            var parents = []; // tracks parent references for elements in queue
            while (parents.length < queue.length) {
                parents.push(rootNode);
            }
            while (queue.length !== 0) {
                var current = queue.pop();
                var parent = parents.pop();
                if (ts.isArray(current)) {
                    if (cbNodes) {
                        var res = cbNodes(current, parent);
                        if (res) {
                            if (res === "skip")
                                continue;
                            return res;
                        }
                    }
                    for (var i = current.length - 1; i >= 0; --i) {
                        queue.push(current[i]);
                        parents.push(parent);
                    }
                }
                else {
                    var res = cbNode(current, parent);
                    if (res) {
                        if (res === "skip")
                            continue;
                        return res;
                    }
                    if (current.kind >= 159 /* FirstNode */) {
                        // add children in reverse order to the queue, so popping gives the first child
                        for (var _i = 0, _a = gatherPossibleChildren(current); _i < _a.length; _i++) {
                            var child = _a[_i];
                            queue.push(child);
                            parents.push(current);
                        }
                    }
                }
            }
        }
        ts.forEachChildRecursively = forEachChildRecursively;
        function gatherPossibleChildren(node) {
            var children = [];
            forEachChild(node, addWorkItem, addWorkItem); // By using a stack above and `unshift` here, we emulate a depth-first preorder traversal
            return children;
            function addWorkItem(n) {
                children.unshift(n);
            }
        }
        function createSourceFile(fileName, sourceText, languageVersion, setParentNodes, scriptKind) {
            if (setParentNodes === void 0) {
                setParentNodes = false;
            }
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("parse" /* Parse */, "createSourceFile", { path: fileName }, /*separateBeginAndEnd*/ true);
            ts.performance.mark("beforeParse");
            var result;
            ts.perfLogger.logStartParseSourceFile(fileName);
            if (languageVersion === 100 /* JSON */) {
                result = Parser.parseSourceFile(fileName, sourceText, languageVersion, /*syntaxCursor*/ undefined, setParentNodes, 6 /* JSON */);
            }
            else {
                result = Parser.parseSourceFile(fileName, sourceText, languageVersion, /*syntaxCursor*/ undefined, setParentNodes, scriptKind);
            }
            ts.perfLogger.logStopParseSourceFile();
            ts.performance.mark("afterParse");
            ts.performance.measure("Parse", "beforeParse", "afterParse");
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
            return result;
        }
        ts.createSourceFile = createSourceFile;
        function parseIsolatedEntityName(text, languageVersion) {
            return Parser.parseIsolatedEntityName(text, languageVersion);
        }
        ts.parseIsolatedEntityName = parseIsolatedEntityName;
        /**
         * Parse json text into SyntaxTree and return node and parse errors if any
         * @param fileName
         * @param sourceText
         */
        function parseJsonText(fileName, sourceText) {
            return Parser.parseJsonText(fileName, sourceText);
        }
        ts.parseJsonText = parseJsonText;
        // See also `isExternalOrCommonJsModule` in utilities.ts
        function isExternalModule(file) {
            return file.externalModuleIndicator !== undefined;
        }
        ts.isExternalModule = isExternalModule;
        // Produces a new SourceFile for the 'newText' provided. The 'textChangeRange' parameter
        // indicates what changed between the 'text' that this SourceFile has and the 'newText'.
        // The SourceFile will be created with the compiler attempting to reuse as many nodes from
        // this file as possible.
        //
        // Note: this function mutates nodes from this SourceFile. That means any existing nodes
        // from this SourceFile that are being held onto may change as a result (including
        // becoming detached from any SourceFile).  It is recommended that this SourceFile not
        // be used once 'update' is called on it.
        function updateSourceFile(sourceFile, newText, textChangeRange, aggressiveChecks) {
            if (aggressiveChecks === void 0) {
                aggressiveChecks = false;
            }
            var newSourceFile = IncrementalParser.updateSourceFile(sourceFile, newText, textChangeRange, aggressiveChecks);
            // Because new source file node is created, it may not have the flag PossiblyContainDynamicImport. This is the case if there is no new edit to add dynamic import.
            // We will manually port the flag to the new source file.
            newSourceFile.flags |= (sourceFile.flags & 3145728 /* PermanentlySetIncrementalFlags */);
            return newSourceFile;
        }
        ts.updateSourceFile = updateSourceFile;
        /* @internal */
        function parseIsolatedJSDocComment(content, start, length) {
            var result = Parser.JSDocParser.parseIsolatedJSDocComment(content, start, length);
            if (result && result.jsDoc) {
                // because the jsDocComment was parsed out of the source file, it might
                // not be covered by the fixupParentReferences.
                Parser.fixupParentReferences(result.jsDoc);
            }
            return result;
        }
        ts.parseIsolatedJSDocComment = parseIsolatedJSDocComment;
        /* @internal */
        // Exposed only for testing.
        function parseJSDocTypeExpressionForTests(content, start, length) {
            return Parser.JSDocParser.parseJSDocTypeExpressionForTests(content, start, length);
        }
        ts.parseJSDocTypeExpressionForTests = parseJSDocTypeExpressionForTests;
        // Implement the parser as a singleton module.  We do this for perf reasons because creating
        // parser instances can actually be expensive enough to impact us on projects with many source
        // files.
        var Parser;
        (function (Parser) {
            // Share a single scanner across all calls to parse a source file.  This helps speed things
            // up by avoiding the cost of creating/compiling scanners over and over again.
            var scanner = ts.createScanner(99 /* Latest */, /*skipTrivia*/ true);
            var disallowInAndDecoratorContext = 4096 /* DisallowInContext */ | 16384 /* DecoratorContext */;
            // capture constructors in 'initializeState' to avoid null checks
            // tslint:disable variable-name
            var NodeConstructor;
            var TokenConstructor;
            var IdentifierConstructor;
            var PrivateIdentifierConstructor;
            var SourceFileConstructor;
            // tslint:enable variable-name
            function countNode(node) {
                nodeCount++;
                return node;
            }
            // Rather than using `createBaseNodeFactory` here, we establish a `BaseNodeFactory` that closes over the
            // constructors above, which are reset each time `initializeState` is called.
            var baseNodeFactory = {
                createBaseSourceFileNode: function (kind) { return countNode(new SourceFileConstructor(kind, /*pos*/ 0, /*end*/ 0)); },
                createBaseIdentifierNode: function (kind) { return countNode(new IdentifierConstructor(kind, /*pos*/ 0, /*end*/ 0)); },
                createBasePrivateIdentifierNode: function (kind) { return countNode(new PrivateIdentifierConstructor(kind, /*pos*/ 0, /*end*/ 0)); },
                createBaseTokenNode: function (kind) { return countNode(new TokenConstructor(kind, /*pos*/ 0, /*end*/ 0)); },
                createBaseNode: function (kind) { return countNode(new NodeConstructor(kind, /*pos*/ 0, /*end*/ 0)); }
            };
            var factory = ts.createNodeFactory(1 /* NoParenthesizerRules */ | 2 /* NoNodeConverters */ | 8 /* NoOriginalNode */, baseNodeFactory);
            var fileName;
            var sourceFlags;
            var sourceText;
            var languageVersion;
            var scriptKind;
            var languageVariant;
            var parseDiagnostics;
            var jsDocDiagnostics;
            var syntaxCursor;
            var currentToken;
            var nodeCount;
            var identifiers;
            var privateIdentifiers;
            var identifierCount;
            var parsingContext;
            var notParenthesizedArrow;
            // Flags that dictate what parsing context we're in.  For example:
            // Whether or not we are in strict parsing mode.  All that changes in strict parsing mode is
            // that some tokens that would be considered identifiers may be considered keywords.
            //
            // When adding more parser context flags, consider which is the more common case that the
            // flag will be in.  This should be the 'false' state for that flag.  The reason for this is
            // that we don't store data in our nodes unless the value is in the *non-default* state.  So,
            // for example, more often than code 'allows-in' (or doesn't 'disallow-in').  We opt for
            // 'disallow-in' set to 'false'.  Otherwise, if we had 'allowsIn' set to 'true', then almost
            // all nodes would need extra state on them to store this info.
            //
            // Note: 'allowIn' and 'allowYield' track 1:1 with the [in] and [yield] concepts in the ES6
            // grammar specification.
            //
            // An important thing about these context concepts.  By default they are effectively inherited
            // while parsing through every grammar production.  i.e. if you don't change them, then when
            // you parse a sub-production, it will have the same context values as the parent production.
            // This is great most of the time.  After all, consider all the 'expression' grammar productions
            // and how nearly all of them pass along the 'in' and 'yield' context values:
            //
            // EqualityExpression[In, Yield] :
            //      RelationalExpression[?In, ?Yield]
            //      EqualityExpression[?In, ?Yield] == RelationalExpression[?In, ?Yield]
            //      EqualityExpression[?In, ?Yield] != RelationalExpression[?In, ?Yield]
            //      EqualityExpression[?In, ?Yield] === RelationalExpression[?In, ?Yield]
            //      EqualityExpression[?In, ?Yield] !== RelationalExpression[?In, ?Yield]
            //
            // Where you have to be careful is then understanding what the points are in the grammar
            // where the values are *not* passed along.  For example:
            //
            // SingleNameBinding[Yield,GeneratorParameter]
            //      [+GeneratorParameter]BindingIdentifier[Yield] Initializer[In]opt
            //      [~GeneratorParameter]BindingIdentifier[?Yield]Initializer[In, ?Yield]opt
            //
            // Here this is saying that if the GeneratorParameter context flag is set, that we should
            // explicitly set the 'yield' context flag to false before calling into the BindingIdentifier
            // and we should explicitly unset the 'yield' context flag before calling into the Initializer.
            // production.  Conversely, if the GeneratorParameter context flag is not set, then we
            // should leave the 'yield' context flag alone.
            //
            // Getting this all correct is tricky and requires careful reading of the grammar to
            // understand when these values should be changed versus when they should be inherited.
            //
            // Note: it should not be necessary to save/restore these flags during speculative/lookahead
            // parsing.  These context flags are naturally stored and restored through normal recursive
            // descent parsing and unwinding.
            var contextFlags;
            // Indicates whether we are currently parsing top-level statements.
            var topLevel = true;
            // Whether or not we've had a parse error since creating the last AST node.  If we have
            // encountered an error, it will be stored on the next AST node we create.  Parse errors
            // can be broken down into three categories:
            //
            // 1) An error that occurred during scanning.  For example, an unterminated literal, or a
            //    character that was completely not understood.
            //
            // 2) A token was expected, but was not present.  This type of error is commonly produced
            //    by the 'parseExpected' function.
            //
            // 3) A token was present that no parsing function was able to consume.  This type of error
            //    only occurs in the 'abortParsingListOrMoveToNextToken' function when the parser
            //    decides to skip the token.
            //
            // In all of these cases, we want to mark the next node as having had an error before it.
            // With this mark, we can know in incremental settings if this node can be reused, or if
            // we have to reparse it.  If we don't keep this information around, we may just reuse the
            // node.  in that event we would then not produce the same errors as we did before, causing
            // significant confusion problems.
            //
            // Note: it is necessary that this value be saved/restored during speculative/lookahead
            // parsing.  During lookahead parsing, we will often create a node.  That node will have
            // this value attached, and then this value will be set back to 'false'.  If we decide to
            // rewind, we must get back to the same value we had prior to the lookahead.
            //
            // Note: any errors at the end of the file that do not precede a regular node, should get
            // attached to the EOF token.
            var parseErrorBeforeNextFinishedNode = false;
            function parseSourceFile(fileName, sourceText, languageVersion, syntaxCursor, setParentNodes, scriptKind) {
                var _a;
                if (setParentNodes === void 0) {
                    setParentNodes = false;
                }
                scriptKind = ts.ensureScriptKind(fileName, scriptKind);
                if (scriptKind === 6 /* JSON */) {
                    var result_3 = parseJsonText(fileName, sourceText, languageVersion, syntaxCursor, setParentNodes);
                    ts.convertToObjectWorker(result_3, (_a = result_3.statements[0]) === null || _a === void 0 ? void 0 : _a.expression, result_3.parseDiagnostics, /*returnValue*/ false, /*knownRootOptions*/ undefined, /*jsonConversionNotifier*/ undefined);
                    result_3.referencedFiles = ts.emptyArray;
                    result_3.typeReferenceDirectives = ts.emptyArray;
                    result_3.libReferenceDirectives = ts.emptyArray;
                    result_3.amdDependencies = ts.emptyArray;
                    result_3.hasNoDefaultLib = false;
                    result_3.pragmas = ts.emptyMap;
                    return result_3;
                }
                initializeState(fileName, sourceText, languageVersion, syntaxCursor, scriptKind);
                var result = parseSourceFileWorker(languageVersion, setParentNodes, scriptKind);
                clearState();
                return result;
            }
            Parser.parseSourceFile = parseSourceFile;
            function parseIsolatedEntityName(content, languageVersion) {
                // Choice of `isDeclarationFile` should be arbitrary
                initializeState("", content, languageVersion, /*syntaxCursor*/ undefined, 1 /* JS */);
                // Prime the scanner.
                nextToken();
                var entityName = parseEntityName(/*allowReservedWords*/ true);
                var isInvalid = token() === 1 /* EndOfFileToken */ && !parseDiagnostics.length;
                clearState();
                return isInvalid ? entityName : undefined;
            }
            Parser.parseIsolatedEntityName = parseIsolatedEntityName;
            function parseJsonText(fileName, sourceText, languageVersion, syntaxCursor, setParentNodes) {
                if (languageVersion === void 0) {
                    languageVersion = 2 /* ES2015 */;
                }
                if (setParentNodes === void 0) {
                    setParentNodes = false;
                }
                initializeState(fileName, sourceText, languageVersion, syntaxCursor, 6 /* JSON */);
                sourceFlags = contextFlags;
                // Prime the scanner.
                nextToken();
                var pos = getNodePos();
                var statements, endOfFileToken;
                if (token() === 1 /* EndOfFileToken */) {
                    statements = createNodeArray([], pos, pos);
                    endOfFileToken = parseTokenNode();
                }
                else {
                    // Loop and synthesize an ArrayLiteralExpression if there are more than
                    // one top-level expressions to ensure all input text is consumed.
                    var expressions = void 0;
                    while (token() !== 1 /* EndOfFileToken */) {
                        var expression_1 = void 0;
                        switch (token()) {
                            case 22 /* OpenBracketToken */:
                                expression_1 = parseArrayLiteralExpression();
                                break;
                            case 110 /* TrueKeyword */:
                            case 95 /* FalseKeyword */:
                            case 104 /* NullKeyword */:
                                expression_1 = parseTokenNode();
                                break;
                            case 40 /* MinusToken */:
                                if (lookAhead(function () { return nextToken() === 8 /* NumericLiteral */ && nextToken() !== 58 /* ColonToken */; })) {
                                    expression_1 = parsePrefixUnaryExpression();
                                }
                                else {
                                    expression_1 = parseObjectLiteralExpression();
                                }
                                break;
                            case 8 /* NumericLiteral */:
                            case 10 /* StringLiteral */:
                                if (lookAhead(function () { return nextToken() !== 58 /* ColonToken */; })) {
                                    expression_1 = parseLiteralNode();
                                    break;
                                }
                            // falls through
                            default:
                                expression_1 = parseObjectLiteralExpression();
                                break;
                        }
                        // Error recovery: collect multiple top-level expressions
                        if (expressions && ts.isArray(expressions)) {
                            expressions.push(expression_1);
                        }
                        else if (expressions) {
                            expressions = [expressions, expression_1];
                        }
                        else {
                            expressions = expression_1;
                            if (token() !== 1 /* EndOfFileToken */) {
                                parseErrorAtCurrentToken(ts.Diagnostics.Unexpected_token);
                            }
                        }
                    }
                    var expression = ts.isArray(expressions) ? finishNode(factory.createArrayLiteralExpression(expressions), pos) : ts.Debug.checkDefined(expressions);
                    var statement = factory.createExpressionStatement(expression);
                    finishNode(statement, pos);
                    statements = createNodeArray([statement], pos);
                    endOfFileToken = parseExpectedToken(1 /* EndOfFileToken */, ts.Diagnostics.Unexpected_token);
                }
                // Set source file so that errors will be reported with this file name
                var sourceFile = createSourceFile(fileName, 2 /* ES2015 */, 6 /* JSON */, /*isDeclaration*/ false, statements, endOfFileToken, sourceFlags);
                if (setParentNodes) {
                    fixupParentReferences(sourceFile);
                }
                sourceFile.nodeCount = nodeCount;
                sourceFile.identifierCount = identifierCount;
                sourceFile.identifiers = identifiers;
                sourceFile.parseDiagnostics = ts.attachFileToDiagnostics(parseDiagnostics, sourceFile);
                if (jsDocDiagnostics) {
                    sourceFile.jsDocDiagnostics = ts.attachFileToDiagnostics(jsDocDiagnostics, sourceFile);
                }
                var result = sourceFile;
                clearState();
                return result;
            }
            Parser.parseJsonText = parseJsonText;
            function initializeState(_fileName, _sourceText, _languageVersion, _syntaxCursor, _scriptKind) {
                NodeConstructor = ts.objectAllocator.getNodeConstructor();
                TokenConstructor = ts.objectAllocator.getTokenConstructor();
                IdentifierConstructor = ts.objectAllocator.getIdentifierConstructor();
                PrivateIdentifierConstructor = ts.objectAllocator.getPrivateIdentifierConstructor();
                SourceFileConstructor = ts.objectAllocator.getSourceFileConstructor();
                fileName = ts.normalizePath(_fileName);
                sourceText = _sourceText;
                languageVersion = _languageVersion;
                syntaxCursor = _syntaxCursor;
                scriptKind = _scriptKind;
                languageVariant = ts.getLanguageVariant(_scriptKind);
                parseDiagnostics = [];
                parsingContext = 0;
                identifiers = new ts.Map();
                privateIdentifiers = new ts.Map();
                identifierCount = 0;
                nodeCount = 0;
                sourceFlags = 0;
                topLevel = true;
                switch (scriptKind) {
                    case 1 /* JS */:
                    case 2 /* JSX */:
                        contextFlags = 131072 /* JavaScriptFile */;
                        break;
                    case 6 /* JSON */:
                        contextFlags = 131072 /* JavaScriptFile */ | 33554432 /* JsonFile */;
                        break;
                    default:
                        contextFlags = 0 /* None */;
                        break;
                }
                parseErrorBeforeNextFinishedNode = false;
                // Initialize and prime the scanner before parsing the source elements.
                scanner.setText(sourceText);
                scanner.setOnError(scanError);
                scanner.setScriptTarget(languageVersion);
                scanner.setLanguageVariant(languageVariant);
            }
            function clearState() {
                // Clear out the text the scanner is pointing at, so it doesn't keep anything alive unnecessarily.
                scanner.clearCommentDirectives();
                scanner.setText("");
                scanner.setOnError(undefined);
                // Clear any data.  We don't want to accidentally hold onto it for too long.
                sourceText = undefined;
                languageVersion = undefined;
                syntaxCursor = undefined;
                scriptKind = undefined;
                languageVariant = undefined;
                sourceFlags = 0;
                parseDiagnostics = undefined;
                jsDocDiagnostics = undefined;
                parsingContext = 0;
                identifiers = undefined;
                notParenthesizedArrow = undefined;
                topLevel = true;
            }
            function parseSourceFileWorker(languageVersion, setParentNodes, scriptKind) {
                var isDeclarationFile = isDeclarationFileName(fileName);
                if (isDeclarationFile) {
                    contextFlags |= 8388608 /* Ambient */;
                }
                sourceFlags = contextFlags;
                // Prime the scanner.
                nextToken();
                var statements = parseList(0 /* SourceElements */, parseStatement);
                ts.Debug.assert(token() === 1 /* EndOfFileToken */);
                var endOfFileToken = addJSDocComment(parseTokenNode());
                var sourceFile = createSourceFile(fileName, languageVersion, scriptKind, isDeclarationFile, statements, endOfFileToken, sourceFlags);
                // A member of ReadonlyArray<T> isn't assignable to a member of T[] (and prevents a direct cast) - but this is where we set up those members so they can be readonly in the future
                processCommentPragmas(sourceFile, sourceText);
                processPragmasIntoFields(sourceFile, reportPragmaDiagnostic);
                sourceFile.commentDirectives = scanner.getCommentDirectives();
                sourceFile.nodeCount = nodeCount;
                sourceFile.identifierCount = identifierCount;
                sourceFile.identifiers = identifiers;
                sourceFile.parseDiagnostics = ts.attachFileToDiagnostics(parseDiagnostics, sourceFile);
                if (jsDocDiagnostics) {
                    sourceFile.jsDocDiagnostics = ts.attachFileToDiagnostics(jsDocDiagnostics, sourceFile);
                }
                if (setParentNodes) {
                    fixupParentReferences(sourceFile);
                }
                return sourceFile;
                function reportPragmaDiagnostic(pos, end, diagnostic) {
                    parseDiagnostics.push(ts.createDetachedDiagnostic(fileName, pos, end, diagnostic));
                }
            }
            function withJSDoc(node, hasJSDoc) {
                return hasJSDoc ? addJSDocComment(node) : node;
            }
            var hasDeprecatedTag = false;
            function addJSDocComment(node) {
                ts.Debug.assert(!node.jsDoc); // Should only be called once per node
                var jsDoc = ts.mapDefined(ts.getJSDocCommentRanges(node, sourceText), function (comment) { return JSDocParser.parseJSDocComment(node, comment.pos, comment.end - comment.pos); });
                if (jsDoc.length)
                    node.jsDoc = jsDoc;
                if (hasDeprecatedTag) {
                    hasDeprecatedTag = false;
                    node.flags |= 134217728 /* Deprecated */;
                }
                return node;
            }
            function reparseTopLevelAwait(sourceFile) {
                var savedSyntaxCursor = syntaxCursor;
                var baseSyntaxCursor = IncrementalParser.createSyntaxCursor(sourceFile);
                syntaxCursor = { currentNode: currentNode };
                var statements = [];
                var savedParseDiagnostics = parseDiagnostics;
                parseDiagnostics = [];
                var pos = 0;
                var start = findNextStatementWithAwait(sourceFile.statements, 0);
                var _loop_3 = function () {
                    // append all statements between pos and start
                    var prevStatement = sourceFile.statements[pos];
                    var nextStatement = sourceFile.statements[start];
                    ts.addRange(statements, sourceFile.statements, pos, start);
                    pos = findNextStatementWithoutAwait(sourceFile.statements, start);
                    // append all diagnostics associated with the copied range
                    var diagnosticStart = ts.findIndex(savedParseDiagnostics, function (diagnostic) { return diagnostic.start >= prevStatement.pos; });
                    var diagnosticEnd = diagnosticStart >= 0 ? ts.findIndex(savedParseDiagnostics, function (diagnostic) { return diagnostic.start >= nextStatement.pos; }, diagnosticStart) : -1;
                    if (diagnosticStart >= 0) {
                        ts.addRange(parseDiagnostics, savedParseDiagnostics, diagnosticStart, diagnosticEnd >= 0 ? diagnosticEnd : undefined);
                    }
                    // reparse all statements between start and pos. We skip existing diagnostics for the same range and allow the parser to generate new ones.
                    speculationHelper(function () {
                        var savedContextFlags = contextFlags;
                        contextFlags |= 32768 /* AwaitContext */;
                        scanner.setTextPos(nextStatement.pos);
                        nextToken();
                        while (token() !== 1 /* EndOfFileToken */) {
                            var startPos = scanner.getStartPos();
                            var statement = parseListElement(0 /* SourceElements */, parseStatement);
                            statements.push(statement);
                            if (startPos === scanner.getStartPos()) {
                                nextToken();
                            }
                            if (pos >= 0) {
                                var nonAwaitStatement = sourceFile.statements[pos];
                                if (statement.end === nonAwaitStatement.pos) {
                                    // done reparsing this section
                                    break;
                                }
                                if (statement.end > nonAwaitStatement.pos) {
                                    // we ate into the next statement, so we must reparse it.
                                    pos = findNextStatementWithoutAwait(sourceFile.statements, pos + 1);
                                }
                            }
                        }
                        contextFlags = savedContextFlags;
                    }, 2 /* Reparse */);
                    // find the next statement containing an `await`
                    start = pos >= 0 ? findNextStatementWithAwait(sourceFile.statements, pos) : -1;
                };
                while (start !== -1) {
                    _loop_3();
                }
                // append all statements between pos and the end of the list
                if (pos >= 0) {
                    var prevStatement_1 = sourceFile.statements[pos];
                    ts.addRange(statements, sourceFile.statements, pos);
                    // append all diagnostics associated with the copied range
                    var diagnosticStart = ts.findIndex(savedParseDiagnostics, function (diagnostic) { return diagnostic.start >= prevStatement_1.pos; });
                    if (diagnosticStart >= 0) {
                        ts.addRange(parseDiagnostics, savedParseDiagnostics, diagnosticStart);
                    }
                }
                syntaxCursor = savedSyntaxCursor;
                return factory.updateSourceFile(sourceFile, ts.setTextRange(factory.createNodeArray(statements), sourceFile.statements));
                function containsPossibleTopLevelAwait(node) {
                    return !(node.flags & 32768 /* AwaitContext */)
                        && !!(node.transformFlags & 16777216 /* ContainsPossibleTopLevelAwait */);
                }
                function findNextStatementWithAwait(statements, start) {
                    for (var i = start; i < statements.length; i++) {
                        if (containsPossibleTopLevelAwait(statements[i])) {
                            return i;
                        }
                    }
                    return -1;
                }
                function findNextStatementWithoutAwait(statements, start) {
                    for (var i = start; i < statements.length; i++) {
                        if (!containsPossibleTopLevelAwait(statements[i])) {
                            return i;
                        }
                    }
                    return -1;
                }
                function currentNode(position) {
                    var node = baseSyntaxCursor.currentNode(position);
                    if (topLevel && node && containsPossibleTopLevelAwait(node)) {
                        node.intersectsChange = true;
                    }
                    return node;
                }
            }
            function fixupParentReferences(rootNode) {
                // normally parent references are set during binding. However, for clients that only need
                // a syntax tree, and no semantic features, then the binding process is an unnecessary
                // overhead.  This functions allows us to set all the parents, without all the expense of
                // binding.
                ts.setParentRecursive(rootNode, /*incremental*/ true);
            }
            Parser.fixupParentReferences = fixupParentReferences;
            function createSourceFile(fileName, languageVersion, scriptKind, isDeclarationFile, statements, endOfFileToken, flags) {
                // code from createNode is inlined here so createNode won't have to deal with special case of creating source files
                // this is quite rare comparing to other nodes and createNode should be as fast as possible
                var sourceFile = factory.createSourceFile(statements, endOfFileToken, flags);
                ts.setTextRangePosWidth(sourceFile, 0, sourceText.length);
                setExternalModuleIndicator(sourceFile);
                // If we parsed this as an external module, it may contain top-level await
                if (!isDeclarationFile && isExternalModule(sourceFile) && sourceFile.transformFlags & 16777216 /* ContainsPossibleTopLevelAwait */) {
                    sourceFile = reparseTopLevelAwait(sourceFile);
                }
                sourceFile.text = sourceText;
                sourceFile.bindDiagnostics = [];
                sourceFile.bindSuggestionDiagnostics = undefined;
                sourceFile.languageVersion = languageVersion;
                sourceFile.fileName = fileName;
                sourceFile.languageVariant = ts.getLanguageVariant(scriptKind);
                sourceFile.isDeclarationFile = isDeclarationFile;
                sourceFile.scriptKind = scriptKind;
                return sourceFile;
            }
            function setContextFlag(val, flag) {
                if (val) {
                    contextFlags |= flag;
                }
                else {
                    contextFlags &= ~flag;
                }
            }
            function setDisallowInContext(val) {
                setContextFlag(val, 4096 /* DisallowInContext */);
            }
            function setYieldContext(val) {
                setContextFlag(val, 8192 /* YieldContext */);
            }
            function setDecoratorContext(val) {
                setContextFlag(val, 16384 /* DecoratorContext */);
            }
            function setAwaitContext(val) {
                setContextFlag(val, 32768 /* AwaitContext */);
            }
            function doOutsideOfContext(context, func) {
                // contextFlagsToClear will contain only the context flags that are
                // currently set that we need to temporarily clear
                // We don't just blindly reset to the previous flags to ensure
                // that we do not mutate cached flags for the incremental
                // parser (ThisNodeHasError, ThisNodeOrAnySubNodesHasError, and
                // HasAggregatedChildData).
                var contextFlagsToClear = context & contextFlags;
                if (contextFlagsToClear) {
                    // clear the requested context flags
                    setContextFlag(/*val*/ false, contextFlagsToClear);
                    var result = func();
                    // restore the context flags we just cleared
                    setContextFlag(/*val*/ true, contextFlagsToClear);
                    return result;
                }
                // no need to do anything special as we are not in any of the requested contexts
                return func();
            }
            function doInsideOfContext(context, func) {
                // contextFlagsToSet will contain only the context flags that
                // are not currently set that we need to temporarily enable.
                // We don't just blindly reset to the previous flags to ensure
                // that we do not mutate cached flags for the incremental
                // parser (ThisNodeHasError, ThisNodeOrAnySubNodesHasError, and
                // HasAggregatedChildData).
                var contextFlagsToSet = context & ~contextFlags;
                if (contextFlagsToSet) {
                    // set the requested context flags
                    setContextFlag(/*val*/ true, contextFlagsToSet);
                    var result = func();
                    // reset the context flags we just set
                    setContextFlag(/*val*/ false, contextFlagsToSet);
                    return result;
                }
                // no need to do anything special as we are already in all of the requested contexts
                return func();
            }
            function allowInAnd(func) {
                return doOutsideOfContext(4096 /* DisallowInContext */, func);
            }
            function disallowInAnd(func) {
                return doInsideOfContext(4096 /* DisallowInContext */, func);
            }
            function doInYieldContext(func) {
                return doInsideOfContext(8192 /* YieldContext */, func);
            }
            function doInDecoratorContext(func) {
                return doInsideOfContext(16384 /* DecoratorContext */, func);
            }
            function doInAwaitContext(func) {
                return doInsideOfContext(32768 /* AwaitContext */, func);
            }
            function doOutsideOfAwaitContext(func) {
                return doOutsideOfContext(32768 /* AwaitContext */, func);
            }
            function doInYieldAndAwaitContext(func) {
                return doInsideOfContext(8192 /* YieldContext */ | 32768 /* AwaitContext */, func);
            }
            function doOutsideOfYieldAndAwaitContext(func) {
                return doOutsideOfContext(8192 /* YieldContext */ | 32768 /* AwaitContext */, func);
            }
            function inContext(flags) {
                return (contextFlags & flags) !== 0;
            }
            function inYieldContext() {
                return inContext(8192 /* YieldContext */);
            }
            function inDisallowInContext() {
                return inContext(4096 /* DisallowInContext */);
            }
            function inDecoratorContext() {
                return inContext(16384 /* DecoratorContext */);
            }
            function inAwaitContext() {
                return inContext(32768 /* AwaitContext */);
            }
            function parseErrorAtCurrentToken(message, arg0) {
                parseErrorAt(scanner.getTokenPos(), scanner.getTextPos(), message, arg0);
            }
            function parseErrorAtPosition(start, length, message, arg0) {
                // Don't report another error if it would just be at the same position as the last error.
                var lastError = ts.lastOrUndefined(parseDiagnostics);
                if (!lastError || start !== lastError.start) {
                    parseDiagnostics.push(ts.createDetachedDiagnostic(fileName, start, length, message, arg0));
                }
                // Mark that we've encountered an error.  We'll set an appropriate bit on the next
                // node we finish so that it can't be reused incrementally.
                parseErrorBeforeNextFinishedNode = true;
            }
            function parseErrorAt(start, end, message, arg0) {
                parseErrorAtPosition(start, end - start, message, arg0);
            }
            function parseErrorAtRange(range, message, arg0) {
                parseErrorAt(range.pos, range.end, message, arg0);
            }
            function scanError(message, length) {
                parseErrorAtPosition(scanner.getTextPos(), length, message);
            }
            function getNodePos() {
                return scanner.getStartPos();
            }
            function hasPrecedingJSDocComment() {
                return scanner.hasPrecedingJSDocComment();
            }
            // Use this function to access the current token instead of reading the currentToken
            // variable. Since function results aren't narrowed in control flow analysis, this ensures
            // that the type checker doesn't make wrong assumptions about the type of the current
            // token (e.g. a call to nextToken() changes the current token but the checker doesn't
            // reason about this side effect).  Mainstream VMs inline simple functions like this, so
            // there is no performance penalty.
            function token() {
                return currentToken;
            }
            function nextTokenWithoutCheck() {
                return currentToken = scanner.scan();
            }
            function nextTokenAnd(func) {
                nextToken();
                return func();
            }
            function nextToken() {
                // if the keyword had an escape
                if (ts.isKeyword(currentToken) && (scanner.hasUnicodeEscape() || scanner.hasExtendedUnicodeEscape())) {
                    // issue a parse error for the escape
                    parseErrorAt(scanner.getTokenPos(), scanner.getTextPos(), ts.Diagnostics.Keywords_cannot_contain_escape_characters);
                }
                return nextTokenWithoutCheck();
            }
            function nextTokenJSDoc() {
                return currentToken = scanner.scanJsDocToken();
            }
            function reScanGreaterToken() {
                return currentToken = scanner.reScanGreaterToken();
            }
            function reScanSlashToken() {
                return currentToken = scanner.reScanSlashToken();
            }
            function reScanTemplateToken(isTaggedTemplate) {
                return currentToken = scanner.reScanTemplateToken(isTaggedTemplate);
            }
            function reScanTemplateHeadOrNoSubstitutionTemplate() {
                return currentToken = scanner.reScanTemplateHeadOrNoSubstitutionTemplate();
            }
            function reScanLessThanToken() {
                return currentToken = scanner.reScanLessThanToken();
            }
            function reScanHashToken() {
                return currentToken = scanner.reScanHashToken();
            }
            function scanJsxIdentifier() {
                return currentToken = scanner.scanJsxIdentifier();
            }
            function scanJsxText() {
                return currentToken = scanner.scanJsxToken();
            }
            function scanJsxAttributeValue() {
                return currentToken = scanner.scanJsxAttributeValue();
            }
            function speculationHelper(callback, speculationKind) {
                // Keep track of the state we'll need to rollback to if lookahead fails (or if the
                // caller asked us to always reset our state).
                var saveToken = currentToken;
                var saveParseDiagnosticsLength = parseDiagnostics.length;
                var saveParseErrorBeforeNextFinishedNode = parseErrorBeforeNextFinishedNode;
                // Note: it is not actually necessary to save/restore the context flags here.  That's
                // because the saving/restoring of these flags happens naturally through the recursive
                // descent nature of our parser.  However, we still store this here just so we can
                // assert that invariant holds.
                var saveContextFlags = contextFlags;
                // If we're only looking ahead, then tell the scanner to only lookahead as well.
                // Otherwise, if we're actually speculatively parsing, then tell the scanner to do the
                // same.
                var result = speculationKind !== 0 /* TryParse */
                    ? scanner.lookAhead(callback)
                    : scanner.tryScan(callback);
                ts.Debug.assert(saveContextFlags === contextFlags);
                // If our callback returned something 'falsy' or we're just looking ahead,
                // then unconditionally restore us to where we were.
                if (!result || speculationKind !== 0 /* TryParse */) {
                    currentToken = saveToken;
                    if (speculationKind !== 2 /* Reparse */) {
                        parseDiagnostics.length = saveParseDiagnosticsLength;
                    }
                    parseErrorBeforeNextFinishedNode = saveParseErrorBeforeNextFinishedNode;
                }
                return result;
            }
            /** Invokes the provided callback then unconditionally restores the parser to the state it
             * was in immediately prior to invoking the callback.  The result of invoking the callback
             * is returned from this function.
             */
            function lookAhead(callback) {
                return speculationHelper(callback, 1 /* Lookahead */);
            }
            /** Invokes the provided callback.  If the callback returns something falsy, then it restores
             * the parser to the state it was in immediately prior to invoking the callback.  If the
             * callback returns something truthy, then the parser state is not rolled back.  The result
             * of invoking the callback is returned from this function.
             */
            function tryParse(callback) {
                return speculationHelper(callback, 0 /* TryParse */);
            }
            function isBindingIdentifier() {
                if (token() === 79 /* Identifier */) {
                    return true;
                }
                // `let await`/`let yield` in [Yield] or [Await] are allowed here and disallowed in the binder.
                return token() > 116 /* LastReservedWord */;
            }
            // Ignore strict mode flag because we will report an error in type checker instead.
            function isIdentifier() {
                if (token() === 79 /* Identifier */) {
                    return true;
                }
                // If we have a 'yield' keyword, and we're in the [yield] context, then 'yield' is
                // considered a keyword and is not an identifier.
                if (token() === 125 /* YieldKeyword */ && inYieldContext()) {
                    return false;
                }
                // If we have a 'await' keyword, and we're in the [Await] context, then 'await' is
                // considered a keyword and is not an identifier.
                if (token() === 131 /* AwaitKeyword */ && inAwaitContext()) {
                    return false;
                }
                return token() > 116 /* LastReservedWord */;
            }
            function parseExpected(kind, diagnosticMessage, shouldAdvance) {
                if (shouldAdvance === void 0) {
                    shouldAdvance = true;
                }
                if (token() === kind) {
                    if (shouldAdvance) {
                        nextToken();
                    }
                    return true;
                }
                // Report specific message if provided with one.  Otherwise, report generic fallback message.
                if (diagnosticMessage) {
                    parseErrorAtCurrentToken(diagnosticMessage);
                }
                else {
                    parseErrorAtCurrentToken(ts.Diagnostics._0_expected, ts.tokenToString(kind));
                }
                return false;
            }
            var viableKeywordSuggestions = Object.keys(ts.textToKeywordObj).filter(function (keyword) { return keyword.length > 2; });
            /**
             * Provides a better error message than the generic "';' expected" if possible for
             * known common variants of a missing semicolon, such as from a mispelled names.
             *
             * @param node Node preceding the expected semicolon location.
             */
            function parseErrorForMissingSemicolonAfter(node) {
                var _a;
                // Tagged template literals are sometimes used in places where only simple strings are allowed, i.e.:
                //   module `M1` {
                //   ^^^^^^^^^^^ This block is parsed as a template literal like module`M1`.
                if (ts.isTaggedTemplateExpression(node)) {
                    parseErrorAt(ts.skipTrivia(sourceText, node.template.pos), node.template.end, ts.Diagnostics.Module_declaration_names_may_only_use_or_quoted_strings);
                    return;
                }
                // Otherwise, if this isn't a well-known keyword-like identifier, give the generic fallback message.
                var expressionText = ts.isIdentifier(node) ? ts.idText(node) : undefined;
                if (!expressionText || !ts.isIdentifierText(expressionText, languageVersion)) {
                    parseErrorAtCurrentToken(ts.Diagnostics._0_expected, ts.tokenToString(26 /* SemicolonToken */));
                    return;
                }
                var pos = ts.skipTrivia(sourceText, node.pos);
                // Some known keywords are likely signs of syntax being used improperly.
                switch (expressionText) {
                    case "const":
                    case "let":
                    case "var":
                        parseErrorAt(pos, node.end, ts.Diagnostics.Variable_declaration_not_allowed_at_this_location);
                        return;
                    case "declare":
                        // If a declared node failed to parse, it would have emitted a diagnostic already.
                        return;
                    case "interface":
                        parseErrorForInvalidName(ts.Diagnostics.Interface_name_cannot_be_0, ts.Diagnostics.Interface_must_be_given_a_name, 18 /* OpenBraceToken */);
                        return;
                    case "is":
                        parseErrorAt(pos, scanner.getTextPos(), ts.Diagnostics.A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods);
                        return;
                    case "module":
                    case "namespace":
                        parseErrorForInvalidName(ts.Diagnostics.Namespace_name_cannot_be_0, ts.Diagnostics.Namespace_must_be_given_a_name, 18 /* OpenBraceToken */);
                        return;
                    case "type":
                        parseErrorForInvalidName(ts.Diagnostics.Type_alias_name_cannot_be_0, ts.Diagnostics.Type_alias_must_be_given_a_name, 63 /* EqualsToken */);
                        return;
                }
                // The user alternatively might have misspelled or forgotten to add a space after a common keyword.
                var suggestion = (_a = ts.getSpellingSuggestion(expressionText, viableKeywordSuggestions, function (n) { return n; })) !== null && _a !== void 0 ? _a : getSpaceSuggestion(expressionText);
                if (suggestion) {
                    parseErrorAt(pos, node.end, ts.Diagnostics.Unknown_keyword_or_identifier_Did_you_mean_0, suggestion);
                    return;
                }
                // Unknown tokens are handled with their own errors in the scanner
                if (token() === 0 /* Unknown */) {
                    return;
                }
                // Otherwise, we know this some kind of unknown word, not just a missing expected semicolon.
                parseErrorAt(pos, node.end, ts.Diagnostics.Unexpected_keyword_or_identifier);
            }
            /**
             * Reports a diagnostic error for the current token being an invalid name.
             *
             * @param blankDiagnostic Diagnostic to report for the case of the name being blank (matched tokenIfBlankName).
             * @param nameDiagnostic Diagnostic to report for all other cases.
             * @param tokenIfBlankName Current token if the name was invalid for being blank (not provided / skipped).
             */
            function parseErrorForInvalidName(nameDiagnostic, blankDiagnostic, tokenIfBlankName) {
                if (token() === tokenIfBlankName) {
                    parseErrorAtCurrentToken(blankDiagnostic);
                }
                else {
                    parseErrorAtCurrentToken(nameDiagnostic, ts.tokenToString(token()));
                }
            }
            function getSpaceSuggestion(expressionText) {
                for (var _i = 0, viableKeywordSuggestions_1 = viableKeywordSuggestions; _i < viableKeywordSuggestions_1.length; _i++) {
                    var keyword = viableKeywordSuggestions_1[_i];
                    if (expressionText.length > keyword.length + 2 && ts.startsWith(expressionText, keyword)) {
                        return keyword + " " + expressionText.slice(keyword.length);
                    }
                }
                return undefined;
            }
            function parseSemicolonAfterPropertyName(name, type, initializer) {
                if (token() === 59 /* AtToken */ && !scanner.hasPrecedingLineBreak()) {
                    parseErrorAtCurrentToken(ts.Diagnostics.Decorators_must_precede_the_name_and_all_keywords_of_property_declarations);
                    return;
                }
                if (token() === 20 /* OpenParenToken */) {
                    parseErrorAtCurrentToken(ts.Diagnostics.Cannot_start_a_function_call_in_a_type_annotation);
                    nextToken();
                    return;
                }
                if (type && !canParseSemicolon()) {
                    if (initializer) {
                        parseErrorAtCurrentToken(ts.Diagnostics._0_expected, ts.tokenToString(26 /* SemicolonToken */));
                    }
                    else {
                        parseErrorAtCurrentToken(ts.Diagnostics.Expected_for_property_initializer);
                    }
                    return;
                }
                if (tryParseSemicolon()) {
                    return;
                }
                // If an initializer was parsed but there is still an error in finding the next semicolon,
                // we generally know there was an error already reported in the initializer...
                //   class Example { a = new Map([), ) }
                //                                ~
                if (initializer) {
                    // ...unless we've found the start of a block after a property declaration, in which
                    // case we can know that regardless of the initializer we should complain on the block.
                    //   class Example { a = 0 {} }
                    //                         ~
                    if (token() === 18 /* OpenBraceToken */) {
                        parseErrorAtCurrentToken(ts.Diagnostics._0_expected, ts.tokenToString(26 /* SemicolonToken */));
                    }
                    return;
                }
                parseErrorForMissingSemicolonAfter(name);
            }
            function parseExpectedJSDoc(kind) {
                if (token() === kind) {
                    nextTokenJSDoc();
                    return true;
                }
                parseErrorAtCurrentToken(ts.Diagnostics._0_expected, ts.tokenToString(kind));
                return false;
            }
            function parseOptional(t) {
                if (token() === t) {
                    nextToken();
                    return true;
                }
                return false;
            }
            function parseOptionalToken(t) {
                if (token() === t) {
                    return parseTokenNode();
                }
                return undefined;
            }
            function parseOptionalTokenJSDoc(t) {
                if (token() === t) {
                    return parseTokenNodeJSDoc();
                }
                return undefined;
            }
            function parseExpectedToken(t, diagnosticMessage, arg0) {
                return parseOptionalToken(t) ||
                    createMissingNode(t, /*reportAtCurrentPosition*/ false, diagnosticMessage || ts.Diagnostics._0_expected, arg0 || ts.tokenToString(t));
            }
            function parseExpectedTokenJSDoc(t) {
                return parseOptionalTokenJSDoc(t) ||
                    createMissingNode(t, /*reportAtCurrentPosition*/ false, ts.Diagnostics._0_expected, ts.tokenToString(t));
            }
            function parseTokenNode() {
                var pos = getNodePos();
                var kind = token();
                nextToken();
                return finishNode(factory.createToken(kind), pos);
            }
            function parseTokenNodeJSDoc() {
                var pos = getNodePos();
                var kind = token();
                nextTokenJSDoc();
                return finishNode(factory.createToken(kind), pos);
            }
            function canParseSemicolon() {
                // If there's a real semicolon, then we can always parse it out.
                if (token() === 26 /* SemicolonToken */) {
                    return true;
                }
                // We can parse out an optional semicolon in ASI cases in the following cases.
                return token() === 19 /* CloseBraceToken */ || token() === 1 /* EndOfFileToken */ || scanner.hasPrecedingLineBreak();
            }
            function tryParseSemicolon() {
                if (!canParseSemicolon()) {
                    return false;
                }
                if (token() === 26 /* SemicolonToken */) {
                    // consume the semicolon if it was explicitly provided.
                    nextToken();
                }
                return true;
            }
            function parseSemicolon() {
                return tryParseSemicolon() || parseExpected(26 /* SemicolonToken */);
            }
            function createNodeArray(elements, pos, end, hasTrailingComma) {
                var array = factory.createNodeArray(elements, hasTrailingComma);
                ts.setTextRangePosEnd(array, pos, end !== null && end !== void 0 ? end : scanner.getStartPos());
                return array;
            }
            function finishNode(node, pos, end) {
                ts.setTextRangePosEnd(node, pos, end !== null && end !== void 0 ? end : scanner.getStartPos());
                if (contextFlags) {
                    node.flags |= contextFlags;
                }
                // Keep track on the node if we encountered an error while parsing it.  If we did, then
                // we cannot reuse the node incrementally.  Once we've marked this node, clear out the
                // flag so that we don't mark any subsequent nodes.
                if (parseErrorBeforeNextFinishedNode) {
                    parseErrorBeforeNextFinishedNode = false;
                    node.flags |= 65536 /* ThisNodeHasError */;
                }
                return node;
            }
            function createMissingNode(kind, reportAtCurrentPosition, diagnosticMessage, arg0) {
                if (reportAtCurrentPosition) {
                    parseErrorAtPosition(scanner.getStartPos(), 0, diagnosticMessage, arg0);
                }
                else if (diagnosticMessage) {
                    parseErrorAtCurrentToken(diagnosticMessage, arg0);
                }
                var pos = getNodePos();
                var result = kind === 79 /* Identifier */ ? factory.createIdentifier("", /*typeArguments*/ undefined, /*originalKeywordKind*/ undefined) :
                    ts.isTemplateLiteralKind(kind) ? factory.createTemplateLiteralLikeNode(kind, "", "", /*templateFlags*/ undefined) :
                        kind === 8 /* NumericLiteral */ ? factory.createNumericLiteral("", /*numericLiteralFlags*/ undefined) :
                            kind === 10 /* StringLiteral */ ? factory.createStringLiteral("", /*isSingleQuote*/ undefined) :
                                kind === 274 /* MissingDeclaration */ ? factory.createMissingDeclaration() :
                                    factory.createToken(kind);
                return finishNode(result, pos);
            }
            function internIdentifier(text) {
                var identifier = identifiers.get(text);
                if (identifier === undefined) {
                    identifiers.set(text, identifier = text);
                }
                return identifier;
            }
            // An identifier that starts with two underscores has an extra underscore character prepended to it to avoid issues
            // with magic property names like '__proto__'. The 'identifiers' object is used to share a single string instance for
            // each identifier in order to reduce memory consumption.
            function createIdentifier(isIdentifier, diagnosticMessage, privateIdentifierDiagnosticMessage) {
                if (isIdentifier) {
                    identifierCount++;
                    var pos = getNodePos();
                    // Store original token kind if it is not just an Identifier so we can report appropriate error later in type checker
                    var originalKeywordKind = token();
                    var text = internIdentifier(scanner.getTokenValue());
                    nextTokenWithoutCheck();
                    return finishNode(factory.createIdentifier(text, /*typeArguments*/ undefined, originalKeywordKind), pos);
                }
                if (token() === 80 /* PrivateIdentifier */) {
                    parseErrorAtCurrentToken(privateIdentifierDiagnosticMessage || ts.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies);
                    return createIdentifier(/*isIdentifier*/ true);
                }
                if (token() === 0 /* Unknown */ && scanner.tryScan(function () { return scanner.reScanInvalidIdentifier() === 79 /* Identifier */; })) {
                    // Scanner has already recorded an 'Invalid character' error, so no need to add another from the parser.
                    return createIdentifier(/*isIdentifier*/ true);
                }
                identifierCount++;
                // Only for end of file because the error gets reported incorrectly on embedded script tags.
                var reportAtCurrentPosition = token() === 1 /* EndOfFileToken */;
                var isReservedWord = scanner.isReservedWord();
                var msgArg = scanner.getTokenText();
                var defaultMessage = isReservedWord ?
                    ts.Diagnostics.Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here :
                    ts.Diagnostics.Identifier_expected;
                return createMissingNode(79 /* Identifier */, reportAtCurrentPosition, diagnosticMessage || defaultMessage, msgArg);
            }
            function parseBindingIdentifier(privateIdentifierDiagnosticMessage) {
                return createIdentifier(isBindingIdentifier(), /*diagnosticMessage*/ undefined, privateIdentifierDiagnosticMessage);
            }
            function parseIdentifier(diagnosticMessage, privateIdentifierDiagnosticMessage) {
                return createIdentifier(isIdentifier(), diagnosticMessage, privateIdentifierDiagnosticMessage);
            }
            function parseIdentifierName(diagnosticMessage) {
                return createIdentifier(ts.tokenIsIdentifierOrKeyword(token()), diagnosticMessage);
            }
            function isLiteralPropertyName() {
                return ts.tokenIsIdentifierOrKeyword(token()) ||
                    token() === 10 /* StringLiteral */ ||
                    token() === 8 /* NumericLiteral */;
            }
            function parsePropertyNameWorker(allowComputedPropertyNames) {
                if (token() === 10 /* StringLiteral */ || token() === 8 /* NumericLiteral */) {
                    var node = parseLiteralNode();
                    node.text = internIdentifier(node.text);
                    return node;
                }
                if (allowComputedPropertyNames && token() === 22 /* OpenBracketToken */) {
                    return parseComputedPropertyName();
                }
                if (token() === 80 /* PrivateIdentifier */) {
                    return parsePrivateIdentifier();
                }
                return parseIdentifierName();
            }
            function parsePropertyName() {
                return parsePropertyNameWorker(/*allowComputedPropertyNames*/ true);
            }
            function parseComputedPropertyName() {
                // PropertyName [Yield]:
                //      LiteralPropertyName
                //      ComputedPropertyName[?Yield]
                var pos = getNodePos();
                parseExpected(22 /* OpenBracketToken */);
                // We parse any expression (including a comma expression). But the grammar
                // says that only an assignment expression is allowed, so the grammar checker
                // will error if it sees a comma expression.
                var expression = allowInAnd(parseExpression);
                parseExpected(23 /* CloseBracketToken */);
                return finishNode(factory.createComputedPropertyName(expression), pos);
            }
            function internPrivateIdentifier(text) {
                var privateIdentifier = privateIdentifiers.get(text);
                if (privateIdentifier === undefined) {
                    privateIdentifiers.set(text, privateIdentifier = text);
                }
                return privateIdentifier;
            }
            function parsePrivateIdentifier() {
                var pos = getNodePos();
                var node = factory.createPrivateIdentifier(internPrivateIdentifier(scanner.getTokenText()));
                nextToken();
                return finishNode(node, pos);
            }
            function parseContextualModifier(t) {
                return token() === t && tryParse(nextTokenCanFollowModifier);
            }
            function nextTokenIsOnSameLineAndCanFollowModifier() {
                nextToken();
                if (scanner.hasPrecedingLineBreak()) {
                    return false;
                }
                return canFollowModifier();
            }
            function nextTokenCanFollowModifier() {
                switch (token()) {
                    case 85 /* ConstKeyword */:
                        // 'const' is only a modifier if followed by 'enum'.
                        return nextToken() === 92 /* EnumKeyword */;
                    case 93 /* ExportKeyword */:
                        nextToken();
                        if (token() === 88 /* DefaultKeyword */) {
                            return lookAhead(nextTokenCanFollowDefaultKeyword);
                        }
                        if (token() === 150 /* TypeKeyword */) {
                            return lookAhead(nextTokenCanFollowExportModifier);
                        }
                        return canFollowExportModifier();
                    case 88 /* DefaultKeyword */:
                        return nextTokenCanFollowDefaultKeyword();
                    case 124 /* StaticKeyword */:
                        return nextTokenIsOnSameLineAndCanFollowModifier();
                    case 135 /* GetKeyword */:
                    case 147 /* SetKeyword */:
                        nextToken();
                        return canFollowModifier();
                    default:
                        return nextTokenIsOnSameLineAndCanFollowModifier();
                }
            }
            function canFollowExportModifier() {
                return token() !== 41 /* AsteriskToken */
                    && token() !== 127 /* AsKeyword */
                    && token() !== 18 /* OpenBraceToken */
                    && canFollowModifier();
            }
            function nextTokenCanFollowExportModifier() {
                nextToken();
                return canFollowExportModifier();
            }
            function parseAnyContextualModifier() {
                return ts.isModifierKind(token()) && tryParse(nextTokenCanFollowModifier);
            }
            function canFollowModifier() {
                return token() === 22 /* OpenBracketToken */
                    || token() === 18 /* OpenBraceToken */
                    || token() === 41 /* AsteriskToken */
                    || token() === 25 /* DotDotDotToken */
                    || isLiteralPropertyName();
            }
            function nextTokenCanFollowDefaultKeyword() {
                nextToken();
                return token() === 84 /* ClassKeyword */ || token() === 98 /* FunctionKeyword */ ||
                    token() === 118 /* InterfaceKeyword */ ||
                    (token() === 126 /* AbstractKeyword */ && lookAhead(nextTokenIsClassKeywordOnSameLine)) ||
                    (token() === 130 /* AsyncKeyword */ && lookAhead(nextTokenIsFunctionKeywordOnSameLine));
            }
            // True if positioned at the start of a list element
            function isListElement(parsingContext, inErrorRecovery) {
                var node = currentNode(parsingContext);
                if (node) {
                    return true;
                }
                switch (parsingContext) {
                    case 0 /* SourceElements */:
                    case 1 /* BlockStatements */:
                    case 3 /* SwitchClauseStatements */:
                        // If we're in error recovery, then we don't want to treat ';' as an empty statement.
                        // The problem is that ';' can show up in far too many contexts, and if we see one
                        // and assume it's a statement, then we may bail out inappropriately from whatever
                        // we're parsing.  For example, if we have a semicolon in the middle of a class, then
                        // we really don't want to assume the class is over and we're on a statement in the
                        // outer module.  We just want to consume and move on.
                        return !(token() === 26 /* SemicolonToken */ && inErrorRecovery) && isStartOfStatement();
                    case 2 /* SwitchClauses */:
                        return token() === 82 /* CaseKeyword */ || token() === 88 /* DefaultKeyword */;
                    case 4 /* TypeMembers */:
                        return lookAhead(isTypeMemberStart);
                    case 5 /* ClassMembers */:
                        // We allow semicolons as class elements (as specified by ES6) as long as we're
                        // not in error recovery.  If we're in error recovery, we don't want an errant
                        // semicolon to be treated as a class member (since they're almost always used
                        // for statements.
                        return lookAhead(isClassMemberStart) || (token() === 26 /* SemicolonToken */ && !inErrorRecovery);
                    case 6 /* EnumMembers */:
                        // Include open bracket computed properties. This technically also lets in indexers,
                        // which would be a candidate for improved error reporting.
                        return token() === 22 /* OpenBracketToken */ || isLiteralPropertyName();
                    case 12 /* ObjectLiteralMembers */:
                        switch (token()) {
                            case 22 /* OpenBracketToken */:
                            case 41 /* AsteriskToken */:
                            case 25 /* DotDotDotToken */:
                            case 24 /* DotToken */: // Not an object literal member, but don't want to close the object (see `tests/cases/fourslash/completionsDotInObjectLiteral.ts`)
                                return true;
                            default:
                                return isLiteralPropertyName();
                        }
                    case 18 /* RestProperties */:
                        return isLiteralPropertyName();
                    case 9 /* ObjectBindingElements */:
                        return token() === 22 /* OpenBracketToken */ || token() === 25 /* DotDotDotToken */ || isLiteralPropertyName();
                    case 7 /* HeritageClauseElement */:
                        // If we see `{ ... }` then only consume it as an expression if it is followed by `,` or `{`
                        // That way we won't consume the body of a class in its heritage clause.
                        if (token() === 18 /* OpenBraceToken */) {
                            return lookAhead(isValidHeritageClauseObjectLiteral);
                        }
                        if (!inErrorRecovery) {
                            return isStartOfLeftHandSideExpression() && !isHeritageClauseExtendsOrImplementsKeyword();
                        }
                        else {
                            // If we're in error recovery we tighten up what we're willing to match.
                            // That way we don't treat something like "this" as a valid heritage clause
                            // element during recovery.
                            return isIdentifier() && !isHeritageClauseExtendsOrImplementsKeyword();
                        }
                    case 8 /* VariableDeclarations */:
                        return isBindingIdentifierOrPrivateIdentifierOrPattern();
                    case 10 /* ArrayBindingElements */:
                        return token() === 27 /* CommaToken */ || token() === 25 /* DotDotDotToken */ || isBindingIdentifierOrPrivateIdentifierOrPattern();
                    case 19 /* TypeParameters */:
                        return isIdentifier();
                    case 15 /* ArrayLiteralMembers */:
                        switch (token()) {
                            case 27 /* CommaToken */:
                            case 24 /* DotToken */: // Not an array literal member, but don't want to close the array (see `tests/cases/fourslash/completionsDotInArrayLiteralInObjectLiteral.ts`)
                                return true;
                        }
                    // falls through
                    case 11 /* ArgumentExpressions */:
                        return token() === 25 /* DotDotDotToken */ || isStartOfExpression();
                    case 16 /* Parameters */:
                        return isStartOfParameter(/*isJSDocParameter*/ false);
                    case 17 /* JSDocParameters */:
                        return isStartOfParameter(/*isJSDocParameter*/ true);
                    case 20 /* TypeArguments */:
                    case 21 /* TupleElementTypes */:
                        return token() === 27 /* CommaToken */ || isStartOfType();
                    case 22 /* HeritageClauses */:
                        return isHeritageClause();
                    case 23 /* ImportOrExportSpecifiers */:
                        return ts.tokenIsIdentifierOrKeyword(token());
                    case 13 /* JsxAttributes */:
                        return ts.tokenIsIdentifierOrKeyword(token()) || token() === 18 /* OpenBraceToken */;
                    case 14 /* JsxChildren */:
                        return true;
                }
                return ts.Debug.fail("Non-exhaustive case in 'isListElement'.");
            }
            function isValidHeritageClauseObjectLiteral() {
                ts.Debug.assert(token() === 18 /* OpenBraceToken */);
                if (nextToken() === 19 /* CloseBraceToken */) {
                    // if we see "extends {}" then only treat the {} as what we're extending (and not
                    // the class body) if we have:
                    //
                    //      extends {} {
                    //      extends {},
                    //      extends {} extends
                    //      extends {} implements
                    var next = nextToken();
                    return next === 27 /* CommaToken */ || next === 18 /* OpenBraceToken */ || next === 94 /* ExtendsKeyword */ || next === 117 /* ImplementsKeyword */;
                }
                return true;
            }
            function nextTokenIsIdentifier() {
                nextToken();
                return isIdentifier();
            }
            function nextTokenIsIdentifierOrKeyword() {
                nextToken();
                return ts.tokenIsIdentifierOrKeyword(token());
            }
            function nextTokenIsIdentifierOrKeywordOrGreaterThan() {
                nextToken();
                return ts.tokenIsIdentifierOrKeywordOrGreaterThan(token());
            }
            function isHeritageClauseExtendsOrImplementsKeyword() {
                if (token() === 117 /* ImplementsKeyword */ ||
                    token() === 94 /* ExtendsKeyword */) {
                    return lookAhead(nextTokenIsStartOfExpression);
                }
                return false;
            }
            function nextTokenIsStartOfExpression() {
                nextToken();
                return isStartOfExpression();
            }
            function nextTokenIsStartOfType() {
                nextToken();
                return isStartOfType();
            }
            // True if positioned at a list terminator
            function isListTerminator(kind) {
                if (token() === 1 /* EndOfFileToken */) {
                    // Being at the end of the file ends all lists.
                    return true;
                }
                switch (kind) {
                    case 1 /* BlockStatements */:
                    case 2 /* SwitchClauses */:
                    case 4 /* TypeMembers */:
                    case 5 /* ClassMembers */:
                    case 6 /* EnumMembers */:
                    case 12 /* ObjectLiteralMembers */:
                    case 9 /* ObjectBindingElements */:
                    case 23 /* ImportOrExportSpecifiers */:
                        return token() === 19 /* CloseBraceToken */;
                    case 3 /* SwitchClauseStatements */:
                        return token() === 19 /* CloseBraceToken */ || token() === 82 /* CaseKeyword */ || token() === 88 /* DefaultKeyword */;
                    case 7 /* HeritageClauseElement */:
                        return token() === 18 /* OpenBraceToken */ || token() === 94 /* ExtendsKeyword */ || token() === 117 /* ImplementsKeyword */;
                    case 8 /* VariableDeclarations */:
                        return isVariableDeclaratorListTerminator();
                    case 19 /* TypeParameters */:
                        // Tokens other than '>' are here for better error recovery
                        return token() === 31 /* GreaterThanToken */ || token() === 20 /* OpenParenToken */ || token() === 18 /* OpenBraceToken */ || token() === 94 /* ExtendsKeyword */ || token() === 117 /* ImplementsKeyword */;
                    case 11 /* ArgumentExpressions */:
                        // Tokens other than ')' are here for better error recovery
                        return token() === 21 /* CloseParenToken */ || token() === 26 /* SemicolonToken */;
                    case 15 /* ArrayLiteralMembers */:
                    case 21 /* TupleElementTypes */:
                    case 10 /* ArrayBindingElements */:
                        return token() === 23 /* CloseBracketToken */;
                    case 17 /* JSDocParameters */:
                    case 16 /* Parameters */:
                    case 18 /* RestProperties */:
                        // Tokens other than ')' and ']' (the latter for index signatures) are here for better error recovery
                        return token() === 21 /* CloseParenToken */ || token() === 23 /* CloseBracketToken */ /*|| token === SyntaxKind.OpenBraceToken*/;
                    case 20 /* TypeArguments */:
                        // All other tokens should cause the type-argument to terminate except comma token
                        return token() !== 27 /* CommaToken */;
                    case 22 /* HeritageClauses */:
                        return token() === 18 /* OpenBraceToken */ || token() === 19 /* CloseBraceToken */;
                    case 13 /* JsxAttributes */:
                        return token() === 31 /* GreaterThanToken */ || token() === 43 /* SlashToken */;
                    case 14 /* JsxChildren */:
                        return token() === 29 /* LessThanToken */ && lookAhead(nextTokenIsSlash);
                    default:
                        return false;
                }
            }
            function isVariableDeclaratorListTerminator() {
                // If we can consume a semicolon (either explicitly, or with ASI), then consider us done
                // with parsing the list of variable declarators.
                if (canParseSemicolon()) {
                    return true;
                }
                // in the case where we're parsing the variable declarator of a 'for-in' statement, we
                // are done if we see an 'in' keyword in front of us. Same with for-of
                if (isInOrOfKeyword(token())) {
                    return true;
                }
                // ERROR RECOVERY TWEAK:
                // For better error recovery, if we see an '=>' then we just stop immediately.  We've got an
                // arrow function here and it's going to be very unlikely that we'll resynchronize and get
                // another variable declaration.
                if (token() === 38 /* EqualsGreaterThanToken */) {
                    return true;
                }
                // Keep trying to parse out variable declarators.
                return false;
            }
            // True if positioned at element or terminator of the current list or any enclosing list
            function isInSomeParsingContext() {
                for (var kind = 0; kind < 24 /* Count */; kind++) {
                    if (parsingContext & (1 << kind)) {
                        if (isListElement(kind, /*inErrorRecovery*/ true) || isListTerminator(kind)) {
                            return true;
                        }
                    }
                }
                return false;
            }
            // Parses a list of elements
            function parseList(kind, parseElement) {
                var saveParsingContext = parsingContext;
                parsingContext |= 1 << kind;
                var list = [];
                var listPos = getNodePos();
                while (!isListTerminator(kind)) {
                    if (isListElement(kind, /*inErrorRecovery*/ false)) {
                        list.push(parseListElement(kind, parseElement));
                        continue;
                    }
                    if (abortParsingListOrMoveToNextToken(kind)) {
                        break;
                    }
                }
                parsingContext = saveParsingContext;
                return createNodeArray(list, listPos);
            }
            function parseListElement(parsingContext, parseElement) {
                var node = currentNode(parsingContext);
                if (node) {
                    return consumeNode(node);
                }
                return parseElement();
            }
            function currentNode(parsingContext) {
                // If we don't have a cursor or the parsing context isn't reusable, there's nothing to reuse.
                //
                // If there is an outstanding parse error that we've encountered, but not attached to
                // some node, then we cannot get a node from the old source tree.  This is because we
                // want to mark the next node we encounter as being unusable.
                //
                // Note: This may be too conservative.  Perhaps we could reuse the node and set the bit
                // on it (or its leftmost child) as having the error.  For now though, being conservative
                // is nice and likely won't ever affect perf.
                if (!syntaxCursor || !isReusableParsingContext(parsingContext) || parseErrorBeforeNextFinishedNode) {
                    return undefined;
                }
                var node = syntaxCursor.currentNode(scanner.getStartPos());
                // Can't reuse a missing node.
                // Can't reuse a node that intersected the change range.
                // Can't reuse a node that contains a parse error.  This is necessary so that we
                // produce the same set of errors again.
                if (ts.nodeIsMissing(node) || node.intersectsChange || ts.containsParseError(node)) {
                    return undefined;
                }
                // We can only reuse a node if it was parsed under the same strict mode that we're
                // currently in.  i.e. if we originally parsed a node in non-strict mode, but then
                // the user added 'using strict' at the top of the file, then we can't use that node
                // again as the presence of strict mode may cause us to parse the tokens in the file
                // differently.
                //
                // Note: we *can* reuse tokens when the strict mode changes.  That's because tokens
                // are unaffected by strict mode.  It's just the parser will decide what to do with it
                // differently depending on what mode it is in.
                //
                // This also applies to all our other context flags as well.
                var nodeContextFlags = node.flags & 25358336 /* ContextFlags */;
                if (nodeContextFlags !== contextFlags) {
                    return undefined;
                }
                // Ok, we have a node that looks like it could be reused.  Now verify that it is valid
                // in the current list parsing context that we're currently at.
                if (!canReuseNode(node, parsingContext)) {
                    return undefined;
                }
                if (node.jsDocCache) {
                    // jsDocCache may include tags from parent nodes, which might have been modified.
                    node.jsDocCache = undefined;
                }
                return node;
            }
            function consumeNode(node) {
                // Move the scanner so it is after the node we just consumed.
                scanner.setTextPos(node.end);
                nextToken();
                return node;
            }
            function isReusableParsingContext(parsingContext) {
                switch (parsingContext) {
                    case 5 /* ClassMembers */:
                    case 2 /* SwitchClauses */:
                    case 0 /* SourceElements */:
                    case 1 /* BlockStatements */:
                    case 3 /* SwitchClauseStatements */:
                    case 6 /* EnumMembers */:
                    case 4 /* TypeMembers */:
                    case 8 /* VariableDeclarations */:
                    case 17 /* JSDocParameters */:
                    case 16 /* Parameters */:
                        return true;
                }
                return false;
            }
            function canReuseNode(node, parsingContext) {
                switch (parsingContext) {
                    case 5 /* ClassMembers */:
                        return isReusableClassMember(node);
                    case 2 /* SwitchClauses */:
                        return isReusableSwitchClause(node);
                    case 0 /* SourceElements */:
                    case 1 /* BlockStatements */:
                    case 3 /* SwitchClauseStatements */:
                        return isReusableStatement(node);
                    case 6 /* EnumMembers */:
                        return isReusableEnumMember(node);
                    case 4 /* TypeMembers */:
                        return isReusableTypeMember(node);
                    case 8 /* VariableDeclarations */:
                        return isReusableVariableDeclaration(node);
                    case 17 /* JSDocParameters */:
                    case 16 /* Parameters */:
                        return isReusableParameter(node);
                    // Any other lists we do not care about reusing nodes in.  But feel free to add if
                    // you can do so safely.  Danger areas involve nodes that may involve speculative
                    // parsing.  If speculative parsing is involved with the node, then the range the
                    // parser reached while looking ahead might be in the edited range (see the example
                    // in canReuseVariableDeclaratorNode for a good case of this).
                    // case ParsingContext.HeritageClauses:
                    // This would probably be safe to reuse.  There is no speculative parsing with
                    // heritage clauses.
                    // case ParsingContext.TypeParameters:
                    // This would probably be safe to reuse.  There is no speculative parsing with
                    // type parameters.  Note that that's because type *parameters* only occur in
                    // unambiguous *type* contexts.  While type *arguments* occur in very ambiguous
                    // *expression* contexts.
                    // case ParsingContext.TupleElementTypes:
                    // This would probably be safe to reuse.  There is no speculative parsing with
                    // tuple types.
                    // Technically, type argument list types are probably safe to reuse.  While
                    // speculative parsing is involved with them (since type argument lists are only
                    // produced from speculative parsing a < as a type argument list), we only have
                    // the types because speculative parsing succeeded.  Thus, the lookahead never
                    // went past the end of the list and rewound.
                    // case ParsingContext.TypeArguments:
                    // Note: these are almost certainly not safe to ever reuse.  Expressions commonly
                    // need a large amount of lookahead, and we should not reuse them as they may
                    // have actually intersected the edit.
                    // case ParsingContext.ArgumentExpressions:
                    // This is not safe to reuse for the same reason as the 'AssignmentExpression'
                    // cases.  i.e. a property assignment may end with an expression, and thus might
                    // have lookahead far beyond it's old node.
                    // case ParsingContext.ObjectLiteralMembers:
                    // This is probably not safe to reuse.  There can be speculative parsing with
                    // type names in a heritage clause.  There can be generic names in the type
                    // name list, and there can be left hand side expressions (which can have type
                    // arguments.)
                    // case ParsingContext.HeritageClauseElement:
                    // Perhaps safe to reuse, but it's unlikely we'd see more than a dozen attributes
                    // on any given element. Same for children.
                    // case ParsingContext.JsxAttributes:
                    // case ParsingContext.JsxChildren:
                }
                return false;
            }
            function isReusableClassMember(node) {
                if (node) {
                    switch (node.kind) {
                        case 169 /* Constructor */:
                        case 174 /* IndexSignature */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                        case 165 /* PropertyDeclaration */:
                        case 232 /* SemicolonClassElement */:
                            return true;
                        case 167 /* MethodDeclaration */:
                            // Method declarations are not necessarily reusable.  An object-literal
                            // may have a method calls "constructor(...)" and we must reparse that
                            // into an actual .ConstructorDeclaration.
                            var methodDeclaration = node;
                            var nameIsConstructor = methodDeclaration.name.kind === 79 /* Identifier */ &&
                                methodDeclaration.name.originalKeywordKind === 133 /* ConstructorKeyword */;
                            return !nameIsConstructor;
                    }
                }
                return false;
            }
            function isReusableSwitchClause(node) {
                if (node) {
                    switch (node.kind) {
                        case 287 /* CaseClause */:
                        case 288 /* DefaultClause */:
                            return true;
                    }
                }
                return false;
            }
            function isReusableStatement(node) {
                if (node) {
                    switch (node.kind) {
                        case 254 /* FunctionDeclaration */:
                        case 235 /* VariableStatement */:
                        case 233 /* Block */:
                        case 237 /* IfStatement */:
                        case 236 /* ExpressionStatement */:
                        case 249 /* ThrowStatement */:
                        case 245 /* ReturnStatement */:
                        case 247 /* SwitchStatement */:
                        case 244 /* BreakStatement */:
                        case 243 /* ContinueStatement */:
                        case 241 /* ForInStatement */:
                        case 242 /* ForOfStatement */:
                        case 240 /* ForStatement */:
                        case 239 /* WhileStatement */:
                        case 246 /* WithStatement */:
                        case 234 /* EmptyStatement */:
                        case 250 /* TryStatement */:
                        case 248 /* LabeledStatement */:
                        case 238 /* DoStatement */:
                        case 251 /* DebuggerStatement */:
                        case 264 /* ImportDeclaration */:
                        case 263 /* ImportEqualsDeclaration */:
                        case 270 /* ExportDeclaration */:
                        case 269 /* ExportAssignment */:
                        case 259 /* ModuleDeclaration */:
                        case 255 /* ClassDeclaration */:
                        case 256 /* InterfaceDeclaration */:
                        case 258 /* EnumDeclaration */:
                        case 257 /* TypeAliasDeclaration */:
                            return true;
                    }
                }
                return false;
            }
            function isReusableEnumMember(node) {
                return node.kind === 294 /* EnumMember */;
            }
            function isReusableTypeMember(node) {
                if (node) {
                    switch (node.kind) {
                        case 173 /* ConstructSignature */:
                        case 166 /* MethodSignature */:
                        case 174 /* IndexSignature */:
                        case 164 /* PropertySignature */:
                        case 172 /* CallSignature */:
                            return true;
                    }
                }
                return false;
            }
            function isReusableVariableDeclaration(node) {
                if (node.kind !== 252 /* VariableDeclaration */) {
                    return false;
                }
                // Very subtle incremental parsing bug.  Consider the following code:
                //
                //      let v = new List < A, B
                //
                // This is actually legal code.  It's a list of variable declarators "v = new List<A"
                // on one side and "B" on the other. If you then change that to:
                //
                //      let v = new List < A, B >()
                //
                // then we have a problem.  "v = new List<A" doesn't intersect the change range, so we
                // start reparsing at "B" and we completely fail to handle this properly.
                //
                // In order to prevent this, we do not allow a variable declarator to be reused if it
                // has an initializer.
                var variableDeclarator = node;
                return variableDeclarator.initializer === undefined;
            }
            function isReusableParameter(node) {
                if (node.kind !== 162 /* Parameter */) {
                    return false;
                }
                // See the comment in isReusableVariableDeclaration for why we do this.
                var parameter = node;
                return parameter.initializer === undefined;
            }
            // Returns true if we should abort parsing.
            function abortParsingListOrMoveToNextToken(kind) {
                parsingContextErrors(kind);
                if (isInSomeParsingContext()) {
                    return true;
                }
                nextToken();
                return false;
            }
            function parsingContextErrors(context) {
                switch (context) {
                    case 0 /* SourceElements */:
                        return token() === 88 /* DefaultKeyword */
                            ? parseErrorAtCurrentToken(ts.Diagnostics._0_expected, ts.tokenToString(93 /* ExportKeyword */))
                            : parseErrorAtCurrentToken(ts.Diagnostics.Declaration_or_statement_expected);
                    case 1 /* BlockStatements */: return parseErrorAtCurrentToken(ts.Diagnostics.Declaration_or_statement_expected);
                    case 2 /* SwitchClauses */: return parseErrorAtCurrentToken(ts.Diagnostics.case_or_default_expected);
                    case 3 /* SwitchClauseStatements */: return parseErrorAtCurrentToken(ts.Diagnostics.Statement_expected);
                    case 18 /* RestProperties */: // fallthrough
                    case 4 /* TypeMembers */: return parseErrorAtCurrentToken(ts.Diagnostics.Property_or_signature_expected);
                    case 5 /* ClassMembers */: return parseErrorAtCurrentToken(ts.Diagnostics.Unexpected_token_A_constructor_method_accessor_or_property_was_expected);
                    case 6 /* EnumMembers */: return parseErrorAtCurrentToken(ts.Diagnostics.Enum_member_expected);
                    case 7 /* HeritageClauseElement */: return parseErrorAtCurrentToken(ts.Diagnostics.Expression_expected);
                    case 8 /* VariableDeclarations */:
                        return ts.isKeyword(token())
                            ? parseErrorAtCurrentToken(ts.Diagnostics._0_is_not_allowed_as_a_variable_declaration_name, ts.tokenToString(token()))
                            : parseErrorAtCurrentToken(ts.Diagnostics.Variable_declaration_expected);
                    case 9 /* ObjectBindingElements */: return parseErrorAtCurrentToken(ts.Diagnostics.Property_destructuring_pattern_expected);
                    case 10 /* ArrayBindingElements */: return parseErrorAtCurrentToken(ts.Diagnostics.Array_element_destructuring_pattern_expected);
                    case 11 /* ArgumentExpressions */: return parseErrorAtCurrentToken(ts.Diagnostics.Argument_expression_expected);
                    case 12 /* ObjectLiteralMembers */: return parseErrorAtCurrentToken(ts.Diagnostics.Property_assignment_expected);
                    case 15 /* ArrayLiteralMembers */: return parseErrorAtCurrentToken(ts.Diagnostics.Expression_or_comma_expected);
                    case 17 /* JSDocParameters */: return parseErrorAtCurrentToken(ts.Diagnostics.Parameter_declaration_expected);
                    case 16 /* Parameters */: return parseErrorAtCurrentToken(ts.Diagnostics.Parameter_declaration_expected);
                    case 19 /* TypeParameters */: return parseErrorAtCurrentToken(ts.Diagnostics.Type_parameter_declaration_expected);
                    case 20 /* TypeArguments */: return parseErrorAtCurrentToken(ts.Diagnostics.Type_argument_expected);
                    case 21 /* TupleElementTypes */: return parseErrorAtCurrentToken(ts.Diagnostics.Type_expected);
                    case 22 /* HeritageClauses */: return parseErrorAtCurrentToken(ts.Diagnostics.Unexpected_token_expected);
                    case 23 /* ImportOrExportSpecifiers */: return parseErrorAtCurrentToken(ts.Diagnostics.Identifier_expected);
                    case 13 /* JsxAttributes */: return parseErrorAtCurrentToken(ts.Diagnostics.Identifier_expected);
                    case 14 /* JsxChildren */: return parseErrorAtCurrentToken(ts.Diagnostics.Identifier_expected);
                    default: return [undefined]; // TODO: GH#18217 `default: Debug.assertNever(context);`
                }
            }
            // Parses a comma-delimited list of elements
            function parseDelimitedList(kind, parseElement, considerSemicolonAsDelimiter) {
                var saveParsingContext = parsingContext;
                parsingContext |= 1 << kind;
                var list = [];
                var listPos = getNodePos();
                var commaStart = -1; // Meaning the previous token was not a comma
                while (true) {
                    if (isListElement(kind, /*inErrorRecovery*/ false)) {
                        var startPos = scanner.getStartPos();
                        list.push(parseListElement(kind, parseElement));
                        commaStart = scanner.getTokenPos();
                        if (parseOptional(27 /* CommaToken */)) {
                            // No need to check for a zero length node since we know we parsed a comma
                            continue;
                        }
                        commaStart = -1; // Back to the state where the last token was not a comma
                        if (isListTerminator(kind)) {
                            break;
                        }
                        // We didn't get a comma, and the list wasn't terminated, explicitly parse
                        // out a comma so we give a good error message.
                        parseExpected(27 /* CommaToken */, getExpectedCommaDiagnostic(kind));
                        // If the token was a semicolon, and the caller allows that, then skip it and
                        // continue.  This ensures we get back on track and don't result in tons of
                        // parse errors.  For example, this can happen when people do things like use
                        // a semicolon to delimit object literal members.   Note: we'll have already
                        // reported an error when we called parseExpected above.
                        if (considerSemicolonAsDelimiter && token() === 26 /* SemicolonToken */ && !scanner.hasPrecedingLineBreak()) {
                            nextToken();
                        }
                        if (startPos === scanner.getStartPos()) {
                            // What we're parsing isn't actually remotely recognizable as a element and we've consumed no tokens whatsoever
                            // Consume a token to advance the parser in some way and avoid an infinite loop
                            // This can happen when we're speculatively parsing parenthesized expressions which we think may be arrow functions,
                            // or when a modifier keyword which is disallowed as a parameter name (ie, `static` in strict mode) is supplied
                            nextToken();
                        }
                        continue;
                    }
                    if (isListTerminator(kind)) {
                        break;
                    }
                    if (abortParsingListOrMoveToNextToken(kind)) {
                        break;
                    }
                }
                parsingContext = saveParsingContext;
                // Recording the trailing comma is deliberately done after the previous
                // loop, and not just if we see a list terminator. This is because the list
                // may have ended incorrectly, but it is still important to know if there
                // was a trailing comma.
                // Check if the last token was a comma.
                // Always preserve a trailing comma by marking it on the NodeArray
                return createNodeArray(list, listPos, /*end*/ undefined, commaStart >= 0);
            }
            function getExpectedCommaDiagnostic(kind) {
                return kind === 6 /* EnumMembers */ ? ts.Diagnostics.An_enum_member_name_must_be_followed_by_a_or : undefined;
            }
            function createMissingList() {
                var list = createNodeArray([], getNodePos());
                list.isMissingList = true;
                return list;
            }
            function isMissingList(arr) {
                return !!arr.isMissingList;
            }
            function parseBracketedList(kind, parseElement, open, close) {
                if (parseExpected(open)) {
                    var result = parseDelimitedList(kind, parseElement);
                    parseExpected(close);
                    return result;
                }
                return createMissingList();
            }
            function parseEntityName(allowReservedWords, diagnosticMessage) {
                var pos = getNodePos();
                var entity = allowReservedWords ? parseIdentifierName(diagnosticMessage) : parseIdentifier(diagnosticMessage);
                var dotPos = getNodePos();
                while (parseOptional(24 /* DotToken */)) {
                    if (token() === 29 /* LessThanToken */) {
                        // the entity is part of a JSDoc-style generic, so record the trailing dot for later error reporting
                        entity.jsdocDotPos = dotPos;
                        break;
                    }
                    dotPos = getNodePos();
                    entity = finishNode(factory.createQualifiedName(entity, parseRightSideOfDot(allowReservedWords, /* allowPrivateIdentifiers */ false)), pos);
                }
                return entity;
            }
            function createQualifiedName(entity, name) {
                return finishNode(factory.createQualifiedName(entity, name), entity.pos);
            }
            function parseRightSideOfDot(allowIdentifierNames, allowPrivateIdentifiers) {
                // Technically a keyword is valid here as all identifiers and keywords are identifier names.
                // However, often we'll encounter this in error situations when the identifier or keyword
                // is actually starting another valid construct.
                //
                // So, we check for the following specific case:
                //
                //      name.
                //      identifierOrKeyword identifierNameOrKeyword
                //
                // Note: the newlines are important here.  For example, if that above code
                // were rewritten into:
                //
                //      name.identifierOrKeyword
                //      identifierNameOrKeyword
                //
                // Then we would consider it valid.  That's because ASI would take effect and
                // the code would be implicitly: "name.identifierOrKeyword; identifierNameOrKeyword".
                // In the first case though, ASI will not take effect because there is not a
                // line terminator after the identifier or keyword.
                if (scanner.hasPrecedingLineBreak() && ts.tokenIsIdentifierOrKeyword(token())) {
                    var matchesPattern = lookAhead(nextTokenIsIdentifierOrKeywordOnSameLine);
                    if (matchesPattern) {
                        // Report that we need an identifier.  However, report it right after the dot,
                        // and not on the next token.  This is because the next token might actually
                        // be an identifier and the error would be quite confusing.
                        return createMissingNode(79 /* Identifier */, /*reportAtCurrentPosition*/ true, ts.Diagnostics.Identifier_expected);
                    }
                }
                if (token() === 80 /* PrivateIdentifier */) {
                    var node = parsePrivateIdentifier();
                    return allowPrivateIdentifiers ? node : createMissingNode(79 /* Identifier */, /*reportAtCurrentPosition*/ true, ts.Diagnostics.Identifier_expected);
                }
                return allowIdentifierNames ? parseIdentifierName() : parseIdentifier();
            }
            function parseTemplateSpans(isTaggedTemplate) {
                var pos = getNodePos();
                var list = [];
                var node;
                do {
                    node = parseTemplateSpan(isTaggedTemplate);
                    list.push(node);
                } while (node.literal.kind === 16 /* TemplateMiddle */);
                return createNodeArray(list, pos);
            }
            function parseTemplateExpression(isTaggedTemplate) {
                var pos = getNodePos();
                return finishNode(factory.createTemplateExpression(parseTemplateHead(isTaggedTemplate), parseTemplateSpans(isTaggedTemplate)), pos);
            }
            function parseTemplateType() {
                var pos = getNodePos();
                return finishNode(factory.createTemplateLiteralType(parseTemplateHead(/*isTaggedTemplate*/ false), parseTemplateTypeSpans()), pos);
            }
            function parseTemplateTypeSpans() {
                var pos = getNodePos();
                var list = [];
                var node;
                do {
                    node = parseTemplateTypeSpan();
                    list.push(node);
                } while (node.literal.kind === 16 /* TemplateMiddle */);
                return createNodeArray(list, pos);
            }
            function parseTemplateTypeSpan() {
                var pos = getNodePos();
                return finishNode(factory.createTemplateLiteralTypeSpan(parseType(), parseLiteralOfTemplateSpan(/*isTaggedTemplate*/ false)), pos);
            }
            function parseLiteralOfTemplateSpan(isTaggedTemplate) {
                if (token() === 19 /* CloseBraceToken */) {
                    reScanTemplateToken(isTaggedTemplate);
                    return parseTemplateMiddleOrTemplateTail();
                }
                else {
                    // TODO(rbuckton): Do we need to call `parseExpectedToken` or can we just call `createMissingNode` directly?
                    return parseExpectedToken(17 /* TemplateTail */, ts.Diagnostics._0_expected, ts.tokenToString(19 /* CloseBraceToken */));
                }
            }
            function parseTemplateSpan(isTaggedTemplate) {
                var pos = getNodePos();
                return finishNode(factory.createTemplateSpan(allowInAnd(parseExpression), parseLiteralOfTemplateSpan(isTaggedTemplate)), pos);
            }
            function parseLiteralNode() {
                return parseLiteralLikeNode(token());
            }
            function parseTemplateHead(isTaggedTemplate) {
                if (isTaggedTemplate) {
                    reScanTemplateHeadOrNoSubstitutionTemplate();
                }
                var fragment = parseLiteralLikeNode(token());
                ts.Debug.assert(fragment.kind === 15 /* TemplateHead */, "Template head has wrong token kind");
                return fragment;
            }
            function parseTemplateMiddleOrTemplateTail() {
                var fragment = parseLiteralLikeNode(token());
                ts.Debug.assert(fragment.kind === 16 /* TemplateMiddle */ || fragment.kind === 17 /* TemplateTail */, "Template fragment has wrong token kind");
                return fragment;
            }
            function getTemplateLiteralRawText(kind) {
                var isLast = kind === 14 /* NoSubstitutionTemplateLiteral */ || kind === 17 /* TemplateTail */;
                var tokenText = scanner.getTokenText();
                return tokenText.substring(1, tokenText.length - (scanner.isUnterminated() ? 0 : isLast ? 1 : 2));
            }
            function parseLiteralLikeNode(kind) {
                var pos = getNodePos();
                var node = ts.isTemplateLiteralKind(kind) ? factory.createTemplateLiteralLikeNode(kind, scanner.getTokenValue(), getTemplateLiteralRawText(kind), scanner.getTokenFlags() & 2048 /* TemplateLiteralLikeFlags */) :
                    // Octal literals are not allowed in strict mode or ES5
                    // Note that theoretically the following condition would hold true literals like 009,
                    // which is not octal. But because of how the scanner separates the tokens, we would
                    // never get a token like this. Instead, we would get 00 and 9 as two separate tokens.
                    // We also do not need to check for negatives because any prefix operator would be part of a
                    // parent unary expression.
                    kind === 8 /* NumericLiteral */ ? factory.createNumericLiteral(scanner.getTokenValue(), scanner.getNumericLiteralFlags()) :
                        kind === 10 /* StringLiteral */ ? factory.createStringLiteral(scanner.getTokenValue(), /*isSingleQuote*/ undefined, scanner.hasExtendedUnicodeEscape()) :
                            ts.isLiteralKind(kind) ? factory.createLiteralLikeNode(kind, scanner.getTokenValue()) :
                                ts.Debug.fail();
                if (scanner.hasExtendedUnicodeEscape()) {
                    node.hasExtendedUnicodeEscape = true;
                }
                if (scanner.isUnterminated()) {
                    node.isUnterminated = true;
                }
                nextToken();
                return finishNode(node, pos);
            }
            // TYPES
            function parseEntityNameOfTypeReference() {
                return parseEntityName(/*allowReservedWords*/ true, ts.Diagnostics.Type_expected);
            }
            function parseTypeArgumentsOfTypeReference() {
                if (!scanner.hasPrecedingLineBreak() && reScanLessThanToken() === 29 /* LessThanToken */) {
                    return parseBracketedList(20 /* TypeArguments */, parseType, 29 /* LessThanToken */, 31 /* GreaterThanToken */);
                }
            }
            function parseTypeReference() {
                var pos = getNodePos();
                return finishNode(factory.createTypeReferenceNode(parseEntityNameOfTypeReference(), parseTypeArgumentsOfTypeReference()), pos);
            }
            // If true, we should abort parsing an error function.
            function typeHasArrowFunctionBlockingParseError(node) {
                switch (node.kind) {
                    case 176 /* TypeReference */:
                        return ts.nodeIsMissing(node.typeName);
                    case 177 /* FunctionType */:
                    case 178 /* ConstructorType */: {
                        var _a = node, parameters = _a.parameters, type = _a.type;
                        return isMissingList(parameters) || typeHasArrowFunctionBlockingParseError(type);
                    }
                    case 189 /* ParenthesizedType */:
                        return typeHasArrowFunctionBlockingParseError(node.type);
                    default:
                        return false;
                }
            }
            function parseThisTypePredicate(lhs) {
                nextToken();
                return finishNode(factory.createTypePredicateNode(/*assertsModifier*/ undefined, lhs, parseType()), lhs.pos);
            }
            function parseThisTypeNode() {
                var pos = getNodePos();
                nextToken();
                return finishNode(factory.createThisTypeNode(), pos);
            }
            function parseJSDocAllType() {
                var pos = getNodePos();
                nextToken();
                return finishNode(factory.createJSDocAllType(), pos);
            }
            function parseJSDocNonNullableType() {
                var pos = getNodePos();
                nextToken();
                return finishNode(factory.createJSDocNonNullableType(parseNonArrayType()), pos);
            }
            function parseJSDocUnknownOrNullableType() {
                var pos = getNodePos();
                // skip the ?
                nextToken();
                // Need to lookahead to decide if this is a nullable or unknown type.
                // Here are cases where we'll pick the unknown type:
                //
                //      Foo(?,
                //      { a: ? }
                //      Foo(?)
                //      Foo<?>
                //      Foo(?=
                //      (?|
                if (token() === 27 /* CommaToken */ ||
                    token() === 19 /* CloseBraceToken */ ||
                    token() === 21 /* CloseParenToken */ ||
                    token() === 31 /* GreaterThanToken */ ||
                    token() === 63 /* EqualsToken */ ||
                    token() === 51 /* BarToken */) {
                    return finishNode(factory.createJSDocUnknownType(), pos);
                }
                else {
                    return finishNode(factory.createJSDocNullableType(parseType()), pos);
                }
            }
            function parseJSDocFunctionType() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                if (lookAhead(nextTokenIsOpenParen)) {
                    nextToken();
                    var parameters = parseParameters(4 /* Type */ | 32 /* JSDoc */);
                    var type = parseReturnType(58 /* ColonToken */, /*isType*/ false);
                    return withJSDoc(finishNode(factory.createJSDocFunctionType(parameters, type), pos), hasJSDoc);
                }
                return finishNode(factory.createTypeReferenceNode(parseIdentifierName(), /*typeArguments*/ undefined), pos);
            }
            function parseJSDocParameter() {
                var pos = getNodePos();
                var name;
                if (token() === 108 /* ThisKeyword */ || token() === 103 /* NewKeyword */) {
                    name = parseIdentifierName();
                    parseExpected(58 /* ColonToken */);
                }
                return finishNode(factory.createParameterDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*dotDotDotToken*/ undefined, 
                // TODO(rbuckton): JSDoc parameters don't have names (except `this`/`new`), should we manufacture an empty identifier?
                name, 
                /*questionToken*/ undefined, parseJSDocType(), 
                /*initializer*/ undefined), pos);
            }
            function parseJSDocType() {
                scanner.setInJSDocType(true);
                var pos = getNodePos();
                if (parseOptional(140 /* ModuleKeyword */)) {
                    // TODO(rbuckton): We never set the type for a JSDocNamepathType. What should we put here?
                    var moduleTag = factory.createJSDocNamepathType(/*type*/ undefined);
                    terminate: while (true) {
                        switch (token()) {
                            case 19 /* CloseBraceToken */:
                            case 1 /* EndOfFileToken */:
                            case 27 /* CommaToken */:
                            case 5 /* WhitespaceTrivia */:
                                break terminate;
                            default:
                                nextTokenJSDoc();
                        }
                    }
                    scanner.setInJSDocType(false);
                    return finishNode(moduleTag, pos);
                }
                var hasDotDotDot = parseOptional(25 /* DotDotDotToken */);
                var type = parseTypeOrTypePredicate();
                scanner.setInJSDocType(false);
                if (hasDotDotDot) {
                    type = finishNode(factory.createJSDocVariadicType(type), pos);
                }
                if (token() === 63 /* EqualsToken */) {
                    nextToken();
                    return finishNode(factory.createJSDocOptionalType(type), pos);
                }
                return type;
            }
            function parseTypeQuery() {
                var pos = getNodePos();
                parseExpected(112 /* TypeOfKeyword */);
                return finishNode(factory.createTypeQueryNode(parseEntityName(/*allowReservedWords*/ true)), pos);
            }
            function parseTypeParameter() {
                var pos = getNodePos();
                var name = parseIdentifier();
                var constraint;
                var expression;
                if (parseOptional(94 /* ExtendsKeyword */)) {
                    // It's not uncommon for people to write improper constraints to a generic.  If the
                    // user writes a constraint that is an expression and not an actual type, then parse
                    // it out as an expression (so we can recover well), but report that a type is needed
                    // instead.
                    if (isStartOfType() || !isStartOfExpression()) {
                        constraint = parseType();
                    }
                    else {
                        // It was not a type, and it looked like an expression.  Parse out an expression
                        // here so we recover well.  Note: it is important that we call parseUnaryExpression
                        // and not parseExpression here.  If the user has:
                        //
                        //      <T extends "">
                        //
                        // We do *not* want to consume the `>` as we're consuming the expression for "".
                        expression = parseUnaryExpressionOrHigher();
                    }
                }
                var defaultType = parseOptional(63 /* EqualsToken */) ? parseType() : undefined;
                var node = factory.createTypeParameterDeclaration(name, constraint, defaultType);
                node.expression = expression;
                return finishNode(node, pos);
            }
            function parseTypeParameters() {
                if (token() === 29 /* LessThanToken */) {
                    return parseBracketedList(19 /* TypeParameters */, parseTypeParameter, 29 /* LessThanToken */, 31 /* GreaterThanToken */);
                }
            }
            function isStartOfParameter(isJSDocParameter) {
                return token() === 25 /* DotDotDotToken */ ||
                    isBindingIdentifierOrPrivateIdentifierOrPattern() ||
                    ts.isModifierKind(token()) ||
                    token() === 59 /* AtToken */ ||
                    isStartOfType(/*inStartOfParameter*/ !isJSDocParameter);
            }
            function parseNameOfParameter(modifiers) {
                // FormalParameter [Yield,Await]:
                //      BindingElement[?Yield,?Await]
                var name = parseIdentifierOrPattern(ts.Diagnostics.Private_identifiers_cannot_be_used_as_parameters);
                if (ts.getFullWidth(name) === 0 && !ts.some(modifiers) && ts.isModifierKind(token())) {
                    // in cases like
                    // 'use strict'
                    // function foo(static)
                    // isParameter('static') === true, because of isModifier('static')
                    // however 'static' is not a legal identifier in a strict mode.
                    // so result of this function will be ParameterDeclaration (flags = 0, name = missing, type = undefined, initializer = undefined)
                    // and current token will not change => parsing of the enclosing parameter list will last till the end of time (or OOM)
                    // to avoid this we'll advance cursor to the next token.
                    nextToken();
                }
                return name;
            }
            function parseParameterInOuterAwaitContext() {
                return parseParameterWorker(/*inOuterAwaitContext*/ true);
            }
            function parseParameter() {
                return parseParameterWorker(/*inOuterAwaitContext*/ false);
            }
            function parseParameterWorker(inOuterAwaitContext) {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                // FormalParameter [Yield,Await]:
                //      BindingElement[?Yield,?Await]
                // Decorators are parsed in the outer [Await] context, the rest of the parameter is parsed in the function's [Await] context.
                var decorators = inOuterAwaitContext ? doInAwaitContext(parseDecorators) : parseDecorators();
                if (token() === 108 /* ThisKeyword */) {
                    var node_1 = factory.createParameterDeclaration(decorators, 
                    /*modifiers*/ undefined, 
                    /*dotDotDotToken*/ undefined, createIdentifier(/*isIdentifier*/ true), 
                    /*questionToken*/ undefined, parseTypeAnnotation(), 
                    /*initializer*/ undefined);
                    if (decorators) {
                        parseErrorAtRange(decorators[0], ts.Diagnostics.Decorators_may_not_be_applied_to_this_parameters);
                    }
                    return withJSDoc(finishNode(node_1, pos), hasJSDoc);
                }
                var savedTopLevel = topLevel;
                topLevel = false;
                var modifiers = parseModifiers();
                var node = withJSDoc(finishNode(factory.createParameterDeclaration(decorators, modifiers, parseOptionalToken(25 /* DotDotDotToken */), parseNameOfParameter(modifiers), parseOptionalToken(57 /* QuestionToken */), parseTypeAnnotation(), parseInitializer()), pos), hasJSDoc);
                topLevel = savedTopLevel;
                return node;
            }
            function parseReturnType(returnToken, isType) {
                if (shouldParseReturnType(returnToken, isType)) {
                    return parseTypeOrTypePredicate();
                }
            }
            function shouldParseReturnType(returnToken, isType) {
                if (returnToken === 38 /* EqualsGreaterThanToken */) {
                    parseExpected(returnToken);
                    return true;
                }
                else if (parseOptional(58 /* ColonToken */)) {
                    return true;
                }
                else if (isType && token() === 38 /* EqualsGreaterThanToken */) {
                    // This is easy to get backward, especially in type contexts, so parse the type anyway
                    parseErrorAtCurrentToken(ts.Diagnostics._0_expected, ts.tokenToString(58 /* ColonToken */));
                    nextToken();
                    return true;
                }
                return false;
            }
            function parseParametersWorker(flags) {
                // FormalParameters [Yield,Await]: (modified)
                //      [empty]
                //      FormalParameterList[?Yield,Await]
                //
                // FormalParameter[Yield,Await]: (modified)
                //      BindingElement[?Yield,Await]
                //
                // BindingElement [Yield,Await]: (modified)
                //      SingleNameBinding[?Yield,?Await]
                //      BindingPattern[?Yield,?Await]Initializer [In, ?Yield,?Await] opt
                //
                // SingleNameBinding [Yield,Await]:
                //      BindingIdentifier[?Yield,?Await]Initializer [In, ?Yield,?Await] opt
                var savedYieldContext = inYieldContext();
                var savedAwaitContext = inAwaitContext();
                setYieldContext(!!(flags & 1 /* Yield */));
                setAwaitContext(!!(flags & 2 /* Await */));
                var parameters = flags & 32 /* JSDoc */ ?
                    parseDelimitedList(17 /* JSDocParameters */, parseJSDocParameter) :
                    parseDelimitedList(16 /* Parameters */, savedAwaitContext ? parseParameterInOuterAwaitContext : parseParameter);
                setYieldContext(savedYieldContext);
                setAwaitContext(savedAwaitContext);
                return parameters;
            }
            function parseParameters(flags) {
                // FormalParameters [Yield,Await]: (modified)
                //      [empty]
                //      FormalParameterList[?Yield,Await]
                //
                // FormalParameter[Yield,Await]: (modified)
                //      BindingElement[?Yield,Await]
                //
                // BindingElement [Yield,Await]: (modified)
                //      SingleNameBinding[?Yield,?Await]
                //      BindingPattern[?Yield,?Await]Initializer [In, ?Yield,?Await] opt
                //
                // SingleNameBinding [Yield,Await]:
                //      BindingIdentifier[?Yield,?Await]Initializer [In, ?Yield,?Await] opt
                if (!parseExpected(20 /* OpenParenToken */)) {
                    return createMissingList();
                }
                var parameters = parseParametersWorker(flags);
                parseExpected(21 /* CloseParenToken */);
                return parameters;
            }
            function parseTypeMemberSemicolon() {
                // We allow type members to be separated by commas or (possibly ASI) semicolons.
                // First check if it was a comma.  If so, we're done with the member.
                if (parseOptional(27 /* CommaToken */)) {
                    return;
                }
                // Didn't have a comma.  We must have a (possible ASI) semicolon.
                parseSemicolon();
            }
            function parseSignatureMember(kind) {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                if (kind === 173 /* ConstructSignature */) {
                    parseExpected(103 /* NewKeyword */);
                }
                var typeParameters = parseTypeParameters();
                var parameters = parseParameters(4 /* Type */);
                var type = parseReturnType(58 /* ColonToken */, /*isType*/ true);
                parseTypeMemberSemicolon();
                var node = kind === 172 /* CallSignature */
                    ? factory.createCallSignature(typeParameters, parameters, type)
                    : factory.createConstructSignature(typeParameters, parameters, type);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function isIndexSignature() {
                return token() === 22 /* OpenBracketToken */ && lookAhead(isUnambiguouslyIndexSignature);
            }
            function isUnambiguouslyIndexSignature() {
                // The only allowed sequence is:
                //
                //   [id:
                //
                // However, for error recovery, we also check the following cases:
                //
                //   [...
                //   [id,
                //   [id?,
                //   [id?:
                //   [id?]
                //   [public id
                //   [private id
                //   [protected id
                //   []
                //
                nextToken();
                if (token() === 25 /* DotDotDotToken */ || token() === 23 /* CloseBracketToken */) {
                    return true;
                }
                if (ts.isModifierKind(token())) {
                    nextToken();
                    if (isIdentifier()) {
                        return true;
                    }
                }
                else if (!isIdentifier()) {
                    return false;
                }
                else {
                    // Skip the identifier
                    nextToken();
                }
                // A colon signifies a well formed indexer
                // A comma should be a badly formed indexer because comma expressions are not allowed
                // in computed properties.
                if (token() === 58 /* ColonToken */ || token() === 27 /* CommaToken */) {
                    return true;
                }
                // Question mark could be an indexer with an optional property,
                // or it could be a conditional expression in a computed property.
                if (token() !== 57 /* QuestionToken */) {
                    return false;
                }
                // If any of the following tokens are after the question mark, it cannot
                // be a conditional expression, so treat it as an indexer.
                nextToken();
                return token() === 58 /* ColonToken */ || token() === 27 /* CommaToken */ || token() === 23 /* CloseBracketToken */;
            }
            function parseIndexSignatureDeclaration(pos, hasJSDoc, decorators, modifiers) {
                var parameters = parseBracketedList(16 /* Parameters */, parseParameter, 22 /* OpenBracketToken */, 23 /* CloseBracketToken */);
                var type = parseTypeAnnotation();
                parseTypeMemberSemicolon();
                var node = factory.createIndexSignature(decorators, modifiers, parameters, type);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parsePropertyOrMethodSignature(pos, hasJSDoc, modifiers) {
                var name = parsePropertyName();
                var questionToken = parseOptionalToken(57 /* QuestionToken */);
                var node;
                if (token() === 20 /* OpenParenToken */ || token() === 29 /* LessThanToken */) {
                    // Method signatures don't exist in expression contexts.  So they have neither
                    // [Yield] nor [Await]
                    var typeParameters = parseTypeParameters();
                    var parameters = parseParameters(4 /* Type */);
                    var type = parseReturnType(58 /* ColonToken */, /*isType*/ true);
                    node = factory.createMethodSignature(modifiers, name, questionToken, typeParameters, parameters, type);
                }
                else {
                    var type = parseTypeAnnotation();
                    node = factory.createPropertySignature(modifiers, name, questionToken, type);
                    // Although type literal properties cannot not have initializers, we attempt
                    // to parse an initializer so we can report in the checker that an interface
                    // property or type literal property cannot have an initializer.
                    if (token() === 63 /* EqualsToken */)
                        node.initializer = parseInitializer();
                }
                parseTypeMemberSemicolon();
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function isTypeMemberStart() {
                // Return true if we have the start of a signature member
                if (token() === 20 /* OpenParenToken */ ||
                    token() === 29 /* LessThanToken */ ||
                    token() === 135 /* GetKeyword */ ||
                    token() === 147 /* SetKeyword */) {
                    return true;
                }
                var idToken = false;
                // Eat up all modifiers, but hold on to the last one in case it is actually an identifier
                while (ts.isModifierKind(token())) {
                    idToken = true;
                    nextToken();
                }
                // Index signatures and computed property names are type members
                if (token() === 22 /* OpenBracketToken */) {
                    return true;
                }
                // Try to get the first property-like token following all modifiers
                if (isLiteralPropertyName()) {
                    idToken = true;
                    nextToken();
                }
                // If we were able to get any potential identifier, check that it is
                // the start of a member declaration
                if (idToken) {
                    return token() === 20 /* OpenParenToken */ ||
                        token() === 29 /* LessThanToken */ ||
                        token() === 57 /* QuestionToken */ ||
                        token() === 58 /* ColonToken */ ||
                        token() === 27 /* CommaToken */ ||
                        canParseSemicolon();
                }
                return false;
            }
            function parseTypeMember() {
                if (token() === 20 /* OpenParenToken */ || token() === 29 /* LessThanToken */) {
                    return parseSignatureMember(172 /* CallSignature */);
                }
                if (token() === 103 /* NewKeyword */ && lookAhead(nextTokenIsOpenParenOrLessThan)) {
                    return parseSignatureMember(173 /* ConstructSignature */);
                }
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                var modifiers = parseModifiers();
                if (parseContextualModifier(135 /* GetKeyword */)) {
                    return parseAccessorDeclaration(pos, hasJSDoc, /*decorators*/ undefined, modifiers, 170 /* GetAccessor */);
                }
                if (parseContextualModifier(147 /* SetKeyword */)) {
                    return parseAccessorDeclaration(pos, hasJSDoc, /*decorators*/ undefined, modifiers, 171 /* SetAccessor */);
                }
                if (isIndexSignature()) {
                    return parseIndexSignatureDeclaration(pos, hasJSDoc, /*decorators*/ undefined, modifiers);
                }
                return parsePropertyOrMethodSignature(pos, hasJSDoc, modifiers);
            }
            function nextTokenIsOpenParenOrLessThan() {
                nextToken();
                return token() === 20 /* OpenParenToken */ || token() === 29 /* LessThanToken */;
            }
            function nextTokenIsDot() {
                return nextToken() === 24 /* DotToken */;
            }
            function nextTokenIsOpenParenOrLessThanOrDot() {
                switch (nextToken()) {
                    case 20 /* OpenParenToken */:
                    case 29 /* LessThanToken */:
                    case 24 /* DotToken */:
                        return true;
                }
                return false;
            }
            function parseTypeLiteral() {
                var pos = getNodePos();
                return finishNode(factory.createTypeLiteralNode(parseObjectTypeMembers()), pos);
            }
            function parseObjectTypeMembers() {
                var members;
                if (parseExpected(18 /* OpenBraceToken */)) {
                    members = parseList(4 /* TypeMembers */, parseTypeMember);
                    parseExpected(19 /* CloseBraceToken */);
                }
                else {
                    members = createMissingList();
                }
                return members;
            }
            function isStartOfMappedType() {
                nextToken();
                if (token() === 39 /* PlusToken */ || token() === 40 /* MinusToken */) {
                    return nextToken() === 143 /* ReadonlyKeyword */;
                }
                if (token() === 143 /* ReadonlyKeyword */) {
                    nextToken();
                }
                return token() === 22 /* OpenBracketToken */ && nextTokenIsIdentifier() && nextToken() === 101 /* InKeyword */;
            }
            function parseMappedTypeParameter() {
                var pos = getNodePos();
                var name = parseIdentifierName();
                parseExpected(101 /* InKeyword */);
                var type = parseType();
                return finishNode(factory.createTypeParameterDeclaration(name, type, /*defaultType*/ undefined), pos);
            }
            function parseMappedType() {
                var pos = getNodePos();
                parseExpected(18 /* OpenBraceToken */);
                var readonlyToken;
                if (token() === 143 /* ReadonlyKeyword */ || token() === 39 /* PlusToken */ || token() === 40 /* MinusToken */) {
                    readonlyToken = parseTokenNode();
                    if (readonlyToken.kind !== 143 /* ReadonlyKeyword */) {
                        parseExpected(143 /* ReadonlyKeyword */);
                    }
                }
                parseExpected(22 /* OpenBracketToken */);
                var typeParameter = parseMappedTypeParameter();
                var nameType = parseOptional(127 /* AsKeyword */) ? parseType() : undefined;
                parseExpected(23 /* CloseBracketToken */);
                var questionToken;
                if (token() === 57 /* QuestionToken */ || token() === 39 /* PlusToken */ || token() === 40 /* MinusToken */) {
                    questionToken = parseTokenNode();
                    if (questionToken.kind !== 57 /* QuestionToken */) {
                        parseExpected(57 /* QuestionToken */);
                    }
                }
                var type = parseTypeAnnotation();
                parseSemicolon();
                parseExpected(19 /* CloseBraceToken */);
                return finishNode(factory.createMappedTypeNode(readonlyToken, typeParameter, nameType, questionToken, type), pos);
            }
            function parseTupleElementType() {
                var pos = getNodePos();
                if (parseOptional(25 /* DotDotDotToken */)) {
                    return finishNode(factory.createRestTypeNode(parseType()), pos);
                }
                var type = parseType();
                if (ts.isJSDocNullableType(type) && type.pos === type.type.pos) {
                    var node = factory.createOptionalTypeNode(type.type);
                    ts.setTextRange(node, type);
                    node.flags = type.flags;
                    return node;
                }
                return type;
            }
            function isNextTokenColonOrQuestionColon() {
                return nextToken() === 58 /* ColonToken */ || (token() === 57 /* QuestionToken */ && nextToken() === 58 /* ColonToken */);
            }
            function isTupleElementName() {
                if (token() === 25 /* DotDotDotToken */) {
                    return ts.tokenIsIdentifierOrKeyword(nextToken()) && isNextTokenColonOrQuestionColon();
                }
                return ts.tokenIsIdentifierOrKeyword(token()) && isNextTokenColonOrQuestionColon();
            }
            function parseTupleElementNameOrTupleElementType() {
                if (lookAhead(isTupleElementName)) {
                    var pos = getNodePos();
                    var hasJSDoc = hasPrecedingJSDocComment();
                    var dotDotDotToken = parseOptionalToken(25 /* DotDotDotToken */);
                    var name = parseIdentifierName();
                    var questionToken = parseOptionalToken(57 /* QuestionToken */);
                    parseExpected(58 /* ColonToken */);
                    var type = parseTupleElementType();
                    var node = factory.createNamedTupleMember(dotDotDotToken, name, questionToken, type);
                    return withJSDoc(finishNode(node, pos), hasJSDoc);
                }
                return parseTupleElementType();
            }
            function parseTupleType() {
                var pos = getNodePos();
                return finishNode(factory.createTupleTypeNode(parseBracketedList(21 /* TupleElementTypes */, parseTupleElementNameOrTupleElementType, 22 /* OpenBracketToken */, 23 /* CloseBracketToken */)), pos);
            }
            function parseParenthesizedType() {
                var pos = getNodePos();
                parseExpected(20 /* OpenParenToken */);
                var type = parseType();
                parseExpected(21 /* CloseParenToken */);
                return finishNode(factory.createParenthesizedType(type), pos);
            }
            function parseModifiersForConstructorType() {
                var modifiers;
                if (token() === 126 /* AbstractKeyword */) {
                    var pos = getNodePos();
                    nextToken();
                    var modifier = finishNode(factory.createToken(126 /* AbstractKeyword */), pos);
                    modifiers = createNodeArray([modifier], pos);
                }
                return modifiers;
            }
            function parseFunctionOrConstructorType() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                var modifiers = parseModifiersForConstructorType();
                var isConstructorType = parseOptional(103 /* NewKeyword */);
                var typeParameters = parseTypeParameters();
                var parameters = parseParameters(4 /* Type */);
                var type = parseReturnType(38 /* EqualsGreaterThanToken */, /*isType*/ false);
                var node = isConstructorType
                    ? factory.createConstructorTypeNode(modifiers, typeParameters, parameters, type)
                    : factory.createFunctionTypeNode(typeParameters, parameters, type);
                if (!isConstructorType)
                    node.modifiers = modifiers;
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseKeywordAndNoDot() {
                var node = parseTokenNode();
                return token() === 24 /* DotToken */ ? undefined : node;
            }
            function parseLiteralTypeNode(negative) {
                var pos = getNodePos();
                if (negative) {
                    nextToken();
                }
                var expression = token() === 110 /* TrueKeyword */ || token() === 95 /* FalseKeyword */ || token() === 104 /* NullKeyword */ ?
                    parseTokenNode() :
                    parseLiteralLikeNode(token());
                if (negative) {
                    expression = finishNode(factory.createPrefixUnaryExpression(40 /* MinusToken */, expression), pos);
                }
                return finishNode(factory.createLiteralTypeNode(expression), pos);
            }
            function isStartOfTypeOfImportType() {
                nextToken();
                return token() === 100 /* ImportKeyword */;
            }
            function parseImportType() {
                sourceFlags |= 1048576 /* PossiblyContainsDynamicImport */;
                var pos = getNodePos();
                var isTypeOf = parseOptional(112 /* TypeOfKeyword */);
                parseExpected(100 /* ImportKeyword */);
                parseExpected(20 /* OpenParenToken */);
                var type = parseType();
                parseExpected(21 /* CloseParenToken */);
                var qualifier = parseOptional(24 /* DotToken */) ? parseEntityNameOfTypeReference() : undefined;
                var typeArguments = parseTypeArgumentsOfTypeReference();
                return finishNode(factory.createImportTypeNode(type, qualifier, typeArguments, isTypeOf), pos);
            }
            function nextTokenIsNumericOrBigIntLiteral() {
                nextToken();
                return token() === 8 /* NumericLiteral */ || token() === 9 /* BigIntLiteral */;
            }
            function parseNonArrayType() {
                switch (token()) {
                    case 129 /* AnyKeyword */:
                    case 153 /* UnknownKeyword */:
                    case 148 /* StringKeyword */:
                    case 145 /* NumberKeyword */:
                    case 156 /* BigIntKeyword */:
                    case 149 /* SymbolKeyword */:
                    case 132 /* BooleanKeyword */:
                    case 151 /* UndefinedKeyword */:
                    case 142 /* NeverKeyword */:
                    case 146 /* ObjectKeyword */:
                        // If these are followed by a dot, then parse these out as a dotted type reference instead.
                        return tryParse(parseKeywordAndNoDot) || parseTypeReference();
                    case 66 /* AsteriskEqualsToken */:
                        // If there is '*=', treat it as * followed by postfix =
                        scanner.reScanAsteriskEqualsToken();
                    // falls through
                    case 41 /* AsteriskToken */:
                        return parseJSDocAllType();
                    case 60 /* QuestionQuestionToken */:
                        // If there is '??', treat it as prefix-'?' in JSDoc type.
                        scanner.reScanQuestionToken();
                    // falls through
                    case 57 /* QuestionToken */:
                        return parseJSDocUnknownOrNullableType();
                    case 98 /* FunctionKeyword */:
                        return parseJSDocFunctionType();
                    case 53 /* ExclamationToken */:
                        return parseJSDocNonNullableType();
                    case 14 /* NoSubstitutionTemplateLiteral */:
                    case 10 /* StringLiteral */:
                    case 8 /* NumericLiteral */:
                    case 9 /* BigIntLiteral */:
                    case 110 /* TrueKeyword */:
                    case 95 /* FalseKeyword */:
                    case 104 /* NullKeyword */:
                        return parseLiteralTypeNode();
                    case 40 /* MinusToken */:
                        return lookAhead(nextTokenIsNumericOrBigIntLiteral) ? parseLiteralTypeNode(/*negative*/ true) : parseTypeReference();
                    case 114 /* VoidKeyword */:
                        return parseTokenNode();
                    case 108 /* ThisKeyword */: {
                        var thisKeyword = parseThisTypeNode();
                        if (token() === 138 /* IsKeyword */ && !scanner.hasPrecedingLineBreak()) {
                            return parseThisTypePredicate(thisKeyword);
                        }
                        else {
                            return thisKeyword;
                        }
                    }
                    case 112 /* TypeOfKeyword */:
                        return lookAhead(isStartOfTypeOfImportType) ? parseImportType() : parseTypeQuery();
                    case 18 /* OpenBraceToken */:
                        return lookAhead(isStartOfMappedType) ? parseMappedType() : parseTypeLiteral();
                    case 22 /* OpenBracketToken */:
                        return parseTupleType();
                    case 20 /* OpenParenToken */:
                        return parseParenthesizedType();
                    case 100 /* ImportKeyword */:
                        return parseImportType();
                    case 128 /* AssertsKeyword */:
                        return lookAhead(nextTokenIsIdentifierOrKeywordOnSameLine) ? parseAssertsTypePredicate() : parseTypeReference();
                    case 15 /* TemplateHead */:
                        return parseTemplateType();
                    default:
                        return parseTypeReference();
                }
            }
            function isStartOfType(inStartOfParameter) {
                switch (token()) {
                    case 129 /* AnyKeyword */:
                    case 153 /* UnknownKeyword */:
                    case 148 /* StringKeyword */:
                    case 145 /* NumberKeyword */:
                    case 156 /* BigIntKeyword */:
                    case 132 /* BooleanKeyword */:
                    case 143 /* ReadonlyKeyword */:
                    case 149 /* SymbolKeyword */:
                    case 152 /* UniqueKeyword */:
                    case 114 /* VoidKeyword */:
                    case 151 /* UndefinedKeyword */:
                    case 104 /* NullKeyword */:
                    case 108 /* ThisKeyword */:
                    case 112 /* TypeOfKeyword */:
                    case 142 /* NeverKeyword */:
                    case 18 /* OpenBraceToken */:
                    case 22 /* OpenBracketToken */:
                    case 29 /* LessThanToken */:
                    case 51 /* BarToken */:
                    case 50 /* AmpersandToken */:
                    case 103 /* NewKeyword */:
                    case 10 /* StringLiteral */:
                    case 8 /* NumericLiteral */:
                    case 9 /* BigIntLiteral */:
                    case 110 /* TrueKeyword */:
                    case 95 /* FalseKeyword */:
                    case 146 /* ObjectKeyword */:
                    case 41 /* AsteriskToken */:
                    case 57 /* QuestionToken */:
                    case 53 /* ExclamationToken */:
                    case 25 /* DotDotDotToken */:
                    case 136 /* InferKeyword */:
                    case 100 /* ImportKeyword */:
                    case 128 /* AssertsKeyword */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                    case 15 /* TemplateHead */:
                        return true;
                    case 98 /* FunctionKeyword */:
                        return !inStartOfParameter;
                    case 40 /* MinusToken */:
                        return !inStartOfParameter && lookAhead(nextTokenIsNumericOrBigIntLiteral);
                    case 20 /* OpenParenToken */:
                        // Only consider '(' the start of a type if followed by ')', '...', an identifier, a modifier,
                        // or something that starts a type. We don't want to consider things like '(1)' a type.
                        return !inStartOfParameter && lookAhead(isStartOfParenthesizedOrFunctionType);
                    default:
                        return isIdentifier();
                }
            }
            function isStartOfParenthesizedOrFunctionType() {
                nextToken();
                return token() === 21 /* CloseParenToken */ || isStartOfParameter(/*isJSDocParameter*/ false) || isStartOfType();
            }
            function parsePostfixTypeOrHigher() {
                var pos = getNodePos();
                var type = parseNonArrayType();
                while (!scanner.hasPrecedingLineBreak()) {
                    switch (token()) {
                        case 53 /* ExclamationToken */:
                            nextToken();
                            type = finishNode(factory.createJSDocNonNullableType(type), pos);
                            break;
                        case 57 /* QuestionToken */:
                            // If next token is start of a type we have a conditional type
                            if (lookAhead(nextTokenIsStartOfType)) {
                                return type;
                            }
                            nextToken();
                            type = finishNode(factory.createJSDocNullableType(type), pos);
                            break;
                        case 22 /* OpenBracketToken */:
                            parseExpected(22 /* OpenBracketToken */);
                            if (isStartOfType()) {
                                var indexType = parseType();
                                parseExpected(23 /* CloseBracketToken */);
                                type = finishNode(factory.createIndexedAccessTypeNode(type, indexType), pos);
                            }
                            else {
                                parseExpected(23 /* CloseBracketToken */);
                                type = finishNode(factory.createArrayTypeNode(type), pos);
                            }
                            break;
                        default:
                            return type;
                    }
                }
                return type;
            }
            function parseTypeOperator(operator) {
                var pos = getNodePos();
                parseExpected(operator);
                return finishNode(factory.createTypeOperatorNode(operator, parseTypeOperatorOrHigher()), pos);
            }
            function parseTypeParameterOfInferType() {
                var pos = getNodePos();
                return finishNode(factory.createTypeParameterDeclaration(parseIdentifier(), 
                /*constraint*/ undefined, 
                /*defaultType*/ undefined), pos);
            }
            function parseInferType() {
                var pos = getNodePos();
                parseExpected(136 /* InferKeyword */);
                return finishNode(factory.createInferTypeNode(parseTypeParameterOfInferType()), pos);
            }
            function parseTypeOperatorOrHigher() {
                var operator = token();
                switch (operator) {
                    case 139 /* KeyOfKeyword */:
                    case 152 /* UniqueKeyword */:
                    case 143 /* ReadonlyKeyword */:
                        return parseTypeOperator(operator);
                    case 136 /* InferKeyword */:
                        return parseInferType();
                }
                return parsePostfixTypeOrHigher();
            }
            function parseFunctionOrConstructorTypeToError(isInUnionType) {
                // the function type and constructor type shorthand notation
                // are not allowed directly in unions and intersections, but we'll
                // try to parse them gracefully and issue a helpful message.
                if (isStartOfFunctionTypeOrConstructorType()) {
                    var type = parseFunctionOrConstructorType();
                    var diagnostic = void 0;
                    if (ts.isFunctionTypeNode(type)) {
                        diagnostic = isInUnionType
                            ? ts.Diagnostics.Function_type_notation_must_be_parenthesized_when_used_in_a_union_type
                            : ts.Diagnostics.Function_type_notation_must_be_parenthesized_when_used_in_an_intersection_type;
                    }
                    else {
                        diagnostic = isInUnionType
                            ? ts.Diagnostics.Constructor_type_notation_must_be_parenthesized_when_used_in_a_union_type
                            : ts.Diagnostics.Constructor_type_notation_must_be_parenthesized_when_used_in_an_intersection_type;
                    }
                    parseErrorAtRange(type, diagnostic);
                    return type;
                }
                return undefined;
            }
            function parseUnionOrIntersectionType(operator, parseConstituentType, createTypeNode) {
                var pos = getNodePos();
                var isUnionType = operator === 51 /* BarToken */;
                var hasLeadingOperator = parseOptional(operator);
                var type = hasLeadingOperator && parseFunctionOrConstructorTypeToError(isUnionType)
                    || parseConstituentType();
                if (token() === operator || hasLeadingOperator) {
                    var types = [type];
                    while (parseOptional(operator)) {
                        types.push(parseFunctionOrConstructorTypeToError(isUnionType) || parseConstituentType());
                    }
                    type = finishNode(createTypeNode(createNodeArray(types, pos)), pos);
                }
                return type;
            }
            function parseIntersectionTypeOrHigher() {
                return parseUnionOrIntersectionType(50 /* AmpersandToken */, parseTypeOperatorOrHigher, factory.createIntersectionTypeNode);
            }
            function parseUnionTypeOrHigher() {
                return parseUnionOrIntersectionType(51 /* BarToken */, parseIntersectionTypeOrHigher, factory.createUnionTypeNode);
            }
            function nextTokenIsNewKeyword() {
                nextToken();
                return token() === 103 /* NewKeyword */;
            }
            function isStartOfFunctionTypeOrConstructorType() {
                if (token() === 29 /* LessThanToken */) {
                    return true;
                }
                if (token() === 20 /* OpenParenToken */ && lookAhead(isUnambiguouslyStartOfFunctionType)) {
                    return true;
                }
                return token() === 103 /* NewKeyword */ ||
                    token() === 126 /* AbstractKeyword */ && lookAhead(nextTokenIsNewKeyword);
            }
            function skipParameterStart() {
                if (ts.isModifierKind(token())) {
                    // Skip modifiers
                    parseModifiers();
                }
                if (isIdentifier() || token() === 108 /* ThisKeyword */) {
                    nextToken();
                    return true;
                }
                if (token() === 22 /* OpenBracketToken */ || token() === 18 /* OpenBraceToken */) {
                    // Return true if we can parse an array or object binding pattern with no errors
                    var previousErrorCount = parseDiagnostics.length;
                    parseIdentifierOrPattern();
                    return previousErrorCount === parseDiagnostics.length;
                }
                return false;
            }
            function isUnambiguouslyStartOfFunctionType() {
                nextToken();
                if (token() === 21 /* CloseParenToken */ || token() === 25 /* DotDotDotToken */) {
                    // ( )
                    // ( ...
                    return true;
                }
                if (skipParameterStart()) {
                    // We successfully skipped modifiers (if any) and an identifier or binding pattern,
                    // now see if we have something that indicates a parameter declaration
                    if (token() === 58 /* ColonToken */ || token() === 27 /* CommaToken */ ||
                        token() === 57 /* QuestionToken */ || token() === 63 /* EqualsToken */) {
                        // ( xxx :
                        // ( xxx ,
                        // ( xxx ?
                        // ( xxx =
                        return true;
                    }
                    if (token() === 21 /* CloseParenToken */) {
                        nextToken();
                        if (token() === 38 /* EqualsGreaterThanToken */) {
                            // ( xxx ) =>
                            return true;
                        }
                    }
                }
                return false;
            }
            function parseTypeOrTypePredicate() {
                var pos = getNodePos();
                var typePredicateVariable = isIdentifier() && tryParse(parseTypePredicatePrefix);
                var type = parseType();
                if (typePredicateVariable) {
                    return finishNode(factory.createTypePredicateNode(/*assertsModifier*/ undefined, typePredicateVariable, type), pos);
                }
                else {
                    return type;
                }
            }
            function parseTypePredicatePrefix() {
                var id = parseIdentifier();
                if (token() === 138 /* IsKeyword */ && !scanner.hasPrecedingLineBreak()) {
                    nextToken();
                    return id;
                }
            }
            function parseAssertsTypePredicate() {
                var pos = getNodePos();
                var assertsModifier = parseExpectedToken(128 /* AssertsKeyword */);
                var parameterName = token() === 108 /* ThisKeyword */ ? parseThisTypeNode() : parseIdentifier();
                var type = parseOptional(138 /* IsKeyword */) ? parseType() : undefined;
                return finishNode(factory.createTypePredicateNode(assertsModifier, parameterName, type), pos);
            }
            function parseType() {
                // The rules about 'yield' only apply to actual code/expression contexts.  They don't
                // apply to 'type' contexts.  So we disable these parameters here before moving on.
                return doOutsideOfContext(40960 /* TypeExcludesFlags */, parseTypeWorker);
            }
            function parseTypeWorker(noConditionalTypes) {
                if (isStartOfFunctionTypeOrConstructorType()) {
                    return parseFunctionOrConstructorType();
                }
                var pos = getNodePos();
                var type = parseUnionTypeOrHigher();
                if (!noConditionalTypes && !scanner.hasPrecedingLineBreak() && parseOptional(94 /* ExtendsKeyword */)) {
                    // The type following 'extends' is not permitted to be another conditional type
                    var extendsType = parseTypeWorker(/*noConditionalTypes*/ true);
                    parseExpected(57 /* QuestionToken */);
                    var trueType = parseTypeWorker();
                    parseExpected(58 /* ColonToken */);
                    var falseType = parseTypeWorker();
                    return finishNode(factory.createConditionalTypeNode(type, extendsType, trueType, falseType), pos);
                }
                return type;
            }
            function parseTypeAnnotation() {
                return parseOptional(58 /* ColonToken */) ? parseType() : undefined;
            }
            // EXPRESSIONS
            function isStartOfLeftHandSideExpression() {
                switch (token()) {
                    case 108 /* ThisKeyword */:
                    case 106 /* SuperKeyword */:
                    case 104 /* NullKeyword */:
                    case 110 /* TrueKeyword */:
                    case 95 /* FalseKeyword */:
                    case 8 /* NumericLiteral */:
                    case 9 /* BigIntLiteral */:
                    case 10 /* StringLiteral */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                    case 15 /* TemplateHead */:
                    case 20 /* OpenParenToken */:
                    case 22 /* OpenBracketToken */:
                    case 18 /* OpenBraceToken */:
                    case 98 /* FunctionKeyword */:
                    case 84 /* ClassKeyword */:
                    case 103 /* NewKeyword */:
                    case 43 /* SlashToken */:
                    case 68 /* SlashEqualsToken */:
                    case 79 /* Identifier */:
                        return true;
                    case 100 /* ImportKeyword */:
                        return lookAhead(nextTokenIsOpenParenOrLessThanOrDot);
                    default:
                        return isIdentifier();
                }
            }
            function isStartOfExpression() {
                if (isStartOfLeftHandSideExpression()) {
                    return true;
                }
                switch (token()) {
                    case 39 /* PlusToken */:
                    case 40 /* MinusToken */:
                    case 54 /* TildeToken */:
                    case 53 /* ExclamationToken */:
                    case 89 /* DeleteKeyword */:
                    case 112 /* TypeOfKeyword */:
                    case 114 /* VoidKeyword */:
                    case 45 /* PlusPlusToken */:
                    case 46 /* MinusMinusToken */:
                    case 29 /* LessThanToken */:
                    case 131 /* AwaitKeyword */:
                    case 125 /* YieldKeyword */:
                    case 80 /* PrivateIdentifier */:
                        // Yield/await always starts an expression.  Either it is an identifier (in which case
                        // it is definitely an expression).  Or it's a keyword (either because we're in
                        // a generator or async function, or in strict mode (or both)) and it started a yield or await expression.
                        return true;
                    default:
                        // Error tolerance.  If we see the start of some binary operator, we consider
                        // that the start of an expression.  That way we'll parse out a missing identifier,
                        // give a good message about an identifier being missing, and then consume the
                        // rest of the binary expression.
                        if (isBinaryOperator()) {
                            return true;
                        }
                        return isIdentifier();
                }
            }
            function isStartOfExpressionStatement() {
                // As per the grammar, none of '{' or 'function' or 'class' can start an expression statement.
                return token() !== 18 /* OpenBraceToken */ &&
                    token() !== 98 /* FunctionKeyword */ &&
                    token() !== 84 /* ClassKeyword */ &&
                    token() !== 59 /* AtToken */ &&
                    isStartOfExpression();
            }
            function parseExpression() {
                // Expression[in]:
                //      AssignmentExpression[in]
                //      Expression[in] , AssignmentExpression[in]
                // clear the decorator context when parsing Expression, as it should be unambiguous when parsing a decorator
                var saveDecoratorContext = inDecoratorContext();
                if (saveDecoratorContext) {
                    setDecoratorContext(/*val*/ false);
                }
                var pos = getNodePos();
                var expr = parseAssignmentExpressionOrHigher();
                var operatorToken;
                while ((operatorToken = parseOptionalToken(27 /* CommaToken */))) {
                    expr = makeBinaryExpression(expr, operatorToken, parseAssignmentExpressionOrHigher(), pos);
                }
                if (saveDecoratorContext) {
                    setDecoratorContext(/*val*/ true);
                }
                return expr;
            }
            function parseInitializer() {
                return parseOptional(63 /* EqualsToken */) ? parseAssignmentExpressionOrHigher() : undefined;
            }
            function parseAssignmentExpressionOrHigher() {
                //  AssignmentExpression[in,yield]:
                //      1) ConditionalExpression[?in,?yield]
                //      2) LeftHandSideExpression = AssignmentExpression[?in,?yield]
                //      3) LeftHandSideExpression AssignmentOperator AssignmentExpression[?in,?yield]
                //      4) ArrowFunctionExpression[?in,?yield]
                //      5) AsyncArrowFunctionExpression[in,yield,await]
                //      6) [+Yield] YieldExpression[?In]
                //
                // Note: for ease of implementation we treat productions '2' and '3' as the same thing.
                // (i.e. they're both BinaryExpressions with an assignment operator in it).
                // First, do the simple check if we have a YieldExpression (production '6').
                if (isYieldExpression()) {
                    return parseYieldExpression();
                }
                // Then, check if we have an arrow function (production '4' and '5') that starts with a parenthesized
                // parameter list or is an async arrow function.
                // AsyncArrowFunctionExpression:
                //      1) async[no LineTerminator here]AsyncArrowBindingIdentifier[?Yield][no LineTerminator here]=>AsyncConciseBody[?In]
                //      2) CoverCallExpressionAndAsyncArrowHead[?Yield, ?Await][no LineTerminator here]=>AsyncConciseBody[?In]
                // Production (1) of AsyncArrowFunctionExpression is parsed in "tryParseAsyncSimpleArrowFunctionExpression".
                // And production (2) is parsed in "tryParseParenthesizedArrowFunctionExpression".
                //
                // If we do successfully parse arrow-function, we must *not* recurse for productions 1, 2 or 3. An ArrowFunction is
                // not a LeftHandSideExpression, nor does it start a ConditionalExpression.  So we are done
                // with AssignmentExpression if we see one.
                var arrowExpression = tryParseParenthesizedArrowFunctionExpression() || tryParseAsyncSimpleArrowFunctionExpression();
                if (arrowExpression) {
                    return arrowExpression;
                }
                // Now try to see if we're in production '1', '2' or '3'.  A conditional expression can
                // start with a LogicalOrExpression, while the assignment productions can only start with
                // LeftHandSideExpressions.
                //
                // So, first, we try to just parse out a BinaryExpression.  If we get something that is a
                // LeftHandSide or higher, then we can try to parse out the assignment expression part.
                // Otherwise, we try to parse out the conditional expression bit.  We want to allow any
                // binary expression here, so we pass in the 'lowest' precedence here so that it matches
                // and consumes anything.
                var pos = getNodePos();
                var expr = parseBinaryExpressionOrHigher(0 /* Lowest */);
                // To avoid a look-ahead, we did not handle the case of an arrow function with a single un-parenthesized
                // parameter ('x => ...') above. We handle it here by checking if the parsed expression was a single
                // identifier and the current token is an arrow.
                if (expr.kind === 79 /* Identifier */ && token() === 38 /* EqualsGreaterThanToken */) {
                    return parseSimpleArrowFunctionExpression(pos, expr, /*asyncModifier*/ undefined);
                }
                // Now see if we might be in cases '2' or '3'.
                // If the expression was a LHS expression, and we have an assignment operator, then
                // we're in '2' or '3'. Consume the assignment and return.
                //
                // Note: we call reScanGreaterToken so that we get an appropriately merged token
                // for cases like `> > =` becoming `>>=`
                if (ts.isLeftHandSideExpression(expr) && ts.isAssignmentOperator(reScanGreaterToken())) {
                    return makeBinaryExpression(expr, parseTokenNode(), parseAssignmentExpressionOrHigher(), pos);
                }
                // It wasn't an assignment or a lambda.  This is a conditional expression:
                return parseConditionalExpressionRest(expr, pos);
            }
            function isYieldExpression() {
                if (token() === 125 /* YieldKeyword */) {
                    // If we have a 'yield' keyword, and this is a context where yield expressions are
                    // allowed, then definitely parse out a yield expression.
                    if (inYieldContext()) {
                        return true;
                    }
                    // We're in a context where 'yield expr' is not allowed.  However, if we can
                    // definitely tell that the user was trying to parse a 'yield expr' and not
                    // just a normal expr that start with a 'yield' identifier, then parse out
                    // a 'yield expr'.  We can then report an error later that they are only
                    // allowed in generator expressions.
                    //
                    // for example, if we see 'yield(foo)', then we'll have to treat that as an
                    // invocation expression of something called 'yield'.  However, if we have
                    // 'yield foo' then that is not legal as a normal expression, so we can
                    // definitely recognize this as a yield expression.
                    //
                    // for now we just check if the next token is an identifier.  More heuristics
                    // can be added here later as necessary.  We just need to make sure that we
                    // don't accidentally consume something legal.
                    return lookAhead(nextTokenIsIdentifierOrKeywordOrLiteralOnSameLine);
                }
                return false;
            }
            function nextTokenIsIdentifierOnSameLine() {
                nextToken();
                return !scanner.hasPrecedingLineBreak() && isIdentifier();
            }
            function parseYieldExpression() {
                var pos = getNodePos();
                // YieldExpression[In] :
                //      yield
                //      yield [no LineTerminator here] [Lexical goal InputElementRegExp]AssignmentExpression[?In, Yield]
                //      yield [no LineTerminator here] * [Lexical goal InputElementRegExp]AssignmentExpression[?In, Yield]
                nextToken();
                if (!scanner.hasPrecedingLineBreak() &&
                    (token() === 41 /* AsteriskToken */ || isStartOfExpression())) {
                    return finishNode(factory.createYieldExpression(parseOptionalToken(41 /* AsteriskToken */), parseAssignmentExpressionOrHigher()), pos);
                }
                else {
                    // if the next token is not on the same line as yield.  or we don't have an '*' or
                    // the start of an expression, then this is just a simple "yield" expression.
                    return finishNode(factory.createYieldExpression(/*asteriskToken*/ undefined, /*expression*/ undefined), pos);
                }
            }
            function parseSimpleArrowFunctionExpression(pos, identifier, asyncModifier) {
                ts.Debug.assert(token() === 38 /* EqualsGreaterThanToken */, "parseSimpleArrowFunctionExpression should only have been called if we had a =>");
                var parameter = factory.createParameterDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*dotDotDotToken*/ undefined, identifier, 
                /*questionToken*/ undefined, 
                /*type*/ undefined, 
                /*initializer*/ undefined);
                finishNode(parameter, identifier.pos);
                var parameters = createNodeArray([parameter], parameter.pos, parameter.end);
                var equalsGreaterThanToken = parseExpectedToken(38 /* EqualsGreaterThanToken */);
                var body = parseArrowFunctionExpressionBody(/*isAsync*/ !!asyncModifier);
                var node = factory.createArrowFunction(asyncModifier, /*typeParameters*/ undefined, parameters, /*type*/ undefined, equalsGreaterThanToken, body);
                return addJSDocComment(finishNode(node, pos));
            }
            function tryParseParenthesizedArrowFunctionExpression() {
                var triState = isParenthesizedArrowFunctionExpression();
                if (triState === 0 /* False */) {
                    // It's definitely not a parenthesized arrow function expression.
                    return undefined;
                }
                // If we definitely have an arrow function, then we can just parse one, not requiring a
                // following => or { token. Otherwise, we *might* have an arrow function.  Try to parse
                // it out, but don't allow any ambiguity, and return 'undefined' if this could be an
                // expression instead.
                return triState === 1 /* True */ ?
                    parseParenthesizedArrowFunctionExpression(/*allowAmbiguity*/ true) :
                    tryParse(parsePossibleParenthesizedArrowFunctionExpression);
            }
            //  True        -> We definitely expect a parenthesized arrow function here.
            //  False       -> There *cannot* be a parenthesized arrow function here.
            //  Unknown     -> There *might* be a parenthesized arrow function here.
            //                 Speculatively look ahead to be sure, and rollback if not.
            function isParenthesizedArrowFunctionExpression() {
                if (token() === 20 /* OpenParenToken */ || token() === 29 /* LessThanToken */ || token() === 130 /* AsyncKeyword */) {
                    return lookAhead(isParenthesizedArrowFunctionExpressionWorker);
                }
                if (token() === 38 /* EqualsGreaterThanToken */) {
                    // ERROR RECOVERY TWEAK:
                    // If we see a standalone => try to parse it as an arrow function expression as that's
                    // likely what the user intended to write.
                    return 1 /* True */;
                }
                // Definitely not a parenthesized arrow function.
                return 0 /* False */;
            }
            function isParenthesizedArrowFunctionExpressionWorker() {
                if (token() === 130 /* AsyncKeyword */) {
                    nextToken();
                    if (scanner.hasPrecedingLineBreak()) {
                        return 0 /* False */;
                    }
                    if (token() !== 20 /* OpenParenToken */ && token() !== 29 /* LessThanToken */) {
                        return 0 /* False */;
                    }
                }
                var first = token();
                var second = nextToken();
                if (first === 20 /* OpenParenToken */) {
                    if (second === 21 /* CloseParenToken */) {
                        // Simple cases: "() =>", "(): ", and "() {".
                        // This is an arrow function with no parameters.
                        // The last one is not actually an arrow function,
                        // but this is probably what the user intended.
                        var third = nextToken();
                        switch (third) {
                            case 38 /* EqualsGreaterThanToken */:
                            case 58 /* ColonToken */:
                            case 18 /* OpenBraceToken */:
                                return 1 /* True */;
                            default:
                                return 0 /* False */;
                        }
                    }
                    // If encounter "([" or "({", this could be the start of a binding pattern.
                    // Examples:
                    //      ([ x ]) => { }
                    //      ({ x }) => { }
                    //      ([ x ])
                    //      ({ x })
                    if (second === 22 /* OpenBracketToken */ || second === 18 /* OpenBraceToken */) {
                        return 2 /* Unknown */;
                    }
                    // Simple case: "(..."
                    // This is an arrow function with a rest parameter.
                    if (second === 25 /* DotDotDotToken */) {
                        return 1 /* True */;
                    }
                    // Check for "(xxx yyy", where xxx is a modifier and yyy is an identifier. This
                    // isn't actually allowed, but we want to treat it as a lambda so we can provide
                    // a good error message.
                    if (ts.isModifierKind(second) && second !== 130 /* AsyncKeyword */ && lookAhead(nextTokenIsIdentifier)) {
                        return 1 /* True */;
                    }
                    // If we had "(" followed by something that's not an identifier,
                    // then this definitely doesn't look like a lambda.  "this" is not
                    // valid, but we want to parse it and then give a semantic error.
                    if (!isIdentifier() && second !== 108 /* ThisKeyword */) {
                        return 0 /* False */;
                    }
                    switch (nextToken()) {
                        case 58 /* ColonToken */:
                            // If we have something like "(a:", then we must have a
                            // type-annotated parameter in an arrow function expression.
                            return 1 /* True */;
                        case 57 /* QuestionToken */:
                            nextToken();
                            // If we have "(a?:" or "(a?," or "(a?=" or "(a?)" then it is definitely a lambda.
                            if (token() === 58 /* ColonToken */ || token() === 27 /* CommaToken */ || token() === 63 /* EqualsToken */ || token() === 21 /* CloseParenToken */) {
                                return 1 /* True */;
                            }
                            // Otherwise it is definitely not a lambda.
                            return 0 /* False */;
                        case 27 /* CommaToken */:
                        case 63 /* EqualsToken */:
                        case 21 /* CloseParenToken */:
                            // If we have "(a," or "(a=" or "(a)" this *could* be an arrow function
                            return 2 /* Unknown */;
                    }
                    // It is definitely not an arrow function
                    return 0 /* False */;
                }
                else {
                    ts.Debug.assert(first === 29 /* LessThanToken */);
                    // If we have "<" not followed by an identifier,
                    // then this definitely is not an arrow function.
                    if (!isIdentifier()) {
                        return 0 /* False */;
                    }
                    // JSX overrides
                    if (languageVariant === 1 /* JSX */) {
                        var isArrowFunctionInJsx = lookAhead(function () {
                            var third = nextToken();
                            if (third === 94 /* ExtendsKeyword */) {
                                var fourth = nextToken();
                                switch (fourth) {
                                    case 63 /* EqualsToken */:
                                    case 31 /* GreaterThanToken */:
                                        return false;
                                    default:
                                        return true;
                                }
                            }
                            else if (third === 27 /* CommaToken */) {
                                return true;
                            }
                            return false;
                        });
                        if (isArrowFunctionInJsx) {
                            return 1 /* True */;
                        }
                        return 0 /* False */;
                    }
                    // This *could* be a parenthesized arrow function.
                    return 2 /* Unknown */;
                }
            }
            function parsePossibleParenthesizedArrowFunctionExpression() {
                var tokenPos = scanner.getTokenPos();
                if (notParenthesizedArrow === null || notParenthesizedArrow === void 0 ? void 0 : notParenthesizedArrow.has(tokenPos)) {
                    return undefined;
                }
                var result = parseParenthesizedArrowFunctionExpression(/*allowAmbiguity*/ false);
                if (!result) {
                    (notParenthesizedArrow || (notParenthesizedArrow = new ts.Set())).add(tokenPos);
                }
                return result;
            }
            function tryParseAsyncSimpleArrowFunctionExpression() {
                // We do a check here so that we won't be doing unnecessarily call to "lookAhead"
                if (token() === 130 /* AsyncKeyword */) {
                    if (lookAhead(isUnParenthesizedAsyncArrowFunctionWorker) === 1 /* True */) {
                        var pos = getNodePos();
                        var asyncModifier = parseModifiersForArrowFunction();
                        var expr = parseBinaryExpressionOrHigher(0 /* Lowest */);
                        return parseSimpleArrowFunctionExpression(pos, expr, asyncModifier);
                    }
                }
                return undefined;
            }
            function isUnParenthesizedAsyncArrowFunctionWorker() {
                // AsyncArrowFunctionExpression:
                //      1) async[no LineTerminator here]AsyncArrowBindingIdentifier[?Yield][no LineTerminator here]=>AsyncConciseBody[?In]
                //      2) CoverCallExpressionAndAsyncArrowHead[?Yield, ?Await][no LineTerminator here]=>AsyncConciseBody[?In]
                if (token() === 130 /* AsyncKeyword */) {
                    nextToken();
                    // If the "async" is followed by "=>" token then it is not a beginning of an async arrow-function
                    // but instead a simple arrow-function which will be parsed inside "parseAssignmentExpressionOrHigher"
                    if (scanner.hasPrecedingLineBreak() || token() === 38 /* EqualsGreaterThanToken */) {
                        return 0 /* False */;
                    }
                    // Check for un-parenthesized AsyncArrowFunction
                    var expr = parseBinaryExpressionOrHigher(0 /* Lowest */);
                    if (!scanner.hasPrecedingLineBreak() && expr.kind === 79 /* Identifier */ && token() === 38 /* EqualsGreaterThanToken */) {
                        return 1 /* True */;
                    }
                }
                return 0 /* False */;
            }
            function parseParenthesizedArrowFunctionExpression(allowAmbiguity) {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                var modifiers = parseModifiersForArrowFunction();
                var isAsync = ts.some(modifiers, ts.isAsyncModifier) ? 2 /* Await */ : 0 /* None */;
                // Arrow functions are never generators.
                //
                // If we're speculatively parsing a signature for a parenthesized arrow function, then
                // we have to have a complete parameter list.  Otherwise we might see something like
                // a => (b => c)
                // And think that "(b =>" was actually a parenthesized arrow function with a missing
                // close paren.
                var typeParameters = parseTypeParameters();
                var parameters;
                if (!parseExpected(20 /* OpenParenToken */)) {
                    if (!allowAmbiguity) {
                        return undefined;
                    }
                    parameters = createMissingList();
                }
                else {
                    parameters = parseParametersWorker(isAsync);
                    if (!parseExpected(21 /* CloseParenToken */) && !allowAmbiguity) {
                        return undefined;
                    }
                }
                var type = parseReturnType(58 /* ColonToken */, /*isType*/ false);
                if (type && !allowAmbiguity && typeHasArrowFunctionBlockingParseError(type)) {
                    return undefined;
                }
                // Parsing a signature isn't enough.
                // Parenthesized arrow signatures often look like other valid expressions.
                // For instance:
                //  - "(x = 10)" is an assignment expression parsed as a signature with a default parameter value.
                //  - "(x,y)" is a comma expression parsed as a signature with two parameters.
                //  - "a ? (b): c" will have "(b):" parsed as a signature with a return type annotation.
                //  - "a ? (b): function() {}" will too, since function() is a valid JSDoc function type.
                //
                // So we need just a bit of lookahead to ensure that it can only be a signature.
                var hasJSDocFunctionType = type && ts.isJSDocFunctionType(type);
                if (!allowAmbiguity && token() !== 38 /* EqualsGreaterThanToken */ && (hasJSDocFunctionType || token() !== 18 /* OpenBraceToken */)) {
                    // Returning undefined here will cause our caller to rewind to where we started from.
                    return undefined;
                }
                // If we have an arrow, then try to parse the body. Even if not, try to parse if we
                // have an opening brace, just in case we're in an error state.
                var lastToken = token();
                var equalsGreaterThanToken = parseExpectedToken(38 /* EqualsGreaterThanToken */);
                var body = (lastToken === 38 /* EqualsGreaterThanToken */ || lastToken === 18 /* OpenBraceToken */)
                    ? parseArrowFunctionExpressionBody(ts.some(modifiers, ts.isAsyncModifier))
                    : parseIdentifier();
                var node = factory.createArrowFunction(modifiers, typeParameters, parameters, type, equalsGreaterThanToken, body);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseArrowFunctionExpressionBody(isAsync) {
                if (token() === 18 /* OpenBraceToken */) {
                    return parseFunctionBlock(isAsync ? 2 /* Await */ : 0 /* None */);
                }
                if (token() !== 26 /* SemicolonToken */ &&
                    token() !== 98 /* FunctionKeyword */ &&
                    token() !== 84 /* ClassKeyword */ &&
                    isStartOfStatement() &&
                    !isStartOfExpressionStatement()) {
                    // Check if we got a plain statement (i.e. no expression-statements, no function/class expressions/declarations)
                    //
                    // Here we try to recover from a potential error situation in the case where the
                    // user meant to supply a block. For example, if the user wrote:
                    //
                    //  a =>
                    //      let v = 0;
                    //  }
                    //
                    // they may be missing an open brace.  Check to see if that's the case so we can
                    // try to recover better.  If we don't do this, then the next close curly we see may end
                    // up preemptively closing the containing construct.
                    //
                    // Note: even when 'IgnoreMissingOpenBrace' is passed, parseBody will still error.
                    return parseFunctionBlock(16 /* IgnoreMissingOpenBrace */ | (isAsync ? 2 /* Await */ : 0 /* None */));
                }
                var savedTopLevel = topLevel;
                topLevel = false;
                var node = isAsync
                    ? doInAwaitContext(parseAssignmentExpressionOrHigher)
                    : doOutsideOfAwaitContext(parseAssignmentExpressionOrHigher);
                topLevel = savedTopLevel;
                return node;
            }
            function parseConditionalExpressionRest(leftOperand, pos) {
                // Note: we are passed in an expression which was produced from parseBinaryExpressionOrHigher.
                var questionToken = parseOptionalToken(57 /* QuestionToken */);
                if (!questionToken) {
                    return leftOperand;
                }
                // Note: we explicitly 'allowIn' in the whenTrue part of the condition expression, and
                // we do not that for the 'whenFalse' part.
                var colonToken;
                return finishNode(factory.createConditionalExpression(leftOperand, questionToken, doOutsideOfContext(disallowInAndDecoratorContext, parseAssignmentExpressionOrHigher), colonToken = parseExpectedToken(58 /* ColonToken */), ts.nodeIsPresent(colonToken)
                    ? parseAssignmentExpressionOrHigher()
                    : createMissingNode(79 /* Identifier */, /*reportAtCurrentPosition*/ false, ts.Diagnostics._0_expected, ts.tokenToString(58 /* ColonToken */))), pos);
            }
            function parseBinaryExpressionOrHigher(precedence) {
                var pos = getNodePos();
                var leftOperand = parseUnaryExpressionOrHigher();
                return parseBinaryExpressionRest(precedence, leftOperand, pos);
            }
            function isInOrOfKeyword(t) {
                return t === 101 /* InKeyword */ || t === 158 /* OfKeyword */;
            }
            function parseBinaryExpressionRest(precedence, leftOperand, pos) {
                while (true) {
                    // We either have a binary operator here, or we're finished.  We call
                    // reScanGreaterToken so that we merge token sequences like > and = into >=
                    reScanGreaterToken();
                    var newPrecedence = ts.getBinaryOperatorPrecedence(token());
                    // Check the precedence to see if we should "take" this operator
                    // - For left associative operator (all operator but **), consume the operator,
                    //   recursively call the function below, and parse binaryExpression as a rightOperand
                    //   of the caller if the new precedence of the operator is greater then or equal to the current precedence.
                    //   For example:
                    //      a - b - c;
                    //            ^token; leftOperand = b. Return b to the caller as a rightOperand
                    //      a * b - c
                    //            ^token; leftOperand = b. Return b to the caller as a rightOperand
                    //      a - b * c;
                    //            ^token; leftOperand = b. Return b * c to the caller as a rightOperand
                    // - For right associative operator (**), consume the operator, recursively call the function
                    //   and parse binaryExpression as a rightOperand of the caller if the new precedence of
                    //   the operator is strictly grater than the current precedence
                    //   For example:
                    //      a ** b ** c;
                    //             ^^token; leftOperand = b. Return b ** c to the caller as a rightOperand
                    //      a - b ** c;
                    //            ^^token; leftOperand = b. Return b ** c to the caller as a rightOperand
                    //      a ** b - c
                    //             ^token; leftOperand = b. Return b to the caller as a rightOperand
                    var consumeCurrentOperator = token() === 42 /* AsteriskAsteriskToken */ ?
                        newPrecedence >= precedence :
                        newPrecedence > precedence;
                    if (!consumeCurrentOperator) {
                        break;
                    }
                    if (token() === 101 /* InKeyword */ && inDisallowInContext()) {
                        break;
                    }
                    if (token() === 127 /* AsKeyword */) {
                        // Make sure we *do* perform ASI for constructs like this:
                        //    var x = foo
                        //    as (Bar)
                        // This should be parsed as an initialized variable, followed
                        // by a function call to 'as' with the argument 'Bar'
                        if (scanner.hasPrecedingLineBreak()) {
                            break;
                        }
                        else {
                            nextToken();
                            leftOperand = makeAsExpression(leftOperand, parseType());
                        }
                    }
                    else {
                        leftOperand = makeBinaryExpression(leftOperand, parseTokenNode(), parseBinaryExpressionOrHigher(newPrecedence), pos);
                    }
                }
                return leftOperand;
            }
            function isBinaryOperator() {
                if (inDisallowInContext() && token() === 101 /* InKeyword */) {
                    return false;
                }
                return ts.getBinaryOperatorPrecedence(token()) > 0;
            }
            function makeBinaryExpression(left, operatorToken, right, pos) {
                return finishNode(factory.createBinaryExpression(left, operatorToken, right), pos);
            }
            function makeAsExpression(left, right) {
                return finishNode(factory.createAsExpression(left, right), left.pos);
            }
            function parsePrefixUnaryExpression() {
                var pos = getNodePos();
                return finishNode(factory.createPrefixUnaryExpression(token(), nextTokenAnd(parseSimpleUnaryExpression)), pos);
            }
            function parseDeleteExpression() {
                var pos = getNodePos();
                return finishNode(factory.createDeleteExpression(nextTokenAnd(parseSimpleUnaryExpression)), pos);
            }
            function parseTypeOfExpression() {
                var pos = getNodePos();
                return finishNode(factory.createTypeOfExpression(nextTokenAnd(parseSimpleUnaryExpression)), pos);
            }
            function parseVoidExpression() {
                var pos = getNodePos();
                return finishNode(factory.createVoidExpression(nextTokenAnd(parseSimpleUnaryExpression)), pos);
            }
            function isAwaitExpression() {
                if (token() === 131 /* AwaitKeyword */) {
                    if (inAwaitContext()) {
                        return true;
                    }
                    // here we are using similar heuristics as 'isYieldExpression'
                    return lookAhead(nextTokenIsIdentifierOrKeywordOrLiteralOnSameLine);
                }
                return false;
            }
            function parseAwaitExpression() {
                var pos = getNodePos();
                return finishNode(factory.createAwaitExpression(nextTokenAnd(parseSimpleUnaryExpression)), pos);
            }
            /**
             * Parse ES7 exponential expression and await expression
             *
             * ES7 ExponentiationExpression:
             *      1) UnaryExpression[?Yield]
             *      2) UpdateExpression[?Yield] ** ExponentiationExpression[?Yield]
             *
             */
            function parseUnaryExpressionOrHigher() {
                /**
                 * ES7 UpdateExpression:
                 *      1) LeftHandSideExpression[?Yield]
                 *      2) LeftHandSideExpression[?Yield][no LineTerminator here]++
                 *      3) LeftHandSideExpression[?Yield][no LineTerminator here]--
                 *      4) ++UnaryExpression[?Yield]
                 *      5) --UnaryExpression[?Yield]
                 */
                if (isUpdateExpression()) {
                    var pos = getNodePos();
                    var updateExpression = parseUpdateExpression();
                    return token() === 42 /* AsteriskAsteriskToken */ ?
                        parseBinaryExpressionRest(ts.getBinaryOperatorPrecedence(token()), updateExpression, pos) :
                        updateExpression;
                }
                /**
                 * ES7 UnaryExpression:
                 *      1) UpdateExpression[?yield]
                 *      2) delete UpdateExpression[?yield]
                 *      3) void UpdateExpression[?yield]
                 *      4) typeof UpdateExpression[?yield]
                 *      5) + UpdateExpression[?yield]
                 *      6) - UpdateExpression[?yield]
                 *      7) ~ UpdateExpression[?yield]
                 *      8) ! UpdateExpression[?yield]
                 */
                var unaryOperator = token();
                var simpleUnaryExpression = parseSimpleUnaryExpression();
                if (token() === 42 /* AsteriskAsteriskToken */) {
                    var pos = ts.skipTrivia(sourceText, simpleUnaryExpression.pos);
                    var end = simpleUnaryExpression.end;
                    if (simpleUnaryExpression.kind === 209 /* TypeAssertionExpression */) {
                        parseErrorAt(pos, end, ts.Diagnostics.A_type_assertion_expression_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses);
                    }
                    else {
                        parseErrorAt(pos, end, ts.Diagnostics.An_unary_expression_with_the_0_operator_is_not_allowed_in_the_left_hand_side_of_an_exponentiation_expression_Consider_enclosing_the_expression_in_parentheses, ts.tokenToString(unaryOperator));
                    }
                }
                return simpleUnaryExpression;
            }
            /**
             * Parse ES7 simple-unary expression or higher:
             *
             * ES7 UnaryExpression:
             *      1) UpdateExpression[?yield]
             *      2) delete UnaryExpression[?yield]
             *      3) void UnaryExpression[?yield]
             *      4) typeof UnaryExpression[?yield]
             *      5) + UnaryExpression[?yield]
             *      6) - UnaryExpression[?yield]
             *      7) ~ UnaryExpression[?yield]
             *      8) ! UnaryExpression[?yield]
             *      9) [+Await] await UnaryExpression[?yield]
             */
            function parseSimpleUnaryExpression() {
                switch (token()) {
                    case 39 /* PlusToken */:
                    case 40 /* MinusToken */:
                    case 54 /* TildeToken */:
                    case 53 /* ExclamationToken */:
                        return parsePrefixUnaryExpression();
                    case 89 /* DeleteKeyword */:
                        return parseDeleteExpression();
                    case 112 /* TypeOfKeyword */:
                        return parseTypeOfExpression();
                    case 114 /* VoidKeyword */:
                        return parseVoidExpression();
                    case 29 /* LessThanToken */:
                        // This is modified UnaryExpression grammar in TypeScript
                        //  UnaryExpression (modified):
                        //      < type > UnaryExpression
                        return parseTypeAssertion();
                    case 131 /* AwaitKeyword */:
                        if (isAwaitExpression()) {
                            return parseAwaitExpression();
                        }
                    // falls through
                    default:
                        return parseUpdateExpression();
                }
            }
            /**
             * Check if the current token can possibly be an ES7 increment expression.
             *
             * ES7 UpdateExpression:
             *      LeftHandSideExpression[?Yield]
             *      LeftHandSideExpression[?Yield][no LineTerminator here]++
             *      LeftHandSideExpression[?Yield][no LineTerminator here]--
             *      ++LeftHandSideExpression[?Yield]
             *      --LeftHandSideExpression[?Yield]
             */
            function isUpdateExpression() {
                // This function is called inside parseUnaryExpression to decide
                // whether to call parseSimpleUnaryExpression or call parseUpdateExpression directly
                switch (token()) {
                    case 39 /* PlusToken */:
                    case 40 /* MinusToken */:
                    case 54 /* TildeToken */:
                    case 53 /* ExclamationToken */:
                    case 89 /* DeleteKeyword */:
                    case 112 /* TypeOfKeyword */:
                    case 114 /* VoidKeyword */:
                    case 131 /* AwaitKeyword */:
                        return false;
                    case 29 /* LessThanToken */:
                        // If we are not in JSX context, we are parsing TypeAssertion which is an UnaryExpression
                        if (languageVariant !== 1 /* JSX */) {
                            return false;
                        }
                    // We are in JSX context and the token is part of JSXElement.
                    // falls through
                    default:
                        return true;
                }
            }
            /**
             * Parse ES7 UpdateExpression. UpdateExpression is used instead of ES6's PostFixExpression.
             *
             * ES7 UpdateExpression[yield]:
             *      1) LeftHandSideExpression[?yield]
             *      2) LeftHandSideExpression[?yield] [[no LineTerminator here]]++
             *      3) LeftHandSideExpression[?yield] [[no LineTerminator here]]--
             *      4) ++LeftHandSideExpression[?yield]
             *      5) --LeftHandSideExpression[?yield]
             * In TypeScript (2), (3) are parsed as PostfixUnaryExpression. (4), (5) are parsed as PrefixUnaryExpression
             */
            function parseUpdateExpression() {
                if (token() === 45 /* PlusPlusToken */ || token() === 46 /* MinusMinusToken */) {
                    var pos = getNodePos();
                    return finishNode(factory.createPrefixUnaryExpression(token(), nextTokenAnd(parseLeftHandSideExpressionOrHigher)), pos);
                }
                else if (languageVariant === 1 /* JSX */ && token() === 29 /* LessThanToken */ && lookAhead(nextTokenIsIdentifierOrKeywordOrGreaterThan)) {
                    // JSXElement is part of primaryExpression
                    return parseJsxElementOrSelfClosingElementOrFragment(/*inExpressionContext*/ true);
                }
                var expression = parseLeftHandSideExpressionOrHigher();
                ts.Debug.assert(ts.isLeftHandSideExpression(expression));
                if ((token() === 45 /* PlusPlusToken */ || token() === 46 /* MinusMinusToken */) && !scanner.hasPrecedingLineBreak()) {
                    var operator = token();
                    nextToken();
                    return finishNode(factory.createPostfixUnaryExpression(expression, operator), expression.pos);
                }
                return expression;
            }
            function parseLeftHandSideExpressionOrHigher() {
                // Original Ecma:
                // LeftHandSideExpression: See 11.2
                //      NewExpression
                //      CallExpression
                //
                // Our simplification:
                //
                // LeftHandSideExpression: See 11.2
                //      MemberExpression
                //      CallExpression
                //
                // See comment in parseMemberExpressionOrHigher on how we replaced NewExpression with
                // MemberExpression to make our lives easier.
                //
                // to best understand the below code, it's important to see how CallExpression expands
                // out into its own productions:
                //
                // CallExpression:
                //      MemberExpression Arguments
                //      CallExpression Arguments
                //      CallExpression[Expression]
                //      CallExpression.IdentifierName
                //      import (AssignmentExpression)
                //      super Arguments
                //      super.IdentifierName
                //
                // Because of the recursion in these calls, we need to bottom out first. There are three
                // bottom out states we can run into: 1) We see 'super' which must start either of
                // the last two CallExpression productions. 2) We see 'import' which must start import call.
                // 3)we have a MemberExpression which either completes the LeftHandSideExpression,
                // or starts the beginning of the first four CallExpression productions.
                var pos = getNodePos();
                var expression;
                if (token() === 100 /* ImportKeyword */) {
                    if (lookAhead(nextTokenIsOpenParenOrLessThan)) {
                        // We don't want to eagerly consume all import keyword as import call expression so we look ahead to find "("
                        // For example:
                        //      var foo3 = require("subfolder
                        //      import * as foo1 from "module-from-node
                        // We want this import to be a statement rather than import call expression
                        sourceFlags |= 1048576 /* PossiblyContainsDynamicImport */;
                        expression = parseTokenNode();
                    }
                    else if (lookAhead(nextTokenIsDot)) {
                        // This is an 'import.*' metaproperty (i.e. 'import.meta')
                        nextToken(); // advance past the 'import'
                        nextToken(); // advance past the dot
                        expression = finishNode(factory.createMetaProperty(100 /* ImportKeyword */, parseIdentifierName()), pos);
                        sourceFlags |= 2097152 /* PossiblyContainsImportMeta */;
                    }
                    else {
                        expression = parseMemberExpressionOrHigher();
                    }
                }
                else {
                    expression = token() === 106 /* SuperKeyword */ ? parseSuperExpression() : parseMemberExpressionOrHigher();
                }
                // Now, we *may* be complete.  However, we might have consumed the start of a
                // CallExpression or OptionalExpression.  As such, we need to consume the rest
                // of it here to be complete.
                return parseCallExpressionRest(pos, expression);
            }
            function parseMemberExpressionOrHigher() {
                // Note: to make our lives simpler, we decompose the NewExpression productions and
                // place ObjectCreationExpression and FunctionExpression into PrimaryExpression.
                // like so:
                //
                //   PrimaryExpression : See 11.1
                //      this
                //      Identifier
                //      Literal
                //      ArrayLiteral
                //      ObjectLiteral
                //      (Expression)
                //      FunctionExpression
                //      new MemberExpression Arguments?
                //
                //   MemberExpression : See 11.2
                //      PrimaryExpression
                //      MemberExpression[Expression]
                //      MemberExpression.IdentifierName
                //
                //   CallExpression : See 11.2
                //      MemberExpression
                //      CallExpression Arguments
                //      CallExpression[Expression]
                //      CallExpression.IdentifierName
                //
                // Technically this is ambiguous.  i.e. CallExpression defines:
                //
                //   CallExpression:
                //      CallExpression Arguments
                //
                // If you see: "new Foo()"
                //
                // Then that could be treated as a single ObjectCreationExpression, or it could be
                // treated as the invocation of "new Foo".  We disambiguate that in code (to match
                // the original grammar) by making sure that if we see an ObjectCreationExpression
                // we always consume arguments if they are there. So we treat "new Foo()" as an
                // object creation only, and not at all as an invocation.  Another way to think
                // about this is that for every "new" that we see, we will consume an argument list if
                // it is there as part of the *associated* object creation node.  Any additional
                // argument lists we see, will become invocation expressions.
                //
                // Because there are no other places in the grammar now that refer to FunctionExpression
                // or ObjectCreationExpression, it is safe to push down into the PrimaryExpression
                // production.
                //
                // Because CallExpression and MemberExpression are left recursive, we need to bottom out
                // of the recursion immediately.  So we parse out a primary expression to start with.
                var pos = getNodePos();
                var expression = parsePrimaryExpression();
                return parseMemberExpressionRest(pos, expression, /*allowOptionalChain*/ true);
            }
            function parseSuperExpression() {
                var pos = getNodePos();
                var expression = parseTokenNode();
                if (token() === 29 /* LessThanToken */) {
                    var startPos = getNodePos();
                    var typeArguments = tryParse(parseTypeArgumentsInExpression);
                    if (typeArguments !== undefined) {
                        parseErrorAt(startPos, getNodePos(), ts.Diagnostics.super_may_not_use_type_arguments);
                    }
                }
                if (token() === 20 /* OpenParenToken */ || token() === 24 /* DotToken */ || token() === 22 /* OpenBracketToken */) {
                    return expression;
                }
                // If we have seen "super" it must be followed by '(' or '.'.
                // If it wasn't then just try to parse out a '.' and report an error.
                parseExpectedToken(24 /* DotToken */, ts.Diagnostics.super_must_be_followed_by_an_argument_list_or_member_access);
                // private names will never work with `super` (`super.#foo`), but that's a semantic error, not syntactic
                return finishNode(factory.createPropertyAccessExpression(expression, parseRightSideOfDot(/*allowIdentifierNames*/ true, /*allowPrivateIdentifiers*/ true)), pos);
            }
            function parseJsxElementOrSelfClosingElementOrFragment(inExpressionContext, topInvalidNodePosition, openingTag) {
                var pos = getNodePos();
                var opening = parseJsxOpeningOrSelfClosingElementOrOpeningFragment(inExpressionContext);
                var result;
                if (opening.kind === 278 /* JsxOpeningElement */) {
                    var children = parseJsxChildren(opening);
                    var closingElement = void 0;
                    var lastChild = children[children.length - 1];
                    if ((lastChild === null || lastChild === void 0 ? void 0 : lastChild.kind) === 276 /* JsxElement */
                        && !tagNamesAreEquivalent(lastChild.openingElement.tagName, lastChild.closingElement.tagName)
                        && tagNamesAreEquivalent(opening.tagName, lastChild.closingElement.tagName)) {
                        // when an unclosed JsxOpeningElement incorrectly parses its parent's JsxClosingElement,
                        // restructure (<div>(...<span></div>)) --> (<div>(...<span></span>)</div>)
                        // (no need to error; the parent will error)
                        var end = lastChild.openingElement.end; // newly-created children and closing are both zero-width end/end
                        var newLast = finishNode(factory.createJsxElement(lastChild.openingElement, createNodeArray([], end, end), finishNode(factory.createJsxClosingElement(finishNode(factory.createIdentifier(""), end, end)), end, end)), lastChild.openingElement.pos, end);
                        children = createNodeArray(__spreadArray(__spreadArray([], children.slice(0, children.length - 1), true), [newLast], false), children.pos, end);
                        closingElement = lastChild.closingElement;
                    }
                    else {
                        closingElement = parseJsxClosingElement(opening, inExpressionContext);
                        if (!tagNamesAreEquivalent(opening.tagName, closingElement.tagName)) {
                            if (openingTag && ts.isJsxOpeningElement(openingTag) && tagNamesAreEquivalent(closingElement.tagName, openingTag.tagName)) {
                                // opening incorrectly matched with its parent's closing -- put error on opening
                                parseErrorAtRange(opening.tagName, ts.Diagnostics.JSX_element_0_has_no_corresponding_closing_tag, ts.getTextOfNodeFromSourceText(sourceText, opening.tagName));
                            }
                            else {
                                // other opening/closing mismatches -- put error on closing
                                parseErrorAtRange(closingElement.tagName, ts.Diagnostics.Expected_corresponding_JSX_closing_tag_for_0, ts.getTextOfNodeFromSourceText(sourceText, opening.tagName));
                            }
                        }
                    }
                    result = finishNode(factory.createJsxElement(opening, children, closingElement), pos);
                }
                else if (opening.kind === 281 /* JsxOpeningFragment */) {
                    result = finishNode(factory.createJsxFragment(opening, parseJsxChildren(opening), parseJsxClosingFragment(inExpressionContext)), pos);
                }
                else {
                    ts.Debug.assert(opening.kind === 277 /* JsxSelfClosingElement */);
                    // Nothing else to do for self-closing elements
                    result = opening;
                }
                // If the user writes the invalid code '<div></div><div></div>' in an expression context (i.e. not wrapped in
                // an enclosing tag), we'll naively try to parse   ^ this as a 'less than' operator and the remainder of the tag
                // as garbage, which will cause the formatter to badly mangle the JSX. Perform a speculative parse of a JSX
                // element if we see a < token so that we can wrap it in a synthetic binary expression so the formatter
                // does less damage and we can report a better error.
                // Since JSX elements are invalid < operands anyway, this lookahead parse will only occur in error scenarios
                // of one sort or another.
                if (inExpressionContext && token() === 29 /* LessThanToken */) {
                    var topBadPos_1 = typeof topInvalidNodePosition === "undefined" ? result.pos : topInvalidNodePosition;
                    var invalidElement = tryParse(function () { return parseJsxElementOrSelfClosingElementOrFragment(/*inExpressionContext*/ true, topBadPos_1); });
                    if (invalidElement) {
                        var operatorToken = createMissingNode(27 /* CommaToken */, /*reportAtCurrentPosition*/ false);
                        ts.setTextRangePosWidth(operatorToken, invalidElement.pos, 0);
                        parseErrorAt(ts.skipTrivia(sourceText, topBadPos_1), invalidElement.end, ts.Diagnostics.JSX_expressions_must_have_one_parent_element);
                        return finishNode(factory.createBinaryExpression(result, operatorToken, invalidElement), pos);
                    }
                }
                return result;
            }
            function parseJsxText() {
                var pos = getNodePos();
                var node = factory.createJsxText(scanner.getTokenValue(), currentToken === 12 /* JsxTextAllWhiteSpaces */);
                currentToken = scanner.scanJsxToken();
                return finishNode(node, pos);
            }
            function parseJsxChild(openingTag, token) {
                switch (token) {
                    case 1 /* EndOfFileToken */:
                        // If we hit EOF, issue the error at the tag that lacks the closing element
                        // rather than at the end of the file (which is useless)
                        if (ts.isJsxOpeningFragment(openingTag)) {
                            parseErrorAtRange(openingTag, ts.Diagnostics.JSX_fragment_has_no_corresponding_closing_tag);
                        }
                        else {
                            // We want the error span to cover only 'Foo.Bar' in < Foo.Bar >
                            // or to cover only 'Foo' in < Foo >
                            var tag = openingTag.tagName;
                            var start = ts.skipTrivia(sourceText, tag.pos);
                            parseErrorAt(start, tag.end, ts.Diagnostics.JSX_element_0_has_no_corresponding_closing_tag, ts.getTextOfNodeFromSourceText(sourceText, openingTag.tagName));
                        }
                        return undefined;
                    case 30 /* LessThanSlashToken */:
                    case 7 /* ConflictMarkerTrivia */:
                        return undefined;
                    case 11 /* JsxText */:
                    case 12 /* JsxTextAllWhiteSpaces */:
                        return parseJsxText();
                    case 18 /* OpenBraceToken */:
                        return parseJsxExpression(/*inExpressionContext*/ false);
                    case 29 /* LessThanToken */:
                        return parseJsxElementOrSelfClosingElementOrFragment(/*inExpressionContext*/ false, /*topInvalidNodePosition*/ undefined, openingTag);
                    default:
                        return ts.Debug.assertNever(token);
                }
            }
            function parseJsxChildren(openingTag) {
                var list = [];
                var listPos = getNodePos();
                var saveParsingContext = parsingContext;
                parsingContext |= 1 << 14 /* JsxChildren */;
                while (true) {
                    var child = parseJsxChild(openingTag, currentToken = scanner.reScanJsxToken());
                    if (!child)
                        break;
                    list.push(child);
                    if (ts.isJsxOpeningElement(openingTag)
                        && (child === null || child === void 0 ? void 0 : child.kind) === 276 /* JsxElement */
                        && !tagNamesAreEquivalent(child.openingElement.tagName, child.closingElement.tagName)
                        && tagNamesAreEquivalent(openingTag.tagName, child.closingElement.tagName)) {
                        // stop after parsing a mismatched child like <div>...(<span></div>) in order to reattach the </div> higher
                        break;
                    }
                }
                parsingContext = saveParsingContext;
                return createNodeArray(list, listPos);
            }
            function parseJsxAttributes() {
                var pos = getNodePos();
                return finishNode(factory.createJsxAttributes(parseList(13 /* JsxAttributes */, parseJsxAttribute)), pos);
            }
            function parseJsxOpeningOrSelfClosingElementOrOpeningFragment(inExpressionContext) {
                var pos = getNodePos();
                parseExpected(29 /* LessThanToken */);
                if (token() === 31 /* GreaterThanToken */) {
                    // See below for explanation of scanJsxText
                    scanJsxText();
                    return finishNode(factory.createJsxOpeningFragment(), pos);
                }
                var tagName = parseJsxElementName();
                var typeArguments = (contextFlags & 131072 /* JavaScriptFile */) === 0 ? tryParseTypeArguments() : undefined;
                var attributes = parseJsxAttributes();
                var node;
                if (token() === 31 /* GreaterThanToken */) {
                    // Closing tag, so scan the immediately-following text with the JSX scanning instead
                    // of regular scanning to avoid treating illegal characters (e.g. '#') as immediate
                    // scanning errors
                    scanJsxText();
                    node = factory.createJsxOpeningElement(tagName, typeArguments, attributes);
                }
                else {
                    parseExpected(43 /* SlashToken */);
                    if (parseExpected(31 /* GreaterThanToken */, /*diagnostic*/ undefined, /*shouldAdvance*/ false)) {
                        // manually advance the scanner in order to look for jsx text inside jsx
                        if (inExpressionContext) {
                            nextToken();
                        }
                        else {
                            scanJsxText();
                        }
                    }
                    node = factory.createJsxSelfClosingElement(tagName, typeArguments, attributes);
                }
                return finishNode(node, pos);
            }
            function parseJsxElementName() {
                var pos = getNodePos();
                scanJsxIdentifier();
                // JsxElement can have name in the form of
                //      propertyAccessExpression
                //      primaryExpression in the form of an identifier and "this" keyword
                // We can't just simply use parseLeftHandSideExpressionOrHigher because then we will start consider class,function etc as a keyword
                // We only want to consider "this" as a primaryExpression
                var expression = token() === 108 /* ThisKeyword */ ?
                    parseTokenNode() : parseIdentifierName();
                while (parseOptional(24 /* DotToken */)) {
                    expression = finishNode(factory.createPropertyAccessExpression(expression, parseRightSideOfDot(/*allowIdentifierNames*/ true, /*allowPrivateIdentifiers*/ false)), pos);
                }
                return expression;
            }
            function parseJsxExpression(inExpressionContext) {
                var pos = getNodePos();
                if (!parseExpected(18 /* OpenBraceToken */)) {
                    return undefined;
                }
                var dotDotDotToken;
                var expression;
                if (token() !== 19 /* CloseBraceToken */) {
                    dotDotDotToken = parseOptionalToken(25 /* DotDotDotToken */);
                    // Only an AssignmentExpression is valid here per the JSX spec,
                    // but we can unambiguously parse a comma sequence and provide
                    // a better error message in grammar checking.
                    expression = parseExpression();
                }
                if (inExpressionContext) {
                    parseExpected(19 /* CloseBraceToken */);
                }
                else {
                    if (parseExpected(19 /* CloseBraceToken */, /*message*/ undefined, /*shouldAdvance*/ false)) {
                        scanJsxText();
                    }
                }
                return finishNode(factory.createJsxExpression(dotDotDotToken, expression), pos);
            }
            function parseJsxAttribute() {
                if (token() === 18 /* OpenBraceToken */) {
                    return parseJsxSpreadAttribute();
                }
                scanJsxIdentifier();
                var pos = getNodePos();
                return finishNode(factory.createJsxAttribute(parseIdentifierName(), token() !== 63 /* EqualsToken */ ? undefined :
                    scanJsxAttributeValue() === 10 /* StringLiteral */ ? parseLiteralNode() :
                        parseJsxExpression(/*inExpressionContext*/ true)), pos);
            }
            function parseJsxSpreadAttribute() {
                var pos = getNodePos();
                parseExpected(18 /* OpenBraceToken */);
                parseExpected(25 /* DotDotDotToken */);
                var expression = parseExpression();
                parseExpected(19 /* CloseBraceToken */);
                return finishNode(factory.createJsxSpreadAttribute(expression), pos);
            }
            function parseJsxClosingElement(open, inExpressionContext) {
                var pos = getNodePos();
                parseExpected(30 /* LessThanSlashToken */);
                var tagName = parseJsxElementName();
                if (parseExpected(31 /* GreaterThanToken */, /*diagnostic*/ undefined, /*shouldAdvance*/ false)) {
                    // manually advance the scanner in order to look for jsx text inside jsx
                    if (inExpressionContext || !tagNamesAreEquivalent(open.tagName, tagName)) {
                        nextToken();
                    }
                    else {
                        scanJsxText();
                    }
                }
                return finishNode(factory.createJsxClosingElement(tagName), pos);
            }
            function parseJsxClosingFragment(inExpressionContext) {
                var pos = getNodePos();
                parseExpected(30 /* LessThanSlashToken */);
                if (ts.tokenIsIdentifierOrKeyword(token())) {
                    parseErrorAtRange(parseJsxElementName(), ts.Diagnostics.Expected_corresponding_closing_tag_for_JSX_fragment);
                }
                if (parseExpected(31 /* GreaterThanToken */, /*diagnostic*/ undefined, /*shouldAdvance*/ false)) {
                    // manually advance the scanner in order to look for jsx text inside jsx
                    if (inExpressionContext) {
                        nextToken();
                    }
                    else {
                        scanJsxText();
                    }
                }
                return finishNode(factory.createJsxJsxClosingFragment(), pos);
            }
            function parseTypeAssertion() {
                var pos = getNodePos();
                parseExpected(29 /* LessThanToken */);
                var type = parseType();
                parseExpected(31 /* GreaterThanToken */);
                var expression = parseSimpleUnaryExpression();
                return finishNode(factory.createTypeAssertion(type, expression), pos);
            }
            function nextTokenIsIdentifierOrKeywordOrOpenBracketOrTemplate() {
                nextToken();
                return ts.tokenIsIdentifierOrKeyword(token())
                    || token() === 22 /* OpenBracketToken */
                    || isTemplateStartOfTaggedTemplate();
            }
            function isStartOfOptionalPropertyOrElementAccessChain() {
                return token() === 28 /* QuestionDotToken */
                    && lookAhead(nextTokenIsIdentifierOrKeywordOrOpenBracketOrTemplate);
            }
            function tryReparseOptionalChain(node) {
                if (node.flags & 32 /* OptionalChain */) {
                    return true;
                }
                // check for an optional chain in a non-null expression
                if (ts.isNonNullExpression(node)) {
                    var expr = node.expression;
                    while (ts.isNonNullExpression(expr) && !(expr.flags & 32 /* OptionalChain */)) {
                        expr = expr.expression;
                    }
                    if (expr.flags & 32 /* OptionalChain */) {
                        // this is part of an optional chain. Walk down from `node` to `expression` and set the flag.
                        while (ts.isNonNullExpression(node)) {
                            node.flags |= 32 /* OptionalChain */;
                            node = node.expression;
                        }
                        return true;
                    }
                }
                return false;
            }
            function parsePropertyAccessExpressionRest(pos, expression, questionDotToken) {
                var name = parseRightSideOfDot(/*allowIdentifierNames*/ true, /*allowPrivateIdentifiers*/ true);
                var isOptionalChain = questionDotToken || tryReparseOptionalChain(expression);
                var propertyAccess = isOptionalChain ?
                    factory.createPropertyAccessChain(expression, questionDotToken, name) :
                    factory.createPropertyAccessExpression(expression, name);
                if (isOptionalChain && ts.isPrivateIdentifier(propertyAccess.name)) {
                    parseErrorAtRange(propertyAccess.name, ts.Diagnostics.An_optional_chain_cannot_contain_private_identifiers);
                }
                return finishNode(propertyAccess, pos);
            }
            function parseElementAccessExpressionRest(pos, expression, questionDotToken) {
                var argumentExpression;
                if (token() === 23 /* CloseBracketToken */) {
                    argumentExpression = createMissingNode(79 /* Identifier */, /*reportAtCurrentPosition*/ true, ts.Diagnostics.An_element_access_expression_should_take_an_argument);
                }
                else {
                    var argument = allowInAnd(parseExpression);
                    if (ts.isStringOrNumericLiteralLike(argument)) {
                        argument.text = internIdentifier(argument.text);
                    }
                    argumentExpression = argument;
                }
                parseExpected(23 /* CloseBracketToken */);
                var indexedAccess = questionDotToken || tryReparseOptionalChain(expression) ?
                    factory.createElementAccessChain(expression, questionDotToken, argumentExpression) :
                    factory.createElementAccessExpression(expression, argumentExpression);
                return finishNode(indexedAccess, pos);
            }
            function parseMemberExpressionRest(pos, expression, allowOptionalChain) {
                while (true) {
                    var questionDotToken = void 0;
                    var isPropertyAccess = false;
                    if (allowOptionalChain && isStartOfOptionalPropertyOrElementAccessChain()) {
                        questionDotToken = parseExpectedToken(28 /* QuestionDotToken */);
                        isPropertyAccess = ts.tokenIsIdentifierOrKeyword(token());
                    }
                    else {
                        isPropertyAccess = parseOptional(24 /* DotToken */);
                    }
                    if (isPropertyAccess) {
                        expression = parsePropertyAccessExpressionRest(pos, expression, questionDotToken);
                        continue;
                    }
                    if (!questionDotToken && token() === 53 /* ExclamationToken */ && !scanner.hasPrecedingLineBreak()) {
                        nextToken();
                        expression = finishNode(factory.createNonNullExpression(expression), pos);
                        continue;
                    }
                    // when in the [Decorator] context, we do not parse ElementAccess as it could be part of a ComputedPropertyName
                    if ((questionDotToken || !inDecoratorContext()) && parseOptional(22 /* OpenBracketToken */)) {
                        expression = parseElementAccessExpressionRest(pos, expression, questionDotToken);
                        continue;
                    }
                    if (isTemplateStartOfTaggedTemplate()) {
                        expression = parseTaggedTemplateRest(pos, expression, questionDotToken, /*typeArguments*/ undefined);
                        continue;
                    }
                    return expression;
                }
            }
            function isTemplateStartOfTaggedTemplate() {
                return token() === 14 /* NoSubstitutionTemplateLiteral */ || token() === 15 /* TemplateHead */;
            }
            function parseTaggedTemplateRest(pos, tag, questionDotToken, typeArguments) {
                var tagExpression = factory.createTaggedTemplateExpression(tag, typeArguments, token() === 14 /* NoSubstitutionTemplateLiteral */ ?
                    (reScanTemplateHeadOrNoSubstitutionTemplate(), parseLiteralNode()) :
                    parseTemplateExpression(/*isTaggedTemplate*/ true));
                if (questionDotToken || tag.flags & 32 /* OptionalChain */) {
                    tagExpression.flags |= 32 /* OptionalChain */;
                }
                tagExpression.questionDotToken = questionDotToken;
                return finishNode(tagExpression, pos);
            }
            function parseCallExpressionRest(pos, expression) {
                while (true) {
                    expression = parseMemberExpressionRest(pos, expression, /*allowOptionalChain*/ true);
                    var questionDotToken = parseOptionalToken(28 /* QuestionDotToken */);
                    // handle 'foo<<T>()'
                    // parse template arguments only in TypeScript files (not in JavaScript files).
                    if ((contextFlags & 131072 /* JavaScriptFile */) === 0 && (token() === 29 /* LessThanToken */ || token() === 47 /* LessThanLessThanToken */)) {
                        // See if this is the start of a generic invocation.  If so, consume it and
                        // keep checking for postfix expressions.  Otherwise, it's just a '<' that's
                        // part of an arithmetic expression.  Break out so we consume it higher in the
                        // stack.
                        var typeArguments = tryParse(parseTypeArgumentsInExpression);
                        if (typeArguments) {
                            if (isTemplateStartOfTaggedTemplate()) {
                                expression = parseTaggedTemplateRest(pos, expression, questionDotToken, typeArguments);
                                continue;
                            }
                            var argumentList = parseArgumentList();
                            var callExpr = questionDotToken || tryReparseOptionalChain(expression) ?
                                factory.createCallChain(expression, questionDotToken, typeArguments, argumentList) :
                                factory.createCallExpression(expression, typeArguments, argumentList);
                            expression = finishNode(callExpr, pos);
                            continue;
                        }
                    }
                    else if (token() === 20 /* OpenParenToken */) {
                        var argumentList = parseArgumentList();
                        var callExpr = questionDotToken || tryReparseOptionalChain(expression) ?
                            factory.createCallChain(expression, questionDotToken, /*typeArguments*/ undefined, argumentList) :
                            factory.createCallExpression(expression, /*typeArguments*/ undefined, argumentList);
                        expression = finishNode(callExpr, pos);
                        continue;
                    }
                    if (questionDotToken) {
                        // We failed to parse anything, so report a missing identifier here.
                        var name = createMissingNode(79 /* Identifier */, /*reportAtCurrentPosition*/ false, ts.Diagnostics.Identifier_expected);
                        expression = finishNode(factory.createPropertyAccessChain(expression, questionDotToken, name), pos);
                    }
                    break;
                }
                return expression;
            }
            function parseArgumentList() {
                parseExpected(20 /* OpenParenToken */);
                var result = parseDelimitedList(11 /* ArgumentExpressions */, parseArgumentExpression);
                parseExpected(21 /* CloseParenToken */);
                return result;
            }
            function parseTypeArgumentsInExpression() {
                if ((contextFlags & 131072 /* JavaScriptFile */) !== 0) {
                    // TypeArguments must not be parsed in JavaScript files to avoid ambiguity with binary operators.
                    return undefined;
                }
                if (reScanLessThanToken() !== 29 /* LessThanToken */) {
                    return undefined;
                }
                nextToken();
                var typeArguments = parseDelimitedList(20 /* TypeArguments */, parseType);
                if (!parseExpected(31 /* GreaterThanToken */)) {
                    // If it doesn't have the closing `>` then it's definitely not an type argument list.
                    return undefined;
                }
                // If we have a '<', then only parse this as a argument list if the type arguments
                // are complete and we have an open paren.  if we don't, rewind and return nothing.
                return typeArguments && canFollowTypeArgumentsInExpression()
                    ? typeArguments
                    : undefined;
            }
            function canFollowTypeArgumentsInExpression() {
                switch (token()) {
                    case 20 /* OpenParenToken */: // foo<x>(
                    case 14 /* NoSubstitutionTemplateLiteral */: // foo<T> `...`
                    case 15 /* TemplateHead */: // foo<T> `...${100}...`
                    // these are the only tokens can legally follow a type argument
                    // list. So we definitely want to treat them as type arg lists.
                    // falls through
                    case 24 /* DotToken */: // foo<x>.
                    case 21 /* CloseParenToken */: // foo<x>)
                    case 23 /* CloseBracketToken */: // foo<x>]
                    case 58 /* ColonToken */: // foo<x>:
                    case 26 /* SemicolonToken */: // foo<x>;
                    case 57 /* QuestionToken */: // foo<x>?
                    case 34 /* EqualsEqualsToken */: // foo<x> ==
                    case 36 /* EqualsEqualsEqualsToken */: // foo<x> ===
                    case 35 /* ExclamationEqualsToken */: // foo<x> !=
                    case 37 /* ExclamationEqualsEqualsToken */: // foo<x> !==
                    case 55 /* AmpersandAmpersandToken */: // foo<x> &&
                    case 56 /* BarBarToken */: // foo<x> ||
                    case 60 /* QuestionQuestionToken */: // foo<x> ??
                    case 52 /* CaretToken */: // foo<x> ^
                    case 50 /* AmpersandToken */: // foo<x> &
                    case 51 /* BarToken */: // foo<x> |
                    case 19 /* CloseBraceToken */: // foo<x> }
                    case 1 /* EndOfFileToken */: // foo<x>
                        // these cases can't legally follow a type arg list.  However, they're not legal
                        // expressions either.  The user is probably in the middle of a generic type. So
                        // treat it as such.
                        return true;
                    case 27 /* CommaToken */: // foo<x>,
                    case 18 /* OpenBraceToken */: // foo<x> {
                    // We don't want to treat these as type arguments.  Otherwise we'll parse this
                    // as an invocation expression.  Instead, we want to parse out the expression
                    // in isolation from the type arguments.
                    // falls through
                    default:
                        // Anything else treat as an expression.
                        return false;
                }
            }
            function parsePrimaryExpression() {
                switch (token()) {
                    case 8 /* NumericLiteral */:
                    case 9 /* BigIntLiteral */:
                    case 10 /* StringLiteral */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                        return parseLiteralNode();
                    case 108 /* ThisKeyword */:
                    case 106 /* SuperKeyword */:
                    case 104 /* NullKeyword */:
                    case 110 /* TrueKeyword */:
                    case 95 /* FalseKeyword */:
                        return parseTokenNode();
                    case 20 /* OpenParenToken */:
                        return parseParenthesizedExpression();
                    case 22 /* OpenBracketToken */:
                        return parseArrayLiteralExpression();
                    case 18 /* OpenBraceToken */:
                        return parseObjectLiteralExpression();
                    case 130 /* AsyncKeyword */:
                        // Async arrow functions are parsed earlier in parseAssignmentExpressionOrHigher.
                        // If we encounter `async [no LineTerminator here] function` then this is an async
                        // function; otherwise, its an identifier.
                        if (!lookAhead(nextTokenIsFunctionKeywordOnSameLine)) {
                            break;
                        }
                        return parseFunctionExpression();
                    case 84 /* ClassKeyword */:
                        return parseClassExpression();
                    case 98 /* FunctionKeyword */:
                        return parseFunctionExpression();
                    case 103 /* NewKeyword */:
                        return parseNewExpressionOrNewDotTarget();
                    case 43 /* SlashToken */:
                    case 68 /* SlashEqualsToken */:
                        if (reScanSlashToken() === 13 /* RegularExpressionLiteral */) {
                            return parseLiteralNode();
                        }
                        break;
                    case 15 /* TemplateHead */:
                        return parseTemplateExpression(/* isTaggedTemplate */ false);
                }
                return parseIdentifier(ts.Diagnostics.Expression_expected);
            }
            function parseParenthesizedExpression() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(20 /* OpenParenToken */);
                var expression = allowInAnd(parseExpression);
                parseExpected(21 /* CloseParenToken */);
                return withJSDoc(finishNode(factory.createParenthesizedExpression(expression), pos), hasJSDoc);
            }
            function parseSpreadElement() {
                var pos = getNodePos();
                parseExpected(25 /* DotDotDotToken */);
                var expression = parseAssignmentExpressionOrHigher();
                return finishNode(factory.createSpreadElement(expression), pos);
            }
            function parseArgumentOrArrayLiteralElement() {
                return token() === 25 /* DotDotDotToken */ ? parseSpreadElement() :
                    token() === 27 /* CommaToken */ ? finishNode(factory.createOmittedExpression(), getNodePos()) :
                        parseAssignmentExpressionOrHigher();
            }
            function parseArgumentExpression() {
                return doOutsideOfContext(disallowInAndDecoratorContext, parseArgumentOrArrayLiteralElement);
            }
            function parseArrayLiteralExpression() {
                var pos = getNodePos();
                parseExpected(22 /* OpenBracketToken */);
                var multiLine = scanner.hasPrecedingLineBreak();
                var elements = parseDelimitedList(15 /* ArrayLiteralMembers */, parseArgumentOrArrayLiteralElement);
                parseExpected(23 /* CloseBracketToken */);
                return finishNode(factory.createArrayLiteralExpression(elements, multiLine), pos);
            }
            function parseObjectLiteralElement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                if (parseOptionalToken(25 /* DotDotDotToken */)) {
                    var expression = parseAssignmentExpressionOrHigher();
                    return withJSDoc(finishNode(factory.createSpreadAssignment(expression), pos), hasJSDoc);
                }
                var decorators = parseDecorators();
                var modifiers = parseModifiers();
                if (parseContextualModifier(135 /* GetKeyword */)) {
                    return parseAccessorDeclaration(pos, hasJSDoc, decorators, modifiers, 170 /* GetAccessor */);
                }
                if (parseContextualModifier(147 /* SetKeyword */)) {
                    return parseAccessorDeclaration(pos, hasJSDoc, decorators, modifiers, 171 /* SetAccessor */);
                }
                var asteriskToken = parseOptionalToken(41 /* AsteriskToken */);
                var tokenIsIdentifier = isIdentifier();
                var name = parsePropertyName();
                // Disallowing of optional property assignments and definite assignment assertion happens in the grammar checker.
                var questionToken = parseOptionalToken(57 /* QuestionToken */);
                var exclamationToken = parseOptionalToken(53 /* ExclamationToken */);
                if (asteriskToken || token() === 20 /* OpenParenToken */ || token() === 29 /* LessThanToken */) {
                    return parseMethodDeclaration(pos, hasJSDoc, decorators, modifiers, asteriskToken, name, questionToken, exclamationToken);
                }
                // check if it is short-hand property assignment or normal property assignment
                // NOTE: if token is EqualsToken it is interpreted as CoverInitializedName production
                // CoverInitializedName[Yield] :
                //     IdentifierReference[?Yield] Initializer[In, ?Yield]
                // this is necessary because ObjectLiteral productions are also used to cover grammar for ObjectAssignmentPattern
                var node;
                var isShorthandPropertyAssignment = tokenIsIdentifier && (token() !== 58 /* ColonToken */);
                if (isShorthandPropertyAssignment) {
                    var equalsToken = parseOptionalToken(63 /* EqualsToken */);
                    var objectAssignmentInitializer = equalsToken ? allowInAnd(parseAssignmentExpressionOrHigher) : undefined;
                    node = factory.createShorthandPropertyAssignment(name, objectAssignmentInitializer);
                    // Save equals token for error reporting.
                    // TODO(rbuckton): Consider manufacturing this when we need to report an error as it is otherwise not useful.
                    node.equalsToken = equalsToken;
                }
                else {
                    parseExpected(58 /* ColonToken */);
                    var initializer = allowInAnd(parseAssignmentExpressionOrHigher);
                    node = factory.createPropertyAssignment(name, initializer);
                }
                // Decorators, Modifiers, questionToken, and exclamationToken are not supported by property assignments and are reported in the grammar checker
                node.decorators = decorators;
                node.modifiers = modifiers;
                node.questionToken = questionToken;
                node.exclamationToken = exclamationToken;
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseObjectLiteralExpression() {
                var pos = getNodePos();
                var openBracePosition = scanner.getTokenPos();
                parseExpected(18 /* OpenBraceToken */);
                var multiLine = scanner.hasPrecedingLineBreak();
                var properties = parseDelimitedList(12 /* ObjectLiteralMembers */, parseObjectLiteralElement, /*considerSemicolonAsDelimiter*/ true);
                if (!parseExpected(19 /* CloseBraceToken */)) {
                    var lastError = ts.lastOrUndefined(parseDiagnostics);
                    if (lastError && lastError.code === ts.Diagnostics._0_expected.code) {
                        ts.addRelatedInfo(lastError, ts.createDetachedDiagnostic(fileName, openBracePosition, 1, ts.Diagnostics.The_parser_expected_to_find_a_to_match_the_token_here));
                    }
                }
                return finishNode(factory.createObjectLiteralExpression(properties, multiLine), pos);
            }
            function parseFunctionExpression() {
                // GeneratorExpression:
                //      function* BindingIdentifier [Yield][opt](FormalParameters[Yield]){ GeneratorBody }
                //
                // FunctionExpression:
                //      function BindingIdentifier[opt](FormalParameters){ FunctionBody }
                var savedDecoratorContext = inDecoratorContext();
                setDecoratorContext(/*val*/ false);
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                var modifiers = parseModifiers();
                parseExpected(98 /* FunctionKeyword */);
                var asteriskToken = parseOptionalToken(41 /* AsteriskToken */);
                var isGenerator = asteriskToken ? 1 /* Yield */ : 0 /* None */;
                var isAsync = ts.some(modifiers, ts.isAsyncModifier) ? 2 /* Await */ : 0 /* None */;
                var name = isGenerator && isAsync ? doInYieldAndAwaitContext(parseOptionalBindingIdentifier) :
                    isGenerator ? doInYieldContext(parseOptionalBindingIdentifier) :
                        isAsync ? doInAwaitContext(parseOptionalBindingIdentifier) :
                            parseOptionalBindingIdentifier();
                var typeParameters = parseTypeParameters();
                var parameters = parseParameters(isGenerator | isAsync);
                var type = parseReturnType(58 /* ColonToken */, /*isType*/ false);
                var body = parseFunctionBlock(isGenerator | isAsync);
                setDecoratorContext(savedDecoratorContext);
                var node = factory.createFunctionExpression(modifiers, asteriskToken, name, typeParameters, parameters, type, body);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseOptionalBindingIdentifier() {
                return isBindingIdentifier() ? parseBindingIdentifier() : undefined;
            }
            function parseNewExpressionOrNewDotTarget() {
                var pos = getNodePos();
                parseExpected(103 /* NewKeyword */);
                if (parseOptional(24 /* DotToken */)) {
                    var name = parseIdentifierName();
                    return finishNode(factory.createMetaProperty(103 /* NewKeyword */, name), pos);
                }
                var expressionPos = getNodePos();
                var expression = parsePrimaryExpression();
                var typeArguments;
                while (true) {
                    expression = parseMemberExpressionRest(expressionPos, expression, /*allowOptionalChain*/ false);
                    typeArguments = tryParse(parseTypeArgumentsInExpression);
                    if (isTemplateStartOfTaggedTemplate()) {
                        ts.Debug.assert(!!typeArguments, "Expected a type argument list; all plain tagged template starts should be consumed in 'parseMemberExpressionRest'");
                        expression = parseTaggedTemplateRest(expressionPos, expression, /*optionalChain*/ undefined, typeArguments);
                        typeArguments = undefined;
                    }
                    break;
                }
                var argumentsArray;
                if (token() === 20 /* OpenParenToken */) {
                    argumentsArray = parseArgumentList();
                }
                else if (typeArguments) {
                    parseErrorAt(pos, scanner.getStartPos(), ts.Diagnostics.A_new_expression_with_type_arguments_must_always_be_followed_by_a_parenthesized_argument_list);
                }
                return finishNode(factory.createNewExpression(expression, typeArguments, argumentsArray), pos);
            }
            // STATEMENTS
            function parseBlock(ignoreMissingOpenBrace, diagnosticMessage) {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                var openBracePosition = scanner.getTokenPos();
                if (parseExpected(18 /* OpenBraceToken */, diagnosticMessage) || ignoreMissingOpenBrace) {
                    var multiLine = scanner.hasPrecedingLineBreak();
                    var statements = parseList(1 /* BlockStatements */, parseStatement);
                    if (!parseExpected(19 /* CloseBraceToken */)) {
                        var lastError = ts.lastOrUndefined(parseDiagnostics);
                        if (lastError && lastError.code === ts.Diagnostics._0_expected.code) {
                            ts.addRelatedInfo(lastError, ts.createDetachedDiagnostic(fileName, openBracePosition, 1, ts.Diagnostics.The_parser_expected_to_find_a_to_match_the_token_here));
                        }
                    }
                    var result = withJSDoc(finishNode(factory.createBlock(statements, multiLine), pos), hasJSDoc);
                    if (token() === 63 /* EqualsToken */) {
                        parseErrorAtCurrentToken(ts.Diagnostics.Declaration_or_statement_expected_This_follows_a_block_of_statements_so_if_you_intended_to_write_a_destructuring_assignment_you_might_need_to_wrap_the_the_whole_assignment_in_parentheses);
                        nextToken();
                    }
                    return result;
                }
                else {
                    var statements = createMissingList();
                    return withJSDoc(finishNode(factory.createBlock(statements, /*multiLine*/ undefined), pos), hasJSDoc);
                }
            }
            function parseFunctionBlock(flags, diagnosticMessage) {
                var savedYieldContext = inYieldContext();
                setYieldContext(!!(flags & 1 /* Yield */));
                var savedAwaitContext = inAwaitContext();
                setAwaitContext(!!(flags & 2 /* Await */));
                var savedTopLevel = topLevel;
                topLevel = false;
                // We may be in a [Decorator] context when parsing a function expression or
                // arrow function. The body of the function is not in [Decorator] context.
                var saveDecoratorContext = inDecoratorContext();
                if (saveDecoratorContext) {
                    setDecoratorContext(/*val*/ false);
                }
                var block = parseBlock(!!(flags & 16 /* IgnoreMissingOpenBrace */), diagnosticMessage);
                if (saveDecoratorContext) {
                    setDecoratorContext(/*val*/ true);
                }
                topLevel = savedTopLevel;
                setYieldContext(savedYieldContext);
                setAwaitContext(savedAwaitContext);
                return block;
            }
            function parseEmptyStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(26 /* SemicolonToken */);
                return withJSDoc(finishNode(factory.createEmptyStatement(), pos), hasJSDoc);
            }
            function parseIfStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(99 /* IfKeyword */);
                parseExpected(20 /* OpenParenToken */);
                var expression = allowInAnd(parseExpression);
                parseExpected(21 /* CloseParenToken */);
                var thenStatement = parseStatement();
                var elseStatement = parseOptional(91 /* ElseKeyword */) ? parseStatement() : undefined;
                return withJSDoc(finishNode(factory.createIfStatement(expression, thenStatement, elseStatement), pos), hasJSDoc);
            }
            function parseDoStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(90 /* DoKeyword */);
                var statement = parseStatement();
                parseExpected(115 /* WhileKeyword */);
                parseExpected(20 /* OpenParenToken */);
                var expression = allowInAnd(parseExpression);
                parseExpected(21 /* CloseParenToken */);
                // From: https://mail.mozilla.org/pipermail/es-discuss/2011-August/016188.html
                // 157 min --- All allen at wirfs-brock.com CONF --- "do{;}while(false)false" prohibited in
                // spec but allowed in consensus reality. Approved -- this is the de-facto standard whereby
                //  do;while(0)x will have a semicolon inserted before x.
                parseOptional(26 /* SemicolonToken */);
                return withJSDoc(finishNode(factory.createDoStatement(statement, expression), pos), hasJSDoc);
            }
            function parseWhileStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(115 /* WhileKeyword */);
                parseExpected(20 /* OpenParenToken */);
                var expression = allowInAnd(parseExpression);
                parseExpected(21 /* CloseParenToken */);
                var statement = parseStatement();
                return withJSDoc(finishNode(factory.createWhileStatement(expression, statement), pos), hasJSDoc);
            }
            function parseForOrForInOrForOfStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(97 /* ForKeyword */);
                var awaitToken = parseOptionalToken(131 /* AwaitKeyword */);
                parseExpected(20 /* OpenParenToken */);
                var initializer;
                if (token() !== 26 /* SemicolonToken */) {
                    if (token() === 113 /* VarKeyword */ || token() === 119 /* LetKeyword */ || token() === 85 /* ConstKeyword */) {
                        initializer = parseVariableDeclarationList(/*inForStatementInitializer*/ true);
                    }
                    else {
                        initializer = disallowInAnd(parseExpression);
                    }
                }
                var node;
                if (awaitToken ? parseExpected(158 /* OfKeyword */) : parseOptional(158 /* OfKeyword */)) {
                    var expression = allowInAnd(parseAssignmentExpressionOrHigher);
                    parseExpected(21 /* CloseParenToken */);
                    node = factory.createForOfStatement(awaitToken, initializer, expression, parseStatement());
                }
                else if (parseOptional(101 /* InKeyword */)) {
                    var expression = allowInAnd(parseExpression);
                    parseExpected(21 /* CloseParenToken */);
                    node = factory.createForInStatement(initializer, expression, parseStatement());
                }
                else {
                    parseExpected(26 /* SemicolonToken */);
                    var condition = token() !== 26 /* SemicolonToken */ && token() !== 21 /* CloseParenToken */
                        ? allowInAnd(parseExpression)
                        : undefined;
                    parseExpected(26 /* SemicolonToken */);
                    var incrementor = token() !== 21 /* CloseParenToken */
                        ? allowInAnd(parseExpression)
                        : undefined;
                    parseExpected(21 /* CloseParenToken */);
                    node = factory.createForStatement(initializer, condition, incrementor, parseStatement());
                }
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseBreakOrContinueStatement(kind) {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(kind === 244 /* BreakStatement */ ? 81 /* BreakKeyword */ : 86 /* ContinueKeyword */);
                var label = canParseSemicolon() ? undefined : parseIdentifier();
                parseSemicolon();
                var node = kind === 244 /* BreakStatement */
                    ? factory.createBreakStatement(label)
                    : factory.createContinueStatement(label);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseReturnStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(105 /* ReturnKeyword */);
                var expression = canParseSemicolon() ? undefined : allowInAnd(parseExpression);
                parseSemicolon();
                return withJSDoc(finishNode(factory.createReturnStatement(expression), pos), hasJSDoc);
            }
            function parseWithStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(116 /* WithKeyword */);
                parseExpected(20 /* OpenParenToken */);
                var expression = allowInAnd(parseExpression);
                parseExpected(21 /* CloseParenToken */);
                var statement = doInsideOfContext(16777216 /* InWithStatement */, parseStatement);
                return withJSDoc(finishNode(factory.createWithStatement(expression, statement), pos), hasJSDoc);
            }
            function parseCaseClause() {
                var pos = getNodePos();
                parseExpected(82 /* CaseKeyword */);
                var expression = allowInAnd(parseExpression);
                parseExpected(58 /* ColonToken */);
                var statements = parseList(3 /* SwitchClauseStatements */, parseStatement);
                return finishNode(factory.createCaseClause(expression, statements), pos);
            }
            function parseDefaultClause() {
                var pos = getNodePos();
                parseExpected(88 /* DefaultKeyword */);
                parseExpected(58 /* ColonToken */);
                var statements = parseList(3 /* SwitchClauseStatements */, parseStatement);
                return finishNode(factory.createDefaultClause(statements), pos);
            }
            function parseCaseOrDefaultClause() {
                return token() === 82 /* CaseKeyword */ ? parseCaseClause() : parseDefaultClause();
            }
            function parseCaseBlock() {
                var pos = getNodePos();
                parseExpected(18 /* OpenBraceToken */);
                var clauses = parseList(2 /* SwitchClauses */, parseCaseOrDefaultClause);
                parseExpected(19 /* CloseBraceToken */);
                return finishNode(factory.createCaseBlock(clauses), pos);
            }
            function parseSwitchStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(107 /* SwitchKeyword */);
                parseExpected(20 /* OpenParenToken */);
                var expression = allowInAnd(parseExpression);
                parseExpected(21 /* CloseParenToken */);
                var caseBlock = parseCaseBlock();
                return withJSDoc(finishNode(factory.createSwitchStatement(expression, caseBlock), pos), hasJSDoc);
            }
            function parseThrowStatement() {
                // ThrowStatement[Yield] :
                //      throw [no LineTerminator here]Expression[In, ?Yield];
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(109 /* ThrowKeyword */);
                // Because of automatic semicolon insertion, we need to report error if this
                // throw could be terminated with a semicolon.  Note: we can't call 'parseExpression'
                // directly as that might consume an expression on the following line.
                // Instead, we create a "missing" identifier, but don't report an error. The actual error
                // will be reported in the grammar walker.
                var expression = scanner.hasPrecedingLineBreak() ? undefined : allowInAnd(parseExpression);
                if (expression === undefined) {
                    identifierCount++;
                    expression = finishNode(factory.createIdentifier(""), getNodePos());
                }
                if (!tryParseSemicolon()) {
                    parseErrorForMissingSemicolonAfter(expression);
                }
                return withJSDoc(finishNode(factory.createThrowStatement(expression), pos), hasJSDoc);
            }
            // TODO: Review for error recovery
            function parseTryStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(111 /* TryKeyword */);
                var tryBlock = parseBlock(/*ignoreMissingOpenBrace*/ false);
                var catchClause = token() === 83 /* CatchKeyword */ ? parseCatchClause() : undefined;
                // If we don't have a catch clause, then we must have a finally clause.  Try to parse
                // one out no matter what.
                var finallyBlock;
                if (!catchClause || token() === 96 /* FinallyKeyword */) {
                    parseExpected(96 /* FinallyKeyword */);
                    finallyBlock = parseBlock(/*ignoreMissingOpenBrace*/ false);
                }
                return withJSDoc(finishNode(factory.createTryStatement(tryBlock, catchClause, finallyBlock), pos), hasJSDoc);
            }
            function parseCatchClause() {
                var pos = getNodePos();
                parseExpected(83 /* CatchKeyword */);
                var variableDeclaration;
                if (parseOptional(20 /* OpenParenToken */)) {
                    variableDeclaration = parseVariableDeclaration();
                    parseExpected(21 /* CloseParenToken */);
                }
                else {
                    // Keep shape of node to avoid degrading performance.
                    variableDeclaration = undefined;
                }
                var block = parseBlock(/*ignoreMissingOpenBrace*/ false);
                return finishNode(factory.createCatchClause(variableDeclaration, block), pos);
            }
            function parseDebuggerStatement() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                parseExpected(87 /* DebuggerKeyword */);
                parseSemicolon();
                return withJSDoc(finishNode(factory.createDebuggerStatement(), pos), hasJSDoc);
            }
            function parseExpressionOrLabeledStatement() {
                // Avoiding having to do the lookahead for a labeled statement by just trying to parse
                // out an expression, seeing if it is identifier and then seeing if it is followed by
                // a colon.
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                var node;
                var hasParen = token() === 20 /* OpenParenToken */;
                var expression = allowInAnd(parseExpression);
                if (ts.isIdentifier(expression) && parseOptional(58 /* ColonToken */)) {
                    node = factory.createLabeledStatement(expression, parseStatement());
                }
                else {
                    if (!tryParseSemicolon()) {
                        parseErrorForMissingSemicolonAfter(expression);
                    }
                    node = factory.createExpressionStatement(expression);
                    if (hasParen) {
                        // do not parse the same jsdoc twice
                        hasJSDoc = false;
                    }
                }
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function nextTokenIsIdentifierOrKeywordOnSameLine() {
                nextToken();
                return ts.tokenIsIdentifierOrKeyword(token()) && !scanner.hasPrecedingLineBreak();
            }
            function nextTokenIsClassKeywordOnSameLine() {
                nextToken();
                return token() === 84 /* ClassKeyword */ && !scanner.hasPrecedingLineBreak();
            }
            function nextTokenIsFunctionKeywordOnSameLine() {
                nextToken();
                return token() === 98 /* FunctionKeyword */ && !scanner.hasPrecedingLineBreak();
            }
            function nextTokenIsIdentifierOrKeywordOrLiteralOnSameLine() {
                nextToken();
                return (ts.tokenIsIdentifierOrKeyword(token()) || token() === 8 /* NumericLiteral */ || token() === 9 /* BigIntLiteral */ || token() === 10 /* StringLiteral */) && !scanner.hasPrecedingLineBreak();
            }
            function isDeclaration() {
                while (true) {
                    switch (token()) {
                        case 113 /* VarKeyword */:
                        case 119 /* LetKeyword */:
                        case 85 /* ConstKeyword */:
                        case 98 /* FunctionKeyword */:
                        case 84 /* ClassKeyword */:
                        case 92 /* EnumKeyword */:
                            return true;
                        // 'declare', 'module', 'namespace', 'interface'* and 'type' are all legal JavaScript identifiers;
                        // however, an identifier cannot be followed by another identifier on the same line. This is what we
                        // count on to parse out the respective declarations. For instance, we exploit this to say that
                        //
                        //    namespace n
                        //
                        // can be none other than the beginning of a namespace declaration, but need to respect that JavaScript sees
                        //
                        //    namespace
                        //    n
                        //
                        // as the identifier 'namespace' on one line followed by the identifier 'n' on another.
                        // We need to look one token ahead to see if it permissible to try parsing a declaration.
                        //
                        // *Note*: 'interface' is actually a strict mode reserved word. So while
                        //
                        //   "use strict"
                        //   interface
                        //   I {}
                        //
                        // could be legal, it would add complexity for very little gain.
                        case 118 /* InterfaceKeyword */:
                        case 150 /* TypeKeyword */:
                            return nextTokenIsIdentifierOnSameLine();
                        case 140 /* ModuleKeyword */:
                        case 141 /* NamespaceKeyword */:
                            return nextTokenIsIdentifierOrStringLiteralOnSameLine();
                        case 126 /* AbstractKeyword */:
                        case 130 /* AsyncKeyword */:
                        case 134 /* DeclareKeyword */:
                        case 121 /* PrivateKeyword */:
                        case 122 /* ProtectedKeyword */:
                        case 123 /* PublicKeyword */:
                        case 143 /* ReadonlyKeyword */:
                            nextToken();
                            // ASI takes effect for this modifier.
                            if (scanner.hasPrecedingLineBreak()) {
                                return false;
                            }
                            continue;
                        case 155 /* GlobalKeyword */:
                            nextToken();
                            return token() === 18 /* OpenBraceToken */ || token() === 79 /* Identifier */ || token() === 93 /* ExportKeyword */;
                        case 100 /* ImportKeyword */:
                            nextToken();
                            return token() === 10 /* StringLiteral */ || token() === 41 /* AsteriskToken */ ||
                                token() === 18 /* OpenBraceToken */ || ts.tokenIsIdentifierOrKeyword(token());
                        case 93 /* ExportKeyword */:
                            var currentToken_1 = nextToken();
                            if (currentToken_1 === 150 /* TypeKeyword */) {
                                currentToken_1 = lookAhead(nextToken);
                            }
                            if (currentToken_1 === 63 /* EqualsToken */ || currentToken_1 === 41 /* AsteriskToken */ ||
                                currentToken_1 === 18 /* OpenBraceToken */ || currentToken_1 === 88 /* DefaultKeyword */ ||
                                currentToken_1 === 127 /* AsKeyword */) {
                                return true;
                            }
                            continue;
                        case 124 /* StaticKeyword */:
                            nextToken();
                            continue;
                        default:
                            return false;
                    }
                }
            }
            function isStartOfDeclaration() {
                return lookAhead(isDeclaration);
            }
            function isStartOfStatement() {
                switch (token()) {
                    case 59 /* AtToken */:
                    case 26 /* SemicolonToken */:
                    case 18 /* OpenBraceToken */:
                    case 113 /* VarKeyword */:
                    case 119 /* LetKeyword */:
                    case 98 /* FunctionKeyword */:
                    case 84 /* ClassKeyword */:
                    case 92 /* EnumKeyword */:
                    case 99 /* IfKeyword */:
                    case 90 /* DoKeyword */:
                    case 115 /* WhileKeyword */:
                    case 97 /* ForKeyword */:
                    case 86 /* ContinueKeyword */:
                    case 81 /* BreakKeyword */:
                    case 105 /* ReturnKeyword */:
                    case 116 /* WithKeyword */:
                    case 107 /* SwitchKeyword */:
                    case 109 /* ThrowKeyword */:
                    case 111 /* TryKeyword */:
                    case 87 /* DebuggerKeyword */:
                    // 'catch' and 'finally' do not actually indicate that the code is part of a statement,
                    // however, we say they are here so that we may gracefully parse them and error later.
                    // falls through
                    case 83 /* CatchKeyword */:
                    case 96 /* FinallyKeyword */:
                        return true;
                    case 100 /* ImportKeyword */:
                        return isStartOfDeclaration() || lookAhead(nextTokenIsOpenParenOrLessThanOrDot);
                    case 85 /* ConstKeyword */:
                    case 93 /* ExportKeyword */:
                        return isStartOfDeclaration();
                    case 130 /* AsyncKeyword */:
                    case 134 /* DeclareKeyword */:
                    case 118 /* InterfaceKeyword */:
                    case 140 /* ModuleKeyword */:
                    case 141 /* NamespaceKeyword */:
                    case 150 /* TypeKeyword */:
                    case 155 /* GlobalKeyword */:
                        // When these don't start a declaration, they're an identifier in an expression statement
                        return true;
                    case 123 /* PublicKeyword */:
                    case 121 /* PrivateKeyword */:
                    case 122 /* ProtectedKeyword */:
                    case 124 /* StaticKeyword */:
                    case 143 /* ReadonlyKeyword */:
                        // When these don't start a declaration, they may be the start of a class member if an identifier
                        // immediately follows. Otherwise they're an identifier in an expression statement.
                        return isStartOfDeclaration() || !lookAhead(nextTokenIsIdentifierOrKeywordOnSameLine);
                    default:
                        return isStartOfExpression();
                }
            }
            function nextTokenIsBindingIdentifierOrStartOfDestructuring() {
                nextToken();
                return isBindingIdentifier() || token() === 18 /* OpenBraceToken */ || token() === 22 /* OpenBracketToken */;
            }
            function isLetDeclaration() {
                // In ES6 'let' always starts a lexical declaration if followed by an identifier or {
                // or [.
                return lookAhead(nextTokenIsBindingIdentifierOrStartOfDestructuring);
            }
            function parseStatement() {
                switch (token()) {
                    case 26 /* SemicolonToken */:
                        return parseEmptyStatement();
                    case 18 /* OpenBraceToken */:
                        return parseBlock(/*ignoreMissingOpenBrace*/ false);
                    case 113 /* VarKeyword */:
                        return parseVariableStatement(getNodePos(), hasPrecedingJSDocComment(), /*decorators*/ undefined, /*modifiers*/ undefined);
                    case 119 /* LetKeyword */:
                        if (isLetDeclaration()) {
                            return parseVariableStatement(getNodePos(), hasPrecedingJSDocComment(), /*decorators*/ undefined, /*modifiers*/ undefined);
                        }
                        break;
                    case 98 /* FunctionKeyword */:
                        return parseFunctionDeclaration(getNodePos(), hasPrecedingJSDocComment(), /*decorators*/ undefined, /*modifiers*/ undefined);
                    case 84 /* ClassKeyword */:
                        return parseClassDeclaration(getNodePos(), hasPrecedingJSDocComment(), /*decorators*/ undefined, /*modifiers*/ undefined);
                    case 99 /* IfKeyword */:
                        return parseIfStatement();
                    case 90 /* DoKeyword */:
                        return parseDoStatement();
                    case 115 /* WhileKeyword */:
                        return parseWhileStatement();
                    case 97 /* ForKeyword */:
                        return parseForOrForInOrForOfStatement();
                    case 86 /* ContinueKeyword */:
                        return parseBreakOrContinueStatement(243 /* ContinueStatement */);
                    case 81 /* BreakKeyword */:
                        return parseBreakOrContinueStatement(244 /* BreakStatement */);
                    case 105 /* ReturnKeyword */:
                        return parseReturnStatement();
                    case 116 /* WithKeyword */:
                        return parseWithStatement();
                    case 107 /* SwitchKeyword */:
                        return parseSwitchStatement();
                    case 109 /* ThrowKeyword */:
                        return parseThrowStatement();
                    case 111 /* TryKeyword */:
                    // Include 'catch' and 'finally' for error recovery.
                    // falls through
                    case 83 /* CatchKeyword */:
                    case 96 /* FinallyKeyword */:
                        return parseTryStatement();
                    case 87 /* DebuggerKeyword */:
                        return parseDebuggerStatement();
                    case 59 /* AtToken */:
                        return parseDeclaration();
                    case 130 /* AsyncKeyword */:
                    case 118 /* InterfaceKeyword */:
                    case 150 /* TypeKeyword */:
                    case 140 /* ModuleKeyword */:
                    case 141 /* NamespaceKeyword */:
                    case 134 /* DeclareKeyword */:
                    case 85 /* ConstKeyword */:
                    case 92 /* EnumKeyword */:
                    case 93 /* ExportKeyword */:
                    case 100 /* ImportKeyword */:
                    case 121 /* PrivateKeyword */:
                    case 122 /* ProtectedKeyword */:
                    case 123 /* PublicKeyword */:
                    case 126 /* AbstractKeyword */:
                    case 124 /* StaticKeyword */:
                    case 143 /* ReadonlyKeyword */:
                    case 155 /* GlobalKeyword */:
                        if (isStartOfDeclaration()) {
                            return parseDeclaration();
                        }
                        break;
                }
                return parseExpressionOrLabeledStatement();
            }
            function isDeclareModifier(modifier) {
                return modifier.kind === 134 /* DeclareKeyword */;
            }
            function parseDeclaration() {
                // TODO: Can we hold onto the parsed decorators/modifiers and advance the scanner
                //       if we can't reuse the declaration, so that we don't do this work twice?
                //
                // `parseListElement` attempted to get the reused node at this position,
                // but the ambient context flag was not yet set, so the node appeared
                // not reusable in that context.
                var isAmbient = ts.some(lookAhead(function () { return (parseDecorators(), parseModifiers()); }), isDeclareModifier);
                if (isAmbient) {
                    var node = tryReuseAmbientDeclaration();
                    if (node) {
                        return node;
                    }
                }
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                var decorators = parseDecorators();
                var modifiers = parseModifiers();
                if (isAmbient) {
                    for (var _i = 0, _a = modifiers; _i < _a.length; _i++) {
                        var m = _a[_i];
                        m.flags |= 8388608 /* Ambient */;
                    }
                    return doInsideOfContext(8388608 /* Ambient */, function () { return parseDeclarationWorker(pos, hasJSDoc, decorators, modifiers); });
                }
                else {
                    return parseDeclarationWorker(pos, hasJSDoc, decorators, modifiers);
                }
            }
            function tryReuseAmbientDeclaration() {
                return doInsideOfContext(8388608 /* Ambient */, function () {
                    var node = currentNode(parsingContext);
                    if (node) {
                        return consumeNode(node);
                    }
                });
            }
            function parseDeclarationWorker(pos, hasJSDoc, decorators, modifiers) {
                switch (token()) {
                    case 113 /* VarKeyword */:
                    case 119 /* LetKeyword */:
                    case 85 /* ConstKeyword */:
                        return parseVariableStatement(pos, hasJSDoc, decorators, modifiers);
                    case 98 /* FunctionKeyword */:
                        return parseFunctionDeclaration(pos, hasJSDoc, decorators, modifiers);
                    case 84 /* ClassKeyword */:
                        return parseClassDeclaration(pos, hasJSDoc, decorators, modifiers);
                    case 118 /* InterfaceKeyword */:
                        return parseInterfaceDeclaration(pos, hasJSDoc, decorators, modifiers);
                    case 150 /* TypeKeyword */:
                        return parseTypeAliasDeclaration(pos, hasJSDoc, decorators, modifiers);
                    case 92 /* EnumKeyword */:
                        return parseEnumDeclaration(pos, hasJSDoc, decorators, modifiers);
                    case 155 /* GlobalKeyword */:
                    case 140 /* ModuleKeyword */:
                    case 141 /* NamespaceKeyword */:
                        return parseModuleDeclaration(pos, hasJSDoc, decorators, modifiers);
                    case 100 /* ImportKeyword */:
                        return parseImportDeclarationOrImportEqualsDeclaration(pos, hasJSDoc, decorators, modifiers);
                    case 93 /* ExportKeyword */:
                        nextToken();
                        switch (token()) {
                            case 88 /* DefaultKeyword */:
                            case 63 /* EqualsToken */:
                                return parseExportAssignment(pos, hasJSDoc, decorators, modifiers);
                            case 127 /* AsKeyword */:
                                return parseNamespaceExportDeclaration(pos, hasJSDoc, decorators, modifiers);
                            default:
                                return parseExportDeclaration(pos, hasJSDoc, decorators, modifiers);
                        }
                    default:
                        if (decorators || modifiers) {
                            // We reached this point because we encountered decorators and/or modifiers and assumed a declaration
                            // would follow. For recovery and error reporting purposes, return an incomplete declaration.
                            var missing = createMissingNode(274 /* MissingDeclaration */, /*reportAtCurrentPosition*/ true, ts.Diagnostics.Declaration_expected);
                            ts.setTextRangePos(missing, pos);
                            missing.decorators = decorators;
                            missing.modifiers = modifiers;
                            return missing;
                        }
                        return undefined; // TODO: GH#18217
                }
            }
            function nextTokenIsIdentifierOrStringLiteralOnSameLine() {
                nextToken();
                return !scanner.hasPrecedingLineBreak() && (isIdentifier() || token() === 10 /* StringLiteral */);
            }
            function parseFunctionBlockOrSemicolon(flags, diagnosticMessage) {
                if (token() !== 18 /* OpenBraceToken */ && canParseSemicolon()) {
                    parseSemicolon();
                    return;
                }
                return parseFunctionBlock(flags, diagnosticMessage);
            }
            // DECLARATIONS
            function parseArrayBindingElement() {
                var pos = getNodePos();
                if (token() === 27 /* CommaToken */) {
                    return finishNode(factory.createOmittedExpression(), pos);
                }
                var dotDotDotToken = parseOptionalToken(25 /* DotDotDotToken */);
                var name = parseIdentifierOrPattern();
                var initializer = parseInitializer();
                return finishNode(factory.createBindingElement(dotDotDotToken, /*propertyName*/ undefined, name, initializer), pos);
            }
            function parseObjectBindingElement() {
                var pos = getNodePos();
                var dotDotDotToken = parseOptionalToken(25 /* DotDotDotToken */);
                var tokenIsIdentifier = isBindingIdentifier();
                var propertyName = parsePropertyName();
                var name;
                if (tokenIsIdentifier && token() !== 58 /* ColonToken */) {
                    name = propertyName;
                    propertyName = undefined;
                }
                else {
                    parseExpected(58 /* ColonToken */);
                    name = parseIdentifierOrPattern();
                }
                var initializer = parseInitializer();
                return finishNode(factory.createBindingElement(dotDotDotToken, propertyName, name, initializer), pos);
            }
            function parseObjectBindingPattern() {
                var pos = getNodePos();
                parseExpected(18 /* OpenBraceToken */);
                var elements = parseDelimitedList(9 /* ObjectBindingElements */, parseObjectBindingElement);
                parseExpected(19 /* CloseBraceToken */);
                return finishNode(factory.createObjectBindingPattern(elements), pos);
            }
            function parseArrayBindingPattern() {
                var pos = getNodePos();
                parseExpected(22 /* OpenBracketToken */);
                var elements = parseDelimitedList(10 /* ArrayBindingElements */, parseArrayBindingElement);
                parseExpected(23 /* CloseBracketToken */);
                return finishNode(factory.createArrayBindingPattern(elements), pos);
            }
            function isBindingIdentifierOrPrivateIdentifierOrPattern() {
                return token() === 18 /* OpenBraceToken */
                    || token() === 22 /* OpenBracketToken */
                    || token() === 80 /* PrivateIdentifier */
                    || isBindingIdentifier();
            }
            function parseIdentifierOrPattern(privateIdentifierDiagnosticMessage) {
                if (token() === 22 /* OpenBracketToken */) {
                    return parseArrayBindingPattern();
                }
                if (token() === 18 /* OpenBraceToken */) {
                    return parseObjectBindingPattern();
                }
                return parseBindingIdentifier(privateIdentifierDiagnosticMessage);
            }
            function parseVariableDeclarationAllowExclamation() {
                return parseVariableDeclaration(/*allowExclamation*/ true);
            }
            function parseVariableDeclaration(allowExclamation) {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                var name = parseIdentifierOrPattern(ts.Diagnostics.Private_identifiers_are_not_allowed_in_variable_declarations);
                var exclamationToken;
                if (allowExclamation && name.kind === 79 /* Identifier */ &&
                    token() === 53 /* ExclamationToken */ && !scanner.hasPrecedingLineBreak()) {
                    exclamationToken = parseTokenNode();
                }
                var type = parseTypeAnnotation();
                var initializer = isInOrOfKeyword(token()) ? undefined : parseInitializer();
                var node = factory.createVariableDeclaration(name, exclamationToken, type, initializer);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseVariableDeclarationList(inForStatementInitializer) {
                var pos = getNodePos();
                var flags = 0;
                switch (token()) {
                    case 113 /* VarKeyword */:
                        break;
                    case 119 /* LetKeyword */:
                        flags |= 1 /* Let */;
                        break;
                    case 85 /* ConstKeyword */:
                        flags |= 2 /* Const */;
                        break;
                    default:
                        ts.Debug.fail();
                }
                nextToken();
                // The user may have written the following:
                //
                //    for (let of X) { }
                //
                // In this case, we want to parse an empty declaration list, and then parse 'of'
                // as a keyword. The reason this is not automatic is that 'of' is a valid identifier.
                // So we need to look ahead to determine if 'of' should be treated as a keyword in
                // this context.
                // The checker will then give an error that there is an empty declaration list.
                var declarations;
                if (token() === 158 /* OfKeyword */ && lookAhead(canFollowContextualOfKeyword)) {
                    declarations = createMissingList();
                }
                else {
                    var savedDisallowIn = inDisallowInContext();
                    setDisallowInContext(inForStatementInitializer);
                    declarations = parseDelimitedList(8 /* VariableDeclarations */, inForStatementInitializer ? parseVariableDeclaration : parseVariableDeclarationAllowExclamation);
                    setDisallowInContext(savedDisallowIn);
                }
                return finishNode(factory.createVariableDeclarationList(declarations, flags), pos);
            }
            function canFollowContextualOfKeyword() {
                return nextTokenIsIdentifier() && nextToken() === 21 /* CloseParenToken */;
            }
            function parseVariableStatement(pos, hasJSDoc, decorators, modifiers) {
                var declarationList = parseVariableDeclarationList(/*inForStatementInitializer*/ false);
                parseSemicolon();
                var node = factory.createVariableStatement(modifiers, declarationList);
                // Decorators are not allowed on a variable statement, so we keep track of them to report them in the grammar checker.
                node.decorators = decorators;
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseFunctionDeclaration(pos, hasJSDoc, decorators, modifiers) {
                var savedAwaitContext = inAwaitContext();
                var modifierFlags = ts.modifiersToFlags(modifiers);
                parseExpected(98 /* FunctionKeyword */);
                var asteriskToken = parseOptionalToken(41 /* AsteriskToken */);
                // We don't parse the name here in await context, instead we will report a grammar error in the checker.
                var name = modifierFlags & 512 /* Default */ ? parseOptionalBindingIdentifier() : parseBindingIdentifier();
                var isGenerator = asteriskToken ? 1 /* Yield */ : 0 /* None */;
                var isAsync = modifierFlags & 256 /* Async */ ? 2 /* Await */ : 0 /* None */;
                var typeParameters = parseTypeParameters();
                if (modifierFlags & 1 /* Export */)
                    setAwaitContext(/*value*/ true);
                var parameters = parseParameters(isGenerator | isAsync);
                var type = parseReturnType(58 /* ColonToken */, /*isType*/ false);
                var body = parseFunctionBlockOrSemicolon(isGenerator | isAsync, ts.Diagnostics.or_expected);
                setAwaitContext(savedAwaitContext);
                var node = factory.createFunctionDeclaration(decorators, modifiers, asteriskToken, name, typeParameters, parameters, type, body);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseConstructorName() {
                if (token() === 133 /* ConstructorKeyword */) {
                    return parseExpected(133 /* ConstructorKeyword */);
                }
                if (token() === 10 /* StringLiteral */ && lookAhead(nextToken) === 20 /* OpenParenToken */) {
                    return tryParse(function () {
                        var literalNode = parseLiteralNode();
                        return literalNode.text === "constructor" ? literalNode : undefined;
                    });
                }
            }
            function tryParseConstructorDeclaration(pos, hasJSDoc, decorators, modifiers) {
                return tryParse(function () {
                    if (parseConstructorName()) {
                        var typeParameters = parseTypeParameters();
                        var parameters = parseParameters(0 /* None */);
                        var type = parseReturnType(58 /* ColonToken */, /*isType*/ false);
                        var body = parseFunctionBlockOrSemicolon(0 /* None */, ts.Diagnostics.or_expected);
                        var node = factory.createConstructorDeclaration(decorators, modifiers, parameters, body);
                        // Attach `typeParameters` and `type` if they exist so that we can report them in the grammar checker.
                        node.typeParameters = typeParameters;
                        node.type = type;
                        return withJSDoc(finishNode(node, pos), hasJSDoc);
                    }
                });
            }
            function parseMethodDeclaration(pos, hasJSDoc, decorators, modifiers, asteriskToken, name, questionToken, exclamationToken, diagnosticMessage) {
                var isGenerator = asteriskToken ? 1 /* Yield */ : 0 /* None */;
                var isAsync = ts.some(modifiers, ts.isAsyncModifier) ? 2 /* Await */ : 0 /* None */;
                var typeParameters = parseTypeParameters();
                var parameters = parseParameters(isGenerator | isAsync);
                var type = parseReturnType(58 /* ColonToken */, /*isType*/ false);
                var body = parseFunctionBlockOrSemicolon(isGenerator | isAsync, diagnosticMessage);
                var node = factory.createMethodDeclaration(decorators, modifiers, asteriskToken, name, questionToken, typeParameters, parameters, type, body);
                // An exclamation token on a method is invalid syntax and will be handled by the grammar checker
                node.exclamationToken = exclamationToken;
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parsePropertyDeclaration(pos, hasJSDoc, decorators, modifiers, name, questionToken) {
                var exclamationToken = !questionToken && !scanner.hasPrecedingLineBreak() ? parseOptionalToken(53 /* ExclamationToken */) : undefined;
                var type = parseTypeAnnotation();
                var initializer = doOutsideOfContext(8192 /* YieldContext */ | 32768 /* AwaitContext */ | 4096 /* DisallowInContext */, parseInitializer);
                parseSemicolonAfterPropertyName(name, type, initializer);
                var node = factory.createPropertyDeclaration(decorators, modifiers, name, questionToken || exclamationToken, type, initializer);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parsePropertyOrMethodDeclaration(pos, hasJSDoc, decorators, modifiers) {
                var asteriskToken = parseOptionalToken(41 /* AsteriskToken */);
                var name = parsePropertyName();
                // Note: this is not legal as per the grammar.  But we allow it in the parser and
                // report an error in the grammar checker.
                var questionToken = parseOptionalToken(57 /* QuestionToken */);
                if (asteriskToken || token() === 20 /* OpenParenToken */ || token() === 29 /* LessThanToken */) {
                    return parseMethodDeclaration(pos, hasJSDoc, decorators, modifiers, asteriskToken, name, questionToken, /*exclamationToken*/ undefined, ts.Diagnostics.or_expected);
                }
                return parsePropertyDeclaration(pos, hasJSDoc, decorators, modifiers, name, questionToken);
            }
            function parseAccessorDeclaration(pos, hasJSDoc, decorators, modifiers, kind) {
                var name = parsePropertyName();
                var typeParameters = parseTypeParameters();
                var parameters = parseParameters(0 /* None */);
                var type = parseReturnType(58 /* ColonToken */, /*isType*/ false);
                var body = parseFunctionBlockOrSemicolon(0 /* None */);
                var node = kind === 170 /* GetAccessor */
                    ? factory.createGetAccessorDeclaration(decorators, modifiers, name, parameters, type, body)
                    : factory.createSetAccessorDeclaration(decorators, modifiers, name, parameters, body);
                // Keep track of `typeParameters` (for both) and `type` (for setters) if they were parsed those indicate grammar errors
                node.typeParameters = typeParameters;
                if (type && node.kind === 171 /* SetAccessor */)
                    node.type = type;
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function isClassMemberStart() {
                var idToken;
                if (token() === 59 /* AtToken */) {
                    return true;
                }
                // Eat up all modifiers, but hold on to the last one in case it is actually an identifier.
                while (ts.isModifierKind(token())) {
                    idToken = token();
                    // If the idToken is a class modifier (protected, private, public, and static), it is
                    // certain that we are starting to parse class member. This allows better error recovery
                    // Example:
                    //      public foo() ...     // true
                    //      public @dec blah ... // true; we will then report an error later
                    //      export public ...    // true; we will then report an error later
                    if (ts.isClassMemberModifier(idToken)) {
                        return true;
                    }
                    nextToken();
                }
                if (token() === 41 /* AsteriskToken */) {
                    return true;
                }
                // Try to get the first property-like token following all modifiers.
                // This can either be an identifier or the 'get' or 'set' keywords.
                if (isLiteralPropertyName()) {
                    idToken = token();
                    nextToken();
                }
                // Index signatures and computed properties are class members; we can parse.
                if (token() === 22 /* OpenBracketToken */) {
                    return true;
                }
                // If we were able to get any potential identifier...
                if (idToken !== undefined) {
                    // If we have a non-keyword identifier, or if we have an accessor, then it's safe to parse.
                    if (!ts.isKeyword(idToken) || idToken === 147 /* SetKeyword */ || idToken === 135 /* GetKeyword */) {
                        return true;
                    }
                    // If it *is* a keyword, but not an accessor, check a little farther along
                    // to see if it should actually be parsed as a class member.
                    switch (token()) {
                        case 20 /* OpenParenToken */: // Method declaration
                        case 29 /* LessThanToken */: // Generic Method declaration
                        case 53 /* ExclamationToken */: // Non-null assertion on property name
                        case 58 /* ColonToken */: // Type Annotation for declaration
                        case 63 /* EqualsToken */: // Initializer for declaration
                        case 57 /* QuestionToken */: // Not valid, but permitted so that it gets caught later on.
                            return true;
                        default:
                            // Covers
                            //  - Semicolons     (declaration termination)
                            //  - Closing braces (end-of-class, must be declaration)
                            //  - End-of-files   (not valid, but permitted so that it gets caught later on)
                            //  - Line-breaks    (enabling *automatic semicolon insertion*)
                            return canParseSemicolon();
                    }
                }
                return false;
            }
            function parseClassStaticBlockDeclaration(pos, hasJSDoc, decorators, modifiers) {
                parseExpectedToken(124 /* StaticKeyword */);
                var body = parseClassStaticBlockBody();
                return withJSDoc(finishNode(factory.createClassStaticBlockDeclaration(decorators, modifiers, body), pos), hasJSDoc);
            }
            function parseClassStaticBlockBody() {
                var savedYieldContext = inYieldContext();
                var savedAwaitContext = inAwaitContext();
                setYieldContext(false);
                setAwaitContext(true);
                var body = parseBlock(/*ignoreMissingOpenBrace*/ false);
                setYieldContext(savedYieldContext);
                setAwaitContext(savedAwaitContext);
                return body;
            }
            function parseDecoratorExpression() {
                if (inAwaitContext() && token() === 131 /* AwaitKeyword */) {
                    // `@await` is is disallowed in an [Await] context, but can cause parsing to go off the rails
                    // This simply parses the missing identifier and moves on.
                    var pos = getNodePos();
                    var awaitExpression = parseIdentifier(ts.Diagnostics.Expression_expected);
                    nextToken();
                    var memberExpression = parseMemberExpressionRest(pos, awaitExpression, /*allowOptionalChain*/ true);
                    return parseCallExpressionRest(pos, memberExpression);
                }
                return parseLeftHandSideExpressionOrHigher();
            }
            function tryParseDecorator() {
                var pos = getNodePos();
                if (!parseOptional(59 /* AtToken */)) {
                    return undefined;
                }
                var expression = doInDecoratorContext(parseDecoratorExpression);
                return finishNode(factory.createDecorator(expression), pos);
            }
            function parseDecorators() {
                var pos = getNodePos();
                var list, decorator;
                while (decorator = tryParseDecorator()) {
                    list = ts.append(list, decorator);
                }
                return list && createNodeArray(list, pos);
            }
            function tryParseModifier(permitInvalidConstAsModifier, stopOnStartOfClassStaticBlock) {
                var pos = getNodePos();
                var kind = token();
                if (token() === 85 /* ConstKeyword */ && permitInvalidConstAsModifier) {
                    // We need to ensure that any subsequent modifiers appear on the same line
                    // so that when 'const' is a standalone declaration, we don't issue an error.
                    if (!tryParse(nextTokenIsOnSameLineAndCanFollowModifier)) {
                        return undefined;
                    }
                }
                else if (stopOnStartOfClassStaticBlock && token() === 124 /* StaticKeyword */ && lookAhead(nextTokenIsOpenBrace)) {
                    return undefined;
                }
                else {
                    if (!parseAnyContextualModifier()) {
                        return undefined;
                    }
                }
                return finishNode(factory.createToken(kind), pos);
            }
            /*
             * There are situations in which a modifier like 'const' will appear unexpectedly, such as on a class member.
             * In those situations, if we are entirely sure that 'const' is not valid on its own (such as when ASI takes effect
             * and turns it into a standalone declaration), then it is better to parse it and report an error later.
             *
             * In such situations, 'permitInvalidConstAsModifier' should be set to true.
             */
            function parseModifiers(permitInvalidConstAsModifier, stopOnStartOfClassStaticBlock) {
                var pos = getNodePos();
                var list, modifier;
                while (modifier = tryParseModifier(permitInvalidConstAsModifier, stopOnStartOfClassStaticBlock)) {
                    list = ts.append(list, modifier);
                }
                return list && createNodeArray(list, pos);
            }
            function parseModifiersForArrowFunction() {
                var modifiers;
                if (token() === 130 /* AsyncKeyword */) {
                    var pos = getNodePos();
                    nextToken();
                    var modifier = finishNode(factory.createToken(130 /* AsyncKeyword */), pos);
                    modifiers = createNodeArray([modifier], pos);
                }
                return modifiers;
            }
            function parseClassElement() {
                var pos = getNodePos();
                if (token() === 26 /* SemicolonToken */) {
                    nextToken();
                    return finishNode(factory.createSemicolonClassElement(), pos);
                }
                var hasJSDoc = hasPrecedingJSDocComment();
                var decorators = parseDecorators();
                var modifiers = parseModifiers(/*permitInvalidConstAsModifier*/ true, /*stopOnStartOfClassStaticBlock*/ true);
                if (token() === 124 /* StaticKeyword */ && lookAhead(nextTokenIsOpenBrace)) {
                    return parseClassStaticBlockDeclaration(pos, hasJSDoc, decorators, modifiers);
                }
                if (parseContextualModifier(135 /* GetKeyword */)) {
                    return parseAccessorDeclaration(pos, hasJSDoc, decorators, modifiers, 170 /* GetAccessor */);
                }
                if (parseContextualModifier(147 /* SetKeyword */)) {
                    return parseAccessorDeclaration(pos, hasJSDoc, decorators, modifiers, 171 /* SetAccessor */);
                }
                if (token() === 133 /* ConstructorKeyword */ || token() === 10 /* StringLiteral */) {
                    var constructorDeclaration = tryParseConstructorDeclaration(pos, hasJSDoc, decorators, modifiers);
                    if (constructorDeclaration) {
                        return constructorDeclaration;
                    }
                }
                if (isIndexSignature()) {
                    return parseIndexSignatureDeclaration(pos, hasJSDoc, decorators, modifiers);
                }
                // It is very important that we check this *after* checking indexers because
                // the [ token can start an index signature or a computed property name
                if (ts.tokenIsIdentifierOrKeyword(token()) ||
                    token() === 10 /* StringLiteral */ ||
                    token() === 8 /* NumericLiteral */ ||
                    token() === 41 /* AsteriskToken */ ||
                    token() === 22 /* OpenBracketToken */) {
                    var isAmbient = ts.some(modifiers, isDeclareModifier);
                    if (isAmbient) {
                        for (var _i = 0, _a = modifiers; _i < _a.length; _i++) {
                            var m = _a[_i];
                            m.flags |= 8388608 /* Ambient */;
                        }
                        return doInsideOfContext(8388608 /* Ambient */, function () { return parsePropertyOrMethodDeclaration(pos, hasJSDoc, decorators, modifiers); });
                    }
                    else {
                        return parsePropertyOrMethodDeclaration(pos, hasJSDoc, decorators, modifiers);
                    }
                }
                if (decorators || modifiers) {
                    // treat this as a property declaration with a missing name.
                    var name = createMissingNode(79 /* Identifier */, /*reportAtCurrentPosition*/ true, ts.Diagnostics.Declaration_expected);
                    return parsePropertyDeclaration(pos, hasJSDoc, decorators, modifiers, name, /*questionToken*/ undefined);
                }
                // 'isClassMemberStart' should have hinted not to attempt parsing.
                return ts.Debug.fail("Should not have attempted to parse class member declaration.");
            }
            function parseClassExpression() {
                return parseClassDeclarationOrExpression(getNodePos(), hasPrecedingJSDocComment(), /*decorators*/ undefined, /*modifiers*/ undefined, 224 /* ClassExpression */);
            }
            function parseClassDeclaration(pos, hasJSDoc, decorators, modifiers) {
                return parseClassDeclarationOrExpression(pos, hasJSDoc, decorators, modifiers, 255 /* ClassDeclaration */);
            }
            function parseClassDeclarationOrExpression(pos, hasJSDoc, decorators, modifiers, kind) {
                var savedAwaitContext = inAwaitContext();
                parseExpected(84 /* ClassKeyword */);
                // We don't parse the name here in await context, instead we will report a grammar error in the checker.
                var name = parseNameOfClassDeclarationOrExpression();
                var typeParameters = parseTypeParameters();
                if (ts.some(modifiers, ts.isExportModifier))
                    setAwaitContext(/*value*/ true);
                var heritageClauses = parseHeritageClauses();
                var members;
                if (parseExpected(18 /* OpenBraceToken */)) {
                    // ClassTail[Yield,Await] : (Modified) See 14.5
                    //      ClassHeritage[?Yield,?Await]opt { ClassBody[?Yield,?Await]opt }
                    members = parseClassMembers();
                    parseExpected(19 /* CloseBraceToken */);
                }
                else {
                    members = createMissingList();
                }
                setAwaitContext(savedAwaitContext);
                var node = kind === 255 /* ClassDeclaration */
                    ? factory.createClassDeclaration(decorators, modifiers, name, typeParameters, heritageClauses, members)
                    : factory.createClassExpression(decorators, modifiers, name, typeParameters, heritageClauses, members);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseNameOfClassDeclarationOrExpression() {
                // implements is a future reserved word so
                // 'class implements' might mean either
                // - class expression with omitted name, 'implements' starts heritage clause
                // - class with name 'implements'
                // 'isImplementsClause' helps to disambiguate between these two cases
                return isBindingIdentifier() && !isImplementsClause()
                    ? createIdentifier(isBindingIdentifier())
                    : undefined;
            }
            function isImplementsClause() {
                return token() === 117 /* ImplementsKeyword */ && lookAhead(nextTokenIsIdentifierOrKeyword);
            }
            function parseHeritageClauses() {
                // ClassTail[Yield,Await] : (Modified) See 14.5
                //      ClassHeritage[?Yield,?Await]opt { ClassBody[?Yield,?Await]opt }
                if (isHeritageClause()) {
                    return parseList(22 /* HeritageClauses */, parseHeritageClause);
                }
                return undefined;
            }
            function parseHeritageClause() {
                var pos = getNodePos();
                var tok = token();
                ts.Debug.assert(tok === 94 /* ExtendsKeyword */ || tok === 117 /* ImplementsKeyword */); // isListElement() should ensure this.
                nextToken();
                var types = parseDelimitedList(7 /* HeritageClauseElement */, parseExpressionWithTypeArguments);
                return finishNode(factory.createHeritageClause(tok, types), pos);
            }
            function parseExpressionWithTypeArguments() {
                var pos = getNodePos();
                var expression = parseLeftHandSideExpressionOrHigher();
                var typeArguments = tryParseTypeArguments();
                return finishNode(factory.createExpressionWithTypeArguments(expression, typeArguments), pos);
            }
            function tryParseTypeArguments() {
                return token() === 29 /* LessThanToken */ ?
                    parseBracketedList(20 /* TypeArguments */, parseType, 29 /* LessThanToken */, 31 /* GreaterThanToken */) : undefined;
            }
            function isHeritageClause() {
                return token() === 94 /* ExtendsKeyword */ || token() === 117 /* ImplementsKeyword */;
            }
            function parseClassMembers() {
                return parseList(5 /* ClassMembers */, parseClassElement);
            }
            function parseInterfaceDeclaration(pos, hasJSDoc, decorators, modifiers) {
                parseExpected(118 /* InterfaceKeyword */);
                var name = parseIdentifier();
                var typeParameters = parseTypeParameters();
                var heritageClauses = parseHeritageClauses();
                var members = parseObjectTypeMembers();
                var node = factory.createInterfaceDeclaration(decorators, modifiers, name, typeParameters, heritageClauses, members);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseTypeAliasDeclaration(pos, hasJSDoc, decorators, modifiers) {
                parseExpected(150 /* TypeKeyword */);
                var name = parseIdentifier();
                var typeParameters = parseTypeParameters();
                parseExpected(63 /* EqualsToken */);
                var type = token() === 137 /* IntrinsicKeyword */ && tryParse(parseKeywordAndNoDot) || parseType();
                parseSemicolon();
                var node = factory.createTypeAliasDeclaration(decorators, modifiers, name, typeParameters, type);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            // In an ambient declaration, the grammar only allows integer literals as initializers.
            // In a non-ambient declaration, the grammar allows uninitialized members only in a
            // ConstantEnumMemberSection, which starts at the beginning of an enum declaration
            // or any time an integer literal initializer is encountered.
            function parseEnumMember() {
                var pos = getNodePos();
                var hasJSDoc = hasPrecedingJSDocComment();
                var name = parsePropertyName();
                var initializer = allowInAnd(parseInitializer);
                return withJSDoc(finishNode(factory.createEnumMember(name, initializer), pos), hasJSDoc);
            }
            function parseEnumDeclaration(pos, hasJSDoc, decorators, modifiers) {
                parseExpected(92 /* EnumKeyword */);
                var name = parseIdentifier();
                var members;
                if (parseExpected(18 /* OpenBraceToken */)) {
                    members = doOutsideOfYieldAndAwaitContext(function () { return parseDelimitedList(6 /* EnumMembers */, parseEnumMember); });
                    parseExpected(19 /* CloseBraceToken */);
                }
                else {
                    members = createMissingList();
                }
                var node = factory.createEnumDeclaration(decorators, modifiers, name, members);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseModuleBlock() {
                var pos = getNodePos();
                var statements;
                if (parseExpected(18 /* OpenBraceToken */)) {
                    statements = parseList(1 /* BlockStatements */, parseStatement);
                    parseExpected(19 /* CloseBraceToken */);
                }
                else {
                    statements = createMissingList();
                }
                return finishNode(factory.createModuleBlock(statements), pos);
            }
            function parseModuleOrNamespaceDeclaration(pos, hasJSDoc, decorators, modifiers, flags) {
                // If we are parsing a dotted namespace name, we want to
                // propagate the 'Namespace' flag across the names if set.
                var namespaceFlag = flags & 16 /* Namespace */;
                var name = parseIdentifier();
                var body = parseOptional(24 /* DotToken */)
                    ? parseModuleOrNamespaceDeclaration(getNodePos(), /*hasJSDoc*/ false, /*decorators*/ undefined, /*modifiers*/ undefined, 4 /* NestedNamespace */ | namespaceFlag)
                    : parseModuleBlock();
                var node = factory.createModuleDeclaration(decorators, modifiers, name, body, flags);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseAmbientExternalModuleDeclaration(pos, hasJSDoc, decorators, modifiers) {
                var flags = 0;
                var name;
                if (token() === 155 /* GlobalKeyword */) {
                    // parse 'global' as name of global scope augmentation
                    name = parseIdentifier();
                    flags |= 1024 /* GlobalAugmentation */;
                }
                else {
                    name = parseLiteralNode();
                    name.text = internIdentifier(name.text);
                }
                var body;
                if (token() === 18 /* OpenBraceToken */) {
                    body = parseModuleBlock();
                }
                else {
                    parseSemicolon();
                }
                var node = factory.createModuleDeclaration(decorators, modifiers, name, body, flags);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseModuleDeclaration(pos, hasJSDoc, decorators, modifiers) {
                var flags = 0;
                if (token() === 155 /* GlobalKeyword */) {
                    // global augmentation
                    return parseAmbientExternalModuleDeclaration(pos, hasJSDoc, decorators, modifiers);
                }
                else if (parseOptional(141 /* NamespaceKeyword */)) {
                    flags |= 16 /* Namespace */;
                }
                else {
                    parseExpected(140 /* ModuleKeyword */);
                    if (token() === 10 /* StringLiteral */) {
                        return parseAmbientExternalModuleDeclaration(pos, hasJSDoc, decorators, modifiers);
                    }
                }
                return parseModuleOrNamespaceDeclaration(pos, hasJSDoc, decorators, modifiers, flags);
            }
            function isExternalModuleReference() {
                return token() === 144 /* RequireKeyword */ &&
                    lookAhead(nextTokenIsOpenParen);
            }
            function nextTokenIsOpenParen() {
                return nextToken() === 20 /* OpenParenToken */;
            }
            function nextTokenIsOpenBrace() {
                return nextToken() === 18 /* OpenBraceToken */;
            }
            function nextTokenIsSlash() {
                return nextToken() === 43 /* SlashToken */;
            }
            function parseNamespaceExportDeclaration(pos, hasJSDoc, decorators, modifiers) {
                parseExpected(127 /* AsKeyword */);
                parseExpected(141 /* NamespaceKeyword */);
                var name = parseIdentifier();
                parseSemicolon();
                var node = factory.createNamespaceExportDeclaration(name);
                // NamespaceExportDeclaration nodes cannot have decorators or modifiers, so we attach them here so we can report them in the grammar checker
                node.decorators = decorators;
                node.modifiers = modifiers;
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseImportDeclarationOrImportEqualsDeclaration(pos, hasJSDoc, decorators, modifiers) {
                parseExpected(100 /* ImportKeyword */);
                var afterImportPos = scanner.getStartPos();
                // We don't parse the identifier here in await context, instead we will report a grammar error in the checker.
                var identifier;
                if (isIdentifier()) {
                    identifier = parseIdentifier();
                }
                var isTypeOnly = false;
                if (token() !== 154 /* FromKeyword */ &&
                    (identifier === null || identifier === void 0 ? void 0 : identifier.escapedText) === "type" &&
                    (isIdentifier() || tokenAfterImportDefinitelyProducesImportDeclaration())) {
                    isTypeOnly = true;
                    identifier = isIdentifier() ? parseIdentifier() : undefined;
                }
                if (identifier && !tokenAfterImportedIdentifierDefinitelyProducesImportDeclaration()) {
                    return parseImportEqualsDeclaration(pos, hasJSDoc, decorators, modifiers, identifier, isTypeOnly);
                }
                // ImportDeclaration:
                //  import ImportClause from ModuleSpecifier ;
                //  import ModuleSpecifier;
                var importClause;
                if (identifier || // import id
                    token() === 41 /* AsteriskToken */ || // import *
                    token() === 18 /* OpenBraceToken */ // import {
                ) {
                    importClause = parseImportClause(identifier, afterImportPos, isTypeOnly);
                    parseExpected(154 /* FromKeyword */);
                }
                var moduleSpecifier = parseModuleSpecifier();
                parseSemicolon();
                var node = factory.createImportDeclaration(decorators, modifiers, importClause, moduleSpecifier);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function tokenAfterImportDefinitelyProducesImportDeclaration() {
                return token() === 41 /* AsteriskToken */ || token() === 18 /* OpenBraceToken */;
            }
            function tokenAfterImportedIdentifierDefinitelyProducesImportDeclaration() {
                // In `import id ___`, the current token decides whether to produce
                // an ImportDeclaration or ImportEqualsDeclaration.
                return token() === 27 /* CommaToken */ || token() === 154 /* FromKeyword */;
            }
            function parseImportEqualsDeclaration(pos, hasJSDoc, decorators, modifiers, identifier, isTypeOnly) {
                parseExpected(63 /* EqualsToken */);
                var moduleReference = parseModuleReference();
                parseSemicolon();
                var node = factory.createImportEqualsDeclaration(decorators, modifiers, isTypeOnly, identifier, moduleReference);
                var finished = withJSDoc(finishNode(node, pos), hasJSDoc);
                return finished;
            }
            function parseImportClause(identifier, pos, isTypeOnly) {
                // ImportClause:
                //  ImportedDefaultBinding
                //  NameSpaceImport
                //  NamedImports
                //  ImportedDefaultBinding, NameSpaceImport
                //  ImportedDefaultBinding, NamedImports
                // If there was no default import or if there is comma token after default import
                // parse namespace or named imports
                var namedBindings;
                if (!identifier ||
                    parseOptional(27 /* CommaToken */)) {
                    namedBindings = token() === 41 /* AsteriskToken */ ? parseNamespaceImport() : parseNamedImportsOrExports(267 /* NamedImports */);
                }
                return finishNode(factory.createImportClause(isTypeOnly, identifier, namedBindings), pos);
            }
            function parseModuleReference() {
                return isExternalModuleReference()
                    ? parseExternalModuleReference()
                    : parseEntityName(/*allowReservedWords*/ false);
            }
            function parseExternalModuleReference() {
                var pos = getNodePos();
                parseExpected(144 /* RequireKeyword */);
                parseExpected(20 /* OpenParenToken */);
                var expression = parseModuleSpecifier();
                parseExpected(21 /* CloseParenToken */);
                return finishNode(factory.createExternalModuleReference(expression), pos);
            }
            function parseModuleSpecifier() {
                if (token() === 10 /* StringLiteral */) {
                    var result = parseLiteralNode();
                    result.text = internIdentifier(result.text);
                    return result;
                }
                else {
                    // We allow arbitrary expressions here, even though the grammar only allows string
                    // literals.  We check to ensure that it is only a string literal later in the grammar
                    // check pass.
                    return parseExpression();
                }
            }
            function parseNamespaceImport() {
                // NameSpaceImport:
                //  * as ImportedBinding
                var pos = getNodePos();
                parseExpected(41 /* AsteriskToken */);
                parseExpected(127 /* AsKeyword */);
                var name = parseIdentifier();
                return finishNode(factory.createNamespaceImport(name), pos);
            }
            function parseNamedImportsOrExports(kind) {
                var pos = getNodePos();
                // NamedImports:
                //  { }
                //  { ImportsList }
                //  { ImportsList, }
                // ImportsList:
                //  ImportSpecifier
                //  ImportsList, ImportSpecifier
                var node = kind === 267 /* NamedImports */
                    ? factory.createNamedImports(parseBracketedList(23 /* ImportOrExportSpecifiers */, parseImportSpecifier, 18 /* OpenBraceToken */, 19 /* CloseBraceToken */))
                    : factory.createNamedExports(parseBracketedList(23 /* ImportOrExportSpecifiers */, parseExportSpecifier, 18 /* OpenBraceToken */, 19 /* CloseBraceToken */));
                return finishNode(node, pos);
            }
            function parseExportSpecifier() {
                return parseImportOrExportSpecifier(273 /* ExportSpecifier */);
            }
            function parseImportSpecifier() {
                return parseImportOrExportSpecifier(268 /* ImportSpecifier */);
            }
            function parseImportOrExportSpecifier(kind) {
                var pos = getNodePos();
                // ImportSpecifier:
                //   BindingIdentifier
                //   IdentifierName as BindingIdentifier
                // ExportSpecifier:
                //   IdentifierName
                //   IdentifierName as IdentifierName
                var checkIdentifierIsKeyword = ts.isKeyword(token()) && !isIdentifier();
                var checkIdentifierStart = scanner.getTokenPos();
                var checkIdentifierEnd = scanner.getTextPos();
                var identifierName = parseIdentifierName();
                var propertyName;
                var name;
                if (token() === 127 /* AsKeyword */) {
                    propertyName = identifierName;
                    parseExpected(127 /* AsKeyword */);
                    checkIdentifierIsKeyword = ts.isKeyword(token()) && !isIdentifier();
                    checkIdentifierStart = scanner.getTokenPos();
                    checkIdentifierEnd = scanner.getTextPos();
                    name = parseIdentifierName();
                }
                else {
                    name = identifierName;
                }
                if (kind === 268 /* ImportSpecifier */ && checkIdentifierIsKeyword) {
                    parseErrorAt(checkIdentifierStart, checkIdentifierEnd, ts.Diagnostics.Identifier_expected);
                }
                var node = kind === 268 /* ImportSpecifier */
                    ? factory.createImportSpecifier(propertyName, name)
                    : factory.createExportSpecifier(propertyName, name);
                return finishNode(node, pos);
            }
            function parseNamespaceExport(pos) {
                return finishNode(factory.createNamespaceExport(parseIdentifierName()), pos);
            }
            function parseExportDeclaration(pos, hasJSDoc, decorators, modifiers) {
                var savedAwaitContext = inAwaitContext();
                setAwaitContext(/*value*/ true);
                var exportClause;
                var moduleSpecifier;
                var isTypeOnly = parseOptional(150 /* TypeKeyword */);
                var namespaceExportPos = getNodePos();
                if (parseOptional(41 /* AsteriskToken */)) {
                    if (parseOptional(127 /* AsKeyword */)) {
                        exportClause = parseNamespaceExport(namespaceExportPos);
                    }
                    parseExpected(154 /* FromKeyword */);
                    moduleSpecifier = parseModuleSpecifier();
                }
                else {
                    exportClause = parseNamedImportsOrExports(271 /* NamedExports */);
                    // It is not uncommon to accidentally omit the 'from' keyword. Additionally, in editing scenarios,
                    // the 'from' keyword can be parsed as a named export when the export clause is unterminated (i.e. `export { from "moduleName";`)
                    // If we don't have a 'from' keyword, see if we have a string literal such that ASI won't take effect.
                    if (token() === 154 /* FromKeyword */ || (token() === 10 /* StringLiteral */ && !scanner.hasPrecedingLineBreak())) {
                        parseExpected(154 /* FromKeyword */);
                        moduleSpecifier = parseModuleSpecifier();
                    }
                }
                parseSemicolon();
                setAwaitContext(savedAwaitContext);
                var node = factory.createExportDeclaration(decorators, modifiers, isTypeOnly, exportClause, moduleSpecifier);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function parseExportAssignment(pos, hasJSDoc, decorators, modifiers) {
                var savedAwaitContext = inAwaitContext();
                setAwaitContext(/*value*/ true);
                var isExportEquals;
                if (parseOptional(63 /* EqualsToken */)) {
                    isExportEquals = true;
                }
                else {
                    parseExpected(88 /* DefaultKeyword */);
                }
                var expression = parseAssignmentExpressionOrHigher();
                parseSemicolon();
                setAwaitContext(savedAwaitContext);
                var node = factory.createExportAssignment(decorators, modifiers, isExportEquals, expression);
                return withJSDoc(finishNode(node, pos), hasJSDoc);
            }
            function setExternalModuleIndicator(sourceFile) {
                // Try to use the first top-level import/export when available, then
                // fall back to looking for an 'import.meta' somewhere in the tree if necessary.
                sourceFile.externalModuleIndicator =
                    ts.forEach(sourceFile.statements, isAnExternalModuleIndicatorNode) ||
                        getImportMetaIfNecessary(sourceFile);
            }
            function isAnExternalModuleIndicatorNode(node) {
                return hasModifierOfKind(node, 93 /* ExportKeyword */)
                    || ts.isImportEqualsDeclaration(node) && ts.isExternalModuleReference(node.moduleReference)
                    || ts.isImportDeclaration(node)
                    || ts.isExportAssignment(node)
                    || ts.isExportDeclaration(node) ? node : undefined;
            }
            function getImportMetaIfNecessary(sourceFile) {
                return sourceFile.flags & 2097152 /* PossiblyContainsImportMeta */ ?
                    walkTreeForExternalModuleIndicators(sourceFile) :
                    undefined;
            }
            function walkTreeForExternalModuleIndicators(node) {
                return isImportMeta(node) ? node : forEachChild(node, walkTreeForExternalModuleIndicators);
            }
            /** Do not use hasModifier inside the parser; it relies on parent pointers. Use this instead. */
            function hasModifierOfKind(node, kind) {
                return ts.some(node.modifiers, function (m) { return m.kind === kind; });
            }
            function isImportMeta(node) {
                return ts.isMetaProperty(node) && node.keywordToken === 100 /* ImportKeyword */ && node.name.escapedText === "meta";
            }
            var ParsingContext;
            (function (ParsingContext) {
                ParsingContext[ParsingContext["SourceElements"] = 0] = "SourceElements";
                ParsingContext[ParsingContext["BlockStatements"] = 1] = "BlockStatements";
                ParsingContext[ParsingContext["SwitchClauses"] = 2] = "SwitchClauses";
                ParsingContext[ParsingContext["SwitchClauseStatements"] = 3] = "SwitchClauseStatements";
                ParsingContext[ParsingContext["TypeMembers"] = 4] = "TypeMembers";
                ParsingContext[ParsingContext["ClassMembers"] = 5] = "ClassMembers";
                ParsingContext[ParsingContext["EnumMembers"] = 6] = "EnumMembers";
                ParsingContext[ParsingContext["HeritageClauseElement"] = 7] = "HeritageClauseElement";
                ParsingContext[ParsingContext["VariableDeclarations"] = 8] = "VariableDeclarations";
                ParsingContext[ParsingContext["ObjectBindingElements"] = 9] = "ObjectBindingElements";
                ParsingContext[ParsingContext["ArrayBindingElements"] = 10] = "ArrayBindingElements";
                ParsingContext[ParsingContext["ArgumentExpressions"] = 11] = "ArgumentExpressions";
                ParsingContext[ParsingContext["ObjectLiteralMembers"] = 12] = "ObjectLiteralMembers";
                ParsingContext[ParsingContext["JsxAttributes"] = 13] = "JsxAttributes";
                ParsingContext[ParsingContext["JsxChildren"] = 14] = "JsxChildren";
                ParsingContext[ParsingContext["ArrayLiteralMembers"] = 15] = "ArrayLiteralMembers";
                ParsingContext[ParsingContext["Parameters"] = 16] = "Parameters";
                ParsingContext[ParsingContext["JSDocParameters"] = 17] = "JSDocParameters";
                ParsingContext[ParsingContext["RestProperties"] = 18] = "RestProperties";
                ParsingContext[ParsingContext["TypeParameters"] = 19] = "TypeParameters";
                ParsingContext[ParsingContext["TypeArguments"] = 20] = "TypeArguments";
                ParsingContext[ParsingContext["TupleElementTypes"] = 21] = "TupleElementTypes";
                ParsingContext[ParsingContext["HeritageClauses"] = 22] = "HeritageClauses";
                ParsingContext[ParsingContext["ImportOrExportSpecifiers"] = 23] = "ImportOrExportSpecifiers";
                ParsingContext[ParsingContext["Count"] = 24] = "Count"; // Number of parsing contexts
            })(ParsingContext || (ParsingContext = {}));
            var Tristate;
            (function (Tristate) {
                Tristate[Tristate["False"] = 0] = "False";
                Tristate[Tristate["True"] = 1] = "True";
                Tristate[Tristate["Unknown"] = 2] = "Unknown";
            })(Tristate || (Tristate = {}));
            var JSDocParser;
            (function (JSDocParser) {
                function parseJSDocTypeExpressionForTests(content, start, length) {
                    initializeState("file.js", content, 99 /* Latest */, /*_syntaxCursor:*/ undefined, 1 /* JS */);
                    scanner.setText(content, start, length);
                    currentToken = scanner.scan();
                    var jsDocTypeExpression = parseJSDocTypeExpression();
                    var sourceFile = createSourceFile("file.js", 99 /* Latest */, 1 /* JS */, /*isDeclarationFile*/ false, [], factory.createToken(1 /* EndOfFileToken */), 0 /* None */);
                    var diagnostics = ts.attachFileToDiagnostics(parseDiagnostics, sourceFile);
                    if (jsDocDiagnostics) {
                        sourceFile.jsDocDiagnostics = ts.attachFileToDiagnostics(jsDocDiagnostics, sourceFile);
                    }
                    clearState();
                    return jsDocTypeExpression ? { jsDocTypeExpression: jsDocTypeExpression, diagnostics: diagnostics } : undefined;
                }
                JSDocParser.parseJSDocTypeExpressionForTests = parseJSDocTypeExpressionForTests;
                // Parses out a JSDoc type expression.
                function parseJSDocTypeExpression(mayOmitBraces) {
                    var pos = getNodePos();
                    var hasBrace = (mayOmitBraces ? parseOptional : parseExpected)(18 /* OpenBraceToken */);
                    var type = doInsideOfContext(4194304 /* JSDoc */, parseJSDocType);
                    if (!mayOmitBraces || hasBrace) {
                        parseExpectedJSDoc(19 /* CloseBraceToken */);
                    }
                    var result = factory.createJSDocTypeExpression(type);
                    fixupParentReferences(result);
                    return finishNode(result, pos);
                }
                JSDocParser.parseJSDocTypeExpression = parseJSDocTypeExpression;
                function parseJSDocNameReference() {
                    var pos = getNodePos();
                    var hasBrace = parseOptional(18 /* OpenBraceToken */);
                    var p2 = getNodePos();
                    var entityName = parseEntityName(/* allowReservedWords*/ false);
                    while (token() === 80 /* PrivateIdentifier */) {
                        reScanHashToken(); // rescan #id as # id
                        nextTokenJSDoc(); // then skip the #
                        entityName = finishNode(factory.createJSDocMemberName(entityName, parseIdentifier()), p2);
                    }
                    if (hasBrace) {
                        parseExpectedJSDoc(19 /* CloseBraceToken */);
                    }
                    var result = factory.createJSDocNameReference(entityName);
                    fixupParentReferences(result);
                    return finishNode(result, pos);
                }
                JSDocParser.parseJSDocNameReference = parseJSDocNameReference;
                function parseIsolatedJSDocComment(content, start, length) {
                    initializeState("", content, 99 /* Latest */, /*_syntaxCursor:*/ undefined, 1 /* JS */);
                    var jsDoc = doInsideOfContext(4194304 /* JSDoc */, function () { return parseJSDocCommentWorker(start, length); });
                    var sourceFile = { languageVariant: 0 /* Standard */, text: content };
                    var diagnostics = ts.attachFileToDiagnostics(parseDiagnostics, sourceFile);
                    clearState();
                    return jsDoc ? { jsDoc: jsDoc, diagnostics: diagnostics } : undefined;
                }
                JSDocParser.parseIsolatedJSDocComment = parseIsolatedJSDocComment;
                function parseJSDocComment(parent, start, length) {
                    var saveToken = currentToken;
                    var saveParseDiagnosticsLength = parseDiagnostics.length;
                    var saveParseErrorBeforeNextFinishedNode = parseErrorBeforeNextFinishedNode;
                    var comment = doInsideOfContext(4194304 /* JSDoc */, function () { return parseJSDocCommentWorker(start, length); });
                    ts.setParent(comment, parent);
                    if (contextFlags & 131072 /* JavaScriptFile */) {
                        if (!jsDocDiagnostics) {
                            jsDocDiagnostics = [];
                        }
                        jsDocDiagnostics.push.apply(jsDocDiagnostics, parseDiagnostics);
                    }
                    currentToken = saveToken;
                    parseDiagnostics.length = saveParseDiagnosticsLength;
                    parseErrorBeforeNextFinishedNode = saveParseErrorBeforeNextFinishedNode;
                    return comment;
                }
                JSDocParser.parseJSDocComment = parseJSDocComment;
                var JSDocState;
                (function (JSDocState) {
                    JSDocState[JSDocState["BeginningOfLine"] = 0] = "BeginningOfLine";
                    JSDocState[JSDocState["SawAsterisk"] = 1] = "SawAsterisk";
                    JSDocState[JSDocState["SavingComments"] = 2] = "SavingComments";
                    JSDocState[JSDocState["SavingBackticks"] = 3] = "SavingBackticks";
                })(JSDocState || (JSDocState = {}));
                var PropertyLikeParse;
                (function (PropertyLikeParse) {
                    PropertyLikeParse[PropertyLikeParse["Property"] = 1] = "Property";
                    PropertyLikeParse[PropertyLikeParse["Parameter"] = 2] = "Parameter";
                    PropertyLikeParse[PropertyLikeParse["CallbackParameter"] = 4] = "CallbackParameter";
                })(PropertyLikeParse || (PropertyLikeParse = {}));
                function parseJSDocCommentWorker(start, length) {
                    if (start === void 0) {
                        start = 0;
                    }
                    var content = sourceText;
                    var end = length === undefined ? content.length : start + length;
                    length = end - start;
                    ts.Debug.assert(start >= 0);
                    ts.Debug.assert(start <= end);
                    ts.Debug.assert(end <= content.length);
                    // Check for /** (JSDoc opening part)
                    if (!isJSDocLikeText(content, start)) {
                        return undefined;
                    }
                    var tags;
                    var tagsPos;
                    var tagsEnd;
                    var linkEnd;
                    var commentsPos;
                    var comments = [];
                    var parts = [];
                    // + 3 for leading /**, - 5 in total for /** */
                    return scanner.scanRange(start + 3, length - 5, function () {
                        // Initially we can parse out a tag.  We also have seen a starting asterisk.
                        // This is so that /** * @type */ doesn't parse.
                        var state = 1 /* SawAsterisk */;
                        var margin;
                        // + 4 for leading '/** '
                        // + 1 because the last index of \n is always one index before the first character in the line and coincidentally, if there is no \n before start, it is -1, which is also one index before the first character
                        var indent = start - (content.lastIndexOf("\n", start) + 1) + 4;
                        function pushComment(text) {
                            if (!margin) {
                                margin = indent;
                            }
                            comments.push(text);
                            indent += text.length;
                        }
                        nextTokenJSDoc();
                        while (parseOptionalJsdoc(5 /* WhitespaceTrivia */))
                            ;
                        if (parseOptionalJsdoc(4 /* NewLineTrivia */)) {
                            state = 0 /* BeginningOfLine */;
                            indent = 0;
                        }
                        loop: while (true) {
                            switch (token()) {
                                case 59 /* AtToken */:
                                    if (state === 0 /* BeginningOfLine */ || state === 1 /* SawAsterisk */) {
                                        removeTrailingWhitespace(comments);
                                        if (!commentsPos)
                                            commentsPos = getNodePos();
                                        addTag(parseTag(indent));
                                        // NOTE: According to usejsdoc.org, a tag goes to end of line, except the last tag.
                                        // Real-world comments may break this rule, so "BeginningOfLine" will not be a real line beginning
                                        // for malformed examples like `/** @param {string} x @returns {number} the length */`
                                        state = 0 /* BeginningOfLine */;
                                        margin = undefined;
                                    }
                                    else {
                                        pushComment(scanner.getTokenText());
                                    }
                                    break;
                                case 4 /* NewLineTrivia */:
                                    comments.push(scanner.getTokenText());
                                    state = 0 /* BeginningOfLine */;
                                    indent = 0;
                                    break;
                                case 41 /* AsteriskToken */:
                                    var asterisk = scanner.getTokenText();
                                    if (state === 1 /* SawAsterisk */ || state === 2 /* SavingComments */) {
                                        // If we've already seen an asterisk, then we can no longer parse a tag on this line
                                        state = 2 /* SavingComments */;
                                        pushComment(asterisk);
                                    }
                                    else {
                                        // Ignore the first asterisk on a line
                                        state = 1 /* SawAsterisk */;
                                        indent += asterisk.length;
                                    }
                                    break;
                                case 5 /* WhitespaceTrivia */:
                                    // only collect whitespace if we're already saving comments or have just crossed the comment indent margin
                                    var whitespace = scanner.getTokenText();
                                    if (state === 2 /* SavingComments */) {
                                        comments.push(whitespace);
                                    }
                                    else if (margin !== undefined && indent + whitespace.length > margin) {
                                        comments.push(whitespace.slice(margin - indent));
                                    }
                                    indent += whitespace.length;
                                    break;
                                case 1 /* EndOfFileToken */:
                                    break loop;
                                case 18 /* OpenBraceToken */:
                                    state = 2 /* SavingComments */;
                                    var commentEnd = scanner.getStartPos();
                                    var linkStart = scanner.getTextPos() - 1;
                                    var link = parseJSDocLink(linkStart);
                                    if (link) {
                                        if (!linkEnd) {
                                            removeLeadingNewlines(comments);
                                        }
                                        parts.push(finishNode(factory.createJSDocText(comments.join("")), linkEnd !== null && linkEnd !== void 0 ? linkEnd : start, commentEnd));
                                        parts.push(link);
                                        comments = [];
                                        linkEnd = scanner.getTextPos();
                                        break;
                                    }
                                // fallthrough if it's not a {@link sequence
                                default:
                                    // Anything else is doc comment text. We just save it. Because it
                                    // wasn't a tag, we can no longer parse a tag on this line until we hit the next
                                    // line break.
                                    state = 2 /* SavingComments */;
                                    pushComment(scanner.getTokenText());
                                    break;
                            }
                            nextTokenJSDoc();
                        }
                        removeTrailingWhitespace(comments);
                        if (parts.length && comments.length) {
                            parts.push(finishNode(factory.createJSDocText(comments.join("")), linkEnd !== null && linkEnd !== void 0 ? linkEnd : start, commentsPos));
                        }
                        if (parts.length && tags)
                            ts.Debug.assertIsDefined(commentsPos, "having parsed tags implies that the end of the comment span should be set");
                        var tagsArray = tags && createNodeArray(tags, tagsPos, tagsEnd);
                        return finishNode(factory.createJSDocComment(parts.length ? createNodeArray(parts, start, commentsPos) : comments.length ? comments.join("") : undefined, tagsArray), start, end);
                    });
                    function removeLeadingNewlines(comments) {
                        while (comments.length && (comments[0] === "\n" || comments[0] === "\r")) {
                            comments.shift();
                        }
                    }
                    function removeTrailingWhitespace(comments) {
                        while (comments.length && comments[comments.length - 1].trim() === "") {
                            comments.pop();
                        }
                    }
                    function isNextNonwhitespaceTokenEndOfFile() {
                        // We must use infinite lookahead, as there could be any number of newlines :(
                        while (true) {
                            nextTokenJSDoc();
                            if (token() === 1 /* EndOfFileToken */) {
                                return true;
                            }
                            if (!(token() === 5 /* WhitespaceTrivia */ || token() === 4 /* NewLineTrivia */)) {
                                return false;
                            }
                        }
                    }
                    function skipWhitespace() {
                        if (token() === 5 /* WhitespaceTrivia */ || token() === 4 /* NewLineTrivia */) {
                            if (lookAhead(isNextNonwhitespaceTokenEndOfFile)) {
                                return; // Don't skip whitespace prior to EoF (or end of comment) - that shouldn't be included in any node's range
                            }
                        }
                        while (token() === 5 /* WhitespaceTrivia */ || token() === 4 /* NewLineTrivia */) {
                            nextTokenJSDoc();
                        }
                    }
                    function skipWhitespaceOrAsterisk() {
                        if (token() === 5 /* WhitespaceTrivia */ || token() === 4 /* NewLineTrivia */) {
                            if (lookAhead(isNextNonwhitespaceTokenEndOfFile)) {
                                return ""; // Don't skip whitespace prior to EoF (or end of comment) - that shouldn't be included in any node's range
                            }
                        }
                        var precedingLineBreak = scanner.hasPrecedingLineBreak();
                        var seenLineBreak = false;
                        var indentText = "";
                        while ((precedingLineBreak && token() === 41 /* AsteriskToken */) || token() === 5 /* WhitespaceTrivia */ || token() === 4 /* NewLineTrivia */) {
                            indentText += scanner.getTokenText();
                            if (token() === 4 /* NewLineTrivia */) {
                                precedingLineBreak = true;
                                seenLineBreak = true;
                                indentText = "";
                            }
                            else if (token() === 41 /* AsteriskToken */) {
                                precedingLineBreak = false;
                            }
                            nextTokenJSDoc();
                        }
                        return seenLineBreak ? indentText : "";
                    }
                    function parseTag(margin) {
                        ts.Debug.assert(token() === 59 /* AtToken */);
                        var start = scanner.getTokenPos();
                        nextTokenJSDoc();
                        var tagName = parseJSDocIdentifierName(/*message*/ undefined);
                        var indentText = skipWhitespaceOrAsterisk();
                        var tag;
                        switch (tagName.escapedText) {
                            case "author":
                                tag = parseAuthorTag(start, tagName, margin, indentText);
                                break;
                            case "implements":
                                tag = parseImplementsTag(start, tagName, margin, indentText);
                                break;
                            case "augments":
                            case "extends":
                                tag = parseAugmentsTag(start, tagName, margin, indentText);
                                break;
                            case "class":
                            case "constructor":
                                tag = parseSimpleTag(start, factory.createJSDocClassTag, tagName, margin, indentText);
                                break;
                            case "public":
                                tag = parseSimpleTag(start, factory.createJSDocPublicTag, tagName, margin, indentText);
                                break;
                            case "private":
                                tag = parseSimpleTag(start, factory.createJSDocPrivateTag, tagName, margin, indentText);
                                break;
                            case "protected":
                                tag = parseSimpleTag(start, factory.createJSDocProtectedTag, tagName, margin, indentText);
                                break;
                            case "readonly":
                                tag = parseSimpleTag(start, factory.createJSDocReadonlyTag, tagName, margin, indentText);
                                break;
                            case "override":
                                tag = parseSimpleTag(start, factory.createJSDocOverrideTag, tagName, margin, indentText);
                                break;
                            case "deprecated":
                                hasDeprecatedTag = true;
                                tag = parseSimpleTag(start, factory.createJSDocDeprecatedTag, tagName, margin, indentText);
                                break;
                            case "this":
                                tag = parseThisTag(start, tagName, margin, indentText);
                                break;
                            case "enum":
                                tag = parseEnumTag(start, tagName, margin, indentText);
                                break;
                            case "arg":
                            case "argument":
                            case "param":
                                return parseParameterOrPropertyTag(start, tagName, 2 /* Parameter */, margin);
                            case "return":
                            case "returns":
                                tag = parseReturnTag(start, tagName, margin, indentText);
                                break;
                            case "template":
                                tag = parseTemplateTag(start, tagName, margin, indentText);
                                break;
                            case "type":
                                tag = parseTypeTag(start, tagName, margin, indentText);
                                break;
                            case "typedef":
                                tag = parseTypedefTag(start, tagName, margin, indentText);
                                break;
                            case "callback":
                                tag = parseCallbackTag(start, tagName, margin, indentText);
                                break;
                            case "see":
                                tag = parseSeeTag(start, tagName, margin, indentText);
                                break;
                            default:
                                tag = parseUnknownTag(start, tagName, margin, indentText);
                                break;
                        }
                        return tag;
                    }
                    function parseTrailingTagComments(pos, end, margin, indentText) {
                        // some tags, like typedef and callback, have already parsed their comments earlier
                        if (!indentText) {
                            margin += end - pos;
                        }
                        return parseTagComments(margin, indentText.slice(margin));
                    }
                    function parseTagComments(indent, initialMargin) {
                        var commentsPos = getNodePos();
                        var comments = [];
                        var parts = [];
                        var linkEnd;
                        var state = 0 /* BeginningOfLine */;
                        var previousWhitespace = true;
                        var margin;
                        function pushComment(text) {
                            if (!margin) {
                                margin = indent;
                            }
                            comments.push(text);
                            indent += text.length;
                        }
                        if (initialMargin !== undefined) {
                            // jump straight to saving comments if there is some initial indentation
                            if (initialMargin !== "") {
                                pushComment(initialMargin);
                            }
                            state = 1 /* SawAsterisk */;
                        }
                        var tok = token();
                        loop: while (true) {
                            switch (tok) {
                                case 4 /* NewLineTrivia */:
                                    state = 0 /* BeginningOfLine */;
                                    // don't use pushComment here because we want to keep the margin unchanged
                                    comments.push(scanner.getTokenText());
                                    indent = 0;
                                    break;
                                case 59 /* AtToken */:
                                    if (state === 3 /* SavingBackticks */
                                        || state === 2 /* SavingComments */ && (!previousWhitespace || lookAhead(isNextJSDocTokenWhitespace))) {
                                        // @ doesn't start a new tag inside ``, and inside a comment, only after whitespace or not before whitespace
                                        comments.push(scanner.getTokenText());
                                        break;
                                    }
                                    scanner.setTextPos(scanner.getTextPos() - 1);
                                // falls through
                                case 1 /* EndOfFileToken */:
                                    // Done
                                    break loop;
                                case 5 /* WhitespaceTrivia */:
                                    if (state === 2 /* SavingComments */ || state === 3 /* SavingBackticks */) {
                                        pushComment(scanner.getTokenText());
                                    }
                                    else {
                                        var whitespace = scanner.getTokenText();
                                        // if the whitespace crosses the margin, take only the whitespace that passes the margin
                                        if (margin !== undefined && indent + whitespace.length > margin) {
                                            comments.push(whitespace.slice(margin - indent));
                                        }
                                        indent += whitespace.length;
                                    }
                                    break;
                                case 18 /* OpenBraceToken */:
                                    state = 2 /* SavingComments */;
                                    var commentEnd = scanner.getStartPos();
                                    var linkStart = scanner.getTextPos() - 1;
                                    var link = parseJSDocLink(linkStart);
                                    if (link) {
                                        parts.push(finishNode(factory.createJSDocText(comments.join("")), linkEnd !== null && linkEnd !== void 0 ? linkEnd : commentsPos, commentEnd));
                                        parts.push(link);
                                        comments = [];
                                        linkEnd = scanner.getTextPos();
                                    }
                                    else {
                                        pushComment(scanner.getTokenText());
                                    }
                                    break;
                                case 61 /* BacktickToken */:
                                    if (state === 3 /* SavingBackticks */) {
                                        state = 2 /* SavingComments */;
                                    }
                                    else {
                                        state = 3 /* SavingBackticks */;
                                    }
                                    pushComment(scanner.getTokenText());
                                    break;
                                case 41 /* AsteriskToken */:
                                    if (state === 0 /* BeginningOfLine */) {
                                        // leading asterisks start recording on the *next* (non-whitespace) token
                                        state = 1 /* SawAsterisk */;
                                        indent += 1;
                                        break;
                                    }
                                // record the * as a comment
                                // falls through
                                default:
                                    if (state !== 3 /* SavingBackticks */) {
                                        state = 2 /* SavingComments */; // leading identifiers start recording as well
                                    }
                                    pushComment(scanner.getTokenText());
                                    break;
                            }
                            previousWhitespace = token() === 5 /* WhitespaceTrivia */;
                            tok = nextTokenJSDoc();
                        }
                        removeLeadingNewlines(comments);
                        removeTrailingWhitespace(comments);
                        if (parts.length) {
                            if (comments.length) {
                                parts.push(finishNode(factory.createJSDocText(comments.join("")), linkEnd !== null && linkEnd !== void 0 ? linkEnd : commentsPos));
                            }
                            return createNodeArray(parts, commentsPos, scanner.getTextPos());
                        }
                        else if (comments.length) {
                            return comments.join("");
                        }
                    }
                    function isNextJSDocTokenWhitespace() {
                        var next = nextTokenJSDoc();
                        return next === 5 /* WhitespaceTrivia */ || next === 4 /* NewLineTrivia */;
                    }
                    function parseJSDocLink(start) {
                        var linkType = tryParse(parseJSDocLinkPrefix);
                        if (!linkType) {
                            return undefined;
                        }
                        nextTokenJSDoc(); // start at token after link, then skip any whitespace
                        skipWhitespace();
                        // parseEntityName logs an error for non-identifier, so create a MissingNode ourselves to avoid the error
                        var p2 = getNodePos();
                        var name = ts.tokenIsIdentifierOrKeyword(token())
                            ? parseEntityName(/*allowReservedWords*/ true)
                            : undefined;
                        if (name) {
                            while (token() === 80 /* PrivateIdentifier */) {
                                reScanHashToken(); // rescan #id as # id
                                nextTokenJSDoc(); // then skip the #
                                name = finishNode(factory.createJSDocMemberName(name, parseIdentifier()), p2);
                            }
                        }
                        var text = [];
                        while (token() !== 19 /* CloseBraceToken */ && token() !== 4 /* NewLineTrivia */ && token() !== 1 /* EndOfFileToken */) {
                            text.push(scanner.getTokenText());
                            nextTokenJSDoc();
                        }
                        var create = linkType === "link" ? factory.createJSDocLink
                            : linkType === "linkcode" ? factory.createJSDocLinkCode
                                : factory.createJSDocLinkPlain;
                        return finishNode(create(name, text.join("")), start, scanner.getTextPos());
                    }
                    function parseJSDocLinkPrefix() {
                        skipWhitespaceOrAsterisk();
                        if (token() === 18 /* OpenBraceToken */
                            && nextTokenJSDoc() === 59 /* AtToken */
                            && ts.tokenIsIdentifierOrKeyword(nextTokenJSDoc())) {
                            var kind = scanner.getTokenValue();
                            if (kind === "link" || kind === "linkcode" || kind === "linkplain") {
                                return kind;
                            }
                        }
                    }
                    function parseUnknownTag(start, tagName, indent, indentText) {
                        return finishNode(factory.createJSDocUnknownTag(tagName, parseTrailingTagComments(start, getNodePos(), indent, indentText)), start);
                    }
                    function addTag(tag) {
                        if (!tag) {
                            return;
                        }
                        if (!tags) {
                            tags = [tag];
                            tagsPos = tag.pos;
                        }
                        else {
                            tags.push(tag);
                        }
                        tagsEnd = tag.end;
                    }
                    function tryParseTypeExpression() {
                        skipWhitespaceOrAsterisk();
                        return token() === 18 /* OpenBraceToken */ ? parseJSDocTypeExpression() : undefined;
                    }
                    function parseBracketNameInPropertyAndParamTag() {
                        // Looking for something like '[foo]', 'foo', '[foo.bar]' or 'foo.bar'
                        var isBracketed = parseOptionalJsdoc(22 /* OpenBracketToken */);
                        if (isBracketed) {
                            skipWhitespace();
                        }
                        // a markdown-quoted name: `arg` is not legal jsdoc, but occurs in the wild
                        var isBackquoted = parseOptionalJsdoc(61 /* BacktickToken */);
                        var name = parseJSDocEntityName();
                        if (isBackquoted) {
                            parseExpectedTokenJSDoc(61 /* BacktickToken */);
                        }
                        if (isBracketed) {
                            skipWhitespace();
                            // May have an optional default, e.g. '[foo = 42]'
                            if (parseOptionalToken(63 /* EqualsToken */)) {
                                parseExpression();
                            }
                            parseExpected(23 /* CloseBracketToken */);
                        }
                        return { name: name, isBracketed: isBracketed };
                    }
                    function isObjectOrObjectArrayTypeReference(node) {
                        switch (node.kind) {
                            case 146 /* ObjectKeyword */:
                                return true;
                            case 181 /* ArrayType */:
                                return isObjectOrObjectArrayTypeReference(node.elementType);
                            default:
                                return ts.isTypeReferenceNode(node) && ts.isIdentifier(node.typeName) && node.typeName.escapedText === "Object" && !node.typeArguments;
                        }
                    }
                    function parseParameterOrPropertyTag(start, tagName, target, indent) {
                        var typeExpression = tryParseTypeExpression();
                        var isNameFirst = !typeExpression;
                        skipWhitespaceOrAsterisk();
                        var _a = parseBracketNameInPropertyAndParamTag(), name = _a.name, isBracketed = _a.isBracketed;
                        var indentText = skipWhitespaceOrAsterisk();
                        if (isNameFirst && !lookAhead(parseJSDocLinkPrefix)) {
                            typeExpression = tryParseTypeExpression();
                        }
                        var comment = parseTrailingTagComments(start, getNodePos(), indent, indentText);
                        var nestedTypeLiteral = target !== 4 /* CallbackParameter */ && parseNestedTypeLiteral(typeExpression, name, target, indent);
                        if (nestedTypeLiteral) {
                            typeExpression = nestedTypeLiteral;
                            isNameFirst = true;
                        }
                        var result = target === 1 /* Property */
                            ? factory.createJSDocPropertyTag(tagName, name, isBracketed, typeExpression, isNameFirst, comment)
                            : factory.createJSDocParameterTag(tagName, name, isBracketed, typeExpression, isNameFirst, comment);
                        return finishNode(result, start);
                    }
                    function parseNestedTypeLiteral(typeExpression, name, target, indent) {
                        if (typeExpression && isObjectOrObjectArrayTypeReference(typeExpression.type)) {
                            var pos = getNodePos();
                            var child = void 0;
                            var children = void 0;
                            while (child = tryParse(function () { return parseChildParameterOrPropertyTag(target, indent, name); })) {
                                if (child.kind === 335 /* JSDocParameterTag */ || child.kind === 342 /* JSDocPropertyTag */) {
                                    children = ts.append(children, child);
                                }
                            }
                            if (children) {
                                var literal = finishNode(factory.createJSDocTypeLiteral(children, typeExpression.type.kind === 181 /* ArrayType */), pos);
                                return finishNode(factory.createJSDocTypeExpression(literal), pos);
                            }
                        }
                    }
                    function parseReturnTag(start, tagName, indent, indentText) {
                        if (ts.some(tags, ts.isJSDocReturnTag)) {
                            parseErrorAt(tagName.pos, scanner.getTokenPos(), ts.Diagnostics._0_tag_already_specified, tagName.escapedText);
                        }
                        var typeExpression = tryParseTypeExpression();
                        return finishNode(factory.createJSDocReturnTag(tagName, typeExpression, parseTrailingTagComments(start, getNodePos(), indent, indentText)), start);
                    }
                    function parseTypeTag(start, tagName, indent, indentText) {
                        if (ts.some(tags, ts.isJSDocTypeTag)) {
                            parseErrorAt(tagName.pos, scanner.getTokenPos(), ts.Diagnostics._0_tag_already_specified, tagName.escapedText);
                        }
                        var typeExpression = parseJSDocTypeExpression(/*mayOmitBraces*/ true);
                        var comments = indent !== undefined && indentText !== undefined ? parseTrailingTagComments(start, getNodePos(), indent, indentText) : undefined;
                        return finishNode(factory.createJSDocTypeTag(tagName, typeExpression, comments), start);
                    }
                    function parseSeeTag(start, tagName, indent, indentText) {
                        var isLink = lookAhead(function () { return nextTokenJSDoc() === 59 /* AtToken */ && ts.tokenIsIdentifierOrKeyword(nextTokenJSDoc()) && scanner.getTokenValue() === "link"; });
                        var nameExpression = isLink ? undefined : parseJSDocNameReference();
                        var comments = indent !== undefined && indentText !== undefined ? parseTrailingTagComments(start, getNodePos(), indent, indentText) : undefined;
                        return finishNode(factory.createJSDocSeeTag(tagName, nameExpression, comments), start);
                    }
                    function parseAuthorTag(start, tagName, indent, indentText) {
                        var commentStart = getNodePos();
                        var textOnly = parseAuthorNameAndEmail();
                        var commentEnd = scanner.getStartPos();
                        var comments = parseTrailingTagComments(start, commentEnd, indent, indentText);
                        if (!comments) {
                            commentEnd = scanner.getStartPos();
                        }
                        var allParts = typeof comments !== "string"
                            ? createNodeArray(ts.concatenate([finishNode(textOnly, commentStart, commentEnd)], comments), commentStart) // cast away readonly
                            : textOnly.text + comments;
                        return finishNode(factory.createJSDocAuthorTag(tagName, allParts), start);
                    }
                    function parseAuthorNameAndEmail() {
                        var comments = [];
                        var inEmail = false;
                        var token = scanner.getToken();
                        while (token !== 1 /* EndOfFileToken */ && token !== 4 /* NewLineTrivia */) {
                            if (token === 29 /* LessThanToken */) {
                                inEmail = true;
                            }
                            else if (token === 59 /* AtToken */ && !inEmail) {
                                break;
                            }
                            else if (token === 31 /* GreaterThanToken */ && inEmail) {
                                comments.push(scanner.getTokenText());
                                scanner.setTextPos(scanner.getTokenPos() + 1);
                                break;
                            }
                            comments.push(scanner.getTokenText());
                            token = nextTokenJSDoc();
                        }
                        return factory.createJSDocText(comments.join(""));
                    }
                    function parseImplementsTag(start, tagName, margin, indentText) {
                        var className = parseExpressionWithTypeArgumentsForAugments();
                        return finishNode(factory.createJSDocImplementsTag(tagName, className, parseTrailingTagComments(start, getNodePos(), margin, indentText)), start);
                    }
                    function parseAugmentsTag(start, tagName, margin, indentText) {
                        var className = parseExpressionWithTypeArgumentsForAugments();
                        return finishNode(factory.createJSDocAugmentsTag(tagName, className, parseTrailingTagComments(start, getNodePos(), margin, indentText)), start);
                    }
                    function parseExpressionWithTypeArgumentsForAugments() {
                        var usedBrace = parseOptional(18 /* OpenBraceToken */);
                        var pos = getNodePos();
                        var expression = parsePropertyAccessEntityNameExpression();
                        var typeArguments = tryParseTypeArguments();
                        var node = factory.createExpressionWithTypeArguments(expression, typeArguments);
                        var res = finishNode(node, pos);
                        if (usedBrace) {
                            parseExpected(19 /* CloseBraceToken */);
                        }
                        return res;
                    }
                    function parsePropertyAccessEntityNameExpression() {
                        var pos = getNodePos();
                        var node = parseJSDocIdentifierName();
                        while (parseOptional(24 /* DotToken */)) {
                            var name = parseJSDocIdentifierName();
                            node = finishNode(factory.createPropertyAccessExpression(node, name), pos);
                        }
                        return node;
                    }
                    function parseSimpleTag(start, createTag, tagName, margin, indentText) {
                        return finishNode(createTag(tagName, parseTrailingTagComments(start, getNodePos(), margin, indentText)), start);
                    }
                    function parseThisTag(start, tagName, margin, indentText) {
                        var typeExpression = parseJSDocTypeExpression(/*mayOmitBraces*/ true);
                        skipWhitespace();
                        return finishNode(factory.createJSDocThisTag(tagName, typeExpression, parseTrailingTagComments(start, getNodePos(), margin, indentText)), start);
                    }
                    function parseEnumTag(start, tagName, margin, indentText) {
                        var typeExpression = parseJSDocTypeExpression(/*mayOmitBraces*/ true);
                        skipWhitespace();
                        return finishNode(factory.createJSDocEnumTag(tagName, typeExpression, parseTrailingTagComments(start, getNodePos(), margin, indentText)), start);
                    }
                    function parseTypedefTag(start, tagName, indent, indentText) {
                        var _a;
                        var typeExpression = tryParseTypeExpression();
                        skipWhitespaceOrAsterisk();
                        var fullName = parseJSDocTypeNameWithNamespace();
                        skipWhitespace();
                        var comment = parseTagComments(indent);
                        var end;
                        if (!typeExpression || isObjectOrObjectArrayTypeReference(typeExpression.type)) {
                            var child = void 0;
                            var childTypeTag = void 0;
                            var jsDocPropertyTags = void 0;
                            var hasChildren = false;
                            while (child = tryParse(function () { return parseChildPropertyTag(indent); })) {
                                hasChildren = true;
                                if (child.kind === 338 /* JSDocTypeTag */) {
                                    if (childTypeTag) {
                                        parseErrorAtCurrentToken(ts.Diagnostics.A_JSDoc_typedef_comment_may_not_contain_multiple_type_tags);
                                        var lastError = ts.lastOrUndefined(parseDiagnostics);
                                        if (lastError) {
                                            ts.addRelatedInfo(lastError, ts.createDetachedDiagnostic(fileName, 0, 0, ts.Diagnostics.The_tag_was_first_specified_here));
                                        }
                                        break;
                                    }
                                    else {
                                        childTypeTag = child;
                                    }
                                }
                                else {
                                    jsDocPropertyTags = ts.append(jsDocPropertyTags, child);
                                }
                            }
                            if (hasChildren) {
                                var isArrayType = typeExpression && typeExpression.type.kind === 181 /* ArrayType */;
                                var jsdocTypeLiteral = factory.createJSDocTypeLiteral(jsDocPropertyTags, isArrayType);
                                typeExpression = childTypeTag && childTypeTag.typeExpression && !isObjectOrObjectArrayTypeReference(childTypeTag.typeExpression.type) ?
                                    childTypeTag.typeExpression :
                                    finishNode(jsdocTypeLiteral, start);
                                end = typeExpression.end;
                            }
                        }
                        // Only include the characters between the name end and the next token if a comment was actually parsed out - otherwise it's just whitespace
                        end = end || comment !== undefined ?
                            getNodePos() :
                            ((_a = fullName !== null && fullName !== void 0 ? fullName : typeExpression) !== null && _a !== void 0 ? _a : tagName).end;
                        if (!comment) {
                            comment = parseTrailingTagComments(start, end, indent, indentText);
                        }
                        var typedefTag = factory.createJSDocTypedefTag(tagName, typeExpression, fullName, comment);
                        return finishNode(typedefTag, start, end);
                    }
                    function parseJSDocTypeNameWithNamespace(nested) {
                        var pos = scanner.getTokenPos();
                        if (!ts.tokenIsIdentifierOrKeyword(token())) {
                            return undefined;
                        }
                        var typeNameOrNamespaceName = parseJSDocIdentifierName();
                        if (parseOptional(24 /* DotToken */)) {
                            var body = parseJSDocTypeNameWithNamespace(/*nested*/ true);
                            var jsDocNamespaceNode = factory.createModuleDeclaration(
                            /*decorators*/ undefined, 
                            /*modifiers*/ undefined, typeNameOrNamespaceName, body, nested ? 4 /* NestedNamespace */ : undefined);
                            return finishNode(jsDocNamespaceNode, pos);
                        }
                        if (nested) {
                            typeNameOrNamespaceName.isInJSDocNamespace = true;
                        }
                        return typeNameOrNamespaceName;
                    }
                    function parseCallbackTagParameters(indent) {
                        var pos = getNodePos();
                        var child;
                        var parameters;
                        while (child = tryParse(function () { return parseChildParameterOrPropertyTag(4 /* CallbackParameter */, indent); })) {
                            parameters = ts.append(parameters, child);
                        }
                        return createNodeArray(parameters || [], pos);
                    }
                    function parseCallbackTag(start, tagName, indent, indentText) {
                        var fullName = parseJSDocTypeNameWithNamespace();
                        skipWhitespace();
                        var comment = parseTagComments(indent);
                        var parameters = parseCallbackTagParameters(indent);
                        var returnTag = tryParse(function () {
                            if (parseOptionalJsdoc(59 /* AtToken */)) {
                                var tag = parseTag(indent);
                                if (tag && tag.kind === 336 /* JSDocReturnTag */) {
                                    return tag;
                                }
                            }
                        });
                        var typeExpression = finishNode(factory.createJSDocSignature(/*typeParameters*/ undefined, parameters, returnTag), start);
                        if (!comment) {
                            comment = parseTrailingTagComments(start, getNodePos(), indent, indentText);
                        }
                        return finishNode(factory.createJSDocCallbackTag(tagName, typeExpression, fullName, comment), start);
                    }
                    function escapedTextsEqual(a, b) {
                        while (!ts.isIdentifier(a) || !ts.isIdentifier(b)) {
                            if (!ts.isIdentifier(a) && !ts.isIdentifier(b) && a.right.escapedText === b.right.escapedText) {
                                a = a.left;
                                b = b.left;
                            }
                            else {
                                return false;
                            }
                        }
                        return a.escapedText === b.escapedText;
                    }
                    function parseChildPropertyTag(indent) {
                        return parseChildParameterOrPropertyTag(1 /* Property */, indent);
                    }
                    function parseChildParameterOrPropertyTag(target, indent, name) {
                        var canParseTag = true;
                        var seenAsterisk = false;
                        while (true) {
                            switch (nextTokenJSDoc()) {
                                case 59 /* AtToken */:
                                    if (canParseTag) {
                                        var child = tryParseChildTag(target, indent);
                                        if (child && (child.kind === 335 /* JSDocParameterTag */ || child.kind === 342 /* JSDocPropertyTag */) &&
                                            target !== 4 /* CallbackParameter */ &&
                                            name && (ts.isIdentifier(child.name) || !escapedTextsEqual(name, child.name.left))) {
                                            return false;
                                        }
                                        return child;
                                    }
                                    seenAsterisk = false;
                                    break;
                                case 4 /* NewLineTrivia */:
                                    canParseTag = true;
                                    seenAsterisk = false;
                                    break;
                                case 41 /* AsteriskToken */:
                                    if (seenAsterisk) {
                                        canParseTag = false;
                                    }
                                    seenAsterisk = true;
                                    break;
                                case 79 /* Identifier */:
                                    canParseTag = false;
                                    break;
                                case 1 /* EndOfFileToken */:
                                    return false;
                            }
                        }
                    }
                    function tryParseChildTag(target, indent) {
                        ts.Debug.assert(token() === 59 /* AtToken */);
                        var start = scanner.getStartPos();
                        nextTokenJSDoc();
                        var tagName = parseJSDocIdentifierName();
                        skipWhitespace();
                        var t;
                        switch (tagName.escapedText) {
                            case "type":
                                return target === 1 /* Property */ && parseTypeTag(start, tagName);
                            case "prop":
                            case "property":
                                t = 1 /* Property */;
                                break;
                            case "arg":
                            case "argument":
                            case "param":
                                t = 2 /* Parameter */ | 4 /* CallbackParameter */;
                                break;
                            default:
                                return false;
                        }
                        if (!(target & t)) {
                            return false;
                        }
                        return parseParameterOrPropertyTag(start, tagName, target, indent);
                    }
                    function parseTemplateTagTypeParameter() {
                        var typeParameterPos = getNodePos();
                        var name = parseJSDocIdentifierName(ts.Diagnostics.Unexpected_token_A_type_parameter_name_was_expected_without_curly_braces);
                        if (ts.nodeIsMissing(name)) {
                            return undefined;
                        }
                        return finishNode(factory.createTypeParameterDeclaration(name, /*constraint*/ undefined, /*defaultType*/ undefined), typeParameterPos);
                    }
                    function parseTemplateTagTypeParameters() {
                        var pos = getNodePos();
                        var typeParameters = [];
                        do {
                            skipWhitespace();
                            var node = parseTemplateTagTypeParameter();
                            if (node !== undefined) {
                                typeParameters.push(node);
                            }
                            skipWhitespaceOrAsterisk();
                        } while (parseOptionalJsdoc(27 /* CommaToken */));
                        return createNodeArray(typeParameters, pos);
                    }
                    function parseTemplateTag(start, tagName, indent, indentText) {
                        // The template tag looks like one of the following:
                        //   @template T,U,V
                        //   @template {Constraint} T
                        //
                        // According to the [closure docs](https://github.com/google/closure-compiler/wiki/Generic-Types#multiple-bounded-template-types):
                        //   > Multiple bounded generics cannot be declared on the same line. For the sake of clarity, if multiple templates share the same
                        //   > type bound they must be declared on separate lines.
                        //
                        // TODO: Determine whether we should enforce this in the checker.
                        // TODO: Consider moving the `constraint` to the first type parameter as we could then remove `getEffectiveConstraintOfTypeParameter`.
                        // TODO: Consider only parsing a single type parameter if there is a constraint.
                        var constraint = token() === 18 /* OpenBraceToken */ ? parseJSDocTypeExpression() : undefined;
                        var typeParameters = parseTemplateTagTypeParameters();
                        return finishNode(factory.createJSDocTemplateTag(tagName, constraint, typeParameters, parseTrailingTagComments(start, getNodePos(), indent, indentText)), start);
                    }
                    function parseOptionalJsdoc(t) {
                        if (token() === t) {
                            nextTokenJSDoc();
                            return true;
                        }
                        return false;
                    }
                    function parseJSDocEntityName() {
                        var entity = parseJSDocIdentifierName();
                        if (parseOptional(22 /* OpenBracketToken */)) {
                            parseExpected(23 /* CloseBracketToken */);
                            // Note that y[] is accepted as an entity name, but the postfix brackets are not saved for checking.
                            // Technically usejsdoc.org requires them for specifying a property of a type equivalent to Array<{ x: ...}>
                            // but it's not worth it to enforce that restriction.
                        }
                        while (parseOptional(24 /* DotToken */)) {
                            var name = parseJSDocIdentifierName();
                            if (parseOptional(22 /* OpenBracketToken */)) {
                                parseExpected(23 /* CloseBracketToken */);
                            }
                            entity = createQualifiedName(entity, name);
                        }
                        return entity;
                    }
                    function parseJSDocIdentifierName(message) {
                        if (!ts.tokenIsIdentifierOrKeyword(token())) {
                            return createMissingNode(79 /* Identifier */, /*reportAtCurrentPosition*/ !message, message || ts.Diagnostics.Identifier_expected);
                        }
                        identifierCount++;
                        var pos = scanner.getTokenPos();
                        var end = scanner.getTextPos();
                        var originalKeywordKind = token();
                        var text = internIdentifier(scanner.getTokenValue());
                        var result = finishNode(factory.createIdentifier(text, /*typeArguments*/ undefined, originalKeywordKind), pos, end);
                        nextTokenJSDoc();
                        return result;
                    }
                }
            })(JSDocParser = Parser.JSDocParser || (Parser.JSDocParser = {}));
        })(Parser || (Parser = {}));
        var IncrementalParser;
        (function (IncrementalParser) {
            function updateSourceFile(sourceFile, newText, textChangeRange, aggressiveChecks) {
                aggressiveChecks = aggressiveChecks || ts.Debug.shouldAssert(2 /* Aggressive */);
                checkChangeRange(sourceFile, newText, textChangeRange, aggressiveChecks);
                if (ts.textChangeRangeIsUnchanged(textChangeRange)) {
                    // if the text didn't change, then we can just return our current source file as-is.
                    return sourceFile;
                }
                if (sourceFile.statements.length === 0) {
                    // If we don't have any statements in the current source file, then there's no real
                    // way to incrementally parse.  So just do a full parse instead.
                    return Parser.parseSourceFile(sourceFile.fileName, newText, sourceFile.languageVersion, /*syntaxCursor*/ undefined, /*setParentNodes*/ true, sourceFile.scriptKind);
                }
                // Make sure we're not trying to incrementally update a source file more than once.  Once
                // we do an update the original source file is considered unusable from that point onwards.
                //
                // This is because we do incremental parsing in-place.  i.e. we take nodes from the old
                // tree and give them new positions and parents.  From that point on, trusting the old
                // tree at all is not possible as far too much of it may violate invariants.
                var incrementalSourceFile = sourceFile;
                ts.Debug.assert(!incrementalSourceFile.hasBeenIncrementallyParsed);
                incrementalSourceFile.hasBeenIncrementallyParsed = true;
                Parser.fixupParentReferences(incrementalSourceFile);
                var oldText = sourceFile.text;
                var syntaxCursor = createSyntaxCursor(sourceFile);
                // Make the actual change larger so that we know to reparse anything whose lookahead
                // might have intersected the change.
                var changeRange = extendToAffectedRange(sourceFile, textChangeRange);
                checkChangeRange(sourceFile, newText, changeRange, aggressiveChecks);
                // Ensure that extending the affected range only moved the start of the change range
                // earlier in the file.
                ts.Debug.assert(changeRange.span.start <= textChangeRange.span.start);
                ts.Debug.assert(ts.textSpanEnd(changeRange.span) === ts.textSpanEnd(textChangeRange.span));
                ts.Debug.assert(ts.textSpanEnd(ts.textChangeRangeNewSpan(changeRange)) === ts.textSpanEnd(ts.textChangeRangeNewSpan(textChangeRange)));
                // The is the amount the nodes after the edit range need to be adjusted.  It can be
                // positive (if the edit added characters), negative (if the edit deleted characters)
                // or zero (if this was a pure overwrite with nothing added/removed).
                var delta = ts.textChangeRangeNewSpan(changeRange).length - changeRange.span.length;
                // If we added or removed characters during the edit, then we need to go and adjust all
                // the nodes after the edit.  Those nodes may move forward (if we inserted chars) or they
                // may move backward (if we deleted chars).
                //
                // Doing this helps us out in two ways.  First, it means that any nodes/tokens we want
                // to reuse are already at the appropriate position in the new text.  That way when we
                // reuse them, we don't have to figure out if they need to be adjusted.  Second, it makes
                // it very easy to determine if we can reuse a node.  If the node's position is at where
                // we are in the text, then we can reuse it.  Otherwise we can't.  If the node's position
                // is ahead of us, then we'll need to rescan tokens.  If the node's position is behind
                // us, then we'll need to skip it or crumble it as appropriate
                //
                // We will also adjust the positions of nodes that intersect the change range as well.
                // By doing this, we ensure that all the positions in the old tree are consistent, not
                // just the positions of nodes entirely before/after the change range.  By being
                // consistent, we can then easily map from positions to nodes in the old tree easily.
                //
                // Also, mark any syntax elements that intersect the changed span.  We know, up front,
                // that we cannot reuse these elements.
                updateTokenPositionsAndMarkElements(incrementalSourceFile, changeRange.span.start, ts.textSpanEnd(changeRange.span), ts.textSpanEnd(ts.textChangeRangeNewSpan(changeRange)), delta, oldText, newText, aggressiveChecks);
                // Now that we've set up our internal incremental state just proceed and parse the
                // source file in the normal fashion.  When possible the parser will retrieve and
                // reuse nodes from the old tree.
                //
                // Note: passing in 'true' for setNodeParents is very important.  When incrementally
                // parsing, we will be reusing nodes from the old tree, and placing it into new
                // parents.  If we don't set the parents now, we'll end up with an observably
                // inconsistent tree.  Setting the parents on the new tree should be very fast.  We
                // will immediately bail out of walking any subtrees when we can see that their parents
                // are already correct.
                var result = Parser.parseSourceFile(sourceFile.fileName, newText, sourceFile.languageVersion, syntaxCursor, /*setParentNodes*/ true, sourceFile.scriptKind);
                result.commentDirectives = getNewCommentDirectives(sourceFile.commentDirectives, result.commentDirectives, changeRange.span.start, ts.textSpanEnd(changeRange.span), delta, oldText, newText, aggressiveChecks);
                return result;
            }
            IncrementalParser.updateSourceFile = updateSourceFile;
            function getNewCommentDirectives(oldDirectives, newDirectives, changeStart, changeRangeOldEnd, delta, oldText, newText, aggressiveChecks) {
                if (!oldDirectives)
                    return newDirectives;
                var commentDirectives;
                var addedNewlyScannedDirectives = false;
                for (var _i = 0, oldDirectives_1 = oldDirectives; _i < oldDirectives_1.length; _i++) {
                    var directive = oldDirectives_1[_i];
                    var range = directive.range, type = directive.type;
                    // Range before the change
                    if (range.end < changeStart) {
                        commentDirectives = ts.append(commentDirectives, directive);
                    }
                    else if (range.pos > changeRangeOldEnd) {
                        addNewlyScannedDirectives();
                        // Node is entirely past the change range.  We need to move both its pos and
                        // end, forward or backward appropriately.
                        var updatedDirective = {
                            range: { pos: range.pos + delta, end: range.end + delta },
                            type: type
                        };
                        commentDirectives = ts.append(commentDirectives, updatedDirective);
                        if (aggressiveChecks) {
                            ts.Debug.assert(oldText.substring(range.pos, range.end) === newText.substring(updatedDirective.range.pos, updatedDirective.range.end));
                        }
                    }
                    // Ignore ranges that fall in change range
                }
                addNewlyScannedDirectives();
                return commentDirectives;
                function addNewlyScannedDirectives() {
                    if (addedNewlyScannedDirectives)
                        return;
                    addedNewlyScannedDirectives = true;
                    if (!commentDirectives) {
                        commentDirectives = newDirectives;
                    }
                    else if (newDirectives) {
                        commentDirectives.push.apply(commentDirectives, newDirectives);
                    }
                }
            }
            function moveElementEntirelyPastChangeRange(element, isArray, delta, oldText, newText, aggressiveChecks) {
                if (isArray) {
                    visitArray(element);
                }
                else {
                    visitNode(element);
                }
                return;
                function visitNode(node) {
                    var text = "";
                    if (aggressiveChecks && shouldCheckNode(node)) {
                        text = oldText.substring(node.pos, node.end);
                    }
                    // Ditch any existing LS children we may have created.  This way we can avoid
                    // moving them forward.
                    if (node._children) {
                        node._children = undefined;
                    }
                    ts.setTextRangePosEnd(node, node.pos + delta, node.end + delta);
                    if (aggressiveChecks && shouldCheckNode(node)) {
                        ts.Debug.assert(text === newText.substring(node.pos, node.end));
                    }
                    forEachChild(node, visitNode, visitArray);
                    if (ts.hasJSDocNodes(node)) {
                        for (var _i = 0, _a = node.jsDoc; _i < _a.length; _i++) {
                            var jsDocComment = _a[_i];
                            visitNode(jsDocComment);
                        }
                    }
                    checkNodePositions(node, aggressiveChecks);
                }
                function visitArray(array) {
                    array._children = undefined;
                    ts.setTextRangePosEnd(array, array.pos + delta, array.end + delta);
                    for (var _i = 0, array_9 = array; _i < array_9.length; _i++) {
                        var node = array_9[_i];
                        visitNode(node);
                    }
                }
            }
            function shouldCheckNode(node) {
                switch (node.kind) {
                    case 10 /* StringLiteral */:
                    case 8 /* NumericLiteral */:
                    case 79 /* Identifier */:
                        return true;
                }
                return false;
            }
            function adjustIntersectingElement(element, changeStart, changeRangeOldEnd, changeRangeNewEnd, delta) {
                ts.Debug.assert(element.end >= changeStart, "Adjusting an element that was entirely before the change range");
                ts.Debug.assert(element.pos <= changeRangeOldEnd, "Adjusting an element that was entirely after the change range");
                ts.Debug.assert(element.pos <= element.end);
                // We have an element that intersects the change range in some way.  It may have its
                // start, or its end (or both) in the changed range.  We want to adjust any part
                // that intersects such that the final tree is in a consistent state.  i.e. all
                // children have spans within the span of their parent, and all siblings are ordered
                // properly.
                // We may need to update both the 'pos' and the 'end' of the element.
                // If the 'pos' is before the start of the change, then we don't need to touch it.
                // If it isn't, then the 'pos' must be inside the change.  How we update it will
                // depend if delta is positive or negative. If delta is positive then we have
                // something like:
                //
                //  -------------------AAA-----------------
                //  -------------------BBBCCCCCCC-----------------
                //
                // In this case, we consider any node that started in the change range to still be
                // starting at the same position.
                //
                // however, if the delta is negative, then we instead have something like this:
                //
                //  -------------------XXXYYYYYYY-----------------
                //  -------------------ZZZ-----------------
                //
                // In this case, any element that started in the 'X' range will keep its position.
                // However any element that started after that will have their pos adjusted to be
                // at the end of the new range.  i.e. any node that started in the 'Y' range will
                // be adjusted to have their start at the end of the 'Z' range.
                //
                // The element will keep its position if possible.  Or Move backward to the new-end
                // if it's in the 'Y' range.
                var pos = Math.min(element.pos, changeRangeNewEnd);
                // If the 'end' is after the change range, then we always adjust it by the delta
                // amount.  However, if the end is in the change range, then how we adjust it
                // will depend on if delta is positive or negative.  If delta is positive then we
                // have something like:
                //
                //  -------------------AAA-----------------
                //  -------------------BBBCCCCCCC-----------------
                //
                // In this case, we consider any node that ended inside the change range to keep its
                // end position.
                //
                // however, if the delta is negative, then we instead have something like this:
                //
                //  -------------------XXXYYYYYYY-----------------
                //  -------------------ZZZ-----------------
                //
                // In this case, any element that ended in the 'X' range will keep its position.
                // However any element that ended after that will have their pos adjusted to be
                // at the end of the new range.  i.e. any node that ended in the 'Y' range will
                // be adjusted to have their end at the end of the 'Z' range.
                var end = element.end >= changeRangeOldEnd ?
                    // Element ends after the change range.  Always adjust the end pos.
                    element.end + delta :
                    // Element ends in the change range.  The element will keep its position if
                    // possible. Or Move backward to the new-end if it's in the 'Y' range.
                    Math.min(element.end, changeRangeNewEnd);
                ts.Debug.assert(pos <= end);
                if (element.parent) {
                    ts.Debug.assertGreaterThanOrEqual(pos, element.parent.pos);
                    ts.Debug.assertLessThanOrEqual(end, element.parent.end);
                }
                ts.setTextRangePosEnd(element, pos, end);
            }
            function checkNodePositions(node, aggressiveChecks) {
                if (aggressiveChecks) {
                    var pos_2 = node.pos;
                    var visitNode_1 = function (child) {
                        ts.Debug.assert(child.pos >= pos_2);
                        pos_2 = child.end;
                    };
                    if (ts.hasJSDocNodes(node)) {
                        for (var _i = 0, _a = node.jsDoc; _i < _a.length; _i++) {
                            var jsDocComment = _a[_i];
                            visitNode_1(jsDocComment);
                        }
                    }
                    forEachChild(node, visitNode_1);
                    ts.Debug.assert(pos_2 <= node.end);
                }
            }
            function updateTokenPositionsAndMarkElements(sourceFile, changeStart, changeRangeOldEnd, changeRangeNewEnd, delta, oldText, newText, aggressiveChecks) {
                visitNode(sourceFile);
                return;
                function visitNode(child) {
                    ts.Debug.assert(child.pos <= child.end);
                    if (child.pos > changeRangeOldEnd) {
                        // Node is entirely past the change range.  We need to move both its pos and
                        // end, forward or backward appropriately.
                        moveElementEntirelyPastChangeRange(child, /*isArray*/ false, delta, oldText, newText, aggressiveChecks);
                        return;
                    }
                    // Check if the element intersects the change range.  If it does, then it is not
                    // reusable.  Also, we'll need to recurse to see what constituent portions we may
                    // be able to use.
                    var fullEnd = child.end;
                    if (fullEnd >= changeStart) {
                        child.intersectsChange = true;
                        child._children = undefined;
                        // Adjust the pos or end (or both) of the intersecting element accordingly.
                        adjustIntersectingElement(child, changeStart, changeRangeOldEnd, changeRangeNewEnd, delta);
                        forEachChild(child, visitNode, visitArray);
                        if (ts.hasJSDocNodes(child)) {
                            for (var _i = 0, _a = child.jsDoc; _i < _a.length; _i++) {
                                var jsDocComment = _a[_i];
                                visitNode(jsDocComment);
                            }
                        }
                        checkNodePositions(child, aggressiveChecks);
                        return;
                    }
                    // Otherwise, the node is entirely before the change range.  No need to do anything with it.
                    ts.Debug.assert(fullEnd < changeStart);
                }
                function visitArray(array) {
                    ts.Debug.assert(array.pos <= array.end);
                    if (array.pos > changeRangeOldEnd) {
                        // Array is entirely after the change range.  We need to move it, and move any of
                        // its children.
                        moveElementEntirelyPastChangeRange(array, /*isArray*/ true, delta, oldText, newText, aggressiveChecks);
                        return;
                    }
                    // Check if the element intersects the change range.  If it does, then it is not
                    // reusable.  Also, we'll need to recurse to see what constituent portions we may
                    // be able to use.
                    var fullEnd = array.end;
                    if (fullEnd >= changeStart) {
                        array.intersectsChange = true;
                        array._children = undefined;
                        // Adjust the pos or end (or both) of the intersecting array accordingly.
                        adjustIntersectingElement(array, changeStart, changeRangeOldEnd, changeRangeNewEnd, delta);
                        for (var _i = 0, array_10 = array; _i < array_10.length; _i++) {
                            var node = array_10[_i];
                            visitNode(node);
                        }
                        return;
                    }
                    // Otherwise, the array is entirely before the change range.  No need to do anything with it.
                    ts.Debug.assert(fullEnd < changeStart);
                }
            }
            function extendToAffectedRange(sourceFile, changeRange) {
                // Consider the following code:
                //      void foo() { /; }
                //
                // If the text changes with an insertion of / just before the semicolon then we end up with:
                //      void foo() { //; }
                //
                // If we were to just use the changeRange a is, then we would not rescan the { token
                // (as it does not intersect the actual original change range).  Because an edit may
                // change the token touching it, we actually need to look back *at least* one token so
                // that the prior token sees that change.
                var maxLookahead = 1;
                var start = changeRange.span.start;
                // the first iteration aligns us with the change start. subsequent iteration move us to
                // the left by maxLookahead tokens.  We only need to do this as long as we're not at the
                // start of the tree.
                for (var i = 0; start > 0 && i <= maxLookahead; i++) {
                    var nearestNode = findNearestNodeStartingBeforeOrAtPosition(sourceFile, start);
                    ts.Debug.assert(nearestNode.pos <= start);
                    var position = nearestNode.pos;
                    start = Math.max(0, position - 1);
                }
                var finalSpan = ts.createTextSpanFromBounds(start, ts.textSpanEnd(changeRange.span));
                var finalLength = changeRange.newLength + (changeRange.span.start - start);
                return ts.createTextChangeRange(finalSpan, finalLength);
            }
            function findNearestNodeStartingBeforeOrAtPosition(sourceFile, position) {
                var bestResult = sourceFile;
                var lastNodeEntirelyBeforePosition;
                forEachChild(sourceFile, visit);
                if (lastNodeEntirelyBeforePosition) {
                    var lastChildOfLastEntireNodeBeforePosition = getLastDescendant(lastNodeEntirelyBeforePosition);
                    if (lastChildOfLastEntireNodeBeforePosition.pos > bestResult.pos) {
                        bestResult = lastChildOfLastEntireNodeBeforePosition;
                    }
                }
                return bestResult;
                function getLastDescendant(node) {
                    while (true) {
                        var lastChild = ts.getLastChild(node);
                        if (lastChild) {
                            node = lastChild;
                        }
                        else {
                            return node;
                        }
                    }
                }
                function visit(child) {
                    if (ts.nodeIsMissing(child)) {
                        // Missing nodes are effectively invisible to us.  We never even consider them
                        // When trying to find the nearest node before us.
                        return;
                    }
                    // If the child intersects this position, then this node is currently the nearest
                    // node that starts before the position.
                    if (child.pos <= position) {
                        if (child.pos >= bestResult.pos) {
                            // This node starts before the position, and is closer to the position than
                            // the previous best node we found.  It is now the new best node.
                            bestResult = child;
                        }
                        // Now, the node may overlap the position, or it may end entirely before the
                        // position.  If it overlaps with the position, then either it, or one of its
                        // children must be the nearest node before the position.  So we can just
                        // recurse into this child to see if we can find something better.
                        if (position < child.end) {
                            // The nearest node is either this child, or one of the children inside
                            // of it.  We've already marked this child as the best so far.  Recurse
                            // in case one of the children is better.
                            forEachChild(child, visit);
                            // Once we look at the children of this node, then there's no need to
                            // continue any further.
                            return true;
                        }
                        else {
                            ts.Debug.assert(child.end <= position);
                            // The child ends entirely before this position.  Say you have the following
                            // (where $ is the position)
                            //
                            //      <complex expr 1> ? <complex expr 2> $ : <...> <...>
                            //
                            // We would want to find the nearest preceding node in "complex expr 2".
                            // To support that, we keep track of this node, and once we're done searching
                            // for a best node, we recurse down this node to see if we can find a good
                            // result in it.
                            //
                            // This approach allows us to quickly skip over nodes that are entirely
                            // before the position, while still allowing us to find any nodes in the
                            // last one that might be what we want.
                            lastNodeEntirelyBeforePosition = child;
                        }
                    }
                    else {
                        ts.Debug.assert(child.pos > position);
                        // We're now at a node that is entirely past the position we're searching for.
                        // This node (and all following nodes) could never contribute to the result,
                        // so just skip them by returning 'true' here.
                        return true;
                    }
                }
            }
            function checkChangeRange(sourceFile, newText, textChangeRange, aggressiveChecks) {
                var oldText = sourceFile.text;
                if (textChangeRange) {
                    ts.Debug.assert((oldText.length - textChangeRange.span.length + textChangeRange.newLength) === newText.length);
                    if (aggressiveChecks || ts.Debug.shouldAssert(3 /* VeryAggressive */)) {
                        var oldTextPrefix = oldText.substr(0, textChangeRange.span.start);
                        var newTextPrefix = newText.substr(0, textChangeRange.span.start);
                        ts.Debug.assert(oldTextPrefix === newTextPrefix);
                        var oldTextSuffix = oldText.substring(ts.textSpanEnd(textChangeRange.span), oldText.length);
                        var newTextSuffix = newText.substring(ts.textSpanEnd(ts.textChangeRangeNewSpan(textChangeRange)), newText.length);
                        ts.Debug.assert(oldTextSuffix === newTextSuffix);
                    }
                }
            }
            function createSyntaxCursor(sourceFile) {
                var currentArray = sourceFile.statements;
                var currentArrayIndex = 0;
                ts.Debug.assert(currentArrayIndex < currentArray.length);
                var current = currentArray[currentArrayIndex];
                var lastQueriedPosition = -1 /* Value */;
                return {
                    currentNode: function (position) {
                        // Only compute the current node if the position is different than the last time
                        // we were asked.  The parser commonly asks for the node at the same position
                        // twice.  Once to know if can read an appropriate list element at a certain point,
                        // and then to actually read and consume the node.
                        if (position !== lastQueriedPosition) {
                            // Much of the time the parser will need the very next node in the array that
                            // we just returned a node from.So just simply check for that case and move
                            // forward in the array instead of searching for the node again.
                            if (current && current.end === position && currentArrayIndex < (currentArray.length - 1)) {
                                currentArrayIndex++;
                                current = currentArray[currentArrayIndex];
                            }
                            // If we don't have a node, or the node we have isn't in the right position,
                            // then try to find a viable node at the position requested.
                            if (!current || current.pos !== position) {
                                findHighestListElementThatStartsAtPosition(position);
                            }
                        }
                        // Cache this query so that we don't do any extra work if the parser calls back
                        // into us.  Note: this is very common as the parser will make pairs of calls like
                        // 'isListElement -> parseListElement'.  If we were unable to find a node when
                        // called with 'isListElement', we don't want to redo the work when parseListElement
                        // is called immediately after.
                        lastQueriedPosition = position;
                        // Either we don'd have a node, or we have a node at the position being asked for.
                        ts.Debug.assert(!current || current.pos === position);
                        return current;
                    }
                };
                // Finds the highest element in the tree we can find that starts at the provided position.
                // The element must be a direct child of some node list in the tree.  This way after we
                // return it, we can easily return its next sibling in the list.
                function findHighestListElementThatStartsAtPosition(position) {
                    // Clear out any cached state about the last node we found.
                    currentArray = undefined;
                    currentArrayIndex = -1 /* Value */;
                    current = undefined;
                    // Recurse into the source file to find the highest node at this position.
                    forEachChild(sourceFile, visitNode, visitArray);
                    return;
                    function visitNode(node) {
                        if (position >= node.pos && position < node.end) {
                            // Position was within this node.  Keep searching deeper to find the node.
                            forEachChild(node, visitNode, visitArray);
                            // don't proceed any further in the search.
                            return true;
                        }
                        // position wasn't in this node, have to keep searching.
                        return false;
                    }
                    function visitArray(array) {
                        if (position >= array.pos && position < array.end) {
                            // position was in this array.  Search through this array to see if we find a
                            // viable element.
                            for (var i = 0; i < array.length; i++) {
                                var child = array[i];
                                if (child) {
                                    if (child.pos === position) {
                                        // Found the right node.  We're done.
                                        currentArray = array;
                                        currentArrayIndex = i;
                                        current = child;
                                        return true;
                                    }
                                    else {
                                        if (child.pos < position && position < child.end) {
                                            // Position in somewhere within this child.  Search in it and
                                            // stop searching in this array.
                                            forEachChild(child, visitNode, visitArray);
                                            return true;
                                        }
                                    }
                                }
                            }
                        }
                        // position wasn't in this array, have to keep searching.
                        return false;
                    }
                }
            }
            IncrementalParser.createSyntaxCursor = createSyntaxCursor;
            var InvalidPosition;
            (function (InvalidPosition) {
                InvalidPosition[InvalidPosition["Value"] = -1] = "Value";
            })(InvalidPosition || (InvalidPosition = {}));
        })(IncrementalParser || (IncrementalParser = {}));
        /** @internal */
        function isDeclarationFileName(fileName) {
            return ts.fileExtensionIs(fileName, ".d.ts" /* Dts */);
        }
        ts.isDeclarationFileName = isDeclarationFileName;
        /*@internal*/
        function processCommentPragmas(context, sourceText) {
            var pragmas = [];
            for (var _i = 0, _a = ts.getLeadingCommentRanges(sourceText, 0) || ts.emptyArray; _i < _a.length; _i++) {
                var range = _a[_i];
                var comment = sourceText.substring(range.pos, range.end);
                extractPragmas(pragmas, range, comment);
            }
            context.pragmas = new ts.Map();
            for (var _b = 0, pragmas_1 = pragmas; _b < pragmas_1.length; _b++) {
                var pragma = pragmas_1[_b];
                if (context.pragmas.has(pragma.name)) {
                    var currentValue = context.pragmas.get(pragma.name);
                    if (currentValue instanceof Array) {
                        currentValue.push(pragma.args);
                    }
                    else {
                        context.pragmas.set(pragma.name, [currentValue, pragma.args]);
                    }
                    continue;
                }
                context.pragmas.set(pragma.name, pragma.args);
            }
        }
        ts.processCommentPragmas = processCommentPragmas;
        /*@internal*/
        function processPragmasIntoFields(context, reportDiagnostic) {
            context.checkJsDirective = undefined;
            context.referencedFiles = [];
            context.typeReferenceDirectives = [];
            context.libReferenceDirectives = [];
            context.amdDependencies = [];
            context.hasNoDefaultLib = false;
            context.pragmas.forEach(function (entryOrList, key) {
                // TODO: The below should be strongly type-guarded and not need casts/explicit annotations, since entryOrList is related to
                // key and key is constrained to a union; but it's not (see GH#21483 for at least partial fix) :(
                switch (key) {
                    case "reference": {
                        var referencedFiles_1 = context.referencedFiles;
                        var typeReferenceDirectives_1 = context.typeReferenceDirectives;
                        var libReferenceDirectives_1 = context.libReferenceDirectives;
                        ts.forEach(ts.toArray(entryOrList), function (arg) {
                            var _a = arg.arguments, types = _a.types, lib = _a.lib, path = _a.path;
                            if (arg.arguments["no-default-lib"]) {
                                context.hasNoDefaultLib = true;
                            }
                            else if (types) {
                                typeReferenceDirectives_1.push({ pos: types.pos, end: types.end, fileName: types.value });
                            }
                            else if (lib) {
                                libReferenceDirectives_1.push({ pos: lib.pos, end: lib.end, fileName: lib.value });
                            }
                            else if (path) {
                                referencedFiles_1.push({ pos: path.pos, end: path.end, fileName: path.value });
                            }
                            else {
                                reportDiagnostic(arg.range.pos, arg.range.end - arg.range.pos, ts.Diagnostics.Invalid_reference_directive_syntax);
                            }
                        });
                        break;
                    }
                    case "amd-dependency": {
                        context.amdDependencies = ts.map(ts.toArray(entryOrList), function (x) { return ({ name: x.arguments.name, path: x.arguments.path }); });
                        break;
                    }
                    case "amd-module": {
                        if (entryOrList instanceof Array) {
                            for (var _i = 0, entryOrList_1 = entryOrList; _i < entryOrList_1.length; _i++) {
                                var entry = entryOrList_1[_i];
                                if (context.moduleName) {
                                    // TODO: It's probably fine to issue this diagnostic on all instances of the pragma
                                    reportDiagnostic(entry.range.pos, entry.range.end - entry.range.pos, ts.Diagnostics.An_AMD_module_cannot_have_multiple_name_assignments);
                                }
                                context.moduleName = entry.arguments.name;
                            }
                        }
                        else {
                            context.moduleName = entryOrList.arguments.name;
                        }
                        break;
                    }
                    case "ts-nocheck":
                    case "ts-check": {
                        // _last_ of either nocheck or check in a file is the "winner"
                        ts.forEach(ts.toArray(entryOrList), function (entry) {
                            if (!context.checkJsDirective || entry.range.pos > context.checkJsDirective.pos) {
                                context.checkJsDirective = {
                                    enabled: key === "ts-check",
                                    end: entry.range.end,
                                    pos: entry.range.pos
                                };
                            }
                        });
                        break;
                    }
                    case "jsx":
                    case "jsxfrag":
                    case "jsximportsource":
                    case "jsxruntime":
                        return; // Accessed directly
                    default: ts.Debug.fail("Unhandled pragma kind"); // Can this be made into an assertNever in the future?
                }
            });
        }
        ts.processPragmasIntoFields = processPragmasIntoFields;
        var namedArgRegExCache = new ts.Map();
        function getNamedArgRegEx(name) {
            if (namedArgRegExCache.has(name)) {
                return namedArgRegExCache.get(name);
            }
            var result = new RegExp("(\\s" + name + "\\s*=\\s*)(?:(?:'([^']*)')|(?:\"([^\"]*)\"))", "im");
            namedArgRegExCache.set(name, result);
            return result;
        }
        var tripleSlashXMLCommentStartRegEx = /^\/\/\/\s*<(\S+)\s.*?\/>/im;
        var singleLinePragmaRegEx = /^\/\/\/?\s*@(\S+)\s*(.*)\s*$/im;
        function extractPragmas(pragmas, range, text) {
            var tripleSlash = range.kind === 2 /* SingleLineCommentTrivia */ && tripleSlashXMLCommentStartRegEx.exec(text);
            if (tripleSlash) {
                var name = tripleSlash[1].toLowerCase(); // Technically unsafe cast, but we do it so the below check to make it safe typechecks
                var pragma = ts.commentPragmas[name];
                if (!pragma || !(pragma.kind & 1 /* TripleSlashXML */)) {
                    return;
                }
                if (pragma.args) {
                    var argument = {};
                    for (var _i = 0, _a = pragma.args; _i < _a.length; _i++) {
                        var arg = _a[_i];
                        var matcher = getNamedArgRegEx(arg.name);
                        var matchResult = matcher.exec(text);
                        if (!matchResult && !arg.optional) {
                            return; // Missing required argument, don't parse
                        }
                        else if (matchResult) {
                            var value = matchResult[2] || matchResult[3];
                            if (arg.captureSpan) {
                                var startPos = range.pos + matchResult.index + matchResult[1].length + 1;
                                argument[arg.name] = {
                                    value: value,
                                    pos: startPos,
                                    end: startPos + value.length
                                };
                            }
                            else {
                                argument[arg.name] = value;
                            }
                        }
                    }
                    pragmas.push({ name: name, args: { arguments: argument, range: range } });
                }
                else {
                    pragmas.push({ name: name, args: { arguments: {}, range: range } });
                }
                return;
            }
            var singleLine = range.kind === 2 /* SingleLineCommentTrivia */ && singleLinePragmaRegEx.exec(text);
            if (singleLine) {
                return addPragmaForMatch(pragmas, range, 2 /* SingleLine */, singleLine);
            }
            if (range.kind === 3 /* MultiLineCommentTrivia */) {
                var multiLinePragmaRegEx = /@(\S+)(\s+.*)?$/gim; // Defined inline since it uses the "g" flag, which keeps a persistent index (for iterating)
                var multiLineMatch = void 0;
                while (multiLineMatch = multiLinePragmaRegEx.exec(text)) {
                    addPragmaForMatch(pragmas, range, 4 /* MultiLine */, multiLineMatch);
                }
            }
        }
        function addPragmaForMatch(pragmas, range, kind, match) {
            if (!match)
                return;
            var name = match[1].toLowerCase(); // Technically unsafe cast, but we do it so they below check to make it safe typechecks
            var pragma = ts.commentPragmas[name];
            if (!pragma || !(pragma.kind & kind)) {
                return;
            }
            var args = match[2]; // Split on spaces and match up positionally with definition
            var argument = getNamedPragmaArguments(pragma, args);
            if (argument === "fail")
                return; // Missing required argument, fail to parse it
            pragmas.push({ name: name, args: { arguments: argument, range: range } });
            return;
        }
        function getNamedPragmaArguments(pragma, text) {
            if (!text)
                return {};
            if (!pragma.args)
                return {};
            var args = ts.trimString(text).split(/\s+/);
            var argMap = {};
            for (var i = 0; i < pragma.args.length; i++) {
                var argument = pragma.args[i];
                if (!args[i] && !argument.optional) {
                    return "fail";
                }
                if (argument.captureSpan) {
                    return ts.Debug.fail("Capture spans not yet implemented for non-xml pragmas");
                }
                argMap[argument.name] = args[i];
            }
            return argMap;
        }
        /** @internal */
        function tagNamesAreEquivalent(lhs, rhs) {
            if (lhs.kind !== rhs.kind) {
                return false;
            }
            if (lhs.kind === 79 /* Identifier */) {
                return lhs.escapedText === rhs.escapedText;
            }
            if (lhs.kind === 108 /* ThisKeyword */) {
                return true;
            }
            // If we are at this statement then we must have PropertyAccessExpression and because tag name in Jsx element can only
            // take forms of JsxTagNameExpression which includes an identifier, "this" expression, or another propertyAccessExpression
            // it is safe to case the expression property as such. See parseJsxElementName for how we parse tag name in Jsx element
            return lhs.name.escapedText === rhs.name.escapedText &&
                tagNamesAreEquivalent(lhs.expression, rhs.expression);
        }
        ts.tagNamesAreEquivalent = tagNamesAreEquivalent;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        /* @internal */
        ts.compileOnSaveCommandLineOption = { name: "compileOnSave", type: "boolean" };
        var jsxOptionMap = new ts.Map(ts.getEntries({
            "preserve": 1 /* Preserve */,
            "react-native": 3 /* ReactNative */,
            "react": 2 /* React */,
            "react-jsx": 4 /* ReactJSX */,
            "react-jsxdev": 5 /* ReactJSXDev */,
        }));
        /* @internal */
        ts.inverseJsxOptionMap = new ts.Map(ts.arrayFrom(ts.mapIterator(jsxOptionMap.entries(), function (_a) {
            var key = _a[0], value = _a[1];
            return ["" + value, key];
        })));
        // NOTE: The order here is important to default lib ordering as entries will have the same
        //       order in the generated program (see `getDefaultLibPriority` in program.ts). This
        //       order also affects overload resolution when a type declared in one lib is
        //       augmented in another lib.
        var libEntries = [
            // JavaScript only
            ["es5", "lib.es5.d.ts"],
            ["es6", "lib.es2015.d.ts"],
            ["es2015", "lib.es2015.d.ts"],
            ["es7", "lib.es2016.d.ts"],
            ["es2016", "lib.es2016.d.ts"],
            ["es2017", "lib.es2017.d.ts"],
            ["es2018", "lib.es2018.d.ts"],
            ["es2019", "lib.es2019.d.ts"],
            ["es2020", "lib.es2020.d.ts"],
            ["es2021", "lib.es2021.d.ts"],
            ["esnext", "lib.esnext.d.ts"],
            // Host only
            ["dom", "lib.dom.d.ts"],
            ["dom.iterable", "lib.dom.iterable.d.ts"],
            ["webworker", "lib.webworker.d.ts"],
            ["webworker.importscripts", "lib.webworker.importscripts.d.ts"],
            ["webworker.iterable", "lib.webworker.iterable.d.ts"],
            ["scripthost", "lib.scripthost.d.ts"],
            // ES2015 Or ESNext By-feature options
            ["es2015.core", "lib.es2015.core.d.ts"],
            ["es2015.collection", "lib.es2015.collection.d.ts"],
            ["es2015.generator", "lib.es2015.generator.d.ts"],
            ["es2015.iterable", "lib.es2015.iterable.d.ts"],
            ["es2015.promise", "lib.es2015.promise.d.ts"],
            ["es2015.proxy", "lib.es2015.proxy.d.ts"],
            ["es2015.reflect", "lib.es2015.reflect.d.ts"],
            ["es2015.symbol", "lib.es2015.symbol.d.ts"],
            ["es2015.symbol.wellknown", "lib.es2015.symbol.wellknown.d.ts"],
            ["es2016.array.include", "lib.es2016.array.include.d.ts"],
            ["es2017.object", "lib.es2017.object.d.ts"],
            ["es2017.sharedmemory", "lib.es2017.sharedmemory.d.ts"],
            ["es2017.string", "lib.es2017.string.d.ts"],
            ["es2017.intl", "lib.es2017.intl.d.ts"],
            ["es2017.typedarrays", "lib.es2017.typedarrays.d.ts"],
            ["es2018.asyncgenerator", "lib.es2018.asyncgenerator.d.ts"],
            ["es2018.asynciterable", "lib.es2018.asynciterable.d.ts"],
            ["es2018.intl", "lib.es2018.intl.d.ts"],
            ["es2018.promise", "lib.es2018.promise.d.ts"],
            ["es2018.regexp", "lib.es2018.regexp.d.ts"],
            ["es2019.array", "lib.es2019.array.d.ts"],
            ["es2019.object", "lib.es2019.object.d.ts"],
            ["es2019.string", "lib.es2019.string.d.ts"],
            ["es2019.symbol", "lib.es2019.symbol.d.ts"],
            ["es2020.bigint", "lib.es2020.bigint.d.ts"],
            ["es2020.promise", "lib.es2020.promise.d.ts"],
            ["es2020.sharedmemory", "lib.es2020.sharedmemory.d.ts"],
            ["es2020.string", "lib.es2020.string.d.ts"],
            ["es2020.symbol.wellknown", "lib.es2020.symbol.wellknown.d.ts"],
            ["es2020.intl", "lib.es2020.intl.d.ts"],
            ["es2021.promise", "lib.es2021.promise.d.ts"],
            ["es2021.string", "lib.es2021.string.d.ts"],
            ["es2021.weakref", "lib.es2021.weakref.d.ts"],
            ["esnext.array", "lib.es2019.array.d.ts"],
            ["esnext.symbol", "lib.es2019.symbol.d.ts"],
            ["esnext.asynciterable", "lib.es2018.asynciterable.d.ts"],
            ["esnext.intl", "lib.esnext.intl.d.ts"],
            ["esnext.bigint", "lib.es2020.bigint.d.ts"],
            ["esnext.string", "lib.es2021.string.d.ts"],
            ["esnext.promise", "lib.es2021.promise.d.ts"],
            ["esnext.weakref", "lib.es2021.weakref.d.ts"]
        ];
        /**
         * An array of supported "lib" reference file names used to determine the order for inclusion
         * when referenced, as well as for spelling suggestions. This ensures the correct ordering for
         * overload resolution when a type declared in one lib is extended by another.
         */
        /* @internal */
        ts.libs = libEntries.map(function (entry) { return entry[0]; });
        /**
         * A map of lib names to lib files. This map is used both for parsing the "lib" command line
         * option as well as for resolving lib reference directives.
         */
        /* @internal */
        ts.libMap = new ts.Map(libEntries);
        // Watch related options
        /* @internal */
        ts.optionsForWatch = [
            {
                name: "watchFile",
                type: new ts.Map(ts.getEntries({
                    fixedpollinginterval: ts.WatchFileKind.FixedPollingInterval,
                    prioritypollinginterval: ts.WatchFileKind.PriorityPollingInterval,
                    dynamicprioritypolling: ts.WatchFileKind.DynamicPriorityPolling,
                    fixedchunksizepolling: ts.WatchFileKind.FixedChunkSizePolling,
                    usefsevents: ts.WatchFileKind.UseFsEvents,
                    usefseventsonparentdirectory: ts.WatchFileKind.UseFsEventsOnParentDirectory,
                })),
                category: ts.Diagnostics.Watch_and_Build_Modes,
                description: ts.Diagnostics.Specify_how_the_TypeScript_watch_mode_works,
            },
            {
                name: "watchDirectory",
                type: new ts.Map(ts.getEntries({
                    usefsevents: ts.WatchDirectoryKind.UseFsEvents,
                    fixedpollinginterval: ts.WatchDirectoryKind.FixedPollingInterval,
                    dynamicprioritypolling: ts.WatchDirectoryKind.DynamicPriorityPolling,
                    fixedchunksizepolling: ts.WatchDirectoryKind.FixedChunkSizePolling,
                })),
                category: ts.Diagnostics.Watch_and_Build_Modes,
                description: ts.Diagnostics.Specify_how_directories_are_watched_on_systems_that_lack_recursive_file_watching_functionality,
            },
            {
                name: "fallbackPolling",
                type: new ts.Map(ts.getEntries({
                    fixedinterval: ts.PollingWatchKind.FixedInterval,
                    priorityinterval: ts.PollingWatchKind.PriorityInterval,
                    dynamicpriority: ts.PollingWatchKind.DynamicPriority,
                    fixedchunksize: ts.PollingWatchKind.FixedChunkSize,
                })),
                category: ts.Diagnostics.Watch_and_Build_Modes,
                description: ts.Diagnostics.Specify_what_approach_the_watcher_should_use_if_the_system_runs_out_of_native_file_watchers,
            },
            {
                name: "synchronousWatchDirectory",
                type: "boolean",
                category: ts.Diagnostics.Watch_and_Build_Modes,
                description: ts.Diagnostics.Synchronously_call_callbacks_and_update_the_state_of_directory_watchers_on_platforms_that_don_t_support_recursive_watching_natively,
            },
            {
                name: "excludeDirectories",
                type: "list",
                element: {
                    name: "excludeDirectory",
                    type: "string",
                    isFilePath: true,
                    extraValidation: specToDiagnostic
                },
                category: ts.Diagnostics.Watch_and_Build_Modes,
                description: ts.Diagnostics.Remove_a_list_of_directories_from_the_watch_process,
            },
            {
                name: "excludeFiles",
                type: "list",
                element: {
                    name: "excludeFile",
                    type: "string",
                    isFilePath: true,
                    extraValidation: specToDiagnostic
                },
                category: ts.Diagnostics.Watch_and_Build_Modes,
                description: ts.Diagnostics.Remove_a_list_of_files_from_the_watch_mode_s_processing,
            },
        ];
        /* @internal */
        ts.commonOptionsWithBuild = [
            {
                name: "help",
                shortName: "h",
                type: "boolean",
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Print_this_message,
            },
            {
                name: "help",
                shortName: "?",
                type: "boolean"
            },
            {
                name: "watch",
                shortName: "w",
                type: "boolean",
                showInSimplifiedHelpView: true,
                isCommandLineOnly: true,
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Watch_input_files,
            },
            {
                name: "preserveWatchOutput",
                type: "boolean",
                showInSimplifiedHelpView: false,
                category: ts.Diagnostics.Output_Formatting,
                description: ts.Diagnostics.Disable_wiping_the_console_in_watch_mode,
                defaultValueDescription: "n/a"
            },
            {
                name: "listFiles",
                type: "boolean",
                category: ts.Diagnostics.Compiler_Diagnostics,
                description: ts.Diagnostics.Print_all_of_the_files_read_during_the_compilation,
                defaultValueDescription: "false"
            },
            {
                name: "explainFiles",
                type: "boolean",
                category: ts.Diagnostics.Compiler_Diagnostics,
                description: ts.Diagnostics.Print_files_read_during_the_compilation_including_why_it_was_included
            },
            {
                name: "listEmittedFiles",
                type: "boolean",
                category: ts.Diagnostics.Compiler_Diagnostics,
                description: ts.Diagnostics.Print_the_names_of_emitted_files_after_a_compilation,
                defaultValueDescription: "false"
            },
            {
                name: "pretty",
                type: "boolean",
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Output_Formatting,
                description: ts.Diagnostics.Enable_color_and_formatting_in_TypeScript_s_output_to_make_compiler_errors_easier_to_read,
                defaultValueDescription: "true"
            },
            {
                name: "traceResolution",
                type: "boolean",
                category: ts.Diagnostics.Compiler_Diagnostics,
                description: ts.Diagnostics.Log_paths_used_during_the_moduleResolution_process,
                defaultValueDescription: "false"
            },
            {
                name: "diagnostics",
                type: "boolean",
                category: ts.Diagnostics.Compiler_Diagnostics,
                description: ts.Diagnostics.Output_compiler_performance_information_after_building,
                defaultValueDescription: "false"
            },
            {
                name: "extendedDiagnostics",
                type: "boolean",
                category: ts.Diagnostics.Compiler_Diagnostics,
                description: ts.Diagnostics.Output_more_detailed_compiler_performance_information_after_building,
                defaultValueDescription: "false"
            },
            {
                name: "generateCpuProfile",
                type: "string",
                isFilePath: true,
                paramType: ts.Diagnostics.FILE_OR_DIRECTORY,
                category: ts.Diagnostics.Compiler_Diagnostics,
                description: ts.Diagnostics.Emit_a_v8_CPU_profile_of_the_compiler_run_for_debugging,
                defaultValueDescription: "profile.cpuprofile"
            },
            {
                name: "generateTrace",
                type: "string",
                isFilePath: true,
                isCommandLineOnly: true,
                paramType: ts.Diagnostics.DIRECTORY,
                category: ts.Diagnostics.Compiler_Diagnostics,
                description: ts.Diagnostics.Generates_an_event_trace_and_a_list_of_types
            },
            {
                name: "incremental",
                shortName: "i",
                type: "boolean",
                category: ts.Diagnostics.Projects,
                description: ts.Diagnostics.Enable_incremental_compilation,
                transpileOptionValue: undefined,
                defaultValueDescription: ts.Diagnostics.false_unless_composite_is_set
            },
            {
                name: "assumeChangesOnlyAffectDirectDependencies",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                affectsEmit: true,
                category: ts.Diagnostics.Watch_and_Build_Modes,
                description: ts.Diagnostics.Have_recompiles_in_projects_that_use_incremental_and_watch_mode_assume_that_changes_within_a_file_will_only_affect_files_directly_depending_on_it
            },
            {
                name: "locale",
                type: "string",
                category: ts.Diagnostics.Command_line_Options,
                isCommandLineOnly: true,
                description: ts.Diagnostics.Set_the_language_of_the_messaging_from_TypeScript_This_does_not_affect_emit,
                defaultValueDescription: ts.Diagnostics.Platform_specific
            },
        ];
        /* @internal */
        ts.targetOptionDeclaration = {
            name: "target",
            shortName: "t",
            type: new ts.Map(ts.getEntries({
                es3: 0 /* ES3 */,
                es5: 1 /* ES5 */,
                es6: 2 /* ES2015 */,
                es2015: 2 /* ES2015 */,
                es2016: 3 /* ES2016 */,
                es2017: 4 /* ES2017 */,
                es2018: 5 /* ES2018 */,
                es2019: 6 /* ES2019 */,
                es2020: 7 /* ES2020 */,
                es2021: 8 /* ES2021 */,
                esnext: 99 /* ESNext */,
            })),
            affectsSourceFile: true,
            affectsModuleResolution: true,
            affectsEmit: true,
            paramType: ts.Diagnostics.VERSION,
            showInSimplifiedHelpView: true,
            category: ts.Diagnostics.Language_and_Environment,
            description: ts.Diagnostics.Set_the_JavaScript_language_version_for_emitted_JavaScript_and_include_compatible_library_declarations,
            defaultValueDescription: "ES3"
        };
        var commandOptionsWithoutBuild = [
            // CommandLine only options
            {
                name: "all",
                type: "boolean",
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Show_all_compiler_options,
            },
            {
                name: "version",
                shortName: "v",
                type: "boolean",
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Print_the_compiler_s_version,
            },
            {
                name: "init",
                type: "boolean",
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Initializes_a_TypeScript_project_and_creates_a_tsconfig_json_file,
            },
            {
                name: "project",
                shortName: "p",
                type: "string",
                isFilePath: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Command_line_Options,
                paramType: ts.Diagnostics.FILE_OR_DIRECTORY,
                description: ts.Diagnostics.Compile_the_project_given_the_path_to_its_configuration_file_or_to_a_folder_with_a_tsconfig_json,
            },
            {
                name: "build",
                type: "boolean",
                shortName: "b",
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Build_one_or_more_projects_and_their_dependencies_if_out_of_date
            },
            {
                name: "showConfig",
                type: "boolean",
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Command_line_Options,
                isCommandLineOnly: true,
                description: ts.Diagnostics.Print_the_final_configuration_instead_of_building
            },
            {
                name: "listFilesOnly",
                type: "boolean",
                category: ts.Diagnostics.Command_line_Options,
                affectsSemanticDiagnostics: true,
                affectsEmit: true,
                isCommandLineOnly: true,
                description: ts.Diagnostics.Print_names_of_files_that_are_part_of_the_compilation_and_then_stop_processing
            },
            // Basic
            ts.targetOptionDeclaration,
            {
                name: "module",
                shortName: "m",
                type: new ts.Map(ts.getEntries({
                    none: ts.ModuleKind.None,
                    commonjs: ts.ModuleKind.CommonJS,
                    amd: ts.ModuleKind.AMD,
                    system: ts.ModuleKind.System,
                    umd: ts.ModuleKind.UMD,
                    es6: ts.ModuleKind.ES2015,
                    es2015: ts.ModuleKind.ES2015,
                    es2020: ts.ModuleKind.ES2020,
                    esnext: ts.ModuleKind.ESNext
                })),
                affectsModuleResolution: true,
                affectsEmit: true,
                paramType: ts.Diagnostics.KIND,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Specify_what_module_code_is_generated,
            },
            {
                name: "lib",
                type: "list",
                element: {
                    name: "lib",
                    type: ts.libMap
                },
                affectsProgramStructure: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Language_and_Environment,
                description: ts.Diagnostics.Specify_a_set_of_bundled_library_declaration_files_that_describe_the_target_runtime_environment,
                transpileOptionValue: undefined
            },
            {
                name: "allowJs",
                type: "boolean",
                affectsModuleResolution: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.JavaScript_Support,
                description: ts.Diagnostics.Allow_JavaScript_files_to_be_a_part_of_your_program_Use_the_checkJS_option_to_get_errors_from_these_files,
                defaultValueDescription: "false"
            },
            {
                name: "checkJs",
                type: "boolean",
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.JavaScript_Support,
                description: ts.Diagnostics.Enable_error_reporting_in_type_checked_JavaScript_files,
                defaultValueDescription: "false"
            },
            {
                name: "jsx",
                type: jsxOptionMap,
                affectsSourceFile: true,
                affectsEmit: true,
                affectsModuleResolution: true,
                paramType: ts.Diagnostics.KIND,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Language_and_Environment,
                description: ts.Diagnostics.Specify_what_JSX_code_is_generated,
                defaultValueDescription: "undefined"
            },
            {
                name: "declaration",
                shortName: "d",
                type: "boolean",
                affectsEmit: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Emit,
                transpileOptionValue: undefined,
                description: ts.Diagnostics.Generate_d_ts_files_from_TypeScript_and_JavaScript_files_in_your_project,
                defaultValueDescription: ts.Diagnostics.false_unless_composite_is_set,
            },
            {
                name: "declarationMap",
                type: "boolean",
                affectsEmit: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Emit,
                transpileOptionValue: undefined,
                defaultValueDescription: "false",
                description: ts.Diagnostics.Create_sourcemaps_for_d_ts_files
            },
            {
                name: "emitDeclarationOnly",
                type: "boolean",
                affectsEmit: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Only_output_d_ts_files_and_not_JavaScript_files,
                transpileOptionValue: undefined,
                defaultValueDescription: "false",
            },
            {
                name: "sourceMap",
                type: "boolean",
                affectsEmit: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Emit,
                defaultValueDescription: "false",
                description: ts.Diagnostics.Create_source_map_files_for_emitted_JavaScript_files,
            },
            {
                name: "outFile",
                type: "string",
                affectsEmit: true,
                isFilePath: true,
                paramType: ts.Diagnostics.FILE,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Specify_a_file_that_bundles_all_outputs_into_one_JavaScript_file_If_declaration_is_true_also_designates_a_file_that_bundles_all_d_ts_output,
                transpileOptionValue: undefined,
                defaultValueDescription: "n/a"
            },
            {
                name: "outDir",
                type: "string",
                affectsEmit: true,
                isFilePath: true,
                paramType: ts.Diagnostics.DIRECTORY,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Specify_an_output_folder_for_all_emitted_files,
                defaultValueDescription: "n/a"
            },
            {
                name: "rootDir",
                type: "string",
                affectsEmit: true,
                isFilePath: true,
                paramType: ts.Diagnostics.LOCATION,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Specify_the_root_folder_within_your_source_files,
                defaultValueDescription: ts.Diagnostics.Computed_from_the_list_of_input_files
            },
            {
                name: "composite",
                type: "boolean",
                affectsEmit: true,
                isTSConfigOnly: true,
                category: ts.Diagnostics.Projects,
                transpileOptionValue: undefined,
                defaultValueDescription: "false",
                description: ts.Diagnostics.Enable_constraints_that_allow_a_TypeScript_project_to_be_used_with_project_references,
            },
            {
                name: "tsBuildInfoFile",
                type: "string",
                affectsEmit: true,
                isFilePath: true,
                paramType: ts.Diagnostics.FILE,
                category: ts.Diagnostics.Projects,
                transpileOptionValue: undefined,
                defaultValueDescription: ".tsbuildinfo",
                description: ts.Diagnostics.Specify_the_folder_for_tsbuildinfo_incremental_compilation_files,
            },
            {
                name: "removeComments",
                type: "boolean",
                affectsEmit: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Emit,
                defaultValueDescription: "false",
                description: ts.Diagnostics.Disable_emitting_comments,
            },
            {
                name: "noEmit",
                type: "boolean",
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Disable_emitting_files_from_a_compilation,
                transpileOptionValue: undefined,
                defaultValueDescription: "false"
            },
            {
                name: "importHelpers",
                type: "boolean",
                affectsEmit: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Allow_importing_helper_functions_from_tslib_once_per_project_instead_of_including_them_per_file,
                defaultValueDescription: "false"
            },
            {
                name: "importsNotUsedAsValues",
                type: new ts.Map(ts.getEntries({
                    remove: 0 /* Remove */,
                    preserve: 1 /* Preserve */,
                    error: 2 /* Error */
                })),
                affectsEmit: true,
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Specify_emit_Slashchecking_behavior_for_imports_that_are_only_used_for_types
            },
            {
                name: "downlevelIteration",
                type: "boolean",
                affectsEmit: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Emit_more_compliant_but_verbose_and_less_performant_JavaScript_for_iteration,
                defaultValueDescription: "false"
            },
            {
                name: "isolatedModules",
                type: "boolean",
                category: ts.Diagnostics.Interop_Constraints,
                description: ts.Diagnostics.Ensure_that_each_file_can_be_safely_transpiled_without_relying_on_other_imports,
                transpileOptionValue: true,
                defaultValueDescription: "false"
            },
            // Strict Type Checks
            {
                name: "strict",
                type: "boolean",
                // Though this affects semantic diagnostics, affectsSemanticDiagnostics is not set here
                // The value of each strictFlag depends on own strictFlag value or this and never accessed directly.
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Enable_all_strict_type_checking_options,
                defaultValueDescription: "false"
            },
            {
                name: "noImplicitAny",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                strictFlag: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Enable_error_reporting_for_expressions_and_declarations_with_an_implied_any_type,
                defaultValueDescription: ts.Diagnostics.false_unless_strict_is_set
            },
            {
                name: "strictNullChecks",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                strictFlag: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.When_type_checking_take_into_account_null_and_undefined,
                defaultValueDescription: ts.Diagnostics.false_unless_strict_is_set
            },
            {
                name: "strictFunctionTypes",
                type: "boolean",
                strictFlag: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.When_assigning_functions_check_to_ensure_parameters_and_the_return_values_are_subtype_compatible,
                defaultValueDescription: ts.Diagnostics.false_unless_strict_is_set
            },
            {
                name: "strictBindCallApply",
                type: "boolean",
                strictFlag: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Check_that_the_arguments_for_bind_call_and_apply_methods_match_the_original_function,
                defaultValueDescription: ts.Diagnostics.false_unless_strict_is_set
            },
            {
                name: "strictPropertyInitialization",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                strictFlag: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Check_for_class_properties_that_are_declared_but_not_set_in_the_constructor,
                defaultValueDescription: ts.Diagnostics.false_unless_strict_is_set
            },
            {
                name: "noImplicitThis",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                strictFlag: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Enable_error_reporting_when_this_is_given_the_type_any,
                defaultValueDescription: ts.Diagnostics.false_unless_strict_is_set
            },
            {
                name: "useUnknownInCatchVariables",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                strictFlag: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Type_catch_clause_variables_as_unknown_instead_of_any,
            },
            {
                name: "alwaysStrict",
                type: "boolean",
                affectsSourceFile: true,
                strictFlag: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Ensure_use_strict_is_always_emitted,
                defaultValueDescription: ts.Diagnostics.false_unless_strict_is_set
            },
            // Additional Checks
            {
                name: "noUnusedLocals",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Enable_error_reporting_when_a_local_variables_aren_t_read,
                defaultValueDescription: "false"
            },
            {
                name: "noUnusedParameters",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Raise_an_error_when_a_function_parameter_isn_t_read,
                defaultValueDescription: "false"
            },
            {
                name: "exactOptionalPropertyTypes",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Interpret_optional_property_types_as_written_rather_than_adding_undefined
            },
            {
                name: "noImplicitReturns",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Enable_error_reporting_for_codepaths_that_do_not_explicitly_return_in_a_function,
                defaultValueDescription: "false"
            },
            {
                name: "noFallthroughCasesInSwitch",
                type: "boolean",
                affectsBindDiagnostics: true,
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Enable_error_reporting_for_fallthrough_cases_in_switch_statements
            },
            {
                name: "noUncheckedIndexedAccess",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Include_undefined_in_index_signature_results
            },
            {
                name: "noImplicitOverride",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Ensure_overriding_members_in_derived_classes_are_marked_with_an_override_modifier
            },
            {
                name: "noPropertyAccessFromIndexSignature",
                type: "boolean",
                showInSimplifiedHelpView: false,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Enforces_using_indexed_accessors_for_keys_declared_using_an_indexed_type,
                defaultValueDescription: "false"
            },
            // Module Resolution
            {
                name: "moduleResolution",
                type: new ts.Map(ts.getEntries({
                    node: ts.ModuleResolutionKind.NodeJs,
                    classic: ts.ModuleResolutionKind.Classic,
                })),
                affectsModuleResolution: true,
                paramType: ts.Diagnostics.STRATEGY,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Specify_how_TypeScript_looks_up_a_file_from_a_given_module_specifier,
                defaultValueDescription: ts.Diagnostics.module_AMD_or_UMD_or_System_or_ES6_then_Classic_Otherwise_Node
            },
            {
                name: "baseUrl",
                type: "string",
                affectsModuleResolution: true,
                isFilePath: true,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Specify_the_base_directory_to_resolve_non_relative_module_names
            },
            {
                // this option can only be specified in tsconfig.json
                // use type = object to copy the value as-is
                name: "paths",
                type: "object",
                affectsModuleResolution: true,
                isTSConfigOnly: true,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Specify_a_set_of_entries_that_re_map_imports_to_additional_lookup_locations,
                transpileOptionValue: undefined
            },
            {
                // this option can only be specified in tsconfig.json
                // use type = object to copy the value as-is
                name: "rootDirs",
                type: "list",
                isTSConfigOnly: true,
                element: {
                    name: "rootDirs",
                    type: "string",
                    isFilePath: true
                },
                affectsModuleResolution: true,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Allow_multiple_folders_to_be_treated_as_one_when_resolving_modules,
                transpileOptionValue: undefined,
                defaultValueDescription: ts.Diagnostics.Computed_from_the_list_of_input_files
            },
            {
                name: "typeRoots",
                type: "list",
                element: {
                    name: "typeRoots",
                    type: "string",
                    isFilePath: true
                },
                affectsModuleResolution: true,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Specify_multiple_folders_that_act_like_Slashnode_modules_Slash_types
            },
            {
                name: "types",
                type: "list",
                element: {
                    name: "types",
                    type: "string"
                },
                affectsProgramStructure: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Specify_type_package_names_to_be_included_without_being_referenced_in_a_source_file,
                transpileOptionValue: undefined
            },
            {
                name: "allowSyntheticDefaultImports",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Interop_Constraints,
                description: ts.Diagnostics.Allow_import_x_from_y_when_a_module_doesn_t_have_a_default_export,
                defaultValueDescription: ts.Diagnostics.module_system_or_esModuleInterop
            },
            {
                name: "esModuleInterop",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                affectsEmit: true,
                showInSimplifiedHelpView: true,
                category: ts.Diagnostics.Interop_Constraints,
                description: ts.Diagnostics.Emit_additional_JavaScript_to_ease_support_for_importing_CommonJS_modules_This_enables_allowSyntheticDefaultImports_for_type_compatibility,
                defaultValueDescription: "false"
            },
            {
                name: "preserveSymlinks",
                type: "boolean",
                category: ts.Diagnostics.Interop_Constraints,
                description: ts.Diagnostics.Disable_resolving_symlinks_to_their_realpath_This_correlates_to_the_same_flag_in_node,
                defaultValueDescription: "n/a"
            },
            {
                name: "allowUmdGlobalAccess",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Allow_accessing_UMD_globals_from_modules,
                defaultValueDescription: "false"
            },
            // Source Maps
            {
                name: "sourceRoot",
                type: "string",
                affectsEmit: true,
                paramType: ts.Diagnostics.LOCATION,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Specify_the_root_path_for_debuggers_to_find_the_reference_source_code,
            },
            {
                name: "mapRoot",
                type: "string",
                affectsEmit: true,
                paramType: ts.Diagnostics.LOCATION,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Specify_the_location_where_debugger_should_locate_map_files_instead_of_generated_locations,
            },
            {
                name: "inlineSourceMap",
                type: "boolean",
                affectsEmit: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Include_sourcemap_files_inside_the_emitted_JavaScript,
                defaultValueDescription: "false"
            },
            {
                name: "inlineSources",
                type: "boolean",
                affectsEmit: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Include_source_code_in_the_sourcemaps_inside_the_emitted_JavaScript,
                defaultValueDescription: "false"
            },
            // Experimental
            {
                name: "experimentalDecorators",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Language_and_Environment,
                description: ts.Diagnostics.Enable_experimental_support_for_TC39_stage_2_draft_decorators
            },
            {
                name: "emitDecoratorMetadata",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                affectsEmit: true,
                category: ts.Diagnostics.Language_and_Environment,
                description: ts.Diagnostics.Emit_design_type_metadata_for_decorated_declarations_in_source_files
            },
            // Advanced
            {
                name: "jsxFactory",
                type: "string",
                category: ts.Diagnostics.Language_and_Environment,
                description: ts.Diagnostics.Specify_the_JSX_factory_function_used_when_targeting_React_JSX_emit_e_g_React_createElement_or_h,
                defaultValueDescription: "`React.createElement`"
            },
            {
                name: "jsxFragmentFactory",
                type: "string",
                category: ts.Diagnostics.Language_and_Environment,
                description: ts.Diagnostics.Specify_the_JSX_Fragment_reference_used_for_fragments_when_targeting_React_JSX_emit_e_g_React_Fragment_or_Fragment
            },
            {
                name: "jsxImportSource",
                type: "string",
                affectsSemanticDiagnostics: true,
                affectsEmit: true,
                affectsModuleResolution: true,
                category: ts.Diagnostics.Language_and_Environment,
                description: ts.Diagnostics.Specify_module_specifier_used_to_import_the_JSX_factory_functions_when_using_jsx_Colon_react_jsx_Asterisk,
                defaultValueDescription: "react"
            },
            {
                name: "resolveJsonModule",
                type: "boolean",
                affectsModuleResolution: true,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Enable_importing_json_files,
                defaultValueDescription: "false"
            },
            {
                name: "out",
                type: "string",
                affectsEmit: true,
                isFilePath: false,
                // for correct behaviour, please use outFile
                category: ts.Diagnostics.Backwards_Compatibility,
                paramType: ts.Diagnostics.FILE,
                transpileOptionValue: undefined,
                defaultValueDescription: "n/a",
                description: ts.Diagnostics.Deprecated_setting_Use_outFile_instead,
            },
            {
                name: "reactNamespace",
                type: "string",
                affectsEmit: true,
                category: ts.Diagnostics.Language_and_Environment,
                description: ts.Diagnostics.Specify_the_object_invoked_for_createElement_This_only_applies_when_targeting_react_JSX_emit,
                defaultValueDescription: "`React`",
            },
            {
                name: "skipDefaultLibCheck",
                type: "boolean",
                category: ts.Diagnostics.Completeness,
                description: ts.Diagnostics.Skip_type_checking_d_ts_files_that_are_included_with_TypeScript,
                defaultValueDescription: "false",
            },
            {
                name: "charset",
                type: "string",
                category: ts.Diagnostics.Backwards_Compatibility,
                description: ts.Diagnostics.No_longer_supported_In_early_versions_manually_set_the_text_encoding_for_reading_files,
                defaultValueDescription: "utf8"
            },
            {
                name: "emitBOM",
                type: "boolean",
                affectsEmit: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Emit_a_UTF_8_Byte_Order_Mark_BOM_in_the_beginning_of_output_files,
                defaultValueDescription: "false"
            },
            {
                name: "newLine",
                type: new ts.Map(ts.getEntries({
                    crlf: 0 /* CarriageReturnLineFeed */,
                    lf: 1 /* LineFeed */
                })),
                affectsEmit: true,
                paramType: ts.Diagnostics.NEWLINE,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Set_the_newline_character_for_emitting_files,
                defaultValueDescription: ts.Diagnostics.Platform_specific
            },
            {
                name: "noErrorTruncation",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Output_Formatting,
                description: ts.Diagnostics.Disable_truncating_types_in_error_messages,
                defaultValueDescription: "false"
            },
            {
                name: "noLib",
                type: "boolean",
                category: ts.Diagnostics.Language_and_Environment,
                affectsProgramStructure: true,
                description: ts.Diagnostics.Disable_including_any_library_files_including_the_default_lib_d_ts,
                // We are not returning a sourceFile for lib file when asked by the program,
                // so pass --noLib to avoid reporting a file not found error.
                transpileOptionValue: true,
                defaultValueDescription: "false"
            },
            {
                name: "noResolve",
                type: "boolean",
                affectsModuleResolution: true,
                category: ts.Diagnostics.Modules,
                description: ts.Diagnostics.Disallow_import_s_require_s_or_reference_s_from_expanding_the_number_of_files_TypeScript_should_add_to_a_project,
                // We are not doing a full typecheck, we are not resolving the whole context,
                // so pass --noResolve to avoid reporting missing file errors.
                transpileOptionValue: true,
                defaultValueDescription: "false"
            },
            {
                name: "stripInternal",
                type: "boolean",
                affectsEmit: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Disable_emitting_declarations_that_have_internal_in_their_JSDoc_comments,
            },
            {
                name: "disableSizeLimit",
                type: "boolean",
                affectsProgramStructure: true,
                category: ts.Diagnostics.Editor_Support,
                description: ts.Diagnostics.Remove_the_20mb_cap_on_total_source_code_size_for_JavaScript_files_in_the_TypeScript_language_server,
                defaultValueDescription: "false"
            },
            {
                name: "disableSourceOfProjectReferenceRedirect",
                type: "boolean",
                isTSConfigOnly: true,
                category: ts.Diagnostics.Projects,
                description: ts.Diagnostics.Disable_preferring_source_files_instead_of_declaration_files_when_referencing_composite_projects
            },
            {
                name: "disableSolutionSearching",
                type: "boolean",
                isTSConfigOnly: true,
                category: ts.Diagnostics.Projects,
                description: ts.Diagnostics.Opt_a_project_out_of_multi_project_reference_checking_when_editing
            },
            {
                name: "disableReferencedProjectLoad",
                type: "boolean",
                isTSConfigOnly: true,
                category: ts.Diagnostics.Projects,
                description: ts.Diagnostics.Reduce_the_number_of_projects_loaded_automatically_by_TypeScript
            },
            {
                name: "noImplicitUseStrict",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Backwards_Compatibility,
                description: ts.Diagnostics.Disable_adding_use_strict_directives_in_emitted_JavaScript_files,
                defaultValueDescription: "false"
            },
            {
                name: "noEmitHelpers",
                type: "boolean",
                affectsEmit: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Disable_generating_custom_helper_functions_like_extends_in_compiled_output,
                defaultValueDescription: "false"
            },
            {
                name: "noEmitOnError",
                type: "boolean",
                affectsEmit: true,
                category: ts.Diagnostics.Emit,
                transpileOptionValue: undefined,
                description: ts.Diagnostics.Disable_emitting_files_if_any_type_checking_errors_are_reported,
                defaultValueDescription: "false"
            },
            {
                name: "preserveConstEnums",
                type: "boolean",
                affectsEmit: true,
                category: ts.Diagnostics.Emit,
                description: ts.Diagnostics.Disable_erasing_const_enum_declarations_in_generated_code,
                defaultValueDescription: "n/a"
            },
            {
                name: "declarationDir",
                type: "string",
                affectsEmit: true,
                isFilePath: true,
                paramType: ts.Diagnostics.DIRECTORY,
                category: ts.Diagnostics.Emit,
                transpileOptionValue: undefined,
                description: ts.Diagnostics.Specify_the_output_directory_for_generated_declaration_files,
                defaultValueDescription: "n/a"
            },
            {
                name: "skipLibCheck",
                type: "boolean",
                category: ts.Diagnostics.Completeness,
                description: ts.Diagnostics.Skip_type_checking_all_d_ts_files,
                defaultValueDescription: "false"
            },
            {
                name: "allowUnusedLabels",
                type: "boolean",
                affectsBindDiagnostics: true,
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Disable_error_reporting_for_unused_labels,
                defaultValueDescription: "undefined"
            },
            {
                name: "allowUnreachableCode",
                type: "boolean",
                affectsBindDiagnostics: true,
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Type_Checking,
                description: ts.Diagnostics.Disable_error_reporting_for_unreachable_code,
                defaultValueDescription: "undefined"
            },
            {
                name: "suppressExcessPropertyErrors",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Backwards_Compatibility,
                description: ts.Diagnostics.Disable_reporting_of_excess_property_errors_during_the_creation_of_object_literals,
                defaultValueDescription: "false"
            },
            {
                name: "suppressImplicitAnyIndexErrors",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Backwards_Compatibility,
                description: ts.Diagnostics.Suppress_noImplicitAny_errors_when_indexing_objects_that_lack_index_signatures,
                defaultValueDescription: "false"
            },
            {
                name: "forceConsistentCasingInFileNames",
                type: "boolean",
                affectsModuleResolution: true,
                category: ts.Diagnostics.Interop_Constraints,
                description: ts.Diagnostics.Ensure_that_casing_is_correct_in_imports,
                defaultValueDescription: "false"
            },
            {
                name: "maxNodeModuleJsDepth",
                type: "number",
                affectsModuleResolution: true,
                category: ts.Diagnostics.JavaScript_Support,
                description: ts.Diagnostics.Specify_the_maximum_folder_depth_used_for_checking_JavaScript_files_from_node_modules_Only_applicable_with_allowJs,
                defaultValueDescription: "0"
            },
            {
                name: "noStrictGenericChecks",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                category: ts.Diagnostics.Backwards_Compatibility,
                description: ts.Diagnostics.Disable_strict_checking_of_generic_signatures_in_function_types,
                defaultValueDescription: "false"
            },
            {
                name: "useDefineForClassFields",
                type: "boolean",
                affectsSemanticDiagnostics: true,
                affectsEmit: true,
                category: ts.Diagnostics.Language_and_Environment,
                description: ts.Diagnostics.Emit_ECMAScript_standard_compliant_class_fields,
                defaultValueDescription: "false"
            },
            {
                name: "keyofStringsOnly",
                type: "boolean",
                category: ts.Diagnostics.Backwards_Compatibility,
                description: ts.Diagnostics.Make_keyof_only_return_strings_instead_of_string_numbers_or_symbols_Legacy_option,
                defaultValueDescription: "false"
            },
            {
                // A list of plugins to load in the language service
                name: "plugins",
                type: "list",
                isTSConfigOnly: true,
                element: {
                    name: "plugin",
                    type: "object"
                },
                description: ts.Diagnostics.List_of_language_service_plugins,
                category: ts.Diagnostics.Editor_Support,
            },
        ];
        /* @internal */
        ts.optionDeclarations = __spreadArray(__spreadArray([], ts.commonOptionsWithBuild, true), commandOptionsWithoutBuild, true);
        /* @internal */
        ts.semanticDiagnosticsOptionDeclarations = ts.optionDeclarations.filter(function (option) { return !!option.affectsSemanticDiagnostics; });
        /* @internal */
        ts.affectsEmitOptionDeclarations = ts.optionDeclarations.filter(function (option) { return !!option.affectsEmit; });
        /* @internal */
        ts.moduleResolutionOptionDeclarations = ts.optionDeclarations.filter(function (option) { return !!option.affectsModuleResolution; });
        /* @internal */
        ts.sourceFileAffectingCompilerOptions = ts.optionDeclarations.filter(function (option) {
            return !!option.affectsSourceFile || !!option.affectsModuleResolution || !!option.affectsBindDiagnostics;
        });
        /* @internal */
        ts.optionsAffectingProgramStructure = ts.optionDeclarations.filter(function (option) { return !!option.affectsProgramStructure; });
        /* @internal */
        ts.transpileOptionValueCompilerOptions = ts.optionDeclarations.filter(function (option) {
            return ts.hasProperty(option, "transpileOptionValue");
        });
        // Build related options
        /* @internal */
        ts.optionsForBuild = [
            {
                name: "verbose",
                shortName: "v",
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Enable_verbose_logging,
                type: "boolean"
            },
            {
                name: "dry",
                shortName: "d",
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Show_what_would_be_built_or_deleted_if_specified_with_clean,
                type: "boolean"
            },
            {
                name: "force",
                shortName: "f",
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Build_all_projects_including_those_that_appear_to_be_up_to_date,
                type: "boolean"
            },
            {
                name: "clean",
                category: ts.Diagnostics.Command_line_Options,
                description: ts.Diagnostics.Delete_the_outputs_of_all_projects,
                type: "boolean"
            }
        ];
        /* @internal */
        ts.buildOpts = __spreadArray(__spreadArray([], ts.commonOptionsWithBuild, true), ts.optionsForBuild, true);
        /* @internal */
        ts.typeAcquisitionDeclarations = [
            {
                /* @deprecated typingOptions.enableAutoDiscovery
                 * Use typeAcquisition.enable instead.
                 */
                name: "enableAutoDiscovery",
                type: "boolean",
            },
            {
                name: "enable",
                type: "boolean",
            },
            {
                name: "include",
                type: "list",
                element: {
                    name: "include",
                    type: "string"
                }
            },
            {
                name: "exclude",
                type: "list",
                element: {
                    name: "exclude",
                    type: "string"
                }
            },
            {
                name: "disableFilenameBasedTypeAcquisition",
                type: "boolean",
            },
        ];
        /*@internal*/
        function createOptionNameMap(optionDeclarations) {
            var optionsNameMap = new ts.Map();
            var shortOptionNames = new ts.Map();
            ts.forEach(optionDeclarations, function (option) {
                optionsNameMap.set(option.name.toLowerCase(), option);
                if (option.shortName) {
                    shortOptionNames.set(option.shortName, option.name);
                }
            });
            return { optionsNameMap: optionsNameMap, shortOptionNames: shortOptionNames };
        }
        ts.createOptionNameMap = createOptionNameMap;
        var optionsNameMapCache;
        /* @internal */
        function getOptionsNameMap() {
            return optionsNameMapCache || (optionsNameMapCache = createOptionNameMap(ts.optionDeclarations));
        }
        ts.getOptionsNameMap = getOptionsNameMap;
        var compilerOptionsAlternateMode = {
            diagnostic: ts.Diagnostics.Compiler_option_0_may_only_be_used_with_build,
            getOptionsNameMap: getBuildOptionsNameMap
        };
        /* @internal */
        ts.defaultInitCompilerOptions = {
            module: ts.ModuleKind.CommonJS,
            target: 1 /* ES5 */,
            strict: true,
            esModuleInterop: true,
            forceConsistentCasingInFileNames: true,
            skipLibCheck: true
        };
        /* @internal */
        function convertEnableAutoDiscoveryToEnable(typeAcquisition) {
            // Convert deprecated typingOptions.enableAutoDiscovery to typeAcquisition.enable
            if (typeAcquisition && typeAcquisition.enableAutoDiscovery !== undefined && typeAcquisition.enable === undefined) {
                return {
                    enable: typeAcquisition.enableAutoDiscovery,
                    include: typeAcquisition.include || [],
                    exclude: typeAcquisition.exclude || []
                };
            }
            return typeAcquisition;
        }
        ts.convertEnableAutoDiscoveryToEnable = convertEnableAutoDiscoveryToEnable;
        /* @internal */
        function createCompilerDiagnosticForInvalidCustomType(opt) {
            return createDiagnosticForInvalidCustomType(opt, ts.createCompilerDiagnostic);
        }
        ts.createCompilerDiagnosticForInvalidCustomType = createCompilerDiagnosticForInvalidCustomType;
        function createDiagnosticForInvalidCustomType(opt, createDiagnostic) {
            var namesOfType = ts.arrayFrom(opt.type.keys()).map(function (key) { return "'" + key + "'"; }).join(", ");
            return createDiagnostic(ts.Diagnostics.Argument_for_0_option_must_be_Colon_1, "--" + opt.name, namesOfType);
        }
        /* @internal */
        function parseCustomTypeOption(opt, value, errors) {
            return convertJsonOptionOfCustomType(opt, ts.trimString(value || ""), errors);
        }
        ts.parseCustomTypeOption = parseCustomTypeOption;
        /* @internal */
        function parseListTypeOption(opt, value, errors) {
            if (value === void 0) {
                value = "";
            }
            value = ts.trimString(value);
            if (ts.startsWith(value, "-")) {
                return undefined;
            }
            if (value === "") {
                return [];
            }
            var values = value.split(",");
            switch (opt.element.type) {
                case "number":
                    return ts.mapDefined(values, function (v) { return validateJsonOptionValue(opt.element, parseInt(v), errors); });
                case "string":
                    return ts.mapDefined(values, function (v) { return validateJsonOptionValue(opt.element, v || "", errors); });
                default:
                    return ts.mapDefined(values, function (v) { return parseCustomTypeOption(opt.element, v, errors); });
            }
        }
        ts.parseListTypeOption = parseListTypeOption;
        function getOptionName(option) {
            return option.name;
        }
        function createUnknownOptionError(unknownOption, diagnostics, createDiagnostics, unknownOptionErrorText) {
            var _a;
            if ((_a = diagnostics.alternateMode) === null || _a === void 0 ? void 0 : _a.getOptionsNameMap().optionsNameMap.has(unknownOption.toLowerCase())) {
                return createDiagnostics(diagnostics.alternateMode.diagnostic, unknownOption);
            }
            var possibleOption = ts.getSpellingSuggestion(unknownOption, diagnostics.optionDeclarations, getOptionName);
            return possibleOption ?
                createDiagnostics(diagnostics.unknownDidYouMeanDiagnostic, unknownOptionErrorText || unknownOption, possibleOption.name) :
                createDiagnostics(diagnostics.unknownOptionDiagnostic, unknownOptionErrorText || unknownOption);
        }
        /*@internal*/
        function parseCommandLineWorker(diagnostics, commandLine, readFile) {
            var options = {};
            var watchOptions;
            var fileNames = [];
            var errors = [];
            parseStrings(commandLine);
            return {
                options: options,
                watchOptions: watchOptions,
                fileNames: fileNames,
                errors: errors
            };
            function parseStrings(args) {
                var i = 0;
                while (i < args.length) {
                    var s = args[i];
                    i++;
                    if (s.charCodeAt(0) === 64 /* at */) {
                        parseResponseFile(s.slice(1));
                    }
                    else if (s.charCodeAt(0) === 45 /* minus */) {
                        var inputOptionName = s.slice(s.charCodeAt(1) === 45 /* minus */ ? 2 : 1);
                        var opt = getOptionDeclarationFromName(diagnostics.getOptionsNameMap, inputOptionName, /*allowShort*/ true);
                        if (opt) {
                            i = parseOptionValue(args, i, diagnostics, opt, options, errors);
                        }
                        else {
                            var watchOpt = getOptionDeclarationFromName(watchOptionsDidYouMeanDiagnostics.getOptionsNameMap, inputOptionName, /*allowShort*/ true);
                            if (watchOpt) {
                                i = parseOptionValue(args, i, watchOptionsDidYouMeanDiagnostics, watchOpt, watchOptions || (watchOptions = {}), errors);
                            }
                            else {
                                errors.push(createUnknownOptionError(inputOptionName, diagnostics, ts.createCompilerDiagnostic, s));
                            }
                        }
                    }
                    else {
                        fileNames.push(s);
                    }
                }
            }
            function parseResponseFile(fileName) {
                var text = tryReadFile(fileName, readFile || (function (fileName) { return ts.sys.readFile(fileName); }));
                if (!ts.isString(text)) {
                    errors.push(text);
                    return;
                }
                var args = [];
                var pos = 0;
                while (true) {
                    while (pos < text.length && text.charCodeAt(pos) <= 32 /* space */)
                        pos++;
                    if (pos >= text.length)
                        break;
                    var start = pos;
                    if (text.charCodeAt(start) === 34 /* doubleQuote */) {
                        pos++;
                        while (pos < text.length && text.charCodeAt(pos) !== 34 /* doubleQuote */)
                            pos++;
                        if (pos < text.length) {
                            args.push(text.substring(start + 1, pos));
                            pos++;
                        }
                        else {
                            errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Unterminated_quoted_string_in_response_file_0, fileName));
                        }
                    }
                    else {
                        while (text.charCodeAt(pos) > 32 /* space */)
                            pos++;
                        args.push(text.substring(start, pos));
                    }
                }
                parseStrings(args);
            }
        }
        ts.parseCommandLineWorker = parseCommandLineWorker;
        function parseOptionValue(args, i, diagnostics, opt, options, errors) {
            if (opt.isTSConfigOnly) {
                var optValue = args[i];
                if (optValue === "null") {
                    options[opt.name] = undefined;
                    i++;
                }
                else if (opt.type === "boolean") {
                    if (optValue === "false") {
                        options[opt.name] = validateJsonOptionValue(opt, /*value*/ false, errors);
                        i++;
                    }
                    else {
                        if (optValue === "true")
                            i++;
                        errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_false_or_null_on_command_line, opt.name));
                    }
                }
                else {
                    errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Option_0_can_only_be_specified_in_tsconfig_json_file_or_set_to_null_on_command_line, opt.name));
                    if (optValue && !ts.startsWith(optValue, "-"))
                        i++;
                }
            }
            else {
                // Check to see if no argument was provided (e.g. "--locale" is the last command-line argument).
                if (!args[i] && opt.type !== "boolean") {
                    errors.push(ts.createCompilerDiagnostic(diagnostics.optionTypeMismatchDiagnostic, opt.name, getCompilerOptionValueTypeString(opt)));
                }
                if (args[i] !== "null") {
                    switch (opt.type) {
                        case "number":
                            options[opt.name] = validateJsonOptionValue(opt, parseInt(args[i]), errors);
                            i++;
                            break;
                        case "boolean":
                            // boolean flag has optional value true, false, others
                            var optValue = args[i];
                            options[opt.name] = validateJsonOptionValue(opt, optValue !== "false", errors);
                            // consume next argument as boolean flag value
                            if (optValue === "false" || optValue === "true") {
                                i++;
                            }
                            break;
                        case "string":
                            options[opt.name] = validateJsonOptionValue(opt, args[i] || "", errors);
                            i++;
                            break;
                        case "list":
                            var result = parseListTypeOption(opt, args[i], errors);
                            options[opt.name] = result || [];
                            if (result) {
                                i++;
                            }
                            break;
                        // If not a primitive, the possible types are specified in what is effectively a map of options.
                        default:
                            options[opt.name] = parseCustomTypeOption(opt, args[i], errors);
                            i++;
                            break;
                    }
                }
                else {
                    options[opt.name] = undefined;
                    i++;
                }
            }
            return i;
        }
        /*@internal*/
        ts.compilerOptionsDidYouMeanDiagnostics = {
            alternateMode: compilerOptionsAlternateMode,
            getOptionsNameMap: getOptionsNameMap,
            optionDeclarations: ts.optionDeclarations,
            unknownOptionDiagnostic: ts.Diagnostics.Unknown_compiler_option_0,
            unknownDidYouMeanDiagnostic: ts.Diagnostics.Unknown_compiler_option_0_Did_you_mean_1,
            optionTypeMismatchDiagnostic: ts.Diagnostics.Compiler_option_0_expects_an_argument
        };
        function parseCommandLine(commandLine, readFile) {
            return parseCommandLineWorker(ts.compilerOptionsDidYouMeanDiagnostics, commandLine, readFile);
        }
        ts.parseCommandLine = parseCommandLine;
        /** @internal */
        function getOptionFromName(optionName, allowShort) {
            return getOptionDeclarationFromName(getOptionsNameMap, optionName, allowShort);
        }
        ts.getOptionFromName = getOptionFromName;
        function getOptionDeclarationFromName(getOptionNameMap, optionName, allowShort) {
            if (allowShort === void 0) {
                allowShort = false;
            }
            optionName = optionName.toLowerCase();
            var _a = getOptionNameMap(), optionsNameMap = _a.optionsNameMap, shortOptionNames = _a.shortOptionNames;
            // Try to translate short option names to their full equivalents.
            if (allowShort) {
                var short = shortOptionNames.get(optionName);
                if (short !== undefined) {
                    optionName = short;
                }
            }
            return optionsNameMap.get(optionName);
        }
        var buildOptionsNameMapCache;
        function getBuildOptionsNameMap() {
            return buildOptionsNameMapCache || (buildOptionsNameMapCache = createOptionNameMap(ts.buildOpts));
        }
        var buildOptionsAlternateMode = {
            diagnostic: ts.Diagnostics.Compiler_option_0_may_not_be_used_with_build,
            getOptionsNameMap: getOptionsNameMap
        };
        var buildOptionsDidYouMeanDiagnostics = {
            alternateMode: buildOptionsAlternateMode,
            getOptionsNameMap: getBuildOptionsNameMap,
            optionDeclarations: ts.buildOpts,
            unknownOptionDiagnostic: ts.Diagnostics.Unknown_build_option_0,
            unknownDidYouMeanDiagnostic: ts.Diagnostics.Unknown_build_option_0_Did_you_mean_1,
            optionTypeMismatchDiagnostic: ts.Diagnostics.Build_option_0_requires_a_value_of_type_1
        };
        /*@internal*/
        function parseBuildCommand(args) {
            var _a = parseCommandLineWorker(buildOptionsDidYouMeanDiagnostics, args), options = _a.options, watchOptions = _a.watchOptions, projects = _a.fileNames, errors = _a.errors;
            var buildOptions = options;
            if (projects.length === 0) {
                // tsc -b invoked with no extra arguments; act as if invoked with "tsc -b ."
                projects.push(".");
            }
            // Nonsensical combinations
            if (buildOptions.clean && buildOptions.force) {
                errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Options_0_and_1_cannot_be_combined, "clean", "force"));
            }
            if (buildOptions.clean && buildOptions.verbose) {
                errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Options_0_and_1_cannot_be_combined, "clean", "verbose"));
            }
            if (buildOptions.clean && buildOptions.watch) {
                errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Options_0_and_1_cannot_be_combined, "clean", "watch"));
            }
            if (buildOptions.watch && buildOptions.dry) {
                errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Options_0_and_1_cannot_be_combined, "watch", "dry"));
            }
            return { buildOptions: buildOptions, watchOptions: watchOptions, projects: projects, errors: errors };
        }
        ts.parseBuildCommand = parseBuildCommand;
        /* @internal */
        function getDiagnosticText(_message) {
            var _args = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                _args[_i - 1] = arguments[_i];
            }
            var diagnostic = ts.createCompilerDiagnostic.apply(undefined, arguments);
            return diagnostic.messageText;
        }
        ts.getDiagnosticText = getDiagnosticText;
        /**
         * Reads the config file, reports errors if any and exits if the config file cannot be found
         */
        function getParsedCommandLineOfConfigFile(configFileName, optionsToExtend, host, extendedConfigCache, watchOptionsToExtend, extraFileExtensions) {
            var configFileText = tryReadFile(configFileName, function (fileName) { return host.readFile(fileName); });
            if (!ts.isString(configFileText)) {
                host.onUnRecoverableConfigFileDiagnostic(configFileText);
                return undefined;
            }
            var result = ts.parseJsonText(configFileName, configFileText);
            var cwd = host.getCurrentDirectory();
            result.path = ts.toPath(configFileName, cwd, ts.createGetCanonicalFileName(host.useCaseSensitiveFileNames));
            result.resolvedPath = result.path;
            result.originalFileName = result.fileName;
            return parseJsonSourceFileConfigFileContent(result, host, ts.getNormalizedAbsolutePath(ts.getDirectoryPath(configFileName), cwd), optionsToExtend, ts.getNormalizedAbsolutePath(configFileName, cwd), 
            /*resolutionStack*/ undefined, extraFileExtensions, extendedConfigCache, watchOptionsToExtend);
        }
        ts.getParsedCommandLineOfConfigFile = getParsedCommandLineOfConfigFile;
        /**
         * Read tsconfig.json file
         * @param fileName The path to the config file
         */
        function readConfigFile(fileName, readFile) {
            var textOrDiagnostic = tryReadFile(fileName, readFile);
            return ts.isString(textOrDiagnostic) ? parseConfigFileTextToJson(fileName, textOrDiagnostic) : { config: {}, error: textOrDiagnostic };
        }
        ts.readConfigFile = readConfigFile;
        /**
         * Parse the text of the tsconfig.json file
         * @param fileName The path to the config file
         * @param jsonText The text of the config file
         */
        function parseConfigFileTextToJson(fileName, jsonText) {
            var jsonSourceFile = ts.parseJsonText(fileName, jsonText);
            return {
                config: convertConfigFileToObject(jsonSourceFile, jsonSourceFile.parseDiagnostics, /*reportOptionsErrors*/ false, /*optionsIterator*/ undefined),
                error: jsonSourceFile.parseDiagnostics.length ? jsonSourceFile.parseDiagnostics[0] : undefined
            };
        }
        ts.parseConfigFileTextToJson = parseConfigFileTextToJson;
        /**
         * Read tsconfig.json file
         * @param fileName The path to the config file
         */
        function readJsonConfigFile(fileName, readFile) {
            var textOrDiagnostic = tryReadFile(fileName, readFile);
            return ts.isString(textOrDiagnostic) ? ts.parseJsonText(fileName, textOrDiagnostic) : { fileName: fileName, parseDiagnostics: [textOrDiagnostic] };
        }
        ts.readJsonConfigFile = readJsonConfigFile;
        /*@internal*/
        function tryReadFile(fileName, readFile) {
            var text;
            try {
                text = readFile(fileName);
            }
            catch (e) {
                return ts.createCompilerDiagnostic(ts.Diagnostics.Cannot_read_file_0_Colon_1, fileName, e.message);
            }
            return text === undefined ? ts.createCompilerDiagnostic(ts.Diagnostics.Cannot_read_file_0, fileName) : text;
        }
        ts.tryReadFile = tryReadFile;
        function commandLineOptionsToMap(options) {
            return ts.arrayToMap(options, getOptionName);
        }
        var typeAcquisitionDidYouMeanDiagnostics = {
            optionDeclarations: ts.typeAcquisitionDeclarations,
            unknownOptionDiagnostic: ts.Diagnostics.Unknown_type_acquisition_option_0,
            unknownDidYouMeanDiagnostic: ts.Diagnostics.Unknown_type_acquisition_option_0_Did_you_mean_1,
        };
        var watchOptionsNameMapCache;
        function getWatchOptionsNameMap() {
            return watchOptionsNameMapCache || (watchOptionsNameMapCache = createOptionNameMap(ts.optionsForWatch));
        }
        var watchOptionsDidYouMeanDiagnostics = {
            getOptionsNameMap: getWatchOptionsNameMap,
            optionDeclarations: ts.optionsForWatch,
            unknownOptionDiagnostic: ts.Diagnostics.Unknown_watch_option_0,
            unknownDidYouMeanDiagnostic: ts.Diagnostics.Unknown_watch_option_0_Did_you_mean_1,
            optionTypeMismatchDiagnostic: ts.Diagnostics.Watch_option_0_requires_a_value_of_type_1
        };
        var commandLineCompilerOptionsMapCache;
        function getCommandLineCompilerOptionsMap() {
            return commandLineCompilerOptionsMapCache || (commandLineCompilerOptionsMapCache = commandLineOptionsToMap(ts.optionDeclarations));
        }
        var commandLineWatchOptionsMapCache;
        function getCommandLineWatchOptionsMap() {
            return commandLineWatchOptionsMapCache || (commandLineWatchOptionsMapCache = commandLineOptionsToMap(ts.optionsForWatch));
        }
        var commandLineTypeAcquisitionMapCache;
        function getCommandLineTypeAcquisitionMap() {
            return commandLineTypeAcquisitionMapCache || (commandLineTypeAcquisitionMapCache = commandLineOptionsToMap(ts.typeAcquisitionDeclarations));
        }
        var _tsconfigRootOptions;
        function getTsconfigRootOptionsMap() {
            if (_tsconfigRootOptions === undefined) {
                _tsconfigRootOptions = {
                    name: undefined,
                    type: "object",
                    elementOptions: commandLineOptionsToMap([
                        {
                            name: "compilerOptions",
                            type: "object",
                            elementOptions: getCommandLineCompilerOptionsMap(),
                            extraKeyDiagnostics: ts.compilerOptionsDidYouMeanDiagnostics,
                        },
                        {
                            name: "watchOptions",
                            type: "object",
                            elementOptions: getCommandLineWatchOptionsMap(),
                            extraKeyDiagnostics: watchOptionsDidYouMeanDiagnostics,
                        },
                        {
                            name: "typingOptions",
                            type: "object",
                            elementOptions: getCommandLineTypeAcquisitionMap(),
                            extraKeyDiagnostics: typeAcquisitionDidYouMeanDiagnostics,
                        },
                        {
                            name: "typeAcquisition",
                            type: "object",
                            elementOptions: getCommandLineTypeAcquisitionMap(),
                            extraKeyDiagnostics: typeAcquisitionDidYouMeanDiagnostics
                        },
                        {
                            name: "extends",
                            type: "string",
                            category: ts.Diagnostics.File_Management,
                        },
                        {
                            name: "references",
                            type: "list",
                            element: {
                                name: "references",
                                type: "object"
                            },
                            category: ts.Diagnostics.Projects,
                        },
                        {
                            name: "files",
                            type: "list",
                            element: {
                                name: "files",
                                type: "string"
                            },
                            category: ts.Diagnostics.File_Management,
                        },
                        {
                            name: "include",
                            type: "list",
                            element: {
                                name: "include",
                                type: "string"
                            },
                            category: ts.Diagnostics.File_Management,
                            defaultValueDescription: ts.Diagnostics.if_files_is_specified_otherwise_Asterisk_Asterisk_Slash_Asterisk
                        },
                        {
                            name: "exclude",
                            type: "list",
                            element: {
                                name: "exclude",
                                type: "string"
                            },
                            category: ts.Diagnostics.File_Management,
                            defaultValueDescription: ts.Diagnostics.node_modules_bower_components_jspm_packages_plus_the_value_of_outDir_if_one_is_specified
                        },
                        ts.compileOnSaveCommandLineOption
                    ])
                };
            }
            return _tsconfigRootOptions;
        }
        function convertConfigFileToObject(sourceFile, errors, reportOptionsErrors, optionsIterator) {
            var _a;
            var rootExpression = (_a = sourceFile.statements[0]) === null || _a === void 0 ? void 0 : _a.expression;
            var knownRootOptions = reportOptionsErrors ? getTsconfigRootOptionsMap() : undefined;
            if (rootExpression && rootExpression.kind !== 203 /* ObjectLiteralExpression */) {
                errors.push(ts.createDiagnosticForNodeInSourceFile(sourceFile, rootExpression, ts.Diagnostics.The_root_value_of_a_0_file_must_be_an_object, ts.getBaseFileName(sourceFile.fileName) === "jsconfig.json" ? "jsconfig.json" : "tsconfig.json"));
                // Last-ditch error recovery. Somewhat useful because the JSON parser will recover from some parse errors by
                // synthesizing a top-level array literal expression. There's a reasonable chance the first element of that
                // array is a well-formed configuration object, made into an array element by stray characters.
                if (ts.isArrayLiteralExpression(rootExpression)) {
                    var firstObject = ts.find(rootExpression.elements, ts.isObjectLiteralExpression);
                    if (firstObject) {
                        return convertToObjectWorker(sourceFile, firstObject, errors, /*returnValue*/ true, knownRootOptions, optionsIterator);
                    }
                }
                return {};
            }
            return convertToObjectWorker(sourceFile, rootExpression, errors, /*returnValue*/ true, knownRootOptions, optionsIterator);
        }
        /**
         * Convert the json syntax tree into the json value
         */
        function convertToObject(sourceFile, errors) {
            var _a;
            return convertToObjectWorker(sourceFile, (_a = sourceFile.statements[0]) === null || _a === void 0 ? void 0 : _a.expression, errors, /*returnValue*/ true, /*knownRootOptions*/ undefined, /*jsonConversionNotifier*/ undefined);
        }
        ts.convertToObject = convertToObject;
        /**
         * Convert the json syntax tree into the json value and report errors
         * This returns the json value (apart from checking errors) only if returnValue provided is true.
         * Otherwise it just checks the errors and returns undefined
         */
        /*@internal*/
        function convertToObjectWorker(sourceFile, rootExpression, errors, returnValue, knownRootOptions, jsonConversionNotifier) {
            if (!rootExpression) {
                return returnValue ? {} : undefined;
            }
            return convertPropertyValueToJson(rootExpression, knownRootOptions);
            function isRootOptionMap(knownOptions) {
                return knownRootOptions && knownRootOptions.elementOptions === knownOptions;
            }
            function convertObjectLiteralExpressionToJson(node, knownOptions, extraKeyDiagnostics, parentOption) {
                var result = returnValue ? {} : undefined;
                var _loop_4 = function (element) {
                    if (element.kind !== 291 /* PropertyAssignment */) {
                        errors.push(ts.createDiagnosticForNodeInSourceFile(sourceFile, element, ts.Diagnostics.Property_assignment_expected));
                        return "continue";
                    }
                    if (element.questionToken) {
                        errors.push(ts.createDiagnosticForNodeInSourceFile(sourceFile, element.questionToken, ts.Diagnostics.The_0_modifier_can_only_be_used_in_TypeScript_files, "?"));
                    }
                    if (!isDoubleQuotedString(element.name)) {
                        errors.push(ts.createDiagnosticForNodeInSourceFile(sourceFile, element.name, ts.Diagnostics.String_literal_with_double_quotes_expected));
                    }
                    var textOfKey = ts.isComputedNonLiteralName(element.name) ? undefined : ts.getTextOfPropertyName(element.name);
                    var keyText = textOfKey && ts.unescapeLeadingUnderscores(textOfKey);
                    var option = keyText && knownOptions ? knownOptions.get(keyText) : undefined;
                    if (keyText && extraKeyDiagnostics && !option) {
                        if (knownOptions) {
                            errors.push(createUnknownOptionError(keyText, extraKeyDiagnostics, function (message, arg0, arg1) { return ts.createDiagnosticForNodeInSourceFile(sourceFile, element.name, message, arg0, arg1); }));
                        }
                        else {
                            errors.push(ts.createDiagnosticForNodeInSourceFile(sourceFile, element.name, extraKeyDiagnostics.unknownOptionDiagnostic, keyText));
                        }
                    }
                    var value = convertPropertyValueToJson(element.initializer, option);
                    if (typeof keyText !== "undefined") {
                        if (returnValue) {
                            result[keyText] = value;
                        }
                        // Notify key value set, if user asked for it
                        if (jsonConversionNotifier &&
                            // Current callbacks are only on known parent option or if we are setting values in the root
                            (parentOption || isRootOptionMap(knownOptions))) {
                            var isValidOptionValue = isCompilerOptionsValue(option, value);
                            if (parentOption) {
                                if (isValidOptionValue) {
                                    // Notify option set in the parent if its a valid option value
                                    jsonConversionNotifier.onSetValidOptionKeyValueInParent(parentOption, option, value);
                                }
                            }
                            else if (isRootOptionMap(knownOptions)) {
                                if (isValidOptionValue) {
                                    // Notify about the valid root key value being set
                                    jsonConversionNotifier.onSetValidOptionKeyValueInRoot(keyText, element.name, value, element.initializer);
                                }
                                else if (!option) {
                                    // Notify about the unknown root key value being set
                                    jsonConversionNotifier.onSetUnknownOptionKeyValueInRoot(keyText, element.name, value, element.initializer);
                                }
                            }
                        }
                    }
                };
                for (var _i = 0, _a = node.properties; _i < _a.length; _i++) {
                    var element = _a[_i];
                    _loop_4(element);
                }
                return result;
            }
            function convertArrayLiteralExpressionToJson(elements, elementOption) {
                if (!returnValue) {
                    elements.forEach(function (element) { return convertPropertyValueToJson(element, elementOption); });
                    return undefined;
                }
                // Filter out invalid values
                return ts.filter(elements.map(function (element) { return convertPropertyValueToJson(element, elementOption); }), function (v) { return v !== undefined; });
            }
            function convertPropertyValueToJson(valueExpression, option) {
                var invalidReported;
                switch (valueExpression.kind) {
                    case 110 /* TrueKeyword */:
                        reportInvalidOptionValue(option && option.type !== "boolean");
                        return validateValue(/*value*/ true);
                    case 95 /* FalseKeyword */:
                        reportInvalidOptionValue(option && option.type !== "boolean");
                        return validateValue(/*value*/ false);
                    case 104 /* NullKeyword */:
                        reportInvalidOptionValue(option && option.name === "extends"); // "extends" is the only option we don't allow null/undefined for
                        return validateValue(/*value*/ null); // eslint-disable-line no-null/no-null
                    case 10 /* StringLiteral */:
                        if (!isDoubleQuotedString(valueExpression)) {
                            errors.push(ts.createDiagnosticForNodeInSourceFile(sourceFile, valueExpression, ts.Diagnostics.String_literal_with_double_quotes_expected));
                        }
                        reportInvalidOptionValue(option && (ts.isString(option.type) && option.type !== "string"));
                        var text = valueExpression.text;
                        if (option && !ts.isString(option.type)) {
                            var customOption = option;
                            // Validate custom option type
                            if (!customOption.type.has(text.toLowerCase())) {
                                errors.push(createDiagnosticForInvalidCustomType(customOption, function (message, arg0, arg1) { return ts.createDiagnosticForNodeInSourceFile(sourceFile, valueExpression, message, arg0, arg1); }));
                                invalidReported = true;
                            }
                        }
                        return validateValue(text);
                    case 8 /* NumericLiteral */:
                        reportInvalidOptionValue(option && option.type !== "number");
                        return validateValue(Number(valueExpression.text));
                    case 217 /* PrefixUnaryExpression */:
                        if (valueExpression.operator !== 40 /* MinusToken */ || valueExpression.operand.kind !== 8 /* NumericLiteral */) {
                            break; // not valid JSON syntax
                        }
                        reportInvalidOptionValue(option && option.type !== "number");
                        return validateValue(-Number(valueExpression.operand.text));
                    case 203 /* ObjectLiteralExpression */:
                        reportInvalidOptionValue(option && option.type !== "object");
                        var objectLiteralExpression = valueExpression;
                        // Currently having element option declaration in the tsconfig with type "object"
                        // determines if it needs onSetValidOptionKeyValueInParent callback or not
                        // At moment there are only "compilerOptions", "typeAcquisition" and "typingOptions"
                        // that satifies it and need it to modify options set in them (for normalizing file paths)
                        // vs what we set in the json
                        // If need arises, we can modify this interface and callbacks as needed
                        if (option) {
                            var _a = option, elementOptions = _a.elementOptions, extraKeyDiagnostics = _a.extraKeyDiagnostics, optionName = _a.name;
                            return validateValue(convertObjectLiteralExpressionToJson(objectLiteralExpression, elementOptions, extraKeyDiagnostics, optionName));
                        }
                        else {
                            return validateValue(convertObjectLiteralExpressionToJson(objectLiteralExpression, /* knownOptions*/ undefined, 
                            /*extraKeyDiagnosticMessage */ undefined, /*parentOption*/ undefined));
                        }
                    case 202 /* ArrayLiteralExpression */:
                        reportInvalidOptionValue(option && option.type !== "list");
                        return validateValue(convertArrayLiteralExpressionToJson(valueExpression.elements, option && option.element));
                }
                // Not in expected format
                if (option) {
                    reportInvalidOptionValue(/*isError*/ true);
                }
                else {
                    errors.push(ts.createDiagnosticForNodeInSourceFile(sourceFile, valueExpression, ts.Diagnostics.Property_value_can_only_be_string_literal_numeric_literal_true_false_null_object_literal_or_array_literal));
                }
                return undefined;
                function validateValue(value) {
                    var _a;
                    if (!invalidReported) {
                        var diagnostic = (_a = option === null || option === void 0 ? void 0 : option.extraValidation) === null || _a === void 0 ? void 0 : _a.call(option, value);
                        if (diagnostic) {
                            errors.push(ts.createDiagnosticForNodeInSourceFile.apply(void 0, __spreadArray([sourceFile, valueExpression], diagnostic, false)));
                            return undefined;
                        }
                    }
                    return value;
                }
                function reportInvalidOptionValue(isError) {
                    if (isError) {
                        errors.push(ts.createDiagnosticForNodeInSourceFile(sourceFile, valueExpression, ts.Diagnostics.Compiler_option_0_requires_a_value_of_type_1, option.name, getCompilerOptionValueTypeString(option)));
                        invalidReported = true;
                    }
                }
            }
            function isDoubleQuotedString(node) {
                return ts.isStringLiteral(node) && ts.isStringDoubleQuoted(node, sourceFile);
            }
        }
        ts.convertToObjectWorker = convertToObjectWorker;
        function getCompilerOptionValueTypeString(option) {
            return option.type === "list" ?
                "Array" :
                ts.isString(option.type) ? option.type : "string";
        }
        function isCompilerOptionsValue(option, value) {
            if (option) {
                if (isNullOrUndefined(value))
                    return true; // All options are undefinable/nullable
                if (option.type === "list") {
                    return ts.isArray(value);
                }
                var expectedType = ts.isString(option.type) ? option.type : "string";
                return typeof value === expectedType;
            }
            return false;
        }
        /**
         * Generate an uncommented, complete tsconfig for use with "--showConfig"
         * @param configParseResult options to be generated into tsconfig.json
         * @param configFileName name of the parsed config file - output paths will be generated relative to this
         * @param host provides current directory and case sensitivity services
         */
        /** @internal */
        function convertToTSConfig(configParseResult, configFileName, host) {
            var _a, _b, _c;
            var getCanonicalFileName = ts.createGetCanonicalFileName(host.useCaseSensitiveFileNames);
            var files = ts.map(ts.filter(configParseResult.fileNames, !((_b = (_a = configParseResult.options.configFile) === null || _a === void 0 ? void 0 : _a.configFileSpecs) === null || _b === void 0 ? void 0 : _b.validatedIncludeSpecs) ? ts.returnTrue : matchesSpecs(configFileName, configParseResult.options.configFile.configFileSpecs.validatedIncludeSpecs, configParseResult.options.configFile.configFileSpecs.validatedExcludeSpecs, host)), function (f) { return ts.getRelativePathFromFile(ts.getNormalizedAbsolutePath(configFileName, host.getCurrentDirectory()), ts.getNormalizedAbsolutePath(f, host.getCurrentDirectory()), getCanonicalFileName); });
            var optionMap = serializeCompilerOptions(configParseResult.options, { configFilePath: ts.getNormalizedAbsolutePath(configFileName, host.getCurrentDirectory()), useCaseSensitiveFileNames: host.useCaseSensitiveFileNames });
            var watchOptionMap = configParseResult.watchOptions && serializeWatchOptions(configParseResult.watchOptions);
            var config = __assign(__assign({ compilerOptions: __assign(__assign({}, optionMapToObject(optionMap)), { showConfig: undefined, configFile: undefined, configFilePath: undefined, help: undefined, init: undefined, listFiles: undefined, listEmittedFiles: undefined, project: undefined, build: undefined, version: undefined }), watchOptions: watchOptionMap && optionMapToObject(watchOptionMap), references: ts.map(configParseResult.projectReferences, function (r) { return (__assign(__assign({}, r), { path: r.originalPath ? r.originalPath : "", originalPath: undefined })); }), files: ts.length(files) ? files : undefined }, (((_c = configParseResult.options.configFile) === null || _c === void 0 ? void 0 : _c.configFileSpecs) ? {
                include: filterSameAsDefaultInclude(configParseResult.options.configFile.configFileSpecs.validatedIncludeSpecs),
                exclude: configParseResult.options.configFile.configFileSpecs.validatedExcludeSpecs
            } : {})), { compileOnSave: !!configParseResult.compileOnSave ? true : undefined });
            return config;
        }
        ts.convertToTSConfig = convertToTSConfig;
        function optionMapToObject(optionMap) {
            return __assign({}, ts.arrayFrom(optionMap.entries()).reduce(function (prev, cur) {
                var _a;
                return (__assign(__assign({}, prev), (_a = {}, _a[cur[0]] = cur[1], _a)));
            }, {}));
        }
        function filterSameAsDefaultInclude(specs) {
            if (!ts.length(specs))
                return undefined;
            if (ts.length(specs) !== 1)
                return specs;
            if (specs[0] === "**/*")
                return undefined;
            return specs;
        }
        function matchesSpecs(path, includeSpecs, excludeSpecs, host) {
            if (!includeSpecs)
                return ts.returnTrue;
            var patterns = ts.getFileMatcherPatterns(path, excludeSpecs, includeSpecs, host.useCaseSensitiveFileNames, host.getCurrentDirectory());
            var excludeRe = patterns.excludePattern && ts.getRegexFromPattern(patterns.excludePattern, host.useCaseSensitiveFileNames);
            var includeRe = patterns.includeFilePattern && ts.getRegexFromPattern(patterns.includeFilePattern, host.useCaseSensitiveFileNames);
            if (includeRe) {
                if (excludeRe) {
                    return function (path) { return !(includeRe.test(path) && !excludeRe.test(path)); };
                }
                return function (path) { return !includeRe.test(path); };
            }
            if (excludeRe) {
                return function (path) { return excludeRe.test(path); };
            }
            return ts.returnTrue;
        }
        function getCustomTypeMapOfCommandLineOption(optionDefinition) {
            if (optionDefinition.type === "string" || optionDefinition.type === "number" || optionDefinition.type === "boolean" || optionDefinition.type === "object") {
                // this is of a type CommandLineOptionOfPrimitiveType
                return undefined;
            }
            else if (optionDefinition.type === "list") {
                return getCustomTypeMapOfCommandLineOption(optionDefinition.element);
            }
            else {
                return optionDefinition.type;
            }
        }
        function getNameOfCompilerOptionValue(value, customTypeMap) {
            // There is a typeMap associated with this command-line option so use it to map value back to its name
            return ts.forEachEntry(customTypeMap, function (mapValue, key) {
                if (mapValue === value) {
                    return key;
                }
            });
        }
        function serializeCompilerOptions(options, pathOptions) {
            return serializeOptionBaseObject(options, getOptionsNameMap(), pathOptions);
        }
        function serializeWatchOptions(options) {
            return serializeOptionBaseObject(options, getWatchOptionsNameMap());
        }
        function serializeOptionBaseObject(options, _a, pathOptions) {
            var optionsNameMap = _a.optionsNameMap;
            var result = new ts.Map();
            var getCanonicalFileName = pathOptions && ts.createGetCanonicalFileName(pathOptions.useCaseSensitiveFileNames);
            var _loop_5 = function (name) {
                if (ts.hasProperty(options, name)) {
                    // tsconfig only options cannot be specified via command line,
                    // so we can assume that only types that can appear here string | number | boolean
                    if (optionsNameMap.has(name) && (optionsNameMap.get(name).category === ts.Diagnostics.Command_line_Options || optionsNameMap.get(name).category === ts.Diagnostics.Output_Formatting)) {
                        return "continue";
                    }
                    var value = options[name];
                    var optionDefinition = optionsNameMap.get(name.toLowerCase());
                    if (optionDefinition) {
                        var customTypeMap_1 = getCustomTypeMapOfCommandLineOption(optionDefinition);
                        if (!customTypeMap_1) {
                            // There is no map associated with this compiler option then use the value as-is
                            // This is the case if the value is expect to be string, number, boolean or list of string
                            if (pathOptions && optionDefinition.isFilePath) {
                                result.set(name, ts.getRelativePathFromFile(pathOptions.configFilePath, ts.getNormalizedAbsolutePath(value, ts.getDirectoryPath(pathOptions.configFilePath)), getCanonicalFileName));
                            }
                            else {
                                result.set(name, value);
                            }
                        }
                        else {
                            if (optionDefinition.type === "list") {
                                result.set(name, value.map(function (element) { return getNameOfCompilerOptionValue(element, customTypeMap_1); })); // TODO: GH#18217
                            }
                            else {
                                // There is a typeMap associated with this command-line option so use it to map value back to its name
                                result.set(name, getNameOfCompilerOptionValue(value, customTypeMap_1));
                            }
                        }
                    }
                }
            };
            for (var name in options) {
                _loop_5(name);
            }
            return result;
        }
        /**
         * Generate tsconfig configuration when running command line "--init"
         * @param options commandlineOptions to be generated into tsconfig.json
         * @param fileNames array of filenames to be generated into tsconfig.json
         */
        /* @internal */
        function generateTSConfig(options, fileNames, newLine) {
            var compilerOptions = ts.extend(options, ts.defaultInitCompilerOptions);
            var compilerOptionsMap = serializeCompilerOptions(compilerOptions);
            return writeConfigurations();
            function getDefaultValueForOption(option) {
                switch (option.type) {
                    case "number":
                        return 1;
                    case "boolean":
                        return true;
                    case "string":
                        return option.isFilePath ? "./" : "";
                    case "list":
                        return [];
                    case "object":
                        return {};
                    default:
                        var iterResult = option.type.keys().next();
                        if (!iterResult.done)
                            return iterResult.value;
                        return ts.Debug.fail("Expected 'option.type' to have entries.");
                }
            }
            function makePadding(paddingLength) {
                return Array(paddingLength + 1).join(" ");
            }
            function isAllowedOptionForOutput(_a) {
                var category = _a.category, name = _a.name, isCommandLineOnly = _a.isCommandLineOnly;
                // Skip options which do not have a category or have categories which are more niche
                var categoriesToSkip = [ts.Diagnostics.Command_line_Options, ts.Diagnostics.Editor_Support, ts.Diagnostics.Compiler_Diagnostics, ts.Diagnostics.Backwards_Compatibility, ts.Diagnostics.Watch_and_Build_Modes, ts.Diagnostics.Output_Formatting];
                return !isCommandLineOnly && category !== undefined && (!categoriesToSkip.includes(category) || compilerOptionsMap.has(name));
            }
            function writeConfigurations() {
                // Filter applicable options to place in the file
                var categorizedOptions = ts.createMultiMap();
                for (var _i = 0, optionDeclarations_1 = ts.optionDeclarations; _i < optionDeclarations_1.length; _i++) {
                    var option = optionDeclarations_1[_i];
                    var category = option.category;
                    if (isAllowedOptionForOutput(option)) {
                        categorizedOptions.add(ts.getLocaleSpecificMessage(category), option);
                    }
                }
                // Serialize all options and their descriptions
                var marginLength = 0;
                var seenKnownKeys = 0;
                var entries = [];
                categorizedOptions.forEach(function (options, category) {
                    if (entries.length !== 0) {
                        entries.push({ value: "" });
                    }
                    entries.push({ value: "/* " + category + " */" });
                    for (var _i = 0, options_1 = options; _i < options_1.length; _i++) {
                        var option = options_1[_i];
                        var optionName = void 0;
                        if (compilerOptionsMap.has(option.name)) {
                            optionName = "\"" + option.name + "\": " + JSON.stringify(compilerOptionsMap.get(option.name)) + ((seenKnownKeys += 1) === compilerOptionsMap.size ? "" : ",");
                        }
                        else {
                            optionName = "// \"" + option.name + "\": " + JSON.stringify(getDefaultValueForOption(option)) + ",";
                        }
                        entries.push({
                            value: optionName,
                            description: "/* " + (option.description && ts.getLocaleSpecificMessage(option.description) || option.name) + " */"
                        });
                        marginLength = Math.max(optionName.length, marginLength);
                    }
                });
                // Write the output
                var tab = makePadding(2);
                var result = [];
                result.push("{");
                result.push(tab + "\"compilerOptions\": {");
                result.push("" + tab + tab + "/* " + ts.getLocaleSpecificMessage(ts.Diagnostics.Visit_https_Colon_Slash_Slashaka_ms_Slashtsconfig_json_to_read_more_about_this_file) + " */");
                result.push("");
                // Print out each row, aligning all the descriptions on the same column.
                for (var _a = 0, entries_2 = entries; _a < entries_2.length; _a++) {
                    var entry = entries_2[_a];
                    var value = entry.value, _b = entry.description, description = _b === void 0 ? "" : _b;
                    result.push(value && "" + tab + tab + value + (description && (makePadding(marginLength - value.length + 2) + description)));
                }
                if (fileNames.length) {
                    result.push(tab + "},");
                    result.push(tab + "\"files\": [");
                    for (var i = 0; i < fileNames.length; i++) {
                        result.push("" + tab + tab + JSON.stringify(fileNames[i]) + (i === fileNames.length - 1 ? "" : ","));
                    }
                    result.push(tab + "]");
                }
                else {
                    result.push(tab + "}");
                }
                result.push("}");
                return result.join(newLine) + newLine;
            }
        }
        ts.generateTSConfig = generateTSConfig;
        /* @internal */
        function convertToOptionsWithAbsolutePaths(options, toAbsolutePath) {
            var result = {};
            var optionsNameMap = getOptionsNameMap().optionsNameMap;
            for (var name in options) {
                if (ts.hasProperty(options, name)) {
                    result[name] = convertToOptionValueWithAbsolutePaths(optionsNameMap.get(name.toLowerCase()), options[name], toAbsolutePath);
                }
            }
            if (result.configFilePath) {
                result.configFilePath = toAbsolutePath(result.configFilePath);
            }
            return result;
        }
        ts.convertToOptionsWithAbsolutePaths = convertToOptionsWithAbsolutePaths;
        function convertToOptionValueWithAbsolutePaths(option, value, toAbsolutePath) {
            if (option && !isNullOrUndefined(value)) {
                if (option.type === "list") {
                    var values = value;
                    if (option.element.isFilePath && values.length) {
                        return values.map(toAbsolutePath);
                    }
                }
                else if (option.isFilePath) {
                    return toAbsolutePath(value);
                }
            }
            return value;
        }
        /**
         * Parse the contents of a config file (tsconfig.json).
         * @param json The contents of the config file to parse
         * @param host Instance of ParseConfigHost used to enumerate files in folder.
         * @param basePath A root directory to resolve relative path entries in the config
         *    file to. e.g. outDir
         */
        function parseJsonConfigFileContent(json, host, basePath, existingOptions, configFileName, resolutionStack, extraFileExtensions, extendedConfigCache, existingWatchOptions) {
            return parseJsonConfigFileContentWorker(json, /*sourceFile*/ undefined, host, basePath, existingOptions, existingWatchOptions, configFileName, resolutionStack, extraFileExtensions, extendedConfigCache);
        }
        ts.parseJsonConfigFileContent = parseJsonConfigFileContent;
        /**
         * Parse the contents of a config file (tsconfig.json).
         * @param jsonNode The contents of the config file to parse
         * @param host Instance of ParseConfigHost used to enumerate files in folder.
         * @param basePath A root directory to resolve relative path entries in the config
         *    file to. e.g. outDir
         */
        function parseJsonSourceFileConfigFileContent(sourceFile, host, basePath, existingOptions, configFileName, resolutionStack, extraFileExtensions, extendedConfigCache, existingWatchOptions) {
            return parseJsonConfigFileContentWorker(/*json*/ undefined, sourceFile, host, basePath, existingOptions, existingWatchOptions, configFileName, resolutionStack, extraFileExtensions, extendedConfigCache);
        }
        ts.parseJsonSourceFileConfigFileContent = parseJsonSourceFileConfigFileContent;
        /*@internal*/
        function setConfigFileInOptions(options, configFile) {
            if (configFile) {
                Object.defineProperty(options, "configFile", { enumerable: false, writable: false, value: configFile });
            }
        }
        ts.setConfigFileInOptions = setConfigFileInOptions;
        function isNullOrUndefined(x) {
            return x === undefined || x === null; // eslint-disable-line no-null/no-null
        }
        function directoryOfCombinedPath(fileName, basePath) {
            // Use the `getNormalizedAbsolutePath` function to avoid canonicalizing the path, as it must remain noncanonical
            // until consistent casing errors are reported
            return ts.getDirectoryPath(ts.getNormalizedAbsolutePath(fileName, basePath));
        }
        /**
         * Parse the contents of a config file from json or json source file (tsconfig.json).
         * @param json The contents of the config file to parse
         * @param sourceFile sourceFile corresponding to the Json
         * @param host Instance of ParseConfigHost used to enumerate files in folder.
         * @param basePath A root directory to resolve relative path entries in the config
         *    file to. e.g. outDir
         * @param resolutionStack Only present for backwards-compatibility. Should be empty.
         */
        function parseJsonConfigFileContentWorker(json, sourceFile, host, basePath, existingOptions, existingWatchOptions, configFileName, resolutionStack, extraFileExtensions, extendedConfigCache) {
            if (existingOptions === void 0) {
                existingOptions = {};
            }
            if (resolutionStack === void 0) {
                resolutionStack = [];
            }
            if (extraFileExtensions === void 0) {
                extraFileExtensions = [];
            }
            ts.Debug.assert((json === undefined && sourceFile !== undefined) || (json !== undefined && sourceFile === undefined));
            var errors = [];
            var parsedConfig = parseConfig(json, sourceFile, host, basePath, configFileName, resolutionStack, errors, extendedConfigCache);
            var raw = parsedConfig.raw;
            var options = ts.extend(existingOptions, parsedConfig.options || {});
            var watchOptions = existingWatchOptions && parsedConfig.watchOptions ?
                ts.extend(existingWatchOptions, parsedConfig.watchOptions) :
                parsedConfig.watchOptions || existingWatchOptions;
            options.configFilePath = configFileName && ts.normalizeSlashes(configFileName);
            var configFileSpecs = getConfigFileSpecs();
            if (sourceFile)
                sourceFile.configFileSpecs = configFileSpecs;
            setConfigFileInOptions(options, sourceFile);
            var basePathForFileNames = ts.normalizePath(configFileName ? directoryOfCombinedPath(configFileName, basePath) : basePath);
            return {
                options: options,
                watchOptions: watchOptions,
                fileNames: getFileNames(basePathForFileNames),
                projectReferences: getProjectReferences(basePathForFileNames),
                typeAcquisition: parsedConfig.typeAcquisition || getDefaultTypeAcquisition(),
                raw: raw,
                errors: errors,
                // Wildcard directories (provided as part of a wildcard path) are stored in a
                // file map that marks whether it was a regular wildcard match (with a `*` or `?` token),
                // or a recursive directory. This information is used by filesystem watchers to monitor for
                // new entries in these paths.
                wildcardDirectories: getWildcardDirectories(configFileSpecs, basePathForFileNames, host.useCaseSensitiveFileNames),
                compileOnSave: !!raw.compileOnSave,
            };
            function getConfigFileSpecs() {
                var referencesOfRaw = getPropFromRaw("references", function (element) { return typeof element === "object"; }, "object");
                var filesSpecs = toPropValue(getSpecsFromRaw("files"));
                if (filesSpecs) {
                    var hasZeroOrNoReferences = referencesOfRaw === "no-prop" || ts.isArray(referencesOfRaw) && referencesOfRaw.length === 0;
                    var hasExtends = ts.hasProperty(raw, "extends");
                    if (filesSpecs.length === 0 && hasZeroOrNoReferences && !hasExtends) {
                        if (sourceFile) {
                            var fileName = configFileName || "tsconfig.json";
                            var diagnosticMessage = ts.Diagnostics.The_files_list_in_config_file_0_is_empty;
                            var nodeValue = ts.firstDefined(ts.getTsConfigPropArray(sourceFile, "files"), function (property) { return property.initializer; });
                            var error = nodeValue
                                ? ts.createDiagnosticForNodeInSourceFile(sourceFile, nodeValue, diagnosticMessage, fileName)
                                : ts.createCompilerDiagnostic(diagnosticMessage, fileName);
                            errors.push(error);
                        }
                        else {
                            createCompilerDiagnosticOnlyIfJson(ts.Diagnostics.The_files_list_in_config_file_0_is_empty, configFileName || "tsconfig.json");
                        }
                    }
                }
                var includeSpecs = toPropValue(getSpecsFromRaw("include"));
                var excludeOfRaw = getSpecsFromRaw("exclude");
                var excludeSpecs = toPropValue(excludeOfRaw);
                if (excludeOfRaw === "no-prop" && raw.compilerOptions) {
                    var outDir = raw.compilerOptions.outDir;
                    var declarationDir = raw.compilerOptions.declarationDir;
                    if (outDir || declarationDir) {
                        excludeSpecs = [outDir, declarationDir].filter(function (d) { return !!d; });
                    }
                }
                if (filesSpecs === undefined && includeSpecs === undefined) {
                    includeSpecs = ["**/*"];
                }
                var validatedIncludeSpecs, validatedExcludeSpecs;
                // The exclude spec list is converted into a regular expression, which allows us to quickly
                // test whether a file or directory should be excluded before recursively traversing the
                // file system.
                if (includeSpecs) {
                    validatedIncludeSpecs = validateSpecs(includeSpecs, errors, /*disallowTrailingRecursion*/ true, sourceFile, "include");
                }
                if (excludeSpecs) {
                    validatedExcludeSpecs = validateSpecs(excludeSpecs, errors, /*disallowTrailingRecursion*/ false, sourceFile, "exclude");
                }
                return {
                    filesSpecs: filesSpecs,
                    includeSpecs: includeSpecs,
                    excludeSpecs: excludeSpecs,
                    validatedFilesSpec: ts.filter(filesSpecs, ts.isString),
                    validatedIncludeSpecs: validatedIncludeSpecs,
                    validatedExcludeSpecs: validatedExcludeSpecs,
                    pathPatterns: undefined, // Initialized on first use
                };
            }
            function getFileNames(basePath) {
                var fileNames = getFileNamesFromConfigSpecs(configFileSpecs, basePath, options, host, extraFileExtensions);
                if (shouldReportNoInputFiles(fileNames, canJsonReportNoInputFiles(raw), resolutionStack)) {
                    errors.push(getErrorForNoInputFiles(configFileSpecs, configFileName));
                }
                return fileNames;
            }
            function getProjectReferences(basePath) {
                var projectReferences;
                var referencesOfRaw = getPropFromRaw("references", function (element) { return typeof element === "object"; }, "object");
                if (ts.isArray(referencesOfRaw)) {
                    for (var _i = 0, referencesOfRaw_1 = referencesOfRaw; _i < referencesOfRaw_1.length; _i++) {
                        var ref = referencesOfRaw_1[_i];
                        if (typeof ref.path !== "string") {
                            createCompilerDiagnosticOnlyIfJson(ts.Diagnostics.Compiler_option_0_requires_a_value_of_type_1, "reference.path", "string");
                        }
                        else {
                            (projectReferences || (projectReferences = [])).push({
                                path: ts.getNormalizedAbsolutePath(ref.path, basePath),
                                originalPath: ref.path,
                                prepend: ref.prepend,
                                circular: ref.circular
                            });
                        }
                    }
                }
                return projectReferences;
            }
            function toPropValue(specResult) {
                return ts.isArray(specResult) ? specResult : undefined;
            }
            function getSpecsFromRaw(prop) {
                return getPropFromRaw(prop, ts.isString, "string");
            }
            function getPropFromRaw(prop, validateElement, elementTypeName) {
                if (ts.hasProperty(raw, prop) && !isNullOrUndefined(raw[prop])) {
                    if (ts.isArray(raw[prop])) {
                        var result = raw[prop];
                        if (!sourceFile && !ts.every(result, validateElement)) {
                            errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Compiler_option_0_requires_a_value_of_type_1, prop, elementTypeName));
                        }
                        return result;
                    }
                    else {
                        createCompilerDiagnosticOnlyIfJson(ts.Diagnostics.Compiler_option_0_requires_a_value_of_type_1, prop, "Array");
                        return "not-array";
                    }
                }
                return "no-prop";
            }
            function createCompilerDiagnosticOnlyIfJson(message, arg0, arg1) {
                if (!sourceFile) {
                    errors.push(ts.createCompilerDiagnostic(message, arg0, arg1));
                }
            }
        }
        function isErrorNoInputFiles(error) {
            return error.code === ts.Diagnostics.No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2.code;
        }
        function getErrorForNoInputFiles(_a, configFileName) {
            var includeSpecs = _a.includeSpecs, excludeSpecs = _a.excludeSpecs;
            return ts.createCompilerDiagnostic(ts.Diagnostics.No_inputs_were_found_in_config_file_0_Specified_include_paths_were_1_and_exclude_paths_were_2, configFileName || "tsconfig.json", JSON.stringify(includeSpecs || []), JSON.stringify(excludeSpecs || []));
        }
        function shouldReportNoInputFiles(fileNames, canJsonReportNoInutFiles, resolutionStack) {
            return fileNames.length === 0 && canJsonReportNoInutFiles && (!resolutionStack || resolutionStack.length === 0);
        }
        /*@internal*/
        function canJsonReportNoInputFiles(raw) {
            return !ts.hasProperty(raw, "files") && !ts.hasProperty(raw, "references");
        }
        ts.canJsonReportNoInputFiles = canJsonReportNoInputFiles;
        /*@internal*/
        function updateErrorForNoInputFiles(fileNames, configFileName, configFileSpecs, configParseDiagnostics, canJsonReportNoInutFiles) {
            var existingErrors = configParseDiagnostics.length;
            if (shouldReportNoInputFiles(fileNames, canJsonReportNoInutFiles)) {
                configParseDiagnostics.push(getErrorForNoInputFiles(configFileSpecs, configFileName));
            }
            else {
                ts.filterMutate(configParseDiagnostics, function (error) { return !isErrorNoInputFiles(error); });
            }
            return existingErrors !== configParseDiagnostics.length;
        }
        ts.updateErrorForNoInputFiles = updateErrorForNoInputFiles;
        function isSuccessfulParsedTsconfig(value) {
            return !!value.options;
        }
        /**
         * This *just* extracts options/include/exclude/files out of a config file.
         * It does *not* resolve the included files.
         */
        function parseConfig(json, sourceFile, host, basePath, configFileName, resolutionStack, errors, extendedConfigCache) {
            var _a;
            basePath = ts.normalizeSlashes(basePath);
            var resolvedPath = ts.getNormalizedAbsolutePath(configFileName || "", basePath);
            if (resolutionStack.indexOf(resolvedPath) >= 0) {
                errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Circularity_detected_while_resolving_configuration_Colon_0, __spreadArray(__spreadArray([], resolutionStack, true), [resolvedPath], false).join(" -> ")));
                return { raw: json || convertToObject(sourceFile, errors) };
            }
            var ownConfig = json ?
                parseOwnConfigOfJson(json, host, basePath, configFileName, errors) :
                parseOwnConfigOfJsonSourceFile(sourceFile, host, basePath, configFileName, errors);
            if ((_a = ownConfig.options) === null || _a === void 0 ? void 0 : _a.paths) {
                // If we end up needing to resolve relative paths from 'paths' relative to
                // the config file location, we'll need to know where that config file was.
                // Since 'paths' can be inherited from an extended config in another directory,
                // we wouldn't know which directory to use unless we store it here.
                ownConfig.options.pathsBasePath = basePath;
            }
            if (ownConfig.extendedConfigPath) {
                // copy the resolution stack so it is never reused between branches in potential diamond-problem scenarios.
                resolutionStack = resolutionStack.concat([resolvedPath]);
                var extendedConfig = getExtendedConfig(sourceFile, ownConfig.extendedConfigPath, host, resolutionStack, errors, extendedConfigCache);
                if (extendedConfig && isSuccessfulParsedTsconfig(extendedConfig)) {
                    var baseRaw_1 = extendedConfig.raw;
                    var raw_1 = ownConfig.raw;
                    var relativeDifference_1;
                    var setPropertyInRawIfNotUndefined = function (propertyName) {
                        if (!raw_1[propertyName] && baseRaw_1[propertyName]) {
                            raw_1[propertyName] = ts.map(baseRaw_1[propertyName], function (path) { return ts.isRootedDiskPath(path) ? path : ts.combinePaths(relativeDifference_1 || (relativeDifference_1 = ts.convertToRelativePath(ts.getDirectoryPath(ownConfig.extendedConfigPath), basePath, ts.createGetCanonicalFileName(host.useCaseSensitiveFileNames))), path); });
                        }
                    };
                    setPropertyInRawIfNotUndefined("include");
                    setPropertyInRawIfNotUndefined("exclude");
                    setPropertyInRawIfNotUndefined("files");
                    if (raw_1.compileOnSave === undefined) {
                        raw_1.compileOnSave = baseRaw_1.compileOnSave;
                    }
                    ownConfig.options = ts.assign({}, extendedConfig.options, ownConfig.options);
                    ownConfig.watchOptions = ownConfig.watchOptions && extendedConfig.watchOptions ?
                        ts.assign({}, extendedConfig.watchOptions, ownConfig.watchOptions) :
                        ownConfig.watchOptions || extendedConfig.watchOptions;
                    // TODO extend type typeAcquisition
                }
            }
            return ownConfig;
        }
        function parseOwnConfigOfJson(json, host, basePath, configFileName, errors) {
            if (ts.hasProperty(json, "excludes")) {
                errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Unknown_option_excludes_Did_you_mean_exclude));
            }
            var options = convertCompilerOptionsFromJsonWorker(json.compilerOptions, basePath, errors, configFileName);
            // typingOptions has been deprecated and is only supported for backward compatibility purposes.
            // It should be removed in future releases - use typeAcquisition instead.
            var typeAcquisition = convertTypeAcquisitionFromJsonWorker(json.typeAcquisition || json.typingOptions, basePath, errors, configFileName);
            var watchOptions = convertWatchOptionsFromJsonWorker(json.watchOptions, basePath, errors);
            json.compileOnSave = convertCompileOnSaveOptionFromJson(json, basePath, errors);
            var extendedConfigPath;
            if (json.extends) {
                if (!ts.isString(json.extends)) {
                    errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Compiler_option_0_requires_a_value_of_type_1, "extends", "string"));
                }
                else {
                    var newBase = configFileName ? directoryOfCombinedPath(configFileName, basePath) : basePath;
                    extendedConfigPath = getExtendsConfigPath(json.extends, host, newBase, errors, ts.createCompilerDiagnostic);
                }
            }
            return { raw: json, options: options, watchOptions: watchOptions, typeAcquisition: typeAcquisition, extendedConfigPath: extendedConfigPath };
        }
        function parseOwnConfigOfJsonSourceFile(sourceFile, host, basePath, configFileName, errors) {
            var options = getDefaultCompilerOptions(configFileName);
            var typeAcquisition, typingOptionstypeAcquisition;
            var watchOptions;
            var extendedConfigPath;
            var optionsIterator = {
                onSetValidOptionKeyValueInParent: function (parentOption, option, value) {
                    var currentOption;
                    switch (parentOption) {
                        case "compilerOptions":
                            currentOption = options;
                            break;
                        case "watchOptions":
                            currentOption = (watchOptions || (watchOptions = {}));
                            break;
                        case "typeAcquisition":
                            currentOption = (typeAcquisition || (typeAcquisition = getDefaultTypeAcquisition(configFileName)));
                            break;
                        case "typingOptions":
                            currentOption = (typingOptionstypeAcquisition || (typingOptionstypeAcquisition = getDefaultTypeAcquisition(configFileName)));
                            break;
                        default:
                            ts.Debug.fail("Unknown option");
                    }
                    currentOption[option.name] = normalizeOptionValue(option, basePath, value);
                },
                onSetValidOptionKeyValueInRoot: function (key, _keyNode, value, valueNode) {
                    switch (key) {
                        case "extends":
                            var newBase = configFileName ? directoryOfCombinedPath(configFileName, basePath) : basePath;
                            extendedConfigPath = getExtendsConfigPath(value, host, newBase, errors, function (message, arg0) {
                                return ts.createDiagnosticForNodeInSourceFile(sourceFile, valueNode, message, arg0);
                            });
                            return;
                    }
                },
                onSetUnknownOptionKeyValueInRoot: function (key, keyNode, _value, _valueNode) {
                    if (key === "excludes") {
                        errors.push(ts.createDiagnosticForNodeInSourceFile(sourceFile, keyNode, ts.Diagnostics.Unknown_option_excludes_Did_you_mean_exclude));
                    }
                }
            };
            var json = convertConfigFileToObject(sourceFile, errors, /*reportOptionsErrors*/ true, optionsIterator);
            if (!typeAcquisition) {
                if (typingOptionstypeAcquisition) {
                    typeAcquisition = (typingOptionstypeAcquisition.enableAutoDiscovery !== undefined) ?
                        {
                            enable: typingOptionstypeAcquisition.enableAutoDiscovery,
                            include: typingOptionstypeAcquisition.include,
                            exclude: typingOptionstypeAcquisition.exclude
                        } :
                        typingOptionstypeAcquisition;
                }
                else {
                    typeAcquisition = getDefaultTypeAcquisition(configFileName);
                }
            }
            return { raw: json, options: options, watchOptions: watchOptions, typeAcquisition: typeAcquisition, extendedConfigPath: extendedConfigPath };
        }
        function getExtendsConfigPath(extendedConfig, host, basePath, errors, createDiagnostic) {
            extendedConfig = ts.normalizeSlashes(extendedConfig);
            if (ts.isRootedDiskPath(extendedConfig) || ts.startsWith(extendedConfig, "./") || ts.startsWith(extendedConfig, "../")) {
                var extendedConfigPath = ts.getNormalizedAbsolutePath(extendedConfig, basePath);
                if (!host.fileExists(extendedConfigPath) && !ts.endsWith(extendedConfigPath, ".json" /* Json */)) {
                    extendedConfigPath = extendedConfigPath + ".json";
                    if (!host.fileExists(extendedConfigPath)) {
                        errors.push(createDiagnostic(ts.Diagnostics.File_0_not_found, extendedConfig));
                        return undefined;
                    }
                }
                return extendedConfigPath;
            }
            // If the path isn't a rooted or relative path, resolve like a module
            var resolved = ts.nodeModuleNameResolver(extendedConfig, ts.combinePaths(basePath, "tsconfig.json"), { moduleResolution: ts.ModuleResolutionKind.NodeJs }, host, /*cache*/ undefined, /*projectRefs*/ undefined, /*lookupConfig*/ true);
            if (resolved.resolvedModule) {
                return resolved.resolvedModule.resolvedFileName;
            }
            errors.push(createDiagnostic(ts.Diagnostics.File_0_not_found, extendedConfig));
            return undefined;
        }
        function getExtendedConfig(sourceFile, extendedConfigPath, host, resolutionStack, errors, extendedConfigCache) {
            var _a;
            var path = host.useCaseSensitiveFileNames ? extendedConfigPath : ts.toFileNameLowerCase(extendedConfigPath);
            var value;
            var extendedResult;
            var extendedConfig;
            if (extendedConfigCache && (value = extendedConfigCache.get(path))) {
                (extendedResult = value.extendedResult, extendedConfig = value.extendedConfig);
            }
            else {
                extendedResult = readJsonConfigFile(extendedConfigPath, function (path) { return host.readFile(path); });
                if (!extendedResult.parseDiagnostics.length) {
                    extendedConfig = parseConfig(/*json*/ undefined, extendedResult, host, ts.getDirectoryPath(extendedConfigPath), ts.getBaseFileName(extendedConfigPath), resolutionStack, errors, extendedConfigCache);
                }
                if (extendedConfigCache) {
                    extendedConfigCache.set(path, { extendedResult: extendedResult, extendedConfig: extendedConfig });
                }
            }
            if (sourceFile) {
                sourceFile.extendedSourceFiles = [extendedResult.fileName];
                if (extendedResult.extendedSourceFiles) {
                    (_a = sourceFile.extendedSourceFiles).push.apply(_a, extendedResult.extendedSourceFiles);
                }
            }
            if (extendedResult.parseDiagnostics.length) {
                errors.push.apply(errors, extendedResult.parseDiagnostics);
                return undefined;
            }
            return extendedConfig;
        }
        function convertCompileOnSaveOptionFromJson(jsonOption, basePath, errors) {
            if (!ts.hasProperty(jsonOption, ts.compileOnSaveCommandLineOption.name)) {
                return false;
            }
            var result = convertJsonOption(ts.compileOnSaveCommandLineOption, jsonOption.compileOnSave, basePath, errors);
            return typeof result === "boolean" && result;
        }
        function convertCompilerOptionsFromJson(jsonOptions, basePath, configFileName) {
            var errors = [];
            var options = convertCompilerOptionsFromJsonWorker(jsonOptions, basePath, errors, configFileName);
            return { options: options, errors: errors };
        }
        ts.convertCompilerOptionsFromJson = convertCompilerOptionsFromJson;
        function convertTypeAcquisitionFromJson(jsonOptions, basePath, configFileName) {
            var errors = [];
            var options = convertTypeAcquisitionFromJsonWorker(jsonOptions, basePath, errors, configFileName);
            return { options: options, errors: errors };
        }
        ts.convertTypeAcquisitionFromJson = convertTypeAcquisitionFromJson;
        function getDefaultCompilerOptions(configFileName) {
            var options = configFileName && ts.getBaseFileName(configFileName) === "jsconfig.json"
                ? { allowJs: true, maxNodeModuleJsDepth: 2, allowSyntheticDefaultImports: true, skipLibCheck: true, noEmit: true }
                : {};
            return options;
        }
        function convertCompilerOptionsFromJsonWorker(jsonOptions, basePath, errors, configFileName) {
            var options = getDefaultCompilerOptions(configFileName);
            convertOptionsFromJson(getCommandLineCompilerOptionsMap(), jsonOptions, basePath, options, ts.compilerOptionsDidYouMeanDiagnostics, errors);
            if (configFileName) {
                options.configFilePath = ts.normalizeSlashes(configFileName);
            }
            return options;
        }
        function getDefaultTypeAcquisition(configFileName) {
            return { enable: !!configFileName && ts.getBaseFileName(configFileName) === "jsconfig.json", include: [], exclude: [] };
        }
        function convertTypeAcquisitionFromJsonWorker(jsonOptions, basePath, errors, configFileName) {
            var options = getDefaultTypeAcquisition(configFileName);
            var typeAcquisition = convertEnableAutoDiscoveryToEnable(jsonOptions);
            convertOptionsFromJson(getCommandLineTypeAcquisitionMap(), typeAcquisition, basePath, options, typeAcquisitionDidYouMeanDiagnostics, errors);
            return options;
        }
        function convertWatchOptionsFromJsonWorker(jsonOptions, basePath, errors) {
            return convertOptionsFromJson(getCommandLineWatchOptionsMap(), jsonOptions, basePath, /*defaultOptions*/ undefined, watchOptionsDidYouMeanDiagnostics, errors);
        }
        function convertOptionsFromJson(optionsNameMap, jsonOptions, basePath, defaultOptions, diagnostics, errors) {
            if (!jsonOptions) {
                return;
            }
            for (var id in jsonOptions) {
                var opt = optionsNameMap.get(id);
                if (opt) {
                    (defaultOptions || (defaultOptions = {}))[opt.name] = convertJsonOption(opt, jsonOptions[id], basePath, errors);
                }
                else {
                    errors.push(createUnknownOptionError(id, diagnostics, ts.createCompilerDiagnostic));
                }
            }
            return defaultOptions;
        }
        /*@internal*/
        function convertJsonOption(opt, value, basePath, errors) {
            if (isCompilerOptionsValue(opt, value)) {
                var optType = opt.type;
                if (optType === "list" && ts.isArray(value)) {
                    return convertJsonOptionOfListType(opt, value, basePath, errors);
                }
                else if (!ts.isString(optType)) {
                    return convertJsonOptionOfCustomType(opt, value, errors);
                }
                var validatedValue = validateJsonOptionValue(opt, value, errors);
                return isNullOrUndefined(validatedValue) ? validatedValue : normalizeNonListOptionValue(opt, basePath, validatedValue);
            }
            else {
                errors.push(ts.createCompilerDiagnostic(ts.Diagnostics.Compiler_option_0_requires_a_value_of_type_1, opt.name, getCompilerOptionValueTypeString(opt)));
            }
        }
        ts.convertJsonOption = convertJsonOption;
        function normalizeOptionValue(option, basePath, value) {
            if (isNullOrUndefined(value))
                return undefined;
            if (option.type === "list") {
                var listOption_1 = option;
                if (listOption_1.element.isFilePath || !ts.isString(listOption_1.element.type)) {
                    return ts.filter(ts.map(value, function (v) { return normalizeOptionValue(listOption_1.element, basePath, v); }), function (v) { return !!v; });
                }
                return value;
            }
            else if (!ts.isString(option.type)) {
                return option.type.get(ts.isString(value) ? value.toLowerCase() : value);
            }
            return normalizeNonListOptionValue(option, basePath, value);
        }
        function normalizeNonListOptionValue(option, basePath, value) {
            if (option.isFilePath) {
                value = ts.getNormalizedAbsolutePath(value, basePath);
                if (value === "") {
                    value = ".";
                }
            }
            return value;
        }
        function validateJsonOptionValue(opt, value, errors) {
            var _a;
            if (isNullOrUndefined(value))
                return undefined;
            var d = (_a = opt.extraValidation) === null || _a === void 0 ? void 0 : _a.call(opt, value);
            if (!d)
                return value;
            errors.push(ts.createCompilerDiagnostic.apply(void 0, d));
            return undefined;
        }
        function convertJsonOptionOfCustomType(opt, value, errors) {
            if (isNullOrUndefined(value))
                return undefined;
            var key = value.toLowerCase();
            var val = opt.type.get(key);
            if (val !== undefined) {
                return validateJsonOptionValue(opt, val, errors);
            }
            else {
                errors.push(createCompilerDiagnosticForInvalidCustomType(opt));
            }
        }
        function convertJsonOptionOfListType(option, values, basePath, errors) {
            return ts.filter(ts.map(values, function (v) { return convertJsonOption(option.element, v, basePath, errors); }), function (v) { return !!v; });
        }
        /**
         * Tests for a path that ends in a recursive directory wildcard.
         * Matches **, \**, **\, and \**\, but not a**b.
         *
         * NOTE: used \ in place of / above to avoid issues with multiline comments.
         *
         * Breakdown:
         *  (^|\/)      # matches either the beginning of the string or a directory separator.
         *  \*\*        # matches the recursive directory wildcard "**".
         *  \/?$        # matches an optional trailing directory separator at the end of the string.
         */
        var invalidTrailingRecursionPattern = /(^|\/)\*\*\/?$/;
        /**
         * Matches the portion of a wildcard path that does not contain wildcards.
         * Matches \a of \a\*, or \a\b\c of \a\b\c\?\d.
         *
         * NOTE: used \ in place of / above to avoid issues with multiline comments.
         *
         * Breakdown:
         *  ^                   # matches the beginning of the string
         *  [^*?]*              # matches any number of non-wildcard characters
         *  (?=\/[^/]*[*?])     # lookahead that matches a directory separator followed by
         *                      # a path component that contains at least one wildcard character (* or ?).
         */
        var wildcardDirectoryPattern = /^[^*?]*(?=\/[^/]*[*?])/;
        /**
         * Gets the file names from the provided config file specs that contain, files, include, exclude and
         * other properties needed to resolve the file names
         * @param configFileSpecs The config file specs extracted with file names to include, wildcards to include/exclude and other details
         * @param basePath The base path for any relative file specifications.
         * @param options Compiler options.
         * @param host The host used to resolve files and directories.
         * @param extraFileExtensions optionaly file extra file extension information from host
         */
        /* @internal */
        function getFileNamesFromConfigSpecs(configFileSpecs, basePath, options, host, extraFileExtensions) {
            if (extraFileExtensions === void 0) {
                extraFileExtensions = ts.emptyArray;
            }
            basePath = ts.normalizePath(basePath);
            var keyMapper = ts.createGetCanonicalFileName(host.useCaseSensitiveFileNames);
            // Literal file names (provided via the "files" array in tsconfig.json) are stored in a
            // file map with a possibly case insensitive key. We use this map later when when including
            // wildcard paths.
            var literalFileMap = new ts.Map();
            // Wildcard paths (provided via the "includes" array in tsconfig.json) are stored in a
            // file map with a possibly case insensitive key. We use this map to store paths matched
            // via wildcard, and to handle extension priority.
            var wildcardFileMap = new ts.Map();
            // Wildcard paths of json files (provided via the "includes" array in tsconfig.json) are stored in a
            // file map with a possibly case insensitive key. We use this map to store paths matched
            // via wildcard of *.json kind
            var wildCardJsonFileMap = new ts.Map();
            var validatedFilesSpec = configFileSpecs.validatedFilesSpec, validatedIncludeSpecs = configFileSpecs.validatedIncludeSpecs, validatedExcludeSpecs = configFileSpecs.validatedExcludeSpecs;
            // Rather than re-query this for each file and filespec, we query the supported extensions
            // once and store it on the expansion context.
            var supportedExtensions = ts.getSupportedExtensions(options, extraFileExtensions);
            var supportedExtensionsWithJsonIfResolveJsonModule = ts.getSuppoertedExtensionsWithJsonIfResolveJsonModule(options, supportedExtensions);
            // Literal files are always included verbatim. An "include" or "exclude" specification cannot
            // remove a literal file.
            if (validatedFilesSpec) {
                for (var _i = 0, validatedFilesSpec_1 = validatedFilesSpec; _i < validatedFilesSpec_1.length; _i++) {
                    var fileName = validatedFilesSpec_1[_i];
                    var file = ts.getNormalizedAbsolutePath(fileName, basePath);
                    literalFileMap.set(keyMapper(file), file);
                }
            }
            var jsonOnlyIncludeRegexes;
            if (validatedIncludeSpecs && validatedIncludeSpecs.length > 0) {
                var _loop_6 = function (file) {
                    if (ts.fileExtensionIs(file, ".json" /* Json */)) {
                        // Valid only if *.json specified
                        if (!jsonOnlyIncludeRegexes) {
                            var includes = validatedIncludeSpecs.filter(function (s) { return ts.endsWith(s, ".json" /* Json */); });
                            var includeFilePatterns = ts.map(ts.getRegularExpressionsForWildcards(includes, basePath, "files"), function (pattern) { return "^" + pattern + "$"; });
                            jsonOnlyIncludeRegexes = includeFilePatterns ? includeFilePatterns.map(function (pattern) { return ts.getRegexFromPattern(pattern, host.useCaseSensitiveFileNames); }) : ts.emptyArray;
                        }
                        var includeIndex = ts.findIndex(jsonOnlyIncludeRegexes, function (re) { return re.test(file); });
                        if (includeIndex !== -1) {
                            var key_1 = keyMapper(file);
                            if (!literalFileMap.has(key_1) && !wildCardJsonFileMap.has(key_1)) {
                                wildCardJsonFileMap.set(key_1, file);
                            }
                        }
                        return "continue";
                    }
                    // If we have already included a literal or wildcard path with a
                    // higher priority extension, we should skip this file.
                    //
                    // This handles cases where we may encounter both <file>.ts and
                    // <file>.d.ts (or <file>.js if "allowJs" is enabled) in the same
                    // directory when they are compilation outputs.
                    if (hasFileWithHigherPriorityExtension(file, literalFileMap, wildcardFileMap, supportedExtensions, keyMapper)) {
                        return "continue";
                    }
                    // We may have included a wildcard path with a lower priority
                    // extension due to the user-defined order of entries in the
                    // "include" array. If there is a lower priority extension in the
                    // same directory, we should remove it.
                    removeWildcardFilesWithLowerPriorityExtension(file, wildcardFileMap, supportedExtensions, keyMapper);
                    var key = keyMapper(file);
                    if (!literalFileMap.has(key) && !wildcardFileMap.has(key)) {
                        wildcardFileMap.set(key, file);
                    }
                };
                for (var _a = 0, _b = host.readDirectory(basePath, supportedExtensionsWithJsonIfResolveJsonModule, validatedExcludeSpecs, validatedIncludeSpecs, /*depth*/ undefined); _a < _b.length; _a++) {
                    var file = _b[_a];
                    _loop_6(file);
                }
            }
            var literalFiles = ts.arrayFrom(literalFileMap.values());
            var wildcardFiles = ts.arrayFrom(wildcardFileMap.values());
            return literalFiles.concat(wildcardFiles, ts.arrayFrom(wildCardJsonFileMap.values()));
        }
        ts.getFileNamesFromConfigSpecs = getFileNamesFromConfigSpecs;
        /* @internal */
        function isExcludedFile(pathToCheck, spec, basePath, useCaseSensitiveFileNames, currentDirectory) {
            var validatedFilesSpec = spec.validatedFilesSpec, validatedIncludeSpecs = spec.validatedIncludeSpecs, validatedExcludeSpecs = spec.validatedExcludeSpecs;
            if (!ts.length(validatedIncludeSpecs) || !ts.length(validatedExcludeSpecs))
                return false;
            basePath = ts.normalizePath(basePath);
            var keyMapper = ts.createGetCanonicalFileName(useCaseSensitiveFileNames);
            if (validatedFilesSpec) {
                for (var _i = 0, validatedFilesSpec_2 = validatedFilesSpec; _i < validatedFilesSpec_2.length; _i++) {
                    var fileName = validatedFilesSpec_2[_i];
                    if (keyMapper(ts.getNormalizedAbsolutePath(fileName, basePath)) === pathToCheck)
                        return false;
                }
            }
            return matchesExcludeWorker(pathToCheck, validatedExcludeSpecs, useCaseSensitiveFileNames, currentDirectory, basePath);
        }
        ts.isExcludedFile = isExcludedFile;
        function invalidDotDotAfterRecursiveWildcard(s) {
            // We used to use the regex /(^|\/)\*\*\/(.*\/)?\.\.($|\/)/ to check for this case, but
            // in v8, that has polynomial performance because the recursive wildcard match - **/ -
            // can be matched in many arbitrary positions when multiple are present, resulting
            // in bad backtracking (and we don't care which is matched - just that some /.. segment
            // comes after some **/ segment).
            var wildcardIndex = ts.startsWith(s, "**/") ? 0 : s.indexOf("/**/");
            if (wildcardIndex === -1) {
                return false;
            }
            var lastDotIndex = ts.endsWith(s, "/..") ? s.length : s.lastIndexOf("/../");
            return lastDotIndex > wildcardIndex;
        }
        /* @internal */
        function matchesExclude(pathToCheck, excludeSpecs, useCaseSensitiveFileNames, currentDirectory) {
            return matchesExcludeWorker(pathToCheck, ts.filter(excludeSpecs, function (spec) { return !invalidDotDotAfterRecursiveWildcard(spec); }), useCaseSensitiveFileNames, currentDirectory);
        }
        ts.matchesExclude = matchesExclude;
        function matchesExcludeWorker(pathToCheck, excludeSpecs, useCaseSensitiveFileNames, currentDirectory, basePath) {
            var excludePattern = ts.getRegularExpressionForWildcard(excludeSpecs, ts.combinePaths(ts.normalizePath(currentDirectory), basePath), "exclude");
            var excludeRegex = excludePattern && ts.getRegexFromPattern(excludePattern, useCaseSensitiveFileNames);
            if (!excludeRegex)
                return false;
            if (excludeRegex.test(pathToCheck))
                return true;
            return !ts.hasExtension(pathToCheck) && excludeRegex.test(ts.ensureTrailingDirectorySeparator(pathToCheck));
        }
        function validateSpecs(specs, errors, disallowTrailingRecursion, jsonSourceFile, specKey) {
            return specs.filter(function (spec) {
                if (!ts.isString(spec))
                    return false;
                var diag = specToDiagnostic(spec, disallowTrailingRecursion);
                if (diag !== undefined) {
                    errors.push(createDiagnostic.apply(void 0, diag));
                }
                return diag === undefined;
            });
            function createDiagnostic(message, spec) {
                var element = ts.getTsConfigPropArrayElementValue(jsonSourceFile, specKey, spec);
                return element ?
                    ts.createDiagnosticForNodeInSourceFile(jsonSourceFile, element, message, spec) :
                    ts.createCompilerDiagnostic(message, spec);
            }
        }
        function specToDiagnostic(spec, disallowTrailingRecursion) {
            if (disallowTrailingRecursion && invalidTrailingRecursionPattern.test(spec)) {
                return [ts.Diagnostics.File_specification_cannot_end_in_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0, spec];
            }
            else if (invalidDotDotAfterRecursiveWildcard(spec)) {
                return [ts.Diagnostics.File_specification_cannot_contain_a_parent_directory_that_appears_after_a_recursive_directory_wildcard_Asterisk_Asterisk_Colon_0, spec];
            }
        }
        /**
         * Gets directories in a set of include patterns that should be watched for changes.
         */
        function getWildcardDirectories(_a, path, useCaseSensitiveFileNames) {
            var include = _a.validatedIncludeSpecs, exclude = _a.validatedExcludeSpecs;
            // We watch a directory recursively if it contains a wildcard anywhere in a directory segment
            // of the pattern:
            //
            //  /a/b/**/d   - Watch /a/b recursively to catch changes to any d in any subfolder recursively
            //  /a/b/*/d    - Watch /a/b recursively to catch any d in any immediate subfolder, even if a new subfolder is added
            //  /a/b        - Watch /a/b recursively to catch changes to anything in any recursive subfoler
            //
            // We watch a directory without recursion if it contains a wildcard in the file segment of
            // the pattern:
            //
            //  /a/b/*      - Watch /a/b directly to catch any new file
            //  /a/b/a?z    - Watch /a/b directly to catch any new file matching a?z
            var rawExcludeRegex = ts.getRegularExpressionForWildcard(exclude, path, "exclude");
            var excludeRegex = rawExcludeRegex && new RegExp(rawExcludeRegex, useCaseSensitiveFileNames ? "" : "i");
            var wildcardDirectories = {};
            if (include !== undefined) {
                var recursiveKeys = [];
                for (var _i = 0, include_1 = include; _i < include_1.length; _i++) {
                    var file = include_1[_i];
                    var spec = ts.normalizePath(ts.combinePaths(path, file));
                    if (excludeRegex && excludeRegex.test(spec)) {
                        continue;
                    }
                    var match = getWildcardDirectoryFromSpec(spec, useCaseSensitiveFileNames);
                    if (match) {
                        var key = match.key, flags = match.flags;
                        var existingFlags = wildcardDirectories[key];
                        if (existingFlags === undefined || existingFlags < flags) {
                            wildcardDirectories[key] = flags;
                            if (flags === 1 /* Recursive */) {
                                recursiveKeys.push(key);
                            }
                        }
                    }
                }
                // Remove any subpaths under an existing recursively watched directory.
                for (var key in wildcardDirectories) {
                    if (ts.hasProperty(wildcardDirectories, key)) {
                        for (var _b = 0, recursiveKeys_1 = recursiveKeys; _b < recursiveKeys_1.length; _b++) {
                            var recursiveKey = recursiveKeys_1[_b];
                            if (key !== recursiveKey && ts.containsPath(recursiveKey, key, path, !useCaseSensitiveFileNames)) {
                                delete wildcardDirectories[key];
                            }
                        }
                    }
                }
            }
            return wildcardDirectories;
        }
        function getWildcardDirectoryFromSpec(spec, useCaseSensitiveFileNames) {
            var match = wildcardDirectoryPattern.exec(spec);
            if (match) {
                // We check this with a few `indexOf` calls because 3 `indexOf`/`lastIndexOf` calls is
                // less algorithmically complex (roughly O(3n) worst-case) than the regex we used to use,
                // \/[^/]*?[*?][^/]*\/ which was polynominal in v8, since arbitrary sequences of wildcard
                // characters could match any of the central patterns, resulting in bad backtracking.
                var questionWildcardIndex = spec.indexOf("?");
                var starWildcardIndex = spec.indexOf("*");
                var lastDirectorySeperatorIndex = spec.lastIndexOf(ts.directorySeparator);
                return {
                    key: useCaseSensitiveFileNames ? match[0] : ts.toFileNameLowerCase(match[0]),
                    flags: (questionWildcardIndex !== -1 && questionWildcardIndex < lastDirectorySeperatorIndex)
                        || (starWildcardIndex !== -1 && starWildcardIndex < lastDirectorySeperatorIndex)
                        ? 1 /* Recursive */ : 0 /* None */
                };
            }
            if (ts.isImplicitGlob(spec)) {
                return {
                    key: useCaseSensitiveFileNames ? spec : ts.toFileNameLowerCase(spec),
                    flags: 1 /* Recursive */
                };
            }
            return undefined;
        }
        /**
         * Determines whether a literal or wildcard file has already been included that has a higher
         * extension priority.
         *
         * @param file The path to the file.
         * @param extensionPriority The priority of the extension.
         * @param context The expansion context.
         */
        function hasFileWithHigherPriorityExtension(file, literalFiles, wildcardFiles, extensions, keyMapper) {
            var extensionPriority = ts.getExtensionPriority(file, extensions);
            var adjustedExtensionPriority = ts.adjustExtensionPriority(extensionPriority, extensions);
            for (var i = 0 /* Highest */; i < adjustedExtensionPriority; i++) {
                var higherPriorityExtension = extensions[i];
                var higherPriorityPath = keyMapper(ts.changeExtension(file, higherPriorityExtension));
                if (literalFiles.has(higherPriorityPath) || wildcardFiles.has(higherPriorityPath)) {
                    return true;
                }
            }
            return false;
        }
        /**
         * Removes files included via wildcard expansion with a lower extension priority that have
         * already been included.
         *
         * @param file The path to the file.
         * @param extensionPriority The priority of the extension.
         * @param context The expansion context.
         */
        function removeWildcardFilesWithLowerPriorityExtension(file, wildcardFiles, extensions, keyMapper) {
            var extensionPriority = ts.getExtensionPriority(file, extensions);
            var nextExtensionPriority = ts.getNextLowestExtensionPriority(extensionPriority, extensions);
            for (var i = nextExtensionPriority; i < extensions.length; i++) {
                var lowerPriorityExtension = extensions[i];
                var lowerPriorityPath = keyMapper(ts.changeExtension(file, lowerPriorityExtension));
                wildcardFiles.delete(lowerPriorityPath);
            }
        }
        /**
         * Produces a cleaned version of compiler options with personally identifying info (aka, paths) removed.
         * Also converts enum values back to strings.
         */
        /* @internal */
        function convertCompilerOptionsForTelemetry(opts) {
            var out = {};
            for (var key in opts) {
                if (opts.hasOwnProperty(key)) {
                    var type = getOptionFromName(key);
                    if (type !== undefined) { // Ignore unknown options
                        out[key] = getOptionValueWithEmptyStrings(opts[key], type);
                    }
                }
            }
            return out;
        }
        ts.convertCompilerOptionsForTelemetry = convertCompilerOptionsForTelemetry;
        function getOptionValueWithEmptyStrings(value, option) {
            switch (option.type) {
                case "object": // "paths". Can't get any useful information from the value since we blank out strings, so just return "".
                    return "";
                case "string": // Could be any arbitrary string -- use empty string instead.
                    return "";
                case "number": // Allow numbers, but be sure to check it's actually a number.
                    return typeof value === "number" ? value : "";
                case "boolean":
                    return typeof value === "boolean" ? value : "";
                case "list":
                    var elementType_1 = option.element;
                    return ts.isArray(value) ? value.map(function (v) { return getOptionValueWithEmptyStrings(v, elementType_1); }) : "";
                default:
                    return ts.forEachEntry(option.type, function (optionEnumValue, optionStringValue) {
                        if (optionEnumValue === value) {
                            return optionStringValue;
                        }
                    }); // TODO: GH#18217
            }
        }
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        function trace(host) {
            host.trace(ts.formatMessage.apply(undefined, arguments));
        }
        ts.trace = trace;
        /* @internal */
        function isTraceEnabled(compilerOptions, host) {
            return !!compilerOptions.traceResolution && host.trace !== undefined;
        }
        ts.isTraceEnabled = isTraceEnabled;
        function withPackageId(packageInfo, r) {
            var packageId;
            if (r && packageInfo) {
                var packageJsonContent = packageInfo.packageJsonContent;
                if (typeof packageJsonContent.name === "string" && typeof packageJsonContent.version === "string") {
                    packageId = {
                        name: packageJsonContent.name,
                        subModuleName: r.path.slice(packageInfo.packageDirectory.length + ts.directorySeparator.length),
                        version: packageJsonContent.version
                    };
                }
            }
            return r && { path: r.path, extension: r.ext, packageId: packageId };
        }
        function noPackageId(r) {
            return withPackageId(/*packageInfo*/ undefined, r);
        }
        function removeIgnoredPackageId(r) {
            if (r) {
                ts.Debug.assert(r.packageId === undefined);
                return { path: r.path, ext: r.extension };
            }
        }
        /**
         * Kinds of file that we are currently looking for.
         * Typically there is one pass with Extensions.TypeScript, then a second pass with Extensions.JavaScript.
         */
        var Extensions;
        (function (Extensions) {
            Extensions[Extensions["TypeScript"] = 0] = "TypeScript";
            Extensions[Extensions["JavaScript"] = 1] = "JavaScript";
            Extensions[Extensions["Json"] = 2] = "Json";
            Extensions[Extensions["TSConfig"] = 3] = "TSConfig";
            Extensions[Extensions["DtsOnly"] = 4] = "DtsOnly"; /** Only '.d.ts' */
        })(Extensions || (Extensions = {}));
        /** Used with `Extensions.DtsOnly` to extract the path from TypeScript results. */
        function resolvedTypeScriptOnly(resolved) {
            if (!resolved) {
                return undefined;
            }
            ts.Debug.assert(ts.extensionIsTS(resolved.extension));
            return { fileName: resolved.path, packageId: resolved.packageId };
        }
        function createResolvedModuleWithFailedLookupLocations(resolved, isExternalLibraryImport, failedLookupLocations, resultFromCache) {
            var _a;
            if (resultFromCache) {
                (_a = resultFromCache.failedLookupLocations).push.apply(_a, failedLookupLocations);
                return resultFromCache;
            }
            return {
                resolvedModule: resolved && { resolvedFileName: resolved.path, originalPath: resolved.originalPath === true ? undefined : resolved.originalPath, extension: resolved.extension, isExternalLibraryImport: isExternalLibraryImport, packageId: resolved.packageId },
                failedLookupLocations: failedLookupLocations
            };
        }
        function readPackageJsonField(jsonContent, fieldName, typeOfTag, state) {
            if (!ts.hasProperty(jsonContent, fieldName)) {
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.package_json_does_not_have_a_0_field, fieldName);
                }
                return;
            }
            var value = jsonContent[fieldName];
            if (typeof value !== typeOfTag || value === null) { // eslint-disable-line no-null/no-null
                if (state.traceEnabled) {
                    // eslint-disable-next-line no-null/no-null
                    trace(state.host, ts.Diagnostics.Expected_type_of_0_field_in_package_json_to_be_1_got_2, fieldName, typeOfTag, value === null ? "null" : typeof value);
                }
                return;
            }
            return value;
        }
        function readPackageJsonPathField(jsonContent, fieldName, baseDirectory, state) {
            var fileName = readPackageJsonField(jsonContent, fieldName, "string", state);
            if (fileName === undefined) {
                return;
            }
            if (!fileName) {
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.package_json_had_a_falsy_0_field, fieldName);
                }
                return;
            }
            var path = ts.normalizePath(ts.combinePaths(baseDirectory, fileName));
            if (state.traceEnabled) {
                trace(state.host, ts.Diagnostics.package_json_has_0_field_1_that_references_2, fieldName, fileName, path);
            }
            return path;
        }
        function readPackageJsonTypesFields(jsonContent, baseDirectory, state) {
            return readPackageJsonPathField(jsonContent, "typings", baseDirectory, state)
                || readPackageJsonPathField(jsonContent, "types", baseDirectory, state);
        }
        function readPackageJsonTSConfigField(jsonContent, baseDirectory, state) {
            return readPackageJsonPathField(jsonContent, "tsconfig", baseDirectory, state);
        }
        function readPackageJsonMainField(jsonContent, baseDirectory, state) {
            return readPackageJsonPathField(jsonContent, "main", baseDirectory, state);
        }
        function readPackageJsonTypesVersionsField(jsonContent, state) {
            var typesVersions = readPackageJsonField(jsonContent, "typesVersions", "object", state);
            if (typesVersions === undefined)
                return;
            if (state.traceEnabled) {
                trace(state.host, ts.Diagnostics.package_json_has_a_typesVersions_field_with_version_specific_path_mappings);
            }
            return typesVersions;
        }
        function readPackageJsonTypesVersionPaths(jsonContent, state) {
            var typesVersions = readPackageJsonTypesVersionsField(jsonContent, state);
            if (typesVersions === undefined)
                return;
            if (state.traceEnabled) {
                for (var key in typesVersions) {
                    if (ts.hasProperty(typesVersions, key) && !ts.VersionRange.tryParse(key)) {
                        trace(state.host, ts.Diagnostics.package_json_has_a_typesVersions_entry_0_that_is_not_a_valid_semver_range, key);
                    }
                }
            }
            var result = getPackageJsonTypesVersionsPaths(typesVersions);
            if (!result) {
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.package_json_does_not_have_a_typesVersions_entry_that_matches_version_0, ts.versionMajorMinor);
                }
                return;
            }
            var bestVersionKey = result.version, bestVersionPaths = result.paths;
            if (typeof bestVersionPaths !== "object") {
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.Expected_type_of_0_field_in_package_json_to_be_1_got_2, "typesVersions['" + bestVersionKey + "']", "object", typeof bestVersionPaths);
                }
                return;
            }
            return result;
        }
        var typeScriptVersion;
        /* @internal */
        function getPackageJsonTypesVersionsPaths(typesVersions) {
            if (!typeScriptVersion)
                typeScriptVersion = new ts.Version(ts.version);
            for (var key in typesVersions) {
                if (!ts.hasProperty(typesVersions, key))
                    continue;
                var keyRange = ts.VersionRange.tryParse(key);
                if (keyRange === undefined) {
                    continue;
                }
                // return the first entry whose range matches the current compiler version.
                if (keyRange.test(typeScriptVersion)) {
                    return { version: key, paths: typesVersions[key] };
                }
            }
        }
        ts.getPackageJsonTypesVersionsPaths = getPackageJsonTypesVersionsPaths;
        function getEffectiveTypeRoots(options, host) {
            if (options.typeRoots) {
                return options.typeRoots;
            }
            var currentDirectory;
            if (options.configFilePath) {
                currentDirectory = ts.getDirectoryPath(options.configFilePath);
            }
            else if (host.getCurrentDirectory) {
                currentDirectory = host.getCurrentDirectory();
            }
            if (currentDirectory !== undefined) {
                return getDefaultTypeRoots(currentDirectory, host);
            }
        }
        ts.getEffectiveTypeRoots = getEffectiveTypeRoots;
        /**
         * Returns the path to every node_modules/@types directory from some ancestor directory.
         * Returns undefined if there are none.
         */
        function getDefaultTypeRoots(currentDirectory, host) {
            if (!host.directoryExists) {
                return [ts.combinePaths(currentDirectory, nodeModulesAtTypes)];
                // And if it doesn't exist, tough.
            }
            var typeRoots;
            ts.forEachAncestorDirectory(ts.normalizePath(currentDirectory), function (directory) {
                var atTypes = ts.combinePaths(directory, nodeModulesAtTypes);
                if (host.directoryExists(atTypes)) {
                    (typeRoots || (typeRoots = [])).push(atTypes);
                }
                return undefined;
            });
            return typeRoots;
        }
        var nodeModulesAtTypes = ts.combinePaths("node_modules", "@types");
        /**
         * @param {string | undefined} containingFile - file that contains type reference directive, can be undefined if containing file is unknown.
         * This is possible in case if resolution is performed for directives specified via 'types' parameter. In this case initial path for secondary lookups
         * is assumed to be the same as root directory of the project.
         */
        function resolveTypeReferenceDirective(typeReferenceDirectiveName, containingFile, options, host, redirectedReference, cache) {
            var traceEnabled = isTraceEnabled(options, host);
            if (redirectedReference) {
                options = redirectedReference.commandLine.options;
            }
            var containingDirectory = containingFile ? ts.getDirectoryPath(containingFile) : undefined;
            var perFolderCache = containingDirectory ? cache && cache.getOrCreateCacheForDirectory(containingDirectory, redirectedReference) : undefined;
            var result = perFolderCache && perFolderCache.get(typeReferenceDirectiveName);
            if (result) {
                if (traceEnabled) {
                    trace(host, ts.Diagnostics.Resolving_type_reference_directive_0_containing_file_1, typeReferenceDirectiveName, containingFile);
                    if (redirectedReference)
                        trace(host, ts.Diagnostics.Using_compiler_options_of_project_reference_redirect_0, redirectedReference.sourceFile.fileName);
                    trace(host, ts.Diagnostics.Resolution_for_type_reference_directive_0_was_found_in_cache_from_location_1, typeReferenceDirectiveName, containingDirectory);
                    traceResult(result);
                }
                return result;
            }
            var typeRoots = getEffectiveTypeRoots(options, host);
            if (traceEnabled) {
                if (containingFile === undefined) {
                    if (typeRoots === undefined) {
                        trace(host, ts.Diagnostics.Resolving_type_reference_directive_0_containing_file_not_set_root_directory_not_set, typeReferenceDirectiveName);
                    }
                    else {
                        trace(host, ts.Diagnostics.Resolving_type_reference_directive_0_containing_file_not_set_root_directory_1, typeReferenceDirectiveName, typeRoots);
                    }
                }
                else {
                    if (typeRoots === undefined) {
                        trace(host, ts.Diagnostics.Resolving_type_reference_directive_0_containing_file_1_root_directory_not_set, typeReferenceDirectiveName, containingFile);
                    }
                    else {
                        trace(host, ts.Diagnostics.Resolving_type_reference_directive_0_containing_file_1_root_directory_2, typeReferenceDirectiveName, containingFile, typeRoots);
                    }
                }
                if (redirectedReference) {
                    trace(host, ts.Diagnostics.Using_compiler_options_of_project_reference_redirect_0, redirectedReference.sourceFile.fileName);
                }
            }
            var failedLookupLocations = [];
            var moduleResolutionState = { compilerOptions: options, host: host, traceEnabled: traceEnabled, failedLookupLocations: failedLookupLocations, packageJsonInfoCache: cache };
            var resolved = primaryLookup();
            var primary = true;
            if (!resolved) {
                resolved = secondaryLookup();
                primary = false;
            }
            var resolvedTypeReferenceDirective;
            if (resolved) {
                var fileName = resolved.fileName, packageId = resolved.packageId;
                var resolvedFileName = options.preserveSymlinks ? fileName : realPath(fileName, host, traceEnabled);
                resolvedTypeReferenceDirective = {
                    primary: primary,
                    resolvedFileName: resolvedFileName,
                    originalPath: fileName === resolvedFileName ? undefined : fileName,
                    packageId: packageId,
                    isExternalLibraryImport: pathContainsNodeModules(fileName),
                };
            }
            result = { resolvedTypeReferenceDirective: resolvedTypeReferenceDirective, failedLookupLocations: failedLookupLocations };
            perFolderCache === null || perFolderCache === void 0 ? void 0 : perFolderCache.set(typeReferenceDirectiveName, result);
            if (traceEnabled)
                traceResult(result);
            return result;
            function traceResult(result) {
                var _a;
                if (!((_a = result.resolvedTypeReferenceDirective) === null || _a === void 0 ? void 0 : _a.resolvedFileName)) {
                    trace(host, ts.Diagnostics.Type_reference_directive_0_was_not_resolved, typeReferenceDirectiveName);
                }
                else if (result.resolvedTypeReferenceDirective.packageId) {
                    trace(host, ts.Diagnostics.Type_reference_directive_0_was_successfully_resolved_to_1_with_Package_ID_2_primary_Colon_3, typeReferenceDirectiveName, result.resolvedTypeReferenceDirective.resolvedFileName, ts.packageIdToString(result.resolvedTypeReferenceDirective.packageId), result.resolvedTypeReferenceDirective.primary);
                }
                else {
                    trace(host, ts.Diagnostics.Type_reference_directive_0_was_successfully_resolved_to_1_primary_Colon_2, typeReferenceDirectiveName, result.resolvedTypeReferenceDirective.resolvedFileName, result.resolvedTypeReferenceDirective.primary);
                }
            }
            function primaryLookup() {
                // Check primary library paths
                if (typeRoots && typeRoots.length) {
                    if (traceEnabled) {
                        trace(host, ts.Diagnostics.Resolving_with_primary_search_path_0, typeRoots.join(", "));
                    }
                    return ts.firstDefined(typeRoots, function (typeRoot) {
                        var candidate = ts.combinePaths(typeRoot, typeReferenceDirectiveName);
                        var candidateDirectory = ts.getDirectoryPath(candidate);
                        var directoryExists = ts.directoryProbablyExists(candidateDirectory, host);
                        if (!directoryExists && traceEnabled) {
                            trace(host, ts.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it, candidateDirectory);
                        }
                        return resolvedTypeScriptOnly(loadNodeModuleFromDirectory(Extensions.DtsOnly, candidate, !directoryExists, moduleResolutionState));
                    });
                }
                else {
                    if (traceEnabled) {
                        trace(host, ts.Diagnostics.Root_directory_cannot_be_determined_skipping_primary_search_paths);
                    }
                }
            }
            function secondaryLookup() {
                var initialLocationForSecondaryLookup = containingFile && ts.getDirectoryPath(containingFile);
                if (initialLocationForSecondaryLookup !== undefined) {
                    // check secondary locations
                    if (traceEnabled) {
                        trace(host, ts.Diagnostics.Looking_up_in_node_modules_folder_initial_location_0, initialLocationForSecondaryLookup);
                    }
                    var result_4;
                    if (!ts.isExternalModuleNameRelative(typeReferenceDirectiveName)) {
                        var searchResult = loadModuleFromNearestNodeModulesDirectory(Extensions.DtsOnly, typeReferenceDirectiveName, initialLocationForSecondaryLookup, moduleResolutionState, /*cache*/ undefined, /*redirectedReference*/ undefined);
                        result_4 = searchResult && searchResult.value;
                    }
                    else {
                        var candidate = ts.normalizePathAndParts(ts.combinePaths(initialLocationForSecondaryLookup, typeReferenceDirectiveName)).path;
                        result_4 = nodeLoadModuleByRelativeName(Extensions.DtsOnly, candidate, /*onlyRecordFailures*/ false, moduleResolutionState, /*considerPackageJson*/ true);
                    }
                    return resolvedTypeScriptOnly(result_4);
                }
                else {
                    if (traceEnabled) {
                        trace(host, ts.Diagnostics.Containing_file_is_not_specified_and_root_directory_cannot_be_determined_skipping_lookup_in_node_modules_folder);
                    }
                }
            }
        }
        ts.resolveTypeReferenceDirective = resolveTypeReferenceDirective;
        /**
         * Given a set of options, returns the set of type directive names
         *   that should be included for this program automatically.
         * This list could either come from the config file,
         *   or from enumerating the types root + initial secondary types lookup location.
         * More type directives might appear in the program later as a result of loading actual source files;
         *   this list is only the set of defaults that are implicitly included.
         */
        function getAutomaticTypeDirectiveNames(options, host) {
            // Use explicit type list from tsconfig.json
            if (options.types) {
                return options.types;
            }
            // Walk the primary type lookup locations
            var result = [];
            if (host.directoryExists && host.getDirectories) {
                var typeRoots = getEffectiveTypeRoots(options, host);
                if (typeRoots) {
                    for (var _i = 0, typeRoots_1 = typeRoots; _i < typeRoots_1.length; _i++) {
                        var root = typeRoots_1[_i];
                        if (host.directoryExists(root)) {
                            for (var _a = 0, _b = host.getDirectories(root); _a < _b.length; _a++) {
                                var typeDirectivePath = _b[_a];
                                var normalized = ts.normalizePath(typeDirectivePath);
                                var packageJsonPath = ts.combinePaths(root, normalized, "package.json");
                                // `types-publisher` sometimes creates packages with `"typings": null` for packages that don't provide their own types.
                                // See `createNotNeededPackageJSON` in the types-publisher` repo.
                                // eslint-disable-next-line no-null/no-null
                                var isNotNeededPackage = host.fileExists(packageJsonPath) && ts.readJson(packageJsonPath, host).typings === null;
                                if (!isNotNeededPackage) {
                                    var baseFileName = ts.getBaseFileName(normalized);
                                    // At this stage, skip results with leading dot.
                                    if (baseFileName.charCodeAt(0) !== 46 /* dot */) {
                                        // Return just the type directive names
                                        result.push(baseFileName);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            return result;
        }
        ts.getAutomaticTypeDirectiveNames = getAutomaticTypeDirectiveNames;
        /*@internal*/
        function createCacheWithRedirects(options) {
            var ownMap = new ts.Map();
            var redirectsMap = new ts.Map();
            return {
                getOwnMap: getOwnMap,
                redirectsMap: redirectsMap,
                getOrCreateMapOfCacheRedirects: getOrCreateMapOfCacheRedirects,
                clear: clear,
                setOwnOptions: setOwnOptions,
                setOwnMap: setOwnMap
            };
            function getOwnMap() {
                return ownMap;
            }
            function setOwnOptions(newOptions) {
                options = newOptions;
            }
            function setOwnMap(newOwnMap) {
                ownMap = newOwnMap;
            }
            function getOrCreateMapOfCacheRedirects(redirectedReference) {
                if (!redirectedReference) {
                    return ownMap;
                }
                var path = redirectedReference.sourceFile.path;
                var redirects = redirectsMap.get(path);
                if (!redirects) {
                    // Reuse map if redirected reference map uses same resolution
                    redirects = !options || ts.optionsHaveModuleResolutionChanges(options, redirectedReference.commandLine.options) ? new ts.Map() : ownMap;
                    redirectsMap.set(path, redirects);
                }
                return redirects;
            }
            function clear() {
                ownMap.clear();
                redirectsMap.clear();
            }
        }
        ts.createCacheWithRedirects = createCacheWithRedirects;
        function createPackageJsonInfoCache(currentDirectory, getCanonicalFileName) {
            var cache;
            return { getPackageJsonInfo: getPackageJsonInfo, setPackageJsonInfo: setPackageJsonInfo, clear: clear, entries: entries };
            function getPackageJsonInfo(packageJsonPath) {
                return cache === null || cache === void 0 ? void 0 : cache.get(ts.toPath(packageJsonPath, currentDirectory, getCanonicalFileName));
            }
            function setPackageJsonInfo(packageJsonPath, info) {
                (cache || (cache = new ts.Map())).set(ts.toPath(packageJsonPath, currentDirectory, getCanonicalFileName), info);
            }
            function clear() {
                cache = undefined;
            }
            function entries() {
                var iter = cache === null || cache === void 0 ? void 0 : cache.entries();
                return iter ? ts.arrayFrom(iter) : [];
            }
        }
        function getOrCreateCache(cacheWithRedirects, redirectedReference, key, create) {
            var cache = cacheWithRedirects.getOrCreateMapOfCacheRedirects(redirectedReference);
            var result = cache.get(key);
            if (!result) {
                result = create();
                cache.set(key, result);
            }
            return result;
        }
        function updateRedirectsMap(options, directoryToModuleNameMap, moduleNameToDirectoryMap) {
            if (!options.configFile)
                return;
            if (directoryToModuleNameMap.redirectsMap.size === 0) {
                // The own map will be for projectCompilerOptions
                ts.Debug.assert(!moduleNameToDirectoryMap || moduleNameToDirectoryMap.redirectsMap.size === 0);
                ts.Debug.assert(directoryToModuleNameMap.getOwnMap().size === 0);
                ts.Debug.assert(!moduleNameToDirectoryMap || moduleNameToDirectoryMap.getOwnMap().size === 0);
                directoryToModuleNameMap.redirectsMap.set(options.configFile.path, directoryToModuleNameMap.getOwnMap());
                moduleNameToDirectoryMap === null || moduleNameToDirectoryMap === void 0 ? void 0 : moduleNameToDirectoryMap.redirectsMap.set(options.configFile.path, moduleNameToDirectoryMap.getOwnMap());
            }
            else {
                // Set correct own map
                ts.Debug.assert(!moduleNameToDirectoryMap || moduleNameToDirectoryMap.redirectsMap.size > 0);
                var ref = {
                    sourceFile: options.configFile,
                    commandLine: { options: options }
                };
                directoryToModuleNameMap.setOwnMap(directoryToModuleNameMap.getOrCreateMapOfCacheRedirects(ref));
                moduleNameToDirectoryMap === null || moduleNameToDirectoryMap === void 0 ? void 0 : moduleNameToDirectoryMap.setOwnMap(moduleNameToDirectoryMap.getOrCreateMapOfCacheRedirects(ref));
            }
            directoryToModuleNameMap.setOwnOptions(options);
            moduleNameToDirectoryMap === null || moduleNameToDirectoryMap === void 0 ? void 0 : moduleNameToDirectoryMap.setOwnOptions(options);
        }
        function createPerDirectoryResolutionCache(currentDirectory, getCanonicalFileName, directoryToModuleNameMap) {
            return {
                getOrCreateCacheForDirectory: getOrCreateCacheForDirectory,
                clear: clear,
                update: update,
            };
            function clear() {
                directoryToModuleNameMap.clear();
            }
            function update(options) {
                updateRedirectsMap(options, directoryToModuleNameMap);
            }
            function getOrCreateCacheForDirectory(directoryName, redirectedReference) {
                var path = ts.toPath(directoryName, currentDirectory, getCanonicalFileName);
                return getOrCreateCache(directoryToModuleNameMap, redirectedReference, path, function () { return new ts.Map(); });
            }
        }
        function createModuleResolutionCache(currentDirectory, getCanonicalFileName, options, directoryToModuleNameMap, moduleNameToDirectoryMap) {
            var preDirectoryResolutionCache = createPerDirectoryResolutionCache(currentDirectory, getCanonicalFileName, directoryToModuleNameMap || (directoryToModuleNameMap = createCacheWithRedirects(options)));
            moduleNameToDirectoryMap || (moduleNameToDirectoryMap = createCacheWithRedirects(options));
            var packageJsonInfoCache = createPackageJsonInfoCache(currentDirectory, getCanonicalFileName);
            return __assign(__assign(__assign({}, packageJsonInfoCache), preDirectoryResolutionCache), { getOrCreateCacheForModuleName: getOrCreateCacheForModuleName, clear: clear, update: update, getPackageJsonInfoCache: function () { return packageJsonInfoCache; } });
            function clear() {
                preDirectoryResolutionCache.clear();
                moduleNameToDirectoryMap.clear();
                packageJsonInfoCache.clear();
            }
            function update(options) {
                updateRedirectsMap(options, directoryToModuleNameMap, moduleNameToDirectoryMap);
            }
            function getOrCreateCacheForModuleName(nonRelativeModuleName, redirectedReference) {
                ts.Debug.assert(!ts.isExternalModuleNameRelative(nonRelativeModuleName));
                return getOrCreateCache(moduleNameToDirectoryMap, redirectedReference, nonRelativeModuleName, createPerModuleNameCache);
            }
            function createPerModuleNameCache() {
                var directoryPathMap = new ts.Map();
                return { get: get, set: set };
                function get(directory) {
                    return directoryPathMap.get(ts.toPath(directory, currentDirectory, getCanonicalFileName));
                }
                /**
                 * At first this function add entry directory -> module resolution result to the table.
                 * Then it computes the set of parent folders for 'directory' that should have the same module resolution result
                 * and for every parent folder in set it adds entry: parent -> module resolution. .
                 * Lets say we first directory name: /a/b/c/d/e and resolution result is: /a/b/bar.ts.
                 * Set of parent folders that should have the same result will be:
                 * [
                 *     /a/b/c/d, /a/b/c, /a/b
                 * ]
                 * this means that request for module resolution from file in any of these folder will be immediately found in cache.
                 */
                function set(directory, result) {
                    var path = ts.toPath(directory, currentDirectory, getCanonicalFileName);
                    // if entry is already in cache do nothing
                    if (directoryPathMap.has(path)) {
                        return;
                    }
                    directoryPathMap.set(path, result);
                    var resolvedFileName = result.resolvedModule &&
                        (result.resolvedModule.originalPath || result.resolvedModule.resolvedFileName);
                    // find common prefix between directory and resolved file name
                    // this common prefix should be the shortest path that has the same resolution
                    // directory: /a/b/c/d/e
                    // resolvedFileName: /a/b/foo.d.ts
                    // commonPrefix: /a/b
                    // for failed lookups cache the result for every directory up to root
                    var commonPrefix = resolvedFileName && getCommonPrefix(path, resolvedFileName);
                    var current = path;
                    while (current !== commonPrefix) {
                        var parent = ts.getDirectoryPath(current);
                        if (parent === current || directoryPathMap.has(parent)) {
                            break;
                        }
                        directoryPathMap.set(parent, result);
                        current = parent;
                    }
                }
                function getCommonPrefix(directory, resolution) {
                    var resolutionDirectory = ts.toPath(ts.getDirectoryPath(resolution), currentDirectory, getCanonicalFileName);
                    // find first position where directory and resolution differs
                    var i = 0;
                    var limit = Math.min(directory.length, resolutionDirectory.length);
                    while (i < limit && directory.charCodeAt(i) === resolutionDirectory.charCodeAt(i)) {
                        i++;
                    }
                    if (i === directory.length && (resolutionDirectory.length === i || resolutionDirectory[i] === ts.directorySeparator)) {
                        return directory;
                    }
                    var rootLength = ts.getRootLength(directory);
                    if (i < rootLength) {
                        return undefined;
                    }
                    var sep = directory.lastIndexOf(ts.directorySeparator, i - 1);
                    if (sep === -1) {
                        return undefined;
                    }
                    return directory.substr(0, Math.max(sep, rootLength));
                }
            }
        }
        ts.createModuleResolutionCache = createModuleResolutionCache;
        function createTypeReferenceDirectiveResolutionCache(currentDirectory, getCanonicalFileName, options, packageJsonInfoCache, directoryToModuleNameMap) {
            var preDirectoryResolutionCache = createPerDirectoryResolutionCache(currentDirectory, getCanonicalFileName, directoryToModuleNameMap || (directoryToModuleNameMap = createCacheWithRedirects(options)));
            packageJsonInfoCache || (packageJsonInfoCache = createPackageJsonInfoCache(currentDirectory, getCanonicalFileName));
            return __assign(__assign(__assign({}, packageJsonInfoCache), preDirectoryResolutionCache), { clear: clear });
            function clear() {
                preDirectoryResolutionCache.clear();
                packageJsonInfoCache.clear();
            }
        }
        ts.createTypeReferenceDirectiveResolutionCache = createTypeReferenceDirectiveResolutionCache;
        function resolveModuleNameFromCache(moduleName, containingFile, cache) {
            var containingDirectory = ts.getDirectoryPath(containingFile);
            var perFolderCache = cache && cache.getOrCreateCacheForDirectory(containingDirectory);
            return perFolderCache && perFolderCache.get(moduleName);
        }
        ts.resolveModuleNameFromCache = resolveModuleNameFromCache;
        function resolveModuleName(moduleName, containingFile, compilerOptions, host, cache, redirectedReference) {
            var traceEnabled = isTraceEnabled(compilerOptions, host);
            if (redirectedReference) {
                compilerOptions = redirectedReference.commandLine.options;
            }
            if (traceEnabled) {
                trace(host, ts.Diagnostics.Resolving_module_0_from_1, moduleName, containingFile);
                if (redirectedReference) {
                    trace(host, ts.Diagnostics.Using_compiler_options_of_project_reference_redirect_0, redirectedReference.sourceFile.fileName);
                }
            }
            var containingDirectory = ts.getDirectoryPath(containingFile);
            var perFolderCache = cache && cache.getOrCreateCacheForDirectory(containingDirectory, redirectedReference);
            var result = perFolderCache && perFolderCache.get(moduleName);
            if (result) {
                if (traceEnabled) {
                    trace(host, ts.Diagnostics.Resolution_for_module_0_was_found_in_cache_from_location_1, moduleName, containingDirectory);
                }
            }
            else {
                var moduleResolution = compilerOptions.moduleResolution;
                if (moduleResolution === undefined) {
                    moduleResolution = ts.getEmitModuleKind(compilerOptions) === ts.ModuleKind.CommonJS ? ts.ModuleResolutionKind.NodeJs : ts.ModuleResolutionKind.Classic;
                    if (traceEnabled) {
                        trace(host, ts.Diagnostics.Module_resolution_kind_is_not_specified_using_0, ts.ModuleResolutionKind[moduleResolution]);
                    }
                }
                else {
                    if (traceEnabled) {
                        trace(host, ts.Diagnostics.Explicitly_specified_module_resolution_kind_Colon_0, ts.ModuleResolutionKind[moduleResolution]);
                    }
                }
                ts.perfLogger.logStartResolveModule(moduleName /* , containingFile, ModuleResolutionKind[moduleResolution]*/);
                switch (moduleResolution) {
                    case ts.ModuleResolutionKind.NodeJs:
                        result = nodeModuleNameResolver(moduleName, containingFile, compilerOptions, host, cache, redirectedReference);
                        break;
                    case ts.ModuleResolutionKind.Classic:
                        result = classicNameResolver(moduleName, containingFile, compilerOptions, host, cache, redirectedReference);
                        break;
                    default:
                        return ts.Debug.fail("Unexpected moduleResolution: " + moduleResolution);
                }
                if (result && result.resolvedModule)
                    ts.perfLogger.logInfoEvent("Module \"" + moduleName + "\" resolved to \"" + result.resolvedModule.resolvedFileName + "\"");
                ts.perfLogger.logStopResolveModule((result && result.resolvedModule) ? "" + result.resolvedModule.resolvedFileName : "null");
                if (perFolderCache) {
                    perFolderCache.set(moduleName, result);
                    if (!ts.isExternalModuleNameRelative(moduleName)) {
                        // put result in per-module name cache
                        cache.getOrCreateCacheForModuleName(moduleName, redirectedReference).set(containingDirectory, result);
                    }
                }
            }
            if (traceEnabled) {
                if (result.resolvedModule) {
                    if (result.resolvedModule.packageId) {
                        trace(host, ts.Diagnostics.Module_name_0_was_successfully_resolved_to_1_with_Package_ID_2, moduleName, result.resolvedModule.resolvedFileName, ts.packageIdToString(result.resolvedModule.packageId));
                    }
                    else {
                        trace(host, ts.Diagnostics.Module_name_0_was_successfully_resolved_to_1, moduleName, result.resolvedModule.resolvedFileName);
                    }
                }
                else {
                    trace(host, ts.Diagnostics.Module_name_0_was_not_resolved, moduleName);
                }
            }
            return result;
        }
        ts.resolveModuleName = resolveModuleName;
        /**
         * Any module resolution kind can be augmented with optional settings: 'baseUrl', 'paths' and 'rootDirs' - they are used to
         * mitigate differences between design time structure of the project and its runtime counterpart so the same import name
         * can be resolved successfully by TypeScript compiler and runtime module loader.
         * If these settings are set then loading procedure will try to use them to resolve module name and it can of failure it will
         * fallback to standard resolution routine.
         *
         * - baseUrl - this setting controls how non-relative module names are resolved. If this setting is specified then non-relative
         * names will be resolved relative to baseUrl: i.e. if baseUrl is '/a/b' then candidate location to resolve module name 'c/d' will
         * be '/a/b/c/d'
         * - paths - this setting can only be used when baseUrl is specified. allows to tune how non-relative module names
         * will be resolved based on the content of the module name.
         * Structure of 'paths' compiler options
         * 'paths': {
         *    pattern-1: [...substitutions],
         *    pattern-2: [...substitutions],
         *    ...
         *    pattern-n: [...substitutions]
         * }
         * Pattern here is a string that can contain zero or one '*' character. During module resolution module name will be matched against
         * all patterns in the list. Matching for patterns that don't contain '*' means that module name must be equal to pattern respecting the case.
         * If pattern contains '*' then to match pattern "<prefix>*<suffix>" module name must start with the <prefix> and end with <suffix>.
         * <MatchedStar> denotes part of the module name between <prefix> and <suffix>.
         * If module name can be matches with multiple patterns then pattern with the longest prefix will be picked.
         * After selecting pattern we'll use list of substitutions to get candidate locations of the module and the try to load module
         * from the candidate location.
         * Substitution is a string that can contain zero or one '*'. To get candidate location from substitution we'll pick every
         * substitution in the list and replace '*' with <MatchedStar> string. If candidate location is not rooted it
         * will be converted to absolute using baseUrl.
         * For example:
         * baseUrl: /a/b/c
         * "paths": {
         *     // match all module names
         *     "*": [
         *         "*",        // use matched name as is,
         *                     // <matched name> will be looked as /a/b/c/<matched name>
         *
         *         "folder1/*" // substitution will convert matched name to 'folder1/<matched name>',
         *                     // since it is not rooted then final candidate location will be /a/b/c/folder1/<matched name>
         *     ],
         *     // match module names that start with 'components/'
         *     "components/*": [ "/root/components/*" ] // substitution will convert /components/folder1/<matched name> to '/root/components/folder1/<matched name>',
         *                                              // it is rooted so it will be final candidate location
         * }
         *
         * 'rootDirs' allows the project to be spreaded across multiple locations and resolve modules with relative names as if
         * they were in the same location. For example lets say there are two files
         * '/local/src/content/file1.ts'
         * '/shared/components/contracts/src/content/protocols/file2.ts'
         * After bundling content of '/shared/components/contracts/src' will be merged with '/local/src' so
         * if file1 has the following import 'import {x} from "./protocols/file2"' it will be resolved successfully in runtime.
         * 'rootDirs' provides the way to tell compiler that in order to get the whole project it should behave as if content of all
         * root dirs were merged together.
         * I.e. for the example above 'rootDirs' will have two entries: [ '/local/src', '/shared/components/contracts/src' ].
         * Compiler will first convert './protocols/file2' into absolute path relative to the location of containing file:
         * '/local/src/content/protocols/file2' and try to load it - failure.
         * Then it will search 'rootDirs' looking for a longest matching prefix of this absolute path and if such prefix is found - absolute path will
         * be converted to a path relative to found rootDir entry './content/protocols/file2' (*). As a last step compiler will check all remaining
         * entries in 'rootDirs', use them to build absolute path out of (*) and try to resolve module from this location.
         */
        function tryLoadModuleUsingOptionalResolutionSettings(extensions, moduleName, containingDirectory, loader, state) {
            var resolved = tryLoadModuleUsingPathsIfEligible(extensions, moduleName, loader, state);
            if (resolved)
                return resolved.value;
            if (!ts.isExternalModuleNameRelative(moduleName)) {
                return tryLoadModuleUsingBaseUrl(extensions, moduleName, loader, state);
            }
            else {
                return tryLoadModuleUsingRootDirs(extensions, moduleName, containingDirectory, loader, state);
            }
        }
        function tryLoadModuleUsingPathsIfEligible(extensions, moduleName, loader, state) {
            var _a;
            var _b = state.compilerOptions, baseUrl = _b.baseUrl, paths = _b.paths, configFile = _b.configFile;
            if (paths && !ts.pathIsRelative(moduleName)) {
                if (state.traceEnabled) {
                    if (baseUrl) {
                        trace(state.host, ts.Diagnostics.baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1, baseUrl, moduleName);
                    }
                    trace(state.host, ts.Diagnostics.paths_option_is_specified_looking_for_a_pattern_to_match_module_name_0, moduleName);
                }
                var baseDirectory = ts.getPathsBasePath(state.compilerOptions, state.host); // Always defined when 'paths' is defined
                var pathPatterns = (configFile === null || configFile === void 0 ? void 0 : configFile.configFileSpecs) ? (_a = configFile.configFileSpecs).pathPatterns || (_a.pathPatterns = ts.tryParsePatterns(paths)) : undefined;
                return tryLoadModuleUsingPaths(extensions, moduleName, baseDirectory, paths, pathPatterns, loader, /*onlyRecordFailures*/ false, state);
            }
        }
        function tryLoadModuleUsingRootDirs(extensions, moduleName, containingDirectory, loader, state) {
            if (!state.compilerOptions.rootDirs) {
                return undefined;
            }
            if (state.traceEnabled) {
                trace(state.host, ts.Diagnostics.rootDirs_option_is_set_using_it_to_resolve_relative_module_name_0, moduleName);
            }
            var candidate = ts.normalizePath(ts.combinePaths(containingDirectory, moduleName));
            var matchedRootDir;
            var matchedNormalizedPrefix;
            for (var _i = 0, _a = state.compilerOptions.rootDirs; _i < _a.length; _i++) {
                var rootDir = _a[_i];
                // rootDirs are expected to be absolute
                // in case of tsconfig.json this will happen automatically - compiler will expand relative names
                // using location of tsconfig.json as base location
                var normalizedRoot = ts.normalizePath(rootDir);
                if (!ts.endsWith(normalizedRoot, ts.directorySeparator)) {
                    normalizedRoot += ts.directorySeparator;
                }
                var isLongestMatchingPrefix = ts.startsWith(candidate, normalizedRoot) &&
                    (matchedNormalizedPrefix === undefined || matchedNormalizedPrefix.length < normalizedRoot.length);
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.Checking_if_0_is_the_longest_matching_prefix_for_1_2, normalizedRoot, candidate, isLongestMatchingPrefix);
                }
                if (isLongestMatchingPrefix) {
                    matchedNormalizedPrefix = normalizedRoot;
                    matchedRootDir = rootDir;
                }
            }
            if (matchedNormalizedPrefix) {
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.Longest_matching_prefix_for_0_is_1, candidate, matchedNormalizedPrefix);
                }
                var suffix = candidate.substr(matchedNormalizedPrefix.length);
                // first - try to load from a initial location
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.Loading_0_from_the_root_dir_1_candidate_location_2, suffix, matchedNormalizedPrefix, candidate);
                }
                var resolvedFileName = loader(extensions, candidate, !ts.directoryProbablyExists(containingDirectory, state.host), state);
                if (resolvedFileName) {
                    return resolvedFileName;
                }
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.Trying_other_entries_in_rootDirs);
                }
                // then try to resolve using remaining entries in rootDirs
                for (var _b = 0, _c = state.compilerOptions.rootDirs; _b < _c.length; _b++) {
                    var rootDir = _c[_b];
                    if (rootDir === matchedRootDir) {
                        // skip the initially matched entry
                        continue;
                    }
                    var candidate_1 = ts.combinePaths(ts.normalizePath(rootDir), suffix);
                    if (state.traceEnabled) {
                        trace(state.host, ts.Diagnostics.Loading_0_from_the_root_dir_1_candidate_location_2, suffix, rootDir, candidate_1);
                    }
                    var baseDirectory = ts.getDirectoryPath(candidate_1);
                    var resolvedFileName_1 = loader(extensions, candidate_1, !ts.directoryProbablyExists(baseDirectory, state.host), state);
                    if (resolvedFileName_1) {
                        return resolvedFileName_1;
                    }
                }
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.Module_resolution_using_rootDirs_has_failed);
                }
            }
            return undefined;
        }
        function tryLoadModuleUsingBaseUrl(extensions, moduleName, loader, state) {
            var baseUrl = state.compilerOptions.baseUrl;
            if (!baseUrl) {
                return undefined;
            }
            if (state.traceEnabled) {
                trace(state.host, ts.Diagnostics.baseUrl_option_is_set_to_0_using_this_value_to_resolve_non_relative_module_name_1, baseUrl, moduleName);
            }
            var candidate = ts.normalizePath(ts.combinePaths(baseUrl, moduleName));
            if (state.traceEnabled) {
                trace(state.host, ts.Diagnostics.Resolving_module_name_0_relative_to_base_url_1_2, moduleName, baseUrl, candidate);
            }
            return loader(extensions, candidate, !ts.directoryProbablyExists(ts.getDirectoryPath(candidate), state.host), state);
        }
        /**
         * Expose resolution logic to allow us to use Node module resolution logic from arbitrary locations.
         * No way to do this with `require()`: https://github.com/nodejs/node/issues/5963
         * Throws an error if the module can't be resolved.
         */
        /* @internal */
        function resolveJSModule(moduleName, initialDir, host) {
            var _a = tryResolveJSModuleWorker(moduleName, initialDir, host), resolvedModule = _a.resolvedModule, failedLookupLocations = _a.failedLookupLocations;
            if (!resolvedModule) {
                throw new Error("Could not resolve JS module '" + moduleName + "' starting at '" + initialDir + "'. Looked in: " + failedLookupLocations.join(", "));
            }
            return resolvedModule.resolvedFileName;
        }
        ts.resolveJSModule = resolveJSModule;
        /* @internal */
        function tryResolveJSModule(moduleName, initialDir, host) {
            var resolvedModule = tryResolveJSModuleWorker(moduleName, initialDir, host).resolvedModule;
            return resolvedModule && resolvedModule.resolvedFileName;
        }
        ts.tryResolveJSModule = tryResolveJSModule;
        var jsOnlyExtensions = [Extensions.JavaScript];
        var tsExtensions = [Extensions.TypeScript, Extensions.JavaScript];
        var tsPlusJsonExtensions = __spreadArray(__spreadArray([], tsExtensions, true), [Extensions.Json], false);
        var tsconfigExtensions = [Extensions.TSConfig];
        function tryResolveJSModuleWorker(moduleName, initialDir, host) {
            return nodeModuleNameResolverWorker(moduleName, initialDir, { moduleResolution: ts.ModuleResolutionKind.NodeJs, allowJs: true }, host, /*cache*/ undefined, jsOnlyExtensions, /*redirectedReferences*/ undefined);
        }
        function nodeModuleNameResolver(moduleName, containingFile, compilerOptions, host, cache, redirectedReference, lookupConfig) {
            return nodeModuleNameResolverWorker(moduleName, ts.getDirectoryPath(containingFile), compilerOptions, host, cache, lookupConfig ? tsconfigExtensions : (compilerOptions.resolveJsonModule ? tsPlusJsonExtensions : tsExtensions), redirectedReference);
        }
        ts.nodeModuleNameResolver = nodeModuleNameResolver;
        function nodeModuleNameResolverWorker(moduleName, containingDirectory, compilerOptions, host, cache, extensions, redirectedReference) {
            var _a, _b;
            var traceEnabled = isTraceEnabled(compilerOptions, host);
            var failedLookupLocations = [];
            var state = { compilerOptions: compilerOptions, host: host, traceEnabled: traceEnabled, failedLookupLocations: failedLookupLocations, packageJsonInfoCache: cache };
            var result = ts.forEach(extensions, function (ext) { return tryResolve(ext); });
            return createResolvedModuleWithFailedLookupLocations((_a = result === null || result === void 0 ? void 0 : result.value) === null || _a === void 0 ? void 0 : _a.resolved, (_b = result === null || result === void 0 ? void 0 : result.value) === null || _b === void 0 ? void 0 : _b.isExternalLibraryImport, failedLookupLocations, state.resultFromCache);
            function tryResolve(extensions) {
                var loader = function (extensions, candidate, onlyRecordFailures, state) { return nodeLoadModuleByRelativeName(extensions, candidate, onlyRecordFailures, state, /*considerPackageJson*/ true); };
                var resolved = tryLoadModuleUsingOptionalResolutionSettings(extensions, moduleName, containingDirectory, loader, state);
                if (resolved) {
                    return toSearchResult({ resolved: resolved, isExternalLibraryImport: pathContainsNodeModules(resolved.path) });
                }
                if (!ts.isExternalModuleNameRelative(moduleName)) {
                    if (traceEnabled) {
                        trace(host, ts.Diagnostics.Loading_module_0_from_node_modules_folder_target_file_type_1, moduleName, Extensions[extensions]);
                    }
                    var resolved_1 = loadModuleFromNearestNodeModulesDirectory(extensions, moduleName, containingDirectory, state, cache, redirectedReference);
                    if (!resolved_1)
                        return undefined;
                    var resolvedValue = resolved_1.value;
                    if (!compilerOptions.preserveSymlinks && resolvedValue && !resolvedValue.originalPath) {
                        var path = realPath(resolvedValue.path, host, traceEnabled);
                        var originalPath = path === resolvedValue.path ? undefined : resolvedValue.path;
                        resolvedValue = __assign(__assign({}, resolvedValue), { path: path, originalPath: originalPath });
                    }
                    // For node_modules lookups, get the real path so that multiple accesses to an `npm link`-ed module do not create duplicate files.
                    return { value: resolvedValue && { resolved: resolvedValue, isExternalLibraryImport: true } };
                }
                else {
                    var _a = ts.normalizePathAndParts(ts.combinePaths(containingDirectory, moduleName)), candidate = _a.path, parts = _a.parts;
                    var resolved_2 = nodeLoadModuleByRelativeName(extensions, candidate, /*onlyRecordFailures*/ false, state, /*considerPackageJson*/ true);
                    // Treat explicit "node_modules" import as an external library import.
                    return resolved_2 && toSearchResult({ resolved: resolved_2, isExternalLibraryImport: ts.contains(parts, "node_modules") });
                }
            }
        }
        function realPath(path, host, traceEnabled) {
            if (!host.realpath) {
                return path;
            }
            var real = ts.normalizePath(host.realpath(path));
            if (traceEnabled) {
                trace(host, ts.Diagnostics.Resolving_real_path_for_0_result_1, path, real);
            }
            ts.Debug.assert(host.fileExists(real), path + " linked to nonexistent file " + real);
            return real;
        }
        function nodeLoadModuleByRelativeName(extensions, candidate, onlyRecordFailures, state, considerPackageJson) {
            if (state.traceEnabled) {
                trace(state.host, ts.Diagnostics.Loading_module_as_file_Slash_folder_candidate_module_location_0_target_file_type_1, candidate, Extensions[extensions]);
            }
            if (!ts.hasTrailingDirectorySeparator(candidate)) {
                if (!onlyRecordFailures) {
                    var parentOfCandidate = ts.getDirectoryPath(candidate);
                    if (!ts.directoryProbablyExists(parentOfCandidate, state.host)) {
                        if (state.traceEnabled) {
                            trace(state.host, ts.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it, parentOfCandidate);
                        }
                        onlyRecordFailures = true;
                    }
                }
                var resolvedFromFile = loadModuleFromFile(extensions, candidate, onlyRecordFailures, state);
                if (resolvedFromFile) {
                    var packageDirectory = considerPackageJson ? parseNodeModuleFromPath(resolvedFromFile.path) : undefined;
                    var packageInfo = packageDirectory ? getPackageJsonInfo(packageDirectory, /*onlyRecordFailures*/ false, state) : undefined;
                    return withPackageId(packageInfo, resolvedFromFile);
                }
            }
            if (!onlyRecordFailures) {
                var candidateExists = ts.directoryProbablyExists(candidate, state.host);
                if (!candidateExists) {
                    if (state.traceEnabled) {
                        trace(state.host, ts.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it, candidate);
                    }
                    onlyRecordFailures = true;
                }
            }
            return loadNodeModuleFromDirectory(extensions, candidate, onlyRecordFailures, state, considerPackageJson);
        }
        /*@internal*/
        ts.nodeModulesPathPart = "/node_modules/";
        /*@internal*/
        function pathContainsNodeModules(path) {
            return ts.stringContains(path, ts.nodeModulesPathPart);
        }
        ts.pathContainsNodeModules = pathContainsNodeModules;
        /**
         * This will be called on the successfully resolved path from `loadModuleFromFile`.
         * (Not needed for `loadModuleFromNodeModules` as that looks up the `package.json` as part of resolution.)
         *
         * packageDirectory is the directory of the package itself.
         *   For `blah/node_modules/foo/index.d.ts` this is packageDirectory: "foo"
         *   For `/node_modules/foo/bar.d.ts` this is packageDirectory: "foo"
         *   For `/node_modules/@types/foo/bar/index.d.ts` this is packageDirectory: "@types/foo"
         *   For `/node_modules/foo/bar/index.d.ts` this is packageDirectory: "foo"
         */
        /* @internal */
        function parseNodeModuleFromPath(resolved) {
            var path = ts.normalizePath(resolved);
            var idx = path.lastIndexOf(ts.nodeModulesPathPart);
            if (idx === -1) {
                return undefined;
            }
            var indexAfterNodeModules = idx + ts.nodeModulesPathPart.length;
            var indexAfterPackageName = moveToNextDirectorySeparatorIfAvailable(path, indexAfterNodeModules);
            if (path.charCodeAt(indexAfterNodeModules) === 64 /* at */) {
                indexAfterPackageName = moveToNextDirectorySeparatorIfAvailable(path, indexAfterPackageName);
            }
            return path.slice(0, indexAfterPackageName);
        }
        ts.parseNodeModuleFromPath = parseNodeModuleFromPath;
        function moveToNextDirectorySeparatorIfAvailable(path, prevSeparatorIndex) {
            var nextSeparatorIndex = path.indexOf(ts.directorySeparator, prevSeparatorIndex + 1);
            return nextSeparatorIndex === -1 ? prevSeparatorIndex : nextSeparatorIndex;
        }
        function loadModuleFromFileNoPackageId(extensions, candidate, onlyRecordFailures, state) {
            return noPackageId(loadModuleFromFile(extensions, candidate, onlyRecordFailures, state));
        }
        /**
         * @param {boolean} onlyRecordFailures - if true then function won't try to actually load files but instead record all attempts as failures. This flag is necessary
         * in cases when we know upfront that all load attempts will fail (because containing folder does not exists) however we still need to record all failed lookup locations.
         */
        function loadModuleFromFile(extensions, candidate, onlyRecordFailures, state) {
            if (extensions === Extensions.Json || extensions === Extensions.TSConfig) {
                var extensionLess = ts.tryRemoveExtension(candidate, ".json" /* Json */);
                return (extensionLess === undefined && extensions === Extensions.Json) ? undefined : tryAddingExtensions(extensionLess || candidate, extensions, onlyRecordFailures, state);
            }
            // First, try adding an extension. An import of "foo" could be matched by a file "foo.ts", or "foo.js" by "foo.js.ts"
            var resolvedByAddingExtension = tryAddingExtensions(candidate, extensions, onlyRecordFailures, state);
            if (resolvedByAddingExtension) {
                return resolvedByAddingExtension;
            }
            // If that didn't work, try stripping a ".js" or ".jsx" extension and replacing it with a TypeScript one;
            // e.g. "./foo.js" can be matched by "./foo.ts" or "./foo.d.ts"
            if (ts.hasJSFileExtension(candidate)) {
                var extensionless = ts.removeFileExtension(candidate);
                if (state.traceEnabled) {
                    var extension = candidate.substring(extensionless.length);
                    trace(state.host, ts.Diagnostics.File_name_0_has_a_1_extension_stripping_it, candidate, extension);
                }
                return tryAddingExtensions(extensionless, extensions, onlyRecordFailures, state);
            }
        }
        /** Try to return an existing file that adds one of the `extensions` to `candidate`. */
        function tryAddingExtensions(candidate, extensions, onlyRecordFailures, state) {
            if (!onlyRecordFailures) {
                // check if containing folder exists - if it doesn't then just record failures for all supported extensions without disk probing
                var directory = ts.getDirectoryPath(candidate);
                if (directory) {
                    onlyRecordFailures = !ts.directoryProbablyExists(directory, state.host);
                }
            }
            switch (extensions) {
                case Extensions.DtsOnly:
                    return tryExtension(".d.ts" /* Dts */);
                case Extensions.TypeScript:
                    return tryExtension(".ts" /* Ts */) || tryExtension(".tsx" /* Tsx */) || tryExtension(".d.ts" /* Dts */);
                case Extensions.JavaScript:
                    return tryExtension(".js" /* Js */) || tryExtension(".jsx" /* Jsx */);
                case Extensions.TSConfig:
                case Extensions.Json:
                    return tryExtension(".json" /* Json */);
            }
            function tryExtension(ext) {
                var path = tryFile(candidate + ext, onlyRecordFailures, state);
                return path === undefined ? undefined : { path: path, ext: ext };
            }
        }
        /** Return the file if it exists. */
        function tryFile(fileName, onlyRecordFailures, state) {
            if (!onlyRecordFailures) {
                if (state.host.fileExists(fileName)) {
                    if (state.traceEnabled) {
                        trace(state.host, ts.Diagnostics.File_0_exist_use_it_as_a_name_resolution_result, fileName);
                    }
                    return fileName;
                }
                else {
                    if (state.traceEnabled) {
                        trace(state.host, ts.Diagnostics.File_0_does_not_exist, fileName);
                    }
                }
            }
            state.failedLookupLocations.push(fileName);
            return undefined;
        }
        function loadNodeModuleFromDirectory(extensions, candidate, onlyRecordFailures, state, considerPackageJson) {
            if (considerPackageJson === void 0) {
                considerPackageJson = true;
            }
            var packageInfo = considerPackageJson ? getPackageJsonInfo(candidate, onlyRecordFailures, state) : undefined;
            var packageJsonContent = packageInfo && packageInfo.packageJsonContent;
            var versionPaths = packageInfo && packageInfo.versionPaths;
            return withPackageId(packageInfo, loadNodeModuleFromDirectoryWorker(extensions, candidate, onlyRecordFailures, state, packageJsonContent, versionPaths));
        }
        function getPackageJsonInfo(packageDirectory, onlyRecordFailures, state) {
            var _a, _b, _c;
            var host = state.host, traceEnabled = state.traceEnabled;
            var packageJsonPath = ts.combinePaths(packageDirectory, "package.json");
            if (onlyRecordFailures) {
                state.failedLookupLocations.push(packageJsonPath);
                return undefined;
            }
            var existing = (_a = state.packageJsonInfoCache) === null || _a === void 0 ? void 0 : _a.getPackageJsonInfo(packageJsonPath);
            if (existing !== undefined) {
                if (typeof existing !== "boolean") {
                    if (traceEnabled)
                        trace(host, ts.Diagnostics.File_0_exists_according_to_earlier_cached_lookups, packageJsonPath);
                    return existing;
                }
                else {
                    if (existing && traceEnabled)
                        trace(host, ts.Diagnostics.File_0_does_not_exist_according_to_earlier_cached_lookups, packageJsonPath);
                    state.failedLookupLocations.push(packageJsonPath);
                    return undefined;
                }
            }
            var directoryExists = ts.directoryProbablyExists(packageDirectory, host);
            if (directoryExists && host.fileExists(packageJsonPath)) {
                var packageJsonContent = ts.readJson(packageJsonPath, host);
                if (traceEnabled) {
                    trace(host, ts.Diagnostics.Found_package_json_at_0, packageJsonPath);
                }
                var versionPaths = readPackageJsonTypesVersionPaths(packageJsonContent, state);
                var result = { packageDirectory: packageDirectory, packageJsonContent: packageJsonContent, versionPaths: versionPaths };
                (_b = state.packageJsonInfoCache) === null || _b === void 0 ? void 0 : _b.setPackageJsonInfo(packageJsonPath, result);
                return result;
            }
            else {
                if (directoryExists && traceEnabled) {
                    trace(host, ts.Diagnostics.File_0_does_not_exist, packageJsonPath);
                }
                (_c = state.packageJsonInfoCache) === null || _c === void 0 ? void 0 : _c.setPackageJsonInfo(packageJsonPath, directoryExists);
                // record package json as one of failed lookup locations - in the future if this file will appear it will invalidate resolution results
                state.failedLookupLocations.push(packageJsonPath);
            }
        }
        function loadNodeModuleFromDirectoryWorker(extensions, candidate, onlyRecordFailures, state, jsonContent, versionPaths) {
            var packageFile;
            if (jsonContent) {
                switch (extensions) {
                    case Extensions.JavaScript:
                    case Extensions.Json:
                        packageFile = readPackageJsonMainField(jsonContent, candidate, state);
                        break;
                    case Extensions.TypeScript:
                        // When resolving typescript modules, try resolving using main field as well
                        packageFile = readPackageJsonTypesFields(jsonContent, candidate, state) || readPackageJsonMainField(jsonContent, candidate, state);
                        break;
                    case Extensions.DtsOnly:
                        packageFile = readPackageJsonTypesFields(jsonContent, candidate, state);
                        break;
                    case Extensions.TSConfig:
                        packageFile = readPackageJsonTSConfigField(jsonContent, candidate, state);
                        break;
                    default:
                        return ts.Debug.assertNever(extensions);
                }
            }
            var loader = function (extensions, candidate, onlyRecordFailures, state) {
                var fromFile = tryFile(candidate, onlyRecordFailures, state);
                if (fromFile) {
                    var resolved = resolvedIfExtensionMatches(extensions, fromFile);
                    if (resolved) {
                        return noPackageId(resolved);
                    }
                    if (state.traceEnabled) {
                        trace(state.host, ts.Diagnostics.File_0_has_an_unsupported_extension_so_skipping_it, fromFile);
                    }
                }
                // Even if extensions is DtsOnly, we can still look up a .ts file as a result of package.json "types"
                var nextExtensions = extensions === Extensions.DtsOnly ? Extensions.TypeScript : extensions;
                // Don't do package.json lookup recursively, because Node.js' package lookup doesn't.
                return nodeLoadModuleByRelativeName(nextExtensions, candidate, onlyRecordFailures, state, /*considerPackageJson*/ false);
            };
            var onlyRecordFailuresForPackageFile = packageFile ? !ts.directoryProbablyExists(ts.getDirectoryPath(packageFile), state.host) : undefined;
            var onlyRecordFailuresForIndex = onlyRecordFailures || !ts.directoryProbablyExists(candidate, state.host);
            var indexPath = ts.combinePaths(candidate, extensions === Extensions.TSConfig ? "tsconfig" : "index");
            if (versionPaths && (!packageFile || ts.containsPath(candidate, packageFile))) {
                var moduleName = ts.getRelativePathFromDirectory(candidate, packageFile || indexPath, /*ignoreCase*/ false);
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_match_module_name_2, versionPaths.version, ts.version, moduleName);
                }
                var result = tryLoadModuleUsingPaths(extensions, moduleName, candidate, versionPaths.paths, /*pathPatterns*/ undefined, loader, onlyRecordFailuresForPackageFile || onlyRecordFailuresForIndex, state);
                if (result) {
                    return removeIgnoredPackageId(result.value);
                }
            }
            // It won't have a `packageId` set, because we disabled `considerPackageJson`.
            var packageFileResult = packageFile && removeIgnoredPackageId(loader(extensions, packageFile, onlyRecordFailuresForPackageFile, state));
            if (packageFileResult)
                return packageFileResult;
            return loadModuleFromFile(extensions, indexPath, onlyRecordFailuresForIndex, state);
        }
        /** Resolve from an arbitrarily specified file. Return `undefined` if it has an unsupported extension. */
        function resolvedIfExtensionMatches(extensions, path) {
            var ext = ts.tryGetExtensionFromPath(path);
            return ext !== undefined && extensionIsOk(extensions, ext) ? { path: path, ext: ext } : undefined;
        }
        /** True if `extension` is one of the supported `extensions`. */
        function extensionIsOk(extensions, extension) {
            switch (extensions) {
                case Extensions.JavaScript:
                    return extension === ".js" /* Js */ || extension === ".jsx" /* Jsx */;
                case Extensions.TSConfig:
                case Extensions.Json:
                    return extension === ".json" /* Json */;
                case Extensions.TypeScript:
                    return extension === ".ts" /* Ts */ || extension === ".tsx" /* Tsx */ || extension === ".d.ts" /* Dts */;
                case Extensions.DtsOnly:
                    return extension === ".d.ts" /* Dts */;
            }
        }
        /* @internal */
        function parsePackageName(moduleName) {
            var idx = moduleName.indexOf(ts.directorySeparator);
            if (moduleName[0] === "@") {
                idx = moduleName.indexOf(ts.directorySeparator, idx + 1);
            }
            return idx === -1 ? { packageName: moduleName, rest: "" } : { packageName: moduleName.slice(0, idx), rest: moduleName.slice(idx + 1) };
        }
        ts.parsePackageName = parsePackageName;
        function loadModuleFromNearestNodeModulesDirectory(extensions, moduleName, directory, state, cache, redirectedReference) {
            return loadModuleFromNearestNodeModulesDirectoryWorker(extensions, moduleName, directory, state, /*typesScopeOnly*/ false, cache, redirectedReference);
        }
        function loadModuleFromNearestNodeModulesDirectoryTypesScope(moduleName, directory, state) {
            // Extensions parameter here doesn't actually matter, because typesOnly ensures we're just doing @types lookup, which is always DtsOnly.
            return loadModuleFromNearestNodeModulesDirectoryWorker(Extensions.DtsOnly, moduleName, directory, state, /*typesScopeOnly*/ true, /*cache*/ undefined, /*redirectedReference*/ undefined);
        }
        function loadModuleFromNearestNodeModulesDirectoryWorker(extensions, moduleName, directory, state, typesScopeOnly, cache, redirectedReference) {
            var perModuleNameCache = cache && cache.getOrCreateCacheForModuleName(moduleName, redirectedReference);
            return ts.forEachAncestorDirectory(ts.normalizeSlashes(directory), function (ancestorDirectory) {
                if (ts.getBaseFileName(ancestorDirectory) !== "node_modules") {
                    var resolutionFromCache = tryFindNonRelativeModuleNameInCache(perModuleNameCache, moduleName, ancestorDirectory, state);
                    if (resolutionFromCache) {
                        return resolutionFromCache;
                    }
                    return toSearchResult(loadModuleFromImmediateNodeModulesDirectory(extensions, moduleName, ancestorDirectory, state, typesScopeOnly));
                }
            });
        }
        function loadModuleFromImmediateNodeModulesDirectory(extensions, moduleName, directory, state, typesScopeOnly) {
            var nodeModulesFolder = ts.combinePaths(directory, "node_modules");
            var nodeModulesFolderExists = ts.directoryProbablyExists(nodeModulesFolder, state.host);
            if (!nodeModulesFolderExists && state.traceEnabled) {
                trace(state.host, ts.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it, nodeModulesFolder);
            }
            var packageResult = typesScopeOnly ? undefined : loadModuleFromSpecificNodeModulesDirectory(extensions, moduleName, nodeModulesFolder, nodeModulesFolderExists, state);
            if (packageResult) {
                return packageResult;
            }
            if (extensions === Extensions.TypeScript || extensions === Extensions.DtsOnly) {
                var nodeModulesAtTypes_1 = ts.combinePaths(nodeModulesFolder, "@types");
                var nodeModulesAtTypesExists = nodeModulesFolderExists;
                if (nodeModulesFolderExists && !ts.directoryProbablyExists(nodeModulesAtTypes_1, state.host)) {
                    if (state.traceEnabled) {
                        trace(state.host, ts.Diagnostics.Directory_0_does_not_exist_skipping_all_lookups_in_it, nodeModulesAtTypes_1);
                    }
                    nodeModulesAtTypesExists = false;
                }
                return loadModuleFromSpecificNodeModulesDirectory(Extensions.DtsOnly, mangleScopedPackageNameWithTrace(moduleName, state), nodeModulesAtTypes_1, nodeModulesAtTypesExists, state);
            }
        }
        function loadModuleFromSpecificNodeModulesDirectory(extensions, moduleName, nodeModulesDirectory, nodeModulesDirectoryExists, state) {
            var candidate = ts.normalizePath(ts.combinePaths(nodeModulesDirectory, moduleName));
            // First look for a nested package.json, as in `node_modules/foo/bar/package.json`.
            var packageInfo = getPackageJsonInfo(candidate, !nodeModulesDirectoryExists, state);
            if (packageInfo) {
                var fromFile = loadModuleFromFile(extensions, candidate, !nodeModulesDirectoryExists, state);
                if (fromFile) {
                    return noPackageId(fromFile);
                }
                var fromDirectory = loadNodeModuleFromDirectoryWorker(extensions, candidate, !nodeModulesDirectoryExists, state, packageInfo.packageJsonContent, packageInfo.versionPaths);
                return withPackageId(packageInfo, fromDirectory);
            }
            var loader = function (extensions, candidate, onlyRecordFailures, state) {
                var pathAndExtension = loadModuleFromFile(extensions, candidate, onlyRecordFailures, state) ||
                    loadNodeModuleFromDirectoryWorker(extensions, candidate, onlyRecordFailures, state, packageInfo && packageInfo.packageJsonContent, packageInfo && packageInfo.versionPaths);
                return withPackageId(packageInfo, pathAndExtension);
            };
            var _a = parsePackageName(moduleName), packageName = _a.packageName, rest = _a.rest;
            if (rest !== "") { // If "rest" is empty, we just did this search above.
                var packageDirectory = ts.combinePaths(nodeModulesDirectory, packageName);
                // Don't use a "types" or "main" from here because we're not loading the root, but a subdirectory -- just here for the packageId and path mappings.
                packageInfo = getPackageJsonInfo(packageDirectory, !nodeModulesDirectoryExists, state);
                if (packageInfo && packageInfo.versionPaths) {
                    if (state.traceEnabled) {
                        trace(state.host, ts.Diagnostics.package_json_has_a_typesVersions_entry_0_that_matches_compiler_version_1_looking_for_a_pattern_to_match_module_name_2, packageInfo.versionPaths.version, ts.version, rest);
                    }
                    var packageDirectoryExists = nodeModulesDirectoryExists && ts.directoryProbablyExists(packageDirectory, state.host);
                    var fromPaths = tryLoadModuleUsingPaths(extensions, rest, packageDirectory, packageInfo.versionPaths.paths, /*pathPatterns*/ undefined, loader, !packageDirectoryExists, state);
                    if (fromPaths) {
                        return fromPaths.value;
                    }
                }
            }
            return loader(extensions, candidate, !nodeModulesDirectoryExists, state);
        }
        function tryLoadModuleUsingPaths(extensions, moduleName, baseDirectory, paths, pathPatterns, loader, onlyRecordFailures, state) {
            pathPatterns || (pathPatterns = ts.tryParsePatterns(paths));
            var matchedPattern = ts.matchPatternOrExact(pathPatterns, moduleName);
            if (matchedPattern) {
                var matchedStar_1 = ts.isString(matchedPattern) ? undefined : ts.matchedText(matchedPattern, moduleName);
                var matchedPatternText = ts.isString(matchedPattern) ? matchedPattern : ts.patternText(matchedPattern);
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.Module_name_0_matched_pattern_1, moduleName, matchedPatternText);
                }
                var resolved = ts.forEach(paths[matchedPatternText], function (subst) {
                    var path = matchedStar_1 ? subst.replace("*", matchedStar_1) : subst;
                    // When baseUrl is not specified, the command line parser resolves relative paths to the config file location.
                    var candidate = ts.normalizePath(ts.combinePaths(baseDirectory, path));
                    if (state.traceEnabled) {
                        trace(state.host, ts.Diagnostics.Trying_substitution_0_candidate_module_location_Colon_1, subst, path);
                    }
                    // A path mapping may have an extension, in contrast to an import, which should omit it.
                    var extension = ts.tryGetExtensionFromPath(subst);
                    if (extension !== undefined) {
                        var path_1 = tryFile(candidate, onlyRecordFailures, state);
                        if (path_1 !== undefined) {
                            return noPackageId({ path: path_1, ext: extension });
                        }
                    }
                    return loader(extensions, candidate, onlyRecordFailures || !ts.directoryProbablyExists(ts.getDirectoryPath(candidate), state.host), state);
                });
                return { value: resolved };
            }
        }
        /** Double underscores are used in DefinitelyTyped to delimit scoped packages. */
        var mangledScopedPackageSeparator = "__";
        /** For a scoped package, we must look in `@types/foo__bar` instead of `@types/@foo/bar`. */
        function mangleScopedPackageNameWithTrace(packageName, state) {
            var mangled = mangleScopedPackageName(packageName);
            if (state.traceEnabled && mangled !== packageName) {
                trace(state.host, ts.Diagnostics.Scoped_package_detected_looking_in_0, mangled);
            }
            return mangled;
        }
        /* @internal */
        function getTypesPackageName(packageName) {
            return "@types/" + mangleScopedPackageName(packageName);
        }
        ts.getTypesPackageName = getTypesPackageName;
        /* @internal */
        function mangleScopedPackageName(packageName) {
            if (ts.startsWith(packageName, "@")) {
                var replaceSlash = packageName.replace(ts.directorySeparator, mangledScopedPackageSeparator);
                if (replaceSlash !== packageName) {
                    return replaceSlash.slice(1); // Take off the "@"
                }
            }
            return packageName;
        }
        ts.mangleScopedPackageName = mangleScopedPackageName;
        /* @internal */
        function getPackageNameFromTypesPackageName(mangledName) {
            var withoutAtTypePrefix = ts.removePrefix(mangledName, "@types/");
            if (withoutAtTypePrefix !== mangledName) {
                return unmangleScopedPackageName(withoutAtTypePrefix);
            }
            return mangledName;
        }
        ts.getPackageNameFromTypesPackageName = getPackageNameFromTypesPackageName;
        /* @internal */
        function unmangleScopedPackageName(typesPackageName) {
            return ts.stringContains(typesPackageName, mangledScopedPackageSeparator) ?
                "@" + typesPackageName.replace(mangledScopedPackageSeparator, ts.directorySeparator) :
                typesPackageName;
        }
        ts.unmangleScopedPackageName = unmangleScopedPackageName;
        function tryFindNonRelativeModuleNameInCache(cache, moduleName, containingDirectory, state) {
            var result = cache && cache.get(containingDirectory);
            if (result) {
                if (state.traceEnabled) {
                    trace(state.host, ts.Diagnostics.Resolution_for_module_0_was_found_in_cache_from_location_1, moduleName, containingDirectory);
                }
                state.resultFromCache = result;
                return { value: result.resolvedModule && { path: result.resolvedModule.resolvedFileName, originalPath: result.resolvedModule.originalPath || true, extension: result.resolvedModule.extension, packageId: result.resolvedModule.packageId } };
            }
        }
        function classicNameResolver(moduleName, containingFile, compilerOptions, host, cache, redirectedReference) {
            var traceEnabled = isTraceEnabled(compilerOptions, host);
            var failedLookupLocations = [];
            var state = { compilerOptions: compilerOptions, host: host, traceEnabled: traceEnabled, failedLookupLocations: failedLookupLocations, packageJsonInfoCache: cache };
            var containingDirectory = ts.getDirectoryPath(containingFile);
            var resolved = tryResolve(Extensions.TypeScript) || tryResolve(Extensions.JavaScript);
            // No originalPath because classic resolution doesn't resolve realPath
            return createResolvedModuleWithFailedLookupLocations(resolved && resolved.value, /*isExternalLibraryImport*/ false, failedLookupLocations, state.resultFromCache);
            function tryResolve(extensions) {
                var resolvedUsingSettings = tryLoadModuleUsingOptionalResolutionSettings(extensions, moduleName, containingDirectory, loadModuleFromFileNoPackageId, state);
                if (resolvedUsingSettings) {
                    return { value: resolvedUsingSettings };
                }
                if (!ts.isExternalModuleNameRelative(moduleName)) {
                    var perModuleNameCache_1 = cache && cache.getOrCreateCacheForModuleName(moduleName, redirectedReference);
                    // Climb up parent directories looking for a module.
                    var resolved_3 = ts.forEachAncestorDirectory(containingDirectory, function (directory) {
                        var resolutionFromCache = tryFindNonRelativeModuleNameInCache(perModuleNameCache_1, moduleName, directory, state);
                        if (resolutionFromCache) {
                            return resolutionFromCache;
                        }
                        var searchName = ts.normalizePath(ts.combinePaths(directory, moduleName));
                        return toSearchResult(loadModuleFromFileNoPackageId(extensions, searchName, /*onlyRecordFailures*/ false, state));
                    });
                    if (resolved_3) {
                        return resolved_3;
                    }
                    if (extensions === Extensions.TypeScript) {
                        // If we didn't find the file normally, look it up in @types.
                        return loadModuleFromNearestNodeModulesDirectoryTypesScope(moduleName, containingDirectory, state);
                    }
                }
                else {
                    var candidate = ts.normalizePath(ts.combinePaths(containingDirectory, moduleName));
                    return toSearchResult(loadModuleFromFileNoPackageId(extensions, candidate, /*onlyRecordFailures*/ false, state));
                }
            }
        }
        ts.classicNameResolver = classicNameResolver;
        /**
         * A host may load a module from a global cache of typings.
         * This is the minumum code needed to expose that functionality; the rest is in the host.
         */
        /* @internal */
        function loadModuleFromGlobalCache(moduleName, projectName, compilerOptions, host, globalCache, packageJsonInfoCache) {
            var traceEnabled = isTraceEnabled(compilerOptions, host);
            if (traceEnabled) {
                trace(host, ts.Diagnostics.Auto_discovery_for_typings_is_enabled_in_project_0_Running_extra_resolution_pass_for_module_1_using_cache_location_2, projectName, moduleName, globalCache);
            }
            var failedLookupLocations = [];
            var state = { compilerOptions: compilerOptions, host: host, traceEnabled: traceEnabled, failedLookupLocations: failedLookupLocations, packageJsonInfoCache: packageJsonInfoCache };
            var resolved = loadModuleFromImmediateNodeModulesDirectory(Extensions.DtsOnly, moduleName, globalCache, state, /*typesScopeOnly*/ false);
            return createResolvedModuleWithFailedLookupLocations(resolved, /*isExternalLibraryImport*/ true, failedLookupLocations, state.resultFromCache);
        }
        ts.loadModuleFromGlobalCache = loadModuleFromGlobalCache;
        /**
         * Wraps value to SearchResult.
         * @returns undefined if value is undefined or { value } otherwise
         */
        function toSearchResult(value) {
            return value !== undefined ? { value: value } : undefined;
        }
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var ModuleInstanceState;
        (function (ModuleInstanceState) {
            ModuleInstanceState[ModuleInstanceState["NonInstantiated"] = 0] = "NonInstantiated";
            ModuleInstanceState[ModuleInstanceState["Instantiated"] = 1] = "Instantiated";
            ModuleInstanceState[ModuleInstanceState["ConstEnumOnly"] = 2] = "ConstEnumOnly";
        })(ModuleInstanceState = ts.ModuleInstanceState || (ts.ModuleInstanceState = {}));
        function getModuleInstanceState(node, visited) {
            if (node.body && !node.body.parent) {
                // getModuleInstanceStateForAliasTarget needs to walk up the parent chain, so parent pointers must be set on this tree already
                ts.setParent(node.body, node);
                ts.setParentRecursive(node.body, /*incremental*/ false);
            }
            return node.body ? getModuleInstanceStateCached(node.body, visited) : 1 /* Instantiated */;
        }
        ts.getModuleInstanceState = getModuleInstanceState;
        function getModuleInstanceStateCached(node, visited) {
            if (visited === void 0) {
                visited = new ts.Map();
            }
            var nodeId = ts.getNodeId(node);
            if (visited.has(nodeId)) {
                return visited.get(nodeId) || 0 /* NonInstantiated */;
            }
            visited.set(nodeId, undefined);
            var result = getModuleInstanceStateWorker(node, visited);
            visited.set(nodeId, result);
            return result;
        }
        function getModuleInstanceStateWorker(node, visited) {
            // A module is uninstantiated if it contains only
            switch (node.kind) {
                // 1. interface declarations, type alias declarations
                case 256 /* InterfaceDeclaration */:
                case 257 /* TypeAliasDeclaration */:
                    return 0 /* NonInstantiated */;
                // 2. const enum declarations
                case 258 /* EnumDeclaration */:
                    if (ts.isEnumConst(node)) {
                        return 2 /* ConstEnumOnly */;
                    }
                    break;
                // 3. non-exported import declarations
                case 264 /* ImportDeclaration */:
                case 263 /* ImportEqualsDeclaration */:
                    if (!(ts.hasSyntacticModifier(node, 1 /* Export */))) {
                        return 0 /* NonInstantiated */;
                    }
                    break;
                // 4. Export alias declarations pointing at only uninstantiated modules or things uninstantiated modules contain
                case 270 /* ExportDeclaration */:
                    var exportDeclaration = node;
                    if (!exportDeclaration.moduleSpecifier && exportDeclaration.exportClause && exportDeclaration.exportClause.kind === 271 /* NamedExports */) {
                        var state = 0 /* NonInstantiated */;
                        for (var _i = 0, _a = exportDeclaration.exportClause.elements; _i < _a.length; _i++) {
                            var specifier = _a[_i];
                            var specifierState = getModuleInstanceStateForAliasTarget(specifier, visited);
                            if (specifierState > state) {
                                state = specifierState;
                            }
                            if (state === 1 /* Instantiated */) {
                                return state;
                            }
                        }
                        return state;
                    }
                    break;
                // 5. other uninstantiated module declarations.
                case 260 /* ModuleBlock */: {
                    var state_1 = 0 /* NonInstantiated */;
                    ts.forEachChild(node, function (n) {
                        var childState = getModuleInstanceStateCached(n, visited);
                        switch (childState) {
                            case 0 /* NonInstantiated */:
                                // child is non-instantiated - continue searching
                                return;
                            case 2 /* ConstEnumOnly */:
                                // child is const enum only - record state and continue searching
                                state_1 = 2 /* ConstEnumOnly */;
                                return;
                            case 1 /* Instantiated */:
                                // child is instantiated - record state and stop
                                state_1 = 1 /* Instantiated */;
                                return true;
                            default:
                                ts.Debug.assertNever(childState);
                        }
                    });
                    return state_1;
                }
                case 259 /* ModuleDeclaration */:
                    return getModuleInstanceState(node, visited);
                case 79 /* Identifier */:
                    // Only jsdoc typedef definition can exist in jsdoc namespace, and it should
                    // be considered the same as type alias
                    if (node.isInJSDocNamespace) {
                        return 0 /* NonInstantiated */;
                    }
            }
            return 1 /* Instantiated */;
        }
        function getModuleInstanceStateForAliasTarget(specifier, visited) {
            var name = specifier.propertyName || specifier.name;
            var p = specifier.parent;
            while (p) {
                if (ts.isBlock(p) || ts.isModuleBlock(p) || ts.isSourceFile(p)) {
                    var statements = p.statements;
                    var found = void 0;
                    for (var _i = 0, statements_2 = statements; _i < statements_2.length; _i++) {
                        var statement = statements_2[_i];
                        if (ts.nodeHasName(statement, name)) {
                            if (!statement.parent) {
                                ts.setParent(statement, p);
                                ts.setParentRecursive(statement, /*incremental*/ false);
                            }
                            var state = getModuleInstanceStateCached(statement, visited);
                            if (found === undefined || state > found) {
                                found = state;
                            }
                            if (found === 1 /* Instantiated */) {
                                return found;
                            }
                        }
                    }
                    if (found !== undefined) {
                        return found;
                    }
                }
                p = p.parent;
            }
            return 1 /* Instantiated */; // Couldn't locate, assume could refer to a value
        }
        var ContainerFlags;
        (function (ContainerFlags) {
            // The current node is not a container, and no container manipulation should happen before
            // recursing into it.
            ContainerFlags[ContainerFlags["None"] = 0] = "None";
            // The current node is a container.  It should be set as the current container (and block-
            // container) before recursing into it.  The current node does not have locals.  Examples:
            //
            //      Classes, ObjectLiterals, TypeLiterals, Interfaces...
            ContainerFlags[ContainerFlags["IsContainer"] = 1] = "IsContainer";
            // The current node is a block-scoped-container.  It should be set as the current block-
            // container before recursing into it.  Examples:
            //
            //      Blocks (when not parented by functions), Catch clauses, For/For-in/For-of statements...
            ContainerFlags[ContainerFlags["IsBlockScopedContainer"] = 2] = "IsBlockScopedContainer";
            // The current node is the container of a control flow path. The current control flow should
            // be saved and restored, and a new control flow initialized within the container.
            ContainerFlags[ContainerFlags["IsControlFlowContainer"] = 4] = "IsControlFlowContainer";
            ContainerFlags[ContainerFlags["IsFunctionLike"] = 8] = "IsFunctionLike";
            ContainerFlags[ContainerFlags["IsFunctionExpression"] = 16] = "IsFunctionExpression";
            ContainerFlags[ContainerFlags["HasLocals"] = 32] = "HasLocals";
            ContainerFlags[ContainerFlags["IsInterface"] = 64] = "IsInterface";
            ContainerFlags[ContainerFlags["IsObjectLiteralOrClassExpressionMethod"] = 128] = "IsObjectLiteralOrClassExpressionMethod";
        })(ContainerFlags || (ContainerFlags = {}));
        function initFlowNode(node) {
            ts.Debug.attachFlowNodeDebugInfo(node);
            return node;
        }
        var binder = createBinder();
        function bindSourceFile(file, options) {
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("bind" /* Bind */, "bindSourceFile", { path: file.path }, /*separateBeginAndEnd*/ true);
            ts.performance.mark("beforeBind");
            ts.perfLogger.logStartBindFile("" + file.fileName);
            binder(file, options);
            ts.perfLogger.logStopBindFile();
            ts.performance.mark("afterBind");
            ts.performance.measure("Bind", "beforeBind", "afterBind");
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
        }
        ts.bindSourceFile = bindSourceFile;
        function createBinder() {
            var file;
            var options;
            var languageVersion;
            var parent;
            var container;
            var thisParentContainer; // Container one level up
            var blockScopeContainer;
            var lastContainer;
            var delayedTypeAliases;
            var seenThisKeyword;
            // state used by control flow analysis
            var currentFlow;
            var currentBreakTarget;
            var currentContinueTarget;
            var currentReturnTarget;
            var currentTrueTarget;
            var currentFalseTarget;
            var currentExceptionTarget;
            var preSwitchCaseFlow;
            var activeLabelList;
            var hasExplicitReturn;
            // state used for emit helpers
            var emitFlags;
            // If this file is an external module, then it is automatically in strict-mode according to
            // ES6.  If it is not an external module, then we'll determine if it is in strict mode or
            // not depending on if we see "use strict" in certain places or if we hit a class/namespace
            // or if compiler options contain alwaysStrict.
            var inStrictMode;
            // If we are binding an assignment pattern, we will bind certain expressions differently.
            var inAssignmentPattern = false;
            var symbolCount = 0;
            var Symbol;
            var classifiableNames;
            var unreachableFlow = { flags: 1 /* Unreachable */ };
            var reportedUnreachableFlow = { flags: 1 /* Unreachable */ };
            var bindBinaryExpressionFlow = createBindBinaryExpressionFlow();
            /**
             * Inside the binder, we may create a diagnostic for an as-yet unbound node (with potentially no parent pointers, implying no accessible source file)
             * If so, the node _must_ be in the current file (as that's the only way anything could have traversed to it to yield it as the error node)
             * This version of `createDiagnosticForNode` uses the binder's context to account for this, and always yields correct diagnostics even in these situations.
             */
            function createDiagnosticForNode(node, message, arg0, arg1, arg2) {
                return ts.createDiagnosticForNodeInSourceFile(ts.getSourceFileOfNode(node) || file, node, message, arg0, arg1, arg2);
            }
            function bindSourceFile(f, opts) {
                file = f;
                options = opts;
                languageVersion = ts.getEmitScriptTarget(options);
                inStrictMode = bindInStrictMode(file, opts);
                classifiableNames = new ts.Set();
                symbolCount = 0;
                Symbol = ts.objectAllocator.getSymbolConstructor();
                // Attach debugging information if necessary
                ts.Debug.attachFlowNodeDebugInfo(unreachableFlow);
                ts.Debug.attachFlowNodeDebugInfo(reportedUnreachableFlow);
                if (!file.locals) {
                    bind(file);
                    file.symbolCount = symbolCount;
                    file.classifiableNames = classifiableNames;
                    delayedBindJSDocTypedefTag();
                }
                file = undefined;
                options = undefined;
                languageVersion = undefined;
                parent = undefined;
                container = undefined;
                thisParentContainer = undefined;
                blockScopeContainer = undefined;
                lastContainer = undefined;
                delayedTypeAliases = undefined;
                seenThisKeyword = false;
                currentFlow = undefined;
                currentBreakTarget = undefined;
                currentContinueTarget = undefined;
                currentReturnTarget = undefined;
                currentTrueTarget = undefined;
                currentFalseTarget = undefined;
                currentExceptionTarget = undefined;
                activeLabelList = undefined;
                hasExplicitReturn = false;
                inAssignmentPattern = false;
                emitFlags = 0 /* None */;
            }
            return bindSourceFile;
            function bindInStrictMode(file, opts) {
                if (ts.getStrictOptionValue(opts, "alwaysStrict") && !file.isDeclarationFile) {
                    // bind in strict mode source files with alwaysStrict option
                    return true;
                }
                else {
                    return !!file.externalModuleIndicator;
                }
            }
            function createSymbol(flags, name) {
                symbolCount++;
                return new Symbol(flags, name);
            }
            function addDeclarationToSymbol(symbol, node, symbolFlags) {
                symbol.flags |= symbolFlags;
                node.symbol = symbol;
                symbol.declarations = ts.appendIfUnique(symbol.declarations, node);
                if (symbolFlags & (32 /* Class */ | 384 /* Enum */ | 1536 /* Module */ | 3 /* Variable */) && !symbol.exports) {
                    symbol.exports = ts.createSymbolTable();
                }
                if (symbolFlags & (32 /* Class */ | 64 /* Interface */ | 2048 /* TypeLiteral */ | 4096 /* ObjectLiteral */) && !symbol.members) {
                    symbol.members = ts.createSymbolTable();
                }
                // On merge of const enum module with class or function, reset const enum only flag (namespaces will already recalculate)
                if (symbol.constEnumOnlyModule && (symbol.flags & (16 /* Function */ | 32 /* Class */ | 256 /* RegularEnum */))) {
                    symbol.constEnumOnlyModule = false;
                }
                if (symbolFlags & 111551 /* Value */) {
                    ts.setValueDeclaration(symbol, node);
                }
            }
            // Should not be called on a declaration with a computed property name,
            // unless it is a well known Symbol.
            function getDeclarationName(node) {
                if (node.kind === 269 /* ExportAssignment */) {
                    return node.isExportEquals ? "export=" /* ExportEquals */ : "default" /* Default */;
                }
                var name = ts.getNameOfDeclaration(node);
                if (name) {
                    if (ts.isAmbientModule(node)) {
                        var moduleName = ts.getTextOfIdentifierOrLiteral(name);
                        return (ts.isGlobalScopeAugmentation(node) ? "__global" : "\"" + moduleName + "\"");
                    }
                    if (name.kind === 160 /* ComputedPropertyName */) {
                        var nameExpression = name.expression;
                        // treat computed property names where expression is string/numeric literal as just string/numeric literal
                        if (ts.isStringOrNumericLiteralLike(nameExpression)) {
                            return ts.escapeLeadingUnderscores(nameExpression.text);
                        }
                        if (ts.isSignedNumericLiteral(nameExpression)) {
                            return ts.tokenToString(nameExpression.operator) + nameExpression.operand.text;
                        }
                        else {
                            ts.Debug.fail("Only computed properties with literal names have declaration names");
                        }
                    }
                    if (ts.isPrivateIdentifier(name)) {
                        // containingClass exists because private names only allowed inside classes
                        var containingClass = ts.getContainingClass(node);
                        if (!containingClass) {
                            // we can get here in cases where there is already a parse error.
                            return undefined;
                        }
                        var containingClassSymbol = containingClass.symbol;
                        return ts.getSymbolNameForPrivateIdentifier(containingClassSymbol, name.escapedText);
                    }
                    return ts.isPropertyNameLiteral(name) ? ts.getEscapedTextOfIdentifierOrLiteral(name) : undefined;
                }
                switch (node.kind) {
                    case 169 /* Constructor */:
                        return "__constructor" /* Constructor */;
                    case 177 /* FunctionType */:
                    case 172 /* CallSignature */:
                    case 318 /* JSDocSignature */:
                        return "__call" /* Call */;
                    case 178 /* ConstructorType */:
                    case 173 /* ConstructSignature */:
                        return "__new" /* New */;
                    case 174 /* IndexSignature */:
                        return "__index" /* Index */;
                    case 270 /* ExportDeclaration */:
                        return "__export" /* ExportStar */;
                    case 300 /* SourceFile */:
                        // json file should behave as
                        // module.exports = ...
                        return "export=" /* ExportEquals */;
                    case 219 /* BinaryExpression */:
                        if (ts.getAssignmentDeclarationKind(node) === 2 /* ModuleExports */) {
                            // module.exports = ...
                            return "export=" /* ExportEquals */;
                        }
                        ts.Debug.fail("Unknown binary declaration kind");
                        break;
                    case 312 /* JSDocFunctionType */:
                        return (ts.isJSDocConstructSignature(node) ? "__new" /* New */ : "__call" /* Call */);
                    case 162 /* Parameter */:
                        // Parameters with names are handled at the top of this function.  Parameters
                        // without names can only come from JSDocFunctionTypes.
                        ts.Debug.assert(node.parent.kind === 312 /* JSDocFunctionType */, "Impossible parameter parent kind", function () { return "parent is: " + (ts.SyntaxKind ? ts.SyntaxKind[node.parent.kind] : node.parent.kind) + ", expected JSDocFunctionType"; });
                        var functionType = node.parent;
                        var index = functionType.parameters.indexOf(node);
                        return "arg" + index;
                }
            }
            function getDisplayName(node) {
                return ts.isNamedDeclaration(node) ? ts.declarationNameToString(node.name) : ts.unescapeLeadingUnderscores(ts.Debug.checkDefined(getDeclarationName(node)));
            }
            /**
             * Declares a Symbol for the node and adds it to symbols. Reports errors for conflicting identifier names.
             * @param symbolTable - The symbol table which node will be added to.
             * @param parent - node's parent declaration.
             * @param node - The declaration to be added to the symbol table
             * @param includes - The SymbolFlags that node has in addition to its declaration type (eg: export, ambient, etc.)
             * @param excludes - The flags which node cannot be declared alongside in a symbol table. Used to report forbidden declarations.
             */
            function declareSymbol(symbolTable, parent, node, includes, excludes, isReplaceableByMethod, isComputedName) {
                ts.Debug.assert(isComputedName || !ts.hasDynamicName(node));
                var isDefaultExport = ts.hasSyntacticModifier(node, 512 /* Default */) || ts.isExportSpecifier(node) && node.name.escapedText === "default";
                // The exported symbol for an export default function/class node is always named "default"
                var name = isComputedName ? "__computed" /* Computed */
                    : isDefaultExport && parent ? "default" /* Default */
                        : getDeclarationName(node);
                var symbol;
                if (name === undefined) {
                    symbol = createSymbol(0 /* None */, "__missing" /* Missing */);
                }
                else {
                    // Check and see if the symbol table already has a symbol with this name.  If not,
                    // create a new symbol with this name and add it to the table.  Note that we don't
                    // give the new symbol any flags *yet*.  This ensures that it will not conflict
                    // with the 'excludes' flags we pass in.
                    //
                    // If we do get an existing symbol, see if it conflicts with the new symbol we're
                    // creating.  For example, a 'var' symbol and a 'class' symbol will conflict within
                    // the same symbol table.  If we have a conflict, report the issue on each
                    // declaration we have for this symbol, and then create a new symbol for this
                    // declaration.
                    //
                    // Note that when properties declared in Javascript constructors
                    // (marked by isReplaceableByMethod) conflict with another symbol, the property loses.
                    // Always. This allows the common Javascript pattern of overwriting a prototype method
                    // with an bound instance method of the same type: `this.method = this.method.bind(this)`
                    //
                    // If we created a new symbol, either because we didn't have a symbol with this name
                    // in the symbol table, or we conflicted with an existing symbol, then just add this
                    // node as the sole declaration of the new symbol.
                    //
                    // Otherwise, we'll be merging into a compatible existing symbol (for example when
                    // you have multiple 'vars' with the same name in the same container).  In this case
                    // just add this node into the declarations list of the symbol.
                    symbol = symbolTable.get(name);
                    if (includes & 2885600 /* Classifiable */) {
                        classifiableNames.add(name);
                    }
                    if (!symbol) {
                        symbolTable.set(name, symbol = createSymbol(0 /* None */, name));
                        if (isReplaceableByMethod)
                            symbol.isReplaceableByMethod = true;
                    }
                    else if (isReplaceableByMethod && !symbol.isReplaceableByMethod) {
                        // A symbol already exists, so don't add this as a declaration.
                        return symbol;
                    }
                    else if (symbol.flags & excludes) {
                        if (symbol.isReplaceableByMethod) {
                            // Javascript constructor-declared symbols can be discarded in favor of
                            // prototype symbols like methods.
                            symbolTable.set(name, symbol = createSymbol(0 /* None */, name));
                        }
                        else if (!(includes & 3 /* Variable */ && symbol.flags & 67108864 /* Assignment */)) {
                            // Assignment declarations are allowed to merge with variables, no matter what other flags they have.
                            if (ts.isNamedDeclaration(node)) {
                                ts.setParent(node.name, node);
                            }
                            // Report errors every position with duplicate declaration
                            // Report errors on previous encountered declarations
                            var message_1 = symbol.flags & 2 /* BlockScopedVariable */
                                ? ts.Diagnostics.Cannot_redeclare_block_scoped_variable_0
                                : ts.Diagnostics.Duplicate_identifier_0;
                            var messageNeedsName_1 = true;
                            if (symbol.flags & 384 /* Enum */ || includes & 384 /* Enum */) {
                                message_1 = ts.Diagnostics.Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations;
                                messageNeedsName_1 = false;
                            }
                            var multipleDefaultExports_1 = false;
                            if (ts.length(symbol.declarations)) {
                                // If the current node is a default export of some sort, then check if
                                // there are any other default exports that we need to error on.
                                // We'll know whether we have other default exports depending on if `symbol` already has a declaration list set.
                                if (isDefaultExport) {
                                    message_1 = ts.Diagnostics.A_module_cannot_have_multiple_default_exports;
                                    messageNeedsName_1 = false;
                                    multipleDefaultExports_1 = true;
                                }
                                else {
                                    // This is to properly report an error in the case "export default { }" is after export default of class declaration or function declaration.
                                    // Error on multiple export default in the following case:
                                    // 1. multiple export default of class declaration or function declaration by checking NodeFlags.Default
                                    // 2. multiple export default of export assignment. This one doesn't have NodeFlags.Default on (as export default doesn't considered as modifiers)
                                    if (symbol.declarations && symbol.declarations.length &&
                                        (node.kind === 269 /* ExportAssignment */ && !node.isExportEquals)) {
                                        message_1 = ts.Diagnostics.A_module_cannot_have_multiple_default_exports;
                                        messageNeedsName_1 = false;
                                        multipleDefaultExports_1 = true;
                                    }
                                }
                            }
                            var relatedInformation_1 = [];
                            if (ts.isTypeAliasDeclaration(node) && ts.nodeIsMissing(node.type) && ts.hasSyntacticModifier(node, 1 /* Export */) && symbol.flags & (2097152 /* Alias */ | 788968 /* Type */ | 1920 /* Namespace */)) {
                                // export type T; - may have meant export type { T }?
                                relatedInformation_1.push(createDiagnosticForNode(node, ts.Diagnostics.Did_you_mean_0, "export type { " + ts.unescapeLeadingUnderscores(node.name.escapedText) + " }"));
                            }
                            var declarationName_1 = ts.getNameOfDeclaration(node) || node;
                            ts.forEach(symbol.declarations, function (declaration, index) {
                                var decl = ts.getNameOfDeclaration(declaration) || declaration;
                                var diag = createDiagnosticForNode(decl, message_1, messageNeedsName_1 ? getDisplayName(declaration) : undefined);
                                file.bindDiagnostics.push(multipleDefaultExports_1 ? ts.addRelatedInfo(diag, createDiagnosticForNode(declarationName_1, index === 0 ? ts.Diagnostics.Another_export_default_is_here : ts.Diagnostics.and_here)) : diag);
                                if (multipleDefaultExports_1) {
                                    relatedInformation_1.push(createDiagnosticForNode(decl, ts.Diagnostics.The_first_export_default_is_here));
                                }
                            });
                            var diag = createDiagnosticForNode(declarationName_1, message_1, messageNeedsName_1 ? getDisplayName(node) : undefined);
                            file.bindDiagnostics.push(ts.addRelatedInfo.apply(void 0, __spreadArray([diag], relatedInformation_1, false)));
                            symbol = createSymbol(0 /* None */, name);
                        }
                    }
                }
                addDeclarationToSymbol(symbol, node, includes);
                if (symbol.parent) {
                    ts.Debug.assert(symbol.parent === parent, "Existing symbol parent should match new one");
                }
                else {
                    symbol.parent = parent;
                }
                return symbol;
            }
            function declareModuleMember(node, symbolFlags, symbolExcludes) {
                var hasExportModifier = !!(ts.getCombinedModifierFlags(node) & 1 /* Export */) || jsdocTreatAsExported(node);
                if (symbolFlags & 2097152 /* Alias */) {
                    if (node.kind === 273 /* ExportSpecifier */ || (node.kind === 263 /* ImportEqualsDeclaration */ && hasExportModifier)) {
                        return declareSymbol(container.symbol.exports, container.symbol, node, symbolFlags, symbolExcludes);
                    }
                    else {
                        return declareSymbol(container.locals, /*parent*/ undefined, node, symbolFlags, symbolExcludes);
                    }
                }
                else {
                    // Exported module members are given 2 symbols: A local symbol that is classified with an ExportValue flag,
                    // and an associated export symbol with all the correct flags set on it. There are 2 main reasons:
                    //
                    //   1. We treat locals and exports of the same name as mutually exclusive within a container.
                    //      That means the binder will issue a Duplicate Identifier error if you mix locals and exports
                    //      with the same name in the same container.
                    //      TODO: Make this a more specific error and decouple it from the exclusion logic.
                    //   2. When we checkIdentifier in the checker, we set its resolved symbol to the local symbol,
                    //      but return the export symbol (by calling getExportSymbolOfValueSymbolIfExported). That way
                    //      when the emitter comes back to it, it knows not to qualify the name if it was found in a containing scope.
                    // NOTE: Nested ambient modules always should go to to 'locals' table to prevent their automatic merge
                    //       during global merging in the checker. Why? The only case when ambient module is permitted inside another module is module augmentation
                    //       and this case is specially handled. Module augmentations should only be merged with original module definition
                    //       and should never be merged directly with other augmentation, and the latter case would be possible if automatic merge is allowed.
                    if (ts.isJSDocTypeAlias(node))
                        ts.Debug.assert(ts.isInJSFile(node)); // We shouldn't add symbols for JSDoc nodes if not in a JS file.
                    if (!ts.isAmbientModule(node) && (hasExportModifier || container.flags & 64 /* ExportContext */)) {
                        if (!container.locals || (ts.hasSyntacticModifier(node, 512 /* Default */) && !getDeclarationName(node))) {
                            return declareSymbol(container.symbol.exports, container.symbol, node, symbolFlags, symbolExcludes); // No local symbol for an unnamed default!
                        }
                        var exportKind = symbolFlags & 111551 /* Value */ ? 1048576 /* ExportValue */ : 0;
                        var local = declareSymbol(container.locals, /*parent*/ undefined, node, exportKind, symbolExcludes);
                        local.exportSymbol = declareSymbol(container.symbol.exports, container.symbol, node, symbolFlags, symbolExcludes);
                        node.localSymbol = local;
                        return local;
                    }
                    else {
                        return declareSymbol(container.locals, /*parent*/ undefined, node, symbolFlags, symbolExcludes);
                    }
                }
            }
            function jsdocTreatAsExported(node) {
                if (node.parent && ts.isModuleDeclaration(node)) {
                    node = node.parent;
                }
                if (!ts.isJSDocTypeAlias(node))
                    return false;
                // jsdoc typedef handling is a bit of a doozy, but to summarize, treat the typedef as exported if:
                // 1. It has an explicit name (since by default typedefs are always directly exported, either at the top level or in a container), or
                if (!ts.isJSDocEnumTag(node) && !!node.fullName)
                    return true;
                // 2. The thing a nameless typedef pulls its name from is implicitly a direct export (either by assignment or actual export flag).
                var declName = ts.getNameOfDeclaration(node);
                if (!declName)
                    return false;
                if (ts.isPropertyAccessEntityNameExpression(declName.parent) && isTopLevelNamespaceAssignment(declName.parent))
                    return true;
                if (ts.isDeclaration(declName.parent) && ts.getCombinedModifierFlags(declName.parent) & 1 /* Export */)
                    return true;
                // This could potentially be simplified by having `delayedBindJSDocTypedefTag` pass in an override for `hasExportModifier`, since it should
                // already have calculated and branched on most of this.
                return false;
            }
            // All container nodes are kept on a linked list in declaration order. This list is used by
            // the getLocalNameOfContainer function in the type checker to validate that the local name
            // used for a container is unique.
            function bindContainer(node, containerFlags) {
                // Before we recurse into a node's children, we first save the existing parent, container
                // and block-container.  Then after we pop out of processing the children, we restore
                // these saved values.
                var saveContainer = container;
                var saveThisParentContainer = thisParentContainer;
                var savedBlockScopeContainer = blockScopeContainer;
                // Depending on what kind of node this is, we may have to adjust the current container
                // and block-container.   If the current node is a container, then it is automatically
                // considered the current block-container as well.  Also, for containers that we know
                // may contain locals, we eagerly initialize the .locals field. We do this because
                // it's highly likely that the .locals will be needed to place some child in (for example,
                // a parameter, or variable declaration).
                //
                // However, we do not proactively create the .locals for block-containers because it's
                // totally normal and common for block-containers to never actually have a block-scoped
                // variable in them.  We don't want to end up allocating an object for every 'block' we
                // run into when most of them won't be necessary.
                //
                // Finally, if this is a block-container, then we clear out any existing .locals object
                // it may contain within it.  This happens in incremental scenarios.  Because we can be
                // reusing a node from a previous compilation, that node may have had 'locals' created
                // for it.  We must clear this so we don't accidentally move any stale data forward from
                // a previous compilation.
                if (containerFlags & 1 /* IsContainer */) {
                    if (node.kind !== 212 /* ArrowFunction */) {
                        thisParentContainer = container;
                    }
                    container = blockScopeContainer = node;
                    if (containerFlags & 32 /* HasLocals */) {
                        container.locals = ts.createSymbolTable();
                    }
                    addToContainerChain(container);
                }
                else if (containerFlags & 2 /* IsBlockScopedContainer */) {
                    blockScopeContainer = node;
                    blockScopeContainer.locals = undefined;
                }
                if (containerFlags & 4 /* IsControlFlowContainer */) {
                    var saveCurrentFlow = currentFlow;
                    var saveBreakTarget = currentBreakTarget;
                    var saveContinueTarget = currentContinueTarget;
                    var saveReturnTarget = currentReturnTarget;
                    var saveExceptionTarget = currentExceptionTarget;
                    var saveActiveLabelList = activeLabelList;
                    var saveHasExplicitReturn = hasExplicitReturn;
                    var isIIFE = containerFlags & 16 /* IsFunctionExpression */ && !ts.hasSyntacticModifier(node, 256 /* Async */) &&
                        !node.asteriskToken && !!ts.getImmediatelyInvokedFunctionExpression(node);
                    // A non-async, non-generator IIFE is considered part of the containing control flow. Return statements behave
                    // similarly to break statements that exit to a label just past the statement body.
                    if (!isIIFE) {
                        currentFlow = initFlowNode({ flags: 2 /* Start */ });
                        if (containerFlags & (16 /* IsFunctionExpression */ | 128 /* IsObjectLiteralOrClassExpressionMethod */)) {
                            currentFlow.node = node;
                        }
                    }
                    // We create a return control flow graph for IIFEs and constructors. For constructors
                    // we use the return control flow graph in strict property initialization checks.
                    currentReturnTarget = isIIFE || node.kind === 169 /* Constructor */ || node.kind === 168 /* ClassStaticBlockDeclaration */ || (ts.isInJSFile(node) && (node.kind === 254 /* FunctionDeclaration */ || node.kind === 211 /* FunctionExpression */)) ? createBranchLabel() : undefined;
                    currentExceptionTarget = undefined;
                    currentBreakTarget = undefined;
                    currentContinueTarget = undefined;
                    activeLabelList = undefined;
                    hasExplicitReturn = false;
                    bindChildren(node);
                    // Reset all reachability check related flags on node (for incremental scenarios)
                    node.flags &= ~2816 /* ReachabilityAndEmitFlags */;
                    if (!(currentFlow.flags & 1 /* Unreachable */) && containerFlags & 8 /* IsFunctionLike */ && ts.nodeIsPresent(node.body)) {
                        node.flags |= 256 /* HasImplicitReturn */;
                        if (hasExplicitReturn)
                            node.flags |= 512 /* HasExplicitReturn */;
                        node.endFlowNode = currentFlow;
                    }
                    if (node.kind === 300 /* SourceFile */) {
                        node.flags |= emitFlags;
                        node.endFlowNode = currentFlow;
                    }
                    if (currentReturnTarget) {
                        addAntecedent(currentReturnTarget, currentFlow);
                        currentFlow = finishFlowLabel(currentReturnTarget);
                        if (node.kind === 169 /* Constructor */ || node.kind === 168 /* ClassStaticBlockDeclaration */ || (ts.isInJSFile(node) && (node.kind === 254 /* FunctionDeclaration */ || node.kind === 211 /* FunctionExpression */))) {
                            node.returnFlowNode = currentFlow;
                        }
                    }
                    if (!isIIFE) {
                        currentFlow = saveCurrentFlow;
                    }
                    currentBreakTarget = saveBreakTarget;
                    currentContinueTarget = saveContinueTarget;
                    currentReturnTarget = saveReturnTarget;
                    currentExceptionTarget = saveExceptionTarget;
                    activeLabelList = saveActiveLabelList;
                    hasExplicitReturn = saveHasExplicitReturn;
                }
                else if (containerFlags & 64 /* IsInterface */) {
                    seenThisKeyword = false;
                    bindChildren(node);
                    node.flags = seenThisKeyword ? node.flags | 128 /* ContainsThis */ : node.flags & ~128 /* ContainsThis */;
                }
                else {
                    bindChildren(node);
                }
                container = saveContainer;
                thisParentContainer = saveThisParentContainer;
                blockScopeContainer = savedBlockScopeContainer;
            }
            function bindEachFunctionsFirst(nodes) {
                bindEach(nodes, function (n) { return n.kind === 254 /* FunctionDeclaration */ ? bind(n) : undefined; });
                bindEach(nodes, function (n) { return n.kind !== 254 /* FunctionDeclaration */ ? bind(n) : undefined; });
            }
            function bindEach(nodes, bindFunction) {
                if (bindFunction === void 0) {
                    bindFunction = bind;
                }
                if (nodes === undefined) {
                    return;
                }
                ts.forEach(nodes, bindFunction);
            }
            function bindEachChild(node) {
                ts.forEachChild(node, bind, bindEach);
            }
            function bindChildren(node) {
                var saveInAssignmentPattern = inAssignmentPattern;
                // Most nodes aren't valid in an assignment pattern, so we clear the value here
                // and set it before we descend into nodes that could actually be part of an assignment pattern.
                inAssignmentPattern = false;
                if (checkUnreachable(node)) {
                    bindEachChild(node);
                    bindJSDoc(node);
                    inAssignmentPattern = saveInAssignmentPattern;
                    return;
                }
                if (node.kind >= 235 /* FirstStatement */ && node.kind <= 251 /* LastStatement */ && !options.allowUnreachableCode) {
                    node.flowNode = currentFlow;
                }
                switch (node.kind) {
                    case 239 /* WhileStatement */:
                        bindWhileStatement(node);
                        break;
                    case 238 /* DoStatement */:
                        bindDoStatement(node);
                        break;
                    case 240 /* ForStatement */:
                        bindForStatement(node);
                        break;
                    case 241 /* ForInStatement */:
                    case 242 /* ForOfStatement */:
                        bindForInOrForOfStatement(node);
                        break;
                    case 237 /* IfStatement */:
                        bindIfStatement(node);
                        break;
                    case 245 /* ReturnStatement */:
                    case 249 /* ThrowStatement */:
                        bindReturnOrThrow(node);
                        break;
                    case 244 /* BreakStatement */:
                    case 243 /* ContinueStatement */:
                        bindBreakOrContinueStatement(node);
                        break;
                    case 250 /* TryStatement */:
                        bindTryStatement(node);
                        break;
                    case 247 /* SwitchStatement */:
                        bindSwitchStatement(node);
                        break;
                    case 261 /* CaseBlock */:
                        bindCaseBlock(node);
                        break;
                    case 287 /* CaseClause */:
                        bindCaseClause(node);
                        break;
                    case 236 /* ExpressionStatement */:
                        bindExpressionStatement(node);
                        break;
                    case 248 /* LabeledStatement */:
                        bindLabeledStatement(node);
                        break;
                    case 217 /* PrefixUnaryExpression */:
                        bindPrefixUnaryExpressionFlow(node);
                        break;
                    case 218 /* PostfixUnaryExpression */:
                        bindPostfixUnaryExpressionFlow(node);
                        break;
                    case 219 /* BinaryExpression */:
                        if (ts.isDestructuringAssignment(node)) {
                            // Carry over whether we are in an assignment pattern to
                            // binary expressions that could actually be an initializer
                            inAssignmentPattern = saveInAssignmentPattern;
                            bindDestructuringAssignmentFlow(node);
                            return;
                        }
                        bindBinaryExpressionFlow(node);
                        break;
                    case 213 /* DeleteExpression */:
                        bindDeleteExpressionFlow(node);
                        break;
                    case 220 /* ConditionalExpression */:
                        bindConditionalExpressionFlow(node);
                        break;
                    case 252 /* VariableDeclaration */:
                        bindVariableDeclarationFlow(node);
                        break;
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */:
                        bindAccessExpressionFlow(node);
                        break;
                    case 206 /* CallExpression */:
                        bindCallExpressionFlow(node);
                        break;
                    case 228 /* NonNullExpression */:
                        bindNonNullExpressionFlow(node);
                        break;
                    case 340 /* JSDocTypedefTag */:
                    case 333 /* JSDocCallbackTag */:
                    case 334 /* JSDocEnumTag */:
                        bindJSDocTypeAlias(node);
                        break;
                    // In source files and blocks, bind functions first to match hoisting that occurs at runtime
                    case 300 /* SourceFile */: {
                        bindEachFunctionsFirst(node.statements);
                        bind(node.endOfFileToken);
                        break;
                    }
                    case 233 /* Block */:
                    case 260 /* ModuleBlock */:
                        bindEachFunctionsFirst(node.statements);
                        break;
                    case 201 /* BindingElement */:
                        bindBindingElementFlow(node);
                        break;
                    case 203 /* ObjectLiteralExpression */:
                    case 202 /* ArrayLiteralExpression */:
                    case 291 /* PropertyAssignment */:
                    case 223 /* SpreadElement */:
                        // Carry over whether we are in an assignment pattern of Object and Array literals
                        // as well as their children that are valid assignment targets.
                        inAssignmentPattern = saveInAssignmentPattern;
                    // falls through
                    default:
                        bindEachChild(node);
                        break;
                }
                bindJSDoc(node);
                inAssignmentPattern = saveInAssignmentPattern;
            }
            function isNarrowingExpression(expr) {
                switch (expr.kind) {
                    case 79 /* Identifier */:
                    case 80 /* PrivateIdentifier */:
                    case 108 /* ThisKeyword */:
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */:
                        return containsNarrowableReference(expr);
                    case 206 /* CallExpression */:
                        return hasNarrowableArgument(expr);
                    case 210 /* ParenthesizedExpression */:
                    case 228 /* NonNullExpression */:
                        return isNarrowingExpression(expr.expression);
                    case 219 /* BinaryExpression */:
                        return isNarrowingBinaryExpression(expr);
                    case 217 /* PrefixUnaryExpression */:
                        return expr.operator === 53 /* ExclamationToken */ && isNarrowingExpression(expr.operand);
                    case 214 /* TypeOfExpression */:
                        return isNarrowingExpression(expr.expression);
                }
                return false;
            }
            function isNarrowableReference(expr) {
                return ts.isDottedName(expr)
                    || (ts.isPropertyAccessExpression(expr) || ts.isNonNullExpression(expr) || ts.isParenthesizedExpression(expr)) && isNarrowableReference(expr.expression)
                    || ts.isBinaryExpression(expr) && expr.operatorToken.kind === 27 /* CommaToken */ && isNarrowableReference(expr.right)
                    || ts.isElementAccessExpression(expr) && ts.isStringOrNumericLiteralLike(expr.argumentExpression) && isNarrowableReference(expr.expression)
                    || ts.isAssignmentExpression(expr) && isNarrowableReference(expr.left);
            }
            function containsNarrowableReference(expr) {
                return isNarrowableReference(expr) || ts.isOptionalChain(expr) && containsNarrowableReference(expr.expression);
            }
            function hasNarrowableArgument(expr) {
                if (expr.arguments) {
                    for (var _i = 0, _a = expr.arguments; _i < _a.length; _i++) {
                        var argument = _a[_i];
                        if (containsNarrowableReference(argument)) {
                            return true;
                        }
                    }
                }
                if (expr.expression.kind === 204 /* PropertyAccessExpression */ &&
                    containsNarrowableReference(expr.expression.expression)) {
                    return true;
                }
                return false;
            }
            function isNarrowingTypeofOperands(expr1, expr2) {
                return ts.isTypeOfExpression(expr1) && isNarrowableOperand(expr1.expression) && ts.isStringLiteralLike(expr2);
            }
            function isNarrowingBinaryExpression(expr) {
                switch (expr.operatorToken.kind) {
                    case 63 /* EqualsToken */:
                    case 75 /* BarBarEqualsToken */:
                    case 76 /* AmpersandAmpersandEqualsToken */:
                    case 77 /* QuestionQuestionEqualsToken */:
                        return containsNarrowableReference(expr.left);
                    case 34 /* EqualsEqualsToken */:
                    case 35 /* ExclamationEqualsToken */:
                    case 36 /* EqualsEqualsEqualsToken */:
                    case 37 /* ExclamationEqualsEqualsToken */:
                        return isNarrowableOperand(expr.left) || isNarrowableOperand(expr.right) ||
                            isNarrowingTypeofOperands(expr.right, expr.left) || isNarrowingTypeofOperands(expr.left, expr.right);
                    case 102 /* InstanceOfKeyword */:
                        return isNarrowableOperand(expr.left);
                    case 101 /* InKeyword */:
                        return isNarrowingExpression(expr.right);
                    case 27 /* CommaToken */:
                        return isNarrowingExpression(expr.right);
                }
                return false;
            }
            function isNarrowableOperand(expr) {
                switch (expr.kind) {
                    case 210 /* ParenthesizedExpression */:
                        return isNarrowableOperand(expr.expression);
                    case 219 /* BinaryExpression */:
                        switch (expr.operatorToken.kind) {
                            case 63 /* EqualsToken */:
                                return isNarrowableOperand(expr.left);
                            case 27 /* CommaToken */:
                                return isNarrowableOperand(expr.right);
                        }
                }
                return containsNarrowableReference(expr);
            }
            function createBranchLabel() {
                return initFlowNode({ flags: 4 /* BranchLabel */, antecedents: undefined });
            }
            function createLoopLabel() {
                return initFlowNode({ flags: 8 /* LoopLabel */, antecedents: undefined });
            }
            function createReduceLabel(target, antecedents, antecedent) {
                return initFlowNode({ flags: 1024 /* ReduceLabel */, target: target, antecedents: antecedents, antecedent: antecedent });
            }
            function setFlowNodeReferenced(flow) {
                // On first reference we set the Referenced flag, thereafter we set the Shared flag
                flow.flags |= flow.flags & 2048 /* Referenced */ ? 4096 /* Shared */ : 2048 /* Referenced */;
            }
            function addAntecedent(label, antecedent) {
                if (!(antecedent.flags & 1 /* Unreachable */) && !ts.contains(label.antecedents, antecedent)) {
                    (label.antecedents || (label.antecedents = [])).push(antecedent);
                    setFlowNodeReferenced(antecedent);
                }
            }
            function createFlowCondition(flags, antecedent, expression) {
                if (antecedent.flags & 1 /* Unreachable */) {
                    return antecedent;
                }
                if (!expression) {
                    return flags & 32 /* TrueCondition */ ? antecedent : unreachableFlow;
                }
                if ((expression.kind === 110 /* TrueKeyword */ && flags & 64 /* FalseCondition */ ||
                    expression.kind === 95 /* FalseKeyword */ && flags & 32 /* TrueCondition */) &&
                    !ts.isExpressionOfOptionalChainRoot(expression) && !ts.isNullishCoalesce(expression.parent)) {
                    return unreachableFlow;
                }
                if (!isNarrowingExpression(expression)) {
                    return antecedent;
                }
                setFlowNodeReferenced(antecedent);
                return initFlowNode({ flags: flags, antecedent: antecedent, node: expression });
            }
            function createFlowSwitchClause(antecedent, switchStatement, clauseStart, clauseEnd) {
                setFlowNodeReferenced(antecedent);
                return initFlowNode({ flags: 128 /* SwitchClause */, antecedent: antecedent, switchStatement: switchStatement, clauseStart: clauseStart, clauseEnd: clauseEnd });
            }
            function createFlowMutation(flags, antecedent, node) {
                setFlowNodeReferenced(antecedent);
                var result = initFlowNode({ flags: flags, antecedent: antecedent, node: node });
                if (currentExceptionTarget) {
                    addAntecedent(currentExceptionTarget, result);
                }
                return result;
            }
            function createFlowCall(antecedent, node) {
                setFlowNodeReferenced(antecedent);
                return initFlowNode({ flags: 512 /* Call */, antecedent: antecedent, node: node });
            }
            function finishFlowLabel(flow) {
                var antecedents = flow.antecedents;
                if (!antecedents) {
                    return unreachableFlow;
                }
                if (antecedents.length === 1) {
                    return antecedents[0];
                }
                return flow;
            }
            function isStatementCondition(node) {
                var parent = node.parent;
                switch (parent.kind) {
                    case 237 /* IfStatement */:
                    case 239 /* WhileStatement */:
                    case 238 /* DoStatement */:
                        return parent.expression === node;
                    case 240 /* ForStatement */:
                    case 220 /* ConditionalExpression */:
                        return parent.condition === node;
                }
                return false;
            }
            function isLogicalExpression(node) {
                while (true) {
                    if (node.kind === 210 /* ParenthesizedExpression */) {
                        node = node.expression;
                    }
                    else if (node.kind === 217 /* PrefixUnaryExpression */ && node.operator === 53 /* ExclamationToken */) {
                        node = node.operand;
                    }
                    else {
                        return node.kind === 219 /* BinaryExpression */ && (node.operatorToken.kind === 55 /* AmpersandAmpersandToken */ ||
                            node.operatorToken.kind === 56 /* BarBarToken */ ||
                            node.operatorToken.kind === 60 /* QuestionQuestionToken */);
                    }
                }
            }
            function isLogicalAssignmentExpression(node) {
                node = ts.skipParentheses(node);
                return ts.isBinaryExpression(node) && ts.isLogicalOrCoalescingAssignmentOperator(node.operatorToken.kind);
            }
            function isTopLevelLogicalExpression(node) {
                while (ts.isParenthesizedExpression(node.parent) ||
                    ts.isPrefixUnaryExpression(node.parent) && node.parent.operator === 53 /* ExclamationToken */) {
                    node = node.parent;
                }
                return !isStatementCondition(node) &&
                    !isLogicalAssignmentExpression(node.parent) &&
                    !isLogicalExpression(node.parent) &&
                    !(ts.isOptionalChain(node.parent) && node.parent.expression === node);
            }
            function doWithConditionalBranches(action, value, trueTarget, falseTarget) {
                var savedTrueTarget = currentTrueTarget;
                var savedFalseTarget = currentFalseTarget;
                currentTrueTarget = trueTarget;
                currentFalseTarget = falseTarget;
                action(value);
                currentTrueTarget = savedTrueTarget;
                currentFalseTarget = savedFalseTarget;
            }
            function bindCondition(node, trueTarget, falseTarget) {
                doWithConditionalBranches(bind, node, trueTarget, falseTarget);
                if (!node || !isLogicalAssignmentExpression(node) && !isLogicalExpression(node) && !(ts.isOptionalChain(node) && ts.isOutermostOptionalChain(node))) {
                    addAntecedent(trueTarget, createFlowCondition(32 /* TrueCondition */, currentFlow, node));
                    addAntecedent(falseTarget, createFlowCondition(64 /* FalseCondition */, currentFlow, node));
                }
            }
            function bindIterativeStatement(node, breakTarget, continueTarget) {
                var saveBreakTarget = currentBreakTarget;
                var saveContinueTarget = currentContinueTarget;
                currentBreakTarget = breakTarget;
                currentContinueTarget = continueTarget;
                bind(node);
                currentBreakTarget = saveBreakTarget;
                currentContinueTarget = saveContinueTarget;
            }
            function setContinueTarget(node, target) {
                var label = activeLabelList;
                while (label && node.parent.kind === 248 /* LabeledStatement */) {
                    label.continueTarget = target;
                    label = label.next;
                    node = node.parent;
                }
                return target;
            }
            function bindWhileStatement(node) {
                var preWhileLabel = setContinueTarget(node, createLoopLabel());
                var preBodyLabel = createBranchLabel();
                var postWhileLabel = createBranchLabel();
                addAntecedent(preWhileLabel, currentFlow);
                currentFlow = preWhileLabel;
                bindCondition(node.expression, preBodyLabel, postWhileLabel);
                currentFlow = finishFlowLabel(preBodyLabel);
                bindIterativeStatement(node.statement, postWhileLabel, preWhileLabel);
                addAntecedent(preWhileLabel, currentFlow);
                currentFlow = finishFlowLabel(postWhileLabel);
            }
            function bindDoStatement(node) {
                var preDoLabel = createLoopLabel();
                var preConditionLabel = setContinueTarget(node, createBranchLabel());
                var postDoLabel = createBranchLabel();
                addAntecedent(preDoLabel, currentFlow);
                currentFlow = preDoLabel;
                bindIterativeStatement(node.statement, postDoLabel, preConditionLabel);
                addAntecedent(preConditionLabel, currentFlow);
                currentFlow = finishFlowLabel(preConditionLabel);
                bindCondition(node.expression, preDoLabel, postDoLabel);
                currentFlow = finishFlowLabel(postDoLabel);
            }
            function bindForStatement(node) {
                var preLoopLabel = setContinueTarget(node, createLoopLabel());
                var preBodyLabel = createBranchLabel();
                var postLoopLabel = createBranchLabel();
                bind(node.initializer);
                addAntecedent(preLoopLabel, currentFlow);
                currentFlow = preLoopLabel;
                bindCondition(node.condition, preBodyLabel, postLoopLabel);
                currentFlow = finishFlowLabel(preBodyLabel);
                bindIterativeStatement(node.statement, postLoopLabel, preLoopLabel);
                bind(node.incrementor);
                addAntecedent(preLoopLabel, currentFlow);
                currentFlow = finishFlowLabel(postLoopLabel);
            }
            function bindForInOrForOfStatement(node) {
                var preLoopLabel = setContinueTarget(node, createLoopLabel());
                var postLoopLabel = createBranchLabel();
                bind(node.expression);
                addAntecedent(preLoopLabel, currentFlow);
                currentFlow = preLoopLabel;
                if (node.kind === 242 /* ForOfStatement */) {
                    bind(node.awaitModifier);
                }
                addAntecedent(postLoopLabel, currentFlow);
                bind(node.initializer);
                if (node.initializer.kind !== 253 /* VariableDeclarationList */) {
                    bindAssignmentTargetFlow(node.initializer);
                }
                bindIterativeStatement(node.statement, postLoopLabel, preLoopLabel);
                addAntecedent(preLoopLabel, currentFlow);
                currentFlow = finishFlowLabel(postLoopLabel);
            }
            function bindIfStatement(node) {
                var thenLabel = createBranchLabel();
                var elseLabel = createBranchLabel();
                var postIfLabel = createBranchLabel();
                bindCondition(node.expression, thenLabel, elseLabel);
                currentFlow = finishFlowLabel(thenLabel);
                bind(node.thenStatement);
                addAntecedent(postIfLabel, currentFlow);
                currentFlow = finishFlowLabel(elseLabel);
                bind(node.elseStatement);
                addAntecedent(postIfLabel, currentFlow);
                currentFlow = finishFlowLabel(postIfLabel);
            }
            function bindReturnOrThrow(node) {
                bind(node.expression);
                if (node.kind === 245 /* ReturnStatement */) {
                    hasExplicitReturn = true;
                    if (currentReturnTarget) {
                        addAntecedent(currentReturnTarget, currentFlow);
                    }
                }
                currentFlow = unreachableFlow;
            }
            function findActiveLabel(name) {
                for (var label = activeLabelList; label; label = label.next) {
                    if (label.name === name) {
                        return label;
                    }
                }
                return undefined;
            }
            function bindBreakOrContinueFlow(node, breakTarget, continueTarget) {
                var flowLabel = node.kind === 244 /* BreakStatement */ ? breakTarget : continueTarget;
                if (flowLabel) {
                    addAntecedent(flowLabel, currentFlow);
                    currentFlow = unreachableFlow;
                }
            }
            function bindBreakOrContinueStatement(node) {
                bind(node.label);
                if (node.label) {
                    var activeLabel = findActiveLabel(node.label.escapedText);
                    if (activeLabel) {
                        activeLabel.referenced = true;
                        bindBreakOrContinueFlow(node, activeLabel.breakTarget, activeLabel.continueTarget);
                    }
                }
                else {
                    bindBreakOrContinueFlow(node, currentBreakTarget, currentContinueTarget);
                }
            }
            function bindTryStatement(node) {
                // We conservatively assume that *any* code in the try block can cause an exception, but we only need
                // to track code that causes mutations (because only mutations widen the possible control flow type of
                // a variable). The exceptionLabel is the target label for control flows that result from exceptions.
                // We add all mutation flow nodes as antecedents of this label such that we can analyze them as possible
                // antecedents of the start of catch or finally blocks. Furthermore, we add the current control flow to
                // represent exceptions that occur before any mutations.
                var saveReturnTarget = currentReturnTarget;
                var saveExceptionTarget = currentExceptionTarget;
                var normalExitLabel = createBranchLabel();
                var returnLabel = createBranchLabel();
                var exceptionLabel = createBranchLabel();
                if (node.finallyBlock) {
                    currentReturnTarget = returnLabel;
                }
                addAntecedent(exceptionLabel, currentFlow);
                currentExceptionTarget = exceptionLabel;
                bind(node.tryBlock);
                addAntecedent(normalExitLabel, currentFlow);
                if (node.catchClause) {
                    // Start of catch clause is the target of exceptions from try block.
                    currentFlow = finishFlowLabel(exceptionLabel);
                    // The currentExceptionTarget now represents control flows from exceptions in the catch clause.
                    // Effectively, in a try-catch-finally, if an exception occurs in the try block, the catch block
                    // acts like a second try block.
                    exceptionLabel = createBranchLabel();
                    addAntecedent(exceptionLabel, currentFlow);
                    currentExceptionTarget = exceptionLabel;
                    bind(node.catchClause);
                    addAntecedent(normalExitLabel, currentFlow);
                }
                currentReturnTarget = saveReturnTarget;
                currentExceptionTarget = saveExceptionTarget;
                if (node.finallyBlock) {
                    // Possible ways control can reach the finally block:
                    // 1) Normal completion of try block of a try-finally or try-catch-finally
                    // 2) Normal completion of catch block (following exception in try block) of a try-catch-finally
                    // 3) Return in try or catch block of a try-finally or try-catch-finally
                    // 4) Exception in try block of a try-finally
                    // 5) Exception in catch block of a try-catch-finally
                    // When analyzing a control flow graph that starts inside a finally block we want to consider all
                    // five possibilities above. However, when analyzing a control flow graph that starts outside (past)
                    // the finally block, we only want to consider the first two (if we're past a finally block then it
                    // must have completed normally). Likewise, when analyzing a control flow graph from return statements
                    // in try or catch blocks in an IIFE, we only want to consider the third. To make this possible, we
                    // inject a ReduceLabel node into the control flow graph. This node contains an alternate reduced
                    // set of antecedents for the pre-finally label. As control flow analysis passes by a ReduceLabel
                    // node, the pre-finally label is temporarily switched to the reduced antecedent set.
                    var finallyLabel = createBranchLabel();
                    finallyLabel.antecedents = ts.concatenate(ts.concatenate(normalExitLabel.antecedents, exceptionLabel.antecedents), returnLabel.antecedents);
                    currentFlow = finallyLabel;
                    bind(node.finallyBlock);
                    if (currentFlow.flags & 1 /* Unreachable */) {
                        // If the end of the finally block is unreachable, the end of the entire try statement is unreachable.
                        currentFlow = unreachableFlow;
                    }
                    else {
                        // If we have an IIFE return target and return statements in the try or catch blocks, add a control
                        // flow that goes back through the finally block and back through only the return statements.
                        if (currentReturnTarget && returnLabel.antecedents) {
                            addAntecedent(currentReturnTarget, createReduceLabel(finallyLabel, returnLabel.antecedents, currentFlow));
                        }
                        // If we have an outer exception target (i.e. a containing try-finally or try-catch-finally), add a
                        // control flow that goes back through the finally blok and back through each possible exception source.
                        if (currentExceptionTarget && exceptionLabel.antecedents) {
                            addAntecedent(currentExceptionTarget, createReduceLabel(finallyLabel, exceptionLabel.antecedents, currentFlow));
                        }
                        // If the end of the finally block is reachable, but the end of the try and catch blocks are not,
                        // convert the current flow to unreachable. For example, 'try { return 1; } finally { ... }' should
                        // result in an unreachable current control flow.
                        currentFlow = normalExitLabel.antecedents ? createReduceLabel(finallyLabel, normalExitLabel.antecedents, currentFlow) : unreachableFlow;
                    }
                }
                else {
                    currentFlow = finishFlowLabel(normalExitLabel);
                }
            }
            function bindSwitchStatement(node) {
                var postSwitchLabel = createBranchLabel();
                bind(node.expression);
                var saveBreakTarget = currentBreakTarget;
                var savePreSwitchCaseFlow = preSwitchCaseFlow;
                currentBreakTarget = postSwitchLabel;
                preSwitchCaseFlow = currentFlow;
                bind(node.caseBlock);
                addAntecedent(postSwitchLabel, currentFlow);
                var hasDefault = ts.forEach(node.caseBlock.clauses, function (c) { return c.kind === 288 /* DefaultClause */; });
                // We mark a switch statement as possibly exhaustive if it has no default clause and if all
                // case clauses have unreachable end points (e.g. they all return). Note, we no longer need
                // this property in control flow analysis, it's there only for backwards compatibility.
                node.possiblyExhaustive = !hasDefault && !postSwitchLabel.antecedents;
                if (!hasDefault) {
                    addAntecedent(postSwitchLabel, createFlowSwitchClause(preSwitchCaseFlow, node, 0, 0));
                }
                currentBreakTarget = saveBreakTarget;
                preSwitchCaseFlow = savePreSwitchCaseFlow;
                currentFlow = finishFlowLabel(postSwitchLabel);
            }
            function bindCaseBlock(node) {
                var clauses = node.clauses;
                var isNarrowingSwitch = isNarrowingExpression(node.parent.expression);
                var fallthroughFlow = unreachableFlow;
                for (var i = 0; i < clauses.length; i++) {
                    var clauseStart = i;
                    while (!clauses[i].statements.length && i + 1 < clauses.length) {
                        bind(clauses[i]);
                        i++;
                    }
                    var preCaseLabel = createBranchLabel();
                    addAntecedent(preCaseLabel, isNarrowingSwitch ? createFlowSwitchClause(preSwitchCaseFlow, node.parent, clauseStart, i + 1) : preSwitchCaseFlow);
                    addAntecedent(preCaseLabel, fallthroughFlow);
                    currentFlow = finishFlowLabel(preCaseLabel);
                    var clause = clauses[i];
                    bind(clause);
                    fallthroughFlow = currentFlow;
                    if (!(currentFlow.flags & 1 /* Unreachable */) && i !== clauses.length - 1 && options.noFallthroughCasesInSwitch) {
                        clause.fallthroughFlowNode = currentFlow;
                    }
                }
            }
            function bindCaseClause(node) {
                var saveCurrentFlow = currentFlow;
                currentFlow = preSwitchCaseFlow;
                bind(node.expression);
                currentFlow = saveCurrentFlow;
                bindEach(node.statements);
            }
            function bindExpressionStatement(node) {
                bind(node.expression);
                maybeBindExpressionFlowIfCall(node.expression);
            }
            function maybeBindExpressionFlowIfCall(node) {
                // A top level or LHS of comma expression call expression with a dotted function name and at least one argument
                // is potentially an assertion and is therefore included in the control flow.
                if (node.kind === 206 /* CallExpression */) {
                    var call = node;
                    if (call.expression.kind !== 106 /* SuperKeyword */ && ts.isDottedName(call.expression)) {
                        currentFlow = createFlowCall(currentFlow, call);
                    }
                }
            }
            function bindLabeledStatement(node) {
                var postStatementLabel = createBranchLabel();
                activeLabelList = {
                    next: activeLabelList,
                    name: node.label.escapedText,
                    breakTarget: postStatementLabel,
                    continueTarget: undefined,
                    referenced: false
                };
                bind(node.label);
                bind(node.statement);
                if (!activeLabelList.referenced && !options.allowUnusedLabels) {
                    errorOrSuggestionOnNode(ts.unusedLabelIsError(options), node.label, ts.Diagnostics.Unused_label);
                }
                activeLabelList = activeLabelList.next;
                addAntecedent(postStatementLabel, currentFlow);
                currentFlow = finishFlowLabel(postStatementLabel);
            }
            function bindDestructuringTargetFlow(node) {
                if (node.kind === 219 /* BinaryExpression */ && node.operatorToken.kind === 63 /* EqualsToken */) {
                    bindAssignmentTargetFlow(node.left);
                }
                else {
                    bindAssignmentTargetFlow(node);
                }
            }
            function bindAssignmentTargetFlow(node) {
                if (isNarrowableReference(node)) {
                    currentFlow = createFlowMutation(16 /* Assignment */, currentFlow, node);
                }
                else if (node.kind === 202 /* ArrayLiteralExpression */) {
                    for (var _i = 0, _a = node.elements; _i < _a.length; _i++) {
                        var e = _a[_i];
                        if (e.kind === 223 /* SpreadElement */) {
                            bindAssignmentTargetFlow(e.expression);
                        }
                        else {
                            bindDestructuringTargetFlow(e);
                        }
                    }
                }
                else if (node.kind === 203 /* ObjectLiteralExpression */) {
                    for (var _b = 0, _c = node.properties; _b < _c.length; _b++) {
                        var p = _c[_b];
                        if (p.kind === 291 /* PropertyAssignment */) {
                            bindDestructuringTargetFlow(p.initializer);
                        }
                        else if (p.kind === 292 /* ShorthandPropertyAssignment */) {
                            bindAssignmentTargetFlow(p.name);
                        }
                        else if (p.kind === 293 /* SpreadAssignment */) {
                            bindAssignmentTargetFlow(p.expression);
                        }
                    }
                }
            }
            function bindLogicalLikeExpression(node, trueTarget, falseTarget) {
                var preRightLabel = createBranchLabel();
                if (node.operatorToken.kind === 55 /* AmpersandAmpersandToken */ || node.operatorToken.kind === 76 /* AmpersandAmpersandEqualsToken */) {
                    bindCondition(node.left, preRightLabel, falseTarget);
                }
                else {
                    bindCondition(node.left, trueTarget, preRightLabel);
                }
                currentFlow = finishFlowLabel(preRightLabel);
                bind(node.operatorToken);
                if (ts.isLogicalOrCoalescingAssignmentOperator(node.operatorToken.kind)) {
                    doWithConditionalBranches(bind, node.right, trueTarget, falseTarget);
                    bindAssignmentTargetFlow(node.left);
                    addAntecedent(trueTarget, createFlowCondition(32 /* TrueCondition */, currentFlow, node));
                    addAntecedent(falseTarget, createFlowCondition(64 /* FalseCondition */, currentFlow, node));
                }
                else {
                    bindCondition(node.right, trueTarget, falseTarget);
                }
            }
            function bindPrefixUnaryExpressionFlow(node) {
                if (node.operator === 53 /* ExclamationToken */) {
                    var saveTrueTarget = currentTrueTarget;
                    currentTrueTarget = currentFalseTarget;
                    currentFalseTarget = saveTrueTarget;
                    bindEachChild(node);
                    currentFalseTarget = currentTrueTarget;
                    currentTrueTarget = saveTrueTarget;
                }
                else {
                    bindEachChild(node);
                    if (node.operator === 45 /* PlusPlusToken */ || node.operator === 46 /* MinusMinusToken */) {
                        bindAssignmentTargetFlow(node.operand);
                    }
                }
            }
            function bindPostfixUnaryExpressionFlow(node) {
                bindEachChild(node);
                if (node.operator === 45 /* PlusPlusToken */ || node.operator === 46 /* MinusMinusToken */) {
                    bindAssignmentTargetFlow(node.operand);
                }
            }
            function bindDestructuringAssignmentFlow(node) {
                if (inAssignmentPattern) {
                    inAssignmentPattern = false;
                    bind(node.operatorToken);
                    bind(node.right);
                    inAssignmentPattern = true;
                    bind(node.left);
                }
                else {
                    inAssignmentPattern = true;
                    bind(node.left);
                    inAssignmentPattern = false;
                    bind(node.operatorToken);
                    bind(node.right);
                }
                bindAssignmentTargetFlow(node.left);
            }
            function createBindBinaryExpressionFlow() {
                return ts.createBinaryExpressionTrampoline(onEnter, onLeft, onOperator, onRight, onExit, /*foldState*/ undefined);
                function onEnter(node, state) {
                    if (state) {
                        state.stackIndex++;
                        // Emulate the work that `bind` does before reaching `bindChildren`. A normal call to
                        // `bindBinaryExpressionFlow` will already have done this work.
                        ts.setParent(node, parent);
                        var saveInStrictMode = inStrictMode;
                        bindWorker(node);
                        var saveParent = parent;
                        parent = node;
                        state.skip = false;
                        state.inStrictModeStack[state.stackIndex] = saveInStrictMode;
                        state.parentStack[state.stackIndex] = saveParent;
                    }
                    else {
                        state = {
                            stackIndex: 0,
                            skip: false,
                            inStrictModeStack: [undefined],
                            parentStack: [undefined]
                        };
                    }
                    // TODO: bindLogicalExpression is recursive - if we want to handle deeply nested `&&` expressions
                    // we'll need to handle the `bindLogicalExpression` scenarios in this state machine, too
                    // For now, though, since the common cases are chained `+`, leaving it recursive is fine
                    var operator = node.operatorToken.kind;
                    if (operator === 55 /* AmpersandAmpersandToken */ ||
                        operator === 56 /* BarBarToken */ ||
                        operator === 60 /* QuestionQuestionToken */ ||
                        ts.isLogicalOrCoalescingAssignmentOperator(operator)) {
                        if (isTopLevelLogicalExpression(node)) {
                            var postExpressionLabel = createBranchLabel();
                            bindLogicalLikeExpression(node, postExpressionLabel, postExpressionLabel);
                            currentFlow = finishFlowLabel(postExpressionLabel);
                        }
                        else {
                            bindLogicalLikeExpression(node, currentTrueTarget, currentFalseTarget);
                        }
                        state.skip = true;
                    }
                    return state;
                }
                function onLeft(left, state, _node) {
                    if (!state.skip) {
                        return maybeBind(left);
                    }
                }
                function onOperator(operatorToken, state, node) {
                    if (!state.skip) {
                        if (operatorToken.kind === 27 /* CommaToken */) {
                            maybeBindExpressionFlowIfCall(node.left);
                        }
                        bind(operatorToken);
                    }
                }
                function onRight(right, state, _node) {
                    if (!state.skip) {
                        return maybeBind(right);
                    }
                }
                function onExit(node, state) {
                    if (!state.skip) {
                        var operator = node.operatorToken.kind;
                        if (ts.isAssignmentOperator(operator) && !ts.isAssignmentTarget(node)) {
                            bindAssignmentTargetFlow(node.left);
                            if (operator === 63 /* EqualsToken */ && node.left.kind === 205 /* ElementAccessExpression */) {
                                var elementAccess = node.left;
                                if (isNarrowableOperand(elementAccess.expression)) {
                                    currentFlow = createFlowMutation(256 /* ArrayMutation */, currentFlow, node);
                                }
                            }
                        }
                    }
                    var savedInStrictMode = state.inStrictModeStack[state.stackIndex];
                    var savedParent = state.parentStack[state.stackIndex];
                    if (savedInStrictMode !== undefined) {
                        inStrictMode = savedInStrictMode;
                    }
                    if (savedParent !== undefined) {
                        parent = savedParent;
                    }
                    state.skip = false;
                    state.stackIndex--;
                }
                function maybeBind(node) {
                    if (node && ts.isBinaryExpression(node) && !ts.isDestructuringAssignment(node)) {
                        return node;
                    }
                    bind(node);
                }
            }
            function bindDeleteExpressionFlow(node) {
                bindEachChild(node);
                if (node.expression.kind === 204 /* PropertyAccessExpression */) {
                    bindAssignmentTargetFlow(node.expression);
                }
            }
            function bindConditionalExpressionFlow(node) {
                var trueLabel = createBranchLabel();
                var falseLabel = createBranchLabel();
                var postExpressionLabel = createBranchLabel();
                bindCondition(node.condition, trueLabel, falseLabel);
                currentFlow = finishFlowLabel(trueLabel);
                bind(node.questionToken);
                bind(node.whenTrue);
                addAntecedent(postExpressionLabel, currentFlow);
                currentFlow = finishFlowLabel(falseLabel);
                bind(node.colonToken);
                bind(node.whenFalse);
                addAntecedent(postExpressionLabel, currentFlow);
                currentFlow = finishFlowLabel(postExpressionLabel);
            }
            function bindInitializedVariableFlow(node) {
                var name = !ts.isOmittedExpression(node) ? node.name : undefined;
                if (ts.isBindingPattern(name)) {
                    for (var _i = 0, _a = name.elements; _i < _a.length; _i++) {
                        var child = _a[_i];
                        bindInitializedVariableFlow(child);
                    }
                }
                else {
                    currentFlow = createFlowMutation(16 /* Assignment */, currentFlow, node);
                }
            }
            function bindVariableDeclarationFlow(node) {
                bindEachChild(node);
                if (node.initializer || ts.isForInOrOfStatement(node.parent.parent)) {
                    bindInitializedVariableFlow(node);
                }
            }
            function bindBindingElementFlow(node) {
                if (ts.isBindingPattern(node.name)) {
                    // When evaluating a binding pattern, the initializer is evaluated before the binding pattern, per:
                    // - https://tc39.es/ecma262/#sec-destructuring-binding-patterns-runtime-semantics-iteratorbindinginitialization
                    //   - `BindingElement: BindingPattern Initializer?`
                    // - https://tc39.es/ecma262/#sec-runtime-semantics-keyedbindinginitialization
                    //   - `BindingElement: BindingPattern Initializer?`
                    bindEach(node.decorators);
                    bindEach(node.modifiers);
                    bind(node.dotDotDotToken);
                    bind(node.propertyName);
                    bind(node.initializer);
                    bind(node.name);
                }
                else {
                    bindEachChild(node);
                }
            }
            function bindJSDocTypeAlias(node) {
                bind(node.tagName);
                if (node.kind !== 334 /* JSDocEnumTag */ && node.fullName) {
                    // don't bind the type name yet; that's delayed until delayedBindJSDocTypedefTag
                    ts.setParent(node.fullName, node);
                    ts.setParentRecursive(node.fullName, /*incremental*/ false);
                }
                if (typeof node.comment !== "string") {
                    bindEach(node.comment);
                }
            }
            function bindJSDocClassTag(node) {
                bindEachChild(node);
                var host = ts.getHostSignatureFromJSDoc(node);
                if (host && host.kind !== 167 /* MethodDeclaration */) {
                    addDeclarationToSymbol(host.symbol, host, 32 /* Class */);
                }
            }
            function bindOptionalExpression(node, trueTarget, falseTarget) {
                doWithConditionalBranches(bind, node, trueTarget, falseTarget);
                if (!ts.isOptionalChain(node) || ts.isOutermostOptionalChain(node)) {
                    addAntecedent(trueTarget, createFlowCondition(32 /* TrueCondition */, currentFlow, node));
                    addAntecedent(falseTarget, createFlowCondition(64 /* FalseCondition */, currentFlow, node));
                }
            }
            function bindOptionalChainRest(node) {
                switch (node.kind) {
                    case 204 /* PropertyAccessExpression */:
                        bind(node.questionDotToken);
                        bind(node.name);
                        break;
                    case 205 /* ElementAccessExpression */:
                        bind(node.questionDotToken);
                        bind(node.argumentExpression);
                        break;
                    case 206 /* CallExpression */:
                        bind(node.questionDotToken);
                        bindEach(node.typeArguments);
                        bindEach(node.arguments);
                        break;
                }
            }
            function bindOptionalChain(node, trueTarget, falseTarget) {
                // For an optional chain, we emulate the behavior of a logical expression:
                //
                // a?.b         -> a && a.b
                // a?.b.c       -> a && a.b.c
                // a?.b?.c      -> a && a.b && a.b.c
                // a?.[x = 1]   -> a && a[x = 1]
                //
                // To do this we descend through the chain until we reach the root of a chain (the expression with a `?.`)
                // and build it's CFA graph as if it were the first condition (`a && ...`). Then we bind the rest
                // of the node as part of the "true" branch, and continue to do so as we ascend back up to the outermost
                // chain node. We then treat the entire node as the right side of the expression.
                var preChainLabel = ts.isOptionalChainRoot(node) ? createBranchLabel() : undefined;
                bindOptionalExpression(node.expression, preChainLabel || trueTarget, falseTarget);
                if (preChainLabel) {
                    currentFlow = finishFlowLabel(preChainLabel);
                }
                doWithConditionalBranches(bindOptionalChainRest, node, trueTarget, falseTarget);
                if (ts.isOutermostOptionalChain(node)) {
                    addAntecedent(trueTarget, createFlowCondition(32 /* TrueCondition */, currentFlow, node));
                    addAntecedent(falseTarget, createFlowCondition(64 /* FalseCondition */, currentFlow, node));
                }
            }
            function bindOptionalChainFlow(node) {
                if (isTopLevelLogicalExpression(node)) {
                    var postExpressionLabel = createBranchLabel();
                    bindOptionalChain(node, postExpressionLabel, postExpressionLabel);
                    currentFlow = finishFlowLabel(postExpressionLabel);
                }
                else {
                    bindOptionalChain(node, currentTrueTarget, currentFalseTarget);
                }
            }
            function bindNonNullExpressionFlow(node) {
                if (ts.isOptionalChain(node)) {
                    bindOptionalChainFlow(node);
                }
                else {
                    bindEachChild(node);
                }
            }
            function bindAccessExpressionFlow(node) {
                if (ts.isOptionalChain(node)) {
                    bindOptionalChainFlow(node);
                }
                else {
                    bindEachChild(node);
                }
            }
            function bindCallExpressionFlow(node) {
                if (ts.isOptionalChain(node)) {
                    bindOptionalChainFlow(node);
                }
                else {
                    // If the target of the call expression is a function expression or arrow function we have
                    // an immediately invoked function expression (IIFE). Initialize the flowNode property to
                    // the current control flow (which includes evaluation of the IIFE arguments).
                    var expr = ts.skipParentheses(node.expression);
                    if (expr.kind === 211 /* FunctionExpression */ || expr.kind === 212 /* ArrowFunction */) {
                        bindEach(node.typeArguments);
                        bindEach(node.arguments);
                        bind(node.expression);
                    }
                    else {
                        bindEachChild(node);
                        if (node.expression.kind === 106 /* SuperKeyword */) {
                            currentFlow = createFlowCall(currentFlow, node);
                        }
                    }
                }
                if (node.expression.kind === 204 /* PropertyAccessExpression */) {
                    var propertyAccess = node.expression;
                    if (ts.isIdentifier(propertyAccess.name) && isNarrowableOperand(propertyAccess.expression) && ts.isPushOrUnshiftIdentifier(propertyAccess.name)) {
                        currentFlow = createFlowMutation(256 /* ArrayMutation */, currentFlow, node);
                    }
                }
            }
            function getContainerFlags(node) {
                switch (node.kind) {
                    case 224 /* ClassExpression */:
                    case 255 /* ClassDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 203 /* ObjectLiteralExpression */:
                    case 180 /* TypeLiteral */:
                    case 317 /* JSDocTypeLiteral */:
                    case 284 /* JsxAttributes */:
                        return 1 /* IsContainer */;
                    case 256 /* InterfaceDeclaration */:
                        return 1 /* IsContainer */ | 64 /* IsInterface */;
                    case 259 /* ModuleDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                    case 193 /* MappedType */:
                        return 1 /* IsContainer */ | 32 /* HasLocals */;
                    case 300 /* SourceFile */:
                        return 1 /* IsContainer */ | 4 /* IsControlFlowContainer */ | 32 /* HasLocals */;
                    case 167 /* MethodDeclaration */:
                        if (ts.isObjectLiteralOrClassExpressionMethod(node)) {
                            return 1 /* IsContainer */ | 4 /* IsControlFlowContainer */ | 32 /* HasLocals */ | 8 /* IsFunctionLike */ | 128 /* IsObjectLiteralOrClassExpressionMethod */;
                        }
                    // falls through
                    case 169 /* Constructor */:
                    case 254 /* FunctionDeclaration */:
                    case 166 /* MethodSignature */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 172 /* CallSignature */:
                    case 318 /* JSDocSignature */:
                    case 312 /* JSDocFunctionType */:
                    case 177 /* FunctionType */:
                    case 173 /* ConstructSignature */:
                    case 174 /* IndexSignature */:
                    case 178 /* ConstructorType */:
                    case 168 /* ClassStaticBlockDeclaration */:
                        return 1 /* IsContainer */ | 4 /* IsControlFlowContainer */ | 32 /* HasLocals */ | 8 /* IsFunctionLike */;
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        return 1 /* IsContainer */ | 4 /* IsControlFlowContainer */ | 32 /* HasLocals */ | 8 /* IsFunctionLike */ | 16 /* IsFunctionExpression */;
                    case 260 /* ModuleBlock */:
                        return 4 /* IsControlFlowContainer */;
                    case 165 /* PropertyDeclaration */:
                        return node.initializer ? 4 /* IsControlFlowContainer */ : 0;
                    case 290 /* CatchClause */:
                    case 240 /* ForStatement */:
                    case 241 /* ForInStatement */:
                    case 242 /* ForOfStatement */:
                    case 261 /* CaseBlock */:
                        return 2 /* IsBlockScopedContainer */;
                    case 233 /* Block */:
                        // do not treat blocks directly inside a function as a block-scoped-container.
                        // Locals that reside in this block should go to the function locals. Otherwise 'x'
                        // would not appear to be a redeclaration of a block scoped local in the following
                        // example:
                        //
                        //      function foo() {
                        //          var x;
                        //          let x;
                        //      }
                        //
                        // If we placed 'var x' into the function locals and 'let x' into the locals of
                        // the block, then there would be no collision.
                        //
                        // By not creating a new block-scoped-container here, we ensure that both 'var x'
                        // and 'let x' go into the Function-container's locals, and we do get a collision
                        // conflict.
                        return ts.isFunctionLike(node.parent) || ts.isClassStaticBlockDeclaration(node.parent) ? 0 /* None */ : 2 /* IsBlockScopedContainer */;
                }
                return 0 /* None */;
            }
            function addToContainerChain(next) {
                if (lastContainer) {
                    lastContainer.nextContainer = next;
                }
                lastContainer = next;
            }
            function declareSymbolAndAddToSymbolTable(node, symbolFlags, symbolExcludes) {
                switch (container.kind) {
                    // Modules, source files, and classes need specialized handling for how their
                    // members are declared (for example, a member of a class will go into a specific
                    // symbol table depending on if it is static or not). We defer to specialized
                    // handlers to take care of declaring these child members.
                    case 259 /* ModuleDeclaration */:
                        return declareModuleMember(node, symbolFlags, symbolExcludes);
                    case 300 /* SourceFile */:
                        return declareSourceFileMember(node, symbolFlags, symbolExcludes);
                    case 224 /* ClassExpression */:
                    case 255 /* ClassDeclaration */:
                        return declareClassMember(node, symbolFlags, symbolExcludes);
                    case 258 /* EnumDeclaration */:
                        return declareSymbol(container.symbol.exports, container.symbol, node, symbolFlags, symbolExcludes);
                    case 180 /* TypeLiteral */:
                    case 317 /* JSDocTypeLiteral */:
                    case 203 /* ObjectLiteralExpression */:
                    case 256 /* InterfaceDeclaration */:
                    case 284 /* JsxAttributes */:
                        // Interface/Object-types always have their children added to the 'members' of
                        // their container. They are only accessible through an instance of their
                        // container, and are never in scope otherwise (even inside the body of the
                        // object / type / interface declaring them). An exception is type parameters,
                        // which are in scope without qualification (similar to 'locals').
                        return declareSymbol(container.symbol.members, container.symbol, node, symbolFlags, symbolExcludes);
                    case 177 /* FunctionType */:
                    case 178 /* ConstructorType */:
                    case 172 /* CallSignature */:
                    case 173 /* ConstructSignature */:
                    case 318 /* JSDocSignature */:
                    case 174 /* IndexSignature */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 169 /* Constructor */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                    case 312 /* JSDocFunctionType */:
                    case 340 /* JSDocTypedefTag */:
                    case 333 /* JSDocCallbackTag */:
                    case 168 /* ClassStaticBlockDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                    case 193 /* MappedType */:
                        // All the children of these container types are never visible through another
                        // symbol (i.e. through another symbol's 'exports' or 'members').  Instead,
                        // they're only accessed 'lexically' (i.e. from code that exists underneath
                        // their container in the tree). To accomplish this, we simply add their declared
                        // symbol to the 'locals' of the container.  These symbols can then be found as
                        // the type checker walks up the containers, checking them for matching names.
                        return declareSymbol(container.locals, /*parent*/ undefined, node, symbolFlags, symbolExcludes);
                }
            }
            function declareClassMember(node, symbolFlags, symbolExcludes) {
                return ts.isStatic(node)
                    ? declareSymbol(container.symbol.exports, container.symbol, node, symbolFlags, symbolExcludes)
                    : declareSymbol(container.symbol.members, container.symbol, node, symbolFlags, symbolExcludes);
            }
            function declareSourceFileMember(node, symbolFlags, symbolExcludes) {
                return ts.isExternalModule(file)
                    ? declareModuleMember(node, symbolFlags, symbolExcludes)
                    : declareSymbol(file.locals, /*parent*/ undefined, node, symbolFlags, symbolExcludes);
            }
            function hasExportDeclarations(node) {
                var body = ts.isSourceFile(node) ? node : ts.tryCast(node.body, ts.isModuleBlock);
                return !!body && body.statements.some(function (s) { return ts.isExportDeclaration(s) || ts.isExportAssignment(s); });
            }
            function setExportContextFlag(node) {
                // A declaration source file or ambient module declaration that contains no export declarations (but possibly regular
                // declarations with export modifiers) is an export context in which declarations are implicitly exported.
                if (node.flags & 8388608 /* Ambient */ && !hasExportDeclarations(node)) {
                    node.flags |= 64 /* ExportContext */;
                }
                else {
                    node.flags &= ~64 /* ExportContext */;
                }
            }
            function bindModuleDeclaration(node) {
                setExportContextFlag(node);
                if (ts.isAmbientModule(node)) {
                    if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                        errorOnFirstToken(node, ts.Diagnostics.export_modifier_cannot_be_applied_to_ambient_modules_and_module_augmentations_since_they_are_always_visible);
                    }
                    if (ts.isModuleAugmentationExternal(node)) {
                        declareModuleSymbol(node);
                    }
                    else {
                        var pattern = void 0;
                        if (node.name.kind === 10 /* StringLiteral */) {
                            var text = node.name.text;
                            pattern = ts.tryParsePattern(text);
                            if (pattern === undefined) {
                                errorOnFirstToken(node.name, ts.Diagnostics.Pattern_0_can_have_at_most_one_Asterisk_character, text);
                            }
                        }
                        var symbol = declareSymbolAndAddToSymbolTable(node, 512 /* ValueModule */, 110735 /* ValueModuleExcludes */);
                        file.patternAmbientModules = ts.append(file.patternAmbientModules, pattern && !ts.isString(pattern) ? { pattern: pattern, symbol: symbol } : undefined);
                    }
                }
                else {
                    var state = declareModuleSymbol(node);
                    if (state !== 0 /* NonInstantiated */) {
                        var symbol = node.symbol;
                        // if module was already merged with some function, class or non-const enum, treat it as non-const-enum-only
                        symbol.constEnumOnlyModule = (!(symbol.flags & (16 /* Function */ | 32 /* Class */ | 256 /* RegularEnum */)))
                            // Current must be `const enum` only
                            && state === 2 /* ConstEnumOnly */
                            // Can't have been set to 'false' in a previous merged symbol. ('undefined' OK)
                            && symbol.constEnumOnlyModule !== false;
                    }
                }
            }
            function declareModuleSymbol(node) {
                var state = getModuleInstanceState(node);
                var instantiated = state !== 0 /* NonInstantiated */;
                declareSymbolAndAddToSymbolTable(node, instantiated ? 512 /* ValueModule */ : 1024 /* NamespaceModule */, instantiated ? 110735 /* ValueModuleExcludes */ : 0 /* NamespaceModuleExcludes */);
                return state;
            }
            function bindFunctionOrConstructorType(node) {
                // For a given function symbol "<...>(...) => T" we want to generate a symbol identical
                // to the one we would get for: { <...>(...): T }
                //
                // We do that by making an anonymous type literal symbol, and then setting the function
                // symbol as its sole member. To the rest of the system, this symbol will be indistinguishable
                // from an actual type literal symbol you would have gotten had you used the long form.
                var symbol = createSymbol(131072 /* Signature */, getDeclarationName(node)); // TODO: GH#18217
                addDeclarationToSymbol(symbol, node, 131072 /* Signature */);
                var typeLiteralSymbol = createSymbol(2048 /* TypeLiteral */, "__type" /* Type */);
                addDeclarationToSymbol(typeLiteralSymbol, node, 2048 /* TypeLiteral */);
                typeLiteralSymbol.members = ts.createSymbolTable();
                typeLiteralSymbol.members.set(symbol.escapedName, symbol);
            }
            function bindObjectLiteralExpression(node) {
                var ElementKind;
                (function (ElementKind) {
                    ElementKind[ElementKind["Property"] = 1] = "Property";
                    ElementKind[ElementKind["Accessor"] = 2] = "Accessor";
                })(ElementKind || (ElementKind = {}));
                if (inStrictMode && !ts.isAssignmentTarget(node)) {
                    var seen = new ts.Map();
                    for (var _i = 0, _a = node.properties; _i < _a.length; _i++) {
                        var prop = _a[_i];
                        if (prop.kind === 293 /* SpreadAssignment */ || prop.name.kind !== 79 /* Identifier */) {
                            continue;
                        }
                        var identifier = prop.name;
                        // ECMA-262 11.1.5 Object Initializer
                        // If previous is not undefined then throw a SyntaxError exception if any of the following conditions are true
                        // a.This production is contained in strict code and IsDataDescriptor(previous) is true and
                        // IsDataDescriptor(propId.descriptor) is true.
                        //    b.IsDataDescriptor(previous) is true and IsAccessorDescriptor(propId.descriptor) is true.
                        //    c.IsAccessorDescriptor(previous) is true and IsDataDescriptor(propId.descriptor) is true.
                        //    d.IsAccessorDescriptor(previous) is true and IsAccessorDescriptor(propId.descriptor) is true
                        // and either both previous and propId.descriptor have[[Get]] fields or both previous and propId.descriptor have[[Set]] fields
                        var currentKind = prop.kind === 291 /* PropertyAssignment */ || prop.kind === 292 /* ShorthandPropertyAssignment */ || prop.kind === 167 /* MethodDeclaration */
                            ? 1 /* Property */
                            : 2 /* Accessor */;
                        var existingKind = seen.get(identifier.escapedText);
                        if (!existingKind) {
                            seen.set(identifier.escapedText, currentKind);
                            continue;
                        }
                        if (currentKind === 1 /* Property */ && existingKind === 1 /* Property */) {
                            var span = ts.getErrorSpanForNode(file, identifier);
                            file.bindDiagnostics.push(ts.createFileDiagnostic(file, span.start, span.length, ts.Diagnostics.An_object_literal_cannot_have_multiple_properties_with_the_same_name_in_strict_mode));
                        }
                    }
                }
                return bindAnonymousDeclaration(node, 4096 /* ObjectLiteral */, "__object" /* Object */);
            }
            function bindJsxAttributes(node) {
                return bindAnonymousDeclaration(node, 4096 /* ObjectLiteral */, "__jsxAttributes" /* JSXAttributes */);
            }
            function bindJsxAttribute(node, symbolFlags, symbolExcludes) {
                return declareSymbolAndAddToSymbolTable(node, symbolFlags, symbolExcludes);
            }
            function bindAnonymousDeclaration(node, symbolFlags, name) {
                var symbol = createSymbol(symbolFlags, name);
                if (symbolFlags & (8 /* EnumMember */ | 106500 /* ClassMember */)) {
                    symbol.parent = container.symbol;
                }
                addDeclarationToSymbol(symbol, node, symbolFlags);
                return symbol;
            }
            function bindBlockScopedDeclaration(node, symbolFlags, symbolExcludes) {
                switch (blockScopeContainer.kind) {
                    case 259 /* ModuleDeclaration */:
                        declareModuleMember(node, symbolFlags, symbolExcludes);
                        break;
                    case 300 /* SourceFile */:
                        if (ts.isExternalOrCommonJsModule(container)) {
                            declareModuleMember(node, symbolFlags, symbolExcludes);
                            break;
                        }
                    // falls through
                    default:
                        if (!blockScopeContainer.locals) {
                            blockScopeContainer.locals = ts.createSymbolTable();
                            addToContainerChain(blockScopeContainer);
                        }
                        declareSymbol(blockScopeContainer.locals, /*parent*/ undefined, node, symbolFlags, symbolExcludes);
                }
            }
            function delayedBindJSDocTypedefTag() {
                if (!delayedTypeAliases) {
                    return;
                }
                var saveContainer = container;
                var saveLastContainer = lastContainer;
                var saveBlockScopeContainer = blockScopeContainer;
                var saveParent = parent;
                var saveCurrentFlow = currentFlow;
                for (var _i = 0, delayedTypeAliases_1 = delayedTypeAliases; _i < delayedTypeAliases_1.length; _i++) {
                    var typeAlias = delayedTypeAliases_1[_i];
                    var host = typeAlias.parent.parent;
                    container = ts.findAncestor(host.parent, function (n) { return !!(getContainerFlags(n) & 1 /* IsContainer */); }) || file;
                    blockScopeContainer = ts.getEnclosingBlockScopeContainer(host) || file;
                    currentFlow = initFlowNode({ flags: 2 /* Start */ });
                    parent = typeAlias;
                    bind(typeAlias.typeExpression);
                    var declName = ts.getNameOfDeclaration(typeAlias);
                    if ((ts.isJSDocEnumTag(typeAlias) || !typeAlias.fullName) && declName && ts.isPropertyAccessEntityNameExpression(declName.parent)) {
                        // typedef anchored to an A.B.C assignment - we need to bind into B's namespace under name C
                        var isTopLevel = isTopLevelNamespaceAssignment(declName.parent);
                        if (isTopLevel) {
                            bindPotentiallyMissingNamespaces(file.symbol, declName.parent, isTopLevel, !!ts.findAncestor(declName, function (d) { return ts.isPropertyAccessExpression(d) && d.name.escapedText === "prototype"; }), /*containerIsClass*/ false);
                            var oldContainer = container;
                            switch (ts.getAssignmentDeclarationPropertyAccessKind(declName.parent)) {
                                case 1 /* ExportsProperty */:
                                case 2 /* ModuleExports */:
                                    if (!ts.isExternalOrCommonJsModule(file)) {
                                        container = undefined;
                                    }
                                    else {
                                        container = file;
                                    }
                                    break;
                                case 4 /* ThisProperty */:
                                    container = declName.parent.expression;
                                    break;
                                case 3 /* PrototypeProperty */:
                                    container = declName.parent.expression.name;
                                    break;
                                case 5 /* Property */:
                                    container = isExportsOrModuleExportsOrAlias(file, declName.parent.expression) ? file
                                        : ts.isPropertyAccessExpression(declName.parent.expression) ? declName.parent.expression.name
                                            : declName.parent.expression;
                                    break;
                                case 0 /* None */:
                                    return ts.Debug.fail("Shouldn't have detected typedef or enum on non-assignment declaration");
                            }
                            if (container) {
                                declareModuleMember(typeAlias, 524288 /* TypeAlias */, 788968 /* TypeAliasExcludes */);
                            }
                            container = oldContainer;
                        }
                    }
                    else if (ts.isJSDocEnumTag(typeAlias) || !typeAlias.fullName || typeAlias.fullName.kind === 79 /* Identifier */) {
                        parent = typeAlias.parent;
                        bindBlockScopedDeclaration(typeAlias, 524288 /* TypeAlias */, 788968 /* TypeAliasExcludes */);
                    }
                    else {
                        bind(typeAlias.fullName);
                    }
                }
                container = saveContainer;
                lastContainer = saveLastContainer;
                blockScopeContainer = saveBlockScopeContainer;
                parent = saveParent;
                currentFlow = saveCurrentFlow;
            }
            // The binder visits every node in the syntax tree so it is a convenient place to perform a single localized
            // check for reserved words used as identifiers in strict mode code, as well as `yield` or `await` in
            // [Yield] or [Await] contexts, respectively.
            function checkContextualIdentifier(node) {
                // Report error only if there are no parse errors in file
                if (!file.parseDiagnostics.length &&
                    !(node.flags & 8388608 /* Ambient */) &&
                    !(node.flags & 4194304 /* JSDoc */) &&
                    !ts.isIdentifierName(node)) {
                    // strict mode identifiers
                    if (inStrictMode &&
                        node.originalKeywordKind >= 117 /* FirstFutureReservedWord */ &&
                        node.originalKeywordKind <= 125 /* LastFutureReservedWord */) {
                        file.bindDiagnostics.push(createDiagnosticForNode(node, getStrictModeIdentifierMessage(node), ts.declarationNameToString(node)));
                    }
                    else if (node.originalKeywordKind === 131 /* AwaitKeyword */) {
                        if (ts.isExternalModule(file) && ts.isInTopLevelContext(node)) {
                            file.bindDiagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.Identifier_expected_0_is_a_reserved_word_at_the_top_level_of_a_module, ts.declarationNameToString(node)));
                        }
                        else if (node.flags & 32768 /* AwaitContext */) {
                            file.bindDiagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here, ts.declarationNameToString(node)));
                        }
                    }
                    else if (node.originalKeywordKind === 125 /* YieldKeyword */ && node.flags & 8192 /* YieldContext */) {
                        file.bindDiagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.Identifier_expected_0_is_a_reserved_word_that_cannot_be_used_here, ts.declarationNameToString(node)));
                    }
                }
            }
            function getStrictModeIdentifierMessage(node) {
                // Provide specialized messages to help the user understand why we think they're in
                // strict mode.
                if (ts.getContainingClass(node)) {
                    return ts.Diagnostics.Identifier_expected_0_is_a_reserved_word_in_strict_mode_Class_definitions_are_automatically_in_strict_mode;
                }
                if (file.externalModuleIndicator) {
                    return ts.Diagnostics.Identifier_expected_0_is_a_reserved_word_in_strict_mode_Modules_are_automatically_in_strict_mode;
                }
                return ts.Diagnostics.Identifier_expected_0_is_a_reserved_word_in_strict_mode;
            }
            // The binder visits every node, so this is a good place to check for
            // the reserved private name (there is only one)
            function checkPrivateIdentifier(node) {
                if (node.escapedText === "#constructor") {
                    // Report error only if there are no parse errors in file
                    if (!file.parseDiagnostics.length) {
                        file.bindDiagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.constructor_is_a_reserved_word, ts.declarationNameToString(node)));
                    }
                }
            }
            function checkStrictModeBinaryExpression(node) {
                if (inStrictMode && ts.isLeftHandSideExpression(node.left) && ts.isAssignmentOperator(node.operatorToken.kind)) {
                    // ECMA 262 (Annex C) The identifier eval or arguments may not appear as the LeftHandSideExpression of an
                    // Assignment operator(11.13) or of a PostfixExpression(11.3)
                    checkStrictModeEvalOrArguments(node, node.left);
                }
            }
            function checkStrictModeCatchClause(node) {
                // It is a SyntaxError if a TryStatement with a Catch occurs within strict code and the Identifier of the
                // Catch production is eval or arguments
                if (inStrictMode && node.variableDeclaration) {
                    checkStrictModeEvalOrArguments(node, node.variableDeclaration.name);
                }
            }
            function checkStrictModeDeleteExpression(node) {
                // Grammar checking
                if (inStrictMode && node.expression.kind === 79 /* Identifier */) {
                    // When a delete operator occurs within strict mode code, a SyntaxError is thrown if its
                    // UnaryExpression is a direct reference to a variable, function argument, or function name
                    var span = ts.getErrorSpanForNode(file, node.expression);
                    file.bindDiagnostics.push(ts.createFileDiagnostic(file, span.start, span.length, ts.Diagnostics.delete_cannot_be_called_on_an_identifier_in_strict_mode));
                }
            }
            function isEvalOrArgumentsIdentifier(node) {
                return ts.isIdentifier(node) && (node.escapedText === "eval" || node.escapedText === "arguments");
            }
            function checkStrictModeEvalOrArguments(contextNode, name) {
                if (name && name.kind === 79 /* Identifier */) {
                    var identifier = name;
                    if (isEvalOrArgumentsIdentifier(identifier)) {
                        // We check first if the name is inside class declaration or class expression; if so give explicit message
                        // otherwise report generic error message.
                        var span = ts.getErrorSpanForNode(file, name);
                        file.bindDiagnostics.push(ts.createFileDiagnostic(file, span.start, span.length, getStrictModeEvalOrArgumentsMessage(contextNode), ts.idText(identifier)));
                    }
                }
            }
            function getStrictModeEvalOrArgumentsMessage(node) {
                // Provide specialized messages to help the user understand why we think they're in
                // strict mode.
                if (ts.getContainingClass(node)) {
                    return ts.Diagnostics.Code_contained_in_a_class_is_evaluated_in_JavaScript_s_strict_mode_which_does_not_allow_this_use_of_0_For_more_information_see_https_Colon_Slash_Slashdeveloper_mozilla_org_Slashen_US_Slashdocs_SlashWeb_SlashJavaScript_SlashReference_SlashStrict_mode;
                }
                if (file.externalModuleIndicator) {
                    return ts.Diagnostics.Invalid_use_of_0_Modules_are_automatically_in_strict_mode;
                }
                return ts.Diagnostics.Invalid_use_of_0_in_strict_mode;
            }
            function checkStrictModeFunctionName(node) {
                if (inStrictMode) {
                    // It is a SyntaxError if the identifier eval or arguments appears within a FormalParameterList of a strict mode FunctionDeclaration or FunctionExpression (13.1))
                    checkStrictModeEvalOrArguments(node, node.name);
                }
            }
            function getStrictModeBlockScopeFunctionDeclarationMessage(node) {
                // Provide specialized messages to help the user understand why we think they're in
                // strict mode.
                if (ts.getContainingClass(node)) {
                    return ts.Diagnostics.Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Class_definitions_are_automatically_in_strict_mode;
                }
                if (file.externalModuleIndicator) {
                    return ts.Diagnostics.Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5_Modules_are_automatically_in_strict_mode;
                }
                return ts.Diagnostics.Function_declarations_are_not_allowed_inside_blocks_in_strict_mode_when_targeting_ES3_or_ES5;
            }
            function checkStrictModeFunctionDeclaration(node) {
                if (languageVersion < 2 /* ES2015 */) {
                    // Report error if function is not top level function declaration
                    if (blockScopeContainer.kind !== 300 /* SourceFile */ &&
                        blockScopeContainer.kind !== 259 /* ModuleDeclaration */ &&
                        !ts.isFunctionLikeOrClassStaticBlockDeclaration(blockScopeContainer)) {
                        // We check first if the name is inside class declaration or class expression; if so give explicit message
                        // otherwise report generic error message.
                        var errorSpan = ts.getErrorSpanForNode(file, node);
                        file.bindDiagnostics.push(ts.createFileDiagnostic(file, errorSpan.start, errorSpan.length, getStrictModeBlockScopeFunctionDeclarationMessage(node)));
                    }
                }
            }
            function checkStrictModeNumericLiteral(node) {
                if (inStrictMode && node.numericLiteralFlags & 32 /* Octal */) {
                    file.bindDiagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.Octal_literals_are_not_allowed_in_strict_mode));
                }
            }
            function checkStrictModePostfixUnaryExpression(node) {
                // Grammar checking
                // The identifier eval or arguments may not appear as the LeftHandSideExpression of an
                // Assignment operator(11.13) or of a PostfixExpression(11.3) or as the UnaryExpression
                // operated upon by a Prefix Increment(11.4.4) or a Prefix Decrement(11.4.5) operator.
                if (inStrictMode) {
                    checkStrictModeEvalOrArguments(node, node.operand);
                }
            }
            function checkStrictModePrefixUnaryExpression(node) {
                // Grammar checking
                if (inStrictMode) {
                    if (node.operator === 45 /* PlusPlusToken */ || node.operator === 46 /* MinusMinusToken */) {
                        checkStrictModeEvalOrArguments(node, node.operand);
                    }
                }
            }
            function checkStrictModeWithStatement(node) {
                // Grammar checking for withStatement
                if (inStrictMode) {
                    errorOnFirstToken(node, ts.Diagnostics.with_statements_are_not_allowed_in_strict_mode);
                }
            }
            function checkStrictModeLabeledStatement(node) {
                // Grammar checking for labeledStatement
                if (inStrictMode && options.target >= 2 /* ES2015 */) {
                    if (ts.isDeclarationStatement(node.statement) || ts.isVariableStatement(node.statement)) {
                        errorOnFirstToken(node.label, ts.Diagnostics.A_label_is_not_allowed_here);
                    }
                }
            }
            function errorOnFirstToken(node, message, arg0, arg1, arg2) {
                var span = ts.getSpanOfTokenAtPosition(file, node.pos);
                file.bindDiagnostics.push(ts.createFileDiagnostic(file, span.start, span.length, message, arg0, arg1, arg2));
            }
            function errorOrSuggestionOnNode(isError, node, message) {
                errorOrSuggestionOnRange(isError, node, node, message);
            }
            function errorOrSuggestionOnRange(isError, startNode, endNode, message) {
                addErrorOrSuggestionDiagnostic(isError, { pos: ts.getTokenPosOfNode(startNode, file), end: endNode.end }, message);
            }
            function addErrorOrSuggestionDiagnostic(isError, range, message) {
                var diag = ts.createFileDiagnostic(file, range.pos, range.end - range.pos, message);
                if (isError) {
                    file.bindDiagnostics.push(diag);
                }
                else {
                    file.bindSuggestionDiagnostics = ts.append(file.bindSuggestionDiagnostics, __assign(__assign({}, diag), { category: ts.DiagnosticCategory.Suggestion }));
                }
            }
            function bind(node) {
                if (!node) {
                    return;
                }
                ts.setParent(node, parent);
                var saveInStrictMode = inStrictMode;
                // Even though in the AST the jsdoc @typedef node belongs to the current node,
                // its symbol might be in the same scope with the current node's symbol. Consider:
                //
                //     /** @typedef {string | number} MyType */
                //     function foo();
                //
                // Here the current node is "foo", which is a container, but the scope of "MyType" should
                // not be inside "foo". Therefore we always bind @typedef before bind the parent node,
                // and skip binding this tag later when binding all the other jsdoc tags.
                // First we bind declaration nodes to a symbol if possible. We'll both create a symbol
                // and then potentially add the symbol to an appropriate symbol table. Possible
                // destination symbol tables are:
                //
                //  1) The 'exports' table of the current container's symbol.
                //  2) The 'members' table of the current container's symbol.
                //  3) The 'locals' table of the current container.
                //
                // However, not all symbols will end up in any of these tables. 'Anonymous' symbols
                // (like TypeLiterals for example) will not be put in any table.
                bindWorker(node);
                // Then we recurse into the children of the node to bind them as well. For certain
                // symbols we do specialized work when we recurse. For example, we'll keep track of
                // the current 'container' node when it changes. This helps us know which symbol table
                // a local should go into for example. Since terminal nodes are known not to have
                // children, as an optimization we don't process those.
                if (node.kind > 158 /* LastToken */) {
                    var saveParent = parent;
                    parent = node;
                    var containerFlags = getContainerFlags(node);
                    if (containerFlags === 0 /* None */) {
                        bindChildren(node);
                    }
                    else {
                        bindContainer(node, containerFlags);
                    }
                    parent = saveParent;
                }
                else {
                    var saveParent = parent;
                    if (node.kind === 1 /* EndOfFileToken */)
                        parent = node;
                    bindJSDoc(node);
                    parent = saveParent;
                }
                inStrictMode = saveInStrictMode;
            }
            function bindJSDoc(node) {
                if (ts.hasJSDocNodes(node)) {
                    if (ts.isInJSFile(node)) {
                        for (var _i = 0, _a = node.jsDoc; _i < _a.length; _i++) {
                            var j = _a[_i];
                            bind(j);
                        }
                    }
                    else {
                        for (var _b = 0, _c = node.jsDoc; _b < _c.length; _b++) {
                            var j = _c[_b];
                            ts.setParent(j, node);
                            ts.setParentRecursive(j, /*incremental*/ false);
                        }
                    }
                }
            }
            function updateStrictModeStatementList(statements) {
                if (!inStrictMode) {
                    for (var _i = 0, statements_3 = statements; _i < statements_3.length; _i++) {
                        var statement = statements_3[_i];
                        if (!ts.isPrologueDirective(statement)) {
                            return;
                        }
                        if (isUseStrictPrologueDirective(statement)) {
                            inStrictMode = true;
                            return;
                        }
                    }
                }
            }
            /// Should be called only on prologue directives (isPrologueDirective(node) should be true)
            function isUseStrictPrologueDirective(node) {
                var nodeText = ts.getSourceTextOfNodeFromSourceFile(file, node.expression);
                // Note: the node text must be exactly "use strict" or 'use strict'.  It is not ok for the
                // string to contain unicode escapes (as per ES5).
                return nodeText === '"use strict"' || nodeText === "'use strict'";
            }
            function bindWorker(node) {
                switch (node.kind) {
                    /* Strict mode checks */
                    case 79 /* Identifier */:
                        // for typedef type names with namespaces, bind the new jsdoc type symbol here
                        // because it requires all containing namespaces to be in effect, namely the
                        // current "blockScopeContainer" needs to be set to its immediate namespace parent.
                        if (node.isInJSDocNamespace) {
                            var parentNode = node.parent;
                            while (parentNode && !ts.isJSDocTypeAlias(parentNode)) {
                                parentNode = parentNode.parent;
                            }
                            bindBlockScopedDeclaration(parentNode, 524288 /* TypeAlias */, 788968 /* TypeAliasExcludes */);
                            break;
                        }
                    // falls through
                    case 108 /* ThisKeyword */:
                        if (currentFlow && (ts.isExpression(node) || parent.kind === 292 /* ShorthandPropertyAssignment */)) {
                            node.flowNode = currentFlow;
                        }
                        return checkContextualIdentifier(node);
                    case 159 /* QualifiedName */:
                        if (currentFlow && ts.isPartOfTypeQuery(node)) {
                            node.flowNode = currentFlow;
                        }
                        break;
                    case 229 /* MetaProperty */:
                    case 106 /* SuperKeyword */:
                        node.flowNode = currentFlow;
                        break;
                    case 80 /* PrivateIdentifier */:
                        return checkPrivateIdentifier(node);
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */:
                        var expr = node;
                        if (currentFlow && isNarrowableReference(expr)) {
                            expr.flowNode = currentFlow;
                        }
                        if (ts.isSpecialPropertyDeclaration(expr)) {
                            bindSpecialPropertyDeclaration(expr);
                        }
                        if (ts.isInJSFile(expr) &&
                            file.commonJsModuleIndicator &&
                            ts.isModuleExportsAccessExpression(expr) &&
                            !lookupSymbolForName(blockScopeContainer, "module")) {
                            declareSymbol(file.locals, /*parent*/ undefined, expr.expression, 1 /* FunctionScopedVariable */ | 134217728 /* ModuleExports */, 111550 /* FunctionScopedVariableExcludes */);
                        }
                        break;
                    case 219 /* BinaryExpression */:
                        var specialKind = ts.getAssignmentDeclarationKind(node);
                        switch (specialKind) {
                            case 1 /* ExportsProperty */:
                                bindExportsPropertyAssignment(node);
                                break;
                            case 2 /* ModuleExports */:
                                bindModuleExportsAssignment(node);
                                break;
                            case 3 /* PrototypeProperty */:
                                bindPrototypePropertyAssignment(node.left, node);
                                break;
                            case 6 /* Prototype */:
                                bindPrototypeAssignment(node);
                                break;
                            case 4 /* ThisProperty */:
                                bindThisPropertyAssignment(node);
                                break;
                            case 5 /* Property */:
                                var expression = node.left.expression;
                                if (ts.isInJSFile(node) && ts.isIdentifier(expression)) {
                                    var symbol = lookupSymbolForName(blockScopeContainer, expression.escapedText);
                                    if (ts.isThisInitializedDeclaration(symbol === null || symbol === void 0 ? void 0 : symbol.valueDeclaration)) {
                                        bindThisPropertyAssignment(node);
                                        break;
                                    }
                                }
                                bindSpecialPropertyAssignment(node);
                                break;
                            case 0 /* None */:
                                // Nothing to do
                                break;
                            default:
                                ts.Debug.fail("Unknown binary expression special property assignment kind");
                        }
                        return checkStrictModeBinaryExpression(node);
                    case 290 /* CatchClause */:
                        return checkStrictModeCatchClause(node);
                    case 213 /* DeleteExpression */:
                        return checkStrictModeDeleteExpression(node);
                    case 8 /* NumericLiteral */:
                        return checkStrictModeNumericLiteral(node);
                    case 218 /* PostfixUnaryExpression */:
                        return checkStrictModePostfixUnaryExpression(node);
                    case 217 /* PrefixUnaryExpression */:
                        return checkStrictModePrefixUnaryExpression(node);
                    case 246 /* WithStatement */:
                        return checkStrictModeWithStatement(node);
                    case 248 /* LabeledStatement */:
                        return checkStrictModeLabeledStatement(node);
                    case 190 /* ThisType */:
                        seenThisKeyword = true;
                        return;
                    case 175 /* TypePredicate */:
                        break; // Binding the children will handle everything
                    case 161 /* TypeParameter */:
                        return bindTypeParameter(node);
                    case 162 /* Parameter */:
                        return bindParameter(node);
                    case 252 /* VariableDeclaration */:
                        return bindVariableDeclarationOrBindingElement(node);
                    case 201 /* BindingElement */:
                        node.flowNode = currentFlow;
                        return bindVariableDeclarationOrBindingElement(node);
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                        return bindPropertyWorker(node);
                    case 291 /* PropertyAssignment */:
                    case 292 /* ShorthandPropertyAssignment */:
                        return bindPropertyOrMethodOrAccessor(node, 4 /* Property */, 0 /* PropertyExcludes */);
                    case 294 /* EnumMember */:
                        return bindPropertyOrMethodOrAccessor(node, 8 /* EnumMember */, 900095 /* EnumMemberExcludes */);
                    case 172 /* CallSignature */:
                    case 173 /* ConstructSignature */:
                    case 174 /* IndexSignature */:
                        return declareSymbolAndAddToSymbolTable(node, 131072 /* Signature */, 0 /* None */);
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                        // If this is an ObjectLiteralExpression method, then it sits in the same space
                        // as other properties in the object literal.  So we use SymbolFlags.PropertyExcludes
                        // so that it will conflict with any other object literal members with the same
                        // name.
                        return bindPropertyOrMethodOrAccessor(node, 8192 /* Method */ | (node.questionToken ? 16777216 /* Optional */ : 0 /* None */), ts.isObjectLiteralMethod(node) ? 0 /* PropertyExcludes */ : 103359 /* MethodExcludes */);
                    case 254 /* FunctionDeclaration */:
                        return bindFunctionDeclaration(node);
                    case 169 /* Constructor */:
                        return declareSymbolAndAddToSymbolTable(node, 16384 /* Constructor */, /*symbolExcludes:*/ 0 /* None */);
                    case 170 /* GetAccessor */:
                        return bindPropertyOrMethodOrAccessor(node, 32768 /* GetAccessor */, 46015 /* GetAccessorExcludes */);
                    case 171 /* SetAccessor */:
                        return bindPropertyOrMethodOrAccessor(node, 65536 /* SetAccessor */, 78783 /* SetAccessorExcludes */);
                    case 177 /* FunctionType */:
                    case 312 /* JSDocFunctionType */:
                    case 318 /* JSDocSignature */:
                    case 178 /* ConstructorType */:
                        return bindFunctionOrConstructorType(node);
                    case 180 /* TypeLiteral */:
                    case 317 /* JSDocTypeLiteral */:
                    case 193 /* MappedType */:
                        return bindAnonymousTypeWorker(node);
                    case 327 /* JSDocClassTag */:
                        return bindJSDocClassTag(node);
                    case 203 /* ObjectLiteralExpression */:
                        return bindObjectLiteralExpression(node);
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        return bindFunctionExpression(node);
                    case 206 /* CallExpression */:
                        var assignmentKind = ts.getAssignmentDeclarationKind(node);
                        switch (assignmentKind) {
                            case 7 /* ObjectDefinePropertyValue */:
                                return bindObjectDefinePropertyAssignment(node);
                            case 8 /* ObjectDefinePropertyExports */:
                                return bindObjectDefinePropertyExport(node);
                            case 9 /* ObjectDefinePrototypeProperty */:
                                return bindObjectDefinePrototypeProperty(node);
                            case 0 /* None */:
                                break; // Nothing to do
                            default:
                                return ts.Debug.fail("Unknown call expression assignment declaration kind");
                        }
                        if (ts.isInJSFile(node)) {
                            bindCallExpression(node);
                        }
                        break;
                    // Members of classes, interfaces, and modules
                    case 224 /* ClassExpression */:
                    case 255 /* ClassDeclaration */:
                        // All classes are automatically in strict mode in ES6.
                        inStrictMode = true;
                        return bindClassLikeDeclaration(node);
                    case 256 /* InterfaceDeclaration */:
                        return bindBlockScopedDeclaration(node, 64 /* Interface */, 788872 /* InterfaceExcludes */);
                    case 257 /* TypeAliasDeclaration */:
                        return bindBlockScopedDeclaration(node, 524288 /* TypeAlias */, 788968 /* TypeAliasExcludes */);
                    case 258 /* EnumDeclaration */:
                        return bindEnumDeclaration(node);
                    case 259 /* ModuleDeclaration */:
                        return bindModuleDeclaration(node);
                    // Jsx-attributes
                    case 284 /* JsxAttributes */:
                        return bindJsxAttributes(node);
                    case 283 /* JsxAttribute */:
                        return bindJsxAttribute(node, 4 /* Property */, 0 /* PropertyExcludes */);
                    // Imports and exports
                    case 263 /* ImportEqualsDeclaration */:
                    case 266 /* NamespaceImport */:
                    case 268 /* ImportSpecifier */:
                    case 273 /* ExportSpecifier */:
                        return declareSymbolAndAddToSymbolTable(node, 2097152 /* Alias */, 2097152 /* AliasExcludes */);
                    case 262 /* NamespaceExportDeclaration */:
                        return bindNamespaceExportDeclaration(node);
                    case 265 /* ImportClause */:
                        return bindImportClause(node);
                    case 270 /* ExportDeclaration */:
                        return bindExportDeclaration(node);
                    case 269 /* ExportAssignment */:
                        return bindExportAssignment(node);
                    case 300 /* SourceFile */:
                        updateStrictModeStatementList(node.statements);
                        return bindSourceFileIfExternalModule();
                    case 233 /* Block */:
                        if (!ts.isFunctionLikeOrClassStaticBlockDeclaration(node.parent)) {
                            return;
                        }
                    // falls through
                    case 260 /* ModuleBlock */:
                        return updateStrictModeStatementList(node.statements);
                    case 335 /* JSDocParameterTag */:
                        if (node.parent.kind === 318 /* JSDocSignature */) {
                            return bindParameter(node);
                        }
                        if (node.parent.kind !== 317 /* JSDocTypeLiteral */) {
                            break;
                        }
                    // falls through
                    case 342 /* JSDocPropertyTag */:
                        var propTag = node;
                        var flags = propTag.isBracketed || propTag.typeExpression && propTag.typeExpression.type.kind === 311 /* JSDocOptionalType */ ?
                            4 /* Property */ | 16777216 /* Optional */ :
                            4 /* Property */;
                        return declareSymbolAndAddToSymbolTable(propTag, flags, 0 /* PropertyExcludes */);
                    case 340 /* JSDocTypedefTag */:
                    case 333 /* JSDocCallbackTag */:
                    case 334 /* JSDocEnumTag */:
                        return (delayedTypeAliases || (delayedTypeAliases = [])).push(node);
                }
            }
            function bindPropertyWorker(node) {
                return bindPropertyOrMethodOrAccessor(node, 4 /* Property */ | (node.questionToken ? 16777216 /* Optional */ : 0 /* None */), 0 /* PropertyExcludes */);
            }
            function bindAnonymousTypeWorker(node) {
                return bindAnonymousDeclaration(node, 2048 /* TypeLiteral */, "__type" /* Type */);
            }
            function bindSourceFileIfExternalModule() {
                setExportContextFlag(file);
                if (ts.isExternalModule(file)) {
                    bindSourceFileAsExternalModule();
                }
                else if (ts.isJsonSourceFile(file)) {
                    bindSourceFileAsExternalModule();
                    // Create symbol equivalent for the module.exports = {}
                    var originalSymbol = file.symbol;
                    declareSymbol(file.symbol.exports, file.symbol, file, 4 /* Property */, 67108863 /* All */);
                    file.symbol = originalSymbol;
                }
            }
            function bindSourceFileAsExternalModule() {
                bindAnonymousDeclaration(file, 512 /* ValueModule */, "\"" + ts.removeFileExtension(file.fileName) + "\"");
            }
            function bindExportAssignment(node) {
                if (!container.symbol || !container.symbol.exports) {
                    // Incorrect export assignment in some sort of block construct
                    bindAnonymousDeclaration(node, 111551 /* Value */, getDeclarationName(node));
                }
                else {
                    var flags = ts.exportAssignmentIsAlias(node)
                        // An export default clause with an EntityNameExpression or a class expression exports all meanings of that identifier or expression;
                        ? 2097152 /* Alias */
                        // An export default clause with any other expression exports a value
                        : 4 /* Property */;
                    // If there is an `export default x;` alias declaration, can't `export default` anything else.
                    // (In contrast, you can still have `export default function f() {}` and `export default interface I {}`.)
                    var symbol = declareSymbol(container.symbol.exports, container.symbol, node, flags, 67108863 /* All */);
                    if (node.isExportEquals) {
                        // Will be an error later, since the module already has other exports. Just make sure this has a valueDeclaration set.
                        ts.setValueDeclaration(symbol, node);
                    }
                }
            }
            function bindNamespaceExportDeclaration(node) {
                if (node.modifiers && node.modifiers.length) {
                    file.bindDiagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.Modifiers_cannot_appear_here));
                }
                var diag = !ts.isSourceFile(node.parent) ? ts.Diagnostics.Global_module_exports_may_only_appear_at_top_level
                    : !ts.isExternalModule(node.parent) ? ts.Diagnostics.Global_module_exports_may_only_appear_in_module_files
                        : !node.parent.isDeclarationFile ? ts.Diagnostics.Global_module_exports_may_only_appear_in_declaration_files
                            : undefined;
                if (diag) {
                    file.bindDiagnostics.push(createDiagnosticForNode(node, diag));
                }
                else {
                    file.symbol.globalExports = file.symbol.globalExports || ts.createSymbolTable();
                    declareSymbol(file.symbol.globalExports, file.symbol, node, 2097152 /* Alias */, 2097152 /* AliasExcludes */);
                }
            }
            function bindExportDeclaration(node) {
                if (!container.symbol || !container.symbol.exports) {
                    // Export * in some sort of block construct
                    bindAnonymousDeclaration(node, 8388608 /* ExportStar */, getDeclarationName(node));
                }
                else if (!node.exportClause) {
                    // All export * declarations are collected in an __export symbol
                    declareSymbol(container.symbol.exports, container.symbol, node, 8388608 /* ExportStar */, 0 /* None */);
                }
                else if (ts.isNamespaceExport(node.exportClause)) {
                    // declareSymbol walks up parents to find name text, parent _must_ be set
                    // but won't be set by the normal binder walk until `bindChildren` later on.
                    ts.setParent(node.exportClause, node);
                    declareSymbol(container.symbol.exports, container.symbol, node.exportClause, 2097152 /* Alias */, 2097152 /* AliasExcludes */);
                }
            }
            function bindImportClause(node) {
                if (node.name) {
                    declareSymbolAndAddToSymbolTable(node, 2097152 /* Alias */, 2097152 /* AliasExcludes */);
                }
            }
            function setCommonJsModuleIndicator(node) {
                if (file.externalModuleIndicator) {
                    return false;
                }
                if (!file.commonJsModuleIndicator) {
                    file.commonJsModuleIndicator = node;
                    bindSourceFileAsExternalModule();
                }
                return true;
            }
            function bindObjectDefinePropertyExport(node) {
                if (!setCommonJsModuleIndicator(node)) {
                    return;
                }
                var symbol = forEachIdentifierInEntityName(node.arguments[0], /*parent*/ undefined, function (id, symbol) {
                    if (symbol) {
                        addDeclarationToSymbol(symbol, id, 1536 /* Module */ | 67108864 /* Assignment */);
                    }
                    return symbol;
                });
                if (symbol) {
                    var flags = 4 /* Property */ | 1048576 /* ExportValue */;
                    declareSymbol(symbol.exports, symbol, node, flags, 0 /* None */);
                }
            }
            function bindExportsPropertyAssignment(node) {
                // When we create a property via 'exports.foo = bar', the 'exports.foo' property access
                // expression is the declaration
                if (!setCommonJsModuleIndicator(node)) {
                    return;
                }
                var symbol = forEachIdentifierInEntityName(node.left.expression, /*parent*/ undefined, function (id, symbol) {
                    if (symbol) {
                        addDeclarationToSymbol(symbol, id, 1536 /* Module */ | 67108864 /* Assignment */);
                    }
                    return symbol;
                });
                if (symbol) {
                    var isAlias = ts.isAliasableExpression(node.right) && (ts.isExportsIdentifier(node.left.expression) || ts.isModuleExportsAccessExpression(node.left.expression));
                    var flags = isAlias ? 2097152 /* Alias */ : 4 /* Property */ | 1048576 /* ExportValue */;
                    ts.setParent(node.left, node);
                    declareSymbol(symbol.exports, symbol, node.left, flags, 0 /* None */);
                }
            }
            function bindModuleExportsAssignment(node) {
                // A common practice in node modules is to set 'export = module.exports = {}', this ensures that 'exports'
                // is still pointing to 'module.exports'.
                // We do not want to consider this as 'export=' since a module can have only one of these.
                // Similarly we do not want to treat 'module.exports = exports' as an 'export='.
                if (!setCommonJsModuleIndicator(node)) {
                    return;
                }
                var assignedExpression = ts.getRightMostAssignedExpression(node.right);
                if (ts.isEmptyObjectLiteral(assignedExpression) || container === file && isExportsOrModuleExportsOrAlias(file, assignedExpression)) {
                    return;
                }
                if (ts.isObjectLiteralExpression(assignedExpression) && ts.every(assignedExpression.properties, ts.isShorthandPropertyAssignment)) {
                    ts.forEach(assignedExpression.properties, bindExportAssignedObjectMemberAlias);
                    return;
                }
                // 'module.exports = expr' assignment
                var flags = ts.exportAssignmentIsAlias(node)
                    ? 2097152 /* Alias */
                    : 4 /* Property */ | 1048576 /* ExportValue */ | 512 /* ValueModule */;
                var symbol = declareSymbol(file.symbol.exports, file.symbol, node, flags | 67108864 /* Assignment */, 0 /* None */);
                ts.setValueDeclaration(symbol, node);
            }
            function bindExportAssignedObjectMemberAlias(node) {
                declareSymbol(file.symbol.exports, file.symbol, node, 2097152 /* Alias */ | 67108864 /* Assignment */, 0 /* None */);
            }
            function bindThisPropertyAssignment(node) {
                ts.Debug.assert(ts.isInJSFile(node));
                // private identifiers *must* be declared (even in JS files)
                var hasPrivateIdentifier = (ts.isBinaryExpression(node) && ts.isPropertyAccessExpression(node.left) && ts.isPrivateIdentifier(node.left.name))
                    || (ts.isPropertyAccessExpression(node) && ts.isPrivateIdentifier(node.name));
                if (hasPrivateIdentifier) {
                    return;
                }
                var thisContainer = ts.getThisContainer(node, /*includeArrowFunctions*/ false);
                switch (thisContainer.kind) {
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                        var constructorSymbol = thisContainer.symbol;
                        // For `f.prototype.m = function() { this.x = 0; }`, `this.x = 0` should modify `f`'s members, not the function expression.
                        if (ts.isBinaryExpression(thisContainer.parent) && thisContainer.parent.operatorToken.kind === 63 /* EqualsToken */) {
                            var l = thisContainer.parent.left;
                            if (ts.isBindableStaticAccessExpression(l) && ts.isPrototypeAccess(l.expression)) {
                                constructorSymbol = lookupSymbolForPropertyAccess(l.expression.expression, thisParentContainer);
                            }
                        }
                        if (constructorSymbol && constructorSymbol.valueDeclaration) {
                            // Declare a 'member' if the container is an ES5 class or ES6 constructor
                            constructorSymbol.members = constructorSymbol.members || ts.createSymbolTable();
                            // It's acceptable for multiple 'this' assignments of the same identifier to occur
                            if (ts.hasDynamicName(node)) {
                                bindDynamicallyNamedThisPropertyAssignment(node, constructorSymbol, constructorSymbol.members);
                            }
                            else {
                                declareSymbol(constructorSymbol.members, constructorSymbol, node, 4 /* Property */ | 67108864 /* Assignment */, 0 /* PropertyExcludes */ & ~4 /* Property */);
                            }
                            addDeclarationToSymbol(constructorSymbol, constructorSymbol.valueDeclaration, 32 /* Class */);
                        }
                        break;
                    case 169 /* Constructor */:
                    case 165 /* PropertyDeclaration */:
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        // this.foo assignment in a JavaScript class
                        // Bind this property to the containing class
                        var containingClass = thisContainer.parent;
                        var symbolTable = ts.isStatic(thisContainer) ? containingClass.symbol.exports : containingClass.symbol.members;
                        if (ts.hasDynamicName(node)) {
                            bindDynamicallyNamedThisPropertyAssignment(node, containingClass.symbol, symbolTable);
                        }
                        else {
                            declareSymbol(symbolTable, containingClass.symbol, node, 4 /* Property */ | 67108864 /* Assignment */, 0 /* None */, /*isReplaceableByMethod*/ true);
                        }
                        break;
                    case 300 /* SourceFile */:
                        // this.property = assignment in a source file -- declare symbol in exports for a module, in locals for a script
                        if (ts.hasDynamicName(node)) {
                            break;
                        }
                        else if (thisContainer.commonJsModuleIndicator) {
                            declareSymbol(thisContainer.symbol.exports, thisContainer.symbol, node, 4 /* Property */ | 1048576 /* ExportValue */, 0 /* None */);
                        }
                        else {
                            declareSymbolAndAddToSymbolTable(node, 1 /* FunctionScopedVariable */, 111550 /* FunctionScopedVariableExcludes */);
                        }
                        break;
                    default:
                        ts.Debug.failBadSyntaxKind(thisContainer);
                }
            }
            function bindDynamicallyNamedThisPropertyAssignment(node, symbol, symbolTable) {
                declareSymbol(symbolTable, symbol, node, 4 /* Property */, 0 /* None */, /*isReplaceableByMethod*/ true, /*isComputedName*/ true);
                addLateBoundAssignmentDeclarationToSymbol(node, symbol);
            }
            function addLateBoundAssignmentDeclarationToSymbol(node, symbol) {
                if (symbol) {
                    (symbol.assignmentDeclarationMembers || (symbol.assignmentDeclarationMembers = new ts.Map())).set(ts.getNodeId(node), node);
                }
            }
            function bindSpecialPropertyDeclaration(node) {
                if (node.expression.kind === 108 /* ThisKeyword */) {
                    bindThisPropertyAssignment(node);
                }
                else if (ts.isBindableStaticAccessExpression(node) && node.parent.parent.kind === 300 /* SourceFile */) {
                    if (ts.isPrototypeAccess(node.expression)) {
                        bindPrototypePropertyAssignment(node, node.parent);
                    }
                    else {
                        bindStaticPropertyAssignment(node);
                    }
                }
            }
            /** For `x.prototype = { p, ... }`, declare members p,... if `x` is function/class/{}, or not declared. */
            function bindPrototypeAssignment(node) {
                ts.setParent(node.left, node);
                ts.setParent(node.right, node);
                bindPropertyAssignment(node.left.expression, node.left, /*isPrototypeProperty*/ false, /*containerIsClass*/ true);
            }
            function bindObjectDefinePrototypeProperty(node) {
                var namespaceSymbol = lookupSymbolForPropertyAccess(node.arguments[0].expression);
                if (namespaceSymbol && namespaceSymbol.valueDeclaration) {
                    // Ensure the namespace symbol becomes class-like
                    addDeclarationToSymbol(namespaceSymbol, namespaceSymbol.valueDeclaration, 32 /* Class */);
                }
                bindPotentiallyNewExpandoMemberToNamespace(node, namespaceSymbol, /*isPrototypeProperty*/ true);
            }
            /**
             * For `x.prototype.y = z`, declare a member `y` on `x` if `x` is a function or class, or not declared.
             * Note that jsdoc preceding an ExpressionStatement like `x.prototype.y;` is also treated as a declaration.
             */
            function bindPrototypePropertyAssignment(lhs, parent) {
                // Look up the function in the local scope, since prototype assignments should
                // follow the function declaration
                var classPrototype = lhs.expression;
                var constructorFunction = classPrototype.expression;
                // Fix up parent pointers since we're going to use these nodes before we bind into them
                ts.setParent(constructorFunction, classPrototype);
                ts.setParent(classPrototype, lhs);
                ts.setParent(lhs, parent);
                bindPropertyAssignment(constructorFunction, lhs, /*isPrototypeProperty*/ true, /*containerIsClass*/ true);
            }
            function bindObjectDefinePropertyAssignment(node) {
                var namespaceSymbol = lookupSymbolForPropertyAccess(node.arguments[0]);
                var isToplevel = node.parent.parent.kind === 300 /* SourceFile */;
                namespaceSymbol = bindPotentiallyMissingNamespaces(namespaceSymbol, node.arguments[0], isToplevel, /*isPrototypeProperty*/ false, /*containerIsClass*/ false);
                bindPotentiallyNewExpandoMemberToNamespace(node, namespaceSymbol, /*isPrototypeProperty*/ false);
            }
            function bindSpecialPropertyAssignment(node) {
                var _a;
                // Class declarations in Typescript do not allow property declarations
                var parentSymbol = lookupSymbolForPropertyAccess(node.left.expression, container) || lookupSymbolForPropertyAccess(node.left.expression, blockScopeContainer);
                if (!ts.isInJSFile(node) && !ts.isFunctionSymbol(parentSymbol)) {
                    return;
                }
                var rootExpr = ts.getLeftmostAccessExpression(node.left);
                if (ts.isIdentifier(rootExpr) && ((_a = lookupSymbolForName(container, rootExpr.escapedText)) === null || _a === void 0 ? void 0 : _a.flags) & 2097152 /* Alias */) {
                    return;
                }
                // Fix up parent pointers since we're going to use these nodes before we bind into them
                ts.setParent(node.left, node);
                ts.setParent(node.right, node);
                if (ts.isIdentifier(node.left.expression) && container === file && isExportsOrModuleExportsOrAlias(file, node.left.expression)) {
                    // This can be an alias for the 'exports' or 'module.exports' names, e.g.
                    //    var util = module.exports;
                    //    util.property = function ...
                    bindExportsPropertyAssignment(node);
                }
                else if (ts.hasDynamicName(node)) {
                    bindAnonymousDeclaration(node, 4 /* Property */ | 67108864 /* Assignment */, "__computed" /* Computed */);
                    var sym = bindPotentiallyMissingNamespaces(parentSymbol, node.left.expression, isTopLevelNamespaceAssignment(node.left), /*isPrototype*/ false, /*containerIsClass*/ false);
                    addLateBoundAssignmentDeclarationToSymbol(node, sym);
                }
                else {
                    bindStaticPropertyAssignment(ts.cast(node.left, ts.isBindableStaticNameExpression));
                }
            }
            /**
             * For nodes like `x.y = z`, declare a member 'y' on 'x' if x is a function (or IIFE) or class or {}, or not declared.
             * Also works for expression statements preceded by JSDoc, like / ** @type number * / x.y;
             */
            function bindStaticPropertyAssignment(node) {
                ts.Debug.assert(!ts.isIdentifier(node));
                ts.setParent(node.expression, node);
                bindPropertyAssignment(node.expression, node, /*isPrototypeProperty*/ false, /*containerIsClass*/ false);
            }
            function bindPotentiallyMissingNamespaces(namespaceSymbol, entityName, isToplevel, isPrototypeProperty, containerIsClass) {
                if ((namespaceSymbol === null || namespaceSymbol === void 0 ? void 0 : namespaceSymbol.flags) & 2097152 /* Alias */) {
                    return namespaceSymbol;
                }
                if (isToplevel && !isPrototypeProperty) {
                    // make symbols or add declarations for intermediate containers
                    var flags_2 = 1536 /* Module */ | 67108864 /* Assignment */;
                    var excludeFlags_1 = 110735 /* ValueModuleExcludes */ & ~67108864 /* Assignment */;
                    namespaceSymbol = forEachIdentifierInEntityName(entityName, namespaceSymbol, function (id, symbol, parent) {
                        if (symbol) {
                            addDeclarationToSymbol(symbol, id, flags_2);
                            return symbol;
                        }
                        else {
                            var table = parent ? parent.exports :
                                file.jsGlobalAugmentations || (file.jsGlobalAugmentations = ts.createSymbolTable());
                            return declareSymbol(table, parent, id, flags_2, excludeFlags_1);
                        }
                    });
                }
                if (containerIsClass && namespaceSymbol && namespaceSymbol.valueDeclaration) {
                    addDeclarationToSymbol(namespaceSymbol, namespaceSymbol.valueDeclaration, 32 /* Class */);
                }
                return namespaceSymbol;
            }
            function bindPotentiallyNewExpandoMemberToNamespace(declaration, namespaceSymbol, isPrototypeProperty) {
                if (!namespaceSymbol || !isExpandoSymbol(namespaceSymbol)) {
                    return;
                }
                // Set up the members collection if it doesn't exist already
                var symbolTable = isPrototypeProperty ?
                    (namespaceSymbol.members || (namespaceSymbol.members = ts.createSymbolTable())) :
                    (namespaceSymbol.exports || (namespaceSymbol.exports = ts.createSymbolTable()));
                var includes = 0 /* None */;
                var excludes = 0 /* None */;
                // Method-like
                if (ts.isFunctionLikeDeclaration(ts.getAssignedExpandoInitializer(declaration))) {
                    includes = 8192 /* Method */;
                    excludes = 103359 /* MethodExcludes */;
                }
                // Maybe accessor-like
                else if (ts.isCallExpression(declaration) && ts.isBindableObjectDefinePropertyCall(declaration)) {
                    if (ts.some(declaration.arguments[2].properties, function (p) {
                        var id = ts.getNameOfDeclaration(p);
                        return !!id && ts.isIdentifier(id) && ts.idText(id) === "set";
                    })) {
                        // We mix in `SymbolFLags.Property` so in the checker `getTypeOfVariableParameterOrProperty` is used for this
                        // symbol, instead of `getTypeOfAccessor` (which will assert as there is no real accessor declaration)
                        includes |= 65536 /* SetAccessor */ | 4 /* Property */;
                        excludes |= 78783 /* SetAccessorExcludes */;
                    }
                    if (ts.some(declaration.arguments[2].properties, function (p) {
                        var id = ts.getNameOfDeclaration(p);
                        return !!id && ts.isIdentifier(id) && ts.idText(id) === "get";
                    })) {
                        includes |= 32768 /* GetAccessor */ | 4 /* Property */;
                        excludes |= 46015 /* GetAccessorExcludes */;
                    }
                }
                if (includes === 0 /* None */) {
                    includes = 4 /* Property */;
                    excludes = 0 /* PropertyExcludes */;
                }
                declareSymbol(symbolTable, namespaceSymbol, declaration, includes | 67108864 /* Assignment */, excludes & ~67108864 /* Assignment */);
            }
            function isTopLevelNamespaceAssignment(propertyAccess) {
                return ts.isBinaryExpression(propertyAccess.parent)
                    ? getParentOfBinaryExpression(propertyAccess.parent).parent.kind === 300 /* SourceFile */
                    : propertyAccess.parent.parent.kind === 300 /* SourceFile */;
            }
            function bindPropertyAssignment(name, propertyAccess, isPrototypeProperty, containerIsClass) {
                var namespaceSymbol = lookupSymbolForPropertyAccess(name, container) || lookupSymbolForPropertyAccess(name, blockScopeContainer);
                var isToplevel = isTopLevelNamespaceAssignment(propertyAccess);
                namespaceSymbol = bindPotentiallyMissingNamespaces(namespaceSymbol, propertyAccess.expression, isToplevel, isPrototypeProperty, containerIsClass);
                bindPotentiallyNewExpandoMemberToNamespace(propertyAccess, namespaceSymbol, isPrototypeProperty);
            }
            /**
             * Javascript expando values are:
             * - Functions
             * - classes
             * - namespaces
             * - variables initialized with function expressions
             * -                       with class expressions
             * -                       with empty object literals
             * -                       with non-empty object literals if assigned to the prototype property
             */
            function isExpandoSymbol(symbol) {
                if (symbol.flags & (16 /* Function */ | 32 /* Class */ | 1024 /* NamespaceModule */)) {
                    return true;
                }
                var node = symbol.valueDeclaration;
                if (node && ts.isCallExpression(node)) {
                    return !!ts.getAssignedExpandoInitializer(node);
                }
                var init = !node ? undefined :
                    ts.isVariableDeclaration(node) ? node.initializer :
                        ts.isBinaryExpression(node) ? node.right :
                            ts.isPropertyAccessExpression(node) && ts.isBinaryExpression(node.parent) ? node.parent.right :
                                undefined;
                init = init && ts.getRightMostAssignedExpression(init);
                if (init) {
                    var isPrototypeAssignment = ts.isPrototypeAccess(ts.isVariableDeclaration(node) ? node.name : ts.isBinaryExpression(node) ? node.left : node);
                    return !!ts.getExpandoInitializer(ts.isBinaryExpression(init) && (init.operatorToken.kind === 56 /* BarBarToken */ || init.operatorToken.kind === 60 /* QuestionQuestionToken */) ? init.right : init, isPrototypeAssignment);
                }
                return false;
            }
            function getParentOfBinaryExpression(expr) {
                while (ts.isBinaryExpression(expr.parent)) {
                    expr = expr.parent;
                }
                return expr.parent;
            }
            function lookupSymbolForPropertyAccess(node, lookupContainer) {
                if (lookupContainer === void 0) {
                    lookupContainer = container;
                }
                if (ts.isIdentifier(node)) {
                    return lookupSymbolForName(lookupContainer, node.escapedText);
                }
                else {
                    var symbol = lookupSymbolForPropertyAccess(node.expression);
                    return symbol && symbol.exports && symbol.exports.get(ts.getElementOrPropertyAccessName(node));
                }
            }
            function forEachIdentifierInEntityName(e, parent, action) {
                if (isExportsOrModuleExportsOrAlias(file, e)) {
                    return file.symbol;
                }
                else if (ts.isIdentifier(e)) {
                    return action(e, lookupSymbolForPropertyAccess(e), parent);
                }
                else {
                    var s = forEachIdentifierInEntityName(e.expression, parent, action);
                    var name = ts.getNameOrArgument(e);
                    // unreachable
                    if (ts.isPrivateIdentifier(name)) {
                        ts.Debug.fail("unexpected PrivateIdentifier");
                    }
                    return action(name, s && s.exports && s.exports.get(ts.getElementOrPropertyAccessName(e)), s);
                }
            }
            function bindCallExpression(node) {
                // We're only inspecting call expressions to detect CommonJS modules, so we can skip
                // this check if we've already seen the module indicator
                if (!file.commonJsModuleIndicator && ts.isRequireCall(node, /*checkArgumentIsStringLiteralLike*/ false)) {
                    setCommonJsModuleIndicator(node);
                }
            }
            function bindClassLikeDeclaration(node) {
                if (node.kind === 255 /* ClassDeclaration */) {
                    bindBlockScopedDeclaration(node, 32 /* Class */, 899503 /* ClassExcludes */);
                }
                else {
                    var bindingName = node.name ? node.name.escapedText : "__class" /* Class */;
                    bindAnonymousDeclaration(node, 32 /* Class */, bindingName);
                    // Add name of class expression into the map for semantic classifier
                    if (node.name) {
                        classifiableNames.add(node.name.escapedText);
                    }
                }
                var symbol = node.symbol;
                // TypeScript 1.0 spec (April 2014): 8.4
                // Every class automatically contains a static property member named 'prototype', the
                // type of which is an instantiation of the class type with type Any supplied as a type
                // argument for each type parameter. It is an error to explicitly declare a static
                // property member with the name 'prototype'.
                //
                // Note: we check for this here because this class may be merging into a module.  The
                // module might have an exported variable called 'prototype'.  We can't allow that as
                // that would clash with the built-in 'prototype' for the class.
                var prototypeSymbol = createSymbol(4 /* Property */ | 4194304 /* Prototype */, "prototype");
                var symbolExport = symbol.exports.get(prototypeSymbol.escapedName);
                if (symbolExport) {
                    if (node.name) {
                        ts.setParent(node.name, node);
                    }
                    file.bindDiagnostics.push(createDiagnosticForNode(symbolExport.declarations[0], ts.Diagnostics.Duplicate_identifier_0, ts.symbolName(prototypeSymbol)));
                }
                symbol.exports.set(prototypeSymbol.escapedName, prototypeSymbol);
                prototypeSymbol.parent = symbol;
            }
            function bindEnumDeclaration(node) {
                return ts.isEnumConst(node)
                    ? bindBlockScopedDeclaration(node, 128 /* ConstEnum */, 899967 /* ConstEnumExcludes */)
                    : bindBlockScopedDeclaration(node, 256 /* RegularEnum */, 899327 /* RegularEnumExcludes */);
            }
            function bindVariableDeclarationOrBindingElement(node) {
                if (inStrictMode) {
                    checkStrictModeEvalOrArguments(node, node.name);
                }
                if (!ts.isBindingPattern(node.name)) {
                    if (ts.isInJSFile(node) && ts.isRequireVariableDeclaration(node) && !ts.getJSDocTypeTag(node)) {
                        declareSymbolAndAddToSymbolTable(node, 2097152 /* Alias */, 2097152 /* AliasExcludes */);
                    }
                    else if (ts.isBlockOrCatchScoped(node)) {
                        bindBlockScopedDeclaration(node, 2 /* BlockScopedVariable */, 111551 /* BlockScopedVariableExcludes */);
                    }
                    else if (ts.isParameterDeclaration(node)) {
                        // It is safe to walk up parent chain to find whether the node is a destructuring parameter declaration
                        // because its parent chain has already been set up, since parents are set before descending into children.
                        //
                        // If node is a binding element in parameter declaration, we need to use ParameterExcludes.
                        // Using ParameterExcludes flag allows the compiler to report an error on duplicate identifiers in Parameter Declaration
                        // For example:
                        //      function foo([a,a]) {} // Duplicate Identifier error
                        //      function bar(a,a) {}   // Duplicate Identifier error, parameter declaration in this case is handled in bindParameter
                        //                             // which correctly set excluded symbols
                        declareSymbolAndAddToSymbolTable(node, 1 /* FunctionScopedVariable */, 111551 /* ParameterExcludes */);
                    }
                    else {
                        declareSymbolAndAddToSymbolTable(node, 1 /* FunctionScopedVariable */, 111550 /* FunctionScopedVariableExcludes */);
                    }
                }
            }
            function bindParameter(node) {
                if (node.kind === 335 /* JSDocParameterTag */ && container.kind !== 318 /* JSDocSignature */) {
                    return;
                }
                if (inStrictMode && !(node.flags & 8388608 /* Ambient */)) {
                    // It is a SyntaxError if the identifier eval or arguments appears within a FormalParameterList of a
                    // strict mode FunctionLikeDeclaration or FunctionExpression(13.1)
                    checkStrictModeEvalOrArguments(node, node.name);
                }
                if (ts.isBindingPattern(node.name)) {
                    bindAnonymousDeclaration(node, 1 /* FunctionScopedVariable */, "__" + node.parent.parameters.indexOf(node));
                }
                else {
                    declareSymbolAndAddToSymbolTable(node, 1 /* FunctionScopedVariable */, 111551 /* ParameterExcludes */);
                }
                // If this is a property-parameter, then also declare the property symbol into the
                // containing class.
                if (ts.isParameterPropertyDeclaration(node, node.parent)) {
                    var classDeclaration = node.parent.parent;
                    declareSymbol(classDeclaration.symbol.members, classDeclaration.symbol, node, 4 /* Property */ | (node.questionToken ? 16777216 /* Optional */ : 0 /* None */), 0 /* PropertyExcludes */);
                }
            }
            function bindFunctionDeclaration(node) {
                if (!file.isDeclarationFile && !(node.flags & 8388608 /* Ambient */)) {
                    if (ts.isAsyncFunction(node)) {
                        emitFlags |= 2048 /* HasAsyncFunctions */;
                    }
                }
                checkStrictModeFunctionName(node);
                if (inStrictMode) {
                    checkStrictModeFunctionDeclaration(node);
                    bindBlockScopedDeclaration(node, 16 /* Function */, 110991 /* FunctionExcludes */);
                }
                else {
                    declareSymbolAndAddToSymbolTable(node, 16 /* Function */, 110991 /* FunctionExcludes */);
                }
            }
            function bindFunctionExpression(node) {
                if (!file.isDeclarationFile && !(node.flags & 8388608 /* Ambient */)) {
                    if (ts.isAsyncFunction(node)) {
                        emitFlags |= 2048 /* HasAsyncFunctions */;
                    }
                }
                if (currentFlow) {
                    node.flowNode = currentFlow;
                }
                checkStrictModeFunctionName(node);
                var bindingName = node.name ? node.name.escapedText : "__function" /* Function */;
                return bindAnonymousDeclaration(node, 16 /* Function */, bindingName);
            }
            function bindPropertyOrMethodOrAccessor(node, symbolFlags, symbolExcludes) {
                if (!file.isDeclarationFile && !(node.flags & 8388608 /* Ambient */) && ts.isAsyncFunction(node)) {
                    emitFlags |= 2048 /* HasAsyncFunctions */;
                }
                if (currentFlow && ts.isObjectLiteralOrClassExpressionMethod(node)) {
                    node.flowNode = currentFlow;
                }
                return ts.hasDynamicName(node)
                    ? bindAnonymousDeclaration(node, symbolFlags, "__computed" /* Computed */)
                    : declareSymbolAndAddToSymbolTable(node, symbolFlags, symbolExcludes);
            }
            function getInferTypeContainer(node) {
                var extendsType = ts.findAncestor(node, function (n) { return n.parent && ts.isConditionalTypeNode(n.parent) && n.parent.extendsType === n; });
                return extendsType && extendsType.parent;
            }
            function bindTypeParameter(node) {
                if (ts.isJSDocTemplateTag(node.parent)) {
                    var container_1 = ts.find(node.parent.parent.tags, ts.isJSDocTypeAlias) || ts.getHostSignatureFromJSDoc(node.parent); // TODO: GH#18217
                    if (container_1) {
                        if (!container_1.locals) {
                            container_1.locals = ts.createSymbolTable();
                        }
                        declareSymbol(container_1.locals, /*parent*/ undefined, node, 262144 /* TypeParameter */, 526824 /* TypeParameterExcludes */);
                    }
                    else {
                        declareSymbolAndAddToSymbolTable(node, 262144 /* TypeParameter */, 526824 /* TypeParameterExcludes */);
                    }
                }
                else if (node.parent.kind === 188 /* InferType */) {
                    var container_2 = getInferTypeContainer(node.parent);
                    if (container_2) {
                        if (!container_2.locals) {
                            container_2.locals = ts.createSymbolTable();
                        }
                        declareSymbol(container_2.locals, /*parent*/ undefined, node, 262144 /* TypeParameter */, 526824 /* TypeParameterExcludes */);
                    }
                    else {
                        bindAnonymousDeclaration(node, 262144 /* TypeParameter */, getDeclarationName(node)); // TODO: GH#18217
                    }
                }
                else {
                    declareSymbolAndAddToSymbolTable(node, 262144 /* TypeParameter */, 526824 /* TypeParameterExcludes */);
                }
            }
            // reachability checks
            function shouldReportErrorOnModuleDeclaration(node) {
                var instanceState = getModuleInstanceState(node);
                return instanceState === 1 /* Instantiated */ || (instanceState === 2 /* ConstEnumOnly */ && ts.shouldPreserveConstEnums(options));
            }
            function checkUnreachable(node) {
                if (!(currentFlow.flags & 1 /* Unreachable */)) {
                    return false;
                }
                if (currentFlow === unreachableFlow) {
                    var reportError = 
                    // report error on all statements except empty ones
                    (ts.isStatementButNotDeclaration(node) && node.kind !== 234 /* EmptyStatement */) ||
                        // report error on class declarations
                        node.kind === 255 /* ClassDeclaration */ ||
                        // report error on instantiated modules or const-enums only modules if preserveConstEnums is set
                        (node.kind === 259 /* ModuleDeclaration */ && shouldReportErrorOnModuleDeclaration(node));
                    if (reportError) {
                        currentFlow = reportedUnreachableFlow;
                        if (!options.allowUnreachableCode) {
                            // unreachable code is reported if
                            // - user has explicitly asked about it AND
                            // - statement is in not ambient context (statements in ambient context is already an error
                            //   so we should not report extras) AND
                            //   - node is not variable statement OR
                            //   - node is block scoped variable statement OR
                            //   - node is not block scoped variable statement and at least one variable declaration has initializer
                            //   Rationale: we don't want to report errors on non-initialized var's since they are hoisted
                            //   On the other side we do want to report errors on non-initialized 'lets' because of TDZ
                            var isError_1 = ts.unreachableCodeIsError(options) &&
                                !(node.flags & 8388608 /* Ambient */) &&
                                (!ts.isVariableStatement(node) ||
                                    !!(ts.getCombinedNodeFlags(node.declarationList) & 3 /* BlockScoped */) ||
                                    node.declarationList.declarations.some(function (d) { return !!d.initializer; }));
                            eachUnreachableRange(node, function (start, end) { return errorOrSuggestionOnRange(isError_1, start, end, ts.Diagnostics.Unreachable_code_detected); });
                        }
                    }
                }
                return true;
            }
        }
        function eachUnreachableRange(node, cb) {
            if (ts.isStatement(node) && isExecutableStatement(node) && ts.isBlock(node.parent)) {
                var statements = node.parent.statements;
                var slice_1 = ts.sliceAfter(statements, node);
                ts.getRangesWhere(slice_1, isExecutableStatement, function (start, afterEnd) { return cb(slice_1[start], slice_1[afterEnd - 1]); });
            }
            else {
                cb(node, node);
            }
        }
        // As opposed to a pure declaration like an `interface`
        function isExecutableStatement(s) {
            // Don't remove statements that can validly be used before they appear.
            return !ts.isFunctionDeclaration(s) && !isPurelyTypeDeclaration(s) && !ts.isEnumDeclaration(s) &&
                // `var x;` may declare a variable used above
                !(ts.isVariableStatement(s) && !(ts.getCombinedNodeFlags(s) & (1 /* Let */ | 2 /* Const */)) && s.declarationList.declarations.some(function (d) { return !d.initializer; }));
        }
        function isPurelyTypeDeclaration(s) {
            switch (s.kind) {
                case 256 /* InterfaceDeclaration */:
                case 257 /* TypeAliasDeclaration */:
                    return true;
                case 259 /* ModuleDeclaration */:
                    return getModuleInstanceState(s) !== 1 /* Instantiated */;
                case 258 /* EnumDeclaration */:
                    return ts.hasSyntacticModifier(s, 2048 /* Const */);
                default:
                    return false;
            }
        }
        function isExportsOrModuleExportsOrAlias(sourceFile, node) {
            var i = 0;
            var q = [node];
            while (q.length && i < 100) {
                i++;
                node = q.shift();
                if (ts.isExportsIdentifier(node) || ts.isModuleExportsAccessExpression(node)) {
                    return true;
                }
                else if (ts.isIdentifier(node)) {
                    var symbol = lookupSymbolForName(sourceFile, node.escapedText);
                    if (!!symbol && !!symbol.valueDeclaration && ts.isVariableDeclaration(symbol.valueDeclaration) && !!symbol.valueDeclaration.initializer) {
                        var init = symbol.valueDeclaration.initializer;
                        q.push(init);
                        if (ts.isAssignmentExpression(init, /*excludeCompoundAssignment*/ true)) {
                            q.push(init.left);
                            q.push(init.right);
                        }
                    }
                }
            }
            return false;
        }
        ts.isExportsOrModuleExportsOrAlias = isExportsOrModuleExportsOrAlias;
        function lookupSymbolForName(container, name) {
            var local = container.locals && container.locals.get(name);
            if (local) {
                return local.exportSymbol || local;
            }
            if (ts.isSourceFile(container) && container.jsGlobalAugmentations && container.jsGlobalAugmentations.has(name)) {
                return container.jsGlobalAugmentations.get(name);
            }
            return container.symbol && container.symbol.exports && container.symbol.exports.get(name);
        }
    })(ts || (ts = {}));
    /** @internal */
    var ts;
    (function (ts) {
        function createGetSymbolWalker(getRestTypeOfSignature, getTypePredicateOfSignature, getReturnTypeOfSignature, getBaseTypes, resolveStructuredTypeMembers, getTypeOfSymbol, getResolvedSymbol, getConstraintOfTypeParameter, getFirstIdentifier, getTypeArguments) {
            return getSymbolWalker;
            function getSymbolWalker(accept) {
                if (accept === void 0) {
                    accept = function () { return true; };
                }
                var visitedTypes = []; // Sparse array from id to type
                var visitedSymbols = []; // Sparse array from id to symbol
                return {
                    walkType: function (type) {
                        try {
                            visitType(type);
                            return { visitedTypes: ts.getOwnValues(visitedTypes), visitedSymbols: ts.getOwnValues(visitedSymbols) };
                        }
                        finally {
                            ts.clear(visitedTypes);
                            ts.clear(visitedSymbols);
                        }
                    },
                    walkSymbol: function (symbol) {
                        try {
                            visitSymbol(symbol);
                            return { visitedTypes: ts.getOwnValues(visitedTypes), visitedSymbols: ts.getOwnValues(visitedSymbols) };
                        }
                        finally {
                            ts.clear(visitedTypes);
                            ts.clear(visitedSymbols);
                        }
                    },
                };
                function visitType(type) {
                    if (!type) {
                        return;
                    }
                    if (visitedTypes[type.id]) {
                        return;
                    }
                    visitedTypes[type.id] = type;
                    // Reuse visitSymbol to visit the type's symbol,
                    //  but be sure to bail on recuring into the type if accept declines the symbol.
                    var shouldBail = visitSymbol(type.symbol);
                    if (shouldBail)
                        return;
                    // Visit the type's related types, if any
                    if (type.flags & 524288 /* Object */) {
                        var objectType = type;
                        var objectFlags = objectType.objectFlags;
                        if (objectFlags & 4 /* Reference */) {
                            visitTypeReference(type);
                        }
                        if (objectFlags & 32 /* Mapped */) {
                            visitMappedType(type);
                        }
                        if (objectFlags & (1 /* Class */ | 2 /* Interface */)) {
                            visitInterfaceType(type);
                        }
                        if (objectFlags & (8 /* Tuple */ | 16 /* Anonymous */)) {
                            visitObjectType(objectType);
                        }
                    }
                    if (type.flags & 262144 /* TypeParameter */) {
                        visitTypeParameter(type);
                    }
                    if (type.flags & 3145728 /* UnionOrIntersection */) {
                        visitUnionOrIntersectionType(type);
                    }
                    if (type.flags & 4194304 /* Index */) {
                        visitIndexType(type);
                    }
                    if (type.flags & 8388608 /* IndexedAccess */) {
                        visitIndexedAccessType(type);
                    }
                }
                function visitTypeReference(type) {
                    visitType(type.target);
                    ts.forEach(getTypeArguments(type), visitType);
                }
                function visitTypeParameter(type) {
                    visitType(getConstraintOfTypeParameter(type));
                }
                function visitUnionOrIntersectionType(type) {
                    ts.forEach(type.types, visitType);
                }
                function visitIndexType(type) {
                    visitType(type.type);
                }
                function visitIndexedAccessType(type) {
                    visitType(type.objectType);
                    visitType(type.indexType);
                    visitType(type.constraint);
                }
                function visitMappedType(type) {
                    visitType(type.typeParameter);
                    visitType(type.constraintType);
                    visitType(type.templateType);
                    visitType(type.modifiersType);
                }
                function visitSignature(signature) {
                    var typePredicate = getTypePredicateOfSignature(signature);
                    if (typePredicate) {
                        visitType(typePredicate.type);
                    }
                    ts.forEach(signature.typeParameters, visitType);
                    for (var _i = 0, _a = signature.parameters; _i < _a.length; _i++) {
                        var parameter = _a[_i];
                        visitSymbol(parameter);
                    }
                    visitType(getRestTypeOfSignature(signature));
                    visitType(getReturnTypeOfSignature(signature));
                }
                function visitInterfaceType(interfaceT) {
                    visitObjectType(interfaceT);
                    ts.forEach(interfaceT.typeParameters, visitType);
                    ts.forEach(getBaseTypes(interfaceT), visitType);
                    visitType(interfaceT.thisType);
                }
                function visitObjectType(type) {
                    var resolved = resolveStructuredTypeMembers(type);
                    for (var _i = 0, _a = resolved.indexInfos; _i < _a.length; _i++) {
                        var info = _a[_i];
                        visitType(info.keyType);
                        visitType(info.type);
                    }
                    for (var _b = 0, _c = resolved.callSignatures; _b < _c.length; _b++) {
                        var signature = _c[_b];
                        visitSignature(signature);
                    }
                    for (var _d = 0, _e = resolved.constructSignatures; _d < _e.length; _d++) {
                        var signature = _e[_d];
                        visitSignature(signature);
                    }
                    for (var _f = 0, _g = resolved.properties; _f < _g.length; _f++) {
                        var p = _g[_f];
                        visitSymbol(p);
                    }
                }
                function visitSymbol(symbol) {
                    if (!symbol) {
                        return false;
                    }
                    var symbolId = ts.getSymbolId(symbol);
                    if (visitedSymbols[symbolId]) {
                        return false;
                    }
                    visitedSymbols[symbolId] = symbol;
                    if (!accept(symbol)) {
                        return true;
                    }
                    var t = getTypeOfSymbol(symbol);
                    visitType(t); // Should handle members on classes and such
                    if (symbol.exports) {
                        symbol.exports.forEach(visitSymbol);
                    }
                    ts.forEach(symbol.declarations, function (d) {
                        // Type queries are too far resolved when we just visit the symbol's type
                        //  (their type resolved directly to the member deeply referenced)
                        // So to get the intervening symbols, we need to check if there's a type
                        // query node on any of the symbol's declarations and get symbols there
                        if (d.type && d.type.kind === 179 /* TypeQuery */) {
                            var query = d.type;
                            var entity = getResolvedSymbol(getFirstIdentifier(query.exprName));
                            visitSymbol(entity);
                        }
                    });
                    return false;
                }
            }
        }
        ts.createGetSymbolWalker = createGetSymbolWalker;
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var ambientModuleSymbolRegex = /^".+"$/;
        var anon = "(anonymous)";
        var nextSymbolId = 1;
        var nextNodeId = 1;
        var nextMergeId = 1;
        var nextFlowId = 1;
        var IterationUse;
        (function (IterationUse) {
            IterationUse[IterationUse["AllowsSyncIterablesFlag"] = 1] = "AllowsSyncIterablesFlag";
            IterationUse[IterationUse["AllowsAsyncIterablesFlag"] = 2] = "AllowsAsyncIterablesFlag";
            IterationUse[IterationUse["AllowsStringInputFlag"] = 4] = "AllowsStringInputFlag";
            IterationUse[IterationUse["ForOfFlag"] = 8] = "ForOfFlag";
            IterationUse[IterationUse["YieldStarFlag"] = 16] = "YieldStarFlag";
            IterationUse[IterationUse["SpreadFlag"] = 32] = "SpreadFlag";
            IterationUse[IterationUse["DestructuringFlag"] = 64] = "DestructuringFlag";
            IterationUse[IterationUse["PossiblyOutOfBounds"] = 128] = "PossiblyOutOfBounds";
            // Spread, Destructuring, Array element assignment
            IterationUse[IterationUse["Element"] = 1] = "Element";
            IterationUse[IterationUse["Spread"] = 33] = "Spread";
            IterationUse[IterationUse["Destructuring"] = 65] = "Destructuring";
            IterationUse[IterationUse["ForOf"] = 13] = "ForOf";
            IterationUse[IterationUse["ForAwaitOf"] = 15] = "ForAwaitOf";
            IterationUse[IterationUse["YieldStar"] = 17] = "YieldStar";
            IterationUse[IterationUse["AsyncYieldStar"] = 19] = "AsyncYieldStar";
            IterationUse[IterationUse["GeneratorReturnType"] = 1] = "GeneratorReturnType";
            IterationUse[IterationUse["AsyncGeneratorReturnType"] = 2] = "AsyncGeneratorReturnType";
        })(IterationUse || (IterationUse = {}));
        var IterationTypeKind;
        (function (IterationTypeKind) {
            IterationTypeKind[IterationTypeKind["Yield"] = 0] = "Yield";
            IterationTypeKind[IterationTypeKind["Return"] = 1] = "Return";
            IterationTypeKind[IterationTypeKind["Next"] = 2] = "Next";
        })(IterationTypeKind || (IterationTypeKind = {}));
        var WideningKind;
        (function (WideningKind) {
            WideningKind[WideningKind["Normal"] = 0] = "Normal";
            WideningKind[WideningKind["FunctionReturn"] = 1] = "FunctionReturn";
            WideningKind[WideningKind["GeneratorNext"] = 2] = "GeneratorNext";
            WideningKind[WideningKind["GeneratorYield"] = 3] = "GeneratorYield";
        })(WideningKind || (WideningKind = {}));
        var TypeFacts;
        (function (TypeFacts) {
            TypeFacts[TypeFacts["None"] = 0] = "None";
            TypeFacts[TypeFacts["TypeofEQString"] = 1] = "TypeofEQString";
            TypeFacts[TypeFacts["TypeofEQNumber"] = 2] = "TypeofEQNumber";
            TypeFacts[TypeFacts["TypeofEQBigInt"] = 4] = "TypeofEQBigInt";
            TypeFacts[TypeFacts["TypeofEQBoolean"] = 8] = "TypeofEQBoolean";
            TypeFacts[TypeFacts["TypeofEQSymbol"] = 16] = "TypeofEQSymbol";
            TypeFacts[TypeFacts["TypeofEQObject"] = 32] = "TypeofEQObject";
            TypeFacts[TypeFacts["TypeofEQFunction"] = 64] = "TypeofEQFunction";
            TypeFacts[TypeFacts["TypeofEQHostObject"] = 128] = "TypeofEQHostObject";
            TypeFacts[TypeFacts["TypeofNEString"] = 256] = "TypeofNEString";
            TypeFacts[TypeFacts["TypeofNENumber"] = 512] = "TypeofNENumber";
            TypeFacts[TypeFacts["TypeofNEBigInt"] = 1024] = "TypeofNEBigInt";
            TypeFacts[TypeFacts["TypeofNEBoolean"] = 2048] = "TypeofNEBoolean";
            TypeFacts[TypeFacts["TypeofNESymbol"] = 4096] = "TypeofNESymbol";
            TypeFacts[TypeFacts["TypeofNEObject"] = 8192] = "TypeofNEObject";
            TypeFacts[TypeFacts["TypeofNEFunction"] = 16384] = "TypeofNEFunction";
            TypeFacts[TypeFacts["TypeofNEHostObject"] = 32768] = "TypeofNEHostObject";
            TypeFacts[TypeFacts["EQUndefined"] = 65536] = "EQUndefined";
            TypeFacts[TypeFacts["EQNull"] = 131072] = "EQNull";
            TypeFacts[TypeFacts["EQUndefinedOrNull"] = 262144] = "EQUndefinedOrNull";
            TypeFacts[TypeFacts["NEUndefined"] = 524288] = "NEUndefined";
            TypeFacts[TypeFacts["NENull"] = 1048576] = "NENull";
            TypeFacts[TypeFacts["NEUndefinedOrNull"] = 2097152] = "NEUndefinedOrNull";
            TypeFacts[TypeFacts["Truthy"] = 4194304] = "Truthy";
            TypeFacts[TypeFacts["Falsy"] = 8388608] = "Falsy";
            TypeFacts[TypeFacts["All"] = 16777215] = "All";
            // The following members encode facts about particular kinds of types for use in the getTypeFacts function.
            // The presence of a particular fact means that the given test is true for some (and possibly all) values
            // of that kind of type.
            TypeFacts[TypeFacts["BaseStringStrictFacts"] = 3735041] = "BaseStringStrictFacts";
            TypeFacts[TypeFacts["BaseStringFacts"] = 12582401] = "BaseStringFacts";
            TypeFacts[TypeFacts["StringStrictFacts"] = 16317953] = "StringStrictFacts";
            TypeFacts[TypeFacts["StringFacts"] = 16776705] = "StringFacts";
            TypeFacts[TypeFacts["EmptyStringStrictFacts"] = 12123649] = "EmptyStringStrictFacts";
            TypeFacts[TypeFacts["EmptyStringFacts"] = 12582401] = "EmptyStringFacts";
            TypeFacts[TypeFacts["NonEmptyStringStrictFacts"] = 7929345] = "NonEmptyStringStrictFacts";
            TypeFacts[TypeFacts["NonEmptyStringFacts"] = 16776705] = "NonEmptyStringFacts";
            TypeFacts[TypeFacts["BaseNumberStrictFacts"] = 3734786] = "BaseNumberStrictFacts";
            TypeFacts[TypeFacts["BaseNumberFacts"] = 12582146] = "BaseNumberFacts";
            TypeFacts[TypeFacts["NumberStrictFacts"] = 16317698] = "NumberStrictFacts";
            TypeFacts[TypeFacts["NumberFacts"] = 16776450] = "NumberFacts";
            TypeFacts[TypeFacts["ZeroNumberStrictFacts"] = 12123394] = "ZeroNumberStrictFacts";
            TypeFacts[TypeFacts["ZeroNumberFacts"] = 12582146] = "ZeroNumberFacts";
            TypeFacts[TypeFacts["NonZeroNumberStrictFacts"] = 7929090] = "NonZeroNumberStrictFacts";
            TypeFacts[TypeFacts["NonZeroNumberFacts"] = 16776450] = "NonZeroNumberFacts";
            TypeFacts[TypeFacts["BaseBigIntStrictFacts"] = 3734276] = "BaseBigIntStrictFacts";
            TypeFacts[TypeFacts["BaseBigIntFacts"] = 12581636] = "BaseBigIntFacts";
            TypeFacts[TypeFacts["BigIntStrictFacts"] = 16317188] = "BigIntStrictFacts";
            TypeFacts[TypeFacts["BigIntFacts"] = 16775940] = "BigIntFacts";
            TypeFacts[TypeFacts["ZeroBigIntStrictFacts"] = 12122884] = "ZeroBigIntStrictFacts";
            TypeFacts[TypeFacts["ZeroBigIntFacts"] = 12581636] = "ZeroBigIntFacts";
            TypeFacts[TypeFacts["NonZeroBigIntStrictFacts"] = 7928580] = "NonZeroBigIntStrictFacts";
            TypeFacts[TypeFacts["NonZeroBigIntFacts"] = 16775940] = "NonZeroBigIntFacts";
            TypeFacts[TypeFacts["BaseBooleanStrictFacts"] = 3733256] = "BaseBooleanStrictFacts";
            TypeFacts[TypeFacts["BaseBooleanFacts"] = 12580616] = "BaseBooleanFacts";
            TypeFacts[TypeFacts["BooleanStrictFacts"] = 16316168] = "BooleanStrictFacts";
            TypeFacts[TypeFacts["BooleanFacts"] = 16774920] = "BooleanFacts";
            TypeFacts[TypeFacts["FalseStrictFacts"] = 12121864] = "FalseStrictFacts";
            TypeFacts[TypeFacts["FalseFacts"] = 12580616] = "FalseFacts";
            TypeFacts[TypeFacts["TrueStrictFacts"] = 7927560] = "TrueStrictFacts";
            TypeFacts[TypeFacts["TrueFacts"] = 16774920] = "TrueFacts";
            TypeFacts[TypeFacts["SymbolStrictFacts"] = 7925520] = "SymbolStrictFacts";
            TypeFacts[TypeFacts["SymbolFacts"] = 16772880] = "SymbolFacts";
            TypeFacts[TypeFacts["ObjectStrictFacts"] = 7888800] = "ObjectStrictFacts";
            TypeFacts[TypeFacts["ObjectFacts"] = 16736160] = "ObjectFacts";
            TypeFacts[TypeFacts["FunctionStrictFacts"] = 7880640] = "FunctionStrictFacts";
            TypeFacts[TypeFacts["FunctionFacts"] = 16728000] = "FunctionFacts";
            TypeFacts[TypeFacts["UndefinedFacts"] = 9830144] = "UndefinedFacts";
            TypeFacts[TypeFacts["NullFacts"] = 9363232] = "NullFacts";
            TypeFacts[TypeFacts["EmptyObjectStrictFacts"] = 16318463] = "EmptyObjectStrictFacts";
            TypeFacts[TypeFacts["AllTypeofNE"] = 556800] = "AllTypeofNE";
            TypeFacts[TypeFacts["EmptyObjectFacts"] = 16777215] = "EmptyObjectFacts";
        })(TypeFacts || (TypeFacts = {}));
        var typeofEQFacts = new ts.Map(ts.getEntries({
            string: 1 /* TypeofEQString */,
            number: 2 /* TypeofEQNumber */,
            bigint: 4 /* TypeofEQBigInt */,
            boolean: 8 /* TypeofEQBoolean */,
            symbol: 16 /* TypeofEQSymbol */,
            undefined: 65536 /* EQUndefined */,
            object: 32 /* TypeofEQObject */,
            function: 64 /* TypeofEQFunction */
        }));
        var typeofNEFacts = new ts.Map(ts.getEntries({
            string: 256 /* TypeofNEString */,
            number: 512 /* TypeofNENumber */,
            bigint: 1024 /* TypeofNEBigInt */,
            boolean: 2048 /* TypeofNEBoolean */,
            symbol: 4096 /* TypeofNESymbol */,
            undefined: 524288 /* NEUndefined */,
            object: 8192 /* TypeofNEObject */,
            function: 16384 /* TypeofNEFunction */
        }));
        var TypeSystemPropertyName;
        (function (TypeSystemPropertyName) {
            TypeSystemPropertyName[TypeSystemPropertyName["Type"] = 0] = "Type";
            TypeSystemPropertyName[TypeSystemPropertyName["ResolvedBaseConstructorType"] = 1] = "ResolvedBaseConstructorType";
            TypeSystemPropertyName[TypeSystemPropertyName["DeclaredType"] = 2] = "DeclaredType";
            TypeSystemPropertyName[TypeSystemPropertyName["ResolvedReturnType"] = 3] = "ResolvedReturnType";
            TypeSystemPropertyName[TypeSystemPropertyName["ImmediateBaseConstraint"] = 4] = "ImmediateBaseConstraint";
            TypeSystemPropertyName[TypeSystemPropertyName["EnumTagType"] = 5] = "EnumTagType";
            TypeSystemPropertyName[TypeSystemPropertyName["ResolvedTypeArguments"] = 6] = "ResolvedTypeArguments";
            TypeSystemPropertyName[TypeSystemPropertyName["ResolvedBaseTypes"] = 7] = "ResolvedBaseTypes";
        })(TypeSystemPropertyName || (TypeSystemPropertyName = {}));
        var CheckMode;
        (function (CheckMode) {
            CheckMode[CheckMode["Normal"] = 0] = "Normal";
            CheckMode[CheckMode["Contextual"] = 1] = "Contextual";
            CheckMode[CheckMode["Inferential"] = 2] = "Inferential";
            CheckMode[CheckMode["SkipContextSensitive"] = 4] = "SkipContextSensitive";
            CheckMode[CheckMode["SkipGenericFunctions"] = 8] = "SkipGenericFunctions";
            CheckMode[CheckMode["IsForSignatureHelp"] = 16] = "IsForSignatureHelp";
        })(CheckMode || (CheckMode = {}));
        var SignatureCheckMode;
        (function (SignatureCheckMode) {
            SignatureCheckMode[SignatureCheckMode["BivariantCallback"] = 1] = "BivariantCallback";
            SignatureCheckMode[SignatureCheckMode["StrictCallback"] = 2] = "StrictCallback";
            SignatureCheckMode[SignatureCheckMode["IgnoreReturnTypes"] = 4] = "IgnoreReturnTypes";
            SignatureCheckMode[SignatureCheckMode["StrictArity"] = 8] = "StrictArity";
            SignatureCheckMode[SignatureCheckMode["Callback"] = 3] = "Callback";
        })(SignatureCheckMode || (SignatureCheckMode = {}));
        var IntersectionState;
        (function (IntersectionState) {
            IntersectionState[IntersectionState["None"] = 0] = "None";
            IntersectionState[IntersectionState["Source"] = 1] = "Source";
            IntersectionState[IntersectionState["Target"] = 2] = "Target";
            IntersectionState[IntersectionState["PropertyCheck"] = 4] = "PropertyCheck";
            IntersectionState[IntersectionState["UnionIntersectionCheck"] = 8] = "UnionIntersectionCheck";
            IntersectionState[IntersectionState["InPropertyCheck"] = 16] = "InPropertyCheck";
        })(IntersectionState || (IntersectionState = {}));
        var MappedTypeModifiers;
        (function (MappedTypeModifiers) {
            MappedTypeModifiers[MappedTypeModifiers["IncludeReadonly"] = 1] = "IncludeReadonly";
            MappedTypeModifiers[MappedTypeModifiers["ExcludeReadonly"] = 2] = "ExcludeReadonly";
            MappedTypeModifiers[MappedTypeModifiers["IncludeOptional"] = 4] = "IncludeOptional";
            MappedTypeModifiers[MappedTypeModifiers["ExcludeOptional"] = 8] = "ExcludeOptional";
        })(MappedTypeModifiers || (MappedTypeModifiers = {}));
        var ExpandingFlags;
        (function (ExpandingFlags) {
            ExpandingFlags[ExpandingFlags["None"] = 0] = "None";
            ExpandingFlags[ExpandingFlags["Source"] = 1] = "Source";
            ExpandingFlags[ExpandingFlags["Target"] = 2] = "Target";
            ExpandingFlags[ExpandingFlags["Both"] = 3] = "Both";
        })(ExpandingFlags || (ExpandingFlags = {}));
        var MembersOrExportsResolutionKind;
        (function (MembersOrExportsResolutionKind) {
            MembersOrExportsResolutionKind["resolvedExports"] = "resolvedExports";
            MembersOrExportsResolutionKind["resolvedMembers"] = "resolvedMembers";
        })(MembersOrExportsResolutionKind || (MembersOrExportsResolutionKind = {}));
        var UnusedKind;
        (function (UnusedKind) {
            UnusedKind[UnusedKind["Local"] = 0] = "Local";
            UnusedKind[UnusedKind["Parameter"] = 1] = "Parameter";
        })(UnusedKind || (UnusedKind = {}));
        var isNotOverloadAndNotAccessor = ts.and(isNotOverload, isNotAccessor);
        var DeclarationMeaning;
        (function (DeclarationMeaning) {
            DeclarationMeaning[DeclarationMeaning["GetAccessor"] = 1] = "GetAccessor";
            DeclarationMeaning[DeclarationMeaning["SetAccessor"] = 2] = "SetAccessor";
            DeclarationMeaning[DeclarationMeaning["PropertyAssignment"] = 4] = "PropertyAssignment";
            DeclarationMeaning[DeclarationMeaning["Method"] = 8] = "Method";
            DeclarationMeaning[DeclarationMeaning["PrivateStatic"] = 16] = "PrivateStatic";
            DeclarationMeaning[DeclarationMeaning["GetOrSetAccessor"] = 3] = "GetOrSetAccessor";
            DeclarationMeaning[DeclarationMeaning["PropertyAssignmentOrMethod"] = 12] = "PropertyAssignmentOrMethod";
        })(DeclarationMeaning || (DeclarationMeaning = {}));
        var DeclarationSpaces;
        (function (DeclarationSpaces) {
            DeclarationSpaces[DeclarationSpaces["None"] = 0] = "None";
            DeclarationSpaces[DeclarationSpaces["ExportValue"] = 1] = "ExportValue";
            DeclarationSpaces[DeclarationSpaces["ExportType"] = 2] = "ExportType";
            DeclarationSpaces[DeclarationSpaces["ExportNamespace"] = 4] = "ExportNamespace";
        })(DeclarationSpaces || (DeclarationSpaces = {}));
        var MinArgumentCountFlags;
        (function (MinArgumentCountFlags) {
            MinArgumentCountFlags[MinArgumentCountFlags["None"] = 0] = "None";
            MinArgumentCountFlags[MinArgumentCountFlags["StrongArityForUntypedJS"] = 1] = "StrongArityForUntypedJS";
            MinArgumentCountFlags[MinArgumentCountFlags["VoidIsNonOptional"] = 2] = "VoidIsNonOptional";
        })(MinArgumentCountFlags || (MinArgumentCountFlags = {}));
        var IntrinsicTypeKind;
        (function (IntrinsicTypeKind) {
            IntrinsicTypeKind[IntrinsicTypeKind["Uppercase"] = 0] = "Uppercase";
            IntrinsicTypeKind[IntrinsicTypeKind["Lowercase"] = 1] = "Lowercase";
            IntrinsicTypeKind[IntrinsicTypeKind["Capitalize"] = 2] = "Capitalize";
            IntrinsicTypeKind[IntrinsicTypeKind["Uncapitalize"] = 3] = "Uncapitalize";
        })(IntrinsicTypeKind || (IntrinsicTypeKind = {}));
        var intrinsicTypeKinds = new ts.Map(ts.getEntries({
            Uppercase: 0 /* Uppercase */,
            Lowercase: 1 /* Lowercase */,
            Capitalize: 2 /* Capitalize */,
            Uncapitalize: 3 /* Uncapitalize */
        }));
        function SymbolLinks() {
        }
        function NodeLinks() {
            this.flags = 0;
        }
        function getNodeId(node) {
            if (!node.id) {
                node.id = nextNodeId;
                nextNodeId++;
            }
            return node.id;
        }
        ts.getNodeId = getNodeId;
        function getSymbolId(symbol) {
            if (!symbol.id) {
                symbol.id = nextSymbolId;
                nextSymbolId++;
            }
            return symbol.id;
        }
        ts.getSymbolId = getSymbolId;
        function isInstantiatedModule(node, preserveConstEnums) {
            var moduleState = ts.getModuleInstanceState(node);
            return moduleState === 1 /* Instantiated */ ||
                (preserveConstEnums && moduleState === 2 /* ConstEnumOnly */);
        }
        ts.isInstantiatedModule = isInstantiatedModule;
        function createTypeChecker(host, produceDiagnostics) {
            var getPackagesSet = ts.memoize(function () {
                var set = new ts.Set();
                host.getSourceFiles().forEach(function (sf) {
                    if (!sf.resolvedModules)
                        return;
                    ts.forEachEntry(sf.resolvedModules, function (r) {
                        if (r && r.packageId)
                            set.add(r.packageId.name);
                    });
                });
                return set;
            });
            // Cancellation that controls whether or not we can cancel in the middle of type checking.
            // In general cancelling is *not* safe for the type checker.  We might be in the middle of
            // computing something, and we will leave our internals in an inconsistent state.  Callers
            // who set the cancellation token should catch if a cancellation exception occurs, and
            // should throw away and create a new TypeChecker.
            //
            // Currently we only support setting the cancellation token when getting diagnostics.  This
            // is because diagnostics can be quite expensive, and we want to allow hosts to bail out if
            // they no longer need the information (for example, if the user started editing again).
            var cancellationToken;
            var requestedExternalEmitHelpers;
            var externalHelpersModule;
            var Symbol = ts.objectAllocator.getSymbolConstructor();
            var Type = ts.objectAllocator.getTypeConstructor();
            var Signature = ts.objectAllocator.getSignatureConstructor();
            var typeCount = 0;
            var symbolCount = 0;
            var enumCount = 0;
            var totalInstantiationCount = 0;
            var instantiationCount = 0;
            var instantiationDepth = 0;
            var currentNode;
            var emptySymbols = ts.createSymbolTable();
            var arrayVariances = [1 /* Covariant */];
            var compilerOptions = host.getCompilerOptions();
            var languageVersion = ts.getEmitScriptTarget(compilerOptions);
            var moduleKind = ts.getEmitModuleKind(compilerOptions);
            var useDefineForClassFields = ts.getUseDefineForClassFields(compilerOptions);
            var allowSyntheticDefaultImports = ts.getAllowSyntheticDefaultImports(compilerOptions);
            var strictNullChecks = ts.getStrictOptionValue(compilerOptions, "strictNullChecks");
            var strictFunctionTypes = ts.getStrictOptionValue(compilerOptions, "strictFunctionTypes");
            var strictBindCallApply = ts.getStrictOptionValue(compilerOptions, "strictBindCallApply");
            var strictPropertyInitialization = ts.getStrictOptionValue(compilerOptions, "strictPropertyInitialization");
            var noImplicitAny = ts.getStrictOptionValue(compilerOptions, "noImplicitAny");
            var noImplicitThis = ts.getStrictOptionValue(compilerOptions, "noImplicitThis");
            var useUnknownInCatchVariables = ts.getStrictOptionValue(compilerOptions, "useUnknownInCatchVariables");
            var keyofStringsOnly = !!compilerOptions.keyofStringsOnly;
            var freshObjectLiteralFlag = compilerOptions.suppressExcessPropertyErrors ? 0 : 16384 /* FreshLiteral */;
            var exactOptionalPropertyTypes = compilerOptions.exactOptionalPropertyTypes;
            var checkBinaryExpression = createCheckBinaryExpression();
            var emitResolver = createResolver();
            var nodeBuilder = createNodeBuilder();
            var globals = ts.createSymbolTable();
            var undefinedSymbol = createSymbol(4 /* Property */, "undefined");
            undefinedSymbol.declarations = [];
            var globalThisSymbol = createSymbol(1536 /* Module */, "globalThis", 8 /* Readonly */);
            globalThisSymbol.exports = globals;
            globalThisSymbol.declarations = [];
            globals.set(globalThisSymbol.escapedName, globalThisSymbol);
            var argumentsSymbol = createSymbol(4 /* Property */, "arguments");
            var requireSymbol = createSymbol(4 /* Property */, "require");
            /** This will be set during calls to `getResolvedSignature` where services determines an apparent number of arguments greater than what is actually provided. */
            var apparentArgumentCount;
            // for public members that accept a Node or one of its subtypes, we must guard against
            // synthetic nodes created during transformations by calling `getParseTreeNode`.
            // for most of these, we perform the guard only on `checker` to avoid any possible
            // extra cost of calling `getParseTreeNode` when calling these functions from inside the
            // checker.
            var checker = {
                getNodeCount: function () { return ts.sum(host.getSourceFiles(), "nodeCount"); },
                getIdentifierCount: function () { return ts.sum(host.getSourceFiles(), "identifierCount"); },
                getSymbolCount: function () { return ts.sum(host.getSourceFiles(), "symbolCount") + symbolCount; },
                getTypeCount: function () { return typeCount; },
                getInstantiationCount: function () { return totalInstantiationCount; },
                getRelationCacheSizes: function () {
                    return ({
                        assignable: assignableRelation.size,
                        identity: identityRelation.size,
                        subtype: subtypeRelation.size,
                        strictSubtype: strictSubtypeRelation.size,
                    });
                },
                isUndefinedSymbol: function (symbol) { return symbol === undefinedSymbol; },
                isArgumentsSymbol: function (symbol) { return symbol === argumentsSymbol; },
                isUnknownSymbol: function (symbol) { return symbol === unknownSymbol; },
                getMergedSymbol: getMergedSymbol,
                getDiagnostics: getDiagnostics,
                getGlobalDiagnostics: getGlobalDiagnostics,
                getRecursionIdentity: getRecursionIdentity,
                getUnmatchedProperties: getUnmatchedProperties,
                getTypeOfSymbolAtLocation: function (symbol, locationIn) {
                    var location = ts.getParseTreeNode(locationIn);
                    return location ? getTypeOfSymbolAtLocation(symbol, location) : errorType;
                },
                getSymbolsOfParameterPropertyDeclaration: function (parameterIn, parameterName) {
                    var parameter = ts.getParseTreeNode(parameterIn, ts.isParameter);
                    if (parameter === undefined)
                        return ts.Debug.fail("Cannot get symbols of a synthetic parameter that cannot be resolved to a parse-tree node.");
                    return getSymbolsOfParameterPropertyDeclaration(parameter, ts.escapeLeadingUnderscores(parameterName));
                },
                getDeclaredTypeOfSymbol: getDeclaredTypeOfSymbol,
                getPropertiesOfType: getPropertiesOfType,
                getPropertyOfType: function (type, name) { return getPropertyOfType(type, ts.escapeLeadingUnderscores(name)); },
                getPrivateIdentifierPropertyOfType: function (leftType, name, location) {
                    var node = ts.getParseTreeNode(location);
                    if (!node) {
                        return undefined;
                    }
                    var propName = ts.escapeLeadingUnderscores(name);
                    var lexicallyScopedIdentifier = lookupSymbolForPrivateIdentifierDeclaration(propName, node);
                    return lexicallyScopedIdentifier ? getPrivateIdentifierPropertyOfType(leftType, lexicallyScopedIdentifier) : undefined;
                },
                getTypeOfPropertyOfType: function (type, name) { return getTypeOfPropertyOfType(type, ts.escapeLeadingUnderscores(name)); },
                getIndexInfoOfType: function (type, kind) { return getIndexInfoOfType(type, kind === 0 /* String */ ? stringType : numberType); },
                getIndexInfosOfType: getIndexInfosOfType,
                getSignaturesOfType: getSignaturesOfType,
                getIndexTypeOfType: function (type, kind) { return getIndexTypeOfType(type, kind === 0 /* String */ ? stringType : numberType); },
                getBaseTypes: getBaseTypes,
                getBaseTypeOfLiteralType: getBaseTypeOfLiteralType,
                getWidenedType: getWidenedType,
                getTypeFromTypeNode: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isTypeNode);
                    return node ? getTypeFromTypeNode(node) : errorType;
                },
                getParameterType: getTypeAtPosition,
                getParameterIdentifierNameAtPosition: getParameterIdentifierNameAtPosition,
                getPromisedTypeOfPromise: getPromisedTypeOfPromise,
                getAwaitedType: function (type) { return getAwaitedType(type); },
                getReturnTypeOfSignature: getReturnTypeOfSignature,
                isNullableType: isNullableType,
                getNullableType: getNullableType,
                getNonNullableType: getNonNullableType,
                getNonOptionalType: removeOptionalTypeMarker,
                getTypeArguments: getTypeArguments,
                typeToTypeNode: nodeBuilder.typeToTypeNode,
                indexInfoToIndexSignatureDeclaration: nodeBuilder.indexInfoToIndexSignatureDeclaration,
                signatureToSignatureDeclaration: nodeBuilder.signatureToSignatureDeclaration,
                symbolToEntityName: nodeBuilder.symbolToEntityName,
                symbolToExpression: nodeBuilder.symbolToExpression,
                symbolToTypeParameterDeclarations: nodeBuilder.symbolToTypeParameterDeclarations,
                symbolToParameterDeclaration: nodeBuilder.symbolToParameterDeclaration,
                typeParameterToDeclaration: nodeBuilder.typeParameterToDeclaration,
                getSymbolsInScope: function (locationIn, meaning) {
                    var location = ts.getParseTreeNode(locationIn);
                    return location ? getSymbolsInScope(location, meaning) : [];
                },
                getSymbolAtLocation: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn);
                    // set ignoreErrors: true because any lookups invoked by the API shouldn't cause any new errors
                    return node ? getSymbolAtLocation(node, /*ignoreErrors*/ true) : undefined;
                },
                getIndexInfosAtLocation: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn);
                    return node ? getIndexInfosAtLocation(node) : undefined;
                },
                getShorthandAssignmentValueSymbol: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn);
                    return node ? getShorthandAssignmentValueSymbol(node) : undefined;
                },
                getExportSpecifierLocalTargetSymbol: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isExportSpecifier);
                    return node ? getExportSpecifierLocalTargetSymbol(node) : undefined;
                },
                getExportSymbolOfSymbol: function (symbol) {
                    return getMergedSymbol(symbol.exportSymbol || symbol);
                },
                getTypeAtLocation: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn);
                    return node ? getTypeOfNode(node) : errorType;
                },
                getTypeOfAssignmentPattern: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isAssignmentPattern);
                    return node && getTypeOfAssignmentPattern(node) || errorType;
                },
                getPropertySymbolOfDestructuringAssignment: function (locationIn) {
                    var location = ts.getParseTreeNode(locationIn, ts.isIdentifier);
                    return location ? getPropertySymbolOfDestructuringAssignment(location) : undefined;
                },
                signatureToString: function (signature, enclosingDeclaration, flags, kind) {
                    return signatureToString(signature, ts.getParseTreeNode(enclosingDeclaration), flags, kind);
                },
                typeToString: function (type, enclosingDeclaration, flags) {
                    return typeToString(type, ts.getParseTreeNode(enclosingDeclaration), flags);
                },
                symbolToString: function (symbol, enclosingDeclaration, meaning, flags) {
                    return symbolToString(symbol, ts.getParseTreeNode(enclosingDeclaration), meaning, flags);
                },
                typePredicateToString: function (predicate, enclosingDeclaration, flags) {
                    return typePredicateToString(predicate, ts.getParseTreeNode(enclosingDeclaration), flags);
                },
                writeSignature: function (signature, enclosingDeclaration, flags, kind, writer) {
                    return signatureToString(signature, ts.getParseTreeNode(enclosingDeclaration), flags, kind, writer);
                },
                writeType: function (type, enclosingDeclaration, flags, writer) {
                    return typeToString(type, ts.getParseTreeNode(enclosingDeclaration), flags, writer);
                },
                writeSymbol: function (symbol, enclosingDeclaration, meaning, flags, writer) {
                    return symbolToString(symbol, ts.getParseTreeNode(enclosingDeclaration), meaning, flags, writer);
                },
                writeTypePredicate: function (predicate, enclosingDeclaration, flags, writer) {
                    return typePredicateToString(predicate, ts.getParseTreeNode(enclosingDeclaration), flags, writer);
                },
                getAugmentedPropertiesOfType: getAugmentedPropertiesOfType,
                getRootSymbols: getRootSymbols,
                getSymbolOfExpando: getSymbolOfExpando,
                getContextualType: function (nodeIn, contextFlags) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isExpression);
                    if (!node) {
                        return undefined;
                    }
                    var containingCall = ts.findAncestor(node, ts.isCallLikeExpression);
                    var containingCallResolvedSignature = containingCall && getNodeLinks(containingCall).resolvedSignature;
                    if (contextFlags & 4 /* Completions */ && containingCall) {
                        var toMarkSkip = node;
                        do {
                            getNodeLinks(toMarkSkip).skipDirectInference = true;
                            toMarkSkip = toMarkSkip.parent;
                        } while (toMarkSkip && toMarkSkip !== containingCall);
                        getNodeLinks(containingCall).resolvedSignature = undefined;
                    }
                    var result = getContextualType(node, contextFlags);
                    if (contextFlags & 4 /* Completions */ && containingCall) {
                        var toMarkSkip = node;
                        do {
                            getNodeLinks(toMarkSkip).skipDirectInference = undefined;
                            toMarkSkip = toMarkSkip.parent;
                        } while (toMarkSkip && toMarkSkip !== containingCall);
                        getNodeLinks(containingCall).resolvedSignature = containingCallResolvedSignature;
                    }
                    return result;
                },
                getContextualTypeForObjectLiteralElement: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isObjectLiteralElementLike);
                    return node ? getContextualTypeForObjectLiteralElement(node) : undefined;
                },
                getContextualTypeForArgumentAtIndex: function (nodeIn, argIndex) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isCallLikeExpression);
                    return node && getContextualTypeForArgumentAtIndex(node, argIndex);
                },
                getContextualTypeForJsxAttribute: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isJsxAttributeLike);
                    return node && getContextualTypeForJsxAttribute(node);
                },
                isContextSensitive: isContextSensitive,
                getTypeOfPropertyOfContextualType: getTypeOfPropertyOfContextualType,
                getFullyQualifiedName: getFullyQualifiedName,
                getResolvedSignature: function (node, candidatesOutArray, argumentCount) {
                    return getResolvedSignatureWorker(node, candidatesOutArray, argumentCount, 0 /* Normal */);
                },
                getResolvedSignatureForSignatureHelp: function (node, candidatesOutArray, argumentCount) {
                    return getResolvedSignatureWorker(node, candidatesOutArray, argumentCount, 16 /* IsForSignatureHelp */);
                },
                getExpandedParameters: getExpandedParameters,
                hasEffectiveRestParameter: hasEffectiveRestParameter,
                containsArgumentsReference: containsArgumentsReference,
                getConstantValue: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn, canHaveConstantValue);
                    return node ? getConstantValue(node) : undefined;
                },
                isValidPropertyAccess: function (nodeIn, propertyName) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isPropertyAccessOrQualifiedNameOrImportTypeNode);
                    return !!node && isValidPropertyAccess(node, ts.escapeLeadingUnderscores(propertyName));
                },
                isValidPropertyAccessForCompletions: function (nodeIn, type, property) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isPropertyAccessExpression);
                    return !!node && isValidPropertyAccessForCompletions(node, type, property);
                },
                getSignatureFromDeclaration: function (declarationIn) {
                    var declaration = ts.getParseTreeNode(declarationIn, ts.isFunctionLike);
                    return declaration ? getSignatureFromDeclaration(declaration) : undefined;
                },
                isImplementationOfOverload: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isFunctionLike);
                    return node ? isImplementationOfOverload(node) : undefined;
                },
                getImmediateAliasedSymbol: getImmediateAliasedSymbol,
                getAliasedSymbol: resolveAlias,
                getEmitResolver: getEmitResolver,
                getExportsOfModule: getExportsOfModuleAsArray,
                getExportsAndPropertiesOfModule: getExportsAndPropertiesOfModule,
                forEachExportAndPropertyOfModule: forEachExportAndPropertyOfModule,
                getSymbolWalker: ts.createGetSymbolWalker(getRestTypeOfSignature, getTypePredicateOfSignature, getReturnTypeOfSignature, getBaseTypes, resolveStructuredTypeMembers, getTypeOfSymbol, getResolvedSymbol, getConstraintOfTypeParameter, ts.getFirstIdentifier, getTypeArguments),
                getAmbientModules: getAmbientModules,
                getJsxIntrinsicTagNamesAt: getJsxIntrinsicTagNamesAt,
                isOptionalParameter: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isParameter);
                    return node ? isOptionalParameter(node) : false;
                },
                tryGetMemberInModuleExports: function (name, symbol) { return tryGetMemberInModuleExports(ts.escapeLeadingUnderscores(name), symbol); },
                tryGetMemberInModuleExportsAndProperties: function (name, symbol) { return tryGetMemberInModuleExportsAndProperties(ts.escapeLeadingUnderscores(name), symbol); },
                tryFindAmbientModule: function (moduleName) { return tryFindAmbientModule(moduleName, /*withAugmentations*/ true); },
                tryFindAmbientModuleWithoutAugmentations: function (moduleName) {
                    // we deliberately exclude augmentations
                    // since we are only interested in declarations of the module itself
                    return tryFindAmbientModule(moduleName, /*withAugmentations*/ false);
                },
                getApparentType: getApparentType,
                getUnionType: getUnionType,
                isTypeAssignableTo: isTypeAssignableTo,
                createAnonymousType: createAnonymousType,
                createSignature: createSignature,
                createSymbol: createSymbol,
                createIndexInfo: createIndexInfo,
                getAnyType: function () { return anyType; },
                getStringType: function () { return stringType; },
                getNumberType: function () { return numberType; },
                createPromiseType: createPromiseType,
                createArrayType: createArrayType,
                getElementTypeOfArrayType: getElementTypeOfArrayType,
                getBooleanType: function () { return booleanType; },
                getFalseType: function (fresh) { return fresh ? falseType : regularFalseType; },
                getTrueType: function (fresh) { return fresh ? trueType : regularTrueType; },
                getVoidType: function () { return voidType; },
                getUndefinedType: function () { return undefinedType; },
                getNullType: function () { return nullType; },
                getESSymbolType: function () { return esSymbolType; },
                getNeverType: function () { return neverType; },
                getOptionalType: function () { return optionalType; },
                isSymbolAccessible: isSymbolAccessible,
                isArrayType: isArrayType,
                isTupleType: isTupleType,
                isArrayLikeType: isArrayLikeType,
                isTypeInvalidDueToUnionDiscriminant: isTypeInvalidDueToUnionDiscriminant,
                getAllPossiblePropertiesOfTypes: getAllPossiblePropertiesOfTypes,
                getSuggestedSymbolForNonexistentProperty: getSuggestedSymbolForNonexistentProperty,
                getSuggestionForNonexistentProperty: getSuggestionForNonexistentProperty,
                getSuggestedSymbolForNonexistentJSXAttribute: getSuggestedSymbolForNonexistentJSXAttribute,
                getSuggestedSymbolForNonexistentSymbol: function (location, name, meaning) { return getSuggestedSymbolForNonexistentSymbol(location, ts.escapeLeadingUnderscores(name), meaning); },
                getSuggestionForNonexistentSymbol: function (location, name, meaning) { return getSuggestionForNonexistentSymbol(location, ts.escapeLeadingUnderscores(name), meaning); },
                getSuggestedSymbolForNonexistentModule: getSuggestedSymbolForNonexistentModule,
                getSuggestionForNonexistentExport: getSuggestionForNonexistentExport,
                getSuggestedSymbolForNonexistentClassMember: getSuggestedSymbolForNonexistentClassMember,
                getBaseConstraintOfType: getBaseConstraintOfType,
                getDefaultFromTypeParameter: function (type) { return type && type.flags & 262144 /* TypeParameter */ ? getDefaultFromTypeParameter(type) : undefined; },
                resolveName: function (name, location, meaning, excludeGlobals) {
                    return resolveName(location, ts.escapeLeadingUnderscores(name), meaning, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false, excludeGlobals);
                },
                getJsxNamespace: function (n) { return ts.unescapeLeadingUnderscores(getJsxNamespace(n)); },
                getJsxFragmentFactory: function (n) {
                    var jsxFragmentFactory = getJsxFragmentFactoryEntity(n);
                    return jsxFragmentFactory && ts.unescapeLeadingUnderscores(ts.getFirstIdentifier(jsxFragmentFactory).escapedText);
                },
                getAccessibleSymbolChain: getAccessibleSymbolChain,
                getTypePredicateOfSignature: getTypePredicateOfSignature,
                resolveExternalModuleName: function (moduleSpecifierIn) {
                    var moduleSpecifier = ts.getParseTreeNode(moduleSpecifierIn, ts.isExpression);
                    return moduleSpecifier && resolveExternalModuleName(moduleSpecifier, moduleSpecifier, /*ignoreErrors*/ true);
                },
                resolveExternalModuleSymbol: resolveExternalModuleSymbol,
                tryGetThisTypeAt: function (nodeIn, includeGlobalThis) {
                    var node = ts.getParseTreeNode(nodeIn);
                    return node && tryGetThisTypeAt(node, includeGlobalThis);
                },
                getTypeArgumentConstraint: function (nodeIn) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isTypeNode);
                    return node && getTypeArgumentConstraint(node);
                },
                getSuggestionDiagnostics: function (fileIn, ct) {
                    var file = ts.getParseTreeNode(fileIn, ts.isSourceFile) || ts.Debug.fail("Could not determine parsed source file.");
                    if (ts.skipTypeChecking(file, compilerOptions, host)) {
                        return ts.emptyArray;
                    }
                    var diagnostics;
                    try {
                        // Record the cancellation token so it can be checked later on during checkSourceElement.
                        // Do this in a finally block so we can ensure that it gets reset back to nothing after
                        // this call is done.
                        cancellationToken = ct;
                        // Ensure file is type checked
                        checkSourceFile(file);
                        ts.Debug.assert(!!(getNodeLinks(file).flags & 1 /* TypeChecked */));
                        diagnostics = ts.addRange(diagnostics, suggestionDiagnostics.getDiagnostics(file.fileName));
                        checkUnusedIdentifiers(getPotentiallyUnusedIdentifiers(file), function (containingNode, kind, diag) {
                            if (!ts.containsParseError(containingNode) && !unusedIsError(kind, !!(containingNode.flags & 8388608 /* Ambient */))) {
                                (diagnostics || (diagnostics = [])).push(__assign(__assign({}, diag), { category: ts.DiagnosticCategory.Suggestion }));
                            }
                        });
                        return diagnostics || ts.emptyArray;
                    }
                    finally {
                        cancellationToken = undefined;
                    }
                },
                runWithCancellationToken: function (token, callback) {
                    try {
                        cancellationToken = token;
                        return callback(checker);
                    }
                    finally {
                        cancellationToken = undefined;
                    }
                },
                getLocalTypeParametersOfClassOrInterfaceOrTypeAlias: getLocalTypeParametersOfClassOrInterfaceOrTypeAlias,
                isDeclarationVisible: isDeclarationVisible,
            };
            function getResolvedSignatureWorker(nodeIn, candidatesOutArray, argumentCount, checkMode) {
                var node = ts.getParseTreeNode(nodeIn, ts.isCallLikeExpression);
                apparentArgumentCount = argumentCount;
                var res = node ? getResolvedSignature(node, candidatesOutArray, checkMode) : undefined;
                apparentArgumentCount = undefined;
                return res;
            }
            var tupleTypes = new ts.Map();
            var unionTypes = new ts.Map();
            var intersectionTypes = new ts.Map();
            var stringLiteralTypes = new ts.Map();
            var numberLiteralTypes = new ts.Map();
            var bigIntLiteralTypes = new ts.Map();
            var enumLiteralTypes = new ts.Map();
            var indexedAccessTypes = new ts.Map();
            var templateLiteralTypes = new ts.Map();
            var stringMappingTypes = new ts.Map();
            var substitutionTypes = new ts.Map();
            var subtypeReductionCache = new ts.Map();
            var evolvingArrayTypes = [];
            var undefinedProperties = new ts.Map();
            var unknownSymbol = createSymbol(4 /* Property */, "unknown");
            var resolvingSymbol = createSymbol(0, "__resolving__" /* Resolving */);
            var anyType = createIntrinsicType(1 /* Any */, "any");
            var autoType = createIntrinsicType(1 /* Any */, "any");
            var wildcardType = createIntrinsicType(1 /* Any */, "any");
            var errorType = createIntrinsicType(1 /* Any */, "error");
            var nonInferrableAnyType = createIntrinsicType(1 /* Any */, "any", 131072 /* ContainsWideningType */);
            var intrinsicMarkerType = createIntrinsicType(1 /* Any */, "intrinsic");
            var unknownType = createIntrinsicType(2 /* Unknown */, "unknown");
            var undefinedType = createIntrinsicType(32768 /* Undefined */, "undefined");
            var undefinedWideningType = strictNullChecks ? undefinedType : createIntrinsicType(32768 /* Undefined */, "undefined", 131072 /* ContainsWideningType */);
            var optionalType = createIntrinsicType(32768 /* Undefined */, "undefined");
            var missingType = exactOptionalPropertyTypes ? createIntrinsicType(32768 /* Undefined */, "undefined") : undefinedType;
            var nullType = createIntrinsicType(65536 /* Null */, "null");
            var nullWideningType = strictNullChecks ? nullType : createIntrinsicType(65536 /* Null */, "null", 131072 /* ContainsWideningType */);
            var stringType = createIntrinsicType(4 /* String */, "string");
            var numberType = createIntrinsicType(8 /* Number */, "number");
            var bigintType = createIntrinsicType(64 /* BigInt */, "bigint");
            var falseType = createIntrinsicType(512 /* BooleanLiteral */, "false");
            var regularFalseType = createIntrinsicType(512 /* BooleanLiteral */, "false");
            var trueType = createIntrinsicType(512 /* BooleanLiteral */, "true");
            var regularTrueType = createIntrinsicType(512 /* BooleanLiteral */, "true");
            trueType.regularType = regularTrueType;
            trueType.freshType = trueType;
            regularTrueType.regularType = regularTrueType;
            regularTrueType.freshType = trueType;
            falseType.regularType = regularFalseType;
            falseType.freshType = falseType;
            regularFalseType.regularType = regularFalseType;
            regularFalseType.freshType = falseType;
            var booleanType = getUnionType([regularFalseType, regularTrueType]);
            var esSymbolType = createIntrinsicType(4096 /* ESSymbol */, "symbol");
            var voidType = createIntrinsicType(16384 /* Void */, "void");
            var neverType = createIntrinsicType(131072 /* Never */, "never");
            var silentNeverType = createIntrinsicType(131072 /* Never */, "never");
            var nonInferrableType = createIntrinsicType(131072 /* Never */, "never", 524288 /* NonInferrableType */);
            var implicitNeverType = createIntrinsicType(131072 /* Never */, "never");
            var unreachableNeverType = createIntrinsicType(131072 /* Never */, "never");
            var nonPrimitiveType = createIntrinsicType(67108864 /* NonPrimitive */, "object");
            var stringOrNumberType = getUnionType([stringType, numberType]);
            var stringNumberSymbolType = getUnionType([stringType, numberType, esSymbolType]);
            var keyofConstraintType = keyofStringsOnly ? stringType : stringNumberSymbolType;
            var numberOrBigIntType = getUnionType([numberType, bigintType]);
            var templateConstraintType = getUnionType([stringType, numberType, booleanType, bigintType, nullType, undefinedType]);
            var restrictiveMapper = makeFunctionTypeMapper(function (t) { return t.flags & 262144 /* TypeParameter */ ? getRestrictiveTypeParameter(t) : t; });
            var permissiveMapper = makeFunctionTypeMapper(function (t) { return t.flags & 262144 /* TypeParameter */ ? wildcardType : t; });
            var emptyObjectType = createAnonymousType(undefined, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
            var emptyJsxObjectType = createAnonymousType(undefined, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
            emptyJsxObjectType.objectFlags |= 2048 /* JsxAttributes */;
            var emptyTypeLiteralSymbol = createSymbol(2048 /* TypeLiteral */, "__type" /* Type */);
            emptyTypeLiteralSymbol.members = ts.createSymbolTable();
            var emptyTypeLiteralType = createAnonymousType(emptyTypeLiteralSymbol, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
            var emptyGenericType = createAnonymousType(undefined, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
            emptyGenericType.instantiations = new ts.Map();
            var anyFunctionType = createAnonymousType(undefined, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
            // The anyFunctionType contains the anyFunctionType by definition. The flag is further propagated
            // in getPropagatingFlagsOfTypes, and it is checked in inferFromTypes.
            anyFunctionType.objectFlags |= 524288 /* NonInferrableType */;
            var noConstraintType = createAnonymousType(undefined, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
            var circularConstraintType = createAnonymousType(undefined, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
            var resolvingDefaultType = createAnonymousType(undefined, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
            var markerSuperType = createTypeParameter();
            var markerSubType = createTypeParameter();
            markerSubType.constraint = markerSuperType;
            var markerOtherType = createTypeParameter();
            var noTypePredicate = createTypePredicate(1 /* Identifier */, "<<unresolved>>", 0, anyType);
            var anySignature = createSignature(undefined, undefined, undefined, ts.emptyArray, anyType, /*resolvedTypePredicate*/ undefined, 0, 0 /* None */);
            var unknownSignature = createSignature(undefined, undefined, undefined, ts.emptyArray, errorType, /*resolvedTypePredicate*/ undefined, 0, 0 /* None */);
            var resolvingSignature = createSignature(undefined, undefined, undefined, ts.emptyArray, anyType, /*resolvedTypePredicate*/ undefined, 0, 0 /* None */);
            var silentNeverSignature = createSignature(undefined, undefined, undefined, ts.emptyArray, silentNeverType, /*resolvedTypePredicate*/ undefined, 0, 0 /* None */);
            var enumNumberIndexInfo = createIndexInfo(numberType, stringType, /*isReadonly*/ true);
            var iterationTypesCache = new ts.Map(); // cache for common IterationTypes instances
            var noIterationTypes = {
                get yieldType() { return ts.Debug.fail("Not supported"); },
                get returnType() { return ts.Debug.fail("Not supported"); },
                get nextType() { return ts.Debug.fail("Not supported"); },
            };
            var anyIterationTypes = createIterationTypes(anyType, anyType, anyType);
            var anyIterationTypesExceptNext = createIterationTypes(anyType, anyType, unknownType);
            var defaultIterationTypes = createIterationTypes(neverType, anyType, undefinedType); // default iteration types for `Iterator`.
            var asyncIterationTypesResolver = {
                iterableCacheKey: "iterationTypesOfAsyncIterable",
                iteratorCacheKey: "iterationTypesOfAsyncIterator",
                iteratorSymbolName: "asyncIterator",
                getGlobalIteratorType: getGlobalAsyncIteratorType,
                getGlobalIterableType: getGlobalAsyncIterableType,
                getGlobalIterableIteratorType: getGlobalAsyncIterableIteratorType,
                getGlobalGeneratorType: getGlobalAsyncGeneratorType,
                resolveIterationType: getAwaitedType,
                mustHaveANextMethodDiagnostic: ts.Diagnostics.An_async_iterator_must_have_a_next_method,
                mustBeAMethodDiagnostic: ts.Diagnostics.The_0_property_of_an_async_iterator_must_be_a_method,
                mustHaveAValueDiagnostic: ts.Diagnostics.The_type_returned_by_the_0_method_of_an_async_iterator_must_be_a_promise_for_a_type_with_a_value_property,
            };
            var syncIterationTypesResolver = {
                iterableCacheKey: "iterationTypesOfIterable",
                iteratorCacheKey: "iterationTypesOfIterator",
                iteratorSymbolName: "iterator",
                getGlobalIteratorType: getGlobalIteratorType,
                getGlobalIterableType: getGlobalIterableType,
                getGlobalIterableIteratorType: getGlobalIterableIteratorType,
                getGlobalGeneratorType: getGlobalGeneratorType,
                resolveIterationType: function (type, _errorNode) { return type; },
                mustHaveANextMethodDiagnostic: ts.Diagnostics.An_iterator_must_have_a_next_method,
                mustBeAMethodDiagnostic: ts.Diagnostics.The_0_property_of_an_iterator_must_be_a_method,
                mustHaveAValueDiagnostic: ts.Diagnostics.The_type_returned_by_the_0_method_of_an_iterator_must_have_a_value_property,
            };
            /** Key is "/path/to/a.ts|/path/to/b.ts". */
            var amalgamatedDuplicates;
            var reverseMappedCache = new ts.Map();
            var inInferTypeForHomomorphicMappedType = false;
            var ambientModulesCache;
            /**
             * List of every ambient module with a "*" wildcard.
             * Unlike other ambient modules, these can't be stored in `globals` because symbol tables only deal with exact matches.
             * This is only used if there is no exact match.
             */
            var patternAmbientModules;
            var patternAmbientModuleAugmentations;
            var globalObjectType;
            var globalFunctionType;
            var globalCallableFunctionType;
            var globalNewableFunctionType;
            var globalArrayType;
            var globalReadonlyArrayType;
            var globalStringType;
            var globalNumberType;
            var globalBooleanType;
            var globalRegExpType;
            var globalThisType;
            var anyArrayType;
            var autoArrayType;
            var anyReadonlyArrayType;
            var deferredGlobalNonNullableTypeAlias;
            // The library files are only loaded when the feature is used.
            // This allows users to just specify library files they want to used through --lib
            // and they will not get an error from not having unrelated library files
            var deferredGlobalESSymbolConstructorSymbol;
            var deferredGlobalESSymbolConstructorTypeSymbol;
            var deferredGlobalESSymbolType;
            var deferredGlobalTypedPropertyDescriptorType;
            var deferredGlobalPromiseType;
            var deferredGlobalPromiseLikeType;
            var deferredGlobalPromiseConstructorSymbol;
            var deferredGlobalPromiseConstructorLikeType;
            var deferredGlobalIterableType;
            var deferredGlobalIteratorType;
            var deferredGlobalIterableIteratorType;
            var deferredGlobalGeneratorType;
            var deferredGlobalIteratorYieldResultType;
            var deferredGlobalIteratorReturnResultType;
            var deferredGlobalAsyncIterableType;
            var deferredGlobalAsyncIteratorType;
            var deferredGlobalAsyncIterableIteratorType;
            var deferredGlobalAsyncGeneratorType;
            var deferredGlobalTemplateStringsArrayType;
            var deferredGlobalImportMetaType;
            var deferredGlobalImportMetaExpressionType;
            var deferredGlobalExtractSymbol;
            var deferredGlobalOmitSymbol;
            var deferredGlobalBigIntType;
            var allPotentiallyUnusedIdentifiers = new ts.Map(); // key is file name
            var flowLoopStart = 0;
            var flowLoopCount = 0;
            var sharedFlowCount = 0;
            var flowAnalysisDisabled = false;
            var flowInvocationCount = 0;
            var lastFlowNode;
            var lastFlowNodeReachable;
            var flowTypeCache;
            var emptyStringType = getStringLiteralType("");
            var zeroType = getNumberLiteralType(0);
            var zeroBigIntType = getBigIntLiteralType({ negative: false, base10Value: "0" });
            var resolutionTargets = [];
            var resolutionResults = [];
            var resolutionPropertyNames = [];
            var suggestionCount = 0;
            var maximumSuggestionCount = 10;
            var mergedSymbols = [];
            var symbolLinks = [];
            var nodeLinks = [];
            var flowLoopCaches = [];
            var flowLoopNodes = [];
            var flowLoopKeys = [];
            var flowLoopTypes = [];
            var sharedFlowNodes = [];
            var sharedFlowTypes = [];
            var flowNodeReachable = [];
            var flowNodePostSuper = [];
            var potentialThisCollisions = [];
            var potentialNewTargetCollisions = [];
            var potentialWeakMapSetCollisions = [];
            var potentialReflectCollisions = [];
            var awaitedTypeStack = [];
            var diagnostics = ts.createDiagnosticCollection();
            var suggestionDiagnostics = ts.createDiagnosticCollection();
            var typeofTypesByName = new ts.Map(ts.getEntries({
                string: stringType,
                number: numberType,
                bigint: bigintType,
                boolean: booleanType,
                symbol: esSymbolType,
                undefined: undefinedType
            }));
            var typeofType = createTypeofType();
            var _jsxNamespace;
            var _jsxFactoryEntity;
            var outofbandVarianceMarkerHandler;
            var subtypeRelation = new ts.Map();
            var strictSubtypeRelation = new ts.Map();
            var assignableRelation = new ts.Map();
            var comparableRelation = new ts.Map();
            var identityRelation = new ts.Map();
            var enumRelation = new ts.Map();
            var builtinGlobals = ts.createSymbolTable();
            builtinGlobals.set(undefinedSymbol.escapedName, undefinedSymbol);
            initializeTypeChecker();
            return checker;
            function getJsxNamespace(location) {
                if (location) {
                    var file = ts.getSourceFileOfNode(location);
                    if (file) {
                        if (ts.isJsxOpeningFragment(location)) {
                            if (file.localJsxFragmentNamespace) {
                                return file.localJsxFragmentNamespace;
                            }
                            var jsxFragmentPragma = file.pragmas.get("jsxfrag");
                            if (jsxFragmentPragma) {
                                var chosenPragma = ts.isArray(jsxFragmentPragma) ? jsxFragmentPragma[0] : jsxFragmentPragma;
                                file.localJsxFragmentFactory = ts.parseIsolatedEntityName(chosenPragma.arguments.factory, languageVersion);
                                ts.visitNode(file.localJsxFragmentFactory, markAsSynthetic);
                                if (file.localJsxFragmentFactory) {
                                    return file.localJsxFragmentNamespace = ts.getFirstIdentifier(file.localJsxFragmentFactory).escapedText;
                                }
                            }
                            var entity = getJsxFragmentFactoryEntity(location);
                            if (entity) {
                                file.localJsxFragmentFactory = entity;
                                return file.localJsxFragmentNamespace = ts.getFirstIdentifier(entity).escapedText;
                            }
                        }
                        else {
                            if (file.localJsxNamespace) {
                                return file.localJsxNamespace;
                            }
                            var jsxPragma = file.pragmas.get("jsx");
                            if (jsxPragma) {
                                var chosenPragma = ts.isArray(jsxPragma) ? jsxPragma[0] : jsxPragma;
                                file.localJsxFactory = ts.parseIsolatedEntityName(chosenPragma.arguments.factory, languageVersion);
                                ts.visitNode(file.localJsxFactory, markAsSynthetic);
                                if (file.localJsxFactory) {
                                    return file.localJsxNamespace = ts.getFirstIdentifier(file.localJsxFactory).escapedText;
                                }
                            }
                        }
                    }
                }
                if (!_jsxNamespace) {
                    _jsxNamespace = "React";
                    if (compilerOptions.jsxFactory) {
                        _jsxFactoryEntity = ts.parseIsolatedEntityName(compilerOptions.jsxFactory, languageVersion);
                        ts.visitNode(_jsxFactoryEntity, markAsSynthetic);
                        if (_jsxFactoryEntity) {
                            _jsxNamespace = ts.getFirstIdentifier(_jsxFactoryEntity).escapedText;
                        }
                    }
                    else if (compilerOptions.reactNamespace) {
                        _jsxNamespace = ts.escapeLeadingUnderscores(compilerOptions.reactNamespace);
                    }
                }
                if (!_jsxFactoryEntity) {
                    _jsxFactoryEntity = ts.factory.createQualifiedName(ts.factory.createIdentifier(ts.unescapeLeadingUnderscores(_jsxNamespace)), "createElement");
                }
                return _jsxNamespace;
                function markAsSynthetic(node) {
                    ts.setTextRangePosEnd(node, -1, -1);
                    return ts.visitEachChild(node, markAsSynthetic, ts.nullTransformationContext);
                }
            }
            function getEmitResolver(sourceFile, cancellationToken) {
                // Ensure we have all the type information in place for this file so that all the
                // emitter questions of this resolver will return the right information.
                getDiagnostics(sourceFile, cancellationToken);
                return emitResolver;
            }
            function lookupOrIssueError(location, message, arg0, arg1, arg2, arg3) {
                var diagnostic = location
                    ? ts.createDiagnosticForNode(location, message, arg0, arg1, arg2, arg3)
                    : ts.createCompilerDiagnostic(message, arg0, arg1, arg2, arg3);
                var existing = diagnostics.lookup(diagnostic);
                if (existing) {
                    return existing;
                }
                else {
                    diagnostics.add(diagnostic);
                    return diagnostic;
                }
            }
            function errorSkippedOn(key, location, message, arg0, arg1, arg2, arg3) {
                var diagnostic = error(location, message, arg0, arg1, arg2, arg3);
                diagnostic.skippedOn = key;
                return diagnostic;
            }
            function createError(location, message, arg0, arg1, arg2, arg3) {
                return location
                    ? ts.createDiagnosticForNode(location, message, arg0, arg1, arg2, arg3)
                    : ts.createCompilerDiagnostic(message, arg0, arg1, arg2, arg3);
            }
            function error(location, message, arg0, arg1, arg2, arg3) {
                var diagnostic = createError(location, message, arg0, arg1, arg2, arg3);
                diagnostics.add(diagnostic);
                return diagnostic;
            }
            function addErrorOrSuggestion(isError, diagnostic) {
                if (isError) {
                    diagnostics.add(diagnostic);
                }
                else {
                    suggestionDiagnostics.add(__assign(__assign({}, diagnostic), { category: ts.DiagnosticCategory.Suggestion }));
                }
            }
            function errorOrSuggestion(isError, location, message, arg0, arg1, arg2, arg3) {
                // Pseudo-synthesized input node
                if (location.pos < 0 || location.end < 0) {
                    if (!isError) {
                        return; // Drop suggestions (we have no span to suggest on)
                    }
                    // Issue errors globally
                    var file = ts.getSourceFileOfNode(location);
                    addErrorOrSuggestion(isError, "message" in message ? ts.createFileDiagnostic(file, 0, 0, message, arg0, arg1, arg2, arg3) : ts.createDiagnosticForFileFromMessageChain(file, message)); // eslint-disable-line no-in-operator
                    return;
                }
                addErrorOrSuggestion(isError, "message" in message ? ts.createDiagnosticForNode(location, message, arg0, arg1, arg2, arg3) : ts.createDiagnosticForNodeFromMessageChain(location, message)); // eslint-disable-line no-in-operator
            }
            function errorAndMaybeSuggestAwait(location, maybeMissingAwait, message, arg0, arg1, arg2, arg3) {
                var diagnostic = error(location, message, arg0, arg1, arg2, arg3);
                if (maybeMissingAwait) {
                    var related = ts.createDiagnosticForNode(location, ts.Diagnostics.Did_you_forget_to_use_await);
                    ts.addRelatedInfo(diagnostic, related);
                }
                return diagnostic;
            }
            function addDeprecatedSuggestionWorker(declarations, diagnostic) {
                var deprecatedTag = Array.isArray(declarations) ? ts.forEach(declarations, ts.getJSDocDeprecatedTag) : ts.getJSDocDeprecatedTag(declarations);
                if (deprecatedTag) {
                    ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(deprecatedTag, ts.Diagnostics.The_declaration_was_marked_as_deprecated_here));
                }
                // We call `addRelatedInfo()` before adding the diagnostic to prevent duplicates.
                suggestionDiagnostics.add(diagnostic);
                return diagnostic;
            }
            function addDeprecatedSuggestion(location, declarations, deprecatedEntity) {
                var diagnostic = ts.createDiagnosticForNode(location, ts.Diagnostics._0_is_deprecated, deprecatedEntity);
                return addDeprecatedSuggestionWorker(declarations, diagnostic);
            }
            function addDeprecatedSuggestionWithSignature(location, declaration, deprecatedEntity, signatureString) {
                var diagnostic = deprecatedEntity
                    ? ts.createDiagnosticForNode(location, ts.Diagnostics.The_signature_0_of_1_is_deprecated, signatureString, deprecatedEntity)
                    : ts.createDiagnosticForNode(location, ts.Diagnostics._0_is_deprecated, signatureString);
                return addDeprecatedSuggestionWorker(declaration, diagnostic);
            }
            function createSymbol(flags, name, checkFlags) {
                symbolCount++;
                var symbol = new Symbol(flags | 33554432 /* Transient */, name);
                symbol.checkFlags = checkFlags || 0;
                return symbol;
            }
            function getExcludedSymbolFlags(flags) {
                var result = 0;
                if (flags & 2 /* BlockScopedVariable */)
                    result |= 111551 /* BlockScopedVariableExcludes */;
                if (flags & 1 /* FunctionScopedVariable */)
                    result |= 111550 /* FunctionScopedVariableExcludes */;
                if (flags & 4 /* Property */)
                    result |= 0 /* PropertyExcludes */;
                if (flags & 8 /* EnumMember */)
                    result |= 900095 /* EnumMemberExcludes */;
                if (flags & 16 /* Function */)
                    result |= 110991 /* FunctionExcludes */;
                if (flags & 32 /* Class */)
                    result |= 899503 /* ClassExcludes */;
                if (flags & 64 /* Interface */)
                    result |= 788872 /* InterfaceExcludes */;
                if (flags & 256 /* RegularEnum */)
                    result |= 899327 /* RegularEnumExcludes */;
                if (flags & 128 /* ConstEnum */)
                    result |= 899967 /* ConstEnumExcludes */;
                if (flags & 512 /* ValueModule */)
                    result |= 110735 /* ValueModuleExcludes */;
                if (flags & 8192 /* Method */)
                    result |= 103359 /* MethodExcludes */;
                if (flags & 32768 /* GetAccessor */)
                    result |= 46015 /* GetAccessorExcludes */;
                if (flags & 65536 /* SetAccessor */)
                    result |= 78783 /* SetAccessorExcludes */;
                if (flags & 262144 /* TypeParameter */)
                    result |= 526824 /* TypeParameterExcludes */;
                if (flags & 524288 /* TypeAlias */)
                    result |= 788968 /* TypeAliasExcludes */;
                if (flags & 2097152 /* Alias */)
                    result |= 2097152 /* AliasExcludes */;
                return result;
            }
            function recordMergedSymbol(target, source) {
                if (!source.mergeId) {
                    source.mergeId = nextMergeId;
                    nextMergeId++;
                }
                mergedSymbols[source.mergeId] = target;
            }
            function cloneSymbol(symbol) {
                var result = createSymbol(symbol.flags, symbol.escapedName);
                result.declarations = symbol.declarations ? symbol.declarations.slice() : [];
                result.parent = symbol.parent;
                if (symbol.valueDeclaration)
                    result.valueDeclaration = symbol.valueDeclaration;
                if (symbol.constEnumOnlyModule)
                    result.constEnumOnlyModule = true;
                if (symbol.members)
                    result.members = new ts.Map(symbol.members);
                if (symbol.exports)
                    result.exports = new ts.Map(symbol.exports);
                recordMergedSymbol(result, symbol);
                return result;
            }
            /**
             * Note: if target is transient, then it is mutable, and mergeSymbol with both mutate and return it.
             * If target is not transient, mergeSymbol will produce a transient clone, mutate that and return it.
             */
            function mergeSymbol(target, source, unidirectional) {
                if (unidirectional === void 0) {
                    unidirectional = false;
                }
                if (!(target.flags & getExcludedSymbolFlags(source.flags)) ||
                    (source.flags | target.flags) & 67108864 /* Assignment */) {
                    if (source === target) {
                        // This can happen when an export assigned namespace exports something also erroneously exported at the top level
                        // See `declarationFileNoCrashOnExtraExportModifier` for an example
                        return target;
                    }
                    if (!(target.flags & 33554432 /* Transient */)) {
                        var resolvedTarget = resolveSymbol(target);
                        if (resolvedTarget === unknownSymbol) {
                            return source;
                        }
                        target = cloneSymbol(resolvedTarget);
                    }
                    // Javascript static-property-assignment declarations always merge, even though they are also values
                    if (source.flags & 512 /* ValueModule */ && target.flags & 512 /* ValueModule */ && target.constEnumOnlyModule && !source.constEnumOnlyModule) {
                        // reset flag when merging instantiated module into value module that has only const enums
                        target.constEnumOnlyModule = false;
                    }
                    target.flags |= source.flags;
                    if (source.valueDeclaration) {
                        ts.setValueDeclaration(target, source.valueDeclaration);
                    }
                    ts.addRange(target.declarations, source.declarations);
                    if (source.members) {
                        if (!target.members)
                            target.members = ts.createSymbolTable();
                        mergeSymbolTable(target.members, source.members, unidirectional);
                    }
                    if (source.exports) {
                        if (!target.exports)
                            target.exports = ts.createSymbolTable();
                        mergeSymbolTable(target.exports, source.exports, unidirectional);
                    }
                    if (!unidirectional) {
                        recordMergedSymbol(target, source);
                    }
                }
                else if (target.flags & 1024 /* NamespaceModule */) {
                    // Do not report an error when merging `var globalThis` with the built-in `globalThis`,
                    // as we will already report a "Declaration name conflicts..." error, and this error
                    // won't make much sense.
                    if (target !== globalThisSymbol) {
                        error(source.declarations && ts.getNameOfDeclaration(source.declarations[0]), ts.Diagnostics.Cannot_augment_module_0_with_value_exports_because_it_resolves_to_a_non_module_entity, symbolToString(target));
                    }
                }
                else { // error
                    var isEitherEnum = !!(target.flags & 384 /* Enum */ || source.flags & 384 /* Enum */);
                    var isEitherBlockScoped_1 = !!(target.flags & 2 /* BlockScopedVariable */ || source.flags & 2 /* BlockScopedVariable */);
                    var message = isEitherEnum
                        ? ts.Diagnostics.Enum_declarations_can_only_merge_with_namespace_or_other_enum_declarations
                        : isEitherBlockScoped_1
                            ? ts.Diagnostics.Cannot_redeclare_block_scoped_variable_0
                            : ts.Diagnostics.Duplicate_identifier_0;
                    var sourceSymbolFile = source.declarations && ts.getSourceFileOfNode(source.declarations[0]);
                    var targetSymbolFile = target.declarations && ts.getSourceFileOfNode(target.declarations[0]);
                    var symbolName_1 = symbolToString(source);
                    // Collect top-level duplicate identifier errors into one mapping, so we can then merge their diagnostics if there are a bunch
                    if (sourceSymbolFile && targetSymbolFile && amalgamatedDuplicates && !isEitherEnum && sourceSymbolFile !== targetSymbolFile) {
                        var firstFile_1 = ts.comparePaths(sourceSymbolFile.path, targetSymbolFile.path) === -1 /* LessThan */ ? sourceSymbolFile : targetSymbolFile;
                        var secondFile_1 = firstFile_1 === sourceSymbolFile ? targetSymbolFile : sourceSymbolFile;
                        var filesDuplicates = ts.getOrUpdate(amalgamatedDuplicates, firstFile_1.path + "|" + secondFile_1.path, function () {
                            return ({ firstFile: firstFile_1, secondFile: secondFile_1, conflictingSymbols: new ts.Map() });
                        });
                        var conflictingSymbolInfo = ts.getOrUpdate(filesDuplicates.conflictingSymbols, symbolName_1, function () {
                            return ({ isBlockScoped: isEitherBlockScoped_1, firstFileLocations: [], secondFileLocations: [] });
                        });
                        addDuplicateLocations(conflictingSymbolInfo.firstFileLocations, source);
                        addDuplicateLocations(conflictingSymbolInfo.secondFileLocations, target);
                    }
                    else {
                        addDuplicateDeclarationErrorsForSymbols(source, message, symbolName_1, target);
                        addDuplicateDeclarationErrorsForSymbols(target, message, symbolName_1, source);
                    }
                }
                return target;
                function addDuplicateLocations(locs, symbol) {
                    if (symbol.declarations) {
                        for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                            var decl = _a[_i];
                            ts.pushIfUnique(locs, decl);
                        }
                    }
                }
            }
            function addDuplicateDeclarationErrorsForSymbols(target, message, symbolName, source) {
                ts.forEach(target.declarations, function (node) {
                    addDuplicateDeclarationError(node, message, symbolName, source.declarations);
                });
            }
            function addDuplicateDeclarationError(node, message, symbolName, relatedNodes) {
                var errorNode = (ts.getExpandoInitializer(node, /*isPrototypeAssignment*/ false) ? ts.getNameOfExpando(node) : ts.getNameOfDeclaration(node)) || node;
                var err = lookupOrIssueError(errorNode, message, symbolName);
                var _loop_7 = function (relatedNode) {
                    var adjustedNode = (ts.getExpandoInitializer(relatedNode, /*isPrototypeAssignment*/ false) ? ts.getNameOfExpando(relatedNode) : ts.getNameOfDeclaration(relatedNode)) || relatedNode;
                    if (adjustedNode === errorNode)
                        return "continue";
                    err.relatedInformation = err.relatedInformation || [];
                    var leadingMessage = ts.createDiagnosticForNode(adjustedNode, ts.Diagnostics._0_was_also_declared_here, symbolName);
                    var followOnMessage = ts.createDiagnosticForNode(adjustedNode, ts.Diagnostics.and_here);
                    if (ts.length(err.relatedInformation) >= 5 || ts.some(err.relatedInformation, function (r) { return ts.compareDiagnostics(r, followOnMessage) === 0 /* EqualTo */ || ts.compareDiagnostics(r, leadingMessage) === 0 /* EqualTo */; }))
                        return "continue";
                    ts.addRelatedInfo(err, !ts.length(err.relatedInformation) ? leadingMessage : followOnMessage);
                };
                for (var _i = 0, _a = relatedNodes || ts.emptyArray; _i < _a.length; _i++) {
                    var relatedNode = _a[_i];
                    _loop_7(relatedNode);
                }
            }
            function combineSymbolTables(first, second) {
                if (!(first === null || first === void 0 ? void 0 : first.size))
                    return second;
                if (!(second === null || second === void 0 ? void 0 : second.size))
                    return first;
                var combined = ts.createSymbolTable();
                mergeSymbolTable(combined, first);
                mergeSymbolTable(combined, second);
                return combined;
            }
            function mergeSymbolTable(target, source, unidirectional) {
                if (unidirectional === void 0) {
                    unidirectional = false;
                }
                source.forEach(function (sourceSymbol, id) {
                    var targetSymbol = target.get(id);
                    target.set(id, targetSymbol ? mergeSymbol(targetSymbol, sourceSymbol, unidirectional) : sourceSymbol);
                });
            }
            function mergeModuleAugmentation(moduleName) {
                var _a, _b, _c;
                var moduleAugmentation = moduleName.parent;
                if (((_a = moduleAugmentation.symbol.declarations) === null || _a === void 0 ? void 0 : _a[0]) !== moduleAugmentation) {
                    // this is a combined symbol for multiple augmentations within the same file.
                    // its symbol already has accumulated information for all declarations
                    // so we need to add it just once - do the work only for first declaration
                    ts.Debug.assert(moduleAugmentation.symbol.declarations.length > 1);
                    return;
                }
                if (ts.isGlobalScopeAugmentation(moduleAugmentation)) {
                    mergeSymbolTable(globals, moduleAugmentation.symbol.exports);
                }
                else {
                    // find a module that about to be augmented
                    // do not validate names of augmentations that are defined in ambient context
                    var moduleNotFoundError = !(moduleName.parent.parent.flags & 8388608 /* Ambient */)
                        ? ts.Diagnostics.Invalid_module_name_in_augmentation_module_0_cannot_be_found
                        : undefined;
                    var mainModule_1 = resolveExternalModuleNameWorker(moduleName, moduleName, moduleNotFoundError, /*isForAugmentation*/ true);
                    if (!mainModule_1) {
                        return;
                    }
                    // obtain item referenced by 'export='
                    mainModule_1 = resolveExternalModuleSymbol(mainModule_1);
                    if (mainModule_1.flags & 1920 /* Namespace */) {
                        // If we're merging an augmentation to a pattern ambient module, we want to
                        // perform the merge unidirectionally from the augmentation ('a.foo') to
                        // the pattern ('*.foo'), so that 'getMergedSymbol()' on a.foo gives you
                        // all the exports both from the pattern and from the augmentation, but
                        // 'getMergedSymbol()' on *.foo only gives you exports from *.foo.
                        if (ts.some(patternAmbientModules, function (module) { return mainModule_1 === module.symbol; })) {
                            var merged = mergeSymbol(moduleAugmentation.symbol, mainModule_1, /*unidirectional*/ true);
                            if (!patternAmbientModuleAugmentations) {
                                patternAmbientModuleAugmentations = new ts.Map();
                            }
                            // moduleName will be a StringLiteral since this is not `declare global`.
                            patternAmbientModuleAugmentations.set(moduleName.text, merged);
                        }
                        else {
                            if (((_b = mainModule_1.exports) === null || _b === void 0 ? void 0 : _b.get("__export" /* ExportStar */)) && ((_c = moduleAugmentation.symbol.exports) === null || _c === void 0 ? void 0 : _c.size)) {
                                // We may need to merge the module augmentation's exports into the target symbols of the resolved exports
                                var resolvedExports = getResolvedMembersOrExportsOfSymbol(mainModule_1, "resolvedExports" /* resolvedExports */);
                                for (var _i = 0, _d = ts.arrayFrom(moduleAugmentation.symbol.exports.entries()); _i < _d.length; _i++) {
                                    var _e = _d[_i], key = _e[0], value = _e[1];
                                    if (resolvedExports.has(key) && !mainModule_1.exports.has(key)) {
                                        mergeSymbol(resolvedExports.get(key), value);
                                    }
                                }
                            }
                            mergeSymbol(mainModule_1, moduleAugmentation.symbol);
                        }
                    }
                    else {
                        // moduleName will be a StringLiteral since this is not `declare global`.
                        error(moduleName, ts.Diagnostics.Cannot_augment_module_0_because_it_resolves_to_a_non_module_entity, moduleName.text);
                    }
                }
            }
            function addToSymbolTable(target, source, message) {
                source.forEach(function (sourceSymbol, id) {
                    var targetSymbol = target.get(id);
                    if (targetSymbol) {
                        // Error on redeclarations
                        ts.forEach(targetSymbol.declarations, addDeclarationDiagnostic(ts.unescapeLeadingUnderscores(id), message));
                    }
                    else {
                        target.set(id, sourceSymbol);
                    }
                });
                function addDeclarationDiagnostic(id, message) {
                    return function (declaration) { return diagnostics.add(ts.createDiagnosticForNode(declaration, message, id)); };
                }
            }
            function getSymbolLinks(symbol) {
                if (symbol.flags & 33554432 /* Transient */)
                    return symbol;
                var id = getSymbolId(symbol);
                return symbolLinks[id] || (symbolLinks[id] = new SymbolLinks());
            }
            function getNodeLinks(node) {
                var nodeId = getNodeId(node);
                return nodeLinks[nodeId] || (nodeLinks[nodeId] = new NodeLinks());
            }
            function isGlobalSourceFile(node) {
                return node.kind === 300 /* SourceFile */ && !ts.isExternalOrCommonJsModule(node);
            }
            function getSymbol(symbols, name, meaning) {
                if (meaning) {
                    var symbol = getMergedSymbol(symbols.get(name));
                    if (symbol) {
                        ts.Debug.assert((ts.getCheckFlags(symbol) & 1 /* Instantiated */) === 0, "Should never get an instantiated symbol here.");
                        if (symbol.flags & meaning) {
                            return symbol;
                        }
                        if (symbol.flags & 2097152 /* Alias */) {
                            var target = resolveAlias(symbol);
                            // Unknown symbol means an error occurred in alias resolution, treat it as positive answer to avoid cascading errors
                            if (target === unknownSymbol || target.flags & meaning) {
                                return symbol;
                            }
                        }
                    }
                }
                // return undefined if we can't find a symbol.
            }
            /**
             * Get symbols that represent parameter-property-declaration as parameter and as property declaration
             * @param parameter a parameterDeclaration node
             * @param parameterName a name of the parameter to get the symbols for.
             * @return a tuple of two symbols
             */
            function getSymbolsOfParameterPropertyDeclaration(parameter, parameterName) {
                var constructorDeclaration = parameter.parent;
                var classDeclaration = parameter.parent.parent;
                var parameterSymbol = getSymbol(constructorDeclaration.locals, parameterName, 111551 /* Value */);
                var propertySymbol = getSymbol(getMembersOfSymbol(classDeclaration.symbol), parameterName, 111551 /* Value */);
                if (parameterSymbol && propertySymbol) {
                    return [parameterSymbol, propertySymbol];
                }
                return ts.Debug.fail("There should exist two symbols, one as property declaration and one as parameter declaration");
            }
            function isBlockScopedNameDeclaredBeforeUse(declaration, usage) {
                var declarationFile = ts.getSourceFileOfNode(declaration);
                var useFile = ts.getSourceFileOfNode(usage);
                var declContainer = ts.getEnclosingBlockScopeContainer(declaration);
                if (declarationFile !== useFile) {
                    if ((moduleKind && (declarationFile.externalModuleIndicator || useFile.externalModuleIndicator)) ||
                        (!ts.outFile(compilerOptions)) ||
                        isInTypeQuery(usage) ||
                        declaration.flags & 8388608 /* Ambient */) {
                        // nodes are in different files and order cannot be determined
                        return true;
                    }
                    // declaration is after usage
                    // can be legal if usage is deferred (i.e. inside function or in initializer of instance property)
                    if (isUsedInFunctionOrInstanceProperty(usage, declaration)) {
                        return true;
                    }
                    var sourceFiles = host.getSourceFiles();
                    return sourceFiles.indexOf(declarationFile) <= sourceFiles.indexOf(useFile);
                }
                if (declaration.pos <= usage.pos && !(ts.isPropertyDeclaration(declaration) && ts.isThisProperty(usage.parent) && !declaration.initializer && !declaration.exclamationToken)) {
                    // declaration is before usage
                    if (declaration.kind === 201 /* BindingElement */) {
                        // still might be illegal if declaration and usage are both binding elements (eg var [a = b, b = b] = [1, 2])
                        var errorBindingElement = ts.getAncestor(usage, 201 /* BindingElement */);
                        if (errorBindingElement) {
                            return ts.findAncestor(errorBindingElement, ts.isBindingElement) !== ts.findAncestor(declaration, ts.isBindingElement) ||
                                declaration.pos < errorBindingElement.pos;
                        }
                        // or it might be illegal if usage happens before parent variable is declared (eg var [a] = a)
                        return isBlockScopedNameDeclaredBeforeUse(ts.getAncestor(declaration, 252 /* VariableDeclaration */), usage);
                    }
                    else if (declaration.kind === 252 /* VariableDeclaration */) {
                        // still might be illegal if usage is in the initializer of the variable declaration (eg var a = a)
                        return !isImmediatelyUsedInInitializerOfBlockScopedVariable(declaration, usage);
                    }
                    else if (ts.isClassDeclaration(declaration)) {
                        // still might be illegal if the usage is within a computed property name in the class (eg class A { static p = "a"; [A.p]() {} })
                        return !ts.findAncestor(usage, function (n) { return ts.isComputedPropertyName(n) && n.parent.parent === declaration; });
                    }
                    else if (ts.isPropertyDeclaration(declaration)) {
                        // still might be illegal if a self-referencing property initializer (eg private x = this.x)
                        return !isPropertyImmediatelyReferencedWithinDeclaration(declaration, usage, /*stopAtAnyPropertyDeclaration*/ false);
                    }
                    else if (ts.isParameterPropertyDeclaration(declaration, declaration.parent)) {
                        // foo = this.bar is illegal in esnext+useDefineForClassFields when bar is a parameter property
                        return !(compilerOptions.target === 99 /* ESNext */ && useDefineForClassFields
                            && ts.getContainingClass(declaration) === ts.getContainingClass(usage)
                            && isUsedInFunctionOrInstanceProperty(usage, declaration));
                    }
                    return true;
                }
                // declaration is after usage, but it can still be legal if usage is deferred:
                // 1. inside an export specifier
                // 2. inside a function
                // 3. inside an instance property initializer, a reference to a non-instance property
                //    (except when target: "esnext" and useDefineForClassFields: true and the reference is to a parameter property)
                // 4. inside a static property initializer, a reference to a static method in the same class
                // 5. inside a TS export= declaration (since we will move the export statement during emit to avoid TDZ)
                // or if usage is in a type context:
                // 1. inside a type query (typeof in type position)
                // 2. inside a jsdoc comment
                if (usage.parent.kind === 273 /* ExportSpecifier */ || (usage.parent.kind === 269 /* ExportAssignment */ && usage.parent.isExportEquals)) {
                    // export specifiers do not use the variable, they only make it available for use
                    return true;
                }
                // When resolving symbols for exports, the `usage` location passed in can be the export site directly
                if (usage.kind === 269 /* ExportAssignment */ && usage.isExportEquals) {
                    return true;
                }
                if (!!(usage.flags & 4194304 /* JSDoc */) || isInTypeQuery(usage) || usageInTypeDeclaration()) {
                    return true;
                }
                if (isUsedInFunctionOrInstanceProperty(usage, declaration)) {
                    if (compilerOptions.target === 99 /* ESNext */ && useDefineForClassFields
                        && ts.getContainingClass(declaration)
                        && (ts.isPropertyDeclaration(declaration) || ts.isParameterPropertyDeclaration(declaration, declaration.parent))) {
                        return !isPropertyImmediatelyReferencedWithinDeclaration(declaration, usage, /*stopAtAnyPropertyDeclaration*/ true);
                    }
                    else {
                        return true;
                    }
                }
                return false;
                function usageInTypeDeclaration() {
                    return !!ts.findAncestor(usage, function (node) { return ts.isInterfaceDeclaration(node) || ts.isTypeAliasDeclaration(node); });
                }
                function isImmediatelyUsedInInitializerOfBlockScopedVariable(declaration, usage) {
                    switch (declaration.parent.parent.kind) {
                        case 235 /* VariableStatement */:
                        case 240 /* ForStatement */:
                        case 242 /* ForOfStatement */:
                            // variable statement/for/for-of statement case,
                            // use site should not be inside variable declaration (initializer of declaration or binding element)
                            if (isSameScopeDescendentOf(usage, declaration, declContainer)) {
                                return true;
                            }
                            break;
                    }
                    // ForIn/ForOf case - use site should not be used in expression part
                    var grandparent = declaration.parent.parent;
                    return ts.isForInOrOfStatement(grandparent) && isSameScopeDescendentOf(usage, grandparent.expression, declContainer);
                }
                function isUsedInFunctionOrInstanceProperty(usage, declaration) {
                    return !!ts.findAncestor(usage, function (current) {
                        if (current === declContainer) {
                            return "quit";
                        }
                        if (ts.isFunctionLike(current)) {
                            return true;
                        }
                        if (ts.isClassStaticBlockDeclaration(current)) {
                            return declaration.pos < usage.pos;
                        }
                        var propertyDeclaration = ts.tryCast(current.parent, ts.isPropertyDeclaration);
                        if (propertyDeclaration) {
                            var initializerOfProperty = propertyDeclaration.initializer === current;
                            if (initializerOfProperty) {
                                if (ts.isStatic(current.parent)) {
                                    if (declaration.kind === 167 /* MethodDeclaration */) {
                                        return true;
                                    }
                                    if (ts.isPropertyDeclaration(declaration) && ts.getContainingClass(usage) === ts.getContainingClass(declaration)) {
                                        var propName = declaration.name;
                                        if (ts.isIdentifier(propName) || ts.isPrivateIdentifier(propName)) {
                                            var type = getTypeOfSymbol(getSymbolOfNode(declaration));
                                            var staticBlocks = ts.filter(declaration.parent.members, ts.isClassStaticBlockDeclaration);
                                            if (isPropertyInitializedInStaticBlocks(propName, type, staticBlocks, declaration.parent.pos, current.pos)) {
                                                return true;
                                            }
                                        }
                                    }
                                }
                                else {
                                    var isDeclarationInstanceProperty = declaration.kind === 165 /* PropertyDeclaration */ && !ts.isStatic(declaration);
                                    if (!isDeclarationInstanceProperty || ts.getContainingClass(usage) !== ts.getContainingClass(declaration)) {
                                        return true;
                                    }
                                }
                            }
                        }
                        return false;
                    });
                }
                /** stopAtAnyPropertyDeclaration is used for detecting ES-standard class field use-before-def errors */
                function isPropertyImmediatelyReferencedWithinDeclaration(declaration, usage, stopAtAnyPropertyDeclaration) {
                    // always legal if usage is after declaration
                    if (usage.end > declaration.end) {
                        return false;
                    }
                    // still might be legal if usage is deferred (e.g. x: any = () => this.x)
                    // otherwise illegal if immediately referenced within the declaration (e.g. x: any = this.x)
                    var ancestorChangingReferenceScope = ts.findAncestor(usage, function (node) {
                        if (node === declaration) {
                            return "quit";
                        }
                        switch (node.kind) {
                            case 212 /* ArrowFunction */:
                                return true;
                            case 165 /* PropertyDeclaration */:
                                // even when stopping at any property declaration, they need to come from the same class
                                return stopAtAnyPropertyDeclaration &&
                                    (ts.isPropertyDeclaration(declaration) && node.parent === declaration.parent
                                        || ts.isParameterPropertyDeclaration(declaration, declaration.parent) && node.parent === declaration.parent.parent)
                                    ? "quit" : true;
                            case 233 /* Block */:
                                switch (node.parent.kind) {
                                    case 170 /* GetAccessor */:
                                    case 167 /* MethodDeclaration */:
                                    case 171 /* SetAccessor */:
                                        return true;
                                    default:
                                        return false;
                                }
                            default:
                                return false;
                        }
                    });
                    return ancestorChangingReferenceScope === undefined;
                }
            }
            function useOuterVariableScopeInParameter(result, location, lastLocation) {
                var target = ts.getEmitScriptTarget(compilerOptions);
                var functionLocation = location;
                if (ts.isParameter(lastLocation)
                    && functionLocation.body
                    && result.valueDeclaration
                    && result.valueDeclaration.pos >= functionLocation.body.pos
                    && result.valueDeclaration.end <= functionLocation.body.end) {
                    // check for several cases where we introduce temporaries that require moving the name/initializer of the parameter to the body
                    // - static field in a class expression
                    // - optional chaining pre-es2020
                    // - nullish coalesce pre-es2020
                    // - spread assignment in binding pattern pre-es2017
                    if (target >= 2 /* ES2015 */) {
                        var links = getNodeLinks(functionLocation);
                        if (links.declarationRequiresScopeChange === undefined) {
                            links.declarationRequiresScopeChange = ts.forEach(functionLocation.parameters, requiresScopeChange) || false;
                        }
                        return !links.declarationRequiresScopeChange;
                    }
                }
                return false;
                function requiresScopeChange(node) {
                    return requiresScopeChangeWorker(node.name)
                        || !!node.initializer && requiresScopeChangeWorker(node.initializer);
                }
                function requiresScopeChangeWorker(node) {
                    switch (node.kind) {
                        case 212 /* ArrowFunction */:
                        case 211 /* FunctionExpression */:
                        case 254 /* FunctionDeclaration */:
                        case 169 /* Constructor */:
                            // do not descend into these
                            return false;
                        case 167 /* MethodDeclaration */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                        case 291 /* PropertyAssignment */:
                            return requiresScopeChangeWorker(node.name);
                        case 165 /* PropertyDeclaration */:
                            // static properties in classes introduce temporary variables
                            if (ts.hasStaticModifier(node)) {
                                return target < 99 /* ESNext */ || !useDefineForClassFields;
                            }
                            return requiresScopeChangeWorker(node.name);
                        default:
                            // null coalesce and optional chain pre-es2020 produce temporary variables
                            if (ts.isNullishCoalesce(node) || ts.isOptionalChain(node)) {
                                return target < 7 /* ES2020 */;
                            }
                            if (ts.isBindingElement(node) && node.dotDotDotToken && ts.isObjectBindingPattern(node.parent)) {
                                return target < 4 /* ES2017 */;
                            }
                            if (ts.isTypeNode(node))
                                return false;
                            return ts.forEachChild(node, requiresScopeChangeWorker) || false;
                    }
                }
            }
            /**
             * Resolve a given name for a given meaning at a given location. An error is reported if the name was not found and
             * the nameNotFoundMessage argument is not undefined. Returns the resolved symbol, or undefined if no symbol with
             * the given name can be found.
             *
             * @param isUse If true, this will count towards --noUnusedLocals / --noUnusedParameters.
             */
            function resolveName(location, name, meaning, nameNotFoundMessage, nameArg, isUse, excludeGlobals, issueSuggestions) {
                if (excludeGlobals === void 0) {
                    excludeGlobals = false;
                }
                return resolveNameHelper(location, name, meaning, nameNotFoundMessage, nameArg, isUse, excludeGlobals, getSymbol, issueSuggestions);
            }
            function resolveNameHelper(location, name, meaning, nameNotFoundMessage, nameArg, isUse, excludeGlobals, lookup, issueSuggestions) {
                var _a;
                var originalLocation = location; // needed for did-you-mean error reporting, which gathers candidates starting from the original location
                var result;
                var lastLocation;
                var lastSelfReferenceLocation;
                var propertyWithInvalidInitializer;
                var associatedDeclarationForContainingInitializerOrBindingName;
                var withinDeferredContext = false;
                var errorLocation = location;
                var grandparent;
                var isInExternalModule = false;
                loop: while (location) {
                    // Locals of a source file are not in scope (because they get merged into the global symbol table)
                    if (location.locals && !isGlobalSourceFile(location)) {
                        if (result = lookup(location.locals, name, meaning)) {
                            var useResult = true;
                            if (ts.isFunctionLike(location) && lastLocation && lastLocation !== location.body) {
                                // symbol lookup restrictions for function-like declarations
                                // - Type parameters of a function are in scope in the entire function declaration, including the parameter
                                //   list and return type. However, local types are only in scope in the function body.
                                // - parameters are only in the scope of function body
                                // This restriction does not apply to JSDoc comment types because they are parented
                                // at a higher level than type parameters would normally be
                                if (meaning & result.flags & 788968 /* Type */ && lastLocation.kind !== 315 /* JSDocComment */) {
                                    useResult = result.flags & 262144 /* TypeParameter */
                                        // type parameters are visible in parameter list, return type and type parameter list
                                        ? lastLocation === location.type ||
                                            lastLocation.kind === 162 /* Parameter */ ||
                                            lastLocation.kind === 161 /* TypeParameter */
                                        // local types not visible outside the function body
                                        : false;
                                }
                                if (meaning & result.flags & 3 /* Variable */) {
                                    // expression inside parameter will lookup as normal variable scope when targeting es2015+
                                    if (useOuterVariableScopeInParameter(result, location, lastLocation)) {
                                        useResult = false;
                                    }
                                    else if (result.flags & 1 /* FunctionScopedVariable */) {
                                        // parameters are visible only inside function body, parameter list and return type
                                        // technically for parameter list case here we might mix parameters and variables declared in function,
                                        // however it is detected separately when checking initializers of parameters
                                        // to make sure that they reference no variables declared after them.
                                        useResult =
                                            lastLocation.kind === 162 /* Parameter */ ||
                                                (lastLocation === location.type &&
                                                    !!ts.findAncestor(result.valueDeclaration, ts.isParameter));
                                    }
                                }
                            }
                            else if (location.kind === 187 /* ConditionalType */) {
                                // A type parameter declared using 'infer T' in a conditional type is visible only in
                                // the true branch of the conditional type.
                                useResult = lastLocation === location.trueType;
                            }
                            if (useResult) {
                                break loop;
                            }
                            else {
                                result = undefined;
                            }
                        }
                    }
                    withinDeferredContext = withinDeferredContext || getIsDeferredContext(location, lastLocation);
                    switch (location.kind) {
                        case 300 /* SourceFile */:
                            if (!ts.isExternalOrCommonJsModule(location))
                                break;
                            isInExternalModule = true;
                        // falls through
                        case 259 /* ModuleDeclaration */:
                            var moduleExports = getSymbolOfNode(location).exports || emptySymbols;
                            if (location.kind === 300 /* SourceFile */ || (ts.isModuleDeclaration(location) && location.flags & 8388608 /* Ambient */ && !ts.isGlobalScopeAugmentation(location))) {
                                // It's an external module. First see if the module has an export default and if the local
                                // name of that export default matches.
                                if (result = moduleExports.get("default" /* Default */)) {
                                    var localSymbol = ts.getLocalSymbolForExportDefault(result);
                                    if (localSymbol && (result.flags & meaning) && localSymbol.escapedName === name) {
                                        break loop;
                                    }
                                    result = undefined;
                                }
                                // Because of module/namespace merging, a module's exports are in scope,
                                // yet we never want to treat an export specifier as putting a member in scope.
                                // Therefore, if the name we find is purely an export specifier, it is not actually considered in scope.
                                // Two things to note about this:
                                //     1. We have to check this without calling getSymbol. The problem with calling getSymbol
                                //        on an export specifier is that it might find the export specifier itself, and try to
                                //        resolve it as an alias. This will cause the checker to consider the export specifier
                                //        a circular alias reference when it might not be.
                                //     2. We check === SymbolFlags.Alias in order to check that the symbol is *purely*
                                //        an alias. If we used &, we'd be throwing out symbols that have non alias aspects,
                                //        which is not the desired behavior.
                                var moduleExport = moduleExports.get(name);
                                if (moduleExport &&
                                    moduleExport.flags === 2097152 /* Alias */ &&
                                    (ts.getDeclarationOfKind(moduleExport, 273 /* ExportSpecifier */) || ts.getDeclarationOfKind(moduleExport, 272 /* NamespaceExport */))) {
                                    break;
                                }
                            }
                            // ES6 exports are also visible locally (except for 'default'), but commonjs exports are not (except typedefs)
                            if (name !== "default" /* Default */ && (result = lookup(moduleExports, name, meaning & 2623475 /* ModuleMember */))) {
                                if (ts.isSourceFile(location) && location.commonJsModuleIndicator && !((_a = result.declarations) === null || _a === void 0 ? void 0 : _a.some(ts.isJSDocTypeAlias))) {
                                    result = undefined;
                                }
                                else {
                                    break loop;
                                }
                            }
                            break;
                        case 258 /* EnumDeclaration */:
                            if (result = lookup(getSymbolOfNode(location).exports, name, meaning & 8 /* EnumMember */)) {
                                break loop;
                            }
                            break;
                        case 165 /* PropertyDeclaration */:
                            // TypeScript 1.0 spec (April 2014): 8.4.1
                            // Initializer expressions for instance member variables are evaluated in the scope
                            // of the class constructor body but are not permitted to reference parameters or
                            // local variables of the constructor. This effectively means that entities from outer scopes
                            // by the same name as a constructor parameter or local variable are inaccessible
                            // in initializer expressions for instance member variables.
                            if (!ts.isStatic(location)) {
                                var ctor = findConstructorDeclaration(location.parent);
                                if (ctor && ctor.locals) {
                                    if (lookup(ctor.locals, name, meaning & 111551 /* Value */)) {
                                        // Remember the property node, it will be used later to report appropriate error
                                        propertyWithInvalidInitializer = location;
                                    }
                                }
                            }
                            break;
                        case 255 /* ClassDeclaration */:
                        case 224 /* ClassExpression */:
                        case 256 /* InterfaceDeclaration */:
                            // The below is used to lookup type parameters within a class or interface, as they are added to the class/interface locals
                            // These can never be latebound, so the symbol's raw members are sufficient. `getMembersOfNode` cannot be used, as it would
                            // trigger resolving late-bound names, which we may already be in the process of doing while we're here!
                            if (result = lookup(getSymbolOfNode(location).members || emptySymbols, name, meaning & 788968 /* Type */)) {
                                if (!isTypeParameterSymbolDeclaredInContainer(result, location)) {
                                    // ignore type parameters not declared in this container
                                    result = undefined;
                                    break;
                                }
                                if (lastLocation && ts.isStatic(lastLocation)) {
                                    // TypeScript 1.0 spec (April 2014): 3.4.1
                                    // The scope of a type parameter extends over the entire declaration with which the type
                                    // parameter list is associated, with the exception of static member declarations in classes.
                                    error(errorLocation, ts.Diagnostics.Static_members_cannot_reference_class_type_parameters);
                                    return undefined;
                                }
                                break loop;
                            }
                            if (location.kind === 224 /* ClassExpression */ && meaning & 32 /* Class */) {
                                var className = location.name;
                                if (className && name === className.escapedText) {
                                    result = location.symbol;
                                    break loop;
                                }
                            }
                            break;
                        case 226 /* ExpressionWithTypeArguments */:
                            // The type parameters of a class are not in scope in the base class expression.
                            if (lastLocation === location.expression && location.parent.token === 94 /* ExtendsKeyword */) {
                                var container = location.parent.parent;
                                if (ts.isClassLike(container) && (result = lookup(getSymbolOfNode(container).members, name, meaning & 788968 /* Type */))) {
                                    if (nameNotFoundMessage) {
                                        error(errorLocation, ts.Diagnostics.Base_class_expressions_cannot_reference_class_type_parameters);
                                    }
                                    return undefined;
                                }
                            }
                            break;
                        // It is not legal to reference a class's own type parameters from a computed property name that
                        // belongs to the class. For example:
                        //
                        //   function foo<T>() { return '' }
                        //   class C<T> { // <-- Class's own type parameter T
                        //       [foo<T>()]() { } // <-- Reference to T from class's own computed property
                        //   }
                        //
                        case 160 /* ComputedPropertyName */:
                            grandparent = location.parent.parent;
                            if (ts.isClassLike(grandparent) || grandparent.kind === 256 /* InterfaceDeclaration */) {
                                // A reference to this grandparent's type parameters would be an error
                                if (result = lookup(getSymbolOfNode(grandparent).members, name, meaning & 788968 /* Type */)) {
                                    error(errorLocation, ts.Diagnostics.A_computed_property_name_cannot_reference_a_type_parameter_from_its_containing_type);
                                    return undefined;
                                }
                            }
                            break;
                        case 212 /* ArrowFunction */:
                            // when targeting ES6 or higher there is no 'arguments' in an arrow function
                            // for lower compile targets the resolved symbol is used to emit an error
                            if (compilerOptions.target >= 2 /* ES2015 */) {
                                break;
                            }
                        // falls through
                        case 167 /* MethodDeclaration */:
                        case 169 /* Constructor */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                        case 254 /* FunctionDeclaration */:
                            if (meaning & 3 /* Variable */ && name === "arguments") {
                                result = argumentsSymbol;
                                break loop;
                            }
                            break;
                        case 211 /* FunctionExpression */:
                            if (meaning & 3 /* Variable */ && name === "arguments") {
                                result = argumentsSymbol;
                                break loop;
                            }
                            if (meaning & 16 /* Function */) {
                                var functionName = location.name;
                                if (functionName && name === functionName.escapedText) {
                                    result = location.symbol;
                                    break loop;
                                }
                            }
                            break;
                        case 163 /* Decorator */:
                            // Decorators are resolved at the class declaration. Resolving at the parameter
                            // or member would result in looking up locals in the method.
                            //
                            //   function y() {}
                            //   class C {
                            //       method(@y x, y) {} // <-- decorator y should be resolved at the class declaration, not the parameter.
                            //   }
                            //
                            if (location.parent && location.parent.kind === 162 /* Parameter */) {
                                location = location.parent;
                            }
                            //
                            //   function y() {}
                            //   class C {
                            //       @y method(x, y) {} // <-- decorator y should be resolved at the class declaration, not the method.
                            //   }
                            //
                            // class Decorators are resolved outside of the class to avoid referencing type parameters of that class.
                            //
                            //   type T = number;
                            //   declare function y(x: T): any;
                            //   @param(1 as T) // <-- T should resolve to the type alias outside of class C
                            //   class C<T> {}
                            if (location.parent && (ts.isClassElement(location.parent) || location.parent.kind === 255 /* ClassDeclaration */)) {
                                location = location.parent;
                            }
                            break;
                        case 340 /* JSDocTypedefTag */:
                        case 333 /* JSDocCallbackTag */:
                        case 334 /* JSDocEnumTag */:
                            // js type aliases do not resolve names from their host, so skip past it
                            var root = ts.getJSDocRoot(location);
                            if (root) {
                                location = root.parent;
                            }
                            break;
                        case 162 /* Parameter */:
                            if (lastLocation && (lastLocation === location.initializer ||
                                lastLocation === location.name && ts.isBindingPattern(lastLocation))) {
                                if (!associatedDeclarationForContainingInitializerOrBindingName) {
                                    associatedDeclarationForContainingInitializerOrBindingName = location;
                                }
                            }
                            break;
                        case 201 /* BindingElement */:
                            if (lastLocation && (lastLocation === location.initializer ||
                                lastLocation === location.name && ts.isBindingPattern(lastLocation))) {
                                if (ts.isParameterDeclaration(location) && !associatedDeclarationForContainingInitializerOrBindingName) {
                                    associatedDeclarationForContainingInitializerOrBindingName = location;
                                }
                            }
                            break;
                        case 188 /* InferType */:
                            if (meaning & 262144 /* TypeParameter */) {
                                var parameterName = location.typeParameter.name;
                                if (parameterName && name === parameterName.escapedText) {
                                    result = location.typeParameter.symbol;
                                    break loop;
                                }
                            }
                            break;
                    }
                    if (isSelfReferenceLocation(location)) {
                        lastSelfReferenceLocation = location;
                    }
                    lastLocation = location;
                    location = location.parent;
                }
                // We just climbed up parents looking for the name, meaning that we started in a descendant node of `lastLocation`.
                // If `result === lastSelfReferenceLocation.symbol`, that means that we are somewhere inside `lastSelfReferenceLocation` looking up a name, and resolving to `lastLocation` itself.
                // That means that this is a self-reference of `lastLocation`, and shouldn't count this when considering whether `lastLocation` is used.
                if (isUse && result && (!lastSelfReferenceLocation || result !== lastSelfReferenceLocation.symbol)) {
                    result.isReferenced |= meaning;
                }
                if (!result) {
                    if (lastLocation) {
                        ts.Debug.assert(lastLocation.kind === 300 /* SourceFile */);
                        if (lastLocation.commonJsModuleIndicator && name === "exports" && meaning & lastLocation.symbol.flags) {
                            return lastLocation.symbol;
                        }
                    }
                    if (!excludeGlobals) {
                        result = lookup(globals, name, meaning);
                    }
                }
                if (!result) {
                    if (originalLocation && ts.isInJSFile(originalLocation) && originalLocation.parent) {
                        if (ts.isRequireCall(originalLocation.parent, /*checkArgumentIsStringLiteralLike*/ false)) {
                            return requireSymbol;
                        }
                    }
                }
                if (!result) {
                    if (nameNotFoundMessage) {
                        if (!errorLocation ||
                            !checkAndReportErrorForMissingPrefix(errorLocation, name, nameArg) && // TODO: GH#18217
                                !checkAndReportErrorForExtendingInterface(errorLocation) &&
                                !checkAndReportErrorForUsingTypeAsNamespace(errorLocation, name, meaning) &&
                                !checkAndReportErrorForExportingPrimitiveType(errorLocation, name) &&
                                !checkAndReportErrorForUsingTypeAsValue(errorLocation, name, meaning) &&
                                !checkAndReportErrorForUsingNamespaceModuleAsValue(errorLocation, name, meaning) &&
                                !checkAndReportErrorForUsingValueAsType(errorLocation, name, meaning)) {
                            var suggestion = void 0;
                            if (issueSuggestions && suggestionCount < maximumSuggestionCount) {
                                suggestion = getSuggestedSymbolForNonexistentSymbol(originalLocation, name, meaning);
                                var isGlobalScopeAugmentationDeclaration = (suggestion === null || suggestion === void 0 ? void 0 : suggestion.valueDeclaration) && ts.isAmbientModule(suggestion.valueDeclaration) && ts.isGlobalScopeAugmentation(suggestion.valueDeclaration);
                                if (isGlobalScopeAugmentationDeclaration) {
                                    suggestion = undefined;
                                }
                                if (suggestion) {
                                    var suggestionName = symbolToString(suggestion);
                                    var isUncheckedJS = isUncheckedJSSuggestion(originalLocation, suggestion, /*excludeClasses*/ false);
                                    var message = isUncheckedJS ? ts.Diagnostics.Could_not_find_name_0_Did_you_mean_1 : ts.Diagnostics.Cannot_find_name_0_Did_you_mean_1;
                                    var diagnostic = createError(errorLocation, message, diagnosticName(nameArg), suggestionName);
                                    addErrorOrSuggestion(!isUncheckedJS, diagnostic);
                                    if (suggestion.valueDeclaration) {
                                        ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(suggestion.valueDeclaration, ts.Diagnostics._0_is_declared_here, suggestionName));
                                    }
                                }
                            }
                            if (!suggestion) {
                                if (nameArg) {
                                    var lib = getSuggestedLibForNonExistentName(nameArg);
                                    if (lib) {
                                        error(errorLocation, nameNotFoundMessage, diagnosticName(nameArg), lib);
                                    }
                                    else {
                                        error(errorLocation, nameNotFoundMessage, diagnosticName(nameArg));
                                    }
                                }
                            }
                            suggestionCount++;
                        }
                    }
                    return undefined;
                }
                // Perform extra checks only if error reporting was requested
                if (nameNotFoundMessage) {
                    if (propertyWithInvalidInitializer && !(compilerOptions.target === 99 /* ESNext */ && useDefineForClassFields)) {
                        // We have a match, but the reference occurred within a property initializer and the identifier also binds
                        // to a local variable in the constructor where the code will be emitted. Note that this is actually allowed
                        // with ESNext+useDefineForClassFields because the scope semantics are different.
                        var propertyName = propertyWithInvalidInitializer.name;
                        error(errorLocation, ts.Diagnostics.Initializer_of_instance_member_variable_0_cannot_reference_identifier_1_declared_in_the_constructor, ts.declarationNameToString(propertyName), diagnosticName(nameArg));
                        return undefined;
                    }
                    // Only check for block-scoped variable if we have an error location and are looking for the
                    // name with variable meaning
                    //      For example,
                    //          declare module foo {
                    //              interface bar {}
                    //          }
                    //      const foo/*1*/: foo/*2*/.bar;
                    // The foo at /*1*/ and /*2*/ will share same symbol with two meanings:
                    // block-scoped variable and namespace module. However, only when we
                    // try to resolve name in /*1*/ which is used in variable position,
                    // we want to check for block-scoped
                    if (errorLocation &&
                        (meaning & 2 /* BlockScopedVariable */ ||
                            ((meaning & 32 /* Class */ || meaning & 384 /* Enum */) && (meaning & 111551 /* Value */) === 111551 /* Value */))) {
                        var exportOrLocalSymbol = getExportSymbolOfValueSymbolIfExported(result);
                        if (exportOrLocalSymbol.flags & 2 /* BlockScopedVariable */ || exportOrLocalSymbol.flags & 32 /* Class */ || exportOrLocalSymbol.flags & 384 /* Enum */) {
                            checkResolvedBlockScopedVariable(exportOrLocalSymbol, errorLocation);
                        }
                    }
                    // If we're in an external module, we can't reference value symbols created from UMD export declarations
                    if (result && isInExternalModule && (meaning & 111551 /* Value */) === 111551 /* Value */ && !(originalLocation.flags & 4194304 /* JSDoc */)) {
                        var merged = getMergedSymbol(result);
                        if (ts.length(merged.declarations) && ts.every(merged.declarations, function (d) { return ts.isNamespaceExportDeclaration(d) || ts.isSourceFile(d) && !!d.symbol.globalExports; })) {
                            errorOrSuggestion(!compilerOptions.allowUmdGlobalAccess, errorLocation, ts.Diagnostics._0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead, ts.unescapeLeadingUnderscores(name));
                        }
                    }
                    // If we're in a parameter initializer or binding name, we can't reference the values of the parameter whose initializer we're within or parameters to the right
                    if (result && associatedDeclarationForContainingInitializerOrBindingName && !withinDeferredContext && (meaning & 111551 /* Value */) === 111551 /* Value */) {
                        var candidate = getMergedSymbol(getLateBoundSymbol(result));
                        var root = ts.getRootDeclaration(associatedDeclarationForContainingInitializerOrBindingName);
                        // A parameter initializer or binding pattern initializer within a parameter cannot refer to itself
                        if (candidate === getSymbolOfNode(associatedDeclarationForContainingInitializerOrBindingName)) {
                            error(errorLocation, ts.Diagnostics.Parameter_0_cannot_reference_itself, ts.declarationNameToString(associatedDeclarationForContainingInitializerOrBindingName.name));
                        }
                        // And it cannot refer to any declarations which come after it
                        else if (candidate.valueDeclaration && candidate.valueDeclaration.pos > associatedDeclarationForContainingInitializerOrBindingName.pos && root.parent.locals && lookup(root.parent.locals, candidate.escapedName, meaning) === candidate) {
                            error(errorLocation, ts.Diagnostics.Parameter_0_cannot_reference_identifier_1_declared_after_it, ts.declarationNameToString(associatedDeclarationForContainingInitializerOrBindingName.name), ts.declarationNameToString(errorLocation));
                        }
                    }
                    if (result && errorLocation && meaning & 111551 /* Value */ && result.flags & 2097152 /* Alias */) {
                        checkSymbolUsageInExpressionContext(result, name, errorLocation);
                    }
                }
                return result;
            }
            function checkSymbolUsageInExpressionContext(symbol, name, useSite) {
                if (!ts.isValidTypeOnlyAliasUseSite(useSite)) {
                    var typeOnlyDeclaration = getTypeOnlyAliasDeclaration(symbol);
                    if (typeOnlyDeclaration) {
                        var isExport = ts.typeOnlyDeclarationIsExport(typeOnlyDeclaration);
                        var message = isExport
                            ? ts.Diagnostics._0_cannot_be_used_as_a_value_because_it_was_exported_using_export_type
                            : ts.Diagnostics._0_cannot_be_used_as_a_value_because_it_was_imported_using_import_type;
                        var relatedMessage = isExport
                            ? ts.Diagnostics._0_was_exported_here
                            : ts.Diagnostics._0_was_imported_here;
                        var unescapedName = ts.unescapeLeadingUnderscores(name);
                        ts.addRelatedInfo(error(useSite, message, unescapedName), ts.createDiagnosticForNode(typeOnlyDeclaration, relatedMessage, unescapedName));
                    }
                }
            }
            function getIsDeferredContext(location, lastLocation) {
                if (location.kind !== 212 /* ArrowFunction */ && location.kind !== 211 /* FunctionExpression */) {
                    // initializers in instance property declaration of class like entities are executed in constructor and thus deferred
                    return ts.isTypeQueryNode(location) || ((ts.isFunctionLikeDeclaration(location) ||
                        (location.kind === 165 /* PropertyDeclaration */ && !ts.isStatic(location))) && (!lastLocation || lastLocation !== location.name)); // A name is evaluated within the enclosing scope - so it shouldn't count as deferred
                }
                if (lastLocation && lastLocation === location.name) {
                    return false;
                }
                // generator functions and async functions are not inlined in control flow when immediately invoked
                if (location.asteriskToken || ts.hasSyntacticModifier(location, 256 /* Async */)) {
                    return true;
                }
                return !ts.getImmediatelyInvokedFunctionExpression(location);
            }
            function isSelfReferenceLocation(node) {
                switch (node.kind) {
                    case 254 /* FunctionDeclaration */:
                    case 255 /* ClassDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                    case 259 /* ModuleDeclaration */: // For `namespace N { N; }`
                        return true;
                    default:
                        return false;
                }
            }
            function diagnosticName(nameArg) {
                return ts.isString(nameArg) ? ts.unescapeLeadingUnderscores(nameArg) : ts.declarationNameToString(nameArg);
            }
            function isTypeParameterSymbolDeclaredInContainer(symbol, container) {
                if (symbol.declarations) {
                    for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                        var decl = _a[_i];
                        if (decl.kind === 161 /* TypeParameter */) {
                            var parent = ts.isJSDocTemplateTag(decl.parent) ? ts.getJSDocHost(decl.parent) : decl.parent;
                            if (parent === container) {
                                return !(ts.isJSDocTemplateTag(decl.parent) && ts.find(decl.parent.parent.tags, ts.isJSDocTypeAlias)); // TODO: GH#18217
                            }
                        }
                    }
                }
                return false;
            }
            function checkAndReportErrorForMissingPrefix(errorLocation, name, nameArg) {
                if (!ts.isIdentifier(errorLocation) || errorLocation.escapedText !== name || isTypeReferenceIdentifier(errorLocation) || isInTypeQuery(errorLocation)) {
                    return false;
                }
                var container = ts.getThisContainer(errorLocation, /*includeArrowFunctions*/ false);
                var location = container;
                while (location) {
                    if (ts.isClassLike(location.parent)) {
                        var classSymbol = getSymbolOfNode(location.parent);
                        if (!classSymbol) {
                            break;
                        }
                        // Check to see if a static member exists.
                        var constructorType = getTypeOfSymbol(classSymbol);
                        if (getPropertyOfType(constructorType, name)) {
                            error(errorLocation, ts.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0, diagnosticName(nameArg), symbolToString(classSymbol));
                            return true;
                        }
                        // No static member is present.
                        // Check if we're in an instance method and look for a relevant instance member.
                        if (location === container && !ts.isStatic(location)) {
                            var instanceType = getDeclaredTypeOfSymbol(classSymbol).thisType; // TODO: GH#18217
                            if (getPropertyOfType(instanceType, name)) {
                                error(errorLocation, ts.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0, diagnosticName(nameArg));
                                return true;
                            }
                        }
                    }
                    location = location.parent;
                }
                return false;
            }
            function checkAndReportErrorForExtendingInterface(errorLocation) {
                var expression = getEntityNameForExtendingInterface(errorLocation);
                if (expression && resolveEntityName(expression, 64 /* Interface */, /*ignoreErrors*/ true)) {
                    error(errorLocation, ts.Diagnostics.Cannot_extend_an_interface_0_Did_you_mean_implements, ts.getTextOfNode(expression));
                    return true;
                }
                return false;
            }
            /**
             * Climbs up parents to an ExpressionWithTypeArguments, and returns its expression,
             * but returns undefined if that expression is not an EntityNameExpression.
             */
            function getEntityNameForExtendingInterface(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                    case 204 /* PropertyAccessExpression */:
                        return node.parent ? getEntityNameForExtendingInterface(node.parent) : undefined;
                    case 226 /* ExpressionWithTypeArguments */:
                        if (ts.isEntityNameExpression(node.expression)) {
                            return node.expression;
                        }
                    // falls through
                    default:
                        return undefined;
                }
            }
            function checkAndReportErrorForUsingTypeAsNamespace(errorLocation, name, meaning) {
                var namespaceMeaning = 1920 /* Namespace */ | (ts.isInJSFile(errorLocation) ? 111551 /* Value */ : 0);
                if (meaning === namespaceMeaning) {
                    var symbol = resolveSymbol(resolveName(errorLocation, name, 788968 /* Type */ & ~namespaceMeaning, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false));
                    var parent = errorLocation.parent;
                    if (symbol) {
                        if (ts.isQualifiedName(parent)) {
                            ts.Debug.assert(parent.left === errorLocation, "Should only be resolving left side of qualified name as a namespace");
                            var propName = parent.right.escapedText;
                            var propType = getPropertyOfType(getDeclaredTypeOfSymbol(symbol), propName);
                            if (propType) {
                                error(parent, ts.Diagnostics.Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_property_1_in_0_with_0_1, ts.unescapeLeadingUnderscores(name), ts.unescapeLeadingUnderscores(propName));
                                return true;
                            }
                        }
                        error(errorLocation, ts.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_namespace_here, ts.unescapeLeadingUnderscores(name));
                        return true;
                    }
                }
                return false;
            }
            function checkAndReportErrorForUsingValueAsType(errorLocation, name, meaning) {
                if (meaning & (788968 /* Type */ & ~1920 /* Namespace */)) {
                    var symbol = resolveSymbol(resolveName(errorLocation, name, ~788968 /* Type */ & 111551 /* Value */, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false));
                    if (symbol && !(symbol.flags & 1920 /* Namespace */)) {
                        error(errorLocation, ts.Diagnostics._0_refers_to_a_value_but_is_being_used_as_a_type_here_Did_you_mean_typeof_0, ts.unescapeLeadingUnderscores(name));
                        return true;
                    }
                }
                return false;
            }
            function isPrimitiveTypeName(name) {
                return name === "any" || name === "string" || name === "number" || name === "boolean" || name === "never" || name === "unknown";
            }
            function checkAndReportErrorForExportingPrimitiveType(errorLocation, name) {
                if (isPrimitiveTypeName(name) && errorLocation.parent.kind === 273 /* ExportSpecifier */) {
                    error(errorLocation, ts.Diagnostics.Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module, name);
                    return true;
                }
                return false;
            }
            function checkAndReportErrorForUsingTypeAsValue(errorLocation, name, meaning) {
                if (meaning & (111551 /* Value */ & ~1024 /* NamespaceModule */)) {
                    if (isPrimitiveTypeName(name)) {
                        error(errorLocation, ts.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here, ts.unescapeLeadingUnderscores(name));
                        return true;
                    }
                    var symbol = resolveSymbol(resolveName(errorLocation, name, 788968 /* Type */ & ~111551 /* Value */, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false));
                    if (symbol && !(symbol.flags & 1024 /* NamespaceModule */)) {
                        var rawName = ts.unescapeLeadingUnderscores(name);
                        if (isES2015OrLaterConstructorName(name)) {
                            error(errorLocation, ts.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_es2015_or_later, rawName);
                        }
                        else if (maybeMappedType(errorLocation, symbol)) {
                            error(errorLocation, ts.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Did_you_mean_to_use_1_in_0, rawName, rawName === "K" ? "P" : "K");
                        }
                        else {
                            error(errorLocation, ts.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here, rawName);
                        }
                        return true;
                    }
                }
                return false;
            }
            function maybeMappedType(node, symbol) {
                var container = ts.findAncestor(node.parent, function (n) {
                    return ts.isComputedPropertyName(n) || ts.isPropertySignature(n) ? false : ts.isTypeLiteralNode(n) || "quit";
                });
                if (container && container.members.length === 1) {
                    var type = getDeclaredTypeOfSymbol(symbol);
                    return !!(type.flags & 1048576 /* Union */) && allTypesAssignableToKind(type, 384 /* StringOrNumberLiteral */, /*strict*/ true);
                }
                return false;
            }
            function isES2015OrLaterConstructorName(n) {
                switch (n) {
                    case "Promise":
                    case "Symbol":
                    case "Map":
                    case "WeakMap":
                    case "Set":
                    case "WeakSet":
                        return true;
                }
                return false;
            }
            function checkAndReportErrorForUsingNamespaceModuleAsValue(errorLocation, name, meaning) {
                if (meaning & (111551 /* Value */ & ~1024 /* NamespaceModule */ & ~788968 /* Type */)) {
                    var symbol = resolveSymbol(resolveName(errorLocation, name, 1024 /* NamespaceModule */ & ~111551 /* Value */, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false));
                    if (symbol) {
                        error(errorLocation, ts.Diagnostics.Cannot_use_namespace_0_as_a_value, ts.unescapeLeadingUnderscores(name));
                        return true;
                    }
                }
                else if (meaning & (788968 /* Type */ & ~1024 /* NamespaceModule */ & ~111551 /* Value */)) {
                    var symbol = resolveSymbol(resolveName(errorLocation, name, (512 /* ValueModule */ | 1024 /* NamespaceModule */) & ~788968 /* Type */, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false));
                    if (symbol) {
                        error(errorLocation, ts.Diagnostics.Cannot_use_namespace_0_as_a_type, ts.unescapeLeadingUnderscores(name));
                        return true;
                    }
                }
                return false;
            }
            function checkResolvedBlockScopedVariable(result, errorLocation) {
                var _a;
                ts.Debug.assert(!!(result.flags & 2 /* BlockScopedVariable */ || result.flags & 32 /* Class */ || result.flags & 384 /* Enum */));
                if (result.flags & (16 /* Function */ | 1 /* FunctionScopedVariable */ | 67108864 /* Assignment */) && result.flags & 32 /* Class */) {
                    // constructor functions aren't block scoped
                    return;
                }
                // Block-scoped variables cannot be used before their definition
                var declaration = (_a = result.declarations) === null || _a === void 0 ? void 0 : _a.find(function (d) { return ts.isBlockOrCatchScoped(d) || ts.isClassLike(d) || (d.kind === 258 /* EnumDeclaration */); });
                if (declaration === undefined)
                    return ts.Debug.fail("checkResolvedBlockScopedVariable could not find block-scoped declaration");
                if (!(declaration.flags & 8388608 /* Ambient */) && !isBlockScopedNameDeclaredBeforeUse(declaration, errorLocation)) {
                    var diagnosticMessage = void 0;
                    var declarationName = ts.declarationNameToString(ts.getNameOfDeclaration(declaration));
                    if (result.flags & 2 /* BlockScopedVariable */) {
                        diagnosticMessage = error(errorLocation, ts.Diagnostics.Block_scoped_variable_0_used_before_its_declaration, declarationName);
                    }
                    else if (result.flags & 32 /* Class */) {
                        diagnosticMessage = error(errorLocation, ts.Diagnostics.Class_0_used_before_its_declaration, declarationName);
                    }
                    else if (result.flags & 256 /* RegularEnum */) {
                        diagnosticMessage = error(errorLocation, ts.Diagnostics.Enum_0_used_before_its_declaration, declarationName);
                    }
                    else {
                        ts.Debug.assert(!!(result.flags & 128 /* ConstEnum */));
                        if (ts.shouldPreserveConstEnums(compilerOptions)) {
                            diagnosticMessage = error(errorLocation, ts.Diagnostics.Enum_0_used_before_its_declaration, declarationName);
                        }
                    }
                    if (diagnosticMessage) {
                        ts.addRelatedInfo(diagnosticMessage, ts.createDiagnosticForNode(declaration, ts.Diagnostics._0_is_declared_here, declarationName));
                    }
                }
            }
            /* Starting from 'initial' node walk up the parent chain until 'stopAt' node is reached.
             * If at any point current node is equal to 'parent' node - return true.
             * Return false if 'stopAt' node is reached or isFunctionLike(current) === true.
             */
            function isSameScopeDescendentOf(initial, parent, stopAt) {
                return !!parent && !!ts.findAncestor(initial, function (n) { return n === stopAt || ts.isFunctionLike(n) ? "quit" : n === parent; });
            }
            function getAnyImportSyntax(node) {
                switch (node.kind) {
                    case 263 /* ImportEqualsDeclaration */:
                        return node;
                    case 265 /* ImportClause */:
                        return node.parent;
                    case 266 /* NamespaceImport */:
                        return node.parent.parent;
                    case 268 /* ImportSpecifier */:
                        return node.parent.parent.parent;
                    default:
                        return undefined;
                }
            }
            function getDeclarationOfAliasSymbol(symbol) {
                return symbol.declarations && ts.findLast(symbol.declarations, isAliasSymbolDeclaration);
            }
            /**
             * An alias symbol is created by one of the following declarations:
             * import <symbol> = ...
             * import <symbol> from ...
             * import * as <symbol> from ...
             * import { x as <symbol> } from ...
             * export { x as <symbol> } from ...
             * export * as ns <symbol> from ...
             * export = <EntityNameExpression>
             * export default <EntityNameExpression>
             * module.exports = <EntityNameExpression>
             * {<Identifier>}
             * {name: <EntityNameExpression>}
             * const { x } = require ...
             */
            function isAliasSymbolDeclaration(node) {
                return node.kind === 263 /* ImportEqualsDeclaration */
                    || node.kind === 262 /* NamespaceExportDeclaration */
                    || node.kind === 265 /* ImportClause */ && !!node.name
                    || node.kind === 266 /* NamespaceImport */
                    || node.kind === 272 /* NamespaceExport */
                    || node.kind === 268 /* ImportSpecifier */
                    || node.kind === 273 /* ExportSpecifier */
                    || node.kind === 269 /* ExportAssignment */ && ts.exportAssignmentIsAlias(node)
                    || ts.isBinaryExpression(node) && ts.getAssignmentDeclarationKind(node) === 2 /* ModuleExports */ && ts.exportAssignmentIsAlias(node)
                    || ts.isAccessExpression(node)
                        && ts.isBinaryExpression(node.parent)
                        && node.parent.left === node
                        && node.parent.operatorToken.kind === 63 /* EqualsToken */
                        && isAliasableOrJsExpression(node.parent.right)
                    || node.kind === 292 /* ShorthandPropertyAssignment */
                    || node.kind === 291 /* PropertyAssignment */ && isAliasableOrJsExpression(node.initializer)
                    || ts.isRequireVariableDeclaration(node);
            }
            function isAliasableOrJsExpression(e) {
                return ts.isAliasableExpression(e) || ts.isFunctionExpression(e) && isJSConstructor(e);
            }
            function getTargetOfImportEqualsDeclaration(node, dontResolveAlias) {
                var commonJSPropertyAccess = getCommonJSPropertyAccess(node);
                if (commonJSPropertyAccess) {
                    var name = ts.getLeftmostAccessExpression(commonJSPropertyAccess.expression).arguments[0];
                    return ts.isIdentifier(commonJSPropertyAccess.name)
                        ? resolveSymbol(getPropertyOfType(resolveExternalModuleTypeByLiteral(name), commonJSPropertyAccess.name.escapedText))
                        : undefined;
                }
                if (ts.isVariableDeclaration(node) || node.moduleReference.kind === 275 /* ExternalModuleReference */) {
                    var immediate = resolveExternalModuleName(node, ts.getExternalModuleRequireArgument(node) || ts.getExternalModuleImportEqualsDeclarationExpression(node));
                    var resolved_4 = resolveExternalModuleSymbol(immediate);
                    markSymbolOfAliasDeclarationIfTypeOnly(node, immediate, resolved_4, /*overwriteEmpty*/ false);
                    return resolved_4;
                }
                var resolved = getSymbolOfPartOfRightHandSideOfImportEquals(node.moduleReference, dontResolveAlias);
                checkAndReportErrorForResolvingImportAliasToTypeOnlySymbol(node, resolved);
                return resolved;
            }
            function checkAndReportErrorForResolvingImportAliasToTypeOnlySymbol(node, resolved) {
                if (markSymbolOfAliasDeclarationIfTypeOnly(node, /*immediateTarget*/ undefined, resolved, /*overwriteEmpty*/ false) && !node.isTypeOnly) {
                    var typeOnlyDeclaration = getTypeOnlyAliasDeclaration(getSymbolOfNode(node));
                    var isExport = ts.typeOnlyDeclarationIsExport(typeOnlyDeclaration);
                    var message = isExport
                        ? ts.Diagnostics.An_import_alias_cannot_reference_a_declaration_that_was_exported_using_export_type
                        : ts.Diagnostics.An_import_alias_cannot_reference_a_declaration_that_was_imported_using_import_type;
                    var relatedMessage = isExport
                        ? ts.Diagnostics._0_was_exported_here
                        : ts.Diagnostics._0_was_imported_here;
                    // Non-null assertion is safe because the optionality comes from ImportClause,
                    // but if an ImportClause was the typeOnlyDeclaration, it had to have a `name`.
                    var name = ts.unescapeLeadingUnderscores(typeOnlyDeclaration.name.escapedText);
                    ts.addRelatedInfo(error(node.moduleReference, message), ts.createDiagnosticForNode(typeOnlyDeclaration, relatedMessage, name));
                }
            }
            function resolveExportByName(moduleSymbol, name, sourceNode, dontResolveAlias) {
                var exportValue = moduleSymbol.exports.get("export=" /* ExportEquals */);
                var exportSymbol = exportValue ? getPropertyOfType(getTypeOfSymbol(exportValue), name) : moduleSymbol.exports.get(name);
                var resolved = resolveSymbol(exportSymbol, dontResolveAlias);
                markSymbolOfAliasDeclarationIfTypeOnly(sourceNode, exportSymbol, resolved, /*overwriteEmpty*/ false);
                return resolved;
            }
            function isSyntacticDefault(node) {
                return ((ts.isExportAssignment(node) && !node.isExportEquals) || ts.hasSyntacticModifier(node, 512 /* Default */) || ts.isExportSpecifier(node));
            }
            function canHaveSyntheticDefault(file, moduleSymbol, dontResolveAlias) {
                if (!allowSyntheticDefaultImports) {
                    return false;
                }
                // Declaration files (and ambient modules)
                if (!file || file.isDeclarationFile) {
                    // Definitely cannot have a synthetic default if they have a syntactic default member specified
                    var defaultExportSymbol = resolveExportByName(moduleSymbol, "default" /* Default */, /*sourceNode*/ undefined, /*dontResolveAlias*/ true); // Dont resolve alias because we want the immediately exported symbol's declaration
                    if (defaultExportSymbol && ts.some(defaultExportSymbol.declarations, isSyntacticDefault)) {
                        return false;
                    }
                    // It _might_ still be incorrect to assume there is no __esModule marker on the import at runtime, even if there is no `default` member
                    // So we check a bit more,
                    if (resolveExportByName(moduleSymbol, ts.escapeLeadingUnderscores("__esModule"), /*sourceNode*/ undefined, dontResolveAlias)) {
                        // If there is an `__esModule` specified in the declaration (meaning someone explicitly added it or wrote it in their code),
                        // it definitely is a module and does not have a synthetic default
                        return false;
                    }
                    // There are _many_ declaration files not written with esmodules in mind that still get compiled into a format with __esModule set
                    // Meaning there may be no default at runtime - however to be on the permissive side, we allow access to a synthetic default member
                    // as there is no marker to indicate if the accompanying JS has `__esModule` or not, or is even native esm
                    return true;
                }
                // TypeScript files never have a synthetic default (as they are always emitted with an __esModule marker) _unless_ they contain an export= statement
                if (!ts.isSourceFileJS(file)) {
                    return hasExportAssignmentSymbol(moduleSymbol);
                }
                // JS files have a synthetic default if they do not contain ES2015+ module syntax (export = is not valid in js) _and_ do not have an __esModule marker
                return !file.externalModuleIndicator && !resolveExportByName(moduleSymbol, ts.escapeLeadingUnderscores("__esModule"), /*sourceNode*/ undefined, dontResolveAlias);
            }
            function getTargetOfImportClause(node, dontResolveAlias) {
                var _a;
                var moduleSymbol = resolveExternalModuleName(node, node.parent.moduleSpecifier);
                if (moduleSymbol) {
                    var exportDefaultSymbol = void 0;
                    if (ts.isShorthandAmbientModuleSymbol(moduleSymbol)) {
                        exportDefaultSymbol = moduleSymbol;
                    }
                    else {
                        exportDefaultSymbol = resolveExportByName(moduleSymbol, "default" /* Default */, node, dontResolveAlias);
                    }
                    var file = (_a = moduleSymbol.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.isSourceFile);
                    var hasSyntheticDefault = canHaveSyntheticDefault(file, moduleSymbol, dontResolveAlias);
                    if (!exportDefaultSymbol && !hasSyntheticDefault) {
                        if (hasExportAssignmentSymbol(moduleSymbol)) {
                            var compilerOptionName = moduleKind >= ts.ModuleKind.ES2015 ? "allowSyntheticDefaultImports" : "esModuleInterop";
                            var exportEqualsSymbol = moduleSymbol.exports.get("export=" /* ExportEquals */);
                            var exportAssignment = exportEqualsSymbol.valueDeclaration;
                            var err = error(node.name, ts.Diagnostics.Module_0_can_only_be_default_imported_using_the_1_flag, symbolToString(moduleSymbol), compilerOptionName);
                            if (exportAssignment) {
                                ts.addRelatedInfo(err, ts.createDiagnosticForNode(exportAssignment, ts.Diagnostics.This_module_is_declared_with_using_export_and_can_only_be_used_with_a_default_import_when_using_the_0_flag, compilerOptionName));
                            }
                        }
                        else {
                            reportNonDefaultExport(moduleSymbol, node);
                        }
                    }
                    else if (hasSyntheticDefault) {
                        // per emit behavior, a synthetic default overrides a "real" .default member if `__esModule` is not present
                        var resolved = resolveExternalModuleSymbol(moduleSymbol, dontResolveAlias) || resolveSymbol(moduleSymbol, dontResolveAlias);
                        markSymbolOfAliasDeclarationIfTypeOnly(node, moduleSymbol, resolved, /*overwriteTypeOnly*/ false);
                        return resolved;
                    }
                    markSymbolOfAliasDeclarationIfTypeOnly(node, exportDefaultSymbol, /*finalTarget*/ undefined, /*overwriteTypeOnly*/ false);
                    return exportDefaultSymbol;
                }
            }
            function reportNonDefaultExport(moduleSymbol, node) {
                var _a, _b, _c;
                if ((_a = moduleSymbol.exports) === null || _a === void 0 ? void 0 : _a.has(node.symbol.escapedName)) {
                    error(node.name, ts.Diagnostics.Module_0_has_no_default_export_Did_you_mean_to_use_import_1_from_0_instead, symbolToString(moduleSymbol), symbolToString(node.symbol));
                }
                else {
                    var diagnostic = error(node.name, ts.Diagnostics.Module_0_has_no_default_export, symbolToString(moduleSymbol));
                    var exportStar = (_b = moduleSymbol.exports) === null || _b === void 0 ? void 0 : _b.get("__export" /* ExportStar */);
                    if (exportStar) {
                        var defaultExport = (_c = exportStar.declarations) === null || _c === void 0 ? void 0 : _c.find(function (decl) {
                            var _a, _b;
                            return !!(ts.isExportDeclaration(decl) && decl.moduleSpecifier &&
                                ((_b = (_a = resolveExternalModuleName(decl, decl.moduleSpecifier)) === null || _a === void 0 ? void 0 : _a.exports) === null || _b === void 0 ? void 0 : _b.has("default" /* Default */)));
                        });
                        if (defaultExport) {
                            ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(defaultExport, ts.Diagnostics.export_Asterisk_does_not_re_export_a_default));
                        }
                    }
                }
            }
            function getTargetOfNamespaceImport(node, dontResolveAlias) {
                var moduleSpecifier = node.parent.parent.moduleSpecifier;
                var immediate = resolveExternalModuleName(node, moduleSpecifier);
                var resolved = resolveESModuleSymbol(immediate, moduleSpecifier, dontResolveAlias, /*suppressUsageError*/ false);
                markSymbolOfAliasDeclarationIfTypeOnly(node, immediate, resolved, /*overwriteEmpty*/ false);
                return resolved;
            }
            function getTargetOfNamespaceExport(node, dontResolveAlias) {
                var moduleSpecifier = node.parent.moduleSpecifier;
                var immediate = moduleSpecifier && resolveExternalModuleName(node, moduleSpecifier);
                var resolved = moduleSpecifier && resolveESModuleSymbol(immediate, moduleSpecifier, dontResolveAlias, /*suppressUsageError*/ false);
                markSymbolOfAliasDeclarationIfTypeOnly(node, immediate, resolved, /*overwriteEmpty*/ false);
                return resolved;
            }
            // This function creates a synthetic symbol that combines the value side of one symbol with the
            // type/namespace side of another symbol. Consider this example:
            //
            //   declare module graphics {
            //       interface Point {
            //           x: number;
            //           y: number;
            //       }
            //   }
            //   declare var graphics: {
            //       Point: new (x: number, y: number) => graphics.Point;
            //   }
            //   declare module "graphics" {
            //       export = graphics;
            //   }
            //
            // An 'import { Point } from "graphics"' needs to create a symbol that combines the value side 'Point'
            // property with the type/namespace side interface 'Point'.
            function combineValueAndTypeSymbols(valueSymbol, typeSymbol) {
                if (valueSymbol === unknownSymbol && typeSymbol === unknownSymbol) {
                    return unknownSymbol;
                }
                if (valueSymbol.flags & (788968 /* Type */ | 1920 /* Namespace */)) {
                    return valueSymbol;
                }
                var result = createSymbol(valueSymbol.flags | typeSymbol.flags, valueSymbol.escapedName);
                result.declarations = ts.deduplicate(ts.concatenate(valueSymbol.declarations, typeSymbol.declarations), ts.equateValues);
                result.parent = valueSymbol.parent || typeSymbol.parent;
                if (valueSymbol.valueDeclaration)
                    result.valueDeclaration = valueSymbol.valueDeclaration;
                if (typeSymbol.members)
                    result.members = new ts.Map(typeSymbol.members);
                if (valueSymbol.exports)
                    result.exports = new ts.Map(valueSymbol.exports);
                return result;
            }
            function getExportOfModule(symbol, name, specifier, dontResolveAlias) {
                if (symbol.flags & 1536 /* Module */) {
                    var exportSymbol = getExportsOfSymbol(symbol).get(name.escapedText);
                    var resolved = resolveSymbol(exportSymbol, dontResolveAlias);
                    markSymbolOfAliasDeclarationIfTypeOnly(specifier, exportSymbol, resolved, /*overwriteEmpty*/ false);
                    return resolved;
                }
            }
            function getPropertyOfVariable(symbol, name) {
                if (symbol.flags & 3 /* Variable */) {
                    var typeAnnotation = symbol.valueDeclaration.type;
                    if (typeAnnotation) {
                        return resolveSymbol(getPropertyOfType(getTypeFromTypeNode(typeAnnotation), name));
                    }
                }
            }
            function getExternalModuleMember(node, specifier, dontResolveAlias) {
                var _a, _b;
                if (dontResolveAlias === void 0) {
                    dontResolveAlias = false;
                }
                var moduleSpecifier = ts.getExternalModuleRequireArgument(node) || node.moduleSpecifier;
                var moduleSymbol = resolveExternalModuleName(node, moduleSpecifier); // TODO: GH#18217
                var name = !ts.isPropertyAccessExpression(specifier) && specifier.propertyName || specifier.name;
                if (!ts.isIdentifier(name)) {
                    return undefined;
                }
                var suppressInteropError = name.escapedText === "default" /* Default */ && !!(compilerOptions.allowSyntheticDefaultImports || compilerOptions.esModuleInterop);
                var targetSymbol = resolveESModuleSymbol(moduleSymbol, moduleSpecifier, /*dontResolveAlias*/ false, suppressInteropError);
                if (targetSymbol) {
                    if (name.escapedText) {
                        if (ts.isShorthandAmbientModuleSymbol(moduleSymbol)) {
                            return moduleSymbol;
                        }
                        var symbolFromVariable = void 0;
                        // First check if module was specified with "export=". If so, get the member from the resolved type
                        if (moduleSymbol && moduleSymbol.exports && moduleSymbol.exports.get("export=" /* ExportEquals */)) {
                            symbolFromVariable = getPropertyOfType(getTypeOfSymbol(targetSymbol), name.escapedText, /*skipObjectFunctionPropertyAugment*/ true);
                        }
                        else {
                            symbolFromVariable = getPropertyOfVariable(targetSymbol, name.escapedText);
                        }
                        // if symbolFromVariable is export - get its final target
                        symbolFromVariable = resolveSymbol(symbolFromVariable, dontResolveAlias);
                        var symbolFromModule = getExportOfModule(targetSymbol, name, specifier, dontResolveAlias);
                        if (symbolFromModule === undefined && name.escapedText === "default" /* Default */) {
                            var file = (_a = moduleSymbol.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.isSourceFile);
                            if (canHaveSyntheticDefault(file, moduleSymbol, dontResolveAlias)) {
                                symbolFromModule = resolveExternalModuleSymbol(moduleSymbol, dontResolveAlias) || resolveSymbol(moduleSymbol, dontResolveAlias);
                            }
                        }
                        var symbol = symbolFromModule && symbolFromVariable && symbolFromModule !== symbolFromVariable ?
                            combineValueAndTypeSymbols(symbolFromVariable, symbolFromModule) :
                            symbolFromModule || symbolFromVariable;
                        if (!symbol) {
                            var moduleName = getFullyQualifiedName(moduleSymbol, node);
                            var declarationName = ts.declarationNameToString(name);
                            var suggestion = getSuggestedSymbolForNonexistentModule(name, targetSymbol);
                            if (suggestion !== undefined) {
                                var suggestionName = symbolToString(suggestion);
                                var diagnostic = error(name, ts.Diagnostics._0_has_no_exported_member_named_1_Did_you_mean_2, moduleName, declarationName, suggestionName);
                                if (suggestion.valueDeclaration) {
                                    ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(suggestion.valueDeclaration, ts.Diagnostics._0_is_declared_here, suggestionName));
                                }
                            }
                            else {
                                if ((_b = moduleSymbol.exports) === null || _b === void 0 ? void 0 : _b.has("default" /* Default */)) {
                                    error(name, ts.Diagnostics.Module_0_has_no_exported_member_1_Did_you_mean_to_use_import_1_from_0_instead, moduleName, declarationName);
                                }
                                else {
                                    reportNonExportedMember(node, name, declarationName, moduleSymbol, moduleName);
                                }
                            }
                        }
                        return symbol;
                    }
                }
            }
            function reportNonExportedMember(node, name, declarationName, moduleSymbol, moduleName) {
                var _a, _b;
                var localSymbol = (_b = (_a = moduleSymbol.valueDeclaration) === null || _a === void 0 ? void 0 : _a.locals) === null || _b === void 0 ? void 0 : _b.get(name.escapedText);
                var exports = moduleSymbol.exports;
                if (localSymbol) {
                    var exportedEqualsSymbol = exports === null || exports === void 0 ? void 0 : exports.get("export=" /* ExportEquals */);
                    if (exportedEqualsSymbol) {
                        getSymbolIfSameReference(exportedEqualsSymbol, localSymbol) ? reportInvalidImportEqualsExportMember(node, name, declarationName, moduleName) :
                            error(name, ts.Diagnostics.Module_0_has_no_exported_member_1, moduleName, declarationName);
                    }
                    else {
                        var exportedSymbol = exports ? ts.find(symbolsToArray(exports), function (symbol) { return !!getSymbolIfSameReference(symbol, localSymbol); }) : undefined;
                        var diagnostic = exportedSymbol ? error(name, ts.Diagnostics.Module_0_declares_1_locally_but_it_is_exported_as_2, moduleName, declarationName, symbolToString(exportedSymbol)) :
                            error(name, ts.Diagnostics.Module_0_declares_1_locally_but_it_is_not_exported, moduleName, declarationName);
                        if (localSymbol.declarations) {
                            ts.addRelatedInfo.apply(void 0, __spreadArray([diagnostic], ts.map(localSymbol.declarations, function (decl, index) {
                                return ts.createDiagnosticForNode(decl, index === 0 ? ts.Diagnostics._0_is_declared_here : ts.Diagnostics.and_here, declarationName);
                            }), false));
                        }
                    }
                }
                else {
                    error(name, ts.Diagnostics.Module_0_has_no_exported_member_1, moduleName, declarationName);
                }
            }
            function reportInvalidImportEqualsExportMember(node, name, declarationName, moduleName) {
                if (moduleKind >= ts.ModuleKind.ES2015) {
                    var message = compilerOptions.esModuleInterop ? ts.Diagnostics._0_can_only_be_imported_by_using_a_default_import :
                        ts.Diagnostics._0_can_only_be_imported_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import;
                    error(name, message, declarationName);
                }
                else {
                    if (ts.isInJSFile(node)) {
                        var message = compilerOptions.esModuleInterop ? ts.Diagnostics._0_can_only_be_imported_by_using_a_require_call_or_by_using_a_default_import :
                            ts.Diagnostics._0_can_only_be_imported_by_using_a_require_call_or_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import;
                        error(name, message, declarationName);
                    }
                    else {
                        var message = compilerOptions.esModuleInterop ? ts.Diagnostics._0_can_only_be_imported_by_using_import_1_require_2_or_a_default_import :
                            ts.Diagnostics._0_can_only_be_imported_by_using_import_1_require_2_or_by_turning_on_the_esModuleInterop_flag_and_using_a_default_import;
                        error(name, message, declarationName, declarationName, moduleName);
                    }
                }
            }
            function getTargetOfImportSpecifier(node, dontResolveAlias) {
                var root = ts.isBindingElement(node) ? ts.getRootDeclaration(node) : node.parent.parent.parent;
                var commonJSPropertyAccess = getCommonJSPropertyAccess(root);
                var resolved = getExternalModuleMember(root, commonJSPropertyAccess || node, dontResolveAlias);
                var name = node.propertyName || node.name;
                if (commonJSPropertyAccess && resolved && ts.isIdentifier(name)) {
                    return resolveSymbol(getPropertyOfType(getTypeOfSymbol(resolved), name.escapedText), dontResolveAlias);
                }
                markSymbolOfAliasDeclarationIfTypeOnly(node, /*immediateTarget*/ undefined, resolved, /*overwriteEmpty*/ false);
                return resolved;
            }
            function getCommonJSPropertyAccess(node) {
                if (ts.isVariableDeclaration(node) && node.initializer && ts.isPropertyAccessExpression(node.initializer)) {
                    return node.initializer;
                }
            }
            function getTargetOfNamespaceExportDeclaration(node, dontResolveAlias) {
                var resolved = resolveExternalModuleSymbol(node.parent.symbol, dontResolveAlias);
                markSymbolOfAliasDeclarationIfTypeOnly(node, /*immediateTarget*/ undefined, resolved, /*overwriteEmpty*/ false);
                return resolved;
            }
            function getTargetOfExportSpecifier(node, meaning, dontResolveAlias) {
                var resolved = node.parent.parent.moduleSpecifier ?
                    getExternalModuleMember(node.parent.parent, node, dontResolveAlias) :
                    resolveEntityName(node.propertyName || node.name, meaning, /*ignoreErrors*/ false, dontResolveAlias);
                markSymbolOfAliasDeclarationIfTypeOnly(node, /*immediateTarget*/ undefined, resolved, /*overwriteEmpty*/ false);
                return resolved;
            }
            function getTargetOfExportAssignment(node, dontResolveAlias) {
                var expression = ts.isExportAssignment(node) ? node.expression : node.right;
                var resolved = getTargetOfAliasLikeExpression(expression, dontResolveAlias);
                markSymbolOfAliasDeclarationIfTypeOnly(node, /*immediateTarget*/ undefined, resolved, /*overwriteEmpty*/ false);
                return resolved;
            }
            function getTargetOfAliasLikeExpression(expression, dontResolveAlias) {
                if (ts.isClassExpression(expression)) {
                    return checkExpressionCached(expression).symbol;
                }
                if (!ts.isEntityName(expression) && !ts.isEntityNameExpression(expression)) {
                    return undefined;
                }
                var aliasLike = resolveEntityName(expression, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */, /*ignoreErrors*/ true, dontResolveAlias);
                if (aliasLike) {
                    return aliasLike;
                }
                checkExpressionCached(expression);
                return getNodeLinks(expression).resolvedSymbol;
            }
            function getTargetOfPropertyAssignment(node, dontRecursivelyResolve) {
                var expression = node.initializer;
                return getTargetOfAliasLikeExpression(expression, dontRecursivelyResolve);
            }
            function getTargetOfAccessExpression(node, dontRecursivelyResolve) {
                if (!(ts.isBinaryExpression(node.parent) && node.parent.left === node && node.parent.operatorToken.kind === 63 /* EqualsToken */)) {
                    return undefined;
                }
                return getTargetOfAliasLikeExpression(node.parent.right, dontRecursivelyResolve);
            }
            function getTargetOfAliasDeclaration(node, dontRecursivelyResolve) {
                if (dontRecursivelyResolve === void 0) {
                    dontRecursivelyResolve = false;
                }
                switch (node.kind) {
                    case 263 /* ImportEqualsDeclaration */:
                    case 252 /* VariableDeclaration */:
                        return getTargetOfImportEqualsDeclaration(node, dontRecursivelyResolve);
                    case 265 /* ImportClause */:
                        return getTargetOfImportClause(node, dontRecursivelyResolve);
                    case 266 /* NamespaceImport */:
                        return getTargetOfNamespaceImport(node, dontRecursivelyResolve);
                    case 272 /* NamespaceExport */:
                        return getTargetOfNamespaceExport(node, dontRecursivelyResolve);
                    case 268 /* ImportSpecifier */:
                    case 201 /* BindingElement */:
                        return getTargetOfImportSpecifier(node, dontRecursivelyResolve);
                    case 273 /* ExportSpecifier */:
                        return getTargetOfExportSpecifier(node, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */, dontRecursivelyResolve);
                    case 269 /* ExportAssignment */:
                    case 219 /* BinaryExpression */:
                        return getTargetOfExportAssignment(node, dontRecursivelyResolve);
                    case 262 /* NamespaceExportDeclaration */:
                        return getTargetOfNamespaceExportDeclaration(node, dontRecursivelyResolve);
                    case 292 /* ShorthandPropertyAssignment */:
                        return resolveEntityName(node.name, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */, /*ignoreErrors*/ true, dontRecursivelyResolve);
                    case 291 /* PropertyAssignment */:
                        return getTargetOfPropertyAssignment(node, dontRecursivelyResolve);
                    case 205 /* ElementAccessExpression */:
                    case 204 /* PropertyAccessExpression */:
                        return getTargetOfAccessExpression(node, dontRecursivelyResolve);
                    default:
                        return ts.Debug.fail();
                }
            }
            /**
             * Indicates that a symbol is an alias that does not merge with a local declaration.
             * OR Is a JSContainer which may merge an alias with a local declaration
             */
            function isNonLocalAlias(symbol, excludes) {
                if (excludes === void 0) {
                    excludes = 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */;
                }
                if (!symbol)
                    return false;
                return (symbol.flags & (2097152 /* Alias */ | excludes)) === 2097152 /* Alias */ || !!(symbol.flags & 2097152 /* Alias */ && symbol.flags & 67108864 /* Assignment */);
            }
            function resolveSymbol(symbol, dontResolveAlias) {
                return !dontResolveAlias && isNonLocalAlias(symbol) ? resolveAlias(symbol) : symbol;
            }
            function resolveAlias(symbol) {
                ts.Debug.assert((symbol.flags & 2097152 /* Alias */) !== 0, "Should only get Alias here.");
                var links = getSymbolLinks(symbol);
                if (!links.target) {
                    links.target = resolvingSymbol;
                    var node = getDeclarationOfAliasSymbol(symbol);
                    if (!node)
                        return ts.Debug.fail();
                    var target = getTargetOfAliasDeclaration(node);
                    if (links.target === resolvingSymbol) {
                        links.target = target || unknownSymbol;
                    }
                    else {
                        error(node, ts.Diagnostics.Circular_definition_of_import_alias_0, symbolToString(symbol));
                    }
                }
                else if (links.target === resolvingSymbol) {
                    links.target = unknownSymbol;
                }
                return links.target;
            }
            function tryResolveAlias(symbol) {
                var links = getSymbolLinks(symbol);
                if (links.target !== resolvingSymbol) {
                    return resolveAlias(symbol);
                }
                return undefined;
            }
            /**
             * Marks a symbol as type-only if its declaration is syntactically type-only.
             * If it is not itself marked type-only, but resolves to a type-only alias
             * somewhere in its resolution chain, save a reference to the type-only alias declaration
             * so the alias _not_ marked type-only can be identified as _transitively_ type-only.
             *
             * This function is called on each alias declaration that could be type-only or resolve to
             * another type-only alias during `resolveAlias`, so that later, when an alias is used in a
             * JS-emitting expression, we can quickly determine if that symbol is effectively type-only
             * and issue an error if so.
             *
             * @param aliasDeclaration The alias declaration not marked as type-only
             * has already been marked as not resolving to a type-only alias. Used when recursively resolving qualified
             * names of import aliases, e.g. `import C = a.b.C`. If namespace `a` is not found to be type-only, the
             * import declaration will initially be marked as not resolving to a type-only symbol. But, namespace `b`
             * must still be checked for a type-only marker, overwriting the previous negative result if found.
             * @param immediateTarget The symbol to which the alias declaration immediately resolves
             * @param finalTarget The symbol to which the alias declaration ultimately resolves
             * @param overwriteEmpty Checks `resolvesToSymbol` for type-only declarations even if `aliasDeclaration`
             */
            function markSymbolOfAliasDeclarationIfTypeOnly(aliasDeclaration, immediateTarget, finalTarget, overwriteEmpty) {
                if (!aliasDeclaration || ts.isPropertyAccessExpression(aliasDeclaration))
                    return false;
                // If the declaration itself is type-only, mark it and return.
                // No need to check what it resolves to.
                var sourceSymbol = getSymbolOfNode(aliasDeclaration);
                if (ts.isTypeOnlyImportOrExportDeclaration(aliasDeclaration)) {
                    var links_1 = getSymbolLinks(sourceSymbol);
                    links_1.typeOnlyDeclaration = aliasDeclaration;
                    return true;
                }
                var links = getSymbolLinks(sourceSymbol);
                return markSymbolOfAliasDeclarationIfTypeOnlyWorker(links, immediateTarget, overwriteEmpty)
                    || markSymbolOfAliasDeclarationIfTypeOnlyWorker(links, finalTarget, overwriteEmpty);
            }
            function markSymbolOfAliasDeclarationIfTypeOnlyWorker(aliasDeclarationLinks, target, overwriteEmpty) {
                var _a, _b, _c;
                if (target && (aliasDeclarationLinks.typeOnlyDeclaration === undefined || overwriteEmpty && aliasDeclarationLinks.typeOnlyDeclaration === false)) {
                    var exportSymbol = (_b = (_a = target.exports) === null || _a === void 0 ? void 0 : _a.get("export=" /* ExportEquals */)) !== null && _b !== void 0 ? _b : target;
                    var typeOnly = exportSymbol.declarations && ts.find(exportSymbol.declarations, ts.isTypeOnlyImportOrExportDeclaration);
                    aliasDeclarationLinks.typeOnlyDeclaration = (_c = typeOnly !== null && typeOnly !== void 0 ? typeOnly : getSymbolLinks(exportSymbol).typeOnlyDeclaration) !== null && _c !== void 0 ? _c : false;
                }
                return !!aliasDeclarationLinks.typeOnlyDeclaration;
            }
            /** Indicates that a symbol directly or indirectly resolves to a type-only import or export. */
            function getTypeOnlyAliasDeclaration(symbol) {
                if (!(symbol.flags & 2097152 /* Alias */)) {
                    return undefined;
                }
                var links = getSymbolLinks(symbol);
                return links.typeOnlyDeclaration || undefined;
            }
            function markExportAsReferenced(node) {
                var symbol = getSymbolOfNode(node);
                var target = resolveAlias(symbol);
                if (target) {
                    var markAlias = target === unknownSymbol ||
                        ((target.flags & 111551 /* Value */) && !isConstEnumOrConstEnumOnlyModule(target) && !getTypeOnlyAliasDeclaration(symbol));
                    if (markAlias) {
                        markAliasSymbolAsReferenced(symbol);
                    }
                }
            }
            // When an alias symbol is referenced, we need to mark the entity it references as referenced and in turn repeat that until
            // we reach a non-alias or an exported entity (which is always considered referenced). We do this by checking the target of
            // the alias as an expression (which recursively takes us back here if the target references another alias).
            function markAliasSymbolAsReferenced(symbol) {
                var links = getSymbolLinks(symbol);
                if (!links.referenced) {
                    links.referenced = true;
                    var node = getDeclarationOfAliasSymbol(symbol);
                    if (!node)
                        return ts.Debug.fail();
                    // We defer checking of the reference of an `import =` until the import itself is referenced,
                    // This way a chain of imports can be elided if ultimately the final input is only used in a type
                    // position.
                    if (ts.isInternalModuleImportEqualsDeclaration(node)) {
                        var target = resolveSymbol(symbol);
                        if (target === unknownSymbol || target.flags & 111551 /* Value */) {
                            // import foo = <symbol>
                            checkExpressionCached(node.moduleReference);
                        }
                    }
                }
            }
            // Aliases that resolve to const enums are not marked as referenced because they are not emitted,
            // but their usage in value positions must be tracked to determine if the import can be type-only.
            function markConstEnumAliasAsReferenced(symbol) {
                var links = getSymbolLinks(symbol);
                if (!links.constEnumReferenced) {
                    links.constEnumReferenced = true;
                }
            }
            // This function is only for imports with entity names
            function getSymbolOfPartOfRightHandSideOfImportEquals(entityName, dontResolveAlias) {
                // There are three things we might try to look for. In the following examples,
                // the search term is enclosed in |...|:
                //
                //     import a = |b|; // Namespace
                //     import a = |b.c|; // Value, type, namespace
                //     import a = |b.c|.d; // Namespace
                if (entityName.kind === 79 /* Identifier */ && ts.isRightSideOfQualifiedNameOrPropertyAccess(entityName)) {
                    entityName = entityName.parent;
                }
                // Check for case 1 and 3 in the above example
                if (entityName.kind === 79 /* Identifier */ || entityName.parent.kind === 159 /* QualifiedName */) {
                    return resolveEntityName(entityName, 1920 /* Namespace */, /*ignoreErrors*/ false, dontResolveAlias);
                }
                else {
                    // Case 2 in above example
                    // entityName.kind could be a QualifiedName or a Missing identifier
                    ts.Debug.assert(entityName.parent.kind === 263 /* ImportEqualsDeclaration */);
                    return resolveEntityName(entityName, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */, /*ignoreErrors*/ false, dontResolveAlias);
                }
            }
            function getFullyQualifiedName(symbol, containingLocation) {
                return symbol.parent ? getFullyQualifiedName(symbol.parent, containingLocation) + "." + symbolToString(symbol) : symbolToString(symbol, containingLocation, /*meaning*/ undefined, 16 /* DoNotIncludeSymbolChain */ | 4 /* AllowAnyNodeKind */);
            }
            /**
             * Resolves a qualified name and any involved aliases.
             */
            function resolveEntityName(name, meaning, ignoreErrors, dontResolveAlias, location) {
                if (ts.nodeIsMissing(name)) {
                    return undefined;
                }
                var namespaceMeaning = 1920 /* Namespace */ | (ts.isInJSFile(name) ? meaning & 111551 /* Value */ : 0);
                var symbol;
                if (name.kind === 79 /* Identifier */) {
                    var message = meaning === namespaceMeaning || ts.nodeIsSynthesized(name) ? ts.Diagnostics.Cannot_find_namespace_0 : getCannotFindNameDiagnosticForName(ts.getFirstIdentifier(name));
                    var symbolFromJSPrototype = ts.isInJSFile(name) && !ts.nodeIsSynthesized(name) ? resolveEntityNameFromAssignmentDeclaration(name, meaning) : undefined;
                    symbol = getMergedSymbol(resolveName(location || name, name.escapedText, meaning, ignoreErrors || symbolFromJSPrototype ? undefined : message, name, /*isUse*/ true));
                    if (!symbol) {
                        return getMergedSymbol(symbolFromJSPrototype);
                    }
                }
                else if (name.kind === 159 /* QualifiedName */ || name.kind === 204 /* PropertyAccessExpression */) {
                    var left = name.kind === 159 /* QualifiedName */ ? name.left : name.expression;
                    var right = name.kind === 159 /* QualifiedName */ ? name.right : name.name;
                    var namespace = resolveEntityName(left, namespaceMeaning, ignoreErrors, /*dontResolveAlias*/ false, location);
                    if (!namespace || ts.nodeIsMissing(right)) {
                        return undefined;
                    }
                    else if (namespace === unknownSymbol) {
                        return namespace;
                    }
                    if (namespace.valueDeclaration &&
                        ts.isInJSFile(namespace.valueDeclaration) &&
                        ts.isVariableDeclaration(namespace.valueDeclaration) &&
                        namespace.valueDeclaration.initializer &&
                        isCommonJsRequire(namespace.valueDeclaration.initializer)) {
                        var moduleName = namespace.valueDeclaration.initializer.arguments[0];
                        var moduleSym = resolveExternalModuleName(moduleName, moduleName);
                        if (moduleSym) {
                            var resolvedModuleSymbol = resolveExternalModuleSymbol(moduleSym);
                            if (resolvedModuleSymbol) {
                                namespace = resolvedModuleSymbol;
                            }
                        }
                    }
                    symbol = getMergedSymbol(getSymbol(getExportsOfSymbol(namespace), right.escapedText, meaning));
                    if (!symbol) {
                        if (!ignoreErrors) {
                            var namespaceName = getFullyQualifiedName(namespace);
                            var declarationName = ts.declarationNameToString(right);
                            var suggestion = getSuggestedSymbolForNonexistentModule(right, namespace);
                            suggestion ?
                                error(right, ts.Diagnostics._0_has_no_exported_member_named_1_Did_you_mean_2, namespaceName, declarationName, symbolToString(suggestion)) :
                                error(right, ts.Diagnostics.Namespace_0_has_no_exported_member_1, namespaceName, declarationName);
                        }
                        return undefined;
                    }
                }
                else {
                    throw ts.Debug.assertNever(name, "Unknown entity name kind.");
                }
                ts.Debug.assert((ts.getCheckFlags(symbol) & 1 /* Instantiated */) === 0, "Should never get an instantiated symbol here.");
                if (!ts.nodeIsSynthesized(name) && ts.isEntityName(name) && (symbol.flags & 2097152 /* Alias */ || name.parent.kind === 269 /* ExportAssignment */)) {
                    markSymbolOfAliasDeclarationIfTypeOnly(ts.getAliasDeclarationFromName(name), symbol, /*finalTarget*/ undefined, /*overwriteEmpty*/ true);
                }
                return (symbol.flags & meaning) || dontResolveAlias ? symbol : resolveAlias(symbol);
            }
            /**
             * 1. For prototype-property methods like `A.prototype.m = function () ...`, try to resolve names in the scope of `A` too.
             * Note that prototype-property assignment to locations outside the current file (eg globals) doesn't work, so
             * name resolution won't work either.
             * 2. For property assignments like `{ x: function f () { } }`, try to resolve names in the scope of `f` too.
             */
            function resolveEntityNameFromAssignmentDeclaration(name, meaning) {
                if (isJSDocTypeReference(name.parent)) {
                    var secondaryLocation = getAssignmentDeclarationLocation(name.parent);
                    if (secondaryLocation) {
                        return resolveName(secondaryLocation, name.escapedText, meaning, /*nameNotFoundMessage*/ undefined, name, /*isUse*/ true);
                    }
                }
            }
            function getAssignmentDeclarationLocation(node) {
                var typeAlias = ts.findAncestor(node, function (node) { return !(ts.isJSDocNode(node) || node.flags & 4194304 /* JSDoc */) ? "quit" : ts.isJSDocTypeAlias(node); });
                if (typeAlias) {
                    return;
                }
                var host = ts.getJSDocHost(node);
                if (host &&
                    ts.isExpressionStatement(host) &&
                    ts.isBinaryExpression(host.expression) &&
                    ts.getAssignmentDeclarationKind(host.expression) === 3 /* PrototypeProperty */) {
                    // X.prototype.m = /** @param {K} p */ function () { } <-- look for K on X's declaration
                    var symbol = getSymbolOfNode(host.expression.left);
                    if (symbol) {
                        return getDeclarationOfJSPrototypeContainer(symbol);
                    }
                }
                if (host && (ts.isObjectLiteralMethod(host) || ts.isPropertyAssignment(host)) &&
                    ts.isBinaryExpression(host.parent.parent) &&
                    ts.getAssignmentDeclarationKind(host.parent.parent) === 6 /* Prototype */) {
                    // X.prototype = { /** @param {K} p */m() { } } <-- look for K on X's declaration
                    var symbol = getSymbolOfNode(host.parent.parent.left);
                    if (symbol) {
                        return getDeclarationOfJSPrototypeContainer(symbol);
                    }
                }
                var sig = ts.getEffectiveJSDocHost(node);
                if (sig && ts.isFunctionLike(sig)) {
                    var symbol = getSymbolOfNode(sig);
                    return symbol && symbol.valueDeclaration;
                }
            }
            function getDeclarationOfJSPrototypeContainer(symbol) {
                var decl = symbol.parent.valueDeclaration;
                if (!decl) {
                    return undefined;
                }
                var initializer = ts.isAssignmentDeclaration(decl) ? ts.getAssignedExpandoInitializer(decl) :
                    ts.hasOnlyExpressionInitializer(decl) ? ts.getDeclaredExpandoInitializer(decl) :
                        undefined;
                return initializer || decl;
            }
            /**
             * Get the real symbol of a declaration with an expando initializer.
             *
             * Normally, declarations have an associated symbol, but when a declaration has an expando
             * initializer, the expando's symbol is the one that has all the members merged into it.
             */
            function getExpandoSymbol(symbol) {
                var decl = symbol.valueDeclaration;
                if (!decl || !ts.isInJSFile(decl) || symbol.flags & 524288 /* TypeAlias */ || ts.getExpandoInitializer(decl, /*isPrototypeAssignment*/ false)) {
                    return undefined;
                }
                var init = ts.isVariableDeclaration(decl) ? ts.getDeclaredExpandoInitializer(decl) : ts.getAssignedExpandoInitializer(decl);
                if (init) {
                    var initSymbol = getSymbolOfNode(init);
                    if (initSymbol) {
                        return mergeJSSymbols(initSymbol, symbol);
                    }
                }
            }
            function resolveExternalModuleName(location, moduleReferenceExpression, ignoreErrors) {
                var isClassic = ts.getEmitModuleResolutionKind(compilerOptions) === ts.ModuleResolutionKind.Classic;
                var errorMessage = isClassic ?
                    ts.Diagnostics.Cannot_find_module_0_Did_you_mean_to_set_the_moduleResolution_option_to_node_or_to_add_aliases_to_the_paths_option
                    : ts.Diagnostics.Cannot_find_module_0_or_its_corresponding_type_declarations;
                return resolveExternalModuleNameWorker(location, moduleReferenceExpression, ignoreErrors ? undefined : errorMessage);
            }
            function resolveExternalModuleNameWorker(location, moduleReferenceExpression, moduleNotFoundError, isForAugmentation) {
                if (isForAugmentation === void 0) {
                    isForAugmentation = false;
                }
                return ts.isStringLiteralLike(moduleReferenceExpression)
                    ? resolveExternalModule(location, moduleReferenceExpression.text, moduleNotFoundError, moduleReferenceExpression, isForAugmentation)
                    : undefined;
            }
            function resolveExternalModule(location, moduleReference, moduleNotFoundError, errorNode, isForAugmentation) {
                if (isForAugmentation === void 0) {
                    isForAugmentation = false;
                }
                if (ts.startsWith(moduleReference, "@types/")) {
                    var diag = ts.Diagnostics.Cannot_import_type_declaration_files_Consider_importing_0_instead_of_1;
                    var withoutAtTypePrefix = ts.removePrefix(moduleReference, "@types/");
                    error(errorNode, diag, withoutAtTypePrefix, moduleReference);
                }
                var ambientModule = tryFindAmbientModule(moduleReference, /*withAugmentations*/ true);
                if (ambientModule) {
                    return ambientModule;
                }
                var currentSourceFile = ts.getSourceFileOfNode(location);
                var resolvedModule = ts.getResolvedModule(currentSourceFile, moduleReference); // TODO: GH#18217
                var resolutionDiagnostic = resolvedModule && ts.getResolutionDiagnostic(compilerOptions, resolvedModule);
                var sourceFile = resolvedModule && !resolutionDiagnostic && host.getSourceFile(resolvedModule.resolvedFileName);
                if (sourceFile) {
                    if (sourceFile.symbol) {
                        if (resolvedModule.isExternalLibraryImport && !ts.resolutionExtensionIsTSOrJson(resolvedModule.extension)) {
                            errorOnImplicitAnyModule(/*isError*/ false, errorNode, resolvedModule, moduleReference);
                        }
                        // merged symbol is module declaration symbol combined with all augmentations
                        return getMergedSymbol(sourceFile.symbol);
                    }
                    if (moduleNotFoundError) {
                        // report errors only if it was requested
                        error(errorNode, ts.Diagnostics.File_0_is_not_a_module, sourceFile.fileName);
                    }
                    return undefined;
                }
                if (patternAmbientModules) {
                    var pattern = ts.findBestPatternMatch(patternAmbientModules, function (_) { return _.pattern; }, moduleReference);
                    if (pattern) {
                        // If the module reference matched a pattern ambient module ('*.foo') but there's also a
                        // module augmentation by the specific name requested ('a.foo'), we store the merged symbol
                        // by the augmentation name ('a.foo'), because asking for *.foo should not give you exports
                        // from a.foo.
                        var augmentation = patternAmbientModuleAugmentations && patternAmbientModuleAugmentations.get(moduleReference);
                        if (augmentation) {
                            return getMergedSymbol(augmentation);
                        }
                        return getMergedSymbol(pattern.symbol);
                    }
                }
                // May be an untyped module. If so, ignore resolutionDiagnostic.
                if (resolvedModule && !ts.resolutionExtensionIsTSOrJson(resolvedModule.extension) && resolutionDiagnostic === undefined || resolutionDiagnostic === ts.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type) {
                    if (isForAugmentation) {
                        var diag = ts.Diagnostics.Invalid_module_name_in_augmentation_Module_0_resolves_to_an_untyped_module_at_1_which_cannot_be_augmented;
                        error(errorNode, diag, moduleReference, resolvedModule.resolvedFileName);
                    }
                    else {
                        errorOnImplicitAnyModule(/*isError*/ noImplicitAny && !!moduleNotFoundError, errorNode, resolvedModule, moduleReference);
                    }
                    // Failed imports and untyped modules are both treated in an untyped manner; only difference is whether we give a diagnostic first.
                    return undefined;
                }
                if (moduleNotFoundError) {
                    // See if this was possibly a projectReference redirect
                    if (resolvedModule) {
                        var redirect = host.getProjectReferenceRedirect(resolvedModule.resolvedFileName);
                        if (redirect) {
                            error(errorNode, ts.Diagnostics.Output_file_0_has_not_been_built_from_source_file_1, redirect, resolvedModule.resolvedFileName);
                            return undefined;
                        }
                    }
                    if (resolutionDiagnostic) {
                        error(errorNode, resolutionDiagnostic, moduleReference, resolvedModule.resolvedFileName);
                    }
                    else {
                        var tsExtension = ts.tryExtractTSExtension(moduleReference);
                        if (tsExtension) {
                            var diag = ts.Diagnostics.An_import_path_cannot_end_with_a_0_extension_Consider_importing_1_instead;
                            var importSourceWithoutExtension = ts.removeExtension(moduleReference, tsExtension);
                            var replacedImportSource = importSourceWithoutExtension;
                            /**
                             * Direct users to import source with .js extension if outputting an ES module.
                             * @see https://github.com/microsoft/TypeScript/issues/42151
                             */
                            var moduleKind_1 = ts.getEmitModuleKind(compilerOptions);
                            if (moduleKind_1 >= ts.ModuleKind.ES2015) {
                                replacedImportSource += ".js";
                            }
                            error(errorNode, diag, tsExtension, replacedImportSource);
                        }
                        else if (!compilerOptions.resolveJsonModule &&
                            ts.fileExtensionIs(moduleReference, ".json" /* Json */) &&
                            ts.getEmitModuleResolutionKind(compilerOptions) === ts.ModuleResolutionKind.NodeJs &&
                            ts.hasJsonModuleEmitEnabled(compilerOptions)) {
                            error(errorNode, ts.Diagnostics.Cannot_find_module_0_Consider_using_resolveJsonModule_to_import_module_with_json_extension, moduleReference);
                        }
                        else {
                            error(errorNode, moduleNotFoundError, moduleReference);
                        }
                    }
                }
                return undefined;
            }
            function errorOnImplicitAnyModule(isError, errorNode, _a, moduleReference) {
                var packageId = _a.packageId, resolvedFileName = _a.resolvedFileName;
                var errorInfo = !ts.isExternalModuleNameRelative(moduleReference) && packageId
                    ? typesPackageExists(packageId.name)
                        ? ts.chainDiagnosticMessages(
                        /*details*/ undefined, ts.Diagnostics.If_the_0_package_actually_exposes_this_module_consider_sending_a_pull_request_to_amend_https_Colon_Slash_Slashgithub_com_SlashDefinitelyTyped_SlashDefinitelyTyped_Slashtree_Slashmaster_Slashtypes_Slash_1, packageId.name, ts.mangleScopedPackageName(packageId.name))
                        : ts.chainDiagnosticMessages(
                        /*details*/ undefined, ts.Diagnostics.Try_npm_i_save_dev_types_Slash_1_if_it_exists_or_add_a_new_declaration_d_ts_file_containing_declare_module_0, moduleReference, ts.mangleScopedPackageName(packageId.name))
                    : undefined;
                errorOrSuggestion(isError, errorNode, ts.chainDiagnosticMessages(errorInfo, ts.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type, moduleReference, resolvedFileName));
            }
            function typesPackageExists(packageName) {
                return getPackagesSet().has(ts.getTypesPackageName(packageName));
            }
            function resolveExternalModuleSymbol(moduleSymbol, dontResolveAlias) {
                if (moduleSymbol === null || moduleSymbol === void 0 ? void 0 : moduleSymbol.exports) {
                    var exportEquals = resolveSymbol(moduleSymbol.exports.get("export=" /* ExportEquals */), dontResolveAlias);
                    var exported = getCommonJsExportEquals(getMergedSymbol(exportEquals), getMergedSymbol(moduleSymbol));
                    return getMergedSymbol(exported) || moduleSymbol;
                }
                return undefined;
            }
            function getCommonJsExportEquals(exported, moduleSymbol) {
                if (!exported || exported === unknownSymbol || exported === moduleSymbol || moduleSymbol.exports.size === 1 || exported.flags & 2097152 /* Alias */) {
                    return exported;
                }
                var links = getSymbolLinks(exported);
                if (links.cjsExportMerged) {
                    return links.cjsExportMerged;
                }
                var merged = exported.flags & 33554432 /* Transient */ ? exported : cloneSymbol(exported);
                merged.flags = merged.flags | 512 /* ValueModule */;
                if (merged.exports === undefined) {
                    merged.exports = ts.createSymbolTable();
                }
                moduleSymbol.exports.forEach(function (s, name) {
                    if (name === "export=" /* ExportEquals */)
                        return;
                    merged.exports.set(name, merged.exports.has(name) ? mergeSymbol(merged.exports.get(name), s) : s);
                });
                getSymbolLinks(merged).cjsExportMerged = merged;
                return links.cjsExportMerged = merged;
            }
            // An external module with an 'export =' declaration may be referenced as an ES6 module provided the 'export ='
            // references a symbol that is at least declared as a module or a variable. The target of the 'export =' may
            // combine other declarations with the module or variable (e.g. a class/module, function/module, interface/variable).
            function resolveESModuleSymbol(moduleSymbol, referencingLocation, dontResolveAlias, suppressInteropError) {
                var symbol = resolveExternalModuleSymbol(moduleSymbol, dontResolveAlias);
                if (!dontResolveAlias && symbol) {
                    if (!suppressInteropError && !(symbol.flags & (1536 /* Module */ | 3 /* Variable */)) && !ts.getDeclarationOfKind(symbol, 300 /* SourceFile */)) {
                        var compilerOptionName = moduleKind >= ts.ModuleKind.ES2015
                            ? "allowSyntheticDefaultImports"
                            : "esModuleInterop";
                        error(referencingLocation, ts.Diagnostics.This_module_can_only_be_referenced_with_ECMAScript_imports_Slashexports_by_turning_on_the_0_flag_and_referencing_its_default_export, compilerOptionName);
                        return symbol;
                    }
                    if (compilerOptions.esModuleInterop) {
                        var referenceParent = referencingLocation.parent;
                        if ((ts.isImportDeclaration(referenceParent) && ts.getNamespaceDeclarationNode(referenceParent)) ||
                            ts.isImportCall(referenceParent)) {
                            var type = getTypeOfSymbol(symbol);
                            var sigs = getSignaturesOfStructuredType(type, 0 /* Call */);
                            if (!sigs || !sigs.length) {
                                sigs = getSignaturesOfStructuredType(type, 1 /* Construct */);
                            }
                            if (sigs && sigs.length) {
                                var moduleType = getTypeWithSyntheticDefaultImportType(type, symbol, moduleSymbol);
                                // Create a new symbol which has the module's type less the call and construct signatures
                                var result = createSymbol(symbol.flags, symbol.escapedName);
                                result.declarations = symbol.declarations ? symbol.declarations.slice() : [];
                                result.parent = symbol.parent;
                                result.target = symbol;
                                result.originatingImport = referenceParent;
                                if (symbol.valueDeclaration)
                                    result.valueDeclaration = symbol.valueDeclaration;
                                if (symbol.constEnumOnlyModule)
                                    result.constEnumOnlyModule = true;
                                if (symbol.members)
                                    result.members = new ts.Map(symbol.members);
                                if (symbol.exports)
                                    result.exports = new ts.Map(symbol.exports);
                                var resolvedModuleType = resolveStructuredTypeMembers(moduleType); // Should already be resolved from the signature checks above
                                result.type = createAnonymousType(result, resolvedModuleType.members, ts.emptyArray, ts.emptyArray, resolvedModuleType.indexInfos);
                                return result;
                            }
                        }
                    }
                }
                return symbol;
            }
            function hasExportAssignmentSymbol(moduleSymbol) {
                return moduleSymbol.exports.get("export=" /* ExportEquals */) !== undefined;
            }
            function getExportsOfModuleAsArray(moduleSymbol) {
                return symbolsToArray(getExportsOfModule(moduleSymbol));
            }
            function getExportsAndPropertiesOfModule(moduleSymbol) {
                var exports = getExportsOfModuleAsArray(moduleSymbol);
                var exportEquals = resolveExternalModuleSymbol(moduleSymbol);
                if (exportEquals !== moduleSymbol) {
                    var type = getTypeOfSymbol(exportEquals);
                    if (shouldTreatPropertiesOfExternalModuleAsExports(type)) {
                        ts.addRange(exports, getPropertiesOfType(type));
                    }
                }
                return exports;
            }
            function forEachExportAndPropertyOfModule(moduleSymbol, cb) {
                var exports = getExportsOfModule(moduleSymbol);
                exports.forEach(function (symbol, key) {
                    if (!isReservedMemberName(key)) {
                        cb(symbol, key);
                    }
                });
                var exportEquals = resolveExternalModuleSymbol(moduleSymbol);
                if (exportEquals !== moduleSymbol) {
                    var type = getTypeOfSymbol(exportEquals);
                    if (shouldTreatPropertiesOfExternalModuleAsExports(type)) {
                        getPropertiesOfType(type).forEach(function (symbol) {
                            cb(symbol, symbol.escapedName);
                        });
                    }
                }
            }
            function tryGetMemberInModuleExports(memberName, moduleSymbol) {
                var symbolTable = getExportsOfModule(moduleSymbol);
                if (symbolTable) {
                    return symbolTable.get(memberName);
                }
            }
            function tryGetMemberInModuleExportsAndProperties(memberName, moduleSymbol) {
                var symbol = tryGetMemberInModuleExports(memberName, moduleSymbol);
                if (symbol) {
                    return symbol;
                }
                var exportEquals = resolveExternalModuleSymbol(moduleSymbol);
                if (exportEquals === moduleSymbol) {
                    return undefined;
                }
                var type = getTypeOfSymbol(exportEquals);
                return shouldTreatPropertiesOfExternalModuleAsExports(type) ? getPropertyOfType(type, memberName) : undefined;
            }
            function shouldTreatPropertiesOfExternalModuleAsExports(resolvedExternalModuleType) {
                return !(resolvedExternalModuleType.flags & 131068 /* Primitive */ ||
                    ts.getObjectFlags(resolvedExternalModuleType) & 1 /* Class */ ||
                    // `isArrayOrTupleLikeType` is too expensive to use in this auto-imports hot path
                    isArrayType(resolvedExternalModuleType) ||
                    isTupleType(resolvedExternalModuleType));
            }
            function getExportsOfSymbol(symbol) {
                return symbol.flags & 6256 /* LateBindingContainer */ ? getResolvedMembersOrExportsOfSymbol(symbol, "resolvedExports" /* resolvedExports */) :
                    symbol.flags & 1536 /* Module */ ? getExportsOfModule(symbol) :
                        symbol.exports || emptySymbols;
            }
            function getExportsOfModule(moduleSymbol) {
                var links = getSymbolLinks(moduleSymbol);
                return links.resolvedExports || (links.resolvedExports = getExportsOfModuleWorker(moduleSymbol));
            }
            /**
             * Extends one symbol table with another while collecting information on name collisions for error message generation into the `lookupTable` argument
             * Not passing `lookupTable` and `exportNode` disables this collection, and just extends the tables
             */
            function extendExportSymbols(target, source, lookupTable, exportNode) {
                if (!source)
                    return;
                source.forEach(function (sourceSymbol, id) {
                    if (id === "default" /* Default */)
                        return;
                    var targetSymbol = target.get(id);
                    if (!targetSymbol) {
                        target.set(id, sourceSymbol);
                        if (lookupTable && exportNode) {
                            lookupTable.set(id, {
                                specifierText: ts.getTextOfNode(exportNode.moduleSpecifier)
                            });
                        }
                    }
                    else if (lookupTable && exportNode && targetSymbol && resolveSymbol(targetSymbol) !== resolveSymbol(sourceSymbol)) {
                        var collisionTracker = lookupTable.get(id);
                        if (!collisionTracker.exportsWithDuplicate) {
                            collisionTracker.exportsWithDuplicate = [exportNode];
                        }
                        else {
                            collisionTracker.exportsWithDuplicate.push(exportNode);
                        }
                    }
                });
            }
            function getExportsOfModuleWorker(moduleSymbol) {
                var visitedSymbols = [];
                // A module defined by an 'export=' consists of one export that needs to be resolved
                moduleSymbol = resolveExternalModuleSymbol(moduleSymbol);
                return visit(moduleSymbol) || emptySymbols;
                // The ES6 spec permits export * declarations in a module to circularly reference the module itself. For example,
                // module 'a' can 'export * from "b"' and 'b' can 'export * from "a"' without error.
                function visit(symbol) {
                    if (!(symbol && symbol.exports && ts.pushIfUnique(visitedSymbols, symbol))) {
                        return;
                    }
                    var symbols = new ts.Map(symbol.exports);
                    // All export * declarations are collected in an __export symbol by the binder
                    var exportStars = symbol.exports.get("__export" /* ExportStar */);
                    if (exportStars) {
                        var nestedSymbols = ts.createSymbolTable();
                        var lookupTable_1 = new ts.Map();
                        if (exportStars.declarations) {
                            for (var _i = 0, _a = exportStars.declarations; _i < _a.length; _i++) {
                                var node = _a[_i];
                                var resolvedModule = resolveExternalModuleName(node, node.moduleSpecifier);
                                var exportedSymbols = visit(resolvedModule);
                                extendExportSymbols(nestedSymbols, exportedSymbols, lookupTable_1, node);
                            }
                        }
                        lookupTable_1.forEach(function (_a, id) {
                            var exportsWithDuplicate = _a.exportsWithDuplicate;
                            // It's not an error if the file with multiple `export *`s with duplicate names exports a member with that name itself
                            if (id === "export=" || !(exportsWithDuplicate && exportsWithDuplicate.length) || symbols.has(id)) {
                                return;
                            }
                            for (var _i = 0, exportsWithDuplicate_1 = exportsWithDuplicate; _i < exportsWithDuplicate_1.length; _i++) {
                                var node = exportsWithDuplicate_1[_i];
                                diagnostics.add(ts.createDiagnosticForNode(node, ts.Diagnostics.Module_0_has_already_exported_a_member_named_1_Consider_explicitly_re_exporting_to_resolve_the_ambiguity, lookupTable_1.get(id).specifierText, ts.unescapeLeadingUnderscores(id)));
                            }
                        });
                        extendExportSymbols(symbols, nestedSymbols);
                    }
                    return symbols;
                }
            }
            function getMergedSymbol(symbol) {
                var merged;
                return symbol && symbol.mergeId && (merged = mergedSymbols[symbol.mergeId]) ? merged : symbol;
            }
            function getSymbolOfNode(node) {
                return getMergedSymbol(node.symbol && getLateBoundSymbol(node.symbol));
            }
            function getParentOfSymbol(symbol) {
                return getMergedSymbol(symbol.parent && getLateBoundSymbol(symbol.parent));
            }
            function getAlternativeContainingModules(symbol, enclosingDeclaration) {
                var containingFile = ts.getSourceFileOfNode(enclosingDeclaration);
                var id = getNodeId(containingFile);
                var links = getSymbolLinks(symbol);
                var results;
                if (links.extendedContainersByFile && (results = links.extendedContainersByFile.get(id))) {
                    return results;
                }
                if (containingFile && containingFile.imports) {
                    // Try to make an import using an import already in the enclosing file, if possible
                    for (var _i = 0, _a = containingFile.imports; _i < _a.length; _i++) {
                        var importRef = _a[_i];
                        if (ts.nodeIsSynthesized(importRef))
                            continue; // Synthetic names can't be resolved by `resolveExternalModuleName` - they'll cause a debug assert if they error
                        var resolvedModule = resolveExternalModuleName(enclosingDeclaration, importRef, /*ignoreErrors*/ true);
                        if (!resolvedModule)
                            continue;
                        var ref = getAliasForSymbolInContainer(resolvedModule, symbol);
                        if (!ref)
                            continue;
                        results = ts.append(results, resolvedModule);
                    }
                    if (ts.length(results)) {
                        (links.extendedContainersByFile || (links.extendedContainersByFile = new ts.Map())).set(id, results);
                        return results;
                    }
                }
                if (links.extendedContainers) {
                    return links.extendedContainers;
                }
                // No results from files already being imported by this file - expand search (expensive, but not location-specific, so cached)
                var otherFiles = host.getSourceFiles();
                for (var _b = 0, otherFiles_1 = otherFiles; _b < otherFiles_1.length; _b++) {
                    var file = otherFiles_1[_b];
                    if (!ts.isExternalModule(file))
                        continue;
                    var sym = getSymbolOfNode(file);
                    var ref = getAliasForSymbolInContainer(sym, symbol);
                    if (!ref)
                        continue;
                    results = ts.append(results, sym);
                }
                return links.extendedContainers = results || ts.emptyArray;
            }
            /**
             * Attempts to find the symbol corresponding to the container a symbol is in - usually this
             * is just its' `.parent`, but for locals, this value is `undefined`
             */
            function getContainersOfSymbol(symbol, enclosingDeclaration, meaning) {
                var container = getParentOfSymbol(symbol);
                // Type parameters end up in the `members` lists but are not externally visible
                if (container && !(symbol.flags & 262144 /* TypeParameter */)) {
                    var additionalContainers = ts.mapDefined(container.declarations, fileSymbolIfFileSymbolExportEqualsContainer);
                    var reexportContainers = enclosingDeclaration && getAlternativeContainingModules(symbol, enclosingDeclaration);
                    var objectLiteralContainer = getVariableDeclarationOfObjectLiteral(container, meaning);
                    if (enclosingDeclaration &&
                        container.flags & getQualifiedLeftMeaning(meaning) &&
                        getAccessibleSymbolChain(container, enclosingDeclaration, 1920 /* Namespace */, /*externalOnly*/ false)) {
                        return ts.append(ts.concatenate(ts.concatenate([container], additionalContainers), reexportContainers), objectLiteralContainer); // This order expresses a preference for the real container if it is in scope
                    }
                    // we potentially have a symbol which is a member of the instance side of something - look for a variable in scope with the container's type
                    // which may be acting like a namespace (eg, `Symbol` acts like a namespace when looking up `Symbol.toStringTag`)
                    var firstVariableMatch = !(container.flags & getQualifiedLeftMeaning(meaning))
                        && container.flags & 788968 /* Type */
                        && getDeclaredTypeOfSymbol(container).flags & 524288 /* Object */
                        && meaning === 111551 /* Value */
                        ? forEachSymbolTableInScope(enclosingDeclaration, function (t) {
                            return ts.forEachEntry(t, function (s) {
                                if (s.flags & getQualifiedLeftMeaning(meaning) && getTypeOfSymbol(s) === getDeclaredTypeOfSymbol(container)) {
                                    return s;
                                }
                            });
                        }) : undefined;
                    var res = firstVariableMatch ? __spreadArray(__spreadArray([firstVariableMatch], additionalContainers, true), [container], false) : __spreadArray(__spreadArray([], additionalContainers, true), [container], false);
                    res = ts.append(res, objectLiteralContainer);
                    res = ts.addRange(res, reexportContainers);
                    return res;
                }
                var candidates = ts.mapDefined(symbol.declarations, function (d) {
                    if (!ts.isAmbientModule(d) && d.parent && hasNonGlobalAugmentationExternalModuleSymbol(d.parent)) {
                        return getSymbolOfNode(d.parent);
                    }
                    if (ts.isClassExpression(d) && ts.isBinaryExpression(d.parent) && d.parent.operatorToken.kind === 63 /* EqualsToken */ && ts.isAccessExpression(d.parent.left) && ts.isEntityNameExpression(d.parent.left.expression)) {
                        if (ts.isModuleExportsAccessExpression(d.parent.left) || ts.isExportsIdentifier(d.parent.left.expression)) {
                            return getSymbolOfNode(ts.getSourceFileOfNode(d));
                        }
                        checkExpressionCached(d.parent.left.expression);
                        return getNodeLinks(d.parent.left.expression).resolvedSymbol;
                    }
                });
                if (!ts.length(candidates)) {
                    return undefined;
                }
                return ts.mapDefined(candidates, function (candidate) { return getAliasForSymbolInContainer(candidate, symbol) ? candidate : undefined; });
                function fileSymbolIfFileSymbolExportEqualsContainer(d) {
                    return container && getFileSymbolIfFileSymbolExportEqualsContainer(d, container);
                }
            }
            function getVariableDeclarationOfObjectLiteral(symbol, meaning) {
                // If we're trying to reference some object literal in, eg `var a = { x: 1 }`, the symbol for the literal, `__object`, is distinct
                // from the symbol of the declaration it is being assigned to. Since we can use the declaration to refer to the literal, however,
                // we'd like to make that connection here - potentially causing us to paint the declaration's visibility, and therefore the literal.
                var firstDecl = !!ts.length(symbol.declarations) && ts.first(symbol.declarations);
                if (meaning & 111551 /* Value */ && firstDecl && firstDecl.parent && ts.isVariableDeclaration(firstDecl.parent)) {
                    if (ts.isObjectLiteralExpression(firstDecl) && firstDecl === firstDecl.parent.initializer || ts.isTypeLiteralNode(firstDecl) && firstDecl === firstDecl.parent.type) {
                        return getSymbolOfNode(firstDecl.parent);
                    }
                }
            }
            function getFileSymbolIfFileSymbolExportEqualsContainer(d, container) {
                var fileSymbol = getExternalModuleContainer(d);
                var exported = fileSymbol && fileSymbol.exports && fileSymbol.exports.get("export=" /* ExportEquals */);
                return exported && getSymbolIfSameReference(exported, container) ? fileSymbol : undefined;
            }
            function getAliasForSymbolInContainer(container, symbol) {
                if (container === getParentOfSymbol(symbol)) {
                    // fast path, `symbol` is either already the alias or isn't aliased
                    return symbol;
                }
                // Check if container is a thing with an `export=` which points directly at `symbol`, and if so, return
                // the container itself as the alias for the symbol
                var exportEquals = container.exports && container.exports.get("export=" /* ExportEquals */);
                if (exportEquals && getSymbolIfSameReference(exportEquals, symbol)) {
                    return container;
                }
                var exports = getExportsOfSymbol(container);
                var quick = exports.get(symbol.escapedName);
                if (quick && getSymbolIfSameReference(quick, symbol)) {
                    return quick;
                }
                return ts.forEachEntry(exports, function (exported) {
                    if (getSymbolIfSameReference(exported, symbol)) {
                        return exported;
                    }
                });
            }
            /**
             * Checks if two symbols, through aliasing and/or merging, refer to the same thing
             */
            function getSymbolIfSameReference(s1, s2) {
                if (getMergedSymbol(resolveSymbol(getMergedSymbol(s1))) === getMergedSymbol(resolveSymbol(getMergedSymbol(s2)))) {
                    return s1;
                }
            }
            function getExportSymbolOfValueSymbolIfExported(symbol) {
                return getMergedSymbol(symbol && (symbol.flags & 1048576 /* ExportValue */) !== 0 ? symbol.exportSymbol : symbol);
            }
            function symbolIsValue(symbol) {
                return !!(symbol.flags & 111551 /* Value */ || symbol.flags & 2097152 /* Alias */ && resolveAlias(symbol).flags & 111551 /* Value */ && !getTypeOnlyAliasDeclaration(symbol));
            }
            function findConstructorDeclaration(node) {
                var members = node.members;
                for (var _i = 0, members_3 = members; _i < members_3.length; _i++) {
                    var member = members_3[_i];
                    if (member.kind === 169 /* Constructor */ && ts.nodeIsPresent(member.body)) {
                        return member;
                    }
                }
            }
            function createType(flags) {
                var result = new Type(checker, flags);
                typeCount++;
                result.id = typeCount;
                if (produceDiagnostics) { // Only record types from one checker
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.recordType(result);
                }
                return result;
            }
            function createOriginType(flags) {
                return new Type(checker, flags);
            }
            function createIntrinsicType(kind, intrinsicName, objectFlags) {
                if (objectFlags === void 0) {
                    objectFlags = 0;
                }
                var type = createType(kind);
                type.intrinsicName = intrinsicName;
                type.objectFlags = objectFlags;
                return type;
            }
            function createObjectType(objectFlags, symbol) {
                var type = createType(524288 /* Object */);
                type.objectFlags = objectFlags;
                type.symbol = symbol;
                type.members = undefined;
                type.properties = undefined;
                type.callSignatures = undefined;
                type.constructSignatures = undefined;
                type.indexInfos = undefined;
                return type;
            }
            function createTypeofType() {
                return getUnionType(ts.arrayFrom(typeofEQFacts.keys(), getStringLiteralType));
            }
            function createTypeParameter(symbol) {
                var type = createType(262144 /* TypeParameter */);
                if (symbol)
                    type.symbol = symbol;
                return type;
            }
            // A reserved member name starts with two underscores, but the third character cannot be an underscore,
            // @, or #. A third underscore indicates an escaped form of an identifier that started
            // with at least two underscores. The @ character indicates that the name is denoted by a well known ES
            // Symbol instance and the # character indicates that the name is a PrivateIdentifier.
            function isReservedMemberName(name) {
                return name.charCodeAt(0) === 95 /* _ */ &&
                    name.charCodeAt(1) === 95 /* _ */ &&
                    name.charCodeAt(2) !== 95 /* _ */ &&
                    name.charCodeAt(2) !== 64 /* at */ &&
                    name.charCodeAt(2) !== 35 /* hash */;
            }
            function getNamedMembers(members) {
                var result;
                members.forEach(function (symbol, id) {
                    if (!isReservedMemberName(id) && symbolIsValue(symbol)) {
                        (result || (result = [])).push(symbol);
                    }
                });
                return result || ts.emptyArray;
            }
            function getNamedOrIndexSignatureMembers(members) {
                var result = getNamedMembers(members);
                var index = getIndexSymbolFromSymbolTable(members);
                return index ? ts.concatenate(result, [index]) : result;
            }
            function setStructuredTypeMembers(type, members, callSignatures, constructSignatures, indexInfos) {
                var resolved = type;
                resolved.members = members;
                resolved.properties = ts.emptyArray;
                resolved.callSignatures = callSignatures;
                resolved.constructSignatures = constructSignatures;
                resolved.indexInfos = indexInfos;
                // This can loop back to getPropertyOfType() which would crash if `callSignatures` & `constructSignatures` are not initialized.
                if (members !== emptySymbols)
                    resolved.properties = getNamedMembers(members);
                return resolved;
            }
            function createAnonymousType(symbol, members, callSignatures, constructSignatures, indexInfos) {
                return setStructuredTypeMembers(createObjectType(16 /* Anonymous */, symbol), members, callSignatures, constructSignatures, indexInfos);
            }
            function getResolvedTypeWithoutAbstractConstructSignatures(type) {
                if (type.constructSignatures.length === 0)
                    return type;
                if (type.objectTypeWithoutAbstractConstructSignatures)
                    return type.objectTypeWithoutAbstractConstructSignatures;
                var constructSignatures = ts.filter(type.constructSignatures, function (signature) { return !(signature.flags & 4 /* Abstract */); });
                if (type.constructSignatures === constructSignatures)
                    return type;
                var typeCopy = createAnonymousType(type.symbol, type.members, type.callSignatures, ts.some(constructSignatures) ? constructSignatures : ts.emptyArray, type.indexInfos);
                type.objectTypeWithoutAbstractConstructSignatures = typeCopy;
                typeCopy.objectTypeWithoutAbstractConstructSignatures = typeCopy;
                return typeCopy;
            }
            function forEachSymbolTableInScope(enclosingDeclaration, callback) {
                var result;
                var _loop_8 = function (location) {
                    // Locals of a source file are not in scope (because they get merged into the global symbol table)
                    if (location.locals && !isGlobalSourceFile(location)) {
                        if (result = callback(location.locals, /*ignoreQualification*/ undefined, /*isLocalNameLookup*/ true, location)) {
                            return { value: result };
                        }
                    }
                    switch (location.kind) {
                        case 300 /* SourceFile */:
                            if (!ts.isExternalOrCommonJsModule(location)) {
                                break;
                            }
                        // falls through
                        case 259 /* ModuleDeclaration */:
                            var sym = getSymbolOfNode(location);
                            // `sym` may not have exports if this module declaration is backed by the symbol for a `const` that's being rewritten
                            // into a namespace - in such cases, it's best to just let the namespace appear empty (the const members couldn't have referred
                            // to one another anyway)
                            if (result = callback((sym === null || sym === void 0 ? void 0 : sym.exports) || emptySymbols, /*ignoreQualification*/ undefined, /*isLocalNameLookup*/ true, location)) {
                                return { value: result };
                            }
                            break;
                        case 255 /* ClassDeclaration */:
                        case 224 /* ClassExpression */:
                        case 256 /* InterfaceDeclaration */:
                            // Type parameters are bound into `members` lists so they can merge across declarations
                            // This is troublesome, since in all other respects, they behave like locals :cries:
                            // TODO: the below is shared with similar code in `resolveName` - in fact, rephrasing all this symbol
                            // lookup logic in terms of `resolveName` would be nice
                            // The below is used to lookup type parameters within a class or interface, as they are added to the class/interface locals
                            // These can never be latebound, so the symbol's raw members are sufficient. `getMembersOfNode` cannot be used, as it would
                            // trigger resolving late-bound names, which we may already be in the process of doing while we're here!
                            var table_1;
                            // TODO: Should this filtered table be cached in some way?
                            (getSymbolOfNode(location).members || emptySymbols).forEach(function (memberSymbol, key) {
                                if (memberSymbol.flags & (788968 /* Type */ & ~67108864 /* Assignment */)) {
                                    (table_1 || (table_1 = ts.createSymbolTable())).set(key, memberSymbol);
                                }
                            });
                            if (table_1 && (result = callback(table_1, /*ignoreQualification*/ undefined, /*isLocalNameLookup*/ false, location))) {
                                return { value: result };
                            }
                            break;
                    }
                };
                for (var location = enclosingDeclaration; location; location = location.parent) {
                    var state_2 = _loop_8(location);
                    if (typeof state_2 === "object")
                        return state_2.value;
                }
                return callback(globals, /*ignoreQualification*/ undefined, /*isLocalNameLookup*/ true);
            }
            function getQualifiedLeftMeaning(rightMeaning) {
                // If we are looking in value space, the parent meaning is value, other wise it is namespace
                return rightMeaning === 111551 /* Value */ ? 111551 /* Value */ : 1920 /* Namespace */;
            }
            function getAccessibleSymbolChain(symbol, enclosingDeclaration, meaning, useOnlyExternalAliasing, visitedSymbolTablesMap) {
                if (visitedSymbolTablesMap === void 0) {
                    visitedSymbolTablesMap = new ts.Map();
                }
                if (!(symbol && !isPropertyOrMethodDeclarationSymbol(symbol))) {
                    return undefined;
                }
                var links = getSymbolLinks(symbol);
                var cache = (links.accessibleChainCache || (links.accessibleChainCache = new ts.Map()));
                // Go from enclosingDeclaration to the first scope we check, so the cache is keyed off the scope and thus shared more
                var firstRelevantLocation = forEachSymbolTableInScope(enclosingDeclaration, function (_, __, ___, node) { return node; });
                var key = (useOnlyExternalAliasing ? 0 : 1) + "|" + (firstRelevantLocation && getNodeId(firstRelevantLocation)) + "|" + meaning;
                if (cache.has(key)) {
                    return cache.get(key);
                }
                var id = getSymbolId(symbol);
                var visitedSymbolTables = visitedSymbolTablesMap.get(id);
                if (!visitedSymbolTables) {
                    visitedSymbolTablesMap.set(id, visitedSymbolTables = []);
                }
                var result = forEachSymbolTableInScope(enclosingDeclaration, getAccessibleSymbolChainFromSymbolTable);
                cache.set(key, result);
                return result;
                /**
                 * @param {ignoreQualification} boolean Set when a symbol is being looked for through the exports of another symbol (meaning we have a route to qualify it already)
                 */
                function getAccessibleSymbolChainFromSymbolTable(symbols, ignoreQualification, isLocalNameLookup) {
                    if (!ts.pushIfUnique(visitedSymbolTables, symbols)) {
                        return undefined;
                    }
                    var result = trySymbolTable(symbols, ignoreQualification, isLocalNameLookup);
                    visitedSymbolTables.pop();
                    return result;
                }
                function canQualifySymbol(symbolFromSymbolTable, meaning) {
                    // If the symbol is equivalent and doesn't need further qualification, this symbol is accessible
                    return !needsQualification(symbolFromSymbolTable, enclosingDeclaration, meaning) ||
                        // If symbol needs qualification, make sure that parent is accessible, if it is then this symbol is accessible too
                        !!getAccessibleSymbolChain(symbolFromSymbolTable.parent, enclosingDeclaration, getQualifiedLeftMeaning(meaning), useOnlyExternalAliasing, visitedSymbolTablesMap);
                }
                function isAccessible(symbolFromSymbolTable, resolvedAliasSymbol, ignoreQualification) {
                    return (symbol === (resolvedAliasSymbol || symbolFromSymbolTable) || getMergedSymbol(symbol) === getMergedSymbol(resolvedAliasSymbol || symbolFromSymbolTable)) &&
                        // if the symbolFromSymbolTable is not external module (it could be if it was determined as ambient external module and would be in globals table)
                        // and if symbolFromSymbolTable or alias resolution matches the symbol,
                        // check the symbol can be qualified, it is only then this symbol is accessible
                        !ts.some(symbolFromSymbolTable.declarations, hasNonGlobalAugmentationExternalModuleSymbol) &&
                        (ignoreQualification || canQualifySymbol(getMergedSymbol(symbolFromSymbolTable), meaning));
                }
                function trySymbolTable(symbols, ignoreQualification, isLocalNameLookup) {
                    // If symbol is directly available by its name in the symbol table
                    if (isAccessible(symbols.get(symbol.escapedName), /*resolvedAliasSymbol*/ undefined, ignoreQualification)) {
                        return [symbol];
                    }
                    // Check if symbol is any of the aliases in scope
                    var result = ts.forEachEntry(symbols, function (symbolFromSymbolTable) {
                        if (symbolFromSymbolTable.flags & 2097152 /* Alias */
                            && symbolFromSymbolTable.escapedName !== "export=" /* ExportEquals */
                            && symbolFromSymbolTable.escapedName !== "default" /* Default */
                            && !(ts.isUMDExportSymbol(symbolFromSymbolTable) && enclosingDeclaration && ts.isExternalModule(ts.getSourceFileOfNode(enclosingDeclaration)))
                            // If `!useOnlyExternalAliasing`, we can use any type of alias to get the name
                            && (!useOnlyExternalAliasing || ts.some(symbolFromSymbolTable.declarations, ts.isExternalModuleImportEqualsDeclaration))
                            // If we're looking up a local name to reference directly, omit namespace reexports, otherwise when we're trawling through an export list to make a dotted name, we can keep it
                            && (isLocalNameLookup ? !ts.some(symbolFromSymbolTable.declarations, ts.isNamespaceReexportDeclaration) : true)
                            // While exports are generally considered to be in scope, export-specifier declared symbols are _not_
                            // See similar comment in `resolveName` for details
                            && (ignoreQualification || !ts.getDeclarationOfKind(symbolFromSymbolTable, 273 /* ExportSpecifier */))) {
                            var resolvedImportedSymbol = resolveAlias(symbolFromSymbolTable);
                            var candidate = getCandidateListForSymbol(symbolFromSymbolTable, resolvedImportedSymbol, ignoreQualification);
                            if (candidate) {
                                return candidate;
                            }
                        }
                        if (symbolFromSymbolTable.escapedName === symbol.escapedName && symbolFromSymbolTable.exportSymbol) {
                            if (isAccessible(getMergedSymbol(symbolFromSymbolTable.exportSymbol), /*aliasSymbol*/ undefined, ignoreQualification)) {
                                return [symbol];
                            }
                        }
                    });
                    // If there's no result and we're looking at the global symbol table, treat `globalThis` like an alias and try to lookup thru that
                    return result || (symbols === globals ? getCandidateListForSymbol(globalThisSymbol, globalThisSymbol, ignoreQualification) : undefined);
                }
                function getCandidateListForSymbol(symbolFromSymbolTable, resolvedImportedSymbol, ignoreQualification) {
                    if (isAccessible(symbolFromSymbolTable, resolvedImportedSymbol, ignoreQualification)) {
                        return [symbolFromSymbolTable];
                    }
                    // Look in the exported members, if we can find accessibleSymbolChain, symbol is accessible using this chain
                    // but only if the symbolFromSymbolTable can be qualified
                    var candidateTable = getExportsOfSymbol(resolvedImportedSymbol);
                    var accessibleSymbolsFromExports = candidateTable && getAccessibleSymbolChainFromSymbolTable(candidateTable, /*ignoreQualification*/ true);
                    if (accessibleSymbolsFromExports && canQualifySymbol(symbolFromSymbolTable, getQualifiedLeftMeaning(meaning))) {
                        return [symbolFromSymbolTable].concat(accessibleSymbolsFromExports);
                    }
                }
            }
            function needsQualification(symbol, enclosingDeclaration, meaning) {
                var qualify = false;
                forEachSymbolTableInScope(enclosingDeclaration, function (symbolTable) {
                    // If symbol of this name is not available in the symbol table we are ok
                    var symbolFromSymbolTable = getMergedSymbol(symbolTable.get(symbol.escapedName));
                    if (!symbolFromSymbolTable) {
                        // Continue to the next symbol table
                        return false;
                    }
                    // If the symbol with this name is present it should refer to the symbol
                    if (symbolFromSymbolTable === symbol) {
                        // No need to qualify
                        return true;
                    }
                    // Qualify if the symbol from symbol table has same meaning as expected
                    symbolFromSymbolTable = (symbolFromSymbolTable.flags & 2097152 /* Alias */ && !ts.getDeclarationOfKind(symbolFromSymbolTable, 273 /* ExportSpecifier */)) ? resolveAlias(symbolFromSymbolTable) : symbolFromSymbolTable;
                    if (symbolFromSymbolTable.flags & meaning) {
                        qualify = true;
                        return true;
                    }
                    // Continue to the next symbol table
                    return false;
                });
                return qualify;
            }
            function isPropertyOrMethodDeclarationSymbol(symbol) {
                if (symbol.declarations && symbol.declarations.length) {
                    for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                        var declaration = _a[_i];
                        switch (declaration.kind) {
                            case 165 /* PropertyDeclaration */:
                            case 167 /* MethodDeclaration */:
                            case 170 /* GetAccessor */:
                            case 171 /* SetAccessor */:
                                continue;
                            default:
                                return false;
                        }
                    }
                    return true;
                }
                return false;
            }
            function isTypeSymbolAccessible(typeSymbol, enclosingDeclaration) {
                var access = isSymbolAccessibleWorker(typeSymbol, enclosingDeclaration, 788968 /* Type */, /*shouldComputeAliasesToMakeVisible*/ false, /*allowModules*/ true);
                return access.accessibility === 0 /* Accessible */;
            }
            function isValueSymbolAccessible(typeSymbol, enclosingDeclaration) {
                var access = isSymbolAccessibleWorker(typeSymbol, enclosingDeclaration, 111551 /* Value */, /*shouldComputeAliasesToMakeVisible*/ false, /*allowModules*/ true);
                return access.accessibility === 0 /* Accessible */;
            }
            function isSymbolAccessibleByFlags(typeSymbol, enclosingDeclaration, flags) {
                var access = isSymbolAccessibleWorker(typeSymbol, enclosingDeclaration, flags, /*shouldComputeAliasesToMakeVisible*/ false, /*allowModules*/ false);
                return access.accessibility === 0 /* Accessible */;
            }
            function isAnySymbolAccessible(symbols, enclosingDeclaration, initialSymbol, meaning, shouldComputeAliasesToMakeVisible, allowModules) {
                if (!ts.length(symbols))
                    return;
                var hadAccessibleChain;
                var earlyModuleBail = false;
                for (var _i = 0, _a = symbols; _i < _a.length; _i++) {
                    var symbol = _a[_i];
                    // Symbol is accessible if it by itself is accessible
                    var accessibleSymbolChain = getAccessibleSymbolChain(symbol, enclosingDeclaration, meaning, /*useOnlyExternalAliasing*/ false);
                    if (accessibleSymbolChain) {
                        hadAccessibleChain = symbol;
                        var hasAccessibleDeclarations = hasVisibleDeclarations(accessibleSymbolChain[0], shouldComputeAliasesToMakeVisible);
                        if (hasAccessibleDeclarations) {
                            return hasAccessibleDeclarations;
                        }
                    }
                    if (allowModules) {
                        if (ts.some(symbol.declarations, hasNonGlobalAugmentationExternalModuleSymbol)) {
                            if (shouldComputeAliasesToMakeVisible) {
                                earlyModuleBail = true;
                                // Generally speaking, we want to use the aliases that already exist to refer to a module, if present
                                // In order to do so, we need to find those aliases in order to retain them in declaration emit; so
                                // if we are in declaration emit, we cannot use the fast path for module visibility until we've exhausted
                                // all other visibility options (in order to capture the possible aliases used to reference the module)
                                continue;
                            }
                            // Any meaning of a module symbol is always accessible via an `import` type
                            return {
                                accessibility: 0 /* Accessible */
                            };
                        }
                    }
                    // If we haven't got the accessible symbol, it doesn't mean the symbol is actually inaccessible.
                    // It could be a qualified symbol and hence verify the path
                    // e.g.:
                    // module m {
                    //     export class c {
                    //     }
                    // }
                    // const x: typeof m.c
                    // In the above example when we start with checking if typeof m.c symbol is accessible,
                    // we are going to see if c can be accessed in scope directly.
                    // But it can't, hence the accessible is going to be undefined, but that doesn't mean m.c is inaccessible
                    // It is accessible if the parent m is accessible because then m.c can be accessed through qualification
                    var containers = getContainersOfSymbol(symbol, enclosingDeclaration, meaning);
                    var parentResult = isAnySymbolAccessible(containers, enclosingDeclaration, initialSymbol, initialSymbol === symbol ? getQualifiedLeftMeaning(meaning) : meaning, shouldComputeAliasesToMakeVisible, allowModules);
                    if (parentResult) {
                        return parentResult;
                    }
                }
                if (earlyModuleBail) {
                    return {
                        accessibility: 0 /* Accessible */
                    };
                }
                if (hadAccessibleChain) {
                    return {
                        accessibility: 1 /* NotAccessible */,
                        errorSymbolName: symbolToString(initialSymbol, enclosingDeclaration, meaning),
                        errorModuleName: hadAccessibleChain !== initialSymbol ? symbolToString(hadAccessibleChain, enclosingDeclaration, 1920 /* Namespace */) : undefined,
                    };
                }
            }
            /**
             * Check if the given symbol in given enclosing declaration is accessible and mark all associated alias to be visible if requested
             *
             * @param symbol a Symbol to check if accessible
             * @param enclosingDeclaration a Node containing reference to the symbol
             * @param meaning a SymbolFlags to check if such meaning of the symbol is accessible
             * @param shouldComputeAliasToMakeVisible a boolean value to indicate whether to return aliases to be mark visible in case the symbol is accessible
             */
            function isSymbolAccessible(symbol, enclosingDeclaration, meaning, shouldComputeAliasesToMakeVisible) {
                return isSymbolAccessibleWorker(symbol, enclosingDeclaration, meaning, shouldComputeAliasesToMakeVisible, /*allowModules*/ true);
            }
            function isSymbolAccessibleWorker(symbol, enclosingDeclaration, meaning, shouldComputeAliasesToMakeVisible, allowModules) {
                if (symbol && enclosingDeclaration) {
                    var result = isAnySymbolAccessible([symbol], enclosingDeclaration, symbol, meaning, shouldComputeAliasesToMakeVisible, allowModules);
                    if (result) {
                        return result;
                    }
                    // This could be a symbol that is not exported in the external module
                    // or it could be a symbol from different external module that is not aliased and hence cannot be named
                    var symbolExternalModule = ts.forEach(symbol.declarations, getExternalModuleContainer);
                    if (symbolExternalModule) {
                        var enclosingExternalModule = getExternalModuleContainer(enclosingDeclaration);
                        if (symbolExternalModule !== enclosingExternalModule) {
                            // name from different external module that is not visible
                            return {
                                accessibility: 2 /* CannotBeNamed */,
                                errorSymbolName: symbolToString(symbol, enclosingDeclaration, meaning),
                                errorModuleName: symbolToString(symbolExternalModule),
                                errorNode: ts.isInJSFile(enclosingDeclaration) ? enclosingDeclaration : undefined,
                            };
                        }
                    }
                    // Just a local name that is not accessible
                    return {
                        accessibility: 1 /* NotAccessible */,
                        errorSymbolName: symbolToString(symbol, enclosingDeclaration, meaning),
                    };
                }
                return { accessibility: 0 /* Accessible */ };
            }
            function getExternalModuleContainer(declaration) {
                var node = ts.findAncestor(declaration, hasExternalModuleSymbol);
                return node && getSymbolOfNode(node);
            }
            function hasExternalModuleSymbol(declaration) {
                return ts.isAmbientModule(declaration) || (declaration.kind === 300 /* SourceFile */ && ts.isExternalOrCommonJsModule(declaration));
            }
            function hasNonGlobalAugmentationExternalModuleSymbol(declaration) {
                return ts.isModuleWithStringLiteralName(declaration) || (declaration.kind === 300 /* SourceFile */ && ts.isExternalOrCommonJsModule(declaration));
            }
            function hasVisibleDeclarations(symbol, shouldComputeAliasToMakeVisible) {
                var aliasesToMakeVisible;
                if (!ts.every(ts.filter(symbol.declarations, function (d) { return d.kind !== 79 /* Identifier */; }), getIsDeclarationVisible)) {
                    return undefined;
                }
                return { accessibility: 0 /* Accessible */, aliasesToMakeVisible: aliasesToMakeVisible };
                function getIsDeclarationVisible(declaration) {
                    var _a, _b;
                    if (!isDeclarationVisible(declaration)) {
                        // Mark the unexported alias as visible if its parent is visible
                        // because these kind of aliases can be used to name types in declaration file
                        var anyImportSyntax = getAnyImportSyntax(declaration);
                        if (anyImportSyntax &&
                            !ts.hasSyntacticModifier(anyImportSyntax, 1 /* Export */) && // import clause without export
                            isDeclarationVisible(anyImportSyntax.parent)) {
                            return addVisibleAlias(declaration, anyImportSyntax);
                        }
                        else if (ts.isVariableDeclaration(declaration) && ts.isVariableStatement(declaration.parent.parent) &&
                            !ts.hasSyntacticModifier(declaration.parent.parent, 1 /* Export */) && // unexported variable statement
                            isDeclarationVisible(declaration.parent.parent.parent)) {
                            return addVisibleAlias(declaration, declaration.parent.parent);
                        }
                        else if (ts.isLateVisibilityPaintedStatement(declaration) // unexported top-level statement
                            && !ts.hasSyntacticModifier(declaration, 1 /* Export */)
                            && isDeclarationVisible(declaration.parent)) {
                            return addVisibleAlias(declaration, declaration);
                        }
                        else if (symbol.flags & 2097152 /* Alias */ && ts.isBindingElement(declaration) && ts.isInJSFile(declaration) && ((_a = declaration.parent) === null || _a === void 0 ? void 0 : _a.parent) // exported import-like top-level JS require statement
                            && ts.isVariableDeclaration(declaration.parent.parent)
                            && ((_b = declaration.parent.parent.parent) === null || _b === void 0 ? void 0 : _b.parent) && ts.isVariableStatement(declaration.parent.parent.parent.parent)
                            && !ts.hasSyntacticModifier(declaration.parent.parent.parent.parent, 1 /* Export */)
                            && declaration.parent.parent.parent.parent.parent // check if the thing containing the variable statement is visible (ie, the file)
                            && isDeclarationVisible(declaration.parent.parent.parent.parent.parent)) {
                            return addVisibleAlias(declaration, declaration.parent.parent.parent.parent);
                        }
                        // Declaration is not visible
                        return false;
                    }
                    return true;
                }
                function addVisibleAlias(declaration, aliasingStatement) {
                    // In function "buildTypeDisplay" where we decide whether to write type-alias or serialize types,
                    // we want to just check if type- alias is accessible or not but we don't care about emitting those alias at that time
                    // since we will do the emitting later in trackSymbol.
                    if (shouldComputeAliasToMakeVisible) {
                        getNodeLinks(declaration).isVisible = true;
                        aliasesToMakeVisible = ts.appendIfUnique(aliasesToMakeVisible, aliasingStatement);
                    }
                    return true;
                }
            }
            function isEntityNameVisible(entityName, enclosingDeclaration) {
                // get symbol of the first identifier of the entityName
                var meaning;
                if (entityName.parent.kind === 179 /* TypeQuery */ ||
                    ts.isExpressionWithTypeArgumentsInClassExtendsClause(entityName.parent) ||
                    entityName.parent.kind === 160 /* ComputedPropertyName */) {
                    // Typeof value
                    meaning = 111551 /* Value */ | 1048576 /* ExportValue */;
                }
                else if (entityName.kind === 159 /* QualifiedName */ || entityName.kind === 204 /* PropertyAccessExpression */ ||
                    entityName.parent.kind === 263 /* ImportEqualsDeclaration */) {
                    // Left identifier from type reference or TypeAlias
                    // Entity name of the import declaration
                    meaning = 1920 /* Namespace */;
                }
                else {
                    // Type Reference or TypeAlias entity = Identifier
                    meaning = 788968 /* Type */;
                }
                var firstIdentifier = ts.getFirstIdentifier(entityName);
                var symbol = resolveName(enclosingDeclaration, firstIdentifier.escapedText, meaning, /*nodeNotFoundErrorMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false);
                if (symbol && symbol.flags & 262144 /* TypeParameter */ && meaning & 788968 /* Type */) {
                    return { accessibility: 0 /* Accessible */ };
                }
                // Verify if the symbol is accessible
                return (symbol && hasVisibleDeclarations(symbol, /*shouldComputeAliasToMakeVisible*/ true)) || {
                    accessibility: 1 /* NotAccessible */,
                    errorSymbolName: ts.getTextOfNode(firstIdentifier),
                    errorNode: firstIdentifier
                };
            }
            function symbolToString(symbol, enclosingDeclaration, meaning, flags, writer) {
                if (flags === void 0) {
                    flags = 4 /* AllowAnyNodeKind */;
                }
                var nodeFlags = 70221824 /* IgnoreErrors */;
                if (flags & 2 /* UseOnlyExternalAliasing */) {
                    nodeFlags |= 128 /* UseOnlyExternalAliasing */;
                }
                if (flags & 1 /* WriteTypeParametersOrArguments */) {
                    nodeFlags |= 512 /* WriteTypeParametersInQualifiedName */;
                }
                if (flags & 8 /* UseAliasDefinedOutsideCurrentScope */) {
                    nodeFlags |= 16384 /* UseAliasDefinedOutsideCurrentScope */;
                }
                if (flags & 16 /* DoNotIncludeSymbolChain */) {
                    nodeFlags |= 134217728 /* DoNotIncludeSymbolChain */;
                }
                var builder = flags & 4 /* AllowAnyNodeKind */ ? nodeBuilder.symbolToExpression : nodeBuilder.symbolToEntityName;
                return writer ? symbolToStringWorker(writer).getText() : ts.usingSingleLineStringWriter(symbolToStringWorker);
                function symbolToStringWorker(writer) {
                    var entity = builder(symbol, meaning, enclosingDeclaration, nodeFlags); // TODO: GH#18217
                    // add neverAsciiEscape for GH#39027
                    var printer = (enclosingDeclaration === null || enclosingDeclaration === void 0 ? void 0 : enclosingDeclaration.kind) === 300 /* SourceFile */ ? ts.createPrinter({ removeComments: true, neverAsciiEscape: true }) : ts.createPrinter({ removeComments: true });
                    var sourceFile = enclosingDeclaration && ts.getSourceFileOfNode(enclosingDeclaration);
                    printer.writeNode(4 /* Unspecified */, entity, /*sourceFile*/ sourceFile, writer);
                    return writer;
                }
            }
            function signatureToString(signature, enclosingDeclaration, flags, kind, writer) {
                if (flags === void 0) {
                    flags = 0 /* None */;
                }
                return writer ? signatureToStringWorker(writer).getText() : ts.usingSingleLineStringWriter(signatureToStringWorker);
                function signatureToStringWorker(writer) {
                    var sigOutput;
                    if (flags & 262144 /* WriteArrowStyleSignature */) {
                        sigOutput = kind === 1 /* Construct */ ? 178 /* ConstructorType */ : 177 /* FunctionType */;
                    }
                    else {
                        sigOutput = kind === 1 /* Construct */ ? 173 /* ConstructSignature */ : 172 /* CallSignature */;
                    }
                    var sig = nodeBuilder.signatureToSignatureDeclaration(signature, sigOutput, enclosingDeclaration, toNodeBuilderFlags(flags) | 70221824 /* IgnoreErrors */ | 512 /* WriteTypeParametersInQualifiedName */);
                    var printer = ts.createPrinter({ removeComments: true, omitTrailingSemicolon: true });
                    var sourceFile = enclosingDeclaration && ts.getSourceFileOfNode(enclosingDeclaration);
                    printer.writeNode(4 /* Unspecified */, sig, /*sourceFile*/ sourceFile, ts.getTrailingSemicolonDeferringWriter(writer)); // TODO: GH#18217
                    return writer;
                }
            }
            function typeToString(type, enclosingDeclaration, flags, writer) {
                if (flags === void 0) {
                    flags = 1048576 /* AllowUniqueESSymbolType */ | 16384 /* UseAliasDefinedOutsideCurrentScope */;
                }
                if (writer === void 0) {
                    writer = ts.createTextWriter("");
                }
                var noTruncation = compilerOptions.noErrorTruncation || flags & 1 /* NoTruncation */;
                var typeNode = nodeBuilder.typeToTypeNode(type, enclosingDeclaration, toNodeBuilderFlags(flags) | 70221824 /* IgnoreErrors */ | (noTruncation ? 1 /* NoTruncation */ : 0), writer);
                if (typeNode === undefined)
                    return ts.Debug.fail("should always get typenode");
                var options = { removeComments: true };
                var printer = ts.createPrinter(options);
                var sourceFile = enclosingDeclaration && ts.getSourceFileOfNode(enclosingDeclaration);
                printer.writeNode(4 /* Unspecified */, typeNode, /*sourceFile*/ sourceFile, writer);
                var result = writer.getText();
                var maxLength = noTruncation ? ts.noTruncationMaximumTruncationLength * 2 : ts.defaultMaximumTruncationLength * 2;
                if (maxLength && result && result.length >= maxLength) {
                    return result.substr(0, maxLength - "...".length) + "...";
                }
                return result;
            }
            function getTypeNamesForErrorDisplay(left, right) {
                var leftStr = symbolValueDeclarationIsContextSensitive(left.symbol) ? typeToString(left, left.symbol.valueDeclaration) : typeToString(left);
                var rightStr = symbolValueDeclarationIsContextSensitive(right.symbol) ? typeToString(right, right.symbol.valueDeclaration) : typeToString(right);
                if (leftStr === rightStr) {
                    leftStr = getTypeNameForErrorDisplay(left);
                    rightStr = getTypeNameForErrorDisplay(right);
                }
                return [leftStr, rightStr];
            }
            function getTypeNameForErrorDisplay(type) {
                return typeToString(type, /*enclosingDeclaration*/ undefined, 64 /* UseFullyQualifiedType */);
            }
            function symbolValueDeclarationIsContextSensitive(symbol) {
                return symbol && !!symbol.valueDeclaration && ts.isExpression(symbol.valueDeclaration) && !isContextSensitive(symbol.valueDeclaration);
            }
            function toNodeBuilderFlags(flags) {
                if (flags === void 0) {
                    flags = 0 /* None */;
                }
                return flags & 814775659 /* NodeBuilderFlagsMask */;
            }
            function isClassInstanceSide(type) {
                return !!type.symbol && !!(type.symbol.flags & 32 /* Class */) && (type === getDeclaredTypeOfClassOrInterface(type.symbol) || (!!(type.flags & 524288 /* Object */) && !!(ts.getObjectFlags(type) & 16777216 /* IsClassInstanceClone */)));
            }
            function createNodeBuilder() {
                return {
                    typeToTypeNode: function (type, enclosingDeclaration, flags, tracker) {
                        return withContext(enclosingDeclaration, flags, tracker, function (context) { return typeToTypeNodeHelper(type, context); });
                    },
                    indexInfoToIndexSignatureDeclaration: function (indexInfo, enclosingDeclaration, flags, tracker) {
                        return withContext(enclosingDeclaration, flags, tracker, function (context) { return indexInfoToIndexSignatureDeclarationHelper(indexInfo, context, /*typeNode*/ undefined); });
                    },
                    signatureToSignatureDeclaration: function (signature, kind, enclosingDeclaration, flags, tracker) {
                        return withContext(enclosingDeclaration, flags, tracker, function (context) { return signatureToSignatureDeclarationHelper(signature, kind, context); });
                    },
                    symbolToEntityName: function (symbol, meaning, enclosingDeclaration, flags, tracker) {
                        return withContext(enclosingDeclaration, flags, tracker, function (context) { return symbolToName(symbol, context, meaning, /*expectsIdentifier*/ false); });
                    },
                    symbolToExpression: function (symbol, meaning, enclosingDeclaration, flags, tracker) {
                        return withContext(enclosingDeclaration, flags, tracker, function (context) { return symbolToExpression(symbol, context, meaning); });
                    },
                    symbolToTypeParameterDeclarations: function (symbol, enclosingDeclaration, flags, tracker) {
                        return withContext(enclosingDeclaration, flags, tracker, function (context) { return typeParametersToTypeParameterDeclarations(symbol, context); });
                    },
                    symbolToParameterDeclaration: function (symbol, enclosingDeclaration, flags, tracker) {
                        return withContext(enclosingDeclaration, flags, tracker, function (context) { return symbolToParameterDeclaration(symbol, context); });
                    },
                    typeParameterToDeclaration: function (parameter, enclosingDeclaration, flags, tracker) {
                        return withContext(enclosingDeclaration, flags, tracker, function (context) { return typeParameterToDeclaration(parameter, context); });
                    },
                    symbolTableToDeclarationStatements: function (symbolTable, enclosingDeclaration, flags, tracker, bundled) {
                        return withContext(enclosingDeclaration, flags, tracker, function (context) { return symbolTableToDeclarationStatements(symbolTable, context, bundled); });
                    },
                };
                function withContext(enclosingDeclaration, flags, tracker, cb) {
                    var _a, _b;
                    ts.Debug.assert(enclosingDeclaration === undefined || (enclosingDeclaration.flags & 8 /* Synthesized */) === 0);
                    var context = {
                        enclosingDeclaration: enclosingDeclaration,
                        flags: flags || 0 /* None */,
                        // If no full tracker is provided, fake up a dummy one with a basic limited-functionality moduleResolverHost
                        tracker: tracker && tracker.trackSymbol ? tracker : { trackSymbol: function () { return false; }, moduleResolverHost: flags & 134217728 /* DoNotIncludeSymbolChain */ ? {
                                getCommonSourceDirectory: !!host.getCommonSourceDirectory ? function () { return host.getCommonSourceDirectory(); } : function () { return ""; },
                                getCurrentDirectory: function () { return host.getCurrentDirectory(); },
                                getSymlinkCache: ts.maybeBind(host, host.getSymlinkCache),
                                useCaseSensitiveFileNames: ts.maybeBind(host, host.useCaseSensitiveFileNames),
                                redirectTargetsMap: host.redirectTargetsMap,
                                getProjectReferenceRedirect: function (fileName) { return host.getProjectReferenceRedirect(fileName); },
                                isSourceOfProjectReferenceRedirect: function (fileName) { return host.isSourceOfProjectReferenceRedirect(fileName); },
                                fileExists: function (fileName) { return host.fileExists(fileName); },
                                getFileIncludeReasons: function () { return host.getFileIncludeReasons(); },
                            } : undefined },
                        encounteredError: false,
                        reportedDiagnostic: false,
                        visitedTypes: undefined,
                        symbolDepth: undefined,
                        inferTypeParameters: undefined,
                        approximateLength: 0
                    };
                    context.tracker = wrapSymbolTrackerToReportForContext(context, context.tracker);
                    var resultingNode = cb(context);
                    if (context.truncating && context.flags & 1 /* NoTruncation */) {
                        (_b = (_a = context.tracker) === null || _a === void 0 ? void 0 : _a.reportTruncationError) === null || _b === void 0 ? void 0 : _b.call(_a);
                    }
                    return context.encounteredError ? undefined : resultingNode;
                }
                function wrapSymbolTrackerToReportForContext(context, tracker) {
                    var oldTrackSymbol = tracker.trackSymbol;
                    return __assign(__assign({}, tracker), { reportCyclicStructureError: wrapReportedDiagnostic(tracker.reportCyclicStructureError), reportInaccessibleThisError: wrapReportedDiagnostic(tracker.reportInaccessibleThisError), reportInaccessibleUniqueSymbolError: wrapReportedDiagnostic(tracker.reportInaccessibleUniqueSymbolError), reportLikelyUnsafeImportRequiredError: wrapReportedDiagnostic(tracker.reportLikelyUnsafeImportRequiredError), reportNonlocalAugmentation: wrapReportedDiagnostic(tracker.reportNonlocalAugmentation), reportPrivateInBaseOfClassExpression: wrapReportedDiagnostic(tracker.reportPrivateInBaseOfClassExpression), reportNonSerializableProperty: wrapReportedDiagnostic(tracker.reportNonSerializableProperty), trackSymbol: oldTrackSymbol && (function () {
                            var args = [];
                            for (var _i = 0; _i < arguments.length; _i++) {
                                args[_i] = arguments[_i];
                            }
                            var result = oldTrackSymbol.apply(void 0, args);
                            if (result) {
                                context.reportedDiagnostic = true;
                            }
                            return result;
                        }) });
                    function wrapReportedDiagnostic(method) {
                        if (!method) {
                            return method;
                        }
                        return (function () {
                            var args = [];
                            for (var _i = 0; _i < arguments.length; _i++) {
                                args[_i] = arguments[_i];
                            }
                            context.reportedDiagnostic = true;
                            return method.apply(void 0, args);
                        });
                    }
                }
                function checkTruncationLength(context) {
                    if (context.truncating)
                        return context.truncating;
                    return context.truncating = context.approximateLength > ((context.flags & 1 /* NoTruncation */) ? ts.noTruncationMaximumTruncationLength : ts.defaultMaximumTruncationLength);
                }
                function typeToTypeNodeHelper(type, context) {
                    if (cancellationToken && cancellationToken.throwIfCancellationRequested) {
                        cancellationToken.throwIfCancellationRequested();
                    }
                    var inTypeAlias = context.flags & 8388608 /* InTypeAlias */;
                    context.flags &= ~8388608 /* InTypeAlias */;
                    if (!type) {
                        if (!(context.flags & 262144 /* AllowEmptyUnionOrIntersection */)) {
                            context.encounteredError = true;
                            return undefined; // TODO: GH#18217
                        }
                        context.approximateLength += 3;
                        return ts.factory.createKeywordTypeNode(129 /* AnyKeyword */);
                    }
                    if (!(context.flags & 536870912 /* NoTypeReduction */)) {
                        type = getReducedType(type);
                    }
                    if (type.flags & 1 /* Any */) {
                        context.approximateLength += 3;
                        return ts.factory.createKeywordTypeNode(type === intrinsicMarkerType ? 137 /* IntrinsicKeyword */ : 129 /* AnyKeyword */);
                    }
                    if (type.flags & 2 /* Unknown */) {
                        return ts.factory.createKeywordTypeNode(153 /* UnknownKeyword */);
                    }
                    if (type.flags & 4 /* String */) {
                        context.approximateLength += 6;
                        return ts.factory.createKeywordTypeNode(148 /* StringKeyword */);
                    }
                    if (type.flags & 8 /* Number */) {
                        context.approximateLength += 6;
                        return ts.factory.createKeywordTypeNode(145 /* NumberKeyword */);
                    }
                    if (type.flags & 64 /* BigInt */) {
                        context.approximateLength += 6;
                        return ts.factory.createKeywordTypeNode(156 /* BigIntKeyword */);
                    }
                    if (type.flags & 16 /* Boolean */ && !type.aliasSymbol) {
                        context.approximateLength += 7;
                        return ts.factory.createKeywordTypeNode(132 /* BooleanKeyword */);
                    }
                    if (type.flags & 1024 /* EnumLiteral */ && !(type.flags & 1048576 /* Union */)) {
                        var parentSymbol = getParentOfSymbol(type.symbol);
                        var parentName = symbolToTypeNode(parentSymbol, context, 788968 /* Type */);
                        if (getDeclaredTypeOfSymbol(parentSymbol) === type) {
                            return parentName;
                        }
                        var memberName = ts.symbolName(type.symbol);
                        if (ts.isIdentifierText(memberName, 0 /* ES3 */)) {
                            return appendReferenceToType(parentName, ts.factory.createTypeReferenceNode(memberName, /*typeArguments*/ undefined));
                        }
                        if (ts.isImportTypeNode(parentName)) {
                            parentName.isTypeOf = true; // mutably update, node is freshly manufactured anyhow
                            return ts.factory.createIndexedAccessTypeNode(parentName, ts.factory.createLiteralTypeNode(ts.factory.createStringLiteral(memberName)));
                        }
                        else if (ts.isTypeReferenceNode(parentName)) {
                            return ts.factory.createIndexedAccessTypeNode(ts.factory.createTypeQueryNode(parentName.typeName), ts.factory.createLiteralTypeNode(ts.factory.createStringLiteral(memberName)));
                        }
                        else {
                            return ts.Debug.fail("Unhandled type node kind returned from `symbolToTypeNode`.");
                        }
                    }
                    if (type.flags & 1056 /* EnumLike */) {
                        return symbolToTypeNode(type.symbol, context, 788968 /* Type */);
                    }
                    if (type.flags & 128 /* StringLiteral */) {
                        context.approximateLength += (type.value.length + 2);
                        return ts.factory.createLiteralTypeNode(ts.setEmitFlags(ts.factory.createStringLiteral(type.value, !!(context.flags & 268435456 /* UseSingleQuotesForStringLiteralType */)), 16777216 /* NoAsciiEscaping */));
                    }
                    if (type.flags & 256 /* NumberLiteral */) {
                        var value = type.value;
                        context.approximateLength += ("" + value).length;
                        return ts.factory.createLiteralTypeNode(value < 0 ? ts.factory.createPrefixUnaryExpression(40 /* MinusToken */, ts.factory.createNumericLiteral(-value)) : ts.factory.createNumericLiteral(value));
                    }
                    if (type.flags & 2048 /* BigIntLiteral */) {
                        context.approximateLength += (ts.pseudoBigIntToString(type.value).length) + 1;
                        return ts.factory.createLiteralTypeNode((ts.factory.createBigIntLiteral(type.value)));
                    }
                    if (type.flags & 512 /* BooleanLiteral */) {
                        context.approximateLength += type.intrinsicName.length;
                        return ts.factory.createLiteralTypeNode(type.intrinsicName === "true" ? ts.factory.createTrue() : ts.factory.createFalse());
                    }
                    if (type.flags & 8192 /* UniqueESSymbol */) {
                        if (!(context.flags & 1048576 /* AllowUniqueESSymbolType */)) {
                            if (isValueSymbolAccessible(type.symbol, context.enclosingDeclaration)) {
                                context.approximateLength += 6;
                                return symbolToTypeNode(type.symbol, context, 111551 /* Value */);
                            }
                            if (context.tracker.reportInaccessibleUniqueSymbolError) {
                                context.tracker.reportInaccessibleUniqueSymbolError();
                            }
                        }
                        context.approximateLength += 13;
                        return ts.factory.createTypeOperatorNode(152 /* UniqueKeyword */, ts.factory.createKeywordTypeNode(149 /* SymbolKeyword */));
                    }
                    if (type.flags & 16384 /* Void */) {
                        context.approximateLength += 4;
                        return ts.factory.createKeywordTypeNode(114 /* VoidKeyword */);
                    }
                    if (type.flags & 32768 /* Undefined */) {
                        context.approximateLength += 9;
                        return ts.factory.createKeywordTypeNode(151 /* UndefinedKeyword */);
                    }
                    if (type.flags & 65536 /* Null */) {
                        context.approximateLength += 4;
                        return ts.factory.createLiteralTypeNode(ts.factory.createNull());
                    }
                    if (type.flags & 131072 /* Never */) {
                        context.approximateLength += 5;
                        return ts.factory.createKeywordTypeNode(142 /* NeverKeyword */);
                    }
                    if (type.flags & 4096 /* ESSymbol */) {
                        context.approximateLength += 6;
                        return ts.factory.createKeywordTypeNode(149 /* SymbolKeyword */);
                    }
                    if (type.flags & 67108864 /* NonPrimitive */) {
                        context.approximateLength += 6;
                        return ts.factory.createKeywordTypeNode(146 /* ObjectKeyword */);
                    }
                    if (isThisTypeParameter(type)) {
                        if (context.flags & 4194304 /* InObjectTypeLiteral */) {
                            if (!context.encounteredError && !(context.flags & 32768 /* AllowThisInObjectLiteral */)) {
                                context.encounteredError = true;
                            }
                            if (context.tracker.reportInaccessibleThisError) {
                                context.tracker.reportInaccessibleThisError();
                            }
                        }
                        context.approximateLength += 4;
                        return ts.factory.createThisTypeNode();
                    }
                    if (!inTypeAlias && type.aliasSymbol && (context.flags & 16384 /* UseAliasDefinedOutsideCurrentScope */ || isTypeSymbolAccessible(type.aliasSymbol, context.enclosingDeclaration))) {
                        var typeArgumentNodes = mapToTypeNodes(type.aliasTypeArguments, context);
                        if (isReservedMemberName(type.aliasSymbol.escapedName) && !(type.aliasSymbol.flags & 32 /* Class */))
                            return ts.factory.createTypeReferenceNode(ts.factory.createIdentifier(""), typeArgumentNodes);
                        return symbolToTypeNode(type.aliasSymbol, context, 788968 /* Type */, typeArgumentNodes);
                    }
                    var objectFlags = ts.getObjectFlags(type);
                    if (objectFlags & 4 /* Reference */) {
                        ts.Debug.assert(!!(type.flags & 524288 /* Object */));
                        return type.node ? visitAndTransformType(type, typeReferenceToTypeNode) : typeReferenceToTypeNode(type);
                    }
                    if (type.flags & 262144 /* TypeParameter */ || objectFlags & 3 /* ClassOrInterface */) {
                        if (type.flags & 262144 /* TypeParameter */ && ts.contains(context.inferTypeParameters, type)) {
                            context.approximateLength += (ts.symbolName(type.symbol).length + 6);
                            return ts.factory.createInferTypeNode(typeParameterToDeclarationWithConstraint(type, context, /*constraintNode*/ undefined));
                        }
                        if (context.flags & 4 /* GenerateNamesForShadowedTypeParams */ &&
                            type.flags & 262144 /* TypeParameter */ &&
                            !isTypeSymbolAccessible(type.symbol, context.enclosingDeclaration)) {
                            var name = typeParameterToName(type, context);
                            context.approximateLength += ts.idText(name).length;
                            return ts.factory.createTypeReferenceNode(ts.factory.createIdentifier(ts.idText(name)), /*typeArguments*/ undefined);
                        }
                        // Ignore constraint/default when creating a usage (as opposed to declaration) of a type parameter.
                        return type.symbol
                            ? symbolToTypeNode(type.symbol, context, 788968 /* Type */)
                            : ts.factory.createTypeReferenceNode(ts.factory.createIdentifier("?"), /*typeArguments*/ undefined);
                    }
                    if (type.flags & 1048576 /* Union */ && type.origin) {
                        type = type.origin;
                    }
                    if (type.flags & (1048576 /* Union */ | 2097152 /* Intersection */)) {
                        var types = type.flags & 1048576 /* Union */ ? formatUnionTypes(type.types) : type.types;
                        if (ts.length(types) === 1) {
                            return typeToTypeNodeHelper(types[0], context);
                        }
                        var typeNodes = mapToTypeNodes(types, context, /*isBareList*/ true);
                        if (typeNodes && typeNodes.length > 0) {
                            return type.flags & 1048576 /* Union */ ? ts.factory.createUnionTypeNode(typeNodes) : ts.factory.createIntersectionTypeNode(typeNodes);
                        }
                        else {
                            if (!context.encounteredError && !(context.flags & 262144 /* AllowEmptyUnionOrIntersection */)) {
                                context.encounteredError = true;
                            }
                            return undefined; // TODO: GH#18217
                        }
                    }
                    if (objectFlags & (16 /* Anonymous */ | 32 /* Mapped */)) {
                        ts.Debug.assert(!!(type.flags & 524288 /* Object */));
                        // The type is an object literal type.
                        return createAnonymousTypeNode(type);
                    }
                    if (type.flags & 4194304 /* Index */) {
                        var indexedType = type.type;
                        context.approximateLength += 6;
                        var indexTypeNode = typeToTypeNodeHelper(indexedType, context);
                        return ts.factory.createTypeOperatorNode(139 /* KeyOfKeyword */, indexTypeNode);
                    }
                    if (type.flags & 134217728 /* TemplateLiteral */) {
                        var texts_1 = type.texts;
                        var types_1 = type.types;
                        var templateHead = ts.factory.createTemplateHead(texts_1[0]);
                        var templateSpans = ts.factory.createNodeArray(ts.map(types_1, function (t, i) { return ts.factory.createTemplateLiteralTypeSpan(typeToTypeNodeHelper(t, context), (i < types_1.length - 1 ? ts.factory.createTemplateMiddle : ts.factory.createTemplateTail)(texts_1[i + 1])); }));
                        context.approximateLength += 2;
                        return ts.factory.createTemplateLiteralType(templateHead, templateSpans);
                    }
                    if (type.flags & 268435456 /* StringMapping */) {
                        var typeNode = typeToTypeNodeHelper(type.type, context);
                        return symbolToTypeNode(type.symbol, context, 788968 /* Type */, [typeNode]);
                    }
                    if (type.flags & 8388608 /* IndexedAccess */) {
                        var objectTypeNode = typeToTypeNodeHelper(type.objectType, context);
                        var indexTypeNode = typeToTypeNodeHelper(type.indexType, context);
                        context.approximateLength += 2;
                        return ts.factory.createIndexedAccessTypeNode(objectTypeNode, indexTypeNode);
                    }
                    if (type.flags & 16777216 /* Conditional */) {
                        return visitAndTransformType(type, function (type) { return conditionalTypeToTypeNode(type); });
                    }
                    if (type.flags & 33554432 /* Substitution */) {
                        return typeToTypeNodeHelper(type.baseType, context);
                    }
                    return ts.Debug.fail("Should be unreachable.");
                    function conditionalTypeToTypeNode(type) {
                        var checkTypeNode = typeToTypeNodeHelper(type.checkType, context);
                        var saveInferTypeParameters = context.inferTypeParameters;
                        context.inferTypeParameters = type.root.inferTypeParameters;
                        var extendsTypeNode = typeToTypeNodeHelper(type.extendsType, context);
                        context.inferTypeParameters = saveInferTypeParameters;
                        var trueTypeNode = typeToTypeNodeOrCircularityElision(getTrueTypeFromConditionalType(type));
                        var falseTypeNode = typeToTypeNodeOrCircularityElision(getFalseTypeFromConditionalType(type));
                        context.approximateLength += 15;
                        return ts.factory.createConditionalTypeNode(checkTypeNode, extendsTypeNode, trueTypeNode, falseTypeNode);
                    }
                    function typeToTypeNodeOrCircularityElision(type) {
                        var _a, _b, _c;
                        if (type.flags & 1048576 /* Union */) {
                            if ((_a = context.visitedTypes) === null || _a === void 0 ? void 0 : _a.has(getTypeId(type))) {
                                if (!(context.flags & 131072 /* AllowAnonymousIdentifier */)) {
                                    context.encounteredError = true;
                                    (_c = (_b = context.tracker) === null || _b === void 0 ? void 0 : _b.reportCyclicStructureError) === null || _c === void 0 ? void 0 : _c.call(_b);
                                }
                                return createElidedInformationPlaceholder(context);
                            }
                            return visitAndTransformType(type, function (type) { return typeToTypeNodeHelper(type, context); });
                        }
                        return typeToTypeNodeHelper(type, context);
                    }
                    function createMappedTypeNodeFromType(type) {
                        ts.Debug.assert(!!(type.flags & 524288 /* Object */));
                        var readonlyToken = type.declaration.readonlyToken ? ts.factory.createToken(type.declaration.readonlyToken.kind) : undefined;
                        var questionToken = type.declaration.questionToken ? ts.factory.createToken(type.declaration.questionToken.kind) : undefined;
                        var appropriateConstraintTypeNode;
                        if (isMappedTypeWithKeyofConstraintDeclaration(type)) {
                            // We have a { [P in keyof T]: X }
                            // We do this to ensure we retain the toplevel keyof-ness of the type which may be lost due to keyof distribution during `getConstraintTypeFromMappedType`
                            appropriateConstraintTypeNode = ts.factory.createTypeOperatorNode(139 /* KeyOfKeyword */, typeToTypeNodeHelper(getModifiersTypeFromMappedType(type), context));
                        }
                        else {
                            appropriateConstraintTypeNode = typeToTypeNodeHelper(getConstraintTypeFromMappedType(type), context);
                        }
                        var typeParameterNode = typeParameterToDeclarationWithConstraint(getTypeParameterFromMappedType(type), context, appropriateConstraintTypeNode);
                        var nameTypeNode = type.declaration.nameType ? typeToTypeNodeHelper(getNameTypeFromMappedType(type), context) : undefined;
                        var templateTypeNode = typeToTypeNodeHelper(removeMissingType(getTemplateTypeFromMappedType(type), !!(getMappedTypeModifiers(type) & 4 /* IncludeOptional */)), context);
                        var mappedTypeNode = ts.factory.createMappedTypeNode(readonlyToken, typeParameterNode, nameTypeNode, questionToken, templateTypeNode);
                        context.approximateLength += 10;
                        return ts.setEmitFlags(mappedTypeNode, 1 /* SingleLine */);
                    }
                    function createAnonymousTypeNode(type) {
                        var _a;
                        var typeId = type.id;
                        var symbol = type.symbol;
                        if (symbol) {
                            var isInstanceType = isClassInstanceSide(type) ? 788968 /* Type */ : 111551 /* Value */;
                            if (isJSConstructor(symbol.valueDeclaration)) {
                                // Instance and static types share the same symbol; only add 'typeof' for the static side.
                                return symbolToTypeNode(symbol, context, isInstanceType);
                            }
                            // Always use 'typeof T' for type of class, enum, and module objects
                            else if (symbol.flags & 32 /* Class */
                                && !getBaseTypeVariableOfClass(symbol)
                                && !(symbol.valueDeclaration && symbol.valueDeclaration.kind === 224 /* ClassExpression */ && context.flags & 2048 /* WriteClassExpressionAsTypeLiteral */) ||
                                symbol.flags & (384 /* Enum */ | 512 /* ValueModule */) ||
                                shouldWriteTypeOfFunctionSymbol()) {
                                return symbolToTypeNode(symbol, context, isInstanceType);
                            }
                            else if ((_a = context.visitedTypes) === null || _a === void 0 ? void 0 : _a.has(typeId)) {
                                // If type is an anonymous type literal in a type alias declaration, use type alias name
                                var typeAlias = getTypeAliasForTypeLiteral(type);
                                if (typeAlias) {
                                    // The specified symbol flags need to be reinterpreted as type flags
                                    return symbolToTypeNode(typeAlias, context, 788968 /* Type */);
                                }
                                else {
                                    return createElidedInformationPlaceholder(context);
                                }
                            }
                            else {
                                return visitAndTransformType(type, createTypeNodeFromObjectType);
                            }
                        }
                        else {
                            // Anonymous types without a symbol are never circular.
                            return createTypeNodeFromObjectType(type);
                        }
                        function shouldWriteTypeOfFunctionSymbol() {
                            var _a;
                            var isStaticMethodSymbol = !!(symbol.flags & 8192 /* Method */) && // typeof static method
                                ts.some(symbol.declarations, function (declaration) { return ts.isStatic(declaration); });
                            var isNonLocalFunctionSymbol = !!(symbol.flags & 16 /* Function */) &&
                                (symbol.parent || // is exported function symbol
                                    ts.forEach(symbol.declarations, function (declaration) {
                                        return declaration.parent.kind === 300 /* SourceFile */ || declaration.parent.kind === 260 /* ModuleBlock */;
                                    }));
                            if (isStaticMethodSymbol || isNonLocalFunctionSymbol) {
                                // typeof is allowed only for static/non local functions
                                return (!!(context.flags & 4096 /* UseTypeOfFunction */) || ((_a = context.visitedTypes) === null || _a === void 0 ? void 0 : _a.has(typeId))) && // it is type of the symbol uses itself recursively
                                    (!(context.flags & 8 /* UseStructuralFallback */) || isValueSymbolAccessible(symbol, context.enclosingDeclaration)); // And the build is going to succeed without visibility error or there is no structural fallback allowed
                            }
                        }
                    }
                    function visitAndTransformType(type, transform) {
                        var _a, _b;
                        var typeId = type.id;
                        var isConstructorObject = ts.getObjectFlags(type) & 16 /* Anonymous */ && type.symbol && type.symbol.flags & 32 /* Class */;
                        var id = ts.getObjectFlags(type) & 4 /* Reference */ && type.node ? "N" + getNodeId(type.node) :
                            type.flags & 16777216 /* Conditional */ ? "N" + getNodeId(type.root.node) :
                                type.symbol ? (isConstructorObject ? "+" : "") + getSymbolId(type.symbol) :
                                    undefined;
                        // Since instantiations of the same anonymous type have the same symbol, tracking symbols instead
                        // of types allows us to catch circular references to instantiations of the same anonymous type
                        if (!context.visitedTypes) {
                            context.visitedTypes = new ts.Set();
                        }
                        if (id && !context.symbolDepth) {
                            context.symbolDepth = new ts.Map();
                        }
                        var links = context.enclosingDeclaration && getNodeLinks(context.enclosingDeclaration);
                        var key = getTypeId(type) + "|" + context.flags;
                        if (links) {
                            links.serializedTypes || (links.serializedTypes = new ts.Map());
                        }
                        var cachedResult = (_a = links === null || links === void 0 ? void 0 : links.serializedTypes) === null || _a === void 0 ? void 0 : _a.get(key);
                        if (cachedResult) {
                            if (cachedResult.truncating) {
                                context.truncating = true;
                            }
                            context.approximateLength += cachedResult.addedLength;
                            return deepCloneOrReuseNode(cachedResult);
                        }
                        var depth;
                        if (id) {
                            depth = context.symbolDepth.get(id) || 0;
                            if (depth > 10) {
                                return createElidedInformationPlaceholder(context);
                            }
                            context.symbolDepth.set(id, depth + 1);
                        }
                        context.visitedTypes.add(typeId);
                        var startLength = context.approximateLength;
                        var result = transform(type);
                        var addedLength = context.approximateLength - startLength;
                        if (!context.reportedDiagnostic && !context.encounteredError) {
                            if (context.truncating) {
                                result.truncating = true;
                            }
                            result.addedLength = addedLength;
                            (_b = links === null || links === void 0 ? void 0 : links.serializedTypes) === null || _b === void 0 ? void 0 : _b.set(key, result);
                        }
                        context.visitedTypes.delete(typeId);
                        if (id) {
                            context.symbolDepth.set(id, depth);
                        }
                        return result;
                        function deepCloneOrReuseNode(node) {
                            if (!ts.nodeIsSynthesized(node) && ts.getParseTreeNode(node) === node) {
                                return node;
                            }
                            return ts.setTextRange(ts.factory.cloneNode(ts.visitEachChild(node, deepCloneOrReuseNode, ts.nullTransformationContext)), node);
                        }
                    }
                    function createTypeNodeFromObjectType(type) {
                        if (isGenericMappedType(type) || type.containsError) {
                            return createMappedTypeNodeFromType(type);
                        }
                        var resolved = resolveStructuredTypeMembers(type);
                        if (!resolved.properties.length && !resolved.indexInfos.length) {
                            if (!resolved.callSignatures.length && !resolved.constructSignatures.length) {
                                context.approximateLength += 2;
                                return ts.setEmitFlags(ts.factory.createTypeLiteralNode(/*members*/ undefined), 1 /* SingleLine */);
                            }
                            if (resolved.callSignatures.length === 1 && !resolved.constructSignatures.length) {
                                var signature = resolved.callSignatures[0];
                                var signatureNode = signatureToSignatureDeclarationHelper(signature, 177 /* FunctionType */, context);
                                return signatureNode;
                            }
                            if (resolved.constructSignatures.length === 1 && !resolved.callSignatures.length) {
                                var signature = resolved.constructSignatures[0];
                                var signatureNode = signatureToSignatureDeclarationHelper(signature, 178 /* ConstructorType */, context);
                                return signatureNode;
                            }
                        }
                        var abstractSignatures = ts.filter(resolved.constructSignatures, function (signature) { return !!(signature.flags & 4 /* Abstract */); });
                        if (ts.some(abstractSignatures)) {
                            var types = ts.map(abstractSignatures, getOrCreateTypeFromSignature);
                            // count the number of type elements excluding abstract constructors
                            var typeElementCount = resolved.callSignatures.length +
                                (resolved.constructSignatures.length - abstractSignatures.length) +
                                resolved.indexInfos.length +
                                // exclude `prototype` when writing a class expression as a type literal, as per
                                // the logic in `createTypeNodesFromResolvedType`.
                                (context.flags & 2048 /* WriteClassExpressionAsTypeLiteral */ ?
                                    ts.countWhere(resolved.properties, function (p) { return !(p.flags & 4194304 /* Prototype */); }) :
                                    ts.length(resolved.properties));
                            // don't include an empty object literal if there were no other static-side
                            // properties to write, i.e. `abstract class C { }` becomes `abstract new () => {}`
                            // and not `(abstract new () => {}) & {}`
                            if (typeElementCount) {
                                // create a copy of the object type without any abstract construct signatures.
                                types.push(getResolvedTypeWithoutAbstractConstructSignatures(resolved));
                            }
                            return typeToTypeNodeHelper(getIntersectionType(types), context);
                        }
                        var savedFlags = context.flags;
                        context.flags |= 4194304 /* InObjectTypeLiteral */;
                        var members = createTypeNodesFromResolvedType(resolved);
                        context.flags = savedFlags;
                        var typeLiteralNode = ts.factory.createTypeLiteralNode(members);
                        context.approximateLength += 2;
                        ts.setEmitFlags(typeLiteralNode, (context.flags & 1024 /* MultilineObjectLiterals */) ? 0 : 1 /* SingleLine */);
                        return typeLiteralNode;
                    }
                    function typeReferenceToTypeNode(type) {
                        var typeArguments = getTypeArguments(type);
                        if (type.target === globalArrayType || type.target === globalReadonlyArrayType) {
                            if (context.flags & 2 /* WriteArrayAsGenericType */) {
                                var typeArgumentNode = typeToTypeNodeHelper(typeArguments[0], context);
                                return ts.factory.createTypeReferenceNode(type.target === globalArrayType ? "Array" : "ReadonlyArray", [typeArgumentNode]);
                            }
                            var elementType = typeToTypeNodeHelper(typeArguments[0], context);
                            var arrayType = ts.factory.createArrayTypeNode(elementType);
                            return type.target === globalArrayType ? arrayType : ts.factory.createTypeOperatorNode(143 /* ReadonlyKeyword */, arrayType);
                        }
                        else if (type.target.objectFlags & 8 /* Tuple */) {
                            typeArguments = ts.sameMap(typeArguments, function (t, i) { return removeMissingType(t, !!(type.target.elementFlags[i] & 2 /* Optional */)); });
                            if (typeArguments.length > 0) {
                                var arity = getTypeReferenceArity(type);
                                var tupleConstituentNodes = mapToTypeNodes(typeArguments.slice(0, arity), context);
                                if (tupleConstituentNodes) {
                                    if (type.target.labeledElementDeclarations) {
                                        for (var i = 0; i < tupleConstituentNodes.length; i++) {
                                            var flags = type.target.elementFlags[i];
                                            tupleConstituentNodes[i] = ts.factory.createNamedTupleMember(flags & 12 /* Variable */ ? ts.factory.createToken(25 /* DotDotDotToken */) : undefined, ts.factory.createIdentifier(ts.unescapeLeadingUnderscores(getTupleElementLabel(type.target.labeledElementDeclarations[i]))), flags & 2 /* Optional */ ? ts.factory.createToken(57 /* QuestionToken */) : undefined, flags & 4 /* Rest */ ? ts.factory.createArrayTypeNode(tupleConstituentNodes[i]) :
                                                tupleConstituentNodes[i]);
                                        }
                                    }
                                    else {
                                        for (var i = 0; i < Math.min(arity, tupleConstituentNodes.length); i++) {
                                            var flags = type.target.elementFlags[i];
                                            tupleConstituentNodes[i] =
                                                flags & 12 /* Variable */ ? ts.factory.createRestTypeNode(flags & 4 /* Rest */ ? ts.factory.createArrayTypeNode(tupleConstituentNodes[i]) : tupleConstituentNodes[i]) :
                                                    flags & 2 /* Optional */ ? ts.factory.createOptionalTypeNode(tupleConstituentNodes[i]) :
                                                        tupleConstituentNodes[i];
                                        }
                                    }
                                    var tupleTypeNode = ts.setEmitFlags(ts.factory.createTupleTypeNode(tupleConstituentNodes), 1 /* SingleLine */);
                                    return type.target.readonly ? ts.factory.createTypeOperatorNode(143 /* ReadonlyKeyword */, tupleTypeNode) : tupleTypeNode;
                                }
                            }
                            if (context.encounteredError || (context.flags & 524288 /* AllowEmptyTuple */)) {
                                var tupleTypeNode = ts.setEmitFlags(ts.factory.createTupleTypeNode([]), 1 /* SingleLine */);
                                return type.target.readonly ? ts.factory.createTypeOperatorNode(143 /* ReadonlyKeyword */, tupleTypeNode) : tupleTypeNode;
                            }
                            context.encounteredError = true;
                            return undefined; // TODO: GH#18217
                        }
                        else if (context.flags & 2048 /* WriteClassExpressionAsTypeLiteral */ &&
                            type.symbol.valueDeclaration &&
                            ts.isClassLike(type.symbol.valueDeclaration) &&
                            !isValueSymbolAccessible(type.symbol, context.enclosingDeclaration)) {
                            return createAnonymousTypeNode(type);
                        }
                        else {
                            var outerTypeParameters = type.target.outerTypeParameters;
                            var i = 0;
                            var resultType = void 0;
                            if (outerTypeParameters) {
                                var length_2 = outerTypeParameters.length;
                                while (i < length_2) {
                                    // Find group of type arguments for type parameters with the same declaring container.
                                    var start = i;
                                    var parent = getParentSymbolOfTypeParameter(outerTypeParameters[i]);
                                    do {
                                        i++;
                                    } while (i < length_2 && getParentSymbolOfTypeParameter(outerTypeParameters[i]) === parent);
                                    // When type parameters are their own type arguments for the whole group (i.e. we have
                                    // the default outer type arguments), we don't show the group.
                                    if (!ts.rangeEquals(outerTypeParameters, typeArguments, start, i)) {
                                        var typeArgumentSlice = mapToTypeNodes(typeArguments.slice(start, i), context);
                                        var flags_3 = context.flags;
                                        context.flags |= 16 /* ForbidIndexedAccessSymbolReferences */;
                                        var ref = symbolToTypeNode(parent, context, 788968 /* Type */, typeArgumentSlice);
                                        context.flags = flags_3;
                                        resultType = !resultType ? ref : appendReferenceToType(resultType, ref);
                                    }
                                }
                            }
                            var typeArgumentNodes = void 0;
                            if (typeArguments.length > 0) {
                                var typeParameterCount = (type.target.typeParameters || ts.emptyArray).length;
                                typeArgumentNodes = mapToTypeNodes(typeArguments.slice(i, typeParameterCount), context);
                            }
                            var flags = context.flags;
                            context.flags |= 16 /* ForbidIndexedAccessSymbolReferences */;
                            var finalRef = symbolToTypeNode(type.symbol, context, 788968 /* Type */, typeArgumentNodes);
                            context.flags = flags;
                            return !resultType ? finalRef : appendReferenceToType(resultType, finalRef);
                        }
                    }
                    function appendReferenceToType(root, ref) {
                        if (ts.isImportTypeNode(root)) {
                            // first shift type arguments
                            var typeArguments = root.typeArguments;
                            var qualifier = root.qualifier;
                            if (qualifier) {
                                if (ts.isIdentifier(qualifier)) {
                                    qualifier = ts.factory.updateIdentifier(qualifier, typeArguments);
                                }
                                else {
                                    qualifier = ts.factory.updateQualifiedName(qualifier, qualifier.left, ts.factory.updateIdentifier(qualifier.right, typeArguments));
                                }
                            }
                            typeArguments = ref.typeArguments;
                            // then move qualifiers
                            var ids = getAccessStack(ref);
                            for (var _i = 0, ids_1 = ids; _i < ids_1.length; _i++) {
                                var id = ids_1[_i];
                                qualifier = qualifier ? ts.factory.createQualifiedName(qualifier, id) : id;
                            }
                            return ts.factory.updateImportTypeNode(root, root.argument, qualifier, typeArguments, root.isTypeOf);
                        }
                        else {
                            // first shift type arguments
                            var typeArguments = root.typeArguments;
                            var typeName = root.typeName;
                            if (ts.isIdentifier(typeName)) {
                                typeName = ts.factory.updateIdentifier(typeName, typeArguments);
                            }
                            else {
                                typeName = ts.factory.updateQualifiedName(typeName, typeName.left, ts.factory.updateIdentifier(typeName.right, typeArguments));
                            }
                            typeArguments = ref.typeArguments;
                            // then move qualifiers
                            var ids = getAccessStack(ref);
                            for (var _a = 0, ids_2 = ids; _a < ids_2.length; _a++) {
                                var id = ids_2[_a];
                                typeName = ts.factory.createQualifiedName(typeName, id);
                            }
                            return ts.factory.updateTypeReferenceNode(root, typeName, typeArguments);
                        }
                    }
                    function getAccessStack(ref) {
                        var state = ref.typeName;
                        var ids = [];
                        while (!ts.isIdentifier(state)) {
                            ids.unshift(state.right);
                            state = state.left;
                        }
                        ids.unshift(state);
                        return ids;
                    }
                    function createTypeNodesFromResolvedType(resolvedType) {
                        if (checkTruncationLength(context)) {
                            return [ts.factory.createPropertySignature(/*modifiers*/ undefined, "...", /*questionToken*/ undefined, /*type*/ undefined)];
                        }
                        var typeElements = [];
                        for (var _i = 0, _a = resolvedType.callSignatures; _i < _a.length; _i++) {
                            var signature = _a[_i];
                            typeElements.push(signatureToSignatureDeclarationHelper(signature, 172 /* CallSignature */, context));
                        }
                        for (var _b = 0, _c = resolvedType.constructSignatures; _b < _c.length; _b++) {
                            var signature = _c[_b];
                            if (signature.flags & 4 /* Abstract */)
                                continue;
                            typeElements.push(signatureToSignatureDeclarationHelper(signature, 173 /* ConstructSignature */, context));
                        }
                        for (var _d = 0, _e = resolvedType.indexInfos; _d < _e.length; _d++) {
                            var info = _e[_d];
                            typeElements.push(indexInfoToIndexSignatureDeclarationHelper(info, context, resolvedType.objectFlags & 1024 /* ReverseMapped */ ? createElidedInformationPlaceholder(context) : undefined));
                        }
                        var properties = resolvedType.properties;
                        if (!properties) {
                            return typeElements;
                        }
                        var i = 0;
                        for (var _f = 0, properties_1 = properties; _f < properties_1.length; _f++) {
                            var propertySymbol = properties_1[_f];
                            i++;
                            if (context.flags & 2048 /* WriteClassExpressionAsTypeLiteral */) {
                                if (propertySymbol.flags & 4194304 /* Prototype */) {
                                    continue;
                                }
                                if (ts.getDeclarationModifierFlagsFromSymbol(propertySymbol) & (8 /* Private */ | 16 /* Protected */) && context.tracker.reportPrivateInBaseOfClassExpression) {
                                    context.tracker.reportPrivateInBaseOfClassExpression(ts.unescapeLeadingUnderscores(propertySymbol.escapedName));
                                }
                            }
                            if (checkTruncationLength(context) && (i + 2 < properties.length - 1)) {
                                typeElements.push(ts.factory.createPropertySignature(/*modifiers*/ undefined, "... " + (properties.length - i) + " more ...", /*questionToken*/ undefined, /*type*/ undefined));
                                addPropertyToElementList(properties[properties.length - 1], context, typeElements);
                                break;
                            }
                            addPropertyToElementList(propertySymbol, context, typeElements);
                        }
                        return typeElements.length ? typeElements : undefined;
                    }
                }
                function createElidedInformationPlaceholder(context) {
                    context.approximateLength += 3;
                    if (!(context.flags & 1 /* NoTruncation */)) {
                        return ts.factory.createTypeReferenceNode(ts.factory.createIdentifier("..."), /*typeArguments*/ undefined);
                    }
                    return ts.factory.createKeywordTypeNode(129 /* AnyKeyword */);
                }
                function shouldUsePlaceholderForProperty(propertySymbol, context) {
                    var _a;
                    // Use placeholders for reverse mapped types we've either already descended into, or which
                    // are nested reverse mappings within a mapping over a non-anonymous type. The later is a restriction mostly just to
                    // reduce the blowup in printback size from doing, eg, a deep reverse mapping over `Window`.
                    // Since anonymous types usually come from expressions, this allows us to preserve the output
                    // for deep mappings which likely come from expressions, while truncating those parts which
                    // come from mappings over library functions.
                    return !!(ts.getCheckFlags(propertySymbol) & 8192 /* ReverseMapped */)
                        && (ts.contains(context.reverseMappedStack, propertySymbol)
                            || (((_a = context.reverseMappedStack) === null || _a === void 0 ? void 0 : _a[0])
                                && !(ts.getObjectFlags(ts.last(context.reverseMappedStack).propertyType) & 16 /* Anonymous */)));
                }
                function addPropertyToElementList(propertySymbol, context, typeElements) {
                    var _a, _b;
                    var propertyIsReverseMapped = !!(ts.getCheckFlags(propertySymbol) & 8192 /* ReverseMapped */);
                    var propertyType = shouldUsePlaceholderForProperty(propertySymbol, context) ?
                        anyType : getNonMissingTypeOfSymbol(propertySymbol);
                    var saveEnclosingDeclaration = context.enclosingDeclaration;
                    context.enclosingDeclaration = undefined;
                    if (context.tracker.trackSymbol && ts.getCheckFlags(propertySymbol) & 4096 /* Late */ && isLateBoundName(propertySymbol.escapedName)) {
                        if (propertySymbol.declarations) {
                            var decl = ts.first(propertySymbol.declarations);
                            if (hasLateBindableName(decl)) {
                                if (ts.isBinaryExpression(decl)) {
                                    var name = ts.getNameOfDeclaration(decl);
                                    if (name && ts.isElementAccessExpression(name) && ts.isPropertyAccessEntityNameExpression(name.argumentExpression)) {
                                        trackComputedName(name.argumentExpression, saveEnclosingDeclaration, context);
                                    }
                                }
                                else {
                                    trackComputedName(decl.name.expression, saveEnclosingDeclaration, context);
                                }
                            }
                        }
                        else if ((_a = context.tracker) === null || _a === void 0 ? void 0 : _a.reportNonSerializableProperty) {
                            context.tracker.reportNonSerializableProperty(symbolToString(propertySymbol));
                        }
                    }
                    context.enclosingDeclaration = propertySymbol.valueDeclaration || ((_b = propertySymbol.declarations) === null || _b === void 0 ? void 0 : _b[0]) || saveEnclosingDeclaration;
                    var propertyName = getPropertyNameNodeForSymbol(propertySymbol, context);
                    context.enclosingDeclaration = saveEnclosingDeclaration;
                    context.approximateLength += (ts.symbolName(propertySymbol).length + 1);
                    var optionalToken = propertySymbol.flags & 16777216 /* Optional */ ? ts.factory.createToken(57 /* QuestionToken */) : undefined;
                    if (propertySymbol.flags & (16 /* Function */ | 8192 /* Method */) && !getPropertiesOfObjectType(propertyType).length && !isReadonlySymbol(propertySymbol)) {
                        var signatures = getSignaturesOfType(filterType(propertyType, function (t) { return !(t.flags & 32768 /* Undefined */); }), 0 /* Call */);
                        for (var _i = 0, signatures_1 = signatures; _i < signatures_1.length; _i++) {
                            var signature = signatures_1[_i];
                            var methodDeclaration = signatureToSignatureDeclarationHelper(signature, 166 /* MethodSignature */, context, { name: propertyName, questionToken: optionalToken });
                            typeElements.push(preserveCommentsOn(methodDeclaration));
                        }
                    }
                    else {
                        var propertyTypeNode = void 0;
                        if (shouldUsePlaceholderForProperty(propertySymbol, context)) {
                            propertyTypeNode = createElidedInformationPlaceholder(context);
                        }
                        else {
                            if (propertyIsReverseMapped) {
                                context.reverseMappedStack || (context.reverseMappedStack = []);
                                context.reverseMappedStack.push(propertySymbol);
                            }
                            propertyTypeNode = propertyType ? serializeTypeForDeclaration(context, propertyType, propertySymbol, saveEnclosingDeclaration) : ts.factory.createKeywordTypeNode(129 /* AnyKeyword */);
                            if (propertyIsReverseMapped) {
                                context.reverseMappedStack.pop();
                            }
                        }
                        var modifiers = isReadonlySymbol(propertySymbol) ? [ts.factory.createToken(143 /* ReadonlyKeyword */)] : undefined;
                        if (modifiers) {
                            context.approximateLength += 9;
                        }
                        var propertySignature = ts.factory.createPropertySignature(modifiers, propertyName, optionalToken, propertyTypeNode);
                        typeElements.push(preserveCommentsOn(propertySignature));
                    }
                    function preserveCommentsOn(node) {
                        var _a;
                        if (ts.some(propertySymbol.declarations, function (d) { return d.kind === 342 /* JSDocPropertyTag */; })) {
                            var d = (_a = propertySymbol.declarations) === null || _a === void 0 ? void 0 : _a.find(function (d) { return d.kind === 342 /* JSDocPropertyTag */; });
                            var commentText = ts.getTextOfJSDocComment(d.comment);
                            if (commentText) {
                                ts.setSyntheticLeadingComments(node, [{ kind: 3 /* MultiLineCommentTrivia */, text: "*\n * " + commentText.replace(/\n/g, "\n * ") + "\n ", pos: -1, end: -1, hasTrailingNewLine: true }]);
                            }
                        }
                        else if (propertySymbol.valueDeclaration) {
                            // Copy comments to node for declaration emit
                            ts.setCommentRange(node, propertySymbol.valueDeclaration);
                        }
                        return node;
                    }
                }
                function mapToTypeNodes(types, context, isBareList) {
                    if (ts.some(types)) {
                        if (checkTruncationLength(context)) {
                            if (!isBareList) {
                                return [ts.factory.createTypeReferenceNode("...", /*typeArguments*/ undefined)];
                            }
                            else if (types.length > 2) {
                                return [
                                    typeToTypeNodeHelper(types[0], context),
                                    ts.factory.createTypeReferenceNode("... " + (types.length - 2) + " more ...", /*typeArguments*/ undefined),
                                    typeToTypeNodeHelper(types[types.length - 1], context)
                                ];
                            }
                        }
                        var mayHaveNameCollisions = !(context.flags & 64 /* UseFullyQualifiedType */);
                        /** Map from type reference identifier text to [type, index in `result` where the type node is] */
                        var seenNames = mayHaveNameCollisions ? ts.createUnderscoreEscapedMultiMap() : undefined;
                        var result_5 = [];
                        var i = 0;
                        for (var _i = 0, types_2 = types; _i < types_2.length; _i++) {
                            var type = types_2[_i];
                            i++;
                            if (checkTruncationLength(context) && (i + 2 < types.length - 1)) {
                                result_5.push(ts.factory.createTypeReferenceNode("... " + (types.length - i) + " more ...", /*typeArguments*/ undefined));
                                var typeNode_1 = typeToTypeNodeHelper(types[types.length - 1], context);
                                if (typeNode_1) {
                                    result_5.push(typeNode_1);
                                }
                                break;
                            }
                            context.approximateLength += 2; // Account for whitespace + separator
                            var typeNode = typeToTypeNodeHelper(type, context);
                            if (typeNode) {
                                result_5.push(typeNode);
                                if (seenNames && ts.isIdentifierTypeReference(typeNode)) {
                                    seenNames.add(typeNode.typeName.escapedText, [type, result_5.length - 1]);
                                }
                            }
                        }
                        if (seenNames) {
                            // To avoid printing types like `[Foo, Foo]` or `Bar & Bar` where
                            // occurrences of the same name actually come from different
                            // namespaces, go through the single-identifier type reference nodes
                            // we just generated, and see if any names were generated more than
                            // once while referring to different types. If so, regenerate the
                            // type node for each entry by that name with the
                            // `UseFullyQualifiedType` flag enabled.
                            var saveContextFlags = context.flags;
                            context.flags |= 64 /* UseFullyQualifiedType */;
                            seenNames.forEach(function (types) {
                                if (!ts.arrayIsHomogeneous(types, function (_a, _b) {
                                    var a = _a[0];
                                    var b = _b[0];
                                    return typesAreSameReference(a, b);
                                })) {
                                    for (var _i = 0, types_3 = types; _i < types_3.length; _i++) {
                                        var _a = types_3[_i], type = _a[0], resultIndex = _a[1];
                                        result_5[resultIndex] = typeToTypeNodeHelper(type, context);
                                    }
                                }
                            });
                            context.flags = saveContextFlags;
                        }
                        return result_5;
                    }
                }
                function typesAreSameReference(a, b) {
                    return a === b
                        || !!a.symbol && a.symbol === b.symbol
                        || !!a.aliasSymbol && a.aliasSymbol === b.aliasSymbol;
                }
                function indexInfoToIndexSignatureDeclarationHelper(indexInfo, context, typeNode) {
                    var name = ts.getNameFromIndexInfo(indexInfo) || "x";
                    var indexerTypeNode = typeToTypeNodeHelper(indexInfo.keyType, context);
                    var indexingParameter = ts.factory.createParameterDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, 
                    /*dotDotDotToken*/ undefined, name, 
                    /*questionToken*/ undefined, indexerTypeNode, 
                    /*initializer*/ undefined);
                    if (!typeNode) {
                        typeNode = typeToTypeNodeHelper(indexInfo.type || anyType, context);
                    }
                    if (!indexInfo.type && !(context.flags & 2097152 /* AllowEmptyIndexInfoType */)) {
                        context.encounteredError = true;
                    }
                    context.approximateLength += (name.length + 4);
                    return ts.factory.createIndexSignature(
                    /*decorators*/ undefined, indexInfo.isReadonly ? [ts.factory.createToken(143 /* ReadonlyKeyword */)] : undefined, [indexingParameter], typeNode);
                }
                function signatureToSignatureDeclarationHelper(signature, kind, context, options) {
                    var _a, _b, _c, _d;
                    var suppressAny = context.flags & 256 /* SuppressAnyReturnType */;
                    if (suppressAny)
                        context.flags &= ~256 /* SuppressAnyReturnType */; // suppress only toplevel `any`s
                    context.approximateLength += 3; // Usually a signature contributes a few more characters than this, but 3 is the minimum
                    var typeParameters;
                    var typeArguments;
                    if (context.flags & 32 /* WriteTypeArgumentsOfSignature */ && signature.target && signature.mapper && signature.target.typeParameters) {
                        typeArguments = signature.target.typeParameters.map(function (parameter) { return typeToTypeNodeHelper(instantiateType(parameter, signature.mapper), context); });
                    }
                    else {
                        typeParameters = signature.typeParameters && signature.typeParameters.map(function (parameter) { return typeParameterToDeclaration(parameter, context); });
                    }
                    var expandedParams = getExpandedParameters(signature, /*skipUnionExpanding*/ true)[0];
                    // If the expanded parameter list had a variadic in a non-trailing position, don't expand it
                    var parameters = (ts.some(expandedParams, function (p) { return p !== expandedParams[expandedParams.length - 1] && !!(ts.getCheckFlags(p) & 32768 /* RestParameter */); }) ? signature.parameters : expandedParams).map(function (parameter) { return symbolToParameterDeclaration(parameter, context, kind === 169 /* Constructor */, options === null || options === void 0 ? void 0 : options.privateSymbolVisitor, options === null || options === void 0 ? void 0 : options.bundledImports); });
                    if (signature.thisParameter) {
                        var thisParameter = symbolToParameterDeclaration(signature.thisParameter, context);
                        parameters.unshift(thisParameter);
                    }
                    var returnTypeNode;
                    var typePredicate = getTypePredicateOfSignature(signature);
                    if (typePredicate) {
                        var assertsModifier = typePredicate.kind === 2 /* AssertsThis */ || typePredicate.kind === 3 /* AssertsIdentifier */ ?
                            ts.factory.createToken(128 /* AssertsKeyword */) :
                            undefined;
                        var parameterName = typePredicate.kind === 1 /* Identifier */ || typePredicate.kind === 3 /* AssertsIdentifier */ ?
                            ts.setEmitFlags(ts.factory.createIdentifier(typePredicate.parameterName), 16777216 /* NoAsciiEscaping */) :
                            ts.factory.createThisTypeNode();
                        var typeNode = typePredicate.type && typeToTypeNodeHelper(typePredicate.type, context);
                        returnTypeNode = ts.factory.createTypePredicateNode(assertsModifier, parameterName, typeNode);
                    }
                    else {
                        var returnType = getReturnTypeOfSignature(signature);
                        if (returnType && !(suppressAny && isTypeAny(returnType))) {
                            returnTypeNode = serializeReturnTypeForSignature(context, returnType, signature, options === null || options === void 0 ? void 0 : options.privateSymbolVisitor, options === null || options === void 0 ? void 0 : options.bundledImports);
                        }
                        else if (!suppressAny) {
                            returnTypeNode = ts.factory.createKeywordTypeNode(129 /* AnyKeyword */);
                        }
                    }
                    var modifiers = options === null || options === void 0 ? void 0 : options.modifiers;
                    if ((kind === 178 /* ConstructorType */) && signature.flags & 4 /* Abstract */) {
                        var flags = ts.modifiersToFlags(modifiers);
                        modifiers = ts.factory.createModifiersFromModifierFlags(flags | 128 /* Abstract */);
                    }
                    var node = kind === 172 /* CallSignature */ ? ts.factory.createCallSignature(typeParameters, parameters, returnTypeNode) :
                        kind === 173 /* ConstructSignature */ ? ts.factory.createConstructSignature(typeParameters, parameters, returnTypeNode) :
                            kind === 166 /* MethodSignature */ ? ts.factory.createMethodSignature(modifiers, (_a = options === null || options === void 0 ? void 0 : options.name) !== null && _a !== void 0 ? _a : ts.factory.createIdentifier(""), options === null || options === void 0 ? void 0 : options.questionToken, typeParameters, parameters, returnTypeNode) :
                                kind === 167 /* MethodDeclaration */ ? ts.factory.createMethodDeclaration(/*decorators*/ undefined, modifiers, /*asteriskToken*/ undefined, (_b = options === null || options === void 0 ? void 0 : options.name) !== null && _b !== void 0 ? _b : ts.factory.createIdentifier(""), /*questionToken*/ undefined, typeParameters, parameters, returnTypeNode, /*body*/ undefined) :
                                    kind === 169 /* Constructor */ ? ts.factory.createConstructorDeclaration(/*decorators*/ undefined, modifiers, parameters, /*body*/ undefined) :
                                        kind === 170 /* GetAccessor */ ? ts.factory.createGetAccessorDeclaration(/*decorators*/ undefined, modifiers, (_c = options === null || options === void 0 ? void 0 : options.name) !== null && _c !== void 0 ? _c : ts.factory.createIdentifier(""), parameters, returnTypeNode, /*body*/ undefined) :
                                            kind === 171 /* SetAccessor */ ? ts.factory.createSetAccessorDeclaration(/*decorators*/ undefined, modifiers, (_d = options === null || options === void 0 ? void 0 : options.name) !== null && _d !== void 0 ? _d : ts.factory.createIdentifier(""), parameters, /*body*/ undefined) :
                                                kind === 174 /* IndexSignature */ ? ts.factory.createIndexSignature(/*decorators*/ undefined, modifiers, parameters, returnTypeNode) :
                                                    kind === 312 /* JSDocFunctionType */ ? ts.factory.createJSDocFunctionType(parameters, returnTypeNode) :
                                                        kind === 177 /* FunctionType */ ? ts.factory.createFunctionTypeNode(typeParameters, parameters, returnTypeNode !== null && returnTypeNode !== void 0 ? returnTypeNode : ts.factory.createTypeReferenceNode(ts.factory.createIdentifier(""))) :
                                                            kind === 178 /* ConstructorType */ ? ts.factory.createConstructorTypeNode(modifiers, typeParameters, parameters, returnTypeNode !== null && returnTypeNode !== void 0 ? returnTypeNode : ts.factory.createTypeReferenceNode(ts.factory.createIdentifier(""))) :
                                                                kind === 254 /* FunctionDeclaration */ ? ts.factory.createFunctionDeclaration(/*decorators*/ undefined, modifiers, /*asteriskToken*/ undefined, (options === null || options === void 0 ? void 0 : options.name) ? ts.cast(options.name, ts.isIdentifier) : ts.factory.createIdentifier(""), typeParameters, parameters, returnTypeNode, /*body*/ undefined) :
                                                                    kind === 211 /* FunctionExpression */ ? ts.factory.createFunctionExpression(modifiers, /*asteriskToken*/ undefined, (options === null || options === void 0 ? void 0 : options.name) ? ts.cast(options.name, ts.isIdentifier) : ts.factory.createIdentifier(""), typeParameters, parameters, returnTypeNode, ts.factory.createBlock([])) :
                                                                        kind === 212 /* ArrowFunction */ ? ts.factory.createArrowFunction(modifiers, typeParameters, parameters, returnTypeNode, /*equalsGreaterThanToken*/ undefined, ts.factory.createBlock([])) :
                                                                            ts.Debug.assertNever(kind);
                    if (typeArguments) {
                        node.typeArguments = ts.factory.createNodeArray(typeArguments);
                    }
                    return node;
                }
                function typeParameterToDeclarationWithConstraint(type, context, constraintNode) {
                    var savedContextFlags = context.flags;
                    context.flags &= ~512 /* WriteTypeParametersInQualifiedName */; // Avoids potential infinite loop when building for a claimspace with a generic
                    var name = typeParameterToName(type, context);
                    var defaultParameter = getDefaultFromTypeParameter(type);
                    var defaultParameterNode = defaultParameter && typeToTypeNodeHelper(defaultParameter, context);
                    context.flags = savedContextFlags;
                    return ts.factory.createTypeParameterDeclaration(name, constraintNode, defaultParameterNode);
                }
                function typeParameterToDeclaration(type, context, constraint) {
                    if (constraint === void 0) {
                        constraint = getConstraintOfTypeParameter(type);
                    }
                    var constraintNode = constraint && typeToTypeNodeHelper(constraint, context);
                    return typeParameterToDeclarationWithConstraint(type, context, constraintNode);
                }
                function symbolToParameterDeclaration(parameterSymbol, context, preserveModifierFlags, privateSymbolVisitor, bundledImports) {
                    var parameterDeclaration = ts.getDeclarationOfKind(parameterSymbol, 162 /* Parameter */);
                    if (!parameterDeclaration && !ts.isTransientSymbol(parameterSymbol)) {
                        parameterDeclaration = ts.getDeclarationOfKind(parameterSymbol, 335 /* JSDocParameterTag */);
                    }
                    var parameterType = getTypeOfSymbol(parameterSymbol);
                    if (parameterDeclaration && isRequiredInitializedParameter(parameterDeclaration)) {
                        parameterType = getOptionalType(parameterType);
                    }
                    if ((context.flags & 1073741824 /* NoUndefinedOptionalParameterType */) && parameterDeclaration && !ts.isJSDocParameterTag(parameterDeclaration) && isOptionalUninitializedParameter(parameterDeclaration)) {
                        parameterType = getTypeWithFacts(parameterType, 524288 /* NEUndefined */);
                    }
                    var parameterTypeNode = serializeTypeForDeclaration(context, parameterType, parameterSymbol, context.enclosingDeclaration, privateSymbolVisitor, bundledImports);
                    var modifiers = !(context.flags & 8192 /* OmitParameterModifiers */) && preserveModifierFlags && parameterDeclaration && parameterDeclaration.modifiers ? parameterDeclaration.modifiers.map(ts.factory.cloneNode) : undefined;
                    var isRest = parameterDeclaration && ts.isRestParameter(parameterDeclaration) || ts.getCheckFlags(parameterSymbol) & 32768 /* RestParameter */;
                    var dotDotDotToken = isRest ? ts.factory.createToken(25 /* DotDotDotToken */) : undefined;
                    var name = parameterDeclaration ? parameterDeclaration.name ?
                        parameterDeclaration.name.kind === 79 /* Identifier */ ? ts.setEmitFlags(ts.factory.cloneNode(parameterDeclaration.name), 16777216 /* NoAsciiEscaping */) :
                            parameterDeclaration.name.kind === 159 /* QualifiedName */ ? ts.setEmitFlags(ts.factory.cloneNode(parameterDeclaration.name.right), 16777216 /* NoAsciiEscaping */) :
                                cloneBindingName(parameterDeclaration.name) :
                        ts.symbolName(parameterSymbol) :
                        ts.symbolName(parameterSymbol);
                    var isOptional = parameterDeclaration && isOptionalParameter(parameterDeclaration) || ts.getCheckFlags(parameterSymbol) & 16384 /* OptionalParameter */;
                    var questionToken = isOptional ? ts.factory.createToken(57 /* QuestionToken */) : undefined;
                    var parameterNode = ts.factory.createParameterDeclaration(
                    /*decorators*/ undefined, modifiers, dotDotDotToken, name, questionToken, parameterTypeNode, 
                    /*initializer*/ undefined);
                    context.approximateLength += ts.symbolName(parameterSymbol).length + 3;
                    return parameterNode;
                    function cloneBindingName(node) {
                        return elideInitializerAndSetEmitFlags(node);
                        function elideInitializerAndSetEmitFlags(node) {
                            if (context.tracker.trackSymbol && ts.isComputedPropertyName(node) && isLateBindableName(node)) {
                                trackComputedName(node.expression, context.enclosingDeclaration, context);
                            }
                            var visited = ts.visitEachChild(node, elideInitializerAndSetEmitFlags, ts.nullTransformationContext, /*nodesVisitor*/ undefined, elideInitializerAndSetEmitFlags);
                            if (ts.isBindingElement(visited)) {
                                visited = ts.factory.updateBindingElement(visited, visited.dotDotDotToken, visited.propertyName, visited.name, 
                                /*initializer*/ undefined);
                            }
                            if (!ts.nodeIsSynthesized(visited)) {
                                visited = ts.factory.cloneNode(visited);
                            }
                            return ts.setEmitFlags(visited, 1 /* SingleLine */ | 16777216 /* NoAsciiEscaping */);
                        }
                    }
                }
                function trackComputedName(accessExpression, enclosingDeclaration, context) {
                    if (!context.tracker.trackSymbol)
                        return;
                    // get symbol of the first identifier of the entityName
                    var firstIdentifier = ts.getFirstIdentifier(accessExpression);
                    var name = resolveName(firstIdentifier, firstIdentifier.escapedText, 111551 /* Value */ | 1048576 /* ExportValue */, /*nodeNotFoundErrorMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ true);
                    if (name) {
                        context.tracker.trackSymbol(name, enclosingDeclaration, 111551 /* Value */);
                    }
                }
                function lookupSymbolChain(symbol, context, meaning, yieldModuleSymbol) {
                    context.tracker.trackSymbol(symbol, context.enclosingDeclaration, meaning); // TODO: GH#18217
                    return lookupSymbolChainWorker(symbol, context, meaning, yieldModuleSymbol);
                }
                function lookupSymbolChainWorker(symbol, context, meaning, yieldModuleSymbol) {
                    // Try to get qualified name if the symbol is not a type parameter and there is an enclosing declaration.
                    var chain;
                    var isTypeParameter = symbol.flags & 262144 /* TypeParameter */;
                    if (!isTypeParameter && (context.enclosingDeclaration || context.flags & 64 /* UseFullyQualifiedType */) && !(context.flags & 134217728 /* DoNotIncludeSymbolChain */)) {
                        chain = ts.Debug.checkDefined(getSymbolChain(symbol, meaning, /*endOfChain*/ true));
                        ts.Debug.assert(chain && chain.length > 0);
                    }
                    else {
                        chain = [symbol];
                    }
                    return chain;
                    /** @param endOfChain Set to false for recursive calls; non-recursive calls should always output something. */
                    function getSymbolChain(symbol, meaning, endOfChain) {
                        var accessibleSymbolChain = getAccessibleSymbolChain(symbol, context.enclosingDeclaration, meaning, !!(context.flags & 128 /* UseOnlyExternalAliasing */));
                        var parentSpecifiers;
                        if (!accessibleSymbolChain ||
                            needsQualification(accessibleSymbolChain[0], context.enclosingDeclaration, accessibleSymbolChain.length === 1 ? meaning : getQualifiedLeftMeaning(meaning))) {
                            // Go up and add our parent.
                            var parents_1 = getContainersOfSymbol(accessibleSymbolChain ? accessibleSymbolChain[0] : symbol, context.enclosingDeclaration, meaning);
                            if (ts.length(parents_1)) {
                                parentSpecifiers = parents_1.map(function (symbol) {
                                    return ts.some(symbol.declarations, hasNonGlobalAugmentationExternalModuleSymbol)
                                        ? getSpecifierForModuleSymbol(symbol, context)
                                        : undefined;
                                });
                                var indices = parents_1.map(function (_, i) { return i; });
                                indices.sort(sortByBestName);
                                var sortedParents = indices.map(function (i) { return parents_1[i]; });
                                for (var _i = 0, sortedParents_1 = sortedParents; _i < sortedParents_1.length; _i++) {
                                    var parent = sortedParents_1[_i];
                                    var parentChain = getSymbolChain(parent, getQualifiedLeftMeaning(meaning), /*endOfChain*/ false);
                                    if (parentChain) {
                                        if (parent.exports && parent.exports.get("export=" /* ExportEquals */) &&
                                            getSymbolIfSameReference(parent.exports.get("export=" /* ExportEquals */), symbol)) {
                                            // parentChain root _is_ symbol - symbol is a module export=, so it kinda looks like it's own parent
                                            // No need to lookup an alias for the symbol in itself
                                            accessibleSymbolChain = parentChain;
                                            break;
                                        }
                                        accessibleSymbolChain = parentChain.concat(accessibleSymbolChain || [getAliasForSymbolInContainer(parent, symbol) || symbol]);
                                        break;
                                    }
                                }
                            }
                        }
                        if (accessibleSymbolChain) {
                            return accessibleSymbolChain;
                        }
                        if (
                        // If this is the last part of outputting the symbol, always output. The cases apply only to parent symbols.
                        endOfChain ||
                            // If a parent symbol is an anonymous type, don't write it.
                            !(symbol.flags & (2048 /* TypeLiteral */ | 4096 /* ObjectLiteral */))) {
                            // If a parent symbol is an external module, don't write it. (We prefer just `x` vs `"foo/bar".x`.)
                            if (!endOfChain && !yieldModuleSymbol && !!ts.forEach(symbol.declarations, hasNonGlobalAugmentationExternalModuleSymbol)) {
                                return;
                            }
                            return [symbol];
                        }
                        function sortByBestName(a, b) {
                            var specifierA = parentSpecifiers[a];
                            var specifierB = parentSpecifiers[b];
                            if (specifierA && specifierB) {
                                var isBRelative = ts.pathIsRelative(specifierB);
                                if (ts.pathIsRelative(specifierA) === isBRelative) {
                                    // Both relative or both non-relative, sort by number of parts
                                    return ts.moduleSpecifiers.countPathComponents(specifierA) - ts.moduleSpecifiers.countPathComponents(specifierB);
                                }
                                if (isBRelative) {
                                    // A is non-relative, B is relative: prefer A
                                    return -1;
                                }
                                // A is relative, B is non-relative: prefer B
                                return 1;
                            }
                            return 0;
                        }
                    }
                }
                function typeParametersToTypeParameterDeclarations(symbol, context) {
                    var typeParameterNodes;
                    var targetSymbol = getTargetSymbol(symbol);
                    if (targetSymbol.flags & (32 /* Class */ | 64 /* Interface */ | 524288 /* TypeAlias */)) {
                        typeParameterNodes = ts.factory.createNodeArray(ts.map(getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(symbol), function (tp) { return typeParameterToDeclaration(tp, context); }));
                    }
                    return typeParameterNodes;
                }
                function lookupTypeParameterNodes(chain, index, context) {
                    var _a;
                    ts.Debug.assert(chain && 0 <= index && index < chain.length);
                    var symbol = chain[index];
                    var symbolId = getSymbolId(symbol);
                    if ((_a = context.typeParameterSymbolList) === null || _a === void 0 ? void 0 : _a.has(symbolId)) {
                        return undefined;
                    }
                    (context.typeParameterSymbolList || (context.typeParameterSymbolList = new ts.Set())).add(symbolId);
                    var typeParameterNodes;
                    if (context.flags & 512 /* WriteTypeParametersInQualifiedName */ && index < (chain.length - 1)) {
                        var parentSymbol = symbol;
                        var nextSymbol_1 = chain[index + 1];
                        if (ts.getCheckFlags(nextSymbol_1) & 1 /* Instantiated */) {
                            var params = getTypeParametersOfClassOrInterface(parentSymbol.flags & 2097152 /* Alias */ ? resolveAlias(parentSymbol) : parentSymbol);
                            typeParameterNodes = mapToTypeNodes(ts.map(params, function (t) { return getMappedType(t, nextSymbol_1.mapper); }), context);
                        }
                        else {
                            typeParameterNodes = typeParametersToTypeParameterDeclarations(symbol, context);
                        }
                    }
                    return typeParameterNodes;
                }
                /**
                 * Given A[B][C][D], finds A[B]
                 */
                function getTopmostIndexedAccessType(top) {
                    if (ts.isIndexedAccessTypeNode(top.objectType)) {
                        return getTopmostIndexedAccessType(top.objectType);
                    }
                    return top;
                }
                function getSpecifierForModuleSymbol(symbol, context) {
                    var _a;
                    var file = ts.getDeclarationOfKind(symbol, 300 /* SourceFile */);
                    if (!file) {
                        var equivalentFileSymbol = ts.firstDefined(symbol.declarations, function (d) { return getFileSymbolIfFileSymbolExportEqualsContainer(d, symbol); });
                        if (equivalentFileSymbol) {
                            file = ts.getDeclarationOfKind(equivalentFileSymbol, 300 /* SourceFile */);
                        }
                    }
                    if (file && file.moduleName !== undefined) {
                        // Use the amd name if it is available
                        return file.moduleName;
                    }
                    if (!file) {
                        if (context.tracker.trackReferencedAmbientModule) {
                            var ambientDecls = ts.filter(symbol.declarations, ts.isAmbientModule);
                            if (ts.length(ambientDecls)) {
                                for (var _i = 0, _b = ambientDecls; _i < _b.length; _i++) {
                                    var decl = _b[_i];
                                    context.tracker.trackReferencedAmbientModule(decl, symbol);
                                }
                            }
                        }
                        if (ambientModuleSymbolRegex.test(symbol.escapedName)) {
                            return symbol.escapedName.substring(1, symbol.escapedName.length - 1);
                        }
                    }
                    if (!context.enclosingDeclaration || !context.tracker.moduleResolverHost) {
                        // If there's no context declaration, we can't lookup a non-ambient specifier, so we just use the symbol name
                        if (ambientModuleSymbolRegex.test(symbol.escapedName)) {
                            return symbol.escapedName.substring(1, symbol.escapedName.length - 1);
                        }
                        return ts.getSourceFileOfNode(ts.getNonAugmentationDeclaration(symbol)).fileName; // A resolver may not be provided for baselines and errors - in those cases we use the fileName in full
                    }
                    var contextFile = ts.getSourceFileOfNode(ts.getOriginalNode(context.enclosingDeclaration));
                    var links = getSymbolLinks(symbol);
                    var specifier = links.specifierCache && links.specifierCache.get(contextFile.path);
                    if (!specifier) {
                        var isBundle_1 = !!ts.outFile(compilerOptions);
                        // For declaration bundles, we need to generate absolute paths relative to the common source dir for imports,
                        // just like how the declaration emitter does for the ambient module declarations - we can easily accomplish this
                        // using the `baseUrl` compiler option (which we would otherwise never use in declaration emit) and a non-relative
                        // specifier preference
                        var moduleResolverHost = context.tracker.moduleResolverHost;
                        var specifierCompilerOptions = isBundle_1 ? __assign(__assign({}, compilerOptions), { baseUrl: moduleResolverHost.getCommonSourceDirectory() }) : compilerOptions;
                        specifier = ts.first(ts.moduleSpecifiers.getModuleSpecifiers(symbol, checker, specifierCompilerOptions, contextFile, moduleResolverHost, { importModuleSpecifierPreference: isBundle_1 ? "non-relative" : "project-relative", importModuleSpecifierEnding: isBundle_1 ? "minimal" : undefined }));
                        (_a = links.specifierCache) !== null && _a !== void 0 ? _a : (links.specifierCache = new ts.Map());
                        links.specifierCache.set(contextFile.path, specifier);
                    }
                    return specifier;
                }
                function symbolToTypeNode(symbol, context, meaning, overrideTypeArguments) {
                    var chain = lookupSymbolChain(symbol, context, meaning, !(context.flags & 16384 /* UseAliasDefinedOutsideCurrentScope */)); // If we're using aliases outside the current scope, dont bother with the module
                    var isTypeOf = meaning === 111551 /* Value */;
                    if (ts.some(chain[0].declarations, hasNonGlobalAugmentationExternalModuleSymbol)) {
                        // module is root, must use `ImportTypeNode`
                        var nonRootParts = chain.length > 1 ? createAccessFromSymbolChain(chain, chain.length - 1, 1) : undefined;
                        var typeParameterNodes = overrideTypeArguments || lookupTypeParameterNodes(chain, 0, context);
                        var specifier = getSpecifierForModuleSymbol(chain[0], context);
                        if (!(context.flags & 67108864 /* AllowNodeModulesRelativePaths */) && ts.getEmitModuleResolutionKind(compilerOptions) === ts.ModuleResolutionKind.NodeJs && specifier.indexOf("/node_modules/") >= 0) {
                            // If ultimately we can only name the symbol with a reference that dives into a `node_modules` folder, we should error
                            // since declaration files with these kinds of references are liable to fail when published :(
                            context.encounteredError = true;
                            if (context.tracker.reportLikelyUnsafeImportRequiredError) {
                                context.tracker.reportLikelyUnsafeImportRequiredError(specifier);
                            }
                        }
                        var lit = ts.factory.createLiteralTypeNode(ts.factory.createStringLiteral(specifier));
                        if (context.tracker.trackExternalModuleSymbolOfImportTypeNode)
                            context.tracker.trackExternalModuleSymbolOfImportTypeNode(chain[0]);
                        context.approximateLength += specifier.length + 10; // specifier + import("")
                        if (!nonRootParts || ts.isEntityName(nonRootParts)) {
                            if (nonRootParts) {
                                var lastId = ts.isIdentifier(nonRootParts) ? nonRootParts : nonRootParts.right;
                                lastId.typeArguments = undefined;
                            }
                            return ts.factory.createImportTypeNode(lit, nonRootParts, typeParameterNodes, isTypeOf);
                        }
                        else {
                            var splitNode = getTopmostIndexedAccessType(nonRootParts);
                            var qualifier = splitNode.objectType.typeName;
                            return ts.factory.createIndexedAccessTypeNode(ts.factory.createImportTypeNode(lit, qualifier, typeParameterNodes, isTypeOf), splitNode.indexType);
                        }
                    }
                    var entityName = createAccessFromSymbolChain(chain, chain.length - 1, 0);
                    if (ts.isIndexedAccessTypeNode(entityName)) {
                        return entityName; // Indexed accesses can never be `typeof`
                    }
                    if (isTypeOf) {
                        return ts.factory.createTypeQueryNode(entityName);
                    }
                    else {
                        var lastId = ts.isIdentifier(entityName) ? entityName : entityName.right;
                        var lastTypeArgs = lastId.typeArguments;
                        lastId.typeArguments = undefined;
                        return ts.factory.createTypeReferenceNode(entityName, lastTypeArgs);
                    }
                    function createAccessFromSymbolChain(chain, index, stopper) {
                        var typeParameterNodes = index === (chain.length - 1) ? overrideTypeArguments : lookupTypeParameterNodes(chain, index, context);
                        var symbol = chain[index];
                        var parent = chain[index - 1];
                        var symbolName;
                        if (index === 0) {
                            context.flags |= 16777216 /* InInitialEntityName */;
                            symbolName = getNameOfSymbolAsWritten(symbol, context);
                            context.approximateLength += (symbolName ? symbolName.length : 0) + 1;
                            context.flags ^= 16777216 /* InInitialEntityName */;
                        }
                        else {
                            if (parent && getExportsOfSymbol(parent)) {
                                var exports_1 = getExportsOfSymbol(parent);
                                ts.forEachEntry(exports_1, function (ex, name) {
                                    if (getSymbolIfSameReference(ex, symbol) && !isLateBoundName(name) && name !== "export=" /* ExportEquals */) {
                                        symbolName = ts.unescapeLeadingUnderscores(name);
                                        return true;
                                    }
                                });
                            }
                        }
                        if (!symbolName) {
                            symbolName = getNameOfSymbolAsWritten(symbol, context);
                        }
                        context.approximateLength += symbolName.length + 1;
                        if (!(context.flags & 16 /* ForbidIndexedAccessSymbolReferences */) && parent &&
                            getMembersOfSymbol(parent) && getMembersOfSymbol(parent).get(symbol.escapedName) &&
                            getSymbolIfSameReference(getMembersOfSymbol(parent).get(symbol.escapedName), symbol)) {
                            // Should use an indexed access
                            var LHS = createAccessFromSymbolChain(chain, index - 1, stopper);
                            if (ts.isIndexedAccessTypeNode(LHS)) {
                                return ts.factory.createIndexedAccessTypeNode(LHS, ts.factory.createLiteralTypeNode(ts.factory.createStringLiteral(symbolName)));
                            }
                            else {
                                return ts.factory.createIndexedAccessTypeNode(ts.factory.createTypeReferenceNode(LHS, typeParameterNodes), ts.factory.createLiteralTypeNode(ts.factory.createStringLiteral(symbolName)));
                            }
                        }
                        var identifier = ts.setEmitFlags(ts.factory.createIdentifier(symbolName, typeParameterNodes), 16777216 /* NoAsciiEscaping */);
                        identifier.symbol = symbol;
                        if (index > stopper) {
                            var LHS = createAccessFromSymbolChain(chain, index - 1, stopper);
                            if (!ts.isEntityName(LHS)) {
                                return ts.Debug.fail("Impossible construct - an export of an indexed access cannot be reachable");
                            }
                            return ts.factory.createQualifiedName(LHS, identifier);
                        }
                        return identifier;
                    }
                }
                function typeParameterShadowsNameInScope(escapedName, context, type) {
                    var result = resolveName(context.enclosingDeclaration, escapedName, 788968 /* Type */, /*nameNotFoundArg*/ undefined, escapedName, /*isUse*/ false);
                    if (result) {
                        if (result.flags & 262144 /* TypeParameter */ && result === type.symbol) {
                            return false;
                        }
                        return true;
                    }
                    return false;
                }
                function typeParameterToName(type, context) {
                    var _a, _b;
                    if (context.flags & 4 /* GenerateNamesForShadowedTypeParams */ && context.typeParameterNames) {
                        var cached = context.typeParameterNames.get(getTypeId(type));
                        if (cached) {
                            return cached;
                        }
                    }
                    var result = symbolToName(type.symbol, context, 788968 /* Type */, /*expectsIdentifier*/ true);
                    if (!(result.kind & 79 /* Identifier */)) {
                        return ts.factory.createIdentifier("(Missing type parameter)");
                    }
                    if (context.flags & 4 /* GenerateNamesForShadowedTypeParams */) {
                        var rawtext = result.escapedText;
                        var i = ((_a = context.typeParameterNamesByTextNextNameCount) === null || _a === void 0 ? void 0 : _a.get(rawtext)) || 0;
                        var text = rawtext;
                        while (((_b = context.typeParameterNamesByText) === null || _b === void 0 ? void 0 : _b.has(text)) || typeParameterShadowsNameInScope(text, context, type)) {
                            i++;
                            text = rawtext + "_" + i;
                        }
                        if (text !== rawtext) {
                            result = ts.factory.createIdentifier(text, result.typeArguments);
                        }
                        // avoiding iterations of the above loop turns out to be worth it when `i` starts to get large, so we cache the max
                        // `i` we've used thus far, to save work later
                        (context.typeParameterNamesByTextNextNameCount || (context.typeParameterNamesByTextNextNameCount = new ts.Map())).set(rawtext, i);
                        (context.typeParameterNames || (context.typeParameterNames = new ts.Map())).set(getTypeId(type), result);
                        (context.typeParameterNamesByText || (context.typeParameterNamesByText = new ts.Set())).add(rawtext);
                    }
                    return result;
                }
                function symbolToName(symbol, context, meaning, expectsIdentifier) {
                    var chain = lookupSymbolChain(symbol, context, meaning);
                    if (expectsIdentifier && chain.length !== 1
                        && !context.encounteredError
                        && !(context.flags & 65536 /* AllowQualifiedNameInPlaceOfIdentifier */)) {
                        context.encounteredError = true;
                    }
                    return createEntityNameFromSymbolChain(chain, chain.length - 1);
                    function createEntityNameFromSymbolChain(chain, index) {
                        var typeParameterNodes = lookupTypeParameterNodes(chain, index, context);
                        var symbol = chain[index];
                        if (index === 0) {
                            context.flags |= 16777216 /* InInitialEntityName */;
                        }
                        var symbolName = getNameOfSymbolAsWritten(symbol, context);
                        if (index === 0) {
                            context.flags ^= 16777216 /* InInitialEntityName */;
                        }
                        var identifier = ts.setEmitFlags(ts.factory.createIdentifier(symbolName, typeParameterNodes), 16777216 /* NoAsciiEscaping */);
                        identifier.symbol = symbol;
                        return index > 0 ? ts.factory.createQualifiedName(createEntityNameFromSymbolChain(chain, index - 1), identifier) : identifier;
                    }
                }
                function symbolToExpression(symbol, context, meaning) {
                    var chain = lookupSymbolChain(symbol, context, meaning);
                    return createExpressionFromSymbolChain(chain, chain.length - 1);
                    function createExpressionFromSymbolChain(chain, index) {
                        var typeParameterNodes = lookupTypeParameterNodes(chain, index, context);
                        var symbol = chain[index];
                        if (index === 0) {
                            context.flags |= 16777216 /* InInitialEntityName */;
                        }
                        var symbolName = getNameOfSymbolAsWritten(symbol, context);
                        if (index === 0) {
                            context.flags ^= 16777216 /* InInitialEntityName */;
                        }
                        var firstChar = symbolName.charCodeAt(0);
                        if (ts.isSingleOrDoubleQuote(firstChar) && ts.some(symbol.declarations, hasNonGlobalAugmentationExternalModuleSymbol)) {
                            return ts.factory.createStringLiteral(getSpecifierForModuleSymbol(symbol, context));
                        }
                        var canUsePropertyAccess = firstChar === 35 /* hash */ ?
                            symbolName.length > 1 && ts.isIdentifierStart(symbolName.charCodeAt(1), languageVersion) :
                            ts.isIdentifierStart(firstChar, languageVersion);
                        if (index === 0 || canUsePropertyAccess) {
                            var identifier = ts.setEmitFlags(ts.factory.createIdentifier(symbolName, typeParameterNodes), 16777216 /* NoAsciiEscaping */);
                            identifier.symbol = symbol;
                            return index > 0 ? ts.factory.createPropertyAccessExpression(createExpressionFromSymbolChain(chain, index - 1), identifier) : identifier;
                        }
                        else {
                            if (firstChar === 91 /* openBracket */) {
                                symbolName = symbolName.substring(1, symbolName.length - 1);
                                firstChar = symbolName.charCodeAt(0);
                            }
                            var expression = void 0;
                            if (ts.isSingleOrDoubleQuote(firstChar)) {
                                expression = ts.factory.createStringLiteral(symbolName
                                    .substring(1, symbolName.length - 1)
                                    .replace(/\\./g, function (s) { return s.substring(1); }), firstChar === 39 /* singleQuote */);
                            }
                            else if (("" + +symbolName) === symbolName) {
                                expression = ts.factory.createNumericLiteral(+symbolName);
                            }
                            if (!expression) {
                                expression = ts.setEmitFlags(ts.factory.createIdentifier(symbolName, typeParameterNodes), 16777216 /* NoAsciiEscaping */);
                                expression.symbol = symbol;
                            }
                            return ts.factory.createElementAccessExpression(createExpressionFromSymbolChain(chain, index - 1), expression);
                        }
                    }
                }
                function isStringNamed(d) {
                    var name = ts.getNameOfDeclaration(d);
                    return !!name && ts.isStringLiteral(name);
                }
                function isSingleQuotedStringNamed(d) {
                    var name = ts.getNameOfDeclaration(d);
                    return !!(name && ts.isStringLiteral(name) && (name.singleQuote || !ts.nodeIsSynthesized(name) && ts.startsWith(ts.getTextOfNode(name, /*includeTrivia*/ false), "'")));
                }
                function getPropertyNameNodeForSymbol(symbol, context) {
                    var singleQuote = !!ts.length(symbol.declarations) && ts.every(symbol.declarations, isSingleQuotedStringNamed);
                    var fromNameType = getPropertyNameNodeForSymbolFromNameType(symbol, context, singleQuote);
                    if (fromNameType) {
                        return fromNameType;
                    }
                    var rawName = ts.unescapeLeadingUnderscores(symbol.escapedName);
                    var stringNamed = !!ts.length(symbol.declarations) && ts.every(symbol.declarations, isStringNamed);
                    return createPropertyNameNodeForIdentifierOrLiteral(rawName, stringNamed, singleQuote);
                }
                // See getNameForSymbolFromNameType for a stringy equivalent
                function getPropertyNameNodeForSymbolFromNameType(symbol, context, singleQuote) {
                    var nameType = getSymbolLinks(symbol).nameType;
                    if (nameType) {
                        if (nameType.flags & 384 /* StringOrNumberLiteral */) {
                            var name = "" + nameType.value;
                            if (!ts.isIdentifierText(name, compilerOptions.target) && !isNumericLiteralName(name)) {
                                return ts.factory.createStringLiteral(name, !!singleQuote);
                            }
                            if (isNumericLiteralName(name) && ts.startsWith(name, "-")) {
                                return ts.factory.createComputedPropertyName(ts.factory.createNumericLiteral(+name));
                            }
                            return createPropertyNameNodeForIdentifierOrLiteral(name);
                        }
                        if (nameType.flags & 8192 /* UniqueESSymbol */) {
                            return ts.factory.createComputedPropertyName(symbolToExpression(nameType.symbol, context, 111551 /* Value */));
                        }
                    }
                }
                function createPropertyNameNodeForIdentifierOrLiteral(name, stringNamed, singleQuote) {
                    return ts.isIdentifierText(name, compilerOptions.target) ? ts.factory.createIdentifier(name) :
                        !stringNamed && isNumericLiteralName(name) && +name >= 0 ? ts.factory.createNumericLiteral(+name) :
                            ts.factory.createStringLiteral(name, !!singleQuote);
                }
                function cloneNodeBuilderContext(context) {
                    var initial = __assign({}, context);
                    // Make type parameters created within this context not consume the name outside this context
                    // The symbol serializer ends up creating many sibling scopes that all need "separate" contexts when
                    // it comes to naming things - within a normal `typeToTypeNode` call, the node builder only ever descends
                    // through the type tree, so the only cases where we could have used distinct sibling scopes was when there
                    // were multiple generic overloads with similar generated type parameter names
                    // The effect:
                    // When we write out
                    // export const x: <T>(x: T) => T
                    // export const y: <T>(x: T) => T
                    // we write it out like that, rather than as
                    // export const x: <T>(x: T) => T
                    // export const y: <T_1>(x: T_1) => T_1
                    if (initial.typeParameterNames) {
                        initial.typeParameterNames = new ts.Map(initial.typeParameterNames);
                    }
                    if (initial.typeParameterNamesByText) {
                        initial.typeParameterNamesByText = new ts.Set(initial.typeParameterNamesByText);
                    }
                    if (initial.typeParameterSymbolList) {
                        initial.typeParameterSymbolList = new ts.Set(initial.typeParameterSymbolList);
                    }
                    initial.tracker = wrapSymbolTrackerToReportForContext(initial, initial.tracker);
                    return initial;
                }
                function getDeclarationWithTypeAnnotation(symbol, enclosingDeclaration) {
                    return symbol.declarations && ts.find(symbol.declarations, function (s) { return !!ts.getEffectiveTypeAnnotationNode(s) && (!enclosingDeclaration || !!ts.findAncestor(s, function (n) { return n === enclosingDeclaration; })); });
                }
                function existingTypeNodeIsNotReferenceOrIsReferenceWithCompatibleTypeArgumentCount(existing, type) {
                    return !(ts.getObjectFlags(type) & 4 /* Reference */) || !ts.isTypeReferenceNode(existing) || ts.length(existing.typeArguments) >= getMinTypeArgumentCount(type.target.typeParameters);
                }
                /**
                 * Unlike `typeToTypeNodeHelper`, this handles setting up the `AllowUniqueESSymbolType` flag
                 * so a `unique symbol` is returned when appropriate for the input symbol, rather than `typeof sym`
                 */
                function serializeTypeForDeclaration(context, type, symbol, enclosingDeclaration, includePrivateSymbol, bundled) {
                    if (type !== errorType && enclosingDeclaration) {
                        var declWithExistingAnnotation = getDeclarationWithTypeAnnotation(symbol, enclosingDeclaration);
                        if (declWithExistingAnnotation && !ts.isFunctionLikeDeclaration(declWithExistingAnnotation) && !ts.isGetAccessorDeclaration(declWithExistingAnnotation)) {
                            // try to reuse the existing annotation
                            var existing = ts.getEffectiveTypeAnnotationNode(declWithExistingAnnotation);
                            if (getTypeFromTypeNode(existing) === type && existingTypeNodeIsNotReferenceOrIsReferenceWithCompatibleTypeArgumentCount(existing, type)) {
                                var result_6 = serializeExistingTypeNode(context, existing, includePrivateSymbol, bundled);
                                if (result_6) {
                                    return result_6;
                                }
                            }
                        }
                    }
                    var oldFlags = context.flags;
                    if (type.flags & 8192 /* UniqueESSymbol */ &&
                        type.symbol === symbol && (!context.enclosingDeclaration || ts.some(symbol.declarations, function (d) { return ts.getSourceFileOfNode(d) === ts.getSourceFileOfNode(context.enclosingDeclaration); }))) {
                        context.flags |= 1048576 /* AllowUniqueESSymbolType */;
                    }
                    var result = typeToTypeNodeHelper(type, context);
                    context.flags = oldFlags;
                    return result;
                }
                function serializeReturnTypeForSignature(context, type, signature, includePrivateSymbol, bundled) {
                    if (type !== errorType && context.enclosingDeclaration) {
                        var annotation = signature.declaration && ts.getEffectiveReturnTypeNode(signature.declaration);
                        if (!!ts.findAncestor(annotation, function (n) { return n === context.enclosingDeclaration; }) && annotation) {
                            var annotated = getTypeFromTypeNode(annotation);
                            var thisInstantiated = annotated.flags & 262144 /* TypeParameter */ && annotated.isThisType ? instantiateType(annotated, signature.mapper) : annotated;
                            if (thisInstantiated === type && existingTypeNodeIsNotReferenceOrIsReferenceWithCompatibleTypeArgumentCount(annotation, type)) {
                                var result = serializeExistingTypeNode(context, annotation, includePrivateSymbol, bundled);
                                if (result) {
                                    return result;
                                }
                            }
                        }
                    }
                    return typeToTypeNodeHelper(type, context);
                }
                function trackExistingEntityName(node, context, includePrivateSymbol) {
                    var _a, _b;
                    var introducesError = false;
                    var leftmost = ts.getFirstIdentifier(node);
                    if (ts.isInJSFile(node) && (ts.isExportsIdentifier(leftmost) || ts.isModuleExportsAccessExpression(leftmost.parent) || (ts.isQualifiedName(leftmost.parent) && ts.isModuleIdentifier(leftmost.parent.left) && ts.isExportsIdentifier(leftmost.parent.right)))) {
                        introducesError = true;
                        return { introducesError: introducesError, node: node };
                    }
                    var sym = resolveEntityName(leftmost, 67108863 /* All */, /*ignoreErrors*/ true, /*dontResolveALias*/ true);
                    if (sym) {
                        if (isSymbolAccessible(sym, context.enclosingDeclaration, 67108863 /* All */, /*shouldComputeAliasesToMakeVisible*/ false).accessibility !== 0 /* Accessible */) {
                            introducesError = true;
                        }
                        else {
                            (_b = (_a = context.tracker) === null || _a === void 0 ? void 0 : _a.trackSymbol) === null || _b === void 0 ? void 0 : _b.call(_a, sym, context.enclosingDeclaration, 67108863 /* All */);
                            includePrivateSymbol === null || includePrivateSymbol === void 0 ? void 0 : includePrivateSymbol(sym);
                        }
                        if (ts.isIdentifier(node)) {
                            var name = sym.flags & 262144 /* TypeParameter */ ? typeParameterToName(getDeclaredTypeOfSymbol(sym), context) : ts.factory.cloneNode(node);
                            name.symbol = sym; // for quickinfo, which uses identifier symbol information
                            return { introducesError: introducesError, node: ts.setEmitFlags(ts.setOriginalNode(name, node), 16777216 /* NoAsciiEscaping */) };
                        }
                    }
                    return { introducesError: introducesError, node: node };
                }
                function serializeExistingTypeNode(context, existing, includePrivateSymbol, bundled) {
                    if (cancellationToken && cancellationToken.throwIfCancellationRequested) {
                        cancellationToken.throwIfCancellationRequested();
                    }
                    var hadError = false;
                    var file = ts.getSourceFileOfNode(existing);
                    var transformed = ts.visitNode(existing, visitExistingNodeTreeSymbols);
                    if (hadError) {
                        return undefined;
                    }
                    return transformed === existing ? ts.setTextRange(ts.factory.cloneNode(existing), existing) : transformed;
                    function visitExistingNodeTreeSymbols(node) {
                        // We don't _actually_ support jsdoc namepath types, emit `any` instead
                        if (ts.isJSDocAllType(node) || node.kind === 314 /* JSDocNamepathType */) {
                            return ts.factory.createKeywordTypeNode(129 /* AnyKeyword */);
                        }
                        if (ts.isJSDocUnknownType(node)) {
                            return ts.factory.createKeywordTypeNode(153 /* UnknownKeyword */);
                        }
                        if (ts.isJSDocNullableType(node)) {
                            return ts.factory.createUnionTypeNode([ts.visitNode(node.type, visitExistingNodeTreeSymbols), ts.factory.createLiteralTypeNode(ts.factory.createNull())]);
                        }
                        if (ts.isJSDocOptionalType(node)) {
                            return ts.factory.createUnionTypeNode([ts.visitNode(node.type, visitExistingNodeTreeSymbols), ts.factory.createKeywordTypeNode(151 /* UndefinedKeyword */)]);
                        }
                        if (ts.isJSDocNonNullableType(node)) {
                            return ts.visitNode(node.type, visitExistingNodeTreeSymbols);
                        }
                        if (ts.isJSDocVariadicType(node)) {
                            return ts.factory.createArrayTypeNode(ts.visitNode(node.type, visitExistingNodeTreeSymbols));
                        }
                        if (ts.isJSDocTypeLiteral(node)) {
                            return ts.factory.createTypeLiteralNode(ts.map(node.jsDocPropertyTags, function (t) {
                                var name = ts.isIdentifier(t.name) ? t.name : t.name.right;
                                var typeViaParent = getTypeOfPropertyOfType(getTypeFromTypeNode(node), name.escapedText);
                                var overrideTypeNode = typeViaParent && t.typeExpression && getTypeFromTypeNode(t.typeExpression.type) !== typeViaParent ? typeToTypeNodeHelper(typeViaParent, context) : undefined;
                                return ts.factory.createPropertySignature(
                                /*modifiers*/ undefined, name, t.isBracketed || t.typeExpression && ts.isJSDocOptionalType(t.typeExpression.type) ? ts.factory.createToken(57 /* QuestionToken */) : undefined, overrideTypeNode || (t.typeExpression && ts.visitNode(t.typeExpression.type, visitExistingNodeTreeSymbols)) || ts.factory.createKeywordTypeNode(129 /* AnyKeyword */));
                            }));
                        }
                        if (ts.isTypeReferenceNode(node) && ts.isIdentifier(node.typeName) && node.typeName.escapedText === "") {
                            return ts.setOriginalNode(ts.factory.createKeywordTypeNode(129 /* AnyKeyword */), node);
                        }
                        if ((ts.isExpressionWithTypeArguments(node) || ts.isTypeReferenceNode(node)) && ts.isJSDocIndexSignature(node)) {
                            return ts.factory.createTypeLiteralNode([ts.factory.createIndexSignature(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, [ts.factory.createParameterDeclaration(
                                    /*decorators*/ undefined, 
                                    /*modifiers*/ undefined, 
                                    /*dotdotdotToken*/ undefined, "x", 
                                    /*questionToken*/ undefined, ts.visitNode(node.typeArguments[0], visitExistingNodeTreeSymbols))], ts.visitNode(node.typeArguments[1], visitExistingNodeTreeSymbols))]);
                        }
                        if (ts.isJSDocFunctionType(node)) {
                            if (ts.isJSDocConstructSignature(node)) {
                                var newTypeNode_1;
                                return ts.factory.createConstructorTypeNode(node.modifiers, ts.visitNodes(node.typeParameters, visitExistingNodeTreeSymbols), ts.mapDefined(node.parameters, function (p, i) {
                                    return p.name && ts.isIdentifier(p.name) && p.name.escapedText === "new" ? (newTypeNode_1 = p.type, undefined) : ts.factory.createParameterDeclaration(
                                    /*decorators*/ undefined, 
                                    /*modifiers*/ undefined, getEffectiveDotDotDotForParameter(p), getNameForJSDocFunctionParameter(p, i), p.questionToken, ts.visitNode(p.type, visitExistingNodeTreeSymbols), 
                                    /*initializer*/ undefined);
                                }), ts.visitNode(newTypeNode_1 || node.type, visitExistingNodeTreeSymbols) || ts.factory.createKeywordTypeNode(129 /* AnyKeyword */));
                            }
                            else {
                                return ts.factory.createFunctionTypeNode(ts.visitNodes(node.typeParameters, visitExistingNodeTreeSymbols), ts.map(node.parameters, function (p, i) {
                                    return ts.factory.createParameterDeclaration(
                                    /*decorators*/ undefined, 
                                    /*modifiers*/ undefined, getEffectiveDotDotDotForParameter(p), getNameForJSDocFunctionParameter(p, i), p.questionToken, ts.visitNode(p.type, visitExistingNodeTreeSymbols), 
                                    /*initializer*/ undefined);
                                }), ts.visitNode(node.type, visitExistingNodeTreeSymbols) || ts.factory.createKeywordTypeNode(129 /* AnyKeyword */));
                            }
                        }
                        if (ts.isTypeReferenceNode(node) && ts.isInJSDoc(node) && (!existingTypeNodeIsNotReferenceOrIsReferenceWithCompatibleTypeArgumentCount(node, getTypeFromTypeNode(node)) || getIntendedTypeFromJSDocTypeReference(node) || unknownSymbol === resolveTypeReferenceName(getTypeReferenceName(node), 788968 /* Type */, /*ignoreErrors*/ true))) {
                            return ts.setOriginalNode(typeToTypeNodeHelper(getTypeFromTypeNode(node), context), node);
                        }
                        if (ts.isLiteralImportTypeNode(node)) {
                            var nodeSymbol = getNodeLinks(node).resolvedSymbol;
                            if (ts.isInJSDoc(node) &&
                                nodeSymbol &&
                                (
                                // The import type resolved using jsdoc fallback logic
                                (!node.isTypeOf && !(nodeSymbol.flags & 788968 /* Type */)) ||
                                    // The import type had type arguments autofilled by js fallback logic
                                    !(ts.length(node.typeArguments) >= getMinTypeArgumentCount(getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(nodeSymbol))))) {
                                return ts.setOriginalNode(typeToTypeNodeHelper(getTypeFromTypeNode(node), context), node);
                            }
                            return ts.factory.updateImportTypeNode(node, ts.factory.updateLiteralTypeNode(node.argument, rewriteModuleSpecifier(node, node.argument.literal)), node.qualifier, ts.visitNodes(node.typeArguments, visitExistingNodeTreeSymbols, ts.isTypeNode), node.isTypeOf);
                        }
                        if (ts.isEntityName(node) || ts.isEntityNameExpression(node)) {
                            var _a = trackExistingEntityName(node, context, includePrivateSymbol), introducesError = _a.introducesError, result = _a.node;
                            hadError = hadError || introducesError;
                            if (result !== node) {
                                return result;
                            }
                        }
                        if (file && ts.isTupleTypeNode(node) && (ts.getLineAndCharacterOfPosition(file, node.pos).line === ts.getLineAndCharacterOfPosition(file, node.end).line)) {
                            ts.setEmitFlags(node, 1 /* SingleLine */);
                        }
                        return ts.visitEachChild(node, visitExistingNodeTreeSymbols, ts.nullTransformationContext);
                        function getEffectiveDotDotDotForParameter(p) {
                            return p.dotDotDotToken || (p.type && ts.isJSDocVariadicType(p.type) ? ts.factory.createToken(25 /* DotDotDotToken */) : undefined);
                        }
                        /** Note that `new:T` parameters are not handled, but should be before calling this function. */
                        function getNameForJSDocFunctionParameter(p, index) {
                            return p.name && ts.isIdentifier(p.name) && p.name.escapedText === "this" ? "this"
                                : getEffectiveDotDotDotForParameter(p) ? "args"
                                    : "arg" + index;
                        }
                        function rewriteModuleSpecifier(parent, lit) {
                            if (bundled) {
                                if (context.tracker && context.tracker.moduleResolverHost) {
                                    var targetFile = getExternalModuleFileFromDeclaration(parent);
                                    if (targetFile) {
                                        var getCanonicalFileName = ts.createGetCanonicalFileName(!!host.useCaseSensitiveFileNames);
                                        var resolverHost = {
                                            getCanonicalFileName: getCanonicalFileName,
                                            getCurrentDirectory: function () { return context.tracker.moduleResolverHost.getCurrentDirectory(); },
                                            getCommonSourceDirectory: function () { return context.tracker.moduleResolverHost.getCommonSourceDirectory(); }
                                        };
                                        var newName = ts.getResolvedExternalModuleName(resolverHost, targetFile);
                                        return ts.factory.createStringLiteral(newName);
                                    }
                                }
                            }
                            else {
                                if (context.tracker && context.tracker.trackExternalModuleSymbolOfImportTypeNode) {
                                    var moduleSym = resolveExternalModuleNameWorker(lit, lit, /*moduleNotFoundError*/ undefined);
                                    if (moduleSym) {
                                        context.tracker.trackExternalModuleSymbolOfImportTypeNode(moduleSym);
                                    }
                                }
                            }
                            return lit;
                        }
                    }
                }
                function symbolTableToDeclarationStatements(symbolTable, context, bundled) {
                    var serializePropertySymbolForClass = makeSerializePropertySymbol(ts.factory.createPropertyDeclaration, 167 /* MethodDeclaration */, /*useAcessors*/ true);
                    var serializePropertySymbolForInterfaceWorker = makeSerializePropertySymbol(function (_decorators, mods, name, question, type) { return ts.factory.createPropertySignature(mods, name, question, type); }, 166 /* MethodSignature */, /*useAcessors*/ false);
                    // TODO: Use `setOriginalNode` on original declaration names where possible so these declarations see some kind of
                    // declaration mapping
                    // We save the enclosing declaration off here so it's not adjusted by well-meaning declaration
                    // emit codepaths which want to apply more specific contexts (so we can still refer to the root real declaration
                    // we're trying to emit from later on)
                    var enclosingDeclaration = context.enclosingDeclaration;
                    var results = [];
                    var visitedSymbols = new ts.Set();
                    var deferredPrivatesStack = [];
                    var oldcontext = context;
                    context = __assign(__assign({}, oldcontext), { usedSymbolNames: new ts.Set(oldcontext.usedSymbolNames), remappedSymbolNames: new ts.Map(), tracker: __assign(__assign({}, oldcontext.tracker), { trackSymbol: function (sym, decl, meaning) {
                                var accessibleResult = isSymbolAccessible(sym, decl, meaning, /*computeAliases*/ false);
                                if (accessibleResult.accessibility === 0 /* Accessible */) {
                                    // Lookup the root symbol of the chain of refs we'll use to access it and serialize it
                                    var chain = lookupSymbolChainWorker(sym, context, meaning);
                                    if (!(sym.flags & 4 /* Property */)) {
                                        includePrivateSymbol(chain[0]);
                                    }
                                }
                                else if (oldcontext.tracker && oldcontext.tracker.trackSymbol) {
                                    return oldcontext.tracker.trackSymbol(sym, decl, meaning);
                                }
                                return false;
                            } }) });
                    context.tracker = wrapSymbolTrackerToReportForContext(context, context.tracker);
                    ts.forEachEntry(symbolTable, function (symbol, name) {
                        var baseName = ts.unescapeLeadingUnderscores(name);
                        void getInternalSymbolName(symbol, baseName); // Called to cache values into `usedSymbolNames` and `remappedSymbolNames`
                    });
                    var addingDeclare = !bundled;
                    var exportEquals = symbolTable.get("export=" /* ExportEquals */);
                    if (exportEquals && symbolTable.size > 1 && exportEquals.flags & 2097152 /* Alias */) {
                        symbolTable = ts.createSymbolTable();
                        // Remove extraneous elements from root symbol table (they'll be mixed back in when the target of the `export=` is looked up)
                        symbolTable.set("export=" /* ExportEquals */, exportEquals);
                    }
                    visitSymbolTable(symbolTable);
                    return mergeRedundantStatements(results);
                    function isIdentifierAndNotUndefined(node) {
                        return !!node && node.kind === 79 /* Identifier */;
                    }
                    function getNamesOfDeclaration(statement) {
                        if (ts.isVariableStatement(statement)) {
                            return ts.filter(ts.map(statement.declarationList.declarations, ts.getNameOfDeclaration), isIdentifierAndNotUndefined);
                        }
                        return ts.filter([ts.getNameOfDeclaration(statement)], isIdentifierAndNotUndefined);
                    }
                    function flattenExportAssignedNamespace(statements) {
                        var exportAssignment = ts.find(statements, ts.isExportAssignment);
                        var nsIndex = ts.findIndex(statements, ts.isModuleDeclaration);
                        var ns = nsIndex !== -1 ? statements[nsIndex] : undefined;
                        if (ns && exportAssignment && exportAssignment.isExportEquals &&
                            ts.isIdentifier(exportAssignment.expression) && ts.isIdentifier(ns.name) && ts.idText(ns.name) === ts.idText(exportAssignment.expression) &&
                            ns.body && ts.isModuleBlock(ns.body)) {
                            // Pass 0: Correct situations where a module has both an `export = ns` and multiple top-level exports by stripping the export modifiers from
                            //  the top-level exports and exporting them in the targeted ns, as can occur when a js file has both typedefs and `module.export` assignments
                            var excessExports = ts.filter(statements, function (s) { return !!(ts.getEffectiveModifierFlags(s) & 1 /* Export */); });
                            var name_2 = ns.name;
                            var body = ns.body;
                            if (ts.length(excessExports)) {
                                ns = ts.factory.updateModuleDeclaration(ns, ns.decorators, ns.modifiers, ns.name, body = ts.factory.updateModuleBlock(body, ts.factory.createNodeArray(__spreadArray(__spreadArray([], ns.body.statements, true), [ts.factory.createExportDeclaration(
                                    /*decorators*/ undefined, 
                                    /*modifiers*/ undefined, 
                                    /*isTypeOnly*/ false, ts.factory.createNamedExports(ts.map(ts.flatMap(excessExports, function (e) { return getNamesOfDeclaration(e); }), function (id) { return ts.factory.createExportSpecifier(/*alias*/ undefined, id); })), 
                                    /*moduleSpecifier*/ undefined)], false))));
                                statements = __spreadArray(__spreadArray(__spreadArray([], statements.slice(0, nsIndex), true), [ns], false), statements.slice(nsIndex + 1), true);
                            }
                            // Pass 1: Flatten `export namespace _exports {} export = _exports;` so long as the `export=` only points at a single namespace declaration
                            if (!ts.find(statements, function (s) { return s !== ns && ts.nodeHasName(s, name_2); })) {
                                results = [];
                                // If the namespace contains no export assignments or declarations, and no declarations flagged with `export`, then _everything_ is exported -
                                // to respect this as the top level, we need to add an `export` modifier to everything
                                var mixinExportFlag_1 = !ts.some(body.statements, function (s) { return ts.hasSyntacticModifier(s, 1 /* Export */) || ts.isExportAssignment(s) || ts.isExportDeclaration(s); });
                                ts.forEach(body.statements, function (s) {
                                    addResult(s, mixinExportFlag_1 ? 1 /* Export */ : 0 /* None */); // Recalculates the ambient (and export, if applicable from above) flag
                                });
                                statements = __spreadArray(__spreadArray([], ts.filter(statements, function (s) { return s !== ns && s !== exportAssignment; }), true), results, true);
                            }
                        }
                        return statements;
                    }
                    function mergeExportDeclarations(statements) {
                        // Pass 2: Combine all `export {}` declarations
                        var exports = ts.filter(statements, function (d) { return ts.isExportDeclaration(d) && !d.moduleSpecifier && !!d.exportClause && ts.isNamedExports(d.exportClause); });
                        if (ts.length(exports) > 1) {
                            var nonExports = ts.filter(statements, function (d) { return !ts.isExportDeclaration(d) || !!d.moduleSpecifier || !d.exportClause; });
                            statements = __spreadArray(__spreadArray([], nonExports, true), [ts.factory.createExportDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, 
                                /*isTypeOnly*/ false, ts.factory.createNamedExports(ts.flatMap(exports, function (e) { return ts.cast(e.exportClause, ts.isNamedExports).elements; })), 
                                /*moduleSpecifier*/ undefined)], false);
                        }
                        // Pass 2b: Also combine all `export {} from "..."` declarations as needed
                        var reexports = ts.filter(statements, function (d) { return ts.isExportDeclaration(d) && !!d.moduleSpecifier && !!d.exportClause && ts.isNamedExports(d.exportClause); });
                        if (ts.length(reexports) > 1) {
                            var groups = ts.group(reexports, function (decl) { return ts.isStringLiteral(decl.moduleSpecifier) ? ">" + decl.moduleSpecifier.text : ">"; });
                            if (groups.length !== reexports.length) {
                                var _loop_9 = function (group_1) {
                                    if (group_1.length > 1) {
                                        // remove group members from statements and then merge group members and add back to statements
                                        statements = __spreadArray(__spreadArray([], ts.filter(statements, function (s) { return group_1.indexOf(s) === -1; }), true), [
                                            ts.factory.createExportDeclaration(
                                            /*decorators*/ undefined, 
                                            /*modifiers*/ undefined, 
                                            /*isTypeOnly*/ false, ts.factory.createNamedExports(ts.flatMap(group_1, function (e) { return ts.cast(e.exportClause, ts.isNamedExports).elements; })), group_1[0].moduleSpecifier)
                                        ], false);
                                    }
                                };
                                for (var _i = 0, groups_1 = groups; _i < groups_1.length; _i++) {
                                    var group_1 = groups_1[_i];
                                    _loop_9(group_1);
                                }
                            }
                        }
                        return statements;
                    }
                    function inlineExportModifiers(statements) {
                        // Pass 3: Move all `export {}`'s to `export` modifiers where possible
                        var index = ts.findIndex(statements, function (d) { return ts.isExportDeclaration(d) && !d.moduleSpecifier && !!d.exportClause && ts.isNamedExports(d.exportClause); });
                        if (index >= 0) {
                            var exportDecl = statements[index];
                            var replacements = ts.mapDefined(exportDecl.exportClause.elements, function (e) {
                                if (!e.propertyName) {
                                    // export {name} - look thru `statements` for `name`, and if all results can take an `export` modifier, do so and filter it
                                    var indices = ts.indicesOf(statements);
                                    var associatedIndices = ts.filter(indices, function (i) { return ts.nodeHasName(statements[i], e.name); });
                                    if (ts.length(associatedIndices) && ts.every(associatedIndices, function (i) { return canHaveExportModifier(statements[i]); })) {
                                        for (var _i = 0, associatedIndices_1 = associatedIndices; _i < associatedIndices_1.length; _i++) {
                                            var index_1 = associatedIndices_1[_i];
                                            statements[index_1] = addExportModifier(statements[index_1]);
                                        }
                                        return undefined;
                                    }
                                }
                                return e;
                            });
                            if (!ts.length(replacements)) {
                                // all clauses removed, remove the export declaration
                                ts.orderedRemoveItemAt(statements, index);
                            }
                            else {
                                // some items filtered, others not - update the export declaration
                                statements[index] = ts.factory.updateExportDeclaration(exportDecl, exportDecl.decorators, exportDecl.modifiers, exportDecl.isTypeOnly, ts.factory.updateNamedExports(exportDecl.exportClause, replacements), exportDecl.moduleSpecifier);
                            }
                        }
                        return statements;
                    }
                    function mergeRedundantStatements(statements) {
                        statements = flattenExportAssignedNamespace(statements);
                        statements = mergeExportDeclarations(statements);
                        statements = inlineExportModifiers(statements);
                        // Not a cleanup, but as a final step: If there is a mix of `export` and non-`export` declarations, but no `export =` or `export {}` add a `export {};` so
                        // declaration privacy is respected.
                        if (enclosingDeclaration &&
                            ((ts.isSourceFile(enclosingDeclaration) && ts.isExternalOrCommonJsModule(enclosingDeclaration)) || ts.isModuleDeclaration(enclosingDeclaration)) &&
                            (!ts.some(statements, ts.isExternalModuleIndicator) || (!ts.hasScopeMarker(statements) && ts.some(statements, ts.needsScopeMarker)))) {
                            statements.push(ts.createEmptyExports(ts.factory));
                        }
                        return statements;
                    }
                    function canHaveExportModifier(node) {
                        return ts.isEnumDeclaration(node) ||
                            ts.isVariableStatement(node) ||
                            ts.isFunctionDeclaration(node) ||
                            ts.isClassDeclaration(node) ||
                            (ts.isModuleDeclaration(node) && !ts.isExternalModuleAugmentation(node) && !ts.isGlobalScopeAugmentation(node)) ||
                            ts.isInterfaceDeclaration(node) ||
                            isTypeDeclaration(node);
                    }
                    function addExportModifier(node) {
                        var flags = (ts.getEffectiveModifierFlags(node) | 1 /* Export */) & ~2 /* Ambient */;
                        return ts.factory.updateModifiers(node, flags);
                    }
                    function removeExportModifier(node) {
                        var flags = ts.getEffectiveModifierFlags(node) & ~1 /* Export */;
                        return ts.factory.updateModifiers(node, flags);
                    }
                    function visitSymbolTable(symbolTable, suppressNewPrivateContext, propertyAsAlias) {
                        if (!suppressNewPrivateContext) {
                            deferredPrivatesStack.push(new ts.Map());
                        }
                        symbolTable.forEach(function (symbol) {
                            serializeSymbol(symbol, /*isPrivate*/ false, !!propertyAsAlias);
                        });
                        if (!suppressNewPrivateContext) {
                            // deferredPrivates will be filled up by visiting the symbol table
                            // And will continue to iterate as elements are added while visited `deferredPrivates`
                            // (As that's how a map iterator is defined to work)
                            deferredPrivatesStack[deferredPrivatesStack.length - 1].forEach(function (symbol) {
                                serializeSymbol(symbol, /*isPrivate*/ true, !!propertyAsAlias);
                            });
                            deferredPrivatesStack.pop();
                        }
                    }
                    function serializeSymbol(symbol, isPrivate, propertyAsAlias) {
                        // cache visited list based on merged symbol, since we want to use the unmerged top-level symbol, but
                        // still skip reserializing it if we encounter the merged product later on
                        var visitedSym = getMergedSymbol(symbol);
                        if (visitedSymbols.has(getSymbolId(visitedSym))) {
                            return; // Already printed
                        }
                        visitedSymbols.add(getSymbolId(visitedSym));
                        // Only actually serialize symbols within the correct enclosing declaration, otherwise do nothing with the out-of-context symbol
                        var skipMembershipCheck = !isPrivate; // We only call this on exported symbols when we know they're in the correct scope
                        if (skipMembershipCheck || (!!ts.length(symbol.declarations) && ts.some(symbol.declarations, function (d) { return !!ts.findAncestor(d, function (n) { return n === enclosingDeclaration; }); }))) {
                            var oldContext = context;
                            context = cloneNodeBuilderContext(context);
                            var result = serializeSymbolWorker(symbol, isPrivate, propertyAsAlias);
                            if (context.reportedDiagnostic) {
                                oldcontext.reportedDiagnostic = context.reportedDiagnostic; // hoist diagnostic result into outer context
                            }
                            context = oldContext;
                            return result;
                        }
                    }
                    // Synthesize declarations for a symbol - might be an Interface, a Class, a Namespace, a Type, a Variable (const, let, or var), an Alias
                    // or a merge of some number of those.
                    // An interesting challenge is ensuring that when classes merge with namespaces and interfaces, is keeping
                    // each symbol in only one of the representations
                    // Also, synthesizing a default export of some kind
                    // If it's an alias: emit `export default ref`
                    // If it's a property: emit `export default _default` with a `_default` prop
                    // If it's a class/interface/function: emit a class/interface/function with a `default` modifier
                    // These forms can merge, eg (`export default 12; export default interface A {}`)
                    function serializeSymbolWorker(symbol, isPrivate, propertyAsAlias) {
                        var _a, _b;
                        var symbolName = ts.unescapeLeadingUnderscores(symbol.escapedName);
                        var isDefault = symbol.escapedName === "default" /* Default */;
                        if (isPrivate && !(context.flags & 131072 /* AllowAnonymousIdentifier */) && ts.isStringANonContextualKeyword(symbolName) && !isDefault) {
                            // Oh no. We cannot use this symbol's name as it's name... It's likely some jsdoc had an invalid name like `export` or `default` :(
                            context.encounteredError = true;
                            // TODO: Issue error via symbol tracker?
                            return; // If we need to emit a private with a keyword name, we're done for, since something else will try to refer to it by that name
                        }
                        var needsPostExportDefault = isDefault && !!(symbol.flags & -113 /* ExportDoesNotSupportDefaultModifier */
                            || (symbol.flags & 16 /* Function */ && ts.length(getPropertiesOfType(getTypeOfSymbol(symbol))))) && !(symbol.flags & 2097152 /* Alias */); // An alias symbol should preclude needing to make an alias ourselves
                        var needsExportDeclaration = !needsPostExportDefault && !isPrivate && ts.isStringANonContextualKeyword(symbolName) && !isDefault;
                        // `serializeVariableOrProperty` will handle adding the export declaration if it is run (since `getInternalSymbolName` will create the name mapping), so we need to ensuer we unset `needsExportDeclaration` if it is
                        if (needsPostExportDefault || needsExportDeclaration) {
                            isPrivate = true;
                        }
                        var modifierFlags = (!isPrivate ? 1 /* Export */ : 0) | (isDefault && !needsPostExportDefault ? 512 /* Default */ : 0);
                        var isConstMergedWithNS = symbol.flags & 1536 /* Module */ &&
                            symbol.flags & (2 /* BlockScopedVariable */ | 1 /* FunctionScopedVariable */ | 4 /* Property */) &&
                            symbol.escapedName !== "export=" /* ExportEquals */;
                        var isConstMergedWithNSPrintableAsSignatureMerge = isConstMergedWithNS && isTypeRepresentableAsFunctionNamespaceMerge(getTypeOfSymbol(symbol), symbol);
                        if (symbol.flags & (16 /* Function */ | 8192 /* Method */) || isConstMergedWithNSPrintableAsSignatureMerge) {
                            serializeAsFunctionNamespaceMerge(getTypeOfSymbol(symbol), symbol, getInternalSymbolName(symbol, symbolName), modifierFlags);
                        }
                        if (symbol.flags & 524288 /* TypeAlias */) {
                            serializeTypeAlias(symbol, symbolName, modifierFlags);
                        }
                        // Need to skip over export= symbols below - json source files get a single `Property` flagged
                        // symbol of name `export=` which needs to be handled like an alias. It's not great, but it is what it is.
                        if (symbol.flags & (2 /* BlockScopedVariable */ | 1 /* FunctionScopedVariable */ | 4 /* Property */)
                            && symbol.escapedName !== "export=" /* ExportEquals */
                            && !(symbol.flags & 4194304 /* Prototype */)
                            && !(symbol.flags & 32 /* Class */)
                            && !isConstMergedWithNSPrintableAsSignatureMerge) {
                            if (propertyAsAlias) {
                                var createdExport = serializeMaybeAliasAssignment(symbol);
                                if (createdExport) {
                                    needsExportDeclaration = false;
                                    needsPostExportDefault = false;
                                }
                            }
                            else {
                                var type = getTypeOfSymbol(symbol);
                                var localName = getInternalSymbolName(symbol, symbolName);
                                if (!(symbol.flags & 16 /* Function */) && isTypeRepresentableAsFunctionNamespaceMerge(type, symbol)) {
                                    // If the type looks like a function declaration + ns could represent it, and it's type is sourced locally, rewrite it into a function declaration + ns
                                    serializeAsFunctionNamespaceMerge(type, symbol, localName, modifierFlags);
                                }
                                else {
                                    // A Class + Property merge is made for a `module.exports.Member = class {}`, and it doesn't serialize well as either a class _or_ a property symbol - in fact, _it behaves like an alias!_
                                    // `var` is `FunctionScopedVariable`, `const` and `let` are `BlockScopedVariable`, and `module.exports.thing =` is `Property`
                                    var flags = !(symbol.flags & 2 /* BlockScopedVariable */) ? undefined
                                        : isConstVariable(symbol) ? 2 /* Const */
                                            : 1 /* Let */;
                                    var name = (needsPostExportDefault || !(symbol.flags & 4 /* Property */)) ? localName : getUnusedName(localName, symbol);
                                    var textRange = symbol.declarations && ts.find(symbol.declarations, function (d) { return ts.isVariableDeclaration(d); });
                                    if (textRange && ts.isVariableDeclarationList(textRange.parent) && textRange.parent.declarations.length === 1) {
                                        textRange = textRange.parent.parent;
                                    }
                                    var propertyAccessRequire = (_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.isPropertyAccessExpression);
                                    if (propertyAccessRequire && ts.isBinaryExpression(propertyAccessRequire.parent) && ts.isIdentifier(propertyAccessRequire.parent.right)
                                        && ((_b = type.symbol) === null || _b === void 0 ? void 0 : _b.valueDeclaration) && ts.isSourceFile(type.symbol.valueDeclaration)) {
                                        var alias = localName === propertyAccessRequire.parent.right.escapedText ? undefined : propertyAccessRequire.parent.right;
                                        addResult(ts.factory.createExportDeclaration(
                                        /*decorators*/ undefined, 
                                        /*modifiers*/ undefined, 
                                        /*isTypeOnly*/ false, ts.factory.createNamedExports([ts.factory.createExportSpecifier(alias, localName)])), 0 /* None */);
                                        context.tracker.trackSymbol(type.symbol, context.enclosingDeclaration, 111551 /* Value */);
                                    }
                                    else {
                                        var statement = ts.setTextRange(ts.factory.createVariableStatement(/*modifiers*/ undefined, ts.factory.createVariableDeclarationList([
                                            ts.factory.createVariableDeclaration(name, /*exclamationToken*/ undefined, serializeTypeForDeclaration(context, type, symbol, enclosingDeclaration, includePrivateSymbol, bundled))
                                        ], flags)), textRange);
                                        addResult(statement, name !== localName ? modifierFlags & ~1 /* Export */ : modifierFlags);
                                        if (name !== localName && !isPrivate) {
                                            // We rename the variable declaration we generate for Property symbols since they may have a name which
                                            // conflicts with a local declaration. For example, given input:
                                            // ```
                                            // function g() {}
                                            // module.exports.g = g
                                            // ```
                                            // In such a situation, we have a local variable named `g`, and a separate exported variable named `g`.
                                            // Naively, we would emit
                                            // ```
                                            // function g() {}
                                            // export const g: typeof g;
                                            // ```
                                            // That's obviously incorrect - the `g` in the type annotation needs to refer to the local `g`, but
                                            // the export declaration shadows it.
                                            // To work around that, we instead write
                                            // ```
                                            // function g() {}
                                            // const g_1: typeof g;
                                            // export { g_1 as g };
                                            // ```
                                            // To create an export named `g` that does _not_ shadow the local `g`
                                            addResult(ts.factory.createExportDeclaration(
                                            /*decorators*/ undefined, 
                                            /*modifiers*/ undefined, 
                                            /*isTypeOnly*/ false, ts.factory.createNamedExports([ts.factory.createExportSpecifier(name, localName)])), 0 /* None */);
                                            needsExportDeclaration = false;
                                            needsPostExportDefault = false;
                                        }
                                    }
                                }
                            }
                        }
                        if (symbol.flags & 384 /* Enum */) {
                            serializeEnum(symbol, symbolName, modifierFlags);
                        }
                        if (symbol.flags & 32 /* Class */) {
                            if (symbol.flags & 4 /* Property */
                                && symbol.valueDeclaration
                                && ts.isBinaryExpression(symbol.valueDeclaration.parent)
                                && ts.isClassExpression(symbol.valueDeclaration.parent.right)) {
                                // Looks like a `module.exports.Sub = class {}` - if we serialize `symbol` as a class, the result will have no members,
                                // since the classiness is actually from the target of the effective alias the symbol is. yes. A BlockScopedVariable|Class|Property
                                // _really_ acts like an Alias, and none of a BlockScopedVariable, Class, or Property. This is the travesty of JS binding today.
                                serializeAsAlias(symbol, getInternalSymbolName(symbol, symbolName), modifierFlags);
                            }
                            else {
                                serializeAsClass(symbol, getInternalSymbolName(symbol, symbolName), modifierFlags);
                            }
                        }
                        if ((symbol.flags & (512 /* ValueModule */ | 1024 /* NamespaceModule */) && (!isConstMergedWithNS || isTypeOnlyNamespace(symbol))) || isConstMergedWithNSPrintableAsSignatureMerge) {
                            serializeModule(symbol, symbolName, modifierFlags);
                        }
                        // The class meaning serialization should handle serializing all interface members
                        if (symbol.flags & 64 /* Interface */ && !(symbol.flags & 32 /* Class */)) {
                            serializeInterface(symbol, symbolName, modifierFlags);
                        }
                        if (symbol.flags & 2097152 /* Alias */) {
                            serializeAsAlias(symbol, getInternalSymbolName(symbol, symbolName), modifierFlags);
                        }
                        if (symbol.flags & 4 /* Property */ && symbol.escapedName === "export=" /* ExportEquals */) {
                            serializeMaybeAliasAssignment(symbol);
                        }
                        if (symbol.flags & 8388608 /* ExportStar */) {
                            // synthesize export * from "moduleReference"
                            // Straightforward - only one thing to do - make an export declaration
                            if (symbol.declarations) {
                                for (var _i = 0, _c = symbol.declarations; _i < _c.length; _i++) {
                                    var node = _c[_i];
                                    var resolvedModule = resolveExternalModuleName(node, node.moduleSpecifier);
                                    if (!resolvedModule)
                                        continue;
                                    addResult(ts.factory.createExportDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*isTypeOnly*/ false, /*exportClause*/ undefined, ts.factory.createStringLiteral(getSpecifierForModuleSymbol(resolvedModule, context))), 0 /* None */);
                                }
                            }
                        }
                        if (needsPostExportDefault) {
                            addResult(ts.factory.createExportAssignment(/*decorators*/ undefined, /*modifiers*/ undefined, /*isExportAssignment*/ false, ts.factory.createIdentifier(getInternalSymbolName(symbol, symbolName))), 0 /* None */);
                        }
                        else if (needsExportDeclaration) {
                            addResult(ts.factory.createExportDeclaration(
                            /*decorators*/ undefined, 
                            /*modifiers*/ undefined, 
                            /*isTypeOnly*/ false, ts.factory.createNamedExports([ts.factory.createExportSpecifier(getInternalSymbolName(symbol, symbolName), symbolName)])), 0 /* None */);
                        }
                    }
                    function includePrivateSymbol(symbol) {
                        if (ts.some(symbol.declarations, ts.isParameterDeclaration))
                            return;
                        ts.Debug.assertIsDefined(deferredPrivatesStack[deferredPrivatesStack.length - 1]);
                        getUnusedName(ts.unescapeLeadingUnderscores(symbol.escapedName), symbol); // Call to cache unique name for symbol
                        // Blanket moving (import) aliases into the root private context should work, since imports are not valid within namespaces
                        // (so they must have been in the root to begin with if they were real imports) cjs `require` aliases (an upcoming feature)
                        // will throw a wrench in this, since those may have been nested, but we'll need to synthesize them in the outer scope
                        // anyway, as that's the only place the import they translate to is valid. In such a case, we might need to use a unique name
                        // for the moved import; which hopefully the above `getUnusedName` call should produce.
                        var isExternalImportAlias = !!(symbol.flags & 2097152 /* Alias */) && !ts.some(symbol.declarations, function (d) {
                            return !!ts.findAncestor(d, ts.isExportDeclaration) ||
                                ts.isNamespaceExport(d) ||
                                (ts.isImportEqualsDeclaration(d) && !ts.isExternalModuleReference(d.moduleReference));
                        });
                        deferredPrivatesStack[isExternalImportAlias ? 0 : (deferredPrivatesStack.length - 1)].set(getSymbolId(symbol), symbol);
                    }
                    function isExportingScope(enclosingDeclaration) {
                        return ((ts.isSourceFile(enclosingDeclaration) && (ts.isExternalOrCommonJsModule(enclosingDeclaration) || ts.isJsonSourceFile(enclosingDeclaration))) ||
                            (ts.isAmbientModule(enclosingDeclaration) && !ts.isGlobalScopeAugmentation(enclosingDeclaration)));
                    }
                    // Prepends a `declare` and/or `export` modifier if the context requires it, and then adds `node` to `result` and returns `node`
                    function addResult(node, additionalModifierFlags) {
                        if (ts.canHaveModifiers(node)) {
                            var newModifierFlags = 0 /* None */;
                            var enclosingDeclaration_1 = context.enclosingDeclaration &&
                                (ts.isJSDocTypeAlias(context.enclosingDeclaration) ? ts.getSourceFileOfNode(context.enclosingDeclaration) : context.enclosingDeclaration);
                            if (additionalModifierFlags & 1 /* Export */ &&
                                enclosingDeclaration_1 && (isExportingScope(enclosingDeclaration_1) || ts.isModuleDeclaration(enclosingDeclaration_1)) &&
                                canHaveExportModifier(node)) {
                                // Classes, namespaces, variables, functions, interfaces, and types should all be `export`ed in a module context if not private
                                newModifierFlags |= 1 /* Export */;
                            }
                            if (addingDeclare && !(newModifierFlags & 1 /* Export */) &&
                                (!enclosingDeclaration_1 || !(enclosingDeclaration_1.flags & 8388608 /* Ambient */)) &&
                                (ts.isEnumDeclaration(node) || ts.isVariableStatement(node) || ts.isFunctionDeclaration(node) || ts.isClassDeclaration(node) || ts.isModuleDeclaration(node))) {
                                // Classes, namespaces, variables, enums, and functions all need `declare` modifiers to be valid in a declaration file top-level scope
                                newModifierFlags |= 2 /* Ambient */;
                            }
                            if ((additionalModifierFlags & 512 /* Default */) && (ts.isClassDeclaration(node) || ts.isInterfaceDeclaration(node) || ts.isFunctionDeclaration(node))) {
                                newModifierFlags |= 512 /* Default */;
                            }
                            if (newModifierFlags) {
                                node = ts.factory.updateModifiers(node, newModifierFlags | ts.getEffectiveModifierFlags(node));
                            }
                        }
                        results.push(node);
                    }
                    function serializeTypeAlias(symbol, symbolName, modifierFlags) {
                        var _a;
                        var aliasType = getDeclaredTypeOfTypeAlias(symbol);
                        var typeParams = getSymbolLinks(symbol).typeParameters;
                        var typeParamDecls = ts.map(typeParams, function (p) { return typeParameterToDeclaration(p, context); });
                        var jsdocAliasDecl = (_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.isJSDocTypeAlias);
                        var commentText = ts.getTextOfJSDocComment(jsdocAliasDecl ? jsdocAliasDecl.comment || jsdocAliasDecl.parent.comment : undefined);
                        var oldFlags = context.flags;
                        context.flags |= 8388608 /* InTypeAlias */;
                        var oldEnclosingDecl = context.enclosingDeclaration;
                        context.enclosingDeclaration = jsdocAliasDecl;
                        var typeNode = jsdocAliasDecl && jsdocAliasDecl.typeExpression
                            && ts.isJSDocTypeExpression(jsdocAliasDecl.typeExpression)
                            && serializeExistingTypeNode(context, jsdocAliasDecl.typeExpression.type, includePrivateSymbol, bundled)
                            || typeToTypeNodeHelper(aliasType, context);
                        addResult(ts.setSyntheticLeadingComments(ts.factory.createTypeAliasDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, getInternalSymbolName(symbol, symbolName), typeParamDecls, typeNode), !commentText ? [] : [{ kind: 3 /* MultiLineCommentTrivia */, text: "*\n * " + commentText.replace(/\n/g, "\n * ") + "\n ", pos: -1, end: -1, hasTrailingNewLine: true }]), modifierFlags);
                        context.flags = oldFlags;
                        context.enclosingDeclaration = oldEnclosingDecl;
                    }
                    function serializeInterface(symbol, symbolName, modifierFlags) {
                        var interfaceType = getDeclaredTypeOfClassOrInterface(symbol);
                        var localParams = getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(symbol);
                        var typeParamDecls = ts.map(localParams, function (p) { return typeParameterToDeclaration(p, context); });
                        var baseTypes = getBaseTypes(interfaceType);
                        var baseType = ts.length(baseTypes) ? getIntersectionType(baseTypes) : undefined;
                        var members = ts.flatMap(getPropertiesOfType(interfaceType), function (p) { return serializePropertySymbolForInterface(p, baseType); });
                        var callSignatures = serializeSignatures(0 /* Call */, interfaceType, baseType, 172 /* CallSignature */);
                        var constructSignatures = serializeSignatures(1 /* Construct */, interfaceType, baseType, 173 /* ConstructSignature */);
                        var indexSignatures = serializeIndexSignatures(interfaceType, baseType);
                        var heritageClauses = !ts.length(baseTypes) ? undefined : [ts.factory.createHeritageClause(94 /* ExtendsKeyword */, ts.mapDefined(baseTypes, function (b) { return trySerializeAsTypeReference(b, 111551 /* Value */); }))];
                        addResult(ts.factory.createInterfaceDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, getInternalSymbolName(symbol, symbolName), typeParamDecls, heritageClauses, __spreadArray(__spreadArray(__spreadArray(__spreadArray([], indexSignatures, true), constructSignatures, true), callSignatures, true), members, true)), modifierFlags);
                    }
                    function getNamespaceMembersForSerialization(symbol) {
                        return !symbol.exports ? [] : ts.filter(ts.arrayFrom(symbol.exports.values()), isNamespaceMember);
                    }
                    function isTypeOnlyNamespace(symbol) {
                        return ts.every(getNamespaceMembersForSerialization(symbol), function (m) { return !(resolveSymbol(m).flags & 111551 /* Value */); });
                    }
                    function serializeModule(symbol, symbolName, modifierFlags) {
                        var members = getNamespaceMembersForSerialization(symbol);
                        // Split NS members up by declaration - members whose parent symbol is the ns symbol vs those whose is not (but were added in later via merging)
                        var locationMap = ts.arrayToMultiMap(members, function (m) { return m.parent && m.parent === symbol ? "real" : "merged"; });
                        var realMembers = locationMap.get("real") || ts.emptyArray;
                        var mergedMembers = locationMap.get("merged") || ts.emptyArray;
                        // TODO: `suppressNewPrivateContext` is questionable -we need to simply be emitting privates in whatever scope they were declared in, rather
                        // than whatever scope we traverse to them in. That's a bit of a complex rewrite, since we're not _actually_ tracking privates at all in advance,
                        // so we don't even have placeholders to fill in.
                        if (ts.length(realMembers)) {
                            var localName = getInternalSymbolName(symbol, symbolName);
                            serializeAsNamespaceDeclaration(realMembers, localName, modifierFlags, !!(symbol.flags & (16 /* Function */ | 67108864 /* Assignment */)));
                        }
                        if (ts.length(mergedMembers)) {
                            var containingFile_1 = ts.getSourceFileOfNode(context.enclosingDeclaration);
                            var localName = getInternalSymbolName(symbol, symbolName);
                            var nsBody = ts.factory.createModuleBlock([ts.factory.createExportDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, 
                                /*isTypeOnly*/ false, ts.factory.createNamedExports(ts.mapDefined(ts.filter(mergedMembers, function (n) { return n.escapedName !== "export=" /* ExportEquals */; }), function (s) {
                                    var _a, _b;
                                    var name = ts.unescapeLeadingUnderscores(s.escapedName);
                                    var localName = getInternalSymbolName(s, name);
                                    var aliasDecl = s.declarations && getDeclarationOfAliasSymbol(s);
                                    if (containingFile_1 && (aliasDecl ? containingFile_1 !== ts.getSourceFileOfNode(aliasDecl) : !ts.some(s.declarations, function (d) { return ts.getSourceFileOfNode(d) === containingFile_1; }))) {
                                        (_b = (_a = context.tracker) === null || _a === void 0 ? void 0 : _a.reportNonlocalAugmentation) === null || _b === void 0 ? void 0 : _b.call(_a, containingFile_1, symbol, s);
                                        return undefined;
                                    }
                                    var target = aliasDecl && getTargetOfAliasDeclaration(aliasDecl, /*dontRecursivelyResolve*/ true);
                                    includePrivateSymbol(target || s);
                                    var targetName = target ? getInternalSymbolName(target, ts.unescapeLeadingUnderscores(target.escapedName)) : localName;
                                    return ts.factory.createExportSpecifier(name === targetName ? undefined : targetName, name);
                                })))]);
                            addResult(ts.factory.createModuleDeclaration(
                            /*decorators*/ undefined, 
                            /*modifiers*/ undefined, ts.factory.createIdentifier(localName), nsBody, 16 /* Namespace */), 0 /* None */);
                        }
                    }
                    function serializeEnum(symbol, symbolName, modifierFlags) {
                        addResult(ts.factory.createEnumDeclaration(
                        /*decorators*/ undefined, ts.factory.createModifiersFromModifierFlags(isConstEnumSymbol(symbol) ? 2048 /* Const */ : 0), getInternalSymbolName(symbol, symbolName), ts.map(ts.filter(getPropertiesOfType(getTypeOfSymbol(symbol)), function (p) { return !!(p.flags & 8 /* EnumMember */); }), function (p) {
                            // TODO: Handle computed names
                            // I hate that to get the initialized value we need to walk back to the declarations here; but there's no
                            // other way to get the possible const value of an enum member that I'm aware of, as the value is cached
                            // _on the declaration_, not on the declaration's symbol...
                            var initializedValue = p.declarations && p.declarations[0] && ts.isEnumMember(p.declarations[0]) ? getConstantValue(p.declarations[0]) : undefined;
                            return ts.factory.createEnumMember(ts.unescapeLeadingUnderscores(p.escapedName), initializedValue === undefined ? undefined :
                                typeof initializedValue === "string" ? ts.factory.createStringLiteral(initializedValue) :
                                    ts.factory.createNumericLiteral(initializedValue));
                        })), modifierFlags);
                    }
                    function serializeAsFunctionNamespaceMerge(type, symbol, localName, modifierFlags) {
                        var signatures = getSignaturesOfType(type, 0 /* Call */);
                        for (var _i = 0, signatures_2 = signatures; _i < signatures_2.length; _i++) {
                            var sig = signatures_2[_i];
                            // Each overload becomes a separate function declaration, in order
                            var decl = signatureToSignatureDeclarationHelper(sig, 254 /* FunctionDeclaration */, context, { name: ts.factory.createIdentifier(localName), privateSymbolVisitor: includePrivateSymbol, bundledImports: bundled });
                            addResult(ts.setTextRange(decl, getSignatureTextRangeLocation(sig)), modifierFlags);
                        }
                        // Module symbol emit will take care of module-y members, provided it has exports
                        if (!(symbol.flags & (512 /* ValueModule */ | 1024 /* NamespaceModule */) && !!symbol.exports && !!symbol.exports.size)) {
                            var props = ts.filter(getPropertiesOfType(type), isNamespaceMember);
                            serializeAsNamespaceDeclaration(props, localName, modifierFlags, /*suppressNewPrivateContext*/ true);
                        }
                    }
                    function getSignatureTextRangeLocation(signature) {
                        if (signature.declaration && signature.declaration.parent) {
                            if (ts.isBinaryExpression(signature.declaration.parent) && ts.getAssignmentDeclarationKind(signature.declaration.parent) === 5 /* Property */) {
                                return signature.declaration.parent;
                            }
                            // for expressions assigned to `var`s, use the `var` as the text range
                            if (ts.isVariableDeclaration(signature.declaration.parent) && signature.declaration.parent.parent) {
                                return signature.declaration.parent.parent;
                            }
                        }
                        return signature.declaration;
                    }
                    function serializeAsNamespaceDeclaration(props, localName, modifierFlags, suppressNewPrivateContext) {
                        if (ts.length(props)) {
                            var localVsRemoteMap = ts.arrayToMultiMap(props, function (p) {
                                return !ts.length(p.declarations) || ts.some(p.declarations, function (d) {
                                    return ts.getSourceFileOfNode(d) === ts.getSourceFileOfNode(context.enclosingDeclaration);
                                }) ? "local" : "remote";
                            });
                            var localProps = localVsRemoteMap.get("local") || ts.emptyArray;
                            // handle remote props first - we need to make an `import` declaration that points at the module containing each remote
                            // prop in the outermost scope (TODO: a namespace within a namespace would need to be appropriately handled by this)
                            // Example:
                            // import Foo_1 = require("./exporter");
                            // export namespace ns {
                            //     import Foo = Foo_1.Foo;
                            //     export { Foo };
                            //     export const c: number;
                            // }
                            // This is needed because in JS, statements like `const x = require("./f")` support both type and value lookup, even if they're
                            // normally just value lookup (so it functions kinda like an alias even when it's not an alias)
                            // _Usually_, we'll simply print the top-level as an alias instead of a `var` in such situations, however is is theoretically
                            // possible to encounter a situation where a type has members from both the current file and other files - in those situations,
                            // emit akin to the above would be needed.
                            // Add a namespace
                            // Create namespace as non-synthetic so it is usable as an enclosing declaration
                            var fakespace = ts.parseNodeFactory.createModuleDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, ts.factory.createIdentifier(localName), ts.factory.createModuleBlock([]), 16 /* Namespace */);
                            ts.setParent(fakespace, enclosingDeclaration);
                            fakespace.locals = ts.createSymbolTable(props);
                            fakespace.symbol = props[0].parent;
                            var oldResults = results;
                            results = [];
                            var oldAddingDeclare = addingDeclare;
                            addingDeclare = false;
                            var subcontext = __assign(__assign({}, context), { enclosingDeclaration: fakespace });
                            var oldContext = context;
                            context = subcontext;
                            // TODO: implement handling for the localVsRemoteMap.get("remote") - should be difficult to trigger (see comment above), as only interesting cross-file js merges should make this possible
                            visitSymbolTable(ts.createSymbolTable(localProps), suppressNewPrivateContext, /*propertyAsAlias*/ true);
                            context = oldContext;
                            addingDeclare = oldAddingDeclare;
                            var declarations = results;
                            results = oldResults;
                            // replace namespace with synthetic version
                            var defaultReplaced = ts.map(declarations, function (d) {
                                return ts.isExportAssignment(d) && !d.isExportEquals && ts.isIdentifier(d.expression) ? ts.factory.createExportDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, 
                                /*isTypeOnly*/ false, ts.factory.createNamedExports([ts.factory.createExportSpecifier(d.expression, ts.factory.createIdentifier("default" /* Default */))])) : d;
                            });
                            var exportModifierStripped = ts.every(defaultReplaced, function (d) { return ts.hasSyntacticModifier(d, 1 /* Export */); }) ? ts.map(defaultReplaced, removeExportModifier) : defaultReplaced;
                            fakespace = ts.factory.updateModuleDeclaration(fakespace, fakespace.decorators, fakespace.modifiers, fakespace.name, ts.factory.createModuleBlock(exportModifierStripped));
                            addResult(fakespace, modifierFlags); // namespaces can never be default exported
                        }
                    }
                    function isNamespaceMember(p) {
                        return !!(p.flags & (788968 /* Type */ | 1920 /* Namespace */ | 2097152 /* Alias */)) ||
                            !(p.flags & 4194304 /* Prototype */ || p.escapedName === "prototype" || p.valueDeclaration && ts.isStatic(p.valueDeclaration) && ts.isClassLike(p.valueDeclaration.parent));
                    }
                    function sanitizeJSDocImplements(clauses) {
                        var result = ts.mapDefined(clauses, function (e) {
                            var _a;
                            var oldEnclosing = context.enclosingDeclaration;
                            context.enclosingDeclaration = e;
                            var expr = e.expression;
                            if (ts.isEntityNameExpression(expr)) {
                                if (ts.isIdentifier(expr) && ts.idText(expr) === "") {
                                    return cleanup(/*result*/ undefined); // Empty heritage clause, should be an error, but prefer emitting no heritage clauses to reemitting the empty one
                                }
                                var introducesError = void 0;
                                (_a = trackExistingEntityName(expr, context, includePrivateSymbol), introducesError = _a.introducesError, expr = _a.node);
                                if (introducesError) {
                                    return cleanup(/*result*/ undefined);
                                }
                            }
                            return cleanup(ts.factory.createExpressionWithTypeArguments(expr, ts.map(e.typeArguments, function (a) {
                                return serializeExistingTypeNode(context, a, includePrivateSymbol, bundled)
                                    || typeToTypeNodeHelper(getTypeFromTypeNode(a), context);
                            })));
                            function cleanup(result) {
                                context.enclosingDeclaration = oldEnclosing;
                                return result;
                            }
                        });
                        if (result.length === clauses.length) {
                            return result;
                        }
                        return undefined;
                    }
                    function serializeAsClass(symbol, localName, modifierFlags) {
                        var _a, _b;
                        var originalDecl = (_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.isClassLike);
                        var oldEnclosing = context.enclosingDeclaration;
                        context.enclosingDeclaration = originalDecl || oldEnclosing;
                        var localParams = getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(symbol);
                        var typeParamDecls = ts.map(localParams, function (p) { return typeParameterToDeclaration(p, context); });
                        var classType = getDeclaredTypeOfClassOrInterface(symbol);
                        var baseTypes = getBaseTypes(classType);
                        var originalImplements = originalDecl && ts.getEffectiveImplementsTypeNodes(originalDecl);
                        var implementsExpressions = originalImplements && sanitizeJSDocImplements(originalImplements)
                            || ts.mapDefined(getImplementsTypes(classType), serializeImplementedType);
                        var staticType = getTypeOfSymbol(symbol);
                        var isClass = !!((_b = staticType.symbol) === null || _b === void 0 ? void 0 : _b.valueDeclaration) && ts.isClassLike(staticType.symbol.valueDeclaration);
                        var staticBaseType = isClass
                            ? getBaseConstructorTypeOfClass(staticType)
                            : anyType;
                        var heritageClauses = __spreadArray(__spreadArray([], !ts.length(baseTypes) ? [] : [ts.factory.createHeritageClause(94 /* ExtendsKeyword */, ts.map(baseTypes, function (b) { return serializeBaseType(b, staticBaseType, localName); }))], true), !ts.length(implementsExpressions) ? [] : [ts.factory.createHeritageClause(117 /* ImplementsKeyword */, implementsExpressions)], true);
                        var symbolProps = getNonInterhitedProperties(classType, baseTypes, getPropertiesOfType(classType));
                        var publicSymbolProps = ts.filter(symbolProps, function (s) {
                            // `valueDeclaration` could be undefined if inherited from
                            // a union/intersection base type, but inherited properties
                            // don't matter here.
                            var valueDecl = s.valueDeclaration;
                            return !!valueDecl && !(ts.isNamedDeclaration(valueDecl) && ts.isPrivateIdentifier(valueDecl.name));
                        });
                        var hasPrivateIdentifier = ts.some(symbolProps, function (s) {
                            // `valueDeclaration` could be undefined if inherited from
                            // a union/intersection base type, but inherited properties
                            // don't matter here.
                            var valueDecl = s.valueDeclaration;
                            return !!valueDecl && ts.isNamedDeclaration(valueDecl) && ts.isPrivateIdentifier(valueDecl.name);
                        });
                        // Boil down all private properties into a single one.
                        var privateProperties = hasPrivateIdentifier ?
                            [ts.factory.createPropertyDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, ts.factory.createPrivateIdentifier("#private"), 
                                /*questionOrExclamationToken*/ undefined, 
                                /*type*/ undefined, 
                                /*initializer*/ undefined)] :
                            ts.emptyArray;
                        var publicProperties = ts.flatMap(publicSymbolProps, function (p) { return serializePropertySymbolForClass(p, /*isStatic*/ false, baseTypes[0]); });
                        // Consider static members empty if symbol also has function or module meaning - function namespacey emit will handle statics
                        var staticMembers = ts.flatMap(ts.filter(getPropertiesOfType(staticType), function (p) { return !(p.flags & 4194304 /* Prototype */) && p.escapedName !== "prototype" && !isNamespaceMember(p); }), function (p) { return serializePropertySymbolForClass(p, /*isStatic*/ true, staticBaseType); });
                        // When we encounter an `X.prototype.y` assignment in a JS file, we bind `X` as a class regardless as to whether
                        // the value is ever initialized with a class or function-like value. For cases where `X` could never be
                        // created via `new`, we will inject a `private constructor()` declaration to indicate it is not createable.
                        var isNonConstructableClassLikeInJsFile = !isClass &&
                            !!symbol.valueDeclaration &&
                            ts.isInJSFile(symbol.valueDeclaration) &&
                            !ts.some(getSignaturesOfType(staticType, 1 /* Construct */));
                        var constructors = isNonConstructableClassLikeInJsFile ?
                            [ts.factory.createConstructorDeclaration(/*decorators*/ undefined, ts.factory.createModifiersFromModifierFlags(8 /* Private */), [], /*body*/ undefined)] :
                            serializeSignatures(1 /* Construct */, staticType, staticBaseType, 169 /* Constructor */);
                        var indexSignatures = serializeIndexSignatures(classType, baseTypes[0]);
                        context.enclosingDeclaration = oldEnclosing;
                        addResult(ts.setTextRange(ts.factory.createClassDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, localName, typeParamDecls, heritageClauses, __spreadArray(__spreadArray(__spreadArray(__spreadArray(__spreadArray([], indexSignatures, true), staticMembers, true), constructors, true), publicProperties, true), privateProperties, true)), symbol.declarations && ts.filter(symbol.declarations, function (d) { return ts.isClassDeclaration(d) || ts.isClassExpression(d); })[0]), modifierFlags);
                    }
                    function serializeAsAlias(symbol, localName, modifierFlags) {
                        var _a, _b, _c, _d, _e;
                        // synthesize an alias, eg `export { symbolName as Name }`
                        // need to mark the alias `symbol` points at
                        // as something we need to serialize as a private declaration as well
                        var node = getDeclarationOfAliasSymbol(symbol);
                        if (!node)
                            return ts.Debug.fail();
                        var target = getMergedSymbol(getTargetOfAliasDeclaration(node, /*dontRecursivelyResolve*/ true));
                        if (!target) {
                            return;
                        }
                        var verbatimTargetName = ts.unescapeLeadingUnderscores(target.escapedName);
                        if (verbatimTargetName === "export=" /* ExportEquals */ && (compilerOptions.esModuleInterop || compilerOptions.allowSyntheticDefaultImports)) {
                            // target refers to an `export=` symbol that was hoisted into a synthetic default - rename here to match
                            verbatimTargetName = "default" /* Default */;
                        }
                        var targetName = getInternalSymbolName(target, verbatimTargetName);
                        includePrivateSymbol(target); // the target may be within the same scope - attempt to serialize it first
                        switch (node.kind) {
                            case 201 /* BindingElement */:
                                if (((_b = (_a = node.parent) === null || _a === void 0 ? void 0 : _a.parent) === null || _b === void 0 ? void 0 : _b.kind) === 252 /* VariableDeclaration */) {
                                    // const { SomeClass } = require('./lib');
                                    var specifier_1 = getSpecifierForModuleSymbol(target.parent || target, context); // './lib'
                                    var propertyName = node.propertyName;
                                    addResult(ts.factory.createImportDeclaration(
                                    /*decorators*/ undefined, 
                                    /*modifiers*/ undefined, ts.factory.createImportClause(/*isTypeOnly*/ false, /*name*/ undefined, ts.factory.createNamedImports([ts.factory.createImportSpecifier(propertyName && ts.isIdentifier(propertyName) ? ts.factory.createIdentifier(ts.idText(propertyName)) : undefined, ts.factory.createIdentifier(localName))])), ts.factory.createStringLiteral(specifier_1)), 0 /* None */);
                                    break;
                                }
                                // We don't know how to serialize this (nested?) binding element
                                ts.Debug.failBadSyntaxKind(((_c = node.parent) === null || _c === void 0 ? void 0 : _c.parent) || node, "Unhandled binding element grandparent kind in declaration serialization");
                                break;
                            case 292 /* ShorthandPropertyAssignment */:
                                if (((_e = (_d = node.parent) === null || _d === void 0 ? void 0 : _d.parent) === null || _e === void 0 ? void 0 : _e.kind) === 219 /* BinaryExpression */) {
                                    // module.exports = { SomeClass }
                                    serializeExportSpecifier(ts.unescapeLeadingUnderscores(symbol.escapedName), targetName);
                                }
                                break;
                            case 252 /* VariableDeclaration */:
                                // commonjs require: const x = require('y')
                                if (ts.isPropertyAccessExpression(node.initializer)) {
                                    // const x = require('y').z
                                    var initializer = node.initializer; // require('y').z
                                    var uniqueName = ts.factory.createUniqueName(localName); // _x
                                    var specifier_2 = getSpecifierForModuleSymbol(target.parent || target, context); // 'y'
                                    // import _x = require('y');
                                    addResult(ts.factory.createImportEqualsDeclaration(
                                    /*decorators*/ undefined, 
                                    /*modifiers*/ undefined, 
                                    /*isTypeOnly*/ false, uniqueName, ts.factory.createExternalModuleReference(ts.factory.createStringLiteral(specifier_2))), 0 /* None */);
                                    // import x = _x.z
                                    addResult(ts.factory.createImportEqualsDeclaration(
                                    /*decorators*/ undefined, 
                                    /*modifiers*/ undefined, 
                                    /*isTypeOnly*/ false, ts.factory.createIdentifier(localName), ts.factory.createQualifiedName(uniqueName, initializer.name)), modifierFlags);
                                    break;
                                }
                            // else fall through and treat commonjs require just like import=
                            case 263 /* ImportEqualsDeclaration */:
                                // This _specifically_ only exists to handle json declarations - where we make aliases, but since
                                // we emit no declarations for the json document, must not refer to it in the declarations
                                if (target.escapedName === "export=" /* ExportEquals */ && ts.some(target.declarations, ts.isJsonSourceFile)) {
                                    serializeMaybeAliasAssignment(symbol);
                                    break;
                                }
                                // Could be a local `import localName = ns.member` or
                                // an external `import localName = require("whatever")`
                                var isLocalImport = !(target.flags & 512 /* ValueModule */) && !ts.isVariableDeclaration(node);
                                addResult(ts.factory.createImportEqualsDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, 
                                /*isTypeOnly*/ false, ts.factory.createIdentifier(localName), isLocalImport
                                    ? symbolToName(target, context, 67108863 /* All */, /*expectsIdentifier*/ false)
                                    : ts.factory.createExternalModuleReference(ts.factory.createStringLiteral(getSpecifierForModuleSymbol(target, context)))), isLocalImport ? modifierFlags : 0 /* None */);
                                break;
                            case 262 /* NamespaceExportDeclaration */:
                                // export as namespace foo
                                // TODO: Not part of a file's local or export symbol tables
                                // Is bound into file.symbol.globalExports instead, which we don't currently traverse
                                addResult(ts.factory.createNamespaceExportDeclaration(ts.idText(node.name)), 0 /* None */);
                                break;
                            case 265 /* ImportClause */:
                                addResult(ts.factory.createImportDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, ts.factory.createImportClause(/*isTypeOnly*/ false, ts.factory.createIdentifier(localName), /*namedBindings*/ undefined), 
                                // We use `target.parent || target` below as `target.parent` is unset when the target is a module which has been export assigned
                                // And then made into a default by the `esModuleInterop` or `allowSyntheticDefaultImports` flag
                                // In such cases, the `target` refers to the module itself already
                                ts.factory.createStringLiteral(getSpecifierForModuleSymbol(target.parent || target, context))), 0 /* None */);
                                break;
                            case 266 /* NamespaceImport */:
                                addResult(ts.factory.createImportDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, ts.factory.createImportClause(/*isTypeOnly*/ false, /*importClause*/ undefined, ts.factory.createNamespaceImport(ts.factory.createIdentifier(localName))), ts.factory.createStringLiteral(getSpecifierForModuleSymbol(target, context))), 0 /* None */);
                                break;
                            case 272 /* NamespaceExport */:
                                addResult(ts.factory.createExportDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, 
                                /*isTypeOnly*/ false, ts.factory.createNamespaceExport(ts.factory.createIdentifier(localName)), ts.factory.createStringLiteral(getSpecifierForModuleSymbol(target, context))), 0 /* None */);
                                break;
                            case 268 /* ImportSpecifier */:
                                addResult(ts.factory.createImportDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, ts.factory.createImportClause(
                                /*isTypeOnly*/ false, 
                                /*importClause*/ undefined, ts.factory.createNamedImports([
                                    ts.factory.createImportSpecifier(localName !== verbatimTargetName ? ts.factory.createIdentifier(verbatimTargetName) : undefined, ts.factory.createIdentifier(localName))
                                ])), ts.factory.createStringLiteral(getSpecifierForModuleSymbol(target.parent || target, context))), 0 /* None */);
                                break;
                            case 273 /* ExportSpecifier */:
                                // does not use localName because the symbol name in this case refers to the name in the exports table,
                                // which we must exactly preserve
                                var specifier = node.parent.parent.moduleSpecifier;
                                // targetName is only used when the target is local, as otherwise the target is an alias that points at
                                // another file
                                serializeExportSpecifier(ts.unescapeLeadingUnderscores(symbol.escapedName), specifier ? verbatimTargetName : targetName, specifier && ts.isStringLiteralLike(specifier) ? ts.factory.createStringLiteral(specifier.text) : undefined);
                                break;
                            case 269 /* ExportAssignment */:
                                serializeMaybeAliasAssignment(symbol);
                                break;
                            case 219 /* BinaryExpression */:
                            case 204 /* PropertyAccessExpression */:
                            case 205 /* ElementAccessExpression */:
                                // Could be best encoded as though an export specifier or as though an export assignment
                                // If name is default or export=, do an export assignment
                                // Otherwise do an export specifier
                                if (symbol.escapedName === "default" /* Default */ || symbol.escapedName === "export=" /* ExportEquals */) {
                                    serializeMaybeAliasAssignment(symbol);
                                }
                                else {
                                    serializeExportSpecifier(localName, targetName);
                                }
                                break;
                            default:
                                return ts.Debug.failBadSyntaxKind(node, "Unhandled alias declaration kind in symbol serializer!");
                        }
                    }
                    function serializeExportSpecifier(localName, targetName, specifier) {
                        addResult(ts.factory.createExportDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, 
                        /*isTypeOnly*/ false, ts.factory.createNamedExports([ts.factory.createExportSpecifier(localName !== targetName ? targetName : undefined, localName)]), specifier), 0 /* None */);
                    }
                    /**
                     * Returns `true` if an export assignment or declaration was produced for the symbol
                     */
                    function serializeMaybeAliasAssignment(symbol) {
                        if (symbol.flags & 4194304 /* Prototype */) {
                            return false;
                        }
                        var name = ts.unescapeLeadingUnderscores(symbol.escapedName);
                        var isExportEquals = name === "export=" /* ExportEquals */;
                        var isDefault = name === "default" /* Default */;
                        var isExportAssignmentCompatibleSymbolName = isExportEquals || isDefault;
                        // synthesize export = ref
                        // ref should refer to either be a locally scoped symbol which we need to emit, or
                        // a reference to another namespace/module which we may need to emit an `import` statement for
                        var aliasDecl = symbol.declarations && getDeclarationOfAliasSymbol(symbol);
                        // serialize what the alias points to, preserve the declaration's initializer
                        var target = aliasDecl && getTargetOfAliasDeclaration(aliasDecl, /*dontRecursivelyResolve*/ true);
                        // If the target resolves and resolves to a thing defined in this file, emit as an alias, otherwise emit as a const
                        if (target && ts.length(target.declarations) && ts.some(target.declarations, function (d) { return ts.getSourceFileOfNode(d) === ts.getSourceFileOfNode(enclosingDeclaration); })) {
                            // In case `target` refers to a namespace member, look at the declaration and serialize the leftmost symbol in it
                            // eg, `namespace A { export class B {} }; exports = A.B;`
                            // Technically, this is all that's required in the case where the assignment is an entity name expression
                            var expr = aliasDecl && ((ts.isExportAssignment(aliasDecl) || ts.isBinaryExpression(aliasDecl)) ? ts.getExportAssignmentExpression(aliasDecl) : ts.getPropertyAssignmentAliasLikeExpression(aliasDecl));
                            var first_1 = expr && ts.isEntityNameExpression(expr) ? getFirstNonModuleExportsIdentifier(expr) : undefined;
                            var referenced = first_1 && resolveEntityName(first_1, 67108863 /* All */, /*ignoreErrors*/ true, /*dontResolveAlias*/ true, enclosingDeclaration);
                            if (referenced || target) {
                                includePrivateSymbol(referenced || target);
                            }
                            // We disable the context's symbol tracker for the duration of this name serialization
                            // as, by virtue of being here, the name is required to print something, and we don't want to
                            // issue a visibility error on it. Only anonymous classes that an alias points at _would_ issue
                            // a visibility error here (as they're not visible within any scope), but we want to hoist them
                            // into the containing scope anyway, so we want to skip the visibility checks.
                            var oldTrack = context.tracker.trackSymbol;
                            context.tracker.trackSymbol = function () { return false; };
                            if (isExportAssignmentCompatibleSymbolName) {
                                results.push(ts.factory.createExportAssignment(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, isExportEquals, symbolToExpression(target, context, 67108863 /* All */)));
                            }
                            else {
                                if (first_1 === expr && first_1) {
                                    // serialize as `export {target as name}`
                                    serializeExportSpecifier(name, ts.idText(first_1));
                                }
                                else if (expr && ts.isClassExpression(expr)) {
                                    serializeExportSpecifier(name, getInternalSymbolName(target, ts.symbolName(target)));
                                }
                                else {
                                    // serialize as `import _Ref = t.arg.et; export { _Ref as name }`
                                    var varName = getUnusedName(name, symbol);
                                    addResult(ts.factory.createImportEqualsDeclaration(
                                    /*decorators*/ undefined, 
                                    /*modifiers*/ undefined, 
                                    /*isTypeOnly*/ false, ts.factory.createIdentifier(varName), symbolToName(target, context, 67108863 /* All */, /*expectsIdentifier*/ false)), 0 /* None */);
                                    serializeExportSpecifier(name, varName);
                                }
                            }
                            context.tracker.trackSymbol = oldTrack;
                            return true;
                        }
                        else {
                            // serialize as an anonymous property declaration
                            var varName = getUnusedName(name, symbol);
                            // We have to use `getWidenedType` here since the object within a json file is unwidened within the file
                            // (Unwidened types can only exist in expression contexts and should never be serialized)
                            var typeToSerialize = getWidenedType(getTypeOfSymbol(getMergedSymbol(symbol)));
                            if (isTypeRepresentableAsFunctionNamespaceMerge(typeToSerialize, symbol)) {
                                // If there are no index signatures and `typeToSerialize` is an object type, emit as a namespace instead of a const
                                serializeAsFunctionNamespaceMerge(typeToSerialize, symbol, varName, isExportAssignmentCompatibleSymbolName ? 0 /* None */ : 1 /* Export */);
                            }
                            else {
                                var statement = ts.factory.createVariableStatement(/*modifiers*/ undefined, ts.factory.createVariableDeclarationList([
                                    ts.factory.createVariableDeclaration(varName, /*exclamationToken*/ undefined, serializeTypeForDeclaration(context, typeToSerialize, symbol, enclosingDeclaration, includePrivateSymbol, bundled))
                                ], 2 /* Const */));
                                // Inlined JSON types exported with [module.]exports= will already emit an export=, so should use `declare`.
                                // Otherwise, the type itself should be exported.
                                addResult(statement, target && target.flags & 4 /* Property */ && target.escapedName === "export=" /* ExportEquals */ ? 2 /* Ambient */
                                    : name === varName ? 1 /* Export */
                                        : 0 /* None */);
                            }
                            if (isExportAssignmentCompatibleSymbolName) {
                                results.push(ts.factory.createExportAssignment(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, isExportEquals, ts.factory.createIdentifier(varName)));
                                return true;
                            }
                            else if (name !== varName) {
                                serializeExportSpecifier(name, varName);
                                return true;
                            }
                            return false;
                        }
                    }
                    function isTypeRepresentableAsFunctionNamespaceMerge(typeToSerialize, hostSymbol) {
                        // Only object types which are not constructable, or indexable, whose members all come from the
                        // context source file, and whose property names are all valid identifiers and not late-bound, _and_
                        // whose input is not type annotated (if the input symbol has an annotation we can reuse, we should prefer it)
                        var ctxSrc = ts.getSourceFileOfNode(context.enclosingDeclaration);
                        return ts.getObjectFlags(typeToSerialize) & (16 /* Anonymous */ | 32 /* Mapped */) &&
                            !ts.length(getIndexInfosOfType(typeToSerialize)) &&
                            !isClassInstanceSide(typeToSerialize) && // While a class instance is potentially representable as a NS, prefer printing a reference to the instance type and serializing the class
                            !!(ts.length(ts.filter(getPropertiesOfType(typeToSerialize), isNamespaceMember)) || ts.length(getSignaturesOfType(typeToSerialize, 0 /* Call */))) &&
                            !ts.length(getSignaturesOfType(typeToSerialize, 1 /* Construct */)) && // TODO: could probably serialize as function + ns + class, now that that's OK
                            !getDeclarationWithTypeAnnotation(hostSymbol, enclosingDeclaration) &&
                            !(typeToSerialize.symbol && ts.some(typeToSerialize.symbol.declarations, function (d) { return ts.getSourceFileOfNode(d) !== ctxSrc; })) &&
                            !ts.some(getPropertiesOfType(typeToSerialize), function (p) { return isLateBoundName(p.escapedName); }) &&
                            !ts.some(getPropertiesOfType(typeToSerialize), function (p) { return ts.some(p.declarations, function (d) { return ts.getSourceFileOfNode(d) !== ctxSrc; }); }) &&
                            ts.every(getPropertiesOfType(typeToSerialize), function (p) { return ts.isIdentifierText(ts.symbolName(p), languageVersion); });
                    }
                    function makeSerializePropertySymbol(createProperty, methodKind, useAccessors) {
                        return function serializePropertySymbol(p, isStatic, baseType) {
                            var _a, _b, _c, _d, _e;
                            var modifierFlags = ts.getDeclarationModifierFlagsFromSymbol(p);
                            var isPrivate = !!(modifierFlags & 8 /* Private */);
                            if (isStatic && (p.flags & (788968 /* Type */ | 1920 /* Namespace */ | 2097152 /* Alias */))) {
                                // Only value-only-meaning symbols can be correctly encoded as class statics, type/namespace/alias meaning symbols
                                // need to be merged namespace members
                                return [];
                            }
                            if (p.flags & 4194304 /* Prototype */ ||
                                (baseType && getPropertyOfType(baseType, p.escapedName)
                                    && isReadonlySymbol(getPropertyOfType(baseType, p.escapedName)) === isReadonlySymbol(p)
                                    && (p.flags & 16777216 /* Optional */) === (getPropertyOfType(baseType, p.escapedName).flags & 16777216 /* Optional */)
                                    && isTypeIdenticalTo(getTypeOfSymbol(p), getTypeOfPropertyOfType(baseType, p.escapedName)))) {
                                return [];
                            }
                            var flag = (modifierFlags & ~256 /* Async */) | (isStatic ? 32 /* Static */ : 0);
                            var name = getPropertyNameNodeForSymbol(p, context);
                            var firstPropertyLikeDecl = (_a = p.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.or(ts.isPropertyDeclaration, ts.isAccessor, ts.isVariableDeclaration, ts.isPropertySignature, ts.isBinaryExpression, ts.isPropertyAccessExpression));
                            if (p.flags & 98304 /* Accessor */ && useAccessors) {
                                var result = [];
                                if (p.flags & 65536 /* SetAccessor */) {
                                    result.push(ts.setTextRange(ts.factory.createSetAccessorDeclaration(
                                    /*decorators*/ undefined, ts.factory.createModifiersFromModifierFlags(flag), name, [ts.factory.createParameterDeclaration(
                                        /*decorators*/ undefined, 
                                        /*modifiers*/ undefined, 
                                        /*dotDotDotToken*/ undefined, "arg", 
                                        /*questionToken*/ undefined, isPrivate ? undefined : serializeTypeForDeclaration(context, getTypeOfSymbol(p), p, enclosingDeclaration, includePrivateSymbol, bundled))], 
                                    /*body*/ undefined), ((_b = p.declarations) === null || _b === void 0 ? void 0 : _b.find(ts.isSetAccessor)) || firstPropertyLikeDecl));
                                }
                                if (p.flags & 32768 /* GetAccessor */) {
                                    var isPrivate_1 = modifierFlags & 8 /* Private */;
                                    result.push(ts.setTextRange(ts.factory.createGetAccessorDeclaration(
                                    /*decorators*/ undefined, ts.factory.createModifiersFromModifierFlags(flag), name, [], isPrivate_1 ? undefined : serializeTypeForDeclaration(context, getTypeOfSymbol(p), p, enclosingDeclaration, includePrivateSymbol, bundled), 
                                    /*body*/ undefined), ((_c = p.declarations) === null || _c === void 0 ? void 0 : _c.find(ts.isGetAccessor)) || firstPropertyLikeDecl));
                                }
                                return result;
                            }
                            // This is an else/if as accessors and properties can't merge in TS, but might in JS
                            // If this happens, we assume the accessor takes priority, as it imposes more constraints
                            else if (p.flags & (4 /* Property */ | 3 /* Variable */ | 98304 /* Accessor */)) {
                                return ts.setTextRange(createProperty(
                                /*decorators*/ undefined, ts.factory.createModifiersFromModifierFlags((isReadonlySymbol(p) ? 64 /* Readonly */ : 0) | flag), name, p.flags & 16777216 /* Optional */ ? ts.factory.createToken(57 /* QuestionToken */) : undefined, isPrivate ? undefined : serializeTypeForDeclaration(context, getTypeOfSymbol(p), p, enclosingDeclaration, includePrivateSymbol, bundled), 
                                // TODO: https://github.com/microsoft/TypeScript/pull/32372#discussion_r328386357
                                // interface members can't have initializers, however class members _can_
                                /*initializer*/ undefined), ((_d = p.declarations) === null || _d === void 0 ? void 0 : _d.find(ts.or(ts.isPropertyDeclaration, ts.isVariableDeclaration))) || firstPropertyLikeDecl);
                            }
                            if (p.flags & (8192 /* Method */ | 16 /* Function */)) {
                                var type = getTypeOfSymbol(p);
                                var signatures = getSignaturesOfType(type, 0 /* Call */);
                                if (flag & 8 /* Private */) {
                                    return ts.setTextRange(createProperty(
                                    /*decorators*/ undefined, ts.factory.createModifiersFromModifierFlags((isReadonlySymbol(p) ? 64 /* Readonly */ : 0) | flag), name, p.flags & 16777216 /* Optional */ ? ts.factory.createToken(57 /* QuestionToken */) : undefined, 
                                    /*type*/ undefined, 
                                    /*initializer*/ undefined), ((_e = p.declarations) === null || _e === void 0 ? void 0 : _e.find(ts.isFunctionLikeDeclaration)) || signatures[0] && signatures[0].declaration || p.declarations && p.declarations[0]);
                                }
                                var results_1 = [];
                                for (var _i = 0, signatures_3 = signatures; _i < signatures_3.length; _i++) {
                                    var sig = signatures_3[_i];
                                    // Each overload becomes a separate method declaration, in order
                                    var decl = signatureToSignatureDeclarationHelper(sig, methodKind, context, {
                                        name: name,
                                        questionToken: p.flags & 16777216 /* Optional */ ? ts.factory.createToken(57 /* QuestionToken */) : undefined,
                                        modifiers: flag ? ts.factory.createModifiersFromModifierFlags(flag) : undefined
                                    });
                                    var location = sig.declaration && ts.isPrototypePropertyAssignment(sig.declaration.parent) ? sig.declaration.parent : sig.declaration;
                                    results_1.push(ts.setTextRange(decl, location));
                                }
                                return results_1;
                            }
                            // The `Constructor`'s symbol isn't in the class's properties lists, obviously, since it's a signature on the static
                            return ts.Debug.fail("Unhandled class member kind! " + (p.__debugFlags || p.flags));
                        };
                    }
                    function serializePropertySymbolForInterface(p, baseType) {
                        return serializePropertySymbolForInterfaceWorker(p, /*isStatic*/ false, baseType);
                    }
                    function serializeSignatures(kind, input, baseType, outputKind) {
                        var signatures = getSignaturesOfType(input, kind);
                        if (kind === 1 /* Construct */) {
                            if (!baseType && ts.every(signatures, function (s) { return ts.length(s.parameters) === 0; })) {
                                return []; // No base type, every constructor is empty - elide the extraneous `constructor()`
                            }
                            if (baseType) {
                                // If there is a base type, if every signature in the class is identical to a signature in the baseType, elide all the declarations
                                var baseSigs = getSignaturesOfType(baseType, 1 /* Construct */);
                                if (!ts.length(baseSigs) && ts.every(signatures, function (s) { return ts.length(s.parameters) === 0; })) {
                                    return []; // Base had no explicit signatures, if all our signatures are also implicit, return an empty list
                                }
                                if (baseSigs.length === signatures.length) {
                                    var failed = false;
                                    for (var i = 0; i < baseSigs.length; i++) {
                                        if (!compareSignaturesIdentical(signatures[i], baseSigs[i], /*partialMatch*/ false, /*ignoreThisTypes*/ false, /*ignoreReturnTypes*/ true, compareTypesIdentical)) {
                                            failed = true;
                                            break;
                                        }
                                    }
                                    if (!failed) {
                                        return []; // Every signature was identical - elide constructor list as it is inherited
                                    }
                                }
                            }
                            var privateProtected = 0;
                            for (var _i = 0, signatures_4 = signatures; _i < signatures_4.length; _i++) {
                                var s = signatures_4[_i];
                                if (s.declaration) {
                                    privateProtected |= ts.getSelectedEffectiveModifierFlags(s.declaration, 8 /* Private */ | 16 /* Protected */);
                                }
                            }
                            if (privateProtected) {
                                return [ts.setTextRange(ts.factory.createConstructorDeclaration(
                                    /*decorators*/ undefined, ts.factory.createModifiersFromModifierFlags(privateProtected), 
                                    /*parameters*/ [], 
                                    /*body*/ undefined), signatures[0].declaration)];
                            }
                        }
                        var results = [];
                        for (var _a = 0, signatures_5 = signatures; _a < signatures_5.length; _a++) {
                            var sig = signatures_5[_a];
                            // Each overload becomes a separate constructor declaration, in order
                            var decl = signatureToSignatureDeclarationHelper(sig, outputKind, context);
                            results.push(ts.setTextRange(decl, sig.declaration));
                        }
                        return results;
                    }
                    function serializeIndexSignatures(input, baseType) {
                        var results = [];
                        for (var _i = 0, _a = getIndexInfosOfType(input); _i < _a.length; _i++) {
                            var info = _a[_i];
                            if (baseType) {
                                var baseInfo = getIndexInfoOfType(baseType, info.keyType);
                                if (baseInfo) {
                                    if (isTypeIdenticalTo(info.type, baseInfo.type)) {
                                        continue; // elide identical index signatures
                                    }
                                }
                            }
                            results.push(indexInfoToIndexSignatureDeclarationHelper(info, context, /*typeNode*/ undefined));
                        }
                        return results;
                    }
                    function serializeBaseType(t, staticType, rootName) {
                        var ref = trySerializeAsTypeReference(t, 111551 /* Value */);
                        if (ref) {
                            return ref;
                        }
                        var tempName = getUnusedName(rootName + "_base");
                        var statement = ts.factory.createVariableStatement(/*modifiers*/ undefined, ts.factory.createVariableDeclarationList([
                            ts.factory.createVariableDeclaration(tempName, /*exclamationToken*/ undefined, typeToTypeNodeHelper(staticType, context))
                        ], 2 /* Const */));
                        addResult(statement, 0 /* None */);
                        return ts.factory.createExpressionWithTypeArguments(ts.factory.createIdentifier(tempName), /*typeArgs*/ undefined);
                    }
                    function trySerializeAsTypeReference(t, flags) {
                        var typeArgs;
                        var reference;
                        // We don't use `isValueSymbolAccessible` below. since that considers alternative containers (like modules)
                        // which we can't write out in a syntactically valid way as an expression
                        if (t.target && isSymbolAccessibleByFlags(t.target.symbol, enclosingDeclaration, flags)) {
                            typeArgs = ts.map(getTypeArguments(t), function (t) { return typeToTypeNodeHelper(t, context); });
                            reference = symbolToExpression(t.target.symbol, context, 788968 /* Type */);
                        }
                        else if (t.symbol && isSymbolAccessibleByFlags(t.symbol, enclosingDeclaration, flags)) {
                            reference = symbolToExpression(t.symbol, context, 788968 /* Type */);
                        }
                        if (reference) {
                            return ts.factory.createExpressionWithTypeArguments(reference, typeArgs);
                        }
                    }
                    function serializeImplementedType(t) {
                        var ref = trySerializeAsTypeReference(t, 788968 /* Type */);
                        if (ref) {
                            return ref;
                        }
                        if (t.symbol) {
                            return ts.factory.createExpressionWithTypeArguments(symbolToExpression(t.symbol, context, 788968 /* Type */), /*typeArgs*/ undefined);
                        }
                    }
                    function getUnusedName(input, symbol) {
                        var _a, _b;
                        var id = symbol ? getSymbolId(symbol) : undefined;
                        if (id) {
                            if (context.remappedSymbolNames.has(id)) {
                                return context.remappedSymbolNames.get(id);
                            }
                        }
                        if (symbol) {
                            input = getNameCandidateWorker(symbol, input);
                        }
                        var i = 0;
                        var original = input;
                        while ((_a = context.usedSymbolNames) === null || _a === void 0 ? void 0 : _a.has(input)) {
                            i++;
                            input = original + "_" + i;
                        }
                        (_b = context.usedSymbolNames) === null || _b === void 0 ? void 0 : _b.add(input);
                        if (id) {
                            context.remappedSymbolNames.set(id, input);
                        }
                        return input;
                    }
                    function getNameCandidateWorker(symbol, localName) {
                        if (localName === "default" /* Default */ || localName === "__class" /* Class */ || localName === "__function" /* Function */) {
                            var flags = context.flags;
                            context.flags |= 16777216 /* InInitialEntityName */;
                            var nameCandidate = getNameOfSymbolAsWritten(symbol, context);
                            context.flags = flags;
                            localName = nameCandidate.length > 0 && ts.isSingleOrDoubleQuote(nameCandidate.charCodeAt(0)) ? ts.stripQuotes(nameCandidate) : nameCandidate;
                        }
                        if (localName === "default" /* Default */) {
                            localName = "_default";
                        }
                        else if (localName === "export=" /* ExportEquals */) {
                            localName = "_exports";
                        }
                        localName = ts.isIdentifierText(localName, languageVersion) && !ts.isStringANonContextualKeyword(localName) ? localName : "_" + localName.replace(/[^a-zA-Z0-9]/g, "_");
                        return localName;
                    }
                    function getInternalSymbolName(symbol, localName) {
                        var id = getSymbolId(symbol);
                        if (context.remappedSymbolNames.has(id)) {
                            return context.remappedSymbolNames.get(id);
                        }
                        localName = getNameCandidateWorker(symbol, localName);
                        // The result of this is going to be used as the symbol's name - lock it in, so `getUnusedName` will also pick it up
                        context.remappedSymbolNames.set(id, localName);
                        return localName;
                    }
                }
            }
            function typePredicateToString(typePredicate, enclosingDeclaration, flags, writer) {
                if (flags === void 0) {
                    flags = 16384 /* UseAliasDefinedOutsideCurrentScope */;
                }
                return writer ? typePredicateToStringWorker(writer).getText() : ts.usingSingleLineStringWriter(typePredicateToStringWorker);
                function typePredicateToStringWorker(writer) {
                    var predicate = ts.factory.createTypePredicateNode(typePredicate.kind === 2 /* AssertsThis */ || typePredicate.kind === 3 /* AssertsIdentifier */ ? ts.factory.createToken(128 /* AssertsKeyword */) : undefined, typePredicate.kind === 1 /* Identifier */ || typePredicate.kind === 3 /* AssertsIdentifier */ ? ts.factory.createIdentifier(typePredicate.parameterName) : ts.factory.createThisTypeNode(), typePredicate.type && nodeBuilder.typeToTypeNode(typePredicate.type, enclosingDeclaration, toNodeBuilderFlags(flags) | 70221824 /* IgnoreErrors */ | 512 /* WriteTypeParametersInQualifiedName */) // TODO: GH#18217
                    );
                    var printer = ts.createPrinter({ removeComments: true });
                    var sourceFile = enclosingDeclaration && ts.getSourceFileOfNode(enclosingDeclaration);
                    printer.writeNode(4 /* Unspecified */, predicate, /*sourceFile*/ sourceFile, writer);
                    return writer;
                }
            }
            function formatUnionTypes(types) {
                var result = [];
                var flags = 0;
                for (var i = 0; i < types.length; i++) {
                    var t = types[i];
                    flags |= t.flags;
                    if (!(t.flags & 98304 /* Nullable */)) {
                        if (t.flags & (512 /* BooleanLiteral */ | 1024 /* EnumLiteral */)) {
                            var baseType = t.flags & 512 /* BooleanLiteral */ ? booleanType : getBaseTypeOfEnumLiteralType(t);
                            if (baseType.flags & 1048576 /* Union */) {
                                var count = baseType.types.length;
                                if (i + count <= types.length && getRegularTypeOfLiteralType(types[i + count - 1]) === getRegularTypeOfLiteralType(baseType.types[count - 1])) {
                                    result.push(baseType);
                                    i += count - 1;
                                    continue;
                                }
                            }
                        }
                        result.push(t);
                    }
                }
                if (flags & 65536 /* Null */)
                    result.push(nullType);
                if (flags & 32768 /* Undefined */)
                    result.push(undefinedType);
                return result || types;
            }
            function visibilityToString(flags) {
                if (flags === 8 /* Private */) {
                    return "private";
                }
                if (flags === 16 /* Protected */) {
                    return "protected";
                }
                return "public";
            }
            function getTypeAliasForTypeLiteral(type) {
                if (type.symbol && type.symbol.flags & 2048 /* TypeLiteral */ && type.symbol.declarations) {
                    var node = ts.walkUpParenthesizedTypes(type.symbol.declarations[0].parent);
                    if (node.kind === 257 /* TypeAliasDeclaration */) {
                        return getSymbolOfNode(node);
                    }
                }
                return undefined;
            }
            function isTopLevelInExternalModuleAugmentation(node) {
                return node && node.parent &&
                    node.parent.kind === 260 /* ModuleBlock */ &&
                    ts.isExternalModuleAugmentation(node.parent.parent);
            }
            function isDefaultBindingContext(location) {
                return location.kind === 300 /* SourceFile */ || ts.isAmbientModule(location);
            }
            function getNameOfSymbolFromNameType(symbol, context) {
                var nameType = getSymbolLinks(symbol).nameType;
                if (nameType) {
                    if (nameType.flags & 384 /* StringOrNumberLiteral */) {
                        var name = "" + nameType.value;
                        if (!ts.isIdentifierText(name, compilerOptions.target) && !isNumericLiteralName(name)) {
                            return "\"" + ts.escapeString(name, 34 /* doubleQuote */) + "\"";
                        }
                        if (isNumericLiteralName(name) && ts.startsWith(name, "-")) {
                            return "[" + name + "]";
                        }
                        return name;
                    }
                    if (nameType.flags & 8192 /* UniqueESSymbol */) {
                        return "[" + getNameOfSymbolAsWritten(nameType.symbol, context) + "]";
                    }
                }
            }
            /**
             * Gets a human-readable name for a symbol.
             * Should *not* be used for the right-hand side of a `.` -- use `symbolName(symbol)` for that instead.
             *
             * Unlike `symbolName(symbol)`, this will include quotes if the name is from a string literal.
             * It will also use a representation of a number as written instead of a decimal form, e.g. `0o11` instead of `9`.
             */
            function getNameOfSymbolAsWritten(symbol, context) {
                if (context && symbol.escapedName === "default" /* Default */ && !(context.flags & 16384 /* UseAliasDefinedOutsideCurrentScope */) &&
                    // If it's not the first part of an entity name, it must print as `default`
                    (!(context.flags & 16777216 /* InInitialEntityName */) ||
                        // if the symbol is synthesized, it will only be referenced externally it must print as `default`
                        !symbol.declarations ||
                        // if not in the same binding context (source file, module declaration), it must print as `default`
                        (context.enclosingDeclaration && ts.findAncestor(symbol.declarations[0], isDefaultBindingContext) !== ts.findAncestor(context.enclosingDeclaration, isDefaultBindingContext)))) {
                    return "default";
                }
                if (symbol.declarations && symbol.declarations.length) {
                    var declaration = ts.firstDefined(symbol.declarations, function (d) { return ts.getNameOfDeclaration(d) ? d : undefined; }); // Try using a declaration with a name, first
                    var name_3 = declaration && ts.getNameOfDeclaration(declaration);
                    if (declaration && name_3) {
                        if (ts.isCallExpression(declaration) && ts.isBindableObjectDefinePropertyCall(declaration)) {
                            return ts.symbolName(symbol);
                        }
                        if (ts.isComputedPropertyName(name_3) && !(ts.getCheckFlags(symbol) & 4096 /* Late */)) {
                            var nameType = getSymbolLinks(symbol).nameType;
                            if (nameType && nameType.flags & 384 /* StringOrNumberLiteral */) {
                                // Computed property name isn't late bound, but has a well-known name type - use name type to generate a symbol name
                                var result = getNameOfSymbolFromNameType(symbol, context);
                                if (result !== undefined) {
                                    return result;
                                }
                            }
                        }
                        return ts.declarationNameToString(name_3);
                    }
                    if (!declaration) {
                        declaration = symbol.declarations[0]; // Declaration may be nameless, but we'll try anyway
                    }
                    if (declaration.parent && declaration.parent.kind === 252 /* VariableDeclaration */) {
                        return ts.declarationNameToString(declaration.parent.name);
                    }
                    switch (declaration.kind) {
                        case 224 /* ClassExpression */:
                        case 211 /* FunctionExpression */:
                        case 212 /* ArrowFunction */:
                            if (context && !context.encounteredError && !(context.flags & 131072 /* AllowAnonymousIdentifier */)) {
                                context.encounteredError = true;
                            }
                            return declaration.kind === 224 /* ClassExpression */ ? "(Anonymous class)" : "(Anonymous function)";
                    }
                }
                var name = getNameOfSymbolFromNameType(symbol, context);
                return name !== undefined ? name : ts.symbolName(symbol);
            }
            function isDeclarationVisible(node) {
                if (node) {
                    var links = getNodeLinks(node);
                    if (links.isVisible === undefined) {
                        links.isVisible = !!determineIfDeclarationIsVisible();
                    }
                    return links.isVisible;
                }
                return false;
                function determineIfDeclarationIsVisible() {
                    switch (node.kind) {
                        case 333 /* JSDocCallbackTag */:
                        case 340 /* JSDocTypedefTag */:
                        case 334 /* JSDocEnumTag */:
                            // Top-level jsdoc type aliases are considered exported
                            // First parent is comment node, second is hosting declaration or token; we only care about those tokens or declarations whose parent is a source file
                            return !!(node.parent && node.parent.parent && node.parent.parent.parent && ts.isSourceFile(node.parent.parent.parent));
                        case 201 /* BindingElement */:
                            return isDeclarationVisible(node.parent.parent);
                        case 252 /* VariableDeclaration */:
                            if (ts.isBindingPattern(node.name) &&
                                !node.name.elements.length) {
                                // If the binding pattern is empty, this variable declaration is not visible
                                return false;
                            }
                        // falls through
                        case 259 /* ModuleDeclaration */:
                        case 255 /* ClassDeclaration */:
                        case 256 /* InterfaceDeclaration */:
                        case 257 /* TypeAliasDeclaration */:
                        case 254 /* FunctionDeclaration */:
                        case 258 /* EnumDeclaration */:
                        case 263 /* ImportEqualsDeclaration */:
                            // external module augmentation is always visible
                            if (ts.isExternalModuleAugmentation(node)) {
                                return true;
                            }
                            var parent = getDeclarationContainer(node);
                            // If the node is not exported or it is not ambient module element (except import declaration)
                            if (!(ts.getCombinedModifierFlags(node) & 1 /* Export */) &&
                                !(node.kind !== 263 /* ImportEqualsDeclaration */ && parent.kind !== 300 /* SourceFile */ && parent.flags & 8388608 /* Ambient */)) {
                                return isGlobalSourceFile(parent);
                            }
                            // Exported members/ambient module elements (exception import declaration) are visible if parent is visible
                            return isDeclarationVisible(parent);
                        case 165 /* PropertyDeclaration */:
                        case 164 /* PropertySignature */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                        case 167 /* MethodDeclaration */:
                        case 166 /* MethodSignature */:
                            if (ts.hasEffectiveModifier(node, 8 /* Private */ | 16 /* Protected */)) {
                                // Private/protected properties/methods are not visible
                                return false;
                            }
                        // Public properties/methods are visible if its parents are visible, so:
                        // falls through
                        case 169 /* Constructor */:
                        case 173 /* ConstructSignature */:
                        case 172 /* CallSignature */:
                        case 174 /* IndexSignature */:
                        case 162 /* Parameter */:
                        case 260 /* ModuleBlock */:
                        case 177 /* FunctionType */:
                        case 178 /* ConstructorType */:
                        case 180 /* TypeLiteral */:
                        case 176 /* TypeReference */:
                        case 181 /* ArrayType */:
                        case 182 /* TupleType */:
                        case 185 /* UnionType */:
                        case 186 /* IntersectionType */:
                        case 189 /* ParenthesizedType */:
                        case 195 /* NamedTupleMember */:
                            return isDeclarationVisible(node.parent);
                        // Default binding, import specifier and namespace import is visible
                        // only on demand so by default it is not visible
                        case 265 /* ImportClause */:
                        case 266 /* NamespaceImport */:
                        case 268 /* ImportSpecifier */:
                            return false;
                        // Type parameters are always visible
                        case 161 /* TypeParameter */:
                        // Source file and namespace export are always visible
                        // falls through
                        case 300 /* SourceFile */:
                        case 262 /* NamespaceExportDeclaration */:
                            return true;
                        // Export assignments do not create name bindings outside the module
                        case 269 /* ExportAssignment */:
                            return false;
                        default:
                            return false;
                    }
                }
            }
            function collectLinkedAliases(node, setVisibility) {
                var exportSymbol;
                if (node.parent && node.parent.kind === 269 /* ExportAssignment */) {
                    exportSymbol = resolveName(node, node.escapedText, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */ | 2097152 /* Alias */, /*nameNotFoundMessage*/ undefined, node, /*isUse*/ false);
                }
                else if (node.parent.kind === 273 /* ExportSpecifier */) {
                    exportSymbol = getTargetOfExportSpecifier(node.parent, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */ | 2097152 /* Alias */);
                }
                var result;
                var visited;
                if (exportSymbol) {
                    visited = new ts.Set();
                    visited.add(getSymbolId(exportSymbol));
                    buildVisibleNodeList(exportSymbol.declarations);
                }
                return result;
                function buildVisibleNodeList(declarations) {
                    ts.forEach(declarations, function (declaration) {
                        var resultNode = getAnyImportSyntax(declaration) || declaration;
                        if (setVisibility) {
                            getNodeLinks(declaration).isVisible = true;
                        }
                        else {
                            result = result || [];
                            ts.pushIfUnique(result, resultNode);
                        }
                        if (ts.isInternalModuleImportEqualsDeclaration(declaration)) {
                            // Add the referenced top container visible
                            var internalModuleReference = declaration.moduleReference;
                            var firstIdentifier = ts.getFirstIdentifier(internalModuleReference);
                            var importSymbol = resolveName(declaration, firstIdentifier.escapedText, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */, undefined, undefined, /*isUse*/ false);
                            if (importSymbol && visited) {
                                if (ts.tryAddToSet(visited, getSymbolId(importSymbol))) {
                                    buildVisibleNodeList(importSymbol.declarations);
                                }
                            }
                        }
                    });
                }
            }
            /**
             * Push an entry on the type resolution stack. If an entry with the given target and the given property name
             * is already on the stack, and no entries in between already have a type, then a circularity has occurred.
             * In this case, the result values of the existing entry and all entries pushed after it are changed to false,
             * and the value false is returned. Otherwise, the new entry is just pushed onto the stack, and true is returned.
             * In order to see if the same query has already been done before, the target object and the propertyName both
             * must match the one passed in.
             *
             * @param target The symbol, type, or signature whose type is being queried
             * @param propertyName The property name that should be used to query the target for its type
             */
            function pushTypeResolution(target, propertyName) {
                var resolutionCycleStartIndex = findResolutionCycleStartIndex(target, propertyName);
                if (resolutionCycleStartIndex >= 0) {
                    // A cycle was found
                    var length_3 = resolutionTargets.length;
                    for (var i = resolutionCycleStartIndex; i < length_3; i++) {
                        resolutionResults[i] = false;
                    }
                    return false;
                }
                resolutionTargets.push(target);
                resolutionResults.push(/*items*/ true);
                resolutionPropertyNames.push(propertyName);
                return true;
            }
            function findResolutionCycleStartIndex(target, propertyName) {
                for (var i = resolutionTargets.length - 1; i >= 0; i--) {
                    if (hasType(resolutionTargets[i], resolutionPropertyNames[i])) {
                        return -1;
                    }
                    if (resolutionTargets[i] === target && resolutionPropertyNames[i] === propertyName) {
                        return i;
                    }
                }
                return -1;
            }
            function hasType(target, propertyName) {
                switch (propertyName) {
                    case 0 /* Type */:
                        return !!getSymbolLinks(target).type;
                    case 5 /* EnumTagType */:
                        return !!(getNodeLinks(target).resolvedEnumType);
                    case 2 /* DeclaredType */:
                        return !!getSymbolLinks(target).declaredType;
                    case 1 /* ResolvedBaseConstructorType */:
                        return !!target.resolvedBaseConstructorType;
                    case 3 /* ResolvedReturnType */:
                        return !!target.resolvedReturnType;
                    case 4 /* ImmediateBaseConstraint */:
                        return !!target.immediateBaseConstraint;
                    case 6 /* ResolvedTypeArguments */:
                        return !!target.resolvedTypeArguments;
                    case 7 /* ResolvedBaseTypes */:
                        return !!target.baseTypesResolved;
                }
                return ts.Debug.assertNever(propertyName);
            }
            /**
             * Pop an entry from the type resolution stack and return its associated result value. The result value will
             * be true if no circularities were detected, or false if a circularity was found.
             */
            function popTypeResolution() {
                resolutionTargets.pop();
                resolutionPropertyNames.pop();
                return resolutionResults.pop();
            }
            function getDeclarationContainer(node) {
                return ts.findAncestor(ts.getRootDeclaration(node), function (node) {
                    switch (node.kind) {
                        case 252 /* VariableDeclaration */:
                        case 253 /* VariableDeclarationList */:
                        case 268 /* ImportSpecifier */:
                        case 267 /* NamedImports */:
                        case 266 /* NamespaceImport */:
                        case 265 /* ImportClause */:
                            return false;
                        default:
                            return true;
                    }
                }).parent;
            }
            function getTypeOfPrototypeProperty(prototype) {
                // TypeScript 1.0 spec (April 2014): 8.4
                // Every class automatically contains a static property member named 'prototype',
                // the type of which is an instantiation of the class type with type Any supplied as a type argument for each type parameter.
                // It is an error to explicitly declare a static property member with the name 'prototype'.
                var classType = getDeclaredTypeOfSymbol(getParentOfSymbol(prototype));
                return classType.typeParameters ? createTypeReference(classType, ts.map(classType.typeParameters, function (_) { return anyType; })) : classType;
            }
            // Return the type of the given property in the given type, or undefined if no such property exists
            function getTypeOfPropertyOfType(type, name) {
                var prop = getPropertyOfType(type, name);
                return prop ? getTypeOfSymbol(prop) : undefined;
            }
            function getTypeOfPropertyOrIndexSignature(type, name) {
                var _a;
                return getTypeOfPropertyOfType(type, name) || ((_a = getApplicableIndexInfoForName(type, name)) === null || _a === void 0 ? void 0 : _a.type) || unknownType;
            }
            function isTypeAny(type) {
                return type && (type.flags & 1 /* Any */) !== 0;
            }
            // Return the type of a binding element parent. We check SymbolLinks first to see if a type has been
            // assigned by contextual typing.
            function getTypeForBindingElementParent(node) {
                var symbol = getSymbolOfNode(node);
                return symbol && getSymbolLinks(symbol).type || getTypeForVariableLikeDeclaration(node, /*includeOptionality*/ false);
            }
            function getRestType(source, properties, symbol) {
                source = filterType(source, function (t) { return !(t.flags & 98304 /* Nullable */); });
                if (source.flags & 131072 /* Never */) {
                    return emptyObjectType;
                }
                if (source.flags & 1048576 /* Union */) {
                    return mapType(source, function (t) { return getRestType(t, properties, symbol); });
                }
                var omitKeyType = getUnionType(ts.map(properties, getLiteralTypeFromPropertyName));
                if (isGenericObjectType(source) || isGenericIndexType(omitKeyType)) {
                    if (omitKeyType.flags & 131072 /* Never */) {
                        return source;
                    }
                    var omitTypeAlias = getGlobalOmitSymbol();
                    if (!omitTypeAlias) {
                        return errorType;
                    }
                    return getTypeAliasInstantiation(omitTypeAlias, [source, omitKeyType]);
                }
                var members = ts.createSymbolTable();
                for (var _i = 0, _a = getPropertiesOfType(source); _i < _a.length; _i++) {
                    var prop = _a[_i];
                    if (!isTypeAssignableTo(getLiteralTypeFromProperty(prop, 8576 /* StringOrNumberLiteralOrUnique */), omitKeyType)
                        && !(ts.getDeclarationModifierFlagsFromSymbol(prop) & (8 /* Private */ | 16 /* Protected */))
                        && isSpreadableProperty(prop)) {
                        members.set(prop.escapedName, getSpreadSymbol(prop, /*readonly*/ false));
                    }
                }
                var result = createAnonymousType(symbol, members, ts.emptyArray, ts.emptyArray, getIndexInfosOfType(source));
                result.objectFlags |= 8388608 /* ObjectRestType */;
                return result;
            }
            function isGenericTypeWithUndefinedConstraint(type) {
                return !!(type.flags & 465829888 /* Instantiable */) && maybeTypeOfKind(getBaseConstraintOfType(type) || unknownType, 32768 /* Undefined */);
            }
            function getNonUndefinedType(type) {
                var typeOrConstraint = someType(type, isGenericTypeWithUndefinedConstraint) ? mapType(type, function (t) { return t.flags & 465829888 /* Instantiable */ ? getBaseConstraintOrType(t) : t; }) : type;
                return getTypeWithFacts(typeOrConstraint, 524288 /* NEUndefined */);
            }
            // Determine the control flow type associated with a destructuring declaration or assignment. The following
            // forms of destructuring are possible:
            //   let { x } = obj;  // BindingElement
            //   let [ x ] = obj;  // BindingElement
            //   { x } = obj;      // ShorthandPropertyAssignment
            //   { x: v } = obj;   // PropertyAssignment
            //   [ x ] = obj;      // Expression
            // We construct a synthetic element access expression corresponding to 'obj.x' such that the control
            // flow analyzer doesn't have to handle all the different syntactic forms.
            function getFlowTypeOfDestructuring(node, declaredType) {
                var reference = getSyntheticElementAccess(node);
                return reference ? getFlowTypeOfReference(reference, declaredType) : declaredType;
            }
            function getSyntheticElementAccess(node) {
                var parentAccess = getParentElementAccess(node);
                if (parentAccess && parentAccess.flowNode) {
                    var propName = getDestructuringPropertyName(node);
                    if (propName) {
                        var literal = ts.setTextRange(ts.parseNodeFactory.createStringLiteral(propName), node);
                        var lhsExpr = ts.isLeftHandSideExpression(parentAccess) ? parentAccess : ts.parseNodeFactory.createParenthesizedExpression(parentAccess);
                        var result = ts.setTextRange(ts.parseNodeFactory.createElementAccessExpression(lhsExpr, literal), node);
                        ts.setParent(literal, result);
                        ts.setParent(result, node);
                        if (lhsExpr !== parentAccess) {
                            ts.setParent(lhsExpr, result);
                        }
                        result.flowNode = parentAccess.flowNode;
                        return result;
                    }
                }
            }
            function getParentElementAccess(node) {
                var ancestor = node.parent.parent;
                switch (ancestor.kind) {
                    case 201 /* BindingElement */:
                    case 291 /* PropertyAssignment */:
                        return getSyntheticElementAccess(ancestor);
                    case 202 /* ArrayLiteralExpression */:
                        return getSyntheticElementAccess(node.parent);
                    case 252 /* VariableDeclaration */:
                        return ancestor.initializer;
                    case 219 /* BinaryExpression */:
                        return ancestor.right;
                }
            }
            function getDestructuringPropertyName(node) {
                var parent = node.parent;
                if (node.kind === 201 /* BindingElement */ && parent.kind === 199 /* ObjectBindingPattern */) {
                    return getLiteralPropertyNameText(node.propertyName || node.name);
                }
                if (node.kind === 291 /* PropertyAssignment */ || node.kind === 292 /* ShorthandPropertyAssignment */) {
                    return getLiteralPropertyNameText(node.name);
                }
                return "" + parent.elements.indexOf(node);
            }
            function getLiteralPropertyNameText(name) {
                var type = getLiteralTypeFromPropertyName(name);
                return type.flags & (128 /* StringLiteral */ | 256 /* NumberLiteral */) ? "" + type.value : undefined;
            }
            /** Return the inferred type for a binding element */
            function getTypeForBindingElement(declaration) {
                var pattern = declaration.parent;
                var parentType = getTypeForBindingElementParent(pattern.parent);
                // If no type or an any type was inferred for parent, infer that for the binding element
                if (!parentType || isTypeAny(parentType)) {
                    return parentType;
                }
                // Relax null check on ambient destructuring parameters, since the parameters have no implementation and are just documentation
                if (strictNullChecks && declaration.flags & 8388608 /* Ambient */ && ts.isParameterDeclaration(declaration)) {
                    parentType = getNonNullableType(parentType);
                }
                // Filter `undefined` from the type we check against if the parent has an initializer and that initializer is not possibly `undefined`
                else if (strictNullChecks && pattern.parent.initializer && !(getTypeFacts(getTypeOfInitializer(pattern.parent.initializer)) & 65536 /* EQUndefined */)) {
                    parentType = getTypeWithFacts(parentType, 524288 /* NEUndefined */);
                }
                var type;
                if (pattern.kind === 199 /* ObjectBindingPattern */) {
                    if (declaration.dotDotDotToken) {
                        parentType = getReducedType(parentType);
                        if (parentType.flags & 2 /* Unknown */ || !isValidSpreadType(parentType)) {
                            error(declaration, ts.Diagnostics.Rest_types_may_only_be_created_from_object_types);
                            return errorType;
                        }
                        var literalMembers = [];
                        for (var _i = 0, _a = pattern.elements; _i < _a.length; _i++) {
                            var element = _a[_i];
                            if (!element.dotDotDotToken) {
                                literalMembers.push(element.propertyName || element.name);
                            }
                        }
                        type = getRestType(parentType, literalMembers, declaration.symbol);
                    }
                    else {
                        // Use explicitly specified property name ({ p: xxx } form), or otherwise the implied name ({ p } form)
                        var name = declaration.propertyName || declaration.name;
                        var indexType = getLiteralTypeFromPropertyName(name);
                        var declaredType = getIndexedAccessType(parentType, indexType, 32 /* ExpressionPosition */, name);
                        type = getFlowTypeOfDestructuring(declaration, declaredType);
                    }
                }
                else {
                    // This elementType will be used if the specific property corresponding to this index is not
                    // present (aka the tuple element property). This call also checks that the parentType is in
                    // fact an iterable or array (depending on target language).
                    var elementType = checkIteratedTypeOrElementType(65 /* Destructuring */ | (declaration.dotDotDotToken ? 0 : 128 /* PossiblyOutOfBounds */), parentType, undefinedType, pattern);
                    var index_2 = pattern.elements.indexOf(declaration);
                    if (declaration.dotDotDotToken) {
                        // If the parent is a tuple type, the rest element has a tuple type of the
                        // remaining tuple element types. Otherwise, the rest element has an array type with same
                        // element type as the parent type.
                        type = everyType(parentType, isTupleType) ?
                            mapType(parentType, function (t) { return sliceTupleType(t, index_2); }) :
                            createArrayType(elementType);
                    }
                    else if (isArrayLikeType(parentType)) {
                        var indexType = getNumberLiteralType(index_2);
                        var accessFlags = 32 /* ExpressionPosition */ | (hasDefaultValue(declaration) ? 16 /* NoTupleBoundsCheck */ : 0);
                        var declaredType = getIndexedAccessTypeOrUndefined(parentType, indexType, accessFlags, declaration.name) || errorType;
                        type = getFlowTypeOfDestructuring(declaration, declaredType);
                    }
                    else {
                        type = elementType;
                    }
                }
                if (!declaration.initializer) {
                    return type;
                }
                if (ts.getEffectiveTypeAnnotationNode(ts.walkUpBindingElementsAndPatterns(declaration))) {
                    // In strict null checking mode, if a default value of a non-undefined type is specified, remove
                    // undefined from the final type.
                    return strictNullChecks && !(getFalsyFlags(checkDeclarationInitializer(declaration)) & 32768 /* Undefined */) ? getNonUndefinedType(type) : type;
                }
                return widenTypeInferredFromInitializer(declaration, getUnionType([getNonUndefinedType(type), checkDeclarationInitializer(declaration)], 2 /* Subtype */));
            }
            function getTypeForDeclarationFromJSDocComment(declaration) {
                var jsdocType = ts.getJSDocType(declaration);
                if (jsdocType) {
                    return getTypeFromTypeNode(jsdocType);
                }
                return undefined;
            }
            function isNullOrUndefined(node) {
                var expr = ts.skipParentheses(node);
                return expr.kind === 104 /* NullKeyword */ || expr.kind === 79 /* Identifier */ && getResolvedSymbol(expr) === undefinedSymbol;
            }
            function isEmptyArrayLiteral(node) {
                var expr = ts.skipParentheses(node);
                return expr.kind === 202 /* ArrayLiteralExpression */ && expr.elements.length === 0;
            }
            function addOptionality(type, isProperty, isOptional) {
                if (isProperty === void 0) {
                    isProperty = false;
                }
                if (isOptional === void 0) {
                    isOptional = true;
                }
                return strictNullChecks && isOptional ? getOptionalType(type, isProperty) : type;
            }
            // Return the inferred type for a variable, parameter, or property declaration
            function getTypeForVariableLikeDeclaration(declaration, includeOptionality) {
                // A variable declared in a for..in statement is of type string, or of type keyof T when the
                // right hand expression is of a type parameter type.
                if (ts.isVariableDeclaration(declaration) && declaration.parent.parent.kind === 241 /* ForInStatement */) {
                    var indexType = getIndexType(getNonNullableTypeIfNeeded(checkExpression(declaration.parent.parent.expression)));
                    return indexType.flags & (262144 /* TypeParameter */ | 4194304 /* Index */) ? getExtractStringType(indexType) : stringType;
                }
                if (ts.isVariableDeclaration(declaration) && declaration.parent.parent.kind === 242 /* ForOfStatement */) {
                    // checkRightHandSideOfForOf will return undefined if the for-of expression type was
                    // missing properties/signatures required to get its iteratedType (like
                    // [Symbol.iterator] or next). This may be because we accessed properties from anyType,
                    // or it may have led to an error inside getElementTypeOfIterable.
                    var forOfStatement = declaration.parent.parent;
                    return checkRightHandSideOfForOf(forOfStatement) || anyType;
                }
                if (ts.isBindingPattern(declaration.parent)) {
                    return getTypeForBindingElement(declaration);
                }
                var isProperty = ts.isPropertyDeclaration(declaration) || ts.isPropertySignature(declaration);
                var isOptional = includeOptionality && (isProperty && !!declaration.questionToken ||
                    ts.isParameter(declaration) && (!!declaration.questionToken || isJSDocOptionalParameter(declaration)) ||
                    isOptionalJSDocPropertyLikeTag(declaration));
                // Use type from type annotation if one is present
                var declaredType = tryGetTypeFromEffectiveTypeNode(declaration);
                if (declaredType) {
                    return addOptionality(declaredType, isProperty, isOptional);
                }
                if ((noImplicitAny || ts.isInJSFile(declaration)) &&
                    ts.isVariableDeclaration(declaration) && !ts.isBindingPattern(declaration.name) &&
                    !(ts.getCombinedModifierFlags(declaration) & 1 /* Export */) && !(declaration.flags & 8388608 /* Ambient */)) {
                    // If --noImplicitAny is on or the declaration is in a Javascript file,
                    // use control flow tracked 'any' type for non-ambient, non-exported var or let variables with no
                    // initializer or a 'null' or 'undefined' initializer.
                    if (!(ts.getCombinedNodeFlags(declaration) & 2 /* Const */) && (!declaration.initializer || isNullOrUndefined(declaration.initializer))) {
                        return autoType;
                    }
                    // Use control flow tracked 'any[]' type for non-ambient, non-exported variables with an empty array
                    // literal initializer.
                    if (declaration.initializer && isEmptyArrayLiteral(declaration.initializer)) {
                        return autoArrayType;
                    }
                }
                if (ts.isParameter(declaration)) {
                    var func = declaration.parent;
                    // For a parameter of a set accessor, use the type of the get accessor if one is present
                    if (func.kind === 171 /* SetAccessor */ && hasBindableName(func)) {
                        var getter = ts.getDeclarationOfKind(getSymbolOfNode(declaration.parent), 170 /* GetAccessor */);
                        if (getter) {
                            var getterSignature = getSignatureFromDeclaration(getter);
                            var thisParameter = getAccessorThisParameter(func);
                            if (thisParameter && declaration === thisParameter) {
                                // Use the type from the *getter*
                                ts.Debug.assert(!thisParameter.type);
                                return getTypeOfSymbol(getterSignature.thisParameter);
                            }
                            return getReturnTypeOfSignature(getterSignature);
                        }
                    }
                    if (ts.isInJSFile(declaration)) {
                        var typeTag = ts.getJSDocType(func);
                        if (typeTag && ts.isFunctionTypeNode(typeTag)) {
                            var signature = getSignatureFromDeclaration(typeTag);
                            var pos = func.parameters.indexOf(declaration);
                            return declaration.dotDotDotToken ? getRestTypeAtPosition(signature, pos) : getTypeAtPosition(signature, pos);
                        }
                    }
                    // Use contextual parameter type if one is available
                    var type = declaration.symbol.escapedName === "this" /* This */ ? getContextualThisParameterType(func) : getContextuallyTypedParameterType(declaration);
                    if (type) {
                        return addOptionality(type, /*isProperty*/ false, isOptional);
                    }
                }
                // Use the type of the initializer expression if one is present and the declaration is
                // not a parameter of a contextually typed function
                if (ts.hasOnlyExpressionInitializer(declaration) && !!declaration.initializer) {
                    if (ts.isInJSFile(declaration) && !ts.isParameter(declaration)) {
                        var containerObjectType = getJSContainerObjectType(declaration, getSymbolOfNode(declaration), ts.getDeclaredExpandoInitializer(declaration));
                        if (containerObjectType) {
                            return containerObjectType;
                        }
                    }
                    var type = widenTypeInferredFromInitializer(declaration, checkDeclarationInitializer(declaration));
                    return addOptionality(type, isProperty, isOptional);
                }
                if (ts.isPropertyDeclaration(declaration) && (noImplicitAny || ts.isInJSFile(declaration))) {
                    // We have a property declaration with no type annotation or initializer, in noImplicitAny mode or a .js file.
                    // Use control flow analysis of this.xxx assignments in the constructor or static block to determine the type of the property.
                    if (!ts.hasStaticModifier(declaration)) {
                        var constructor = findConstructorDeclaration(declaration.parent);
                        var type = constructor ? getFlowTypeInConstructor(declaration.symbol, constructor) :
                            ts.getEffectiveModifierFlags(declaration) & 2 /* Ambient */ ? getTypeOfPropertyInBaseClass(declaration.symbol) :
                                undefined;
                        return type && addOptionality(type, /*isProperty*/ true, isOptional);
                    }
                    else {
                        var staticBlocks = ts.filter(declaration.parent.members, ts.isClassStaticBlockDeclaration);
                        var type = staticBlocks.length ? getFlowTypeInStaticBlocks(declaration.symbol, staticBlocks) :
                            ts.getEffectiveModifierFlags(declaration) & 2 /* Ambient */ ? getTypeOfPropertyInBaseClass(declaration.symbol) :
                                undefined;
                        return type && addOptionality(type, /*isProperty*/ true, isOptional);
                    }
                }
                if (ts.isJsxAttribute(declaration)) {
                    // if JSX attribute doesn't have initializer, by default the attribute will have boolean value of true.
                    // I.e <Elem attr /> is sugar for <Elem attr={true} />
                    return trueType;
                }
                // If the declaration specifies a binding pattern and is not a parameter of a contextually
                // typed function, use the type implied by the binding pattern
                if (ts.isBindingPattern(declaration.name)) {
                    return getTypeFromBindingPattern(declaration.name, /*includePatternInType*/ false, /*reportErrors*/ true);
                }
                // No type specified and nothing can be inferred
                return undefined;
            }
            function isConstructorDeclaredProperty(symbol) {
                // A property is considered a constructor declared property when all declaration sites are this.xxx assignments,
                // when no declaration sites have JSDoc type annotations, and when at least one declaration site is in the body of
                // a class constructor.
                if (symbol.valueDeclaration && ts.isBinaryExpression(symbol.valueDeclaration)) {
                    var links = getSymbolLinks(symbol);
                    if (links.isConstructorDeclaredProperty === undefined) {
                        links.isConstructorDeclaredProperty = false;
                        links.isConstructorDeclaredProperty = !!getDeclaringConstructor(symbol) && ts.every(symbol.declarations, function (declaration) {
                            return ts.isBinaryExpression(declaration) &&
                                isPossiblyAliasedThisProperty(declaration) &&
                                (declaration.left.kind !== 205 /* ElementAccessExpression */ || ts.isStringOrNumericLiteralLike(declaration.left.argumentExpression)) &&
                                !getAnnotatedTypeForAssignmentDeclaration(/*declaredType*/ undefined, declaration, symbol, declaration);
                        });
                    }
                    return links.isConstructorDeclaredProperty;
                }
                return false;
            }
            function isAutoTypedProperty(symbol) {
                // A property is auto-typed when its declaration has no type annotation or initializer and we're in
                // noImplicitAny mode or a .js file.
                var declaration = symbol.valueDeclaration;
                return declaration && ts.isPropertyDeclaration(declaration) && !ts.getEffectiveTypeAnnotationNode(declaration) &&
                    !declaration.initializer && (noImplicitAny || ts.isInJSFile(declaration));
            }
            function getDeclaringConstructor(symbol) {
                if (!symbol.declarations) {
                    return;
                }
                for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                    var declaration = _a[_i];
                    var container = ts.getThisContainer(declaration, /*includeArrowFunctions*/ false);
                    if (container && (container.kind === 169 /* Constructor */ || isJSConstructor(container))) {
                        return container;
                    }
                }
                ;
            }
            /** Create a synthetic property access flow node after the last statement of the file */
            function getFlowTypeFromCommonJSExport(symbol) {
                var file = ts.getSourceFileOfNode(symbol.declarations[0]);
                var accessName = ts.unescapeLeadingUnderscores(symbol.escapedName);
                var areAllModuleExports = symbol.declarations.every(function (d) { return ts.isInJSFile(d) && ts.isAccessExpression(d) && ts.isModuleExportsAccessExpression(d.expression); });
                var reference = areAllModuleExports
                    ? ts.factory.createPropertyAccessExpression(ts.factory.createPropertyAccessExpression(ts.factory.createIdentifier("module"), ts.factory.createIdentifier("exports")), accessName)
                    : ts.factory.createPropertyAccessExpression(ts.factory.createIdentifier("exports"), accessName);
                if (areAllModuleExports) {
                    ts.setParent(reference.expression.expression, reference.expression);
                }
                ts.setParent(reference.expression, reference);
                ts.setParent(reference, file);
                reference.flowNode = file.endFlowNode;
                return getFlowTypeOfReference(reference, autoType, undefinedType);
            }
            function getFlowTypeInStaticBlocks(symbol, staticBlocks) {
                var accessName = ts.startsWith(symbol.escapedName, "__#")
                    ? ts.factory.createPrivateIdentifier(symbol.escapedName.split("@")[1])
                    : ts.unescapeLeadingUnderscores(symbol.escapedName);
                for (var _i = 0, staticBlocks_1 = staticBlocks; _i < staticBlocks_1.length; _i++) {
                    var staticBlock = staticBlocks_1[_i];
                    var reference = ts.factory.createPropertyAccessExpression(ts.factory.createThis(), accessName);
                    ts.setParent(reference.expression, reference);
                    ts.setParent(reference, staticBlock);
                    reference.flowNode = staticBlock.returnFlowNode;
                    var flowType = getFlowTypeOfProperty(reference, symbol);
                    if (noImplicitAny && (flowType === autoType || flowType === autoArrayType)) {
                        error(symbol.valueDeclaration, ts.Diagnostics.Member_0_implicitly_has_an_1_type, symbolToString(symbol), typeToString(flowType));
                    }
                    // We don't infer a type if assignments are only null or undefined.
                    if (everyType(flowType, isNullableType)) {
                        continue;
                    }
                    return convertAutoToAny(flowType);
                }
            }
            function getFlowTypeInConstructor(symbol, constructor) {
                var accessName = ts.startsWith(symbol.escapedName, "__#")
                    ? ts.factory.createPrivateIdentifier(symbol.escapedName.split("@")[1])
                    : ts.unescapeLeadingUnderscores(symbol.escapedName);
                var reference = ts.factory.createPropertyAccessExpression(ts.factory.createThis(), accessName);
                ts.setParent(reference.expression, reference);
                ts.setParent(reference, constructor);
                reference.flowNode = constructor.returnFlowNode;
                var flowType = getFlowTypeOfProperty(reference, symbol);
                if (noImplicitAny && (flowType === autoType || flowType === autoArrayType)) {
                    error(symbol.valueDeclaration, ts.Diagnostics.Member_0_implicitly_has_an_1_type, symbolToString(symbol), typeToString(flowType));
                }
                // We don't infer a type if assignments are only null or undefined.
                return everyType(flowType, isNullableType) ? undefined : convertAutoToAny(flowType);
            }
            function getFlowTypeOfProperty(reference, prop) {
                var initialType = (prop === null || prop === void 0 ? void 0 : prop.valueDeclaration)
                    && (!isAutoTypedProperty(prop) || ts.getEffectiveModifierFlags(prop.valueDeclaration) & 2 /* Ambient */)
                    && getTypeOfPropertyInBaseClass(prop)
                    || undefinedType;
                return getFlowTypeOfReference(reference, autoType, initialType);
            }
            function getWidenedTypeForAssignmentDeclaration(symbol, resolvedSymbol) {
                // function/class/{} initializers are themselves containers, so they won't merge in the same way as other initializers
                var container = ts.getAssignedExpandoInitializer(symbol.valueDeclaration);
                if (container) {
                    var tag = ts.getJSDocTypeTag(container);
                    if (tag && tag.typeExpression) {
                        return getTypeFromTypeNode(tag.typeExpression);
                    }
                    var containerObjectType = symbol.valueDeclaration && getJSContainerObjectType(symbol.valueDeclaration, symbol, container);
                    return containerObjectType || getWidenedLiteralType(checkExpressionCached(container));
                }
                var type;
                var definedInConstructor = false;
                var definedInMethod = false;
                // We use control flow analysis to determine the type of the property if the property qualifies as a constructor
                // declared property and the resulting control flow type isn't just undefined or null.
                if (isConstructorDeclaredProperty(symbol)) {
                    type = getFlowTypeInConstructor(symbol, getDeclaringConstructor(symbol));
                }
                if (!type) {
                    var types = void 0;
                    if (symbol.declarations) {
                        var jsdocType = void 0;
                        for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                            var declaration = _a[_i];
                            var expression = (ts.isBinaryExpression(declaration) || ts.isCallExpression(declaration)) ? declaration :
                                ts.isAccessExpression(declaration) ? ts.isBinaryExpression(declaration.parent) ? declaration.parent : declaration :
                                    undefined;
                            if (!expression) {
                                continue; // Non-assignment declaration merged in (eg, an Identifier to mark the thing as a namespace) - skip over it and pull type info from elsewhere
                            }
                            var kind = ts.isAccessExpression(expression)
                                ? ts.getAssignmentDeclarationPropertyAccessKind(expression)
                                : ts.getAssignmentDeclarationKind(expression);
                            if (kind === 4 /* ThisProperty */ || ts.isBinaryExpression(expression) && isPossiblyAliasedThisProperty(expression, kind)) {
                                if (isDeclarationInConstructor(expression)) {
                                    definedInConstructor = true;
                                }
                                else {
                                    definedInMethod = true;
                                }
                            }
                            if (!ts.isCallExpression(expression)) {
                                jsdocType = getAnnotatedTypeForAssignmentDeclaration(jsdocType, expression, symbol, declaration);
                            }
                            if (!jsdocType) {
                                (types || (types = [])).push((ts.isBinaryExpression(expression) || ts.isCallExpression(expression)) ? getInitializerTypeFromAssignmentDeclaration(symbol, resolvedSymbol, expression, kind) : neverType);
                            }
                        }
                        type = jsdocType;
                    }
                    if (!type) {
                        if (!ts.length(types)) {
                            return errorType; // No types from any declarations :(
                        }
                        var constructorTypes = definedInConstructor && symbol.declarations ? getConstructorDefinedThisAssignmentTypes(types, symbol.declarations) : undefined;
                        // use only the constructor types unless they were only assigned null | undefined (including widening variants)
                        if (definedInMethod) {
                            var propType = getTypeOfPropertyInBaseClass(symbol);
                            if (propType) {
                                (constructorTypes || (constructorTypes = [])).push(propType);
                                definedInConstructor = true;
                            }
                        }
                        var sourceTypes = ts.some(constructorTypes, function (t) { return !!(t.flags & ~98304 /* Nullable */); }) ? constructorTypes : types; // TODO: GH#18217
                        type = getUnionType(sourceTypes, 2 /* Subtype */);
                    }
                }
                var widened = getWidenedType(addOptionality(type, /*isProperty*/ false, definedInMethod && !definedInConstructor));
                if (symbol.valueDeclaration && filterType(widened, function (t) { return !!(t.flags & ~98304 /* Nullable */); }) === neverType) {
                    reportImplicitAny(symbol.valueDeclaration, anyType);
                    return anyType;
                }
                return widened;
            }
            function getJSContainerObjectType(decl, symbol, init) {
                var _a, _b;
                if (!ts.isInJSFile(decl) || !init || !ts.isObjectLiteralExpression(init) || init.properties.length) {
                    return undefined;
                }
                var exports = ts.createSymbolTable();
                while (ts.isBinaryExpression(decl) || ts.isPropertyAccessExpression(decl)) {
                    var s_2 = getSymbolOfNode(decl);
                    if ((_a = s_2 === null || s_2 === void 0 ? void 0 : s_2.exports) === null || _a === void 0 ? void 0 : _a.size) {
                        mergeSymbolTable(exports, s_2.exports);
                    }
                    decl = ts.isBinaryExpression(decl) ? decl.parent : decl.parent.parent;
                }
                var s = getSymbolOfNode(decl);
                if ((_b = s === null || s === void 0 ? void 0 : s.exports) === null || _b === void 0 ? void 0 : _b.size) {
                    mergeSymbolTable(exports, s.exports);
                }
                var type = createAnonymousType(symbol, exports, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                type.objectFlags |= 8192 /* JSLiteral */;
                return type;
            }
            function getAnnotatedTypeForAssignmentDeclaration(declaredType, expression, symbol, declaration) {
                var _a;
                var typeNode = ts.getEffectiveTypeAnnotationNode(expression.parent);
                if (typeNode) {
                    var type = getWidenedType(getTypeFromTypeNode(typeNode));
                    if (!declaredType) {
                        return type;
                    }
                    else if (declaredType !== errorType && type !== errorType && !isTypeIdenticalTo(declaredType, type)) {
                        errorNextVariableOrPropertyDeclarationMustHaveSameType(/*firstDeclaration*/ undefined, declaredType, declaration, type);
                    }
                }
                if ((_a = symbol.parent) === null || _a === void 0 ? void 0 : _a.valueDeclaration) {
                    var typeNode_2 = ts.getEffectiveTypeAnnotationNode(symbol.parent.valueDeclaration);
                    if (typeNode_2) {
                        var annotationSymbol = getPropertyOfType(getTypeFromTypeNode(typeNode_2), symbol.escapedName);
                        if (annotationSymbol) {
                            return getNonMissingTypeOfSymbol(annotationSymbol);
                        }
                    }
                }
                return declaredType;
            }
            /** If we don't have an explicit JSDoc type, get the type from the initializer. */
            function getInitializerTypeFromAssignmentDeclaration(symbol, resolvedSymbol, expression, kind) {
                if (ts.isCallExpression(expression)) {
                    if (resolvedSymbol) {
                        return getTypeOfSymbol(resolvedSymbol); // This shouldn't happen except under some hopefully forbidden merges of export assignments and object define assignments
                    }
                    var objectLitType = checkExpressionCached(expression.arguments[2]);
                    var valueType = getTypeOfPropertyOfType(objectLitType, "value");
                    if (valueType) {
                        return valueType;
                    }
                    var getFunc = getTypeOfPropertyOfType(objectLitType, "get");
                    if (getFunc) {
                        var getSig = getSingleCallSignature(getFunc);
                        if (getSig) {
                            return getReturnTypeOfSignature(getSig);
                        }
                    }
                    var setFunc = getTypeOfPropertyOfType(objectLitType, "set");
                    if (setFunc) {
                        var setSig = getSingleCallSignature(setFunc);
                        if (setSig) {
                            return getTypeOfFirstParameterOfSignature(setSig);
                        }
                    }
                    return anyType;
                }
                if (containsSameNamedThisProperty(expression.left, expression.right)) {
                    return anyType;
                }
                var type = resolvedSymbol ? getTypeOfSymbol(resolvedSymbol) : getWidenedLiteralType(checkExpressionCached(expression.right));
                if (type.flags & 524288 /* Object */ &&
                    kind === 2 /* ModuleExports */ &&
                    symbol.escapedName === "export=" /* ExportEquals */) {
                    var exportedType = resolveStructuredTypeMembers(type);
                    var members_4 = ts.createSymbolTable();
                    ts.copyEntries(exportedType.members, members_4);
                    var initialSize = members_4.size;
                    if (resolvedSymbol && !resolvedSymbol.exports) {
                        resolvedSymbol.exports = ts.createSymbolTable();
                    }
                    (resolvedSymbol || symbol).exports.forEach(function (s, name) {
                        var _a;
                        var exportedMember = members_4.get(name);
                        if (exportedMember && exportedMember !== s) {
                            if (s.flags & 111551 /* Value */ && exportedMember.flags & 111551 /* Value */) {
                                // If the member has an additional value-like declaration, union the types from the two declarations,
                                // but issue an error if they occurred in two different files. The purpose is to support a JS file with
                                // a pattern like:
                                //
                                // module.exports = { a: true };
                                // module.exports.a = 3;
                                //
                                // but we may have a JS file with `module.exports = { a: true }` along with a TypeScript module augmentation
                                // declaring an `export const a: number`. In that case, we issue a duplicate identifier error, because
                                // it's unclear what that's supposed to mean, so it's probably a mistake.
                                if (s.valueDeclaration && exportedMember.valueDeclaration && ts.getSourceFileOfNode(s.valueDeclaration) !== ts.getSourceFileOfNode(exportedMember.valueDeclaration)) {
                                    var unescapedName = ts.unescapeLeadingUnderscores(s.escapedName);
                                    var exportedMemberName = ((_a = ts.tryCast(exportedMember.valueDeclaration, ts.isNamedDeclaration)) === null || _a === void 0 ? void 0 : _a.name) || exportedMember.valueDeclaration;
                                    ts.addRelatedInfo(error(s.valueDeclaration, ts.Diagnostics.Duplicate_identifier_0, unescapedName), ts.createDiagnosticForNode(exportedMemberName, ts.Diagnostics._0_was_also_declared_here, unescapedName));
                                    ts.addRelatedInfo(error(exportedMemberName, ts.Diagnostics.Duplicate_identifier_0, unescapedName), ts.createDiagnosticForNode(s.valueDeclaration, ts.Diagnostics._0_was_also_declared_here, unescapedName));
                                }
                                var union = createSymbol(s.flags | exportedMember.flags, name);
                                union.type = getUnionType([getTypeOfSymbol(s), getTypeOfSymbol(exportedMember)]);
                                union.valueDeclaration = exportedMember.valueDeclaration;
                                union.declarations = ts.concatenate(exportedMember.declarations, s.declarations);
                                members_4.set(name, union);
                            }
                            else {
                                members_4.set(name, mergeSymbol(s, exportedMember));
                            }
                        }
                        else {
                            members_4.set(name, s);
                        }
                    });
                    var result = createAnonymousType(initialSize !== members_4.size ? undefined : exportedType.symbol, // Only set the type's symbol if it looks to be the same as the original type
                    members_4, exportedType.callSignatures, exportedType.constructSignatures, exportedType.indexInfos);
                    result.objectFlags |= (ts.getObjectFlags(type) & 8192 /* JSLiteral */); // Propagate JSLiteral flag
                    if (result.symbol && result.symbol.flags & 32 /* Class */ && type === getDeclaredTypeOfClassOrInterface(result.symbol)) {
                        result.objectFlags |= 16777216 /* IsClassInstanceClone */; // Propagate the knowledge that this type is equivalent to the symbol's class instance type
                    }
                    return result;
                }
                if (isEmptyArrayLiteralType(type)) {
                    reportImplicitAny(expression, anyArrayType);
                    return anyArrayType;
                }
                return type;
            }
            function containsSameNamedThisProperty(thisProperty, expression) {
                return ts.isPropertyAccessExpression(thisProperty)
                    && thisProperty.expression.kind === 108 /* ThisKeyword */
                    && ts.forEachChildRecursively(expression, function (n) { return isMatchingReference(thisProperty, n); });
            }
            function isDeclarationInConstructor(expression) {
                var thisContainer = ts.getThisContainer(expression, /*includeArrowFunctions*/ false);
                // Properties defined in a constructor (or base constructor, or javascript constructor function) don't get undefined added.
                // Function expressions that are assigned to the prototype count as methods.
                return thisContainer.kind === 169 /* Constructor */ ||
                    thisContainer.kind === 254 /* FunctionDeclaration */ ||
                    (thisContainer.kind === 211 /* FunctionExpression */ && !ts.isPrototypePropertyAssignment(thisContainer.parent));
            }
            function getConstructorDefinedThisAssignmentTypes(types, declarations) {
                ts.Debug.assert(types.length === declarations.length);
                return types.filter(function (_, i) {
                    var declaration = declarations[i];
                    var expression = ts.isBinaryExpression(declaration) ? declaration :
                        ts.isBinaryExpression(declaration.parent) ? declaration.parent : undefined;
                    return expression && isDeclarationInConstructor(expression);
                });
            }
            // Return the type implied by a binding pattern element. This is the type of the initializer of the element if
            // one is present. Otherwise, if the element is itself a binding pattern, it is the type implied by the binding
            // pattern. Otherwise, it is the type any.
            function getTypeFromBindingElement(element, includePatternInType, reportErrors) {
                if (element.initializer) {
                    // The type implied by a binding pattern is independent of context, so we check the initializer with no
                    // contextual type or, if the element itself is a binding pattern, with the type implied by that binding
                    // pattern.
                    var contextualType = ts.isBindingPattern(element.name) ? getTypeFromBindingPattern(element.name, /*includePatternInType*/ true, /*reportErrors*/ false) : unknownType;
                    return addOptionality(widenTypeInferredFromInitializer(element, checkDeclarationInitializer(element, contextualType)));
                }
                if (ts.isBindingPattern(element.name)) {
                    return getTypeFromBindingPattern(element.name, includePatternInType, reportErrors);
                }
                if (reportErrors && !declarationBelongsToPrivateAmbientMember(element)) {
                    reportImplicitAny(element, anyType);
                }
                // When we're including the pattern in the type (an indication we're obtaining a contextual type), we
                // use the non-inferrable any type. Inference will never directly infer this type, but it is possible
                // to infer a type that contains it, e.g. for a binding pattern like [foo] or { foo }. In such cases,
                // widening of the binding pattern type substitutes a regular any for the non-inferrable any.
                return includePatternInType ? nonInferrableAnyType : anyType;
            }
            // Return the type implied by an object binding pattern
            function getTypeFromObjectBindingPattern(pattern, includePatternInType, reportErrors) {
                var members = ts.createSymbolTable();
                var stringIndexInfo;
                var objectFlags = 128 /* ObjectLiteral */ | 262144 /* ContainsObjectOrArrayLiteral */;
                ts.forEach(pattern.elements, function (e) {
                    var name = e.propertyName || e.name;
                    if (e.dotDotDotToken) {
                        stringIndexInfo = createIndexInfo(stringType, anyType, /*isReadonly*/ false);
                        return;
                    }
                    var exprType = getLiteralTypeFromPropertyName(name);
                    if (!isTypeUsableAsPropertyName(exprType)) {
                        // do not include computed properties in the implied type
                        objectFlags |= 512 /* ObjectLiteralPatternWithComputedProperties */;
                        return;
                    }
                    var text = getPropertyNameFromType(exprType);
                    var flags = 4 /* Property */ | (e.initializer ? 16777216 /* Optional */ : 0);
                    var symbol = createSymbol(flags, text);
                    symbol.type = getTypeFromBindingElement(e, includePatternInType, reportErrors);
                    symbol.bindingElement = e;
                    members.set(symbol.escapedName, symbol);
                });
                var result = createAnonymousType(undefined, members, ts.emptyArray, ts.emptyArray, stringIndexInfo ? [stringIndexInfo] : ts.emptyArray);
                result.objectFlags |= objectFlags;
                if (includePatternInType) {
                    result.pattern = pattern;
                    result.objectFlags |= 262144 /* ContainsObjectOrArrayLiteral */;
                }
                return result;
            }
            // Return the type implied by an array binding pattern
            function getTypeFromArrayBindingPattern(pattern, includePatternInType, reportErrors) {
                var elements = pattern.elements;
                var lastElement = ts.lastOrUndefined(elements);
                var restElement = lastElement && lastElement.kind === 201 /* BindingElement */ && lastElement.dotDotDotToken ? lastElement : undefined;
                if (elements.length === 0 || elements.length === 1 && restElement) {
                    return languageVersion >= 2 /* ES2015 */ ? createIterableType(anyType) : anyArrayType;
                }
                var elementTypes = ts.map(elements, function (e) { return ts.isOmittedExpression(e) ? anyType : getTypeFromBindingElement(e, includePatternInType, reportErrors); });
                var minLength = ts.findLastIndex(elements, function (e) { return !(e === restElement || ts.isOmittedExpression(e) || hasDefaultValue(e)); }, elements.length - 1) + 1;
                var elementFlags = ts.map(elements, function (e, i) { return e === restElement ? 4 /* Rest */ : i >= minLength ? 2 /* Optional */ : 1 /* Required */; });
                var result = createTupleType(elementTypes, elementFlags);
                if (includePatternInType) {
                    result = cloneTypeReference(result);
                    result.pattern = pattern;
                    result.objectFlags |= 262144 /* ContainsObjectOrArrayLiteral */;
                }
                return result;
            }
            // Return the type implied by a binding pattern. This is the type implied purely by the binding pattern itself
            // and without regard to its context (i.e. without regard any type annotation or initializer associated with the
            // declaration in which the binding pattern is contained). For example, the implied type of [x, y] is [any, any]
            // and the implied type of { x, y: z = 1 } is { x: any; y: number; }. The type implied by a binding pattern is
            // used as the contextual type of an initializer associated with the binding pattern. Also, for a destructuring
            // parameter with no type annotation or initializer, the type implied by the binding pattern becomes the type of
            // the parameter.
            function getTypeFromBindingPattern(pattern, includePatternInType, reportErrors) {
                if (includePatternInType === void 0) {
                    includePatternInType = false;
                }
                if (reportErrors === void 0) {
                    reportErrors = false;
                }
                return pattern.kind === 199 /* ObjectBindingPattern */
                    ? getTypeFromObjectBindingPattern(pattern, includePatternInType, reportErrors)
                    : getTypeFromArrayBindingPattern(pattern, includePatternInType, reportErrors);
            }
            // Return the type associated with a variable, parameter, or property declaration. In the simple case this is the type
            // specified in a type annotation or inferred from an initializer. However, in the case of a destructuring declaration it
            // is a bit more involved. For example:
            //
            //   var [x, s = ""] = [1, "one"];
            //
            // Here, the array literal [1, "one"] is contextually typed by the type [any, string], which is the implied type of the
            // binding pattern [x, s = ""]. Because the contextual type is a tuple type, the resulting type of [1, "one"] is the
            // tuple type [number, string]. Thus, the type inferred for 'x' is number and the type inferred for 's' is string.
            function getWidenedTypeForVariableLikeDeclaration(declaration, reportErrors) {
                return widenTypeForVariableLikeDeclaration(getTypeForVariableLikeDeclaration(declaration, /*includeOptionality*/ true), declaration, reportErrors);
            }
            function isGlobalSymbolConstructor(node) {
                var symbol = getSymbolOfNode(node);
                var globalSymbol = getGlobalESSymbolConstructorTypeSymbol(/*reportErrors*/ false);
                return globalSymbol && symbol && symbol === globalSymbol;
            }
            function widenTypeForVariableLikeDeclaration(type, declaration, reportErrors) {
                if (type) {
                    // TODO: If back compat with pre-3.0/4.0 libs isn't required, remove the following SymbolConstructor special case transforming `symbol` into `unique symbol`
                    if (type.flags & 4096 /* ESSymbol */ && isGlobalSymbolConstructor(declaration.parent)) {
                        type = getESSymbolLikeTypeForNode(declaration);
                    }
                    if (reportErrors) {
                        reportErrorsFromWidening(declaration, type);
                    }
                    // always widen a 'unique symbol' type if the type was created for a different declaration.
                    if (type.flags & 8192 /* UniqueESSymbol */ && (ts.isBindingElement(declaration) || !declaration.type) && type.symbol !== getSymbolOfNode(declaration)) {
                        type = esSymbolType;
                    }
                    return getWidenedType(type);
                }
                // Rest parameters default to type any[], other parameters default to type any
                type = ts.isParameter(declaration) && declaration.dotDotDotToken ? anyArrayType : anyType;
                // Report implicit any errors unless this is a private property within an ambient declaration
                if (reportErrors) {
                    if (!declarationBelongsToPrivateAmbientMember(declaration)) {
                        reportImplicitAny(declaration, type);
                    }
                }
                return type;
            }
            function declarationBelongsToPrivateAmbientMember(declaration) {
                var root = ts.getRootDeclaration(declaration);
                var memberDeclaration = root.kind === 162 /* Parameter */ ? root.parent : root;
                return isPrivateWithinAmbient(memberDeclaration);
            }
            function tryGetTypeFromEffectiveTypeNode(declaration) {
                var typeNode = ts.getEffectiveTypeAnnotationNode(declaration);
                if (typeNode) {
                    return getTypeFromTypeNode(typeNode);
                }
            }
            function getTypeOfVariableOrParameterOrProperty(symbol) {
                var links = getSymbolLinks(symbol);
                if (!links.type) {
                    var type = getTypeOfVariableOrParameterOrPropertyWorker(symbol);
                    // For a contextually typed parameter it is possible that a type has already
                    // been assigned (in assignTypeToParameterAndFixTypeParameters), and we want
                    // to preserve this type.
                    if (!links.type) {
                        links.type = type;
                    }
                }
                return links.type;
            }
            function getTypeOfVariableOrParameterOrPropertyWorker(symbol) {
                // Handle prototype property
                if (symbol.flags & 4194304 /* Prototype */) {
                    return getTypeOfPrototypeProperty(symbol);
                }
                // CommonsJS require and module both have type any.
                if (symbol === requireSymbol) {
                    return anyType;
                }
                if (symbol.flags & 134217728 /* ModuleExports */ && symbol.valueDeclaration) {
                    var fileSymbol = getSymbolOfNode(ts.getSourceFileOfNode(symbol.valueDeclaration));
                    var result = createSymbol(fileSymbol.flags, "exports");
                    result.declarations = fileSymbol.declarations ? fileSymbol.declarations.slice() : [];
                    result.parent = symbol;
                    result.target = fileSymbol;
                    if (fileSymbol.valueDeclaration)
                        result.valueDeclaration = fileSymbol.valueDeclaration;
                    if (fileSymbol.members)
                        result.members = new ts.Map(fileSymbol.members);
                    if (fileSymbol.exports)
                        result.exports = new ts.Map(fileSymbol.exports);
                    var members = ts.createSymbolTable();
                    members.set("exports", result);
                    return createAnonymousType(symbol, members, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                }
                // Handle catch clause variables
                ts.Debug.assertIsDefined(symbol.valueDeclaration);
                var declaration = symbol.valueDeclaration;
                if (ts.isCatchClauseVariableDeclarationOrBindingElement(declaration)) {
                    var typeNode = ts.getEffectiveTypeAnnotationNode(declaration);
                    if (typeNode === undefined) {
                        return useUnknownInCatchVariables ? unknownType : anyType;
                    }
                    var type_1 = getTypeOfNode(typeNode);
                    // an errorType will make `checkTryStatement` issue an error
                    return isTypeAny(type_1) || type_1 === unknownType ? type_1 : errorType;
                }
                // Handle export default expressions
                if (ts.isSourceFile(declaration) && ts.isJsonSourceFile(declaration)) {
                    if (!declaration.statements.length) {
                        return emptyObjectType;
                    }
                    return getWidenedType(getWidenedLiteralType(checkExpression(declaration.statements[0].expression)));
                }
                // Handle variable, parameter or property
                if (!pushTypeResolution(symbol, 0 /* Type */)) {
                    // Symbol is property of some kind that is merged with something - should use `getTypeOfFuncClassEnumModule` and not `getTypeOfVariableOrParameterOrProperty`
                    if (symbol.flags & 512 /* ValueModule */ && !(symbol.flags & 67108864 /* Assignment */)) {
                        return getTypeOfFuncClassEnumModule(symbol);
                    }
                    return reportCircularityError(symbol);
                }
                var type;
                if (declaration.kind === 269 /* ExportAssignment */) {
                    type = widenTypeForVariableLikeDeclaration(checkExpressionCached(declaration.expression), declaration);
                }
                else if (ts.isBinaryExpression(declaration) ||
                    (ts.isInJSFile(declaration) &&
                        (ts.isCallExpression(declaration) || (ts.isPropertyAccessExpression(declaration) || ts.isBindableStaticElementAccessExpression(declaration)) && ts.isBinaryExpression(declaration.parent)))) {
                    type = getWidenedTypeForAssignmentDeclaration(symbol);
                }
                else if (ts.isPropertyAccessExpression(declaration)
                    || ts.isElementAccessExpression(declaration)
                    || ts.isIdentifier(declaration)
                    || ts.isStringLiteralLike(declaration)
                    || ts.isNumericLiteral(declaration)
                    || ts.isClassDeclaration(declaration)
                    || ts.isFunctionDeclaration(declaration)
                    || (ts.isMethodDeclaration(declaration) && !ts.isObjectLiteralMethod(declaration))
                    || ts.isMethodSignature(declaration)
                    || ts.isSourceFile(declaration)) {
                    // Symbol is property of some kind that is merged with something - should use `getTypeOfFuncClassEnumModule` and not `getTypeOfVariableOrParameterOrProperty`
                    if (symbol.flags & (16 /* Function */ | 8192 /* Method */ | 32 /* Class */ | 384 /* Enum */ | 512 /* ValueModule */)) {
                        return getTypeOfFuncClassEnumModule(symbol);
                    }
                    type = ts.isBinaryExpression(declaration.parent) ?
                        getWidenedTypeForAssignmentDeclaration(symbol) :
                        tryGetTypeFromEffectiveTypeNode(declaration) || anyType;
                }
                else if (ts.isPropertyAssignment(declaration)) {
                    type = tryGetTypeFromEffectiveTypeNode(declaration) || checkPropertyAssignment(declaration);
                }
                else if (ts.isJsxAttribute(declaration)) {
                    type = tryGetTypeFromEffectiveTypeNode(declaration) || checkJsxAttribute(declaration);
                }
                else if (ts.isShorthandPropertyAssignment(declaration)) {
                    type = tryGetTypeFromEffectiveTypeNode(declaration) || checkExpressionForMutableLocation(declaration.name, 0 /* Normal */);
                }
                else if (ts.isObjectLiteralMethod(declaration)) {
                    type = tryGetTypeFromEffectiveTypeNode(declaration) || checkObjectLiteralMethod(declaration, 0 /* Normal */);
                }
                else if (ts.isParameter(declaration)
                    || ts.isPropertyDeclaration(declaration)
                    || ts.isPropertySignature(declaration)
                    || ts.isVariableDeclaration(declaration)
                    || ts.isBindingElement(declaration)
                    || ts.isJSDocPropertyLikeTag(declaration)) {
                    type = getWidenedTypeForVariableLikeDeclaration(declaration, /*includeOptionality*/ true);
                }
                // getTypeOfSymbol dispatches some JS merges incorrectly because their symbol flags are not mutually exclusive.
                // Re-dispatch based on valueDeclaration.kind instead.
                else if (ts.isEnumDeclaration(declaration)) {
                    type = getTypeOfFuncClassEnumModule(symbol);
                }
                else if (ts.isEnumMember(declaration)) {
                    type = getTypeOfEnumMember(symbol);
                }
                else if (ts.isAccessor(declaration)) {
                    type = resolveTypeOfAccessors(symbol) || ts.Debug.fail("Non-write accessor resolution must always produce a type");
                }
                else {
                    return ts.Debug.fail("Unhandled declaration kind! " + ts.Debug.formatSyntaxKind(declaration.kind) + " for " + ts.Debug.formatSymbol(symbol));
                }
                if (!popTypeResolution()) {
                    // Symbol is property of some kind that is merged with something - should use `getTypeOfFuncClassEnumModule` and not `getTypeOfVariableOrParameterOrProperty`
                    if (symbol.flags & 512 /* ValueModule */ && !(symbol.flags & 67108864 /* Assignment */)) {
                        return getTypeOfFuncClassEnumModule(symbol);
                    }
                    return reportCircularityError(symbol);
                }
                return type;
            }
            function getAnnotatedAccessorTypeNode(accessor) {
                if (accessor) {
                    if (accessor.kind === 170 /* GetAccessor */) {
                        var getterTypeAnnotation = ts.getEffectiveReturnTypeNode(accessor);
                        return getterTypeAnnotation;
                    }
                    else {
                        var setterTypeAnnotation = ts.getEffectiveSetAccessorTypeAnnotationNode(accessor);
                        return setterTypeAnnotation;
                    }
                }
                return undefined;
            }
            function getAnnotatedAccessorType(accessor) {
                var node = getAnnotatedAccessorTypeNode(accessor);
                return node && getTypeFromTypeNode(node);
            }
            function getAnnotatedAccessorThisParameter(accessor) {
                var parameter = getAccessorThisParameter(accessor);
                return parameter && parameter.symbol;
            }
            function getThisTypeOfDeclaration(declaration) {
                return getThisTypeOfSignature(getSignatureFromDeclaration(declaration));
            }
            function getTypeOfAccessors(symbol) {
                var links = getSymbolLinks(symbol);
                return links.type || (links.type = getTypeOfAccessorsWorker(symbol) || ts.Debug.fail("Read type of accessor must always produce a type"));
            }
            function getTypeOfSetAccessor(symbol) {
                var links = getSymbolLinks(symbol);
                return links.writeType || (links.writeType = getTypeOfAccessorsWorker(symbol, /*writing*/ true));
            }
            function getTypeOfAccessorsWorker(symbol, writing) {
                if (writing === void 0) {
                    writing = false;
                }
                if (!pushTypeResolution(symbol, 0 /* Type */)) {
                    return errorType;
                }
                var type = resolveTypeOfAccessors(symbol, writing);
                if (!popTypeResolution()) {
                    type = anyType;
                    if (noImplicitAny) {
                        var getter = ts.getDeclarationOfKind(symbol, 170 /* GetAccessor */);
                        error(getter, ts.Diagnostics._0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions, symbolToString(symbol));
                    }
                }
                return type;
            }
            function resolveTypeOfAccessors(symbol, writing) {
                if (writing === void 0) {
                    writing = false;
                }
                var getter = ts.getDeclarationOfKind(symbol, 170 /* GetAccessor */);
                var setter = ts.getDeclarationOfKind(symbol, 171 /* SetAccessor */);
                var setterType = getAnnotatedAccessorType(setter);
                // For write operations, prioritize type annotations on the setter
                if (writing && setterType) {
                    return instantiateTypeIfNeeded(setterType, symbol);
                }
                // Else defer to the getter type
                if (getter && ts.isInJSFile(getter)) {
                    var jsDocType = getTypeForDeclarationFromJSDocComment(getter);
                    if (jsDocType) {
                        return instantiateTypeIfNeeded(jsDocType, symbol);
                    }
                }
                // Try to see if the user specified a return type on the get-accessor.
                var getterType = getAnnotatedAccessorType(getter);
                if (getterType) {
                    return instantiateTypeIfNeeded(getterType, symbol);
                }
                // If the user didn't specify a return type, try to use the set-accessor's parameter type.
                if (setterType) {
                    return setterType;
                }
                // If there are no specified types, try to infer it from the body of the get accessor if it exists.
                if (getter && getter.body) {
                    var returnTypeFromBody = getReturnTypeFromBody(getter);
                    return instantiateTypeIfNeeded(returnTypeFromBody, symbol);
                }
                // Otherwise, fall back to 'any'.
                if (setter) {
                    if (!isPrivateWithinAmbient(setter)) {
                        errorOrSuggestion(noImplicitAny, setter, ts.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation, symbolToString(symbol));
                    }
                    return anyType;
                }
                else if (getter) {
                    ts.Debug.assert(!!getter, "there must exist a getter as we are current checking either setter or getter in this function");
                    if (!isPrivateWithinAmbient(getter)) {
                        errorOrSuggestion(noImplicitAny, getter, ts.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation, symbolToString(symbol));
                    }
                    return anyType;
                }
                return undefined;
                function instantiateTypeIfNeeded(type, symbol) {
                    if (ts.getCheckFlags(symbol) & 1 /* Instantiated */) {
                        var links = getSymbolLinks(symbol);
                        return instantiateType(type, links.mapper);
                    }
                    return type;
                }
            }
            function getBaseTypeVariableOfClass(symbol) {
                var baseConstructorType = getBaseConstructorTypeOfClass(getDeclaredTypeOfClassOrInterface(symbol));
                return baseConstructorType.flags & 8650752 /* TypeVariable */ ? baseConstructorType :
                    baseConstructorType.flags & 2097152 /* Intersection */ ? ts.find(baseConstructorType.types, function (t) { return !!(t.flags & 8650752 /* TypeVariable */); }) :
                        undefined;
            }
            function getTypeOfFuncClassEnumModule(symbol) {
                var links = getSymbolLinks(symbol);
                var originalLinks = links;
                if (!links.type) {
                    var expando = symbol.valueDeclaration && getSymbolOfExpando(symbol.valueDeclaration, /*allowDeclaration*/ false);
                    if (expando) {
                        var merged = mergeJSSymbols(symbol, expando);
                        if (merged) {
                            // note:we overwrite links because we just cloned the symbol
                            symbol = links = merged;
                        }
                    }
                    originalLinks.type = links.type = getTypeOfFuncClassEnumModuleWorker(symbol);
                }
                return links.type;
            }
            function getTypeOfFuncClassEnumModuleWorker(symbol) {
                var declaration = symbol.valueDeclaration;
                if (symbol.flags & 1536 /* Module */ && ts.isShorthandAmbientModuleSymbol(symbol)) {
                    return anyType;
                }
                else if (declaration && (declaration.kind === 219 /* BinaryExpression */ ||
                    ts.isAccessExpression(declaration) &&
                        declaration.parent.kind === 219 /* BinaryExpression */)) {
                    return getWidenedTypeForAssignmentDeclaration(symbol);
                }
                else if (symbol.flags & 512 /* ValueModule */ && declaration && ts.isSourceFile(declaration) && declaration.commonJsModuleIndicator) {
                    var resolvedModule = resolveExternalModuleSymbol(symbol);
                    if (resolvedModule !== symbol) {
                        if (!pushTypeResolution(symbol, 0 /* Type */)) {
                            return errorType;
                        }
                        var exportEquals = getMergedSymbol(symbol.exports.get("export=" /* ExportEquals */));
                        var type_2 = getWidenedTypeForAssignmentDeclaration(exportEquals, exportEquals === resolvedModule ? undefined : resolvedModule);
                        if (!popTypeResolution()) {
                            return reportCircularityError(symbol);
                        }
                        return type_2;
                    }
                }
                var type = createObjectType(16 /* Anonymous */, symbol);
                if (symbol.flags & 32 /* Class */) {
                    var baseTypeVariable = getBaseTypeVariableOfClass(symbol);
                    return baseTypeVariable ? getIntersectionType([type, baseTypeVariable]) : type;
                }
                else {
                    return strictNullChecks && symbol.flags & 16777216 /* Optional */ ? getOptionalType(type) : type;
                }
            }
            function getTypeOfEnumMember(symbol) {
                var links = getSymbolLinks(symbol);
                return links.type || (links.type = getDeclaredTypeOfEnumMember(symbol));
            }
            function getTypeOfAlias(symbol) {
                var links = getSymbolLinks(symbol);
                if (!links.type) {
                    var targetSymbol = resolveAlias(symbol);
                    var exportSymbol = symbol.declarations && getTargetOfAliasDeclaration(getDeclarationOfAliasSymbol(symbol), /*dontResolveAlias*/ true);
                    // It only makes sense to get the type of a value symbol. If the result of resolving
                    // the alias is not a value, then it has no type. To get the type associated with a
                    // type symbol, call getDeclaredTypeOfSymbol.
                    // This check is important because without it, a call to getTypeOfSymbol could end
                    // up recursively calling getTypeOfAlias, causing a stack overflow.
                    links.type = (exportSymbol === null || exportSymbol === void 0 ? void 0 : exportSymbol.declarations) && isDuplicatedCommonJSExport(exportSymbol.declarations) && symbol.declarations.length ? getFlowTypeFromCommonJSExport(exportSymbol)
                        : isDuplicatedCommonJSExport(symbol.declarations) ? autoType
                            : targetSymbol.flags & 111551 /* Value */ ? getTypeOfSymbol(targetSymbol)
                                : errorType;
                }
                return links.type;
            }
            function getTypeOfInstantiatedSymbol(symbol) {
                var links = getSymbolLinks(symbol);
                if (!links.type) {
                    if (!pushTypeResolution(symbol, 0 /* Type */)) {
                        return links.type = errorType;
                    }
                    var type = instantiateType(getTypeOfSymbol(links.target), links.mapper);
                    if (!popTypeResolution()) {
                        type = reportCircularityError(symbol);
                    }
                    links.type = type;
                }
                return links.type;
            }
            function reportCircularityError(symbol) {
                var declaration = symbol.valueDeclaration;
                // Check if variable has type annotation that circularly references the variable itself
                if (ts.getEffectiveTypeAnnotationNode(declaration)) {
                    error(symbol.valueDeclaration, ts.Diagnostics._0_is_referenced_directly_or_indirectly_in_its_own_type_annotation, symbolToString(symbol));
                    return errorType;
                }
                // Check if variable has initializer that circularly references the variable itself
                if (noImplicitAny && (declaration.kind !== 162 /* Parameter */ || declaration.initializer)) {
                    error(symbol.valueDeclaration, ts.Diagnostics._0_implicitly_has_type_any_because_it_does_not_have_a_type_annotation_and_is_referenced_directly_or_indirectly_in_its_own_initializer, symbolToString(symbol));
                }
                // Circularities could also result from parameters in function expressions that end up
                // having themselves as contextual types following type argument inference. In those cases
                // we have already reported an implicit any error so we don't report anything here.
                return anyType;
            }
            function getTypeOfSymbolWithDeferredType(symbol) {
                var links = getSymbolLinks(symbol);
                if (!links.type) {
                    ts.Debug.assertIsDefined(links.deferralParent);
                    ts.Debug.assertIsDefined(links.deferralConstituents);
                    links.type = links.deferralParent.flags & 1048576 /* Union */ ? getUnionType(links.deferralConstituents) : getIntersectionType(links.deferralConstituents);
                }
                return links.type;
            }
            function getSetAccessorTypeOfSymbol(symbol) {
                if (symbol.flags & 98304 /* Accessor */) {
                    var type = getTypeOfSetAccessor(symbol);
                    if (type) {
                        return type;
                    }
                }
                return getTypeOfSymbol(symbol);
            }
            function getTypeOfSymbol(symbol) {
                var checkFlags = ts.getCheckFlags(symbol);
                if (checkFlags & 65536 /* DeferredType */) {
                    return getTypeOfSymbolWithDeferredType(symbol);
                }
                if (checkFlags & 1 /* Instantiated */) {
                    return getTypeOfInstantiatedSymbol(symbol);
                }
                if (checkFlags & 262144 /* Mapped */) {
                    return getTypeOfMappedSymbol(symbol);
                }
                if (checkFlags & 8192 /* ReverseMapped */) {
                    return getTypeOfReverseMappedSymbol(symbol);
                }
                if (symbol.flags & (3 /* Variable */ | 4 /* Property */)) {
                    return getTypeOfVariableOrParameterOrProperty(symbol);
                }
                if (symbol.flags & (16 /* Function */ | 8192 /* Method */ | 32 /* Class */ | 384 /* Enum */ | 512 /* ValueModule */)) {
                    return getTypeOfFuncClassEnumModule(symbol);
                }
                if (symbol.flags & 8 /* EnumMember */) {
                    return getTypeOfEnumMember(symbol);
                }
                if (symbol.flags & 98304 /* Accessor */) {
                    return getTypeOfAccessors(symbol);
                }
                if (symbol.flags & 2097152 /* Alias */) {
                    return getTypeOfAlias(symbol);
                }
                return errorType;
            }
            function getNonMissingTypeOfSymbol(symbol) {
                return removeMissingType(getTypeOfSymbol(symbol), !!(symbol.flags & 16777216 /* Optional */));
            }
            function isReferenceToType(type, target) {
                return type !== undefined
                    && target !== undefined
                    && (ts.getObjectFlags(type) & 4 /* Reference */) !== 0
                    && type.target === target;
            }
            function getTargetType(type) {
                return ts.getObjectFlags(type) & 4 /* Reference */ ? type.target : type;
            }
            // TODO: GH#18217 If `checkBase` is undefined, we should not call this because this will always return false.
            function hasBaseType(type, checkBase) {
                return check(type);
                function check(type) {
                    if (ts.getObjectFlags(type) & (3 /* ClassOrInterface */ | 4 /* Reference */)) {
                        var target = getTargetType(type);
                        return target === checkBase || ts.some(getBaseTypes(target), check);
                    }
                    else if (type.flags & 2097152 /* Intersection */) {
                        return ts.some(type.types, check);
                    }
                    return false;
                }
            }
            // Appends the type parameters given by a list of declarations to a set of type parameters and returns the resulting set.
            // The function allocates a new array if the input type parameter set is undefined, but otherwise it modifies the set
            // in-place and returns the same array.
            function appendTypeParameters(typeParameters, declarations) {
                for (var _i = 0, declarations_2 = declarations; _i < declarations_2.length; _i++) {
                    var declaration = declarations_2[_i];
                    typeParameters = ts.appendIfUnique(typeParameters, getDeclaredTypeOfTypeParameter(getSymbolOfNode(declaration)));
                }
                return typeParameters;
            }
            // Return the outer type parameters of a node or undefined if the node has no outer type parameters.
            function getOuterTypeParameters(node, includeThisTypes) {
                while (true) {
                    node = node.parent; // TODO: GH#18217 Use SourceFile kind check instead
                    if (node && ts.isBinaryExpression(node)) {
                        // prototype assignments get the outer type parameters of their constructor function
                        var assignmentKind = ts.getAssignmentDeclarationKind(node);
                        if (assignmentKind === 6 /* Prototype */ || assignmentKind === 3 /* PrototypeProperty */) {
                            var symbol = getSymbolOfNode(node.left);
                            if (symbol && symbol.parent && !ts.findAncestor(symbol.parent.valueDeclaration, function (d) { return node === d; })) {
                                node = symbol.parent.valueDeclaration;
                            }
                        }
                    }
                    if (!node) {
                        return undefined;
                    }
                    switch (node.kind) {
                        case 255 /* ClassDeclaration */:
                        case 224 /* ClassExpression */:
                        case 256 /* InterfaceDeclaration */:
                        case 172 /* CallSignature */:
                        case 173 /* ConstructSignature */:
                        case 166 /* MethodSignature */:
                        case 177 /* FunctionType */:
                        case 178 /* ConstructorType */:
                        case 312 /* JSDocFunctionType */:
                        case 254 /* FunctionDeclaration */:
                        case 167 /* MethodDeclaration */:
                        case 211 /* FunctionExpression */:
                        case 212 /* ArrowFunction */:
                        case 257 /* TypeAliasDeclaration */:
                        case 339 /* JSDocTemplateTag */:
                        case 340 /* JSDocTypedefTag */:
                        case 334 /* JSDocEnumTag */:
                        case 333 /* JSDocCallbackTag */:
                        case 193 /* MappedType */:
                        case 187 /* ConditionalType */: {
                            var outerTypeParameters = getOuterTypeParameters(node, includeThisTypes);
                            if (node.kind === 193 /* MappedType */) {
                                return ts.append(outerTypeParameters, getDeclaredTypeOfTypeParameter(getSymbolOfNode(node.typeParameter)));
                            }
                            else if (node.kind === 187 /* ConditionalType */) {
                                return ts.concatenate(outerTypeParameters, getInferTypeParameters(node));
                            }
                            var outerAndOwnTypeParameters = appendTypeParameters(outerTypeParameters, ts.getEffectiveTypeParameterDeclarations(node));
                            var thisType = includeThisTypes &&
                                (node.kind === 255 /* ClassDeclaration */ || node.kind === 224 /* ClassExpression */ || node.kind === 256 /* InterfaceDeclaration */ || isJSConstructor(node)) &&
                                getDeclaredTypeOfClassOrInterface(getSymbolOfNode(node)).thisType;
                            return thisType ? ts.append(outerAndOwnTypeParameters, thisType) : outerAndOwnTypeParameters;
                        }
                        case 335 /* JSDocParameterTag */:
                            var paramSymbol = ts.getParameterSymbolFromJSDoc(node);
                            if (paramSymbol) {
                                node = paramSymbol.valueDeclaration;
                            }
                            break;
                        case 315 /* JSDocComment */: {
                            var outerTypeParameters = getOuterTypeParameters(node, includeThisTypes);
                            return node.tags
                                ? appendTypeParameters(outerTypeParameters, ts.flatMap(node.tags, function (t) { return ts.isJSDocTemplateTag(t) ? t.typeParameters : undefined; }))
                                : outerTypeParameters;
                        }
                    }
                }
            }
            // The outer type parameters are those defined by enclosing generic classes, methods, or functions.
            function getOuterTypeParametersOfClassOrInterface(symbol) {
                var declaration = symbol.flags & 32 /* Class */ ? symbol.valueDeclaration : ts.getDeclarationOfKind(symbol, 256 /* InterfaceDeclaration */);
                ts.Debug.assert(!!declaration, "Class was missing valueDeclaration -OR- non-class had no interface declarations");
                return getOuterTypeParameters(declaration);
            }
            // The local type parameters are the combined set of type parameters from all declarations of the class,
            // interface, or type alias.
            function getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(symbol) {
                if (!symbol.declarations) {
                    return;
                }
                var result;
                for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                    var node = _a[_i];
                    if (node.kind === 256 /* InterfaceDeclaration */ ||
                        node.kind === 255 /* ClassDeclaration */ ||
                        node.kind === 224 /* ClassExpression */ ||
                        isJSConstructor(node) ||
                        ts.isTypeAlias(node)) {
                        var declaration = node;
                        result = appendTypeParameters(result, ts.getEffectiveTypeParameterDeclarations(declaration));
                    }
                }
                return result;
            }
            // The full set of type parameters for a generic class or interface type consists of its outer type parameters plus
            // its locally declared type parameters.
            function getTypeParametersOfClassOrInterface(symbol) {
                return ts.concatenate(getOuterTypeParametersOfClassOrInterface(symbol), getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(symbol));
            }
            // A type is a mixin constructor if it has a single construct signature taking no type parameters and a single
            // rest parameter of type any[].
            function isMixinConstructorType(type) {
                var signatures = getSignaturesOfType(type, 1 /* Construct */);
                if (signatures.length === 1) {
                    var s = signatures[0];
                    if (!s.typeParameters && s.parameters.length === 1 && signatureHasRestParameter(s)) {
                        var paramType = getTypeOfParameter(s.parameters[0]);
                        return isTypeAny(paramType) || getElementTypeOfArrayType(paramType) === anyType;
                    }
                }
                return false;
            }
            function isConstructorType(type) {
                if (getSignaturesOfType(type, 1 /* Construct */).length > 0) {
                    return true;
                }
                if (type.flags & 8650752 /* TypeVariable */) {
                    var constraint = getBaseConstraintOfType(type);
                    return !!constraint && isMixinConstructorType(constraint);
                }
                return false;
            }
            function getBaseTypeNodeOfClass(type) {
                return ts.getEffectiveBaseTypeNode(type.symbol.valueDeclaration);
            }
            function getConstructorsForTypeArguments(type, typeArgumentNodes, location) {
                var typeArgCount = ts.length(typeArgumentNodes);
                var isJavascript = ts.isInJSFile(location);
                return ts.filter(getSignaturesOfType(type, 1 /* Construct */), function (sig) { return (isJavascript || typeArgCount >= getMinTypeArgumentCount(sig.typeParameters)) && typeArgCount <= ts.length(sig.typeParameters); });
            }
            function getInstantiatedConstructorsForTypeArguments(type, typeArgumentNodes, location) {
                var signatures = getConstructorsForTypeArguments(type, typeArgumentNodes, location);
                var typeArguments = ts.map(typeArgumentNodes, getTypeFromTypeNode);
                return ts.sameMap(signatures, function (sig) { return ts.some(sig.typeParameters) ? getSignatureInstantiation(sig, typeArguments, ts.isInJSFile(location)) : sig; });
            }
            /**
             * The base constructor of a class can resolve to
             * * undefinedType if the class has no extends clause,
             * * unknownType if an error occurred during resolution of the extends expression,
             * * nullType if the extends expression is the null value,
             * * anyType if the extends expression has type any, or
             * * an object type with at least one construct signature.
             */
            function getBaseConstructorTypeOfClass(type) {
                if (!type.resolvedBaseConstructorType) {
                    var decl = type.symbol.valueDeclaration;
                    var extended = ts.getEffectiveBaseTypeNode(decl);
                    var baseTypeNode = getBaseTypeNodeOfClass(type);
                    if (!baseTypeNode) {
                        return type.resolvedBaseConstructorType = undefinedType;
                    }
                    if (!pushTypeResolution(type, 1 /* ResolvedBaseConstructorType */)) {
                        return errorType;
                    }
                    var baseConstructorType = checkExpression(baseTypeNode.expression);
                    if (extended && baseTypeNode !== extended) {
                        ts.Debug.assert(!extended.typeArguments); // Because this is in a JS file, and baseTypeNode is in an @extends tag
                        checkExpression(extended.expression);
                    }
                    if (baseConstructorType.flags & (524288 /* Object */ | 2097152 /* Intersection */)) {
                        // Resolving the members of a class requires us to resolve the base class of that class.
                        // We force resolution here such that we catch circularities now.
                        resolveStructuredTypeMembers(baseConstructorType);
                    }
                    if (!popTypeResolution()) {
                        error(type.symbol.valueDeclaration, ts.Diagnostics._0_is_referenced_directly_or_indirectly_in_its_own_base_expression, symbolToString(type.symbol));
                        return type.resolvedBaseConstructorType = errorType;
                    }
                    if (!(baseConstructorType.flags & 1 /* Any */) && baseConstructorType !== nullWideningType && !isConstructorType(baseConstructorType)) {
                        var err = error(baseTypeNode.expression, ts.Diagnostics.Type_0_is_not_a_constructor_function_type, typeToString(baseConstructorType));
                        if (baseConstructorType.flags & 262144 /* TypeParameter */) {
                            var constraint = getConstraintFromTypeParameter(baseConstructorType);
                            var ctorReturn = unknownType;
                            if (constraint) {
                                var ctorSig = getSignaturesOfType(constraint, 1 /* Construct */);
                                if (ctorSig[0]) {
                                    ctorReturn = getReturnTypeOfSignature(ctorSig[0]);
                                }
                            }
                            if (baseConstructorType.symbol.declarations) {
                                ts.addRelatedInfo(err, ts.createDiagnosticForNode(baseConstructorType.symbol.declarations[0], ts.Diagnostics.Did_you_mean_for_0_to_be_constrained_to_type_new_args_Colon_any_1, symbolToString(baseConstructorType.symbol), typeToString(ctorReturn)));
                            }
                        }
                        return type.resolvedBaseConstructorType = errorType;
                    }
                    type.resolvedBaseConstructorType = baseConstructorType;
                }
                return type.resolvedBaseConstructorType;
            }
            function getImplementsTypes(type) {
                var resolvedImplementsTypes = ts.emptyArray;
                if (type.symbol.declarations) {
                    for (var _i = 0, _a = type.symbol.declarations; _i < _a.length; _i++) {
                        var declaration = _a[_i];
                        var implementsTypeNodes = ts.getEffectiveImplementsTypeNodes(declaration);
                        if (!implementsTypeNodes)
                            continue;
                        for (var _b = 0, implementsTypeNodes_1 = implementsTypeNodes; _b < implementsTypeNodes_1.length; _b++) {
                            var node = implementsTypeNodes_1[_b];
                            var implementsType = getTypeFromTypeNode(node);
                            if (implementsType !== errorType) {
                                if (resolvedImplementsTypes === ts.emptyArray) {
                                    resolvedImplementsTypes = [implementsType];
                                }
                                else {
                                    resolvedImplementsTypes.push(implementsType);
                                }
                            }
                        }
                    }
                }
                return resolvedImplementsTypes;
            }
            function reportCircularBaseType(node, type) {
                error(node, ts.Diagnostics.Type_0_recursively_references_itself_as_a_base_type, typeToString(type, /*enclosingDeclaration*/ undefined, 2 /* WriteArrayAsGenericType */));
            }
            function getBaseTypes(type) {
                if (!type.baseTypesResolved) {
                    if (pushTypeResolution(type, 7 /* ResolvedBaseTypes */)) {
                        if (type.objectFlags & 8 /* Tuple */) {
                            type.resolvedBaseTypes = [getTupleBaseType(type)];
                        }
                        else if (type.symbol.flags & (32 /* Class */ | 64 /* Interface */)) {
                            if (type.symbol.flags & 32 /* Class */) {
                                resolveBaseTypesOfClass(type);
                            }
                            if (type.symbol.flags & 64 /* Interface */) {
                                resolveBaseTypesOfInterface(type);
                            }
                        }
                        else {
                            ts.Debug.fail("type must be class or interface");
                        }
                        if (!popTypeResolution() && type.symbol.declarations) {
                            for (var _i = 0, _a = type.symbol.declarations; _i < _a.length; _i++) {
                                var declaration = _a[_i];
                                if (declaration.kind === 255 /* ClassDeclaration */ || declaration.kind === 256 /* InterfaceDeclaration */) {
                                    reportCircularBaseType(declaration, type);
                                }
                            }
                        }
                    }
                    type.baseTypesResolved = true;
                }
                return type.resolvedBaseTypes;
            }
            function getTupleBaseType(type) {
                var elementTypes = ts.sameMap(type.typeParameters, function (t, i) { return type.elementFlags[i] & 8 /* Variadic */ ? getIndexedAccessType(t, numberType) : t; });
                return createArrayType(getUnionType(elementTypes || ts.emptyArray), type.readonly);
            }
            function resolveBaseTypesOfClass(type) {
                type.resolvedBaseTypes = ts.resolvingEmptyArray;
                var baseConstructorType = getApparentType(getBaseConstructorTypeOfClass(type));
                if (!(baseConstructorType.flags & (524288 /* Object */ | 2097152 /* Intersection */ | 1 /* Any */))) {
                    return type.resolvedBaseTypes = ts.emptyArray;
                }
                var baseTypeNode = getBaseTypeNodeOfClass(type);
                var baseType;
                var originalBaseType = baseConstructorType.symbol ? getDeclaredTypeOfSymbol(baseConstructorType.symbol) : undefined;
                if (baseConstructorType.symbol && baseConstructorType.symbol.flags & 32 /* Class */ &&
                    areAllOuterTypeParametersApplied(originalBaseType)) {
                    // When base constructor type is a class with no captured type arguments we know that the constructors all have the same type parameters as the
                    // class and all return the instance type of the class. There is no need for further checks and we can apply the
                    // type arguments in the same manner as a type reference to get the same error reporting experience.
                    baseType = getTypeFromClassOrInterfaceReference(baseTypeNode, baseConstructorType.symbol);
                }
                else if (baseConstructorType.flags & 1 /* Any */) {
                    baseType = baseConstructorType;
                }
                else {
                    // The class derives from a "class-like" constructor function, check that we have at least one construct signature
                    // with a matching number of type parameters and use the return type of the first instantiated signature. Elsewhere
                    // we check that all instantiated signatures return the same type.
                    var constructors = getInstantiatedConstructorsForTypeArguments(baseConstructorType, baseTypeNode.typeArguments, baseTypeNode);
                    if (!constructors.length) {
                        error(baseTypeNode.expression, ts.Diagnostics.No_base_constructor_has_the_specified_number_of_type_arguments);
                        return type.resolvedBaseTypes = ts.emptyArray;
                    }
                    baseType = getReturnTypeOfSignature(constructors[0]);
                }
                if (baseType === errorType) {
                    return type.resolvedBaseTypes = ts.emptyArray;
                }
                var reducedBaseType = getReducedType(baseType);
                if (!isValidBaseType(reducedBaseType)) {
                    var elaboration = elaborateNeverIntersection(/*errorInfo*/ undefined, baseType);
                    var diagnostic = ts.chainDiagnosticMessages(elaboration, ts.Diagnostics.Base_constructor_return_type_0_is_not_an_object_type_or_intersection_of_object_types_with_statically_known_members, typeToString(reducedBaseType));
                    diagnostics.add(ts.createDiagnosticForNodeFromMessageChain(baseTypeNode.expression, diagnostic));
                    return type.resolvedBaseTypes = ts.emptyArray;
                }
                if (type === reducedBaseType || hasBaseType(reducedBaseType, type)) {
                    error(type.symbol.valueDeclaration, ts.Diagnostics.Type_0_recursively_references_itself_as_a_base_type, typeToString(type, /*enclosingDeclaration*/ undefined, 2 /* WriteArrayAsGenericType */));
                    return type.resolvedBaseTypes = ts.emptyArray;
                }
                if (type.resolvedBaseTypes === ts.resolvingEmptyArray) {
                    // Circular reference, likely through instantiation of default parameters
                    // (otherwise there'd be an error from hasBaseType) - this is fine, but `.members` should be reset
                    // as `getIndexedAccessType` via `instantiateType` via `getTypeFromClassOrInterfaceReference` forces a
                    // partial instantiation of the members without the base types fully resolved
                    type.members = undefined;
                }
                return type.resolvedBaseTypes = [reducedBaseType];
            }
            function areAllOuterTypeParametersApplied(type) {
                // An unapplied type parameter has its symbol still the same as the matching argument symbol.
                // Since parameters are applied outer-to-inner, only the last outer parameter needs to be checked.
                var outerTypeParameters = type.outerTypeParameters;
                if (outerTypeParameters) {
                    var last_1 = outerTypeParameters.length - 1;
                    var typeArguments = getTypeArguments(type);
                    return outerTypeParameters[last_1].symbol !== typeArguments[last_1].symbol;
                }
                return true;
            }
            // A valid base type is `any`, an object type or intersection of object types.
            function isValidBaseType(type) {
                if (type.flags & 262144 /* TypeParameter */) {
                    var constraint = getBaseConstraintOfType(type);
                    if (constraint) {
                        return isValidBaseType(constraint);
                    }
                }
                // TODO: Given that we allow type parmeters here now, is this `!isGenericMappedType(type)` check really needed?
                // There's no reason a `T` should be allowed while a `Readonly<T>` should not.
                return !!(type.flags & (524288 /* Object */ | 67108864 /* NonPrimitive */ | 1 /* Any */) && !isGenericMappedType(type) ||
                    type.flags & 2097152 /* Intersection */ && ts.every(type.types, isValidBaseType));
            }
            function resolveBaseTypesOfInterface(type) {
                type.resolvedBaseTypes = type.resolvedBaseTypes || ts.emptyArray;
                if (type.symbol.declarations) {
                    for (var _i = 0, _a = type.symbol.declarations; _i < _a.length; _i++) {
                        var declaration = _a[_i];
                        if (declaration.kind === 256 /* InterfaceDeclaration */ && ts.getInterfaceBaseTypeNodes(declaration)) {
                            for (var _b = 0, _c = ts.getInterfaceBaseTypeNodes(declaration); _b < _c.length; _b++) {
                                var node = _c[_b];
                                var baseType = getReducedType(getTypeFromTypeNode(node));
                                if (baseType !== errorType) {
                                    if (isValidBaseType(baseType)) {
                                        if (type !== baseType && !hasBaseType(baseType, type)) {
                                            if (type.resolvedBaseTypes === ts.emptyArray) {
                                                type.resolvedBaseTypes = [baseType];
                                            }
                                            else {
                                                type.resolvedBaseTypes.push(baseType);
                                            }
                                        }
                                        else {
                                            reportCircularBaseType(declaration, type);
                                        }
                                    }
                                    else {
                                        error(node, ts.Diagnostics.An_interface_can_only_extend_an_object_type_or_intersection_of_object_types_with_statically_known_members);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            /**
             * Returns true if the interface given by the symbol is free of "this" references.
             *
             * Specifically, the result is true if the interface itself contains no references
             * to "this" in its body, if all base types are interfaces,
             * and if none of the base interfaces have a "this" type.
             */
            function isThislessInterface(symbol) {
                if (!symbol.declarations) {
                    return true;
                }
                for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                    var declaration = _a[_i];
                    if (declaration.kind === 256 /* InterfaceDeclaration */) {
                        if (declaration.flags & 128 /* ContainsThis */) {
                            return false;
                        }
                        var baseTypeNodes = ts.getInterfaceBaseTypeNodes(declaration);
                        if (baseTypeNodes) {
                            for (var _b = 0, baseTypeNodes_1 = baseTypeNodes; _b < baseTypeNodes_1.length; _b++) {
                                var node = baseTypeNodes_1[_b];
                                if (ts.isEntityNameExpression(node.expression)) {
                                    var baseSymbol = resolveEntityName(node.expression, 788968 /* Type */, /*ignoreErrors*/ true);
                                    if (!baseSymbol || !(baseSymbol.flags & 64 /* Interface */) || getDeclaredTypeOfClassOrInterface(baseSymbol).thisType) {
                                        return false;
                                    }
                                }
                            }
                        }
                    }
                }
                return true;
            }
            function getDeclaredTypeOfClassOrInterface(symbol) {
                var links = getSymbolLinks(symbol);
                var originalLinks = links;
                if (!links.declaredType) {
                    var kind = symbol.flags & 32 /* Class */ ? 1 /* Class */ : 2 /* Interface */;
                    var merged = mergeJSSymbols(symbol, symbol.valueDeclaration && getAssignedClassSymbol(symbol.valueDeclaration));
                    if (merged) {
                        // note:we overwrite links because we just cloned the symbol
                        symbol = links = merged;
                    }
                    var type = originalLinks.declaredType = links.declaredType = createObjectType(kind, symbol);
                    var outerTypeParameters = getOuterTypeParametersOfClassOrInterface(symbol);
                    var localTypeParameters = getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(symbol);
                    // A class or interface is generic if it has type parameters or a "this" type. We always give classes a "this" type
                    // because it is not feasible to analyze all members to determine if the "this" type escapes the class (in particular,
                    // property types inferred from initializers and method return types inferred from return statements are very hard
                    // to exhaustively analyze). We give interfaces a "this" type if we can't definitely determine that they are free of
                    // "this" references.
                    if (outerTypeParameters || localTypeParameters || kind === 1 /* Class */ || !isThislessInterface(symbol)) {
                        type.objectFlags |= 4 /* Reference */;
                        type.typeParameters = ts.concatenate(outerTypeParameters, localTypeParameters);
                        type.outerTypeParameters = outerTypeParameters;
                        type.localTypeParameters = localTypeParameters;
                        type.instantiations = new ts.Map();
                        type.instantiations.set(getTypeListId(type.typeParameters), type);
                        type.target = type;
                        type.resolvedTypeArguments = type.typeParameters;
                        type.thisType = createTypeParameter(symbol);
                        type.thisType.isThisType = true;
                        type.thisType.constraint = type;
                    }
                }
                return links.declaredType;
            }
            function getDeclaredTypeOfTypeAlias(symbol) {
                var _a;
                var links = getSymbolLinks(symbol);
                if (!links.declaredType) {
                    // Note that we use the links object as the target here because the symbol object is used as the unique
                    // identity for resolution of the 'type' property in SymbolLinks.
                    if (!pushTypeResolution(symbol, 2 /* DeclaredType */)) {
                        return errorType;
                    }
                    var declaration = ts.Debug.checkDefined((_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.isTypeAlias), "Type alias symbol with no valid declaration found");
                    var typeNode = ts.isJSDocTypeAlias(declaration) ? declaration.typeExpression : declaration.type;
                    // If typeNode is missing, we will error in checkJSDocTypedefTag.
                    var type = typeNode ? getTypeFromTypeNode(typeNode) : errorType;
                    if (popTypeResolution()) {
                        var typeParameters = getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(symbol);
                        if (typeParameters) {
                            // Initialize the instantiation cache for generic type aliases. The declared type corresponds to
                            // an instantiation of the type alias with the type parameters supplied as type arguments.
                            links.typeParameters = typeParameters;
                            links.instantiations = new ts.Map();
                            links.instantiations.set(getTypeListId(typeParameters), type);
                        }
                    }
                    else {
                        type = errorType;
                        if (declaration.kind === 334 /* JSDocEnumTag */) {
                            error(declaration.typeExpression.type, ts.Diagnostics.Type_alias_0_circularly_references_itself, symbolToString(symbol));
                        }
                        else {
                            error(ts.isNamedDeclaration(declaration) ? declaration.name : declaration || declaration, ts.Diagnostics.Type_alias_0_circularly_references_itself, symbolToString(symbol));
                        }
                    }
                    links.declaredType = type;
                }
                return links.declaredType;
            }
            function isStringConcatExpression(expr) {
                if (ts.isStringLiteralLike(expr)) {
                    return true;
                }
                else if (expr.kind === 219 /* BinaryExpression */) {
                    return isStringConcatExpression(expr.left) && isStringConcatExpression(expr.right);
                }
                return false;
            }
            function isLiteralEnumMember(member) {
                var expr = member.initializer;
                if (!expr) {
                    return !(member.flags & 8388608 /* Ambient */);
                }
                switch (expr.kind) {
                    case 10 /* StringLiteral */:
                    case 8 /* NumericLiteral */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                        return true;
                    case 217 /* PrefixUnaryExpression */:
                        return expr.operator === 40 /* MinusToken */ &&
                            expr.operand.kind === 8 /* NumericLiteral */;
                    case 79 /* Identifier */:
                        return ts.nodeIsMissing(expr) || !!getSymbolOfNode(member.parent).exports.get(expr.escapedText);
                    case 219 /* BinaryExpression */:
                        return isStringConcatExpression(expr);
                    default:
                        return false;
                }
            }
            function getEnumKind(symbol) {
                var links = getSymbolLinks(symbol);
                if (links.enumKind !== undefined) {
                    return links.enumKind;
                }
                var hasNonLiteralMember = false;
                if (symbol.declarations) {
                    for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                        var declaration = _a[_i];
                        if (declaration.kind === 258 /* EnumDeclaration */) {
                            for (var _b = 0, _c = declaration.members; _b < _c.length; _b++) {
                                var member = _c[_b];
                                if (member.initializer && ts.isStringLiteralLike(member.initializer)) {
                                    return links.enumKind = 1 /* Literal */;
                                }
                                if (!isLiteralEnumMember(member)) {
                                    hasNonLiteralMember = true;
                                }
                            }
                        }
                    }
                }
                return links.enumKind = hasNonLiteralMember ? 0 /* Numeric */ : 1 /* Literal */;
            }
            function getBaseTypeOfEnumLiteralType(type) {
                return type.flags & 1024 /* EnumLiteral */ && !(type.flags & 1048576 /* Union */) ? getDeclaredTypeOfSymbol(getParentOfSymbol(type.symbol)) : type;
            }
            function getDeclaredTypeOfEnum(symbol) {
                var links = getSymbolLinks(symbol);
                if (links.declaredType) {
                    return links.declaredType;
                }
                if (getEnumKind(symbol) === 1 /* Literal */) {
                    enumCount++;
                    var memberTypeList = [];
                    if (symbol.declarations) {
                        for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                            var declaration = _a[_i];
                            if (declaration.kind === 258 /* EnumDeclaration */) {
                                for (var _b = 0, _c = declaration.members; _b < _c.length; _b++) {
                                    var member = _c[_b];
                                    var value = getEnumMemberValue(member);
                                    var memberType = getFreshTypeOfLiteralType(getEnumLiteralType(value !== undefined ? value : 0, enumCount, getSymbolOfNode(member)));
                                    getSymbolLinks(getSymbolOfNode(member)).declaredType = memberType;
                                    memberTypeList.push(getRegularTypeOfLiteralType(memberType));
                                }
                            }
                        }
                    }
                    if (memberTypeList.length) {
                        var enumType_1 = getUnionType(memberTypeList, 1 /* Literal */, symbol, /*aliasTypeArguments*/ undefined);
                        if (enumType_1.flags & 1048576 /* Union */) {
                            enumType_1.flags |= 1024 /* EnumLiteral */;
                            enumType_1.symbol = symbol;
                        }
                        return links.declaredType = enumType_1;
                    }
                }
                var enumType = createType(32 /* Enum */);
                enumType.symbol = symbol;
                return links.declaredType = enumType;
            }
            function getDeclaredTypeOfEnumMember(symbol) {
                var links = getSymbolLinks(symbol);
                if (!links.declaredType) {
                    var enumType = getDeclaredTypeOfEnum(getParentOfSymbol(symbol));
                    if (!links.declaredType) {
                        links.declaredType = enumType;
                    }
                }
                return links.declaredType;
            }
            function getDeclaredTypeOfTypeParameter(symbol) {
                var links = getSymbolLinks(symbol);
                return links.declaredType || (links.declaredType = createTypeParameter(symbol));
            }
            function getDeclaredTypeOfAlias(symbol) {
                var links = getSymbolLinks(symbol);
                return links.declaredType || (links.declaredType = getDeclaredTypeOfSymbol(resolveAlias(symbol)));
            }
            function getDeclaredTypeOfSymbol(symbol) {
                return tryGetDeclaredTypeOfSymbol(symbol) || errorType;
            }
            function tryGetDeclaredTypeOfSymbol(symbol) {
                if (symbol.flags & (32 /* Class */ | 64 /* Interface */)) {
                    return getDeclaredTypeOfClassOrInterface(symbol);
                }
                if (symbol.flags & 524288 /* TypeAlias */) {
                    return getDeclaredTypeOfTypeAlias(symbol);
                }
                if (symbol.flags & 262144 /* TypeParameter */) {
                    return getDeclaredTypeOfTypeParameter(symbol);
                }
                if (symbol.flags & 384 /* Enum */) {
                    return getDeclaredTypeOfEnum(symbol);
                }
                if (symbol.flags & 8 /* EnumMember */) {
                    return getDeclaredTypeOfEnumMember(symbol);
                }
                if (symbol.flags & 2097152 /* Alias */) {
                    return getDeclaredTypeOfAlias(symbol);
                }
                return undefined;
            }
            /**
             * A type is free of this references if it's the any, string, number, boolean, symbol, or void keyword, a string
             * literal type, an array with an element type that is free of this references, or a type reference that is
             * free of this references.
             */
            function isThislessType(node) {
                switch (node.kind) {
                    case 129 /* AnyKeyword */:
                    case 153 /* UnknownKeyword */:
                    case 148 /* StringKeyword */:
                    case 145 /* NumberKeyword */:
                    case 156 /* BigIntKeyword */:
                    case 132 /* BooleanKeyword */:
                    case 149 /* SymbolKeyword */:
                    case 146 /* ObjectKeyword */:
                    case 114 /* VoidKeyword */:
                    case 151 /* UndefinedKeyword */:
                    case 142 /* NeverKeyword */:
                    case 194 /* LiteralType */:
                        return true;
                    case 181 /* ArrayType */:
                        return isThislessType(node.elementType);
                    case 176 /* TypeReference */:
                        return !node.typeArguments || node.typeArguments.every(isThislessType);
                }
                return false;
            }
            /** A type parameter is thisless if its constraint is thisless, or if it has no constraint. */
            function isThislessTypeParameter(node) {
                var constraint = ts.getEffectiveConstraintOfTypeParameter(node);
                return !constraint || isThislessType(constraint);
            }
            /**
             * A variable-like declaration is free of this references if it has a type annotation
             * that is thisless, or if it has no type annotation and no initializer (and is thus of type any).
             */
            function isThislessVariableLikeDeclaration(node) {
                var typeNode = ts.getEffectiveTypeAnnotationNode(node);
                return typeNode ? isThislessType(typeNode) : !ts.hasInitializer(node);
            }
            /**
             * A function-like declaration is considered free of `this` references if it has a return type
             * annotation that is free of this references and if each parameter is thisless and if
             * each type parameter (if present) is thisless.
             */
            function isThislessFunctionLikeDeclaration(node) {
                var returnType = ts.getEffectiveReturnTypeNode(node);
                var typeParameters = ts.getEffectiveTypeParameterDeclarations(node);
                return (node.kind === 169 /* Constructor */ || (!!returnType && isThislessType(returnType))) &&
                    node.parameters.every(isThislessVariableLikeDeclaration) &&
                    typeParameters.every(isThislessTypeParameter);
            }
            /**
             * Returns true if the class or interface member given by the symbol is free of "this" references. The
             * function may return false for symbols that are actually free of "this" references because it is not
             * feasible to perform a complete analysis in all cases. In particular, property members with types
             * inferred from their initializers and function members with inferred return types are conservatively
             * assumed not to be free of "this" references.
             */
            function isThisless(symbol) {
                if (symbol.declarations && symbol.declarations.length === 1) {
                    var declaration = symbol.declarations[0];
                    if (declaration) {
                        switch (declaration.kind) {
                            case 165 /* PropertyDeclaration */:
                            case 164 /* PropertySignature */:
                                return isThislessVariableLikeDeclaration(declaration);
                            case 167 /* MethodDeclaration */:
                            case 166 /* MethodSignature */:
                            case 169 /* Constructor */:
                            case 170 /* GetAccessor */:
                            case 171 /* SetAccessor */:
                                return isThislessFunctionLikeDeclaration(declaration);
                        }
                    }
                }
                return false;
            }
            // The mappingThisOnly flag indicates that the only type parameter being mapped is "this". When the flag is true,
            // we check symbols to see if we can quickly conclude they are free of "this" references, thus needing no instantiation.
            function createInstantiatedSymbolTable(symbols, mapper, mappingThisOnly) {
                var result = ts.createSymbolTable();
                for (var _i = 0, symbols_2 = symbols; _i < symbols_2.length; _i++) {
                    var symbol = symbols_2[_i];
                    result.set(symbol.escapedName, mappingThisOnly && isThisless(symbol) ? symbol : instantiateSymbol(symbol, mapper));
                }
                return result;
            }
            function addInheritedMembers(symbols, baseSymbols) {
                for (var _i = 0, baseSymbols_1 = baseSymbols; _i < baseSymbols_1.length; _i++) {
                    var s = baseSymbols_1[_i];
                    if (!symbols.has(s.escapedName) && !isStaticPrivateIdentifierProperty(s)) {
                        symbols.set(s.escapedName, s);
                    }
                }
            }
            function isStaticPrivateIdentifierProperty(s) {
                return !!s.valueDeclaration && ts.isPrivateIdentifierClassElementDeclaration(s.valueDeclaration) && ts.isStatic(s.valueDeclaration);
            }
            function resolveDeclaredMembers(type) {
                if (!type.declaredProperties) {
                    var symbol = type.symbol;
                    var members = getMembersOfSymbol(symbol);
                    type.declaredProperties = getNamedMembers(members);
                    // Start with signatures at empty array in case of recursive types
                    type.declaredCallSignatures = ts.emptyArray;
                    type.declaredConstructSignatures = ts.emptyArray;
                    type.declaredIndexInfos = ts.emptyArray;
                    type.declaredCallSignatures = getSignaturesOfSymbol(members.get("__call" /* Call */));
                    type.declaredConstructSignatures = getSignaturesOfSymbol(members.get("__new" /* New */));
                    type.declaredIndexInfos = getIndexInfosOfSymbol(symbol);
                }
                return type;
            }
            /**
             * Indicates whether a type can be used as a property name.
             */
            function isTypeUsableAsPropertyName(type) {
                return !!(type.flags & 8576 /* StringOrNumberLiteralOrUnique */);
            }
            /**
             * Indicates whether a declaration name is definitely late-bindable.
             * A declaration name is only late-bindable if:
             * - It is a `ComputedPropertyName`.
             * - Its expression is an `Identifier` or either a `PropertyAccessExpression` an
             * `ElementAccessExpression` consisting only of these same three types of nodes.
             * - The type of its expression is a string or numeric literal type, or is a `unique symbol` type.
             */
            function isLateBindableName(node) {
                if (!ts.isComputedPropertyName(node) && !ts.isElementAccessExpression(node)) {
                    return false;
                }
                var expr = ts.isComputedPropertyName(node) ? node.expression : node.argumentExpression;
                return ts.isEntityNameExpression(expr)
                    && isTypeUsableAsPropertyName(ts.isComputedPropertyName(node) ? checkComputedPropertyName(node) : checkExpressionCached(expr));
            }
            function isLateBoundName(name) {
                return name.charCodeAt(0) === 95 /* _ */ &&
                    name.charCodeAt(1) === 95 /* _ */ &&
                    name.charCodeAt(2) === 64 /* at */;
            }
            /**
             * Indicates whether a declaration has a late-bindable dynamic name.
             */
            function hasLateBindableName(node) {
                var name = ts.getNameOfDeclaration(node);
                return !!name && isLateBindableName(name);
            }
            /**
             * Indicates whether a declaration has an early-bound name or a dynamic name that can be late-bound.
             */
            function hasBindableName(node) {
                return !ts.hasDynamicName(node) || hasLateBindableName(node);
            }
            /**
             * Indicates whether a declaration name is a dynamic name that cannot be late-bound.
             */
            function isNonBindableDynamicName(node) {
                return ts.isDynamicName(node) && !isLateBindableName(node);
            }
            /**
             * Gets the symbolic name for a member from its type.
             */
            function getPropertyNameFromType(type) {
                if (type.flags & 8192 /* UniqueESSymbol */) {
                    return type.escapedName;
                }
                if (type.flags & (128 /* StringLiteral */ | 256 /* NumberLiteral */)) {
                    return ts.escapeLeadingUnderscores("" + type.value);
                }
                return ts.Debug.fail();
            }
            /**
             * Adds a declaration to a late-bound dynamic member. This performs the same function for
             * late-bound members that `addDeclarationToSymbol` in binder.ts performs for early-bound
             * members.
             */
            function addDeclarationToLateBoundSymbol(symbol, member, symbolFlags) {
                ts.Debug.assert(!!(ts.getCheckFlags(symbol) & 4096 /* Late */), "Expected a late-bound symbol.");
                symbol.flags |= symbolFlags;
                getSymbolLinks(member.symbol).lateSymbol = symbol;
                if (!symbol.declarations) {
                    symbol.declarations = [member];
                }
                else if (!member.symbol.isReplaceableByMethod) {
                    symbol.declarations.push(member);
                }
                if (symbolFlags & 111551 /* Value */) {
                    if (!symbol.valueDeclaration || symbol.valueDeclaration.kind !== member.kind) {
                        symbol.valueDeclaration = member;
                    }
                }
            }
            /**
             * Performs late-binding of a dynamic member. This performs the same function for
             * late-bound members that `declareSymbol` in binder.ts performs for early-bound
             * members.
             *
             * If a symbol is a dynamic name from a computed property, we perform an additional "late"
             * binding phase to attempt to resolve the name for the symbol from the type of the computed
             * property's expression. If the type of the expression is a string-literal, numeric-literal,
             * or unique symbol type, we can use that type as the name of the symbol.
             *
             * For example, given:
             *
             *   const x = Symbol();
             *
             *   interface I {
             *     [x]: number;
             *   }
             *
             * The binder gives the property `[x]: number` a special symbol with the name "__computed".
             * In the late-binding phase we can type-check the expression `x` and see that it has a
             * unique symbol type which we can then use as the name of the member. This allows users
             * to define custom symbols that can be used in the members of an object type.
             *
             * @param parent The containing symbol for the member.
             * @param earlySymbols The early-bound symbols of the parent.
             * @param lateSymbols The late-bound symbols of the parent.
             * @param decl The member to bind.
             */
            function lateBindMember(parent, earlySymbols, lateSymbols, decl) {
                ts.Debug.assert(!!decl.symbol, "The member is expected to have a symbol.");
                var links = getNodeLinks(decl);
                if (!links.resolvedSymbol) {
                    // In the event we attempt to resolve the late-bound name of this member recursively,
                    // fall back to the early-bound name of this member.
                    links.resolvedSymbol = decl.symbol;
                    var declName = ts.isBinaryExpression(decl) ? decl.left : decl.name;
                    var type = ts.isElementAccessExpression(declName) ? checkExpressionCached(declName.argumentExpression) : checkComputedPropertyName(declName);
                    if (isTypeUsableAsPropertyName(type)) {
                        var memberName = getPropertyNameFromType(type);
                        var symbolFlags = decl.symbol.flags;
                        // Get or add a late-bound symbol for the member. This allows us to merge late-bound accessor declarations.
                        var lateSymbol = lateSymbols.get(memberName);
                        if (!lateSymbol)
                            lateSymbols.set(memberName, lateSymbol = createSymbol(0 /* None */, memberName, 4096 /* Late */));
                        // Report an error if a late-bound member has the same name as an early-bound member,
                        // or if we have another early-bound symbol declaration with the same name and
                        // conflicting flags.
                        var earlySymbol = earlySymbols && earlySymbols.get(memberName);
                        if (lateSymbol.flags & getExcludedSymbolFlags(symbolFlags) || earlySymbol) {
                            // If we have an existing early-bound member, combine its declarations so that we can
                            // report an error at each declaration.
                            var declarations = earlySymbol ? ts.concatenate(earlySymbol.declarations, lateSymbol.declarations) : lateSymbol.declarations;
                            var name_4 = !(type.flags & 8192 /* UniqueESSymbol */) && ts.unescapeLeadingUnderscores(memberName) || ts.declarationNameToString(declName);
                            ts.forEach(declarations, function (declaration) { return error(ts.getNameOfDeclaration(declaration) || declaration, ts.Diagnostics.Property_0_was_also_declared_here, name_4); });
                            error(declName || decl, ts.Diagnostics.Duplicate_property_0, name_4);
                            lateSymbol = createSymbol(0 /* None */, memberName, 4096 /* Late */);
                        }
                        lateSymbol.nameType = type;
                        addDeclarationToLateBoundSymbol(lateSymbol, decl, symbolFlags);
                        if (lateSymbol.parent) {
                            ts.Debug.assert(lateSymbol.parent === parent, "Existing symbol parent should match new one");
                        }
                        else {
                            lateSymbol.parent = parent;
                        }
                        return links.resolvedSymbol = lateSymbol;
                    }
                }
                return links.resolvedSymbol;
            }
            function getResolvedMembersOrExportsOfSymbol(symbol, resolutionKind) {
                var links = getSymbolLinks(symbol);
                if (!links[resolutionKind]) {
                    var isStatic_1 = resolutionKind === "resolvedExports" /* resolvedExports */;
                    var earlySymbols = !isStatic_1 ? symbol.members :
                        symbol.flags & 1536 /* Module */ ? getExportsOfModuleWorker(symbol) :
                            symbol.exports;
                    // In the event we recursively resolve the members/exports of the symbol, we
                    // set the initial value of resolvedMembers/resolvedExports to the early-bound
                    // members/exports of the symbol.
                    links[resolutionKind] = earlySymbols || emptySymbols;
                    // fill in any as-yet-unresolved late-bound members.
                    var lateSymbols = ts.createSymbolTable();
                    for (var _i = 0, _a = symbol.declarations || ts.emptyArray; _i < _a.length; _i++) {
                        var decl = _a[_i];
                        var members = ts.getMembersOfDeclaration(decl);
                        if (members) {
                            for (var _b = 0, members_5 = members; _b < members_5.length; _b++) {
                                var member = members_5[_b];
                                if (isStatic_1 === ts.hasStaticModifier(member) && hasLateBindableName(member)) {
                                    lateBindMember(symbol, earlySymbols, lateSymbols, member);
                                }
                            }
                        }
                    }
                    var assignments = symbol.assignmentDeclarationMembers;
                    if (assignments) {
                        var decls = ts.arrayFrom(assignments.values());
                        for (var _c = 0, decls_1 = decls; _c < decls_1.length; _c++) {
                            var member = decls_1[_c];
                            var assignmentKind = ts.getAssignmentDeclarationKind(member);
                            var isInstanceMember = assignmentKind === 3 /* PrototypeProperty */
                                || ts.isBinaryExpression(member) && isPossiblyAliasedThisProperty(member, assignmentKind)
                                || assignmentKind === 9 /* ObjectDefinePrototypeProperty */
                                || assignmentKind === 6 /* Prototype */; // A straight `Prototype` assignment probably can never have a computed name
                            if (isStatic_1 === !isInstanceMember && hasLateBindableName(member)) {
                                lateBindMember(symbol, earlySymbols, lateSymbols, member);
                            }
                        }
                    }
                    links[resolutionKind] = combineSymbolTables(earlySymbols, lateSymbols) || emptySymbols;
                }
                return links[resolutionKind];
            }
            /**
             * Gets a SymbolTable containing both the early- and late-bound members of a symbol.
             *
             * For a description of late-binding, see `lateBindMember`.
             */
            function getMembersOfSymbol(symbol) {
                return symbol.flags & 6256 /* LateBindingContainer */
                    ? getResolvedMembersOrExportsOfSymbol(symbol, "resolvedMembers" /* resolvedMembers */)
                    : symbol.members || emptySymbols;
            }
            /**
             * If a symbol is the dynamic name of the member of an object type, get the late-bound
             * symbol of the member.
             *
             * For a description of late-binding, see `lateBindMember`.
             */
            function getLateBoundSymbol(symbol) {
                if (symbol.flags & 106500 /* ClassMember */ && symbol.escapedName === "__computed" /* Computed */) {
                    var links = getSymbolLinks(symbol);
                    if (!links.lateSymbol && ts.some(symbol.declarations, hasLateBindableName)) {
                        // force late binding of members/exports. This will set the late-bound symbol
                        var parent = getMergedSymbol(symbol.parent);
                        if (ts.some(symbol.declarations, ts.hasStaticModifier)) {
                            getExportsOfSymbol(parent);
                        }
                        else {
                            getMembersOfSymbol(parent);
                        }
                    }
                    return links.lateSymbol || (links.lateSymbol = symbol);
                }
                return symbol;
            }
            function getTypeWithThisArgument(type, thisArgument, needApparentType) {
                if (ts.getObjectFlags(type) & 4 /* Reference */) {
                    var target = type.target;
                    var typeArguments = getTypeArguments(type);
                    if (ts.length(target.typeParameters) === ts.length(typeArguments)) {
                        var ref = createTypeReference(target, ts.concatenate(typeArguments, [thisArgument || target.thisType]));
                        return needApparentType ? getApparentType(ref) : ref;
                    }
                }
                else if (type.flags & 2097152 /* Intersection */) {
                    var types = ts.sameMap(type.types, function (t) { return getTypeWithThisArgument(t, thisArgument, needApparentType); });
                    return types !== type.types ? getIntersectionType(types) : type;
                }
                return needApparentType ? getApparentType(type) : type;
            }
            function resolveObjectTypeMembers(type, source, typeParameters, typeArguments) {
                var mapper;
                var members;
                var callSignatures;
                var constructSignatures;
                var indexInfos;
                if (ts.rangeEquals(typeParameters, typeArguments, 0, typeParameters.length)) {
                    members = source.symbol ? getMembersOfSymbol(source.symbol) : ts.createSymbolTable(source.declaredProperties);
                    callSignatures = source.declaredCallSignatures;
                    constructSignatures = source.declaredConstructSignatures;
                    indexInfos = source.declaredIndexInfos;
                }
                else {
                    mapper = createTypeMapper(typeParameters, typeArguments);
                    members = createInstantiatedSymbolTable(source.declaredProperties, mapper, /*mappingThisOnly*/ typeParameters.length === 1);
                    callSignatures = instantiateSignatures(source.declaredCallSignatures, mapper);
                    constructSignatures = instantiateSignatures(source.declaredConstructSignatures, mapper);
                    indexInfos = instantiateIndexInfos(source.declaredIndexInfos, mapper);
                }
                var baseTypes = getBaseTypes(source);
                if (baseTypes.length) {
                    if (source.symbol && members === getMembersOfSymbol(source.symbol)) {
                        members = ts.createSymbolTable(source.declaredProperties);
                    }
                    setStructuredTypeMembers(type, members, callSignatures, constructSignatures, indexInfos);
                    var thisArgument = ts.lastOrUndefined(typeArguments);
                    for (var _i = 0, baseTypes_1 = baseTypes; _i < baseTypes_1.length; _i++) {
                        var baseType = baseTypes_1[_i];
                        var instantiatedBaseType = thisArgument ? getTypeWithThisArgument(instantiateType(baseType, mapper), thisArgument) : baseType;
                        addInheritedMembers(members, getPropertiesOfType(instantiatedBaseType));
                        callSignatures = ts.concatenate(callSignatures, getSignaturesOfType(instantiatedBaseType, 0 /* Call */));
                        constructSignatures = ts.concatenate(constructSignatures, getSignaturesOfType(instantiatedBaseType, 1 /* Construct */));
                        var inheritedIndexInfos = instantiatedBaseType !== anyType ? getIndexInfosOfType(instantiatedBaseType) : [createIndexInfo(stringType, anyType, /*isReadonly*/ false)];
                        indexInfos = ts.concatenate(indexInfos, ts.filter(inheritedIndexInfos, function (info) { return !findIndexInfo(indexInfos, info.keyType); }));
                    }
                }
                setStructuredTypeMembers(type, members, callSignatures, constructSignatures, indexInfos);
            }
            function resolveClassOrInterfaceMembers(type) {
                resolveObjectTypeMembers(type, resolveDeclaredMembers(type), ts.emptyArray, ts.emptyArray);
            }
            function resolveTypeReferenceMembers(type) {
                var source = resolveDeclaredMembers(type.target);
                var typeParameters = ts.concatenate(source.typeParameters, [source.thisType]);
                var typeArguments = getTypeArguments(type);
                var paddedTypeArguments = typeArguments.length === typeParameters.length ? typeArguments : ts.concatenate(typeArguments, [type]);
                resolveObjectTypeMembers(type, source, typeParameters, paddedTypeArguments);
            }
            function createSignature(declaration, typeParameters, thisParameter, parameters, resolvedReturnType, resolvedTypePredicate, minArgumentCount, flags) {
                var sig = new Signature(checker, flags);
                sig.declaration = declaration;
                sig.typeParameters = typeParameters;
                sig.parameters = parameters;
                sig.thisParameter = thisParameter;
                sig.resolvedReturnType = resolvedReturnType;
                sig.resolvedTypePredicate = resolvedTypePredicate;
                sig.minArgumentCount = minArgumentCount;
                sig.resolvedMinArgumentCount = undefined;
                sig.target = undefined;
                sig.mapper = undefined;
                sig.compositeSignatures = undefined;
                sig.compositeKind = undefined;
                return sig;
            }
            function cloneSignature(sig) {
                var result = createSignature(sig.declaration, sig.typeParameters, sig.thisParameter, sig.parameters, /*resolvedReturnType*/ undefined, 
                /*resolvedTypePredicate*/ undefined, sig.minArgumentCount, sig.flags & 39 /* PropagatingFlags */);
                result.target = sig.target;
                result.mapper = sig.mapper;
                result.compositeSignatures = sig.compositeSignatures;
                result.compositeKind = sig.compositeKind;
                return result;
            }
            function createUnionSignature(signature, unionSignatures) {
                var result = cloneSignature(signature);
                result.compositeSignatures = unionSignatures;
                result.compositeKind = 1048576 /* Union */;
                result.target = undefined;
                result.mapper = undefined;
                return result;
            }
            function getOptionalCallSignature(signature, callChainFlags) {
                if ((signature.flags & 24 /* CallChainFlags */) === callChainFlags) {
                    return signature;
                }
                if (!signature.optionalCallSignatureCache) {
                    signature.optionalCallSignatureCache = {};
                }
                var key = callChainFlags === 8 /* IsInnerCallChain */ ? "inner" : "outer";
                return signature.optionalCallSignatureCache[key]
                    || (signature.optionalCallSignatureCache[key] = createOptionalCallSignature(signature, callChainFlags));
            }
            function createOptionalCallSignature(signature, callChainFlags) {
                ts.Debug.assert(callChainFlags === 8 /* IsInnerCallChain */ || callChainFlags === 16 /* IsOuterCallChain */, "An optional call signature can either be for an inner call chain or an outer call chain, but not both.");
                var result = cloneSignature(signature);
                result.flags |= callChainFlags;
                return result;
            }
            function getExpandedParameters(sig, skipUnionExpanding) {
                if (signatureHasRestParameter(sig)) {
                    var restIndex_1 = sig.parameters.length - 1;
                    var restType = getTypeOfSymbol(sig.parameters[restIndex_1]);
                    if (isTupleType(restType)) {
                        return [expandSignatureParametersWithTupleMembers(restType, restIndex_1)];
                    }
                    else if (!skipUnionExpanding && restType.flags & 1048576 /* Union */ && ts.every(restType.types, isTupleType)) {
                        return ts.map(restType.types, function (t) { return expandSignatureParametersWithTupleMembers(t, restIndex_1); });
                    }
                }
                return [sig.parameters];
                function expandSignatureParametersWithTupleMembers(restType, restIndex) {
                    var elementTypes = getTypeArguments(restType);
                    var associatedNames = restType.target.labeledElementDeclarations;
                    var restParams = ts.map(elementTypes, function (t, i) {
                        // Lookup the label from the individual tuple passed in before falling back to the signature `rest` parameter name
                        var tupleLabelName = !!associatedNames && getTupleElementLabel(associatedNames[i]);
                        var name = tupleLabelName || getParameterNameAtPosition(sig, restIndex + i, restType);
                        var flags = restType.target.elementFlags[i];
                        var checkFlags = flags & 12 /* Variable */ ? 32768 /* RestParameter */ :
                            flags & 2 /* Optional */ ? 16384 /* OptionalParameter */ : 0;
                        var symbol = createSymbol(1 /* FunctionScopedVariable */, name, checkFlags);
                        symbol.type = flags & 4 /* Rest */ ? createArrayType(t) : t;
                        return symbol;
                    });
                    return ts.concatenate(sig.parameters.slice(0, restIndex), restParams);
                }
            }
            function getDefaultConstructSignatures(classType) {
                var baseConstructorType = getBaseConstructorTypeOfClass(classType);
                var baseSignatures = getSignaturesOfType(baseConstructorType, 1 /* Construct */);
                var declaration = ts.getClassLikeDeclarationOfSymbol(classType.symbol);
                var isAbstract = !!declaration && ts.hasSyntacticModifier(declaration, 128 /* Abstract */);
                if (baseSignatures.length === 0) {
                    return [createSignature(undefined, classType.localTypeParameters, undefined, ts.emptyArray, classType, /*resolvedTypePredicate*/ undefined, 0, isAbstract ? 4 /* Abstract */ : 0 /* None */)];
                }
                var baseTypeNode = getBaseTypeNodeOfClass(classType);
                var isJavaScript = ts.isInJSFile(baseTypeNode);
                var typeArguments = typeArgumentsFromTypeReferenceNode(baseTypeNode);
                var typeArgCount = ts.length(typeArguments);
                var result = [];
                for (var _i = 0, baseSignatures_1 = baseSignatures; _i < baseSignatures_1.length; _i++) {
                    var baseSig = baseSignatures_1[_i];
                    var minTypeArgumentCount = getMinTypeArgumentCount(baseSig.typeParameters);
                    var typeParamCount = ts.length(baseSig.typeParameters);
                    if (isJavaScript || typeArgCount >= minTypeArgumentCount && typeArgCount <= typeParamCount) {
                        var sig = typeParamCount ? createSignatureInstantiation(baseSig, fillMissingTypeArguments(typeArguments, baseSig.typeParameters, minTypeArgumentCount, isJavaScript)) : cloneSignature(baseSig);
                        sig.typeParameters = classType.localTypeParameters;
                        sig.resolvedReturnType = classType;
                        sig.flags = isAbstract ? sig.flags | 4 /* Abstract */ : sig.flags & ~4 /* Abstract */;
                        result.push(sig);
                    }
                }
                return result;
            }
            function findMatchingSignature(signatureList, signature, partialMatch, ignoreThisTypes, ignoreReturnTypes) {
                for (var _i = 0, signatureList_1 = signatureList; _i < signatureList_1.length; _i++) {
                    var s = signatureList_1[_i];
                    if (compareSignaturesIdentical(s, signature, partialMatch, ignoreThisTypes, ignoreReturnTypes, partialMatch ? compareTypesSubtypeOf : compareTypesIdentical)) {
                        return s;
                    }
                }
            }
            function findMatchingSignatures(signatureLists, signature, listIndex) {
                if (signature.typeParameters) {
                    // We require an exact match for generic signatures, so we only return signatures from the first
                    // signature list and only if they have exact matches in the other signature lists.
                    if (listIndex > 0) {
                        return undefined;
                    }
                    for (var i = 1; i < signatureLists.length; i++) {
                        if (!findMatchingSignature(signatureLists[i], signature, /*partialMatch*/ false, /*ignoreThisTypes*/ false, /*ignoreReturnTypes*/ false)) {
                            return undefined;
                        }
                    }
                    return [signature];
                }
                var result;
                for (var i = 0; i < signatureLists.length; i++) {
                    // Allow matching non-generic signatures to have excess parameters and different return types.
                    // Prefer matching this types if possible.
                    var match = i === listIndex ? signature : findMatchingSignature(signatureLists[i], signature, /*partialMatch*/ true, /*ignoreThisTypes*/ false, /*ignoreReturnTypes*/ true);
                    if (!match) {
                        return undefined;
                    }
                    result = ts.appendIfUnique(result, match);
                }
                return result;
            }
            // The signatures of a union type are those signatures that are present in each of the constituent types.
            // Generic signatures must match exactly, but non-generic signatures are allowed to have extra optional
            // parameters and may differ in return types. When signatures differ in return types, the resulting return
            // type is the union of the constituent return types.
            function getUnionSignatures(signatureLists) {
                var result;
                var indexWithLengthOverOne;
                for (var i = 0; i < signatureLists.length; i++) {
                    if (signatureLists[i].length === 0)
                        return ts.emptyArray;
                    if (signatureLists[i].length > 1) {
                        indexWithLengthOverOne = indexWithLengthOverOne === undefined ? i : -1; // -1 is a signal there are multiple overload sets
                    }
                    for (var _i = 0, _a = signatureLists[i]; _i < _a.length; _i++) {
                        var signature = _a[_i];
                        // Only process signatures with parameter lists that aren't already in the result list
                        if (!result || !findMatchingSignature(result, signature, /*partialMatch*/ false, /*ignoreThisTypes*/ false, /*ignoreReturnTypes*/ true)) {
                            var unionSignatures = findMatchingSignatures(signatureLists, signature, i);
                            if (unionSignatures) {
                                var s = signature;
                                // Union the result types when more than one signature matches
                                if (unionSignatures.length > 1) {
                                    var thisParameter = signature.thisParameter;
                                    var firstThisParameterOfUnionSignatures = ts.forEach(unionSignatures, function (sig) { return sig.thisParameter; });
                                    if (firstThisParameterOfUnionSignatures) {
                                        var thisType = getIntersectionType(ts.mapDefined(unionSignatures, function (sig) { return sig.thisParameter && getTypeOfSymbol(sig.thisParameter); }));
                                        thisParameter = createSymbolWithType(firstThisParameterOfUnionSignatures, thisType);
                                    }
                                    s = createUnionSignature(signature, unionSignatures);
                                    s.thisParameter = thisParameter;
                                }
                                (result || (result = [])).push(s);
                            }
                        }
                    }
                }
                if (!ts.length(result) && indexWithLengthOverOne !== -1) {
                    // No sufficiently similar signature existed to subsume all the other signatures in the union - time to see if we can make a single
                    // signature that handles all over them. We only do this when there are overloads in only one constituent.
                    // (Overloads are conditional in nature and having overloads in multiple constituents would necessitate making a power set of
                    // signatures from the type, whose ordering would be non-obvious)
                    var masterList = signatureLists[indexWithLengthOverOne !== undefined ? indexWithLengthOverOne : 0];
                    var results = masterList.slice();
                    var _loop_10 = function (signatures) {
                        if (signatures !== masterList) {
                            var signature_1 = signatures[0];
                            ts.Debug.assert(!!signature_1, "getUnionSignatures bails early on empty signature lists and should not have empty lists on second pass");
                            results = !!signature_1.typeParameters && ts.some(results, function (s) { return !!s.typeParameters && !compareTypeParametersIdentical(signature_1.typeParameters, s.typeParameters); }) ? undefined : ts.map(results, function (sig) { return combineSignaturesOfUnionMembers(sig, signature_1); });
                            if (!results) {
                                return "break";
                            }
                        }
                    };
                    for (var _b = 0, signatureLists_1 = signatureLists; _b < signatureLists_1.length; _b++) {
                        var signatures = signatureLists_1[_b];
                        var state_3 = _loop_10(signatures);
                        if (state_3 === "break")
                            break;
                    }
                    result = results;
                }
                return result || ts.emptyArray;
            }
            function compareTypeParametersIdentical(sourceParams, targetParams) {
                if (ts.length(sourceParams) !== ts.length(targetParams)) {
                    return false;
                }
                if (!sourceParams || !targetParams) {
                    return true;
                }
                var mapper = createTypeMapper(targetParams, sourceParams);
                for (var i = 0; i < sourceParams.length; i++) {
                    var source = sourceParams[i];
                    var target = targetParams[i];
                    if (source === target)
                        continue;
                    // We instantiate the target type parameter constraints into the source types so we can recognize `<T, U extends T>` as the same as `<A, B extends A>`
                    if (!isTypeIdenticalTo(getConstraintFromTypeParameter(source) || unknownType, instantiateType(getConstraintFromTypeParameter(target) || unknownType, mapper)))
                        return false;
                    // We don't compare defaults - we just use the type parameter defaults from the first signature that seems to match.
                    // It might make sense to combine these defaults in the future, but doing so intelligently requires knowing
                    // if the parameter is used covariantly or contravariantly (so we intersect if it's used like a parameter or union if used like a return type)
                    // and, since it's just an inference _default_, just picking one arbitrarily works OK.
                }
                return true;
            }
            function combineUnionThisParam(left, right, mapper) {
                if (!left || !right) {
                    return left || right;
                }
                // A signature `this` type might be a read or a write position... It's very possible that it should be invariant
                // and we should refuse to merge signatures if there are `this` types and they do not match. However, so as to be
                // permissive when calling, for now, we'll intersect the `this` types just like we do for param types in union signatures.
                var thisType = getIntersectionType([getTypeOfSymbol(left), instantiateType(getTypeOfSymbol(right), mapper)]);
                return createSymbolWithType(left, thisType);
            }
            function combineUnionParameters(left, right, mapper) {
                var leftCount = getParameterCount(left);
                var rightCount = getParameterCount(right);
                var longest = leftCount >= rightCount ? left : right;
                var shorter = longest === left ? right : left;
                var longestCount = longest === left ? leftCount : rightCount;
                var eitherHasEffectiveRest = (hasEffectiveRestParameter(left) || hasEffectiveRestParameter(right));
                var needsExtraRestElement = eitherHasEffectiveRest && !hasEffectiveRestParameter(longest);
                var params = new Array(longestCount + (needsExtraRestElement ? 1 : 0));
                for (var i = 0; i < longestCount; i++) {
                    var longestParamType = tryGetTypeAtPosition(longest, i);
                    if (longest === right) {
                        longestParamType = instantiateType(longestParamType, mapper);
                    }
                    var shorterParamType = tryGetTypeAtPosition(shorter, i) || unknownType;
                    if (shorter === right) {
                        shorterParamType = instantiateType(shorterParamType, mapper);
                    }
                    var unionParamType = getIntersectionType([longestParamType, shorterParamType]);
                    var isRestParam = eitherHasEffectiveRest && !needsExtraRestElement && i === (longestCount - 1);
                    var isOptional = i >= getMinArgumentCount(longest) && i >= getMinArgumentCount(shorter);
                    var leftName = i >= leftCount ? undefined : getParameterNameAtPosition(left, i);
                    var rightName = i >= rightCount ? undefined : getParameterNameAtPosition(right, i);
                    var paramName = leftName === rightName ? leftName :
                        !leftName ? rightName :
                            !rightName ? leftName :
                                undefined;
                    var paramSymbol = createSymbol(1 /* FunctionScopedVariable */ | (isOptional && !isRestParam ? 16777216 /* Optional */ : 0), paramName || "arg" + i);
                    paramSymbol.type = isRestParam ? createArrayType(unionParamType) : unionParamType;
                    params[i] = paramSymbol;
                }
                if (needsExtraRestElement) {
                    var restParamSymbol = createSymbol(1 /* FunctionScopedVariable */, "args");
                    restParamSymbol.type = createArrayType(getTypeAtPosition(shorter, longestCount));
                    if (shorter === right) {
                        restParamSymbol.type = instantiateType(restParamSymbol.type, mapper);
                    }
                    params[longestCount] = restParamSymbol;
                }
                return params;
            }
            function combineSignaturesOfUnionMembers(left, right) {
                var typeParams = left.typeParameters || right.typeParameters;
                var paramMapper;
                if (left.typeParameters && right.typeParameters) {
                    paramMapper = createTypeMapper(right.typeParameters, left.typeParameters);
                    // We just use the type parameter defaults from the first signature
                }
                var declaration = left.declaration;
                var params = combineUnionParameters(left, right, paramMapper);
                var thisParam = combineUnionThisParam(left.thisParameter, right.thisParameter, paramMapper);
                var minArgCount = Math.max(left.minArgumentCount, right.minArgumentCount);
                var result = createSignature(declaration, typeParams, thisParam, params, 
                /*resolvedReturnType*/ undefined, 
                /*resolvedTypePredicate*/ undefined, minArgCount, (left.flags | right.flags) & 39 /* PropagatingFlags */);
                result.compositeKind = 1048576 /* Union */;
                result.compositeSignatures = ts.concatenate(left.compositeKind !== 2097152 /* Intersection */ && left.compositeSignatures || [left], [right]);
                if (paramMapper) {
                    result.mapper = left.compositeKind !== 2097152 /* Intersection */ && left.mapper && left.compositeSignatures ? combineTypeMappers(left.mapper, paramMapper) : paramMapper;
                }
                return result;
            }
            function getUnionIndexInfos(types) {
                var sourceInfos = getIndexInfosOfType(types[0]);
                if (sourceInfos) {
                    var result = [];
                    var _loop_11 = function (info) {
                        var indexType = info.keyType;
                        if (ts.every(types, function (t) { return !!getIndexInfoOfType(t, indexType); })) {
                            result.push(createIndexInfo(indexType, getUnionType(ts.map(types, function (t) { return getIndexTypeOfType(t, indexType); })), ts.some(types, function (t) { return getIndexInfoOfType(t, indexType).isReadonly; })));
                        }
                    };
                    for (var _i = 0, sourceInfos_1 = sourceInfos; _i < sourceInfos_1.length; _i++) {
                        var info = sourceInfos_1[_i];
                        _loop_11(info);
                    }
                    return result;
                }
                return ts.emptyArray;
            }
            function resolveUnionTypeMembers(type) {
                // The members and properties collections are empty for union types. To get all properties of a union
                // type use getPropertiesOfType (only the language service uses this).
                var callSignatures = getUnionSignatures(ts.map(type.types, function (t) { return t === globalFunctionType ? [unknownSignature] : getSignaturesOfType(t, 0 /* Call */); }));
                var constructSignatures = getUnionSignatures(ts.map(type.types, function (t) { return getSignaturesOfType(t, 1 /* Construct */); }));
                var indexInfos = getUnionIndexInfos(type.types);
                setStructuredTypeMembers(type, emptySymbols, callSignatures, constructSignatures, indexInfos);
            }
            function intersectTypes(type1, type2) {
                return !type1 ? type2 : !type2 ? type1 : getIntersectionType([type1, type2]);
            }
            function findMixins(types) {
                var constructorTypeCount = ts.countWhere(types, function (t) { return getSignaturesOfType(t, 1 /* Construct */).length > 0; });
                var mixinFlags = ts.map(types, isMixinConstructorType);
                if (constructorTypeCount > 0 && constructorTypeCount === ts.countWhere(mixinFlags, function (b) { return b; })) {
                    var firstMixinIndex = mixinFlags.indexOf(/*searchElement*/ true);
                    mixinFlags[firstMixinIndex] = false;
                }
                return mixinFlags;
            }
            function includeMixinType(type, types, mixinFlags, index) {
                var mixedTypes = [];
                for (var i = 0; i < types.length; i++) {
                    if (i === index) {
                        mixedTypes.push(type);
                    }
                    else if (mixinFlags[i]) {
                        mixedTypes.push(getReturnTypeOfSignature(getSignaturesOfType(types[i], 1 /* Construct */)[0]));
                    }
                }
                return getIntersectionType(mixedTypes);
            }
            function resolveIntersectionTypeMembers(type) {
                // The members and properties collections are empty for intersection types. To get all properties of an
                // intersection type use getPropertiesOfType (only the language service uses this).
                var callSignatures;
                var constructSignatures;
                var indexInfos;
                var types = type.types;
                var mixinFlags = findMixins(types);
                var mixinCount = ts.countWhere(mixinFlags, function (b) { return b; });
                var _loop_12 = function (i) {
                    var t = type.types[i];
                    // When an intersection type contains mixin constructor types, the construct signatures from
                    // those types are discarded and their return types are mixed into the return types of all
                    // other construct signatures in the intersection type. For example, the intersection type
                    // '{ new(...args: any[]) => A } & { new(s: string) => B }' has a single construct signature
                    // 'new(s: string) => A & B'.
                    if (!mixinFlags[i]) {
                        var signatures = getSignaturesOfType(t, 1 /* Construct */);
                        if (signatures.length && mixinCount > 0) {
                            signatures = ts.map(signatures, function (s) {
                                var clone = cloneSignature(s);
                                clone.resolvedReturnType = includeMixinType(getReturnTypeOfSignature(s), types, mixinFlags, i);
                                return clone;
                            });
                        }
                        constructSignatures = appendSignatures(constructSignatures, signatures);
                    }
                    callSignatures = appendSignatures(callSignatures, getSignaturesOfType(t, 0 /* Call */));
                    indexInfos = ts.reduceLeft(getIndexInfosOfType(t), function (infos, newInfo) { return appendIndexInfo(infos, newInfo, /*union*/ false); }, indexInfos);
                };
                for (var i = 0; i < types.length; i++) {
                    _loop_12(i);
                }
                setStructuredTypeMembers(type, emptySymbols, callSignatures || ts.emptyArray, constructSignatures || ts.emptyArray, indexInfos || ts.emptyArray);
            }
            function appendSignatures(signatures, newSignatures) {
                var _loop_13 = function (sig) {
                    if (!signatures || ts.every(signatures, function (s) { return !compareSignaturesIdentical(s, sig, /*partialMatch*/ false, /*ignoreThisTypes*/ false, /*ignoreReturnTypes*/ false, compareTypesIdentical); })) {
                        signatures = ts.append(signatures, sig);
                    }
                };
                for (var _i = 0, newSignatures_1 = newSignatures; _i < newSignatures_1.length; _i++) {
                    var sig = newSignatures_1[_i];
                    _loop_13(sig);
                }
                return signatures;
            }
            function appendIndexInfo(indexInfos, newInfo, union) {
                if (indexInfos) {
                    for (var i = 0; i < indexInfos.length; i++) {
                        var info = indexInfos[i];
                        if (info.keyType === newInfo.keyType) {
                            indexInfos[i] = createIndexInfo(info.keyType, union ? getUnionType([info.type, newInfo.type]) : getIntersectionType([info.type, newInfo.type]), union ? info.isReadonly || newInfo.isReadonly : info.isReadonly && newInfo.isReadonly);
                            return indexInfos;
                        }
                    }
                }
                return ts.append(indexInfos, newInfo);
            }
            /**
             * Converts an AnonymousType to a ResolvedType.
             */
            function resolveAnonymousTypeMembers(type) {
                var symbol = getMergedSymbol(type.symbol);
                if (type.target) {
                    setStructuredTypeMembers(type, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                    var members = createInstantiatedSymbolTable(getPropertiesOfObjectType(type.target), type.mapper, /*mappingThisOnly*/ false);
                    var callSignatures = instantiateSignatures(getSignaturesOfType(type.target, 0 /* Call */), type.mapper);
                    var constructSignatures = instantiateSignatures(getSignaturesOfType(type.target, 1 /* Construct */), type.mapper);
                    var indexInfos = instantiateIndexInfos(getIndexInfosOfType(type.target), type.mapper);
                    setStructuredTypeMembers(type, members, callSignatures, constructSignatures, indexInfos);
                }
                else if (symbol.flags & 2048 /* TypeLiteral */) {
                    setStructuredTypeMembers(type, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                    var members = getMembersOfSymbol(symbol);
                    var callSignatures = getSignaturesOfSymbol(members.get("__call" /* Call */));
                    var constructSignatures = getSignaturesOfSymbol(members.get("__new" /* New */));
                    var indexInfos = getIndexInfosOfSymbol(symbol);
                    setStructuredTypeMembers(type, members, callSignatures, constructSignatures, indexInfos);
                }
                else {
                    // Combinations of function, class, enum and module
                    var members = emptySymbols;
                    var indexInfos = void 0;
                    if (symbol.exports) {
                        members = getExportsOfSymbol(symbol);
                        if (symbol === globalThisSymbol) {
                            var varsOnly_1 = new ts.Map();
                            members.forEach(function (p) {
                                if (!(p.flags & 418 /* BlockScoped */)) {
                                    varsOnly_1.set(p.escapedName, p);
                                }
                            });
                            members = varsOnly_1;
                        }
                    }
                    var baseConstructorIndexInfo = void 0;
                    setStructuredTypeMembers(type, members, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                    if (symbol.flags & 32 /* Class */) {
                        var classType = getDeclaredTypeOfClassOrInterface(symbol);
                        var baseConstructorType = getBaseConstructorTypeOfClass(classType);
                        if (baseConstructorType.flags & (524288 /* Object */ | 2097152 /* Intersection */ | 8650752 /* TypeVariable */)) {
                            members = ts.createSymbolTable(getNamedOrIndexSignatureMembers(members));
                            addInheritedMembers(members, getPropertiesOfType(baseConstructorType));
                        }
                        else if (baseConstructorType === anyType) {
                            baseConstructorIndexInfo = createIndexInfo(stringType, anyType, /*isReadonly*/ false);
                        }
                    }
                    var indexSymbol = getIndexSymbolFromSymbolTable(members);
                    if (indexSymbol) {
                        indexInfos = getIndexInfosOfIndexSymbol(indexSymbol);
                    }
                    else {
                        if (baseConstructorIndexInfo) {
                            indexInfos = ts.append(indexInfos, baseConstructorIndexInfo);
                        }
                        if (symbol.flags & 384 /* Enum */ && (getDeclaredTypeOfSymbol(symbol).flags & 32 /* Enum */ ||
                            ts.some(type.properties, function (prop) { return !!(getTypeOfSymbol(prop).flags & 296 /* NumberLike */); }))) {
                            indexInfos = ts.append(indexInfos, enumNumberIndexInfo);
                        }
                    }
                    setStructuredTypeMembers(type, members, ts.emptyArray, ts.emptyArray, indexInfos || ts.emptyArray);
                    // We resolve the members before computing the signatures because a signature may use
                    // typeof with a qualified name expression that circularly references the type we are
                    // in the process of resolving (see issue #6072). The temporarily empty signature list
                    // will never be observed because a qualified name can't reference signatures.
                    if (symbol.flags & (16 /* Function */ | 8192 /* Method */)) {
                        type.callSignatures = getSignaturesOfSymbol(symbol);
                    }
                    // And likewise for construct signatures for classes
                    if (symbol.flags & 32 /* Class */) {
                        var classType_1 = getDeclaredTypeOfClassOrInterface(symbol);
                        var constructSignatures = symbol.members ? getSignaturesOfSymbol(symbol.members.get("__constructor" /* Constructor */)) : ts.emptyArray;
                        if (symbol.flags & 16 /* Function */) {
                            constructSignatures = ts.addRange(constructSignatures.slice(), ts.mapDefined(type.callSignatures, function (sig) {
                                return isJSConstructor(sig.declaration) ?
                                    createSignature(sig.declaration, sig.typeParameters, sig.thisParameter, sig.parameters, classType_1, /*resolvedTypePredicate*/ undefined, sig.minArgumentCount, sig.flags & 39 /* PropagatingFlags */) :
                                    undefined;
                            }));
                        }
                        if (!constructSignatures.length) {
                            constructSignatures = getDefaultConstructSignatures(classType_1);
                        }
                        type.constructSignatures = constructSignatures;
                    }
                }
            }
            function replaceIndexedAccess(instantiable, type, replacement) {
                // map type.indexType to 0
                // map type.objectType to `[TReplacement]`
                // thus making the indexed access `[TReplacement][0]` or `TReplacement`
                return instantiateType(instantiable, createTypeMapper([type.indexType, type.objectType], [getNumberLiteralType(0), createTupleType([replacement])]));
            }
            function resolveReverseMappedTypeMembers(type) {
                var indexInfo = getIndexInfoOfType(type.source, stringType);
                var modifiers = getMappedTypeModifiers(type.mappedType);
                var readonlyMask = modifiers & 1 /* IncludeReadonly */ ? false : true;
                var optionalMask = modifiers & 4 /* IncludeOptional */ ? 0 : 16777216 /* Optional */;
                var indexInfos = indexInfo ? [createIndexInfo(stringType, inferReverseMappedType(indexInfo.type, type.mappedType, type.constraintType), readonlyMask && indexInfo.isReadonly)] : ts.emptyArray;
                var members = ts.createSymbolTable();
                for (var _i = 0, _a = getPropertiesOfType(type.source); _i < _a.length; _i++) {
                    var prop = _a[_i];
                    var checkFlags = 8192 /* ReverseMapped */ | (readonlyMask && isReadonlySymbol(prop) ? 8 /* Readonly */ : 0);
                    var inferredProp = createSymbol(4 /* Property */ | prop.flags & optionalMask, prop.escapedName, checkFlags);
                    inferredProp.declarations = prop.declarations;
                    inferredProp.nameType = getSymbolLinks(prop).nameType;
                    inferredProp.propertyType = getTypeOfSymbol(prop);
                    if (type.constraintType.type.flags & 8388608 /* IndexedAccess */
                        && type.constraintType.type.objectType.flags & 262144 /* TypeParameter */
                        && type.constraintType.type.indexType.flags & 262144 /* TypeParameter */) {
                        // A reverse mapping of `{[K in keyof T[K_1]]: T[K_1]}` is the same as that of `{[K in keyof T]: T}`, since all we care about is
                        // inferring to the "type parameter" (or indexed access) shared by the constraint and template. So, to reduce the number of
                        // type identities produced, we simplify such indexed access occurences
                        var newTypeParam = type.constraintType.type.objectType;
                        var newMappedType = replaceIndexedAccess(type.mappedType, type.constraintType.type, newTypeParam);
                        inferredProp.mappedType = newMappedType;
                        inferredProp.constraintType = getIndexType(newTypeParam);
                    }
                    else {
                        inferredProp.mappedType = type.mappedType;
                        inferredProp.constraintType = type.constraintType;
                    }
                    members.set(prop.escapedName, inferredProp);
                }
                setStructuredTypeMembers(type, members, ts.emptyArray, ts.emptyArray, indexInfos);
            }
            // Return the lower bound of the key type in a mapped type. Intuitively, the lower
            // bound includes those keys that are known to always be present, for example because
            // because of constraints on type parameters (e.g. 'keyof T' for a constrained T).
            function getLowerBoundOfKeyType(type) {
                if (type.flags & 4194304 /* Index */) {
                    var t = getApparentType(type.type);
                    return isGenericTupleType(t) ? getKnownKeysOfTupleType(t) : getIndexType(t);
                }
                if (type.flags & 16777216 /* Conditional */) {
                    if (type.root.isDistributive) {
                        var checkType = type.checkType;
                        var constraint = getLowerBoundOfKeyType(checkType);
                        if (constraint !== checkType) {
                            return getConditionalTypeInstantiation(type, prependTypeMapping(type.root.checkType, constraint, type.mapper));
                        }
                    }
                    return type;
                }
                if (type.flags & 1048576 /* Union */) {
                    return mapType(type, getLowerBoundOfKeyType);
                }
                if (type.flags & 2097152 /* Intersection */) {
                    return getIntersectionType(ts.sameMap(type.types, getLowerBoundOfKeyType));
                }
                return type;
            }
            function getIsLateCheckFlag(s) {
                return ts.getCheckFlags(s) & 4096 /* Late */;
            }
            /** Resolve the members of a mapped type { [P in K]: T } */
            function resolveMappedTypeMembers(type) {
                var members = ts.createSymbolTable();
                var indexInfos;
                // Resolve upfront such that recursive references see an empty object type.
                setStructuredTypeMembers(type, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                // In { [P in K]: T }, we refer to P as the type parameter type, K as the constraint type,
                // and T as the template type.
                var typeParameter = getTypeParameterFromMappedType(type);
                var constraintType = getConstraintTypeFromMappedType(type);
                var nameType = getNameTypeFromMappedType(type.target || type);
                var templateType = getTemplateTypeFromMappedType(type.target || type);
                var modifiersType = getApparentType(getModifiersTypeFromMappedType(type)); // The 'T' in 'keyof T'
                var templateModifiers = getMappedTypeModifiers(type);
                var include = keyofStringsOnly ? 128 /* StringLiteral */ : 8576 /* StringOrNumberLiteralOrUnique */;
                if (isMappedTypeWithKeyofConstraintDeclaration(type)) {
                    // We have a { [P in keyof T]: X }
                    for (var _i = 0, _a = getPropertiesOfType(modifiersType); _i < _a.length; _i++) {
                        var prop = _a[_i];
                        addMemberForKeyType(getLiteralTypeFromProperty(prop, include));
                    }
                    if (modifiersType.flags & 1 /* Any */) {
                        addMemberForKeyType(stringType);
                    }
                    else {
                        for (var _b = 0, _c = getIndexInfosOfType(modifiersType); _b < _c.length; _b++) {
                            var info = _c[_b];
                            if (!keyofStringsOnly || info.keyType.flags & (4 /* String */ | 134217728 /* TemplateLiteral */)) {
                                addMemberForKeyType(info.keyType);
                            }
                        }
                    }
                }
                else {
                    forEachType(getLowerBoundOfKeyType(constraintType), addMemberForKeyType);
                }
                setStructuredTypeMembers(type, members, ts.emptyArray, ts.emptyArray, indexInfos || ts.emptyArray);
                function addMemberForKeyType(keyType) {
                    var propNameType = nameType ? instantiateType(nameType, appendTypeMapping(type.mapper, typeParameter, keyType)) : keyType;
                    forEachType(propNameType, function (t) { return addMemberForKeyTypeWorker(keyType, t); });
                }
                function addMemberForKeyTypeWorker(keyType, propNameType) {
                    // If the current iteration type constituent is a string literal type, create a property.
                    // Otherwise, for type string create a string index signature.
                    if (isTypeUsableAsPropertyName(propNameType)) {
                        var propName = getPropertyNameFromType(propNameType);
                        // String enum members from separate enums with identical values
                        // are distinct types with the same property name. Make the resulting
                        // property symbol's name type be the union of those enum member types.
                        var existingProp = members.get(propName);
                        if (existingProp) {
                            existingProp.nameType = getUnionType([existingProp.nameType, propNameType]);
                            existingProp.keyType = getUnionType([existingProp.keyType, keyType]);
                        }
                        else {
                            var modifiersProp = isTypeUsableAsPropertyName(keyType) ? getPropertyOfType(modifiersType, getPropertyNameFromType(keyType)) : undefined;
                            var isOptional = !!(templateModifiers & 4 /* IncludeOptional */ ||
                                !(templateModifiers & 8 /* ExcludeOptional */) && modifiersProp && modifiersProp.flags & 16777216 /* Optional */);
                            var isReadonly = !!(templateModifiers & 1 /* IncludeReadonly */ ||
                                !(templateModifiers & 2 /* ExcludeReadonly */) && modifiersProp && isReadonlySymbol(modifiersProp));
                            var stripOptional = strictNullChecks && !isOptional && modifiersProp && modifiersProp.flags & 16777216 /* Optional */;
                            var lateFlag = modifiersProp ? getIsLateCheckFlag(modifiersProp) : 0;
                            var prop = createSymbol(4 /* Property */ | (isOptional ? 16777216 /* Optional */ : 0), propName, lateFlag | 262144 /* Mapped */ | (isReadonly ? 8 /* Readonly */ : 0) | (stripOptional ? 524288 /* StripOptional */ : 0));
                            prop.mappedType = type;
                            prop.nameType = propNameType;
                            prop.keyType = keyType;
                            if (modifiersProp) {
                                prop.syntheticOrigin = modifiersProp;
                                // If the mapped type has an `as XXX` clause, the property name likely won't match the declaration name and
                                // multiple properties may map to the same name. Thus, we attach no declarations to the symbol.
                                prop.declarations = nameType ? undefined : modifiersProp.declarations;
                            }
                            members.set(propName, prop);
                        }
                    }
                    else if (isValidIndexKeyType(propNameType) || propNameType.flags & (1 /* Any */ | 32 /* Enum */)) {
                        var indexKeyType = propNameType.flags & (1 /* Any */ | 4 /* String */) ? stringType :
                            propNameType.flags & (8 /* Number */ | 32 /* Enum */) ? numberType :
                                propNameType;
                        var propType = instantiateType(templateType, appendTypeMapping(type.mapper, typeParameter, keyType));
                        var indexInfo = createIndexInfo(indexKeyType, propType, !!(templateModifiers & 1 /* IncludeReadonly */));
                        indexInfos = appendIndexInfo(indexInfos, indexInfo, /*union*/ true);
                    }
                }
            }
            function getTypeOfMappedSymbol(symbol) {
                if (!symbol.type) {
                    var mappedType = symbol.mappedType;
                    if (!pushTypeResolution(symbol, 0 /* Type */)) {
                        mappedType.containsError = true;
                        return errorType;
                    }
                    var templateType = getTemplateTypeFromMappedType(mappedType.target || mappedType);
                    var mapper = appendTypeMapping(mappedType.mapper, getTypeParameterFromMappedType(mappedType), symbol.keyType);
                    var propType = instantiateType(templateType, mapper);
                    // When creating an optional property in strictNullChecks mode, if 'undefined' isn't assignable to the
                    // type, we include 'undefined' in the type. Similarly, when creating a non-optional property in strictNullChecks
                    // mode, if the underlying property is optional we remove 'undefined' from the type.
                    var type = strictNullChecks && symbol.flags & 16777216 /* Optional */ && !maybeTypeOfKind(propType, 32768 /* Undefined */ | 16384 /* Void */) ? getOptionalType(propType, /*isProperty*/ true) :
                        symbol.checkFlags & 524288 /* StripOptional */ ? removeMissingOrUndefinedType(propType) :
                            propType;
                    if (!popTypeResolution()) {
                        error(currentNode, ts.Diagnostics.Type_of_property_0_circularly_references_itself_in_mapped_type_1, symbolToString(symbol), typeToString(mappedType));
                        type = errorType;
                    }
                    symbol.type = type;
                }
                return symbol.type;
            }
            function getTypeParameterFromMappedType(type) {
                return type.typeParameter ||
                    (type.typeParameter = getDeclaredTypeOfTypeParameter(getSymbolOfNode(type.declaration.typeParameter)));
            }
            function getConstraintTypeFromMappedType(type) {
                return type.constraintType ||
                    (type.constraintType = getConstraintOfTypeParameter(getTypeParameterFromMappedType(type)) || errorType);
            }
            function getNameTypeFromMappedType(type) {
                return type.declaration.nameType ?
                    type.nameType || (type.nameType = instantiateType(getTypeFromTypeNode(type.declaration.nameType), type.mapper)) :
                    undefined;
            }
            function getTemplateTypeFromMappedType(type) {
                return type.templateType ||
                    (type.templateType = type.declaration.type ?
                        instantiateType(addOptionality(getTypeFromTypeNode(type.declaration.type), /*isProperty*/ true, !!(getMappedTypeModifiers(type) & 4 /* IncludeOptional */)), type.mapper) :
                        errorType);
            }
            function getConstraintDeclarationForMappedType(type) {
                return ts.getEffectiveConstraintOfTypeParameter(type.declaration.typeParameter);
            }
            function isMappedTypeWithKeyofConstraintDeclaration(type) {
                var constraintDeclaration = getConstraintDeclarationForMappedType(type); // TODO: GH#18217
                return constraintDeclaration.kind === 191 /* TypeOperator */ &&
                    constraintDeclaration.operator === 139 /* KeyOfKeyword */;
            }
            function getModifiersTypeFromMappedType(type) {
                if (!type.modifiersType) {
                    if (isMappedTypeWithKeyofConstraintDeclaration(type)) {
                        // If the constraint declaration is a 'keyof T' node, the modifiers type is T. We check
                        // AST nodes here because, when T is a non-generic type, the logic below eagerly resolves
                        // 'keyof T' to a literal union type and we can't recover T from that type.
                        type.modifiersType = instantiateType(getTypeFromTypeNode(getConstraintDeclarationForMappedType(type).type), type.mapper);
                    }
                    else {
                        // Otherwise, get the declared constraint type, and if the constraint type is a type parameter,
                        // get the constraint of that type parameter. If the resulting type is an indexed type 'keyof T',
                        // the modifiers type is T. Otherwise, the modifiers type is unknown.
                        var declaredType = getTypeFromMappedTypeNode(type.declaration);
                        var constraint = getConstraintTypeFromMappedType(declaredType);
                        var extendedConstraint = constraint && constraint.flags & 262144 /* TypeParameter */ ? getConstraintOfTypeParameter(constraint) : constraint;
                        type.modifiersType = extendedConstraint && extendedConstraint.flags & 4194304 /* Index */ ? instantiateType(extendedConstraint.type, type.mapper) : unknownType;
                    }
                }
                return type.modifiersType;
            }
            function getMappedTypeModifiers(type) {
                var declaration = type.declaration;
                return (declaration.readonlyToken ? declaration.readonlyToken.kind === 40 /* MinusToken */ ? 2 /* ExcludeReadonly */ : 1 /* IncludeReadonly */ : 0) |
                    (declaration.questionToken ? declaration.questionToken.kind === 40 /* MinusToken */ ? 8 /* ExcludeOptional */ : 4 /* IncludeOptional */ : 0);
            }
            function getMappedTypeOptionality(type) {
                var modifiers = getMappedTypeModifiers(type);
                return modifiers & 8 /* ExcludeOptional */ ? -1 : modifiers & 4 /* IncludeOptional */ ? 1 : 0;
            }
            function getCombinedMappedTypeOptionality(type) {
                var optionality = getMappedTypeOptionality(type);
                var modifiersType = getModifiersTypeFromMappedType(type);
                return optionality || (isGenericMappedType(modifiersType) ? getMappedTypeOptionality(modifiersType) : 0);
            }
            function isPartialMappedType(type) {
                return !!(ts.getObjectFlags(type) & 32 /* Mapped */ && getMappedTypeModifiers(type) & 4 /* IncludeOptional */);
            }
            function isGenericMappedType(type) {
                return !!(ts.getObjectFlags(type) & 32 /* Mapped */) && isGenericIndexType(getConstraintTypeFromMappedType(type));
            }
            function resolveStructuredTypeMembers(type) {
                if (!type.members) {
                    if (type.flags & 524288 /* Object */) {
                        if (type.objectFlags & 4 /* Reference */) {
                            resolveTypeReferenceMembers(type);
                        }
                        else if (type.objectFlags & 3 /* ClassOrInterface */) {
                            resolveClassOrInterfaceMembers(type);
                        }
                        else if (type.objectFlags & 1024 /* ReverseMapped */) {
                            resolveReverseMappedTypeMembers(type);
                        }
                        else if (type.objectFlags & 16 /* Anonymous */) {
                            resolveAnonymousTypeMembers(type);
                        }
                        else if (type.objectFlags & 32 /* Mapped */) {
                            resolveMappedTypeMembers(type);
                        }
                    }
                    else if (type.flags & 1048576 /* Union */) {
                        resolveUnionTypeMembers(type);
                    }
                    else if (type.flags & 2097152 /* Intersection */) {
                        resolveIntersectionTypeMembers(type);
                    }
                }
                return type;
            }
            /** Return properties of an object type or an empty array for other types */
            function getPropertiesOfObjectType(type) {
                if (type.flags & 524288 /* Object */) {
                    return resolveStructuredTypeMembers(type).properties;
                }
                return ts.emptyArray;
            }
            /** If the given type is an object type and that type has a property by the given name,
             * return the symbol for that property. Otherwise return undefined.
             */
            function getPropertyOfObjectType(type, name) {
                if (type.flags & 524288 /* Object */) {
                    var resolved = resolveStructuredTypeMembers(type);
                    var symbol = resolved.members.get(name);
                    if (symbol && symbolIsValue(symbol)) {
                        return symbol;
                    }
                }
            }
            function getPropertiesOfUnionOrIntersectionType(type) {
                if (!type.resolvedProperties) {
                    var members = ts.createSymbolTable();
                    for (var _i = 0, _a = type.types; _i < _a.length; _i++) {
                        var current = _a[_i];
                        for (var _b = 0, _c = getPropertiesOfType(current); _b < _c.length; _b++) {
                            var prop = _c[_b];
                            if (!members.has(prop.escapedName)) {
                                var combinedProp = getPropertyOfUnionOrIntersectionType(type, prop.escapedName);
                                if (combinedProp) {
                                    members.set(prop.escapedName, combinedProp);
                                }
                            }
                        }
                        // The properties of a union type are those that are present in all constituent types, so
                        // we only need to check the properties of the first type without index signature
                        if (type.flags & 1048576 /* Union */ && getIndexInfosOfType(current).length === 0) {
                            break;
                        }
                    }
                    type.resolvedProperties = getNamedMembers(members);
                }
                return type.resolvedProperties;
            }
            function getPropertiesOfType(type) {
                type = getReducedApparentType(type);
                return type.flags & 3145728 /* UnionOrIntersection */ ?
                    getPropertiesOfUnionOrIntersectionType(type) :
                    getPropertiesOfObjectType(type);
            }
            function isTypeInvalidDueToUnionDiscriminant(contextualType, obj) {
                var list = obj.properties;
                return list.some(function (property) {
                    var nameType = property.name && getLiteralTypeFromPropertyName(property.name);
                    var name = nameType && isTypeUsableAsPropertyName(nameType) ? getPropertyNameFromType(nameType) : undefined;
                    var expected = name === undefined ? undefined : getTypeOfPropertyOfType(contextualType, name);
                    return !!expected && isLiteralType(expected) && !isTypeAssignableTo(getTypeOfNode(property), expected);
                });
            }
            function getAllPossiblePropertiesOfTypes(types) {
                var unionType = getUnionType(types);
                if (!(unionType.flags & 1048576 /* Union */)) {
                    return getAugmentedPropertiesOfType(unionType);
                }
                var props = ts.createSymbolTable();
                for (var _i = 0, types_4 = types; _i < types_4.length; _i++) {
                    var memberType = types_4[_i];
                    for (var _a = 0, _b = getAugmentedPropertiesOfType(memberType); _a < _b.length; _a++) {
                        var escapedName = _b[_a].escapedName;
                        if (!props.has(escapedName)) {
                            var prop = createUnionOrIntersectionProperty(unionType, escapedName);
                            // May be undefined if the property is private
                            if (prop)
                                props.set(escapedName, prop);
                        }
                    }
                }
                return ts.arrayFrom(props.values());
            }
            function getConstraintOfType(type) {
                return type.flags & 262144 /* TypeParameter */ ? getConstraintOfTypeParameter(type) :
                    type.flags & 8388608 /* IndexedAccess */ ? getConstraintOfIndexedAccess(type) :
                        type.flags & 16777216 /* Conditional */ ? getConstraintOfConditionalType(type) :
                            getBaseConstraintOfType(type);
            }
            function getConstraintOfTypeParameter(typeParameter) {
                return hasNonCircularBaseConstraint(typeParameter) ? getConstraintFromTypeParameter(typeParameter) : undefined;
            }
            function getConstraintOfIndexedAccess(type) {
                return hasNonCircularBaseConstraint(type) ? getConstraintFromIndexedAccess(type) : undefined;
            }
            function getSimplifiedTypeOrConstraint(type) {
                var simplified = getSimplifiedType(type, /*writing*/ false);
                return simplified !== type ? simplified : getConstraintOfType(type);
            }
            function getConstraintFromIndexedAccess(type) {
                var indexConstraint = getSimplifiedTypeOrConstraint(type.indexType);
                if (indexConstraint && indexConstraint !== type.indexType) {
                    var indexedAccess = getIndexedAccessTypeOrUndefined(type.objectType, indexConstraint, type.accessFlags);
                    if (indexedAccess) {
                        return indexedAccess;
                    }
                }
                var objectConstraint = getSimplifiedTypeOrConstraint(type.objectType);
                if (objectConstraint && objectConstraint !== type.objectType) {
                    return getIndexedAccessTypeOrUndefined(objectConstraint, type.indexType, type.accessFlags);
                }
                return undefined;
            }
            function getDefaultConstraintOfConditionalType(type) {
                if (!type.resolvedDefaultConstraint) {
                    // An `any` branch of a conditional type would normally be viral - specifically, without special handling here,
                    // a conditional type with a single branch of type `any` would be assignable to anything, since it's constraint would simplify to
                    // just `any`. This result is _usually_ unwanted - so instead here we elide an `any` branch from the constraint type,
                    // in effect treating `any` like `never` rather than `unknown` in this location.
                    var trueConstraint = getInferredTrueTypeFromConditionalType(type);
                    var falseConstraint = getFalseTypeFromConditionalType(type);
                    type.resolvedDefaultConstraint = isTypeAny(trueConstraint) ? falseConstraint : isTypeAny(falseConstraint) ? trueConstraint : getUnionType([trueConstraint, falseConstraint]);
                }
                return type.resolvedDefaultConstraint;
            }
            function getConstraintOfDistributiveConditionalType(type) {
                // Check if we have a conditional type of the form 'T extends U ? X : Y', where T is a constrained
                // type parameter. If so, create an instantiation of the conditional type where T is replaced
                // with its constraint. We do this because if the constraint is a union type it will be distributed
                // over the conditional type and possibly reduced. For example, 'T extends undefined ? never : T'
                // removes 'undefined' from T.
                // We skip returning a distributive constraint for a restrictive instantiation of a conditional type
                // as the constraint for all type params (check type included) have been replace with `unknown`, which
                // is going to produce even more false positive/negative results than the distribute constraint already does.
                // Please note: the distributive constraint is a kludge for emulating what a negated type could to do filter
                // a union - once negated types exist and are applied to the conditional false branch, this "constraint"
                // likely doesn't need to exist.
                if (type.root.isDistributive && type.restrictiveInstantiation !== type) {
                    var simplified = getSimplifiedType(type.checkType, /*writing*/ false);
                    var constraint = simplified === type.checkType ? getConstraintOfType(simplified) : simplified;
                    if (constraint && constraint !== type.checkType) {
                        var instantiated = getConditionalTypeInstantiation(type, prependTypeMapping(type.root.checkType, constraint, type.mapper));
                        if (!(instantiated.flags & 131072 /* Never */)) {
                            return instantiated;
                        }
                    }
                }
                return undefined;
            }
            function getConstraintFromConditionalType(type) {
                return getConstraintOfDistributiveConditionalType(type) || getDefaultConstraintOfConditionalType(type);
            }
            function getConstraintOfConditionalType(type) {
                return hasNonCircularBaseConstraint(type) ? getConstraintFromConditionalType(type) : undefined;
            }
            function getEffectiveConstraintOfIntersection(types, targetIsUnion) {
                var constraints;
                var hasDisjointDomainType = false;
                for (var _i = 0, types_5 = types; _i < types_5.length; _i++) {
                    var t = types_5[_i];
                    if (t.flags & 465829888 /* Instantiable */) {
                        // We keep following constraints as long as we have an instantiable type that is known
                        // not to be circular or infinite (hence we stop on index access types).
                        var constraint = getConstraintOfType(t);
                        while (constraint && constraint.flags & (262144 /* TypeParameter */ | 4194304 /* Index */ | 16777216 /* Conditional */)) {
                            constraint = getConstraintOfType(constraint);
                        }
                        if (constraint) {
                            constraints = ts.append(constraints, constraint);
                            if (targetIsUnion) {
                                constraints = ts.append(constraints, t);
                            }
                        }
                    }
                    else if (t.flags & 469892092 /* DisjointDomains */) {
                        hasDisjointDomainType = true;
                    }
                }
                // If the target is a union type or if we are intersecting with types belonging to one of the
                // disjoint domains, we may end up producing a constraint that hasn't been examined before.
                if (constraints && (targetIsUnion || hasDisjointDomainType)) {
                    if (hasDisjointDomainType) {
                        // We add any types belong to one of the disjoint domains because they might cause the final
                        // intersection operation to reduce the union constraints.
                        for (var _a = 0, types_6 = types; _a < types_6.length; _a++) {
                            var t = types_6[_a];
                            if (t.flags & 469892092 /* DisjointDomains */) {
                                constraints = ts.append(constraints, t);
                            }
                        }
                    }
                    return getIntersectionType(constraints);
                }
                return undefined;
            }
            function getBaseConstraintOfType(type) {
                if (type.flags & (58982400 /* InstantiableNonPrimitive */ | 3145728 /* UnionOrIntersection */ | 134217728 /* TemplateLiteral */ | 268435456 /* StringMapping */)) {
                    var constraint = getResolvedBaseConstraint(type);
                    return constraint !== noConstraintType && constraint !== circularConstraintType ? constraint : undefined;
                }
                return type.flags & 4194304 /* Index */ ? keyofConstraintType : undefined;
            }
            /**
             * This is similar to `getBaseConstraintOfType` except it returns the input type if there's no base constraint, instead of `undefined`
             * It also doesn't map indexes to `string`, as where this is used this would be unneeded (and likely undesirable)
             */
            function getBaseConstraintOrType(type) {
                return getBaseConstraintOfType(type) || type;
            }
            function hasNonCircularBaseConstraint(type) {
                return getResolvedBaseConstraint(type) !== circularConstraintType;
            }
            /**
             * Return the resolved base constraint of a type variable. The noConstraintType singleton is returned if the
             * type variable has no constraint, and the circularConstraintType singleton is returned if the constraint
             * circularly references the type variable.
             */
            function getResolvedBaseConstraint(type) {
                if (type.resolvedBaseConstraint) {
                    return type.resolvedBaseConstraint;
                }
                var stack = [];
                return type.resolvedBaseConstraint = getTypeWithThisArgument(getImmediateBaseConstraint(type), type);
                function getImmediateBaseConstraint(t) {
                    if (!t.immediateBaseConstraint) {
                        if (!pushTypeResolution(t, 4 /* ImmediateBaseConstraint */)) {
                            return circularConstraintType;
                        }
                        var result = void 0;
                        // We always explore at least 10 levels of nested constraints. Thereafter, we continue to explore
                        // up to 50 levels of nested constraints provided there are no "deeply nested" types on the stack
                        // (i.e. no types for which five instantiations have been recorded on the stack). If we reach 50
                        // levels of nesting, we are presumably exploring a repeating pattern with a long cycle that hasn't
                        // yet triggered the deeply nested limiter. We have no test cases that actually get to 50 levels of
                        // nesting, so it is effectively just a safety stop.
                        if (stack.length < 10 || stack.length < 50 && !isDeeplyNestedType(t, stack, stack.length)) {
                            stack.push(t);
                            result = computeBaseConstraint(getSimplifiedType(t, /*writing*/ false));
                            stack.pop();
                        }
                        if (!popTypeResolution()) {
                            if (t.flags & 262144 /* TypeParameter */) {
                                var errorNode = getConstraintDeclaration(t);
                                if (errorNode) {
                                    var diagnostic = error(errorNode, ts.Diagnostics.Type_parameter_0_has_a_circular_constraint, typeToString(t));
                                    if (currentNode && !ts.isNodeDescendantOf(errorNode, currentNode) && !ts.isNodeDescendantOf(currentNode, errorNode)) {
                                        ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(currentNode, ts.Diagnostics.Circularity_originates_in_type_at_this_location));
                                    }
                                }
                            }
                            result = circularConstraintType;
                        }
                        t.immediateBaseConstraint = result || noConstraintType;
                    }
                    return t.immediateBaseConstraint;
                }
                function getBaseConstraint(t) {
                    var c = getImmediateBaseConstraint(t);
                    return c !== noConstraintType && c !== circularConstraintType ? c : undefined;
                }
                function computeBaseConstraint(t) {
                    if (t.flags & 262144 /* TypeParameter */) {
                        var constraint = getConstraintFromTypeParameter(t);
                        return t.isThisType || !constraint ?
                            constraint :
                            getBaseConstraint(constraint);
                    }
                    if (t.flags & 3145728 /* UnionOrIntersection */) {
                        var types = t.types;
                        var baseTypes = [];
                        var different = false;
                        for (var _i = 0, types_7 = types; _i < types_7.length; _i++) {
                            var type_3 = types_7[_i];
                            var baseType = getBaseConstraint(type_3);
                            if (baseType) {
                                if (baseType !== type_3) {
                                    different = true;
                                }
                                baseTypes.push(baseType);
                            }
                            else {
                                different = true;
                            }
                        }
                        if (!different) {
                            return t;
                        }
                        return t.flags & 1048576 /* Union */ && baseTypes.length === types.length ? getUnionType(baseTypes) :
                            t.flags & 2097152 /* Intersection */ && baseTypes.length ? getIntersectionType(baseTypes) :
                                undefined;
                    }
                    if (t.flags & 4194304 /* Index */) {
                        return keyofConstraintType;
                    }
                    if (t.flags & 134217728 /* TemplateLiteral */) {
                        var types = t.types;
                        var constraints = ts.mapDefined(types, getBaseConstraint);
                        return constraints.length === types.length ? getTemplateLiteralType(t.texts, constraints) : stringType;
                    }
                    if (t.flags & 268435456 /* StringMapping */) {
                        var constraint = getBaseConstraint(t.type);
                        return constraint ? getStringMappingType(t.symbol, constraint) : stringType;
                    }
                    if (t.flags & 8388608 /* IndexedAccess */) {
                        var baseObjectType = getBaseConstraint(t.objectType);
                        var baseIndexType = getBaseConstraint(t.indexType);
                        var baseIndexedAccess = baseObjectType && baseIndexType && getIndexedAccessTypeOrUndefined(baseObjectType, baseIndexType, t.accessFlags);
                        return baseIndexedAccess && getBaseConstraint(baseIndexedAccess);
                    }
                    if (t.flags & 16777216 /* Conditional */) {
                        var constraint = getConstraintFromConditionalType(t);
                        return constraint && getBaseConstraint(constraint);
                    }
                    if (t.flags & 33554432 /* Substitution */) {
                        return getBaseConstraint(t.substitute);
                    }
                    return t;
                }
            }
            function getApparentTypeOfIntersectionType(type) {
                return type.resolvedApparentType || (type.resolvedApparentType = getTypeWithThisArgument(type, type, /*apparentType*/ true));
            }
            function getResolvedTypeParameterDefault(typeParameter) {
                if (!typeParameter.default) {
                    if (typeParameter.target) {
                        var targetDefault = getResolvedTypeParameterDefault(typeParameter.target);
                        typeParameter.default = targetDefault ? instantiateType(targetDefault, typeParameter.mapper) : noConstraintType;
                    }
                    else {
                        // To block recursion, set the initial value to the resolvingDefaultType.
                        typeParameter.default = resolvingDefaultType;
                        var defaultDeclaration = typeParameter.symbol && ts.forEach(typeParameter.symbol.declarations, function (decl) { return ts.isTypeParameterDeclaration(decl) && decl.default; });
                        var defaultType = defaultDeclaration ? getTypeFromTypeNode(defaultDeclaration) : noConstraintType;
                        if (typeParameter.default === resolvingDefaultType) {
                            // If we have not been called recursively, set the correct default type.
                            typeParameter.default = defaultType;
                        }
                    }
                }
                else if (typeParameter.default === resolvingDefaultType) {
                    // If we are called recursively for this type parameter, mark the default as circular.
                    typeParameter.default = circularConstraintType;
                }
                return typeParameter.default;
            }
            /**
             * Gets the default type for a type parameter.
             *
             * If the type parameter is the result of an instantiation, this gets the instantiated
             * default type of its target. If the type parameter has no default type or the default is
             * circular, `undefined` is returned.
             */
            function getDefaultFromTypeParameter(typeParameter) {
                var defaultType = getResolvedTypeParameterDefault(typeParameter);
                return defaultType !== noConstraintType && defaultType !== circularConstraintType ? defaultType : undefined;
            }
            function hasNonCircularTypeParameterDefault(typeParameter) {
                return getResolvedTypeParameterDefault(typeParameter) !== circularConstraintType;
            }
            /**
             * Indicates whether the declaration of a typeParameter has a default type.
             */
            function hasTypeParameterDefault(typeParameter) {
                return !!(typeParameter.symbol && ts.forEach(typeParameter.symbol.declarations, function (decl) { return ts.isTypeParameterDeclaration(decl) && decl.default; }));
            }
            function getApparentTypeOfMappedType(type) {
                return type.resolvedApparentType || (type.resolvedApparentType = getResolvedApparentTypeOfMappedType(type));
            }
            function getResolvedApparentTypeOfMappedType(type) {
                var typeVariable = getHomomorphicTypeVariable(type);
                if (typeVariable && !type.declaration.nameType) {
                    var constraint = getConstraintOfTypeParameter(typeVariable);
                    if (constraint && (isArrayType(constraint) || isTupleType(constraint))) {
                        return instantiateType(type, prependTypeMapping(typeVariable, constraint, type.mapper));
                    }
                }
                return type;
            }
            /**
             * For a type parameter, return the base constraint of the type parameter. For the string, number,
             * boolean, and symbol primitive types, return the corresponding object types. Otherwise return the
             * type itself.
             */
            function getApparentType(type) {
                var t = type.flags & 465829888 /* Instantiable */ ? getBaseConstraintOfType(type) || unknownType : type;
                return ts.getObjectFlags(t) & 32 /* Mapped */ ? getApparentTypeOfMappedType(t) :
                    t.flags & 2097152 /* Intersection */ ? getApparentTypeOfIntersectionType(t) :
                        t.flags & 402653316 /* StringLike */ ? globalStringType :
                            t.flags & 296 /* NumberLike */ ? globalNumberType :
                                t.flags & 2112 /* BigIntLike */ ? getGlobalBigIntType(/*reportErrors*/ languageVersion >= 7 /* ES2020 */) :
                                    t.flags & 528 /* BooleanLike */ ? globalBooleanType :
                                        t.flags & 12288 /* ESSymbolLike */ ? getGlobalESSymbolType(/*reportErrors*/ languageVersion >= 2 /* ES2015 */) :
                                            t.flags & 67108864 /* NonPrimitive */ ? emptyObjectType :
                                                t.flags & 4194304 /* Index */ ? keyofConstraintType :
                                                    t.flags & 2 /* Unknown */ && !strictNullChecks ? emptyObjectType :
                                                        t;
            }
            function getReducedApparentType(type) {
                // Since getApparentType may return a non-reduced union or intersection type, we need to perform
                // type reduction both before and after obtaining the apparent type. For example, given a type parameter
                // 'T extends A | B', the type 'T & X' becomes 'A & X | B & X' after obtaining the apparent type, and
                // that type may need further reduction to remove empty intersections.
                return getReducedType(getApparentType(getReducedType(type)));
            }
            function createUnionOrIntersectionProperty(containingType, name, skipObjectFunctionPropertyAugment) {
                var _a, _b;
                var singleProp;
                var propSet;
                var indexTypes;
                var isUnion = containingType.flags & 1048576 /* Union */;
                // Flags we want to propagate to the result if they exist in all source symbols
                var optionalFlag = isUnion ? 0 /* None */ : 16777216 /* Optional */;
                var syntheticFlag = 4 /* SyntheticMethod */;
                var checkFlags = 0;
                var mergedInstantiations = false;
                for (var _i = 0, _c = containingType.types; _i < _c.length; _i++) {
                    var current = _c[_i];
                    var type = getApparentType(current);
                    if (!(type === errorType || type.flags & 131072 /* Never */)) {
                        var prop = getPropertyOfType(type, name, skipObjectFunctionPropertyAugment);
                        var modifiers = prop ? ts.getDeclarationModifierFlagsFromSymbol(prop) : 0;
                        if (prop) {
                            if (isUnion) {
                                optionalFlag |= (prop.flags & 16777216 /* Optional */);
                            }
                            else {
                                optionalFlag &= prop.flags;
                            }
                            if (!singleProp) {
                                singleProp = prop;
                            }
                            else if (prop !== singleProp) {
                                var isInstantiation = (getTargetSymbol(prop) || prop) === (getTargetSymbol(singleProp) || singleProp);
                                // If the symbols are instances of one another with identical types - consider the symbols
                                // equivalent and just use the first one, which thus allows us to avoid eliding private
                                // members when intersecting a (this-)instantiations of a class with it's raw base or another instance
                                if (isInstantiation && compareProperties(singleProp, prop, function (a, b) { return a === b ? -1 /* True */ : 0 /* False */; }) === -1 /* True */) {
                                    // If we merged instantiations of a generic type, we replicate the symbol parent resetting behavior we used
                                    // to do when we recorded multiple distinct symbols so that we still get, eg, `Array<T>.length` printed
                                    // back and not `Array<string>.length` when we're looking at a `.length` access on a `string[] | number[]`
                                    mergedInstantiations = !!singleProp.parent && !!ts.length(getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(singleProp.parent));
                                }
                                else {
                                    if (!propSet) {
                                        propSet = new ts.Map();
                                        propSet.set(getSymbolId(singleProp), singleProp);
                                    }
                                    var id = getSymbolId(prop);
                                    if (!propSet.has(id)) {
                                        propSet.set(id, prop);
                                    }
                                }
                            }
                            checkFlags |= (isReadonlySymbol(prop) ? 8 /* Readonly */ : 0) |
                                (!(modifiers & 24 /* NonPublicAccessibilityModifier */) ? 256 /* ContainsPublic */ : 0) |
                                (modifiers & 16 /* Protected */ ? 512 /* ContainsProtected */ : 0) |
                                (modifiers & 8 /* Private */ ? 1024 /* ContainsPrivate */ : 0) |
                                (modifiers & 32 /* Static */ ? 2048 /* ContainsStatic */ : 0);
                            if (!isPrototypeProperty(prop)) {
                                syntheticFlag = 2 /* SyntheticProperty */;
                            }
                        }
                        else if (isUnion) {
                            var indexInfo = !isLateBoundName(name) && getApplicableIndexInfoForName(type, name);
                            if (indexInfo) {
                                checkFlags |= 32 /* WritePartial */ | (indexInfo.isReadonly ? 8 /* Readonly */ : 0);
                                indexTypes = ts.append(indexTypes, isTupleType(type) ? getRestTypeOfTupleType(type) || undefinedType : indexInfo.type);
                            }
                            else if (isObjectLiteralType(type) && !(ts.getObjectFlags(type) & 4194304 /* ContainsSpread */)) {
                                checkFlags |= 32 /* WritePartial */;
                                indexTypes = ts.append(indexTypes, undefinedType);
                            }
                            else {
                                checkFlags |= 16 /* ReadPartial */;
                            }
                        }
                    }
                }
                if (!singleProp || isUnion && (propSet || checkFlags & 48 /* Partial */) && checkFlags & (1024 /* ContainsPrivate */ | 512 /* ContainsProtected */)) {
                    // No property was found, or, in a union, a property has a private or protected declaration in one
                    // constituent, but is missing or has a different declaration in another constituent.
                    return undefined;
                }
                if (!propSet && !(checkFlags & 16 /* ReadPartial */) && !indexTypes) {
                    if (mergedInstantiations) {
                        // No symbol from a union/intersection should have a `.parent` set (since unions/intersections don't act as symbol parents)
                        // Unless that parent is "reconstituted" from the "first value declaration" on the symbol (which is likely different than its instantiated parent!)
                        // They also have a `.containingType` set, which affects some services endpoints behavior, like `getRootSymbol`
                        var clone_1 = createSymbolWithType(singleProp, singleProp.type);
                        clone_1.parent = (_b = (_a = singleProp.valueDeclaration) === null || _a === void 0 ? void 0 : _a.symbol) === null || _b === void 0 ? void 0 : _b.parent;
                        clone_1.containingType = containingType;
                        clone_1.mapper = singleProp.mapper;
                        return clone_1;
                    }
                    else {
                        return singleProp;
                    }
                }
                var props = propSet ? ts.arrayFrom(propSet.values()) : [singleProp];
                var declarations;
                var firstType;
                var nameType;
                var propTypes = [];
                var firstValueDeclaration;
                var hasNonUniformValueDeclaration = false;
                for (var _d = 0, props_1 = props; _d < props_1.length; _d++) {
                    var prop = props_1[_d];
                    if (!firstValueDeclaration) {
                        firstValueDeclaration = prop.valueDeclaration;
                    }
                    else if (prop.valueDeclaration && prop.valueDeclaration !== firstValueDeclaration) {
                        hasNonUniformValueDeclaration = true;
                    }
                    declarations = ts.addRange(declarations, prop.declarations);
                    var type = getTypeOfSymbol(prop);
                    if (!firstType) {
                        firstType = type;
                        nameType = getSymbolLinks(prop).nameType;
                    }
                    else if (type !== firstType) {
                        checkFlags |= 64 /* HasNonUniformType */;
                    }
                    if (isLiteralType(type)) {
                        checkFlags |= 128 /* HasLiteralType */;
                    }
                    if (type.flags & 131072 /* Never */) {
                        checkFlags |= 131072 /* HasNeverType */;
                    }
                    propTypes.push(type);
                }
                ts.addRange(propTypes, indexTypes);
                var result = createSymbol(4 /* Property */ | optionalFlag, name, syntheticFlag | checkFlags);
                result.containingType = containingType;
                if (!hasNonUniformValueDeclaration && firstValueDeclaration) {
                    result.valueDeclaration = firstValueDeclaration;
                    // Inherit information about parent type.
                    if (firstValueDeclaration.symbol.parent) {
                        result.parent = firstValueDeclaration.symbol.parent;
                    }
                }
                result.declarations = declarations;
                result.nameType = nameType;
                if (propTypes.length > 2) {
                    // When `propTypes` has the potential to explode in size when normalized, defer normalization until absolutely needed
                    result.checkFlags |= 65536 /* DeferredType */;
                    result.deferralParent = containingType;
                    result.deferralConstituents = propTypes;
                }
                else {
                    result.type = isUnion ? getUnionType(propTypes) : getIntersectionType(propTypes);
                }
                return result;
            }
            // Return the symbol for a given property in a union or intersection type, or undefined if the property
            // does not exist in any constituent type. Note that the returned property may only be present in some
            // constituents, in which case the isPartial flag is set when the containing type is union type. We need
            // these partial properties when identifying discriminant properties, but otherwise they are filtered out
            // and do not appear to be present in the union type.
            function getUnionOrIntersectionProperty(type, name, skipObjectFunctionPropertyAugment) {
                var _a, _b;
                var property = ((_a = type.propertyCacheWithoutObjectFunctionPropertyAugment) === null || _a === void 0 ? void 0 : _a.get(name)) ||
                    !skipObjectFunctionPropertyAugment ? (_b = type.propertyCache) === null || _b === void 0 ? void 0 : _b.get(name) : undefined;
                if (!property) {
                    property = createUnionOrIntersectionProperty(type, name, skipObjectFunctionPropertyAugment);
                    if (property) {
                        var properties = skipObjectFunctionPropertyAugment ? type.propertyCacheWithoutObjectFunctionPropertyAugment || (type.propertyCacheWithoutObjectFunctionPropertyAugment = ts.createSymbolTable()) : type.propertyCache || (type.propertyCache = ts.createSymbolTable());
                        properties.set(name, property);
                    }
                }
                return property;
            }
            function getPropertyOfUnionOrIntersectionType(type, name, skipObjectFunctionPropertyAugment) {
                var property = getUnionOrIntersectionProperty(type, name, skipObjectFunctionPropertyAugment);
                // We need to filter out partial properties in union types
                return property && !(ts.getCheckFlags(property) & 16 /* ReadPartial */) ? property : undefined;
            }
            /**
             * Return the reduced form of the given type. For a union type, it is a union of the normalized constituent types.
             * For an intersection of types containing one or more mututally exclusive discriminant properties, it is 'never'.
             * For all other types, it is simply the type itself. Discriminant properties are considered mutually exclusive when
             * no constituent property has type 'never', but the intersection of the constituent property types is 'never'.
             */
            function getReducedType(type) {
                if (type.flags & 1048576 /* Union */ && type.objectFlags & 33554432 /* ContainsIntersections */) {
                    return type.resolvedReducedType || (type.resolvedReducedType = getReducedUnionType(type));
                }
                else if (type.flags & 2097152 /* Intersection */) {
                    if (!(type.objectFlags & 33554432 /* IsNeverIntersectionComputed */)) {
                        type.objectFlags |= 33554432 /* IsNeverIntersectionComputed */ |
                            (ts.some(getPropertiesOfUnionOrIntersectionType(type), isNeverReducedProperty) ? 67108864 /* IsNeverIntersection */ : 0);
                    }
                    return type.objectFlags & 67108864 /* IsNeverIntersection */ ? neverType : type;
                }
                return type;
            }
            function getReducedUnionType(unionType) {
                var reducedTypes = ts.sameMap(unionType.types, getReducedType);
                if (reducedTypes === unionType.types) {
                    return unionType;
                }
                var reduced = getUnionType(reducedTypes);
                if (reduced.flags & 1048576 /* Union */) {
                    reduced.resolvedReducedType = reduced;
                }
                return reduced;
            }
            function isNeverReducedProperty(prop) {
                return isDiscriminantWithNeverType(prop) || isConflictingPrivateProperty(prop);
            }
            function isDiscriminantWithNeverType(prop) {
                // Return true for a synthetic non-optional property with non-uniform types, where at least one is
                // a literal type and none is never, that reduces to never.
                return !(prop.flags & 16777216 /* Optional */) &&
                    (ts.getCheckFlags(prop) & (192 /* Discriminant */ | 131072 /* HasNeverType */)) === 192 /* Discriminant */ &&
                    !!(getTypeOfSymbol(prop).flags & 131072 /* Never */);
            }
            function isConflictingPrivateProperty(prop) {
                // Return true for a synthetic property with multiple declarations, at least one of which is private.
                return !prop.valueDeclaration && !!(ts.getCheckFlags(prop) & 1024 /* ContainsPrivate */);
            }
            function elaborateNeverIntersection(errorInfo, type) {
                if (type.flags & 2097152 /* Intersection */ && ts.getObjectFlags(type) & 67108864 /* IsNeverIntersection */) {
                    var neverProp = ts.find(getPropertiesOfUnionOrIntersectionType(type), isDiscriminantWithNeverType);
                    if (neverProp) {
                        return ts.chainDiagnosticMessages(errorInfo, ts.Diagnostics.The_intersection_0_was_reduced_to_never_because_property_1_has_conflicting_types_in_some_constituents, typeToString(type, /*enclosingDeclaration*/ undefined, 536870912 /* NoTypeReduction */), symbolToString(neverProp));
                    }
                    var privateProp = ts.find(getPropertiesOfUnionOrIntersectionType(type), isConflictingPrivateProperty);
                    if (privateProp) {
                        return ts.chainDiagnosticMessages(errorInfo, ts.Diagnostics.The_intersection_0_was_reduced_to_never_because_property_1_exists_in_multiple_constituents_and_is_private_in_some, typeToString(type, /*enclosingDeclaration*/ undefined, 536870912 /* NoTypeReduction */), symbolToString(privateProp));
                    }
                }
                return errorInfo;
            }
            /**
             * Return the symbol for the property with the given name in the given type. Creates synthetic union properties when
             * necessary, maps primitive types and type parameters are to their apparent types, and augments with properties from
             * Object and Function as appropriate.
             *
             * @param type a type to look up property from
             * @param name a name of property to look up in a given type
             */
            function getPropertyOfType(type, name, skipObjectFunctionPropertyAugment) {
                type = getReducedApparentType(type);
                if (type.flags & 524288 /* Object */) {
                    var resolved = resolveStructuredTypeMembers(type);
                    var symbol = resolved.members.get(name);
                    if (symbol && symbolIsValue(symbol)) {
                        return symbol;
                    }
                    if (skipObjectFunctionPropertyAugment)
                        return undefined;
                    var functionType = resolved === anyFunctionType ? globalFunctionType :
                        resolved.callSignatures.length ? globalCallableFunctionType :
                            resolved.constructSignatures.length ? globalNewableFunctionType :
                                undefined;
                    if (functionType) {
                        var symbol_1 = getPropertyOfObjectType(functionType, name);
                        if (symbol_1) {
                            return symbol_1;
                        }
                    }
                    return getPropertyOfObjectType(globalObjectType, name);
                }
                if (type.flags & 3145728 /* UnionOrIntersection */) {
                    return getPropertyOfUnionOrIntersectionType(type, name, skipObjectFunctionPropertyAugment);
                }
                return undefined;
            }
            function getSignaturesOfStructuredType(type, kind) {
                if (type.flags & 3670016 /* StructuredType */) {
                    var resolved = resolveStructuredTypeMembers(type);
                    return kind === 0 /* Call */ ? resolved.callSignatures : resolved.constructSignatures;
                }
                return ts.emptyArray;
            }
            /**
             * Return the signatures of the given kind in the given type. Creates synthetic union signatures when necessary and
             * maps primitive types and type parameters are to their apparent types.
             */
            function getSignaturesOfType(type, kind) {
                return getSignaturesOfStructuredType(getReducedApparentType(type), kind);
            }
            function findIndexInfo(indexInfos, keyType) {
                return ts.find(indexInfos, function (info) { return info.keyType === keyType; });
            }
            function findApplicableIndexInfo(indexInfos, keyType) {
                // Index signatures for type 'string' are considered only when no other index signatures apply.
                var stringIndexInfo;
                var applicableInfo;
                var applicableInfos;
                for (var _i = 0, indexInfos_1 = indexInfos; _i < indexInfos_1.length; _i++) {
                    var info = indexInfos_1[_i];
                    if (info.keyType === stringType) {
                        stringIndexInfo = info;
                    }
                    else if (isApplicableIndexType(keyType, info.keyType)) {
                        if (!applicableInfo) {
                            applicableInfo = info;
                        }
                        else {
                            (applicableInfos || (applicableInfos = [applicableInfo])).push(info);
                        }
                    }
                }
                // When more than one index signature is applicable we create a synthetic IndexInfo. Instead of computing
                // the intersected key type, we just use unknownType for the key type as nothing actually depends on the
                // keyType property of the returned IndexInfo.
                return applicableInfos ? createIndexInfo(unknownType, getIntersectionType(ts.map(applicableInfos, function (info) { return info.type; })), ts.reduceLeft(applicableInfos, function (isReadonly, info) { return isReadonly && info.isReadonly; }, /*initial*/ true)) :
                    applicableInfo ? applicableInfo :
                        stringIndexInfo && isApplicableIndexType(keyType, stringType) ? stringIndexInfo :
                            undefined;
            }
            function isApplicableIndexType(source, target) {
                // A 'string' index signature applies to types assignable to 'string' or 'number', and a 'number' index
                // signature applies to types assignable to 'number' and numeric string literal types.
                return isTypeAssignableTo(source, target) ||
                    target === stringType && isTypeAssignableTo(source, numberType) ||
                    target === numberType && !!(source.flags & 128 /* StringLiteral */) && isNumericLiteralName(source.value);
            }
            function getIndexInfosOfStructuredType(type) {
                if (type.flags & 3670016 /* StructuredType */) {
                    var resolved = resolveStructuredTypeMembers(type);
                    return resolved.indexInfos;
                }
                return ts.emptyArray;
            }
            function getIndexInfosOfType(type) {
                return getIndexInfosOfStructuredType(getReducedApparentType(type));
            }
            // Return the indexing info of the given kind in the given type. Creates synthetic union index types when necessary and
            // maps primitive types and type parameters are to their apparent types.
            function getIndexInfoOfType(type, keyType) {
                return findIndexInfo(getIndexInfosOfType(type), keyType);
            }
            // Return the index type of the given kind in the given type. Creates synthetic union index types when necessary and
            // maps primitive types and type parameters are to their apparent types.
            function getIndexTypeOfType(type, keyType) {
                var _a;
                return (_a = getIndexInfoOfType(type, keyType)) === null || _a === void 0 ? void 0 : _a.type;
            }
            function getApplicableIndexInfos(type, keyType) {
                return getIndexInfosOfType(type).filter(function (info) { return isApplicableIndexType(keyType, info.keyType); });
            }
            function getApplicableIndexInfo(type, keyType) {
                return findApplicableIndexInfo(getIndexInfosOfType(type), keyType);
            }
            function getApplicableIndexInfoForName(type, name) {
                return getApplicableIndexInfo(type, isLateBoundName(name) ? esSymbolType : getStringLiteralType(ts.unescapeLeadingUnderscores(name)));
            }
            // Return list of type parameters with duplicates removed (duplicate identifier errors are generated in the actual
            // type checking functions).
            function getTypeParametersFromDeclaration(declaration) {
                var result;
                for (var _i = 0, _a = ts.getEffectiveTypeParameterDeclarations(declaration); _i < _a.length; _i++) {
                    var node = _a[_i];
                    result = ts.appendIfUnique(result, getDeclaredTypeOfTypeParameter(node.symbol));
                }
                return result;
            }
            function symbolsToArray(symbols) {
                var result = [];
                symbols.forEach(function (symbol, id) {
                    if (!isReservedMemberName(id)) {
                        result.push(symbol);
                    }
                });
                return result;
            }
            function isJSDocOptionalParameter(node) {
                return ts.isInJSFile(node) && (
                // node.type should only be a JSDocOptionalType when node is a parameter of a JSDocFunctionType
                node.type && node.type.kind === 311 /* JSDocOptionalType */
                    || ts.getJSDocParameterTags(node).some(function (_a) {
                        var isBracketed = _a.isBracketed, typeExpression = _a.typeExpression;
                        return isBracketed || !!typeExpression && typeExpression.type.kind === 311 /* JSDocOptionalType */;
                    }));
            }
            function tryFindAmbientModule(moduleName, withAugmentations) {
                if (ts.isExternalModuleNameRelative(moduleName)) {
                    return undefined;
                }
                var symbol = getSymbol(globals, '"' + moduleName + '"', 512 /* ValueModule */);
                // merged symbol is module declaration symbol combined with all augmentations
                return symbol && withAugmentations ? getMergedSymbol(symbol) : symbol;
            }
            function isOptionalParameter(node) {
                if (ts.hasQuestionToken(node) || isOptionalJSDocPropertyLikeTag(node) || isJSDocOptionalParameter(node)) {
                    return true;
                }
                if (node.initializer) {
                    var signature = getSignatureFromDeclaration(node.parent);
                    var parameterIndex = node.parent.parameters.indexOf(node);
                    ts.Debug.assert(parameterIndex >= 0);
                    // Only consider syntactic or instantiated parameters as optional, not `void` parameters as this function is used
                    // in grammar checks and checking for `void` too early results in parameter types widening too early
                    // and causes some noImplicitAny errors to be lost.
                    return parameterIndex >= getMinArgumentCount(signature, 1 /* StrongArityForUntypedJS */ | 2 /* VoidIsNonOptional */);
                }
                var iife = ts.getImmediatelyInvokedFunctionExpression(node.parent);
                if (iife) {
                    return !node.type &&
                        !node.dotDotDotToken &&
                        node.parent.parameters.indexOf(node) >= iife.arguments.length;
                }
                return false;
            }
            function isOptionalPropertyDeclaration(node) {
                return ts.isPropertyDeclaration(node) && node.questionToken;
            }
            function isOptionalJSDocPropertyLikeTag(node) {
                if (!ts.isJSDocPropertyLikeTag(node)) {
                    return false;
                }
                var isBracketed = node.isBracketed, typeExpression = node.typeExpression;
                return isBracketed || !!typeExpression && typeExpression.type.kind === 311 /* JSDocOptionalType */;
            }
            function createTypePredicate(kind, parameterName, parameterIndex, type) {
                return { kind: kind, parameterName: parameterName, parameterIndex: parameterIndex, type: type };
            }
            /**
             * Gets the minimum number of type arguments needed to satisfy all non-optional type
             * parameters.
             */
            function getMinTypeArgumentCount(typeParameters) {
                var minTypeArgumentCount = 0;
                if (typeParameters) {
                    for (var i = 0; i < typeParameters.length; i++) {
                        if (!hasTypeParameterDefault(typeParameters[i])) {
                            minTypeArgumentCount = i + 1;
                        }
                    }
                }
                return minTypeArgumentCount;
            }
            function fillMissingTypeArguments(typeArguments, typeParameters, minTypeArgumentCount, isJavaScriptImplicitAny) {
                var numTypeParameters = ts.length(typeParameters);
                if (!numTypeParameters) {
                    return [];
                }
                var numTypeArguments = ts.length(typeArguments);
                if (isJavaScriptImplicitAny || (numTypeArguments >= minTypeArgumentCount && numTypeArguments <= numTypeParameters)) {
                    var result = typeArguments ? typeArguments.slice() : [];
                    // Map invalid forward references in default types to the error type
                    for (var i = numTypeArguments; i < numTypeParameters; i++) {
                        result[i] = errorType;
                    }
                    var baseDefaultType = getDefaultTypeArgumentType(isJavaScriptImplicitAny);
                    for (var i = numTypeArguments; i < numTypeParameters; i++) {
                        var defaultType = getDefaultFromTypeParameter(typeParameters[i]);
                        if (isJavaScriptImplicitAny && defaultType && (isTypeIdenticalTo(defaultType, unknownType) || isTypeIdenticalTo(defaultType, emptyObjectType))) {
                            defaultType = anyType;
                        }
                        result[i] = defaultType ? instantiateType(defaultType, createTypeMapper(typeParameters, result)) : baseDefaultType;
                    }
                    result.length = typeParameters.length;
                    return result;
                }
                return typeArguments && typeArguments.slice();
            }
            function getSignatureFromDeclaration(declaration) {
                var links = getNodeLinks(declaration);
                if (!links.resolvedSignature) {
                    var parameters = [];
                    var flags = 0 /* None */;
                    var minArgumentCount = 0;
                    var thisParameter = void 0;
                    var hasThisParameter = false;
                    var iife = ts.getImmediatelyInvokedFunctionExpression(declaration);
                    var isJSConstructSignature = ts.isJSDocConstructSignature(declaration);
                    var isUntypedSignatureInJSFile = !iife &&
                        ts.isInJSFile(declaration) &&
                        ts.isValueSignatureDeclaration(declaration) &&
                        !ts.hasJSDocParameterTags(declaration) &&
                        !ts.getJSDocType(declaration);
                    if (isUntypedSignatureInJSFile) {
                        flags |= 32 /* IsUntypedSignatureInJSFile */;
                    }
                    // If this is a JSDoc construct signature, then skip the first parameter in the
                    // parameter list.  The first parameter represents the return type of the construct
                    // signature.
                    for (var i = isJSConstructSignature ? 1 : 0; i < declaration.parameters.length; i++) {
                        var param = declaration.parameters[i];
                        var paramSymbol = param.symbol;
                        var type = ts.isJSDocParameterTag(param) ? (param.typeExpression && param.typeExpression.type) : param.type;
                        // Include parameter symbol instead of property symbol in the signature
                        if (paramSymbol && !!(paramSymbol.flags & 4 /* Property */) && !ts.isBindingPattern(param.name)) {
                            var resolvedSymbol = resolveName(param, paramSymbol.escapedName, 111551 /* Value */, undefined, undefined, /*isUse*/ false);
                            paramSymbol = resolvedSymbol;
                        }
                        if (i === 0 && paramSymbol.escapedName === "this" /* This */) {
                            hasThisParameter = true;
                            thisParameter = param.symbol;
                        }
                        else {
                            parameters.push(paramSymbol);
                        }
                        if (type && type.kind === 194 /* LiteralType */) {
                            flags |= 2 /* HasLiteralTypes */;
                        }
                        // Record a new minimum argument count if this is not an optional parameter
                        var isOptionalParameter_1 = isOptionalJSDocPropertyLikeTag(param) ||
                            param.initializer || param.questionToken || ts.isRestParameter(param) ||
                            iife && parameters.length > iife.arguments.length && !type ||
                            isJSDocOptionalParameter(param);
                        if (!isOptionalParameter_1) {
                            minArgumentCount = parameters.length;
                        }
                    }
                    // If only one accessor includes a this-type annotation, the other behaves as if it had the same type annotation
                    if ((declaration.kind === 170 /* GetAccessor */ || declaration.kind === 171 /* SetAccessor */) &&
                        hasBindableName(declaration) &&
                        (!hasThisParameter || !thisParameter)) {
                        var otherKind = declaration.kind === 170 /* GetAccessor */ ? 171 /* SetAccessor */ : 170 /* GetAccessor */;
                        var other = ts.getDeclarationOfKind(getSymbolOfNode(declaration), otherKind);
                        if (other) {
                            thisParameter = getAnnotatedAccessorThisParameter(other);
                        }
                    }
                    var classType = declaration.kind === 169 /* Constructor */ ?
                        getDeclaredTypeOfClassOrInterface(getMergedSymbol(declaration.parent.symbol))
                        : undefined;
                    var typeParameters = classType ? classType.localTypeParameters : getTypeParametersFromDeclaration(declaration);
                    if (ts.hasRestParameter(declaration) || ts.isInJSFile(declaration) && maybeAddJsSyntheticRestParameter(declaration, parameters)) {
                        flags |= 1 /* HasRestParameter */;
                    }
                    if (ts.isConstructorTypeNode(declaration) && ts.hasSyntacticModifier(declaration, 128 /* Abstract */) ||
                        ts.isConstructorDeclaration(declaration) && ts.hasSyntacticModifier(declaration.parent, 128 /* Abstract */)) {
                        flags |= 4 /* Abstract */;
                    }
                    links.resolvedSignature = createSignature(declaration, typeParameters, thisParameter, parameters, 
                    /*resolvedReturnType*/ undefined, /*resolvedTypePredicate*/ undefined, minArgumentCount, flags);
                }
                return links.resolvedSignature;
            }
            /**
             * A JS function gets a synthetic rest parameter if it references `arguments` AND:
             * 1. It has no parameters but at least one `@param` with a type that starts with `...`
             * OR
             * 2. It has at least one parameter, and the last parameter has a matching `@param` with a type that starts with `...`
             */
            function maybeAddJsSyntheticRestParameter(declaration, parameters) {
                if (ts.isJSDocSignature(declaration) || !containsArgumentsReference(declaration)) {
                    return false;
                }
                var lastParam = ts.lastOrUndefined(declaration.parameters);
                var lastParamTags = lastParam ? ts.getJSDocParameterTags(lastParam) : ts.getJSDocTags(declaration).filter(ts.isJSDocParameterTag);
                var lastParamVariadicType = ts.firstDefined(lastParamTags, function (p) {
                    return p.typeExpression && ts.isJSDocVariadicType(p.typeExpression.type) ? p.typeExpression.type : undefined;
                });
                var syntheticArgsSymbol = createSymbol(3 /* Variable */, "args", 32768 /* RestParameter */);
                syntheticArgsSymbol.type = lastParamVariadicType ? createArrayType(getTypeFromTypeNode(lastParamVariadicType.type)) : anyArrayType;
                if (lastParamVariadicType) {
                    // Replace the last parameter with a rest parameter.
                    parameters.pop();
                }
                parameters.push(syntheticArgsSymbol);
                return true;
            }
            function getSignatureOfTypeTag(node) {
                // should be attached to a function declaration or expression
                if (!(ts.isInJSFile(node) && ts.isFunctionLikeDeclaration(node)))
                    return undefined;
                var typeTag = ts.getJSDocTypeTag(node);
                return (typeTag === null || typeTag === void 0 ? void 0 : typeTag.typeExpression) && getSingleCallSignature(getTypeFromTypeNode(typeTag.typeExpression));
            }
            function getReturnTypeOfTypeTag(node) {
                var signature = getSignatureOfTypeTag(node);
                return signature && getReturnTypeOfSignature(signature);
            }
            function containsArgumentsReference(declaration) {
                var links = getNodeLinks(declaration);
                if (links.containsArgumentsReference === undefined) {
                    if (links.flags & 8192 /* CaptureArguments */) {
                        links.containsArgumentsReference = true;
                    }
                    else {
                        links.containsArgumentsReference = traverse(declaration.body);
                    }
                }
                return links.containsArgumentsReference;
                function traverse(node) {
                    if (!node)
                        return false;
                    switch (node.kind) {
                        case 79 /* Identifier */:
                            return node.escapedText === argumentsSymbol.escapedName && getResolvedSymbol(node) === argumentsSymbol;
                        case 165 /* PropertyDeclaration */:
                        case 167 /* MethodDeclaration */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            return node.name.kind === 160 /* ComputedPropertyName */
                                && traverse(node.name);
                        case 204 /* PropertyAccessExpression */:
                        case 205 /* ElementAccessExpression */:
                            return traverse(node.expression);
                        default:
                            return !ts.nodeStartsNewLexicalEnvironment(node) && !ts.isPartOfTypeNode(node) && !!ts.forEachChild(node, traverse);
                    }
                }
            }
            function getSignaturesOfSymbol(symbol) {
                if (!symbol || !symbol.declarations)
                    return ts.emptyArray;
                var result = [];
                for (var i = 0; i < symbol.declarations.length; i++) {
                    var decl = symbol.declarations[i];
                    if (!ts.isFunctionLike(decl))
                        continue;
                    // Don't include signature if node is the implementation of an overloaded function. A node is considered
                    // an implementation node if it has a body and the previous node is of the same kind and immediately
                    // precedes the implementation node (i.e. has the same parent and ends where the implementation starts).
                    if (i > 0 && decl.body) {
                        var previous = symbol.declarations[i - 1];
                        if (decl.parent === previous.parent && decl.kind === previous.kind && decl.pos === previous.end) {
                            continue;
                        }
                    }
                    result.push(getSignatureFromDeclaration(decl));
                }
                return result;
            }
            function resolveExternalModuleTypeByLiteral(name) {
                var moduleSym = resolveExternalModuleName(name, name);
                if (moduleSym) {
                    var resolvedModuleSymbol = resolveExternalModuleSymbol(moduleSym);
                    if (resolvedModuleSymbol) {
                        return getTypeOfSymbol(resolvedModuleSymbol);
                    }
                }
                return anyType;
            }
            function getThisTypeOfSignature(signature) {
                if (signature.thisParameter) {
                    return getTypeOfSymbol(signature.thisParameter);
                }
            }
            function getTypePredicateOfSignature(signature) {
                if (!signature.resolvedTypePredicate) {
                    if (signature.target) {
                        var targetTypePredicate = getTypePredicateOfSignature(signature.target);
                        signature.resolvedTypePredicate = targetTypePredicate ? instantiateTypePredicate(targetTypePredicate, signature.mapper) : noTypePredicate;
                    }
                    else if (signature.compositeSignatures) {
                        signature.resolvedTypePredicate = getUnionOrIntersectionTypePredicate(signature.compositeSignatures, signature.compositeKind) || noTypePredicate;
                    }
                    else {
                        var type = signature.declaration && ts.getEffectiveReturnTypeNode(signature.declaration);
                        var jsdocPredicate = void 0;
                        if (!type && ts.isInJSFile(signature.declaration)) {
                            var jsdocSignature = getSignatureOfTypeTag(signature.declaration);
                            if (jsdocSignature && signature !== jsdocSignature) {
                                jsdocPredicate = getTypePredicateOfSignature(jsdocSignature);
                            }
                        }
                        signature.resolvedTypePredicate = type && ts.isTypePredicateNode(type) ?
                            createTypePredicateFromTypePredicateNode(type, signature) :
                            jsdocPredicate || noTypePredicate;
                    }
                    ts.Debug.assert(!!signature.resolvedTypePredicate);
                }
                return signature.resolvedTypePredicate === noTypePredicate ? undefined : signature.resolvedTypePredicate;
            }
            function createTypePredicateFromTypePredicateNode(node, signature) {
                var parameterName = node.parameterName;
                var type = node.type && getTypeFromTypeNode(node.type);
                return parameterName.kind === 190 /* ThisType */ ?
                    createTypePredicate(node.assertsModifier ? 2 /* AssertsThis */ : 0 /* This */, /*parameterName*/ undefined, /*parameterIndex*/ undefined, type) :
                    createTypePredicate(node.assertsModifier ? 3 /* AssertsIdentifier */ : 1 /* Identifier */, parameterName.escapedText, ts.findIndex(signature.parameters, function (p) { return p.escapedName === parameterName.escapedText; }), type);
            }
            function getUnionOrIntersectionType(types, kind, unionReduction) {
                return kind !== 2097152 /* Intersection */ ? getUnionType(types, unionReduction) : getIntersectionType(types);
            }
            function getReturnTypeOfSignature(signature) {
                if (!signature.resolvedReturnType) {
                    if (!pushTypeResolution(signature, 3 /* ResolvedReturnType */)) {
                        return errorType;
                    }
                    var type = signature.target ? instantiateType(getReturnTypeOfSignature(signature.target), signature.mapper) :
                        signature.compositeSignatures ? instantiateType(getUnionOrIntersectionType(ts.map(signature.compositeSignatures, getReturnTypeOfSignature), signature.compositeKind, 2 /* Subtype */), signature.mapper) :
                            getReturnTypeFromAnnotation(signature.declaration) ||
                                (ts.nodeIsMissing(signature.declaration.body) ? anyType : getReturnTypeFromBody(signature.declaration));
                    if (signature.flags & 8 /* IsInnerCallChain */) {
                        type = addOptionalTypeMarker(type);
                    }
                    else if (signature.flags & 16 /* IsOuterCallChain */) {
                        type = getOptionalType(type);
                    }
                    if (!popTypeResolution()) {
                        if (signature.declaration) {
                            var typeNode = ts.getEffectiveReturnTypeNode(signature.declaration);
                            if (typeNode) {
                                error(typeNode, ts.Diagnostics.Return_type_annotation_circularly_references_itself);
                            }
                            else if (noImplicitAny) {
                                var declaration = signature.declaration;
                                var name = ts.getNameOfDeclaration(declaration);
                                if (name) {
                                    error(name, ts.Diagnostics._0_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions, ts.declarationNameToString(name));
                                }
                                else {
                                    error(declaration, ts.Diagnostics.Function_implicitly_has_return_type_any_because_it_does_not_have_a_return_type_annotation_and_is_referenced_directly_or_indirectly_in_one_of_its_return_expressions);
                                }
                            }
                        }
                        type = anyType;
                    }
                    signature.resolvedReturnType = type;
                }
                return signature.resolvedReturnType;
            }
            function getReturnTypeFromAnnotation(declaration) {
                if (declaration.kind === 169 /* Constructor */) {
                    return getDeclaredTypeOfClassOrInterface(getMergedSymbol(declaration.parent.symbol));
                }
                if (ts.isJSDocConstructSignature(declaration)) {
                    return getTypeFromTypeNode(declaration.parameters[0].type); // TODO: GH#18217
                }
                var typeNode = ts.getEffectiveReturnTypeNode(declaration);
                if (typeNode) {
                    return getTypeFromTypeNode(typeNode);
                }
                if (declaration.kind === 170 /* GetAccessor */ && hasBindableName(declaration)) {
                    var jsDocType = ts.isInJSFile(declaration) && getTypeForDeclarationFromJSDocComment(declaration);
                    if (jsDocType) {
                        return jsDocType;
                    }
                    var setter = ts.getDeclarationOfKind(getSymbolOfNode(declaration), 171 /* SetAccessor */);
                    var setterType = getAnnotatedAccessorType(setter);
                    if (setterType) {
                        return setterType;
                    }
                }
                return getReturnTypeOfTypeTag(declaration);
            }
            function isResolvingReturnTypeOfSignature(signature) {
                return !signature.resolvedReturnType && findResolutionCycleStartIndex(signature, 3 /* ResolvedReturnType */) >= 0;
            }
            function getRestTypeOfSignature(signature) {
                return tryGetRestTypeOfSignature(signature) || anyType;
            }
            function tryGetRestTypeOfSignature(signature) {
                if (signatureHasRestParameter(signature)) {
                    var sigRestType = getTypeOfSymbol(signature.parameters[signature.parameters.length - 1]);
                    var restType = isTupleType(sigRestType) ? getRestTypeOfTupleType(sigRestType) : sigRestType;
                    return restType && getIndexTypeOfType(restType, numberType);
                }
                return undefined;
            }
            function getSignatureInstantiation(signature, typeArguments, isJavascript, inferredTypeParameters) {
                var instantiatedSignature = getSignatureInstantiationWithoutFillingInTypeArguments(signature, fillMissingTypeArguments(typeArguments, signature.typeParameters, getMinTypeArgumentCount(signature.typeParameters), isJavascript));
                if (inferredTypeParameters) {
                    var returnSignature = getSingleCallOrConstructSignature(getReturnTypeOfSignature(instantiatedSignature));
                    if (returnSignature) {
                        var newReturnSignature = cloneSignature(returnSignature);
                        newReturnSignature.typeParameters = inferredTypeParameters;
                        var newInstantiatedSignature = cloneSignature(instantiatedSignature);
                        newInstantiatedSignature.resolvedReturnType = getOrCreateTypeFromSignature(newReturnSignature);
                        return newInstantiatedSignature;
                    }
                }
                return instantiatedSignature;
            }
            function getSignatureInstantiationWithoutFillingInTypeArguments(signature, typeArguments) {
                var instantiations = signature.instantiations || (signature.instantiations = new ts.Map());
                var id = getTypeListId(typeArguments);
                var instantiation = instantiations.get(id);
                if (!instantiation) {
                    instantiations.set(id, instantiation = createSignatureInstantiation(signature, typeArguments));
                }
                return instantiation;
            }
            function createSignatureInstantiation(signature, typeArguments) {
                return instantiateSignature(signature, createSignatureTypeMapper(signature, typeArguments), /*eraseTypeParameters*/ true);
            }
            function createSignatureTypeMapper(signature, typeArguments) {
                return createTypeMapper(signature.typeParameters, typeArguments);
            }
            function getErasedSignature(signature) {
                return signature.typeParameters ?
                    signature.erasedSignatureCache || (signature.erasedSignatureCache = createErasedSignature(signature)) :
                    signature;
            }
            function createErasedSignature(signature) {
                // Create an instantiation of the signature where all type arguments are the any type.
                return instantiateSignature(signature, createTypeEraser(signature.typeParameters), /*eraseTypeParameters*/ true);
            }
            function getCanonicalSignature(signature) {
                return signature.typeParameters ?
                    signature.canonicalSignatureCache || (signature.canonicalSignatureCache = createCanonicalSignature(signature)) :
                    signature;
            }
            function createCanonicalSignature(signature) {
                // Create an instantiation of the signature where each unconstrained type parameter is replaced with
                // its original. When a generic class or interface is instantiated, each generic method in the class or
                // interface is instantiated with a fresh set of cloned type parameters (which we need to handle scenarios
                // where different generations of the same type parameter are in scope). This leads to a lot of new type
                // identities, and potentially a lot of work comparing those identities, so here we create an instantiation
                // that uses the original type identities for all unconstrained type parameters.
                return getSignatureInstantiation(signature, ts.map(signature.typeParameters, function (tp) { return tp.target && !getConstraintOfTypeParameter(tp.target) ? tp.target : tp; }), ts.isInJSFile(signature.declaration));
            }
            function getBaseSignature(signature) {
                var typeParameters = signature.typeParameters;
                if (typeParameters) {
                    if (signature.baseSignatureCache) {
                        return signature.baseSignatureCache;
                    }
                    var typeEraser = createTypeEraser(typeParameters);
                    var baseConstraintMapper_1 = createTypeMapper(typeParameters, ts.map(typeParameters, function (tp) { return getConstraintOfTypeParameter(tp) || unknownType; }));
                    var baseConstraints = ts.map(typeParameters, function (tp) { return instantiateType(tp, baseConstraintMapper_1) || unknownType; });
                    // Run N type params thru the immediate constraint mapper up to N times
                    // This way any noncircular interdependent type parameters are definitely resolved to their external dependencies
                    for (var i = 0; i < typeParameters.length - 1; i++) {
                        baseConstraints = instantiateTypes(baseConstraints, baseConstraintMapper_1);
                    }
                    // and then apply a type eraser to remove any remaining circularly dependent type parameters
                    baseConstraints = instantiateTypes(baseConstraints, typeEraser);
                    return signature.baseSignatureCache = instantiateSignature(signature, createTypeMapper(typeParameters, baseConstraints), /*eraseTypeParameters*/ true);
                }
                return signature;
            }
            function getOrCreateTypeFromSignature(signature) {
                // There are two ways to declare a construct signature, one is by declaring a class constructor
                // using the constructor keyword, and the other is declaring a bare construct signature in an
                // object type literal or interface (using the new keyword). Each way of declaring a constructor
                // will result in a different declaration kind.
                if (!signature.isolatedSignatureType) {
                    var kind = signature.declaration ? signature.declaration.kind : 0 /* Unknown */;
                    var isConstructor = kind === 169 /* Constructor */ || kind === 173 /* ConstructSignature */ || kind === 178 /* ConstructorType */;
                    var type = createObjectType(16 /* Anonymous */);
                    type.members = emptySymbols;
                    type.properties = ts.emptyArray;
                    type.callSignatures = !isConstructor ? [signature] : ts.emptyArray;
                    type.constructSignatures = isConstructor ? [signature] : ts.emptyArray;
                    type.indexInfos = ts.emptyArray;
                    signature.isolatedSignatureType = type;
                }
                return signature.isolatedSignatureType;
            }
            function getIndexSymbol(symbol) {
                return symbol.members ? getIndexSymbolFromSymbolTable(symbol.members) : undefined;
            }
            function getIndexSymbolFromSymbolTable(symbolTable) {
                return symbolTable.get("__index" /* Index */);
            }
            function createIndexInfo(keyType, type, isReadonly, declaration) {
                return { keyType: keyType, type: type, isReadonly: isReadonly, declaration: declaration };
            }
            function getIndexInfosOfSymbol(symbol) {
                var indexSymbol = getIndexSymbol(symbol);
                return indexSymbol ? getIndexInfosOfIndexSymbol(indexSymbol) : ts.emptyArray;
            }
            function getIndexInfosOfIndexSymbol(indexSymbol) {
                if (indexSymbol.declarations) {
                    var indexInfos_2 = [];
                    var _loop_14 = function (declaration) {
                        if (declaration.parameters.length === 1) {
                            var parameter = declaration.parameters[0];
                            if (parameter.type) {
                                forEachType(getTypeFromTypeNode(parameter.type), function (keyType) {
                                    if (isValidIndexKeyType(keyType) && !findIndexInfo(indexInfos_2, keyType)) {
                                        indexInfos_2.push(createIndexInfo(keyType, declaration.type ? getTypeFromTypeNode(declaration.type) : anyType, ts.hasEffectiveModifier(declaration, 64 /* Readonly */), declaration));
                                    }
                                });
                            }
                        }
                    };
                    for (var _i = 0, _a = indexSymbol.declarations; _i < _a.length; _i++) {
                        var declaration = _a[_i];
                        _loop_14(declaration);
                    }
                    return indexInfos_2;
                }
                return ts.emptyArray;
            }
            function isValidIndexKeyType(type) {
                return !!(type.flags & (4 /* String */ | 8 /* Number */ | 4096 /* ESSymbol */)) || isPatternLiteralType(type) ||
                    !!(type.flags & 2097152 /* Intersection */) && !isGenericIndexType(type) && !isGenericObjectType(type) && ts.some(type.types, isValidIndexKeyType);
            }
            function getConstraintDeclaration(type) {
                return ts.mapDefined(ts.filter(type.symbol && type.symbol.declarations, ts.isTypeParameterDeclaration), ts.getEffectiveConstraintOfTypeParameter)[0];
            }
            function getInferredTypeParameterConstraint(typeParameter) {
                var _a;
                var inferences;
                if ((_a = typeParameter.symbol) === null || _a === void 0 ? void 0 : _a.declarations) {
                    for (var _i = 0, _b = typeParameter.symbol.declarations; _i < _b.length; _i++) {
                        var declaration = _b[_i];
                        if (declaration.parent.kind === 188 /* InferType */) {
                            // When an 'infer T' declaration is immediately contained in a type reference node
                            // (such as 'Foo<infer T>'), T's constraint is inferred from the constraint of the
                            // corresponding type parameter in 'Foo'. When multiple 'infer T' declarations are
                            // present, we form an intersection of the inferred constraint types.
                            var _c = ts.walkUpParenthesizedTypesAndGetParentAndChild(declaration.parent.parent), _d = _c[0], childTypeParameter = _d === void 0 ? declaration.parent : _d, grandParent = _c[1];
                            if (grandParent.kind === 176 /* TypeReference */) {
                                var typeReference = grandParent;
                                var typeParameters = getTypeParametersForTypeReference(typeReference);
                                if (typeParameters) {
                                    var index = typeReference.typeArguments.indexOf(childTypeParameter);
                                    if (index < typeParameters.length) {
                                        var declaredConstraint = getConstraintOfTypeParameter(typeParameters[index]);
                                        if (declaredConstraint) {
                                            // Type parameter constraints can reference other type parameters so
                                            // constraints need to be instantiated. If instantiation produces the
                                            // type parameter itself, we discard that inference. For example, in
                                            //   type Foo<T extends string, U extends T> = [T, U];
                                            //   type Bar<T> = T extends Foo<infer X, infer X> ? Foo<X, X> : T;
                                            // the instantiated constraint for U is X, so we discard that inference.
                                            var mapper = createTypeMapper(typeParameters, getEffectiveTypeArguments(typeReference, typeParameters));
                                            var constraint = instantiateType(declaredConstraint, mapper);
                                            if (constraint !== typeParameter) {
                                                inferences = ts.append(inferences, constraint);
                                            }
                                        }
                                    }
                                }
                            }
                            // When an 'infer T' declaration is immediately contained in a rest parameter declaration, a rest type
                            // or a named rest tuple element, we infer an 'unknown[]' constraint.
                            else if (grandParent.kind === 162 /* Parameter */ && grandParent.dotDotDotToken ||
                                grandParent.kind === 184 /* RestType */ ||
                                grandParent.kind === 195 /* NamedTupleMember */ && grandParent.dotDotDotToken) {
                                inferences = ts.append(inferences, createArrayType(unknownType));
                            }
                            // When an 'infer T' declaration is immediately contained in a string template type, we infer a 'string'
                            // constraint.
                            else if (grandParent.kind === 197 /* TemplateLiteralTypeSpan */) {
                                inferences = ts.append(inferences, stringType);
                            }
                            // When an 'infer T' declaration is in the constraint position of a mapped type, we infer a 'keyof any'
                            // constraint.
                            else if (grandParent.kind === 161 /* TypeParameter */ && grandParent.parent.kind === 193 /* MappedType */) {
                                inferences = ts.append(inferences, keyofConstraintType);
                            }
                            // When an 'infer T' declaration is the template of a mapped type, and that mapped type is the extends
                            // clause of a conditional whose check type is also a mapped type, give it a constraint equal to the template
                            // of the check type's mapped type
                            else if (grandParent.kind === 193 /* MappedType */ && grandParent.type &&
                                ts.skipParentheses(grandParent.type) === declaration.parent && grandParent.parent.kind === 187 /* ConditionalType */ &&
                                grandParent.parent.extendsType === grandParent && grandParent.parent.checkType.kind === 193 /* MappedType */ &&
                                grandParent.parent.checkType.type) {
                                var checkMappedType_1 = grandParent.parent.checkType;
                                var nodeType = getTypeFromTypeNode(checkMappedType_1.type);
                                inferences = ts.append(inferences, instantiateType(nodeType, makeUnaryTypeMapper(getDeclaredTypeOfTypeParameter(getSymbolOfNode(checkMappedType_1.typeParameter)), checkMappedType_1.typeParameter.constraint ? getTypeFromTypeNode(checkMappedType_1.typeParameter.constraint) : keyofConstraintType)));
                            }
                        }
                    }
                }
                return inferences && getIntersectionType(inferences);
            }
            /** This is a worker function. Use getConstraintOfTypeParameter which guards against circular constraints. */
            function getConstraintFromTypeParameter(typeParameter) {
                if (!typeParameter.constraint) {
                    if (typeParameter.target) {
                        var targetConstraint = getConstraintOfTypeParameter(typeParameter.target);
                        typeParameter.constraint = targetConstraint ? instantiateType(targetConstraint, typeParameter.mapper) : noConstraintType;
                    }
                    else {
                        var constraintDeclaration = getConstraintDeclaration(typeParameter);
                        if (!constraintDeclaration) {
                            typeParameter.constraint = getInferredTypeParameterConstraint(typeParameter) || noConstraintType;
                        }
                        else {
                            var type = getTypeFromTypeNode(constraintDeclaration);
                            if (type.flags & 1 /* Any */ && type !== errorType) { // Allow errorType to propegate to keep downstream errors suppressed
                                // use keyofConstraintType as the base constraint for mapped type key constraints (unknown isn;t assignable to that, but `any` was),
                                // use unknown otherwise
                                type = constraintDeclaration.parent.parent.kind === 193 /* MappedType */ ? keyofConstraintType : unknownType;
                            }
                            typeParameter.constraint = type;
                        }
                    }
                }
                return typeParameter.constraint === noConstraintType ? undefined : typeParameter.constraint;
            }
            function getParentSymbolOfTypeParameter(typeParameter) {
                var tp = ts.getDeclarationOfKind(typeParameter.symbol, 161 /* TypeParameter */);
                var host = ts.isJSDocTemplateTag(tp.parent) ? ts.getHostSignatureFromJSDoc(tp.parent) : tp.parent;
                return host && getSymbolOfNode(host);
            }
            function getTypeListId(types) {
                var result = "";
                if (types) {
                    var length_4 = types.length;
                    var i = 0;
                    while (i < length_4) {
                        var startId = types[i].id;
                        var count = 1;
                        while (i + count < length_4 && types[i + count].id === startId + count) {
                            count++;
                        }
                        if (result.length) {
                            result += ",";
                        }
                        result += startId;
                        if (count > 1) {
                            result += ":" + count;
                        }
                        i += count;
                    }
                }
                return result;
            }
            function getAliasId(aliasSymbol, aliasTypeArguments) {
                return aliasSymbol ? "@" + getSymbolId(aliasSymbol) + (aliasTypeArguments ? ":" + getTypeListId(aliasTypeArguments) : "") : "";
            }
            // This function is used to propagate certain flags when creating new object type references and union types.
            // It is only necessary to do so if a constituent type might be the undefined type, the null type, the type
            // of an object literal or the anyFunctionType. This is because there are operations in the type checker
            // that care about the presence of such types at arbitrary depth in a containing type.
            function getPropagatingFlagsOfTypes(types, excludeKinds) {
                var result = 0;
                for (var _i = 0, types_8 = types; _i < types_8.length; _i++) {
                    var type = types_8[_i];
                    if (!(type.flags & excludeKinds)) {
                        result |= ts.getObjectFlags(type);
                    }
                }
                return result & 917504 /* PropagatingFlags */;
            }
            function createTypeReference(target, typeArguments) {
                var id = getTypeListId(typeArguments);
                var type = target.instantiations.get(id);
                if (!type) {
                    type = createObjectType(4 /* Reference */, target.symbol);
                    target.instantiations.set(id, type);
                    type.objectFlags |= typeArguments ? getPropagatingFlagsOfTypes(typeArguments, /*excludeKinds*/ 0) : 0;
                    type.target = target;
                    type.resolvedTypeArguments = typeArguments;
                }
                return type;
            }
            function cloneTypeReference(source) {
                var type = createType(source.flags);
                type.symbol = source.symbol;
                type.objectFlags = source.objectFlags;
                type.target = source.target;
                type.resolvedTypeArguments = source.resolvedTypeArguments;
                return type;
            }
            function createDeferredTypeReference(target, node, mapper, aliasSymbol, aliasTypeArguments) {
                if (!aliasSymbol) {
                    aliasSymbol = getAliasSymbolForTypeNode(node);
                    var localAliasTypeArguments = getTypeArgumentsForAliasSymbol(aliasSymbol);
                    aliasTypeArguments = mapper ? instantiateTypes(localAliasTypeArguments, mapper) : localAliasTypeArguments;
                }
                var type = createObjectType(4 /* Reference */, target.symbol);
                type.target = target;
                type.node = node;
                type.mapper = mapper;
                type.aliasSymbol = aliasSymbol;
                type.aliasTypeArguments = aliasTypeArguments;
                return type;
            }
            function getTypeArguments(type) {
                var _a, _b;
                if (!type.resolvedTypeArguments) {
                    if (!pushTypeResolution(type, 6 /* ResolvedTypeArguments */)) {
                        return ((_a = type.target.localTypeParameters) === null || _a === void 0 ? void 0 : _a.map(function () { return errorType; })) || ts.emptyArray;
                    }
                    var node = type.node;
                    var typeArguments = !node ? ts.emptyArray :
                        node.kind === 176 /* TypeReference */ ? ts.concatenate(type.target.outerTypeParameters, getEffectiveTypeArguments(node, type.target.localTypeParameters)) :
                            node.kind === 181 /* ArrayType */ ? [getTypeFromTypeNode(node.elementType)] :
                                ts.map(node.elements, getTypeFromTypeNode);
                    if (popTypeResolution()) {
                        type.resolvedTypeArguments = type.mapper ? instantiateTypes(typeArguments, type.mapper) : typeArguments;
                    }
                    else {
                        type.resolvedTypeArguments = ((_b = type.target.localTypeParameters) === null || _b === void 0 ? void 0 : _b.map(function () { return errorType; })) || ts.emptyArray;
                        error(type.node || currentNode, type.target.symbol ? ts.Diagnostics.Type_arguments_for_0_circularly_reference_themselves : ts.Diagnostics.Tuple_type_arguments_circularly_reference_themselves, type.target.symbol && symbolToString(type.target.symbol));
                    }
                }
                return type.resolvedTypeArguments;
            }
            function getTypeReferenceArity(type) {
                return ts.length(type.target.typeParameters);
            }
            /**
             * Get type from type-reference that reference to class or interface
             */
            function getTypeFromClassOrInterfaceReference(node, symbol) {
                var type = getDeclaredTypeOfSymbol(getMergedSymbol(symbol));
                var typeParameters = type.localTypeParameters;
                if (typeParameters) {
                    var numTypeArguments = ts.length(node.typeArguments);
                    var minTypeArgumentCount = getMinTypeArgumentCount(typeParameters);
                    var isJs = ts.isInJSFile(node);
                    var isJsImplicitAny = !noImplicitAny && isJs;
                    if (!isJsImplicitAny && (numTypeArguments < minTypeArgumentCount || numTypeArguments > typeParameters.length)) {
                        var missingAugmentsTag = isJs && ts.isExpressionWithTypeArguments(node) && !ts.isJSDocAugmentsTag(node.parent);
                        var diag = minTypeArgumentCount === typeParameters.length ?
                            missingAugmentsTag ?
                                ts.Diagnostics.Expected_0_type_arguments_provide_these_with_an_extends_tag :
                                ts.Diagnostics.Generic_type_0_requires_1_type_argument_s :
                            missingAugmentsTag ?
                                ts.Diagnostics.Expected_0_1_type_arguments_provide_these_with_an_extends_tag :
                                ts.Diagnostics.Generic_type_0_requires_between_1_and_2_type_arguments;
                        var typeStr = typeToString(type, /*enclosingDeclaration*/ undefined, 2 /* WriteArrayAsGenericType */);
                        error(node, diag, typeStr, minTypeArgumentCount, typeParameters.length);
                        if (!isJs) {
                            // TODO: Adopt same permissive behavior in TS as in JS to reduce follow-on editing experience failures (requires editing fillMissingTypeArguments)
                            return errorType;
                        }
                    }
                    if (node.kind === 176 /* TypeReference */ && isDeferredTypeReferenceNode(node, ts.length(node.typeArguments) !== typeParameters.length)) {
                        return createDeferredTypeReference(type, node, /*mapper*/ undefined);
                    }
                    // In a type reference, the outer type parameters of the referenced class or interface are automatically
                    // supplied as type arguments and the type reference only specifies arguments for the local type parameters
                    // of the class or interface.
                    var typeArguments = ts.concatenate(type.outerTypeParameters, fillMissingTypeArguments(typeArgumentsFromTypeReferenceNode(node), typeParameters, minTypeArgumentCount, isJs));
                    return createTypeReference(type, typeArguments);
                }
                return checkNoTypeArguments(node, symbol) ? type : errorType;
            }
            function getTypeAliasInstantiation(symbol, typeArguments, aliasSymbol, aliasTypeArguments) {
                var type = getDeclaredTypeOfSymbol(symbol);
                if (type === intrinsicMarkerType && intrinsicTypeKinds.has(symbol.escapedName) && typeArguments && typeArguments.length === 1) {
                    return getStringMappingType(symbol, typeArguments[0]);
                }
                var links = getSymbolLinks(symbol);
                var typeParameters = links.typeParameters;
                var id = getTypeListId(typeArguments) + getAliasId(aliasSymbol, aliasTypeArguments);
                var instantiation = links.instantiations.get(id);
                if (!instantiation) {
                    links.instantiations.set(id, instantiation = instantiateTypeWithAlias(type, createTypeMapper(typeParameters, fillMissingTypeArguments(typeArguments, typeParameters, getMinTypeArgumentCount(typeParameters), ts.isInJSFile(symbol.valueDeclaration))), aliasSymbol, aliasTypeArguments));
                }
                return instantiation;
            }
            /**
             * Get type from reference to type alias. When a type alias is generic, the declared type of the type alias may include
             * references to the type parameters of the alias. We replace those with the actual type arguments by instantiating the
             * declared type. Instantiations are cached using the type identities of the type arguments as the key.
             */
            function getTypeFromTypeAliasReference(node, symbol) {
                var type = getDeclaredTypeOfSymbol(symbol);
                var typeParameters = getSymbolLinks(symbol).typeParameters;
                if (typeParameters) {
                    var numTypeArguments = ts.length(node.typeArguments);
                    var minTypeArgumentCount = getMinTypeArgumentCount(typeParameters);
                    if (numTypeArguments < minTypeArgumentCount || numTypeArguments > typeParameters.length) {
                        error(node, minTypeArgumentCount === typeParameters.length ?
                            ts.Diagnostics.Generic_type_0_requires_1_type_argument_s :
                            ts.Diagnostics.Generic_type_0_requires_between_1_and_2_type_arguments, symbolToString(symbol), minTypeArgumentCount, typeParameters.length);
                        return errorType;
                    }
                    // We refrain from associating a local type alias with an instantiation of a top-level type alias
                    // because the local alias may end up being referenced in an inferred return type where it is not
                    // accessible--which in turn may lead to a large structural expansion of the type when generating
                    // a .d.ts file. See #43622 for an example.
                    var aliasSymbol = getAliasSymbolForTypeNode(node);
                    var newAliasSymbol = aliasSymbol && (isLocalTypeAlias(symbol) || !isLocalTypeAlias(aliasSymbol)) ? aliasSymbol : undefined;
                    return getTypeAliasInstantiation(symbol, typeArgumentsFromTypeReferenceNode(node), newAliasSymbol, getTypeArgumentsForAliasSymbol(newAliasSymbol));
                }
                return checkNoTypeArguments(node, symbol) ? type : errorType;
            }
            function isLocalTypeAlias(symbol) {
                var _a;
                var declaration = (_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.isTypeAlias);
                return !!(declaration && ts.getContainingFunction(declaration));
            }
            function getTypeReferenceName(node) {
                switch (node.kind) {
                    case 176 /* TypeReference */:
                        return node.typeName;
                    case 226 /* ExpressionWithTypeArguments */:
                        // We only support expressions that are simple qualified names. For other
                        // expressions this produces undefined.
                        var expr = node.expression;
                        if (ts.isEntityNameExpression(expr)) {
                            return expr;
                        }
                    // fall through;
                }
                return undefined;
            }
            function resolveTypeReferenceName(typeReferenceName, meaning, ignoreErrors) {
                if (!typeReferenceName) {
                    return unknownSymbol;
                }
                return resolveEntityName(typeReferenceName, meaning, ignoreErrors) || unknownSymbol;
            }
            function getTypeReferenceType(node, symbol) {
                if (symbol === unknownSymbol) {
                    return errorType;
                }
                symbol = getExpandoSymbol(symbol) || symbol;
                if (symbol.flags & (32 /* Class */ | 64 /* Interface */)) {
                    return getTypeFromClassOrInterfaceReference(node, symbol);
                }
                if (symbol.flags & 524288 /* TypeAlias */) {
                    return getTypeFromTypeAliasReference(node, symbol);
                }
                // Get type from reference to named type that cannot be generic (enum or type parameter)
                var res = tryGetDeclaredTypeOfSymbol(symbol);
                if (res) {
                    return checkNoTypeArguments(node, symbol) ? getRegularTypeOfLiteralType(res) : errorType;
                }
                if (symbol.flags & 111551 /* Value */ && isJSDocTypeReference(node)) {
                    var jsdocType = getTypeFromJSDocValueReference(node, symbol);
                    if (jsdocType) {
                        return jsdocType;
                    }
                    else {
                        // Resolve the type reference as a Type for the purpose of reporting errors.
                        resolveTypeReferenceName(getTypeReferenceName(node), 788968 /* Type */);
                        return getTypeOfSymbol(symbol);
                    }
                }
                return errorType;
            }
            /**
             * A JSdoc TypeReference may be to a value, but resolve it as a type anyway.
             * Example: import('./b').ConstructorFunction
             */
            function getTypeFromJSDocValueReference(node, symbol) {
                var links = getNodeLinks(node);
                if (!links.resolvedJSDocType) {
                    var valueType = getTypeOfSymbol(symbol);
                    var typeType = valueType;
                    if (symbol.valueDeclaration) {
                        var isImportTypeWithQualifier = node.kind === 198 /* ImportType */ && node.qualifier;
                        // valueType might not have a symbol, eg, {import('./b').STRING_LITERAL}
                        if (valueType.symbol && valueType.symbol !== symbol && isImportTypeWithQualifier) {
                            typeType = getTypeReferenceType(node, valueType.symbol);
                        }
                    }
                    links.resolvedJSDocType = typeType;
                }
                return links.resolvedJSDocType;
            }
            function getSubstitutionType(baseType, substitute) {
                if (substitute.flags & 3 /* AnyOrUnknown */ || substitute === baseType) {
                    return baseType;
                }
                var id = getTypeId(baseType) + ">" + getTypeId(substitute);
                var cached = substitutionTypes.get(id);
                if (cached) {
                    return cached;
                }
                var result = createType(33554432 /* Substitution */);
                result.baseType = baseType;
                result.substitute = substitute;
                substitutionTypes.set(id, result);
                return result;
            }
            function isUnaryTupleTypeNode(node) {
                return node.kind === 182 /* TupleType */ && node.elements.length === 1;
            }
            function getImpliedConstraint(type, checkNode, extendsNode) {
                return isUnaryTupleTypeNode(checkNode) && isUnaryTupleTypeNode(extendsNode) ? getImpliedConstraint(type, checkNode.elements[0], extendsNode.elements[0]) :
                    getActualTypeVariable(getTypeFromTypeNode(checkNode)) === type ? getTypeFromTypeNode(extendsNode) :
                        undefined;
            }
            function getConditionalFlowTypeOfType(type, node) {
                var constraints;
                var covariant = true;
                while (node && !ts.isStatement(node) && node.kind !== 315 /* JSDocComment */) {
                    var parent = node.parent;
                    // only consider variance flipped by parameter locations - `keyof` types would usually be considered variance inverting, but
                    // often get used in indexed accesses where they behave sortof invariantly, but our checking is lax
                    if (parent.kind === 162 /* Parameter */) {
                        covariant = !covariant;
                    }
                    // Always substitute on type parameters, regardless of variance, since even
                    // in contravariant positions, they may rely on substituted constraints to be valid
                    if ((covariant || type.flags & 8650752 /* TypeVariable */) && parent.kind === 187 /* ConditionalType */ && node === parent.trueType) {
                        var constraint = getImpliedConstraint(type, parent.checkType, parent.extendsType);
                        if (constraint) {
                            constraints = ts.append(constraints, constraint);
                        }
                    }
                    node = parent;
                }
                return constraints ? getSubstitutionType(type, getIntersectionType(ts.append(constraints, type))) : type;
            }
            function isJSDocTypeReference(node) {
                return !!(node.flags & 4194304 /* JSDoc */) && (node.kind === 176 /* TypeReference */ || node.kind === 198 /* ImportType */);
            }
            function checkNoTypeArguments(node, symbol) {
                if (node.typeArguments) {
                    error(node, ts.Diagnostics.Type_0_is_not_generic, symbol ? symbolToString(symbol) : node.typeName ? ts.declarationNameToString(node.typeName) : anon);
                    return false;
                }
                return true;
            }
            function getIntendedTypeFromJSDocTypeReference(node) {
                if (ts.isIdentifier(node.typeName)) {
                    var typeArgs = node.typeArguments;
                    switch (node.typeName.escapedText) {
                        case "String":
                            checkNoTypeArguments(node);
                            return stringType;
                        case "Number":
                            checkNoTypeArguments(node);
                            return numberType;
                        case "Boolean":
                            checkNoTypeArguments(node);
                            return booleanType;
                        case "Void":
                            checkNoTypeArguments(node);
                            return voidType;
                        case "Undefined":
                            checkNoTypeArguments(node);
                            return undefinedType;
                        case "Null":
                            checkNoTypeArguments(node);
                            return nullType;
                        case "Function":
                        case "function":
                            checkNoTypeArguments(node);
                            return globalFunctionType;
                        case "array":
                            return (!typeArgs || !typeArgs.length) && !noImplicitAny ? anyArrayType : undefined;
                        case "promise":
                            return (!typeArgs || !typeArgs.length) && !noImplicitAny ? createPromiseType(anyType) : undefined;
                        case "Object":
                            if (typeArgs && typeArgs.length === 2) {
                                if (ts.isJSDocIndexSignature(node)) {
                                    var indexed = getTypeFromTypeNode(typeArgs[0]);
                                    var target = getTypeFromTypeNode(typeArgs[1]);
                                    var indexInfo = indexed === stringType || indexed === numberType ? [createIndexInfo(indexed, target, /*isReadonly*/ false)] : ts.emptyArray;
                                    return createAnonymousType(undefined, emptySymbols, ts.emptyArray, ts.emptyArray, indexInfo);
                                }
                                return anyType;
                            }
                            checkNoTypeArguments(node);
                            return !noImplicitAny ? anyType : undefined;
                    }
                }
            }
            function getTypeFromJSDocNullableTypeNode(node) {
                var type = getTypeFromTypeNode(node.type);
                return strictNullChecks ? getNullableType(type, 65536 /* Null */) : type;
            }
            function getTypeFromTypeReference(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    // handle LS queries on the `const` in `x as const` by resolving to the type of `x`
                    if (ts.isConstTypeReference(node) && ts.isAssertionExpression(node.parent)) {
                        links.resolvedSymbol = unknownSymbol;
                        return links.resolvedType = checkExpressionCached(node.parent.expression);
                    }
                    var symbol = void 0;
                    var type = void 0;
                    var meaning = 788968 /* Type */;
                    if (isJSDocTypeReference(node)) {
                        type = getIntendedTypeFromJSDocTypeReference(node);
                        if (!type) {
                            symbol = resolveTypeReferenceName(getTypeReferenceName(node), meaning, /*ignoreErrors*/ true);
                            if (symbol === unknownSymbol) {
                                symbol = resolveTypeReferenceName(getTypeReferenceName(node), meaning | 111551 /* Value */);
                            }
                            else {
                                resolveTypeReferenceName(getTypeReferenceName(node), meaning); // Resolve again to mark errors, if any
                            }
                            type = getTypeReferenceType(node, symbol);
                        }
                    }
                    if (!type) {
                        symbol = resolveTypeReferenceName(getTypeReferenceName(node), meaning);
                        type = getTypeReferenceType(node, symbol);
                    }
                    // Cache both the resolved symbol and the resolved type. The resolved symbol is needed when we check the
                    // type reference in checkTypeReferenceNode.
                    links.resolvedSymbol = symbol;
                    links.resolvedType = type;
                }
                return links.resolvedType;
            }
            function typeArgumentsFromTypeReferenceNode(node) {
                return ts.map(node.typeArguments, getTypeFromTypeNode);
            }
            function getTypeFromTypeQueryNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    // TypeScript 1.0 spec (April 2014): 3.6.3
                    // The expression is processed as an identifier expression (section 4.3)
                    // or property access expression(section 4.10),
                    // the widened type(section 3.9) of which becomes the result.
                    var type = ts.isThisIdentifier(node.exprName) ? checkThisExpression(node.exprName) : checkExpression(node.exprName);
                    links.resolvedType = getRegularTypeOfLiteralType(getWidenedType(type));
                }
                return links.resolvedType;
            }
            function getTypeOfGlobalSymbol(symbol, arity) {
                function getTypeDeclaration(symbol) {
                    var declarations = symbol.declarations;
                    if (declarations) {
                        for (var _i = 0, declarations_3 = declarations; _i < declarations_3.length; _i++) {
                            var declaration = declarations_3[_i];
                            switch (declaration.kind) {
                                case 255 /* ClassDeclaration */:
                                case 256 /* InterfaceDeclaration */:
                                case 258 /* EnumDeclaration */:
                                    return declaration;
                            }
                        }
                    }
                }
                if (!symbol) {
                    return arity ? emptyGenericType : emptyObjectType;
                }
                var type = getDeclaredTypeOfSymbol(symbol);
                if (!(type.flags & 524288 /* Object */)) {
                    error(getTypeDeclaration(symbol), ts.Diagnostics.Global_type_0_must_be_a_class_or_interface_type, ts.symbolName(symbol));
                    return arity ? emptyGenericType : emptyObjectType;
                }
                if (ts.length(type.typeParameters) !== arity) {
                    error(getTypeDeclaration(symbol), ts.Diagnostics.Global_type_0_must_have_1_type_parameter_s, ts.symbolName(symbol), arity);
                    return arity ? emptyGenericType : emptyObjectType;
                }
                return type;
            }
            function getGlobalValueSymbol(name, reportErrors) {
                return getGlobalSymbol(name, 111551 /* Value */, reportErrors ? ts.Diagnostics.Cannot_find_global_value_0 : undefined);
            }
            function getGlobalTypeSymbol(name, reportErrors) {
                return getGlobalSymbol(name, 788968 /* Type */, reportErrors ? ts.Diagnostics.Cannot_find_global_type_0 : undefined);
            }
            function getGlobalSymbol(name, meaning, diagnostic) {
                // Don't track references for global symbols anyway, so value if `isReference` is arbitrary
                return resolveName(undefined, name, meaning, diagnostic, name, /*isUse*/ false);
            }
            function getGlobalType(name, arity, reportErrors) {
                var symbol = getGlobalTypeSymbol(name, reportErrors);
                return symbol || reportErrors ? getTypeOfGlobalSymbol(symbol, arity) : undefined;
            }
            function getGlobalTypedPropertyDescriptorType() {
                return deferredGlobalTypedPropertyDescriptorType || (deferredGlobalTypedPropertyDescriptorType = getGlobalType("TypedPropertyDescriptor", /*arity*/ 1, /*reportErrors*/ true)) || emptyGenericType;
            }
            function getGlobalTemplateStringsArrayType() {
                return deferredGlobalTemplateStringsArrayType || (deferredGlobalTemplateStringsArrayType = getGlobalType("TemplateStringsArray", /*arity*/ 0, /*reportErrors*/ true)) || emptyObjectType;
            }
            function getGlobalImportMetaType() {
                return deferredGlobalImportMetaType || (deferredGlobalImportMetaType = getGlobalType("ImportMeta", /*arity*/ 0, /*reportErrors*/ true)) || emptyObjectType;
            }
            function getGlobalImportMetaExpressionType() {
                if (!deferredGlobalImportMetaExpressionType) {
                    // Create a synthetic type `ImportMetaExpression { meta: MetaProperty }`
                    var symbol = createSymbol(0 /* None */, "ImportMetaExpression");
                    var importMetaType = getGlobalImportMetaType();
                    var metaPropertySymbol = createSymbol(4 /* Property */, "meta", 8 /* Readonly */);
                    metaPropertySymbol.parent = symbol;
                    metaPropertySymbol.type = importMetaType;
                    var members = ts.createSymbolTable([metaPropertySymbol]);
                    symbol.members = members;
                    deferredGlobalImportMetaExpressionType = createAnonymousType(symbol, members, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                }
                return deferredGlobalImportMetaExpressionType;
            }
            function getGlobalESSymbolConstructorSymbol(reportErrors) {
                return deferredGlobalESSymbolConstructorSymbol || (deferredGlobalESSymbolConstructorSymbol = getGlobalValueSymbol("Symbol", reportErrors));
            }
            function getGlobalESSymbolConstructorTypeSymbol(reportErrors) {
                return deferredGlobalESSymbolConstructorTypeSymbol || (deferredGlobalESSymbolConstructorTypeSymbol = getGlobalTypeSymbol("SymbolConstructor", reportErrors));
            }
            function getGlobalESSymbolType(reportErrors) {
                return deferredGlobalESSymbolType || (deferredGlobalESSymbolType = getGlobalType("Symbol", /*arity*/ 0, reportErrors)) || emptyObjectType;
            }
            function getGlobalPromiseType(reportErrors) {
                return deferredGlobalPromiseType || (deferredGlobalPromiseType = getGlobalType("Promise", /*arity*/ 1, reportErrors)) || emptyGenericType;
            }
            function getGlobalPromiseLikeType(reportErrors) {
                return deferredGlobalPromiseLikeType || (deferredGlobalPromiseLikeType = getGlobalType("PromiseLike", /*arity*/ 1, reportErrors)) || emptyGenericType;
            }
            function getGlobalPromiseConstructorSymbol(reportErrors) {
                return deferredGlobalPromiseConstructorSymbol || (deferredGlobalPromiseConstructorSymbol = getGlobalValueSymbol("Promise", reportErrors));
            }
            function getGlobalPromiseConstructorLikeType(reportErrors) {
                return deferredGlobalPromiseConstructorLikeType || (deferredGlobalPromiseConstructorLikeType = getGlobalType("PromiseConstructorLike", /*arity*/ 0, reportErrors)) || emptyObjectType;
            }
            function getGlobalAsyncIterableType(reportErrors) {
                return deferredGlobalAsyncIterableType || (deferredGlobalAsyncIterableType = getGlobalType("AsyncIterable", /*arity*/ 1, reportErrors)) || emptyGenericType;
            }
            function getGlobalAsyncIteratorType(reportErrors) {
                return deferredGlobalAsyncIteratorType || (deferredGlobalAsyncIteratorType = getGlobalType("AsyncIterator", /*arity*/ 3, reportErrors)) || emptyGenericType;
            }
            function getGlobalAsyncIterableIteratorType(reportErrors) {
                return deferredGlobalAsyncIterableIteratorType || (deferredGlobalAsyncIterableIteratorType = getGlobalType("AsyncIterableIterator", /*arity*/ 1, reportErrors)) || emptyGenericType;
            }
            function getGlobalAsyncGeneratorType(reportErrors) {
                return deferredGlobalAsyncGeneratorType || (deferredGlobalAsyncGeneratorType = getGlobalType("AsyncGenerator", /*arity*/ 3, reportErrors)) || emptyGenericType;
            }
            function getGlobalIterableType(reportErrors) {
                return deferredGlobalIterableType || (deferredGlobalIterableType = getGlobalType("Iterable", /*arity*/ 1, reportErrors)) || emptyGenericType;
            }
            function getGlobalIteratorType(reportErrors) {
                return deferredGlobalIteratorType || (deferredGlobalIteratorType = getGlobalType("Iterator", /*arity*/ 3, reportErrors)) || emptyGenericType;
            }
            function getGlobalIterableIteratorType(reportErrors) {
                return deferredGlobalIterableIteratorType || (deferredGlobalIterableIteratorType = getGlobalType("IterableIterator", /*arity*/ 1, reportErrors)) || emptyGenericType;
            }
            function getGlobalGeneratorType(reportErrors) {
                return deferredGlobalGeneratorType || (deferredGlobalGeneratorType = getGlobalType("Generator", /*arity*/ 3, reportErrors)) || emptyGenericType;
            }
            function getGlobalIteratorYieldResultType(reportErrors) {
                return deferredGlobalIteratorYieldResultType || (deferredGlobalIteratorYieldResultType = getGlobalType("IteratorYieldResult", /*arity*/ 1, reportErrors)) || emptyGenericType;
            }
            function getGlobalIteratorReturnResultType(reportErrors) {
                return deferredGlobalIteratorReturnResultType || (deferredGlobalIteratorReturnResultType = getGlobalType("IteratorReturnResult", /*arity*/ 1, reportErrors)) || emptyGenericType;
            }
            function getGlobalTypeOrUndefined(name, arity) {
                if (arity === void 0) {
                    arity = 0;
                }
                var symbol = getGlobalSymbol(name, 788968 /* Type */, /*diagnostic*/ undefined);
                return symbol && getTypeOfGlobalSymbol(symbol, arity);
            }
            function getGlobalExtractSymbol() {
                return deferredGlobalExtractSymbol || (deferredGlobalExtractSymbol = getGlobalSymbol("Extract", 524288 /* TypeAlias */, ts.Diagnostics.Cannot_find_global_type_0)); // TODO: GH#18217
            }
            function getGlobalOmitSymbol() {
                return deferredGlobalOmitSymbol || (deferredGlobalOmitSymbol = getGlobalSymbol("Omit", 524288 /* TypeAlias */, ts.Diagnostics.Cannot_find_global_type_0)); // TODO: GH#18217
            }
            function getGlobalBigIntType(reportErrors) {
                return deferredGlobalBigIntType || (deferredGlobalBigIntType = getGlobalType("BigInt", /*arity*/ 0, reportErrors)) || emptyObjectType;
            }
            /**
             * Instantiates a global type that is generic with some element type, and returns that instantiation.
             */
            function createTypeFromGenericGlobalType(genericGlobalType, typeArguments) {
                return genericGlobalType !== emptyGenericType ? createTypeReference(genericGlobalType, typeArguments) : emptyObjectType;
            }
            function createTypedPropertyDescriptorType(propertyType) {
                return createTypeFromGenericGlobalType(getGlobalTypedPropertyDescriptorType(), [propertyType]);
            }
            function createIterableType(iteratedType) {
                return createTypeFromGenericGlobalType(getGlobalIterableType(/*reportErrors*/ true), [iteratedType]);
            }
            function createArrayType(elementType, readonly) {
                return createTypeFromGenericGlobalType(readonly ? globalReadonlyArrayType : globalArrayType, [elementType]);
            }
            function getTupleElementFlags(node) {
                switch (node.kind) {
                    case 183 /* OptionalType */:
                        return 2 /* Optional */;
                    case 184 /* RestType */:
                        return getRestTypeElementFlags(node);
                    case 195 /* NamedTupleMember */:
                        return node.questionToken ? 2 /* Optional */ :
                            node.dotDotDotToken ? getRestTypeElementFlags(node) :
                                1 /* Required */;
                    default:
                        return 1 /* Required */;
                }
            }
            function getRestTypeElementFlags(node) {
                return getArrayElementTypeNode(node.type) ? 4 /* Rest */ : 8 /* Variadic */;
            }
            function getArrayOrTupleTargetType(node) {
                var readonly = isReadonlyTypeOperator(node.parent);
                var elementType = getArrayElementTypeNode(node);
                if (elementType) {
                    return readonly ? globalReadonlyArrayType : globalArrayType;
                }
                var elementFlags = ts.map(node.elements, getTupleElementFlags);
                var missingName = ts.some(node.elements, function (e) { return e.kind !== 195 /* NamedTupleMember */; });
                return getTupleTargetType(elementFlags, readonly, /*associatedNames*/ missingName ? undefined : node.elements);
            }
            // Return true if the given type reference node is directly aliased or if it needs to be deferred
            // because it is possibly contained in a circular chain of eagerly resolved types.
            function isDeferredTypeReferenceNode(node, hasDefaultTypeArguments) {
                return !!getAliasSymbolForTypeNode(node) || isResolvedByTypeAlias(node) && (node.kind === 181 /* ArrayType */ ? mayResolveTypeAlias(node.elementType) :
                    node.kind === 182 /* TupleType */ ? ts.some(node.elements, mayResolveTypeAlias) :
                        hasDefaultTypeArguments || ts.some(node.typeArguments, mayResolveTypeAlias));
            }
            // Return true when the given node is transitively contained in type constructs that eagerly
            // resolve their constituent types. We include SyntaxKind.TypeReference because type arguments
            // of type aliases are eagerly resolved.
            function isResolvedByTypeAlias(node) {
                var parent = node.parent;
                switch (parent.kind) {
                    case 189 /* ParenthesizedType */:
                    case 195 /* NamedTupleMember */:
                    case 176 /* TypeReference */:
                    case 185 /* UnionType */:
                    case 186 /* IntersectionType */:
                    case 192 /* IndexedAccessType */:
                    case 187 /* ConditionalType */:
                    case 191 /* TypeOperator */:
                    case 181 /* ArrayType */:
                    case 182 /* TupleType */:
                        return isResolvedByTypeAlias(parent);
                    case 257 /* TypeAliasDeclaration */:
                        return true;
                }
                return false;
            }
            // Return true if resolving the given node (i.e. getTypeFromTypeNode) possibly causes resolution
            // of a type alias.
            function mayResolveTypeAlias(node) {
                switch (node.kind) {
                    case 176 /* TypeReference */:
                        return isJSDocTypeReference(node) || !!(resolveTypeReferenceName(node.typeName, 788968 /* Type */).flags & 524288 /* TypeAlias */);
                    case 179 /* TypeQuery */:
                        return true;
                    case 191 /* TypeOperator */:
                        return node.operator !== 152 /* UniqueKeyword */ && mayResolveTypeAlias(node.type);
                    case 189 /* ParenthesizedType */:
                    case 183 /* OptionalType */:
                    case 195 /* NamedTupleMember */:
                    case 311 /* JSDocOptionalType */:
                    case 309 /* JSDocNullableType */:
                    case 310 /* JSDocNonNullableType */:
                    case 304 /* JSDocTypeExpression */:
                        return mayResolveTypeAlias(node.type);
                    case 184 /* RestType */:
                        return node.type.kind !== 181 /* ArrayType */ || mayResolveTypeAlias(node.type.elementType);
                    case 185 /* UnionType */:
                    case 186 /* IntersectionType */:
                        return ts.some(node.types, mayResolveTypeAlias);
                    case 192 /* IndexedAccessType */:
                        return mayResolveTypeAlias(node.objectType) || mayResolveTypeAlias(node.indexType);
                    case 187 /* ConditionalType */:
                        return mayResolveTypeAlias(node.checkType) || mayResolveTypeAlias(node.extendsType) ||
                            mayResolveTypeAlias(node.trueType) || mayResolveTypeAlias(node.falseType);
                }
                return false;
            }
            function getTypeFromArrayOrTupleTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    var target = getArrayOrTupleTargetType(node);
                    if (target === emptyGenericType) {
                        links.resolvedType = emptyObjectType;
                    }
                    else if (!(node.kind === 182 /* TupleType */ && ts.some(node.elements, function (e) { return !!(getTupleElementFlags(e) & 8 /* Variadic */); })) && isDeferredTypeReferenceNode(node)) {
                        links.resolvedType = node.kind === 182 /* TupleType */ && node.elements.length === 0 ? target :
                            createDeferredTypeReference(target, node, /*mapper*/ undefined);
                    }
                    else {
                        var elementTypes = node.kind === 181 /* ArrayType */ ? [getTypeFromTypeNode(node.elementType)] : ts.map(node.elements, getTypeFromTypeNode);
                        links.resolvedType = createNormalizedTypeReference(target, elementTypes);
                    }
                }
                return links.resolvedType;
            }
            function isReadonlyTypeOperator(node) {
                return ts.isTypeOperatorNode(node) && node.operator === 143 /* ReadonlyKeyword */;
            }
            function createTupleType(elementTypes, elementFlags, readonly, namedMemberDeclarations) {
                if (readonly === void 0) {
                    readonly = false;
                }
                var tupleTarget = getTupleTargetType(elementFlags || ts.map(elementTypes, function (_) { return 1 /* Required */; }), readonly, namedMemberDeclarations);
                return tupleTarget === emptyGenericType ? emptyObjectType :
                    elementTypes.length ? createNormalizedTypeReference(tupleTarget, elementTypes) :
                        tupleTarget;
            }
            function getTupleTargetType(elementFlags, readonly, namedMemberDeclarations) {
                if (elementFlags.length === 1 && elementFlags[0] & 4 /* Rest */) {
                    // [...X[]] is equivalent to just X[]
                    return readonly ? globalReadonlyArrayType : globalArrayType;
                }
                var key = ts.map(elementFlags, function (f) { return f & 1 /* Required */ ? "#" : f & 2 /* Optional */ ? "?" : f & 4 /* Rest */ ? "." : "*"; }).join() +
                    (readonly ? "R" : "") +
                    (namedMemberDeclarations && namedMemberDeclarations.length ? "," + ts.map(namedMemberDeclarations, getNodeId).join(",") : "");
                var type = tupleTypes.get(key);
                if (!type) {
                    tupleTypes.set(key, type = createTupleTargetType(elementFlags, readonly, namedMemberDeclarations));
                }
                return type;
            }
            // We represent tuple types as type references to synthesized generic interface types created by
            // this function. The types are of the form:
            //
            //   interface Tuple<T0, T1, T2, ...> extends Array<T0 | T1 | T2 | ...> { 0: T0, 1: T1, 2: T2, ... }
            //
            // Note that the generic type created by this function has no symbol associated with it. The same
            // is true for each of the synthesized type parameters.
            function createTupleTargetType(elementFlags, readonly, namedMemberDeclarations) {
                var arity = elementFlags.length;
                var minLength = ts.countWhere(elementFlags, function (f) { return !!(f & (1 /* Required */ | 8 /* Variadic */)); });
                var typeParameters;
                var properties = [];
                var combinedFlags = 0;
                if (arity) {
                    typeParameters = new Array(arity);
                    for (var i = 0; i < arity; i++) {
                        var typeParameter = typeParameters[i] = createTypeParameter();
                        var flags = elementFlags[i];
                        combinedFlags |= flags;
                        if (!(combinedFlags & 12 /* Variable */)) {
                            var property = createSymbol(4 /* Property */ | (flags & 2 /* Optional */ ? 16777216 /* Optional */ : 0), "" + i, readonly ? 8 /* Readonly */ : 0);
                            property.tupleLabelDeclaration = namedMemberDeclarations === null || namedMemberDeclarations === void 0 ? void 0 : namedMemberDeclarations[i];
                            property.type = typeParameter;
                            properties.push(property);
                        }
                    }
                }
                var fixedLength = properties.length;
                var lengthSymbol = createSymbol(4 /* Property */, "length");
                if (combinedFlags & 12 /* Variable */) {
                    lengthSymbol.type = numberType;
                }
                else {
                    var literalTypes = [];
                    for (var i = minLength; i <= arity; i++)
                        literalTypes.push(getNumberLiteralType(i));
                    lengthSymbol.type = getUnionType(literalTypes);
                }
                properties.push(lengthSymbol);
                var type = createObjectType(8 /* Tuple */ | 4 /* Reference */);
                type.typeParameters = typeParameters;
                type.outerTypeParameters = undefined;
                type.localTypeParameters = typeParameters;
                type.instantiations = new ts.Map();
                type.instantiations.set(getTypeListId(type.typeParameters), type);
                type.target = type;
                type.resolvedTypeArguments = type.typeParameters;
                type.thisType = createTypeParameter();
                type.thisType.isThisType = true;
                type.thisType.constraint = type;
                type.declaredProperties = properties;
                type.declaredCallSignatures = ts.emptyArray;
                type.declaredConstructSignatures = ts.emptyArray;
                type.declaredIndexInfos = ts.emptyArray;
                type.elementFlags = elementFlags;
                type.minLength = minLength;
                type.fixedLength = fixedLength;
                type.hasRestElement = !!(combinedFlags & 12 /* Variable */);
                type.combinedFlags = combinedFlags;
                type.readonly = readonly;
                type.labeledElementDeclarations = namedMemberDeclarations;
                return type;
            }
            function createNormalizedTypeReference(target, typeArguments) {
                return target.objectFlags & 8 /* Tuple */ ? createNormalizedTupleType(target, typeArguments) : createTypeReference(target, typeArguments);
            }
            function createNormalizedTupleType(target, elementTypes) {
                var _a, _b, _c;
                if (!(target.combinedFlags & 14 /* NonRequired */)) {
                    // No need to normalize when we only have regular required elements
                    return createTypeReference(target, elementTypes);
                }
                if (target.combinedFlags & 8 /* Variadic */) {
                    // Transform [A, ...(X | Y | Z)] into [A, ...X] | [A, ...Y] | [A, ...Z]
                    var unionIndex_1 = ts.findIndex(elementTypes, function (t, i) { return !!(target.elementFlags[i] & 8 /* Variadic */ && t.flags & (131072 /* Never */ | 1048576 /* Union */)); });
                    if (unionIndex_1 >= 0) {
                        return checkCrossProductUnion(ts.map(elementTypes, function (t, i) { return target.elementFlags[i] & 8 /* Variadic */ ? t : unknownType; })) ?
                            mapType(elementTypes[unionIndex_1], function (t) { return createNormalizedTupleType(target, ts.replaceElement(elementTypes, unionIndex_1, t)); }) :
                            errorType;
                    }
                }
                // We have optional, rest, or variadic elements that may need normalizing. Normalization ensures that all variadic
                // elements are generic and that the tuple type has one of the following layouts, disregarding variadic elements:
                // (1) Zero or more required elements, followed by zero or more optional elements, followed by zero or one rest element.
                // (2) Zero or more required elements, followed by a rest element, followed by zero or more required elements.
                // In either layout, zero or more generic variadic elements may be present at any location.
                var expandedTypes = [];
                var expandedFlags = [];
                var expandedDeclarations = [];
                var lastRequiredIndex = -1;
                var firstRestIndex = -1;
                var lastOptionalOrRestIndex = -1;
                var _loop_15 = function (i) {
                    var type = elementTypes[i];
                    var flags = target.elementFlags[i];
                    if (flags & 8 /* Variadic */) {
                        if (type.flags & 58982400 /* InstantiableNonPrimitive */ || isGenericMappedType(type)) {
                            // Generic variadic elements stay as they are.
                            addElement(type, 8 /* Variadic */, (_a = target.labeledElementDeclarations) === null || _a === void 0 ? void 0 : _a[i]);
                        }
                        else if (isTupleType(type)) {
                            var elements = getTypeArguments(type);
                            if (elements.length + expandedTypes.length >= 10000) {
                                error(currentNode, ts.isPartOfTypeNode(currentNode)
                                    ? ts.Diagnostics.Type_produces_a_tuple_type_that_is_too_large_to_represent
                                    : ts.Diagnostics.Expression_produces_a_tuple_type_that_is_too_large_to_represent);
                                return { value: errorType };
                            }
                            // Spread variadic elements with tuple types into the resulting tuple.
                            ts.forEach(elements, function (t, n) { var _a; return addElement(t, type.target.elementFlags[n], (_a = type.target.labeledElementDeclarations) === null || _a === void 0 ? void 0 : _a[n]); });
                        }
                        else {
                            // Treat everything else as an array type and create a rest element.
                            addElement(isArrayLikeType(type) && getIndexTypeOfType(type, numberType) || errorType, 4 /* Rest */, (_b = target.labeledElementDeclarations) === null || _b === void 0 ? void 0 : _b[i]);
                        }
                    }
                    else {
                        // Copy other element kinds with no change.
                        addElement(type, flags, (_c = target.labeledElementDeclarations) === null || _c === void 0 ? void 0 : _c[i]);
                    }
                };
                for (var i = 0; i < elementTypes.length; i++) {
                    var state_4 = _loop_15(i);
                    if (typeof state_4 === "object")
                        return state_4.value;
                }
                // Turn optional elements preceding the last required element into required elements
                for (var i = 0; i < lastRequiredIndex; i++) {
                    if (expandedFlags[i] & 2 /* Optional */)
                        expandedFlags[i] = 1 /* Required */;
                }
                if (firstRestIndex >= 0 && firstRestIndex < lastOptionalOrRestIndex) {
                    // Turn elements between first rest and last optional/rest into a single rest element
                    expandedTypes[firstRestIndex] = getUnionType(ts.sameMap(expandedTypes.slice(firstRestIndex, lastOptionalOrRestIndex + 1), function (t, i) { return expandedFlags[firstRestIndex + i] & 8 /* Variadic */ ? getIndexedAccessType(t, numberType) : t; }));
                    expandedTypes.splice(firstRestIndex + 1, lastOptionalOrRestIndex - firstRestIndex);
                    expandedFlags.splice(firstRestIndex + 1, lastOptionalOrRestIndex - firstRestIndex);
                    expandedDeclarations === null || expandedDeclarations === void 0 ? void 0 : expandedDeclarations.splice(firstRestIndex + 1, lastOptionalOrRestIndex - firstRestIndex);
                }
                var tupleTarget = getTupleTargetType(expandedFlags, target.readonly, expandedDeclarations);
                return tupleTarget === emptyGenericType ? emptyObjectType :
                    expandedFlags.length ? createTypeReference(tupleTarget, expandedTypes) :
                        tupleTarget;
                function addElement(type, flags, declaration) {
                    if (flags & 1 /* Required */) {
                        lastRequiredIndex = expandedFlags.length;
                    }
                    if (flags & 4 /* Rest */ && firstRestIndex < 0) {
                        firstRestIndex = expandedFlags.length;
                    }
                    if (flags & (2 /* Optional */ | 4 /* Rest */)) {
                        lastOptionalOrRestIndex = expandedFlags.length;
                    }
                    expandedTypes.push(type);
                    expandedFlags.push(flags);
                    if (expandedDeclarations && declaration) {
                        expandedDeclarations.push(declaration);
                    }
                    else {
                        expandedDeclarations = undefined;
                    }
                }
            }
            function sliceTupleType(type, index, endSkipCount) {
                if (endSkipCount === void 0) {
                    endSkipCount = 0;
                }
                var target = type.target;
                var endIndex = getTypeReferenceArity(type) - endSkipCount;
                return index > target.fixedLength ? getRestArrayTypeOfTupleType(type) || createTupleType(ts.emptyArray) :
                    createTupleType(getTypeArguments(type).slice(index, endIndex), target.elementFlags.slice(index, endIndex), 
                    /*readonly*/ false, target.labeledElementDeclarations && target.labeledElementDeclarations.slice(index, endIndex));
            }
            function getKnownKeysOfTupleType(type) {
                return getUnionType(ts.append(ts.arrayOf(type.target.fixedLength, function (i) { return getStringLiteralType("" + i); }), getIndexType(type.target.readonly ? globalReadonlyArrayType : globalArrayType)));
            }
            // Return count of starting consecutive tuple elements of the given kind(s)
            function getStartElementCount(type, flags) {
                var index = ts.findIndex(type.elementFlags, function (f) { return !(f & flags); });
                return index >= 0 ? index : type.elementFlags.length;
            }
            // Return count of ending consecutive tuple elements of the given kind(s)
            function getEndElementCount(type, flags) {
                return type.elementFlags.length - ts.findLastIndex(type.elementFlags, function (f) { return !(f & flags); }) - 1;
            }
            function getTypeFromOptionalTypeNode(node) {
                return addOptionality(getTypeFromTypeNode(node.type), /*isProperty*/ true);
            }
            function getTypeId(type) {
                return type.id;
            }
            function containsType(types, type) {
                return ts.binarySearch(types, type, getTypeId, ts.compareValues) >= 0;
            }
            function insertType(types, type) {
                var index = ts.binarySearch(types, type, getTypeId, ts.compareValues);
                if (index < 0) {
                    types.splice(~index, 0, type);
                    return true;
                }
                return false;
            }
            function addTypeToUnion(typeSet, includes, type) {
                var flags = type.flags;
                if (flags & 1048576 /* Union */) {
                    return addTypesToUnion(typeSet, includes | (isNamedUnionType(type) ? 1048576 /* Union */ : 0), type.types);
                }
                // We ignore 'never' types in unions
                if (!(flags & 131072 /* Never */)) {
                    includes |= flags & 205258751 /* IncludesMask */;
                    if (flags & 469499904 /* StructuredOrInstantiable */)
                        includes |= 262144 /* IncludesStructuredOrInstantiable */;
                    if (type === wildcardType)
                        includes |= 8388608 /* IncludesWildcard */;
                    if (!strictNullChecks && flags & 98304 /* Nullable */) {
                        if (!(ts.getObjectFlags(type) & 131072 /* ContainsWideningType */))
                            includes |= 4194304 /* IncludesNonWideningType */;
                    }
                    else {
                        var len = typeSet.length;
                        var index = len && type.id > typeSet[len - 1].id ? ~len : ts.binarySearch(typeSet, type, getTypeId, ts.compareValues);
                        if (index < 0) {
                            typeSet.splice(~index, 0, type);
                        }
                    }
                }
                return includes;
            }
            // Add the given types to the given type set. Order is preserved, duplicates are removed,
            // and nested types of the given kind are flattened into the set.
            function addTypesToUnion(typeSet, includes, types) {
                for (var _i = 0, types_9 = types; _i < types_9.length; _i++) {
                    var type = types_9[_i];
                    includes = addTypeToUnion(typeSet, includes, type);
                }
                return includes;
            }
            function removeSubtypes(types, hasObjectTypes) {
                var id = getTypeListId(types);
                var match = subtypeReductionCache.get(id);
                if (match) {
                    return match;
                }
                // We assume that redundant primitive types have already been removed from the types array and that there
                // are no any and unknown types in the array. Thus, the only possible supertypes for primitive types are empty
                // object types, and if none of those are present we can exclude primitive types from the subtype check.
                var hasEmptyObject = hasObjectTypes && ts.some(types, function (t) { return !!(t.flags & 524288 /* Object */) && !isGenericMappedType(t) && isEmptyResolvedType(resolveStructuredTypeMembers(t)); });
                var len = types.length;
                var i = len;
                var count = 0;
                while (i > 0) {
                    i--;
                    var source = types[i];
                    if (hasEmptyObject || source.flags & 469499904 /* StructuredOrInstantiable */) {
                        // Find the first property with a unit type, if any. When constituents have a property by the same name
                        // but of a different unit type, we can quickly disqualify them from subtype checks. This helps subtype
                        // reduction of large discriminated union types.
                        var keyProperty = source.flags & (524288 /* Object */ | 2097152 /* Intersection */ | 58982400 /* InstantiableNonPrimitive */) ?
                            ts.find(getPropertiesOfType(source), function (p) { return isUnitType(getTypeOfSymbol(p)); }) :
                            undefined;
                        var keyPropertyType = keyProperty && getRegularTypeOfLiteralType(getTypeOfSymbol(keyProperty));
                        for (var _i = 0, types_10 = types; _i < types_10.length; _i++) {
                            var target = types_10[_i];
                            if (source !== target) {
                                if (count === 100000) {
                                    // After 100000 subtype checks we estimate the remaining amount of work by assuming the
                                    // same ratio of checks per element. If the estimated number of remaining type checks is
                                    // greater than 1M we deem the union type too complex to represent. This for example
                                    // caps union types at 1000 unique object types.
                                    var estimatedCount = (count / (len - i)) * len;
                                    if (estimatedCount > 1000000) {
                                        ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.instant("checkTypes" /* CheckTypes */, "removeSubtypes_DepthLimit", { typeIds: types.map(function (t) { return t.id; }) });
                                        error(currentNode, ts.Diagnostics.Expression_produces_a_union_type_that_is_too_complex_to_represent);
                                        return undefined;
                                    }
                                }
                                count++;
                                if (keyProperty && target.flags & (524288 /* Object */ | 2097152 /* Intersection */ | 58982400 /* InstantiableNonPrimitive */)) {
                                    var t = getTypeOfPropertyOfType(target, keyProperty.escapedName);
                                    if (t && isUnitType(t) && getRegularTypeOfLiteralType(t) !== keyPropertyType) {
                                        continue;
                                    }
                                }
                                if (isTypeRelatedTo(source, target, strictSubtypeRelation) && (!(ts.getObjectFlags(getTargetType(source)) & 1 /* Class */) ||
                                    !(ts.getObjectFlags(getTargetType(target)) & 1 /* Class */) ||
                                    isTypeDerivedFrom(source, target))) {
                                    ts.orderedRemoveItemAt(types, i);
                                    break;
                                }
                            }
                        }
                    }
                }
                subtypeReductionCache.set(id, types);
                return types;
            }
            function removeRedundantLiteralTypes(types, includes, reduceVoidUndefined) {
                var i = types.length;
                while (i > 0) {
                    i--;
                    var t = types[i];
                    var flags = t.flags;
                    var remove = flags & (128 /* StringLiteral */ | 134217728 /* TemplateLiteral */ | 268435456 /* StringMapping */) && includes & 4 /* String */ ||
                        flags & 256 /* NumberLiteral */ && includes & 8 /* Number */ ||
                        flags & 2048 /* BigIntLiteral */ && includes & 64 /* BigInt */ ||
                        flags & 8192 /* UniqueESSymbol */ && includes & 4096 /* ESSymbol */ ||
                        reduceVoidUndefined && flags & 32768 /* Undefined */ && includes & 16384 /* Void */ ||
                        isFreshLiteralType(t) && containsType(types, t.regularType);
                    if (remove) {
                        ts.orderedRemoveItemAt(types, i);
                    }
                }
            }
            function removeStringLiteralsMatchedByTemplateLiterals(types) {
                var templates = ts.filter(types, isPatternLiteralType);
                if (templates.length) {
                    var i = types.length;
                    var _loop_16 = function () {
                        i--;
                        var t = types[i];
                        if (t.flags & 128 /* StringLiteral */ && ts.some(templates, function (template) { return isTypeSubtypeOf(t, template); })) {
                            ts.orderedRemoveItemAt(types, i);
                        }
                    };
                    while (i > 0) {
                        _loop_16();
                    }
                }
            }
            function isNamedUnionType(type) {
                return !!(type.flags & 1048576 /* Union */ && (type.aliasSymbol || type.origin));
            }
            function addNamedUnions(namedUnions, types) {
                for (var _i = 0, types_11 = types; _i < types_11.length; _i++) {
                    var t = types_11[_i];
                    if (t.flags & 1048576 /* Union */) {
                        var origin = t.origin;
                        if (t.aliasSymbol || origin && !(origin.flags & 1048576 /* Union */)) {
                            ts.pushIfUnique(namedUnions, t);
                        }
                        else if (origin && origin.flags & 1048576 /* Union */) {
                            addNamedUnions(namedUnions, origin.types);
                        }
                    }
                }
            }
            function createOriginUnionOrIntersectionType(flags, types) {
                var result = createOriginType(flags);
                result.types = types;
                return result;
            }
            // We sort and deduplicate the constituent types based on object identity. If the subtypeReduction
            // flag is specified we also reduce the constituent type set to only include types that aren't subtypes
            // of other types. Subtype reduction is expensive for large union types and is possible only when union
            // types are known not to circularly reference themselves (as is the case with union types created by
            // expression constructs such as array literals and the || and ?: operators). Named types can
            // circularly reference themselves and therefore cannot be subtype reduced during their declaration.
            // For example, "type Item = string | (() => Item" is a named type that circularly references itself.
            function getUnionType(types, unionReduction, aliasSymbol, aliasTypeArguments, origin) {
                if (unionReduction === void 0) {
                    unionReduction = 1 /* Literal */;
                }
                if (types.length === 0) {
                    return neverType;
                }
                if (types.length === 1) {
                    return types[0];
                }
                var typeSet = [];
                var includes = addTypesToUnion(typeSet, 0, types);
                if (unionReduction !== 0 /* None */) {
                    if (includes & 3 /* AnyOrUnknown */) {
                        return includes & 1 /* Any */ ? includes & 8388608 /* IncludesWildcard */ ? wildcardType : anyType : unknownType;
                    }
                    if (exactOptionalPropertyTypes && includes & 32768 /* Undefined */) {
                        var missingIndex = ts.binarySearch(typeSet, missingType, getTypeId, ts.compareValues);
                        if (missingIndex >= 0 && containsType(typeSet, undefinedType)) {
                            ts.orderedRemoveItemAt(typeSet, missingIndex);
                        }
                    }
                    if (includes & (2944 /* Literal */ | 8192 /* UniqueESSymbol */ | 134217728 /* TemplateLiteral */ | 268435456 /* StringMapping */) || includes & 16384 /* Void */ && includes & 32768 /* Undefined */) {
                        removeRedundantLiteralTypes(typeSet, includes, !!(unionReduction & 2 /* Subtype */));
                    }
                    if (includes & 128 /* StringLiteral */ && includes & 134217728 /* TemplateLiteral */) {
                        removeStringLiteralsMatchedByTemplateLiterals(typeSet);
                    }
                    if (unionReduction === 2 /* Subtype */) {
                        typeSet = removeSubtypes(typeSet, !!(includes & 524288 /* Object */));
                        if (!typeSet) {
                            return errorType;
                        }
                    }
                    if (typeSet.length === 0) {
                        return includes & 65536 /* Null */ ? includes & 4194304 /* IncludesNonWideningType */ ? nullType : nullWideningType :
                            includes & 32768 /* Undefined */ ? includes & 4194304 /* IncludesNonWideningType */ ? undefinedType : undefinedWideningType :
                                neverType;
                    }
                }
                if (!origin && includes & 1048576 /* Union */) {
                    var namedUnions = [];
                    addNamedUnions(namedUnions, types);
                    var reducedTypes = [];
                    var _loop_17 = function (t) {
                        if (!ts.some(namedUnions, function (union) { return containsType(union.types, t); })) {
                            reducedTypes.push(t);
                        }
                    };
                    for (var _i = 0, typeSet_1 = typeSet; _i < typeSet_1.length; _i++) {
                        var t = typeSet_1[_i];
                        _loop_17(t);
                    }
                    if (!aliasSymbol && namedUnions.length === 1 && reducedTypes.length === 0) {
                        return namedUnions[0];
                    }
                    // We create a denormalized origin type only when the union was created from one or more named unions
                    // (unions with alias symbols or origins) and when there is no overlap between those named unions.
                    var namedTypesCount = ts.reduceLeft(namedUnions, function (sum, union) { return sum + union.types.length; }, 0);
                    if (namedTypesCount + reducedTypes.length === typeSet.length) {
                        for (var _a = 0, namedUnions_1 = namedUnions; _a < namedUnions_1.length; _a++) {
                            var t = namedUnions_1[_a];
                            insertType(reducedTypes, t);
                        }
                        origin = createOriginUnionOrIntersectionType(1048576 /* Union */, reducedTypes);
                    }
                }
                var objectFlags = (includes & 468598819 /* NotPrimitiveUnion */ ? 0 : 65536 /* PrimitiveUnion */) |
                    (includes & 2097152 /* Intersection */ ? 33554432 /* ContainsIntersections */ : 0);
                return getUnionTypeFromSortedList(typeSet, objectFlags, aliasSymbol, aliasTypeArguments, origin);
            }
            function getUnionOrIntersectionTypePredicate(signatures, kind) {
                var first;
                var types = [];
                for (var _i = 0, signatures_6 = signatures; _i < signatures_6.length; _i++) {
                    var sig = signatures_6[_i];
                    var pred = getTypePredicateOfSignature(sig);
                    if (!pred || pred.kind === 2 /* AssertsThis */ || pred.kind === 3 /* AssertsIdentifier */) {
                        if (kind !== 2097152 /* Intersection */) {
                            continue;
                        }
                        else {
                            return; // intersections demand all members be type predicates for the result to have a predicate
                        }
                    }
                    if (first) {
                        if (!typePredicateKindsMatch(first, pred)) {
                            // No common type predicate.
                            return undefined;
                        }
                    }
                    else {
                        first = pred;
                    }
                    types.push(pred.type);
                }
                if (!first) {
                    // No signatures had a type predicate.
                    return undefined;
                }
                var compositeType = getUnionOrIntersectionType(types, kind);
                return createTypePredicate(first.kind, first.parameterName, first.parameterIndex, compositeType);
            }
            function typePredicateKindsMatch(a, b) {
                return a.kind === b.kind && a.parameterIndex === b.parameterIndex;
            }
            // This function assumes the constituent type list is sorted and deduplicated.
            function getUnionTypeFromSortedList(types, objectFlags, aliasSymbol, aliasTypeArguments, origin) {
                if (types.length === 0) {
                    return neverType;
                }
                if (types.length === 1) {
                    return types[0];
                }
                var typeKey = !origin ? getTypeListId(types) :
                    origin.flags & 1048576 /* Union */ ? "|" + getTypeListId(origin.types) :
                        origin.flags & 2097152 /* Intersection */ ? "&" + getTypeListId(origin.types) :
                            "#" + origin.type.id + "|" + getTypeListId(types); // origin type id alone is insufficient, as `keyof x` may resolve to multiple WIP values while `x` is still resolving
                var id = typeKey + getAliasId(aliasSymbol, aliasTypeArguments);
                var type = unionTypes.get(id);
                if (!type) {
                    type = createType(1048576 /* Union */);
                    type.objectFlags = objectFlags | getPropagatingFlagsOfTypes(types, /*excludeKinds*/ 98304 /* Nullable */);
                    type.types = types;
                    type.origin = origin;
                    type.aliasSymbol = aliasSymbol;
                    type.aliasTypeArguments = aliasTypeArguments;
                    if (types.length === 2 && types[0].flags & 512 /* BooleanLiteral */ && types[1].flags & 512 /* BooleanLiteral */) {
                        type.flags |= 16 /* Boolean */;
                        type.intrinsicName = "boolean";
                    }
                    unionTypes.set(id, type);
                }
                return type;
            }
            function getTypeFromUnionTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    var aliasSymbol = getAliasSymbolForTypeNode(node);
                    links.resolvedType = getUnionType(ts.map(node.types, getTypeFromTypeNode), 1 /* Literal */, aliasSymbol, getTypeArgumentsForAliasSymbol(aliasSymbol));
                }
                return links.resolvedType;
            }
            function addTypeToIntersection(typeSet, includes, type) {
                var flags = type.flags;
                if (flags & 2097152 /* Intersection */) {
                    return addTypesToIntersection(typeSet, includes, type.types);
                }
                if (isEmptyAnonymousObjectType(type)) {
                    if (!(includes & 16777216 /* IncludesEmptyObject */)) {
                        includes |= 16777216 /* IncludesEmptyObject */;
                        typeSet.set(type.id.toString(), type);
                    }
                }
                else {
                    if (flags & 3 /* AnyOrUnknown */) {
                        if (type === wildcardType)
                            includes |= 8388608 /* IncludesWildcard */;
                    }
                    else if ((strictNullChecks || !(flags & 98304 /* Nullable */)) && !typeSet.has(type.id.toString())) {
                        if (type.flags & 109440 /* Unit */ && includes & 109440 /* Unit */) {
                            // We have seen two distinct unit types which means we should reduce to an
                            // empty intersection. Adding TypeFlags.NonPrimitive causes that to happen.
                            includes |= 67108864 /* NonPrimitive */;
                        }
                        typeSet.set(type.id.toString(), type);
                    }
                    includes |= flags & 205258751 /* IncludesMask */;
                }
                return includes;
            }
            // Add the given types to the given type set. Order is preserved, freshness is removed from literal
            // types, duplicates are removed, and nested types of the given kind are flattened into the set.
            function addTypesToIntersection(typeSet, includes, types) {
                for (var _i = 0, types_12 = types; _i < types_12.length; _i++) {
                    var type = types_12[_i];
                    includes = addTypeToIntersection(typeSet, includes, getRegularTypeOfLiteralType(type));
                }
                return includes;
            }
            function removeRedundantPrimitiveTypes(types, includes) {
                var i = types.length;
                while (i > 0) {
                    i--;
                    var t = types[i];
                    var remove = t.flags & 4 /* String */ && includes & 128 /* StringLiteral */ ||
                        t.flags & 8 /* Number */ && includes & 256 /* NumberLiteral */ ||
                        t.flags & 64 /* BigInt */ && includes & 2048 /* BigIntLiteral */ ||
                        t.flags & 4096 /* ESSymbol */ && includes & 8192 /* UniqueESSymbol */;
                    if (remove) {
                        ts.orderedRemoveItemAt(types, i);
                    }
                }
            }
            // Check that the given type has a match in every union. A given type is matched by
            // an identical type, and a literal type is additionally matched by its corresponding
            // primitive type.
            function eachUnionContains(unionTypes, type) {
                for (var _i = 0, unionTypes_1 = unionTypes; _i < unionTypes_1.length; _i++) {
                    var u = unionTypes_1[_i];
                    if (!containsType(u.types, type)) {
                        var primitive = type.flags & 128 /* StringLiteral */ ? stringType :
                            type.flags & 256 /* NumberLiteral */ ? numberType :
                                type.flags & 2048 /* BigIntLiteral */ ? bigintType :
                                    type.flags & 8192 /* UniqueESSymbol */ ? esSymbolType :
                                        undefined;
                        if (!primitive || !containsType(u.types, primitive)) {
                            return false;
                        }
                    }
                }
                return true;
            }
            /**
             * Returns `true` if the intersection of the template literals and string literals is the empty set, eg `get${string}` & "setX", and should reduce to `never`
             */
            function extractRedundantTemplateLiterals(types) {
                var i = types.length;
                var literals = ts.filter(types, function (t) { return !!(t.flags & 128 /* StringLiteral */); });
                while (i > 0) {
                    i--;
                    var t = types[i];
                    if (!(t.flags & 134217728 /* TemplateLiteral */))
                        continue;
                    for (var _i = 0, literals_1 = literals; _i < literals_1.length; _i++) {
                        var t2 = literals_1[_i];
                        if (isTypeSubtypeOf(t2, t)) {
                            // eg, ``get${T}` & "getX"` is just `"getX"`
                            ts.orderedRemoveItemAt(types, i);
                            break;
                        }
                        else if (isPatternLiteralType(t)) {
                            return true;
                        }
                    }
                }
                return false;
            }
            function extractIrreducible(types, flag) {
                if (ts.every(types, function (t) { return !!(t.flags & 1048576 /* Union */) && ts.some(t.types, function (tt) { return !!(tt.flags & flag); }); })) {
                    for (var i = 0; i < types.length; i++) {
                        types[i] = filterType(types[i], function (t) { return !(t.flags & flag); });
                    }
                    return true;
                }
                return false;
            }
            // If the given list of types contains more than one union of primitive types, replace the
            // first with a union containing an intersection of those primitive types, then remove the
            // other unions and return true. Otherwise, do nothing and return false.
            function intersectUnionsOfPrimitiveTypes(types) {
                var unionTypes;
                var index = ts.findIndex(types, function (t) { return !!(ts.getObjectFlags(t) & 65536 /* PrimitiveUnion */); });
                if (index < 0) {
                    return false;
                }
                var i = index + 1;
                // Remove all but the first union of primitive types and collect them in
                // the unionTypes array.
                while (i < types.length) {
                    var t = types[i];
                    if (ts.getObjectFlags(t) & 65536 /* PrimitiveUnion */) {
                        (unionTypes || (unionTypes = [types[index]])).push(t);
                        ts.orderedRemoveItemAt(types, i);
                    }
                    else {
                        i++;
                    }
                }
                // Return false if there was only one union of primitive types
                if (!unionTypes) {
                    return false;
                }
                // We have more than one union of primitive types, now intersect them. For each
                // type in each union we check if the type is matched in every union and if so
                // we include it in the result.
                var checked = [];
                var result = [];
                for (var _i = 0, unionTypes_2 = unionTypes; _i < unionTypes_2.length; _i++) {
                    var u = unionTypes_2[_i];
                    for (var _a = 0, _b = u.types; _a < _b.length; _a++) {
                        var t = _b[_a];
                        if (insertType(checked, t)) {
                            if (eachUnionContains(unionTypes, t)) {
                                insertType(result, t);
                            }
                        }
                    }
                }
                // Finally replace the first union with the result
                types[index] = getUnionTypeFromSortedList(result, 65536 /* PrimitiveUnion */);
                return true;
            }
            function createIntersectionType(types, aliasSymbol, aliasTypeArguments) {
                var result = createType(2097152 /* Intersection */);
                result.objectFlags = getPropagatingFlagsOfTypes(types, /*excludeKinds*/ 98304 /* Nullable */);
                result.types = types;
                result.aliasSymbol = aliasSymbol;
                result.aliasTypeArguments = aliasTypeArguments;
                return result;
            }
            // We normalize combinations of intersection and union types based on the distributive property of the '&'
            // operator. Specifically, because X & (A | B) is equivalent to X & A | X & B, we can transform intersection
            // types with union type constituents into equivalent union types with intersection type constituents and
            // effectively ensure that union types are always at the top level in type representations.
            //
            // We do not perform structural deduplication on intersection types. Intersection types are created only by the &
            // type operator and we can't reduce those because we want to support recursive intersection types. For example,
            // a type alias of the form "type List<T> = T & { next: List<T> }" cannot be reduced during its declaration.
            // Also, unlike union types, the order of the constituent types is preserved in order that overload resolution
            // for intersections of types with signatures can be deterministic.
            function getIntersectionType(types, aliasSymbol, aliasTypeArguments) {
                var typeMembershipMap = new ts.Map();
                var includes = addTypesToIntersection(typeMembershipMap, 0, types);
                var typeSet = ts.arrayFrom(typeMembershipMap.values());
                // An intersection type is considered empty if it contains
                // the type never, or
                // more than one unit type or,
                // an object type and a nullable type (null or undefined), or
                // a string-like type and a type known to be non-string-like, or
                // a number-like type and a type known to be non-number-like, or
                // a symbol-like type and a type known to be non-symbol-like, or
                // a void-like type and a type known to be non-void-like, or
                // a non-primitive type and a type known to be primitive.
                if (includes & 131072 /* Never */) {
                    return ts.contains(typeSet, silentNeverType) ? silentNeverType : neverType;
                }
                if (strictNullChecks && includes & 98304 /* Nullable */ && includes & (524288 /* Object */ | 67108864 /* NonPrimitive */ | 16777216 /* IncludesEmptyObject */) ||
                    includes & 67108864 /* NonPrimitive */ && includes & (469892092 /* DisjointDomains */ & ~67108864 /* NonPrimitive */) ||
                    includes & 402653316 /* StringLike */ && includes & (469892092 /* DisjointDomains */ & ~402653316 /* StringLike */) ||
                    includes & 296 /* NumberLike */ && includes & (469892092 /* DisjointDomains */ & ~296 /* NumberLike */) ||
                    includes & 2112 /* BigIntLike */ && includes & (469892092 /* DisjointDomains */ & ~2112 /* BigIntLike */) ||
                    includes & 12288 /* ESSymbolLike */ && includes & (469892092 /* DisjointDomains */ & ~12288 /* ESSymbolLike */) ||
                    includes & 49152 /* VoidLike */ && includes & (469892092 /* DisjointDomains */ & ~49152 /* VoidLike */)) {
                    return neverType;
                }
                if (includes & 134217728 /* TemplateLiteral */ && includes & 128 /* StringLiteral */ && extractRedundantTemplateLiterals(typeSet)) {
                    return neverType;
                }
                if (includes & 1 /* Any */) {
                    return includes & 8388608 /* IncludesWildcard */ ? wildcardType : anyType;
                }
                if (!strictNullChecks && includes & 98304 /* Nullable */) {
                    return includes & 32768 /* Undefined */ ? undefinedType : nullType;
                }
                if (includes & 4 /* String */ && includes & 128 /* StringLiteral */ ||
                    includes & 8 /* Number */ && includes & 256 /* NumberLiteral */ ||
                    includes & 64 /* BigInt */ && includes & 2048 /* BigIntLiteral */ ||
                    includes & 4096 /* ESSymbol */ && includes & 8192 /* UniqueESSymbol */) {
                    removeRedundantPrimitiveTypes(typeSet, includes);
                }
                if (includes & 16777216 /* IncludesEmptyObject */ && includes & 524288 /* Object */) {
                    ts.orderedRemoveItemAt(typeSet, ts.findIndex(typeSet, isEmptyAnonymousObjectType));
                }
                if (typeSet.length === 0) {
                    return unknownType;
                }
                if (typeSet.length === 1) {
                    return typeSet[0];
                }
                var id = getTypeListId(typeSet) + getAliasId(aliasSymbol, aliasTypeArguments);
                var result = intersectionTypes.get(id);
                if (!result) {
                    if (includes & 1048576 /* Union */) {
                        if (intersectUnionsOfPrimitiveTypes(typeSet)) {
                            // When the intersection creates a reduced set (which might mean that *all* union types have
                            // disappeared), we restart the operation to get a new set of combined flags. Once we have
                            // reduced we'll never reduce again, so this occurs at most once.
                            result = getIntersectionType(typeSet, aliasSymbol, aliasTypeArguments);
                        }
                        else if (extractIrreducible(typeSet, 32768 /* Undefined */)) {
                            result = getUnionType([getIntersectionType(typeSet), undefinedType], 1 /* Literal */, aliasSymbol, aliasTypeArguments);
                        }
                        else if (extractIrreducible(typeSet, 65536 /* Null */)) {
                            result = getUnionType([getIntersectionType(typeSet), nullType], 1 /* Literal */, aliasSymbol, aliasTypeArguments);
                        }
                        else {
                            // We are attempting to construct a type of the form X & (A | B) & (C | D). Transform this into a type of
                            // the form X & A & C | X & A & D | X & B & C | X & B & D. If the estimated size of the resulting union type
                            // exceeds 100000 constituents, report an error.
                            if (!checkCrossProductUnion(typeSet)) {
                                return errorType;
                            }
                            var constituents = getCrossProductIntersections(typeSet);
                            // We attach a denormalized origin type when at least one constituent of the cross-product union is an
                            // intersection (i.e. when the intersection didn't just reduce one or more unions to smaller unions).
                            var origin = ts.some(constituents, function (t) { return !!(t.flags & 2097152 /* Intersection */); }) ? createOriginUnionOrIntersectionType(2097152 /* Intersection */, typeSet) : undefined;
                            result = getUnionType(constituents, 1 /* Literal */, aliasSymbol, aliasTypeArguments, origin);
                        }
                    }
                    else {
                        result = createIntersectionType(typeSet, aliasSymbol, aliasTypeArguments);
                    }
                    intersectionTypes.set(id, result);
                }
                return result;
            }
            function getCrossProductUnionSize(types) {
                return ts.reduceLeft(types, function (n, t) { return t.flags & 1048576 /* Union */ ? n * t.types.length : t.flags & 131072 /* Never */ ? 0 : n; }, 1);
            }
            function checkCrossProductUnion(types) {
                var size = getCrossProductUnionSize(types);
                if (size >= 100000) {
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.instant("checkTypes" /* CheckTypes */, "checkCrossProductUnion_DepthLimit", { typeIds: types.map(function (t) { return t.id; }), size: size });
                    error(currentNode, ts.Diagnostics.Expression_produces_a_union_type_that_is_too_complex_to_represent);
                    return false;
                }
                return true;
            }
            function getCrossProductIntersections(types) {
                var count = getCrossProductUnionSize(types);
                var intersections = [];
                for (var i = 0; i < count; i++) {
                    var constituents = types.slice();
                    var n = i;
                    for (var j = types.length - 1; j >= 0; j--) {
                        if (types[j].flags & 1048576 /* Union */) {
                            var sourceTypes = types[j].types;
                            var length_5 = sourceTypes.length;
                            constituents[j] = sourceTypes[n % length_5];
                            n = Math.floor(n / length_5);
                        }
                    }
                    var t = getIntersectionType(constituents);
                    if (!(t.flags & 131072 /* Never */))
                        intersections.push(t);
                }
                return intersections;
            }
            function getTypeFromIntersectionTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    var aliasSymbol = getAliasSymbolForTypeNode(node);
                    links.resolvedType = getIntersectionType(ts.map(node.types, getTypeFromTypeNode), aliasSymbol, getTypeArgumentsForAliasSymbol(aliasSymbol));
                }
                return links.resolvedType;
            }
            function createIndexType(type, stringsOnly) {
                var result = createType(4194304 /* Index */);
                result.type = type;
                result.stringsOnly = stringsOnly;
                return result;
            }
            function createOriginIndexType(type) {
                var result = createOriginType(4194304 /* Index */);
                result.type = type;
                return result;
            }
            function getIndexTypeForGenericType(type, stringsOnly) {
                return stringsOnly ?
                    type.resolvedStringIndexType || (type.resolvedStringIndexType = createIndexType(type, /*stringsOnly*/ true)) :
                    type.resolvedIndexType || (type.resolvedIndexType = createIndexType(type, /*stringsOnly*/ false));
            }
            function instantiateTypeAsMappedNameType(nameType, type, t) {
                return instantiateType(nameType, appendTypeMapping(type.mapper, getTypeParameterFromMappedType(type), t));
            }
            function getIndexTypeForMappedType(type, noIndexSignatures) {
                var constraint = filterType(getConstraintTypeFromMappedType(type), function (t) { return !(noIndexSignatures && t.flags & (1 /* Any */ | 4 /* String */)); });
                var nameType = type.declaration.nameType && getTypeFromTypeNode(type.declaration.nameType);
                // If the constraint is exclusively string/number/never type(s), we need to pull the property names from the modified type and run them through the `nameType` mapper as well
                // since they won't appear in the constraint, due to subtype reducing with the string/number index types
                var properties = nameType && everyType(constraint, function (t) { return !!(t.flags & (4 /* String */ | 8 /* Number */ | 131072 /* Never */)); }) && getPropertiesOfType(getApparentType(getModifiersTypeFromMappedType(type)));
                return nameType ?
                    getUnionType([mapType(constraint, function (t) { return instantiateTypeAsMappedNameType(nameType, type, t); }), mapType(getUnionType(ts.map(properties || ts.emptyArray, function (p) { return getLiteralTypeFromProperty(p, 8576 /* StringOrNumberLiteralOrUnique */); })), function (t) { return instantiateTypeAsMappedNameType(nameType, type, t); })]) :
                    constraint;
            }
            // Ordinarily we reduce a keyof M, where M is a mapped type { [P in K as N<P>]: X }, to simply N<K>. This however presumes
            // that N distributes over union types, i.e. that N<A | B | C> is equivalent to N<A> | N<B> | N<C>. Specifically, we only
            // want to perform the reduction when the name type of a mapped type is distributive with respect to the type variable
            // introduced by the 'in' clause of the mapped type. Note that non-generic types are considered to be distributive because
            // they're the same type regardless of what's being distributed over.
            function hasDistributiveNameType(mappedType) {
                var typeVariable = getTypeParameterFromMappedType(mappedType);
                return isDistributive(getNameTypeFromMappedType(mappedType) || typeVariable);
                function isDistributive(type) {
                    return type.flags & (3 /* AnyOrUnknown */ | 131068 /* Primitive */ | 131072 /* Never */ | 262144 /* TypeParameter */ | 524288 /* Object */ | 67108864 /* NonPrimitive */) ? true :
                        type.flags & 16777216 /* Conditional */ ? type.root.isDistributive && type.checkType === typeVariable :
                            type.flags & (3145728 /* UnionOrIntersection */ | 134217728 /* TemplateLiteral */) ? ts.every(type.types, isDistributive) :
                                type.flags & 8388608 /* IndexedAccess */ ? isDistributive(type.objectType) && isDistributive(type.indexType) :
                                    type.flags & 33554432 /* Substitution */ ? isDistributive(type.substitute) :
                                        type.flags & 268435456 /* StringMapping */ ? isDistributive(type.type) :
                                            false;
                }
            }
            function getLiteralTypeFromPropertyName(name) {
                if (ts.isPrivateIdentifier(name)) {
                    return neverType;
                }
                return ts.isIdentifier(name) ? getStringLiteralType(ts.unescapeLeadingUnderscores(name.escapedText)) :
                    getRegularTypeOfLiteralType(ts.isComputedPropertyName(name) ? checkComputedPropertyName(name) : checkExpression(name));
            }
            function getLiteralTypeFromProperty(prop, include, includeNonPublic) {
                if (includeNonPublic || !(ts.getDeclarationModifierFlagsFromSymbol(prop) & 24 /* NonPublicAccessibilityModifier */)) {
                    var type = getSymbolLinks(getLateBoundSymbol(prop)).nameType;
                    if (!type) {
                        var name = ts.getNameOfDeclaration(prop.valueDeclaration);
                        type = prop.escapedName === "default" /* Default */ ? getStringLiteralType("default") :
                            name && getLiteralTypeFromPropertyName(name) || (!ts.isKnownSymbol(prop) ? getStringLiteralType(ts.symbolName(prop)) : undefined);
                    }
                    if (type && type.flags & include) {
                        return type;
                    }
                }
                return neverType;
            }
            function getLiteralTypeFromProperties(type, include, includeOrigin) {
                var origin = includeOrigin && (ts.getObjectFlags(type) & (3 /* ClassOrInterface */ | 4 /* Reference */) || type.aliasSymbol) ? createOriginIndexType(type) : undefined;
                var propertyTypes = ts.map(getPropertiesOfType(type), function (prop) { return getLiteralTypeFromProperty(prop, include); });
                var indexKeyTypes = ts.map(getIndexInfosOfType(type), function (info) {
                    return info !== enumNumberIndexInfo && info.keyType.flags & include ?
                        info.keyType === stringType && include & 8 /* Number */ ? stringOrNumberType : info.keyType : neverType;
                });
                return getUnionType(ts.concatenate(propertyTypes, indexKeyTypes), 1 /* Literal */, 
                /*aliasSymbol*/ undefined, /*aliasTypeArguments*/ undefined, origin);
            }
            function getIndexType(type, stringsOnly, noIndexSignatures) {
                if (stringsOnly === void 0) {
                    stringsOnly = keyofStringsOnly;
                }
                type = getReducedType(type);
                return type.flags & 1048576 /* Union */ ? getIntersectionType(ts.map(type.types, function (t) { return getIndexType(t, stringsOnly, noIndexSignatures); })) :
                    type.flags & 2097152 /* Intersection */ ? getUnionType(ts.map(type.types, function (t) { return getIndexType(t, stringsOnly, noIndexSignatures); })) :
                        type.flags & 58982400 /* InstantiableNonPrimitive */ || isGenericTupleType(type) || isGenericMappedType(type) && !hasDistributiveNameType(type) ? getIndexTypeForGenericType(type, stringsOnly) :
                            ts.getObjectFlags(type) & 32 /* Mapped */ ? getIndexTypeForMappedType(type, noIndexSignatures) :
                                type === wildcardType ? wildcardType :
                                    type.flags & 2 /* Unknown */ ? neverType :
                                        type.flags & (1 /* Any */ | 131072 /* Never */) ? keyofConstraintType :
                                            getLiteralTypeFromProperties(type, (noIndexSignatures ? 128 /* StringLiteral */ : 402653316 /* StringLike */) | (stringsOnly ? 0 : 296 /* NumberLike */ | 12288 /* ESSymbolLike */), stringsOnly === keyofStringsOnly && !noIndexSignatures);
            }
            function getExtractStringType(type) {
                if (keyofStringsOnly) {
                    return type;
                }
                var extractTypeAlias = getGlobalExtractSymbol();
                return extractTypeAlias ? getTypeAliasInstantiation(extractTypeAlias, [type, stringType]) : stringType;
            }
            function getIndexTypeOrString(type) {
                var indexType = getExtractStringType(getIndexType(type));
                return indexType.flags & 131072 /* Never */ ? stringType : indexType;
            }
            function getTypeFromTypeOperatorNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    switch (node.operator) {
                        case 139 /* KeyOfKeyword */:
                            links.resolvedType = getIndexType(getTypeFromTypeNode(node.type));
                            break;
                        case 152 /* UniqueKeyword */:
                            links.resolvedType = node.type.kind === 149 /* SymbolKeyword */
                                ? getESSymbolLikeTypeForNode(ts.walkUpParenthesizedTypes(node.parent))
                                : errorType;
                            break;
                        case 143 /* ReadonlyKeyword */:
                            links.resolvedType = getTypeFromTypeNode(node.type);
                            break;
                        default:
                            throw ts.Debug.assertNever(node.operator);
                    }
                }
                return links.resolvedType;
            }
            function getTypeFromTemplateTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    links.resolvedType = getTemplateLiteralType(__spreadArray([node.head.text], ts.map(node.templateSpans, function (span) { return span.literal.text; }), true), ts.map(node.templateSpans, function (span) { return getTypeFromTypeNode(span.type); }));
                }
                return links.resolvedType;
            }
            function getTemplateLiteralType(texts, types) {
                var unionIndex = ts.findIndex(types, function (t) { return !!(t.flags & (131072 /* Never */ | 1048576 /* Union */)); });
                if (unionIndex >= 0) {
                    return checkCrossProductUnion(types) ?
                        mapType(types[unionIndex], function (t) { return getTemplateLiteralType(texts, ts.replaceElement(types, unionIndex, t)); }) :
                        errorType;
                }
                if (ts.contains(types, wildcardType)) {
                    return wildcardType;
                }
                var newTypes = [];
                var newTexts = [];
                var text = texts[0];
                if (!addSpans(texts, types)) {
                    return stringType;
                }
                if (newTypes.length === 0) {
                    return getStringLiteralType(text);
                }
                newTexts.push(text);
                if (ts.every(newTexts, function (t) { return t === ""; }) && ts.every(newTypes, function (t) { return !!(t.flags & 4 /* String */); })) {
                    return stringType;
                }
                var id = getTypeListId(newTypes) + "|" + ts.map(newTexts, function (t) { return t.length; }).join(",") + "|" + newTexts.join("");
                var type = templateLiteralTypes.get(id);
                if (!type) {
                    templateLiteralTypes.set(id, type = createTemplateLiteralType(newTexts, newTypes));
                }
                return type;
                function addSpans(texts, types) {
                    for (var i = 0; i < types.length; i++) {
                        var t = types[i];
                        if (t.flags & (2944 /* Literal */ | 65536 /* Null */ | 32768 /* Undefined */)) {
                            text += getTemplateStringForType(t) || "";
                            text += texts[i + 1];
                        }
                        else if (t.flags & 134217728 /* TemplateLiteral */) {
                            text += t.texts[0];
                            if (!addSpans(t.texts, t.types))
                                return false;
                            text += texts[i + 1];
                        }
                        else if (isGenericIndexType(t) || isPatternLiteralPlaceholderType(t)) {
                            newTypes.push(t);
                            newTexts.push(text);
                            text = texts[i + 1];
                        }
                        else {
                            return false;
                        }
                    }
                    return true;
                }
            }
            function getTemplateStringForType(type) {
                return type.flags & 128 /* StringLiteral */ ? type.value :
                    type.flags & 256 /* NumberLiteral */ ? "" + type.value :
                        type.flags & 2048 /* BigIntLiteral */ ? ts.pseudoBigIntToString(type.value) :
                            type.flags & (512 /* BooleanLiteral */ | 98304 /* Nullable */) ? type.intrinsicName :
                                undefined;
            }
            function createTemplateLiteralType(texts, types) {
                var type = createType(134217728 /* TemplateLiteral */);
                type.texts = texts;
                type.types = types;
                return type;
            }
            function getStringMappingType(symbol, type) {
                return type.flags & (1048576 /* Union */ | 131072 /* Never */) ? mapType(type, function (t) { return getStringMappingType(symbol, t); }) :
                    isGenericIndexType(type) ? getStringMappingTypeForGenericType(symbol, type) :
                        type.flags & 128 /* StringLiteral */ ? getStringLiteralType(applyStringMapping(symbol, type.value)) :
                            type;
            }
            function applyStringMapping(symbol, str) {
                switch (intrinsicTypeKinds.get(symbol.escapedName)) {
                    case 0 /* Uppercase */: return str.toUpperCase();
                    case 1 /* Lowercase */: return str.toLowerCase();
                    case 2 /* Capitalize */: return str.charAt(0).toUpperCase() + str.slice(1);
                    case 3 /* Uncapitalize */: return str.charAt(0).toLowerCase() + str.slice(1);
                }
                return str;
            }
            function getStringMappingTypeForGenericType(symbol, type) {
                var id = getSymbolId(symbol) + "," + getTypeId(type);
                var result = stringMappingTypes.get(id);
                if (!result) {
                    stringMappingTypes.set(id, result = createStringMappingType(symbol, type));
                }
                return result;
            }
            function createStringMappingType(symbol, type) {
                var result = createType(268435456 /* StringMapping */);
                result.symbol = symbol;
                result.type = type;
                return result;
            }
            function createIndexedAccessType(objectType, indexType, accessFlags, aliasSymbol, aliasTypeArguments) {
                var type = createType(8388608 /* IndexedAccess */);
                type.objectType = objectType;
                type.indexType = indexType;
                type.accessFlags = accessFlags;
                type.aliasSymbol = aliasSymbol;
                type.aliasTypeArguments = aliasTypeArguments;
                return type;
            }
            /**
             * Returns if a type is or consists of a JSLiteral object type
             * In addition to objects which are directly literals,
             * * unions where every element is a jsliteral
             * * intersections where at least one element is a jsliteral
             * * and instantiable types constrained to a jsliteral
             * Should all count as literals and not print errors on access or assignment of possibly existing properties.
             * This mirrors the behavior of the index signature propagation, to which this behaves similarly (but doesn't affect assignability or inference).
             */
            function isJSLiteralType(type) {
                if (noImplicitAny) {
                    return false; // Flag is meaningless under `noImplicitAny` mode
                }
                if (ts.getObjectFlags(type) & 8192 /* JSLiteral */) {
                    return true;
                }
                if (type.flags & 1048576 /* Union */) {
                    return ts.every(type.types, isJSLiteralType);
                }
                if (type.flags & 2097152 /* Intersection */) {
                    return ts.some(type.types, isJSLiteralType);
                }
                if (type.flags & 465829888 /* Instantiable */) {
                    var constraint = getResolvedBaseConstraint(type);
                    return constraint !== type && isJSLiteralType(constraint);
                }
                return false;
            }
            function getPropertyNameFromIndex(indexType, accessNode) {
                return isTypeUsableAsPropertyName(indexType) ?
                    getPropertyNameFromType(indexType) :
                    accessNode && ts.isPropertyName(accessNode) ?
                        // late bound names are handled in the first branch, so here we only need to handle normal names
                        ts.getPropertyNameForPropertyNameNode(accessNode) :
                        undefined;
            }
            function isUncalledFunctionReference(node, symbol) {
                if (symbol.flags & (16 /* Function */ | 8192 /* Method */)) {
                    var parent = ts.findAncestor(node.parent, function (n) { return !ts.isAccessExpression(n); }) || node.parent;
                    if (ts.isCallLikeExpression(parent)) {
                        return ts.isCallOrNewExpression(parent) && ts.isIdentifier(node) && hasMatchingArgument(parent, node);
                    }
                    return ts.every(symbol.declarations, function (d) { return !ts.isFunctionLike(d) || !!(ts.getCombinedNodeFlags(d) & 134217728 /* Deprecated */); });
                }
                return true;
            }
            function getPropertyTypeForIndexType(originalObjectType, objectType, indexType, fullIndexType, accessNode, accessFlags) {
                var _a;
                var accessExpression = accessNode && accessNode.kind === 205 /* ElementAccessExpression */ ? accessNode : undefined;
                var propName = accessNode && ts.isPrivateIdentifier(accessNode) ? undefined : getPropertyNameFromIndex(indexType, accessNode);
                if (propName !== undefined) {
                    if (accessFlags & 256 /* Contextual */) {
                        return getTypeOfPropertyOfContextualType(objectType, propName) || anyType;
                    }
                    var prop = getPropertyOfType(objectType, propName);
                    if (prop) {
                        if (accessFlags & 64 /* ReportDeprecated */ && accessNode && prop.declarations && getDeclarationNodeFlagsFromSymbol(prop) & 134217728 /* Deprecated */ && isUncalledFunctionReference(accessNode, prop)) {
                            var deprecatedNode = (_a = accessExpression === null || accessExpression === void 0 ? void 0 : accessExpression.argumentExpression) !== null && _a !== void 0 ? _a : (ts.isIndexedAccessTypeNode(accessNode) ? accessNode.indexType : accessNode);
                            addDeprecatedSuggestion(deprecatedNode, prop.declarations, propName);
                        }
                        if (accessExpression) {
                            markPropertyAsReferenced(prop, accessExpression, isSelfTypeAccess(accessExpression.expression, objectType.symbol));
                            if (isAssignmentToReadonlyEntity(accessExpression, prop, ts.getAssignmentTargetKind(accessExpression))) {
                                error(accessExpression.argumentExpression, ts.Diagnostics.Cannot_assign_to_0_because_it_is_a_read_only_property, symbolToString(prop));
                                return undefined;
                            }
                            if (accessFlags & 8 /* CacheSymbol */) {
                                getNodeLinks(accessNode).resolvedSymbol = prop;
                            }
                            if (isThisPropertyAccessInConstructor(accessExpression, prop)) {
                                return autoType;
                            }
                        }
                        var propType = getTypeOfSymbol(prop);
                        return accessExpression && ts.getAssignmentTargetKind(accessExpression) !== 1 /* Definite */ ?
                            getFlowTypeOfReference(accessExpression, propType) :
                            propType;
                    }
                    if (everyType(objectType, isTupleType) && isNumericLiteralName(propName) && +propName >= 0) {
                        if (accessNode && everyType(objectType, function (t) { return !t.target.hasRestElement; }) && !(accessFlags & 16 /* NoTupleBoundsCheck */)) {
                            var indexNode = getIndexNodeForAccessExpression(accessNode);
                            if (isTupleType(objectType)) {
                                error(indexNode, ts.Diagnostics.Tuple_type_0_of_length_1_has_no_element_at_index_2, typeToString(objectType), getTypeReferenceArity(objectType), ts.unescapeLeadingUnderscores(propName));
                            }
                            else {
                                error(indexNode, ts.Diagnostics.Property_0_does_not_exist_on_type_1, ts.unescapeLeadingUnderscores(propName), typeToString(objectType));
                            }
                        }
                        errorIfWritingToReadonlyIndex(getIndexInfoOfType(objectType, numberType));
                        return mapType(objectType, function (t) {
                            var restType = getRestTypeOfTupleType(t) || undefinedType;
                            return accessFlags & 1 /* IncludeUndefined */ ? getUnionType([restType, undefinedType]) : restType;
                        });
                    }
                }
                if (!(indexType.flags & 98304 /* Nullable */) && isTypeAssignableToKind(indexType, 402653316 /* StringLike */ | 296 /* NumberLike */ | 12288 /* ESSymbolLike */)) {
                    if (objectType.flags & (1 /* Any */ | 131072 /* Never */)) {
                        return objectType;
                    }
                    // If no index signature is applicable, we default to the string index signature. In effect, this means the string
                    // index signature applies even when accessing with a symbol-like type.
                    var indexInfo = getApplicableIndexInfo(objectType, indexType) || getIndexInfoOfType(objectType, stringType);
                    if (indexInfo) {
                        if (accessFlags & 2 /* NoIndexSignatures */ && indexInfo.keyType !== numberType) {
                            if (accessExpression) {
                                error(accessExpression, ts.Diagnostics.Type_0_cannot_be_used_to_index_type_1, typeToString(indexType), typeToString(originalObjectType));
                            }
                            return undefined;
                        }
                        if (accessNode && indexInfo.keyType === stringType && !isTypeAssignableToKind(indexType, 4 /* String */ | 8 /* Number */)) {
                            var indexNode = getIndexNodeForAccessExpression(accessNode);
                            error(indexNode, ts.Diagnostics.Type_0_cannot_be_used_as_an_index_type, typeToString(indexType));
                            return accessFlags & 1 /* IncludeUndefined */ ? getUnionType([indexInfo.type, undefinedType]) : indexInfo.type;
                        }
                        errorIfWritingToReadonlyIndex(indexInfo);
                        return accessFlags & 1 /* IncludeUndefined */ ? getUnionType([indexInfo.type, undefinedType]) : indexInfo.type;
                    }
                    if (indexType.flags & 131072 /* Never */) {
                        return neverType;
                    }
                    if (isJSLiteralType(objectType)) {
                        return anyType;
                    }
                    if (accessExpression && !isConstEnumObjectType(objectType)) {
                        if (isObjectLiteralType(objectType)) {
                            if (noImplicitAny && indexType.flags & (128 /* StringLiteral */ | 256 /* NumberLiteral */)) {
                                diagnostics.add(ts.createDiagnosticForNode(accessExpression, ts.Diagnostics.Property_0_does_not_exist_on_type_1, indexType.value, typeToString(objectType)));
                                return undefinedType;
                            }
                            else if (indexType.flags & (8 /* Number */ | 4 /* String */)) {
                                var types = ts.map(objectType.properties, function (property) {
                                    return getTypeOfSymbol(property);
                                });
                                return getUnionType(ts.append(types, undefinedType));
                            }
                        }
                        if (objectType.symbol === globalThisSymbol && propName !== undefined && globalThisSymbol.exports.has(propName) && (globalThisSymbol.exports.get(propName).flags & 418 /* BlockScoped */)) {
                            error(accessExpression, ts.Diagnostics.Property_0_does_not_exist_on_type_1, ts.unescapeLeadingUnderscores(propName), typeToString(objectType));
                        }
                        else if (noImplicitAny && !compilerOptions.suppressImplicitAnyIndexErrors && !(accessFlags & 128 /* SuppressNoImplicitAnyError */)) {
                            if (propName !== undefined && typeHasStaticProperty(propName, objectType)) {
                                var typeName = typeToString(objectType);
                                error(accessExpression, ts.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_to_access_the_static_member_2_instead, propName, typeName, typeName + "[" + ts.getTextOfNode(accessExpression.argumentExpression) + "]");
                            }
                            else if (getIndexTypeOfType(objectType, numberType)) {
                                error(accessExpression.argumentExpression, ts.Diagnostics.Element_implicitly_has_an_any_type_because_index_expression_is_not_of_type_number);
                            }
                            else {
                                var suggestion = void 0;
                                if (propName !== undefined && (suggestion = getSuggestionForNonexistentProperty(propName, objectType))) {
                                    if (suggestion !== undefined) {
                                        error(accessExpression.argumentExpression, ts.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2, propName, typeToString(objectType), suggestion);
                                    }
                                }
                                else {
                                    var suggestion_1 = getSuggestionForNonexistentIndexSignature(objectType, accessExpression, indexType);
                                    if (suggestion_1 !== undefined) {
                                        error(accessExpression, ts.Diagnostics.Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature_Did_you_mean_to_call_1, typeToString(objectType), suggestion_1);
                                    }
                                    else {
                                        var errorInfo = void 0;
                                        if (indexType.flags & 1024 /* EnumLiteral */) {
                                            errorInfo = ts.chainDiagnosticMessages(/* details */ undefined, ts.Diagnostics.Property_0_does_not_exist_on_type_1, "[" + typeToString(indexType) + "]", typeToString(objectType));
                                        }
                                        else if (indexType.flags & 8192 /* UniqueESSymbol */) {
                                            var symbolName_2 = getFullyQualifiedName(indexType.symbol, accessExpression);
                                            errorInfo = ts.chainDiagnosticMessages(/* details */ undefined, ts.Diagnostics.Property_0_does_not_exist_on_type_1, "[" + symbolName_2 + "]", typeToString(objectType));
                                        }
                                        else if (indexType.flags & 128 /* StringLiteral */) {
                                            errorInfo = ts.chainDiagnosticMessages(/* details */ undefined, ts.Diagnostics.Property_0_does_not_exist_on_type_1, indexType.value, typeToString(objectType));
                                        }
                                        else if (indexType.flags & 256 /* NumberLiteral */) {
                                            errorInfo = ts.chainDiagnosticMessages(/* details */ undefined, ts.Diagnostics.Property_0_does_not_exist_on_type_1, indexType.value, typeToString(objectType));
                                        }
                                        else if (indexType.flags & (8 /* Number */ | 4 /* String */)) {
                                            errorInfo = ts.chainDiagnosticMessages(/* details */ undefined, ts.Diagnostics.No_index_signature_with_a_parameter_of_type_0_was_found_on_type_1, typeToString(indexType), typeToString(objectType));
                                        }
                                        errorInfo = ts.chainDiagnosticMessages(errorInfo, ts.Diagnostics.Element_implicitly_has_an_any_type_because_expression_of_type_0_can_t_be_used_to_index_type_1, typeToString(fullIndexType), typeToString(objectType));
                                        diagnostics.add(ts.createDiagnosticForNodeFromMessageChain(accessExpression, errorInfo));
                                    }
                                }
                            }
                        }
                        return undefined;
                    }
                }
                if (isJSLiteralType(objectType)) {
                    return anyType;
                }
                if (accessNode) {
                    var indexNode = getIndexNodeForAccessExpression(accessNode);
                    if (indexType.flags & (128 /* StringLiteral */ | 256 /* NumberLiteral */)) {
                        error(indexNode, ts.Diagnostics.Property_0_does_not_exist_on_type_1, "" + indexType.value, typeToString(objectType));
                    }
                    else if (indexType.flags & (4 /* String */ | 8 /* Number */)) {
                        error(indexNode, ts.Diagnostics.Type_0_has_no_matching_index_signature_for_type_1, typeToString(objectType), typeToString(indexType));
                    }
                    else {
                        error(indexNode, ts.Diagnostics.Type_0_cannot_be_used_as_an_index_type, typeToString(indexType));
                    }
                }
                if (isTypeAny(indexType)) {
                    return indexType;
                }
                return undefined;
                function errorIfWritingToReadonlyIndex(indexInfo) {
                    if (indexInfo && indexInfo.isReadonly && accessExpression && (ts.isAssignmentTarget(accessExpression) || ts.isDeleteTarget(accessExpression))) {
                        error(accessExpression, ts.Diagnostics.Index_signature_in_type_0_only_permits_reading, typeToString(objectType));
                    }
                }
            }
            function getIndexNodeForAccessExpression(accessNode) {
                return accessNode.kind === 205 /* ElementAccessExpression */ ? accessNode.argumentExpression :
                    accessNode.kind === 192 /* IndexedAccessType */ ? accessNode.indexType :
                        accessNode.kind === 160 /* ComputedPropertyName */ ? accessNode.expression :
                            accessNode;
            }
            function isPatternLiteralPlaceholderType(type) {
                return !!(type.flags & (1 /* Any */ | 4 /* String */ | 8 /* Number */ | 64 /* BigInt */));
            }
            function isPatternLiteralType(type) {
                return !!(type.flags & 134217728 /* TemplateLiteral */) && ts.every(type.types, isPatternLiteralPlaceholderType);
            }
            function isGenericType(type) {
                return !!getGenericObjectFlags(type);
            }
            function isGenericObjectType(type) {
                return !!(getGenericObjectFlags(type) & 8388608 /* IsGenericObjectType */);
            }
            function isGenericIndexType(type) {
                return !!(getGenericObjectFlags(type) & 16777216 /* IsGenericIndexType */);
            }
            function getGenericObjectFlags(type) {
                if (type.flags & 3145728 /* UnionOrIntersection */) {
                    if (!(type.objectFlags & 4194304 /* IsGenericTypeComputed */)) {
                        type.objectFlags |= 4194304 /* IsGenericTypeComputed */ |
                            ts.reduceLeft(type.types, function (flags, t) { return flags | getGenericObjectFlags(t); }, 0);
                    }
                    return type.objectFlags & 25165824 /* IsGenericType */;
                }
                if (type.flags & 33554432 /* Substitution */) {
                    if (!(type.objectFlags & 4194304 /* IsGenericTypeComputed */)) {
                        type.objectFlags |= 4194304 /* IsGenericTypeComputed */ |
                            getGenericObjectFlags(type.substitute) | getGenericObjectFlags(type.baseType);
                    }
                    return type.objectFlags & 25165824 /* IsGenericType */;
                }
                return (type.flags & 58982400 /* InstantiableNonPrimitive */ || isGenericMappedType(type) || isGenericTupleType(type) ? 8388608 /* IsGenericObjectType */ : 0) |
                    (type.flags & (58982400 /* InstantiableNonPrimitive */ | 4194304 /* Index */ | 134217728 /* TemplateLiteral */ | 268435456 /* StringMapping */) && !isPatternLiteralType(type) ? 16777216 /* IsGenericIndexType */ : 0);
            }
            function isThisTypeParameter(type) {
                return !!(type.flags & 262144 /* TypeParameter */ && type.isThisType);
            }
            function getSimplifiedType(type, writing) {
                return type.flags & 8388608 /* IndexedAccess */ ? getSimplifiedIndexedAccessType(type, writing) :
                    type.flags & 16777216 /* Conditional */ ? getSimplifiedConditionalType(type, writing) :
                        type;
            }
            function distributeIndexOverObjectType(objectType, indexType, writing) {
                // (T | U)[K] -> T[K] | U[K] (reading)
                // (T | U)[K] -> T[K] & U[K] (writing)
                // (T & U)[K] -> T[K] & U[K]
                if (objectType.flags & 3145728 /* UnionOrIntersection */) {
                    var types = ts.map(objectType.types, function (t) { return getSimplifiedType(getIndexedAccessType(t, indexType), writing); });
                    return objectType.flags & 2097152 /* Intersection */ || writing ? getIntersectionType(types) : getUnionType(types);
                }
            }
            function distributeObjectOverIndexType(objectType, indexType, writing) {
                // T[A | B] -> T[A] | T[B] (reading)
                // T[A | B] -> T[A] & T[B] (writing)
                if (indexType.flags & 1048576 /* Union */) {
                    var types = ts.map(indexType.types, function (t) { return getSimplifiedType(getIndexedAccessType(objectType, t), writing); });
                    return writing ? getIntersectionType(types) : getUnionType(types);
                }
            }
            // Transform an indexed access to a simpler form, if possible. Return the simpler form, or return
            // the type itself if no transformation is possible. The writing flag indicates that the type is
            // the target of an assignment.
            function getSimplifiedIndexedAccessType(type, writing) {
                var cache = writing ? "simplifiedForWriting" : "simplifiedForReading";
                if (type[cache]) {
                    return type[cache] === circularConstraintType ? type : type[cache];
                }
                type[cache] = circularConstraintType;
                // We recursively simplify the object type as it may in turn be an indexed access type. For example, with
                // '{ [P in T]: { [Q in U]: number } }[T][U]' we want to first simplify the inner indexed access type.
                var objectType = getSimplifiedType(type.objectType, writing);
                var indexType = getSimplifiedType(type.indexType, writing);
                // T[A | B] -> T[A] | T[B] (reading)
                // T[A | B] -> T[A] & T[B] (writing)
                var distributedOverIndex = distributeObjectOverIndexType(objectType, indexType, writing);
                if (distributedOverIndex) {
                    return type[cache] = distributedOverIndex;
                }
                // Only do the inner distributions if the index can no longer be instantiated to cause index distribution again
                if (!(indexType.flags & 465829888 /* Instantiable */)) {
                    // (T | U)[K] -> T[K] | U[K] (reading)
                    // (T | U)[K] -> T[K] & U[K] (writing)
                    // (T & U)[K] -> T[K] & U[K]
                    var distributedOverObject = distributeIndexOverObjectType(objectType, indexType, writing);
                    if (distributedOverObject) {
                        return type[cache] = distributedOverObject;
                    }
                }
                // So ultimately (reading):
                // ((A & B) | C)[K1 | K2] -> ((A & B) | C)[K1] | ((A & B) | C)[K2] -> (A & B)[K1] | C[K1] | (A & B)[K2] | C[K2] -> (A[K1] & B[K1]) | C[K1] | (A[K2] & B[K2]) | C[K2]
                // A generic tuple type indexed by a number exists only when the index type doesn't select a
                // fixed element. We simplify to either the combined type of all elements (when the index type
                // the actual number type) or to the combined type of all non-fixed elements.
                if (isGenericTupleType(objectType) && indexType.flags & 296 /* NumberLike */) {
                    var elementType = getElementTypeOfSliceOfTupleType(objectType, indexType.flags & 8 /* Number */ ? 0 : objectType.target.fixedLength, /*endSkipCount*/ 0, writing);
                    if (elementType) {
                        return type[cache] = elementType;
                    }
                }
                // If the object type is a mapped type { [P in K]: E }, where K is generic, instantiate E using a mapper
                // that substitutes the index type for P. For example, for an index access { [P in K]: Box<T[P]> }[X], we
                // construct the type Box<T[X]>.
                if (isGenericMappedType(objectType)) {
                    return type[cache] = mapType(substituteIndexedMappedType(objectType, type.indexType), function (t) { return getSimplifiedType(t, writing); });
                }
                return type[cache] = type;
            }
            function isConditionalTypeAlwaysTrueDisregardingInferTypes(type) {
                var extendsInferParamMapper = type.root.inferTypeParameters && createTypeMapper(type.root.inferTypeParameters, ts.map(type.root.inferTypeParameters, function () { return wildcardType; }));
                var checkType = type.checkType;
                var extendsType = type.extendsType;
                return isTypeAssignableTo(getRestrictiveInstantiation(checkType), getRestrictiveInstantiation(instantiateType(extendsType, extendsInferParamMapper)));
            }
            function getSimplifiedConditionalType(type, writing) {
                var checkType = type.checkType;
                var extendsType = type.extendsType;
                var trueType = getTrueTypeFromConditionalType(type);
                var falseType = getFalseTypeFromConditionalType(type);
                // Simplifications for types of the form `T extends U ? T : never` and `T extends U ? never : T`.
                if (falseType.flags & 131072 /* Never */ && getActualTypeVariable(trueType) === getActualTypeVariable(checkType)) {
                    if (checkType.flags & 1 /* Any */ || isTypeAssignableTo(getRestrictiveInstantiation(checkType), getRestrictiveInstantiation(extendsType))) { // Always true
                        return getSimplifiedType(trueType, writing);
                    }
                    else if (isIntersectionEmpty(checkType, extendsType)) { // Always false
                        return neverType;
                    }
                }
                else if (trueType.flags & 131072 /* Never */ && getActualTypeVariable(falseType) === getActualTypeVariable(checkType)) {
                    if (!(checkType.flags & 1 /* Any */) && isTypeAssignableTo(getRestrictiveInstantiation(checkType), getRestrictiveInstantiation(extendsType))) { // Always true
                        return neverType;
                    }
                    else if (checkType.flags & 1 /* Any */ || isIntersectionEmpty(checkType, extendsType)) { // Always false
                        return getSimplifiedType(falseType, writing);
                    }
                }
                return type;
            }
            /**
             * Invokes union simplification logic to determine if an intersection is considered empty as a union constituent
             */
            function isIntersectionEmpty(type1, type2) {
                return !!(getUnionType([intersectTypes(type1, type2), neverType]).flags & 131072 /* Never */);
            }
            function substituteIndexedMappedType(objectType, index) {
                var mapper = createTypeMapper([getTypeParameterFromMappedType(objectType)], [index]);
                var templateMapper = combineTypeMappers(objectType.mapper, mapper);
                return instantiateType(getTemplateTypeFromMappedType(objectType), templateMapper);
            }
            function getIndexedAccessType(objectType, indexType, accessFlags, accessNode, aliasSymbol, aliasTypeArguments) {
                if (accessFlags === void 0) {
                    accessFlags = 0 /* None */;
                }
                return getIndexedAccessTypeOrUndefined(objectType, indexType, accessFlags, accessNode, aliasSymbol, aliasTypeArguments) || (accessNode ? errorType : unknownType);
            }
            function indexTypeLessThan(indexType, limit) {
                return everyType(indexType, function (t) {
                    if (t.flags & 384 /* StringOrNumberLiteral */) {
                        var propName = getPropertyNameFromType(t);
                        if (isNumericLiteralName(propName)) {
                            var index = +propName;
                            return index >= 0 && index < limit;
                        }
                    }
                    return false;
                });
            }
            function getIndexedAccessTypeOrUndefined(objectType, indexType, accessFlags, accessNode, aliasSymbol, aliasTypeArguments) {
                if (accessFlags === void 0) {
                    accessFlags = 0 /* None */;
                }
                if (objectType === wildcardType || indexType === wildcardType) {
                    return wildcardType;
                }
                // If the object type has a string index signature and no other members we know that the result will
                // always be the type of that index signature and we can simplify accordingly.
                if (isStringIndexSignatureOnlyType(objectType) && !(indexType.flags & 98304 /* Nullable */) && isTypeAssignableToKind(indexType, 4 /* String */ | 8 /* Number */)) {
                    indexType = stringType;
                }
                // In noUncheckedIndexedAccess mode, indexed access operations that occur in an expression in a read position and resolve to
                // an index signature have 'undefined' included in their type.
                if (compilerOptions.noUncheckedIndexedAccess && accessFlags & 32 /* ExpressionPosition */)
                    accessFlags |= 1 /* IncludeUndefined */;
                // If the index type is generic, or if the object type is generic and doesn't originate in an expression and
                // the operation isn't exclusively indexing the fixed (non-variadic) portion of a tuple type, we are performing
                // a higher-order index access where we cannot meaningfully access the properties of the object type. Note that
                // for a generic T and a non-generic K, we eagerly resolve T[K] if it originates in an expression. This is to
                // preserve backwards compatibility. For example, an element access 'this["foo"]' has always been resolved
                // eagerly using the constraint type of 'this' at the given location.
                if (isGenericIndexType(indexType) || (accessNode && accessNode.kind !== 192 /* IndexedAccessType */ ?
                    isGenericTupleType(objectType) && !indexTypeLessThan(indexType, objectType.target.fixedLength) :
                    isGenericObjectType(objectType) && !(isTupleType(objectType) && indexTypeLessThan(indexType, objectType.target.fixedLength)))) {
                    if (objectType.flags & 3 /* AnyOrUnknown */) {
                        return objectType;
                    }
                    // Defer the operation by creating an indexed access type.
                    var persistentAccessFlags = accessFlags & 1 /* Persistent */;
                    var id = objectType.id + "," + indexType.id + "," + persistentAccessFlags + getAliasId(aliasSymbol, aliasTypeArguments);
                    var type = indexedAccessTypes.get(id);
                    if (!type) {
                        indexedAccessTypes.set(id, type = createIndexedAccessType(objectType, indexType, persistentAccessFlags, aliasSymbol, aliasTypeArguments));
                    }
                    return type;
                }
                // In the following we resolve T[K] to the type of the property in T selected by K.
                // We treat boolean as different from other unions to improve errors;
                // skipping straight to getPropertyTypeForIndexType gives errors with 'boolean' instead of 'true'.
                var apparentObjectType = getReducedApparentType(objectType);
                if (indexType.flags & 1048576 /* Union */ && !(indexType.flags & 16 /* Boolean */)) {
                    var propTypes = [];
                    var wasMissingProp = false;
                    for (var _i = 0, _a = indexType.types; _i < _a.length; _i++) {
                        var t = _a[_i];
                        var propType = getPropertyTypeForIndexType(objectType, apparentObjectType, t, indexType, accessNode, accessFlags | (wasMissingProp ? 128 /* SuppressNoImplicitAnyError */ : 0));
                        if (propType) {
                            propTypes.push(propType);
                        }
                        else if (!accessNode) {
                            // If there's no error node, we can immeditely stop, since error reporting is off
                            return undefined;
                        }
                        else {
                            // Otherwise we set a flag and return at the end of the loop so we still mark all errors
                            wasMissingProp = true;
                        }
                    }
                    if (wasMissingProp) {
                        return undefined;
                    }
                    return accessFlags & 4 /* Writing */
                        ? getIntersectionType(propTypes, aliasSymbol, aliasTypeArguments)
                        : getUnionType(propTypes, 1 /* Literal */, aliasSymbol, aliasTypeArguments);
                }
                return getPropertyTypeForIndexType(objectType, apparentObjectType, indexType, indexType, accessNode, accessFlags | 8 /* CacheSymbol */ | 64 /* ReportDeprecated */);
            }
            function getTypeFromIndexedAccessTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    var objectType = getTypeFromTypeNode(node.objectType);
                    var indexType = getTypeFromTypeNode(node.indexType);
                    var potentialAlias = getAliasSymbolForTypeNode(node);
                    var resolved = getIndexedAccessType(objectType, indexType, 0 /* None */, node, potentialAlias, getTypeArgumentsForAliasSymbol(potentialAlias));
                    links.resolvedType = resolved.flags & 8388608 /* IndexedAccess */ &&
                        resolved.objectType === objectType &&
                        resolved.indexType === indexType ?
                        getConditionalFlowTypeOfType(resolved, node) : resolved;
                }
                return links.resolvedType;
            }
            function getTypeFromMappedTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    var type = createObjectType(32 /* Mapped */, node.symbol);
                    type.declaration = node;
                    type.aliasSymbol = getAliasSymbolForTypeNode(node);
                    type.aliasTypeArguments = getTypeArgumentsForAliasSymbol(type.aliasSymbol);
                    links.resolvedType = type;
                    // Eagerly resolve the constraint type which forces an error if the constraint type circularly
                    // references itself through one or more type aliases.
                    getConstraintTypeFromMappedType(type);
                }
                return links.resolvedType;
            }
            function getActualTypeVariable(type) {
                if (type.flags & 33554432 /* Substitution */) {
                    return type.baseType;
                }
                if (type.flags & 8388608 /* IndexedAccess */ && (type.objectType.flags & 33554432 /* Substitution */ ||
                    type.indexType.flags & 33554432 /* Substitution */)) {
                    return getIndexedAccessType(getActualTypeVariable(type.objectType), getActualTypeVariable(type.indexType));
                }
                return type;
            }
            function isTypicalNondistributiveConditional(root) {
                return !root.isDistributive && isSingletonTupleType(root.node.checkType) && isSingletonTupleType(root.node.extendsType);
            }
            function isSingletonTupleType(node) {
                return ts.isTupleTypeNode(node) && ts.length(node.elements) === 1 && !ts.isOptionalTypeNode(node.elements[0]) && !ts.isRestTypeNode(node.elements[0]);
            }
            /**
             * We syntactually check for common nondistributive conditional shapes and unwrap them into
             * the intended comparison - we do this so we can check if the unwrapped types are generic or
             * not and appropriately defer condition calculation
             */
            function unwrapNondistributiveConditionalTuple(root, type) {
                return isTypicalNondistributiveConditional(root) && isTupleType(type) ? getTypeArguments(type)[0] : type;
            }
            function getConditionalType(root, mapper, aliasSymbol, aliasTypeArguments) {
                var result;
                var extraTypes;
                // We loop here for an immediately nested conditional type in the false position, effectively treating
                // types of the form 'A extends B ? X : C extends D ? Y : E extends F ? Z : ...' as a single construct for
                // purposes of resolution. This means such types aren't subject to the instantiation depth limiter.
                while (true) {
                    var isUnwrapped = isTypicalNondistributiveConditional(root);
                    var checkType = instantiateType(unwrapNondistributiveConditionalTuple(root, getActualTypeVariable(root.checkType)), mapper);
                    var checkTypeInstantiable = isGenericType(checkType);
                    var extendsType = instantiateType(unwrapNondistributiveConditionalTuple(root, root.extendsType), mapper);
                    if (checkType === wildcardType || extendsType === wildcardType) {
                        return wildcardType;
                    }
                    var combinedMapper = void 0;
                    if (root.inferTypeParameters) {
                        var context = createInferenceContext(root.inferTypeParameters, /*signature*/ undefined, 0 /* None */);
                        if (!checkTypeInstantiable) {
                            // We don't want inferences from constraints as they may cause us to eagerly resolve the
                            // conditional type instead of deferring resolution. Also, we always want strict function
                            // types rules (i.e. proper contravariance) for inferences.
                            inferTypes(context.inferences, checkType, extendsType, 512 /* NoConstraints */ | 1024 /* AlwaysStrict */);
                        }
                        // It's possible for 'infer T' type paramteters to be given uninstantiated constraints when the
                        // those type parameters are used in type references (see getInferredTypeParameterConstraint). For
                        // that reason we need context.mapper to be first in the combined mapper. See #42636 for examples.
                        combinedMapper = mapper ? combineTypeMappers(context.mapper, mapper) : context.mapper;
                    }
                    // Instantiate the extends type including inferences for 'infer T' type parameters
                    var inferredExtendsType = combinedMapper ? instantiateType(unwrapNondistributiveConditionalTuple(root, root.extendsType), combinedMapper) : extendsType;
                    // We attempt to resolve the conditional type only when the check and extends types are non-generic
                    if (!checkTypeInstantiable && !isGenericType(inferredExtendsType)) {
                        // Return falseType for a definitely false extends check. We check an instantiations of the two
                        // types with type parameters mapped to the wildcard type, the most permissive instantiations
                        // possible (the wildcard type is assignable to and from all types). If those are not related,
                        // then no instantiations will be and we can just return the false branch type.
                        if (!(inferredExtendsType.flags & 3 /* AnyOrUnknown */) && ((checkType.flags & 1 /* Any */ && !isUnwrapped) || !isTypeAssignableTo(getPermissiveInstantiation(checkType), getPermissiveInstantiation(inferredExtendsType)))) {
                            // Return union of trueType and falseType for 'any' since it matches anything
                            if (checkType.flags & 1 /* Any */ && !isUnwrapped) {
                                (extraTypes || (extraTypes = [])).push(instantiateType(getTypeFromTypeNode(root.node.trueType), combinedMapper || mapper));
                            }
                            // If falseType is an immediately nested conditional type that isn't distributive or has an
                            // identical checkType, switch to that type and loop.
                            var falseType_1 = getTypeFromTypeNode(root.node.falseType);
                            if (falseType_1.flags & 16777216 /* Conditional */) {
                                var newRoot = falseType_1.root;
                                if (newRoot.node.parent === root.node && (!newRoot.isDistributive || newRoot.checkType === root.checkType)) {
                                    root = newRoot;
                                    continue;
                                }
                            }
                            result = instantiateType(falseType_1, mapper);
                            break;
                        }
                        // Return trueType for a definitely true extends check. We check instantiations of the two
                        // types with type parameters mapped to their restrictive form, i.e. a form of the type parameter
                        // that has no constraint. This ensures that, for example, the type
                        //   type Foo<T extends { x: any }> = T extends { x: string } ? string : number
                        // doesn't immediately resolve to 'string' instead of being deferred.
                        if (inferredExtendsType.flags & 3 /* AnyOrUnknown */ || isTypeAssignableTo(getRestrictiveInstantiation(checkType), getRestrictiveInstantiation(inferredExtendsType))) {
                            result = instantiateType(getTypeFromTypeNode(root.node.trueType), combinedMapper || mapper);
                            break;
                        }
                    }
                    // Return a deferred type for a check that is neither definitely true nor definitely false
                    result = createType(16777216 /* Conditional */);
                    result.root = root;
                    result.checkType = instantiateType(root.checkType, mapper);
                    result.extendsType = instantiateType(root.extendsType, mapper);
                    result.mapper = mapper;
                    result.combinedMapper = combinedMapper;
                    result.aliasSymbol = aliasSymbol || root.aliasSymbol;
                    result.aliasTypeArguments = aliasSymbol ? aliasTypeArguments : instantiateTypes(root.aliasTypeArguments, mapper); // TODO: GH#18217
                    break;
                }
                return extraTypes ? getUnionType(ts.append(extraTypes, result)) : result;
            }
            function getTrueTypeFromConditionalType(type) {
                return type.resolvedTrueType || (type.resolvedTrueType = instantiateType(getTypeFromTypeNode(type.root.node.trueType), type.mapper));
            }
            function getFalseTypeFromConditionalType(type) {
                return type.resolvedFalseType || (type.resolvedFalseType = instantiateType(getTypeFromTypeNode(type.root.node.falseType), type.mapper));
            }
            function getInferredTrueTypeFromConditionalType(type) {
                return type.resolvedInferredTrueType || (type.resolvedInferredTrueType = type.combinedMapper ? instantiateType(getTypeFromTypeNode(type.root.node.trueType), type.combinedMapper) : getTrueTypeFromConditionalType(type));
            }
            function getInferTypeParameters(node) {
                var result;
                if (node.locals) {
                    node.locals.forEach(function (symbol) {
                        if (symbol.flags & 262144 /* TypeParameter */) {
                            result = ts.append(result, getDeclaredTypeOfSymbol(symbol));
                        }
                    });
                }
                return result;
            }
            function getTypeFromConditionalTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    var checkType = getTypeFromTypeNode(node.checkType);
                    var aliasSymbol = getAliasSymbolForTypeNode(node);
                    var aliasTypeArguments = getTypeArgumentsForAliasSymbol(aliasSymbol);
                    var allOuterTypeParameters = getOuterTypeParameters(node, /*includeThisTypes*/ true);
                    var outerTypeParameters = aliasTypeArguments ? allOuterTypeParameters : ts.filter(allOuterTypeParameters, function (tp) { return isTypeParameterPossiblyReferenced(tp, node); });
                    var root = {
                        node: node,
                        checkType: checkType,
                        extendsType: getTypeFromTypeNode(node.extendsType),
                        isDistributive: !!(checkType.flags & 262144 /* TypeParameter */),
                        inferTypeParameters: getInferTypeParameters(node),
                        outerTypeParameters: outerTypeParameters,
                        instantiations: undefined,
                        aliasSymbol: aliasSymbol,
                        aliasTypeArguments: aliasTypeArguments
                    };
                    links.resolvedType = getConditionalType(root, /*mapper*/ undefined);
                    if (outerTypeParameters) {
                        root.instantiations = new ts.Map();
                        root.instantiations.set(getTypeListId(outerTypeParameters), links.resolvedType);
                    }
                }
                return links.resolvedType;
            }
            function getTypeFromInferTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    links.resolvedType = getDeclaredTypeOfTypeParameter(getSymbolOfNode(node.typeParameter));
                }
                return links.resolvedType;
            }
            function getIdentifierChain(node) {
                if (ts.isIdentifier(node)) {
                    return [node];
                }
                else {
                    return ts.append(getIdentifierChain(node.left), node.right);
                }
            }
            function getTypeFromImportTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    if (node.isTypeOf && node.typeArguments) { // Only the non-typeof form can make use of type arguments
                        error(node, ts.Diagnostics.Type_arguments_cannot_be_used_here);
                        links.resolvedSymbol = unknownSymbol;
                        return links.resolvedType = errorType;
                    }
                    if (!ts.isLiteralImportTypeNode(node)) {
                        error(node.argument, ts.Diagnostics.String_literal_expected);
                        links.resolvedSymbol = unknownSymbol;
                        return links.resolvedType = errorType;
                    }
                    var targetMeaning = node.isTypeOf ? 111551 /* Value */ : node.flags & 4194304 /* JSDoc */ ? 111551 /* Value */ | 788968 /* Type */ : 788968 /* Type */;
                    // TODO: Future work: support unions/generics/whatever via a deferred import-type
                    var innerModuleSymbol = resolveExternalModuleName(node, node.argument.literal);
                    if (!innerModuleSymbol) {
                        links.resolvedSymbol = unknownSymbol;
                        return links.resolvedType = errorType;
                    }
                    var moduleSymbol = resolveExternalModuleSymbol(innerModuleSymbol, /*dontResolveAlias*/ false);
                    if (!ts.nodeIsMissing(node.qualifier)) {
                        var nameStack = getIdentifierChain(node.qualifier);
                        var currentNamespace = moduleSymbol;
                        var current = void 0;
                        while (current = nameStack.shift()) {
                            var meaning = nameStack.length ? 1920 /* Namespace */ : targetMeaning;
                            // typeof a.b.c is normally resolved using `checkExpression` which in turn defers to `checkQualifiedName`
                            // That, in turn, ultimately uses `getPropertyOfType` on the type of the symbol, which differs slightly from
                            // the `exports` lookup process that only looks up namespace members which is used for most type references
                            var mergedResolvedSymbol = getMergedSymbol(resolveSymbol(currentNamespace));
                            var next = node.isTypeOf
                                ? getPropertyOfType(getTypeOfSymbol(mergedResolvedSymbol), current.escapedText)
                                : getSymbol(getExportsOfSymbol(mergedResolvedSymbol), current.escapedText, meaning);
                            if (!next) {
                                error(current, ts.Diagnostics.Namespace_0_has_no_exported_member_1, getFullyQualifiedName(currentNamespace), ts.declarationNameToString(current));
                                return links.resolvedType = errorType;
                            }
                            getNodeLinks(current).resolvedSymbol = next;
                            getNodeLinks(current.parent).resolvedSymbol = next;
                            currentNamespace = next;
                        }
                        links.resolvedType = resolveImportSymbolType(node, links, currentNamespace, targetMeaning);
                    }
                    else {
                        if (moduleSymbol.flags & targetMeaning) {
                            links.resolvedType = resolveImportSymbolType(node, links, moduleSymbol, targetMeaning);
                        }
                        else {
                            var errorMessage = targetMeaning === 111551 /* Value */
                                ? ts.Diagnostics.Module_0_does_not_refer_to_a_value_but_is_used_as_a_value_here
                                : ts.Diagnostics.Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0;
                            error(node, errorMessage, node.argument.literal.text);
                            links.resolvedSymbol = unknownSymbol;
                            links.resolvedType = errorType;
                        }
                    }
                }
                return links.resolvedType;
            }
            function resolveImportSymbolType(node, links, symbol, meaning) {
                var resolvedSymbol = resolveSymbol(symbol);
                links.resolvedSymbol = resolvedSymbol;
                if (meaning === 111551 /* Value */) {
                    return getTypeOfSymbol(symbol); // intentionally doesn't use resolved symbol so type is cached as expected on the alias
                }
                else {
                    return getTypeReferenceType(node, resolvedSymbol); // getTypeReferenceType doesn't handle aliases - it must get the resolved symbol
                }
            }
            function getTypeFromTypeLiteralOrFunctionOrConstructorTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    // Deferred resolution of members is handled by resolveObjectTypeMembers
                    var aliasSymbol = getAliasSymbolForTypeNode(node);
                    if (getMembersOfSymbol(node.symbol).size === 0 && !aliasSymbol) {
                        links.resolvedType = emptyTypeLiteralType;
                    }
                    else {
                        var type = createObjectType(16 /* Anonymous */, node.symbol);
                        type.aliasSymbol = aliasSymbol;
                        type.aliasTypeArguments = getTypeArgumentsForAliasSymbol(aliasSymbol);
                        if (ts.isJSDocTypeLiteral(node) && node.isArrayType) {
                            type = createArrayType(type);
                        }
                        links.resolvedType = type;
                    }
                }
                return links.resolvedType;
            }
            function getAliasSymbolForTypeNode(node) {
                var host = node.parent;
                while (ts.isParenthesizedTypeNode(host) || ts.isJSDocTypeExpression(host) || ts.isTypeOperatorNode(host) && host.operator === 143 /* ReadonlyKeyword */) {
                    host = host.parent;
                }
                return ts.isTypeAlias(host) ? getSymbolOfNode(host) : undefined;
            }
            function getTypeArgumentsForAliasSymbol(symbol) {
                return symbol ? getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(symbol) : undefined;
            }
            function isNonGenericObjectType(type) {
                return !!(type.flags & 524288 /* Object */) && !isGenericMappedType(type);
            }
            function isEmptyObjectTypeOrSpreadsIntoEmptyObject(type) {
                return isEmptyObjectType(type) || !!(type.flags & (65536 /* Null */ | 32768 /* Undefined */ | 528 /* BooleanLike */ | 296 /* NumberLike */ | 2112 /* BigIntLike */ | 402653316 /* StringLike */ | 1056 /* EnumLike */ | 67108864 /* NonPrimitive */ | 4194304 /* Index */));
            }
            function tryMergeUnionOfObjectTypeAndEmptyObject(type, readonly) {
                if (!(type.flags & 1048576 /* Union */)) {
                    return type;
                }
                if (ts.every(type.types, isEmptyObjectTypeOrSpreadsIntoEmptyObject)) {
                    return ts.find(type.types, isEmptyObjectType) || emptyObjectType;
                }
                var firstType = ts.find(type.types, function (t) { return !isEmptyObjectTypeOrSpreadsIntoEmptyObject(t); });
                if (!firstType) {
                    return type;
                }
                var secondType = ts.find(type.types, function (t) { return t !== firstType && !isEmptyObjectTypeOrSpreadsIntoEmptyObject(t); });
                if (secondType) {
                    return type;
                }
                return getAnonymousPartialType(firstType);
                function getAnonymousPartialType(type) {
                    // gets the type as if it had been spread, but where everything in the spread is made optional
                    var members = ts.createSymbolTable();
                    for (var _i = 0, _a = getPropertiesOfType(type); _i < _a.length; _i++) {
                        var prop = _a[_i];
                        if (ts.getDeclarationModifierFlagsFromSymbol(prop) & (8 /* Private */ | 16 /* Protected */)) {
                            // do nothing, skip privates
                        }
                        else if (isSpreadableProperty(prop)) {
                            var isSetonlyAccessor = prop.flags & 65536 /* SetAccessor */ && !(prop.flags & 32768 /* GetAccessor */);
                            var flags = 4 /* Property */ | 16777216 /* Optional */;
                            var result = createSymbol(flags, prop.escapedName, getIsLateCheckFlag(prop) | (readonly ? 8 /* Readonly */ : 0));
                            result.type = isSetonlyAccessor ? undefinedType : addOptionality(getTypeOfSymbol(prop), /*isProperty*/ true);
                            result.declarations = prop.declarations;
                            result.nameType = getSymbolLinks(prop).nameType;
                            result.syntheticOrigin = prop;
                            members.set(prop.escapedName, result);
                        }
                    }
                    var spread = createAnonymousType(type.symbol, members, ts.emptyArray, ts.emptyArray, getIndexInfosOfType(type));
                    spread.objectFlags |= 128 /* ObjectLiteral */ | 262144 /* ContainsObjectOrArrayLiteral */;
                    return spread;
                }
            }
            /**
             * Since the source of spread types are object literals, which are not binary,
             * this function should be called in a left folding style, with left = previous result of getSpreadType
             * and right = the new element to be spread.
             */
            function getSpreadType(left, right, symbol, objectFlags, readonly) {
                if (left.flags & 1 /* Any */ || right.flags & 1 /* Any */) {
                    return anyType;
                }
                if (left.flags & 2 /* Unknown */ || right.flags & 2 /* Unknown */) {
                    return unknownType;
                }
                if (left.flags & 131072 /* Never */) {
                    return right;
                }
                if (right.flags & 131072 /* Never */) {
                    return left;
                }
                left = tryMergeUnionOfObjectTypeAndEmptyObject(left, readonly);
                if (left.flags & 1048576 /* Union */) {
                    return checkCrossProductUnion([left, right])
                        ? mapType(left, function (t) { return getSpreadType(t, right, symbol, objectFlags, readonly); })
                        : errorType;
                }
                right = tryMergeUnionOfObjectTypeAndEmptyObject(right, readonly);
                if (right.flags & 1048576 /* Union */) {
                    return checkCrossProductUnion([left, right])
                        ? mapType(right, function (t) { return getSpreadType(left, t, symbol, objectFlags, readonly); })
                        : errorType;
                }
                if (right.flags & (528 /* BooleanLike */ | 296 /* NumberLike */ | 2112 /* BigIntLike */ | 402653316 /* StringLike */ | 1056 /* EnumLike */ | 67108864 /* NonPrimitive */ | 4194304 /* Index */)) {
                    return left;
                }
                if (isGenericObjectType(left) || isGenericObjectType(right)) {
                    if (isEmptyObjectType(left)) {
                        return right;
                    }
                    // When the left type is an intersection, we may need to merge the last constituent of the
                    // intersection with the right type. For example when the left type is 'T & { a: string }'
                    // and the right type is '{ b: string }' we produce 'T & { a: string, b: string }'.
                    if (left.flags & 2097152 /* Intersection */) {
                        var types = left.types;
                        var lastLeft = types[types.length - 1];
                        if (isNonGenericObjectType(lastLeft) && isNonGenericObjectType(right)) {
                            return getIntersectionType(ts.concatenate(types.slice(0, types.length - 1), [getSpreadType(lastLeft, right, symbol, objectFlags, readonly)]));
                        }
                    }
                    return getIntersectionType([left, right]);
                }
                var members = ts.createSymbolTable();
                var skippedPrivateMembers = new ts.Set();
                var indexInfos = left === emptyObjectType ? getIndexInfosOfType(right) : getUnionIndexInfos([left, right]);
                for (var _i = 0, _a = getPropertiesOfType(right); _i < _a.length; _i++) {
                    var rightProp = _a[_i];
                    if (ts.getDeclarationModifierFlagsFromSymbol(rightProp) & (8 /* Private */ | 16 /* Protected */)) {
                        skippedPrivateMembers.add(rightProp.escapedName);
                    }
                    else if (isSpreadableProperty(rightProp)) {
                        members.set(rightProp.escapedName, getSpreadSymbol(rightProp, readonly));
                    }
                }
                for (var _b = 0, _c = getPropertiesOfType(left); _b < _c.length; _b++) {
                    var leftProp = _c[_b];
                    if (skippedPrivateMembers.has(leftProp.escapedName) || !isSpreadableProperty(leftProp)) {
                        continue;
                    }
                    if (members.has(leftProp.escapedName)) {
                        var rightProp = members.get(leftProp.escapedName);
                        var rightType = getTypeOfSymbol(rightProp);
                        if (rightProp.flags & 16777216 /* Optional */) {
                            var declarations = ts.concatenate(leftProp.declarations, rightProp.declarations);
                            var flags = 4 /* Property */ | (leftProp.flags & 16777216 /* Optional */);
                            var result = createSymbol(flags, leftProp.escapedName);
                            result.type = getUnionType([getTypeOfSymbol(leftProp), removeMissingOrUndefinedType(rightType)]);
                            result.leftSpread = leftProp;
                            result.rightSpread = rightProp;
                            result.declarations = declarations;
                            result.nameType = getSymbolLinks(leftProp).nameType;
                            members.set(leftProp.escapedName, result);
                        }
                    }
                    else {
                        members.set(leftProp.escapedName, getSpreadSymbol(leftProp, readonly));
                    }
                }
                var spread = createAnonymousType(symbol, members, ts.emptyArray, ts.emptyArray, ts.sameMap(indexInfos, function (info) { return getIndexInfoWithReadonly(info, readonly); }));
                spread.objectFlags |= 128 /* ObjectLiteral */ | 262144 /* ContainsObjectOrArrayLiteral */ | 4194304 /* ContainsSpread */ | objectFlags;
                return spread;
            }
            /** We approximate own properties as non-methods plus methods that are inside the object literal */
            function isSpreadableProperty(prop) {
                var _a;
                return !ts.some(prop.declarations, ts.isPrivateIdentifierClassElementDeclaration) &&
                    (!(prop.flags & (8192 /* Method */ | 32768 /* GetAccessor */ | 65536 /* SetAccessor */)) ||
                        !((_a = prop.declarations) === null || _a === void 0 ? void 0 : _a.some(function (decl) { return ts.isClassLike(decl.parent); })));
            }
            function getSpreadSymbol(prop, readonly) {
                var isSetonlyAccessor = prop.flags & 65536 /* SetAccessor */ && !(prop.flags & 32768 /* GetAccessor */);
                if (!isSetonlyAccessor && readonly === isReadonlySymbol(prop)) {
                    return prop;
                }
                var flags = 4 /* Property */ | (prop.flags & 16777216 /* Optional */);
                var result = createSymbol(flags, prop.escapedName, getIsLateCheckFlag(prop) | (readonly ? 8 /* Readonly */ : 0));
                result.type = isSetonlyAccessor ? undefinedType : getTypeOfSymbol(prop);
                result.declarations = prop.declarations;
                result.nameType = getSymbolLinks(prop).nameType;
                result.syntheticOrigin = prop;
                return result;
            }
            function getIndexInfoWithReadonly(info, readonly) {
                return info.isReadonly !== readonly ? createIndexInfo(info.keyType, info.type, readonly, info.declaration) : info;
            }
            function createLiteralType(flags, value, symbol, regularType) {
                var type = createType(flags);
                type.symbol = symbol;
                type.value = value;
                type.regularType = regularType || type;
                return type;
            }
            function getFreshTypeOfLiteralType(type) {
                if (type.flags & 2944 /* Literal */) {
                    if (!type.freshType) {
                        var freshType = createLiteralType(type.flags, type.value, type.symbol, type);
                        freshType.freshType = freshType;
                        type.freshType = freshType;
                    }
                    return type.freshType;
                }
                return type;
            }
            function getRegularTypeOfLiteralType(type) {
                return type.flags & 2944 /* Literal */ ? type.regularType :
                    type.flags & 1048576 /* Union */ ? (type.regularType || (type.regularType = mapType(type, getRegularTypeOfLiteralType))) :
                        type;
            }
            function isFreshLiteralType(type) {
                return !!(type.flags & 2944 /* Literal */) && type.freshType === type;
            }
            function getStringLiteralType(value) {
                var type;
                return stringLiteralTypes.get(value) ||
                    (stringLiteralTypes.set(value, type = createLiteralType(128 /* StringLiteral */, value)), type);
            }
            function getNumberLiteralType(value) {
                var type;
                return numberLiteralTypes.get(value) ||
                    (numberLiteralTypes.set(value, type = createLiteralType(256 /* NumberLiteral */, value)), type);
            }
            function getBigIntLiteralType(value) {
                var type;
                var key = ts.pseudoBigIntToString(value);
                return bigIntLiteralTypes.get(key) ||
                    (bigIntLiteralTypes.set(key, type = createLiteralType(2048 /* BigIntLiteral */, value)), type);
            }
            function getEnumLiteralType(value, enumId, symbol) {
                var type;
                var qualifier = typeof value === "string" ? "@" : "#";
                var key = enumId + qualifier + value;
                var flags = 1024 /* EnumLiteral */ | (typeof value === "string" ? 128 /* StringLiteral */ : 256 /* NumberLiteral */);
                return enumLiteralTypes.get(key) ||
                    (enumLiteralTypes.set(key, type = createLiteralType(flags, value, symbol)), type);
            }
            function getTypeFromLiteralTypeNode(node) {
                if (node.literal.kind === 104 /* NullKeyword */) {
                    return nullType;
                }
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    links.resolvedType = getRegularTypeOfLiteralType(checkExpression(node.literal));
                }
                return links.resolvedType;
            }
            function createUniqueESSymbolType(symbol) {
                var type = createType(8192 /* UniqueESSymbol */);
                type.symbol = symbol;
                type.escapedName = "__@" + type.symbol.escapedName + "@" + getSymbolId(type.symbol);
                return type;
            }
            function getESSymbolLikeTypeForNode(node) {
                if (ts.isValidESSymbolDeclaration(node)) {
                    var symbol = getSymbolOfNode(node);
                    var links = getSymbolLinks(symbol);
                    return links.uniqueESSymbolType || (links.uniqueESSymbolType = createUniqueESSymbolType(symbol));
                }
                return esSymbolType;
            }
            function getThisType(node) {
                var container = ts.getThisContainer(node, /*includeArrowFunctions*/ false);
                var parent = container && container.parent;
                if (parent && (ts.isClassLike(parent) || parent.kind === 256 /* InterfaceDeclaration */)) {
                    if (!ts.isStatic(container) &&
                        (!ts.isConstructorDeclaration(container) || ts.isNodeDescendantOf(node, container.body))) {
                        return getDeclaredTypeOfClassOrInterface(getSymbolOfNode(parent)).thisType;
                    }
                }
                // inside x.prototype = { ... }
                if (parent && ts.isObjectLiteralExpression(parent) && ts.isBinaryExpression(parent.parent) && ts.getAssignmentDeclarationKind(parent.parent) === 6 /* Prototype */) {
                    return getDeclaredTypeOfClassOrInterface(getSymbolOfNode(parent.parent.left).parent).thisType;
                }
                // /** @return {this} */
                // x.prototype.m = function() { ... }
                var host = node.flags & 4194304 /* JSDoc */ ? ts.getHostSignatureFromJSDoc(node) : undefined;
                if (host && ts.isFunctionExpression(host) && ts.isBinaryExpression(host.parent) && ts.getAssignmentDeclarationKind(host.parent) === 3 /* PrototypeProperty */) {
                    return getDeclaredTypeOfClassOrInterface(getSymbolOfNode(host.parent.left).parent).thisType;
                }
                // inside constructor function C() { ... }
                if (isJSConstructor(container) && ts.isNodeDescendantOf(node, container.body)) {
                    return getDeclaredTypeOfClassOrInterface(getSymbolOfNode(container)).thisType;
                }
                error(node, ts.Diagnostics.A_this_type_is_available_only_in_a_non_static_member_of_a_class_or_interface);
                return errorType;
            }
            function getTypeFromThisTypeNode(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    links.resolvedType = getThisType(node);
                }
                return links.resolvedType;
            }
            function getTypeFromRestTypeNode(node) {
                return getTypeFromTypeNode(getArrayElementTypeNode(node.type) || node.type);
            }
            function getArrayElementTypeNode(node) {
                switch (node.kind) {
                    case 189 /* ParenthesizedType */:
                        return getArrayElementTypeNode(node.type);
                    case 182 /* TupleType */:
                        if (node.elements.length === 1) {
                            node = node.elements[0];
                            if (node.kind === 184 /* RestType */ || node.kind === 195 /* NamedTupleMember */ && node.dotDotDotToken) {
                                return getArrayElementTypeNode(node.type);
                            }
                        }
                        break;
                    case 181 /* ArrayType */:
                        return node.elementType;
                }
                return undefined;
            }
            function getTypeFromNamedTupleTypeNode(node) {
                var links = getNodeLinks(node);
                return links.resolvedType || (links.resolvedType =
                    node.dotDotDotToken ? getTypeFromRestTypeNode(node) :
                        addOptionality(getTypeFromTypeNode(node.type), /*isProperty*/ true, !!node.questionToken));
            }
            function getTypeFromTypeNode(node) {
                return getConditionalFlowTypeOfType(getTypeFromTypeNodeWorker(node), node);
            }
            function getTypeFromTypeNodeWorker(node) {
                switch (node.kind) {
                    case 129 /* AnyKeyword */:
                    case 307 /* JSDocAllType */:
                    case 308 /* JSDocUnknownType */:
                        return anyType;
                    case 153 /* UnknownKeyword */:
                        return unknownType;
                    case 148 /* StringKeyword */:
                        return stringType;
                    case 145 /* NumberKeyword */:
                        return numberType;
                    case 156 /* BigIntKeyword */:
                        return bigintType;
                    case 132 /* BooleanKeyword */:
                        return booleanType;
                    case 149 /* SymbolKeyword */:
                        return esSymbolType;
                    case 114 /* VoidKeyword */:
                        return voidType;
                    case 151 /* UndefinedKeyword */:
                        return undefinedType;
                    case 104 /* NullKeyword */:
                        // TODO(rbuckton): `NullKeyword` is no longer a `TypeNode`, but we defensively allow it here because of incorrect casts in the Language Service.
                        return nullType;
                    case 142 /* NeverKeyword */:
                        return neverType;
                    case 146 /* ObjectKeyword */:
                        return node.flags & 131072 /* JavaScriptFile */ && !noImplicitAny ? anyType : nonPrimitiveType;
                    case 137 /* IntrinsicKeyword */:
                        return intrinsicMarkerType;
                    case 190 /* ThisType */:
                    case 108 /* ThisKeyword */:
                        // TODO(rbuckton): `ThisKeyword` is no longer a `TypeNode`, but we defensively allow it here because of incorrect casts in the Language Service and because of `isPartOfTypeNode`.
                        return getTypeFromThisTypeNode(node);
                    case 194 /* LiteralType */:
                        return getTypeFromLiteralTypeNode(node);
                    case 176 /* TypeReference */:
                        return getTypeFromTypeReference(node);
                    case 175 /* TypePredicate */:
                        return node.assertsModifier ? voidType : booleanType;
                    case 226 /* ExpressionWithTypeArguments */:
                        return getTypeFromTypeReference(node);
                    case 179 /* TypeQuery */:
                        return getTypeFromTypeQueryNode(node);
                    case 181 /* ArrayType */:
                    case 182 /* TupleType */:
                        return getTypeFromArrayOrTupleTypeNode(node);
                    case 183 /* OptionalType */:
                        return getTypeFromOptionalTypeNode(node);
                    case 185 /* UnionType */:
                        return getTypeFromUnionTypeNode(node);
                    case 186 /* IntersectionType */:
                        return getTypeFromIntersectionTypeNode(node);
                    case 309 /* JSDocNullableType */:
                        return getTypeFromJSDocNullableTypeNode(node);
                    case 311 /* JSDocOptionalType */:
                        return addOptionality(getTypeFromTypeNode(node.type));
                    case 195 /* NamedTupleMember */:
                        return getTypeFromNamedTupleTypeNode(node);
                    case 189 /* ParenthesizedType */:
                    case 310 /* JSDocNonNullableType */:
                    case 304 /* JSDocTypeExpression */:
                        return getTypeFromTypeNode(node.type);
                    case 184 /* RestType */:
                        return getTypeFromRestTypeNode(node);
                    case 313 /* JSDocVariadicType */:
                        return getTypeFromJSDocVariadicType(node);
                    case 177 /* FunctionType */:
                    case 178 /* ConstructorType */:
                    case 180 /* TypeLiteral */:
                    case 317 /* JSDocTypeLiteral */:
                    case 312 /* JSDocFunctionType */:
                    case 318 /* JSDocSignature */:
                        return getTypeFromTypeLiteralOrFunctionOrConstructorTypeNode(node);
                    case 191 /* TypeOperator */:
                        return getTypeFromTypeOperatorNode(node);
                    case 192 /* IndexedAccessType */:
                        return getTypeFromIndexedAccessTypeNode(node);
                    case 193 /* MappedType */:
                        return getTypeFromMappedTypeNode(node);
                    case 187 /* ConditionalType */:
                        return getTypeFromConditionalTypeNode(node);
                    case 188 /* InferType */:
                        return getTypeFromInferTypeNode(node);
                    case 196 /* TemplateLiteralType */:
                        return getTypeFromTemplateTypeNode(node);
                    case 198 /* ImportType */:
                        return getTypeFromImportTypeNode(node);
                    // This function assumes that an identifier, qualified name, or property access expression is a type expression
                    // Callers should first ensure this by calling `isPartOfTypeNode`
                    // TODO(rbuckton): These aren't valid TypeNodes, but we treat them as such because of `isPartOfTypeNode`, which returns `true` for things that aren't `TypeNode`s.
                    case 79 /* Identifier */:
                    case 159 /* QualifiedName */:
                    case 204 /* PropertyAccessExpression */:
                        var symbol = getSymbolAtLocation(node);
                        return symbol ? getDeclaredTypeOfSymbol(symbol) : errorType;
                    default:
                        return errorType;
                }
            }
            function instantiateList(items, mapper, instantiator) {
                if (items && items.length) {
                    for (var i = 0; i < items.length; i++) {
                        var item = items[i];
                        var mapped = instantiator(item, mapper);
                        if (item !== mapped) {
                            var result = i === 0 ? [] : items.slice(0, i);
                            result.push(mapped);
                            for (i++; i < items.length; i++) {
                                result.push(instantiator(items[i], mapper));
                            }
                            return result;
                        }
                    }
                }
                return items;
            }
            function instantiateTypes(types, mapper) {
                return instantiateList(types, mapper, instantiateType);
            }
            function instantiateSignatures(signatures, mapper) {
                return instantiateList(signatures, mapper, instantiateSignature);
            }
            function instantiateIndexInfos(indexInfos, mapper) {
                return instantiateList(indexInfos, mapper, instantiateIndexInfo);
            }
            function createTypeMapper(sources, targets) {
                return sources.length === 1 ? makeUnaryTypeMapper(sources[0], targets ? targets[0] : anyType) : makeArrayTypeMapper(sources, targets);
            }
            function getMappedType(type, mapper) {
                switch (mapper.kind) {
                    case 0 /* Simple */:
                        return type === mapper.source ? mapper.target : type;
                    case 1 /* Array */:
                        var sources = mapper.sources;
                        var targets = mapper.targets;
                        for (var i = 0; i < sources.length; i++) {
                            if (type === sources[i]) {
                                return targets ? targets[i] : anyType;
                            }
                        }
                        return type;
                    case 2 /* Function */:
                        return mapper.func(type);
                    case 3 /* Composite */:
                    case 4 /* Merged */:
                        var t1 = getMappedType(type, mapper.mapper1);
                        return t1 !== type && mapper.kind === 3 /* Composite */ ? instantiateType(t1, mapper.mapper2) : getMappedType(t1, mapper.mapper2);
                }
            }
            function makeUnaryTypeMapper(source, target) {
                return { kind: 0 /* Simple */, source: source, target: target };
            }
            function makeArrayTypeMapper(sources, targets) {
                return { kind: 1 /* Array */, sources: sources, targets: targets };
            }
            function makeFunctionTypeMapper(func) {
                return { kind: 2 /* Function */, func: func };
            }
            function makeCompositeTypeMapper(kind, mapper1, mapper2) {
                return { kind: kind, mapper1: mapper1, mapper2: mapper2 };
            }
            function createTypeEraser(sources) {
                return createTypeMapper(sources, /*targets*/ undefined);
            }
            /**
             * Maps forward-references to later types parameters to the empty object type.
             * This is used during inference when instantiating type parameter defaults.
             */
            function createBackreferenceMapper(context, index) {
                return makeFunctionTypeMapper(function (t) { return ts.findIndex(context.inferences, function (info) { return info.typeParameter === t; }) >= index ? unknownType : t; });
            }
            function combineTypeMappers(mapper1, mapper2) {
                return mapper1 ? makeCompositeTypeMapper(3 /* Composite */, mapper1, mapper2) : mapper2;
            }
            function mergeTypeMappers(mapper1, mapper2) {
                return mapper1 ? makeCompositeTypeMapper(4 /* Merged */, mapper1, mapper2) : mapper2;
            }
            function prependTypeMapping(source, target, mapper) {
                return !mapper ? makeUnaryTypeMapper(source, target) : makeCompositeTypeMapper(4 /* Merged */, makeUnaryTypeMapper(source, target), mapper);
            }
            function appendTypeMapping(mapper, source, target) {
                return !mapper ? makeUnaryTypeMapper(source, target) : makeCompositeTypeMapper(4 /* Merged */, mapper, makeUnaryTypeMapper(source, target));
            }
            function getRestrictiveTypeParameter(tp) {
                return tp.constraint === unknownType ? tp : tp.restrictiveInstantiation || (tp.restrictiveInstantiation = createTypeParameter(tp.symbol),
                    tp.restrictiveInstantiation.constraint = unknownType,
                    tp.restrictiveInstantiation);
            }
            function cloneTypeParameter(typeParameter) {
                var result = createTypeParameter(typeParameter.symbol);
                result.target = typeParameter;
                return result;
            }
            function instantiateTypePredicate(predicate, mapper) {
                return createTypePredicate(predicate.kind, predicate.parameterName, predicate.parameterIndex, instantiateType(predicate.type, mapper));
            }
            function instantiateSignature(signature, mapper, eraseTypeParameters) {
                var freshTypeParameters;
                if (signature.typeParameters && !eraseTypeParameters) {
                    // First create a fresh set of type parameters, then include a mapping from the old to the
                    // new type parameters in the mapper function. Finally store this mapper in the new type
                    // parameters such that we can use it when instantiating constraints.
                    freshTypeParameters = ts.map(signature.typeParameters, cloneTypeParameter);
                    mapper = combineTypeMappers(createTypeMapper(signature.typeParameters, freshTypeParameters), mapper);
                    for (var _i = 0, freshTypeParameters_1 = freshTypeParameters; _i < freshTypeParameters_1.length; _i++) {
                        var tp = freshTypeParameters_1[_i];
                        tp.mapper = mapper;
                    }
                }
                // Don't compute resolvedReturnType and resolvedTypePredicate now,
                // because using `mapper` now could trigger inferences to become fixed. (See `createInferenceContext`.)
                // See GH#17600.
                var result = createSignature(signature.declaration, freshTypeParameters, signature.thisParameter && instantiateSymbol(signature.thisParameter, mapper), instantiateList(signature.parameters, mapper, instantiateSymbol), 
                /*resolvedReturnType*/ undefined, 
                /*resolvedTypePredicate*/ undefined, signature.minArgumentCount, signature.flags & 39 /* PropagatingFlags */);
                result.target = signature;
                result.mapper = mapper;
                return result;
            }
            function instantiateSymbol(symbol, mapper) {
                var links = getSymbolLinks(symbol);
                if (links.type && !couldContainTypeVariables(links.type)) {
                    // If the type of the symbol is already resolved, and if that type could not possibly
                    // be affected by instantiation, simply return the symbol itself.
                    return symbol;
                }
                if (ts.getCheckFlags(symbol) & 1 /* Instantiated */) {
                    // If symbol being instantiated is itself a instantiation, fetch the original target and combine the
                    // type mappers. This ensures that original type identities are properly preserved and that aliases
                    // always reference a non-aliases.
                    symbol = links.target;
                    mapper = combineTypeMappers(links.mapper, mapper);
                }
                // Keep the flags from the symbol we're instantiating.  Mark that is instantiated, and
                // also transient so that we can just store data on it directly.
                var result = createSymbol(symbol.flags, symbol.escapedName, 1 /* Instantiated */ | ts.getCheckFlags(symbol) & (8 /* Readonly */ | 4096 /* Late */ | 16384 /* OptionalParameter */ | 32768 /* RestParameter */));
                result.declarations = symbol.declarations;
                result.parent = symbol.parent;
                result.target = symbol;
                result.mapper = mapper;
                if (symbol.valueDeclaration) {
                    result.valueDeclaration = symbol.valueDeclaration;
                }
                if (links.nameType) {
                    result.nameType = links.nameType;
                }
                return result;
            }
            function getObjectTypeInstantiation(type, mapper, aliasSymbol, aliasTypeArguments) {
                var declaration = type.objectFlags & 4 /* Reference */ ? type.node : type.symbol.declarations[0];
                var links = getNodeLinks(declaration);
                var target = type.objectFlags & 4 /* Reference */ ? links.resolvedType :
                    type.objectFlags & 64 /* Instantiated */ ? type.target : type;
                var typeParameters = links.outerTypeParameters;
                if (!typeParameters) {
                    // The first time an anonymous type is instantiated we compute and store a list of the type
                    // parameters that are in scope (and therefore potentially referenced). For type literals that
                    // aren't the right hand side of a generic type alias declaration we optimize by reducing the
                    // set of type parameters to those that are possibly referenced in the literal.
                    var outerTypeParameters = getOuterTypeParameters(declaration, /*includeThisTypes*/ true);
                    if (isJSConstructor(declaration)) {
                        var templateTagParameters = getTypeParametersFromDeclaration(declaration);
                        outerTypeParameters = ts.addRange(outerTypeParameters, templateTagParameters);
                    }
                    typeParameters = outerTypeParameters || ts.emptyArray;
                    var allDeclarations_1 = type.objectFlags & 4 /* Reference */ ? [declaration] : type.symbol.declarations;
                    typeParameters = (target.objectFlags & 4 /* Reference */ || target.symbol.flags & 8192 /* Method */ || target.symbol.flags & 2048 /* TypeLiteral */) && !target.aliasTypeArguments ?
                        ts.filter(typeParameters, function (tp) { return ts.some(allDeclarations_1, function (d) { return isTypeParameterPossiblyReferenced(tp, d); }); }) :
                        typeParameters;
                    links.outerTypeParameters = typeParameters;
                }
                if (typeParameters.length) {
                    // We are instantiating an anonymous type that has one or more type parameters in scope. Apply the
                    // mapper to the type parameters to produce the effective list of type arguments, and compute the
                    // instantiation cache key from the type IDs of the type arguments.
                    var combinedMapper_1 = combineTypeMappers(type.mapper, mapper);
                    var typeArguments = ts.map(typeParameters, function (t) { return getMappedType(t, combinedMapper_1); });
                    var newAliasSymbol = aliasSymbol || type.aliasSymbol;
                    var newAliasTypeArguments = aliasSymbol ? aliasTypeArguments : instantiateTypes(type.aliasTypeArguments, mapper);
                    var id = getTypeListId(typeArguments) + getAliasId(newAliasSymbol, newAliasTypeArguments);
                    if (!target.instantiations) {
                        target.instantiations = new ts.Map();
                        target.instantiations.set(getTypeListId(typeParameters) + getAliasId(target.aliasSymbol, target.aliasTypeArguments), target);
                    }
                    var result = target.instantiations.get(id);
                    if (!result) {
                        var newMapper = createTypeMapper(typeParameters, typeArguments);
                        result = target.objectFlags & 4 /* Reference */ ? createDeferredTypeReference(type.target, type.node, newMapper, newAliasSymbol, newAliasTypeArguments) :
                            target.objectFlags & 32 /* Mapped */ ? instantiateMappedType(target, newMapper, newAliasSymbol, newAliasTypeArguments) :
                                instantiateAnonymousType(target, newMapper, newAliasSymbol, newAliasTypeArguments);
                        target.instantiations.set(id, result);
                    }
                    return result;
                }
                return type;
            }
            function maybeTypeParameterReference(node) {
                return !(node.parent.kind === 176 /* TypeReference */ && node.parent.typeArguments && node === node.parent.typeName ||
                    node.parent.kind === 198 /* ImportType */ && node.parent.typeArguments && node === node.parent.qualifier);
            }
            function isTypeParameterPossiblyReferenced(tp, node) {
                // If the type parameter doesn't have exactly one declaration, if there are invening statement blocks
                // between the node and the type parameter declaration, if the node contains actual references to the
                // type parameter, or if the node contains type queries, we consider the type parameter possibly referenced.
                if (tp.symbol && tp.symbol.declarations && tp.symbol.declarations.length === 1) {
                    var container = tp.symbol.declarations[0].parent;
                    for (var n = node; n !== container; n = n.parent) {
                        if (!n || n.kind === 233 /* Block */ || n.kind === 187 /* ConditionalType */ && ts.forEachChild(n.extendsType, containsReference)) {
                            return true;
                        }
                    }
                    return containsReference(node);
                }
                return true;
                function containsReference(node) {
                    switch (node.kind) {
                        case 190 /* ThisType */:
                            return !!tp.isThisType;
                        case 79 /* Identifier */:
                            return !tp.isThisType && ts.isPartOfTypeNode(node) && maybeTypeParameterReference(node) &&
                                getTypeFromTypeNodeWorker(node) === tp; // use worker because we're looking for === equality
                        case 179 /* TypeQuery */:
                            return true;
                        case 167 /* MethodDeclaration */:
                        case 166 /* MethodSignature */:
                            return !node.type && !!node.body ||
                                ts.some(node.typeParameters, containsReference) ||
                                ts.some(node.parameters, containsReference) ||
                                !!node.type && containsReference(node.type);
                    }
                    return !!ts.forEachChild(node, containsReference);
                }
            }
            function getHomomorphicTypeVariable(type) {
                var constraintType = getConstraintTypeFromMappedType(type);
                if (constraintType.flags & 4194304 /* Index */) {
                    var typeVariable = getActualTypeVariable(constraintType.type);
                    if (typeVariable.flags & 262144 /* TypeParameter */) {
                        return typeVariable;
                    }
                }
                return undefined;
            }
            function instantiateMappedType(type, mapper, aliasSymbol, aliasTypeArguments) {
                // For a homomorphic mapped type { [P in keyof T]: X }, where T is some type variable, the mapping
                // operation depends on T as follows:
                // * If T is a primitive type no mapping is performed and the result is simply T.
                // * If T is a union type we distribute the mapped type over the union.
                // * If T is an array we map to an array where the element type has been transformed.
                // * If T is a tuple we map to a tuple where the element types have been transformed.
                // * Otherwise we map to an object type where the type of each property has been transformed.
                // For example, when T is instantiated to a union type A | B, we produce { [P in keyof A]: X } |
                // { [P in keyof B]: X }, and when when T is instantiated to a union type A | undefined, we produce
                // { [P in keyof A]: X } | undefined.
                var typeVariable = getHomomorphicTypeVariable(type);
                if (typeVariable) {
                    var mappedTypeVariable = instantiateType(typeVariable, mapper);
                    if (typeVariable !== mappedTypeVariable) {
                        return mapTypeWithAlias(getReducedType(mappedTypeVariable), function (t) {
                            if (t.flags & (3 /* AnyOrUnknown */ | 58982400 /* InstantiableNonPrimitive */ | 524288 /* Object */ | 2097152 /* Intersection */) && t !== wildcardType && t !== errorType) {
                                if (!type.declaration.nameType) {
                                    if (isArrayType(t)) {
                                        return instantiateMappedArrayType(t, type, prependTypeMapping(typeVariable, t, mapper));
                                    }
                                    if (isGenericTupleType(t)) {
                                        return instantiateMappedGenericTupleType(t, type, typeVariable, mapper);
                                    }
                                    if (isTupleType(t)) {
                                        return instantiateMappedTupleType(t, type, prependTypeMapping(typeVariable, t, mapper));
                                    }
                                }
                                return instantiateAnonymousType(type, prependTypeMapping(typeVariable, t, mapper));
                            }
                            return t;
                        }, aliasSymbol, aliasTypeArguments);
                    }
                }
                // If the constraint type of the instantiation is the wildcard type, return the wildcard type.
                return instantiateType(getConstraintTypeFromMappedType(type), mapper) === wildcardType ? wildcardType : instantiateAnonymousType(type, mapper, aliasSymbol, aliasTypeArguments);
            }
            function getModifiedReadonlyState(state, modifiers) {
                return modifiers & 1 /* IncludeReadonly */ ? true : modifiers & 2 /* ExcludeReadonly */ ? false : state;
            }
            function instantiateMappedGenericTupleType(tupleType, mappedType, typeVariable, mapper) {
                // When a tuple type is generic (i.e. when it contains variadic elements), we want to eagerly map the
                // non-generic elements and defer mapping the generic elements. In order to facilitate this, we transform
                // M<[A, B?, ...T, ...C[]] into [...M<[A]>, ...M<[B?]>, ...M<T>, ...M<C[]>] and then rely on tuple type
                // normalization to resolve the non-generic parts of the resulting tuple.
                var elementFlags = tupleType.target.elementFlags;
                var elementTypes = ts.map(getTypeArguments(tupleType), function (t, i) {
                    var singleton = elementFlags[i] & 8 /* Variadic */ ? t :
                        elementFlags[i] & 4 /* Rest */ ? createArrayType(t) :
                            createTupleType([t], [elementFlags[i]]);
                    // The singleton is never a generic tuple type, so it is safe to recurse here.
                    return instantiateMappedType(mappedType, prependTypeMapping(typeVariable, singleton, mapper));
                });
                var newReadonly = getModifiedReadonlyState(tupleType.target.readonly, getMappedTypeModifiers(mappedType));
                return createTupleType(elementTypes, ts.map(elementTypes, function (_) { return 8 /* Variadic */; }), newReadonly);
            }
            function instantiateMappedArrayType(arrayType, mappedType, mapper) {
                var elementType = instantiateMappedTypeTemplate(mappedType, numberType, /*isOptional*/ true, mapper);
                return elementType === errorType ? errorType :
                    createArrayType(elementType, getModifiedReadonlyState(isReadonlyArrayType(arrayType), getMappedTypeModifiers(mappedType)));
            }
            function instantiateMappedTupleType(tupleType, mappedType, mapper) {
                var elementFlags = tupleType.target.elementFlags;
                var elementTypes = ts.map(getTypeArguments(tupleType), function (_, i) {
                    return instantiateMappedTypeTemplate(mappedType, getStringLiteralType("" + i), !!(elementFlags[i] & 2 /* Optional */), mapper);
                });
                var modifiers = getMappedTypeModifiers(mappedType);
                var newTupleModifiers = modifiers & 4 /* IncludeOptional */ ? ts.map(elementFlags, function (f) { return f & 1 /* Required */ ? 2 /* Optional */ : f; }) :
                    modifiers & 8 /* ExcludeOptional */ ? ts.map(elementFlags, function (f) { return f & 2 /* Optional */ ? 1 /* Required */ : f; }) :
                        elementFlags;
                var newReadonly = getModifiedReadonlyState(tupleType.target.readonly, modifiers);
                return ts.contains(elementTypes, errorType) ? errorType :
                    createTupleType(elementTypes, newTupleModifiers, newReadonly, tupleType.target.labeledElementDeclarations);
            }
            function instantiateMappedTypeTemplate(type, key, isOptional, mapper) {
                var templateMapper = appendTypeMapping(mapper, getTypeParameterFromMappedType(type), key);
                var propType = instantiateType(getTemplateTypeFromMappedType(type.target || type), templateMapper);
                var modifiers = getMappedTypeModifiers(type);
                return strictNullChecks && modifiers & 4 /* IncludeOptional */ && !maybeTypeOfKind(propType, 32768 /* Undefined */ | 16384 /* Void */) ? getOptionalType(propType, /*isProperty*/ true) :
                    strictNullChecks && modifiers & 8 /* ExcludeOptional */ && isOptional ? getTypeWithFacts(propType, 524288 /* NEUndefined */) :
                        propType;
            }
            function instantiateAnonymousType(type, mapper, aliasSymbol, aliasTypeArguments) {
                var result = createObjectType(type.objectFlags | 64 /* Instantiated */, type.symbol);
                if (type.objectFlags & 32 /* Mapped */) {
                    result.declaration = type.declaration;
                    // C.f. instantiateSignature
                    var origTypeParameter = getTypeParameterFromMappedType(type);
                    var freshTypeParameter = cloneTypeParameter(origTypeParameter);
                    result.typeParameter = freshTypeParameter;
                    mapper = combineTypeMappers(makeUnaryTypeMapper(origTypeParameter, freshTypeParameter), mapper);
                    freshTypeParameter.mapper = mapper;
                }
                result.target = type;
                result.mapper = mapper;
                result.aliasSymbol = aliasSymbol || type.aliasSymbol;
                result.aliasTypeArguments = aliasSymbol ? aliasTypeArguments : instantiateTypes(type.aliasTypeArguments, mapper);
                return result;
            }
            function getConditionalTypeInstantiation(type, mapper, aliasSymbol, aliasTypeArguments) {
                var root = type.root;
                if (root.outerTypeParameters) {
                    // We are instantiating a conditional type that has one or more type parameters in scope. Apply the
                    // mapper to the type parameters to produce the effective list of type arguments, and compute the
                    // instantiation cache key from the type IDs of the type arguments.
                    var typeArguments = ts.map(root.outerTypeParameters, function (t) { return getMappedType(t, mapper); });
                    var id = getTypeListId(typeArguments) + getAliasId(aliasSymbol, aliasTypeArguments);
                    var result = root.instantiations.get(id);
                    if (!result) {
                        var newMapper = createTypeMapper(root.outerTypeParameters, typeArguments);
                        result = instantiateConditionalType(root, newMapper, aliasSymbol, aliasTypeArguments);
                        root.instantiations.set(id, result);
                    }
                    return result;
                }
                return type;
            }
            function instantiateConditionalType(root, mapper, aliasSymbol, aliasTypeArguments) {
                // Check if we have a conditional type where the check type is a naked type parameter. If so,
                // the conditional type is distributive over union types and when T is instantiated to a union
                // type A | B, we produce (A extends U ? X : Y) | (B extends U ? X : Y).
                if (root.isDistributive) {
                    var checkType_1 = root.checkType;
                    var instantiatedType = getMappedType(checkType_1, mapper);
                    if (checkType_1 !== instantiatedType && instantiatedType.flags & (1048576 /* Union */ | 131072 /* Never */)) {
                        return mapTypeWithAlias(instantiatedType, function (t) { return getConditionalType(root, prependTypeMapping(checkType_1, t, mapper)); }, aliasSymbol, aliasTypeArguments);
                    }
                }
                return getConditionalType(root, mapper, aliasSymbol, aliasTypeArguments);
            }
            function instantiateType(type, mapper) {
                return type && mapper ? instantiateTypeWithAlias(type, mapper, /*aliasSymbol*/ undefined, /*aliasTypeArguments*/ undefined) : type;
            }
            function instantiateTypeWithAlias(type, mapper, aliasSymbol, aliasTypeArguments) {
                if (!couldContainTypeVariables(type)) {
                    return type;
                }
                if (instantiationDepth === 50 || instantiationCount >= 5000000) {
                    // We have reached 50 recursive type instantiations and there is a very high likelyhood we're dealing
                    // with a combination of infinite generic types that perpetually generate new type identities. We stop
                    // the recursion here by yielding the error type.
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.instant("checkTypes" /* CheckTypes */, "instantiateType_DepthLimit", { typeId: type.id, instantiationDepth: instantiationDepth, instantiationCount: instantiationCount });
                    error(currentNode, ts.Diagnostics.Type_instantiation_is_excessively_deep_and_possibly_infinite);
                    return errorType;
                }
                totalInstantiationCount++;
                instantiationCount++;
                instantiationDepth++;
                var result = instantiateTypeWorker(type, mapper, aliasSymbol, aliasTypeArguments);
                instantiationDepth--;
                return result;
            }
            function instantiateTypeWorker(type, mapper, aliasSymbol, aliasTypeArguments) {
                var flags = type.flags;
                if (flags & 262144 /* TypeParameter */) {
                    return getMappedType(type, mapper);
                }
                if (flags & 524288 /* Object */) {
                    var objectFlags = type.objectFlags;
                    if (objectFlags & (4 /* Reference */ | 16 /* Anonymous */ | 32 /* Mapped */)) {
                        if (objectFlags & 4 /* Reference */ && !type.node) {
                            var resolvedTypeArguments = type.resolvedTypeArguments;
                            var newTypeArguments = instantiateTypes(resolvedTypeArguments, mapper);
                            return newTypeArguments !== resolvedTypeArguments ? createNormalizedTypeReference(type.target, newTypeArguments) : type;
                        }
                        if (objectFlags & 1024 /* ReverseMapped */) {
                            return instantiateReverseMappedType(type, mapper);
                        }
                        return getObjectTypeInstantiation(type, mapper, aliasSymbol, aliasTypeArguments);
                    }
                    return type;
                }
                if (flags & 3145728 /* UnionOrIntersection */) {
                    var origin = type.flags & 1048576 /* Union */ ? type.origin : undefined;
                    var types = origin && origin.flags & 3145728 /* UnionOrIntersection */ ? origin.types : type.types;
                    var newTypes = instantiateTypes(types, mapper);
                    if (newTypes === types && aliasSymbol === type.aliasSymbol) {
                        return type;
                    }
                    var newAliasSymbol = aliasSymbol || type.aliasSymbol;
                    var newAliasTypeArguments = aliasSymbol ? aliasTypeArguments : instantiateTypes(type.aliasTypeArguments, mapper);
                    return flags & 2097152 /* Intersection */ || origin && origin.flags & 2097152 /* Intersection */ ?
                        getIntersectionType(newTypes, newAliasSymbol, newAliasTypeArguments) :
                        getUnionType(newTypes, 1 /* Literal */, newAliasSymbol, newAliasTypeArguments);
                }
                if (flags & 4194304 /* Index */) {
                    return getIndexType(instantiateType(type.type, mapper));
                }
                if (flags & 134217728 /* TemplateLiteral */) {
                    return getTemplateLiteralType(type.texts, instantiateTypes(type.types, mapper));
                }
                if (flags & 268435456 /* StringMapping */) {
                    return getStringMappingType(type.symbol, instantiateType(type.type, mapper));
                }
                if (flags & 8388608 /* IndexedAccess */) {
                    var newAliasSymbol = aliasSymbol || type.aliasSymbol;
                    var newAliasTypeArguments = aliasSymbol ? aliasTypeArguments : instantiateTypes(type.aliasTypeArguments, mapper);
                    return getIndexedAccessType(instantiateType(type.objectType, mapper), instantiateType(type.indexType, mapper), type.accessFlags, /*accessNode*/ undefined, newAliasSymbol, newAliasTypeArguments);
                }
                if (flags & 16777216 /* Conditional */) {
                    return getConditionalTypeInstantiation(type, combineTypeMappers(type.mapper, mapper), aliasSymbol, aliasTypeArguments);
                }
                if (flags & 33554432 /* Substitution */) {
                    var maybeVariable = instantiateType(type.baseType, mapper);
                    if (maybeVariable.flags & 8650752 /* TypeVariable */) {
                        return getSubstitutionType(maybeVariable, instantiateType(type.substitute, mapper));
                    }
                    else {
                        var sub = instantiateType(type.substitute, mapper);
                        if (sub.flags & 3 /* AnyOrUnknown */ || isTypeAssignableTo(getRestrictiveInstantiation(maybeVariable), getRestrictiveInstantiation(sub))) {
                            return maybeVariable;
                        }
                        return sub;
                    }
                }
                return type;
            }
            function instantiateReverseMappedType(type, mapper) {
                var innerMappedType = instantiateType(type.mappedType, mapper);
                if (!(ts.getObjectFlags(innerMappedType) & 32 /* Mapped */)) {
                    return type;
                }
                var innerIndexType = instantiateType(type.constraintType, mapper);
                if (!(innerIndexType.flags & 4194304 /* Index */)) {
                    return type;
                }
                var instantiated = inferTypeForHomomorphicMappedType(instantiateType(type.source, mapper), innerMappedType, innerIndexType);
                if (instantiated) {
                    return instantiated;
                }
                return type; // Nested invocation of `inferTypeForHomomorphicMappedType` or the `source` instantiated into something unmappable
            }
            function getPermissiveInstantiation(type) {
                return type.flags & (131068 /* Primitive */ | 3 /* AnyOrUnknown */ | 131072 /* Never */) ? type :
                    type.permissiveInstantiation || (type.permissiveInstantiation = instantiateType(type, permissiveMapper));
            }
            function getRestrictiveInstantiation(type) {
                if (type.flags & (131068 /* Primitive */ | 3 /* AnyOrUnknown */ | 131072 /* Never */)) {
                    return type;
                }
                if (type.restrictiveInstantiation) {
                    return type.restrictiveInstantiation;
                }
                type.restrictiveInstantiation = instantiateType(type, restrictiveMapper);
                // We set the following so we don't attempt to set the restrictive instance of a restrictive instance
                // which is redundant - we'll produce new type identities, but all type params have already been mapped.
                // This also gives us a way to detect restrictive instances upon comparisons and _disable_ the "distributeive constraint"
                // assignability check for them, which is distinctly unsafe, as once you have a restrctive instance, all the type parameters
                // are constrained to `unknown` and produce tons of false positives/negatives!
                type.restrictiveInstantiation.restrictiveInstantiation = type.restrictiveInstantiation;
                return type.restrictiveInstantiation;
            }
            function instantiateIndexInfo(info, mapper) {
                return createIndexInfo(info.keyType, instantiateType(info.type, mapper), info.isReadonly, info.declaration);
            }
            // Returns true if the given expression contains (at any level of nesting) a function or arrow expression
            // that is subject to contextual typing.
            function isContextSensitive(node) {
                ts.Debug.assert(node.kind !== 167 /* MethodDeclaration */ || ts.isObjectLiteralMethod(node));
                switch (node.kind) {
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                    case 167 /* MethodDeclaration */:
                    case 254 /* FunctionDeclaration */: // Function declarations can have context when annotated with a jsdoc @type
                        return isContextSensitiveFunctionLikeDeclaration(node);
                    case 203 /* ObjectLiteralExpression */:
                        return ts.some(node.properties, isContextSensitive);
                    case 202 /* ArrayLiteralExpression */:
                        return ts.some(node.elements, isContextSensitive);
                    case 220 /* ConditionalExpression */:
                        return isContextSensitive(node.whenTrue) ||
                            isContextSensitive(node.whenFalse);
                    case 219 /* BinaryExpression */:
                        return (node.operatorToken.kind === 56 /* BarBarToken */ || node.operatorToken.kind === 60 /* QuestionQuestionToken */) &&
                            (isContextSensitive(node.left) || isContextSensitive(node.right));
                    case 291 /* PropertyAssignment */:
                        return isContextSensitive(node.initializer);
                    case 210 /* ParenthesizedExpression */:
                        return isContextSensitive(node.expression);
                    case 284 /* JsxAttributes */:
                        return ts.some(node.properties, isContextSensitive) || ts.isJsxOpeningElement(node.parent) && ts.some(node.parent.parent.children, isContextSensitive);
                    case 283 /* JsxAttribute */: {
                        // If there is no initializer, JSX attribute has a boolean value of true which is not context sensitive.
                        var initializer = node.initializer;
                        return !!initializer && isContextSensitive(initializer);
                    }
                    case 286 /* JsxExpression */: {
                        // It is possible to that node.expression is undefined (e.g <div x={} />)
                        var expression = node.expression;
                        return !!expression && isContextSensitive(expression);
                    }
                }
                return false;
            }
            function isContextSensitiveFunctionLikeDeclaration(node) {
                return (!ts.isFunctionDeclaration(node) || ts.isInJSFile(node) && !!getTypeForDeclarationFromJSDocComment(node)) &&
                    (ts.hasContextSensitiveParameters(node) || hasContextSensitiveReturnExpression(node));
            }
            function hasContextSensitiveReturnExpression(node) {
                // TODO(anhans): A block should be context-sensitive if it has a context-sensitive return value.
                return !node.typeParameters && !ts.getEffectiveReturnTypeNode(node) && !!node.body && node.body.kind !== 233 /* Block */ && isContextSensitive(node.body);
            }
            function isContextSensitiveFunctionOrObjectLiteralMethod(func) {
                return (ts.isInJSFile(func) && ts.isFunctionDeclaration(func) || isFunctionExpressionOrArrowFunction(func) || ts.isObjectLiteralMethod(func)) &&
                    isContextSensitiveFunctionLikeDeclaration(func);
            }
            function getTypeWithoutSignatures(type) {
                if (type.flags & 524288 /* Object */) {
                    var resolved = resolveStructuredTypeMembers(type);
                    if (resolved.constructSignatures.length || resolved.callSignatures.length) {
                        var result = createObjectType(16 /* Anonymous */, type.symbol);
                        result.members = resolved.members;
                        result.properties = resolved.properties;
                        result.callSignatures = ts.emptyArray;
                        result.constructSignatures = ts.emptyArray;
                        result.indexInfos = ts.emptyArray;
                        return result;
                    }
                }
                else if (type.flags & 2097152 /* Intersection */) {
                    return getIntersectionType(ts.map(type.types, getTypeWithoutSignatures));
                }
                return type;
            }
            // TYPE CHECKING
            function isTypeIdenticalTo(source, target) {
                return isTypeRelatedTo(source, target, identityRelation);
            }
            function compareTypesIdentical(source, target) {
                return isTypeRelatedTo(source, target, identityRelation) ? -1 /* True */ : 0 /* False */;
            }
            function compareTypesAssignable(source, target) {
                return isTypeRelatedTo(source, target, assignableRelation) ? -1 /* True */ : 0 /* False */;
            }
            function compareTypesSubtypeOf(source, target) {
                return isTypeRelatedTo(source, target, subtypeRelation) ? -1 /* True */ : 0 /* False */;
            }
            function isTypeSubtypeOf(source, target) {
                return isTypeRelatedTo(source, target, subtypeRelation);
            }
            function isTypeAssignableTo(source, target) {
                return isTypeRelatedTo(source, target, assignableRelation);
            }
            // An object type S is considered to be derived from an object type T if
            // S is a union type and every constituent of S is derived from T,
            // T is a union type and S is derived from at least one constituent of T, or
            // S is a type variable with a base constraint that is derived from T,
            // T is one of the global types Object and Function and S is a subtype of T, or
            // T occurs directly or indirectly in an 'extends' clause of S.
            // Note that this check ignores type parameters and only considers the
            // inheritance hierarchy.
            function isTypeDerivedFrom(source, target) {
                return source.flags & 1048576 /* Union */ ? ts.every(source.types, function (t) { return isTypeDerivedFrom(t, target); }) :
                    target.flags & 1048576 /* Union */ ? ts.some(target.types, function (t) { return isTypeDerivedFrom(source, t); }) :
                        source.flags & 58982400 /* InstantiableNonPrimitive */ ? isTypeDerivedFrom(getBaseConstraintOfType(source) || unknownType, target) :
                            target === globalObjectType ? !!(source.flags & (524288 /* Object */ | 67108864 /* NonPrimitive */)) :
                                target === globalFunctionType ? !!(source.flags & 524288 /* Object */) && isFunctionObjectType(source) :
                                    hasBaseType(source, getTargetType(target)) || (isArrayType(target) && !isReadonlyArrayType(target) && isTypeDerivedFrom(source, globalReadonlyArrayType));
            }
            /**
             * This is *not* a bi-directional relationship.
             * If one needs to check both directions for comparability, use a second call to this function or 'checkTypeComparableTo'.
             *
             * A type S is comparable to a type T if some (but not necessarily all) of the possible values of S are also possible values of T.
             * It is used to check following cases:
             *   - the types of the left and right sides of equality/inequality operators (`===`, `!==`, `==`, `!=`).
             *   - the types of `case` clause expressions and their respective `switch` expressions.
             *   - the type of an expression in a type assertion with the type being asserted.
             */
            function isTypeComparableTo(source, target) {
                return isTypeRelatedTo(source, target, comparableRelation);
            }
            function areTypesComparable(type1, type2) {
                return isTypeComparableTo(type1, type2) || isTypeComparableTo(type2, type1);
            }
            function checkTypeAssignableTo(source, target, errorNode, headMessage, containingMessageChain, errorOutputObject) {
                return checkTypeRelatedTo(source, target, assignableRelation, errorNode, headMessage, containingMessageChain, errorOutputObject);
            }
            /**
             * Like `checkTypeAssignableTo`, but if it would issue an error, instead performs structural comparisons of the types using the given expression node to
             * attempt to issue more specific errors on, for example, specific object literal properties or tuple members.
             */
            function checkTypeAssignableToAndOptionallyElaborate(source, target, errorNode, expr, headMessage, containingMessageChain) {
                return checkTypeRelatedToAndOptionallyElaborate(source, target, assignableRelation, errorNode, expr, headMessage, containingMessageChain, /*errorOutputContainer*/ undefined);
            }
            function checkTypeRelatedToAndOptionallyElaborate(source, target, relation, errorNode, expr, headMessage, containingMessageChain, errorOutputContainer) {
                if (isTypeRelatedTo(source, target, relation))
                    return true;
                if (!errorNode || !elaborateError(expr, source, target, relation, headMessage, containingMessageChain, errorOutputContainer)) {
                    return checkTypeRelatedTo(source, target, relation, errorNode, headMessage, containingMessageChain, errorOutputContainer);
                }
                return false;
            }
            function isOrHasGenericConditional(type) {
                return !!(type.flags & 16777216 /* Conditional */ || (type.flags & 2097152 /* Intersection */ && ts.some(type.types, isOrHasGenericConditional)));
            }
            function elaborateError(node, source, target, relation, headMessage, containingMessageChain, errorOutputContainer) {
                if (!node || isOrHasGenericConditional(target))
                    return false;
                if (!checkTypeRelatedTo(source, target, relation, /*errorNode*/ undefined)
                    && elaborateDidYouMeanToCallOrConstruct(node, source, target, relation, headMessage, containingMessageChain, errorOutputContainer)) {
                    return true;
                }
                switch (node.kind) {
                    case 286 /* JsxExpression */:
                    case 210 /* ParenthesizedExpression */:
                        return elaborateError(node.expression, source, target, relation, headMessage, containingMessageChain, errorOutputContainer);
                    case 219 /* BinaryExpression */:
                        switch (node.operatorToken.kind) {
                            case 63 /* EqualsToken */:
                            case 27 /* CommaToken */:
                                return elaborateError(node.right, source, target, relation, headMessage, containingMessageChain, errorOutputContainer);
                        }
                        break;
                    case 203 /* ObjectLiteralExpression */:
                        return elaborateObjectLiteral(node, source, target, relation, containingMessageChain, errorOutputContainer);
                    case 202 /* ArrayLiteralExpression */:
                        return elaborateArrayLiteral(node, source, target, relation, containingMessageChain, errorOutputContainer);
                    case 284 /* JsxAttributes */:
                        return elaborateJsxComponents(node, source, target, relation, containingMessageChain, errorOutputContainer);
                    case 212 /* ArrowFunction */:
                        return elaborateArrowFunction(node, source, target, relation, containingMessageChain, errorOutputContainer);
                }
                return false;
            }
            function elaborateDidYouMeanToCallOrConstruct(node, source, target, relation, headMessage, containingMessageChain, errorOutputContainer) {
                var callSignatures = getSignaturesOfType(source, 0 /* Call */);
                var constructSignatures = getSignaturesOfType(source, 1 /* Construct */);
                for (var _i = 0, _a = [constructSignatures, callSignatures]; _i < _a.length; _i++) {
                    var signatures = _a[_i];
                    if (ts.some(signatures, function (s) {
                        var returnType = getReturnTypeOfSignature(s);
                        return !(returnType.flags & (1 /* Any */ | 131072 /* Never */)) && checkTypeRelatedTo(returnType, target, relation, /*errorNode*/ undefined);
                    })) {
                        var resultObj = errorOutputContainer || {};
                        checkTypeAssignableTo(source, target, node, headMessage, containingMessageChain, resultObj);
                        var diagnostic = resultObj.errors[resultObj.errors.length - 1];
                        ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(node, signatures === constructSignatures ? ts.Diagnostics.Did_you_mean_to_use_new_with_this_expression : ts.Diagnostics.Did_you_mean_to_call_this_expression));
                        return true;
                    }
                }
                return false;
            }
            function elaborateArrowFunction(node, source, target, relation, containingMessageChain, errorOutputContainer) {
                // Don't elaborate blocks
                if (ts.isBlock(node.body)) {
                    return false;
                }
                // Or functions with annotated parameter types
                if (ts.some(node.parameters, ts.hasType)) {
                    return false;
                }
                var sourceSig = getSingleCallSignature(source);
                if (!sourceSig) {
                    return false;
                }
                var targetSignatures = getSignaturesOfType(target, 0 /* Call */);
                if (!ts.length(targetSignatures)) {
                    return false;
                }
                var returnExpression = node.body;
                var sourceReturn = getReturnTypeOfSignature(sourceSig);
                var targetReturn = getUnionType(ts.map(targetSignatures, getReturnTypeOfSignature));
                if (!checkTypeRelatedTo(sourceReturn, targetReturn, relation, /*errorNode*/ undefined)) {
                    var elaborated = returnExpression && elaborateError(returnExpression, sourceReturn, targetReturn, relation, /*headMessage*/ undefined, containingMessageChain, errorOutputContainer);
                    if (elaborated) {
                        return elaborated;
                    }
                    var resultObj = errorOutputContainer || {};
                    checkTypeRelatedTo(sourceReturn, targetReturn, relation, returnExpression, /*message*/ undefined, containingMessageChain, resultObj);
                    if (resultObj.errors) {
                        if (target.symbol && ts.length(target.symbol.declarations)) {
                            ts.addRelatedInfo(resultObj.errors[resultObj.errors.length - 1], ts.createDiagnosticForNode(target.symbol.declarations[0], ts.Diagnostics.The_expected_type_comes_from_the_return_type_of_this_signature));
                        }
                        if ((ts.getFunctionFlags(node) & 2 /* Async */) === 0
                            // exclude cases where source itself is promisy - this way we don't make a suggestion when relating
                            // an IPromise and a Promise that are slightly different
                            && !getTypeOfPropertyOfType(sourceReturn, "then")
                            && checkTypeRelatedTo(createPromiseType(sourceReturn), targetReturn, relation, /*errorNode*/ undefined)) {
                            ts.addRelatedInfo(resultObj.errors[resultObj.errors.length - 1], ts.createDiagnosticForNode(node, ts.Diagnostics.Did_you_mean_to_mark_this_function_as_async));
                        }
                        return true;
                    }
                }
                return false;
            }
            function getBestMatchIndexedAccessTypeOrUndefined(source, target, nameType) {
                var idx = getIndexedAccessTypeOrUndefined(target, nameType);
                if (idx) {
                    return idx;
                }
                if (target.flags & 1048576 /* Union */) {
                    var best = getBestMatchingType(source, target);
                    if (best) {
                        return getIndexedAccessTypeOrUndefined(best, nameType);
                    }
                }
            }
            function checkExpressionForMutableLocationWithContextualType(next, sourcePropType) {
                next.contextualType = sourcePropType;
                try {
                    return checkExpressionForMutableLocation(next, 1 /* Contextual */, sourcePropType);
                }
                finally {
                    next.contextualType = undefined;
                }
            }
            /**
             * For every element returned from the iterator, checks that element to issue an error on a property of that element's type
             * If that element would issue an error, we first attempt to dive into that element's inner expression and issue a more specific error by recuring into `elaborateError`
             * Otherwise, we issue an error on _every_ element which fail the assignability check
             */
            function elaborateElementwise(iterator, source, target, relation, containingMessageChain, errorOutputContainer) {
                // Assignability failure - check each prop individually, and if that fails, fall back on the bad error span
                var reportedError = false;
                for (var status = iterator.next(); !status.done; status = iterator.next()) {
                    var _a = status.value, prop = _a.errorNode, next = _a.innerExpression, nameType = _a.nameType, errorMessage = _a.errorMessage;
                    var targetPropType = getBestMatchIndexedAccessTypeOrUndefined(source, target, nameType);
                    if (!targetPropType || targetPropType.flags & 8388608 /* IndexedAccess */)
                        continue; // Don't elaborate on indexes on generic variables
                    var sourcePropType = getIndexedAccessTypeOrUndefined(source, nameType);
                    if (!sourcePropType)
                        continue;
                    var propName = getPropertyNameFromIndex(nameType, /*accessNode*/ undefined);
                    var targetIsOptional = !!(propName && (getPropertyOfType(target, propName) || unknownSymbol).flags & 16777216 /* Optional */);
                    var sourceIsOptional = !!(propName && (getPropertyOfType(source, propName) || unknownSymbol).flags & 16777216 /* Optional */);
                    targetPropType = removeMissingType(targetPropType, targetIsOptional);
                    sourcePropType = removeMissingType(sourcePropType, targetIsOptional && sourceIsOptional);
                    if (!checkTypeRelatedTo(sourcePropType, targetPropType, relation, /*errorNode*/ undefined)) {
                        var elaborated = next && elaborateError(next, sourcePropType, targetPropType, relation, /*headMessage*/ undefined, containingMessageChain, errorOutputContainer);
                        if (elaborated) {
                            reportedError = true;
                        }
                        else {
                            // Issue error on the prop itself, since the prop couldn't elaborate the error
                            var resultObj = errorOutputContainer || {};
                            // Use the expression type, if available
                            var specificSource = next ? checkExpressionForMutableLocationWithContextualType(next, sourcePropType) : sourcePropType;
                            var result = checkTypeRelatedTo(specificSource, targetPropType, relation, prop, errorMessage, containingMessageChain, resultObj);
                            if (result && specificSource !== sourcePropType) {
                                // If for whatever reason the expression type doesn't yield an error, make sure we still issue an error on the sourcePropType
                                checkTypeRelatedTo(sourcePropType, targetPropType, relation, prop, errorMessage, containingMessageChain, resultObj);
                            }
                            if (resultObj.errors) {
                                var reportedDiag = resultObj.errors[resultObj.errors.length - 1];
                                var propertyName = isTypeUsableAsPropertyName(nameType) ? getPropertyNameFromType(nameType) : undefined;
                                var targetProp = propertyName !== undefined ? getPropertyOfType(target, propertyName) : undefined;
                                var issuedElaboration = false;
                                if (!targetProp) {
                                    var indexInfo = getApplicableIndexInfo(target, nameType);
                                    if (indexInfo && indexInfo.declaration && !ts.getSourceFileOfNode(indexInfo.declaration).hasNoDefaultLib) {
                                        issuedElaboration = true;
                                        ts.addRelatedInfo(reportedDiag, ts.createDiagnosticForNode(indexInfo.declaration, ts.Diagnostics.The_expected_type_comes_from_this_index_signature));
                                    }
                                }
                                if (!issuedElaboration && (targetProp && ts.length(targetProp.declarations) || target.symbol && ts.length(target.symbol.declarations))) {
                                    var targetNode = targetProp && ts.length(targetProp.declarations) ? targetProp.declarations[0] : target.symbol.declarations[0];
                                    if (!ts.getSourceFileOfNode(targetNode).hasNoDefaultLib) {
                                        ts.addRelatedInfo(reportedDiag, ts.createDiagnosticForNode(targetNode, ts.Diagnostics.The_expected_type_comes_from_property_0_which_is_declared_here_on_type_1, propertyName && !(nameType.flags & 8192 /* UniqueESSymbol */) ? ts.unescapeLeadingUnderscores(propertyName) : typeToString(nameType), typeToString(target)));
                                    }
                                }
                            }
                            reportedError = true;
                        }
                    }
                }
                return reportedError;
            }
            function generateJsxAttributes(node) {
                var _i, _a, prop;
                return __generator(this, function (_b) {
                    switch (_b.label) {
                        case 0:
                            if (!ts.length(node.properties))
                                return [2 /*return*/];
                            _i = 0, _a = node.properties;
                            _b.label = 1;
                        case 1:
                            if (!(_i < _a.length))
                                return [3 /*break*/, 4];
                            prop = _a[_i];
                            if (ts.isJsxSpreadAttribute(prop) || isHyphenatedJsxName(ts.idText(prop.name)))
                                return [3 /*break*/, 3];
                            return [4 /*yield*/, { errorNode: prop.name, innerExpression: prop.initializer, nameType: getStringLiteralType(ts.idText(prop.name)) }];
                        case 2:
                            _b.sent();
                            _b.label = 3;
                        case 3:
                            _i++;
                            return [3 /*break*/, 1];
                        case 4: return [2 /*return*/];
                    }
                });
            }
            function generateJsxChildren(node, getInvalidTextDiagnostic) {
                var memberOffset, i, child, nameType, elem;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            if (!ts.length(node.children))
                                return [2 /*return*/];
                            memberOffset = 0;
                            i = 0;
                            _a.label = 1;
                        case 1:
                            if (!(i < node.children.length))
                                return [3 /*break*/, 5];
                            child = node.children[i];
                            nameType = getNumberLiteralType(i - memberOffset);
                            elem = getElaborationElementForJsxChild(child, nameType, getInvalidTextDiagnostic);
                            if (!elem)
                                return [3 /*break*/, 3];
                            return [4 /*yield*/, elem];
                        case 2:
                            _a.sent();
                            return [3 /*break*/, 4];
                        case 3:
                            memberOffset++;
                            _a.label = 4;
                        case 4:
                            i++;
                            return [3 /*break*/, 1];
                        case 5: return [2 /*return*/];
                    }
                });
            }
            function getElaborationElementForJsxChild(child, nameType, getInvalidTextDiagnostic) {
                switch (child.kind) {
                    case 286 /* JsxExpression */:
                        // child is of the type of the expression
                        return { errorNode: child, innerExpression: child.expression, nameType: nameType };
                    case 11 /* JsxText */:
                        if (child.containsOnlyTriviaWhiteSpaces) {
                            break; // Whitespace only jsx text isn't real jsx text
                        }
                        // child is a string
                        return { errorNode: child, innerExpression: undefined, nameType: nameType, errorMessage: getInvalidTextDiagnostic() };
                    case 276 /* JsxElement */:
                    case 277 /* JsxSelfClosingElement */:
                    case 280 /* JsxFragment */:
                        // child is of type JSX.Element
                        return { errorNode: child, innerExpression: child, nameType: nameType };
                    default:
                        return ts.Debug.assertNever(child, "Found invalid jsx child");
                }
            }
            function elaborateJsxComponents(node, source, target, relation, containingMessageChain, errorOutputContainer) {
                var result = elaborateElementwise(generateJsxAttributes(node), source, target, relation, containingMessageChain, errorOutputContainer);
                var invalidTextDiagnostic;
                if (ts.isJsxOpeningElement(node.parent) && ts.isJsxElement(node.parent.parent)) {
                    var containingElement = node.parent.parent;
                    var childPropName = getJsxElementChildrenPropertyName(getJsxNamespaceAt(node));
                    var childrenPropName = childPropName === undefined ? "children" : ts.unescapeLeadingUnderscores(childPropName);
                    var childrenNameType = getStringLiteralType(childrenPropName);
                    var childrenTargetType = getIndexedAccessType(target, childrenNameType);
                    var validChildren = ts.getSemanticJsxChildren(containingElement.children);
                    if (!ts.length(validChildren)) {
                        return result;
                    }
                    var moreThanOneRealChildren = ts.length(validChildren) > 1;
                    var arrayLikeTargetParts = filterType(childrenTargetType, isArrayOrTupleLikeType);
                    var nonArrayLikeTargetParts = filterType(childrenTargetType, function (t) { return !isArrayOrTupleLikeType(t); });
                    if (moreThanOneRealChildren) {
                        if (arrayLikeTargetParts !== neverType) {
                            var realSource = createTupleType(checkJsxChildren(containingElement, 0 /* Normal */));
                            var children = generateJsxChildren(containingElement, getInvalidTextualChildDiagnostic);
                            result = elaborateElementwise(children, realSource, arrayLikeTargetParts, relation, containingMessageChain, errorOutputContainer) || result;
                        }
                        else if (!isTypeRelatedTo(getIndexedAccessType(source, childrenNameType), childrenTargetType, relation)) {
                            // arity mismatch
                            result = true;
                            var diag = error(containingElement.openingElement.tagName, ts.Diagnostics.This_JSX_tag_s_0_prop_expects_a_single_child_of_type_1_but_multiple_children_were_provided, childrenPropName, typeToString(childrenTargetType));
                            if (errorOutputContainer && errorOutputContainer.skipLogging) {
                                (errorOutputContainer.errors || (errorOutputContainer.errors = [])).push(diag);
                            }
                        }
                    }
                    else {
                        if (nonArrayLikeTargetParts !== neverType) {
                            var child = validChildren[0];
                            var elem_1 = getElaborationElementForJsxChild(child, childrenNameType, getInvalidTextualChildDiagnostic);
                            if (elem_1) {
                                result = elaborateElementwise((function () {
                                    return __generator(this, function (_a) {
                                        switch (_a.label) {
                                            case 0: return [4 /*yield*/, elem_1];
                                            case 1:
                                                _a.sent();
                                                return [2 /*return*/];
                                        }
                                    });
                                })(), source, target, relation, containingMessageChain, errorOutputContainer) || result;
                            }
                        }
                        else if (!isTypeRelatedTo(getIndexedAccessType(source, childrenNameType), childrenTargetType, relation)) {
                            // arity mismatch
                            result = true;
                            var diag = error(containingElement.openingElement.tagName, ts.Diagnostics.This_JSX_tag_s_0_prop_expects_type_1_which_requires_multiple_children_but_only_a_single_child_was_provided, childrenPropName, typeToString(childrenTargetType));
                            if (errorOutputContainer && errorOutputContainer.skipLogging) {
                                (errorOutputContainer.errors || (errorOutputContainer.errors = [])).push(diag);
                            }
                        }
                    }
                }
                return result;
                function getInvalidTextualChildDiagnostic() {
                    if (!invalidTextDiagnostic) {
                        var tagNameText = ts.getTextOfNode(node.parent.tagName);
                        var childPropName = getJsxElementChildrenPropertyName(getJsxNamespaceAt(node));
                        var childrenPropName = childPropName === undefined ? "children" : ts.unescapeLeadingUnderscores(childPropName);
                        var childrenTargetType = getIndexedAccessType(target, getStringLiteralType(childrenPropName));
                        var diagnostic = ts.Diagnostics._0_components_don_t_accept_text_as_child_elements_Text_in_JSX_has_the_type_string_but_the_expected_type_of_1_is_2;
                        invalidTextDiagnostic = __assign(__assign({}, diagnostic), { key: "!!ALREADY FORMATTED!!", message: ts.formatMessage(/*_dummy*/ undefined, diagnostic, tagNameText, childrenPropName, typeToString(childrenTargetType)) });
                    }
                    return invalidTextDiagnostic;
                }
            }
            function generateLimitedTupleElements(node, target) {
                var len, i, elem, nameType;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            len = ts.length(node.elements);
                            if (!len)
                                return [2 /*return*/];
                            i = 0;
                            _a.label = 1;
                        case 1:
                            if (!(i < len))
                                return [3 /*break*/, 4];
                            // Skip elements which do not exist in the target - a length error on the tuple overall is likely better than an error on a mismatched index signature
                            if (isTupleLikeType(target) && !getPropertyOfType(target, ("" + i)))
                                return [3 /*break*/, 3];
                            elem = node.elements[i];
                            if (ts.isOmittedExpression(elem))
                                return [3 /*break*/, 3];
                            nameType = getNumberLiteralType(i);
                            return [4 /*yield*/, { errorNode: elem, innerExpression: elem, nameType: nameType }];
                        case 2:
                            _a.sent();
                            _a.label = 3;
                        case 3:
                            i++;
                            return [3 /*break*/, 1];
                        case 4: return [2 /*return*/];
                    }
                });
            }
            function elaborateArrayLiteral(node, source, target, relation, containingMessageChain, errorOutputContainer) {
                if (target.flags & 131068 /* Primitive */)
                    return false;
                if (isTupleLikeType(source)) {
                    return elaborateElementwise(generateLimitedTupleElements(node, target), source, target, relation, containingMessageChain, errorOutputContainer);
                }
                // recreate a tuple from the elements, if possible
                // Since we're re-doing the expression type, we need to reapply the contextual type
                var oldContext = node.contextualType;
                node.contextualType = target;
                try {
                    var tupleizedType = checkArrayLiteral(node, 1 /* Contextual */, /*forceTuple*/ true);
                    node.contextualType = oldContext;
                    if (isTupleLikeType(tupleizedType)) {
                        return elaborateElementwise(generateLimitedTupleElements(node, target), tupleizedType, target, relation, containingMessageChain, errorOutputContainer);
                    }
                    return false;
                }
                finally {
                    node.contextualType = oldContext;
                }
            }
            function generateObjectLiteralElements(node) {
                var _i, _a, prop, type, _b;
                return __generator(this, function (_c) {
                    switch (_c.label) {
                        case 0:
                            if (!ts.length(node.properties))
                                return [2 /*return*/];
                            _i = 0, _a = node.properties;
                            _c.label = 1;
                        case 1:
                            if (!(_i < _a.length))
                                return [3 /*break*/, 8];
                            prop = _a[_i];
                            if (ts.isSpreadAssignment(prop))
                                return [3 /*break*/, 7];
                            type = getLiteralTypeFromProperty(getSymbolOfNode(prop), 8576 /* StringOrNumberLiteralOrUnique */);
                            if (!type || (type.flags & 131072 /* Never */)) {
                                return [3 /*break*/, 7];
                            }
                            _b = prop.kind;
                            switch (_b) {
                                case 171 /* SetAccessor */: return [3 /*break*/, 2];
                                case 170 /* GetAccessor */: return [3 /*break*/, 2];
                                case 167 /* MethodDeclaration */: return [3 /*break*/, 2];
                                case 292 /* ShorthandPropertyAssignment */: return [3 /*break*/, 2];
                                case 291 /* PropertyAssignment */: return [3 /*break*/, 4];
                            }
                            return [3 /*break*/, 6];
                        case 2: return [4 /*yield*/, { errorNode: prop.name, innerExpression: undefined, nameType: type }];
                        case 3:
                            _c.sent();
                            return [3 /*break*/, 7];
                        case 4: return [4 /*yield*/, { errorNode: prop.name, innerExpression: prop.initializer, nameType: type, errorMessage: ts.isComputedNonLiteralName(prop.name) ? ts.Diagnostics.Type_of_computed_property_s_value_is_0_which_is_not_assignable_to_type_1 : undefined }];
                        case 5:
                            _c.sent();
                            return [3 /*break*/, 7];
                        case 6:
                            ts.Debug.assertNever(prop);
                            _c.label = 7;
                        case 7:
                            _i++;
                            return [3 /*break*/, 1];
                        case 8: return [2 /*return*/];
                    }
                });
            }
            function elaborateObjectLiteral(node, source, target, relation, containingMessageChain, errorOutputContainer) {
                if (target.flags & 131068 /* Primitive */)
                    return false;
                return elaborateElementwise(generateObjectLiteralElements(node), source, target, relation, containingMessageChain, errorOutputContainer);
            }
            /**
             * This is *not* a bi-directional relationship.
             * If one needs to check both directions for comparability, use a second call to this function or 'isTypeComparableTo'.
             */
            function checkTypeComparableTo(source, target, errorNode, headMessage, containingMessageChain) {
                return checkTypeRelatedTo(source, target, comparableRelation, errorNode, headMessage, containingMessageChain);
            }
            function isSignatureAssignableTo(source, target, ignoreReturnTypes) {
                return compareSignaturesRelated(source, target, ignoreReturnTypes ? 4 /* IgnoreReturnTypes */ : 0, /*reportErrors*/ false, 
                /*errorReporter*/ undefined, /*errorReporter*/ undefined, compareTypesAssignable, /*reportUnreliableMarkers*/ undefined) !== 0 /* False */;
            }
            /**
             * Returns true if `s` is `(...args: any[]) => any` or `(this: any, ...args: any[]) => any`
             */
            function isAnySignature(s) {
                return !s.typeParameters && (!s.thisParameter || isTypeAny(getTypeOfParameter(s.thisParameter))) && s.parameters.length === 1 &&
                    signatureHasRestParameter(s) && (getTypeOfParameter(s.parameters[0]) === anyArrayType || isTypeAny(getTypeOfParameter(s.parameters[0]))) &&
                    isTypeAny(getReturnTypeOfSignature(s));
            }
            /**
             * See signatureRelatedTo, compareSignaturesIdentical
             */
            function compareSignaturesRelated(source, target, checkMode, reportErrors, errorReporter, incompatibleErrorReporter, compareTypes, reportUnreliableMarkers) {
                // TODO (drosen): De-duplicate code between related functions.
                if (source === target) {
                    return -1 /* True */;
                }
                if (isAnySignature(target)) {
                    return -1 /* True */;
                }
                var targetCount = getParameterCount(target);
                var sourceHasMoreParameters = !hasEffectiveRestParameter(target) &&
                    (checkMode & 8 /* StrictArity */ ? hasEffectiveRestParameter(source) || getParameterCount(source) > targetCount : getMinArgumentCount(source) > targetCount);
                if (sourceHasMoreParameters) {
                    return 0 /* False */;
                }
                if (source.typeParameters && source.typeParameters !== target.typeParameters) {
                    target = getCanonicalSignature(target);
                    source = instantiateSignatureInContextOf(source, target, /*inferenceContext*/ undefined, compareTypes);
                }
                var sourceCount = getParameterCount(source);
                var sourceRestType = getNonArrayRestType(source);
                var targetRestType = getNonArrayRestType(target);
                if (sourceRestType || targetRestType) {
                    void instantiateType(sourceRestType || targetRestType, reportUnreliableMarkers);
                }
                if (sourceRestType && targetRestType && sourceCount !== targetCount) {
                    // We're not able to relate misaligned complex rest parameters
                    return 0 /* False */;
                }
                var kind = target.declaration ? target.declaration.kind : 0 /* Unknown */;
                var strictVariance = !(checkMode & 3 /* Callback */) && strictFunctionTypes && kind !== 167 /* MethodDeclaration */ &&
                    kind !== 166 /* MethodSignature */ && kind !== 169 /* Constructor */;
                var result = -1 /* True */;
                var sourceThisType = getThisTypeOfSignature(source);
                if (sourceThisType && sourceThisType !== voidType) {
                    var targetThisType = getThisTypeOfSignature(target);
                    if (targetThisType) {
                        // void sources are assignable to anything.
                        var related = !strictVariance && compareTypes(sourceThisType, targetThisType, /*reportErrors*/ false)
                            || compareTypes(targetThisType, sourceThisType, reportErrors);
                        if (!related) {
                            if (reportErrors) {
                                errorReporter(ts.Diagnostics.The_this_types_of_each_signature_are_incompatible);
                            }
                            return 0 /* False */;
                        }
                        result &= related;
                    }
                }
                var paramCount = sourceRestType || targetRestType ? Math.min(sourceCount, targetCount) : Math.max(sourceCount, targetCount);
                var restIndex = sourceRestType || targetRestType ? paramCount - 1 : -1;
                for (var i = 0; i < paramCount; i++) {
                    var sourceType = i === restIndex ? getRestTypeAtPosition(source, i) : tryGetTypeAtPosition(source, i);
                    var targetType = i === restIndex ? getRestTypeAtPosition(target, i) : tryGetTypeAtPosition(target, i);
                    if (sourceType && targetType) {
                        // In order to ensure that any generic type Foo<T> is at least co-variant with respect to T no matter
                        // how Foo uses T, we need to relate parameters bi-variantly (given that parameters are input positions,
                        // they naturally relate only contra-variantly). However, if the source and target parameters both have
                        // function types with a single call signature, we know we are relating two callback parameters. In
                        // that case it is sufficient to only relate the parameters of the signatures co-variantly because,
                        // similar to return values, callback parameters are output positions. This means that a Promise<T>,
                        // where T is used only in callback parameter positions, will be co-variant (as opposed to bi-variant)
                        // with respect to T.
                        var sourceSig = checkMode & 3 /* Callback */ ? undefined : getSingleCallSignature(getNonNullableType(sourceType));
                        var targetSig = checkMode & 3 /* Callback */ ? undefined : getSingleCallSignature(getNonNullableType(targetType));
                        var callbacks = sourceSig && targetSig && !getTypePredicateOfSignature(sourceSig) && !getTypePredicateOfSignature(targetSig) &&
                            (getFalsyFlags(sourceType) & 98304 /* Nullable */) === (getFalsyFlags(targetType) & 98304 /* Nullable */);
                        var related = callbacks ?
                            compareSignaturesRelated(targetSig, sourceSig, (checkMode & 8 /* StrictArity */) | (strictVariance ? 2 /* StrictCallback */ : 1 /* BivariantCallback */), reportErrors, errorReporter, incompatibleErrorReporter, compareTypes, reportUnreliableMarkers) :
                            !(checkMode & 3 /* Callback */) && !strictVariance && compareTypes(sourceType, targetType, /*reportErrors*/ false) || compareTypes(targetType, sourceType, reportErrors);
                        // With strict arity, (x: number | undefined) => void is a subtype of (x?: number | undefined) => void
                        if (related && checkMode & 8 /* StrictArity */ && i >= getMinArgumentCount(source) && i < getMinArgumentCount(target) && compareTypes(sourceType, targetType, /*reportErrors*/ false)) {
                            related = 0 /* False */;
                        }
                        if (!related) {
                            if (reportErrors) {
                                errorReporter(ts.Diagnostics.Types_of_parameters_0_and_1_are_incompatible, ts.unescapeLeadingUnderscores(getParameterNameAtPosition(source, i)), ts.unescapeLeadingUnderscores(getParameterNameAtPosition(target, i)));
                            }
                            return 0 /* False */;
                        }
                        result &= related;
                    }
                }
                if (!(checkMode & 4 /* IgnoreReturnTypes */)) {
                    // If a signature resolution is already in-flight, skip issuing a circularity error
                    // here and just use the `any` type directly
                    var targetReturnType = isResolvingReturnTypeOfSignature(target) ? anyType
                        : target.declaration && isJSConstructor(target.declaration) ? getDeclaredTypeOfClassOrInterface(getMergedSymbol(target.declaration.symbol))
                            : getReturnTypeOfSignature(target);
                    if (targetReturnType === voidType) {
                        return result;
                    }
                    var sourceReturnType = isResolvingReturnTypeOfSignature(source) ? anyType
                        : source.declaration && isJSConstructor(source.declaration) ? getDeclaredTypeOfClassOrInterface(getMergedSymbol(source.declaration.symbol))
                            : getReturnTypeOfSignature(source);
                    // The following block preserves behavior forbidding boolean returning functions from being assignable to type guard returning functions
                    var targetTypePredicate = getTypePredicateOfSignature(target);
                    if (targetTypePredicate) {
                        var sourceTypePredicate = getTypePredicateOfSignature(source);
                        if (sourceTypePredicate) {
                            result &= compareTypePredicateRelatedTo(sourceTypePredicate, targetTypePredicate, reportErrors, errorReporter, compareTypes);
                        }
                        else if (ts.isIdentifierTypePredicate(targetTypePredicate)) {
                            if (reportErrors) {
                                errorReporter(ts.Diagnostics.Signature_0_must_be_a_type_predicate, signatureToString(source));
                            }
                            return 0 /* False */;
                        }
                    }
                    else {
                        // When relating callback signatures, we still need to relate return types bi-variantly as otherwise
                        // the containing type wouldn't be co-variant. For example, interface Foo<T> { add(cb: () => T): void }
                        // wouldn't be co-variant for T without this rule.
                        result &= checkMode & 1 /* BivariantCallback */ && compareTypes(targetReturnType, sourceReturnType, /*reportErrors*/ false) ||
                            compareTypes(sourceReturnType, targetReturnType, reportErrors);
                        if (!result && reportErrors && incompatibleErrorReporter) {
                            incompatibleErrorReporter(sourceReturnType, targetReturnType);
                        }
                    }
                }
                return result;
            }
            function compareTypePredicateRelatedTo(source, target, reportErrors, errorReporter, compareTypes) {
                if (source.kind !== target.kind) {
                    if (reportErrors) {
                        errorReporter(ts.Diagnostics.A_this_based_type_guard_is_not_compatible_with_a_parameter_based_type_guard);
                        errorReporter(ts.Diagnostics.Type_predicate_0_is_not_assignable_to_1, typePredicateToString(source), typePredicateToString(target));
                    }
                    return 0 /* False */;
                }
                if (source.kind === 1 /* Identifier */ || source.kind === 3 /* AssertsIdentifier */) {
                    if (source.parameterIndex !== target.parameterIndex) {
                        if (reportErrors) {
                            errorReporter(ts.Diagnostics.Parameter_0_is_not_in_the_same_position_as_parameter_1, source.parameterName, target.parameterName);
                            errorReporter(ts.Diagnostics.Type_predicate_0_is_not_assignable_to_1, typePredicateToString(source), typePredicateToString(target));
                        }
                        return 0 /* False */;
                    }
                }
                var related = source.type === target.type ? -1 /* True */ :
                    source.type && target.type ? compareTypes(source.type, target.type, reportErrors) :
                        0 /* False */;
                if (related === 0 /* False */ && reportErrors) {
                    errorReporter(ts.Diagnostics.Type_predicate_0_is_not_assignable_to_1, typePredicateToString(source), typePredicateToString(target));
                }
                return related;
            }
            function isImplementationCompatibleWithOverload(implementation, overload) {
                var erasedSource = getErasedSignature(implementation);
                var erasedTarget = getErasedSignature(overload);
                // First see if the return types are compatible in either direction.
                var sourceReturnType = getReturnTypeOfSignature(erasedSource);
                var targetReturnType = getReturnTypeOfSignature(erasedTarget);
                if (targetReturnType === voidType
                    || isTypeRelatedTo(targetReturnType, sourceReturnType, assignableRelation)
                    || isTypeRelatedTo(sourceReturnType, targetReturnType, assignableRelation)) {
                    return isSignatureAssignableTo(erasedSource, erasedTarget, /*ignoreReturnTypes*/ true);
                }
                return false;
            }
            function isEmptyResolvedType(t) {
                return t !== anyFunctionType &&
                    t.properties.length === 0 &&
                    t.callSignatures.length === 0 &&
                    t.constructSignatures.length === 0 &&
                    t.indexInfos.length === 0;
            }
            function isEmptyObjectType(type) {
                return type.flags & 524288 /* Object */ ? !isGenericMappedType(type) && isEmptyResolvedType(resolveStructuredTypeMembers(type)) :
                    type.flags & 67108864 /* NonPrimitive */ ? true :
                        type.flags & 1048576 /* Union */ ? ts.some(type.types, isEmptyObjectType) :
                            type.flags & 2097152 /* Intersection */ ? ts.every(type.types, isEmptyObjectType) :
                                false;
            }
            function isEmptyAnonymousObjectType(type) {
                return !!(ts.getObjectFlags(type) & 16 /* Anonymous */ && (type.members && isEmptyResolvedType(type) ||
                    type.symbol && type.symbol.flags & 2048 /* TypeLiteral */ && getMembersOfSymbol(type.symbol).size === 0));
            }
            function isStringIndexSignatureOnlyType(type) {
                return type.flags & 524288 /* Object */ && !isGenericMappedType(type) && getPropertiesOfType(type).length === 0 && getIndexInfosOfType(type).length === 1 && !!getIndexInfoOfType(type, stringType) ||
                    type.flags & 3145728 /* UnionOrIntersection */ && ts.every(type.types, isStringIndexSignatureOnlyType) ||
                    false;
            }
            function isEnumTypeRelatedTo(sourceSymbol, targetSymbol, errorReporter) {
                if (sourceSymbol === targetSymbol) {
                    return true;
                }
                var id = getSymbolId(sourceSymbol) + "," + getSymbolId(targetSymbol);
                var entry = enumRelation.get(id);
                if (entry !== undefined && !(!(entry & 4 /* Reported */) && entry & 2 /* Failed */ && errorReporter)) {
                    return !!(entry & 1 /* Succeeded */);
                }
                if (sourceSymbol.escapedName !== targetSymbol.escapedName || !(sourceSymbol.flags & 256 /* RegularEnum */) || !(targetSymbol.flags & 256 /* RegularEnum */)) {
                    enumRelation.set(id, 2 /* Failed */ | 4 /* Reported */);
                    return false;
                }
                var targetEnumType = getTypeOfSymbol(targetSymbol);
                for (var _i = 0, _a = getPropertiesOfType(getTypeOfSymbol(sourceSymbol)); _i < _a.length; _i++) {
                    var property = _a[_i];
                    if (property.flags & 8 /* EnumMember */) {
                        var targetProperty = getPropertyOfType(targetEnumType, property.escapedName);
                        if (!targetProperty || !(targetProperty.flags & 8 /* EnumMember */)) {
                            if (errorReporter) {
                                errorReporter(ts.Diagnostics.Property_0_is_missing_in_type_1, ts.symbolName(property), typeToString(getDeclaredTypeOfSymbol(targetSymbol), /*enclosingDeclaration*/ undefined, 64 /* UseFullyQualifiedType */));
                                enumRelation.set(id, 2 /* Failed */ | 4 /* Reported */);
                            }
                            else {
                                enumRelation.set(id, 2 /* Failed */);
                            }
                            return false;
                        }
                    }
                }
                enumRelation.set(id, 1 /* Succeeded */);
                return true;
            }
            function isSimpleTypeRelatedTo(source, target, relation, errorReporter) {
                var s = source.flags;
                var t = target.flags;
                if (t & 3 /* AnyOrUnknown */ || s & 131072 /* Never */ || source === wildcardType)
                    return true;
                if (t & 131072 /* Never */)
                    return false;
                if (s & 402653316 /* StringLike */ && t & 4 /* String */)
                    return true;
                if (s & 128 /* StringLiteral */ && s & 1024 /* EnumLiteral */ &&
                    t & 128 /* StringLiteral */ && !(t & 1024 /* EnumLiteral */) &&
                    source.value === target.value)
                    return true;
                if (s & 296 /* NumberLike */ && t & 8 /* Number */)
                    return true;
                if (s & 256 /* NumberLiteral */ && s & 1024 /* EnumLiteral */ &&
                    t & 256 /* NumberLiteral */ && !(t & 1024 /* EnumLiteral */) &&
                    source.value === target.value)
                    return true;
                if (s & 2112 /* BigIntLike */ && t & 64 /* BigInt */)
                    return true;
                if (s & 528 /* BooleanLike */ && t & 16 /* Boolean */)
                    return true;
                if (s & 12288 /* ESSymbolLike */ && t & 4096 /* ESSymbol */)
                    return true;
                if (s & 32 /* Enum */ && t & 32 /* Enum */ && isEnumTypeRelatedTo(source.symbol, target.symbol, errorReporter))
                    return true;
                if (s & 1024 /* EnumLiteral */ && t & 1024 /* EnumLiteral */) {
                    if (s & 1048576 /* Union */ && t & 1048576 /* Union */ && isEnumTypeRelatedTo(source.symbol, target.symbol, errorReporter))
                        return true;
                    if (s & 2944 /* Literal */ && t & 2944 /* Literal */ &&
                        source.value === target.value &&
                        isEnumTypeRelatedTo(getParentOfSymbol(source.symbol), getParentOfSymbol(target.symbol), errorReporter))
                        return true;
                }
                if (s & 32768 /* Undefined */ && (!strictNullChecks || t & (32768 /* Undefined */ | 16384 /* Void */)))
                    return true;
                if (s & 65536 /* Null */ && (!strictNullChecks || t & 65536 /* Null */))
                    return true;
                if (s & 524288 /* Object */ && t & 67108864 /* NonPrimitive */)
                    return true;
                if (relation === assignableRelation || relation === comparableRelation) {
                    if (s & 1 /* Any */)
                        return true;
                    // Type number or any numeric literal type is assignable to any numeric enum type or any
                    // numeric enum literal type. This rule exists for backwards compatibility reasons because
                    // bit-flag enum types sometimes look like literal enum types with numeric literal values.
                    if (s & (8 /* Number */ | 256 /* NumberLiteral */) && !(s & 1024 /* EnumLiteral */) && (t & 32 /* Enum */ || relation === assignableRelation && t & 256 /* NumberLiteral */ && t & 1024 /* EnumLiteral */))
                        return true;
                }
                return false;
            }
            function isTypeRelatedTo(source, target, relation) {
                if (isFreshLiteralType(source)) {
                    source = source.regularType;
                }
                if (isFreshLiteralType(target)) {
                    target = target.regularType;
                }
                if (source === target) {
                    return true;
                }
                if (relation !== identityRelation) {
                    if (relation === comparableRelation && !(target.flags & 131072 /* Never */) && isSimpleTypeRelatedTo(target, source, relation) || isSimpleTypeRelatedTo(source, target, relation)) {
                        return true;
                    }
                }
                else {
                    if (source.flags !== target.flags)
                        return false;
                    if (source.flags & 67358815 /* Singleton */)
                        return true;
                }
                if (source.flags & 524288 /* Object */ && target.flags & 524288 /* Object */) {
                    var related = relation.get(getRelationKey(source, target, 0 /* None */, relation));
                    if (related !== undefined) {
                        return !!(related & 1 /* Succeeded */);
                    }
                }
                if (source.flags & 469499904 /* StructuredOrInstantiable */ || target.flags & 469499904 /* StructuredOrInstantiable */) {
                    return checkTypeRelatedTo(source, target, relation, /*errorNode*/ undefined);
                }
                return false;
            }
            function isIgnoredJsxProperty(source, sourceProp) {
                return ts.getObjectFlags(source) & 2048 /* JsxAttributes */ && isHyphenatedJsxName(sourceProp.escapedName);
            }
            function getNormalizedType(type, writing) {
                while (true) {
                    var t = isFreshLiteralType(type) ? type.regularType :
                        ts.getObjectFlags(type) & 4 /* Reference */ && type.node ? createTypeReference(type.target, getTypeArguments(type)) :
                            type.flags & 3145728 /* UnionOrIntersection */ ? getReducedType(type) :
                                type.flags & 33554432 /* Substitution */ ? writing ? type.baseType : type.substitute :
                                    type.flags & 25165824 /* Simplifiable */ ? getSimplifiedType(type, writing) :
                                        type;
                    t = getSingleBaseForNonAugmentingSubtype(t) || t;
                    if (t === type)
                        break;
                    type = t;
                }
                return type;
            }
            /**
             * Checks if 'source' is related to 'target' (e.g.: is a assignable to).
             * @param source The left-hand-side of the relation.
             * @param target The right-hand-side of the relation.
             * @param relation The relation considered. One of 'identityRelation', 'subtypeRelation', 'assignableRelation', or 'comparableRelation'.
             * Used as both to determine which checks are performed and as a cache of previously computed results.
             * @param errorNode The suggested node upon which all errors will be reported, if defined. This may or may not be the actual node used.
             * @param headMessage If the error chain should be prepended by a head message, then headMessage will be used.
             * @param containingMessageChain A chain of errors to prepend any new errors found.
             * @param errorOutputContainer Return the diagnostic. Do not log if 'skipLogging' is truthy.
             */
            function checkTypeRelatedTo(source, target, relation, errorNode, headMessage, containingMessageChain, errorOutputContainer) {
                var errorInfo;
                var relatedInfo;
                var maybeKeys;
                var sourceStack;
                var targetStack;
                var maybeCount = 0;
                var depth = 0;
                var expandingFlags = 0 /* None */;
                var overflow = false;
                var overrideNextErrorInfo = 0; // How many `reportRelationError` calls should be skipped in the elaboration pyramid
                var lastSkippedInfo;
                var incompatibleStack = [];
                var inPropertyCheck = false;
                ts.Debug.assert(relation !== identityRelation || !errorNode, "no error reporting in identity checking");
                var result = isRelatedTo(source, target, /*reportErrors*/ !!errorNode, headMessage);
                if (incompatibleStack.length) {
                    reportIncompatibleStack();
                }
                if (overflow) {
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.instant("checkTypes" /* CheckTypes */, "checkTypeRelatedTo_DepthLimit", { sourceId: source.id, targetId: target.id, depth: depth });
                    var diag = error(errorNode || currentNode, ts.Diagnostics.Excessive_stack_depth_comparing_types_0_and_1, typeToString(source), typeToString(target));
                    if (errorOutputContainer) {
                        (errorOutputContainer.errors || (errorOutputContainer.errors = [])).push(diag);
                    }
                }
                else if (errorInfo) {
                    if (containingMessageChain) {
                        var chain = containingMessageChain();
                        if (chain) {
                            ts.concatenateDiagnosticMessageChains(chain, errorInfo);
                            errorInfo = chain;
                        }
                    }
                    var relatedInformation = void 0;
                    // Check if we should issue an extra diagnostic to produce a quickfix for a slightly incorrect import statement
                    if (headMessage && errorNode && !result && source.symbol) {
                        var links = getSymbolLinks(source.symbol);
                        if (links.originatingImport && !ts.isImportCall(links.originatingImport)) {
                            var helpfulRetry = checkTypeRelatedTo(getTypeOfSymbol(links.target), target, relation, /*errorNode*/ undefined);
                            if (helpfulRetry) {
                                // Likely an incorrect import. Issue a helpful diagnostic to produce a quickfix to change the import
                                var diag_1 = ts.createDiagnosticForNode(links.originatingImport, ts.Diagnostics.Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cause_a_failure_at_runtime_Consider_using_a_default_import_or_import_require_here_instead);
                                relatedInformation = ts.append(relatedInformation, diag_1); // Cause the error to appear with the error that triggered it
                            }
                        }
                    }
                    var diag = ts.createDiagnosticForNodeFromMessageChain(errorNode, errorInfo, relatedInformation);
                    if (relatedInfo) {
                        ts.addRelatedInfo.apply(void 0, __spreadArray([diag], relatedInfo, false));
                    }
                    if (errorOutputContainer) {
                        (errorOutputContainer.errors || (errorOutputContainer.errors = [])).push(diag);
                    }
                    if (!errorOutputContainer || !errorOutputContainer.skipLogging) {
                        diagnostics.add(diag);
                    }
                }
                if (errorNode && errorOutputContainer && errorOutputContainer.skipLogging && result === 0 /* False */) {
                    ts.Debug.assert(!!errorOutputContainer.errors, "missed opportunity to interact with error.");
                }
                return result !== 0 /* False */;
                function resetErrorInfo(saved) {
                    errorInfo = saved.errorInfo;
                    lastSkippedInfo = saved.lastSkippedInfo;
                    incompatibleStack = saved.incompatibleStack;
                    overrideNextErrorInfo = saved.overrideNextErrorInfo;
                    relatedInfo = saved.relatedInfo;
                }
                function captureErrorCalculationState() {
                    return {
                        errorInfo: errorInfo,
                        lastSkippedInfo: lastSkippedInfo,
                        incompatibleStack: incompatibleStack.slice(),
                        overrideNextErrorInfo: overrideNextErrorInfo,
                        relatedInfo: !relatedInfo ? undefined : relatedInfo.slice()
                    };
                }
                function reportIncompatibleError(message, arg0, arg1, arg2, arg3) {
                    overrideNextErrorInfo++; // Suppress the next relation error
                    lastSkippedInfo = undefined; // Reset skipped info cache
                    incompatibleStack.push([message, arg0, arg1, arg2, arg3]);
                }
                function reportIncompatibleStack() {
                    var stack = incompatibleStack;
                    incompatibleStack = [];
                    var info = lastSkippedInfo;
                    lastSkippedInfo = undefined;
                    if (stack.length === 1) {
                        reportError.apply(void 0, stack[0]);
                        if (info) {
                            // Actually do the last relation error
                            reportRelationError.apply(void 0, __spreadArray([/*headMessage*/ undefined], info, false));
                        }
                        return;
                    }
                    // The first error will be the innermost, while the last will be the outermost - so by popping off the end,
                    // we can build from left to right
                    var path = "";
                    var secondaryRootErrors = [];
                    while (stack.length) {
                        var _a = stack.pop(), msg = _a[0], args = _a.slice(1);
                        switch (msg.code) {
                            case ts.Diagnostics.Types_of_property_0_are_incompatible.code: {
                                // Parenthesize a `new` if there is one
                                if (path.indexOf("new ") === 0) {
                                    path = "(" + path + ")";
                                }
                                var str = "" + args[0];
                                // If leading, just print back the arg (irrespective of if it's a valid identifier)
                                if (path.length === 0) {
                                    path = "" + str;
                                }
                                // Otherwise write a dotted name if possible
                                else if (ts.isIdentifierText(str, compilerOptions.target)) {
                                    path = path + "." + str;
                                }
                                // Failing that, check if the name is already a computed name
                                else if (str[0] === "[" && str[str.length - 1] === "]") {
                                    path = "" + path + str;
                                }
                                // And finally write out a computed name as a last resort
                                else {
                                    path = path + "[" + str + "]";
                                }
                                break;
                            }
                            case ts.Diagnostics.Call_signature_return_types_0_and_1_are_incompatible.code:
                            case ts.Diagnostics.Construct_signature_return_types_0_and_1_are_incompatible.code:
                            case ts.Diagnostics.Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code:
                            case ts.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code: {
                                if (path.length === 0) {
                                    // Don't flatten signature compatability errors at the start of a chain - instead prefer
                                    // to unify (the with no arguments bit is excessive for printback) and print them back
                                    var mappedMsg = msg;
                                    if (msg.code === ts.Diagnostics.Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code) {
                                        mappedMsg = ts.Diagnostics.Call_signature_return_types_0_and_1_are_incompatible;
                                    }
                                    else if (msg.code === ts.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code) {
                                        mappedMsg = ts.Diagnostics.Construct_signature_return_types_0_and_1_are_incompatible;
                                    }
                                    secondaryRootErrors.unshift([mappedMsg, args[0], args[1]]);
                                }
                                else {
                                    var prefix = (msg.code === ts.Diagnostics.Construct_signature_return_types_0_and_1_are_incompatible.code ||
                                        msg.code === ts.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code)
                                        ? "new "
                                        : "";
                                    var params = (msg.code === ts.Diagnostics.Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code ||
                                        msg.code === ts.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1.code)
                                        ? ""
                                        : "...";
                                    path = "" + prefix + path + "(" + params + ")";
                                }
                                break;
                            }
                            case ts.Diagnostics.Type_at_position_0_in_source_is_not_compatible_with_type_at_position_1_in_target.code: {
                                secondaryRootErrors.unshift([ts.Diagnostics.Type_at_position_0_in_source_is_not_compatible_with_type_at_position_1_in_target, args[0], args[1]]);
                                break;
                            }
                            case ts.Diagnostics.Type_at_positions_0_through_1_in_source_is_not_compatible_with_type_at_position_2_in_target.code: {
                                secondaryRootErrors.unshift([ts.Diagnostics.Type_at_positions_0_through_1_in_source_is_not_compatible_with_type_at_position_2_in_target, args[0], args[1], args[2]]);
                                break;
                            }
                            default:
                                return ts.Debug.fail("Unhandled Diagnostic: " + msg.code);
                        }
                    }
                    if (path) {
                        reportError(path[path.length - 1] === ")"
                            ? ts.Diagnostics.The_types_returned_by_0_are_incompatible_between_these_types
                            : ts.Diagnostics.The_types_of_0_are_incompatible_between_these_types, path);
                    }
                    else {
                        // Remove the innermost secondary error as it will duplicate the error already reported by `reportRelationError` on entry
                        secondaryRootErrors.shift();
                    }
                    for (var _i = 0, secondaryRootErrors_1 = secondaryRootErrors; _i < secondaryRootErrors_1.length; _i++) {
                        var _b = secondaryRootErrors_1[_i], msg = _b[0], args = _b.slice(1);
                        var originalValue = msg.elidedInCompatabilityPyramid;
                        msg.elidedInCompatabilityPyramid = false; // Temporarily override elision to ensure error is reported
                        reportError.apply(void 0, __spreadArray([msg], args, false));
                        msg.elidedInCompatabilityPyramid = originalValue;
                    }
                    if (info) {
                        // Actually do the last relation error
                        reportRelationError.apply(void 0, __spreadArray([/*headMessage*/ undefined], info, false));
                    }
                }
                function reportError(message, arg0, arg1, arg2, arg3) {
                    ts.Debug.assert(!!errorNode);
                    if (incompatibleStack.length)
                        reportIncompatibleStack();
                    if (message.elidedInCompatabilityPyramid)
                        return;
                    errorInfo = ts.chainDiagnosticMessages(errorInfo, message, arg0, arg1, arg2, arg3);
                }
                function associateRelatedInfo(info) {
                    ts.Debug.assert(!!errorInfo);
                    if (!relatedInfo) {
                        relatedInfo = [info];
                    }
                    else {
                        relatedInfo.push(info);
                    }
                }
                function reportRelationError(message, source, target) {
                    if (incompatibleStack.length)
                        reportIncompatibleStack();
                    var _a = getTypeNamesForErrorDisplay(source, target), sourceType = _a[0], targetType = _a[1];
                    var generalizedSource = source;
                    var generalizedSourceType = sourceType;
                    if (isLiteralType(source) && !typeCouldHaveTopLevelSingletonTypes(target)) {
                        generalizedSource = getBaseTypeOfLiteralType(source);
                        ts.Debug.assert(!isTypeAssignableTo(generalizedSource, target), "generalized source shouldn't be assignable");
                        generalizedSourceType = getTypeNameForErrorDisplay(generalizedSource);
                    }
                    if (target.flags & 262144 /* TypeParameter */) {
                        var constraint = getBaseConstraintOfType(target);
                        var needsOriginalSource = void 0;
                        if (constraint && (isTypeAssignableTo(generalizedSource, constraint) || (needsOriginalSource = isTypeAssignableTo(source, constraint)))) {
                            reportError(ts.Diagnostics._0_is_assignable_to_the_constraint_of_type_1_but_1_could_be_instantiated_with_a_different_subtype_of_constraint_2, needsOriginalSource ? sourceType : generalizedSourceType, targetType, typeToString(constraint));
                        }
                        else {
                            errorInfo = undefined;
                            reportError(ts.Diagnostics._0_could_be_instantiated_with_an_arbitrary_type_which_could_be_unrelated_to_1, targetType, generalizedSourceType);
                        }
                    }
                    if (!message) {
                        if (relation === comparableRelation) {
                            message = ts.Diagnostics.Type_0_is_not_comparable_to_type_1;
                        }
                        else if (sourceType === targetType) {
                            message = ts.Diagnostics.Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated;
                        }
                        else {
                            message = ts.Diagnostics.Type_0_is_not_assignable_to_type_1;
                        }
                    }
                    reportError(message, generalizedSourceType, targetType);
                }
                function tryElaborateErrorsForPrimitivesAndObjects(source, target) {
                    var sourceType = symbolValueDeclarationIsContextSensitive(source.symbol) ? typeToString(source, source.symbol.valueDeclaration) : typeToString(source);
                    var targetType = symbolValueDeclarationIsContextSensitive(target.symbol) ? typeToString(target, target.symbol.valueDeclaration) : typeToString(target);
                    if ((globalStringType === source && stringType === target) ||
                        (globalNumberType === source && numberType === target) ||
                        (globalBooleanType === source && booleanType === target) ||
                        (getGlobalESSymbolType(/*reportErrors*/ false) === source && esSymbolType === target)) {
                        reportError(ts.Diagnostics._0_is_a_primitive_but_1_is_a_wrapper_object_Prefer_using_0_when_possible, targetType, sourceType);
                    }
                }
                /**
                 * Try and elaborate array and tuple errors. Returns false
                 * if we have found an elaboration, or we should ignore
                 * any other elaborations when relating the `source` and
                 * `target` types.
                 */
                function tryElaborateArrayLikeErrors(source, target, reportErrors) {
                    /**
                     * The spec for elaboration is:
                     * - If the source is a readonly tuple and the target is a mutable array or tuple, elaborate on mutability and skip property elaborations.
                     * - If the source is a tuple then skip property elaborations if the target is an array or tuple.
                     * - If the source is a readonly array and the target is a mutable array or tuple, elaborate on mutability and skip property elaborations.
                     * - If the source an array then skip property elaborations if the target is a tuple.
                     */
                    if (isTupleType(source)) {
                        if (source.target.readonly && isMutableArrayOrTuple(target)) {
                            if (reportErrors) {
                                reportError(ts.Diagnostics.The_type_0_is_readonly_and_cannot_be_assigned_to_the_mutable_type_1, typeToString(source), typeToString(target));
                            }
                            return false;
                        }
                        return isTupleType(target) || isArrayType(target);
                    }
                    if (isReadonlyArrayType(source) && isMutableArrayOrTuple(target)) {
                        if (reportErrors) {
                            reportError(ts.Diagnostics.The_type_0_is_readonly_and_cannot_be_assigned_to_the_mutable_type_1, typeToString(source), typeToString(target));
                        }
                        return false;
                    }
                    if (isTupleType(target)) {
                        return isArrayType(source);
                    }
                    return true;
                }
                /**
                 * Compare two types and return
                 * * Ternary.True if they are related with no assumptions,
                 * * Ternary.Maybe if they are related with assumptions of other relationships, or
                 * * Ternary.False if they are not related.
                 */
                function isRelatedTo(originalSource, originalTarget, reportErrors, headMessage, intersectionState) {
                    if (reportErrors === void 0) {
                        reportErrors = false;
                    }
                    if (intersectionState === void 0) {
                        intersectionState = 0 /* None */;
                    }
                    // Before normalization: if `source` is type an object type, and `target` is primitive,
                    // skip all the checks we don't need and just return `isSimpleTypeRelatedTo` result
                    if (originalSource.flags & 524288 /* Object */ && originalTarget.flags & 131068 /* Primitive */) {
                        if (isSimpleTypeRelatedTo(originalSource, originalTarget, relation, reportErrors ? reportError : undefined)) {
                            return -1 /* True */;
                        }
                        reportErrorResults(originalSource, originalTarget, 0 /* False */, !!(ts.getObjectFlags(originalSource) & 2048 /* JsxAttributes */));
                        return 0 /* False */;
                    }
                    // Normalize the source and target types: Turn fresh literal types into regular literal types,
                    // turn deferred type references into regular type references, simplify indexed access and
                    // conditional types, and resolve substitution types to either the substitution (on the source
                    // side) or the type variable (on the target side).
                    var source = getNormalizedType(originalSource, /*writing*/ false);
                    var target = getNormalizedType(originalTarget, /*writing*/ true);
                    if (source === target)
                        return -1 /* True */;
                    if (relation === identityRelation) {
                        return isIdenticalTo(source, target);
                    }
                    // We fastpath comparing a type parameter to exactly its constraint, as this is _super_ common,
                    // and otherwise, for type parameters in large unions, causes us to need to compare the union to itself,
                    // as we break down the _target_ union first, _then_ get the source constraint - so for every
                    // member of the target, we attempt to find a match in the source. This avoids that in cases where
                    // the target is exactly the constraint.
                    if (source.flags & 262144 /* TypeParameter */ && getConstraintOfType(source) === target) {
                        return -1 /* True */;
                    }
                    // Try to see if we're relating something like `Foo` -> `Bar | null | undefined`.
                    // If so, reporting the `null` and `undefined` in the type is hardly useful.
                    // First, see if we're even relating an object type to a union.
                    // Then see if the target is stripped down to a single non-union type.
                    // Note
                    //  * We actually want to remove null and undefined naively here (rather than using getNonNullableType),
                    //    since we don't want to end up with a worse error like "`Foo` is not assignable to `NonNullable<T>`"
                    //    when dealing with generics.
                    //  * We also don't deal with primitive source types, since we already halt elaboration below.
                    if (target.flags & 1048576 /* Union */ && source.flags & 524288 /* Object */ &&
                        target.types.length <= 3 && maybeTypeOfKind(target, 98304 /* Nullable */)) {
                        var nullStrippedTarget = extractTypesOfKind(target, ~98304 /* Nullable */);
                        if (!(nullStrippedTarget.flags & (1048576 /* Union */ | 131072 /* Never */))) {
                            target = getNormalizedType(nullStrippedTarget, /*writing*/ true);
                        }
                        if (source === nullStrippedTarget)
                            return -1 /* True */;
                    }
                    if (relation === comparableRelation && !(target.flags & 131072 /* Never */) && isSimpleTypeRelatedTo(target, source, relation) ||
                        isSimpleTypeRelatedTo(source, target, relation, reportErrors ? reportError : undefined))
                        return -1 /* True */;
                    var isComparingJsxAttributes = !!(ts.getObjectFlags(source) & 2048 /* JsxAttributes */);
                    var isPerformingExcessPropertyChecks = !(intersectionState & 2 /* Target */) && (isObjectLiteralType(source) && ts.getObjectFlags(source) & 16384 /* FreshLiteral */);
                    if (isPerformingExcessPropertyChecks) {
                        if (hasExcessProperties(source, target, reportErrors)) {
                            if (reportErrors) {
                                reportRelationError(headMessage, source, originalTarget.aliasSymbol ? originalTarget : target);
                            }
                            return 0 /* False */;
                        }
                    }
                    var isPerformingCommonPropertyChecks = relation !== comparableRelation && !(intersectionState & 2 /* Target */) &&
                        source.flags & (131068 /* Primitive */ | 524288 /* Object */ | 2097152 /* Intersection */) && source !== globalObjectType &&
                        target.flags & (524288 /* Object */ | 2097152 /* Intersection */) && isWeakType(target) &&
                        (getPropertiesOfType(source).length > 0 || typeHasCallOrConstructSignatures(source));
                    if (isPerformingCommonPropertyChecks && !hasCommonProperties(source, target, isComparingJsxAttributes)) {
                        if (reportErrors) {
                            var sourceString = typeToString(originalSource.aliasSymbol ? originalSource : source);
                            var targetString = typeToString(originalTarget.aliasSymbol ? originalTarget : target);
                            var calls = getSignaturesOfType(source, 0 /* Call */);
                            var constructs = getSignaturesOfType(source, 1 /* Construct */);
                            if (calls.length > 0 && isRelatedTo(getReturnTypeOfSignature(calls[0]), target, /*reportErrors*/ false) ||
                                constructs.length > 0 && isRelatedTo(getReturnTypeOfSignature(constructs[0]), target, /*reportErrors*/ false)) {
                                reportError(ts.Diagnostics.Value_of_type_0_has_no_properties_in_common_with_type_1_Did_you_mean_to_call_it, sourceString, targetString);
                            }
                            else {
                                reportError(ts.Diagnostics.Type_0_has_no_properties_in_common_with_type_1, sourceString, targetString);
                            }
                        }
                        return 0 /* False */;
                    }
                    traceUnionsOrIntersectionsTooLarge(source, target);
                    var result = 0 /* False */;
                    var saveErrorInfo = captureErrorCalculationState();
                    // Note that these checks are specifically ordered to produce correct results. In particular,
                    // we need to deconstruct unions before intersections (because unions are always at the top),
                    // and we need to handle "each" relations before "some" relations for the same kind of type.
                    if (source.flags & 3145728 /* UnionOrIntersection */ || target.flags & 3145728 /* UnionOrIntersection */) {
                        result = getConstituentCount(source) * getConstituentCount(target) >= 4 ?
                            recursiveTypeRelatedTo(source, target, reportErrors, intersectionState | 8 /* UnionIntersectionCheck */) :
                            structuredTypeRelatedTo(source, target, reportErrors, intersectionState | 8 /* UnionIntersectionCheck */);
                    }
                    if (!result && !(source.flags & 1048576 /* Union */) && (source.flags & (469499904 /* StructuredOrInstantiable */) || target.flags & 469499904 /* StructuredOrInstantiable */)) {
                        if (result = recursiveTypeRelatedTo(source, target, reportErrors, intersectionState)) {
                            resetErrorInfo(saveErrorInfo);
                        }
                    }
                    if (!result && source.flags & (2097152 /* Intersection */ | 262144 /* TypeParameter */)) {
                        // The combined constraint of an intersection type is the intersection of the constraints of
                        // the constituents. When an intersection type contains instantiable types with union type
                        // constraints, there are situations where we need to examine the combined constraint. One is
                        // when the target is a union type. Another is when the intersection contains types belonging
                        // to one of the disjoint domains. For example, given type variables T and U, each with the
                        // constraint 'string | number', the combined constraint of 'T & U' is 'string | number' and
                        // we need to check this constraint against a union on the target side. Also, given a type
                        // variable V constrained to 'string | number', 'V & number' has a combined constraint of
                        // 'string & number | number & number' which reduces to just 'number'.
                        // This also handles type parameters, as a type parameter with a union constraint compared against a union
                        // needs to have its constraint hoisted into an intersection with said type parameter, this way
                        // the type param can be compared with itself in the target (with the influence of its constraint to match other parts)
                        // For example, if `T extends 1 | 2` and `U extends 2 | 3` and we compare `T & U` to `T & U & (1 | 2 | 3)`
                        var constraint = getEffectiveConstraintOfIntersection(source.flags & 2097152 /* Intersection */ ? source.types : [source], !!(target.flags & 1048576 /* Union */));
                        if (constraint && (source.flags & 2097152 /* Intersection */ || target.flags & 1048576 /* Union */)) {
                            if (everyType(constraint, function (c) { return c !== source; })) { // Skip comparison if expansion contains the source itself
                                // TODO: Stack errors so we get a pyramid for the "normal" comparison above, _and_ a second for this
                                if (result = isRelatedTo(constraint, target, /*reportErrors*/ false, /*headMessage*/ undefined, intersectionState)) {
                                    resetErrorInfo(saveErrorInfo);
                                }
                            }
                        }
                    }
                    // For certain combinations involving intersections and optional, excess, or mismatched properties we need
                    // an extra property check where the intersection is viewed as a single object. The following are motivating
                    // examples that all should be errors, but aren't without this extra property check:
                    //
                    //   let obj: { a: { x: string } } & { c: number } = { a: { x: 'hello', y: 2 }, c: 5 };  // Nested excess property
                    //
                    //   declare let wrong: { a: { y: string } };
                    //   let weak: { a?: { x?: number } } & { c?: string } = wrong;  // Nested weak object type
                    //
                    //   function foo<T extends object>(x: { a?: string }, y: T & { a: boolean }) {
                    //     x = y;  // Mismatched property in source intersection
                    //   }
                    //
                    // We suppress recursive intersection property checks because they can generate lots of work when relating
                    // recursive intersections that are structurally similar but not exactly identical. See #37854.
                    if (result && !inPropertyCheck && (target.flags & 2097152 /* Intersection */ && (isPerformingExcessPropertyChecks || isPerformingCommonPropertyChecks) ||
                        isNonGenericObjectType(target) && !isArrayType(target) && !isTupleType(target) && source.flags & 2097152 /* Intersection */ && getApparentType(source).flags & 3670016 /* StructuredType */ && !ts.some(source.types, function (t) { return !!(ts.getObjectFlags(t) & 524288 /* NonInferrableType */); }))) {
                        inPropertyCheck = true;
                        result &= recursiveTypeRelatedTo(source, target, reportErrors, 4 /* PropertyCheck */);
                        inPropertyCheck = false;
                    }
                    reportErrorResults(source, target, result, isComparingJsxAttributes);
                    return result;
                    function reportErrorResults(source, target, result, isComparingJsxAttributes) {
                        if (!result && reportErrors) {
                            var sourceHasBase = !!getSingleBaseForNonAugmentingSubtype(originalSource);
                            var targetHasBase = !!getSingleBaseForNonAugmentingSubtype(originalTarget);
                            source = (originalSource.aliasSymbol || sourceHasBase) ? originalSource : source;
                            target = (originalTarget.aliasSymbol || targetHasBase) ? originalTarget : target;
                            var maybeSuppress = overrideNextErrorInfo > 0;
                            if (maybeSuppress) {
                                overrideNextErrorInfo--;
                            }
                            if (source.flags & 524288 /* Object */ && target.flags & 524288 /* Object */) {
                                var currentError = errorInfo;
                                tryElaborateArrayLikeErrors(source, target, reportErrors);
                                if (errorInfo !== currentError) {
                                    maybeSuppress = !!errorInfo;
                                }
                            }
                            if (source.flags & 524288 /* Object */ && target.flags & 131068 /* Primitive */) {
                                tryElaborateErrorsForPrimitivesAndObjects(source, target);
                            }
                            else if (source.symbol && source.flags & 524288 /* Object */ && globalObjectType === source) {
                                reportError(ts.Diagnostics.The_Object_type_is_assignable_to_very_few_other_types_Did_you_mean_to_use_the_any_type_instead);
                            }
                            else if (isComparingJsxAttributes && target.flags & 2097152 /* Intersection */) {
                                var targetTypes = target.types;
                                var intrinsicAttributes = getJsxType(JsxNames.IntrinsicAttributes, errorNode);
                                var intrinsicClassAttributes = getJsxType(JsxNames.IntrinsicClassAttributes, errorNode);
                                if (intrinsicAttributes !== errorType && intrinsicClassAttributes !== errorType &&
                                    (ts.contains(targetTypes, intrinsicAttributes) || ts.contains(targetTypes, intrinsicClassAttributes))) {
                                    // do not report top error
                                    return result;
                                }
                            }
                            else {
                                errorInfo = elaborateNeverIntersection(errorInfo, originalTarget);
                            }
                            if (!headMessage && maybeSuppress) {
                                lastSkippedInfo = [source, target];
                                // Used by, eg, missing property checking to replace the top-level message with a more informative one
                                return result;
                            }
                            reportRelationError(headMessage, source, target);
                        }
                    }
                }
                function traceUnionsOrIntersectionsTooLarge(source, target) {
                    if (!ts.tracing) {
                        return;
                    }
                    if ((source.flags & 3145728 /* UnionOrIntersection */) && (target.flags & 3145728 /* UnionOrIntersection */)) {
                        var sourceUnionOrIntersection = source;
                        var targetUnionOrIntersection = target;
                        if (sourceUnionOrIntersection.objectFlags & targetUnionOrIntersection.objectFlags & 65536 /* PrimitiveUnion */) {
                            // There's a fast path for comparing primitive unions
                            return;
                        }
                        var sourceSize = sourceUnionOrIntersection.types.length;
                        var targetSize = targetUnionOrIntersection.types.length;
                        if (sourceSize * targetSize > 1E6) {
                            ts.tracing.instant("checkTypes" /* CheckTypes */, "traceUnionsOrIntersectionsTooLarge_DepthLimit", {
                                sourceId: source.id,
                                sourceSize: sourceSize,
                                targetId: target.id,
                                targetSize: targetSize,
                                pos: errorNode === null || errorNode === void 0 ? void 0 : errorNode.pos,
                                end: errorNode === null || errorNode === void 0 ? void 0 : errorNode.end
                            });
                        }
                    }
                }
                function isIdenticalTo(source, target) {
                    if (source.flags !== target.flags)
                        return 0 /* False */;
                    if (source.flags & 67358815 /* Singleton */)
                        return -1 /* True */;
                    traceUnionsOrIntersectionsTooLarge(source, target);
                    if (source.flags & 3145728 /* UnionOrIntersection */) {
                        var result_7 = eachTypeRelatedToSomeType(source, target);
                        if (result_7) {
                            result_7 &= eachTypeRelatedToSomeType(target, source);
                        }
                        return result_7;
                    }
                    return recursiveTypeRelatedTo(source, target, /*reportErrors*/ false, 0 /* None */);
                }
                function getTypeOfPropertyInTypes(types, name) {
                    var appendPropType = function (propTypes, type) {
                        var _a;
                        type = getApparentType(type);
                        var prop = type.flags & 3145728 /* UnionOrIntersection */ ? getPropertyOfUnionOrIntersectionType(type, name) : getPropertyOfObjectType(type, name);
                        var propType = prop && getTypeOfSymbol(prop) || ((_a = getApplicableIndexInfoForName(type, name)) === null || _a === void 0 ? void 0 : _a.type) || undefinedType;
                        return ts.append(propTypes, propType);
                    };
                    return getUnionType(ts.reduceLeft(types, appendPropType, /*initial*/ undefined) || ts.emptyArray);
                }
                function hasExcessProperties(source, target, reportErrors) {
                    var _a;
                    if (!isExcessPropertyCheckTarget(target) || !noImplicitAny && ts.getObjectFlags(target) & 8192 /* JSLiteral */) {
                        return false; // Disable excess property checks on JS literals to simulate having an implicit "index signature" - but only outside of noImplicitAny
                    }
                    var isComparingJsxAttributes = !!(ts.getObjectFlags(source) & 2048 /* JsxAttributes */);
                    if ((relation === assignableRelation || relation === comparableRelation) &&
                        (isTypeSubsetOf(globalObjectType, target) || (!isComparingJsxAttributes && isEmptyObjectType(target)))) {
                        return false;
                    }
                    var reducedTarget = target;
                    var checkTypes;
                    if (target.flags & 1048576 /* Union */) {
                        reducedTarget = findMatchingDiscriminantType(source, target, isRelatedTo) || filterPrimitivesIfContainsNonPrimitive(target);
                        checkTypes = reducedTarget.flags & 1048576 /* Union */ ? reducedTarget.types : [reducedTarget];
                    }
                    var _loop_18 = function (prop) {
                        if (shouldCheckAsExcessProperty(prop, source.symbol) && !isIgnoredJsxProperty(source, prop)) {
                            if (!isKnownProperty(reducedTarget, prop.escapedName, isComparingJsxAttributes)) {
                                if (reportErrors) {
                                    // Report error in terms of object types in the target as those are the only ones
                                    // we check in isKnownProperty.
                                    var errorTarget = filterType(reducedTarget, isExcessPropertyCheckTarget);
                                    // We know *exactly* where things went wrong when comparing the types.
                                    // Use this property as the error node as this will be more helpful in
                                    // reasoning about what went wrong.
                                    if (!errorNode)
                                        return { value: ts.Debug.fail() };
                                    if (ts.isJsxAttributes(errorNode) || ts.isJsxOpeningLikeElement(errorNode) || ts.isJsxOpeningLikeElement(errorNode.parent)) {
                                        // JsxAttributes has an object-literal flag and undergo same type-assignablity check as normal object-literal.
                                        // However, using an object-literal error message will be very confusing to the users so we give different a message.
                                        if (prop.valueDeclaration && ts.isJsxAttribute(prop.valueDeclaration) && ts.getSourceFileOfNode(errorNode) === ts.getSourceFileOfNode(prop.valueDeclaration.name)) {
                                            // Note that extraneous children (as in `<NoChild>extra</NoChild>`) don't pass this check,
                                            // since `children` is a SyntaxKind.PropertySignature instead of a SyntaxKind.JsxAttribute.
                                            errorNode = prop.valueDeclaration.name;
                                        }
                                        var propName = symbolToString(prop);
                                        var suggestionSymbol = getSuggestedSymbolForNonexistentJSXAttribute(propName, errorTarget);
                                        var suggestion = suggestionSymbol ? symbolToString(suggestionSymbol) : undefined;
                                        if (suggestion) {
                                            reportError(ts.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2, propName, typeToString(errorTarget), suggestion);
                                        }
                                        else {
                                            reportError(ts.Diagnostics.Property_0_does_not_exist_on_type_1, propName, typeToString(errorTarget));
                                        }
                                    }
                                    else {
                                        // use the property's value declaration if the property is assigned inside the literal itself
                                        var objectLiteralDeclaration_1 = ((_a = source.symbol) === null || _a === void 0 ? void 0 : _a.declarations) && ts.firstOrUndefined(source.symbol.declarations);
                                        var suggestion = void 0;
                                        if (prop.valueDeclaration && ts.findAncestor(prop.valueDeclaration, function (d) { return d === objectLiteralDeclaration_1; }) && ts.getSourceFileOfNode(objectLiteralDeclaration_1) === ts.getSourceFileOfNode(errorNode)) {
                                            var propDeclaration = prop.valueDeclaration;
                                            ts.Debug.assertNode(propDeclaration, ts.isObjectLiteralElementLike);
                                            errorNode = propDeclaration;
                                            var name = propDeclaration.name;
                                            if (ts.isIdentifier(name)) {
                                                suggestion = getSuggestionForNonexistentProperty(name, errorTarget);
                                            }
                                        }
                                        if (suggestion !== undefined) {
                                            reportError(ts.Diagnostics.Object_literal_may_only_specify_known_properties_but_0_does_not_exist_in_type_1_Did_you_mean_to_write_2, symbolToString(prop), typeToString(errorTarget), suggestion);
                                        }
                                        else {
                                            reportError(ts.Diagnostics.Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1, symbolToString(prop), typeToString(errorTarget));
                                        }
                                    }
                                }
                                return { value: true };
                            }
                            if (checkTypes && !isRelatedTo(getTypeOfSymbol(prop), getTypeOfPropertyInTypes(checkTypes, prop.escapedName), reportErrors)) {
                                if (reportErrors) {
                                    reportIncompatibleError(ts.Diagnostics.Types_of_property_0_are_incompatible, symbolToString(prop));
                                }
                                return { value: true };
                            }
                        }
                    };
                    for (var _i = 0, _b = getPropertiesOfType(source); _i < _b.length; _i++) {
                        var prop = _b[_i];
                        var state_5 = _loop_18(prop);
                        if (typeof state_5 === "object")
                            return state_5.value;
                    }
                    return false;
                }
                function shouldCheckAsExcessProperty(prop, container) {
                    return prop.valueDeclaration && container.valueDeclaration && prop.valueDeclaration.parent === container.valueDeclaration;
                }
                function eachTypeRelatedToSomeType(source, target) {
                    var result = -1 /* True */;
                    var sourceTypes = source.types;
                    for (var _i = 0, sourceTypes_1 = sourceTypes; _i < sourceTypes_1.length; _i++) {
                        var sourceType = sourceTypes_1[_i];
                        var related = typeRelatedToSomeType(sourceType, target, /*reportErrors*/ false);
                        if (!related) {
                            return 0 /* False */;
                        }
                        result &= related;
                    }
                    return result;
                }
                function typeRelatedToSomeType(source, target, reportErrors) {
                    var targetTypes = target.types;
                    if (target.flags & 1048576 /* Union */) {
                        if (containsType(targetTypes, source)) {
                            return -1 /* True */;
                        }
                        var match = getMatchingUnionConstituentForType(target, source);
                        if (match) {
                            var related = isRelatedTo(source, match, /*reportErrors*/ false);
                            if (related) {
                                return related;
                            }
                        }
                    }
                    for (var _i = 0, targetTypes_1 = targetTypes; _i < targetTypes_1.length; _i++) {
                        var type = targetTypes_1[_i];
                        var related = isRelatedTo(source, type, /*reportErrors*/ false);
                        if (related) {
                            return related;
                        }
                    }
                    if (reportErrors) {
                        var bestMatchingType = getBestMatchingType(source, target, isRelatedTo);
                        isRelatedTo(source, bestMatchingType || targetTypes[targetTypes.length - 1], /*reportErrors*/ true);
                    }
                    return 0 /* False */;
                }
                function typeRelatedToEachType(source, target, reportErrors, intersectionState) {
                    var result = -1 /* True */;
                    var targetTypes = target.types;
                    for (var _i = 0, targetTypes_2 = targetTypes; _i < targetTypes_2.length; _i++) {
                        var targetType = targetTypes_2[_i];
                        var related = isRelatedTo(source, targetType, reportErrors, /*headMessage*/ undefined, intersectionState);
                        if (!related) {
                            return 0 /* False */;
                        }
                        result &= related;
                    }
                    return result;
                }
                function someTypeRelatedToType(source, target, reportErrors, intersectionState) {
                    var sourceTypes = source.types;
                    if (source.flags & 1048576 /* Union */ && containsType(sourceTypes, target)) {
                        return -1 /* True */;
                    }
                    var len = sourceTypes.length;
                    for (var i = 0; i < len; i++) {
                        var related = isRelatedTo(sourceTypes[i], target, reportErrors && i === len - 1, /*headMessage*/ undefined, intersectionState);
                        if (related) {
                            return related;
                        }
                    }
                    return 0 /* False */;
                }
                function getUndefinedStrippedTargetIfNeeded(source, target) {
                    // As a builtin type, `undefined` is a very low type ID - making it almsot always first, making this a very fast check to see
                    // if we need to strip `undefined` from the target
                    if (source.flags & 1048576 /* Union */ && target.flags & 1048576 /* Union */ &&
                        !(source.types[0].flags & 32768 /* Undefined */) && target.types[0].flags & 32768 /* Undefined */) {
                        return extractTypesOfKind(target, ~32768 /* Undefined */);
                    }
                    return target;
                }
                function eachTypeRelatedToType(source, target, reportErrors, intersectionState) {
                    var result = -1 /* True */;
                    var sourceTypes = source.types;
                    // We strip `undefined` from the target if the `source` trivially doesn't contain it for our correspondence-checking fastpath
                    // since `undefined` is frequently added by optionality and would otherwise spoil a potentially useful correspondence
                    var undefinedStrippedTarget = getUndefinedStrippedTargetIfNeeded(source, target);
                    for (var i = 0; i < sourceTypes.length; i++) {
                        var sourceType = sourceTypes[i];
                        if (undefinedStrippedTarget.flags & 1048576 /* Union */ && sourceTypes.length >= undefinedStrippedTarget.types.length && sourceTypes.length % undefinedStrippedTarget.types.length === 0) {
                            // many unions are mappings of one another; in such cases, simply comparing members at the same index can shortcut the comparison
                            // such unions will have identical lengths, and their corresponding elements will match up. Another common scenario is where a large
                            // union has a union of objects intersected with it. In such cases, if the input was, eg `("a" | "b" | "c") & (string | boolean | {} | {whatever})`,
                            // the result will have the structure `"a" | "b" | "c" | "a" & {} | "b" & {} | "c" & {} | "a" & {whatever} | "b" & {whatever} | "c" & {whatever}`
                            // - the resulting union has a length which is a multiple of the original union, and the elements correspond modulo the length of the original union
                            var related_1 = isRelatedTo(sourceType, undefinedStrippedTarget.types[i % undefinedStrippedTarget.types.length], /*reportErrors*/ false, /*headMessage*/ undefined, intersectionState);
                            if (related_1) {
                                result &= related_1;
                                continue;
                            }
                        }
                        var related = isRelatedTo(sourceType, target, reportErrors, /*headMessage*/ undefined, intersectionState);
                        if (!related) {
                            return 0 /* False */;
                        }
                        result &= related;
                    }
                    return result;
                }
                function typeArgumentsRelatedTo(sources, targets, variances, reportErrors, intersectionState) {
                    if (sources === void 0) {
                        sources = ts.emptyArray;
                    }
                    if (targets === void 0) {
                        targets = ts.emptyArray;
                    }
                    if (variances === void 0) {
                        variances = ts.emptyArray;
                    }
                    if (sources.length !== targets.length && relation === identityRelation) {
                        return 0 /* False */;
                    }
                    var length = sources.length <= targets.length ? sources.length : targets.length;
                    var result = -1 /* True */;
                    for (var i = 0; i < length; i++) {
                        // When variance information isn't available we default to covariance. This happens
                        // in the process of computing variance information for recursive types and when
                        // comparing 'this' type arguments.
                        var varianceFlags = i < variances.length ? variances[i] : 1 /* Covariant */;
                        var variance = varianceFlags & 7 /* VarianceMask */;
                        // We ignore arguments for independent type parameters (because they're never witnessed).
                        if (variance !== 4 /* Independent */) {
                            var s = sources[i];
                            var t = targets[i];
                            var related = -1 /* True */;
                            if (varianceFlags & 8 /* Unmeasurable */) {
                                // Even an `Unmeasurable` variance works out without a structural check if the source and target are _identical_.
                                // We can't simply assume invariance, because `Unmeasurable` marks nonlinear relations, for example, a relation tained by
                                // the `-?` modifier in a mapped type (where, no matter how the inputs are related, the outputs still might not be)
                                related = relation === identityRelation ? isRelatedTo(s, t, /*reportErrors*/ false) : compareTypesIdentical(s, t);
                            }
                            else if (variance === 1 /* Covariant */) {
                                related = isRelatedTo(s, t, reportErrors, /*headMessage*/ undefined, intersectionState);
                            }
                            else if (variance === 2 /* Contravariant */) {
                                related = isRelatedTo(t, s, reportErrors, /*headMessage*/ undefined, intersectionState);
                            }
                            else if (variance === 3 /* Bivariant */) {
                                // In the bivariant case we first compare contravariantly without reporting
                                // errors. Then, if that doesn't succeed, we compare covariantly with error
                                // reporting. Thus, error elaboration will be based on the the covariant check,
                                // which is generally easier to reason about.
                                related = isRelatedTo(t, s, /*reportErrors*/ false);
                                if (!related) {
                                    related = isRelatedTo(s, t, reportErrors, /*headMessage*/ undefined, intersectionState);
                                }
                            }
                            else {
                                // In the invariant case we first compare covariantly, and only when that
                                // succeeds do we proceed to compare contravariantly. Thus, error elaboration
                                // will typically be based on the covariant check.
                                related = isRelatedTo(s, t, reportErrors, /*headMessage*/ undefined, intersectionState);
                                if (related) {
                                    related &= isRelatedTo(t, s, reportErrors, /*headMessage*/ undefined, intersectionState);
                                }
                            }
                            if (!related) {
                                return 0 /* False */;
                            }
                            result &= related;
                        }
                    }
                    return result;
                }
                // Determine if possibly recursive types are related. First, check if the result is already available in the global cache.
                // Second, check if we have already started a comparison of the given two types in which case we assume the result to be true.
                // Third, check if both types are part of deeply nested chains of generic type instantiations and if so assume the types are
                // equal and infinitely expanding. Fourth, if we have reached a depth of 100 nested comparisons, assume we have runaway recursion
                // and issue an error. Otherwise, actually compare the structure of the two types.
                function recursiveTypeRelatedTo(source, target, reportErrors, intersectionState) {
                    if (overflow) {
                        return 0 /* False */;
                    }
                    var id = getRelationKey(source, target, intersectionState | (inPropertyCheck ? 16 /* InPropertyCheck */ : 0), relation);
                    var entry = relation.get(id);
                    if (entry !== undefined) {
                        if (reportErrors && entry & 2 /* Failed */ && !(entry & 4 /* Reported */)) {
                            // We are elaborating errors and the cached result is an unreported failure. The result will be reported
                            // as a failure, and should be updated as a reported failure by the bottom of this function.
                        }
                        else {
                            if (outofbandVarianceMarkerHandler) {
                                // We're in the middle of variance checking - integrate any unmeasurable/unreliable flags from this cached component
                                var saved = entry & 24 /* ReportsMask */;
                                if (saved & 8 /* ReportsUnmeasurable */) {
                                    instantiateType(source, makeFunctionTypeMapper(reportUnmeasurableMarkers));
                                }
                                if (saved & 16 /* ReportsUnreliable */) {
                                    instantiateType(source, makeFunctionTypeMapper(reportUnreliableMarkers));
                                }
                            }
                            return entry & 1 /* Succeeded */ ? -1 /* True */ : 0 /* False */;
                        }
                    }
                    if (!maybeKeys) {
                        maybeKeys = [];
                        sourceStack = [];
                        targetStack = [];
                    }
                    else {
                        // generate a key where all type parameter id positions are replaced with unconstrained type parameter ids
                        // this isn't perfect - nested type references passed as type arguments will muck up the indexes and thus
                        // prevent finding matches- but it should hit up the common cases
                        var broadestEquivalentId = id.split(",").map(function (i) {
                            return i.replace(/-\d+/g, function (_match, offset) {
                                var index = ts.length(id.slice(0, offset).match(/[-=]/g) || undefined);
                                return "=" + index;
                            });
                        }).join(",");
                        for (var i = 0; i < maybeCount; i++) {
                            // If source and target are already being compared, consider them related with assumptions
                            if (id === maybeKeys[i] || broadestEquivalentId === maybeKeys[i]) {
                                return 3 /* Maybe */;
                            }
                        }
                        if (depth === 100) {
                            overflow = true;
                            return 0 /* False */;
                        }
                    }
                    var maybeStart = maybeCount;
                    maybeKeys[maybeCount] = id;
                    maybeCount++;
                    sourceStack[depth] = source;
                    targetStack[depth] = target;
                    depth++;
                    var saveExpandingFlags = expandingFlags;
                    if (!(expandingFlags & 1 /* Source */) && isDeeplyNestedType(source, sourceStack, depth))
                        expandingFlags |= 1 /* Source */;
                    if (!(expandingFlags & 2 /* Target */) && isDeeplyNestedType(target, targetStack, depth))
                        expandingFlags |= 2 /* Target */;
                    var originalHandler;
                    var propagatingVarianceFlags = 0;
                    if (outofbandVarianceMarkerHandler) {
                        originalHandler = outofbandVarianceMarkerHandler;
                        outofbandVarianceMarkerHandler = function (onlyUnreliable) {
                            propagatingVarianceFlags |= onlyUnreliable ? 16 /* ReportsUnreliable */ : 8 /* ReportsUnmeasurable */;
                            return originalHandler(onlyUnreliable);
                        };
                    }
                    if (expandingFlags === 3 /* Both */) {
                        ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.instant("checkTypes" /* CheckTypes */, "recursiveTypeRelatedTo_DepthLimit", {
                            sourceId: source.id,
                            sourceIdStack: sourceStack.map(function (t) { return t.id; }),
                            targetId: target.id,
                            targetIdStack: targetStack.map(function (t) { return t.id; }),
                            depth: depth,
                        });
                    }
                    var result = expandingFlags !== 3 /* Both */ ? structuredTypeRelatedTo(source, target, reportErrors, intersectionState) : 3 /* Maybe */;
                    if (outofbandVarianceMarkerHandler) {
                        outofbandVarianceMarkerHandler = originalHandler;
                    }
                    expandingFlags = saveExpandingFlags;
                    depth--;
                    if (result) {
                        if (result === -1 /* True */ || depth === 0) {
                            if (result === -1 /* True */ || result === 3 /* Maybe */) {
                                // If result is definitely true, record all maybe keys as having succeeded. Also, record Ternary.Maybe
                                // results as having succeeded once we reach depth 0, but never record Ternary.Unknown results.
                                for (var i = maybeStart; i < maybeCount; i++) {
                                    relation.set(maybeKeys[i], 1 /* Succeeded */ | propagatingVarianceFlags);
                                }
                            }
                            maybeCount = maybeStart;
                        }
                    }
                    else {
                        // A false result goes straight into global cache (when something is false under
                        // assumptions it will also be false without assumptions)
                        relation.set(id, (reportErrors ? 4 /* Reported */ : 0) | 2 /* Failed */ | propagatingVarianceFlags);
                        maybeCount = maybeStart;
                    }
                    return result;
                }
                function structuredTypeRelatedTo(source, target, reportErrors, intersectionState) {
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("checkTypes" /* CheckTypes */, "structuredTypeRelatedTo", { sourceId: source.id, targetId: target.id });
                    var result = structuredTypeRelatedToWorker(source, target, reportErrors, intersectionState);
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                    return result;
                }
                function structuredTypeRelatedToWorker(source, target, reportErrors, intersectionState) {
                    if (intersectionState & 4 /* PropertyCheck */) {
                        return propertiesRelatedTo(source, target, reportErrors, /*excludedProperties*/ undefined, 0 /* None */);
                    }
                    if (intersectionState & 8 /* UnionIntersectionCheck */) {
                        // Note that these checks are specifically ordered to produce correct results. In particular,
                        // we need to deconstruct unions before intersections (because unions are always at the top),
                        // and we need to handle "each" relations before "some" relations for the same kind of type.
                        if (source.flags & 1048576 /* Union */) {
                            return relation === comparableRelation ?
                                someTypeRelatedToType(source, target, reportErrors && !(source.flags & 131068 /* Primitive */), intersectionState & ~8 /* UnionIntersectionCheck */) :
                                eachTypeRelatedToType(source, target, reportErrors && !(source.flags & 131068 /* Primitive */), intersectionState & ~8 /* UnionIntersectionCheck */);
                        }
                        if (target.flags & 1048576 /* Union */) {
                            return typeRelatedToSomeType(getRegularTypeOfObjectLiteral(source), target, reportErrors && !(source.flags & 131068 /* Primitive */) && !(target.flags & 131068 /* Primitive */));
                        }
                        if (target.flags & 2097152 /* Intersection */) {
                            return typeRelatedToEachType(getRegularTypeOfObjectLiteral(source), target, reportErrors, 2 /* Target */);
                        }
                        // Source is an intersection. For the comparable relation, if the target is a primitive type we hoist the
                        // constraints of all non-primitive types in the source into a new intersection. We do this because the
                        // intersection may further constrain the constraints of the non-primitive types. For example, given a type
                        // parameter 'T extends 1 | 2', the intersection 'T & 1' should be reduced to '1' such that it doesn't
                        // appear to be comparable to '2'.
                        if (relation === comparableRelation && target.flags & 131068 /* Primitive */) {
                            var constraints = ts.sameMap(source.types, getBaseConstraintOrType);
                            if (constraints !== source.types) {
                                source = getIntersectionType(constraints);
                                if (!(source.flags & 2097152 /* Intersection */)) {
                                    return isRelatedTo(source, target, /*reportErrors*/ false);
                                }
                            }
                        }
                        // Check to see if any constituents of the intersection are immediately related to the target.
                        //
                        // Don't report errors though. Checking whether a constituent is related to the source is not actually
                        // useful and leads to some confusing error messages. Instead it is better to let the below checks
                        // take care of this, or to not elaborate at all. For instance,
                        //
                        //    - For an object type (such as 'C = A & B'), users are usually more interested in structural errors.
                        //
                        //    - For a union type (such as '(A | B) = (C & D)'), it's better to hold onto the whole intersection
                        //          than to report that 'D' is not assignable to 'A' or 'B'.
                        //
                        //    - For a primitive type or type parameter (such as 'number = A & B') there is no point in
                        //          breaking the intersection apart.
                        return someTypeRelatedToType(source, target, /*reportErrors*/ false, 1 /* Source */);
                    }
                    var flags = source.flags & target.flags;
                    if (relation === identityRelation && !(flags & 524288 /* Object */)) {
                        if (flags & 4194304 /* Index */) {
                            return isRelatedTo(source.type, target.type, /*reportErrors*/ false);
                        }
                        var result_8 = 0 /* False */;
                        if (flags & 8388608 /* IndexedAccess */) {
                            if (result_8 = isRelatedTo(source.objectType, target.objectType, /*reportErrors*/ false)) {
                                if (result_8 &= isRelatedTo(source.indexType, target.indexType, /*reportErrors*/ false)) {
                                    return result_8;
                                }
                            }
                        }
                        if (flags & 16777216 /* Conditional */) {
                            if (source.root.isDistributive === target.root.isDistributive) {
                                if (result_8 = isRelatedTo(source.checkType, target.checkType, /*reportErrors*/ false)) {
                                    if (result_8 &= isRelatedTo(source.extendsType, target.extendsType, /*reportErrors*/ false)) {
                                        if (result_8 &= isRelatedTo(getTrueTypeFromConditionalType(source), getTrueTypeFromConditionalType(target), /*reportErrors*/ false)) {
                                            if (result_8 &= isRelatedTo(getFalseTypeFromConditionalType(source), getFalseTypeFromConditionalType(target), /*reportErrors*/ false)) {
                                                return result_8;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        if (flags & 33554432 /* Substitution */) {
                            return isRelatedTo(source.substitute, target.substitute, /*reportErrors*/ false);
                        }
                        return 0 /* False */;
                    }
                    var result;
                    var originalErrorInfo;
                    var varianceCheckFailed = false;
                    var saveErrorInfo = captureErrorCalculationState();
                    // We limit alias variance probing to only object and conditional types since their alias behavior
                    // is more predictable than other, interned types, which may or may not have an alias depending on
                    // the order in which things were checked.
                    if (source.flags & (524288 /* Object */ | 16777216 /* Conditional */) && source.aliasSymbol &&
                        source.aliasTypeArguments && source.aliasSymbol === target.aliasSymbol &&
                        !(source.aliasTypeArgumentsContainsMarker || target.aliasTypeArgumentsContainsMarker)) {
                        var variances = getAliasVariances(source.aliasSymbol);
                        if (variances === ts.emptyArray) {
                            return 1 /* Unknown */;
                        }
                        var varianceResult = relateVariances(source.aliasTypeArguments, target.aliasTypeArguments, variances, intersectionState);
                        if (varianceResult !== undefined) {
                            return varianceResult;
                        }
                    }
                    // For a generic type T and a type U that is assignable to T, [...U] is assignable to T, U is assignable to readonly [...T],
                    // and U is assignable to [...T] when U is constrained to a mutable array or tuple type.
                    if (isSingleElementGenericTupleType(source) && !source.target.readonly && (result = isRelatedTo(getTypeArguments(source)[0], target)) ||
                        isSingleElementGenericTupleType(target) && (target.target.readonly || isMutableArrayOrTuple(getBaseConstraintOfType(source) || source)) && (result = isRelatedTo(source, getTypeArguments(target)[0]))) {
                        return result;
                    }
                    if (target.flags & 262144 /* TypeParameter */) {
                        // A source type { [P in Q]: X } is related to a target type T if keyof T is related to Q and X is related to T[Q].
                        if (ts.getObjectFlags(source) & 32 /* Mapped */ && !source.declaration.nameType && isRelatedTo(getIndexType(target), getConstraintTypeFromMappedType(source))) {
                            if (!(getMappedTypeModifiers(source) & 4 /* IncludeOptional */)) {
                                var templateType = getTemplateTypeFromMappedType(source);
                                var indexedAccessType = getIndexedAccessType(target, getTypeParameterFromMappedType(source));
                                if (result = isRelatedTo(templateType, indexedAccessType, reportErrors)) {
                                    return result;
                                }
                            }
                        }
                    }
                    else if (target.flags & 4194304 /* Index */) {
                        var targetType = target.type;
                        // A keyof S is related to a keyof T if T is related to S.
                        if (source.flags & 4194304 /* Index */) {
                            if (result = isRelatedTo(targetType, source.type, /*reportErrors*/ false)) {
                                return result;
                            }
                        }
                        if (isTupleType(targetType)) {
                            // An index type can have a tuple type target when the tuple type contains variadic elements.
                            // Check if the source is related to the known keys of the tuple type.
                            if (result = isRelatedTo(source, getKnownKeysOfTupleType(targetType), reportErrors)) {
                                return result;
                            }
                        }
                        else {
                            // A type S is assignable to keyof T if S is assignable to keyof C, where C is the
                            // simplified form of T or, if T doesn't simplify, the constraint of T.
                            var constraint = getSimplifiedTypeOrConstraint(targetType);
                            if (constraint) {
                                // We require Ternary.True here such that circular constraints don't cause
                                // false positives. For example, given 'T extends { [K in keyof T]: string }',
                                // 'keyof T' has itself as its constraint and produces a Ternary.Maybe when
                                // related to other types.
                                if (isRelatedTo(source, getIndexType(constraint, target.stringsOnly), reportErrors) === -1 /* True */) {
                                    return -1 /* True */;
                                }
                            }
                        }
                    }
                    else if (target.flags & 8388608 /* IndexedAccess */) {
                        if (source.flags & 8388608 /* IndexedAccess */) {
                            // Relate components directly before falling back to constraint relationships
                            // A type S[K] is related to a type T[J] if S is related to T and K is related to J.
                            if (result = isRelatedTo(source.objectType, target.objectType, reportErrors)) {
                                result &= isRelatedTo(source.indexType, target.indexType, reportErrors);
                            }
                            if (result) {
                                resetErrorInfo(saveErrorInfo);
                                return result;
                            }
                            if (reportErrors) {
                                originalErrorInfo = errorInfo;
                            }
                        }
                        // A type S is related to a type T[K] if S is related to C, where C is the base
                        // constraint of T[K] for writing.
                        if (relation === assignableRelation || relation === comparableRelation) {
                            var objectType = target.objectType;
                            var indexType = target.indexType;
                            var baseObjectType = getBaseConstraintOfType(objectType) || objectType;
                            var baseIndexType = getBaseConstraintOfType(indexType) || indexType;
                            if (!isGenericObjectType(baseObjectType) && !isGenericIndexType(baseIndexType)) {
                                var accessFlags = 4 /* Writing */ | (baseObjectType !== objectType ? 2 /* NoIndexSignatures */ : 0);
                                var constraint = getIndexedAccessTypeOrUndefined(baseObjectType, baseIndexType, accessFlags);
                                if (constraint) {
                                    if (reportErrors && originalErrorInfo) {
                                        // create a new chain for the constraint error
                                        resetErrorInfo(saveErrorInfo);
                                    }
                                    if (result = isRelatedTo(source, constraint, reportErrors)) {
                                        return result;
                                    }
                                    // prefer the shorter chain of the constraint comparison chain, and the direct comparison chain
                                    if (reportErrors && originalErrorInfo && errorInfo) {
                                        errorInfo = countMessageChainBreadth([originalErrorInfo]) <= countMessageChainBreadth([errorInfo]) ? originalErrorInfo : errorInfo;
                                    }
                                }
                            }
                        }
                        if (reportErrors) {
                            originalErrorInfo = undefined;
                        }
                    }
                    else if (isGenericMappedType(target) && !target.declaration.nameType) {
                        // A source type T is related to a target type { [P in X]: T[P] }
                        var template = getTemplateTypeFromMappedType(target);
                        var modifiers = getMappedTypeModifiers(target);
                        if (!(modifiers & 8 /* ExcludeOptional */)) {
                            if (template.flags & 8388608 /* IndexedAccess */ && template.objectType === source &&
                                template.indexType === getTypeParameterFromMappedType(target)) {
                                return -1 /* True */;
                            }
                            if (!isGenericMappedType(source)) {
                                var targetConstraint = getConstraintTypeFromMappedType(target);
                                var sourceKeys = getIndexType(source, /*stringsOnly*/ undefined, /*noIndexSignatures*/ true);
                                var includeOptional = modifiers & 4 /* IncludeOptional */;
                                var filteredByApplicability = includeOptional ? intersectTypes(targetConstraint, sourceKeys) : undefined;
                                // A source type T is related to a target type { [P in Q]: X } if Q is related to keyof T and T[Q] is related to X.
                                // A source type T is related to a target type { [P in Q]?: X } if some constituent Q' of Q is related to keyof T and T[Q'] is related to X.
                                if (includeOptional
                                    ? !(filteredByApplicability.flags & 131072 /* Never */)
                                    : isRelatedTo(targetConstraint, sourceKeys)) {
                                    var templateType = getTemplateTypeFromMappedType(target);
                                    var typeParameter = getTypeParameterFromMappedType(target);
                                    // Fastpath: When the template has the form Obj[P] where P is the mapped type parameter, directly compare `source` with `Obj`
                                    // to avoid creating the (potentially very large) number of new intermediate types made by manufacturing `source[P]`
                                    var nonNullComponent = extractTypesOfKind(templateType, ~98304 /* Nullable */);
                                    if (nonNullComponent.flags & 8388608 /* IndexedAccess */ && nonNullComponent.indexType === typeParameter) {
                                        if (result = isRelatedTo(source, nonNullComponent.objectType, reportErrors)) {
                                            return result;
                                        }
                                    }
                                    else {
                                        var indexingType = filteredByApplicability ? getIntersectionType([filteredByApplicability, typeParameter]) : typeParameter;
                                        var indexedAccessType = getIndexedAccessType(source, indexingType);
                                        if (result = isRelatedTo(indexedAccessType, templateType, reportErrors)) {
                                            return result;
                                        }
                                    }
                                }
                                originalErrorInfo = errorInfo;
                                resetErrorInfo(saveErrorInfo);
                            }
                        }
                    }
                    else if (target.flags & 16777216 /* Conditional */) {
                        var c = target;
                        // Check if the conditional is always true or always false but still deferred for distribution purposes
                        var skipTrue = !isTypeAssignableTo(getPermissiveInstantiation(c.checkType), getPermissiveInstantiation(c.extendsType));
                        var skipFalse = !skipTrue && isConditionalTypeAlwaysTrueDisregardingInferTypes(c);
                        // Instantiate with a replacement mapper if the conditional is distributive, replacing the check type with a clone of itself,
                        // this way {x: string | number, y: string | number} -> (T extends T ? { x: T, y: T } : never) appropriately _fails_ when
                        // T = string | number (since that will end up distributing and producing `{x: string, y: string} | {x: number, y: number}`,
                        // to which `{x: string | number, y: string | number}` isn't assignable)
                        var distributionMapper = void 0;
                        var checkVar = getActualTypeVariable(c.root.checkType);
                        if (c.root.isDistributive && checkVar.flags & 262144 /* TypeParameter */) {
                            var newParam = cloneTypeParameter(checkVar);
                            distributionMapper = prependTypeMapping(checkVar, newParam, c.mapper);
                            newParam.mapper = distributionMapper;
                        }
                        // TODO: Find a nice way to include potential conditional type breakdowns in error output, if they seem good (they usually don't)
                        var localResult = void 0;
                        if (skipTrue || (localResult = isRelatedTo(source, distributionMapper ? instantiateType(getTypeFromTypeNode(c.root.node.trueType), distributionMapper) : getTrueTypeFromConditionalType(c), /*reportErrors*/ false))) {
                            if (!skipFalse) {
                                localResult = (localResult || 3 /* Maybe */) & isRelatedTo(source, distributionMapper ? instantiateType(getTypeFromTypeNode(c.root.node.falseType), distributionMapper) : getFalseTypeFromConditionalType(c), /*reportErrors*/ false);
                            }
                        }
                        if (localResult) {
                            resetErrorInfo(saveErrorInfo);
                            return localResult;
                        }
                    }
                    else if (target.flags & 134217728 /* TemplateLiteral */) {
                        if (source.flags & 134217728 /* TemplateLiteral */) {
                            // Report unreliable variance for type variables referenced in template literal type placeholders.
                            // For example, `foo-${number}` is related to `foo-${string}` even though number isn't related to string.
                            instantiateType(source, makeFunctionTypeMapper(reportUnreliableMarkers));
                        }
                        var result_9 = inferTypesFromTemplateLiteralType(source, target);
                        if (result_9 && ts.every(result_9, function (r, i) { return isValidTypeForTemplateLiteralPlaceholder(r, target.types[i]); })) {
                            return -1 /* True */;
                        }
                    }
                    if (source.flags & 8650752 /* TypeVariable */) {
                        // IndexedAccess comparisons are handled above in the `target.flags & TypeFlage.IndexedAccess` branch
                        if (!(source.flags & 8388608 /* IndexedAccess */ && target.flags & 8388608 /* IndexedAccess */)) {
                            var constraint = getConstraintOfType(source);
                            if (!constraint || (source.flags & 262144 /* TypeParameter */ && constraint.flags & 1 /* Any */)) {
                                // A type variable with no constraint is not related to the non-primitive object type.
                                if (result = isRelatedTo(emptyObjectType, extractTypesOfKind(target, ~67108864 /* NonPrimitive */))) {
                                    resetErrorInfo(saveErrorInfo);
                                    return result;
                                }
                            }
                            // hi-speed no-this-instantiation check (less accurate, but avoids costly `this`-instantiation when the constraint will suffice), see #28231 for report on why this is needed
                            else if (result = isRelatedTo(constraint, target, /*reportErrors*/ false, /*headMessage*/ undefined, intersectionState)) {
                                resetErrorInfo(saveErrorInfo);
                                return result;
                            }
                            // slower, fuller, this-instantiated check (necessary when comparing raw `this` types from base classes), see `subclassWithPolymorphicThisIsAssignable.ts` test for example
                            else if (result = isRelatedTo(getTypeWithThisArgument(constraint, source), target, reportErrors && !(target.flags & source.flags & 262144 /* TypeParameter */), /*headMessage*/ undefined, intersectionState)) {
                                resetErrorInfo(saveErrorInfo);
                                return result;
                            }
                        }
                    }
                    else if (source.flags & 4194304 /* Index */) {
                        if (result = isRelatedTo(keyofConstraintType, target, reportErrors)) {
                            resetErrorInfo(saveErrorInfo);
                            return result;
                        }
                    }
                    else if (source.flags & 134217728 /* TemplateLiteral */) {
                        if (!(target.flags & 134217728 /* TemplateLiteral */)) {
                            var baseConstraint = getBaseConstraintOfType(source);
                            var constraint = baseConstraint && baseConstraint !== source ? baseConstraint : stringType;
                            if (result = isRelatedTo(constraint, target, reportErrors)) {
                                resetErrorInfo(saveErrorInfo);
                                return result;
                            }
                        }
                    }
                    else if (source.flags & 268435456 /* StringMapping */) {
                        if (target.flags & 268435456 /* StringMapping */ && source.symbol === target.symbol) {
                            if (result = isRelatedTo(source.type, target.type, reportErrors)) {
                                resetErrorInfo(saveErrorInfo);
                                return result;
                            }
                        }
                        else {
                            var constraint = getBaseConstraintOfType(source);
                            if (constraint && (result = isRelatedTo(constraint, target, reportErrors))) {
                                resetErrorInfo(saveErrorInfo);
                                return result;
                            }
                        }
                    }
                    else if (source.flags & 16777216 /* Conditional */) {
                        if (target.flags & 16777216 /* Conditional */) {
                            // Two conditional types 'T1 extends U1 ? X1 : Y1' and 'T2 extends U2 ? X2 : Y2' are related if
                            // one of T1 and T2 is related to the other, U1 and U2 are identical types, X1 is related to X2,
                            // and Y1 is related to Y2.
                            var sourceParams = source.root.inferTypeParameters;
                            var sourceExtends = source.extendsType;
                            var mapper = void 0;
                            if (sourceParams) {
                                // If the source has infer type parameters, we instantiate them in the context of the target
                                var ctx = createInferenceContext(sourceParams, /*signature*/ undefined, 0 /* None */, isRelatedTo);
                                inferTypes(ctx.inferences, target.extendsType, sourceExtends, 512 /* NoConstraints */ | 1024 /* AlwaysStrict */);
                                sourceExtends = instantiateType(sourceExtends, ctx.mapper);
                                mapper = ctx.mapper;
                            }
                            if (isTypeIdenticalTo(sourceExtends, target.extendsType) &&
                                (isRelatedTo(source.checkType, target.checkType) || isRelatedTo(target.checkType, source.checkType))) {
                                if (result = isRelatedTo(instantiateType(getTrueTypeFromConditionalType(source), mapper), getTrueTypeFromConditionalType(target), reportErrors)) {
                                    result &= isRelatedTo(getFalseTypeFromConditionalType(source), getFalseTypeFromConditionalType(target), reportErrors);
                                }
                                if (result) {
                                    resetErrorInfo(saveErrorInfo);
                                    return result;
                                }
                            }
                        }
                        else {
                            // conditionals aren't related to one another via distributive constraint as it is much too inaccurate and allows way
                            // more assignments than are desirable (since it maps the source check type to its constraint, it loses information)
                            var distributiveConstraint = getConstraintOfDistributiveConditionalType(source);
                            if (distributiveConstraint) {
                                if (result = isRelatedTo(distributiveConstraint, target, reportErrors)) {
                                    resetErrorInfo(saveErrorInfo);
                                    return result;
                                }
                            }
                        }
                        // conditionals _can_ be related to one another via normal constraint, as, eg, `A extends B ? O : never` should be assignable to `O`
                        // when `O` is a conditional (`never` is trivially aissgnable to `O`, as is `O`!).
                        var defaultConstraint = getDefaultConstraintOfConditionalType(source);
                        if (defaultConstraint) {
                            if (result = isRelatedTo(defaultConstraint, target, reportErrors)) {
                                resetErrorInfo(saveErrorInfo);
                                return result;
                            }
                        }
                    }
                    else {
                        // An empty object type is related to any mapped type that includes a '?' modifier.
                        if (relation !== subtypeRelation && relation !== strictSubtypeRelation && isPartialMappedType(target) && isEmptyObjectType(source)) {
                            return -1 /* True */;
                        }
                        if (isGenericMappedType(target)) {
                            if (isGenericMappedType(source)) {
                                if (result = mappedTypeRelatedTo(source, target, reportErrors)) {
                                    resetErrorInfo(saveErrorInfo);
                                    return result;
                                }
                            }
                            return 0 /* False */;
                        }
                        var sourceIsPrimitive = !!(source.flags & 131068 /* Primitive */);
                        if (relation !== identityRelation) {
                            source = getApparentType(source);
                        }
                        else if (isGenericMappedType(source)) {
                            return 0 /* False */;
                        }
                        if (ts.getObjectFlags(source) & 4 /* Reference */ && ts.getObjectFlags(target) & 4 /* Reference */ && source.target === target.target &&
                            !isTupleType(source) && !(ts.getObjectFlags(source) & 4096 /* MarkerType */ || ts.getObjectFlags(target) & 4096 /* MarkerType */)) {
                            // We have type references to the same generic type, and the type references are not marker
                            // type references (which are intended by be compared structurally). Obtain the variance
                            // information for the type parameters and relate the type arguments accordingly.
                            var variances = getVariances(source.target);
                            // We return Ternary.Maybe for a recursive invocation of getVariances (signalled by emptyArray). This
                            // effectively means we measure variance only from type parameter occurrences that aren't nested in
                            // recursive instantiations of the generic type.
                            if (variances === ts.emptyArray) {
                                return 1 /* Unknown */;
                            }
                            var varianceResult = relateVariances(getTypeArguments(source), getTypeArguments(target), variances, intersectionState);
                            if (varianceResult !== undefined) {
                                return varianceResult;
                            }
                        }
                        else if (isReadonlyArrayType(target) ? isArrayType(source) || isTupleType(source) : isArrayType(target) && isTupleType(source) && !source.target.readonly) {
                            if (relation !== identityRelation) {
                                return isRelatedTo(getIndexTypeOfType(source, numberType) || anyType, getIndexTypeOfType(target, numberType) || anyType, reportErrors);
                            }
                            else {
                                // By flags alone, we know that the `target` is a readonly array while the source is a normal array or tuple
                                // or `target` is an array and source is a tuple - in both cases the types cannot be identical, by construction
                                return 0 /* False */;
                            }
                        }
                        // Consider a fresh empty object literal type "closed" under the subtype relationship - this way `{} <- {[idx: string]: any} <- fresh({})`
                        // and not `{} <- fresh({}) <- {[idx: string]: any}`
                        else if ((relation === subtypeRelation || relation === strictSubtypeRelation) && isEmptyObjectType(target) && ts.getObjectFlags(target) & 16384 /* FreshLiteral */ && !isEmptyObjectType(source)) {
                            return 0 /* False */;
                        }
                        // Even if relationship doesn't hold for unions, intersections, or generic type references,
                        // it may hold in a structural comparison.
                        // In a check of the form X = A & B, we will have previously checked if A relates to X or B relates
                        // to X. Failing both of those we want to check if the aggregation of A and B's members structurally
                        // relates to X. Thus, we include intersection types on the source side here.
                        if (source.flags & (524288 /* Object */ | 2097152 /* Intersection */) && target.flags & 524288 /* Object */) {
                            // Report structural errors only if we haven't reported any errors yet
                            var reportStructuralErrors = reportErrors && errorInfo === saveErrorInfo.errorInfo && !sourceIsPrimitive;
                            result = propertiesRelatedTo(source, target, reportStructuralErrors, /*excludedProperties*/ undefined, intersectionState);
                            if (result) {
                                result &= signaturesRelatedTo(source, target, 0 /* Call */, reportStructuralErrors);
                                if (result) {
                                    result &= signaturesRelatedTo(source, target, 1 /* Construct */, reportStructuralErrors);
                                    if (result) {
                                        result &= indexSignaturesRelatedTo(source, target, sourceIsPrimitive, reportStructuralErrors, intersectionState);
                                    }
                                }
                            }
                            if (varianceCheckFailed && result) {
                                errorInfo = originalErrorInfo || errorInfo || saveErrorInfo.errorInfo; // Use variance error (there is no structural one) and return false
                            }
                            else if (result) {
                                return result;
                            }
                        }
                        // If S is an object type and T is a discriminated union, S may be related to T if
                        // there exists a constituent of T for every combination of the discriminants of S
                        // with respect to T. We do not report errors here, as we will use the existing
                        // error result from checking each constituent of the union.
                        if (source.flags & (524288 /* Object */ | 2097152 /* Intersection */) && target.flags & 1048576 /* Union */) {
                            var objectOnlyTarget = extractTypesOfKind(target, 524288 /* Object */ | 2097152 /* Intersection */ | 33554432 /* Substitution */);
                            if (objectOnlyTarget.flags & 1048576 /* Union */) {
                                var result_10 = typeRelatedToDiscriminatedType(source, objectOnlyTarget);
                                if (result_10) {
                                    return result_10;
                                }
                            }
                        }
                    }
                    return 0 /* False */;
                    function countMessageChainBreadth(info) {
                        if (!info)
                            return 0;
                        return ts.reduceLeft(info, function (value, chain) { return value + 1 + countMessageChainBreadth(chain.next); }, 0);
                    }
                    function relateVariances(sourceTypeArguments, targetTypeArguments, variances, intersectionState) {
                        if (result = typeArgumentsRelatedTo(sourceTypeArguments, targetTypeArguments, variances, reportErrors, intersectionState)) {
                            return result;
                        }
                        if (ts.some(variances, function (v) { return !!(v & 24 /* AllowsStructuralFallback */); })) {
                            // If some type parameter was `Unmeasurable` or `Unreliable`, and we couldn't pass by assuming it was identical, then we
                            // have to allow a structural fallback check
                            // We elide the variance-based error elaborations, since those might not be too helpful, since we'll potentially
                            // be assuming identity of the type parameter.
                            originalErrorInfo = undefined;
                            resetErrorInfo(saveErrorInfo);
                            return undefined;
                        }
                        var allowStructuralFallback = targetTypeArguments && hasCovariantVoidArgument(targetTypeArguments, variances);
                        varianceCheckFailed = !allowStructuralFallback;
                        // The type arguments did not relate appropriately, but it may be because we have no variance
                        // information (in which case typeArgumentsRelatedTo defaulted to covariance for all type
                        // arguments). It might also be the case that the target type has a 'void' type argument for
                        // a covariant type parameter that is only used in return positions within the generic type
                        // (in which case any type argument is permitted on the source side). In those cases we proceed
                        // with a structural comparison. Otherwise, we know for certain the instantiations aren't
                        // related and we can return here.
                        if (variances !== ts.emptyArray && !allowStructuralFallback) {
                            // In some cases generic types that are covariant in regular type checking mode become
                            // invariant in --strictFunctionTypes mode because one or more type parameters are used in
                            // both co- and contravariant positions. In order to make it easier to diagnose *why* such
                            // types are invariant, if any of the type parameters are invariant we reset the reported
                            // errors and instead force a structural comparison (which will include elaborations that
                            // reveal the reason).
                            // We can switch on `reportErrors` here, since varianceCheckFailed guarantees we return `False`,
                            // we can return `False` early here to skip calculating the structural error message we don't need.
                            if (varianceCheckFailed && !(reportErrors && ts.some(variances, function (v) { return (v & 7 /* VarianceMask */) === 0 /* Invariant */; }))) {
                                return 0 /* False */;
                            }
                            // We remember the original error information so we can restore it in case the structural
                            // comparison unexpectedly succeeds. This can happen when the structural comparison result
                            // is a Ternary.Maybe for example caused by the recursion depth limiter.
                            originalErrorInfo = errorInfo;
                            resetErrorInfo(saveErrorInfo);
                        }
                    }
                }
                function reportUnmeasurableMarkers(p) {
                    if (outofbandVarianceMarkerHandler && (p === markerSuperType || p === markerSubType || p === markerOtherType)) {
                        outofbandVarianceMarkerHandler(/*onlyUnreliable*/ false);
                    }
                    return p;
                }
                function reportUnreliableMarkers(p) {
                    if (outofbandVarianceMarkerHandler && (p === markerSuperType || p === markerSubType || p === markerOtherType)) {
                        outofbandVarianceMarkerHandler(/*onlyUnreliable*/ true);
                    }
                    return p;
                }
                // A type [P in S]: X is related to a type [Q in T]: Y if T is related to S and X' is
                // related to Y, where X' is an instantiation of X in which P is replaced with Q. Notice
                // that S and T are contra-variant whereas X and Y are co-variant.
                function mappedTypeRelatedTo(source, target, reportErrors) {
                    var modifiersRelated = relation === comparableRelation || (relation === identityRelation ? getMappedTypeModifiers(source) === getMappedTypeModifiers(target) :
                        getCombinedMappedTypeOptionality(source) <= getCombinedMappedTypeOptionality(target));
                    if (modifiersRelated) {
                        var result_11;
                        var targetConstraint = getConstraintTypeFromMappedType(target);
                        var sourceConstraint = instantiateType(getConstraintTypeFromMappedType(source), makeFunctionTypeMapper(getCombinedMappedTypeOptionality(source) < 0 ? reportUnmeasurableMarkers : reportUnreliableMarkers));
                        if (result_11 = isRelatedTo(targetConstraint, sourceConstraint, reportErrors)) {
                            var mapper = createTypeMapper([getTypeParameterFromMappedType(source)], [getTypeParameterFromMappedType(target)]);
                            if (instantiateType(getNameTypeFromMappedType(source), mapper) === instantiateType(getNameTypeFromMappedType(target), mapper)) {
                                return result_11 & isRelatedTo(instantiateType(getTemplateTypeFromMappedType(source), mapper), getTemplateTypeFromMappedType(target), reportErrors);
                            }
                        }
                    }
                    return 0 /* False */;
                }
                function typeRelatedToDiscriminatedType(source, target) {
                    // 1. Generate the combinations of discriminant properties & types 'source' can satisfy.
                    //    a. If the number of combinations is above a set limit, the comparison is too complex.
                    // 2. Filter 'target' to the subset of types whose discriminants exist in the matrix.
                    //    a. If 'target' does not satisfy all discriminants in the matrix, 'source' is not related.
                    // 3. For each type in the filtered 'target', determine if all non-discriminant properties of
                    //    'target' are related to a property in 'source'.
                    //
                    // NOTE: See ~/tests/cases/conformance/types/typeRelationships/assignmentCompatibility/assignmentCompatWithDiscriminatedUnion.ts
                    //       for examples.
                    var sourceProperties = getPropertiesOfType(source);
                    var sourcePropertiesFiltered = findDiscriminantProperties(sourceProperties, target);
                    if (!sourcePropertiesFiltered)
                        return 0 /* False */;
                    // Though we could compute the number of combinations as we generate
                    // the matrix, this would incur additional memory overhead due to
                    // array allocations. To reduce this overhead, we first compute
                    // the number of combinations to ensure we will not surpass our
                    // fixed limit before incurring the cost of any allocations:
                    var numCombinations = 1;
                    for (var _i = 0, sourcePropertiesFiltered_1 = sourcePropertiesFiltered; _i < sourcePropertiesFiltered_1.length; _i++) {
                        var sourceProperty = sourcePropertiesFiltered_1[_i];
                        numCombinations *= countTypes(getNonMissingTypeOfSymbol(sourceProperty));
                        if (numCombinations > 25) {
                            // We've reached the complexity limit.
                            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.instant("checkTypes" /* CheckTypes */, "typeRelatedToDiscriminatedType_DepthLimit", { sourceId: source.id, targetId: target.id, numCombinations: numCombinations });
                            return 0 /* False */;
                        }
                    }
                    // Compute the set of types for each discriminant property.
                    var sourceDiscriminantTypes = new Array(sourcePropertiesFiltered.length);
                    var excludedProperties = new ts.Set();
                    for (var i = 0; i < sourcePropertiesFiltered.length; i++) {
                        var sourceProperty = sourcePropertiesFiltered[i];
                        var sourcePropertyType = getNonMissingTypeOfSymbol(sourceProperty);
                        sourceDiscriminantTypes[i] = sourcePropertyType.flags & 1048576 /* Union */
                            ? sourcePropertyType.types
                            : [sourcePropertyType];
                        excludedProperties.add(sourceProperty.escapedName);
                    }
                    // Match each combination of the cartesian product of discriminant properties to one or more
                    // constituents of 'target'. If any combination does not have a match then 'source' is not relatable.
                    var discriminantCombinations = ts.cartesianProduct(sourceDiscriminantTypes);
                    var matchingTypes = [];
                    var _loop_19 = function (combination) {
                        var hasMatch = false;
                        outer: for (var _c = 0, _d = target.types; _c < _d.length; _c++) {
                            var type = _d[_c];
                            var _loop_20 = function (i) {
                                var sourceProperty = sourcePropertiesFiltered[i];
                                var targetProperty = getPropertyOfType(type, sourceProperty.escapedName);
                                if (!targetProperty)
                                    return "continue-outer";
                                if (sourceProperty === targetProperty)
                                    return "continue";
                                // We compare the source property to the target in the context of a single discriminant type.
                                var related = propertyRelatedTo(source, target, sourceProperty, targetProperty, function (_) { return combination[i]; }, /*reportErrors*/ false, 0 /* None */, /*skipOptional*/ strictNullChecks || relation === comparableRelation);
                                // If the target property could not be found, or if the properties were not related,
                                // then this constituent is not a match.
                                if (!related) {
                                    return "continue-outer";
                                }
                            };
                            for (var i = 0; i < sourcePropertiesFiltered.length; i++) {
                                var state_7 = _loop_20(i);
                                switch (state_7) {
                                    case "continue-outer": continue outer;
                                }
                            }
                            ts.pushIfUnique(matchingTypes, type, ts.equateValues);
                            hasMatch = true;
                        }
                        if (!hasMatch) {
                            return { value: 0 /* False */ };
                        }
                    };
                    for (var _a = 0, discriminantCombinations_1 = discriminantCombinations; _a < discriminantCombinations_1.length; _a++) {
                        var combination = discriminantCombinations_1[_a];
                        var state_6 = _loop_19(combination);
                        if (typeof state_6 === "object")
                            return state_6.value;
                    }
                    // Compare the remaining non-discriminant properties of each match.
                    var result = -1 /* True */;
                    for (var _b = 0, matchingTypes_1 = matchingTypes; _b < matchingTypes_1.length; _b++) {
                        var type = matchingTypes_1[_b];
                        result &= propertiesRelatedTo(source, type, /*reportErrors*/ false, excludedProperties, 0 /* None */);
                        if (result) {
                            result &= signaturesRelatedTo(source, type, 0 /* Call */, /*reportStructuralErrors*/ false);
                            if (result) {
                                result &= signaturesRelatedTo(source, type, 1 /* Construct */, /*reportStructuralErrors*/ false);
                                if (result && !(isTupleType(source) && isTupleType(type))) {
                                    // Comparing numeric index types when both `source` and `type` are tuples is unnecessary as the
                                    // element types should be sufficiently covered by `propertiesRelatedTo`. It also causes problems
                                    // with index type assignability as the types for the excluded discriminants are still included
                                    // in the index type.
                                    result &= indexSignaturesRelatedTo(source, type, /*sourceIsPrimitive*/ false, /*reportStructuralErrors*/ false, 0 /* None */);
                                }
                            }
                        }
                        if (!result) {
                            return result;
                        }
                    }
                    return result;
                }
                function excludeProperties(properties, excludedProperties) {
                    if (!excludedProperties || properties.length === 0)
                        return properties;
                    var result;
                    for (var i = 0; i < properties.length; i++) {
                        if (!excludedProperties.has(properties[i].escapedName)) {
                            if (result) {
                                result.push(properties[i]);
                            }
                        }
                        else if (!result) {
                            result = properties.slice(0, i);
                        }
                    }
                    return result || properties;
                }
                function isPropertySymbolTypeRelated(sourceProp, targetProp, getTypeOfSourceProperty, reportErrors, intersectionState) {
                    var targetIsOptional = strictNullChecks && !!(ts.getCheckFlags(targetProp) & 48 /* Partial */);
                    var effectiveTarget = addOptionality(getNonMissingTypeOfSymbol(targetProp), /*isProperty*/ false, targetIsOptional);
                    var effectiveSource = getTypeOfSourceProperty(sourceProp);
                    return isRelatedTo(effectiveSource, effectiveTarget, reportErrors, /*headMessage*/ undefined, intersectionState);
                }
                function propertyRelatedTo(source, target, sourceProp, targetProp, getTypeOfSourceProperty, reportErrors, intersectionState, skipOptional) {
                    var sourcePropFlags = ts.getDeclarationModifierFlagsFromSymbol(sourceProp);
                    var targetPropFlags = ts.getDeclarationModifierFlagsFromSymbol(targetProp);
                    if (sourcePropFlags & 8 /* Private */ || targetPropFlags & 8 /* Private */) {
                        if (sourceProp.valueDeclaration !== targetProp.valueDeclaration) {
                            if (reportErrors) {
                                if (sourcePropFlags & 8 /* Private */ && targetPropFlags & 8 /* Private */) {
                                    reportError(ts.Diagnostics.Types_have_separate_declarations_of_a_private_property_0, symbolToString(targetProp));
                                }
                                else {
                                    reportError(ts.Diagnostics.Property_0_is_private_in_type_1_but_not_in_type_2, symbolToString(targetProp), typeToString(sourcePropFlags & 8 /* Private */ ? source : target), typeToString(sourcePropFlags & 8 /* Private */ ? target : source));
                                }
                            }
                            return 0 /* False */;
                        }
                    }
                    else if (targetPropFlags & 16 /* Protected */) {
                        if (!isValidOverrideOf(sourceProp, targetProp)) {
                            if (reportErrors) {
                                reportError(ts.Diagnostics.Property_0_is_protected_but_type_1_is_not_a_class_derived_from_2, symbolToString(targetProp), typeToString(getDeclaringClass(sourceProp) || source), typeToString(getDeclaringClass(targetProp) || target));
                            }
                            return 0 /* False */;
                        }
                    }
                    else if (sourcePropFlags & 16 /* Protected */) {
                        if (reportErrors) {
                            reportError(ts.Diagnostics.Property_0_is_protected_in_type_1_but_public_in_type_2, symbolToString(targetProp), typeToString(source), typeToString(target));
                        }
                        return 0 /* False */;
                    }
                    // If the target comes from a partial union prop, allow `undefined` in the target type
                    var related = isPropertySymbolTypeRelated(sourceProp, targetProp, getTypeOfSourceProperty, reportErrors, intersectionState);
                    if (!related) {
                        if (reportErrors) {
                            reportIncompatibleError(ts.Diagnostics.Types_of_property_0_are_incompatible, symbolToString(targetProp));
                        }
                        return 0 /* False */;
                    }
                    // When checking for comparability, be more lenient with optional properties.
                    if (!skipOptional && sourceProp.flags & 16777216 /* Optional */ && !(targetProp.flags & 16777216 /* Optional */)) {
                        // TypeScript 1.0 spec (April 2014): 3.8.3
                        // S is a subtype of a type T, and T is a supertype of S if ...
                        // S' and T are object types and, for each member M in T..
                        // M is a property and S' contains a property N where
                        // if M is a required property, N is also a required property
                        // (M - property in T)
                        // (N - property in S)
                        if (reportErrors) {
                            reportError(ts.Diagnostics.Property_0_is_optional_in_type_1_but_required_in_type_2, symbolToString(targetProp), typeToString(source), typeToString(target));
                        }
                        return 0 /* False */;
                    }
                    return related;
                }
                function reportUnmatchedProperty(source, target, unmatchedProperty, requireOptionalProperties) {
                    var shouldSkipElaboration = false;
                    // give specific error in case where private names have the same description
                    if (unmatchedProperty.valueDeclaration
                        && ts.isNamedDeclaration(unmatchedProperty.valueDeclaration)
                        && ts.isPrivateIdentifier(unmatchedProperty.valueDeclaration.name)
                        && source.symbol
                        && source.symbol.flags & 32 /* Class */) {
                        var privateIdentifierDescription = unmatchedProperty.valueDeclaration.name.escapedText;
                        var symbolTableKey = ts.getSymbolNameForPrivateIdentifier(source.symbol, privateIdentifierDescription);
                        if (symbolTableKey && getPropertyOfType(source, symbolTableKey)) {
                            var sourceName = ts.factory.getDeclarationName(source.symbol.valueDeclaration);
                            var targetName = ts.factory.getDeclarationName(target.symbol.valueDeclaration);
                            reportError(ts.Diagnostics.Property_0_in_type_1_refers_to_a_different_member_that_cannot_be_accessed_from_within_type_2, diagnosticName(privateIdentifierDescription), diagnosticName(sourceName.escapedText === "" ? anon : sourceName), diagnosticName(targetName.escapedText === "" ? anon : targetName));
                            return;
                        }
                    }
                    var props = ts.arrayFrom(getUnmatchedProperties(source, target, requireOptionalProperties, /*matchDiscriminantProperties*/ false));
                    if (!headMessage || (headMessage.code !== ts.Diagnostics.Class_0_incorrectly_implements_interface_1.code &&
                        headMessage.code !== ts.Diagnostics.Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass.code)) {
                        shouldSkipElaboration = true; // Retain top-level error for interface implementing issues, otherwise omit it
                    }
                    if (props.length === 1) {
                        var propName = symbolToString(unmatchedProperty);
                        reportError.apply(void 0, __spreadArray([ts.Diagnostics.Property_0_is_missing_in_type_1_but_required_in_type_2, propName], getTypeNamesForErrorDisplay(source, target), false));
                        if (ts.length(unmatchedProperty.declarations)) {
                            associateRelatedInfo(ts.createDiagnosticForNode(unmatchedProperty.declarations[0], ts.Diagnostics._0_is_declared_here, propName));
                        }
                        if (shouldSkipElaboration && errorInfo) {
                            overrideNextErrorInfo++;
                        }
                    }
                    else if (tryElaborateArrayLikeErrors(source, target, /*reportErrors*/ false)) {
                        if (props.length > 5) { // arbitrary cutoff for too-long list form
                            reportError(ts.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more, typeToString(source), typeToString(target), ts.map(props.slice(0, 4), function (p) { return symbolToString(p); }).join(", "), props.length - 4);
                        }
                        else {
                            reportError(ts.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2, typeToString(source), typeToString(target), ts.map(props, function (p) { return symbolToString(p); }).join(", "));
                        }
                        if (shouldSkipElaboration && errorInfo) {
                            overrideNextErrorInfo++;
                        }
                    }
                    // No array like or unmatched property error - just issue top level error (errorInfo = undefined)
                }
                function propertiesRelatedTo(source, target, reportErrors, excludedProperties, intersectionState) {
                    if (relation === identityRelation) {
                        return propertiesIdenticalTo(source, target, excludedProperties);
                    }
                    var result = -1 /* True */;
                    if (isTupleType(target)) {
                        if (isArrayType(source) || isTupleType(source)) {
                            if (!target.target.readonly && (isReadonlyArrayType(source) || isTupleType(source) && source.target.readonly)) {
                                return 0 /* False */;
                            }
                            var sourceArity = getTypeReferenceArity(source);
                            var targetArity = getTypeReferenceArity(target);
                            var sourceRestFlag = isTupleType(source) ? source.target.combinedFlags & 4 /* Rest */ : 4 /* Rest */;
                            var targetRestFlag = target.target.combinedFlags & 4 /* Rest */;
                            var sourceMinLength = isTupleType(source) ? source.target.minLength : 0;
                            var targetMinLength = target.target.minLength;
                            if (!sourceRestFlag && sourceArity < targetMinLength) {
                                if (reportErrors) {
                                    reportError(ts.Diagnostics.Source_has_0_element_s_but_target_requires_1, sourceArity, targetMinLength);
                                }
                                return 0 /* False */;
                            }
                            if (!targetRestFlag && targetArity < sourceMinLength) {
                                if (reportErrors) {
                                    reportError(ts.Diagnostics.Source_has_0_element_s_but_target_allows_only_1, sourceMinLength, targetArity);
                                }
                                return 0 /* False */;
                            }
                            if (!targetRestFlag && (sourceRestFlag || targetArity < sourceArity)) {
                                if (reportErrors) {
                                    if (sourceMinLength < targetMinLength) {
                                        reportError(ts.Diagnostics.Target_requires_0_element_s_but_source_may_have_fewer, targetMinLength);
                                    }
                                    else {
                                        reportError(ts.Diagnostics.Target_allows_only_0_element_s_but_source_may_have_more, targetArity);
                                    }
                                }
                                return 0 /* False */;
                            }
                            var sourceTypeArguments = getTypeArguments(source);
                            var targetTypeArguments = getTypeArguments(target);
                            var startCount = Math.min(isTupleType(source) ? getStartElementCount(source.target, 11 /* NonRest */) : 0, getStartElementCount(target.target, 11 /* NonRest */));
                            var endCount = Math.min(isTupleType(source) ? getEndElementCount(source.target, 11 /* NonRest */) : 0, targetRestFlag ? getEndElementCount(target.target, 11 /* NonRest */) : 0);
                            var canExcludeDiscriminants = !!excludedProperties;
                            for (var i = 0; i < targetArity; i++) {
                                var sourceIndex = i < targetArity - endCount ? i : i + sourceArity - targetArity;
                                var sourceFlags = isTupleType(source) && (i < startCount || i >= targetArity - endCount) ? source.target.elementFlags[sourceIndex] : 4 /* Rest */;
                                var targetFlags = target.target.elementFlags[i];
                                if (targetFlags & 8 /* Variadic */ && !(sourceFlags & 8 /* Variadic */)) {
                                    if (reportErrors) {
                                        reportError(ts.Diagnostics.Source_provides_no_match_for_variadic_element_at_position_0_in_target, i);
                                    }
                                    return 0 /* False */;
                                }
                                if (sourceFlags & 8 /* Variadic */ && !(targetFlags & 12 /* Variable */)) {
                                    if (reportErrors) {
                                        reportError(ts.Diagnostics.Variadic_element_at_position_0_in_source_does_not_match_element_at_position_1_in_target, sourceIndex, i);
                                    }
                                    return 0 /* False */;
                                }
                                if (targetFlags & 1 /* Required */ && !(sourceFlags & 1 /* Required */)) {
                                    if (reportErrors) {
                                        reportError(ts.Diagnostics.Source_provides_no_match_for_required_element_at_position_0_in_target, i);
                                    }
                                    return 0 /* False */;
                                }
                                // We can only exclude discriminant properties if we have not yet encountered a variable-length element.
                                if (canExcludeDiscriminants) {
                                    if (sourceFlags & 12 /* Variable */ || targetFlags & 12 /* Variable */) {
                                        canExcludeDiscriminants = false;
                                    }
                                    if (canExcludeDiscriminants && (excludedProperties === null || excludedProperties === void 0 ? void 0 : excludedProperties.has(("" + i)))) {
                                        continue;
                                    }
                                }
                                var sourceType = !isTupleType(source) ? sourceTypeArguments[0] :
                                    i < startCount || i >= targetArity - endCount ? removeMissingType(sourceTypeArguments[sourceIndex], !!(sourceFlags & targetFlags & 2 /* Optional */)) :
                                        getElementTypeOfSliceOfTupleType(source, startCount, endCount) || neverType;
                                var targetType = targetTypeArguments[i];
                                var targetCheckType = sourceFlags & 8 /* Variadic */ && targetFlags & 4 /* Rest */ ? createArrayType(targetType) :
                                    removeMissingType(targetType, !!(targetFlags & 2 /* Optional */));
                                var related = isRelatedTo(sourceType, targetCheckType, reportErrors, /*headMessage*/ undefined, intersectionState);
                                if (!related) {
                                    if (reportErrors && (targetArity > 1 || sourceArity > 1)) {
                                        if (i < startCount || i >= targetArity - endCount || sourceArity - startCount - endCount === 1) {
                                            reportIncompatibleError(ts.Diagnostics.Type_at_position_0_in_source_is_not_compatible_with_type_at_position_1_in_target, sourceIndex, i);
                                        }
                                        else {
                                            reportIncompatibleError(ts.Diagnostics.Type_at_positions_0_through_1_in_source_is_not_compatible_with_type_at_position_2_in_target, startCount, sourceArity - endCount - 1, i);
                                        }
                                    }
                                    return 0 /* False */;
                                }
                                result &= related;
                            }
                            return result;
                        }
                        if (target.target.combinedFlags & 12 /* Variable */) {
                            return 0 /* False */;
                        }
                    }
                    var requireOptionalProperties = (relation === subtypeRelation || relation === strictSubtypeRelation) && !isObjectLiteralType(source) && !isEmptyArrayLiteralType(source) && !isTupleType(source);
                    var unmatchedProperty = getUnmatchedProperty(source, target, requireOptionalProperties, /*matchDiscriminantProperties*/ false);
                    if (unmatchedProperty) {
                        if (reportErrors) {
                            reportUnmatchedProperty(source, target, unmatchedProperty, requireOptionalProperties);
                        }
                        return 0 /* False */;
                    }
                    if (isObjectLiteralType(target)) {
                        for (var _i = 0, _a = excludeProperties(getPropertiesOfType(source), excludedProperties); _i < _a.length; _i++) {
                            var sourceProp = _a[_i];
                            if (!getPropertyOfObjectType(target, sourceProp.escapedName)) {
                                var sourceType = getTypeOfSymbol(sourceProp);
                                if (!(sourceType.flags & 32768 /* Undefined */)) {
                                    if (reportErrors) {
                                        reportError(ts.Diagnostics.Property_0_does_not_exist_on_type_1, symbolToString(sourceProp), typeToString(target));
                                    }
                                    return 0 /* False */;
                                }
                            }
                        }
                    }
                    // We only call this for union target types when we're attempting to do excess property checking - in those cases, we want to get _all possible props_
                    // from the target union, across all members
                    var properties = getPropertiesOfType(target);
                    var numericNamesOnly = isTupleType(source) && isTupleType(target);
                    for (var _b = 0, _c = excludeProperties(properties, excludedProperties); _b < _c.length; _b++) {
                        var targetProp = _c[_b];
                        var name = targetProp.escapedName;
                        if (!(targetProp.flags & 4194304 /* Prototype */) && (!numericNamesOnly || isNumericLiteralName(name) || name === "length")) {
                            var sourceProp = getPropertyOfType(source, name);
                            if (sourceProp && sourceProp !== targetProp) {
                                var related = propertyRelatedTo(source, target, sourceProp, targetProp, getNonMissingTypeOfSymbol, reportErrors, intersectionState, relation === comparableRelation);
                                if (!related) {
                                    return 0 /* False */;
                                }
                                result &= related;
                            }
                        }
                    }
                    return result;
                }
                function propertiesIdenticalTo(source, target, excludedProperties) {
                    if (!(source.flags & 524288 /* Object */ && target.flags & 524288 /* Object */)) {
                        return 0 /* False */;
                    }
                    var sourceProperties = excludeProperties(getPropertiesOfObjectType(source), excludedProperties);
                    var targetProperties = excludeProperties(getPropertiesOfObjectType(target), excludedProperties);
                    if (sourceProperties.length !== targetProperties.length) {
                        return 0 /* False */;
                    }
                    var result = -1 /* True */;
                    for (var _i = 0, sourceProperties_1 = sourceProperties; _i < sourceProperties_1.length; _i++) {
                        var sourceProp = sourceProperties_1[_i];
                        var targetProp = getPropertyOfObjectType(target, sourceProp.escapedName);
                        if (!targetProp) {
                            return 0 /* False */;
                        }
                        var related = compareProperties(sourceProp, targetProp, isRelatedTo);
                        if (!related) {
                            return 0 /* False */;
                        }
                        result &= related;
                    }
                    return result;
                }
                function signaturesRelatedTo(source, target, kind, reportErrors) {
                    var _a, _b;
                    if (relation === identityRelation) {
                        return signaturesIdenticalTo(source, target, kind);
                    }
                    if (target === anyFunctionType || source === anyFunctionType) {
                        return -1 /* True */;
                    }
                    var sourceIsJSConstructor = source.symbol && isJSConstructor(source.symbol.valueDeclaration);
                    var targetIsJSConstructor = target.symbol && isJSConstructor(target.symbol.valueDeclaration);
                    var sourceSignatures = getSignaturesOfType(source, (sourceIsJSConstructor && kind === 1 /* Construct */) ?
                        0 /* Call */ : kind);
                    var targetSignatures = getSignaturesOfType(target, (targetIsJSConstructor && kind === 1 /* Construct */) ?
                        0 /* Call */ : kind);
                    if (kind === 1 /* Construct */ && sourceSignatures.length && targetSignatures.length) {
                        var sourceIsAbstract = !!(sourceSignatures[0].flags & 4 /* Abstract */);
                        var targetIsAbstract = !!(targetSignatures[0].flags & 4 /* Abstract */);
                        if (sourceIsAbstract && !targetIsAbstract) {
                            // An abstract constructor type is not assignable to a non-abstract constructor type
                            // as it would otherwise be possible to new an abstract class. Note that the assignability
                            // check we perform for an extends clause excludes construct signatures from the target,
                            // so this check never proceeds.
                            if (reportErrors) {
                                reportError(ts.Diagnostics.Cannot_assign_an_abstract_constructor_type_to_a_non_abstract_constructor_type);
                            }
                            return 0 /* False */;
                        }
                        if (!constructorVisibilitiesAreCompatible(sourceSignatures[0], targetSignatures[0], reportErrors)) {
                            return 0 /* False */;
                        }
                    }
                    var result = -1 /* True */;
                    var saveErrorInfo = captureErrorCalculationState();
                    var incompatibleReporter = kind === 1 /* Construct */ ? reportIncompatibleConstructSignatureReturn : reportIncompatibleCallSignatureReturn;
                    var sourceObjectFlags = ts.getObjectFlags(source);
                    var targetObjectFlags = ts.getObjectFlags(target);
                    if (sourceObjectFlags & 64 /* Instantiated */ && targetObjectFlags & 64 /* Instantiated */ && source.symbol === target.symbol) {
                        // We have instantiations of the same anonymous type (which typically will be the type of a
                        // method). Simply do a pairwise comparison of the signatures in the two signature lists instead
                        // of the much more expensive N * M comparison matrix we explore below. We erase type parameters
                        // as they are known to always be the same.
                        for (var i = 0; i < targetSignatures.length; i++) {
                            var related = signatureRelatedTo(sourceSignatures[i], targetSignatures[i], /*erase*/ true, reportErrors, incompatibleReporter(sourceSignatures[i], targetSignatures[i]));
                            if (!related) {
                                return 0 /* False */;
                            }
                            result &= related;
                        }
                    }
                    else if (sourceSignatures.length === 1 && targetSignatures.length === 1) {
                        // For simple functions (functions with a single signature) we only erase type parameters for
                        // the comparable relation. Otherwise, if the source signature is generic, we instantiate it
                        // in the context of the target signature before checking the relationship. Ideally we'd do
                        // this regardless of the number of signatures, but the potential costs are prohibitive due
                        // to the quadratic nature of the logic below.
                        var eraseGenerics = relation === comparableRelation || !!compilerOptions.noStrictGenericChecks;
                        var sourceSignature = ts.first(sourceSignatures);
                        var targetSignature = ts.first(targetSignatures);
                        result = signatureRelatedTo(sourceSignature, targetSignature, eraseGenerics, reportErrors, incompatibleReporter(sourceSignature, targetSignature));
                        if (!result && reportErrors && kind === 1 /* Construct */ && (sourceObjectFlags & targetObjectFlags) &&
                            (((_a = targetSignature.declaration) === null || _a === void 0 ? void 0 : _a.kind) === 169 /* Constructor */ || ((_b = sourceSignature.declaration) === null || _b === void 0 ? void 0 : _b.kind) === 169 /* Constructor */)) {
                            var constructSignatureToString = function (signature) {
                                return signatureToString(signature, /*enclosingDeclaration*/ undefined, 262144 /* WriteArrowStyleSignature */, kind);
                            };
                            reportError(ts.Diagnostics.Type_0_is_not_assignable_to_type_1, constructSignatureToString(sourceSignature), constructSignatureToString(targetSignature));
                            reportError(ts.Diagnostics.Types_of_construct_signatures_are_incompatible);
                            return result;
                        }
                    }
                    else {
                        outer: for (var _i = 0, targetSignatures_1 = targetSignatures; _i < targetSignatures_1.length; _i++) {
                            var t = targetSignatures_1[_i];
                            // Only elaborate errors from the first failure
                            var shouldElaborateErrors = reportErrors;
                            for (var _c = 0, sourceSignatures_1 = sourceSignatures; _c < sourceSignatures_1.length; _c++) {
                                var s = sourceSignatures_1[_c];
                                var related = signatureRelatedTo(s, t, /*erase*/ true, shouldElaborateErrors, incompatibleReporter(s, t));
                                if (related) {
                                    result &= related;
                                    resetErrorInfo(saveErrorInfo);
                                    continue outer;
                                }
                                shouldElaborateErrors = false;
                            }
                            if (shouldElaborateErrors) {
                                reportError(ts.Diagnostics.Type_0_provides_no_match_for_the_signature_1, typeToString(source), signatureToString(t, /*enclosingDeclaration*/ undefined, /*flags*/ undefined, kind));
                            }
                            return 0 /* False */;
                        }
                    }
                    return result;
                }
                function reportIncompatibleCallSignatureReturn(siga, sigb) {
                    if (siga.parameters.length === 0 && sigb.parameters.length === 0) {
                        return function (source, target) { return reportIncompatibleError(ts.Diagnostics.Call_signatures_with_no_arguments_have_incompatible_return_types_0_and_1, typeToString(source), typeToString(target)); };
                    }
                    return function (source, target) { return reportIncompatibleError(ts.Diagnostics.Call_signature_return_types_0_and_1_are_incompatible, typeToString(source), typeToString(target)); };
                }
                function reportIncompatibleConstructSignatureReturn(siga, sigb) {
                    if (siga.parameters.length === 0 && sigb.parameters.length === 0) {
                        return function (source, target) { return reportIncompatibleError(ts.Diagnostics.Construct_signatures_with_no_arguments_have_incompatible_return_types_0_and_1, typeToString(source), typeToString(target)); };
                    }
                    return function (source, target) { return reportIncompatibleError(ts.Diagnostics.Construct_signature_return_types_0_and_1_are_incompatible, typeToString(source), typeToString(target)); };
                }
                /**
                 * See signatureAssignableTo, compareSignaturesIdentical
                 */
                function signatureRelatedTo(source, target, erase, reportErrors, incompatibleReporter) {
                    return compareSignaturesRelated(erase ? getErasedSignature(source) : source, erase ? getErasedSignature(target) : target, relation === strictSubtypeRelation ? 8 /* StrictArity */ : 0, reportErrors, reportError, incompatibleReporter, isRelatedTo, makeFunctionTypeMapper(reportUnreliableMarkers));
                }
                function signaturesIdenticalTo(source, target, kind) {
                    var sourceSignatures = getSignaturesOfType(source, kind);
                    var targetSignatures = getSignaturesOfType(target, kind);
                    if (sourceSignatures.length !== targetSignatures.length) {
                        return 0 /* False */;
                    }
                    var result = -1 /* True */;
                    for (var i = 0; i < sourceSignatures.length; i++) {
                        var related = compareSignaturesIdentical(sourceSignatures[i], targetSignatures[i], /*partialMatch*/ false, /*ignoreThisTypes*/ false, /*ignoreReturnTypes*/ false, isRelatedTo);
                        if (!related) {
                            return 0 /* False */;
                        }
                        result &= related;
                    }
                    return result;
                }
                function membersRelatedToIndexInfo(source, targetInfo, reportErrors) {
                    var result = -1 /* True */;
                    var keyType = targetInfo.keyType;
                    var props = source.flags & 2097152 /* Intersection */ ? getPropertiesOfUnionOrIntersectionType(source) : getPropertiesOfObjectType(source);
                    for (var _i = 0, props_2 = props; _i < props_2.length; _i++) {
                        var prop = props_2[_i];
                        // Skip over ignored JSX and symbol-named members
                        if (isIgnoredJsxProperty(source, prop)) {
                            continue;
                        }
                        if (isApplicableIndexType(getLiteralTypeFromProperty(prop, 8576 /* StringOrNumberLiteralOrUnique */), keyType)) {
                            var propType = getNonMissingTypeOfSymbol(prop);
                            var type = exactOptionalPropertyTypes || propType.flags & 32768 /* Undefined */ || keyType === numberType || !(prop.flags & 16777216 /* Optional */)
                                ? propType
                                : getTypeWithFacts(propType, 524288 /* NEUndefined */);
                            var related = isRelatedTo(type, targetInfo.type, reportErrors);
                            if (!related) {
                                if (reportErrors) {
                                    reportError(ts.Diagnostics.Property_0_is_incompatible_with_index_signature, symbolToString(prop));
                                }
                                return 0 /* False */;
                            }
                            result &= related;
                        }
                    }
                    for (var _a = 0, _b = getIndexInfosOfType(source); _a < _b.length; _a++) {
                        var info = _b[_a];
                        if (isApplicableIndexType(info.keyType, keyType)) {
                            var related = indexInfoRelatedTo(info, targetInfo, reportErrors);
                            if (!related) {
                                return 0 /* False */;
                            }
                            result &= related;
                        }
                    }
                    return result;
                }
                function indexInfoRelatedTo(sourceInfo, targetInfo, reportErrors) {
                    var related = isRelatedTo(sourceInfo.type, targetInfo.type, reportErrors);
                    if (!related && reportErrors) {
                        if (sourceInfo.keyType === targetInfo.keyType) {
                            reportError(ts.Diagnostics._0_index_signatures_are_incompatible, typeToString(sourceInfo.keyType));
                        }
                        else {
                            reportError(ts.Diagnostics._0_and_1_index_signatures_are_incompatible, typeToString(sourceInfo.keyType), typeToString(targetInfo.keyType));
                        }
                    }
                    return related;
                }
                function indexSignaturesRelatedTo(source, target, sourceIsPrimitive, reportErrors, intersectionState) {
                    if (relation === identityRelation) {
                        return indexSignaturesIdenticalTo(source, target);
                    }
                    var indexInfos = getIndexInfosOfType(target);
                    var targetHasStringIndex = ts.some(indexInfos, function (info) { return info.keyType === stringType; });
                    var result = -1 /* True */;
                    for (var _i = 0, indexInfos_3 = indexInfos; _i < indexInfos_3.length; _i++) {
                        var targetInfo = indexInfos_3[_i];
                        var related = !sourceIsPrimitive && targetHasStringIndex && targetInfo.type.flags & 1 /* Any */ ? -1 /* True */ :
                            isGenericMappedType(source) && targetHasStringIndex ? isRelatedTo(getTemplateTypeFromMappedType(source), targetInfo.type, reportErrors) :
                                typeRelatedToIndexInfo(source, targetInfo, reportErrors, intersectionState);
                        if (!related) {
                            return 0 /* False */;
                        }
                        result &= related;
                    }
                    return result;
                }
                function typeRelatedToIndexInfo(source, targetInfo, reportErrors, intersectionState) {
                    var sourceInfo = getApplicableIndexInfo(source, targetInfo.keyType);
                    if (sourceInfo) {
                        return indexInfoRelatedTo(sourceInfo, targetInfo, reportErrors);
                    }
                    if (!(intersectionState & 1 /* Source */) && isObjectTypeWithInferableIndex(source)) {
                        // Intersection constituents are never considered to have an inferred index signature
                        return membersRelatedToIndexInfo(source, targetInfo, reportErrors);
                    }
                    if (reportErrors) {
                        reportError(ts.Diagnostics.Index_signature_for_type_0_is_missing_in_type_1, typeToString(targetInfo.keyType), typeToString(source));
                    }
                    return 0 /* False */;
                }
                function indexSignaturesIdenticalTo(source, target) {
                    var sourceInfos = getIndexInfosOfType(source);
                    var targetInfos = getIndexInfosOfType(target);
                    if (sourceInfos.length !== targetInfos.length) {
                        return 0 /* False */;
                    }
                    for (var _i = 0, targetInfos_1 = targetInfos; _i < targetInfos_1.length; _i++) {
                        var targetInfo = targetInfos_1[_i];
                        var sourceInfo = getIndexInfoOfType(source, targetInfo.keyType);
                        if (!(sourceInfo && isRelatedTo(sourceInfo.type, targetInfo.type) && sourceInfo.isReadonly === targetInfo.isReadonly)) {
                            return 0 /* False */;
                        }
                    }
                    return -1 /* True */;
                }
                function constructorVisibilitiesAreCompatible(sourceSignature, targetSignature, reportErrors) {
                    if (!sourceSignature.declaration || !targetSignature.declaration) {
                        return true;
                    }
                    var sourceAccessibility = ts.getSelectedEffectiveModifierFlags(sourceSignature.declaration, 24 /* NonPublicAccessibilityModifier */);
                    var targetAccessibility = ts.getSelectedEffectiveModifierFlags(targetSignature.declaration, 24 /* NonPublicAccessibilityModifier */);
                    // A public, protected and private signature is assignable to a private signature.
                    if (targetAccessibility === 8 /* Private */) {
                        return true;
                    }
                    // A public and protected signature is assignable to a protected signature.
                    if (targetAccessibility === 16 /* Protected */ && sourceAccessibility !== 8 /* Private */) {
                        return true;
                    }
                    // Only a public signature is assignable to public signature.
                    if (targetAccessibility !== 16 /* Protected */ && !sourceAccessibility) {
                        return true;
                    }
                    if (reportErrors) {
                        reportError(ts.Diagnostics.Cannot_assign_a_0_constructor_type_to_a_1_constructor_type, visibilityToString(sourceAccessibility), visibilityToString(targetAccessibility));
                    }
                    return false;
                }
            }
            function typeCouldHaveTopLevelSingletonTypes(type) {
                // Okay, yes, 'boolean' is a union of 'true | false', but that's not useful
                // in error reporting scenarios. If you need to use this function but that detail matters,
                // feel free to add a flag.
                if (type.flags & 16 /* Boolean */) {
                    return false;
                }
                if (type.flags & 3145728 /* UnionOrIntersection */) {
                    return !!ts.forEach(type.types, typeCouldHaveTopLevelSingletonTypes);
                }
                if (type.flags & 465829888 /* Instantiable */) {
                    var constraint = getConstraintOfType(type);
                    if (constraint && constraint !== type) {
                        return typeCouldHaveTopLevelSingletonTypes(constraint);
                    }
                }
                return isUnitType(type) || !!(type.flags & 134217728 /* TemplateLiteral */);
            }
            function getBestMatchingType(source, target, isRelatedTo) {
                if (isRelatedTo === void 0) {
                    isRelatedTo = compareTypesAssignable;
                }
                return findMatchingDiscriminantType(source, target, isRelatedTo, /*skipPartial*/ true) ||
                    findMatchingTypeReferenceOrTypeAliasReference(source, target) ||
                    findBestTypeForObjectLiteral(source, target) ||
                    findBestTypeForInvokable(source, target) ||
                    findMostOverlappyType(source, target);
            }
            function discriminateTypeByDiscriminableItems(target, discriminators, related, defaultValue, skipPartial) {
                // undefined=unknown, true=discriminated, false=not discriminated
                // The state of each type progresses from left to right. Discriminated types stop at 'true'.
                var discriminable = target.types.map(function (_) { return undefined; });
                for (var _i = 0, discriminators_1 = discriminators; _i < discriminators_1.length; _i++) {
                    var _a = discriminators_1[_i], getDiscriminatingType = _a[0], propertyName = _a[1];
                    var targetProp = getUnionOrIntersectionProperty(target, propertyName);
                    if (skipPartial && targetProp && ts.getCheckFlags(targetProp) & 16 /* ReadPartial */) {
                        continue;
                    }
                    var i = 0;
                    for (var _b = 0, _c = target.types; _b < _c.length; _b++) {
                        var type = _c[_b];
                        var targetType = getTypeOfPropertyOfType(type, propertyName);
                        if (targetType && related(getDiscriminatingType(), targetType)) {
                            discriminable[i] = discriminable[i] === undefined ? true : discriminable[i];
                        }
                        else {
                            discriminable[i] = false;
                        }
                        i++;
                    }
                }
                var match = discriminable.indexOf(/*searchElement*/ true);
                if (match === -1) {
                    return defaultValue;
                }
                // make sure exactly 1 matches before returning it
                var nextMatch = discriminable.indexOf(/*searchElement*/ true, match + 1);
                while (nextMatch !== -1) {
                    if (!isTypeIdenticalTo(target.types[match], target.types[nextMatch])) {
                        return defaultValue;
                    }
                    nextMatch = discriminable.indexOf(/*searchElement*/ true, nextMatch + 1);
                }
                return target.types[match];
            }
            /**
             * A type is 'weak' if it is an object type with at least one optional property
             * and no required properties, call/construct signatures or index signatures
             */
            function isWeakType(type) {
                if (type.flags & 524288 /* Object */) {
                    var resolved = resolveStructuredTypeMembers(type);
                    return resolved.callSignatures.length === 0 && resolved.constructSignatures.length === 0 && resolved.indexInfos.length === 0 &&
                        resolved.properties.length > 0 && ts.every(resolved.properties, function (p) { return !!(p.flags & 16777216 /* Optional */); });
                }
                if (type.flags & 2097152 /* Intersection */) {
                    return ts.every(type.types, isWeakType);
                }
                return false;
            }
            function hasCommonProperties(source, target, isComparingJsxAttributes) {
                for (var _i = 0, _a = getPropertiesOfType(source); _i < _a.length; _i++) {
                    var prop = _a[_i];
                    if (isKnownProperty(target, prop.escapedName, isComparingJsxAttributes)) {
                        return true;
                    }
                }
                return false;
            }
            // Return a type reference where the source type parameter is replaced with the target marker
            // type, and flag the result as a marker type reference.
            function getMarkerTypeReference(type, source, target) {
                var result = createTypeReference(type, ts.map(type.typeParameters, function (t) { return t === source ? target : t; }));
                result.objectFlags |= 4096 /* MarkerType */;
                return result;
            }
            function getAliasVariances(symbol) {
                var links = getSymbolLinks(symbol);
                return getVariancesWorker(links.typeParameters, links, function (_links, param, marker) {
                    var type = getTypeAliasInstantiation(symbol, instantiateTypes(links.typeParameters, makeUnaryTypeMapper(param, marker)));
                    type.aliasTypeArgumentsContainsMarker = true;
                    return type;
                });
            }
            // Return an array containing the variance of each type parameter. The variance is effectively
            // a digest of the type comparisons that occur for each type argument when instantiations of the
            // generic type are structurally compared. We infer the variance information by comparing
            // instantiations of the generic type for type arguments with known relations. The function
            // returns the emptyArray singleton when invoked recursively for the given generic type.
            function getVariancesWorker(typeParameters, cache, createMarkerType) {
                var _a, _b, _c;
                if (typeParameters === void 0) {
                    typeParameters = ts.emptyArray;
                }
                var variances = cache.variances;
                if (!variances) {
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("checkTypes" /* CheckTypes */, "getVariancesWorker", { arity: typeParameters.length, id: (_c = (_a = cache.id) !== null && _a !== void 0 ? _a : (_b = cache.declaredType) === null || _b === void 0 ? void 0 : _b.id) !== null && _c !== void 0 ? _c : -1 });
                    // The emptyArray singleton is used to signal a recursive invocation.
                    cache.variances = ts.emptyArray;
                    variances = [];
                    var _loop_21 = function (tp) {
                        var unmeasurable = false;
                        var unreliable = false;
                        var oldHandler = outofbandVarianceMarkerHandler;
                        outofbandVarianceMarkerHandler = function (onlyUnreliable) { return onlyUnreliable ? unreliable = true : unmeasurable = true; };
                        // We first compare instantiations where the type parameter is replaced with
                        // marker types that have a known subtype relationship. From this we can infer
                        // invariance, covariance, contravariance or bivariance.
                        var typeWithSuper = createMarkerType(cache, tp, markerSuperType);
                        var typeWithSub = createMarkerType(cache, tp, markerSubType);
                        var variance = (isTypeAssignableTo(typeWithSub, typeWithSuper) ? 1 /* Covariant */ : 0) |
                            (isTypeAssignableTo(typeWithSuper, typeWithSub) ? 2 /* Contravariant */ : 0);
                        // If the instantiations appear to be related bivariantly it may be because the
                        // type parameter is independent (i.e. it isn't witnessed anywhere in the generic
                        // type). To determine this we compare instantiations where the type parameter is
                        // replaced with marker types that are known to be unrelated.
                        if (variance === 3 /* Bivariant */ && isTypeAssignableTo(createMarkerType(cache, tp, markerOtherType), typeWithSuper)) {
                            variance = 4 /* Independent */;
                        }
                        outofbandVarianceMarkerHandler = oldHandler;
                        if (unmeasurable || unreliable) {
                            if (unmeasurable) {
                                variance |= 8 /* Unmeasurable */;
                            }
                            if (unreliable) {
                                variance |= 16 /* Unreliable */;
                            }
                        }
                        variances.push(variance);
                    };
                    for (var _i = 0, typeParameters_1 = typeParameters; _i < typeParameters_1.length; _i++) {
                        var tp = typeParameters_1[_i];
                        _loop_21(tp);
                    }
                    cache.variances = variances;
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                }
                return variances;
            }
            function getVariances(type) {
                // Arrays and tuples are known to be covariant, no need to spend time computing this.
                if (type === globalArrayType || type === globalReadonlyArrayType || type.objectFlags & 8 /* Tuple */) {
                    return arrayVariances;
                }
                return getVariancesWorker(type.typeParameters, type, getMarkerTypeReference);
            }
            // Return true if the given type reference has a 'void' type argument for a covariant type parameter.
            // See comment at call in recursiveTypeRelatedTo for when this case matters.
            function hasCovariantVoidArgument(typeArguments, variances) {
                for (var i = 0; i < variances.length; i++) {
                    if ((variances[i] & 7 /* VarianceMask */) === 1 /* Covariant */ && typeArguments[i].flags & 16384 /* Void */) {
                        return true;
                    }
                }
                return false;
            }
            function isUnconstrainedTypeParameter(type) {
                return type.flags & 262144 /* TypeParameter */ && !getConstraintOfTypeParameter(type);
            }
            function isNonDeferredTypeReference(type) {
                return !!(ts.getObjectFlags(type) & 4 /* Reference */) && !type.node;
            }
            function isTypeReferenceWithGenericArguments(type) {
                return isNonDeferredTypeReference(type) && ts.some(getTypeArguments(type), function (t) { return !!(t.flags & 262144 /* TypeParameter */) || isTypeReferenceWithGenericArguments(t); });
            }
            /**
             * getTypeReferenceId(A<T, number, U>) returns "111=0-12=1"
             *   where A.id=111 and number.id=12
             */
            function getTypeReferenceId(type, typeParameters, depth) {
                if (depth === void 0) {
                    depth = 0;
                }
                var result = "" + type.target.id;
                for (var _i = 0, _a = getTypeArguments(type); _i < _a.length; _i++) {
                    var t = _a[_i];
                    if (isUnconstrainedTypeParameter(t)) {
                        var index = typeParameters.indexOf(t);
                        if (index < 0) {
                            index = typeParameters.length;
                            typeParameters.push(t);
                        }
                        result += "=" + index;
                    }
                    else if (depth < 4 && isTypeReferenceWithGenericArguments(t)) {
                        result += "<" + getTypeReferenceId(t, typeParameters, depth + 1) + ">";
                    }
                    else {
                        result += "-" + t.id;
                    }
                }
                return result;
            }
            /**
             * To improve caching, the relation key for two generic types uses the target's id plus ids of the type parameters.
             * For other cases, the types ids are used.
             */
            function getRelationKey(source, target, intersectionState, relation) {
                if (relation === identityRelation && source.id > target.id) {
                    var temp = source;
                    source = target;
                    target = temp;
                }
                var postFix = intersectionState ? ":" + intersectionState : "";
                if (isTypeReferenceWithGenericArguments(source) && isTypeReferenceWithGenericArguments(target)) {
                    var typeParameters = [];
                    return getTypeReferenceId(source, typeParameters) + "," + getTypeReferenceId(target, typeParameters) + postFix;
                }
                return source.id + "," + target.id + postFix;
            }
            // Invoke the callback for each underlying property symbol of the given symbol and return the first
            // value that isn't undefined.
            function forEachProperty(prop, callback) {
                if (ts.getCheckFlags(prop) & 6 /* Synthetic */) {
                    for (var _i = 0, _a = prop.containingType.types; _i < _a.length; _i++) {
                        var t = _a[_i];
                        var p = getPropertyOfType(t, prop.escapedName);
                        var result = p && forEachProperty(p, callback);
                        if (result) {
                            return result;
                        }
                    }
                    return undefined;
                }
                return callback(prop);
            }
            // Return the declaring class type of a property or undefined if property not declared in class
            function getDeclaringClass(prop) {
                return prop.parent && prop.parent.flags & 32 /* Class */ ? getDeclaredTypeOfSymbol(getParentOfSymbol(prop)) : undefined;
            }
            // Return the inherited type of the given property or undefined if property doesn't exist in a base class.
            function getTypeOfPropertyInBaseClass(property) {
                var classType = getDeclaringClass(property);
                var baseClassType = classType && getBaseTypes(classType)[0];
                return baseClassType && getTypeOfPropertyOfType(baseClassType, property.escapedName);
            }
            // Return true if some underlying source property is declared in a class that derives
            // from the given base class.
            function isPropertyInClassDerivedFrom(prop, baseClass) {
                return forEachProperty(prop, function (sp) {
                    var sourceClass = getDeclaringClass(sp);
                    return sourceClass ? hasBaseType(sourceClass, baseClass) : false;
                });
            }
            // Return true if source property is a valid override of protected parts of target property.
            function isValidOverrideOf(sourceProp, targetProp) {
                return !forEachProperty(targetProp, function (tp) {
                    return ts.getDeclarationModifierFlagsFromSymbol(tp) & 16 /* Protected */ ?
                        !isPropertyInClassDerivedFrom(sourceProp, getDeclaringClass(tp)) : false;
                });
            }
            // Return true if the given class derives from each of the declaring classes of the protected
            // constituents of the given property.
            function isClassDerivedFromDeclaringClasses(checkClass, prop, writing) {
                return forEachProperty(prop, function (p) {
                    return ts.getDeclarationModifierFlagsFromSymbol(p, writing) & 16 /* Protected */ ?
                        !hasBaseType(checkClass, getDeclaringClass(p)) : false;
                }) ? undefined : checkClass;
            }
            // Return true if the given type is deeply nested. We consider this to be the case when structural type comparisons
            // for 5 or more occurrences or instantiations of the type have been recorded on the given stack. It is possible,
            // though highly unlikely, for this test to be true in a situation where a chain of instantiations is not infinitely
            // expanding. Effectively, we will generate a false positive when two types are structurally equal to at least 5
            // levels, but unequal at some level beyond that.
            // In addition, this will also detect when an indexed access has been chained off of 5 or more times (which is essentially
            // the dual of the structural comparison), and likewise mark the type as deeply nested, potentially adding false positives
            // for finite but deeply expanding indexed accesses (eg, for `Q[P1][P2][P3][P4][P5]`).
            // It also detects when a recursive type reference has expanded 5 or more times, eg, if the true branch of
            // `type A<T> = null extends T ? [A<NonNullable<T>>] : [T]`
            // has expanded into `[A<NonNullable<NonNullable<NonNullable<NonNullable<NonNullable<T>>>>>>]`
            // in such cases we need to terminate the expansion, and we do so here.
            function isDeeplyNestedType(type, stack, depth) {
                if (depth >= 5) {
                    var identity_1 = getRecursionIdentity(type);
                    var count = 0;
                    for (var i = 0; i < depth; i++) {
                        if (getRecursionIdentity(stack[i]) === identity_1) {
                            count++;
                            if (count >= 5) {
                                return true;
                            }
                        }
                    }
                }
                return false;
            }
            // The recursion identity of a type is an object identity that is shared among multiple instantiations of the type.
            // We track recursion identities in order to identify deeply nested and possibly infinite type instantiations with
            // the same origin. For example, when type parameters are in scope in an object type such as { x: T }, all
            // instantiations of that type have the same recursion identity. The default recursion identity is the object
            // identity of the type, meaning that every type is unique. Generally, types with constituents that could circularly
            // reference the type have a recursion identity that differs from the object identity.
            function getRecursionIdentity(type) {
                // Object and array literals are known not to contain recursive references and don't need a recursion identity.
                if (type.flags & 524288 /* Object */ && !isObjectOrArrayLiteralType(type)) {
                    if (ts.getObjectFlags(type) && 4 /* Reference */ && type.node) {
                        // Deferred type references are tracked through their associated AST node. This gives us finer
                        // granularity than using their associated target because each manifest type reference has a
                        // unique AST node.
                        return type.node;
                    }
                    if (type.symbol && !(ts.getObjectFlags(type) & 16 /* Anonymous */ && type.symbol.flags & 32 /* Class */)) {
                        // We track all object types that have an associated symbol (representing the origin of the type), but
                        // exclude the static side of classes from this check since it shares its symbol with the instance side.
                        return type.symbol;
                    }
                    if (isTupleType(type)) {
                        // Tuple types are tracked through their target type
                        return type.target;
                    }
                }
                if (type.flags & 262144 /* TypeParameter */) {
                    return type.symbol;
                }
                if (type.flags & 8388608 /* IndexedAccess */) {
                    // Identity is the leftmost object type in a chain of indexed accesses, eg, in A[P][Q] it is A
                    do {
                        type = type.objectType;
                    } while (type.flags & 8388608 /* IndexedAccess */);
                    return type;
                }
                if (type.flags & 16777216 /* Conditional */) {
                    // The root object represents the origin of the conditional type
                    return type.root;
                }
                return type;
            }
            function isPropertyIdenticalTo(sourceProp, targetProp) {
                return compareProperties(sourceProp, targetProp, compareTypesIdentical) !== 0 /* False */;
            }
            function compareProperties(sourceProp, targetProp, compareTypes) {
                // Two members are considered identical when
                // - they are public properties with identical names, optionality, and types,
                // - they are private or protected properties originating in the same declaration and having identical types
                if (sourceProp === targetProp) {
                    return -1 /* True */;
                }
                var sourcePropAccessibility = ts.getDeclarationModifierFlagsFromSymbol(sourceProp) & 24 /* NonPublicAccessibilityModifier */;
                var targetPropAccessibility = ts.getDeclarationModifierFlagsFromSymbol(targetProp) & 24 /* NonPublicAccessibilityModifier */;
                if (sourcePropAccessibility !== targetPropAccessibility) {
                    return 0 /* False */;
                }
                if (sourcePropAccessibility) {
                    if (getTargetSymbol(sourceProp) !== getTargetSymbol(targetProp)) {
                        return 0 /* False */;
                    }
                }
                else {
                    if ((sourceProp.flags & 16777216 /* Optional */) !== (targetProp.flags & 16777216 /* Optional */)) {
                        return 0 /* False */;
                    }
                }
                if (isReadonlySymbol(sourceProp) !== isReadonlySymbol(targetProp)) {
                    return 0 /* False */;
                }
                return compareTypes(getTypeOfSymbol(sourceProp), getTypeOfSymbol(targetProp));
            }
            function isMatchingSignature(source, target, partialMatch) {
                var sourceParameterCount = getParameterCount(source);
                var targetParameterCount = getParameterCount(target);
                var sourceMinArgumentCount = getMinArgumentCount(source);
                var targetMinArgumentCount = getMinArgumentCount(target);
                var sourceHasRestParameter = hasEffectiveRestParameter(source);
                var targetHasRestParameter = hasEffectiveRestParameter(target);
                // A source signature matches a target signature if the two signatures have the same number of required,
                // optional, and rest parameters.
                if (sourceParameterCount === targetParameterCount &&
                    sourceMinArgumentCount === targetMinArgumentCount &&
                    sourceHasRestParameter === targetHasRestParameter) {
                    return true;
                }
                // A source signature partially matches a target signature if the target signature has no fewer required
                // parameters
                if (partialMatch && sourceMinArgumentCount <= targetMinArgumentCount) {
                    return true;
                }
                return false;
            }
            /**
             * See signatureRelatedTo, compareSignaturesIdentical
             */
            function compareSignaturesIdentical(source, target, partialMatch, ignoreThisTypes, ignoreReturnTypes, compareTypes) {
                // TODO (drosen): De-duplicate code between related functions.
                if (source === target) {
                    return -1 /* True */;
                }
                if (!(isMatchingSignature(source, target, partialMatch))) {
                    return 0 /* False */;
                }
                // Check that the two signatures have the same number of type parameters.
                if (ts.length(source.typeParameters) !== ts.length(target.typeParameters)) {
                    return 0 /* False */;
                }
                // Check that type parameter constraints and defaults match. If they do, instantiate the source
                // signature with the type parameters of the target signature and continue the comparison.
                if (target.typeParameters) {
                    var mapper = createTypeMapper(source.typeParameters, target.typeParameters);
                    for (var i = 0; i < target.typeParameters.length; i++) {
                        var s = source.typeParameters[i];
                        var t = target.typeParameters[i];
                        if (!(s === t || compareTypes(instantiateType(getConstraintFromTypeParameter(s), mapper) || unknownType, getConstraintFromTypeParameter(t) || unknownType) &&
                            compareTypes(instantiateType(getDefaultFromTypeParameter(s), mapper) || unknownType, getDefaultFromTypeParameter(t) || unknownType))) {
                            return 0 /* False */;
                        }
                    }
                    source = instantiateSignature(source, mapper, /*eraseTypeParameters*/ true);
                }
                var result = -1 /* True */;
                if (!ignoreThisTypes) {
                    var sourceThisType = getThisTypeOfSignature(source);
                    if (sourceThisType) {
                        var targetThisType = getThisTypeOfSignature(target);
                        if (targetThisType) {
                            var related = compareTypes(sourceThisType, targetThisType);
                            if (!related) {
                                return 0 /* False */;
                            }
                            result &= related;
                        }
                    }
                }
                var targetLen = getParameterCount(target);
                for (var i = 0; i < targetLen; i++) {
                    var s = getTypeAtPosition(source, i);
                    var t = getTypeAtPosition(target, i);
                    var related = compareTypes(t, s);
                    if (!related) {
                        return 0 /* False */;
                    }
                    result &= related;
                }
                if (!ignoreReturnTypes) {
                    var sourceTypePredicate = getTypePredicateOfSignature(source);
                    var targetTypePredicate = getTypePredicateOfSignature(target);
                    result &= sourceTypePredicate || targetTypePredicate ?
                        compareTypePredicatesIdentical(sourceTypePredicate, targetTypePredicate, compareTypes) :
                        compareTypes(getReturnTypeOfSignature(source), getReturnTypeOfSignature(target));
                }
                return result;
            }
            function compareTypePredicatesIdentical(source, target, compareTypes) {
                return !(source && target && typePredicateKindsMatch(source, target)) ? 0 /* False */ :
                    source.type === target.type ? -1 /* True */ :
                        source.type && target.type ? compareTypes(source.type, target.type) :
                            0 /* False */;
            }
            function literalTypesWithSameBaseType(types) {
                var commonBaseType;
                for (var _i = 0, types_13 = types; _i < types_13.length; _i++) {
                    var t = types_13[_i];
                    var baseType = getBaseTypeOfLiteralType(t);
                    if (!commonBaseType) {
                        commonBaseType = baseType;
                    }
                    if (baseType === t || baseType !== commonBaseType) {
                        return false;
                    }
                }
                return true;
            }
            // When the candidate types are all literal types with the same base type, return a union
            // of those literal types. Otherwise, return the leftmost type for which no type to the
            // right is a supertype.
            function getSupertypeOrUnion(types) {
                if (types.length === 1) {
                    return types[0];
                }
                return literalTypesWithSameBaseType(types) ?
                    getUnionType(types) :
                    ts.reduceLeft(types, function (s, t) { return isTypeSubtypeOf(s, t) ? t : s; });
            }
            function getCommonSupertype(types) {
                if (!strictNullChecks) {
                    return getSupertypeOrUnion(types);
                }
                var primaryTypes = ts.filter(types, function (t) { return !(t.flags & 98304 /* Nullable */); });
                return primaryTypes.length ?
                    getNullableType(getSupertypeOrUnion(primaryTypes), getFalsyFlagsOfTypes(types) & 98304 /* Nullable */) :
                    getUnionType(types, 2 /* Subtype */);
            }
            // Return the leftmost type for which no type to the right is a subtype.
            function getCommonSubtype(types) {
                return ts.reduceLeft(types, function (s, t) { return isTypeSubtypeOf(t, s) ? t : s; });
            }
            function isArrayType(type) {
                return !!(ts.getObjectFlags(type) & 4 /* Reference */) && (type.target === globalArrayType || type.target === globalReadonlyArrayType);
            }
            function isReadonlyArrayType(type) {
                return !!(ts.getObjectFlags(type) & 4 /* Reference */) && type.target === globalReadonlyArrayType;
            }
            function isMutableArrayOrTuple(type) {
                return isArrayType(type) && !isReadonlyArrayType(type) || isTupleType(type) && !type.target.readonly;
            }
            function getElementTypeOfArrayType(type) {
                return isArrayType(type) ? getTypeArguments(type)[0] : undefined;
            }
            function isArrayLikeType(type) {
                // A type is array-like if it is a reference to the global Array or global ReadonlyArray type,
                // or if it is not the undefined or null type and if it is assignable to ReadonlyArray<any>
                return isArrayType(type) || !(type.flags & 98304 /* Nullable */) && isTypeAssignableTo(type, anyReadonlyArrayType);
            }
            function getSingleBaseForNonAugmentingSubtype(type) {
                if (!(ts.getObjectFlags(type) & 4 /* Reference */) || !(ts.getObjectFlags(type.target) & 3 /* ClassOrInterface */)) {
                    return undefined;
                }
                if (ts.getObjectFlags(type) & 33554432 /* IdenticalBaseTypeCalculated */) {
                    return ts.getObjectFlags(type) & 67108864 /* IdenticalBaseTypeExists */ ? type.cachedEquivalentBaseType : undefined;
                }
                type.objectFlags |= 33554432 /* IdenticalBaseTypeCalculated */;
                var target = type.target;
                if (ts.getObjectFlags(target) & 1 /* Class */) {
                    var baseTypeNode = getBaseTypeNodeOfClass(target);
                    // A base type expression may circularly reference the class itself (e.g. as an argument to function call), so we only
                    // check for base types specified as simple qualified names.
                    if (baseTypeNode && baseTypeNode.expression.kind !== 79 /* Identifier */ && baseTypeNode.expression.kind !== 204 /* PropertyAccessExpression */) {
                        return undefined;
                    }
                }
                var bases = getBaseTypes(target);
                if (bases.length !== 1) {
                    return undefined;
                }
                if (getMembersOfSymbol(type.symbol).size) {
                    return undefined; // If the interface has any members, they may subtype members in the base, so we should do a full structural comparison
                }
                var instantiatedBase = !ts.length(target.typeParameters) ? bases[0] : instantiateType(bases[0], createTypeMapper(target.typeParameters, getTypeArguments(type).slice(0, target.typeParameters.length)));
                if (ts.length(getTypeArguments(type)) > ts.length(target.typeParameters)) {
                    instantiatedBase = getTypeWithThisArgument(instantiatedBase, ts.last(getTypeArguments(type)));
                }
                type.objectFlags |= 67108864 /* IdenticalBaseTypeExists */;
                return type.cachedEquivalentBaseType = instantiatedBase;
            }
            function isEmptyLiteralType(type) {
                return strictNullChecks ? type === implicitNeverType : type === undefinedWideningType;
            }
            function isEmptyArrayLiteralType(type) {
                var elementType = getElementTypeOfArrayType(type);
                return !!elementType && isEmptyLiteralType(elementType);
            }
            function isTupleLikeType(type) {
                return isTupleType(type) || !!getPropertyOfType(type, "0");
            }
            function isArrayOrTupleLikeType(type) {
                return isArrayLikeType(type) || isTupleLikeType(type);
            }
            function getTupleElementType(type, index) {
                var propType = getTypeOfPropertyOfType(type, "" + index);
                if (propType) {
                    return propType;
                }
                if (everyType(type, isTupleType)) {
                    return mapType(type, function (t) { return getRestTypeOfTupleType(t) || undefinedType; });
                }
                return undefined;
            }
            function isNeitherUnitTypeNorNever(type) {
                return !(type.flags & (109440 /* Unit */ | 131072 /* Never */));
            }
            function isUnitType(type) {
                return !!(type.flags & 109440 /* Unit */);
            }
            function isUnitLikeType(type) {
                return type.flags & 2097152 /* Intersection */ ? ts.some(type.types, isUnitType) :
                    !!(type.flags & 109440 /* Unit */);
            }
            function extractUnitType(type) {
                return type.flags & 2097152 /* Intersection */ ? ts.find(type.types, isUnitType) || type : type;
            }
            function isLiteralType(type) {
                return type.flags & 16 /* Boolean */ ? true :
                    type.flags & 1048576 /* Union */ ? type.flags & 1024 /* EnumLiteral */ ? true : ts.every(type.types, isUnitType) :
                        isUnitType(type);
            }
            function getBaseTypeOfLiteralType(type) {
                return type.flags & 1024 /* EnumLiteral */ ? getBaseTypeOfEnumLiteralType(type) :
                    type.flags & 128 /* StringLiteral */ ? stringType :
                        type.flags & 256 /* NumberLiteral */ ? numberType :
                            type.flags & 2048 /* BigIntLiteral */ ? bigintType :
                                type.flags & 512 /* BooleanLiteral */ ? booleanType :
                                    type.flags & 1048576 /* Union */ ? mapType(type, getBaseTypeOfLiteralType) :
                                        type;
            }
            function getWidenedLiteralType(type) {
                return type.flags & 1024 /* EnumLiteral */ && isFreshLiteralType(type) ? getBaseTypeOfEnumLiteralType(type) :
                    type.flags & 128 /* StringLiteral */ && isFreshLiteralType(type) ? stringType :
                        type.flags & 256 /* NumberLiteral */ && isFreshLiteralType(type) ? numberType :
                            type.flags & 2048 /* BigIntLiteral */ && isFreshLiteralType(type) ? bigintType :
                                type.flags & 512 /* BooleanLiteral */ && isFreshLiteralType(type) ? booleanType :
                                    type.flags & 1048576 /* Union */ ? mapType(type, getWidenedLiteralType) :
                                        type;
            }
            function getWidenedUniqueESSymbolType(type) {
                return type.flags & 8192 /* UniqueESSymbol */ ? esSymbolType :
                    type.flags & 1048576 /* Union */ ? mapType(type, getWidenedUniqueESSymbolType) :
                        type;
            }
            function getWidenedLiteralLikeTypeForContextualType(type, contextualType) {
                if (!isLiteralOfContextualType(type, contextualType)) {
                    type = getWidenedUniqueESSymbolType(getWidenedLiteralType(type));
                }
                return type;
            }
            function getWidenedLiteralLikeTypeForContextualReturnTypeIfNeeded(type, contextualSignatureReturnType, isAsync) {
                if (type && isUnitType(type)) {
                    var contextualType = !contextualSignatureReturnType ? undefined :
                        isAsync ? getPromisedTypeOfPromise(contextualSignatureReturnType) :
                            contextualSignatureReturnType;
                    type = getWidenedLiteralLikeTypeForContextualType(type, contextualType);
                }
                return type;
            }
            function getWidenedLiteralLikeTypeForContextualIterationTypeIfNeeded(type, contextualSignatureReturnType, kind, isAsyncGenerator) {
                if (type && isUnitType(type)) {
                    var contextualType = !contextualSignatureReturnType ? undefined :
                        getIterationTypeOfGeneratorFunctionReturnType(kind, contextualSignatureReturnType, isAsyncGenerator);
                    type = getWidenedLiteralLikeTypeForContextualType(type, contextualType);
                }
                return type;
            }
            /**
             * Check if a Type was written as a tuple type literal.
             * Prefer using isTupleLikeType() unless the use of `elementTypes`/`getTypeArguments` is required.
             */
            function isTupleType(type) {
                return !!(ts.getObjectFlags(type) & 4 /* Reference */ && type.target.objectFlags & 8 /* Tuple */);
            }
            function isGenericTupleType(type) {
                return isTupleType(type) && !!(type.target.combinedFlags & 8 /* Variadic */);
            }
            function isSingleElementGenericTupleType(type) {
                return isGenericTupleType(type) && type.target.elementFlags.length === 1;
            }
            function getRestTypeOfTupleType(type) {
                return getElementTypeOfSliceOfTupleType(type, type.target.fixedLength);
            }
            function getRestArrayTypeOfTupleType(type) {
                var restType = getRestTypeOfTupleType(type);
                return restType && createArrayType(restType);
            }
            function getElementTypeOfSliceOfTupleType(type, index, endSkipCount, writing) {
                if (endSkipCount === void 0) {
                    endSkipCount = 0;
                }
                if (writing === void 0) {
                    writing = false;
                }
                var length = getTypeReferenceArity(type) - endSkipCount;
                if (index < length) {
                    var typeArguments = getTypeArguments(type);
                    var elementTypes = [];
                    for (var i = index; i < length; i++) {
                        var t = typeArguments[i];
                        elementTypes.push(type.target.elementFlags[i] & 8 /* Variadic */ ? getIndexedAccessType(t, numberType) : t);
                    }
                    return writing ? getIntersectionType(elementTypes) : getUnionType(elementTypes);
                }
                return undefined;
            }
            function isTupleTypeStructureMatching(t1, t2) {
                return getTypeReferenceArity(t1) === getTypeReferenceArity(t2) &&
                    ts.every(t1.target.elementFlags, function (f, i) { return (f & 12 /* Variable */) === (t2.target.elementFlags[i] & 12 /* Variable */); });
            }
            function isZeroBigInt(_a) {
                var value = _a.value;
                return value.base10Value === "0";
            }
            function getFalsyFlagsOfTypes(types) {
                var result = 0;
                for (var _i = 0, types_14 = types; _i < types_14.length; _i++) {
                    var t = types_14[_i];
                    result |= getFalsyFlags(t);
                }
                return result;
            }
            // Returns the String, Number, Boolean, StringLiteral, NumberLiteral, BooleanLiteral, Void, Undefined, or Null
            // flags for the string, number, boolean, "", 0, false, void, undefined, or null types respectively. Returns
            // no flags for all other types (including non-falsy literal types).
            function getFalsyFlags(type) {
                return type.flags & 1048576 /* Union */ ? getFalsyFlagsOfTypes(type.types) :
                    type.flags & 128 /* StringLiteral */ ? type.value === "" ? 128 /* StringLiteral */ : 0 :
                        type.flags & 256 /* NumberLiteral */ ? type.value === 0 ? 256 /* NumberLiteral */ : 0 :
                            type.flags & 2048 /* BigIntLiteral */ ? isZeroBigInt(type) ? 2048 /* BigIntLiteral */ : 0 :
                                type.flags & 512 /* BooleanLiteral */ ? (type === falseType || type === regularFalseType) ? 512 /* BooleanLiteral */ : 0 :
                                    type.flags & 117724 /* PossiblyFalsy */;
            }
            function removeDefinitelyFalsyTypes(type) {
                return getFalsyFlags(type) & 117632 /* DefinitelyFalsy */ ?
                    filterType(type, function (t) { return !(getFalsyFlags(t) & 117632 /* DefinitelyFalsy */); }) :
                    type;
            }
            function extractDefinitelyFalsyTypes(type) {
                return mapType(type, getDefinitelyFalsyPartOfType);
            }
            function getDefinitelyFalsyPartOfType(type) {
                return type.flags & 4 /* String */ ? emptyStringType :
                    type.flags & 8 /* Number */ ? zeroType :
                        type.flags & 64 /* BigInt */ ? zeroBigIntType :
                            type === regularFalseType ||
                                type === falseType ||
                                type.flags & (16384 /* Void */ | 32768 /* Undefined */ | 65536 /* Null */ | 3 /* AnyOrUnknown */) ||
                                type.flags & 128 /* StringLiteral */ && type.value === "" ||
                                type.flags & 256 /* NumberLiteral */ && type.value === 0 ||
                                type.flags & 2048 /* BigIntLiteral */ && isZeroBigInt(type) ? type :
                                neverType;
            }
            /**
             * Add undefined or null or both to a type if they are missing.
             * @param type - type to add undefined and/or null to if not present
             * @param flags - Either TypeFlags.Undefined or TypeFlags.Null, or both
             */
            function getNullableType(type, flags) {
                var missing = (flags & ~type.flags) & (32768 /* Undefined */ | 65536 /* Null */);
                return missing === 0 ? type :
                    missing === 32768 /* Undefined */ ? getUnionType([type, undefinedType]) :
                        missing === 65536 /* Null */ ? getUnionType([type, nullType]) :
                            getUnionType([type, undefinedType, nullType]);
            }
            function getOptionalType(type, isProperty) {
                if (isProperty === void 0) {
                    isProperty = false;
                }
                ts.Debug.assert(strictNullChecks);
                return type.flags & 32768 /* Undefined */ ? type : getUnionType([type, isProperty ? missingType : undefinedType]);
            }
            function getGlobalNonNullableTypeInstantiation(type) {
                // First reduce away any constituents that are assignable to 'undefined' or 'null'. This not only eliminates
                // 'undefined' and 'null', but also higher-order types such as a type parameter 'U extends undefined | null'
                // that isn't eliminated by a NonNullable<T> instantiation.
                var reducedType = getTypeWithFacts(type, 2097152 /* NEUndefinedOrNull */);
                if (!deferredGlobalNonNullableTypeAlias) {
                    deferredGlobalNonNullableTypeAlias = getGlobalSymbol("NonNullable", 524288 /* TypeAlias */, /*diagnostic*/ undefined) || unknownSymbol;
                }
                // If the NonNullable<T> type is available, return an instantiation. Otherwise just return the reduced type.
                return deferredGlobalNonNullableTypeAlias !== unknownSymbol ?
                    getTypeAliasInstantiation(deferredGlobalNonNullableTypeAlias, [reducedType]) :
                    reducedType;
            }
            function getNonNullableType(type) {
                return strictNullChecks ? getGlobalNonNullableTypeInstantiation(type) : type;
            }
            function addOptionalTypeMarker(type) {
                return strictNullChecks ? getUnionType([type, optionalType]) : type;
            }
            function removeOptionalTypeMarker(type) {
                return strictNullChecks ? removeType(type, optionalType) : type;
            }
            function propagateOptionalTypeMarker(type, node, wasOptional) {
                return wasOptional ? ts.isOutermostOptionalChain(node) ? getOptionalType(type) : addOptionalTypeMarker(type) : type;
            }
            function getOptionalExpressionType(exprType, expression) {
                return ts.isExpressionOfOptionalChainRoot(expression) ? getNonNullableType(exprType) :
                    ts.isOptionalChain(expression) ? removeOptionalTypeMarker(exprType) :
                        exprType;
            }
            function removeMissingType(type, isOptional) {
                return exactOptionalPropertyTypes && isOptional ? removeType(type, missingType) : type;
            }
            function containsMissingType(type) {
                return exactOptionalPropertyTypes && (type === missingType || type.flags & 1048576 /* Union */ && containsType(type.types, missingType));
            }
            function removeMissingOrUndefinedType(type) {
                return exactOptionalPropertyTypes ? removeType(type, missingType) : getTypeWithFacts(type, 524288 /* NEUndefined */);
            }
            /**
             * Is source potentially coercible to target type under `==`.
             * Assumes that `source` is a constituent of a union, hence
             * the boolean literal flag on the LHS, but not on the RHS.
             *
             * This does not fully replicate the semantics of `==`. The
             * intention is to catch cases that are clearly not right.
             *
             * Comparing (string | number) to number should not remove the
             * string element.
             *
             * Comparing (string | number) to 1 will remove the string
             * element, though this is not sound. This is a pragmatic
             * choice.
             *
             * @see narrowTypeByEquality
             *
             * @param source
             * @param target
             */
            function isCoercibleUnderDoubleEquals(source, target) {
                return ((source.flags & (8 /* Number */ | 4 /* String */ | 512 /* BooleanLiteral */)) !== 0)
                    && ((target.flags & (8 /* Number */ | 4 /* String */ | 16 /* Boolean */)) !== 0);
            }
            /**
             * Return true if type was inferred from an object literal, written as an object type literal, or is the shape of a module
             * with no call or construct signatures.
             */
            function isObjectTypeWithInferableIndex(type) {
                return type.flags & 2097152 /* Intersection */ ? ts.every(type.types, isObjectTypeWithInferableIndex) :
                    !!(type.symbol && (type.symbol.flags & (4096 /* ObjectLiteral */ | 2048 /* TypeLiteral */ | 384 /* Enum */ | 512 /* ValueModule */)) !== 0 &&
                        !typeHasCallOrConstructSignatures(type)) || !!(ts.getObjectFlags(type) & 1024 /* ReverseMapped */ && isObjectTypeWithInferableIndex(type.source));
            }
            function createSymbolWithType(source, type) {
                var symbol = createSymbol(source.flags, source.escapedName, ts.getCheckFlags(source) & 8 /* Readonly */);
                symbol.declarations = source.declarations;
                symbol.parent = source.parent;
                symbol.type = type;
                symbol.target = source;
                if (source.valueDeclaration) {
                    symbol.valueDeclaration = source.valueDeclaration;
                }
                var nameType = getSymbolLinks(source).nameType;
                if (nameType) {
                    symbol.nameType = nameType;
                }
                return symbol;
            }
            function transformTypeOfMembers(type, f) {
                var members = ts.createSymbolTable();
                for (var _i = 0, _a = getPropertiesOfObjectType(type); _i < _a.length; _i++) {
                    var property = _a[_i];
                    var original = getTypeOfSymbol(property);
                    var updated = f(original);
                    members.set(property.escapedName, updated === original ? property : createSymbolWithType(property, updated));
                }
                return members;
            }
            /**
             * If the the provided object literal is subject to the excess properties check,
             * create a new that is exempt. Recursively mark object literal members as exempt.
             * Leave signatures alone since they are not subject to the check.
             */
            function getRegularTypeOfObjectLiteral(type) {
                if (!(isObjectLiteralType(type) && ts.getObjectFlags(type) & 16384 /* FreshLiteral */)) {
                    return type;
                }
                var regularType = type.regularType;
                if (regularType) {
                    return regularType;
                }
                var resolved = type;
                var members = transformTypeOfMembers(type, getRegularTypeOfObjectLiteral);
                var regularNew = createAnonymousType(resolved.symbol, members, resolved.callSignatures, resolved.constructSignatures, resolved.indexInfos);
                regularNew.flags = resolved.flags;
                regularNew.objectFlags |= resolved.objectFlags & ~16384 /* FreshLiteral */;
                type.regularType = regularNew;
                return regularNew;
            }
            function createWideningContext(parent, propertyName, siblings) {
                return { parent: parent, propertyName: propertyName, siblings: siblings, resolvedProperties: undefined };
            }
            function getSiblingsOfContext(context) {
                if (!context.siblings) {
                    var siblings_1 = [];
                    for (var _i = 0, _a = getSiblingsOfContext(context.parent); _i < _a.length; _i++) {
                        var type = _a[_i];
                        if (isObjectLiteralType(type)) {
                            var prop = getPropertyOfObjectType(type, context.propertyName);
                            if (prop) {
                                forEachType(getTypeOfSymbol(prop), function (t) {
                                    siblings_1.push(t);
                                });
                            }
                        }
                    }
                    context.siblings = siblings_1;
                }
                return context.siblings;
            }
            function getPropertiesOfContext(context) {
                if (!context.resolvedProperties) {
                    var names = new ts.Map();
                    for (var _i = 0, _a = getSiblingsOfContext(context); _i < _a.length; _i++) {
                        var t = _a[_i];
                        if (isObjectLiteralType(t) && !(ts.getObjectFlags(t) & 4194304 /* ContainsSpread */)) {
                            for (var _b = 0, _c = getPropertiesOfType(t); _b < _c.length; _b++) {
                                var prop = _c[_b];
                                names.set(prop.escapedName, prop);
                            }
                        }
                    }
                    context.resolvedProperties = ts.arrayFrom(names.values());
                }
                return context.resolvedProperties;
            }
            function getWidenedProperty(prop, context) {
                if (!(prop.flags & 4 /* Property */)) {
                    // Since get accessors already widen their return value there is no need to
                    // widen accessor based properties here.
                    return prop;
                }
                var original = getTypeOfSymbol(prop);
                var propContext = context && createWideningContext(context, prop.escapedName, /*siblings*/ undefined);
                var widened = getWidenedTypeWithContext(original, propContext);
                return widened === original ? prop : createSymbolWithType(prop, widened);
            }
            function getUndefinedProperty(prop) {
                var cached = undefinedProperties.get(prop.escapedName);
                if (cached) {
                    return cached;
                }
                var result = createSymbolWithType(prop, missingType);
                result.flags |= 16777216 /* Optional */;
                undefinedProperties.set(prop.escapedName, result);
                return result;
            }
            function getWidenedTypeOfObjectLiteral(type, context) {
                var members = ts.createSymbolTable();
                for (var _i = 0, _a = getPropertiesOfObjectType(type); _i < _a.length; _i++) {
                    var prop = _a[_i];
                    members.set(prop.escapedName, getWidenedProperty(prop, context));
                }
                if (context) {
                    for (var _b = 0, _c = getPropertiesOfContext(context); _b < _c.length; _b++) {
                        var prop = _c[_b];
                        if (!members.has(prop.escapedName)) {
                            members.set(prop.escapedName, getUndefinedProperty(prop));
                        }
                    }
                }
                var result = createAnonymousType(type.symbol, members, ts.emptyArray, ts.emptyArray, ts.sameMap(getIndexInfosOfType(type), function (info) { return createIndexInfo(info.keyType, getWidenedType(info.type), info.isReadonly); }));
                result.objectFlags |= (ts.getObjectFlags(type) & (8192 /* JSLiteral */ | 524288 /* NonInferrableType */)); // Retain js literal flag through widening
                return result;
            }
            function getWidenedType(type) {
                return getWidenedTypeWithContext(type, /*context*/ undefined);
            }
            function getWidenedTypeWithContext(type, context) {
                if (ts.getObjectFlags(type) & 393216 /* RequiresWidening */) {
                    if (context === undefined && type.widened) {
                        return type.widened;
                    }
                    var result = void 0;
                    if (type.flags & (1 /* Any */ | 98304 /* Nullable */)) {
                        result = anyType;
                    }
                    else if (isObjectLiteralType(type)) {
                        result = getWidenedTypeOfObjectLiteral(type, context);
                    }
                    else if (type.flags & 1048576 /* Union */) {
                        var unionContext_1 = context || createWideningContext(/*parent*/ undefined, /*propertyName*/ undefined, type.types);
                        var widenedTypes = ts.sameMap(type.types, function (t) { return t.flags & 98304 /* Nullable */ ? t : getWidenedTypeWithContext(t, unionContext_1); });
                        // Widening an empty object literal transitions from a highly restrictive type to
                        // a highly inclusive one. For that reason we perform subtype reduction here if the
                        // union includes empty object types (e.g. reducing {} | string to just {}).
                        result = getUnionType(widenedTypes, ts.some(widenedTypes, isEmptyObjectType) ? 2 /* Subtype */ : 1 /* Literal */);
                    }
                    else if (type.flags & 2097152 /* Intersection */) {
                        result = getIntersectionType(ts.sameMap(type.types, getWidenedType));
                    }
                    else if (isArrayType(type) || isTupleType(type)) {
                        result = createTypeReference(type.target, ts.sameMap(getTypeArguments(type), getWidenedType));
                    }
                    if (result && context === undefined) {
                        type.widened = result;
                    }
                    return result || type;
                }
                return type;
            }
            /**
             * Reports implicit any errors that occur as a result of widening 'null' and 'undefined'
             * to 'any'. A call to reportWideningErrorsInType is normally accompanied by a call to
             * getWidenedType. But in some cases getWidenedType is called without reporting errors
             * (type argument inference is an example).
             *
             * The return value indicates whether an error was in fact reported. The particular circumstances
             * are on a best effort basis. Currently, if the null or undefined that causes widening is inside
             * an object literal property (arbitrarily deeply), this function reports an error. If no error is
             * reported, reportImplicitAnyError is a suitable fallback to report a general error.
             */
            function reportWideningErrorsInType(type) {
                var errorReported = false;
                if (ts.getObjectFlags(type) & 131072 /* ContainsWideningType */) {
                    if (type.flags & 1048576 /* Union */) {
                        if (ts.some(type.types, isEmptyObjectType)) {
                            errorReported = true;
                        }
                        else {
                            for (var _i = 0, _a = type.types; _i < _a.length; _i++) {
                                var t = _a[_i];
                                if (reportWideningErrorsInType(t)) {
                                    errorReported = true;
                                }
                            }
                        }
                    }
                    if (isArrayType(type) || isTupleType(type)) {
                        for (var _b = 0, _c = getTypeArguments(type); _b < _c.length; _b++) {
                            var t = _c[_b];
                            if (reportWideningErrorsInType(t)) {
                                errorReported = true;
                            }
                        }
                    }
                    if (isObjectLiteralType(type)) {
                        for (var _d = 0, _e = getPropertiesOfObjectType(type); _d < _e.length; _d++) {
                            var p = _e[_d];
                            var t = getTypeOfSymbol(p);
                            if (ts.getObjectFlags(t) & 131072 /* ContainsWideningType */) {
                                if (!reportWideningErrorsInType(t)) {
                                    error(p.valueDeclaration, ts.Diagnostics.Object_literal_s_property_0_implicitly_has_an_1_type, symbolToString(p), typeToString(getWidenedType(t)));
                                }
                                errorReported = true;
                            }
                        }
                    }
                }
                return errorReported;
            }
            function reportImplicitAny(declaration, type, wideningKind) {
                var typeAsString = typeToString(getWidenedType(type));
                if (ts.isInJSFile(declaration) && !ts.isCheckJsEnabledForFile(ts.getSourceFileOfNode(declaration), compilerOptions)) {
                    // Only report implicit any errors/suggestions in TS and ts-check JS files
                    return;
                }
                var diagnostic;
                switch (declaration.kind) {
                    case 219 /* BinaryExpression */:
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                        diagnostic = noImplicitAny ? ts.Diagnostics.Member_0_implicitly_has_an_1_type : ts.Diagnostics.Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage;
                        break;
                    case 162 /* Parameter */:
                        var param = declaration;
                        if (ts.isIdentifier(param.name) &&
                            (ts.isCallSignatureDeclaration(param.parent) || ts.isMethodSignature(param.parent) || ts.isFunctionTypeNode(param.parent)) &&
                            param.parent.parameters.indexOf(param) > -1 &&
                            (resolveName(param, param.name.escapedText, 788968 /* Type */, undefined, param.name.escapedText, /*isUse*/ true) ||
                                param.name.originalKeywordKind && ts.isTypeNodeKind(param.name.originalKeywordKind))) {
                            var newName = "arg" + param.parent.parameters.indexOf(param);
                            errorOrSuggestion(noImplicitAny, declaration, ts.Diagnostics.Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1, newName, ts.declarationNameToString(param.name));
                            return;
                        }
                        diagnostic = declaration.dotDotDotToken ?
                            noImplicitAny ? ts.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type : ts.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage :
                            noImplicitAny ? ts.Diagnostics.Parameter_0_implicitly_has_an_1_type : ts.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage;
                        break;
                    case 201 /* BindingElement */:
                        diagnostic = ts.Diagnostics.Binding_element_0_implicitly_has_an_1_type;
                        if (!noImplicitAny) {
                            // Don't issue a suggestion for binding elements since the codefix doesn't yet support them.
                            return;
                        }
                        break;
                    case 312 /* JSDocFunctionType */:
                        error(declaration, ts.Diagnostics.Function_type_which_lacks_return_type_annotation_implicitly_has_an_0_return_type, typeAsString);
                        return;
                    case 254 /* FunctionDeclaration */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        if (noImplicitAny && !declaration.name) {
                            if (wideningKind === 3 /* GeneratorYield */) {
                                error(declaration, ts.Diagnostics.Generator_implicitly_has_yield_type_0_because_it_does_not_yield_any_values_Consider_supplying_a_return_type_annotation, typeAsString);
                            }
                            else {
                                error(declaration, ts.Diagnostics.Function_expression_which_lacks_return_type_annotation_implicitly_has_an_0_return_type, typeAsString);
                            }
                            return;
                        }
                        diagnostic = !noImplicitAny ? ts.Diagnostics._0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage :
                            wideningKind === 3 /* GeneratorYield */ ? ts.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_yield_type :
                                ts.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type;
                        break;
                    case 193 /* MappedType */:
                        if (noImplicitAny) {
                            error(declaration, ts.Diagnostics.Mapped_object_type_implicitly_has_an_any_template_type);
                        }
                        return;
                    default:
                        diagnostic = noImplicitAny ? ts.Diagnostics.Variable_0_implicitly_has_an_1_type : ts.Diagnostics.Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage;
                }
                errorOrSuggestion(noImplicitAny, declaration, diagnostic, ts.declarationNameToString(ts.getNameOfDeclaration(declaration)), typeAsString);
            }
            function reportErrorsFromWidening(declaration, type, wideningKind) {
                if (produceDiagnostics && noImplicitAny && ts.getObjectFlags(type) & 131072 /* ContainsWideningType */ && (!wideningKind || !getContextualSignatureForFunctionLikeDeclaration(declaration))) {
                    // Report implicit any error within type if possible, otherwise report error on declaration
                    if (!reportWideningErrorsInType(type)) {
                        reportImplicitAny(declaration, type, wideningKind);
                    }
                }
            }
            function applyToParameterTypes(source, target, callback) {
                var sourceCount = getParameterCount(source);
                var targetCount = getParameterCount(target);
                var sourceRestType = getEffectiveRestType(source);
                var targetRestType = getEffectiveRestType(target);
                var targetNonRestCount = targetRestType ? targetCount - 1 : targetCount;
                var paramCount = sourceRestType ? targetNonRestCount : Math.min(sourceCount, targetNonRestCount);
                var sourceThisType = getThisTypeOfSignature(source);
                if (sourceThisType) {
                    var targetThisType = getThisTypeOfSignature(target);
                    if (targetThisType) {
                        callback(sourceThisType, targetThisType);
                    }
                }
                for (var i = 0; i < paramCount; i++) {
                    callback(getTypeAtPosition(source, i), getTypeAtPosition(target, i));
                }
                if (targetRestType) {
                    callback(getRestTypeAtPosition(source, paramCount), targetRestType);
                }
            }
            function applyToReturnTypes(source, target, callback) {
                var sourceTypePredicate = getTypePredicateOfSignature(source);
                var targetTypePredicate = getTypePredicateOfSignature(target);
                if (sourceTypePredicate && targetTypePredicate && typePredicateKindsMatch(sourceTypePredicate, targetTypePredicate) && sourceTypePredicate.type && targetTypePredicate.type) {
                    callback(sourceTypePredicate.type, targetTypePredicate.type);
                }
                else {
                    callback(getReturnTypeOfSignature(source), getReturnTypeOfSignature(target));
                }
            }
            function createInferenceContext(typeParameters, signature, flags, compareTypes) {
                return createInferenceContextWorker(typeParameters.map(createInferenceInfo), signature, flags, compareTypes || compareTypesAssignable);
            }
            function cloneInferenceContext(context, extraFlags) {
                if (extraFlags === void 0) {
                    extraFlags = 0;
                }
                return context && createInferenceContextWorker(ts.map(context.inferences, cloneInferenceInfo), context.signature, context.flags | extraFlags, context.compareTypes);
            }
            function createInferenceContextWorker(inferences, signature, flags, compareTypes) {
                var context = {
                    inferences: inferences,
                    signature: signature,
                    flags: flags,
                    compareTypes: compareTypes,
                    mapper: makeFunctionTypeMapper(function (t) { return mapToInferredType(context, t, /*fix*/ true); }),
                    nonFixingMapper: makeFunctionTypeMapper(function (t) { return mapToInferredType(context, t, /*fix*/ false); }),
                };
                return context;
            }
            function mapToInferredType(context, t, fix) {
                var inferences = context.inferences;
                for (var i = 0; i < inferences.length; i++) {
                    var inference = inferences[i];
                    if (t === inference.typeParameter) {
                        if (fix && !inference.isFixed) {
                            clearCachedInferences(inferences);
                            inference.isFixed = true;
                        }
                        return getInferredType(context, i);
                    }
                }
                return t;
            }
            function clearCachedInferences(inferences) {
                for (var _i = 0, inferences_1 = inferences; _i < inferences_1.length; _i++) {
                    var inference = inferences_1[_i];
                    if (!inference.isFixed) {
                        inference.inferredType = undefined;
                    }
                }
            }
            function createInferenceInfo(typeParameter) {
                return {
                    typeParameter: typeParameter,
                    candidates: undefined,
                    contraCandidates: undefined,
                    inferredType: undefined,
                    priority: undefined,
                    topLevel: true,
                    isFixed: false,
                    impliedArity: undefined
                };
            }
            function cloneInferenceInfo(inference) {
                return {
                    typeParameter: inference.typeParameter,
                    candidates: inference.candidates && inference.candidates.slice(),
                    contraCandidates: inference.contraCandidates && inference.contraCandidates.slice(),
                    inferredType: inference.inferredType,
                    priority: inference.priority,
                    topLevel: inference.topLevel,
                    isFixed: inference.isFixed,
                    impliedArity: inference.impliedArity
                };
            }
            function cloneInferredPartOfContext(context) {
                var inferences = ts.filter(context.inferences, hasInferenceCandidates);
                return inferences.length ?
                    createInferenceContextWorker(ts.map(inferences, cloneInferenceInfo), context.signature, context.flags, context.compareTypes) :
                    undefined;
            }
            function getMapperFromContext(context) {
                return context && context.mapper;
            }
            // Return true if the given type could possibly reference a type parameter for which
            // we perform type inference (i.e. a type parameter of a generic function). We cache
            // results for union and intersection types for performance reasons.
            function couldContainTypeVariables(type) {
                var objectFlags = ts.getObjectFlags(type);
                if (objectFlags & 1048576 /* CouldContainTypeVariablesComputed */) {
                    return !!(objectFlags & 2097152 /* CouldContainTypeVariables */);
                }
                var result = !!(type.flags & 465829888 /* Instantiable */ ||
                    type.flags & 524288 /* Object */ && !isNonGenericTopLevelType(type) && (objectFlags & 4 /* Reference */ && (type.node || ts.forEach(getTypeArguments(type), couldContainTypeVariables)) ||
                        objectFlags & 16 /* Anonymous */ && type.symbol && type.symbol.flags & (16 /* Function */ | 8192 /* Method */ | 32 /* Class */ | 2048 /* TypeLiteral */ | 4096 /* ObjectLiteral */) && type.symbol.declarations ||
                        objectFlags & (32 /* Mapped */ | 1024 /* ReverseMapped */ | 8388608 /* ObjectRestType */)) ||
                    type.flags & 3145728 /* UnionOrIntersection */ && !(type.flags & 1024 /* EnumLiteral */) && !isNonGenericTopLevelType(type) && ts.some(type.types, couldContainTypeVariables));
                if (type.flags & 3899393 /* ObjectFlagsType */) {
                    type.objectFlags |= 1048576 /* CouldContainTypeVariablesComputed */ | (result ? 2097152 /* CouldContainTypeVariables */ : 0);
                }
                return result;
            }
            function isNonGenericTopLevelType(type) {
                if (type.aliasSymbol && !type.aliasTypeArguments) {
                    var declaration = ts.getDeclarationOfKind(type.aliasSymbol, 257 /* TypeAliasDeclaration */);
                    return !!(declaration && ts.findAncestor(declaration.parent, function (n) { return n.kind === 300 /* SourceFile */ ? true : n.kind === 259 /* ModuleDeclaration */ ? false : "quit"; }));
                }
                return false;
            }
            function isTypeParameterAtTopLevel(type, typeParameter) {
                return !!(type === typeParameter ||
                    type.flags & 3145728 /* UnionOrIntersection */ && ts.some(type.types, function (t) { return isTypeParameterAtTopLevel(t, typeParameter); }) ||
                    type.flags & 16777216 /* Conditional */ && (getTrueTypeFromConditionalType(type) === typeParameter || getFalseTypeFromConditionalType(type) === typeParameter));
            }
            /** Create an object with properties named in the string literal type. Every property has type `any` */
            function createEmptyObjectTypeFromStringLiteral(type) {
                var members = ts.createSymbolTable();
                forEachType(type, function (t) {
                    if (!(t.flags & 128 /* StringLiteral */)) {
                        return;
                    }
                    var name = ts.escapeLeadingUnderscores(t.value);
                    var literalProp = createSymbol(4 /* Property */, name);
                    literalProp.type = anyType;
                    if (t.symbol) {
                        literalProp.declarations = t.symbol.declarations;
                        literalProp.valueDeclaration = t.symbol.valueDeclaration;
                    }
                    members.set(name, literalProp);
                });
                var indexInfos = type.flags & 4 /* String */ ? [createIndexInfo(stringType, emptyObjectType, /*isReadonly*/ false)] : ts.emptyArray;
                return createAnonymousType(undefined, members, ts.emptyArray, ts.emptyArray, indexInfos);
            }
            /**
             * Infer a suitable input type for a homomorphic mapped type { [P in keyof T]: X }. We construct
             * an object type with the same set of properties as the source type, where the type of each
             * property is computed by inferring from the source property type to X for the type
             * variable T[P] (i.e. we treat the type T[P] as the type variable we're inferring for).
             */
            function inferTypeForHomomorphicMappedType(source, target, constraint) {
                if (inInferTypeForHomomorphicMappedType) {
                    return undefined;
                }
                var key = source.id + "," + target.id + "," + constraint.id;
                if (reverseMappedCache.has(key)) {
                    return reverseMappedCache.get(key);
                }
                inInferTypeForHomomorphicMappedType = true;
                var type = createReverseMappedType(source, target, constraint);
                inInferTypeForHomomorphicMappedType = false;
                reverseMappedCache.set(key, type);
                return type;
            }
            // We consider a type to be partially inferable if it isn't marked non-inferable or if it is
            // an object literal type with at least one property of an inferable type. For example, an object
            // literal { a: 123, b: x => true } is marked non-inferable because it contains a context sensitive
            // arrow function, but is considered partially inferable because property 'a' has an inferable type.
            function isPartiallyInferableType(type) {
                return !(ts.getObjectFlags(type) & 524288 /* NonInferrableType */) ||
                    isObjectLiteralType(type) && ts.some(getPropertiesOfType(type), function (prop) { return isPartiallyInferableType(getTypeOfSymbol(prop)); }) ||
                    isTupleType(type) && ts.some(getTypeArguments(type), isPartiallyInferableType);
            }
            function createReverseMappedType(source, target, constraint) {
                // We consider a source type reverse mappable if it has a string index signature or if
                // it has one or more properties and is of a partially inferable type.
                if (!(getIndexInfoOfType(source, stringType) || getPropertiesOfType(source).length !== 0 && isPartiallyInferableType(source))) {
                    return undefined;
                }
                // For arrays and tuples we infer new arrays and tuples where the reverse mapping has been
                // applied to the element type(s).
                if (isArrayType(source)) {
                    return createArrayType(inferReverseMappedType(getTypeArguments(source)[0], target, constraint), isReadonlyArrayType(source));
                }
                if (isTupleType(source)) {
                    var elementTypes = ts.map(getTypeArguments(source), function (t) { return inferReverseMappedType(t, target, constraint); });
                    var elementFlags = getMappedTypeModifiers(target) & 4 /* IncludeOptional */ ?
                        ts.sameMap(source.target.elementFlags, function (f) { return f & 2 /* Optional */ ? 1 /* Required */ : f; }) :
                        source.target.elementFlags;
                    return createTupleType(elementTypes, elementFlags, source.target.readonly, source.target.labeledElementDeclarations);
                }
                // For all other object types we infer a new object type where the reverse mapping has been
                // applied to the type of each property.
                var reversed = createObjectType(1024 /* ReverseMapped */ | 16 /* Anonymous */, /*symbol*/ undefined);
                reversed.source = source;
                reversed.mappedType = target;
                reversed.constraintType = constraint;
                return reversed;
            }
            function getTypeOfReverseMappedSymbol(symbol) {
                var links = getSymbolLinks(symbol);
                if (!links.type) {
                    links.type = inferReverseMappedType(symbol.propertyType, symbol.mappedType, symbol.constraintType);
                }
                return links.type;
            }
            function inferReverseMappedType(sourceType, target, constraint) {
                var typeParameter = getIndexedAccessType(constraint.type, getTypeParameterFromMappedType(target));
                var templateType = getTemplateTypeFromMappedType(target);
                var inference = createInferenceInfo(typeParameter);
                inferTypes([inference], sourceType, templateType);
                return getTypeFromInference(inference) || unknownType;
            }
            function getUnmatchedProperties(source, target, requireOptionalProperties, matchDiscriminantProperties) {
                var properties, _i, properties_2, targetProp, sourceProp, targetType, sourceType;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            properties = getPropertiesOfType(target);
                            _i = 0, properties_2 = properties;
                            _a.label = 1;
                        case 1:
                            if (!(_i < properties_2.length))
                                return [3 /*break*/, 6];
                            targetProp = properties_2[_i];
                            // TODO: remove this when we support static private identifier fields and find other solutions to get privateNamesAndStaticFields test to pass
                            if (isStaticPrivateIdentifierProperty(targetProp)) {
                                return [3 /*break*/, 5];
                            }
                            if (!(requireOptionalProperties || !(targetProp.flags & 16777216 /* Optional */ || ts.getCheckFlags(targetProp) & 48 /* Partial */)))
                                return [3 /*break*/, 5];
                            sourceProp = getPropertyOfType(source, targetProp.escapedName);
                            if (!!sourceProp)
                                return [3 /*break*/, 3];
                            return [4 /*yield*/, targetProp];
                        case 2:
                            _a.sent();
                            return [3 /*break*/, 5];
                        case 3:
                            if (!matchDiscriminantProperties)
                                return [3 /*break*/, 5];
                            targetType = getTypeOfSymbol(targetProp);
                            if (!(targetType.flags & 109440 /* Unit */))
                                return [3 /*break*/, 5];
                            sourceType = getTypeOfSymbol(sourceProp);
                            if (!!(sourceType.flags & 1 /* Any */ || getRegularTypeOfLiteralType(sourceType) === getRegularTypeOfLiteralType(targetType)))
                                return [3 /*break*/, 5];
                            return [4 /*yield*/, targetProp];
                        case 4:
                            _a.sent();
                            _a.label = 5;
                        case 5:
                            _i++;
                            return [3 /*break*/, 1];
                        case 6: return [2 /*return*/];
                    }
                });
            }
            function getUnmatchedProperty(source, target, requireOptionalProperties, matchDiscriminantProperties) {
                var result = getUnmatchedProperties(source, target, requireOptionalProperties, matchDiscriminantProperties).next();
                if (!result.done)
                    return result.value;
            }
            function tupleTypesDefinitelyUnrelated(source, target) {
                return !(target.target.combinedFlags & 8 /* Variadic */) && target.target.minLength > source.target.minLength ||
                    !target.target.hasRestElement && (source.target.hasRestElement || target.target.fixedLength < source.target.fixedLength);
            }
            function typesDefinitelyUnrelated(source, target) {
                // Two tuple types with incompatible arities are definitely unrelated.
                // Two object types that each have a property that is unmatched in the other are definitely unrelated.
                return isTupleType(source) && isTupleType(target) ? tupleTypesDefinitelyUnrelated(source, target) :
                    !!getUnmatchedProperty(source, target, /*requireOptionalProperties*/ false, /*matchDiscriminantProperties*/ true) &&
                        !!getUnmatchedProperty(target, source, /*requireOptionalProperties*/ false, /*matchDiscriminantProperties*/ false);
            }
            function getTypeFromInference(inference) {
                return inference.candidates ? getUnionType(inference.candidates, 2 /* Subtype */) :
                    inference.contraCandidates ? getIntersectionType(inference.contraCandidates) :
                        undefined;
            }
            function hasSkipDirectInferenceFlag(node) {
                return !!getNodeLinks(node).skipDirectInference;
            }
            function isFromInferenceBlockedSource(type) {
                return !!(type.symbol && ts.some(type.symbol.declarations, hasSkipDirectInferenceFlag));
            }
            function isValidBigIntString(s) {
                var scanner = ts.createScanner(99 /* ESNext */, /*skipTrivia*/ false);
                var success = true;
                scanner.setOnError(function () { return success = false; });
                scanner.setText(s + "n");
                var result = scanner.scan();
                if (result === 40 /* MinusToken */) {
                    result = scanner.scan();
                }
                var flags = scanner.getTokenFlags();
                // validate that
                // * scanning proceeded without error
                // * a bigint can be scanned, and that when it is scanned, it is
                // * the full length of the input string (so the scanner is one character beyond the augmented input length)
                // * it does not contain a numeric seperator (the `BigInt` constructor does not accept a numeric seperator in its input)
                return success && result === 9 /* BigIntLiteral */ && scanner.getTextPos() === (s.length + 1) && !(flags & 512 /* ContainsSeparator */);
            }
            function isValidTypeForTemplateLiteralPlaceholder(source, target) {
                if (source === target || target.flags & (1 /* Any */ | 4 /* String */)) {
                    return true;
                }
                if (source.flags & 128 /* StringLiteral */) {
                    var value = source.value;
                    return !!(target.flags & 8 /* Number */ && value !== "" && isFinite(+value) ||
                        target.flags & 64 /* BigInt */ && value !== "" && isValidBigIntString(value) ||
                        target.flags & (512 /* BooleanLiteral */ | 98304 /* Nullable */) && value === target.intrinsicName);
                }
                if (source.flags & 134217728 /* TemplateLiteral */) {
                    var texts = source.texts;
                    return texts.length === 2 && texts[0] === "" && texts[1] === "" && isTypeAssignableTo(source.types[0], target);
                }
                return isTypeAssignableTo(source, target);
            }
            function inferTypesFromTemplateLiteralType(source, target) {
                return source.flags & 128 /* StringLiteral */ ? inferFromLiteralPartsToTemplateLiteral([source.value], ts.emptyArray, target) :
                    source.flags & 134217728 /* TemplateLiteral */ ?
                        ts.arraysEqual(source.texts, target.texts) ? ts.map(source.types, getStringLikeTypeForType) :
                            inferFromLiteralPartsToTemplateLiteral(source.texts, source.types, target) :
                        undefined;
            }
            function getStringLikeTypeForType(type) {
                return type.flags & (1 /* Any */ | 402653316 /* StringLike */) ? type : getTemplateLiteralType(["", ""], [type]);
            }
            // This function infers from the text parts and type parts of a source literal to a target template literal. The number
            // of text parts is always one more than the number of type parts, and a source string literal is treated as a source
            // with one text part and zero type parts. The function returns an array of inferred string or template literal types
            // corresponding to the placeholders in the target template literal, or undefined if the source doesn't match the target.
            //
            // We first check that the starting source text part matches the starting target text part, and that the ending source
            // text part ends matches the ending target text part. We then iterate through the remaining target text parts, finding
            // a match for each in the source and inferring string or template literal types created from the segments of the source
            // that occur between the matches. During this iteration, seg holds the index of the current text part in the sourceTexts
            // array and pos holds the current character position in the current text part.
            //
            // Consider inference from type `<<${string}>.<${number}-${number}>>` to type `<${string}.${string}>`, i.e.
            //   sourceTexts = ['<<', '>.<', '-', '>>']
            //   sourceTypes = [string, number, number]
            //   target.texts = ['<', '.', '>']
            // We first match '<' in the target to the start of '<<' in the source and '>' in the target to the end of '>>' in
            // the source. The first match for the '.' in target occurs at character 1 in the source text part at index 1, and thus
            // the first inference is the template literal type `<${string}>`. The remainder of the source makes up the second
            // inference, the template literal type `<${number}-${number}>`.
            function inferFromLiteralPartsToTemplateLiteral(sourceTexts, sourceTypes, target) {
                var lastSourceIndex = sourceTexts.length - 1;
                var sourceStartText = sourceTexts[0];
                var sourceEndText = sourceTexts[lastSourceIndex];
                var targetTexts = target.texts;
                var lastTargetIndex = targetTexts.length - 1;
                var targetStartText = targetTexts[0];
                var targetEndText = targetTexts[lastTargetIndex];
                if (lastSourceIndex === 0 && sourceStartText.length < targetStartText.length + targetEndText.length ||
                    !sourceStartText.startsWith(targetStartText) || !sourceEndText.endsWith(targetEndText))
                    return undefined;
                var remainingEndText = sourceEndText.slice(0, sourceEndText.length - targetEndText.length);
                var matches = [];
                var seg = 0;
                var pos = targetStartText.length;
                for (var i = 1; i < lastTargetIndex; i++) {
                    var delim = targetTexts[i];
                    if (delim.length > 0) {
                        var s = seg;
                        var p = pos;
                        while (true) {
                            p = getSourceText(s).indexOf(delim, p);
                            if (p >= 0)
                                break;
                            s++;
                            if (s === sourceTexts.length)
                                return undefined;
                            p = 0;
                        }
                        addMatch(s, p);
                        pos += delim.length;
                    }
                    else if (pos < getSourceText(seg).length) {
                        addMatch(seg, pos + 1);
                    }
                    else if (seg < lastSourceIndex) {
                        addMatch(seg + 1, 0);
                    }
                    else {
                        return undefined;
                    }
                }
                addMatch(lastSourceIndex, getSourceText(lastSourceIndex).length);
                return matches;
                function getSourceText(index) {
                    return index < lastSourceIndex ? sourceTexts[index] : remainingEndText;
                }
                function addMatch(s, p) {
                    var matchType = s === seg ?
                        getStringLiteralType(getSourceText(s).slice(pos, p)) :
                        getTemplateLiteralType(__spreadArray(__spreadArray([sourceTexts[seg].slice(pos)], sourceTexts.slice(seg + 1, s), true), [getSourceText(s).slice(0, p)], false), sourceTypes.slice(seg, s));
                    matches.push(matchType);
                    seg = s;
                    pos = p;
                }
            }
            function inferTypes(inferences, originalSource, originalTarget, priority, contravariant) {
                if (priority === void 0) {
                    priority = 0;
                }
                if (contravariant === void 0) {
                    contravariant = false;
                }
                var bivariant = false;
                var propagationType;
                var inferencePriority = 2048 /* MaxValue */;
                var allowComplexConstraintInference = true;
                var visited;
                var sourceStack;
                var targetStack;
                var expandingFlags = 0 /* None */;
                inferFromTypes(originalSource, originalTarget);
                function inferFromTypes(source, target) {
                    if (!couldContainTypeVariables(target)) {
                        return;
                    }
                    if (source === wildcardType) {
                        // We are inferring from an 'any' type. We want to infer this type for every type parameter
                        // referenced in the target type, so we record it as the propagation type and infer from the
                        // target to itself. Then, as we find candidates we substitute the propagation type.
                        var savePropagationType = propagationType;
                        propagationType = source;
                        inferFromTypes(target, target);
                        propagationType = savePropagationType;
                        return;
                    }
                    if (source.aliasSymbol && source.aliasTypeArguments && source.aliasSymbol === target.aliasSymbol) {
                        // Source and target are types originating in the same generic type alias declaration.
                        // Simply infer from source type arguments to target type arguments.
                        inferFromTypeArguments(source.aliasTypeArguments, target.aliasTypeArguments, getAliasVariances(source.aliasSymbol));
                        return;
                    }
                    if (source === target && source.flags & 3145728 /* UnionOrIntersection */) {
                        // When source and target are the same union or intersection type, just relate each constituent
                        // type to itself.
                        for (var _i = 0, _a = source.types; _i < _a.length; _i++) {
                            var t = _a[_i];
                            inferFromTypes(t, t);
                        }
                        return;
                    }
                    if (target.flags & 1048576 /* Union */) {
                        // First, infer between identically matching source and target constituents and remove the
                        // matching types.
                        var _b = inferFromMatchingTypes(source.flags & 1048576 /* Union */ ? source.types : [source], target.types, isTypeOrBaseIdenticalTo), tempSources = _b[0], tempTargets = _b[1];
                        // Next, infer between closely matching source and target constituents and remove
                        // the matching types. Types closely match when they are instantiations of the same
                        // object type or instantiations of the same type alias.
                        var _c = inferFromMatchingTypes(tempSources, tempTargets, isTypeCloselyMatchedBy), sources = _c[0], targets = _c[1];
                        if (targets.length === 0) {
                            return;
                        }
                        target = getUnionType(targets);
                        if (sources.length === 0) {
                            // All source constituents have been matched and there is nothing further to infer from.
                            // However, simply making no inferences is undesirable because it could ultimately mean
                            // inferring a type parameter constraint. Instead, make a lower priority inference from
                            // the full source to whatever remains in the target. For example, when inferring from
                            // string to 'string | T', make a lower priority inference of string for T.
                            inferWithPriority(source, target, 1 /* NakedTypeVariable */);
                            return;
                        }
                        source = getUnionType(sources);
                    }
                    else if (target.flags & 2097152 /* Intersection */ && ts.some(target.types, function (t) { return !!getInferenceInfoForType(t) || (isGenericMappedType(t) && !!getInferenceInfoForType(getHomomorphicTypeVariable(t) || neverType)); })) {
                        // We reduce intersection types only when they contain naked type parameters. For example, when
                        // inferring from 'string[] & { extra: any }' to 'string[] & T' we want to remove string[] and
                        // infer { extra: any } for T. But when inferring to 'string[] & Iterable<T>' we want to keep the
                        // string[] on the source side and infer string for T.
                        // Likewise, we consider a homomorphic mapped type constrainted to the target type parameter as similar to a "naked type variable"
                        // in such scenarios.
                        if (!(source.flags & 1048576 /* Union */)) {
                            // Infer between identically matching source and target constituents and remove the matching types.
                            var _d = inferFromMatchingTypes(source.flags & 2097152 /* Intersection */ ? source.types : [source], target.types, isTypeIdenticalTo), sources = _d[0], targets = _d[1];
                            if (sources.length === 0 || targets.length === 0) {
                                return;
                            }
                            source = getIntersectionType(sources);
                            target = getIntersectionType(targets);
                        }
                    }
                    else if (target.flags & (8388608 /* IndexedAccess */ | 33554432 /* Substitution */)) {
                        target = getActualTypeVariable(target);
                    }
                    if (target.flags & 8650752 /* TypeVariable */) {
                        // If target is a type parameter, make an inference, unless the source type contains
                        // the anyFunctionType (the wildcard type that's used to avoid contextually typing functions).
                        // Because the anyFunctionType is internal, it should not be exposed to the user by adding
                        // it as an inference candidate. Hopefully, a better candidate will come along that does
                        // not contain anyFunctionType when we come back to this argument for its second round
                        // of inference. Also, we exclude inferences for silentNeverType (which is used as a wildcard
                        // when constructing types from type parameters that had no inference candidates).
                        if (ts.getObjectFlags(source) & 524288 /* NonInferrableType */ || source === nonInferrableAnyType || source === silentNeverType ||
                            (priority & 128 /* ReturnType */ && (source === autoType || source === autoArrayType)) || isFromInferenceBlockedSource(source)) {
                            return;
                        }
                        var inference = getInferenceInfoForType(target);
                        if (inference) {
                            if (!inference.isFixed) {
                                if (inference.priority === undefined || priority < inference.priority) {
                                    inference.candidates = undefined;
                                    inference.contraCandidates = undefined;
                                    inference.topLevel = true;
                                    inference.priority = priority;
                                }
                                if (priority === inference.priority) {
                                    var candidate = propagationType || source;
                                    // We make contravariant inferences only if we are in a pure contravariant position,
                                    // i.e. only if we have not descended into a bivariant position.
                                    if (contravariant && !bivariant) {
                                        if (!ts.contains(inference.contraCandidates, candidate)) {
                                            inference.contraCandidates = ts.append(inference.contraCandidates, candidate);
                                            clearCachedInferences(inferences);
                                        }
                                    }
                                    else if (!ts.contains(inference.candidates, candidate)) {
                                        inference.candidates = ts.append(inference.candidates, candidate);
                                        clearCachedInferences(inferences);
                                    }
                                }
                                if (!(priority & 128 /* ReturnType */) && target.flags & 262144 /* TypeParameter */ && inference.topLevel && !isTypeParameterAtTopLevel(originalTarget, target)) {
                                    inference.topLevel = false;
                                    clearCachedInferences(inferences);
                                }
                            }
                            inferencePriority = Math.min(inferencePriority, priority);
                            return;
                        }
                        else {
                            // Infer to the simplified version of an indexed access, if possible, to (hopefully) expose more bare type parameters to the inference engine
                            var simplified = getSimplifiedType(target, /*writing*/ false);
                            if (simplified !== target) {
                                invokeOnce(source, simplified, inferFromTypes);
                            }
                            else if (target.flags & 8388608 /* IndexedAccess */) {
                                var indexType = getSimplifiedType(target.indexType, /*writing*/ false);
                                // Generally simplifications of instantiable indexes are avoided to keep relationship checking correct, however if our target is an access, we can consider
                                // that key of that access to be "instantiated", since we're looking to find the infernce goal in any way we can.
                                if (indexType.flags & 465829888 /* Instantiable */) {
                                    var simplified_1 = distributeIndexOverObjectType(getSimplifiedType(target.objectType, /*writing*/ false), indexType, /*writing*/ false);
                                    if (simplified_1 && simplified_1 !== target) {
                                        invokeOnce(source, simplified_1, inferFromTypes);
                                    }
                                }
                            }
                        }
                    }
                    if (ts.getObjectFlags(source) & 4 /* Reference */ && ts.getObjectFlags(target) & 4 /* Reference */ && (source.target === target.target || isArrayType(source) && isArrayType(target)) &&
                        !(source.node && target.node)) {
                        // If source and target are references to the same generic type, infer from type arguments
                        inferFromTypeArguments(getTypeArguments(source), getTypeArguments(target), getVariances(source.target));
                    }
                    else if (source.flags & 4194304 /* Index */ && target.flags & 4194304 /* Index */) {
                        contravariant = !contravariant;
                        inferFromTypes(source.type, target.type);
                        contravariant = !contravariant;
                    }
                    else if ((isLiteralType(source) || source.flags & 4 /* String */) && target.flags & 4194304 /* Index */) {
                        var empty = createEmptyObjectTypeFromStringLiteral(source);
                        contravariant = !contravariant;
                        inferWithPriority(empty, target.type, 256 /* LiteralKeyof */);
                        contravariant = !contravariant;
                    }
                    else if (source.flags & 8388608 /* IndexedAccess */ && target.flags & 8388608 /* IndexedAccess */) {
                        inferFromTypes(source.objectType, target.objectType);
                        inferFromTypes(source.indexType, target.indexType);
                    }
                    else if (source.flags & 268435456 /* StringMapping */ && target.flags & 268435456 /* StringMapping */) {
                        if (source.symbol === target.symbol) {
                            inferFromTypes(source.type, target.type);
                        }
                    }
                    else if (source.flags & 33554432 /* Substitution */) {
                        inferFromTypes(source.baseType, target);
                        var oldPriority = priority;
                        priority |= 4 /* SubstituteSource */;
                        inferFromTypes(source.substitute, target); // Make substitute inference at a lower priority
                        priority = oldPriority;
                    }
                    else if (target.flags & 16777216 /* Conditional */) {
                        invokeOnce(source, target, inferToConditionalType);
                    }
                    else if (target.flags & 3145728 /* UnionOrIntersection */) {
                        inferToMultipleTypes(source, target.types, target.flags);
                    }
                    else if (source.flags & 1048576 /* Union */) {
                        // Source is a union or intersection type, infer from each constituent type
                        var sourceTypes = source.types;
                        for (var _e = 0, sourceTypes_2 = sourceTypes; _e < sourceTypes_2.length; _e++) {
                            var sourceType = sourceTypes_2[_e];
                            inferFromTypes(sourceType, target);
                        }
                    }
                    else if (target.flags & 134217728 /* TemplateLiteral */) {
                        inferToTemplateLiteralType(source, target);
                    }
                    else {
                        source = getReducedType(source);
                        if (!(priority & 512 /* NoConstraints */ && source.flags & (2097152 /* Intersection */ | 465829888 /* Instantiable */))) {
                            var apparentSource = getApparentType(source);
                            // getApparentType can return _any_ type, since an indexed access or conditional may simplify to any other type.
                            // If that occurs and it doesn't simplify to an object or intersection, we'll need to restart `inferFromTypes`
                            // with the simplified source.
                            if (apparentSource !== source && allowComplexConstraintInference && !(apparentSource.flags & (524288 /* Object */ | 2097152 /* Intersection */))) {
                                // TODO: The `allowComplexConstraintInference` flag is a hack! This forbids inference from complex constraints within constraints!
                                // This isn't required algorithmically, but rather is used to lower the memory burden caused by performing inference
                                // that is _too good_ in projects with complicated constraints (eg, fp-ts). In such cases, if we did not limit ourselves
                                // here, we might produce more valid inferences for types, causing us to do more checks and perform more instantiations
                                // (in addition to the extra stack depth here) which, in turn, can push the already close process over its limit.
                                // TL;DR: If we ever become generally more memory efficient (or our resource budget ever increases), we should just
                                // remove this `allowComplexConstraintInference` flag.
                                allowComplexConstraintInference = false;
                                return inferFromTypes(apparentSource, target);
                            }
                            source = apparentSource;
                        }
                        if (source.flags & (524288 /* Object */ | 2097152 /* Intersection */)) {
                            invokeOnce(source, target, inferFromObjectTypes);
                        }
                    }
                }
                function inferWithPriority(source, target, newPriority) {
                    var savePriority = priority;
                    priority |= newPriority;
                    inferFromTypes(source, target);
                    priority = savePriority;
                }
                function invokeOnce(source, target, action) {
                    var key = source.id + "," + target.id;
                    var status = visited && visited.get(key);
                    if (status !== undefined) {
                        inferencePriority = Math.min(inferencePriority, status);
                        return;
                    }
                    (visited || (visited = new ts.Map())).set(key, -1 /* Circularity */);
                    var saveInferencePriority = inferencePriority;
                    inferencePriority = 2048 /* MaxValue */;
                    // We stop inferring and report a circularity if we encounter duplicate recursion identities on both
                    // the source side and the target side.
                    var saveExpandingFlags = expandingFlags;
                    var sourceIdentity = getRecursionIdentity(source);
                    var targetIdentity = getRecursionIdentity(target);
                    if (ts.contains(sourceStack, sourceIdentity))
                        expandingFlags |= 1 /* Source */;
                    if (ts.contains(targetStack, targetIdentity))
                        expandingFlags |= 2 /* Target */;
                    if (expandingFlags !== 3 /* Both */) {
                        (sourceStack || (sourceStack = [])).push(sourceIdentity);
                        (targetStack || (targetStack = [])).push(targetIdentity);
                        action(source, target);
                        targetStack.pop();
                        sourceStack.pop();
                    }
                    else {
                        inferencePriority = -1 /* Circularity */;
                    }
                    expandingFlags = saveExpandingFlags;
                    visited.set(key, inferencePriority);
                    inferencePriority = Math.min(inferencePriority, saveInferencePriority);
                }
                function inferFromMatchingTypes(sources, targets, matches) {
                    var matchedSources;
                    var matchedTargets;
                    for (var _i = 0, targets_1 = targets; _i < targets_1.length; _i++) {
                        var t = targets_1[_i];
                        for (var _a = 0, sources_1 = sources; _a < sources_1.length; _a++) {
                            var s = sources_1[_a];
                            if (matches(s, t)) {
                                inferFromTypes(s, t);
                                matchedSources = ts.appendIfUnique(matchedSources, s);
                                matchedTargets = ts.appendIfUnique(matchedTargets, t);
                            }
                        }
                    }
                    return [
                        matchedSources ? ts.filter(sources, function (t) { return !ts.contains(matchedSources, t); }) : sources,
                        matchedTargets ? ts.filter(targets, function (t) { return !ts.contains(matchedTargets, t); }) : targets,
                    ];
                }
                function inferFromTypeArguments(sourceTypes, targetTypes, variances) {
                    var count = sourceTypes.length < targetTypes.length ? sourceTypes.length : targetTypes.length;
                    for (var i = 0; i < count; i++) {
                        if (i < variances.length && (variances[i] & 7 /* VarianceMask */) === 2 /* Contravariant */) {
                            inferFromContravariantTypes(sourceTypes[i], targetTypes[i]);
                        }
                        else {
                            inferFromTypes(sourceTypes[i], targetTypes[i]);
                        }
                    }
                }
                function inferFromContravariantTypes(source, target) {
                    if (strictFunctionTypes || priority & 1024 /* AlwaysStrict */) {
                        contravariant = !contravariant;
                        inferFromTypes(source, target);
                        contravariant = !contravariant;
                    }
                    else {
                        inferFromTypes(source, target);
                    }
                }
                function getInferenceInfoForType(type) {
                    if (type.flags & 8650752 /* TypeVariable */) {
                        for (var _i = 0, inferences_2 = inferences; _i < inferences_2.length; _i++) {
                            var inference = inferences_2[_i];
                            if (type === inference.typeParameter) {
                                return inference;
                            }
                        }
                    }
                    return undefined;
                }
                function getSingleTypeVariableFromIntersectionTypes(types) {
                    var typeVariable;
                    for (var _i = 0, types_15 = types; _i < types_15.length; _i++) {
                        var type = types_15[_i];
                        var t = type.flags & 2097152 /* Intersection */ && ts.find(type.types, function (t) { return !!getInferenceInfoForType(t); });
                        if (!t || typeVariable && t !== typeVariable) {
                            return undefined;
                        }
                        typeVariable = t;
                    }
                    return typeVariable;
                }
                function inferToMultipleTypes(source, targets, targetFlags) {
                    var typeVariableCount = 0;
                    if (targetFlags & 1048576 /* Union */) {
                        var nakedTypeVariable = void 0;
                        var sources = source.flags & 1048576 /* Union */ ? source.types : [source];
                        var matched_1 = new Array(sources.length);
                        var inferenceCircularity = false;
                        // First infer to types that are not naked type variables. For each source type we
                        // track whether inferences were made from that particular type to some target with
                        // equal priority (i.e. of equal quality) to what we would infer for a naked type
                        // parameter.
                        for (var _i = 0, targets_2 = targets; _i < targets_2.length; _i++) {
                            var t = targets_2[_i];
                            if (getInferenceInfoForType(t)) {
                                nakedTypeVariable = t;
                                typeVariableCount++;
                            }
                            else {
                                for (var i = 0; i < sources.length; i++) {
                                    var saveInferencePriority = inferencePriority;
                                    inferencePriority = 2048 /* MaxValue */;
                                    inferFromTypes(sources[i], t);
                                    if (inferencePriority === priority)
                                        matched_1[i] = true;
                                    inferenceCircularity = inferenceCircularity || inferencePriority === -1 /* Circularity */;
                                    inferencePriority = Math.min(inferencePriority, saveInferencePriority);
                                }
                            }
                        }
                        if (typeVariableCount === 0) {
                            // If every target is an intersection of types containing a single naked type variable,
                            // make a lower priority inference to that type variable. This handles inferring from
                            // 'A | B' to 'T & (X | Y)' where we want to infer 'A | B' for T.
                            var intersectionTypeVariable = getSingleTypeVariableFromIntersectionTypes(targets);
                            if (intersectionTypeVariable) {
                                inferWithPriority(source, intersectionTypeVariable, 1 /* NakedTypeVariable */);
                            }
                            return;
                        }
                        // If the target has a single naked type variable and no inference circularities were
                        // encountered above (meaning we explored the types fully), create a union of the source
                        // types from which no inferences have been made so far and infer from that union to the
                        // naked type variable.
                        if (typeVariableCount === 1 && !inferenceCircularity) {
                            var unmatched = ts.flatMap(sources, function (s, i) { return matched_1[i] ? undefined : s; });
                            if (unmatched.length) {
                                inferFromTypes(getUnionType(unmatched), nakedTypeVariable);
                                return;
                            }
                        }
                    }
                    else {
                        // We infer from types that are not naked type variables first so that inferences we
                        // make from nested naked type variables and given slightly higher priority by virtue
                        // of being first in the candidates array.
                        for (var _a = 0, targets_3 = targets; _a < targets_3.length; _a++) {
                            var t = targets_3[_a];
                            if (getInferenceInfoForType(t)) {
                                typeVariableCount++;
                            }
                            else {
                                inferFromTypes(source, t);
                            }
                        }
                    }
                    // Inferences directly to naked type variables are given lower priority as they are
                    // less specific. For example, when inferring from Promise<string> to T | Promise<T>,
                    // we want to infer string for T, not Promise<string> | string. For intersection types
                    // we only infer to single naked type variables.
                    if (targetFlags & 2097152 /* Intersection */ ? typeVariableCount === 1 : typeVariableCount > 0) {
                        for (var _b = 0, targets_4 = targets; _b < targets_4.length; _b++) {
                            var t = targets_4[_b];
                            if (getInferenceInfoForType(t)) {
                                inferWithPriority(source, t, 1 /* NakedTypeVariable */);
                            }
                        }
                    }
                }
                function inferToMappedType(source, target, constraintType) {
                    if (constraintType.flags & 1048576 /* Union */) {
                        var result = false;
                        for (var _i = 0, _a = constraintType.types; _i < _a.length; _i++) {
                            var type = _a[_i];
                            result = inferToMappedType(source, target, type) || result;
                        }
                        return result;
                    }
                    if (constraintType.flags & 4194304 /* Index */) {
                        // We're inferring from some source type S to a homomorphic mapped type { [P in keyof T]: X },
                        // where T is a type variable. Use inferTypeForHomomorphicMappedType to infer a suitable source
                        // type and then make a secondary inference from that type to T. We make a secondary inference
                        // such that direct inferences to T get priority over inferences to Partial<T>, for example.
                        var inference = getInferenceInfoForType(constraintType.type);
                        if (inference && !inference.isFixed && !isFromInferenceBlockedSource(source)) {
                            var inferredType = inferTypeForHomomorphicMappedType(source, target, constraintType);
                            if (inferredType) {
                                // We assign a lower priority to inferences made from types containing non-inferrable
                                // types because we may only have a partial result (i.e. we may have failed to make
                                // reverse inferences for some properties).
                                inferWithPriority(inferredType, inference.typeParameter, ts.getObjectFlags(source) & 524288 /* NonInferrableType */ ?
                                    16 /* PartialHomomorphicMappedType */ :
                                    8 /* HomomorphicMappedType */);
                            }
                        }
                        return true;
                    }
                    if (constraintType.flags & 262144 /* TypeParameter */) {
                        // We're inferring from some source type S to a mapped type { [P in K]: X }, where K is a type
                        // parameter. First infer from 'keyof S' to K.
                        inferWithPriority(getIndexType(source), constraintType, 32 /* MappedTypeConstraint */);
                        // If K is constrained to a type C, also infer to C. Thus, for a mapped type { [P in K]: X },
                        // where K extends keyof T, we make the same inferences as for a homomorphic mapped type
                        // { [P in keyof T]: X }. This enables us to make meaningful inferences when the target is a
                        // Pick<T, K>.
                        var extendedConstraint = getConstraintOfType(constraintType);
                        if (extendedConstraint && inferToMappedType(source, target, extendedConstraint)) {
                            return true;
                        }
                        // If no inferences can be made to K's constraint, infer from a union of the property types
                        // in the source to the template type X.
                        var propTypes = ts.map(getPropertiesOfType(source), getTypeOfSymbol);
                        var indexTypes = ts.map(getIndexInfosOfType(source), function (info) { return info !== enumNumberIndexInfo ? info.type : neverType; });
                        inferFromTypes(getUnionType(ts.concatenate(propTypes, indexTypes)), getTemplateTypeFromMappedType(target));
                        return true;
                    }
                    return false;
                }
                function inferToConditionalType(source, target) {
                    if (source.flags & 16777216 /* Conditional */) {
                        inferFromTypes(source.checkType, target.checkType);
                        inferFromTypes(source.extendsType, target.extendsType);
                        inferFromTypes(getTrueTypeFromConditionalType(source), getTrueTypeFromConditionalType(target));
                        inferFromTypes(getFalseTypeFromConditionalType(source), getFalseTypeFromConditionalType(target));
                    }
                    else {
                        var savePriority = priority;
                        priority |= contravariant ? 64 /* ContravariantConditional */ : 0;
                        var targetTypes = [getTrueTypeFromConditionalType(target), getFalseTypeFromConditionalType(target)];
                        inferToMultipleTypes(source, targetTypes, target.flags);
                        priority = savePriority;
                    }
                }
                function inferToTemplateLiteralType(source, target) {
                    var matches = inferTypesFromTemplateLiteralType(source, target);
                    var types = target.types;
                    for (var i = 0; i < types.length; i++) {
                        inferFromTypes(matches ? matches[i] : neverType, types[i]);
                    }
                }
                function inferFromObjectTypes(source, target) {
                    if (ts.getObjectFlags(source) & 4 /* Reference */ && ts.getObjectFlags(target) & 4 /* Reference */ && (source.target === target.target || isArrayType(source) && isArrayType(target))) {
                        // If source and target are references to the same generic type, infer from type arguments
                        inferFromTypeArguments(getTypeArguments(source), getTypeArguments(target), getVariances(source.target));
                        return;
                    }
                    if (isGenericMappedType(source) && isGenericMappedType(target)) {
                        // The source and target types are generic types { [P in S]: X } and { [P in T]: Y }, so we infer
                        // from S to T and from X to Y.
                        inferFromTypes(getConstraintTypeFromMappedType(source), getConstraintTypeFromMappedType(target));
                        inferFromTypes(getTemplateTypeFromMappedType(source), getTemplateTypeFromMappedType(target));
                        var sourceNameType = getNameTypeFromMappedType(source);
                        var targetNameType = getNameTypeFromMappedType(target);
                        if (sourceNameType && targetNameType)
                            inferFromTypes(sourceNameType, targetNameType);
                    }
                    if (ts.getObjectFlags(target) & 32 /* Mapped */ && !target.declaration.nameType) {
                        var constraintType = getConstraintTypeFromMappedType(target);
                        if (inferToMappedType(source, target, constraintType)) {
                            return;
                        }
                    }
                    // Infer from the members of source and target only if the two types are possibly related
                    if (!typesDefinitelyUnrelated(source, target)) {
                        if (isArrayType(source) || isTupleType(source)) {
                            if (isTupleType(target)) {
                                var sourceArity = getTypeReferenceArity(source);
                                var targetArity = getTypeReferenceArity(target);
                                var elementTypes = getTypeArguments(target);
                                var elementFlags = target.target.elementFlags;
                                // When source and target are tuple types with the same structure (fixed, variadic, and rest are matched
                                // to the same kind in each position), simply infer between the element types.
                                if (isTupleType(source) && isTupleTypeStructureMatching(source, target)) {
                                    for (var i = 0; i < targetArity; i++) {
                                        inferFromTypes(getTypeArguments(source)[i], elementTypes[i]);
                                    }
                                    return;
                                }
                                var startLength = isTupleType(source) ? Math.min(source.target.fixedLength, target.target.fixedLength) : 0;
                                var endLength = Math.min(isTupleType(source) ? getEndElementCount(source.target, 3 /* Fixed */) : 0, target.target.hasRestElement ? getEndElementCount(target.target, 3 /* Fixed */) : 0);
                                // Infer between starting fixed elements.
                                for (var i = 0; i < startLength; i++) {
                                    inferFromTypes(getTypeArguments(source)[i], elementTypes[i]);
                                }
                                if (!isTupleType(source) || sourceArity - startLength - endLength === 1 && source.target.elementFlags[startLength] & 4 /* Rest */) {
                                    // Single rest element remains in source, infer from that to every element in target
                                    var restType = getTypeArguments(source)[startLength];
                                    for (var i = startLength; i < targetArity - endLength; i++) {
                                        inferFromTypes(elementFlags[i] & 8 /* Variadic */ ? createArrayType(restType) : restType, elementTypes[i]);
                                    }
                                }
                                else {
                                    var middleLength = targetArity - startLength - endLength;
                                    if (middleLength === 2 && elementFlags[startLength] & elementFlags[startLength + 1] & 8 /* Variadic */ && isTupleType(source)) {
                                        // Middle of target is [...T, ...U] and source is tuple type
                                        var targetInfo = getInferenceInfoForType(elementTypes[startLength]);
                                        if (targetInfo && targetInfo.impliedArity !== undefined) {
                                            // Infer slices from source based on implied arity of T.
                                            inferFromTypes(sliceTupleType(source, startLength, endLength + sourceArity - targetInfo.impliedArity), elementTypes[startLength]);
                                            inferFromTypes(sliceTupleType(source, startLength + targetInfo.impliedArity, endLength), elementTypes[startLength + 1]);
                                        }
                                    }
                                    else if (middleLength === 1 && elementFlags[startLength] & 8 /* Variadic */) {
                                        // Middle of target is exactly one variadic element. Infer the slice between the fixed parts in the source.
                                        // If target ends in optional element(s), make a lower priority a speculative inference.
                                        var endsInOptional = target.target.elementFlags[targetArity - 1] & 2 /* Optional */;
                                        var sourceSlice = isTupleType(source) ? sliceTupleType(source, startLength, endLength) : createArrayType(getTypeArguments(source)[0]);
                                        inferWithPriority(sourceSlice, elementTypes[startLength], endsInOptional ? 2 /* SpeculativeTuple */ : 0);
                                    }
                                    else if (middleLength === 1 && elementFlags[startLength] & 4 /* Rest */) {
                                        // Middle of target is exactly one rest element. If middle of source is not empty, infer union of middle element types.
                                        var restType = isTupleType(source) ? getElementTypeOfSliceOfTupleType(source, startLength, endLength) : getTypeArguments(source)[0];
                                        if (restType) {
                                            inferFromTypes(restType, elementTypes[startLength]);
                                        }
                                    }
                                }
                                // Infer between ending fixed elements
                                for (var i = 0; i < endLength; i++) {
                                    inferFromTypes(getTypeArguments(source)[sourceArity - i - 1], elementTypes[targetArity - i - 1]);
                                }
                                return;
                            }
                            if (isArrayType(target)) {
                                inferFromIndexTypes(source, target);
                                return;
                            }
                        }
                        inferFromProperties(source, target);
                        inferFromSignatures(source, target, 0 /* Call */);
                        inferFromSignatures(source, target, 1 /* Construct */);
                        inferFromIndexTypes(source, target);
                    }
                }
                function inferFromProperties(source, target) {
                    var properties = getPropertiesOfObjectType(target);
                    for (var _i = 0, properties_3 = properties; _i < properties_3.length; _i++) {
                        var targetProp = properties_3[_i];
                        var sourceProp = getPropertyOfType(source, targetProp.escapedName);
                        if (sourceProp) {
                            inferFromTypes(getTypeOfSymbol(sourceProp), getTypeOfSymbol(targetProp));
                        }
                    }
                }
                function inferFromSignatures(source, target, kind) {
                    var sourceSignatures = getSignaturesOfType(source, kind);
                    var targetSignatures = getSignaturesOfType(target, kind);
                    var sourceLen = sourceSignatures.length;
                    var targetLen = targetSignatures.length;
                    var len = sourceLen < targetLen ? sourceLen : targetLen;
                    var skipParameters = !!(ts.getObjectFlags(source) & 524288 /* NonInferrableType */);
                    for (var i = 0; i < len; i++) {
                        inferFromSignature(getBaseSignature(sourceSignatures[sourceLen - len + i]), getErasedSignature(targetSignatures[targetLen - len + i]), skipParameters);
                    }
                }
                function inferFromSignature(source, target, skipParameters) {
                    if (!skipParameters) {
                        var saveBivariant = bivariant;
                        var kind = target.declaration ? target.declaration.kind : 0 /* Unknown */;
                        // Once we descend into a bivariant signature we remain bivariant for all nested inferences
                        bivariant = bivariant || kind === 167 /* MethodDeclaration */ || kind === 166 /* MethodSignature */ || kind === 169 /* Constructor */;
                        applyToParameterTypes(source, target, inferFromContravariantTypes);
                        bivariant = saveBivariant;
                    }
                    applyToReturnTypes(source, target, inferFromTypes);
                }
                function inferFromIndexTypes(source, target) {
                    // Inferences across mapped type index signatures are pretty much the same a inferences to homomorphic variables
                    var priority = (ts.getObjectFlags(source) & ts.getObjectFlags(target) & 32 /* Mapped */) ? 8 /* HomomorphicMappedType */ : 0;
                    var indexInfos = getIndexInfosOfType(target);
                    if (isObjectTypeWithInferableIndex(source)) {
                        for (var _i = 0, indexInfos_4 = indexInfos; _i < indexInfos_4.length; _i++) {
                            var targetInfo = indexInfos_4[_i];
                            var propTypes = [];
                            for (var _a = 0, _b = getPropertiesOfType(source); _a < _b.length; _a++) {
                                var prop = _b[_a];
                                if (isApplicableIndexType(getLiteralTypeFromProperty(prop, 8576 /* StringOrNumberLiteralOrUnique */), targetInfo.keyType)) {
                                    var propType = getTypeOfSymbol(prop);
                                    propTypes.push(prop.flags & 16777216 /* Optional */ ? removeMissingOrUndefinedType(propType) : propType);
                                }
                            }
                            for (var _c = 0, _d = getIndexInfosOfType(source); _c < _d.length; _c++) {
                                var info = _d[_c];
                                if (isApplicableIndexType(info.keyType, targetInfo.keyType)) {
                                    propTypes.push(info.type);
                                }
                            }
                            if (propTypes.length) {
                                inferWithPriority(getUnionType(propTypes), targetInfo.type, priority);
                            }
                        }
                    }
                    for (var _e = 0, indexInfos_5 = indexInfos; _e < indexInfos_5.length; _e++) {
                        var targetInfo = indexInfos_5[_e];
                        var sourceInfo = getApplicableIndexInfo(source, targetInfo.keyType);
                        if (sourceInfo) {
                            inferWithPriority(sourceInfo.type, targetInfo.type, priority);
                        }
                    }
                }
            }
            function isTypeOrBaseIdenticalTo(s, t) {
                return exactOptionalPropertyTypes && t === missingType ? s === t :
                    (isTypeIdenticalTo(s, t) || !!(t.flags & 4 /* String */ && s.flags & 128 /* StringLiteral */ || t.flags & 8 /* Number */ && s.flags & 256 /* NumberLiteral */));
            }
            function isTypeCloselyMatchedBy(s, t) {
                return !!(s.flags & 524288 /* Object */ && t.flags & 524288 /* Object */ && s.symbol && s.symbol === t.symbol ||
                    s.aliasSymbol && s.aliasTypeArguments && s.aliasSymbol === t.aliasSymbol);
            }
            function hasPrimitiveConstraint(type) {
                var constraint = getConstraintOfTypeParameter(type);
                return !!constraint && maybeTypeOfKind(constraint.flags & 16777216 /* Conditional */ ? getDefaultConstraintOfConditionalType(constraint) : constraint, 131068 /* Primitive */ | 4194304 /* Index */ | 134217728 /* TemplateLiteral */ | 268435456 /* StringMapping */);
            }
            function isObjectLiteralType(type) {
                return !!(ts.getObjectFlags(type) & 128 /* ObjectLiteral */);
            }
            function isObjectOrArrayLiteralType(type) {
                return !!(ts.getObjectFlags(type) & (128 /* ObjectLiteral */ | 32768 /* ArrayLiteral */));
            }
            function unionObjectAndArrayLiteralCandidates(candidates) {
                if (candidates.length > 1) {
                    var objectLiterals = ts.filter(candidates, isObjectOrArrayLiteralType);
                    if (objectLiterals.length) {
                        var literalsType = getUnionType(objectLiterals, 2 /* Subtype */);
                        return ts.concatenate(ts.filter(candidates, function (t) { return !isObjectOrArrayLiteralType(t); }), [literalsType]);
                    }
                }
                return candidates;
            }
            function getContravariantInference(inference) {
                return inference.priority & 416 /* PriorityImpliesCombination */ ? getIntersectionType(inference.contraCandidates) : getCommonSubtype(inference.contraCandidates);
            }
            function getCovariantInference(inference, signature) {
                // Extract all object and array literal types and replace them with a single widened and normalized type.
                var candidates = unionObjectAndArrayLiteralCandidates(inference.candidates);
                // We widen inferred literal types if
                // all inferences were made to top-level occurrences of the type parameter, and
                // the type parameter has no constraint or its constraint includes no primitive or literal types, and
                // the type parameter was fixed during inference or does not occur at top-level in the return type.
                var primitiveConstraint = hasPrimitiveConstraint(inference.typeParameter);
                var widenLiteralTypes = !primitiveConstraint && inference.topLevel &&
                    (inference.isFixed || !isTypeParameterAtTopLevel(getReturnTypeOfSignature(signature), inference.typeParameter));
                var baseCandidates = primitiveConstraint ? ts.sameMap(candidates, getRegularTypeOfLiteralType) :
                    widenLiteralTypes ? ts.sameMap(candidates, getWidenedLiteralType) :
                        candidates;
                // If all inferences were made from a position that implies a combined result, infer a union type.
                // Otherwise, infer a common supertype.
                var unwidenedType = inference.priority & 416 /* PriorityImpliesCombination */ ?
                    getUnionType(baseCandidates, 2 /* Subtype */) :
                    getCommonSupertype(baseCandidates);
                return getWidenedType(unwidenedType);
            }
            function getInferredType(context, index) {
                var inference = context.inferences[index];
                if (!inference.inferredType) {
                    var inferredType = void 0;
                    var signature = context.signature;
                    if (signature) {
                        var inferredCovariantType = inference.candidates ? getCovariantInference(inference, signature) : undefined;
                        if (inference.contraCandidates) {
                            var inferredContravariantType = getContravariantInference(inference);
                            // If we have both co- and contra-variant inferences, we prefer the contra-variant inference
                            // unless the co-variant inference is a subtype and not 'never'.
                            inferredType = inferredCovariantType && !(inferredCovariantType.flags & 131072 /* Never */) &&
                                isTypeSubtypeOf(inferredCovariantType, inferredContravariantType) ?
                                inferredCovariantType : inferredContravariantType;
                        }
                        else if (inferredCovariantType) {
                            inferredType = inferredCovariantType;
                        }
                        else if (context.flags & 1 /* NoDefault */) {
                            // We use silentNeverType as the wildcard that signals no inferences.
                            inferredType = silentNeverType;
                        }
                        else {
                            // Infer either the default or the empty object type when no inferences were
                            // made. It is important to remember that in this case, inference still
                            // succeeds, meaning there is no error for not having inference candidates. An
                            // inference error only occurs when there are *conflicting* candidates, i.e.
                            // candidates with no common supertype.
                            var defaultType = getDefaultFromTypeParameter(inference.typeParameter);
                            if (defaultType) {
                                // Instantiate the default type. Any forward reference to a type
                                // parameter should be instantiated to the empty object type.
                                inferredType = instantiateType(defaultType, mergeTypeMappers(createBackreferenceMapper(context, index), context.nonFixingMapper));
                            }
                        }
                    }
                    else {
                        inferredType = getTypeFromInference(inference);
                    }
                    inference.inferredType = inferredType || getDefaultTypeArgumentType(!!(context.flags & 2 /* AnyDefault */));
                    var constraint = getConstraintOfTypeParameter(inference.typeParameter);
                    if (constraint) {
                        var instantiatedConstraint = instantiateType(constraint, context.nonFixingMapper);
                        if (!inferredType || !context.compareTypes(inferredType, getTypeWithThisArgument(instantiatedConstraint, inferredType))) {
                            inference.inferredType = inferredType = instantiatedConstraint;
                        }
                    }
                }
                return inference.inferredType;
            }
            function getDefaultTypeArgumentType(isInJavaScriptFile) {
                return isInJavaScriptFile ? anyType : unknownType;
            }
            function getInferredTypes(context) {
                var result = [];
                for (var i = 0; i < context.inferences.length; i++) {
                    result.push(getInferredType(context, i));
                }
                return result;
            }
            // EXPRESSION TYPE CHECKING
            function getCannotFindNameDiagnosticForName(node) {
                switch (node.escapedText) {
                    case "document":
                    case "console":
                        return ts.Diagnostics.Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_include_dom;
                    case "$":
                        return compilerOptions.types
                            ? ts.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slashjquery_and_then_add_jquery_to_the_types_field_in_your_tsconfig
                            : ts.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_jQuery_Try_npm_i_save_dev_types_Slashjquery;
                    case "describe":
                    case "suite":
                    case "it":
                    case "test":
                        return compilerOptions.types
                            ? ts.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_types_Slashjest_or_npm_i_save_dev_types_Slashmocha_and_then_add_jest_or_mocha_to_the_types_field_in_your_tsconfig
                            : ts.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_a_test_runner_Try_npm_i_save_dev_types_Slashjest_or_npm_i_save_dev_types_Slashmocha;
                    case "process":
                    case "require":
                    case "Buffer":
                    case "module":
                        return compilerOptions.types
                            ? ts.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashnode_and_then_add_node_to_the_types_field_in_your_tsconfig
                            : ts.Diagnostics.Cannot_find_name_0_Do_you_need_to_install_type_definitions_for_node_Try_npm_i_save_dev_types_Slashnode;
                    case "Map":
                    case "Set":
                    case "Promise":
                    case "Symbol":
                    case "WeakMap":
                    case "WeakSet":
                    case "Iterator":
                    case "AsyncIterator":
                    case "SharedArrayBuffer":
                    case "Atomics":
                    case "AsyncIterable":
                    case "AsyncIterableIterator":
                    case "AsyncGenerator":
                    case "AsyncGeneratorFunction":
                    case "BigInt":
                    case "Reflect":
                    case "BigInt64Array":
                    case "BigUint64Array":
                        return ts.Diagnostics.Cannot_find_name_0_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_1_or_later;
                    default:
                        if (node.parent.kind === 292 /* ShorthandPropertyAssignment */) {
                            return ts.Diagnostics.No_value_exists_in_scope_for_the_shorthand_property_0_Either_declare_one_or_provide_an_initializer;
                        }
                        else {
                            return ts.Diagnostics.Cannot_find_name_0;
                        }
                }
            }
            function getResolvedSymbol(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedSymbol) {
                    links.resolvedSymbol = !ts.nodeIsMissing(node) &&
                        resolveName(node, node.escapedText, 111551 /* Value */ | 1048576 /* ExportValue */, getCannotFindNameDiagnosticForName(node), node, !ts.isWriteOnlyAccess(node), 
                        /*excludeGlobals*/ false, 
                        /*issueSuggestions*/ true) || unknownSymbol;
                }
                return links.resolvedSymbol;
            }
            function isInTypeQuery(node) {
                // TypeScript 1.0 spec (April 2014): 3.6.3
                // A type query consists of the keyword typeof followed by an expression.
                // The expression is restricted to a single identifier or a sequence of identifiers separated by periods
                return !!ts.findAncestor(node, function (n) { return n.kind === 179 /* TypeQuery */ ? true : n.kind === 79 /* Identifier */ || n.kind === 159 /* QualifiedName */ ? false : "quit"; });
            }
            // Return the flow cache key for a "dotted name" (i.e. a sequence of identifiers
            // separated by dots). The key consists of the id of the symbol referenced by the
            // leftmost identifier followed by zero or more property names separated by dots.
            // The result is undefined if the reference isn't a dotted name.
            function getFlowCacheKey(node, declaredType, initialType, flowContainer) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        var symbol = getResolvedSymbol(node);
                        return symbol !== unknownSymbol ? (flowContainer ? getNodeId(flowContainer) : "-1") + "|" + getTypeId(declaredType) + "|" + getTypeId(initialType) + "|" + getSymbolId(symbol) : undefined;
                    case 108 /* ThisKeyword */:
                        return "0|" + (flowContainer ? getNodeId(flowContainer) : "-1") + "|" + getTypeId(declaredType) + "|" + getTypeId(initialType);
                    case 228 /* NonNullExpression */:
                    case 210 /* ParenthesizedExpression */:
                        return getFlowCacheKey(node.expression, declaredType, initialType, flowContainer);
                    case 159 /* QualifiedName */:
                        var left = getFlowCacheKey(node.left, declaredType, initialType, flowContainer);
                        return left && left + "." + node.right.escapedText;
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */:
                        var propName = getAccessedPropertyName(node);
                        if (propName !== undefined) {
                            var key = getFlowCacheKey(node.expression, declaredType, initialType, flowContainer);
                            return key && key + "." + propName;
                        }
                }
                return undefined;
            }
            function isMatchingReference(source, target) {
                switch (target.kind) {
                    case 210 /* ParenthesizedExpression */:
                    case 228 /* NonNullExpression */:
                        return isMatchingReference(source, target.expression);
                    case 219 /* BinaryExpression */:
                        return (ts.isAssignmentExpression(target) && isMatchingReference(source, target.left)) ||
                            (ts.isBinaryExpression(target) && target.operatorToken.kind === 27 /* CommaToken */ && isMatchingReference(source, target.right));
                }
                switch (source.kind) {
                    case 229 /* MetaProperty */:
                        return target.kind === 229 /* MetaProperty */
                            && source.keywordToken === target.keywordToken
                            && source.name.escapedText === target.name.escapedText;
                    case 79 /* Identifier */:
                    case 80 /* PrivateIdentifier */:
                        return ts.isThisInTypeQuery(source) ?
                            target.kind === 108 /* ThisKeyword */ :
                            target.kind === 79 /* Identifier */ && getResolvedSymbol(source) === getResolvedSymbol(target) ||
                                (target.kind === 252 /* VariableDeclaration */ || target.kind === 201 /* BindingElement */) &&
                                    getExportSymbolOfValueSymbolIfExported(getResolvedSymbol(source)) === getSymbolOfNode(target);
                    case 108 /* ThisKeyword */:
                        return target.kind === 108 /* ThisKeyword */;
                    case 106 /* SuperKeyword */:
                        return target.kind === 106 /* SuperKeyword */;
                    case 228 /* NonNullExpression */:
                    case 210 /* ParenthesizedExpression */:
                        return isMatchingReference(source.expression, target);
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */:
                        return ts.isAccessExpression(target) &&
                            getAccessedPropertyName(source) === getAccessedPropertyName(target) &&
                            isMatchingReference(source.expression, target.expression);
                    case 159 /* QualifiedName */:
                        return ts.isAccessExpression(target) &&
                            source.right.escapedText === getAccessedPropertyName(target) &&
                            isMatchingReference(source.left, target.expression);
                    case 219 /* BinaryExpression */:
                        return (ts.isBinaryExpression(source) && source.operatorToken.kind === 27 /* CommaToken */ && isMatchingReference(source.right, target));
                }
                return false;
            }
            // Given a source x, check if target matches x or is an && operation with an operand that matches x.
            function containsTruthyCheck(source, target) {
                return isMatchingReference(source, target) ||
                    (target.kind === 219 /* BinaryExpression */ && target.operatorToken.kind === 55 /* AmpersandAmpersandToken */ &&
                        (containsTruthyCheck(source, target.left) || containsTruthyCheck(source, target.right)));
            }
            function getPropertyAccess(expr) {
                if (ts.isAccessExpression(expr)) {
                    return expr;
                }
                if (ts.isIdentifier(expr)) {
                    var symbol = getResolvedSymbol(expr);
                    if (isConstVariable(symbol)) {
                        var declaration = symbol.valueDeclaration;
                        // Given 'const x = obj.kind', allow 'x' as an alias for 'obj.kind'
                        if (ts.isVariableDeclaration(declaration) && !declaration.type && declaration.initializer && ts.isAccessExpression(declaration.initializer)) {
                            return declaration.initializer;
                        }
                        // Given 'const { kind: x } = obj', allow 'x' as an alias for 'obj.kind'
                        if (ts.isBindingElement(declaration) && !declaration.initializer) {
                            var parent = declaration.parent.parent;
                            if (ts.isVariableDeclaration(parent) && !parent.type && parent.initializer && (ts.isIdentifier(parent.initializer) || ts.isAccessExpression(parent.initializer))) {
                                return declaration;
                            }
                        }
                    }
                }
                return undefined;
            }
            function getAccessedPropertyName(access) {
                var propertyName;
                return access.kind === 204 /* PropertyAccessExpression */ ? access.name.escapedText :
                    access.kind === 205 /* ElementAccessExpression */ && ts.isStringOrNumericLiteralLike(access.argumentExpression) ? ts.escapeLeadingUnderscores(access.argumentExpression.text) :
                        access.kind === 201 /* BindingElement */ && (propertyName = getDestructuringPropertyName(access)) ? ts.escapeLeadingUnderscores(propertyName) :
                            undefined;
            }
            function containsMatchingReference(source, target) {
                while (ts.isAccessExpression(source)) {
                    source = source.expression;
                    if (isMatchingReference(source, target)) {
                        return true;
                    }
                }
                return false;
            }
            function optionalChainContainsReference(source, target) {
                while (ts.isOptionalChain(source)) {
                    source = source.expression;
                    if (isMatchingReference(source, target)) {
                        return true;
                    }
                }
                return false;
            }
            function isDiscriminantProperty(type, name) {
                if (type && type.flags & 1048576 /* Union */) {
                    var prop = getUnionOrIntersectionProperty(type, name);
                    if (prop && ts.getCheckFlags(prop) & 2 /* SyntheticProperty */) {
                        if (prop.isDiscriminantProperty === undefined) {
                            prop.isDiscriminantProperty =
                                (prop.checkFlags & 192 /* Discriminant */) === 192 /* Discriminant */ &&
                                    !maybeTypeOfKind(getTypeOfSymbol(prop), 465829888 /* Instantiable */ & ~134217728 /* TemplateLiteral */);
                        }
                        return !!prop.isDiscriminantProperty;
                    }
                }
                return false;
            }
            function findDiscriminantProperties(sourceProperties, target) {
                var result;
                for (var _i = 0, sourceProperties_2 = sourceProperties; _i < sourceProperties_2.length; _i++) {
                    var sourceProperty = sourceProperties_2[_i];
                    if (isDiscriminantProperty(target, sourceProperty.escapedName)) {
                        if (result) {
                            result.push(sourceProperty);
                            continue;
                        }
                        result = [sourceProperty];
                    }
                }
                return result;
            }
            // Given a set of constituent types and a property name, create and return a map keyed by the literal
            // types of the property by that name in each constituent type. No map is returned if some key property
            // has a non-literal type or if less than 10 or less than 50% of the constituents have a unique key.
            // Entries with duplicate keys have unknownType as the value.
            function mapTypesByKeyProperty(types, name) {
                var map = new ts.Map();
                var count = 0;
                var _loop_22 = function (type) {
                    if (type.flags & (524288 /* Object */ | 2097152 /* Intersection */ | 58982400 /* InstantiableNonPrimitive */)) {
                        var discriminant = getTypeOfPropertyOfType(type, name);
                        if (discriminant) {
                            if (!isLiteralType(discriminant)) {
                                return { value: undefined };
                            }
                            var duplicate_1 = false;
                            forEachType(discriminant, function (t) {
                                var id = getTypeId(getRegularTypeOfLiteralType(t));
                                var existing = map.get(id);
                                if (!existing) {
                                    map.set(id, type);
                                }
                                else if (existing !== unknownType) {
                                    map.set(id, unknownType);
                                    duplicate_1 = true;
                                }
                            });
                            if (!duplicate_1)
                                count++;
                        }
                    }
                };
                for (var _i = 0, types_16 = types; _i < types_16.length; _i++) {
                    var type = types_16[_i];
                    var state_8 = _loop_22(type);
                    if (typeof state_8 === "object")
                        return state_8.value;
                }
                return count >= 10 && count * 2 >= types.length ? map : undefined;
            }
            // Return the name of a discriminant property for which it was possible and feasible to construct a map of
            // constituent types keyed by the literal types of the property by that name in each constituent type.
            function getKeyPropertyName(unionType) {
                var types = unionType.types;
                // We only construct maps for large unions with non-primitive constituents.
                if (types.length < 10 || ts.getObjectFlags(unionType) & 65536 /* PrimitiveUnion */) {
                    return undefined;
                }
                if (unionType.keyPropertyName === undefined) {
                    // The candidate key property name is the name of the first property with a unit type in one of the
                    // constituent types.
                    var keyPropertyName = ts.forEach(types, function (t) {
                        return t.flags & (524288 /* Object */ | 58982400 /* InstantiableNonPrimitive */) ?
                            ts.forEach(getPropertiesOfType(t), function (p) { return isUnitType(getTypeOfSymbol(p)) ? p.escapedName : undefined; }) :
                            undefined;
                    });
                    var mapByKeyProperty = keyPropertyName && mapTypesByKeyProperty(types, keyPropertyName);
                    unionType.keyPropertyName = mapByKeyProperty ? keyPropertyName : "";
                    unionType.constituentMap = mapByKeyProperty;
                }
                return unionType.keyPropertyName.length ? unionType.keyPropertyName : undefined;
            }
            // Given a union type for which getKeyPropertyName returned a non-undefined result, return the constituent
            // that corresponds to the given key type for that property name.
            function getConstituentTypeForKeyType(unionType, keyType) {
                var _a;
                var result = (_a = unionType.constituentMap) === null || _a === void 0 ? void 0 : _a.get(getTypeId(getRegularTypeOfLiteralType(keyType)));
                return result !== unknownType ? result : undefined;
            }
            function getMatchingUnionConstituentForType(unionType, type) {
                var keyPropertyName = getKeyPropertyName(unionType);
                var propType = keyPropertyName && getTypeOfPropertyOfType(type, keyPropertyName);
                return propType && getConstituentTypeForKeyType(unionType, propType);
            }
            function getMatchingUnionConstituentForObjectLiteral(unionType, node) {
                var keyPropertyName = getKeyPropertyName(unionType);
                var propNode = keyPropertyName && ts.find(node.properties, function (p) {
                    return p.symbol && p.kind === 291 /* PropertyAssignment */ &&
                        p.symbol.escapedName === keyPropertyName && isPossiblyDiscriminantValue(p.initializer);
                });
                var propType = propNode && getTypeOfExpression(propNode.initializer);
                return propType && getConstituentTypeForKeyType(unionType, propType);
            }
            function isOrContainsMatchingReference(source, target) {
                return isMatchingReference(source, target) || containsMatchingReference(source, target);
            }
            function hasMatchingArgument(expression, reference) {
                if (expression.arguments) {
                    for (var _i = 0, _a = expression.arguments; _i < _a.length; _i++) {
                        var argument = _a[_i];
                        if (isOrContainsMatchingReference(reference, argument)) {
                            return true;
                        }
                    }
                }
                if (expression.expression.kind === 204 /* PropertyAccessExpression */ &&
                    isOrContainsMatchingReference(reference, expression.expression.expression)) {
                    return true;
                }
                return false;
            }
            function getFlowNodeId(flow) {
                if (!flow.id || flow.id < 0) {
                    flow.id = nextFlowId;
                    nextFlowId++;
                }
                return flow.id;
            }
            function typeMaybeAssignableTo(source, target) {
                if (!(source.flags & 1048576 /* Union */)) {
                    return isTypeAssignableTo(source, target);
                }
                for (var _i = 0, _a = source.types; _i < _a.length; _i++) {
                    var t = _a[_i];
                    if (isTypeAssignableTo(t, target)) {
                        return true;
                    }
                }
                return false;
            }
            // Remove those constituent types of declaredType to which no constituent type of assignedType is assignable.
            // For example, when a variable of type number | string | boolean is assigned a value of type number | boolean,
            // we remove type string.
            function getAssignmentReducedType(declaredType, assignedType) {
                if (declaredType !== assignedType) {
                    if (assignedType.flags & 131072 /* Never */) {
                        return assignedType;
                    }
                    var reducedType = filterType(declaredType, function (t) { return typeMaybeAssignableTo(assignedType, t); });
                    if (assignedType.flags & 512 /* BooleanLiteral */ && isFreshLiteralType(assignedType)) {
                        reducedType = mapType(reducedType, getFreshTypeOfLiteralType); // Ensure that if the assignment is a fresh type, that we narrow to fresh types
                    }
                    // Our crude heuristic produces an invalid result in some cases: see GH#26130.
                    // For now, when that happens, we give up and don't narrow at all.  (This also
                    // means we'll never narrow for erroneous assignments where the assigned type
                    // is not assignable to the declared type.)
                    if (isTypeAssignableTo(assignedType, reducedType)) {
                        return reducedType;
                    }
                }
                return declaredType;
            }
            function isFunctionObjectType(type) {
                // We do a quick check for a "bind" property before performing the more expensive subtype
                // check. This gives us a quicker out in the common case where an object type is not a function.
                var resolved = resolveStructuredTypeMembers(type);
                return !!(resolved.callSignatures.length || resolved.constructSignatures.length ||
                    resolved.members.get("bind") && isTypeSubtypeOf(type, globalFunctionType));
            }
            function getTypeFacts(type, ignoreObjects) {
                if (ignoreObjects === void 0) {
                    ignoreObjects = false;
                }
                var flags = type.flags;
                if (flags & 4 /* String */) {
                    return strictNullChecks ? 16317953 /* StringStrictFacts */ : 16776705 /* StringFacts */;
                }
                if (flags & 128 /* StringLiteral */) {
                    var isEmpty = type.value === "";
                    return strictNullChecks ?
                        isEmpty ? 12123649 /* EmptyStringStrictFacts */ : 7929345 /* NonEmptyStringStrictFacts */ :
                        isEmpty ? 12582401 /* EmptyStringFacts */ : 16776705 /* NonEmptyStringFacts */;
                }
                if (flags & (8 /* Number */ | 32 /* Enum */)) {
                    return strictNullChecks ? 16317698 /* NumberStrictFacts */ : 16776450 /* NumberFacts */;
                }
                if (flags & 256 /* NumberLiteral */) {
                    var isZero = type.value === 0;
                    return strictNullChecks ?
                        isZero ? 12123394 /* ZeroNumberStrictFacts */ : 7929090 /* NonZeroNumberStrictFacts */ :
                        isZero ? 12582146 /* ZeroNumberFacts */ : 16776450 /* NonZeroNumberFacts */;
                }
                if (flags & 64 /* BigInt */) {
                    return strictNullChecks ? 16317188 /* BigIntStrictFacts */ : 16775940 /* BigIntFacts */;
                }
                if (flags & 2048 /* BigIntLiteral */) {
                    var isZero = isZeroBigInt(type);
                    return strictNullChecks ?
                        isZero ? 12122884 /* ZeroBigIntStrictFacts */ : 7928580 /* NonZeroBigIntStrictFacts */ :
                        isZero ? 12581636 /* ZeroBigIntFacts */ : 16775940 /* NonZeroBigIntFacts */;
                }
                if (flags & 16 /* Boolean */) {
                    return strictNullChecks ? 16316168 /* BooleanStrictFacts */ : 16774920 /* BooleanFacts */;
                }
                if (flags & 528 /* BooleanLike */) {
                    return strictNullChecks ?
                        (type === falseType || type === regularFalseType) ? 12121864 /* FalseStrictFacts */ : 7927560 /* TrueStrictFacts */ :
                        (type === falseType || type === regularFalseType) ? 12580616 /* FalseFacts */ : 16774920 /* TrueFacts */;
                }
                if (flags & 524288 /* Object */ && !ignoreObjects) {
                    return ts.getObjectFlags(type) & 16 /* Anonymous */ && isEmptyObjectType(type) ?
                        strictNullChecks ? 16318463 /* EmptyObjectStrictFacts */ : 16777215 /* EmptyObjectFacts */ :
                        isFunctionObjectType(type) ?
                            strictNullChecks ? 7880640 /* FunctionStrictFacts */ : 16728000 /* FunctionFacts */ :
                            strictNullChecks ? 7888800 /* ObjectStrictFacts */ : 16736160 /* ObjectFacts */;
                }
                if (flags & (16384 /* Void */ | 32768 /* Undefined */)) {
                    return 9830144 /* UndefinedFacts */;
                }
                if (flags & 65536 /* Null */) {
                    return 9363232 /* NullFacts */;
                }
                if (flags & 12288 /* ESSymbolLike */) {
                    return strictNullChecks ? 7925520 /* SymbolStrictFacts */ : 16772880 /* SymbolFacts */;
                }
                if (flags & 67108864 /* NonPrimitive */) {
                    return strictNullChecks ? 7888800 /* ObjectStrictFacts */ : 16736160 /* ObjectFacts */;
                }
                if (flags & 131072 /* Never */) {
                    return 0 /* None */;
                }
                if (flags & 465829888 /* Instantiable */) {
                    return !isPatternLiteralType(type) ? getTypeFacts(getBaseConstraintOfType(type) || unknownType, ignoreObjects) :
                        strictNullChecks ? 7929345 /* NonEmptyStringStrictFacts */ : 16776705 /* NonEmptyStringFacts */;
                }
                if (flags & 1048576 /* Union */) {
                    return ts.reduceLeft(type.types, function (facts, t) { return facts | getTypeFacts(t, ignoreObjects); }, 0 /* None */);
                }
                if (flags & 2097152 /* Intersection */) {
                    // When an intersection contains a primitive type we ignore object type constituents as they are
                    // presumably type tags. For example, in string & { __kind__: "name" } we ignore the object type.
                    ignoreObjects || (ignoreObjects = maybeTypeOfKind(type, 131068 /* Primitive */));
                    return ts.reduceLeft(type.types, function (facts, t) { return facts & getTypeFacts(t, ignoreObjects); }, 16777215 /* All */);
                }
                return 16777215 /* All */;
            }
            function getTypeWithFacts(type, include) {
                return filterType(type, function (t) { return (getTypeFacts(t) & include) !== 0; });
            }
            function getTypeWithDefault(type, defaultExpression) {
                return defaultExpression ?
                    getUnionType([getNonUndefinedType(type), getTypeOfExpression(defaultExpression)]) :
                    type;
            }
            function getTypeOfDestructuredProperty(type, name) {
                var _a;
                var nameType = getLiteralTypeFromPropertyName(name);
                if (!isTypeUsableAsPropertyName(nameType))
                    return errorType;
                var text = getPropertyNameFromType(nameType);
                return getTypeOfPropertyOfType(type, text) || includeUndefinedInIndexSignature((_a = getApplicableIndexInfoForName(type, text)) === null || _a === void 0 ? void 0 : _a.type) || errorType;
            }
            function getTypeOfDestructuredArrayElement(type, index) {
                return everyType(type, isTupleLikeType) && getTupleElementType(type, index) ||
                    includeUndefinedInIndexSignature(checkIteratedTypeOrElementType(65 /* Destructuring */, type, undefinedType, /*errorNode*/ undefined)) ||
                    errorType;
            }
            function includeUndefinedInIndexSignature(type) {
                if (!type)
                    return type;
                return compilerOptions.noUncheckedIndexedAccess ?
                    getUnionType([type, undefinedType]) :
                    type;
            }
            function getTypeOfDestructuredSpreadExpression(type) {
                return createArrayType(checkIteratedTypeOrElementType(65 /* Destructuring */, type, undefinedType, /*errorNode*/ undefined) || errorType);
            }
            function getAssignedTypeOfBinaryExpression(node) {
                var isDestructuringDefaultAssignment = node.parent.kind === 202 /* ArrayLiteralExpression */ && isDestructuringAssignmentTarget(node.parent) ||
                    node.parent.kind === 291 /* PropertyAssignment */ && isDestructuringAssignmentTarget(node.parent.parent);
                return isDestructuringDefaultAssignment ?
                    getTypeWithDefault(getAssignedType(node), node.right) :
                    getTypeOfExpression(node.right);
            }
            function isDestructuringAssignmentTarget(parent) {
                return parent.parent.kind === 219 /* BinaryExpression */ && parent.parent.left === parent ||
                    parent.parent.kind === 242 /* ForOfStatement */ && parent.parent.initializer === parent;
            }
            function getAssignedTypeOfArrayLiteralElement(node, element) {
                return getTypeOfDestructuredArrayElement(getAssignedType(node), node.elements.indexOf(element));
            }
            function getAssignedTypeOfSpreadExpression(node) {
                return getTypeOfDestructuredSpreadExpression(getAssignedType(node.parent));
            }
            function getAssignedTypeOfPropertyAssignment(node) {
                return getTypeOfDestructuredProperty(getAssignedType(node.parent), node.name);
            }
            function getAssignedTypeOfShorthandPropertyAssignment(node) {
                return getTypeWithDefault(getAssignedTypeOfPropertyAssignment(node), node.objectAssignmentInitializer);
            }
            function getAssignedType(node) {
                var parent = node.parent;
                switch (parent.kind) {
                    case 241 /* ForInStatement */:
                        return stringType;
                    case 242 /* ForOfStatement */:
                        return checkRightHandSideOfForOf(parent) || errorType;
                    case 219 /* BinaryExpression */:
                        return getAssignedTypeOfBinaryExpression(parent);
                    case 213 /* DeleteExpression */:
                        return undefinedType;
                    case 202 /* ArrayLiteralExpression */:
                        return getAssignedTypeOfArrayLiteralElement(parent, node);
                    case 223 /* SpreadElement */:
                        return getAssignedTypeOfSpreadExpression(parent);
                    case 291 /* PropertyAssignment */:
                        return getAssignedTypeOfPropertyAssignment(parent);
                    case 292 /* ShorthandPropertyAssignment */:
                        return getAssignedTypeOfShorthandPropertyAssignment(parent);
                }
                return errorType;
            }
            function getInitialTypeOfBindingElement(node) {
                var pattern = node.parent;
                var parentType = getInitialType(pattern.parent);
                var type = pattern.kind === 199 /* ObjectBindingPattern */ ?
                    getTypeOfDestructuredProperty(parentType, node.propertyName || node.name) :
                    !node.dotDotDotToken ?
                        getTypeOfDestructuredArrayElement(parentType, pattern.elements.indexOf(node)) :
                        getTypeOfDestructuredSpreadExpression(parentType);
                return getTypeWithDefault(type, node.initializer);
            }
            function getTypeOfInitializer(node) {
                // Return the cached type if one is available. If the type of the variable was inferred
                // from its initializer, we'll already have cached the type. Otherwise we compute it now
                // without caching such that transient types are reflected.
                var links = getNodeLinks(node);
                return links.resolvedType || getTypeOfExpression(node);
            }
            function getInitialTypeOfVariableDeclaration(node) {
                if (node.initializer) {
                    return getTypeOfInitializer(node.initializer);
                }
                if (node.parent.parent.kind === 241 /* ForInStatement */) {
                    return stringType;
                }
                if (node.parent.parent.kind === 242 /* ForOfStatement */) {
                    return checkRightHandSideOfForOf(node.parent.parent) || errorType;
                }
                return errorType;
            }
            function getInitialType(node) {
                return node.kind === 252 /* VariableDeclaration */ ?
                    getInitialTypeOfVariableDeclaration(node) :
                    getInitialTypeOfBindingElement(node);
            }
            function isEmptyArrayAssignment(node) {
                return node.kind === 252 /* VariableDeclaration */ && node.initializer &&
                    isEmptyArrayLiteral(node.initializer) ||
                    node.kind !== 201 /* BindingElement */ && node.parent.kind === 219 /* BinaryExpression */ &&
                        isEmptyArrayLiteral(node.parent.right);
            }
            function getReferenceCandidate(node) {
                switch (node.kind) {
                    case 210 /* ParenthesizedExpression */:
                        return getReferenceCandidate(node.expression);
                    case 219 /* BinaryExpression */:
                        switch (node.operatorToken.kind) {
                            case 63 /* EqualsToken */:
                            case 75 /* BarBarEqualsToken */:
                            case 76 /* AmpersandAmpersandEqualsToken */:
                            case 77 /* QuestionQuestionEqualsToken */:
                                return getReferenceCandidate(node.left);
                            case 27 /* CommaToken */:
                                return getReferenceCandidate(node.right);
                        }
                }
                return node;
            }
            function getReferenceRoot(node) {
                var parent = node.parent;
                return parent.kind === 210 /* ParenthesizedExpression */ ||
                    parent.kind === 219 /* BinaryExpression */ && parent.operatorToken.kind === 63 /* EqualsToken */ && parent.left === node ||
                    parent.kind === 219 /* BinaryExpression */ && parent.operatorToken.kind === 27 /* CommaToken */ && parent.right === node ?
                    getReferenceRoot(parent) : node;
            }
            function getTypeOfSwitchClause(clause) {
                if (clause.kind === 287 /* CaseClause */) {
                    return getRegularTypeOfLiteralType(getTypeOfExpression(clause.expression));
                }
                return neverType;
            }
            function getSwitchClauseTypes(switchStatement) {
                var links = getNodeLinks(switchStatement);
                if (!links.switchTypes) {
                    links.switchTypes = [];
                    for (var _i = 0, _a = switchStatement.caseBlock.clauses; _i < _a.length; _i++) {
                        var clause = _a[_i];
                        links.switchTypes.push(getTypeOfSwitchClause(clause));
                    }
                }
                return links.switchTypes;
            }
            function getSwitchClauseTypeOfWitnesses(switchStatement, retainDefault) {
                var witnesses = [];
                for (var _i = 0, _a = switchStatement.caseBlock.clauses; _i < _a.length; _i++) {
                    var clause = _a[_i];
                    if (clause.kind === 287 /* CaseClause */) {
                        if (ts.isStringLiteralLike(clause.expression)) {
                            witnesses.push(clause.expression.text);
                            continue;
                        }
                        return ts.emptyArray;
                    }
                    if (retainDefault)
                        witnesses.push(/*explicitDefaultStatement*/ undefined);
                }
                return witnesses;
            }
            function eachTypeContainedIn(source, types) {
                return source.flags & 1048576 /* Union */ ? !ts.forEach(source.types, function (t) { return !ts.contains(types, t); }) : ts.contains(types, source);
            }
            function isTypeSubsetOf(source, target) {
                return source === target || target.flags & 1048576 /* Union */ && isTypeSubsetOfUnion(source, target);
            }
            function isTypeSubsetOfUnion(source, target) {
                if (source.flags & 1048576 /* Union */) {
                    for (var _i = 0, _a = source.types; _i < _a.length; _i++) {
                        var t = _a[_i];
                        if (!containsType(target.types, t)) {
                            return false;
                        }
                    }
                    return true;
                }
                if (source.flags & 1024 /* EnumLiteral */ && getBaseTypeOfEnumLiteralType(source) === target) {
                    return true;
                }
                return containsType(target.types, source);
            }
            function forEachType(type, f) {
                return type.flags & 1048576 /* Union */ ? ts.forEach(type.types, f) : f(type);
            }
            function someType(type, f) {
                return type.flags & 1048576 /* Union */ ? ts.some(type.types, f) : f(type);
            }
            function everyType(type, f) {
                return type.flags & 1048576 /* Union */ ? ts.every(type.types, f) : f(type);
            }
            function everyContainedType(type, f) {
                return type.flags & 3145728 /* UnionOrIntersection */ ? ts.every(type.types, f) : f(type);
            }
            function filterType(type, f) {
                if (type.flags & 1048576 /* Union */) {
                    var types = type.types;
                    var filtered = ts.filter(types, f);
                    if (filtered === types) {
                        return type;
                    }
                    var origin = type.origin;
                    var newOrigin = void 0;
                    if (origin && origin.flags & 1048576 /* Union */) {
                        // If the origin type is a (denormalized) union type, filter its non-union constituents. If that ends
                        // up removing a smaller number of types than in the normalized constituent set (meaning some of the
                        // filtered types are within nested unions in the origin), then we can't construct a new origin type.
                        // Otherwise, if we have exactly one type left in the origin set, return that as the filtered type.
                        // Otherwise, construct a new filtered origin type.
                        var originTypes = origin.types;
                        var originFiltered = ts.filter(originTypes, function (t) { return !!(t.flags & 1048576 /* Union */) || f(t); });
                        if (originTypes.length - originFiltered.length === types.length - filtered.length) {
                            if (originFiltered.length === 1) {
                                return originFiltered[0];
                            }
                            newOrigin = createOriginUnionOrIntersectionType(1048576 /* Union */, originFiltered);
                        }
                    }
                    return getUnionTypeFromSortedList(filtered, type.objectFlags, /*aliasSymbol*/ undefined, /*aliasTypeArguments*/ undefined, newOrigin);
                }
                return type.flags & 131072 /* Never */ || f(type) ? type : neverType;
            }
            function removeType(type, targetType) {
                return filterType(type, function (t) { return t !== targetType; });
            }
            function countTypes(type) {
                return type.flags & 1048576 /* Union */ ? type.types.length : 1;
            }
            function mapType(type, mapper, noReductions) {
                if (type.flags & 131072 /* Never */) {
                    return type;
                }
                if (!(type.flags & 1048576 /* Union */)) {
                    return mapper(type);
                }
                var origin = type.origin;
                var types = origin && origin.flags & 1048576 /* Union */ ? origin.types : type.types;
                var mappedTypes;
                var changed = false;
                for (var _i = 0, types_17 = types; _i < types_17.length; _i++) {
                    var t = types_17[_i];
                    var mapped = t.flags & 1048576 /* Union */ ? mapType(t, mapper, noReductions) : mapper(t);
                    changed || (changed = t !== mapped);
                    if (mapped) {
                        if (!mappedTypes) {
                            mappedTypes = [mapped];
                        }
                        else {
                            mappedTypes.push(mapped);
                        }
                    }
                }
                return changed ? mappedTypes && getUnionType(mappedTypes, noReductions ? 0 /* None */ : 1 /* Literal */) : type;
            }
            function mapTypeWithAlias(type, mapper, aliasSymbol, aliasTypeArguments) {
                return type.flags & 1048576 /* Union */ && aliasSymbol ?
                    getUnionType(ts.map(type.types, mapper), 1 /* Literal */, aliasSymbol, aliasTypeArguments) :
                    mapType(type, mapper);
            }
            function getConstituentCount(type) {
                return type.flags & 3145728 /* UnionOrIntersection */ ? type.types.length : 1;
            }
            function extractTypesOfKind(type, kind) {
                return filterType(type, function (t) { return (t.flags & kind) !== 0; });
            }
            // Return a new type in which occurrences of the string and number primitive types in
            // typeWithPrimitives have been replaced with occurrences of string literals and numeric
            // literals in typeWithLiterals, respectively.
            function replacePrimitivesWithLiterals(typeWithPrimitives, typeWithLiterals) {
                if (isTypeSubsetOf(stringType, typeWithPrimitives) && maybeTypeOfKind(typeWithLiterals, 128 /* StringLiteral */) ||
                    isTypeSubsetOf(numberType, typeWithPrimitives) && maybeTypeOfKind(typeWithLiterals, 256 /* NumberLiteral */) ||
                    isTypeSubsetOf(bigintType, typeWithPrimitives) && maybeTypeOfKind(typeWithLiterals, 2048 /* BigIntLiteral */)) {
                    return mapType(typeWithPrimitives, function (t) {
                        return t.flags & 4 /* String */ ? extractTypesOfKind(typeWithLiterals, 4 /* String */ | 128 /* StringLiteral */) :
                            t.flags & 8 /* Number */ ? extractTypesOfKind(typeWithLiterals, 8 /* Number */ | 256 /* NumberLiteral */) :
                                t.flags & 64 /* BigInt */ ? extractTypesOfKind(typeWithLiterals, 64 /* BigInt */ | 2048 /* BigIntLiteral */) : t;
                    });
                }
                return typeWithPrimitives;
            }
            function isIncomplete(flowType) {
                return flowType.flags === 0;
            }
            function getTypeFromFlowType(flowType) {
                return flowType.flags === 0 ? flowType.type : flowType;
            }
            function createFlowType(type, incomplete) {
                return incomplete ? { flags: 0, type: type.flags & 131072 /* Never */ ? silentNeverType : type } : type;
            }
            // An evolving array type tracks the element types that have so far been seen in an
            // 'x.push(value)' or 'x[n] = value' operation along the control flow graph. Evolving
            // array types are ultimately converted into manifest array types (using getFinalArrayType)
            // and never escape the getFlowTypeOfReference function.
            function createEvolvingArrayType(elementType) {
                var result = createObjectType(256 /* EvolvingArray */);
                result.elementType = elementType;
                return result;
            }
            function getEvolvingArrayType(elementType) {
                return evolvingArrayTypes[elementType.id] || (evolvingArrayTypes[elementType.id] = createEvolvingArrayType(elementType));
            }
            // When adding evolving array element types we do not perform subtype reduction. Instead,
            // we defer subtype reduction until the evolving array type is finalized into a manifest
            // array type.
            function addEvolvingArrayElementType(evolvingArrayType, node) {
                var elementType = getRegularTypeOfObjectLiteral(getBaseTypeOfLiteralType(getContextFreeTypeOfExpression(node)));
                return isTypeSubsetOf(elementType, evolvingArrayType.elementType) ? evolvingArrayType : getEvolvingArrayType(getUnionType([evolvingArrayType.elementType, elementType]));
            }
            function createFinalArrayType(elementType) {
                return elementType.flags & 131072 /* Never */ ?
                    autoArrayType :
                    createArrayType(elementType.flags & 1048576 /* Union */ ?
                        getUnionType(elementType.types, 2 /* Subtype */) :
                        elementType);
            }
            // We perform subtype reduction upon obtaining the final array type from an evolving array type.
            function getFinalArrayType(evolvingArrayType) {
                return evolvingArrayType.finalArrayType || (evolvingArrayType.finalArrayType = createFinalArrayType(evolvingArrayType.elementType));
            }
            function finalizeEvolvingArrayType(type) {
                return ts.getObjectFlags(type) & 256 /* EvolvingArray */ ? getFinalArrayType(type) : type;
            }
            function getElementTypeOfEvolvingArrayType(type) {
                return ts.getObjectFlags(type) & 256 /* EvolvingArray */ ? type.elementType : neverType;
            }
            function isEvolvingArrayTypeList(types) {
                var hasEvolvingArrayType = false;
                for (var _i = 0, types_18 = types; _i < types_18.length; _i++) {
                    var t = types_18[_i];
                    if (!(t.flags & 131072 /* Never */)) {
                        if (!(ts.getObjectFlags(t) & 256 /* EvolvingArray */)) {
                            return false;
                        }
                        hasEvolvingArrayType = true;
                    }
                }
                return hasEvolvingArrayType;
            }
            // Return true if the given node is 'x' in an 'x.length', x.push(value)', 'x.unshift(value)' or
            // 'x[n] = value' operation, where 'n' is an expression of type any, undefined, or a number-like type.
            function isEvolvingArrayOperationTarget(node) {
                var root = getReferenceRoot(node);
                var parent = root.parent;
                var isLengthPushOrUnshift = ts.isPropertyAccessExpression(parent) && (parent.name.escapedText === "length" ||
                    parent.parent.kind === 206 /* CallExpression */
                        && ts.isIdentifier(parent.name)
                        && ts.isPushOrUnshiftIdentifier(parent.name));
                var isElementAssignment = parent.kind === 205 /* ElementAccessExpression */ &&
                    parent.expression === root &&
                    parent.parent.kind === 219 /* BinaryExpression */ &&
                    parent.parent.operatorToken.kind === 63 /* EqualsToken */ &&
                    parent.parent.left === parent &&
                    !ts.isAssignmentTarget(parent.parent) &&
                    isTypeAssignableToKind(getTypeOfExpression(parent.argumentExpression), 296 /* NumberLike */);
                return isLengthPushOrUnshift || isElementAssignment;
            }
            function isDeclarationWithExplicitTypeAnnotation(declaration) {
                return (declaration.kind === 252 /* VariableDeclaration */ || declaration.kind === 162 /* Parameter */ ||
                    declaration.kind === 165 /* PropertyDeclaration */ || declaration.kind === 164 /* PropertySignature */) &&
                    !!ts.getEffectiveTypeAnnotationNode(declaration);
            }
            function getExplicitTypeOfSymbol(symbol, diagnostic) {
                if (symbol.flags & (16 /* Function */ | 8192 /* Method */ | 32 /* Class */ | 512 /* ValueModule */)) {
                    return getTypeOfSymbol(symbol);
                }
                if (symbol.flags & (3 /* Variable */ | 4 /* Property */)) {
                    if (ts.getCheckFlags(symbol) & 262144 /* Mapped */) {
                        var origin = symbol.syntheticOrigin;
                        if (origin && getExplicitTypeOfSymbol(origin)) {
                            return getTypeOfSymbol(symbol);
                        }
                    }
                    var declaration = symbol.valueDeclaration;
                    if (declaration) {
                        if (isDeclarationWithExplicitTypeAnnotation(declaration)) {
                            return getTypeOfSymbol(symbol);
                        }
                        if (ts.isVariableDeclaration(declaration) && declaration.parent.parent.kind === 242 /* ForOfStatement */) {
                            var statement = declaration.parent.parent;
                            var expressionType = getTypeOfDottedName(statement.expression, /*diagnostic*/ undefined);
                            if (expressionType) {
                                var use = statement.awaitModifier ? 15 /* ForAwaitOf */ : 13 /* ForOf */;
                                return checkIteratedTypeOrElementType(use, expressionType, undefinedType, /*errorNode*/ undefined);
                            }
                        }
                        if (diagnostic) {
                            ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(declaration, ts.Diagnostics._0_needs_an_explicit_type_annotation, symbolToString(symbol)));
                        }
                    }
                }
            }
            // We require the dotted function name in an assertion expression to be comprised of identifiers
            // that reference function, method, class or value module symbols; or variable, property or
            // parameter symbols with declarations that have explicit type annotations. Such references are
            // resolvable with no possibility of triggering circularities in control flow analysis.
            function getTypeOfDottedName(node, diagnostic) {
                if (!(node.flags & 16777216 /* InWithStatement */)) {
                    switch (node.kind) {
                        case 79 /* Identifier */:
                            var symbol = getExportSymbolOfValueSymbolIfExported(getResolvedSymbol(node));
                            return getExplicitTypeOfSymbol(symbol.flags & 2097152 /* Alias */ ? resolveAlias(symbol) : symbol, diagnostic);
                        case 108 /* ThisKeyword */:
                            return getExplicitThisType(node);
                        case 106 /* SuperKeyword */:
                            return checkSuperExpression(node);
                        case 204 /* PropertyAccessExpression */: {
                            var type = getTypeOfDottedName(node.expression, diagnostic);
                            if (type) {
                                var name = node.name;
                                var prop = void 0;
                                if (ts.isPrivateIdentifier(name)) {
                                    if (!type.symbol) {
                                        return undefined;
                                    }
                                    prop = getPropertyOfType(type, ts.getSymbolNameForPrivateIdentifier(type.symbol, name.escapedText));
                                }
                                else {
                                    prop = getPropertyOfType(type, name.escapedText);
                                }
                                return prop && getExplicitTypeOfSymbol(prop, diagnostic);
                            }
                            return undefined;
                        }
                        case 210 /* ParenthesizedExpression */:
                            return getTypeOfDottedName(node.expression, diagnostic);
                    }
                }
            }
            function getEffectsSignature(node) {
                var links = getNodeLinks(node);
                var signature = links.effectsSignature;
                if (signature === undefined) {
                    // A call expression parented by an expression statement is a potential assertion. Other call
                    // expressions are potential type predicate function calls. In order to avoid triggering
                    // circularities in control flow analysis, we use getTypeOfDottedName when resolving the call
                    // target expression of an assertion.
                    var funcType = void 0;
                    if (node.parent.kind === 236 /* ExpressionStatement */) {
                        funcType = getTypeOfDottedName(node.expression, /*diagnostic*/ undefined);
                    }
                    else if (node.expression.kind !== 106 /* SuperKeyword */) {
                        if (ts.isOptionalChain(node)) {
                            funcType = checkNonNullType(getOptionalExpressionType(checkExpression(node.expression), node.expression), node.expression);
                        }
                        else {
                            funcType = checkNonNullExpression(node.expression);
                        }
                    }
                    var signatures = getSignaturesOfType(funcType && getApparentType(funcType) || unknownType, 0 /* Call */);
                    var candidate = signatures.length === 1 && !signatures[0].typeParameters ? signatures[0] :
                        ts.some(signatures, hasTypePredicateOrNeverReturnType) ? getResolvedSignature(node) :
                            undefined;
                    signature = links.effectsSignature = candidate && hasTypePredicateOrNeverReturnType(candidate) ? candidate : unknownSignature;
                }
                return signature === unknownSignature ? undefined : signature;
            }
            function hasTypePredicateOrNeverReturnType(signature) {
                return !!(getTypePredicateOfSignature(signature) ||
                    signature.declaration && (getReturnTypeFromAnnotation(signature.declaration) || unknownType).flags & 131072 /* Never */);
            }
            function getTypePredicateArgument(predicate, callExpression) {
                if (predicate.kind === 1 /* Identifier */ || predicate.kind === 3 /* AssertsIdentifier */) {
                    return callExpression.arguments[predicate.parameterIndex];
                }
                var invokedExpression = ts.skipParentheses(callExpression.expression);
                return ts.isAccessExpression(invokedExpression) ? ts.skipParentheses(invokedExpression.expression) : undefined;
            }
            function reportFlowControlError(node) {
                var block = ts.findAncestor(node, ts.isFunctionOrModuleBlock);
                var sourceFile = ts.getSourceFileOfNode(node);
                var span = ts.getSpanOfTokenAtPosition(sourceFile, block.statements.pos);
                diagnostics.add(ts.createFileDiagnostic(sourceFile, span.start, span.length, ts.Diagnostics.The_containing_function_or_module_body_is_too_large_for_control_flow_analysis));
            }
            function isReachableFlowNode(flow) {
                var result = isReachableFlowNodeWorker(flow, /*noCacheCheck*/ false);
                lastFlowNode = flow;
                lastFlowNodeReachable = result;
                return result;
            }
            function isFalseExpression(expr) {
                var node = ts.skipParentheses(expr);
                return node.kind === 95 /* FalseKeyword */ || node.kind === 219 /* BinaryExpression */ && (node.operatorToken.kind === 55 /* AmpersandAmpersandToken */ && (isFalseExpression(node.left) || isFalseExpression(node.right)) ||
                    node.operatorToken.kind === 56 /* BarBarToken */ && isFalseExpression(node.left) && isFalseExpression(node.right));
            }
            function isReachableFlowNodeWorker(flow, noCacheCheck) {
                while (true) {
                    if (flow === lastFlowNode) {
                        return lastFlowNodeReachable;
                    }
                    var flags = flow.flags;
                    if (flags & 4096 /* Shared */) {
                        if (!noCacheCheck) {
                            var id = getFlowNodeId(flow);
                            var reachable = flowNodeReachable[id];
                            return reachable !== undefined ? reachable : (flowNodeReachable[id] = isReachableFlowNodeWorker(flow, /*noCacheCheck*/ true));
                        }
                        noCacheCheck = false;
                    }
                    if (flags & (16 /* Assignment */ | 96 /* Condition */ | 256 /* ArrayMutation */)) {
                        flow = flow.antecedent;
                    }
                    else if (flags & 512 /* Call */) {
                        var signature = getEffectsSignature(flow.node);
                        if (signature) {
                            var predicate = getTypePredicateOfSignature(signature);
                            if (predicate && predicate.kind === 3 /* AssertsIdentifier */ && !predicate.type) {
                                var predicateArgument = flow.node.arguments[predicate.parameterIndex];
                                if (predicateArgument && isFalseExpression(predicateArgument)) {
                                    return false;
                                }
                            }
                            if (getReturnTypeOfSignature(signature).flags & 131072 /* Never */) {
                                return false;
                            }
                        }
                        flow = flow.antecedent;
                    }
                    else if (flags & 4 /* BranchLabel */) {
                        // A branching point is reachable if any branch is reachable.
                        return ts.some(flow.antecedents, function (f) { return isReachableFlowNodeWorker(f, /*noCacheCheck*/ false); });
                    }
                    else if (flags & 8 /* LoopLabel */) {
                        var antecedents = flow.antecedents;
                        if (antecedents === undefined || antecedents.length === 0) {
                            return false;
                        }
                        // A loop is reachable if the control flow path that leads to the top is reachable.
                        flow = antecedents[0];
                    }
                    else if (flags & 128 /* SwitchClause */) {
                        // The control flow path representing an unmatched value in a switch statement with
                        // no default clause is unreachable if the switch statement is exhaustive.
                        if (flow.clauseStart === flow.clauseEnd && isExhaustiveSwitchStatement(flow.switchStatement)) {
                            return false;
                        }
                        flow = flow.antecedent;
                    }
                    else if (flags & 1024 /* ReduceLabel */) {
                        // Cache is unreliable once we start adjusting labels
                        lastFlowNode = undefined;
                        var target = flow.target;
                        var saveAntecedents = target.antecedents;
                        target.antecedents = flow.antecedents;
                        var result = isReachableFlowNodeWorker(flow.antecedent, /*noCacheCheck*/ false);
                        target.antecedents = saveAntecedents;
                        return result;
                    }
                    else {
                        return !(flags & 1 /* Unreachable */);
                    }
                }
            }
            // Return true if the given flow node is preceded by a 'super(...)' call in every possible code path
            // leading to the node.
            function isPostSuperFlowNode(flow, noCacheCheck) {
                while (true) {
                    var flags = flow.flags;
                    if (flags & 4096 /* Shared */) {
                        if (!noCacheCheck) {
                            var id = getFlowNodeId(flow);
                            var postSuper = flowNodePostSuper[id];
                            return postSuper !== undefined ? postSuper : (flowNodePostSuper[id] = isPostSuperFlowNode(flow, /*noCacheCheck*/ true));
                        }
                        noCacheCheck = false;
                    }
                    if (flags & (16 /* Assignment */ | 96 /* Condition */ | 256 /* ArrayMutation */ | 128 /* SwitchClause */)) {
                        flow = flow.antecedent;
                    }
                    else if (flags & 512 /* Call */) {
                        if (flow.node.expression.kind === 106 /* SuperKeyword */) {
                            return true;
                        }
                        flow = flow.antecedent;
                    }
                    else if (flags & 4 /* BranchLabel */) {
                        // A branching point is post-super if every branch is post-super.
                        return ts.every(flow.antecedents, function (f) { return isPostSuperFlowNode(f, /*noCacheCheck*/ false); });
                    }
                    else if (flags & 8 /* LoopLabel */) {
                        // A loop is post-super if the control flow path that leads to the top is post-super.
                        flow = flow.antecedents[0];
                    }
                    else if (flags & 1024 /* ReduceLabel */) {
                        var target = flow.target;
                        var saveAntecedents = target.antecedents;
                        target.antecedents = flow.antecedents;
                        var result = isPostSuperFlowNode(flow.antecedent, /*noCacheCheck*/ false);
                        target.antecedents = saveAntecedents;
                        return result;
                    }
                    else {
                        // Unreachable nodes are considered post-super to silence errors
                        return !!(flags & 1 /* Unreachable */);
                    }
                }
            }
            function isConstantReference(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        var symbol = getResolvedSymbol(node);
                        return isConstVariable(symbol) || !!symbol.valueDeclaration && ts.getRootDeclaration(symbol.valueDeclaration).kind === 162 /* Parameter */ && !isParameterAssigned(symbol);
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */:
                        // The resolvedSymbol property is initialized by checkPropertyAccess or checkElementAccess before we get here.
                        return isConstantReference(node.expression) && isReadonlySymbol(getNodeLinks(node).resolvedSymbol || unknownSymbol);
                }
                return false;
            }
            function getFlowTypeOfReference(reference, declaredType, initialType, flowContainer) {
                if (initialType === void 0) {
                    initialType = declaredType;
                }
                var key;
                var isKeySet = false;
                var flowDepth = 0;
                var inlineLevel = 0;
                if (flowAnalysisDisabled) {
                    return errorType;
                }
                if (!reference.flowNode) {
                    return declaredType;
                }
                flowInvocationCount++;
                var sharedFlowStart = sharedFlowCount;
                var evolvedType = getTypeFromFlowType(getTypeAtFlowNode(reference.flowNode));
                sharedFlowCount = sharedFlowStart;
                // When the reference is 'x' in an 'x.length', 'x.push(value)', 'x.unshift(value)' or x[n] = value' operation,
                // we give type 'any[]' to 'x' instead of using the type determined by control flow analysis such that operations
                // on empty arrays are possible without implicit any errors and new element types can be inferred without
                // type mismatch errors.
                var resultType = ts.getObjectFlags(evolvedType) & 256 /* EvolvingArray */ && isEvolvingArrayOperationTarget(reference) ? autoArrayType : finalizeEvolvingArrayType(evolvedType);
                if (resultType === unreachableNeverType || reference.parent && reference.parent.kind === 228 /* NonNullExpression */ && !(resultType.flags & 131072 /* Never */) && getTypeWithFacts(resultType, 2097152 /* NEUndefinedOrNull */).flags & 131072 /* Never */) {
                    return declaredType;
                }
                return resultType;
                function getOrSetCacheKey() {
                    if (isKeySet) {
                        return key;
                    }
                    isKeySet = true;
                    return key = getFlowCacheKey(reference, declaredType, initialType, flowContainer);
                }
                function getTypeAtFlowNode(flow) {
                    if (flowDepth === 2000) {
                        // We have made 2000 recursive invocations. To avoid overflowing the call stack we report an error
                        // and disable further control flow analysis in the containing function or module body.
                        ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.instant("checkTypes" /* CheckTypes */, "getTypeAtFlowNode_DepthLimit", { flowId: flow.id });
                        flowAnalysisDisabled = true;
                        reportFlowControlError(reference);
                        return errorType;
                    }
                    flowDepth++;
                    var sharedFlow;
                    while (true) {
                        var flags = flow.flags;
                        if (flags & 4096 /* Shared */) {
                            // We cache results of flow type resolution for shared nodes that were previously visited in
                            // the same getFlowTypeOfReference invocation. A node is considered shared when it is the
                            // antecedent of more than one node.
                            for (var i = sharedFlowStart; i < sharedFlowCount; i++) {
                                if (sharedFlowNodes[i] === flow) {
                                    flowDepth--;
                                    return sharedFlowTypes[i];
                                }
                            }
                            sharedFlow = flow;
                        }
                        var type = void 0;
                        if (flags & 16 /* Assignment */) {
                            type = getTypeAtFlowAssignment(flow);
                            if (!type) {
                                flow = flow.antecedent;
                                continue;
                            }
                        }
                        else if (flags & 512 /* Call */) {
                            type = getTypeAtFlowCall(flow);
                            if (!type) {
                                flow = flow.antecedent;
                                continue;
                            }
                        }
                        else if (flags & 96 /* Condition */) {
                            type = getTypeAtFlowCondition(flow);
                        }
                        else if (flags & 128 /* SwitchClause */) {
                            type = getTypeAtSwitchClause(flow);
                        }
                        else if (flags & 12 /* Label */) {
                            if (flow.antecedents.length === 1) {
                                flow = flow.antecedents[0];
                                continue;
                            }
                            type = flags & 4 /* BranchLabel */ ?
                                getTypeAtFlowBranchLabel(flow) :
                                getTypeAtFlowLoopLabel(flow);
                        }
                        else if (flags & 256 /* ArrayMutation */) {
                            type = getTypeAtFlowArrayMutation(flow);
                            if (!type) {
                                flow = flow.antecedent;
                                continue;
                            }
                        }
                        else if (flags & 1024 /* ReduceLabel */) {
                            var target = flow.target;
                            var saveAntecedents = target.antecedents;
                            target.antecedents = flow.antecedents;
                            type = getTypeAtFlowNode(flow.antecedent);
                            target.antecedents = saveAntecedents;
                        }
                        else if (flags & 2 /* Start */) {
                            // Check if we should continue with the control flow of the containing function.
                            var container = flow.node;
                            if (container && container !== flowContainer &&
                                reference.kind !== 204 /* PropertyAccessExpression */ &&
                                reference.kind !== 205 /* ElementAccessExpression */ &&
                                reference.kind !== 108 /* ThisKeyword */) {
                                flow = container.flowNode;
                                continue;
                            }
                            // At the top of the flow we have the initial type.
                            type = initialType;
                        }
                        else {
                            // Unreachable code errors are reported in the binding phase. Here we
                            // simply return the non-auto declared type to reduce follow-on errors.
                            type = convertAutoToAny(declaredType);
                        }
                        if (sharedFlow) {
                            // Record visited node and the associated type in the cache.
                            sharedFlowNodes[sharedFlowCount] = sharedFlow;
                            sharedFlowTypes[sharedFlowCount] = type;
                            sharedFlowCount++;
                        }
                        flowDepth--;
                        return type;
                    }
                }
                function getInitialOrAssignedType(flow) {
                    var node = flow.node;
                    return getNarrowableTypeForReference(node.kind === 252 /* VariableDeclaration */ || node.kind === 201 /* BindingElement */ ?
                        getInitialType(node) :
                        getAssignedType(node), reference);
                }
                function getTypeAtFlowAssignment(flow) {
                    var node = flow.node;
                    // Assignments only narrow the computed type if the declared type is a union type. Thus, we
                    // only need to evaluate the assigned type if the declared type is a union type.
                    if (isMatchingReference(reference, node)) {
                        if (!isReachableFlowNode(flow)) {
                            return unreachableNeverType;
                        }
                        if (ts.getAssignmentTargetKind(node) === 2 /* Compound */) {
                            var flowType = getTypeAtFlowNode(flow.antecedent);
                            return createFlowType(getBaseTypeOfLiteralType(getTypeFromFlowType(flowType)), isIncomplete(flowType));
                        }
                        if (declaredType === autoType || declaredType === autoArrayType) {
                            if (isEmptyArrayAssignment(node)) {
                                return getEvolvingArrayType(neverType);
                            }
                            var assignedType = getWidenedLiteralType(getInitialOrAssignedType(flow));
                            return isTypeAssignableTo(assignedType, declaredType) ? assignedType : anyArrayType;
                        }
                        if (declaredType.flags & 1048576 /* Union */) {
                            return getAssignmentReducedType(declaredType, getInitialOrAssignedType(flow));
                        }
                        return declaredType;
                    }
                    // We didn't have a direct match. However, if the reference is a dotted name, this
                    // may be an assignment to a left hand part of the reference. For example, for a
                    // reference 'x.y.z', we may be at an assignment to 'x.y' or 'x'. In that case,
                    // return the declared type.
                    if (containsMatchingReference(reference, node)) {
                        if (!isReachableFlowNode(flow)) {
                            return unreachableNeverType;
                        }
                        // A matching dotted name might also be an expando property on a function *expression*,
                        // in which case we continue control flow analysis back to the function's declaration
                        if (ts.isVariableDeclaration(node) && (ts.isInJSFile(node) || ts.isVarConst(node))) {
                            var init = ts.getDeclaredExpandoInitializer(node);
                            if (init && (init.kind === 211 /* FunctionExpression */ || init.kind === 212 /* ArrowFunction */)) {
                                return getTypeAtFlowNode(flow.antecedent);
                            }
                        }
                        return declaredType;
                    }
                    // for (const _ in ref) acts as a nonnull on ref
                    if (ts.isVariableDeclaration(node) && node.parent.parent.kind === 241 /* ForInStatement */ && isMatchingReference(reference, node.parent.parent.expression)) {
                        return getNonNullableTypeIfNeeded(getTypeFromFlowType(getTypeAtFlowNode(flow.antecedent)));
                    }
                    // Assignment doesn't affect reference
                    return undefined;
                }
                function narrowTypeByAssertion(type, expr) {
                    var node = ts.skipParentheses(expr);
                    if (node.kind === 95 /* FalseKeyword */) {
                        return unreachableNeverType;
                    }
                    if (node.kind === 219 /* BinaryExpression */) {
                        if (node.operatorToken.kind === 55 /* AmpersandAmpersandToken */) {
                            return narrowTypeByAssertion(narrowTypeByAssertion(type, node.left), node.right);
                        }
                        if (node.operatorToken.kind === 56 /* BarBarToken */) {
                            return getUnionType([narrowTypeByAssertion(type, node.left), narrowTypeByAssertion(type, node.right)]);
                        }
                    }
                    return narrowType(type, node, /*assumeTrue*/ true);
                }
                function getTypeAtFlowCall(flow) {
                    var signature = getEffectsSignature(flow.node);
                    if (signature) {
                        var predicate = getTypePredicateOfSignature(signature);
                        if (predicate && (predicate.kind === 2 /* AssertsThis */ || predicate.kind === 3 /* AssertsIdentifier */)) {
                            var flowType = getTypeAtFlowNode(flow.antecedent);
                            var type = finalizeEvolvingArrayType(getTypeFromFlowType(flowType));
                            var narrowedType = predicate.type ? narrowTypeByTypePredicate(type, predicate, flow.node, /*assumeTrue*/ true) :
                                predicate.kind === 3 /* AssertsIdentifier */ && predicate.parameterIndex >= 0 && predicate.parameterIndex < flow.node.arguments.length ? narrowTypeByAssertion(type, flow.node.arguments[predicate.parameterIndex]) :
                                    type;
                            return narrowedType === type ? flowType : createFlowType(narrowedType, isIncomplete(flowType));
                        }
                        if (getReturnTypeOfSignature(signature).flags & 131072 /* Never */) {
                            return unreachableNeverType;
                        }
                    }
                    return undefined;
                }
                function getTypeAtFlowArrayMutation(flow) {
                    if (declaredType === autoType || declaredType === autoArrayType) {
                        var node = flow.node;
                        var expr = node.kind === 206 /* CallExpression */ ?
                            node.expression.expression :
                            node.left.expression;
                        if (isMatchingReference(reference, getReferenceCandidate(expr))) {
                            var flowType = getTypeAtFlowNode(flow.antecedent);
                            var type = getTypeFromFlowType(flowType);
                            if (ts.getObjectFlags(type) & 256 /* EvolvingArray */) {
                                var evolvedType_1 = type;
                                if (node.kind === 206 /* CallExpression */) {
                                    for (var _i = 0, _a = node.arguments; _i < _a.length; _i++) {
                                        var arg = _a[_i];
                                        evolvedType_1 = addEvolvingArrayElementType(evolvedType_1, arg);
                                    }
                                }
                                else {
                                    // We must get the context free expression type so as to not recur in an uncached fashion on the LHS (which causes exponential blowup in compile time)
                                    var indexType = getContextFreeTypeOfExpression(node.left.argumentExpression);
                                    if (isTypeAssignableToKind(indexType, 296 /* NumberLike */)) {
                                        evolvedType_1 = addEvolvingArrayElementType(evolvedType_1, node.right);
                                    }
                                }
                                return evolvedType_1 === type ? flowType : createFlowType(evolvedType_1, isIncomplete(flowType));
                            }
                            return flowType;
                        }
                    }
                    return undefined;
                }
                function getTypeAtFlowCondition(flow) {
                    var flowType = getTypeAtFlowNode(flow.antecedent);
                    var type = getTypeFromFlowType(flowType);
                    if (type.flags & 131072 /* Never */) {
                        return flowType;
                    }
                    // If we have an antecedent type (meaning we're reachable in some way), we first
                    // attempt to narrow the antecedent type. If that produces the never type, and if
                    // the antecedent type is incomplete (i.e. a transient type in a loop), then we
                    // take the type guard as an indication that control *could* reach here once we
                    // have the complete type. We proceed by switching to the silent never type which
                    // doesn't report errors when operators are applied to it. Note that this is the
                    // *only* place a silent never type is ever generated.
                    var assumeTrue = (flow.flags & 32 /* TrueCondition */) !== 0;
                    var nonEvolvingType = finalizeEvolvingArrayType(type);
                    var narrowedType = narrowType(nonEvolvingType, flow.node, assumeTrue);
                    if (narrowedType === nonEvolvingType) {
                        return flowType;
                    }
                    return createFlowType(narrowedType, isIncomplete(flowType));
                }
                function getTypeAtSwitchClause(flow) {
                    var expr = flow.switchStatement.expression;
                    var flowType = getTypeAtFlowNode(flow.antecedent);
                    var type = getTypeFromFlowType(flowType);
                    if (isMatchingReference(reference, expr)) {
                        type = narrowTypeBySwitchOnDiscriminant(type, flow.switchStatement, flow.clauseStart, flow.clauseEnd);
                    }
                    else if (expr.kind === 214 /* TypeOfExpression */ && isMatchingReference(reference, expr.expression)) {
                        type = narrowBySwitchOnTypeOf(type, flow.switchStatement, flow.clauseStart, flow.clauseEnd);
                    }
                    else {
                        if (strictNullChecks) {
                            if (optionalChainContainsReference(expr, reference)) {
                                type = narrowTypeBySwitchOptionalChainContainment(type, flow.switchStatement, flow.clauseStart, flow.clauseEnd, function (t) { return !(t.flags & (32768 /* Undefined */ | 131072 /* Never */)); });
                            }
                            else if (expr.kind === 214 /* TypeOfExpression */ && optionalChainContainsReference(expr.expression, reference)) {
                                type = narrowTypeBySwitchOptionalChainContainment(type, flow.switchStatement, flow.clauseStart, flow.clauseEnd, function (t) { return !(t.flags & 131072 /* Never */ || t.flags & 128 /* StringLiteral */ && t.value === "undefined"); });
                            }
                        }
                        var access = getDiscriminantPropertyAccess(expr, type);
                        if (access) {
                            type = narrowTypeBySwitchOnDiscriminantProperty(type, access, flow.switchStatement, flow.clauseStart, flow.clauseEnd);
                        }
                    }
                    return createFlowType(type, isIncomplete(flowType));
                }
                function getTypeAtFlowBranchLabel(flow) {
                    var antecedentTypes = [];
                    var subtypeReduction = false;
                    var seenIncomplete = false;
                    var bypassFlow;
                    for (var _i = 0, _a = flow.antecedents; _i < _a.length; _i++) {
                        var antecedent = _a[_i];
                        if (!bypassFlow && antecedent.flags & 128 /* SwitchClause */ && antecedent.clauseStart === antecedent.clauseEnd) {
                            // The antecedent is the bypass branch of a potentially exhaustive switch statement.
                            bypassFlow = antecedent;
                            continue;
                        }
                        var flowType = getTypeAtFlowNode(antecedent);
                        var type = getTypeFromFlowType(flowType);
                        // If the type at a particular antecedent path is the declared type and the
                        // reference is known to always be assigned (i.e. when declared and initial types
                        // are the same), there is no reason to process more antecedents since the only
                        // possible outcome is subtypes that will be removed in the final union type anyway.
                        if (type === declaredType && declaredType === initialType) {
                            return type;
                        }
                        ts.pushIfUnique(antecedentTypes, type);
                        // If an antecedent type is not a subset of the declared type, we need to perform
                        // subtype reduction. This happens when a "foreign" type is injected into the control
                        // flow using the instanceof operator or a user defined type predicate.
                        if (!isTypeSubsetOf(type, declaredType)) {
                            subtypeReduction = true;
                        }
                        if (isIncomplete(flowType)) {
                            seenIncomplete = true;
                        }
                    }
                    if (bypassFlow) {
                        var flowType = getTypeAtFlowNode(bypassFlow);
                        var type = getTypeFromFlowType(flowType);
                        // If the bypass flow contributes a type we haven't seen yet and the switch statement
                        // isn't exhaustive, process the bypass flow type. Since exhaustiveness checks increase
                        // the risk of circularities, we only want to perform them when they make a difference.
                        if (!ts.contains(antecedentTypes, type) && !isExhaustiveSwitchStatement(bypassFlow.switchStatement)) {
                            if (type === declaredType && declaredType === initialType) {
                                return type;
                            }
                            antecedentTypes.push(type);
                            if (!isTypeSubsetOf(type, declaredType)) {
                                subtypeReduction = true;
                            }
                            if (isIncomplete(flowType)) {
                                seenIncomplete = true;
                            }
                        }
                    }
                    return createFlowType(getUnionOrEvolvingArrayType(antecedentTypes, subtypeReduction ? 2 /* Subtype */ : 1 /* Literal */), seenIncomplete);
                }
                function getTypeAtFlowLoopLabel(flow) {
                    // If we have previously computed the control flow type for the reference at
                    // this flow loop junction, return the cached type.
                    var id = getFlowNodeId(flow);
                    var cache = flowLoopCaches[id] || (flowLoopCaches[id] = new ts.Map());
                    var key = getOrSetCacheKey();
                    if (!key) {
                        // No cache key is generated when binding patterns are in unnarrowable situations
                        return declaredType;
                    }
                    var cached = cache.get(key);
                    if (cached) {
                        return cached;
                    }
                    // If this flow loop junction and reference are already being processed, return
                    // the union of the types computed for each branch so far, marked as incomplete.
                    // It is possible to see an empty array in cases where loops are nested and the
                    // back edge of the outer loop reaches an inner loop that is already being analyzed.
                    // In such cases we restart the analysis of the inner loop, which will then see
                    // a non-empty in-process array for the outer loop and eventually terminate because
                    // the first antecedent of a loop junction is always the non-looping control flow
                    // path that leads to the top.
                    for (var i = flowLoopStart; i < flowLoopCount; i++) {
                        if (flowLoopNodes[i] === flow && flowLoopKeys[i] === key && flowLoopTypes[i].length) {
                            return createFlowType(getUnionOrEvolvingArrayType(flowLoopTypes[i], 1 /* Literal */), /*incomplete*/ true);
                        }
                    }
                    // Add the flow loop junction and reference to the in-process stack and analyze
                    // each antecedent code path.
                    var antecedentTypes = [];
                    var subtypeReduction = false;
                    var firstAntecedentType;
                    for (var _i = 0, _a = flow.antecedents; _i < _a.length; _i++) {
                        var antecedent = _a[_i];
                        var flowType = void 0;
                        if (!firstAntecedentType) {
                            // The first antecedent of a loop junction is always the non-looping control
                            // flow path that leads to the top.
                            flowType = firstAntecedentType = getTypeAtFlowNode(antecedent);
                        }
                        else {
                            // All but the first antecedent are the looping control flow paths that lead
                            // back to the loop junction. We track these on the flow loop stack.
                            flowLoopNodes[flowLoopCount] = flow;
                            flowLoopKeys[flowLoopCount] = key;
                            flowLoopTypes[flowLoopCount] = antecedentTypes;
                            flowLoopCount++;
                            var saveFlowTypeCache = flowTypeCache;
                            flowTypeCache = undefined;
                            flowType = getTypeAtFlowNode(antecedent);
                            flowTypeCache = saveFlowTypeCache;
                            flowLoopCount--;
                            // If we see a value appear in the cache it is a sign that control flow analysis
                            // was restarted and completed by checkExpressionCached. We can simply pick up
                            // the resulting type and bail out.
                            var cached_1 = cache.get(key);
                            if (cached_1) {
                                return cached_1;
                            }
                        }
                        var type = getTypeFromFlowType(flowType);
                        ts.pushIfUnique(antecedentTypes, type);
                        // If an antecedent type is not a subset of the declared type, we need to perform
                        // subtype reduction. This happens when a "foreign" type is injected into the control
                        // flow using the instanceof operator or a user defined type predicate.
                        if (!isTypeSubsetOf(type, declaredType)) {
                            subtypeReduction = true;
                        }
                        // If the type at a particular antecedent path is the declared type there is no
                        // reason to process more antecedents since the only possible outcome is subtypes
                        // that will be removed in the final union type anyway.
                        if (type === declaredType) {
                            break;
                        }
                    }
                    // The result is incomplete if the first antecedent (the non-looping control flow path)
                    // is incomplete.
                    var result = getUnionOrEvolvingArrayType(antecedentTypes, subtypeReduction ? 2 /* Subtype */ : 1 /* Literal */);
                    if (isIncomplete(firstAntecedentType)) {
                        return createFlowType(result, /*incomplete*/ true);
                    }
                    cache.set(key, result);
                    return result;
                }
                // At flow control branch or loop junctions, if the type along every antecedent code path
                // is an evolving array type, we construct a combined evolving array type. Otherwise we
                // finalize all evolving array types.
                function getUnionOrEvolvingArrayType(types, subtypeReduction) {
                    if (isEvolvingArrayTypeList(types)) {
                        return getEvolvingArrayType(getUnionType(ts.map(types, getElementTypeOfEvolvingArrayType)));
                    }
                    var result = getUnionType(ts.sameMap(types, finalizeEvolvingArrayType), subtypeReduction);
                    if (result !== declaredType && result.flags & declaredType.flags & 1048576 /* Union */ && ts.arraysEqual(result.types, declaredType.types)) {
                        return declaredType;
                    }
                    return result;
                }
                function getDiscriminantPropertyAccess(expr, computedType) {
                    var access, name;
                    var type = declaredType.flags & 1048576 /* Union */ ? declaredType : computedType;
                    return type.flags & 1048576 /* Union */ && (access = getPropertyAccess(expr)) && (name = getAccessedPropertyName(access)) &&
                        isMatchingReference(reference, ts.isAccessExpression(access) ? access.expression : access.parent.parent.initializer) &&
                        isDiscriminantProperty(type, name) ?
                        access : undefined;
                }
                function narrowTypeByDiscriminant(type, access, narrowType) {
                    var propName = getAccessedPropertyName(access);
                    if (propName === undefined) {
                        return type;
                    }
                    var removeNullable = strictNullChecks && ts.isOptionalChain(access) && maybeTypeOfKind(type, 98304 /* Nullable */);
                    var propType = getTypeOfPropertyOfType(removeNullable ? getTypeWithFacts(type, 2097152 /* NEUndefinedOrNull */) : type, propName);
                    if (!propType) {
                        return type;
                    }
                    propType = removeNullable ? getOptionalType(propType) : propType;
                    var narrowedPropType = narrowType(propType);
                    return filterType(type, function (t) {
                        var discriminantType = getTypeOfPropertyOrIndexSignature(t, propName);
                        return !(discriminantType.flags & 131072 /* Never */) && isTypeComparableTo(discriminantType, narrowedPropType);
                    });
                }
                function narrowTypeByDiscriminantProperty(type, access, operator, value, assumeTrue) {
                    if ((operator === 36 /* EqualsEqualsEqualsToken */ || operator === 37 /* ExclamationEqualsEqualsToken */) && type.flags & 1048576 /* Union */) {
                        var keyPropertyName = getKeyPropertyName(type);
                        if (keyPropertyName && keyPropertyName === getAccessedPropertyName(access)) {
                            var candidate = getConstituentTypeForKeyType(type, getTypeOfExpression(value));
                            if (candidate) {
                                return operator === (assumeTrue ? 36 /* EqualsEqualsEqualsToken */ : 37 /* ExclamationEqualsEqualsToken */) ? candidate :
                                    isUnitType(getTypeOfPropertyOfType(candidate, keyPropertyName) || unknownType) ? removeType(type, candidate) :
                                        type;
                            }
                        }
                    }
                    return narrowTypeByDiscriminant(type, access, function (t) { return narrowTypeByEquality(t, operator, value, assumeTrue); });
                }
                function narrowTypeBySwitchOnDiscriminantProperty(type, access, switchStatement, clauseStart, clauseEnd) {
                    if (clauseStart < clauseEnd && type.flags & 1048576 /* Union */ && getKeyPropertyName(type) === getAccessedPropertyName(access)) {
                        var clauseTypes = getSwitchClauseTypes(switchStatement).slice(clauseStart, clauseEnd);
                        var candidate = getUnionType(ts.map(clauseTypes, function (t) { return getConstituentTypeForKeyType(type, t) || unknownType; }));
                        if (candidate !== unknownType) {
                            return candidate;
                        }
                    }
                    return narrowTypeByDiscriminant(type, access, function (t) { return narrowTypeBySwitchOnDiscriminant(t, switchStatement, clauseStart, clauseEnd); });
                }
                function narrowTypeByTruthiness(type, expr, assumeTrue) {
                    if (isMatchingReference(reference, expr)) {
                        return getTypeWithFacts(type, assumeTrue ? 4194304 /* Truthy */ : 8388608 /* Falsy */);
                    }
                    if (strictNullChecks && assumeTrue && optionalChainContainsReference(expr, reference)) {
                        type = getTypeWithFacts(type, 2097152 /* NEUndefinedOrNull */);
                    }
                    var access = getDiscriminantPropertyAccess(expr, type);
                    if (access) {
                        return narrowTypeByDiscriminant(type, access, function (t) { return getTypeWithFacts(t, assumeTrue ? 4194304 /* Truthy */ : 8388608 /* Falsy */); });
                    }
                    return type;
                }
                function isTypePresencePossible(type, propName, assumeTrue) {
                    var prop = getPropertyOfType(type, propName);
                    if (prop) {
                        return prop.flags & 16777216 /* Optional */ ? true : assumeTrue;
                    }
                    return getApplicableIndexInfoForName(type, propName) ? true : !assumeTrue;
                }
                function narrowByInKeyword(type, name, assumeTrue) {
                    if (type.flags & 1048576 /* Union */
                        || type.flags & 524288 /* Object */ && declaredType !== type
                        || isThisTypeParameter(type)
                        || type.flags & 2097152 /* Intersection */ && ts.every(type.types, function (t) { return t.symbol !== globalThisSymbol; })) {
                        return filterType(type, function (t) { return isTypePresencePossible(t, name, assumeTrue); });
                    }
                    return type;
                }
                function narrowTypeByBinaryExpression(type, expr, assumeTrue) {
                    switch (expr.operatorToken.kind) {
                        case 63 /* EqualsToken */:
                        case 75 /* BarBarEqualsToken */:
                        case 76 /* AmpersandAmpersandEqualsToken */:
                        case 77 /* QuestionQuestionEqualsToken */:
                            return narrowTypeByTruthiness(narrowType(type, expr.right, assumeTrue), expr.left, assumeTrue);
                        case 34 /* EqualsEqualsToken */:
                        case 35 /* ExclamationEqualsToken */:
                        case 36 /* EqualsEqualsEqualsToken */:
                        case 37 /* ExclamationEqualsEqualsToken */:
                            var operator = expr.operatorToken.kind;
                            var left = getReferenceCandidate(expr.left);
                            var right = getReferenceCandidate(expr.right);
                            if (left.kind === 214 /* TypeOfExpression */ && ts.isStringLiteralLike(right)) {
                                return narrowTypeByTypeof(type, left, operator, right, assumeTrue);
                            }
                            if (right.kind === 214 /* TypeOfExpression */ && ts.isStringLiteralLike(left)) {
                                return narrowTypeByTypeof(type, right, operator, left, assumeTrue);
                            }
                            if (isMatchingReference(reference, left)) {
                                return narrowTypeByEquality(type, operator, right, assumeTrue);
                            }
                            if (isMatchingReference(reference, right)) {
                                return narrowTypeByEquality(type, operator, left, assumeTrue);
                            }
                            if (strictNullChecks) {
                                if (optionalChainContainsReference(left, reference)) {
                                    type = narrowTypeByOptionalChainContainment(type, operator, right, assumeTrue);
                                }
                                else if (optionalChainContainsReference(right, reference)) {
                                    type = narrowTypeByOptionalChainContainment(type, operator, left, assumeTrue);
                                }
                            }
                            var leftAccess = getDiscriminantPropertyAccess(left, type);
                            if (leftAccess) {
                                return narrowTypeByDiscriminantProperty(type, leftAccess, operator, right, assumeTrue);
                            }
                            var rightAccess = getDiscriminantPropertyAccess(right, type);
                            if (rightAccess) {
                                return narrowTypeByDiscriminantProperty(type, rightAccess, operator, left, assumeTrue);
                            }
                            if (isMatchingConstructorReference(left)) {
                                return narrowTypeByConstructor(type, operator, right, assumeTrue);
                            }
                            if (isMatchingConstructorReference(right)) {
                                return narrowTypeByConstructor(type, operator, left, assumeTrue);
                            }
                            break;
                        case 102 /* InstanceOfKeyword */:
                            return narrowTypeByInstanceof(type, expr, assumeTrue);
                        case 101 /* InKeyword */:
                            var target = getReferenceCandidate(expr.right);
                            var leftType = getTypeOfNode(expr.left);
                            if (leftType.flags & 128 /* StringLiteral */) {
                                var name = ts.escapeLeadingUnderscores(leftType.value);
                                if (containsMissingType(type) && ts.isAccessExpression(reference) && isMatchingReference(reference.expression, target) &&
                                    getAccessedPropertyName(reference) === name) {
                                    return getTypeWithFacts(type, assumeTrue ? 524288 /* NEUndefined */ : 65536 /* EQUndefined */);
                                }
                                if (isMatchingReference(reference, target)) {
                                    return narrowByInKeyword(type, name, assumeTrue);
                                }
                            }
                            break;
                        case 27 /* CommaToken */:
                            return narrowType(type, expr.right, assumeTrue);
                        // Ordinarily we won't see && and || expressions in control flow analysis because the Binder breaks those
                        // expressions down to individual conditional control flows. However, we may encounter them when analyzing
                        // aliased conditional expressions.
                        case 55 /* AmpersandAmpersandToken */:
                            return assumeTrue ?
                                narrowType(narrowType(type, expr.left, /*assumeTrue*/ true), expr.right, /*assumeTrue*/ true) :
                                getUnionType([narrowType(type, expr.left, /*assumeTrue*/ false), narrowType(type, expr.right, /*assumeTrue*/ false)]);
                        case 56 /* BarBarToken */:
                            return assumeTrue ?
                                getUnionType([narrowType(type, expr.left, /*assumeTrue*/ true), narrowType(type, expr.right, /*assumeTrue*/ true)]) :
                                narrowType(narrowType(type, expr.left, /*assumeTrue*/ false), expr.right, /*assumeTrue*/ false);
                    }
                    return type;
                }
                function narrowTypeByOptionalChainContainment(type, operator, value, assumeTrue) {
                    // We are in a branch of obj?.foo === value (or any one of the other equality operators). We narrow obj as follows:
                    // When operator is === and type of value excludes undefined, null and undefined is removed from type of obj in true branch.
                    // When operator is !== and type of value excludes undefined, null and undefined is removed from type of obj in false branch.
                    // When operator is == and type of value excludes null and undefined, null and undefined is removed from type of obj in true branch.
                    // When operator is != and type of value excludes null and undefined, null and undefined is removed from type of obj in false branch.
                    // When operator is === and type of value is undefined, null and undefined is removed from type of obj in false branch.
                    // When operator is !== and type of value is undefined, null and undefined is removed from type of obj in true branch.
                    // When operator is == and type of value is null or undefined, null and undefined is removed from type of obj in false branch.
                    // When operator is != and type of value is null or undefined, null and undefined is removed from type of obj in true branch.
                    var equalsOperator = operator === 34 /* EqualsEqualsToken */ || operator === 36 /* EqualsEqualsEqualsToken */;
                    var nullableFlags = operator === 34 /* EqualsEqualsToken */ || operator === 35 /* ExclamationEqualsToken */ ? 98304 /* Nullable */ : 32768 /* Undefined */;
                    var valueType = getTypeOfExpression(value);
                    // Note that we include any and unknown in the exclusion test because their domain includes null and undefined.
                    var removeNullable = equalsOperator !== assumeTrue && everyType(valueType, function (t) { return !!(t.flags & nullableFlags); }) ||
                        equalsOperator === assumeTrue && everyType(valueType, function (t) { return !(t.flags & (3 /* AnyOrUnknown */ | nullableFlags)); });
                    return removeNullable ? getTypeWithFacts(type, 2097152 /* NEUndefinedOrNull */) : type;
                }
                function narrowTypeByEquality(type, operator, value, assumeTrue) {
                    if (type.flags & 1 /* Any */) {
                        return type;
                    }
                    if (operator === 35 /* ExclamationEqualsToken */ || operator === 37 /* ExclamationEqualsEqualsToken */) {
                        assumeTrue = !assumeTrue;
                    }
                    var valueType = getTypeOfExpression(value);
                    if ((type.flags & 2 /* Unknown */) && assumeTrue && (operator === 36 /* EqualsEqualsEqualsToken */ || operator === 37 /* ExclamationEqualsEqualsToken */)) {
                        if (valueType.flags & (131068 /* Primitive */ | 67108864 /* NonPrimitive */)) {
                            return valueType;
                        }
                        if (valueType.flags & 524288 /* Object */) {
                            return nonPrimitiveType;
                        }
                        return type;
                    }
                    if (valueType.flags & 98304 /* Nullable */) {
                        if (!strictNullChecks) {
                            return type;
                        }
                        var doubleEquals = operator === 34 /* EqualsEqualsToken */ || operator === 35 /* ExclamationEqualsToken */;
                        var facts = doubleEquals ?
                            assumeTrue ? 262144 /* EQUndefinedOrNull */ : 2097152 /* NEUndefinedOrNull */ :
                            valueType.flags & 65536 /* Null */ ?
                                assumeTrue ? 131072 /* EQNull */ : 1048576 /* NENull */ :
                                assumeTrue ? 65536 /* EQUndefined */ : 524288 /* NEUndefined */;
                        return getTypeWithFacts(type, facts);
                    }
                    if (assumeTrue) {
                        var filterFn = operator === 34 /* EqualsEqualsToken */ ?
                            function (t) { return areTypesComparable(t, valueType) || isCoercibleUnderDoubleEquals(t, valueType); } :
                            function (t) { return areTypesComparable(t, valueType); };
                        return replacePrimitivesWithLiterals(filterType(type, filterFn), valueType);
                    }
                    if (isUnitType(valueType)) {
                        return filterType(type, function (t) { return !(isUnitLikeType(t) && areTypesComparable(t, valueType)); });
                    }
                    return type;
                }
                function narrowTypeByTypeof(type, typeOfExpr, operator, literal, assumeTrue) {
                    // We have '==', '!=', '===', or !==' operator with 'typeof xxx' and string literal operands
                    if (operator === 35 /* ExclamationEqualsToken */ || operator === 37 /* ExclamationEqualsEqualsToken */) {
                        assumeTrue = !assumeTrue;
                    }
                    var target = getReferenceCandidate(typeOfExpr.expression);
                    if (!isMatchingReference(reference, target)) {
                        if (strictNullChecks && optionalChainContainsReference(target, reference) && assumeTrue === (literal.text !== "undefined")) {
                            return getTypeWithFacts(type, 2097152 /* NEUndefinedOrNull */);
                        }
                        return type;
                    }
                    if (type.flags & 1 /* Any */ && literal.text === "function") {
                        return type;
                    }
                    if (assumeTrue && type.flags & 2 /* Unknown */ && literal.text === "object") {
                        // The pattern x && typeof x === 'object', where x is of type unknown, narrows x to type object. We don't
                        // need to check for the reverse typeof x === 'object' && x since that already narrows correctly.
                        if (typeOfExpr.parent.parent.kind === 219 /* BinaryExpression */) {
                            var expr = typeOfExpr.parent.parent;
                            if (expr.operatorToken.kind === 55 /* AmpersandAmpersandToken */ && expr.right === typeOfExpr.parent && containsTruthyCheck(reference, expr.left)) {
                                return nonPrimitiveType;
                            }
                        }
                        return getUnionType([nonPrimitiveType, nullType]);
                    }
                    var facts = assumeTrue ?
                        typeofEQFacts.get(literal.text) || 128 /* TypeofEQHostObject */ :
                        typeofNEFacts.get(literal.text) || 32768 /* TypeofNEHostObject */;
                    var impliedType = getImpliedTypeFromTypeofGuard(type, literal.text);
                    return getTypeWithFacts(assumeTrue && impliedType ? mapType(type, narrowUnionMemberByTypeof(impliedType)) : type, facts);
                }
                function narrowTypeBySwitchOptionalChainContainment(type, switchStatement, clauseStart, clauseEnd, clauseCheck) {
                    var everyClauseChecks = clauseStart !== clauseEnd && ts.every(getSwitchClauseTypes(switchStatement).slice(clauseStart, clauseEnd), clauseCheck);
                    return everyClauseChecks ? getTypeWithFacts(type, 2097152 /* NEUndefinedOrNull */) : type;
                }
                function narrowTypeBySwitchOnDiscriminant(type, switchStatement, clauseStart, clauseEnd) {
                    // We only narrow if all case expressions specify
                    // values with unit types, except for the case where
                    // `type` is unknown. In this instance we map object
                    // types to the nonPrimitive type and narrow with that.
                    var switchTypes = getSwitchClauseTypes(switchStatement);
                    if (!switchTypes.length) {
                        return type;
                    }
                    var clauseTypes = switchTypes.slice(clauseStart, clauseEnd);
                    var hasDefaultClause = clauseStart === clauseEnd || ts.contains(clauseTypes, neverType);
                    if ((type.flags & 2 /* Unknown */) && !hasDefaultClause) {
                        var groundClauseTypes = void 0;
                        for (var i = 0; i < clauseTypes.length; i += 1) {
                            var t = clauseTypes[i];
                            if (t.flags & (131068 /* Primitive */ | 67108864 /* NonPrimitive */)) {
                                if (groundClauseTypes !== undefined) {
                                    groundClauseTypes.push(t);
                                }
                            }
                            else if (t.flags & 524288 /* Object */) {
                                if (groundClauseTypes === undefined) {
                                    groundClauseTypes = clauseTypes.slice(0, i);
                                }
                                groundClauseTypes.push(nonPrimitiveType);
                            }
                            else {
                                return type;
                            }
                        }
                        return getUnionType(groundClauseTypes === undefined ? clauseTypes : groundClauseTypes);
                    }
                    var discriminantType = getUnionType(clauseTypes);
                    var caseType = discriminantType.flags & 131072 /* Never */ ? neverType :
                        replacePrimitivesWithLiterals(filterType(type, function (t) { return areTypesComparable(discriminantType, t); }), discriminantType);
                    if (!hasDefaultClause) {
                        return caseType;
                    }
                    var defaultType = filterType(type, function (t) { return !(isUnitLikeType(t) && ts.contains(switchTypes, getRegularTypeOfLiteralType(extractUnitType(t)))); });
                    return caseType.flags & 131072 /* Never */ ? defaultType : getUnionType([caseType, defaultType]);
                }
                function getImpliedTypeFromTypeofGuard(type, text) {
                    switch (text) {
                        case "function":
                            return type.flags & 1 /* Any */ ? type : globalFunctionType;
                        case "object":
                            return type.flags & 2 /* Unknown */ ? getUnionType([nonPrimitiveType, nullType]) : type;
                        default:
                            return typeofTypesByName.get(text);
                    }
                }
                // When narrowing a union type by a `typeof` guard using type-facts alone, constituent types that are
                // super-types of the implied guard will be retained in the final type: this is because type-facts only
                // filter. Instead, we would like to replace those union constituents with the more precise type implied by
                // the guard. For example: narrowing `{} | undefined` by `"boolean"` should produce the type `boolean`, not
                // the filtered type `{}`. For this reason we narrow constituents of the union individually, in addition to
                // filtering by type-facts.
                function narrowUnionMemberByTypeof(candidate) {
                    return function (type) {
                        if (isTypeSubtypeOf(type, candidate)) {
                            return type;
                        }
                        if (isTypeSubtypeOf(candidate, type)) {
                            return candidate;
                        }
                        if (type.flags & 465829888 /* Instantiable */) {
                            var constraint = getBaseConstraintOfType(type) || anyType;
                            if (isTypeSubtypeOf(candidate, constraint)) {
                                return getIntersectionType([type, candidate]);
                            }
                        }
                        return type;
                    };
                }
                function narrowBySwitchOnTypeOf(type, switchStatement, clauseStart, clauseEnd) {
                    var switchWitnesses = getSwitchClauseTypeOfWitnesses(switchStatement, /*retainDefault*/ true);
                    if (!switchWitnesses.length) {
                        return type;
                    }
                    //  Equal start and end denotes implicit fallthrough; undefined marks explicit default clause
                    var defaultCaseLocation = ts.findIndex(switchWitnesses, function (elem) { return elem === undefined; });
                    var hasDefaultClause = clauseStart === clauseEnd || (defaultCaseLocation >= clauseStart && defaultCaseLocation < clauseEnd);
                    var clauseWitnesses;
                    var switchFacts;
                    if (defaultCaseLocation > -1) {
                        // We no longer need the undefined denoting an explicit default case. Remove the undefined and
                        // fix-up clauseStart and clauseEnd.  This means that we don't have to worry about undefined in the
                        // witness array.
                        var witnesses = switchWitnesses.filter(function (witness) { return witness !== undefined; });
                        // The adjusted clause start and end after removing the `default` statement.
                        var fixedClauseStart = defaultCaseLocation < clauseStart ? clauseStart - 1 : clauseStart;
                        var fixedClauseEnd = defaultCaseLocation < clauseEnd ? clauseEnd - 1 : clauseEnd;
                        clauseWitnesses = witnesses.slice(fixedClauseStart, fixedClauseEnd);
                        switchFacts = getFactsFromTypeofSwitch(fixedClauseStart, fixedClauseEnd, witnesses, hasDefaultClause);
                    }
                    else {
                        clauseWitnesses = switchWitnesses.slice(clauseStart, clauseEnd);
                        switchFacts = getFactsFromTypeofSwitch(clauseStart, clauseEnd, switchWitnesses, hasDefaultClause);
                    }
                    if (hasDefaultClause) {
                        return filterType(type, function (t) { return (getTypeFacts(t) & switchFacts) === switchFacts; });
                    }
                    /*
                      The implied type is the raw type suggested by a
                      value being caught in this clause.
    
                      When the clause contains a default case we ignore
                      the implied type and try to narrow using any facts
                      we can learn: see `switchFacts`.
    
                      Example:
                      switch (typeof x) {
                          case 'number':
                          case 'string': break;
                          default: break;
                          case 'number':
                          case 'boolean': break
                      }
    
                      In the first clause (case `number` and `string`) the
                      implied type is number | string.
    
                      In the default clause we de not compute an implied type.
    
                      In the third clause (case `number` and `boolean`)
                      the naive implied type is number | boolean, however
                      we use the type facts to narrow the implied type to
                      boolean. We know that number cannot be selected
                      because it is caught in the first clause.
                    */
                    var impliedType = getTypeWithFacts(getUnionType(clauseWitnesses.map(function (text) { return getImpliedTypeFromTypeofGuard(type, text) || type; })), switchFacts);
                    return getTypeWithFacts(mapType(type, narrowUnionMemberByTypeof(impliedType)), switchFacts);
                }
                function isMatchingConstructorReference(expr) {
                    return (ts.isPropertyAccessExpression(expr) && ts.idText(expr.name) === "constructor" ||
                        ts.isElementAccessExpression(expr) && ts.isStringLiteralLike(expr.argumentExpression) && expr.argumentExpression.text === "constructor") &&
                        isMatchingReference(reference, expr.expression);
                }
                function narrowTypeByConstructor(type, operator, identifier, assumeTrue) {
                    // Do not narrow when checking inequality.
                    if (assumeTrue ? (operator !== 34 /* EqualsEqualsToken */ && operator !== 36 /* EqualsEqualsEqualsToken */) : (operator !== 35 /* ExclamationEqualsToken */ && operator !== 37 /* ExclamationEqualsEqualsToken */)) {
                        return type;
                    }
                    // Get the type of the constructor identifier expression, if it is not a function then do not narrow.
                    var identifierType = getTypeOfExpression(identifier);
                    if (!isFunctionType(identifierType) && !isConstructorType(identifierType)) {
                        return type;
                    }
                    // Get the prototype property of the type identifier so we can find out its type.
                    var prototypeProperty = getPropertyOfType(identifierType, "prototype");
                    if (!prototypeProperty) {
                        return type;
                    }
                    // Get the type of the prototype, if it is undefined, or the global `Object` or `Function` types then do not narrow.
                    var prototypeType = getTypeOfSymbol(prototypeProperty);
                    var candidate = !isTypeAny(prototypeType) ? prototypeType : undefined;
                    if (!candidate || candidate === globalObjectType || candidate === globalFunctionType) {
                        return type;
                    }
                    // If the type that is being narrowed is `any` then just return the `candidate` type since every type is a subtype of `any`.
                    if (isTypeAny(type)) {
                        return candidate;
                    }
                    // Filter out types that are not considered to be "constructed by" the `candidate` type.
                    return filterType(type, function (t) { return isConstructedBy(t, candidate); });
                    function isConstructedBy(source, target) {
                        // If either the source or target type are a class type then we need to check that they are the same exact type.
                        // This is because you may have a class `A` that defines some set of properties, and another class `B`
                        // that defines the same set of properties as class `A`, in that case they are structurally the same
                        // type, but when you do something like `instanceOfA.constructor === B` it will return false.
                        if (source.flags & 524288 /* Object */ && ts.getObjectFlags(source) & 1 /* Class */ ||
                            target.flags & 524288 /* Object */ && ts.getObjectFlags(target) & 1 /* Class */) {
                            return source.symbol === target.symbol;
                        }
                        // For all other types just check that the `source` type is a subtype of the `target` type.
                        return isTypeSubtypeOf(source, target);
                    }
                }
                function narrowTypeByInstanceof(type, expr, assumeTrue) {
                    var left = getReferenceCandidate(expr.left);
                    if (!isMatchingReference(reference, left)) {
                        if (assumeTrue && strictNullChecks && optionalChainContainsReference(left, reference)) {
                            return getTypeWithFacts(type, 2097152 /* NEUndefinedOrNull */);
                        }
                        return type;
                    }
                    // Check that right operand is a function type with a prototype property
                    var rightType = getTypeOfExpression(expr.right);
                    if (!isTypeDerivedFrom(rightType, globalFunctionType)) {
                        return type;
                    }
                    var targetType;
                    var prototypeProperty = getPropertyOfType(rightType, "prototype");
                    if (prototypeProperty) {
                        // Target type is type of the prototype property
                        var prototypePropertyType = getTypeOfSymbol(prototypeProperty);
                        if (!isTypeAny(prototypePropertyType)) {
                            targetType = prototypePropertyType;
                        }
                    }
                    // Don't narrow from 'any' if the target type is exactly 'Object' or 'Function'
                    if (isTypeAny(type) && (targetType === globalObjectType || targetType === globalFunctionType)) {
                        return type;
                    }
                    if (!targetType) {
                        var constructSignatures = getSignaturesOfType(rightType, 1 /* Construct */);
                        targetType = constructSignatures.length ?
                            getUnionType(ts.map(constructSignatures, function (signature) { return getReturnTypeOfSignature(getErasedSignature(signature)); })) :
                            emptyObjectType;
                    }
                    // We can't narrow a union based off instanceof without negated types see #31576 for more info
                    if (!assumeTrue && rightType.flags & 1048576 /* Union */) {
                        var nonConstructorTypeInUnion = ts.find(rightType.types, function (t) { return !isConstructorType(t); });
                        if (!nonConstructorTypeInUnion)
                            return type;
                    }
                    return getNarrowedType(type, targetType, assumeTrue, isTypeDerivedFrom);
                }
                function getNarrowedType(type, candidate, assumeTrue, isRelated) {
                    if (!assumeTrue) {
                        return filterType(type, function (t) {
                            if (!isRelated(t, candidate)) {
                                return true;
                            }
                            var constraint = getBaseConstraintOfType(t);
                            if (constraint && constraint !== t) {
                                return !isRelated(constraint, candidate);
                            }
                            return false;
                        });
                    }
                    // If the current type is a union type, remove all constituents that couldn't be instances of
                    // the candidate type. If one or more constituents remain, return a union of those.
                    if (type.flags & 1048576 /* Union */) {
                        var assignableType = filterType(type, function (t) { return isRelated(t, candidate); });
                        if (!(assignableType.flags & 131072 /* Never */)) {
                            return assignableType;
                        }
                    }
                    // If the candidate type is a subtype of the target type, narrow to the candidate type.
                    // Otherwise, if the target type is assignable to the candidate type, keep the target type.
                    // Otherwise, if the candidate type is assignable to the target type, narrow to the candidate
                    // type. Otherwise, the types are completely unrelated, so narrow to an intersection of the
                    // two types.
                    return isTypeSubtypeOf(candidate, type) ? candidate :
                        isTypeAssignableTo(type, candidate) ? type :
                            isTypeAssignableTo(candidate, type) ? candidate :
                                getIntersectionType([type, candidate]);
                }
                function narrowTypeByCallExpression(type, callExpression, assumeTrue) {
                    if (hasMatchingArgument(callExpression, reference)) {
                        var signature = assumeTrue || !ts.isCallChain(callExpression) ? getEffectsSignature(callExpression) : undefined;
                        var predicate = signature && getTypePredicateOfSignature(signature);
                        if (predicate && (predicate.kind === 0 /* This */ || predicate.kind === 1 /* Identifier */)) {
                            return narrowTypeByTypePredicate(type, predicate, callExpression, assumeTrue);
                        }
                    }
                    if (containsMissingType(type) && ts.isAccessExpression(reference) && ts.isPropertyAccessExpression(callExpression.expression)) {
                        var callAccess = callExpression.expression;
                        if (isMatchingReference(reference.expression, getReferenceCandidate(callAccess.expression)) &&
                            ts.isIdentifier(callAccess.name) && callAccess.name.escapedText === "hasOwnProperty" && callExpression.arguments.length === 1) {
                            var argument = callExpression.arguments[0];
                            if (ts.isStringLiteralLike(argument) && getAccessedPropertyName(reference) === ts.escapeLeadingUnderscores(argument.text)) {
                                return getTypeWithFacts(type, assumeTrue ? 524288 /* NEUndefined */ : 65536 /* EQUndefined */);
                            }
                        }
                    }
                    return type;
                }
                function narrowTypeByTypePredicate(type, predicate, callExpression, assumeTrue) {
                    // Don't narrow from 'any' if the predicate type is exactly 'Object' or 'Function'
                    if (predicate.type && !(isTypeAny(type) && (predicate.type === globalObjectType || predicate.type === globalFunctionType))) {
                        var predicateArgument = getTypePredicateArgument(predicate, callExpression);
                        if (predicateArgument) {
                            if (isMatchingReference(reference, predicateArgument)) {
                                return getNarrowedType(type, predicate.type, assumeTrue, isTypeSubtypeOf);
                            }
                            if (strictNullChecks && assumeTrue && optionalChainContainsReference(predicateArgument, reference) &&
                                !(getTypeFacts(predicate.type) & 65536 /* EQUndefined */)) {
                                type = getTypeWithFacts(type, 2097152 /* NEUndefinedOrNull */);
                            }
                            var access = getDiscriminantPropertyAccess(predicateArgument, type);
                            if (access) {
                                return narrowTypeByDiscriminant(type, access, function (t) { return getNarrowedType(t, predicate.type, assumeTrue, isTypeSubtypeOf); });
                            }
                        }
                    }
                    return type;
                }
                // Narrow the given type based on the given expression having the assumed boolean value. The returned type
                // will be a subtype or the same type as the argument.
                function narrowType(type, expr, assumeTrue) {
                    // for `a?.b`, we emulate a synthetic `a !== null && a !== undefined` condition for `a`
                    if (ts.isExpressionOfOptionalChainRoot(expr) ||
                        ts.isBinaryExpression(expr.parent) && expr.parent.operatorToken.kind === 60 /* QuestionQuestionToken */ && expr.parent.left === expr) {
                        return narrowTypeByOptionality(type, expr, assumeTrue);
                    }
                    switch (expr.kind) {
                        case 79 /* Identifier */:
                            // When narrowing a reference to a const variable, non-assigned parameter, or readonly property, we inline
                            // up to five levels of aliased conditional expressions that are themselves declared as const variables.
                            if (!isMatchingReference(reference, expr) && inlineLevel < 5) {
                                var symbol = getResolvedSymbol(expr);
                                if (isConstVariable(symbol)) {
                                    var declaration = symbol.valueDeclaration;
                                    if (declaration && ts.isVariableDeclaration(declaration) && !declaration.type && declaration.initializer && isConstantReference(reference)) {
                                        inlineLevel++;
                                        var result = narrowType(type, declaration.initializer, assumeTrue);
                                        inlineLevel--;
                                        return result;
                                    }
                                }
                            }
                        // falls through
                        case 108 /* ThisKeyword */:
                        case 106 /* SuperKeyword */:
                        case 204 /* PropertyAccessExpression */:
                        case 205 /* ElementAccessExpression */:
                            return narrowTypeByTruthiness(type, expr, assumeTrue);
                        case 206 /* CallExpression */:
                            return narrowTypeByCallExpression(type, expr, assumeTrue);
                        case 210 /* ParenthesizedExpression */:
                        case 228 /* NonNullExpression */:
                            return narrowType(type, expr.expression, assumeTrue);
                        case 219 /* BinaryExpression */:
                            return narrowTypeByBinaryExpression(type, expr, assumeTrue);
                        case 217 /* PrefixUnaryExpression */:
                            if (expr.operator === 53 /* ExclamationToken */) {
                                return narrowType(type, expr.operand, !assumeTrue);
                            }
                            break;
                    }
                    return type;
                }
                function narrowTypeByOptionality(type, expr, assumePresent) {
                    if (isMatchingReference(reference, expr)) {
                        return getTypeWithFacts(type, assumePresent ? 2097152 /* NEUndefinedOrNull */ : 262144 /* EQUndefinedOrNull */);
                    }
                    var access = getDiscriminantPropertyAccess(expr, type);
                    if (access) {
                        return narrowTypeByDiscriminant(type, access, function (t) { return getTypeWithFacts(t, assumePresent ? 2097152 /* NEUndefinedOrNull */ : 262144 /* EQUndefinedOrNull */); });
                    }
                    return type;
                }
            }
            function getTypeOfSymbolAtLocation(symbol, location) {
                symbol = symbol.exportSymbol || symbol;
                // If we have an identifier or a property access at the given location, if the location is
                // an dotted name expression, and if the location is not an assignment target, obtain the type
                // of the expression (which will reflect control flow analysis). If the expression indeed
                // resolved to the given symbol, return the narrowed type.
                if (location.kind === 79 /* Identifier */ || location.kind === 80 /* PrivateIdentifier */) {
                    if (ts.isRightSideOfQualifiedNameOrPropertyAccess(location)) {
                        location = location.parent;
                    }
                    if (ts.isExpressionNode(location) && (!ts.isAssignmentTarget(location) || ts.isWriteAccess(location))) {
                        var type = getTypeOfExpression(location);
                        if (getExportSymbolOfValueSymbolIfExported(getNodeLinks(location).resolvedSymbol) === symbol) {
                            return type;
                        }
                    }
                }
                if (ts.isDeclarationName(location) && ts.isSetAccessor(location.parent) && getAnnotatedAccessorTypeNode(location.parent)) {
                    return resolveTypeOfAccessors(location.parent.symbol, /*writing*/ true);
                }
                // The location isn't a reference to the given symbol, meaning we're being asked
                // a hypothetical question of what type the symbol would have if there was a reference
                // to it at the given location. Since we have no control flow information for the
                // hypothetical reference (control flow information is created and attached by the
                // binder), we simply return the declared type of the symbol.
                return getNonMissingTypeOfSymbol(symbol);
            }
            function getControlFlowContainer(node) {
                return ts.findAncestor(node.parent, function (node) {
                    return ts.isFunctionLike(node) && !ts.getImmediatelyInvokedFunctionExpression(node) ||
                        node.kind === 260 /* ModuleBlock */ ||
                        node.kind === 300 /* SourceFile */ ||
                        node.kind === 165 /* PropertyDeclaration */;
                });
            }
            // Check if a parameter is assigned anywhere within its declaring function.
            function isParameterAssigned(symbol) {
                if (!symbol.valueDeclaration) {
                    return false;
                }
                var func = ts.getRootDeclaration(symbol.valueDeclaration).parent;
                var links = getNodeLinks(func);
                if (!(links.flags & 8388608 /* AssignmentsMarked */)) {
                    links.flags |= 8388608 /* AssignmentsMarked */;
                    if (!hasParentWithAssignmentsMarked(func)) {
                        markParameterAssignments(func);
                    }
                }
                return symbol.isAssigned || false;
            }
            function hasParentWithAssignmentsMarked(node) {
                return !!ts.findAncestor(node.parent, function (node) { return ts.isFunctionLike(node) && !!(getNodeLinks(node).flags & 8388608 /* AssignmentsMarked */); });
            }
            function markParameterAssignments(node) {
                if (node.kind === 79 /* Identifier */) {
                    if (ts.isAssignmentTarget(node)) {
                        var symbol = getResolvedSymbol(node);
                        if (symbol.valueDeclaration && ts.getRootDeclaration(symbol.valueDeclaration).kind === 162 /* Parameter */) {
                            symbol.isAssigned = true;
                        }
                    }
                }
                else {
                    ts.forEachChild(node, markParameterAssignments);
                }
            }
            function isConstVariable(symbol) {
                return symbol.flags & 3 /* Variable */ && (getDeclarationNodeFlagsFromSymbol(symbol) & 2 /* Const */) !== 0;
            }
            /** remove undefined from the annotated type of a parameter when there is an initializer (that doesn't include undefined) */
            function removeOptionalityFromDeclaredType(declaredType, declaration) {
                if (pushTypeResolution(declaration.symbol, 2 /* DeclaredType */)) {
                    var annotationIncludesUndefined = strictNullChecks &&
                        declaration.kind === 162 /* Parameter */ &&
                        declaration.initializer &&
                        getFalsyFlags(declaredType) & 32768 /* Undefined */ &&
                        !(getFalsyFlags(checkExpression(declaration.initializer)) & 32768 /* Undefined */);
                    popTypeResolution();
                    return annotationIncludesUndefined ? getTypeWithFacts(declaredType, 524288 /* NEUndefined */) : declaredType;
                }
                else {
                    reportCircularityError(declaration.symbol);
                    return declaredType;
                }
            }
            function isConstraintPosition(type, node) {
                var parent = node.parent;
                // In an element access obj[x], we consider obj to be in a constraint position, except when obj is of
                // a generic type without a nullable constraint and x is a generic type. This is because when both obj
                // and x are of generic types T and K, we want the resulting type to be T[K].
                return parent.kind === 204 /* PropertyAccessExpression */ ||
                    parent.kind === 206 /* CallExpression */ && parent.expression === node ||
                    parent.kind === 205 /* ElementAccessExpression */ && parent.expression === node &&
                        !(isGenericTypeWithoutNullableConstraint(type) && isGenericIndexType(getTypeOfExpression(parent.argumentExpression)));
            }
            function isGenericTypeWithUnionConstraint(type) {
                return !!(type.flags & 465829888 /* Instantiable */ && getBaseConstraintOrType(type).flags & (98304 /* Nullable */ | 1048576 /* Union */));
            }
            function isGenericTypeWithoutNullableConstraint(type) {
                return !!(type.flags & 465829888 /* Instantiable */ && !maybeTypeOfKind(getBaseConstraintOrType(type), 98304 /* Nullable */));
            }
            function hasNonBindingPatternContextualTypeWithNoGenericTypes(node) {
                // Computing the contextual type for a child of a JSX element involves resolving the type of the
                // element's tag name, so we exclude that here to avoid circularities.
                var contextualType = (ts.isIdentifier(node) || ts.isPropertyAccessExpression(node) || ts.isElementAccessExpression(node)) &&
                    !((ts.isJsxOpeningElement(node.parent) || ts.isJsxSelfClosingElement(node.parent)) && node.parent.tagName === node) &&
                    getContextualType(node, 8 /* SkipBindingPatterns */);
                return contextualType && !isGenericType(contextualType);
            }
            function getNarrowableTypeForReference(type, reference, checkMode) {
                // When the type of a reference is or contains an instantiable type with a union type constraint, and
                // when the reference is in a constraint position (where it is known we'll obtain the apparent type) or
                // has a contextual type containing no top-level instantiables (meaning constraints will determine
                // assignability), we substitute constraints for all instantiables in the type of the reference to give
                // control flow analysis an opportunity to narrow it further. For example, for a reference of a type
                // parameter type 'T extends string | undefined' with a contextual type 'string', we substitute
                // 'string | undefined' to give control flow analysis the opportunity to narrow to type 'string'.
                var substituteConstraints = !(checkMode && checkMode & 2 /* Inferential */) &&
                    someType(type, isGenericTypeWithUnionConstraint) &&
                    (isConstraintPosition(type, reference) || hasNonBindingPatternContextualTypeWithNoGenericTypes(reference));
                return substituteConstraints ? mapType(type, function (t) { return t.flags & 465829888 /* Instantiable */ ? getBaseConstraintOrType(t) : t; }) : type;
            }
            function isExportOrExportExpression(location) {
                return !!ts.findAncestor(location, function (e) { return e.parent && ts.isExportAssignment(e.parent) && e.parent.expression === e && ts.isEntityNameExpression(e); });
            }
            function markAliasReferenced(symbol, location) {
                if (isNonLocalAlias(symbol, /*excludes*/ 111551 /* Value */) && !isInTypeQuery(location) && !getTypeOnlyAliasDeclaration(symbol)) {
                    var target = resolveAlias(symbol);
                    if (target.flags & 111551 /* Value */) {
                        // An alias resolving to a const enum cannot be elided if (1) 'isolatedModules' is enabled
                        // (because the const enum value will not be inlined), or if (2) the alias is an export
                        // of a const enum declaration that will be preserved.
                        if (compilerOptions.isolatedModules ||
                            ts.shouldPreserveConstEnums(compilerOptions) && isExportOrExportExpression(location) ||
                            !isConstEnumOrConstEnumOnlyModule(target)) {
                            markAliasSymbolAsReferenced(symbol);
                        }
                        else {
                            markConstEnumAliasAsReferenced(symbol);
                        }
                    }
                }
            }
            function checkIdentifier(node, checkMode) {
                var symbol = getResolvedSymbol(node);
                if (symbol === unknownSymbol) {
                    return errorType;
                }
                // As noted in ECMAScript 6 language spec, arrow functions never have an arguments objects.
                // Although in down-level emit of arrow function, we emit it using function expression which means that
                // arguments objects will be bound to the inner object; emitting arrow function natively in ES6, arguments objects
                // will be bound to non-arrow function that contain this arrow function. This results in inconsistent behavior.
                // To avoid that we will give an error to users if they use arguments objects in arrow function so that they
                // can explicitly bound arguments objects
                if (symbol === argumentsSymbol) {
                    if (isInPropertyInitializerOrClassStaticBlock(node)) {
                        error(node, ts.Diagnostics.arguments_cannot_be_referenced_in_property_initializers);
                        return errorType;
                    }
                    var container = ts.getContainingFunction(node);
                    if (languageVersion < 2 /* ES2015 */) {
                        if (container.kind === 212 /* ArrowFunction */) {
                            error(node, ts.Diagnostics.The_arguments_object_cannot_be_referenced_in_an_arrow_function_in_ES3_and_ES5_Consider_using_a_standard_function_expression);
                        }
                        else if (ts.hasSyntacticModifier(container, 256 /* Async */)) {
                            error(node, ts.Diagnostics.The_arguments_object_cannot_be_referenced_in_an_async_function_or_method_in_ES3_and_ES5_Consider_using_a_standard_function_or_method);
                        }
                    }
                    getNodeLinks(container).flags |= 8192 /* CaptureArguments */;
                    return getTypeOfSymbol(symbol);
                }
                // We should only mark aliases as referenced if there isn't a local value declaration
                // for the symbol. Also, don't mark any property access expression LHS - checkPropertyAccessExpression will handle that
                if (!(node.parent && ts.isPropertyAccessExpression(node.parent) && node.parent.expression === node)) {
                    markAliasReferenced(symbol, node);
                }
                var localOrExportSymbol = getExportSymbolOfValueSymbolIfExported(symbol);
                var sourceSymbol = localOrExportSymbol.flags & 2097152 /* Alias */ ? resolveAlias(localOrExportSymbol) : localOrExportSymbol;
                if (sourceSymbol.declarations && getDeclarationNodeFlagsFromSymbol(sourceSymbol) & 134217728 /* Deprecated */ && isUncalledFunctionReference(node, sourceSymbol)) {
                    addDeprecatedSuggestion(node, sourceSymbol.declarations, node.escapedText);
                }
                var declaration = localOrExportSymbol.valueDeclaration;
                if (declaration && localOrExportSymbol.flags & 32 /* Class */) {
                    // Due to the emit for class decorators, any reference to the class from inside of the class body
                    // must instead be rewritten to point to a temporary variable to avoid issues with the double-bind
                    // behavior of class names in ES6.
                    if (declaration.kind === 255 /* ClassDeclaration */
                        && ts.nodeIsDecorated(declaration)) {
                        var container = ts.getContainingClass(node);
                        while (container !== undefined) {
                            if (container === declaration && container.name !== node) {
                                getNodeLinks(declaration).flags |= 16777216 /* ClassWithConstructorReference */;
                                getNodeLinks(node).flags |= 33554432 /* ConstructorReferenceInClass */;
                                break;
                            }
                            container = ts.getContainingClass(container);
                        }
                    }
                    else if (declaration.kind === 224 /* ClassExpression */) {
                        // When we emit a class expression with static members that contain a reference
                        // to the constructor in the initializer, we will need to substitute that
                        // binding with an alias as the class name is not in scope.
                        var container = ts.getThisContainer(node, /*includeArrowFunctions*/ false);
                        while (container.kind !== 300 /* SourceFile */) {
                            if (container.parent === declaration) {
                                if (ts.isPropertyDeclaration(container) && ts.isStatic(container) || ts.isClassStaticBlockDeclaration(container)) {
                                    getNodeLinks(declaration).flags |= 16777216 /* ClassWithConstructorReference */;
                                    getNodeLinks(node).flags |= 33554432 /* ConstructorReferenceInClass */;
                                }
                                break;
                            }
                            container = ts.getThisContainer(container, /*includeArrowFunctions*/ false);
                        }
                    }
                }
                checkNestedBlockScopedBinding(node, symbol);
                var type = getTypeOfSymbol(localOrExportSymbol);
                var assignmentKind = ts.getAssignmentTargetKind(node);
                if (assignmentKind) {
                    if (!(localOrExportSymbol.flags & 3 /* Variable */) &&
                        !(ts.isInJSFile(node) && localOrExportSymbol.flags & 512 /* ValueModule */)) {
                        var assignmentError = localOrExportSymbol.flags & 384 /* Enum */ ? ts.Diagnostics.Cannot_assign_to_0_because_it_is_an_enum
                            : localOrExportSymbol.flags & 32 /* Class */ ? ts.Diagnostics.Cannot_assign_to_0_because_it_is_a_class
                                : localOrExportSymbol.flags & 1536 /* Module */ ? ts.Diagnostics.Cannot_assign_to_0_because_it_is_a_namespace
                                    : localOrExportSymbol.flags & 16 /* Function */ ? ts.Diagnostics.Cannot_assign_to_0_because_it_is_a_function
                                        : localOrExportSymbol.flags & 2097152 /* Alias */ ? ts.Diagnostics.Cannot_assign_to_0_because_it_is_an_import
                                            : ts.Diagnostics.Cannot_assign_to_0_because_it_is_not_a_variable;
                        error(node, assignmentError, symbolToString(symbol));
                        return errorType;
                    }
                    if (isReadonlySymbol(localOrExportSymbol)) {
                        if (localOrExportSymbol.flags & 3 /* Variable */) {
                            error(node, ts.Diagnostics.Cannot_assign_to_0_because_it_is_a_constant, symbolToString(symbol));
                        }
                        else {
                            error(node, ts.Diagnostics.Cannot_assign_to_0_because_it_is_a_read_only_property, symbolToString(symbol));
                        }
                        return errorType;
                    }
                }
                var isAlias = localOrExportSymbol.flags & 2097152 /* Alias */;
                // We only narrow variables and parameters occurring in a non-assignment position. For all other
                // entities we simply return the declared type.
                if (localOrExportSymbol.flags & 3 /* Variable */) {
                    if (assignmentKind === 1 /* Definite */) {
                        return type;
                    }
                }
                else if (isAlias) {
                    declaration = getDeclarationOfAliasSymbol(symbol);
                }
                else {
                    return type;
                }
                if (!declaration) {
                    return type;
                }
                type = getNarrowableTypeForReference(type, node, checkMode);
                // The declaration container is the innermost function that encloses the declaration of the variable
                // or parameter. The flow container is the innermost function starting with which we analyze the control
                // flow graph to determine the control flow based type.
                var isParameter = ts.getRootDeclaration(declaration).kind === 162 /* Parameter */;
                var declarationContainer = getControlFlowContainer(declaration);
                var flowContainer = getControlFlowContainer(node);
                var isOuterVariable = flowContainer !== declarationContainer;
                var isSpreadDestructuringAssignmentTarget = node.parent && node.parent.parent && ts.isSpreadAssignment(node.parent) && isDestructuringAssignmentTarget(node.parent.parent);
                var isModuleExports = symbol.flags & 134217728 /* ModuleExports */;
                // When the control flow originates in a function expression or arrow function and we are referencing
                // a const variable or parameter from an outer function, we extend the origin of the control flow
                // analysis to include the immediately enclosing function.
                while (flowContainer !== declarationContainer && (flowContainer.kind === 211 /* FunctionExpression */ ||
                    flowContainer.kind === 212 /* ArrowFunction */ || ts.isObjectLiteralOrClassExpressionMethod(flowContainer)) &&
                    (isConstVariable(localOrExportSymbol) && type !== autoArrayType || isParameter && !isParameterAssigned(localOrExportSymbol))) {
                    flowContainer = getControlFlowContainer(flowContainer);
                }
                // We only look for uninitialized variables in strict null checking mode, and only when we can analyze
                // the entire control flow graph from the variable's declaration (i.e. when the flow container and
                // declaration container are the same).
                var assumeInitialized = isParameter || isAlias || isOuterVariable || isSpreadDestructuringAssignmentTarget || isModuleExports || ts.isBindingElement(declaration) ||
                    type !== autoType && type !== autoArrayType && (!strictNullChecks || (type.flags & (3 /* AnyOrUnknown */ | 16384 /* Void */)) !== 0 ||
                        isInTypeQuery(node) || node.parent.kind === 273 /* ExportSpecifier */) ||
                    node.parent.kind === 228 /* NonNullExpression */ ||
                    declaration.kind === 252 /* VariableDeclaration */ && declaration.exclamationToken ||
                    declaration.flags & 8388608 /* Ambient */;
                var initialType = assumeInitialized ? (isParameter ? removeOptionalityFromDeclaredType(type, declaration) : type) :
                    type === autoType || type === autoArrayType ? undefinedType :
                        getOptionalType(type);
                var flowType = getFlowTypeOfReference(node, type, initialType, flowContainer);
                // A variable is considered uninitialized when it is possible to analyze the entire control flow graph
                // from declaration to use, and when the variable's declared type doesn't include undefined but the
                // control flow based type does include undefined.
                if (!isEvolvingArrayOperationTarget(node) && (type === autoType || type === autoArrayType)) {
                    if (flowType === autoType || flowType === autoArrayType) {
                        if (noImplicitAny) {
                            error(ts.getNameOfDeclaration(declaration), ts.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined, symbolToString(symbol), typeToString(flowType));
                            error(node, ts.Diagnostics.Variable_0_implicitly_has_an_1_type, symbolToString(symbol), typeToString(flowType));
                        }
                        return convertAutoToAny(flowType);
                    }
                }
                else if (!assumeInitialized && !(getFalsyFlags(type) & 32768 /* Undefined */) && getFalsyFlags(flowType) & 32768 /* Undefined */) {
                    error(node, ts.Diagnostics.Variable_0_is_used_before_being_assigned, symbolToString(symbol));
                    // Return the declared type to reduce follow-on errors
                    return type;
                }
                return assignmentKind ? getBaseTypeOfLiteralType(flowType) : flowType;
            }
            function isInsideFunctionOrInstancePropertyInitializer(node, threshold) {
                return !!ts.findAncestor(node, function (n) { return n === threshold ? "quit" : ts.isFunctionLike(n) || (n.parent && ts.isPropertyDeclaration(n.parent) && !ts.hasStaticModifier(n.parent) && n.parent.initializer === n); });
            }
            function getPartOfForStatementContainingNode(node, container) {
                return ts.findAncestor(node, function (n) { return n === container ? "quit" : n === container.initializer || n === container.condition || n === container.incrementor || n === container.statement; });
            }
            function getEnclosingIterationStatement(node) {
                return ts.findAncestor(node, function (n) { return (!n || ts.nodeStartsNewLexicalEnvironment(n)) ? "quit" : ts.isIterationStatement(n, /*lookInLabeledStatements*/ false); });
            }
            function checkNestedBlockScopedBinding(node, symbol) {
                if (languageVersion >= 2 /* ES2015 */ ||
                    (symbol.flags & (2 /* BlockScopedVariable */ | 32 /* Class */)) === 0 ||
                    !symbol.valueDeclaration ||
                    ts.isSourceFile(symbol.valueDeclaration) ||
                    symbol.valueDeclaration.parent.kind === 290 /* CatchClause */) {
                    return;
                }
                // 1. walk from the use site up to the declaration and check
                // if there is anything function like between declaration and use-site (is binding/class is captured in function).
                // 2. walk from the declaration up to the boundary of lexical environment and check
                // if there is an iteration statement in between declaration and boundary (is binding/class declared inside iteration statement)
                var container = ts.getEnclosingBlockScopeContainer(symbol.valueDeclaration);
                var isCaptured = isInsideFunctionOrInstancePropertyInitializer(node, container);
                var enclosingIterationStatement = getEnclosingIterationStatement(container);
                if (enclosingIterationStatement) {
                    if (isCaptured) {
                        // mark iteration statement as containing block-scoped binding captured in some function
                        var capturesBlockScopeBindingInLoopBody = true;
                        if (ts.isForStatement(container)) {
                            var varDeclList = ts.getAncestor(symbol.valueDeclaration, 253 /* VariableDeclarationList */);
                            if (varDeclList && varDeclList.parent === container) {
                                var part = getPartOfForStatementContainingNode(node.parent, container);
                                if (part) {
                                    var links = getNodeLinks(part);
                                    links.flags |= 131072 /* ContainsCapturedBlockScopeBinding */;
                                    var capturedBindings = links.capturedBlockScopeBindings || (links.capturedBlockScopeBindings = []);
                                    ts.pushIfUnique(capturedBindings, symbol);
                                    if (part === container.initializer) {
                                        capturesBlockScopeBindingInLoopBody = false; // Initializer is outside of loop body
                                    }
                                }
                            }
                        }
                        if (capturesBlockScopeBindingInLoopBody) {
                            getNodeLinks(enclosingIterationStatement).flags |= 65536 /* LoopWithCapturedBlockScopedBinding */;
                        }
                    }
                    // mark variables that are declared in loop initializer and reassigned inside the body of ForStatement.
                    // if body of ForStatement will be converted to function then we'll need a extra machinery to propagate reassigned values back.
                    if (ts.isForStatement(container)) {
                        var varDeclList = ts.getAncestor(symbol.valueDeclaration, 253 /* VariableDeclarationList */);
                        if (varDeclList && varDeclList.parent === container && isAssignedInBodyOfForStatement(node, container)) {
                            getNodeLinks(symbol.valueDeclaration).flags |= 4194304 /* NeedsLoopOutParameter */;
                        }
                    }
                    // set 'declared inside loop' bit on the block-scoped binding
                    getNodeLinks(symbol.valueDeclaration).flags |= 524288 /* BlockScopedBindingInLoop */;
                }
                if (isCaptured) {
                    getNodeLinks(symbol.valueDeclaration).flags |= 262144 /* CapturedBlockScopedBinding */;
                }
            }
            function isBindingCapturedByNode(node, decl) {
                var links = getNodeLinks(node);
                return !!links && ts.contains(links.capturedBlockScopeBindings, getSymbolOfNode(decl));
            }
            function isAssignedInBodyOfForStatement(node, container) {
                // skip parenthesized nodes
                var current = node;
                while (current.parent.kind === 210 /* ParenthesizedExpression */) {
                    current = current.parent;
                }
                // check if node is used as LHS in some assignment expression
                var isAssigned = false;
                if (ts.isAssignmentTarget(current)) {
                    isAssigned = true;
                }
                else if ((current.parent.kind === 217 /* PrefixUnaryExpression */ || current.parent.kind === 218 /* PostfixUnaryExpression */)) {
                    var expr = current.parent;
                    isAssigned = expr.operator === 45 /* PlusPlusToken */ || expr.operator === 46 /* MinusMinusToken */;
                }
                if (!isAssigned) {
                    return false;
                }
                // at this point we know that node is the target of assignment
                // now check that modification happens inside the statement part of the ForStatement
                return !!ts.findAncestor(current, function (n) { return n === container ? "quit" : n === container.statement; });
            }
            function captureLexicalThis(node, container) {
                getNodeLinks(node).flags |= 2 /* LexicalThis */;
                if (container.kind === 165 /* PropertyDeclaration */ || container.kind === 169 /* Constructor */) {
                    var classNode = container.parent;
                    getNodeLinks(classNode).flags |= 4 /* CaptureThis */;
                }
                else {
                    getNodeLinks(container).flags |= 4 /* CaptureThis */;
                }
            }
            function findFirstSuperCall(node) {
                return ts.isSuperCall(node) ? node :
                    ts.isFunctionLike(node) ? undefined :
                        ts.forEachChild(node, findFirstSuperCall);
            }
            /**
             * Check if the given class-declaration extends null then return true.
             * Otherwise, return false
             * @param classDecl a class declaration to check if it extends null
             */
            function classDeclarationExtendsNull(classDecl) {
                var classSymbol = getSymbolOfNode(classDecl);
                var classInstanceType = getDeclaredTypeOfSymbol(classSymbol);
                var baseConstructorType = getBaseConstructorTypeOfClass(classInstanceType);
                return baseConstructorType === nullWideningType;
            }
            function checkThisBeforeSuper(node, container, diagnosticMessage) {
                var containingClassDecl = container.parent;
                var baseTypeNode = ts.getClassExtendsHeritageElement(containingClassDecl);
                // If a containing class does not have extends clause or the class extends null
                // skip checking whether super statement is called before "this" accessing.
                if (baseTypeNode && !classDeclarationExtendsNull(containingClassDecl)) {
                    if (node.flowNode && !isPostSuperFlowNode(node.flowNode, /*noCacheCheck*/ false)) {
                        error(node, diagnosticMessage);
                    }
                }
            }
            function checkThisInStaticClassFieldInitializerInDecoratedClass(thisExpression, container) {
                if (ts.isPropertyDeclaration(container) && ts.hasStaticModifier(container) &&
                    container.initializer && ts.textRangeContainsPositionInclusive(container.initializer, thisExpression.pos) && ts.length(container.parent.decorators)) {
                    error(thisExpression, ts.Diagnostics.Cannot_use_this_in_a_static_property_initializer_of_a_decorated_class);
                }
            }
            function checkThisExpression(node) {
                var isNodeInTypeQuery = isInTypeQuery(node);
                // Stop at the first arrow function so that we can
                // tell whether 'this' needs to be captured.
                var container = ts.getThisContainer(node, /* includeArrowFunctions */ true);
                var capturedByArrowFunction = false;
                if (container.kind === 169 /* Constructor */) {
                    checkThisBeforeSuper(node, container, ts.Diagnostics.super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class);
                }
                // Now skip arrow functions to get the "real" owner of 'this'.
                if (container.kind === 212 /* ArrowFunction */) {
                    container = ts.getThisContainer(container, /* includeArrowFunctions */ false);
                    capturedByArrowFunction = true;
                }
                checkThisInStaticClassFieldInitializerInDecoratedClass(node, container);
                switch (container.kind) {
                    case 259 /* ModuleDeclaration */:
                        error(node, ts.Diagnostics.this_cannot_be_referenced_in_a_module_or_namespace_body);
                        // do not return here so in case if lexical this is captured - it will be reflected in flags on NodeLinks
                        break;
                    case 258 /* EnumDeclaration */:
                        error(node, ts.Diagnostics.this_cannot_be_referenced_in_current_location);
                        // do not return here so in case if lexical this is captured - it will be reflected in flags on NodeLinks
                        break;
                    case 169 /* Constructor */:
                        if (isInConstructorArgumentInitializer(node, container)) {
                            error(node, ts.Diagnostics.this_cannot_be_referenced_in_constructor_arguments);
                            // do not return here so in case if lexical this is captured - it will be reflected in flags on NodeLinks
                        }
                        break;
                    case 160 /* ComputedPropertyName */:
                        error(node, ts.Diagnostics.this_cannot_be_referenced_in_a_computed_property_name);
                        break;
                }
                // When targeting es6, mark that we'll need to capture `this` in its lexically bound scope.
                if (!isNodeInTypeQuery && capturedByArrowFunction && languageVersion < 2 /* ES2015 */) {
                    captureLexicalThis(node, container);
                }
                var type = tryGetThisTypeAt(node, /*includeGlobalThis*/ true, container);
                if (noImplicitThis) {
                    var globalThisType_1 = getTypeOfSymbol(globalThisSymbol);
                    if (type === globalThisType_1 && capturedByArrowFunction) {
                        error(node, ts.Diagnostics.The_containing_arrow_function_captures_the_global_value_of_this);
                    }
                    else if (!type) {
                        // With noImplicitThis, functions may not reference 'this' if it has type 'any'
                        var diag = error(node, ts.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation);
                        if (!ts.isSourceFile(container)) {
                            var outsideThis = tryGetThisTypeAt(container);
                            if (outsideThis && outsideThis !== globalThisType_1) {
                                ts.addRelatedInfo(diag, ts.createDiagnosticForNode(container, ts.Diagnostics.An_outer_value_of_this_is_shadowed_by_this_container));
                            }
                        }
                    }
                }
                return type || anyType;
            }
            function tryGetThisTypeAt(node, includeGlobalThis, container) {
                if (includeGlobalThis === void 0) {
                    includeGlobalThis = true;
                }
                if (container === void 0) {
                    container = ts.getThisContainer(node, /*includeArrowFunctions*/ false);
                }
                var isInJS = ts.isInJSFile(node);
                if (ts.isFunctionLike(container) &&
                    (!isInParameterInitializerBeforeContainingFunction(node) || ts.getThisParameter(container))) {
                    var thisType = getThisTypeOfDeclaration(container) || isInJS && getTypeForThisExpressionFromJSDoc(container);
                    // Note: a parameter initializer should refer to class-this unless function-this is explicitly annotated.
                    // If this is a function in a JS file, it might be a class method.
                    if (!thisType) {
                        var className = getClassNameFromPrototypeMethod(container);
                        if (isInJS && className) {
                            var classSymbol = checkExpression(className).symbol;
                            if (classSymbol && classSymbol.members && (classSymbol.flags & 16 /* Function */)) {
                                thisType = getDeclaredTypeOfSymbol(classSymbol).thisType;
                            }
                        }
                        else if (isJSConstructor(container)) {
                            thisType = getDeclaredTypeOfSymbol(getMergedSymbol(container.symbol)).thisType;
                        }
                        thisType || (thisType = getContextualThisParameterType(container));
                    }
                    if (thisType) {
                        return getFlowTypeOfReference(node, thisType);
                    }
                }
                if (ts.isClassLike(container.parent)) {
                    var symbol = getSymbolOfNode(container.parent);
                    var type = ts.isStatic(container) ? getTypeOfSymbol(symbol) : getDeclaredTypeOfSymbol(symbol).thisType;
                    return getFlowTypeOfReference(node, type);
                }
                if (ts.isSourceFile(container)) {
                    // look up in the source file's locals or exports
                    if (container.commonJsModuleIndicator) {
                        var fileSymbol = getSymbolOfNode(container);
                        return fileSymbol && getTypeOfSymbol(fileSymbol);
                    }
                    else if (container.externalModuleIndicator) {
                        // TODO: Maybe issue a better error than 'object is possibly undefined'
                        return undefinedType;
                    }
                    else if (includeGlobalThis) {
                        return getTypeOfSymbol(globalThisSymbol);
                    }
                }
            }
            function getExplicitThisType(node) {
                var container = ts.getThisContainer(node, /*includeArrowFunctions*/ false);
                if (ts.isFunctionLike(container)) {
                    var signature = getSignatureFromDeclaration(container);
                    if (signature.thisParameter) {
                        return getExplicitTypeOfSymbol(signature.thisParameter);
                    }
                }
                if (ts.isClassLike(container.parent)) {
                    var symbol = getSymbolOfNode(container.parent);
                    return ts.isStatic(container) ? getTypeOfSymbol(symbol) : getDeclaredTypeOfSymbol(symbol).thisType;
                }
            }
            function getClassNameFromPrototypeMethod(container) {
                // Check if it's the RHS of a x.prototype.y = function [name]() { .... }
                if (container.kind === 211 /* FunctionExpression */ &&
                    ts.isBinaryExpression(container.parent) &&
                    ts.getAssignmentDeclarationKind(container.parent) === 3 /* PrototypeProperty */) {
                    // Get the 'x' of 'x.prototype.y = container'
                    return container.parent // x.prototype.y = container
                        .left // x.prototype.y
                        .expression // x.prototype
                        .expression; // x
                }
                // x.prototype = { method() { } }
                else if (container.kind === 167 /* MethodDeclaration */ &&
                    container.parent.kind === 203 /* ObjectLiteralExpression */ &&
                    ts.isBinaryExpression(container.parent.parent) &&
                    ts.getAssignmentDeclarationKind(container.parent.parent) === 6 /* Prototype */) {
                    return container.parent.parent.left.expression;
                }
                // x.prototype = { method: function() { } }
                else if (container.kind === 211 /* FunctionExpression */ &&
                    container.parent.kind === 291 /* PropertyAssignment */ &&
                    container.parent.parent.kind === 203 /* ObjectLiteralExpression */ &&
                    ts.isBinaryExpression(container.parent.parent.parent) &&
                    ts.getAssignmentDeclarationKind(container.parent.parent.parent) === 6 /* Prototype */) {
                    return container.parent.parent.parent.left.expression;
                }
                // Object.defineProperty(x, "method", { value: function() { } });
                // Object.defineProperty(x, "method", { set: (x: () => void) => void });
                // Object.defineProperty(x, "method", { get: () => function() { }) });
                else if (container.kind === 211 /* FunctionExpression */ &&
                    ts.isPropertyAssignment(container.parent) &&
                    ts.isIdentifier(container.parent.name) &&
                    (container.parent.name.escapedText === "value" || container.parent.name.escapedText === "get" || container.parent.name.escapedText === "set") &&
                    ts.isObjectLiteralExpression(container.parent.parent) &&
                    ts.isCallExpression(container.parent.parent.parent) &&
                    container.parent.parent.parent.arguments[2] === container.parent.parent &&
                    ts.getAssignmentDeclarationKind(container.parent.parent.parent) === 9 /* ObjectDefinePrototypeProperty */) {
                    return container.parent.parent.parent.arguments[0].expression;
                }
                // Object.defineProperty(x, "method", { value() { } });
                // Object.defineProperty(x, "method", { set(x: () => void) {} });
                // Object.defineProperty(x, "method", { get() { return () => {} } });
                else if (ts.isMethodDeclaration(container) &&
                    ts.isIdentifier(container.name) &&
                    (container.name.escapedText === "value" || container.name.escapedText === "get" || container.name.escapedText === "set") &&
                    ts.isObjectLiteralExpression(container.parent) &&
                    ts.isCallExpression(container.parent.parent) &&
                    container.parent.parent.arguments[2] === container.parent &&
                    ts.getAssignmentDeclarationKind(container.parent.parent) === 9 /* ObjectDefinePrototypeProperty */) {
                    return container.parent.parent.arguments[0].expression;
                }
            }
            function getTypeForThisExpressionFromJSDoc(node) {
                var jsdocType = ts.getJSDocType(node);
                if (jsdocType && jsdocType.kind === 312 /* JSDocFunctionType */) {
                    var jsDocFunctionType = jsdocType;
                    if (jsDocFunctionType.parameters.length > 0 &&
                        jsDocFunctionType.parameters[0].name &&
                        jsDocFunctionType.parameters[0].name.escapedText === "this" /* This */) {
                        return getTypeFromTypeNode(jsDocFunctionType.parameters[0].type);
                    }
                }
                var thisTag = ts.getJSDocThisTag(node);
                if (thisTag && thisTag.typeExpression) {
                    return getTypeFromTypeNode(thisTag.typeExpression);
                }
            }
            function isInConstructorArgumentInitializer(node, constructorDecl) {
                return !!ts.findAncestor(node, function (n) { return ts.isFunctionLikeDeclaration(n) ? "quit" : n.kind === 162 /* Parameter */ && n.parent === constructorDecl; });
            }
            function checkSuperExpression(node) {
                var isCallExpression = node.parent.kind === 206 /* CallExpression */ && node.parent.expression === node;
                var immediateContainer = ts.getSuperContainer(node, /*stopOnFunctions*/ true);
                var container = immediateContainer;
                var needToCaptureLexicalThis = false;
                // adjust the container reference in case if super is used inside arrow functions with arbitrarily deep nesting
                if (!isCallExpression) {
                    while (container && container.kind === 212 /* ArrowFunction */) {
                        container = ts.getSuperContainer(container, /*stopOnFunctions*/ true);
                        needToCaptureLexicalThis = languageVersion < 2 /* ES2015 */;
                    }
                }
                var canUseSuperExpression = isLegalUsageOfSuperExpression(container);
                var nodeCheckFlag = 0;
                if (!canUseSuperExpression) {
                    // issue more specific error if super is used in computed property name
                    // class A { foo() { return "1" }}
                    // class B {
                    //     [super.foo()]() {}
                    // }
                    var current = ts.findAncestor(node, function (n) { return n === container ? "quit" : n.kind === 160 /* ComputedPropertyName */; });
                    if (current && current.kind === 160 /* ComputedPropertyName */) {
                        error(node, ts.Diagnostics.super_cannot_be_referenced_in_a_computed_property_name);
                    }
                    else if (isCallExpression) {
                        error(node, ts.Diagnostics.Super_calls_are_not_permitted_outside_constructors_or_in_nested_functions_inside_constructors);
                    }
                    else if (!container || !container.parent || !(ts.isClassLike(container.parent) || container.parent.kind === 203 /* ObjectLiteralExpression */)) {
                        error(node, ts.Diagnostics.super_can_only_be_referenced_in_members_of_derived_classes_or_object_literal_expressions);
                    }
                    else {
                        error(node, ts.Diagnostics.super_property_access_is_permitted_only_in_a_constructor_member_function_or_member_accessor_of_a_derived_class);
                    }
                    return errorType;
                }
                if (!isCallExpression && immediateContainer.kind === 169 /* Constructor */) {
                    checkThisBeforeSuper(node, container, ts.Diagnostics.super_must_be_called_before_accessing_a_property_of_super_in_the_constructor_of_a_derived_class);
                }
                if (ts.isStatic(container) || isCallExpression) {
                    nodeCheckFlag = 512 /* SuperStatic */;
                    if (!isCallExpression &&
                        languageVersion >= 2 /* ES2015 */ && languageVersion <= 8 /* ES2021 */ &&
                        (ts.isPropertyDeclaration(container) || ts.isClassStaticBlockDeclaration(container))) {
                        // for `super.x` or `super[x]` in a static initializer, mark all enclosing
                        // block scope containers so that we can report potential collisions with
                        // `Reflect`.
                        ts.forEachEnclosingBlockScopeContainer(node.parent, function (current) {
                            if (!ts.isSourceFile(current) || ts.isExternalOrCommonJsModule(current)) {
                                getNodeLinks(current).flags |= 134217728 /* ContainsSuperPropertyInStaticInitializer */;
                            }
                        });
                    }
                }
                else {
                    nodeCheckFlag = 256 /* SuperInstance */;
                }
                getNodeLinks(node).flags |= nodeCheckFlag;
                // Due to how we emit async functions, we need to specialize the emit for an async method that contains a `super` reference.
                // This is due to the fact that we emit the body of an async function inside of a generator function. As generator
                // functions cannot reference `super`, we emit a helper inside of the method body, but outside of the generator. This helper
                // uses an arrow function, which is permitted to reference `super`.
                //
                // There are two primary ways we can access `super` from within an async method. The first is getting the value of a property
                // or indexed access on super, either as part of a right-hand-side expression or call expression. The second is when setting the value
                // of a property or indexed access, either as part of an assignment expression or destructuring assignment.
                //
                // The simplest case is reading a value, in which case we will emit something like the following:
                //
                //  // ts
                //  ...
                //  async asyncMethod() {
                //    let x = await super.asyncMethod();
                //    return x;
                //  }
                //  ...
                //
                //  // js
                //  ...
                //  asyncMethod() {
                //      const _super = Object.create(null, {
                //        asyncMethod: { get: () => super.asyncMethod },
                //      });
                //      return __awaiter(this, arguments, Promise, function *() {
                //          let x = yield _super.asyncMethod.call(this);
                //          return x;
                //      });
                //  }
                //  ...
                //
                // The more complex case is when we wish to assign a value, especially as part of a destructuring assignment. As both cases
                // are legal in ES6, but also likely less frequent, we only emit setters if there is an assignment:
                //
                //  // ts
                //  ...
                //  async asyncMethod(ar: Promise<any[]>) {
                //      [super.a, super.b] = await ar;
                //  }
                //  ...
                //
                //  // js
                //  ...
                //  asyncMethod(ar) {
                //      const _super = Object.create(null, {
                //        a: { get: () => super.a, set: (v) => super.a = v },
                //        b: { get: () => super.b, set: (v) => super.b = v }
                //      };
                //      return __awaiter(this, arguments, Promise, function *() {
                //          [_super.a, _super.b] = yield ar;
                //      });
                //  }
                //  ...
                //
                // Creating an object that has getter and setters instead of just an accessor function is required for destructuring assignments
                // as a call expression cannot be used as the target of a destructuring assignment while a property access can.
                //
                // For element access expressions (`super[x]`), we emit a generic helper that forwards the element access in both situations.
                if (container.kind === 167 /* MethodDeclaration */ && ts.hasSyntacticModifier(container, 256 /* Async */)) {
                    if (ts.isSuperProperty(node.parent) && ts.isAssignmentTarget(node.parent)) {
                        getNodeLinks(container).flags |= 4096 /* AsyncMethodWithSuperBinding */;
                    }
                    else {
                        getNodeLinks(container).flags |= 2048 /* AsyncMethodWithSuper */;
                    }
                }
                if (needToCaptureLexicalThis) {
                    // call expressions are allowed only in constructors so they should always capture correct 'this'
                    // super property access expressions can also appear in arrow functions -
                    // in this case they should also use correct lexical this
                    captureLexicalThis(node.parent, container);
                }
                if (container.parent.kind === 203 /* ObjectLiteralExpression */) {
                    if (languageVersion < 2 /* ES2015 */) {
                        error(node, ts.Diagnostics.super_is_only_allowed_in_members_of_object_literal_expressions_when_option_target_is_ES2015_or_higher);
                        return errorType;
                    }
                    else {
                        // for object literal assume that type of 'super' is 'any'
                        return anyType;
                    }
                }
                // at this point the only legal case for parent is ClassLikeDeclaration
                var classLikeDeclaration = container.parent;
                if (!ts.getClassExtendsHeritageElement(classLikeDeclaration)) {
                    error(node, ts.Diagnostics.super_can_only_be_referenced_in_a_derived_class);
                    return errorType;
                }
                var classType = getDeclaredTypeOfSymbol(getSymbolOfNode(classLikeDeclaration));
                var baseClassType = classType && getBaseTypes(classType)[0];
                if (!baseClassType) {
                    return errorType;
                }
                if (container.kind === 169 /* Constructor */ && isInConstructorArgumentInitializer(node, container)) {
                    // issue custom error message for super property access in constructor arguments (to be aligned with old compiler)
                    error(node, ts.Diagnostics.super_cannot_be_referenced_in_constructor_arguments);
                    return errorType;
                }
                return nodeCheckFlag === 512 /* SuperStatic */
                    ? getBaseConstructorTypeOfClass(classType)
                    : getTypeWithThisArgument(baseClassType, classType.thisType);
                function isLegalUsageOfSuperExpression(container) {
                    if (!container) {
                        return false;
                    }
                    if (isCallExpression) {
                        // TS 1.0 SPEC (April 2014): 4.8.1
                        // Super calls are only permitted in constructors of derived classes
                        return container.kind === 169 /* Constructor */;
                    }
                    else {
                        // TS 1.0 SPEC (April 2014)
                        // 'super' property access is allowed
                        // - In a constructor, instance member function, instance member accessor, or instance member variable initializer where this references a derived class instance
                        // - In a static member function or static member accessor
                        // topmost container must be something that is directly nested in the class declaration\object literal expression
                        if (ts.isClassLike(container.parent) || container.parent.kind === 203 /* ObjectLiteralExpression */) {
                            if (ts.isStatic(container)) {
                                return container.kind === 167 /* MethodDeclaration */ ||
                                    container.kind === 166 /* MethodSignature */ ||
                                    container.kind === 170 /* GetAccessor */ ||
                                    container.kind === 171 /* SetAccessor */ ||
                                    container.kind === 165 /* PropertyDeclaration */ ||
                                    container.kind === 168 /* ClassStaticBlockDeclaration */;
                            }
                            else {
                                return container.kind === 167 /* MethodDeclaration */ ||
                                    container.kind === 166 /* MethodSignature */ ||
                                    container.kind === 170 /* GetAccessor */ ||
                                    container.kind === 171 /* SetAccessor */ ||
                                    container.kind === 165 /* PropertyDeclaration */ ||
                                    container.kind === 164 /* PropertySignature */ ||
                                    container.kind === 169 /* Constructor */;
                            }
                        }
                    }
                    return false;
                }
            }
            function getContainingObjectLiteral(func) {
                return (func.kind === 167 /* MethodDeclaration */ ||
                    func.kind === 170 /* GetAccessor */ ||
                    func.kind === 171 /* SetAccessor */) && func.parent.kind === 203 /* ObjectLiteralExpression */ ? func.parent :
                    func.kind === 211 /* FunctionExpression */ && func.parent.kind === 291 /* PropertyAssignment */ ? func.parent.parent :
                        undefined;
            }
            function getThisTypeArgument(type) {
                return ts.getObjectFlags(type) & 4 /* Reference */ && type.target === globalThisType ? getTypeArguments(type)[0] : undefined;
            }
            function getThisTypeFromContextualType(type) {
                return mapType(type, function (t) {
                    return t.flags & 2097152 /* Intersection */ ? ts.forEach(t.types, getThisTypeArgument) : getThisTypeArgument(t);
                });
            }
            function getContextualThisParameterType(func) {
                if (func.kind === 212 /* ArrowFunction */) {
                    return undefined;
                }
                if (isContextSensitiveFunctionOrObjectLiteralMethod(func)) {
                    var contextualSignature = getContextualSignature(func);
                    if (contextualSignature) {
                        var thisParameter = contextualSignature.thisParameter;
                        if (thisParameter) {
                            return getTypeOfSymbol(thisParameter);
                        }
                    }
                }
                var inJs = ts.isInJSFile(func);
                if (noImplicitThis || inJs) {
                    var containingLiteral = getContainingObjectLiteral(func);
                    if (containingLiteral) {
                        // We have an object literal method. Check if the containing object literal has a contextual type
                        // that includes a ThisType<T>. If so, T is the contextual type for 'this'. We continue looking in
                        // any directly enclosing object literals.
                        var contextualType = getApparentTypeOfContextualType(containingLiteral);
                        var literal = containingLiteral;
                        var type = contextualType;
                        while (type) {
                            var thisType = getThisTypeFromContextualType(type);
                            if (thisType) {
                                return instantiateType(thisType, getMapperFromContext(getInferenceContext(containingLiteral)));
                            }
                            if (literal.parent.kind !== 291 /* PropertyAssignment */) {
                                break;
                            }
                            literal = literal.parent.parent;
                            type = getApparentTypeOfContextualType(literal);
                        }
                        // There was no contextual ThisType<T> for the containing object literal, so the contextual type
                        // for 'this' is the non-null form of the contextual type for the containing object literal or
                        // the type of the object literal itself.
                        return getWidenedType(contextualType ? getNonNullableType(contextualType) : checkExpressionCached(containingLiteral));
                    }
                    // In an assignment of the form 'obj.xxx = function(...)' or 'obj[xxx] = function(...)', the
                    // contextual type for 'this' is 'obj'.
                    var parent = ts.walkUpParenthesizedExpressions(func.parent);
                    if (parent.kind === 219 /* BinaryExpression */ && parent.operatorToken.kind === 63 /* EqualsToken */) {
                        var target = parent.left;
                        if (ts.isAccessExpression(target)) {
                            var expression = target.expression;
                            // Don't contextually type `this` as `exports` in `exports.Point = function(x, y) { this.x = x; this.y = y; }`
                            if (inJs && ts.isIdentifier(expression)) {
                                var sourceFile = ts.getSourceFileOfNode(parent);
                                if (sourceFile.commonJsModuleIndicator && getResolvedSymbol(expression) === sourceFile.symbol) {
                                    return undefined;
                                }
                            }
                            return getWidenedType(checkExpressionCached(expression));
                        }
                    }
                }
                return undefined;
            }
            // Return contextual type of parameter or undefined if no contextual type is available
            function getContextuallyTypedParameterType(parameter) {
                var func = parameter.parent;
                if (!isContextSensitiveFunctionOrObjectLiteralMethod(func)) {
                    return undefined;
                }
                var iife = ts.getImmediatelyInvokedFunctionExpression(func);
                if (iife && iife.arguments) {
                    var args = getEffectiveCallArguments(iife);
                    var indexOfParameter = func.parameters.indexOf(parameter);
                    if (parameter.dotDotDotToken) {
                        return getSpreadArgumentType(args, indexOfParameter, args.length, anyType, /*context*/ undefined, 0 /* Normal */);
                    }
                    var links = getNodeLinks(iife);
                    var cached = links.resolvedSignature;
                    links.resolvedSignature = anySignature;
                    var type = indexOfParameter < args.length ?
                        getWidenedLiteralType(checkExpression(args[indexOfParameter])) :
                        parameter.initializer ? undefined : undefinedWideningType;
                    links.resolvedSignature = cached;
                    return type;
                }
                var contextualSignature = getContextualSignature(func);
                if (contextualSignature) {
                    var index = func.parameters.indexOf(parameter) - (ts.getThisParameter(func) ? 1 : 0);
                    return parameter.dotDotDotToken && ts.lastOrUndefined(func.parameters) === parameter ?
                        getRestTypeAtPosition(contextualSignature, index) :
                        tryGetTypeAtPosition(contextualSignature, index);
                }
            }
            function getContextualTypeForVariableLikeDeclaration(declaration) {
                var typeNode = ts.getEffectiveTypeAnnotationNode(declaration);
                if (typeNode) {
                    return getTypeFromTypeNode(typeNode);
                }
                switch (declaration.kind) {
                    case 162 /* Parameter */:
                        return getContextuallyTypedParameterType(declaration);
                    case 201 /* BindingElement */:
                        return getContextualTypeForBindingElement(declaration);
                    case 165 /* PropertyDeclaration */:
                        if (ts.isStatic(declaration)) {
                            return getContextualTypeForStaticPropertyDeclaration(declaration);
                        }
                    // By default, do nothing and return undefined - only the above cases have context implied by a parent
                }
            }
            function getContextualTypeForBindingElement(declaration) {
                var parent = declaration.parent.parent;
                var name = declaration.propertyName || declaration.name;
                var parentType = getContextualTypeForVariableLikeDeclaration(parent) ||
                    parent.kind !== 201 /* BindingElement */ && parent.initializer && checkDeclarationInitializer(parent);
                if (!parentType || ts.isBindingPattern(name) || ts.isComputedNonLiteralName(name))
                    return undefined;
                if (parent.name.kind === 200 /* ArrayBindingPattern */) {
                    var index = ts.indexOfNode(declaration.parent.elements, declaration);
                    if (index < 0)
                        return undefined;
                    return getContextualTypeForElementExpression(parentType, index);
                }
                var nameType = getLiteralTypeFromPropertyName(name);
                if (isTypeUsableAsPropertyName(nameType)) {
                    var text = getPropertyNameFromType(nameType);
                    return getTypeOfPropertyOfType(parentType, text);
                }
            }
            function getContextualTypeForStaticPropertyDeclaration(declaration) {
                var parentType = ts.isExpression(declaration.parent) && getContextualType(declaration.parent);
                if (!parentType)
                    return undefined;
                return getTypeOfPropertyOfContextualType(parentType, getSymbolOfNode(declaration).escapedName);
            }
            // In a variable, parameter or property declaration with a type annotation,
            //   the contextual type of an initializer expression is the type of the variable, parameter or property.
            // Otherwise, in a parameter declaration of a contextually typed function expression,
            //   the contextual type of an initializer expression is the contextual type of the parameter.
            // Otherwise, in a variable or parameter declaration with a binding pattern name,
            //   the contextual type of an initializer expression is the type implied by the binding pattern.
            // Otherwise, in a binding pattern inside a variable or parameter declaration,
            //   the contextual type of an initializer expression is the type annotation of the containing declaration, if present.
            function getContextualTypeForInitializerExpression(node, contextFlags) {
                var declaration = node.parent;
                if (ts.hasInitializer(declaration) && node === declaration.initializer) {
                    var result = getContextualTypeForVariableLikeDeclaration(declaration);
                    if (result) {
                        return result;
                    }
                    if (!(contextFlags & 8 /* SkipBindingPatterns */) && ts.isBindingPattern(declaration.name)) { // This is less a contextual type and more an implied shape - in some cases, this may be undesirable
                        return getTypeFromBindingPattern(declaration.name, /*includePatternInType*/ true, /*reportErrors*/ false);
                    }
                }
                return undefined;
            }
            function getContextualTypeForReturnExpression(node) {
                var func = ts.getContainingFunction(node);
                if (func) {
                    var contextualReturnType = getContextualReturnType(func);
                    if (contextualReturnType) {
                        var functionFlags = ts.getFunctionFlags(func);
                        if (functionFlags & 1 /* Generator */) { // Generator or AsyncGenerator function
                            var use = functionFlags & 2 /* Async */ ? 2 /* AsyncGeneratorReturnType */ : 1 /* GeneratorReturnType */;
                            var iterationTypes = getIterationTypesOfIterable(contextualReturnType, use, /*errorNode*/ undefined);
                            if (!iterationTypes) {
                                return undefined;
                            }
                            contextualReturnType = iterationTypes.returnType;
                            // falls through to unwrap Promise for AsyncGenerators
                        }
                        if (functionFlags & 2 /* Async */) { // Async function or AsyncGenerator function
                            var contextualAwaitedType = mapType(contextualReturnType, getAwaitedType);
                            return contextualAwaitedType && getUnionType([contextualAwaitedType, createPromiseLikeType(contextualAwaitedType)]);
                        }
                        return contextualReturnType; // Regular function or Generator function
                    }
                }
                return undefined;
            }
            function getContextualTypeForAwaitOperand(node, contextFlags) {
                var contextualType = getContextualType(node, contextFlags);
                if (contextualType) {
                    var contextualAwaitedType = getAwaitedType(contextualType);
                    return contextualAwaitedType && getUnionType([contextualAwaitedType, createPromiseLikeType(contextualAwaitedType)]);
                }
                return undefined;
            }
            function getContextualTypeForYieldOperand(node) {
                var func = ts.getContainingFunction(node);
                if (func) {
                    var functionFlags = ts.getFunctionFlags(func);
                    var contextualReturnType = getContextualReturnType(func);
                    if (contextualReturnType) {
                        return node.asteriskToken
                            ? contextualReturnType
                            : getIterationTypeOfGeneratorFunctionReturnType(0 /* Yield */, contextualReturnType, (functionFlags & 2 /* Async */) !== 0);
                    }
                }
                return undefined;
            }
            function isInParameterInitializerBeforeContainingFunction(node) {
                var inBindingInitializer = false;
                while (node.parent && !ts.isFunctionLike(node.parent)) {
                    if (ts.isParameter(node.parent) && (inBindingInitializer || node.parent.initializer === node)) {
                        return true;
                    }
                    if (ts.isBindingElement(node.parent) && node.parent.initializer === node) {
                        inBindingInitializer = true;
                    }
                    node = node.parent;
                }
                return false;
            }
            function getContextualIterationType(kind, functionDecl) {
                var isAsync = !!(ts.getFunctionFlags(functionDecl) & 2 /* Async */);
                var contextualReturnType = getContextualReturnType(functionDecl);
                if (contextualReturnType) {
                    return getIterationTypeOfGeneratorFunctionReturnType(kind, contextualReturnType, isAsync)
                        || undefined;
                }
                return undefined;
            }
            function getContextualReturnType(functionDecl) {
                // If the containing function has a return type annotation, is a constructor, or is a get accessor whose
                // corresponding set accessor has a type annotation, return statements in the function are contextually typed
                var returnType = getReturnTypeFromAnnotation(functionDecl);
                if (returnType) {
                    return returnType;
                }
                // Otherwise, if the containing function is contextually typed by a function type with exactly one call signature
                // and that call signature is non-generic, return statements are contextually typed by the return type of the signature
                var signature = getContextualSignatureForFunctionLikeDeclaration(functionDecl);
                if (signature && !isResolvingReturnTypeOfSignature(signature)) {
                    return getReturnTypeOfSignature(signature);
                }
                return undefined;
            }
            // In a typed function call, an argument or substitution expression is contextually typed by the type of the corresponding parameter.
            function getContextualTypeForArgument(callTarget, arg) {
                var args = getEffectiveCallArguments(callTarget);
                var argIndex = args.indexOf(arg); // -1 for e.g. the expression of a CallExpression, or the tag of a TaggedTemplateExpression
                return argIndex === -1 ? undefined : getContextualTypeForArgumentAtIndex(callTarget, argIndex);
            }
            function getContextualTypeForArgumentAtIndex(callTarget, argIndex) {
                // If we're already in the process of resolving the given signature, don't resolve again as
                // that could cause infinite recursion. Instead, return anySignature.
                var signature = getNodeLinks(callTarget).resolvedSignature === resolvingSignature ? resolvingSignature : getResolvedSignature(callTarget);
                if (ts.isJsxOpeningLikeElement(callTarget) && argIndex === 0) {
                    return getEffectiveFirstArgumentForJsxSignature(signature, callTarget);
                }
                var restIndex = signature.parameters.length - 1;
                return signatureHasRestParameter(signature) && argIndex >= restIndex ?
                    getIndexedAccessType(getTypeOfSymbol(signature.parameters[restIndex]), getNumberLiteralType(argIndex - restIndex), 256 /* Contextual */) :
                    getTypeAtPosition(signature, argIndex);
            }
            function getContextualTypeForSubstitutionExpression(template, substitutionExpression) {
                if (template.parent.kind === 208 /* TaggedTemplateExpression */) {
                    return getContextualTypeForArgument(template.parent, substitutionExpression);
                }
                return undefined;
            }
            function getContextualTypeForBinaryOperand(node, contextFlags) {
                var binaryExpression = node.parent;
                var left = binaryExpression.left, operatorToken = binaryExpression.operatorToken, right = binaryExpression.right;
                switch (operatorToken.kind) {
                    case 63 /* EqualsToken */:
                    case 76 /* AmpersandAmpersandEqualsToken */:
                    case 75 /* BarBarEqualsToken */:
                    case 77 /* QuestionQuestionEqualsToken */:
                        return node === right ? getContextualTypeForAssignmentDeclaration(binaryExpression) : undefined;
                    case 56 /* BarBarToken */:
                    case 60 /* QuestionQuestionToken */:
                        // When an || expression has a contextual type, the operands are contextually typed by that type, except
                        // when that type originates in a binding pattern, the right operand is contextually typed by the type of
                        // the left operand. When an || expression has no contextual type, the right operand is contextually typed
                        // by the type of the left operand, except for the special case of Javascript declarations of the form
                        // `namespace.prop = namespace.prop || {}`.
                        var type = getContextualType(binaryExpression, contextFlags);
                        return node === right && (type && type.pattern || !type && !ts.isDefaultedExpandoInitializer(binaryExpression)) ?
                            getTypeOfExpression(left) : type;
                    case 55 /* AmpersandAmpersandToken */:
                    case 27 /* CommaToken */:
                        return node === right ? getContextualType(binaryExpression, contextFlags) : undefined;
                    default:
                        return undefined;
                }
            }
            /**
             * Try to find a resolved symbol for an expression without also resolving its type, as
             * getSymbolAtLocation would (as that could be reentrant into contextual typing)
             */
            function getSymbolForExpression(e) {
                if (e.symbol) {
                    return e.symbol;
                }
                if (ts.isIdentifier(e)) {
                    return getResolvedSymbol(e);
                }
                if (ts.isPropertyAccessExpression(e)) {
                    var lhsType = getTypeOfExpression(e.expression);
                    return ts.isPrivateIdentifier(e.name) ? tryGetPrivateIdentifierPropertyOfType(lhsType, e.name) : getPropertyOfType(lhsType, e.name.escapedText);
                }
                return undefined;
                function tryGetPrivateIdentifierPropertyOfType(type, id) {
                    var lexicallyScopedSymbol = lookupSymbolForPrivateIdentifierDeclaration(id.escapedText, id);
                    return lexicallyScopedSymbol && getPrivateIdentifierPropertyOfType(type, lexicallyScopedSymbol);
                }
            }
            // In an assignment expression, the right operand is contextually typed by the type of the left operand.
            // Don't do this for assignment declarations unless there is a type tag on the assignment, to avoid circularity from checking the right operand.
            function getContextualTypeForAssignmentDeclaration(binaryExpression) {
                var _a, _b;
                var kind = ts.getAssignmentDeclarationKind(binaryExpression);
                switch (kind) {
                    case 0 /* None */:
                    case 4 /* ThisProperty */:
                        var lhsSymbol = getSymbolForExpression(binaryExpression.left);
                        var decl = lhsSymbol && lhsSymbol.valueDeclaration;
                        // Unannotated, uninitialized property declarations have a type implied by their usage in the constructor.
                        // We avoid calling back into `getTypeOfExpression` and reentering contextual typing to avoid a bogus circularity error in that case.
                        if (decl && (ts.isPropertyDeclaration(decl) || ts.isPropertySignature(decl))) {
                            var overallAnnotation = ts.getEffectiveTypeAnnotationNode(decl);
                            return (overallAnnotation && instantiateType(getTypeFromTypeNode(overallAnnotation), getSymbolLinks(lhsSymbol).mapper)) ||
                                (decl.initializer && getTypeOfExpression(binaryExpression.left));
                        }
                        if (kind === 0 /* None */) {
                            return getTypeOfExpression(binaryExpression.left);
                        }
                        return getContextualTypeForThisPropertyAssignment(binaryExpression);
                    case 5 /* Property */:
                        if (isPossiblyAliasedThisProperty(binaryExpression, kind)) {
                            return getContextualTypeForThisPropertyAssignment(binaryExpression);
                        }
                        // If `binaryExpression.left` was assigned a symbol, then this is a new declaration; otherwise it is an assignment to an existing declaration.
                        // See `bindStaticPropertyAssignment` in `binder.ts`.
                        else if (!binaryExpression.left.symbol) {
                            return getTypeOfExpression(binaryExpression.left);
                        }
                        else {
                            var decl_1 = binaryExpression.left.symbol.valueDeclaration;
                            if (!decl_1) {
                                return undefined;
                            }
                            var lhs = ts.cast(binaryExpression.left, ts.isAccessExpression);
                            var overallAnnotation = ts.getEffectiveTypeAnnotationNode(decl_1);
                            if (overallAnnotation) {
                                return getTypeFromTypeNode(overallAnnotation);
                            }
                            else if (ts.isIdentifier(lhs.expression)) {
                                var id = lhs.expression;
                                var parentSymbol = resolveName(id, id.escapedText, 111551 /* Value */, undefined, id.escapedText, /*isUse*/ true);
                                if (parentSymbol) {
                                    var annotated_1 = parentSymbol.valueDeclaration && ts.getEffectiveTypeAnnotationNode(parentSymbol.valueDeclaration);
                                    if (annotated_1) {
                                        var nameStr = ts.getElementOrPropertyAccessName(lhs);
                                        if (nameStr !== undefined) {
                                            return getTypeOfPropertyOfContextualType(getTypeFromTypeNode(annotated_1), nameStr);
                                        }
                                    }
                                    return undefined;
                                }
                            }
                            return ts.isInJSFile(decl_1) ? undefined : getTypeOfExpression(binaryExpression.left);
                        }
                    case 1 /* ExportsProperty */:
                    case 6 /* Prototype */:
                    case 3 /* PrototypeProperty */:
                        var valueDeclaration = (_a = binaryExpression.left.symbol) === null || _a === void 0 ? void 0 : _a.valueDeclaration;
                    // falls through
                    case 2 /* ModuleExports */:
                        valueDeclaration || (valueDeclaration = (_b = binaryExpression.symbol) === null || _b === void 0 ? void 0 : _b.valueDeclaration);
                        var annotated = valueDeclaration && ts.getEffectiveTypeAnnotationNode(valueDeclaration);
                        return annotated ? getTypeFromTypeNode(annotated) : undefined;
                    case 7 /* ObjectDefinePropertyValue */:
                    case 8 /* ObjectDefinePropertyExports */:
                    case 9 /* ObjectDefinePrototypeProperty */:
                        return ts.Debug.fail("Does not apply");
                    default:
                        return ts.Debug.assertNever(kind);
                }
            }
            function isPossiblyAliasedThisProperty(declaration, kind) {
                if (kind === void 0) {
                    kind = ts.getAssignmentDeclarationKind(declaration);
                }
                if (kind === 4 /* ThisProperty */) {
                    return true;
                }
                if (!ts.isInJSFile(declaration) || kind !== 5 /* Property */ || !ts.isIdentifier(declaration.left.expression)) {
                    return false;
                }
                var name = declaration.left.expression.escapedText;
                var symbol = resolveName(declaration.left, name, 111551 /* Value */, undefined, undefined, /*isUse*/ true, /*excludeGlobals*/ true);
                return ts.isThisInitializedDeclaration(symbol === null || symbol === void 0 ? void 0 : symbol.valueDeclaration);
            }
            function getContextualTypeForThisPropertyAssignment(binaryExpression) {
                if (!binaryExpression.symbol)
                    return getTypeOfExpression(binaryExpression.left);
                if (binaryExpression.symbol.valueDeclaration) {
                    var annotated = ts.getEffectiveTypeAnnotationNode(binaryExpression.symbol.valueDeclaration);
                    if (annotated) {
                        var type = getTypeFromTypeNode(annotated);
                        if (type) {
                            return type;
                        }
                    }
                }
                var thisAccess = ts.cast(binaryExpression.left, ts.isAccessExpression);
                if (!ts.isObjectLiteralMethod(ts.getThisContainer(thisAccess.expression, /*includeArrowFunctions*/ false))) {
                    return undefined;
                }
                var thisType = checkThisExpression(thisAccess.expression);
                var nameStr = ts.getElementOrPropertyAccessName(thisAccess);
                return nameStr !== undefined && getTypeOfPropertyOfContextualType(thisType, nameStr) || undefined;
            }
            function isCircularMappedProperty(symbol) {
                return !!(ts.getCheckFlags(symbol) & 262144 /* Mapped */ && !symbol.type && findResolutionCycleStartIndex(symbol, 0 /* Type */) >= 0);
            }
            function getTypeOfPropertyOfContextualType(type, name) {
                return mapType(type, function (t) {
                    var _a;
                    if (isGenericMappedType(t)) {
                        var constraint = getConstraintTypeFromMappedType(t);
                        var constraintOfConstraint = getBaseConstraintOfType(constraint) || constraint;
                        var propertyNameType = getStringLiteralType(ts.unescapeLeadingUnderscores(name));
                        if (isTypeAssignableTo(propertyNameType, constraintOfConstraint)) {
                            return substituteIndexedMappedType(t, propertyNameType);
                        }
                    }
                    else if (t.flags & 3670016 /* StructuredType */) {
                        var prop = getPropertyOfType(t, name);
                        if (prop) {
                            return isCircularMappedProperty(prop) ? undefined : getTypeOfSymbol(prop);
                        }
                        if (isTupleType(t)) {
                            var restType = getRestTypeOfTupleType(t);
                            if (restType && isNumericLiteralName(name) && +name >= 0) {
                                return restType;
                            }
                        }
                        return (_a = findApplicableIndexInfo(getIndexInfosOfStructuredType(t), getStringLiteralType(ts.unescapeLeadingUnderscores(name)))) === null || _a === void 0 ? void 0 : _a.type;
                    }
                    return undefined;
                }, /*noReductions*/ true);
            }
            // In an object literal contextually typed by a type T, the contextual type of a property assignment is the type of
            // the matching property in T, if one exists. Otherwise, it is the type of the numeric index signature in T, if one
            // exists. Otherwise, it is the type of the string index signature in T, if one exists.
            function getContextualTypeForObjectLiteralMethod(node, contextFlags) {
                ts.Debug.assert(ts.isObjectLiteralMethod(node));
                if (node.flags & 16777216 /* InWithStatement */) {
                    // We cannot answer semantic questions within a with block, do not proceed any further
                    return undefined;
                }
                return getContextualTypeForObjectLiteralElement(node, contextFlags);
            }
            function getContextualTypeForObjectLiteralElement(element, contextFlags) {
                var objectLiteral = element.parent;
                var propertyAssignmentType = ts.isPropertyAssignment(element) && getContextualTypeForVariableLikeDeclaration(element);
                if (propertyAssignmentType) {
                    return propertyAssignmentType;
                }
                var type = getApparentTypeOfContextualType(objectLiteral, contextFlags);
                if (type) {
                    if (hasBindableName(element)) {
                        // For a (non-symbol) computed property, there is no reason to look up the name
                        // in the type. It will just be "__computed", which does not appear in any
                        // SymbolTable.
                        return getTypeOfPropertyOfContextualType(type, getSymbolOfNode(element).escapedName);
                    }
                    if (element.name) {
                        var nameType_1 = getLiteralTypeFromPropertyName(element.name);
                        // We avoid calling getApplicableIndexInfo here because it performs potentially expensive intersection reduction.
                        return mapType(type, function (t) { var _a; return (_a = findApplicableIndexInfo(getIndexInfosOfStructuredType(t), nameType_1)) === null || _a === void 0 ? void 0 : _a.type; }, /*noReductions*/ true);
                    }
                }
                return undefined;
            }
            // In an array literal contextually typed by a type T, the contextual type of an element expression at index N is
            // the type of the property with the numeric name N in T, if one exists. Otherwise, if T has a numeric index signature,
            // it is the type of the numeric index signature in T. Otherwise, in ES6 and higher, the contextual type is the iterated
            // type of T.
            function getContextualTypeForElementExpression(arrayContextualType, index) {
                return arrayContextualType && (getTypeOfPropertyOfContextualType(arrayContextualType, "" + index)
                    || mapType(arrayContextualType, function (t) { return getIteratedTypeOrElementType(1 /* Element */, t, undefinedType, /*errorNode*/ undefined, /*checkAssignability*/ false); }, 
                    /*noReductions*/ true));
            }
            // In a contextually typed conditional expression, the true/false expressions are contextually typed by the same type.
            function getContextualTypeForConditionalOperand(node, contextFlags) {
                var conditional = node.parent;
                return node === conditional.whenTrue || node === conditional.whenFalse ? getContextualType(conditional, contextFlags) : undefined;
            }
            function getContextualTypeForChildJsxExpression(node, child) {
                var attributesType = getApparentTypeOfContextualType(node.openingElement.tagName);
                // JSX expression is in children of JSX Element, we will look for an "children" attribute (we get the name from JSX.ElementAttributesProperty)
                var jsxChildrenPropertyName = getJsxElementChildrenPropertyName(getJsxNamespaceAt(node));
                if (!(attributesType && !isTypeAny(attributesType) && jsxChildrenPropertyName && jsxChildrenPropertyName !== "")) {
                    return undefined;
                }
                var realChildren = ts.getSemanticJsxChildren(node.children);
                var childIndex = realChildren.indexOf(child);
                var childFieldType = getTypeOfPropertyOfContextualType(attributesType, jsxChildrenPropertyName);
                return childFieldType && (realChildren.length === 1 ? childFieldType : mapType(childFieldType, function (t) {
                    if (isArrayLikeType(t)) {
                        return getIndexedAccessType(t, getNumberLiteralType(childIndex));
                    }
                    else {
                        return t;
                    }
                }, /*noReductions*/ true));
            }
            function getContextualTypeForJsxExpression(node) {
                var exprParent = node.parent;
                return ts.isJsxAttributeLike(exprParent)
                    ? getContextualType(node)
                    : ts.isJsxElement(exprParent)
                        ? getContextualTypeForChildJsxExpression(exprParent, node)
                        : undefined;
            }
            function getContextualTypeForJsxAttribute(attribute) {
                // When we trying to resolve JsxOpeningLikeElement as a stateless function element, we will already give its attributes a contextual type
                // which is a type of the parameter of the signature we are trying out.
                // If there is no contextual type (e.g. we are trying to resolve stateful component), get attributes type from resolving element's tagName
                if (ts.isJsxAttribute(attribute)) {
                    var attributesType = getApparentTypeOfContextualType(attribute.parent);
                    if (!attributesType || isTypeAny(attributesType)) {
                        return undefined;
                    }
                    return getTypeOfPropertyOfContextualType(attributesType, attribute.name.escapedText);
                }
                else {
                    return getContextualType(attribute.parent);
                }
            }
            // Return true if the given expression is possibly a discriminant value. We limit the kinds of
            // expressions we check to those that don't depend on their contextual type in order not to cause
            // recursive (and possibly infinite) invocations of getContextualType.
            function isPossiblyDiscriminantValue(node) {
                switch (node.kind) {
                    case 10 /* StringLiteral */:
                    case 8 /* NumericLiteral */:
                    case 9 /* BigIntLiteral */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                    case 110 /* TrueKeyword */:
                    case 95 /* FalseKeyword */:
                    case 104 /* NullKeyword */:
                    case 79 /* Identifier */:
                    case 151 /* UndefinedKeyword */:
                        return true;
                    case 204 /* PropertyAccessExpression */:
                    case 210 /* ParenthesizedExpression */:
                        return isPossiblyDiscriminantValue(node.expression);
                    case 286 /* JsxExpression */:
                        return !node.expression || isPossiblyDiscriminantValue(node.expression);
                }
                return false;
            }
            function discriminateContextualTypeByObjectMembers(node, contextualType) {
                return getMatchingUnionConstituentForObjectLiteral(contextualType, node) || discriminateTypeByDiscriminableItems(contextualType, ts.concatenate(ts.map(ts.filter(node.properties, function (p) { return !!p.symbol && p.kind === 291 /* PropertyAssignment */ && isPossiblyDiscriminantValue(p.initializer) && isDiscriminantProperty(contextualType, p.symbol.escapedName); }), function (prop) { return [function () { return getContextFreeTypeOfExpression(prop.initializer); }, prop.symbol.escapedName]; }), ts.map(ts.filter(getPropertiesOfType(contextualType), function (s) { var _a; return !!(s.flags & 16777216 /* Optional */) && !!((_a = node === null || node === void 0 ? void 0 : node.symbol) === null || _a === void 0 ? void 0 : _a.members) && !node.symbol.members.has(s.escapedName) && isDiscriminantProperty(contextualType, s.escapedName); }), function (s) { return [function () { return undefinedType; }, s.escapedName]; })), isTypeAssignableTo, contextualType);
            }
            function discriminateContextualTypeByJSXAttributes(node, contextualType) {
                return discriminateTypeByDiscriminableItems(contextualType, ts.concatenate(ts.map(ts.filter(node.properties, function (p) { return !!p.symbol && p.kind === 283 /* JsxAttribute */ && isDiscriminantProperty(contextualType, p.symbol.escapedName) && (!p.initializer || isPossiblyDiscriminantValue(p.initializer)); }), function (prop) { return [!prop.initializer ? (function () { return trueType; }) : (function () { return checkExpression(prop.initializer); }), prop.symbol.escapedName]; }), ts.map(ts.filter(getPropertiesOfType(contextualType), function (s) { var _a; return !!(s.flags & 16777216 /* Optional */) && !!((_a = node === null || node === void 0 ? void 0 : node.symbol) === null || _a === void 0 ? void 0 : _a.members) && !node.symbol.members.has(s.escapedName) && isDiscriminantProperty(contextualType, s.escapedName); }), function (s) { return [function () { return undefinedType; }, s.escapedName]; })), isTypeAssignableTo, contextualType);
            }
            // Return the contextual type for a given expression node. During overload resolution, a contextual type may temporarily
            // be "pushed" onto a node using the contextualType property.
            function getApparentTypeOfContextualType(node, contextFlags) {
                var contextualType = ts.isObjectLiteralMethod(node) ?
                    getContextualTypeForObjectLiteralMethod(node, contextFlags) :
                    getContextualType(node, contextFlags);
                var instantiatedType = instantiateContextualType(contextualType, node, contextFlags);
                if (instantiatedType && !(contextFlags && contextFlags & 2 /* NoConstraints */ && instantiatedType.flags & 8650752 /* TypeVariable */)) {
                    var apparentType = mapType(instantiatedType, getApparentType, /*noReductions*/ true);
                    return apparentType.flags & 1048576 /* Union */ && ts.isObjectLiteralExpression(node) ? discriminateContextualTypeByObjectMembers(node, apparentType) :
                        apparentType.flags & 1048576 /* Union */ && ts.isJsxAttributes(node) ? discriminateContextualTypeByJSXAttributes(node, apparentType) :
                            apparentType;
                }
            }
            // If the given contextual type contains instantiable types and if a mapper representing
            // return type inferences is available, instantiate those types using that mapper.
            function instantiateContextualType(contextualType, node, contextFlags) {
                if (contextualType && maybeTypeOfKind(contextualType, 465829888 /* Instantiable */)) {
                    var inferenceContext = getInferenceContext(node);
                    // If no inferences have been made, nothing is gained from instantiating as type parameters
                    // would just be replaced with their defaults similar to the apparent type.
                    if (inferenceContext && ts.some(inferenceContext.inferences, hasInferenceCandidates)) {
                        // For contextual signatures we incorporate all inferences made so far, e.g. from return
                        // types as well as arguments to the left in a function call.
                        if (contextFlags && contextFlags & 1 /* Signature */) {
                            return instantiateInstantiableTypes(contextualType, inferenceContext.nonFixingMapper);
                        }
                        // For other purposes (e.g. determining whether to produce literal types) we only
                        // incorporate inferences made from the return type in a function call.
                        if (inferenceContext.returnMapper) {
                            return instantiateInstantiableTypes(contextualType, inferenceContext.returnMapper);
                        }
                    }
                }
                return contextualType;
            }
            // This function is similar to instantiateType, except that (a) it only instantiates types that
            // are classified as instantiable (i.e. it doesn't instantiate object types), and (b) it performs
            // no reductions on instantiated union types.
            function instantiateInstantiableTypes(type, mapper) {
                if (type.flags & 465829888 /* Instantiable */) {
                    return instantiateType(type, mapper);
                }
                if (type.flags & 1048576 /* Union */) {
                    return getUnionType(ts.map(type.types, function (t) { return instantiateInstantiableTypes(t, mapper); }), 0 /* None */);
                }
                if (type.flags & 2097152 /* Intersection */) {
                    return getIntersectionType(ts.map(type.types, function (t) { return instantiateInstantiableTypes(t, mapper); }));
                }
                return type;
            }
            /**
             * Whoa! Do you really want to use this function?
             *
             * Unless you're trying to get the *non-apparent* type for a
             * value-literal type or you're authoring relevant portions of this algorithm,
             * you probably meant to use 'getApparentTypeOfContextualType'.
             * Otherwise this may not be very useful.
             *
             * In cases where you *are* working on this function, you should understand
             * when it is appropriate to use 'getContextualType' and 'getApparentTypeOfContextualType'.
             *
             *   - Use 'getContextualType' when you are simply going to propagate the result to the expression.
             *   - Use 'getApparentTypeOfContextualType' when you're going to need the members of the type.
             *
             * @param node the expression whose contextual type will be returned.
             * @returns the contextual type of an expression.
             */
            function getContextualType(node, contextFlags) {
                if (node.flags & 16777216 /* InWithStatement */) {
                    // We cannot answer semantic questions within a with block, do not proceed any further
                    return undefined;
                }
                if (node.contextualType) {
                    return node.contextualType;
                }
                var parent = node.parent;
                switch (parent.kind) {
                    case 252 /* VariableDeclaration */:
                    case 162 /* Parameter */:
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                    case 201 /* BindingElement */:
                        return getContextualTypeForInitializerExpression(node, contextFlags);
                    case 212 /* ArrowFunction */:
                    case 245 /* ReturnStatement */:
                        return getContextualTypeForReturnExpression(node);
                    case 222 /* YieldExpression */:
                        return getContextualTypeForYieldOperand(parent);
                    case 216 /* AwaitExpression */:
                        return getContextualTypeForAwaitOperand(parent, contextFlags);
                    case 206 /* CallExpression */:
                        if (parent.expression.kind === 100 /* ImportKeyword */) {
                            return stringType;
                        }
                    /* falls through */
                    case 207 /* NewExpression */:
                        return getContextualTypeForArgument(parent, node);
                    case 209 /* TypeAssertionExpression */:
                    case 227 /* AsExpression */:
                        return ts.isConstTypeReference(parent.type) ? tryFindWhenConstTypeReference(parent) : getTypeFromTypeNode(parent.type);
                    case 219 /* BinaryExpression */:
                        return getContextualTypeForBinaryOperand(node, contextFlags);
                    case 291 /* PropertyAssignment */:
                    case 292 /* ShorthandPropertyAssignment */:
                        return getContextualTypeForObjectLiteralElement(parent, contextFlags);
                    case 293 /* SpreadAssignment */:
                        return getContextualType(parent.parent, contextFlags);
                    case 202 /* ArrayLiteralExpression */: {
                        var arrayLiteral = parent;
                        var type = getApparentTypeOfContextualType(arrayLiteral, contextFlags);
                        return getContextualTypeForElementExpression(type, ts.indexOfNode(arrayLiteral.elements, node));
                    }
                    case 220 /* ConditionalExpression */:
                        return getContextualTypeForConditionalOperand(node, contextFlags);
                    case 231 /* TemplateSpan */:
                        ts.Debug.assert(parent.parent.kind === 221 /* TemplateExpression */);
                        return getContextualTypeForSubstitutionExpression(parent.parent, node);
                    case 210 /* ParenthesizedExpression */: {
                        // Like in `checkParenthesizedExpression`, an `/** @type {xyz} */` comment before a parenthesized expression acts as a type cast.
                        var tag = ts.isInJSFile(parent) ? ts.getJSDocTypeTag(parent) : undefined;
                        return tag ? getTypeFromTypeNode(tag.typeExpression.type) : getContextualType(parent, contextFlags);
                    }
                    case 228 /* NonNullExpression */:
                        return getContextualType(parent, contextFlags);
                    case 286 /* JsxExpression */:
                        return getContextualTypeForJsxExpression(parent);
                    case 283 /* JsxAttribute */:
                    case 285 /* JsxSpreadAttribute */:
                        return getContextualTypeForJsxAttribute(parent);
                    case 278 /* JsxOpeningElement */:
                    case 277 /* JsxSelfClosingElement */:
                        return getContextualJsxElementAttributesType(parent, contextFlags);
                }
                return undefined;
                function tryFindWhenConstTypeReference(node) {
                    return getContextualType(node);
                }
            }
            function getInferenceContext(node) {
                var ancestor = ts.findAncestor(node, function (n) { return !!n.inferenceContext; });
                return ancestor && ancestor.inferenceContext;
            }
            function getContextualJsxElementAttributesType(node, contextFlags) {
                if (ts.isJsxOpeningElement(node) && node.parent.contextualType && contextFlags !== 4 /* Completions */) {
                    // Contextually applied type is moved from attributes up to the outer jsx attributes so when walking up from the children they get hit
                    // _However_ to hit them from the _attributes_ we must look for them here; otherwise we'll used the declared type
                    // (as below) instead!
                    return node.parent.contextualType;
                }
                return getContextualTypeForArgumentAtIndex(node, 0);
            }
            function getEffectiveFirstArgumentForJsxSignature(signature, node) {
                return getJsxReferenceKind(node) !== 0 /* Component */
                    ? getJsxPropsTypeFromCallSignature(signature, node)
                    : getJsxPropsTypeFromClassType(signature, node);
            }
            function getJsxPropsTypeFromCallSignature(sig, context) {
                var propsType = getTypeOfFirstParameterOfSignatureWithFallback(sig, unknownType);
                propsType = getJsxManagedAttributesFromLocatedAttributes(context, getJsxNamespaceAt(context), propsType);
                var intrinsicAttribs = getJsxType(JsxNames.IntrinsicAttributes, context);
                if (intrinsicAttribs !== errorType) {
                    propsType = intersectTypes(intrinsicAttribs, propsType);
                }
                return propsType;
            }
            function getJsxPropsTypeForSignatureFromMember(sig, forcedLookupLocation) {
                if (sig.compositeSignatures) {
                    // JSX Elements using the legacy `props`-field based lookup (eg, react class components) need to treat the `props` member as an input
                    // instead of an output position when resolving the signature. We need to go back to the input signatures of the composite signature,
                    // get the type of `props` on each return type individually, and then _intersect them_, rather than union them (as would normally occur
                    // for a union signature). It's an unfortunate quirk of looking in the output of the signature for the type we want to use for the input.
                    // The default behavior of `getTypeOfFirstParameterOfSignatureWithFallback` when no `props` member name is defined is much more sane.
                    var results = [];
                    for (var _i = 0, _a = sig.compositeSignatures; _i < _a.length; _i++) {
                        var signature = _a[_i];
                        var instance = getReturnTypeOfSignature(signature);
                        if (isTypeAny(instance)) {
                            return instance;
                        }
                        var propType = getTypeOfPropertyOfType(instance, forcedLookupLocation);
                        if (!propType) {
                            return;
                        }
                        results.push(propType);
                    }
                    return getIntersectionType(results); // Same result for both union and intersection signatures
                }
                var instanceType = getReturnTypeOfSignature(sig);
                return isTypeAny(instanceType) ? instanceType : getTypeOfPropertyOfType(instanceType, forcedLookupLocation);
            }
            function getStaticTypeOfReferencedJsxConstructor(context) {
                if (isJsxIntrinsicIdentifier(context.tagName)) {
                    var result = getIntrinsicAttributesTypeFromJsxOpeningLikeElement(context);
                    var fakeSignature = createSignatureForJSXIntrinsic(context, result);
                    return getOrCreateTypeFromSignature(fakeSignature);
                }
                var tagType = checkExpressionCached(context.tagName);
                if (tagType.flags & 128 /* StringLiteral */) {
                    var result = getIntrinsicAttributesTypeFromStringLiteralType(tagType, context);
                    if (!result) {
                        return errorType;
                    }
                    var fakeSignature = createSignatureForJSXIntrinsic(context, result);
                    return getOrCreateTypeFromSignature(fakeSignature);
                }
                return tagType;
            }
            function getJsxManagedAttributesFromLocatedAttributes(context, ns, attributesType) {
                var managedSym = getJsxLibraryManagedAttributes(ns);
                if (managedSym) {
                    var declaredManagedType = getDeclaredTypeOfSymbol(managedSym); // fetches interface type, or initializes symbol links type parmaeters
                    var ctorType = getStaticTypeOfReferencedJsxConstructor(context);
                    if (managedSym.flags & 524288 /* TypeAlias */) {
                        var params = getSymbolLinks(managedSym).typeParameters;
                        if (ts.length(params) >= 2) {
                            var args = fillMissingTypeArguments([ctorType, attributesType], params, 2, ts.isInJSFile(context));
                            return getTypeAliasInstantiation(managedSym, args);
                        }
                    }
                    if (ts.length(declaredManagedType.typeParameters) >= 2) {
                        var args = fillMissingTypeArguments([ctorType, attributesType], declaredManagedType.typeParameters, 2, ts.isInJSFile(context));
                        return createTypeReference(declaredManagedType, args);
                    }
                }
                return attributesType;
            }
            function getJsxPropsTypeFromClassType(sig, context) {
                var ns = getJsxNamespaceAt(context);
                var forcedLookupLocation = getJsxElementPropertiesName(ns);
                var attributesType = forcedLookupLocation === undefined
                    // If there is no type ElementAttributesProperty, return the type of the first parameter of the signature, which should be the props type
                    ? getTypeOfFirstParameterOfSignatureWithFallback(sig, unknownType)
                    : forcedLookupLocation === ""
                        // If there is no e.g. 'props' member in ElementAttributesProperty, use the element class type instead
                        ? getReturnTypeOfSignature(sig)
                        // Otherwise get the type of the property on the signature return type
                        : getJsxPropsTypeForSignatureFromMember(sig, forcedLookupLocation);
                if (!attributesType) {
                    // There is no property named 'props' on this instance type
                    if (!!forcedLookupLocation && !!ts.length(context.attributes.properties)) {
                        error(context, ts.Diagnostics.JSX_element_class_does_not_support_attributes_because_it_does_not_have_a_0_property, ts.unescapeLeadingUnderscores(forcedLookupLocation));
                    }
                    return unknownType;
                }
                attributesType = getJsxManagedAttributesFromLocatedAttributes(context, ns, attributesType);
                if (isTypeAny(attributesType)) {
                    // Props is of type 'any' or unknown
                    return attributesType;
                }
                else {
                    // Normal case -- add in IntrinsicClassElements<T> and IntrinsicElements
                    var apparentAttributesType = attributesType;
                    var intrinsicClassAttribs = getJsxType(JsxNames.IntrinsicClassAttributes, context);
                    if (intrinsicClassAttribs !== errorType) {
                        var typeParams = getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(intrinsicClassAttribs.symbol);
                        var hostClassType = getReturnTypeOfSignature(sig);
                        apparentAttributesType = intersectTypes(typeParams
                            ? createTypeReference(intrinsicClassAttribs, fillMissingTypeArguments([hostClassType], typeParams, getMinTypeArgumentCount(typeParams), ts.isInJSFile(context)))
                            : intrinsicClassAttribs, apparentAttributesType);
                    }
                    var intrinsicAttribs = getJsxType(JsxNames.IntrinsicAttributes, context);
                    if (intrinsicAttribs !== errorType) {
                        apparentAttributesType = intersectTypes(intrinsicAttribs, apparentAttributesType);
                    }
                    return apparentAttributesType;
                }
            }
            function getIntersectedSignatures(signatures) {
                return ts.getStrictOptionValue(compilerOptions, "noImplicitAny")
                    ? ts.reduceLeft(signatures, function (left, right) {
                        return left === right || !left ? left
                            : compareTypeParametersIdentical(left.typeParameters, right.typeParameters) ? combineSignaturesOfIntersectionMembers(left, right)
                                : undefined;
                    })
                    : undefined;
            }
            function combineIntersectionThisParam(left, right, mapper) {
                if (!left || !right) {
                    return left || right;
                }
                // A signature `this` type might be a read or a write position... It's very possible that it should be invariant
                // and we should refuse to merge signatures if there are `this` types and they do not match. However, so as to be
                // pessimistic when contextual typing, for now, we'll union the `this` types.
                var thisType = getUnionType([getTypeOfSymbol(left), instantiateType(getTypeOfSymbol(right), mapper)]);
                return createSymbolWithType(left, thisType);
            }
            function combineIntersectionParameters(left, right, mapper) {
                var leftCount = getParameterCount(left);
                var rightCount = getParameterCount(right);
                var longest = leftCount >= rightCount ? left : right;
                var shorter = longest === left ? right : left;
                var longestCount = longest === left ? leftCount : rightCount;
                var eitherHasEffectiveRest = (hasEffectiveRestParameter(left) || hasEffectiveRestParameter(right));
                var needsExtraRestElement = eitherHasEffectiveRest && !hasEffectiveRestParameter(longest);
                var params = new Array(longestCount + (needsExtraRestElement ? 1 : 0));
                for (var i = 0; i < longestCount; i++) {
                    var longestParamType = tryGetTypeAtPosition(longest, i);
                    if (longest === right) {
                        longestParamType = instantiateType(longestParamType, mapper);
                    }
                    var shorterParamType = tryGetTypeAtPosition(shorter, i) || unknownType;
                    if (shorter === right) {
                        shorterParamType = instantiateType(shorterParamType, mapper);
                    }
                    var unionParamType = getUnionType([longestParamType, shorterParamType]);
                    var isRestParam = eitherHasEffectiveRest && !needsExtraRestElement && i === (longestCount - 1);
                    var isOptional = i >= getMinArgumentCount(longest) && i >= getMinArgumentCount(shorter);
                    var leftName = i >= leftCount ? undefined : getParameterNameAtPosition(left, i);
                    var rightName = i >= rightCount ? undefined : getParameterNameAtPosition(right, i);
                    var paramName = leftName === rightName ? leftName :
                        !leftName ? rightName :
                            !rightName ? leftName :
                                undefined;
                    var paramSymbol = createSymbol(1 /* FunctionScopedVariable */ | (isOptional && !isRestParam ? 16777216 /* Optional */ : 0), paramName || "arg" + i);
                    paramSymbol.type = isRestParam ? createArrayType(unionParamType) : unionParamType;
                    params[i] = paramSymbol;
                }
                if (needsExtraRestElement) {
                    var restParamSymbol = createSymbol(1 /* FunctionScopedVariable */, "args");
                    restParamSymbol.type = createArrayType(getTypeAtPosition(shorter, longestCount));
                    if (shorter === right) {
                        restParamSymbol.type = instantiateType(restParamSymbol.type, mapper);
                    }
                    params[longestCount] = restParamSymbol;
                }
                return params;
            }
            function combineSignaturesOfIntersectionMembers(left, right) {
                var typeParams = left.typeParameters || right.typeParameters;
                var paramMapper;
                if (left.typeParameters && right.typeParameters) {
                    paramMapper = createTypeMapper(right.typeParameters, left.typeParameters);
                    // We just use the type parameter defaults from the first signature
                }
                var declaration = left.declaration;
                var params = combineIntersectionParameters(left, right, paramMapper);
                var thisParam = combineIntersectionThisParam(left.thisParameter, right.thisParameter, paramMapper);
                var minArgCount = Math.max(left.minArgumentCount, right.minArgumentCount);
                var result = createSignature(declaration, typeParams, thisParam, params, 
                /*resolvedReturnType*/ undefined, 
                /*resolvedTypePredicate*/ undefined, minArgCount, (left.flags | right.flags) & 39 /* PropagatingFlags */);
                result.compositeKind = 2097152 /* Intersection */;
                result.compositeSignatures = ts.concatenate(left.compositeKind === 2097152 /* Intersection */ && left.compositeSignatures || [left], [right]);
                if (paramMapper) {
                    result.mapper = left.compositeKind === 2097152 /* Intersection */ && left.mapper && left.compositeSignatures ? combineTypeMappers(left.mapper, paramMapper) : paramMapper;
                }
                return result;
            }
            // If the given type is an object or union type with a single signature, and if that signature has at
            // least as many parameters as the given function, return the signature. Otherwise return undefined.
            function getContextualCallSignature(type, node) {
                var signatures = getSignaturesOfType(type, 0 /* Call */);
                var applicableByArity = ts.filter(signatures, function (s) { return !isAritySmaller(s, node); });
                return applicableByArity.length === 1 ? applicableByArity[0] : getIntersectedSignatures(applicableByArity);
            }
            /** If the contextual signature has fewer parameters than the function expression, do not use it */
            function isAritySmaller(signature, target) {
                var targetParameterCount = 0;
                for (; targetParameterCount < target.parameters.length; targetParameterCount++) {
                    var param = target.parameters[targetParameterCount];
                    if (param.initializer || param.questionToken || param.dotDotDotToken || isJSDocOptionalParameter(param)) {
                        break;
                    }
                }
                if (target.parameters.length && ts.parameterIsThisKeyword(target.parameters[0])) {
                    targetParameterCount--;
                }
                return !hasEffectiveRestParameter(signature) && getParameterCount(signature) < targetParameterCount;
            }
            function isFunctionExpressionOrArrowFunction(node) {
                return node.kind === 211 /* FunctionExpression */ || node.kind === 212 /* ArrowFunction */;
            }
            function getContextualSignatureForFunctionLikeDeclaration(node) {
                // Only function expressions, arrow functions, and object literal methods are contextually typed.
                return isFunctionExpressionOrArrowFunction(node) || ts.isObjectLiteralMethod(node)
                    ? getContextualSignature(node)
                    : undefined;
            }
            // Return the contextual signature for a given expression node. A contextual type provides a
            // contextual signature if it has a single call signature and if that call signature is non-generic.
            // If the contextual type is a union type, get the signature from each type possible and if they are
            // all identical ignoring their return type, the result is same signature but with return type as
            // union type of return types from these signatures
            function getContextualSignature(node) {
                ts.Debug.assert(node.kind !== 167 /* MethodDeclaration */ || ts.isObjectLiteralMethod(node));
                var typeTagSignature = getSignatureOfTypeTag(node);
                if (typeTagSignature) {
                    return typeTagSignature;
                }
                var type = getApparentTypeOfContextualType(node, 1 /* Signature */);
                if (!type) {
                    return undefined;
                }
                if (!(type.flags & 1048576 /* Union */)) {
                    return getContextualCallSignature(type, node);
                }
                var signatureList;
                var types = type.types;
                for (var _i = 0, types_19 = types; _i < types_19.length; _i++) {
                    var current = types_19[_i];
                    var signature = getContextualCallSignature(current, node);
                    if (signature) {
                        if (!signatureList) {
                            // This signature will contribute to contextual union signature
                            signatureList = [signature];
                        }
                        else if (!compareSignaturesIdentical(signatureList[0], signature, /*partialMatch*/ false, /*ignoreThisTypes*/ true, /*ignoreReturnTypes*/ true, compareTypesIdentical)) {
                            // Signatures aren't identical, do not use
                            return undefined;
                        }
                        else {
                            // Use this signature for contextual union signature
                            signatureList.push(signature);
                        }
                    }
                }
                // Result is union of signatures collected (return type is union of return types of this signature set)
                if (signatureList) {
                    return signatureList.length === 1 ? signatureList[0] : createUnionSignature(signatureList[0], signatureList);
                }
            }
            function checkSpreadExpression(node, checkMode) {
                if (languageVersion < 2 /* ES2015 */) {
                    checkExternalEmitHelpers(node, compilerOptions.downlevelIteration ? 1536 /* SpreadIncludes */ : 1024 /* SpreadArray */);
                }
                var arrayOrIterableType = checkExpression(node.expression, checkMode);
                return checkIteratedTypeOrElementType(33 /* Spread */, arrayOrIterableType, undefinedType, node.expression);
            }
            function checkSyntheticExpression(node) {
                return node.isSpread ? getIndexedAccessType(node.type, numberType) : node.type;
            }
            function hasDefaultValue(node) {
                return (node.kind === 201 /* BindingElement */ && !!node.initializer) ||
                    (node.kind === 219 /* BinaryExpression */ && node.operatorToken.kind === 63 /* EqualsToken */);
            }
            function checkArrayLiteral(node, checkMode, forceTuple) {
                var elements = node.elements;
                var elementCount = elements.length;
                var elementTypes = [];
                var elementFlags = [];
                var contextualType = getApparentTypeOfContextualType(node);
                var inDestructuringPattern = ts.isAssignmentTarget(node);
                var inConstContext = isConstContext(node);
                var hasOmittedExpression = false;
                for (var i = 0; i < elementCount; i++) {
                    var e = elements[i];
                    if (e.kind === 223 /* SpreadElement */) {
                        if (languageVersion < 2 /* ES2015 */) {
                            checkExternalEmitHelpers(e, compilerOptions.downlevelIteration ? 1536 /* SpreadIncludes */ : 1024 /* SpreadArray */);
                        }
                        var spreadType = checkExpression(e.expression, checkMode, forceTuple);
                        if (isArrayLikeType(spreadType)) {
                            elementTypes.push(spreadType);
                            elementFlags.push(8 /* Variadic */);
                        }
                        else if (inDestructuringPattern) {
                            // Given the following situation:
                            //    var c: {};
                            //    [...c] = ["", 0];
                            //
                            // c is represented in the tree as a spread element in an array literal.
                            // But c really functions as a rest element, and its purpose is to provide
                            // a contextual type for the right hand side of the assignment. Therefore,
                            // instead of calling checkExpression on "...c", which will give an error
                            // if c is not iterable/array-like, we need to act as if we are trying to
                            // get the contextual element type from it. So we do something similar to
                            // getContextualTypeForElementExpression, which will crucially not error
                            // if there is no index type / iterated type.
                            var restElementType = getIndexTypeOfType(spreadType, numberType) ||
                                getIteratedTypeOrElementType(65 /* Destructuring */, spreadType, undefinedType, /*errorNode*/ undefined, /*checkAssignability*/ false) ||
                                unknownType;
                            elementTypes.push(restElementType);
                            elementFlags.push(4 /* Rest */);
                        }
                        else {
                            elementTypes.push(checkIteratedTypeOrElementType(33 /* Spread */, spreadType, undefinedType, e.expression));
                            elementFlags.push(4 /* Rest */);
                        }
                    }
                    else if (exactOptionalPropertyTypes && e.kind === 225 /* OmittedExpression */) {
                        hasOmittedExpression = true;
                        elementTypes.push(missingType);
                        elementFlags.push(2 /* Optional */);
                    }
                    else {
                        var elementContextualType = getContextualTypeForElementExpression(contextualType, elementTypes.length);
                        var type = checkExpressionForMutableLocation(e, checkMode, elementContextualType, forceTuple);
                        elementTypes.push(addOptionality(type, /*isProperty*/ true, hasOmittedExpression));
                        elementFlags.push(hasOmittedExpression ? 2 /* Optional */ : 1 /* Required */);
                    }
                }
                if (inDestructuringPattern) {
                    return createTupleType(elementTypes, elementFlags);
                }
                if (forceTuple || inConstContext || contextualType && someType(contextualType, isTupleLikeType)) {
                    return createArrayLiteralType(createTupleType(elementTypes, elementFlags, /*readonly*/ inConstContext));
                }
                return createArrayLiteralType(createArrayType(elementTypes.length ?
                    getUnionType(ts.sameMap(elementTypes, function (t, i) { return elementFlags[i] & 8 /* Variadic */ ? getIndexedAccessTypeOrUndefined(t, numberType) || anyType : t; }), 2 /* Subtype */) :
                    strictNullChecks ? implicitNeverType : undefinedWideningType, inConstContext));
            }
            function createArrayLiteralType(type) {
                if (!(ts.getObjectFlags(type) & 4 /* Reference */)) {
                    return type;
                }
                var literalType = type.literalType;
                if (!literalType) {
                    literalType = type.literalType = cloneTypeReference(type);
                    literalType.objectFlags |= 32768 /* ArrayLiteral */ | 262144 /* ContainsObjectOrArrayLiteral */;
                }
                return literalType;
            }
            function isNumericName(name) {
                switch (name.kind) {
                    case 160 /* ComputedPropertyName */:
                        return isNumericComputedName(name);
                    case 79 /* Identifier */:
                        return isNumericLiteralName(name.escapedText);
                    case 8 /* NumericLiteral */:
                    case 10 /* StringLiteral */:
                        return isNumericLiteralName(name.text);
                    default:
                        return false;
                }
            }
            function isNumericComputedName(name) {
                // It seems odd to consider an expression of type Any to result in a numeric name,
                // but this behavior is consistent with checkIndexedAccess
                return isTypeAssignableToKind(checkComputedPropertyName(name), 296 /* NumberLike */);
            }
            function isInfinityOrNaNString(name) {
                return name === "Infinity" || name === "-Infinity" || name === "NaN";
            }
            function isNumericLiteralName(name) {
                // The intent of numeric names is that
                //     - they are names with text in a numeric form, and that
                //     - setting properties/indexing with them is always equivalent to doing so with the numeric literal 'numLit',
                //         acquired by applying the abstract 'ToNumber' operation on the name's text.
                //
                // The subtlety is in the latter portion, as we cannot reliably say that anything that looks like a numeric literal is a numeric name.
                // In fact, it is the case that the text of the name must be equal to 'ToString(numLit)' for this to hold.
                //
                // Consider the property name '"0xF00D"'. When one indexes with '0xF00D', they are actually indexing with the value of 'ToString(0xF00D)'
                // according to the ECMAScript specification, so it is actually as if the user indexed with the string '"61453"'.
                // Thus, the text of all numeric literals equivalent to '61543' such as '0xF00D', '0xf00D', '0170015', etc. are not valid numeric names
                // because their 'ToString' representation is not equal to their original text.
                // This is motivated by ECMA-262 sections 9.3.1, 9.8.1, 11.1.5, and 11.2.1.
                //
                // Here, we test whether 'ToString(ToNumber(name))' is exactly equal to 'name'.
                // The '+' prefix operator is equivalent here to applying the abstract ToNumber operation.
                // Applying the 'toString()' method on a number gives us the abstract ToString operation on a number.
                //
                // Note that this accepts the values 'Infinity', '-Infinity', and 'NaN', and that this is intentional.
                // This is desired behavior, because when indexing with them as numeric entities, you are indexing
                // with the strings '"Infinity"', '"-Infinity"', and '"NaN"' respectively.
                return (+name).toString() === name;
            }
            function checkComputedPropertyName(node) {
                var links = getNodeLinks(node.expression);
                if (!links.resolvedType) {
                    links.resolvedType = checkExpression(node.expression);
                    // The computed property name of a non-static class field within a loop must be stored in a block-scoped binding.
                    // (It needs to be bound at class evaluation time.)
                    if (ts.isPropertyDeclaration(node.parent) && !ts.hasStaticModifier(node.parent) && ts.isClassExpression(node.parent.parent)) {
                        var container = ts.getEnclosingBlockScopeContainer(node.parent.parent);
                        var enclosingIterationStatement = getEnclosingIterationStatement(container);
                        if (enclosingIterationStatement) {
                            // The computed field name will use a block scoped binding which can be unique for each iteration of the loop.
                            getNodeLinks(enclosingIterationStatement).flags |= 65536 /* LoopWithCapturedBlockScopedBinding */;
                            // The generated variable which stores the computed field name must be block-scoped.
                            getNodeLinks(node).flags |= 524288 /* BlockScopedBindingInLoop */;
                            // The generated variable which stores the class must be block-scoped.
                            getNodeLinks(node.parent.parent).flags |= 524288 /* BlockScopedBindingInLoop */;
                        }
                    }
                    // This will allow types number, string, symbol or any. It will also allow enums, the unknown
                    // type, and any union of these types (like string | number).
                    if (links.resolvedType.flags & 98304 /* Nullable */ ||
                        !isTypeAssignableToKind(links.resolvedType, 402653316 /* StringLike */ | 296 /* NumberLike */ | 12288 /* ESSymbolLike */) &&
                            !isTypeAssignableTo(links.resolvedType, stringNumberSymbolType)) {
                        error(node, ts.Diagnostics.A_computed_property_name_must_be_of_type_string_number_symbol_or_any);
                    }
                }
                return links.resolvedType;
            }
            function isSymbolWithNumericName(symbol) {
                var _a;
                var firstDecl = (_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a[0];
                return isNumericLiteralName(symbol.escapedName) || (firstDecl && ts.isNamedDeclaration(firstDecl) && isNumericName(firstDecl.name));
            }
            function isSymbolWithSymbolName(symbol) {
                var _a;
                var firstDecl = (_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a[0];
                return ts.isKnownSymbol(symbol) || (firstDecl && ts.isNamedDeclaration(firstDecl) && ts.isComputedPropertyName(firstDecl.name) &&
                    isTypeAssignableToKind(checkComputedPropertyName(firstDecl.name), 4096 /* ESSymbol */));
            }
            function getObjectLiteralIndexInfo(node, offset, properties, keyType) {
                var propTypes = [];
                for (var i = offset; i < properties.length; i++) {
                    var prop = properties[i];
                    if (keyType === stringType && !isSymbolWithSymbolName(prop) ||
                        keyType === numberType && isSymbolWithNumericName(prop) ||
                        keyType === esSymbolType && isSymbolWithSymbolName(prop)) {
                        propTypes.push(getTypeOfSymbol(properties[i]));
                    }
                }
                var unionType = propTypes.length ? getUnionType(propTypes, 2 /* Subtype */) : undefinedType;
                return createIndexInfo(keyType, unionType, isConstContext(node));
            }
            function getImmediateAliasedSymbol(symbol) {
                ts.Debug.assert((symbol.flags & 2097152 /* Alias */) !== 0, "Should only get Alias here.");
                var links = getSymbolLinks(symbol);
                if (!links.immediateTarget) {
                    var node = getDeclarationOfAliasSymbol(symbol);
                    if (!node)
                        return ts.Debug.fail();
                    links.immediateTarget = getTargetOfAliasDeclaration(node, /*dontRecursivelyResolve*/ true);
                }
                return links.immediateTarget;
            }
            function checkObjectLiteral(node, checkMode) {
                var inDestructuringPattern = ts.isAssignmentTarget(node);
                // Grammar checking
                checkGrammarObjectLiteralExpression(node, inDestructuringPattern);
                var allPropertiesTable = strictNullChecks ? ts.createSymbolTable() : undefined;
                var propertiesTable = ts.createSymbolTable();
                var propertiesArray = [];
                var spread = emptyObjectType;
                var contextualType = getApparentTypeOfContextualType(node);
                var contextualTypeHasPattern = contextualType && contextualType.pattern &&
                    (contextualType.pattern.kind === 199 /* ObjectBindingPattern */ || contextualType.pattern.kind === 203 /* ObjectLiteralExpression */);
                var inConstContext = isConstContext(node);
                var checkFlags = inConstContext ? 8 /* Readonly */ : 0;
                var isInJavascript = ts.isInJSFile(node) && !ts.isInJsonFile(node);
                var enumTag = ts.getJSDocEnumTag(node);
                var isJSObjectLiteral = !contextualType && isInJavascript && !enumTag;
                var objectFlags = freshObjectLiteralFlag;
                var patternWithComputedProperties = false;
                var hasComputedStringProperty = false;
                var hasComputedNumberProperty = false;
                var hasComputedSymbolProperty = false;
                // Spreads may cause an early bail; ensure computed names are always checked (this is cached)
                // As otherwise they may not be checked until exports for the type at this position are retrieved,
                // which may never occur.
                for (var _i = 0, _a = node.properties; _i < _a.length; _i++) {
                    var elem = _a[_i];
                    if (elem.name && ts.isComputedPropertyName(elem.name)) {
                        checkComputedPropertyName(elem.name);
                    }
                }
                var offset = 0;
                for (var _b = 0, _c = node.properties; _b < _c.length; _b++) {
                    var memberDecl = _c[_b];
                    var member = getSymbolOfNode(memberDecl);
                    var computedNameType = memberDecl.name && memberDecl.name.kind === 160 /* ComputedPropertyName */ ?
                        checkComputedPropertyName(memberDecl.name) : undefined;
                    if (memberDecl.kind === 291 /* PropertyAssignment */ ||
                        memberDecl.kind === 292 /* ShorthandPropertyAssignment */ ||
                        ts.isObjectLiteralMethod(memberDecl)) {
                        var type = memberDecl.kind === 291 /* PropertyAssignment */ ? checkPropertyAssignment(memberDecl, checkMode) :
                            // avoid resolving the left side of the ShorthandPropertyAssignment outside of the destructuring
                            // for error recovery purposes. For example, if a user wrote `{ a = 100 }` instead of `{ a: 100 }`.
                            // we don't want to say "could not find 'a'".
                            memberDecl.kind === 292 /* ShorthandPropertyAssignment */ ? checkExpressionForMutableLocation(!inDestructuringPattern && memberDecl.objectAssignmentInitializer ? memberDecl.objectAssignmentInitializer : memberDecl.name, checkMode) :
                                checkObjectLiteralMethod(memberDecl, checkMode);
                        if (isInJavascript) {
                            var jsDocType = getTypeForDeclarationFromJSDocComment(memberDecl);
                            if (jsDocType) {
                                checkTypeAssignableTo(type, jsDocType, memberDecl);
                                type = jsDocType;
                            }
                            else if (enumTag && enumTag.typeExpression) {
                                checkTypeAssignableTo(type, getTypeFromTypeNode(enumTag.typeExpression), memberDecl);
                            }
                        }
                        objectFlags |= ts.getObjectFlags(type) & 917504 /* PropagatingFlags */;
                        var nameType = computedNameType && isTypeUsableAsPropertyName(computedNameType) ? computedNameType : undefined;
                        var prop = nameType ?
                            createSymbol(4 /* Property */ | member.flags, getPropertyNameFromType(nameType), checkFlags | 4096 /* Late */) :
                            createSymbol(4 /* Property */ | member.flags, member.escapedName, checkFlags);
                        if (nameType) {
                            prop.nameType = nameType;
                        }
                        if (inDestructuringPattern) {
                            // If object literal is an assignment pattern and if the assignment pattern specifies a default value
                            // for the property, make the property optional.
                            var isOptional = (memberDecl.kind === 291 /* PropertyAssignment */ && hasDefaultValue(memberDecl.initializer)) ||
                                (memberDecl.kind === 292 /* ShorthandPropertyAssignment */ && memberDecl.objectAssignmentInitializer);
                            if (isOptional) {
                                prop.flags |= 16777216 /* Optional */;
                            }
                        }
                        else if (contextualTypeHasPattern && !(ts.getObjectFlags(contextualType) & 512 /* ObjectLiteralPatternWithComputedProperties */)) {
                            // If object literal is contextually typed by the implied type of a binding pattern, and if the
                            // binding pattern specifies a default value for the property, make the property optional.
                            var impliedProp = getPropertyOfType(contextualType, member.escapedName);
                            if (impliedProp) {
                                prop.flags |= impliedProp.flags & 16777216 /* Optional */;
                            }
                            else if (!compilerOptions.suppressExcessPropertyErrors && !getIndexInfoOfType(contextualType, stringType)) {
                                error(memberDecl.name, ts.Diagnostics.Object_literal_may_only_specify_known_properties_and_0_does_not_exist_in_type_1, symbolToString(member), typeToString(contextualType));
                            }
                        }
                        prop.declarations = member.declarations;
                        prop.parent = member.parent;
                        if (member.valueDeclaration) {
                            prop.valueDeclaration = member.valueDeclaration;
                        }
                        prop.type = type;
                        prop.target = member;
                        member = prop;
                        allPropertiesTable === null || allPropertiesTable === void 0 ? void 0 : allPropertiesTable.set(prop.escapedName, prop);
                    }
                    else if (memberDecl.kind === 293 /* SpreadAssignment */) {
                        if (languageVersion < 2 /* ES2015 */) {
                            checkExternalEmitHelpers(memberDecl, 2 /* Assign */);
                        }
                        if (propertiesArray.length > 0) {
                            spread = getSpreadType(spread, createObjectLiteralType(), node.symbol, objectFlags, inConstContext);
                            propertiesArray = [];
                            propertiesTable = ts.createSymbolTable();
                            hasComputedStringProperty = false;
                            hasComputedNumberProperty = false;
                            hasComputedSymbolProperty = false;
                        }
                        var type = getReducedType(checkExpression(memberDecl.expression));
                        if (isValidSpreadType(type)) {
                            var mergedType = tryMergeUnionOfObjectTypeAndEmptyObject(type, inConstContext);
                            if (allPropertiesTable) {
                                checkSpreadPropOverrides(mergedType, allPropertiesTable, memberDecl);
                            }
                            offset = propertiesArray.length;
                            if (spread === errorType) {
                                continue;
                            }
                            spread = getSpreadType(spread, mergedType, node.symbol, objectFlags, inConstContext);
                        }
                        else {
                            error(memberDecl, ts.Diagnostics.Spread_types_may_only_be_created_from_object_types);
                            spread = errorType;
                        }
                        continue;
                    }
                    else {
                        // TypeScript 1.0 spec (April 2014)
                        // A get accessor declaration is processed in the same manner as
                        // an ordinary function declaration(section 6.1) with no parameters.
                        // A set accessor declaration is processed in the same manner
                        // as an ordinary function declaration with a single parameter and a Void return type.
                        ts.Debug.assert(memberDecl.kind === 170 /* GetAccessor */ || memberDecl.kind === 171 /* SetAccessor */);
                        checkNodeDeferred(memberDecl);
                    }
                    if (computedNameType && !(computedNameType.flags & 8576 /* StringOrNumberLiteralOrUnique */)) {
                        if (isTypeAssignableTo(computedNameType, stringNumberSymbolType)) {
                            if (isTypeAssignableTo(computedNameType, numberType)) {
                                hasComputedNumberProperty = true;
                            }
                            else if (isTypeAssignableTo(computedNameType, esSymbolType)) {
                                hasComputedSymbolProperty = true;
                            }
                            else {
                                hasComputedStringProperty = true;
                            }
                            if (inDestructuringPattern) {
                                patternWithComputedProperties = true;
                            }
                        }
                    }
                    else {
                        propertiesTable.set(member.escapedName, member);
                    }
                    propertiesArray.push(member);
                }
                // If object literal is contextually typed by the implied type of a binding pattern, augment the result
                // type with those properties for which the binding pattern specifies a default value.
                // If the object literal is spread into another object literal, skip this step and let the top-level object
                // literal handle it instead.
                if (contextualTypeHasPattern && node.parent.kind !== 293 /* SpreadAssignment */) {
                    for (var _d = 0, _e = getPropertiesOfType(contextualType); _d < _e.length; _d++) {
                        var prop = _e[_d];
                        if (!propertiesTable.get(prop.escapedName) && !getPropertyOfType(spread, prop.escapedName)) {
                            if (!(prop.flags & 16777216 /* Optional */)) {
                                error(prop.valueDeclaration || prop.bindingElement, ts.Diagnostics.Initializer_provides_no_value_for_this_binding_element_and_the_binding_element_has_no_default_value);
                            }
                            propertiesTable.set(prop.escapedName, prop);
                            propertiesArray.push(prop);
                        }
                    }
                }
                if (spread === errorType) {
                    return errorType;
                }
                if (spread !== emptyObjectType) {
                    if (propertiesArray.length > 0) {
                        spread = getSpreadType(spread, createObjectLiteralType(), node.symbol, objectFlags, inConstContext);
                        propertiesArray = [];
                        propertiesTable = ts.createSymbolTable();
                        hasComputedStringProperty = false;
                        hasComputedNumberProperty = false;
                    }
                    // remap the raw emptyObjectType fed in at the top into a fresh empty object literal type, unique to this use site
                    return mapType(spread, function (t) { return t === emptyObjectType ? createObjectLiteralType() : t; });
                }
                return createObjectLiteralType();
                function createObjectLiteralType() {
                    var indexInfos = [];
                    if (hasComputedStringProperty)
                        indexInfos.push(getObjectLiteralIndexInfo(node, offset, propertiesArray, stringType));
                    if (hasComputedNumberProperty)
                        indexInfos.push(getObjectLiteralIndexInfo(node, offset, propertiesArray, numberType));
                    if (hasComputedSymbolProperty)
                        indexInfos.push(getObjectLiteralIndexInfo(node, offset, propertiesArray, esSymbolType));
                    var result = createAnonymousType(node.symbol, propertiesTable, ts.emptyArray, ts.emptyArray, indexInfos);
                    result.objectFlags |= objectFlags | 128 /* ObjectLiteral */ | 262144 /* ContainsObjectOrArrayLiteral */;
                    if (isJSObjectLiteral) {
                        result.objectFlags |= 8192 /* JSLiteral */;
                    }
                    if (patternWithComputedProperties) {
                        result.objectFlags |= 512 /* ObjectLiteralPatternWithComputedProperties */;
                    }
                    if (inDestructuringPattern) {
                        result.pattern = node;
                    }
                    return result;
                }
            }
            function isValidSpreadType(type) {
                if (type.flags & 465829888 /* Instantiable */) {
                    var constraint = getBaseConstraintOfType(type);
                    if (constraint !== undefined) {
                        return isValidSpreadType(constraint);
                    }
                }
                return !!(type.flags & (1 /* Any */ | 67108864 /* NonPrimitive */ | 524288 /* Object */ | 58982400 /* InstantiableNonPrimitive */) ||
                    getFalsyFlags(type) & 117632 /* DefinitelyFalsy */ && isValidSpreadType(removeDefinitelyFalsyTypes(type)) ||
                    type.flags & 3145728 /* UnionOrIntersection */ && ts.every(type.types, isValidSpreadType));
            }
            function checkJsxSelfClosingElementDeferred(node) {
                checkJsxOpeningLikeElementOrOpeningFragment(node);
            }
            function checkJsxSelfClosingElement(node, _checkMode) {
                checkNodeDeferred(node);
                return getJsxElementTypeAt(node) || anyType;
            }
            function checkJsxElementDeferred(node) {
                // Check attributes
                checkJsxOpeningLikeElementOrOpeningFragment(node.openingElement);
                // Perform resolution on the closing tag so that rename/go to definition/etc work
                if (isJsxIntrinsicIdentifier(node.closingElement.tagName)) {
                    getIntrinsicTagSymbol(node.closingElement);
                }
                else {
                    checkExpression(node.closingElement.tagName);
                }
                checkJsxChildren(node);
            }
            function checkJsxElement(node, _checkMode) {
                checkNodeDeferred(node);
                return getJsxElementTypeAt(node) || anyType;
            }
            function checkJsxFragment(node) {
                checkJsxOpeningLikeElementOrOpeningFragment(node.openingFragment);
                // by default, jsx:'react' will use jsxFactory = React.createElement and jsxFragmentFactory = React.Fragment
                // if jsxFactory compiler option is provided, ensure jsxFragmentFactory compiler option or @jsxFrag pragma is provided too
                var nodeSourceFile = ts.getSourceFileOfNode(node);
                if (ts.getJSXTransformEnabled(compilerOptions) && (compilerOptions.jsxFactory || nodeSourceFile.pragmas.has("jsx"))
                    && !compilerOptions.jsxFragmentFactory && !nodeSourceFile.pragmas.has("jsxfrag")) {
                    error(node, compilerOptions.jsxFactory
                        ? ts.Diagnostics.The_jsxFragmentFactory_compiler_option_must_be_provided_to_use_JSX_fragments_with_the_jsxFactory_compiler_option
                        : ts.Diagnostics.An_jsxFrag_pragma_is_required_when_using_an_jsx_pragma_with_JSX_fragments);
                }
                checkJsxChildren(node);
                return getJsxElementTypeAt(node) || anyType;
            }
            function isHyphenatedJsxName(name) {
                return ts.stringContains(name, "-");
            }
            /**
             * Returns true iff React would emit this tag name as a string rather than an identifier or qualified name
             */
            function isJsxIntrinsicIdentifier(tagName) {
                return tagName.kind === 79 /* Identifier */ && ts.isIntrinsicJsxName(tagName.escapedText);
            }
            function checkJsxAttribute(node, checkMode) {
                return node.initializer
                    ? checkExpressionForMutableLocation(node.initializer, checkMode)
                    : trueType; // <Elem attr /> is sugar for <Elem attr={true} />
            }
            /**
             * Get attributes type of the JSX opening-like element. The result is from resolving "attributes" property of the opening-like element.
             *
             * @param openingLikeElement a JSX opening-like element
             * @param filter a function to remove attributes that will not participate in checking whether attributes are assignable
             * @return an anonymous type (similar to the one returned by checkObjectLiteral) in which its properties are attributes property.
             * @remarks Because this function calls getSpreadType, it needs to use the same checks as checkObjectLiteral,
             * which also calls getSpreadType.
             */
            function createJsxAttributesTypeFromAttributesProperty(openingLikeElement, checkMode) {
                var attributes = openingLikeElement.attributes;
                var allAttributesTable = strictNullChecks ? ts.createSymbolTable() : undefined;
                var attributesTable = ts.createSymbolTable();
                var spread = emptyJsxObjectType;
                var hasSpreadAnyType = false;
                var typeToIntersect;
                var explicitlySpecifyChildrenAttribute = false;
                var objectFlags = 2048 /* JsxAttributes */;
                var jsxChildrenPropertyName = getJsxElementChildrenPropertyName(getJsxNamespaceAt(openingLikeElement));
                for (var _i = 0, _a = attributes.properties; _i < _a.length; _i++) {
                    var attributeDecl = _a[_i];
                    var member = attributeDecl.symbol;
                    if (ts.isJsxAttribute(attributeDecl)) {
                        var exprType = checkJsxAttribute(attributeDecl, checkMode);
                        objectFlags |= ts.getObjectFlags(exprType) & 917504 /* PropagatingFlags */;
                        var attributeSymbol = createSymbol(4 /* Property */ | member.flags, member.escapedName);
                        attributeSymbol.declarations = member.declarations;
                        attributeSymbol.parent = member.parent;
                        if (member.valueDeclaration) {
                            attributeSymbol.valueDeclaration = member.valueDeclaration;
                        }
                        attributeSymbol.type = exprType;
                        attributeSymbol.target = member;
                        attributesTable.set(attributeSymbol.escapedName, attributeSymbol);
                        allAttributesTable === null || allAttributesTable === void 0 ? void 0 : allAttributesTable.set(attributeSymbol.escapedName, attributeSymbol);
                        if (attributeDecl.name.escapedText === jsxChildrenPropertyName) {
                            explicitlySpecifyChildrenAttribute = true;
                        }
                    }
                    else {
                        ts.Debug.assert(attributeDecl.kind === 285 /* JsxSpreadAttribute */);
                        if (attributesTable.size > 0) {
                            spread = getSpreadType(spread, createJsxAttributesType(), attributes.symbol, objectFlags, /*readonly*/ false);
                            attributesTable = ts.createSymbolTable();
                        }
                        var exprType = getReducedType(checkExpressionCached(attributeDecl.expression, checkMode));
                        if (isTypeAny(exprType)) {
                            hasSpreadAnyType = true;
                        }
                        if (isValidSpreadType(exprType)) {
                            spread = getSpreadType(spread, exprType, attributes.symbol, objectFlags, /*readonly*/ false);
                            if (allAttributesTable) {
                                checkSpreadPropOverrides(exprType, allAttributesTable, attributeDecl);
                            }
                        }
                        else {
                            typeToIntersect = typeToIntersect ? getIntersectionType([typeToIntersect, exprType]) : exprType;
                        }
                    }
                }
                if (!hasSpreadAnyType) {
                    if (attributesTable.size > 0) {
                        spread = getSpreadType(spread, createJsxAttributesType(), attributes.symbol, objectFlags, /*readonly*/ false);
                    }
                }
                // Handle children attribute
                var parent = openingLikeElement.parent.kind === 276 /* JsxElement */ ? openingLikeElement.parent : undefined;
                // We have to check that openingElement of the parent is the one we are visiting as this may not be true for selfClosingElement
                if (parent && parent.openingElement === openingLikeElement && parent.children.length > 0) {
                    var childrenTypes = checkJsxChildren(parent, checkMode);
                    if (!hasSpreadAnyType && jsxChildrenPropertyName && jsxChildrenPropertyName !== "") {
                        // Error if there is a attribute named "children" explicitly specified and children element.
                        // This is because children element will overwrite the value from attributes.
                        // Note: we will not warn "children" attribute overwritten if "children" attribute is specified in object spread.
                        if (explicitlySpecifyChildrenAttribute) {
                            error(attributes, ts.Diagnostics._0_are_specified_twice_The_attribute_named_0_will_be_overwritten, ts.unescapeLeadingUnderscores(jsxChildrenPropertyName));
                        }
                        var contextualType = getApparentTypeOfContextualType(openingLikeElement.attributes);
                        var childrenContextualType = contextualType && getTypeOfPropertyOfContextualType(contextualType, jsxChildrenPropertyName);
                        // If there are children in the body of JSX element, create dummy attribute "children" with the union of children types so that it will pass the attribute checking process
                        var childrenPropSymbol = createSymbol(4 /* Property */, jsxChildrenPropertyName);
                        childrenPropSymbol.type = childrenTypes.length === 1 ? childrenTypes[0] :
                            childrenContextualType && someType(childrenContextualType, isTupleLikeType) ? createTupleType(childrenTypes) :
                                createArrayType(getUnionType(childrenTypes));
                        // Fake up a property declaration for the children
                        childrenPropSymbol.valueDeclaration = ts.factory.createPropertySignature(/*modifiers*/ undefined, ts.unescapeLeadingUnderscores(jsxChildrenPropertyName), /*questionToken*/ undefined, /*type*/ undefined);
                        ts.setParent(childrenPropSymbol.valueDeclaration, attributes);
                        childrenPropSymbol.valueDeclaration.symbol = childrenPropSymbol;
                        var childPropMap = ts.createSymbolTable();
                        childPropMap.set(jsxChildrenPropertyName, childrenPropSymbol);
                        spread = getSpreadType(spread, createAnonymousType(attributes.symbol, childPropMap, ts.emptyArray, ts.emptyArray, ts.emptyArray), attributes.symbol, objectFlags, /*readonly*/ false);
                    }
                }
                if (hasSpreadAnyType) {
                    return anyType;
                }
                if (typeToIntersect && spread !== emptyJsxObjectType) {
                    return getIntersectionType([typeToIntersect, spread]);
                }
                return typeToIntersect || (spread === emptyJsxObjectType ? createJsxAttributesType() : spread);
                /**
                 * Create anonymous type from given attributes symbol table.
                 * @param symbol a symbol of JsxAttributes containing attributes corresponding to attributesTable
                 * @param attributesTable a symbol table of attributes property
                 */
                function createJsxAttributesType() {
                    objectFlags |= freshObjectLiteralFlag;
                    var result = createAnonymousType(attributes.symbol, attributesTable, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                    result.objectFlags |= objectFlags | 128 /* ObjectLiteral */ | 262144 /* ContainsObjectOrArrayLiteral */;
                    return result;
                }
            }
            function checkJsxChildren(node, checkMode) {
                var childrenTypes = [];
                for (var _i = 0, _a = node.children; _i < _a.length; _i++) {
                    var child = _a[_i];
                    // In React, JSX text that contains only whitespaces will be ignored so we don't want to type-check that
                    // because then type of children property will have constituent of string type.
                    if (child.kind === 11 /* JsxText */) {
                        if (!child.containsOnlyTriviaWhiteSpaces) {
                            childrenTypes.push(stringType);
                        }
                    }
                    else if (child.kind === 286 /* JsxExpression */ && !child.expression) {
                        continue; // empty jsx expressions don't *really* count as present children
                    }
                    else {
                        childrenTypes.push(checkExpressionForMutableLocation(child, checkMode));
                    }
                }
                return childrenTypes;
            }
            function checkSpreadPropOverrides(type, props, spread) {
                for (var _i = 0, _a = getPropertiesOfType(type); _i < _a.length; _i++) {
                    var right = _a[_i];
                    if (!(right.flags & 16777216 /* Optional */)) {
                        var left = props.get(right.escapedName);
                        if (left) {
                            var diagnostic = error(left.valueDeclaration, ts.Diagnostics._0_is_specified_more_than_once_so_this_usage_will_be_overwritten, ts.unescapeLeadingUnderscores(left.escapedName));
                            ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(spread, ts.Diagnostics.This_spread_always_overwrites_this_property));
                        }
                    }
                }
            }
            /**
             * Check attributes property of opening-like element. This function is called during chooseOverload to get call signature of a JSX opening-like element.
             * (See "checkApplicableSignatureForJsxOpeningLikeElement" for how the function is used)
             * @param node a JSXAttributes to be resolved of its type
             */
            function checkJsxAttributes(node, checkMode) {
                return createJsxAttributesTypeFromAttributesProperty(node.parent, checkMode);
            }
            function getJsxType(name, location) {
                var namespace = getJsxNamespaceAt(location);
                var exports = namespace && getExportsOfSymbol(namespace);
                var typeSymbol = exports && getSymbol(exports, name, 788968 /* Type */);
                return typeSymbol ? getDeclaredTypeOfSymbol(typeSymbol) : errorType;
            }
            /**
             * Looks up an intrinsic tag name and returns a symbol that either points to an intrinsic
             * property (in which case nodeLinks.jsxFlags will be IntrinsicNamedElement) or an intrinsic
             * string index signature (in which case nodeLinks.jsxFlags will be IntrinsicIndexedElement).
             * May also return unknownSymbol if both of these lookups fail.
             */
            function getIntrinsicTagSymbol(node) {
                var links = getNodeLinks(node);
                if (!links.resolvedSymbol) {
                    var intrinsicElementsType = getJsxType(JsxNames.IntrinsicElements, node);
                    if (intrinsicElementsType !== errorType) {
                        // Property case
                        if (!ts.isIdentifier(node.tagName))
                            return ts.Debug.fail();
                        var intrinsicProp = getPropertyOfType(intrinsicElementsType, node.tagName.escapedText);
                        if (intrinsicProp) {
                            links.jsxFlags |= 1 /* IntrinsicNamedElement */;
                            return links.resolvedSymbol = intrinsicProp;
                        }
                        // Intrinsic string indexer case
                        var indexSignatureType = getIndexTypeOfType(intrinsicElementsType, stringType);
                        if (indexSignatureType) {
                            links.jsxFlags |= 2 /* IntrinsicIndexedElement */;
                            return links.resolvedSymbol = intrinsicElementsType.symbol;
                        }
                        // Wasn't found
                        error(node, ts.Diagnostics.Property_0_does_not_exist_on_type_1, ts.idText(node.tagName), "JSX." + JsxNames.IntrinsicElements);
                        return links.resolvedSymbol = unknownSymbol;
                    }
                    else {
                        if (noImplicitAny) {
                            error(node, ts.Diagnostics.JSX_element_implicitly_has_type_any_because_no_interface_JSX_0_exists, ts.unescapeLeadingUnderscores(JsxNames.IntrinsicElements));
                        }
                        return links.resolvedSymbol = unknownSymbol;
                    }
                }
                return links.resolvedSymbol;
            }
            function getJsxNamespaceContainerForImplicitImport(location) {
                var file = location && ts.getSourceFileOfNode(location);
                var links = file && getNodeLinks(file);
                if (links && links.jsxImplicitImportContainer === false) {
                    return undefined;
                }
                if (links && links.jsxImplicitImportContainer) {
                    return links.jsxImplicitImportContainer;
                }
                var runtimeImportSpecifier = ts.getJSXRuntimeImport(ts.getJSXImplicitImportBase(compilerOptions, file), compilerOptions);
                if (!runtimeImportSpecifier) {
                    return undefined;
                }
                var isClassic = ts.getEmitModuleResolutionKind(compilerOptions) === ts.ModuleResolutionKind.Classic;
                var errorMessage = isClassic
                    ? ts.Diagnostics.Cannot_find_module_0_Did_you_mean_to_set_the_moduleResolution_option_to_node_or_to_add_aliases_to_the_paths_option
                    : ts.Diagnostics.Cannot_find_module_0_or_its_corresponding_type_declarations;
                var mod = resolveExternalModule(location, runtimeImportSpecifier, errorMessage, location);
                var result = mod && mod !== unknownSymbol ? getMergedSymbol(resolveSymbol(mod)) : undefined;
                if (links) {
                    links.jsxImplicitImportContainer = result || false;
                }
                return result;
            }
            function getJsxNamespaceAt(location) {
                var links = location && getNodeLinks(location);
                if (links && links.jsxNamespace) {
                    return links.jsxNamespace;
                }
                if (!links || links.jsxNamespace !== false) {
                    var resolvedNamespace = getJsxNamespaceContainerForImplicitImport(location);
                    if (!resolvedNamespace || resolvedNamespace === unknownSymbol) {
                        var namespaceName = getJsxNamespace(location);
                        resolvedNamespace = resolveName(location, namespaceName, 1920 /* Namespace */, /*diagnosticMessage*/ undefined, namespaceName, /*isUse*/ false);
                    }
                    if (resolvedNamespace) {
                        var candidate = resolveSymbol(getSymbol(getExportsOfSymbol(resolveSymbol(resolvedNamespace)), JsxNames.JSX, 1920 /* Namespace */));
                        if (candidate && candidate !== unknownSymbol) {
                            if (links) {
                                links.jsxNamespace = candidate;
                            }
                            return candidate;
                        }
                    }
                    if (links) {
                        links.jsxNamespace = false;
                    }
                }
                // JSX global fallback
                var s = resolveSymbol(getGlobalSymbol(JsxNames.JSX, 1920 /* Namespace */, /*diagnosticMessage*/ undefined));
                if (s === unknownSymbol) {
                    return undefined; // TODO: GH#18217
                }
                return s; // TODO: GH#18217
            }
            /**
             * Look into JSX namespace and then look for container with matching name as nameOfAttribPropContainer.
             * Get a single property from that container if existed. Report an error if there are more than one property.
             *
             * @param nameOfAttribPropContainer a string of value JsxNames.ElementAttributesPropertyNameContainer or JsxNames.ElementChildrenAttributeNameContainer
             *          if other string is given or the container doesn't exist, return undefined.
             */
            function getNameFromJsxElementAttributesContainer(nameOfAttribPropContainer, jsxNamespace) {
                // JSX.ElementAttributesProperty | JSX.ElementChildrenAttribute [symbol]
                var jsxElementAttribPropInterfaceSym = jsxNamespace && getSymbol(jsxNamespace.exports, nameOfAttribPropContainer, 788968 /* Type */);
                // JSX.ElementAttributesProperty | JSX.ElementChildrenAttribute [type]
                var jsxElementAttribPropInterfaceType = jsxElementAttribPropInterfaceSym && getDeclaredTypeOfSymbol(jsxElementAttribPropInterfaceSym);
                // The properties of JSX.ElementAttributesProperty | JSX.ElementChildrenAttribute
                var propertiesOfJsxElementAttribPropInterface = jsxElementAttribPropInterfaceType && getPropertiesOfType(jsxElementAttribPropInterfaceType);
                if (propertiesOfJsxElementAttribPropInterface) {
                    // Element Attributes has zero properties, so the element attributes type will be the class instance type
                    if (propertiesOfJsxElementAttribPropInterface.length === 0) {
                        return "";
                    }
                    // Element Attributes has one property, so the element attributes type will be the type of the corresponding
                    // property of the class instance type
                    else if (propertiesOfJsxElementAttribPropInterface.length === 1) {
                        return propertiesOfJsxElementAttribPropInterface[0].escapedName;
                    }
                    else if (propertiesOfJsxElementAttribPropInterface.length > 1 && jsxElementAttribPropInterfaceSym.declarations) {
                        // More than one property on ElementAttributesProperty is an error
                        error(jsxElementAttribPropInterfaceSym.declarations[0], ts.Diagnostics.The_global_type_JSX_0_may_not_have_more_than_one_property, ts.unescapeLeadingUnderscores(nameOfAttribPropContainer));
                    }
                }
                return undefined;
            }
            function getJsxLibraryManagedAttributes(jsxNamespace) {
                // JSX.LibraryManagedAttributes [symbol]
                return jsxNamespace && getSymbol(jsxNamespace.exports, JsxNames.LibraryManagedAttributes, 788968 /* Type */);
            }
            /// e.g. "props" for React.d.ts,
            /// or 'undefined' if ElementAttributesProperty doesn't exist (which means all
            ///     non-intrinsic elements' attributes type is 'any'),
            /// or '' if it has 0 properties (which means every
            ///     non-intrinsic elements' attributes type is the element instance type)
            function getJsxElementPropertiesName(jsxNamespace) {
                return getNameFromJsxElementAttributesContainer(JsxNames.ElementAttributesPropertyNameContainer, jsxNamespace);
            }
            function getJsxElementChildrenPropertyName(jsxNamespace) {
                return getNameFromJsxElementAttributesContainer(JsxNames.ElementChildrenAttributeNameContainer, jsxNamespace);
            }
            function getUninstantiatedJsxSignaturesOfType(elementType, caller) {
                if (elementType.flags & 4 /* String */) {
                    return [anySignature];
                }
                else if (elementType.flags & 128 /* StringLiteral */) {
                    var intrinsicType = getIntrinsicAttributesTypeFromStringLiteralType(elementType, caller);
                    if (!intrinsicType) {
                        error(caller, ts.Diagnostics.Property_0_does_not_exist_on_type_1, elementType.value, "JSX." + JsxNames.IntrinsicElements);
                        return ts.emptyArray;
                    }
                    else {
                        var fakeSignature = createSignatureForJSXIntrinsic(caller, intrinsicType);
                        return [fakeSignature];
                    }
                }
                var apparentElemType = getApparentType(elementType);
                // Resolve the signatures, preferring constructor
                var signatures = getSignaturesOfType(apparentElemType, 1 /* Construct */);
                if (signatures.length === 0) {
                    // No construct signatures, try call signatures
                    signatures = getSignaturesOfType(apparentElemType, 0 /* Call */);
                }
                if (signatures.length === 0 && apparentElemType.flags & 1048576 /* Union */) {
                    // If each member has some combination of new/call signatures; make a union signature list for those
                    signatures = getUnionSignatures(ts.map(apparentElemType.types, function (t) { return getUninstantiatedJsxSignaturesOfType(t, caller); }));
                }
                return signatures;
            }
            function getIntrinsicAttributesTypeFromStringLiteralType(type, location) {
                // If the elemType is a stringLiteral type, we can then provide a check to make sure that the string literal type is one of the Jsx intrinsic element type
                // For example:
                //      var CustomTag: "h1" = "h1";
                //      <CustomTag> Hello World </CustomTag>
                var intrinsicElementsType = getJsxType(JsxNames.IntrinsicElements, location);
                if (intrinsicElementsType !== errorType) {
                    var stringLiteralTypeName = type.value;
                    var intrinsicProp = getPropertyOfType(intrinsicElementsType, ts.escapeLeadingUnderscores(stringLiteralTypeName));
                    if (intrinsicProp) {
                        return getTypeOfSymbol(intrinsicProp);
                    }
                    var indexSignatureType = getIndexTypeOfType(intrinsicElementsType, stringType);
                    if (indexSignatureType) {
                        return indexSignatureType;
                    }
                    return undefined;
                }
                // If we need to report an error, we already done so here. So just return any to prevent any more error downstream
                return anyType;
            }
            function checkJsxReturnAssignableToAppropriateBound(refKind, elemInstanceType, openingLikeElement) {
                if (refKind === 1 /* Function */) {
                    var sfcReturnConstraint = getJsxStatelessElementTypeAt(openingLikeElement);
                    if (sfcReturnConstraint) {
                        checkTypeRelatedTo(elemInstanceType, sfcReturnConstraint, assignableRelation, openingLikeElement.tagName, ts.Diagnostics.Its_return_type_0_is_not_a_valid_JSX_element, generateInitialErrorChain);
                    }
                }
                else if (refKind === 0 /* Component */) {
                    var classConstraint = getJsxElementClassTypeAt(openingLikeElement);
                    if (classConstraint) {
                        // Issue an error if this return type isn't assignable to JSX.ElementClass, failing that
                        checkTypeRelatedTo(elemInstanceType, classConstraint, assignableRelation, openingLikeElement.tagName, ts.Diagnostics.Its_instance_type_0_is_not_a_valid_JSX_element, generateInitialErrorChain);
                    }
                }
                else { // Mixed
                    var sfcReturnConstraint = getJsxStatelessElementTypeAt(openingLikeElement);
                    var classConstraint = getJsxElementClassTypeAt(openingLikeElement);
                    if (!sfcReturnConstraint || !classConstraint) {
                        return;
                    }
                    var combined = getUnionType([sfcReturnConstraint, classConstraint]);
                    checkTypeRelatedTo(elemInstanceType, combined, assignableRelation, openingLikeElement.tagName, ts.Diagnostics.Its_element_type_0_is_not_a_valid_JSX_element, generateInitialErrorChain);
                }
                function generateInitialErrorChain() {
                    var componentName = ts.getTextOfNode(openingLikeElement.tagName);
                    return ts.chainDiagnosticMessages(/* details */ undefined, ts.Diagnostics._0_cannot_be_used_as_a_JSX_component, componentName);
                }
            }
            /**
             * Get attributes type of the given intrinsic opening-like Jsx element by resolving the tag name.
             * The function is intended to be called from a function which has checked that the opening element is an intrinsic element.
             * @param node an intrinsic JSX opening-like element
             */
            function getIntrinsicAttributesTypeFromJsxOpeningLikeElement(node) {
                ts.Debug.assert(isJsxIntrinsicIdentifier(node.tagName));
                var links = getNodeLinks(node);
                if (!links.resolvedJsxElementAttributesType) {
                    var symbol = getIntrinsicTagSymbol(node);
                    if (links.jsxFlags & 1 /* IntrinsicNamedElement */) {
                        return links.resolvedJsxElementAttributesType = getTypeOfSymbol(symbol) || errorType;
                    }
                    else if (links.jsxFlags & 2 /* IntrinsicIndexedElement */) {
                        return links.resolvedJsxElementAttributesType =
                            getIndexTypeOfType(getJsxType(JsxNames.IntrinsicElements, node), stringType) || errorType;
                    }
                    else {
                        return links.resolvedJsxElementAttributesType = errorType;
                    }
                }
                return links.resolvedJsxElementAttributesType;
            }
            function getJsxElementClassTypeAt(location) {
                var type = getJsxType(JsxNames.ElementClass, location);
                if (type === errorType)
                    return undefined;
                return type;
            }
            function getJsxElementTypeAt(location) {
                return getJsxType(JsxNames.Element, location);
            }
            function getJsxStatelessElementTypeAt(location) {
                var jsxElementType = getJsxElementTypeAt(location);
                if (jsxElementType) {
                    return getUnionType([jsxElementType, nullType]);
                }
            }
            /**
             * Returns all the properties of the Jsx.IntrinsicElements interface
             */
            function getJsxIntrinsicTagNamesAt(location) {
                var intrinsics = getJsxType(JsxNames.IntrinsicElements, location);
                return intrinsics ? getPropertiesOfType(intrinsics) : ts.emptyArray;
            }
            function checkJsxPreconditions(errorNode) {
                // Preconditions for using JSX
                if ((compilerOptions.jsx || 0 /* None */) === 0 /* None */) {
                    error(errorNode, ts.Diagnostics.Cannot_use_JSX_unless_the_jsx_flag_is_provided);
                }
                if (getJsxElementTypeAt(errorNode) === undefined) {
                    if (noImplicitAny) {
                        error(errorNode, ts.Diagnostics.JSX_element_implicitly_has_type_any_because_the_global_type_JSX_Element_does_not_exist);
                    }
                }
            }
            function checkJsxOpeningLikeElementOrOpeningFragment(node) {
                var isNodeOpeningLikeElement = ts.isJsxOpeningLikeElement(node);
                if (isNodeOpeningLikeElement) {
                    checkGrammarJsxElement(node);
                }
                checkJsxPreconditions(node);
                if (!getJsxNamespaceContainerForImplicitImport(node)) {
                    // The reactNamespace/jsxFactory's root symbol should be marked as 'used' so we don't incorrectly elide its import.
                    // And if there is no reactNamespace/jsxFactory's symbol in scope when targeting React emit, we should issue an error.
                    var jsxFactoryRefErr = diagnostics && compilerOptions.jsx === 2 /* React */ ? ts.Diagnostics.Cannot_find_name_0 : undefined;
                    var jsxFactoryNamespace = getJsxNamespace(node);
                    var jsxFactoryLocation = isNodeOpeningLikeElement ? node.tagName : node;
                    // allow null as jsxFragmentFactory
                    var jsxFactorySym = void 0;
                    if (!(ts.isJsxOpeningFragment(node) && jsxFactoryNamespace === "null")) {
                        jsxFactorySym = resolveName(jsxFactoryLocation, jsxFactoryNamespace, 111551 /* Value */, jsxFactoryRefErr, jsxFactoryNamespace, /*isUse*/ true);
                    }
                    if (jsxFactorySym) {
                        // Mark local symbol as referenced here because it might not have been marked
                        // if jsx emit was not jsxFactory as there wont be error being emitted
                        jsxFactorySym.isReferenced = 67108863 /* All */;
                        // If react/jsxFactory symbol is alias, mark it as refereced
                        if (jsxFactorySym.flags & 2097152 /* Alias */ && !getTypeOnlyAliasDeclaration(jsxFactorySym)) {
                            markAliasSymbolAsReferenced(jsxFactorySym);
                        }
                    }
                }
                if (isNodeOpeningLikeElement) {
                    var jsxOpeningLikeNode = node;
                    var sig = getResolvedSignature(jsxOpeningLikeNode);
                    checkDeprecatedSignature(sig, node);
                    checkJsxReturnAssignableToAppropriateBound(getJsxReferenceKind(jsxOpeningLikeNode), getReturnTypeOfSignature(sig), jsxOpeningLikeNode);
                }
            }
            /**
             * Check if a property with the given name is known anywhere in the given type. In an object type, a property
             * is considered known if
             * 1. the object type is empty and the check is for assignability, or
             * 2. if the object type has index signatures, or
             * 3. if the property is actually declared in the object type
             *    (this means that 'toString', for example, is not usually a known property).
             * 4. In a union or intersection type,
             *    a property is considered known if it is known in any constituent type.
             * @param targetType a type to search a given name in
             * @param name a property name to search
             * @param isComparingJsxAttributes a boolean flag indicating whether we are searching in JsxAttributesType
             */
            function isKnownProperty(targetType, name, isComparingJsxAttributes) {
                if (targetType.flags & 524288 /* Object */) {
                    // For backwards compatibility a symbol-named property is satisfied by a string index signature. This
                    // is incorrect and inconsistent with element access expressions, where it is an error, so eventually
                    // we should remove this exception.
                    if (getPropertyOfObjectType(targetType, name) ||
                        getApplicableIndexInfoForName(targetType, name) ||
                        isLateBoundName(name) && getIndexInfoOfType(targetType, stringType) ||
                        isComparingJsxAttributes && isHyphenatedJsxName(name)) {
                        // For JSXAttributes, if the attribute has a hyphenated name, consider that the attribute to be known.
                        return true;
                    }
                }
                else if (targetType.flags & 3145728 /* UnionOrIntersection */ && isExcessPropertyCheckTarget(targetType)) {
                    for (var _i = 0, _a = targetType.types; _i < _a.length; _i++) {
                        var t = _a[_i];
                        if (isKnownProperty(t, name, isComparingJsxAttributes)) {
                            return true;
                        }
                    }
                }
                return false;
            }
            function isExcessPropertyCheckTarget(type) {
                return !!(type.flags & 524288 /* Object */ && !(ts.getObjectFlags(type) & 512 /* ObjectLiteralPatternWithComputedProperties */) ||
                    type.flags & 67108864 /* NonPrimitive */ ||
                    type.flags & 1048576 /* Union */ && ts.some(type.types, isExcessPropertyCheckTarget) ||
                    type.flags & 2097152 /* Intersection */ && ts.every(type.types, isExcessPropertyCheckTarget));
            }
            function checkJsxExpression(node, checkMode) {
                checkGrammarJsxExpression(node);
                if (node.expression) {
                    var type = checkExpression(node.expression, checkMode);
                    if (node.dotDotDotToken && type !== anyType && !isArrayType(type)) {
                        error(node, ts.Diagnostics.JSX_spread_child_must_be_an_array_type);
                    }
                    return type;
                }
                else {
                    return errorType;
                }
            }
            function getDeclarationNodeFlagsFromSymbol(s) {
                return s.valueDeclaration ? ts.getCombinedNodeFlags(s.valueDeclaration) : 0;
            }
            /**
             * Return whether this symbol is a member of a prototype somewhere
             * Note that this is not tracked well within the compiler, so the answer may be incorrect.
             */
            function isPrototypeProperty(symbol) {
                if (symbol.flags & 8192 /* Method */ || ts.getCheckFlags(symbol) & 4 /* SyntheticMethod */) {
                    return true;
                }
                if (ts.isInJSFile(symbol.valueDeclaration)) {
                    var parent = symbol.valueDeclaration.parent;
                    return parent && ts.isBinaryExpression(parent) &&
                        ts.getAssignmentDeclarationKind(parent) === 3 /* PrototypeProperty */;
                }
            }
            /**
             * Check whether the requested property access is valid.
             * Returns true if node is a valid property access, and false otherwise.
             * @param node The node to be checked.
             * @param isSuper True if the access is from `super.`.
             * @param type The type of the object whose property is being accessed. (Not the type of the property.)
             * @param prop The symbol for the property being accessed.
             */
            function checkPropertyAccessibility(node, isSuper, writing, type, prop, reportError) {
                if (reportError === void 0) {
                    reportError = true;
                }
                var flags = ts.getDeclarationModifierFlagsFromSymbol(prop, writing);
                var errorNode = node.kind === 159 /* QualifiedName */ ? node.right :
                    node.kind === 198 /* ImportType */ ? node :
                        node.kind === 201 /* BindingElement */ && node.propertyName ? node.propertyName : node.name;
                if (isSuper) {
                    // TS 1.0 spec (April 2014): 4.8.2
                    // - In a constructor, instance member function, instance member accessor, or
                    //   instance member variable initializer where this references a derived class instance,
                    //   a super property access is permitted and must specify a public instance member function of the base class.
                    // - In a static member function or static member accessor
                    //   where this references the constructor function object of a derived class,
                    //   a super property access is permitted and must specify a public static member function of the base class.
                    if (languageVersion < 2 /* ES2015 */) {
                        if (symbolHasNonMethodDeclaration(prop)) {
                            if (reportError) {
                                error(errorNode, ts.Diagnostics.Only_public_and_protected_methods_of_the_base_class_are_accessible_via_the_super_keyword);
                            }
                            return false;
                        }
                    }
                    if (flags & 128 /* Abstract */) {
                        // A method cannot be accessed in a super property access if the method is abstract.
                        // This error could mask a private property access error. But, a member
                        // cannot simultaneously be private and abstract, so this will trigger an
                        // additional error elsewhere.
                        if (reportError) {
                            error(errorNode, ts.Diagnostics.Abstract_method_0_in_class_1_cannot_be_accessed_via_super_expression, symbolToString(prop), typeToString(getDeclaringClass(prop)));
                        }
                        return false;
                    }
                }
                // Referencing abstract properties within their own constructors is not allowed
                if ((flags & 128 /* Abstract */) && symbolHasNonMethodDeclaration(prop) &&
                    (ts.isThisProperty(node) || ts.isThisInitializedObjectBindingExpression(node) || ts.isObjectBindingPattern(node.parent) && ts.isThisInitializedDeclaration(node.parent.parent))) {
                    var declaringClassDeclaration = ts.getClassLikeDeclarationOfSymbol(getParentOfSymbol(prop));
                    if (declaringClassDeclaration && isNodeUsedDuringClassInitialization(node)) {
                        if (reportError) {
                            error(errorNode, ts.Diagnostics.Abstract_property_0_in_class_1_cannot_be_accessed_in_the_constructor, symbolToString(prop), ts.getTextOfIdentifierOrLiteral(declaringClassDeclaration.name)); // TODO: GH#18217
                        }
                        return false;
                    }
                }
                // Public properties are otherwise accessible.
                if (!(flags & 24 /* NonPublicAccessibilityModifier */)) {
                    return true;
                }
                // Property is known to be private or protected at this point
                // Private property is accessible if the property is within the declaring class
                if (flags & 8 /* Private */) {
                    var declaringClassDeclaration = ts.getClassLikeDeclarationOfSymbol(getParentOfSymbol(prop));
                    if (!isNodeWithinClass(node, declaringClassDeclaration)) {
                        if (reportError) {
                            error(errorNode, ts.Diagnostics.Property_0_is_private_and_only_accessible_within_class_1, symbolToString(prop), typeToString(getDeclaringClass(prop)));
                        }
                        return false;
                    }
                    return true;
                }
                // Property is known to be protected at this point
                // All protected properties of a supertype are accessible in a super access
                if (isSuper) {
                    return true;
                }
                // Find the first enclosing class that has the declaring classes of the protected constituents
                // of the property as base classes
                var enclosingClass = forEachEnclosingClass(node, function (enclosingDeclaration) {
                    var enclosingClass = getDeclaredTypeOfSymbol(getSymbolOfNode(enclosingDeclaration));
                    return isClassDerivedFromDeclaringClasses(enclosingClass, prop, writing) ? enclosingClass : undefined;
                });
                // A protected property is accessible if the property is within the declaring class or classes derived from it
                if (!enclosingClass) {
                    // allow PropertyAccessibility if context is in function with this parameter
                    // static member access is disallow
                    var thisParameter = void 0;
                    if (flags & 32 /* Static */ || !(thisParameter = getThisParameterFromNodeContext(node)) || !thisParameter.type) {
                        if (reportError) {
                            error(errorNode, ts.Diagnostics.Property_0_is_protected_and_only_accessible_within_class_1_and_its_subclasses, symbolToString(prop), typeToString(getDeclaringClass(prop) || type));
                        }
                        return false;
                    }
                    var thisType = getTypeFromTypeNode(thisParameter.type);
                    enclosingClass = ((thisType.flags & 262144 /* TypeParameter */) ? getConstraintOfTypeParameter(thisType) : thisType).target;
                }
                // No further restrictions for static properties
                if (flags & 32 /* Static */) {
                    return true;
                }
                if (type.flags & 262144 /* TypeParameter */) {
                    // get the original type -- represented as the type constraint of the 'this' type
                    type = type.isThisType ? getConstraintOfTypeParameter(type) : getBaseConstraintOfType(type); // TODO: GH#18217 Use a different variable that's allowed to be undefined
                }
                if (!type || !hasBaseType(type, enclosingClass)) {
                    if (reportError) {
                        error(errorNode, ts.Diagnostics.Property_0_is_protected_and_only_accessible_through_an_instance_of_class_1_This_is_an_instance_of_class_2, symbolToString(prop), typeToString(enclosingClass), typeToString(type));
                    }
                    return false;
                }
                return true;
            }
            function getThisParameterFromNodeContext(node) {
                var thisContainer = ts.getThisContainer(node, /* includeArrowFunctions */ false);
                return thisContainer && ts.isFunctionLike(thisContainer) ? ts.getThisParameter(thisContainer) : undefined;
            }
            function symbolHasNonMethodDeclaration(symbol) {
                return !!forEachProperty(symbol, function (prop) { return !(prop.flags & 8192 /* Method */); });
            }
            function checkNonNullExpression(node) {
                return checkNonNullType(checkExpression(node), node);
            }
            function isNullableType(type) {
                return !!((strictNullChecks ? getFalsyFlags(type) : type.flags) & 98304 /* Nullable */);
            }
            function getNonNullableTypeIfNeeded(type) {
                return isNullableType(type) ? getNonNullableType(type) : type;
            }
            function reportObjectPossiblyNullOrUndefinedError(node, flags) {
                error(node, flags & 32768 /* Undefined */ ? flags & 65536 /* Null */ ?
                    ts.Diagnostics.Object_is_possibly_null_or_undefined :
                    ts.Diagnostics.Object_is_possibly_undefined :
                    ts.Diagnostics.Object_is_possibly_null);
            }
            function reportCannotInvokePossiblyNullOrUndefinedError(node, flags) {
                error(node, flags & 32768 /* Undefined */ ? flags & 65536 /* Null */ ?
                    ts.Diagnostics.Cannot_invoke_an_object_which_is_possibly_null_or_undefined :
                    ts.Diagnostics.Cannot_invoke_an_object_which_is_possibly_undefined :
                    ts.Diagnostics.Cannot_invoke_an_object_which_is_possibly_null);
            }
            function checkNonNullTypeWithReporter(type, node, reportError) {
                if (strictNullChecks && type.flags & 2 /* Unknown */) {
                    error(node, ts.Diagnostics.Object_is_of_type_unknown);
                    return errorType;
                }
                var kind = (strictNullChecks ? getFalsyFlags(type) : type.flags) & 98304 /* Nullable */;
                if (kind) {
                    reportError(node, kind);
                    var t = getNonNullableType(type);
                    return t.flags & (98304 /* Nullable */ | 131072 /* Never */) ? errorType : t;
                }
                return type;
            }
            function checkNonNullType(type, node) {
                return checkNonNullTypeWithReporter(type, node, reportObjectPossiblyNullOrUndefinedError);
            }
            function checkNonNullNonVoidType(type, node) {
                var nonNullType = checkNonNullType(type, node);
                if (nonNullType !== errorType && nonNullType.flags & 16384 /* Void */) {
                    error(node, ts.Diagnostics.Object_is_possibly_undefined);
                }
                return nonNullType;
            }
            function checkPropertyAccessExpression(node, checkMode) {
                return node.flags & 32 /* OptionalChain */ ? checkPropertyAccessChain(node, checkMode) :
                    checkPropertyAccessExpressionOrQualifiedName(node, node.expression, checkNonNullExpression(node.expression), node.name, checkMode);
            }
            function checkPropertyAccessChain(node, checkMode) {
                var leftType = checkExpression(node.expression);
                var nonOptionalType = getOptionalExpressionType(leftType, node.expression);
                return propagateOptionalTypeMarker(checkPropertyAccessExpressionOrQualifiedName(node, node.expression, checkNonNullType(nonOptionalType, node.expression), node.name, checkMode), node, nonOptionalType !== leftType);
            }
            function checkQualifiedName(node, checkMode) {
                var leftType = ts.isPartOfTypeQuery(node) && ts.isThisIdentifier(node.left) ? checkNonNullType(checkThisExpression(node.left), node.left) : checkNonNullExpression(node.left);
                return checkPropertyAccessExpressionOrQualifiedName(node, node.left, leftType, node.right, checkMode);
            }
            function isMethodAccessForCall(node) {
                while (node.parent.kind === 210 /* ParenthesizedExpression */) {
                    node = node.parent;
                }
                return ts.isCallOrNewExpression(node.parent) && node.parent.expression === node;
            }
            // Lookup the private identifier lexically.
            function lookupSymbolForPrivateIdentifierDeclaration(propName, location) {
                for (var containingClass = ts.getContainingClass(location); !!containingClass; containingClass = ts.getContainingClass(containingClass)) {
                    var symbol = containingClass.symbol;
                    var name = ts.getSymbolNameForPrivateIdentifier(symbol, propName);
                    var prop = (symbol.members && symbol.members.get(name)) || (symbol.exports && symbol.exports.get(name));
                    if (prop) {
                        return prop;
                    }
                }
            }
            function getPrivateIdentifierPropertyOfType(leftType, lexicallyScopedIdentifier) {
                return getPropertyOfType(leftType, lexicallyScopedIdentifier.escapedName);
            }
            function checkPrivateIdentifierPropertyAccess(leftType, right, lexicallyScopedIdentifier) {
                // Either the identifier could not be looked up in the lexical scope OR the lexically scoped identifier did not exist on the type.
                // Find a private identifier with the same description on the type.
                var propertyOnType;
                var properties = getPropertiesOfType(leftType);
                if (properties) {
                    ts.forEach(properties, function (symbol) {
                        var decl = symbol.valueDeclaration;
                        if (decl && ts.isNamedDeclaration(decl) && ts.isPrivateIdentifier(decl.name) && decl.name.escapedText === right.escapedText) {
                            propertyOnType = symbol;
                            return true;
                        }
                    });
                }
                var diagName = diagnosticName(right);
                if (propertyOnType) {
                    var typeValueDecl = ts.Debug.checkDefined(propertyOnType.valueDeclaration);
                    var typeClass_1 = ts.Debug.checkDefined(ts.getContainingClass(typeValueDecl));
                    // We found a private identifier property with the same description.
                    // Either:
                    // - There is a lexically scoped private identifier AND it shadows the one we found on the type.
                    // - It is an attempt to access the private identifier outside of the class.
                    if (lexicallyScopedIdentifier === null || lexicallyScopedIdentifier === void 0 ? void 0 : lexicallyScopedIdentifier.valueDeclaration) {
                        var lexicalValueDecl = lexicallyScopedIdentifier.valueDeclaration;
                        var lexicalClass = ts.getContainingClass(lexicalValueDecl);
                        ts.Debug.assert(!!lexicalClass);
                        if (ts.findAncestor(lexicalClass, function (n) { return typeClass_1 === n; })) {
                            var diagnostic = error(right, ts.Diagnostics.The_property_0_cannot_be_accessed_on_type_1_within_this_class_because_it_is_shadowed_by_another_private_identifier_with_the_same_spelling, diagName, typeToString(leftType));
                            ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(lexicalValueDecl, ts.Diagnostics.The_shadowing_declaration_of_0_is_defined_here, diagName), ts.createDiagnosticForNode(typeValueDecl, ts.Diagnostics.The_declaration_of_0_that_you_probably_intended_to_use_is_defined_here, diagName));
                            return true;
                        }
                    }
                    error(right, ts.Diagnostics.Property_0_is_not_accessible_outside_class_1_because_it_has_a_private_identifier, diagName, diagnosticName(typeClass_1.name || anon));
                    return true;
                }
                return false;
            }
            function isThisPropertyAccessInConstructor(node, prop) {
                return (isConstructorDeclaredProperty(prop) || ts.isThisProperty(node) && isAutoTypedProperty(prop))
                    && ts.getThisContainer(node, /*includeArrowFunctions*/ true) === getDeclaringConstructor(prop);
            }
            function checkPropertyAccessExpressionOrQualifiedName(node, left, leftType, right, checkMode) {
                var parentSymbol = getNodeLinks(left).resolvedSymbol;
                var assignmentKind = ts.getAssignmentTargetKind(node);
                var apparentType = getApparentType(assignmentKind !== 0 /* None */ || isMethodAccessForCall(node) ? getWidenedType(leftType) : leftType);
                var isAnyLike = isTypeAny(apparentType) || apparentType === silentNeverType;
                var prop;
                if (ts.isPrivateIdentifier(right)) {
                    if (languageVersion < 99 /* ESNext */) {
                        if (assignmentKind !== 0 /* None */) {
                            checkExternalEmitHelpers(node, 1048576 /* ClassPrivateFieldSet */);
                        }
                        if (assignmentKind !== 1 /* Definite */) {
                            checkExternalEmitHelpers(node, 524288 /* ClassPrivateFieldGet */);
                        }
                    }
                    var lexicallyScopedSymbol = lookupSymbolForPrivateIdentifierDeclaration(right.escapedText, right);
                    if (assignmentKind && lexicallyScopedSymbol && lexicallyScopedSymbol.valueDeclaration && ts.isMethodDeclaration(lexicallyScopedSymbol.valueDeclaration)) {
                        grammarErrorOnNode(right, ts.Diagnostics.Cannot_assign_to_private_method_0_Private_methods_are_not_writable, ts.idText(right));
                    }
                    if ((lexicallyScopedSymbol === null || lexicallyScopedSymbol === void 0 ? void 0 : lexicallyScopedSymbol.valueDeclaration) && (compilerOptions.target === 99 /* ESNext */ && !useDefineForClassFields)) {
                        var lexicalClass_1 = ts.getContainingClass(lexicallyScopedSymbol.valueDeclaration);
                        var parentStaticFieldInitializer = ts.findAncestor(node, function (n) {
                            if (n === lexicalClass_1)
                                return "quit";
                            if (ts.isPropertyDeclaration(n.parent) && ts.hasStaticModifier(n.parent) && n.parent.initializer === n && n.parent.parent === lexicalClass_1) {
                                return true;
                            }
                            return false;
                        });
                        if (parentStaticFieldInitializer) {
                            var parentStaticFieldInitializerSymbol = getSymbolOfNode(parentStaticFieldInitializer.parent);
                            ts.Debug.assert(parentStaticFieldInitializerSymbol, "Initializer without declaration symbol");
                            var diagnostic = error(node, ts.Diagnostics.Property_0_may_not_be_used_in_a_static_property_s_initializer_in_the_same_class_when_target_is_esnext_and_useDefineForClassFields_is_false, ts.symbolName(lexicallyScopedSymbol));
                            ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(parentStaticFieldInitializer.parent, ts.Diagnostics.Initializer_for_property_0, ts.symbolName(parentStaticFieldInitializerSymbol)));
                        }
                    }
                    if (isAnyLike) {
                        if (lexicallyScopedSymbol) {
                            return apparentType;
                        }
                        if (!ts.getContainingClass(right)) {
                            grammarErrorOnNode(right, ts.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies);
                            return anyType;
                        }
                    }
                    prop = lexicallyScopedSymbol ? getPrivateIdentifierPropertyOfType(leftType, lexicallyScopedSymbol) : undefined;
                    // Check for private-identifier-specific shadowing and lexical-scoping errors.
                    if (!prop && checkPrivateIdentifierPropertyAccess(leftType, right, lexicallyScopedSymbol)) {
                        return errorType;
                    }
                    else {
                        var isSetonlyAccessor = prop && prop.flags & 65536 /* SetAccessor */ && !(prop.flags & 32768 /* GetAccessor */);
                        if (isSetonlyAccessor && assignmentKind !== 1 /* Definite */) {
                            error(node, ts.Diagnostics.Private_accessor_was_defined_without_a_getter);
                        }
                    }
                }
                else {
                    if (isAnyLike) {
                        if (ts.isIdentifier(left) && parentSymbol) {
                            markAliasReferenced(parentSymbol, node);
                        }
                        return apparentType;
                    }
                    prop = getPropertyOfType(apparentType, right.escapedText);
                }
                // In `Foo.Bar.Baz`, 'Foo' is not referenced if 'Bar' is a const enum or a module containing only const enums.
                // The exceptions are:
                //   1. if 'isolatedModules' is enabled, because the const enum value will not be inlined, and
                //   2. if 'preserveConstEnums' is enabled and the expression is itself an export, e.g. `export = Foo.Bar.Baz`.
                if (ts.isIdentifier(left) && parentSymbol && (compilerOptions.isolatedModules || !(prop && isConstEnumOrConstEnumOnlyModule(prop)) || ts.shouldPreserveConstEnums(compilerOptions) && isExportOrExportExpression(node))) {
                    markAliasReferenced(parentSymbol, node);
                }
                var propType;
                if (!prop) {
                    var indexInfo = !ts.isPrivateIdentifier(right) && (assignmentKind === 0 /* None */ || !isGenericObjectType(leftType) || isThisTypeParameter(leftType)) ?
                        getApplicableIndexInfoForName(apparentType, right.escapedText) : undefined;
                    if (!(indexInfo && indexInfo.type)) {
                        var isUncheckedJS = isUncheckedJSSuggestion(node, leftType.symbol, /*excludeClasses*/ true);
                        if (!isUncheckedJS && isJSLiteralType(leftType)) {
                            return anyType;
                        }
                        if (leftType.symbol === globalThisSymbol) {
                            if (globalThisSymbol.exports.has(right.escapedText) && (globalThisSymbol.exports.get(right.escapedText).flags & 418 /* BlockScoped */)) {
                                error(right, ts.Diagnostics.Property_0_does_not_exist_on_type_1, ts.unescapeLeadingUnderscores(right.escapedText), typeToString(leftType));
                            }
                            else if (noImplicitAny) {
                                error(right, ts.Diagnostics.Element_implicitly_has_an_any_type_because_type_0_has_no_index_signature, typeToString(leftType));
                            }
                            return anyType;
                        }
                        if (right.escapedText && !checkAndReportErrorForExtendingInterface(node)) {
                            reportNonexistentProperty(right, isThisTypeParameter(leftType) ? apparentType : leftType, isUncheckedJS);
                        }
                        return errorType;
                    }
                    if (indexInfo.isReadonly && (ts.isAssignmentTarget(node) || ts.isDeleteTarget(node))) {
                        error(node, ts.Diagnostics.Index_signature_in_type_0_only_permits_reading, typeToString(apparentType));
                    }
                    propType = (compilerOptions.noUncheckedIndexedAccess && !ts.isAssignmentTarget(node)) ? getUnionType([indexInfo.type, undefinedType]) : indexInfo.type;
                    if (compilerOptions.noPropertyAccessFromIndexSignature && ts.isPropertyAccessExpression(node)) {
                        error(right, ts.Diagnostics.Property_0_comes_from_an_index_signature_so_it_must_be_accessed_with_0, ts.unescapeLeadingUnderscores(right.escapedText));
                    }
                }
                else {
                    if (prop.declarations && getDeclarationNodeFlagsFromSymbol(prop) & 134217728 /* Deprecated */ && isUncalledFunctionReference(node, prop)) {
                        addDeprecatedSuggestion(right, prop.declarations, right.escapedText);
                    }
                    checkPropertyNotUsedBeforeDeclaration(prop, node, right);
                    markPropertyAsReferenced(prop, node, isSelfTypeAccess(left, parentSymbol));
                    getNodeLinks(node).resolvedSymbol = prop;
                    var writing = ts.isWriteAccess(node);
                    checkPropertyAccessibility(node, left.kind === 106 /* SuperKeyword */, writing, apparentType, prop);
                    if (isAssignmentToReadonlyEntity(node, prop, assignmentKind)) {
                        error(right, ts.Diagnostics.Cannot_assign_to_0_because_it_is_a_read_only_property, ts.idText(right));
                        return errorType;
                    }
                    propType = isThisPropertyAccessInConstructor(node, prop) ? autoType : writing ? getSetAccessorTypeOfSymbol(prop) : getTypeOfSymbol(prop);
                }
                return getFlowTypeOfAccessExpression(node, prop, propType, right, checkMode);
            }
            /**
             * Determines whether a did-you-mean error should be a suggestion in an unchecked JS file.
             * Only applies to unchecked JS files without checkJS, // @ts-check or // @ts-nocheck
             * It does not suggest when the suggestion:
             * - Is from a global file that is different from the reference file, or
             * - (optionally) Is a class, or is a this.x property access expression
             */
            function isUncheckedJSSuggestion(node, suggestion, excludeClasses) {
                var file = ts.getSourceFileOfNode(node);
                if (file) {
                    if (compilerOptions.checkJs === undefined && file.checkJsDirective === undefined && (file.scriptKind === 1 /* JS */ || file.scriptKind === 2 /* JSX */)) {
                        var declarationFile = ts.forEach(suggestion === null || suggestion === void 0 ? void 0 : suggestion.declarations, ts.getSourceFileOfNode);
                        return !(file !== declarationFile && !!declarationFile && isGlobalSourceFile(declarationFile))
                            && !(excludeClasses && suggestion && suggestion.flags & 32 /* Class */)
                            && !(!!node && excludeClasses && ts.isPropertyAccessExpression(node) && node.expression.kind === 108 /* ThisKeyword */);
                    }
                }
                return false;
            }
            function getFlowTypeOfAccessExpression(node, prop, propType, errorNode, checkMode) {
                // Only compute control flow type if this is a property access expression that isn't an
                // assignment target, and the referenced property was declared as a variable, property,
                // accessor, or optional method.
                var assignmentKind = ts.getAssignmentTargetKind(node);
                if (assignmentKind === 1 /* Definite */) {
                    return removeMissingType(propType, !!(prop && prop.flags & 16777216 /* Optional */));
                }
                if (prop &&
                    !(prop.flags & (3 /* Variable */ | 4 /* Property */ | 98304 /* Accessor */))
                    && !(prop.flags & 8192 /* Method */ && propType.flags & 1048576 /* Union */)
                    && !isDuplicatedCommonJSExport(prop.declarations)) {
                    return propType;
                }
                if (propType === autoType) {
                    return getFlowTypeOfProperty(node, prop);
                }
                propType = getNarrowableTypeForReference(propType, node, checkMode);
                // If strict null checks and strict property initialization checks are enabled, if we have
                // a this.xxx property access, if the property is an instance property without an initializer,
                // and if we are in a constructor of the same class as the property declaration, assume that
                // the property is uninitialized at the top of the control flow.
                var assumeUninitialized = false;
                if (strictNullChecks && strictPropertyInitialization && ts.isAccessExpression(node) && node.expression.kind === 108 /* ThisKeyword */) {
                    var declaration = prop && prop.valueDeclaration;
                    if (declaration && isPropertyWithoutInitializer(declaration)) {
                        if (!ts.isStatic(declaration)) {
                            var flowContainer = getControlFlowContainer(node);
                            if (flowContainer.kind === 169 /* Constructor */ && flowContainer.parent === declaration.parent && !(declaration.flags & 8388608 /* Ambient */)) {
                                assumeUninitialized = true;
                            }
                        }
                    }
                }
                else if (strictNullChecks && prop && prop.valueDeclaration &&
                    ts.isPropertyAccessExpression(prop.valueDeclaration) &&
                    ts.getAssignmentDeclarationPropertyAccessKind(prop.valueDeclaration) &&
                    getControlFlowContainer(node) === getControlFlowContainer(prop.valueDeclaration)) {
                    assumeUninitialized = true;
                }
                var flowType = getFlowTypeOfReference(node, propType, assumeUninitialized ? getOptionalType(propType) : propType);
                if (assumeUninitialized && !(getFalsyFlags(propType) & 32768 /* Undefined */) && getFalsyFlags(flowType) & 32768 /* Undefined */) {
                    error(errorNode, ts.Diagnostics.Property_0_is_used_before_being_assigned, symbolToString(prop)); // TODO: GH#18217
                    // Return the declared type to reduce follow-on errors
                    return propType;
                }
                return assignmentKind ? getBaseTypeOfLiteralType(flowType) : flowType;
            }
            function checkPropertyNotUsedBeforeDeclaration(prop, node, right) {
                var valueDeclaration = prop.valueDeclaration;
                if (!valueDeclaration || ts.getSourceFileOfNode(node).isDeclarationFile) {
                    return;
                }
                var diagnosticMessage;
                var declarationName = ts.idText(right);
                if (isInPropertyInitializerOrClassStaticBlock(node)
                    && !isOptionalPropertyDeclaration(valueDeclaration)
                    && !(ts.isAccessExpression(node) && ts.isAccessExpression(node.expression))
                    && !isBlockScopedNameDeclaredBeforeUse(valueDeclaration, right)
                    && (compilerOptions.useDefineForClassFields || !isPropertyDeclaredInAncestorClass(prop))) {
                    diagnosticMessage = error(right, ts.Diagnostics.Property_0_is_used_before_its_initialization, declarationName);
                }
                else if (valueDeclaration.kind === 255 /* ClassDeclaration */ &&
                    node.parent.kind !== 176 /* TypeReference */ &&
                    !(valueDeclaration.flags & 8388608 /* Ambient */) &&
                    !isBlockScopedNameDeclaredBeforeUse(valueDeclaration, right)) {
                    diagnosticMessage = error(right, ts.Diagnostics.Class_0_used_before_its_declaration, declarationName);
                }
                if (diagnosticMessage) {
                    ts.addRelatedInfo(diagnosticMessage, ts.createDiagnosticForNode(valueDeclaration, ts.Diagnostics._0_is_declared_here, declarationName));
                }
            }
            function isInPropertyInitializerOrClassStaticBlock(node) {
                return !!ts.findAncestor(node, function (node) {
                    switch (node.kind) {
                        case 165 /* PropertyDeclaration */:
                            return true;
                        case 291 /* PropertyAssignment */:
                        case 167 /* MethodDeclaration */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                        case 293 /* SpreadAssignment */:
                        case 160 /* ComputedPropertyName */:
                        case 231 /* TemplateSpan */:
                        case 286 /* JsxExpression */:
                        case 283 /* JsxAttribute */:
                        case 284 /* JsxAttributes */:
                        case 285 /* JsxSpreadAttribute */:
                        case 278 /* JsxOpeningElement */:
                        case 226 /* ExpressionWithTypeArguments */:
                        case 289 /* HeritageClause */:
                            return false;
                        case 212 /* ArrowFunction */:
                        case 236 /* ExpressionStatement */:
                            return ts.isBlock(node.parent) && ts.isClassStaticBlockDeclaration(node.parent.parent) ? true : "quit";
                        default:
                            return ts.isExpressionNode(node) ? false : "quit";
                    }
                });
            }
            /**
             * It's possible that "prop.valueDeclaration" is a local declaration, but the property was also declared in a superclass.
             * In that case we won't consider it used before its declaration, because it gets its value from the superclass' declaration.
             */
            function isPropertyDeclaredInAncestorClass(prop) {
                if (!(prop.parent.flags & 32 /* Class */)) {
                    return false;
                }
                var classType = getTypeOfSymbol(prop.parent);
                while (true) {
                    classType = classType.symbol && getSuperClass(classType);
                    if (!classType) {
                        return false;
                    }
                    var superProperty = getPropertyOfType(classType, prop.escapedName);
                    if (superProperty && superProperty.valueDeclaration) {
                        return true;
                    }
                }
            }
            function getSuperClass(classType) {
                var x = getBaseTypes(classType);
                if (x.length === 0) {
                    return undefined;
                }
                return getIntersectionType(x);
            }
            function reportNonexistentProperty(propNode, containingType, isUncheckedJS) {
                var errorInfo;
                var relatedInfo;
                if (!ts.isPrivateIdentifier(propNode) && containingType.flags & 1048576 /* Union */ && !(containingType.flags & 131068 /* Primitive */)) {
                    for (var _i = 0, _a = containingType.types; _i < _a.length; _i++) {
                        var subtype = _a[_i];
                        if (!getPropertyOfType(subtype, propNode.escapedText) && !getApplicableIndexInfoForName(subtype, propNode.escapedText)) {
                            errorInfo = ts.chainDiagnosticMessages(errorInfo, ts.Diagnostics.Property_0_does_not_exist_on_type_1, ts.declarationNameToString(propNode), typeToString(subtype));
                            break;
                        }
                    }
                }
                if (typeHasStaticProperty(propNode.escapedText, containingType)) {
                    var propName = ts.declarationNameToString(propNode);
                    var typeName = typeToString(containingType);
                    errorInfo = ts.chainDiagnosticMessages(errorInfo, ts.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_to_access_the_static_member_2_instead, propName, typeName, typeName + "." + propName);
                }
                else {
                    var promisedType = getPromisedTypeOfPromise(containingType);
                    if (promisedType && getPropertyOfType(promisedType, propNode.escapedText)) {
                        errorInfo = ts.chainDiagnosticMessages(errorInfo, ts.Diagnostics.Property_0_does_not_exist_on_type_1, ts.declarationNameToString(propNode), typeToString(containingType));
                        relatedInfo = ts.createDiagnosticForNode(propNode, ts.Diagnostics.Did_you_forget_to_use_await);
                    }
                    else {
                        var missingProperty = ts.declarationNameToString(propNode);
                        var container = typeToString(containingType);
                        var libSuggestion = getSuggestedLibForNonExistentProperty(missingProperty, containingType);
                        if (libSuggestion !== undefined) {
                            errorInfo = ts.chainDiagnosticMessages(errorInfo, ts.Diagnostics.Property_0_does_not_exist_on_type_1_Do_you_need_to_change_your_target_library_Try_changing_the_lib_compiler_option_to_2_or_later, missingProperty, container, libSuggestion);
                        }
                        else {
                            var suggestion = getSuggestedSymbolForNonexistentProperty(propNode, containingType);
                            if (suggestion !== undefined) {
                                var suggestedName = ts.symbolName(suggestion);
                                var message = isUncheckedJS ? ts.Diagnostics.Property_0_may_not_exist_on_type_1_Did_you_mean_2 : ts.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2;
                                errorInfo = ts.chainDiagnosticMessages(errorInfo, message, missingProperty, container, suggestedName);
                                relatedInfo = suggestion.valueDeclaration && ts.createDiagnosticForNode(suggestion.valueDeclaration, ts.Diagnostics._0_is_declared_here, suggestedName);
                            }
                            else {
                                var diagnostic = containerSeemsToBeEmptyDomElement(containingType)
                                    ? ts.Diagnostics.Property_0_does_not_exist_on_type_1_Try_changing_the_lib_compiler_option_to_include_dom
                                    : ts.Diagnostics.Property_0_does_not_exist_on_type_1;
                                errorInfo = ts.chainDiagnosticMessages(elaborateNeverIntersection(errorInfo, containingType), diagnostic, missingProperty, container);
                            }
                        }
                    }
                }
                var resultDiagnostic = ts.createDiagnosticForNodeFromMessageChain(propNode, errorInfo);
                if (relatedInfo) {
                    ts.addRelatedInfo(resultDiagnostic, relatedInfo);
                }
                addErrorOrSuggestion(!isUncheckedJS, resultDiagnostic);
            }
            function containerSeemsToBeEmptyDomElement(containingType) {
                return (compilerOptions.lib && !compilerOptions.lib.includes("dom")) &&
                    everyContainedType(containingType, function (type) { return type.symbol && /^(EventTarget|Node|((HTML[a-zA-Z]*)?Element))$/.test(ts.unescapeLeadingUnderscores(type.symbol.escapedName)); }) &&
                    isEmptyObjectType(containingType);
            }
            function typeHasStaticProperty(propName, containingType) {
                var prop = containingType.symbol && getPropertyOfType(getTypeOfSymbol(containingType.symbol), propName);
                return prop !== undefined && !!prop.valueDeclaration && ts.isStatic(prop.valueDeclaration);
            }
            function getSuggestedLibForNonExistentName(name) {
                var missingName = diagnosticName(name);
                var allFeatures = ts.getScriptTargetFeatures();
                var libTargets = ts.getOwnKeys(allFeatures);
                for (var _i = 0, libTargets_1 = libTargets; _i < libTargets_1.length; _i++) {
                    var libTarget = libTargets_1[_i];
                    var containingTypes = ts.getOwnKeys(allFeatures[libTarget]);
                    if (containingTypes !== undefined && ts.contains(containingTypes, missingName)) {
                        return libTarget;
                    }
                }
            }
            function getSuggestedLibForNonExistentProperty(missingProperty, containingType) {
                var container = getApparentType(containingType).symbol;
                if (!container) {
                    return undefined;
                }
                var allFeatures = ts.getScriptTargetFeatures();
                var libTargets = ts.getOwnKeys(allFeatures);
                for (var _i = 0, libTargets_2 = libTargets; _i < libTargets_2.length; _i++) {
                    var libTarget = libTargets_2[_i];
                    var featuresOfLib = allFeatures[libTarget];
                    var featuresOfContainingType = featuresOfLib[ts.symbolName(container)];
                    if (featuresOfContainingType !== undefined && ts.contains(featuresOfContainingType, missingProperty)) {
                        return libTarget;
                    }
                }
            }
            function getSuggestedSymbolForNonexistentClassMember(name, baseType) {
                return getSpellingSuggestionForName(name, getPropertiesOfType(baseType), 106500 /* ClassMember */);
            }
            function getSuggestedSymbolForNonexistentProperty(name, containingType) {
                var props = getPropertiesOfType(containingType);
                if (typeof name !== "string") {
                    var parent_2 = name.parent;
                    if (ts.isPropertyAccessExpression(parent_2)) {
                        props = ts.filter(props, function (prop) { return isValidPropertyAccessForCompletions(parent_2, containingType, prop); });
                    }
                    name = ts.idText(name);
                }
                return getSpellingSuggestionForName(name, props, 111551 /* Value */);
            }
            function getSuggestedSymbolForNonexistentJSXAttribute(name, containingType) {
                var strName = ts.isString(name) ? name : ts.idText(name);
                var properties = getPropertiesOfType(containingType);
                var jsxSpecific = strName === "for" ? ts.find(properties, function (x) { return ts.symbolName(x) === "htmlFor"; })
                    : strName === "class" ? ts.find(properties, function (x) { return ts.symbolName(x) === "className"; })
                        : undefined;
                return jsxSpecific !== null && jsxSpecific !== void 0 ? jsxSpecific : getSpellingSuggestionForName(strName, properties, 111551 /* Value */);
            }
            function getSuggestionForNonexistentProperty(name, containingType) {
                var suggestion = getSuggestedSymbolForNonexistentProperty(name, containingType);
                return suggestion && ts.symbolName(suggestion);
            }
            function getSuggestedSymbolForNonexistentSymbol(location, outerName, meaning) {
                ts.Debug.assert(outerName !== undefined, "outername should always be defined");
                var result = resolveNameHelper(location, outerName, meaning, /*nameNotFoundMessage*/ undefined, outerName, /*isUse*/ false, /*excludeGlobals*/ false, function (symbols, name, meaning) {
                    ts.Debug.assertEqual(outerName, name, "name should equal outerName");
                    var symbol = getSymbol(symbols, name, meaning);
                    // Sometimes the symbol is found when location is a return type of a function: `typeof x` and `x` is declared in the body of the function
                    // So the table *contains* `x` but `x` isn't actually in scope.
                    // However, resolveNameHelper will continue and call this callback again, so we'll eventually get a correct suggestion.
                    return symbol || getSpellingSuggestionForName(ts.unescapeLeadingUnderscores(name), ts.arrayFrom(symbols.values()), meaning);
                });
                return result;
            }
            function getSuggestionForNonexistentSymbol(location, outerName, meaning) {
                var symbolResult = getSuggestedSymbolForNonexistentSymbol(location, outerName, meaning);
                return symbolResult && ts.symbolName(symbolResult);
            }
            function getSuggestedSymbolForNonexistentModule(name, targetModule) {
                return targetModule.exports && getSpellingSuggestionForName(ts.idText(name), getExportsOfModuleAsArray(targetModule), 2623475 /* ModuleMember */);
            }
            function getSuggestionForNonexistentExport(name, targetModule) {
                var suggestion = getSuggestedSymbolForNonexistentModule(name, targetModule);
                return suggestion && ts.symbolName(suggestion);
            }
            function getSuggestionForNonexistentIndexSignature(objectType, expr, keyedType) {
                // check if object type has setter or getter
                function hasProp(name) {
                    var prop = getPropertyOfObjectType(objectType, name);
                    if (prop) {
                        var s = getSingleCallSignature(getTypeOfSymbol(prop));
                        return !!s && getMinArgumentCount(s) >= 1 && isTypeAssignableTo(keyedType, getTypeAtPosition(s, 0));
                    }
                    return false;
                }
                ;
                var suggestedMethod = ts.isAssignmentTarget(expr) ? "set" : "get";
                if (!hasProp(suggestedMethod)) {
                    return undefined;
                }
                var suggestion = ts.tryGetPropertyAccessOrIdentifierToString(expr.expression);
                if (suggestion === undefined) {
                    suggestion = suggestedMethod;
                }
                else {
                    suggestion += "." + suggestedMethod;
                }
                return suggestion;
            }
            /**
             * Given a name and a list of symbols whose names are *not* equal to the name, return a spelling suggestion if there is one that is close enough.
             * Names less than length 3 only check for case-insensitive equality, not levenshtein distance.
             *
             * If there is a candidate that's the same except for case, return that.
             * If there is a candidate that's within one edit of the name, return that.
             * Otherwise, return the candidate with the smallest Levenshtein distance,
             *    except for candidates:
             *      * With no name
             *      * Whose meaning doesn't match the `meaning` parameter.
             *      * Whose length differs from the target name by more than 0.34 of the length of the name.
             *      * Whose levenshtein distance is more than 0.4 of the length of the name
             *        (0.4 allows 1 substitution/transposition for every 5 characters,
             *         and 1 insertion/deletion at 3 characters)
             */
            function getSpellingSuggestionForName(name, symbols, meaning) {
                return ts.getSpellingSuggestion(name, symbols, getCandidateName);
                function getCandidateName(candidate) {
                    var candidateName = ts.symbolName(candidate);
                    if (ts.startsWith(candidateName, "\"")) {
                        return undefined;
                    }
                    if (candidate.flags & meaning) {
                        return candidateName;
                    }
                    if (candidate.flags & 2097152 /* Alias */) {
                        var alias = tryResolveAlias(candidate);
                        if (alias && alias.flags & meaning) {
                            return candidateName;
                        }
                    }
                    return undefined;
                }
            }
            function markPropertyAsReferenced(prop, nodeForCheckWriteOnly, isSelfTypeAccess) {
                var valueDeclaration = prop && (prop.flags & 106500 /* ClassMember */) && prop.valueDeclaration;
                if (!valueDeclaration) {
                    return;
                }
                var hasPrivateModifier = ts.hasEffectiveModifier(valueDeclaration, 8 /* Private */);
                var hasPrivateIdentifier = prop.valueDeclaration && ts.isNamedDeclaration(prop.valueDeclaration) && ts.isPrivateIdentifier(prop.valueDeclaration.name);
                if (!hasPrivateModifier && !hasPrivateIdentifier) {
                    return;
                }
                if (nodeForCheckWriteOnly && ts.isWriteOnlyAccess(nodeForCheckWriteOnly) && !(prop.flags & 65536 /* SetAccessor */)) {
                    return;
                }
                if (isSelfTypeAccess) {
                    // Find any FunctionLikeDeclaration because those create a new 'this' binding. But this should only matter for methods (or getters/setters).
                    var containingMethod = ts.findAncestor(nodeForCheckWriteOnly, ts.isFunctionLikeDeclaration);
                    if (containingMethod && containingMethod.symbol === prop) {
                        return;
                    }
                }
                (ts.getCheckFlags(prop) & 1 /* Instantiated */ ? getSymbolLinks(prop).target : prop).isReferenced = 67108863 /* All */;
            }
            function isSelfTypeAccess(name, parent) {
                return name.kind === 108 /* ThisKeyword */
                    || !!parent && ts.isEntityNameExpression(name) && parent === getResolvedSymbol(ts.getFirstIdentifier(name));
            }
            function isValidPropertyAccess(node, propertyName) {
                switch (node.kind) {
                    case 204 /* PropertyAccessExpression */:
                        return isValidPropertyAccessWithType(node, node.expression.kind === 106 /* SuperKeyword */, propertyName, getWidenedType(checkExpression(node.expression)));
                    case 159 /* QualifiedName */:
                        return isValidPropertyAccessWithType(node, /*isSuper*/ false, propertyName, getWidenedType(checkExpression(node.left)));
                    case 198 /* ImportType */:
                        return isValidPropertyAccessWithType(node, /*isSuper*/ false, propertyName, getTypeFromTypeNode(node));
                }
            }
            function isValidPropertyAccessForCompletions(node, type, property) {
                return isValidPropertyAccessWithType(node, node.kind === 204 /* PropertyAccessExpression */ && node.expression.kind === 106 /* SuperKeyword */, property.escapedName, type);
                // Previously we validated the 'this' type of methods but this adversely affected performance. See #31377 for more context.
            }
            function isValidPropertyAccessWithType(node, isSuper, propertyName, type) {
                if (type === errorType || isTypeAny(type)) {
                    return true;
                }
                var prop = getPropertyOfType(type, propertyName);
                if (prop) {
                    if (prop.valueDeclaration && ts.isPrivateIdentifierClassElementDeclaration(prop.valueDeclaration)) {
                        var declClass_1 = ts.getContainingClass(prop.valueDeclaration);
                        return !ts.isOptionalChain(node) && !!ts.findAncestor(node, function (parent) { return parent === declClass_1; });
                    }
                    return checkPropertyAccessibility(node, isSuper, /*writing*/ false, type, prop, /* reportError */ false);
                }
                // In js files properties of unions are allowed in completion
                return ts.isInJSFile(node) && (type.flags & 1048576 /* Union */) !== 0 && type.types.some(function (elementType) { return isValidPropertyAccessWithType(node, isSuper, propertyName, elementType); });
            }
            /**
             * Return the symbol of the for-in variable declared or referenced by the given for-in statement.
             */
            function getForInVariableSymbol(node) {
                var initializer = node.initializer;
                if (initializer.kind === 253 /* VariableDeclarationList */) {
                    var variable = initializer.declarations[0];
                    if (variable && !ts.isBindingPattern(variable.name)) {
                        return getSymbolOfNode(variable);
                    }
                }
                else if (initializer.kind === 79 /* Identifier */) {
                    return getResolvedSymbol(initializer);
                }
                return undefined;
            }
            /**
             * Return true if the given type is considered to have numeric property names.
             */
            function hasNumericPropertyNames(type) {
                return getIndexInfosOfType(type).length === 1 && !!getIndexInfoOfType(type, numberType);
            }
            /**
             * Return true if given node is an expression consisting of an identifier (possibly parenthesized)
             * that references a for-in variable for an object with numeric property names.
             */
            function isForInVariableForNumericPropertyNames(expr) {
                var e = ts.skipParentheses(expr);
                if (e.kind === 79 /* Identifier */) {
                    var symbol = getResolvedSymbol(e);
                    if (symbol.flags & 3 /* Variable */) {
                        var child = expr;
                        var node = expr.parent;
                        while (node) {
                            if (node.kind === 241 /* ForInStatement */ &&
                                child === node.statement &&
                                getForInVariableSymbol(node) === symbol &&
                                hasNumericPropertyNames(getTypeOfExpression(node.expression))) {
                                return true;
                            }
                            child = node;
                            node = node.parent;
                        }
                    }
                }
                return false;
            }
            function checkIndexedAccess(node, checkMode) {
                return node.flags & 32 /* OptionalChain */ ? checkElementAccessChain(node, checkMode) :
                    checkElementAccessExpression(node, checkNonNullExpression(node.expression), checkMode);
            }
            function checkElementAccessChain(node, checkMode) {
                var exprType = checkExpression(node.expression);
                var nonOptionalType = getOptionalExpressionType(exprType, node.expression);
                return propagateOptionalTypeMarker(checkElementAccessExpression(node, checkNonNullType(nonOptionalType, node.expression), checkMode), node, nonOptionalType !== exprType);
            }
            function checkElementAccessExpression(node, exprType, checkMode) {
                var objectType = ts.getAssignmentTargetKind(node) !== 0 /* None */ || isMethodAccessForCall(node) ? getWidenedType(exprType) : exprType;
                var indexExpression = node.argumentExpression;
                var indexType = checkExpression(indexExpression);
                if (objectType === errorType || objectType === silentNeverType) {
                    return objectType;
                }
                if (isConstEnumObjectType(objectType) && !ts.isStringLiteralLike(indexExpression)) {
                    error(indexExpression, ts.Diagnostics.A_const_enum_member_can_only_be_accessed_using_a_string_literal);
                    return errorType;
                }
                var effectiveIndexType = isForInVariableForNumericPropertyNames(indexExpression) ? numberType : indexType;
                var accessFlags = ts.isAssignmentTarget(node) ?
                    4 /* Writing */ | (isGenericObjectType(objectType) && !isThisTypeParameter(objectType) ? 2 /* NoIndexSignatures */ : 0) :
                    32 /* ExpressionPosition */;
                var indexedAccessType = getIndexedAccessTypeOrUndefined(objectType, effectiveIndexType, accessFlags, node) || errorType;
                return checkIndexedAccessIndexType(getFlowTypeOfAccessExpression(node, getNodeLinks(node).resolvedSymbol, indexedAccessType, indexExpression, checkMode), node);
            }
            function callLikeExpressionMayHaveTypeArguments(node) {
                return ts.isCallOrNewExpression(node) || ts.isTaggedTemplateExpression(node) || ts.isJsxOpeningLikeElement(node);
            }
            function resolveUntypedCall(node) {
                if (callLikeExpressionMayHaveTypeArguments(node)) {
                    // Check type arguments even though we will give an error that untyped calls may not accept type arguments.
                    // This gets us diagnostics for the type arguments and marks them as referenced.
                    ts.forEach(node.typeArguments, checkSourceElement);
                }
                if (node.kind === 208 /* TaggedTemplateExpression */) {
                    checkExpression(node.template);
                }
                else if (ts.isJsxOpeningLikeElement(node)) {
                    checkExpression(node.attributes);
                }
                else if (node.kind !== 163 /* Decorator */) {
                    ts.forEach(node.arguments, function (argument) {
                        checkExpression(argument);
                    });
                }
                return anySignature;
            }
            function resolveErrorCall(node) {
                resolveUntypedCall(node);
                return unknownSignature;
            }
            // Re-order candidate signatures into the result array. Assumes the result array to be empty.
            // The candidate list orders groups in reverse, but within a group signatures are kept in declaration order
            // A nit here is that we reorder only signatures that belong to the same symbol,
            // so order how inherited signatures are processed is still preserved.
            // interface A { (x: string): void }
            // interface B extends A { (x: 'foo'): string }
            // const b: B;
            // b('foo') // <- here overloads should be processed as [(x:'foo'): string, (x: string): void]
            function reorderCandidates(signatures, result, callChainFlags) {
                var lastParent;
                var lastSymbol;
                var cutoffIndex = 0;
                var index;
                var specializedIndex = -1;
                var spliceIndex;
                ts.Debug.assert(!result.length);
                for (var _i = 0, signatures_7 = signatures; _i < signatures_7.length; _i++) {
                    var signature = signatures_7[_i];
                    var symbol = signature.declaration && getSymbolOfNode(signature.declaration);
                    var parent = signature.declaration && signature.declaration.parent;
                    if (!lastSymbol || symbol === lastSymbol) {
                        if (lastParent && parent === lastParent) {
                            index = index + 1;
                        }
                        else {
                            lastParent = parent;
                            index = cutoffIndex;
                        }
                    }
                    else {
                        // current declaration belongs to a different symbol
                        // set cutoffIndex so re-orderings in the future won't change result set from 0 to cutoffIndex
                        index = cutoffIndex = result.length;
                        lastParent = parent;
                    }
                    lastSymbol = symbol;
                    // specialized signatures always need to be placed before non-specialized signatures regardless
                    // of the cutoff position; see GH#1133
                    if (signatureHasLiteralTypes(signature)) {
                        specializedIndex++;
                        spliceIndex = specializedIndex;
                        // The cutoff index always needs to be greater than or equal to the specialized signature index
                        // in order to prevent non-specialized signatures from being added before a specialized
                        // signature.
                        cutoffIndex++;
                    }
                    else {
                        spliceIndex = index;
                    }
                    result.splice(spliceIndex, 0, callChainFlags ? getOptionalCallSignature(signature, callChainFlags) : signature);
                }
            }
            function isSpreadArgument(arg) {
                return !!arg && (arg.kind === 223 /* SpreadElement */ || arg.kind === 230 /* SyntheticExpression */ && arg.isSpread);
            }
            function getSpreadArgumentIndex(args) {
                return ts.findIndex(args, isSpreadArgument);
            }
            function acceptsVoid(t) {
                return !!(t.flags & 16384 /* Void */);
            }
            function acceptsVoidUndefinedUnknownOrAny(t) {
                return !!(t.flags & (16384 /* Void */ | 32768 /* Undefined */ | 2 /* Unknown */ | 1 /* Any */));
            }
            function hasCorrectArity(node, args, signature, signatureHelpTrailingComma) {
                if (signatureHelpTrailingComma === void 0) {
                    signatureHelpTrailingComma = false;
                }
                var argCount;
                var callIsIncomplete = false; // In incomplete call we want to be lenient when we have too few arguments
                var effectiveParameterCount = getParameterCount(signature);
                var effectiveMinimumArguments = getMinArgumentCount(signature);
                if (node.kind === 208 /* TaggedTemplateExpression */) {
                    argCount = args.length;
                    if (node.template.kind === 221 /* TemplateExpression */) {
                        // If a tagged template expression lacks a tail literal, the call is incomplete.
                        // Specifically, a template only can end in a TemplateTail or a Missing literal.
                        var lastSpan = ts.last(node.template.templateSpans); // we should always have at least one span.
                        callIsIncomplete = ts.nodeIsMissing(lastSpan.literal) || !!lastSpan.literal.isUnterminated;
                    }
                    else {
                        // If the template didn't end in a backtick, or its beginning occurred right prior to EOF,
                        // then this might actually turn out to be a TemplateHead in the future;
                        // so we consider the call to be incomplete.
                        var templateLiteral = node.template;
                        ts.Debug.assert(templateLiteral.kind === 14 /* NoSubstitutionTemplateLiteral */);
                        callIsIncomplete = !!templateLiteral.isUnterminated;
                    }
                }
                else if (node.kind === 163 /* Decorator */) {
                    argCount = getDecoratorArgumentCount(node, signature);
                }
                else if (ts.isJsxOpeningLikeElement(node)) {
                    callIsIncomplete = node.attributes.end === node.end;
                    if (callIsIncomplete) {
                        return true;
                    }
                    argCount = effectiveMinimumArguments === 0 ? args.length : 1;
                    effectiveParameterCount = args.length === 0 ? effectiveParameterCount : 1; // class may have argumentless ctor functions - still resolve ctor and compare vs props member type
                    effectiveMinimumArguments = Math.min(effectiveMinimumArguments, 1); // sfc may specify context argument - handled by framework and not typechecked
                }
                else if (!node.arguments) {
                    // This only happens when we have something of the form: 'new C'
                    ts.Debug.assert(node.kind === 207 /* NewExpression */);
                    return getMinArgumentCount(signature) === 0;
                }
                else {
                    argCount = signatureHelpTrailingComma ? args.length + 1 : args.length;
                    // If we are missing the close parenthesis, the call is incomplete.
                    callIsIncomplete = node.arguments.end === node.end;
                    // If a spread argument is present, check that it corresponds to a rest parameter or at least that it's in the valid range.
                    var spreadArgIndex = getSpreadArgumentIndex(args);
                    if (spreadArgIndex >= 0) {
                        return spreadArgIndex >= getMinArgumentCount(signature) && (hasEffectiveRestParameter(signature) || spreadArgIndex < getParameterCount(signature));
                    }
                }
                // Too many arguments implies incorrect arity.
                if (!hasEffectiveRestParameter(signature) && argCount > effectiveParameterCount) {
                    return false;
                }
                // If the call is incomplete, we should skip the lower bound check.
                // JSX signatures can have extra parameters provided by the library which we don't check
                if (callIsIncomplete || argCount >= effectiveMinimumArguments) {
                    return true;
                }
                for (var i = argCount; i < effectiveMinimumArguments; i++) {
                    var type = getTypeAtPosition(signature, i);
                    if (filterType(type, ts.isInJSFile(node) && !strictNullChecks ? acceptsVoidUndefinedUnknownOrAny : acceptsVoid).flags & 131072 /* Never */) {
                        return false;
                    }
                }
                return true;
            }
            function hasCorrectTypeArgumentArity(signature, typeArguments) {
                // If the user supplied type arguments, but the number of type arguments does not match
                // the declared number of type parameters, the call has an incorrect arity.
                var numTypeParameters = ts.length(signature.typeParameters);
                var minTypeArgumentCount = getMinTypeArgumentCount(signature.typeParameters);
                return !ts.some(typeArguments) ||
                    (typeArguments.length >= minTypeArgumentCount && typeArguments.length <= numTypeParameters);
            }
            // If type has a single call signature and no other members, return that signature. Otherwise, return undefined.
            function getSingleCallSignature(type) {
                return getSingleSignature(type, 0 /* Call */, /*allowMembers*/ false);
            }
            function getSingleCallOrConstructSignature(type) {
                return getSingleSignature(type, 0 /* Call */, /*allowMembers*/ false) ||
                    getSingleSignature(type, 1 /* Construct */, /*allowMembers*/ false);
            }
            function getSingleSignature(type, kind, allowMembers) {
                if (type.flags & 524288 /* Object */) {
                    var resolved = resolveStructuredTypeMembers(type);
                    if (allowMembers || resolved.properties.length === 0 && resolved.indexInfos.length === 0) {
                        if (kind === 0 /* Call */ && resolved.callSignatures.length === 1 && resolved.constructSignatures.length === 0) {
                            return resolved.callSignatures[0];
                        }
                        if (kind === 1 /* Construct */ && resolved.constructSignatures.length === 1 && resolved.callSignatures.length === 0) {
                            return resolved.constructSignatures[0];
                        }
                    }
                }
                return undefined;
            }
            // Instantiate a generic signature in the context of a non-generic signature (section 3.8.5 in TypeScript spec)
            function instantiateSignatureInContextOf(signature, contextualSignature, inferenceContext, compareTypes) {
                var context = createInferenceContext(signature.typeParameters, signature, 0 /* None */, compareTypes);
                // We clone the inferenceContext to avoid fixing. For example, when the source signature is <T>(x: T) => T[] and
                // the contextual signature is (...args: A) => B, we want to infer the element type of A's constraint (say 'any')
                // for T but leave it possible to later infer '[any]' back to A.
                var restType = getEffectiveRestType(contextualSignature);
                var mapper = inferenceContext && (restType && restType.flags & 262144 /* TypeParameter */ ? inferenceContext.nonFixingMapper : inferenceContext.mapper);
                var sourceSignature = mapper ? instantiateSignature(contextualSignature, mapper) : contextualSignature;
                applyToParameterTypes(sourceSignature, signature, function (source, target) {
                    // Type parameters from outer context referenced by source type are fixed by instantiation of the source type
                    inferTypes(context.inferences, source, target);
                });
                if (!inferenceContext) {
                    applyToReturnTypes(contextualSignature, signature, function (source, target) {
                        inferTypes(context.inferences, source, target, 128 /* ReturnType */);
                    });
                }
                return getSignatureInstantiation(signature, getInferredTypes(context), ts.isInJSFile(contextualSignature.declaration));
            }
            function inferJsxTypeArguments(node, signature, checkMode, context) {
                var paramType = getEffectiveFirstArgumentForJsxSignature(signature, node);
                var checkAttrType = checkExpressionWithContextualType(node.attributes, paramType, context, checkMode);
                inferTypes(context.inferences, checkAttrType, paramType);
                return getInferredTypes(context);
            }
            function getThisArgumentType(thisArgumentNode) {
                if (!thisArgumentNode) {
                    return voidType;
                }
                var thisArgumentType = checkExpression(thisArgumentNode);
                return ts.isOptionalChainRoot(thisArgumentNode.parent) ? getNonNullableType(thisArgumentType) :
                    ts.isOptionalChain(thisArgumentNode.parent) ? removeOptionalTypeMarker(thisArgumentType) :
                        thisArgumentType;
            }
            function inferTypeArguments(node, signature, args, checkMode, context) {
                if (ts.isJsxOpeningLikeElement(node)) {
                    return inferJsxTypeArguments(node, signature, checkMode, context);
                }
                // If a contextual type is available, infer from that type to the return type of the call expression. For
                // example, given a 'function wrap<T, U>(cb: (x: T) => U): (x: T) => U' and a call expression
                // 'let f: (x: string) => number = wrap(s => s.length)', we infer from the declared type of 'f' to the
                // return type of 'wrap'.
                if (node.kind !== 163 /* Decorator */) {
                    var contextualType = getContextualType(node, ts.every(signature.typeParameters, function (p) { return !!getDefaultFromTypeParameter(p); }) ? 8 /* SkipBindingPatterns */ : 0 /* None */);
                    if (contextualType) {
                        // We clone the inference context to avoid disturbing a resolution in progress for an
                        // outer call expression. Effectively we just want a snapshot of whatever has been
                        // inferred for any outer call expression so far.
                        var outerContext = getInferenceContext(node);
                        var outerMapper = getMapperFromContext(cloneInferenceContext(outerContext, 1 /* NoDefault */));
                        var instantiatedType = instantiateType(contextualType, outerMapper);
                        // If the contextual type is a generic function type with a single call signature, we
                        // instantiate the type with its own type parameters and type arguments. This ensures that
                        // the type parameters are not erased to type any during type inference such that they can
                        // be inferred as actual types from the contextual type. For example:
                        //   declare function arrayMap<T, U>(f: (x: T) => U): (a: T[]) => U[];
                        //   const boxElements: <A>(a: A[]) => { value: A }[] = arrayMap(value => ({ value }));
                        // Above, the type of the 'value' parameter is inferred to be 'A'.
                        var contextualSignature = getSingleCallSignature(instantiatedType);
                        var inferenceSourceType = contextualSignature && contextualSignature.typeParameters ?
                            getOrCreateTypeFromSignature(getSignatureInstantiationWithoutFillingInTypeArguments(contextualSignature, contextualSignature.typeParameters)) :
                            instantiatedType;
                        var inferenceTargetType = getReturnTypeOfSignature(signature);
                        // Inferences made from return types have lower priority than all other inferences.
                        inferTypes(context.inferences, inferenceSourceType, inferenceTargetType, 128 /* ReturnType */);
                        // Create a type mapper for instantiating generic contextual types using the inferences made
                        // from the return type. We need a separate inference pass here because (a) instantiation of
                        // the source type uses the outer context's return mapper (which excludes inferences made from
                        // outer arguments), and (b) we don't want any further inferences going into this context.
                        var returnContext = createInferenceContext(signature.typeParameters, signature, context.flags);
                        var returnSourceType = instantiateType(contextualType, outerContext && outerContext.returnMapper);
                        inferTypes(returnContext.inferences, returnSourceType, inferenceTargetType);
                        context.returnMapper = ts.some(returnContext.inferences, hasInferenceCandidates) ? getMapperFromContext(cloneInferredPartOfContext(returnContext)) : undefined;
                    }
                }
                var restType = getNonArrayRestType(signature);
                var argCount = restType ? Math.min(getParameterCount(signature) - 1, args.length) : args.length;
                if (restType && restType.flags & 262144 /* TypeParameter */) {
                    var info = ts.find(context.inferences, function (info) { return info.typeParameter === restType; });
                    if (info) {
                        info.impliedArity = ts.findIndex(args, isSpreadArgument, argCount) < 0 ? args.length - argCount : undefined;
                    }
                }
                var thisType = getThisTypeOfSignature(signature);
                if (thisType) {
                    var thisArgumentNode = getThisArgumentOfCall(node);
                    inferTypes(context.inferences, getThisArgumentType(thisArgumentNode), thisType);
                }
                for (var i = 0; i < argCount; i++) {
                    var arg = args[i];
                    if (arg.kind !== 225 /* OmittedExpression */) {
                        var paramType = getTypeAtPosition(signature, i);
                        var argType = checkExpressionWithContextualType(arg, paramType, context, checkMode);
                        inferTypes(context.inferences, argType, paramType);
                    }
                }
                if (restType) {
                    var spreadType = getSpreadArgumentType(args, argCount, args.length, restType, context, checkMode);
                    inferTypes(context.inferences, spreadType, restType);
                }
                return getInferredTypes(context);
            }
            function getMutableArrayOrTupleType(type) {
                return type.flags & 1048576 /* Union */ ? mapType(type, getMutableArrayOrTupleType) :
                    type.flags & 1 /* Any */ || isMutableArrayOrTuple(getBaseConstraintOfType(type) || type) ? type :
                        isTupleType(type) ? createTupleType(getTypeArguments(type), type.target.elementFlags, /*readonly*/ false, type.target.labeledElementDeclarations) :
                            createTupleType([type], [8 /* Variadic */]);
            }
            function getSpreadArgumentType(args, index, argCount, restType, context, checkMode) {
                if (index >= argCount - 1) {
                    var arg = args[argCount - 1];
                    if (isSpreadArgument(arg)) {
                        // We are inferring from a spread expression in the last argument position, i.e. both the parameter
                        // and the argument are ...x forms.
                        return getMutableArrayOrTupleType(arg.kind === 230 /* SyntheticExpression */ ? arg.type :
                            checkExpressionWithContextualType(arg.expression, restType, context, checkMode));
                    }
                }
                var types = [];
                var flags = [];
                var names = [];
                for (var i = index; i < argCount; i++) {
                    var arg = args[i];
                    if (isSpreadArgument(arg)) {
                        var spreadType = arg.kind === 230 /* SyntheticExpression */ ? arg.type : checkExpression(arg.expression);
                        if (isArrayLikeType(spreadType)) {
                            types.push(spreadType);
                            flags.push(8 /* Variadic */);
                        }
                        else {
                            types.push(checkIteratedTypeOrElementType(33 /* Spread */, spreadType, undefinedType, arg.kind === 223 /* SpreadElement */ ? arg.expression : arg));
                            flags.push(4 /* Rest */);
                        }
                    }
                    else {
                        var contextualType = getIndexedAccessType(restType, getNumberLiteralType(i - index), 256 /* Contextual */);
                        var argType = checkExpressionWithContextualType(arg, contextualType, context, checkMode);
                        var hasPrimitiveContextualType = maybeTypeOfKind(contextualType, 131068 /* Primitive */ | 4194304 /* Index */ | 134217728 /* TemplateLiteral */ | 268435456 /* StringMapping */);
                        types.push(hasPrimitiveContextualType ? getRegularTypeOfLiteralType(argType) : getWidenedLiteralType(argType));
                        flags.push(1 /* Required */);
                    }
                    if (arg.kind === 230 /* SyntheticExpression */ && arg.tupleNameSource) {
                        names.push(arg.tupleNameSource);
                    }
                }
                return createTupleType(types, flags, /*readonly*/ false, ts.length(names) === ts.length(types) ? names : undefined);
            }
            function checkTypeArguments(signature, typeArgumentNodes, reportErrors, headMessage) {
                var isJavascript = ts.isInJSFile(signature.declaration);
                var typeParameters = signature.typeParameters;
                var typeArgumentTypes = fillMissingTypeArguments(ts.map(typeArgumentNodes, getTypeFromTypeNode), typeParameters, getMinTypeArgumentCount(typeParameters), isJavascript);
                var mapper;
                for (var i = 0; i < typeArgumentNodes.length; i++) {
                    ts.Debug.assert(typeParameters[i] !== undefined, "Should not call checkTypeArguments with too many type arguments");
                    var constraint = getConstraintOfTypeParameter(typeParameters[i]);
                    if (constraint) {
                        var errorInfo = reportErrors && headMessage ? (function () { return ts.chainDiagnosticMessages(/*details*/ undefined, ts.Diagnostics.Type_0_does_not_satisfy_the_constraint_1); }) : undefined;
                        var typeArgumentHeadMessage = headMessage || ts.Diagnostics.Type_0_does_not_satisfy_the_constraint_1;
                        if (!mapper) {
                            mapper = createTypeMapper(typeParameters, typeArgumentTypes);
                        }
                        var typeArgument = typeArgumentTypes[i];
                        if (!checkTypeAssignableTo(typeArgument, getTypeWithThisArgument(instantiateType(constraint, mapper), typeArgument), reportErrors ? typeArgumentNodes[i] : undefined, typeArgumentHeadMessage, errorInfo)) {
                            return undefined;
                        }
                    }
                }
                return typeArgumentTypes;
            }
            function getJsxReferenceKind(node) {
                if (isJsxIntrinsicIdentifier(node.tagName)) {
                    return 2 /* Mixed */;
                }
                var tagType = getApparentType(checkExpression(node.tagName));
                if (ts.length(getSignaturesOfType(tagType, 1 /* Construct */))) {
                    return 0 /* Component */;
                }
                if (ts.length(getSignaturesOfType(tagType, 0 /* Call */))) {
                    return 1 /* Function */;
                }
                return 2 /* Mixed */;
            }
            /**
             * Check if the given signature can possibly be a signature called by the JSX opening-like element.
             * @param node a JSX opening-like element we are trying to figure its call signature
             * @param signature a candidate signature we are trying whether it is a call signature
             * @param relation a relationship to check parameter and argument type
             */
            function checkApplicableSignatureForJsxOpeningLikeElement(node, signature, relation, checkMode, reportErrors, containingMessageChain, errorOutputContainer) {
                // Stateless function components can have maximum of three arguments: "props", "context", and "updater".
                // However "context" and "updater" are implicit and can't be specify by users. Only the first parameter, props,
                // can be specified by users through attributes property.
                var paramType = getEffectiveFirstArgumentForJsxSignature(signature, node);
                var attributesType = checkExpressionWithContextualType(node.attributes, paramType, /*inferenceContext*/ undefined, checkMode);
                return checkTagNameDoesNotExpectTooManyArguments() && checkTypeRelatedToAndOptionallyElaborate(attributesType, paramType, relation, reportErrors ? node.tagName : undefined, node.attributes, 
                /*headMessage*/ undefined, containingMessageChain, errorOutputContainer);
                function checkTagNameDoesNotExpectTooManyArguments() {
                    var _a;
                    if (getJsxNamespaceContainerForImplicitImport(node)) {
                        return true; // factory is implicitly jsx/jsxdev - assume it fits the bill, since we don't strongly look for the jsx/jsxs/jsxDEV factory APIs anywhere else (at least not yet)
                    }
                    var tagType = ts.isJsxOpeningElement(node) || ts.isJsxSelfClosingElement(node) && !isJsxIntrinsicIdentifier(node.tagName) ? checkExpression(node.tagName) : undefined;
                    if (!tagType) {
                        return true;
                    }
                    var tagCallSignatures = getSignaturesOfType(tagType, 0 /* Call */);
                    if (!ts.length(tagCallSignatures)) {
                        return true;
                    }
                    var factory = getJsxFactoryEntity(node);
                    if (!factory) {
                        return true;
                    }
                    var factorySymbol = resolveEntityName(factory, 111551 /* Value */, /*ignoreErrors*/ true, /*dontResolveAlias*/ false, node);
                    if (!factorySymbol) {
                        return true;
                    }
                    var factoryType = getTypeOfSymbol(factorySymbol);
                    var callSignatures = getSignaturesOfType(factoryType, 0 /* Call */);
                    if (!ts.length(callSignatures)) {
                        return true;
                    }
                    var hasFirstParamSignatures = false;
                    var maxParamCount = 0;
                    // Check that _some_ first parameter expects a FC-like thing, and that some overload of the SFC expects an acceptable number of arguments
                    for (var _i = 0, callSignatures_1 = callSignatures; _i < callSignatures_1.length; _i++) {
                        var sig = callSignatures_1[_i];
                        var firstparam = getTypeAtPosition(sig, 0);
                        var signaturesOfParam = getSignaturesOfType(firstparam, 0 /* Call */);
                        if (!ts.length(signaturesOfParam))
                            continue;
                        for (var _b = 0, signaturesOfParam_1 = signaturesOfParam; _b < signaturesOfParam_1.length; _b++) {
                            var paramSig = signaturesOfParam_1[_b];
                            hasFirstParamSignatures = true;
                            if (hasEffectiveRestParameter(paramSig)) {
                                return true; // some signature has a rest param, so function components can have an arbitrary number of arguments
                            }
                            var paramCount = getParameterCount(paramSig);
                            if (paramCount > maxParamCount) {
                                maxParamCount = paramCount;
                            }
                        }
                    }
                    if (!hasFirstParamSignatures) {
                        // Not a single signature had a first parameter which expected a signature - for back compat, and
                        // to guard against generic factories which won't have signatures directly, do not error
                        return true;
                    }
                    var absoluteMinArgCount = Infinity;
                    for (var _c = 0, tagCallSignatures_1 = tagCallSignatures; _c < tagCallSignatures_1.length; _c++) {
                        var tagSig = tagCallSignatures_1[_c];
                        var tagRequiredArgCount = getMinArgumentCount(tagSig);
                        if (tagRequiredArgCount < absoluteMinArgCount) {
                            absoluteMinArgCount = tagRequiredArgCount;
                        }
                    }
                    if (absoluteMinArgCount <= maxParamCount) {
                        return true; // some signature accepts the number of arguments the function component provides
                    }
                    if (reportErrors) {
                        var diag = ts.createDiagnosticForNode(node.tagName, ts.Diagnostics.Tag_0_expects_at_least_1_arguments_but_the_JSX_factory_2_provides_at_most_3, ts.entityNameToString(node.tagName), absoluteMinArgCount, ts.entityNameToString(factory), maxParamCount);
                        var tagNameDeclaration = (_a = getSymbolAtLocation(node.tagName)) === null || _a === void 0 ? void 0 : _a.valueDeclaration;
                        if (tagNameDeclaration) {
                            ts.addRelatedInfo(diag, ts.createDiagnosticForNode(tagNameDeclaration, ts.Diagnostics._0_is_declared_here, ts.entityNameToString(node.tagName)));
                        }
                        if (errorOutputContainer && errorOutputContainer.skipLogging) {
                            (errorOutputContainer.errors || (errorOutputContainer.errors = [])).push(diag);
                        }
                        if (!errorOutputContainer.skipLogging) {
                            diagnostics.add(diag);
                        }
                    }
                    return false;
                }
            }
            function getSignatureApplicabilityError(node, args, signature, relation, checkMode, reportErrors, containingMessageChain) {
                var errorOutputContainer = { errors: undefined, skipLogging: true };
                if (ts.isJsxOpeningLikeElement(node)) {
                    if (!checkApplicableSignatureForJsxOpeningLikeElement(node, signature, relation, checkMode, reportErrors, containingMessageChain, errorOutputContainer)) {
                        ts.Debug.assert(!reportErrors || !!errorOutputContainer.errors, "jsx should have errors when reporting errors");
                        return errorOutputContainer.errors || ts.emptyArray;
                    }
                    return undefined;
                }
                var thisType = getThisTypeOfSignature(signature);
                if (thisType && thisType !== voidType && node.kind !== 207 /* NewExpression */) {
                    // If the called expression is not of the form `x.f` or `x["f"]`, then sourceType = voidType
                    // If the signature's 'this' type is voidType, then the check is skipped -- anything is compatible.
                    // If the expression is a new expression, then the check is skipped.
                    var thisArgumentNode = getThisArgumentOfCall(node);
                    var thisArgumentType = getThisArgumentType(thisArgumentNode);
                    var errorNode = reportErrors ? (thisArgumentNode || node) : undefined;
                    var headMessage_1 = ts.Diagnostics.The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1;
                    if (!checkTypeRelatedTo(thisArgumentType, thisType, relation, errorNode, headMessage_1, containingMessageChain, errorOutputContainer)) {
                        ts.Debug.assert(!reportErrors || !!errorOutputContainer.errors, "this parameter should have errors when reporting errors");
                        return errorOutputContainer.errors || ts.emptyArray;
                    }
                }
                var headMessage = ts.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1;
                var restType = getNonArrayRestType(signature);
                var argCount = restType ? Math.min(getParameterCount(signature) - 1, args.length) : args.length;
                for (var i = 0; i < argCount; i++) {
                    var arg = args[i];
                    if (arg.kind !== 225 /* OmittedExpression */) {
                        var paramType = getTypeAtPosition(signature, i);
                        var argType = checkExpressionWithContextualType(arg, paramType, /*inferenceContext*/ undefined, checkMode);
                        // If one or more arguments are still excluded (as indicated by CheckMode.SkipContextSensitive),
                        // we obtain the regular type of any object literal arguments because we may not have inferred complete
                        // parameter types yet and therefore excess property checks may yield false positives (see #17041).
                        var checkArgType = checkMode & 4 /* SkipContextSensitive */ ? getRegularTypeOfObjectLiteral(argType) : argType;
                        if (!checkTypeRelatedToAndOptionallyElaborate(checkArgType, paramType, relation, reportErrors ? arg : undefined, arg, headMessage, containingMessageChain, errorOutputContainer)) {
                            ts.Debug.assert(!reportErrors || !!errorOutputContainer.errors, "parameter should have errors when reporting errors");
                            maybeAddMissingAwaitInfo(arg, checkArgType, paramType);
                            return errorOutputContainer.errors || ts.emptyArray;
                        }
                    }
                }
                if (restType) {
                    var spreadType = getSpreadArgumentType(args, argCount, args.length, restType, /*context*/ undefined, checkMode);
                    var restArgCount = args.length - argCount;
                    var errorNode = !reportErrors ? undefined :
                        restArgCount === 0 ? node :
                            restArgCount === 1 ? args[argCount] :
                                ts.setTextRangePosEnd(createSyntheticExpression(node, spreadType), args[argCount].pos, args[args.length - 1].end);
                    if (!checkTypeRelatedTo(spreadType, restType, relation, errorNode, headMessage, /*containingMessageChain*/ undefined, errorOutputContainer)) {
                        ts.Debug.assert(!reportErrors || !!errorOutputContainer.errors, "rest parameter should have errors when reporting errors");
                        maybeAddMissingAwaitInfo(errorNode, spreadType, restType);
                        return errorOutputContainer.errors || ts.emptyArray;
                    }
                }
                return undefined;
                function maybeAddMissingAwaitInfo(errorNode, source, target) {
                    if (errorNode && reportErrors && errorOutputContainer.errors && errorOutputContainer.errors.length) {
                        // Bail if target is Promise-like---something else is wrong
                        if (getAwaitedTypeOfPromise(target)) {
                            return;
                        }
                        var awaitedTypeOfSource = getAwaitedTypeOfPromise(source);
                        if (awaitedTypeOfSource && isTypeRelatedTo(awaitedTypeOfSource, target, relation)) {
                            ts.addRelatedInfo(errorOutputContainer.errors[0], ts.createDiagnosticForNode(errorNode, ts.Diagnostics.Did_you_forget_to_use_await));
                        }
                    }
                }
            }
            /**
             * Returns the this argument in calls like x.f(...) and x[f](...). Undefined otherwise.
             */
            function getThisArgumentOfCall(node) {
                var expression = node.kind === 206 /* CallExpression */ ? node.expression :
                    node.kind === 208 /* TaggedTemplateExpression */ ? node.tag : undefined;
                if (expression) {
                    var callee = ts.skipOuterExpressions(expression);
                    if (ts.isAccessExpression(callee)) {
                        return callee.expression;
                    }
                }
            }
            function createSyntheticExpression(parent, type, isSpread, tupleNameSource) {
                var result = ts.parseNodeFactory.createSyntheticExpression(type, isSpread, tupleNameSource);
                ts.setTextRange(result, parent);
                ts.setParent(result, parent);
                return result;
            }
            /**
             * Returns the effective arguments for an expression that works like a function invocation.
             */
            function getEffectiveCallArguments(node) {
                if (node.kind === 208 /* TaggedTemplateExpression */) {
                    var template = node.template;
                    var args_3 = [createSyntheticExpression(template, getGlobalTemplateStringsArrayType())];
                    if (template.kind === 221 /* TemplateExpression */) {
                        ts.forEach(template.templateSpans, function (span) {
                            args_3.push(span.expression);
                        });
                    }
                    return args_3;
                }
                if (node.kind === 163 /* Decorator */) {
                    return getEffectiveDecoratorArguments(node);
                }
                if (ts.isJsxOpeningLikeElement(node)) {
                    return node.attributes.properties.length > 0 || (ts.isJsxOpeningElement(node) && node.parent.children.length > 0) ? [node.attributes] : ts.emptyArray;
                }
                var args = node.arguments || ts.emptyArray;
                var spreadIndex = getSpreadArgumentIndex(args);
                if (spreadIndex >= 0) {
                    // Create synthetic arguments from spreads of tuple types.
                    var effectiveArgs_1 = args.slice(0, spreadIndex);
                    var _loop_23 = function (i) {
                        var arg = args[i];
                        // We can call checkExpressionCached because spread expressions never have a contextual type.
                        var spreadType = arg.kind === 223 /* SpreadElement */ && (flowLoopCount ? checkExpression(arg.expression) : checkExpressionCached(arg.expression));
                        if (spreadType && isTupleType(spreadType)) {
                            ts.forEach(getTypeArguments(spreadType), function (t, i) {
                                var _a;
                                var flags = spreadType.target.elementFlags[i];
                                var syntheticArg = createSyntheticExpression(arg, flags & 4 /* Rest */ ? createArrayType(t) : t, !!(flags & 12 /* Variable */), (_a = spreadType.target.labeledElementDeclarations) === null || _a === void 0 ? void 0 : _a[i]);
                                effectiveArgs_1.push(syntheticArg);
                            });
                        }
                        else {
                            effectiveArgs_1.push(arg);
                        }
                    };
                    for (var i = spreadIndex; i < args.length; i++) {
                        _loop_23(i);
                    }
                    return effectiveArgs_1;
                }
                return args;
            }
            /**
             * Returns the synthetic argument list for a decorator invocation.
             */
            function getEffectiveDecoratorArguments(node) {
                var parent = node.parent;
                var expr = node.expression;
                switch (parent.kind) {
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                        // For a class decorator, the `target` is the type of the class (e.g. the
                        // "static" or "constructor" side of the class).
                        return [
                            createSyntheticExpression(expr, getTypeOfSymbol(getSymbolOfNode(parent)))
                        ];
                    case 162 /* Parameter */:
                        // A parameter declaration decorator will have three arguments (see
                        // `ParameterDecorator` in core.d.ts).
                        var func = parent.parent;
                        return [
                            createSyntheticExpression(expr, parent.parent.kind === 169 /* Constructor */ ? getTypeOfSymbol(getSymbolOfNode(func)) : errorType),
                            createSyntheticExpression(expr, anyType),
                            createSyntheticExpression(expr, numberType)
                        ];
                    case 165 /* PropertyDeclaration */:
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        // A method or accessor declaration decorator will have two or three arguments (see
                        // `PropertyDecorator` and `MethodDecorator` in core.d.ts). If we are emitting decorators
                        // for ES3, we will only pass two arguments.
                        var hasPropDesc = parent.kind !== 165 /* PropertyDeclaration */ && languageVersion !== 0 /* ES3 */;
                        return [
                            createSyntheticExpression(expr, getParentTypeOfClassElement(parent)),
                            createSyntheticExpression(expr, getClassElementPropertyKeyType(parent)),
                            createSyntheticExpression(expr, hasPropDesc ? createTypedPropertyDescriptorType(getTypeOfNode(parent)) : anyType)
                        ];
                }
                return ts.Debug.fail();
            }
            /**
             * Returns the argument count for a decorator node that works like a function invocation.
             */
            function getDecoratorArgumentCount(node, signature) {
                switch (node.parent.kind) {
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                        return 1;
                    case 165 /* PropertyDeclaration */:
                        return 2;
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        // For ES3 or decorators with only two parameters we supply only two arguments
                        return languageVersion === 0 /* ES3 */ || signature.parameters.length <= 2 ? 2 : 3;
                    case 162 /* Parameter */:
                        return 3;
                    default:
                        return ts.Debug.fail();
                }
            }
            function getDiagnosticSpanForCallNode(node, doNotIncludeArguments) {
                var start;
                var length;
                var sourceFile = ts.getSourceFileOfNode(node);
                if (ts.isPropertyAccessExpression(node.expression)) {
                    var nameSpan = ts.getErrorSpanForNode(sourceFile, node.expression.name);
                    start = nameSpan.start;
                    length = doNotIncludeArguments ? nameSpan.length : node.end - start;
                }
                else {
                    var expressionSpan = ts.getErrorSpanForNode(sourceFile, node.expression);
                    start = expressionSpan.start;
                    length = doNotIncludeArguments ? expressionSpan.length : node.end - start;
                }
                return { start: start, length: length, sourceFile: sourceFile };
            }
            function getDiagnosticForCallNode(node, message, arg0, arg1, arg2, arg3) {
                if (ts.isCallExpression(node)) {
                    var _a = getDiagnosticSpanForCallNode(node), sourceFile = _a.sourceFile, start = _a.start, length_6 = _a.length;
                    return ts.createFileDiagnostic(sourceFile, start, length_6, message, arg0, arg1, arg2, arg3);
                }
                else {
                    return ts.createDiagnosticForNode(node, message, arg0, arg1, arg2, arg3);
                }
            }
            function isPromiseResolveArityError(node) {
                if (!ts.isCallExpression(node) || !ts.isIdentifier(node.expression))
                    return false;
                var symbol = resolveName(node.expression, node.expression.escapedText, 111551 /* Value */, undefined, undefined, false);
                var decl = symbol === null || symbol === void 0 ? void 0 : symbol.valueDeclaration;
                if (!decl || !ts.isParameter(decl) || !isFunctionExpressionOrArrowFunction(decl.parent) || !ts.isNewExpression(decl.parent.parent) || !ts.isIdentifier(decl.parent.parent.expression)) {
                    return false;
                }
                var globalPromiseSymbol = getGlobalPromiseConstructorSymbol(/*reportErrors*/ false);
                if (!globalPromiseSymbol)
                    return false;
                var constructorSymbol = getSymbolAtLocation(decl.parent.parent.expression, /*ignoreErrors*/ true);
                return constructorSymbol === globalPromiseSymbol;
            }
            function getArgumentArityError(node, signatures, args) {
                var _a;
                var spreadIndex = getSpreadArgumentIndex(args);
                if (spreadIndex > -1) {
                    return ts.createDiagnosticForNode(args[spreadIndex], ts.Diagnostics.A_spread_argument_must_either_have_a_tuple_type_or_be_passed_to_a_rest_parameter);
                }
                var min = Number.POSITIVE_INFINITY; // smallest parameter count
                var max = Number.NEGATIVE_INFINITY; // largest parameter count
                var maxBelow = Number.NEGATIVE_INFINITY; // largest parameter count that is smaller than the number of arguments
                var minAbove = Number.POSITIVE_INFINITY; // smallest parameter count that is larger than the number of arguments
                var closestSignature;
                for (var _i = 0, signatures_8 = signatures; _i < signatures_8.length; _i++) {
                    var sig = signatures_8[_i];
                    var minParameter = getMinArgumentCount(sig);
                    var maxParameter = getParameterCount(sig);
                    // smallest/largest parameter counts
                    if (minParameter < min) {
                        min = minParameter;
                        closestSignature = sig;
                    }
                    max = Math.max(max, maxParameter);
                    // shortest parameter count *longer than the call*/longest parameter count *shorter than the call*
                    if (minParameter < args.length && minParameter > maxBelow)
                        maxBelow = minParameter;
                    if (args.length < maxParameter && maxParameter < minAbove)
                        minAbove = maxParameter;
                }
                var hasRestParameter = ts.some(signatures, hasEffectiveRestParameter);
                var parameterRange = hasRestParameter ? min
                    : min < max ? min + "-" + max
                        : min;
                var error = hasRestParameter ? ts.Diagnostics.Expected_at_least_0_arguments_but_got_1
                    : parameterRange === 1 && args.length === 0 && isPromiseResolveArityError(node) ? ts.Diagnostics.Expected_0_arguments_but_got_1_Did_you_forget_to_include_void_in_your_type_argument_to_Promise
                        : ts.Diagnostics.Expected_0_arguments_but_got_1;
                if (min < args.length && args.length < max) {
                    // between min and max, but with no matching overload
                    return getDiagnosticForCallNode(node, ts.Diagnostics.No_overload_expects_0_arguments_but_overloads_do_exist_that_expect_either_1_or_2_arguments, args.length, maxBelow, minAbove);
                }
                else if (args.length < min) {
                    // too short: put the error span on the call expression, not any of the args
                    var diagnostic = getDiagnosticForCallNode(node, error, parameterRange, args.length);
                    var parameter = (_a = closestSignature === null || closestSignature === void 0 ? void 0 : closestSignature.declaration) === null || _a === void 0 ? void 0 : _a.parameters[closestSignature.thisParameter ? args.length + 1 : args.length];
                    if (parameter) {
                        var parameterError = ts.createDiagnosticForNode(parameter, ts.isBindingPattern(parameter.name) ? ts.Diagnostics.An_argument_matching_this_binding_pattern_was_not_provided
                            : ts.isRestParameter(parameter) ? ts.Diagnostics.Arguments_for_the_rest_parameter_0_were_not_provided
                                : ts.Diagnostics.An_argument_for_0_was_not_provided, !parameter.name ? args.length : !ts.isBindingPattern(parameter.name) ? ts.idText(ts.getFirstIdentifier(parameter.name)) : undefined);
                        return ts.addRelatedInfo(diagnostic, parameterError);
                    }
                    return diagnostic;
                }
                else {
                    // too long; error goes on the excess parameters
                    var errorSpan = ts.factory.createNodeArray(args.slice(max));
                    var pos = ts.first(errorSpan).pos;
                    var end = ts.last(errorSpan).end;
                    if (end === pos) {
                        end++;
                    }
                    ts.setTextRangePosEnd(errorSpan, pos, end);
                    return ts.createDiagnosticForNodeArray(ts.getSourceFileOfNode(node), errorSpan, error, parameterRange, args.length);
                }
            }
            function getTypeArgumentArityError(node, signatures, typeArguments) {
                var argCount = typeArguments.length;
                // No overloads exist
                if (signatures.length === 1) {
                    var sig = signatures[0];
                    var min_1 = getMinTypeArgumentCount(sig.typeParameters);
                    var max = ts.length(sig.typeParameters);
                    return ts.createDiagnosticForNodeArray(ts.getSourceFileOfNode(node), typeArguments, ts.Diagnostics.Expected_0_type_arguments_but_got_1, min_1 < max ? min_1 + "-" + max : min_1, argCount);
                }
                // Overloads exist
                var belowArgCount = -Infinity;
                var aboveArgCount = Infinity;
                for (var _i = 0, signatures_9 = signatures; _i < signatures_9.length; _i++) {
                    var sig = signatures_9[_i];
                    var min_2 = getMinTypeArgumentCount(sig.typeParameters);
                    var max = ts.length(sig.typeParameters);
                    if (min_2 > argCount) {
                        aboveArgCount = Math.min(aboveArgCount, min_2);
                    }
                    else if (max < argCount) {
                        belowArgCount = Math.max(belowArgCount, max);
                    }
                }
                if (belowArgCount !== -Infinity && aboveArgCount !== Infinity) {
                    return ts.createDiagnosticForNodeArray(ts.getSourceFileOfNode(node), typeArguments, ts.Diagnostics.No_overload_expects_0_type_arguments_but_overloads_do_exist_that_expect_either_1_or_2_type_arguments, argCount, belowArgCount, aboveArgCount);
                }
                return ts.createDiagnosticForNodeArray(ts.getSourceFileOfNode(node), typeArguments, ts.Diagnostics.Expected_0_type_arguments_but_got_1, belowArgCount === -Infinity ? aboveArgCount : belowArgCount, argCount);
            }
            function resolveCall(node, signatures, candidatesOutArray, checkMode, callChainFlags, fallbackError) {
                var isTaggedTemplate = node.kind === 208 /* TaggedTemplateExpression */;
                var isDecorator = node.kind === 163 /* Decorator */;
                var isJsxOpeningOrSelfClosingElement = ts.isJsxOpeningLikeElement(node);
                var reportErrors = !candidatesOutArray && produceDiagnostics;
                var typeArguments;
                if (!isDecorator) {
                    typeArguments = node.typeArguments;
                    // We already perform checking on the type arguments on the class declaration itself.
                    if (isTaggedTemplate || isJsxOpeningOrSelfClosingElement || node.expression.kind !== 106 /* SuperKeyword */) {
                        ts.forEach(typeArguments, checkSourceElement);
                    }
                }
                var candidates = candidatesOutArray || [];
                // reorderCandidates fills up the candidates array directly
                reorderCandidates(signatures, candidates, callChainFlags);
                if (!candidates.length) {
                    if (reportErrors) {
                        diagnostics.add(getDiagnosticForCallNode(node, ts.Diagnostics.Call_target_does_not_contain_any_signatures));
                    }
                    return resolveErrorCall(node);
                }
                var args = getEffectiveCallArguments(node);
                // The excludeArgument array contains true for each context sensitive argument (an argument
                // is context sensitive it is susceptible to a one-time permanent contextual typing).
                //
                // The idea is that we will perform type argument inference & assignability checking once
                // without using the susceptible parameters that are functions, and once more for those
                // parameters, contextually typing each as we go along.
                //
                // For a tagged template, then the first argument be 'undefined' if necessary because it
                // represents a TemplateStringsArray.
                //
                // For a decorator, no arguments are susceptible to contextual typing due to the fact
                // decorators are applied to a declaration by the emitter, and not to an expression.
                var isSingleNonGenericCandidate = candidates.length === 1 && !candidates[0].typeParameters;
                var argCheckMode = !isDecorator && !isSingleNonGenericCandidate && ts.some(args, isContextSensitive) ? 4 /* SkipContextSensitive */ : 0 /* Normal */;
                // The following variables are captured and modified by calls to chooseOverload.
                // If overload resolution or type argument inference fails, we want to report the
                // best error possible. The best error is one which says that an argument was not
                // assignable to a parameter. This implies that everything else about the overload
                // was fine. So if there is any overload that is only incorrect because of an
                // argument, we will report an error on that one.
                //
                //     function foo(s: string): void;
                //     function foo(n: number): void; // Report argument error on this overload
                //     function foo(): void;
                //     foo(true);
                //
                // If none of the overloads even made it that far, there are two possibilities.
                // There was a problem with type arguments for some overload, in which case
                // report an error on that. Or none of the overloads even had correct arity,
                // in which case give an arity error.
                //
                //     function foo<T extends string>(x: T): void; // Report type argument error
                //     function foo(): void;
                //     foo<number>(0);
                //
                var candidatesForArgumentError;
                var candidateForArgumentArityError;
                var candidateForTypeArgumentError;
                var result;
                // If we are in signature help, a trailing comma indicates that we intend to provide another argument,
                // so we will only accept overloads with arity at least 1 higher than the current number of provided arguments.
                var signatureHelpTrailingComma = !!(checkMode & 16 /* IsForSignatureHelp */) && node.kind === 206 /* CallExpression */ && node.arguments.hasTrailingComma;
                // Section 4.12.1:
                // if the candidate list contains one or more signatures for which the type of each argument
                // expression is a subtype of each corresponding parameter type, the return type of the first
                // of those signatures becomes the return type of the function call.
                // Otherwise, the return type of the first signature in the candidate list becomes the return
                // type of the function call.
                //
                // Whether the call is an error is determined by assignability of the arguments. The subtype pass
                // is just important for choosing the best signature. So in the case where there is only one
                // signature, the subtype pass is useless. So skipping it is an optimization.
                if (candidates.length > 1) {
                    result = chooseOverload(candidates, subtypeRelation, isSingleNonGenericCandidate, signatureHelpTrailingComma);
                }
                if (!result) {
                    result = chooseOverload(candidates, assignableRelation, isSingleNonGenericCandidate, signatureHelpTrailingComma);
                }
                if (result) {
                    return result;
                }
                // No signatures were applicable. Now report errors based on the last applicable signature with
                // no arguments excluded from assignability checks.
                // If candidate is undefined, it means that no candidates had a suitable arity. In that case,
                // skip the checkApplicableSignature check.
                if (reportErrors) {
                    if (candidatesForArgumentError) {
                        if (candidatesForArgumentError.length === 1 || candidatesForArgumentError.length > 3) {
                            var last_2 = candidatesForArgumentError[candidatesForArgumentError.length - 1];
                            var chain_1;
                            if (candidatesForArgumentError.length > 3) {
                                chain_1 = ts.chainDiagnosticMessages(chain_1, ts.Diagnostics.The_last_overload_gave_the_following_error);
                                chain_1 = ts.chainDiagnosticMessages(chain_1, ts.Diagnostics.No_overload_matches_this_call);
                            }
                            var diags = getSignatureApplicabilityError(node, args, last_2, assignableRelation, 0 /* Normal */, /*reportErrors*/ true, function () { return chain_1; });
                            if (diags) {
                                for (var _i = 0, diags_1 = diags; _i < diags_1.length; _i++) {
                                    var d = diags_1[_i];
                                    if (last_2.declaration && candidatesForArgumentError.length > 3) {
                                        ts.addRelatedInfo(d, ts.createDiagnosticForNode(last_2.declaration, ts.Diagnostics.The_last_overload_is_declared_here));
                                    }
                                    addImplementationSuccessElaboration(last_2, d);
                                    diagnostics.add(d);
                                }
                            }
                            else {
                                ts.Debug.fail("No error for last overload signature");
                            }
                        }
                        else {
                            var allDiagnostics = [];
                            var max = 0;
                            var min_3 = Number.MAX_VALUE;
                            var minIndex = 0;
                            var i_1 = 0;
                            var _loop_24 = function (c) {
                                var chain_2 = function () { return ts.chainDiagnosticMessages(/*details*/ undefined, ts.Diagnostics.Overload_0_of_1_2_gave_the_following_error, i_1 + 1, candidates.length, signatureToString(c)); };
                                var diags_2 = getSignatureApplicabilityError(node, args, c, assignableRelation, 0 /* Normal */, /*reportErrors*/ true, chain_2);
                                if (diags_2) {
                                    if (diags_2.length <= min_3) {
                                        min_3 = diags_2.length;
                                        minIndex = i_1;
                                    }
                                    max = Math.max(max, diags_2.length);
                                    allDiagnostics.push(diags_2);
                                }
                                else {
                                    ts.Debug.fail("No error for 3 or fewer overload signatures");
                                }
                                i_1++;
                            };
                            for (var _a = 0, candidatesForArgumentError_1 = candidatesForArgumentError; _a < candidatesForArgumentError_1.length; _a++) {
                                var c = candidatesForArgumentError_1[_a];
                                _loop_24(c);
                            }
                            var diags_3 = max > 1 ? allDiagnostics[minIndex] : ts.flatten(allDiagnostics);
                            ts.Debug.assert(diags_3.length > 0, "No errors reported for 3 or fewer overload signatures");
                            var chain = ts.chainDiagnosticMessages(ts.map(diags_3, function (d) { return typeof d.messageText === "string" ? d : d.messageText; }), ts.Diagnostics.No_overload_matches_this_call);
                            // The below is a spread to guarantee we get a new (mutable) array - our `flatMap` helper tries to do "smart" optimizations where it reuses input
                            // arrays and the emptyArray singleton where possible, which is decidedly not what we want while we're still constructing this diagnostic
                            var related = __spreadArray([], ts.flatMap(diags_3, function (d) { return d.relatedInformation; }), true);
                            var diag = void 0;
                            if (ts.every(diags_3, function (d) { return d.start === diags_3[0].start && d.length === diags_3[0].length && d.file === diags_3[0].file; })) {
                                var _b = diags_3[0], file = _b.file, start = _b.start, length_7 = _b.length;
                                diag = { file: file, start: start, length: length_7, code: chain.code, category: chain.category, messageText: chain, relatedInformation: related };
                            }
                            else {
                                diag = ts.createDiagnosticForNodeFromMessageChain(node, chain, related);
                            }
                            addImplementationSuccessElaboration(candidatesForArgumentError[0], diag);
                            diagnostics.add(diag);
                        }
                    }
                    else if (candidateForArgumentArityError) {
                        diagnostics.add(getArgumentArityError(node, [candidateForArgumentArityError], args));
                    }
                    else if (candidateForTypeArgumentError) {
                        checkTypeArguments(candidateForTypeArgumentError, node.typeArguments, /*reportErrors*/ true, fallbackError);
                    }
                    else {
                        var signaturesWithCorrectTypeArgumentArity = ts.filter(signatures, function (s) { return hasCorrectTypeArgumentArity(s, typeArguments); });
                        if (signaturesWithCorrectTypeArgumentArity.length === 0) {
                            diagnostics.add(getTypeArgumentArityError(node, signatures, typeArguments));
                        }
                        else if (!isDecorator) {
                            diagnostics.add(getArgumentArityError(node, signaturesWithCorrectTypeArgumentArity, args));
                        }
                        else if (fallbackError) {
                            diagnostics.add(getDiagnosticForCallNode(node, fallbackError));
                        }
                    }
                }
                return getCandidateForOverloadFailure(node, candidates, args, !!candidatesOutArray);
                function addImplementationSuccessElaboration(failed, diagnostic) {
                    var _a, _b;
                    var oldCandidatesForArgumentError = candidatesForArgumentError;
                    var oldCandidateForArgumentArityError = candidateForArgumentArityError;
                    var oldCandidateForTypeArgumentError = candidateForTypeArgumentError;
                    var failedSignatureDeclarations = ((_b = (_a = failed.declaration) === null || _a === void 0 ? void 0 : _a.symbol) === null || _b === void 0 ? void 0 : _b.declarations) || ts.emptyArray;
                    var isOverload = failedSignatureDeclarations.length > 1;
                    var implDecl = isOverload ? ts.find(failedSignatureDeclarations, function (d) { return ts.isFunctionLikeDeclaration(d) && ts.nodeIsPresent(d.body); }) : undefined;
                    if (implDecl) {
                        var candidate = getSignatureFromDeclaration(implDecl);
                        var isSingleNonGenericCandidate_1 = !candidate.typeParameters;
                        if (chooseOverload([candidate], assignableRelation, isSingleNonGenericCandidate_1)) {
                            ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(implDecl, ts.Diagnostics.The_call_would_have_succeeded_against_this_implementation_but_implementation_signatures_of_overloads_are_not_externally_visible));
                        }
                    }
                    candidatesForArgumentError = oldCandidatesForArgumentError;
                    candidateForArgumentArityError = oldCandidateForArgumentArityError;
                    candidateForTypeArgumentError = oldCandidateForTypeArgumentError;
                }
                function chooseOverload(candidates, relation, isSingleNonGenericCandidate, signatureHelpTrailingComma) {
                    if (signatureHelpTrailingComma === void 0) {
                        signatureHelpTrailingComma = false;
                    }
                    candidatesForArgumentError = undefined;
                    candidateForArgumentArityError = undefined;
                    candidateForTypeArgumentError = undefined;
                    if (isSingleNonGenericCandidate) {
                        var candidate = candidates[0];
                        if (ts.some(typeArguments) || !hasCorrectArity(node, args, candidate, signatureHelpTrailingComma)) {
                            return undefined;
                        }
                        if (getSignatureApplicabilityError(node, args, candidate, relation, 0 /* Normal */, /*reportErrors*/ false, /*containingMessageChain*/ undefined)) {
                            candidatesForArgumentError = [candidate];
                            return undefined;
                        }
                        return candidate;
                    }
                    for (var candidateIndex = 0; candidateIndex < candidates.length; candidateIndex++) {
                        var candidate = candidates[candidateIndex];
                        if (!hasCorrectTypeArgumentArity(candidate, typeArguments) || !hasCorrectArity(node, args, candidate, signatureHelpTrailingComma)) {
                            continue;
                        }
                        var checkCandidate = void 0;
                        var inferenceContext = void 0;
                        if (candidate.typeParameters) {
                            var typeArgumentTypes = void 0;
                            if (ts.some(typeArguments)) {
                                typeArgumentTypes = checkTypeArguments(candidate, typeArguments, /*reportErrors*/ false);
                                if (!typeArgumentTypes) {
                                    candidateForTypeArgumentError = candidate;
                                    continue;
                                }
                            }
                            else {
                                inferenceContext = createInferenceContext(candidate.typeParameters, candidate, /*flags*/ ts.isInJSFile(node) ? 2 /* AnyDefault */ : 0 /* None */);
                                typeArgumentTypes = inferTypeArguments(node, candidate, args, argCheckMode | 8 /* SkipGenericFunctions */, inferenceContext);
                                argCheckMode |= inferenceContext.flags & 4 /* SkippedGenericFunction */ ? 8 /* SkipGenericFunctions */ : 0 /* Normal */;
                            }
                            checkCandidate = getSignatureInstantiation(candidate, typeArgumentTypes, ts.isInJSFile(candidate.declaration), inferenceContext && inferenceContext.inferredTypeParameters);
                            // If the original signature has a generic rest type, instantiation may produce a
                            // signature with different arity and we need to perform another arity check.
                            if (getNonArrayRestType(candidate) && !hasCorrectArity(node, args, checkCandidate, signatureHelpTrailingComma)) {
                                candidateForArgumentArityError = checkCandidate;
                                continue;
                            }
                        }
                        else {
                            checkCandidate = candidate;
                        }
                        if (getSignatureApplicabilityError(node, args, checkCandidate, relation, argCheckMode, /*reportErrors*/ false, /*containingMessageChain*/ undefined)) {
                            // Give preference to error candidates that have no rest parameters (as they are more specific)
                            (candidatesForArgumentError || (candidatesForArgumentError = [])).push(checkCandidate);
                            continue;
                        }
                        if (argCheckMode) {
                            // If one or more context sensitive arguments were excluded, we start including
                            // them now (and keeping do so for any subsequent candidates) and perform a second
                            // round of type inference and applicability checking for this particular candidate.
                            argCheckMode = 0 /* Normal */;
                            if (inferenceContext) {
                                var typeArgumentTypes = inferTypeArguments(node, candidate, args, argCheckMode, inferenceContext);
                                checkCandidate = getSignatureInstantiation(candidate, typeArgumentTypes, ts.isInJSFile(candidate.declaration), inferenceContext && inferenceContext.inferredTypeParameters);
                                // If the original signature has a generic rest type, instantiation may produce a
                                // signature with different arity and we need to perform another arity check.
                                if (getNonArrayRestType(candidate) && !hasCorrectArity(node, args, checkCandidate, signatureHelpTrailingComma)) {
                                    candidateForArgumentArityError = checkCandidate;
                                    continue;
                                }
                            }
                            if (getSignatureApplicabilityError(node, args, checkCandidate, relation, argCheckMode, /*reportErrors*/ false, /*containingMessageChain*/ undefined)) {
                                // Give preference to error candidates that have no rest parameters (as they are more specific)
                                (candidatesForArgumentError || (candidatesForArgumentError = [])).push(checkCandidate);
                                continue;
                            }
                        }
                        candidates[candidateIndex] = checkCandidate;
                        return checkCandidate;
                    }
                    return undefined;
                }
            }
            // No signature was applicable. We have already reported the errors for the invalid signature.
            function getCandidateForOverloadFailure(node, candidates, args, hasCandidatesOutArray) {
                ts.Debug.assert(candidates.length > 0); // Else should not have called this.
                checkNodeDeferred(node);
                // Normally we will combine overloads. Skip this if they have type parameters since that's hard to combine.
                // Don't do this if there is a `candidatesOutArray`,
                // because then we want the chosen best candidate to be one of the overloads, not a combination.
                return hasCandidatesOutArray || candidates.length === 1 || candidates.some(function (c) { return !!c.typeParameters; })
                    ? pickLongestCandidateSignature(node, candidates, args)
                    : createUnionOfSignaturesForOverloadFailure(candidates);
            }
            function createUnionOfSignaturesForOverloadFailure(candidates) {
                var thisParameters = ts.mapDefined(candidates, function (c) { return c.thisParameter; });
                var thisParameter;
                if (thisParameters.length) {
                    thisParameter = createCombinedSymbolFromTypes(thisParameters, thisParameters.map(getTypeOfParameter));
                }
                var _a = ts.minAndMax(candidates, getNumNonRestParameters), minArgumentCount = _a.min, maxNonRestParam = _a.max;
                var parameters = [];
                var _loop_25 = function (i) {
                    var symbols = ts.mapDefined(candidates, function (s) {
                        return signatureHasRestParameter(s) ?
                            i < s.parameters.length - 1 ? s.parameters[i] : ts.last(s.parameters) :
                            i < s.parameters.length ? s.parameters[i] : undefined;
                    });
                    ts.Debug.assert(symbols.length !== 0);
                    parameters.push(createCombinedSymbolFromTypes(symbols, ts.mapDefined(candidates, function (candidate) { return tryGetTypeAtPosition(candidate, i); })));
                };
                for (var i = 0; i < maxNonRestParam; i++) {
                    _loop_25(i);
                }
                var restParameterSymbols = ts.mapDefined(candidates, function (c) { return signatureHasRestParameter(c) ? ts.last(c.parameters) : undefined; });
                var flags = 0 /* None */;
                if (restParameterSymbols.length !== 0) {
                    var type = createArrayType(getUnionType(ts.mapDefined(candidates, tryGetRestTypeOfSignature), 2 /* Subtype */));
                    parameters.push(createCombinedSymbolForOverloadFailure(restParameterSymbols, type));
                    flags |= 1 /* HasRestParameter */;
                }
                if (candidates.some(signatureHasLiteralTypes)) {
                    flags |= 2 /* HasLiteralTypes */;
                }
                return createSignature(candidates[0].declaration, 
                /*typeParameters*/ undefined, // Before calling this we tested for `!candidates.some(c => !!c.typeParameters)`.
                thisParameter, parameters, 
                /*resolvedReturnType*/ getIntersectionType(candidates.map(getReturnTypeOfSignature)), 
                /*typePredicate*/ undefined, minArgumentCount, flags);
            }
            function getNumNonRestParameters(signature) {
                var numParams = signature.parameters.length;
                return signatureHasRestParameter(signature) ? numParams - 1 : numParams;
            }
            function createCombinedSymbolFromTypes(sources, types) {
                return createCombinedSymbolForOverloadFailure(sources, getUnionType(types, 2 /* Subtype */));
            }
            function createCombinedSymbolForOverloadFailure(sources, type) {
                // This function is currently only used for erroneous overloads, so it's good enough to just use the first source.
                return createSymbolWithType(ts.first(sources), type);
            }
            function pickLongestCandidateSignature(node, candidates, args) {
                // Pick the longest signature. This way we can get a contextual type for cases like:
                //     declare function f(a: { xa: number; xb: number; }, b: number);
                //     f({ |
                // Also, use explicitly-supplied type arguments if they are provided, so we can get a contextual signature in cases like:
                //     declare function f<T>(k: keyof T);
                //     f<Foo>("
                var bestIndex = getLongestCandidateIndex(candidates, apparentArgumentCount === undefined ? args.length : apparentArgumentCount);
                var candidate = candidates[bestIndex];
                var typeParameters = candidate.typeParameters;
                if (!typeParameters) {
                    return candidate;
                }
                var typeArgumentNodes = callLikeExpressionMayHaveTypeArguments(node) ? node.typeArguments : undefined;
                var instantiated = typeArgumentNodes
                    ? createSignatureInstantiation(candidate, getTypeArgumentsFromNodes(typeArgumentNodes, typeParameters, ts.isInJSFile(node)))
                    : inferSignatureInstantiationForOverloadFailure(node, typeParameters, candidate, args);
                candidates[bestIndex] = instantiated;
                return instantiated;
            }
            function getTypeArgumentsFromNodes(typeArgumentNodes, typeParameters, isJs) {
                var typeArguments = typeArgumentNodes.map(getTypeOfNode);
                while (typeArguments.length > typeParameters.length) {
                    typeArguments.pop();
                }
                while (typeArguments.length < typeParameters.length) {
                    typeArguments.push(getConstraintOfTypeParameter(typeParameters[typeArguments.length]) || getDefaultTypeArgumentType(isJs));
                }
                return typeArguments;
            }
            function inferSignatureInstantiationForOverloadFailure(node, typeParameters, candidate, args) {
                var inferenceContext = createInferenceContext(typeParameters, candidate, /*flags*/ ts.isInJSFile(node) ? 2 /* AnyDefault */ : 0 /* None */);
                var typeArgumentTypes = inferTypeArguments(node, candidate, args, 4 /* SkipContextSensitive */ | 8 /* SkipGenericFunctions */, inferenceContext);
                return createSignatureInstantiation(candidate, typeArgumentTypes);
            }
            function getLongestCandidateIndex(candidates, argsCount) {
                var maxParamsIndex = -1;
                var maxParams = -1;
                for (var i = 0; i < candidates.length; i++) {
                    var candidate = candidates[i];
                    var paramCount = getParameterCount(candidate);
                    if (hasEffectiveRestParameter(candidate) || paramCount >= argsCount) {
                        return i;
                    }
                    if (paramCount > maxParams) {
                        maxParams = paramCount;
                        maxParamsIndex = i;
                    }
                }
                return maxParamsIndex;
            }
            function resolveCallExpression(node, candidatesOutArray, checkMode) {
                if (node.expression.kind === 106 /* SuperKeyword */) {
                    var superType = checkSuperExpression(node.expression);
                    if (isTypeAny(superType)) {
                        for (var _i = 0, _a = node.arguments; _i < _a.length; _i++) {
                            var arg = _a[_i];
                            checkExpression(arg); // Still visit arguments so they get marked for visibility, etc
                        }
                        return anySignature;
                    }
                    if (superType !== errorType) {
                        // In super call, the candidate signatures are the matching arity signatures of the base constructor function instantiated
                        // with the type arguments specified in the extends clause.
                        var baseTypeNode = ts.getEffectiveBaseTypeNode(ts.getContainingClass(node));
                        if (baseTypeNode) {
                            var baseConstructors = getInstantiatedConstructorsForTypeArguments(superType, baseTypeNode.typeArguments, baseTypeNode);
                            return resolveCall(node, baseConstructors, candidatesOutArray, checkMode, 0 /* None */);
                        }
                    }
                    return resolveUntypedCall(node);
                }
                var callChainFlags;
                var funcType = checkExpression(node.expression);
                if (ts.isCallChain(node)) {
                    var nonOptionalType = getOptionalExpressionType(funcType, node.expression);
                    callChainFlags = nonOptionalType === funcType ? 0 /* None */ :
                        ts.isOutermostOptionalChain(node) ? 16 /* IsOuterCallChain */ :
                            8 /* IsInnerCallChain */;
                    funcType = nonOptionalType;
                }
                else {
                    callChainFlags = 0 /* None */;
                }
                funcType = checkNonNullTypeWithReporter(funcType, node.expression, reportCannotInvokePossiblyNullOrUndefinedError);
                if (funcType === silentNeverType) {
                    return silentNeverSignature;
                }
                var apparentType = getApparentType(funcType);
                if (apparentType === errorType) {
                    // Another error has already been reported
                    return resolveErrorCall(node);
                }
                // Technically, this signatures list may be incomplete. We are taking the apparent type,
                // but we are not including call signatures that may have been added to the Object or
                // Function interface, since they have none by default. This is a bit of a leap of faith
                // that the user will not add any.
                var callSignatures = getSignaturesOfType(apparentType, 0 /* Call */);
                var numConstructSignatures = getSignaturesOfType(apparentType, 1 /* Construct */).length;
                // TS 1.0 Spec: 4.12
                // In an untyped function call no TypeArgs are permitted, Args can be any argument list, no contextual
                // types are provided for the argument expressions, and the result is always of type Any.
                if (isUntypedFunctionCall(funcType, apparentType, callSignatures.length, numConstructSignatures)) {
                    // The unknownType indicates that an error already occurred (and was reported).  No
                    // need to report another error in this case.
                    if (funcType !== errorType && node.typeArguments) {
                        error(node, ts.Diagnostics.Untyped_function_calls_may_not_accept_type_arguments);
                    }
                    return resolveUntypedCall(node);
                }
                // If FuncExpr's apparent type(section 3.8.1) is a function type, the call is a typed function call.
                // TypeScript employs overload resolution in typed function calls in order to support functions
                // with multiple call signatures.
                if (!callSignatures.length) {
                    if (numConstructSignatures) {
                        error(node, ts.Diagnostics.Value_of_type_0_is_not_callable_Did_you_mean_to_include_new, typeToString(funcType));
                    }
                    else {
                        var relatedInformation = void 0;
                        if (node.arguments.length === 1) {
                            var text = ts.getSourceFileOfNode(node).text;
                            if (ts.isLineBreak(text.charCodeAt(ts.skipTrivia(text, node.expression.end, /* stopAfterLineBreak */ true) - 1))) {
                                relatedInformation = ts.createDiagnosticForNode(node.expression, ts.Diagnostics.Are_you_missing_a_semicolon);
                            }
                        }
                        invocationError(node.expression, apparentType, 0 /* Call */, relatedInformation);
                    }
                    return resolveErrorCall(node);
                }
                // When a call to a generic function is an argument to an outer call to a generic function for which
                // inference is in process, we have a choice to make. If the inner call relies on inferences made from
                // its contextual type to its return type, deferring the inner call processing allows the best possible
                // contextual type to accumulate. But if the outer call relies on inferences made from the return type of
                // the inner call, the inner call should be processed early. There's no sure way to know which choice is
                // right (only a full unification algorithm can determine that), so we resort to the following heuristic:
                // If no type arguments are specified in the inner call and at least one call signature is generic and
                // returns a function type, we choose to defer processing. This narrowly permits function composition
                // operators to flow inferences through return types, but otherwise processes calls right away. We
                // use the resolvingSignature singleton to indicate that we deferred processing. This result will be
                // propagated out and eventually turned into nonInferrableType (a type that is assignable to anything and
                // from which we never make inferences).
                if (checkMode & 8 /* SkipGenericFunctions */ && !node.typeArguments && callSignatures.some(isGenericFunctionReturningFunction)) {
                    skippedGenericFunction(node, checkMode);
                    return resolvingSignature;
                }
                // If the function is explicitly marked with `@class`, then it must be constructed.
                if (callSignatures.some(function (sig) { return ts.isInJSFile(sig.declaration) && !!ts.getJSDocClassTag(sig.declaration); })) {
                    error(node, ts.Diagnostics.Value_of_type_0_is_not_callable_Did_you_mean_to_include_new, typeToString(funcType));
                    return resolveErrorCall(node);
                }
                return resolveCall(node, callSignatures, candidatesOutArray, checkMode, callChainFlags);
            }
            function isGenericFunctionReturningFunction(signature) {
                return !!(signature.typeParameters && isFunctionType(getReturnTypeOfSignature(signature)));
            }
            /**
             * TS 1.0 spec: 4.12
             * If FuncExpr is of type Any, or of an object type that has no call or construct signatures
             * but is a subtype of the Function interface, the call is an untyped function call.
             */
            function isUntypedFunctionCall(funcType, apparentFuncType, numCallSignatures, numConstructSignatures) {
                // We exclude union types because we may have a union of function types that happen to have no common signatures.
                return isTypeAny(funcType) || isTypeAny(apparentFuncType) && !!(funcType.flags & 262144 /* TypeParameter */) ||
                    !numCallSignatures && !numConstructSignatures && !(apparentFuncType.flags & 1048576 /* Union */) && !(getReducedType(apparentFuncType).flags & 131072 /* Never */) && isTypeAssignableTo(funcType, globalFunctionType);
            }
            function resolveNewExpression(node, candidatesOutArray, checkMode) {
                if (node.arguments && languageVersion < 1 /* ES5 */) {
                    var spreadIndex = getSpreadArgumentIndex(node.arguments);
                    if (spreadIndex >= 0) {
                        error(node.arguments[spreadIndex], ts.Diagnostics.Spread_operator_in_new_expressions_is_only_available_when_targeting_ECMAScript_5_and_higher);
                    }
                }
                var expressionType = checkNonNullExpression(node.expression);
                if (expressionType === silentNeverType) {
                    return silentNeverSignature;
                }
                // If expressionType's apparent type(section 3.8.1) is an object type with one or
                // more construct signatures, the expression is processed in the same manner as a
                // function call, but using the construct signatures as the initial set of candidate
                // signatures for overload resolution. The result type of the function call becomes
                // the result type of the operation.
                expressionType = getApparentType(expressionType);
                if (expressionType === errorType) {
                    // Another error has already been reported
                    return resolveErrorCall(node);
                }
                // TS 1.0 spec: 4.11
                // If expressionType is of type Any, Args can be any argument
                // list and the result of the operation is of type Any.
                if (isTypeAny(expressionType)) {
                    if (node.typeArguments) {
                        error(node, ts.Diagnostics.Untyped_function_calls_may_not_accept_type_arguments);
                    }
                    return resolveUntypedCall(node);
                }
                // Technically, this signatures list may be incomplete. We are taking the apparent type,
                // but we are not including construct signatures that may have been added to the Object or
                // Function interface, since they have none by default. This is a bit of a leap of faith
                // that the user will not add any.
                var constructSignatures = getSignaturesOfType(expressionType, 1 /* Construct */);
                if (constructSignatures.length) {
                    if (!isConstructorAccessible(node, constructSignatures[0])) {
                        return resolveErrorCall(node);
                    }
                    // If the expression is a class of abstract type, or an abstract construct signature,
                    // then it cannot be instantiated.
                    // In the case of a merged class-module or class-interface declaration,
                    // only the class declaration node will have the Abstract flag set.
                    if (constructSignatures.some(function (signature) { return signature.flags & 4 /* Abstract */; })) {
                        error(node, ts.Diagnostics.Cannot_create_an_instance_of_an_abstract_class);
                        return resolveErrorCall(node);
                    }
                    var valueDecl = expressionType.symbol && ts.getClassLikeDeclarationOfSymbol(expressionType.symbol);
                    if (valueDecl && ts.hasSyntacticModifier(valueDecl, 128 /* Abstract */)) {
                        error(node, ts.Diagnostics.Cannot_create_an_instance_of_an_abstract_class);
                        return resolveErrorCall(node);
                    }
                    return resolveCall(node, constructSignatures, candidatesOutArray, checkMode, 0 /* None */);
                }
                // If expressionType's apparent type is an object type with no construct signatures but
                // one or more call signatures, the expression is processed as a function call. A compile-time
                // error occurs if the result of the function call is not Void. The type of the result of the
                // operation is Any. It is an error to have a Void this type.
                var callSignatures = getSignaturesOfType(expressionType, 0 /* Call */);
                if (callSignatures.length) {
                    var signature = resolveCall(node, callSignatures, candidatesOutArray, checkMode, 0 /* None */);
                    if (!noImplicitAny) {
                        if (signature.declaration && !isJSConstructor(signature.declaration) && getReturnTypeOfSignature(signature) !== voidType) {
                            error(node, ts.Diagnostics.Only_a_void_function_can_be_called_with_the_new_keyword);
                        }
                        if (getThisTypeOfSignature(signature) === voidType) {
                            error(node, ts.Diagnostics.A_function_that_is_called_with_the_new_keyword_cannot_have_a_this_type_that_is_void);
                        }
                    }
                    return signature;
                }
                invocationError(node.expression, expressionType, 1 /* Construct */);
                return resolveErrorCall(node);
            }
            function typeHasProtectedAccessibleBase(target, type) {
                var baseTypes = getBaseTypes(type);
                if (!ts.length(baseTypes)) {
                    return false;
                }
                var firstBase = baseTypes[0];
                if (firstBase.flags & 2097152 /* Intersection */) {
                    var types = firstBase.types;
                    var mixinFlags = findMixins(types);
                    var i = 0;
                    for (var _i = 0, _a = firstBase.types; _i < _a.length; _i++) {
                        var intersectionMember = _a[_i];
                        // We want to ignore mixin ctors
                        if (!mixinFlags[i]) {
                            if (ts.getObjectFlags(intersectionMember) & (1 /* Class */ | 2 /* Interface */)) {
                                if (intersectionMember.symbol === target) {
                                    return true;
                                }
                                if (typeHasProtectedAccessibleBase(target, intersectionMember)) {
                                    return true;
                                }
                            }
                        }
                        i++;
                    }
                    return false;
                }
                if (firstBase.symbol === target) {
                    return true;
                }
                return typeHasProtectedAccessibleBase(target, firstBase);
            }
            function isConstructorAccessible(node, signature) {
                if (!signature || !signature.declaration) {
                    return true;
                }
                var declaration = signature.declaration;
                var modifiers = ts.getSelectedEffectiveModifierFlags(declaration, 24 /* NonPublicAccessibilityModifier */);
                // (1) Public constructors and (2) constructor functions are always accessible.
                if (!modifiers || declaration.kind !== 169 /* Constructor */) {
                    return true;
                }
                var declaringClassDeclaration = ts.getClassLikeDeclarationOfSymbol(declaration.parent.symbol);
                var declaringClass = getDeclaredTypeOfSymbol(declaration.parent.symbol);
                // A private or protected constructor can only be instantiated within its own class (or a subclass, for protected)
                if (!isNodeWithinClass(node, declaringClassDeclaration)) {
                    var containingClass = ts.getContainingClass(node);
                    if (containingClass && modifiers & 16 /* Protected */) {
                        var containingType = getTypeOfNode(containingClass);
                        if (typeHasProtectedAccessibleBase(declaration.parent.symbol, containingType)) {
                            return true;
                        }
                    }
                    if (modifiers & 8 /* Private */) {
                        error(node, ts.Diagnostics.Constructor_of_class_0_is_private_and_only_accessible_within_the_class_declaration, typeToString(declaringClass));
                    }
                    if (modifiers & 16 /* Protected */) {
                        error(node, ts.Diagnostics.Constructor_of_class_0_is_protected_and_only_accessible_within_the_class_declaration, typeToString(declaringClass));
                    }
                    return false;
                }
                return true;
            }
            function invocationErrorDetails(errorTarget, apparentType, kind) {
                var errorInfo;
                var isCall = kind === 0 /* Call */;
                var awaitedType = getAwaitedType(apparentType);
                var maybeMissingAwait = awaitedType && getSignaturesOfType(awaitedType, kind).length > 0;
                if (apparentType.flags & 1048576 /* Union */) {
                    var types = apparentType.types;
                    var hasSignatures = false;
                    for (var _i = 0, types_20 = types; _i < types_20.length; _i++) {
                        var constituent = types_20[_i];
                        var signatures = getSignaturesOfType(constituent, kind);
                        if (signatures.length !== 0) {
                            hasSignatures = true;
                            if (errorInfo) {
                                // Bail early if we already have an error, no chance of "No constituent of type is callable"
                                break;
                            }
                        }
                        else {
                            // Error on the first non callable constituent only
                            if (!errorInfo) {
                                errorInfo = ts.chainDiagnosticMessages(errorInfo, isCall ?
                                    ts.Diagnostics.Type_0_has_no_call_signatures :
                                    ts.Diagnostics.Type_0_has_no_construct_signatures, typeToString(constituent));
                                errorInfo = ts.chainDiagnosticMessages(errorInfo, isCall ?
                                    ts.Diagnostics.Not_all_constituents_of_type_0_are_callable :
                                    ts.Diagnostics.Not_all_constituents_of_type_0_are_constructable, typeToString(apparentType));
                            }
                            if (hasSignatures) {
                                // Bail early if we already found a siganture, no chance of "No constituent of type is callable"
                                break;
                            }
                        }
                    }
                    if (!hasSignatures) {
                        errorInfo = ts.chainDiagnosticMessages(
                        /* detials */ undefined, isCall ?
                            ts.Diagnostics.No_constituent_of_type_0_is_callable :
                            ts.Diagnostics.No_constituent_of_type_0_is_constructable, typeToString(apparentType));
                    }
                    if (!errorInfo) {
                        errorInfo = ts.chainDiagnosticMessages(errorInfo, isCall ?
                            ts.Diagnostics.Each_member_of_the_union_type_0_has_signatures_but_none_of_those_signatures_are_compatible_with_each_other :
                            ts.Diagnostics.Each_member_of_the_union_type_0_has_construct_signatures_but_none_of_those_signatures_are_compatible_with_each_other, typeToString(apparentType));
                    }
                }
                else {
                    errorInfo = ts.chainDiagnosticMessages(errorInfo, isCall ?
                        ts.Diagnostics.Type_0_has_no_call_signatures :
                        ts.Diagnostics.Type_0_has_no_construct_signatures, typeToString(apparentType));
                }
                var headMessage = isCall ? ts.Diagnostics.This_expression_is_not_callable : ts.Diagnostics.This_expression_is_not_constructable;
                // Diagnose get accessors incorrectly called as functions
                if (ts.isCallExpression(errorTarget.parent) && errorTarget.parent.arguments.length === 0) {
                    var resolvedSymbol = getNodeLinks(errorTarget).resolvedSymbol;
                    if (resolvedSymbol && resolvedSymbol.flags & 32768 /* GetAccessor */) {
                        headMessage = ts.Diagnostics.This_expression_is_not_callable_because_it_is_a_get_accessor_Did_you_mean_to_use_it_without;
                    }
                }
                return {
                    messageChain: ts.chainDiagnosticMessages(errorInfo, headMessage),
                    relatedMessage: maybeMissingAwait ? ts.Diagnostics.Did_you_forget_to_use_await : undefined,
                };
            }
            function invocationError(errorTarget, apparentType, kind, relatedInformation) {
                var _a = invocationErrorDetails(errorTarget, apparentType, kind), messageChain = _a.messageChain, relatedInfo = _a.relatedMessage;
                var diagnostic = ts.createDiagnosticForNodeFromMessageChain(errorTarget, messageChain);
                if (relatedInfo) {
                    ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(errorTarget, relatedInfo));
                }
                if (ts.isCallExpression(errorTarget.parent)) {
                    var _b = getDiagnosticSpanForCallNode(errorTarget.parent, /* doNotIncludeArguments */ true), start = _b.start, length_8 = _b.length;
                    diagnostic.start = start;
                    diagnostic.length = length_8;
                }
                diagnostics.add(diagnostic);
                invocationErrorRecovery(apparentType, kind, relatedInformation ? ts.addRelatedInfo(diagnostic, relatedInformation) : diagnostic);
            }
            function invocationErrorRecovery(apparentType, kind, diagnostic) {
                if (!apparentType.symbol) {
                    return;
                }
                var importNode = getSymbolLinks(apparentType.symbol).originatingImport;
                // Create a diagnostic on the originating import if possible onto which we can attach a quickfix
                //  An import call expression cannot be rewritten into another form to correct the error - the only solution is to use `.default` at the use-site
                if (importNode && !ts.isImportCall(importNode)) {
                    var sigs = getSignaturesOfType(getTypeOfSymbol(getSymbolLinks(apparentType.symbol).target), kind);
                    if (!sigs || !sigs.length)
                        return;
                    ts.addRelatedInfo(diagnostic, ts.createDiagnosticForNode(importNode, ts.Diagnostics.Type_originates_at_this_import_A_namespace_style_import_cannot_be_called_or_constructed_and_will_cause_a_failure_at_runtime_Consider_using_a_default_import_or_import_require_here_instead));
                }
            }
            function resolveTaggedTemplateExpression(node, candidatesOutArray, checkMode) {
                var tagType = checkExpression(node.tag);
                var apparentType = getApparentType(tagType);
                if (apparentType === errorType) {
                    // Another error has already been reported
                    return resolveErrorCall(node);
                }
                var callSignatures = getSignaturesOfType(apparentType, 0 /* Call */);
                var numConstructSignatures = getSignaturesOfType(apparentType, 1 /* Construct */).length;
                if (isUntypedFunctionCall(tagType, apparentType, callSignatures.length, numConstructSignatures)) {
                    return resolveUntypedCall(node);
                }
                if (!callSignatures.length) {
                    if (ts.isArrayLiteralExpression(node.parent)) {
                        var diagnostic = ts.createDiagnosticForNode(node.tag, ts.Diagnostics.It_is_likely_that_you_are_missing_a_comma_to_separate_these_two_template_expressions_They_form_a_tagged_template_expression_which_cannot_be_invoked);
                        diagnostics.add(diagnostic);
                        return resolveErrorCall(node);
                    }
                    invocationError(node.tag, apparentType, 0 /* Call */);
                    return resolveErrorCall(node);
                }
                return resolveCall(node, callSignatures, candidatesOutArray, checkMode, 0 /* None */);
            }
            /**
             * Gets the localized diagnostic head message to use for errors when resolving a decorator as a call expression.
             */
            function getDiagnosticHeadMessageForDecoratorResolution(node) {
                switch (node.parent.kind) {
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                        return ts.Diagnostics.Unable_to_resolve_signature_of_class_decorator_when_called_as_an_expression;
                    case 162 /* Parameter */:
                        return ts.Diagnostics.Unable_to_resolve_signature_of_parameter_decorator_when_called_as_an_expression;
                    case 165 /* PropertyDeclaration */:
                        return ts.Diagnostics.Unable_to_resolve_signature_of_property_decorator_when_called_as_an_expression;
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return ts.Diagnostics.Unable_to_resolve_signature_of_method_decorator_when_called_as_an_expression;
                    default:
                        return ts.Debug.fail();
                }
            }
            /**
             * Resolves a decorator as if it were a call expression.
             */
            function resolveDecorator(node, candidatesOutArray, checkMode) {
                var funcType = checkExpression(node.expression);
                var apparentType = getApparentType(funcType);
                if (apparentType === errorType) {
                    return resolveErrorCall(node);
                }
                var callSignatures = getSignaturesOfType(apparentType, 0 /* Call */);
                var numConstructSignatures = getSignaturesOfType(apparentType, 1 /* Construct */).length;
                if (isUntypedFunctionCall(funcType, apparentType, callSignatures.length, numConstructSignatures)) {
                    return resolveUntypedCall(node);
                }
                if (isPotentiallyUncalledDecorator(node, callSignatures)) {
                    var nodeStr = ts.getTextOfNode(node.expression, /*includeTrivia*/ false);
                    error(node, ts.Diagnostics._0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write_0, nodeStr);
                    return resolveErrorCall(node);
                }
                var headMessage = getDiagnosticHeadMessageForDecoratorResolution(node);
                if (!callSignatures.length) {
                    var errorDetails = invocationErrorDetails(node.expression, apparentType, 0 /* Call */);
                    var messageChain = ts.chainDiagnosticMessages(errorDetails.messageChain, headMessage);
                    var diag = ts.createDiagnosticForNodeFromMessageChain(node.expression, messageChain);
                    if (errorDetails.relatedMessage) {
                        ts.addRelatedInfo(diag, ts.createDiagnosticForNode(node.expression, errorDetails.relatedMessage));
                    }
                    diagnostics.add(diag);
                    invocationErrorRecovery(apparentType, 0 /* Call */, diag);
                    return resolveErrorCall(node);
                }
                return resolveCall(node, callSignatures, candidatesOutArray, checkMode, 0 /* None */, headMessage);
            }
            function createSignatureForJSXIntrinsic(node, result) {
                var namespace = getJsxNamespaceAt(node);
                var exports = namespace && getExportsOfSymbol(namespace);
                // We fake up a SFC signature for each intrinsic, however a more specific per-element signature drawn from the JSX declaration
                // file would probably be preferable.
                var typeSymbol = exports && getSymbol(exports, JsxNames.Element, 788968 /* Type */);
                var returnNode = typeSymbol && nodeBuilder.symbolToEntityName(typeSymbol, 788968 /* Type */, node);
                var declaration = ts.factory.createFunctionTypeNode(/*typeParameters*/ undefined, [ts.factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotdotdot*/ undefined, "props", /*questionMark*/ undefined, nodeBuilder.typeToTypeNode(result, node))], returnNode ? ts.factory.createTypeReferenceNode(returnNode, /*typeArguments*/ undefined) : ts.factory.createKeywordTypeNode(129 /* AnyKeyword */));
                var parameterSymbol = createSymbol(1 /* FunctionScopedVariable */, "props");
                parameterSymbol.type = result;
                return createSignature(declaration, 
                /*typeParameters*/ undefined, 
                /*thisParameter*/ undefined, [parameterSymbol], typeSymbol ? getDeclaredTypeOfSymbol(typeSymbol) : errorType, 
                /*returnTypePredicate*/ undefined, 1, 0 /* None */);
            }
            function resolveJsxOpeningLikeElement(node, candidatesOutArray, checkMode) {
                if (isJsxIntrinsicIdentifier(node.tagName)) {
                    var result = getIntrinsicAttributesTypeFromJsxOpeningLikeElement(node);
                    var fakeSignature = createSignatureForJSXIntrinsic(node, result);
                    checkTypeAssignableToAndOptionallyElaborate(checkExpressionWithContextualType(node.attributes, getEffectiveFirstArgumentForJsxSignature(fakeSignature, node), /*mapper*/ undefined, 0 /* Normal */), result, node.tagName, node.attributes);
                    if (ts.length(node.typeArguments)) {
                        ts.forEach(node.typeArguments, checkSourceElement);
                        diagnostics.add(ts.createDiagnosticForNodeArray(ts.getSourceFileOfNode(node), node.typeArguments, ts.Diagnostics.Expected_0_type_arguments_but_got_1, 0, ts.length(node.typeArguments)));
                    }
                    return fakeSignature;
                }
                var exprTypes = checkExpression(node.tagName);
                var apparentType = getApparentType(exprTypes);
                if (apparentType === errorType) {
                    return resolveErrorCall(node);
                }
                var signatures = getUninstantiatedJsxSignaturesOfType(exprTypes, node);
                if (isUntypedFunctionCall(exprTypes, apparentType, signatures.length, /*constructSignatures*/ 0)) {
                    return resolveUntypedCall(node);
                }
                if (signatures.length === 0) {
                    // We found no signatures at all, which is an error
                    error(node.tagName, ts.Diagnostics.JSX_element_type_0_does_not_have_any_construct_or_call_signatures, ts.getTextOfNode(node.tagName));
                    return resolveErrorCall(node);
                }
                return resolveCall(node, signatures, candidatesOutArray, checkMode, 0 /* None */);
            }
            /**
             * Sometimes, we have a decorator that could accept zero arguments,
             * but is receiving too many arguments as part of the decorator invocation.
             * In those cases, a user may have meant to *call* the expression before using it as a decorator.
             */
            function isPotentiallyUncalledDecorator(decorator, signatures) {
                return signatures.length && ts.every(signatures, function (signature) {
                    return signature.minArgumentCount === 0 &&
                        !signatureHasRestParameter(signature) &&
                        signature.parameters.length < getDecoratorArgumentCount(decorator, signature);
                });
            }
            function resolveSignature(node, candidatesOutArray, checkMode) {
                switch (node.kind) {
                    case 206 /* CallExpression */:
                        return resolveCallExpression(node, candidatesOutArray, checkMode);
                    case 207 /* NewExpression */:
                        return resolveNewExpression(node, candidatesOutArray, checkMode);
                    case 208 /* TaggedTemplateExpression */:
                        return resolveTaggedTemplateExpression(node, candidatesOutArray, checkMode);
                    case 163 /* Decorator */:
                        return resolveDecorator(node, candidatesOutArray, checkMode);
                    case 278 /* JsxOpeningElement */:
                    case 277 /* JsxSelfClosingElement */:
                        return resolveJsxOpeningLikeElement(node, candidatesOutArray, checkMode);
                }
                throw ts.Debug.assertNever(node, "Branch in 'resolveSignature' should be unreachable.");
            }
            /**
             * Resolve a signature of a given call-like expression.
             * @param node a call-like expression to try resolve a signature for
             * @param candidatesOutArray an array of signature to be filled in by the function. It is passed by signature help in the language service;
             *                           the function will fill it up with appropriate candidate signatures
             * @return a signature of the call-like expression or undefined if one can't be found
             */
            function getResolvedSignature(node, candidatesOutArray, checkMode) {
                var links = getNodeLinks(node);
                // If getResolvedSignature has already been called, we will have cached the resolvedSignature.
                // However, it is possible that either candidatesOutArray was not passed in the first time,
                // or that a different candidatesOutArray was passed in. Therefore, we need to redo the work
                // to correctly fill the candidatesOutArray.
                var cached = links.resolvedSignature;
                if (cached && cached !== resolvingSignature && !candidatesOutArray) {
                    return cached;
                }
                links.resolvedSignature = resolvingSignature;
                var result = resolveSignature(node, candidatesOutArray, checkMode || 0 /* Normal */);
                // When CheckMode.SkipGenericFunctions is set we use resolvingSignature to indicate that call
                // resolution should be deferred.
                if (result !== resolvingSignature) {
                    // If signature resolution originated in control flow type analysis (for example to compute the
                    // assigned type in a flow assignment) we don't cache the result as it may be based on temporary
                    // types from the control flow analysis.
                    links.resolvedSignature = flowLoopStart === flowLoopCount ? result : cached;
                }
                return result;
            }
            /**
             * Indicates whether a declaration can be treated as a constructor in a JavaScript
             * file.
             */
            function isJSConstructor(node) {
                var _a;
                if (!node || !ts.isInJSFile(node)) {
                    return false;
                }
                var func = ts.isFunctionDeclaration(node) || ts.isFunctionExpression(node) ? node :
                    ts.isVariableDeclaration(node) && node.initializer && ts.isFunctionExpression(node.initializer) ? node.initializer :
                        undefined;
                if (func) {
                    // If the node has a @class tag, treat it like a constructor.
                    if (ts.getJSDocClassTag(node))
                        return true;
                    // If the symbol of the node has members, treat it like a constructor.
                    var symbol = getSymbolOfNode(func);
                    return !!((_a = symbol === null || symbol === void 0 ? void 0 : symbol.members) === null || _a === void 0 ? void 0 : _a.size);
                }
                return false;
            }
            function mergeJSSymbols(target, source) {
                var _a, _b;
                if (source) {
                    var links = getSymbolLinks(source);
                    if (!links.inferredClassSymbol || !links.inferredClassSymbol.has(getSymbolId(target))) {
                        var inferred = ts.isTransientSymbol(target) ? target : cloneSymbol(target);
                        inferred.exports = inferred.exports || ts.createSymbolTable();
                        inferred.members = inferred.members || ts.createSymbolTable();
                        inferred.flags |= source.flags & 32 /* Class */;
                        if ((_a = source.exports) === null || _a === void 0 ? void 0 : _a.size) {
                            mergeSymbolTable(inferred.exports, source.exports);
                        }
                        if ((_b = source.members) === null || _b === void 0 ? void 0 : _b.size) {
                            mergeSymbolTable(inferred.members, source.members);
                        }
                        (links.inferredClassSymbol || (links.inferredClassSymbol = new ts.Map())).set(getSymbolId(inferred), inferred);
                        return inferred;
                    }
                    return links.inferredClassSymbol.get(getSymbolId(target));
                }
            }
            function getAssignedClassSymbol(decl) {
                var _a;
                var assignmentSymbol = decl && getSymbolOfExpando(decl, /*allowDeclaration*/ true);
                var prototype = (_a = assignmentSymbol === null || assignmentSymbol === void 0 ? void 0 : assignmentSymbol.exports) === null || _a === void 0 ? void 0 : _a.get("prototype");
                var init = (prototype === null || prototype === void 0 ? void 0 : prototype.valueDeclaration) && getAssignedJSPrototype(prototype.valueDeclaration);
                return init ? getSymbolOfNode(init) : undefined;
            }
            function getSymbolOfExpando(node, allowDeclaration) {
                if (!node.parent) {
                    return undefined;
                }
                var name;
                var decl;
                if (ts.isVariableDeclaration(node.parent) && node.parent.initializer === node) {
                    if (!ts.isInJSFile(node) && !(ts.isVarConst(node.parent) && ts.isFunctionLikeDeclaration(node))) {
                        return undefined;
                    }
                    name = node.parent.name;
                    decl = node.parent;
                }
                else if (ts.isBinaryExpression(node.parent)) {
                    var parentNode = node.parent;
                    var parentNodeOperator = node.parent.operatorToken.kind;
                    if (parentNodeOperator === 63 /* EqualsToken */ && (allowDeclaration || parentNode.right === node)) {
                        name = parentNode.left;
                        decl = name;
                    }
                    else if (parentNodeOperator === 56 /* BarBarToken */ || parentNodeOperator === 60 /* QuestionQuestionToken */) {
                        if (ts.isVariableDeclaration(parentNode.parent) && parentNode.parent.initializer === parentNode) {
                            name = parentNode.parent.name;
                            decl = parentNode.parent;
                        }
                        else if (ts.isBinaryExpression(parentNode.parent) && parentNode.parent.operatorToken.kind === 63 /* EqualsToken */ && (allowDeclaration || parentNode.parent.right === parentNode)) {
                            name = parentNode.parent.left;
                            decl = name;
                        }
                        if (!name || !ts.isBindableStaticNameExpression(name) || !ts.isSameEntityName(name, parentNode.left)) {
                            return undefined;
                        }
                    }
                }
                else if (allowDeclaration && ts.isFunctionDeclaration(node)) {
                    name = node.name;
                    decl = node;
                }
                if (!decl || !name || (!allowDeclaration && !ts.getExpandoInitializer(node, ts.isPrototypeAccess(name)))) {
                    return undefined;
                }
                return getSymbolOfNode(decl);
            }
            function getAssignedJSPrototype(node) {
                if (!node.parent) {
                    return false;
                }
                var parent = node.parent;
                while (parent && parent.kind === 204 /* PropertyAccessExpression */) {
                    parent = parent.parent;
                }
                if (parent && ts.isBinaryExpression(parent) && ts.isPrototypeAccess(parent.left) && parent.operatorToken.kind === 63 /* EqualsToken */) {
                    var right = ts.getInitializerOfBinaryExpression(parent);
                    return ts.isObjectLiteralExpression(right) && right;
                }
            }
            /**
             * Syntactically and semantically checks a call or new expression.
             * @param node The call/new expression to be checked.
             * @returns On success, the expression's signature's return type. On failure, anyType.
             */
            function checkCallExpression(node, checkMode) {
                var _a;
                if (!checkGrammarTypeArguments(node, node.typeArguments))
                    checkGrammarArguments(node.arguments);
                var signature = getResolvedSignature(node, /*candidatesOutArray*/ undefined, checkMode);
                if (signature === resolvingSignature) {
                    // CheckMode.SkipGenericFunctions is enabled and this is a call to a generic function that
                    // returns a function type. We defer checking and return nonInferrableType.
                    return nonInferrableType;
                }
                checkDeprecatedSignature(signature, node);
                if (node.expression.kind === 106 /* SuperKeyword */) {
                    return voidType;
                }
                if (node.kind === 207 /* NewExpression */) {
                    var declaration = signature.declaration;
                    if (declaration &&
                        declaration.kind !== 169 /* Constructor */ &&
                        declaration.kind !== 173 /* ConstructSignature */ &&
                        declaration.kind !== 178 /* ConstructorType */ &&
                        !ts.isJSDocConstructSignature(declaration) &&
                        !isJSConstructor(declaration)) {
                        // When resolved signature is a call signature (and not a construct signature) the result type is any
                        if (noImplicitAny) {
                            error(node, ts.Diagnostics.new_expression_whose_target_lacks_a_construct_signature_implicitly_has_an_any_type);
                        }
                        return anyType;
                    }
                }
                // In JavaScript files, calls to any identifier 'require' are treated as external module imports
                if (ts.isInJSFile(node) && isCommonJsRequire(node)) {
                    return resolveExternalModuleTypeByLiteral(node.arguments[0]);
                }
                var returnType = getReturnTypeOfSignature(signature);
                // Treat any call to the global 'Symbol' function that is part of a const variable or readonly property
                // as a fresh unique symbol literal type.
                if (returnType.flags & 12288 /* ESSymbolLike */ && isSymbolOrSymbolForCall(node)) {
                    return getESSymbolLikeTypeForNode(ts.walkUpParenthesizedExpressions(node.parent));
                }
                if (node.kind === 206 /* CallExpression */ && !node.questionDotToken && node.parent.kind === 236 /* ExpressionStatement */ &&
                    returnType.flags & 16384 /* Void */ && getTypePredicateOfSignature(signature)) {
                    if (!ts.isDottedName(node.expression)) {
                        error(node.expression, ts.Diagnostics.Assertions_require_the_call_target_to_be_an_identifier_or_qualified_name);
                    }
                    else if (!getEffectsSignature(node)) {
                        var diagnostic = error(node.expression, ts.Diagnostics.Assertions_require_every_name_in_the_call_target_to_be_declared_with_an_explicit_type_annotation);
                        getTypeOfDottedName(node.expression, diagnostic);
                    }
                }
                if (ts.isInJSFile(node)) {
                    var jsSymbol = getSymbolOfExpando(node, /*allowDeclaration*/ false);
                    if ((_a = jsSymbol === null || jsSymbol === void 0 ? void 0 : jsSymbol.exports) === null || _a === void 0 ? void 0 : _a.size) {
                        var jsAssignmentType = createAnonymousType(jsSymbol, jsSymbol.exports, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                        jsAssignmentType.objectFlags |= 8192 /* JSLiteral */;
                        return getIntersectionType([returnType, jsAssignmentType]);
                    }
                }
                return returnType;
            }
            function checkDeprecatedSignature(signature, node) {
                if (signature.declaration && signature.declaration.flags & 134217728 /* Deprecated */) {
                    var suggestionNode = getDeprecatedSuggestionNode(node);
                    var name = ts.tryGetPropertyAccessOrIdentifierToString(ts.getInvokedExpression(node));
                    addDeprecatedSuggestionWithSignature(suggestionNode, signature.declaration, name, signatureToString(signature));
                }
            }
            function getDeprecatedSuggestionNode(node) {
                node = ts.skipParentheses(node);
                switch (node.kind) {
                    case 206 /* CallExpression */:
                    case 163 /* Decorator */:
                    case 207 /* NewExpression */:
                        return getDeprecatedSuggestionNode(node.expression);
                    case 208 /* TaggedTemplateExpression */:
                        return getDeprecatedSuggestionNode(node.tag);
                    case 278 /* JsxOpeningElement */:
                    case 277 /* JsxSelfClosingElement */:
                        return getDeprecatedSuggestionNode(node.tagName);
                    case 205 /* ElementAccessExpression */:
                        return node.argumentExpression;
                    case 204 /* PropertyAccessExpression */:
                        return node.name;
                    case 176 /* TypeReference */:
                        var typeReference = node;
                        return ts.isQualifiedName(typeReference.typeName) ? typeReference.typeName.right : typeReference;
                    default:
                        return node;
                }
            }
            function isSymbolOrSymbolForCall(node) {
                if (!ts.isCallExpression(node))
                    return false;
                var left = node.expression;
                if (ts.isPropertyAccessExpression(left) && left.name.escapedText === "for") {
                    left = left.expression;
                }
                if (!ts.isIdentifier(left) || left.escapedText !== "Symbol") {
                    return false;
                }
                // make sure `Symbol` is the global symbol
                var globalESSymbol = getGlobalESSymbolConstructorSymbol(/*reportErrors*/ false);
                if (!globalESSymbol) {
                    return false;
                }
                return globalESSymbol === resolveName(left, "Symbol", 111551 /* Value */, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false);
            }
            function checkImportCallExpression(node) {
                // Check grammar of dynamic import
                if (!checkGrammarArguments(node.arguments))
                    checkGrammarImportCallExpression(node);
                if (node.arguments.length === 0) {
                    return createPromiseReturnType(node, anyType);
                }
                var specifier = node.arguments[0];
                var specifierType = checkExpressionCached(specifier);
                // Even though multiple arguments is grammatically incorrect, type-check extra arguments for completion
                for (var i = 1; i < node.arguments.length; ++i) {
                    checkExpressionCached(node.arguments[i]);
                }
                if (specifierType.flags & 32768 /* Undefined */ || specifierType.flags & 65536 /* Null */ || !isTypeAssignableTo(specifierType, stringType)) {
                    error(specifier, ts.Diagnostics.Dynamic_import_s_specifier_must_be_of_type_string_but_here_has_type_0, typeToString(specifierType));
                }
                // resolveExternalModuleName will return undefined if the moduleReferenceExpression is not a string literal
                var moduleSymbol = resolveExternalModuleName(node, specifier);
                if (moduleSymbol) {
                    var esModuleSymbol = resolveESModuleSymbol(moduleSymbol, specifier, /*dontRecursivelyResolve*/ true, /*suppressUsageError*/ false);
                    if (esModuleSymbol) {
                        return createPromiseReturnType(node, getTypeWithSyntheticDefaultImportType(getTypeOfSymbol(esModuleSymbol), esModuleSymbol, moduleSymbol));
                    }
                }
                return createPromiseReturnType(node, anyType);
            }
            function getTypeWithSyntheticDefaultImportType(type, symbol, originalSymbol) {
                var _a;
                if (allowSyntheticDefaultImports && type && type !== errorType) {
                    var synthType = type;
                    if (!synthType.syntheticType) {
                        var file = (_a = originalSymbol.declarations) === null || _a === void 0 ? void 0 : _a.find(ts.isSourceFile);
                        var hasSyntheticDefault = canHaveSyntheticDefault(file, originalSymbol, /*dontResolveAlias*/ false);
                        if (hasSyntheticDefault) {
                            var memberTable = ts.createSymbolTable();
                            var newSymbol = createSymbol(2097152 /* Alias */, "default" /* Default */);
                            newSymbol.parent = originalSymbol;
                            newSymbol.nameType = getStringLiteralType("default");
                            newSymbol.target = resolveSymbol(symbol);
                            memberTable.set("default" /* Default */, newSymbol);
                            var anonymousSymbol = createSymbol(2048 /* TypeLiteral */, "__type" /* Type */);
                            var defaultContainingObject = createAnonymousType(anonymousSymbol, memberTable, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                            anonymousSymbol.type = defaultContainingObject;
                            synthType.syntheticType = isValidSpreadType(type) ? getSpreadType(type, defaultContainingObject, anonymousSymbol, /*objectFlags*/ 0, /*readonly*/ false) : defaultContainingObject;
                        }
                        else {
                            synthType.syntheticType = type;
                        }
                    }
                    return synthType.syntheticType;
                }
                return type;
            }
            function isCommonJsRequire(node) {
                if (!ts.isRequireCall(node, /*checkArgumentIsStringLiteralLike*/ true)) {
                    return false;
                }
                // Make sure require is not a local function
                if (!ts.isIdentifier(node.expression))
                    return ts.Debug.fail();
                var resolvedRequire = resolveName(node.expression, node.expression.escapedText, 111551 /* Value */, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ true); // TODO: GH#18217
                if (resolvedRequire === requireSymbol) {
                    return true;
                }
                // project includes symbol named 'require' - make sure that it is ambient and local non-alias
                if (resolvedRequire.flags & 2097152 /* Alias */) {
                    return false;
                }
                var targetDeclarationKind = resolvedRequire.flags & 16 /* Function */
                    ? 254 /* FunctionDeclaration */
                    : resolvedRequire.flags & 3 /* Variable */
                        ? 252 /* VariableDeclaration */
                        : 0 /* Unknown */;
                if (targetDeclarationKind !== 0 /* Unknown */) {
                    var decl = ts.getDeclarationOfKind(resolvedRequire, targetDeclarationKind);
                    // function/variable declaration should be ambient
                    return !!decl && !!(decl.flags & 8388608 /* Ambient */);
                }
                return false;
            }
            function checkTaggedTemplateExpression(node) {
                if (!checkGrammarTaggedTemplateChain(node))
                    checkGrammarTypeArguments(node, node.typeArguments);
                if (languageVersion < 2 /* ES2015 */) {
                    checkExternalEmitHelpers(node, 262144 /* MakeTemplateObject */);
                }
                var signature = getResolvedSignature(node);
                checkDeprecatedSignature(signature, node);
                return getReturnTypeOfSignature(signature);
            }
            function checkAssertion(node) {
                return checkAssertionWorker(node, node.type, node.expression);
            }
            function isValidConstAssertionArgument(node) {
                switch (node.kind) {
                    case 10 /* StringLiteral */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                    case 8 /* NumericLiteral */:
                    case 9 /* BigIntLiteral */:
                    case 110 /* TrueKeyword */:
                    case 95 /* FalseKeyword */:
                    case 202 /* ArrayLiteralExpression */:
                    case 203 /* ObjectLiteralExpression */:
                    case 221 /* TemplateExpression */:
                        return true;
                    case 210 /* ParenthesizedExpression */:
                        return isValidConstAssertionArgument(node.expression);
                    case 217 /* PrefixUnaryExpression */:
                        var op = node.operator;
                        var arg = node.operand;
                        return op === 40 /* MinusToken */ && (arg.kind === 8 /* NumericLiteral */ || arg.kind === 9 /* BigIntLiteral */) ||
                            op === 39 /* PlusToken */ && arg.kind === 8 /* NumericLiteral */;
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */:
                        var expr = node.expression;
                        if (ts.isIdentifier(expr)) {
                            var symbol = getSymbolAtLocation(expr);
                            if (symbol && symbol.flags & 2097152 /* Alias */) {
                                symbol = resolveAlias(symbol);
                            }
                            return !!(symbol && (symbol.flags & 384 /* Enum */) && getEnumKind(symbol) === 1 /* Literal */);
                        }
                }
                return false;
            }
            function checkAssertionWorker(errNode, type, expression, checkMode) {
                var exprType = checkExpression(expression, checkMode);
                if (ts.isConstTypeReference(type)) {
                    if (!isValidConstAssertionArgument(expression)) {
                        error(expression, ts.Diagnostics.A_const_assertions_can_only_be_applied_to_references_to_enum_members_or_string_number_boolean_array_or_object_literals);
                    }
                    return getRegularTypeOfLiteralType(exprType);
                }
                checkSourceElement(type);
                exprType = getRegularTypeOfObjectLiteral(getBaseTypeOfLiteralType(exprType));
                var targetType = getTypeFromTypeNode(type);
                if (produceDiagnostics && targetType !== errorType) {
                    var widenedType = getWidenedType(exprType);
                    if (!isTypeComparableTo(targetType, widenedType)) {
                        checkTypeComparableTo(exprType, targetType, errNode, ts.Diagnostics.Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the_other_If_this_was_intentional_convert_the_expression_to_unknown_first);
                    }
                }
                return targetType;
            }
            function checkNonNullChain(node) {
                var leftType = checkExpression(node.expression);
                var nonOptionalType = getOptionalExpressionType(leftType, node.expression);
                return propagateOptionalTypeMarker(getNonNullableType(nonOptionalType), node, nonOptionalType !== leftType);
            }
            function checkNonNullAssertion(node) {
                return node.flags & 32 /* OptionalChain */ ? checkNonNullChain(node) :
                    getNonNullableType(checkExpression(node.expression));
            }
            function checkMetaProperty(node) {
                checkGrammarMetaProperty(node);
                if (node.keywordToken === 103 /* NewKeyword */) {
                    return checkNewTargetMetaProperty(node);
                }
                if (node.keywordToken === 100 /* ImportKeyword */) {
                    return checkImportMetaProperty(node);
                }
                return ts.Debug.assertNever(node.keywordToken);
            }
            function checkMetaPropertyKeyword(node) {
                switch (node.keywordToken) {
                    case 100 /* ImportKeyword */:
                        return getGlobalImportMetaExpressionType();
                    case 103 /* NewKeyword */:
                        var type = checkNewTargetMetaProperty(node);
                        return type === errorType ? errorType : createNewTargetExpressionType(type);
                    default:
                        ts.Debug.assertNever(node.keywordToken);
                }
            }
            function checkNewTargetMetaProperty(node) {
                var container = ts.getNewTargetContainer(node);
                if (!container) {
                    error(node, ts.Diagnostics.Meta_property_0_is_only_allowed_in_the_body_of_a_function_declaration_function_expression_or_constructor, "new.target");
                    return errorType;
                }
                else if (container.kind === 169 /* Constructor */) {
                    var symbol = getSymbolOfNode(container.parent);
                    return getTypeOfSymbol(symbol);
                }
                else {
                    var symbol = getSymbolOfNode(container);
                    return getTypeOfSymbol(symbol);
                }
            }
            function checkImportMetaProperty(node) {
                if (moduleKind !== ts.ModuleKind.ES2020 && moduleKind !== ts.ModuleKind.ESNext && moduleKind !== ts.ModuleKind.System) {
                    error(node, ts.Diagnostics.The_import_meta_meta_property_is_only_allowed_when_the_module_option_is_es2020_esnext_or_system);
                }
                var file = ts.getSourceFileOfNode(node);
                ts.Debug.assert(!!(file.flags & 2097152 /* PossiblyContainsImportMeta */), "Containing file is missing import meta node flag.");
                return node.name.escapedText === "meta" ? getGlobalImportMetaType() : errorType;
            }
            function getTypeOfParameter(symbol) {
                var type = getTypeOfSymbol(symbol);
                if (strictNullChecks) {
                    var declaration = symbol.valueDeclaration;
                    if (declaration && ts.hasInitializer(declaration)) {
                        return getOptionalType(type);
                    }
                }
                return type;
            }
            function getTupleElementLabel(d) {
                ts.Debug.assert(ts.isIdentifier(d.name)); // Parameter declarations could be binding patterns, but we only allow identifier names
                return d.name.escapedText;
            }
            function getParameterNameAtPosition(signature, pos, overrideRestType) {
                var paramCount = signature.parameters.length - (signatureHasRestParameter(signature) ? 1 : 0);
                if (pos < paramCount) {
                    return signature.parameters[pos].escapedName;
                }
                var restParameter = signature.parameters[paramCount] || unknownSymbol;
                var restType = overrideRestType || getTypeOfSymbol(restParameter);
                if (isTupleType(restType)) {
                    var associatedNames = restType.target.labeledElementDeclarations;
                    var index = pos - paramCount;
                    return associatedNames && getTupleElementLabel(associatedNames[index]) || restParameter.escapedName + "_" + index;
                }
                return restParameter.escapedName;
            }
            function getParameterIdentifierNameAtPosition(signature, pos) {
                var paramCount = signature.parameters.length - (signatureHasRestParameter(signature) ? 1 : 0);
                if (pos < paramCount) {
                    var param = signature.parameters[pos];
                    return isParameterDeclarationWithIdentifierName(param) ? [param.escapedName, false] : undefined;
                }
                var restParameter = signature.parameters[paramCount] || unknownSymbol;
                if (!isParameterDeclarationWithIdentifierName(restParameter)) {
                    return undefined;
                }
                var restType = getTypeOfSymbol(restParameter);
                if (isTupleType(restType)) {
                    var associatedNames = restType.target.labeledElementDeclarations;
                    var index = pos - paramCount;
                    var associatedName = associatedNames === null || associatedNames === void 0 ? void 0 : associatedNames[index];
                    var isRestTupleElement = !!(associatedName === null || associatedName === void 0 ? void 0 : associatedName.dotDotDotToken);
                    return associatedName ? [
                        getTupleElementLabel(associatedName),
                        isRestTupleElement
                    ] : undefined;
                }
                if (pos === paramCount) {
                    return [restParameter.escapedName, true];
                }
                return undefined;
            }
            function isParameterDeclarationWithIdentifierName(symbol) {
                return symbol.valueDeclaration && ts.isParameter(symbol.valueDeclaration) && ts.isIdentifier(symbol.valueDeclaration.name);
            }
            function isValidDeclarationForTupleLabel(d) {
                return d.kind === 195 /* NamedTupleMember */ || (ts.isParameter(d) && d.name && ts.isIdentifier(d.name));
            }
            function getNameableDeclarationAtPosition(signature, pos) {
                var paramCount = signature.parameters.length - (signatureHasRestParameter(signature) ? 1 : 0);
                if (pos < paramCount) {
                    var decl = signature.parameters[pos].valueDeclaration;
                    return decl && isValidDeclarationForTupleLabel(decl) ? decl : undefined;
                }
                var restParameter = signature.parameters[paramCount] || unknownSymbol;
                var restType = getTypeOfSymbol(restParameter);
                if (isTupleType(restType)) {
                    var associatedNames = restType.target.labeledElementDeclarations;
                    var index = pos - paramCount;
                    return associatedNames && associatedNames[index];
                }
                return restParameter.valueDeclaration && isValidDeclarationForTupleLabel(restParameter.valueDeclaration) ? restParameter.valueDeclaration : undefined;
            }
            function getTypeAtPosition(signature, pos) {
                return tryGetTypeAtPosition(signature, pos) || anyType;
            }
            function tryGetTypeAtPosition(signature, pos) {
                var paramCount = signature.parameters.length - (signatureHasRestParameter(signature) ? 1 : 0);
                if (pos < paramCount) {
                    return getTypeOfParameter(signature.parameters[pos]);
                }
                if (signatureHasRestParameter(signature)) {
                    // We want to return the value undefined for an out of bounds parameter position,
                    // so we need to check bounds here before calling getIndexedAccessType (which
                    // otherwise would return the type 'undefined').
                    var restType = getTypeOfSymbol(signature.parameters[paramCount]);
                    var index = pos - paramCount;
                    if (!isTupleType(restType) || restType.target.hasRestElement || index < restType.target.fixedLength) {
                        return getIndexedAccessType(restType, getNumberLiteralType(index));
                    }
                }
                return undefined;
            }
            function getRestTypeAtPosition(source, pos) {
                var parameterCount = getParameterCount(source);
                var minArgumentCount = getMinArgumentCount(source);
                var restType = getEffectiveRestType(source);
                if (restType && pos >= parameterCount - 1) {
                    return pos === parameterCount - 1 ? restType : createArrayType(getIndexedAccessType(restType, numberType));
                }
                var types = [];
                var flags = [];
                var names = [];
                for (var i = pos; i < parameterCount; i++) {
                    if (!restType || i < parameterCount - 1) {
                        types.push(getTypeAtPosition(source, i));
                        flags.push(i < minArgumentCount ? 1 /* Required */ : 2 /* Optional */);
                    }
                    else {
                        types.push(restType);
                        flags.push(8 /* Variadic */);
                    }
                    var name = getNameableDeclarationAtPosition(source, i);
                    if (name) {
                        names.push(name);
                    }
                }
                return createTupleType(types, flags, /*readonly*/ false, ts.length(names) === ts.length(types) ? names : undefined);
            }
            // Return the number of parameters in a signature. The rest parameter, if present, counts as one
            // parameter. For example, the parameter count of (x: number, y: number, ...z: string[]) is 3 and
            // the parameter count of (x: number, ...args: [number, ...string[], boolean])) is also 3. In the
            // latter example, the effective rest type is [...string[], boolean].
            function getParameterCount(signature) {
                var length = signature.parameters.length;
                if (signatureHasRestParameter(signature)) {
                    var restType = getTypeOfSymbol(signature.parameters[length - 1]);
                    if (isTupleType(restType)) {
                        return length + restType.target.fixedLength - (restType.target.hasRestElement ? 0 : 1);
                    }
                }
                return length;
            }
            function getMinArgumentCount(signature, flags) {
                var strongArityForUntypedJS = flags & 1 /* StrongArityForUntypedJS */;
                var voidIsNonOptional = flags & 2 /* VoidIsNonOptional */;
                if (voidIsNonOptional || signature.resolvedMinArgumentCount === undefined) {
                    var minArgumentCount = void 0;
                    if (signatureHasRestParameter(signature)) {
                        var restType = getTypeOfSymbol(signature.parameters[signature.parameters.length - 1]);
                        if (isTupleType(restType)) {
                            var firstOptionalIndex = ts.findIndex(restType.target.elementFlags, function (f) { return !(f & 1 /* Required */); });
                            var requiredCount = firstOptionalIndex < 0 ? restType.target.fixedLength : firstOptionalIndex;
                            if (requiredCount > 0) {
                                minArgumentCount = signature.parameters.length - 1 + requiredCount;
                            }
                        }
                    }
                    if (minArgumentCount === undefined) {
                        if (!strongArityForUntypedJS && signature.flags & 32 /* IsUntypedSignatureInJSFile */) {
                            return 0;
                        }
                        minArgumentCount = signature.minArgumentCount;
                    }
                    if (voidIsNonOptional) {
                        return minArgumentCount;
                    }
                    for (var i = minArgumentCount - 1; i >= 0; i--) {
                        var type = getTypeAtPosition(signature, i);
                        if (filterType(type, acceptsVoid).flags & 131072 /* Never */) {
                            break;
                        }
                        minArgumentCount = i;
                    }
                    signature.resolvedMinArgumentCount = minArgumentCount;
                }
                return signature.resolvedMinArgumentCount;
            }
            function hasEffectiveRestParameter(signature) {
                if (signatureHasRestParameter(signature)) {
                    var restType = getTypeOfSymbol(signature.parameters[signature.parameters.length - 1]);
                    return !isTupleType(restType) || restType.target.hasRestElement;
                }
                return false;
            }
            function getEffectiveRestType(signature) {
                if (signatureHasRestParameter(signature)) {
                    var restType = getTypeOfSymbol(signature.parameters[signature.parameters.length - 1]);
                    if (!isTupleType(restType)) {
                        return restType;
                    }
                    if (restType.target.hasRestElement) {
                        return sliceTupleType(restType, restType.target.fixedLength);
                    }
                }
                return undefined;
            }
            function getNonArrayRestType(signature) {
                var restType = getEffectiveRestType(signature);
                return restType && !isArrayType(restType) && !isTypeAny(restType) && (getReducedType(restType).flags & 131072 /* Never */) === 0 ? restType : undefined;
            }
            function getTypeOfFirstParameterOfSignature(signature) {
                return getTypeOfFirstParameterOfSignatureWithFallback(signature, neverType);
            }
            function getTypeOfFirstParameterOfSignatureWithFallback(signature, fallbackType) {
                return signature.parameters.length > 0 ? getTypeAtPosition(signature, 0) : fallbackType;
            }
            function inferFromAnnotatedParameters(signature, context, inferenceContext) {
                var len = signature.parameters.length - (signatureHasRestParameter(signature) ? 1 : 0);
                for (var i = 0; i < len; i++) {
                    var declaration = signature.parameters[i].valueDeclaration;
                    if (declaration.type) {
                        var typeNode = ts.getEffectiveTypeAnnotationNode(declaration);
                        if (typeNode) {
                            inferTypes(inferenceContext.inferences, getTypeFromTypeNode(typeNode), getTypeAtPosition(context, i));
                        }
                    }
                }
                var restType = getEffectiveRestType(context);
                if (restType && restType.flags & 262144 /* TypeParameter */) {
                    // The contextual signature has a generic rest parameter. We first instantiate the contextual
                    // signature (without fixing type parameters) and assign types to contextually typed parameters.
                    var instantiatedContext = instantiateSignature(context, inferenceContext.nonFixingMapper);
                    assignContextualParameterTypes(signature, instantiatedContext);
                    // We then infer from a tuple type representing the parameters that correspond to the contextual
                    // rest parameter.
                    var restPos = getParameterCount(context) - 1;
                    inferTypes(inferenceContext.inferences, getRestTypeAtPosition(signature, restPos), restType);
                }
            }
            function assignContextualParameterTypes(signature, context) {
                if (context.typeParameters) {
                    if (!signature.typeParameters) {
                        signature.typeParameters = context.typeParameters;
                    }
                    else {
                        return; // This signature has already has a contextual inference performed and cached on it!
                    }
                }
                if (context.thisParameter) {
                    var parameter = signature.thisParameter;
                    if (!parameter || parameter.valueDeclaration && !parameter.valueDeclaration.type) {
                        if (!parameter) {
                            signature.thisParameter = createSymbolWithType(context.thisParameter, /*type*/ undefined);
                        }
                        assignParameterType(signature.thisParameter, getTypeOfSymbol(context.thisParameter));
                    }
                }
                var len = signature.parameters.length - (signatureHasRestParameter(signature) ? 1 : 0);
                for (var i = 0; i < len; i++) {
                    var parameter = signature.parameters[i];
                    if (!ts.getEffectiveTypeAnnotationNode(parameter.valueDeclaration)) {
                        var contextualParameterType = tryGetTypeAtPosition(context, i);
                        assignParameterType(parameter, contextualParameterType);
                    }
                }
                if (signatureHasRestParameter(signature)) {
                    // parameter might be a transient symbol generated by use of `arguments` in the function body.
                    var parameter = ts.last(signature.parameters);
                    if (ts.isTransientSymbol(parameter) || !ts.getEffectiveTypeAnnotationNode(parameter.valueDeclaration)) {
                        var contextualParameterType = getRestTypeAtPosition(context, len);
                        assignParameterType(parameter, contextualParameterType);
                    }
                }
            }
            function assignNonContextualParameterTypes(signature) {
                if (signature.thisParameter) {
                    assignParameterType(signature.thisParameter);
                }
                for (var _i = 0, _a = signature.parameters; _i < _a.length; _i++) {
                    var parameter = _a[_i];
                    assignParameterType(parameter);
                }
            }
            function assignParameterType(parameter, type) {
                var links = getSymbolLinks(parameter);
                if (!links.type) {
                    var declaration = parameter.valueDeclaration;
                    links.type = type || getWidenedTypeForVariableLikeDeclaration(declaration, /*includeOptionality*/ true);
                    if (declaration.name.kind !== 79 /* Identifier */) {
                        // if inference didn't come up with anything but unknown, fall back to the binding pattern if present.
                        if (links.type === unknownType) {
                            links.type = getTypeFromBindingPattern(declaration.name);
                        }
                        assignBindingElementTypes(declaration.name);
                    }
                }
            }
            // When contextual typing assigns a type to a parameter that contains a binding pattern, we also need to push
            // the destructured type into the contained binding elements.
            function assignBindingElementTypes(pattern) {
                for (var _i = 0, _a = pattern.elements; _i < _a.length; _i++) {
                    var element = _a[_i];
                    if (!ts.isOmittedExpression(element)) {
                        if (element.name.kind === 79 /* Identifier */) {
                            getSymbolLinks(getSymbolOfNode(element)).type = getTypeForBindingElement(element);
                        }
                        else {
                            assignBindingElementTypes(element.name);
                        }
                    }
                }
            }
            function createPromiseType(promisedType) {
                // creates a `Promise<T>` type where `T` is the promisedType argument
                var globalPromiseType = getGlobalPromiseType(/*reportErrors*/ true);
                if (globalPromiseType !== emptyGenericType) {
                    // if the promised type is itself a promise, get the underlying type; otherwise, fallback to the promised type
                    promisedType = getAwaitedType(promisedType) || unknownType;
                    return createTypeReference(globalPromiseType, [promisedType]);
                }
                return unknownType;
            }
            function createPromiseLikeType(promisedType) {
                // creates a `PromiseLike<T>` type where `T` is the promisedType argument
                var globalPromiseLikeType = getGlobalPromiseLikeType(/*reportErrors*/ true);
                if (globalPromiseLikeType !== emptyGenericType) {
                    // if the promised type is itself a promise, get the underlying type; otherwise, fallback to the promised type
                    promisedType = getAwaitedType(promisedType) || unknownType;
                    return createTypeReference(globalPromiseLikeType, [promisedType]);
                }
                return unknownType;
            }
            function createPromiseReturnType(func, promisedType) {
                var promiseType = createPromiseType(promisedType);
                if (promiseType === unknownType) {
                    error(func, ts.isImportCall(func) ?
                        ts.Diagnostics.A_dynamic_import_call_returns_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option :
                        ts.Diagnostics.An_async_function_or_method_must_return_a_Promise_Make_sure_you_have_a_declaration_for_Promise_or_include_ES2015_in_your_lib_option);
                    return errorType;
                }
                else if (!getGlobalPromiseConstructorSymbol(/*reportErrors*/ true)) {
                    error(func, ts.isImportCall(func) ?
                        ts.Diagnostics.A_dynamic_import_call_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option :
                        ts.Diagnostics.An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option);
                }
                return promiseType;
            }
            function createNewTargetExpressionType(targetType) {
                // Create a synthetic type `NewTargetExpression { target: TargetType; }`
                var symbol = createSymbol(0 /* None */, "NewTargetExpression");
                var targetPropertySymbol = createSymbol(4 /* Property */, "target", 8 /* Readonly */);
                targetPropertySymbol.parent = symbol;
                targetPropertySymbol.type = targetType;
                var members = ts.createSymbolTable([targetPropertySymbol]);
                symbol.members = members;
                return createAnonymousType(symbol, members, ts.emptyArray, ts.emptyArray, ts.emptyArray);
            }
            function getReturnTypeFromBody(func, checkMode) {
                if (!func.body) {
                    return errorType;
                }
                var functionFlags = ts.getFunctionFlags(func);
                var isAsync = (functionFlags & 2 /* Async */) !== 0;
                var isGenerator = (functionFlags & 1 /* Generator */) !== 0;
                var returnType;
                var yieldType;
                var nextType;
                var fallbackReturnType = voidType;
                if (func.body.kind !== 233 /* Block */) { // Async or normal arrow function
                    returnType = checkExpressionCached(func.body, checkMode && checkMode & ~8 /* SkipGenericFunctions */);
                    if (isAsync) {
                        // From within an async function you can return either a non-promise value or a promise. Any
                        // Promise/A+ compatible implementation will always assimilate any foreign promise, so the
                        // return type of the body should be unwrapped to its awaited type, which we will wrap in
                        // the native Promise<T> type later in this function.
                        returnType = checkAwaitedType(returnType, /*errorNode*/ func, ts.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member);
                    }
                }
                else if (isGenerator) { // Generator or AsyncGenerator function
                    var returnTypes = checkAndAggregateReturnExpressionTypes(func, checkMode);
                    if (!returnTypes) {
                        fallbackReturnType = neverType;
                    }
                    else if (returnTypes.length > 0) {
                        returnType = getUnionType(returnTypes, 2 /* Subtype */);
                    }
                    var _a = checkAndAggregateYieldOperandTypes(func, checkMode), yieldTypes = _a.yieldTypes, nextTypes = _a.nextTypes;
                    yieldType = ts.some(yieldTypes) ? getUnionType(yieldTypes, 2 /* Subtype */) : undefined;
                    nextType = ts.some(nextTypes) ? getIntersectionType(nextTypes) : undefined;
                }
                else { // Async or normal function
                    var types = checkAndAggregateReturnExpressionTypes(func, checkMode);
                    if (!types) {
                        // For an async function, the return type will not be never, but rather a Promise for never.
                        return functionFlags & 2 /* Async */
                            ? createPromiseReturnType(func, neverType) // Async function
                            : neverType; // Normal function
                    }
                    if (types.length === 0) {
                        // For an async function, the return type will not be void, but rather a Promise for void.
                        return functionFlags & 2 /* Async */
                            ? createPromiseReturnType(func, voidType) // Async function
                            : voidType; // Normal function
                    }
                    // Return a union of the return expression types.
                    returnType = getUnionType(types, 2 /* Subtype */);
                }
                if (returnType || yieldType || nextType) {
                    if (yieldType)
                        reportErrorsFromWidening(func, yieldType, 3 /* GeneratorYield */);
                    if (returnType)
                        reportErrorsFromWidening(func, returnType, 1 /* FunctionReturn */);
                    if (nextType)
                        reportErrorsFromWidening(func, nextType, 2 /* GeneratorNext */);
                    if (returnType && isUnitType(returnType) ||
                        yieldType && isUnitType(yieldType) ||
                        nextType && isUnitType(nextType)) {
                        var contextualSignature = getContextualSignatureForFunctionLikeDeclaration(func);
                        var contextualType = !contextualSignature ? undefined :
                            contextualSignature === getSignatureFromDeclaration(func) ? isGenerator ? undefined : returnType :
                                instantiateContextualType(getReturnTypeOfSignature(contextualSignature), func);
                        if (isGenerator) {
                            yieldType = getWidenedLiteralLikeTypeForContextualIterationTypeIfNeeded(yieldType, contextualType, 0 /* Yield */, isAsync);
                            returnType = getWidenedLiteralLikeTypeForContextualIterationTypeIfNeeded(returnType, contextualType, 1 /* Return */, isAsync);
                            nextType = getWidenedLiteralLikeTypeForContextualIterationTypeIfNeeded(nextType, contextualType, 2 /* Next */, isAsync);
                        }
                        else {
                            returnType = getWidenedLiteralLikeTypeForContextualReturnTypeIfNeeded(returnType, contextualType, isAsync);
                        }
                    }
                    if (yieldType)
                        yieldType = getWidenedType(yieldType);
                    if (returnType)
                        returnType = getWidenedType(returnType);
                    if (nextType)
                        nextType = getWidenedType(nextType);
                }
                if (isGenerator) {
                    return createGeneratorReturnType(yieldType || neverType, returnType || fallbackReturnType, nextType || getContextualIterationType(2 /* Next */, func) || unknownType, isAsync);
                }
                else {
                    // From within an async function you can return either a non-promise value or a promise. Any
                    // Promise/A+ compatible implementation will always assimilate any foreign promise, so the
                    // return type of the body is awaited type of the body, wrapped in a native Promise<T> type.
                    return isAsync
                        ? createPromiseType(returnType || fallbackReturnType)
                        : returnType || fallbackReturnType;
                }
            }
            function createGeneratorReturnType(yieldType, returnType, nextType, isAsyncGenerator) {
                var resolver = isAsyncGenerator ? asyncIterationTypesResolver : syncIterationTypesResolver;
                var globalGeneratorType = resolver.getGlobalGeneratorType(/*reportErrors*/ false);
                yieldType = resolver.resolveIterationType(yieldType, /*errorNode*/ undefined) || unknownType;
                returnType = resolver.resolveIterationType(returnType, /*errorNode*/ undefined) || unknownType;
                nextType = resolver.resolveIterationType(nextType, /*errorNode*/ undefined) || unknownType;
                if (globalGeneratorType === emptyGenericType) {
                    // Fall back to the global IterableIterator if returnType is assignable to the expected return iteration
                    // type of IterableIterator, and the expected next iteration type of IterableIterator is assignable to
                    // nextType.
                    var globalType = resolver.getGlobalIterableIteratorType(/*reportErrors*/ false);
                    var iterationTypes = globalType !== emptyGenericType ? getIterationTypesOfGlobalIterableType(globalType, resolver) : undefined;
                    var iterableIteratorReturnType = iterationTypes ? iterationTypes.returnType : anyType;
                    var iterableIteratorNextType = iterationTypes ? iterationTypes.nextType : undefinedType;
                    if (isTypeAssignableTo(returnType, iterableIteratorReturnType) &&
                        isTypeAssignableTo(iterableIteratorNextType, nextType)) {
                        if (globalType !== emptyGenericType) {
                            return createTypeFromGenericGlobalType(globalType, [yieldType]);
                        }
                        // The global IterableIterator type doesn't exist, so report an error
                        resolver.getGlobalIterableIteratorType(/*reportErrors*/ true);
                        return emptyObjectType;
                    }
                    // The global Generator type doesn't exist, so report an error
                    resolver.getGlobalGeneratorType(/*reportErrors*/ true);
                    return emptyObjectType;
                }
                return createTypeFromGenericGlobalType(globalGeneratorType, [yieldType, returnType, nextType]);
            }
            function checkAndAggregateYieldOperandTypes(func, checkMode) {
                var yieldTypes = [];
                var nextTypes = [];
                var isAsync = (ts.getFunctionFlags(func) & 2 /* Async */) !== 0;
                ts.forEachYieldExpression(func.body, function (yieldExpression) {
                    var yieldExpressionType = yieldExpression.expression ? checkExpression(yieldExpression.expression, checkMode) : undefinedWideningType;
                    ts.pushIfUnique(yieldTypes, getYieldedTypeOfYieldExpression(yieldExpression, yieldExpressionType, anyType, isAsync));
                    var nextType;
                    if (yieldExpression.asteriskToken) {
                        var iterationTypes = getIterationTypesOfIterable(yieldExpressionType, isAsync ? 19 /* AsyncYieldStar */ : 17 /* YieldStar */, yieldExpression.expression);
                        nextType = iterationTypes && iterationTypes.nextType;
                    }
                    else {
                        nextType = getContextualType(yieldExpression);
                    }
                    if (nextType)
                        ts.pushIfUnique(nextTypes, nextType);
                });
                return { yieldTypes: yieldTypes, nextTypes: nextTypes };
            }
            function getYieldedTypeOfYieldExpression(node, expressionType, sentType, isAsync) {
                var errorNode = node.expression || node;
                // A `yield*` expression effectively yields everything that its operand yields
                var yieldedType = node.asteriskToken ? checkIteratedTypeOrElementType(isAsync ? 19 /* AsyncYieldStar */ : 17 /* YieldStar */, expressionType, sentType, errorNode) : expressionType;
                return !isAsync ? yieldedType : getAwaitedType(yieldedType, errorNode, node.asteriskToken
                    ? ts.Diagnostics.Type_of_iterated_elements_of_a_yield_Asterisk_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member
                    : ts.Diagnostics.Type_of_yield_operand_in_an_async_generator_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member);
            }
            /**
             * Collect the TypeFacts learned from a typeof switch with
             * total clauses `witnesses`, and the active clause ranging
             * from `start` to `end`. Parameter `hasDefault` denotes
             * whether the active clause contains a default clause.
             */
            function getFactsFromTypeofSwitch(start, end, witnesses, hasDefault) {
                var facts = 0 /* None */;
                // When in the default we only collect inequality facts
                // because default is 'in theory' a set of infinite
                // equalities.
                if (hasDefault) {
                    // Value is not equal to any types after the active clause.
                    for (var i = end; i < witnesses.length; i++) {
                        facts |= typeofNEFacts.get(witnesses[i]) || 32768 /* TypeofNEHostObject */;
                    }
                    // Remove inequalities for types that appear in the
                    // active clause because they appear before other
                    // types collected so far.
                    for (var i = start; i < end; i++) {
                        facts &= ~(typeofNEFacts.get(witnesses[i]) || 0);
                    }
                    // Add inequalities for types before the active clause unconditionally.
                    for (var i = 0; i < start; i++) {
                        facts |= typeofNEFacts.get(witnesses[i]) || 32768 /* TypeofNEHostObject */;
                    }
                }
                // When in an active clause without default the set of
                // equalities is finite.
                else {
                    // Add equalities for all types in the active clause.
                    for (var i = start; i < end; i++) {
                        facts |= typeofEQFacts.get(witnesses[i]) || 128 /* TypeofEQHostObject */;
                    }
                    // Remove equalities for types that appear before the
                    // active clause.
                    for (var i = 0; i < start; i++) {
                        facts &= ~(typeofEQFacts.get(witnesses[i]) || 0);
                    }
                }
                return facts;
            }
            function isExhaustiveSwitchStatement(node) {
                var links = getNodeLinks(node);
                return links.isExhaustive !== undefined ? links.isExhaustive : (links.isExhaustive = computeExhaustiveSwitchStatement(node));
            }
            function computeExhaustiveSwitchStatement(node) {
                if (node.expression.kind === 214 /* TypeOfExpression */) {
                    var operandType = getTypeOfExpression(node.expression.expression);
                    var witnesses = getSwitchClauseTypeOfWitnesses(node, /*retainDefault*/ false);
                    // notEqualFacts states that the type of the switched value is not equal to every type in the switch.
                    var notEqualFacts_1 = getFactsFromTypeofSwitch(0, 0, witnesses, /*hasDefault*/ true);
                    var type_4 = getBaseConstraintOfType(operandType) || operandType;
                    // Take any/unknown as a special condition. Or maybe we could change `type` to a union containing all primitive types.
                    if (type_4.flags & 3 /* AnyOrUnknown */) {
                        return (556800 /* AllTypeofNE */ & notEqualFacts_1) === 556800 /* AllTypeofNE */;
                    }
                    return !!(filterType(type_4, function (t) { return (getTypeFacts(t) & notEqualFacts_1) === notEqualFacts_1; }).flags & 131072 /* Never */);
                }
                var type = getTypeOfExpression(node.expression);
                if (!isLiteralType(type)) {
                    return false;
                }
                var switchTypes = getSwitchClauseTypes(node);
                if (!switchTypes.length || ts.some(switchTypes, isNeitherUnitTypeNorNever)) {
                    return false;
                }
                return eachTypeContainedIn(mapType(type, getRegularTypeOfLiteralType), switchTypes);
            }
            function functionHasImplicitReturn(func) {
                return func.endFlowNode && isReachableFlowNode(func.endFlowNode);
            }
            /** NOTE: Return value of `[]` means a different thing than `undefined`. `[]` means func returns `void`, `undefined` means it returns `never`. */
            function checkAndAggregateReturnExpressionTypes(func, checkMode) {
                var functionFlags = ts.getFunctionFlags(func);
                var aggregatedTypes = [];
                var hasReturnWithNoExpression = functionHasImplicitReturn(func);
                var hasReturnOfTypeNever = false;
                ts.forEachReturnStatement(func.body, function (returnStatement) {
                    var expr = returnStatement.expression;
                    if (expr) {
                        var type = checkExpressionCached(expr, checkMode && checkMode & ~8 /* SkipGenericFunctions */);
                        if (functionFlags & 2 /* Async */) {
                            // From within an async function you can return either a non-promise value or a promise. Any
                            // Promise/A+ compatible implementation will always assimilate any foreign promise, so the
                            // return type of the body should be unwrapped to its awaited type, which should be wrapped in
                            // the native Promise<T> type by the caller.
                            type = checkAwaitedType(type, func, ts.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member);
                        }
                        if (type.flags & 131072 /* Never */) {
                            hasReturnOfTypeNever = true;
                        }
                        ts.pushIfUnique(aggregatedTypes, type);
                    }
                    else {
                        hasReturnWithNoExpression = true;
                    }
                });
                if (aggregatedTypes.length === 0 && !hasReturnWithNoExpression && (hasReturnOfTypeNever || mayReturnNever(func))) {
                    return undefined;
                }
                if (strictNullChecks && aggregatedTypes.length && hasReturnWithNoExpression &&
                    !(isJSConstructor(func) && aggregatedTypes.some(function (t) { return t.symbol === func.symbol; }))) {
                    // Javascript "callable constructors", containing eg `if (!(this instanceof A)) return new A()` should not add undefined
                    ts.pushIfUnique(aggregatedTypes, undefinedType);
                }
                return aggregatedTypes;
            }
            function mayReturnNever(func) {
                switch (func.kind) {
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        return true;
                    case 167 /* MethodDeclaration */:
                        return func.parent.kind === 203 /* ObjectLiteralExpression */;
                    default:
                        return false;
                }
            }
            /**
             * TypeScript Specification 1.0 (6.3) - July 2014
             *   An explicitly typed function whose return type isn't the Void type,
             *   the Any type, or a union type containing the Void or Any type as a constituent
             *   must have at least one return statement somewhere in its body.
             *   An exception to this rule is if the function implementation consists of a single 'throw' statement.
             *
             * @param returnType - return type of the function, can be undefined if return type is not explicitly specified
             */
            function checkAllCodePathsInNonVoidFunctionReturnOrThrow(func, returnType) {
                if (!produceDiagnostics) {
                    return;
                }
                var functionFlags = ts.getFunctionFlags(func);
                var type = returnType && unwrapReturnType(returnType, functionFlags);
                // Functions with with an explicitly specified 'void' or 'any' return type don't need any return expressions.
                if (type && maybeTypeOfKind(type, 1 /* Any */ | 16384 /* Void */)) {
                    return;
                }
                // If all we have is a function signature, or an arrow function with an expression body, then there is nothing to check.
                // also if HasImplicitReturn flag is not set this means that all codepaths in function body end with return or throw
                if (func.kind === 166 /* MethodSignature */ || ts.nodeIsMissing(func.body) || func.body.kind !== 233 /* Block */ || !functionHasImplicitReturn(func)) {
                    return;
                }
                var hasExplicitReturn = func.flags & 512 /* HasExplicitReturn */;
                var errorNode = ts.getEffectiveReturnTypeNode(func) || func;
                if (type && type.flags & 131072 /* Never */) {
                    error(errorNode, ts.Diagnostics.A_function_returning_never_cannot_have_a_reachable_end_point);
                }
                else if (type && !hasExplicitReturn) {
                    // minimal check: function has syntactic return type annotation and no explicit return statements in the body
                    // this function does not conform to the specification.
                    error(errorNode, ts.Diagnostics.A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value);
                }
                else if (type && strictNullChecks && !isTypeAssignableTo(undefinedType, type)) {
                    error(errorNode, ts.Diagnostics.Function_lacks_ending_return_statement_and_return_type_does_not_include_undefined);
                }
                else if (compilerOptions.noImplicitReturns) {
                    if (!type) {
                        // If return type annotation is omitted check if function has any explicit return statements.
                        // If it does not have any - its inferred return type is void - don't do any checks.
                        // Otherwise get inferred return type from function body and report error only if it is not void / anytype
                        if (!hasExplicitReturn) {
                            return;
                        }
                        var inferredReturnType = getReturnTypeOfSignature(getSignatureFromDeclaration(func));
                        if (isUnwrappedReturnTypeVoidOrAny(func, inferredReturnType)) {
                            return;
                        }
                    }
                    error(errorNode, ts.Diagnostics.Not_all_code_paths_return_a_value);
                }
            }
            function checkFunctionExpressionOrObjectLiteralMethod(node, checkMode) {
                ts.Debug.assert(node.kind !== 167 /* MethodDeclaration */ || ts.isObjectLiteralMethod(node));
                checkNodeDeferred(node);
                if (ts.isFunctionExpression(node)) {
                    checkCollisionsForDeclarationName(node, node.name);
                }
                // The identityMapper object is used to indicate that function expressions are wildcards
                if (checkMode && checkMode & 4 /* SkipContextSensitive */ && isContextSensitive(node)) {
                    // Skip parameters, return signature with return type that retains noncontextual parts so inferences can still be drawn in an early stage
                    if (!ts.getEffectiveReturnTypeNode(node) && !ts.hasContextSensitiveParameters(node)) {
                        // Return plain anyFunctionType if there is no possibility we'll make inferences from the return type
                        var contextualSignature = getContextualSignature(node);
                        if (contextualSignature && couldContainTypeVariables(getReturnTypeOfSignature(contextualSignature))) {
                            var links = getNodeLinks(node);
                            if (links.contextFreeType) {
                                return links.contextFreeType;
                            }
                            var returnType = getReturnTypeFromBody(node, checkMode);
                            var returnOnlySignature = createSignature(undefined, undefined, undefined, ts.emptyArray, returnType, /*resolvedTypePredicate*/ undefined, 0, 0 /* None */);
                            var returnOnlyType = createAnonymousType(node.symbol, emptySymbols, [returnOnlySignature], ts.emptyArray, ts.emptyArray);
                            returnOnlyType.objectFlags |= 524288 /* NonInferrableType */;
                            return links.contextFreeType = returnOnlyType;
                        }
                    }
                    return anyFunctionType;
                }
                // Grammar checking
                var hasGrammarError = checkGrammarFunctionLikeDeclaration(node);
                if (!hasGrammarError && node.kind === 211 /* FunctionExpression */) {
                    checkGrammarForGenerator(node);
                }
                contextuallyCheckFunctionExpressionOrObjectLiteralMethod(node, checkMode);
                return getTypeOfSymbol(getSymbolOfNode(node));
            }
            function contextuallyCheckFunctionExpressionOrObjectLiteralMethod(node, checkMode) {
                var links = getNodeLinks(node);
                // Check if function expression is contextually typed and assign parameter types if so.
                if (!(links.flags & 1024 /* ContextChecked */)) {
                    var contextualSignature = getContextualSignature(node);
                    // If a type check is started at a function expression that is an argument of a function call, obtaining the
                    // contextual type may recursively get back to here during overload resolution of the call. If so, we will have
                    // already assigned contextual types.
                    if (!(links.flags & 1024 /* ContextChecked */)) {
                        links.flags |= 1024 /* ContextChecked */;
                        var signature = ts.firstOrUndefined(getSignaturesOfType(getTypeOfSymbol(getSymbolOfNode(node)), 0 /* Call */));
                        if (!signature) {
                            return;
                        }
                        if (isContextSensitive(node)) {
                            if (contextualSignature) {
                                var inferenceContext = getInferenceContext(node);
                                if (checkMode && checkMode & 2 /* Inferential */) {
                                    inferFromAnnotatedParameters(signature, contextualSignature, inferenceContext);
                                }
                                var instantiatedContextualSignature = inferenceContext ?
                                    instantiateSignature(contextualSignature, inferenceContext.mapper) : contextualSignature;
                                assignContextualParameterTypes(signature, instantiatedContextualSignature);
                            }
                            else {
                                // Force resolution of all parameter types such that the absence of a contextual type is consistently reflected.
                                assignNonContextualParameterTypes(signature);
                            }
                        }
                        if (contextualSignature && !getReturnTypeFromAnnotation(node) && !signature.resolvedReturnType) {
                            var returnType = getReturnTypeFromBody(node, checkMode);
                            if (!signature.resolvedReturnType) {
                                signature.resolvedReturnType = returnType;
                            }
                        }
                        checkSignatureDeclaration(node);
                    }
                }
            }
            function checkFunctionExpressionOrObjectLiteralMethodDeferred(node) {
                ts.Debug.assert(node.kind !== 167 /* MethodDeclaration */ || ts.isObjectLiteralMethod(node));
                var functionFlags = ts.getFunctionFlags(node);
                var returnType = getReturnTypeFromAnnotation(node);
                checkAllCodePathsInNonVoidFunctionReturnOrThrow(node, returnType);
                if (node.body) {
                    if (!ts.getEffectiveReturnTypeNode(node)) {
                        // There are some checks that are only performed in getReturnTypeFromBody, that may produce errors
                        // we need. An example is the noImplicitAny errors resulting from widening the return expression
                        // of a function. Because checking of function expression bodies is deferred, there was never an
                        // appropriate time to do this during the main walk of the file (see the comment at the top of
                        // checkFunctionExpressionBodies). So it must be done now.
                        getReturnTypeOfSignature(getSignatureFromDeclaration(node));
                    }
                    if (node.body.kind === 233 /* Block */) {
                        checkSourceElement(node.body);
                    }
                    else {
                        // From within an async function you can return either a non-promise value or a promise. Any
                        // Promise/A+ compatible implementation will always assimilate any foreign promise, so we
                        // should not be checking assignability of a promise to the return type. Instead, we need to
                        // check assignability of the awaited type of the expression body against the promised type of
                        // its return type annotation.
                        var exprType = checkExpression(node.body);
                        var returnOrPromisedType = returnType && unwrapReturnType(returnType, functionFlags);
                        if (returnOrPromisedType) {
                            if ((functionFlags & 3 /* AsyncGenerator */) === 2 /* Async */) { // Async function
                                var awaitedType = checkAwaitedType(exprType, node.body, ts.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member);
                                checkTypeAssignableToAndOptionallyElaborate(awaitedType, returnOrPromisedType, node.body, node.body);
                            }
                            else { // Normal function
                                checkTypeAssignableToAndOptionallyElaborate(exprType, returnOrPromisedType, node.body, node.body);
                            }
                        }
                    }
                }
            }
            function checkArithmeticOperandType(operand, type, diagnostic, isAwaitValid) {
                if (isAwaitValid === void 0) {
                    isAwaitValid = false;
                }
                if (!isTypeAssignableTo(type, numberOrBigIntType)) {
                    var awaitedType = isAwaitValid && getAwaitedTypeOfPromise(type);
                    errorAndMaybeSuggestAwait(operand, !!awaitedType && isTypeAssignableTo(awaitedType, numberOrBigIntType), diagnostic);
                    return false;
                }
                return true;
            }
            function isReadonlyAssignmentDeclaration(d) {
                if (!ts.isCallExpression(d)) {
                    return false;
                }
                if (!ts.isBindableObjectDefinePropertyCall(d)) {
                    return false;
                }
                var objectLitType = checkExpressionCached(d.arguments[2]);
                var valueType = getTypeOfPropertyOfType(objectLitType, "value");
                if (valueType) {
                    var writableProp = getPropertyOfType(objectLitType, "writable");
                    var writableType = writableProp && getTypeOfSymbol(writableProp);
                    if (!writableType || writableType === falseType || writableType === regularFalseType) {
                        return true;
                    }
                    // We include this definition whereupon we walk back and check the type at the declaration because
                    // The usual definition of `Object.defineProperty` will _not_ cause literal types to be preserved in the
                    // argument types, should the type be contextualized by the call itself.
                    if (writableProp && writableProp.valueDeclaration && ts.isPropertyAssignment(writableProp.valueDeclaration)) {
                        var initializer = writableProp.valueDeclaration.initializer;
                        var rawOriginalType = checkExpression(initializer);
                        if (rawOriginalType === falseType || rawOriginalType === regularFalseType) {
                            return true;
                        }
                    }
                    return false;
                }
                var setProp = getPropertyOfType(objectLitType, "set");
                return !setProp;
            }
            function isReadonlySymbol(symbol) {
                // The following symbols are considered read-only:
                // Properties with a 'readonly' modifier
                // Variables declared with 'const'
                // Get accessors without matching set accessors
                // Enum members
                // Object.defineProperty assignments with writable false or no setter
                // Unions and intersections of the above (unions and intersections eagerly set isReadonly on creation)
                return !!(ts.getCheckFlags(symbol) & 8 /* Readonly */ ||
                    symbol.flags & 4 /* Property */ && ts.getDeclarationModifierFlagsFromSymbol(symbol) & 64 /* Readonly */ ||
                    symbol.flags & 3 /* Variable */ && getDeclarationNodeFlagsFromSymbol(symbol) & 2 /* Const */ ||
                    symbol.flags & 98304 /* Accessor */ && !(symbol.flags & 65536 /* SetAccessor */) ||
                    symbol.flags & 8 /* EnumMember */ ||
                    ts.some(symbol.declarations, isReadonlyAssignmentDeclaration));
            }
            function isAssignmentToReadonlyEntity(expr, symbol, assignmentKind) {
                var _a, _b;
                if (assignmentKind === 0 /* None */) {
                    // no assigment means it doesn't matter whether the entity is readonly
                    return false;
                }
                if (isReadonlySymbol(symbol)) {
                    // Allow assignments to readonly properties within constructors of the same class declaration.
                    if (symbol.flags & 4 /* Property */ &&
                        ts.isAccessExpression(expr) &&
                        expr.expression.kind === 108 /* ThisKeyword */) {
                        // Look for if this is the constructor for the class that `symbol` is a property of.
                        var ctor = ts.getContainingFunction(expr);
                        if (!(ctor && (ctor.kind === 169 /* Constructor */ || isJSConstructor(ctor)))) {
                            return true;
                        }
                        if (symbol.valueDeclaration) {
                            var isAssignmentDeclaration_1 = ts.isBinaryExpression(symbol.valueDeclaration);
                            var isLocalPropertyDeclaration = ctor.parent === symbol.valueDeclaration.parent;
                            var isLocalParameterProperty = ctor === symbol.valueDeclaration.parent;
                            var isLocalThisPropertyAssignment = isAssignmentDeclaration_1 && ((_a = symbol.parent) === null || _a === void 0 ? void 0 : _a.valueDeclaration) === ctor.parent;
                            var isLocalThisPropertyAssignmentConstructorFunction = isAssignmentDeclaration_1 && ((_b = symbol.parent) === null || _b === void 0 ? void 0 : _b.valueDeclaration) === ctor;
                            var isWriteableSymbol = isLocalPropertyDeclaration
                                || isLocalParameterProperty
                                || isLocalThisPropertyAssignment
                                || isLocalThisPropertyAssignmentConstructorFunction;
                            return !isWriteableSymbol;
                        }
                    }
                    return true;
                }
                if (ts.isAccessExpression(expr)) {
                    // references through namespace import should be readonly
                    var node = ts.skipParentheses(expr.expression);
                    if (node.kind === 79 /* Identifier */) {
                        var symbol_2 = getNodeLinks(node).resolvedSymbol;
                        if (symbol_2.flags & 2097152 /* Alias */) {
                            var declaration = getDeclarationOfAliasSymbol(symbol_2);
                            return !!declaration && declaration.kind === 266 /* NamespaceImport */;
                        }
                    }
                }
                return false;
            }
            function checkReferenceExpression(expr, invalidReferenceMessage, invalidOptionalChainMessage) {
                // References are combinations of identifiers, parentheses, and property accesses.
                var node = ts.skipOuterExpressions(expr, 6 /* Assertions */ | 1 /* Parentheses */);
                if (node.kind !== 79 /* Identifier */ && !ts.isAccessExpression(node)) {
                    error(expr, invalidReferenceMessage);
                    return false;
                }
                if (node.flags & 32 /* OptionalChain */) {
                    error(expr, invalidOptionalChainMessage);
                    return false;
                }
                return true;
            }
            function checkDeleteExpression(node) {
                checkExpression(node.expression);
                var expr = ts.skipParentheses(node.expression);
                if (!ts.isAccessExpression(expr)) {
                    error(expr, ts.Diagnostics.The_operand_of_a_delete_operator_must_be_a_property_reference);
                    return booleanType;
                }
                if (ts.isPropertyAccessExpression(expr) && ts.isPrivateIdentifier(expr.name)) {
                    error(expr, ts.Diagnostics.The_operand_of_a_delete_operator_cannot_be_a_private_identifier);
                }
                var links = getNodeLinks(expr);
                var symbol = getExportSymbolOfValueSymbolIfExported(links.resolvedSymbol);
                if (symbol) {
                    if (isReadonlySymbol(symbol)) {
                        error(expr, ts.Diagnostics.The_operand_of_a_delete_operator_cannot_be_a_read_only_property);
                    }
                    checkDeleteExpressionMustBeOptional(expr, symbol);
                }
                return booleanType;
            }
            function checkDeleteExpressionMustBeOptional(expr, symbol) {
                var type = getTypeOfSymbol(symbol);
                if (strictNullChecks &&
                    !(type.flags & (3 /* AnyOrUnknown */ | 131072 /* Never */)) &&
                    !(exactOptionalPropertyTypes ? symbol.flags & 16777216 /* Optional */ : getFalsyFlags(type) & 32768 /* Undefined */)) {
                    error(expr, ts.Diagnostics.The_operand_of_a_delete_operator_must_be_optional);
                }
            }
            function checkTypeOfExpression(node) {
                checkExpression(node.expression);
                return typeofType;
            }
            function checkVoidExpression(node) {
                checkExpression(node.expression);
                return undefinedWideningType;
            }
            function checkAwaitExpression(node) {
                // Grammar checking
                if (produceDiagnostics) {
                    var container = ts.getContainingFunctionOrClassStaticBlock(node);
                    if (container && ts.isClassStaticBlockDeclaration(container)) {
                        error(node, ts.Diagnostics.Await_expression_cannot_be_used_inside_a_class_static_block);
                    }
                    else if (!(node.flags & 32768 /* AwaitContext */)) {
                        if (ts.isInTopLevelContext(node)) {
                            var sourceFile = ts.getSourceFileOfNode(node);
                            if (!hasParseDiagnostics(sourceFile)) {
                                var span = void 0;
                                if (!ts.isEffectiveExternalModule(sourceFile, compilerOptions)) {
                                    if (!span)
                                        span = ts.getSpanOfTokenAtPosition(sourceFile, node.pos);
                                    var diagnostic = ts.createFileDiagnostic(sourceFile, span.start, span.length, ts.Diagnostics.await_expressions_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_has_no_imports_or_exports_Consider_adding_an_empty_export_to_make_this_file_a_module);
                                    diagnostics.add(diagnostic);
                                }
                                if ((moduleKind !== ts.ModuleKind.ESNext && moduleKind !== ts.ModuleKind.System) || languageVersion < 4 /* ES2017 */) {
                                    span = ts.getSpanOfTokenAtPosition(sourceFile, node.pos);
                                    var diagnostic = ts.createFileDiagnostic(sourceFile, span.start, span.length, ts.Diagnostics.Top_level_await_expressions_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_target_option_is_set_to_es2017_or_higher);
                                    diagnostics.add(diagnostic);
                                }
                            }
                        }
                        else {
                            // use of 'await' in non-async function
                            var sourceFile = ts.getSourceFileOfNode(node);
                            if (!hasParseDiagnostics(sourceFile)) {
                                var span = ts.getSpanOfTokenAtPosition(sourceFile, node.pos);
                                var diagnostic = ts.createFileDiagnostic(sourceFile, span.start, span.length, ts.Diagnostics.await_expressions_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules);
                                if (container && container.kind !== 169 /* Constructor */ && (ts.getFunctionFlags(container) & 2 /* Async */) === 0) {
                                    var relatedInfo = ts.createDiagnosticForNode(container, ts.Diagnostics.Did_you_mean_to_mark_this_function_as_async);
                                    ts.addRelatedInfo(diagnostic, relatedInfo);
                                }
                                diagnostics.add(diagnostic);
                            }
                        }
                    }
                    if (isInParameterInitializerBeforeContainingFunction(node)) {
                        error(node, ts.Diagnostics.await_expressions_cannot_be_used_in_a_parameter_initializer);
                    }
                }
                var operandType = checkExpression(node.expression);
                var awaitedType = checkAwaitedType(operandType, node, ts.Diagnostics.Type_of_await_operand_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member);
                if (awaitedType === operandType && awaitedType !== errorType && !(operandType.flags & 3 /* AnyOrUnknown */)) {
                    addErrorOrSuggestion(/*isError*/ false, ts.createDiagnosticForNode(node, ts.Diagnostics.await_has_no_effect_on_the_type_of_this_expression));
                }
                return awaitedType;
            }
            function checkPrefixUnaryExpression(node) {
                var operandType = checkExpression(node.operand);
                if (operandType === silentNeverType) {
                    return silentNeverType;
                }
                switch (node.operand.kind) {
                    case 8 /* NumericLiteral */:
                        switch (node.operator) {
                            case 40 /* MinusToken */:
                                return getFreshTypeOfLiteralType(getNumberLiteralType(-node.operand.text));
                            case 39 /* PlusToken */:
                                return getFreshTypeOfLiteralType(getNumberLiteralType(+node.operand.text));
                        }
                        break;
                    case 9 /* BigIntLiteral */:
                        if (node.operator === 40 /* MinusToken */) {
                            return getFreshTypeOfLiteralType(getBigIntLiteralType({
                                negative: true,
                                base10Value: ts.parsePseudoBigInt(node.operand.text)
                            }));
                        }
                }
                switch (node.operator) {
                    case 39 /* PlusToken */:
                    case 40 /* MinusToken */:
                    case 54 /* TildeToken */:
                        checkNonNullType(operandType, node.operand);
                        if (maybeTypeOfKind(operandType, 12288 /* ESSymbolLike */)) {
                            error(node.operand, ts.Diagnostics.The_0_operator_cannot_be_applied_to_type_symbol, ts.tokenToString(node.operator));
                        }
                        if (node.operator === 39 /* PlusToken */) {
                            if (maybeTypeOfKind(operandType, 2112 /* BigIntLike */)) {
                                error(node.operand, ts.Diagnostics.Operator_0_cannot_be_applied_to_type_1, ts.tokenToString(node.operator), typeToString(getBaseTypeOfLiteralType(operandType)));
                            }
                            return numberType;
                        }
                        return getUnaryResultType(operandType);
                    case 53 /* ExclamationToken */:
                        checkTruthinessExpression(node.operand);
                        var facts = getTypeFacts(operandType) & (4194304 /* Truthy */ | 8388608 /* Falsy */);
                        return facts === 4194304 /* Truthy */ ? falseType :
                            facts === 8388608 /* Falsy */ ? trueType :
                                booleanType;
                    case 45 /* PlusPlusToken */:
                    case 46 /* MinusMinusToken */:
                        var ok = checkArithmeticOperandType(node.operand, checkNonNullType(operandType, node.operand), ts.Diagnostics.An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type);
                        if (ok) {
                            // run check only if former checks succeeded to avoid reporting cascading errors
                            checkReferenceExpression(node.operand, ts.Diagnostics.The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access, ts.Diagnostics.The_operand_of_an_increment_or_decrement_operator_may_not_be_an_optional_property_access);
                        }
                        return getUnaryResultType(operandType);
                }
                return errorType;
            }
            function checkPostfixUnaryExpression(node) {
                var operandType = checkExpression(node.operand);
                if (operandType === silentNeverType) {
                    return silentNeverType;
                }
                var ok = checkArithmeticOperandType(node.operand, checkNonNullType(operandType, node.operand), ts.Diagnostics.An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type);
                if (ok) {
                    // run check only if former checks succeeded to avoid reporting cascading errors
                    checkReferenceExpression(node.operand, ts.Diagnostics.The_operand_of_an_increment_or_decrement_operator_must_be_a_variable_or_a_property_access, ts.Diagnostics.The_operand_of_an_increment_or_decrement_operator_may_not_be_an_optional_property_access);
                }
                return getUnaryResultType(operandType);
            }
            function getUnaryResultType(operandType) {
                if (maybeTypeOfKind(operandType, 2112 /* BigIntLike */)) {
                    return isTypeAssignableToKind(operandType, 3 /* AnyOrUnknown */) || maybeTypeOfKind(operandType, 296 /* NumberLike */)
                        ? numberOrBigIntType
                        : bigintType;
                }
                // If it's not a bigint type, implicit coercion will result in a number
                return numberType;
            }
            // Return true if type might be of the given kind. A union or intersection type might be of a given
            // kind if at least one constituent type is of the given kind.
            function maybeTypeOfKind(type, kind) {
                if (type.flags & kind) {
                    return true;
                }
                if (type.flags & 3145728 /* UnionOrIntersection */) {
                    var types = type.types;
                    for (var _i = 0, types_21 = types; _i < types_21.length; _i++) {
                        var t = types_21[_i];
                        if (maybeTypeOfKind(t, kind)) {
                            return true;
                        }
                    }
                }
                return false;
            }
            function isTypeAssignableToKind(source, kind, strict) {
                if (source.flags & kind) {
                    return true;
                }
                if (strict && source.flags & (3 /* AnyOrUnknown */ | 16384 /* Void */ | 32768 /* Undefined */ | 65536 /* Null */)) {
                    return false;
                }
                return !!(kind & 296 /* NumberLike */) && isTypeAssignableTo(source, numberType) ||
                    !!(kind & 2112 /* BigIntLike */) && isTypeAssignableTo(source, bigintType) ||
                    !!(kind & 402653316 /* StringLike */) && isTypeAssignableTo(source, stringType) ||
                    !!(kind & 528 /* BooleanLike */) && isTypeAssignableTo(source, booleanType) ||
                    !!(kind & 16384 /* Void */) && isTypeAssignableTo(source, voidType) ||
                    !!(kind & 131072 /* Never */) && isTypeAssignableTo(source, neverType) ||
                    !!(kind & 65536 /* Null */) && isTypeAssignableTo(source, nullType) ||
                    !!(kind & 32768 /* Undefined */) && isTypeAssignableTo(source, undefinedType) ||
                    !!(kind & 4096 /* ESSymbol */) && isTypeAssignableTo(source, esSymbolType) ||
                    !!(kind & 67108864 /* NonPrimitive */) && isTypeAssignableTo(source, nonPrimitiveType);
            }
            function allTypesAssignableToKind(source, kind, strict) {
                return source.flags & 1048576 /* Union */ ?
                    ts.every(source.types, function (subType) { return allTypesAssignableToKind(subType, kind, strict); }) :
                    isTypeAssignableToKind(source, kind, strict);
            }
            function isConstEnumObjectType(type) {
                return !!(ts.getObjectFlags(type) & 16 /* Anonymous */) && !!type.symbol && isConstEnumSymbol(type.symbol);
            }
            function isConstEnumSymbol(symbol) {
                return (symbol.flags & 128 /* ConstEnum */) !== 0;
            }
            function checkInstanceOfExpression(left, right, leftType, rightType) {
                if (leftType === silentNeverType || rightType === silentNeverType) {
                    return silentNeverType;
                }
                // TypeScript 1.0 spec (April 2014): 4.15.4
                // The instanceof operator requires the left operand to be of type Any, an object type, or a type parameter type,
                // and the right operand to be of type Any, a subtype of the 'Function' interface type, or have a call or construct signature.
                // The result is always of the Boolean primitive type.
                // NOTE: do not raise error if leftType is unknown as related error was already reported
                if (!isTypeAny(leftType) &&
                    allTypesAssignableToKind(leftType, 131068 /* Primitive */)) {
                    error(left, ts.Diagnostics.The_left_hand_side_of_an_instanceof_expression_must_be_of_type_any_an_object_type_or_a_type_parameter);
                }
                // NOTE: do not raise error if right is unknown as related error was already reported
                if (!(isTypeAny(rightType) || typeHasCallOrConstructSignatures(rightType) || isTypeSubtypeOf(rightType, globalFunctionType))) {
                    error(right, ts.Diagnostics.The_right_hand_side_of_an_instanceof_expression_must_be_of_type_any_or_of_a_type_assignable_to_the_Function_interface_type);
                }
                return booleanType;
            }
            function checkInExpression(left, right, leftType, rightType) {
                if (leftType === silentNeverType || rightType === silentNeverType) {
                    return silentNeverType;
                }
                leftType = checkNonNullType(leftType, left);
                rightType = checkNonNullType(rightType, right);
                // TypeScript 1.0 spec (April 2014): 4.15.5
                // The in operator requires the left operand to be of type Any, the String primitive type, or the Number primitive type,
                // and the right operand to be
                //
                //   1. assignable to the non-primitive type,
                //   2. an unconstrained type parameter,
                //   3. a union or intersection including one or more type parameters, whose constituents are all assignable to the
                //      the non-primitive type, or are unconstrainted type parameters, or have constraints assignable to the
                //      non-primitive type, or
                //   4. a type parameter whose constraint is
                //      i. an object type,
                //     ii. the non-primitive type, or
                //    iii. a union or intersection with at least one constituent assignable to an object or non-primitive type.
                //
                // The divergent behavior for type parameters and unions containing type parameters is a workaround for type
                // parameters not being narrowable. If the right operand is a concrete type, we can error if there is any chance
                // it is a primitive. But if the operand is a type parameter, it cannot be narrowed, so we don't issue an error
                // unless *all* instantiations would result in an error.
                //
                // The result is always of the Boolean primitive type.
                if (!(allTypesAssignableToKind(leftType, 402653316 /* StringLike */ | 296 /* NumberLike */ | 12288 /* ESSymbolLike */) ||
                    isTypeAssignableToKind(leftType, 4194304 /* Index */ | 134217728 /* TemplateLiteral */ | 268435456 /* StringMapping */ | 262144 /* TypeParameter */))) {
                    error(left, ts.Diagnostics.The_left_hand_side_of_an_in_expression_must_be_of_type_any_string_number_or_symbol);
                }
                var rightTypeConstraint = getConstraintOfType(rightType);
                if (!allTypesAssignableToKind(rightType, 67108864 /* NonPrimitive */ | 58982400 /* InstantiableNonPrimitive */) ||
                    rightTypeConstraint && (isTypeAssignableToKind(rightType, 3145728 /* UnionOrIntersection */) && !allTypesAssignableToKind(rightTypeConstraint, 67108864 /* NonPrimitive */ | 58982400 /* InstantiableNonPrimitive */) ||
                        !maybeTypeOfKind(rightTypeConstraint, 67108864 /* NonPrimitive */ | 58982400 /* InstantiableNonPrimitive */ | 524288 /* Object */))) {
                    error(right, ts.Diagnostics.The_right_hand_side_of_an_in_expression_must_not_be_a_primitive);
                }
                return booleanType;
            }
            function checkObjectLiteralAssignment(node, sourceType, rightIsThis) {
                var properties = node.properties;
                if (strictNullChecks && properties.length === 0) {
                    return checkNonNullType(sourceType, node);
                }
                for (var i = 0; i < properties.length; i++) {
                    checkObjectLiteralDestructuringPropertyAssignment(node, sourceType, i, properties, rightIsThis);
                }
                return sourceType;
            }
            /** Note: If property cannot be a SpreadAssignment, then allProperties does not need to be provided */
            function checkObjectLiteralDestructuringPropertyAssignment(node, objectLiteralType, propertyIndex, allProperties, rightIsThis) {
                if (rightIsThis === void 0) {
                    rightIsThis = false;
                }
                var properties = node.properties;
                var property = properties[propertyIndex];
                if (property.kind === 291 /* PropertyAssignment */ || property.kind === 292 /* ShorthandPropertyAssignment */) {
                    var name = property.name;
                    var exprType = getLiteralTypeFromPropertyName(name);
                    if (isTypeUsableAsPropertyName(exprType)) {
                        var text = getPropertyNameFromType(exprType);
                        var prop = getPropertyOfType(objectLiteralType, text);
                        if (prop) {
                            markPropertyAsReferenced(prop, property, rightIsThis);
                            checkPropertyAccessibility(property, /*isSuper*/ false, /*writing*/ true, objectLiteralType, prop);
                        }
                    }
                    var elementType = getIndexedAccessType(objectLiteralType, exprType, 32 /* ExpressionPosition */, name);
                    var type = getFlowTypeOfDestructuring(property, elementType);
                    return checkDestructuringAssignment(property.kind === 292 /* ShorthandPropertyAssignment */ ? property : property.initializer, type);
                }
                else if (property.kind === 293 /* SpreadAssignment */) {
                    if (propertyIndex < properties.length - 1) {
                        error(property, ts.Diagnostics.A_rest_element_must_be_last_in_a_destructuring_pattern);
                    }
                    else {
                        if (languageVersion < 99 /* ESNext */) {
                            checkExternalEmitHelpers(property, 4 /* Rest */);
                        }
                        var nonRestNames = [];
                        if (allProperties) {
                            for (var _i = 0, allProperties_1 = allProperties; _i < allProperties_1.length; _i++) {
                                var otherProperty = allProperties_1[_i];
                                if (!ts.isSpreadAssignment(otherProperty)) {
                                    nonRestNames.push(otherProperty.name);
                                }
                            }
                        }
                        var type = getRestType(objectLiteralType, nonRestNames, objectLiteralType.symbol);
                        checkGrammarForDisallowedTrailingComma(allProperties, ts.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma);
                        return checkDestructuringAssignment(property.expression, type);
                    }
                }
                else {
                    error(property, ts.Diagnostics.Property_assignment_expected);
                }
            }
            function checkArrayLiteralAssignment(node, sourceType, checkMode) {
                var elements = node.elements;
                if (languageVersion < 2 /* ES2015 */ && compilerOptions.downlevelIteration) {
                    checkExternalEmitHelpers(node, 512 /* Read */);
                }
                // This elementType will be used if the specific property corresponding to this index is not
                // present (aka the tuple element property). This call also checks that the parentType is in
                // fact an iterable or array (depending on target language).
                var possiblyOutOfBoundsType = checkIteratedTypeOrElementType(65 /* Destructuring */ | 128 /* PossiblyOutOfBounds */, sourceType, undefinedType, node) || errorType;
                var inBoundsType = compilerOptions.noUncheckedIndexedAccess ? undefined : possiblyOutOfBoundsType;
                for (var i = 0; i < elements.length; i++) {
                    var type = possiblyOutOfBoundsType;
                    if (node.elements[i].kind === 223 /* SpreadElement */) {
                        type = inBoundsType = inBoundsType !== null && inBoundsType !== void 0 ? inBoundsType : (checkIteratedTypeOrElementType(65 /* Destructuring */, sourceType, undefinedType, node) || errorType);
                    }
                    checkArrayLiteralDestructuringElementAssignment(node, sourceType, i, type, checkMode);
                }
                return sourceType;
            }
            function checkArrayLiteralDestructuringElementAssignment(node, sourceType, elementIndex, elementType, checkMode) {
                var elements = node.elements;
                var element = elements[elementIndex];
                if (element.kind !== 225 /* OmittedExpression */) {
                    if (element.kind !== 223 /* SpreadElement */) {
                        var indexType = getNumberLiteralType(elementIndex);
                        if (isArrayLikeType(sourceType)) {
                            // We create a synthetic expression so that getIndexedAccessType doesn't get confused
                            // when the element is a SyntaxKind.ElementAccessExpression.
                            var accessFlags = 32 /* ExpressionPosition */ | (hasDefaultValue(element) ? 16 /* NoTupleBoundsCheck */ : 0);
                            var elementType_2 = getIndexedAccessTypeOrUndefined(sourceType, indexType, accessFlags, createSyntheticExpression(element, indexType)) || errorType;
                            var assignedType = hasDefaultValue(element) ? getTypeWithFacts(elementType_2, 524288 /* NEUndefined */) : elementType_2;
                            var type = getFlowTypeOfDestructuring(element, assignedType);
                            return checkDestructuringAssignment(element, type, checkMode);
                        }
                        return checkDestructuringAssignment(element, elementType, checkMode);
                    }
                    if (elementIndex < elements.length - 1) {
                        error(element, ts.Diagnostics.A_rest_element_must_be_last_in_a_destructuring_pattern);
                    }
                    else {
                        var restExpression = element.expression;
                        if (restExpression.kind === 219 /* BinaryExpression */ && restExpression.operatorToken.kind === 63 /* EqualsToken */) {
                            error(restExpression.operatorToken, ts.Diagnostics.A_rest_element_cannot_have_an_initializer);
                        }
                        else {
                            checkGrammarForDisallowedTrailingComma(node.elements, ts.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma);
                            var type = everyType(sourceType, isTupleType) ?
                                mapType(sourceType, function (t) { return sliceTupleType(t, elementIndex); }) :
                                createArrayType(elementType);
                            return checkDestructuringAssignment(restExpression, type, checkMode);
                        }
                    }
                }
                return undefined;
            }
            function checkDestructuringAssignment(exprOrAssignment, sourceType, checkMode, rightIsThis) {
                var target;
                if (exprOrAssignment.kind === 292 /* ShorthandPropertyAssignment */) {
                    var prop = exprOrAssignment;
                    if (prop.objectAssignmentInitializer) {
                        // In strict null checking mode, if a default value of a non-undefined type is specified, remove
                        // undefined from the final type.
                        if (strictNullChecks &&
                            !(getFalsyFlags(checkExpression(prop.objectAssignmentInitializer)) & 32768 /* Undefined */)) {
                            sourceType = getTypeWithFacts(sourceType, 524288 /* NEUndefined */);
                        }
                        checkBinaryLikeExpression(prop.name, prop.equalsToken, prop.objectAssignmentInitializer, checkMode);
                    }
                    target = exprOrAssignment.name;
                }
                else {
                    target = exprOrAssignment;
                }
                if (target.kind === 219 /* BinaryExpression */ && target.operatorToken.kind === 63 /* EqualsToken */) {
                    checkBinaryExpression(target, checkMode);
                    target = target.left;
                }
                if (target.kind === 203 /* ObjectLiteralExpression */) {
                    return checkObjectLiteralAssignment(target, sourceType, rightIsThis);
                }
                if (target.kind === 202 /* ArrayLiteralExpression */) {
                    return checkArrayLiteralAssignment(target, sourceType, checkMode);
                }
                return checkReferenceAssignment(target, sourceType, checkMode);
            }
            function checkReferenceAssignment(target, sourceType, checkMode) {
                var targetType = checkExpression(target, checkMode);
                var error = target.parent.kind === 293 /* SpreadAssignment */ ?
                    ts.Diagnostics.The_target_of_an_object_rest_assignment_must_be_a_variable_or_a_property_access :
                    ts.Diagnostics.The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access;
                var optionalError = target.parent.kind === 293 /* SpreadAssignment */ ?
                    ts.Diagnostics.The_target_of_an_object_rest_assignment_may_not_be_an_optional_property_access :
                    ts.Diagnostics.The_left_hand_side_of_an_assignment_expression_may_not_be_an_optional_property_access;
                if (checkReferenceExpression(target, error, optionalError)) {
                    checkTypeAssignableToAndOptionallyElaborate(sourceType, targetType, target, target);
                }
                if (ts.isPrivateIdentifierPropertyAccessExpression(target)) {
                    checkExternalEmitHelpers(target.parent, 1048576 /* ClassPrivateFieldSet */);
                }
                return sourceType;
            }
            /**
             * This is a *shallow* check: An expression is side-effect-free if the
             * evaluation of the expression *itself* cannot produce side effects.
             * For example, x++ / 3 is side-effect free because the / operator
             * does not have side effects.
             * The intent is to "smell test" an expression for correctness in positions where
             * its value is discarded (e.g. the left side of the comma operator).
             */
            function isSideEffectFree(node) {
                node = ts.skipParentheses(node);
                switch (node.kind) {
                    case 79 /* Identifier */:
                    case 10 /* StringLiteral */:
                    case 13 /* RegularExpressionLiteral */:
                    case 208 /* TaggedTemplateExpression */:
                    case 221 /* TemplateExpression */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                    case 8 /* NumericLiteral */:
                    case 9 /* BigIntLiteral */:
                    case 110 /* TrueKeyword */:
                    case 95 /* FalseKeyword */:
                    case 104 /* NullKeyword */:
                    case 151 /* UndefinedKeyword */:
                    case 211 /* FunctionExpression */:
                    case 224 /* ClassExpression */:
                    case 212 /* ArrowFunction */:
                    case 202 /* ArrayLiteralExpression */:
                    case 203 /* ObjectLiteralExpression */:
                    case 214 /* TypeOfExpression */:
                    case 228 /* NonNullExpression */:
                    case 277 /* JsxSelfClosingElement */:
                    case 276 /* JsxElement */:
                        return true;
                    case 220 /* ConditionalExpression */:
                        return isSideEffectFree(node.whenTrue) &&
                            isSideEffectFree(node.whenFalse);
                    case 219 /* BinaryExpression */:
                        if (ts.isAssignmentOperator(node.operatorToken.kind)) {
                            return false;
                        }
                        return isSideEffectFree(node.left) &&
                            isSideEffectFree(node.right);
                    case 217 /* PrefixUnaryExpression */:
                    case 218 /* PostfixUnaryExpression */:
                        // Unary operators ~, !, +, and - have no side effects.
                        // The rest do.
                        switch (node.operator) {
                            case 53 /* ExclamationToken */:
                            case 39 /* PlusToken */:
                            case 40 /* MinusToken */:
                            case 54 /* TildeToken */:
                                return true;
                        }
                        return false;
                    // Some forms listed here for clarity
                    case 215 /* VoidExpression */: // Explicit opt-out
                    case 209 /* TypeAssertionExpression */: // Not SEF, but can produce useful type warnings
                    case 227 /* AsExpression */: // Not SEF, but can produce useful type warnings
                    default:
                        return false;
                }
            }
            function isTypeEqualityComparableTo(source, target) {
                return (target.flags & 98304 /* Nullable */) !== 0 || isTypeComparableTo(source, target);
            }
            function createCheckBinaryExpression() {
                var trampoline = ts.createBinaryExpressionTrampoline(onEnter, onLeft, onOperator, onRight, onExit, foldState);
                return function (node, checkMode) {
                    var result = trampoline(node, checkMode);
                    ts.Debug.assertIsDefined(result);
                    return result;
                };
                function onEnter(node, state, checkMode) {
                    if (state) {
                        state.stackIndex++;
                        state.skip = false;
                        setLeftType(state, /*type*/ undefined);
                        setLastResult(state, /*type*/ undefined);
                    }
                    else {
                        state = {
                            checkMode: checkMode,
                            skip: false,
                            stackIndex: 0,
                            typeStack: [undefined, undefined],
                        };
                    }
                    if (ts.isInJSFile(node) && ts.getAssignedExpandoInitializer(node)) {
                        state.skip = true;
                        setLastResult(state, checkExpression(node.right, checkMode));
                        return state;
                    }
                    checkGrammarNullishCoalesceWithLogicalExpression(node);
                    var operator = node.operatorToken.kind;
                    if (operator === 63 /* EqualsToken */ && (node.left.kind === 203 /* ObjectLiteralExpression */ || node.left.kind === 202 /* ArrayLiteralExpression */)) {
                        state.skip = true;
                        setLastResult(state, checkDestructuringAssignment(node.left, checkExpression(node.right, checkMode), checkMode, node.right.kind === 108 /* ThisKeyword */));
                        return state;
                    }
                    return state;
                }
                function onLeft(left, state, _node) {
                    if (!state.skip) {
                        return maybeCheckExpression(state, left);
                    }
                }
                function onOperator(operatorToken, state, node) {
                    if (!state.skip) {
                        var leftType = getLastResult(state);
                        ts.Debug.assertIsDefined(leftType);
                        setLeftType(state, leftType);
                        setLastResult(state, /*type*/ undefined);
                        var operator = operatorToken.kind;
                        if (operator === 55 /* AmpersandAmpersandToken */ || operator === 56 /* BarBarToken */ || operator === 60 /* QuestionQuestionToken */) {
                            if (operator === 55 /* AmpersandAmpersandToken */) {
                                var parent = ts.walkUpParenthesizedExpressions(node.parent);
                                checkTestingKnownTruthyCallableOrAwaitableType(node.left, leftType, ts.isIfStatement(parent) ? parent.thenStatement : undefined);
                            }
                            checkTruthinessOfType(leftType, node.left);
                        }
                    }
                }
                function onRight(right, state, _node) {
                    if (!state.skip) {
                        return maybeCheckExpression(state, right);
                    }
                }
                function onExit(node, state) {
                    var result;
                    if (state.skip) {
                        result = getLastResult(state);
                    }
                    else {
                        var leftType = getLeftType(state);
                        ts.Debug.assertIsDefined(leftType);
                        var rightType = getLastResult(state);
                        ts.Debug.assertIsDefined(rightType);
                        result = checkBinaryLikeExpressionWorker(node.left, node.operatorToken, node.right, leftType, rightType, node);
                    }
                    state.skip = false;
                    setLeftType(state, /*type*/ undefined);
                    setLastResult(state, /*type*/ undefined);
                    state.stackIndex--;
                    return result;
                }
                function foldState(state, result, _side) {
                    setLastResult(state, result);
                    return state;
                }
                function maybeCheckExpression(state, node) {
                    if (ts.isBinaryExpression(node)) {
                        return node;
                    }
                    setLastResult(state, checkExpression(node, state.checkMode));
                }
                function getLeftType(state) {
                    return state.typeStack[state.stackIndex];
                }
                function setLeftType(state, type) {
                    state.typeStack[state.stackIndex] = type;
                }
                function getLastResult(state) {
                    return state.typeStack[state.stackIndex + 1];
                }
                function setLastResult(state, type) {
                    // To reduce overhead, reuse the next stack entry to store the
                    // last result. This avoids the overhead of an additional property
                    // on `WorkArea` and reuses empty stack entries as we walk back up
                    // the stack.
                    state.typeStack[state.stackIndex + 1] = type;
                }
            }
            function checkGrammarNullishCoalesceWithLogicalExpression(node) {
                var left = node.left, operatorToken = node.operatorToken, right = node.right;
                if (operatorToken.kind === 60 /* QuestionQuestionToken */) {
                    if (ts.isBinaryExpression(left) && (left.operatorToken.kind === 56 /* BarBarToken */ || left.operatorToken.kind === 55 /* AmpersandAmpersandToken */)) {
                        grammarErrorOnNode(left, ts.Diagnostics._0_and_1_operations_cannot_be_mixed_without_parentheses, ts.tokenToString(left.operatorToken.kind), ts.tokenToString(operatorToken.kind));
                    }
                    if (ts.isBinaryExpression(right) && (right.operatorToken.kind === 56 /* BarBarToken */ || right.operatorToken.kind === 55 /* AmpersandAmpersandToken */)) {
                        grammarErrorOnNode(right, ts.Diagnostics._0_and_1_operations_cannot_be_mixed_without_parentheses, ts.tokenToString(right.operatorToken.kind), ts.tokenToString(operatorToken.kind));
                    }
                }
            }
            // Note that this and `checkBinaryExpression` above should behave mostly the same, except this elides some
            // expression-wide checks and does not use a work stack to fold nested binary expressions into the same callstack frame
            function checkBinaryLikeExpression(left, operatorToken, right, checkMode, errorNode) {
                var operator = operatorToken.kind;
                if (operator === 63 /* EqualsToken */ && (left.kind === 203 /* ObjectLiteralExpression */ || left.kind === 202 /* ArrayLiteralExpression */)) {
                    return checkDestructuringAssignment(left, checkExpression(right, checkMode), checkMode, right.kind === 108 /* ThisKeyword */);
                }
                var leftType;
                if (operator === 55 /* AmpersandAmpersandToken */ || operator === 56 /* BarBarToken */ || operator === 60 /* QuestionQuestionToken */) {
                    leftType = checkTruthinessExpression(left, checkMode);
                }
                else {
                    leftType = checkExpression(left, checkMode);
                }
                var rightType = checkExpression(right, checkMode);
                return checkBinaryLikeExpressionWorker(left, operatorToken, right, leftType, rightType, errorNode);
            }
            function checkBinaryLikeExpressionWorker(left, operatorToken, right, leftType, rightType, errorNode) {
                var operator = operatorToken.kind;
                switch (operator) {
                    case 41 /* AsteriskToken */:
                    case 42 /* AsteriskAsteriskToken */:
                    case 66 /* AsteriskEqualsToken */:
                    case 67 /* AsteriskAsteriskEqualsToken */:
                    case 43 /* SlashToken */:
                    case 68 /* SlashEqualsToken */:
                    case 44 /* PercentToken */:
                    case 69 /* PercentEqualsToken */:
                    case 40 /* MinusToken */:
                    case 65 /* MinusEqualsToken */:
                    case 47 /* LessThanLessThanToken */:
                    case 70 /* LessThanLessThanEqualsToken */:
                    case 48 /* GreaterThanGreaterThanToken */:
                    case 71 /* GreaterThanGreaterThanEqualsToken */:
                    case 49 /* GreaterThanGreaterThanGreaterThanToken */:
                    case 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */:
                    case 51 /* BarToken */:
                    case 74 /* BarEqualsToken */:
                    case 52 /* CaretToken */:
                    case 78 /* CaretEqualsToken */:
                    case 50 /* AmpersandToken */:
                    case 73 /* AmpersandEqualsToken */:
                        if (leftType === silentNeverType || rightType === silentNeverType) {
                            return silentNeverType;
                        }
                        leftType = checkNonNullType(leftType, left);
                        rightType = checkNonNullType(rightType, right);
                        var suggestedOperator = void 0;
                        // if a user tries to apply a bitwise operator to 2 boolean operands
                        // try and return them a helpful suggestion
                        if ((leftType.flags & 528 /* BooleanLike */) &&
                            (rightType.flags & 528 /* BooleanLike */) &&
                            (suggestedOperator = getSuggestedBooleanOperator(operatorToken.kind)) !== undefined) {
                            error(errorNode || operatorToken, ts.Diagnostics.The_0_operator_is_not_allowed_for_boolean_types_Consider_using_1_instead, ts.tokenToString(operatorToken.kind), ts.tokenToString(suggestedOperator));
                            return numberType;
                        }
                        else {
                            // otherwise just check each operand separately and report errors as normal
                            var leftOk = checkArithmeticOperandType(left, leftType, ts.Diagnostics.The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type, /*isAwaitValid*/ true);
                            var rightOk = checkArithmeticOperandType(right, rightType, ts.Diagnostics.The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type, /*isAwaitValid*/ true);
                            var resultType_1;
                            // If both are any or unknown, allow operation; assume it will resolve to number
                            if ((isTypeAssignableToKind(leftType, 3 /* AnyOrUnknown */) && isTypeAssignableToKind(rightType, 3 /* AnyOrUnknown */)) ||
                                // Or, if neither could be bigint, implicit coercion results in a number result
                                !(maybeTypeOfKind(leftType, 2112 /* BigIntLike */) || maybeTypeOfKind(rightType, 2112 /* BigIntLike */))) {
                                resultType_1 = numberType;
                            }
                            // At least one is assignable to bigint, so check that both are
                            else if (bothAreBigIntLike(leftType, rightType)) {
                                switch (operator) {
                                    case 49 /* GreaterThanGreaterThanGreaterThanToken */:
                                    case 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */:
                                        reportOperatorError();
                                        break;
                                    case 42 /* AsteriskAsteriskToken */:
                                    case 67 /* AsteriskAsteriskEqualsToken */:
                                        if (languageVersion < 3 /* ES2016 */) {
                                            error(errorNode, ts.Diagnostics.Exponentiation_cannot_be_performed_on_bigint_values_unless_the_target_option_is_set_to_es2016_or_later);
                                        }
                                }
                                resultType_1 = bigintType;
                            }
                            // Exactly one of leftType/rightType is assignable to bigint
                            else {
                                reportOperatorError(bothAreBigIntLike);
                                resultType_1 = errorType;
                            }
                            if (leftOk && rightOk) {
                                checkAssignmentOperator(resultType_1);
                            }
                            return resultType_1;
                        }
                    case 39 /* PlusToken */:
                    case 64 /* PlusEqualsToken */:
                        if (leftType === silentNeverType || rightType === silentNeverType) {
                            return silentNeverType;
                        }
                        if (!isTypeAssignableToKind(leftType, 402653316 /* StringLike */) && !isTypeAssignableToKind(rightType, 402653316 /* StringLike */)) {
                            leftType = checkNonNullType(leftType, left);
                            rightType = checkNonNullType(rightType, right);
                        }
                        var resultType = void 0;
                        if (isTypeAssignableToKind(leftType, 296 /* NumberLike */, /*strict*/ true) && isTypeAssignableToKind(rightType, 296 /* NumberLike */, /*strict*/ true)) {
                            // Operands of an enum type are treated as having the primitive type Number.
                            // If both operands are of the Number primitive type, the result is of the Number primitive type.
                            resultType = numberType;
                        }
                        else if (isTypeAssignableToKind(leftType, 2112 /* BigIntLike */, /*strict*/ true) && isTypeAssignableToKind(rightType, 2112 /* BigIntLike */, /*strict*/ true)) {
                            // If both operands are of the BigInt primitive type, the result is of the BigInt primitive type.
                            resultType = bigintType;
                        }
                        else if (isTypeAssignableToKind(leftType, 402653316 /* StringLike */, /*strict*/ true) || isTypeAssignableToKind(rightType, 402653316 /* StringLike */, /*strict*/ true)) {
                            // If one or both operands are of the String primitive type, the result is of the String primitive type.
                            resultType = stringType;
                        }
                        else if (isTypeAny(leftType) || isTypeAny(rightType)) {
                            // Otherwise, the result is of type Any.
                            // NOTE: unknown type here denotes error type. Old compiler treated this case as any type so do we.
                            resultType = leftType === errorType || rightType === errorType ? errorType : anyType;
                        }
                        // Symbols are not allowed at all in arithmetic expressions
                        if (resultType && !checkForDisallowedESSymbolOperand(operator)) {
                            return resultType;
                        }
                        if (!resultType) {
                            // Types that have a reasonably good chance of being a valid operand type.
                            // If both types have an awaited type of one of these, we'll assume the user
                            // might be missing an await without doing an exhaustive check that inserting
                            // await(s) will actually be a completely valid binary expression.
                            var closeEnoughKind_1 = 296 /* NumberLike */ | 2112 /* BigIntLike */ | 402653316 /* StringLike */ | 3 /* AnyOrUnknown */;
                            reportOperatorError(function (left, right) {
                                return isTypeAssignableToKind(left, closeEnoughKind_1) &&
                                    isTypeAssignableToKind(right, closeEnoughKind_1);
                            });
                            return anyType;
                        }
                        if (operator === 64 /* PlusEqualsToken */) {
                            checkAssignmentOperator(resultType);
                        }
                        return resultType;
                    case 29 /* LessThanToken */:
                    case 31 /* GreaterThanToken */:
                    case 32 /* LessThanEqualsToken */:
                    case 33 /* GreaterThanEqualsToken */:
                        if (checkForDisallowedESSymbolOperand(operator)) {
                            leftType = getBaseTypeOfLiteralType(checkNonNullType(leftType, left));
                            rightType = getBaseTypeOfLiteralType(checkNonNullType(rightType, right));
                            reportOperatorErrorUnless(function (left, right) {
                                return isTypeComparableTo(left, right) || isTypeComparableTo(right, left) || (isTypeAssignableTo(left, numberOrBigIntType) && isTypeAssignableTo(right, numberOrBigIntType));
                            });
                        }
                        return booleanType;
                    case 34 /* EqualsEqualsToken */:
                    case 35 /* ExclamationEqualsToken */:
                    case 36 /* EqualsEqualsEqualsToken */:
                    case 37 /* ExclamationEqualsEqualsToken */:
                        reportOperatorErrorUnless(function (left, right) { return isTypeEqualityComparableTo(left, right) || isTypeEqualityComparableTo(right, left); });
                        return booleanType;
                    case 102 /* InstanceOfKeyword */:
                        return checkInstanceOfExpression(left, right, leftType, rightType);
                    case 101 /* InKeyword */:
                        return checkInExpression(left, right, leftType, rightType);
                    case 55 /* AmpersandAmpersandToken */:
                    case 76 /* AmpersandAmpersandEqualsToken */: {
                        var resultType_2 = getTypeFacts(leftType) & 4194304 /* Truthy */ ?
                            getUnionType([extractDefinitelyFalsyTypes(strictNullChecks ? leftType : getBaseTypeOfLiteralType(rightType)), rightType]) :
                            leftType;
                        if (operator === 76 /* AmpersandAmpersandEqualsToken */) {
                            checkAssignmentOperator(rightType);
                        }
                        return resultType_2;
                    }
                    case 56 /* BarBarToken */:
                    case 75 /* BarBarEqualsToken */: {
                        var resultType_3 = getTypeFacts(leftType) & 8388608 /* Falsy */ ?
                            getUnionType([removeDefinitelyFalsyTypes(leftType), rightType], 2 /* Subtype */) :
                            leftType;
                        if (operator === 75 /* BarBarEqualsToken */) {
                            checkAssignmentOperator(rightType);
                        }
                        return resultType_3;
                    }
                    case 60 /* QuestionQuestionToken */:
                    case 77 /* QuestionQuestionEqualsToken */: {
                        var resultType_4 = getTypeFacts(leftType) & 262144 /* EQUndefinedOrNull */ ?
                            getUnionType([getNonNullableType(leftType), rightType], 2 /* Subtype */) :
                            leftType;
                        if (operator === 77 /* QuestionQuestionEqualsToken */) {
                            checkAssignmentOperator(rightType);
                        }
                        return resultType_4;
                    }
                    case 63 /* EqualsToken */:
                        var declKind = ts.isBinaryExpression(left.parent) ? ts.getAssignmentDeclarationKind(left.parent) : 0 /* None */;
                        checkAssignmentDeclaration(declKind, rightType);
                        if (isAssignmentDeclaration(declKind)) {
                            if (!(rightType.flags & 524288 /* Object */) ||
                                declKind !== 2 /* ModuleExports */ &&
                                    declKind !== 6 /* Prototype */ &&
                                    !isEmptyObjectType(rightType) &&
                                    !isFunctionObjectType(rightType) &&
                                    !(ts.getObjectFlags(rightType) & 1 /* Class */)) {
                                // don't check assignability of module.exports=, C.prototype=, or expando types because they will necessarily be incomplete
                                checkAssignmentOperator(rightType);
                            }
                            return leftType;
                        }
                        else {
                            checkAssignmentOperator(rightType);
                            return getRegularTypeOfObjectLiteral(rightType);
                        }
                    case 27 /* CommaToken */:
                        if (!compilerOptions.allowUnreachableCode && isSideEffectFree(left) && !isEvalNode(right)) {
                            var sf = ts.getSourceFileOfNode(left);
                            var sourceText = sf.text;
                            var start_3 = ts.skipTrivia(sourceText, left.pos);
                            var isInDiag2657 = sf.parseDiagnostics.some(function (diag) {
                                if (diag.code !== ts.Diagnostics.JSX_expressions_must_have_one_parent_element.code)
                                    return false;
                                return ts.textSpanContainsPosition(diag, start_3);
                            });
                            if (!isInDiag2657)
                                error(left, ts.Diagnostics.Left_side_of_comma_operator_is_unused_and_has_no_side_effects);
                        }
                        return rightType;
                    default:
                        return ts.Debug.fail();
                }
                function bothAreBigIntLike(left, right) {
                    return isTypeAssignableToKind(left, 2112 /* BigIntLike */) && isTypeAssignableToKind(right, 2112 /* BigIntLike */);
                }
                function checkAssignmentDeclaration(kind, rightType) {
                    if (kind === 2 /* ModuleExports */) {
                        for (var _i = 0, _a = getPropertiesOfObjectType(rightType); _i < _a.length; _i++) {
                            var prop = _a[_i];
                            var propType = getTypeOfSymbol(prop);
                            if (propType.symbol && propType.symbol.flags & 32 /* Class */) {
                                var name = prop.escapedName;
                                var symbol = resolveName(prop.valueDeclaration, name, 788968 /* Type */, undefined, name, /*isUse*/ false);
                                if ((symbol === null || symbol === void 0 ? void 0 : symbol.declarations) && symbol.declarations.some(ts.isJSDocTypedefTag)) {
                                    addDuplicateDeclarationErrorsForSymbols(symbol, ts.Diagnostics.Duplicate_identifier_0, ts.unescapeLeadingUnderscores(name), prop);
                                    addDuplicateDeclarationErrorsForSymbols(prop, ts.Diagnostics.Duplicate_identifier_0, ts.unescapeLeadingUnderscores(name), symbol);
                                }
                            }
                        }
                    }
                }
                function isEvalNode(node) {
                    return node.kind === 79 /* Identifier */ && node.escapedText === "eval";
                }
                // Return true if there was no error, false if there was an error.
                function checkForDisallowedESSymbolOperand(operator) {
                    var offendingSymbolOperand = maybeTypeOfKind(leftType, 12288 /* ESSymbolLike */) ? left :
                        maybeTypeOfKind(rightType, 12288 /* ESSymbolLike */) ? right :
                            undefined;
                    if (offendingSymbolOperand) {
                        error(offendingSymbolOperand, ts.Diagnostics.The_0_operator_cannot_be_applied_to_type_symbol, ts.tokenToString(operator));
                        return false;
                    }
                    return true;
                }
                function getSuggestedBooleanOperator(operator) {
                    switch (operator) {
                        case 51 /* BarToken */:
                        case 74 /* BarEqualsToken */:
                            return 56 /* BarBarToken */;
                        case 52 /* CaretToken */:
                        case 78 /* CaretEqualsToken */:
                            return 37 /* ExclamationEqualsEqualsToken */;
                        case 50 /* AmpersandToken */:
                        case 73 /* AmpersandEqualsToken */:
                            return 55 /* AmpersandAmpersandToken */;
                        default:
                            return undefined;
                    }
                }
                function checkAssignmentOperator(valueType) {
                    if (produceDiagnostics && ts.isAssignmentOperator(operator)) {
                        // TypeScript 1.0 spec (April 2014): 4.17
                        // An assignment of the form
                        //    VarExpr = ValueExpr
                        // requires VarExpr to be classified as a reference
                        // A compound assignment furthermore requires VarExpr to be classified as a reference (section 4.1)
                        // and the type of the non-compound operation to be assignable to the type of VarExpr.
                        if (checkReferenceExpression(left, ts.Diagnostics.The_left_hand_side_of_an_assignment_expression_must_be_a_variable_or_a_property_access, ts.Diagnostics.The_left_hand_side_of_an_assignment_expression_may_not_be_an_optional_property_access)
                            && (!ts.isIdentifier(left) || ts.unescapeLeadingUnderscores(left.escapedText) !== "exports")) {
                            // to avoid cascading errors check assignability only if 'isReference' check succeeded and no errors were reported
                            checkTypeAssignableToAndOptionallyElaborate(valueType, leftType, left, right);
                        }
                    }
                }
                function isAssignmentDeclaration(kind) {
                    var _a;
                    switch (kind) {
                        case 2 /* ModuleExports */:
                            return true;
                        case 1 /* ExportsProperty */:
                        case 5 /* Property */:
                        case 6 /* Prototype */:
                        case 3 /* PrototypeProperty */:
                        case 4 /* ThisProperty */:
                            var symbol = getSymbolOfNode(left);
                            var init = ts.getAssignedExpandoInitializer(right);
                            return !!init && ts.isObjectLiteralExpression(init) &&
                                !!((_a = symbol === null || symbol === void 0 ? void 0 : symbol.exports) === null || _a === void 0 ? void 0 : _a.size);
                        default:
                            return false;
                    }
                }
                /**
                 * Returns true if an error is reported
                 */
                function reportOperatorErrorUnless(typesAreCompatible) {
                    if (!typesAreCompatible(leftType, rightType)) {
                        reportOperatorError(typesAreCompatible);
                        return true;
                    }
                    return false;
                }
                function reportOperatorError(isRelated) {
                    var _a;
                    var wouldWorkWithAwait = false;
                    var errNode = errorNode || operatorToken;
                    if (isRelated) {
                        var awaitedLeftType = getAwaitedType(leftType);
                        var awaitedRightType = getAwaitedType(rightType);
                        wouldWorkWithAwait = !(awaitedLeftType === leftType && awaitedRightType === rightType)
                            && !!(awaitedLeftType && awaitedRightType)
                            && isRelated(awaitedLeftType, awaitedRightType);
                    }
                    var effectiveLeft = leftType;
                    var effectiveRight = rightType;
                    if (!wouldWorkWithAwait && isRelated) {
                        _a = getBaseTypesIfUnrelated(leftType, rightType, isRelated), effectiveLeft = _a[0], effectiveRight = _a[1];
                    }
                    var _b = getTypeNamesForErrorDisplay(effectiveLeft, effectiveRight), leftStr = _b[0], rightStr = _b[1];
                    if (!tryGiveBetterPrimaryError(errNode, wouldWorkWithAwait, leftStr, rightStr)) {
                        errorAndMaybeSuggestAwait(errNode, wouldWorkWithAwait, ts.Diagnostics.Operator_0_cannot_be_applied_to_types_1_and_2, ts.tokenToString(operatorToken.kind), leftStr, rightStr);
                    }
                }
                function tryGiveBetterPrimaryError(errNode, maybeMissingAwait, leftStr, rightStr) {
                    var typeName;
                    switch (operatorToken.kind) {
                        case 36 /* EqualsEqualsEqualsToken */:
                        case 34 /* EqualsEqualsToken */:
                            typeName = "false";
                            break;
                        case 37 /* ExclamationEqualsEqualsToken */:
                        case 35 /* ExclamationEqualsToken */:
                            typeName = "true";
                    }
                    if (typeName) {
                        return errorAndMaybeSuggestAwait(errNode, maybeMissingAwait, ts.Diagnostics.This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap, typeName, leftStr, rightStr);
                    }
                    return undefined;
                }
            }
            function getBaseTypesIfUnrelated(leftType, rightType, isRelated) {
                var effectiveLeft = leftType;
                var effectiveRight = rightType;
                var leftBase = getBaseTypeOfLiteralType(leftType);
                var rightBase = getBaseTypeOfLiteralType(rightType);
                if (!isRelated(leftBase, rightBase)) {
                    effectiveLeft = leftBase;
                    effectiveRight = rightBase;
                }
                return [effectiveLeft, effectiveRight];
            }
            function checkYieldExpression(node) {
                // Grammar checking
                if (produceDiagnostics) {
                    if (!(node.flags & 8192 /* YieldContext */)) {
                        grammarErrorOnFirstToken(node, ts.Diagnostics.A_yield_expression_is_only_allowed_in_a_generator_body);
                    }
                    if (isInParameterInitializerBeforeContainingFunction(node)) {
                        error(node, ts.Diagnostics.yield_expressions_cannot_be_used_in_a_parameter_initializer);
                    }
                }
                var func = ts.getContainingFunction(node);
                if (!func)
                    return anyType;
                var functionFlags = ts.getFunctionFlags(func);
                if (!(functionFlags & 1 /* Generator */)) {
                    // If the user's code is syntactically correct, the func should always have a star. After all, we are in a yield context.
                    return anyType;
                }
                var isAsync = (functionFlags & 2 /* Async */) !== 0;
                if (node.asteriskToken) {
                    // Async generator functions prior to ESNext require the __await, __asyncDelegator,
                    // and __asyncValues helpers
                    if (isAsync && languageVersion < 99 /* ESNext */) {
                        checkExternalEmitHelpers(node, 26624 /* AsyncDelegatorIncludes */);
                    }
                    // Generator functions prior to ES2015 require the __values helper
                    if (!isAsync && languageVersion < 2 /* ES2015 */ && compilerOptions.downlevelIteration) {
                        checkExternalEmitHelpers(node, 256 /* Values */);
                    }
                }
                // There is no point in doing an assignability check if the function
                // has no explicit return type because the return type is directly computed
                // from the yield expressions.
                var returnType = getReturnTypeFromAnnotation(func);
                var iterationTypes = returnType && getIterationTypesOfGeneratorFunctionReturnType(returnType, isAsync);
                var signatureYieldType = iterationTypes && iterationTypes.yieldType || anyType;
                var signatureNextType = iterationTypes && iterationTypes.nextType || anyType;
                var resolvedSignatureNextType = isAsync ? getAwaitedType(signatureNextType) || anyType : signatureNextType;
                var yieldExpressionType = node.expression ? checkExpression(node.expression) : undefinedWideningType;
                var yieldedType = getYieldedTypeOfYieldExpression(node, yieldExpressionType, resolvedSignatureNextType, isAsync);
                if (returnType && yieldedType) {
                    checkTypeAssignableToAndOptionallyElaborate(yieldedType, signatureYieldType, node.expression || node, node.expression);
                }
                if (node.asteriskToken) {
                    var use = isAsync ? 19 /* AsyncYieldStar */ : 17 /* YieldStar */;
                    return getIterationTypeOfIterable(use, 1 /* Return */, yieldExpressionType, node.expression)
                        || anyType;
                }
                else if (returnType) {
                    return getIterationTypeOfGeneratorFunctionReturnType(2 /* Next */, returnType, isAsync)
                        || anyType;
                }
                var type = getContextualIterationType(2 /* Next */, func);
                if (!type) {
                    type = anyType;
                    if (produceDiagnostics && noImplicitAny && !ts.expressionResultIsUnused(node)) {
                        var contextualType = getContextualType(node);
                        if (!contextualType || isTypeAny(contextualType)) {
                            error(node, ts.Diagnostics.yield_expression_implicitly_results_in_an_any_type_because_its_containing_generator_lacks_a_return_type_annotation);
                        }
                    }
                }
                return type;
            }
            function checkConditionalExpression(node, checkMode) {
                var type = checkTruthinessExpression(node.condition);
                checkTestingKnownTruthyCallableOrAwaitableType(node.condition, type, node.whenTrue);
                var type1 = checkExpression(node.whenTrue, checkMode);
                var type2 = checkExpression(node.whenFalse, checkMode);
                return getUnionType([type1, type2], 2 /* Subtype */);
            }
            function isTemplateLiteralContext(node) {
                var parent = node.parent;
                return ts.isParenthesizedExpression(parent) && isTemplateLiteralContext(parent) ||
                    ts.isElementAccessExpression(parent) && parent.argumentExpression === node;
            }
            function checkTemplateExpression(node) {
                var texts = [node.head.text];
                var types = [];
                for (var _i = 0, _a = node.templateSpans; _i < _a.length; _i++) {
                    var span = _a[_i];
                    var type = checkExpression(span.expression);
                    if (maybeTypeOfKind(type, 12288 /* ESSymbolLike */)) {
                        error(span.expression, ts.Diagnostics.Implicit_conversion_of_a_symbol_to_a_string_will_fail_at_runtime_Consider_wrapping_this_expression_in_String);
                    }
                    texts.push(span.literal.text);
                    types.push(isTypeAssignableTo(type, templateConstraintType) ? type : stringType);
                }
                return isConstContext(node) || isTemplateLiteralContext(node) || someType(getContextualType(node) || unknownType, isTemplateLiteralContextualType) ? getTemplateLiteralType(texts, types) : stringType;
            }
            function isTemplateLiteralContextualType(type) {
                return !!(type.flags & (128 /* StringLiteral */ | 134217728 /* TemplateLiteral */) ||
                    type.flags & 58982400 /* InstantiableNonPrimitive */ && maybeTypeOfKind(getBaseConstraintOfType(type) || unknownType, 402653316 /* StringLike */));
            }
            function getContextNode(node) {
                if (node.kind === 284 /* JsxAttributes */ && !ts.isJsxSelfClosingElement(node.parent)) {
                    return node.parent.parent; // Needs to be the root JsxElement, so it encompasses the attributes _and_ the children (which are essentially part of the attributes)
                }
                return node;
            }
            function checkExpressionWithContextualType(node, contextualType, inferenceContext, checkMode) {
                var context = getContextNode(node);
                var saveContextualType = context.contextualType;
                var saveInferenceContext = context.inferenceContext;
                try {
                    context.contextualType = contextualType;
                    context.inferenceContext = inferenceContext;
                    var type = checkExpression(node, checkMode | 1 /* Contextual */ | (inferenceContext ? 2 /* Inferential */ : 0));
                    // We strip literal freshness when an appropriate contextual type is present such that contextually typed
                    // literals always preserve their literal types (otherwise they might widen during type inference). An alternative
                    // here would be to not mark contextually typed literals as fresh in the first place.
                    var result = maybeTypeOfKind(type, 2944 /* Literal */) && isLiteralOfContextualType(type, instantiateContextualType(contextualType, node)) ?
                        getRegularTypeOfLiteralType(type) : type;
                    return result;
                }
                finally {
                    // In the event our operation is canceled or some other exception occurs, reset the contextual type
                    // so that we do not accidentally hold onto an instance of the checker, as a Type created in the services layer
                    // may hold onto the checker that created it.
                    context.contextualType = saveContextualType;
                    context.inferenceContext = saveInferenceContext;
                }
            }
            function checkExpressionCached(node, checkMode) {
                var links = getNodeLinks(node);
                if (!links.resolvedType) {
                    if (checkMode && checkMode !== 0 /* Normal */) {
                        return checkExpression(node, checkMode);
                    }
                    // When computing a type that we're going to cache, we need to ignore any ongoing control flow
                    // analysis because variables may have transient types in indeterminable states. Moving flowLoopStart
                    // to the top of the stack ensures all transient types are computed from a known point.
                    var saveFlowLoopStart = flowLoopStart;
                    var saveFlowTypeCache = flowTypeCache;
                    flowLoopStart = flowLoopCount;
                    flowTypeCache = undefined;
                    links.resolvedType = checkExpression(node, checkMode);
                    flowTypeCache = saveFlowTypeCache;
                    flowLoopStart = saveFlowLoopStart;
                }
                return links.resolvedType;
            }
            function isTypeAssertion(node) {
                node = ts.skipParentheses(node);
                return node.kind === 209 /* TypeAssertionExpression */ || node.kind === 227 /* AsExpression */;
            }
            function checkDeclarationInitializer(declaration, contextualType) {
                var initializer = ts.getEffectiveInitializer(declaration);
                var type = getQuickTypeOfExpression(initializer) ||
                    (contextualType ? checkExpressionWithContextualType(initializer, contextualType, /*inferenceContext*/ undefined, 0 /* Normal */) : checkExpressionCached(initializer));
                return ts.isParameter(declaration) && declaration.name.kind === 200 /* ArrayBindingPattern */ &&
                    isTupleType(type) && !type.target.hasRestElement && getTypeReferenceArity(type) < declaration.name.elements.length ?
                    padTupleType(type, declaration.name) : type;
            }
            function padTupleType(type, pattern) {
                var patternElements = pattern.elements;
                var elementTypes = getTypeArguments(type).slice();
                var elementFlags = type.target.elementFlags.slice();
                for (var i = getTypeReferenceArity(type); i < patternElements.length; i++) {
                    var e = patternElements[i];
                    if (i < patternElements.length - 1 || !(e.kind === 201 /* BindingElement */ && e.dotDotDotToken)) {
                        elementTypes.push(!ts.isOmittedExpression(e) && hasDefaultValue(e) ? getTypeFromBindingElement(e, /*includePatternInType*/ false, /*reportErrors*/ false) : anyType);
                        elementFlags.push(2 /* Optional */);
                        if (!ts.isOmittedExpression(e) && !hasDefaultValue(e)) {
                            reportImplicitAny(e, anyType);
                        }
                    }
                }
                return createTupleType(elementTypes, elementFlags, type.target.readonly);
            }
            function widenTypeInferredFromInitializer(declaration, type) {
                var widened = ts.getCombinedNodeFlags(declaration) & 2 /* Const */ || ts.isDeclarationReadonly(declaration) ? type : getWidenedLiteralType(type);
                if (ts.isInJSFile(declaration)) {
                    if (isEmptyLiteralType(widened)) {
                        reportImplicitAny(declaration, anyType);
                        return anyType;
                    }
                    else if (isEmptyArrayLiteralType(widened)) {
                        reportImplicitAny(declaration, anyArrayType);
                        return anyArrayType;
                    }
                }
                return widened;
            }
            function isLiteralOfContextualType(candidateType, contextualType) {
                if (contextualType) {
                    if (contextualType.flags & 3145728 /* UnionOrIntersection */) {
                        var types = contextualType.types;
                        return ts.some(types, function (t) { return isLiteralOfContextualType(candidateType, t); });
                    }
                    if (contextualType.flags & 58982400 /* InstantiableNonPrimitive */) {
                        // If the contextual type is a type variable constrained to a primitive type, consider
                        // this a literal context for literals of that primitive type. For example, given a
                        // type parameter 'T extends string', infer string literal types for T.
                        var constraint = getBaseConstraintOfType(contextualType) || unknownType;
                        return maybeTypeOfKind(constraint, 4 /* String */) && maybeTypeOfKind(candidateType, 128 /* StringLiteral */) ||
                            maybeTypeOfKind(constraint, 8 /* Number */) && maybeTypeOfKind(candidateType, 256 /* NumberLiteral */) ||
                            maybeTypeOfKind(constraint, 64 /* BigInt */) && maybeTypeOfKind(candidateType, 2048 /* BigIntLiteral */) ||
                            maybeTypeOfKind(constraint, 4096 /* ESSymbol */) && maybeTypeOfKind(candidateType, 8192 /* UniqueESSymbol */) ||
                            isLiteralOfContextualType(candidateType, constraint);
                    }
                    // If the contextual type is a literal of a particular primitive type, we consider this a
                    // literal context for all literals of that primitive type.
                    return !!(contextualType.flags & (128 /* StringLiteral */ | 4194304 /* Index */ | 134217728 /* TemplateLiteral */ | 268435456 /* StringMapping */) && maybeTypeOfKind(candidateType, 128 /* StringLiteral */) ||
                        contextualType.flags & 256 /* NumberLiteral */ && maybeTypeOfKind(candidateType, 256 /* NumberLiteral */) ||
                        contextualType.flags & 2048 /* BigIntLiteral */ && maybeTypeOfKind(candidateType, 2048 /* BigIntLiteral */) ||
                        contextualType.flags & 512 /* BooleanLiteral */ && maybeTypeOfKind(candidateType, 512 /* BooleanLiteral */) ||
                        contextualType.flags & 8192 /* UniqueESSymbol */ && maybeTypeOfKind(candidateType, 8192 /* UniqueESSymbol */));
                }
                return false;
            }
            function isConstContext(node) {
                var parent = node.parent;
                return ts.isAssertionExpression(parent) && ts.isConstTypeReference(parent.type) ||
                    (ts.isParenthesizedExpression(parent) || ts.isArrayLiteralExpression(parent) || ts.isSpreadElement(parent)) && isConstContext(parent) ||
                    (ts.isPropertyAssignment(parent) || ts.isShorthandPropertyAssignment(parent) || ts.isTemplateSpan(parent)) && isConstContext(parent.parent);
            }
            function checkExpressionForMutableLocation(node, checkMode, contextualType, forceTuple) {
                var type = checkExpression(node, checkMode, forceTuple);
                return isConstContext(node) ? getRegularTypeOfLiteralType(type) :
                    isTypeAssertion(node) ? type :
                        getWidenedLiteralLikeTypeForContextualType(type, instantiateContextualType(arguments.length === 2 ? getContextualType(node) : contextualType, node));
            }
            function checkPropertyAssignment(node, checkMode) {
                // Do not use hasDynamicName here, because that returns false for well known symbols.
                // We want to perform checkComputedPropertyName for all computed properties, including
                // well known symbols.
                if (node.name.kind === 160 /* ComputedPropertyName */) {
                    checkComputedPropertyName(node.name);
                }
                return checkExpressionForMutableLocation(node.initializer, checkMode);
            }
            function checkObjectLiteralMethod(node, checkMode) {
                // Grammar checking
                checkGrammarMethod(node);
                // Do not use hasDynamicName here, because that returns false for well known symbols.
                // We want to perform checkComputedPropertyName for all computed properties, including
                // well known symbols.
                if (node.name.kind === 160 /* ComputedPropertyName */) {
                    checkComputedPropertyName(node.name);
                }
                var uninstantiatedType = checkFunctionExpressionOrObjectLiteralMethod(node, checkMode);
                return instantiateTypeWithSingleGenericCallSignature(node, uninstantiatedType, checkMode);
            }
            function instantiateTypeWithSingleGenericCallSignature(node, type, checkMode) {
                if (checkMode && checkMode & (2 /* Inferential */ | 8 /* SkipGenericFunctions */)) {
                    var callSignature = getSingleSignature(type, 0 /* Call */, /*allowMembers*/ true);
                    var constructSignature = getSingleSignature(type, 1 /* Construct */, /*allowMembers*/ true);
                    var signature = callSignature || constructSignature;
                    if (signature && signature.typeParameters) {
                        var contextualType = getApparentTypeOfContextualType(node, 2 /* NoConstraints */);
                        if (contextualType) {
                            var contextualSignature = getSingleSignature(getNonNullableType(contextualType), callSignature ? 0 /* Call */ : 1 /* Construct */, /*allowMembers*/ false);
                            if (contextualSignature && !contextualSignature.typeParameters) {
                                if (checkMode & 8 /* SkipGenericFunctions */) {
                                    skippedGenericFunction(node, checkMode);
                                    return anyFunctionType;
                                }
                                var context = getInferenceContext(node);
                                // We have an expression that is an argument of a generic function for which we are performing
                                // type argument inference. The expression is of a function type with a single generic call
                                // signature and a contextual function type with a single non-generic call signature. Now check
                                // if the outer function returns a function type with a single non-generic call signature and
                                // if some of the outer function type parameters have no inferences so far. If so, we can
                                // potentially add inferred type parameters to the outer function return type.
                                var returnType = context.signature && getReturnTypeOfSignature(context.signature);
                                var returnSignature = returnType && getSingleCallOrConstructSignature(returnType);
                                if (returnSignature && !returnSignature.typeParameters && !ts.every(context.inferences, hasInferenceCandidates)) {
                                    // Instantiate the signature with its own type parameters as type arguments, possibly
                                    // renaming the type parameters to ensure they have unique names.
                                    var uniqueTypeParameters = getUniqueTypeParameters(context, signature.typeParameters);
                                    var instantiatedSignature = getSignatureInstantiationWithoutFillingInTypeArguments(signature, uniqueTypeParameters);
                                    // Infer from the parameters of the instantiated signature to the parameters of the
                                    // contextual signature starting with an empty set of inference candidates.
                                    var inferences_3 = ts.map(context.inferences, function (info) { return createInferenceInfo(info.typeParameter); });
                                    applyToParameterTypes(instantiatedSignature, contextualSignature, function (source, target) {
                                        inferTypes(inferences_3, source, target, /*priority*/ 0, /*contravariant*/ true);
                                    });
                                    if (ts.some(inferences_3, hasInferenceCandidates)) {
                                        // We have inference candidates, indicating that one or more type parameters are referenced
                                        // in the parameter types of the contextual signature. Now also infer from the return type.
                                        applyToReturnTypes(instantiatedSignature, contextualSignature, function (source, target) {
                                            inferTypes(inferences_3, source, target);
                                        });
                                        // If the type parameters for which we produced candidates do not have any inferences yet,
                                        // we adopt the new inference candidates and add the type parameters of the expression type
                                        // to the set of inferred type parameters for the outer function return type.
                                        if (!hasOverlappingInferences(context.inferences, inferences_3)) {
                                            mergeInferences(context.inferences, inferences_3);
                                            context.inferredTypeParameters = ts.concatenate(context.inferredTypeParameters, uniqueTypeParameters);
                                            return getOrCreateTypeFromSignature(instantiatedSignature);
                                        }
                                    }
                                }
                                return getOrCreateTypeFromSignature(instantiateSignatureInContextOf(signature, contextualSignature, context));
                            }
                        }
                    }
                }
                return type;
            }
            function skippedGenericFunction(node, checkMode) {
                if (checkMode & 2 /* Inferential */) {
                    // We have skipped a generic function during inferential typing. Obtain the inference context and
                    // indicate this has occurred such that we know a second pass of inference is be needed.
                    var context = getInferenceContext(node);
                    context.flags |= 4 /* SkippedGenericFunction */;
                }
            }
            function hasInferenceCandidates(info) {
                return !!(info.candidates || info.contraCandidates);
            }
            function hasOverlappingInferences(a, b) {
                for (var i = 0; i < a.length; i++) {
                    if (hasInferenceCandidates(a[i]) && hasInferenceCandidates(b[i])) {
                        return true;
                    }
                }
                return false;
            }
            function mergeInferences(target, source) {
                for (var i = 0; i < target.length; i++) {
                    if (!hasInferenceCandidates(target[i]) && hasInferenceCandidates(source[i])) {
                        target[i] = source[i];
                    }
                }
            }
            function getUniqueTypeParameters(context, typeParameters) {
                var result = [];
                var oldTypeParameters;
                var newTypeParameters;
                for (var _i = 0, typeParameters_2 = typeParameters; _i < typeParameters_2.length; _i++) {
                    var tp = typeParameters_2[_i];
                    var name = tp.symbol.escapedName;
                    if (hasTypeParameterByName(context.inferredTypeParameters, name) || hasTypeParameterByName(result, name)) {
                        var newName = getUniqueTypeParameterName(ts.concatenate(context.inferredTypeParameters, result), name);
                        var symbol = createSymbol(262144 /* TypeParameter */, newName);
                        var newTypeParameter = createTypeParameter(symbol);
                        newTypeParameter.target = tp;
                        oldTypeParameters = ts.append(oldTypeParameters, tp);
                        newTypeParameters = ts.append(newTypeParameters, newTypeParameter);
                        result.push(newTypeParameter);
                    }
                    else {
                        result.push(tp);
                    }
                }
                if (newTypeParameters) {
                    var mapper = createTypeMapper(oldTypeParameters, newTypeParameters);
                    for (var _a = 0, newTypeParameters_1 = newTypeParameters; _a < newTypeParameters_1.length; _a++) {
                        var tp = newTypeParameters_1[_a];
                        tp.mapper = mapper;
                    }
                }
                return result;
            }
            function hasTypeParameterByName(typeParameters, name) {
                return ts.some(typeParameters, function (tp) { return tp.symbol.escapedName === name; });
            }
            function getUniqueTypeParameterName(typeParameters, baseName) {
                var len = baseName.length;
                while (len > 1 && baseName.charCodeAt(len - 1) >= 48 /* _0 */ && baseName.charCodeAt(len - 1) <= 57 /* _9 */)
                    len--;
                var s = baseName.slice(0, len);
                for (var index = 1; true; index++) {
                    var augmentedName = s + index;
                    if (!hasTypeParameterByName(typeParameters, augmentedName)) {
                        return augmentedName;
                    }
                }
            }
            function getReturnTypeOfSingleNonGenericCallSignature(funcType) {
                var signature = getSingleCallSignature(funcType);
                if (signature && !signature.typeParameters) {
                    return getReturnTypeOfSignature(signature);
                }
            }
            function getReturnTypeOfSingleNonGenericSignatureOfCallChain(expr) {
                var funcType = checkExpression(expr.expression);
                var nonOptionalType = getOptionalExpressionType(funcType, expr.expression);
                var returnType = getReturnTypeOfSingleNonGenericCallSignature(funcType);
                return returnType && propagateOptionalTypeMarker(returnType, expr, nonOptionalType !== funcType);
            }
            /**
             * Returns the type of an expression. Unlike checkExpression, this function is simply concerned
             * with computing the type and may not fully check all contained sub-expressions for errors.
             */
            function getTypeOfExpression(node) {
                // Don't bother caching types that require no flow analysis and are quick to compute.
                var quickType = getQuickTypeOfExpression(node);
                if (quickType) {
                    return quickType;
                }
                // If a type has been cached for the node, return it.
                if (node.flags & 67108864 /* TypeCached */ && flowTypeCache) {
                    var cachedType = flowTypeCache[getNodeId(node)];
                    if (cachedType) {
                        return cachedType;
                    }
                }
                var startInvocationCount = flowInvocationCount;
                var type = checkExpression(node);
                // If control flow analysis was required to determine the type, it is worth caching.
                if (flowInvocationCount !== startInvocationCount) {
                    var cache = flowTypeCache || (flowTypeCache = []);
                    cache[getNodeId(node)] = type;
                    ts.setNodeFlags(node, node.flags | 67108864 /* TypeCached */);
                }
                return type;
            }
            function getQuickTypeOfExpression(node) {
                var expr = ts.skipParentheses(node);
                // Optimize for the common case of a call to a function with a single non-generic call
                // signature where we can just fetch the return type without checking the arguments.
                if (ts.isCallExpression(expr) && expr.expression.kind !== 106 /* SuperKeyword */ && !ts.isRequireCall(expr, /*checkArgumentIsStringLiteralLike*/ true) && !isSymbolOrSymbolForCall(expr)) {
                    var type = ts.isCallChain(expr) ? getReturnTypeOfSingleNonGenericSignatureOfCallChain(expr) :
                        getReturnTypeOfSingleNonGenericCallSignature(checkNonNullExpression(expr.expression));
                    if (type) {
                        return type;
                    }
                }
                else if (ts.isAssertionExpression(expr) && !ts.isConstTypeReference(expr.type)) {
                    return getTypeFromTypeNode(expr.type);
                }
                else if (node.kind === 8 /* NumericLiteral */ || node.kind === 10 /* StringLiteral */ ||
                    node.kind === 110 /* TrueKeyword */ || node.kind === 95 /* FalseKeyword */) {
                    return checkExpression(node);
                }
                return undefined;
            }
            /**
             * Returns the type of an expression. Unlike checkExpression, this function is simply concerned
             * with computing the type and may not fully check all contained sub-expressions for errors.
             * It is intended for uses where you know there is no contextual type,
             * and requesting the contextual type might cause a circularity or other bad behaviour.
             * It sets the contextual type of the node to any before calling getTypeOfExpression.
             */
            function getContextFreeTypeOfExpression(node) {
                var links = getNodeLinks(node);
                if (links.contextFreeType) {
                    return links.contextFreeType;
                }
                var saveContextualType = node.contextualType;
                node.contextualType = anyType;
                try {
                    var type = links.contextFreeType = checkExpression(node, 4 /* SkipContextSensitive */);
                    return type;
                }
                finally {
                    // In the event our operation is canceled or some other exception occurs, reset the contextual type
                    // so that we do not accidentally hold onto an instance of the checker, as a Type created in the services layer
                    // may hold onto the checker that created it.
                    node.contextualType = saveContextualType;
                }
            }
            function checkExpression(node, checkMode, forceTuple) {
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("check" /* Check */, "checkExpression", { kind: node.kind, pos: node.pos, end: node.end });
                var saveCurrentNode = currentNode;
                currentNode = node;
                instantiationCount = 0;
                var uninstantiatedType = checkExpressionWorker(node, checkMode, forceTuple);
                var type = instantiateTypeWithSingleGenericCallSignature(node, uninstantiatedType, checkMode);
                if (isConstEnumObjectType(type)) {
                    checkConstEnumAccess(node, type);
                }
                currentNode = saveCurrentNode;
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                return type;
            }
            function checkConstEnumAccess(node, type) {
                // enum object type for const enums are only permitted in:
                // - 'left' in property access
                // - 'object' in indexed access
                // - target in rhs of import statement
                var ok = (node.parent.kind === 204 /* PropertyAccessExpression */ && node.parent.expression === node) ||
                    (node.parent.kind === 205 /* ElementAccessExpression */ && node.parent.expression === node) ||
                    ((node.kind === 79 /* Identifier */ || node.kind === 159 /* QualifiedName */) && isInRightSideOfImportOrExportAssignment(node) ||
                        (node.parent.kind === 179 /* TypeQuery */ && node.parent.exprName === node)) ||
                    (node.parent.kind === 273 /* ExportSpecifier */); // We allow reexporting const enums
                if (!ok) {
                    error(node, ts.Diagnostics.const_enums_can_only_be_used_in_property_or_index_access_expressions_or_the_right_hand_side_of_an_import_declaration_or_export_assignment_or_type_query);
                }
                if (compilerOptions.isolatedModules) {
                    ts.Debug.assert(!!(type.symbol.flags & 128 /* ConstEnum */));
                    var constEnumDeclaration = type.symbol.valueDeclaration;
                    if (constEnumDeclaration.flags & 8388608 /* Ambient */) {
                        error(node, ts.Diagnostics.Cannot_access_ambient_const_enums_when_the_isolatedModules_flag_is_provided);
                    }
                }
            }
            function checkParenthesizedExpression(node, checkMode) {
                var tag = ts.isInJSFile(node) ? ts.getJSDocTypeTag(node) : undefined;
                if (tag) {
                    return checkAssertionWorker(tag.typeExpression.type, tag.typeExpression.type, node.expression, checkMode);
                }
                return checkExpression(node.expression, checkMode);
            }
            function checkExpressionWorker(node, checkMode, forceTuple) {
                var kind = node.kind;
                if (cancellationToken) {
                    // Only bother checking on a few construct kinds.  We don't want to be excessively
                    // hitting the cancellation token on every node we check.
                    switch (kind) {
                        case 224 /* ClassExpression */:
                        case 211 /* FunctionExpression */:
                        case 212 /* ArrowFunction */:
                            cancellationToken.throwIfCancellationRequested();
                    }
                }
                switch (kind) {
                    case 79 /* Identifier */:
                        return checkIdentifier(node, checkMode);
                    case 108 /* ThisKeyword */:
                        return checkThisExpression(node);
                    case 106 /* SuperKeyword */:
                        return checkSuperExpression(node);
                    case 104 /* NullKeyword */:
                        return nullWideningType;
                    case 14 /* NoSubstitutionTemplateLiteral */:
                    case 10 /* StringLiteral */:
                        return getFreshTypeOfLiteralType(getStringLiteralType(node.text));
                    case 8 /* NumericLiteral */:
                        checkGrammarNumericLiteral(node);
                        return getFreshTypeOfLiteralType(getNumberLiteralType(+node.text));
                    case 9 /* BigIntLiteral */:
                        checkGrammarBigIntLiteral(node);
                        return getFreshTypeOfLiteralType(getBigIntLiteralType({
                            negative: false,
                            base10Value: ts.parsePseudoBigInt(node.text)
                        }));
                    case 110 /* TrueKeyword */:
                        return trueType;
                    case 95 /* FalseKeyword */:
                        return falseType;
                    case 221 /* TemplateExpression */:
                        return checkTemplateExpression(node);
                    case 13 /* RegularExpressionLiteral */:
                        return globalRegExpType;
                    case 202 /* ArrayLiteralExpression */:
                        return checkArrayLiteral(node, checkMode, forceTuple);
                    case 203 /* ObjectLiteralExpression */:
                        return checkObjectLiteral(node, checkMode);
                    case 204 /* PropertyAccessExpression */:
                        return checkPropertyAccessExpression(node, checkMode);
                    case 159 /* QualifiedName */:
                        return checkQualifiedName(node, checkMode);
                    case 205 /* ElementAccessExpression */:
                        return checkIndexedAccess(node, checkMode);
                    case 206 /* CallExpression */:
                        if (node.expression.kind === 100 /* ImportKeyword */) {
                            return checkImportCallExpression(node);
                        }
                    // falls through
                    case 207 /* NewExpression */:
                        return checkCallExpression(node, checkMode);
                    case 208 /* TaggedTemplateExpression */:
                        return checkTaggedTemplateExpression(node);
                    case 210 /* ParenthesizedExpression */:
                        return checkParenthesizedExpression(node, checkMode);
                    case 224 /* ClassExpression */:
                        return checkClassExpression(node);
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        return checkFunctionExpressionOrObjectLiteralMethod(node, checkMode);
                    case 214 /* TypeOfExpression */:
                        return checkTypeOfExpression(node);
                    case 209 /* TypeAssertionExpression */:
                    case 227 /* AsExpression */:
                        return checkAssertion(node);
                    case 228 /* NonNullExpression */:
                        return checkNonNullAssertion(node);
                    case 229 /* MetaProperty */:
                        return checkMetaProperty(node);
                    case 213 /* DeleteExpression */:
                        return checkDeleteExpression(node);
                    case 215 /* VoidExpression */:
                        return checkVoidExpression(node);
                    case 216 /* AwaitExpression */:
                        return checkAwaitExpression(node);
                    case 217 /* PrefixUnaryExpression */:
                        return checkPrefixUnaryExpression(node);
                    case 218 /* PostfixUnaryExpression */:
                        return checkPostfixUnaryExpression(node);
                    case 219 /* BinaryExpression */:
                        return checkBinaryExpression(node, checkMode);
                    case 220 /* ConditionalExpression */:
                        return checkConditionalExpression(node, checkMode);
                    case 223 /* SpreadElement */:
                        return checkSpreadExpression(node, checkMode);
                    case 225 /* OmittedExpression */:
                        return undefinedWideningType;
                    case 222 /* YieldExpression */:
                        return checkYieldExpression(node);
                    case 230 /* SyntheticExpression */:
                        return checkSyntheticExpression(node);
                    case 286 /* JsxExpression */:
                        return checkJsxExpression(node, checkMode);
                    case 276 /* JsxElement */:
                        return checkJsxElement(node, checkMode);
                    case 277 /* JsxSelfClosingElement */:
                        return checkJsxSelfClosingElement(node, checkMode);
                    case 280 /* JsxFragment */:
                        return checkJsxFragment(node);
                    case 284 /* JsxAttributes */:
                        return checkJsxAttributes(node, checkMode);
                    case 278 /* JsxOpeningElement */:
                        ts.Debug.fail("Shouldn't ever directly check a JsxOpeningElement");
                }
                return errorType;
            }
            // DECLARATION AND STATEMENT TYPE CHECKING
            function checkTypeParameter(node) {
                // Grammar Checking
                if (node.expression) {
                    grammarErrorOnFirstToken(node.expression, ts.Diagnostics.Type_expected);
                }
                checkSourceElement(node.constraint);
                checkSourceElement(node.default);
                var typeParameter = getDeclaredTypeOfTypeParameter(getSymbolOfNode(node));
                // Resolve base constraint to reveal circularity errors
                getBaseConstraintOfType(typeParameter);
                if (!hasNonCircularTypeParameterDefault(typeParameter)) {
                    error(node.default, ts.Diagnostics.Type_parameter_0_has_a_circular_default, typeToString(typeParameter));
                }
                var constraintType = getConstraintOfTypeParameter(typeParameter);
                var defaultType = getDefaultFromTypeParameter(typeParameter);
                if (constraintType && defaultType) {
                    checkTypeAssignableTo(defaultType, getTypeWithThisArgument(instantiateType(constraintType, makeUnaryTypeMapper(typeParameter, defaultType)), defaultType), node.default, ts.Diagnostics.Type_0_does_not_satisfy_the_constraint_1);
                }
                if (produceDiagnostics) {
                    checkTypeNameIsReserved(node.name, ts.Diagnostics.Type_parameter_name_cannot_be_0);
                }
            }
            function checkParameter(node) {
                // Grammar checking
                // It is a SyntaxError if the Identifier "eval" or the Identifier "arguments" occurs as the
                // Identifier in a PropertySetParameterList of a PropertyAssignment that is contained in strict code
                // or if its FunctionBody is strict code(11.1.5).
                checkGrammarDecoratorsAndModifiers(node);
                checkVariableLikeDeclaration(node);
                var func = ts.getContainingFunction(node);
                if (ts.hasSyntacticModifier(node, 16476 /* ParameterPropertyModifier */)) {
                    if (!(func.kind === 169 /* Constructor */ && ts.nodeIsPresent(func.body))) {
                        error(node, ts.Diagnostics.A_parameter_property_is_only_allowed_in_a_constructor_implementation);
                    }
                    if (func.kind === 169 /* Constructor */ && ts.isIdentifier(node.name) && node.name.escapedText === "constructor") {
                        error(node.name, ts.Diagnostics.constructor_cannot_be_used_as_a_parameter_property_name);
                    }
                }
                if (node.questionToken && ts.isBindingPattern(node.name) && func.body) {
                    error(node, ts.Diagnostics.A_binding_pattern_parameter_cannot_be_optional_in_an_implementation_signature);
                }
                if (node.name && ts.isIdentifier(node.name) && (node.name.escapedText === "this" || node.name.escapedText === "new")) {
                    if (func.parameters.indexOf(node) !== 0) {
                        error(node, ts.Diagnostics.A_0_parameter_must_be_the_first_parameter, node.name.escapedText);
                    }
                    if (func.kind === 169 /* Constructor */ || func.kind === 173 /* ConstructSignature */ || func.kind === 178 /* ConstructorType */) {
                        error(node, ts.Diagnostics.A_constructor_cannot_have_a_this_parameter);
                    }
                    if (func.kind === 212 /* ArrowFunction */) {
                        error(node, ts.Diagnostics.An_arrow_function_cannot_have_a_this_parameter);
                    }
                    if (func.kind === 170 /* GetAccessor */ || func.kind === 171 /* SetAccessor */) {
                        error(node, ts.Diagnostics.get_and_set_accessors_cannot_declare_this_parameters);
                    }
                }
                // Only check rest parameter type if it's not a binding pattern. Since binding patterns are
                // not allowed in a rest parameter, we already have an error from checkGrammarParameterList.
                if (node.dotDotDotToken && !ts.isBindingPattern(node.name) && !isTypeAssignableTo(getReducedType(getTypeOfSymbol(node.symbol)), anyReadonlyArrayType)) {
                    error(node, ts.Diagnostics.A_rest_parameter_must_be_of_an_array_type);
                }
            }
            function checkTypePredicate(node) {
                var parent = getTypePredicateParent(node);
                if (!parent) {
                    // The parent must not be valid.
                    error(node, ts.Diagnostics.A_type_predicate_is_only_allowed_in_return_type_position_for_functions_and_methods);
                    return;
                }
                var signature = getSignatureFromDeclaration(parent);
                var typePredicate = getTypePredicateOfSignature(signature);
                if (!typePredicate) {
                    return;
                }
                checkSourceElement(node.type);
                var parameterName = node.parameterName;
                if (typePredicate.kind === 0 /* This */ || typePredicate.kind === 2 /* AssertsThis */) {
                    getTypeFromThisTypeNode(parameterName);
                }
                else {
                    if (typePredicate.parameterIndex >= 0) {
                        if (signatureHasRestParameter(signature) && typePredicate.parameterIndex === signature.parameters.length - 1) {
                            error(parameterName, ts.Diagnostics.A_type_predicate_cannot_reference_a_rest_parameter);
                        }
                        else {
                            if (typePredicate.type) {
                                var leadingError = function () { return ts.chainDiagnosticMessages(/*details*/ undefined, ts.Diagnostics.A_type_predicate_s_type_must_be_assignable_to_its_parameter_s_type); };
                                checkTypeAssignableTo(typePredicate.type, getTypeOfSymbol(signature.parameters[typePredicate.parameterIndex]), node.type, 
                                /*headMessage*/ undefined, leadingError);
                            }
                        }
                    }
                    else if (parameterName) {
                        var hasReportedError = false;
                        for (var _i = 0, _a = parent.parameters; _i < _a.length; _i++) {
                            var name = _a[_i].name;
                            if (ts.isBindingPattern(name) &&
                                checkIfTypePredicateVariableIsDeclaredInBindingPattern(name, parameterName, typePredicate.parameterName)) {
                                hasReportedError = true;
                                break;
                            }
                        }
                        if (!hasReportedError) {
                            error(node.parameterName, ts.Diagnostics.Cannot_find_parameter_0, typePredicate.parameterName);
                        }
                    }
                }
            }
            function getTypePredicateParent(node) {
                switch (node.parent.kind) {
                    case 212 /* ArrowFunction */:
                    case 172 /* CallSignature */:
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 177 /* FunctionType */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                        var parent = node.parent;
                        if (node === parent.type) {
                            return parent;
                        }
                }
            }
            function checkIfTypePredicateVariableIsDeclaredInBindingPattern(pattern, predicateVariableNode, predicateVariableName) {
                for (var _i = 0, _a = pattern.elements; _i < _a.length; _i++) {
                    var element = _a[_i];
                    if (ts.isOmittedExpression(element)) {
                        continue;
                    }
                    var name = element.name;
                    if (name.kind === 79 /* Identifier */ && name.escapedText === predicateVariableName) {
                        error(predicateVariableNode, ts.Diagnostics.A_type_predicate_cannot_reference_element_0_in_a_binding_pattern, predicateVariableName);
                        return true;
                    }
                    else if (name.kind === 200 /* ArrayBindingPattern */ || name.kind === 199 /* ObjectBindingPattern */) {
                        if (checkIfTypePredicateVariableIsDeclaredInBindingPattern(name, predicateVariableNode, predicateVariableName)) {
                            return true;
                        }
                    }
                }
            }
            function checkSignatureDeclaration(node) {
                // Grammar checking
                if (node.kind === 174 /* IndexSignature */) {
                    checkGrammarIndexSignature(node);
                }
                // TODO (yuisu): Remove this check in else-if when SyntaxKind.Construct is moved and ambient context is handled
                else if (node.kind === 177 /* FunctionType */ || node.kind === 254 /* FunctionDeclaration */ || node.kind === 178 /* ConstructorType */ ||
                    node.kind === 172 /* CallSignature */ || node.kind === 169 /* Constructor */ ||
                    node.kind === 173 /* ConstructSignature */) {
                    checkGrammarFunctionLikeDeclaration(node);
                }
                var functionFlags = ts.getFunctionFlags(node);
                if (!(functionFlags & 4 /* Invalid */)) {
                    // Async generators prior to ESNext require the __await and __asyncGenerator helpers
                    if ((functionFlags & 3 /* AsyncGenerator */) === 3 /* AsyncGenerator */ && languageVersion < 99 /* ESNext */) {
                        checkExternalEmitHelpers(node, 6144 /* AsyncGeneratorIncludes */);
                    }
                    // Async functions prior to ES2017 require the __awaiter helper
                    if ((functionFlags & 3 /* AsyncGenerator */) === 2 /* Async */ && languageVersion < 4 /* ES2017 */) {
                        checkExternalEmitHelpers(node, 64 /* Awaiter */);
                    }
                    // Generator functions, Async functions, and Async Generator functions prior to
                    // ES2015 require the __generator helper
                    if ((functionFlags & 3 /* AsyncGenerator */) !== 0 /* Normal */ && languageVersion < 2 /* ES2015 */) {
                        checkExternalEmitHelpers(node, 128 /* Generator */);
                    }
                }
                checkTypeParameters(node.typeParameters);
                ts.forEach(node.parameters, checkParameter);
                // TODO(rbuckton): Should we start checking JSDoc types?
                if (node.type) {
                    checkSourceElement(node.type);
                }
                if (produceDiagnostics) {
                    checkCollisionWithArgumentsInGeneratedCode(node);
                    var returnTypeNode = ts.getEffectiveReturnTypeNode(node);
                    if (noImplicitAny && !returnTypeNode) {
                        switch (node.kind) {
                            case 173 /* ConstructSignature */:
                                error(node, ts.Diagnostics.Construct_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type);
                                break;
                            case 172 /* CallSignature */:
                                error(node, ts.Diagnostics.Call_signature_which_lacks_return_type_annotation_implicitly_has_an_any_return_type);
                                break;
                        }
                    }
                    if (returnTypeNode) {
                        var functionFlags_1 = ts.getFunctionFlags(node);
                        if ((functionFlags_1 & (4 /* Invalid */ | 1 /* Generator */)) === 1 /* Generator */) {
                            var returnType = getTypeFromTypeNode(returnTypeNode);
                            if (returnType === voidType) {
                                error(returnTypeNode, ts.Diagnostics.A_generator_cannot_have_a_void_type_annotation);
                            }
                            else {
                                // Naively, one could check that Generator<any, any, any> is assignable to the return type annotation.
                                // However, that would not catch the error in the following case.
                                //
                                //    interface BadGenerator extends Iterable<number>, Iterator<string> { }
                                //    function* g(): BadGenerator { } // Iterable and Iterator have different types!
                                //
                                var generatorYieldType = getIterationTypeOfGeneratorFunctionReturnType(0 /* Yield */, returnType, (functionFlags_1 & 2 /* Async */) !== 0) || anyType;
                                var generatorReturnType = getIterationTypeOfGeneratorFunctionReturnType(1 /* Return */, returnType, (functionFlags_1 & 2 /* Async */) !== 0) || generatorYieldType;
                                var generatorNextType = getIterationTypeOfGeneratorFunctionReturnType(2 /* Next */, returnType, (functionFlags_1 & 2 /* Async */) !== 0) || unknownType;
                                var generatorInstantiation = createGeneratorReturnType(generatorYieldType, generatorReturnType, generatorNextType, !!(functionFlags_1 & 2 /* Async */));
                                checkTypeAssignableTo(generatorInstantiation, returnType, returnTypeNode);
                            }
                        }
                        else if ((functionFlags_1 & 3 /* AsyncGenerator */) === 2 /* Async */) {
                            checkAsyncFunctionReturnType(node, returnTypeNode);
                        }
                    }
                    if (node.kind !== 174 /* IndexSignature */ && node.kind !== 312 /* JSDocFunctionType */) {
                        registerForUnusedIdentifiersCheck(node);
                    }
                }
            }
            function checkClassForDuplicateDeclarations(node) {
                var instanceNames = new ts.Map();
                var staticNames = new ts.Map();
                // instance and static private identifiers share the same scope
                var privateIdentifiers = new ts.Map();
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    if (member.kind === 169 /* Constructor */) {
                        for (var _b = 0, _c = member.parameters; _b < _c.length; _b++) {
                            var param = _c[_b];
                            if (ts.isParameterPropertyDeclaration(param, member) && !ts.isBindingPattern(param.name)) {
                                addName(instanceNames, param.name, param.name.escapedText, 3 /* GetOrSetAccessor */);
                            }
                        }
                    }
                    else {
                        var isStaticMember = ts.isStatic(member);
                        var name = member.name;
                        if (!name) {
                            continue;
                        }
                        var isPrivate = ts.isPrivateIdentifier(name);
                        var privateStaticFlags = isPrivate && isStaticMember ? 16 /* PrivateStatic */ : 0;
                        var names = isPrivate ? privateIdentifiers :
                            isStaticMember ? staticNames :
                                instanceNames;
                        var memberName = name && ts.getPropertyNameForPropertyNameNode(name);
                        if (memberName) {
                            switch (member.kind) {
                                case 170 /* GetAccessor */:
                                    addName(names, name, memberName, 1 /* GetAccessor */ | privateStaticFlags);
                                    break;
                                case 171 /* SetAccessor */:
                                    addName(names, name, memberName, 2 /* SetAccessor */ | privateStaticFlags);
                                    break;
                                case 165 /* PropertyDeclaration */:
                                    addName(names, name, memberName, 3 /* GetOrSetAccessor */ | privateStaticFlags);
                                    break;
                                case 167 /* MethodDeclaration */:
                                    addName(names, name, memberName, 8 /* Method */ | privateStaticFlags);
                                    break;
                            }
                        }
                    }
                }
                function addName(names, location, name, meaning) {
                    var prev = names.get(name);
                    if (prev) {
                        // For private identifiers, do not allow mixing of static and instance members with the same name
                        if ((prev & 16 /* PrivateStatic */) !== (meaning & 16 /* PrivateStatic */)) {
                            error(location, ts.Diagnostics.Duplicate_identifier_0_Static_and_instance_elements_cannot_share_the_same_private_name, ts.getTextOfNode(location));
                        }
                        else {
                            var prevIsMethod = !!(prev & 8 /* Method */);
                            var isMethod = !!(meaning & 8 /* Method */);
                            if (prevIsMethod || isMethod) {
                                if (prevIsMethod !== isMethod) {
                                    error(location, ts.Diagnostics.Duplicate_identifier_0, ts.getTextOfNode(location));
                                }
                                // If this is a method/method duplication is might be an overload, so this will be handled when overloads are considered
                            }
                            else if (prev & meaning & ~16 /* PrivateStatic */) {
                                error(location, ts.Diagnostics.Duplicate_identifier_0, ts.getTextOfNode(location));
                            }
                            else {
                                names.set(name, prev | meaning);
                            }
                        }
                    }
                    else {
                        names.set(name, meaning);
                    }
                }
            }
            /**
             * Static members being set on a constructor function may conflict with built-in properties
             * of Function. Esp. in ECMAScript 5 there are non-configurable and non-writable
             * built-in properties. This check issues a transpile error when a class has a static
             * member with the same name as a non-writable built-in property.
             *
             * @see http://www.ecma-international.org/ecma-262/5.1/#sec-15.3.3
             * @see http://www.ecma-international.org/ecma-262/5.1/#sec-15.3.5
             * @see http://www.ecma-international.org/ecma-262/6.0/#sec-properties-of-the-function-constructor
             * @see http://www.ecma-international.org/ecma-262/6.0/#sec-function-instances
             */
            function checkClassForStaticPropertyNameConflicts(node) {
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    var memberNameNode = member.name;
                    var isStaticMember = ts.isStatic(member);
                    if (isStaticMember && memberNameNode) {
                        var memberName = ts.getPropertyNameForPropertyNameNode(memberNameNode);
                        switch (memberName) {
                            case "name":
                            case "length":
                            case "caller":
                            case "arguments":
                            case "prototype":
                                var message = ts.Diagnostics.Static_property_0_conflicts_with_built_in_property_Function_0_of_constructor_function_1;
                                var className = getNameOfSymbolAsWritten(getSymbolOfNode(node));
                                error(memberNameNode, message, memberName, className);
                                break;
                        }
                    }
                }
            }
            function checkObjectTypeForDuplicateDeclarations(node) {
                var names = new ts.Map();
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    if (member.kind === 164 /* PropertySignature */) {
                        var memberName = void 0;
                        var name = member.name;
                        switch (name.kind) {
                            case 10 /* StringLiteral */:
                            case 8 /* NumericLiteral */:
                                memberName = name.text;
                                break;
                            case 79 /* Identifier */:
                                memberName = ts.idText(name);
                                break;
                            default:
                                continue;
                        }
                        if (names.get(memberName)) {
                            error(ts.getNameOfDeclaration(member.symbol.valueDeclaration), ts.Diagnostics.Duplicate_identifier_0, memberName);
                            error(member.name, ts.Diagnostics.Duplicate_identifier_0, memberName);
                        }
                        else {
                            names.set(memberName, true);
                        }
                    }
                }
            }
            function checkTypeForDuplicateIndexSignatures(node) {
                if (node.kind === 256 /* InterfaceDeclaration */) {
                    var nodeSymbol = getSymbolOfNode(node);
                    // in case of merging interface declaration it is possible that we'll enter this check procedure several times for every declaration
                    // to prevent this run check only for the first declaration of a given kind
                    if (nodeSymbol.declarations && nodeSymbol.declarations.length > 0 && nodeSymbol.declarations[0] !== node) {
                        return;
                    }
                }
                // TypeScript 1.0 spec (April 2014)
                // 3.7.4: An object type can contain at most one string index signature and one numeric index signature.
                // 8.5: A class declaration can have at most one string index member declaration and one numeric index member declaration
                var indexSymbol = getIndexSymbol(getSymbolOfNode(node));
                if (indexSymbol === null || indexSymbol === void 0 ? void 0 : indexSymbol.declarations) {
                    var indexSignatureMap_1 = new ts.Map();
                    var _loop_26 = function (declaration) {
                        if (declaration.parameters.length === 1 && declaration.parameters[0].type) {
                            forEachType(getTypeFromTypeNode(declaration.parameters[0].type), function (type) {
                                var entry = indexSignatureMap_1.get(getTypeId(type));
                                if (entry) {
                                    entry.declarations.push(declaration);
                                }
                                else {
                                    indexSignatureMap_1.set(getTypeId(type), { type: type, declarations: [declaration] });
                                }
                            });
                        }
                    };
                    for (var _i = 0, _a = indexSymbol.declarations; _i < _a.length; _i++) {
                        var declaration = _a[_i];
                        _loop_26(declaration);
                    }
                    indexSignatureMap_1.forEach(function (entry) {
                        if (entry.declarations.length > 1) {
                            for (var _i = 0, _a = entry.declarations; _i < _a.length; _i++) {
                                var declaration = _a[_i];
                                error(declaration, ts.Diagnostics.Duplicate_index_signature_for_type_0, typeToString(entry.type));
                            }
                        }
                    });
                }
            }
            function checkPropertyDeclaration(node) {
                // Grammar checking
                if (!checkGrammarDecoratorsAndModifiers(node) && !checkGrammarProperty(node))
                    checkGrammarComputedPropertyName(node.name);
                checkVariableLikeDeclaration(node);
                setNodeLinksForPrivateIdentifierScope(node);
                if (ts.isPrivateIdentifier(node.name) && ts.hasStaticModifier(node) && node.initializer && languageVersion === 99 /* ESNext */ && !compilerOptions.useDefineForClassFields) {
                    error(node.initializer, ts.Diagnostics.Static_fields_with_private_names_can_t_have_initializers_when_the_useDefineForClassFields_flag_is_not_specified_with_a_target_of_esnext_Consider_adding_the_useDefineForClassFields_flag);
                }
                // property signatures already report "initializer not allowed in ambient context" elsewhere
                if (ts.hasSyntacticModifier(node, 128 /* Abstract */) && node.kind === 165 /* PropertyDeclaration */ && node.initializer) {
                    error(node, ts.Diagnostics.Property_0_cannot_have_an_initializer_because_it_is_marked_abstract, ts.declarationNameToString(node.name));
                }
            }
            function checkPropertySignature(node) {
                if (ts.isPrivateIdentifier(node.name)) {
                    error(node, ts.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies);
                }
                return checkPropertyDeclaration(node);
            }
            function checkMethodDeclaration(node) {
                // Grammar checking
                if (!checkGrammarMethod(node))
                    checkGrammarComputedPropertyName(node.name);
                // Grammar checking for modifiers is done inside the function checkGrammarFunctionLikeDeclaration
                checkFunctionOrMethodDeclaration(node);
                // method signatures already report "implementation not allowed in ambient context" elsewhere
                if (ts.hasSyntacticModifier(node, 128 /* Abstract */) && node.kind === 167 /* MethodDeclaration */ && node.body) {
                    error(node, ts.Diagnostics.Method_0_cannot_have_an_implementation_because_it_is_marked_abstract, ts.declarationNameToString(node.name));
                }
                // Private named methods are only allowed in class declarations
                if (ts.isPrivateIdentifier(node.name) && !ts.getContainingClass(node)) {
                    error(node, ts.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies);
                }
                setNodeLinksForPrivateIdentifierScope(node);
            }
            function setNodeLinksForPrivateIdentifierScope(node) {
                if (ts.isPrivateIdentifier(node.name) && languageVersion < 99 /* ESNext */) {
                    for (var lexicalScope = ts.getEnclosingBlockScopeContainer(node); !!lexicalScope; lexicalScope = ts.getEnclosingBlockScopeContainer(lexicalScope)) {
                        getNodeLinks(lexicalScope).flags |= 67108864 /* ContainsClassWithPrivateIdentifiers */;
                    }
                    // If this is a private element in a class expression inside the body of a loop,
                    // then we must use a block-scoped binding to store the additional variables required
                    // to transform private elements.
                    if (ts.isClassExpression(node.parent)) {
                        var enclosingIterationStatement = getEnclosingIterationStatement(node.parent);
                        if (enclosingIterationStatement) {
                            getNodeLinks(node.name).flags |= 524288 /* BlockScopedBindingInLoop */;
                            getNodeLinks(enclosingIterationStatement).flags |= 65536 /* LoopWithCapturedBlockScopedBinding */;
                        }
                    }
                }
            }
            function checkClassStaticBlockDeclaration(node) {
                checkGrammarDecoratorsAndModifiers(node);
                ts.forEachChild(node, checkSourceElement);
            }
            function checkConstructorDeclaration(node) {
                // Grammar check on signature of constructor and modifier of the constructor is done in checkSignatureDeclaration function.
                checkSignatureDeclaration(node);
                // Grammar check for checking only related to constructorDeclaration
                if (!checkGrammarConstructorTypeParameters(node))
                    checkGrammarConstructorTypeAnnotation(node);
                checkSourceElement(node.body);
                var symbol = getSymbolOfNode(node);
                var firstDeclaration = ts.getDeclarationOfKind(symbol, node.kind);
                // Only type check the symbol once
                if (node === firstDeclaration) {
                    checkFunctionOrConstructorSymbol(symbol);
                }
                // exit early in the case of signature - super checks are not relevant to them
                if (ts.nodeIsMissing(node.body)) {
                    return;
                }
                if (!produceDiagnostics) {
                    return;
                }
                function isInstancePropertyWithInitializerOrPrivateIdentifierProperty(n) {
                    if (ts.isPrivateIdentifierClassElementDeclaration(n)) {
                        return true;
                    }
                    return n.kind === 165 /* PropertyDeclaration */ &&
                        !ts.isStatic(n) &&
                        !!n.initializer;
                }
                // TS 1.0 spec (April 2014): 8.3.2
                // Constructors of classes with no extends clause may not contain super calls, whereas
                // constructors of derived classes must contain at least one super call somewhere in their function body.
                var containingClassDecl = node.parent;
                if (ts.getClassExtendsHeritageElement(containingClassDecl)) {
                    captureLexicalThis(node.parent, containingClassDecl);
                    var classExtendsNull = classDeclarationExtendsNull(containingClassDecl);
                    var superCall = findFirstSuperCall(node.body);
                    if (superCall) {
                        if (classExtendsNull) {
                            error(superCall, ts.Diagnostics.A_constructor_cannot_contain_a_super_call_when_its_class_extends_null);
                        }
                        // The first statement in the body of a constructor (excluding prologue directives) must be a super call
                        // if both of the following are true:
                        // - The containing class is a derived class.
                        // - The constructor declares parameter properties
                        //   or the containing class declares instance member variables with initializers.
                        var superCallShouldBeFirst = (compilerOptions.target !== 99 /* ESNext */ || !useDefineForClassFields) &&
                            (ts.some(node.parent.members, isInstancePropertyWithInitializerOrPrivateIdentifierProperty) ||
                                ts.some(node.parameters, function (p) { return ts.hasSyntacticModifier(p, 16476 /* ParameterPropertyModifier */); }));
                        // Skip past any prologue directives to find the first statement
                        // to ensure that it was a super call.
                        if (superCallShouldBeFirst) {
                            var statements = node.body.statements;
                            var superCallStatement = void 0;
                            for (var _i = 0, statements_4 = statements; _i < statements_4.length; _i++) {
                                var statement = statements_4[_i];
                                if (statement.kind === 236 /* ExpressionStatement */ && ts.isSuperCall(statement.expression)) {
                                    superCallStatement = statement;
                                    break;
                                }
                                if (!ts.isPrologueDirective(statement)) {
                                    break;
                                }
                            }
                            if (!superCallStatement) {
                                error(node, ts.Diagnostics.A_super_call_must_be_the_first_statement_in_the_constructor_when_a_class_contains_initialized_properties_parameter_properties_or_private_identifiers);
                            }
                        }
                    }
                    else if (!classExtendsNull) {
                        error(node, ts.Diagnostics.Constructors_for_derived_classes_must_contain_a_super_call);
                    }
                }
            }
            function checkAccessorDeclaration(node) {
                if (produceDiagnostics) {
                    // Grammar checking accessors
                    if (!checkGrammarFunctionLikeDeclaration(node) && !checkGrammarAccessor(node))
                        checkGrammarComputedPropertyName(node.name);
                    checkDecorators(node);
                    checkSignatureDeclaration(node);
                    if (node.kind === 170 /* GetAccessor */) {
                        if (!(node.flags & 8388608 /* Ambient */) && ts.nodeIsPresent(node.body) && (node.flags & 256 /* HasImplicitReturn */)) {
                            if (!(node.flags & 512 /* HasExplicitReturn */)) {
                                error(node.name, ts.Diagnostics.A_get_accessor_must_return_a_value);
                            }
                        }
                    }
                    // Do not use hasDynamicName here, because that returns false for well known symbols.
                    // We want to perform checkComputedPropertyName for all computed properties, including
                    // well known symbols.
                    if (node.name.kind === 160 /* ComputedPropertyName */) {
                        checkComputedPropertyName(node.name);
                    }
                    if (hasBindableName(node)) {
                        // TypeScript 1.0 spec (April 2014): 8.4.3
                        // Accessors for the same member name must specify the same accessibility.
                        var symbol = getSymbolOfNode(node);
                        var getter = ts.getDeclarationOfKind(symbol, 170 /* GetAccessor */);
                        var setter = ts.getDeclarationOfKind(symbol, 171 /* SetAccessor */);
                        if (getter && setter && !(getNodeCheckFlags(getter) & 1 /* TypeChecked */)) {
                            getNodeLinks(getter).flags |= 1 /* TypeChecked */;
                            var getterFlags = ts.getEffectiveModifierFlags(getter);
                            var setterFlags = ts.getEffectiveModifierFlags(setter);
                            if ((getterFlags & 128 /* Abstract */) !== (setterFlags & 128 /* Abstract */)) {
                                error(getter.name, ts.Diagnostics.Accessors_must_both_be_abstract_or_non_abstract);
                                error(setter.name, ts.Diagnostics.Accessors_must_both_be_abstract_or_non_abstract);
                            }
                            if (((getterFlags & 16 /* Protected */) && !(setterFlags & (16 /* Protected */ | 8 /* Private */))) ||
                                ((getterFlags & 8 /* Private */) && !(setterFlags & 8 /* Private */))) {
                                error(getter.name, ts.Diagnostics.A_get_accessor_must_be_at_least_as_accessible_as_the_setter);
                                error(setter.name, ts.Diagnostics.A_get_accessor_must_be_at_least_as_accessible_as_the_setter);
                            }
                            var getterType = getAnnotatedAccessorType(getter);
                            var setterType = getAnnotatedAccessorType(setter);
                            if (getterType && setterType) {
                                checkTypeAssignableTo(getterType, setterType, getter, ts.Diagnostics.The_return_type_of_a_get_accessor_must_be_assignable_to_its_set_accessor_type);
                            }
                        }
                    }
                    var returnType = getTypeOfAccessors(getSymbolOfNode(node));
                    if (node.kind === 170 /* GetAccessor */) {
                        checkAllCodePathsInNonVoidFunctionReturnOrThrow(node, returnType);
                    }
                }
                checkSourceElement(node.body);
                setNodeLinksForPrivateIdentifierScope(node);
            }
            function checkMissingDeclaration(node) {
                checkDecorators(node);
            }
            function getEffectiveTypeArguments(node, typeParameters) {
                return fillMissingTypeArguments(ts.map(node.typeArguments, getTypeFromTypeNode), typeParameters, getMinTypeArgumentCount(typeParameters), ts.isInJSFile(node));
            }
            function checkTypeArgumentConstraints(node, typeParameters) {
                var typeArguments;
                var mapper;
                var result = true;
                for (var i = 0; i < typeParameters.length; i++) {
                    var constraint = getConstraintOfTypeParameter(typeParameters[i]);
                    if (constraint) {
                        if (!typeArguments) {
                            typeArguments = getEffectiveTypeArguments(node, typeParameters);
                            mapper = createTypeMapper(typeParameters, typeArguments);
                        }
                        result = result && checkTypeAssignableTo(typeArguments[i], instantiateType(constraint, mapper), node.typeArguments[i], ts.Diagnostics.Type_0_does_not_satisfy_the_constraint_1);
                    }
                }
                return result;
            }
            function getTypeParametersForTypeReference(node) {
                var type = getTypeFromTypeReference(node);
                if (type !== errorType) {
                    var symbol = getNodeLinks(node).resolvedSymbol;
                    if (symbol) {
                        return symbol.flags & 524288 /* TypeAlias */ && getSymbolLinks(symbol).typeParameters ||
                            (ts.getObjectFlags(type) & 4 /* Reference */ ? type.target.localTypeParameters : undefined);
                    }
                }
                return undefined;
            }
            function checkTypeReferenceNode(node) {
                checkGrammarTypeArguments(node, node.typeArguments);
                if (node.kind === 176 /* TypeReference */ && node.typeName.jsdocDotPos !== undefined && !ts.isInJSFile(node) && !ts.isInJSDoc(node)) {
                    grammarErrorAtPos(node, node.typeName.jsdocDotPos, 1, ts.Diagnostics.JSDoc_types_can_only_be_used_inside_documentation_comments);
                }
                ts.forEach(node.typeArguments, checkSourceElement);
                var type = getTypeFromTypeReference(node);
                if (type !== errorType) {
                    if (node.typeArguments && produceDiagnostics) {
                        var typeParameters = getTypeParametersForTypeReference(node);
                        if (typeParameters) {
                            checkTypeArgumentConstraints(node, typeParameters);
                        }
                    }
                    var symbol = getNodeLinks(node).resolvedSymbol;
                    if (symbol) {
                        if (ts.some(symbol.declarations, function (d) { return isTypeDeclaration(d) && !!(d.flags & 134217728 /* Deprecated */); })) {
                            addDeprecatedSuggestion(getDeprecatedSuggestionNode(node), symbol.declarations, symbol.escapedName);
                        }
                        if (type.flags & 32 /* Enum */ && symbol.flags & 8 /* EnumMember */) {
                            error(node, ts.Diagnostics.Enum_type_0_has_members_with_initializers_that_are_not_literals, typeToString(type));
                        }
                    }
                }
            }
            function getTypeArgumentConstraint(node) {
                var typeReferenceNode = ts.tryCast(node.parent, ts.isTypeReferenceType);
                if (!typeReferenceNode)
                    return undefined;
                var typeParameters = getTypeParametersForTypeReference(typeReferenceNode); // TODO: GH#18217
                var constraint = getConstraintOfTypeParameter(typeParameters[typeReferenceNode.typeArguments.indexOf(node)]);
                return constraint && instantiateType(constraint, createTypeMapper(typeParameters, getEffectiveTypeArguments(typeReferenceNode, typeParameters)));
            }
            function checkTypeQuery(node) {
                getTypeFromTypeQueryNode(node);
            }
            function checkTypeLiteral(node) {
                ts.forEach(node.members, checkSourceElement);
                if (produceDiagnostics) {
                    var type = getTypeFromTypeLiteralOrFunctionOrConstructorTypeNode(node);
                    checkIndexConstraints(type);
                    checkTypeForDuplicateIndexSignatures(node);
                    checkObjectTypeForDuplicateDeclarations(node);
                }
            }
            function checkArrayType(node) {
                checkSourceElement(node.elementType);
            }
            function checkTupleType(node) {
                var elementTypes = node.elements;
                var seenOptionalElement = false;
                var seenRestElement = false;
                var hasNamedElement = ts.some(elementTypes, ts.isNamedTupleMember);
                for (var _i = 0, elementTypes_1 = elementTypes; _i < elementTypes_1.length; _i++) {
                    var e = elementTypes_1[_i];
                    if (e.kind !== 195 /* NamedTupleMember */ && hasNamedElement) {
                        grammarErrorOnNode(e, ts.Diagnostics.Tuple_members_must_all_have_names_or_all_not_have_names);
                        break;
                    }
                    var flags = getTupleElementFlags(e);
                    if (flags & 8 /* Variadic */) {
                        var type = getTypeFromTypeNode(e.type);
                        if (!isArrayLikeType(type)) {
                            error(e, ts.Diagnostics.A_rest_element_type_must_be_an_array_type);
                            break;
                        }
                        if (isArrayType(type) || isTupleType(type) && type.target.combinedFlags & 4 /* Rest */) {
                            seenRestElement = true;
                        }
                    }
                    else if (flags & 4 /* Rest */) {
                        if (seenRestElement) {
                            grammarErrorOnNode(e, ts.Diagnostics.A_rest_element_cannot_follow_another_rest_element);
                            break;
                        }
                        seenRestElement = true;
                    }
                    else if (flags & 2 /* Optional */) {
                        if (seenRestElement) {
                            grammarErrorOnNode(e, ts.Diagnostics.An_optional_element_cannot_follow_a_rest_element);
                            break;
                        }
                        seenOptionalElement = true;
                    }
                    else if (seenOptionalElement) {
                        grammarErrorOnNode(e, ts.Diagnostics.A_required_element_cannot_follow_an_optional_element);
                        break;
                    }
                }
                ts.forEach(node.elements, checkSourceElement);
                getTypeFromTypeNode(node);
            }
            function checkUnionOrIntersectionType(node) {
                ts.forEach(node.types, checkSourceElement);
                getTypeFromTypeNode(node);
            }
            function checkIndexedAccessIndexType(type, accessNode) {
                if (!(type.flags & 8388608 /* IndexedAccess */)) {
                    return type;
                }
                // Check if the index type is assignable to 'keyof T' for the object type.
                var objectType = type.objectType;
                var indexType = type.indexType;
                if (isTypeAssignableTo(indexType, getIndexType(objectType, /*stringsOnly*/ false))) {
                    if (accessNode.kind === 205 /* ElementAccessExpression */ && ts.isAssignmentTarget(accessNode) &&
                        ts.getObjectFlags(objectType) & 32 /* Mapped */ && getMappedTypeModifiers(objectType) & 1 /* IncludeReadonly */) {
                        error(accessNode, ts.Diagnostics.Index_signature_in_type_0_only_permits_reading, typeToString(objectType));
                    }
                    return type;
                }
                // Check if we're indexing with a numeric type and if either object or index types
                // is a generic type with a constraint that has a numeric index signature.
                var apparentObjectType = getApparentType(objectType);
                if (getIndexInfoOfType(apparentObjectType, numberType) && isTypeAssignableToKind(indexType, 296 /* NumberLike */)) {
                    return type;
                }
                if (isGenericObjectType(objectType)) {
                    var propertyName_1 = getPropertyNameFromIndex(indexType, accessNode);
                    if (propertyName_1) {
                        var propertySymbol = forEachType(apparentObjectType, function (t) { return getPropertyOfType(t, propertyName_1); });
                        if (propertySymbol && ts.getDeclarationModifierFlagsFromSymbol(propertySymbol) & 24 /* NonPublicAccessibilityModifier */) {
                            error(accessNode, ts.Diagnostics.Private_or_protected_member_0_cannot_be_accessed_on_a_type_parameter, ts.unescapeLeadingUnderscores(propertyName_1));
                            return errorType;
                        }
                    }
                }
                error(accessNode, ts.Diagnostics.Type_0_cannot_be_used_to_index_type_1, typeToString(indexType), typeToString(objectType));
                return errorType;
            }
            function checkIndexedAccessType(node) {
                checkSourceElement(node.objectType);
                checkSourceElement(node.indexType);
                checkIndexedAccessIndexType(getTypeFromIndexedAccessTypeNode(node), node);
            }
            function checkMappedType(node) {
                checkSourceElement(node.typeParameter);
                checkSourceElement(node.nameType);
                checkSourceElement(node.type);
                if (!node.type) {
                    reportImplicitAny(node, anyType);
                }
                var type = getTypeFromMappedTypeNode(node);
                var nameType = getNameTypeFromMappedType(type);
                if (nameType) {
                    checkTypeAssignableTo(nameType, keyofConstraintType, node.nameType);
                }
                else {
                    var constraintType = getConstraintTypeFromMappedType(type);
                    checkTypeAssignableTo(constraintType, keyofConstraintType, ts.getEffectiveConstraintOfTypeParameter(node.typeParameter));
                }
            }
            function checkThisType(node) {
                getTypeFromThisTypeNode(node);
            }
            function checkTypeOperator(node) {
                checkGrammarTypeOperatorNode(node);
                checkSourceElement(node.type);
            }
            function checkConditionalType(node) {
                ts.forEachChild(node, checkSourceElement);
            }
            function checkInferType(node) {
                if (!ts.findAncestor(node, function (n) { return n.parent && n.parent.kind === 187 /* ConditionalType */ && n.parent.extendsType === n; })) {
                    grammarErrorOnNode(node, ts.Diagnostics.infer_declarations_are_only_permitted_in_the_extends_clause_of_a_conditional_type);
                }
                checkSourceElement(node.typeParameter);
                registerForUnusedIdentifiersCheck(node);
            }
            function checkTemplateLiteralType(node) {
                for (var _i = 0, _a = node.templateSpans; _i < _a.length; _i++) {
                    var span = _a[_i];
                    checkSourceElement(span.type);
                    var type = getTypeFromTypeNode(span.type);
                    checkTypeAssignableTo(type, templateConstraintType, span.type);
                }
                getTypeFromTypeNode(node);
            }
            function checkImportType(node) {
                checkSourceElement(node.argument);
                getTypeFromTypeNode(node);
            }
            function checkNamedTupleMember(node) {
                if (node.dotDotDotToken && node.questionToken) {
                    grammarErrorOnNode(node, ts.Diagnostics.A_tuple_member_cannot_be_both_optional_and_rest);
                }
                if (node.type.kind === 183 /* OptionalType */) {
                    grammarErrorOnNode(node.type, ts.Diagnostics.A_labeled_tuple_element_is_declared_as_optional_with_a_question_mark_after_the_name_and_before_the_colon_rather_than_after_the_type);
                }
                if (node.type.kind === 184 /* RestType */) {
                    grammarErrorOnNode(node.type, ts.Diagnostics.A_labeled_tuple_element_is_declared_as_rest_with_a_before_the_name_rather_than_before_the_type);
                }
                checkSourceElement(node.type);
                getTypeFromTypeNode(node);
            }
            function isPrivateWithinAmbient(node) {
                return (ts.hasEffectiveModifier(node, 8 /* Private */) || ts.isPrivateIdentifierClassElementDeclaration(node)) && !!(node.flags & 8388608 /* Ambient */);
            }
            function getEffectiveDeclarationFlags(n, flagsToCheck) {
                var flags = ts.getCombinedModifierFlags(n);
                // children of classes (even ambient classes) should not be marked as ambient or export
                // because those flags have no useful semantics there.
                if (n.parent.kind !== 256 /* InterfaceDeclaration */ &&
                    n.parent.kind !== 255 /* ClassDeclaration */ &&
                    n.parent.kind !== 224 /* ClassExpression */ &&
                    n.flags & 8388608 /* Ambient */) {
                    if (!(flags & 2 /* Ambient */) && !(ts.isModuleBlock(n.parent) && ts.isModuleDeclaration(n.parent.parent) && ts.isGlobalScopeAugmentation(n.parent.parent))) {
                        // It is nested in an ambient context, which means it is automatically exported
                        flags |= 1 /* Export */;
                    }
                    flags |= 2 /* Ambient */;
                }
                return flags & flagsToCheck;
            }
            function checkFunctionOrConstructorSymbol(symbol) {
                if (!produceDiagnostics) {
                    return;
                }
                function getCanonicalOverload(overloads, implementation) {
                    // Consider the canonical set of flags to be the flags of the bodyDeclaration or the first declaration
                    // Error on all deviations from this canonical set of flags
                    // The caveat is that if some overloads are defined in lib.d.ts, we don't want to
                    // report the errors on those. To achieve this, we will say that the implementation is
                    // the canonical signature only if it is in the same container as the first overload
                    var implementationSharesContainerWithFirstOverload = implementation !== undefined && implementation.parent === overloads[0].parent;
                    return implementationSharesContainerWithFirstOverload ? implementation : overloads[0];
                }
                function checkFlagAgreementBetweenOverloads(overloads, implementation, flagsToCheck, someOverloadFlags, allOverloadFlags) {
                    // Error if some overloads have a flag that is not shared by all overloads. To find the
                    // deviations, we XOR someOverloadFlags with allOverloadFlags
                    var someButNotAllOverloadFlags = someOverloadFlags ^ allOverloadFlags;
                    if (someButNotAllOverloadFlags !== 0) {
                        var canonicalFlags_1 = getEffectiveDeclarationFlags(getCanonicalOverload(overloads, implementation), flagsToCheck);
                        ts.forEach(overloads, function (o) {
                            var deviation = getEffectiveDeclarationFlags(o, flagsToCheck) ^ canonicalFlags_1;
                            if (deviation & 1 /* Export */) {
                                error(ts.getNameOfDeclaration(o), ts.Diagnostics.Overload_signatures_must_all_be_exported_or_non_exported);
                            }
                            else if (deviation & 2 /* Ambient */) {
                                error(ts.getNameOfDeclaration(o), ts.Diagnostics.Overload_signatures_must_all_be_ambient_or_non_ambient);
                            }
                            else if (deviation & (8 /* Private */ | 16 /* Protected */)) {
                                error(ts.getNameOfDeclaration(o) || o, ts.Diagnostics.Overload_signatures_must_all_be_public_private_or_protected);
                            }
                            else if (deviation & 128 /* Abstract */) {
                                error(ts.getNameOfDeclaration(o), ts.Diagnostics.Overload_signatures_must_all_be_abstract_or_non_abstract);
                            }
                        });
                    }
                }
                function checkQuestionTokenAgreementBetweenOverloads(overloads, implementation, someHaveQuestionToken, allHaveQuestionToken) {
                    if (someHaveQuestionToken !== allHaveQuestionToken) {
                        var canonicalHasQuestionToken_1 = ts.hasQuestionToken(getCanonicalOverload(overloads, implementation));
                        ts.forEach(overloads, function (o) {
                            var deviation = ts.hasQuestionToken(o) !== canonicalHasQuestionToken_1;
                            if (deviation) {
                                error(ts.getNameOfDeclaration(o), ts.Diagnostics.Overload_signatures_must_all_be_optional_or_required);
                            }
                        });
                    }
                }
                var flagsToCheck = 1 /* Export */ | 2 /* Ambient */ | 8 /* Private */ | 16 /* Protected */ | 128 /* Abstract */;
                var someNodeFlags = 0 /* None */;
                var allNodeFlags = flagsToCheck;
                var someHaveQuestionToken = false;
                var allHaveQuestionToken = true;
                var hasOverloads = false;
                var bodyDeclaration;
                var lastSeenNonAmbientDeclaration;
                var previousDeclaration;
                var declarations = symbol.declarations;
                var isConstructor = (symbol.flags & 16384 /* Constructor */) !== 0;
                function reportImplementationExpectedError(node) {
                    if (node.name && ts.nodeIsMissing(node.name)) {
                        return;
                    }
                    var seen = false;
                    var subsequentNode = ts.forEachChild(node.parent, function (c) {
                        if (seen) {
                            return c;
                        }
                        else {
                            seen = c === node;
                        }
                    });
                    // We may be here because of some extra nodes between overloads that could not be parsed into a valid node.
                    // In this case the subsequent node is not really consecutive (.pos !== node.end), and we must ignore it here.
                    if (subsequentNode && subsequentNode.pos === node.end) {
                        if (subsequentNode.kind === node.kind) {
                            var errorNode_1 = subsequentNode.name || subsequentNode;
                            var subsequentName = subsequentNode.name;
                            if (node.name && subsequentName && (
                            // both are private identifiers
                            ts.isPrivateIdentifier(node.name) && ts.isPrivateIdentifier(subsequentName) && node.name.escapedText === subsequentName.escapedText ||
                                // Both are computed property names
                                // TODO: GH#17345: These are methods, so handle computed name case. (`Always allowing computed property names is *not* the correct behavior!)
                                ts.isComputedPropertyName(node.name) && ts.isComputedPropertyName(subsequentName) ||
                                // Both are literal property names that are the same.
                                ts.isPropertyNameLiteral(node.name) && ts.isPropertyNameLiteral(subsequentName) &&
                                    ts.getEscapedTextOfIdentifierOrLiteral(node.name) === ts.getEscapedTextOfIdentifierOrLiteral(subsequentName))) {
                                var reportError = (node.kind === 167 /* MethodDeclaration */ || node.kind === 166 /* MethodSignature */) &&
                                    ts.isStatic(node) !== ts.isStatic(subsequentNode);
                                // we can get here in two cases
                                // 1. mixed static and instance class members
                                // 2. something with the same name was defined before the set of overloads that prevents them from merging
                                // here we'll report error only for the first case since for second we should already report error in binder
                                if (reportError) {
                                    var diagnostic = ts.isStatic(node) ? ts.Diagnostics.Function_overload_must_be_static : ts.Diagnostics.Function_overload_must_not_be_static;
                                    error(errorNode_1, diagnostic);
                                }
                                return;
                            }
                            if (ts.nodeIsPresent(subsequentNode.body)) {
                                error(errorNode_1, ts.Diagnostics.Function_implementation_name_must_be_0, ts.declarationNameToString(node.name));
                                return;
                            }
                        }
                    }
                    var errorNode = node.name || node;
                    if (isConstructor) {
                        error(errorNode, ts.Diagnostics.Constructor_implementation_is_missing);
                    }
                    else {
                        // Report different errors regarding non-consecutive blocks of declarations depending on whether
                        // the node in question is abstract.
                        if (ts.hasSyntacticModifier(node, 128 /* Abstract */)) {
                            error(errorNode, ts.Diagnostics.All_declarations_of_an_abstract_method_must_be_consecutive);
                        }
                        else {
                            error(errorNode, ts.Diagnostics.Function_implementation_is_missing_or_not_immediately_following_the_declaration);
                        }
                    }
                }
                var duplicateFunctionDeclaration = false;
                var multipleConstructorImplementation = false;
                var hasNonAmbientClass = false;
                var functionDeclarations = [];
                if (declarations) {
                    for (var _i = 0, declarations_4 = declarations; _i < declarations_4.length; _i++) {
                        var current = declarations_4[_i];
                        var node = current;
                        var inAmbientContext = node.flags & 8388608 /* Ambient */;
                        var inAmbientContextOrInterface = node.parent && (node.parent.kind === 256 /* InterfaceDeclaration */ || node.parent.kind === 180 /* TypeLiteral */) || inAmbientContext;
                        if (inAmbientContextOrInterface) {
                            // check if declarations are consecutive only if they are non-ambient
                            // 1. ambient declarations can be interleaved
                            // i.e. this is legal
                            //     declare function foo();
                            //     declare function bar();
                            //     declare function foo();
                            // 2. mixing ambient and non-ambient declarations is a separate error that will be reported - do not want to report an extra one
                            previousDeclaration = undefined;
                        }
                        if ((node.kind === 255 /* ClassDeclaration */ || node.kind === 224 /* ClassExpression */) && !inAmbientContext) {
                            hasNonAmbientClass = true;
                        }
                        if (node.kind === 254 /* FunctionDeclaration */ || node.kind === 167 /* MethodDeclaration */ || node.kind === 166 /* MethodSignature */ || node.kind === 169 /* Constructor */) {
                            functionDeclarations.push(node);
                            var currentNodeFlags = getEffectiveDeclarationFlags(node, flagsToCheck);
                            someNodeFlags |= currentNodeFlags;
                            allNodeFlags &= currentNodeFlags;
                            someHaveQuestionToken = someHaveQuestionToken || ts.hasQuestionToken(node);
                            allHaveQuestionToken = allHaveQuestionToken && ts.hasQuestionToken(node);
                            var bodyIsPresent = ts.nodeIsPresent(node.body);
                            if (bodyIsPresent && bodyDeclaration) {
                                if (isConstructor) {
                                    multipleConstructorImplementation = true;
                                }
                                else {
                                    duplicateFunctionDeclaration = true;
                                }
                            }
                            else if ((previousDeclaration === null || previousDeclaration === void 0 ? void 0 : previousDeclaration.parent) === node.parent && previousDeclaration.end !== node.pos) {
                                reportImplementationExpectedError(previousDeclaration);
                            }
                            if (bodyIsPresent) {
                                if (!bodyDeclaration) {
                                    bodyDeclaration = node;
                                }
                            }
                            else {
                                hasOverloads = true;
                            }
                            previousDeclaration = node;
                            if (!inAmbientContextOrInterface) {
                                lastSeenNonAmbientDeclaration = node;
                            }
                        }
                    }
                }
                if (multipleConstructorImplementation) {
                    ts.forEach(functionDeclarations, function (declaration) {
                        error(declaration, ts.Diagnostics.Multiple_constructor_implementations_are_not_allowed);
                    });
                }
                if (duplicateFunctionDeclaration) {
                    ts.forEach(functionDeclarations, function (declaration) {
                        error(ts.getNameOfDeclaration(declaration) || declaration, ts.Diagnostics.Duplicate_function_implementation);
                    });
                }
                if (hasNonAmbientClass && !isConstructor && symbol.flags & 16 /* Function */ && declarations) {
                    var relatedDiagnostics_1 = ts.filter(declarations, function (d) { return d.kind === 255 /* ClassDeclaration */; })
                        .map(function (d) { return ts.createDiagnosticForNode(d, ts.Diagnostics.Consider_adding_a_declare_modifier_to_this_class); });
                    ts.forEach(declarations, function (declaration) {
                        var diagnostic = declaration.kind === 255 /* ClassDeclaration */
                            ? ts.Diagnostics.Class_declaration_cannot_implement_overload_list_for_0
                            : declaration.kind === 254 /* FunctionDeclaration */
                                ? ts.Diagnostics.Function_with_bodies_can_only_merge_with_classes_that_are_ambient
                                : undefined;
                        if (diagnostic) {
                            ts.addRelatedInfo.apply(void 0, __spreadArray([error(ts.getNameOfDeclaration(declaration) || declaration, diagnostic, ts.symbolName(symbol))], relatedDiagnostics_1, false));
                        }
                    });
                }
                // Abstract methods can't have an implementation -- in particular, they don't need one.
                if (lastSeenNonAmbientDeclaration && !lastSeenNonAmbientDeclaration.body &&
                    !ts.hasSyntacticModifier(lastSeenNonAmbientDeclaration, 128 /* Abstract */) && !lastSeenNonAmbientDeclaration.questionToken) {
                    reportImplementationExpectedError(lastSeenNonAmbientDeclaration);
                }
                if (hasOverloads) {
                    if (declarations) {
                        checkFlagAgreementBetweenOverloads(declarations, bodyDeclaration, flagsToCheck, someNodeFlags, allNodeFlags);
                        checkQuestionTokenAgreementBetweenOverloads(declarations, bodyDeclaration, someHaveQuestionToken, allHaveQuestionToken);
                    }
                    if (bodyDeclaration) {
                        var signatures = getSignaturesOfSymbol(symbol);
                        var bodySignature = getSignatureFromDeclaration(bodyDeclaration);
                        for (var _a = 0, signatures_10 = signatures; _a < signatures_10.length; _a++) {
                            var signature = signatures_10[_a];
                            if (!isImplementationCompatibleWithOverload(bodySignature, signature)) {
                                ts.addRelatedInfo(error(signature.declaration, ts.Diagnostics.This_overload_signature_is_not_compatible_with_its_implementation_signature), ts.createDiagnosticForNode(bodyDeclaration, ts.Diagnostics.The_implementation_signature_is_declared_here));
                                break;
                            }
                        }
                    }
                }
            }
            function checkExportsOnMergedDeclarations(node) {
                if (!produceDiagnostics) {
                    return;
                }
                // if localSymbol is defined on node then node itself is exported - check is required
                var symbol = node.localSymbol;
                if (!symbol) {
                    // local symbol is undefined => this declaration is non-exported.
                    // however symbol might contain other declarations that are exported
                    symbol = getSymbolOfNode(node);
                    if (!symbol.exportSymbol) {
                        // this is a pure local symbol (all declarations are non-exported) - no need to check anything
                        return;
                    }
                }
                // run the check only for the first declaration in the list
                if (ts.getDeclarationOfKind(symbol, node.kind) !== node) {
                    return;
                }
                var exportedDeclarationSpaces = 0 /* None */;
                var nonExportedDeclarationSpaces = 0 /* None */;
                var defaultExportedDeclarationSpaces = 0 /* None */;
                for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                    var d = _a[_i];
                    var declarationSpaces = getDeclarationSpaces(d);
                    var effectiveDeclarationFlags = getEffectiveDeclarationFlags(d, 1 /* Export */ | 512 /* Default */);
                    if (effectiveDeclarationFlags & 1 /* Export */) {
                        if (effectiveDeclarationFlags & 512 /* Default */) {
                            defaultExportedDeclarationSpaces |= declarationSpaces;
                        }
                        else {
                            exportedDeclarationSpaces |= declarationSpaces;
                        }
                    }
                    else {
                        nonExportedDeclarationSpaces |= declarationSpaces;
                    }
                }
                // Spaces for anything not declared a 'default export'.
                var nonDefaultExportedDeclarationSpaces = exportedDeclarationSpaces | nonExportedDeclarationSpaces;
                var commonDeclarationSpacesForExportsAndLocals = exportedDeclarationSpaces & nonExportedDeclarationSpaces;
                var commonDeclarationSpacesForDefaultAndNonDefault = defaultExportedDeclarationSpaces & nonDefaultExportedDeclarationSpaces;
                if (commonDeclarationSpacesForExportsAndLocals || commonDeclarationSpacesForDefaultAndNonDefault) {
                    // declaration spaces for exported and non-exported declarations intersect
                    for (var _b = 0, _c = symbol.declarations; _b < _c.length; _b++) {
                        var d = _c[_b];
                        var declarationSpaces = getDeclarationSpaces(d);
                        var name = ts.getNameOfDeclaration(d);
                        // Only error on the declarations that contributed to the intersecting spaces.
                        if (declarationSpaces & commonDeclarationSpacesForDefaultAndNonDefault) {
                            error(name, ts.Diagnostics.Merged_declaration_0_cannot_include_a_default_export_declaration_Consider_adding_a_separate_export_default_0_declaration_instead, ts.declarationNameToString(name));
                        }
                        else if (declarationSpaces & commonDeclarationSpacesForExportsAndLocals) {
                            error(name, ts.Diagnostics.Individual_declarations_in_merged_declaration_0_must_be_all_exported_or_all_local, ts.declarationNameToString(name));
                        }
                    }
                }
                function getDeclarationSpaces(decl) {
                    var d = decl;
                    switch (d.kind) {
                        case 256 /* InterfaceDeclaration */:
                        case 257 /* TypeAliasDeclaration */:
                        // A jsdoc typedef and callback are, by definition, type aliases.
                        // falls through
                        case 340 /* JSDocTypedefTag */:
                        case 333 /* JSDocCallbackTag */:
                        case 334 /* JSDocEnumTag */:
                            return 2 /* ExportType */;
                        case 259 /* ModuleDeclaration */:
                            return ts.isAmbientModule(d) || ts.getModuleInstanceState(d) !== 0 /* NonInstantiated */
                                ? 4 /* ExportNamespace */ | 1 /* ExportValue */
                                : 4 /* ExportNamespace */;
                        case 255 /* ClassDeclaration */:
                        case 258 /* EnumDeclaration */:
                        case 294 /* EnumMember */:
                            return 2 /* ExportType */ | 1 /* ExportValue */;
                        case 300 /* SourceFile */:
                            return 2 /* ExportType */ | 1 /* ExportValue */ | 4 /* ExportNamespace */;
                        case 269 /* ExportAssignment */:
                        case 219 /* BinaryExpression */:
                            var node_2 = d;
                            var expression = ts.isExportAssignment(node_2) ? node_2.expression : node_2.right;
                            // Export assigned entity name expressions act as aliases and should fall through, otherwise they export values
                            if (!ts.isEntityNameExpression(expression)) {
                                return 1 /* ExportValue */;
                            }
                            d = expression;
                        // The below options all declare an Alias, which is allowed to merge with other values within the importing module.
                        // falls through
                        case 263 /* ImportEqualsDeclaration */:
                        case 266 /* NamespaceImport */:
                        case 265 /* ImportClause */:
                            var result_12 = 0 /* None */;
                            var target = resolveAlias(getSymbolOfNode(d));
                            ts.forEach(target.declarations, function (d) { result_12 |= getDeclarationSpaces(d); });
                            return result_12;
                        case 252 /* VariableDeclaration */:
                        case 201 /* BindingElement */:
                        case 254 /* FunctionDeclaration */:
                        case 268 /* ImportSpecifier */: // https://github.com/Microsoft/TypeScript/pull/7591
                        case 79 /* Identifier */: // https://github.com/microsoft/TypeScript/issues/36098
                            // Identifiers are used as declarations of assignment declarations whose parents may be
                            // SyntaxKind.CallExpression - `Object.defineProperty(thing, "aField", {value: 42});`
                            // SyntaxKind.ElementAccessExpression - `thing["aField"] = 42;` or `thing["aField"];` (with a doc comment on it)
                            // or SyntaxKind.PropertyAccessExpression - `thing.aField = 42;`
                            // all of which are pretty much always values, or at least imply a value meaning.
                            // It may be apprpriate to treat these as aliases in the future.
                            return 1 /* ExportValue */;
                        default:
                            return ts.Debug.failBadSyntaxKind(d);
                    }
                }
            }
            function getAwaitedTypeOfPromise(type, errorNode, diagnosticMessage, arg0) {
                var promisedType = getPromisedTypeOfPromise(type, errorNode);
                return promisedType && getAwaitedType(promisedType, errorNode, diagnosticMessage, arg0);
            }
            /**
             * Gets the "promised type" of a promise.
             * @param type The type of the promise.
             * @remarks The "promised type" of a type is the type of the "value" parameter of the "onfulfilled" callback.
             */
            function getPromisedTypeOfPromise(type, errorNode) {
                //
                //  { // type
                //      then( // thenFunction
                //          onfulfilled: ( // onfulfilledParameterType
                //              value: T // valueParameterType
                //          ) => any
                //      ): any;
                //  }
                //
                if (isTypeAny(type)) {
                    return undefined;
                }
                var typeAsPromise = type;
                if (typeAsPromise.promisedTypeOfPromise) {
                    return typeAsPromise.promisedTypeOfPromise;
                }
                if (isReferenceToType(type, getGlobalPromiseType(/*reportErrors*/ false))) {
                    return typeAsPromise.promisedTypeOfPromise = getTypeArguments(type)[0];
                }
                var thenFunction = getTypeOfPropertyOfType(type, "then"); // TODO: GH#18217
                if (isTypeAny(thenFunction)) {
                    return undefined;
                }
                var thenSignatures = thenFunction ? getSignaturesOfType(thenFunction, 0 /* Call */) : ts.emptyArray;
                if (thenSignatures.length === 0) {
                    if (errorNode) {
                        error(errorNode, ts.Diagnostics.A_promise_must_have_a_then_method);
                    }
                    return undefined;
                }
                var onfulfilledParameterType = getTypeWithFacts(getUnionType(ts.map(thenSignatures, getTypeOfFirstParameterOfSignature)), 2097152 /* NEUndefinedOrNull */);
                if (isTypeAny(onfulfilledParameterType)) {
                    return undefined;
                }
                var onfulfilledParameterSignatures = getSignaturesOfType(onfulfilledParameterType, 0 /* Call */);
                if (onfulfilledParameterSignatures.length === 0) {
                    if (errorNode) {
                        error(errorNode, ts.Diagnostics.The_first_parameter_of_the_then_method_of_a_promise_must_be_a_callback);
                    }
                    return undefined;
                }
                return typeAsPromise.promisedTypeOfPromise = getUnionType(ts.map(onfulfilledParameterSignatures, getTypeOfFirstParameterOfSignature), 2 /* Subtype */);
            }
            /**
             * Gets the "awaited type" of a type.
             * @param type The type to await.
             * @remarks The "awaited type" of an expression is its "promised type" if the expression is a
             * Promise-like type; otherwise, it is the type of the expression. This is used to reflect
             * The runtime behavior of the `await` keyword.
             */
            function checkAwaitedType(type, errorNode, diagnosticMessage, arg0) {
                var awaitedType = getAwaitedType(type, errorNode, diagnosticMessage, arg0);
                return awaitedType || errorType;
            }
            /**
             * Determines whether a type has a callable `then` member.
             */
            function isThenableType(type) {
                var thenFunction = getTypeOfPropertyOfType(type, "then");
                return !!thenFunction && getSignaturesOfType(getTypeWithFacts(thenFunction, 2097152 /* NEUndefinedOrNull */), 0 /* Call */).length > 0;
            }
            /**
             * Gets the "awaited type" of a type.
             *
             * The "awaited type" of an expression is its "promised type" if the expression is a
             * Promise-like type; otherwise, it is the type of the expression. If the "promised
             * type" is itself a Promise-like, the "promised type" is recursively unwrapped until a
             * non-promise type is found.
             *
             * This is used to reflect the runtime behavior of the `await` keyword.
             */
            function getAwaitedType(type, errorNode, diagnosticMessage, arg0) {
                if (isTypeAny(type)) {
                    return type;
                }
                var typeAsAwaitable = type;
                if (typeAsAwaitable.awaitedTypeOfType) {
                    return typeAsAwaitable.awaitedTypeOfType;
                }
                // For a union, get a union of the awaited types of each constituent.
                //
                return typeAsAwaitable.awaitedTypeOfType =
                    mapType(type, errorNode ? function (constituentType) { return getAwaitedTypeWorker(constituentType, errorNode, diagnosticMessage, arg0); } : getAwaitedTypeWorker);
            }
            function getAwaitedTypeWorker(type, errorNode, diagnosticMessage, arg0) {
                var typeAsAwaitable = type;
                if (typeAsAwaitable.awaitedTypeOfType) {
                    return typeAsAwaitable.awaitedTypeOfType;
                }
                var promisedType = getPromisedTypeOfPromise(type);
                if (promisedType) {
                    if (type.id === promisedType.id || awaitedTypeStack.lastIndexOf(promisedType.id) >= 0) {
                        // Verify that we don't have a bad actor in the form of a promise whose
                        // promised type is the same as the promise type, or a mutually recursive
                        // promise. If so, we return undefined as we cannot guess the shape. If this
                        // were the actual case in the JavaScript, this Promise would never resolve.
                        //
                        // An example of a bad actor with a singly-recursive promise type might
                        // be:
                        //
                        //  interface BadPromise {
                        //      then(
                        //          onfulfilled: (value: BadPromise) => any,
                        //          onrejected: (error: any) => any): BadPromise;
                        //  }
                        //
                        // The above interface will pass the PromiseLike check, and return a
                        // promised type of `BadPromise`. Since this is a self reference, we
                        // don't want to keep recursing ad infinitum.
                        //
                        // An example of a bad actor in the form of a mutually-recursive
                        // promise type might be:
                        //
                        //  interface BadPromiseA {
                        //      then(
                        //          onfulfilled: (value: BadPromiseB) => any,
                        //          onrejected: (error: any) => any): BadPromiseB;
                        //  }
                        //
                        //  interface BadPromiseB {
                        //      then(
                        //          onfulfilled: (value: BadPromiseA) => any,
                        //          onrejected: (error: any) => any): BadPromiseA;
                        //  }
                        //
                        if (errorNode) {
                            error(errorNode, ts.Diagnostics.Type_is_referenced_directly_or_indirectly_in_the_fulfillment_callback_of_its_own_then_method);
                        }
                        return undefined;
                    }
                    // Keep track of the type we're about to unwrap to avoid bad recursive promise types.
                    // See the comments above for more information.
                    awaitedTypeStack.push(type.id);
                    var awaitedType = getAwaitedType(promisedType, errorNode, diagnosticMessage, arg0);
                    awaitedTypeStack.pop();
                    if (!awaitedType) {
                        return undefined;
                    }
                    return typeAsAwaitable.awaitedTypeOfType = awaitedType;
                }
                // The type was not a promise, so it could not be unwrapped any further.
                // As long as the type does not have a callable "then" property, it is
                // safe to return the type; otherwise, an error is reported and we return
                // undefined.
                //
                // An example of a non-promise "thenable" might be:
                //
                //  await { then(): void {} }
                //
                // The "thenable" does not match the minimal definition for a promise. When
                // a Promise/A+-compatible or ES6 promise tries to adopt this value, the promise
                // will never settle. We treat this as an error to help flag an early indicator
                // of a runtime problem. If the user wants to return this value from an async
                // function, they would need to wrap it in some other value. If they want it to
                // be treated as a promise, they can cast to <any>.
                if (isThenableType(type)) {
                    if (errorNode) {
                        if (!diagnosticMessage)
                            return ts.Debug.fail();
                        error(errorNode, diagnosticMessage, arg0);
                    }
                    return undefined;
                }
                return typeAsAwaitable.awaitedTypeOfType = type;
            }
            /**
             * Checks the return type of an async function to ensure it is a compatible
             * Promise implementation.
             *
             * This checks that an async function has a valid Promise-compatible return type.
             * An async function has a valid Promise-compatible return type if the resolved value
             * of the return type has a construct signature that takes in an `initializer` function
             * that in turn supplies a `resolve` function as one of its arguments and results in an
             * object with a callable `then` signature.
             *
             * @param node The signature to check
             */
            function checkAsyncFunctionReturnType(node, returnTypeNode) {
                // As part of our emit for an async function, we will need to emit the entity name of
                // the return type annotation as an expression. To meet the necessary runtime semantics
                // for __awaiter, we must also check that the type of the declaration (e.g. the static
                // side or "constructor" of the promise type) is compatible `PromiseConstructorLike`.
                //
                // An example might be (from lib.es6.d.ts):
                //
                //  interface Promise<T> { ... }
                //  interface PromiseConstructor {
                //      new <T>(...): Promise<T>;
                //  }
                //  declare var Promise: PromiseConstructor;
                //
                // When an async function declares a return type annotation of `Promise<T>`, we
                // need to get the type of the `Promise` variable declaration above, which would
                // be `PromiseConstructor`.
                //
                // The same case applies to a class:
                //
                //  declare class Promise<T> {
                //      constructor(...);
                //      then<U>(...): Promise<U>;
                //  }
                //
                var returnType = getTypeFromTypeNode(returnTypeNode);
                if (languageVersion >= 2 /* ES2015 */) {
                    if (returnType === errorType) {
                        return;
                    }
                    var globalPromiseType = getGlobalPromiseType(/*reportErrors*/ true);
                    if (globalPromiseType !== emptyGenericType && !isReferenceToType(returnType, globalPromiseType)) {
                        // The promise type was not a valid type reference to the global promise type, so we
                        // report an error and return the unknown type.
                        error(returnTypeNode, ts.Diagnostics.The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type_Did_you_mean_to_write_Promise_0, typeToString(getAwaitedType(returnType) || voidType));
                        return;
                    }
                }
                else {
                    // Always mark the type node as referenced if it points to a value
                    markTypeNodeAsReferenced(returnTypeNode);
                    if (returnType === errorType) {
                        return;
                    }
                    var promiseConstructorName = ts.getEntityNameFromTypeNode(returnTypeNode);
                    if (promiseConstructorName === undefined) {
                        error(returnTypeNode, ts.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value, typeToString(returnType));
                        return;
                    }
                    var promiseConstructorSymbol = resolveEntityName(promiseConstructorName, 111551 /* Value */, /*ignoreErrors*/ true);
                    var promiseConstructorType = promiseConstructorSymbol ? getTypeOfSymbol(promiseConstructorSymbol) : errorType;
                    if (promiseConstructorType === errorType) {
                        if (promiseConstructorName.kind === 79 /* Identifier */ && promiseConstructorName.escapedText === "Promise" && getTargetType(returnType) === getGlobalPromiseType(/*reportErrors*/ false)) {
                            error(returnTypeNode, ts.Diagnostics.An_async_function_or_method_in_ES5_SlashES3_requires_the_Promise_constructor_Make_sure_you_have_a_declaration_for_the_Promise_constructor_or_include_ES2015_in_your_lib_option);
                        }
                        else {
                            error(returnTypeNode, ts.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value, ts.entityNameToString(promiseConstructorName));
                        }
                        return;
                    }
                    var globalPromiseConstructorLikeType = getGlobalPromiseConstructorLikeType(/*reportErrors*/ true);
                    if (globalPromiseConstructorLikeType === emptyObjectType) {
                        // If we couldn't resolve the global PromiseConstructorLike type we cannot verify
                        // compatibility with __awaiter.
                        error(returnTypeNode, ts.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value, ts.entityNameToString(promiseConstructorName));
                        return;
                    }
                    if (!checkTypeAssignableTo(promiseConstructorType, globalPromiseConstructorLikeType, returnTypeNode, ts.Diagnostics.Type_0_is_not_a_valid_async_function_return_type_in_ES5_SlashES3_because_it_does_not_refer_to_a_Promise_compatible_constructor_value)) {
                        return;
                    }
                    // Verify there is no local declaration that could collide with the promise constructor.
                    var rootName = promiseConstructorName && ts.getFirstIdentifier(promiseConstructorName);
                    var collidingSymbol = getSymbol(node.locals, rootName.escapedText, 111551 /* Value */);
                    if (collidingSymbol) {
                        error(collidingSymbol.valueDeclaration, ts.Diagnostics.Duplicate_identifier_0_Compiler_uses_declaration_1_to_support_async_functions, ts.idText(rootName), ts.entityNameToString(promiseConstructorName));
                        return;
                    }
                }
                checkAwaitedType(returnType, node, ts.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member);
            }
            /** Check a decorator */
            function checkDecorator(node) {
                var signature = getResolvedSignature(node);
                checkDeprecatedSignature(signature, node);
                var returnType = getReturnTypeOfSignature(signature);
                if (returnType.flags & 1 /* Any */) {
                    return;
                }
                var expectedReturnType;
                var headMessage = getDiagnosticHeadMessageForDecoratorResolution(node);
                var errorInfo;
                switch (node.parent.kind) {
                    case 255 /* ClassDeclaration */:
                        var classSymbol = getSymbolOfNode(node.parent);
                        var classConstructorType = getTypeOfSymbol(classSymbol);
                        expectedReturnType = getUnionType([classConstructorType, voidType]);
                        break;
                    case 162 /* Parameter */:
                        expectedReturnType = voidType;
                        errorInfo = ts.chainDiagnosticMessages(
                        /*details*/ undefined, ts.Diagnostics.The_return_type_of_a_parameter_decorator_function_must_be_either_void_or_any);
                        break;
                    case 165 /* PropertyDeclaration */:
                        expectedReturnType = voidType;
                        errorInfo = ts.chainDiagnosticMessages(
                        /*details*/ undefined, ts.Diagnostics.The_return_type_of_a_property_decorator_function_must_be_either_void_or_any);
                        break;
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        var methodType = getTypeOfNode(node.parent);
                        var descriptorType = createTypedPropertyDescriptorType(methodType);
                        expectedReturnType = getUnionType([descriptorType, voidType]);
                        break;
                    default:
                        return ts.Debug.fail();
                }
                checkTypeAssignableTo(returnType, expectedReturnType, node, headMessage, function () { return errorInfo; });
            }
            /**
             * If a TypeNode can be resolved to a value symbol imported from an external module, it is
             * marked as referenced to prevent import elision.
             */
            function markTypeNodeAsReferenced(node) {
                markEntityNameOrEntityExpressionAsReference(node && ts.getEntityNameFromTypeNode(node));
            }
            function markEntityNameOrEntityExpressionAsReference(typeName) {
                if (!typeName)
                    return;
                var rootName = ts.getFirstIdentifier(typeName);
                var meaning = (typeName.kind === 79 /* Identifier */ ? 788968 /* Type */ : 1920 /* Namespace */) | 2097152 /* Alias */;
                var rootSymbol = resolveName(rootName, rootName.escapedText, meaning, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isReference*/ true);
                if (rootSymbol
                    && rootSymbol.flags & 2097152 /* Alias */
                    && symbolIsValue(rootSymbol)
                    && !isConstEnumOrConstEnumOnlyModule(resolveAlias(rootSymbol))
                    && !getTypeOnlyAliasDeclaration(rootSymbol)) {
                    markAliasSymbolAsReferenced(rootSymbol);
                }
            }
            /**
             * This function marks the type used for metadata decorator as referenced if it is import
             * from external module.
             * This is different from markTypeNodeAsReferenced because it tries to simplify type nodes in
             * union and intersection type
             * @param node
             */
            function markDecoratorMedataDataTypeNodeAsReferenced(node) {
                var entityName = getEntityNameForDecoratorMetadata(node);
                if (entityName && ts.isEntityName(entityName)) {
                    markEntityNameOrEntityExpressionAsReference(entityName);
                }
            }
            function getEntityNameForDecoratorMetadata(node) {
                if (node) {
                    switch (node.kind) {
                        case 186 /* IntersectionType */:
                        case 185 /* UnionType */:
                            return getEntityNameForDecoratorMetadataFromTypeList(node.types);
                        case 187 /* ConditionalType */:
                            return getEntityNameForDecoratorMetadataFromTypeList([node.trueType, node.falseType]);
                        case 189 /* ParenthesizedType */:
                        case 195 /* NamedTupleMember */:
                            return getEntityNameForDecoratorMetadata(node.type);
                        case 176 /* TypeReference */:
                            return node.typeName;
                    }
                }
            }
            function getEntityNameForDecoratorMetadataFromTypeList(types) {
                var commonEntityName;
                for (var _i = 0, types_22 = types; _i < types_22.length; _i++) {
                    var typeNode = types_22[_i];
                    while (typeNode.kind === 189 /* ParenthesizedType */ || typeNode.kind === 195 /* NamedTupleMember */) {
                        typeNode = typeNode.type; // Skip parens if need be
                    }
                    if (typeNode.kind === 142 /* NeverKeyword */) {
                        continue; // Always elide `never` from the union/intersection if possible
                    }
                    if (!strictNullChecks && (typeNode.kind === 194 /* LiteralType */ && typeNode.literal.kind === 104 /* NullKeyword */ || typeNode.kind === 151 /* UndefinedKeyword */)) {
                        continue; // Elide null and undefined from unions for metadata, just like what we did prior to the implementation of strict null checks
                    }
                    var individualEntityName = getEntityNameForDecoratorMetadata(typeNode);
                    if (!individualEntityName) {
                        // Individual is something like string number
                        // So it would be serialized to either that type or object
                        // Safe to return here
                        return undefined;
                    }
                    if (commonEntityName) {
                        // Note this is in sync with the transformation that happens for type node.
                        // Keep this in sync with serializeUnionOrIntersectionType
                        // Verify if they refer to same entity and is identifier
                        // return undefined if they dont match because we would emit object
                        if (!ts.isIdentifier(commonEntityName) ||
                            !ts.isIdentifier(individualEntityName) ||
                            commonEntityName.escapedText !== individualEntityName.escapedText) {
                            return undefined;
                        }
                    }
                    else {
                        commonEntityName = individualEntityName;
                    }
                }
                return commonEntityName;
            }
            function getParameterTypeNodeForDecoratorCheck(node) {
                var typeNode = ts.getEffectiveTypeAnnotationNode(node);
                return ts.isRestParameter(node) ? ts.getRestParameterElementType(typeNode) : typeNode;
            }
            /** Check the decorators of a node */
            function checkDecorators(node) {
                if (!node.decorators) {
                    return;
                }
                // skip this check for nodes that cannot have decorators. These should have already had an error reported by
                // checkGrammarDecorators.
                if (!ts.nodeCanBeDecorated(node, node.parent, node.parent.parent)) {
                    return;
                }
                if (!compilerOptions.experimentalDecorators) {
                    error(node, ts.Diagnostics.Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_in_your_tsconfig_or_jsconfig_to_remove_this_warning);
                }
                var firstDecorator = node.decorators[0];
                checkExternalEmitHelpers(firstDecorator, 8 /* Decorate */);
                if (node.kind === 162 /* Parameter */) {
                    checkExternalEmitHelpers(firstDecorator, 32 /* Param */);
                }
                if (compilerOptions.emitDecoratorMetadata) {
                    checkExternalEmitHelpers(firstDecorator, 16 /* Metadata */);
                    // we only need to perform these checks if we are emitting serialized type metadata for the target of a decorator.
                    switch (node.kind) {
                        case 255 /* ClassDeclaration */:
                            var constructor = ts.getFirstConstructorWithBody(node);
                            if (constructor) {
                                for (var _i = 0, _a = constructor.parameters; _i < _a.length; _i++) {
                                    var parameter = _a[_i];
                                    markDecoratorMedataDataTypeNodeAsReferenced(getParameterTypeNodeForDecoratorCheck(parameter));
                                }
                            }
                            break;
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            var otherKind = node.kind === 170 /* GetAccessor */ ? 171 /* SetAccessor */ : 170 /* GetAccessor */;
                            var otherAccessor = ts.getDeclarationOfKind(getSymbolOfNode(node), otherKind);
                            markDecoratorMedataDataTypeNodeAsReferenced(getAnnotatedAccessorTypeNode(node) || otherAccessor && getAnnotatedAccessorTypeNode(otherAccessor));
                            break;
                        case 167 /* MethodDeclaration */:
                            for (var _b = 0, _c = node.parameters; _b < _c.length; _b++) {
                                var parameter = _c[_b];
                                markDecoratorMedataDataTypeNodeAsReferenced(getParameterTypeNodeForDecoratorCheck(parameter));
                            }
                            markDecoratorMedataDataTypeNodeAsReferenced(ts.getEffectiveReturnTypeNode(node));
                            break;
                        case 165 /* PropertyDeclaration */:
                            markDecoratorMedataDataTypeNodeAsReferenced(ts.getEffectiveTypeAnnotationNode(node));
                            break;
                        case 162 /* Parameter */:
                            markDecoratorMedataDataTypeNodeAsReferenced(getParameterTypeNodeForDecoratorCheck(node));
                            var containingSignature = node.parent;
                            for (var _d = 0, _e = containingSignature.parameters; _d < _e.length; _d++) {
                                var parameter = _e[_d];
                                markDecoratorMedataDataTypeNodeAsReferenced(getParameterTypeNodeForDecoratorCheck(parameter));
                            }
                            break;
                    }
                }
                ts.forEach(node.decorators, checkDecorator);
            }
            function checkFunctionDeclaration(node) {
                if (produceDiagnostics) {
                    checkFunctionOrMethodDeclaration(node);
                    checkGrammarForGenerator(node);
                    checkCollisionsForDeclarationName(node, node.name);
                }
            }
            function checkJSDocTypeAliasTag(node) {
                if (!node.typeExpression) {
                    // If the node had `@property` tags, `typeExpression` would have been set to the first property tag.
                    error(node.name, ts.Diagnostics.JSDoc_typedef_tag_should_either_have_a_type_annotation_or_be_followed_by_property_or_member_tags);
                }
                if (node.name) {
                    checkTypeNameIsReserved(node.name, ts.Diagnostics.Type_alias_name_cannot_be_0);
                }
                checkSourceElement(node.typeExpression);
            }
            function checkJSDocTemplateTag(node) {
                checkSourceElement(node.constraint);
                for (var _i = 0, _a = node.typeParameters; _i < _a.length; _i++) {
                    var tp = _a[_i];
                    checkSourceElement(tp);
                }
            }
            function checkJSDocTypeTag(node) {
                checkSourceElement(node.typeExpression);
            }
            function checkJSDocParameterTag(node) {
                checkSourceElement(node.typeExpression);
                if (!ts.getParameterSymbolFromJSDoc(node)) {
                    var decl = ts.getHostSignatureFromJSDoc(node);
                    // don't issue an error for invalid hosts -- just functions --
                    // and give a better error message when the host function mentions `arguments`
                    // but the tag doesn't have an array type
                    if (decl) {
                        var i = ts.getJSDocTags(decl).filter(ts.isJSDocParameterTag).indexOf(node);
                        if (i > -1 && i < decl.parameters.length && ts.isBindingPattern(decl.parameters[i].name)) {
                            return;
                        }
                        if (!containsArgumentsReference(decl)) {
                            if (ts.isQualifiedName(node.name)) {
                                error(node.name, ts.Diagnostics.Qualified_name_0_is_not_allowed_without_a_leading_param_object_1, ts.entityNameToString(node.name), ts.entityNameToString(node.name.left));
                            }
                            else {
                                error(node.name, ts.Diagnostics.JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name, ts.idText(node.name));
                            }
                        }
                        else if (ts.findLast(ts.getJSDocTags(decl), ts.isJSDocParameterTag) === node &&
                            node.typeExpression && node.typeExpression.type &&
                            !isArrayType(getTypeFromTypeNode(node.typeExpression.type))) {
                            error(node.name, ts.Diagnostics.JSDoc_param_tag_has_name_0_but_there_is_no_parameter_with_that_name_It_would_match_arguments_if_it_had_an_array_type, ts.idText(node.name.kind === 159 /* QualifiedName */ ? node.name.right : node.name));
                        }
                    }
                }
            }
            function checkJSDocPropertyTag(node) {
                checkSourceElement(node.typeExpression);
            }
            function checkJSDocFunctionType(node) {
                if (produceDiagnostics && !node.type && !ts.isJSDocConstructSignature(node)) {
                    reportImplicitAny(node, anyType);
                }
                checkSignatureDeclaration(node);
            }
            function checkJSDocImplementsTag(node) {
                var classLike = ts.getEffectiveJSDocHost(node);
                if (!classLike || !ts.isClassDeclaration(classLike) && !ts.isClassExpression(classLike)) {
                    error(classLike, ts.Diagnostics.JSDoc_0_is_not_attached_to_a_class, ts.idText(node.tagName));
                }
            }
            function checkJSDocAugmentsTag(node) {
                var classLike = ts.getEffectiveJSDocHost(node);
                if (!classLike || !ts.isClassDeclaration(classLike) && !ts.isClassExpression(classLike)) {
                    error(classLike, ts.Diagnostics.JSDoc_0_is_not_attached_to_a_class, ts.idText(node.tagName));
                    return;
                }
                var augmentsTags = ts.getJSDocTags(classLike).filter(ts.isJSDocAugmentsTag);
                ts.Debug.assert(augmentsTags.length > 0);
                if (augmentsTags.length > 1) {
                    error(augmentsTags[1], ts.Diagnostics.Class_declarations_cannot_have_more_than_one_augments_or_extends_tag);
                }
                var name = getIdentifierFromEntityNameExpression(node.class.expression);
                var extend = ts.getClassExtendsHeritageElement(classLike);
                if (extend) {
                    var className = getIdentifierFromEntityNameExpression(extend.expression);
                    if (className && name.escapedText !== className.escapedText) {
                        error(name, ts.Diagnostics.JSDoc_0_1_does_not_match_the_extends_2_clause, ts.idText(node.tagName), ts.idText(name), ts.idText(className));
                    }
                }
            }
            function getIdentifierFromEntityNameExpression(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        return node;
                    case 204 /* PropertyAccessExpression */:
                        return node.name;
                    default:
                        return undefined;
                }
            }
            function checkFunctionOrMethodDeclaration(node) {
                var _a;
                checkDecorators(node);
                checkSignatureDeclaration(node);
                var functionFlags = ts.getFunctionFlags(node);
                // Do not use hasDynamicName here, because that returns false for well known symbols.
                // We want to perform checkComputedPropertyName for all computed properties, including
                // well known symbols.
                if (node.name && node.name.kind === 160 /* ComputedPropertyName */) {
                    // This check will account for methods in class/interface declarations,
                    // as well as accessors in classes/object literals
                    checkComputedPropertyName(node.name);
                }
                if (hasBindableName(node)) {
                    // first we want to check the local symbol that contain this declaration
                    // - if node.localSymbol !== undefined - this is current declaration is exported and localSymbol points to the local symbol
                    // - if node.localSymbol === undefined - this node is non-exported so we can just pick the result of getSymbolOfNode
                    var symbol = getSymbolOfNode(node);
                    var localSymbol = node.localSymbol || symbol;
                    // Since the javascript won't do semantic analysis like typescript,
                    // if the javascript file comes before the typescript file and both contain same name functions,
                    // checkFunctionOrConstructorSymbol wouldn't be called if we didnt ignore javascript function.
                    var firstDeclaration = (_a = localSymbol.declarations) === null || _a === void 0 ? void 0 : _a.find(
                    // Get first non javascript function declaration
                    function (declaration) { return declaration.kind === node.kind && !(declaration.flags & 131072 /* JavaScriptFile */); });
                    // Only type check the symbol once
                    if (node === firstDeclaration) {
                        checkFunctionOrConstructorSymbol(localSymbol);
                    }
                    if (symbol.parent) {
                        // run check on export symbol to check that modifiers agree across all exported declarations
                        checkFunctionOrConstructorSymbol(symbol);
                    }
                }
                var body = node.kind === 166 /* MethodSignature */ ? undefined : node.body;
                checkSourceElement(body);
                checkAllCodePathsInNonVoidFunctionReturnOrThrow(node, getReturnTypeFromAnnotation(node));
                if (produceDiagnostics && !ts.getEffectiveReturnTypeNode(node)) {
                    // Report an implicit any error if there is no body, no explicit return type, and node is not a private method
                    // in an ambient context
                    if (ts.nodeIsMissing(body) && !isPrivateWithinAmbient(node)) {
                        reportImplicitAny(node, anyType);
                    }
                    if (functionFlags & 1 /* Generator */ && ts.nodeIsPresent(body)) {
                        // A generator with a body and no type annotation can still cause errors. It can error if the
                        // yielded values have no common supertype, or it can give an implicit any error if it has no
                        // yielded values. The only way to trigger these errors is to try checking its return type.
                        getReturnTypeOfSignature(getSignatureFromDeclaration(node));
                    }
                }
                // A js function declaration can have a @type tag instead of a return type node, but that type must have a call signature
                if (ts.isInJSFile(node)) {
                    var typeTag = ts.getJSDocTypeTag(node);
                    if (typeTag && typeTag.typeExpression && !getContextualCallSignature(getTypeFromTypeNode(typeTag.typeExpression), node)) {
                        error(typeTag.typeExpression.type, ts.Diagnostics.The_type_of_a_function_declaration_must_match_the_function_s_signature);
                    }
                }
            }
            function registerForUnusedIdentifiersCheck(node) {
                // May be in a call such as getTypeOfNode that happened to call this. But potentiallyUnusedIdentifiers is only defined in the scope of `checkSourceFile`.
                if (produceDiagnostics) {
                    var sourceFile = ts.getSourceFileOfNode(node);
                    var potentiallyUnusedIdentifiers = allPotentiallyUnusedIdentifiers.get(sourceFile.path);
                    if (!potentiallyUnusedIdentifiers) {
                        potentiallyUnusedIdentifiers = [];
                        allPotentiallyUnusedIdentifiers.set(sourceFile.path, potentiallyUnusedIdentifiers);
                    }
                    // TODO: GH#22580
                    // Debug.assert(addToSeen(seenPotentiallyUnusedIdentifiers, getNodeId(node)), "Adding potentially-unused identifier twice");
                    potentiallyUnusedIdentifiers.push(node);
                }
            }
            function checkUnusedIdentifiers(potentiallyUnusedIdentifiers, addDiagnostic) {
                for (var _i = 0, potentiallyUnusedIdentifiers_1 = potentiallyUnusedIdentifiers; _i < potentiallyUnusedIdentifiers_1.length; _i++) {
                    var node = potentiallyUnusedIdentifiers_1[_i];
                    switch (node.kind) {
                        case 255 /* ClassDeclaration */:
                        case 224 /* ClassExpression */:
                            checkUnusedClassMembers(node, addDiagnostic);
                            checkUnusedTypeParameters(node, addDiagnostic);
                            break;
                        case 300 /* SourceFile */:
                        case 259 /* ModuleDeclaration */:
                        case 233 /* Block */:
                        case 261 /* CaseBlock */:
                        case 240 /* ForStatement */:
                        case 241 /* ForInStatement */:
                        case 242 /* ForOfStatement */:
                            checkUnusedLocalsAndParameters(node, addDiagnostic);
                            break;
                        case 169 /* Constructor */:
                        case 211 /* FunctionExpression */:
                        case 254 /* FunctionDeclaration */:
                        case 212 /* ArrowFunction */:
                        case 167 /* MethodDeclaration */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            if (node.body) { // Don't report unused parameters in overloads
                                checkUnusedLocalsAndParameters(node, addDiagnostic);
                            }
                            checkUnusedTypeParameters(node, addDiagnostic);
                            break;
                        case 166 /* MethodSignature */:
                        case 172 /* CallSignature */:
                        case 173 /* ConstructSignature */:
                        case 177 /* FunctionType */:
                        case 178 /* ConstructorType */:
                        case 257 /* TypeAliasDeclaration */:
                        case 256 /* InterfaceDeclaration */:
                            checkUnusedTypeParameters(node, addDiagnostic);
                            break;
                        case 188 /* InferType */:
                            checkUnusedInferTypeParameter(node, addDiagnostic);
                            break;
                        default:
                            ts.Debug.assertNever(node, "Node should not have been registered for unused identifiers check");
                    }
                }
            }
            function errorUnusedLocal(declaration, name, addDiagnostic) {
                var node = ts.getNameOfDeclaration(declaration) || declaration;
                var message = isTypeDeclaration(declaration) ? ts.Diagnostics._0_is_declared_but_never_used : ts.Diagnostics._0_is_declared_but_its_value_is_never_read;
                addDiagnostic(declaration, 0 /* Local */, ts.createDiagnosticForNode(node, message, name));
            }
            function isIdentifierThatStartsWithUnderscore(node) {
                return ts.isIdentifier(node) && ts.idText(node).charCodeAt(0) === 95 /* _ */;
            }
            function checkUnusedClassMembers(node, addDiagnostic) {
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    switch (member.kind) {
                        case 167 /* MethodDeclaration */:
                        case 165 /* PropertyDeclaration */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            if (member.kind === 171 /* SetAccessor */ && member.symbol.flags & 32768 /* GetAccessor */) {
                                // Already would have reported an error on the getter.
                                break;
                            }
                            var symbol = getSymbolOfNode(member);
                            if (!symbol.isReferenced
                                && (ts.hasEffectiveModifier(member, 8 /* Private */) || ts.isNamedDeclaration(member) && ts.isPrivateIdentifier(member.name))
                                && !(member.flags & 8388608 /* Ambient */)) {
                                addDiagnostic(member, 0 /* Local */, ts.createDiagnosticForNode(member.name, ts.Diagnostics._0_is_declared_but_its_value_is_never_read, symbolToString(symbol)));
                            }
                            break;
                        case 169 /* Constructor */:
                            for (var _b = 0, _c = member.parameters; _b < _c.length; _b++) {
                                var parameter = _c[_b];
                                if (!parameter.symbol.isReferenced && ts.hasSyntacticModifier(parameter, 8 /* Private */)) {
                                    addDiagnostic(parameter, 0 /* Local */, ts.createDiagnosticForNode(parameter.name, ts.Diagnostics.Property_0_is_declared_but_its_value_is_never_read, ts.symbolName(parameter.symbol)));
                                }
                            }
                            break;
                        case 174 /* IndexSignature */:
                        case 232 /* SemicolonClassElement */:
                        case 168 /* ClassStaticBlockDeclaration */:
                            // Can't be private
                            break;
                        default:
                            ts.Debug.fail("Unexpected class member");
                    }
                }
            }
            function checkUnusedInferTypeParameter(node, addDiagnostic) {
                var typeParameter = node.typeParameter;
                if (isTypeParameterUnused(typeParameter)) {
                    addDiagnostic(node, 1 /* Parameter */, ts.createDiagnosticForNode(node, ts.Diagnostics._0_is_declared_but_its_value_is_never_read, ts.idText(typeParameter.name)));
                }
            }
            function checkUnusedTypeParameters(node, addDiagnostic) {
                // Only report errors on the last declaration for the type parameter container;
                // this ensures that all uses have been accounted for.
                var declarations = getSymbolOfNode(node).declarations;
                if (!declarations || ts.last(declarations) !== node)
                    return;
                var typeParameters = ts.getEffectiveTypeParameterDeclarations(node);
                var seenParentsWithEveryUnused = new ts.Set();
                for (var _i = 0, typeParameters_3 = typeParameters; _i < typeParameters_3.length; _i++) {
                    var typeParameter = typeParameters_3[_i];
                    if (!isTypeParameterUnused(typeParameter))
                        continue;
                    var name = ts.idText(typeParameter.name);
                    var parent = typeParameter.parent;
                    if (parent.kind !== 188 /* InferType */ && parent.typeParameters.every(isTypeParameterUnused)) {
                        if (ts.tryAddToSet(seenParentsWithEveryUnused, parent)) {
                            var sourceFile = ts.getSourceFileOfNode(parent);
                            var range = ts.isJSDocTemplateTag(parent)
                                // Whole @template tag
                                ? ts.rangeOfNode(parent)
                                // Include the `<>` in the error message
                                : ts.rangeOfTypeParameters(sourceFile, parent.typeParameters);
                            var only = parent.typeParameters.length === 1;
                            //TODO: following line is possible reason for bug #41974, unusedTypeParameters_TemplateTag
                            var message = only ? ts.Diagnostics._0_is_declared_but_its_value_is_never_read : ts.Diagnostics.All_type_parameters_are_unused;
                            var arg0 = only ? name : undefined;
                            addDiagnostic(typeParameter, 1 /* Parameter */, ts.createFileDiagnostic(sourceFile, range.pos, range.end - range.pos, message, arg0));
                        }
                    }
                    else {
                        //TODO: following line is possible reason for bug #41974, unusedTypeParameters_TemplateTag
                        addDiagnostic(typeParameter, 1 /* Parameter */, ts.createDiagnosticForNode(typeParameter, ts.Diagnostics._0_is_declared_but_its_value_is_never_read, name));
                    }
                }
            }
            function isTypeParameterUnused(typeParameter) {
                return !(getMergedSymbol(typeParameter.symbol).isReferenced & 262144 /* TypeParameter */) && !isIdentifierThatStartsWithUnderscore(typeParameter.name);
            }
            function addToGroup(map, key, value, getKey) {
                var keyString = String(getKey(key));
                var group = map.get(keyString);
                if (group) {
                    group[1].push(value);
                }
                else {
                    map.set(keyString, [key, [value]]);
                }
            }
            function tryGetRootParameterDeclaration(node) {
                return ts.tryCast(ts.getRootDeclaration(node), ts.isParameter);
            }
            function isValidUnusedLocalDeclaration(declaration) {
                if (ts.isBindingElement(declaration)) {
                    if (ts.isObjectBindingPattern(declaration.parent)) {
                        /**
                         * ignore starts with underscore names _
                         * const { a: _a } = { a: 1 }
                         */
                        return !!(declaration.propertyName && isIdentifierThatStartsWithUnderscore(declaration.name));
                    }
                    return isIdentifierThatStartsWithUnderscore(declaration.name);
                }
                return ts.isAmbientModule(declaration) ||
                    (ts.isVariableDeclaration(declaration) && ts.isForInOrOfStatement(declaration.parent.parent) || isImportedDeclaration(declaration)) && isIdentifierThatStartsWithUnderscore(declaration.name);
            }
            function checkUnusedLocalsAndParameters(nodeWithLocals, addDiagnostic) {
                // Ideally we could use the ImportClause directly as a key, but must wait until we have full ES6 maps. So must store key along with value.
                var unusedImports = new ts.Map();
                var unusedDestructures = new ts.Map();
                var unusedVariables = new ts.Map();
                nodeWithLocals.locals.forEach(function (local) {
                    // If it's purely a type parameter, ignore, will be checked in `checkUnusedTypeParameters`.
                    // If it's a type parameter merged with a parameter, check if the parameter-side is used.
                    if (local.flags & 262144 /* TypeParameter */ ? !(local.flags & 3 /* Variable */ && !(local.isReferenced & 3 /* Variable */)) : local.isReferenced || local.exportSymbol) {
                        return;
                    }
                    if (local.declarations) {
                        for (var _i = 0, _a = local.declarations; _i < _a.length; _i++) {
                            var declaration = _a[_i];
                            if (isValidUnusedLocalDeclaration(declaration)) {
                                continue;
                            }
                            if (isImportedDeclaration(declaration)) {
                                addToGroup(unusedImports, importClauseFromImported(declaration), declaration, getNodeId);
                            }
                            else if (ts.isBindingElement(declaration) && ts.isObjectBindingPattern(declaration.parent)) {
                                // In `{ a, ...b }, `a` is considered used since it removes a property from `b`. `b` may still be unused though.
                                var lastElement = ts.last(declaration.parent.elements);
                                if (declaration === lastElement || !ts.last(declaration.parent.elements).dotDotDotToken) {
                                    addToGroup(unusedDestructures, declaration.parent, declaration, getNodeId);
                                }
                            }
                            else if (ts.isVariableDeclaration(declaration)) {
                                addToGroup(unusedVariables, declaration.parent, declaration, getNodeId);
                            }
                            else {
                                var parameter = local.valueDeclaration && tryGetRootParameterDeclaration(local.valueDeclaration);
                                var name = local.valueDeclaration && ts.getNameOfDeclaration(local.valueDeclaration);
                                if (parameter && name) {
                                    if (!ts.isParameterPropertyDeclaration(parameter, parameter.parent) && !ts.parameterIsThisKeyword(parameter) && !isIdentifierThatStartsWithUnderscore(name)) {
                                        if (ts.isBindingElement(declaration) && ts.isArrayBindingPattern(declaration.parent)) {
                                            addToGroup(unusedDestructures, declaration.parent, declaration, getNodeId);
                                        }
                                        else {
                                            addDiagnostic(parameter, 1 /* Parameter */, ts.createDiagnosticForNode(name, ts.Diagnostics._0_is_declared_but_its_value_is_never_read, ts.symbolName(local)));
                                        }
                                    }
                                }
                                else {
                                    errorUnusedLocal(declaration, ts.symbolName(local), addDiagnostic);
                                }
                            }
                        }
                    }
                });
                unusedImports.forEach(function (_a) {
                    var importClause = _a[0], unuseds = _a[1];
                    var importDecl = importClause.parent;
                    var nDeclarations = (importClause.name ? 1 : 0) +
                        (importClause.namedBindings ?
                            (importClause.namedBindings.kind === 266 /* NamespaceImport */ ? 1 : importClause.namedBindings.elements.length)
                            : 0);
                    if (nDeclarations === unuseds.length) {
                        addDiagnostic(importDecl, 0 /* Local */, unuseds.length === 1
                            ? ts.createDiagnosticForNode(importDecl, ts.Diagnostics._0_is_declared_but_its_value_is_never_read, ts.idText(ts.first(unuseds).name))
                            : ts.createDiagnosticForNode(importDecl, ts.Diagnostics.All_imports_in_import_declaration_are_unused));
                    }
                    else {
                        for (var _i = 0, unuseds_1 = unuseds; _i < unuseds_1.length; _i++) {
                            var unused = unuseds_1[_i];
                            errorUnusedLocal(unused, ts.idText(unused.name), addDiagnostic);
                        }
                    }
                });
                unusedDestructures.forEach(function (_a) {
                    var bindingPattern = _a[0], bindingElements = _a[1];
                    var kind = tryGetRootParameterDeclaration(bindingPattern.parent) ? 1 /* Parameter */ : 0 /* Local */;
                    if (bindingPattern.elements.length === bindingElements.length) {
                        if (bindingElements.length === 1 && bindingPattern.parent.kind === 252 /* VariableDeclaration */ && bindingPattern.parent.parent.kind === 253 /* VariableDeclarationList */) {
                            addToGroup(unusedVariables, bindingPattern.parent.parent, bindingPattern.parent, getNodeId);
                        }
                        else {
                            addDiagnostic(bindingPattern, kind, bindingElements.length === 1
                                ? ts.createDiagnosticForNode(bindingPattern, ts.Diagnostics._0_is_declared_but_its_value_is_never_read, bindingNameText(ts.first(bindingElements).name))
                                : ts.createDiagnosticForNode(bindingPattern, ts.Diagnostics.All_destructured_elements_are_unused));
                        }
                    }
                    else {
                        for (var _i = 0, bindingElements_1 = bindingElements; _i < bindingElements_1.length; _i++) {
                            var e = bindingElements_1[_i];
                            addDiagnostic(e, kind, ts.createDiagnosticForNode(e, ts.Diagnostics._0_is_declared_but_its_value_is_never_read, bindingNameText(e.name)));
                        }
                    }
                });
                unusedVariables.forEach(function (_a) {
                    var declarationList = _a[0], declarations = _a[1];
                    if (declarationList.declarations.length === declarations.length) {
                        addDiagnostic(declarationList, 0 /* Local */, declarations.length === 1
                            ? ts.createDiagnosticForNode(ts.first(declarations).name, ts.Diagnostics._0_is_declared_but_its_value_is_never_read, bindingNameText(ts.first(declarations).name))
                            : ts.createDiagnosticForNode(declarationList.parent.kind === 235 /* VariableStatement */ ? declarationList.parent : declarationList, ts.Diagnostics.All_variables_are_unused));
                    }
                    else {
                        for (var _i = 0, declarations_5 = declarations; _i < declarations_5.length; _i++) {
                            var decl = declarations_5[_i];
                            addDiagnostic(decl, 0 /* Local */, ts.createDiagnosticForNode(decl, ts.Diagnostics._0_is_declared_but_its_value_is_never_read, bindingNameText(decl.name)));
                        }
                    }
                });
            }
            function bindingNameText(name) {
                switch (name.kind) {
                    case 79 /* Identifier */:
                        return ts.idText(name);
                    case 200 /* ArrayBindingPattern */:
                    case 199 /* ObjectBindingPattern */:
                        return bindingNameText(ts.cast(ts.first(name.elements), ts.isBindingElement).name);
                    default:
                        return ts.Debug.assertNever(name);
                }
            }
            function isImportedDeclaration(node) {
                return node.kind === 265 /* ImportClause */ || node.kind === 268 /* ImportSpecifier */ || node.kind === 266 /* NamespaceImport */;
            }
            function importClauseFromImported(decl) {
                return decl.kind === 265 /* ImportClause */ ? decl : decl.kind === 266 /* NamespaceImport */ ? decl.parent : decl.parent.parent;
            }
            function checkBlock(node) {
                // Grammar checking for SyntaxKind.Block
                if (node.kind === 233 /* Block */) {
                    checkGrammarStatementInAmbientContext(node);
                }
                if (ts.isFunctionOrModuleBlock(node)) {
                    var saveFlowAnalysisDisabled = flowAnalysisDisabled;
                    ts.forEach(node.statements, checkSourceElement);
                    flowAnalysisDisabled = saveFlowAnalysisDisabled;
                }
                else {
                    ts.forEach(node.statements, checkSourceElement);
                }
                if (node.locals) {
                    registerForUnusedIdentifiersCheck(node);
                }
            }
            function checkCollisionWithArgumentsInGeneratedCode(node) {
                // no rest parameters \ declaration context \ overload - no codegen impact
                if (languageVersion >= 2 /* ES2015 */ || !ts.hasRestParameter(node) || node.flags & 8388608 /* Ambient */ || ts.nodeIsMissing(node.body)) {
                    return;
                }
                ts.forEach(node.parameters, function (p) {
                    if (p.name && !ts.isBindingPattern(p.name) && p.name.escapedText === argumentsSymbol.escapedName) {
                        errorSkippedOn("noEmit", p, ts.Diagnostics.Duplicate_identifier_arguments_Compiler_uses_arguments_to_initialize_rest_parameters);
                    }
                });
            }
            /**
             * Checks whether an {@link Identifier}, in the context of another {@link Node}, would collide with a runtime value
             * of {@link name} in an outer scope. This is used to check for collisions for downlevel transformations that
             * require names like `Object`, `Promise`, `Reflect`, `require`, `exports`, etc.
             */
            function needCollisionCheckForIdentifier(node, identifier, name) {
                if ((identifier === null || identifier === void 0 ? void 0 : identifier.escapedText) !== name) {
                    return false;
                }
                if (node.kind === 165 /* PropertyDeclaration */ ||
                    node.kind === 164 /* PropertySignature */ ||
                    node.kind === 167 /* MethodDeclaration */ ||
                    node.kind === 166 /* MethodSignature */ ||
                    node.kind === 170 /* GetAccessor */ ||
                    node.kind === 171 /* SetAccessor */ ||
                    node.kind === 291 /* PropertyAssignment */) {
                    // it is ok to have member named '_super', '_this', `Promise`, etc. - member access is always qualified
                    return false;
                }
                if (node.flags & 8388608 /* Ambient */) {
                    // ambient context - no codegen impact
                    return false;
                }
                if (ts.isImportClause(node) || ts.isImportEqualsDeclaration(node) || ts.isImportSpecifier(node)) {
                    // type-only imports do not require collision checks against runtime values.
                    if (ts.isTypeOnlyImportOrExportDeclaration(node)) {
                        return false;
                    }
                }
                var root = ts.getRootDeclaration(node);
                if (ts.isParameter(root) && ts.nodeIsMissing(root.parent.body)) {
                    // just an overload - no codegen impact
                    return false;
                }
                return true;
            }
            // this function will run after checking the source file so 'CaptureThis' is correct for all nodes
            function checkIfThisIsCapturedInEnclosingScope(node) {
                ts.findAncestor(node, function (current) {
                    if (getNodeCheckFlags(current) & 4 /* CaptureThis */) {
                        var isDeclaration_1 = node.kind !== 79 /* Identifier */;
                        if (isDeclaration_1) {
                            error(ts.getNameOfDeclaration(node), ts.Diagnostics.Duplicate_identifier_this_Compiler_uses_variable_declaration_this_to_capture_this_reference);
                        }
                        else {
                            error(node, ts.Diagnostics.Expression_resolves_to_variable_declaration_this_that_compiler_uses_to_capture_this_reference);
                        }
                        return true;
                    }
                    return false;
                });
            }
            function checkIfNewTargetIsCapturedInEnclosingScope(node) {
                ts.findAncestor(node, function (current) {
                    if (getNodeCheckFlags(current) & 8 /* CaptureNewTarget */) {
                        var isDeclaration_2 = node.kind !== 79 /* Identifier */;
                        if (isDeclaration_2) {
                            error(ts.getNameOfDeclaration(node), ts.Diagnostics.Duplicate_identifier_newTarget_Compiler_uses_variable_declaration_newTarget_to_capture_new_target_meta_property_reference);
                        }
                        else {
                            error(node, ts.Diagnostics.Expression_resolves_to_variable_declaration_newTarget_that_compiler_uses_to_capture_new_target_meta_property_reference);
                        }
                        return true;
                    }
                    return false;
                });
            }
            function checkCollisionWithRequireExportsInGeneratedCode(node, name) {
                // No need to check for require or exports for ES6 modules and later
                if (moduleKind >= ts.ModuleKind.ES2015) {
                    return;
                }
                if (!name || !needCollisionCheckForIdentifier(node, name, "require") && !needCollisionCheckForIdentifier(node, name, "exports")) {
                    return;
                }
                // Uninstantiated modules shouldnt do this check
                if (ts.isModuleDeclaration(node) && ts.getModuleInstanceState(node) !== 1 /* Instantiated */) {
                    return;
                }
                // In case of variable declaration, node.parent is variable statement so look at the variable statement's parent
                var parent = getDeclarationContainer(node);
                if (parent.kind === 300 /* SourceFile */ && ts.isExternalOrCommonJsModule(parent)) {
                    // If the declaration happens to be in external module, report error that require and exports are reserved keywords
                    errorSkippedOn("noEmit", name, ts.Diagnostics.Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module, ts.declarationNameToString(name), ts.declarationNameToString(name));
                }
            }
            function checkCollisionWithGlobalPromiseInGeneratedCode(node, name) {
                if (!name || languageVersion >= 4 /* ES2017 */ || !needCollisionCheckForIdentifier(node, name, "Promise")) {
                    return;
                }
                // Uninstantiated modules shouldnt do this check
                if (ts.isModuleDeclaration(node) && ts.getModuleInstanceState(node) !== 1 /* Instantiated */) {
                    return;
                }
                // In case of variable declaration, node.parent is variable statement so look at the variable statement's parent
                var parent = getDeclarationContainer(node);
                if (parent.kind === 300 /* SourceFile */ && ts.isExternalOrCommonJsModule(parent) && parent.flags & 2048 /* HasAsyncFunctions */) {
                    // If the declaration happens to be in external module, report error that Promise is a reserved identifier.
                    errorSkippedOn("noEmit", name, ts.Diagnostics.Duplicate_identifier_0_Compiler_reserves_name_1_in_top_level_scope_of_a_module_containing_async_functions, ts.declarationNameToString(name), ts.declarationNameToString(name));
                }
            }
            function recordPotentialCollisionWithWeakMapSetInGeneratedCode(node, name) {
                if (languageVersion <= 8 /* ES2021 */
                    && (needCollisionCheckForIdentifier(node, name, "WeakMap") || needCollisionCheckForIdentifier(node, name, "WeakSet"))) {
                    potentialWeakMapSetCollisions.push(node);
                }
            }
            function checkWeakMapSetCollision(node) {
                var enclosingBlockScope = ts.getEnclosingBlockScopeContainer(node);
                if (getNodeCheckFlags(enclosingBlockScope) & 67108864 /* ContainsClassWithPrivateIdentifiers */) {
                    ts.Debug.assert(ts.isNamedDeclaration(node) && ts.isIdentifier(node.name) && typeof node.name.escapedText === "string", "The target of a WeakMap/WeakSet collision check should be an identifier");
                    errorSkippedOn("noEmit", node, ts.Diagnostics.Compiler_reserves_name_0_when_emitting_private_identifier_downlevel, node.name.escapedText);
                }
            }
            function recordPotentialCollisionWithReflectInGeneratedCode(node, name) {
                if (name && languageVersion >= 2 /* ES2015 */ && languageVersion <= 8 /* ES2021 */
                    && needCollisionCheckForIdentifier(node, name, "Reflect")) {
                    potentialReflectCollisions.push(node);
                }
            }
            function checkReflectCollision(node) {
                var hasCollision = false;
                if (ts.isClassExpression(node)) {
                    // ClassExpression names don't contribute to their containers, but do matter for any of their block-scoped members.
                    for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                        var member = _a[_i];
                        if (getNodeCheckFlags(member) & 134217728 /* ContainsSuperPropertyInStaticInitializer */) {
                            hasCollision = true;
                            break;
                        }
                    }
                }
                else if (ts.isFunctionExpression(node)) {
                    // FunctionExpression names don't contribute to their containers, but do matter for their contents
                    if (getNodeCheckFlags(node) & 134217728 /* ContainsSuperPropertyInStaticInitializer */) {
                        hasCollision = true;
                    }
                }
                else {
                    var container = ts.getEnclosingBlockScopeContainer(node);
                    if (container && getNodeCheckFlags(container) & 134217728 /* ContainsSuperPropertyInStaticInitializer */) {
                        hasCollision = true;
                    }
                }
                if (hasCollision) {
                    ts.Debug.assert(ts.isNamedDeclaration(node) && ts.isIdentifier(node.name), "The target of a Reflect collision check should be an identifier");
                    errorSkippedOn("noEmit", node, ts.Diagnostics.Duplicate_identifier_0_Compiler_reserves_name_1_when_emitting_super_references_in_static_initializers, ts.declarationNameToString(node.name), "Reflect");
                }
            }
            function checkCollisionsForDeclarationName(node, name) {
                if (!name)
                    return;
                checkCollisionWithRequireExportsInGeneratedCode(node, name);
                checkCollisionWithGlobalPromiseInGeneratedCode(node, name);
                recordPotentialCollisionWithWeakMapSetInGeneratedCode(node, name);
                recordPotentialCollisionWithReflectInGeneratedCode(node, name);
                if (ts.isClassLike(node)) {
                    checkTypeNameIsReserved(name, ts.Diagnostics.Class_name_cannot_be_0);
                    if (!(node.flags & 8388608 /* Ambient */)) {
                        checkClassNameCollisionWithObject(name);
                    }
                }
                else if (ts.isEnumDeclaration(node)) {
                    checkTypeNameIsReserved(name, ts.Diagnostics.Enum_name_cannot_be_0);
                }
            }
            function checkVarDeclaredNamesNotShadowed(node) {
                // - ScriptBody : StatementList
                // It is a Syntax Error if any element of the LexicallyDeclaredNames of StatementList
                // also occurs in the VarDeclaredNames of StatementList.
                // - Block : { StatementList }
                // It is a Syntax Error if any element of the LexicallyDeclaredNames of StatementList
                // also occurs in the VarDeclaredNames of StatementList.
                // Variable declarations are hoisted to the top of their function scope. They can shadow
                // block scoped declarations, which bind tighter. this will not be flagged as duplicate definition
                // by the binder as the declaration scope is different.
                // A non-initialized declaration is a no-op as the block declaration will resolve before the var
                // declaration. the problem is if the declaration has an initializer. this will act as a write to the
                // block declared value. this is fine for let, but not const.
                // Only consider declarations with initializers, uninitialized const declarations will not
                // step on a let/const variable.
                // Do not consider const and const declarations, as duplicate block-scoped declarations
                // are handled by the binder.
                // We are only looking for const declarations that step on let\const declarations from a
                // different scope. e.g.:
                //      {
                //          const x = 0; // localDeclarationSymbol obtained after name resolution will correspond to this declaration
                //          const x = 0; // symbol for this declaration will be 'symbol'
                //      }
                // skip block-scoped variables and parameters
                if ((ts.getCombinedNodeFlags(node) & 3 /* BlockScoped */) !== 0 || ts.isParameterDeclaration(node)) {
                    return;
                }
                // skip variable declarations that don't have initializers
                // NOTE: in ES6 spec initializer is required in variable declarations where name is binding pattern
                // so we'll always treat binding elements as initialized
                if (node.kind === 252 /* VariableDeclaration */ && !node.initializer) {
                    return;
                }
                var symbol = getSymbolOfNode(node);
                if (symbol.flags & 1 /* FunctionScopedVariable */) {
                    if (!ts.isIdentifier(node.name))
                        return ts.Debug.fail();
                    var localDeclarationSymbol = resolveName(node, node.name.escapedText, 3 /* Variable */, /*nodeNotFoundErrorMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false);
                    if (localDeclarationSymbol &&
                        localDeclarationSymbol !== symbol &&
                        localDeclarationSymbol.flags & 2 /* BlockScopedVariable */) {
                        if (getDeclarationNodeFlagsFromSymbol(localDeclarationSymbol) & 3 /* BlockScoped */) {
                            var varDeclList = ts.getAncestor(localDeclarationSymbol.valueDeclaration, 253 /* VariableDeclarationList */);
                            var container = varDeclList.parent.kind === 235 /* VariableStatement */ && varDeclList.parent.parent
                                ? varDeclList.parent.parent
                                : undefined;
                            // names of block-scoped and function scoped variables can collide only
                            // if block scoped variable is defined in the function\module\source file scope (because of variable hoisting)
                            var namesShareScope = container &&
                                (container.kind === 233 /* Block */ && ts.isFunctionLike(container.parent) ||
                                    container.kind === 260 /* ModuleBlock */ ||
                                    container.kind === 259 /* ModuleDeclaration */ ||
                                    container.kind === 300 /* SourceFile */);
                            // here we know that function scoped variable is shadowed by block scoped one
                            // if they are defined in the same scope - binder has already reported redeclaration error
                            // otherwise if variable has an initializer - show error that initialization will fail
                            // since LHS will be block scoped name instead of function scoped
                            if (!namesShareScope) {
                                var name = symbolToString(localDeclarationSymbol);
                                error(node, ts.Diagnostics.Cannot_initialize_outer_scoped_variable_0_in_the_same_scope_as_block_scoped_declaration_1, name, name);
                            }
                        }
                    }
                }
            }
            function convertAutoToAny(type) {
                return type === autoType ? anyType : type === autoArrayType ? anyArrayType : type;
            }
            // Check variable, parameter, or property declaration
            function checkVariableLikeDeclaration(node) {
                var _a;
                checkDecorators(node);
                if (!ts.isBindingElement(node)) {
                    checkSourceElement(node.type);
                }
                // JSDoc `function(string, string): string` syntax results in parameters with no name
                if (!node.name) {
                    return;
                }
                // For a computed property, just check the initializer and exit
                // Do not use hasDynamicName here, because that returns false for well known symbols.
                // We want to perform checkComputedPropertyName for all computed properties, including
                // well known symbols.
                if (node.name.kind === 160 /* ComputedPropertyName */) {
                    checkComputedPropertyName(node.name);
                    if (node.initializer) {
                        checkExpressionCached(node.initializer);
                    }
                }
                if (ts.isBindingElement(node)) {
                    if (ts.isObjectBindingPattern(node.parent) && node.dotDotDotToken && languageVersion < 5 /* ES2018 */) {
                        checkExternalEmitHelpers(node, 4 /* Rest */);
                    }
                    // check computed properties inside property names of binding elements
                    if (node.propertyName && node.propertyName.kind === 160 /* ComputedPropertyName */) {
                        checkComputedPropertyName(node.propertyName);
                    }
                    // check private/protected variable access
                    var parent = node.parent.parent;
                    var parentType = getTypeForBindingElementParent(parent);
                    var name = node.propertyName || node.name;
                    if (parentType && !ts.isBindingPattern(name)) {
                        var exprType = getLiteralTypeFromPropertyName(name);
                        if (isTypeUsableAsPropertyName(exprType)) {
                            var nameText = getPropertyNameFromType(exprType);
                            var property = getPropertyOfType(parentType, nameText);
                            if (property) {
                                markPropertyAsReferenced(property, /*nodeForCheckWriteOnly*/ undefined, /*isSelfTypeAccess*/ false); // A destructuring is never a write-only reference.
                                checkPropertyAccessibility(node, !!parent.initializer && parent.initializer.kind === 106 /* SuperKeyword */, /*writing*/ false, parentType, property);
                            }
                        }
                    }
                }
                // For a binding pattern, check contained binding elements
                if (ts.isBindingPattern(node.name)) {
                    if (node.name.kind === 200 /* ArrayBindingPattern */ && languageVersion < 2 /* ES2015 */ && compilerOptions.downlevelIteration) {
                        checkExternalEmitHelpers(node, 512 /* Read */);
                    }
                    ts.forEach(node.name.elements, checkSourceElement);
                }
                // For a parameter declaration with an initializer, error and exit if the containing function doesn't have a body
                if (node.initializer && ts.isParameterDeclaration(node) && ts.nodeIsMissing(ts.getContainingFunction(node).body)) {
                    error(node, ts.Diagnostics.A_parameter_initializer_is_only_allowed_in_a_function_or_constructor_implementation);
                    return;
                }
                // For a binding pattern, validate the initializer and exit
                if (ts.isBindingPattern(node.name)) {
                    var needCheckInitializer = node.initializer && node.parent.parent.kind !== 241 /* ForInStatement */;
                    var needCheckWidenedType = node.name.elements.length === 0;
                    if (needCheckInitializer || needCheckWidenedType) {
                        // Don't validate for-in initializer as it is already an error
                        var widenedType = getWidenedTypeForVariableLikeDeclaration(node);
                        if (needCheckInitializer) {
                            var initializerType = checkExpressionCached(node.initializer);
                            if (strictNullChecks && needCheckWidenedType) {
                                checkNonNullNonVoidType(initializerType, node);
                            }
                            else {
                                checkTypeAssignableToAndOptionallyElaborate(initializerType, getWidenedTypeForVariableLikeDeclaration(node), node, node.initializer);
                            }
                        }
                        // check the binding pattern with empty elements
                        if (needCheckWidenedType) {
                            if (ts.isArrayBindingPattern(node.name)) {
                                checkIteratedTypeOrElementType(65 /* Destructuring */, widenedType, undefinedType, node);
                            }
                            else if (strictNullChecks) {
                                checkNonNullNonVoidType(widenedType, node);
                            }
                        }
                    }
                    return;
                }
                // For a commonjs `const x = require`, validate the alias and exit
                var symbol = getSymbolOfNode(node);
                if (symbol.flags & 2097152 /* Alias */ && ts.isRequireVariableDeclaration(node)) {
                    checkAliasSymbol(node);
                    return;
                }
                var type = convertAutoToAny(getTypeOfSymbol(symbol));
                if (node === symbol.valueDeclaration) {
                    // Node is the primary declaration of the symbol, just validate the initializer
                    // Don't validate for-in initializer as it is already an error
                    var initializer = ts.getEffectiveInitializer(node);
                    if (initializer) {
                        var isJSObjectLiteralInitializer = ts.isInJSFile(node) &&
                            ts.isObjectLiteralExpression(initializer) &&
                            (initializer.properties.length === 0 || ts.isPrototypeAccess(node.name)) &&
                            !!((_a = symbol.exports) === null || _a === void 0 ? void 0 : _a.size);
                        if (!isJSObjectLiteralInitializer && node.parent.parent.kind !== 241 /* ForInStatement */) {
                            checkTypeAssignableToAndOptionallyElaborate(checkExpressionCached(initializer), type, node, initializer, /*headMessage*/ undefined);
                        }
                    }
                    if (symbol.declarations && symbol.declarations.length > 1) {
                        if (ts.some(symbol.declarations, function (d) { return d !== node && ts.isVariableLike(d) && !areDeclarationFlagsIdentical(d, node); })) {
                            error(node.name, ts.Diagnostics.All_declarations_of_0_must_have_identical_modifiers, ts.declarationNameToString(node.name));
                        }
                    }
                }
                else {
                    // Node is a secondary declaration, check that type is identical to primary declaration and check that
                    // initializer is consistent with type associated with the node
                    var declarationType = convertAutoToAny(getWidenedTypeForVariableLikeDeclaration(node));
                    if (type !== errorType && declarationType !== errorType &&
                        !isTypeIdenticalTo(type, declarationType) &&
                        !(symbol.flags & 67108864 /* Assignment */)) {
                        errorNextVariableOrPropertyDeclarationMustHaveSameType(symbol.valueDeclaration, type, node, declarationType);
                    }
                    if (node.initializer) {
                        checkTypeAssignableToAndOptionallyElaborate(checkExpressionCached(node.initializer), declarationType, node, node.initializer, /*headMessage*/ undefined);
                    }
                    if (symbol.valueDeclaration && !areDeclarationFlagsIdentical(node, symbol.valueDeclaration)) {
                        error(node.name, ts.Diagnostics.All_declarations_of_0_must_have_identical_modifiers, ts.declarationNameToString(node.name));
                    }
                }
                if (node.kind !== 165 /* PropertyDeclaration */ && node.kind !== 164 /* PropertySignature */) {
                    // We know we don't have a binding pattern or computed name here
                    checkExportsOnMergedDeclarations(node);
                    if (node.kind === 252 /* VariableDeclaration */ || node.kind === 201 /* BindingElement */) {
                        checkVarDeclaredNamesNotShadowed(node);
                    }
                    checkCollisionsForDeclarationName(node, node.name);
                }
            }
            function errorNextVariableOrPropertyDeclarationMustHaveSameType(firstDeclaration, firstType, nextDeclaration, nextType) {
                var nextDeclarationName = ts.getNameOfDeclaration(nextDeclaration);
                var message = nextDeclaration.kind === 165 /* PropertyDeclaration */ || nextDeclaration.kind === 164 /* PropertySignature */
                    ? ts.Diagnostics.Subsequent_property_declarations_must_have_the_same_type_Property_0_must_be_of_type_1_but_here_has_type_2
                    : ts.Diagnostics.Subsequent_variable_declarations_must_have_the_same_type_Variable_0_must_be_of_type_1_but_here_has_type_2;
                var declName = ts.declarationNameToString(nextDeclarationName);
                var err = error(nextDeclarationName, message, declName, typeToString(firstType), typeToString(nextType));
                if (firstDeclaration) {
                    ts.addRelatedInfo(err, ts.createDiagnosticForNode(firstDeclaration, ts.Diagnostics._0_was_also_declared_here, declName));
                }
            }
            function areDeclarationFlagsIdentical(left, right) {
                if ((left.kind === 162 /* Parameter */ && right.kind === 252 /* VariableDeclaration */) ||
                    (left.kind === 252 /* VariableDeclaration */ && right.kind === 162 /* Parameter */)) {
                    // Differences in optionality between parameters and variables are allowed.
                    return true;
                }
                if (ts.hasQuestionToken(left) !== ts.hasQuestionToken(right)) {
                    return false;
                }
                var interestingFlags = 8 /* Private */ |
                    16 /* Protected */ |
                    256 /* Async */ |
                    128 /* Abstract */ |
                    64 /* Readonly */ |
                    32 /* Static */;
                return ts.getSelectedEffectiveModifierFlags(left, interestingFlags) === ts.getSelectedEffectiveModifierFlags(right, interestingFlags);
            }
            function checkVariableDeclaration(node) {
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("check" /* Check */, "checkVariableDeclaration", { kind: node.kind, pos: node.pos, end: node.end });
                checkGrammarVariableDeclaration(node);
                checkVariableLikeDeclaration(node);
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
            }
            function checkBindingElement(node) {
                checkGrammarBindingElement(node);
                return checkVariableLikeDeclaration(node);
            }
            function checkVariableStatement(node) {
                // Grammar checking
                if (!checkGrammarDecoratorsAndModifiers(node) && !checkGrammarVariableDeclarationList(node.declarationList))
                    checkGrammarForDisallowedLetOrConstStatement(node);
                ts.forEach(node.declarationList.declarations, checkSourceElement);
            }
            function checkExpressionStatement(node) {
                // Grammar checking
                checkGrammarStatementInAmbientContext(node);
                checkExpression(node.expression);
            }
            function checkIfStatement(node) {
                // Grammar checking
                checkGrammarStatementInAmbientContext(node);
                var type = checkTruthinessExpression(node.expression);
                checkTestingKnownTruthyCallableOrAwaitableType(node.expression, type, node.thenStatement);
                checkSourceElement(node.thenStatement);
                if (node.thenStatement.kind === 234 /* EmptyStatement */) {
                    error(node.thenStatement, ts.Diagnostics.The_body_of_an_if_statement_cannot_be_the_empty_statement);
                }
                checkSourceElement(node.elseStatement);
            }
            function checkTestingKnownTruthyCallableOrAwaitableType(condExpr, type, body) {
                if (!strictNullChecks)
                    return;
                if (getFalsyFlags(type))
                    return;
                var location = ts.isBinaryExpression(condExpr) ? condExpr.right : condExpr;
                if (ts.isPropertyAccessExpression(location) && ts.isAssertionExpression(ts.skipParentheses(location.expression))) {
                    return;
                }
                var testedNode = ts.isIdentifier(location) ? location
                    : ts.isPropertyAccessExpression(location) ? location.name
                        : ts.isBinaryExpression(location) && ts.isIdentifier(location.right) ? location.right
                            : undefined;
                // While it technically should be invalid for any known-truthy value
                // to be tested, we de-scope to functions and Promises unreferenced in
                // the block as a heuristic to identify the most common bugs. There
                // are too many false positives for values sourced from type
                // definitions without strictNullChecks otherwise.
                var callSignatures = getSignaturesOfType(type, 0 /* Call */);
                var isPromise = !!getAwaitedTypeOfPromise(type);
                if (callSignatures.length === 0 && !isPromise) {
                    return;
                }
                var testedSymbol = testedNode && getSymbolAtLocation(testedNode);
                if (!testedSymbol && !isPromise) {
                    return;
                }
                var isUsed = testedSymbol && ts.isBinaryExpression(condExpr.parent) && isSymbolUsedInBinaryExpressionChain(condExpr.parent, testedSymbol)
                    || testedSymbol && body && isSymbolUsedInConditionBody(condExpr, body, testedNode, testedSymbol);
                if (!isUsed) {
                    if (isPromise) {
                        errorAndMaybeSuggestAwait(location, 
                        /*maybeMissingAwait*/ true, ts.Diagnostics.This_condition_will_always_return_true_since_this_0_is_always_defined, getTypeNameForErrorDisplay(type));
                    }
                    else {
                        error(location, ts.Diagnostics.This_condition_will_always_return_true_since_this_function_is_always_defined_Did_you_mean_to_call_it_instead);
                    }
                }
            }
            function isSymbolUsedInConditionBody(expr, body, testedNode, testedSymbol) {
                return !!ts.forEachChild(body, function check(childNode) {
                    if (ts.isIdentifier(childNode)) {
                        var childSymbol = getSymbolAtLocation(childNode);
                        if (childSymbol && childSymbol === testedSymbol) {
                            // If the test was a simple identifier, the above check is sufficient
                            if (ts.isIdentifier(expr)) {
                                return true;
                            }
                            // Otherwise we need to ensure the symbol is called on the same target
                            var testedExpression = testedNode.parent;
                            var childExpression = childNode.parent;
                            while (testedExpression && childExpression) {
                                if (ts.isIdentifier(testedExpression) && ts.isIdentifier(childExpression) ||
                                    testedExpression.kind === 108 /* ThisKeyword */ && childExpression.kind === 108 /* ThisKeyword */) {
                                    return getSymbolAtLocation(testedExpression) === getSymbolAtLocation(childExpression);
                                }
                                else if (ts.isPropertyAccessExpression(testedExpression) && ts.isPropertyAccessExpression(childExpression)) {
                                    if (getSymbolAtLocation(testedExpression.name) !== getSymbolAtLocation(childExpression.name)) {
                                        return false;
                                    }
                                    childExpression = childExpression.expression;
                                    testedExpression = testedExpression.expression;
                                }
                                else if (ts.isCallExpression(testedExpression) && ts.isCallExpression(childExpression)) {
                                    childExpression = childExpression.expression;
                                    testedExpression = testedExpression.expression;
                                }
                                else {
                                    return false;
                                }
                            }
                        }
                    }
                    return ts.forEachChild(childNode, check);
                });
            }
            function isSymbolUsedInBinaryExpressionChain(node, testedSymbol) {
                while (ts.isBinaryExpression(node) && node.operatorToken.kind === 55 /* AmpersandAmpersandToken */) {
                    var isUsed = ts.forEachChild(node.right, function visit(child) {
                        if (ts.isIdentifier(child)) {
                            var symbol = getSymbolAtLocation(child);
                            if (symbol && symbol === testedSymbol) {
                                return true;
                            }
                        }
                        return ts.forEachChild(child, visit);
                    });
                    if (isUsed) {
                        return true;
                    }
                    node = node.parent;
                }
                return false;
            }
            function checkDoStatement(node) {
                // Grammar checking
                checkGrammarStatementInAmbientContext(node);
                checkSourceElement(node.statement);
                checkTruthinessExpression(node.expression);
            }
            function checkWhileStatement(node) {
                // Grammar checking
                checkGrammarStatementInAmbientContext(node);
                checkTruthinessExpression(node.expression);
                checkSourceElement(node.statement);
            }
            function checkTruthinessOfType(type, node) {
                if (type.flags & 16384 /* Void */) {
                    error(node, ts.Diagnostics.An_expression_of_type_void_cannot_be_tested_for_truthiness);
                }
                return type;
            }
            function checkTruthinessExpression(node, checkMode) {
                return checkTruthinessOfType(checkExpression(node, checkMode), node);
            }
            function checkForStatement(node) {
                // Grammar checking
                if (!checkGrammarStatementInAmbientContext(node)) {
                    if (node.initializer && node.initializer.kind === 253 /* VariableDeclarationList */) {
                        checkGrammarVariableDeclarationList(node.initializer);
                    }
                }
                if (node.initializer) {
                    if (node.initializer.kind === 253 /* VariableDeclarationList */) {
                        ts.forEach(node.initializer.declarations, checkVariableDeclaration);
                    }
                    else {
                        checkExpression(node.initializer);
                    }
                }
                if (node.condition)
                    checkTruthinessExpression(node.condition);
                if (node.incrementor)
                    checkExpression(node.incrementor);
                checkSourceElement(node.statement);
                if (node.locals) {
                    registerForUnusedIdentifiersCheck(node);
                }
            }
            function checkForOfStatement(node) {
                checkGrammarForInOrForOfStatement(node);
                var container = ts.getContainingFunctionOrClassStaticBlock(node);
                if (node.awaitModifier) {
                    if (container && ts.isClassStaticBlockDeclaration(container)) {
                        grammarErrorOnNode(node.awaitModifier, ts.Diagnostics.For_await_loops_cannot_be_used_inside_a_class_static_block);
                    }
                    else {
                        var functionFlags = ts.getFunctionFlags(container);
                        if ((functionFlags & (4 /* Invalid */ | 2 /* Async */)) === 2 /* Async */ && languageVersion < 99 /* ESNext */) {
                            // for..await..of in an async function or async generator function prior to ESNext requires the __asyncValues helper
                            checkExternalEmitHelpers(node, 16384 /* ForAwaitOfIncludes */);
                        }
                    }
                }
                else if (compilerOptions.downlevelIteration && languageVersion < 2 /* ES2015 */) {
                    // for..of prior to ES2015 requires the __values helper when downlevelIteration is enabled
                    checkExternalEmitHelpers(node, 256 /* ForOfIncludes */);
                }
                // Check the LHS and RHS
                // If the LHS is a declaration, just check it as a variable declaration, which will in turn check the RHS
                // via checkRightHandSideOfForOf.
                // If the LHS is an expression, check the LHS, as a destructuring assignment or as a reference.
                // Then check that the RHS is assignable to it.
                if (node.initializer.kind === 253 /* VariableDeclarationList */) {
                    checkForInOrForOfVariableDeclaration(node);
                }
                else {
                    var varExpr = node.initializer;
                    var iteratedType = checkRightHandSideOfForOf(node);
                    // There may be a destructuring assignment on the left side
                    if (varExpr.kind === 202 /* ArrayLiteralExpression */ || varExpr.kind === 203 /* ObjectLiteralExpression */) {
                        // iteratedType may be undefined. In this case, we still want to check the structure of
                        // varExpr, in particular making sure it's a valid LeftHandSideExpression. But we'd like
                        // to short circuit the type relation checking as much as possible, so we pass the unknownType.
                        checkDestructuringAssignment(varExpr, iteratedType || errorType);
                    }
                    else {
                        var leftType = checkExpression(varExpr);
                        checkReferenceExpression(varExpr, ts.Diagnostics.The_left_hand_side_of_a_for_of_statement_must_be_a_variable_or_a_property_access, ts.Diagnostics.The_left_hand_side_of_a_for_of_statement_may_not_be_an_optional_property_access);
                        // iteratedType will be undefined if the rightType was missing properties/signatures
                        // required to get its iteratedType (like [Symbol.iterator] or next). This may be
                        // because we accessed properties from anyType, or it may have led to an error inside
                        // getElementTypeOfIterable.
                        if (iteratedType) {
                            checkTypeAssignableToAndOptionallyElaborate(iteratedType, leftType, varExpr, node.expression);
                        }
                    }
                }
                checkSourceElement(node.statement);
                if (node.locals) {
                    registerForUnusedIdentifiersCheck(node);
                }
            }
            function checkForInStatement(node) {
                // Grammar checking
                checkGrammarForInOrForOfStatement(node);
                var rightType = getNonNullableTypeIfNeeded(checkExpression(node.expression));
                // TypeScript 1.0 spec (April 2014): 5.4
                // In a 'for-in' statement of the form
                // for (let VarDecl in Expr) Statement
                //   VarDecl must be a variable declaration without a type annotation that declares a variable of type Any,
                //   and Expr must be an expression of type Any, an object type, or a type parameter type.
                if (node.initializer.kind === 253 /* VariableDeclarationList */) {
                    var variable = node.initializer.declarations[0];
                    if (variable && ts.isBindingPattern(variable.name)) {
                        error(variable.name, ts.Diagnostics.The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern);
                    }
                    checkForInOrForOfVariableDeclaration(node);
                }
                else {
                    // In a 'for-in' statement of the form
                    // for (Var in Expr) Statement
                    //   Var must be an expression classified as a reference of type Any or the String primitive type,
                    //   and Expr must be an expression of type Any, an object type, or a type parameter type.
                    var varExpr = node.initializer;
                    var leftType = checkExpression(varExpr);
                    if (varExpr.kind === 202 /* ArrayLiteralExpression */ || varExpr.kind === 203 /* ObjectLiteralExpression */) {
                        error(varExpr, ts.Diagnostics.The_left_hand_side_of_a_for_in_statement_cannot_be_a_destructuring_pattern);
                    }
                    else if (!isTypeAssignableTo(getIndexTypeOrString(rightType), leftType)) {
                        error(varExpr, ts.Diagnostics.The_left_hand_side_of_a_for_in_statement_must_be_of_type_string_or_any);
                    }
                    else {
                        // run check only former check succeeded to avoid cascading errors
                        checkReferenceExpression(varExpr, ts.Diagnostics.The_left_hand_side_of_a_for_in_statement_must_be_a_variable_or_a_property_access, ts.Diagnostics.The_left_hand_side_of_a_for_in_statement_may_not_be_an_optional_property_access);
                    }
                }
                // unknownType is returned i.e. if node.expression is identifier whose name cannot be resolved
                // in this case error about missing name is already reported - do not report extra one
                if (rightType === neverType || !isTypeAssignableToKind(rightType, 67108864 /* NonPrimitive */ | 58982400 /* InstantiableNonPrimitive */)) {
                    error(node.expression, ts.Diagnostics.The_right_hand_side_of_a_for_in_statement_must_be_of_type_any_an_object_type_or_a_type_parameter_but_here_has_type_0, typeToString(rightType));
                }
                checkSourceElement(node.statement);
                if (node.locals) {
                    registerForUnusedIdentifiersCheck(node);
                }
            }
            function checkForInOrForOfVariableDeclaration(iterationStatement) {
                var variableDeclarationList = iterationStatement.initializer;
                // checkGrammarForInOrForOfStatement will check that there is exactly one declaration.
                if (variableDeclarationList.declarations.length >= 1) {
                    var decl = variableDeclarationList.declarations[0];
                    checkVariableDeclaration(decl);
                }
            }
            function checkRightHandSideOfForOf(statement) {
                var use = statement.awaitModifier ? 15 /* ForAwaitOf */ : 13 /* ForOf */;
                return checkIteratedTypeOrElementType(use, checkNonNullExpression(statement.expression), undefinedType, statement.expression);
            }
            function checkIteratedTypeOrElementType(use, inputType, sentType, errorNode) {
                if (isTypeAny(inputType)) {
                    return inputType;
                }
                return getIteratedTypeOrElementType(use, inputType, sentType, errorNode, /*checkAssignability*/ true) || anyType;
            }
            /**
             * When consuming an iterable type in a for..of, spread, or iterator destructuring assignment
             * we want to get the iterated type of an iterable for ES2015 or later, or the iterated type
             * of a iterable (if defined globally) or element type of an array like for ES2015 or earlier.
             */
            function getIteratedTypeOrElementType(use, inputType, sentType, errorNode, checkAssignability) {
                var allowAsyncIterables = (use & 2 /* AllowsAsyncIterablesFlag */) !== 0;
                if (inputType === neverType) {
                    reportTypeNotIterableError(errorNode, inputType, allowAsyncIterables); // TODO: GH#18217
                    return undefined;
                }
                var uplevelIteration = languageVersion >= 2 /* ES2015 */;
                var downlevelIteration = !uplevelIteration && compilerOptions.downlevelIteration;
                var possibleOutOfBounds = compilerOptions.noUncheckedIndexedAccess && !!(use & 128 /* PossiblyOutOfBounds */);
                // Get the iterated type of an `Iterable<T>` or `IterableIterator<T>` only in ES2015
                // or higher, when inside of an async generator or for-await-if, or when
                // downlevelIteration is requested.
                if (uplevelIteration || downlevelIteration || allowAsyncIterables) {
                    // We only report errors for an invalid iterable type in ES2015 or higher.
                    var iterationTypes = getIterationTypesOfIterable(inputType, use, uplevelIteration ? errorNode : undefined);
                    if (checkAssignability) {
                        if (iterationTypes) {
                            var diagnostic = use & 8 /* ForOfFlag */ ? ts.Diagnostics.Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_for_of_will_always_send_0 :
                                use & 32 /* SpreadFlag */ ? ts.Diagnostics.Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_spread_will_always_send_0 :
                                    use & 64 /* DestructuringFlag */ ? ts.Diagnostics.Cannot_iterate_value_because_the_next_method_of_its_iterator_expects_type_1_but_array_destructuring_will_always_send_0 :
                                        use & 16 /* YieldStarFlag */ ? ts.Diagnostics.Cannot_delegate_iteration_to_value_because_the_next_method_of_its_iterator_expects_type_1_but_the_containing_generator_will_always_send_0 :
                                            undefined;
                            if (diagnostic) {
                                checkTypeAssignableTo(sentType, iterationTypes.nextType, errorNode, diagnostic);
                            }
                        }
                    }
                    if (iterationTypes || uplevelIteration) {
                        return possibleOutOfBounds ? includeUndefinedInIndexSignature(iterationTypes && iterationTypes.yieldType) : (iterationTypes && iterationTypes.yieldType);
                    }
                }
                var arrayType = inputType;
                var reportedError = false;
                var hasStringConstituent = false;
                // If strings are permitted, remove any string-like constituents from the array type.
                // This allows us to find other non-string element types from an array unioned with
                // a string.
                if (use & 4 /* AllowsStringInputFlag */) {
                    if (arrayType.flags & 1048576 /* Union */) {
                        // After we remove all types that are StringLike, we will know if there was a string constituent
                        // based on whether the result of filter is a new array.
                        var arrayTypes = inputType.types;
                        var filteredTypes = ts.filter(arrayTypes, function (t) { return !(t.flags & 402653316 /* StringLike */); });
                        if (filteredTypes !== arrayTypes) {
                            arrayType = getUnionType(filteredTypes, 2 /* Subtype */);
                        }
                    }
                    else if (arrayType.flags & 402653316 /* StringLike */) {
                        arrayType = neverType;
                    }
                    hasStringConstituent = arrayType !== inputType;
                    if (hasStringConstituent) {
                        if (languageVersion < 1 /* ES5 */) {
                            if (errorNode) {
                                error(errorNode, ts.Diagnostics.Using_a_string_in_a_for_of_statement_is_only_supported_in_ECMAScript_5_and_higher);
                                reportedError = true;
                            }
                        }
                        // Now that we've removed all the StringLike types, if no constituents remain, then the entire
                        // arrayOrStringType was a string.
                        if (arrayType.flags & 131072 /* Never */) {
                            return possibleOutOfBounds ? includeUndefinedInIndexSignature(stringType) : stringType;
                        }
                    }
                }
                if (!isArrayLikeType(arrayType)) {
                    if (errorNode && !reportedError) {
                        // Which error we report depends on whether we allow strings or if there was a
                        // string constituent. For example, if the input type is number | string, we
                        // want to say that number is not an array type. But if the input was just
                        // number and string input is allowed, we want to say that number is not an
                        // array type or a string type.
                        var allowsStrings = !!(use & 4 /* AllowsStringInputFlag */) && !hasStringConstituent;
                        var _a = getIterationDiagnosticDetails(allowsStrings, downlevelIteration), defaultDiagnostic = _a[0], maybeMissingAwait = _a[1];
                        errorAndMaybeSuggestAwait(errorNode, maybeMissingAwait && !!getAwaitedTypeOfPromise(arrayType), defaultDiagnostic, typeToString(arrayType));
                    }
                    return hasStringConstituent ? possibleOutOfBounds ? includeUndefinedInIndexSignature(stringType) : stringType : undefined;
                }
                var arrayElementType = getIndexTypeOfType(arrayType, numberType);
                if (hasStringConstituent && arrayElementType) {
                    // This is just an optimization for the case where arrayOrStringType is string | string[]
                    if (arrayElementType.flags & 402653316 /* StringLike */ && !compilerOptions.noUncheckedIndexedAccess) {
                        return stringType;
                    }
                    return getUnionType(possibleOutOfBounds ? [arrayElementType, stringType, undefinedType] : [arrayElementType, stringType], 2 /* Subtype */);
                }
                return (use & 128 /* PossiblyOutOfBounds */) ? includeUndefinedInIndexSignature(arrayElementType) : arrayElementType;
                function getIterationDiagnosticDetails(allowsStrings, downlevelIteration) {
                    var _a;
                    if (downlevelIteration) {
                        return allowsStrings
                            ? [ts.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator, true]
                            : [ts.Diagnostics.Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator, true];
                    }
                    var yieldType = getIterationTypeOfIterable(use, 0 /* Yield */, inputType, /*errorNode*/ undefined);
                    if (yieldType) {
                        return [ts.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators, false];
                    }
                    if (isES2015OrLaterIterable((_a = inputType.symbol) === null || _a === void 0 ? void 0 : _a.escapedName)) {
                        return [ts.Diagnostics.Type_0_can_only_be_iterated_through_when_using_the_downlevelIteration_flag_or_with_a_target_of_es2015_or_higher, true];
                    }
                    return allowsStrings
                        ? [ts.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type, true]
                        : [ts.Diagnostics.Type_0_is_not_an_array_type, true];
                }
            }
            function isES2015OrLaterIterable(n) {
                switch (n) {
                    case "Float32Array":
                    case "Float64Array":
                    case "Int16Array":
                    case "Int32Array":
                    case "Int8Array":
                    case "NodeList":
                    case "Uint16Array":
                    case "Uint32Array":
                    case "Uint8Array":
                    case "Uint8ClampedArray":
                        return true;
                }
                return false;
            }
            /**
             * Gets the requested "iteration type" from an `Iterable`-like or `AsyncIterable`-like type.
             */
            function getIterationTypeOfIterable(use, typeKind, inputType, errorNode) {
                if (isTypeAny(inputType)) {
                    return undefined;
                }
                var iterationTypes = getIterationTypesOfIterable(inputType, use, errorNode);
                return iterationTypes && iterationTypes[getIterationTypesKeyFromIterationTypeKind(typeKind)];
            }
            function createIterationTypes(yieldType, returnType, nextType) {
                // `yieldType` and `returnType` are defaulted to `neverType` they each will be combined
                // via `getUnionType` when merging iteration types. `nextType` is defined as `unknownType`
                // as it is combined via `getIntersectionType` when merging iteration types.
                if (yieldType === void 0) {
                    yieldType = neverType;
                }
                if (returnType === void 0) {
                    returnType = neverType;
                }
                if (nextType === void 0) {
                    nextType = unknownType;
                }
                // Use the cache only for intrinsic types to keep it small as they are likely to be
                // more frequently created (i.e. `Iterator<number, void, unknown>`). Iteration types
                // are also cached on the type they are requested for, so we shouldn't need to maintain
                // the cache for less-frequently used types.
                if (yieldType.flags & 67359327 /* Intrinsic */ &&
                    returnType.flags & (1 /* Any */ | 131072 /* Never */ | 2 /* Unknown */ | 16384 /* Void */ | 32768 /* Undefined */) &&
                    nextType.flags & (1 /* Any */ | 131072 /* Never */ | 2 /* Unknown */ | 16384 /* Void */ | 32768 /* Undefined */)) {
                    var id = getTypeListId([yieldType, returnType, nextType]);
                    var iterationTypes = iterationTypesCache.get(id);
                    if (!iterationTypes) {
                        iterationTypes = { yieldType: yieldType, returnType: returnType, nextType: nextType };
                        iterationTypesCache.set(id, iterationTypes);
                    }
                    return iterationTypes;
                }
                return { yieldType: yieldType, returnType: returnType, nextType: nextType };
            }
            /**
             * Combines multiple `IterationTypes` records.
             *
             * If `array` is empty or all elements are missing or are references to `noIterationTypes`,
             * then `noIterationTypes` is returned. Otherwise, an `IterationTypes` record is returned
             * for the combined iteration types.
             */
            function combineIterationTypes(array) {
                var yieldTypes;
                var returnTypes;
                var nextTypes;
                for (var _i = 0, array_11 = array; _i < array_11.length; _i++) {
                    var iterationTypes = array_11[_i];
                    if (iterationTypes === undefined || iterationTypes === noIterationTypes) {
                        continue;
                    }
                    if (iterationTypes === anyIterationTypes) {
                        return anyIterationTypes;
                    }
                    yieldTypes = ts.append(yieldTypes, iterationTypes.yieldType);
                    returnTypes = ts.append(returnTypes, iterationTypes.returnType);
                    nextTypes = ts.append(nextTypes, iterationTypes.nextType);
                }
                if (yieldTypes || returnTypes || nextTypes) {
                    return createIterationTypes(yieldTypes && getUnionType(yieldTypes), returnTypes && getUnionType(returnTypes), nextTypes && getIntersectionType(nextTypes));
                }
                return noIterationTypes;
            }
            function getCachedIterationTypes(type, cacheKey) {
                return type[cacheKey];
            }
            function setCachedIterationTypes(type, cacheKey, cachedTypes) {
                return type[cacheKey] = cachedTypes;
            }
            /**
             * Gets the *yield*, *return*, and *next* types from an `Iterable`-like or `AsyncIterable`-like type.
             *
             * At every level that involves analyzing return types of signatures, we union the return types of all the signatures.
             *
             * Another thing to note is that at any step of this process, we could run into a dead end,
             * meaning either the property is missing, or we run into the anyType. If either of these things
             * happens, we return `undefined` to signal that we could not find the iteration type. If a property
             * is missing, and the previous step did not result in `any`, then we also give an error if the
             * caller requested it. Then the caller can decide what to do in the case where there is no iterated
             * type.
             *
             * For a **for-of** statement, `yield*` (in a normal generator), spread, array
             * destructuring, or normal generator we will only ever look for a `[Symbol.iterator]()`
             * method.
             *
             * For an async generator we will only ever look at the `[Symbol.asyncIterator]()` method.
             *
             * For a **for-await-of** statement or a `yield*` in an async generator we will look for
             * the `[Symbol.asyncIterator]()` method first, and then the `[Symbol.iterator]()` method.
             */
            function getIterationTypesOfIterable(type, use, errorNode) {
                if (isTypeAny(type)) {
                    return anyIterationTypes;
                }
                if (!(type.flags & 1048576 /* Union */)) {
                    var iterationTypes_1 = getIterationTypesOfIterableWorker(type, use, errorNode);
                    if (iterationTypes_1 === noIterationTypes) {
                        if (errorNode) {
                            reportTypeNotIterableError(errorNode, type, !!(use & 2 /* AllowsAsyncIterablesFlag */));
                        }
                        return undefined;
                    }
                    return iterationTypes_1;
                }
                var cacheKey = use & 2 /* AllowsAsyncIterablesFlag */ ? "iterationTypesOfAsyncIterable" : "iterationTypesOfIterable";
                var cachedTypes = getCachedIterationTypes(type, cacheKey);
                if (cachedTypes)
                    return cachedTypes === noIterationTypes ? undefined : cachedTypes;
                var allIterationTypes;
                for (var _i = 0, _a = type.types; _i < _a.length; _i++) {
                    var constituent = _a[_i];
                    var iterationTypes_2 = getIterationTypesOfIterableWorker(constituent, use, errorNode);
                    if (iterationTypes_2 === noIterationTypes) {
                        if (errorNode) {
                            reportTypeNotIterableError(errorNode, type, !!(use & 2 /* AllowsAsyncIterablesFlag */));
                        }
                        setCachedIterationTypes(type, cacheKey, noIterationTypes);
                        return undefined;
                    }
                    else {
                        allIterationTypes = ts.append(allIterationTypes, iterationTypes_2);
                    }
                }
                var iterationTypes = allIterationTypes ? combineIterationTypes(allIterationTypes) : noIterationTypes;
                setCachedIterationTypes(type, cacheKey, iterationTypes);
                return iterationTypes === noIterationTypes ? undefined : iterationTypes;
            }
            function getAsyncFromSyncIterationTypes(iterationTypes, errorNode) {
                if (iterationTypes === noIterationTypes)
                    return noIterationTypes;
                if (iterationTypes === anyIterationTypes)
                    return anyIterationTypes;
                var yieldType = iterationTypes.yieldType, returnType = iterationTypes.returnType, nextType = iterationTypes.nextType;
                return createIterationTypes(getAwaitedType(yieldType, errorNode) || anyType, getAwaitedType(returnType, errorNode) || anyType, nextType);
            }
            /**
             * Gets the *yield*, *return*, and *next* types from a non-union type.
             *
             * If we are unable to find the *yield*, *return*, and *next* types, `noIterationTypes` is
             * returned to indicate to the caller that it should report an error. Otherwise, an
             * `IterationTypes` record is returned.
             *
             * NOTE: You probably don't want to call this directly and should be calling
             * `getIterationTypesOfIterable` instead.
             */
            function getIterationTypesOfIterableWorker(type, use, errorNode) {
                if (isTypeAny(type)) {
                    return anyIterationTypes;
                }
                if (use & 2 /* AllowsAsyncIterablesFlag */) {
                    var iterationTypes = getIterationTypesOfIterableCached(type, asyncIterationTypesResolver) ||
                        getIterationTypesOfIterableFast(type, asyncIterationTypesResolver);
                    if (iterationTypes) {
                        return iterationTypes;
                    }
                }
                if (use & 1 /* AllowsSyncIterablesFlag */) {
                    var iterationTypes = getIterationTypesOfIterableCached(type, syncIterationTypesResolver) ||
                        getIterationTypesOfIterableFast(type, syncIterationTypesResolver);
                    if (iterationTypes) {
                        if (use & 2 /* AllowsAsyncIterablesFlag */) {
                            // for a sync iterable in an async context, only use the cached types if they are valid.
                            if (iterationTypes !== noIterationTypes) {
                                return setCachedIterationTypes(type, "iterationTypesOfAsyncIterable", getAsyncFromSyncIterationTypes(iterationTypes, errorNode));
                            }
                        }
                        else {
                            return iterationTypes;
                        }
                    }
                }
                if (use & 2 /* AllowsAsyncIterablesFlag */) {
                    var iterationTypes = getIterationTypesOfIterableSlow(type, asyncIterationTypesResolver, errorNode);
                    if (iterationTypes !== noIterationTypes) {
                        return iterationTypes;
                    }
                }
                if (use & 1 /* AllowsSyncIterablesFlag */) {
                    var iterationTypes = getIterationTypesOfIterableSlow(type, syncIterationTypesResolver, errorNode);
                    if (iterationTypes !== noIterationTypes) {
                        if (use & 2 /* AllowsAsyncIterablesFlag */) {
                            return setCachedIterationTypes(type, "iterationTypesOfAsyncIterable", iterationTypes
                                ? getAsyncFromSyncIterationTypes(iterationTypes, errorNode)
                                : noIterationTypes);
                        }
                        else {
                            return iterationTypes;
                        }
                    }
                }
                return noIterationTypes;
            }
            /**
             * Gets the *yield*, *return*, and *next* types of an `Iterable`-like or
             * `AsyncIterable`-like type from the cache.
             *
             * NOTE: You probably don't want to call this directly and should be calling
             * `getIterationTypesOfIterable` instead.
             */
            function getIterationTypesOfIterableCached(type, resolver) {
                return getCachedIterationTypes(type, resolver.iterableCacheKey);
            }
            function getIterationTypesOfGlobalIterableType(globalType, resolver) {
                var globalIterationTypes = getIterationTypesOfIterableCached(globalType, resolver) ||
                    getIterationTypesOfIterableSlow(globalType, resolver, /*errorNode*/ undefined);
                return globalIterationTypes === noIterationTypes ? defaultIterationTypes : globalIterationTypes;
            }
            /**
             * Gets the *yield*, *return*, and *next* types of an `Iterable`-like or `AsyncIterable`-like
             * type from from common heuristics.
             *
             * If we previously analyzed this type and found no iteration types, `noIterationTypes` is
             * returned. If we found iteration types, an `IterationTypes` record is returned.
             * Otherwise, we return `undefined` to indicate to the caller it should perform a more
             * exhaustive analysis.
             *
             * NOTE: You probably don't want to call this directly and should be calling
             * `getIterationTypesOfIterable` instead.
             */
            function getIterationTypesOfIterableFast(type, resolver) {
                // As an optimization, if the type is an instantiation of one of the following global types, then
                // just grab its related type argument:
                // - `Iterable<T>` or `AsyncIterable<T>`
                // - `IterableIterator<T>` or `AsyncIterableIterator<T>`
                var globalType;
                if (isReferenceToType(type, globalType = resolver.getGlobalIterableType(/*reportErrors*/ false)) ||
                    isReferenceToType(type, globalType = resolver.getGlobalIterableIteratorType(/*reportErrors*/ false))) {
                    var yieldType = getTypeArguments(type)[0];
                    // The "return" and "next" types of `Iterable` and `IterableIterator` are defined by the
                    // iteration types of their `[Symbol.iterator]()` method. The same is true for their async cousins.
                    // While we define these as `any` and `undefined` in our libs by default, a custom lib *could* use
                    // different definitions.
                    var _a = getIterationTypesOfGlobalIterableType(globalType, resolver), returnType = _a.returnType, nextType = _a.nextType;
                    return setCachedIterationTypes(type, resolver.iterableCacheKey, createIterationTypes(yieldType, returnType, nextType));
                }
                // As an optimization, if the type is an instantiation of the following global type, then
                // just grab its related type arguments:
                // - `Generator<T, TReturn, TNext>` or `AsyncGenerator<T, TReturn, TNext>`
                if (isReferenceToType(type, resolver.getGlobalGeneratorType(/*reportErrors*/ false))) {
                    var _b = getTypeArguments(type), yieldType = _b[0], returnType = _b[1], nextType = _b[2];
                    return setCachedIterationTypes(type, resolver.iterableCacheKey, createIterationTypes(yieldType, returnType, nextType));
                }
            }
            function getPropertyNameForKnownSymbolName(symbolName) {
                var ctorType = getGlobalESSymbolConstructorSymbol(/*reportErrors*/ false);
                var uniqueType = ctorType && getTypeOfPropertyOfType(getTypeOfSymbol(ctorType), ts.escapeLeadingUnderscores(symbolName));
                return uniqueType && isTypeUsableAsPropertyName(uniqueType) ? getPropertyNameFromType(uniqueType) : "__@" + symbolName;
            }
            /**
             * Gets the *yield*, *return*, and *next* types of an `Iterable`-like or `AsyncIterable`-like
             * type from its members.
             *
             * If we successfully found the *yield*, *return*, and *next* types, an `IterationTypes`
             * record is returned. Otherwise, `noIterationTypes` is returned.
             *
             * NOTE: You probably don't want to call this directly and should be calling
             * `getIterationTypesOfIterable` instead.
             */
            function getIterationTypesOfIterableSlow(type, resolver, errorNode) {
                var _a;
                var method = getPropertyOfType(type, getPropertyNameForKnownSymbolName(resolver.iteratorSymbolName));
                var methodType = method && !(method.flags & 16777216 /* Optional */) ? getTypeOfSymbol(method) : undefined;
                if (isTypeAny(methodType)) {
                    return setCachedIterationTypes(type, resolver.iterableCacheKey, anyIterationTypes);
                }
                var signatures = methodType ? getSignaturesOfType(methodType, 0 /* Call */) : undefined;
                if (!ts.some(signatures)) {
                    return setCachedIterationTypes(type, resolver.iterableCacheKey, noIterationTypes);
                }
                var iteratorType = getIntersectionType(ts.map(signatures, getReturnTypeOfSignature));
                var iterationTypes = (_a = getIterationTypesOfIterator(iteratorType, resolver, errorNode)) !== null && _a !== void 0 ? _a : noIterationTypes;
                return setCachedIterationTypes(type, resolver.iterableCacheKey, iterationTypes);
            }
            function reportTypeNotIterableError(errorNode, type, allowAsyncIterables) {
                var message = allowAsyncIterables
                    ? ts.Diagnostics.Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator
                    : ts.Diagnostics.Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator;
                errorAndMaybeSuggestAwait(errorNode, !!getAwaitedTypeOfPromise(type), message, typeToString(type));
            }
            /**
             * Gets the *yield*, *return*, and *next* types from an `Iterator`-like or `AsyncIterator`-like type.
             *
             * If we successfully found the *yield*, *return*, and *next* types, an `IterationTypes`
             * record is returned. Otherwise, `undefined` is returned.
             */
            function getIterationTypesOfIterator(type, resolver, errorNode) {
                if (isTypeAny(type)) {
                    return anyIterationTypes;
                }
                var iterationTypes = getIterationTypesOfIteratorCached(type, resolver) ||
                    getIterationTypesOfIteratorFast(type, resolver) ||
                    getIterationTypesOfIteratorSlow(type, resolver, errorNode);
                return iterationTypes === noIterationTypes ? undefined : iterationTypes;
            }
            /**
             * Gets the iteration types of an `Iterator`-like or `AsyncIterator`-like type from the
             * cache.
             *
             * NOTE: You probably don't want to call this directly and should be calling
             * `getIterationTypesOfIterator` instead.
             */
            function getIterationTypesOfIteratorCached(type, resolver) {
                return getCachedIterationTypes(type, resolver.iteratorCacheKey);
            }
            /**
             * Gets the iteration types of an `Iterator`-like or `AsyncIterator`-like type from the
             * cache or from common heuristics.
             *
             * If we previously analyzed this type and found no iteration types, `noIterationTypes` is
             * returned. If we found iteration types, an `IterationTypes` record is returned.
             * Otherwise, we return `undefined` to indicate to the caller it should perform a more
             * exhaustive analysis.
             *
             * NOTE: You probably don't want to call this directly and should be calling
             * `getIterationTypesOfIterator` instead.
             */
            function getIterationTypesOfIteratorFast(type, resolver) {
                // As an optimization, if the type is an instantiation of one of the following global types,
                // then just grab its related type argument:
                // - `IterableIterator<T>` or `AsyncIterableIterator<T>`
                // - `Iterator<T, TReturn, TNext>` or `AsyncIterator<T, TReturn, TNext>`
                // - `Generator<T, TReturn, TNext>` or `AsyncGenerator<T, TReturn, TNext>`
                var globalType = resolver.getGlobalIterableIteratorType(/*reportErrors*/ false);
                if (isReferenceToType(type, globalType)) {
                    var yieldType = getTypeArguments(type)[0];
                    // The "return" and "next" types of `IterableIterator` and `AsyncIterableIterator` are defined by the
                    // iteration types of their `next`, `return`, and `throw` methods. While we define these as `any`
                    // and `undefined` in our libs by default, a custom lib *could* use different definitions.
                    var globalIterationTypes = getIterationTypesOfIteratorCached(globalType, resolver) ||
                        getIterationTypesOfIteratorSlow(globalType, resolver, /*errorNode*/ undefined);
                    var _a = globalIterationTypes === noIterationTypes ? defaultIterationTypes : globalIterationTypes, returnType = _a.returnType, nextType = _a.nextType;
                    return setCachedIterationTypes(type, resolver.iteratorCacheKey, createIterationTypes(yieldType, returnType, nextType));
                }
                if (isReferenceToType(type, resolver.getGlobalIteratorType(/*reportErrors*/ false)) ||
                    isReferenceToType(type, resolver.getGlobalGeneratorType(/*reportErrors*/ false))) {
                    var _b = getTypeArguments(type), yieldType = _b[0], returnType = _b[1], nextType = _b[2];
                    return setCachedIterationTypes(type, resolver.iteratorCacheKey, createIterationTypes(yieldType, returnType, nextType));
                }
            }
            function isIteratorResult(type, kind) {
                // From https://tc39.github.io/ecma262/#sec-iteratorresult-interface:
                // > [done] is the result status of an iterator `next` method call. If the end of the iterator was reached `done` is `true`.
                // > If the end was not reached `done` is `false` and a value is available.
                // > If a `done` property (either own or inherited) does not exist, it is consider to have the value `false`.
                var doneType = getTypeOfPropertyOfType(type, "done") || falseType;
                return isTypeAssignableTo(kind === 0 /* Yield */ ? falseType : trueType, doneType);
            }
            function isYieldIteratorResult(type) {
                return isIteratorResult(type, 0 /* Yield */);
            }
            function isReturnIteratorResult(type) {
                return isIteratorResult(type, 1 /* Return */);
            }
            /**
             * Gets the *yield* and *return* types of an `IteratorResult`-like type.
             *
             * If we are unable to determine a *yield* or a *return* type, `noIterationTypes` is
             * returned to indicate to the caller that it should handle the error. Otherwise, an
             * `IterationTypes` record is returned.
             */
            function getIterationTypesOfIteratorResult(type) {
                if (isTypeAny(type)) {
                    return anyIterationTypes;
                }
                var cachedTypes = getCachedIterationTypes(type, "iterationTypesOfIteratorResult");
                if (cachedTypes) {
                    return cachedTypes;
                }
                // As an optimization, if the type is an instantiation of one of the global `IteratorYieldResult<T>`
                // or `IteratorReturnResult<TReturn>` types, then just grab its type argument.
                if (isReferenceToType(type, getGlobalIteratorYieldResultType(/*reportErrors*/ false))) {
                    var yieldType_1 = getTypeArguments(type)[0];
                    return setCachedIterationTypes(type, "iterationTypesOfIteratorResult", createIterationTypes(yieldType_1, /*returnType*/ undefined, /*nextType*/ undefined));
                }
                if (isReferenceToType(type, getGlobalIteratorReturnResultType(/*reportErrors*/ false))) {
                    var returnType_1 = getTypeArguments(type)[0];
                    return setCachedIterationTypes(type, "iterationTypesOfIteratorResult", createIterationTypes(/*yieldType*/ undefined, returnType_1, /*nextType*/ undefined));
                }
                // Choose any constituents that can produce the requested iteration type.
                var yieldIteratorResult = filterType(type, isYieldIteratorResult);
                var yieldType = yieldIteratorResult !== neverType ? getTypeOfPropertyOfType(yieldIteratorResult, "value") : undefined;
                var returnIteratorResult = filterType(type, isReturnIteratorResult);
                var returnType = returnIteratorResult !== neverType ? getTypeOfPropertyOfType(returnIteratorResult, "value") : undefined;
                if (!yieldType && !returnType) {
                    return setCachedIterationTypes(type, "iterationTypesOfIteratorResult", noIterationTypes);
                }
                // From https://tc39.github.io/ecma262/#sec-iteratorresult-interface
                // > ... If the iterator does not have a return value, `value` is `undefined`. In that case, the
                // > `value` property may be absent from the conforming object if it does not inherit an explicit
                // > `value` property.
                return setCachedIterationTypes(type, "iterationTypesOfIteratorResult", createIterationTypes(yieldType, returnType || voidType, /*nextType*/ undefined));
            }
            /**
             * Gets the *yield*, *return*, and *next* types of a the `next()`, `return()`, or
             * `throw()` method of an `Iterator`-like or `AsyncIterator`-like type.
             *
             * If we successfully found the *yield*, *return*, and *next* types, an `IterationTypes`
             * record is returned. Otherwise, we return `undefined`.
             */
            function getIterationTypesOfMethod(type, resolver, methodName, errorNode) {
                var _a, _b, _c, _d;
                var method = getPropertyOfType(type, methodName);
                // Ignore 'return' or 'throw' if they are missing.
                if (!method && methodName !== "next") {
                    return undefined;
                }
                var methodType = method && !(methodName === "next" && (method.flags & 16777216 /* Optional */))
                    ? methodName === "next" ? getTypeOfSymbol(method) : getTypeWithFacts(getTypeOfSymbol(method), 2097152 /* NEUndefinedOrNull */)
                    : undefined;
                if (isTypeAny(methodType)) {
                    // `return()` and `throw()` don't provide a *next* type.
                    return methodName === "next" ? anyIterationTypes : anyIterationTypesExceptNext;
                }
                // Both async and non-async iterators *must* have a `next` method.
                var methodSignatures = methodType ? getSignaturesOfType(methodType, 0 /* Call */) : ts.emptyArray;
                if (methodSignatures.length === 0) {
                    if (errorNode) {
                        var diagnostic = methodName === "next"
                            ? resolver.mustHaveANextMethodDiagnostic
                            : resolver.mustBeAMethodDiagnostic;
                        error(errorNode, diagnostic, methodName);
                    }
                    return methodName === "next" ? anyIterationTypes : undefined;
                }
                // If the method signature comes exclusively from the global iterator or generator type,
                // create iteration types from its type arguments like `getIterationTypesOfIteratorFast`
                // does (so as to remove `undefined` from the next and return types). We arrive here when
                // a contextual type for a generator was not a direct reference to one of those global types,
                // but looking up `methodType` referred to one of them (and nothing else). E.g., in
                // `interface SpecialIterator extends Iterator<number> {}`, `SpecialIterator` is not a
                // reference to `Iterator`, but its `next` member derives exclusively from `Iterator`.
                if ((methodType === null || methodType === void 0 ? void 0 : methodType.symbol) && methodSignatures.length === 1) {
                    var globalGeneratorType = resolver.getGlobalGeneratorType(/*reportErrors*/ false);
                    var globalIteratorType = resolver.getGlobalIteratorType(/*reportErrors*/ false);
                    var isGeneratorMethod = ((_b = (_a = globalGeneratorType.symbol) === null || _a === void 0 ? void 0 : _a.members) === null || _b === void 0 ? void 0 : _b.get(methodName)) === methodType.symbol;
                    var isIteratorMethod = !isGeneratorMethod && ((_d = (_c = globalIteratorType.symbol) === null || _c === void 0 ? void 0 : _c.members) === null || _d === void 0 ? void 0 : _d.get(methodName)) === methodType.symbol;
                    if (isGeneratorMethod || isIteratorMethod) {
                        var globalType = isGeneratorMethod ? globalGeneratorType : globalIteratorType;
                        var mapper = methodType.mapper;
                        return createIterationTypes(getMappedType(globalType.typeParameters[0], mapper), getMappedType(globalType.typeParameters[1], mapper), methodName === "next" ? getMappedType(globalType.typeParameters[2], mapper) : undefined);
                    }
                }
                // Extract the first parameter and return type of each signature.
                var methodParameterTypes;
                var methodReturnTypes;
                for (var _i = 0, methodSignatures_1 = methodSignatures; _i < methodSignatures_1.length; _i++) {
                    var signature = methodSignatures_1[_i];
                    if (methodName !== "throw" && ts.some(signature.parameters)) {
                        methodParameterTypes = ts.append(methodParameterTypes, getTypeAtPosition(signature, 0));
                    }
                    methodReturnTypes = ts.append(methodReturnTypes, getReturnTypeOfSignature(signature));
                }
                // Resolve the *next* or *return* type from the first parameter of a `next()` or
                // `return()` method, respectively.
                var returnTypes;
                var nextType;
                if (methodName !== "throw") {
                    var methodParameterType = methodParameterTypes ? getUnionType(methodParameterTypes) : unknownType;
                    if (methodName === "next") {
                        // The value of `next(value)` is *not* awaited by async generators
                        nextType = methodParameterType;
                    }
                    else if (methodName === "return") {
                        // The value of `return(value)` *is* awaited by async generators
                        var resolvedMethodParameterType = resolver.resolveIterationType(methodParameterType, errorNode) || anyType;
                        returnTypes = ts.append(returnTypes, resolvedMethodParameterType);
                    }
                }
                // Resolve the *yield* and *return* types from the return type of the method (i.e. `IteratorResult`)
                var yieldType;
                var methodReturnType = methodReturnTypes ? getIntersectionType(methodReturnTypes) : neverType;
                var resolvedMethodReturnType = resolver.resolveIterationType(methodReturnType, errorNode) || anyType;
                var iterationTypes = getIterationTypesOfIteratorResult(resolvedMethodReturnType);
                if (iterationTypes === noIterationTypes) {
                    if (errorNode) {
                        error(errorNode, resolver.mustHaveAValueDiagnostic, methodName);
                    }
                    yieldType = anyType;
                    returnTypes = ts.append(returnTypes, anyType);
                }
                else {
                    yieldType = iterationTypes.yieldType;
                    returnTypes = ts.append(returnTypes, iterationTypes.returnType);
                }
                return createIterationTypes(yieldType, getUnionType(returnTypes), nextType);
            }
            /**
             * Gets the *yield*, *return*, and *next* types of an `Iterator`-like or `AsyncIterator`-like
             * type from its members.
             *
             * If we successfully found the *yield*, *return*, and *next* types, an `IterationTypes`
             * record is returned. Otherwise, `noIterationTypes` is returned.
             *
             * NOTE: You probably don't want to call this directly and should be calling
             * `getIterationTypesOfIterator` instead.
             */
            function getIterationTypesOfIteratorSlow(type, resolver, errorNode) {
                var iterationTypes = combineIterationTypes([
                    getIterationTypesOfMethod(type, resolver, "next", errorNode),
                    getIterationTypesOfMethod(type, resolver, "return", errorNode),
                    getIterationTypesOfMethod(type, resolver, "throw", errorNode),
                ]);
                return setCachedIterationTypes(type, resolver.iteratorCacheKey, iterationTypes);
            }
            /**
             * Gets the requested "iteration type" from a type that is either `Iterable`-like, `Iterator`-like,
             * `IterableIterator`-like, or `Generator`-like (for a non-async generator); or `AsyncIterable`-like,
             * `AsyncIterator`-like, `AsyncIterableIterator`-like, or `AsyncGenerator`-like (for an async generator).
             */
            function getIterationTypeOfGeneratorFunctionReturnType(kind, returnType, isAsyncGenerator) {
                if (isTypeAny(returnType)) {
                    return undefined;
                }
                var iterationTypes = getIterationTypesOfGeneratorFunctionReturnType(returnType, isAsyncGenerator);
                return iterationTypes && iterationTypes[getIterationTypesKeyFromIterationTypeKind(kind)];
            }
            function getIterationTypesOfGeneratorFunctionReturnType(type, isAsyncGenerator) {
                if (isTypeAny(type)) {
                    return anyIterationTypes;
                }
                var use = isAsyncGenerator ? 2 /* AsyncGeneratorReturnType */ : 1 /* GeneratorReturnType */;
                var resolver = isAsyncGenerator ? asyncIterationTypesResolver : syncIterationTypesResolver;
                return getIterationTypesOfIterable(type, use, /*errorNode*/ undefined) ||
                    getIterationTypesOfIterator(type, resolver, /*errorNode*/ undefined);
            }
            function checkBreakOrContinueStatement(node) {
                // Grammar checking
                if (!checkGrammarStatementInAmbientContext(node))
                    checkGrammarBreakOrContinueStatement(node);
                // TODO: Check that target label is valid
            }
            function unwrapReturnType(returnType, functionFlags) {
                var _a, _b;
                var isGenerator = !!(functionFlags & 1 /* Generator */);
                var isAsync = !!(functionFlags & 2 /* Async */);
                return isGenerator ? (_a = getIterationTypeOfGeneratorFunctionReturnType(1 /* Return */, returnType, isAsync)) !== null && _a !== void 0 ? _a : errorType :
                    isAsync ? (_b = getAwaitedType(returnType)) !== null && _b !== void 0 ? _b : errorType :
                        returnType;
            }
            function isUnwrappedReturnTypeVoidOrAny(func, returnType) {
                var unwrappedReturnType = unwrapReturnType(returnType, ts.getFunctionFlags(func));
                return !!unwrappedReturnType && maybeTypeOfKind(unwrappedReturnType, 16384 /* Void */ | 3 /* AnyOrUnknown */);
            }
            function checkReturnStatement(node) {
                var _a;
                // Grammar checking
                if (checkGrammarStatementInAmbientContext(node)) {
                    return;
                }
                var container = ts.getContainingFunctionOrClassStaticBlock(node);
                if (container && ts.isClassStaticBlockDeclaration(container)) {
                    grammarErrorOnFirstToken(node, ts.Diagnostics.A_return_statement_cannot_be_used_inside_a_class_static_block);
                    return;
                }
                if (!container) {
                    grammarErrorOnFirstToken(node, ts.Diagnostics.A_return_statement_can_only_be_used_within_a_function_body);
                    return;
                }
                var signature = getSignatureFromDeclaration(container);
                var returnType = getReturnTypeOfSignature(signature);
                var functionFlags = ts.getFunctionFlags(container);
                if (strictNullChecks || node.expression || returnType.flags & 131072 /* Never */) {
                    var exprType = node.expression ? checkExpressionCached(node.expression) : undefinedType;
                    if (container.kind === 171 /* SetAccessor */) {
                        if (node.expression) {
                            error(node, ts.Diagnostics.Setters_cannot_return_a_value);
                        }
                    }
                    else if (container.kind === 169 /* Constructor */) {
                        if (node.expression && !checkTypeAssignableToAndOptionallyElaborate(exprType, returnType, node, node.expression)) {
                            error(node, ts.Diagnostics.Return_type_of_constructor_signature_must_be_assignable_to_the_instance_type_of_the_class);
                        }
                    }
                    else if (getReturnTypeFromAnnotation(container)) {
                        var unwrappedReturnType = (_a = unwrapReturnType(returnType, functionFlags)) !== null && _a !== void 0 ? _a : returnType;
                        var unwrappedExprType = functionFlags & 2 /* Async */
                            ? checkAwaitedType(exprType, node, ts.Diagnostics.The_return_type_of_an_async_function_must_either_be_a_valid_promise_or_must_not_contain_a_callable_then_member)
                            : exprType;
                        if (unwrappedReturnType) {
                            // If the function has a return type, but promisedType is
                            // undefined, an error will be reported in checkAsyncFunctionReturnType
                            // so we don't need to report one here.
                            checkTypeAssignableToAndOptionallyElaborate(unwrappedExprType, unwrappedReturnType, node, node.expression);
                        }
                    }
                }
                else if (container.kind !== 169 /* Constructor */ && compilerOptions.noImplicitReturns && !isUnwrappedReturnTypeVoidOrAny(container, returnType)) {
                    // The function has a return type, but the return statement doesn't have an expression.
                    error(node, ts.Diagnostics.Not_all_code_paths_return_a_value);
                }
            }
            function checkWithStatement(node) {
                // Grammar checking for withStatement
                if (!checkGrammarStatementInAmbientContext(node)) {
                    if (node.flags & 32768 /* AwaitContext */) {
                        grammarErrorOnFirstToken(node, ts.Diagnostics.with_statements_are_not_allowed_in_an_async_function_block);
                    }
                }
                checkExpression(node.expression);
                var sourceFile = ts.getSourceFileOfNode(node);
                if (!hasParseDiagnostics(sourceFile)) {
                    var start = ts.getSpanOfTokenAtPosition(sourceFile, node.pos).start;
                    var end = node.statement.pos;
                    grammarErrorAtPos(sourceFile, start, end - start, ts.Diagnostics.The_with_statement_is_not_supported_All_symbols_in_a_with_block_will_have_type_any);
                }
            }
            function checkSwitchStatement(node) {
                // Grammar checking
                checkGrammarStatementInAmbientContext(node);
                var firstDefaultClause;
                var hasDuplicateDefaultClause = false;
                var expressionType = checkExpression(node.expression);
                var expressionIsLiteral = isLiteralType(expressionType);
                ts.forEach(node.caseBlock.clauses, function (clause) {
                    // Grammar check for duplicate default clauses, skip if we already report duplicate default clause
                    if (clause.kind === 288 /* DefaultClause */ && !hasDuplicateDefaultClause) {
                        if (firstDefaultClause === undefined) {
                            firstDefaultClause = clause;
                        }
                        else {
                            grammarErrorOnNode(clause, ts.Diagnostics.A_default_clause_cannot_appear_more_than_once_in_a_switch_statement);
                            hasDuplicateDefaultClause = true;
                        }
                    }
                    if (produceDiagnostics && clause.kind === 287 /* CaseClause */) {
                        // TypeScript 1.0 spec (April 2014): 5.9
                        // In a 'switch' statement, each 'case' expression must be of a type that is comparable
                        // to or from the type of the 'switch' expression.
                        var caseType = checkExpression(clause.expression);
                        var caseIsLiteral = isLiteralType(caseType);
                        var comparedExpressionType = expressionType;
                        if (!caseIsLiteral || !expressionIsLiteral) {
                            caseType = caseIsLiteral ? getBaseTypeOfLiteralType(caseType) : caseType;
                            comparedExpressionType = getBaseTypeOfLiteralType(expressionType);
                        }
                        if (!isTypeEqualityComparableTo(comparedExpressionType, caseType)) {
                            // expressionType is not comparable to caseType, try the reversed check and report errors if it fails
                            checkTypeComparableTo(caseType, comparedExpressionType, clause.expression, /*headMessage*/ undefined);
                        }
                    }
                    ts.forEach(clause.statements, checkSourceElement);
                    if (compilerOptions.noFallthroughCasesInSwitch && clause.fallthroughFlowNode && isReachableFlowNode(clause.fallthroughFlowNode)) {
                        error(clause, ts.Diagnostics.Fallthrough_case_in_switch);
                    }
                });
                if (node.caseBlock.locals) {
                    registerForUnusedIdentifiersCheck(node.caseBlock);
                }
            }
            function checkLabeledStatement(node) {
                // Grammar checking
                if (!checkGrammarStatementInAmbientContext(node)) {
                    ts.findAncestor(node.parent, function (current) {
                        if (ts.isFunctionLike(current)) {
                            return "quit";
                        }
                        if (current.kind === 248 /* LabeledStatement */ && current.label.escapedText === node.label.escapedText) {
                            grammarErrorOnNode(node.label, ts.Diagnostics.Duplicate_label_0, ts.getTextOfNode(node.label));
                            return true;
                        }
                        return false;
                    });
                }
                // ensure that label is unique
                checkSourceElement(node.statement);
            }
            function checkThrowStatement(node) {
                // Grammar checking
                if (!checkGrammarStatementInAmbientContext(node)) {
                    if (ts.isIdentifier(node.expression) && !node.expression.escapedText) {
                        grammarErrorAfterFirstToken(node, ts.Diagnostics.Line_break_not_permitted_here);
                    }
                }
                if (node.expression) {
                    checkExpression(node.expression);
                }
            }
            function checkTryStatement(node) {
                // Grammar checking
                checkGrammarStatementInAmbientContext(node);
                checkBlock(node.tryBlock);
                var catchClause = node.catchClause;
                if (catchClause) {
                    // Grammar checking
                    if (catchClause.variableDeclaration) {
                        var declaration = catchClause.variableDeclaration;
                        var typeNode = ts.getEffectiveTypeAnnotationNode(ts.getRootDeclaration(declaration));
                        if (typeNode) {
                            var type = getTypeForVariableLikeDeclaration(declaration, /*includeOptionality*/ false);
                            if (type && !(type.flags & 3 /* AnyOrUnknown */)) {
                                grammarErrorOnFirstToken(typeNode, ts.Diagnostics.Catch_clause_variable_type_annotation_must_be_any_or_unknown_if_specified);
                            }
                        }
                        else if (declaration.initializer) {
                            grammarErrorOnFirstToken(declaration.initializer, ts.Diagnostics.Catch_clause_variable_cannot_have_an_initializer);
                        }
                        else {
                            var blockLocals_1 = catchClause.block.locals;
                            if (blockLocals_1) {
                                ts.forEachKey(catchClause.locals, function (caughtName) {
                                    var blockLocal = blockLocals_1.get(caughtName);
                                    if ((blockLocal === null || blockLocal === void 0 ? void 0 : blockLocal.valueDeclaration) && (blockLocal.flags & 2 /* BlockScopedVariable */) !== 0) {
                                        grammarErrorOnNode(blockLocal.valueDeclaration, ts.Diagnostics.Cannot_redeclare_identifier_0_in_catch_clause, caughtName);
                                    }
                                });
                            }
                        }
                    }
                    checkBlock(catchClause.block);
                }
                if (node.finallyBlock) {
                    checkBlock(node.finallyBlock);
                }
            }
            function checkIndexConstraints(type, isStaticIndex) {
                var indexInfos = getIndexInfosOfType(type);
                if (indexInfos.length === 0) {
                    return;
                }
                for (var _i = 0, _a = getPropertiesOfObjectType(type); _i < _a.length; _i++) {
                    var prop = _a[_i];
                    if (!(isStaticIndex && prop.flags & 4194304 /* Prototype */)) {
                        checkIndexConstraintForProperty(type, prop, getLiteralTypeFromProperty(prop, 8576 /* StringOrNumberLiteralOrUnique */, /*includeNonPublic*/ true), getNonMissingTypeOfSymbol(prop));
                    }
                }
                var typeDeclaration = type.symbol.valueDeclaration;
                if (typeDeclaration && ts.isClassLike(typeDeclaration)) {
                    for (var _b = 0, _c = typeDeclaration.members; _b < _c.length; _b++) {
                        var member = _c[_b];
                        // Only process instance properties with computed names here. Static properties cannot be in conflict with indexers,
                        // and properties with literal names were already checked.
                        if (!ts.isStatic(member) && !hasBindableName(member)) {
                            var symbol = getSymbolOfNode(member);
                            checkIndexConstraintForProperty(type, symbol, getTypeOfExpression(member.name.expression), getNonMissingTypeOfSymbol(symbol));
                        }
                    }
                }
                if (indexInfos.length > 1) {
                    for (var _d = 0, indexInfos_6 = indexInfos; _d < indexInfos_6.length; _d++) {
                        var info = indexInfos_6[_d];
                        checkIndexConstraintForIndexSignature(type, info);
                    }
                }
            }
            function checkIndexConstraintForProperty(type, prop, propNameType, propType) {
                var declaration = prop.valueDeclaration;
                var name = ts.getNameOfDeclaration(declaration);
                if (name && ts.isPrivateIdentifier(name)) {
                    return;
                }
                var indexInfos = getApplicableIndexInfos(type, propNameType);
                var interfaceDeclaration = ts.getObjectFlags(type) & 2 /* Interface */ ? ts.getDeclarationOfKind(type.symbol, 256 /* InterfaceDeclaration */) : undefined;
                var localPropDeclaration = declaration && declaration.kind === 219 /* BinaryExpression */ ||
                    name && name.kind === 160 /* ComputedPropertyName */ || getParentOfSymbol(prop) === type.symbol ? declaration : undefined;
                var _loop_27 = function (info) {
                    var localIndexDeclaration = info.declaration && getParentOfSymbol(getSymbolOfNode(info.declaration)) === type.symbol ? info.declaration : undefined;
                    // We check only when (a) the property is declared in the containing type, or (b) the applicable index signature is declared
                    // in the containing type, or (c) the containing type is an interface and no base interface contains both the property and
                    // the index signature (i.e. property and index signature are declared in separate inherited interfaces).
                    var errorNode = localPropDeclaration || localIndexDeclaration ||
                        (interfaceDeclaration && !ts.some(getBaseTypes(type), function (base) { return !!getPropertyOfObjectType(base, prop.escapedName) && !!getIndexTypeOfType(base, info.keyType); }) ? interfaceDeclaration : undefined);
                    if (errorNode && !isTypeAssignableTo(propType, info.type)) {
                        error(errorNode, ts.Diagnostics.Property_0_of_type_1_is_not_assignable_to_2_index_type_3, symbolToString(prop), typeToString(propType), typeToString(info.keyType), typeToString(info.type));
                    }
                };
                for (var _i = 0, indexInfos_7 = indexInfos; _i < indexInfos_7.length; _i++) {
                    var info = indexInfos_7[_i];
                    _loop_27(info);
                }
            }
            function checkIndexConstraintForIndexSignature(type, checkInfo) {
                var declaration = checkInfo.declaration;
                var indexInfos = getApplicableIndexInfos(type, checkInfo.keyType);
                var interfaceDeclaration = ts.getObjectFlags(type) & 2 /* Interface */ ? ts.getDeclarationOfKind(type.symbol, 256 /* InterfaceDeclaration */) : undefined;
                var localCheckDeclaration = declaration && getParentOfSymbol(getSymbolOfNode(declaration)) === type.symbol ? declaration : undefined;
                var _loop_28 = function (info) {
                    if (info === checkInfo)
                        return "continue";
                    var localIndexDeclaration = info.declaration && getParentOfSymbol(getSymbolOfNode(info.declaration)) === type.symbol ? info.declaration : undefined;
                    // We check only when (a) the check index signature is declared in the containing type, or (b) the applicable index
                    // signature is declared in the containing type, or (c) the containing type is an interface and no base interface contains
                    // both index signatures (i.e. the index signatures are declared in separate inherited interfaces).
                    var errorNode = localCheckDeclaration || localIndexDeclaration ||
                        (interfaceDeclaration && !ts.some(getBaseTypes(type), function (base) { return !!getIndexInfoOfType(base, checkInfo.keyType) && !!getIndexTypeOfType(base, info.keyType); }) ? interfaceDeclaration : undefined);
                    if (errorNode && !isTypeAssignableTo(checkInfo.type, info.type)) {
                        error(errorNode, ts.Diagnostics._0_index_type_1_is_not_assignable_to_2_index_type_3, typeToString(checkInfo.keyType), typeToString(checkInfo.type), typeToString(info.keyType), typeToString(info.type));
                    }
                };
                for (var _i = 0, indexInfos_8 = indexInfos; _i < indexInfos_8.length; _i++) {
                    var info = indexInfos_8[_i];
                    _loop_28(info);
                }
            }
            function checkTypeNameIsReserved(name, message) {
                // TS 1.0 spec (April 2014): 3.6.1
                // The predefined type keywords are reserved and cannot be used as names of user defined types.
                switch (name.escapedText) {
                    case "any":
                    case "unknown":
                    case "never":
                    case "number":
                    case "bigint":
                    case "boolean":
                    case "string":
                    case "symbol":
                    case "void":
                    case "object":
                        error(name, message, name.escapedText);
                }
            }
            /**
             * The name cannot be used as 'Object' of user defined types with special target.
             */
            function checkClassNameCollisionWithObject(name) {
                if (languageVersion === 1 /* ES5 */ && name.escapedText === "Object"
                    && moduleKind < ts.ModuleKind.ES2015) {
                    error(name, ts.Diagnostics.Class_name_cannot_be_Object_when_targeting_ES5_with_module_0, ts.ModuleKind[moduleKind]); // https://github.com/Microsoft/TypeScript/issues/17494
                }
            }
            /**
             * Check each type parameter and check that type parameters have no duplicate type parameter declarations
             */
            function checkTypeParameters(typeParameterDeclarations) {
                if (typeParameterDeclarations) {
                    var seenDefault = false;
                    for (var i = 0; i < typeParameterDeclarations.length; i++) {
                        var node = typeParameterDeclarations[i];
                        checkTypeParameter(node);
                        if (produceDiagnostics) {
                            if (node.default) {
                                seenDefault = true;
                                checkTypeParametersNotReferenced(node.default, typeParameterDeclarations, i);
                            }
                            else if (seenDefault) {
                                error(node, ts.Diagnostics.Required_type_parameters_may_not_follow_optional_type_parameters);
                            }
                            for (var j = 0; j < i; j++) {
                                if (typeParameterDeclarations[j].symbol === node.symbol) {
                                    error(node.name, ts.Diagnostics.Duplicate_identifier_0, ts.declarationNameToString(node.name));
                                }
                            }
                        }
                    }
                }
            }
            /** Check that type parameter defaults only reference previously declared type parameters */
            function checkTypeParametersNotReferenced(root, typeParameters, index) {
                visit(root);
                function visit(node) {
                    if (node.kind === 176 /* TypeReference */) {
                        var type = getTypeFromTypeReference(node);
                        if (type.flags & 262144 /* TypeParameter */) {
                            for (var i = index; i < typeParameters.length; i++) {
                                if (type.symbol === getSymbolOfNode(typeParameters[i])) {
                                    error(node, ts.Diagnostics.Type_parameter_defaults_can_only_reference_previously_declared_type_parameters);
                                }
                            }
                        }
                    }
                    ts.forEachChild(node, visit);
                }
            }
            /** Check that type parameter lists are identical across multiple declarations */
            function checkTypeParameterListsIdentical(symbol) {
                if (symbol.declarations && symbol.declarations.length === 1) {
                    return;
                }
                var links = getSymbolLinks(symbol);
                if (!links.typeParametersChecked) {
                    links.typeParametersChecked = true;
                    var declarations = getClassOrInterfaceDeclarationsOfSymbol(symbol);
                    if (!declarations || declarations.length <= 1) {
                        return;
                    }
                    var type = getDeclaredTypeOfSymbol(symbol);
                    if (!areTypeParametersIdentical(declarations, type.localTypeParameters)) {
                        // Report an error on every conflicting declaration.
                        var name = symbolToString(symbol);
                        for (var _i = 0, declarations_6 = declarations; _i < declarations_6.length; _i++) {
                            var declaration = declarations_6[_i];
                            error(declaration.name, ts.Diagnostics.All_declarations_of_0_must_have_identical_type_parameters, name);
                        }
                    }
                }
            }
            function areTypeParametersIdentical(declarations, targetParameters) {
                var maxTypeArgumentCount = ts.length(targetParameters);
                var minTypeArgumentCount = getMinTypeArgumentCount(targetParameters);
                for (var _i = 0, declarations_7 = declarations; _i < declarations_7.length; _i++) {
                    var declaration = declarations_7[_i];
                    // If this declaration has too few or too many type parameters, we report an error
                    var sourceParameters = ts.getEffectiveTypeParameterDeclarations(declaration);
                    var numTypeParameters = sourceParameters.length;
                    if (numTypeParameters < minTypeArgumentCount || numTypeParameters > maxTypeArgumentCount) {
                        return false;
                    }
                    for (var i = 0; i < numTypeParameters; i++) {
                        var source = sourceParameters[i];
                        var target = targetParameters[i];
                        // If the type parameter node does not have the same as the resolved type
                        // parameter at this position, we report an error.
                        if (source.name.escapedText !== target.symbol.escapedName) {
                            return false;
                        }
                        // If the type parameter node does not have an identical constraint as the resolved
                        // type parameter at this position, we report an error.
                        var constraint = ts.getEffectiveConstraintOfTypeParameter(source);
                        var sourceConstraint = constraint && getTypeFromTypeNode(constraint);
                        var targetConstraint = getConstraintOfTypeParameter(target);
                        // relax check if later interface augmentation has no constraint, it's more broad and is OK to merge with
                        // a more constrained interface (this could be generalized to a full hierarchy check, but that's maybe overkill)
                        if (sourceConstraint && targetConstraint && !isTypeIdenticalTo(sourceConstraint, targetConstraint)) {
                            return false;
                        }
                        // If the type parameter node has a default and it is not identical to the default
                        // for the type parameter at this position, we report an error.
                        var sourceDefault = source.default && getTypeFromTypeNode(source.default);
                        var targetDefault = getDefaultFromTypeParameter(target);
                        if (sourceDefault && targetDefault && !isTypeIdenticalTo(sourceDefault, targetDefault)) {
                            return false;
                        }
                    }
                }
                return true;
            }
            function checkClassExpression(node) {
                checkClassLikeDeclaration(node);
                checkNodeDeferred(node);
                return getTypeOfSymbol(getSymbolOfNode(node));
            }
            function checkClassExpressionDeferred(node) {
                ts.forEach(node.members, checkSourceElement);
                registerForUnusedIdentifiersCheck(node);
            }
            function checkClassDeclaration(node) {
                if (ts.some(node.decorators) && ts.some(node.members, function (p) { return ts.hasStaticModifier(p) && ts.isPrivateIdentifierClassElementDeclaration(p); })) {
                    grammarErrorOnNode(node.decorators[0], ts.Diagnostics.Class_decorators_can_t_be_used_with_static_private_identifier_Consider_removing_the_experimental_decorator);
                }
                if (!node.name && !ts.hasSyntacticModifier(node, 512 /* Default */)) {
                    grammarErrorOnFirstToken(node, ts.Diagnostics.A_class_declaration_without_the_default_modifier_must_have_a_name);
                }
                checkClassLikeDeclaration(node);
                ts.forEach(node.members, checkSourceElement);
                registerForUnusedIdentifiersCheck(node);
            }
            function checkClassLikeDeclaration(node) {
                checkGrammarClassLikeDeclaration(node);
                checkDecorators(node);
                checkCollisionsForDeclarationName(node, node.name);
                checkTypeParameters(ts.getEffectiveTypeParameterDeclarations(node));
                checkExportsOnMergedDeclarations(node);
                var symbol = getSymbolOfNode(node);
                var type = getDeclaredTypeOfSymbol(symbol);
                var typeWithThis = getTypeWithThisArgument(type);
                var staticType = getTypeOfSymbol(symbol);
                checkTypeParameterListsIdentical(symbol);
                checkFunctionOrConstructorSymbol(symbol);
                checkClassForDuplicateDeclarations(node);
                // Only check for reserved static identifiers on non-ambient context.
                var nodeInAmbientContext = !!(node.flags & 8388608 /* Ambient */);
                if (!nodeInAmbientContext) {
                    checkClassForStaticPropertyNameConflicts(node);
                }
                var baseTypeNode = ts.getEffectiveBaseTypeNode(node);
                if (baseTypeNode) {
                    ts.forEach(baseTypeNode.typeArguments, checkSourceElement);
                    if (languageVersion < 2 /* ES2015 */) {
                        checkExternalEmitHelpers(baseTypeNode.parent, 1 /* Extends */);
                    }
                    // check both @extends and extends if both are specified.
                    var extendsNode = ts.getClassExtendsHeritageElement(node);
                    if (extendsNode && extendsNode !== baseTypeNode) {
                        checkExpression(extendsNode.expression);
                    }
                    var baseTypes = getBaseTypes(type);
                    if (baseTypes.length && produceDiagnostics) {
                        var baseType_1 = baseTypes[0];
                        var baseConstructorType = getBaseConstructorTypeOfClass(type);
                        var staticBaseType = getApparentType(baseConstructorType);
                        checkBaseTypeAccessibility(staticBaseType, baseTypeNode);
                        checkSourceElement(baseTypeNode.expression);
                        if (ts.some(baseTypeNode.typeArguments)) {
                            ts.forEach(baseTypeNode.typeArguments, checkSourceElement);
                            for (var _i = 0, _a = getConstructorsForTypeArguments(staticBaseType, baseTypeNode.typeArguments, baseTypeNode); _i < _a.length; _i++) {
                                var constructor = _a[_i];
                                if (!checkTypeArgumentConstraints(baseTypeNode, constructor.typeParameters)) {
                                    break;
                                }
                            }
                        }
                        var baseWithThis = getTypeWithThisArgument(baseType_1, type.thisType);
                        if (!checkTypeAssignableTo(typeWithThis, baseWithThis, /*errorNode*/ undefined)) {
                            issueMemberSpecificError(node, typeWithThis, baseWithThis, ts.Diagnostics.Class_0_incorrectly_extends_base_class_1);
                        }
                        else {
                            // Report static side error only when instance type is assignable
                            checkTypeAssignableTo(staticType, getTypeWithoutSignatures(staticBaseType), node.name || node, ts.Diagnostics.Class_static_side_0_incorrectly_extends_base_class_static_side_1);
                        }
                        if (baseConstructorType.flags & 8650752 /* TypeVariable */) {
                            if (!isMixinConstructorType(staticType)) {
                                error(node.name || node, ts.Diagnostics.A_mixin_class_must_have_a_constructor_with_a_single_rest_parameter_of_type_any);
                            }
                            else {
                                var constructSignatures = getSignaturesOfType(baseConstructorType, 1 /* Construct */);
                                if (constructSignatures.some(function (signature) { return signature.flags & 4 /* Abstract */; }) && !ts.hasSyntacticModifier(node, 128 /* Abstract */)) {
                                    error(node.name || node, ts.Diagnostics.A_mixin_class_that_extends_from_a_type_variable_containing_an_abstract_construct_signature_must_also_be_declared_abstract);
                                }
                            }
                        }
                        if (!(staticBaseType.symbol && staticBaseType.symbol.flags & 32 /* Class */) && !(baseConstructorType.flags & 8650752 /* TypeVariable */)) {
                            // When the static base type is a "class-like" constructor function (but not actually a class), we verify
                            // that all instantiated base constructor signatures return the same type.
                            var constructors = getInstantiatedConstructorsForTypeArguments(staticBaseType, baseTypeNode.typeArguments, baseTypeNode);
                            if (ts.forEach(constructors, function (sig) { return !isJSConstructor(sig.declaration) && !isTypeIdenticalTo(getReturnTypeOfSignature(sig), baseType_1); })) {
                                error(baseTypeNode.expression, ts.Diagnostics.Base_constructors_must_all_have_the_same_return_type);
                            }
                        }
                        checkKindsOfPropertyMemberOverrides(type, baseType_1);
                    }
                }
                checkMembersForMissingOverrideModifier(node, type, typeWithThis, staticType);
                var implementedTypeNodes = ts.getEffectiveImplementsTypeNodes(node);
                if (implementedTypeNodes) {
                    for (var _b = 0, implementedTypeNodes_1 = implementedTypeNodes; _b < implementedTypeNodes_1.length; _b++) {
                        var typeRefNode = implementedTypeNodes_1[_b];
                        if (!ts.isEntityNameExpression(typeRefNode.expression) || ts.isOptionalChain(typeRefNode.expression)) {
                            error(typeRefNode.expression, ts.Diagnostics.A_class_can_only_implement_an_identifier_Slashqualified_name_with_optional_type_arguments);
                        }
                        checkTypeReferenceNode(typeRefNode);
                        if (produceDiagnostics) {
                            var t = getReducedType(getTypeFromTypeNode(typeRefNode));
                            if (t !== errorType) {
                                if (isValidBaseType(t)) {
                                    var genericDiag = t.symbol && t.symbol.flags & 32 /* Class */ ?
                                        ts.Diagnostics.Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass :
                                        ts.Diagnostics.Class_0_incorrectly_implements_interface_1;
                                    var baseWithThis = getTypeWithThisArgument(t, type.thisType);
                                    if (!checkTypeAssignableTo(typeWithThis, baseWithThis, /*errorNode*/ undefined)) {
                                        issueMemberSpecificError(node, typeWithThis, baseWithThis, genericDiag);
                                    }
                                }
                                else {
                                    error(typeRefNode, ts.Diagnostics.A_class_can_only_implement_an_object_type_or_intersection_of_object_types_with_statically_known_members);
                                }
                            }
                        }
                    }
                }
                if (produceDiagnostics) {
                    checkIndexConstraints(type);
                    checkIndexConstraints(staticType, /*isStaticIndex*/ true);
                    checkTypeForDuplicateIndexSignatures(node);
                    checkPropertyInitialization(node);
                }
            }
            function checkMembersForMissingOverrideModifier(node, type, typeWithThis, staticType) {
                var nodeInAmbientContext = !!(node.flags & 8388608 /* Ambient */);
                var baseTypeNode = ts.getEffectiveBaseTypeNode(node);
                var baseTypes = baseTypeNode && getBaseTypes(type);
                var baseWithThis = (baseTypes === null || baseTypes === void 0 ? void 0 : baseTypes.length) ? getTypeWithThisArgument(ts.first(baseTypes), type.thisType) : undefined;
                var baseStaticType = getBaseConstructorTypeOfClass(type);
                var _loop_29 = function (member) {
                    if (ts.hasAmbientModifier(member)) {
                        return "continue";
                    }
                    if (ts.isConstructorDeclaration(member)) {
                        ts.forEach(member.parameters, function (param) {
                            if (ts.isParameterPropertyDeclaration(param, member)) {
                                checkClassMember(param, /*memberIsParameterProperty*/ true);
                            }
                        });
                    }
                    checkClassMember(member);
                };
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    _loop_29(member);
                }
                function checkClassMember(member, memberIsParameterProperty) {
                    var hasOverride = ts.hasOverrideModifier(member);
                    var hasStatic = ts.isStatic(member);
                    if (baseWithThis && (hasOverride || compilerOptions.noImplicitOverride)) {
                        var declaredProp = member.name && getSymbolAtLocation(member.name) || getSymbolAtLocation(member);
                        if (!declaredProp) {
                            return;
                        }
                        var thisType = hasStatic ? staticType : typeWithThis;
                        var baseType = hasStatic ? baseStaticType : baseWithThis;
                        var prop = getPropertyOfType(thisType, declaredProp.escapedName);
                        var baseProp = getPropertyOfType(baseType, declaredProp.escapedName);
                        var baseClassName = typeToString(baseWithThis);
                        if (prop && !baseProp && hasOverride) {
                            var suggestion = getSuggestedSymbolForNonexistentClassMember(ts.symbolName(declaredProp), baseType);
                            suggestion ?
                                error(member, ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0_Did_you_mean_1, baseClassName, symbolToString(suggestion)) :
                                error(member, ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0, baseClassName);
                        }
                        else if (prop && (baseProp === null || baseProp === void 0 ? void 0 : baseProp.declarations) && compilerOptions.noImplicitOverride && !nodeInAmbientContext) {
                            var baseHasAbstract = ts.some(baseProp.declarations, function (d) { return ts.hasAbstractModifier(d); });
                            if (hasOverride) {
                                return;
                            }
                            if (!baseHasAbstract) {
                                var diag = memberIsParameterProperty ?
                                    ts.Diagnostics.This_parameter_property_must_have_an_override_modifier_because_it_overrides_a_member_in_base_class_0 :
                                    ts.Diagnostics.This_member_must_have_an_override_modifier_because_it_overrides_a_member_in_the_base_class_0;
                                error(member, diag, baseClassName);
                            }
                            else if (ts.hasAbstractModifier(member) && baseHasAbstract) {
                                error(member, ts.Diagnostics.This_member_must_have_an_override_modifier_because_it_overrides_an_abstract_method_that_is_declared_in_the_base_class_0, baseClassName);
                            }
                        }
                    }
                    else if (hasOverride) {
                        var className = typeToString(type);
                        error(member, ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_its_containing_class_0_does_not_extend_another_class, className);
                    }
                }
            }
            function issueMemberSpecificError(node, typeWithThis, baseWithThis, broadDiag) {
                // iterate over all implemented properties and issue errors on each one which isn't compatible, rather than the class as a whole, if possible
                var issuedMemberError = false;
                var _loop_30 = function (member) {
                    if (ts.isStatic(member)) {
                        return "continue";
                    }
                    var declaredProp = member.name && getSymbolAtLocation(member.name) || getSymbolAtLocation(member);
                    if (declaredProp) {
                        var prop = getPropertyOfType(typeWithThis, declaredProp.escapedName);
                        var baseProp = getPropertyOfType(baseWithThis, declaredProp.escapedName);
                        if (prop && baseProp) {
                            var rootChain = function () {
                                return ts.chainDiagnosticMessages(
                                /*details*/ undefined, ts.Diagnostics.Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2, symbolToString(declaredProp), typeToString(typeWithThis), typeToString(baseWithThis));
                            };
                            if (!checkTypeAssignableTo(getTypeOfSymbol(prop), getTypeOfSymbol(baseProp), member.name || member, /*message*/ undefined, rootChain)) {
                                issuedMemberError = true;
                            }
                        }
                    }
                };
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    _loop_30(member);
                }
                if (!issuedMemberError) {
                    // check again with diagnostics to generate a less-specific error
                    checkTypeAssignableTo(typeWithThis, baseWithThis, node.name || node, broadDiag);
                }
            }
            function checkBaseTypeAccessibility(type, node) {
                var signatures = getSignaturesOfType(type, 1 /* Construct */);
                if (signatures.length) {
                    var declaration = signatures[0].declaration;
                    if (declaration && ts.hasEffectiveModifier(declaration, 8 /* Private */)) {
                        var typeClassDeclaration = ts.getClassLikeDeclarationOfSymbol(type.symbol);
                        if (!isNodeWithinClass(node, typeClassDeclaration)) {
                            error(node, ts.Diagnostics.Cannot_extend_a_class_0_Class_constructor_is_marked_as_private, getFullyQualifiedName(type.symbol));
                        }
                    }
                }
            }
            function getTargetSymbol(s) {
                // if symbol is instantiated its flags are not copied from the 'target'
                // so we'll need to get back original 'target' symbol to work with correct set of flags
                return ts.getCheckFlags(s) & 1 /* Instantiated */ ? s.target : s;
            }
            function getClassOrInterfaceDeclarationsOfSymbol(symbol) {
                return ts.filter(symbol.declarations, function (d) {
                    return d.kind === 255 /* ClassDeclaration */ || d.kind === 256 /* InterfaceDeclaration */;
                });
            }
            function checkKindsOfPropertyMemberOverrides(type, baseType) {
                // TypeScript 1.0 spec (April 2014): 8.2.3
                // A derived class inherits all members from its base class it doesn't override.
                // Inheritance means that a derived class implicitly contains all non - overridden members of the base class.
                // Both public and private property members are inherited, but only public property members can be overridden.
                // A property member in a derived class is said to override a property member in a base class
                // when the derived class property member has the same name and kind(instance or static)
                // as the base class property member.
                // The type of an overriding property member must be assignable(section 3.8.4)
                // to the type of the overridden property member, or otherwise a compile - time error occurs.
                // Base class instance member functions can be overridden by derived class instance member functions,
                // but not by other kinds of members.
                // Base class instance member variables and accessors can be overridden by
                // derived class instance member variables and accessors, but not by other kinds of members.
                var _a, _b;
                // NOTE: assignability is checked in checkClassDeclaration
                var baseProperties = getPropertiesOfType(baseType);
                basePropertyCheck: for (var _i = 0, baseProperties_1 = baseProperties; _i < baseProperties_1.length; _i++) {
                    var baseProperty = baseProperties_1[_i];
                    var base = getTargetSymbol(baseProperty);
                    if (base.flags & 4194304 /* Prototype */) {
                        continue;
                    }
                    var baseSymbol = getPropertyOfObjectType(type, base.escapedName);
                    if (!baseSymbol) {
                        continue;
                    }
                    var derived = getTargetSymbol(baseSymbol);
                    var baseDeclarationFlags = ts.getDeclarationModifierFlagsFromSymbol(base);
                    ts.Debug.assert(!!derived, "derived should point to something, even if it is the base class' declaration.");
                    // In order to resolve whether the inherited method was overridden in the base class or not,
                    // we compare the Symbols obtained. Since getTargetSymbol returns the symbol on the *uninstantiated*
                    // type declaration, derived and base resolve to the same symbol even in the case of generic classes.
                    if (derived === base) {
                        // derived class inherits base without override/redeclaration
                        var derivedClassDecl = ts.getClassLikeDeclarationOfSymbol(type.symbol);
                        // It is an error to inherit an abstract member without implementing it or being declared abstract.
                        // If there is no declaration for the derived class (as in the case of class expressions),
                        // then the class cannot be declared abstract.
                        if (baseDeclarationFlags & 128 /* Abstract */ && (!derivedClassDecl || !ts.hasSyntacticModifier(derivedClassDecl, 128 /* Abstract */))) {
                            // Searches other base types for a declaration that would satisfy the inherited abstract member.
                            // (The class may have more than one base type via declaration merging with an interface with the
                            // same name.)
                            for (var _c = 0, _d = getBaseTypes(type); _c < _d.length; _c++) {
                                var otherBaseType = _d[_c];
                                if (otherBaseType === baseType)
                                    continue;
                                var baseSymbol_1 = getPropertyOfObjectType(otherBaseType, base.escapedName);
                                var derivedElsewhere = baseSymbol_1 && getTargetSymbol(baseSymbol_1);
                                if (derivedElsewhere && derivedElsewhere !== base) {
                                    continue basePropertyCheck;
                                }
                            }
                            if (derivedClassDecl.kind === 224 /* ClassExpression */) {
                                error(derivedClassDecl, ts.Diagnostics.Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1, symbolToString(baseProperty), typeToString(baseType));
                            }
                            else {
                                error(derivedClassDecl, ts.Diagnostics.Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2, typeToString(type), symbolToString(baseProperty), typeToString(baseType));
                            }
                        }
                    }
                    else {
                        // derived overrides base.
                        var derivedDeclarationFlags = ts.getDeclarationModifierFlagsFromSymbol(derived);
                        if (baseDeclarationFlags & 8 /* Private */ || derivedDeclarationFlags & 8 /* Private */) {
                            // either base or derived property is private - not override, skip it
                            continue;
                        }
                        var errorMessage = void 0;
                        var basePropertyFlags = base.flags & 98308 /* PropertyOrAccessor */;
                        var derivedPropertyFlags = derived.flags & 98308 /* PropertyOrAccessor */;
                        if (basePropertyFlags && derivedPropertyFlags) {
                            // property/accessor is overridden with property/accessor
                            if (baseDeclarationFlags & 128 /* Abstract */ && !(base.valueDeclaration && ts.isPropertyDeclaration(base.valueDeclaration) && base.valueDeclaration.initializer)
                                || base.valueDeclaration && base.valueDeclaration.parent.kind === 256 /* InterfaceDeclaration */
                                || derived.valueDeclaration && ts.isBinaryExpression(derived.valueDeclaration)) {
                                // when the base property is abstract or from an interface, base/derived flags don't need to match
                                // same when the derived property is from an assignment
                                continue;
                            }
                            var overriddenInstanceProperty = basePropertyFlags !== 4 /* Property */ && derivedPropertyFlags === 4 /* Property */;
                            var overriddenInstanceAccessor = basePropertyFlags === 4 /* Property */ && derivedPropertyFlags !== 4 /* Property */;
                            if (overriddenInstanceProperty || overriddenInstanceAccessor) {
                                var errorMessage_1 = overriddenInstanceProperty ?
                                    ts.Diagnostics._0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property :
                                    ts.Diagnostics._0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor;
                                error(ts.getNameOfDeclaration(derived.valueDeclaration) || derived.valueDeclaration, errorMessage_1, symbolToString(base), typeToString(baseType), typeToString(type));
                            }
                            else if (useDefineForClassFields) {
                                var uninitialized = (_a = derived.declarations) === null || _a === void 0 ? void 0 : _a.find(function (d) { return d.kind === 165 /* PropertyDeclaration */ && !d.initializer; });
                                if (uninitialized
                                    && !(derived.flags & 33554432 /* Transient */)
                                    && !(baseDeclarationFlags & 128 /* Abstract */)
                                    && !(derivedDeclarationFlags & 128 /* Abstract */)
                                    && !((_b = derived.declarations) === null || _b === void 0 ? void 0 : _b.some(function (d) { return !!(d.flags & 8388608 /* Ambient */); }))) {
                                    var constructor = findConstructorDeclaration(ts.getClassLikeDeclarationOfSymbol(type.symbol));
                                    var propName = uninitialized.name;
                                    if (uninitialized.exclamationToken
                                        || !constructor
                                        || !ts.isIdentifier(propName)
                                        || !strictNullChecks
                                        || !isPropertyInitializedInConstructor(propName, type, constructor)) {
                                        var errorMessage_2 = ts.Diagnostics.Property_0_will_overwrite_the_base_property_in_1_If_this_is_intentional_add_an_initializer_Otherwise_add_a_declare_modifier_or_remove_the_redundant_declaration;
                                        error(ts.getNameOfDeclaration(derived.valueDeclaration) || derived.valueDeclaration, errorMessage_2, symbolToString(base), typeToString(baseType));
                                    }
                                }
                            }
                            // correct case
                            continue;
                        }
                        else if (isPrototypeProperty(base)) {
                            if (isPrototypeProperty(derived) || derived.flags & 4 /* Property */) {
                                // method is overridden with method or property -- correct case
                                continue;
                            }
                            else {
                                ts.Debug.assert(!!(derived.flags & 98304 /* Accessor */));
                                errorMessage = ts.Diagnostics.Class_0_defines_instance_member_function_1_but_extended_class_2_defines_it_as_instance_member_accessor;
                            }
                        }
                        else if (base.flags & 98304 /* Accessor */) {
                            errorMessage = ts.Diagnostics.Class_0_defines_instance_member_accessor_1_but_extended_class_2_defines_it_as_instance_member_function;
                        }
                        else {
                            errorMessage = ts.Diagnostics.Class_0_defines_instance_member_property_1_but_extended_class_2_defines_it_as_instance_member_function;
                        }
                        error(ts.getNameOfDeclaration(derived.valueDeclaration) || derived.valueDeclaration, errorMessage, typeToString(baseType), symbolToString(base), typeToString(type));
                    }
                }
            }
            function getNonInterhitedProperties(type, baseTypes, properties) {
                if (!ts.length(baseTypes)) {
                    return properties;
                }
                var seen = new ts.Map();
                ts.forEach(properties, function (p) { seen.set(p.escapedName, p); });
                for (var _i = 0, baseTypes_2 = baseTypes; _i < baseTypes_2.length; _i++) {
                    var base = baseTypes_2[_i];
                    var properties_5 = getPropertiesOfType(getTypeWithThisArgument(base, type.thisType));
                    for (var _a = 0, properties_4 = properties_5; _a < properties_4.length; _a++) {
                        var prop = properties_4[_a];
                        var existing = seen.get(prop.escapedName);
                        if (existing && !isPropertyIdenticalTo(existing, prop)) {
                            seen.delete(prop.escapedName);
                        }
                    }
                }
                return ts.arrayFrom(seen.values());
            }
            function checkInheritedPropertiesAreIdentical(type, typeNode) {
                var baseTypes = getBaseTypes(type);
                if (baseTypes.length < 2) {
                    return true;
                }
                var seen = new ts.Map();
                ts.forEach(resolveDeclaredMembers(type).declaredProperties, function (p) { seen.set(p.escapedName, { prop: p, containingType: type }); });
                var ok = true;
                for (var _i = 0, baseTypes_3 = baseTypes; _i < baseTypes_3.length; _i++) {
                    var base = baseTypes_3[_i];
                    var properties = getPropertiesOfType(getTypeWithThisArgument(base, type.thisType));
                    for (var _a = 0, properties_6 = properties; _a < properties_6.length; _a++) {
                        var prop = properties_6[_a];
                        var existing = seen.get(prop.escapedName);
                        if (!existing) {
                            seen.set(prop.escapedName, { prop: prop, containingType: base });
                        }
                        else {
                            var isInheritedProperty = existing.containingType !== type;
                            if (isInheritedProperty && !isPropertyIdenticalTo(existing.prop, prop)) {
                                ok = false;
                                var typeName1 = typeToString(existing.containingType);
                                var typeName2 = typeToString(base);
                                var errorInfo = ts.chainDiagnosticMessages(/*details*/ undefined, ts.Diagnostics.Named_property_0_of_types_1_and_2_are_not_identical, symbolToString(prop), typeName1, typeName2);
                                errorInfo = ts.chainDiagnosticMessages(errorInfo, ts.Diagnostics.Interface_0_cannot_simultaneously_extend_types_1_and_2, typeToString(type), typeName1, typeName2);
                                diagnostics.add(ts.createDiagnosticForNodeFromMessageChain(typeNode, errorInfo));
                            }
                        }
                    }
                }
                return ok;
            }
            function checkPropertyInitialization(node) {
                if (!strictNullChecks || !strictPropertyInitialization || node.flags & 8388608 /* Ambient */) {
                    return;
                }
                var constructor = findConstructorDeclaration(node);
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    if (ts.getEffectiveModifierFlags(member) & 2 /* Ambient */) {
                        continue;
                    }
                    if (!ts.isStatic(member) && isPropertyWithoutInitializer(member)) {
                        var propName = member.name;
                        if (ts.isIdentifier(propName) || ts.isPrivateIdentifier(propName)) {
                            var type = getTypeOfSymbol(getSymbolOfNode(member));
                            if (!(type.flags & 3 /* AnyOrUnknown */ || getFalsyFlags(type) & 32768 /* Undefined */)) {
                                if (!constructor || !isPropertyInitializedInConstructor(propName, type, constructor)) {
                                    error(member.name, ts.Diagnostics.Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor, ts.declarationNameToString(propName));
                                }
                            }
                        }
                    }
                }
            }
            function isPropertyWithoutInitializer(node) {
                return node.kind === 165 /* PropertyDeclaration */ &&
                    !ts.hasAbstractModifier(node) &&
                    !node.exclamationToken &&
                    !node.initializer;
            }
            function isPropertyInitializedInStaticBlocks(propName, propType, staticBlocks, startPos, endPos) {
                for (var _i = 0, staticBlocks_2 = staticBlocks; _i < staticBlocks_2.length; _i++) {
                    var staticBlock = staticBlocks_2[_i];
                    // static block must be within the provided range as they are evaluated in document order (unlike constructors)
                    if (staticBlock.pos >= startPos && staticBlock.pos <= endPos) {
                        var reference = ts.factory.createPropertyAccessExpression(ts.factory.createThis(), propName);
                        ts.setParent(reference.expression, reference);
                        ts.setParent(reference, staticBlock);
                        reference.flowNode = staticBlock.returnFlowNode;
                        var flowType = getFlowTypeOfReference(reference, propType, getOptionalType(propType));
                        if (!(getFalsyFlags(flowType) & 32768 /* Undefined */)) {
                            return true;
                        }
                    }
                }
                return false;
            }
            function isPropertyInitializedInConstructor(propName, propType, constructor) {
                var reference = ts.factory.createPropertyAccessExpression(ts.factory.createThis(), propName);
                ts.setParent(reference.expression, reference);
                ts.setParent(reference, constructor);
                reference.flowNode = constructor.returnFlowNode;
                var flowType = getFlowTypeOfReference(reference, propType, getOptionalType(propType));
                return !(getFalsyFlags(flowType) & 32768 /* Undefined */);
            }
            function checkInterfaceDeclaration(node) {
                // Grammar checking
                if (!checkGrammarDecoratorsAndModifiers(node))
                    checkGrammarInterfaceDeclaration(node);
                checkTypeParameters(node.typeParameters);
                if (produceDiagnostics) {
                    checkTypeNameIsReserved(node.name, ts.Diagnostics.Interface_name_cannot_be_0);
                    checkExportsOnMergedDeclarations(node);
                    var symbol = getSymbolOfNode(node);
                    checkTypeParameterListsIdentical(symbol);
                    // Only check this symbol once
                    var firstInterfaceDecl = ts.getDeclarationOfKind(symbol, 256 /* InterfaceDeclaration */);
                    if (node === firstInterfaceDecl) {
                        var type = getDeclaredTypeOfSymbol(symbol);
                        var typeWithThis = getTypeWithThisArgument(type);
                        // run subsequent checks only if first set succeeded
                        if (checkInheritedPropertiesAreIdentical(type, node.name)) {
                            for (var _i = 0, _a = getBaseTypes(type); _i < _a.length; _i++) {
                                var baseType = _a[_i];
                                checkTypeAssignableTo(typeWithThis, getTypeWithThisArgument(baseType, type.thisType), node.name, ts.Diagnostics.Interface_0_incorrectly_extends_interface_1);
                            }
                            checkIndexConstraints(type);
                        }
                    }
                    checkObjectTypeForDuplicateDeclarations(node);
                }
                ts.forEach(ts.getInterfaceBaseTypeNodes(node), function (heritageElement) {
                    if (!ts.isEntityNameExpression(heritageElement.expression) || ts.isOptionalChain(heritageElement.expression)) {
                        error(heritageElement.expression, ts.Diagnostics.An_interface_can_only_extend_an_identifier_Slashqualified_name_with_optional_type_arguments);
                    }
                    checkTypeReferenceNode(heritageElement);
                });
                ts.forEach(node.members, checkSourceElement);
                if (produceDiagnostics) {
                    checkTypeForDuplicateIndexSignatures(node);
                    registerForUnusedIdentifiersCheck(node);
                }
            }
            function checkTypeAliasDeclaration(node) {
                // Grammar checking
                checkGrammarDecoratorsAndModifiers(node);
                checkTypeNameIsReserved(node.name, ts.Diagnostics.Type_alias_name_cannot_be_0);
                checkExportsOnMergedDeclarations(node);
                checkTypeParameters(node.typeParameters);
                if (node.type.kind === 137 /* IntrinsicKeyword */) {
                    if (!intrinsicTypeKinds.has(node.name.escapedText) || ts.length(node.typeParameters) !== 1) {
                        error(node.type, ts.Diagnostics.The_intrinsic_keyword_can_only_be_used_to_declare_compiler_provided_intrinsic_types);
                    }
                }
                else {
                    checkSourceElement(node.type);
                    registerForUnusedIdentifiersCheck(node);
                }
            }
            function computeEnumMemberValues(node) {
                var nodeLinks = getNodeLinks(node);
                if (!(nodeLinks.flags & 16384 /* EnumValuesComputed */)) {
                    nodeLinks.flags |= 16384 /* EnumValuesComputed */;
                    var autoValue = 0;
                    for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                        var member = _a[_i];
                        var value = computeMemberValue(member, autoValue);
                        getNodeLinks(member).enumMemberValue = value;
                        autoValue = typeof value === "number" ? value + 1 : undefined;
                    }
                }
            }
            function computeMemberValue(member, autoValue) {
                if (ts.isComputedNonLiteralName(member.name)) {
                    error(member.name, ts.Diagnostics.Computed_property_names_are_not_allowed_in_enums);
                }
                else {
                    var text = ts.getTextOfPropertyName(member.name);
                    if (isNumericLiteralName(text) && !isInfinityOrNaNString(text)) {
                        error(member.name, ts.Diagnostics.An_enum_member_cannot_have_a_numeric_name);
                    }
                }
                if (member.initializer) {
                    return computeConstantValue(member);
                }
                // In ambient non-const numeric enum declarations, enum members without initializers are
                // considered computed members (as opposed to having auto-incremented values).
                if (member.parent.flags & 8388608 /* Ambient */ && !ts.isEnumConst(member.parent) && getEnumKind(getSymbolOfNode(member.parent)) === 0 /* Numeric */) {
                    return undefined;
                }
                // If the member declaration specifies no value, the member is considered a constant enum member.
                // If the member is the first member in the enum declaration, it is assigned the value zero.
                // Otherwise, it is assigned the value of the immediately preceding member plus one, and an error
                // occurs if the immediately preceding member is not a constant enum member.
                if (autoValue !== undefined) {
                    return autoValue;
                }
                error(member.name, ts.Diagnostics.Enum_member_must_have_initializer);
                return undefined;
            }
            function computeConstantValue(member) {
                var enumKind = getEnumKind(getSymbolOfNode(member.parent));
                var isConstEnum = ts.isEnumConst(member.parent);
                var initializer = member.initializer;
                var value = enumKind === 1 /* Literal */ && !isLiteralEnumMember(member) ? undefined : evaluate(initializer);
                if (value !== undefined) {
                    if (isConstEnum && typeof value === "number" && !isFinite(value)) {
                        error(initializer, isNaN(value) ?
                            ts.Diagnostics.const_enum_member_initializer_was_evaluated_to_disallowed_value_NaN :
                            ts.Diagnostics.const_enum_member_initializer_was_evaluated_to_a_non_finite_value);
                    }
                }
                else if (enumKind === 1 /* Literal */) {
                    error(initializer, ts.Diagnostics.Computed_values_are_not_permitted_in_an_enum_with_string_valued_members);
                    return 0;
                }
                else if (isConstEnum) {
                    error(initializer, ts.Diagnostics.const_enum_member_initializers_can_only_contain_literal_values_and_other_computed_enum_values);
                }
                else if (member.parent.flags & 8388608 /* Ambient */) {
                    error(initializer, ts.Diagnostics.In_ambient_enum_declarations_member_initializer_must_be_constant_expression);
                }
                else {
                    // Only here do we need to check that the initializer is assignable to the enum type.
                    var source = checkExpression(initializer);
                    if (!isTypeAssignableToKind(source, 296 /* NumberLike */)) {
                        error(initializer, ts.Diagnostics.Only_numeric_enums_can_have_computed_members_but_this_expression_has_type_0_If_you_do_not_need_exhaustiveness_checks_consider_using_an_object_literal_instead, typeToString(source));
                    }
                    else {
                        checkTypeAssignableTo(source, getDeclaredTypeOfSymbol(getSymbolOfNode(member.parent)), initializer, /*headMessage*/ undefined);
                    }
                }
                return value;
                function evaluate(expr) {
                    switch (expr.kind) {
                        case 217 /* PrefixUnaryExpression */:
                            var value_2 = evaluate(expr.operand);
                            if (typeof value_2 === "number") {
                                switch (expr.operator) {
                                    case 39 /* PlusToken */: return value_2;
                                    case 40 /* MinusToken */: return -value_2;
                                    case 54 /* TildeToken */: return ~value_2;
                                }
                            }
                            break;
                        case 219 /* BinaryExpression */:
                            var left = evaluate(expr.left);
                            var right = evaluate(expr.right);
                            if (typeof left === "number" && typeof right === "number") {
                                switch (expr.operatorToken.kind) {
                                    case 51 /* BarToken */: return left | right;
                                    case 50 /* AmpersandToken */: return left & right;
                                    case 48 /* GreaterThanGreaterThanToken */: return left >> right;
                                    case 49 /* GreaterThanGreaterThanGreaterThanToken */: return left >>> right;
                                    case 47 /* LessThanLessThanToken */: return left << right;
                                    case 52 /* CaretToken */: return left ^ right;
                                    case 41 /* AsteriskToken */: return left * right;
                                    case 43 /* SlashToken */: return left / right;
                                    case 39 /* PlusToken */: return left + right;
                                    case 40 /* MinusToken */: return left - right;
                                    case 44 /* PercentToken */: return left % right;
                                    case 42 /* AsteriskAsteriskToken */: return Math.pow(left, right);
                                }
                            }
                            else if (typeof left === "string" && typeof right === "string" && expr.operatorToken.kind === 39 /* PlusToken */) {
                                return left + right;
                            }
                            break;
                        case 10 /* StringLiteral */:
                        case 14 /* NoSubstitutionTemplateLiteral */:
                            return expr.text;
                        case 8 /* NumericLiteral */:
                            checkGrammarNumericLiteral(expr);
                            return +expr.text;
                        case 210 /* ParenthesizedExpression */:
                            return evaluate(expr.expression);
                        case 79 /* Identifier */:
                            var identifier = expr;
                            if (isInfinityOrNaNString(identifier.escapedText)) {
                                return +(identifier.escapedText);
                            }
                            return ts.nodeIsMissing(expr) ? 0 : evaluateEnumMember(expr, getSymbolOfNode(member.parent), identifier.escapedText);
                        case 205 /* ElementAccessExpression */:
                        case 204 /* PropertyAccessExpression */:
                            var ex = expr;
                            if (isConstantMemberAccess(ex)) {
                                var type = getTypeOfExpression(ex.expression);
                                if (type.symbol && type.symbol.flags & 384 /* Enum */) {
                                    var name = void 0;
                                    if (ex.kind === 204 /* PropertyAccessExpression */) {
                                        name = ex.name.escapedText;
                                    }
                                    else {
                                        name = ts.escapeLeadingUnderscores(ts.cast(ex.argumentExpression, ts.isLiteralExpression).text);
                                    }
                                    return evaluateEnumMember(expr, type.symbol, name);
                                }
                            }
                            break;
                    }
                    return undefined;
                }
                function evaluateEnumMember(expr, enumSymbol, name) {
                    var memberSymbol = enumSymbol.exports.get(name);
                    if (memberSymbol) {
                        var declaration = memberSymbol.valueDeclaration;
                        if (declaration !== member) {
                            if (declaration && isBlockScopedNameDeclaredBeforeUse(declaration, member)) {
                                return getEnumMemberValue(declaration);
                            }
                            error(expr, ts.Diagnostics.A_member_initializer_in_a_enum_declaration_cannot_reference_members_declared_after_it_including_members_defined_in_other_enums);
                            return 0;
                        }
                        else {
                            error(expr, ts.Diagnostics.Property_0_is_used_before_being_assigned, symbolToString(memberSymbol));
                        }
                    }
                    return undefined;
                }
            }
            function isConstantMemberAccess(node) {
                return node.kind === 79 /* Identifier */ ||
                    node.kind === 204 /* PropertyAccessExpression */ && isConstantMemberAccess(node.expression) ||
                    node.kind === 205 /* ElementAccessExpression */ && isConstantMemberAccess(node.expression) &&
                        ts.isStringLiteralLike(node.argumentExpression);
            }
            function checkEnumDeclaration(node) {
                if (!produceDiagnostics) {
                    return;
                }
                // Grammar checking
                checkGrammarDecoratorsAndModifiers(node);
                checkCollisionsForDeclarationName(node, node.name);
                checkExportsOnMergedDeclarations(node);
                node.members.forEach(checkEnumMember);
                computeEnumMemberValues(node);
                // Spec 2014 - Section 9.3:
                // It isn't possible for one enum declaration to continue the automatic numbering sequence of another,
                // and when an enum type has multiple declarations, only one declaration is permitted to omit a value
                // for the first member.
                //
                // Only perform this check once per symbol
                var enumSymbol = getSymbolOfNode(node);
                var firstDeclaration = ts.getDeclarationOfKind(enumSymbol, node.kind);
                if (node === firstDeclaration) {
                    if (enumSymbol.declarations && enumSymbol.declarations.length > 1) {
                        var enumIsConst_1 = ts.isEnumConst(node);
                        // check that const is placed\omitted on all enum declarations
                        ts.forEach(enumSymbol.declarations, function (decl) {
                            if (ts.isEnumDeclaration(decl) && ts.isEnumConst(decl) !== enumIsConst_1) {
                                error(ts.getNameOfDeclaration(decl), ts.Diagnostics.Enum_declarations_must_all_be_const_or_non_const);
                            }
                        });
                    }
                    var seenEnumMissingInitialInitializer_1 = false;
                    ts.forEach(enumSymbol.declarations, function (declaration) {
                        // return true if we hit a violation of the rule, false otherwise
                        if (declaration.kind !== 258 /* EnumDeclaration */) {
                            return false;
                        }
                        var enumDeclaration = declaration;
                        if (!enumDeclaration.members.length) {
                            return false;
                        }
                        var firstEnumMember = enumDeclaration.members[0];
                        if (!firstEnumMember.initializer) {
                            if (seenEnumMissingInitialInitializer_1) {
                                error(firstEnumMember.name, ts.Diagnostics.In_an_enum_with_multiple_declarations_only_one_declaration_can_omit_an_initializer_for_its_first_enum_element);
                            }
                            else {
                                seenEnumMissingInitialInitializer_1 = true;
                            }
                        }
                    });
                }
            }
            function checkEnumMember(node) {
                if (ts.isPrivateIdentifier(node.name)) {
                    error(node, ts.Diagnostics.An_enum_member_cannot_be_named_with_a_private_identifier);
                }
            }
            function getFirstNonAmbientClassOrFunctionDeclaration(symbol) {
                var declarations = symbol.declarations;
                if (declarations) {
                    for (var _i = 0, declarations_8 = declarations; _i < declarations_8.length; _i++) {
                        var declaration = declarations_8[_i];
                        if ((declaration.kind === 255 /* ClassDeclaration */ ||
                            (declaration.kind === 254 /* FunctionDeclaration */ && ts.nodeIsPresent(declaration.body))) &&
                            !(declaration.flags & 8388608 /* Ambient */)) {
                            return declaration;
                        }
                    }
                }
                return undefined;
            }
            function inSameLexicalScope(node1, node2) {
                var container1 = ts.getEnclosingBlockScopeContainer(node1);
                var container2 = ts.getEnclosingBlockScopeContainer(node2);
                if (isGlobalSourceFile(container1)) {
                    return isGlobalSourceFile(container2);
                }
                else if (isGlobalSourceFile(container2)) {
                    return false;
                }
                else {
                    return container1 === container2;
                }
            }
            function checkModuleDeclaration(node) {
                if (produceDiagnostics) {
                    // Grammar checking
                    var isGlobalAugmentation = ts.isGlobalScopeAugmentation(node);
                    var inAmbientContext = node.flags & 8388608 /* Ambient */;
                    if (isGlobalAugmentation && !inAmbientContext) {
                        error(node.name, ts.Diagnostics.Augmentations_for_the_global_scope_should_have_declare_modifier_unless_they_appear_in_already_ambient_context);
                    }
                    var isAmbientExternalModule = ts.isAmbientModule(node);
                    var contextErrorMessage = isAmbientExternalModule
                        ? ts.Diagnostics.An_ambient_module_declaration_is_only_allowed_at_the_top_level_in_a_file
                        : ts.Diagnostics.A_namespace_declaration_is_only_allowed_in_a_namespace_or_module;
                    if (checkGrammarModuleElementContext(node, contextErrorMessage)) {
                        // If we hit a module declaration in an illegal context, just bail out to avoid cascading errors.
                        return;
                    }
                    if (!checkGrammarDecoratorsAndModifiers(node)) {
                        if (!inAmbientContext && node.name.kind === 10 /* StringLiteral */) {
                            grammarErrorOnNode(node.name, ts.Diagnostics.Only_ambient_modules_can_use_quoted_names);
                        }
                    }
                    if (ts.isIdentifier(node.name)) {
                        checkCollisionsForDeclarationName(node, node.name);
                    }
                    checkExportsOnMergedDeclarations(node);
                    var symbol = getSymbolOfNode(node);
                    // The following checks only apply on a non-ambient instantiated module declaration.
                    if (symbol.flags & 512 /* ValueModule */
                        && !inAmbientContext
                        && symbol.declarations
                        && symbol.declarations.length > 1
                        && isInstantiatedModule(node, ts.shouldPreserveConstEnums(compilerOptions))) {
                        var firstNonAmbientClassOrFunc = getFirstNonAmbientClassOrFunctionDeclaration(symbol);
                        if (firstNonAmbientClassOrFunc) {
                            if (ts.getSourceFileOfNode(node) !== ts.getSourceFileOfNode(firstNonAmbientClassOrFunc)) {
                                error(node.name, ts.Diagnostics.A_namespace_declaration_cannot_be_in_a_different_file_from_a_class_or_function_with_which_it_is_merged);
                            }
                            else if (node.pos < firstNonAmbientClassOrFunc.pos) {
                                error(node.name, ts.Diagnostics.A_namespace_declaration_cannot_be_located_prior_to_a_class_or_function_with_which_it_is_merged);
                            }
                        }
                        // if the module merges with a class declaration in the same lexical scope,
                        // we need to track this to ensure the correct emit.
                        var mergedClass = ts.getDeclarationOfKind(symbol, 255 /* ClassDeclaration */);
                        if (mergedClass &&
                            inSameLexicalScope(node, mergedClass)) {
                            getNodeLinks(node).flags |= 32768 /* LexicalModuleMergesWithClass */;
                        }
                    }
                    if (isAmbientExternalModule) {
                        if (ts.isExternalModuleAugmentation(node)) {
                            // body of the augmentation should be checked for consistency only if augmentation was applied to its target (either global scope or module)
                            // otherwise we'll be swamped in cascading errors.
                            // We can detect if augmentation was applied using following rules:
                            // - augmentation for a global scope is always applied
                            // - augmentation for some external module is applied if symbol for augmentation is merged (it was combined with target module).
                            var checkBody = isGlobalAugmentation || (getSymbolOfNode(node).flags & 33554432 /* Transient */);
                            if (checkBody && node.body) {
                                for (var _i = 0, _a = node.body.statements; _i < _a.length; _i++) {
                                    var statement = _a[_i];
                                    checkModuleAugmentationElement(statement, isGlobalAugmentation);
                                }
                            }
                        }
                        else if (isGlobalSourceFile(node.parent)) {
                            if (isGlobalAugmentation) {
                                error(node.name, ts.Diagnostics.Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_declarations);
                            }
                            else if (ts.isExternalModuleNameRelative(ts.getTextOfIdentifierOrLiteral(node.name))) {
                                error(node.name, ts.Diagnostics.Ambient_module_declaration_cannot_specify_relative_module_name);
                            }
                        }
                        else {
                            if (isGlobalAugmentation) {
                                error(node.name, ts.Diagnostics.Augmentations_for_the_global_scope_can_only_be_directly_nested_in_external_modules_or_ambient_module_declarations);
                            }
                            else {
                                // Node is not an augmentation and is not located on the script level.
                                // This means that this is declaration of ambient module that is located in other module or namespace which is prohibited.
                                error(node.name, ts.Diagnostics.Ambient_modules_cannot_be_nested_in_other_modules_or_namespaces);
                            }
                        }
                    }
                }
                if (node.body) {
                    checkSourceElement(node.body);
                    if (!ts.isGlobalScopeAugmentation(node)) {
                        registerForUnusedIdentifiersCheck(node);
                    }
                }
            }
            function checkModuleAugmentationElement(node, isGlobalAugmentation) {
                var _a;
                switch (node.kind) {
                    case 235 /* VariableStatement */:
                        // error each individual name in variable statement instead of marking the entire variable statement
                        for (var _i = 0, _b = node.declarationList.declarations; _i < _b.length; _i++) {
                            var decl = _b[_i];
                            checkModuleAugmentationElement(decl, isGlobalAugmentation);
                        }
                        break;
                    case 269 /* ExportAssignment */:
                    case 270 /* ExportDeclaration */:
                        grammarErrorOnFirstToken(node, ts.Diagnostics.Exports_and_export_assignments_are_not_permitted_in_module_augmentations);
                        break;
                    case 263 /* ImportEqualsDeclaration */:
                    case 264 /* ImportDeclaration */:
                        grammarErrorOnFirstToken(node, ts.Diagnostics.Imports_are_not_permitted_in_module_augmentations_Consider_moving_them_to_the_enclosing_external_module);
                        break;
                    case 201 /* BindingElement */:
                    case 252 /* VariableDeclaration */:
                        var name = node.name;
                        if (ts.isBindingPattern(name)) {
                            for (var _c = 0, _d = name.elements; _c < _d.length; _c++) {
                                var el = _d[_c];
                                // mark individual names in binding pattern
                                checkModuleAugmentationElement(el, isGlobalAugmentation);
                            }
                            break;
                        }
                    // falls through
                    case 255 /* ClassDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 254 /* FunctionDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 259 /* ModuleDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                        if (isGlobalAugmentation) {
                            return;
                        }
                        var symbol = getSymbolOfNode(node);
                        if (symbol) {
                            // module augmentations cannot introduce new names on the top level scope of the module
                            // this is done it two steps
                            // 1. quick check - if symbol for node is not merged - this is local symbol to this augmentation - report error
                            // 2. main check - report error if value declaration of the parent symbol is module augmentation)
                            var reportError = !(symbol.flags & 33554432 /* Transient */);
                            if (!reportError) {
                                // symbol should not originate in augmentation
                                reportError = !!((_a = symbol.parent) === null || _a === void 0 ? void 0 : _a.declarations) && ts.isExternalModuleAugmentation(symbol.parent.declarations[0]);
                            }
                        }
                        break;
                }
            }
            function getFirstNonModuleExportsIdentifier(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        return node;
                    case 159 /* QualifiedName */:
                        do {
                            node = node.left;
                        } while (node.kind !== 79 /* Identifier */);
                        return node;
                    case 204 /* PropertyAccessExpression */:
                        do {
                            if (ts.isModuleExportsAccessExpression(node.expression) && !ts.isPrivateIdentifier(node.name)) {
                                return node.name;
                            }
                            node = node.expression;
                        } while (node.kind !== 79 /* Identifier */);
                        return node;
                }
            }
            function checkExternalImportOrExportDeclaration(node) {
                var moduleName = ts.getExternalModuleName(node);
                if (!moduleName || ts.nodeIsMissing(moduleName)) {
                    // Should be a parse error.
                    return false;
                }
                if (!ts.isStringLiteral(moduleName)) {
                    error(moduleName, ts.Diagnostics.String_literal_expected);
                    return false;
                }
                var inAmbientExternalModule = node.parent.kind === 260 /* ModuleBlock */ && ts.isAmbientModule(node.parent.parent);
                if (node.parent.kind !== 300 /* SourceFile */ && !inAmbientExternalModule) {
                    error(moduleName, node.kind === 270 /* ExportDeclaration */ ?
                        ts.Diagnostics.Export_declarations_are_not_permitted_in_a_namespace :
                        ts.Diagnostics.Import_declarations_in_a_namespace_cannot_reference_a_module);
                    return false;
                }
                if (inAmbientExternalModule && ts.isExternalModuleNameRelative(moduleName.text)) {
                    // we have already reported errors on top level imports/exports in external module augmentations in checkModuleDeclaration
                    // no need to do this again.
                    if (!isTopLevelInExternalModuleAugmentation(node)) {
                        // TypeScript 1.0 spec (April 2013): 12.1.6
                        // An ExternalImportDeclaration in an AmbientExternalModuleDeclaration may reference
                        // other external modules only through top - level external module names.
                        // Relative external module names are not permitted.
                        error(node, ts.Diagnostics.Import_or_export_declaration_in_an_ambient_module_declaration_cannot_reference_module_through_relative_module_name);
                        return false;
                    }
                }
                return true;
            }
            function checkAliasSymbol(node) {
                var _a;
                var symbol = getSymbolOfNode(node);
                var target = resolveAlias(symbol);
                if (target !== unknownSymbol) {
                    // For external modules, `symbol` represents the local symbol for an alias.
                    // This local symbol will merge any other local declarations (excluding other aliases)
                    // and symbol.flags will contains combined representation for all merged declaration.
                    // Based on symbol.flags we can compute a set of excluded meanings (meaning that resolved alias should not have,
                    // otherwise it will conflict with some local declaration). Note that in addition to normal flags we include matching SymbolFlags.Export*
                    // in order to prevent collisions with declarations that were exported from the current module (they still contribute to local names).
                    symbol = getMergedSymbol(symbol.exportSymbol || symbol);
                    var excludedMeanings = (symbol.flags & (111551 /* Value */ | 1048576 /* ExportValue */) ? 111551 /* Value */ : 0) |
                        (symbol.flags & 788968 /* Type */ ? 788968 /* Type */ : 0) |
                        (symbol.flags & 1920 /* Namespace */ ? 1920 /* Namespace */ : 0);
                    if (target.flags & excludedMeanings) {
                        var message = node.kind === 273 /* ExportSpecifier */ ?
                            ts.Diagnostics.Export_declaration_conflicts_with_exported_declaration_of_0 :
                            ts.Diagnostics.Import_declaration_conflicts_with_local_declaration_of_0;
                        error(node, message, symbolToString(symbol));
                    }
                    // Don't allow to re-export something with no value side when `--isolatedModules` is set.
                    if (compilerOptions.isolatedModules
                        && node.kind === 273 /* ExportSpecifier */
                        && !node.parent.parent.isTypeOnly
                        && !(target.flags & 111551 /* Value */)
                        && !(node.flags & 8388608 /* Ambient */)) {
                        error(node, ts.Diagnostics.Re_exporting_a_type_when_the_isolatedModules_flag_is_provided_requires_using_export_type);
                    }
                    if (ts.isImportSpecifier(node) && ((_a = target.declarations) === null || _a === void 0 ? void 0 : _a.every(function (d) { return !!(ts.getCombinedNodeFlags(d) & 134217728 /* Deprecated */); }))) {
                        addDeprecatedSuggestion(node.name, target.declarations, symbol.escapedName);
                    }
                }
            }
            function checkImportBinding(node) {
                checkCollisionsForDeclarationName(node, node.name);
                checkAliasSymbol(node);
                if (node.kind === 268 /* ImportSpecifier */ &&
                    ts.idText(node.propertyName || node.name) === "default" &&
                    compilerOptions.esModuleInterop &&
                    moduleKind !== ts.ModuleKind.System && moduleKind < ts.ModuleKind.ES2015) {
                    checkExternalEmitHelpers(node, 131072 /* ImportDefault */);
                }
            }
            function checkImportDeclaration(node) {
                if (checkGrammarModuleElementContext(node, ts.Diagnostics.An_import_declaration_can_only_be_used_in_a_namespace_or_module)) {
                    // If we hit an import declaration in an illegal context, just bail out to avoid cascading errors.
                    return;
                }
                if (!checkGrammarDecoratorsAndModifiers(node) && ts.hasEffectiveModifiers(node)) {
                    grammarErrorOnFirstToken(node, ts.Diagnostics.An_import_declaration_cannot_have_modifiers);
                }
                if (checkExternalImportOrExportDeclaration(node)) {
                    var importClause = node.importClause;
                    if (importClause && !checkGrammarImportClause(importClause)) {
                        if (importClause.name) {
                            checkImportBinding(importClause);
                        }
                        if (importClause.namedBindings) {
                            if (importClause.namedBindings.kind === 266 /* NamespaceImport */) {
                                checkImportBinding(importClause.namedBindings);
                                if (moduleKind !== ts.ModuleKind.System && moduleKind < ts.ModuleKind.ES2015 && compilerOptions.esModuleInterop) {
                                    // import * as ns from "foo";
                                    checkExternalEmitHelpers(node, 65536 /* ImportStar */);
                                }
                            }
                            else {
                                var moduleExisted = resolveExternalModuleName(node, node.moduleSpecifier);
                                if (moduleExisted) {
                                    ts.forEach(importClause.namedBindings.elements, checkImportBinding);
                                }
                            }
                        }
                    }
                }
            }
            function checkImportEqualsDeclaration(node) {
                if (checkGrammarModuleElementContext(node, ts.Diagnostics.An_import_declaration_can_only_be_used_in_a_namespace_or_module)) {
                    // If we hit an import declaration in an illegal context, just bail out to avoid cascading errors.
                    return;
                }
                checkGrammarDecoratorsAndModifiers(node);
                if (ts.isInternalModuleImportEqualsDeclaration(node) || checkExternalImportOrExportDeclaration(node)) {
                    checkImportBinding(node);
                    if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                        markExportAsReferenced(node);
                    }
                    if (node.moduleReference.kind !== 275 /* ExternalModuleReference */) {
                        var target = resolveAlias(getSymbolOfNode(node));
                        if (target !== unknownSymbol) {
                            if (target.flags & 111551 /* Value */) {
                                // Target is a value symbol, check that it is not hidden by a local declaration with the same name
                                var moduleName = ts.getFirstIdentifier(node.moduleReference);
                                if (!(resolveEntityName(moduleName, 111551 /* Value */ | 1920 /* Namespace */).flags & 1920 /* Namespace */)) {
                                    error(moduleName, ts.Diagnostics.Module_0_is_hidden_by_a_local_declaration_with_the_same_name, ts.declarationNameToString(moduleName));
                                }
                            }
                            if (target.flags & 788968 /* Type */) {
                                checkTypeNameIsReserved(node.name, ts.Diagnostics.Import_name_cannot_be_0);
                            }
                        }
                        if (node.isTypeOnly) {
                            grammarErrorOnNode(node, ts.Diagnostics.An_import_alias_cannot_use_import_type);
                        }
                    }
                    else {
                        if (moduleKind >= ts.ModuleKind.ES2015 && !node.isTypeOnly && !(node.flags & 8388608 /* Ambient */)) {
                            // Import equals declaration is deprecated in es6 or above
                            grammarErrorOnNode(node, ts.Diagnostics.Import_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_import_Asterisk_as_ns_from_mod_import_a_from_mod_import_d_from_mod_or_another_module_format_instead);
                        }
                    }
                }
            }
            function checkExportDeclaration(node) {
                if (checkGrammarModuleElementContext(node, ts.Diagnostics.An_export_declaration_can_only_be_used_in_a_module)) {
                    // If we hit an export in an illegal context, just bail out to avoid cascading errors.
                    return;
                }
                if (!checkGrammarDecoratorsAndModifiers(node) && ts.hasEffectiveModifiers(node)) {
                    grammarErrorOnFirstToken(node, ts.Diagnostics.An_export_declaration_cannot_have_modifiers);
                }
                if (node.moduleSpecifier && node.exportClause && ts.isNamedExports(node.exportClause) && ts.length(node.exportClause.elements) && languageVersion === 0 /* ES3 */) {
                    checkExternalEmitHelpers(node, 2097152 /* CreateBinding */);
                }
                checkGrammarExportDeclaration(node);
                if (!node.moduleSpecifier || checkExternalImportOrExportDeclaration(node)) {
                    if (node.exportClause && !ts.isNamespaceExport(node.exportClause)) {
                        // export { x, y }
                        // export { x, y } from "foo"
                        ts.forEach(node.exportClause.elements, checkExportSpecifier);
                        var inAmbientExternalModule = node.parent.kind === 260 /* ModuleBlock */ && ts.isAmbientModule(node.parent.parent);
                        var inAmbientNamespaceDeclaration = !inAmbientExternalModule && node.parent.kind === 260 /* ModuleBlock */ &&
                            !node.moduleSpecifier && node.flags & 8388608 /* Ambient */;
                        if (node.parent.kind !== 300 /* SourceFile */ && !inAmbientExternalModule && !inAmbientNamespaceDeclaration) {
                            error(node, ts.Diagnostics.Export_declarations_are_not_permitted_in_a_namespace);
                        }
                    }
                    else {
                        // export * from "foo"
                        // export * as ns from "foo";
                        var moduleSymbol = resolveExternalModuleName(node, node.moduleSpecifier);
                        if (moduleSymbol && hasExportAssignmentSymbol(moduleSymbol)) {
                            error(node.moduleSpecifier, ts.Diagnostics.Module_0_uses_export_and_cannot_be_used_with_export_Asterisk, symbolToString(moduleSymbol));
                        }
                        else if (node.exportClause) {
                            checkAliasSymbol(node.exportClause);
                        }
                        if (moduleKind !== ts.ModuleKind.System && moduleKind < ts.ModuleKind.ES2015) {
                            if (node.exportClause) {
                                // export * as ns from "foo";
                                // For ES2015 modules, we emit it as a pair of `import * as a_1 ...; export { a_1 as ns }` and don't need the helper.
                                // We only use the helper here when in esModuleInterop
                                if (compilerOptions.esModuleInterop) {
                                    checkExternalEmitHelpers(node, 65536 /* ImportStar */);
                                }
                            }
                            else {
                                // export * from "foo"
                                checkExternalEmitHelpers(node, 32768 /* ExportStar */);
                            }
                        }
                    }
                }
            }
            function checkGrammarExportDeclaration(node) {
                var _a;
                var isTypeOnlyExportStar = node.isTypeOnly && ((_a = node.exportClause) === null || _a === void 0 ? void 0 : _a.kind) !== 271 /* NamedExports */;
                if (isTypeOnlyExportStar) {
                    grammarErrorOnNode(node, ts.Diagnostics.Only_named_exports_may_use_export_type);
                }
                return !isTypeOnlyExportStar;
            }
            function checkGrammarModuleElementContext(node, errorMessage) {
                var isInAppropriateContext = node.parent.kind === 300 /* SourceFile */ || node.parent.kind === 260 /* ModuleBlock */ || node.parent.kind === 259 /* ModuleDeclaration */;
                if (!isInAppropriateContext) {
                    grammarErrorOnFirstToken(node, errorMessage);
                }
                return !isInAppropriateContext;
            }
            function importClauseContainsReferencedImport(importClause) {
                return ts.forEachImportClauseDeclaration(importClause, function (declaration) {
                    return !!getSymbolOfNode(declaration).isReferenced;
                });
            }
            function importClauseContainsConstEnumUsedAsValue(importClause) {
                return ts.forEachImportClauseDeclaration(importClause, function (declaration) {
                    return !!getSymbolLinks(getSymbolOfNode(declaration)).constEnumReferenced;
                });
            }
            function canConvertImportDeclarationToTypeOnly(statement) {
                return ts.isImportDeclaration(statement) &&
                    statement.importClause &&
                    !statement.importClause.isTypeOnly &&
                    importClauseContainsReferencedImport(statement.importClause) &&
                    !isReferencedAliasDeclaration(statement.importClause, /*checkChildren*/ true) &&
                    !importClauseContainsConstEnumUsedAsValue(statement.importClause);
            }
            function canConvertImportEqualsDeclarationToTypeOnly(statement) {
                return ts.isImportEqualsDeclaration(statement) &&
                    ts.isExternalModuleReference(statement.moduleReference) &&
                    !statement.isTypeOnly &&
                    getSymbolOfNode(statement).isReferenced &&
                    !isReferencedAliasDeclaration(statement, /*checkChildren*/ false) &&
                    !getSymbolLinks(getSymbolOfNode(statement)).constEnumReferenced;
            }
            function checkImportsForTypeOnlyConversion(sourceFile) {
                for (var _i = 0, _a = sourceFile.statements; _i < _a.length; _i++) {
                    var statement = _a[_i];
                    if (canConvertImportDeclarationToTypeOnly(statement) || canConvertImportEqualsDeclarationToTypeOnly(statement)) {
                        error(statement, ts.Diagnostics.This_import_is_never_used_as_a_value_and_must_use_import_type_because_importsNotUsedAsValues_is_set_to_error);
                    }
                }
            }
            function checkExportSpecifier(node) {
                checkAliasSymbol(node);
                if (ts.getEmitDeclarations(compilerOptions)) {
                    collectLinkedAliases(node.propertyName || node.name, /*setVisibility*/ true);
                }
                if (!node.parent.parent.moduleSpecifier) {
                    var exportedName = node.propertyName || node.name;
                    // find immediate value referenced by exported name (SymbolFlags.Alias is set so we don't chase down aliases)
                    var symbol = resolveName(exportedName, exportedName.escapedText, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */ | 2097152 /* Alias */, 
                    /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ true);
                    if (symbol && (symbol === undefinedSymbol || symbol === globalThisSymbol || symbol.declarations && isGlobalSourceFile(getDeclarationContainer(symbol.declarations[0])))) {
                        error(exportedName, ts.Diagnostics.Cannot_export_0_Only_local_declarations_can_be_exported_from_a_module, ts.idText(exportedName));
                    }
                    else {
                        markExportAsReferenced(node);
                        var target = symbol && (symbol.flags & 2097152 /* Alias */ ? resolveAlias(symbol) : symbol);
                        if (!target || target === unknownSymbol || target.flags & 111551 /* Value */) {
                            checkExpressionCached(node.propertyName || node.name);
                        }
                    }
                }
                else {
                    if (compilerOptions.esModuleInterop &&
                        moduleKind !== ts.ModuleKind.System &&
                        moduleKind < ts.ModuleKind.ES2015 &&
                        ts.idText(node.propertyName || node.name) === "default") {
                        checkExternalEmitHelpers(node, 131072 /* ImportDefault */);
                    }
                }
            }
            function checkExportAssignment(node) {
                var illegalContextMessage = node.isExportEquals
                    ? ts.Diagnostics.An_export_assignment_must_be_at_the_top_level_of_a_file_or_module_declaration
                    : ts.Diagnostics.A_default_export_must_be_at_the_top_level_of_a_file_or_module_declaration;
                if (checkGrammarModuleElementContext(node, illegalContextMessage)) {
                    // If we hit an export assignment in an illegal context, just bail out to avoid cascading errors.
                    return;
                }
                var container = node.parent.kind === 300 /* SourceFile */ ? node.parent : node.parent.parent;
                if (container.kind === 259 /* ModuleDeclaration */ && !ts.isAmbientModule(container)) {
                    if (node.isExportEquals) {
                        error(node, ts.Diagnostics.An_export_assignment_cannot_be_used_in_a_namespace);
                    }
                    else {
                        error(node, ts.Diagnostics.A_default_export_can_only_be_used_in_an_ECMAScript_style_module);
                    }
                    return;
                }
                // Grammar checking
                if (!checkGrammarDecoratorsAndModifiers(node) && ts.hasEffectiveModifiers(node)) {
                    grammarErrorOnFirstToken(node, ts.Diagnostics.An_export_assignment_cannot_have_modifiers);
                }
                if (node.expression.kind === 79 /* Identifier */) {
                    var id = node.expression;
                    var sym = resolveEntityName(id, 67108863 /* All */, /*ignoreErrors*/ true, /*dontResolveAlias*/ true, node);
                    if (sym) {
                        markAliasReferenced(sym, id);
                        // If not a value, we're interpreting the identifier as a type export, along the lines of (`export { Id as default }`)
                        var target = sym.flags & 2097152 /* Alias */ ? resolveAlias(sym) : sym;
                        if (target === unknownSymbol || target.flags & 111551 /* Value */) {
                            // However if it is a value, we need to check it's being used correctly
                            checkExpressionCached(node.expression);
                        }
                    }
                    else {
                        checkExpressionCached(node.expression); // doesn't resolve, check as expression to mark as error
                    }
                    if (ts.getEmitDeclarations(compilerOptions)) {
                        collectLinkedAliases(node.expression, /*setVisibility*/ true);
                    }
                }
                else {
                    checkExpressionCached(node.expression);
                }
                checkExternalModuleExports(container);
                if ((node.flags & 8388608 /* Ambient */) && !ts.isEntityNameExpression(node.expression)) {
                    grammarErrorOnNode(node.expression, ts.Diagnostics.The_expression_of_an_export_assignment_must_be_an_identifier_or_qualified_name_in_an_ambient_context);
                }
                if (node.isExportEquals && !(node.flags & 8388608 /* Ambient */)) {
                    if (moduleKind >= ts.ModuleKind.ES2015) {
                        // export assignment is not supported in es6 modules
                        grammarErrorOnNode(node, ts.Diagnostics.Export_assignment_cannot_be_used_when_targeting_ECMAScript_modules_Consider_using_export_default_or_another_module_format_instead);
                    }
                    else if (moduleKind === ts.ModuleKind.System) {
                        // system modules does not support export assignment
                        grammarErrorOnNode(node, ts.Diagnostics.Export_assignment_is_not_supported_when_module_flag_is_system);
                    }
                }
            }
            function hasExportedMembers(moduleSymbol) {
                return ts.forEachEntry(moduleSymbol.exports, function (_, id) { return id !== "export="; });
            }
            function checkExternalModuleExports(node) {
                var moduleSymbol = getSymbolOfNode(node);
                var links = getSymbolLinks(moduleSymbol);
                if (!links.exportsChecked) {
                    var exportEqualsSymbol = moduleSymbol.exports.get("export=");
                    if (exportEqualsSymbol && hasExportedMembers(moduleSymbol)) {
                        var declaration = getDeclarationOfAliasSymbol(exportEqualsSymbol) || exportEqualsSymbol.valueDeclaration;
                        if (declaration && !isTopLevelInExternalModuleAugmentation(declaration) && !ts.isInJSFile(declaration)) {
                            error(declaration, ts.Diagnostics.An_export_assignment_cannot_be_used_in_a_module_with_other_exported_elements);
                        }
                    }
                    // Checks for export * conflicts
                    var exports_2 = getExportsOfModule(moduleSymbol);
                    if (exports_2) {
                        exports_2.forEach(function (_a, id) {
                            var declarations = _a.declarations, flags = _a.flags;
                            if (id === "__export") {
                                return;
                            }
                            // ECMA262: 15.2.1.1 It is a Syntax Error if the ExportedNames of ModuleItemList contains any duplicate entries.
                            // (TS Exceptions: namespaces, function overloads, enums, and interfaces)
                            if (flags & (1920 /* Namespace */ | 64 /* Interface */ | 384 /* Enum */)) {
                                return;
                            }
                            var exportedDeclarationsCount = ts.countWhere(declarations, isNotOverloadAndNotAccessor);
                            if (flags & 524288 /* TypeAlias */ && exportedDeclarationsCount <= 2) {
                                // it is legal to merge type alias with other values
                                // so count should be either 1 (just type alias) or 2 (type alias + merged value)
                                return;
                            }
                            if (exportedDeclarationsCount > 1) {
                                if (!isDuplicatedCommonJSExport(declarations)) {
                                    for (var _i = 0, _b = declarations; _i < _b.length; _i++) {
                                        var declaration = _b[_i];
                                        if (isNotOverload(declaration)) {
                                            diagnostics.add(ts.createDiagnosticForNode(declaration, ts.Diagnostics.Cannot_redeclare_exported_variable_0, ts.unescapeLeadingUnderscores(id)));
                                        }
                                    }
                                }
                            }
                        });
                    }
                    links.exportsChecked = true;
                }
            }
            function isDuplicatedCommonJSExport(declarations) {
                return declarations
                    && declarations.length > 1
                    && declarations.every(function (d) { return ts.isInJSFile(d) && ts.isAccessExpression(d) && (ts.isExportsIdentifier(d.expression) || ts.isModuleExportsAccessExpression(d.expression)); });
            }
            function checkSourceElement(node) {
                if (node) {
                    var saveCurrentNode = currentNode;
                    currentNode = node;
                    instantiationCount = 0;
                    checkSourceElementWorker(node);
                    currentNode = saveCurrentNode;
                }
            }
            function checkSourceElementWorker(node) {
                if (ts.isInJSFile(node)) {
                    ts.forEach(node.jsDoc, function (_a) {
                        var tags = _a.tags;
                        return ts.forEach(tags, checkSourceElement);
                    });
                }
                var kind = node.kind;
                if (cancellationToken) {
                    // Only bother checking on a few construct kinds.  We don't want to be excessively
                    // hitting the cancellation token on every node we check.
                    switch (kind) {
                        case 259 /* ModuleDeclaration */:
                        case 255 /* ClassDeclaration */:
                        case 256 /* InterfaceDeclaration */:
                        case 254 /* FunctionDeclaration */:
                            cancellationToken.throwIfCancellationRequested();
                    }
                }
                if (kind >= 235 /* FirstStatement */ && kind <= 251 /* LastStatement */ && node.flowNode && !isReachableFlowNode(node.flowNode)) {
                    errorOrSuggestion(compilerOptions.allowUnreachableCode === false, node, ts.Diagnostics.Unreachable_code_detected);
                }
                switch (kind) {
                    case 161 /* TypeParameter */:
                        return checkTypeParameter(node);
                    case 162 /* Parameter */:
                        return checkParameter(node);
                    case 165 /* PropertyDeclaration */:
                        return checkPropertyDeclaration(node);
                    case 164 /* PropertySignature */:
                        return checkPropertySignature(node);
                    case 178 /* ConstructorType */:
                    case 177 /* FunctionType */:
                    case 172 /* CallSignature */:
                    case 173 /* ConstructSignature */:
                    case 174 /* IndexSignature */:
                        return checkSignatureDeclaration(node);
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                        return checkMethodDeclaration(node);
                    case 168 /* ClassStaticBlockDeclaration */:
                        return checkClassStaticBlockDeclaration(node);
                    case 169 /* Constructor */:
                        return checkConstructorDeclaration(node);
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return checkAccessorDeclaration(node);
                    case 176 /* TypeReference */:
                        return checkTypeReferenceNode(node);
                    case 175 /* TypePredicate */:
                        return checkTypePredicate(node);
                    case 179 /* TypeQuery */:
                        return checkTypeQuery(node);
                    case 180 /* TypeLiteral */:
                        return checkTypeLiteral(node);
                    case 181 /* ArrayType */:
                        return checkArrayType(node);
                    case 182 /* TupleType */:
                        return checkTupleType(node);
                    case 185 /* UnionType */:
                    case 186 /* IntersectionType */:
                        return checkUnionOrIntersectionType(node);
                    case 189 /* ParenthesizedType */:
                    case 183 /* OptionalType */:
                    case 184 /* RestType */:
                        return checkSourceElement(node.type);
                    case 190 /* ThisType */:
                        return checkThisType(node);
                    case 191 /* TypeOperator */:
                        return checkTypeOperator(node);
                    case 187 /* ConditionalType */:
                        return checkConditionalType(node);
                    case 188 /* InferType */:
                        return checkInferType(node);
                    case 196 /* TemplateLiteralType */:
                        return checkTemplateLiteralType(node);
                    case 198 /* ImportType */:
                        return checkImportType(node);
                    case 195 /* NamedTupleMember */:
                        return checkNamedTupleMember(node);
                    case 323 /* JSDocAugmentsTag */:
                        return checkJSDocAugmentsTag(node);
                    case 324 /* JSDocImplementsTag */:
                        return checkJSDocImplementsTag(node);
                    case 340 /* JSDocTypedefTag */:
                    case 333 /* JSDocCallbackTag */:
                    case 334 /* JSDocEnumTag */:
                        return checkJSDocTypeAliasTag(node);
                    case 339 /* JSDocTemplateTag */:
                        return checkJSDocTemplateTag(node);
                    case 338 /* JSDocTypeTag */:
                        return checkJSDocTypeTag(node);
                    case 335 /* JSDocParameterTag */:
                        return checkJSDocParameterTag(node);
                    case 342 /* JSDocPropertyTag */:
                        return checkJSDocPropertyTag(node);
                    case 312 /* JSDocFunctionType */:
                        checkJSDocFunctionType(node);
                    // falls through
                    case 310 /* JSDocNonNullableType */:
                    case 309 /* JSDocNullableType */:
                    case 307 /* JSDocAllType */:
                    case 308 /* JSDocUnknownType */:
                    case 317 /* JSDocTypeLiteral */:
                        checkJSDocTypeIsInJsFile(node);
                        ts.forEachChild(node, checkSourceElement);
                        return;
                    case 313 /* JSDocVariadicType */:
                        checkJSDocVariadicType(node);
                        return;
                    case 304 /* JSDocTypeExpression */:
                        return checkSourceElement(node.type);
                    case 192 /* IndexedAccessType */:
                        return checkIndexedAccessType(node);
                    case 193 /* MappedType */:
                        return checkMappedType(node);
                    case 254 /* FunctionDeclaration */:
                        return checkFunctionDeclaration(node);
                    case 233 /* Block */:
                    case 260 /* ModuleBlock */:
                        return checkBlock(node);
                    case 235 /* VariableStatement */:
                        return checkVariableStatement(node);
                    case 236 /* ExpressionStatement */:
                        return checkExpressionStatement(node);
                    case 237 /* IfStatement */:
                        return checkIfStatement(node);
                    case 238 /* DoStatement */:
                        return checkDoStatement(node);
                    case 239 /* WhileStatement */:
                        return checkWhileStatement(node);
                    case 240 /* ForStatement */:
                        return checkForStatement(node);
                    case 241 /* ForInStatement */:
                        return checkForInStatement(node);
                    case 242 /* ForOfStatement */:
                        return checkForOfStatement(node);
                    case 243 /* ContinueStatement */:
                    case 244 /* BreakStatement */:
                        return checkBreakOrContinueStatement(node);
                    case 245 /* ReturnStatement */:
                        return checkReturnStatement(node);
                    case 246 /* WithStatement */:
                        return checkWithStatement(node);
                    case 247 /* SwitchStatement */:
                        return checkSwitchStatement(node);
                    case 248 /* LabeledStatement */:
                        return checkLabeledStatement(node);
                    case 249 /* ThrowStatement */:
                        return checkThrowStatement(node);
                    case 250 /* TryStatement */:
                        return checkTryStatement(node);
                    case 252 /* VariableDeclaration */:
                        return checkVariableDeclaration(node);
                    case 201 /* BindingElement */:
                        return checkBindingElement(node);
                    case 255 /* ClassDeclaration */:
                        return checkClassDeclaration(node);
                    case 256 /* InterfaceDeclaration */:
                        return checkInterfaceDeclaration(node);
                    case 257 /* TypeAliasDeclaration */:
                        return checkTypeAliasDeclaration(node);
                    case 258 /* EnumDeclaration */:
                        return checkEnumDeclaration(node);
                    case 259 /* ModuleDeclaration */:
                        return checkModuleDeclaration(node);
                    case 264 /* ImportDeclaration */:
                        return checkImportDeclaration(node);
                    case 263 /* ImportEqualsDeclaration */:
                        return checkImportEqualsDeclaration(node);
                    case 270 /* ExportDeclaration */:
                        return checkExportDeclaration(node);
                    case 269 /* ExportAssignment */:
                        return checkExportAssignment(node);
                    case 234 /* EmptyStatement */:
                    case 251 /* DebuggerStatement */:
                        checkGrammarStatementInAmbientContext(node);
                        return;
                    case 274 /* MissingDeclaration */:
                        return checkMissingDeclaration(node);
                }
            }
            function checkJSDocTypeIsInJsFile(node) {
                if (!ts.isInJSFile(node)) {
                    grammarErrorOnNode(node, ts.Diagnostics.JSDoc_types_can_only_be_used_inside_documentation_comments);
                }
            }
            function checkJSDocVariadicType(node) {
                checkJSDocTypeIsInJsFile(node);
                checkSourceElement(node.type);
                // Only legal location is in the *last* parameter tag or last parameter of a JSDoc function.
                var parent = node.parent;
                if (ts.isParameter(parent) && ts.isJSDocFunctionType(parent.parent)) {
                    if (ts.last(parent.parent.parameters) !== parent) {
                        error(node, ts.Diagnostics.A_rest_parameter_must_be_last_in_a_parameter_list);
                    }
                    return;
                }
                if (!ts.isJSDocTypeExpression(parent)) {
                    error(node, ts.Diagnostics.JSDoc_may_only_appear_in_the_last_parameter_of_a_signature);
                }
                var paramTag = node.parent.parent;
                if (!ts.isJSDocParameterTag(paramTag)) {
                    error(node, ts.Diagnostics.JSDoc_may_only_appear_in_the_last_parameter_of_a_signature);
                    return;
                }
                var param = ts.getParameterSymbolFromJSDoc(paramTag);
                if (!param) {
                    // We will error in `checkJSDocParameterTag`.
                    return;
                }
                var host = ts.getHostSignatureFromJSDoc(paramTag);
                if (!host || ts.last(host.parameters).symbol !== param) {
                    error(node, ts.Diagnostics.A_rest_parameter_must_be_last_in_a_parameter_list);
                }
            }
            function getTypeFromJSDocVariadicType(node) {
                var type = getTypeFromTypeNode(node.type);
                var parent = node.parent;
                var paramTag = node.parent.parent;
                if (ts.isJSDocTypeExpression(node.parent) && ts.isJSDocParameterTag(paramTag)) {
                    // Else we will add a diagnostic, see `checkJSDocVariadicType`.
                    var host_1 = ts.getHostSignatureFromJSDoc(paramTag);
                    var isCallbackTag = ts.isJSDocCallbackTag(paramTag.parent.parent);
                    if (host_1 || isCallbackTag) {
                        /*
                        Only return an array type if the corresponding parameter is marked as a rest parameter, or if there are no parameters.
                        So in the following situation we will not create an array type:
                            /** @param {...number} a * /
                            function f(a) {}
                        Because `a` will just be of type `number | undefined`. A synthetic `...args` will also be added, which *will* get an array type.
                        */
                        var lastParamDeclaration = isCallbackTag
                            ? ts.lastOrUndefined(paramTag.parent.parent.typeExpression.parameters)
                            : ts.lastOrUndefined(host_1.parameters);
                        var symbol = ts.getParameterSymbolFromJSDoc(paramTag);
                        if (!lastParamDeclaration ||
                            symbol && lastParamDeclaration.symbol === symbol && ts.isRestParameter(lastParamDeclaration)) {
                            return createArrayType(type);
                        }
                    }
                }
                if (ts.isParameter(parent) && ts.isJSDocFunctionType(parent.parent)) {
                    return createArrayType(type);
                }
                return addOptionality(type);
            }
            // Function and class expression bodies are checked after all statements in the enclosing body. This is
            // to ensure constructs like the following are permitted:
            //     const foo = function () {
            //        const s = foo();
            //        return "hello";
            //     }
            // Here, performing a full type check of the body of the function expression whilst in the process of
            // determining the type of foo would cause foo to be given type any because of the recursive reference.
            // Delaying the type check of the body ensures foo has been assigned a type.
            function checkNodeDeferred(node) {
                var enclosingFile = ts.getSourceFileOfNode(node);
                var links = getNodeLinks(enclosingFile);
                if (!(links.flags & 1 /* TypeChecked */)) {
                    links.deferredNodes = links.deferredNodes || new ts.Map();
                    var id = getNodeId(node);
                    links.deferredNodes.set(id, node);
                }
            }
            function checkDeferredNodes(context) {
                var links = getNodeLinks(context);
                if (links.deferredNodes) {
                    links.deferredNodes.forEach(checkDeferredNode);
                }
            }
            function checkDeferredNode(node) {
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("check" /* Check */, "checkDeferredNode", { kind: node.kind, pos: node.pos, end: node.end });
                var saveCurrentNode = currentNode;
                currentNode = node;
                instantiationCount = 0;
                switch (node.kind) {
                    case 206 /* CallExpression */:
                    case 207 /* NewExpression */:
                    case 208 /* TaggedTemplateExpression */:
                    case 163 /* Decorator */:
                    case 278 /* JsxOpeningElement */:
                        // These node kinds are deferred checked when overload resolution fails
                        // To save on work, we ensure the arguments are checked just once, in
                        // a deferred way
                        resolveUntypedCall(node);
                        break;
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                        checkFunctionExpressionOrObjectLiteralMethodDeferred(node);
                        break;
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        checkAccessorDeclaration(node);
                        break;
                    case 224 /* ClassExpression */:
                        checkClassExpressionDeferred(node);
                        break;
                    case 277 /* JsxSelfClosingElement */:
                        checkJsxSelfClosingElementDeferred(node);
                        break;
                    case 276 /* JsxElement */:
                        checkJsxElementDeferred(node);
                        break;
                }
                currentNode = saveCurrentNode;
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
            }
            function checkSourceFile(node) {
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("check" /* Check */, "checkSourceFile", { path: node.path }, /*separateBeginAndEnd*/ true);
                ts.performance.mark("beforeCheck");
                checkSourceFileWorker(node);
                ts.performance.mark("afterCheck");
                ts.performance.measure("Check", "beforeCheck", "afterCheck");
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
            }
            function unusedIsError(kind, isAmbient) {
                if (isAmbient) {
                    return false;
                }
                switch (kind) {
                    case 0 /* Local */:
                        return !!compilerOptions.noUnusedLocals;
                    case 1 /* Parameter */:
                        return !!compilerOptions.noUnusedParameters;
                    default:
                        return ts.Debug.assertNever(kind);
                }
            }
            function getPotentiallyUnusedIdentifiers(sourceFile) {
                return allPotentiallyUnusedIdentifiers.get(sourceFile.path) || ts.emptyArray;
            }
            // Fully type check a source file and collect the relevant diagnostics.
            function checkSourceFileWorker(node) {
                var links = getNodeLinks(node);
                if (!(links.flags & 1 /* TypeChecked */)) {
                    if (ts.skipTypeChecking(node, compilerOptions, host)) {
                        return;
                    }
                    // Grammar checking
                    checkGrammarSourceFile(node);
                    ts.clear(potentialThisCollisions);
                    ts.clear(potentialNewTargetCollisions);
                    ts.clear(potentialWeakMapSetCollisions);
                    ts.clear(potentialReflectCollisions);
                    ts.forEach(node.statements, checkSourceElement);
                    checkSourceElement(node.endOfFileToken);
                    checkDeferredNodes(node);
                    if (ts.isExternalOrCommonJsModule(node)) {
                        registerForUnusedIdentifiersCheck(node);
                    }
                    if (!node.isDeclarationFile && (compilerOptions.noUnusedLocals || compilerOptions.noUnusedParameters)) {
                        checkUnusedIdentifiers(getPotentiallyUnusedIdentifiers(node), function (containingNode, kind, diag) {
                            if (!ts.containsParseError(containingNode) && unusedIsError(kind, !!(containingNode.flags & 8388608 /* Ambient */))) {
                                diagnostics.add(diag);
                            }
                        });
                    }
                    if (compilerOptions.importsNotUsedAsValues === 2 /* Error */ &&
                        !node.isDeclarationFile &&
                        ts.isExternalModule(node)) {
                        checkImportsForTypeOnlyConversion(node);
                    }
                    if (ts.isExternalOrCommonJsModule(node)) {
                        checkExternalModuleExports(node);
                    }
                    if (potentialThisCollisions.length) {
                        ts.forEach(potentialThisCollisions, checkIfThisIsCapturedInEnclosingScope);
                        ts.clear(potentialThisCollisions);
                    }
                    if (potentialNewTargetCollisions.length) {
                        ts.forEach(potentialNewTargetCollisions, checkIfNewTargetIsCapturedInEnclosingScope);
                        ts.clear(potentialNewTargetCollisions);
                    }
                    if (potentialWeakMapSetCollisions.length) {
                        ts.forEach(potentialWeakMapSetCollisions, checkWeakMapSetCollision);
                        ts.clear(potentialWeakMapSetCollisions);
                    }
                    if (potentialReflectCollisions.length) {
                        ts.forEach(potentialReflectCollisions, checkReflectCollision);
                        ts.clear(potentialReflectCollisions);
                    }
                    links.flags |= 1 /* TypeChecked */;
                }
            }
            function getDiagnostics(sourceFile, ct) {
                try {
                    // Record the cancellation token so it can be checked later on during checkSourceElement.
                    // Do this in a finally block so we can ensure that it gets reset back to nothing after
                    // this call is done.
                    cancellationToken = ct;
                    return getDiagnosticsWorker(sourceFile);
                }
                finally {
                    cancellationToken = undefined;
                }
            }
            function getDiagnosticsWorker(sourceFile) {
                throwIfNonDiagnosticsProducing();
                if (sourceFile) {
                    // Some global diagnostics are deferred until they are needed and
                    // may not be reported in the first call to getGlobalDiagnostics.
                    // We should catch these changes and report them.
                    var previousGlobalDiagnostics = diagnostics.getGlobalDiagnostics();
                    var previousGlobalDiagnosticsSize = previousGlobalDiagnostics.length;
                    checkSourceFile(sourceFile);
                    var semanticDiagnostics = diagnostics.getDiagnostics(sourceFile.fileName);
                    var currentGlobalDiagnostics = diagnostics.getGlobalDiagnostics();
                    if (currentGlobalDiagnostics !== previousGlobalDiagnostics) {
                        // If the arrays are not the same reference, new diagnostics were added.
                        var deferredGlobalDiagnostics = ts.relativeComplement(previousGlobalDiagnostics, currentGlobalDiagnostics, ts.compareDiagnostics);
                        return ts.concatenate(deferredGlobalDiagnostics, semanticDiagnostics);
                    }
                    else if (previousGlobalDiagnosticsSize === 0 && currentGlobalDiagnostics.length > 0) {
                        // If the arrays are the same reference, but the length has changed, a single
                        // new diagnostic was added as DiagnosticCollection attempts to reuse the
                        // same array.
                        return ts.concatenate(currentGlobalDiagnostics, semanticDiagnostics);
                    }
                    return semanticDiagnostics;
                }
                // Global diagnostics are always added when a file is not provided to
                // getDiagnostics
                ts.forEach(host.getSourceFiles(), checkSourceFile);
                return diagnostics.getDiagnostics();
            }
            function getGlobalDiagnostics() {
                throwIfNonDiagnosticsProducing();
                return diagnostics.getGlobalDiagnostics();
            }
            function throwIfNonDiagnosticsProducing() {
                if (!produceDiagnostics) {
                    throw new Error("Trying to get diagnostics from a type checker that does not produce them.");
                }
            }
            // Language service support
            function getSymbolsInScope(location, meaning) {
                if (location.flags & 16777216 /* InWithStatement */) {
                    // We cannot answer semantic questions within a with block, do not proceed any further
                    return [];
                }
                var symbols = ts.createSymbolTable();
                var isStaticSymbol = false;
                populateSymbols();
                symbols.delete("this" /* This */); // Not a symbol, a keyword
                return symbolsToArray(symbols);
                function populateSymbols() {
                    while (location) {
                        if (location.locals && !isGlobalSourceFile(location)) {
                            copySymbols(location.locals, meaning);
                        }
                        switch (location.kind) {
                            case 300 /* SourceFile */:
                                if (!ts.isExternalModule(location))
                                    break;
                            // falls through
                            case 259 /* ModuleDeclaration */:
                                copyLocallyVisibleExportSymbols(getSymbolOfNode(location).exports, meaning & 2623475 /* ModuleMember */);
                                break;
                            case 258 /* EnumDeclaration */:
                                copySymbols(getSymbolOfNode(location).exports, meaning & 8 /* EnumMember */);
                                break;
                            case 224 /* ClassExpression */:
                                var className = location.name;
                                if (className) {
                                    copySymbol(location.symbol, meaning);
                                }
                            // this fall-through is necessary because we would like to handle
                            // type parameter inside class expression similar to how we handle it in classDeclaration and interface Declaration.
                            // falls through
                            case 255 /* ClassDeclaration */:
                            case 256 /* InterfaceDeclaration */:
                                // If we didn't come from static member of class or interface,
                                // add the type parameters into the symbol table
                                // (type parameters of classDeclaration/classExpression and interface are in member property of the symbol.
                                // Note: that the memberFlags come from previous iteration.
                                if (!isStaticSymbol) {
                                    copySymbols(getMembersOfSymbol(getSymbolOfNode(location)), meaning & 788968 /* Type */);
                                }
                                break;
                            case 211 /* FunctionExpression */:
                                var funcName = location.name;
                                if (funcName) {
                                    copySymbol(location.symbol, meaning);
                                }
                                break;
                        }
                        if (ts.introducesArgumentsExoticObject(location)) {
                            copySymbol(argumentsSymbol, meaning);
                        }
                        isStaticSymbol = ts.isStatic(location);
                        location = location.parent;
                    }
                    copySymbols(globals, meaning);
                }
                /**
                 * Copy the given symbol into symbol tables if the symbol has the given meaning
                 * and it doesn't already existed in the symbol table
                 * @param key a key for storing in symbol table; if undefined, use symbol.name
                 * @param symbol the symbol to be added into symbol table
                 * @param meaning meaning of symbol to filter by before adding to symbol table
                 */
                function copySymbol(symbol, meaning) {
                    if (ts.getCombinedLocalAndExportSymbolFlags(symbol) & meaning) {
                        var id = symbol.escapedName;
                        // We will copy all symbol regardless of its reserved name because
                        // symbolsToArray will check whether the key is a reserved name and
                        // it will not copy symbol with reserved name to the array
                        if (!symbols.has(id)) {
                            symbols.set(id, symbol);
                        }
                    }
                }
                function copySymbols(source, meaning) {
                    if (meaning) {
                        source.forEach(function (symbol) {
                            copySymbol(symbol, meaning);
                        });
                    }
                }
                function copyLocallyVisibleExportSymbols(source, meaning) {
                    if (meaning) {
                        source.forEach(function (symbol) {
                            // Similar condition as in `resolveNameHelper`
                            if (!ts.getDeclarationOfKind(symbol, 273 /* ExportSpecifier */) && !ts.getDeclarationOfKind(symbol, 272 /* NamespaceExport */)) {
                                copySymbol(symbol, meaning);
                            }
                        });
                    }
                }
            }
            function isTypeDeclarationName(name) {
                return name.kind === 79 /* Identifier */ &&
                    isTypeDeclaration(name.parent) &&
                    ts.getNameOfDeclaration(name.parent) === name;
            }
            function isTypeDeclaration(node) {
                switch (node.kind) {
                    case 161 /* TypeParameter */:
                    case 255 /* ClassDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 340 /* JSDocTypedefTag */:
                    case 333 /* JSDocCallbackTag */:
                    case 334 /* JSDocEnumTag */:
                        return true;
                    case 265 /* ImportClause */:
                        return node.isTypeOnly;
                    case 268 /* ImportSpecifier */:
                    case 273 /* ExportSpecifier */:
                        return node.parent.parent.isTypeOnly;
                    default:
                        return false;
                }
            }
            // True if the given identifier is part of a type reference
            function isTypeReferenceIdentifier(node) {
                while (node.parent.kind === 159 /* QualifiedName */) {
                    node = node.parent;
                }
                return node.parent.kind === 176 /* TypeReference */;
            }
            function isHeritageClauseElementIdentifier(node) {
                while (node.parent.kind === 204 /* PropertyAccessExpression */) {
                    node = node.parent;
                }
                return node.parent.kind === 226 /* ExpressionWithTypeArguments */;
            }
            function forEachEnclosingClass(node, callback) {
                var result;
                while (true) {
                    node = ts.getContainingClass(node);
                    if (!node)
                        break;
                    if (result = callback(node))
                        break;
                }
                return result;
            }
            function isNodeUsedDuringClassInitialization(node) {
                return !!ts.findAncestor(node, function (element) {
                    if (ts.isConstructorDeclaration(element) && ts.nodeIsPresent(element.body) || ts.isPropertyDeclaration(element)) {
                        return true;
                    }
                    else if (ts.isClassLike(element) || ts.isFunctionLikeDeclaration(element)) {
                        return "quit";
                    }
                    return false;
                });
            }
            function isNodeWithinClass(node, classDeclaration) {
                return !!forEachEnclosingClass(node, function (n) { return n === classDeclaration; });
            }
            function getLeftSideOfImportEqualsOrExportAssignment(nodeOnRightSide) {
                while (nodeOnRightSide.parent.kind === 159 /* QualifiedName */) {
                    nodeOnRightSide = nodeOnRightSide.parent;
                }
                if (nodeOnRightSide.parent.kind === 263 /* ImportEqualsDeclaration */) {
                    return nodeOnRightSide.parent.moduleReference === nodeOnRightSide ? nodeOnRightSide.parent : undefined;
                }
                if (nodeOnRightSide.parent.kind === 269 /* ExportAssignment */) {
                    return nodeOnRightSide.parent.expression === nodeOnRightSide ? nodeOnRightSide.parent : undefined;
                }
                return undefined;
            }
            function isInRightSideOfImportOrExportAssignment(node) {
                return getLeftSideOfImportEqualsOrExportAssignment(node) !== undefined;
            }
            function getSpecialPropertyAssignmentSymbolFromEntityName(entityName) {
                var specialPropertyAssignmentKind = ts.getAssignmentDeclarationKind(entityName.parent.parent);
                switch (specialPropertyAssignmentKind) {
                    case 1 /* ExportsProperty */:
                    case 3 /* PrototypeProperty */:
                        return getSymbolOfNode(entityName.parent);
                    case 4 /* ThisProperty */:
                    case 2 /* ModuleExports */:
                    case 5 /* Property */:
                        return getSymbolOfNode(entityName.parent.parent);
                }
            }
            function isImportTypeQualifierPart(node) {
                var parent = node.parent;
                while (ts.isQualifiedName(parent)) {
                    node = parent;
                    parent = parent.parent;
                }
                if (parent && parent.kind === 198 /* ImportType */ && parent.qualifier === node) {
                    return parent;
                }
                return undefined;
            }
            function getSymbolOfNameOrPropertyAccessExpression(name) {
                if (ts.isDeclarationName(name)) {
                    return getSymbolOfNode(name.parent);
                }
                if (ts.isInJSFile(name) &&
                    name.parent.kind === 204 /* PropertyAccessExpression */ &&
                    name.parent === name.parent.parent.left) {
                    // Check if this is a special property assignment
                    if (!ts.isPrivateIdentifier(name) && !ts.isJSDocMemberName(name)) {
                        var specialPropertyAssignmentSymbol = getSpecialPropertyAssignmentSymbolFromEntityName(name);
                        if (specialPropertyAssignmentSymbol) {
                            return specialPropertyAssignmentSymbol;
                        }
                    }
                }
                if (name.parent.kind === 269 /* ExportAssignment */ && ts.isEntityNameExpression(name)) {
                    // Even an entity name expression that doesn't resolve as an entityname may still typecheck as a property access expression
                    var success = resolveEntityName(name, 
                    /*all meanings*/ 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */ | 2097152 /* Alias */, /*ignoreErrors*/ true);
                    if (success && success !== unknownSymbol) {
                        return success;
                    }
                }
                else if (ts.isEntityName(name) && isInRightSideOfImportOrExportAssignment(name)) {
                    // Since we already checked for ExportAssignment, this really could only be an Import
                    var importEqualsDeclaration = ts.getAncestor(name, 263 /* ImportEqualsDeclaration */);
                    ts.Debug.assert(importEqualsDeclaration !== undefined);
                    return getSymbolOfPartOfRightHandSideOfImportEquals(name, /*dontResolveAlias*/ true);
                }
                if (ts.isEntityName(name)) {
                    var possibleImportNode = isImportTypeQualifierPart(name);
                    if (possibleImportNode) {
                        getTypeFromTypeNode(possibleImportNode);
                        var sym = getNodeLinks(name).resolvedSymbol;
                        return sym === unknownSymbol ? undefined : sym;
                    }
                }
                while (ts.isRightSideOfQualifiedNameOrPropertyAccessOrJSDocMemberName(name)) {
                    name = name.parent;
                }
                if (isHeritageClauseElementIdentifier(name)) {
                    var meaning = 0 /* None */;
                    // In an interface or class, we're definitely interested in a type.
                    if (name.parent.kind === 226 /* ExpressionWithTypeArguments */) {
                        meaning = 788968 /* Type */;
                        // In a class 'extends' clause we are also looking for a value.
                        if (ts.isExpressionWithTypeArgumentsInClassExtendsClause(name.parent)) {
                            meaning |= 111551 /* Value */;
                        }
                    }
                    else {
                        meaning = 1920 /* Namespace */;
                    }
                    meaning |= 2097152 /* Alias */;
                    var entityNameSymbol = ts.isEntityNameExpression(name) ? resolveEntityName(name, meaning) : undefined;
                    if (entityNameSymbol) {
                        return entityNameSymbol;
                    }
                }
                if (name.parent.kind === 335 /* JSDocParameterTag */) {
                    return ts.getParameterSymbolFromJSDoc(name.parent);
                }
                if (name.parent.kind === 161 /* TypeParameter */ && name.parent.parent.kind === 339 /* JSDocTemplateTag */) {
                    ts.Debug.assert(!ts.isInJSFile(name)); // Otherwise `isDeclarationName` would have been true.
                    var typeParameter = ts.getTypeParameterFromJsDoc(name.parent);
                    return typeParameter && typeParameter.symbol;
                }
                if (ts.isExpressionNode(name)) {
                    if (ts.nodeIsMissing(name)) {
                        // Missing entity name.
                        return undefined;
                    }
                    var isJSDoc_1 = ts.findAncestor(name, ts.or(ts.isJSDocLinkLike, ts.isJSDocNameReference, ts.isJSDocMemberName));
                    var meaning = isJSDoc_1 ? 788968 /* Type */ | 1920 /* Namespace */ | 111551 /* Value */ : 111551 /* Value */;
                    if (name.kind === 79 /* Identifier */) {
                        if (ts.isJSXTagName(name) && isJsxIntrinsicIdentifier(name)) {
                            var symbol = getIntrinsicTagSymbol(name.parent);
                            return symbol === unknownSymbol ? undefined : symbol;
                        }
                        var result = resolveEntityName(name, meaning, /*ignoreErrors*/ false, /*dontResolveAlias*/ !isJSDoc_1, ts.getHostSignatureFromJSDoc(name));
                        if (!result && isJSDoc_1) {
                            var container = ts.findAncestor(name, ts.or(ts.isClassLike, ts.isInterfaceDeclaration));
                            if (container) {
                                return resolveJSDocMemberName(name, getSymbolOfNode(container));
                            }
                        }
                        return result;
                    }
                    else if (name.kind === 204 /* PropertyAccessExpression */ || name.kind === 159 /* QualifiedName */) {
                        var links = getNodeLinks(name);
                        if (links.resolvedSymbol) {
                            return links.resolvedSymbol;
                        }
                        if (name.kind === 204 /* PropertyAccessExpression */) {
                            checkPropertyAccessExpression(name, 0 /* Normal */);
                        }
                        else {
                            checkQualifiedName(name, 0 /* Normal */);
                        }
                        if (!links.resolvedSymbol && isJSDoc_1 && ts.isQualifiedName(name)) {
                            return resolveJSDocMemberName(name);
                        }
                        return links.resolvedSymbol;
                    }
                    else if (ts.isJSDocMemberName(name)) {
                        return resolveJSDocMemberName(name);
                    }
                }
                else if (isTypeReferenceIdentifier(name)) {
                    var meaning = name.parent.kind === 176 /* TypeReference */ ? 788968 /* Type */ : 1920 /* Namespace */;
                    return resolveEntityName(name, meaning, /*ignoreErrors*/ false, /*dontResolveAlias*/ true);
                }
                if (name.parent.kind === 175 /* TypePredicate */) {
                    return resolveEntityName(name, /*meaning*/ 1 /* FunctionScopedVariable */);
                }
                return undefined;
            }
            /**
             * Recursively resolve entity names and jsdoc instance references:
             * 1. K#m as K.prototype.m for a class (or other value) K
             * 2. K.m as K.prototype.m
             * 3. I.m as I.m for a type I, or any other I.m that fails to resolve in (1) or (2)
             *
             * For unqualified names, a container K may be provided as a second argument.
             */
            function resolveJSDocMemberName(name, container) {
                if (ts.isEntityName(name)) {
                    // resolve static values first
                    var meaning = 788968 /* Type */ | 1920 /* Namespace */ | 111551 /* Value */;
                    var symbol = resolveEntityName(name, meaning, /*ignoreErrors*/ false, /*dontResolveAlias*/ true, ts.getHostSignatureFromJSDoc(name));
                    if (!symbol && ts.isIdentifier(name) && container) {
                        symbol = getMergedSymbol(getSymbol(getExportsOfSymbol(container), name.escapedText, meaning));
                    }
                    if (symbol) {
                        return symbol;
                    }
                }
                var left = ts.isIdentifier(name) ? container : resolveJSDocMemberName(name.left);
                var right = ts.isIdentifier(name) ? name.escapedText : name.right.escapedText;
                if (left) {
                    var proto = left.flags & 111551 /* Value */ && getPropertyOfType(getTypeOfSymbol(left), "prototype");
                    var t = proto ? getTypeOfSymbol(proto) : getDeclaredTypeOfSymbol(left);
                    return getPropertyOfType(t, right);
                }
            }
            function getSymbolAtLocation(node, ignoreErrors) {
                if (node.kind === 300 /* SourceFile */) {
                    return ts.isExternalModule(node) ? getMergedSymbol(node.symbol) : undefined;
                }
                var parent = node.parent;
                var grandParent = parent.parent;
                if (node.flags & 16777216 /* InWithStatement */) {
                    // We cannot answer semantic questions within a with block, do not proceed any further
                    return undefined;
                }
                if (isDeclarationNameOrImportPropertyName(node)) {
                    // This is a declaration, call getSymbolOfNode
                    var parentSymbol = getSymbolOfNode(parent);
                    return ts.isImportOrExportSpecifier(node.parent) && node.parent.propertyName === node
                        ? getImmediateAliasedSymbol(parentSymbol)
                        : parentSymbol;
                }
                else if (ts.isLiteralComputedPropertyDeclarationName(node)) {
                    return getSymbolOfNode(parent.parent);
                }
                if (node.kind === 79 /* Identifier */) {
                    if (isInRightSideOfImportOrExportAssignment(node)) {
                        return getSymbolOfNameOrPropertyAccessExpression(node);
                    }
                    else if (parent.kind === 201 /* BindingElement */ &&
                        grandParent.kind === 199 /* ObjectBindingPattern */ &&
                        node === parent.propertyName) {
                        var typeOfPattern = getTypeOfNode(grandParent);
                        var propertyDeclaration = getPropertyOfType(typeOfPattern, node.escapedText);
                        if (propertyDeclaration) {
                            return propertyDeclaration;
                        }
                    }
                    else if (ts.isMetaProperty(parent)) {
                        var parentType = getTypeOfNode(parent);
                        var propertyDeclaration = getPropertyOfType(parentType, node.escapedText);
                        if (propertyDeclaration) {
                            return propertyDeclaration;
                        }
                        if (parent.keywordToken === 103 /* NewKeyword */) {
                            return checkNewTargetMetaProperty(parent).symbol;
                        }
                    }
                }
                switch (node.kind) {
                    case 79 /* Identifier */:
                    case 80 /* PrivateIdentifier */:
                    case 204 /* PropertyAccessExpression */:
                    case 159 /* QualifiedName */:
                        return getSymbolOfNameOrPropertyAccessExpression(node);
                    case 108 /* ThisKeyword */:
                        var container = ts.getThisContainer(node, /*includeArrowFunctions*/ false);
                        if (ts.isFunctionLike(container)) {
                            var sig = getSignatureFromDeclaration(container);
                            if (sig.thisParameter) {
                                return sig.thisParameter;
                            }
                        }
                        if (ts.isInExpressionContext(node)) {
                            return checkExpression(node).symbol;
                        }
                    // falls through
                    case 190 /* ThisType */:
                        return getTypeFromThisTypeNode(node).symbol;
                    case 106 /* SuperKeyword */:
                        return checkExpression(node).symbol;
                    case 133 /* ConstructorKeyword */:
                        // constructor keyword for an overload, should take us to the definition if it exist
                        var constructorDeclaration = node.parent;
                        if (constructorDeclaration && constructorDeclaration.kind === 169 /* Constructor */) {
                            return constructorDeclaration.parent.symbol;
                        }
                        return undefined;
                    case 10 /* StringLiteral */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                        // 1). import x = require("./mo/*gotToDefinitionHere*/d")
                        // 2). External module name in an import declaration
                        // 3). Dynamic import call or require in javascript
                        // 4). type A = import("./f/*gotToDefinitionHere*/oo")
                        if ((ts.isExternalModuleImportEqualsDeclaration(node.parent.parent) && ts.getExternalModuleImportEqualsDeclarationExpression(node.parent.parent) === node) ||
                            ((node.parent.kind === 264 /* ImportDeclaration */ || node.parent.kind === 270 /* ExportDeclaration */) && node.parent.moduleSpecifier === node) ||
                            ((ts.isInJSFile(node) && ts.isRequireCall(node.parent, /*checkArgumentIsStringLiteralLike*/ false)) || ts.isImportCall(node.parent)) ||
                            (ts.isLiteralTypeNode(node.parent) && ts.isLiteralImportTypeNode(node.parent.parent) && node.parent.parent.argument === node.parent)) {
                            return resolveExternalModuleName(node, node, ignoreErrors);
                        }
                        if (ts.isCallExpression(parent) && ts.isBindableObjectDefinePropertyCall(parent) && parent.arguments[1] === node) {
                            return getSymbolOfNode(parent);
                        }
                    // falls through
                    case 8 /* NumericLiteral */:
                        // index access
                        var objectType = ts.isElementAccessExpression(parent)
                            ? parent.argumentExpression === node ? getTypeOfExpression(parent.expression) : undefined
                            : ts.isLiteralTypeNode(parent) && ts.isIndexedAccessTypeNode(grandParent)
                                ? getTypeFromTypeNode(grandParent.objectType)
                                : undefined;
                        return objectType && getPropertyOfType(objectType, ts.escapeLeadingUnderscores(node.text));
                    case 88 /* DefaultKeyword */:
                    case 98 /* FunctionKeyword */:
                    case 38 /* EqualsGreaterThanToken */:
                    case 84 /* ClassKeyword */:
                        return getSymbolOfNode(node.parent);
                    case 198 /* ImportType */:
                        return ts.isLiteralImportTypeNode(node) ? getSymbolAtLocation(node.argument.literal, ignoreErrors) : undefined;
                    case 93 /* ExportKeyword */:
                        return ts.isExportAssignment(node.parent) ? ts.Debug.checkDefined(node.parent.symbol) : undefined;
                    case 100 /* ImportKeyword */:
                    case 103 /* NewKeyword */:
                        return ts.isMetaProperty(node.parent) ? checkMetaPropertyKeyword(node.parent).symbol : undefined;
                    case 229 /* MetaProperty */:
                        return checkExpression(node).symbol;
                    default:
                        return undefined;
                }
            }
            function getIndexInfosAtLocation(node) {
                if (ts.isIdentifier(node) && ts.isPropertyAccessExpression(node.parent) && node.parent.name === node) {
                    var keyType_1 = getLiteralTypeFromPropertyName(node);
                    var objectType = getTypeOfExpression(node.parent.expression);
                    var objectTypes = objectType.flags & 1048576 /* Union */ ? objectType.types : [objectType];
                    return ts.flatMap(objectTypes, function (t) { return ts.filter(getIndexInfosOfType(t), function (info) { return isApplicableIndexType(keyType_1, info.keyType); }); });
                }
                return undefined;
            }
            function getShorthandAssignmentValueSymbol(location) {
                if (location && location.kind === 292 /* ShorthandPropertyAssignment */) {
                    return resolveEntityName(location.name, 111551 /* Value */ | 2097152 /* Alias */);
                }
                return undefined;
            }
            /** Returns the target of an export specifier without following aliases */
            function getExportSpecifierLocalTargetSymbol(node) {
                if (ts.isExportSpecifier(node)) {
                    return node.parent.parent.moduleSpecifier ?
                        getExternalModuleMember(node.parent.parent, node) :
                        resolveEntityName(node.propertyName || node.name, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */ | 2097152 /* Alias */);
                }
                else {
                    return resolveEntityName(node, 111551 /* Value */ | 788968 /* Type */ | 1920 /* Namespace */ | 2097152 /* Alias */);
                }
            }
            function getTypeOfNode(node) {
                if (ts.isSourceFile(node) && !ts.isExternalModule(node)) {
                    return errorType;
                }
                if (node.flags & 16777216 /* InWithStatement */) {
                    // We cannot answer semantic questions within a with block, do not proceed any further
                    return errorType;
                }
                var classDecl = ts.tryGetClassImplementingOrExtendingExpressionWithTypeArguments(node);
                var classType = classDecl && getDeclaredTypeOfClassOrInterface(getSymbolOfNode(classDecl.class));
                if (ts.isPartOfTypeNode(node)) {
                    var typeFromTypeNode = getTypeFromTypeNode(node);
                    return classType ? getTypeWithThisArgument(typeFromTypeNode, classType.thisType) : typeFromTypeNode;
                }
                if (ts.isExpressionNode(node)) {
                    return getRegularTypeOfExpression(node);
                }
                if (classType && !classDecl.isImplements) {
                    // A SyntaxKind.ExpressionWithTypeArguments is considered a type node, except when it occurs in the
                    // extends clause of a class. We handle that case here.
                    var baseType = ts.firstOrUndefined(getBaseTypes(classType));
                    return baseType ? getTypeWithThisArgument(baseType, classType.thisType) : errorType;
                }
                if (isTypeDeclaration(node)) {
                    // In this case, we call getSymbolOfNode instead of getSymbolAtLocation because it is a declaration
                    var symbol = getSymbolOfNode(node);
                    return getDeclaredTypeOfSymbol(symbol);
                }
                if (isTypeDeclarationName(node)) {
                    var symbol = getSymbolAtLocation(node);
                    return symbol ? getDeclaredTypeOfSymbol(symbol) : errorType;
                }
                if (ts.isDeclaration(node)) {
                    // In this case, we call getSymbolOfNode instead of getSymbolAtLocation because it is a declaration
                    var symbol = getSymbolOfNode(node);
                    return getTypeOfSymbol(symbol);
                }
                if (isDeclarationNameOrImportPropertyName(node)) {
                    var symbol = getSymbolAtLocation(node);
                    if (symbol) {
                        return getTypeOfSymbol(symbol);
                    }
                    return errorType;
                }
                if (ts.isBindingPattern(node)) {
                    return getTypeForVariableLikeDeclaration(node.parent, /*includeOptionality*/ true) || errorType;
                }
                if (isInRightSideOfImportOrExportAssignment(node)) {
                    var symbol = getSymbolAtLocation(node);
                    if (symbol) {
                        var declaredType = getDeclaredTypeOfSymbol(symbol);
                        return declaredType !== errorType ? declaredType : getTypeOfSymbol(symbol);
                    }
                }
                if (ts.isMetaProperty(node.parent) && node.parent.keywordToken === node.kind) {
                    return checkMetaPropertyKeyword(node.parent);
                }
                return errorType;
            }
            // Gets the type of object literal or array literal of destructuring assignment.
            // { a } from
            //     for ( { a } of elems) {
            //     }
            // [ a ] from
            //     [a] = [ some array ...]
            function getTypeOfAssignmentPattern(expr) {
                ts.Debug.assert(expr.kind === 203 /* ObjectLiteralExpression */ || expr.kind === 202 /* ArrayLiteralExpression */);
                // If this is from "for of"
                //     for ( { a } of elems) {
                //     }
                if (expr.parent.kind === 242 /* ForOfStatement */) {
                    var iteratedType = checkRightHandSideOfForOf(expr.parent);
                    return checkDestructuringAssignment(expr, iteratedType || errorType);
                }
                // If this is from "for" initializer
                //     for ({a } = elems[0];.....) { }
                if (expr.parent.kind === 219 /* BinaryExpression */) {
                    var iteratedType = getTypeOfExpression(expr.parent.right);
                    return checkDestructuringAssignment(expr, iteratedType || errorType);
                }
                // If this is from nested object binding pattern
                //     for ({ skills: { primary, secondary } } = multiRobot, i = 0; i < 1; i++) {
                if (expr.parent.kind === 291 /* PropertyAssignment */) {
                    var node_3 = ts.cast(expr.parent.parent, ts.isObjectLiteralExpression);
                    var typeOfParentObjectLiteral = getTypeOfAssignmentPattern(node_3) || errorType;
                    var propertyIndex = ts.indexOfNode(node_3.properties, expr.parent);
                    return checkObjectLiteralDestructuringPropertyAssignment(node_3, typeOfParentObjectLiteral, propertyIndex);
                }
                // Array literal assignment - array destructuring pattern
                var node = ts.cast(expr.parent, ts.isArrayLiteralExpression);
                //    [{ property1: p1, property2 }] = elems;
                var typeOfArrayLiteral = getTypeOfAssignmentPattern(node) || errorType;
                var elementType = checkIteratedTypeOrElementType(65 /* Destructuring */, typeOfArrayLiteral, undefinedType, expr.parent) || errorType;
                return checkArrayLiteralDestructuringElementAssignment(node, typeOfArrayLiteral, node.elements.indexOf(expr), elementType);
            }
            // Gets the property symbol corresponding to the property in destructuring assignment
            // 'property1' from
            //     for ( { property1: a } of elems) {
            //     }
            // 'property1' at location 'a' from:
            //     [a] = [ property1, property2 ]
            function getPropertySymbolOfDestructuringAssignment(location) {
                // Get the type of the object or array literal and then look for property of given name in the type
                var typeOfObjectLiteral = getTypeOfAssignmentPattern(ts.cast(location.parent.parent, ts.isAssignmentPattern));
                return typeOfObjectLiteral && getPropertyOfType(typeOfObjectLiteral, location.escapedText);
            }
            function getRegularTypeOfExpression(expr) {
                if (ts.isRightSideOfQualifiedNameOrPropertyAccess(expr)) {
                    expr = expr.parent;
                }
                return getRegularTypeOfLiteralType(getTypeOfExpression(expr));
            }
            /**
             * Gets either the static or instance type of a class element, based on
             * whether the element is declared as "static".
             */
            function getParentTypeOfClassElement(node) {
                var classSymbol = getSymbolOfNode(node.parent);
                return ts.isStatic(node)
                    ? getTypeOfSymbol(classSymbol)
                    : getDeclaredTypeOfSymbol(classSymbol);
            }
            function getClassElementPropertyKeyType(element) {
                var name = element.name;
                switch (name.kind) {
                    case 79 /* Identifier */:
                        return getStringLiteralType(ts.idText(name));
                    case 8 /* NumericLiteral */:
                    case 10 /* StringLiteral */:
                        return getStringLiteralType(name.text);
                    case 160 /* ComputedPropertyName */:
                        var nameType = checkComputedPropertyName(name);
                        return isTypeAssignableToKind(nameType, 12288 /* ESSymbolLike */) ? nameType : stringType;
                    default:
                        return ts.Debug.fail("Unsupported property name.");
                }
            }
            // Return the list of properties of the given type, augmented with properties from Function
            // if the type has call or construct signatures
            function getAugmentedPropertiesOfType(type) {
                type = getApparentType(type);
                var propsByName = ts.createSymbolTable(getPropertiesOfType(type));
                var functionType = getSignaturesOfType(type, 0 /* Call */).length ? globalCallableFunctionType :
                    getSignaturesOfType(type, 1 /* Construct */).length ? globalNewableFunctionType :
                        undefined;
                if (functionType) {
                    ts.forEach(getPropertiesOfType(functionType), function (p) {
                        if (!propsByName.has(p.escapedName)) {
                            propsByName.set(p.escapedName, p);
                        }
                    });
                }
                return getNamedMembers(propsByName);
            }
            function typeHasCallOrConstructSignatures(type) {
                return ts.typeHasCallOrConstructSignatures(type, checker);
            }
            function getRootSymbols(symbol) {
                var roots = getImmediateRootSymbols(symbol);
                return roots ? ts.flatMap(roots, getRootSymbols) : [symbol];
            }
            function getImmediateRootSymbols(symbol) {
                if (ts.getCheckFlags(symbol) & 6 /* Synthetic */) {
                    return ts.mapDefined(getSymbolLinks(symbol).containingType.types, function (type) { return getPropertyOfType(type, symbol.escapedName); });
                }
                else if (symbol.flags & 33554432 /* Transient */) {
                    var _a = symbol, leftSpread = _a.leftSpread, rightSpread = _a.rightSpread, syntheticOrigin = _a.syntheticOrigin;
                    return leftSpread ? [leftSpread, rightSpread]
                        : syntheticOrigin ? [syntheticOrigin]
                            : ts.singleElementArray(tryGetAliasTarget(symbol));
                }
                return undefined;
            }
            function tryGetAliasTarget(symbol) {
                var target;
                var next = symbol;
                while (next = getSymbolLinks(next).target) {
                    target = next;
                }
                return target;
            }
            // Emitter support
            function isArgumentsLocalBinding(nodeIn) {
                // Note: does not handle isShorthandPropertyAssignment (and probably a few more)
                if (ts.isGeneratedIdentifier(nodeIn))
                    return false;
                var node = ts.getParseTreeNode(nodeIn, ts.isIdentifier);
                if (!node)
                    return false;
                var parent = node.parent;
                if (!parent)
                    return false;
                var isPropertyName = ((ts.isPropertyAccessExpression(parent)
                    || ts.isPropertyAssignment(parent))
                    && parent.name === node);
                return !isPropertyName && getReferencedValueSymbol(node) === argumentsSymbol;
            }
            function moduleExportsSomeValue(moduleReferenceExpression) {
                var moduleSymbol = resolveExternalModuleName(moduleReferenceExpression.parent, moduleReferenceExpression);
                if (!moduleSymbol || ts.isShorthandAmbientModuleSymbol(moduleSymbol)) {
                    // If the module is not found or is shorthand, assume that it may export a value.
                    return true;
                }
                var hasExportAssignment = hasExportAssignmentSymbol(moduleSymbol);
                // if module has export assignment then 'resolveExternalModuleSymbol' will return resolved symbol for export assignment
                // otherwise it will return moduleSymbol itself
                moduleSymbol = resolveExternalModuleSymbol(moduleSymbol);
                var symbolLinks = getSymbolLinks(moduleSymbol);
                if (symbolLinks.exportsSomeValue === undefined) {
                    // for export assignments - check if resolved symbol for RHS is itself a value
                    // otherwise - check if at least one export is value
                    symbolLinks.exportsSomeValue = hasExportAssignment
                        ? !!(moduleSymbol.flags & 111551 /* Value */)
                        : ts.forEachEntry(getExportsOfModule(moduleSymbol), isValue);
                }
                return symbolLinks.exportsSomeValue;
                function isValue(s) {
                    s = resolveSymbol(s);
                    return s && !!(s.flags & 111551 /* Value */);
                }
            }
            function isNameOfModuleOrEnumDeclaration(node) {
                return ts.isModuleOrEnumDeclaration(node.parent) && node === node.parent.name;
            }
            // When resolved as an expression identifier, if the given node references an exported entity, return the declaration
            // node of the exported entity's container. Otherwise, return undefined.
            function getReferencedExportContainer(nodeIn, prefixLocals) {
                var _a;
                var node = ts.getParseTreeNode(nodeIn, ts.isIdentifier);
                if (node) {
                    // When resolving the export container for the name of a module or enum
                    // declaration, we need to start resolution at the declaration's container.
                    // Otherwise, we could incorrectly resolve the export container as the
                    // declaration if it contains an exported member with the same name.
                    var symbol = getReferencedValueSymbol(node, /*startInDeclarationContainer*/ isNameOfModuleOrEnumDeclaration(node));
                    if (symbol) {
                        if (symbol.flags & 1048576 /* ExportValue */) {
                            // If we reference an exported entity within the same module declaration, then whether
                            // we prefix depends on the kind of entity. SymbolFlags.ExportHasLocal encompasses all the
                            // kinds that we do NOT prefix.
                            var exportSymbol = getMergedSymbol(symbol.exportSymbol);
                            if (!prefixLocals && exportSymbol.flags & 944 /* ExportHasLocal */ && !(exportSymbol.flags & 3 /* Variable */)) {
                                return undefined;
                            }
                            symbol = exportSymbol;
                        }
                        var parentSymbol_1 = getParentOfSymbol(symbol);
                        if (parentSymbol_1) {
                            if (parentSymbol_1.flags & 512 /* ValueModule */ && ((_a = parentSymbol_1.valueDeclaration) === null || _a === void 0 ? void 0 : _a.kind) === 300 /* SourceFile */) {
                                var symbolFile = parentSymbol_1.valueDeclaration;
                                var referenceFile = ts.getSourceFileOfNode(node);
                                // If `node` accesses an export and that export isn't in the same file, then symbol is a namespace export, so return undefined.
                                var symbolIsUmdExport = symbolFile !== referenceFile;
                                return symbolIsUmdExport ? undefined : symbolFile;
                            }
                            return ts.findAncestor(node.parent, function (n) { return ts.isModuleOrEnumDeclaration(n) && getSymbolOfNode(n) === parentSymbol_1; });
                        }
                    }
                }
            }
            // When resolved as an expression identifier, if the given node references an import, return the declaration of
            // that import. Otherwise, return undefined.
            function getReferencedImportDeclaration(nodeIn) {
                if (nodeIn.generatedImportReference) {
                    return nodeIn.generatedImportReference;
                }
                var node = ts.getParseTreeNode(nodeIn, ts.isIdentifier);
                if (node) {
                    var symbol = getReferencedValueSymbol(node);
                    // We should only get the declaration of an alias if there isn't a local value
                    // declaration for the symbol
                    if (isNonLocalAlias(symbol, /*excludes*/ 111551 /* Value */) && !getTypeOnlyAliasDeclaration(symbol)) {
                        return getDeclarationOfAliasSymbol(symbol);
                    }
                }
                return undefined;
            }
            function isSymbolOfDestructuredElementOfCatchBinding(symbol) {
                return symbol.valueDeclaration
                    && ts.isBindingElement(symbol.valueDeclaration)
                    && ts.walkUpBindingElementsAndPatterns(symbol.valueDeclaration).parent.kind === 290 /* CatchClause */;
            }
            function isSymbolOfDeclarationWithCollidingName(symbol) {
                if (symbol.flags & 418 /* BlockScoped */ && symbol.valueDeclaration && !ts.isSourceFile(symbol.valueDeclaration)) {
                    var links = getSymbolLinks(symbol);
                    if (links.isDeclarationWithCollidingName === undefined) {
                        var container = ts.getEnclosingBlockScopeContainer(symbol.valueDeclaration);
                        if (ts.isStatementWithLocals(container) || isSymbolOfDestructuredElementOfCatchBinding(symbol)) {
                            var nodeLinks_1 = getNodeLinks(symbol.valueDeclaration);
                            if (resolveName(container.parent, symbol.escapedName, 111551 /* Value */, /*nameNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ false)) {
                                // redeclaration - always should be renamed
                                links.isDeclarationWithCollidingName = true;
                            }
                            else if (nodeLinks_1.flags & 262144 /* CapturedBlockScopedBinding */) {
                                // binding is captured in the function
                                // should be renamed if:
                                // - binding is not top level - top level bindings never collide with anything
                                // AND
                                //   - binding is not declared in loop, should be renamed to avoid name reuse across siblings
                                //     let a, b
                                //     { let x = 1; a = () => x; }
                                //     { let x = 100; b = () => x; }
                                //     console.log(a()); // should print '1'
                                //     console.log(b()); // should print '100'
                                //     OR
                                //   - binding is declared inside loop but not in inside initializer of iteration statement or directly inside loop body
                                //     * variables from initializer are passed to rewritten loop body as parameters so they are not captured directly
                                //     * variables that are declared immediately in loop body will become top level variable after loop is rewritten and thus
                                //       they will not collide with anything
                                var isDeclaredInLoop = nodeLinks_1.flags & 524288 /* BlockScopedBindingInLoop */;
                                var inLoopInitializer = ts.isIterationStatement(container, /*lookInLabeledStatements*/ false);
                                var inLoopBodyBlock = container.kind === 233 /* Block */ && ts.isIterationStatement(container.parent, /*lookInLabeledStatements*/ false);
                                links.isDeclarationWithCollidingName = !ts.isBlockScopedContainerTopLevel(container) && (!isDeclaredInLoop || (!inLoopInitializer && !inLoopBodyBlock));
                            }
                            else {
                                links.isDeclarationWithCollidingName = false;
                            }
                        }
                    }
                    return links.isDeclarationWithCollidingName;
                }
                return false;
            }
            // When resolved as an expression identifier, if the given node references a nested block scoped entity with
            // a name that either hides an existing name or might hide it when compiled downlevel,
            // return the declaration of that entity. Otherwise, return undefined.
            function getReferencedDeclarationWithCollidingName(nodeIn) {
                if (!ts.isGeneratedIdentifier(nodeIn)) {
                    var node = ts.getParseTreeNode(nodeIn, ts.isIdentifier);
                    if (node) {
                        var symbol = getReferencedValueSymbol(node);
                        if (symbol && isSymbolOfDeclarationWithCollidingName(symbol)) {
                            return symbol.valueDeclaration;
                        }
                    }
                }
                return undefined;
            }
            // Return true if the given node is a declaration of a nested block scoped entity with a name that either hides an
            // existing name or might hide a name when compiled downlevel
            function isDeclarationWithCollidingName(nodeIn) {
                var node = ts.getParseTreeNode(nodeIn, ts.isDeclaration);
                if (node) {
                    var symbol = getSymbolOfNode(node);
                    if (symbol) {
                        return isSymbolOfDeclarationWithCollidingName(symbol);
                    }
                }
                return false;
            }
            function isValueAliasDeclaration(node) {
                switch (node.kind) {
                    case 263 /* ImportEqualsDeclaration */:
                        return isAliasResolvedToValue(getSymbolOfNode(node) || unknownSymbol);
                    case 265 /* ImportClause */:
                    case 266 /* NamespaceImport */:
                    case 268 /* ImportSpecifier */:
                    case 273 /* ExportSpecifier */:
                        var symbol = getSymbolOfNode(node) || unknownSymbol;
                        return isAliasResolvedToValue(symbol) && !getTypeOnlyAliasDeclaration(symbol);
                    case 270 /* ExportDeclaration */:
                        var exportClause = node.exportClause;
                        return !!exportClause && (ts.isNamespaceExport(exportClause) ||
                            ts.some(exportClause.elements, isValueAliasDeclaration));
                    case 269 /* ExportAssignment */:
                        return node.expression && node.expression.kind === 79 /* Identifier */ ?
                            isAliasResolvedToValue(getSymbolOfNode(node) || unknownSymbol) :
                            true;
                }
                return false;
            }
            function isTopLevelValueImportEqualsWithEntityName(nodeIn) {
                var node = ts.getParseTreeNode(nodeIn, ts.isImportEqualsDeclaration);
                if (node === undefined || node.parent.kind !== 300 /* SourceFile */ || !ts.isInternalModuleImportEqualsDeclaration(node)) {
                    // parent is not source file or it is not reference to internal module
                    return false;
                }
                var isValue = isAliasResolvedToValue(getSymbolOfNode(node));
                return isValue && node.moduleReference && !ts.nodeIsMissing(node.moduleReference);
            }
            function isAliasResolvedToValue(symbol) {
                var target = resolveAlias(symbol);
                if (target === unknownSymbol) {
                    return true;
                }
                // const enums and modules that contain only const enums are not considered values from the emit perspective
                // unless 'preserveConstEnums' option is set to true
                return !!(target.flags & 111551 /* Value */) &&
                    (ts.shouldPreserveConstEnums(compilerOptions) || !isConstEnumOrConstEnumOnlyModule(target));
            }
            function isConstEnumOrConstEnumOnlyModule(s) {
                return isConstEnumSymbol(s) || !!s.constEnumOnlyModule;
            }
            function isReferencedAliasDeclaration(node, checkChildren) {
                if (isAliasSymbolDeclaration(node)) {
                    var symbol = getSymbolOfNode(node);
                    var links = symbol && getSymbolLinks(symbol);
                    if (links === null || links === void 0 ? void 0 : links.referenced) {
                        return true;
                    }
                    var target = getSymbolLinks(symbol).target; // TODO: GH#18217
                    if (target && ts.getEffectiveModifierFlags(node) & 1 /* Export */ &&
                        target.flags & 111551 /* Value */ &&
                        (ts.shouldPreserveConstEnums(compilerOptions) || !isConstEnumOrConstEnumOnlyModule(target))) {
                        // An `export import ... =` of a value symbol is always considered referenced
                        return true;
                    }
                }
                if (checkChildren) {
                    return !!ts.forEachChild(node, function (node) { return isReferencedAliasDeclaration(node, checkChildren); });
                }
                return false;
            }
            function isImplementationOfOverload(node) {
                if (ts.nodeIsPresent(node.body)) {
                    if (ts.isGetAccessor(node) || ts.isSetAccessor(node))
                        return false; // Get or set accessors can never be overload implementations, but can have up to 2 signatures
                    var symbol = getSymbolOfNode(node);
                    var signaturesOfSymbol = getSignaturesOfSymbol(symbol);
                    // If this function body corresponds to function with multiple signature, it is implementation of overload
                    // e.g.: function foo(a: string): string;
                    //       function foo(a: number): number;
                    //       function foo(a: any) { // This is implementation of the overloads
                    //           return a;
                    //       }
                    return signaturesOfSymbol.length > 1 ||
                        // If there is single signature for the symbol, it is overload if that signature isn't coming from the node
                        // e.g.: function foo(a: string): string;
                        //       function foo(a: any) { // This is implementation of the overloads
                        //           return a;
                        //       }
                        (signaturesOfSymbol.length === 1 && signaturesOfSymbol[0].declaration !== node);
                }
                return false;
            }
            function isRequiredInitializedParameter(parameter) {
                return !!strictNullChecks &&
                    !isOptionalParameter(parameter) &&
                    !ts.isJSDocParameterTag(parameter) &&
                    !!parameter.initializer &&
                    !ts.hasSyntacticModifier(parameter, 16476 /* ParameterPropertyModifier */);
            }
            function isOptionalUninitializedParameterProperty(parameter) {
                return strictNullChecks &&
                    isOptionalParameter(parameter) &&
                    !parameter.initializer &&
                    ts.hasSyntacticModifier(parameter, 16476 /* ParameterPropertyModifier */);
            }
            function isOptionalUninitializedParameter(parameter) {
                return !!strictNullChecks &&
                    isOptionalParameter(parameter) &&
                    !parameter.initializer;
            }
            function isExpandoFunctionDeclaration(node) {
                var declaration = ts.getParseTreeNode(node, ts.isFunctionDeclaration);
                if (!declaration) {
                    return false;
                }
                var symbol = getSymbolOfNode(declaration);
                if (!symbol || !(symbol.flags & 16 /* Function */)) {
                    return false;
                }
                return !!ts.forEachEntry(getExportsOfSymbol(symbol), function (p) { return p.flags & 111551 /* Value */ && p.valueDeclaration && ts.isPropertyAccessExpression(p.valueDeclaration); });
            }
            function getPropertiesOfContainerFunction(node) {
                var declaration = ts.getParseTreeNode(node, ts.isFunctionDeclaration);
                if (!declaration) {
                    return ts.emptyArray;
                }
                var symbol = getSymbolOfNode(declaration);
                return symbol && getPropertiesOfType(getTypeOfSymbol(symbol)) || ts.emptyArray;
            }
            function getNodeCheckFlags(node) {
                var _a;
                var nodeId = node.id || 0;
                if (nodeId < 0 || nodeId >= nodeLinks.length)
                    return 0;
                return ((_a = nodeLinks[nodeId]) === null || _a === void 0 ? void 0 : _a.flags) || 0;
            }
            function getEnumMemberValue(node) {
                computeEnumMemberValues(node.parent);
                return getNodeLinks(node).enumMemberValue;
            }
            function canHaveConstantValue(node) {
                switch (node.kind) {
                    case 294 /* EnumMember */:
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */:
                        return true;
                }
                return false;
            }
            function getConstantValue(node) {
                if (node.kind === 294 /* EnumMember */) {
                    return getEnumMemberValue(node);
                }
                var symbol = getNodeLinks(node).resolvedSymbol;
                if (symbol && (symbol.flags & 8 /* EnumMember */)) {
                    // inline property\index accesses only for const enums
                    var member = symbol.valueDeclaration;
                    if (ts.isEnumConst(member.parent)) {
                        return getEnumMemberValue(member);
                    }
                }
                return undefined;
            }
            function isFunctionType(type) {
                return !!(type.flags & 524288 /* Object */) && getSignaturesOfType(type, 0 /* Call */).length > 0;
            }
            function getTypeReferenceSerializationKind(typeNameIn, location) {
                var _a, _b;
                // ensure both `typeName` and `location` are parse tree nodes.
                var typeName = ts.getParseTreeNode(typeNameIn, ts.isEntityName);
                if (!typeName)
                    return ts.TypeReferenceSerializationKind.Unknown;
                if (location) {
                    location = ts.getParseTreeNode(location);
                    if (!location)
                        return ts.TypeReferenceSerializationKind.Unknown;
                }
                // Resolve the symbol as a value to ensure the type can be reached at runtime during emit.
                var isTypeOnly = false;
                if (ts.isQualifiedName(typeName)) {
                    var rootValueSymbol = resolveEntityName(ts.getFirstIdentifier(typeName), 111551 /* Value */, /*ignoreErrors*/ true, /*dontResolveAlias*/ true, location);
                    isTypeOnly = !!((_a = rootValueSymbol === null || rootValueSymbol === void 0 ? void 0 : rootValueSymbol.declarations) === null || _a === void 0 ? void 0 : _a.every(ts.isTypeOnlyImportOrExportDeclaration));
                }
                var valueSymbol = resolveEntityName(typeName, 111551 /* Value */, /*ignoreErrors*/ true, /*dontResolveAlias*/ true, location);
                var resolvedSymbol = valueSymbol && valueSymbol.flags & 2097152 /* Alias */ ? resolveAlias(valueSymbol) : valueSymbol;
                isTypeOnly || (isTypeOnly = !!((_b = valueSymbol === null || valueSymbol === void 0 ? void 0 : valueSymbol.declarations) === null || _b === void 0 ? void 0 : _b.every(ts.isTypeOnlyImportOrExportDeclaration)));
                // Resolve the symbol as a type so that we can provide a more useful hint for the type serializer.
                var typeSymbol = resolveEntityName(typeName, 788968 /* Type */, /*ignoreErrors*/ true, /*dontResolveAlias*/ false, location);
                if (resolvedSymbol && resolvedSymbol === typeSymbol) {
                    var globalPromiseSymbol = getGlobalPromiseConstructorSymbol(/*reportErrors*/ false);
                    if (globalPromiseSymbol && resolvedSymbol === globalPromiseSymbol) {
                        return ts.TypeReferenceSerializationKind.Promise;
                    }
                    var constructorType = getTypeOfSymbol(resolvedSymbol);
                    if (constructorType && isConstructorType(constructorType)) {
                        return isTypeOnly ? ts.TypeReferenceSerializationKind.TypeWithCallSignature : ts.TypeReferenceSerializationKind.TypeWithConstructSignatureAndValue;
                    }
                }
                // We might not be able to resolve type symbol so use unknown type in that case (eg error case)
                if (!typeSymbol) {
                    return isTypeOnly ? ts.TypeReferenceSerializationKind.ObjectType : ts.TypeReferenceSerializationKind.Unknown;
                }
                var type = getDeclaredTypeOfSymbol(typeSymbol);
                if (type === errorType) {
                    return isTypeOnly ? ts.TypeReferenceSerializationKind.ObjectType : ts.TypeReferenceSerializationKind.Unknown;
                }
                else if (type.flags & 3 /* AnyOrUnknown */) {
                    return ts.TypeReferenceSerializationKind.ObjectType;
                }
                else if (isTypeAssignableToKind(type, 16384 /* Void */ | 98304 /* Nullable */ | 131072 /* Never */)) {
                    return ts.TypeReferenceSerializationKind.VoidNullableOrNeverType;
                }
                else if (isTypeAssignableToKind(type, 528 /* BooleanLike */)) {
                    return ts.TypeReferenceSerializationKind.BooleanType;
                }
                else if (isTypeAssignableToKind(type, 296 /* NumberLike */)) {
                    return ts.TypeReferenceSerializationKind.NumberLikeType;
                }
                else if (isTypeAssignableToKind(type, 2112 /* BigIntLike */)) {
                    return ts.TypeReferenceSerializationKind.BigIntLikeType;
                }
                else if (isTypeAssignableToKind(type, 402653316 /* StringLike */)) {
                    return ts.TypeReferenceSerializationKind.StringLikeType;
                }
                else if (isTupleType(type)) {
                    return ts.TypeReferenceSerializationKind.ArrayLikeType;
                }
                else if (isTypeAssignableToKind(type, 12288 /* ESSymbolLike */)) {
                    return ts.TypeReferenceSerializationKind.ESSymbolType;
                }
                else if (isFunctionType(type)) {
                    return ts.TypeReferenceSerializationKind.TypeWithCallSignature;
                }
                else if (isArrayType(type)) {
                    return ts.TypeReferenceSerializationKind.ArrayLikeType;
                }
                else {
                    return ts.TypeReferenceSerializationKind.ObjectType;
                }
            }
            function createTypeOfDeclaration(declarationIn, enclosingDeclaration, flags, tracker, addUndefined) {
                var declaration = ts.getParseTreeNode(declarationIn, ts.isVariableLikeOrAccessor);
                if (!declaration) {
                    return ts.factory.createToken(129 /* AnyKeyword */);
                }
                // Get type of the symbol if this is the valid symbol otherwise get type at location
                var symbol = getSymbolOfNode(declaration);
                var type = symbol && !(symbol.flags & (2048 /* TypeLiteral */ | 131072 /* Signature */))
                    ? getWidenedLiteralType(getTypeOfSymbol(symbol))
                    : errorType;
                if (type.flags & 8192 /* UniqueESSymbol */ &&
                    type.symbol === symbol) {
                    flags |= 1048576 /* AllowUniqueESSymbolType */;
                }
                if (addUndefined) {
                    type = getOptionalType(type);
                }
                return nodeBuilder.typeToTypeNode(type, enclosingDeclaration, flags | 1024 /* MultilineObjectLiterals */, tracker);
            }
            function createReturnTypeOfSignatureDeclaration(signatureDeclarationIn, enclosingDeclaration, flags, tracker) {
                var signatureDeclaration = ts.getParseTreeNode(signatureDeclarationIn, ts.isFunctionLike);
                if (!signatureDeclaration) {
                    return ts.factory.createToken(129 /* AnyKeyword */);
                }
                var signature = getSignatureFromDeclaration(signatureDeclaration);
                return nodeBuilder.typeToTypeNode(getReturnTypeOfSignature(signature), enclosingDeclaration, flags | 1024 /* MultilineObjectLiterals */, tracker);
            }
            function createTypeOfExpression(exprIn, enclosingDeclaration, flags, tracker) {
                var expr = ts.getParseTreeNode(exprIn, ts.isExpression);
                if (!expr) {
                    return ts.factory.createToken(129 /* AnyKeyword */);
                }
                var type = getWidenedType(getRegularTypeOfExpression(expr));
                return nodeBuilder.typeToTypeNode(type, enclosingDeclaration, flags | 1024 /* MultilineObjectLiterals */, tracker);
            }
            function hasGlobalName(name) {
                return globals.has(ts.escapeLeadingUnderscores(name));
            }
            function getReferencedValueSymbol(reference, startInDeclarationContainer) {
                var resolvedSymbol = getNodeLinks(reference).resolvedSymbol;
                if (resolvedSymbol) {
                    return resolvedSymbol;
                }
                var location = reference;
                if (startInDeclarationContainer) {
                    // When resolving the name of a declaration as a value, we need to start resolution
                    // at a point outside of the declaration.
                    var parent = reference.parent;
                    if (ts.isDeclaration(parent) && reference === parent.name) {
                        location = getDeclarationContainer(parent);
                    }
                }
                return resolveName(location, reference.escapedText, 111551 /* Value */ | 1048576 /* ExportValue */ | 2097152 /* Alias */, /*nodeNotFoundMessage*/ undefined, /*nameArg*/ undefined, /*isUse*/ true);
            }
            function getReferencedValueDeclaration(referenceIn) {
                if (!ts.isGeneratedIdentifier(referenceIn)) {
                    var reference = ts.getParseTreeNode(referenceIn, ts.isIdentifier);
                    if (reference) {
                        var symbol = getReferencedValueSymbol(reference);
                        if (symbol) {
                            return getExportSymbolOfValueSymbolIfExported(symbol).valueDeclaration;
                        }
                    }
                }
                return undefined;
            }
            function isLiteralConstDeclaration(node) {
                if (ts.isDeclarationReadonly(node) || ts.isVariableDeclaration(node) && ts.isVarConst(node)) {
                    return isFreshLiteralType(getTypeOfSymbol(getSymbolOfNode(node)));
                }
                return false;
            }
            function literalTypeToNode(type, enclosing, tracker) {
                var enumResult = type.flags & 1024 /* EnumLiteral */ ? nodeBuilder.symbolToExpression(type.symbol, 111551 /* Value */, enclosing, /*flags*/ undefined, tracker)
                    : type === trueType ? ts.factory.createTrue() : type === falseType && ts.factory.createFalse();
                if (enumResult)
                    return enumResult;
                var literalValue = type.value;
                return typeof literalValue === "object" ? ts.factory.createBigIntLiteral(literalValue) :
                    typeof literalValue === "number" ? ts.factory.createNumericLiteral(literalValue) :
                        ts.factory.createStringLiteral(literalValue);
            }
            function createLiteralConstValue(node, tracker) {
                var type = getTypeOfSymbol(getSymbolOfNode(node));
                return literalTypeToNode(type, node, tracker);
            }
            function getJsxFactoryEntity(location) {
                return location ? (getJsxNamespace(location), (ts.getSourceFileOfNode(location).localJsxFactory || _jsxFactoryEntity)) : _jsxFactoryEntity;
            }
            function getJsxFragmentFactoryEntity(location) {
                if (location) {
                    var file = ts.getSourceFileOfNode(location);
                    if (file) {
                        if (file.localJsxFragmentFactory) {
                            return file.localJsxFragmentFactory;
                        }
                        var jsxFragPragmas = file.pragmas.get("jsxfrag");
                        var jsxFragPragma = ts.isArray(jsxFragPragmas) ? jsxFragPragmas[0] : jsxFragPragmas;
                        if (jsxFragPragma) {
                            file.localJsxFragmentFactory = ts.parseIsolatedEntityName(jsxFragPragma.arguments.factory, languageVersion);
                            return file.localJsxFragmentFactory;
                        }
                    }
                }
                if (compilerOptions.jsxFragmentFactory) {
                    return ts.parseIsolatedEntityName(compilerOptions.jsxFragmentFactory, languageVersion);
                }
            }
            function createResolver() {
                // this variable and functions that use it are deliberately moved here from the outer scope
                // to avoid scope pollution
                var resolvedTypeReferenceDirectives = host.getResolvedTypeReferenceDirectives();
                var fileToDirective;
                if (resolvedTypeReferenceDirectives) {
                    // populate reverse mapping: file path -> type reference directive that was resolved to this file
                    fileToDirective = new ts.Map();
                    resolvedTypeReferenceDirectives.forEach(function (resolvedDirective, key) {
                        if (!resolvedDirective || !resolvedDirective.resolvedFileName) {
                            return;
                        }
                        var file = host.getSourceFile(resolvedDirective.resolvedFileName);
                        if (file) {
                            // Add the transitive closure of path references loaded by this file (as long as they are not)
                            // part of an existing type reference.
                            addReferencedFilesToTypeDirective(file, key);
                        }
                    });
                }
                return {
                    getReferencedExportContainer: getReferencedExportContainer,
                    getReferencedImportDeclaration: getReferencedImportDeclaration,
                    getReferencedDeclarationWithCollidingName: getReferencedDeclarationWithCollidingName,
                    isDeclarationWithCollidingName: isDeclarationWithCollidingName,
                    isValueAliasDeclaration: function (nodeIn) {
                        var node = ts.getParseTreeNode(nodeIn);
                        // Synthesized nodes are always treated like values.
                        return node ? isValueAliasDeclaration(node) : true;
                    },
                    hasGlobalName: hasGlobalName,
                    isReferencedAliasDeclaration: function (nodeIn, checkChildren) {
                        var node = ts.getParseTreeNode(nodeIn);
                        // Synthesized nodes are always treated as referenced.
                        return node ? isReferencedAliasDeclaration(node, checkChildren) : true;
                    },
                    getNodeCheckFlags: function (nodeIn) {
                        var node = ts.getParseTreeNode(nodeIn);
                        return node ? getNodeCheckFlags(node) : 0;
                    },
                    isTopLevelValueImportEqualsWithEntityName: isTopLevelValueImportEqualsWithEntityName,
                    isDeclarationVisible: isDeclarationVisible,
                    isImplementationOfOverload: isImplementationOfOverload,
                    isRequiredInitializedParameter: isRequiredInitializedParameter,
                    isOptionalUninitializedParameterProperty: isOptionalUninitializedParameterProperty,
                    isExpandoFunctionDeclaration: isExpandoFunctionDeclaration,
                    getPropertiesOfContainerFunction: getPropertiesOfContainerFunction,
                    createTypeOfDeclaration: createTypeOfDeclaration,
                    createReturnTypeOfSignatureDeclaration: createReturnTypeOfSignatureDeclaration,
                    createTypeOfExpression: createTypeOfExpression,
                    createLiteralConstValue: createLiteralConstValue,
                    isSymbolAccessible: isSymbolAccessible,
                    isEntityNameVisible: isEntityNameVisible,
                    getConstantValue: function (nodeIn) {
                        var node = ts.getParseTreeNode(nodeIn, canHaveConstantValue);
                        return node ? getConstantValue(node) : undefined;
                    },
                    collectLinkedAliases: collectLinkedAliases,
                    getReferencedValueDeclaration: getReferencedValueDeclaration,
                    getTypeReferenceSerializationKind: getTypeReferenceSerializationKind,
                    isOptionalParameter: isOptionalParameter,
                    moduleExportsSomeValue: moduleExportsSomeValue,
                    isArgumentsLocalBinding: isArgumentsLocalBinding,
                    getExternalModuleFileFromDeclaration: function (nodeIn) {
                        var node = ts.getParseTreeNode(nodeIn, ts.hasPossibleExternalModuleReference);
                        return node && getExternalModuleFileFromDeclaration(node);
                    },
                    getTypeReferenceDirectivesForEntityName: getTypeReferenceDirectivesForEntityName,
                    getTypeReferenceDirectivesForSymbol: getTypeReferenceDirectivesForSymbol,
                    isLiteralConstDeclaration: isLiteralConstDeclaration,
                    isLateBound: function (nodeIn) {
                        var node = ts.getParseTreeNode(nodeIn, ts.isDeclaration);
                        var symbol = node && getSymbolOfNode(node);
                        return !!(symbol && ts.getCheckFlags(symbol) & 4096 /* Late */);
                    },
                    getJsxFactoryEntity: getJsxFactoryEntity,
                    getJsxFragmentFactoryEntity: getJsxFragmentFactoryEntity,
                    getAllAccessorDeclarations: function (accessor) {
                        accessor = ts.getParseTreeNode(accessor, ts.isGetOrSetAccessorDeclaration); // TODO: GH#18217
                        var otherKind = accessor.kind === 171 /* SetAccessor */ ? 170 /* GetAccessor */ : 171 /* SetAccessor */;
                        var otherAccessor = ts.getDeclarationOfKind(getSymbolOfNode(accessor), otherKind);
                        var firstAccessor = otherAccessor && (otherAccessor.pos < accessor.pos) ? otherAccessor : accessor;
                        var secondAccessor = otherAccessor && (otherAccessor.pos < accessor.pos) ? accessor : otherAccessor;
                        var setAccessor = accessor.kind === 171 /* SetAccessor */ ? accessor : otherAccessor;
                        var getAccessor = accessor.kind === 170 /* GetAccessor */ ? accessor : otherAccessor;
                        return {
                            firstAccessor: firstAccessor,
                            secondAccessor: secondAccessor,
                            setAccessor: setAccessor,
                            getAccessor: getAccessor
                        };
                    },
                    getSymbolOfExternalModuleSpecifier: function (moduleName) { return resolveExternalModuleNameWorker(moduleName, moduleName, /*moduleNotFoundError*/ undefined); },
                    isBindingCapturedByNode: function (node, decl) {
                        var parseNode = ts.getParseTreeNode(node);
                        var parseDecl = ts.getParseTreeNode(decl);
                        return !!parseNode && !!parseDecl && (ts.isVariableDeclaration(parseDecl) || ts.isBindingElement(parseDecl)) && isBindingCapturedByNode(parseNode, parseDecl);
                    },
                    getDeclarationStatementsForSourceFile: function (node, flags, tracker, bundled) {
                        var n = ts.getParseTreeNode(node);
                        ts.Debug.assert(n && n.kind === 300 /* SourceFile */, "Non-sourcefile node passed into getDeclarationsForSourceFile");
                        var sym = getSymbolOfNode(node);
                        if (!sym) {
                            return !node.locals ? [] : nodeBuilder.symbolTableToDeclarationStatements(node.locals, node, flags, tracker, bundled);
                        }
                        return !sym.exports ? [] : nodeBuilder.symbolTableToDeclarationStatements(sym.exports, node, flags, tracker, bundled);
                    },
                    isImportRequiredByAugmentation: isImportRequiredByAugmentation,
                };
                function isImportRequiredByAugmentation(node) {
                    var file = ts.getSourceFileOfNode(node);
                    if (!file.symbol)
                        return false;
                    var importTarget = getExternalModuleFileFromDeclaration(node);
                    if (!importTarget)
                        return false;
                    if (importTarget === file)
                        return false;
                    var exports = getExportsOfModule(file.symbol);
                    for (var _i = 0, _a = ts.arrayFrom(exports.values()); _i < _a.length; _i++) {
                        var s = _a[_i];
                        if (s.mergeId) {
                            var merged = getMergedSymbol(s);
                            if (merged.declarations) {
                                for (var _b = 0, _c = merged.declarations; _b < _c.length; _b++) {
                                    var d = _c[_b];
                                    var declFile = ts.getSourceFileOfNode(d);
                                    if (declFile === importTarget) {
                                        return true;
                                    }
                                }
                            }
                        }
                    }
                    return false;
                }
                function isInHeritageClause(node) {
                    return node.parent && node.parent.kind === 226 /* ExpressionWithTypeArguments */ && node.parent.parent && node.parent.parent.kind === 289 /* HeritageClause */;
                }
                // defined here to avoid outer scope pollution
                function getTypeReferenceDirectivesForEntityName(node) {
                    // program does not have any files with type reference directives - bail out
                    if (!fileToDirective) {
                        return undefined;
                    }
                    // property access can only be used as values, or types when within an expression with type arguments inside a heritage clause
                    // qualified names can only be used as types\namespaces
                    // identifiers are treated as values only if they appear in type queries
                    var meaning = 788968 /* Type */ | 1920 /* Namespace */;
                    if ((node.kind === 79 /* Identifier */ && isInTypeQuery(node)) || (node.kind === 204 /* PropertyAccessExpression */ && !isInHeritageClause(node))) {
                        meaning = 111551 /* Value */ | 1048576 /* ExportValue */;
                    }
                    var symbol = resolveEntityName(node, meaning, /*ignoreErrors*/ true);
                    return symbol && symbol !== unknownSymbol ? getTypeReferenceDirectivesForSymbol(symbol, meaning) : undefined;
                }
                // defined here to avoid outer scope pollution
                function getTypeReferenceDirectivesForSymbol(symbol, meaning) {
                    // program does not have any files with type reference directives - bail out
                    if (!fileToDirective || !isSymbolFromTypeDeclarationFile(symbol)) {
                        return undefined;
                    }
                    // check what declarations in the symbol can contribute to the target meaning
                    var typeReferenceDirectives;
                    for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                        var decl = _a[_i];
                        // check meaning of the local symbol to see if declaration needs to be analyzed further
                        if (decl.symbol && decl.symbol.flags & meaning) {
                            var file = ts.getSourceFileOfNode(decl);
                            var typeReferenceDirective = fileToDirective.get(file.path);
                            if (typeReferenceDirective) {
                                (typeReferenceDirectives || (typeReferenceDirectives = [])).push(typeReferenceDirective);
                            }
                            else {
                                // found at least one entry that does not originate from type reference directive
                                return undefined;
                            }
                        }
                    }
                    return typeReferenceDirectives;
                }
                function isSymbolFromTypeDeclarationFile(symbol) {
                    // bail out if symbol does not have associated declarations (i.e. this is transient symbol created for property in binding pattern)
                    if (!symbol.declarations) {
                        return false;
                    }
                    // walk the parent chain for symbols to make sure that top level parent symbol is in the global scope
                    // external modules cannot define or contribute to type declaration files
                    var current = symbol;
                    while (true) {
                        var parent = getParentOfSymbol(current);
                        if (parent) {
                            current = parent;
                        }
                        else {
                            break;
                        }
                    }
                    if (current.valueDeclaration && current.valueDeclaration.kind === 300 /* SourceFile */ && current.flags & 512 /* ValueModule */) {
                        return false;
                    }
                    // check that at least one declaration of top level symbol originates from type declaration file
                    for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                        var decl = _a[_i];
                        var file = ts.getSourceFileOfNode(decl);
                        if (fileToDirective.has(file.path)) {
                            return true;
                        }
                    }
                    return false;
                }
                function addReferencedFilesToTypeDirective(file, key) {
                    if (fileToDirective.has(file.path))
                        return;
                    fileToDirective.set(file.path, key);
                    for (var _i = 0, _a = file.referencedFiles; _i < _a.length; _i++) {
                        var fileName = _a[_i].fileName;
                        var resolvedFile = ts.resolveTripleslashReference(fileName, file.fileName);
                        var referencedFile = host.getSourceFile(resolvedFile);
                        if (referencedFile) {
                            addReferencedFilesToTypeDirective(referencedFile, key);
                        }
                    }
                }
            }
            function getExternalModuleFileFromDeclaration(declaration) {
                var specifier = declaration.kind === 259 /* ModuleDeclaration */ ? ts.tryCast(declaration.name, ts.isStringLiteral) : ts.getExternalModuleName(declaration);
                var moduleSymbol = resolveExternalModuleNameWorker(specifier, specifier, /*moduleNotFoundError*/ undefined); // TODO: GH#18217
                if (!moduleSymbol) {
                    return undefined;
                }
                return ts.getDeclarationOfKind(moduleSymbol, 300 /* SourceFile */);
            }
            function initializeTypeChecker() {
                // Bind all source files and propagate errors
                for (var _i = 0, _a = host.getSourceFiles(); _i < _a.length; _i++) {
                    var file = _a[_i];
                    ts.bindSourceFile(file, compilerOptions);
                }
                amalgamatedDuplicates = new ts.Map();
                // Initialize global symbol table
                var augmentations;
                for (var _b = 0, _c = host.getSourceFiles(); _b < _c.length; _b++) {
                    var file = _c[_b];
                    if (file.redirectInfo) {
                        continue;
                    }
                    if (!ts.isExternalOrCommonJsModule(file)) {
                        // It is an error for a non-external-module (i.e. script) to declare its own `globalThis`.
                        // We can't use `builtinGlobals` for this due to synthetic expando-namespace generation in JS files.
                        var fileGlobalThisSymbol = file.locals.get("globalThis");
                        if (fileGlobalThisSymbol === null || fileGlobalThisSymbol === void 0 ? void 0 : fileGlobalThisSymbol.declarations) {
                            for (var _d = 0, _e = fileGlobalThisSymbol.declarations; _d < _e.length; _d++) {
                                var declaration = _e[_d];
                                diagnostics.add(ts.createDiagnosticForNode(declaration, ts.Diagnostics.Declaration_name_conflicts_with_built_in_global_identifier_0, "globalThis"));
                            }
                        }
                        mergeSymbolTable(globals, file.locals);
                    }
                    if (file.jsGlobalAugmentations) {
                        mergeSymbolTable(globals, file.jsGlobalAugmentations);
                    }
                    if (file.patternAmbientModules && file.patternAmbientModules.length) {
                        patternAmbientModules = ts.concatenate(patternAmbientModules, file.patternAmbientModules);
                    }
                    if (file.moduleAugmentations.length) {
                        (augmentations || (augmentations = [])).push(file.moduleAugmentations);
                    }
                    if (file.symbol && file.symbol.globalExports) {
                        // Merge in UMD exports with first-in-wins semantics (see #9771)
                        var source = file.symbol.globalExports;
                        source.forEach(function (sourceSymbol, id) {
                            if (!globals.has(id)) {
                                globals.set(id, sourceSymbol);
                            }
                        });
                    }
                }
                // We do global augmentations separately from module augmentations (and before creating global types) because they
                //  1. Affect global types. We won't have the correct global types until global augmentations are merged. Also,
                //  2. Module augmentation instantiation requires creating the type of a module, which, in turn, can require
                //       checking for an export or property on the module (if export=) which, in turn, can fall back to the
                //       apparent type of the module - either globalObjectType or globalFunctionType - which wouldn't exist if we
                //       did module augmentations prior to finalizing the global types.
                if (augmentations) {
                    // merge _global_ module augmentations.
                    // this needs to be done after global symbol table is initialized to make sure that all ambient modules are indexed
                    for (var _f = 0, augmentations_1 = augmentations; _f < augmentations_1.length; _f++) {
                        var list = augmentations_1[_f];
                        for (var _g = 0, list_1 = list; _g < list_1.length; _g++) {
                            var augmentation = list_1[_g];
                            if (!ts.isGlobalScopeAugmentation(augmentation.parent))
                                continue;
                            mergeModuleAugmentation(augmentation);
                        }
                    }
                }
                // Setup global builtins
                addToSymbolTable(globals, builtinGlobals, ts.Diagnostics.Declaration_name_conflicts_with_built_in_global_identifier_0);
                getSymbolLinks(undefinedSymbol).type = undefinedWideningType;
                getSymbolLinks(argumentsSymbol).type = getGlobalType("IArguments", /*arity*/ 0, /*reportErrors*/ true);
                getSymbolLinks(unknownSymbol).type = errorType;
                getSymbolLinks(globalThisSymbol).type = createObjectType(16 /* Anonymous */, globalThisSymbol);
                // Initialize special types
                globalArrayType = getGlobalType("Array", /*arity*/ 1, /*reportErrors*/ true);
                globalObjectType = getGlobalType("Object", /*arity*/ 0, /*reportErrors*/ true);
                globalFunctionType = getGlobalType("Function", /*arity*/ 0, /*reportErrors*/ true);
                globalCallableFunctionType = strictBindCallApply && getGlobalType("CallableFunction", /*arity*/ 0, /*reportErrors*/ true) || globalFunctionType;
                globalNewableFunctionType = strictBindCallApply && getGlobalType("NewableFunction", /*arity*/ 0, /*reportErrors*/ true) || globalFunctionType;
                globalStringType = getGlobalType("String", /*arity*/ 0, /*reportErrors*/ true);
                globalNumberType = getGlobalType("Number", /*arity*/ 0, /*reportErrors*/ true);
                globalBooleanType = getGlobalType("Boolean", /*arity*/ 0, /*reportErrors*/ true);
                globalRegExpType = getGlobalType("RegExp", /*arity*/ 0, /*reportErrors*/ true);
                anyArrayType = createArrayType(anyType);
                autoArrayType = createArrayType(autoType);
                if (autoArrayType === emptyObjectType) {
                    // autoArrayType is used as a marker, so even if global Array type is not defined, it needs to be a unique type
                    autoArrayType = createAnonymousType(undefined, emptySymbols, ts.emptyArray, ts.emptyArray, ts.emptyArray);
                }
                globalReadonlyArrayType = getGlobalTypeOrUndefined("ReadonlyArray", /*arity*/ 1) || globalArrayType;
                anyReadonlyArrayType = globalReadonlyArrayType ? createTypeFromGenericGlobalType(globalReadonlyArrayType, [anyType]) : anyArrayType;
                globalThisType = getGlobalTypeOrUndefined("ThisType", /*arity*/ 1);
                if (augmentations) {
                    // merge _nonglobal_ module augmentations.
                    // this needs to be done after global symbol table is initialized to make sure that all ambient modules are indexed
                    for (var _h = 0, augmentations_2 = augmentations; _h < augmentations_2.length; _h++) {
                        var list = augmentations_2[_h];
                        for (var _j = 0, list_2 = list; _j < list_2.length; _j++) {
                            var augmentation = list_2[_j];
                            if (ts.isGlobalScopeAugmentation(augmentation.parent))
                                continue;
                            mergeModuleAugmentation(augmentation);
                        }
                    }
                }
                amalgamatedDuplicates.forEach(function (_a) {
                    var firstFile = _a.firstFile, secondFile = _a.secondFile, conflictingSymbols = _a.conflictingSymbols;
                    // If not many things conflict, issue individual errors
                    if (conflictingSymbols.size < 8) {
                        conflictingSymbols.forEach(function (_a, symbolName) {
                            var isBlockScoped = _a.isBlockScoped, firstFileLocations = _a.firstFileLocations, secondFileLocations = _a.secondFileLocations;
                            var message = isBlockScoped ? ts.Diagnostics.Cannot_redeclare_block_scoped_variable_0 : ts.Diagnostics.Duplicate_identifier_0;
                            for (var _i = 0, firstFileLocations_1 = firstFileLocations; _i < firstFileLocations_1.length; _i++) {
                                var node = firstFileLocations_1[_i];
                                addDuplicateDeclarationError(node, message, symbolName, secondFileLocations);
                            }
                            for (var _b = 0, secondFileLocations_1 = secondFileLocations; _b < secondFileLocations_1.length; _b++) {
                                var node = secondFileLocations_1[_b];
                                addDuplicateDeclarationError(node, message, symbolName, firstFileLocations);
                            }
                        });
                    }
                    else {
                        // Otherwise issue top-level error since the files appear very identical in terms of what they contain
                        var list = ts.arrayFrom(conflictingSymbols.keys()).join(", ");
                        diagnostics.add(ts.addRelatedInfo(ts.createDiagnosticForNode(firstFile, ts.Diagnostics.Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0, list), ts.createDiagnosticForNode(secondFile, ts.Diagnostics.Conflicts_are_in_this_file)));
                        diagnostics.add(ts.addRelatedInfo(ts.createDiagnosticForNode(secondFile, ts.Diagnostics.Definitions_of_the_following_identifiers_conflict_with_those_in_another_file_Colon_0, list), ts.createDiagnosticForNode(firstFile, ts.Diagnostics.Conflicts_are_in_this_file)));
                    }
                });
                amalgamatedDuplicates = undefined;
            }
            function checkExternalEmitHelpers(location, helpers) {
                if ((requestedExternalEmitHelpers & helpers) !== helpers && compilerOptions.importHelpers) {
                    var sourceFile = ts.getSourceFileOfNode(location);
                    if (ts.isEffectiveExternalModule(sourceFile, compilerOptions) && !(location.flags & 8388608 /* Ambient */)) {
                        var helpersModule = resolveHelpersModule(sourceFile, location);
                        if (helpersModule !== unknownSymbol) {
                            var uncheckedHelpers = helpers & ~requestedExternalEmitHelpers;
                            for (var helper = 1 /* FirstEmitHelper */; helper <= 2097152 /* LastEmitHelper */; helper <<= 1) {
                                if (uncheckedHelpers & helper) {
                                    var name = getHelperName(helper);
                                    var symbol = getSymbol(helpersModule.exports, ts.escapeLeadingUnderscores(name), 111551 /* Value */);
                                    if (!symbol) {
                                        error(location, ts.Diagnostics.This_syntax_requires_an_imported_helper_named_1_which_does_not_exist_in_0_Consider_upgrading_your_version_of_0, ts.externalHelpersModuleNameText, name);
                                    }
                                    else if (helper & 524288 /* ClassPrivateFieldGet */) {
                                        if (!ts.some(getSignaturesOfSymbol(symbol), function (signature) { return getParameterCount(signature) > 3; })) {
                                            error(location, ts.Diagnostics.This_syntax_requires_an_imported_helper_named_1_with_2_parameters_which_is_not_compatible_with_the_one_in_0_Consider_upgrading_your_version_of_0, ts.externalHelpersModuleNameText, name, 4);
                                        }
                                    }
                                    else if (helper & 1048576 /* ClassPrivateFieldSet */) {
                                        if (!ts.some(getSignaturesOfSymbol(symbol), function (signature) { return getParameterCount(signature) > 4; })) {
                                            error(location, ts.Diagnostics.This_syntax_requires_an_imported_helper_named_1_with_2_parameters_which_is_not_compatible_with_the_one_in_0_Consider_upgrading_your_version_of_0, ts.externalHelpersModuleNameText, name, 5);
                                        }
                                    }
                                    else if (helper & 1024 /* SpreadArray */) {
                                        if (!ts.some(getSignaturesOfSymbol(symbol), function (signature) { return getParameterCount(signature) > 2; })) {
                                            error(location, ts.Diagnostics.This_syntax_requires_an_imported_helper_named_1_with_2_parameters_which_is_not_compatible_with_the_one_in_0_Consider_upgrading_your_version_of_0, ts.externalHelpersModuleNameText, name, 3);
                                        }
                                    }
                                }
                            }
                        }
                        requestedExternalEmitHelpers |= helpers;
                    }
                }
            }
            function getHelperName(helper) {
                switch (helper) {
                    case 1 /* Extends */: return "__extends";
                    case 2 /* Assign */: return "__assign";
                    case 4 /* Rest */: return "__rest";
                    case 8 /* Decorate */: return "__decorate";
                    case 16 /* Metadata */: return "__metadata";
                    case 32 /* Param */: return "__param";
                    case 64 /* Awaiter */: return "__awaiter";
                    case 128 /* Generator */: return "__generator";
                    case 256 /* Values */: return "__values";
                    case 512 /* Read */: return "__read";
                    case 1024 /* SpreadArray */: return "__spreadArray";
                    case 2048 /* Await */: return "__await";
                    case 4096 /* AsyncGenerator */: return "__asyncGenerator";
                    case 8192 /* AsyncDelegator */: return "__asyncDelegator";
                    case 16384 /* AsyncValues */: return "__asyncValues";
                    case 32768 /* ExportStar */: return "__exportStar";
                    case 65536 /* ImportStar */: return "__importStar";
                    case 131072 /* ImportDefault */: return "__importDefault";
                    case 262144 /* MakeTemplateObject */: return "__makeTemplateObject";
                    case 524288 /* ClassPrivateFieldGet */: return "__classPrivateFieldGet";
                    case 1048576 /* ClassPrivateFieldSet */: return "__classPrivateFieldSet";
                    case 2097152 /* CreateBinding */: return "__createBinding";
                    default: return ts.Debug.fail("Unrecognized helper");
                }
            }
            function resolveHelpersModule(node, errorNode) {
                if (!externalHelpersModule) {
                    externalHelpersModule = resolveExternalModule(node, ts.externalHelpersModuleNameText, ts.Diagnostics.This_syntax_requires_an_imported_helper_but_module_0_cannot_be_found, errorNode) || unknownSymbol;
                }
                return externalHelpersModule;
            }
            // GRAMMAR CHECKING
            function checkGrammarDecoratorsAndModifiers(node) {
                return checkGrammarDecorators(node) || checkGrammarModifiers(node);
            }
            function checkGrammarDecorators(node) {
                if (!node.decorators) {
                    return false;
                }
                if (!ts.nodeCanBeDecorated(node, node.parent, node.parent.parent)) {
                    if (node.kind === 167 /* MethodDeclaration */ && !ts.nodeIsPresent(node.body)) {
                        return grammarErrorOnFirstToken(node, ts.Diagnostics.A_decorator_can_only_decorate_a_method_implementation_not_an_overload);
                    }
                    else {
                        return grammarErrorOnFirstToken(node, ts.Diagnostics.Decorators_are_not_valid_here);
                    }
                }
                else if (node.kind === 170 /* GetAccessor */ || node.kind === 171 /* SetAccessor */) {
                    var accessors = ts.getAllAccessorDeclarations(node.parent.members, node);
                    if (accessors.firstAccessor.decorators && node === accessors.secondAccessor) {
                        return grammarErrorOnFirstToken(node, ts.Diagnostics.Decorators_cannot_be_applied_to_multiple_get_Slashset_accessors_of_the_same_name);
                    }
                }
                return false;
            }
            function checkGrammarModifiers(node) {
                var quickResult = reportObviousModifierErrors(node);
                if (quickResult !== undefined) {
                    return quickResult;
                }
                var lastStatic, lastDeclare, lastAsync, lastReadonly, lastOverride;
                var flags = 0 /* None */;
                for (var _i = 0, _a = node.modifiers; _i < _a.length; _i++) {
                    var modifier = _a[_i];
                    if (modifier.kind !== 143 /* ReadonlyKeyword */) {
                        if (node.kind === 164 /* PropertySignature */ || node.kind === 166 /* MethodSignature */) {
                            return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_a_type_member, ts.tokenToString(modifier.kind));
                        }
                        if (node.kind === 174 /* IndexSignature */ && (modifier.kind !== 124 /* StaticKeyword */ || !ts.isClassLike(node.parent))) {
                            return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_an_index_signature, ts.tokenToString(modifier.kind));
                        }
                    }
                    switch (modifier.kind) {
                        case 85 /* ConstKeyword */:
                            if (node.kind !== 258 /* EnumDeclaration */) {
                                return grammarErrorOnNode(node, ts.Diagnostics.A_class_member_cannot_have_the_0_keyword, ts.tokenToString(85 /* ConstKeyword */));
                            }
                            break;
                        case 157 /* OverrideKeyword */:
                            // If node.kind === SyntaxKind.Parameter, checkParameter reports an error if it's not a parameter property.
                            if (flags & 16384 /* Override */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_already_seen, "override");
                            }
                            else if (flags & 2 /* Ambient */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_with_1_modifier, "override", "declare");
                            }
                            else if (flags & 64 /* Readonly */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "override", "readonly");
                            }
                            else if (flags & 256 /* Async */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "override", "async");
                            }
                            flags |= 16384 /* Override */;
                            lastOverride = modifier;
                            break;
                        case 123 /* PublicKeyword */:
                        case 122 /* ProtectedKeyword */:
                        case 121 /* PrivateKeyword */:
                            var text = visibilityToString(ts.modifierToFlag(modifier.kind));
                            if (flags & 28 /* AccessibilityModifier */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics.Accessibility_modifier_already_seen);
                            }
                            else if (flags & 16384 /* Override */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, text, "override");
                            }
                            else if (flags & 32 /* Static */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, text, "static");
                            }
                            else if (flags & 64 /* Readonly */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, text, "readonly");
                            }
                            else if (flags & 256 /* Async */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, text, "async");
                            }
                            else if (node.parent.kind === 260 /* ModuleBlock */ || node.parent.kind === 300 /* SourceFile */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_a_module_or_namespace_element, text);
                            }
                            else if (flags & 128 /* Abstract */) {
                                if (modifier.kind === 121 /* PrivateKeyword */) {
                                    return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_with_1_modifier, text, "abstract");
                                }
                                else {
                                    return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, text, "abstract");
                                }
                            }
                            else if (ts.isPrivateIdentifierClassElementDeclaration(node)) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics.An_accessibility_modifier_cannot_be_used_with_a_private_identifier);
                            }
                            flags |= ts.modifierToFlag(modifier.kind);
                            break;
                        case 124 /* StaticKeyword */:
                            if (flags & 32 /* Static */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_already_seen, "static");
                            }
                            else if (flags & 64 /* Readonly */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "static", "readonly");
                            }
                            else if (flags & 256 /* Async */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "static", "async");
                            }
                            else if (node.parent.kind === 260 /* ModuleBlock */ || node.parent.kind === 300 /* SourceFile */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_a_module_or_namespace_element, "static");
                            }
                            else if (node.kind === 162 /* Parameter */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_a_parameter, "static");
                            }
                            else if (flags & 128 /* Abstract */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_with_1_modifier, "static", "abstract");
                            }
                            else if (flags & 16384 /* Override */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "static", "override");
                            }
                            flags |= 32 /* Static */;
                            lastStatic = modifier;
                            break;
                        case 143 /* ReadonlyKeyword */:
                            if (flags & 64 /* Readonly */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_already_seen, "readonly");
                            }
                            else if (node.kind !== 165 /* PropertyDeclaration */ && node.kind !== 164 /* PropertySignature */ && node.kind !== 174 /* IndexSignature */ && node.kind !== 162 /* Parameter */) {
                                // If node.kind === SyntaxKind.Parameter, checkParameter reports an error if it's not a parameter property.
                                return grammarErrorOnNode(modifier, ts.Diagnostics.readonly_modifier_can_only_appear_on_a_property_declaration_or_index_signature);
                            }
                            flags |= 64 /* Readonly */;
                            lastReadonly = modifier;
                            break;
                        case 93 /* ExportKeyword */:
                            if (flags & 1 /* Export */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_already_seen, "export");
                            }
                            else if (flags & 2 /* Ambient */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "export", "declare");
                            }
                            else if (flags & 128 /* Abstract */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "export", "abstract");
                            }
                            else if (flags & 256 /* Async */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "export", "async");
                            }
                            else if (ts.isClassLike(node.parent)) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_class_elements_of_this_kind, "export");
                            }
                            else if (node.kind === 162 /* Parameter */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_a_parameter, "export");
                            }
                            flags |= 1 /* Export */;
                            break;
                        case 88 /* DefaultKeyword */:
                            var container = node.parent.kind === 300 /* SourceFile */ ? node.parent : node.parent.parent;
                            if (container.kind === 259 /* ModuleDeclaration */ && !ts.isAmbientModule(container)) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics.A_default_export_can_only_be_used_in_an_ECMAScript_style_module);
                            }
                            else if (!(flags & 1 /* Export */)) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "export", "default");
                            }
                            flags |= 512 /* Default */;
                            break;
                        case 134 /* DeclareKeyword */:
                            if (flags & 2 /* Ambient */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_already_seen, "declare");
                            }
                            else if (flags & 256 /* Async */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_in_an_ambient_context, "async");
                            }
                            else if (flags & 16384 /* Override */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_in_an_ambient_context, "override");
                            }
                            else if (ts.isClassLike(node.parent) && !ts.isPropertyDeclaration(node)) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_class_elements_of_this_kind, "declare");
                            }
                            else if (node.kind === 162 /* Parameter */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_a_parameter, "declare");
                            }
                            else if ((node.parent.flags & 8388608 /* Ambient */) && node.parent.kind === 260 /* ModuleBlock */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics.A_declare_modifier_cannot_be_used_in_an_already_ambient_context);
                            }
                            else if (ts.isPrivateIdentifierClassElementDeclaration(node)) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_with_a_private_identifier, "declare");
                            }
                            flags |= 2 /* Ambient */;
                            lastDeclare = modifier;
                            break;
                        case 126 /* AbstractKeyword */:
                            if (flags & 128 /* Abstract */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_already_seen, "abstract");
                            }
                            if (node.kind !== 255 /* ClassDeclaration */ &&
                                node.kind !== 178 /* ConstructorType */) {
                                if (node.kind !== 167 /* MethodDeclaration */ &&
                                    node.kind !== 165 /* PropertyDeclaration */ &&
                                    node.kind !== 170 /* GetAccessor */ &&
                                    node.kind !== 171 /* SetAccessor */) {
                                    return grammarErrorOnNode(modifier, ts.Diagnostics.abstract_modifier_can_only_appear_on_a_class_method_or_property_declaration);
                                }
                                if (!(node.parent.kind === 255 /* ClassDeclaration */ && ts.hasSyntacticModifier(node.parent, 128 /* Abstract */))) {
                                    return grammarErrorOnNode(modifier, ts.Diagnostics.Abstract_methods_can_only_appear_within_an_abstract_class);
                                }
                                if (flags & 32 /* Static */) {
                                    return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_with_1_modifier, "static", "abstract");
                                }
                                if (flags & 8 /* Private */) {
                                    return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_with_1_modifier, "private", "abstract");
                                }
                                if (flags & 256 /* Async */ && lastAsync) {
                                    return grammarErrorOnNode(lastAsync, ts.Diagnostics._0_modifier_cannot_be_used_with_1_modifier, "async", "abstract");
                                }
                                if (flags & 16384 /* Override */) {
                                    return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_must_precede_1_modifier, "abstract", "override");
                                }
                            }
                            if (ts.isNamedDeclaration(node) && node.name.kind === 80 /* PrivateIdentifier */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_with_a_private_identifier, "abstract");
                            }
                            flags |= 128 /* Abstract */;
                            break;
                        case 130 /* AsyncKeyword */:
                            if (flags & 256 /* Async */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_already_seen, "async");
                            }
                            else if (flags & 2 /* Ambient */ || node.parent.flags & 8388608 /* Ambient */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_in_an_ambient_context, "async");
                            }
                            else if (node.kind === 162 /* Parameter */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_appear_on_a_parameter, "async");
                            }
                            if (flags & 128 /* Abstract */) {
                                return grammarErrorOnNode(modifier, ts.Diagnostics._0_modifier_cannot_be_used_with_1_modifier, "async", "abstract");
                            }
                            flags |= 256 /* Async */;
                            lastAsync = modifier;
                            break;
                    }
                }
                if (node.kind === 169 /* Constructor */) {
                    if (flags & 32 /* Static */) {
                        return grammarErrorOnNode(lastStatic, ts.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration, "static");
                    }
                    if (flags & 128 /* Abstract */) {
                        return grammarErrorOnNode(lastStatic, ts.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration, "abstract"); // TODO: GH#18217
                    }
                    if (flags & 16384 /* Override */) {
                        return grammarErrorOnNode(lastOverride, ts.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration, "override"); // TODO: GH#18217
                    }
                    else if (flags & 256 /* Async */) {
                        return grammarErrorOnNode(lastAsync, ts.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration, "async");
                    }
                    else if (flags & 64 /* Readonly */) {
                        return grammarErrorOnNode(lastReadonly, ts.Diagnostics._0_modifier_cannot_appear_on_a_constructor_declaration, "readonly");
                    }
                    return false;
                }
                else if ((node.kind === 264 /* ImportDeclaration */ || node.kind === 263 /* ImportEqualsDeclaration */) && flags & 2 /* Ambient */) {
                    return grammarErrorOnNode(lastDeclare, ts.Diagnostics.A_0_modifier_cannot_be_used_with_an_import_declaration, "declare");
                }
                else if (node.kind === 162 /* Parameter */ && (flags & 16476 /* ParameterPropertyModifier */) && ts.isBindingPattern(node.name)) {
                    return grammarErrorOnNode(node, ts.Diagnostics.A_parameter_property_may_not_be_declared_using_a_binding_pattern);
                }
                else if (node.kind === 162 /* Parameter */ && (flags & 16476 /* ParameterPropertyModifier */) && node.dotDotDotToken) {
                    return grammarErrorOnNode(node, ts.Diagnostics.A_parameter_property_cannot_be_declared_using_a_rest_parameter);
                }
                if (flags & 256 /* Async */) {
                    return checkGrammarAsyncModifier(node, lastAsync);
                }
                return false;
            }
            /**
             * true | false: Early return this value from checkGrammarModifiers.
             * undefined: Need to do full checking on the modifiers.
             */
            function reportObviousModifierErrors(node) {
                return !node.modifiers
                    ? false
                    : shouldReportBadModifier(node)
                        ? grammarErrorOnFirstToken(node, ts.Diagnostics.Modifiers_cannot_appear_here)
                        : undefined;
            }
            function shouldReportBadModifier(node) {
                switch (node.kind) {
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 169 /* Constructor */:
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 174 /* IndexSignature */:
                    case 259 /* ModuleDeclaration */:
                    case 264 /* ImportDeclaration */:
                    case 263 /* ImportEqualsDeclaration */:
                    case 270 /* ExportDeclaration */:
                    case 269 /* ExportAssignment */:
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                    case 162 /* Parameter */:
                        return false;
                    default:
                        if (node.parent.kind === 260 /* ModuleBlock */ || node.parent.kind === 300 /* SourceFile */) {
                            return false;
                        }
                        switch (node.kind) {
                            case 254 /* FunctionDeclaration */:
                                return nodeHasAnyModifiersExcept(node, 130 /* AsyncKeyword */);
                            case 255 /* ClassDeclaration */:
                            case 178 /* ConstructorType */:
                                return nodeHasAnyModifiersExcept(node, 126 /* AbstractKeyword */);
                            case 256 /* InterfaceDeclaration */:
                            case 235 /* VariableStatement */:
                            case 257 /* TypeAliasDeclaration */:
                            case 168 /* ClassStaticBlockDeclaration */:
                                return true;
                            case 258 /* EnumDeclaration */:
                                return nodeHasAnyModifiersExcept(node, 85 /* ConstKeyword */);
                            default:
                                ts.Debug.fail();
                        }
                }
            }
            function nodeHasAnyModifiersExcept(node, allowedModifier) {
                return node.modifiers.length > 1 || node.modifiers[0].kind !== allowedModifier;
            }
            function checkGrammarAsyncModifier(node, asyncModifier) {
                switch (node.kind) {
                    case 167 /* MethodDeclaration */:
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        return false;
                }
                return grammarErrorOnNode(asyncModifier, ts.Diagnostics._0_modifier_cannot_be_used_here, "async");
            }
            function checkGrammarForDisallowedTrailingComma(list, diag) {
                if (diag === void 0) {
                    diag = ts.Diagnostics.Trailing_comma_not_allowed;
                }
                if (list && list.hasTrailingComma) {
                    return grammarErrorAtPos(list[0], list.end - ",".length, ",".length, diag);
                }
                return false;
            }
            function checkGrammarTypeParameterList(typeParameters, file) {
                if (typeParameters && typeParameters.length === 0) {
                    var start = typeParameters.pos - "<".length;
                    var end = ts.skipTrivia(file.text, typeParameters.end) + ">".length;
                    return grammarErrorAtPos(file, start, end - start, ts.Diagnostics.Type_parameter_list_cannot_be_empty);
                }
                return false;
            }
            function checkGrammarParameterList(parameters) {
                var seenOptionalParameter = false;
                var parameterCount = parameters.length;
                for (var i = 0; i < parameterCount; i++) {
                    var parameter = parameters[i];
                    if (parameter.dotDotDotToken) {
                        if (i !== (parameterCount - 1)) {
                            return grammarErrorOnNode(parameter.dotDotDotToken, ts.Diagnostics.A_rest_parameter_must_be_last_in_a_parameter_list);
                        }
                        if (!(parameter.flags & 8388608 /* Ambient */)) { // Allow `...foo,` in ambient declarations; see GH#23070
                            checkGrammarForDisallowedTrailingComma(parameters, ts.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma);
                        }
                        if (parameter.questionToken) {
                            return grammarErrorOnNode(parameter.questionToken, ts.Diagnostics.A_rest_parameter_cannot_be_optional);
                        }
                        if (parameter.initializer) {
                            return grammarErrorOnNode(parameter.name, ts.Diagnostics.A_rest_parameter_cannot_have_an_initializer);
                        }
                    }
                    else if (isOptionalParameter(parameter)) {
                        seenOptionalParameter = true;
                        if (parameter.questionToken && parameter.initializer) {
                            return grammarErrorOnNode(parameter.name, ts.Diagnostics.Parameter_cannot_have_question_mark_and_initializer);
                        }
                    }
                    else if (seenOptionalParameter && !parameter.initializer) {
                        return grammarErrorOnNode(parameter.name, ts.Diagnostics.A_required_parameter_cannot_follow_an_optional_parameter);
                    }
                }
            }
            function getNonSimpleParameters(parameters) {
                return ts.filter(parameters, function (parameter) { return !!parameter.initializer || ts.isBindingPattern(parameter.name) || ts.isRestParameter(parameter); });
            }
            function checkGrammarForUseStrictSimpleParameterList(node) {
                if (languageVersion >= 3 /* ES2016 */) {
                    var useStrictDirective_1 = node.body && ts.isBlock(node.body) && ts.findUseStrictPrologue(node.body.statements);
                    if (useStrictDirective_1) {
                        var nonSimpleParameters = getNonSimpleParameters(node.parameters);
                        if (ts.length(nonSimpleParameters)) {
                            ts.forEach(nonSimpleParameters, function (parameter) {
                                ts.addRelatedInfo(error(parameter, ts.Diagnostics.This_parameter_is_not_allowed_with_use_strict_directive), ts.createDiagnosticForNode(useStrictDirective_1, ts.Diagnostics.use_strict_directive_used_here));
                            });
                            var diagnostics_2 = nonSimpleParameters.map(function (parameter, index) { return (index === 0 ? ts.createDiagnosticForNode(parameter, ts.Diagnostics.Non_simple_parameter_declared_here) : ts.createDiagnosticForNode(parameter, ts.Diagnostics.and_here)); });
                            ts.addRelatedInfo.apply(void 0, __spreadArray([error(useStrictDirective_1, ts.Diagnostics.use_strict_directive_cannot_be_used_with_non_simple_parameter_list)], diagnostics_2, false));
                            return true;
                        }
                    }
                }
                return false;
            }
            function checkGrammarFunctionLikeDeclaration(node) {
                // Prevent cascading error by short-circuit
                var file = ts.getSourceFileOfNode(node);
                return checkGrammarDecoratorsAndModifiers(node) ||
                    checkGrammarTypeParameterList(node.typeParameters, file) ||
                    checkGrammarParameterList(node.parameters) ||
                    checkGrammarArrowFunction(node, file) ||
                    (ts.isFunctionLikeDeclaration(node) && checkGrammarForUseStrictSimpleParameterList(node));
            }
            function checkGrammarClassLikeDeclaration(node) {
                var file = ts.getSourceFileOfNode(node);
                return checkGrammarClassDeclarationHeritageClauses(node) ||
                    checkGrammarTypeParameterList(node.typeParameters, file);
            }
            function checkGrammarArrowFunction(node, file) {
                if (!ts.isArrowFunction(node)) {
                    return false;
                }
                var equalsGreaterThanToken = node.equalsGreaterThanToken;
                var startLine = ts.getLineAndCharacterOfPosition(file, equalsGreaterThanToken.pos).line;
                var endLine = ts.getLineAndCharacterOfPosition(file, equalsGreaterThanToken.end).line;
                return startLine !== endLine && grammarErrorOnNode(equalsGreaterThanToken, ts.Diagnostics.Line_terminator_not_permitted_before_arrow);
            }
            function checkGrammarIndexSignatureParameters(node) {
                var parameter = node.parameters[0];
                if (node.parameters.length !== 1) {
                    if (parameter) {
                        return grammarErrorOnNode(parameter.name, ts.Diagnostics.An_index_signature_must_have_exactly_one_parameter);
                    }
                    else {
                        return grammarErrorOnNode(node, ts.Diagnostics.An_index_signature_must_have_exactly_one_parameter);
                    }
                }
                checkGrammarForDisallowedTrailingComma(node.parameters, ts.Diagnostics.An_index_signature_cannot_have_a_trailing_comma);
                if (parameter.dotDotDotToken) {
                    return grammarErrorOnNode(parameter.dotDotDotToken, ts.Diagnostics.An_index_signature_cannot_have_a_rest_parameter);
                }
                if (ts.hasEffectiveModifiers(parameter)) {
                    return grammarErrorOnNode(parameter.name, ts.Diagnostics.An_index_signature_parameter_cannot_have_an_accessibility_modifier);
                }
                if (parameter.questionToken) {
                    return grammarErrorOnNode(parameter.questionToken, ts.Diagnostics.An_index_signature_parameter_cannot_have_a_question_mark);
                }
                if (parameter.initializer) {
                    return grammarErrorOnNode(parameter.name, ts.Diagnostics.An_index_signature_parameter_cannot_have_an_initializer);
                }
                if (!parameter.type) {
                    return grammarErrorOnNode(parameter.name, ts.Diagnostics.An_index_signature_parameter_must_have_a_type_annotation);
                }
                var type = getTypeFromTypeNode(parameter.type);
                if (someType(type, function (t) { return !!(t.flags & 8576 /* StringOrNumberLiteralOrUnique */); }) || isGenericType(type)) {
                    return grammarErrorOnNode(parameter.name, ts.Diagnostics.An_index_signature_parameter_type_cannot_be_a_literal_type_or_generic_type_Consider_using_a_mapped_object_type_instead);
                }
                if (!everyType(type, isValidIndexKeyType)) {
                    return grammarErrorOnNode(parameter.name, ts.Diagnostics.An_index_signature_parameter_type_must_be_string_number_symbol_or_a_template_literal_type);
                }
                if (!node.type) {
                    return grammarErrorOnNode(node, ts.Diagnostics.An_index_signature_must_have_a_type_annotation);
                }
                return false;
            }
            function checkGrammarIndexSignature(node) {
                // Prevent cascading error by short-circuit
                return checkGrammarDecoratorsAndModifiers(node) || checkGrammarIndexSignatureParameters(node);
            }
            function checkGrammarForAtLeastOneTypeArgument(node, typeArguments) {
                if (typeArguments && typeArguments.length === 0) {
                    var sourceFile = ts.getSourceFileOfNode(node);
                    var start = typeArguments.pos - "<".length;
                    var end = ts.skipTrivia(sourceFile.text, typeArguments.end) + ">".length;
                    return grammarErrorAtPos(sourceFile, start, end - start, ts.Diagnostics.Type_argument_list_cannot_be_empty);
                }
                return false;
            }
            function checkGrammarTypeArguments(node, typeArguments) {
                return checkGrammarForDisallowedTrailingComma(typeArguments) ||
                    checkGrammarForAtLeastOneTypeArgument(node, typeArguments);
            }
            function checkGrammarTaggedTemplateChain(node) {
                if (node.questionDotToken || node.flags & 32 /* OptionalChain */) {
                    return grammarErrorOnNode(node.template, ts.Diagnostics.Tagged_template_expressions_are_not_permitted_in_an_optional_chain);
                }
                return false;
            }
            function checkGrammarForOmittedArgument(args) {
                if (args) {
                    for (var _i = 0, args_4 = args; _i < args_4.length; _i++) {
                        var arg = args_4[_i];
                        if (arg.kind === 225 /* OmittedExpression */) {
                            return grammarErrorAtPos(arg, arg.pos, 0, ts.Diagnostics.Argument_expression_expected);
                        }
                    }
                }
                return false;
            }
            function checkGrammarArguments(args) {
                return checkGrammarForOmittedArgument(args);
            }
            function checkGrammarHeritageClause(node) {
                var types = node.types;
                if (checkGrammarForDisallowedTrailingComma(types)) {
                    return true;
                }
                if (types && types.length === 0) {
                    var listType = ts.tokenToString(node.token);
                    return grammarErrorAtPos(node, types.pos, 0, ts.Diagnostics._0_list_cannot_be_empty, listType);
                }
                return ts.some(types, checkGrammarExpressionWithTypeArguments);
            }
            function checkGrammarExpressionWithTypeArguments(node) {
                return checkGrammarTypeArguments(node, node.typeArguments);
            }
            function checkGrammarClassDeclarationHeritageClauses(node) {
                var seenExtendsClause = false;
                var seenImplementsClause = false;
                if (!checkGrammarDecoratorsAndModifiers(node) && node.heritageClauses) {
                    for (var _i = 0, _a = node.heritageClauses; _i < _a.length; _i++) {
                        var heritageClause = _a[_i];
                        if (heritageClause.token === 94 /* ExtendsKeyword */) {
                            if (seenExtendsClause) {
                                return grammarErrorOnFirstToken(heritageClause, ts.Diagnostics.extends_clause_already_seen);
                            }
                            if (seenImplementsClause) {
                                return grammarErrorOnFirstToken(heritageClause, ts.Diagnostics.extends_clause_must_precede_implements_clause);
                            }
                            if (heritageClause.types.length > 1) {
                                return grammarErrorOnFirstToken(heritageClause.types[1], ts.Diagnostics.Classes_can_only_extend_a_single_class);
                            }
                            seenExtendsClause = true;
                        }
                        else {
                            ts.Debug.assert(heritageClause.token === 117 /* ImplementsKeyword */);
                            if (seenImplementsClause) {
                                return grammarErrorOnFirstToken(heritageClause, ts.Diagnostics.implements_clause_already_seen);
                            }
                            seenImplementsClause = true;
                        }
                        // Grammar checking heritageClause inside class declaration
                        checkGrammarHeritageClause(heritageClause);
                    }
                }
            }
            function checkGrammarInterfaceDeclaration(node) {
                var seenExtendsClause = false;
                if (node.heritageClauses) {
                    for (var _i = 0, _a = node.heritageClauses; _i < _a.length; _i++) {
                        var heritageClause = _a[_i];
                        if (heritageClause.token === 94 /* ExtendsKeyword */) {
                            if (seenExtendsClause) {
                                return grammarErrorOnFirstToken(heritageClause, ts.Diagnostics.extends_clause_already_seen);
                            }
                            seenExtendsClause = true;
                        }
                        else {
                            ts.Debug.assert(heritageClause.token === 117 /* ImplementsKeyword */);
                            return grammarErrorOnFirstToken(heritageClause, ts.Diagnostics.Interface_declaration_cannot_have_implements_clause);
                        }
                        // Grammar checking heritageClause inside class declaration
                        checkGrammarHeritageClause(heritageClause);
                    }
                }
                return false;
            }
            function checkGrammarComputedPropertyName(node) {
                // If node is not a computedPropertyName, just skip the grammar checking
                if (node.kind !== 160 /* ComputedPropertyName */) {
                    return false;
                }
                var computedPropertyName = node;
                if (computedPropertyName.expression.kind === 219 /* BinaryExpression */ && computedPropertyName.expression.operatorToken.kind === 27 /* CommaToken */) {
                    return grammarErrorOnNode(computedPropertyName.expression, ts.Diagnostics.A_comma_expression_is_not_allowed_in_a_computed_property_name);
                }
                return false;
            }
            function checkGrammarForGenerator(node) {
                if (node.asteriskToken) {
                    ts.Debug.assert(node.kind === 254 /* FunctionDeclaration */ ||
                        node.kind === 211 /* FunctionExpression */ ||
                        node.kind === 167 /* MethodDeclaration */);
                    if (node.flags & 8388608 /* Ambient */) {
                        return grammarErrorOnNode(node.asteriskToken, ts.Diagnostics.Generators_are_not_allowed_in_an_ambient_context);
                    }
                    if (!node.body) {
                        return grammarErrorOnNode(node.asteriskToken, ts.Diagnostics.An_overload_signature_cannot_be_declared_as_a_generator);
                    }
                }
            }
            function checkGrammarForInvalidQuestionMark(questionToken, message) {
                return !!questionToken && grammarErrorOnNode(questionToken, message);
            }
            function checkGrammarForInvalidExclamationToken(exclamationToken, message) {
                return !!exclamationToken && grammarErrorOnNode(exclamationToken, message);
            }
            function checkGrammarObjectLiteralExpression(node, inDestructuring) {
                var seen = new ts.Map();
                for (var _i = 0, _a = node.properties; _i < _a.length; _i++) {
                    var prop = _a[_i];
                    if (prop.kind === 293 /* SpreadAssignment */) {
                        if (inDestructuring) {
                            // a rest property cannot be destructured any further
                            var expression = ts.skipParentheses(prop.expression);
                            if (ts.isArrayLiteralExpression(expression) || ts.isObjectLiteralExpression(expression)) {
                                return grammarErrorOnNode(prop.expression, ts.Diagnostics.A_rest_element_cannot_contain_a_binding_pattern);
                            }
                        }
                        continue;
                    }
                    var name = prop.name;
                    if (name.kind === 160 /* ComputedPropertyName */) {
                        // If the name is not a ComputedPropertyName, the grammar checking will skip it
                        checkGrammarComputedPropertyName(name);
                    }
                    if (prop.kind === 292 /* ShorthandPropertyAssignment */ && !inDestructuring && prop.objectAssignmentInitializer) {
                        // having objectAssignmentInitializer is only valid in ObjectAssignmentPattern
                        // outside of destructuring it is a syntax error
                        return grammarErrorOnNode(prop.equalsToken, ts.Diagnostics.Did_you_mean_to_use_a_Colon_An_can_only_follow_a_property_name_when_the_containing_object_literal_is_part_of_a_destructuring_pattern);
                    }
                    if (name.kind === 80 /* PrivateIdentifier */) {
                        grammarErrorOnNode(name, ts.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies);
                    }
                    // Modifiers are never allowed on properties except for 'async' on a method declaration
                    if (prop.modifiers) {
                        // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-assertion
                        for (var _b = 0, _c = prop.modifiers; _b < _c.length; _b++) { // TODO: GH#19955
                            var mod = _c[_b];
                            if (mod.kind !== 130 /* AsyncKeyword */ || prop.kind !== 167 /* MethodDeclaration */) {
                                grammarErrorOnNode(mod, ts.Diagnostics._0_modifier_cannot_be_used_here, ts.getTextOfNode(mod));
                            }
                        }
                    }
                    // ECMA-262 11.1.5 Object Initializer
                    // If previous is not undefined then throw a SyntaxError exception if any of the following conditions are true
                    // a.This production is contained in strict code and IsDataDescriptor(previous) is true and
                    // IsDataDescriptor(propId.descriptor) is true.
                    //    b.IsDataDescriptor(previous) is true and IsAccessorDescriptor(propId.descriptor) is true.
                    //    c.IsAccessorDescriptor(previous) is true and IsDataDescriptor(propId.descriptor) is true.
                    //    d.IsAccessorDescriptor(previous) is true and IsAccessorDescriptor(propId.descriptor) is true
                    // and either both previous and propId.descriptor have[[Get]] fields or both previous and propId.descriptor have[[Set]] fields
                    var currentKind = void 0;
                    switch (prop.kind) {
                        case 292 /* ShorthandPropertyAssignment */:
                            checkGrammarForInvalidExclamationToken(prop.exclamationToken, ts.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context);
                        // falls through
                        case 291 /* PropertyAssignment */:
                            // Grammar checking for computedPropertyName and shorthandPropertyAssignment
                            checkGrammarForInvalidQuestionMark(prop.questionToken, ts.Diagnostics.An_object_member_cannot_be_declared_optional);
                            if (name.kind === 8 /* NumericLiteral */) {
                                checkGrammarNumericLiteral(name);
                            }
                            currentKind = 4 /* PropertyAssignment */;
                            break;
                        case 167 /* MethodDeclaration */:
                            currentKind = 8 /* Method */;
                            break;
                        case 170 /* GetAccessor */:
                            currentKind = 1 /* GetAccessor */;
                            break;
                        case 171 /* SetAccessor */:
                            currentKind = 2 /* SetAccessor */;
                            break;
                        default:
                            throw ts.Debug.assertNever(prop, "Unexpected syntax kind:" + prop.kind);
                    }
                    if (!inDestructuring) {
                        var effectiveName = ts.getPropertyNameForPropertyNameNode(name);
                        if (effectiveName === undefined) {
                            continue;
                        }
                        var existingKind = seen.get(effectiveName);
                        if (!existingKind) {
                            seen.set(effectiveName, currentKind);
                        }
                        else {
                            if ((currentKind & 12 /* PropertyAssignmentOrMethod */) && (existingKind & 12 /* PropertyAssignmentOrMethod */)) {
                                grammarErrorOnNode(name, ts.Diagnostics.Duplicate_identifier_0, ts.getTextOfNode(name));
                            }
                            else if ((currentKind & 3 /* GetOrSetAccessor */) && (existingKind & 3 /* GetOrSetAccessor */)) {
                                if (existingKind !== 3 /* GetOrSetAccessor */ && currentKind !== existingKind) {
                                    seen.set(effectiveName, currentKind | existingKind);
                                }
                                else {
                                    return grammarErrorOnNode(name, ts.Diagnostics.An_object_literal_cannot_have_multiple_get_Slashset_accessors_with_the_same_name);
                                }
                            }
                            else {
                                return grammarErrorOnNode(name, ts.Diagnostics.An_object_literal_cannot_have_property_and_accessor_with_the_same_name);
                            }
                        }
                    }
                }
            }
            function checkGrammarJsxElement(node) {
                checkGrammarJsxName(node.tagName);
                checkGrammarTypeArguments(node, node.typeArguments);
                var seen = new ts.Map();
                for (var _i = 0, _a = node.attributes.properties; _i < _a.length; _i++) {
                    var attr = _a[_i];
                    if (attr.kind === 285 /* JsxSpreadAttribute */) {
                        continue;
                    }
                    var name = attr.name, initializer = attr.initializer;
                    if (!seen.get(name.escapedText)) {
                        seen.set(name.escapedText, true);
                    }
                    else {
                        return grammarErrorOnNode(name, ts.Diagnostics.JSX_elements_cannot_have_multiple_attributes_with_the_same_name);
                    }
                    if (initializer && initializer.kind === 286 /* JsxExpression */ && !initializer.expression) {
                        return grammarErrorOnNode(initializer, ts.Diagnostics.JSX_attributes_must_only_be_assigned_a_non_empty_expression);
                    }
                }
            }
            function checkGrammarJsxName(node) {
                if (ts.isPropertyAccessExpression(node)) {
                    var propName = node;
                    do {
                        var check_1 = checkGrammarJsxNestedIdentifier(propName.name);
                        if (check_1) {
                            return check_1;
                        }
                        propName = propName.expression;
                    } while (ts.isPropertyAccessExpression(propName));
                    var check = checkGrammarJsxNestedIdentifier(propName);
                    if (check) {
                        return check;
                    }
                }
                function checkGrammarJsxNestedIdentifier(name) {
                    if (ts.isIdentifier(name) && ts.idText(name).indexOf(":") !== -1) {
                        return grammarErrorOnNode(name, ts.Diagnostics.JSX_property_access_expressions_cannot_include_JSX_namespace_names);
                    }
                }
            }
            function checkGrammarJsxExpression(node) {
                if (node.expression && ts.isCommaSequence(node.expression)) {
                    return grammarErrorOnNode(node.expression, ts.Diagnostics.JSX_expressions_may_not_use_the_comma_operator_Did_you_mean_to_write_an_array);
                }
            }
            function checkGrammarForInOrForOfStatement(forInOrOfStatement) {
                if (checkGrammarStatementInAmbientContext(forInOrOfStatement)) {
                    return true;
                }
                if (forInOrOfStatement.kind === 242 /* ForOfStatement */ && forInOrOfStatement.awaitModifier) {
                    if (!(forInOrOfStatement.flags & 32768 /* AwaitContext */)) {
                        var sourceFile = ts.getSourceFileOfNode(forInOrOfStatement);
                        if (ts.isInTopLevelContext(forInOrOfStatement)) {
                            if (!hasParseDiagnostics(sourceFile)) {
                                if (!ts.isEffectiveExternalModule(sourceFile, compilerOptions)) {
                                    diagnostics.add(ts.createDiagnosticForNode(forInOrOfStatement.awaitModifier, ts.Diagnostics.for_await_loops_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_has_no_imports_or_exports_Consider_adding_an_empty_export_to_make_this_file_a_module));
                                }
                                if ((moduleKind !== ts.ModuleKind.ESNext && moduleKind !== ts.ModuleKind.System) || languageVersion < 4 /* ES2017 */) {
                                    diagnostics.add(ts.createDiagnosticForNode(forInOrOfStatement.awaitModifier, ts.Diagnostics.Top_level_for_await_loops_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_target_option_is_set_to_es2017_or_higher));
                                }
                            }
                        }
                        else {
                            // use of 'for-await-of' in non-async function
                            if (!hasParseDiagnostics(sourceFile)) {
                                var diagnostic = ts.createDiagnosticForNode(forInOrOfStatement.awaitModifier, ts.Diagnostics.for_await_loops_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules);
                                var func = ts.getContainingFunction(forInOrOfStatement);
                                if (func && func.kind !== 169 /* Constructor */) {
                                    ts.Debug.assert((ts.getFunctionFlags(func) & 2 /* Async */) === 0, "Enclosing function should never be an async function.");
                                    var relatedInfo = ts.createDiagnosticForNode(func, ts.Diagnostics.Did_you_mean_to_mark_this_function_as_async);
                                    ts.addRelatedInfo(diagnostic, relatedInfo);
                                }
                                diagnostics.add(diagnostic);
                                return true;
                            }
                        }
                        return false;
                    }
                }
                if (ts.isForOfStatement(forInOrOfStatement) && !(forInOrOfStatement.flags & 32768 /* AwaitContext */) &&
                    ts.isIdentifier(forInOrOfStatement.initializer) && forInOrOfStatement.initializer.escapedText === "async") {
                    grammarErrorOnNode(forInOrOfStatement.initializer, ts.Diagnostics.The_left_hand_side_of_a_for_of_statement_may_not_be_async);
                    return false;
                }
                if (forInOrOfStatement.initializer.kind === 253 /* VariableDeclarationList */) {
                    var variableList = forInOrOfStatement.initializer;
                    if (!checkGrammarVariableDeclarationList(variableList)) {
                        var declarations = variableList.declarations;
                        // declarations.length can be zero if there is an error in variable declaration in for-of or for-in
                        // See http://www.ecma-international.org/ecma-262/6.0/#sec-for-in-and-for-of-statements for details
                        // For example:
                        //      var let = 10;
                        //      for (let of [1,2,3]) {} // this is invalid ES6 syntax
                        //      for (let in [1,2,3]) {} // this is invalid ES6 syntax
                        // We will then want to skip on grammar checking on variableList declaration
                        if (!declarations.length) {
                            return false;
                        }
                        if (declarations.length > 1) {
                            var diagnostic = forInOrOfStatement.kind === 241 /* ForInStatement */
                                ? ts.Diagnostics.Only_a_single_variable_declaration_is_allowed_in_a_for_in_statement
                                : ts.Diagnostics.Only_a_single_variable_declaration_is_allowed_in_a_for_of_statement;
                            return grammarErrorOnFirstToken(variableList.declarations[1], diagnostic);
                        }
                        var firstDeclaration = declarations[0];
                        if (firstDeclaration.initializer) {
                            var diagnostic = forInOrOfStatement.kind === 241 /* ForInStatement */
                                ? ts.Diagnostics.The_variable_declaration_of_a_for_in_statement_cannot_have_an_initializer
                                : ts.Diagnostics.The_variable_declaration_of_a_for_of_statement_cannot_have_an_initializer;
                            return grammarErrorOnNode(firstDeclaration.name, diagnostic);
                        }
                        if (firstDeclaration.type) {
                            var diagnostic = forInOrOfStatement.kind === 241 /* ForInStatement */
                                ? ts.Diagnostics.The_left_hand_side_of_a_for_in_statement_cannot_use_a_type_annotation
                                : ts.Diagnostics.The_left_hand_side_of_a_for_of_statement_cannot_use_a_type_annotation;
                            return grammarErrorOnNode(firstDeclaration, diagnostic);
                        }
                    }
                }
                return false;
            }
            function checkGrammarAccessor(accessor) {
                if (!(accessor.flags & 8388608 /* Ambient */) && (accessor.parent.kind !== 180 /* TypeLiteral */) && (accessor.parent.kind !== 256 /* InterfaceDeclaration */)) {
                    if (languageVersion < 1 /* ES5 */) {
                        return grammarErrorOnNode(accessor.name, ts.Diagnostics.Accessors_are_only_available_when_targeting_ECMAScript_5_and_higher);
                    }
                    if (languageVersion < 2 /* ES2015 */ && ts.isPrivateIdentifier(accessor.name)) {
                        return grammarErrorOnNode(accessor.name, ts.Diagnostics.Private_identifiers_are_only_available_when_targeting_ECMAScript_2015_and_higher);
                    }
                    if (accessor.body === undefined && !ts.hasSyntacticModifier(accessor, 128 /* Abstract */)) {
                        return grammarErrorAtPos(accessor, accessor.end - 1, ";".length, ts.Diagnostics._0_expected, "{");
                    }
                }
                if (accessor.body) {
                    if (ts.hasSyntacticModifier(accessor, 128 /* Abstract */)) {
                        return grammarErrorOnNode(accessor, ts.Diagnostics.An_abstract_accessor_cannot_have_an_implementation);
                    }
                    if (accessor.parent.kind === 180 /* TypeLiteral */ || accessor.parent.kind === 256 /* InterfaceDeclaration */) {
                        return grammarErrorOnNode(accessor.body, ts.Diagnostics.An_implementation_cannot_be_declared_in_ambient_contexts);
                    }
                }
                if (accessor.typeParameters) {
                    return grammarErrorOnNode(accessor.name, ts.Diagnostics.An_accessor_cannot_have_type_parameters);
                }
                if (!doesAccessorHaveCorrectParameterCount(accessor)) {
                    return grammarErrorOnNode(accessor.name, accessor.kind === 170 /* GetAccessor */ ?
                        ts.Diagnostics.A_get_accessor_cannot_have_parameters :
                        ts.Diagnostics.A_set_accessor_must_have_exactly_one_parameter);
                }
                if (accessor.kind === 171 /* SetAccessor */) {
                    if (accessor.type) {
                        return grammarErrorOnNode(accessor.name, ts.Diagnostics.A_set_accessor_cannot_have_a_return_type_annotation);
                    }
                    var parameter = ts.Debug.checkDefined(ts.getSetAccessorValueParameter(accessor), "Return value does not match parameter count assertion.");
                    if (parameter.dotDotDotToken) {
                        return grammarErrorOnNode(parameter.dotDotDotToken, ts.Diagnostics.A_set_accessor_cannot_have_rest_parameter);
                    }
                    if (parameter.questionToken) {
                        return grammarErrorOnNode(parameter.questionToken, ts.Diagnostics.A_set_accessor_cannot_have_an_optional_parameter);
                    }
                    if (parameter.initializer) {
                        return grammarErrorOnNode(accessor.name, ts.Diagnostics.A_set_accessor_parameter_cannot_have_an_initializer);
                    }
                }
                return false;
            }
            /** Does the accessor have the right number of parameters?
             * A get accessor has no parameters or a single `this` parameter.
             * A set accessor has one parameter or a `this` parameter and one more parameter.
             */
            function doesAccessorHaveCorrectParameterCount(accessor) {
                return getAccessorThisParameter(accessor) || accessor.parameters.length === (accessor.kind === 170 /* GetAccessor */ ? 0 : 1);
            }
            function getAccessorThisParameter(accessor) {
                if (accessor.parameters.length === (accessor.kind === 170 /* GetAccessor */ ? 1 : 2)) {
                    return ts.getThisParameter(accessor);
                }
            }
            function checkGrammarTypeOperatorNode(node) {
                if (node.operator === 152 /* UniqueKeyword */) {
                    if (node.type.kind !== 149 /* SymbolKeyword */) {
                        return grammarErrorOnNode(node.type, ts.Diagnostics._0_expected, ts.tokenToString(149 /* SymbolKeyword */));
                    }
                    var parent = ts.walkUpParenthesizedTypes(node.parent);
                    if (ts.isInJSFile(parent) && ts.isJSDocTypeExpression(parent)) {
                        parent = parent.parent;
                        if (ts.isJSDocTypeTag(parent)) {
                            // walk up past JSDoc comment node
                            parent = parent.parent.parent;
                        }
                    }
                    switch (parent.kind) {
                        case 252 /* VariableDeclaration */:
                            var decl = parent;
                            if (decl.name.kind !== 79 /* Identifier */) {
                                return grammarErrorOnNode(node, ts.Diagnostics.unique_symbol_types_may_not_be_used_on_a_variable_declaration_with_a_binding_name);
                            }
                            if (!ts.isVariableDeclarationInVariableStatement(decl)) {
                                return grammarErrorOnNode(node, ts.Diagnostics.unique_symbol_types_are_only_allowed_on_variables_in_a_variable_statement);
                            }
                            if (!(decl.parent.flags & 2 /* Const */)) {
                                return grammarErrorOnNode(parent.name, ts.Diagnostics.A_variable_whose_type_is_a_unique_symbol_type_must_be_const);
                            }
                            break;
                        case 165 /* PropertyDeclaration */:
                            if (!ts.isStatic(parent) ||
                                !ts.hasEffectiveReadonlyModifier(parent)) {
                                return grammarErrorOnNode(parent.name, ts.Diagnostics.A_property_of_a_class_whose_type_is_a_unique_symbol_type_must_be_both_static_and_readonly);
                            }
                            break;
                        case 164 /* PropertySignature */:
                            if (!ts.hasSyntacticModifier(parent, 64 /* Readonly */)) {
                                return grammarErrorOnNode(parent.name, ts.Diagnostics.A_property_of_an_interface_or_type_literal_whose_type_is_a_unique_symbol_type_must_be_readonly);
                            }
                            break;
                        default:
                            return grammarErrorOnNode(node, ts.Diagnostics.unique_symbol_types_are_not_allowed_here);
                    }
                }
                else if (node.operator === 143 /* ReadonlyKeyword */) {
                    if (node.type.kind !== 181 /* ArrayType */ && node.type.kind !== 182 /* TupleType */) {
                        return grammarErrorOnFirstToken(node, ts.Diagnostics.readonly_type_modifier_is_only_permitted_on_array_and_tuple_literal_types, ts.tokenToString(149 /* SymbolKeyword */));
                    }
                }
            }
            function checkGrammarForInvalidDynamicName(node, message) {
                if (isNonBindableDynamicName(node)) {
                    return grammarErrorOnNode(node, message);
                }
            }
            function checkGrammarMethod(node) {
                if (checkGrammarFunctionLikeDeclaration(node)) {
                    return true;
                }
                if (node.kind === 167 /* MethodDeclaration */) {
                    if (node.parent.kind === 203 /* ObjectLiteralExpression */) {
                        // We only disallow modifier on a method declaration if it is a property of object-literal-expression
                        if (node.modifiers && !(node.modifiers.length === 1 && ts.first(node.modifiers).kind === 130 /* AsyncKeyword */)) {
                            return grammarErrorOnFirstToken(node, ts.Diagnostics.Modifiers_cannot_appear_here);
                        }
                        else if (checkGrammarForInvalidQuestionMark(node.questionToken, ts.Diagnostics.An_object_member_cannot_be_declared_optional)) {
                            return true;
                        }
                        else if (checkGrammarForInvalidExclamationToken(node.exclamationToken, ts.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context)) {
                            return true;
                        }
                        else if (node.body === undefined) {
                            return grammarErrorAtPos(node, node.end - 1, ";".length, ts.Diagnostics._0_expected, "{");
                        }
                    }
                    if (checkGrammarForGenerator(node)) {
                        return true;
                    }
                }
                if (ts.isClassLike(node.parent)) {
                    if (languageVersion < 2 /* ES2015 */ && ts.isPrivateIdentifier(node.name)) {
                        return grammarErrorOnNode(node.name, ts.Diagnostics.Private_identifiers_are_only_available_when_targeting_ECMAScript_2015_and_higher);
                    }
                    // Technically, computed properties in ambient contexts is disallowed
                    // for property declarations and accessors too, not just methods.
                    // However, property declarations disallow computed names in general,
                    // and accessors are not allowed in ambient contexts in general,
                    // so this error only really matters for methods.
                    if (node.flags & 8388608 /* Ambient */) {
                        return checkGrammarForInvalidDynamicName(node.name, ts.Diagnostics.A_computed_property_name_in_an_ambient_context_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type);
                    }
                    else if (node.kind === 167 /* MethodDeclaration */ && !node.body) {
                        return checkGrammarForInvalidDynamicName(node.name, ts.Diagnostics.A_computed_property_name_in_a_method_overload_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type);
                    }
                }
                else if (node.parent.kind === 256 /* InterfaceDeclaration */) {
                    return checkGrammarForInvalidDynamicName(node.name, ts.Diagnostics.A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type);
                }
                else if (node.parent.kind === 180 /* TypeLiteral */) {
                    return checkGrammarForInvalidDynamicName(node.name, ts.Diagnostics.A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type);
                }
            }
            function checkGrammarBreakOrContinueStatement(node) {
                var current = node;
                while (current) {
                    if (ts.isFunctionLikeOrClassStaticBlockDeclaration(current)) {
                        return grammarErrorOnNode(node, ts.Diagnostics.Jump_target_cannot_cross_function_boundary);
                    }
                    switch (current.kind) {
                        case 248 /* LabeledStatement */:
                            if (node.label && current.label.escapedText === node.label.escapedText) {
                                // found matching label - verify that label usage is correct
                                // continue can only target labels that are on iteration statements
                                var isMisplacedContinueLabel = node.kind === 243 /* ContinueStatement */
                                    && !ts.isIterationStatement(current.statement, /*lookInLabeledStatement*/ true);
                                if (isMisplacedContinueLabel) {
                                    return grammarErrorOnNode(node, ts.Diagnostics.A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement);
                                }
                                return false;
                            }
                            break;
                        case 247 /* SwitchStatement */:
                            if (node.kind === 244 /* BreakStatement */ && !node.label) {
                                // unlabeled break within switch statement - ok
                                return false;
                            }
                            break;
                        default:
                            if (ts.isIterationStatement(current, /*lookInLabeledStatement*/ false) && !node.label) {
                                // unlabeled break or continue within iteration statement - ok
                                return false;
                            }
                            break;
                    }
                    current = current.parent;
                }
                if (node.label) {
                    var message = node.kind === 244 /* BreakStatement */
                        ? ts.Diagnostics.A_break_statement_can_only_jump_to_a_label_of_an_enclosing_statement
                        : ts.Diagnostics.A_continue_statement_can_only_jump_to_a_label_of_an_enclosing_iteration_statement;
                    return grammarErrorOnNode(node, message);
                }
                else {
                    var message = node.kind === 244 /* BreakStatement */
                        ? ts.Diagnostics.A_break_statement_can_only_be_used_within_an_enclosing_iteration_or_switch_statement
                        : ts.Diagnostics.A_continue_statement_can_only_be_used_within_an_enclosing_iteration_statement;
                    return grammarErrorOnNode(node, message);
                }
            }
            function checkGrammarBindingElement(node) {
                if (node.dotDotDotToken) {
                    var elements = node.parent.elements;
                    if (node !== ts.last(elements)) {
                        return grammarErrorOnNode(node, ts.Diagnostics.A_rest_element_must_be_last_in_a_destructuring_pattern);
                    }
                    checkGrammarForDisallowedTrailingComma(elements, ts.Diagnostics.A_rest_parameter_or_binding_pattern_may_not_have_a_trailing_comma);
                    if (node.propertyName) {
                        return grammarErrorOnNode(node.name, ts.Diagnostics.A_rest_element_cannot_have_a_property_name);
                    }
                }
                if (node.dotDotDotToken && node.initializer) {
                    // Error on equals token which immediately precedes the initializer
                    return grammarErrorAtPos(node, node.initializer.pos - 1, 1, ts.Diagnostics.A_rest_element_cannot_have_an_initializer);
                }
            }
            function isStringOrNumberLiteralExpression(expr) {
                return ts.isStringOrNumericLiteralLike(expr) ||
                    expr.kind === 217 /* PrefixUnaryExpression */ && expr.operator === 40 /* MinusToken */ &&
                        expr.operand.kind === 8 /* NumericLiteral */;
            }
            function isBigIntLiteralExpression(expr) {
                return expr.kind === 9 /* BigIntLiteral */ ||
                    expr.kind === 217 /* PrefixUnaryExpression */ && expr.operator === 40 /* MinusToken */ &&
                        expr.operand.kind === 9 /* BigIntLiteral */;
            }
            function isSimpleLiteralEnumReference(expr) {
                if ((ts.isPropertyAccessExpression(expr) || (ts.isElementAccessExpression(expr) && isStringOrNumberLiteralExpression(expr.argumentExpression))) &&
                    ts.isEntityNameExpression(expr.expression)) {
                    return !!(checkExpressionCached(expr).flags & 1024 /* EnumLiteral */);
                }
            }
            function checkAmbientInitializer(node) {
                var initializer = node.initializer;
                if (initializer) {
                    var isInvalidInitializer = !(isStringOrNumberLiteralExpression(initializer) ||
                        isSimpleLiteralEnumReference(initializer) ||
                        initializer.kind === 110 /* TrueKeyword */ || initializer.kind === 95 /* FalseKeyword */ ||
                        isBigIntLiteralExpression(initializer));
                    var isConstOrReadonly = ts.isDeclarationReadonly(node) || ts.isVariableDeclaration(node) && ts.isVarConst(node);
                    if (isConstOrReadonly && !node.type) {
                        if (isInvalidInitializer) {
                            return grammarErrorOnNode(initializer, ts.Diagnostics.A_const_initializer_in_an_ambient_context_must_be_a_string_or_numeric_literal_or_literal_enum_reference);
                        }
                    }
                    else {
                        return grammarErrorOnNode(initializer, ts.Diagnostics.Initializers_are_not_allowed_in_ambient_contexts);
                    }
                    if (!isConstOrReadonly || isInvalidInitializer) {
                        return grammarErrorOnNode(initializer, ts.Diagnostics.Initializers_are_not_allowed_in_ambient_contexts);
                    }
                }
            }
            function checkGrammarVariableDeclaration(node) {
                if (node.parent.parent.kind !== 241 /* ForInStatement */ && node.parent.parent.kind !== 242 /* ForOfStatement */) {
                    if (node.flags & 8388608 /* Ambient */) {
                        checkAmbientInitializer(node);
                    }
                    else if (!node.initializer) {
                        if (ts.isBindingPattern(node.name) && !ts.isBindingPattern(node.parent)) {
                            return grammarErrorOnNode(node, ts.Diagnostics.A_destructuring_declaration_must_have_an_initializer);
                        }
                        if (ts.isVarConst(node)) {
                            return grammarErrorOnNode(node, ts.Diagnostics.const_declarations_must_be_initialized);
                        }
                    }
                }
                if (node.exclamationToken && (node.parent.parent.kind !== 235 /* VariableStatement */ || !node.type || node.initializer || node.flags & 8388608 /* Ambient */)) {
                    var message = node.initializer
                        ? ts.Diagnostics.Declarations_with_initializers_cannot_also_have_definite_assignment_assertions
                        : !node.type
                            ? ts.Diagnostics.Declarations_with_definite_assignment_assertions_must_also_have_type_annotations
                            : ts.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context;
                    return grammarErrorOnNode(node.exclamationToken, message);
                }
                var moduleKind = ts.getEmitModuleKind(compilerOptions);
                if (moduleKind < ts.ModuleKind.ES2015 && moduleKind !== ts.ModuleKind.System &&
                    !(node.parent.parent.flags & 8388608 /* Ambient */) && ts.hasSyntacticModifier(node.parent.parent, 1 /* Export */)) {
                    checkESModuleMarker(node.name);
                }
                var checkLetConstNames = (ts.isLet(node) || ts.isVarConst(node));
                // 1. LexicalDeclaration : LetOrConst BindingList ;
                // It is a Syntax Error if the BoundNames of BindingList contains "let".
                // 2. ForDeclaration: ForDeclaration : LetOrConst ForBinding
                // It is a Syntax Error if the BoundNames of ForDeclaration contains "let".
                // It is a SyntaxError if a VariableDeclaration or VariableDeclarationNoIn occurs within strict code
                // and its Identifier is eval or arguments
                return checkLetConstNames && checkGrammarNameInLetOrConstDeclarations(node.name);
            }
            function checkESModuleMarker(name) {
                if (name.kind === 79 /* Identifier */) {
                    if (ts.idText(name) === "__esModule") {
                        return grammarErrorOnNodeSkippedOn("noEmit", name, ts.Diagnostics.Identifier_expected_esModule_is_reserved_as_an_exported_marker_when_transforming_ECMAScript_modules);
                    }
                }
                else {
                    var elements = name.elements;
                    for (var _i = 0, elements_1 = elements; _i < elements_1.length; _i++) {
                        var element = elements_1[_i];
                        if (!ts.isOmittedExpression(element)) {
                            return checkESModuleMarker(element.name);
                        }
                    }
                }
                return false;
            }
            function checkGrammarNameInLetOrConstDeclarations(name) {
                if (name.kind === 79 /* Identifier */) {
                    if (name.originalKeywordKind === 119 /* LetKeyword */) {
                        return grammarErrorOnNode(name, ts.Diagnostics.let_is_not_allowed_to_be_used_as_a_name_in_let_or_const_declarations);
                    }
                }
                else {
                    var elements = name.elements;
                    for (var _i = 0, elements_2 = elements; _i < elements_2.length; _i++) {
                        var element = elements_2[_i];
                        if (!ts.isOmittedExpression(element)) {
                            checkGrammarNameInLetOrConstDeclarations(element.name);
                        }
                    }
                }
                return false;
            }
            function checkGrammarVariableDeclarationList(declarationList) {
                var declarations = declarationList.declarations;
                if (checkGrammarForDisallowedTrailingComma(declarationList.declarations)) {
                    return true;
                }
                if (!declarationList.declarations.length) {
                    return grammarErrorAtPos(declarationList, declarations.pos, declarations.end - declarations.pos, ts.Diagnostics.Variable_declaration_list_cannot_be_empty);
                }
                return false;
            }
            function allowLetAndConstDeclarations(parent) {
                switch (parent.kind) {
                    case 237 /* IfStatement */:
                    case 238 /* DoStatement */:
                    case 239 /* WhileStatement */:
                    case 246 /* WithStatement */:
                    case 240 /* ForStatement */:
                    case 241 /* ForInStatement */:
                    case 242 /* ForOfStatement */:
                        return false;
                    case 248 /* LabeledStatement */:
                        return allowLetAndConstDeclarations(parent.parent);
                }
                return true;
            }
            function checkGrammarForDisallowedLetOrConstStatement(node) {
                if (!allowLetAndConstDeclarations(node.parent)) {
                    if (ts.isLet(node.declarationList)) {
                        return grammarErrorOnNode(node, ts.Diagnostics.let_declarations_can_only_be_declared_inside_a_block);
                    }
                    else if (ts.isVarConst(node.declarationList)) {
                        return grammarErrorOnNode(node, ts.Diagnostics.const_declarations_can_only_be_declared_inside_a_block);
                    }
                }
            }
            function checkGrammarMetaProperty(node) {
                var escapedText = node.name.escapedText;
                switch (node.keywordToken) {
                    case 103 /* NewKeyword */:
                        if (escapedText !== "target") {
                            return grammarErrorOnNode(node.name, ts.Diagnostics._0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2, node.name.escapedText, ts.tokenToString(node.keywordToken), "target");
                        }
                        break;
                    case 100 /* ImportKeyword */:
                        if (escapedText !== "meta") {
                            return grammarErrorOnNode(node.name, ts.Diagnostics._0_is_not_a_valid_meta_property_for_keyword_1_Did_you_mean_2, node.name.escapedText, ts.tokenToString(node.keywordToken), "meta");
                        }
                        break;
                }
            }
            function hasParseDiagnostics(sourceFile) {
                return sourceFile.parseDiagnostics.length > 0;
            }
            function grammarErrorOnFirstToken(node, message, arg0, arg1, arg2) {
                var sourceFile = ts.getSourceFileOfNode(node);
                if (!hasParseDiagnostics(sourceFile)) {
                    var span = ts.getSpanOfTokenAtPosition(sourceFile, node.pos);
                    diagnostics.add(ts.createFileDiagnostic(sourceFile, span.start, span.length, message, arg0, arg1, arg2));
                    return true;
                }
                return false;
            }
            function grammarErrorAtPos(nodeForSourceFile, start, length, message, arg0, arg1, arg2) {
                var sourceFile = ts.getSourceFileOfNode(nodeForSourceFile);
                if (!hasParseDiagnostics(sourceFile)) {
                    diagnostics.add(ts.createFileDiagnostic(sourceFile, start, length, message, arg0, arg1, arg2));
                    return true;
                }
                return false;
            }
            function grammarErrorOnNodeSkippedOn(key, node, message, arg0, arg1, arg2) {
                var sourceFile = ts.getSourceFileOfNode(node);
                if (!hasParseDiagnostics(sourceFile)) {
                    errorSkippedOn(key, node, message, arg0, arg1, arg2);
                    return true;
                }
                return false;
            }
            function grammarErrorOnNode(node, message, arg0, arg1, arg2) {
                var sourceFile = ts.getSourceFileOfNode(node);
                if (!hasParseDiagnostics(sourceFile)) {
                    diagnostics.add(ts.createDiagnosticForNode(node, message, arg0, arg1, arg2));
                    return true;
                }
                return false;
            }
            function checkGrammarConstructorTypeParameters(node) {
                var jsdocTypeParameters = ts.isInJSFile(node) ? ts.getJSDocTypeParameterDeclarations(node) : undefined;
                var range = node.typeParameters || jsdocTypeParameters && ts.firstOrUndefined(jsdocTypeParameters);
                if (range) {
                    var pos = range.pos === range.end ? range.pos : ts.skipTrivia(ts.getSourceFileOfNode(node).text, range.pos);
                    return grammarErrorAtPos(node, pos, range.end - pos, ts.Diagnostics.Type_parameters_cannot_appear_on_a_constructor_declaration);
                }
            }
            function checkGrammarConstructorTypeAnnotation(node) {
                var type = ts.getEffectiveReturnTypeNode(node);
                if (type) {
                    return grammarErrorOnNode(type, ts.Diagnostics.Type_annotation_cannot_appear_on_a_constructor_declaration);
                }
            }
            function checkGrammarProperty(node) {
                if (ts.isClassLike(node.parent)) {
                    if (ts.isStringLiteral(node.name) && node.name.text === "constructor") {
                        return grammarErrorOnNode(node.name, ts.Diagnostics.Classes_may_not_have_a_field_named_constructor);
                    }
                    if (checkGrammarForInvalidDynamicName(node.name, ts.Diagnostics.A_computed_property_name_in_a_class_property_declaration_must_have_a_simple_literal_type_or_a_unique_symbol_type)) {
                        return true;
                    }
                    if (languageVersion < 2 /* ES2015 */ && ts.isPrivateIdentifier(node.name)) {
                        return grammarErrorOnNode(node.name, ts.Diagnostics.Private_identifiers_are_only_available_when_targeting_ECMAScript_2015_and_higher);
                    }
                }
                else if (node.parent.kind === 256 /* InterfaceDeclaration */) {
                    if (checkGrammarForInvalidDynamicName(node.name, ts.Diagnostics.A_computed_property_name_in_an_interface_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type)) {
                        return true;
                    }
                    if (node.initializer) {
                        return grammarErrorOnNode(node.initializer, ts.Diagnostics.An_interface_property_cannot_have_an_initializer);
                    }
                }
                else if (node.parent.kind === 180 /* TypeLiteral */) {
                    if (checkGrammarForInvalidDynamicName(node.name, ts.Diagnostics.A_computed_property_name_in_a_type_literal_must_refer_to_an_expression_whose_type_is_a_literal_type_or_a_unique_symbol_type)) {
                        return true;
                    }
                    if (node.initializer) {
                        return grammarErrorOnNode(node.initializer, ts.Diagnostics.A_type_literal_property_cannot_have_an_initializer);
                    }
                }
                if (node.flags & 8388608 /* Ambient */) {
                    checkAmbientInitializer(node);
                }
                if (ts.isPropertyDeclaration(node) && node.exclamationToken && (!ts.isClassLike(node.parent) || !node.type || node.initializer ||
                    node.flags & 8388608 /* Ambient */ || ts.isStatic(node) || ts.hasAbstractModifier(node))) {
                    var message = node.initializer
                        ? ts.Diagnostics.Declarations_with_initializers_cannot_also_have_definite_assignment_assertions
                        : !node.type
                            ? ts.Diagnostics.Declarations_with_definite_assignment_assertions_must_also_have_type_annotations
                            : ts.Diagnostics.A_definite_assignment_assertion_is_not_permitted_in_this_context;
                    return grammarErrorOnNode(node.exclamationToken, message);
                }
            }
            function checkGrammarTopLevelElementForRequiredDeclareModifier(node) {
                // A declare modifier is required for any top level .d.ts declaration except export=, export default, export as namespace
                // interfaces and imports categories:
                //
                //  DeclarationElement:
                //     ExportAssignment
                //     export_opt   InterfaceDeclaration
                //     export_opt   TypeAliasDeclaration
                //     export_opt   ImportDeclaration
                //     export_opt   ExternalImportDeclaration
                //     export_opt   AmbientDeclaration
                //
                // TODO: The spec needs to be amended to reflect this grammar.
                if (node.kind === 256 /* InterfaceDeclaration */ ||
                    node.kind === 257 /* TypeAliasDeclaration */ ||
                    node.kind === 264 /* ImportDeclaration */ ||
                    node.kind === 263 /* ImportEqualsDeclaration */ ||
                    node.kind === 270 /* ExportDeclaration */ ||
                    node.kind === 269 /* ExportAssignment */ ||
                    node.kind === 262 /* NamespaceExportDeclaration */ ||
                    ts.hasSyntacticModifier(node, 2 /* Ambient */ | 1 /* Export */ | 512 /* Default */)) {
                    return false;
                }
                return grammarErrorOnFirstToken(node, ts.Diagnostics.Top_level_declarations_in_d_ts_files_must_start_with_either_a_declare_or_export_modifier);
            }
            function checkGrammarTopLevelElementsForRequiredDeclareModifier(file) {
                for (var _i = 0, _a = file.statements; _i < _a.length; _i++) {
                    var decl = _a[_i];
                    if (ts.isDeclaration(decl) || decl.kind === 235 /* VariableStatement */) {
                        if (checkGrammarTopLevelElementForRequiredDeclareModifier(decl)) {
                            return true;
                        }
                    }
                }
                return false;
            }
            function checkGrammarSourceFile(node) {
                return !!(node.flags & 8388608 /* Ambient */) && checkGrammarTopLevelElementsForRequiredDeclareModifier(node);
            }
            function checkGrammarStatementInAmbientContext(node) {
                if (node.flags & 8388608 /* Ambient */) {
                    // Find containing block which is either Block, ModuleBlock, SourceFile
                    var links = getNodeLinks(node);
                    if (!links.hasReportedStatementInAmbientContext && (ts.isFunctionLike(node.parent) || ts.isAccessor(node.parent))) {
                        return getNodeLinks(node).hasReportedStatementInAmbientContext = grammarErrorOnFirstToken(node, ts.Diagnostics.An_implementation_cannot_be_declared_in_ambient_contexts);
                    }
                    // We are either parented by another statement, or some sort of block.
                    // If we're in a block, we only want to really report an error once
                    // to prevent noisiness.  So use a bit on the block to indicate if
                    // this has already been reported, and don't report if it has.
                    //
                    if (node.parent.kind === 233 /* Block */ || node.parent.kind === 260 /* ModuleBlock */ || node.parent.kind === 300 /* SourceFile */) {
                        var links_2 = getNodeLinks(node.parent);
                        // Check if the containing block ever report this error
                        if (!links_2.hasReportedStatementInAmbientContext) {
                            return links_2.hasReportedStatementInAmbientContext = grammarErrorOnFirstToken(node, ts.Diagnostics.Statements_are_not_allowed_in_ambient_contexts);
                        }
                    }
                    else {
                        // We must be parented by a statement.  If so, there's no need
                        // to report the error as our parent will have already done it.
                        // Debug.assert(isStatement(node.parent));
                    }
                }
                return false;
            }
            function checkGrammarNumericLiteral(node) {
                // Grammar checking
                if (node.numericLiteralFlags & 32 /* Octal */) {
                    var diagnosticMessage = void 0;
                    if (languageVersion >= 1 /* ES5 */) {
                        diagnosticMessage = ts.Diagnostics.Octal_literals_are_not_available_when_targeting_ECMAScript_5_and_higher_Use_the_syntax_0;
                    }
                    else if (ts.isChildOfNodeWithKind(node, 194 /* LiteralType */)) {
                        diagnosticMessage = ts.Diagnostics.Octal_literal_types_must_use_ES2015_syntax_Use_the_syntax_0;
                    }
                    else if (ts.isChildOfNodeWithKind(node, 294 /* EnumMember */)) {
                        diagnosticMessage = ts.Diagnostics.Octal_literals_are_not_allowed_in_enums_members_initializer_Use_the_syntax_0;
                    }
                    if (diagnosticMessage) {
                        var withMinus = ts.isPrefixUnaryExpression(node.parent) && node.parent.operator === 40 /* MinusToken */;
                        var literal = (withMinus ? "-" : "") + "0o" + node.text;
                        return grammarErrorOnNode(withMinus ? node.parent : node, diagnosticMessage, literal);
                    }
                }
                // Realism (size) checking
                checkNumericLiteralValueSize(node);
                return false;
            }
            function checkNumericLiteralValueSize(node) {
                // Scientific notation (e.g. 2e54 and 1e00000000010) can't be converted to bigint
                // Literals with 15 or fewer characters aren't long enough to reach past 2^53 - 1
                // Fractional numbers (e.g. 9000000000000000.001) are inherently imprecise anyway
                if (node.numericLiteralFlags & 16 /* Scientific */ || node.text.length <= 15 || node.text.indexOf(".") !== -1) {
                    return;
                }
                // We can't rely on the runtime to accurately store and compare extremely large numeric values
                // Even for internal use, we use getTextOfNode: https://github.com/microsoft/TypeScript/issues/33298
                // Thus, if the runtime claims a too-large number is lower than Number.MAX_SAFE_INTEGER,
                // it's likely addition operations on it will fail too
                var apparentValue = +ts.getTextOfNode(node);
                if (apparentValue <= Math.pow(2, 53) - 1 && apparentValue + 1 > apparentValue) {
                    return;
                }
                addErrorOrSuggestion(/*isError*/ false, ts.createDiagnosticForNode(node, ts.Diagnostics.Numeric_literals_with_absolute_values_equal_to_2_53_or_greater_are_too_large_to_be_represented_accurately_as_integers));
            }
            function checkGrammarBigIntLiteral(node) {
                var literalType = ts.isLiteralTypeNode(node.parent) ||
                    ts.isPrefixUnaryExpression(node.parent) && ts.isLiteralTypeNode(node.parent.parent);
                if (!literalType) {
                    if (languageVersion < 7 /* ES2020 */) {
                        if (grammarErrorOnNode(node, ts.Diagnostics.BigInt_literals_are_not_available_when_targeting_lower_than_ES2020)) {
                            return true;
                        }
                    }
                }
                return false;
            }
            function grammarErrorAfterFirstToken(node, message, arg0, arg1, arg2) {
                var sourceFile = ts.getSourceFileOfNode(node);
                if (!hasParseDiagnostics(sourceFile)) {
                    var span = ts.getSpanOfTokenAtPosition(sourceFile, node.pos);
                    diagnostics.add(ts.createFileDiagnostic(sourceFile, ts.textSpanEnd(span), /*length*/ 0, message, arg0, arg1, arg2));
                    return true;
                }
                return false;
            }
            function getAmbientModules() {
                if (!ambientModulesCache) {
                    ambientModulesCache = [];
                    globals.forEach(function (global, sym) {
                        // No need to `unescapeLeadingUnderscores`, an escaped symbol is never an ambient module.
                        if (ambientModuleSymbolRegex.test(sym)) {
                            ambientModulesCache.push(global);
                        }
                    });
                }
                return ambientModulesCache;
            }
            function checkGrammarImportClause(node) {
                if (node.isTypeOnly && node.name && node.namedBindings) {
                    return grammarErrorOnNode(node, ts.Diagnostics.A_type_only_import_can_specify_a_default_import_or_named_bindings_but_not_both);
                }
                return false;
            }
            function checkGrammarImportCallExpression(node) {
                if (moduleKind === ts.ModuleKind.ES2015) {
                    return grammarErrorOnNode(node, ts.Diagnostics.Dynamic_imports_are_only_supported_when_the_module_flag_is_set_to_es2020_esnext_commonjs_amd_system_or_umd);
                }
                if (node.typeArguments) {
                    return grammarErrorOnNode(node, ts.Diagnostics.Dynamic_import_cannot_have_type_arguments);
                }
                var nodeArguments = node.arguments;
                if (nodeArguments.length !== 1) {
                    return grammarErrorOnNode(node, ts.Diagnostics.Dynamic_import_must_have_one_specifier_as_an_argument);
                }
                checkGrammarForDisallowedTrailingComma(nodeArguments);
                // see: parseArgumentOrArrayLiteralElement...we use this function which parse arguments of callExpression to parse specifier for dynamic import.
                // parseArgumentOrArrayLiteralElement allows spread element to be in an argument list which is not allowed as specifier in dynamic import.
                if (ts.isSpreadElement(nodeArguments[0])) {
                    return grammarErrorOnNode(nodeArguments[0], ts.Diagnostics.Specifier_of_dynamic_import_cannot_be_spread_element);
                }
                return false;
            }
            function findMatchingTypeReferenceOrTypeAliasReference(source, unionTarget) {
                var sourceObjectFlags = ts.getObjectFlags(source);
                if (sourceObjectFlags & (4 /* Reference */ | 16 /* Anonymous */) && unionTarget.flags & 1048576 /* Union */) {
                    return ts.find(unionTarget.types, function (target) {
                        if (target.flags & 524288 /* Object */) {
                            var overlapObjFlags = sourceObjectFlags & ts.getObjectFlags(target);
                            if (overlapObjFlags & 4 /* Reference */) {
                                return source.target === target.target;
                            }
                            if (overlapObjFlags & 16 /* Anonymous */) {
                                return !!source.aliasSymbol && source.aliasSymbol === target.aliasSymbol;
                            }
                        }
                        return false;
                    });
                }
            }
            function findBestTypeForObjectLiteral(source, unionTarget) {
                if (ts.getObjectFlags(source) & 128 /* ObjectLiteral */ && someType(unionTarget, isArrayLikeType)) {
                    return ts.find(unionTarget.types, function (t) { return !isArrayLikeType(t); });
                }
            }
            function findBestTypeForInvokable(source, unionTarget) {
                var signatureKind = 0 /* Call */;
                var hasSignatures = getSignaturesOfType(source, signatureKind).length > 0 ||
                    (signatureKind = 1 /* Construct */, getSignaturesOfType(source, signatureKind).length > 0);
                if (hasSignatures) {
                    return ts.find(unionTarget.types, function (t) { return getSignaturesOfType(t, signatureKind).length > 0; });
                }
            }
            function findMostOverlappyType(source, unionTarget) {
                var bestMatch;
                var matchingCount = 0;
                for (var _i = 0, _a = unionTarget.types; _i < _a.length; _i++) {
                    var target = _a[_i];
                    var overlap = getIntersectionType([getIndexType(source), getIndexType(target)]);
                    if (overlap.flags & 4194304 /* Index */) {
                        // perfect overlap of keys
                        bestMatch = target;
                        matchingCount = Infinity;
                    }
                    else if (overlap.flags & 1048576 /* Union */) {
                        // We only want to account for literal types otherwise.
                        // If we have a union of index types, it seems likely that we
                        // needed to elaborate between two generic mapped types anyway.
                        var len = ts.length(ts.filter(overlap.types, isUnitType));
                        if (len >= matchingCount) {
                            bestMatch = target;
                            matchingCount = len;
                        }
                    }
                    else if (isUnitType(overlap) && 1 >= matchingCount) {
                        bestMatch = target;
                        matchingCount = 1;
                    }
                }
                return bestMatch;
            }
            function filterPrimitivesIfContainsNonPrimitive(type) {
                if (maybeTypeOfKind(type, 67108864 /* NonPrimitive */)) {
                    var result = filterType(type, function (t) { return !(t.flags & 131068 /* Primitive */); });
                    if (!(result.flags & 131072 /* Never */)) {
                        return result;
                    }
                }
                return type;
            }
            // Keep this up-to-date with the same logic within `getApparentTypeOfContextualType`, since they should behave similarly
            function findMatchingDiscriminantType(source, target, isRelatedTo, skipPartial) {
                if (target.flags & 1048576 /* Union */ && source.flags & (2097152 /* Intersection */ | 524288 /* Object */)) {
                    var match = getMatchingUnionConstituentForType(target, source);
                    if (match) {
                        return match;
                    }
                    var sourceProperties = getPropertiesOfType(source);
                    if (sourceProperties) {
                        var sourcePropertiesFiltered = findDiscriminantProperties(sourceProperties, target);
                        if (sourcePropertiesFiltered) {
                            return discriminateTypeByDiscriminableItems(target, ts.map(sourcePropertiesFiltered, function (p) { return [function () { return getTypeOfSymbol(p); }, p.escapedName]; }), isRelatedTo, /*defaultValue*/ undefined, skipPartial);
                        }
                    }
                }
                return undefined;
            }
        }
        ts.createTypeChecker = createTypeChecker;
        function isNotAccessor(declaration) {
            // Accessors check for their own matching duplicates, and in contexts where they are valid, there are already duplicate identifier checks
            return !ts.isAccessor(declaration);
        }
        function isNotOverload(declaration) {
            return (declaration.kind !== 254 /* FunctionDeclaration */ && declaration.kind !== 167 /* MethodDeclaration */) ||
                !!declaration.body;
        }
        /** Like 'isDeclarationName', but returns true for LHS of `import { x as y }` or `export { x as y }`. */
        function isDeclarationNameOrImportPropertyName(name) {
            switch (name.parent.kind) {
                case 268 /* ImportSpecifier */:
                case 273 /* ExportSpecifier */:
                    return ts.isIdentifier(name);
                default:
                    return ts.isDeclarationName(name);
            }
        }
        var JsxNames;
        (function (JsxNames) {
            JsxNames.JSX = "JSX";
            JsxNames.IntrinsicElements = "IntrinsicElements";
            JsxNames.ElementClass = "ElementClass";
            JsxNames.ElementAttributesPropertyNameContainer = "ElementAttributesProperty"; // TODO: Deprecate and remove support
            JsxNames.ElementChildrenAttributeNameContainer = "ElementChildrenAttribute";
            JsxNames.Element = "Element";
            JsxNames.IntrinsicAttributes = "IntrinsicAttributes";
            JsxNames.IntrinsicClassAttributes = "IntrinsicClassAttributes";
            JsxNames.LibraryManagedAttributes = "LibraryManagedAttributes";
        })(JsxNames || (JsxNames = {}));
        function getIterationTypesKeyFromIterationTypeKind(typeKind) {
            switch (typeKind) {
                case 0 /* Yield */: return "yieldType";
                case 1 /* Return */: return "returnType";
                case 2 /* Next */: return "nextType";
            }
        }
        function signatureHasRestParameter(s) {
            return !!(s.flags & 1 /* HasRestParameter */);
        }
        ts.signatureHasRestParameter = signatureHasRestParameter;
        function signatureHasLiteralTypes(s) {
            return !!(s.flags & 2 /* HasLiteralTypes */);
        }
        ts.signatureHasLiteralTypes = signatureHasLiteralTypes;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        function visitNode(node, visitor, test, lift) {
            if (node === undefined || visitor === undefined) {
                return node;
            }
            var visited = visitor(node);
            if (visited === node) {
                return node;
            }
            var visitedNode;
            if (visited === undefined) {
                return undefined;
            }
            else if (ts.isArray(visited)) {
                visitedNode = (lift || extractSingleNode)(visited);
            }
            else {
                visitedNode = visited;
            }
            ts.Debug.assertNode(visitedNode, test);
            return visitedNode;
        }
        ts.visitNode = visitNode;
        /**
         * Visits a NodeArray using the supplied visitor, possibly returning a new NodeArray in its place.
         *
         * @param nodes The NodeArray to visit.
         * @param visitor The callback used to visit a Node.
         * @param test A node test to execute for each node.
         * @param start An optional value indicating the starting offset at which to start visiting.
         * @param count An optional value indicating the maximum number of nodes to visit.
         */
        function visitNodes(nodes, visitor, test, start, count) {
            if (nodes === undefined || visitor === undefined) {
                return nodes;
            }
            var updated;
            // Ensure start and count have valid values
            var length = nodes.length;
            if (start === undefined || start < 0) {
                start = 0;
            }
            if (count === undefined || count > length - start) {
                count = length - start;
            }
            var hasTrailingComma;
            var pos = -1;
            var end = -1;
            if (start > 0 || count < length) {
                // If we are not visiting all of the original nodes, we must always create a new array.
                // Since this is a fragment of a node array, we do not copy over the previous location
                // and will only copy over `hasTrailingComma` if we are including the last element.
                updated = [];
                hasTrailingComma = nodes.hasTrailingComma && start + count === length;
            }
            // Visit each original node.
            for (var i = 0; i < count; i++) {
                var node = nodes[i + start];
                var visited = node !== undefined ? visitor(node) : undefined;
                if (updated !== undefined || visited === undefined || visited !== node) {
                    if (updated === undefined) {
                        // Ensure we have a copy of `nodes`, up to the current index.
                        updated = nodes.slice(0, i);
                        hasTrailingComma = nodes.hasTrailingComma;
                        pos = nodes.pos;
                        end = nodes.end;
                    }
                    if (visited) {
                        if (ts.isArray(visited)) {
                            for (var _i = 0, visited_1 = visited; _i < visited_1.length; _i++) {
                                var visitedNode = visited_1[_i];
                                void ts.Debug.assertNode(visitedNode, test);
                                updated.push(visitedNode);
                            }
                        }
                        else {
                            void ts.Debug.assertNode(visited, test);
                            updated.push(visited);
                        }
                    }
                }
            }
            if (updated) {
                // TODO(rbuckton): Remove dependency on `ts.factory` in favor of a provided factory.
                var updatedArray = ts.factory.createNodeArray(updated, hasTrailingComma);
                ts.setTextRangePosEnd(updatedArray, pos, end);
                return updatedArray;
            }
            return nodes;
        }
        ts.visitNodes = visitNodes;
        /**
         * Starts a new lexical environment and visits a statement list, ending the lexical environment
         * and merging hoisted declarations upon completion.
         */
        function visitLexicalEnvironment(statements, visitor, context, start, ensureUseStrict, nodesVisitor) {
            if (nodesVisitor === void 0) {
                nodesVisitor = visitNodes;
            }
            context.startLexicalEnvironment();
            statements = nodesVisitor(statements, visitor, ts.isStatement, start);
            if (ensureUseStrict)
                statements = context.factory.ensureUseStrict(statements);
            return ts.factory.mergeLexicalEnvironment(statements, context.endLexicalEnvironment());
        }
        ts.visitLexicalEnvironment = visitLexicalEnvironment;
        function visitParameterList(nodes, visitor, context, nodesVisitor) {
            if (nodesVisitor === void 0) {
                nodesVisitor = visitNodes;
            }
            var updated;
            context.startLexicalEnvironment();
            if (nodes) {
                context.setLexicalEnvironmentFlags(1 /* InParameters */, true);
                updated = nodesVisitor(nodes, visitor, ts.isParameterDeclaration);
                // As of ES2015, any runtime execution of that occurs in for a parameter (such as evaluating an
                // initializer or a binding pattern), occurs in its own lexical scope. As a result, any expression
                // that we might transform that introduces a temporary variable would fail as the temporary variable
                // exists in a different lexical scope. To address this, we move any binding patterns and initializers
                // in a parameter list to the body if we detect a variable being hoisted while visiting a parameter list
                // when the emit target is greater than ES2015.
                if (context.getLexicalEnvironmentFlags() & 2 /* VariablesHoistedInParameters */ &&
                    ts.getEmitScriptTarget(context.getCompilerOptions()) >= 2 /* ES2015 */) {
                    updated = addDefaultValueAssignmentsIfNeeded(updated, context);
                }
                context.setLexicalEnvironmentFlags(1 /* InParameters */, false);
            }
            context.suspendLexicalEnvironment();
            return updated;
        }
        ts.visitParameterList = visitParameterList;
        function addDefaultValueAssignmentsIfNeeded(parameters, context) {
            var result;
            for (var i = 0; i < parameters.length; i++) {
                var parameter = parameters[i];
                var updated = addDefaultValueAssignmentIfNeeded(parameter, context);
                if (result || updated !== parameter) {
                    if (!result)
                        result = parameters.slice(0, i);
                    result[i] = updated;
                }
            }
            if (result) {
                return ts.setTextRange(context.factory.createNodeArray(result, parameters.hasTrailingComma), parameters);
            }
            return parameters;
        }
        function addDefaultValueAssignmentIfNeeded(parameter, context) {
            // A rest parameter cannot have a binding pattern or an initializer,
            // so let's just ignore it.
            return parameter.dotDotDotToken ? parameter :
                ts.isBindingPattern(parameter.name) ? addDefaultValueAssignmentForBindingPattern(parameter, context) :
                    parameter.initializer ? addDefaultValueAssignmentForInitializer(parameter, parameter.name, parameter.initializer, context) :
                        parameter;
        }
        function addDefaultValueAssignmentForBindingPattern(parameter, context) {
            var factory = context.factory;
            context.addInitializationStatement(factory.createVariableStatement(
            /*modifiers*/ undefined, factory.createVariableDeclarationList([
                factory.createVariableDeclaration(parameter.name, 
                /*exclamationToken*/ undefined, parameter.type, parameter.initializer ?
                    factory.createConditionalExpression(factory.createStrictEquality(factory.getGeneratedNameForNode(parameter), factory.createVoidZero()), 
                    /*questionToken*/ undefined, parameter.initializer, 
                    /*colonToken*/ undefined, factory.getGeneratedNameForNode(parameter)) :
                    factory.getGeneratedNameForNode(parameter)),
            ])));
            return factory.updateParameterDeclaration(parameter, parameter.decorators, parameter.modifiers, parameter.dotDotDotToken, factory.getGeneratedNameForNode(parameter), parameter.questionToken, parameter.type, 
            /*initializer*/ undefined);
        }
        function addDefaultValueAssignmentForInitializer(parameter, name, initializer, context) {
            var factory = context.factory;
            context.addInitializationStatement(factory.createIfStatement(factory.createTypeCheck(factory.cloneNode(name), "undefined"), ts.setEmitFlags(ts.setTextRange(factory.createBlock([
                factory.createExpressionStatement(ts.setEmitFlags(ts.setTextRange(factory.createAssignment(ts.setEmitFlags(factory.cloneNode(name), 48 /* NoSourceMap */), ts.setEmitFlags(initializer, 48 /* NoSourceMap */ | ts.getEmitFlags(initializer) | 1536 /* NoComments */)), parameter), 1536 /* NoComments */))
            ]), parameter), 1 /* SingleLine */ | 32 /* NoTrailingSourceMap */ | 384 /* NoTokenSourceMaps */ | 1536 /* NoComments */)));
            return factory.updateParameterDeclaration(parameter, parameter.decorators, parameter.modifiers, parameter.dotDotDotToken, parameter.name, parameter.questionToken, parameter.type, 
            /*initializer*/ undefined);
        }
        function visitFunctionBody(node, visitor, context, nodeVisitor) {
            if (nodeVisitor === void 0) {
                nodeVisitor = visitNode;
            }
            context.resumeLexicalEnvironment();
            var updated = nodeVisitor(node, visitor, ts.isConciseBody);
            var declarations = context.endLexicalEnvironment();
            if (ts.some(declarations)) {
                if (!updated) {
                    return context.factory.createBlock(declarations);
                }
                var block = context.factory.converters.convertToFunctionBlock(updated);
                var statements = ts.factory.mergeLexicalEnvironment(block.statements, declarations);
                return context.factory.updateBlock(block, statements);
            }
            return updated;
        }
        ts.visitFunctionBody = visitFunctionBody;
        /**
         * Visits an iteration body, adding any block-scoped variables required by the transformation.
         */
        function visitIterationBody(body, visitor, context) {
            context.startBlockScope();
            var updated = visitNode(body, visitor, ts.isStatement, context.factory.liftToBlock);
            var declarations = context.endBlockScope();
            if (ts.some(declarations)) {
                if (ts.isBlock(updated)) {
                    declarations.push.apply(declarations, updated.statements);
                    return context.factory.updateBlock(updated, declarations);
                }
                declarations.push(updated);
                return context.factory.createBlock(declarations);
            }
            return updated;
        }
        ts.visitIterationBody = visitIterationBody;
        function visitEachChild(node, visitor, context, nodesVisitor, tokenVisitor, nodeVisitor) {
            if (nodesVisitor === void 0) {
                nodesVisitor = visitNodes;
            }
            if (nodeVisitor === void 0) {
                nodeVisitor = visitNode;
            }
            if (node === undefined) {
                return undefined;
            }
            var kind = node.kind;
            // No need to visit nodes with no children.
            if ((kind > 0 /* FirstToken */ && kind <= 158 /* LastToken */) || kind === 190 /* ThisType */) {
                return node;
            }
            var factory = context.factory;
            switch (kind) {
                // Names
                case 79 /* Identifier */:
                    ts.Debug.type(node);
                    return factory.updateIdentifier(node, nodesVisitor(node.typeArguments, visitor, ts.isTypeNodeOrTypeParameterDeclaration));
                case 159 /* QualifiedName */:
                    ts.Debug.type(node);
                    return factory.updateQualifiedName(node, nodeVisitor(node.left, visitor, ts.isEntityName), nodeVisitor(node.right, visitor, ts.isIdentifier));
                case 160 /* ComputedPropertyName */:
                    ts.Debug.type(node);
                    return factory.updateComputedPropertyName(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                // Signature elements
                case 161 /* TypeParameter */:
                    ts.Debug.type(node);
                    return factory.updateTypeParameterDeclaration(node, nodeVisitor(node.name, visitor, ts.isIdentifier), nodeVisitor(node.constraint, visitor, ts.isTypeNode), nodeVisitor(node.default, visitor, ts.isTypeNode));
                case 162 /* Parameter */:
                    ts.Debug.type(node);
                    return factory.updateParameterDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.dotDotDotToken, tokenVisitor, ts.isDotDotDotToken), nodeVisitor(node.name, visitor, ts.isBindingName), nodeVisitor(node.questionToken, tokenVisitor, ts.isQuestionToken), nodeVisitor(node.type, visitor, ts.isTypeNode), nodeVisitor(node.initializer, visitor, ts.isExpression));
                case 163 /* Decorator */:
                    ts.Debug.type(node);
                    return factory.updateDecorator(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                // Type elements
                case 164 /* PropertySignature */:
                    ts.Debug.type(node);
                    return factory.updatePropertySignature(node, nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isPropertyName), nodeVisitor(node.questionToken, tokenVisitor, ts.isToken), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 165 /* PropertyDeclaration */:
                    ts.Debug.type(node);
                    return factory.updatePropertyDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isPropertyName), 
                    // QuestionToken and ExclamationToken is uniqued in Property Declaration and the signature of 'updateProperty' is that too
                    nodeVisitor(node.questionToken || node.exclamationToken, tokenVisitor, ts.isQuestionOrExclamationToken), nodeVisitor(node.type, visitor, ts.isTypeNode), nodeVisitor(node.initializer, visitor, ts.isExpression));
                case 166 /* MethodSignature */:
                    ts.Debug.type(node);
                    return factory.updateMethodSignature(node, nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isPropertyName), nodeVisitor(node.questionToken, tokenVisitor, ts.isQuestionToken), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), nodesVisitor(node.parameters, visitor, ts.isParameterDeclaration), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 167 /* MethodDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateMethodDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.asteriskToken, tokenVisitor, ts.isAsteriskToken), nodeVisitor(node.name, visitor, ts.isPropertyName), nodeVisitor(node.questionToken, tokenVisitor, ts.isQuestionToken), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), visitParameterList(node.parameters, visitor, context, nodesVisitor), nodeVisitor(node.type, visitor, ts.isTypeNode), visitFunctionBody(node.body, visitor, context, nodeVisitor));
                case 169 /* Constructor */:
                    ts.Debug.type(node);
                    return factory.updateConstructorDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), visitParameterList(node.parameters, visitor, context, nodesVisitor), visitFunctionBody(node.body, visitor, context, nodeVisitor));
                case 170 /* GetAccessor */:
                    ts.Debug.type(node);
                    return factory.updateGetAccessorDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isPropertyName), visitParameterList(node.parameters, visitor, context, nodesVisitor), nodeVisitor(node.type, visitor, ts.isTypeNode), visitFunctionBody(node.body, visitor, context, nodeVisitor));
                case 171 /* SetAccessor */:
                    ts.Debug.type(node);
                    return factory.updateSetAccessorDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isPropertyName), visitParameterList(node.parameters, visitor, context, nodesVisitor), visitFunctionBody(node.body, visitor, context, nodeVisitor));
                case 168 /* ClassStaticBlockDeclaration */:
                    ts.Debug.type(node);
                    context.startLexicalEnvironment();
                    context.suspendLexicalEnvironment();
                    return factory.updateClassStaticBlockDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), visitFunctionBody(node.body, visitor, context, nodeVisitor));
                case 172 /* CallSignature */:
                    ts.Debug.type(node);
                    return factory.updateCallSignature(node, nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), nodesVisitor(node.parameters, visitor, ts.isParameterDeclaration), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 173 /* ConstructSignature */:
                    ts.Debug.type(node);
                    return factory.updateConstructSignature(node, nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), nodesVisitor(node.parameters, visitor, ts.isParameterDeclaration), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 174 /* IndexSignature */:
                    ts.Debug.type(node);
                    return factory.updateIndexSignature(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodesVisitor(node.parameters, visitor, ts.isParameterDeclaration), nodeVisitor(node.type, visitor, ts.isTypeNode));
                // Types
                case 175 /* TypePredicate */:
                    ts.Debug.type(node);
                    return factory.updateTypePredicateNode(node, nodeVisitor(node.assertsModifier, visitor, ts.isAssertsKeyword), nodeVisitor(node.parameterName, visitor, ts.isIdentifierOrThisTypeNode), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 176 /* TypeReference */:
                    ts.Debug.type(node);
                    return factory.updateTypeReferenceNode(node, nodeVisitor(node.typeName, visitor, ts.isEntityName), nodesVisitor(node.typeArguments, visitor, ts.isTypeNode));
                case 177 /* FunctionType */:
                    ts.Debug.type(node);
                    return factory.updateFunctionTypeNode(node, nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), nodesVisitor(node.parameters, visitor, ts.isParameterDeclaration), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 178 /* ConstructorType */:
                    ts.Debug.type(node);
                    return factory.updateConstructorTypeNode(node, nodesVisitor(node.modifiers, visitor, ts.isModifier), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), nodesVisitor(node.parameters, visitor, ts.isParameterDeclaration), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 179 /* TypeQuery */:
                    ts.Debug.type(node);
                    return factory.updateTypeQueryNode(node, nodeVisitor(node.exprName, visitor, ts.isEntityName));
                case 180 /* TypeLiteral */:
                    ts.Debug.type(node);
                    return factory.updateTypeLiteralNode(node, nodesVisitor(node.members, visitor, ts.isTypeElement));
                case 181 /* ArrayType */:
                    ts.Debug.type(node);
                    return factory.updateArrayTypeNode(node, nodeVisitor(node.elementType, visitor, ts.isTypeNode));
                case 182 /* TupleType */:
                    ts.Debug.type(node);
                    return factory.updateTupleTypeNode(node, nodesVisitor(node.elements, visitor, ts.isTypeNode));
                case 183 /* OptionalType */:
                    ts.Debug.type(node);
                    return factory.updateOptionalTypeNode(node, nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 184 /* RestType */:
                    ts.Debug.type(node);
                    return factory.updateRestTypeNode(node, nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 185 /* UnionType */:
                    ts.Debug.type(node);
                    return factory.updateUnionTypeNode(node, nodesVisitor(node.types, visitor, ts.isTypeNode));
                case 186 /* IntersectionType */:
                    ts.Debug.type(node);
                    return factory.updateIntersectionTypeNode(node, nodesVisitor(node.types, visitor, ts.isTypeNode));
                case 187 /* ConditionalType */:
                    ts.Debug.type(node);
                    return factory.updateConditionalTypeNode(node, nodeVisitor(node.checkType, visitor, ts.isTypeNode), nodeVisitor(node.extendsType, visitor, ts.isTypeNode), nodeVisitor(node.trueType, visitor, ts.isTypeNode), nodeVisitor(node.falseType, visitor, ts.isTypeNode));
                case 188 /* InferType */:
                    ts.Debug.type(node);
                    return factory.updateInferTypeNode(node, nodeVisitor(node.typeParameter, visitor, ts.isTypeParameterDeclaration));
                case 198 /* ImportType */:
                    ts.Debug.type(node);
                    return factory.updateImportTypeNode(node, nodeVisitor(node.argument, visitor, ts.isTypeNode), nodeVisitor(node.qualifier, visitor, ts.isEntityName), visitNodes(node.typeArguments, visitor, ts.isTypeNode), node.isTypeOf);
                case 195 /* NamedTupleMember */:
                    ts.Debug.type(node);
                    return factory.updateNamedTupleMember(node, visitNode(node.dotDotDotToken, visitor, ts.isDotDotDotToken), visitNode(node.name, visitor, ts.isIdentifier), visitNode(node.questionToken, visitor, ts.isQuestionToken), visitNode(node.type, visitor, ts.isTypeNode));
                case 189 /* ParenthesizedType */:
                    ts.Debug.type(node);
                    return factory.updateParenthesizedType(node, nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 191 /* TypeOperator */:
                    ts.Debug.type(node);
                    return factory.updateTypeOperatorNode(node, nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 192 /* IndexedAccessType */:
                    ts.Debug.type(node);
                    return factory.updateIndexedAccessTypeNode(node, nodeVisitor(node.objectType, visitor, ts.isTypeNode), nodeVisitor(node.indexType, visitor, ts.isTypeNode));
                case 193 /* MappedType */:
                    ts.Debug.type(node);
                    return factory.updateMappedTypeNode(node, nodeVisitor(node.readonlyToken, tokenVisitor, ts.isReadonlyKeywordOrPlusOrMinusToken), nodeVisitor(node.typeParameter, visitor, ts.isTypeParameterDeclaration), nodeVisitor(node.nameType, visitor, ts.isTypeNode), nodeVisitor(node.questionToken, tokenVisitor, ts.isQuestionOrPlusOrMinusToken), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 194 /* LiteralType */:
                    ts.Debug.type(node);
                    return factory.updateLiteralTypeNode(node, nodeVisitor(node.literal, visitor, ts.isExpression));
                case 196 /* TemplateLiteralType */:
                    ts.Debug.type(node);
                    return factory.updateTemplateLiteralType(node, nodeVisitor(node.head, visitor, ts.isTemplateHead), nodesVisitor(node.templateSpans, visitor, ts.isTemplateLiteralTypeSpan));
                case 197 /* TemplateLiteralTypeSpan */:
                    ts.Debug.type(node);
                    return factory.updateTemplateLiteralTypeSpan(node, nodeVisitor(node.type, visitor, ts.isTypeNode), nodeVisitor(node.literal, visitor, ts.isTemplateMiddleOrTemplateTail));
                // Binding patterns
                case 199 /* ObjectBindingPattern */:
                    ts.Debug.type(node);
                    return factory.updateObjectBindingPattern(node, nodesVisitor(node.elements, visitor, ts.isBindingElement));
                case 200 /* ArrayBindingPattern */:
                    ts.Debug.type(node);
                    return factory.updateArrayBindingPattern(node, nodesVisitor(node.elements, visitor, ts.isArrayBindingElement));
                case 201 /* BindingElement */:
                    ts.Debug.type(node);
                    return factory.updateBindingElement(node, nodeVisitor(node.dotDotDotToken, tokenVisitor, ts.isDotDotDotToken), nodeVisitor(node.propertyName, visitor, ts.isPropertyName), nodeVisitor(node.name, visitor, ts.isBindingName), nodeVisitor(node.initializer, visitor, ts.isExpression));
                // Expression
                case 202 /* ArrayLiteralExpression */:
                    ts.Debug.type(node);
                    return factory.updateArrayLiteralExpression(node, nodesVisitor(node.elements, visitor, ts.isExpression));
                case 203 /* ObjectLiteralExpression */:
                    ts.Debug.type(node);
                    return factory.updateObjectLiteralExpression(node, nodesVisitor(node.properties, visitor, ts.isObjectLiteralElementLike));
                case 204 /* PropertyAccessExpression */:
                    if (node.flags & 32 /* OptionalChain */) {
                        ts.Debug.type(node);
                        return factory.updatePropertyAccessChain(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.questionDotToken, tokenVisitor, ts.isQuestionDotToken), nodeVisitor(node.name, visitor, ts.isMemberName));
                    }
                    ts.Debug.type(node);
                    return factory.updatePropertyAccessExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.name, visitor, ts.isMemberName));
                case 205 /* ElementAccessExpression */:
                    if (node.flags & 32 /* OptionalChain */) {
                        ts.Debug.type(node);
                        return factory.updateElementAccessChain(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.questionDotToken, tokenVisitor, ts.isQuestionDotToken), nodeVisitor(node.argumentExpression, visitor, ts.isExpression));
                    }
                    ts.Debug.type(node);
                    return factory.updateElementAccessExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.argumentExpression, visitor, ts.isExpression));
                case 206 /* CallExpression */:
                    if (node.flags & 32 /* OptionalChain */) {
                        ts.Debug.type(node);
                        return factory.updateCallChain(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.questionDotToken, tokenVisitor, ts.isQuestionDotToken), nodesVisitor(node.typeArguments, visitor, ts.isTypeNode), nodesVisitor(node.arguments, visitor, ts.isExpression));
                    }
                    ts.Debug.type(node);
                    return factory.updateCallExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodesVisitor(node.typeArguments, visitor, ts.isTypeNode), nodesVisitor(node.arguments, visitor, ts.isExpression));
                case 207 /* NewExpression */:
                    ts.Debug.type(node);
                    return factory.updateNewExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodesVisitor(node.typeArguments, visitor, ts.isTypeNode), nodesVisitor(node.arguments, visitor, ts.isExpression));
                case 208 /* TaggedTemplateExpression */:
                    ts.Debug.type(node);
                    return factory.updateTaggedTemplateExpression(node, nodeVisitor(node.tag, visitor, ts.isExpression), visitNodes(node.typeArguments, visitor, ts.isTypeNode), nodeVisitor(node.template, visitor, ts.isTemplateLiteral));
                case 209 /* TypeAssertionExpression */:
                    ts.Debug.type(node);
                    return factory.updateTypeAssertion(node, nodeVisitor(node.type, visitor, ts.isTypeNode), nodeVisitor(node.expression, visitor, ts.isExpression));
                case 210 /* ParenthesizedExpression */:
                    ts.Debug.type(node);
                    return factory.updateParenthesizedExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 211 /* FunctionExpression */:
                    ts.Debug.type(node);
                    return factory.updateFunctionExpression(node, nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.asteriskToken, tokenVisitor, ts.isAsteriskToken), nodeVisitor(node.name, visitor, ts.isIdentifier), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), visitParameterList(node.parameters, visitor, context, nodesVisitor), nodeVisitor(node.type, visitor, ts.isTypeNode), visitFunctionBody(node.body, visitor, context, nodeVisitor));
                case 212 /* ArrowFunction */:
                    ts.Debug.type(node);
                    return factory.updateArrowFunction(node, nodesVisitor(node.modifiers, visitor, ts.isModifier), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), visitParameterList(node.parameters, visitor, context, nodesVisitor), nodeVisitor(node.type, visitor, ts.isTypeNode), nodeVisitor(node.equalsGreaterThanToken, tokenVisitor, ts.isEqualsGreaterThanToken), visitFunctionBody(node.body, visitor, context, nodeVisitor));
                case 213 /* DeleteExpression */:
                    ts.Debug.type(node);
                    return factory.updateDeleteExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 214 /* TypeOfExpression */:
                    ts.Debug.type(node);
                    return factory.updateTypeOfExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 215 /* VoidExpression */:
                    ts.Debug.type(node);
                    return factory.updateVoidExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 216 /* AwaitExpression */:
                    ts.Debug.type(node);
                    return factory.updateAwaitExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 217 /* PrefixUnaryExpression */:
                    ts.Debug.type(node);
                    return factory.updatePrefixUnaryExpression(node, nodeVisitor(node.operand, visitor, ts.isExpression));
                case 218 /* PostfixUnaryExpression */:
                    ts.Debug.type(node);
                    return factory.updatePostfixUnaryExpression(node, nodeVisitor(node.operand, visitor, ts.isExpression));
                case 219 /* BinaryExpression */:
                    ts.Debug.type(node);
                    return factory.updateBinaryExpression(node, nodeVisitor(node.left, visitor, ts.isExpression), nodeVisitor(node.operatorToken, tokenVisitor, ts.isBinaryOperatorToken), nodeVisitor(node.right, visitor, ts.isExpression));
                case 220 /* ConditionalExpression */:
                    ts.Debug.type(node);
                    return factory.updateConditionalExpression(node, nodeVisitor(node.condition, visitor, ts.isExpression), nodeVisitor(node.questionToken, tokenVisitor, ts.isQuestionToken), nodeVisitor(node.whenTrue, visitor, ts.isExpression), nodeVisitor(node.colonToken, tokenVisitor, ts.isColonToken), nodeVisitor(node.whenFalse, visitor, ts.isExpression));
                case 221 /* TemplateExpression */:
                    ts.Debug.type(node);
                    return factory.updateTemplateExpression(node, nodeVisitor(node.head, visitor, ts.isTemplateHead), nodesVisitor(node.templateSpans, visitor, ts.isTemplateSpan));
                case 222 /* YieldExpression */:
                    ts.Debug.type(node);
                    return factory.updateYieldExpression(node, nodeVisitor(node.asteriskToken, tokenVisitor, ts.isAsteriskToken), nodeVisitor(node.expression, visitor, ts.isExpression));
                case 223 /* SpreadElement */:
                    ts.Debug.type(node);
                    return factory.updateSpreadElement(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 224 /* ClassExpression */:
                    ts.Debug.type(node);
                    return factory.updateClassExpression(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isIdentifier), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), nodesVisitor(node.heritageClauses, visitor, ts.isHeritageClause), nodesVisitor(node.members, visitor, ts.isClassElement));
                case 226 /* ExpressionWithTypeArguments */:
                    ts.Debug.type(node);
                    return factory.updateExpressionWithTypeArguments(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodesVisitor(node.typeArguments, visitor, ts.isTypeNode));
                case 227 /* AsExpression */:
                    ts.Debug.type(node);
                    return factory.updateAsExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 228 /* NonNullExpression */:
                    if (node.flags & 32 /* OptionalChain */) {
                        ts.Debug.type(node);
                        return factory.updateNonNullChain(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                    }
                    ts.Debug.type(node);
                    return factory.updateNonNullExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 229 /* MetaProperty */:
                    ts.Debug.type(node);
                    return factory.updateMetaProperty(node, nodeVisitor(node.name, visitor, ts.isIdentifier));
                // Misc
                case 231 /* TemplateSpan */:
                    ts.Debug.type(node);
                    return factory.updateTemplateSpan(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.literal, visitor, ts.isTemplateMiddleOrTemplateTail));
                // Element
                case 233 /* Block */:
                    ts.Debug.type(node);
                    return factory.updateBlock(node, nodesVisitor(node.statements, visitor, ts.isStatement));
                case 235 /* VariableStatement */:
                    ts.Debug.type(node);
                    return factory.updateVariableStatement(node, nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.declarationList, visitor, ts.isVariableDeclarationList));
                case 236 /* ExpressionStatement */:
                    ts.Debug.type(node);
                    return factory.updateExpressionStatement(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 237 /* IfStatement */:
                    ts.Debug.type(node);
                    return factory.updateIfStatement(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.thenStatement, visitor, ts.isStatement, factory.liftToBlock), nodeVisitor(node.elseStatement, visitor, ts.isStatement, factory.liftToBlock));
                case 238 /* DoStatement */:
                    ts.Debug.type(node);
                    return factory.updateDoStatement(node, visitIterationBody(node.statement, visitor, context), nodeVisitor(node.expression, visitor, ts.isExpression));
                case 239 /* WhileStatement */:
                    ts.Debug.type(node);
                    return factory.updateWhileStatement(node, nodeVisitor(node.expression, visitor, ts.isExpression), visitIterationBody(node.statement, visitor, context));
                case 240 /* ForStatement */:
                    ts.Debug.type(node);
                    return factory.updateForStatement(node, nodeVisitor(node.initializer, visitor, ts.isForInitializer), nodeVisitor(node.condition, visitor, ts.isExpression), nodeVisitor(node.incrementor, visitor, ts.isExpression), visitIterationBody(node.statement, visitor, context));
                case 241 /* ForInStatement */:
                    ts.Debug.type(node);
                    return factory.updateForInStatement(node, nodeVisitor(node.initializer, visitor, ts.isForInitializer), nodeVisitor(node.expression, visitor, ts.isExpression), visitIterationBody(node.statement, visitor, context));
                case 242 /* ForOfStatement */:
                    ts.Debug.type(node);
                    return factory.updateForOfStatement(node, nodeVisitor(node.awaitModifier, tokenVisitor, ts.isAwaitKeyword), nodeVisitor(node.initializer, visitor, ts.isForInitializer), nodeVisitor(node.expression, visitor, ts.isExpression), visitIterationBody(node.statement, visitor, context));
                case 243 /* ContinueStatement */:
                    ts.Debug.type(node);
                    return factory.updateContinueStatement(node, nodeVisitor(node.label, visitor, ts.isIdentifier));
                case 244 /* BreakStatement */:
                    ts.Debug.type(node);
                    return factory.updateBreakStatement(node, nodeVisitor(node.label, visitor, ts.isIdentifier));
                case 245 /* ReturnStatement */:
                    ts.Debug.type(node);
                    return factory.updateReturnStatement(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 246 /* WithStatement */:
                    ts.Debug.type(node);
                    return factory.updateWithStatement(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.statement, visitor, ts.isStatement, factory.liftToBlock));
                case 247 /* SwitchStatement */:
                    ts.Debug.type(node);
                    return factory.updateSwitchStatement(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodeVisitor(node.caseBlock, visitor, ts.isCaseBlock));
                case 248 /* LabeledStatement */:
                    ts.Debug.type(node);
                    return factory.updateLabeledStatement(node, nodeVisitor(node.label, visitor, ts.isIdentifier), nodeVisitor(node.statement, visitor, ts.isStatement, factory.liftToBlock));
                case 249 /* ThrowStatement */:
                    ts.Debug.type(node);
                    return factory.updateThrowStatement(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 250 /* TryStatement */:
                    ts.Debug.type(node);
                    return factory.updateTryStatement(node, nodeVisitor(node.tryBlock, visitor, ts.isBlock), nodeVisitor(node.catchClause, visitor, ts.isCatchClause), nodeVisitor(node.finallyBlock, visitor, ts.isBlock));
                case 252 /* VariableDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateVariableDeclaration(node, nodeVisitor(node.name, visitor, ts.isBindingName), nodeVisitor(node.exclamationToken, tokenVisitor, ts.isExclamationToken), nodeVisitor(node.type, visitor, ts.isTypeNode), nodeVisitor(node.initializer, visitor, ts.isExpression));
                case 253 /* VariableDeclarationList */:
                    ts.Debug.type(node);
                    return factory.updateVariableDeclarationList(node, nodesVisitor(node.declarations, visitor, ts.isVariableDeclaration));
                case 254 /* FunctionDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateFunctionDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.asteriskToken, tokenVisitor, ts.isAsteriskToken), nodeVisitor(node.name, visitor, ts.isIdentifier), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), visitParameterList(node.parameters, visitor, context, nodesVisitor), nodeVisitor(node.type, visitor, ts.isTypeNode), visitFunctionBody(node.body, visitor, context, nodeVisitor));
                case 255 /* ClassDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateClassDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isIdentifier), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), nodesVisitor(node.heritageClauses, visitor, ts.isHeritageClause), nodesVisitor(node.members, visitor, ts.isClassElement));
                case 256 /* InterfaceDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateInterfaceDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isIdentifier), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), nodesVisitor(node.heritageClauses, visitor, ts.isHeritageClause), nodesVisitor(node.members, visitor, ts.isTypeElement));
                case 257 /* TypeAliasDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateTypeAliasDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isIdentifier), nodesVisitor(node.typeParameters, visitor, ts.isTypeParameterDeclaration), nodeVisitor(node.type, visitor, ts.isTypeNode));
                case 258 /* EnumDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateEnumDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isIdentifier), nodesVisitor(node.members, visitor, ts.isEnumMember));
                case 259 /* ModuleDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateModuleDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.name, visitor, ts.isModuleName), nodeVisitor(node.body, visitor, ts.isModuleBody));
                case 260 /* ModuleBlock */:
                    ts.Debug.type(node);
                    return factory.updateModuleBlock(node, nodesVisitor(node.statements, visitor, ts.isStatement));
                case 261 /* CaseBlock */:
                    ts.Debug.type(node);
                    return factory.updateCaseBlock(node, nodesVisitor(node.clauses, visitor, ts.isCaseOrDefaultClause));
                case 262 /* NamespaceExportDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateNamespaceExportDeclaration(node, nodeVisitor(node.name, visitor, ts.isIdentifier));
                case 263 /* ImportEqualsDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateImportEqualsDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), node.isTypeOnly, nodeVisitor(node.name, visitor, ts.isIdentifier), nodeVisitor(node.moduleReference, visitor, ts.isModuleReference));
                case 264 /* ImportDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateImportDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.importClause, visitor, ts.isImportClause), nodeVisitor(node.moduleSpecifier, visitor, ts.isExpression));
                case 265 /* ImportClause */:
                    ts.Debug.type(node);
                    return factory.updateImportClause(node, node.isTypeOnly, nodeVisitor(node.name, visitor, ts.isIdentifier), nodeVisitor(node.namedBindings, visitor, ts.isNamedImportBindings));
                case 266 /* NamespaceImport */:
                    ts.Debug.type(node);
                    return factory.updateNamespaceImport(node, nodeVisitor(node.name, visitor, ts.isIdentifier));
                case 272 /* NamespaceExport */:
                    ts.Debug.type(node);
                    return factory.updateNamespaceExport(node, nodeVisitor(node.name, visitor, ts.isIdentifier));
                case 267 /* NamedImports */:
                    ts.Debug.type(node);
                    return factory.updateNamedImports(node, nodesVisitor(node.elements, visitor, ts.isImportSpecifier));
                case 268 /* ImportSpecifier */:
                    ts.Debug.type(node);
                    return factory.updateImportSpecifier(node, nodeVisitor(node.propertyName, visitor, ts.isIdentifier), nodeVisitor(node.name, visitor, ts.isIdentifier));
                case 269 /* ExportAssignment */:
                    ts.Debug.type(node);
                    return factory.updateExportAssignment(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), nodeVisitor(node.expression, visitor, ts.isExpression));
                case 270 /* ExportDeclaration */:
                    ts.Debug.type(node);
                    return factory.updateExportDeclaration(node, nodesVisitor(node.decorators, visitor, ts.isDecorator), nodesVisitor(node.modifiers, visitor, ts.isModifier), node.isTypeOnly, nodeVisitor(node.exportClause, visitor, ts.isNamedExportBindings), nodeVisitor(node.moduleSpecifier, visitor, ts.isExpression));
                case 271 /* NamedExports */:
                    ts.Debug.type(node);
                    return factory.updateNamedExports(node, nodesVisitor(node.elements, visitor, ts.isExportSpecifier));
                case 273 /* ExportSpecifier */:
                    ts.Debug.type(node);
                    return factory.updateExportSpecifier(node, nodeVisitor(node.propertyName, visitor, ts.isIdentifier), nodeVisitor(node.name, visitor, ts.isIdentifier));
                // Module references
                case 275 /* ExternalModuleReference */:
                    ts.Debug.type(node);
                    return factory.updateExternalModuleReference(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                // JSX
                case 276 /* JsxElement */:
                    ts.Debug.type(node);
                    return factory.updateJsxElement(node, nodeVisitor(node.openingElement, visitor, ts.isJsxOpeningElement), nodesVisitor(node.children, visitor, ts.isJsxChild), nodeVisitor(node.closingElement, visitor, ts.isJsxClosingElement));
                case 277 /* JsxSelfClosingElement */:
                    ts.Debug.type(node);
                    return factory.updateJsxSelfClosingElement(node, nodeVisitor(node.tagName, visitor, ts.isJsxTagNameExpression), nodesVisitor(node.typeArguments, visitor, ts.isTypeNode), nodeVisitor(node.attributes, visitor, ts.isJsxAttributes));
                case 278 /* JsxOpeningElement */:
                    ts.Debug.type(node);
                    return factory.updateJsxOpeningElement(node, nodeVisitor(node.tagName, visitor, ts.isJsxTagNameExpression), nodesVisitor(node.typeArguments, visitor, ts.isTypeNode), nodeVisitor(node.attributes, visitor, ts.isJsxAttributes));
                case 279 /* JsxClosingElement */:
                    ts.Debug.type(node);
                    return factory.updateJsxClosingElement(node, nodeVisitor(node.tagName, visitor, ts.isJsxTagNameExpression));
                case 280 /* JsxFragment */:
                    ts.Debug.type(node);
                    return factory.updateJsxFragment(node, nodeVisitor(node.openingFragment, visitor, ts.isJsxOpeningFragment), nodesVisitor(node.children, visitor, ts.isJsxChild), nodeVisitor(node.closingFragment, visitor, ts.isJsxClosingFragment));
                case 283 /* JsxAttribute */:
                    ts.Debug.type(node);
                    return factory.updateJsxAttribute(node, nodeVisitor(node.name, visitor, ts.isIdentifier), nodeVisitor(node.initializer, visitor, ts.isStringLiteralOrJsxExpression));
                case 284 /* JsxAttributes */:
                    ts.Debug.type(node);
                    return factory.updateJsxAttributes(node, nodesVisitor(node.properties, visitor, ts.isJsxAttributeLike));
                case 285 /* JsxSpreadAttribute */:
                    ts.Debug.type(node);
                    return factory.updateJsxSpreadAttribute(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 286 /* JsxExpression */:
                    ts.Debug.type(node);
                    return factory.updateJsxExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                // Clauses
                case 287 /* CaseClause */:
                    ts.Debug.type(node);
                    return factory.updateCaseClause(node, nodeVisitor(node.expression, visitor, ts.isExpression), nodesVisitor(node.statements, visitor, ts.isStatement));
                case 288 /* DefaultClause */:
                    ts.Debug.type(node);
                    return factory.updateDefaultClause(node, nodesVisitor(node.statements, visitor, ts.isStatement));
                case 289 /* HeritageClause */:
                    ts.Debug.type(node);
                    return factory.updateHeritageClause(node, nodesVisitor(node.types, visitor, ts.isExpressionWithTypeArguments));
                case 290 /* CatchClause */:
                    ts.Debug.type(node);
                    return factory.updateCatchClause(node, nodeVisitor(node.variableDeclaration, visitor, ts.isVariableDeclaration), nodeVisitor(node.block, visitor, ts.isBlock));
                // Property assignments
                case 291 /* PropertyAssignment */:
                    ts.Debug.type(node);
                    return factory.updatePropertyAssignment(node, nodeVisitor(node.name, visitor, ts.isPropertyName), nodeVisitor(node.initializer, visitor, ts.isExpression));
                case 292 /* ShorthandPropertyAssignment */:
                    ts.Debug.type(node);
                    return factory.updateShorthandPropertyAssignment(node, nodeVisitor(node.name, visitor, ts.isIdentifier), nodeVisitor(node.objectAssignmentInitializer, visitor, ts.isExpression));
                case 293 /* SpreadAssignment */:
                    ts.Debug.type(node);
                    return factory.updateSpreadAssignment(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                // Enum
                case 294 /* EnumMember */:
                    ts.Debug.type(node);
                    return factory.updateEnumMember(node, nodeVisitor(node.name, visitor, ts.isPropertyName), nodeVisitor(node.initializer, visitor, ts.isExpression));
                // Top-level nodes
                case 300 /* SourceFile */:
                    ts.Debug.type(node);
                    return factory.updateSourceFile(node, visitLexicalEnvironment(node.statements, visitor, context));
                // Transformation nodes
                case 345 /* PartiallyEmittedExpression */:
                    ts.Debug.type(node);
                    return factory.updatePartiallyEmittedExpression(node, nodeVisitor(node.expression, visitor, ts.isExpression));
                case 346 /* CommaListExpression */:
                    ts.Debug.type(node);
                    return factory.updateCommaListExpression(node, nodesVisitor(node.elements, visitor, ts.isExpression));
                default:
                    // No need to visit nodes with no children.
                    return node;
            }
        }
        ts.visitEachChild = visitEachChild;
        /**
         * Extracts the single node from a NodeArray.
         *
         * @param nodes The NodeArray.
         */
        function extractSingleNode(nodes) {
            ts.Debug.assert(nodes.length <= 1, "Too many nodes written to output.");
            return ts.singleOrUndefined(nodes);
        }
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function createSourceMapGenerator(host, file, sourceRoot, sourcesDirectoryPath, generatorOptions) {
            var _a = generatorOptions.extendedDiagnostics
                ? ts.performance.createTimer("Source Map", "beforeSourcemap", "afterSourcemap")
                : ts.performance.nullTimer, enter = _a.enter, exit = _a.exit;
            // Current source map file and its index in the sources list
            var rawSources = [];
            var sources = [];
            var sourceToSourceIndexMap = new ts.Map();
            var sourcesContent;
            var names = [];
            var nameToNameIndexMap;
            var mappingCharCodes = [];
            var mappings = "";
            // Last recorded and encoded mappings
            var lastGeneratedLine = 0;
            var lastGeneratedCharacter = 0;
            var lastSourceIndex = 0;
            var lastSourceLine = 0;
            var lastSourceCharacter = 0;
            var lastNameIndex = 0;
            var hasLast = false;
            var pendingGeneratedLine = 0;
            var pendingGeneratedCharacter = 0;
            var pendingSourceIndex = 0;
            var pendingSourceLine = 0;
            var pendingSourceCharacter = 0;
            var pendingNameIndex = 0;
            var hasPending = false;
            var hasPendingSource = false;
            var hasPendingName = false;
            return {
                getSources: function () { return rawSources; },
                addSource: addSource,
                setSourceContent: setSourceContent,
                addName: addName,
                addMapping: addMapping,
                appendSourceMap: appendSourceMap,
                toJSON: toJSON,
                toString: function () { return JSON.stringify(toJSON()); }
            };
            function addSource(fileName) {
                enter();
                var source = ts.getRelativePathToDirectoryOrUrl(sourcesDirectoryPath, fileName, host.getCurrentDirectory(), host.getCanonicalFileName, 
                /*isAbsolutePathAnUrl*/ true);
                var sourceIndex = sourceToSourceIndexMap.get(source);
                if (sourceIndex === undefined) {
                    sourceIndex = sources.length;
                    sources.push(source);
                    rawSources.push(fileName);
                    sourceToSourceIndexMap.set(source, sourceIndex);
                }
                exit();
                return sourceIndex;
            }
            /* eslint-disable boolean-trivia, no-null/no-null */
            function setSourceContent(sourceIndex, content) {
                enter();
                if (content !== null) {
                    if (!sourcesContent)
                        sourcesContent = [];
                    while (sourcesContent.length < sourceIndex) {
                        sourcesContent.push(null);
                    }
                    sourcesContent[sourceIndex] = content;
                }
                exit();
            }
            /* eslint-enable boolean-trivia, no-null/no-null */
            function addName(name) {
                enter();
                if (!nameToNameIndexMap)
                    nameToNameIndexMap = new ts.Map();
                var nameIndex = nameToNameIndexMap.get(name);
                if (nameIndex === undefined) {
                    nameIndex = names.length;
                    names.push(name);
                    nameToNameIndexMap.set(name, nameIndex);
                }
                exit();
                return nameIndex;
            }
            function isNewGeneratedPosition(generatedLine, generatedCharacter) {
                return !hasPending
                    || pendingGeneratedLine !== generatedLine
                    || pendingGeneratedCharacter !== generatedCharacter;
            }
            function isBacktrackingSourcePosition(sourceIndex, sourceLine, sourceCharacter) {
                return sourceIndex !== undefined
                    && sourceLine !== undefined
                    && sourceCharacter !== undefined
                    && pendingSourceIndex === sourceIndex
                    && (pendingSourceLine > sourceLine
                        || pendingSourceLine === sourceLine && pendingSourceCharacter > sourceCharacter);
            }
            function addMapping(generatedLine, generatedCharacter, sourceIndex, sourceLine, sourceCharacter, nameIndex) {
                ts.Debug.assert(generatedLine >= pendingGeneratedLine, "generatedLine cannot backtrack");
                ts.Debug.assert(generatedCharacter >= 0, "generatedCharacter cannot be negative");
                ts.Debug.assert(sourceIndex === undefined || sourceIndex >= 0, "sourceIndex cannot be negative");
                ts.Debug.assert(sourceLine === undefined || sourceLine >= 0, "sourceLine cannot be negative");
                ts.Debug.assert(sourceCharacter === undefined || sourceCharacter >= 0, "sourceCharacter cannot be negative");
                enter();
                // If this location wasn't recorded or the location in source is going backwards, record the mapping
                if (isNewGeneratedPosition(generatedLine, generatedCharacter) ||
                    isBacktrackingSourcePosition(sourceIndex, sourceLine, sourceCharacter)) {
                    commitPendingMapping();
                    pendingGeneratedLine = generatedLine;
                    pendingGeneratedCharacter = generatedCharacter;
                    hasPendingSource = false;
                    hasPendingName = false;
                    hasPending = true;
                }
                if (sourceIndex !== undefined && sourceLine !== undefined && sourceCharacter !== undefined) {
                    pendingSourceIndex = sourceIndex;
                    pendingSourceLine = sourceLine;
                    pendingSourceCharacter = sourceCharacter;
                    hasPendingSource = true;
                    if (nameIndex !== undefined) {
                        pendingNameIndex = nameIndex;
                        hasPendingName = true;
                    }
                }
                exit();
            }
            function appendSourceMap(generatedLine, generatedCharacter, map, sourceMapPath, start, end) {
                ts.Debug.assert(generatedLine >= pendingGeneratedLine, "generatedLine cannot backtrack");
                ts.Debug.assert(generatedCharacter >= 0, "generatedCharacter cannot be negative");
                enter();
                // First, decode the old component sourcemap
                var sourceIndexToNewSourceIndexMap = [];
                var nameIndexToNewNameIndexMap;
                var mappingIterator = decodeMappings(map.mappings);
                for (var iterResult = mappingIterator.next(); !iterResult.done; iterResult = mappingIterator.next()) {
                    var raw = iterResult.value;
                    if (end && (raw.generatedLine > end.line ||
                        (raw.generatedLine === end.line && raw.generatedCharacter > end.character))) {
                        break;
                    }
                    if (start && (raw.generatedLine < start.line ||
                        (start.line === raw.generatedLine && raw.generatedCharacter < start.character))) {
                        continue;
                    }
                    // Then reencode all the updated mappings into the overall map
                    var newSourceIndex = void 0;
                    var newSourceLine = void 0;
                    var newSourceCharacter = void 0;
                    var newNameIndex = void 0;
                    if (raw.sourceIndex !== undefined) {
                        newSourceIndex = sourceIndexToNewSourceIndexMap[raw.sourceIndex];
                        if (newSourceIndex === undefined) {
                            // Apply offsets to each position and fixup source entries
                            var rawPath = map.sources[raw.sourceIndex];
                            var relativePath = map.sourceRoot ? ts.combinePaths(map.sourceRoot, rawPath) : rawPath;
                            var combinedPath = ts.combinePaths(ts.getDirectoryPath(sourceMapPath), relativePath);
                            sourceIndexToNewSourceIndexMap[raw.sourceIndex] = newSourceIndex = addSource(combinedPath);
                            if (map.sourcesContent && typeof map.sourcesContent[raw.sourceIndex] === "string") {
                                setSourceContent(newSourceIndex, map.sourcesContent[raw.sourceIndex]);
                            }
                        }
                        newSourceLine = raw.sourceLine;
                        newSourceCharacter = raw.sourceCharacter;
                        if (map.names && raw.nameIndex !== undefined) {
                            if (!nameIndexToNewNameIndexMap)
                                nameIndexToNewNameIndexMap = [];
                            newNameIndex = nameIndexToNewNameIndexMap[raw.nameIndex];
                            if (newNameIndex === undefined) {
                                nameIndexToNewNameIndexMap[raw.nameIndex] = newNameIndex = addName(map.names[raw.nameIndex]);
                            }
                        }
                    }
                    var rawGeneratedLine = raw.generatedLine - (start ? start.line : 0);
                    var newGeneratedLine = rawGeneratedLine + generatedLine;
                    var rawGeneratedCharacter = start && start.line === raw.generatedLine ? raw.generatedCharacter - start.character : raw.generatedCharacter;
                    var newGeneratedCharacter = rawGeneratedLine === 0 ? rawGeneratedCharacter + generatedCharacter : rawGeneratedCharacter;
                    addMapping(newGeneratedLine, newGeneratedCharacter, newSourceIndex, newSourceLine, newSourceCharacter, newNameIndex);
                }
                exit();
            }
            function shouldCommitMapping() {
                return !hasLast
                    || lastGeneratedLine !== pendingGeneratedLine
                    || lastGeneratedCharacter !== pendingGeneratedCharacter
                    || lastSourceIndex !== pendingSourceIndex
                    || lastSourceLine !== pendingSourceLine
                    || lastSourceCharacter !== pendingSourceCharacter
                    || lastNameIndex !== pendingNameIndex;
            }
            function appendMappingCharCode(charCode) {
                mappingCharCodes.push(charCode);
                // String.fromCharCode accepts its arguments on the stack, so we have to chunk the input,
                // otherwise we can get stack overflows for large source maps
                if (mappingCharCodes.length >= 1024) {
                    flushMappingBuffer();
                }
            }
            function commitPendingMapping() {
                if (!hasPending || !shouldCommitMapping()) {
                    return;
                }
                enter();
                // Line/Comma delimiters
                if (lastGeneratedLine < pendingGeneratedLine) {
                    // Emit line delimiters
                    do {
                        appendMappingCharCode(59 /* semicolon */);
                        lastGeneratedLine++;
                    } while (lastGeneratedLine < pendingGeneratedLine);
                    // Only need to set this once
                    lastGeneratedCharacter = 0;
                }
                else {
                    ts.Debug.assertEqual(lastGeneratedLine, pendingGeneratedLine, "generatedLine cannot backtrack");
                    // Emit comma to separate the entry
                    if (hasLast) {
                        appendMappingCharCode(44 /* comma */);
                    }
                }
                // 1. Relative generated character
                appendBase64VLQ(pendingGeneratedCharacter - lastGeneratedCharacter);
                lastGeneratedCharacter = pendingGeneratedCharacter;
                if (hasPendingSource) {
                    // 2. Relative sourceIndex
                    appendBase64VLQ(pendingSourceIndex - lastSourceIndex);
                    lastSourceIndex = pendingSourceIndex;
                    // 3. Relative source line
                    appendBase64VLQ(pendingSourceLine - lastSourceLine);
                    lastSourceLine = pendingSourceLine;
                    // 4. Relative source character
                    appendBase64VLQ(pendingSourceCharacter - lastSourceCharacter);
                    lastSourceCharacter = pendingSourceCharacter;
                    if (hasPendingName) {
                        // 5. Relative nameIndex
                        appendBase64VLQ(pendingNameIndex - lastNameIndex);
                        lastNameIndex = pendingNameIndex;
                    }
                }
                hasLast = true;
                exit();
            }
            function flushMappingBuffer() {
                if (mappingCharCodes.length > 0) {
                    mappings += String.fromCharCode.apply(undefined, mappingCharCodes);
                    mappingCharCodes.length = 0;
                }
            }
            function toJSON() {
                commitPendingMapping();
                flushMappingBuffer();
                return {
                    version: 3,
                    file: file,
                    sourceRoot: sourceRoot,
                    sources: sources,
                    names: names,
                    mappings: mappings,
                    sourcesContent: sourcesContent,
                };
            }
            function appendBase64VLQ(inValue) {
                // Add a new least significant bit that has the sign of the value.
                // if negative number the least significant bit that gets added to the number has value 1
                // else least significant bit value that gets added is 0
                // eg. -1 changes to binary : 01 [1] => 3
                //     +1 changes to binary : 01 [0] => 2
                if (inValue < 0) {
                    inValue = ((-inValue) << 1) + 1;
                }
                else {
                    inValue = inValue << 1;
                }
                // Encode 5 bits at a time starting from least significant bits
                do {
                    var currentDigit = inValue & 31; // 11111
                    inValue = inValue >> 5;
                    if (inValue > 0) {
                        // There are still more digits to decode, set the msb (6th bit)
                        currentDigit = currentDigit | 32;
                    }
                    appendMappingCharCode(base64FormatEncode(currentDigit));
                } while (inValue > 0);
            }
        }
        ts.createSourceMapGenerator = createSourceMapGenerator;
        // Sometimes tools can see the following line as a source mapping url comment, so we mangle it a bit (the [M])
        var sourceMapCommentRegExp = /^\/\/[@#] source[M]appingURL=(.+)$/;
        var whitespaceOrMapCommentRegExp = /^\s*(\/\/[@#] .*)?$/;
        function getLineInfo(text, lineStarts) {
            return {
                getLineCount: function () { return lineStarts.length; },
                getLineText: function (line) { return text.substring(lineStarts[line], lineStarts[line + 1]); }
            };
        }
        ts.getLineInfo = getLineInfo;
        /**
         * Tries to find the sourceMappingURL comment at the end of a file.
         */
        function tryGetSourceMappingURL(lineInfo) {
            for (var index = lineInfo.getLineCount() - 1; index >= 0; index--) {
                var line = lineInfo.getLineText(index);
                var comment = sourceMapCommentRegExp.exec(line);
                if (comment) {
                    return ts.trimStringEnd(comment[1]);
                }
                // If we see a non-whitespace/map comment-like line, break, to avoid scanning up the entire file
                else if (!line.match(whitespaceOrMapCommentRegExp)) {
                    break;
                }
            }
        }
        ts.tryGetSourceMappingURL = tryGetSourceMappingURL;
        /* eslint-disable no-null/no-null */
        function isStringOrNull(x) {
            return typeof x === "string" || x === null;
        }
        function isRawSourceMap(x) {
            return x !== null
                && typeof x === "object"
                && x.version === 3
                && typeof x.file === "string"
                && typeof x.mappings === "string"
                && ts.isArray(x.sources) && ts.every(x.sources, ts.isString)
                && (x.sourceRoot === undefined || x.sourceRoot === null || typeof x.sourceRoot === "string")
                && (x.sourcesContent === undefined || x.sourcesContent === null || ts.isArray(x.sourcesContent) && ts.every(x.sourcesContent, isStringOrNull))
                && (x.names === undefined || x.names === null || ts.isArray(x.names) && ts.every(x.names, ts.isString));
        }
        ts.isRawSourceMap = isRawSourceMap;
        /* eslint-enable no-null/no-null */
        function tryParseRawSourceMap(text) {
            try {
                var parsed = JSON.parse(text);
                if (isRawSourceMap(parsed)) {
                    return parsed;
                }
            }
            catch (_a) {
                // empty
            }
            return undefined;
        }
        ts.tryParseRawSourceMap = tryParseRawSourceMap;
        function decodeMappings(mappings) {
            var done = false;
            var pos = 0;
            var generatedLine = 0;
            var generatedCharacter = 0;
            var sourceIndex = 0;
            var sourceLine = 0;
            var sourceCharacter = 0;
            var nameIndex = 0;
            var error;
            return {
                get pos() { return pos; },
                get error() { return error; },
                get state() { return captureMapping(/*hasSource*/ true, /*hasName*/ true); },
                next: function () {
                    while (!done && pos < mappings.length) {
                        var ch = mappings.charCodeAt(pos);
                        if (ch === 59 /* semicolon */) {
                            // new line
                            generatedLine++;
                            generatedCharacter = 0;
                            pos++;
                            continue;
                        }
                        if (ch === 44 /* comma */) {
                            // Next entry is on same line - no action needed
                            pos++;
                            continue;
                        }
                        var hasSource = false;
                        var hasName = false;
                        generatedCharacter += base64VLQFormatDecode();
                        if (hasReportedError())
                            return stopIterating();
                        if (generatedCharacter < 0)
                            return setErrorAndStopIterating("Invalid generatedCharacter found");
                        if (!isSourceMappingSegmentEnd()) {
                            hasSource = true;
                            sourceIndex += base64VLQFormatDecode();
                            if (hasReportedError())
                                return stopIterating();
                            if (sourceIndex < 0)
                                return setErrorAndStopIterating("Invalid sourceIndex found");
                            if (isSourceMappingSegmentEnd())
                                return setErrorAndStopIterating("Unsupported Format: No entries after sourceIndex");
                            sourceLine += base64VLQFormatDecode();
                            if (hasReportedError())
                                return stopIterating();
                            if (sourceLine < 0)
                                return setErrorAndStopIterating("Invalid sourceLine found");
                            if (isSourceMappingSegmentEnd())
                                return setErrorAndStopIterating("Unsupported Format: No entries after sourceLine");
                            sourceCharacter += base64VLQFormatDecode();
                            if (hasReportedError())
                                return stopIterating();
                            if (sourceCharacter < 0)
                                return setErrorAndStopIterating("Invalid sourceCharacter found");
                            if (!isSourceMappingSegmentEnd()) {
                                hasName = true;
                                nameIndex += base64VLQFormatDecode();
                                if (hasReportedError())
                                    return stopIterating();
                                if (nameIndex < 0)
                                    return setErrorAndStopIterating("Invalid nameIndex found");
                                if (!isSourceMappingSegmentEnd())
                                    return setErrorAndStopIterating("Unsupported Error Format: Entries after nameIndex");
                            }
                        }
                        return { value: captureMapping(hasSource, hasName), done: done };
                    }
                    return stopIterating();
                }
            };
            function captureMapping(hasSource, hasName) {
                return {
                    generatedLine: generatedLine,
                    generatedCharacter: generatedCharacter,
                    sourceIndex: hasSource ? sourceIndex : undefined,
                    sourceLine: hasSource ? sourceLine : undefined,
                    sourceCharacter: hasSource ? sourceCharacter : undefined,
                    nameIndex: hasName ? nameIndex : undefined
                };
            }
            function stopIterating() {
                done = true;
                return { value: undefined, done: true };
            }
            function setError(message) {
                if (error === undefined) {
                    error = message;
                }
            }
            function setErrorAndStopIterating(message) {
                setError(message);
                return stopIterating();
            }
            function hasReportedError() {
                return error !== undefined;
            }
            function isSourceMappingSegmentEnd() {
                return (pos === mappings.length ||
                    mappings.charCodeAt(pos) === 44 /* comma */ ||
                    mappings.charCodeAt(pos) === 59 /* semicolon */);
            }
            function base64VLQFormatDecode() {
                var moreDigits = true;
                var shiftCount = 0;
                var value = 0;
                for (; moreDigits; pos++) {
                    if (pos >= mappings.length)
                        return setError("Error in decoding base64VLQFormatDecode, past the mapping string"), -1;
                    // 6 digit number
                    var currentByte = base64FormatDecode(mappings.charCodeAt(pos));
                    if (currentByte === -1)
                        return setError("Invalid character in VLQ"), -1;
                    // If msb is set, we still have more bits to continue
                    moreDigits = (currentByte & 32) !== 0;
                    // least significant 5 bits are the next msbs in the final value.
                    value = value | ((currentByte & 31) << shiftCount);
                    shiftCount += 5;
                }
                // Least significant bit if 1 represents negative and rest of the msb is actual absolute value
                if ((value & 1) === 0) {
                    // + number
                    value = value >> 1;
                }
                else {
                    // - number
                    value = value >> 1;
                    value = -value;
                }
                return value;
            }
        }
        ts.decodeMappings = decodeMappings;
        function sameMapping(left, right) {
            return left === right
                || left.generatedLine === right.generatedLine
                    && left.generatedCharacter === right.generatedCharacter
                    && left.sourceIndex === right.sourceIndex
                    && left.sourceLine === right.sourceLine
                    && left.sourceCharacter === right.sourceCharacter
                    && left.nameIndex === right.nameIndex;
        }
        ts.sameMapping = sameMapping;
        function isSourceMapping(mapping) {
            return mapping.sourceIndex !== undefined
                && mapping.sourceLine !== undefined
                && mapping.sourceCharacter !== undefined;
        }
        ts.isSourceMapping = isSourceMapping;
        function base64FormatEncode(value) {
            return value >= 0 && value < 26 ? 65 /* A */ + value :
                value >= 26 && value < 52 ? 97 /* a */ + value - 26 :
                    value >= 52 && value < 62 ? 48 /* _0 */ + value - 52 :
                        value === 62 ? 43 /* plus */ :
                            value === 63 ? 47 /* slash */ :
                                ts.Debug.fail(value + ": not a base64 value");
        }
        function base64FormatDecode(ch) {
            return ch >= 65 /* A */ && ch <= 90 /* Z */ ? ch - 65 /* A */ :
                ch >= 97 /* a */ && ch <= 122 /* z */ ? ch - 97 /* a */ + 26 :
                    ch >= 48 /* _0 */ && ch <= 57 /* _9 */ ? ch - 48 /* _0 */ + 52 :
                        ch === 43 /* plus */ ? 62 :
                            ch === 47 /* slash */ ? 63 :
                                -1;
        }
        function isSourceMappedPosition(value) {
            return value.sourceIndex !== undefined
                && value.sourcePosition !== undefined;
        }
        function sameMappedPosition(left, right) {
            return left.generatedPosition === right.generatedPosition
                && left.sourceIndex === right.sourceIndex
                && left.sourcePosition === right.sourcePosition;
        }
        function compareSourcePositions(left, right) {
            // Compares sourcePosition without comparing sourceIndex
            // since the mappings are grouped by sourceIndex
            ts.Debug.assert(left.sourceIndex === right.sourceIndex);
            return ts.compareValues(left.sourcePosition, right.sourcePosition);
        }
        function compareGeneratedPositions(left, right) {
            return ts.compareValues(left.generatedPosition, right.generatedPosition);
        }
        function getSourcePositionOfMapping(value) {
            return value.sourcePosition;
        }
        function getGeneratedPositionOfMapping(value) {
            return value.generatedPosition;
        }
        function createDocumentPositionMapper(host, map, mapPath) {
            var mapDirectory = ts.getDirectoryPath(mapPath);
            var sourceRoot = map.sourceRoot ? ts.getNormalizedAbsolutePath(map.sourceRoot, mapDirectory) : mapDirectory;
            var generatedAbsoluteFilePath = ts.getNormalizedAbsolutePath(map.file, mapDirectory);
            var generatedFile = host.getSourceFileLike(generatedAbsoluteFilePath);
            var sourceFileAbsolutePaths = map.sources.map(function (source) { return ts.getNormalizedAbsolutePath(source, sourceRoot); });
            var sourceToSourceIndexMap = new ts.Map(sourceFileAbsolutePaths.map(function (source, i) { return [host.getCanonicalFileName(source), i]; }));
            var decodedMappings;
            var generatedMappings;
            var sourceMappings;
            return {
                getSourcePosition: getSourcePosition,
                getGeneratedPosition: getGeneratedPosition
            };
            function processMapping(mapping) {
                var generatedPosition = generatedFile !== undefined
                    ? ts.getPositionOfLineAndCharacter(generatedFile, mapping.generatedLine, mapping.generatedCharacter, /*allowEdits*/ true)
                    : -1;
                var source;
                var sourcePosition;
                if (isSourceMapping(mapping)) {
                    var sourceFile = host.getSourceFileLike(sourceFileAbsolutePaths[mapping.sourceIndex]);
                    source = map.sources[mapping.sourceIndex];
                    sourcePosition = sourceFile !== undefined
                        ? ts.getPositionOfLineAndCharacter(sourceFile, mapping.sourceLine, mapping.sourceCharacter, /*allowEdits*/ true)
                        : -1;
                }
                return {
                    generatedPosition: generatedPosition,
                    source: source,
                    sourceIndex: mapping.sourceIndex,
                    sourcePosition: sourcePosition,
                    nameIndex: mapping.nameIndex
                };
            }
            function getDecodedMappings() {
                if (decodedMappings === undefined) {
                    var decoder = decodeMappings(map.mappings);
                    var mappings = ts.arrayFrom(decoder, processMapping);
                    if (decoder.error !== undefined) {
                        if (host.log) {
                            host.log("Encountered error while decoding sourcemap: " + decoder.error);
                        }
                        decodedMappings = ts.emptyArray;
                    }
                    else {
                        decodedMappings = mappings;
                    }
                }
                return decodedMappings;
            }
            function getSourceMappings(sourceIndex) {
                if (sourceMappings === undefined) {
                    var lists = [];
                    for (var _i = 0, _a = getDecodedMappings(); _i < _a.length; _i++) {
                        var mapping = _a[_i];
                        if (!isSourceMappedPosition(mapping))
                            continue;
                        var list = lists[mapping.sourceIndex];
                        if (!list)
                            lists[mapping.sourceIndex] = list = [];
                        list.push(mapping);
                    }
                    sourceMappings = lists.map(function (list) { return ts.sortAndDeduplicate(list, compareSourcePositions, sameMappedPosition); });
                }
                return sourceMappings[sourceIndex];
            }
            function getGeneratedMappings() {
                if (generatedMappings === undefined) {
                    var list = [];
                    for (var _i = 0, _a = getDecodedMappings(); _i < _a.length; _i++) {
                        var mapping = _a[_i];
                        list.push(mapping);
                    }
                    generatedMappings = ts.sortAndDeduplicate(list, compareGeneratedPositions, sameMappedPosition);
                }
                return generatedMappings;
            }
            function getGeneratedPosition(loc) {
                var sourceIndex = sourceToSourceIndexMap.get(host.getCanonicalFileName(loc.fileName));
                if (sourceIndex === undefined)
                    return loc;
                var sourceMappings = getSourceMappings(sourceIndex);
                if (!ts.some(sourceMappings))
                    return loc;
                var targetIndex = ts.binarySearchKey(sourceMappings, loc.pos, getSourcePositionOfMapping, ts.compareValues);
                if (targetIndex < 0) {
                    // if no exact match, closest is 2's complement of result
                    targetIndex = ~targetIndex;
                }
                var mapping = sourceMappings[targetIndex];
                if (mapping === undefined || mapping.sourceIndex !== sourceIndex) {
                    return loc;
                }
                return { fileName: generatedAbsoluteFilePath, pos: mapping.generatedPosition }; // Closest pos
            }
            function getSourcePosition(loc) {
                var generatedMappings = getGeneratedMappings();
                if (!ts.some(generatedMappings))
                    return loc;
                var targetIndex = ts.binarySearchKey(generatedMappings, loc.pos, getGeneratedPositionOfMapping, ts.compareValues);
                if (targetIndex < 0) {
                    // if no exact match, closest is 2's complement of result
                    targetIndex = ~targetIndex;
                }
                var mapping = generatedMappings[targetIndex];
                if (mapping === undefined || !isSourceMappedPosition(mapping)) {
                    return loc;
                }
                return { fileName: sourceFileAbsolutePaths[mapping.sourceIndex], pos: mapping.sourcePosition }; // Closest pos
            }
        }
        ts.createDocumentPositionMapper = createDocumentPositionMapper;
        ts.identitySourceMapConsumer = {
            getSourcePosition: ts.identity,
            getGeneratedPosition: ts.identity
        };
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function getOriginalNodeId(node) {
            node = ts.getOriginalNode(node);
            return node ? ts.getNodeId(node) : 0;
        }
        ts.getOriginalNodeId = getOriginalNodeId;
        function containsDefaultReference(node) {
            if (!node)
                return false;
            if (!ts.isNamedImports(node))
                return false;
            return ts.some(node.elements, isNamedDefaultReference);
        }
        function isNamedDefaultReference(e) {
            return e.propertyName !== undefined && e.propertyName.escapedText === "default" /* Default */;
        }
        function chainBundle(context, transformSourceFile) {
            return transformSourceFileOrBundle;
            function transformSourceFileOrBundle(node) {
                return node.kind === 300 /* SourceFile */ ? transformSourceFile(node) : transformBundle(node);
            }
            function transformBundle(node) {
                return context.factory.createBundle(ts.map(node.sourceFiles, transformSourceFile), node.prepends);
            }
        }
        ts.chainBundle = chainBundle;
        function getExportNeedsImportStarHelper(node) {
            return !!ts.getNamespaceDeclarationNode(node);
        }
        ts.getExportNeedsImportStarHelper = getExportNeedsImportStarHelper;
        function getImportNeedsImportStarHelper(node) {
            if (!!ts.getNamespaceDeclarationNode(node)) {
                return true;
            }
            var bindings = node.importClause && node.importClause.namedBindings;
            if (!bindings) {
                return false;
            }
            if (!ts.isNamedImports(bindings))
                return false;
            var defaultRefCount = 0;
            for (var _i = 0, _a = bindings.elements; _i < _a.length; _i++) {
                var binding = _a[_i];
                if (isNamedDefaultReference(binding)) {
                    defaultRefCount++;
                }
            }
            // Import star is required if there's default named refs mixed with non-default refs, or if theres non-default refs and it has a default import
            return (defaultRefCount > 0 && defaultRefCount !== bindings.elements.length) || (!!(bindings.elements.length - defaultRefCount) && ts.isDefaultImport(node));
        }
        ts.getImportNeedsImportStarHelper = getImportNeedsImportStarHelper;
        function getImportNeedsImportDefaultHelper(node) {
            // Import default is needed if there's a default import or a default ref and no other refs (meaning an import star helper wasn't requested)
            return !getImportNeedsImportStarHelper(node) && (ts.isDefaultImport(node) || (!!node.importClause && ts.isNamedImports(node.importClause.namedBindings) && containsDefaultReference(node.importClause.namedBindings))); // TODO: GH#18217
        }
        ts.getImportNeedsImportDefaultHelper = getImportNeedsImportDefaultHelper;
        function collectExternalModuleInfo(context, sourceFile, resolver, compilerOptions) {
            var externalImports = [];
            var exportSpecifiers = ts.createMultiMap();
            var exportedBindings = [];
            var uniqueExports = new ts.Map();
            var exportedNames;
            var hasExportDefault = false;
            var exportEquals;
            var hasExportStarsToExportValues = false;
            var hasImportStar = false;
            var hasImportDefault = false;
            for (var _i = 0, _a = sourceFile.statements; _i < _a.length; _i++) {
                var node = _a[_i];
                switch (node.kind) {
                    case 264 /* ImportDeclaration */:
                        // import "mod"
                        // import x from "mod"
                        // import * as x from "mod"
                        // import { x, y } from "mod"
                        externalImports.push(node);
                        if (!hasImportStar && getImportNeedsImportStarHelper(node)) {
                            hasImportStar = true;
                        }
                        if (!hasImportDefault && getImportNeedsImportDefaultHelper(node)) {
                            hasImportDefault = true;
                        }
                        break;
                    case 263 /* ImportEqualsDeclaration */:
                        if (node.moduleReference.kind === 275 /* ExternalModuleReference */) {
                            // import x = require("mod")
                            externalImports.push(node);
                        }
                        break;
                    case 270 /* ExportDeclaration */:
                        if (node.moduleSpecifier) {
                            if (!node.exportClause) {
                                // export * from "mod"
                                externalImports.push(node);
                                hasExportStarsToExportValues = true;
                            }
                            else {
                                // export * as ns from "mod"
                                // export { x, y } from "mod"
                                externalImports.push(node);
                                if (ts.isNamedExports(node.exportClause)) {
                                    addExportedNamesForExportDeclaration(node);
                                }
                                else {
                                    var name = node.exportClause.name;
                                    if (!uniqueExports.get(ts.idText(name))) {
                                        multiMapSparseArrayAdd(exportedBindings, getOriginalNodeId(node), name);
                                        uniqueExports.set(ts.idText(name), true);
                                        exportedNames = ts.append(exportedNames, name);
                                    }
                                    // we use the same helpers for `export * as ns` as we do for `import * as ns`
                                    hasImportStar = true;
                                }
                            }
                        }
                        else {
                            // export { x, y }
                            addExportedNamesForExportDeclaration(node);
                        }
                        break;
                    case 269 /* ExportAssignment */:
                        if (node.isExportEquals && !exportEquals) {
                            // export = x
                            exportEquals = node;
                        }
                        break;
                    case 235 /* VariableStatement */:
                        if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                            for (var _b = 0, _c = node.declarationList.declarations; _b < _c.length; _b++) {
                                var decl = _c[_b];
                                exportedNames = collectExportedVariableInfo(decl, uniqueExports, exportedNames);
                            }
                        }
                        break;
                    case 254 /* FunctionDeclaration */:
                        if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                            if (ts.hasSyntacticModifier(node, 512 /* Default */)) {
                                // export default function() { }
                                if (!hasExportDefault) {
                                    multiMapSparseArrayAdd(exportedBindings, getOriginalNodeId(node), context.factory.getDeclarationName(node));
                                    hasExportDefault = true;
                                }
                            }
                            else {
                                // export function x() { }
                                var name = node.name;
                                if (!uniqueExports.get(ts.idText(name))) {
                                    multiMapSparseArrayAdd(exportedBindings, getOriginalNodeId(node), name);
                                    uniqueExports.set(ts.idText(name), true);
                                    exportedNames = ts.append(exportedNames, name);
                                }
                            }
                        }
                        break;
                    case 255 /* ClassDeclaration */:
                        if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                            if (ts.hasSyntacticModifier(node, 512 /* Default */)) {
                                // export default class { }
                                if (!hasExportDefault) {
                                    multiMapSparseArrayAdd(exportedBindings, getOriginalNodeId(node), context.factory.getDeclarationName(node));
                                    hasExportDefault = true;
                                }
                            }
                            else {
                                // export class x { }
                                var name = node.name;
                                if (name && !uniqueExports.get(ts.idText(name))) {
                                    multiMapSparseArrayAdd(exportedBindings, getOriginalNodeId(node), name);
                                    uniqueExports.set(ts.idText(name), true);
                                    exportedNames = ts.append(exportedNames, name);
                                }
                            }
                        }
                        break;
                }
            }
            var externalHelpersImportDeclaration = ts.createExternalHelpersImportDeclarationIfNeeded(context.factory, context.getEmitHelperFactory(), sourceFile, compilerOptions, hasExportStarsToExportValues, hasImportStar, hasImportDefault);
            if (externalHelpersImportDeclaration) {
                externalImports.unshift(externalHelpersImportDeclaration);
            }
            return { externalImports: externalImports, exportSpecifiers: exportSpecifiers, exportEquals: exportEquals, hasExportStarsToExportValues: hasExportStarsToExportValues, exportedBindings: exportedBindings, exportedNames: exportedNames, externalHelpersImportDeclaration: externalHelpersImportDeclaration };
            function addExportedNamesForExportDeclaration(node) {
                for (var _i = 0, _a = ts.cast(node.exportClause, ts.isNamedExports).elements; _i < _a.length; _i++) {
                    var specifier = _a[_i];
                    if (!uniqueExports.get(ts.idText(specifier.name))) {
                        var name = specifier.propertyName || specifier.name;
                        if (!node.moduleSpecifier) {
                            exportSpecifiers.add(ts.idText(name), specifier);
                        }
                        var decl = resolver.getReferencedImportDeclaration(name)
                            || resolver.getReferencedValueDeclaration(name);
                        if (decl) {
                            multiMapSparseArrayAdd(exportedBindings, getOriginalNodeId(decl), specifier.name);
                        }
                        uniqueExports.set(ts.idText(specifier.name), true);
                        exportedNames = ts.append(exportedNames, specifier.name);
                    }
                }
            }
        }
        ts.collectExternalModuleInfo = collectExternalModuleInfo;
        function collectExportedVariableInfo(decl, uniqueExports, exportedNames) {
            if (ts.isBindingPattern(decl.name)) {
                for (var _i = 0, _a = decl.name.elements; _i < _a.length; _i++) {
                    var element = _a[_i];
                    if (!ts.isOmittedExpression(element)) {
                        exportedNames = collectExportedVariableInfo(element, uniqueExports, exportedNames);
                    }
                }
            }
            else if (!ts.isGeneratedIdentifier(decl.name)) {
                var text = ts.idText(decl.name);
                if (!uniqueExports.get(text)) {
                    uniqueExports.set(text, true);
                    exportedNames = ts.append(exportedNames, decl.name);
                }
            }
            return exportedNames;
        }
        /** Use a sparse array as a multi-map. */
        function multiMapSparseArrayAdd(map, key, value) {
            var values = map[key];
            if (values) {
                values.push(value);
            }
            else {
                map[key] = values = [value];
            }
            return values;
        }
        /**
         * Used in the module transformer to check if an expression is reasonably without sideeffect,
         *  and thus better to copy into multiple places rather than to cache in a temporary variable
         *  - this is mostly subjective beyond the requirement that the expression not be sideeffecting
         */
        function isSimpleCopiableExpression(expression) {
            return ts.isStringLiteralLike(expression) ||
                expression.kind === 8 /* NumericLiteral */ ||
                ts.isKeyword(expression.kind) ||
                ts.isIdentifier(expression);
        }
        ts.isSimpleCopiableExpression = isSimpleCopiableExpression;
        /**
         * A simple inlinable expression is an expression which can be copied into multiple locations
         * without risk of repeating any sideeffects and whose value could not possibly change between
         * any such locations
         */
        function isSimpleInlineableExpression(expression) {
            return !ts.isIdentifier(expression) && isSimpleCopiableExpression(expression);
        }
        ts.isSimpleInlineableExpression = isSimpleInlineableExpression;
        function isCompoundAssignment(kind) {
            return kind >= 64 /* FirstCompoundAssignment */
                && kind <= 78 /* LastCompoundAssignment */;
        }
        ts.isCompoundAssignment = isCompoundAssignment;
        function getNonAssignmentOperatorForCompoundAssignment(kind) {
            switch (kind) {
                case 64 /* PlusEqualsToken */: return 39 /* PlusToken */;
                case 65 /* MinusEqualsToken */: return 40 /* MinusToken */;
                case 66 /* AsteriskEqualsToken */: return 41 /* AsteriskToken */;
                case 67 /* AsteriskAsteriskEqualsToken */: return 42 /* AsteriskAsteriskToken */;
                case 68 /* SlashEqualsToken */: return 43 /* SlashToken */;
                case 69 /* PercentEqualsToken */: return 44 /* PercentToken */;
                case 70 /* LessThanLessThanEqualsToken */: return 47 /* LessThanLessThanToken */;
                case 71 /* GreaterThanGreaterThanEqualsToken */: return 48 /* GreaterThanGreaterThanToken */;
                case 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */: return 49 /* GreaterThanGreaterThanGreaterThanToken */;
                case 73 /* AmpersandEqualsToken */: return 50 /* AmpersandToken */;
                case 74 /* BarEqualsToken */: return 51 /* BarToken */;
                case 78 /* CaretEqualsToken */: return 52 /* CaretToken */;
                case 75 /* BarBarEqualsToken */: return 56 /* BarBarToken */;
                case 76 /* AmpersandAmpersandEqualsToken */: return 55 /* AmpersandAmpersandToken */;
                case 77 /* QuestionQuestionEqualsToken */: return 60 /* QuestionQuestionToken */;
            }
        }
        ts.getNonAssignmentOperatorForCompoundAssignment = getNonAssignmentOperatorForCompoundAssignment;
        /**
         * Adds super call and preceding prologue directives into the list of statements.
         *
         * @param ctor The constructor node.
         * @param result The list of statements.
         * @param visitor The visitor to apply to each node added to the result array.
         * @returns index of the statement that follows super call
         */
        function addPrologueDirectivesAndInitialSuperCall(factory, ctor, result, visitor) {
            if (ctor.body) {
                var statements = ctor.body.statements;
                // add prologue directives to the list (if any)
                var index = factory.copyPrologue(statements, result, /*ensureUseStrict*/ false, visitor);
                if (index === statements.length) {
                    // list contains nothing but prologue directives (or empty) - exit
                    return index;
                }
                var superIndex = ts.findIndex(statements, function (s) { return ts.isExpressionStatement(s) && ts.isSuperCall(s.expression); }, index);
                if (superIndex > -1) {
                    for (var i = index; i <= superIndex; i++) {
                        result.push(ts.visitNode(statements[i], visitor, ts.isStatement));
                    }
                    return superIndex + 1;
                }
                return index;
            }
            return 0;
        }
        ts.addPrologueDirectivesAndInitialSuperCall = addPrologueDirectivesAndInitialSuperCall;
        function getProperties(node, requireInitializer, isStatic) {
            return ts.filter(node.members, function (m) { return isInitializedOrStaticProperty(m, requireInitializer, isStatic); });
        }
        ts.getProperties = getProperties;
        function isStaticPropertyDeclarationOrClassStaticBlockDeclaration(element) {
            return isStaticPropertyDeclaration(element) || ts.isClassStaticBlockDeclaration(element);
        }
        function getStaticPropertiesAndClassStaticBlock(node) {
            return ts.filter(node.members, isStaticPropertyDeclarationOrClassStaticBlockDeclaration);
        }
        ts.getStaticPropertiesAndClassStaticBlock = getStaticPropertiesAndClassStaticBlock;
        /**
         * Is a class element either a static or an instance property declaration with an initializer?
         *
         * @param member The class element node.
         * @param isStatic A value indicating whether the member should be a static or instance member.
         */
        function isInitializedOrStaticProperty(member, requireInitializer, isStatic) {
            return ts.isPropertyDeclaration(member)
                && (!!member.initializer || !requireInitializer)
                && ts.hasStaticModifier(member) === isStatic;
        }
        function isStaticPropertyDeclaration(member) {
            return ts.isPropertyDeclaration(member) && ts.hasStaticModifier(member);
        }
        /**
         * Gets a value indicating whether a class element is either a static or an instance property declaration with an initializer.
         *
         * @param member The class element node.
         * @param isStatic A value indicating whether the member should be a static or instance member.
         */
        function isInitializedProperty(member) {
            return member.kind === 165 /* PropertyDeclaration */
                && member.initializer !== undefined;
        }
        ts.isInitializedProperty = isInitializedProperty;
        /**
         * Gets a value indicating whether a class element is a private instance method or accessor.
         *
         * @param member The class element node.
         */
        function isNonStaticMethodOrAccessorWithPrivateName(member) {
            return !ts.isStatic(member) && ts.isMethodOrAccessor(member) && ts.isPrivateIdentifier(member.name);
        }
        ts.isNonStaticMethodOrAccessorWithPrivateName = isNonStaticMethodOrAccessorWithPrivateName;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var FlattenLevel;
        (function (FlattenLevel) {
            FlattenLevel[FlattenLevel["All"] = 0] = "All";
            FlattenLevel[FlattenLevel["ObjectRest"] = 1] = "ObjectRest";
        })(FlattenLevel = ts.FlattenLevel || (ts.FlattenLevel = {}));
        /**
         * Flattens a DestructuringAssignment or a VariableDeclaration to an expression.
         *
         * @param node The node to flatten.
         * @param visitor An optional visitor used to visit initializers.
         * @param context The transformation context.
         * @param level Indicates the extent to which flattening should occur.
         * @param needsValue An optional value indicating whether the value from the right-hand-side of
         * the destructuring assignment is needed as part of a larger expression.
         * @param createAssignmentCallback An optional callback used to create the assignment expression.
         */
        function flattenDestructuringAssignment(node, visitor, context, level, needsValue, createAssignmentCallback) {
            var location = node;
            var value;
            if (ts.isDestructuringAssignment(node)) {
                value = node.right;
                while (ts.isEmptyArrayLiteral(node.left) || ts.isEmptyObjectLiteral(node.left)) {
                    if (ts.isDestructuringAssignment(value)) {
                        location = node = value;
                        value = node.right;
                    }
                    else {
                        return ts.visitNode(value, visitor, ts.isExpression);
                    }
                }
            }
            var expressions;
            var flattenContext = {
                context: context,
                level: level,
                downlevelIteration: !!context.getCompilerOptions().downlevelIteration,
                hoistTempVariables: true,
                emitExpression: emitExpression,
                emitBindingOrAssignment: emitBindingOrAssignment,
                createArrayBindingOrAssignmentPattern: function (elements) { return makeArrayAssignmentPattern(context.factory, elements); },
                createObjectBindingOrAssignmentPattern: function (elements) { return makeObjectAssignmentPattern(context.factory, elements); },
                createArrayBindingOrAssignmentElement: makeAssignmentElement,
                visitor: visitor
            };
            if (value) {
                value = ts.visitNode(value, visitor, ts.isExpression);
                if (ts.isIdentifier(value) && bindingOrAssignmentElementAssignsToName(node, value.escapedText) ||
                    bindingOrAssignmentElementContainsNonLiteralComputedName(node)) {
                    // If the right-hand value of the assignment is also an assignment target then
                    // we need to cache the right-hand value.
                    value = ensureIdentifier(flattenContext, value, /*reuseIdentifierExpressions*/ false, location);
                }
                else if (needsValue) {
                    // If the right-hand value of the destructuring assignment needs to be preserved (as
                    // is the case when the destructuring assignment is part of a larger expression),
                    // then we need to cache the right-hand value.
                    //
                    // The source map location for the assignment should point to the entire binary
                    // expression.
                    value = ensureIdentifier(flattenContext, value, /*reuseIdentifierExpressions*/ true, location);
                }
                else if (ts.nodeIsSynthesized(node)) {
                    // Generally, the source map location for a destructuring assignment is the root
                    // expression.
                    //
                    // However, if the root expression is synthesized (as in the case
                    // of the initializer when transforming a ForOfStatement), then the source map
                    // location should point to the right-hand value of the expression.
                    location = value;
                }
            }
            flattenBindingOrAssignmentElement(flattenContext, node, value, location, /*skipInitializer*/ ts.isDestructuringAssignment(node));
            if (value && needsValue) {
                if (!ts.some(expressions)) {
                    return value;
                }
                expressions.push(value);
            }
            return context.factory.inlineExpressions(expressions) || context.factory.createOmittedExpression();
            function emitExpression(expression) {
                expressions = ts.append(expressions, expression);
            }
            function emitBindingOrAssignment(target, value, location, original) {
                ts.Debug.assertNode(target, createAssignmentCallback ? ts.isIdentifier : ts.isExpression);
                var expression = createAssignmentCallback
                    ? createAssignmentCallback(target, value, location)
                    : ts.setTextRange(context.factory.createAssignment(ts.visitNode(target, visitor, ts.isExpression), value), location);
                expression.original = original;
                emitExpression(expression);
            }
        }
        ts.flattenDestructuringAssignment = flattenDestructuringAssignment;
        function bindingOrAssignmentElementAssignsToName(element, escapedName) {
            var target = ts.getTargetOfBindingOrAssignmentElement(element); // TODO: GH#18217
            if (ts.isBindingOrAssignmentPattern(target)) {
                return bindingOrAssignmentPatternAssignsToName(target, escapedName);
            }
            else if (ts.isIdentifier(target)) {
                return target.escapedText === escapedName;
            }
            return false;
        }
        function bindingOrAssignmentPatternAssignsToName(pattern, escapedName) {
            var elements = ts.getElementsOfBindingOrAssignmentPattern(pattern);
            for (var _i = 0, elements_3 = elements; _i < elements_3.length; _i++) {
                var element = elements_3[_i];
                if (bindingOrAssignmentElementAssignsToName(element, escapedName)) {
                    return true;
                }
            }
            return false;
        }
        function bindingOrAssignmentElementContainsNonLiteralComputedName(element) {
            var propertyName = ts.tryGetPropertyNameOfBindingOrAssignmentElement(element);
            if (propertyName && ts.isComputedPropertyName(propertyName) && !ts.isLiteralExpression(propertyName.expression)) {
                return true;
            }
            var target = ts.getTargetOfBindingOrAssignmentElement(element);
            return !!target && ts.isBindingOrAssignmentPattern(target) && bindingOrAssignmentPatternContainsNonLiteralComputedName(target);
        }
        function bindingOrAssignmentPatternContainsNonLiteralComputedName(pattern) {
            return !!ts.forEach(ts.getElementsOfBindingOrAssignmentPattern(pattern), bindingOrAssignmentElementContainsNonLiteralComputedName);
        }
        /**
         * Flattens a VariableDeclaration or ParameterDeclaration to one or more variable declarations.
         *
         * @param node The node to flatten.
         * @param visitor An optional visitor used to visit initializers.
         * @param context The transformation context.
         * @param boundValue The value bound to the declaration.
         * @param skipInitializer A value indicating whether to ignore the initializer of `node`.
         * @param hoistTempVariables Indicates whether temporary variables should not be recorded in-line.
         * @param level Indicates the extent to which flattening should occur.
         */
        function flattenDestructuringBinding(node, visitor, context, level, rval, hoistTempVariables, skipInitializer) {
            if (hoistTempVariables === void 0) {
                hoistTempVariables = false;
            }
            var pendingExpressions;
            var pendingDeclarations = [];
            var declarations = [];
            var flattenContext = {
                context: context,
                level: level,
                downlevelIteration: !!context.getCompilerOptions().downlevelIteration,
                hoistTempVariables: hoistTempVariables,
                emitExpression: emitExpression,
                emitBindingOrAssignment: emitBindingOrAssignment,
                createArrayBindingOrAssignmentPattern: function (elements) { return makeArrayBindingPattern(context.factory, elements); },
                createObjectBindingOrAssignmentPattern: function (elements) { return makeObjectBindingPattern(context.factory, elements); },
                createArrayBindingOrAssignmentElement: function (name) { return makeBindingElement(context.factory, name); },
                visitor: visitor
            };
            if (ts.isVariableDeclaration(node)) {
                var initializer = ts.getInitializerOfBindingOrAssignmentElement(node);
                if (initializer && (ts.isIdentifier(initializer) && bindingOrAssignmentElementAssignsToName(node, initializer.escapedText) ||
                    bindingOrAssignmentElementContainsNonLiteralComputedName(node))) {
                    // If the right-hand value of the assignment is also an assignment target then
                    // we need to cache the right-hand value.
                    initializer = ensureIdentifier(flattenContext, ts.visitNode(initializer, flattenContext.visitor), /*reuseIdentifierExpressions*/ false, initializer);
                    node = context.factory.updateVariableDeclaration(node, node.name, /*exclamationToken*/ undefined, /*type*/ undefined, initializer);
                }
            }
            flattenBindingOrAssignmentElement(flattenContext, node, rval, node, skipInitializer);
            if (pendingExpressions) {
                var temp = context.factory.createTempVariable(/*recordTempVariable*/ undefined);
                if (hoistTempVariables) {
                    var value = context.factory.inlineExpressions(pendingExpressions);
                    pendingExpressions = undefined;
                    emitBindingOrAssignment(temp, value, /*location*/ undefined, /*original*/ undefined);
                }
                else {
                    context.hoistVariableDeclaration(temp);
                    var pendingDeclaration = ts.last(pendingDeclarations);
                    pendingDeclaration.pendingExpressions = ts.append(pendingDeclaration.pendingExpressions, context.factory.createAssignment(temp, pendingDeclaration.value));
                    ts.addRange(pendingDeclaration.pendingExpressions, pendingExpressions);
                    pendingDeclaration.value = temp;
                }
            }
            for (var _i = 0, pendingDeclarations_1 = pendingDeclarations; _i < pendingDeclarations_1.length; _i++) {
                var _a = pendingDeclarations_1[_i], pendingExpressions_1 = _a.pendingExpressions, name = _a.name, value = _a.value, location = _a.location, original = _a.original;
                var variable = context.factory.createVariableDeclaration(name, 
                /*exclamationToken*/ undefined, 
                /*type*/ undefined, pendingExpressions_1 ? context.factory.inlineExpressions(ts.append(pendingExpressions_1, value)) : value);
                variable.original = original;
                ts.setTextRange(variable, location);
                declarations.push(variable);
            }
            return declarations;
            function emitExpression(value) {
                pendingExpressions = ts.append(pendingExpressions, value);
            }
            function emitBindingOrAssignment(target, value, location, original) {
                ts.Debug.assertNode(target, ts.isBindingName);
                if (pendingExpressions) {
                    value = context.factory.inlineExpressions(ts.append(pendingExpressions, value));
                    pendingExpressions = undefined;
                }
                pendingDeclarations.push({ pendingExpressions: pendingExpressions, name: target, value: value, location: location, original: original });
            }
        }
        ts.flattenDestructuringBinding = flattenDestructuringBinding;
        /**
         * Flattens a BindingOrAssignmentElement into zero or more bindings or assignments.
         *
         * @param flattenContext Options used to control flattening.
         * @param element The element to flatten.
         * @param value The current RHS value to assign to the element.
         * @param location The location to use for source maps and comments.
         * @param skipInitializer An optional value indicating whether to include the initializer
         * for the element.
         */
        function flattenBindingOrAssignmentElement(flattenContext, element, value, location, skipInitializer) {
            var bindingTarget = ts.getTargetOfBindingOrAssignmentElement(element); // TODO: GH#18217
            if (!skipInitializer) {
                var initializer = ts.visitNode(ts.getInitializerOfBindingOrAssignmentElement(element), flattenContext.visitor, ts.isExpression);
                if (initializer) {
                    // Combine value and initializer
                    if (value) {
                        value = createDefaultValueCheck(flattenContext, value, initializer, location);
                        // If 'value' is not a simple expression, it could contain side-effecting code that should evaluate before an object or array binding pattern.
                        if (!ts.isSimpleInlineableExpression(initializer) && ts.isBindingOrAssignmentPattern(bindingTarget)) {
                            value = ensureIdentifier(flattenContext, value, /*reuseIdentifierExpressions*/ true, location);
                        }
                    }
                    else {
                        value = initializer;
                    }
                }
                else if (!value) {
                    // Use 'void 0' in absence of value and initializer
                    value = flattenContext.context.factory.createVoidZero();
                }
            }
            if (ts.isObjectBindingOrAssignmentPattern(bindingTarget)) {
                flattenObjectBindingOrAssignmentPattern(flattenContext, element, bindingTarget, value, location);
            }
            else if (ts.isArrayBindingOrAssignmentPattern(bindingTarget)) {
                flattenArrayBindingOrAssignmentPattern(flattenContext, element, bindingTarget, value, location);
            }
            else {
                flattenContext.emitBindingOrAssignment(bindingTarget, value, location, /*original*/ element); // TODO: GH#18217
            }
        }
        /**
         * Flattens an ObjectBindingOrAssignmentPattern into zero or more bindings or assignments.
         *
         * @param flattenContext Options used to control flattening.
         * @param parent The parent element of the pattern.
         * @param pattern The ObjectBindingOrAssignmentPattern to flatten.
         * @param value The current RHS value to assign to the element.
         * @param location The location to use for source maps and comments.
         */
        function flattenObjectBindingOrAssignmentPattern(flattenContext, parent, pattern, value, location) {
            var elements = ts.getElementsOfBindingOrAssignmentPattern(pattern);
            var numElements = elements.length;
            if (numElements !== 1) {
                // For anything other than a single-element destructuring we need to generate a temporary
                // to ensure value is evaluated exactly once. Additionally, if we have zero elements
                // we need to emit *something* to ensure that in case a 'var' keyword was already emitted,
                // so in that case, we'll intentionally create that temporary.
                var reuseIdentifierExpressions = !ts.isDeclarationBindingElement(parent) || numElements !== 0;
                value = ensureIdentifier(flattenContext, value, reuseIdentifierExpressions, location);
            }
            var bindingElements;
            var computedTempVariables;
            for (var i = 0; i < numElements; i++) {
                var element = elements[i];
                if (!ts.getRestIndicatorOfBindingOrAssignmentElement(element)) {
                    var propertyName = ts.getPropertyNameOfBindingOrAssignmentElement(element);
                    if (flattenContext.level >= 1 /* ObjectRest */
                        && !(element.transformFlags & (16384 /* ContainsRestOrSpread */ | 32768 /* ContainsObjectRestOrSpread */))
                        && !(ts.getTargetOfBindingOrAssignmentElement(element).transformFlags & (16384 /* ContainsRestOrSpread */ | 32768 /* ContainsObjectRestOrSpread */))
                        && !ts.isComputedPropertyName(propertyName)) {
                        bindingElements = ts.append(bindingElements, ts.visitNode(element, flattenContext.visitor));
                    }
                    else {
                        if (bindingElements) {
                            flattenContext.emitBindingOrAssignment(flattenContext.createObjectBindingOrAssignmentPattern(bindingElements), value, location, pattern);
                            bindingElements = undefined;
                        }
                        var rhsValue = createDestructuringPropertyAccess(flattenContext, value, propertyName);
                        if (ts.isComputedPropertyName(propertyName)) {
                            computedTempVariables = ts.append(computedTempVariables, rhsValue.argumentExpression);
                        }
                        flattenBindingOrAssignmentElement(flattenContext, element, rhsValue, /*location*/ element);
                    }
                }
                else if (i === numElements - 1) {
                    if (bindingElements) {
                        flattenContext.emitBindingOrAssignment(flattenContext.createObjectBindingOrAssignmentPattern(bindingElements), value, location, pattern);
                        bindingElements = undefined;
                    }
                    var rhsValue = flattenContext.context.getEmitHelperFactory().createRestHelper(value, elements, computedTempVariables, pattern);
                    flattenBindingOrAssignmentElement(flattenContext, element, rhsValue, element);
                }
            }
            if (bindingElements) {
                flattenContext.emitBindingOrAssignment(flattenContext.createObjectBindingOrAssignmentPattern(bindingElements), value, location, pattern);
            }
        }
        /**
         * Flattens an ArrayBindingOrAssignmentPattern into zero or more bindings or assignments.
         *
         * @param flattenContext Options used to control flattening.
         * @param parent The parent element of the pattern.
         * @param pattern The ArrayBindingOrAssignmentPattern to flatten.
         * @param value The current RHS value to assign to the element.
         * @param location The location to use for source maps and comments.
         */
        function flattenArrayBindingOrAssignmentPattern(flattenContext, parent, pattern, value, location) {
            var elements = ts.getElementsOfBindingOrAssignmentPattern(pattern);
            var numElements = elements.length;
            if (flattenContext.level < 1 /* ObjectRest */ && flattenContext.downlevelIteration) {
                // Read the elements of the iterable into an array
                value = ensureIdentifier(flattenContext, ts.setTextRange(flattenContext.context.getEmitHelperFactory().createReadHelper(value, numElements > 0 && ts.getRestIndicatorOfBindingOrAssignmentElement(elements[numElements - 1])
                    ? undefined
                    : numElements), location), 
                /*reuseIdentifierExpressions*/ false, location);
            }
            else if (numElements !== 1 && (flattenContext.level < 1 /* ObjectRest */ || numElements === 0)
                || ts.every(elements, ts.isOmittedExpression)) {
                // For anything other than a single-element destructuring we need to generate a temporary
                // to ensure value is evaluated exactly once. Additionally, if we have zero elements
                // we need to emit *something* to ensure that in case a 'var' keyword was already emitted,
                // so in that case, we'll intentionally create that temporary.
                // Or all the elements of the binding pattern are omitted expression such as "var [,] = [1,2]",
                // then we will create temporary variable.
                var reuseIdentifierExpressions = !ts.isDeclarationBindingElement(parent) || numElements !== 0;
                value = ensureIdentifier(flattenContext, value, reuseIdentifierExpressions, location);
            }
            var bindingElements;
            var restContainingElements;
            for (var i = 0; i < numElements; i++) {
                var element = elements[i];
                if (flattenContext.level >= 1 /* ObjectRest */) {
                    // If an array pattern contains an ObjectRest, we must cache the result so that we
                    // can perform the ObjectRest destructuring in a different declaration
                    if (element.transformFlags & 32768 /* ContainsObjectRestOrSpread */ || flattenContext.hasTransformedPriorElement && !isSimpleBindingOrAssignmentElement(element)) {
                        flattenContext.hasTransformedPriorElement = true;
                        var temp = flattenContext.context.factory.createTempVariable(/*recordTempVariable*/ undefined);
                        if (flattenContext.hoistTempVariables) {
                            flattenContext.context.hoistVariableDeclaration(temp);
                        }
                        restContainingElements = ts.append(restContainingElements, [temp, element]);
                        bindingElements = ts.append(bindingElements, flattenContext.createArrayBindingOrAssignmentElement(temp));
                    }
                    else {
                        bindingElements = ts.append(bindingElements, element);
                    }
                }
                else if (ts.isOmittedExpression(element)) {
                    continue;
                }
                else if (!ts.getRestIndicatorOfBindingOrAssignmentElement(element)) {
                    var rhsValue = flattenContext.context.factory.createElementAccessExpression(value, i);
                    flattenBindingOrAssignmentElement(flattenContext, element, rhsValue, /*location*/ element);
                }
                else if (i === numElements - 1) {
                    var rhsValue = flattenContext.context.factory.createArraySliceCall(value, i);
                    flattenBindingOrAssignmentElement(flattenContext, element, rhsValue, /*location*/ element);
                }
            }
            if (bindingElements) {
                flattenContext.emitBindingOrAssignment(flattenContext.createArrayBindingOrAssignmentPattern(bindingElements), value, location, pattern);
            }
            if (restContainingElements) {
                for (var _i = 0, restContainingElements_1 = restContainingElements; _i < restContainingElements_1.length; _i++) {
                    var _a = restContainingElements_1[_i], id = _a[0], element = _a[1];
                    flattenBindingOrAssignmentElement(flattenContext, element, id, element);
                }
            }
        }
        function isSimpleBindingOrAssignmentElement(element) {
            var target = ts.getTargetOfBindingOrAssignmentElement(element);
            if (!target || ts.isOmittedExpression(target))
                return true;
            var propertyName = ts.tryGetPropertyNameOfBindingOrAssignmentElement(element);
            if (propertyName && !ts.isPropertyNameLiteral(propertyName))
                return false;
            var initializer = ts.getInitializerOfBindingOrAssignmentElement(element);
            if (initializer && !ts.isSimpleInlineableExpression(initializer))
                return false;
            if (ts.isBindingOrAssignmentPattern(target))
                return ts.every(ts.getElementsOfBindingOrAssignmentPattern(target), isSimpleBindingOrAssignmentElement);
            return ts.isIdentifier(target);
        }
        /**
         * Creates an expression used to provide a default value if a value is `undefined` at runtime.
         *
         * @param flattenContext Options used to control flattening.
         * @param value The RHS value to test.
         * @param defaultValue The default value to use if `value` is `undefined` at runtime.
         * @param location The location to use for source maps and comments.
         */
        function createDefaultValueCheck(flattenContext, value, defaultValue, location) {
            value = ensureIdentifier(flattenContext, value, /*reuseIdentifierExpressions*/ true, location);
            return flattenContext.context.factory.createConditionalExpression(flattenContext.context.factory.createTypeCheck(value, "undefined"), /*questionToken*/ undefined, defaultValue, /*colonToken*/ undefined, value);
        }
        /**
         * Creates either a PropertyAccessExpression or an ElementAccessExpression for the
         * right-hand side of a transformed destructuring assignment.
         *
         * @link https://tc39.github.io/ecma262/#sec-runtime-semantics-keyeddestructuringassignmentevaluation
         *
         * @param flattenContext Options used to control flattening.
         * @param value The RHS value that is the source of the property.
         * @param propertyName The destructuring property name.
         */
        function createDestructuringPropertyAccess(flattenContext, value, propertyName) {
            if (ts.isComputedPropertyName(propertyName)) {
                var argumentExpression = ensureIdentifier(flattenContext, ts.visitNode(propertyName.expression, flattenContext.visitor), /*reuseIdentifierExpressions*/ false, /*location*/ propertyName);
                return flattenContext.context.factory.createElementAccessExpression(value, argumentExpression);
            }
            else if (ts.isStringOrNumericLiteralLike(propertyName)) {
                var argumentExpression = ts.factory.cloneNode(propertyName);
                return flattenContext.context.factory.createElementAccessExpression(value, argumentExpression);
            }
            else {
                var name = flattenContext.context.factory.createIdentifier(ts.idText(propertyName));
                return flattenContext.context.factory.createPropertyAccessExpression(value, name);
            }
        }
        /**
         * Ensures that there exists a declared identifier whose value holds the given expression.
         * This function is useful to ensure that the expression's value can be read from in subsequent expressions.
         * Unless 'reuseIdentifierExpressions' is false, 'value' will be returned if it is just an identifier.
         *
         * @param flattenContext Options used to control flattening.
         * @param value the expression whose value needs to be bound.
         * @param reuseIdentifierExpressions true if identifier expressions can simply be returned;
         * false if it is necessary to always emit an identifier.
         * @param location The location to use for source maps and comments.
         */
        function ensureIdentifier(flattenContext, value, reuseIdentifierExpressions, location) {
            if (ts.isIdentifier(value) && reuseIdentifierExpressions) {
                return value;
            }
            else {
                var temp = flattenContext.context.factory.createTempVariable(/*recordTempVariable*/ undefined);
                if (flattenContext.hoistTempVariables) {
                    flattenContext.context.hoistVariableDeclaration(temp);
                    flattenContext.emitExpression(ts.setTextRange(flattenContext.context.factory.createAssignment(temp, value), location));
                }
                else {
                    flattenContext.emitBindingOrAssignment(temp, value, location, /*original*/ undefined);
                }
                return temp;
            }
        }
        function makeArrayBindingPattern(factory, elements) {
            ts.Debug.assertEachNode(elements, ts.isArrayBindingElement);
            return factory.createArrayBindingPattern(elements);
        }
        function makeArrayAssignmentPattern(factory, elements) {
            return factory.createArrayLiteralExpression(ts.map(elements, factory.converters.convertToArrayAssignmentElement));
        }
        function makeObjectBindingPattern(factory, elements) {
            ts.Debug.assertEachNode(elements, ts.isBindingElement);
            return factory.createObjectBindingPattern(elements);
        }
        function makeObjectAssignmentPattern(factory, elements) {
            return factory.createObjectLiteralExpression(ts.map(elements, factory.converters.convertToObjectAssignmentElement));
        }
        function makeBindingElement(factory, name) {
            return factory.createBindingElement(/*dotDotDotToken*/ undefined, /*propertyName*/ undefined, name);
        }
        function makeAssignmentElement(name) {
            return name;
        }
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var ProcessLevel;
        (function (ProcessLevel) {
            ProcessLevel[ProcessLevel["LiftRestriction"] = 0] = "LiftRestriction";
            ProcessLevel[ProcessLevel["All"] = 1] = "All";
        })(ProcessLevel = ts.ProcessLevel || (ts.ProcessLevel = {}));
        function processTaggedTemplateExpression(context, node, visitor, currentSourceFile, recordTaggedTemplateString, level) {
            // Visit the tag expression
            var tag = ts.visitNode(node.tag, visitor, ts.isExpression);
            // Build up the template arguments and the raw and cooked strings for the template.
            // We start out with 'undefined' for the first argument and revisit later
            // to avoid walking over the template string twice and shifting all our arguments over after the fact.
            var templateArguments = [undefined];
            var cookedStrings = [];
            var rawStrings = [];
            var template = node.template;
            if (level === ProcessLevel.LiftRestriction && !ts.hasInvalidEscape(template)) {
                return ts.visitEachChild(node, visitor, context);
            }
            if (ts.isNoSubstitutionTemplateLiteral(template)) {
                cookedStrings.push(createTemplateCooked(template));
                rawStrings.push(getRawLiteral(template, currentSourceFile));
            }
            else {
                cookedStrings.push(createTemplateCooked(template.head));
                rawStrings.push(getRawLiteral(template.head, currentSourceFile));
                for (var _i = 0, _a = template.templateSpans; _i < _a.length; _i++) {
                    var templateSpan = _a[_i];
                    cookedStrings.push(createTemplateCooked(templateSpan.literal));
                    rawStrings.push(getRawLiteral(templateSpan.literal, currentSourceFile));
                    templateArguments.push(ts.visitNode(templateSpan.expression, visitor, ts.isExpression));
                }
            }
            var helperCall = context.getEmitHelperFactory().createTemplateObjectHelper(ts.factory.createArrayLiteralExpression(cookedStrings), ts.factory.createArrayLiteralExpression(rawStrings));
            // Create a variable to cache the template object if we're in a module.
            // Do not do this in the global scope, as any variable we currently generate could conflict with
            // variables from outside of the current compilation. In the future, we can revisit this behavior.
            if (ts.isExternalModule(currentSourceFile)) {
                var tempVar = ts.factory.createUniqueName("templateObject");
                recordTaggedTemplateString(tempVar);
                templateArguments[0] = ts.factory.createLogicalOr(tempVar, ts.factory.createAssignment(tempVar, helperCall));
            }
            else {
                templateArguments[0] = helperCall;
            }
            return ts.factory.createCallExpression(tag, /*typeArguments*/ undefined, templateArguments);
        }
        ts.processTaggedTemplateExpression = processTaggedTemplateExpression;
        function createTemplateCooked(template) {
            return template.templateFlags ? ts.factory.createVoidZero() : ts.factory.createStringLiteral(template.text);
        }
        /**
         * Creates an ES5 compatible literal from an ES6 template literal.
         *
         * @param node The ES6 template literal.
         */
        function getRawLiteral(node, currentSourceFile) {
            // Find original source text, since we need to emit the raw strings of the tagged template.
            // The raw strings contain the (escaped) strings of what the user wrote.
            // Examples: `\n` is converted to "\\n", a template string with a newline to "\n".
            var text = node.rawText;
            if (text === undefined) {
                ts.Debug.assertIsDefined(currentSourceFile, "Template literal node is missing 'rawText' and does not have a source file. Possibly bad transform.");
                text = ts.getSourceTextOfNodeFromSourceFile(currentSourceFile, node);
                // text contains the original source, it will also contain quotes ("`"), dolar signs and braces ("${" and "}"),
                // thus we need to remove those characters.
                // First template piece starts with "`", others with "}"
                // Last template piece ends with "`", others with "${"
                var isLast = node.kind === 14 /* NoSubstitutionTemplateLiteral */ || node.kind === 17 /* TemplateTail */;
                text = text.substring(1, text.length - (isLast ? 1 : 2));
            }
            // Newline normalization:
            // ES6 Spec 11.8.6.1 - Static Semantics of TV's and TRV's
            // <CR><LF> and <CR> LineTerminatorSequences are normalized to <LF> for both TV and TRV.
            text = text.replace(/\r\n?/g, "\n");
            return ts.setTextRange(ts.factory.createStringLiteral(text), node);
        }
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        /**
         * Indicates whether to emit type metadata in the new format.
         */
        var USE_NEW_TYPE_METADATA_FORMAT = false;
        var TypeScriptSubstitutionFlags;
        (function (TypeScriptSubstitutionFlags) {
            /** Enables substitutions for decorated classes. */
            TypeScriptSubstitutionFlags[TypeScriptSubstitutionFlags["ClassAliases"] = 1] = "ClassAliases";
            /** Enables substitutions for namespace exports. */
            TypeScriptSubstitutionFlags[TypeScriptSubstitutionFlags["NamespaceExports"] = 2] = "NamespaceExports";
            /* Enables substitutions for unqualified enum members */
            TypeScriptSubstitutionFlags[TypeScriptSubstitutionFlags["NonQualifiedEnumMembers"] = 8] = "NonQualifiedEnumMembers";
        })(TypeScriptSubstitutionFlags || (TypeScriptSubstitutionFlags = {}));
        var ClassFacts;
        (function (ClassFacts) {
            ClassFacts[ClassFacts["None"] = 0] = "None";
            ClassFacts[ClassFacts["HasStaticInitializedProperties"] = 1] = "HasStaticInitializedProperties";
            ClassFacts[ClassFacts["HasConstructorDecorators"] = 2] = "HasConstructorDecorators";
            ClassFacts[ClassFacts["HasMemberDecorators"] = 4] = "HasMemberDecorators";
            ClassFacts[ClassFacts["IsExportOfNamespace"] = 8] = "IsExportOfNamespace";
            ClassFacts[ClassFacts["IsNamedExternalExport"] = 16] = "IsNamedExternalExport";
            ClassFacts[ClassFacts["IsDefaultExternalExport"] = 32] = "IsDefaultExternalExport";
            ClassFacts[ClassFacts["IsDerivedClass"] = 64] = "IsDerivedClass";
            ClassFacts[ClassFacts["UseImmediatelyInvokedFunctionExpression"] = 128] = "UseImmediatelyInvokedFunctionExpression";
            ClassFacts[ClassFacts["HasAnyDecorators"] = 6] = "HasAnyDecorators";
            ClassFacts[ClassFacts["NeedsName"] = 5] = "NeedsName";
            ClassFacts[ClassFacts["MayNeedImmediatelyInvokedFunctionExpression"] = 7] = "MayNeedImmediatelyInvokedFunctionExpression";
            ClassFacts[ClassFacts["IsExported"] = 56] = "IsExported";
        })(ClassFacts || (ClassFacts = {}));
        function transformTypeScript(context) {
            var factory = context.factory, emitHelpers = context.getEmitHelperFactory, startLexicalEnvironment = context.startLexicalEnvironment, resumeLexicalEnvironment = context.resumeLexicalEnvironment, endLexicalEnvironment = context.endLexicalEnvironment, hoistVariableDeclaration = context.hoistVariableDeclaration;
            var resolver = context.getEmitResolver();
            var compilerOptions = context.getCompilerOptions();
            var strictNullChecks = ts.getStrictOptionValue(compilerOptions, "strictNullChecks");
            var languageVersion = ts.getEmitScriptTarget(compilerOptions);
            var moduleKind = ts.getEmitModuleKind(compilerOptions);
            // Save the previous transformation hooks.
            var previousOnEmitNode = context.onEmitNode;
            var previousOnSubstituteNode = context.onSubstituteNode;
            // Set new transformation hooks.
            context.onEmitNode = onEmitNode;
            context.onSubstituteNode = onSubstituteNode;
            // Enable substitution for property/element access to emit const enum values.
            context.enableSubstitution(204 /* PropertyAccessExpression */);
            context.enableSubstitution(205 /* ElementAccessExpression */);
            // These variables contain state that changes as we descend into the tree.
            var currentSourceFile;
            var currentNamespace;
            var currentNamespaceContainerName;
            var currentLexicalScope;
            var currentNameScope;
            var currentScopeFirstDeclarationsOfName;
            var currentClassHasParameterProperties;
            /**
             * Keeps track of whether expression substitution has been enabled for specific edge cases.
             * They are persisted between each SourceFile transformation and should not be reset.
             */
            var enabledSubstitutions;
            /**
             * A map that keeps track of aliases created for classes with decorators to avoid issues
             * with the double-binding behavior of classes.
             */
            var classAliases;
            /**
             * Keeps track of whether we are within any containing namespaces when performing
             * just-in-time substitution while printing an expression identifier.
             */
            var applicableSubstitutions;
            return transformSourceFileOrBundle;
            function transformSourceFileOrBundle(node) {
                if (node.kind === 301 /* Bundle */) {
                    return transformBundle(node);
                }
                return transformSourceFile(node);
            }
            function transformBundle(node) {
                return factory.createBundle(node.sourceFiles.map(transformSourceFile), ts.mapDefined(node.prepends, function (prepend) {
                    if (prepend.kind === 303 /* InputFiles */) {
                        return ts.createUnparsedSourceFile(prepend, "js");
                    }
                    return prepend;
                }));
            }
            /**
             * Transform TypeScript-specific syntax in a SourceFile.
             *
             * @param node A SourceFile node.
             */
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                currentSourceFile = node;
                var visited = saveStateAndInvoke(node, visitSourceFile);
                ts.addEmitHelpers(visited, context.readEmitHelpers());
                currentSourceFile = undefined;
                return visited;
            }
            /**
             * Visits a node, saving and restoring state variables on the stack.
             *
             * @param node The node to visit.
             */
            function saveStateAndInvoke(node, f) {
                // Save state
                var savedCurrentScope = currentLexicalScope;
                var savedCurrentNameScope = currentNameScope;
                var savedCurrentScopeFirstDeclarationsOfName = currentScopeFirstDeclarationsOfName;
                var savedCurrentClassHasParameterProperties = currentClassHasParameterProperties;
                // Handle state changes before visiting a node.
                onBeforeVisitNode(node);
                var visited = f(node);
                // Restore state
                if (currentLexicalScope !== savedCurrentScope) {
                    currentScopeFirstDeclarationsOfName = savedCurrentScopeFirstDeclarationsOfName;
                }
                currentLexicalScope = savedCurrentScope;
                currentNameScope = savedCurrentNameScope;
                currentClassHasParameterProperties = savedCurrentClassHasParameterProperties;
                return visited;
            }
            /**
             * Performs actions that should always occur immediately before visiting a node.
             *
             * @param node The node to visit.
             */
            function onBeforeVisitNode(node) {
                switch (node.kind) {
                    case 300 /* SourceFile */:
                    case 261 /* CaseBlock */:
                    case 260 /* ModuleBlock */:
                    case 233 /* Block */:
                        currentLexicalScope = node;
                        currentNameScope = undefined;
                        currentScopeFirstDeclarationsOfName = undefined;
                        break;
                    case 255 /* ClassDeclaration */:
                    case 254 /* FunctionDeclaration */:
                        if (ts.hasSyntacticModifier(node, 2 /* Ambient */)) {
                            break;
                        }
                        // Record these declarations provided that they have a name.
                        if (node.name) {
                            recordEmittedDeclarationInScope(node);
                        }
                        else {
                            // These nodes should always have names unless they are default-exports;
                            // however, class declaration parsing allows for undefined names, so syntactically invalid
                            // programs may also have an undefined name.
                            ts.Debug.assert(node.kind === 255 /* ClassDeclaration */ || ts.hasSyntacticModifier(node, 512 /* Default */));
                        }
                        if (ts.isClassDeclaration(node)) {
                            // XXX: should probably also cover interfaces and type aliases that can have type variables?
                            currentNameScope = node;
                        }
                        break;
                }
            }
            /**
             * General-purpose node visitor.
             *
             * @param node The node to visit.
             */
            function visitor(node) {
                return saveStateAndInvoke(node, visitorWorker);
            }
            /**
             * Visits and possibly transforms any node.
             *
             * @param node The node to visit.
             */
            function visitorWorker(node) {
                if (node.transformFlags & 1 /* ContainsTypeScript */) {
                    return visitTypeScript(node);
                }
                return node;
            }
            /**
             * Specialized visitor that visits the immediate children of a SourceFile.
             *
             * @param node The node to visit.
             */
            function sourceElementVisitor(node) {
                return saveStateAndInvoke(node, sourceElementVisitorWorker);
            }
            /**
             * Specialized visitor that visits the immediate children of a SourceFile.
             *
             * @param node The node to visit.
             */
            function sourceElementVisitorWorker(node) {
                switch (node.kind) {
                    case 264 /* ImportDeclaration */:
                    case 263 /* ImportEqualsDeclaration */:
                    case 269 /* ExportAssignment */:
                    case 270 /* ExportDeclaration */:
                        return visitElidableStatement(node);
                    default:
                        return visitorWorker(node);
                }
            }
            function visitElidableStatement(node) {
                var parsed = ts.getParseTreeNode(node);
                if (parsed !== node) {
                    // If the node has been transformed by a `before` transformer, perform no ellision on it
                    // As the type information we would attempt to lookup to perform ellision is potentially unavailable for the synthesized nodes
                    // We do not reuse `visitorWorker`, as the ellidable statement syntax kinds are technically unrecognized by the switch-case in `visitTypeScript`,
                    // and will trigger debug failures when debug verbosity is turned up
                    if (node.transformFlags & 1 /* ContainsTypeScript */) {
                        // This node contains TypeScript, so we should visit its children.
                        return ts.visitEachChild(node, visitor, context);
                    }
                    // Otherwise, we can just return the node
                    return node;
                }
                switch (node.kind) {
                    case 264 /* ImportDeclaration */:
                        return visitImportDeclaration(node);
                    case 263 /* ImportEqualsDeclaration */:
                        return visitImportEqualsDeclaration(node);
                    case 269 /* ExportAssignment */:
                        return visitExportAssignment(node);
                    case 270 /* ExportDeclaration */:
                        return visitExportDeclaration(node);
                    default:
                        ts.Debug.fail("Unhandled ellided statement");
                }
            }
            /**
             * Specialized visitor that visits the immediate children of a namespace.
             *
             * @param node The node to visit.
             */
            function namespaceElementVisitor(node) {
                return saveStateAndInvoke(node, namespaceElementVisitorWorker);
            }
            /**
             * Specialized visitor that visits the immediate children of a namespace.
             *
             * @param node The node to visit.
             */
            function namespaceElementVisitorWorker(node) {
                if (node.kind === 270 /* ExportDeclaration */ ||
                    node.kind === 264 /* ImportDeclaration */ ||
                    node.kind === 265 /* ImportClause */ ||
                    (node.kind === 263 /* ImportEqualsDeclaration */ &&
                        node.moduleReference.kind === 275 /* ExternalModuleReference */)) {
                    // do not emit ES6 imports and exports since they are illegal inside a namespace
                    return undefined;
                }
                else if (node.transformFlags & 1 /* ContainsTypeScript */ || ts.hasSyntacticModifier(node, 1 /* Export */)) {
                    return visitTypeScript(node);
                }
                return node;
            }
            /**
             * Specialized visitor that visits the immediate children of a class with TypeScript syntax.
             *
             * @param node The node to visit.
             */
            function classElementVisitor(node) {
                return saveStateAndInvoke(node, classElementVisitorWorker);
            }
            /**
             * Specialized visitor that visits the immediate children of a class with TypeScript syntax.
             *
             * @param node The node to visit.
             */
            function classElementVisitorWorker(node) {
                switch (node.kind) {
                    case 169 /* Constructor */:
                        return visitConstructor(node);
                    case 165 /* PropertyDeclaration */:
                        // Property declarations are not TypeScript syntax, but they must be visited
                        // for the decorator transformation.
                        return visitPropertyDeclaration(node);
                    case 174 /* IndexSignature */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 167 /* MethodDeclaration */:
                    case 168 /* ClassStaticBlockDeclaration */:
                        // Fallback to the default visit behavior.
                        return visitorWorker(node);
                    case 232 /* SemicolonClassElement */:
                        return node;
                    default:
                        return ts.Debug.failBadSyntaxKind(node);
                }
            }
            function modifierVisitor(node) {
                if (ts.modifierToFlag(node.kind) & 18654 /* TypeScriptModifier */) {
                    return undefined;
                }
                else if (currentNamespace && node.kind === 93 /* ExportKeyword */) {
                    return undefined;
                }
                return node;
            }
            /**
             * Branching visitor, visits a TypeScript syntax node.
             *
             * @param node The node to visit.
             */
            function visitTypeScript(node) {
                if (ts.isStatement(node) && ts.hasSyntacticModifier(node, 2 /* Ambient */)) {
                    // TypeScript ambient declarations are elided, but some comments may be preserved.
                    // See the implementation of `getLeadingComments` in comments.ts for more details.
                    return factory.createNotEmittedStatement(node);
                }
                switch (node.kind) {
                    case 93 /* ExportKeyword */:
                    case 88 /* DefaultKeyword */:
                        // ES6 export and default modifiers are elided when inside a namespace.
                        return currentNamespace ? undefined : node;
                    case 123 /* PublicKeyword */:
                    case 121 /* PrivateKeyword */:
                    case 122 /* ProtectedKeyword */:
                    case 126 /* AbstractKeyword */:
                    case 157 /* OverrideKeyword */:
                    case 85 /* ConstKeyword */:
                    case 134 /* DeclareKeyword */:
                    case 143 /* ReadonlyKeyword */:
                    // TypeScript accessibility and readonly modifiers are elided
                    // falls through
                    case 181 /* ArrayType */:
                    case 182 /* TupleType */:
                    case 183 /* OptionalType */:
                    case 184 /* RestType */:
                    case 180 /* TypeLiteral */:
                    case 175 /* TypePredicate */:
                    case 161 /* TypeParameter */:
                    case 129 /* AnyKeyword */:
                    case 153 /* UnknownKeyword */:
                    case 132 /* BooleanKeyword */:
                    case 148 /* StringKeyword */:
                    case 145 /* NumberKeyword */:
                    case 142 /* NeverKeyword */:
                    case 114 /* VoidKeyword */:
                    case 149 /* SymbolKeyword */:
                    case 178 /* ConstructorType */:
                    case 177 /* FunctionType */:
                    case 179 /* TypeQuery */:
                    case 176 /* TypeReference */:
                    case 185 /* UnionType */:
                    case 186 /* IntersectionType */:
                    case 187 /* ConditionalType */:
                    case 189 /* ParenthesizedType */:
                    case 190 /* ThisType */:
                    case 191 /* TypeOperator */:
                    case 192 /* IndexedAccessType */:
                    case 193 /* MappedType */:
                    case 194 /* LiteralType */:
                    // TypeScript type nodes are elided.
                    // falls through
                    case 174 /* IndexSignature */:
                    // TypeScript index signatures are elided.
                    // falls through
                    case 163 /* Decorator */:
                        // TypeScript decorators are elided. They will be emitted as part of visitClassDeclaration.
                        return undefined;
                    case 257 /* TypeAliasDeclaration */:
                        // TypeScript type-only declarations are elided.
                        return factory.createNotEmittedStatement(node);
                    case 165 /* PropertyDeclaration */:
                        // TypeScript property declarations are elided. However their names are still visited, and can potentially be retained if they could have sideeffects
                        return visitPropertyDeclaration(node);
                    case 262 /* NamespaceExportDeclaration */:
                        // TypeScript namespace export declarations are elided.
                        return undefined;
                    case 169 /* Constructor */:
                        return visitConstructor(node);
                    case 256 /* InterfaceDeclaration */:
                        // TypeScript interfaces are elided, but some comments may be preserved.
                        // See the implementation of `getLeadingComments` in comments.ts for more details.
                        return factory.createNotEmittedStatement(node);
                    case 255 /* ClassDeclaration */:
                        // This may be a class declaration with TypeScript syntax extensions.
                        //
                        // TypeScript class syntax extensions include:
                        // - decorators
                        // - optional `implements` heritage clause
                        // - parameter property assignments in the constructor
                        // - index signatures
                        // - method overload signatures
                        return visitClassDeclaration(node);
                    case 224 /* ClassExpression */:
                        // This may be a class expression with TypeScript syntax extensions.
                        //
                        // TypeScript class syntax extensions include:
                        // - decorators
                        // - optional `implements` heritage clause
                        // - parameter property assignments in the constructor
                        // - index signatures
                        // - method overload signatures
                        return visitClassExpression(node);
                    case 289 /* HeritageClause */:
                        // This may be a heritage clause with TypeScript syntax extensions.
                        //
                        // TypeScript heritage clause extensions include:
                        // - `implements` clause
                        return visitHeritageClause(node);
                    case 226 /* ExpressionWithTypeArguments */:
                        // TypeScript supports type arguments on an expression in an `extends` heritage clause.
                        return visitExpressionWithTypeArguments(node);
                    case 167 /* MethodDeclaration */:
                        // TypeScript method declarations may have decorators, modifiers
                        // or type annotations.
                        return visitMethodDeclaration(node);
                    case 170 /* GetAccessor */:
                        // Get Accessors can have TypeScript modifiers, decorators, and type annotations.
                        return visitGetAccessor(node);
                    case 171 /* SetAccessor */:
                        // Set Accessors can have TypeScript modifiers and type annotations.
                        return visitSetAccessor(node);
                    case 254 /* FunctionDeclaration */:
                        // Typescript function declarations can have modifiers, decorators, and type annotations.
                        return visitFunctionDeclaration(node);
                    case 211 /* FunctionExpression */:
                        // TypeScript function expressions can have modifiers and type annotations.
                        return visitFunctionExpression(node);
                    case 212 /* ArrowFunction */:
                        // TypeScript arrow functions can have modifiers and type annotations.
                        return visitArrowFunction(node);
                    case 162 /* Parameter */:
                        // This may be a parameter declaration with TypeScript syntax extensions.
                        //
                        // TypeScript parameter declaration syntax extensions include:
                        // - decorators
                        // - accessibility modifiers
                        // - the question mark (?) token for optional parameters
                        // - type annotations
                        // - this parameters
                        return visitParameter(node);
                    case 210 /* ParenthesizedExpression */:
                        // ParenthesizedExpressions are TypeScript if their expression is a
                        // TypeAssertion or AsExpression
                        return visitParenthesizedExpression(node);
                    case 209 /* TypeAssertionExpression */:
                    case 227 /* AsExpression */:
                        // TypeScript type assertions are removed, but their subtrees are preserved.
                        return visitAssertionExpression(node);
                    case 206 /* CallExpression */:
                        return visitCallExpression(node);
                    case 207 /* NewExpression */:
                        return visitNewExpression(node);
                    case 208 /* TaggedTemplateExpression */:
                        return visitTaggedTemplateExpression(node);
                    case 228 /* NonNullExpression */:
                        // TypeScript non-null expressions are removed, but their subtrees are preserved.
                        return visitNonNullExpression(node);
                    case 258 /* EnumDeclaration */:
                        // TypeScript enum declarations do not exist in ES6 and must be rewritten.
                        return visitEnumDeclaration(node);
                    case 235 /* VariableStatement */:
                        // TypeScript namespace exports for variable statements must be transformed.
                        return visitVariableStatement(node);
                    case 252 /* VariableDeclaration */:
                        return visitVariableDeclaration(node);
                    case 259 /* ModuleDeclaration */:
                        // TypeScript namespace declarations must be transformed.
                        return visitModuleDeclaration(node);
                    case 263 /* ImportEqualsDeclaration */:
                        // TypeScript namespace or external module import.
                        return visitImportEqualsDeclaration(node);
                    case 277 /* JsxSelfClosingElement */:
                        return visitJsxSelfClosingElement(node);
                    case 278 /* JsxOpeningElement */:
                        return visitJsxJsxOpeningElement(node);
                    default:
                        // node contains some other TypeScript syntax
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function visitSourceFile(node) {
                var alwaysStrict = ts.getStrictOptionValue(compilerOptions, "alwaysStrict") &&
                    !(ts.isExternalModule(node) && moduleKind >= ts.ModuleKind.ES2015) &&
                    !ts.isJsonSourceFile(node);
                return factory.updateSourceFile(node, ts.visitLexicalEnvironment(node.statements, sourceElementVisitor, context, /*start*/ 0, alwaysStrict));
            }
            function getClassFacts(node, staticProperties) {
                var facts = 0 /* None */;
                if (ts.some(staticProperties))
                    facts |= 1 /* HasStaticInitializedProperties */;
                var extendsClauseElement = ts.getEffectiveBaseTypeNode(node);
                if (extendsClauseElement && ts.skipOuterExpressions(extendsClauseElement.expression).kind !== 104 /* NullKeyword */)
                    facts |= 64 /* IsDerivedClass */;
                if (ts.classOrConstructorParameterIsDecorated(node))
                    facts |= 2 /* HasConstructorDecorators */;
                if (ts.childIsDecorated(node))
                    facts |= 4 /* HasMemberDecorators */;
                if (isExportOfNamespace(node))
                    facts |= 8 /* IsExportOfNamespace */;
                else if (isDefaultExternalModuleExport(node))
                    facts |= 32 /* IsDefaultExternalExport */;
                else if (isNamedExternalModuleExport(node))
                    facts |= 16 /* IsNamedExternalExport */;
                if (languageVersion <= 1 /* ES5 */ && (facts & 7 /* MayNeedImmediatelyInvokedFunctionExpression */))
                    facts |= 128 /* UseImmediatelyInvokedFunctionExpression */;
                return facts;
            }
            function hasTypeScriptClassSyntax(node) {
                return !!(node.transformFlags & 4096 /* ContainsTypeScriptClassSyntax */);
            }
            function isClassLikeDeclarationWithTypeScriptSyntax(node) {
                return ts.some(node.decorators)
                    || ts.some(node.typeParameters)
                    || ts.some(node.heritageClauses, hasTypeScriptClassSyntax)
                    || ts.some(node.members, hasTypeScriptClassSyntax);
            }
            function visitClassDeclaration(node) {
                if (!isClassLikeDeclarationWithTypeScriptSyntax(node) && !(currentNamespace && ts.hasSyntacticModifier(node, 1 /* Export */))) {
                    return ts.visitEachChild(node, visitor, context);
                }
                var staticProperties = ts.getProperties(node, /*requireInitializer*/ true, /*isStatic*/ true);
                var facts = getClassFacts(node, staticProperties);
                if (facts & 128 /* UseImmediatelyInvokedFunctionExpression */) {
                    context.startLexicalEnvironment();
                }
                var name = node.name || (facts & 5 /* NeedsName */ ? factory.getGeneratedNameForNode(node) : undefined);
                var classStatement = facts & 2 /* HasConstructorDecorators */
                    ? createClassDeclarationHeadWithDecorators(node, name)
                    : createClassDeclarationHeadWithoutDecorators(node, name, facts);
                var statements = [classStatement];
                // Write any decorators of the node.
                addClassElementDecorationStatements(statements, node, /*isStatic*/ false);
                addClassElementDecorationStatements(statements, node, /*isStatic*/ true);
                addConstructorDecorationStatement(statements, node);
                if (facts & 128 /* UseImmediatelyInvokedFunctionExpression */) {
                    // When we emit a TypeScript class down to ES5, we must wrap it in an IIFE so that the
                    // 'es2015' transformer can properly nest static initializers and decorators. The result
                    // looks something like:
                    //
                    //  var C = function () {
                    //      class C {
                    //      }
                    //      C.static_prop = 1;
                    //      return C;
                    //  }();
                    //
                    var closingBraceLocation = ts.createTokenRange(ts.skipTrivia(currentSourceFile.text, node.members.end), 19 /* CloseBraceToken */);
                    var localName = factory.getInternalName(node);
                    // The following partially-emitted expression exists purely to align our sourcemap
                    // emit with the original emitter.
                    var outer = factory.createPartiallyEmittedExpression(localName);
                    ts.setTextRangeEnd(outer, closingBraceLocation.end);
                    ts.setEmitFlags(outer, 1536 /* NoComments */);
                    var statement = factory.createReturnStatement(outer);
                    ts.setTextRangePos(statement, closingBraceLocation.pos);
                    ts.setEmitFlags(statement, 1536 /* NoComments */ | 384 /* NoTokenSourceMaps */);
                    statements.push(statement);
                    ts.insertStatementsAfterStandardPrologue(statements, context.endLexicalEnvironment());
                    var iife = factory.createImmediatelyInvokedArrowFunction(statements);
                    ts.setEmitFlags(iife, 33554432 /* TypeScriptClassWrapper */);
                    var varStatement = factory.createVariableStatement(
                    /*modifiers*/ undefined, factory.createVariableDeclarationList([
                        factory.createVariableDeclaration(factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ false), 
                        /*exclamationToken*/ undefined, 
                        /*type*/ undefined, iife)
                    ]));
                    ts.setOriginalNode(varStatement, node);
                    ts.setCommentRange(varStatement, node);
                    ts.setSourceMapRange(varStatement, ts.moveRangePastDecorators(node));
                    ts.startOnNewLine(varStatement);
                    statements = [varStatement];
                }
                // If the class is exported as part of a TypeScript namespace, emit the namespace export.
                // Otherwise, if the class was exported at the top level and was decorated, emit an export
                // declaration or export default for the class.
                if (facts & 8 /* IsExportOfNamespace */) {
                    addExportMemberAssignment(statements, node);
                }
                else if (facts & 128 /* UseImmediatelyInvokedFunctionExpression */ || facts & 2 /* HasConstructorDecorators */) {
                    if (facts & 32 /* IsDefaultExternalExport */) {
                        statements.push(factory.createExportDefault(factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true)));
                    }
                    else if (facts & 16 /* IsNamedExternalExport */) {
                        statements.push(factory.createExternalModuleExport(factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true)));
                    }
                }
                if (statements.length > 1) {
                    // Add a DeclarationMarker as a marker for the end of the declaration
                    statements.push(factory.createEndOfDeclarationMarker(node));
                    ts.setEmitFlags(classStatement, ts.getEmitFlags(classStatement) | 4194304 /* HasEndOfDeclarationMarker */);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Transforms a non-decorated class declaration and appends the resulting statements.
             *
             * @param node A ClassDeclaration node.
             * @param name The name of the class.
             * @param facts Precomputed facts about the class.
             */
            function createClassDeclarationHeadWithoutDecorators(node, name, facts) {
                //  ${modifiers} class ${name} ${heritageClauses} {
                //      ${members}
                //  }
                // we do not emit modifiers on the declaration if we are emitting an IIFE
                var modifiers = !(facts & 128 /* UseImmediatelyInvokedFunctionExpression */)
                    ? ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier)
                    : undefined;
                var classDeclaration = factory.createClassDeclaration(
                /*decorators*/ undefined, modifiers, name, 
                /*typeParameters*/ undefined, ts.visitNodes(node.heritageClauses, visitor, ts.isHeritageClause), transformClassMembers(node));
                // To better align with the old emitter, we should not emit a trailing source map
                // entry if the class has static properties.
                var emitFlags = ts.getEmitFlags(node);
                if (facts & 1 /* HasStaticInitializedProperties */) {
                    emitFlags |= 32 /* NoTrailingSourceMap */;
                }
                ts.setTextRange(classDeclaration, node);
                ts.setOriginalNode(classDeclaration, node);
                ts.setEmitFlags(classDeclaration, emitFlags);
                return classDeclaration;
            }
            /**
             * Transforms a decorated class declaration and appends the resulting statements. If
             * the class requires an alias to avoid issues with double-binding, the alias is returned.
             */
            function createClassDeclarationHeadWithDecorators(node, name) {
                // When we emit an ES6 class that has a class decorator, we must tailor the
                // emit to certain specific cases.
                //
                // In the simplest case, we emit the class declaration as a let declaration, and
                // evaluate decorators after the close of the class body:
                //
                //  [Example 1]
                //  ---------------------------------------------------------------------
                //  TypeScript                      | Javascript
                //  ---------------------------------------------------------------------
                //  @dec                            | let C = class C {
                //  class C {                       | }
                //  }                               | C = __decorate([dec], C);
                //  ---------------------------------------------------------------------
                //  @dec                            | let C = class C {
                //  export class C {                | }
                //  }                               | C = __decorate([dec], C);
                //                                  | export { C };
                //  ---------------------------------------------------------------------
                //
                // If a class declaration contains a reference to itself *inside* of the class body,
                // this introduces two bindings to the class: One outside of the class body, and one
                // inside of the class body. If we apply decorators as in [Example 1] above, there
                // is the possibility that the decorator `dec` will return a new value for the
                // constructor, which would result in the binding inside of the class no longer
                // pointing to the same reference as the binding outside of the class.
                //
                // As a result, we must instead rewrite all references to the class *inside* of the
                // class body to instead point to a local temporary alias for the class:
                //
                //  [Example 2]
                //  ---------------------------------------------------------------------
                //  TypeScript                      | Javascript
                //  ---------------------------------------------------------------------
                //  @dec                            | let C = C_1 = class C {
                //  class C {                       |   static x() { return C_1.y; }
                //    static x() { return C.y; }    | }
                //    static y = 1;                 | C.y = 1;
                //  }                               | C = C_1 = __decorate([dec], C);
                //                                  | var C_1;
                //  ---------------------------------------------------------------------
                //  @dec                            | let C = class C {
                //  export class C {                |   static x() { return C_1.y; }
                //    static x() { return C.y; }    | }
                //    static y = 1;                 | C.y = 1;
                //  }                               | C = C_1 = __decorate([dec], C);
                //                                  | export { C };
                //                                  | var C_1;
                //  ---------------------------------------------------------------------
                //
                // If a class declaration is the default export of a module, we instead emit
                // the export after the decorated declaration:
                //
                //  [Example 3]
                //  ---------------------------------------------------------------------
                //  TypeScript                      | Javascript
                //  ---------------------------------------------------------------------
                //  @dec                            | let default_1 = class {
                //  export default class {          | }
                //  }                               | default_1 = __decorate([dec], default_1);
                //                                  | export default default_1;
                //  ---------------------------------------------------------------------
                //  @dec                            | let C = class C {
                //  export default class C {        | }
                //  }                               | C = __decorate([dec], C);
                //                                  | export default C;
                //  ---------------------------------------------------------------------
                //
                // If the class declaration is the default export and a reference to itself
                // inside of the class body, we must emit both an alias for the class *and*
                // move the export after the declaration:
                //
                //  [Example 4]
                //  ---------------------------------------------------------------------
                //  TypeScript                      | Javascript
                //  ---------------------------------------------------------------------
                //  @dec                            | let C = class C {
                //  export default class C {        |   static x() { return C_1.y; }
                //    static x() { return C.y; }    | }
                //    static y = 1;                 | C.y = 1;
                //  }                               | C = C_1 = __decorate([dec], C);
                //                                  | export default C;
                //                                  | var C_1;
                //  ---------------------------------------------------------------------
                //
                var location = ts.moveRangePastDecorators(node);
                var classAlias = getClassAliasIfNeeded(node);
                // When we transform to ES5/3 this will be moved inside an IIFE and should reference the name
                // without any block-scoped variable collision handling
                var declName = languageVersion <= 2 /* ES2015 */ ?
                    factory.getInternalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true) :
                    factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true);
                //  ... = class ${name} ${heritageClauses} {
                //      ${members}
                //  }
                var heritageClauses = ts.visitNodes(node.heritageClauses, visitor, ts.isHeritageClause);
                var members = transformClassMembers(node);
                var classExpression = factory.createClassExpression(/*decorators*/ undefined, /*modifiers*/ undefined, name, /*typeParameters*/ undefined, heritageClauses, members);
                ts.setOriginalNode(classExpression, node);
                ts.setTextRange(classExpression, location);
                //  let ${name} = ${classExpression} where name is either declaredName if the class doesn't contain self-reference
                //                                         or decoratedClassAlias if the class contain self-reference.
                var statement = factory.createVariableStatement(
                /*modifiers*/ undefined, factory.createVariableDeclarationList([
                    factory.createVariableDeclaration(declName, 
                    /*exclamationToken*/ undefined, 
                    /*type*/ undefined, classAlias ? factory.createAssignment(classAlias, classExpression) : classExpression)
                ], 1 /* Let */));
                ts.setOriginalNode(statement, node);
                ts.setTextRange(statement, location);
                ts.setCommentRange(statement, node);
                return statement;
            }
            function visitClassExpression(node) {
                if (!isClassLikeDeclarationWithTypeScriptSyntax(node)) {
                    return ts.visitEachChild(node, visitor, context);
                }
                var classExpression = factory.createClassExpression(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, node.name, 
                /*typeParameters*/ undefined, ts.visitNodes(node.heritageClauses, visitor, ts.isHeritageClause), transformClassMembers(node));
                ts.setOriginalNode(classExpression, node);
                ts.setTextRange(classExpression, node);
                return classExpression;
            }
            /**
             * Transforms the members of a class.
             *
             * @param node The current class.
             */
            function transformClassMembers(node) {
                var members = [];
                var constructor = ts.getFirstConstructorWithBody(node);
                var parametersWithPropertyAssignments = constructor &&
                    ts.filter(constructor.parameters, function (p) { return ts.isParameterPropertyDeclaration(p, constructor); });
                if (parametersWithPropertyAssignments) {
                    for (var _i = 0, parametersWithPropertyAssignments_1 = parametersWithPropertyAssignments; _i < parametersWithPropertyAssignments_1.length; _i++) {
                        var parameter = parametersWithPropertyAssignments_1[_i];
                        if (ts.isIdentifier(parameter.name)) {
                            members.push(ts.setOriginalNode(factory.createPropertyDeclaration(
                            /*decorators*/ undefined, 
                            /*modifiers*/ undefined, parameter.name, 
                            /*questionOrExclamationToken*/ undefined, 
                            /*type*/ undefined, 
                            /*initializer*/ undefined), parameter));
                        }
                    }
                }
                ts.addRange(members, ts.visitNodes(node.members, classElementVisitor, ts.isClassElement));
                return ts.setTextRange(factory.createNodeArray(members), /*location*/ node.members);
            }
            /**
             * Gets either the static or instance members of a class that are decorated, or have
             * parameters that are decorated.
             *
             * @param node The class containing the member.
             * @param isStatic A value indicating whether to retrieve static or instance members of
             *                 the class.
             */
            function getDecoratedClassElements(node, isStatic) {
                return ts.filter(node.members, isStatic ? function (m) { return isStaticDecoratedClassElement(m, node); } : function (m) { return isInstanceDecoratedClassElement(m, node); });
            }
            /**
             * Determines whether a class member is a static member of a class that is decorated, or
             * has parameters that are decorated.
             *
             * @param member The class member.
             */
            function isStaticDecoratedClassElement(member, parent) {
                return isDecoratedClassElement(member, /*isStaticElement*/ true, parent);
            }
            /**
             * Determines whether a class member is an instance member of a class that is decorated,
             * or has parameters that are decorated.
             *
             * @param member The class member.
             */
            function isInstanceDecoratedClassElement(member, parent) {
                return isDecoratedClassElement(member, /*isStaticElement*/ false, parent);
            }
            /**
             * Determines whether a class member is either a static or an instance member of a class
             * that is decorated, or has parameters that are decorated.
             *
             * @param member The class member.
             */
            function isDecoratedClassElement(member, isStaticElement, parent) {
                return ts.nodeOrChildIsDecorated(member, parent)
                    && isStaticElement === ts.isStatic(member);
            }
            /**
             * Gets an array of arrays of decorators for the parameters of a function-like node.
             * The offset into the result array should correspond to the offset of the parameter.
             *
             * @param node The function-like node.
             */
            function getDecoratorsOfParameters(node) {
                var decorators;
                if (node) {
                    var parameters = node.parameters;
                    var firstParameterIsThis = parameters.length > 0 && ts.parameterIsThisKeyword(parameters[0]);
                    var firstParameterOffset = firstParameterIsThis ? 1 : 0;
                    var numParameters = firstParameterIsThis ? parameters.length - 1 : parameters.length;
                    for (var i = 0; i < numParameters; i++) {
                        var parameter = parameters[i + firstParameterOffset];
                        if (decorators || parameter.decorators) {
                            if (!decorators) {
                                decorators = new Array(numParameters);
                            }
                            decorators[i] = parameter.decorators;
                        }
                    }
                }
                return decorators;
            }
            /**
             * Gets an AllDecorators object containing the decorators for the class and the decorators for the
             * parameters of the constructor of the class.
             *
             * @param node The class node.
             */
            function getAllDecoratorsOfConstructor(node) {
                var decorators = node.decorators;
                var parameters = getDecoratorsOfParameters(ts.getFirstConstructorWithBody(node));
                if (!decorators && !parameters) {
                    return undefined;
                }
                return {
                    decorators: decorators,
                    parameters: parameters
                };
            }
            /**
             * Gets an AllDecorators object containing the decorators for the member and its parameters.
             *
             * @param node The class node that contains the member.
             * @param member The class member.
             */
            function getAllDecoratorsOfClassElement(node, member) {
                switch (member.kind) {
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return getAllDecoratorsOfAccessors(node, member);
                    case 167 /* MethodDeclaration */:
                        return getAllDecoratorsOfMethod(member);
                    case 165 /* PropertyDeclaration */:
                        return getAllDecoratorsOfProperty(member);
                    default:
                        return undefined;
                }
            }
            /**
             * Gets an AllDecorators object containing the decorators for the accessor and its parameters.
             *
             * @param node The class node that contains the accessor.
             * @param accessor The class accessor member.
             */
            function getAllDecoratorsOfAccessors(node, accessor) {
                if (!accessor.body) {
                    return undefined;
                }
                var _a = ts.getAllAccessorDeclarations(node.members, accessor), firstAccessor = _a.firstAccessor, secondAccessor = _a.secondAccessor, setAccessor = _a.setAccessor;
                var firstAccessorWithDecorators = firstAccessor.decorators ? firstAccessor : secondAccessor && secondAccessor.decorators ? secondAccessor : undefined;
                if (!firstAccessorWithDecorators || accessor !== firstAccessorWithDecorators) {
                    return undefined;
                }
                var decorators = firstAccessorWithDecorators.decorators;
                var parameters = getDecoratorsOfParameters(setAccessor);
                if (!decorators && !parameters) {
                    return undefined;
                }
                return { decorators: decorators, parameters: parameters };
            }
            /**
             * Gets an AllDecorators object containing the decorators for the method and its parameters.
             *
             * @param method The class method member.
             */
            function getAllDecoratorsOfMethod(method) {
                if (!method.body) {
                    return undefined;
                }
                var decorators = method.decorators;
                var parameters = getDecoratorsOfParameters(method);
                if (!decorators && !parameters) {
                    return undefined;
                }
                return { decorators: decorators, parameters: parameters };
            }
            /**
             * Gets an AllDecorators object containing the decorators for the property.
             *
             * @param property The class property member.
             */
            function getAllDecoratorsOfProperty(property) {
                var decorators = property.decorators;
                if (!decorators) {
                    return undefined;
                }
                return { decorators: decorators };
            }
            /**
             * Transforms all of the decorators for a declaration into an array of expressions.
             *
             * @param node The declaration node.
             * @param allDecorators An object containing all of the decorators for the declaration.
             */
            function transformAllDecoratorsOfDeclaration(node, container, allDecorators) {
                if (!allDecorators) {
                    return undefined;
                }
                var decoratorExpressions = [];
                ts.addRange(decoratorExpressions, ts.map(allDecorators.decorators, transformDecorator));
                ts.addRange(decoratorExpressions, ts.flatMap(allDecorators.parameters, transformDecoratorsOfParameter));
                addTypeMetadata(node, container, decoratorExpressions);
                return decoratorExpressions;
            }
            /**
             * Generates statements used to apply decorators to either the static or instance members
             * of a class.
             *
             * @param node The class node.
             * @param isStatic A value indicating whether to generate statements for static or
             *                 instance members.
             */
            function addClassElementDecorationStatements(statements, node, isStatic) {
                ts.addRange(statements, ts.map(generateClassElementDecorationExpressions(node, isStatic), expressionToStatement));
            }
            /**
             * Generates expressions used to apply decorators to either the static or instance members
             * of a class.
             *
             * @param node The class node.
             * @param isStatic A value indicating whether to generate expressions for static or
             *                 instance members.
             */
            function generateClassElementDecorationExpressions(node, isStatic) {
                var members = getDecoratedClassElements(node, isStatic);
                var expressions;
                for (var _i = 0, members_6 = members; _i < members_6.length; _i++) {
                    var member = members_6[_i];
                    var expression = generateClassElementDecorationExpression(node, member);
                    if (expression) {
                        if (!expressions) {
                            expressions = [expression];
                        }
                        else {
                            expressions.push(expression);
                        }
                    }
                }
                return expressions;
            }
            /**
             * Generates an expression used to evaluate class element decorators at runtime.
             *
             * @param node The class node that contains the member.
             * @param member The class member.
             */
            function generateClassElementDecorationExpression(node, member) {
                var allDecorators = getAllDecoratorsOfClassElement(node, member);
                var decoratorExpressions = transformAllDecoratorsOfDeclaration(member, node, allDecorators);
                if (!decoratorExpressions) {
                    return undefined;
                }
                // Emit the call to __decorate. Given the following:
                //
                //   class C {
                //     @dec method(@dec2 x) {}
                //     @dec get accessor() {}
                //     @dec prop;
                //   }
                //
                // The emit for a method is:
                //
                //   __decorate([
                //       dec,
                //       __param(0, dec2),
                //       __metadata("design:type", Function),
                //       __metadata("design:paramtypes", [Object]),
                //       __metadata("design:returntype", void 0)
                //   ], C.prototype, "method", null);
                //
                // The emit for an accessor is:
                //
                //   __decorate([
                //       dec
                //   ], C.prototype, "accessor", null);
                //
                // The emit for a property is:
                //
                //   __decorate([
                //       dec
                //   ], C.prototype, "prop");
                //
                var prefix = getClassMemberPrefix(node, member);
                var memberName = getExpressionForPropertyName(member, /*generateNameForComputedPropertyName*/ true);
                var descriptor = languageVersion > 0 /* ES3 */
                    ? member.kind === 165 /* PropertyDeclaration */
                        // We emit `void 0` here to indicate to `__decorate` that it can invoke `Object.defineProperty` directly, but that it
                        // should not invoke `Object.getOwnPropertyDescriptor`.
                        ? factory.createVoidZero()
                        // We emit `null` here to indicate to `__decorate` that it can invoke `Object.getOwnPropertyDescriptor` directly.
                        // We have this extra argument here so that we can inject an explicit property descriptor at a later date.
                        : factory.createNull()
                    : undefined;
                var helper = emitHelpers().createDecorateHelper(decoratorExpressions, prefix, memberName, descriptor);
                ts.setTextRange(helper, ts.moveRangePastDecorators(member));
                ts.setEmitFlags(helper, 1536 /* NoComments */);
                return helper;
            }
            /**
             * Generates a __decorate helper call for a class constructor.
             *
             * @param node The class node.
             */
            function addConstructorDecorationStatement(statements, node) {
                var expression = generateConstructorDecorationExpression(node);
                if (expression) {
                    statements.push(ts.setOriginalNode(factory.createExpressionStatement(expression), node));
                }
            }
            /**
             * Generates a __decorate helper call for a class constructor.
             *
             * @param node The class node.
             */
            function generateConstructorDecorationExpression(node) {
                var allDecorators = getAllDecoratorsOfConstructor(node);
                var decoratorExpressions = transformAllDecoratorsOfDeclaration(node, node, allDecorators);
                if (!decoratorExpressions) {
                    return undefined;
                }
                var classAlias = classAliases && classAliases[ts.getOriginalNodeId(node)];
                // When we transform to ES5/3 this will be moved inside an IIFE and should reference the name
                // without any block-scoped variable collision handling
                var localName = languageVersion <= 2 /* ES2015 */ ?
                    factory.getInternalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true) :
                    factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true);
                var decorate = emitHelpers().createDecorateHelper(decoratorExpressions, localName);
                var expression = factory.createAssignment(localName, classAlias ? factory.createAssignment(classAlias, decorate) : decorate);
                ts.setEmitFlags(expression, 1536 /* NoComments */);
                ts.setSourceMapRange(expression, ts.moveRangePastDecorators(node));
                return expression;
            }
            /**
             * Transforms a decorator into an expression.
             *
             * @param decorator The decorator node.
             */
            function transformDecorator(decorator) {
                return ts.visitNode(decorator.expression, visitor, ts.isExpression);
            }
            /**
             * Transforms the decorators of a parameter.
             *
             * @param decorators The decorators for the parameter at the provided offset.
             * @param parameterOffset The offset of the parameter.
             */
            function transformDecoratorsOfParameter(decorators, parameterOffset) {
                var expressions;
                if (decorators) {
                    expressions = [];
                    for (var _i = 0, decorators_1 = decorators; _i < decorators_1.length; _i++) {
                        var decorator = decorators_1[_i];
                        var helper = emitHelpers().createParamHelper(transformDecorator(decorator), parameterOffset);
                        ts.setTextRange(helper, decorator.expression);
                        ts.setEmitFlags(helper, 1536 /* NoComments */);
                        expressions.push(helper);
                    }
                }
                return expressions;
            }
            /**
             * Adds optional type metadata for a declaration.
             *
             * @param node The declaration node.
             * @param decoratorExpressions The destination array to which to add new decorator expressions.
             */
            function addTypeMetadata(node, container, decoratorExpressions) {
                if (USE_NEW_TYPE_METADATA_FORMAT) {
                    addNewTypeMetadata(node, container, decoratorExpressions);
                }
                else {
                    addOldTypeMetadata(node, container, decoratorExpressions);
                }
            }
            function addOldTypeMetadata(node, container, decoratorExpressions) {
                if (compilerOptions.emitDecoratorMetadata) {
                    if (shouldAddTypeMetadata(node)) {
                        decoratorExpressions.push(emitHelpers().createMetadataHelper("design:type", serializeTypeOfNode(node)));
                    }
                    if (shouldAddParamTypesMetadata(node)) {
                        decoratorExpressions.push(emitHelpers().createMetadataHelper("design:paramtypes", serializeParameterTypesOfNode(node, container)));
                    }
                    if (shouldAddReturnTypeMetadata(node)) {
                        decoratorExpressions.push(emitHelpers().createMetadataHelper("design:returntype", serializeReturnTypeOfNode(node)));
                    }
                }
            }
            function addNewTypeMetadata(node, container, decoratorExpressions) {
                if (compilerOptions.emitDecoratorMetadata) {
                    var properties = void 0;
                    if (shouldAddTypeMetadata(node)) {
                        (properties || (properties = [])).push(factory.createPropertyAssignment("type", factory.createArrowFunction(/*modifiers*/ undefined, /*typeParameters*/ undefined, [], /*type*/ undefined, factory.createToken(38 /* EqualsGreaterThanToken */), serializeTypeOfNode(node))));
                    }
                    if (shouldAddParamTypesMetadata(node)) {
                        (properties || (properties = [])).push(factory.createPropertyAssignment("paramTypes", factory.createArrowFunction(/*modifiers*/ undefined, /*typeParameters*/ undefined, [], /*type*/ undefined, factory.createToken(38 /* EqualsGreaterThanToken */), serializeParameterTypesOfNode(node, container))));
                    }
                    if (shouldAddReturnTypeMetadata(node)) {
                        (properties || (properties = [])).push(factory.createPropertyAssignment("returnType", factory.createArrowFunction(/*modifiers*/ undefined, /*typeParameters*/ undefined, [], /*type*/ undefined, factory.createToken(38 /* EqualsGreaterThanToken */), serializeReturnTypeOfNode(node))));
                    }
                    if (properties) {
                        decoratorExpressions.push(emitHelpers().createMetadataHelper("design:typeinfo", factory.createObjectLiteralExpression(properties, /*multiLine*/ true)));
                    }
                }
            }
            /**
             * Determines whether to emit the "design:type" metadata based on the node's kind.
             * The caller should have already tested whether the node has decorators and whether the
             * emitDecoratorMetadata compiler option is set.
             *
             * @param node The node to test.
             */
            function shouldAddTypeMetadata(node) {
                var kind = node.kind;
                return kind === 167 /* MethodDeclaration */
                    || kind === 170 /* GetAccessor */
                    || kind === 171 /* SetAccessor */
                    || kind === 165 /* PropertyDeclaration */;
            }
            /**
             * Determines whether to emit the "design:returntype" metadata based on the node's kind.
             * The caller should have already tested whether the node has decorators and whether the
             * emitDecoratorMetadata compiler option is set.
             *
             * @param node The node to test.
             */
            function shouldAddReturnTypeMetadata(node) {
                return node.kind === 167 /* MethodDeclaration */;
            }
            /**
             * Determines whether to emit the "design:paramtypes" metadata based on the node's kind.
             * The caller should have already tested whether the node has decorators and whether the
             * emitDecoratorMetadata compiler option is set.
             *
             * @param node The node to test.
             */
            function shouldAddParamTypesMetadata(node) {
                switch (node.kind) {
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                        return ts.getFirstConstructorWithBody(node) !== undefined;
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return true;
                }
                return false;
            }
            function getAccessorTypeNode(node) {
                var accessors = resolver.getAllAccessorDeclarations(node);
                return accessors.setAccessor && ts.getSetAccessorTypeAnnotationNode(accessors.setAccessor)
                    || accessors.getAccessor && ts.getEffectiveReturnTypeNode(accessors.getAccessor);
            }
            /**
             * Serializes the type of a node for use with decorator type metadata.
             *
             * @param node The node that should have its type serialized.
             */
            function serializeTypeOfNode(node) {
                switch (node.kind) {
                    case 165 /* PropertyDeclaration */:
                    case 162 /* Parameter */:
                        return serializeTypeNode(node.type);
                    case 171 /* SetAccessor */:
                    case 170 /* GetAccessor */:
                        return serializeTypeNode(getAccessorTypeNode(node));
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                    case 167 /* MethodDeclaration */:
                        return factory.createIdentifier("Function");
                    default:
                        return factory.createVoidZero();
                }
            }
            /**
             * Serializes the types of the parameters of a node for use with decorator type metadata.
             *
             * @param node The node that should have its parameter types serialized.
             */
            function serializeParameterTypesOfNode(node, container) {
                var valueDeclaration = ts.isClassLike(node)
                    ? ts.getFirstConstructorWithBody(node)
                    : ts.isFunctionLike(node) && ts.nodeIsPresent(node.body)
                        ? node
                        : undefined;
                var expressions = [];
                if (valueDeclaration) {
                    var parameters = getParametersOfDecoratedDeclaration(valueDeclaration, container);
                    var numParameters = parameters.length;
                    for (var i = 0; i < numParameters; i++) {
                        var parameter = parameters[i];
                        if (i === 0 && ts.isIdentifier(parameter.name) && parameter.name.escapedText === "this") {
                            continue;
                        }
                        if (parameter.dotDotDotToken) {
                            expressions.push(serializeTypeNode(ts.getRestParameterElementType(parameter.type)));
                        }
                        else {
                            expressions.push(serializeTypeOfNode(parameter));
                        }
                    }
                }
                return factory.createArrayLiteralExpression(expressions);
            }
            function getParametersOfDecoratedDeclaration(node, container) {
                if (container && node.kind === 170 /* GetAccessor */) {
                    var setAccessor = ts.getAllAccessorDeclarations(container.members, node).setAccessor;
                    if (setAccessor) {
                        return setAccessor.parameters;
                    }
                }
                return node.parameters;
            }
            /**
             * Serializes the return type of a node for use with decorator type metadata.
             *
             * @param node The node that should have its return type serialized.
             */
            function serializeReturnTypeOfNode(node) {
                if (ts.isFunctionLike(node) && node.type) {
                    return serializeTypeNode(node.type);
                }
                else if (ts.isAsyncFunction(node)) {
                    return factory.createIdentifier("Promise");
                }
                return factory.createVoidZero();
            }
            /**
             * Serializes a type node for use with decorator type metadata.
             *
             * Types are serialized in the following fashion:
             * - Void types point to "undefined" (e.g. "void 0")
             * - Function and Constructor types point to the global "Function" constructor.
             * - Interface types with a call or construct signature types point to the global
             *   "Function" constructor.
             * - Array and Tuple types point to the global "Array" constructor.
             * - Type predicates and booleans point to the global "Boolean" constructor.
             * - String literal types and strings point to the global "String" constructor.
             * - Enum and number types point to the global "Number" constructor.
             * - Symbol types point to the global "Symbol" constructor.
             * - Type references to classes (or class-like variables) point to the constructor for the class.
             * - Anything else points to the global "Object" constructor.
             *
             * @param node The type node to serialize.
             */
            function serializeTypeNode(node) {
                if (node === undefined) {
                    return factory.createIdentifier("Object");
                }
                switch (node.kind) {
                    case 114 /* VoidKeyword */:
                    case 151 /* UndefinedKeyword */:
                    case 142 /* NeverKeyword */:
                        return factory.createVoidZero();
                    case 189 /* ParenthesizedType */:
                        return serializeTypeNode(node.type);
                    case 177 /* FunctionType */:
                    case 178 /* ConstructorType */:
                        return factory.createIdentifier("Function");
                    case 181 /* ArrayType */:
                    case 182 /* TupleType */:
                        return factory.createIdentifier("Array");
                    case 175 /* TypePredicate */:
                    case 132 /* BooleanKeyword */:
                        return factory.createIdentifier("Boolean");
                    case 148 /* StringKeyword */:
                        return factory.createIdentifier("String");
                    case 146 /* ObjectKeyword */:
                        return factory.createIdentifier("Object");
                    case 194 /* LiteralType */:
                        switch (node.literal.kind) {
                            case 10 /* StringLiteral */:
                            case 14 /* NoSubstitutionTemplateLiteral */:
                                return factory.createIdentifier("String");
                            case 217 /* PrefixUnaryExpression */:
                            case 8 /* NumericLiteral */:
                                return factory.createIdentifier("Number");
                            case 9 /* BigIntLiteral */:
                                return getGlobalBigIntNameWithFallback();
                            case 110 /* TrueKeyword */:
                            case 95 /* FalseKeyword */:
                                return factory.createIdentifier("Boolean");
                            case 104 /* NullKeyword */:
                                return factory.createVoidZero();
                            default:
                                return ts.Debug.failBadSyntaxKind(node.literal);
                        }
                    case 145 /* NumberKeyword */:
                        return factory.createIdentifier("Number");
                    case 156 /* BigIntKeyword */:
                        return getGlobalBigIntNameWithFallback();
                    case 149 /* SymbolKeyword */:
                        return languageVersion < 2 /* ES2015 */
                            ? getGlobalSymbolNameWithFallback()
                            : factory.createIdentifier("Symbol");
                    case 176 /* TypeReference */:
                        return serializeTypeReferenceNode(node);
                    case 186 /* IntersectionType */:
                    case 185 /* UnionType */:
                        return serializeTypeList(node.types);
                    case 187 /* ConditionalType */:
                        return serializeTypeList([node.trueType, node.falseType]);
                    case 191 /* TypeOperator */:
                        if (node.operator === 143 /* ReadonlyKeyword */) {
                            return serializeTypeNode(node.type);
                        }
                        break;
                    case 179 /* TypeQuery */:
                    case 192 /* IndexedAccessType */:
                    case 193 /* MappedType */:
                    case 180 /* TypeLiteral */:
                    case 129 /* AnyKeyword */:
                    case 153 /* UnknownKeyword */:
                    case 190 /* ThisType */:
                    case 198 /* ImportType */:
                        break;
                    // handle JSDoc types from an invalid parse
                    case 307 /* JSDocAllType */:
                    case 308 /* JSDocUnknownType */:
                    case 312 /* JSDocFunctionType */:
                    case 313 /* JSDocVariadicType */:
                    case 314 /* JSDocNamepathType */:
                        break;
                    case 309 /* JSDocNullableType */:
                    case 310 /* JSDocNonNullableType */:
                    case 311 /* JSDocOptionalType */:
                        return serializeTypeNode(node.type);
                    default:
                        return ts.Debug.failBadSyntaxKind(node);
                }
                return factory.createIdentifier("Object");
            }
            function serializeTypeList(types) {
                // Note when updating logic here also update getEntityNameForDecoratorMetadata
                // so that aliases can be marked as referenced
                var serializedUnion;
                for (var _i = 0, types_23 = types; _i < types_23.length; _i++) {
                    var typeNode = types_23[_i];
                    while (typeNode.kind === 189 /* ParenthesizedType */) {
                        typeNode = typeNode.type; // Skip parens if need be
                    }
                    if (typeNode.kind === 142 /* NeverKeyword */) {
                        continue; // Always elide `never` from the union/intersection if possible
                    }
                    if (!strictNullChecks && (typeNode.kind === 194 /* LiteralType */ && typeNode.literal.kind === 104 /* NullKeyword */ || typeNode.kind === 151 /* UndefinedKeyword */)) {
                        continue; // Elide null and undefined from unions for metadata, just like what we did prior to the implementation of strict null checks
                    }
                    var serializedIndividual = serializeTypeNode(typeNode);
                    if (ts.isIdentifier(serializedIndividual) && serializedIndividual.escapedText === "Object") {
                        // One of the individual is global object, return immediately
                        return serializedIndividual;
                    }
                    // If there exists union that is not void 0 expression, check if the the common type is identifier.
                    // anything more complex and we will just default to Object
                    else if (serializedUnion) {
                        // Different types
                        if (!ts.isIdentifier(serializedUnion) ||
                            !ts.isIdentifier(serializedIndividual) ||
                            serializedUnion.escapedText !== serializedIndividual.escapedText) {
                            return factory.createIdentifier("Object");
                        }
                    }
                    else {
                        // Initialize the union type
                        serializedUnion = serializedIndividual;
                    }
                }
                // If we were able to find common type, use it
                return serializedUnion || factory.createVoidZero(); // Fallback is only hit if all union constituients are null/undefined/never
            }
            /**
             * Serializes a TypeReferenceNode to an appropriate JS constructor value for use with
             * decorator type metadata.
             *
             * @param node The type reference node.
             */
            function serializeTypeReferenceNode(node) {
                var kind = resolver.getTypeReferenceSerializationKind(node.typeName, currentNameScope || currentLexicalScope);
                switch (kind) {
                    case ts.TypeReferenceSerializationKind.Unknown:
                        // From conditional type type reference that cannot be resolved is Similar to any or unknown
                        if (ts.findAncestor(node, function (n) { return n.parent && ts.isConditionalTypeNode(n.parent) && (n.parent.trueType === n || n.parent.falseType === n); })) {
                            return factory.createIdentifier("Object");
                        }
                        var serialized = serializeEntityNameAsExpressionFallback(node.typeName);
                        var temp = factory.createTempVariable(hoistVariableDeclaration);
                        return factory.createConditionalExpression(factory.createTypeCheck(factory.createAssignment(temp, serialized), "function"), 
                        /*questionToken*/ undefined, temp, 
                        /*colonToken*/ undefined, factory.createIdentifier("Object"));
                    case ts.TypeReferenceSerializationKind.TypeWithConstructSignatureAndValue:
                        return serializeEntityNameAsExpression(node.typeName);
                    case ts.TypeReferenceSerializationKind.VoidNullableOrNeverType:
                        return factory.createVoidZero();
                    case ts.TypeReferenceSerializationKind.BigIntLikeType:
                        return getGlobalBigIntNameWithFallback();
                    case ts.TypeReferenceSerializationKind.BooleanType:
                        return factory.createIdentifier("Boolean");
                    case ts.TypeReferenceSerializationKind.NumberLikeType:
                        return factory.createIdentifier("Number");
                    case ts.TypeReferenceSerializationKind.StringLikeType:
                        return factory.createIdentifier("String");
                    case ts.TypeReferenceSerializationKind.ArrayLikeType:
                        return factory.createIdentifier("Array");
                    case ts.TypeReferenceSerializationKind.ESSymbolType:
                        return languageVersion < 2 /* ES2015 */
                            ? getGlobalSymbolNameWithFallback()
                            : factory.createIdentifier("Symbol");
                    case ts.TypeReferenceSerializationKind.TypeWithCallSignature:
                        return factory.createIdentifier("Function");
                    case ts.TypeReferenceSerializationKind.Promise:
                        return factory.createIdentifier("Promise");
                    case ts.TypeReferenceSerializationKind.ObjectType:
                        return factory.createIdentifier("Object");
                    default:
                        return ts.Debug.assertNever(kind);
                }
            }
            function createCheckedValue(left, right) {
                return factory.createLogicalAnd(factory.createStrictInequality(factory.createTypeOfExpression(left), factory.createStringLiteral("undefined")), right);
            }
            /**
             * Serializes an entity name which may not exist at runtime, but whose access shouldn't throw
             *
             * @param node The entity name to serialize.
             */
            function serializeEntityNameAsExpressionFallback(node) {
                if (node.kind === 79 /* Identifier */) {
                    // A -> typeof A !== undefined && A
                    var copied = serializeEntityNameAsExpression(node);
                    return createCheckedValue(copied, copied);
                }
                if (node.left.kind === 79 /* Identifier */) {
                    // A.B -> typeof A !== undefined && A.B
                    return createCheckedValue(serializeEntityNameAsExpression(node.left), serializeEntityNameAsExpression(node));
                }
                // A.B.C -> typeof A !== undefined && (_a = A.B) !== void 0 && _a.C
                var left = serializeEntityNameAsExpressionFallback(node.left);
                var temp = factory.createTempVariable(hoistVariableDeclaration);
                return factory.createLogicalAnd(factory.createLogicalAnd(left.left, factory.createStrictInequality(factory.createAssignment(temp, left.right), factory.createVoidZero())), factory.createPropertyAccessExpression(temp, node.right));
            }
            /**
             * Serializes an entity name as an expression for decorator type metadata.
             *
             * @param node The entity name to serialize.
             */
            function serializeEntityNameAsExpression(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        // Create a clone of the name with a new parent, and treat it as if it were
                        // a source tree node for the purposes of the checker.
                        var name = ts.setParent(ts.setTextRange(ts.parseNodeFactory.cloneNode(node), node), node.parent);
                        name.original = undefined;
                        ts.setParent(name, ts.getParseTreeNode(currentLexicalScope)); // ensure the parent is set to a parse tree node.
                        return name;
                    case 159 /* QualifiedName */:
                        return serializeQualifiedNameAsExpression(node);
                }
            }
            /**
             * Serializes an qualified name as an expression for decorator type metadata.
             *
             * @param node The qualified name to serialize.
             * @param useFallback A value indicating whether to use logical operators to test for the
             *                    qualified name at runtime.
             */
            function serializeQualifiedNameAsExpression(node) {
                return factory.createPropertyAccessExpression(serializeEntityNameAsExpression(node.left), node.right);
            }
            /**
             * Gets an expression that points to the global "Symbol" constructor at runtime if it is
             * available.
             */
            function getGlobalSymbolNameWithFallback() {
                return factory.createConditionalExpression(factory.createTypeCheck(factory.createIdentifier("Symbol"), "function"), 
                /*questionToken*/ undefined, factory.createIdentifier("Symbol"), 
                /*colonToken*/ undefined, factory.createIdentifier("Object"));
            }
            /**
             * Gets an expression that points to the global "BigInt" constructor at runtime if it is
             * available.
             */
            function getGlobalBigIntNameWithFallback() {
                return languageVersion < 99 /* ESNext */
                    ? factory.createConditionalExpression(factory.createTypeCheck(factory.createIdentifier("BigInt"), "function"), 
                    /*questionToken*/ undefined, factory.createIdentifier("BigInt"), 
                    /*colonToken*/ undefined, factory.createIdentifier("Object"))
                    : factory.createIdentifier("BigInt");
            }
            /**
             * Gets an expression that represents a property name (for decorated properties or enums).
             * For a computed property, a name is generated for the node.
             *
             * @param member The member whose name should be converted into an expression.
             */
            function getExpressionForPropertyName(member, generateNameForComputedPropertyName) {
                var name = member.name;
                if (ts.isPrivateIdentifier(name)) {
                    return factory.createIdentifier("");
                }
                else if (ts.isComputedPropertyName(name)) {
                    return generateNameForComputedPropertyName && !ts.isSimpleInlineableExpression(name.expression)
                        ? factory.getGeneratedNameForNode(name)
                        : name.expression;
                }
                else if (ts.isIdentifier(name)) {
                    return factory.createStringLiteral(ts.idText(name));
                }
                else {
                    return factory.cloneNode(name);
                }
            }
            /**
             * Visits the property name of a class element, for use when emitting property
             * initializers. For a computed property on a node with decorators, a temporary
             * value is stored for later use.
             *
             * @param member The member whose name should be visited.
             */
            function visitPropertyNameOfClassElement(member) {
                var name = member.name;
                // Computed property names need to be transformed into a hoisted variable when they are used more than once.
                // The names are used more than once when:
                //   - the property is non-static and its initializer is moved to the constructor (when there are parameter property assignments).
                //   - the property has a decorator.
                if (ts.isComputedPropertyName(name) && ((!ts.hasStaticModifier(member) && currentClassHasParameterProperties) || ts.some(member.decorators))) {
                    var expression = ts.visitNode(name.expression, visitor, ts.isExpression);
                    var innerExpression = ts.skipPartiallyEmittedExpressions(expression);
                    if (!ts.isSimpleInlineableExpression(innerExpression)) {
                        var generatedName = factory.getGeneratedNameForNode(name);
                        hoistVariableDeclaration(generatedName);
                        return factory.updateComputedPropertyName(name, factory.createAssignment(generatedName, expression));
                    }
                }
                return ts.visitNode(name, visitor, ts.isPropertyName);
            }
            /**
             * Transforms a HeritageClause with TypeScript syntax.
             *
             * This function will only be called when one of the following conditions are met:
             * - The node is a non-`extends` heritage clause that should be elided.
             * - The node is an `extends` heritage clause that should be visited, but only allow a single type.
             *
             * @param node The HeritageClause to transform.
             */
            function visitHeritageClause(node) {
                if (node.token === 117 /* ImplementsKeyword */) {
                    // implements clauses are elided
                    return undefined;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Transforms an ExpressionWithTypeArguments with TypeScript syntax.
             *
             * This function will only be called when one of the following conditions are met:
             * - The node contains type arguments that should be elided.
             *
             * @param node The ExpressionWithTypeArguments to transform.
             */
            function visitExpressionWithTypeArguments(node) {
                return factory.updateExpressionWithTypeArguments(node, ts.visitNode(node.expression, visitor, ts.isLeftHandSideExpression), 
                /*typeArguments*/ undefined);
            }
            /**
             * Determines whether to emit a function-like declaration. We should not emit the
             * declaration if it does not have a body.
             *
             * @param node The declaration node.
             */
            function shouldEmitFunctionLikeDeclaration(node) {
                return !ts.nodeIsMissing(node.body);
            }
            function visitPropertyDeclaration(node) {
                if (node.flags & 8388608 /* Ambient */ || ts.hasSyntacticModifier(node, 128 /* Abstract */)) {
                    return undefined;
                }
                var updated = factory.updatePropertyDeclaration(node, 
                /*decorators*/ undefined, ts.visitNodes(node.modifiers, visitor, ts.isModifier), visitPropertyNameOfClassElement(node), 
                /*questionOrExclamationToken*/ undefined, 
                /*type*/ undefined, ts.visitNode(node.initializer, visitor));
                if (updated !== node) {
                    // While we emit the source map for the node after skipping decorators and modifiers,
                    // we need to emit the comments for the original range.
                    ts.setCommentRange(updated, node);
                    ts.setSourceMapRange(updated, ts.moveRangePastDecorators(node));
                }
                return updated;
            }
            function visitConstructor(node) {
                if (!shouldEmitFunctionLikeDeclaration(node)) {
                    return undefined;
                }
                return factory.updateConstructorDeclaration(node, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, ts.visitParameterList(node.parameters, visitor, context), transformConstructorBody(node.body, node));
            }
            function transformConstructorBody(body, constructor) {
                var parametersWithPropertyAssignments = constructor &&
                    ts.filter(constructor.parameters, function (p) { return ts.isParameterPropertyDeclaration(p, constructor); });
                if (!ts.some(parametersWithPropertyAssignments)) {
                    return ts.visitFunctionBody(body, visitor, context);
                }
                var statements = [];
                var indexOfFirstStatement = 0;
                resumeLexicalEnvironment();
                indexOfFirstStatement = ts.addPrologueDirectivesAndInitialSuperCall(factory, constructor, statements, visitor);
                // Add parameters with property assignments. Transforms this:
                //
                //  constructor (public x, public y) {
                //  }
                //
                // Into this:
                //
                //  constructor (x, y) {
                //      this.x = x;
                //      this.y = y;
                //  }
                //
                ts.addRange(statements, ts.map(parametersWithPropertyAssignments, transformParameterWithPropertyAssignment));
                // Add the existing statements, skipping the initial super call.
                ts.addRange(statements, ts.visitNodes(body.statements, visitor, ts.isStatement, indexOfFirstStatement));
                // End the lexical environment.
                statements = factory.mergeLexicalEnvironment(statements, endLexicalEnvironment());
                var block = factory.createBlock(ts.setTextRange(factory.createNodeArray(statements), body.statements), /*multiLine*/ true);
                ts.setTextRange(block, /*location*/ body);
                ts.setOriginalNode(block, body);
                return block;
            }
            /**
             * Transforms a parameter into a property assignment statement.
             *
             * @param node The parameter declaration.
             */
            function transformParameterWithPropertyAssignment(node) {
                var name = node.name;
                if (!ts.isIdentifier(name)) {
                    return undefined;
                }
                // TODO(rbuckton): Does this need to be parented?
                var propertyName = ts.setParent(ts.setTextRange(factory.cloneNode(name), name), name.parent);
                ts.setEmitFlags(propertyName, 1536 /* NoComments */ | 48 /* NoSourceMap */);
                // TODO(rbuckton): Does this need to be parented?
                var localName = ts.setParent(ts.setTextRange(factory.cloneNode(name), name), name.parent);
                ts.setEmitFlags(localName, 1536 /* NoComments */);
                return ts.startOnNewLine(ts.removeAllComments(ts.setTextRange(ts.setOriginalNode(factory.createExpressionStatement(factory.createAssignment(ts.setTextRange(factory.createPropertyAccessExpression(factory.createThis(), propertyName), node.name), localName)), node), ts.moveRangePos(node, -1))));
            }
            function visitMethodDeclaration(node) {
                if (!shouldEmitFunctionLikeDeclaration(node)) {
                    return undefined;
                }
                var updated = factory.updateMethodDeclaration(node, 
                /*decorators*/ undefined, ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), node.asteriskToken, visitPropertyNameOfClassElement(node), 
                /*questionToken*/ undefined, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, ts.visitFunctionBody(node.body, visitor, context));
                if (updated !== node) {
                    // While we emit the source map for the node after skipping decorators and modifiers,
                    // we need to emit the comments for the original range.
                    ts.setCommentRange(updated, node);
                    ts.setSourceMapRange(updated, ts.moveRangePastDecorators(node));
                }
                return updated;
            }
            /**
             * Determines whether to emit an accessor declaration. We should not emit the
             * declaration if it does not have a body and is abstract.
             *
             * @param node The declaration node.
             */
            function shouldEmitAccessorDeclaration(node) {
                return !(ts.nodeIsMissing(node.body) && ts.hasSyntacticModifier(node, 128 /* Abstract */));
            }
            function visitGetAccessor(node) {
                if (!shouldEmitAccessorDeclaration(node)) {
                    return undefined;
                }
                var updated = factory.updateGetAccessorDeclaration(node, 
                /*decorators*/ undefined, ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), visitPropertyNameOfClassElement(node), ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, ts.visitFunctionBody(node.body, visitor, context) || factory.createBlock([]));
                if (updated !== node) {
                    // While we emit the source map for the node after skipping decorators and modifiers,
                    // we need to emit the comments for the original range.
                    ts.setCommentRange(updated, node);
                    ts.setSourceMapRange(updated, ts.moveRangePastDecorators(node));
                }
                return updated;
            }
            function visitSetAccessor(node) {
                if (!shouldEmitAccessorDeclaration(node)) {
                    return undefined;
                }
                var updated = factory.updateSetAccessorDeclaration(node, 
                /*decorators*/ undefined, ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), visitPropertyNameOfClassElement(node), ts.visitParameterList(node.parameters, visitor, context), ts.visitFunctionBody(node.body, visitor, context) || factory.createBlock([]));
                if (updated !== node) {
                    // While we emit the source map for the node after skipping decorators and modifiers,
                    // we need to emit the comments for the original range.
                    ts.setCommentRange(updated, node);
                    ts.setSourceMapRange(updated, ts.moveRangePastDecorators(node));
                }
                return updated;
            }
            function visitFunctionDeclaration(node) {
                if (!shouldEmitFunctionLikeDeclaration(node)) {
                    return factory.createNotEmittedStatement(node);
                }
                var updated = factory.updateFunctionDeclaration(node, 
                /*decorators*/ undefined, ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), node.asteriskToken, node.name, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, ts.visitFunctionBody(node.body, visitor, context) || factory.createBlock([]));
                if (isExportOfNamespace(node)) {
                    var statements = [updated];
                    addExportMemberAssignment(statements, node);
                    return statements;
                }
                return updated;
            }
            function visitFunctionExpression(node) {
                if (!shouldEmitFunctionLikeDeclaration(node)) {
                    return factory.createOmittedExpression();
                }
                var updated = factory.updateFunctionExpression(node, ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), node.asteriskToken, node.name, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, ts.visitFunctionBody(node.body, visitor, context) || factory.createBlock([]));
                return updated;
            }
            function visitArrowFunction(node) {
                var updated = factory.updateArrowFunction(node, ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, node.equalsGreaterThanToken, ts.visitFunctionBody(node.body, visitor, context));
                return updated;
            }
            function visitParameter(node) {
                if (ts.parameterIsThisKeyword(node)) {
                    return undefined;
                }
                var updated = factory.updateParameterDeclaration(node, 
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, node.dotDotDotToken, ts.visitNode(node.name, visitor, ts.isBindingName), 
                /*questionToken*/ undefined, 
                /*type*/ undefined, ts.visitNode(node.initializer, visitor, ts.isExpression));
                if (updated !== node) {
                    // While we emit the source map for the node after skipping decorators and modifiers,
                    // we need to emit the comments for the original range.
                    ts.setCommentRange(updated, node);
                    ts.setTextRange(updated, ts.moveRangePastModifiers(node));
                    ts.setSourceMapRange(updated, ts.moveRangePastModifiers(node));
                    ts.setEmitFlags(updated.name, 32 /* NoTrailingSourceMap */);
                }
                return updated;
            }
            function visitVariableStatement(node) {
                if (isExportOfNamespace(node)) {
                    var variables = ts.getInitializedVariables(node.declarationList);
                    if (variables.length === 0) {
                        // elide statement if there are no initialized variables.
                        return undefined;
                    }
                    return ts.setTextRange(factory.createExpressionStatement(factory.inlineExpressions(ts.map(variables, transformInitializedVariable))), node);
                }
                else {
                    return ts.visitEachChild(node, visitor, context);
                }
            }
            function transformInitializedVariable(node) {
                var name = node.name;
                if (ts.isBindingPattern(name)) {
                    return ts.flattenDestructuringAssignment(node, visitor, context, 0 /* All */, 
                    /*needsValue*/ false, createNamespaceExportExpression);
                }
                else {
                    return ts.setTextRange(factory.createAssignment(getNamespaceMemberNameWithSourceMapsAndWithoutComments(name), ts.visitNode(node.initializer, visitor, ts.isExpression)), 
                    /*location*/ node);
                }
            }
            function visitVariableDeclaration(node) {
                return factory.updateVariableDeclaration(node, ts.visitNode(node.name, visitor, ts.isBindingName), 
                /*exclamationToken*/ undefined, 
                /*type*/ undefined, ts.visitNode(node.initializer, visitor, ts.isExpression));
            }
            function visitParenthesizedExpression(node) {
                var innerExpression = ts.skipOuterExpressions(node.expression, ~6 /* Assertions */);
                if (ts.isAssertionExpression(innerExpression)) {
                    // Make sure we consider all nested cast expressions, e.g.:
                    // (<any><number><any>-A).x;
                    var expression = ts.visitNode(node.expression, visitor, ts.isExpression);
                    // We have an expression of the form: (<Type>SubExpr). Emitting this as (SubExpr)
                    // is really not desirable. We would like to emit the subexpression as-is. Omitting
                    // the parentheses, however, could cause change in the semantics of the generated
                    // code if the casted expression has a lower precedence than the rest of the
                    // expression.
                    //
                    // To preserve comments, we return a "PartiallyEmittedExpression" here which will
                    // preserve the position information of the original expression.
                    //
                    // Due to the auto-parenthesization rules used by the visitor and factory functions
                    // we can safely elide the parentheses here, as a new synthetic
                    // ParenthesizedExpression will be inserted if we remove parentheses too
                    // aggressively.
                    // HOWEVER - if there are leading comments on the expression itself, to handle ASI
                    // correctly for return and throw, we must keep the parenthesis
                    if (ts.length(ts.getLeadingCommentRangesOfNode(expression, currentSourceFile))) {
                        return factory.updateParenthesizedExpression(node, expression);
                    }
                    return factory.createPartiallyEmittedExpression(expression, node);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitAssertionExpression(node) {
                var expression = ts.visitNode(node.expression, visitor, ts.isExpression);
                return factory.createPartiallyEmittedExpression(expression, node);
            }
            function visitNonNullExpression(node) {
                var expression = ts.visitNode(node.expression, visitor, ts.isLeftHandSideExpression);
                return factory.createPartiallyEmittedExpression(expression, node);
            }
            function visitCallExpression(node) {
                return factory.updateCallExpression(node, ts.visitNode(node.expression, visitor, ts.isExpression), 
                /*typeArguments*/ undefined, ts.visitNodes(node.arguments, visitor, ts.isExpression));
            }
            function visitNewExpression(node) {
                return factory.updateNewExpression(node, ts.visitNode(node.expression, visitor, ts.isExpression), 
                /*typeArguments*/ undefined, ts.visitNodes(node.arguments, visitor, ts.isExpression));
            }
            function visitTaggedTemplateExpression(node) {
                return factory.updateTaggedTemplateExpression(node, ts.visitNode(node.tag, visitor, ts.isExpression), 
                /*typeArguments*/ undefined, ts.visitNode(node.template, visitor, ts.isExpression));
            }
            function visitJsxSelfClosingElement(node) {
                return factory.updateJsxSelfClosingElement(node, ts.visitNode(node.tagName, visitor, ts.isJsxTagNameExpression), 
                /*typeArguments*/ undefined, ts.visitNode(node.attributes, visitor, ts.isJsxAttributes));
            }
            function visitJsxJsxOpeningElement(node) {
                return factory.updateJsxOpeningElement(node, ts.visitNode(node.tagName, visitor, ts.isJsxTagNameExpression), 
                /*typeArguments*/ undefined, ts.visitNode(node.attributes, visitor, ts.isJsxAttributes));
            }
            /**
             * Determines whether to emit an enum declaration.
             *
             * @param node The enum declaration node.
             */
            function shouldEmitEnumDeclaration(node) {
                return !ts.isEnumConst(node)
                    || ts.shouldPreserveConstEnums(compilerOptions);
            }
            /**
             * Visits an enum declaration.
             *
             * This function will be called any time a TypeScript enum is encountered.
             *
             * @param node The enum declaration node.
             */
            function visitEnumDeclaration(node) {
                if (!shouldEmitEnumDeclaration(node)) {
                    return factory.createNotEmittedStatement(node);
                }
                var statements = [];
                // We request to be advised when the printer is about to print this node. This allows
                // us to set up the correct state for later substitutions.
                var emitFlags = 2 /* AdviseOnEmitNode */;
                // If needed, we should emit a variable declaration for the enum. If we emit
                // a leading variable declaration, we should not emit leading comments for the
                // enum body.
                var varAdded = addVarForEnumOrModuleDeclaration(statements, node);
                if (varAdded) {
                    // We should still emit the comments if we are emitting a system module.
                    if (moduleKind !== ts.ModuleKind.System || currentLexicalScope !== currentSourceFile) {
                        emitFlags |= 512 /* NoLeadingComments */;
                    }
                }
                // `parameterName` is the declaration name used inside of the enum.
                var parameterName = getNamespaceParameterName(node);
                // `containerName` is the expression used inside of the enum for assignments.
                var containerName = getNamespaceContainerName(node);
                // `exportName` is the expression used within this node's container for any exported references.
                var exportName = ts.hasSyntacticModifier(node, 1 /* Export */)
                    ? factory.getExternalModuleOrNamespaceExportName(currentNamespaceContainerName, node, /*allowComments*/ false, /*allowSourceMaps*/ true)
                    : factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true);
                //  x || (x = {})
                //  exports.x || (exports.x = {})
                var moduleArg = factory.createLogicalOr(exportName, factory.createAssignment(exportName, factory.createObjectLiteralExpression()));
                if (hasNamespaceQualifiedExportName(node)) {
                    // `localName` is the expression used within this node's containing scope for any local references.
                    var localName = factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true);
                    //  x = (exports.x || (exports.x = {}))
                    moduleArg = factory.createAssignment(localName, moduleArg);
                }
                //  (function (x) {
                //      x[x["y"] = 0] = "y";
                //      ...
                //  })(x || (x = {}));
                var enumStatement = factory.createExpressionStatement(factory.createCallExpression(factory.createFunctionExpression(
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, [factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, parameterName)], 
                /*type*/ undefined, transformEnumBody(node, containerName)), 
                /*typeArguments*/ undefined, [moduleArg]));
                ts.setOriginalNode(enumStatement, node);
                if (varAdded) {
                    // If a variable was added, synthetic comments are emitted on it, not on the moduleStatement.
                    ts.setSyntheticLeadingComments(enumStatement, undefined);
                    ts.setSyntheticTrailingComments(enumStatement, undefined);
                }
                ts.setTextRange(enumStatement, node);
                ts.addEmitFlags(enumStatement, emitFlags);
                statements.push(enumStatement);
                // Add a DeclarationMarker for the enum to preserve trailing comments and mark
                // the end of the declaration.
                statements.push(factory.createEndOfDeclarationMarker(node));
                return statements;
            }
            /**
             * Transforms the body of an enum declaration.
             *
             * @param node The enum declaration node.
             */
            function transformEnumBody(node, localName) {
                var savedCurrentNamespaceLocalName = currentNamespaceContainerName;
                currentNamespaceContainerName = localName;
                var statements = [];
                startLexicalEnvironment();
                var members = ts.map(node.members, transformEnumMember);
                ts.insertStatementsAfterStandardPrologue(statements, endLexicalEnvironment());
                ts.addRange(statements, members);
                currentNamespaceContainerName = savedCurrentNamespaceLocalName;
                return factory.createBlock(ts.setTextRange(factory.createNodeArray(statements), /*location*/ node.members), 
                /*multiLine*/ true);
            }
            /**
             * Transforms an enum member into a statement.
             *
             * @param member The enum member node.
             */
            function transformEnumMember(member) {
                // enums don't support computed properties
                // we pass false as 'generateNameForComputedPropertyName' for a backward compatibility purposes
                // old emitter always generate 'expression' part of the name as-is.
                var name = getExpressionForPropertyName(member, /*generateNameForComputedPropertyName*/ false);
                var valueExpression = transformEnumMemberDeclarationValue(member);
                var innerAssignment = factory.createAssignment(factory.createElementAccessExpression(currentNamespaceContainerName, name), valueExpression);
                var outerAssignment = valueExpression.kind === 10 /* StringLiteral */ ?
                    innerAssignment :
                    factory.createAssignment(factory.createElementAccessExpression(currentNamespaceContainerName, innerAssignment), name);
                return ts.setTextRange(factory.createExpressionStatement(ts.setTextRange(outerAssignment, member)), member);
            }
            /**
             * Transforms the value of an enum member.
             *
             * @param member The enum member node.
             */
            function transformEnumMemberDeclarationValue(member) {
                var value = resolver.getConstantValue(member);
                if (value !== undefined) {
                    return typeof value === "string" ? factory.createStringLiteral(value) : factory.createNumericLiteral(value);
                }
                else {
                    enableSubstitutionForNonQualifiedEnumMembers();
                    if (member.initializer) {
                        return ts.visitNode(member.initializer, visitor, ts.isExpression);
                    }
                    else {
                        return factory.createVoidZero();
                    }
                }
            }
            /**
             * Determines whether to elide a module declaration.
             *
             * @param node The module declaration node.
             */
            function shouldEmitModuleDeclaration(nodeIn) {
                var node = ts.getParseTreeNode(nodeIn, ts.isModuleDeclaration);
                if (!node) {
                    // If we can't find a parse tree node, assume the node is instantiated.
                    return true;
                }
                return ts.isInstantiatedModule(node, ts.shouldPreserveConstEnums(compilerOptions));
            }
            /**
             * Determines whether an exported declaration will have a qualified export name (e.g. `f.x`
             * or `exports.x`).
             */
            function hasNamespaceQualifiedExportName(node) {
                return isExportOfNamespace(node)
                    || (isExternalModuleExport(node)
                        && moduleKind !== ts.ModuleKind.ES2015
                        && moduleKind !== ts.ModuleKind.ES2020
                        && moduleKind !== ts.ModuleKind.ESNext
                        && moduleKind !== ts.ModuleKind.System);
            }
            /**
             * Records that a declaration was emitted in the current scope, if it was the first
             * declaration for the provided symbol.
             */
            function recordEmittedDeclarationInScope(node) {
                if (!currentScopeFirstDeclarationsOfName) {
                    currentScopeFirstDeclarationsOfName = new ts.Map();
                }
                var name = declaredNameInScope(node);
                if (!currentScopeFirstDeclarationsOfName.has(name)) {
                    currentScopeFirstDeclarationsOfName.set(name, node);
                }
            }
            /**
             * Determines whether a declaration is the first declaration with
             * the same name emitted in the current scope.
             */
            function isFirstEmittedDeclarationInScope(node) {
                if (currentScopeFirstDeclarationsOfName) {
                    var name = declaredNameInScope(node);
                    return currentScopeFirstDeclarationsOfName.get(name) === node;
                }
                return true;
            }
            function declaredNameInScope(node) {
                ts.Debug.assertNode(node.name, ts.isIdentifier);
                return node.name.escapedText;
            }
            /**
             * Adds a leading VariableStatement for a enum or module declaration.
             */
            function addVarForEnumOrModuleDeclaration(statements, node) {
                // Emit a variable statement for the module. We emit top-level enums as a `var`
                // declaration to avoid static errors in global scripts scripts due to redeclaration.
                // enums in any other scope are emitted as a `let` declaration.
                var statement = factory.createVariableStatement(ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), factory.createVariableDeclarationList([
                    factory.createVariableDeclaration(factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true))
                ], currentLexicalScope.kind === 300 /* SourceFile */ ? 0 /* None */ : 1 /* Let */));
                ts.setOriginalNode(statement, node);
                recordEmittedDeclarationInScope(node);
                if (isFirstEmittedDeclarationInScope(node)) {
                    // Adjust the source map emit to match the old emitter.
                    if (node.kind === 258 /* EnumDeclaration */) {
                        ts.setSourceMapRange(statement.declarationList, node);
                    }
                    else {
                        ts.setSourceMapRange(statement, node);
                    }
                    // Trailing comments for module declaration should be emitted after the function closure
                    // instead of the variable statement:
                    //
                    //     /** Module comment*/
                    //     module m1 {
                    //         function foo4Export() {
                    //         }
                    //     } // trailing comment module
                    //
                    // Should emit:
                    //
                    //     /** Module comment*/
                    //     var m1;
                    //     (function (m1) {
                    //         function foo4Export() {
                    //         }
                    //     })(m1 || (m1 = {})); // trailing comment module
                    //
                    ts.setCommentRange(statement, node);
                    ts.addEmitFlags(statement, 1024 /* NoTrailingComments */ | 4194304 /* HasEndOfDeclarationMarker */);
                    statements.push(statement);
                    return true;
                }
                else {
                    // For an EnumDeclaration or ModuleDeclaration that merges with a preceeding
                    // declaration we do not emit a leading variable declaration. To preserve the
                    // begin/end semantics of the declararation and to properly handle exports
                    // we wrap the leading variable declaration in a `MergeDeclarationMarker`.
                    var mergeMarker = factory.createMergeDeclarationMarker(statement);
                    ts.setEmitFlags(mergeMarker, 1536 /* NoComments */ | 4194304 /* HasEndOfDeclarationMarker */);
                    statements.push(mergeMarker);
                    return false;
                }
            }
            /**
             * Visits a module declaration node.
             *
             * This function will be called any time a TypeScript namespace (ModuleDeclaration) is encountered.
             *
             * @param node The module declaration node.
             */
            function visitModuleDeclaration(node) {
                if (!shouldEmitModuleDeclaration(node)) {
                    return factory.createNotEmittedStatement(node);
                }
                ts.Debug.assertNode(node.name, ts.isIdentifier, "A TypeScript namespace should have an Identifier name.");
                enableSubstitutionForNamespaceExports();
                var statements = [];
                // We request to be advised when the printer is about to print this node. This allows
                // us to set up the correct state for later substitutions.
                var emitFlags = 2 /* AdviseOnEmitNode */;
                // If needed, we should emit a variable declaration for the module. If we emit
                // a leading variable declaration, we should not emit leading comments for the
                // module body.
                var varAdded = addVarForEnumOrModuleDeclaration(statements, node);
                if (varAdded) {
                    // We should still emit the comments if we are emitting a system module.
                    if (moduleKind !== ts.ModuleKind.System || currentLexicalScope !== currentSourceFile) {
                        emitFlags |= 512 /* NoLeadingComments */;
                    }
                }
                // `parameterName` is the declaration name used inside of the namespace.
                var parameterName = getNamespaceParameterName(node);
                // `containerName` is the expression used inside of the namespace for exports.
                var containerName = getNamespaceContainerName(node);
                // `exportName` is the expression used within this node's container for any exported references.
                var exportName = ts.hasSyntacticModifier(node, 1 /* Export */)
                    ? factory.getExternalModuleOrNamespaceExportName(currentNamespaceContainerName, node, /*allowComments*/ false, /*allowSourceMaps*/ true)
                    : factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true);
                //  x || (x = {})
                //  exports.x || (exports.x = {})
                var moduleArg = factory.createLogicalOr(exportName, factory.createAssignment(exportName, factory.createObjectLiteralExpression()));
                if (hasNamespaceQualifiedExportName(node)) {
                    // `localName` is the expression used within this node's containing scope for any local references.
                    var localName = factory.getLocalName(node, /*allowComments*/ false, /*allowSourceMaps*/ true);
                    //  x = (exports.x || (exports.x = {}))
                    moduleArg = factory.createAssignment(localName, moduleArg);
                }
                //  (function (x_1) {
                //      x_1.y = ...;
                //  })(x || (x = {}));
                var moduleStatement = factory.createExpressionStatement(factory.createCallExpression(factory.createFunctionExpression(
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, [factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, parameterName)], 
                /*type*/ undefined, transformModuleBody(node, containerName)), 
                /*typeArguments*/ undefined, [moduleArg]));
                ts.setOriginalNode(moduleStatement, node);
                if (varAdded) {
                    // If a variable was added, synthetic comments are emitted on it, not on the moduleStatement.
                    ts.setSyntheticLeadingComments(moduleStatement, undefined);
                    ts.setSyntheticTrailingComments(moduleStatement, undefined);
                }
                ts.setTextRange(moduleStatement, node);
                ts.addEmitFlags(moduleStatement, emitFlags);
                statements.push(moduleStatement);
                // Add a DeclarationMarker for the namespace to preserve trailing comments and mark
                // the end of the declaration.
                statements.push(factory.createEndOfDeclarationMarker(node));
                return statements;
            }
            /**
             * Transforms the body of a module declaration.
             *
             * @param node The module declaration node.
             */
            function transformModuleBody(node, namespaceLocalName) {
                var savedCurrentNamespaceContainerName = currentNamespaceContainerName;
                var savedCurrentNamespace = currentNamespace;
                var savedCurrentScopeFirstDeclarationsOfName = currentScopeFirstDeclarationsOfName;
                currentNamespaceContainerName = namespaceLocalName;
                currentNamespace = node;
                currentScopeFirstDeclarationsOfName = undefined;
                var statements = [];
                startLexicalEnvironment();
                var statementsLocation;
                var blockLocation;
                if (node.body) {
                    if (node.body.kind === 260 /* ModuleBlock */) {
                        saveStateAndInvoke(node.body, function (body) { return ts.addRange(statements, ts.visitNodes(body.statements, namespaceElementVisitor, ts.isStatement)); });
                        statementsLocation = node.body.statements;
                        blockLocation = node.body;
                    }
                    else {
                        var result = visitModuleDeclaration(node.body);
                        if (result) {
                            if (ts.isArray(result)) {
                                ts.addRange(statements, result);
                            }
                            else {
                                statements.push(result);
                            }
                        }
                        var moduleBlock = getInnerMostModuleDeclarationFromDottedModule(node).body;
                        statementsLocation = ts.moveRangePos(moduleBlock.statements, -1);
                    }
                }
                ts.insertStatementsAfterStandardPrologue(statements, endLexicalEnvironment());
                currentNamespaceContainerName = savedCurrentNamespaceContainerName;
                currentNamespace = savedCurrentNamespace;
                currentScopeFirstDeclarationsOfName = savedCurrentScopeFirstDeclarationsOfName;
                var block = factory.createBlock(ts.setTextRange(factory.createNodeArray(statements), 
                /*location*/ statementsLocation), 
                /*multiLine*/ true);
                ts.setTextRange(block, blockLocation);
                // namespace hello.hi.world {
                //      function foo() {}
                //
                //      // TODO, blah
                // }
                //
                // should be emitted as
                //
                // var hello;
                // (function (hello) {
                //     var hi;
                //     (function (hi) {
                //         var world;
                //         (function (world) {
                //             function foo() { }
                //             // TODO, blah
                //         })(world = hi.world || (hi.world = {}));
                //     })(hi = hello.hi || (hello.hi = {}));
                // })(hello || (hello = {}));
                // We only want to emit comment on the namespace which contains block body itself, not the containing namespaces.
                if (!node.body || node.body.kind !== 260 /* ModuleBlock */) {
                    ts.setEmitFlags(block, ts.getEmitFlags(block) | 1536 /* NoComments */);
                }
                return block;
            }
            function getInnerMostModuleDeclarationFromDottedModule(moduleDeclaration) {
                if (moduleDeclaration.body.kind === 259 /* ModuleDeclaration */) {
                    var recursiveInnerModule = getInnerMostModuleDeclarationFromDottedModule(moduleDeclaration.body);
                    return recursiveInnerModule || moduleDeclaration.body;
                }
            }
            /**
             * Visits an import declaration, eliding it if it is not referenced and `importsNotUsedAsValues` is not 'preserve'.
             *
             * @param node The import declaration node.
             */
            function visitImportDeclaration(node) {
                if (!node.importClause) {
                    // Do not elide a side-effect only import declaration.
                    //  import "foo";
                    return node;
                }
                if (node.importClause.isTypeOnly) {
                    // Always elide type-only imports
                    return undefined;
                }
                // Elide the declaration if the import clause was elided.
                var importClause = ts.visitNode(node.importClause, visitImportClause, ts.isImportClause);
                return importClause ||
                    compilerOptions.importsNotUsedAsValues === 1 /* Preserve */ ||
                    compilerOptions.importsNotUsedAsValues === 2 /* Error */
                    ? factory.updateImportDeclaration(node, 
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, importClause, node.moduleSpecifier)
                    : undefined;
            }
            /**
             * Visits an import clause, eliding it if it is not referenced.
             *
             * @param node The import clause node.
             */
            function visitImportClause(node) {
                if (node.isTypeOnly) {
                    return undefined;
                }
                // Elide the import clause if we elide both its name and its named bindings.
                var name = resolver.isReferencedAliasDeclaration(node) ? node.name : undefined;
                var namedBindings = ts.visitNode(node.namedBindings, visitNamedImportBindings, ts.isNamedImportBindings);
                return (name || namedBindings) ? factory.updateImportClause(node, /*isTypeOnly*/ false, name, namedBindings) : undefined;
            }
            /**
             * Visits named import bindings, eliding it if it is not referenced.
             *
             * @param node The named import bindings node.
             */
            function visitNamedImportBindings(node) {
                if (node.kind === 266 /* NamespaceImport */) {
                    // Elide a namespace import if it is not referenced.
                    return resolver.isReferencedAliasDeclaration(node) ? node : undefined;
                }
                else {
                    // Elide named imports if all of its import specifiers are elided.
                    var elements = ts.visitNodes(node.elements, visitImportSpecifier, ts.isImportSpecifier);
                    return ts.some(elements) ? factory.updateNamedImports(node, elements) : undefined;
                }
            }
            /**
             * Visits an import specifier, eliding it if it is not referenced.
             *
             * @param node The import specifier node.
             */
            function visitImportSpecifier(node) {
                // Elide an import specifier if it is not referenced.
                return resolver.isReferencedAliasDeclaration(node) ? node : undefined;
            }
            /**
             * Visits an export assignment, eliding it if it does not contain a clause that resolves
             * to a value.
             *
             * @param node The export assignment node.
             */
            function visitExportAssignment(node) {
                // Elide the export assignment if it does not reference a value.
                return resolver.isValueAliasDeclaration(node)
                    ? ts.visitEachChild(node, visitor, context)
                    : undefined;
            }
            /**
             * Visits an export declaration, eliding it if it does not contain a clause that resolves
             * to a value.
             *
             * @param node The export declaration node.
             */
            function visitExportDeclaration(node) {
                if (node.isTypeOnly) {
                    return undefined;
                }
                if (!node.exportClause || ts.isNamespaceExport(node.exportClause)) {
                    // never elide `export <whatever> from <whereever>` declarations -
                    // they should be kept for sideffects/untyped exports, even when the
                    // type checker doesn't know about any exports
                    return node;
                }
                if (!resolver.isValueAliasDeclaration(node)) {
                    // Elide the export declaration if it does not export a value.
                    return undefined;
                }
                // Elide the export declaration if all of its named exports are elided.
                var exportClause = ts.visitNode(node.exportClause, visitNamedExportBindings, ts.isNamedExportBindings);
                return exportClause
                    ? factory.updateExportDeclaration(node, 
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, node.isTypeOnly, exportClause, node.moduleSpecifier)
                    : undefined;
            }
            /**
             * Visits named exports, eliding it if it does not contain an export specifier that
             * resolves to a value.
             *
             * @param node The named exports node.
             */
            function visitNamedExports(node) {
                // Elide the named exports if all of its export specifiers were elided.
                var elements = ts.visitNodes(node.elements, visitExportSpecifier, ts.isExportSpecifier);
                return ts.some(elements) ? factory.updateNamedExports(node, elements) : undefined;
            }
            function visitNamespaceExports(node) {
                return factory.updateNamespaceExport(node, ts.visitNode(node.name, visitor, ts.isIdentifier));
            }
            function visitNamedExportBindings(node) {
                return ts.isNamespaceExport(node) ? visitNamespaceExports(node) : visitNamedExports(node);
            }
            /**
             * Visits an export specifier, eliding it if it does not resolve to a value.
             *
             * @param node The export specifier node.
             */
            function visitExportSpecifier(node) {
                // Elide an export specifier if it does not reference a value.
                return resolver.isValueAliasDeclaration(node) ? node : undefined;
            }
            /**
             * Determines whether to emit an import equals declaration.
             *
             * @param node The import equals declaration node.
             */
            function shouldEmitImportEqualsDeclaration(node) {
                // preserve old compiler's behavior: emit 'var' for import declaration (even if we do not consider them referenced) when
                // - current file is not external module
                // - import declaration is top level and target is value imported by entity name
                return resolver.isReferencedAliasDeclaration(node)
                    || (!ts.isExternalModule(currentSourceFile)
                        && resolver.isTopLevelValueImportEqualsWithEntityName(node));
            }
            /**
             * Visits an import equals declaration.
             *
             * @param node The import equals declaration node.
             */
            function visitImportEqualsDeclaration(node) {
                // Always elide type-only imports
                if (node.isTypeOnly) {
                    return undefined;
                }
                if (ts.isExternalModuleImportEqualsDeclaration(node)) {
                    var isReferenced = resolver.isReferencedAliasDeclaration(node);
                    // If the alias is unreferenced but we want to keep the import, replace with 'import "mod"'.
                    if (!isReferenced && compilerOptions.importsNotUsedAsValues === 1 /* Preserve */) {
                        return ts.setOriginalNode(ts.setTextRange(factory.createImportDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, 
                        /*importClause*/ undefined, node.moduleReference.expression), node), node);
                    }
                    return isReferenced ? ts.visitEachChild(node, visitor, context) : undefined;
                }
                if (!shouldEmitImportEqualsDeclaration(node)) {
                    return undefined;
                }
                var moduleReference = ts.createExpressionFromEntityName(factory, node.moduleReference);
                ts.setEmitFlags(moduleReference, 1536 /* NoComments */ | 2048 /* NoNestedComments */);
                if (isNamedExternalModuleExport(node) || !isExportOfNamespace(node)) {
                    //  export var ${name} = ${moduleReference};
                    //  var ${name} = ${moduleReference};
                    return ts.setOriginalNode(ts.setTextRange(factory.createVariableStatement(ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), factory.createVariableDeclarationList([
                        ts.setOriginalNode(factory.createVariableDeclaration(node.name, 
                        /*exclamationToken*/ undefined, 
                        /*type*/ undefined, moduleReference), node)
                    ])), node), node);
                }
                else {
                    // exports.${name} = ${moduleReference};
                    return ts.setOriginalNode(createNamespaceExport(node.name, moduleReference, node), node);
                }
            }
            /**
             * Gets a value indicating whether the node is exported from a namespace.
             *
             * @param node The node to test.
             */
            function isExportOfNamespace(node) {
                return currentNamespace !== undefined && ts.hasSyntacticModifier(node, 1 /* Export */);
            }
            /**
             * Gets a value indicating whether the node is exported from an external module.
             *
             * @param node The node to test.
             */
            function isExternalModuleExport(node) {
                return currentNamespace === undefined && ts.hasSyntacticModifier(node, 1 /* Export */);
            }
            /**
             * Gets a value indicating whether the node is a named export from an external module.
             *
             * @param node The node to test.
             */
            function isNamedExternalModuleExport(node) {
                return isExternalModuleExport(node)
                    && !ts.hasSyntacticModifier(node, 512 /* Default */);
            }
            /**
             * Gets a value indicating whether the node is the default export of an external module.
             *
             * @param node The node to test.
             */
            function isDefaultExternalModuleExport(node) {
                return isExternalModuleExport(node)
                    && ts.hasSyntacticModifier(node, 512 /* Default */);
            }
            /**
             * Creates a statement for the provided expression. This is used in calls to `map`.
             */
            function expressionToStatement(expression) {
                return factory.createExpressionStatement(expression);
            }
            function addExportMemberAssignment(statements, node) {
                var expression = factory.createAssignment(factory.getExternalModuleOrNamespaceExportName(currentNamespaceContainerName, node, /*allowComments*/ false, /*allowSourceMaps*/ true), factory.getLocalName(node));
                ts.setSourceMapRange(expression, ts.createRange(node.name ? node.name.pos : node.pos, node.end));
                var statement = factory.createExpressionStatement(expression);
                ts.setSourceMapRange(statement, ts.createRange(-1, node.end));
                statements.push(statement);
            }
            function createNamespaceExport(exportName, exportValue, location) {
                return ts.setTextRange(factory.createExpressionStatement(factory.createAssignment(factory.getNamespaceMemberName(currentNamespaceContainerName, exportName, /*allowComments*/ false, /*allowSourceMaps*/ true), exportValue)), location);
            }
            function createNamespaceExportExpression(exportName, exportValue, location) {
                return ts.setTextRange(factory.createAssignment(getNamespaceMemberNameWithSourceMapsAndWithoutComments(exportName), exportValue), location);
            }
            function getNamespaceMemberNameWithSourceMapsAndWithoutComments(name) {
                return factory.getNamespaceMemberName(currentNamespaceContainerName, name, /*allowComments*/ false, /*allowSourceMaps*/ true);
            }
            /**
             * Gets the declaration name used inside of a namespace or enum.
             */
            function getNamespaceParameterName(node) {
                var name = factory.getGeneratedNameForNode(node);
                ts.setSourceMapRange(name, node.name);
                return name;
            }
            /**
             * Gets the expression used to refer to a namespace or enum within the body
             * of its declaration.
             */
            function getNamespaceContainerName(node) {
                return factory.getGeneratedNameForNode(node);
            }
            /**
             * Gets a local alias for a class declaration if it is a decorated class with an internal
             * reference to the static side of the class. This is necessary to avoid issues with
             * double-binding semantics for the class name.
             */
            function getClassAliasIfNeeded(node) {
                if (resolver.getNodeCheckFlags(node) & 16777216 /* ClassWithConstructorReference */) {
                    enableSubstitutionForClassAliases();
                    var classAlias = factory.createUniqueName(node.name && !ts.isGeneratedIdentifier(node.name) ? ts.idText(node.name) : "default");
                    classAliases[ts.getOriginalNodeId(node)] = classAlias;
                    hoistVariableDeclaration(classAlias);
                    return classAlias;
                }
            }
            function getClassPrototype(node) {
                return factory.createPropertyAccessExpression(factory.getDeclarationName(node), "prototype");
            }
            function getClassMemberPrefix(node, member) {
                return ts.isStatic(member)
                    ? factory.getDeclarationName(node)
                    : getClassPrototype(node);
            }
            function enableSubstitutionForNonQualifiedEnumMembers() {
                if ((enabledSubstitutions & 8 /* NonQualifiedEnumMembers */) === 0) {
                    enabledSubstitutions |= 8 /* NonQualifiedEnumMembers */;
                    context.enableSubstitution(79 /* Identifier */);
                }
            }
            function enableSubstitutionForClassAliases() {
                if ((enabledSubstitutions & 1 /* ClassAliases */) === 0) {
                    enabledSubstitutions |= 1 /* ClassAliases */;
                    // We need to enable substitutions for identifiers. This allows us to
                    // substitute class names inside of a class declaration.
                    context.enableSubstitution(79 /* Identifier */);
                    // Keep track of class aliases.
                    classAliases = [];
                }
            }
            function enableSubstitutionForNamespaceExports() {
                if ((enabledSubstitutions & 2 /* NamespaceExports */) === 0) {
                    enabledSubstitutions |= 2 /* NamespaceExports */;
                    // We need to enable substitutions for identifiers and shorthand property assignments. This allows us to
                    // substitute the names of exported members of a namespace.
                    context.enableSubstitution(79 /* Identifier */);
                    context.enableSubstitution(292 /* ShorthandPropertyAssignment */);
                    // We need to be notified when entering and exiting namespaces.
                    context.enableEmitNotification(259 /* ModuleDeclaration */);
                }
            }
            function isTransformedModuleDeclaration(node) {
                return ts.getOriginalNode(node).kind === 259 /* ModuleDeclaration */;
            }
            function isTransformedEnumDeclaration(node) {
                return ts.getOriginalNode(node).kind === 258 /* EnumDeclaration */;
            }
            /**
             * Hook for node emit.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to emit.
             * @param emit A callback used to emit the node in the printer.
             */
            function onEmitNode(hint, node, emitCallback) {
                var savedApplicableSubstitutions = applicableSubstitutions;
                var savedCurrentSourceFile = currentSourceFile;
                if (ts.isSourceFile(node)) {
                    currentSourceFile = node;
                }
                if (enabledSubstitutions & 2 /* NamespaceExports */ && isTransformedModuleDeclaration(node)) {
                    applicableSubstitutions |= 2 /* NamespaceExports */;
                }
                if (enabledSubstitutions & 8 /* NonQualifiedEnumMembers */ && isTransformedEnumDeclaration(node)) {
                    applicableSubstitutions |= 8 /* NonQualifiedEnumMembers */;
                }
                previousOnEmitNode(hint, node, emitCallback);
                applicableSubstitutions = savedApplicableSubstitutions;
                currentSourceFile = savedCurrentSourceFile;
            }
            /**
             * Hooks node substitutions.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to substitute.
             */
            function onSubstituteNode(hint, node) {
                node = previousOnSubstituteNode(hint, node);
                if (hint === 1 /* Expression */) {
                    return substituteExpression(node);
                }
                else if (ts.isShorthandPropertyAssignment(node)) {
                    return substituteShorthandPropertyAssignment(node);
                }
                return node;
            }
            function substituteShorthandPropertyAssignment(node) {
                if (enabledSubstitutions & 2 /* NamespaceExports */) {
                    var name = node.name;
                    var exportedName = trySubstituteNamespaceExportedName(name);
                    if (exportedName) {
                        // A shorthand property with an assignment initializer is probably part of a
                        // destructuring assignment
                        if (node.objectAssignmentInitializer) {
                            var initializer = factory.createAssignment(exportedName, node.objectAssignmentInitializer);
                            return ts.setTextRange(factory.createPropertyAssignment(name, initializer), node);
                        }
                        return ts.setTextRange(factory.createPropertyAssignment(name, exportedName), node);
                    }
                }
                return node;
            }
            function substituteExpression(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        return substituteExpressionIdentifier(node);
                    case 204 /* PropertyAccessExpression */:
                        return substitutePropertyAccessExpression(node);
                    case 205 /* ElementAccessExpression */:
                        return substituteElementAccessExpression(node);
                }
                return node;
            }
            function substituteExpressionIdentifier(node) {
                return trySubstituteClassAlias(node)
                    || trySubstituteNamespaceExportedName(node)
                    || node;
            }
            function trySubstituteClassAlias(node) {
                if (enabledSubstitutions & 1 /* ClassAliases */) {
                    if (resolver.getNodeCheckFlags(node) & 33554432 /* ConstructorReferenceInClass */) {
                        // Due to the emit for class decorators, any reference to the class from inside of the class body
                        // must instead be rewritten to point to a temporary variable to avoid issues with the double-bind
                        // behavior of class names in ES6.
                        // Also, when emitting statics for class expressions, we must substitute a class alias for
                        // constructor references in static property initializers.
                        var declaration = resolver.getReferencedValueDeclaration(node);
                        if (declaration) {
                            var classAlias = classAliases[declaration.id]; // TODO: GH#18217
                            if (classAlias) {
                                var clone_2 = factory.cloneNode(classAlias);
                                ts.setSourceMapRange(clone_2, node);
                                ts.setCommentRange(clone_2, node);
                                return clone_2;
                            }
                        }
                    }
                }
                return undefined;
            }
            function trySubstituteNamespaceExportedName(node) {
                // If this is explicitly a local name, do not substitute.
                if (enabledSubstitutions & applicableSubstitutions && !ts.isGeneratedIdentifier(node) && !ts.isLocalName(node)) {
                    // If we are nested within a namespace declaration, we may need to qualifiy
                    // an identifier that is exported from a merged namespace.
                    var container = resolver.getReferencedExportContainer(node, /*prefixLocals*/ false);
                    if (container && container.kind !== 300 /* SourceFile */) {
                        var substitute = (applicableSubstitutions & 2 /* NamespaceExports */ && container.kind === 259 /* ModuleDeclaration */) ||
                            (applicableSubstitutions & 8 /* NonQualifiedEnumMembers */ && container.kind === 258 /* EnumDeclaration */);
                        if (substitute) {
                            return ts.setTextRange(factory.createPropertyAccessExpression(factory.getGeneratedNameForNode(container), node), 
                            /*location*/ node);
                        }
                    }
                }
                return undefined;
            }
            function substitutePropertyAccessExpression(node) {
                return substituteConstantValue(node);
            }
            function substituteElementAccessExpression(node) {
                return substituteConstantValue(node);
            }
            function substituteConstantValue(node) {
                var constantValue = tryGetConstEnumValue(node);
                if (constantValue !== undefined) {
                    // track the constant value on the node for the printer in needsDotDotForPropertyAccess
                    ts.setConstantValue(node, constantValue);
                    var substitute = typeof constantValue === "string" ? factory.createStringLiteral(constantValue) : factory.createNumericLiteral(constantValue);
                    if (!compilerOptions.removeComments) {
                        var originalNode = ts.getOriginalNode(node, ts.isAccessExpression);
                        var propertyName = ts.isPropertyAccessExpression(originalNode)
                            ? ts.declarationNameToString(originalNode.name)
                            : ts.getTextOfNode(originalNode.argumentExpression);
                        ts.addSyntheticTrailingComment(substitute, 3 /* MultiLineCommentTrivia */, " " + propertyName + " ");
                    }
                    return substitute;
                }
                return node;
            }
            function tryGetConstEnumValue(node) {
                if (compilerOptions.isolatedModules) {
                    return undefined;
                }
                return ts.isPropertyAccessExpression(node) || ts.isElementAccessExpression(node) ? resolver.getConstantValue(node) : undefined;
            }
        }
        ts.transformTypeScript = transformTypeScript;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var ClassPropertySubstitutionFlags;
        (function (ClassPropertySubstitutionFlags) {
            /**
             * Enables substitutions for class expressions with static fields
             * which have initializers that reference the class name.
             */
            ClassPropertySubstitutionFlags[ClassPropertySubstitutionFlags["ClassAliases"] = 1] = "ClassAliases";
            /**
             * Enables substitutions for class expressions with static fields
             * which have initializers that reference the 'this' or 'super'.
             */
            ClassPropertySubstitutionFlags[ClassPropertySubstitutionFlags["ClassStaticThisOrSuperReference"] = 2] = "ClassStaticThisOrSuperReference";
        })(ClassPropertySubstitutionFlags || (ClassPropertySubstitutionFlags = {}));
        var PrivateIdentifierKind;
        (function (PrivateIdentifierKind) {
            PrivateIdentifierKind["Field"] = "f";
            PrivateIdentifierKind["Method"] = "m";
            PrivateIdentifierKind["Accessor"] = "a";
        })(PrivateIdentifierKind = ts.PrivateIdentifierKind || (ts.PrivateIdentifierKind = {}));
        var ClassFacts;
        (function (ClassFacts) {
            ClassFacts[ClassFacts["None"] = 0] = "None";
            ClassFacts[ClassFacts["ClassWasDecorated"] = 1] = "ClassWasDecorated";
            ClassFacts[ClassFacts["NeedsClassConstructorReference"] = 2] = "NeedsClassConstructorReference";
            ClassFacts[ClassFacts["NeedsClassSuperReference"] = 4] = "NeedsClassSuperReference";
            ClassFacts[ClassFacts["NeedsSubstitutionForThisInClassStaticField"] = 8] = "NeedsSubstitutionForThisInClassStaticField";
        })(ClassFacts || (ClassFacts = {}));
        /**
         * Transforms ECMAScript Class Syntax.
         * TypeScript parameter property syntax is transformed in the TypeScript transformer.
         * For now, this transforms public field declarations using TypeScript class semantics,
         * where declarations are elided and initializers are transformed as assignments in the constructor.
         * When --useDefineForClassFields is on, this transforms to ECMAScript semantics, with Object.defineProperty.
         */
        function transformClassFields(context) {
            var factory = context.factory, hoistVariableDeclaration = context.hoistVariableDeclaration, endLexicalEnvironment = context.endLexicalEnvironment, startLexicalEnvironment = context.startLexicalEnvironment, resumeLexicalEnvironment = context.resumeLexicalEnvironment, addBlockScopedVariable = context.addBlockScopedVariable;
            var resolver = context.getEmitResolver();
            var compilerOptions = context.getCompilerOptions();
            var languageVersion = ts.getEmitScriptTarget(compilerOptions);
            var useDefineForClassFields = ts.getUseDefineForClassFields(compilerOptions);
            var shouldTransformPrivateElementsOrClassStaticBlocks = languageVersion < 99 /* ESNext */;
            // We don't need to transform `super` property access when targeting ES5, ES3 because
            // the es2015 transformation handles those.
            var shouldTransformSuperInStaticInitializers = (languageVersion <= 8 /* ES2021 */ || !useDefineForClassFields) && languageVersion >= 2 /* ES2015 */;
            var shouldTransformThisInStaticInitializers = languageVersion <= 8 /* ES2021 */ || !useDefineForClassFields;
            var previousOnSubstituteNode = context.onSubstituteNode;
            context.onSubstituteNode = onSubstituteNode;
            var previousOnEmitNode = context.onEmitNode;
            context.onEmitNode = onEmitNode;
            var enabledSubstitutions;
            var classAliases;
            /**
             * Tracks what computed name expressions originating from elided names must be inlined
             * at the next execution site, in document order
             */
            var pendingExpressions;
            /**
             * Tracks what computed name expression statements and static property initializers must be
             * emitted at the next execution site, in document order (for decorated classes).
             */
            var pendingStatements;
            var classLexicalEnvironmentStack = [];
            var classLexicalEnvironmentMap = new ts.Map();
            var currentClassLexicalEnvironment;
            var currentComputedPropertyNameClassLexicalEnvironment;
            var currentStaticPropertyDeclarationOrStaticBlock;
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                var options = context.getCompilerOptions();
                if (node.isDeclarationFile
                    || useDefineForClassFields && options.target === 99 /* ESNext */) {
                    return node;
                }
                var visited = ts.visitEachChild(node, visitor, context);
                ts.addEmitHelpers(visited, context.readEmitHelpers());
                return visited;
            }
            function visitorWorker(node, valueIsDiscarded) {
                if (node.transformFlags & 8388608 /* ContainsClassFields */) {
                    switch (node.kind) {
                        case 224 /* ClassExpression */:
                        case 255 /* ClassDeclaration */:
                            return visitClassLike(node);
                        case 165 /* PropertyDeclaration */:
                            return visitPropertyDeclaration(node);
                        case 235 /* VariableStatement */:
                            return visitVariableStatement(node);
                        case 80 /* PrivateIdentifier */:
                            return visitPrivateIdentifier(node);
                        case 168 /* ClassStaticBlockDeclaration */:
                            return visitClassStaticBlockDeclaration(node);
                    }
                }
                if (node.transformFlags & 8388608 /* ContainsClassFields */ ||
                    node.transformFlags & 33554432 /* ContainsLexicalSuper */ &&
                        shouldTransformSuperInStaticInitializers &&
                        currentStaticPropertyDeclarationOrStaticBlock &&
                        currentClassLexicalEnvironment) {
                    switch (node.kind) {
                        case 217 /* PrefixUnaryExpression */:
                        case 218 /* PostfixUnaryExpression */:
                            return visitPreOrPostfixUnaryExpression(node, valueIsDiscarded);
                        case 219 /* BinaryExpression */:
                            return visitBinaryExpression(node, valueIsDiscarded);
                        case 206 /* CallExpression */:
                            return visitCallExpression(node);
                        case 208 /* TaggedTemplateExpression */:
                            return visitTaggedTemplateExpression(node);
                        case 204 /* PropertyAccessExpression */:
                            return visitPropertyAccessExpression(node);
                        case 205 /* ElementAccessExpression */:
                            return visitElementAccessExpression(node);
                        case 236 /* ExpressionStatement */:
                            return visitExpressionStatement(node);
                        case 240 /* ForStatement */:
                            return visitForStatement(node);
                        case 254 /* FunctionDeclaration */:
                        case 211 /* FunctionExpression */:
                        case 169 /* Constructor */:
                        case 167 /* MethodDeclaration */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */: {
                            var savedCurrentStaticPropertyDeclarationOrStaticBlock = currentStaticPropertyDeclarationOrStaticBlock;
                            currentStaticPropertyDeclarationOrStaticBlock = undefined;
                            var result = ts.visitEachChild(node, visitor, context);
                            currentStaticPropertyDeclarationOrStaticBlock = savedCurrentStaticPropertyDeclarationOrStaticBlock;
                            return result;
                        }
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function discardedValueVisitor(node) {
                return visitorWorker(node, /*valueIsDiscarded*/ true);
            }
            function visitor(node) {
                return visitorWorker(node, /*valueIsDiscarded*/ false);
            }
            function heritageClauseVisitor(node) {
                switch (node.kind) {
                    case 289 /* HeritageClause */:
                        return ts.visitEachChild(node, heritageClauseVisitor, context);
                    case 226 /* ExpressionWithTypeArguments */:
                        return visitExpressionWithTypeArguments(node);
                }
                return visitor(node);
            }
            function visitorDestructuringTarget(node) {
                switch (node.kind) {
                    case 203 /* ObjectLiteralExpression */:
                    case 202 /* ArrayLiteralExpression */:
                        return visitAssignmentPattern(node);
                    default:
                        return visitor(node);
                }
            }
            /**
             * If we visit a private name, this means it is an undeclared private name.
             * Replace it with an empty identifier to indicate a problem with the code.
             */
            function visitPrivateIdentifier(node) {
                if (!shouldTransformPrivateElementsOrClassStaticBlocks) {
                    return node;
                }
                return ts.setOriginalNode(factory.createIdentifier(""), node);
            }
            /**
             * Visits the members of a class that has fields.
             *
             * @param node The node to visit.
             */
            function classElementVisitor(node) {
                switch (node.kind) {
                    case 169 /* Constructor */:
                        // Constructors for classes using class fields are transformed in
                        // `visitClassDeclaration` or `visitClassExpression`.
                        return undefined;
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 167 /* MethodDeclaration */:
                        return visitMethodOrAccessorDeclaration(node);
                    case 165 /* PropertyDeclaration */:
                        return visitPropertyDeclaration(node);
                    case 160 /* ComputedPropertyName */:
                        return visitComputedPropertyName(node);
                    case 232 /* SemicolonClassElement */:
                        return node;
                    default:
                        return visitor(node);
                }
            }
            function visitVariableStatement(node) {
                var savedPendingStatements = pendingStatements;
                pendingStatements = [];
                var visitedNode = ts.visitEachChild(node, visitor, context);
                var statement = ts.some(pendingStatements) ? __spreadArray([visitedNode], pendingStatements, true) :
                    visitedNode;
                pendingStatements = savedPendingStatements;
                return statement;
            }
            function visitComputedPropertyName(name) {
                var node = ts.visitEachChild(name, visitor, context);
                if (ts.some(pendingExpressions)) {
                    var expressions = pendingExpressions;
                    expressions.push(node.expression);
                    pendingExpressions = [];
                    node = factory.updateComputedPropertyName(node, factory.inlineExpressions(expressions));
                }
                return node;
            }
            function visitMethodOrAccessorDeclaration(node) {
                ts.Debug.assert(!ts.some(node.decorators));
                if (!shouldTransformPrivateElementsOrClassStaticBlocks || !ts.isPrivateIdentifier(node.name)) {
                    return ts.visitEachChild(node, classElementVisitor, context);
                }
                // leave invalid code untransformed
                var info = accessPrivateIdentifier(node.name);
                ts.Debug.assert(info, "Undeclared private name for property declaration.");
                if (!info.isValid) {
                    return node;
                }
                var functionName = getHoistedFunctionName(node);
                if (functionName) {
                    getPendingExpressions().push(factory.createAssignment(functionName, factory.createFunctionExpression(ts.filter(node.modifiers, function (m) { return !ts.isStaticModifier(m); }), node.asteriskToken, functionName, 
                    /* typeParameters */ undefined, ts.visitParameterList(node.parameters, classElementVisitor, context), 
                    /* type */ undefined, ts.visitFunctionBody(node.body, classElementVisitor, context))));
                }
                // remove method declaration from class
                return undefined;
            }
            function getHoistedFunctionName(node) {
                ts.Debug.assert(ts.isPrivateIdentifier(node.name));
                var info = accessPrivateIdentifier(node.name);
                ts.Debug.assert(info, "Undeclared private name for property declaration.");
                if (info.kind === "m" /* Method */) {
                    return info.methodName;
                }
                if (info.kind === "a" /* Accessor */) {
                    if (ts.isGetAccessor(node)) {
                        return info.getterName;
                    }
                    if (ts.isSetAccessor(node)) {
                        return info.setterName;
                    }
                }
            }
            function visitPropertyDeclaration(node) {
                ts.Debug.assert(!ts.some(node.decorators));
                if (ts.isPrivateIdentifier(node.name)) {
                    if (!shouldTransformPrivateElementsOrClassStaticBlocks) {
                        // Initializer is elided as the field is initialized in transformConstructor.
                        return factory.updatePropertyDeclaration(node, 
                        /*decorators*/ undefined, ts.visitNodes(node.modifiers, visitor, ts.isModifier), node.name, 
                        /*questionOrExclamationToken*/ undefined, 
                        /*type*/ undefined, 
                        /*initializer*/ undefined);
                    }
                    // leave invalid code untransformed
                    var info = accessPrivateIdentifier(node.name);
                    ts.Debug.assert(info, "Undeclared private name for property declaration.");
                    if (!info.isValid) {
                        return node;
                    }
                }
                // Create a temporary variable to store a computed property name (if necessary).
                // If it's not inlineable, then we emit an expression after the class which assigns
                // the property name to the temporary variable.
                var expr = getPropertyNameExpressionIfNeeded(node.name, !!node.initializer || useDefineForClassFields);
                if (expr && !ts.isSimpleInlineableExpression(expr)) {
                    getPendingExpressions().push(expr);
                }
                return undefined;
            }
            function createPrivateIdentifierAccess(info, receiver) {
                return createPrivateIdentifierAccessHelper(info, ts.visitNode(receiver, visitor, ts.isExpression));
            }
            function createPrivateIdentifierAccessHelper(info, receiver) {
                ts.setCommentRange(receiver, ts.moveRangePos(receiver, -1));
                switch (info.kind) {
                    case "a" /* Accessor */:
                        return context.getEmitHelperFactory().createClassPrivateFieldGetHelper(receiver, info.brandCheckIdentifier, info.kind, info.getterName);
                    case "m" /* Method */:
                        return context.getEmitHelperFactory().createClassPrivateFieldGetHelper(receiver, info.brandCheckIdentifier, info.kind, info.methodName);
                    case "f" /* Field */:
                        return context.getEmitHelperFactory().createClassPrivateFieldGetHelper(receiver, info.brandCheckIdentifier, info.kind, info.variableName);
                    default:
                        ts.Debug.assertNever(info, "Unknown private element type");
                }
            }
            function visitPropertyAccessExpression(node) {
                if (shouldTransformPrivateElementsOrClassStaticBlocks && ts.isPrivateIdentifier(node.name)) {
                    var privateIdentifierInfo = accessPrivateIdentifier(node.name);
                    if (privateIdentifierInfo) {
                        return ts.setTextRange(ts.setOriginalNode(createPrivateIdentifierAccess(privateIdentifierInfo, node.expression), node), node);
                    }
                }
                if (shouldTransformSuperInStaticInitializers &&
                    ts.isSuperProperty(node) &&
                    ts.isIdentifier(node.name) &&
                    currentStaticPropertyDeclarationOrStaticBlock &&
                    currentClassLexicalEnvironment) {
                    var classConstructor = currentClassLexicalEnvironment.classConstructor, superClassReference = currentClassLexicalEnvironment.superClassReference, facts = currentClassLexicalEnvironment.facts;
                    if (facts & 1 /* ClassWasDecorated */) {
                        return visitInvalidSuperProperty(node);
                    }
                    if (classConstructor && superClassReference) {
                        // converts `super.x` into `Reflect.get(_baseTemp, "x", _classTemp)`
                        var superProperty = factory.createReflectGetCall(superClassReference, factory.createStringLiteralFromNode(node.name), classConstructor);
                        ts.setOriginalNode(superProperty, node.expression);
                        ts.setTextRange(superProperty, node.expression);
                        return superProperty;
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitElementAccessExpression(node) {
                if (shouldTransformSuperInStaticInitializers &&
                    ts.isSuperProperty(node) &&
                    currentStaticPropertyDeclarationOrStaticBlock &&
                    currentClassLexicalEnvironment) {
                    var classConstructor = currentClassLexicalEnvironment.classConstructor, superClassReference = currentClassLexicalEnvironment.superClassReference, facts = currentClassLexicalEnvironment.facts;
                    if (facts & 1 /* ClassWasDecorated */) {
                        return visitInvalidSuperProperty(node);
                    }
                    if (classConstructor && superClassReference) {
                        // converts `super[x]` into `Reflect.get(_baseTemp, x, _classTemp)`
                        var superProperty = factory.createReflectGetCall(superClassReference, ts.visitNode(node.argumentExpression, visitor, ts.isExpression), classConstructor);
                        ts.setOriginalNode(superProperty, node.expression);
                        ts.setTextRange(superProperty, node.expression);
                        return superProperty;
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitPreOrPostfixUnaryExpression(node, valueIsDiscarded) {
                if (node.operator === 45 /* PlusPlusToken */ || node.operator === 46 /* MinusMinusToken */) {
                    if (shouldTransformPrivateElementsOrClassStaticBlocks && ts.isPrivateIdentifierPropertyAccessExpression(node.operand)) {
                        var info = void 0;
                        if (info = accessPrivateIdentifier(node.operand.name)) {
                            var receiver = ts.visitNode(node.operand.expression, visitor, ts.isExpression);
                            var _a = createCopiableReceiverExpr(receiver), readExpression = _a.readExpression, initializeExpression = _a.initializeExpression;
                            var expression = createPrivateIdentifierAccess(info, readExpression);
                            var temp = ts.isPrefixUnaryExpression(node) || valueIsDiscarded ? undefined : factory.createTempVariable(hoistVariableDeclaration);
                            expression = ts.expandPreOrPostfixIncrementOrDecrementExpression(factory, node, expression, hoistVariableDeclaration, temp);
                            expression = createPrivateIdentifierAssignment(info, initializeExpression || readExpression, expression, 63 /* EqualsToken */);
                            ts.setOriginalNode(expression, node);
                            ts.setTextRange(expression, node);
                            if (temp) {
                                expression = factory.createComma(expression, temp);
                                ts.setTextRange(expression, node);
                            }
                            return expression;
                        }
                    }
                    else if (shouldTransformSuperInStaticInitializers &&
                        ts.isSuperProperty(node.operand) &&
                        currentStaticPropertyDeclarationOrStaticBlock &&
                        currentClassLexicalEnvironment) {
                        // converts `++super.a` into `(Reflect.set(_baseTemp, "a", (_a = Reflect.get(_baseTemp, "a", _classTemp), _b = ++_a), _classTemp), _b)`
                        // converts `++super[f()]` into `(Reflect.set(_baseTemp, _a = f(), (_b = Reflect.get(_baseTemp, _a, _classTemp), _c = ++_b), _classTemp), _c)`
                        // converts `--super.a` into `(Reflect.set(_baseTemp, "a", (_a = Reflect.get(_baseTemp, "a", _classTemp), _b = --_a), _classTemp), _b)`
                        // converts `--super[f()]` into `(Reflect.set(_baseTemp, _a = f(), (_b = Reflect.get(_baseTemp, _a, _classTemp), _c = --_b), _classTemp), _c)`
                        // converts `super.a++` into `(Reflect.set(_baseTemp, "a", (_a = Reflect.get(_baseTemp, "a", _classTemp), _b = _a++), _classTemp), _b)`
                        // converts `super[f()]++` into `(Reflect.set(_baseTemp, _a = f(), (_b = Reflect.get(_baseTemp, _a, _classTemp), _c = _b++), _classTemp), _c)`
                        // converts `super.a--` into `(Reflect.set(_baseTemp, "a", (_a = Reflect.get(_baseTemp, "a", _classTemp), _b = _a--), _classTemp), _b)`
                        // converts `super[f()]--` into `(Reflect.set(_baseTemp, _a = f(), (_b = Reflect.get(_baseTemp, _a, _classTemp), _c = _b--), _classTemp), _c)`
                        var classConstructor = currentClassLexicalEnvironment.classConstructor, superClassReference = currentClassLexicalEnvironment.superClassReference, facts = currentClassLexicalEnvironment.facts;
                        if (facts & 1 /* ClassWasDecorated */) {
                            var operand = visitInvalidSuperProperty(node.operand);
                            return ts.isPrefixUnaryExpression(node) ?
                                factory.updatePrefixUnaryExpression(node, operand) :
                                factory.updatePostfixUnaryExpression(node, operand);
                        }
                        if (classConstructor && superClassReference) {
                            var setterName = void 0;
                            var getterName = void 0;
                            if (ts.isPropertyAccessExpression(node.operand)) {
                                if (ts.isIdentifier(node.operand.name)) {
                                    getterName = setterName = factory.createStringLiteralFromNode(node.operand.name);
                                }
                            }
                            else {
                                if (ts.isSimpleInlineableExpression(node.operand.argumentExpression)) {
                                    getterName = setterName = node.operand.argumentExpression;
                                }
                                else {
                                    getterName = factory.createTempVariable(hoistVariableDeclaration);
                                    setterName = factory.createAssignment(getterName, ts.visitNode(node.operand.argumentExpression, visitor, ts.isExpression));
                                }
                            }
                            if (setterName && getterName) {
                                var expression = factory.createReflectGetCall(superClassReference, getterName, classConstructor);
                                ts.setTextRange(expression, node.operand);
                                var temp = valueIsDiscarded ? undefined : factory.createTempVariable(hoistVariableDeclaration);
                                expression = ts.expandPreOrPostfixIncrementOrDecrementExpression(factory, node, expression, hoistVariableDeclaration, temp);
                                expression = factory.createReflectSetCall(superClassReference, setterName, expression, classConstructor);
                                ts.setOriginalNode(expression, node);
                                ts.setTextRange(expression, node);
                                if (temp) {
                                    expression = factory.createComma(expression, temp);
                                    ts.setTextRange(expression, node);
                                }
                                return expression;
                            }
                        }
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitForStatement(node) {
                return factory.updateForStatement(node, ts.visitNode(node.initializer, discardedValueVisitor, ts.isForInitializer), ts.visitNode(node.condition, visitor, ts.isExpression), ts.visitNode(node.incrementor, discardedValueVisitor, ts.isExpression), ts.visitIterationBody(node.statement, visitor, context));
            }
            function visitExpressionStatement(node) {
                return factory.updateExpressionStatement(node, ts.visitNode(node.expression, discardedValueVisitor, ts.isExpression));
            }
            function createCopiableReceiverExpr(receiver) {
                var clone = ts.nodeIsSynthesized(receiver) ? receiver : factory.cloneNode(receiver);
                if (ts.isSimpleInlineableExpression(receiver)) {
                    return { readExpression: clone, initializeExpression: undefined };
                }
                var readExpression = factory.createTempVariable(hoistVariableDeclaration);
                var initializeExpression = factory.createAssignment(readExpression, clone);
                return { readExpression: readExpression, initializeExpression: initializeExpression };
            }
            function visitCallExpression(node) {
                if (shouldTransformPrivateElementsOrClassStaticBlocks && ts.isPrivateIdentifierPropertyAccessExpression(node.expression)) {
                    // Transform call expressions of private names to properly bind the `this` parameter.
                    var _a = factory.createCallBinding(node.expression, hoistVariableDeclaration, languageVersion), thisArg = _a.thisArg, target = _a.target;
                    if (ts.isCallChain(node)) {
                        return factory.updateCallChain(node, factory.createPropertyAccessChain(ts.visitNode(target, visitor), node.questionDotToken, "call"), 
                        /*questionDotToken*/ undefined, 
                        /*typeArguments*/ undefined, __spreadArray([ts.visitNode(thisArg, visitor, ts.isExpression)], ts.visitNodes(node.arguments, visitor, ts.isExpression), true));
                    }
                    return factory.updateCallExpression(node, factory.createPropertyAccessExpression(ts.visitNode(target, visitor), "call"), 
                    /*typeArguments*/ undefined, __spreadArray([ts.visitNode(thisArg, visitor, ts.isExpression)], ts.visitNodes(node.arguments, visitor, ts.isExpression), true));
                }
                if (shouldTransformSuperInStaticInitializers &&
                    ts.isSuperProperty(node.expression) &&
                    currentStaticPropertyDeclarationOrStaticBlock &&
                    (currentClassLexicalEnvironment === null || currentClassLexicalEnvironment === void 0 ? void 0 : currentClassLexicalEnvironment.classConstructor)) {
                    // converts `super.f(...)` into `Reflect.get(_baseTemp, "f", _classTemp).call(_classTemp, ...)`
                    var invocation = factory.createFunctionCallCall(ts.visitNode(node.expression, visitor, ts.isExpression), currentClassLexicalEnvironment.classConstructor, ts.visitNodes(node.arguments, visitor, ts.isExpression));
                    ts.setOriginalNode(invocation, node);
                    ts.setTextRange(invocation, node);
                    return invocation;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitTaggedTemplateExpression(node) {
                if (shouldTransformPrivateElementsOrClassStaticBlocks && ts.isPrivateIdentifierPropertyAccessExpression(node.tag)) {
                    // Bind the `this` correctly for tagged template literals when the tag is a private identifier property access.
                    var _a = factory.createCallBinding(node.tag, hoistVariableDeclaration, languageVersion), thisArg = _a.thisArg, target = _a.target;
                    return factory.updateTaggedTemplateExpression(node, factory.createCallExpression(factory.createPropertyAccessExpression(ts.visitNode(target, visitor), "bind"), 
                    /*typeArguments*/ undefined, [ts.visitNode(thisArg, visitor, ts.isExpression)]), 
                    /*typeArguments*/ undefined, ts.visitNode(node.template, visitor, ts.isTemplateLiteral));
                }
                if (shouldTransformSuperInStaticInitializers &&
                    ts.isSuperProperty(node.tag) &&
                    currentStaticPropertyDeclarationOrStaticBlock &&
                    (currentClassLexicalEnvironment === null || currentClassLexicalEnvironment === void 0 ? void 0 : currentClassLexicalEnvironment.classConstructor)) {
                    // converts `` super.f`x` `` into `` Reflect.get(_baseTemp, "f", _classTemp).bind(_classTemp)`x` ``
                    var invocation = factory.createFunctionBindCall(ts.visitNode(node.tag, visitor, ts.isExpression), currentClassLexicalEnvironment.classConstructor, []);
                    ts.setOriginalNode(invocation, node);
                    ts.setTextRange(invocation, node);
                    return factory.updateTaggedTemplateExpression(node, invocation, 
                    /*typeArguments*/ undefined, ts.visitNode(node.template, visitor, ts.isTemplateLiteral));
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function transformClassStaticBlockDeclaration(node) {
                if (shouldTransformPrivateElementsOrClassStaticBlocks) {
                    if (currentClassLexicalEnvironment) {
                        classLexicalEnvironmentMap.set(ts.getOriginalNodeId(node), currentClassLexicalEnvironment);
                    }
                    startLexicalEnvironment();
                    var savedCurrentStaticPropertyDeclarationOrStaticBlock = currentStaticPropertyDeclarationOrStaticBlock;
                    currentStaticPropertyDeclarationOrStaticBlock = node;
                    var statements = ts.visitNodes(node.body.statements, visitor, ts.isStatement);
                    statements = factory.mergeLexicalEnvironment(statements, endLexicalEnvironment());
                    currentStaticPropertyDeclarationOrStaticBlock = savedCurrentStaticPropertyDeclarationOrStaticBlock;
                    var iife = factory.createImmediatelyInvokedArrowFunction(statements);
                    ts.setOriginalNode(iife, node);
                    ts.setTextRange(iife, node);
                    ts.addEmitFlags(iife, 2 /* AdviseOnEmitNode */);
                    return iife;
                }
            }
            function visitBinaryExpression(node, valueIsDiscarded) {
                if (ts.isDestructuringAssignment(node)) {
                    var savedPendingExpressions = pendingExpressions;
                    pendingExpressions = undefined;
                    node = factory.updateBinaryExpression(node, ts.visitNode(node.left, visitorDestructuringTarget), node.operatorToken, ts.visitNode(node.right, visitor));
                    var expr = ts.some(pendingExpressions) ?
                        factory.inlineExpressions(ts.compact(__spreadArray(__spreadArray([], pendingExpressions, true), [node], false))) :
                        node;
                    pendingExpressions = savedPendingExpressions;
                    return expr;
                }
                if (ts.isAssignmentExpression(node)) {
                    if (shouldTransformPrivateElementsOrClassStaticBlocks && ts.isPrivateIdentifierPropertyAccessExpression(node.left)) {
                        var info = accessPrivateIdentifier(node.left.name);
                        if (info) {
                            return ts.setTextRange(ts.setOriginalNode(createPrivateIdentifierAssignment(info, node.left.expression, node.right, node.operatorToken.kind), node), node);
                        }
                    }
                    else if (shouldTransformSuperInStaticInitializers &&
                        ts.isSuperProperty(node.left) &&
                        currentStaticPropertyDeclarationOrStaticBlock &&
                        currentClassLexicalEnvironment) {
                        var classConstructor = currentClassLexicalEnvironment.classConstructor, superClassReference = currentClassLexicalEnvironment.superClassReference, facts = currentClassLexicalEnvironment.facts;
                        if (facts & 1 /* ClassWasDecorated */) {
                            return factory.updateBinaryExpression(node, visitInvalidSuperProperty(node.left), node.operatorToken, ts.visitNode(node.right, visitor, ts.isExpression));
                        }
                        if (classConstructor && superClassReference) {
                            var setterName = ts.isElementAccessExpression(node.left) ? ts.visitNode(node.left.argumentExpression, visitor, ts.isExpression) :
                                ts.isIdentifier(node.left.name) ? factory.createStringLiteralFromNode(node.left.name) :
                                    undefined;
                            if (setterName) {
                                // converts `super.x = 1` into `(Reflect.set(_baseTemp, "x", _a = 1, _classTemp), _a)`
                                // converts `super[f()] = 1` into `(Reflect.set(_baseTemp, f(), _a = 1, _classTemp), _a)`
                                // converts `super.x += 1` into `(Reflect.set(_baseTemp, "x", _a = Reflect.get(_baseTemp, "x", _classtemp) + 1, _classTemp), _a)`
                                // converts `super[f()] += 1` into `(Reflect.set(_baseTemp, _a = f(), _b = Reflect.get(_baseTemp, _a, _classtemp) + 1, _classTemp), _b)`
                                var expression = ts.visitNode(node.right, visitor, ts.isExpression);
                                if (ts.isCompoundAssignment(node.operatorToken.kind)) {
                                    var getterName = setterName;
                                    if (!ts.isSimpleInlineableExpression(setterName)) {
                                        getterName = factory.createTempVariable(hoistVariableDeclaration);
                                        setterName = factory.createAssignment(getterName, setterName);
                                    }
                                    var superPropertyGet = factory.createReflectGetCall(superClassReference, getterName, classConstructor);
                                    ts.setOriginalNode(superPropertyGet, node.left);
                                    ts.setTextRange(superPropertyGet, node.left);
                                    expression = factory.createBinaryExpression(superPropertyGet, ts.getNonAssignmentOperatorForCompoundAssignment(node.operatorToken.kind), expression);
                                    ts.setTextRange(expression, node);
                                }
                                var temp = valueIsDiscarded ? undefined : factory.createTempVariable(hoistVariableDeclaration);
                                if (temp) {
                                    expression = factory.createAssignment(temp, expression);
                                    ts.setTextRange(temp, node);
                                }
                                expression = factory.createReflectSetCall(superClassReference, setterName, expression, classConstructor);
                                ts.setOriginalNode(expression, node);
                                ts.setTextRange(expression, node);
                                if (temp) {
                                    expression = factory.createComma(expression, temp);
                                    ts.setTextRange(expression, node);
                                }
                                return expression;
                            }
                        }
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function createPrivateIdentifierAssignment(info, receiver, right, operator) {
                receiver = ts.visitNode(receiver, visitor, ts.isExpression);
                right = ts.visitNode(right, visitor, ts.isExpression);
                if (ts.isCompoundAssignment(operator)) {
                    var _a = createCopiableReceiverExpr(receiver), readExpression = _a.readExpression, initializeExpression = _a.initializeExpression;
                    receiver = initializeExpression || readExpression;
                    right = factory.createBinaryExpression(createPrivateIdentifierAccessHelper(info, readExpression), ts.getNonAssignmentOperatorForCompoundAssignment(operator), right);
                }
                ts.setCommentRange(receiver, ts.moveRangePos(receiver, -1));
                switch (info.kind) {
                    case "a" /* Accessor */:
                        return context.getEmitHelperFactory().createClassPrivateFieldSetHelper(receiver, info.brandCheckIdentifier, right, info.kind, info.setterName);
                    case "m" /* Method */:
                        return context.getEmitHelperFactory().createClassPrivateFieldSetHelper(receiver, info.brandCheckIdentifier, right, info.kind, 
                        /* f */ undefined);
                    case "f" /* Field */:
                        return context.getEmitHelperFactory().createClassPrivateFieldSetHelper(receiver, info.brandCheckIdentifier, right, info.kind, info.variableName);
                    default:
                        ts.Debug.assertNever(info, "Unknown private element type");
                }
            }
            /**
             * Set up the environment for a class.
             */
            function visitClassLike(node) {
                if (!ts.forEach(node.members, doesClassElementNeedTransform)) {
                    return ts.visitEachChild(node, visitor, context);
                }
                var savedPendingExpressions = pendingExpressions;
                pendingExpressions = undefined;
                startClassLexicalEnvironment();
                if (shouldTransformPrivateElementsOrClassStaticBlocks) {
                    var name = ts.getNameOfDeclaration(node);
                    if (name && ts.isIdentifier(name)) {
                        getPrivateIdentifierEnvironment().className = ts.idText(name);
                    }
                    var privateInstanceMethodsAndAccessors = getPrivateInstanceMethodsAndAccessors(node);
                    if (ts.some(privateInstanceMethodsAndAccessors)) {
                        getPrivateIdentifierEnvironment().weakSetName = createHoistedVariableForClass("instances", privateInstanceMethodsAndAccessors[0].name);
                    }
                }
                var result = ts.isClassDeclaration(node) ?
                    visitClassDeclaration(node) :
                    visitClassExpression(node);
                endClassLexicalEnvironment();
                pendingExpressions = savedPendingExpressions;
                return result;
            }
            function doesClassElementNeedTransform(node) {
                return ts.isPropertyDeclaration(node) || ts.isClassStaticBlockDeclaration(node) || (shouldTransformPrivateElementsOrClassStaticBlocks && node.name && ts.isPrivateIdentifier(node.name));
            }
            function getPrivateInstanceMethodsAndAccessors(node) {
                return ts.filter(node.members, ts.isNonStaticMethodOrAccessorWithPrivateName);
            }
            function getClassFacts(node) {
                var facts = 0 /* None */;
                var original = ts.getOriginalNode(node);
                if (ts.isClassDeclaration(original) && ts.classOrConstructorParameterIsDecorated(original)) {
                    facts |= 1 /* ClassWasDecorated */;
                }
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    if (!ts.isStatic(member))
                        continue;
                    if (member.name && ts.isPrivateIdentifier(member.name) && shouldTransformPrivateElementsOrClassStaticBlocks) {
                        facts |= 2 /* NeedsClassConstructorReference */;
                    }
                    if (ts.isPropertyDeclaration(member) || ts.isClassStaticBlockDeclaration(member)) {
                        if (shouldTransformThisInStaticInitializers && member.transformFlags & 8192 /* ContainsLexicalThis */) {
                            facts |= 8 /* NeedsSubstitutionForThisInClassStaticField */;
                            if (!(facts & 1 /* ClassWasDecorated */)) {
                                facts |= 2 /* NeedsClassConstructorReference */;
                            }
                        }
                        if (shouldTransformSuperInStaticInitializers && member.transformFlags & 33554432 /* ContainsLexicalSuper */) {
                            if (!(facts & 1 /* ClassWasDecorated */)) {
                                facts |= 2 /* NeedsClassConstructorReference */ | 4 /* NeedsClassSuperReference */;
                            }
                        }
                    }
                }
                return facts;
            }
            function visitExpressionWithTypeArguments(node) {
                var facts = (currentClassLexicalEnvironment === null || currentClassLexicalEnvironment === void 0 ? void 0 : currentClassLexicalEnvironment.facts) || 0 /* None */;
                if (facts & 4 /* NeedsClassSuperReference */) {
                    var temp = factory.createTempVariable(hoistVariableDeclaration, /*reserveInNestedScopes*/ true);
                    getClassLexicalEnvironment().superClassReference = temp;
                    return factory.updateExpressionWithTypeArguments(node, factory.createAssignment(temp, ts.visitNode(node.expression, visitor, ts.isExpression)), 
                    /*typeArguments*/ undefined);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitClassDeclaration(node) {
                var facts = getClassFacts(node);
                if (facts) {
                    getClassLexicalEnvironment().facts = facts;
                }
                if (facts & 8 /* NeedsSubstitutionForThisInClassStaticField */) {
                    enableSubstitutionForClassStaticThisOrSuperReference();
                }
                var staticProperties = ts.getStaticPropertiesAndClassStaticBlock(node);
                // If a class has private static fields, or a static field has a `this` or `super` reference,
                // then we need to allocate a temp variable to hold on to that reference.
                var pendingClassReferenceAssignment;
                if (facts & 2 /* NeedsClassConstructorReference */) {
                    var temp = factory.createTempVariable(hoistVariableDeclaration, /*reservedInNestedScopes*/ true);
                    getClassLexicalEnvironment().classConstructor = factory.cloneNode(temp);
                    pendingClassReferenceAssignment = factory.createAssignment(temp, factory.getInternalName(node));
                }
                var extendsClauseElement = ts.getEffectiveBaseTypeNode(node);
                var isDerivedClass = !!(extendsClauseElement && ts.skipOuterExpressions(extendsClauseElement.expression).kind !== 104 /* NullKeyword */);
                var statements = [
                    factory.updateClassDeclaration(node, 
                    /*decorators*/ undefined, node.modifiers, node.name, 
                    /*typeParameters*/ undefined, ts.visitNodes(node.heritageClauses, heritageClauseVisitor, ts.isHeritageClause), transformClassMembers(node, isDerivedClass))
                ];
                if (pendingClassReferenceAssignment) {
                    getPendingExpressions().unshift(pendingClassReferenceAssignment);
                }
                // Write any pending expressions from elided or moved computed property names
                if (ts.some(pendingExpressions)) {
                    statements.push(factory.createExpressionStatement(factory.inlineExpressions(pendingExpressions)));
                }
                // Emit static property assignment. Because classDeclaration is lexically evaluated,
                // it is safe to emit static property assignment after classDeclaration
                // From ES6 specification:
                //      HasLexicalDeclaration (N) : Determines if the argument identifier has a binding in this environment record that was created using
                //                                  a lexical declaration such as a LexicalDeclaration or a ClassDeclaration.
                if (ts.some(staticProperties)) {
                    addPropertyOrClassStaticBlockStatements(statements, staticProperties, factory.getInternalName(node));
                }
                return statements;
            }
            function visitClassExpression(node) {
                var facts = getClassFacts(node);
                if (facts) {
                    getClassLexicalEnvironment().facts = facts;
                }
                if (facts & 8 /* NeedsSubstitutionForThisInClassStaticField */) {
                    enableSubstitutionForClassStaticThisOrSuperReference();
                }
                // If this class expression is a transformation of a decorated class declaration,
                // then we want to output the pendingExpressions as statements, not as inlined
                // expressions with the class statement.
                //
                // In this case, we use pendingStatements to produce the same output as the
                // class declaration transformation. The VariableStatement visitor will insert
                // these statements after the class expression variable statement.
                var isDecoratedClassDeclaration = !!(facts & 1 /* ClassWasDecorated */);
                var staticPropertiesOrClassStaticBlocks = ts.getStaticPropertiesAndClassStaticBlock(node);
                var extendsClauseElement = ts.getEffectiveBaseTypeNode(node);
                var isDerivedClass = !!(extendsClauseElement && ts.skipOuterExpressions(extendsClauseElement.expression).kind !== 104 /* NullKeyword */);
                var isClassWithConstructorReference = resolver.getNodeCheckFlags(node) & 16777216 /* ClassWithConstructorReference */;
                var temp;
                function createClassTempVar() {
                    var classCheckFlags = resolver.getNodeCheckFlags(node);
                    var isClassWithConstructorReference = classCheckFlags & 16777216 /* ClassWithConstructorReference */;
                    var requiresBlockScopedVar = classCheckFlags & 524288 /* BlockScopedBindingInLoop */;
                    return factory.createTempVariable(requiresBlockScopedVar ? addBlockScopedVariable : hoistVariableDeclaration, !!isClassWithConstructorReference);
                }
                if (facts & 2 /* NeedsClassConstructorReference */) {
                    temp = createClassTempVar();
                    getClassLexicalEnvironment().classConstructor = factory.cloneNode(temp);
                }
                var classExpression = factory.updateClassExpression(node, ts.visitNodes(node.decorators, visitor, ts.isDecorator), node.modifiers, node.name, 
                /*typeParameters*/ undefined, ts.visitNodes(node.heritageClauses, heritageClauseVisitor, ts.isHeritageClause), transformClassMembers(node, isDerivedClass));
                var hasTransformableStatics = ts.some(staticPropertiesOrClassStaticBlocks, function (p) { return ts.isClassStaticBlockDeclaration(p) || !!p.initializer || (shouldTransformPrivateElementsOrClassStaticBlocks && ts.isPrivateIdentifier(p.name)); });
                if (hasTransformableStatics || ts.some(pendingExpressions)) {
                    if (isDecoratedClassDeclaration) {
                        ts.Debug.assertIsDefined(pendingStatements, "Decorated classes transformed by TypeScript are expected to be within a variable declaration.");
                        // Write any pending expressions from elided or moved computed property names
                        if (pendingStatements && pendingExpressions && ts.some(pendingExpressions)) {
                            pendingStatements.push(factory.createExpressionStatement(factory.inlineExpressions(pendingExpressions)));
                        }
                        if (pendingStatements && ts.some(staticPropertiesOrClassStaticBlocks)) {
                            addPropertyOrClassStaticBlockStatements(pendingStatements, staticPropertiesOrClassStaticBlocks, factory.getInternalName(node));
                        }
                        if (temp) {
                            return factory.inlineExpressions([factory.createAssignment(temp, classExpression), temp]);
                        }
                        return classExpression;
                    }
                    else {
                        var expressions = [];
                        temp || (temp = createClassTempVar());
                        if (isClassWithConstructorReference) {
                            // record an alias as the class name is not in scope for statics.
                            enableSubstitutionForClassAliases();
                            var alias = factory.cloneNode(temp);
                            alias.autoGenerateFlags &= ~8 /* ReservedInNestedScopes */;
                            classAliases[ts.getOriginalNodeId(node)] = alias;
                        }
                        // To preserve the behavior of the old emitter, we explicitly indent
                        // the body of a class with static initializers.
                        ts.setEmitFlags(classExpression, 65536 /* Indented */ | ts.getEmitFlags(classExpression));
                        expressions.push(ts.startOnNewLine(factory.createAssignment(temp, classExpression)));
                        // Add any pending expressions leftover from elided or relocated computed property names
                        ts.addRange(expressions, ts.map(pendingExpressions, ts.startOnNewLine));
                        ts.addRange(expressions, generateInitializedPropertyExpressionsOrClassStaticBlock(staticPropertiesOrClassStaticBlocks, temp));
                        expressions.push(ts.startOnNewLine(temp));
                        return factory.inlineExpressions(expressions);
                    }
                }
                return classExpression;
            }
            function visitClassStaticBlockDeclaration(node) {
                if (!shouldTransformPrivateElementsOrClassStaticBlocks) {
                    return ts.visitEachChild(node, classElementVisitor, context);
                }
                // ClassStaticBlockDeclaration for classes are transformed in `visitClassDeclaration` or `visitClassExpression`.
                return undefined;
            }
            function transformClassMembers(node, isDerivedClass) {
                if (shouldTransformPrivateElementsOrClassStaticBlocks) {
                    // Declare private names.
                    for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                        var member = _a[_i];
                        if (ts.isPrivateIdentifierClassElementDeclaration(member)) {
                            addPrivateIdentifierToEnvironment(member);
                        }
                    }
                    if (ts.some(getPrivateInstanceMethodsAndAccessors(node))) {
                        createBrandCheckWeakSetForPrivateMethods();
                    }
                }
                var members = [];
                var constructor = transformConstructor(node, isDerivedClass);
                if (constructor) {
                    members.push(constructor);
                }
                ts.addRange(members, ts.visitNodes(node.members, classElementVisitor, ts.isClassElement));
                return ts.setTextRange(factory.createNodeArray(members), /*location*/ node.members);
            }
            function createBrandCheckWeakSetForPrivateMethods() {
                var weakSetName = getPrivateIdentifierEnvironment().weakSetName;
                ts.Debug.assert(weakSetName, "weakSetName should be set in private identifier environment");
                getPendingExpressions().push(factory.createAssignment(weakSetName, factory.createNewExpression(factory.createIdentifier("WeakSet"), 
                /*typeArguments*/ undefined, [])));
            }
            function isClassElementThatRequiresConstructorStatement(member) {
                if (ts.isStatic(member) || ts.hasSyntacticModifier(ts.getOriginalNode(member), 128 /* Abstract */)) {
                    return false;
                }
                if (useDefineForClassFields) {
                    // If we are using define semantics and targeting ESNext or higher,
                    // then we don't need to transform any class properties.
                    return languageVersion < 99 /* ESNext */;
                }
                return ts.isInitializedProperty(member) || shouldTransformPrivateElementsOrClassStaticBlocks && ts.isPrivateIdentifierClassElementDeclaration(member);
            }
            function transformConstructor(node, isDerivedClass) {
                var constructor = ts.visitNode(ts.getFirstConstructorWithBody(node), visitor, ts.isConstructorDeclaration);
                var elements = node.members.filter(isClassElementThatRequiresConstructorStatement);
                if (!ts.some(elements)) {
                    return constructor;
                }
                var parameters = ts.visitParameterList(constructor ? constructor.parameters : undefined, visitor, context);
                var body = transformConstructorBody(node, constructor, isDerivedClass);
                if (!body) {
                    return undefined;
                }
                return ts.startOnNewLine(ts.setOriginalNode(ts.setTextRange(factory.createConstructorDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, parameters !== null && parameters !== void 0 ? parameters : [], body), constructor || node), constructor));
            }
            function transformConstructorBody(node, constructor, isDerivedClass) {
                var properties = ts.getProperties(node, /*requireInitializer*/ false, /*isStatic*/ false);
                if (!useDefineForClassFields) {
                    properties = ts.filter(properties, function (property) { return !!property.initializer || ts.isPrivateIdentifier(property.name); });
                }
                var privateMethodsAndAccessors = getPrivateInstanceMethodsAndAccessors(node);
                var needsConstructorBody = ts.some(properties) || ts.some(privateMethodsAndAccessors);
                // Only generate synthetic constructor when there are property initializers to move.
                if (!constructor && !needsConstructorBody) {
                    return ts.visitFunctionBody(/*node*/ undefined, visitor, context);
                }
                resumeLexicalEnvironment();
                var indexOfFirstStatement = 0;
                var statements = [];
                if (!constructor && isDerivedClass) {
                    // Add a synthetic `super` call:
                    //
                    //  super(...arguments);
                    //
                    statements.push(factory.createExpressionStatement(factory.createCallExpression(factory.createSuper(), 
                    /*typeArguments*/ undefined, [factory.createSpreadElement(factory.createIdentifier("arguments"))])));
                }
                if (constructor) {
                    indexOfFirstStatement = ts.addPrologueDirectivesAndInitialSuperCall(factory, constructor, statements, visitor);
                }
                // Add the property initializers. Transforms this:
                //
                //  public x = 1;
                //
                // Into this:
                //
                //  constructor() {
                //      this.x = 1;
                //  }
                //
                if (constructor === null || constructor === void 0 ? void 0 : constructor.body) {
                    var afterParameterProperties = ts.findIndex(constructor.body.statements, function (s) { return !ts.isParameterPropertyDeclaration(ts.getOriginalNode(s), constructor); }, indexOfFirstStatement);
                    if (afterParameterProperties === -1) {
                        afterParameterProperties = constructor.body.statements.length;
                    }
                    if (afterParameterProperties > indexOfFirstStatement) {
                        if (!useDefineForClassFields) {
                            ts.addRange(statements, ts.visitNodes(constructor.body.statements, visitor, ts.isStatement, indexOfFirstStatement, afterParameterProperties - indexOfFirstStatement));
                        }
                        indexOfFirstStatement = afterParameterProperties;
                    }
                }
                var receiver = factory.createThis();
                // private methods can be called in property initializers, they should execute first.
                addMethodStatements(statements, privateMethodsAndAccessors, receiver);
                addPropertyOrClassStaticBlockStatements(statements, properties, receiver);
                // Add existing statements, skipping the initial super call.
                if (constructor) {
                    ts.addRange(statements, ts.visitNodes(constructor.body.statements, visitor, ts.isStatement, indexOfFirstStatement));
                }
                statements = factory.mergeLexicalEnvironment(statements, endLexicalEnvironment());
                return ts.setTextRange(factory.createBlock(ts.setTextRange(factory.createNodeArray(statements), 
                /*location*/ constructor ? constructor.body.statements : node.members), 
                /*multiLine*/ true), 
                /*location*/ constructor ? constructor.body : undefined);
            }
            /**
             * Generates assignment statements for property initializers.
             *
             * @param properties An array of property declarations to transform.
             * @param receiver The receiver on which each property should be assigned.
             */
            function addPropertyOrClassStaticBlockStatements(statements, properties, receiver) {
                for (var _i = 0, properties_7 = properties; _i < properties_7.length; _i++) {
                    var property = properties_7[_i];
                    var expression = ts.isClassStaticBlockDeclaration(property) ?
                        transformClassStaticBlockDeclaration(property) :
                        transformProperty(property, receiver);
                    if (!expression) {
                        continue;
                    }
                    var statement = factory.createExpressionStatement(expression);
                    ts.setSourceMapRange(statement, ts.moveRangePastModifiers(property));
                    ts.setCommentRange(statement, property);
                    ts.setOriginalNode(statement, property);
                    statements.push(statement);
                }
            }
            /**
             * Generates assignment expressions for property initializers.
             *
             * @param propertiesOrClassStaticBlocks An array of property declarations to transform.
             * @param receiver The receiver on which each property should be assigned.
             */
            function generateInitializedPropertyExpressionsOrClassStaticBlock(propertiesOrClassStaticBlocks, receiver) {
                var expressions = [];
                for (var _i = 0, propertiesOrClassStaticBlocks_1 = propertiesOrClassStaticBlocks; _i < propertiesOrClassStaticBlocks_1.length; _i++) {
                    var property = propertiesOrClassStaticBlocks_1[_i];
                    var expression = ts.isClassStaticBlockDeclaration(property) ? transformClassStaticBlockDeclaration(property) : transformProperty(property, receiver);
                    if (!expression) {
                        continue;
                    }
                    ts.startOnNewLine(expression);
                    ts.setSourceMapRange(expression, ts.moveRangePastModifiers(property));
                    ts.setCommentRange(expression, property);
                    ts.setOriginalNode(expression, property);
                    expressions.push(expression);
                }
                return expressions;
            }
            /**
             * Transforms a property initializer into an assignment statement.
             *
             * @param property The property declaration.
             * @param receiver The object receiving the property assignment.
             */
            function transformProperty(property, receiver) {
                var savedCurrentStaticPropertyDeclarationOrStaticBlock = currentStaticPropertyDeclarationOrStaticBlock;
                var transformed = transformPropertyWorker(property, receiver);
                if (transformed && ts.hasStaticModifier(property) && (currentClassLexicalEnvironment === null || currentClassLexicalEnvironment === void 0 ? void 0 : currentClassLexicalEnvironment.facts)) {
                    // capture the lexical environment for the member
                    ts.setOriginalNode(transformed, property);
                    ts.addEmitFlags(transformed, 2 /* AdviseOnEmitNode */);
                    classLexicalEnvironmentMap.set(ts.getOriginalNodeId(transformed), currentClassLexicalEnvironment);
                }
                currentStaticPropertyDeclarationOrStaticBlock = savedCurrentStaticPropertyDeclarationOrStaticBlock;
                return transformed;
            }
            function transformPropertyWorker(property, receiver) {
                var _a;
                // We generate a name here in order to reuse the value cached by the relocated computed name expression (which uses the same generated name)
                var emitAssignment = !useDefineForClassFields;
                var propertyName = ts.isComputedPropertyName(property.name) && !ts.isSimpleInlineableExpression(property.name.expression)
                    ? factory.updateComputedPropertyName(property.name, factory.getGeneratedNameForNode(property.name))
                    : property.name;
                if (ts.hasStaticModifier(property)) {
                    currentStaticPropertyDeclarationOrStaticBlock = property;
                }
                if (shouldTransformPrivateElementsOrClassStaticBlocks && ts.isPrivateIdentifier(propertyName)) {
                    var privateIdentifierInfo = accessPrivateIdentifier(propertyName);
                    if (privateIdentifierInfo) {
                        if (privateIdentifierInfo.kind === "f" /* Field */) {
                            if (!privateIdentifierInfo.isStatic) {
                                return createPrivateInstanceFieldInitializer(receiver, ts.visitNode(property.initializer, visitor, ts.isExpression), privateIdentifierInfo.brandCheckIdentifier);
                            }
                            else {
                                return createPrivateStaticFieldInitializer(privateIdentifierInfo.variableName, ts.visitNode(property.initializer, visitor, ts.isExpression));
                            }
                        }
                        else {
                            return undefined;
                        }
                    }
                    else {
                        ts.Debug.fail("Undeclared private name for property declaration.");
                    }
                }
                if ((ts.isPrivateIdentifier(propertyName) || ts.hasStaticModifier(property)) && !property.initializer) {
                    return undefined;
                }
                var propertyOriginalNode = ts.getOriginalNode(property);
                if (ts.hasSyntacticModifier(propertyOriginalNode, 128 /* Abstract */)) {
                    return undefined;
                }
                var initializer = property.initializer || emitAssignment ? (_a = ts.visitNode(property.initializer, visitor, ts.isExpression)) !== null && _a !== void 0 ? _a : factory.createVoidZero()
                    : ts.isParameterPropertyDeclaration(propertyOriginalNode, propertyOriginalNode.parent) && ts.isIdentifier(propertyName) ? propertyName
                        : factory.createVoidZero();
                if (emitAssignment || ts.isPrivateIdentifier(propertyName)) {
                    var memberAccess = ts.createMemberAccessForPropertyName(factory, receiver, propertyName, /*location*/ propertyName);
                    return factory.createAssignment(memberAccess, initializer);
                }
                else {
                    var name = ts.isComputedPropertyName(propertyName) ? propertyName.expression
                        : ts.isIdentifier(propertyName) ? factory.createStringLiteral(ts.unescapeLeadingUnderscores(propertyName.escapedText))
                            : propertyName;
                    var descriptor = factory.createPropertyDescriptor({ value: initializer, configurable: true, writable: true, enumerable: true });
                    return factory.createObjectDefinePropertyCall(receiver, name, descriptor);
                }
            }
            function enableSubstitutionForClassAliases() {
                if ((enabledSubstitutions & 1 /* ClassAliases */) === 0) {
                    enabledSubstitutions |= 1 /* ClassAliases */;
                    // We need to enable substitutions for identifiers. This allows us to
                    // substitute class names inside of a class declaration.
                    context.enableSubstitution(79 /* Identifier */);
                    // Keep track of class aliases.
                    classAliases = [];
                }
            }
            function enableSubstitutionForClassStaticThisOrSuperReference() {
                if ((enabledSubstitutions & 2 /* ClassStaticThisOrSuperReference */) === 0) {
                    enabledSubstitutions |= 2 /* ClassStaticThisOrSuperReference */;
                    // substitute `this` in a static field initializer
                    context.enableSubstitution(108 /* ThisKeyword */);
                    // these push a new lexical environment that is not the class lexical environment
                    context.enableEmitNotification(254 /* FunctionDeclaration */);
                    context.enableEmitNotification(211 /* FunctionExpression */);
                    context.enableEmitNotification(169 /* Constructor */);
                    // these push a new lexical environment that is not the class lexical environment, except
                    // when they have a computed property name
                    context.enableEmitNotification(170 /* GetAccessor */);
                    context.enableEmitNotification(171 /* SetAccessor */);
                    context.enableEmitNotification(167 /* MethodDeclaration */);
                    context.enableEmitNotification(165 /* PropertyDeclaration */);
                    // class lexical environments are restored when entering a computed property name
                    context.enableEmitNotification(160 /* ComputedPropertyName */);
                }
            }
            /**
             * Generates brand-check initializer for private methods.
             *
             * @param statements Statement list that should be used to append new statements.
             * @param methods An array of method declarations.
             * @param receiver The receiver on which each method should be assigned.
             */
            function addMethodStatements(statements, methods, receiver) {
                if (!shouldTransformPrivateElementsOrClassStaticBlocks || !ts.some(methods)) {
                    return;
                }
                var weakSetName = getPrivateIdentifierEnvironment().weakSetName;
                ts.Debug.assert(weakSetName, "weakSetName should be set in private identifier environment");
                statements.push(factory.createExpressionStatement(createPrivateInstanceMethodInitializer(receiver, weakSetName)));
            }
            function visitInvalidSuperProperty(node) {
                return ts.isPropertyAccessExpression(node) ?
                    factory.updatePropertyAccessExpression(node, factory.createVoidZero(), node.name) :
                    factory.updateElementAccessExpression(node, factory.createVoidZero(), ts.visitNode(node.argumentExpression, visitor, ts.isExpression));
            }
            function onEmitNode(hint, node, emitCallback) {
                var original = ts.getOriginalNode(node);
                if (original.id) {
                    var classLexicalEnvironment = classLexicalEnvironmentMap.get(original.id);
                    if (classLexicalEnvironment) {
                        var savedClassLexicalEnvironment = currentClassLexicalEnvironment;
                        var savedCurrentComputedPropertyNameClassLexicalEnvironment = currentComputedPropertyNameClassLexicalEnvironment;
                        currentClassLexicalEnvironment = classLexicalEnvironment;
                        currentComputedPropertyNameClassLexicalEnvironment = classLexicalEnvironment;
                        previousOnEmitNode(hint, node, emitCallback);
                        currentClassLexicalEnvironment = savedClassLexicalEnvironment;
                        currentComputedPropertyNameClassLexicalEnvironment = savedCurrentComputedPropertyNameClassLexicalEnvironment;
                        return;
                    }
                }
                switch (node.kind) {
                    case 211 /* FunctionExpression */:
                        if (ts.isArrowFunction(original) || ts.getEmitFlags(node) & 262144 /* AsyncFunctionBody */) {
                            break;
                        }
                    // falls through
                    case 254 /* FunctionDeclaration */:
                    case 169 /* Constructor */: {
                        var savedClassLexicalEnvironment = currentClassLexicalEnvironment;
                        var savedCurrentComputedPropertyNameClassLexicalEnvironment = currentComputedPropertyNameClassLexicalEnvironment;
                        currentClassLexicalEnvironment = undefined;
                        currentComputedPropertyNameClassLexicalEnvironment = undefined;
                        previousOnEmitNode(hint, node, emitCallback);
                        currentClassLexicalEnvironment = savedClassLexicalEnvironment;
                        currentComputedPropertyNameClassLexicalEnvironment = savedCurrentComputedPropertyNameClassLexicalEnvironment;
                        return;
                    }
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 167 /* MethodDeclaration */:
                    case 165 /* PropertyDeclaration */: {
                        var savedClassLexicalEnvironment = currentClassLexicalEnvironment;
                        var savedCurrentComputedPropertyNameClassLexicalEnvironment = currentComputedPropertyNameClassLexicalEnvironment;
                        currentComputedPropertyNameClassLexicalEnvironment = currentClassLexicalEnvironment;
                        currentClassLexicalEnvironment = undefined;
                        previousOnEmitNode(hint, node, emitCallback);
                        currentClassLexicalEnvironment = savedClassLexicalEnvironment;
                        currentComputedPropertyNameClassLexicalEnvironment = savedCurrentComputedPropertyNameClassLexicalEnvironment;
                        return;
                    }
                    case 160 /* ComputedPropertyName */: {
                        var savedClassLexicalEnvironment = currentClassLexicalEnvironment;
                        var savedCurrentComputedPropertyNameClassLexicalEnvironment = currentComputedPropertyNameClassLexicalEnvironment;
                        currentClassLexicalEnvironment = currentComputedPropertyNameClassLexicalEnvironment;
                        currentComputedPropertyNameClassLexicalEnvironment = undefined;
                        previousOnEmitNode(hint, node, emitCallback);
                        currentClassLexicalEnvironment = savedClassLexicalEnvironment;
                        currentComputedPropertyNameClassLexicalEnvironment = savedCurrentComputedPropertyNameClassLexicalEnvironment;
                        return;
                    }
                }
                previousOnEmitNode(hint, node, emitCallback);
            }
            /**
             * Hooks node substitutions.
             *
             * @param hint The context for the emitter.
             * @param node The node to substitute.
             */
            function onSubstituteNode(hint, node) {
                node = previousOnSubstituteNode(hint, node);
                if (hint === 1 /* Expression */) {
                    return substituteExpression(node);
                }
                return node;
            }
            function substituteExpression(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        return substituteExpressionIdentifier(node);
                    case 108 /* ThisKeyword */:
                        return substituteThisExpression(node);
                }
                return node;
            }
            function substituteThisExpression(node) {
                if (enabledSubstitutions & 2 /* ClassStaticThisOrSuperReference */ && currentClassLexicalEnvironment) {
                    var facts = currentClassLexicalEnvironment.facts, classConstructor = currentClassLexicalEnvironment.classConstructor;
                    if (facts & 1 /* ClassWasDecorated */) {
                        return factory.createParenthesizedExpression(factory.createVoidZero());
                    }
                    if (classConstructor) {
                        return ts.setTextRange(ts.setOriginalNode(factory.cloneNode(classConstructor), node), node);
                    }
                }
                return node;
            }
            function substituteExpressionIdentifier(node) {
                return trySubstituteClassAlias(node) || node;
            }
            function trySubstituteClassAlias(node) {
                if (enabledSubstitutions & 1 /* ClassAliases */) {
                    if (resolver.getNodeCheckFlags(node) & 33554432 /* ConstructorReferenceInClass */) {
                        // Due to the emit for class decorators, any reference to the class from inside of the class body
                        // must instead be rewritten to point to a temporary variable to avoid issues with the double-bind
                        // behavior of class names in ES6.
                        // Also, when emitting statics for class expressions, we must substitute a class alias for
                        // constructor references in static property initializers.
                        var declaration = resolver.getReferencedValueDeclaration(node);
                        if (declaration) {
                            var classAlias = classAliases[declaration.id]; // TODO: GH#18217
                            if (classAlias) {
                                var clone_3 = factory.cloneNode(classAlias);
                                ts.setSourceMapRange(clone_3, node);
                                ts.setCommentRange(clone_3, node);
                                return clone_3;
                            }
                        }
                    }
                }
                return undefined;
            }
            /**
             * If the name is a computed property, this function transforms it, then either returns an expression which caches the
             * value of the result or the expression itself if the value is either unused or safe to inline into multiple locations
             * @param shouldHoist Does the expression need to be reused? (ie, for an initializer or a decorator)
             */
            function getPropertyNameExpressionIfNeeded(name, shouldHoist) {
                if (ts.isComputedPropertyName(name)) {
                    var expression = ts.visitNode(name.expression, visitor, ts.isExpression);
                    var innerExpression = ts.skipPartiallyEmittedExpressions(expression);
                    var inlinable = ts.isSimpleInlineableExpression(innerExpression);
                    var alreadyTransformed = ts.isAssignmentExpression(innerExpression) && ts.isGeneratedIdentifier(innerExpression.left);
                    if (!alreadyTransformed && !inlinable && shouldHoist) {
                        var generatedName = factory.getGeneratedNameForNode(name);
                        if (resolver.getNodeCheckFlags(name) & 524288 /* BlockScopedBindingInLoop */) {
                            addBlockScopedVariable(generatedName);
                        }
                        else {
                            hoistVariableDeclaration(generatedName);
                        }
                        return factory.createAssignment(generatedName, expression);
                    }
                    return (inlinable || ts.isIdentifier(innerExpression)) ? undefined : expression;
                }
            }
            function startClassLexicalEnvironment() {
                classLexicalEnvironmentStack.push(currentClassLexicalEnvironment);
                currentClassLexicalEnvironment = undefined;
            }
            function endClassLexicalEnvironment() {
                currentClassLexicalEnvironment = classLexicalEnvironmentStack.pop();
            }
            function getClassLexicalEnvironment() {
                return currentClassLexicalEnvironment || (currentClassLexicalEnvironment = {
                    facts: 0 /* None */,
                    classConstructor: undefined,
                    superClassReference: undefined,
                    privateIdentifierEnvironment: undefined,
                });
            }
            function getPrivateIdentifierEnvironment() {
                var lex = getClassLexicalEnvironment();
                lex.privateIdentifierEnvironment || (lex.privateIdentifierEnvironment = {
                    className: "",
                    identifiers: new ts.Map()
                });
                return lex.privateIdentifierEnvironment;
            }
            function getPendingExpressions() {
                return pendingExpressions || (pendingExpressions = []);
            }
            function addPrivateIdentifierToEnvironment(node) {
                var _a;
                var text = ts.getTextOfPropertyName(node.name);
                var lex = getClassLexicalEnvironment();
                var classConstructor = lex.classConstructor;
                var privateEnv = getPrivateIdentifierEnvironment();
                var weakSetName = privateEnv.weakSetName;
                var assignmentExpressions = [];
                var privateName = node.name.escapedText;
                var previousInfo = privateEnv.identifiers.get(privateName);
                var isValid = !isReservedPrivateName(node.name) && previousInfo === undefined;
                if (ts.hasStaticModifier(node)) {
                    ts.Debug.assert(classConstructor, "weakSetName should be set in private identifier environment");
                    if (ts.isPropertyDeclaration(node)) {
                        var variableName = createHoistedVariableForPrivateName(text, node);
                        privateEnv.identifiers.set(privateName, {
                            kind: "f" /* Field */,
                            variableName: variableName,
                            brandCheckIdentifier: classConstructor,
                            isStatic: true,
                            isValid: isValid,
                        });
                    }
                    else if (ts.isMethodDeclaration(node)) {
                        var functionName = createHoistedVariableForPrivateName(text, node);
                        privateEnv.identifiers.set(privateName, {
                            kind: "m" /* Method */,
                            methodName: functionName,
                            brandCheckIdentifier: classConstructor,
                            isStatic: true,
                            isValid: isValid,
                        });
                    }
                    else if (ts.isGetAccessorDeclaration(node)) {
                        var getterName = createHoistedVariableForPrivateName(text + "_get", node);
                        if ((previousInfo === null || previousInfo === void 0 ? void 0 : previousInfo.kind) === "a" /* Accessor */ && previousInfo.isStatic && !previousInfo.getterName) {
                            previousInfo.getterName = getterName;
                        }
                        else {
                            privateEnv.identifiers.set(privateName, {
                                kind: "a" /* Accessor */,
                                getterName: getterName,
                                setterName: undefined,
                                brandCheckIdentifier: classConstructor,
                                isStatic: true,
                                isValid: isValid,
                            });
                        }
                    }
                    else if (ts.isSetAccessorDeclaration(node)) {
                        var setterName = createHoistedVariableForPrivateName(text + "_set", node);
                        if ((previousInfo === null || previousInfo === void 0 ? void 0 : previousInfo.kind) === "a" /* Accessor */ && previousInfo.isStatic && !previousInfo.setterName) {
                            previousInfo.setterName = setterName;
                        }
                        else {
                            privateEnv.identifiers.set(privateName, {
                                kind: "a" /* Accessor */,
                                getterName: undefined,
                                setterName: setterName,
                                brandCheckIdentifier: classConstructor,
                                isStatic: true,
                                isValid: isValid,
                            });
                        }
                    }
                    else {
                        ts.Debug.assertNever(node, "Unknown class element type.");
                    }
                }
                else if (ts.isPropertyDeclaration(node)) {
                    var weakMapName = createHoistedVariableForPrivateName(text, node);
                    privateEnv.identifiers.set(privateName, {
                        kind: "f" /* Field */,
                        brandCheckIdentifier: weakMapName,
                        isStatic: false,
                        variableName: undefined,
                        isValid: isValid,
                    });
                    assignmentExpressions.push(factory.createAssignment(weakMapName, factory.createNewExpression(factory.createIdentifier("WeakMap"), 
                    /*typeArguments*/ undefined, [])));
                }
                else if (ts.isMethodDeclaration(node)) {
                    ts.Debug.assert(weakSetName, "weakSetName should be set in private identifier environment");
                    privateEnv.identifiers.set(privateName, {
                        kind: "m" /* Method */,
                        methodName: createHoistedVariableForPrivateName(text, node),
                        brandCheckIdentifier: weakSetName,
                        isStatic: false,
                        isValid: isValid,
                    });
                }
                else if (ts.isAccessor(node)) {
                    ts.Debug.assert(weakSetName, "weakSetName should be set in private identifier environment");
                    if (ts.isGetAccessor(node)) {
                        var getterName = createHoistedVariableForPrivateName(text + "_get", node);
                        if ((previousInfo === null || previousInfo === void 0 ? void 0 : previousInfo.kind) === "a" /* Accessor */ && !previousInfo.isStatic && !previousInfo.getterName) {
                            previousInfo.getterName = getterName;
                        }
                        else {
                            privateEnv.identifiers.set(privateName, {
                                kind: "a" /* Accessor */,
                                getterName: getterName,
                                setterName: undefined,
                                brandCheckIdentifier: weakSetName,
                                isStatic: false,
                                isValid: isValid,
                            });
                        }
                    }
                    else {
                        var setterName = createHoistedVariableForPrivateName(text + "_set", node);
                        if ((previousInfo === null || previousInfo === void 0 ? void 0 : previousInfo.kind) === "a" /* Accessor */ && !previousInfo.isStatic && !previousInfo.setterName) {
                            previousInfo.setterName = setterName;
                        }
                        else {
                            privateEnv.identifiers.set(privateName, {
                                kind: "a" /* Accessor */,
                                getterName: undefined,
                                setterName: setterName,
                                brandCheckIdentifier: weakSetName,
                                isStatic: false,
                                isValid: isValid,
                            });
                        }
                    }
                }
                else {
                    ts.Debug.assertNever(node, "Unknown class element type.");
                }
                (_a = getPendingExpressions()).push.apply(_a, assignmentExpressions);
            }
            function createHoistedVariableForClass(name, node) {
                var className = getPrivateIdentifierEnvironment().className;
                var prefix = className ? "_" + className : "";
                var identifier = factory.createUniqueName(prefix + "_" + name, 16 /* Optimistic */);
                if (resolver.getNodeCheckFlags(node) & 524288 /* BlockScopedBindingInLoop */) {
                    addBlockScopedVariable(identifier);
                }
                else {
                    hoistVariableDeclaration(identifier);
                }
                return identifier;
            }
            function createHoistedVariableForPrivateName(privateName, node) {
                return createHoistedVariableForClass(privateName.substring(1), node.name);
            }
            function accessPrivateIdentifier(name) {
                var _a;
                if (currentClassLexicalEnvironment === null || currentClassLexicalEnvironment === void 0 ? void 0 : currentClassLexicalEnvironment.privateIdentifierEnvironment) {
                    var info = currentClassLexicalEnvironment.privateIdentifierEnvironment.identifiers.get(name.escapedText);
                    if (info) {
                        return info;
                    }
                }
                for (var i = classLexicalEnvironmentStack.length - 1; i >= 0; --i) {
                    var env = classLexicalEnvironmentStack[i];
                    if (!env) {
                        continue;
                    }
                    var info = (_a = env.privateIdentifierEnvironment) === null || _a === void 0 ? void 0 : _a.identifiers.get(name.escapedText);
                    if (info) {
                        return info;
                    }
                }
                return undefined;
            }
            function wrapPrivateIdentifierForDestructuringTarget(node) {
                var parameter = factory.getGeneratedNameForNode(node);
                var info = accessPrivateIdentifier(node.name);
                if (!info) {
                    return ts.visitEachChild(node, visitor, context);
                }
                var receiver = node.expression;
                // We cannot copy `this` or `super` into the function because they will be bound
                // differently inside the function.
                if (ts.isThisProperty(node) || ts.isSuperProperty(node) || !ts.isSimpleCopiableExpression(node.expression)) {
                    receiver = factory.createTempVariable(hoistVariableDeclaration, /*reservedInNestedScopes*/ true);
                    getPendingExpressions().push(factory.createBinaryExpression(receiver, 63 /* EqualsToken */, ts.visitNode(node.expression, visitor, ts.isExpression)));
                }
                return factory.createAssignmentTargetWrapper(parameter, createPrivateIdentifierAssignment(info, receiver, parameter, 63 /* EqualsToken */));
            }
            function visitArrayAssignmentTarget(node) {
                var target = ts.getTargetOfBindingOrAssignmentElement(node);
                if (target) {
                    var wrapped = void 0;
                    if (ts.isPrivateIdentifierPropertyAccessExpression(target)) {
                        wrapped = wrapPrivateIdentifierForDestructuringTarget(target);
                    }
                    else if (shouldTransformSuperInStaticInitializers &&
                        ts.isSuperProperty(target) &&
                        currentStaticPropertyDeclarationOrStaticBlock &&
                        currentClassLexicalEnvironment) {
                        var classConstructor = currentClassLexicalEnvironment.classConstructor, superClassReference = currentClassLexicalEnvironment.superClassReference, facts = currentClassLexicalEnvironment.facts;
                        if (facts & 1 /* ClassWasDecorated */) {
                            wrapped = visitInvalidSuperProperty(target);
                        }
                        else if (classConstructor && superClassReference) {
                            var name = ts.isElementAccessExpression(target) ? ts.visitNode(target.argumentExpression, visitor, ts.isExpression) :
                                ts.isIdentifier(target.name) ? factory.createStringLiteralFromNode(target.name) :
                                    undefined;
                            if (name) {
                                var temp = factory.createTempVariable(/*recordTempVariable*/ undefined);
                                wrapped = factory.createAssignmentTargetWrapper(temp, factory.createReflectSetCall(superClassReference, name, temp, classConstructor));
                            }
                        }
                    }
                    if (wrapped) {
                        if (ts.isAssignmentExpression(node)) {
                            return factory.updateBinaryExpression(node, wrapped, node.operatorToken, ts.visitNode(node.right, visitor, ts.isExpression));
                        }
                        else if (ts.isSpreadElement(node)) {
                            return factory.updateSpreadElement(node, wrapped);
                        }
                        else {
                            return wrapped;
                        }
                    }
                }
                return ts.visitNode(node, visitorDestructuringTarget);
            }
            function visitObjectAssignmentTarget(node) {
                if (ts.isObjectBindingOrAssignmentElement(node) && !ts.isShorthandPropertyAssignment(node)) {
                    var target = ts.getTargetOfBindingOrAssignmentElement(node);
                    var wrapped = void 0;
                    if (target) {
                        if (ts.isPrivateIdentifierPropertyAccessExpression(target)) {
                            wrapped = wrapPrivateIdentifierForDestructuringTarget(target);
                        }
                        else if (shouldTransformSuperInStaticInitializers &&
                            ts.isSuperProperty(target) &&
                            currentStaticPropertyDeclarationOrStaticBlock &&
                            currentClassLexicalEnvironment) {
                            var classConstructor = currentClassLexicalEnvironment.classConstructor, superClassReference = currentClassLexicalEnvironment.superClassReference, facts = currentClassLexicalEnvironment.facts;
                            if (facts & 1 /* ClassWasDecorated */) {
                                wrapped = visitInvalidSuperProperty(target);
                            }
                            else if (classConstructor && superClassReference) {
                                var name = ts.isElementAccessExpression(target) ? ts.visitNode(target.argumentExpression, visitor, ts.isExpression) :
                                    ts.isIdentifier(target.name) ? factory.createStringLiteralFromNode(target.name) :
                                        undefined;
                                if (name) {
                                    var temp = factory.createTempVariable(/*recordTempVariable*/ undefined);
                                    wrapped = factory.createAssignmentTargetWrapper(temp, factory.createReflectSetCall(superClassReference, name, temp, classConstructor));
                                }
                            }
                        }
                    }
                    if (ts.isPropertyAssignment(node)) {
                        var initializer = ts.getInitializerOfBindingOrAssignmentElement(node);
                        return factory.updatePropertyAssignment(node, ts.visitNode(node.name, visitor, ts.isPropertyName), wrapped ?
                            initializer ? factory.createAssignment(wrapped, ts.visitNode(initializer, visitor)) : wrapped :
                            ts.visitNode(node.initializer, visitorDestructuringTarget, ts.isExpression));
                    }
                    if (ts.isSpreadAssignment(node)) {
                        return factory.updateSpreadAssignment(node, wrapped || ts.visitNode(node.expression, visitorDestructuringTarget, ts.isExpression));
                    }
                    ts.Debug.assert(wrapped === undefined, "Should not have generated a wrapped target");
                }
                return ts.visitNode(node, visitor);
            }
            function visitAssignmentPattern(node) {
                if (ts.isArrayLiteralExpression(node)) {
                    // Transforms private names in destructuring assignment array bindings.
                    // Transforms SuperProperty assignments in destructuring assignment array bindings in static initializers.
                    //
                    // Source:
                    // ([ this.#myProp ] = [ "hello" ]);
                    //
                    // Transformation:
                    // [ { set value(x) { this.#myProp = x; } }.value ] = [ "hello" ];
                    return factory.updateArrayLiteralExpression(node, ts.visitNodes(node.elements, visitArrayAssignmentTarget, ts.isExpression));
                }
                else {
                    // Transforms private names in destructuring assignment object bindings.
                    // Transforms SuperProperty assignments in destructuring assignment object bindings in static initializers.
                    //
                    // Source:
                    // ({ stringProperty: this.#myProp } = { stringProperty: "hello" });
                    //
                    // Transformation:
                    // ({ stringProperty: { set value(x) { this.#myProp = x; } }.value }) = { stringProperty: "hello" };
                    return factory.updateObjectLiteralExpression(node, ts.visitNodes(node.properties, visitObjectAssignmentTarget, ts.isObjectLiteralElementLike));
                }
            }
        }
        ts.transformClassFields = transformClassFields;
        function createPrivateStaticFieldInitializer(variableName, initializer) {
            return ts.factory.createAssignment(variableName, ts.factory.createObjectLiteralExpression([
                ts.factory.createPropertyAssignment("value", initializer || ts.factory.createVoidZero())
            ]));
        }
        function createPrivateInstanceFieldInitializer(receiver, initializer, weakMapName) {
            return ts.factory.createCallExpression(ts.factory.createPropertyAccessExpression(weakMapName, "set"), 
            /*typeArguments*/ undefined, [receiver, initializer || ts.factory.createVoidZero()]);
        }
        function createPrivateInstanceMethodInitializer(receiver, weakSetName) {
            return ts.factory.createCallExpression(ts.factory.createPropertyAccessExpression(weakSetName, "add"), 
            /*typeArguments*/ undefined, [receiver]);
        }
        function isReservedPrivateName(node) {
            return node.escapedText === "#constructor";
        }
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var ES2017SubstitutionFlags;
        (function (ES2017SubstitutionFlags) {
            /** Enables substitutions for async methods with `super` calls. */
            ES2017SubstitutionFlags[ES2017SubstitutionFlags["AsyncMethodsWithSuper"] = 1] = "AsyncMethodsWithSuper";
        })(ES2017SubstitutionFlags || (ES2017SubstitutionFlags = {}));
        var ContextFlags;
        (function (ContextFlags) {
            ContextFlags[ContextFlags["NonTopLevel"] = 1] = "NonTopLevel";
            ContextFlags[ContextFlags["HasLexicalThis"] = 2] = "HasLexicalThis";
        })(ContextFlags || (ContextFlags = {}));
        function transformES2017(context) {
            var factory = context.factory, emitHelpers = context.getEmitHelperFactory, resumeLexicalEnvironment = context.resumeLexicalEnvironment, endLexicalEnvironment = context.endLexicalEnvironment, hoistVariableDeclaration = context.hoistVariableDeclaration;
            var resolver = context.getEmitResolver();
            var compilerOptions = context.getCompilerOptions();
            var languageVersion = ts.getEmitScriptTarget(compilerOptions);
            /**
             * Keeps track of whether expression substitution has been enabled for specific edge cases.
             * They are persisted between each SourceFile transformation and should not be reset.
             */
            var enabledSubstitutions;
            /**
             * This keeps track of containers where `super` is valid, for use with
             * just-in-time substitution for `super` expressions inside of async methods.
             */
            var enclosingSuperContainerFlags = 0;
            var enclosingFunctionParameterNames;
            /**
             * Keeps track of property names accessed on super (`super.x`) within async functions.
             */
            var capturedSuperProperties;
            /** Whether the async function contains an element access on super (`super[x]`). */
            var hasSuperElementAccess;
            /** A set of node IDs for generated super accessors (variable statements). */
            var substitutedSuperAccessors = [];
            var contextFlags = 0;
            // Save the previous transformation hooks.
            var previousOnEmitNode = context.onEmitNode;
            var previousOnSubstituteNode = context.onSubstituteNode;
            // Set new transformation hooks.
            context.onEmitNode = onEmitNode;
            context.onSubstituteNode = onSubstituteNode;
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                setContextFlag(1 /* NonTopLevel */, false);
                setContextFlag(2 /* HasLexicalThis */, !ts.isEffectiveStrictModeSourceFile(node, compilerOptions));
                var visited = ts.visitEachChild(node, visitor, context);
                ts.addEmitHelpers(visited, context.readEmitHelpers());
                return visited;
            }
            function setContextFlag(flag, val) {
                contextFlags = val ? contextFlags | flag : contextFlags & ~flag;
            }
            function inContext(flags) {
                return (contextFlags & flags) !== 0;
            }
            function inTopLevelContext() {
                return !inContext(1 /* NonTopLevel */);
            }
            function inHasLexicalThisContext() {
                return inContext(2 /* HasLexicalThis */);
            }
            function doWithContext(flags, cb, value) {
                var contextFlagsToSet = flags & ~contextFlags;
                if (contextFlagsToSet) {
                    setContextFlag(contextFlagsToSet, /*val*/ true);
                    var result = cb(value);
                    setContextFlag(contextFlagsToSet, /*val*/ false);
                    return result;
                }
                return cb(value);
            }
            function visitDefault(node) {
                return ts.visitEachChild(node, visitor, context);
            }
            function visitor(node) {
                if ((node.transformFlags & 128 /* ContainsES2017 */) === 0) {
                    return node;
                }
                switch (node.kind) {
                    case 130 /* AsyncKeyword */:
                        // ES2017 async modifier should be elided for targets < ES2017
                        return undefined;
                    case 216 /* AwaitExpression */:
                        return visitAwaitExpression(node);
                    case 167 /* MethodDeclaration */:
                        return doWithContext(1 /* NonTopLevel */ | 2 /* HasLexicalThis */, visitMethodDeclaration, node);
                    case 254 /* FunctionDeclaration */:
                        return doWithContext(1 /* NonTopLevel */ | 2 /* HasLexicalThis */, visitFunctionDeclaration, node);
                    case 211 /* FunctionExpression */:
                        return doWithContext(1 /* NonTopLevel */ | 2 /* HasLexicalThis */, visitFunctionExpression, node);
                    case 212 /* ArrowFunction */:
                        return doWithContext(1 /* NonTopLevel */, visitArrowFunction, node);
                    case 204 /* PropertyAccessExpression */:
                        if (capturedSuperProperties && ts.isPropertyAccessExpression(node) && node.expression.kind === 106 /* SuperKeyword */) {
                            capturedSuperProperties.add(node.name.escapedText);
                        }
                        return ts.visitEachChild(node, visitor, context);
                    case 205 /* ElementAccessExpression */:
                        if (capturedSuperProperties && node.expression.kind === 106 /* SuperKeyword */) {
                            hasSuperElementAccess = true;
                        }
                        return ts.visitEachChild(node, visitor, context);
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 169 /* Constructor */:
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                        return doWithContext(1 /* NonTopLevel */ | 2 /* HasLexicalThis */, visitDefault, node);
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function asyncBodyVisitor(node) {
                if (ts.isNodeWithPossibleHoistedDeclaration(node)) {
                    switch (node.kind) {
                        case 235 /* VariableStatement */:
                            return visitVariableStatementInAsyncBody(node);
                        case 240 /* ForStatement */:
                            return visitForStatementInAsyncBody(node);
                        case 241 /* ForInStatement */:
                            return visitForInStatementInAsyncBody(node);
                        case 242 /* ForOfStatement */:
                            return visitForOfStatementInAsyncBody(node);
                        case 290 /* CatchClause */:
                            return visitCatchClauseInAsyncBody(node);
                        case 233 /* Block */:
                        case 247 /* SwitchStatement */:
                        case 261 /* CaseBlock */:
                        case 287 /* CaseClause */:
                        case 288 /* DefaultClause */:
                        case 250 /* TryStatement */:
                        case 238 /* DoStatement */:
                        case 239 /* WhileStatement */:
                        case 237 /* IfStatement */:
                        case 246 /* WithStatement */:
                        case 248 /* LabeledStatement */:
                            return ts.visitEachChild(node, asyncBodyVisitor, context);
                        default:
                            return ts.Debug.assertNever(node, "Unhandled node.");
                    }
                }
                return visitor(node);
            }
            function visitCatchClauseInAsyncBody(node) {
                var catchClauseNames = new ts.Set();
                recordDeclarationName(node.variableDeclaration, catchClauseNames); // TODO: GH#18217
                // names declared in a catch variable are block scoped
                var catchClauseUnshadowedNames;
                catchClauseNames.forEach(function (_, escapedName) {
                    if (enclosingFunctionParameterNames.has(escapedName)) {
                        if (!catchClauseUnshadowedNames) {
                            catchClauseUnshadowedNames = new ts.Set(enclosingFunctionParameterNames);
                        }
                        catchClauseUnshadowedNames.delete(escapedName);
                    }
                });
                if (catchClauseUnshadowedNames) {
                    var savedEnclosingFunctionParameterNames = enclosingFunctionParameterNames;
                    enclosingFunctionParameterNames = catchClauseUnshadowedNames;
                    var result = ts.visitEachChild(node, asyncBodyVisitor, context);
                    enclosingFunctionParameterNames = savedEnclosingFunctionParameterNames;
                    return result;
                }
                else {
                    return ts.visitEachChild(node, asyncBodyVisitor, context);
                }
            }
            function visitVariableStatementInAsyncBody(node) {
                if (isVariableDeclarationListWithCollidingName(node.declarationList)) {
                    var expression = visitVariableDeclarationListWithCollidingNames(node.declarationList, /*hasReceiver*/ false);
                    return expression ? factory.createExpressionStatement(expression) : undefined;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitForInStatementInAsyncBody(node) {
                return factory.updateForInStatement(node, isVariableDeclarationListWithCollidingName(node.initializer)
                    ? visitVariableDeclarationListWithCollidingNames(node.initializer, /*hasReceiver*/ true)
                    : ts.visitNode(node.initializer, visitor, ts.isForInitializer), ts.visitNode(node.expression, visitor, ts.isExpression), ts.visitIterationBody(node.statement, asyncBodyVisitor, context));
            }
            function visitForOfStatementInAsyncBody(node) {
                return factory.updateForOfStatement(node, ts.visitNode(node.awaitModifier, visitor, ts.isToken), isVariableDeclarationListWithCollidingName(node.initializer)
                    ? visitVariableDeclarationListWithCollidingNames(node.initializer, /*hasReceiver*/ true)
                    : ts.visitNode(node.initializer, visitor, ts.isForInitializer), ts.visitNode(node.expression, visitor, ts.isExpression), ts.visitIterationBody(node.statement, asyncBodyVisitor, context));
            }
            function visitForStatementInAsyncBody(node) {
                var initializer = node.initializer; // TODO: GH#18217
                return factory.updateForStatement(node, isVariableDeclarationListWithCollidingName(initializer)
                    ? visitVariableDeclarationListWithCollidingNames(initializer, /*hasReceiver*/ false)
                    : ts.visitNode(node.initializer, visitor, ts.isForInitializer), ts.visitNode(node.condition, visitor, ts.isExpression), ts.visitNode(node.incrementor, visitor, ts.isExpression), ts.visitIterationBody(node.statement, asyncBodyVisitor, context));
            }
            /**
             * Visits an AwaitExpression node.
             *
             * This function will be called any time a ES2017 await expression is encountered.
             *
             * @param node The node to visit.
             */
            function visitAwaitExpression(node) {
                // do not downlevel a top-level await as it is module syntax...
                if (inTopLevelContext()) {
                    return ts.visitEachChild(node, visitor, context);
                }
                return ts.setOriginalNode(ts.setTextRange(factory.createYieldExpression(
                /*asteriskToken*/ undefined, ts.visitNode(node.expression, visitor, ts.isExpression)), node), node);
            }
            /**
             * Visits a MethodDeclaration node.
             *
             * This function will be called when one of the following conditions are met:
             * - The node is marked as async
             *
             * @param node The node to visit.
             */
            function visitMethodDeclaration(node) {
                return factory.updateMethodDeclaration(node, 
                /*decorators*/ undefined, ts.visitNodes(node.modifiers, visitor, ts.isModifier), node.asteriskToken, node.name, 
                /*questionToken*/ undefined, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, ts.getFunctionFlags(node) & 2 /* Async */
                    ? transformAsyncFunctionBody(node)
                    : ts.visitFunctionBody(node.body, visitor, context));
            }
            /**
             * Visits a FunctionDeclaration node.
             *
             * This function will be called when one of the following conditions are met:
             * - The node is marked async
             *
             * @param node The node to visit.
             */
            function visitFunctionDeclaration(node) {
                return factory.updateFunctionDeclaration(node, 
                /*decorators*/ undefined, ts.visitNodes(node.modifiers, visitor, ts.isModifier), node.asteriskToken, node.name, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, ts.getFunctionFlags(node) & 2 /* Async */
                    ? transformAsyncFunctionBody(node)
                    : ts.visitFunctionBody(node.body, visitor, context));
            }
            /**
             * Visits a FunctionExpression node.
             *
             * This function will be called when one of the following conditions are met:
             * - The node is marked async
             *
             * @param node The node to visit.
             */
            function visitFunctionExpression(node) {
                return factory.updateFunctionExpression(node, ts.visitNodes(node.modifiers, visitor, ts.isModifier), node.asteriskToken, node.name, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, ts.getFunctionFlags(node) & 2 /* Async */
                    ? transformAsyncFunctionBody(node)
                    : ts.visitFunctionBody(node.body, visitor, context));
            }
            /**
             * Visits an ArrowFunction.
             *
             * This function will be called when one of the following conditions are met:
             * - The node is marked async
             *
             * @param node The node to visit.
             */
            function visitArrowFunction(node) {
                return factory.updateArrowFunction(node, ts.visitNodes(node.modifiers, visitor, ts.isModifier), 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, node.equalsGreaterThanToken, ts.getFunctionFlags(node) & 2 /* Async */
                    ? transformAsyncFunctionBody(node)
                    : ts.visitFunctionBody(node.body, visitor, context));
            }
            function recordDeclarationName(_a, names) {
                var name = _a.name;
                if (ts.isIdentifier(name)) {
                    names.add(name.escapedText);
                }
                else {
                    for (var _i = 0, _b = name.elements; _i < _b.length; _i++) {
                        var element = _b[_i];
                        if (!ts.isOmittedExpression(element)) {
                            recordDeclarationName(element, names);
                        }
                    }
                }
            }
            function isVariableDeclarationListWithCollidingName(node) {
                return !!node
                    && ts.isVariableDeclarationList(node)
                    && !(node.flags & 3 /* BlockScoped */)
                    && node.declarations.some(collidesWithParameterName);
            }
            function visitVariableDeclarationListWithCollidingNames(node, hasReceiver) {
                hoistVariableDeclarationList(node);
                var variables = ts.getInitializedVariables(node);
                if (variables.length === 0) {
                    if (hasReceiver) {
                        return ts.visitNode(factory.converters.convertToAssignmentElementTarget(node.declarations[0].name), visitor, ts.isExpression);
                    }
                    return undefined;
                }
                return factory.inlineExpressions(ts.map(variables, transformInitializedVariable));
            }
            function hoistVariableDeclarationList(node) {
                ts.forEach(node.declarations, hoistVariable);
            }
            function hoistVariable(_a) {
                var name = _a.name;
                if (ts.isIdentifier(name)) {
                    hoistVariableDeclaration(name);
                }
                else {
                    for (var _i = 0, _b = name.elements; _i < _b.length; _i++) {
                        var element = _b[_i];
                        if (!ts.isOmittedExpression(element)) {
                            hoistVariable(element);
                        }
                    }
                }
            }
            function transformInitializedVariable(node) {
                var converted = ts.setSourceMapRange(factory.createAssignment(factory.converters.convertToAssignmentElementTarget(node.name), node.initializer), node);
                return ts.visitNode(converted, visitor, ts.isExpression);
            }
            function collidesWithParameterName(_a) {
                var name = _a.name;
                if (ts.isIdentifier(name)) {
                    return enclosingFunctionParameterNames.has(name.escapedText);
                }
                else {
                    for (var _i = 0, _b = name.elements; _i < _b.length; _i++) {
                        var element = _b[_i];
                        if (!ts.isOmittedExpression(element) && collidesWithParameterName(element)) {
                            return true;
                        }
                    }
                }
                return false;
            }
            function transformAsyncFunctionBody(node) {
                resumeLexicalEnvironment();
                var original = ts.getOriginalNode(node, ts.isFunctionLike);
                var nodeType = original.type;
                var promiseConstructor = languageVersion < 2 /* ES2015 */ ? getPromiseConstructor(nodeType) : undefined;
                var isArrowFunction = node.kind === 212 /* ArrowFunction */;
                var hasLexicalArguments = (resolver.getNodeCheckFlags(node) & 8192 /* CaptureArguments */) !== 0;
                // An async function is emit as an outer function that calls an inner
                // generator function. To preserve lexical bindings, we pass the current
                // `this` and `arguments` objects to `__awaiter`. The generator function
                // passed to `__awaiter` is executed inside of the callback to the
                // promise constructor.
                var savedEnclosingFunctionParameterNames = enclosingFunctionParameterNames;
                enclosingFunctionParameterNames = new ts.Set();
                for (var _i = 0, _a = node.parameters; _i < _a.length; _i++) {
                    var parameter = _a[_i];
                    recordDeclarationName(parameter, enclosingFunctionParameterNames);
                }
                var savedCapturedSuperProperties = capturedSuperProperties;
                var savedHasSuperElementAccess = hasSuperElementAccess;
                if (!isArrowFunction) {
                    capturedSuperProperties = new ts.Set();
                    hasSuperElementAccess = false;
                }
                var result;
                if (!isArrowFunction) {
                    var statements = [];
                    var statementOffset = factory.copyPrologue(node.body.statements, statements, /*ensureUseStrict*/ false, visitor);
                    statements.push(factory.createReturnStatement(emitHelpers().createAwaiterHelper(inHasLexicalThisContext(), hasLexicalArguments, promiseConstructor, transformAsyncFunctionBodyWorker(node.body, statementOffset))));
                    ts.insertStatementsAfterStandardPrologue(statements, endLexicalEnvironment());
                    // Minor optimization, emit `_super` helper to capture `super` access in an arrow.
                    // This step isn't needed if we eventually transform this to ES5.
                    var emitSuperHelpers = languageVersion >= 2 /* ES2015 */ && resolver.getNodeCheckFlags(node) & (4096 /* AsyncMethodWithSuperBinding */ | 2048 /* AsyncMethodWithSuper */);
                    if (emitSuperHelpers) {
                        enableSubstitutionForAsyncMethodsWithSuper();
                        if (capturedSuperProperties.size) {
                            var variableStatement = createSuperAccessVariableStatement(factory, resolver, node, capturedSuperProperties);
                            substitutedSuperAccessors[ts.getNodeId(variableStatement)] = true;
                            ts.insertStatementsAfterStandardPrologue(statements, [variableStatement]);
                        }
                    }
                    var block = factory.createBlock(statements, /*multiLine*/ true);
                    ts.setTextRange(block, node.body);
                    if (emitSuperHelpers && hasSuperElementAccess) {
                        // Emit helpers for super element access expressions (`super[x]`).
                        if (resolver.getNodeCheckFlags(node) & 4096 /* AsyncMethodWithSuperBinding */) {
                            ts.addEmitHelper(block, ts.advancedAsyncSuperHelper);
                        }
                        else if (resolver.getNodeCheckFlags(node) & 2048 /* AsyncMethodWithSuper */) {
                            ts.addEmitHelper(block, ts.asyncSuperHelper);
                        }
                    }
                    result = block;
                }
                else {
                    var expression = emitHelpers().createAwaiterHelper(inHasLexicalThisContext(), hasLexicalArguments, promiseConstructor, transformAsyncFunctionBodyWorker(node.body));
                    var declarations = endLexicalEnvironment();
                    if (ts.some(declarations)) {
                        var block = factory.converters.convertToFunctionBlock(expression);
                        result = factory.updateBlock(block, ts.setTextRange(factory.createNodeArray(ts.concatenate(declarations, block.statements)), block.statements));
                    }
                    else {
                        result = expression;
                    }
                }
                enclosingFunctionParameterNames = savedEnclosingFunctionParameterNames;
                if (!isArrowFunction) {
                    capturedSuperProperties = savedCapturedSuperProperties;
                    hasSuperElementAccess = savedHasSuperElementAccess;
                }
                return result;
            }
            function transformAsyncFunctionBodyWorker(body, start) {
                if (ts.isBlock(body)) {
                    return factory.updateBlock(body, ts.visitNodes(body.statements, asyncBodyVisitor, ts.isStatement, start));
                }
                else {
                    return factory.converters.convertToFunctionBlock(ts.visitNode(body, asyncBodyVisitor, ts.isConciseBody));
                }
            }
            function getPromiseConstructor(type) {
                var typeName = type && ts.getEntityNameFromTypeNode(type);
                if (typeName && ts.isEntityName(typeName)) {
                    var serializationKind = resolver.getTypeReferenceSerializationKind(typeName);
                    if (serializationKind === ts.TypeReferenceSerializationKind.TypeWithConstructSignatureAndValue
                        || serializationKind === ts.TypeReferenceSerializationKind.Unknown) {
                        return typeName;
                    }
                }
                return undefined;
            }
            function enableSubstitutionForAsyncMethodsWithSuper() {
                if ((enabledSubstitutions & 1 /* AsyncMethodsWithSuper */) === 0) {
                    enabledSubstitutions |= 1 /* AsyncMethodsWithSuper */;
                    // We need to enable substitutions for call, property access, and element access
                    // if we need to rewrite super calls.
                    context.enableSubstitution(206 /* CallExpression */);
                    context.enableSubstitution(204 /* PropertyAccessExpression */);
                    context.enableSubstitution(205 /* ElementAccessExpression */);
                    // We need to be notified when entering and exiting declarations that bind super.
                    context.enableEmitNotification(255 /* ClassDeclaration */);
                    context.enableEmitNotification(167 /* MethodDeclaration */);
                    context.enableEmitNotification(170 /* GetAccessor */);
                    context.enableEmitNotification(171 /* SetAccessor */);
                    context.enableEmitNotification(169 /* Constructor */);
                    // We need to be notified when entering the generated accessor arrow functions.
                    context.enableEmitNotification(235 /* VariableStatement */);
                }
            }
            /**
             * Hook for node emit.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to emit.
             * @param emit A callback used to emit the node in the printer.
             */
            function onEmitNode(hint, node, emitCallback) {
                // If we need to support substitutions for `super` in an async method,
                // we should track it here.
                if (enabledSubstitutions & 1 /* AsyncMethodsWithSuper */ && isSuperContainer(node)) {
                    var superContainerFlags = resolver.getNodeCheckFlags(node) & (2048 /* AsyncMethodWithSuper */ | 4096 /* AsyncMethodWithSuperBinding */);
                    if (superContainerFlags !== enclosingSuperContainerFlags) {
                        var savedEnclosingSuperContainerFlags = enclosingSuperContainerFlags;
                        enclosingSuperContainerFlags = superContainerFlags;
                        previousOnEmitNode(hint, node, emitCallback);
                        enclosingSuperContainerFlags = savedEnclosingSuperContainerFlags;
                        return;
                    }
                }
                // Disable substitution in the generated super accessor itself.
                else if (enabledSubstitutions && substitutedSuperAccessors[ts.getNodeId(node)]) {
                    var savedEnclosingSuperContainerFlags = enclosingSuperContainerFlags;
                    enclosingSuperContainerFlags = 0;
                    previousOnEmitNode(hint, node, emitCallback);
                    enclosingSuperContainerFlags = savedEnclosingSuperContainerFlags;
                    return;
                }
                previousOnEmitNode(hint, node, emitCallback);
            }
            /**
             * Hooks node substitutions.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to substitute.
             */
            function onSubstituteNode(hint, node) {
                node = previousOnSubstituteNode(hint, node);
                if (hint === 1 /* Expression */ && enclosingSuperContainerFlags) {
                    return substituteExpression(node);
                }
                return node;
            }
            function substituteExpression(node) {
                switch (node.kind) {
                    case 204 /* PropertyAccessExpression */:
                        return substitutePropertyAccessExpression(node);
                    case 205 /* ElementAccessExpression */:
                        return substituteElementAccessExpression(node);
                    case 206 /* CallExpression */:
                        return substituteCallExpression(node);
                }
                return node;
            }
            function substitutePropertyAccessExpression(node) {
                if (node.expression.kind === 106 /* SuperKeyword */) {
                    return ts.setTextRange(factory.createPropertyAccessExpression(factory.createUniqueName("_super", 16 /* Optimistic */ | 32 /* FileLevel */), node.name), node);
                }
                return node;
            }
            function substituteElementAccessExpression(node) {
                if (node.expression.kind === 106 /* SuperKeyword */) {
                    return createSuperElementAccessInAsyncMethod(node.argumentExpression, node);
                }
                return node;
            }
            function substituteCallExpression(node) {
                var expression = node.expression;
                if (ts.isSuperProperty(expression)) {
                    var argumentExpression = ts.isPropertyAccessExpression(expression)
                        ? substitutePropertyAccessExpression(expression)
                        : substituteElementAccessExpression(expression);
                    return factory.createCallExpression(factory.createPropertyAccessExpression(argumentExpression, "call"), 
                    /*typeArguments*/ undefined, __spreadArray([
                        factory.createThis()
                    ], node.arguments, true));
                }
                return node;
            }
            function isSuperContainer(node) {
                var kind = node.kind;
                return kind === 255 /* ClassDeclaration */
                    || kind === 169 /* Constructor */
                    || kind === 167 /* MethodDeclaration */
                    || kind === 170 /* GetAccessor */
                    || kind === 171 /* SetAccessor */;
            }
            function createSuperElementAccessInAsyncMethod(argumentExpression, location) {
                if (enclosingSuperContainerFlags & 4096 /* AsyncMethodWithSuperBinding */) {
                    return ts.setTextRange(factory.createPropertyAccessExpression(factory.createCallExpression(factory.createUniqueName("_superIndex", 16 /* Optimistic */ | 32 /* FileLevel */), 
                    /*typeArguments*/ undefined, [argumentExpression]), "value"), location);
                }
                else {
                    return ts.setTextRange(factory.createCallExpression(factory.createUniqueName("_superIndex", 16 /* Optimistic */ | 32 /* FileLevel */), 
                    /*typeArguments*/ undefined, [argumentExpression]), location);
                }
            }
        }
        ts.transformES2017 = transformES2017;
        /** Creates a variable named `_super` with accessor properties for the given property names. */
        function createSuperAccessVariableStatement(factory, resolver, node, names) {
            // Create a variable declaration with a getter/setter (if binding) definition for each name:
            //   const _super = Object.create(null, { x: { get: () => super.x, set: (v) => super.x = v }, ... });
            var hasBinding = (resolver.getNodeCheckFlags(node) & 4096 /* AsyncMethodWithSuperBinding */) !== 0;
            var accessors = [];
            names.forEach(function (_, key) {
                var name = ts.unescapeLeadingUnderscores(key);
                var getterAndSetter = [];
                getterAndSetter.push(factory.createPropertyAssignment("get", factory.createArrowFunction(
                /* modifiers */ undefined, 
                /* typeParameters */ undefined, 
                /* parameters */ [], 
                /* type */ undefined, 
                /* equalsGreaterThanToken */ undefined, ts.setEmitFlags(factory.createPropertyAccessExpression(ts.setEmitFlags(factory.createSuper(), 4 /* NoSubstitution */), name), 4 /* NoSubstitution */))));
                if (hasBinding) {
                    getterAndSetter.push(factory.createPropertyAssignment("set", factory.createArrowFunction(
                    /* modifiers */ undefined, 
                    /* typeParameters */ undefined, 
                    /* parameters */ [
                        factory.createParameterDeclaration(
                        /* decorators */ undefined, 
                        /* modifiers */ undefined, 
                        /* dotDotDotToken */ undefined, "v", 
                        /* questionToken */ undefined, 
                        /* type */ undefined, 
                        /* initializer */ undefined)
                    ], 
                    /* type */ undefined, 
                    /* equalsGreaterThanToken */ undefined, factory.createAssignment(ts.setEmitFlags(factory.createPropertyAccessExpression(ts.setEmitFlags(factory.createSuper(), 4 /* NoSubstitution */), name), 4 /* NoSubstitution */), factory.createIdentifier("v")))));
                }
                accessors.push(factory.createPropertyAssignment(name, factory.createObjectLiteralExpression(getterAndSetter)));
            });
            return factory.createVariableStatement(
            /* modifiers */ undefined, factory.createVariableDeclarationList([
                factory.createVariableDeclaration(factory.createUniqueName("_super", 16 /* Optimistic */ | 32 /* FileLevel */), 
                /*exclamationToken*/ undefined, 
                /* type */ undefined, factory.createCallExpression(factory.createPropertyAccessExpression(factory.createIdentifier("Object"), "create"), 
                /* typeArguments */ undefined, [
                    factory.createNull(),
                    factory.createObjectLiteralExpression(accessors, /* multiline */ true)
                ]))
            ], 2 /* Const */));
        }
        ts.createSuperAccessVariableStatement = createSuperAccessVariableStatement;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var ESNextSubstitutionFlags;
        (function (ESNextSubstitutionFlags) {
            /** Enables substitutions for async methods with `super` calls. */
            ESNextSubstitutionFlags[ESNextSubstitutionFlags["AsyncMethodsWithSuper"] = 1] = "AsyncMethodsWithSuper";
        })(ESNextSubstitutionFlags || (ESNextSubstitutionFlags = {}));
        // Facts we track as we traverse the tree
        var HierarchyFacts;
        (function (HierarchyFacts) {
            HierarchyFacts[HierarchyFacts["None"] = 0] = "None";
            //
            // Ancestor facts
            //
            HierarchyFacts[HierarchyFacts["HasLexicalThis"] = 1] = "HasLexicalThis";
            HierarchyFacts[HierarchyFacts["IterationContainer"] = 2] = "IterationContainer";
            // NOTE: do not add more ancestor flags without also updating AncestorFactsMask below.
            //
            // Ancestor masks
            //
            HierarchyFacts[HierarchyFacts["AncestorFactsMask"] = 3] = "AncestorFactsMask";
            HierarchyFacts[HierarchyFacts["SourceFileIncludes"] = 1] = "SourceFileIncludes";
            HierarchyFacts[HierarchyFacts["SourceFileExcludes"] = 2] = "SourceFileExcludes";
            HierarchyFacts[HierarchyFacts["StrictModeSourceFileIncludes"] = 0] = "StrictModeSourceFileIncludes";
            HierarchyFacts[HierarchyFacts["ClassOrFunctionIncludes"] = 1] = "ClassOrFunctionIncludes";
            HierarchyFacts[HierarchyFacts["ClassOrFunctionExcludes"] = 2] = "ClassOrFunctionExcludes";
            HierarchyFacts[HierarchyFacts["ArrowFunctionIncludes"] = 0] = "ArrowFunctionIncludes";
            HierarchyFacts[HierarchyFacts["ArrowFunctionExcludes"] = 2] = "ArrowFunctionExcludes";
            HierarchyFacts[HierarchyFacts["IterationStatementIncludes"] = 2] = "IterationStatementIncludes";
            HierarchyFacts[HierarchyFacts["IterationStatementExcludes"] = 0] = "IterationStatementExcludes";
        })(HierarchyFacts || (HierarchyFacts = {}));
        function transformES2018(context) {
            var factory = context.factory, emitHelpers = context.getEmitHelperFactory, resumeLexicalEnvironment = context.resumeLexicalEnvironment, endLexicalEnvironment = context.endLexicalEnvironment, hoistVariableDeclaration = context.hoistVariableDeclaration;
            var resolver = context.getEmitResolver();
            var compilerOptions = context.getCompilerOptions();
            var languageVersion = ts.getEmitScriptTarget(compilerOptions);
            var previousOnEmitNode = context.onEmitNode;
            context.onEmitNode = onEmitNode;
            var previousOnSubstituteNode = context.onSubstituteNode;
            context.onSubstituteNode = onSubstituteNode;
            var exportedVariableStatement = false;
            var enabledSubstitutions;
            var enclosingFunctionFlags;
            var enclosingSuperContainerFlags = 0;
            var hierarchyFacts = 0;
            var currentSourceFile;
            var taggedTemplateStringDeclarations;
            /** Keeps track of property names accessed on super (`super.x`) within async functions. */
            var capturedSuperProperties;
            /** Whether the async function contains an element access on super (`super[x]`). */
            var hasSuperElementAccess;
            /** A set of node IDs for generated super accessors. */
            var substitutedSuperAccessors = [];
            return ts.chainBundle(context, transformSourceFile);
            function affectsSubtree(excludeFacts, includeFacts) {
                return hierarchyFacts !== (hierarchyFacts & ~excludeFacts | includeFacts);
            }
            /**
             * Sets the `HierarchyFacts` for this node prior to visiting this node's subtree, returning the facts set prior to modification.
             * @param excludeFacts The existing `HierarchyFacts` to reset before visiting the subtree.
             * @param includeFacts The new `HierarchyFacts` to set before visiting the subtree.
             */
            function enterSubtree(excludeFacts, includeFacts) {
                var ancestorFacts = hierarchyFacts;
                hierarchyFacts = (hierarchyFacts & ~excludeFacts | includeFacts) & 3 /* AncestorFactsMask */;
                return ancestorFacts;
            }
            /**
             * Restores the `HierarchyFacts` for this node's ancestor after visiting this node's
             * subtree.
             * @param ancestorFacts The `HierarchyFacts` of the ancestor to restore after visiting the subtree.
             */
            function exitSubtree(ancestorFacts) {
                hierarchyFacts = ancestorFacts;
            }
            function recordTaggedTemplateString(temp) {
                taggedTemplateStringDeclarations = ts.append(taggedTemplateStringDeclarations, factory.createVariableDeclaration(temp));
            }
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                currentSourceFile = node;
                var visited = visitSourceFile(node);
                ts.addEmitHelpers(visited, context.readEmitHelpers());
                currentSourceFile = undefined;
                taggedTemplateStringDeclarations = undefined;
                return visited;
            }
            function visitor(node) {
                return visitorWorker(node, /*expressionResultIsUnused*/ false);
            }
            function visitorWithUnusedExpressionResult(node) {
                return visitorWorker(node, /*expressionResultIsUnused*/ true);
            }
            function visitorNoAsyncModifier(node) {
                if (node.kind === 130 /* AsyncKeyword */) {
                    return undefined;
                }
                return node;
            }
            function doWithHierarchyFacts(cb, value, excludeFacts, includeFacts) {
                if (affectsSubtree(excludeFacts, includeFacts)) {
                    var ancestorFacts = enterSubtree(excludeFacts, includeFacts);
                    var result = cb(value);
                    exitSubtree(ancestorFacts);
                    return result;
                }
                return cb(value);
            }
            function visitDefault(node) {
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * @param expressionResultIsUnused Indicates the result of an expression is unused by the parent node (i.e., the left side of a comma or the
             * expression of an `ExpressionStatement`).
             */
            function visitorWorker(node, expressionResultIsUnused) {
                if ((node.transformFlags & 64 /* ContainsES2018 */) === 0) {
                    return node;
                }
                switch (node.kind) {
                    case 216 /* AwaitExpression */:
                        return visitAwaitExpression(node);
                    case 222 /* YieldExpression */:
                        return visitYieldExpression(node);
                    case 245 /* ReturnStatement */:
                        return visitReturnStatement(node);
                    case 248 /* LabeledStatement */:
                        return visitLabeledStatement(node);
                    case 203 /* ObjectLiteralExpression */:
                        return visitObjectLiteralExpression(node);
                    case 219 /* BinaryExpression */:
                        return visitBinaryExpression(node, expressionResultIsUnused);
                    case 346 /* CommaListExpression */:
                        return visitCommaListExpression(node, expressionResultIsUnused);
                    case 290 /* CatchClause */:
                        return visitCatchClause(node);
                    case 235 /* VariableStatement */:
                        return visitVariableStatement(node);
                    case 252 /* VariableDeclaration */:
                        return visitVariableDeclaration(node);
                    case 238 /* DoStatement */:
                    case 239 /* WhileStatement */:
                    case 241 /* ForInStatement */:
                        return doWithHierarchyFacts(visitDefault, node, 0 /* IterationStatementExcludes */, 2 /* IterationStatementIncludes */);
                    case 242 /* ForOfStatement */:
                        return visitForOfStatement(node, /*outermostLabeledStatement*/ undefined);
                    case 240 /* ForStatement */:
                        return doWithHierarchyFacts(visitForStatement, node, 0 /* IterationStatementExcludes */, 2 /* IterationStatementIncludes */);
                    case 215 /* VoidExpression */:
                        return visitVoidExpression(node);
                    case 169 /* Constructor */:
                        return doWithHierarchyFacts(visitConstructorDeclaration, node, 2 /* ClassOrFunctionExcludes */, 1 /* ClassOrFunctionIncludes */);
                    case 167 /* MethodDeclaration */:
                        return doWithHierarchyFacts(visitMethodDeclaration, node, 2 /* ClassOrFunctionExcludes */, 1 /* ClassOrFunctionIncludes */);
                    case 170 /* GetAccessor */:
                        return doWithHierarchyFacts(visitGetAccessorDeclaration, node, 2 /* ClassOrFunctionExcludes */, 1 /* ClassOrFunctionIncludes */);
                    case 171 /* SetAccessor */:
                        return doWithHierarchyFacts(visitSetAccessorDeclaration, node, 2 /* ClassOrFunctionExcludes */, 1 /* ClassOrFunctionIncludes */);
                    case 254 /* FunctionDeclaration */:
                        return doWithHierarchyFacts(visitFunctionDeclaration, node, 2 /* ClassOrFunctionExcludes */, 1 /* ClassOrFunctionIncludes */);
                    case 211 /* FunctionExpression */:
                        return doWithHierarchyFacts(visitFunctionExpression, node, 2 /* ClassOrFunctionExcludes */, 1 /* ClassOrFunctionIncludes */);
                    case 212 /* ArrowFunction */:
                        return doWithHierarchyFacts(visitArrowFunction, node, 2 /* ArrowFunctionExcludes */, 0 /* ArrowFunctionIncludes */);
                    case 162 /* Parameter */:
                        return visitParameter(node);
                    case 236 /* ExpressionStatement */:
                        return visitExpressionStatement(node);
                    case 210 /* ParenthesizedExpression */:
                        return visitParenthesizedExpression(node, expressionResultIsUnused);
                    case 208 /* TaggedTemplateExpression */:
                        return visitTaggedTemplateExpression(node);
                    case 204 /* PropertyAccessExpression */:
                        if (capturedSuperProperties && ts.isPropertyAccessExpression(node) && node.expression.kind === 106 /* SuperKeyword */) {
                            capturedSuperProperties.add(node.name.escapedText);
                        }
                        return ts.visitEachChild(node, visitor, context);
                    case 205 /* ElementAccessExpression */:
                        if (capturedSuperProperties && node.expression.kind === 106 /* SuperKeyword */) {
                            hasSuperElementAccess = true;
                        }
                        return ts.visitEachChild(node, visitor, context);
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                        return doWithHierarchyFacts(visitDefault, node, 2 /* ClassOrFunctionExcludes */, 1 /* ClassOrFunctionIncludes */);
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function visitAwaitExpression(node) {
                if (enclosingFunctionFlags & 2 /* Async */ && enclosingFunctionFlags & 1 /* Generator */) {
                    return ts.setOriginalNode(ts.setTextRange(factory.createYieldExpression(/*asteriskToken*/ undefined, emitHelpers().createAwaitHelper(ts.visitNode(node.expression, visitor, ts.isExpression))), 
                    /*location*/ node), node);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitYieldExpression(node) {
                if (enclosingFunctionFlags & 2 /* Async */ && enclosingFunctionFlags & 1 /* Generator */) {
                    if (node.asteriskToken) {
                        var expression = ts.visitNode(ts.Debug.assertDefined(node.expression), visitor, ts.isExpression);
                        return ts.setOriginalNode(ts.setTextRange(factory.createYieldExpression(
                        /*asteriskToken*/ undefined, emitHelpers().createAwaitHelper(factory.updateYieldExpression(node, node.asteriskToken, ts.setTextRange(emitHelpers().createAsyncDelegatorHelper(ts.setTextRange(emitHelpers().createAsyncValuesHelper(expression), expression)), expression)))), node), node);
                    }
                    return ts.setOriginalNode(ts.setTextRange(factory.createYieldExpression(
                    /*asteriskToken*/ undefined, createDownlevelAwait(node.expression
                        ? ts.visitNode(node.expression, visitor, ts.isExpression)
                        : factory.createVoidZero())), node), node);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitReturnStatement(node) {
                if (enclosingFunctionFlags & 2 /* Async */ && enclosingFunctionFlags & 1 /* Generator */) {
                    return factory.updateReturnStatement(node, createDownlevelAwait(node.expression ? ts.visitNode(node.expression, visitor, ts.isExpression) : factory.createVoidZero()));
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitLabeledStatement(node) {
                if (enclosingFunctionFlags & 2 /* Async */) {
                    var statement = ts.unwrapInnermostStatementOfLabel(node);
                    if (statement.kind === 242 /* ForOfStatement */ && statement.awaitModifier) {
                        return visitForOfStatement(statement, node);
                    }
                    return factory.restoreEnclosingLabel(ts.visitNode(statement, visitor, ts.isStatement, factory.liftToBlock), node);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function chunkObjectLiteralElements(elements) {
                var chunkObject;
                var objects = [];
                for (var _i = 0, elements_4 = elements; _i < elements_4.length; _i++) {
                    var e = elements_4[_i];
                    if (e.kind === 293 /* SpreadAssignment */) {
                        if (chunkObject) {
                            objects.push(factory.createObjectLiteralExpression(chunkObject));
                            chunkObject = undefined;
                        }
                        var target = e.expression;
                        objects.push(ts.visitNode(target, visitor, ts.isExpression));
                    }
                    else {
                        chunkObject = ts.append(chunkObject, e.kind === 291 /* PropertyAssignment */
                            ? factory.createPropertyAssignment(e.name, ts.visitNode(e.initializer, visitor, ts.isExpression))
                            : ts.visitNode(e, visitor, ts.isObjectLiteralElementLike));
                    }
                }
                if (chunkObject) {
                    objects.push(factory.createObjectLiteralExpression(chunkObject));
                }
                return objects;
            }
            function visitObjectLiteralExpression(node) {
                if (node.transformFlags & 32768 /* ContainsObjectRestOrSpread */) {
                    // spread elements emit like so:
                    // non-spread elements are chunked together into object literals, and then all are passed to __assign:
                    //     { a, ...o, b } => __assign(__assign({a}, o), {b});
                    // If the first element is a spread element, then the first argument to __assign is {}:
                    //     { ...o, a, b, ...o2 } => __assign(__assign(__assign({}, o), {a, b}), o2)
                    //
                    // We cannot call __assign with more than two elements, since any element could cause side effects. For
                    // example:
                    //      var k = { a: 1, b: 2 };
                    //      var o = { a: 3, ...k, b: k.a++ };
                    //      // expected: { a: 1, b: 1 }
                    // If we translate the above to `__assign({ a: 3 }, k, { b: k.a++ })`, the `k.a++` will evaluate before
                    // `k` is spread and we end up with `{ a: 2, b: 1 }`.
                    //
                    // This also occurs for spread elements, not just property assignments:
                    //      var k = { a: 1, get b() { l = { z: 9 }; return 2; } };
                    //      var l = { c: 3 };
                    //      var o = { ...k, ...l };
                    //      // expected: { a: 1, b: 2, z: 9 }
                    // If we translate the above to `__assign({}, k, l)`, the `l` will evaluate before `k` is spread and we
                    // end up with `{ a: 1, b: 2, c: 3 }`
                    var objects = chunkObjectLiteralElements(node.properties);
                    if (objects.length && objects[0].kind !== 203 /* ObjectLiteralExpression */) {
                        objects.unshift(factory.createObjectLiteralExpression());
                    }
                    var expression = objects[0];
                    if (objects.length > 1) {
                        for (var i = 1; i < objects.length; i++) {
                            expression = emitHelpers().createAssignHelper([expression, objects[i]]);
                        }
                        return expression;
                    }
                    else {
                        return emitHelpers().createAssignHelper(objects);
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitExpressionStatement(node) {
                return ts.visitEachChild(node, visitorWithUnusedExpressionResult, context);
            }
            /**
             * @param expressionResultIsUnused Indicates the result of an expression is unused by the parent node (i.e., the left side of a comma or the
             * expression of an `ExpressionStatement`).
             */
            function visitParenthesizedExpression(node, expressionResultIsUnused) {
                return ts.visitEachChild(node, expressionResultIsUnused ? visitorWithUnusedExpressionResult : visitor, context);
            }
            function visitSourceFile(node) {
                var ancestorFacts = enterSubtree(2 /* SourceFileExcludes */, ts.isEffectiveStrictModeSourceFile(node, compilerOptions) ?
                    0 /* StrictModeSourceFileIncludes */ :
                    1 /* SourceFileIncludes */);
                exportedVariableStatement = false;
                var visited = ts.visitEachChild(node, visitor, context);
                var statement = ts.concatenate(visited.statements, taggedTemplateStringDeclarations && [
                    factory.createVariableStatement(/*modifiers*/ undefined, factory.createVariableDeclarationList(taggedTemplateStringDeclarations))
                ]);
                var result = factory.updateSourceFile(visited, ts.setTextRange(factory.createNodeArray(statement), node.statements));
                exitSubtree(ancestorFacts);
                return result;
            }
            function visitTaggedTemplateExpression(node) {
                return ts.processTaggedTemplateExpression(context, node, visitor, currentSourceFile, recordTaggedTemplateString, ts.ProcessLevel.LiftRestriction);
            }
            /**
             * Visits a BinaryExpression that contains a destructuring assignment.
             *
             * @param node A BinaryExpression node.
             * @param expressionResultIsUnused Indicates the result of an expression is unused by the parent node (i.e., the left side of a comma or the
             * expression of an `ExpressionStatement`).
             */
            function visitBinaryExpression(node, expressionResultIsUnused) {
                if (ts.isDestructuringAssignment(node) && node.left.transformFlags & 32768 /* ContainsObjectRestOrSpread */) {
                    return ts.flattenDestructuringAssignment(node, visitor, context, 1 /* ObjectRest */, !expressionResultIsUnused);
                }
                if (node.operatorToken.kind === 27 /* CommaToken */) {
                    return factory.updateBinaryExpression(node, ts.visitNode(node.left, visitorWithUnusedExpressionResult, ts.isExpression), node.operatorToken, ts.visitNode(node.right, expressionResultIsUnused ? visitorWithUnusedExpressionResult : visitor, ts.isExpression));
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * @param expressionResultIsUnused Indicates the result of an expression is unused by the parent node (i.e., the left side of a comma or the
             * expression of an `ExpressionStatement`).
             */
            function visitCommaListExpression(node, expressionResultIsUnused) {
                if (expressionResultIsUnused) {
                    return ts.visitEachChild(node, visitorWithUnusedExpressionResult, context);
                }
                var result;
                for (var i = 0; i < node.elements.length; i++) {
                    var element = node.elements[i];
                    var visited = ts.visitNode(element, i < node.elements.length - 1 ? visitorWithUnusedExpressionResult : visitor, ts.isExpression);
                    if (result || visited !== element) {
                        result || (result = node.elements.slice(0, i));
                        result.push(visited);
                    }
                }
                var elements = result ? ts.setTextRange(factory.createNodeArray(result), node.elements) : node.elements;
                return factory.updateCommaListExpression(node, elements);
            }
            function visitCatchClause(node) {
                if (node.variableDeclaration &&
                    ts.isBindingPattern(node.variableDeclaration.name) &&
                    node.variableDeclaration.name.transformFlags & 32768 /* ContainsObjectRestOrSpread */) {
                    var name = factory.getGeneratedNameForNode(node.variableDeclaration.name);
                    var updatedDecl = factory.updateVariableDeclaration(node.variableDeclaration, node.variableDeclaration.name, /*exclamationToken*/ undefined, /*type*/ undefined, name);
                    var visitedBindings = ts.flattenDestructuringBinding(updatedDecl, visitor, context, 1 /* ObjectRest */);
                    var block = ts.visitNode(node.block, visitor, ts.isBlock);
                    if (ts.some(visitedBindings)) {
                        block = factory.updateBlock(block, __spreadArray([
                            factory.createVariableStatement(/*modifiers*/ undefined, visitedBindings)
                        ], block.statements, true));
                    }
                    return factory.updateCatchClause(node, factory.updateVariableDeclaration(node.variableDeclaration, name, /*exclamationToken*/ undefined, /*type*/ undefined, /*initializer*/ undefined), block);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitVariableStatement(node) {
                if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                    var savedExportedVariableStatement = exportedVariableStatement;
                    exportedVariableStatement = true;
                    var visited = ts.visitEachChild(node, visitor, context);
                    exportedVariableStatement = savedExportedVariableStatement;
                    return visited;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visits a VariableDeclaration node with a binding pattern.
             *
             * @param node A VariableDeclaration node.
             */
            function visitVariableDeclaration(node) {
                if (exportedVariableStatement) {
                    var savedExportedVariableStatement = exportedVariableStatement;
                    exportedVariableStatement = false;
                    var visited = visitVariableDeclarationWorker(node, /*exportedVariableStatement*/ true);
                    exportedVariableStatement = savedExportedVariableStatement;
                    return visited;
                }
                return visitVariableDeclarationWorker(node, /*exportedVariableStatement*/ false);
            }
            function visitVariableDeclarationWorker(node, exportedVariableStatement) {
                // If we are here it is because the name contains a binding pattern with a rest somewhere in it.
                if (ts.isBindingPattern(node.name) && node.name.transformFlags & 32768 /* ContainsObjectRestOrSpread */) {
                    return ts.flattenDestructuringBinding(node, visitor, context, 1 /* ObjectRest */, 
                    /*rval*/ undefined, exportedVariableStatement);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitForStatement(node) {
                return factory.updateForStatement(node, ts.visitNode(node.initializer, visitorWithUnusedExpressionResult, ts.isForInitializer), ts.visitNode(node.condition, visitor, ts.isExpression), ts.visitNode(node.incrementor, visitorWithUnusedExpressionResult, ts.isExpression), ts.visitIterationBody(node.statement, visitor, context));
            }
            function visitVoidExpression(node) {
                return ts.visitEachChild(node, visitorWithUnusedExpressionResult, context);
            }
            /**
             * Visits a ForOfStatement and converts it into a ES2015-compatible ForOfStatement.
             *
             * @param node A ForOfStatement.
             */
            function visitForOfStatement(node, outermostLabeledStatement) {
                var ancestorFacts = enterSubtree(0 /* IterationStatementExcludes */, 2 /* IterationStatementIncludes */);
                if (node.initializer.transformFlags & 32768 /* ContainsObjectRestOrSpread */) {
                    node = transformForOfStatementWithObjectRest(node);
                }
                var result = node.awaitModifier ?
                    transformForAwaitOfStatement(node, outermostLabeledStatement, ancestorFacts) :
                    factory.restoreEnclosingLabel(ts.visitEachChild(node, visitor, context), outermostLabeledStatement);
                exitSubtree(ancestorFacts);
                return result;
            }
            function transformForOfStatementWithObjectRest(node) {
                var initializerWithoutParens = ts.skipParentheses(node.initializer);
                if (ts.isVariableDeclarationList(initializerWithoutParens) || ts.isAssignmentPattern(initializerWithoutParens)) {
                    var bodyLocation = void 0;
                    var statementsLocation = void 0;
                    var temp = factory.createTempVariable(/*recordTempVariable*/ undefined);
                    var statements = [ts.createForOfBindingStatement(factory, initializerWithoutParens, temp)];
                    if (ts.isBlock(node.statement)) {
                        ts.addRange(statements, node.statement.statements);
                        bodyLocation = node.statement;
                        statementsLocation = node.statement.statements;
                    }
                    else if (node.statement) {
                        ts.append(statements, node.statement);
                        bodyLocation = node.statement;
                        statementsLocation = node.statement;
                    }
                    return factory.updateForOfStatement(node, node.awaitModifier, ts.setTextRange(factory.createVariableDeclarationList([
                        ts.setTextRange(factory.createVariableDeclaration(temp), node.initializer)
                    ], 1 /* Let */), node.initializer), node.expression, ts.setTextRange(factory.createBlock(ts.setTextRange(factory.createNodeArray(statements), statementsLocation), 
                    /*multiLine*/ true), bodyLocation));
                }
                return node;
            }
            function convertForOfStatementHead(node, boundValue) {
                var binding = ts.createForOfBindingStatement(factory, node.initializer, boundValue);
                var bodyLocation;
                var statementsLocation;
                var statements = [ts.visitNode(binding, visitor, ts.isStatement)];
                var statement = ts.visitIterationBody(node.statement, visitor, context);
                if (ts.isBlock(statement)) {
                    ts.addRange(statements, statement.statements);
                    bodyLocation = statement;
                    statementsLocation = statement.statements;
                }
                else {
                    statements.push(statement);
                }
                return ts.setEmitFlags(ts.setTextRange(factory.createBlock(ts.setTextRange(factory.createNodeArray(statements), statementsLocation), 
                /*multiLine*/ true), bodyLocation), 48 /* NoSourceMap */ | 384 /* NoTokenSourceMaps */);
            }
            function createDownlevelAwait(expression) {
                return enclosingFunctionFlags & 1 /* Generator */
                    ? factory.createYieldExpression(/*asteriskToken*/ undefined, emitHelpers().createAwaitHelper(expression))
                    : factory.createAwaitExpression(expression);
            }
            function transformForAwaitOfStatement(node, outermostLabeledStatement, ancestorFacts) {
                var expression = ts.visitNode(node.expression, visitor, ts.isExpression);
                var iterator = ts.isIdentifier(expression) ? factory.getGeneratedNameForNode(expression) : factory.createTempVariable(/*recordTempVariable*/ undefined);
                var result = ts.isIdentifier(expression) ? factory.getGeneratedNameForNode(iterator) : factory.createTempVariable(/*recordTempVariable*/ undefined);
                var errorRecord = factory.createUniqueName("e");
                var catchVariable = factory.getGeneratedNameForNode(errorRecord);
                var returnMethod = factory.createTempVariable(/*recordTempVariable*/ undefined);
                var callValues = ts.setTextRange(emitHelpers().createAsyncValuesHelper(expression), node.expression);
                var callNext = factory.createCallExpression(factory.createPropertyAccessExpression(iterator, "next"), /*typeArguments*/ undefined, []);
                var getDone = factory.createPropertyAccessExpression(result, "done");
                var getValue = factory.createPropertyAccessExpression(result, "value");
                var callReturn = factory.createFunctionCallCall(returnMethod, iterator, []);
                hoistVariableDeclaration(errorRecord);
                hoistVariableDeclaration(returnMethod);
                // if we are enclosed in an outer loop ensure we reset 'errorRecord' per each iteration
                var initializer = ancestorFacts & 2 /* IterationContainer */ ?
                    factory.inlineExpressions([factory.createAssignment(errorRecord, factory.createVoidZero()), callValues]) :
                    callValues;
                var forStatement = ts.setEmitFlags(ts.setTextRange(factory.createForStatement(
                /*initializer*/ ts.setEmitFlags(ts.setTextRange(factory.createVariableDeclarationList([
                    ts.setTextRange(factory.createVariableDeclaration(iterator, /*exclamationToken*/ undefined, /*type*/ undefined, initializer), node.expression),
                    factory.createVariableDeclaration(result)
                ]), node.expression), 2097152 /* NoHoisting */), 
                /*condition*/ factory.createComma(factory.createAssignment(result, createDownlevelAwait(callNext)), factory.createLogicalNot(getDone)), 
                /*incrementor*/ undefined, 
                /*statement*/ convertForOfStatementHead(node, getValue)), 
                /*location*/ node), 256 /* NoTokenTrailingSourceMaps */);
                return factory.createTryStatement(factory.createBlock([
                    factory.restoreEnclosingLabel(forStatement, outermostLabeledStatement)
                ]), factory.createCatchClause(factory.createVariableDeclaration(catchVariable), ts.setEmitFlags(factory.createBlock([
                    factory.createExpressionStatement(factory.createAssignment(errorRecord, factory.createObjectLiteralExpression([
                        factory.createPropertyAssignment("error", catchVariable)
                    ])))
                ]), 1 /* SingleLine */)), factory.createBlock([
                    factory.createTryStatement(
                    /*tryBlock*/ factory.createBlock([
                        ts.setEmitFlags(factory.createIfStatement(factory.createLogicalAnd(factory.createLogicalAnd(result, factory.createLogicalNot(getDone)), factory.createAssignment(returnMethod, factory.createPropertyAccessExpression(iterator, "return"))), factory.createExpressionStatement(createDownlevelAwait(callReturn))), 1 /* SingleLine */)
                    ]), 
                    /*catchClause*/ undefined, 
                    /*finallyBlock*/ ts.setEmitFlags(factory.createBlock([
                        ts.setEmitFlags(factory.createIfStatement(errorRecord, factory.createThrowStatement(factory.createPropertyAccessExpression(errorRecord, "error"))), 1 /* SingleLine */)
                    ]), 1 /* SingleLine */))
                ]));
            }
            function visitParameter(node) {
                if (node.transformFlags & 32768 /* ContainsObjectRestOrSpread */) {
                    // Binding patterns are converted into a generated name and are
                    // evaluated inside the function body.
                    return factory.updateParameterDeclaration(node, 
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, node.dotDotDotToken, factory.getGeneratedNameForNode(node), 
                    /*questionToken*/ undefined, 
                    /*type*/ undefined, ts.visitNode(node.initializer, visitor, ts.isExpression));
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitConstructorDeclaration(node) {
                var savedEnclosingFunctionFlags = enclosingFunctionFlags;
                enclosingFunctionFlags = 0 /* Normal */;
                var updated = factory.updateConstructorDeclaration(node, 
                /*decorators*/ undefined, node.modifiers, ts.visitParameterList(node.parameters, visitor, context), transformFunctionBody(node));
                enclosingFunctionFlags = savedEnclosingFunctionFlags;
                return updated;
            }
            function visitGetAccessorDeclaration(node) {
                var savedEnclosingFunctionFlags = enclosingFunctionFlags;
                enclosingFunctionFlags = 0 /* Normal */;
                var updated = factory.updateGetAccessorDeclaration(node, 
                /*decorators*/ undefined, node.modifiers, ts.visitNode(node.name, visitor, ts.isPropertyName), ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, transformFunctionBody(node));
                enclosingFunctionFlags = savedEnclosingFunctionFlags;
                return updated;
            }
            function visitSetAccessorDeclaration(node) {
                var savedEnclosingFunctionFlags = enclosingFunctionFlags;
                enclosingFunctionFlags = 0 /* Normal */;
                var updated = factory.updateSetAccessorDeclaration(node, 
                /*decorators*/ undefined, node.modifiers, ts.visitNode(node.name, visitor, ts.isPropertyName), ts.visitParameterList(node.parameters, visitor, context), transformFunctionBody(node));
                enclosingFunctionFlags = savedEnclosingFunctionFlags;
                return updated;
            }
            function visitMethodDeclaration(node) {
                var savedEnclosingFunctionFlags = enclosingFunctionFlags;
                enclosingFunctionFlags = ts.getFunctionFlags(node);
                var updated = factory.updateMethodDeclaration(node, 
                /*decorators*/ undefined, enclosingFunctionFlags & 1 /* Generator */
                    ? ts.visitNodes(node.modifiers, visitorNoAsyncModifier, ts.isModifier)
                    : node.modifiers, enclosingFunctionFlags & 2 /* Async */
                    ? undefined
                    : node.asteriskToken, ts.visitNode(node.name, visitor, ts.isPropertyName), ts.visitNode(/*questionToken*/ undefined, visitor, ts.isToken), 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, enclosingFunctionFlags & 2 /* Async */ && enclosingFunctionFlags & 1 /* Generator */
                    ? transformAsyncGeneratorFunctionBody(node)
                    : transformFunctionBody(node));
                enclosingFunctionFlags = savedEnclosingFunctionFlags;
                return updated;
            }
            function visitFunctionDeclaration(node) {
                var savedEnclosingFunctionFlags = enclosingFunctionFlags;
                enclosingFunctionFlags = ts.getFunctionFlags(node);
                var updated = factory.updateFunctionDeclaration(node, 
                /*decorators*/ undefined, enclosingFunctionFlags & 1 /* Generator */
                    ? ts.visitNodes(node.modifiers, visitorNoAsyncModifier, ts.isModifier)
                    : node.modifiers, enclosingFunctionFlags & 2 /* Async */
                    ? undefined
                    : node.asteriskToken, node.name, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, enclosingFunctionFlags & 2 /* Async */ && enclosingFunctionFlags & 1 /* Generator */
                    ? transformAsyncGeneratorFunctionBody(node)
                    : transformFunctionBody(node));
                enclosingFunctionFlags = savedEnclosingFunctionFlags;
                return updated;
            }
            function visitArrowFunction(node) {
                var savedEnclosingFunctionFlags = enclosingFunctionFlags;
                enclosingFunctionFlags = ts.getFunctionFlags(node);
                var updated = factory.updateArrowFunction(node, node.modifiers, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, node.equalsGreaterThanToken, transformFunctionBody(node));
                enclosingFunctionFlags = savedEnclosingFunctionFlags;
                return updated;
            }
            function visitFunctionExpression(node) {
                var savedEnclosingFunctionFlags = enclosingFunctionFlags;
                enclosingFunctionFlags = ts.getFunctionFlags(node);
                var updated = factory.updateFunctionExpression(node, enclosingFunctionFlags & 1 /* Generator */
                    ? ts.visitNodes(node.modifiers, visitorNoAsyncModifier, ts.isModifier)
                    : node.modifiers, enclosingFunctionFlags & 2 /* Async */
                    ? undefined
                    : node.asteriskToken, node.name, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, enclosingFunctionFlags & 2 /* Async */ && enclosingFunctionFlags & 1 /* Generator */
                    ? transformAsyncGeneratorFunctionBody(node)
                    : transformFunctionBody(node));
                enclosingFunctionFlags = savedEnclosingFunctionFlags;
                return updated;
            }
            function transformAsyncGeneratorFunctionBody(node) {
                resumeLexicalEnvironment();
                var statements = [];
                var statementOffset = factory.copyPrologue(node.body.statements, statements, /*ensureUseStrict*/ false, visitor);
                appendObjectRestAssignmentsIfNeeded(statements, node);
                var savedCapturedSuperProperties = capturedSuperProperties;
                var savedHasSuperElementAccess = hasSuperElementAccess;
                capturedSuperProperties = new ts.Set();
                hasSuperElementAccess = false;
                var returnStatement = factory.createReturnStatement(emitHelpers().createAsyncGeneratorHelper(factory.createFunctionExpression(
                /*modifiers*/ undefined, factory.createToken(41 /* AsteriskToken */), node.name && factory.getGeneratedNameForNode(node.name), 
                /*typeParameters*/ undefined, 
                /*parameters*/ [], 
                /*type*/ undefined, factory.updateBlock(node.body, ts.visitLexicalEnvironment(node.body.statements, visitor, context, statementOffset))), !!(hierarchyFacts & 1 /* HasLexicalThis */)));
                // Minor optimization, emit `_super` helper to capture `super` access in an arrow.
                // This step isn't needed if we eventually transform this to ES5.
                var emitSuperHelpers = languageVersion >= 2 /* ES2015 */ && resolver.getNodeCheckFlags(node) & (4096 /* AsyncMethodWithSuperBinding */ | 2048 /* AsyncMethodWithSuper */);
                if (emitSuperHelpers) {
                    enableSubstitutionForAsyncMethodsWithSuper();
                    var variableStatement = ts.createSuperAccessVariableStatement(factory, resolver, node, capturedSuperProperties);
                    substitutedSuperAccessors[ts.getNodeId(variableStatement)] = true;
                    ts.insertStatementsAfterStandardPrologue(statements, [variableStatement]);
                }
                statements.push(returnStatement);
                ts.insertStatementsAfterStandardPrologue(statements, endLexicalEnvironment());
                var block = factory.updateBlock(node.body, statements);
                if (emitSuperHelpers && hasSuperElementAccess) {
                    if (resolver.getNodeCheckFlags(node) & 4096 /* AsyncMethodWithSuperBinding */) {
                        ts.addEmitHelper(block, ts.advancedAsyncSuperHelper);
                    }
                    else if (resolver.getNodeCheckFlags(node) & 2048 /* AsyncMethodWithSuper */) {
                        ts.addEmitHelper(block, ts.asyncSuperHelper);
                    }
                }
                capturedSuperProperties = savedCapturedSuperProperties;
                hasSuperElementAccess = savedHasSuperElementAccess;
                return block;
            }
            function transformFunctionBody(node) {
                var _a;
                resumeLexicalEnvironment();
                var statementOffset = 0;
                var statements = [];
                var body = (_a = ts.visitNode(node.body, visitor, ts.isConciseBody)) !== null && _a !== void 0 ? _a : factory.createBlock([]);
                if (ts.isBlock(body)) {
                    statementOffset = factory.copyPrologue(body.statements, statements, /*ensureUseStrict*/ false, visitor);
                }
                ts.addRange(statements, appendObjectRestAssignmentsIfNeeded(/*statements*/ undefined, node));
                var leadingStatements = endLexicalEnvironment();
                if (statementOffset > 0 || ts.some(statements) || ts.some(leadingStatements)) {
                    var block = factory.converters.convertToFunctionBlock(body, /*multiLine*/ true);
                    ts.insertStatementsAfterStandardPrologue(statements, leadingStatements);
                    ts.addRange(statements, block.statements.slice(statementOffset));
                    return factory.updateBlock(block, ts.setTextRange(factory.createNodeArray(statements), block.statements));
                }
                return body;
            }
            function appendObjectRestAssignmentsIfNeeded(statements, node) {
                for (var _i = 0, _a = node.parameters; _i < _a.length; _i++) {
                    var parameter = _a[_i];
                    if (parameter.transformFlags & 32768 /* ContainsObjectRestOrSpread */) {
                        var temp = factory.getGeneratedNameForNode(parameter);
                        var declarations = ts.flattenDestructuringBinding(parameter, visitor, context, 1 /* ObjectRest */, temp, 
                        /*doNotRecordTempVariablesInLine*/ false, 
                        /*skipInitializer*/ true);
                        if (ts.some(declarations)) {
                            var statement = factory.createVariableStatement(
                            /*modifiers*/ undefined, factory.createVariableDeclarationList(declarations));
                            ts.setEmitFlags(statement, 1048576 /* CustomPrologue */);
                            statements = ts.append(statements, statement);
                        }
                    }
                }
                return statements;
            }
            function enableSubstitutionForAsyncMethodsWithSuper() {
                if ((enabledSubstitutions & 1 /* AsyncMethodsWithSuper */) === 0) {
                    enabledSubstitutions |= 1 /* AsyncMethodsWithSuper */;
                    // We need to enable substitutions for call, property access, and element access
                    // if we need to rewrite super calls.
                    context.enableSubstitution(206 /* CallExpression */);
                    context.enableSubstitution(204 /* PropertyAccessExpression */);
                    context.enableSubstitution(205 /* ElementAccessExpression */);
                    // We need to be notified when entering and exiting declarations that bind super.
                    context.enableEmitNotification(255 /* ClassDeclaration */);
                    context.enableEmitNotification(167 /* MethodDeclaration */);
                    context.enableEmitNotification(170 /* GetAccessor */);
                    context.enableEmitNotification(171 /* SetAccessor */);
                    context.enableEmitNotification(169 /* Constructor */);
                    // We need to be notified when entering the generated accessor arrow functions.
                    context.enableEmitNotification(235 /* VariableStatement */);
                }
            }
            /**
             * Called by the printer just before a node is printed.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to be printed.
             * @param emitCallback The callback used to emit the node.
             */
            function onEmitNode(hint, node, emitCallback) {
                // If we need to support substitutions for `super` in an async method,
                // we should track it here.
                if (enabledSubstitutions & 1 /* AsyncMethodsWithSuper */ && isSuperContainer(node)) {
                    var superContainerFlags = resolver.getNodeCheckFlags(node) & (2048 /* AsyncMethodWithSuper */ | 4096 /* AsyncMethodWithSuperBinding */);
                    if (superContainerFlags !== enclosingSuperContainerFlags) {
                        var savedEnclosingSuperContainerFlags = enclosingSuperContainerFlags;
                        enclosingSuperContainerFlags = superContainerFlags;
                        previousOnEmitNode(hint, node, emitCallback);
                        enclosingSuperContainerFlags = savedEnclosingSuperContainerFlags;
                        return;
                    }
                }
                // Disable substitution in the generated super accessor itself.
                else if (enabledSubstitutions && substitutedSuperAccessors[ts.getNodeId(node)]) {
                    var savedEnclosingSuperContainerFlags = enclosingSuperContainerFlags;
                    enclosingSuperContainerFlags = 0;
                    previousOnEmitNode(hint, node, emitCallback);
                    enclosingSuperContainerFlags = savedEnclosingSuperContainerFlags;
                    return;
                }
                previousOnEmitNode(hint, node, emitCallback);
            }
            /**
             * Hooks node substitutions.
             *
             * @param hint The context for the emitter.
             * @param node The node to substitute.
             */
            function onSubstituteNode(hint, node) {
                node = previousOnSubstituteNode(hint, node);
                if (hint === 1 /* Expression */ && enclosingSuperContainerFlags) {
                    return substituteExpression(node);
                }
                return node;
            }
            function substituteExpression(node) {
                switch (node.kind) {
                    case 204 /* PropertyAccessExpression */:
                        return substitutePropertyAccessExpression(node);
                    case 205 /* ElementAccessExpression */:
                        return substituteElementAccessExpression(node);
                    case 206 /* CallExpression */:
                        return substituteCallExpression(node);
                }
                return node;
            }
            function substitutePropertyAccessExpression(node) {
                if (node.expression.kind === 106 /* SuperKeyword */) {
                    return ts.setTextRange(factory.createPropertyAccessExpression(factory.createUniqueName("_super", 16 /* Optimistic */ | 32 /* FileLevel */), node.name), node);
                }
                return node;
            }
            function substituteElementAccessExpression(node) {
                if (node.expression.kind === 106 /* SuperKeyword */) {
                    return createSuperElementAccessInAsyncMethod(node.argumentExpression, node);
                }
                return node;
            }
            function substituteCallExpression(node) {
                var expression = node.expression;
                if (ts.isSuperProperty(expression)) {
                    var argumentExpression = ts.isPropertyAccessExpression(expression)
                        ? substitutePropertyAccessExpression(expression)
                        : substituteElementAccessExpression(expression);
                    return factory.createCallExpression(factory.createPropertyAccessExpression(argumentExpression, "call"), 
                    /*typeArguments*/ undefined, __spreadArray([
                        factory.createThis()
                    ], node.arguments, true));
                }
                return node;
            }
            function isSuperContainer(node) {
                var kind = node.kind;
                return kind === 255 /* ClassDeclaration */
                    || kind === 169 /* Constructor */
                    || kind === 167 /* MethodDeclaration */
                    || kind === 170 /* GetAccessor */
                    || kind === 171 /* SetAccessor */;
            }
            function createSuperElementAccessInAsyncMethod(argumentExpression, location) {
                if (enclosingSuperContainerFlags & 4096 /* AsyncMethodWithSuperBinding */) {
                    return ts.setTextRange(factory.createPropertyAccessExpression(factory.createCallExpression(factory.createIdentifier("_superIndex"), 
                    /*typeArguments*/ undefined, [argumentExpression]), "value"), location);
                }
                else {
                    return ts.setTextRange(factory.createCallExpression(factory.createIdentifier("_superIndex"), 
                    /*typeArguments*/ undefined, [argumentExpression]), location);
                }
            }
        }
        ts.transformES2018 = transformES2018;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function transformES2019(context) {
            var factory = context.factory;
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitor(node) {
                if ((node.transformFlags & 32 /* ContainsES2019 */) === 0) {
                    return node;
                }
                switch (node.kind) {
                    case 290 /* CatchClause */:
                        return visitCatchClause(node);
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function visitCatchClause(node) {
                if (!node.variableDeclaration) {
                    return factory.updateCatchClause(node, factory.createVariableDeclaration(factory.createTempVariable(/*recordTempVariable*/ undefined)), ts.visitNode(node.block, visitor, ts.isBlock));
                }
                return ts.visitEachChild(node, visitor, context);
            }
        }
        ts.transformES2019 = transformES2019;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function transformES2020(context) {
            var factory = context.factory, hoistVariableDeclaration = context.hoistVariableDeclaration;
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitor(node) {
                if ((node.transformFlags & 16 /* ContainsES2020 */) === 0) {
                    return node;
                }
                switch (node.kind) {
                    case 206 /* CallExpression */: {
                        var updated = visitNonOptionalCallExpression(node, /*captureThisArg*/ false);
                        ts.Debug.assertNotNode(updated, ts.isSyntheticReference);
                        return updated;
                    }
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */:
                        if (ts.isOptionalChain(node)) {
                            var updated = visitOptionalExpression(node, /*captureThisArg*/ false, /*isDelete*/ false);
                            ts.Debug.assertNotNode(updated, ts.isSyntheticReference);
                            return updated;
                        }
                        return ts.visitEachChild(node, visitor, context);
                    case 219 /* BinaryExpression */:
                        if (node.operatorToken.kind === 60 /* QuestionQuestionToken */) {
                            return transformNullishCoalescingExpression(node);
                        }
                        return ts.visitEachChild(node, visitor, context);
                    case 213 /* DeleteExpression */:
                        return visitDeleteExpression(node);
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function flattenChain(chain) {
                ts.Debug.assertNotNode(chain, ts.isNonNullChain);
                var links = [chain];
                while (!chain.questionDotToken && !ts.isTaggedTemplateExpression(chain)) {
                    chain = ts.cast(ts.skipPartiallyEmittedExpressions(chain.expression), ts.isOptionalChain);
                    ts.Debug.assertNotNode(chain, ts.isNonNullChain);
                    links.unshift(chain);
                }
                return { expression: chain.expression, chain: links };
            }
            function visitNonOptionalParenthesizedExpression(node, captureThisArg, isDelete) {
                var expression = visitNonOptionalExpression(node.expression, captureThisArg, isDelete);
                if (ts.isSyntheticReference(expression)) {
                    // `(a.b)` -> { expression `((_a = a).b)`, thisArg: `_a` }
                    // `(a[b])` -> { expression `((_a = a)[b])`, thisArg: `_a` }
                    return factory.createSyntheticReferenceExpression(factory.updateParenthesizedExpression(node, expression.expression), expression.thisArg);
                }
                return factory.updateParenthesizedExpression(node, expression);
            }
            function visitNonOptionalPropertyOrElementAccessExpression(node, captureThisArg, isDelete) {
                if (ts.isOptionalChain(node)) {
                    // If `node` is an optional chain, then it is the outermost chain of an optional expression.
                    return visitOptionalExpression(node, captureThisArg, isDelete);
                }
                var expression = ts.visitNode(node.expression, visitor, ts.isExpression);
                ts.Debug.assertNotNode(expression, ts.isSyntheticReference);
                var thisArg;
                if (captureThisArg) {
                    if (!ts.isSimpleCopiableExpression(expression)) {
                        thisArg = factory.createTempVariable(hoistVariableDeclaration);
                        expression = factory.createAssignment(thisArg, expression);
                    }
                    else {
                        thisArg = expression;
                    }
                }
                expression = node.kind === 204 /* PropertyAccessExpression */
                    ? factory.updatePropertyAccessExpression(node, expression, ts.visitNode(node.name, visitor, ts.isIdentifier))
                    : factory.updateElementAccessExpression(node, expression, ts.visitNode(node.argumentExpression, visitor, ts.isExpression));
                return thisArg ? factory.createSyntheticReferenceExpression(expression, thisArg) : expression;
            }
            function visitNonOptionalCallExpression(node, captureThisArg) {
                if (ts.isOptionalChain(node)) {
                    // If `node` is an optional chain, then it is the outermost chain of an optional expression.
                    return visitOptionalExpression(node, captureThisArg, /*isDelete*/ false);
                }
                if (ts.isParenthesizedExpression(node.expression) && ts.isOptionalChain(ts.skipParentheses(node.expression))) {
                    // capture thisArg for calls of parenthesized optional chains like `(foo?.bar)()`
                    var expression = visitNonOptionalParenthesizedExpression(node.expression, /*captureThisArg*/ true, /*isDelete*/ false);
                    var args = ts.visitNodes(node.arguments, visitor, ts.isExpression);
                    if (ts.isSyntheticReference(expression)) {
                        return ts.setTextRange(factory.createFunctionCallCall(expression.expression, expression.thisArg, args), node);
                    }
                    return factory.updateCallExpression(node, expression, /*typeArguments*/ undefined, args);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitNonOptionalExpression(node, captureThisArg, isDelete) {
                switch (node.kind) {
                    case 210 /* ParenthesizedExpression */: return visitNonOptionalParenthesizedExpression(node, captureThisArg, isDelete);
                    case 204 /* PropertyAccessExpression */:
                    case 205 /* ElementAccessExpression */: return visitNonOptionalPropertyOrElementAccessExpression(node, captureThisArg, isDelete);
                    case 206 /* CallExpression */: return visitNonOptionalCallExpression(node, captureThisArg);
                    default: return ts.visitNode(node, visitor, ts.isExpression);
                }
            }
            function visitOptionalExpression(node, captureThisArg, isDelete) {
                var _a = flattenChain(node), expression = _a.expression, chain = _a.chain;
                var left = visitNonOptionalExpression(expression, ts.isCallChain(chain[0]), /*isDelete*/ false);
                var leftThisArg = ts.isSyntheticReference(left) ? left.thisArg : undefined;
                var leftExpression = ts.isSyntheticReference(left) ? left.expression : left;
                var capturedLeft = leftExpression;
                if (!ts.isSimpleCopiableExpression(leftExpression)) {
                    capturedLeft = factory.createTempVariable(hoistVariableDeclaration);
                    leftExpression = factory.createAssignment(capturedLeft, leftExpression);
                }
                var rightExpression = capturedLeft;
                var thisArg;
                for (var i = 0; i < chain.length; i++) {
                    var segment = chain[i];
                    switch (segment.kind) {
                        case 204 /* PropertyAccessExpression */:
                        case 205 /* ElementAccessExpression */:
                            if (i === chain.length - 1 && captureThisArg) {
                                if (!ts.isSimpleCopiableExpression(rightExpression)) {
                                    thisArg = factory.createTempVariable(hoistVariableDeclaration);
                                    rightExpression = factory.createAssignment(thisArg, rightExpression);
                                }
                                else {
                                    thisArg = rightExpression;
                                }
                            }
                            rightExpression = segment.kind === 204 /* PropertyAccessExpression */
                                ? factory.createPropertyAccessExpression(rightExpression, ts.visitNode(segment.name, visitor, ts.isIdentifier))
                                : factory.createElementAccessExpression(rightExpression, ts.visitNode(segment.argumentExpression, visitor, ts.isExpression));
                            break;
                        case 206 /* CallExpression */:
                            if (i === 0 && leftThisArg) {
                                rightExpression = factory.createFunctionCallCall(rightExpression, leftThisArg.kind === 106 /* SuperKeyword */ ? factory.createThis() : leftThisArg, ts.visitNodes(segment.arguments, visitor, ts.isExpression));
                            }
                            else {
                                rightExpression = factory.createCallExpression(rightExpression, 
                                /*typeArguments*/ undefined, ts.visitNodes(segment.arguments, visitor, ts.isExpression));
                            }
                            break;
                    }
                    ts.setOriginalNode(rightExpression, segment);
                }
                var target = isDelete
                    ? factory.createConditionalExpression(createNotNullCondition(leftExpression, capturedLeft, /*invert*/ true), /*questionToken*/ undefined, factory.createTrue(), /*colonToken*/ undefined, factory.createDeleteExpression(rightExpression))
                    : factory.createConditionalExpression(createNotNullCondition(leftExpression, capturedLeft, /*invert*/ true), /*questionToken*/ undefined, factory.createVoidZero(), /*colonToken*/ undefined, rightExpression);
                ts.setTextRange(target, node);
                return thisArg ? factory.createSyntheticReferenceExpression(target, thisArg) : target;
            }
            function createNotNullCondition(left, right, invert) {
                return factory.createBinaryExpression(factory.createBinaryExpression(left, factory.createToken(invert ? 36 /* EqualsEqualsEqualsToken */ : 37 /* ExclamationEqualsEqualsToken */), factory.createNull()), factory.createToken(invert ? 56 /* BarBarToken */ : 55 /* AmpersandAmpersandToken */), factory.createBinaryExpression(right, factory.createToken(invert ? 36 /* EqualsEqualsEqualsToken */ : 37 /* ExclamationEqualsEqualsToken */), factory.createVoidZero()));
            }
            function transformNullishCoalescingExpression(node) {
                var left = ts.visitNode(node.left, visitor, ts.isExpression);
                var right = left;
                if (!ts.isSimpleCopiableExpression(left)) {
                    right = factory.createTempVariable(hoistVariableDeclaration);
                    left = factory.createAssignment(right, left);
                }
                return ts.setTextRange(factory.createConditionalExpression(createNotNullCondition(left, right), 
                /*questionToken*/ undefined, right, 
                /*colonToken*/ undefined, ts.visitNode(node.right, visitor, ts.isExpression)), node);
            }
            function visitDeleteExpression(node) {
                return ts.isOptionalChain(ts.skipParentheses(node.expression))
                    ? ts.setOriginalNode(visitNonOptionalExpression(node.expression, /*captureThisArg*/ false, /*isDelete*/ true), node)
                    : factory.updateDeleteExpression(node, ts.visitNode(node.expression, visitor, ts.isExpression));
            }
        }
        ts.transformES2020 = transformES2020;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function transformES2021(context) {
            var hoistVariableDeclaration = context.hoistVariableDeclaration, factory = context.factory;
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitor(node) {
                if ((node.transformFlags & 8 /* ContainsES2021 */) === 0) {
                    return node;
                }
                switch (node.kind) {
                    case 219 /* BinaryExpression */:
                        var binaryExpression = node;
                        if (ts.isLogicalOrCoalescingAssignmentExpression(binaryExpression)) {
                            return transformLogicalAssignment(binaryExpression);
                        }
                    // falls through
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function transformLogicalAssignment(binaryExpression) {
                var operator = binaryExpression.operatorToken;
                var nonAssignmentOperator = ts.getNonAssignmentOperatorForCompoundAssignment(operator.kind);
                var left = ts.skipParentheses(ts.visitNode(binaryExpression.left, visitor, ts.isLeftHandSideExpression));
                var assignmentTarget = left;
                var right = ts.skipParentheses(ts.visitNode(binaryExpression.right, visitor, ts.isExpression));
                if (ts.isAccessExpression(left)) {
                    var propertyAccessTargetSimpleCopiable = ts.isSimpleCopiableExpression(left.expression);
                    var propertyAccessTarget = propertyAccessTargetSimpleCopiable ? left.expression :
                        factory.createTempVariable(hoistVariableDeclaration);
                    var propertyAccessTargetAssignment = propertyAccessTargetSimpleCopiable ? left.expression : factory.createAssignment(propertyAccessTarget, left.expression);
                    if (ts.isPropertyAccessExpression(left)) {
                        assignmentTarget = factory.createPropertyAccessExpression(propertyAccessTarget, left.name);
                        left = factory.createPropertyAccessExpression(propertyAccessTargetAssignment, left.name);
                    }
                    else {
                        var elementAccessArgumentSimpleCopiable = ts.isSimpleCopiableExpression(left.argumentExpression);
                        var elementAccessArgument = elementAccessArgumentSimpleCopiable ? left.argumentExpression :
                            factory.createTempVariable(hoistVariableDeclaration);
                        assignmentTarget = factory.createElementAccessExpression(propertyAccessTarget, elementAccessArgument);
                        left = factory.createElementAccessExpression(propertyAccessTargetAssignment, elementAccessArgumentSimpleCopiable ? left.argumentExpression : factory.createAssignment(elementAccessArgument, left.argumentExpression));
                    }
                }
                return factory.createBinaryExpression(left, nonAssignmentOperator, factory.createParenthesizedExpression(factory.createAssignment(assignmentTarget, right)));
            }
        }
        ts.transformES2021 = transformES2021;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function transformESNext(context) {
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitor(node) {
                if ((node.transformFlags & 4 /* ContainsESNext */) === 0) {
                    return node;
                }
                switch (node.kind) {
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
        }
        ts.transformESNext = transformESNext;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function transformJsx(context) {
            var factory = context.factory, emitHelpers = context.getEmitHelperFactory;
            var compilerOptions = context.getCompilerOptions();
            var currentSourceFile;
            var currentFileState;
            return ts.chainBundle(context, transformSourceFile);
            function getCurrentFileNameExpression() {
                if (currentFileState.filenameDeclaration) {
                    return currentFileState.filenameDeclaration.name;
                }
                var declaration = factory.createVariableDeclaration(factory.createUniqueName("_jsxFileName", 16 /* Optimistic */ | 32 /* FileLevel */), /*exclaimationToken*/ undefined, /*type*/ undefined, factory.createStringLiteral(currentSourceFile.fileName));
                currentFileState.filenameDeclaration = declaration;
                return currentFileState.filenameDeclaration.name;
            }
            function getJsxFactoryCalleePrimitive(childrenLength) {
                return compilerOptions.jsx === 5 /* ReactJSXDev */ ? "jsxDEV" : childrenLength > 1 ? "jsxs" : "jsx";
            }
            function getJsxFactoryCallee(childrenLength) {
                var type = getJsxFactoryCalleePrimitive(childrenLength);
                return getImplicitImportForName(type);
            }
            function getImplicitJsxFragmentReference() {
                return getImplicitImportForName("Fragment");
            }
            function getImplicitImportForName(name) {
                var _a, _b;
                var importSource = name === "createElement"
                    ? currentFileState.importSpecifier
                    : ts.getJSXRuntimeImport(currentFileState.importSpecifier, compilerOptions);
                var existing = (_b = (_a = currentFileState.utilizedImplicitRuntimeImports) === null || _a === void 0 ? void 0 : _a.get(importSource)) === null || _b === void 0 ? void 0 : _b.get(name);
                if (existing) {
                    return existing.name;
                }
                if (!currentFileState.utilizedImplicitRuntimeImports) {
                    currentFileState.utilizedImplicitRuntimeImports = ts.createMap();
                }
                var specifierSourceImports = currentFileState.utilizedImplicitRuntimeImports.get(importSource);
                if (!specifierSourceImports) {
                    specifierSourceImports = ts.createMap();
                    currentFileState.utilizedImplicitRuntimeImports.set(importSource, specifierSourceImports);
                }
                var generatedName = factory.createUniqueName("_" + name, 16 /* Optimistic */ | 32 /* FileLevel */ | 64 /* AllowNameSubstitution */);
                var specifier = factory.createImportSpecifier(factory.createIdentifier(name), generatedName);
                generatedName.generatedImportReference = specifier;
                specifierSourceImports.set(name, specifier);
                return generatedName;
            }
            /**
             * Transform JSX-specific syntax in a SourceFile.
             *
             * @param node A SourceFile node.
             */
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                currentSourceFile = node;
                currentFileState = {};
                currentFileState.importSpecifier = ts.getJSXImplicitImportBase(compilerOptions, node);
                var visited = ts.visitEachChild(node, visitor, context);
                ts.addEmitHelpers(visited, context.readEmitHelpers());
                var statements = visited.statements;
                if (currentFileState.filenameDeclaration) {
                    statements = ts.insertStatementAfterCustomPrologue(statements.slice(), factory.createVariableStatement(/*modifiers*/ undefined, factory.createVariableDeclarationList([currentFileState.filenameDeclaration], 2 /* Const */)));
                }
                if (currentFileState.utilizedImplicitRuntimeImports) {
                    for (var _i = 0, _a = ts.arrayFrom(currentFileState.utilizedImplicitRuntimeImports.entries()); _i < _a.length; _i++) {
                        var _b = _a[_i], importSource = _b[0], importSpecifiersMap = _b[1];
                        if (ts.isExternalModule(node)) {
                            // Add `import` statement
                            var importStatement = factory.createImportDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, factory.createImportClause(/*typeOnly*/ false, /*name*/ undefined, factory.createNamedImports(ts.arrayFrom(importSpecifiersMap.values()))), factory.createStringLiteral(importSource));
                            ts.setParentRecursive(importStatement, /*incremental*/ false);
                            statements = ts.insertStatementAfterCustomPrologue(statements.slice(), importStatement);
                        }
                        else if (ts.isExternalOrCommonJsModule(node)) {
                            // Add `require` statement
                            var requireStatement = factory.createVariableStatement(/*modifiers*/ undefined, factory.createVariableDeclarationList([
                                factory.createVariableDeclaration(factory.createObjectBindingPattern(ts.map(ts.arrayFrom(importSpecifiersMap.values()), function (s) { return factory.createBindingElement(/*dotdotdot*/ undefined, s.propertyName, s.name); })), 
                                /*exclaimationToken*/ undefined, 
                                /*type*/ undefined, factory.createCallExpression(factory.createIdentifier("require"), /*typeArguments*/ undefined, [factory.createStringLiteral(importSource)]))
                            ], 2 /* Const */));
                            ts.setParentRecursive(requireStatement, /*incremental*/ false);
                            statements = ts.insertStatementAfterCustomPrologue(statements.slice(), requireStatement);
                        }
                        else {
                            // Do nothing (script file) - consider an error in the checker?
                        }
                    }
                }
                if (statements !== visited.statements) {
                    visited = factory.updateSourceFile(visited, statements);
                }
                currentFileState = undefined;
                return visited;
            }
            function visitor(node) {
                if (node.transformFlags & 2 /* ContainsJsx */) {
                    return visitorWorker(node);
                }
                else {
                    return node;
                }
            }
            function visitorWorker(node) {
                switch (node.kind) {
                    case 276 /* JsxElement */:
                        return visitJsxElement(node, /*isChild*/ false);
                    case 277 /* JsxSelfClosingElement */:
                        return visitJsxSelfClosingElement(node, /*isChild*/ false);
                    case 280 /* JsxFragment */:
                        return visitJsxFragment(node, /*isChild*/ false);
                    case 286 /* JsxExpression */:
                        return visitJsxExpression(node);
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function transformJsxChildToExpression(node) {
                switch (node.kind) {
                    case 11 /* JsxText */:
                        return visitJsxText(node);
                    case 286 /* JsxExpression */:
                        return visitJsxExpression(node);
                    case 276 /* JsxElement */:
                        return visitJsxElement(node, /*isChild*/ true);
                    case 277 /* JsxSelfClosingElement */:
                        return visitJsxSelfClosingElement(node, /*isChild*/ true);
                    case 280 /* JsxFragment */:
                        return visitJsxFragment(node, /*isChild*/ true);
                    default:
                        return ts.Debug.failBadSyntaxKind(node);
                }
            }
            /**
             * The react jsx/jsxs transform falls back to `createElement` when an explicit `key` argument comes after a spread
             */
            function hasKeyAfterPropsSpread(node) {
                var spread = false;
                for (var _i = 0, _a = node.attributes.properties; _i < _a.length; _i++) {
                    var elem = _a[_i];
                    if (ts.isJsxSpreadAttribute(elem)) {
                        spread = true;
                    }
                    else if (spread && ts.isJsxAttribute(elem) && elem.name.escapedText === "key") {
                        return true;
                    }
                }
                return false;
            }
            function shouldUseCreateElement(node) {
                return currentFileState.importSpecifier === undefined || hasKeyAfterPropsSpread(node);
            }
            function visitJsxElement(node, isChild) {
                var tagTransform = shouldUseCreateElement(node.openingElement) ? visitJsxOpeningLikeElementCreateElement : visitJsxOpeningLikeElementJSX;
                return tagTransform(node.openingElement, node.children, isChild, /*location*/ node);
            }
            function visitJsxSelfClosingElement(node, isChild) {
                var tagTransform = shouldUseCreateElement(node) ? visitJsxOpeningLikeElementCreateElement : visitJsxOpeningLikeElementJSX;
                return tagTransform(node, /*children*/ undefined, isChild, /*location*/ node);
            }
            function visitJsxFragment(node, isChild) {
                var tagTransform = currentFileState.importSpecifier === undefined ? visitJsxOpeningFragmentCreateElement : visitJsxOpeningFragmentJSX;
                return tagTransform(node.openingFragment, node.children, isChild, /*location*/ node);
            }
            function convertJsxChildrenToChildrenPropObject(children) {
                var nonWhitespaceChildren = ts.getSemanticJsxChildren(children);
                if (ts.length(nonWhitespaceChildren) === 1) {
                    var result_13 = transformJsxChildToExpression(nonWhitespaceChildren[0]);
                    return result_13 && factory.createObjectLiteralExpression([
                        factory.createPropertyAssignment("children", result_13)
                    ]);
                }
                var result = ts.mapDefined(children, transformJsxChildToExpression);
                return !result.length ? undefined : factory.createObjectLiteralExpression([
                    factory.createPropertyAssignment("children", factory.createArrayLiteralExpression(result))
                ]);
            }
            function visitJsxOpeningLikeElementJSX(node, children, isChild, location) {
                var tagName = getTagName(node);
                var objectProperties;
                var keyAttr = ts.find(node.attributes.properties, function (p) { return !!p.name && ts.isIdentifier(p.name) && p.name.escapedText === "key"; });
                var attrs = keyAttr ? ts.filter(node.attributes.properties, function (p) { return p !== keyAttr; }) : node.attributes.properties;
                var segments = [];
                if (attrs.length) {
                    // Map spans of JsxAttribute nodes into object literals and spans
                    // of JsxSpreadAttribute nodes into expressions.
                    segments = ts.flatten(ts.spanMap(attrs, ts.isJsxSpreadAttribute, function (attrs, isSpread) {
                        return isSpread
                            ? ts.map(attrs, transformJsxSpreadAttributeToExpression)
                            : factory.createObjectLiteralExpression(ts.map(attrs, transformJsxAttributeToObjectLiteralElement));
                    }));
                    if (ts.isJsxSpreadAttribute(attrs[0])) {
                        // We must always emit at least one object literal before a spread
                        // argument.factory.createObjectLiteral
                        segments.unshift(factory.createObjectLiteralExpression());
                    }
                }
                if (children && children.length) {
                    var result = convertJsxChildrenToChildrenPropObject(children);
                    if (result) {
                        segments.push(result);
                    }
                }
                if (segments.length === 0) {
                    objectProperties = factory.createObjectLiteralExpression([]);
                    // When there are no attributes, React wants {}
                }
                else {
                    // Either emit one big object literal (no spread attribs), or
                    // a call to the __assign helper.
                    objectProperties = ts.singleOrUndefined(segments) || emitHelpers().createAssignHelper(segments);
                }
                return visitJsxOpeningLikeElementOrFragmentJSX(tagName, objectProperties, keyAttr, ts.length(ts.getSemanticJsxChildren(children || ts.emptyArray)), isChild, location);
            }
            function visitJsxOpeningLikeElementOrFragmentJSX(tagName, objectProperties, keyAttr, childrenLength, isChild, location) {
                var args = [tagName, objectProperties, !keyAttr ? factory.createVoidZero() : transformJsxAttributeInitializer(keyAttr.initializer)];
                if (compilerOptions.jsx === 5 /* ReactJSXDev */) {
                    var originalFile = ts.getOriginalNode(currentSourceFile);
                    if (originalFile && ts.isSourceFile(originalFile)) {
                        // isStaticChildren development flag
                        args.push(childrenLength > 1 ? factory.createTrue() : factory.createFalse());
                        // __source development flag
                        var lineCol = ts.getLineAndCharacterOfPosition(originalFile, location.pos);
                        args.push(factory.createObjectLiteralExpression([
                            factory.createPropertyAssignment("fileName", getCurrentFileNameExpression()),
                            factory.createPropertyAssignment("lineNumber", factory.createNumericLiteral(lineCol.line + 1)),
                            factory.createPropertyAssignment("columnNumber", factory.createNumericLiteral(lineCol.character + 1))
                        ]));
                        // __self development flag
                        args.push(factory.createThis());
                    }
                }
                var element = ts.setTextRange(factory.createCallExpression(getJsxFactoryCallee(childrenLength), /*typeArguments*/ undefined, args), location);
                if (isChild) {
                    ts.startOnNewLine(element);
                }
                return element;
            }
            function visitJsxOpeningLikeElementCreateElement(node, children, isChild, location) {
                var tagName = getTagName(node);
                var objectProperties;
                var attrs = node.attributes.properties;
                if (attrs.length === 0) {
                    objectProperties = factory.createNull();
                    // When there are no attributes, React wants "null"
                }
                else {
                    var target = compilerOptions.target;
                    if (target && target >= 5 /* ES2018 */) {
                        objectProperties = factory.createObjectLiteralExpression(ts.flatten(ts.spanMap(attrs, ts.isJsxSpreadAttribute, function (attrs, isSpread) {
                            return isSpread ? ts.map(attrs, transformJsxSpreadAttributeToSpreadAssignment) : ts.map(attrs, transformJsxAttributeToObjectLiteralElement);
                        })));
                    }
                    else {
                        // Map spans of JsxAttribute nodes into object literals and spans
                        // of JsxSpreadAttribute nodes into expressions.
                        var segments = ts.flatten(ts.spanMap(attrs, ts.isJsxSpreadAttribute, function (attrs, isSpread) {
                            return isSpread
                                ? ts.map(attrs, transformJsxSpreadAttributeToExpression)
                                : factory.createObjectLiteralExpression(ts.map(attrs, transformJsxAttributeToObjectLiteralElement));
                        }));
                        if (ts.isJsxSpreadAttribute(attrs[0])) {
                            // We must always emit at least one object literal before a spread
                            // argument.factory.createObjectLiteral
                            segments.unshift(factory.createObjectLiteralExpression());
                        }
                        // Either emit one big object literal (no spread attribs), or
                        // a call to the __assign helper.
                        objectProperties = ts.singleOrUndefined(segments);
                        if (!objectProperties) {
                            objectProperties = emitHelpers().createAssignHelper(segments);
                        }
                    }
                }
                var callee = currentFileState.importSpecifier === undefined
                    ? ts.createJsxFactoryExpression(factory, context.getEmitResolver().getJsxFactoryEntity(currentSourceFile), compilerOptions.reactNamespace, // TODO: GH#18217
                    node)
                    : getImplicitImportForName("createElement");
                var element = ts.createExpressionForJsxElement(factory, callee, tagName, objectProperties, ts.mapDefined(children, transformJsxChildToExpression), location);
                if (isChild) {
                    ts.startOnNewLine(element);
                }
                return element;
            }
            function visitJsxOpeningFragmentJSX(_node, children, isChild, location) {
                var childrenProps;
                if (children && children.length) {
                    var result = convertJsxChildrenToChildrenPropObject(children);
                    if (result) {
                        childrenProps = result;
                    }
                }
                return visitJsxOpeningLikeElementOrFragmentJSX(getImplicitJsxFragmentReference(), childrenProps || factory.createObjectLiteralExpression([]), 
                /*keyAttr*/ undefined, ts.length(ts.getSemanticJsxChildren(children)), isChild, location);
            }
            function visitJsxOpeningFragmentCreateElement(node, children, isChild, location) {
                var element = ts.createExpressionForJsxFragment(factory, context.getEmitResolver().getJsxFactoryEntity(currentSourceFile), context.getEmitResolver().getJsxFragmentFactoryEntity(currentSourceFile), compilerOptions.reactNamespace, // TODO: GH#18217
                ts.mapDefined(children, transformJsxChildToExpression), node, location);
                if (isChild) {
                    ts.startOnNewLine(element);
                }
                return element;
            }
            function transformJsxSpreadAttributeToSpreadAssignment(node) {
                return factory.createSpreadAssignment(ts.visitNode(node.expression, visitor, ts.isExpression));
            }
            function transformJsxSpreadAttributeToExpression(node) {
                return ts.visitNode(node.expression, visitor, ts.isExpression);
            }
            function transformJsxAttributeToObjectLiteralElement(node) {
                var name = getAttributeName(node);
                var expression = transformJsxAttributeInitializer(node.initializer);
                return factory.createPropertyAssignment(name, expression);
            }
            function transformJsxAttributeInitializer(node) {
                if (node === undefined) {
                    return factory.createTrue();
                }
                else if (node.kind === 10 /* StringLiteral */) {
                    // Always recreate the literal to escape any escape sequences or newlines which may be in the original jsx string and which
                    // Need to be escaped to be handled correctly in a normal string
                    var singleQuote = node.singleQuote !== undefined ? node.singleQuote : !ts.isStringDoubleQuoted(node, currentSourceFile);
                    var literal = factory.createStringLiteral(tryDecodeEntities(node.text) || node.text, singleQuote);
                    return ts.setTextRange(literal, node);
                }
                else if (node.kind === 286 /* JsxExpression */) {
                    if (node.expression === undefined) {
                        return factory.createTrue();
                    }
                    return ts.visitNode(node.expression, visitor, ts.isExpression);
                }
                else {
                    return ts.Debug.failBadSyntaxKind(node);
                }
            }
            function visitJsxText(node) {
                var fixed = fixupWhitespaceAndDecodeEntities(node.text);
                return fixed === undefined ? undefined : factory.createStringLiteral(fixed);
            }
            /**
             * JSX trims whitespace at the end and beginning of lines, except that the
             * start/end of a tag is considered a start/end of a line only if that line is
             * on the same line as the closing tag. See examples in
             * tests/cases/conformance/jsx/tsxReactEmitWhitespace.tsx
             * See also https://www.w3.org/TR/html4/struct/text.html#h-9.1 and https://www.w3.org/TR/CSS2/text.html#white-space-model
             *
             * An equivalent algorithm would be:
             * - If there is only one line, return it.
             * - If there is only whitespace (but multiple lines), return `undefined`.
             * - Split the text into lines.
             * - 'trimRight' the first line, 'trimLeft' the last line, 'trim' middle lines.
             * - Decode entities on each line (individually).
             * - Remove empty lines and join the rest with " ".
             */
            function fixupWhitespaceAndDecodeEntities(text) {
                var acc;
                // First non-whitespace character on this line.
                var firstNonWhitespace = 0;
                // Last non-whitespace character on this line.
                var lastNonWhitespace = -1;
                // These initial values are special because the first line is:
                // firstNonWhitespace = 0 to indicate that we want leading whitsepace,
                // but lastNonWhitespace = -1 as a special flag to indicate that we *don't* include the line if it's all whitespace.
                for (var i = 0; i < text.length; i++) {
                    var c = text.charCodeAt(i);
                    if (ts.isLineBreak(c)) {
                        // If we've seen any non-whitespace characters on this line, add the 'trim' of the line.
                        // (lastNonWhitespace === -1 is a special flag to detect whether the first line is all whitespace.)
                        if (firstNonWhitespace !== -1 && lastNonWhitespace !== -1) {
                            acc = addLineOfJsxText(acc, text.substr(firstNonWhitespace, lastNonWhitespace - firstNonWhitespace + 1));
                        }
                        // Reset firstNonWhitespace for the next line.
                        // Don't bother to reset lastNonWhitespace because we ignore it if firstNonWhitespace = -1.
                        firstNonWhitespace = -1;
                    }
                    else if (!ts.isWhiteSpaceSingleLine(c)) {
                        lastNonWhitespace = i;
                        if (firstNonWhitespace === -1) {
                            firstNonWhitespace = i;
                        }
                    }
                }
                return firstNonWhitespace !== -1
                    // Last line had a non-whitespace character. Emit the 'trimLeft', meaning keep trailing whitespace.
                    ? addLineOfJsxText(acc, text.substr(firstNonWhitespace))
                    // Last line was all whitespace, so ignore it
                    : acc;
            }
            function addLineOfJsxText(acc, trimmedLine) {
                // We do not escape the string here as that is handled by the printer
                // when it emits the literal. We do, however, need to decode JSX entities.
                var decoded = decodeEntities(trimmedLine);
                return acc === undefined ? decoded : acc + " " + decoded;
            }
            /**
             * Replace entities like "&nbsp;", "&#123;", and "&#xDEADBEEF;" with the characters they encode.
             * See https://en.wikipedia.org/wiki/List_of_XML_and_HTML_character_entity_references
             */
            function decodeEntities(text) {
                return text.replace(/&((#((\d+)|x([\da-fA-F]+)))|(\w+));/g, function (match, _all, _number, _digits, decimal, hex, word) {
                    if (decimal) {
                        return ts.utf16EncodeAsString(parseInt(decimal, 10));
                    }
                    else if (hex) {
                        return ts.utf16EncodeAsString(parseInt(hex, 16));
                    }
                    else {
                        var ch = entities.get(word);
                        // If this is not a valid entity, then just use `match` (replace it with itself, i.e. don't replace)
                        return ch ? ts.utf16EncodeAsString(ch) : match;
                    }
                });
            }
            /** Like `decodeEntities` but returns `undefined` if there were no entities to decode. */
            function tryDecodeEntities(text) {
                var decoded = decodeEntities(text);
                return decoded === text ? undefined : decoded;
            }
            function getTagName(node) {
                if (node.kind === 276 /* JsxElement */) {
                    return getTagName(node.openingElement);
                }
                else {
                    var name = node.tagName;
                    if (ts.isIdentifier(name) && ts.isIntrinsicJsxName(name.escapedText)) {
                        return factory.createStringLiteral(ts.idText(name));
                    }
                    else {
                        return ts.createExpressionFromEntityName(factory, name);
                    }
                }
            }
            /**
             * Emit an attribute name, which is quoted if it needs to be quoted. Because
             * these emit into an object literal property name, we don't need to be worried
             * about keywords, just non-identifier characters
             */
            function getAttributeName(node) {
                var name = node.name;
                var text = ts.idText(name);
                if (/^[A-Za-z_]\w*$/.test(text)) {
                    return name;
                }
                else {
                    return factory.createStringLiteral(text);
                }
            }
            function visitJsxExpression(node) {
                return ts.visitNode(node.expression, visitor, ts.isExpression);
            }
        }
        ts.transformJsx = transformJsx;
        var entities = new ts.Map(ts.getEntries({
            quot: 0x0022,
            amp: 0x0026,
            apos: 0x0027,
            lt: 0x003C,
            gt: 0x003E,
            nbsp: 0x00A0,
            iexcl: 0x00A1,
            cent: 0x00A2,
            pound: 0x00A3,
            curren: 0x00A4,
            yen: 0x00A5,
            brvbar: 0x00A6,
            sect: 0x00A7,
            uml: 0x00A8,
            copy: 0x00A9,
            ordf: 0x00AA,
            laquo: 0x00AB,
            not: 0x00AC,
            shy: 0x00AD,
            reg: 0x00AE,
            macr: 0x00AF,
            deg: 0x00B0,
            plusmn: 0x00B1,
            sup2: 0x00B2,
            sup3: 0x00B3,
            acute: 0x00B4,
            micro: 0x00B5,
            para: 0x00B6,
            middot: 0x00B7,
            cedil: 0x00B8,
            sup1: 0x00B9,
            ordm: 0x00BA,
            raquo: 0x00BB,
            frac14: 0x00BC,
            frac12: 0x00BD,
            frac34: 0x00BE,
            iquest: 0x00BF,
            Agrave: 0x00C0,
            Aacute: 0x00C1,
            Acirc: 0x00C2,
            Atilde: 0x00C3,
            Auml: 0x00C4,
            Aring: 0x00C5,
            AElig: 0x00C6,
            Ccedil: 0x00C7,
            Egrave: 0x00C8,
            Eacute: 0x00C9,
            Ecirc: 0x00CA,
            Euml: 0x00CB,
            Igrave: 0x00CC,
            Iacute: 0x00CD,
            Icirc: 0x00CE,
            Iuml: 0x00CF,
            ETH: 0x00D0,
            Ntilde: 0x00D1,
            Ograve: 0x00D2,
            Oacute: 0x00D3,
            Ocirc: 0x00D4,
            Otilde: 0x00D5,
            Ouml: 0x00D6,
            times: 0x00D7,
            Oslash: 0x00D8,
            Ugrave: 0x00D9,
            Uacute: 0x00DA,
            Ucirc: 0x00DB,
            Uuml: 0x00DC,
            Yacute: 0x00DD,
            THORN: 0x00DE,
            szlig: 0x00DF,
            agrave: 0x00E0,
            aacute: 0x00E1,
            acirc: 0x00E2,
            atilde: 0x00E3,
            auml: 0x00E4,
            aring: 0x00E5,
            aelig: 0x00E6,
            ccedil: 0x00E7,
            egrave: 0x00E8,
            eacute: 0x00E9,
            ecirc: 0x00EA,
            euml: 0x00EB,
            igrave: 0x00EC,
            iacute: 0x00ED,
            icirc: 0x00EE,
            iuml: 0x00EF,
            eth: 0x00F0,
            ntilde: 0x00F1,
            ograve: 0x00F2,
            oacute: 0x00F3,
            ocirc: 0x00F4,
            otilde: 0x00F5,
            ouml: 0x00F6,
            divide: 0x00F7,
            oslash: 0x00F8,
            ugrave: 0x00F9,
            uacute: 0x00FA,
            ucirc: 0x00FB,
            uuml: 0x00FC,
            yacute: 0x00FD,
            thorn: 0x00FE,
            yuml: 0x00FF,
            OElig: 0x0152,
            oelig: 0x0153,
            Scaron: 0x0160,
            scaron: 0x0161,
            Yuml: 0x0178,
            fnof: 0x0192,
            circ: 0x02C6,
            tilde: 0x02DC,
            Alpha: 0x0391,
            Beta: 0x0392,
            Gamma: 0x0393,
            Delta: 0x0394,
            Epsilon: 0x0395,
            Zeta: 0x0396,
            Eta: 0x0397,
            Theta: 0x0398,
            Iota: 0x0399,
            Kappa: 0x039A,
            Lambda: 0x039B,
            Mu: 0x039C,
            Nu: 0x039D,
            Xi: 0x039E,
            Omicron: 0x039F,
            Pi: 0x03A0,
            Rho: 0x03A1,
            Sigma: 0x03A3,
            Tau: 0x03A4,
            Upsilon: 0x03A5,
            Phi: 0x03A6,
            Chi: 0x03A7,
            Psi: 0x03A8,
            Omega: 0x03A9,
            alpha: 0x03B1,
            beta: 0x03B2,
            gamma: 0x03B3,
            delta: 0x03B4,
            epsilon: 0x03B5,
            zeta: 0x03B6,
            eta: 0x03B7,
            theta: 0x03B8,
            iota: 0x03B9,
            kappa: 0x03BA,
            lambda: 0x03BB,
            mu: 0x03BC,
            nu: 0x03BD,
            xi: 0x03BE,
            omicron: 0x03BF,
            pi: 0x03C0,
            rho: 0x03C1,
            sigmaf: 0x03C2,
            sigma: 0x03C3,
            tau: 0x03C4,
            upsilon: 0x03C5,
            phi: 0x03C6,
            chi: 0x03C7,
            psi: 0x03C8,
            omega: 0x03C9,
            thetasym: 0x03D1,
            upsih: 0x03D2,
            piv: 0x03D6,
            ensp: 0x2002,
            emsp: 0x2003,
            thinsp: 0x2009,
            zwnj: 0x200C,
            zwj: 0x200D,
            lrm: 0x200E,
            rlm: 0x200F,
            ndash: 0x2013,
            mdash: 0x2014,
            lsquo: 0x2018,
            rsquo: 0x2019,
            sbquo: 0x201A,
            ldquo: 0x201C,
            rdquo: 0x201D,
            bdquo: 0x201E,
            dagger: 0x2020,
            Dagger: 0x2021,
            bull: 0x2022,
            hellip: 0x2026,
            permil: 0x2030,
            prime: 0x2032,
            Prime: 0x2033,
            lsaquo: 0x2039,
            rsaquo: 0x203A,
            oline: 0x203E,
            frasl: 0x2044,
            euro: 0x20AC,
            image: 0x2111,
            weierp: 0x2118,
            real: 0x211C,
            trade: 0x2122,
            alefsym: 0x2135,
            larr: 0x2190,
            uarr: 0x2191,
            rarr: 0x2192,
            darr: 0x2193,
            harr: 0x2194,
            crarr: 0x21B5,
            lArr: 0x21D0,
            uArr: 0x21D1,
            rArr: 0x21D2,
            dArr: 0x21D3,
            hArr: 0x21D4,
            forall: 0x2200,
            part: 0x2202,
            exist: 0x2203,
            empty: 0x2205,
            nabla: 0x2207,
            isin: 0x2208,
            notin: 0x2209,
            ni: 0x220B,
            prod: 0x220F,
            sum: 0x2211,
            minus: 0x2212,
            lowast: 0x2217,
            radic: 0x221A,
            prop: 0x221D,
            infin: 0x221E,
            ang: 0x2220,
            and: 0x2227,
            or: 0x2228,
            cap: 0x2229,
            cup: 0x222A,
            int: 0x222B,
            there4: 0x2234,
            sim: 0x223C,
            cong: 0x2245,
            asymp: 0x2248,
            ne: 0x2260,
            equiv: 0x2261,
            le: 0x2264,
            ge: 0x2265,
            sub: 0x2282,
            sup: 0x2283,
            nsub: 0x2284,
            sube: 0x2286,
            supe: 0x2287,
            oplus: 0x2295,
            otimes: 0x2297,
            perp: 0x22A5,
            sdot: 0x22C5,
            lceil: 0x2308,
            rceil: 0x2309,
            lfloor: 0x230A,
            rfloor: 0x230B,
            lang: 0x2329,
            rang: 0x232A,
            loz: 0x25CA,
            spades: 0x2660,
            clubs: 0x2663,
            hearts: 0x2665,
            diams: 0x2666
        }));
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function transformES2016(context) {
            var factory = context.factory, hoistVariableDeclaration = context.hoistVariableDeclaration;
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitor(node) {
                if ((node.transformFlags & 256 /* ContainsES2016 */) === 0) {
                    return node;
                }
                switch (node.kind) {
                    case 219 /* BinaryExpression */:
                        return visitBinaryExpression(node);
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function visitBinaryExpression(node) {
                switch (node.operatorToken.kind) {
                    case 67 /* AsteriskAsteriskEqualsToken */:
                        return visitExponentiationAssignmentExpression(node);
                    case 42 /* AsteriskAsteriskToken */:
                        return visitExponentiationExpression(node);
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function visitExponentiationAssignmentExpression(node) {
                var target;
                var value;
                var left = ts.visitNode(node.left, visitor, ts.isExpression);
                var right = ts.visitNode(node.right, visitor, ts.isExpression);
                if (ts.isElementAccessExpression(left)) {
                    // Transforms `a[x] **= b` into `(_a = a)[_x = x] = Math.pow(_a[_x], b)`
                    var expressionTemp = factory.createTempVariable(hoistVariableDeclaration);
                    var argumentExpressionTemp = factory.createTempVariable(hoistVariableDeclaration);
                    target = ts.setTextRange(factory.createElementAccessExpression(ts.setTextRange(factory.createAssignment(expressionTemp, left.expression), left.expression), ts.setTextRange(factory.createAssignment(argumentExpressionTemp, left.argumentExpression), left.argumentExpression)), left);
                    value = ts.setTextRange(factory.createElementAccessExpression(expressionTemp, argumentExpressionTemp), left);
                }
                else if (ts.isPropertyAccessExpression(left)) {
                    // Transforms `a.x **= b` into `(_a = a).x = Math.pow(_a.x, b)`
                    var expressionTemp = factory.createTempVariable(hoistVariableDeclaration);
                    target = ts.setTextRange(factory.createPropertyAccessExpression(ts.setTextRange(factory.createAssignment(expressionTemp, left.expression), left.expression), left.name), left);
                    value = ts.setTextRange(factory.createPropertyAccessExpression(expressionTemp, left.name), left);
                }
                else {
                    // Transforms `a **= b` into `a = Math.pow(a, b)`
                    target = left;
                    value = left;
                }
                return ts.setTextRange(factory.createAssignment(target, ts.setTextRange(factory.createGlobalMethodCall("Math", "pow", [value, right]), node)), node);
            }
            function visitExponentiationExpression(node) {
                // Transforms `a ** b` into `Math.pow(a, b)`
                var left = ts.visitNode(node.left, visitor, ts.isExpression);
                var right = ts.visitNode(node.right, visitor, ts.isExpression);
                return ts.setTextRange(factory.createGlobalMethodCall("Math", "pow", [left, right]), node);
            }
        }
        ts.transformES2016 = transformES2016;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var ES2015SubstitutionFlags;
        (function (ES2015SubstitutionFlags) {
            /** Enables substitutions for captured `this` */
            ES2015SubstitutionFlags[ES2015SubstitutionFlags["CapturedThis"] = 1] = "CapturedThis";
            /** Enables substitutions for block-scoped bindings. */
            ES2015SubstitutionFlags[ES2015SubstitutionFlags["BlockScopedBindings"] = 2] = "BlockScopedBindings";
        })(ES2015SubstitutionFlags || (ES2015SubstitutionFlags = {}));
        var LoopOutParameterFlags;
        (function (LoopOutParameterFlags) {
            LoopOutParameterFlags[LoopOutParameterFlags["Body"] = 1] = "Body";
            LoopOutParameterFlags[LoopOutParameterFlags["Initializer"] = 2] = "Initializer";
        })(LoopOutParameterFlags || (LoopOutParameterFlags = {}));
        var CopyDirection;
        (function (CopyDirection) {
            CopyDirection[CopyDirection["ToOriginal"] = 0] = "ToOriginal";
            CopyDirection[CopyDirection["ToOutParameter"] = 1] = "ToOutParameter";
        })(CopyDirection || (CopyDirection = {}));
        var Jump;
        (function (Jump) {
            Jump[Jump["Break"] = 2] = "Break";
            Jump[Jump["Continue"] = 4] = "Continue";
            Jump[Jump["Return"] = 8] = "Return";
        })(Jump || (Jump = {}));
        // Facts we track as we traverse the tree
        var HierarchyFacts;
        (function (HierarchyFacts) {
            HierarchyFacts[HierarchyFacts["None"] = 0] = "None";
            //
            // Ancestor facts
            //
            HierarchyFacts[HierarchyFacts["Function"] = 1] = "Function";
            HierarchyFacts[HierarchyFacts["ArrowFunction"] = 2] = "ArrowFunction";
            HierarchyFacts[HierarchyFacts["AsyncFunctionBody"] = 4] = "AsyncFunctionBody";
            HierarchyFacts[HierarchyFacts["NonStaticClassElement"] = 8] = "NonStaticClassElement";
            HierarchyFacts[HierarchyFacts["CapturesThis"] = 16] = "CapturesThis";
            HierarchyFacts[HierarchyFacts["ExportedVariableStatement"] = 32] = "ExportedVariableStatement";
            HierarchyFacts[HierarchyFacts["TopLevel"] = 64] = "TopLevel";
            HierarchyFacts[HierarchyFacts["Block"] = 128] = "Block";
            HierarchyFacts[HierarchyFacts["IterationStatement"] = 256] = "IterationStatement";
            HierarchyFacts[HierarchyFacts["IterationStatementBlock"] = 512] = "IterationStatementBlock";
            HierarchyFacts[HierarchyFacts["IterationContainer"] = 1024] = "IterationContainer";
            HierarchyFacts[HierarchyFacts["ForStatement"] = 2048] = "ForStatement";
            HierarchyFacts[HierarchyFacts["ForInOrForOfStatement"] = 4096] = "ForInOrForOfStatement";
            HierarchyFacts[HierarchyFacts["ConstructorWithCapturedSuper"] = 8192] = "ConstructorWithCapturedSuper";
            HierarchyFacts[HierarchyFacts["StaticInitializer"] = 16384] = "StaticInitializer";
            // NOTE: do not add more ancestor flags without also updating AncestorFactsMask below.
            // NOTE: when adding a new ancestor flag, be sure to update the subtree flags below.
            //
            // Ancestor masks
            //
            HierarchyFacts[HierarchyFacts["AncestorFactsMask"] = 32767] = "AncestorFactsMask";
            // We are always in *some* kind of block scope, but only specific block-scope containers are
            // top-level or Blocks.
            HierarchyFacts[HierarchyFacts["BlockScopeIncludes"] = 0] = "BlockScopeIncludes";
            HierarchyFacts[HierarchyFacts["BlockScopeExcludes"] = 7104] = "BlockScopeExcludes";
            // A source file is a top-level block scope.
            HierarchyFacts[HierarchyFacts["SourceFileIncludes"] = 64] = "SourceFileIncludes";
            HierarchyFacts[HierarchyFacts["SourceFileExcludes"] = 8064] = "SourceFileExcludes";
            // Functions, methods, and accessors are both new lexical scopes and new block scopes.
            HierarchyFacts[HierarchyFacts["FunctionIncludes"] = 65] = "FunctionIncludes";
            HierarchyFacts[HierarchyFacts["FunctionExcludes"] = 32670] = "FunctionExcludes";
            HierarchyFacts[HierarchyFacts["AsyncFunctionBodyIncludes"] = 69] = "AsyncFunctionBodyIncludes";
            HierarchyFacts[HierarchyFacts["AsyncFunctionBodyExcludes"] = 32662] = "AsyncFunctionBodyExcludes";
            // Arrow functions are lexically scoped to their container, but are new block scopes.
            HierarchyFacts[HierarchyFacts["ArrowFunctionIncludes"] = 66] = "ArrowFunctionIncludes";
            HierarchyFacts[HierarchyFacts["ArrowFunctionExcludes"] = 15232] = "ArrowFunctionExcludes";
            // Constructors are both new lexical scopes and new block scopes. Constructors are also
            // always considered non-static members of a class.
            HierarchyFacts[HierarchyFacts["ConstructorIncludes"] = 73] = "ConstructorIncludes";
            HierarchyFacts[HierarchyFacts["ConstructorExcludes"] = 32662] = "ConstructorExcludes";
            // 'do' and 'while' statements are not block scopes. We track that the subtree is contained
            // within an IterationStatement to indicate whether the embedded statement is an
            // IterationStatementBlock.
            HierarchyFacts[HierarchyFacts["DoOrWhileStatementIncludes"] = 1280] = "DoOrWhileStatementIncludes";
            HierarchyFacts[HierarchyFacts["DoOrWhileStatementExcludes"] = 0] = "DoOrWhileStatementExcludes";
            // 'for' statements are new block scopes and have special handling for 'let' declarations.
            HierarchyFacts[HierarchyFacts["ForStatementIncludes"] = 3328] = "ForStatementIncludes";
            HierarchyFacts[HierarchyFacts["ForStatementExcludes"] = 5056] = "ForStatementExcludes";
            // 'for-in' and 'for-of' statements are new block scopes and have special handling for
            // 'let' declarations.
            HierarchyFacts[HierarchyFacts["ForInOrForOfStatementIncludes"] = 5376] = "ForInOrForOfStatementIncludes";
            HierarchyFacts[HierarchyFacts["ForInOrForOfStatementExcludes"] = 3008] = "ForInOrForOfStatementExcludes";
            // Blocks (other than function bodies) are new block scopes.
            HierarchyFacts[HierarchyFacts["BlockIncludes"] = 128] = "BlockIncludes";
            HierarchyFacts[HierarchyFacts["BlockExcludes"] = 6976] = "BlockExcludes";
            HierarchyFacts[HierarchyFacts["IterationStatementBlockIncludes"] = 512] = "IterationStatementBlockIncludes";
            HierarchyFacts[HierarchyFacts["IterationStatementBlockExcludes"] = 7104] = "IterationStatementBlockExcludes";
            HierarchyFacts[HierarchyFacts["StaticInitializerIncludes"] = 16449] = "StaticInitializerIncludes";
            HierarchyFacts[HierarchyFacts["StaticInitializerExcludes"] = 32670] = "StaticInitializerExcludes";
            //
            // Subtree facts
            //
            HierarchyFacts[HierarchyFacts["NewTarget"] = 32768] = "NewTarget";
            HierarchyFacts[HierarchyFacts["CapturedLexicalThis"] = 65536] = "CapturedLexicalThis";
            //
            // Subtree masks
            //
            HierarchyFacts[HierarchyFacts["SubtreeFactsMask"] = -32768] = "SubtreeFactsMask";
            HierarchyFacts[HierarchyFacts["ArrowFunctionSubtreeExcludes"] = 0] = "ArrowFunctionSubtreeExcludes";
            HierarchyFacts[HierarchyFacts["FunctionSubtreeExcludes"] = 98304] = "FunctionSubtreeExcludes";
        })(HierarchyFacts || (HierarchyFacts = {}));
        var SpreadSegmentKind;
        (function (SpreadSegmentKind) {
            SpreadSegmentKind[SpreadSegmentKind["None"] = 0] = "None";
            SpreadSegmentKind[SpreadSegmentKind["UnpackedSpread"] = 1] = "UnpackedSpread";
            SpreadSegmentKind[SpreadSegmentKind["PackedSpread"] = 2] = "PackedSpread";
        })(SpreadSegmentKind || (SpreadSegmentKind = {}));
        function createSpreadSegment(kind, expression) {
            return { kind: kind, expression: expression };
        }
        function transformES2015(context) {
            var factory = context.factory, emitHelpers = context.getEmitHelperFactory, startLexicalEnvironment = context.startLexicalEnvironment, resumeLexicalEnvironment = context.resumeLexicalEnvironment, endLexicalEnvironment = context.endLexicalEnvironment, hoistVariableDeclaration = context.hoistVariableDeclaration;
            var compilerOptions = context.getCompilerOptions();
            var resolver = context.getEmitResolver();
            var previousOnSubstituteNode = context.onSubstituteNode;
            var previousOnEmitNode = context.onEmitNode;
            context.onEmitNode = onEmitNode;
            context.onSubstituteNode = onSubstituteNode;
            var currentSourceFile;
            var currentText;
            var hierarchyFacts;
            var taggedTemplateStringDeclarations;
            function recordTaggedTemplateString(temp) {
                taggedTemplateStringDeclarations = ts.append(taggedTemplateStringDeclarations, factory.createVariableDeclaration(temp));
            }
            /**
             * Used to track if we are emitting body of the converted loop
             */
            var convertedLoopState;
            /**
             * Keeps track of whether substitutions have been enabled for specific cases.
             * They are persisted between each SourceFile transformation and should not
             * be reset.
             */
            var enabledSubstitutions;
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                currentSourceFile = node;
                currentText = node.text;
                var visited = visitSourceFile(node);
                ts.addEmitHelpers(visited, context.readEmitHelpers());
                currentSourceFile = undefined;
                currentText = undefined;
                taggedTemplateStringDeclarations = undefined;
                hierarchyFacts = 0 /* None */;
                return visited;
            }
            /**
             * Sets the `HierarchyFacts` for this node prior to visiting this node's subtree, returning the facts set prior to modification.
             * @param excludeFacts The existing `HierarchyFacts` to reset before visiting the subtree.
             * @param includeFacts The new `HierarchyFacts` to set before visiting the subtree.
             */
            function enterSubtree(excludeFacts, includeFacts) {
                var ancestorFacts = hierarchyFacts;
                hierarchyFacts = (hierarchyFacts & ~excludeFacts | includeFacts) & 32767 /* AncestorFactsMask */;
                return ancestorFacts;
            }
            /**
             * Restores the `HierarchyFacts` for this node's ancestor after visiting this node's
             * subtree, propagating specific facts from the subtree.
             * @param ancestorFacts The `HierarchyFacts` of the ancestor to restore after visiting the subtree.
             * @param excludeFacts The existing `HierarchyFacts` of the subtree that should not be propagated.
             * @param includeFacts The new `HierarchyFacts` of the subtree that should be propagated.
             */
            function exitSubtree(ancestorFacts, excludeFacts, includeFacts) {
                hierarchyFacts = (hierarchyFacts & ~excludeFacts | includeFacts) & -32768 /* SubtreeFactsMask */ | ancestorFacts;
            }
            function isReturnVoidStatementInConstructorWithCapturedSuper(node) {
                return (hierarchyFacts & 8192 /* ConstructorWithCapturedSuper */) !== 0
                    && node.kind === 245 /* ReturnStatement */
                    && !node.expression;
            }
            function isOrMayContainReturnCompletion(node) {
                return node.transformFlags & 2097152 /* ContainsHoistedDeclarationOrCompletion */
                    && (ts.isReturnStatement(node)
                        || ts.isIfStatement(node)
                        || ts.isWithStatement(node)
                        || ts.isSwitchStatement(node)
                        || ts.isCaseBlock(node)
                        || ts.isCaseClause(node)
                        || ts.isDefaultClause(node)
                        || ts.isTryStatement(node)
                        || ts.isCatchClause(node)
                        || ts.isLabeledStatement(node)
                        || ts.isIterationStatement(node, /*lookInLabeledStatements*/ false)
                        || ts.isBlock(node));
            }
            function shouldVisitNode(node) {
                return (node.transformFlags & 512 /* ContainsES2015 */) !== 0
                    || convertedLoopState !== undefined
                    || (hierarchyFacts & 8192 /* ConstructorWithCapturedSuper */ && isOrMayContainReturnCompletion(node))
                    || (ts.isIterationStatement(node, /*lookInLabeledStatements*/ false) && shouldConvertIterationStatement(node))
                    || (ts.getEmitFlags(node) & 33554432 /* TypeScriptClassWrapper */) !== 0;
            }
            function visitor(node) {
                return shouldVisitNode(node) ? visitorWorker(node, /*expressionResultIsUnused*/ false) : node;
            }
            function visitorWithUnusedExpressionResult(node) {
                return shouldVisitNode(node) ? visitorWorker(node, /*expressionResultIsUnused*/ true) : node;
            }
            function classWrapperStatementVisitor(node) {
                if (shouldVisitNode(node)) {
                    var original = ts.getOriginalNode(node);
                    if (ts.isPropertyDeclaration(original) && ts.hasStaticModifier(original)) {
                        var ancestorFacts = enterSubtree(32670 /* StaticInitializerExcludes */, 16449 /* StaticInitializerIncludes */);
                        var result = visitorWorker(node, /*expressionResultIsUnused*/ false);
                        exitSubtree(ancestorFacts, 98304 /* FunctionSubtreeExcludes */, 0 /* None */);
                        return result;
                    }
                    return visitorWorker(node, /*expressionResultIsUnused*/ false);
                }
                return node;
            }
            function callExpressionVisitor(node) {
                if (node.kind === 106 /* SuperKeyword */) {
                    return visitSuperKeyword(/*isExpressionOfCall*/ true);
                }
                return visitor(node);
            }
            function visitorWorker(node, expressionResultIsUnused) {
                switch (node.kind) {
                    case 124 /* StaticKeyword */:
                        return undefined; // elide static keyword
                    case 255 /* ClassDeclaration */:
                        return visitClassDeclaration(node);
                    case 224 /* ClassExpression */:
                        return visitClassExpression(node);
                    case 162 /* Parameter */:
                        return visitParameter(node);
                    case 254 /* FunctionDeclaration */:
                        return visitFunctionDeclaration(node);
                    case 212 /* ArrowFunction */:
                        return visitArrowFunction(node);
                    case 211 /* FunctionExpression */:
                        return visitFunctionExpression(node);
                    case 252 /* VariableDeclaration */:
                        return visitVariableDeclaration(node);
                    case 79 /* Identifier */:
                        return visitIdentifier(node);
                    case 253 /* VariableDeclarationList */:
                        return visitVariableDeclarationList(node);
                    case 247 /* SwitchStatement */:
                        return visitSwitchStatement(node);
                    case 261 /* CaseBlock */:
                        return visitCaseBlock(node);
                    case 233 /* Block */:
                        return visitBlock(node, /*isFunctionBody*/ false);
                    case 244 /* BreakStatement */:
                    case 243 /* ContinueStatement */:
                        return visitBreakOrContinueStatement(node);
                    case 248 /* LabeledStatement */:
                        return visitLabeledStatement(node);
                    case 238 /* DoStatement */:
                    case 239 /* WhileStatement */:
                        return visitDoOrWhileStatement(node, /*outermostLabeledStatement*/ undefined);
                    case 240 /* ForStatement */:
                        return visitForStatement(node, /*outermostLabeledStatement*/ undefined);
                    case 241 /* ForInStatement */:
                        return visitForInStatement(node, /*outermostLabeledStatement*/ undefined);
                    case 242 /* ForOfStatement */:
                        return visitForOfStatement(node, /*outermostLabeledStatement*/ undefined);
                    case 236 /* ExpressionStatement */:
                        return visitExpressionStatement(node);
                    case 203 /* ObjectLiteralExpression */:
                        return visitObjectLiteralExpression(node);
                    case 290 /* CatchClause */:
                        return visitCatchClause(node);
                    case 292 /* ShorthandPropertyAssignment */:
                        return visitShorthandPropertyAssignment(node);
                    case 160 /* ComputedPropertyName */:
                        return visitComputedPropertyName(node);
                    case 202 /* ArrayLiteralExpression */:
                        return visitArrayLiteralExpression(node);
                    case 206 /* CallExpression */:
                        return visitCallExpression(node);
                    case 207 /* NewExpression */:
                        return visitNewExpression(node);
                    case 210 /* ParenthesizedExpression */:
                        return visitParenthesizedExpression(node, expressionResultIsUnused);
                    case 219 /* BinaryExpression */:
                        return visitBinaryExpression(node, expressionResultIsUnused);
                    case 346 /* CommaListExpression */:
                        return visitCommaListExpression(node, expressionResultIsUnused);
                    case 14 /* NoSubstitutionTemplateLiteral */:
                    case 15 /* TemplateHead */:
                    case 16 /* TemplateMiddle */:
                    case 17 /* TemplateTail */:
                        return visitTemplateLiteral(node);
                    case 10 /* StringLiteral */:
                        return visitStringLiteral(node);
                    case 8 /* NumericLiteral */:
                        return visitNumericLiteral(node);
                    case 208 /* TaggedTemplateExpression */:
                        return visitTaggedTemplateExpression(node);
                    case 221 /* TemplateExpression */:
                        return visitTemplateExpression(node);
                    case 222 /* YieldExpression */:
                        return visitYieldExpression(node);
                    case 223 /* SpreadElement */:
                        return visitSpreadElement(node);
                    case 106 /* SuperKeyword */:
                        return visitSuperKeyword(/*isExpressionOfCall*/ false);
                    case 108 /* ThisKeyword */:
                        return visitThisKeyword(node);
                    case 229 /* MetaProperty */:
                        return visitMetaProperty(node);
                    case 167 /* MethodDeclaration */:
                        return visitMethodDeclaration(node);
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return visitAccessorDeclaration(node);
                    case 235 /* VariableStatement */:
                        return visitVariableStatement(node);
                    case 245 /* ReturnStatement */:
                        return visitReturnStatement(node);
                    case 215 /* VoidExpression */:
                        return visitVoidExpression(node);
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            function visitSourceFile(node) {
                var ancestorFacts = enterSubtree(8064 /* SourceFileExcludes */, 64 /* SourceFileIncludes */);
                var prologue = [];
                var statements = [];
                startLexicalEnvironment();
                var statementOffset = factory.copyPrologue(node.statements, prologue, /*ensureUseStrict*/ false, visitor);
                ts.addRange(statements, ts.visitNodes(node.statements, visitor, ts.isStatement, statementOffset));
                if (taggedTemplateStringDeclarations) {
                    statements.push(factory.createVariableStatement(/*modifiers*/ undefined, factory.createVariableDeclarationList(taggedTemplateStringDeclarations)));
                }
                factory.mergeLexicalEnvironment(prologue, endLexicalEnvironment());
                insertCaptureThisForNodeIfNeeded(prologue, node);
                exitSubtree(ancestorFacts, 0 /* None */, 0 /* None */);
                return factory.updateSourceFile(node, ts.setTextRange(factory.createNodeArray(ts.concatenate(prologue, statements)), node.statements));
            }
            function visitSwitchStatement(node) {
                if (convertedLoopState !== undefined) {
                    var savedAllowedNonLabeledJumps = convertedLoopState.allowedNonLabeledJumps;
                    // for switch statement allow only non-labeled break
                    convertedLoopState.allowedNonLabeledJumps |= 2 /* Break */;
                    var result = ts.visitEachChild(node, visitor, context);
                    convertedLoopState.allowedNonLabeledJumps = savedAllowedNonLabeledJumps;
                    return result;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitCaseBlock(node) {
                var ancestorFacts = enterSubtree(7104 /* BlockScopeExcludes */, 0 /* BlockScopeIncludes */);
                var updated = ts.visitEachChild(node, visitor, context);
                exitSubtree(ancestorFacts, 0 /* None */, 0 /* None */);
                return updated;
            }
            function returnCapturedThis(node) {
                return ts.setOriginalNode(factory.createReturnStatement(factory.createUniqueName("_this", 16 /* Optimistic */ | 32 /* FileLevel */)), node);
            }
            function visitReturnStatement(node) {
                if (convertedLoopState) {
                    convertedLoopState.nonLocalJumps |= 8 /* Return */;
                    if (isReturnVoidStatementInConstructorWithCapturedSuper(node)) {
                        node = returnCapturedThis(node);
                    }
                    return factory.createReturnStatement(factory.createObjectLiteralExpression([
                        factory.createPropertyAssignment(factory.createIdentifier("value"), node.expression
                            ? ts.visitNode(node.expression, visitor, ts.isExpression)
                            : factory.createVoidZero())
                    ]));
                }
                else if (isReturnVoidStatementInConstructorWithCapturedSuper(node)) {
                    return returnCapturedThis(node);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitThisKeyword(node) {
                if (hierarchyFacts & 2 /* ArrowFunction */ && !(hierarchyFacts & 16384 /* StaticInitializer */)) {
                    hierarchyFacts |= 65536 /* CapturedLexicalThis */;
                }
                if (convertedLoopState) {
                    if (hierarchyFacts & 2 /* ArrowFunction */) {
                        // if the enclosing function is an ArrowFunction then we use the captured 'this' keyword.
                        convertedLoopState.containsLexicalThis = true;
                        return node;
                    }
                    return convertedLoopState.thisName || (convertedLoopState.thisName = factory.createUniqueName("this"));
                }
                return node;
            }
            function visitVoidExpression(node) {
                return ts.visitEachChild(node, visitorWithUnusedExpressionResult, context);
            }
            function visitIdentifier(node) {
                if (!convertedLoopState) {
                    return node;
                }
                if (resolver.isArgumentsLocalBinding(node)) {
                    return convertedLoopState.argumentsName || (convertedLoopState.argumentsName = factory.createUniqueName("arguments"));
                }
                return node;
            }
            function visitBreakOrContinueStatement(node) {
                if (convertedLoopState) {
                    // check if we can emit break/continue as is
                    // it is possible if either
                    //   - break/continue is labeled and label is located inside the converted loop
                    //   - break/continue is non-labeled and located in non-converted loop/switch statement
                    var jump = node.kind === 244 /* BreakStatement */ ? 2 /* Break */ : 4 /* Continue */;
                    var canUseBreakOrContinue = (node.label && convertedLoopState.labels && convertedLoopState.labels.get(ts.idText(node.label))) ||
                        (!node.label && (convertedLoopState.allowedNonLabeledJumps & jump));
                    if (!canUseBreakOrContinue) {
                        var labelMarker = void 0;
                        var label = node.label;
                        if (!label) {
                            if (node.kind === 244 /* BreakStatement */) {
                                convertedLoopState.nonLocalJumps |= 2 /* Break */;
                                labelMarker = "break";
                            }
                            else {
                                convertedLoopState.nonLocalJumps |= 4 /* Continue */;
                                // note: return value is emitted only to simplify debugging, call to converted loop body does not do any dispatching on it.
                                labelMarker = "continue";
                            }
                        }
                        else {
                            if (node.kind === 244 /* BreakStatement */) {
                                labelMarker = "break-" + label.escapedText;
                                setLabeledJump(convertedLoopState, /*isBreak*/ true, ts.idText(label), labelMarker);
                            }
                            else {
                                labelMarker = "continue-" + label.escapedText;
                                setLabeledJump(convertedLoopState, /*isBreak*/ false, ts.idText(label), labelMarker);
                            }
                        }
                        var returnExpression = factory.createStringLiteral(labelMarker);
                        if (convertedLoopState.loopOutParameters.length) {
                            var outParams = convertedLoopState.loopOutParameters;
                            var expr = void 0;
                            for (var i = 0; i < outParams.length; i++) {
                                var copyExpr = copyOutParameter(outParams[i], 1 /* ToOutParameter */);
                                if (i === 0) {
                                    expr = copyExpr;
                                }
                                else {
                                    expr = factory.createBinaryExpression(expr, 27 /* CommaToken */, copyExpr);
                                }
                            }
                            returnExpression = factory.createBinaryExpression(expr, 27 /* CommaToken */, returnExpression);
                        }
                        return factory.createReturnStatement(returnExpression);
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visits a ClassDeclaration and transforms it into a variable statement.
             *
             * @param node A ClassDeclaration node.
             */
            function visitClassDeclaration(node) {
                // [source]
                //      class C { }
                //
                // [output]
                //      var C = (function () {
                //          function C() {
                //          }
                //          return C;
                //      }());
                var variable = factory.createVariableDeclaration(factory.getLocalName(node, /*allowComments*/ true), 
                /*exclamationToken*/ undefined, 
                /*type*/ undefined, transformClassLikeDeclarationToExpression(node));
                ts.setOriginalNode(variable, node);
                var statements = [];
                var statement = factory.createVariableStatement(/*modifiers*/ undefined, factory.createVariableDeclarationList([variable]));
                ts.setOriginalNode(statement, node);
                ts.setTextRange(statement, node);
                ts.startOnNewLine(statement);
                statements.push(statement);
                // Add an `export default` statement for default exports (for `--target es5 --module es6`)
                if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                    var exportStatement = ts.hasSyntacticModifier(node, 512 /* Default */)
                        ? factory.createExportDefault(factory.getLocalName(node))
                        : factory.createExternalModuleExport(factory.getLocalName(node));
                    ts.setOriginalNode(exportStatement, statement);
                    statements.push(exportStatement);
                }
                var emitFlags = ts.getEmitFlags(node);
                if ((emitFlags & 4194304 /* HasEndOfDeclarationMarker */) === 0) {
                    // Add a DeclarationMarker as a marker for the end of the declaration
                    statements.push(factory.createEndOfDeclarationMarker(node));
                    ts.setEmitFlags(statement, emitFlags | 4194304 /* HasEndOfDeclarationMarker */);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Visits a ClassExpression and transforms it into an expression.
             *
             * @param node A ClassExpression node.
             */
            function visitClassExpression(node) {
                // [source]
                //      C = class { }
                //
                // [output]
                //      C = (function () {
                //          function class_1() {
                //          }
                //          return class_1;
                //      }())
                return transformClassLikeDeclarationToExpression(node);
            }
            /**
             * Transforms a ClassExpression or ClassDeclaration into an expression.
             *
             * @param node A ClassExpression or ClassDeclaration node.
             */
            function transformClassLikeDeclarationToExpression(node) {
                // [source]
                //      class C extends D {
                //          constructor() {}
                //          method() {}
                //          get prop() {}
                //          set prop(v) {}
                //      }
                //
                // [output]
                //      (function (_super) {
                //          __extends(C, _super);
                //          function C() {
                //          }
                //          C.prototype.method = function () {}
                //          Object.defineProperty(C.prototype, "prop", {
                //              get: function() {},
                //              set: function() {},
                //              enumerable: true,
                //              configurable: true
                //          });
                //          return C;
                //      }(D))
                if (node.name) {
                    enableSubstitutionsForBlockScopedBindings();
                }
                var extendsClauseElement = ts.getClassExtendsHeritageElement(node);
                var classFunction = factory.createFunctionExpression(
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, extendsClauseElement ? [factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, factory.createUniqueName("_super", 16 /* Optimistic */ | 32 /* FileLevel */))] : [], 
                /*type*/ undefined, transformClassBody(node, extendsClauseElement));
                // To preserve the behavior of the old emitter, we explicitly indent
                // the body of the function here if it was requested in an earlier
                // transformation.
                ts.setEmitFlags(classFunction, (ts.getEmitFlags(node) & 65536 /* Indented */) | 524288 /* ReuseTempVariableScope */);
                // "inner" and "outer" below are added purely to preserve source map locations from
                // the old emitter
                var inner = factory.createPartiallyEmittedExpression(classFunction);
                ts.setTextRangeEnd(inner, node.end);
                ts.setEmitFlags(inner, 1536 /* NoComments */);
                var outer = factory.createPartiallyEmittedExpression(inner);
                ts.setTextRangeEnd(outer, ts.skipTrivia(currentText, node.pos));
                ts.setEmitFlags(outer, 1536 /* NoComments */);
                var result = factory.createParenthesizedExpression(factory.createCallExpression(outer, 
                /*typeArguments*/ undefined, extendsClauseElement
                    ? [ts.visitNode(extendsClauseElement.expression, visitor, ts.isExpression)]
                    : []));
                ts.addSyntheticLeadingComment(result, 3 /* MultiLineCommentTrivia */, "* @class ");
                return result;
            }
            /**
             * Transforms a ClassExpression or ClassDeclaration into a function body.
             *
             * @param node A ClassExpression or ClassDeclaration node.
             * @param extendsClauseElement The expression for the class `extends` clause.
             */
            function transformClassBody(node, extendsClauseElement) {
                var statements = [];
                var name = factory.getInternalName(node);
                var constructorLikeName = ts.isIdentifierANonContextualKeyword(name) ? factory.getGeneratedNameForNode(name) : name;
                startLexicalEnvironment();
                addExtendsHelperIfNeeded(statements, node, extendsClauseElement);
                addConstructor(statements, node, constructorLikeName, extendsClauseElement);
                addClassMembers(statements, node);
                // Create a synthetic text range for the return statement.
                var closingBraceLocation = ts.createTokenRange(ts.skipTrivia(currentText, node.members.end), 19 /* CloseBraceToken */);
                // The following partially-emitted expression exists purely to align our sourcemap
                // emit with the original emitter.
                var outer = factory.createPartiallyEmittedExpression(constructorLikeName);
                ts.setTextRangeEnd(outer, closingBraceLocation.end);
                ts.setEmitFlags(outer, 1536 /* NoComments */);
                var statement = factory.createReturnStatement(outer);
                ts.setTextRangePos(statement, closingBraceLocation.pos);
                ts.setEmitFlags(statement, 1536 /* NoComments */ | 384 /* NoTokenSourceMaps */);
                statements.push(statement);
                ts.insertStatementsAfterStandardPrologue(statements, endLexicalEnvironment());
                var block = factory.createBlock(ts.setTextRange(factory.createNodeArray(statements), /*location*/ node.members), /*multiLine*/ true);
                ts.setEmitFlags(block, 1536 /* NoComments */);
                return block;
            }
            /**
             * Adds a call to the `__extends` helper if needed for a class.
             *
             * @param statements The statements of the class body function.
             * @param node The ClassExpression or ClassDeclaration node.
             * @param extendsClauseElement The expression for the class `extends` clause.
             */
            function addExtendsHelperIfNeeded(statements, node, extendsClauseElement) {
                if (extendsClauseElement) {
                    statements.push(ts.setTextRange(factory.createExpressionStatement(emitHelpers().createExtendsHelper(factory.getInternalName(node))), 
                    /*location*/ extendsClauseElement));
                }
            }
            /**
             * Adds the constructor of the class to a class body function.
             *
             * @param statements The statements of the class body function.
             * @param node The ClassExpression or ClassDeclaration node.
             * @param extendsClauseElement The expression for the class `extends` clause.
             */
            function addConstructor(statements, node, name, extendsClauseElement) {
                var savedConvertedLoopState = convertedLoopState;
                convertedLoopState = undefined;
                var ancestorFacts = enterSubtree(32662 /* ConstructorExcludes */, 73 /* ConstructorIncludes */);
                var constructor = ts.getFirstConstructorWithBody(node);
                var hasSynthesizedSuper = hasSynthesizedDefaultSuperCall(constructor, extendsClauseElement !== undefined);
                var constructorFunction = factory.createFunctionDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, name, 
                /*typeParameters*/ undefined, transformConstructorParameters(constructor, hasSynthesizedSuper), 
                /*type*/ undefined, transformConstructorBody(constructor, node, extendsClauseElement, hasSynthesizedSuper));
                ts.setTextRange(constructorFunction, constructor || node);
                if (extendsClauseElement) {
                    ts.setEmitFlags(constructorFunction, 8 /* CapturesThis */);
                }
                statements.push(constructorFunction);
                exitSubtree(ancestorFacts, 98304 /* FunctionSubtreeExcludes */, 0 /* None */);
                convertedLoopState = savedConvertedLoopState;
            }
            /**
             * Transforms the parameters of the constructor declaration of a class.
             *
             * @param constructor The constructor for the class.
             * @param hasSynthesizedSuper A value indicating whether the constructor starts with a
             *                            synthesized `super` call.
             */
            function transformConstructorParameters(constructor, hasSynthesizedSuper) {
                // If the TypeScript transformer needed to synthesize a constructor for property
                // initializers, it would have also added a synthetic `...args` parameter and
                // `super` call.
                // If this is the case, we do not include the synthetic `...args` parameter and
                // will instead use the `arguments` object in ES5/3.
                return ts.visitParameterList(constructor && !hasSynthesizedSuper ? constructor.parameters : undefined, visitor, context)
                    || [];
            }
            function createDefaultConstructorBody(node, isDerivedClass) {
                // We must be here because the user didn't write a constructor
                // but we needed to call 'super(...args)' anyway as per 14.5.14 of the ES2016 spec.
                // If that's the case we can just immediately return the result of a 'super()' call.
                var statements = [];
                resumeLexicalEnvironment();
                factory.mergeLexicalEnvironment(statements, endLexicalEnvironment());
                if (isDerivedClass) {
                    // return _super !== null && _super.apply(this, arguments) || this;
                    statements.push(factory.createReturnStatement(createDefaultSuperCallOrThis()));
                }
                var statementsArray = factory.createNodeArray(statements);
                ts.setTextRange(statementsArray, node.members);
                var block = factory.createBlock(statementsArray, /*multiLine*/ true);
                ts.setTextRange(block, node);
                ts.setEmitFlags(block, 1536 /* NoComments */);
                return block;
            }
            /**
             * Transforms the body of a constructor declaration of a class.
             *
             * @param constructor The constructor for the class.
             * @param node The node which contains the constructor.
             * @param extendsClauseElement The expression for the class `extends` clause.
             * @param hasSynthesizedSuper A value indicating whether the constructor starts with a
             *                            synthesized `super` call.
             */
            function transformConstructorBody(constructor, node, extendsClauseElement, hasSynthesizedSuper) {
                // determine whether the class is known syntactically to be a derived class (e.g. a
                // class that extends a value that is not syntactically known to be `null`).
                var isDerivedClass = !!extendsClauseElement && ts.skipOuterExpressions(extendsClauseElement.expression).kind !== 104 /* NullKeyword */;
                // When the subclass does not have a constructor, we synthesize a *default* constructor using the following
                // representation:
                //
                // ```
                // // es2015 (source)
                // class C extends Base { }
                //
                // // es5 (transformed)
                // var C = (function (_super) {
                //     function C() {
                //         return _super.apply(this, arguments) || this;
                //     }
                //     return C;
                // })(Base);
                // ```
                if (!constructor)
                    return createDefaultConstructorBody(node, isDerivedClass);
                // The prologue will contain all leading standard and custom prologue statements added by this transform
                var prologue = [];
                var statements = [];
                resumeLexicalEnvironment();
                // If a super call has already been synthesized,
                // we're going to assume that we should just transform everything after that.
                // The assumption is that no prior step in the pipeline has added any prologue directives.
                var statementOffset = 0;
                if (!hasSynthesizedSuper)
                    statementOffset = factory.copyStandardPrologue(constructor.body.statements, prologue, /*ensureUseStrict*/ false);
                addDefaultValueAssignmentsIfNeeded(statements, constructor);
                addRestParameterIfNeeded(statements, constructor, hasSynthesizedSuper);
                if (!hasSynthesizedSuper)
                    statementOffset = factory.copyCustomPrologue(constructor.body.statements, statements, statementOffset, visitor);
                // If the first statement is a call to `super()`, visit the statement directly
                var superCallExpression;
                if (hasSynthesizedSuper) {
                    superCallExpression = createDefaultSuperCallOrThis();
                }
                else if (isDerivedClass && statementOffset < constructor.body.statements.length) {
                    var firstStatement = constructor.body.statements[statementOffset];
                    if (ts.isExpressionStatement(firstStatement) && ts.isSuperCall(firstStatement.expression)) {
                        superCallExpression = visitImmediateSuperCallInBody(firstStatement.expression);
                    }
                }
                if (superCallExpression) {
                    hierarchyFacts |= 8192 /* ConstructorWithCapturedSuper */;
                    statementOffset++; // skip this statement, we will add it after visiting the rest of the body.
                }
                // visit the remaining statements
                ts.addRange(statements, ts.visitNodes(constructor.body.statements, visitor, ts.isStatement, /*start*/ statementOffset));
                factory.mergeLexicalEnvironment(prologue, endLexicalEnvironment());
                insertCaptureNewTargetIfNeeded(prologue, constructor, /*copyOnWrite*/ false);
                if (isDerivedClass) {
                    if (superCallExpression && statementOffset === constructor.body.statements.length && !(constructor.body.transformFlags & 8192 /* ContainsLexicalThis */)) {
                        // If the subclass constructor does *not* contain `this` and *ends* with a `super()` call, we will use the
                        // following representation:
                        //
                        // ```
                        // // es2015 (source)
                        // class C extends Base {
                        //     constructor() {
                        //         super("foo");
                        //     }
                        // }
                        //
                        // // es5 (transformed)
                        // var C = (function (_super) {
                        //     function C() {
                        //         return _super.call(this, "foo") || this;
                        //     }
                        //     return C;
                        // })(Base);
                        // ```
                        var superCall = ts.cast(ts.cast(superCallExpression, ts.isBinaryExpression).left, ts.isCallExpression);
                        var returnStatement = factory.createReturnStatement(superCallExpression);
                        ts.setCommentRange(returnStatement, ts.getCommentRange(superCall));
                        ts.setEmitFlags(superCall, 1536 /* NoComments */);
                        statements.push(returnStatement);
                    }
                    else {
                        // Otherwise, we will use the following transformed representation for calls to `super()` in a constructor:
                        //
                        // ```
                        // // es2015 (source)
                        // class C extends Base {
                        //     constructor() {
                        //         super("foo");
                        //         this.x = 1;
                        //     }
                        // }
                        //
                        // // es5 (transformed)
                        // var C = (function (_super) {
                        //     function C() {
                        //         var _this = _super.call(this, "foo") || this;
                        //         _this.x = 1;
                        //         return _this;
                        //     }
                        //     return C;
                        // })(Base);
                        // ```
                        // Since the `super()` call was the first statement, we insert the `this` capturing call to
                        // `super()` at the top of the list of `statements` (after any pre-existing custom prologues).
                        insertCaptureThisForNode(statements, constructor, superCallExpression || createActualThis());
                        if (!isSufficientlyCoveredByReturnStatements(constructor.body)) {
                            statements.push(factory.createReturnStatement(factory.createUniqueName("_this", 16 /* Optimistic */ | 32 /* FileLevel */)));
                        }
                    }
                }
                else {
                    // If a class is not derived from a base class or does not have a call to `super()`, `this` is only
                    // captured when necessitated by an arrow function capturing the lexical `this`:
                    //
                    // ```
                    // // es2015
                    // class C {}
                    //
                    // // es5
                    // var C = (function () {
                    //     function C() {
                    //     }
                    //     return C;
                    // })();
                    // ```
                    insertCaptureThisForNodeIfNeeded(prologue, constructor);
                }
                var block = factory.createBlock(ts.setTextRange(factory.createNodeArray(ts.concatenate(prologue, statements)), 
                /*location*/ constructor.body.statements), 
                /*multiLine*/ true);
                ts.setTextRange(block, constructor.body);
                return block;
            }
            /**
             * We want to try to avoid emitting a return statement in certain cases if a user already returned something.
             * It would generate obviously dead code, so we'll try to make things a little bit prettier
             * by doing a minimal check on whether some common patterns always explicitly return.
             */
            function isSufficientlyCoveredByReturnStatements(statement) {
                // A return statement is considered covered.
                if (statement.kind === 245 /* ReturnStatement */) {
                    return true;
                }
                // An if-statement with two covered branches is covered.
                else if (statement.kind === 237 /* IfStatement */) {
                    var ifStatement = statement;
                    if (ifStatement.elseStatement) {
                        return isSufficientlyCoveredByReturnStatements(ifStatement.thenStatement) &&
                            isSufficientlyCoveredByReturnStatements(ifStatement.elseStatement);
                    }
                }
                // A block is covered if it has a last statement which is covered.
                else if (statement.kind === 233 /* Block */) {
                    var lastStatement = ts.lastOrUndefined(statement.statements);
                    if (lastStatement && isSufficientlyCoveredByReturnStatements(lastStatement)) {
                        return true;
                    }
                }
                return false;
            }
            function createActualThis() {
                return ts.setEmitFlags(factory.createThis(), 4 /* NoSubstitution */);
            }
            function createDefaultSuperCallOrThis() {
                return factory.createLogicalOr(factory.createLogicalAnd(factory.createStrictInequality(factory.createUniqueName("_super", 16 /* Optimistic */ | 32 /* FileLevel */), factory.createNull()), factory.createFunctionApplyCall(factory.createUniqueName("_super", 16 /* Optimistic */ | 32 /* FileLevel */), createActualThis(), factory.createIdentifier("arguments"))), createActualThis());
            }
            /**
             * Visits a parameter declaration.
             *
             * @param node A ParameterDeclaration node.
             */
            function visitParameter(node) {
                if (node.dotDotDotToken) {
                    // rest parameters are elided
                    return undefined;
                }
                else if (ts.isBindingPattern(node.name)) {
                    // Binding patterns are converted into a generated name and are
                    // evaluated inside the function body.
                    return ts.setOriginalNode(ts.setTextRange(factory.createParameterDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, 
                    /*dotDotDotToken*/ undefined, factory.getGeneratedNameForNode(node), 
                    /*questionToken*/ undefined, 
                    /*type*/ undefined, 
                    /*initializer*/ undefined), 
                    /*location*/ node), 
                    /*original*/ node);
                }
                else if (node.initializer) {
                    // Initializers are elided
                    return ts.setOriginalNode(ts.setTextRange(factory.createParameterDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, 
                    /*dotDotDotToken*/ undefined, node.name, 
                    /*questionToken*/ undefined, 
                    /*type*/ undefined, 
                    /*initializer*/ undefined), 
                    /*location*/ node), 
                    /*original*/ node);
                }
                else {
                    return node;
                }
            }
            function hasDefaultValueOrBindingPattern(node) {
                return node.initializer !== undefined
                    || ts.isBindingPattern(node.name);
            }
            /**
             * Adds statements to the body of a function-like node if it contains parameters with
             * binding patterns or initializers.
             *
             * @param statements The statements for the new function body.
             * @param node A function-like node.
             */
            function addDefaultValueAssignmentsIfNeeded(statements, node) {
                if (!ts.some(node.parameters, hasDefaultValueOrBindingPattern)) {
                    return false;
                }
                var added = false;
                for (var _i = 0, _a = node.parameters; _i < _a.length; _i++) {
                    var parameter = _a[_i];
                    var name = parameter.name, initializer = parameter.initializer, dotDotDotToken = parameter.dotDotDotToken;
                    // A rest parameter cannot have a binding pattern or an initializer,
                    // so let's just ignore it.
                    if (dotDotDotToken) {
                        continue;
                    }
                    if (ts.isBindingPattern(name)) {
                        added = insertDefaultValueAssignmentForBindingPattern(statements, parameter, name, initializer) || added;
                    }
                    else if (initializer) {
                        insertDefaultValueAssignmentForInitializer(statements, parameter, name, initializer);
                        added = true;
                    }
                }
                return added;
            }
            /**
             * Adds statements to the body of a function-like node for parameters with binding patterns
             *
             * @param statements The statements for the new function body.
             * @param parameter The parameter for the function.
             * @param name The name of the parameter.
             * @param initializer The initializer for the parameter.
             */
            function insertDefaultValueAssignmentForBindingPattern(statements, parameter, name, initializer) {
                // In cases where a binding pattern is simply '[]' or '{}',
                // we usually don't want to emit a var declaration; however, in the presence
                // of an initializer, we must emit that expression to preserve side effects.
                if (name.elements.length > 0) {
                    ts.insertStatementAfterCustomPrologue(statements, ts.setEmitFlags(factory.createVariableStatement(
                    /*modifiers*/ undefined, factory.createVariableDeclarationList(ts.flattenDestructuringBinding(parameter, visitor, context, 0 /* All */, factory.getGeneratedNameForNode(parameter)))), 1048576 /* CustomPrologue */));
                    return true;
                }
                else if (initializer) {
                    ts.insertStatementAfterCustomPrologue(statements, ts.setEmitFlags(factory.createExpressionStatement(factory.createAssignment(factory.getGeneratedNameForNode(parameter), ts.visitNode(initializer, visitor, ts.isExpression))), 1048576 /* CustomPrologue */));
                    return true;
                }
                return false;
            }
            /**
             * Adds statements to the body of a function-like node for parameters with initializers.
             *
             * @param statements The statements for the new function body.
             * @param parameter The parameter for the function.
             * @param name The name of the parameter.
             * @param initializer The initializer for the parameter.
             */
            function insertDefaultValueAssignmentForInitializer(statements, parameter, name, initializer) {
                initializer = ts.visitNode(initializer, visitor, ts.isExpression);
                var statement = factory.createIfStatement(factory.createTypeCheck(factory.cloneNode(name), "undefined"), ts.setEmitFlags(ts.setTextRange(factory.createBlock([
                    factory.createExpressionStatement(ts.setEmitFlags(ts.setTextRange(factory.createAssignment(
                    // TODO(rbuckton): Does this need to be parented?
                    ts.setEmitFlags(ts.setParent(ts.setTextRange(factory.cloneNode(name), name), name.parent), 48 /* NoSourceMap */), ts.setEmitFlags(initializer, 48 /* NoSourceMap */ | ts.getEmitFlags(initializer) | 1536 /* NoComments */)), parameter), 1536 /* NoComments */))
                ]), parameter), 1 /* SingleLine */ | 32 /* NoTrailingSourceMap */ | 384 /* NoTokenSourceMaps */ | 1536 /* NoComments */));
                ts.startOnNewLine(statement);
                ts.setTextRange(statement, parameter);
                ts.setEmitFlags(statement, 384 /* NoTokenSourceMaps */ | 32 /* NoTrailingSourceMap */ | 1048576 /* CustomPrologue */ | 1536 /* NoComments */);
                ts.insertStatementAfterCustomPrologue(statements, statement);
            }
            /**
             * Gets a value indicating whether we need to add statements to handle a rest parameter.
             *
             * @param node A ParameterDeclaration node.
             * @param inConstructorWithSynthesizedSuper A value indicating whether the parameter is
             *                                          part of a constructor declaration with a
             *                                          synthesized call to `super`
             */
            function shouldAddRestParameter(node, inConstructorWithSynthesizedSuper) {
                return !!(node && node.dotDotDotToken && !inConstructorWithSynthesizedSuper);
            }
            /**
             * Adds statements to the body of a function-like node if it contains a rest parameter.
             *
             * @param statements The statements for the new function body.
             * @param node A function-like node.
             * @param inConstructorWithSynthesizedSuper A value indicating whether the parameter is
             *                                          part of a constructor declaration with a
             *                                          synthesized call to `super`
             */
            function addRestParameterIfNeeded(statements, node, inConstructorWithSynthesizedSuper) {
                var prologueStatements = [];
                var parameter = ts.lastOrUndefined(node.parameters);
                if (!shouldAddRestParameter(parameter, inConstructorWithSynthesizedSuper)) {
                    return false;
                }
                // `declarationName` is the name of the local declaration for the parameter.
                // TODO(rbuckton): Does this need to be parented?
                var declarationName = parameter.name.kind === 79 /* Identifier */ ? ts.setParent(ts.setTextRange(factory.cloneNode(parameter.name), parameter.name), parameter.name.parent) : factory.createTempVariable(/*recordTempVariable*/ undefined);
                ts.setEmitFlags(declarationName, 48 /* NoSourceMap */);
                // `expressionName` is the name of the parameter used in expressions.
                var expressionName = parameter.name.kind === 79 /* Identifier */ ? factory.cloneNode(parameter.name) : declarationName;
                var restIndex = node.parameters.length - 1;
                var temp = factory.createLoopVariable();
                // var param = [];
                prologueStatements.push(ts.setEmitFlags(ts.setTextRange(factory.createVariableStatement(
                /*modifiers*/ undefined, factory.createVariableDeclarationList([
                    factory.createVariableDeclaration(declarationName, 
                    /*exclamationToken*/ undefined, 
                    /*type*/ undefined, factory.createArrayLiteralExpression([]))
                ])), 
                /*location*/ parameter), 1048576 /* CustomPrologue */));
                // for (var _i = restIndex; _i < arguments.length; _i++) {
                //   param[_i - restIndex] = arguments[_i];
                // }
                var forStatement = factory.createForStatement(ts.setTextRange(factory.createVariableDeclarationList([
                    factory.createVariableDeclaration(temp, /*exclamationToken*/ undefined, /*type*/ undefined, factory.createNumericLiteral(restIndex))
                ]), parameter), ts.setTextRange(factory.createLessThan(temp, factory.createPropertyAccessExpression(factory.createIdentifier("arguments"), "length")), parameter), ts.setTextRange(factory.createPostfixIncrement(temp), parameter), factory.createBlock([
                    ts.startOnNewLine(ts.setTextRange(factory.createExpressionStatement(factory.createAssignment(factory.createElementAccessExpression(expressionName, restIndex === 0
                        ? temp
                        : factory.createSubtract(temp, factory.createNumericLiteral(restIndex))), factory.createElementAccessExpression(factory.createIdentifier("arguments"), temp))), 
                    /*location*/ parameter))
                ]));
                ts.setEmitFlags(forStatement, 1048576 /* CustomPrologue */);
                ts.startOnNewLine(forStatement);
                prologueStatements.push(forStatement);
                if (parameter.name.kind !== 79 /* Identifier */) {
                    // do the actual destructuring of the rest parameter if necessary
                    prologueStatements.push(ts.setEmitFlags(ts.setTextRange(factory.createVariableStatement(
                    /*modifiers*/ undefined, factory.createVariableDeclarationList(ts.flattenDestructuringBinding(parameter, visitor, context, 0 /* All */, expressionName))), parameter), 1048576 /* CustomPrologue */));
                }
                ts.insertStatementsAfterCustomPrologue(statements, prologueStatements);
                return true;
            }
            /**
             * Adds a statement to capture the `this` of a function declaration if it is needed.
             * NOTE: This must be executed *after* the subtree has been visited.
             *
             * @param statements The statements for the new function body.
             * @param node A node.
             */
            function insertCaptureThisForNodeIfNeeded(statements, node) {
                if (hierarchyFacts & 65536 /* CapturedLexicalThis */ && node.kind !== 212 /* ArrowFunction */) {
                    insertCaptureThisForNode(statements, node, factory.createThis());
                    return true;
                }
                return false;
            }
            function insertCaptureThisForNode(statements, node, initializer) {
                enableSubstitutionsForCapturedThis();
                var captureThisStatement = factory.createVariableStatement(
                /*modifiers*/ undefined, factory.createVariableDeclarationList([
                    factory.createVariableDeclaration(factory.createUniqueName("_this", 16 /* Optimistic */ | 32 /* FileLevel */), 
                    /*exclamationToken*/ undefined, 
                    /*type*/ undefined, initializer)
                ]));
                ts.setEmitFlags(captureThisStatement, 1536 /* NoComments */ | 1048576 /* CustomPrologue */);
                ts.setSourceMapRange(captureThisStatement, node);
                ts.insertStatementAfterCustomPrologue(statements, captureThisStatement);
            }
            function insertCaptureNewTargetIfNeeded(statements, node, copyOnWrite) {
                if (hierarchyFacts & 32768 /* NewTarget */) {
                    var newTarget = void 0;
                    switch (node.kind) {
                        case 212 /* ArrowFunction */:
                            return statements;
                        case 167 /* MethodDeclaration */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            // Methods and accessors cannot be constructors, so 'new.target' will
                            // always return 'undefined'.
                            newTarget = factory.createVoidZero();
                            break;
                        case 169 /* Constructor */:
                            // Class constructors can only be called with `new`, so `this.constructor`
                            // should be relatively safe to use.
                            newTarget = factory.createPropertyAccessExpression(ts.setEmitFlags(factory.createThis(), 4 /* NoSubstitution */), "constructor");
                            break;
                        case 254 /* FunctionDeclaration */:
                        case 211 /* FunctionExpression */:
                            // Functions can be called or constructed, and may have a `this` due to
                            // being a member or when calling an imported function via `other_1.f()`.
                            newTarget = factory.createConditionalExpression(factory.createLogicalAnd(ts.setEmitFlags(factory.createThis(), 4 /* NoSubstitution */), factory.createBinaryExpression(ts.setEmitFlags(factory.createThis(), 4 /* NoSubstitution */), 102 /* InstanceOfKeyword */, factory.getLocalName(node))), 
                            /*questionToken*/ undefined, factory.createPropertyAccessExpression(ts.setEmitFlags(factory.createThis(), 4 /* NoSubstitution */), "constructor"), 
                            /*colonToken*/ undefined, factory.createVoidZero());
                            break;
                        default:
                            return ts.Debug.failBadSyntaxKind(node);
                    }
                    var captureNewTargetStatement = factory.createVariableStatement(
                    /*modifiers*/ undefined, factory.createVariableDeclarationList([
                        factory.createVariableDeclaration(factory.createUniqueName("_newTarget", 16 /* Optimistic */ | 32 /* FileLevel */), 
                        /*exclamationToken*/ undefined, 
                        /*type*/ undefined, newTarget)
                    ]));
                    ts.setEmitFlags(captureNewTargetStatement, 1536 /* NoComments */ | 1048576 /* CustomPrologue */);
                    if (copyOnWrite) {
                        statements = statements.slice();
                    }
                    ts.insertStatementAfterCustomPrologue(statements, captureNewTargetStatement);
                }
                return statements;
            }
            /**
             * Adds statements to the class body function for a class to define the members of the
             * class.
             *
             * @param statements The statements for the class body function.
             * @param node The ClassExpression or ClassDeclaration node.
             */
            function addClassMembers(statements, node) {
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    switch (member.kind) {
                        case 232 /* SemicolonClassElement */:
                            statements.push(transformSemicolonClassElementToStatement(member));
                            break;
                        case 167 /* MethodDeclaration */:
                            statements.push(transformClassMethodDeclarationToStatement(getClassMemberPrefix(node, member), member, node));
                            break;
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            var accessors = ts.getAllAccessorDeclarations(node.members, member);
                            if (member === accessors.firstAccessor) {
                                statements.push(transformAccessorsToStatement(getClassMemberPrefix(node, member), accessors, node));
                            }
                            break;
                        case 169 /* Constructor */:
                        case 168 /* ClassStaticBlockDeclaration */:
                            // Constructors are handled in visitClassExpression/visitClassDeclaration
                            break;
                        default:
                            ts.Debug.failBadSyntaxKind(member, currentSourceFile && currentSourceFile.fileName);
                            break;
                    }
                }
            }
            /**
             * Transforms a SemicolonClassElement into a statement for a class body function.
             *
             * @param member The SemicolonClassElement node.
             */
            function transformSemicolonClassElementToStatement(member) {
                return ts.setTextRange(factory.createEmptyStatement(), member);
            }
            /**
             * Transforms a MethodDeclaration into a statement for a class body function.
             *
             * @param receiver The receiver for the member.
             * @param member The MethodDeclaration node.
             */
            function transformClassMethodDeclarationToStatement(receiver, member, container) {
                var commentRange = ts.getCommentRange(member);
                var sourceMapRange = ts.getSourceMapRange(member);
                var memberFunction = transformFunctionLikeToExpression(member, /*location*/ member, /*name*/ undefined, container);
                var propertyName = ts.visitNode(member.name, visitor, ts.isPropertyName);
                var e;
                if (!ts.isPrivateIdentifier(propertyName) && ts.getUseDefineForClassFields(context.getCompilerOptions())) {
                    var name = ts.isComputedPropertyName(propertyName) ? propertyName.expression
                        : ts.isIdentifier(propertyName) ? factory.createStringLiteral(ts.unescapeLeadingUnderscores(propertyName.escapedText))
                            : propertyName;
                    e = factory.createObjectDefinePropertyCall(receiver, name, factory.createPropertyDescriptor({ value: memberFunction, enumerable: false, writable: true, configurable: true }));
                }
                else {
                    var memberName = ts.createMemberAccessForPropertyName(factory, receiver, propertyName, /*location*/ member.name);
                    e = factory.createAssignment(memberName, memberFunction);
                }
                ts.setEmitFlags(memberFunction, 1536 /* NoComments */);
                ts.setSourceMapRange(memberFunction, sourceMapRange);
                var statement = ts.setTextRange(factory.createExpressionStatement(e), /*location*/ member);
                ts.setOriginalNode(statement, member);
                ts.setCommentRange(statement, commentRange);
                // The location for the statement is used to emit comments only.
                // No source map should be emitted for this statement to align with the
                // old emitter.
                ts.setEmitFlags(statement, 48 /* NoSourceMap */);
                return statement;
            }
            /**
             * Transforms a set of related of get/set accessors into a statement for a class body function.
             *
             * @param receiver The receiver for the member.
             * @param accessors The set of related get/set accessors.
             */
            function transformAccessorsToStatement(receiver, accessors, container) {
                var statement = factory.createExpressionStatement(transformAccessorsToExpression(receiver, accessors, container, /*startsOnNewLine*/ false));
                // The location for the statement is used to emit source maps only.
                // No comments should be emitted for this statement to align with the
                // old emitter.
                ts.setEmitFlags(statement, 1536 /* NoComments */);
                ts.setSourceMapRange(statement, ts.getSourceMapRange(accessors.firstAccessor));
                return statement;
            }
            /**
             * Transforms a set of related get/set accessors into an expression for either a class
             * body function or an ObjectLiteralExpression with computed properties.
             *
             * @param receiver The receiver for the member.
             */
            function transformAccessorsToExpression(receiver, _a, container, startsOnNewLine) {
                var firstAccessor = _a.firstAccessor, getAccessor = _a.getAccessor, setAccessor = _a.setAccessor;
                // To align with source maps in the old emitter, the receiver and property name
                // arguments are both mapped contiguously to the accessor name.
                // TODO(rbuckton): Does this need to be parented?
                var target = ts.setParent(ts.setTextRange(factory.cloneNode(receiver), receiver), receiver.parent);
                ts.setEmitFlags(target, 1536 /* NoComments */ | 32 /* NoTrailingSourceMap */);
                ts.setSourceMapRange(target, firstAccessor.name);
                var visitedAccessorName = ts.visitNode(firstAccessor.name, visitor, ts.isPropertyName);
                if (ts.isPrivateIdentifier(visitedAccessorName)) {
                    return ts.Debug.failBadSyntaxKind(visitedAccessorName, "Encountered unhandled private identifier while transforming ES2015.");
                }
                var propertyName = ts.createExpressionForPropertyName(factory, visitedAccessorName);
                ts.setEmitFlags(propertyName, 1536 /* NoComments */ | 16 /* NoLeadingSourceMap */);
                ts.setSourceMapRange(propertyName, firstAccessor.name);
                var properties = [];
                if (getAccessor) {
                    var getterFunction = transformFunctionLikeToExpression(getAccessor, /*location*/ undefined, /*name*/ undefined, container);
                    ts.setSourceMapRange(getterFunction, ts.getSourceMapRange(getAccessor));
                    ts.setEmitFlags(getterFunction, 512 /* NoLeadingComments */);
                    var getter = factory.createPropertyAssignment("get", getterFunction);
                    ts.setCommentRange(getter, ts.getCommentRange(getAccessor));
                    properties.push(getter);
                }
                if (setAccessor) {
                    var setterFunction = transformFunctionLikeToExpression(setAccessor, /*location*/ undefined, /*name*/ undefined, container);
                    ts.setSourceMapRange(setterFunction, ts.getSourceMapRange(setAccessor));
                    ts.setEmitFlags(setterFunction, 512 /* NoLeadingComments */);
                    var setter = factory.createPropertyAssignment("set", setterFunction);
                    ts.setCommentRange(setter, ts.getCommentRange(setAccessor));
                    properties.push(setter);
                }
                properties.push(factory.createPropertyAssignment("enumerable", getAccessor || setAccessor ? factory.createFalse() : factory.createTrue()), factory.createPropertyAssignment("configurable", factory.createTrue()));
                var call = factory.createCallExpression(factory.createPropertyAccessExpression(factory.createIdentifier("Object"), "defineProperty"), 
                /*typeArguments*/ undefined, [
                    target,
                    propertyName,
                    factory.createObjectLiteralExpression(properties, /*multiLine*/ true)
                ]);
                if (startsOnNewLine) {
                    ts.startOnNewLine(call);
                }
                return call;
            }
            /**
             * Visits an ArrowFunction and transforms it into a FunctionExpression.
             *
             * @param node An ArrowFunction node.
             */
            function visitArrowFunction(node) {
                if (node.transformFlags & 8192 /* ContainsLexicalThis */ && !(hierarchyFacts & 16384 /* StaticInitializer */)) {
                    hierarchyFacts |= 65536 /* CapturedLexicalThis */;
                }
                var savedConvertedLoopState = convertedLoopState;
                convertedLoopState = undefined;
                var ancestorFacts = enterSubtree(15232 /* ArrowFunctionExcludes */, 66 /* ArrowFunctionIncludes */);
                var func = factory.createFunctionExpression(
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                /*type*/ undefined, transformFunctionBody(node));
                ts.setTextRange(func, node);
                ts.setOriginalNode(func, node);
                ts.setEmitFlags(func, 8 /* CapturesThis */);
                // If an arrow function contains
                exitSubtree(ancestorFacts, 0 /* ArrowFunctionSubtreeExcludes */, 0 /* None */);
                convertedLoopState = savedConvertedLoopState;
                return func;
            }
            /**
             * Visits a FunctionExpression node.
             *
             * @param node a FunctionExpression node.
             */
            function visitFunctionExpression(node) {
                var ancestorFacts = ts.getEmitFlags(node) & 262144 /* AsyncFunctionBody */
                    ? enterSubtree(32662 /* AsyncFunctionBodyExcludes */, 69 /* AsyncFunctionBodyIncludes */)
                    : enterSubtree(32670 /* FunctionExcludes */, 65 /* FunctionIncludes */);
                var savedConvertedLoopState = convertedLoopState;
                convertedLoopState = undefined;
                var parameters = ts.visitParameterList(node.parameters, visitor, context);
                var body = transformFunctionBody(node);
                var name = hierarchyFacts & 32768 /* NewTarget */
                    ? factory.getLocalName(node)
                    : node.name;
                exitSubtree(ancestorFacts, 98304 /* FunctionSubtreeExcludes */, 0 /* None */);
                convertedLoopState = savedConvertedLoopState;
                return factory.updateFunctionExpression(node, 
                /*modifiers*/ undefined, node.asteriskToken, name, 
                /*typeParameters*/ undefined, parameters, 
                /*type*/ undefined, body);
            }
            /**
             * Visits a FunctionDeclaration node.
             *
             * @param node a FunctionDeclaration node.
             */
            function visitFunctionDeclaration(node) {
                var savedConvertedLoopState = convertedLoopState;
                convertedLoopState = undefined;
                var ancestorFacts = enterSubtree(32670 /* FunctionExcludes */, 65 /* FunctionIncludes */);
                var parameters = ts.visitParameterList(node.parameters, visitor, context);
                var body = transformFunctionBody(node);
                var name = hierarchyFacts & 32768 /* NewTarget */
                    ? factory.getLocalName(node)
                    : node.name;
                exitSubtree(ancestorFacts, 98304 /* FunctionSubtreeExcludes */, 0 /* None */);
                convertedLoopState = savedConvertedLoopState;
                return factory.updateFunctionDeclaration(node, 
                /*decorators*/ undefined, ts.visitNodes(node.modifiers, visitor, ts.isModifier), node.asteriskToken, name, 
                /*typeParameters*/ undefined, parameters, 
                /*type*/ undefined, body);
            }
            /**
             * Transforms a function-like node into a FunctionExpression.
             *
             * @param node The function-like node to transform.
             * @param location The source-map location for the new FunctionExpression.
             * @param name The name of the new FunctionExpression.
             */
            function transformFunctionLikeToExpression(node, location, name, container) {
                var savedConvertedLoopState = convertedLoopState;
                convertedLoopState = undefined;
                var ancestorFacts = container && ts.isClassLike(container) && !ts.isStatic(node)
                    ? enterSubtree(32670 /* FunctionExcludes */, 65 /* FunctionIncludes */ | 8 /* NonStaticClassElement */)
                    : enterSubtree(32670 /* FunctionExcludes */, 65 /* FunctionIncludes */);
                var parameters = ts.visitParameterList(node.parameters, visitor, context);
                var body = transformFunctionBody(node);
                if (hierarchyFacts & 32768 /* NewTarget */ && !name && (node.kind === 254 /* FunctionDeclaration */ || node.kind === 211 /* FunctionExpression */)) {
                    name = factory.getGeneratedNameForNode(node);
                }
                exitSubtree(ancestorFacts, 98304 /* FunctionSubtreeExcludes */, 0 /* None */);
                convertedLoopState = savedConvertedLoopState;
                return ts.setOriginalNode(ts.setTextRange(factory.createFunctionExpression(
                /*modifiers*/ undefined, node.asteriskToken, name, 
                /*typeParameters*/ undefined, parameters, 
                /*type*/ undefined, body), location), 
                /*original*/ node);
            }
            /**
             * Transforms the body of a function-like node.
             *
             * @param node A function-like node.
             */
            function transformFunctionBody(node) {
                var multiLine = false; // indicates whether the block *must* be emitted as multiple lines
                var singleLine = false; // indicates whether the block *may* be emitted as a single line
                var statementsLocation;
                var closeBraceLocation;
                var prologue = [];
                var statements = [];
                var body = node.body;
                var statementOffset;
                resumeLexicalEnvironment();
                if (ts.isBlock(body)) {
                    // ensureUseStrict is false because no new prologue-directive should be added.
                    // addStandardPrologue will put already-existing directives at the beginning of the target statement-array
                    statementOffset = factory.copyStandardPrologue(body.statements, prologue, /*ensureUseStrict*/ false);
                    statementOffset = factory.copyCustomPrologue(body.statements, statements, statementOffset, visitor, ts.isHoistedFunction);
                    statementOffset = factory.copyCustomPrologue(body.statements, statements, statementOffset, visitor, ts.isHoistedVariableStatement);
                }
                multiLine = addDefaultValueAssignmentsIfNeeded(statements, node) || multiLine;
                multiLine = addRestParameterIfNeeded(statements, node, /*inConstructorWithSynthesizedSuper*/ false) || multiLine;
                if (ts.isBlock(body)) {
                    // addCustomPrologue puts already-existing directives at the beginning of the target statement-array
                    statementOffset = factory.copyCustomPrologue(body.statements, statements, statementOffset, visitor);
                    statementsLocation = body.statements;
                    ts.addRange(statements, ts.visitNodes(body.statements, visitor, ts.isStatement, statementOffset));
                    // If the original body was a multi-line block, this must be a multi-line block.
                    if (!multiLine && body.multiLine) {
                        multiLine = true;
                    }
                }
                else {
                    ts.Debug.assert(node.kind === 212 /* ArrowFunction */);
                    // To align with the old emitter, we use a synthetic end position on the location
                    // for the statement list we synthesize when we down-level an arrow function with
                    // an expression function body. This prevents both comments and source maps from
                    // being emitted for the end position only.
                    statementsLocation = ts.moveRangeEnd(body, -1);
                    var equalsGreaterThanToken = node.equalsGreaterThanToken;
                    if (!ts.nodeIsSynthesized(equalsGreaterThanToken) && !ts.nodeIsSynthesized(body)) {
                        if (ts.rangeEndIsOnSameLineAsRangeStart(equalsGreaterThanToken, body, currentSourceFile)) {
                            singleLine = true;
                        }
                        else {
                            multiLine = true;
                        }
                    }
                    var expression = ts.visitNode(body, visitor, ts.isExpression);
                    var returnStatement = factory.createReturnStatement(expression);
                    ts.setTextRange(returnStatement, body);
                    ts.moveSyntheticComments(returnStatement, body);
                    ts.setEmitFlags(returnStatement, 384 /* NoTokenSourceMaps */ | 32 /* NoTrailingSourceMap */ | 1024 /* NoTrailingComments */);
                    statements.push(returnStatement);
                    // To align with the source map emit for the old emitter, we set a custom
                    // source map location for the close brace.
                    closeBraceLocation = body;
                }
                factory.mergeLexicalEnvironment(prologue, endLexicalEnvironment());
                insertCaptureNewTargetIfNeeded(prologue, node, /*copyOnWrite*/ false);
                insertCaptureThisForNodeIfNeeded(prologue, node);
                // If we added any final generated statements, this must be a multi-line block
                if (ts.some(prologue)) {
                    multiLine = true;
                }
                statements.unshift.apply(statements, prologue);
                if (ts.isBlock(body) && ts.arrayIsEqualTo(statements, body.statements)) {
                    // no changes were made, preserve the tree
                    return body;
                }
                var block = factory.createBlock(ts.setTextRange(factory.createNodeArray(statements), statementsLocation), multiLine);
                ts.setTextRange(block, node.body);
                if (!multiLine && singleLine) {
                    ts.setEmitFlags(block, 1 /* SingleLine */);
                }
                if (closeBraceLocation) {
                    ts.setTokenSourceMapRange(block, 19 /* CloseBraceToken */, closeBraceLocation);
                }
                ts.setOriginalNode(block, node.body);
                return block;
            }
            function visitBlock(node, isFunctionBody) {
                if (isFunctionBody) {
                    // A function body is not a block scope.
                    return ts.visitEachChild(node, visitor, context);
                }
                var ancestorFacts = hierarchyFacts & 256 /* IterationStatement */
                    ? enterSubtree(7104 /* IterationStatementBlockExcludes */, 512 /* IterationStatementBlockIncludes */)
                    : enterSubtree(6976 /* BlockExcludes */, 128 /* BlockIncludes */);
                var updated = ts.visitEachChild(node, visitor, context);
                exitSubtree(ancestorFacts, 0 /* None */, 0 /* None */);
                return updated;
            }
            /**
             * Visits an ExpressionStatement that contains a destructuring assignment.
             *
             * @param node An ExpressionStatement node.
             */
            function visitExpressionStatement(node) {
                return ts.visitEachChild(node, visitorWithUnusedExpressionResult, context);
            }
            /**
             * Visits a ParenthesizedExpression that may contain a destructuring assignment.
             *
             * @param node A ParenthesizedExpression node.
             * @param expressionResultIsUnused Indicates the result of an expression is unused by the parent node (i.e., the left side of a comma or the
             * expression of an `ExpressionStatement`).
             */
            function visitParenthesizedExpression(node, expressionResultIsUnused) {
                return ts.visitEachChild(node, expressionResultIsUnused ? visitorWithUnusedExpressionResult : visitor, context);
            }
            /**
             * Visits a BinaryExpression that contains a destructuring assignment.
             *
             * @param node A BinaryExpression node.
             * @param expressionResultIsUnused Indicates the result of an expression is unused by the parent node (i.e., the left side of a comma or the
             * expression of an `ExpressionStatement`).
             */
            function visitBinaryExpression(node, expressionResultIsUnused) {
                // If we are here it is because this is a destructuring assignment.
                if (ts.isDestructuringAssignment(node)) {
                    return ts.flattenDestructuringAssignment(node, visitor, context, 0 /* All */, !expressionResultIsUnused);
                }
                if (node.operatorToken.kind === 27 /* CommaToken */) {
                    return factory.updateBinaryExpression(node, ts.visitNode(node.left, visitorWithUnusedExpressionResult, ts.isExpression), node.operatorToken, ts.visitNode(node.right, expressionResultIsUnused ? visitorWithUnusedExpressionResult : visitor, ts.isExpression));
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * @param expressionResultIsUnused Indicates the result of an expression is unused by the parent node (i.e., the left side of a comma or the
             * expression of an `ExpressionStatement`).
             */
            function visitCommaListExpression(node, expressionResultIsUnused) {
                if (expressionResultIsUnused) {
                    return ts.visitEachChild(node, visitorWithUnusedExpressionResult, context);
                }
                var result;
                for (var i = 0; i < node.elements.length; i++) {
                    var element = node.elements[i];
                    var visited = ts.visitNode(element, i < node.elements.length - 1 ? visitorWithUnusedExpressionResult : visitor, ts.isExpression);
                    if (result || visited !== element) {
                        result || (result = node.elements.slice(0, i));
                        result.push(visited);
                    }
                }
                var elements = result ? ts.setTextRange(factory.createNodeArray(result), node.elements) : node.elements;
                return factory.updateCommaListExpression(node, elements);
            }
            function isVariableStatementOfTypeScriptClassWrapper(node) {
                return node.declarationList.declarations.length === 1
                    && !!node.declarationList.declarations[0].initializer
                    && !!(ts.getEmitFlags(node.declarationList.declarations[0].initializer) & 33554432 /* TypeScriptClassWrapper */);
            }
            function visitVariableStatement(node) {
                var ancestorFacts = enterSubtree(0 /* None */, ts.hasSyntacticModifier(node, 1 /* Export */) ? 32 /* ExportedVariableStatement */ : 0 /* None */);
                var updated;
                if (convertedLoopState && (node.declarationList.flags & 3 /* BlockScoped */) === 0 && !isVariableStatementOfTypeScriptClassWrapper(node)) {
                    // we are inside a converted loop - hoist variable declarations
                    var assignments = void 0;
                    for (var _i = 0, _a = node.declarationList.declarations; _i < _a.length; _i++) {
                        var decl = _a[_i];
                        hoistVariableDeclarationDeclaredInConvertedLoop(convertedLoopState, decl);
                        if (decl.initializer) {
                            var assignment = void 0;
                            if (ts.isBindingPattern(decl.name)) {
                                assignment = ts.flattenDestructuringAssignment(decl, visitor, context, 0 /* All */);
                            }
                            else {
                                assignment = factory.createBinaryExpression(decl.name, 63 /* EqualsToken */, ts.visitNode(decl.initializer, visitor, ts.isExpression));
                                ts.setTextRange(assignment, decl);
                            }
                            assignments = ts.append(assignments, assignment);
                        }
                    }
                    if (assignments) {
                        updated = ts.setTextRange(factory.createExpressionStatement(factory.inlineExpressions(assignments)), node);
                    }
                    else {
                        // none of declarations has initializer - the entire variable statement can be deleted
                        updated = undefined;
                    }
                }
                else {
                    updated = ts.visitEachChild(node, visitor, context);
                }
                exitSubtree(ancestorFacts, 0 /* None */, 0 /* None */);
                return updated;
            }
            /**
             * Visits a VariableDeclarationList that is block scoped (e.g. `let` or `const`).
             *
             * @param node A VariableDeclarationList node.
             */
            function visitVariableDeclarationList(node) {
                if (node.flags & 3 /* BlockScoped */ || node.transformFlags & 262144 /* ContainsBindingPattern */) {
                    if (node.flags & 3 /* BlockScoped */) {
                        enableSubstitutionsForBlockScopedBindings();
                    }
                    var declarations = ts.flatMap(node.declarations, node.flags & 1 /* Let */
                        ? visitVariableDeclarationInLetDeclarationList
                        : visitVariableDeclaration);
                    var declarationList = factory.createVariableDeclarationList(declarations);
                    ts.setOriginalNode(declarationList, node);
                    ts.setTextRange(declarationList, node);
                    ts.setCommentRange(declarationList, node);
                    // If the first or last declaration is a binding pattern, we need to modify
                    // the source map range for the declaration list.
                    if (node.transformFlags & 262144 /* ContainsBindingPattern */
                        && (ts.isBindingPattern(node.declarations[0].name) || ts.isBindingPattern(ts.last(node.declarations).name))) {
                        ts.setSourceMapRange(declarationList, getRangeUnion(declarations));
                    }
                    return declarationList;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function getRangeUnion(declarations) {
                // declarations may not be sorted by position.
                // pos should be the minimum* position over all nodes (that's not -1), end should be the maximum end over all nodes.
                var pos = -1, end = -1;
                for (var _i = 0, declarations_9 = declarations; _i < declarations_9.length; _i++) {
                    var node = declarations_9[_i];
                    pos = pos === -1 ? node.pos : node.pos === -1 ? pos : Math.min(pos, node.pos);
                    end = Math.max(end, node.end);
                }
                return ts.createRange(pos, end);
            }
            /**
             * Gets a value indicating whether we should emit an explicit initializer for a variable
             * declaration in a `let` declaration list.
             *
             * @param node A VariableDeclaration node.
             */
            function shouldEmitExplicitInitializerForLetDeclaration(node) {
                // Nested let bindings might need to be initialized explicitly to preserve
                // ES6 semantic:
                //
                //  { let x = 1; }
                //  { let x; } // x here should be undefined. not 1
                //
                // Top level bindings never collide with anything and thus don't require
                // explicit initialization. As for nested let bindings there are two cases:
                //
                // - Nested let bindings that were not renamed definitely should be
                //   initialized explicitly:
                //
                //    { let x = 1; }
                //    { let x; if (some-condition) { x = 1}; if (x) { /*1*/ } }
                //
                //   Without explicit initialization code in /*1*/ can be executed even if
                //   some-condition is evaluated to false.
                //
                // - Renaming introduces fresh name that should not collide with any
                //   existing names, however renamed bindings sometimes also should be
                //   explicitly initialized. One particular case: non-captured binding
                //   declared inside loop body (but not in loop initializer):
                //
                //    let x;
                //    for (;;) {
                //        let x;
                //    }
                //
                //   In downlevel codegen inner 'x' will be renamed so it won't collide
                //   with outer 'x' however it will should be reset on every iteration as
                //   if it was declared anew.
                //
                //   * Why non-captured binding?
                //     - Because if loop contains block scoped binding captured in some
                //       function then loop body will be rewritten to have a fresh scope
                //       on every iteration so everything will just work.
                //
                //   * Why loop initializer is excluded?
                //     - Since we've introduced a fresh name it already will be undefined.
                var flags = resolver.getNodeCheckFlags(node);
                var isCapturedInFunction = flags & 262144 /* CapturedBlockScopedBinding */;
                var isDeclaredInLoop = flags & 524288 /* BlockScopedBindingInLoop */;
                var emittedAsTopLevel = (hierarchyFacts & 64 /* TopLevel */) !== 0
                    || (isCapturedInFunction
                        && isDeclaredInLoop
                        && (hierarchyFacts & 512 /* IterationStatementBlock */) !== 0);
                var emitExplicitInitializer = !emittedAsTopLevel
                    && (hierarchyFacts & 4096 /* ForInOrForOfStatement */) === 0
                    && (!resolver.isDeclarationWithCollidingName(node)
                        || (isDeclaredInLoop
                            && !isCapturedInFunction
                            && (hierarchyFacts & (2048 /* ForStatement */ | 4096 /* ForInOrForOfStatement */)) === 0));
                return emitExplicitInitializer;
            }
            /**
             * Visits a VariableDeclaration in a `let` declaration list.
             *
             * @param node A VariableDeclaration node.
             */
            function visitVariableDeclarationInLetDeclarationList(node) {
                // For binding pattern names that lack initializers there is no point to emit
                // explicit initializer since downlevel codegen for destructuring will fail
                // in the absence of initializer so all binding elements will say uninitialized
                var name = node.name;
                if (ts.isBindingPattern(name)) {
                    return visitVariableDeclaration(node);
                }
                if (!node.initializer && shouldEmitExplicitInitializerForLetDeclaration(node)) {
                    return factory.updateVariableDeclaration(node, node.name, /*exclamationToken*/ undefined, /*type*/ undefined, factory.createVoidZero());
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visits a VariableDeclaration node with a binding pattern.
             *
             * @param node A VariableDeclaration node.
             */
            function visitVariableDeclaration(node) {
                var ancestorFacts = enterSubtree(32 /* ExportedVariableStatement */, 0 /* None */);
                var updated;
                if (ts.isBindingPattern(node.name)) {
                    updated = ts.flattenDestructuringBinding(node, visitor, context, 0 /* All */, 
                    /*value*/ undefined, (ancestorFacts & 32 /* ExportedVariableStatement */) !== 0);
                }
                else {
                    updated = ts.visitEachChild(node, visitor, context);
                }
                exitSubtree(ancestorFacts, 0 /* None */, 0 /* None */);
                return updated;
            }
            function recordLabel(node) {
                convertedLoopState.labels.set(ts.idText(node.label), true);
            }
            function resetLabel(node) {
                convertedLoopState.labels.set(ts.idText(node.label), false);
            }
            function visitLabeledStatement(node) {
                if (convertedLoopState && !convertedLoopState.labels) {
                    convertedLoopState.labels = new ts.Map();
                }
                var statement = ts.unwrapInnermostStatementOfLabel(node, convertedLoopState && recordLabel);
                return ts.isIterationStatement(statement, /*lookInLabeledStatements*/ false)
                    ? visitIterationStatement(statement, /*outermostLabeledStatement*/ node)
                    : factory.restoreEnclosingLabel(ts.visitNode(statement, visitor, ts.isStatement, factory.liftToBlock), node, convertedLoopState && resetLabel);
            }
            function visitIterationStatement(node, outermostLabeledStatement) {
                switch (node.kind) {
                    case 238 /* DoStatement */:
                    case 239 /* WhileStatement */:
                        return visitDoOrWhileStatement(node, outermostLabeledStatement);
                    case 240 /* ForStatement */:
                        return visitForStatement(node, outermostLabeledStatement);
                    case 241 /* ForInStatement */:
                        return visitForInStatement(node, outermostLabeledStatement);
                    case 242 /* ForOfStatement */:
                        return visitForOfStatement(node, outermostLabeledStatement);
                }
            }
            function visitIterationStatementWithFacts(excludeFacts, includeFacts, node, outermostLabeledStatement, convert) {
                var ancestorFacts = enterSubtree(excludeFacts, includeFacts);
                var updated = convertIterationStatementBodyIfNecessary(node, outermostLabeledStatement, ancestorFacts, convert);
                exitSubtree(ancestorFacts, 0 /* None */, 0 /* None */);
                return updated;
            }
            function visitDoOrWhileStatement(node, outermostLabeledStatement) {
                return visitIterationStatementWithFacts(0 /* DoOrWhileStatementExcludes */, 1280 /* DoOrWhileStatementIncludes */, node, outermostLabeledStatement);
            }
            function visitForStatement(node, outermostLabeledStatement) {
                return visitIterationStatementWithFacts(5056 /* ForStatementExcludes */, 3328 /* ForStatementIncludes */, node, outermostLabeledStatement);
            }
            function visitEachChildOfForStatement(node) {
                return factory.updateForStatement(node, ts.visitNode(node.initializer, visitorWithUnusedExpressionResult, ts.isForInitializer), ts.visitNode(node.condition, visitor, ts.isExpression), ts.visitNode(node.incrementor, visitorWithUnusedExpressionResult, ts.isExpression), ts.visitNode(node.statement, visitor, ts.isStatement, factory.liftToBlock));
            }
            function visitForInStatement(node, outermostLabeledStatement) {
                return visitIterationStatementWithFacts(3008 /* ForInOrForOfStatementExcludes */, 5376 /* ForInOrForOfStatementIncludes */, node, outermostLabeledStatement);
            }
            function visitForOfStatement(node, outermostLabeledStatement) {
                return visitIterationStatementWithFacts(3008 /* ForInOrForOfStatementExcludes */, 5376 /* ForInOrForOfStatementIncludes */, node, outermostLabeledStatement, compilerOptions.downlevelIteration ? convertForOfStatementForIterable : convertForOfStatementForArray);
            }
            function convertForOfStatementHead(node, boundValue, convertedLoopBodyStatements) {
                var statements = [];
                var initializer = node.initializer;
                if (ts.isVariableDeclarationList(initializer)) {
                    if (node.initializer.flags & 3 /* BlockScoped */) {
                        enableSubstitutionsForBlockScopedBindings();
                    }
                    var firstOriginalDeclaration = ts.firstOrUndefined(initializer.declarations);
                    if (firstOriginalDeclaration && ts.isBindingPattern(firstOriginalDeclaration.name)) {
                        // This works whether the declaration is a var, let, or const.
                        // It will use rhsIterationValue _a[_i] as the initializer.
                        var declarations = ts.flattenDestructuringBinding(firstOriginalDeclaration, visitor, context, 0 /* All */, boundValue);
                        var declarationList = ts.setTextRange(factory.createVariableDeclarationList(declarations), node.initializer);
                        ts.setOriginalNode(declarationList, node.initializer);
                        // Adjust the source map range for the first declaration to align with the old
                        // emitter.
                        ts.setSourceMapRange(declarationList, ts.createRange(declarations[0].pos, ts.last(declarations).end));
                        statements.push(factory.createVariableStatement(
                        /*modifiers*/ undefined, declarationList));
                    }
                    else {
                        // The following call does not include the initializer, so we have
                        // to emit it separately.
                        statements.push(ts.setTextRange(factory.createVariableStatement(
                        /*modifiers*/ undefined, ts.setOriginalNode(ts.setTextRange(factory.createVariableDeclarationList([
                            factory.createVariableDeclaration(firstOriginalDeclaration ? firstOriginalDeclaration.name : factory.createTempVariable(/*recordTempVariable*/ undefined), 
                            /*exclamationToken*/ undefined, 
                            /*type*/ undefined, boundValue)
                        ]), ts.moveRangePos(initializer, -1)), initializer)), ts.moveRangeEnd(initializer, -1)));
                    }
                }
                else {
                    // Initializer is an expression. Emit the expression in the body, so that it's
                    // evaluated on every iteration.
                    var assignment = factory.createAssignment(initializer, boundValue);
                    if (ts.isDestructuringAssignment(assignment)) {
                        statements.push(factory.createExpressionStatement(visitBinaryExpression(assignment, /*expressionResultIsUnused*/ true)));
                    }
                    else {
                        ts.setTextRangeEnd(assignment, initializer.end);
                        statements.push(ts.setTextRange(factory.createExpressionStatement(ts.visitNode(assignment, visitor, ts.isExpression)), ts.moveRangeEnd(initializer, -1)));
                    }
                }
                if (convertedLoopBodyStatements) {
                    return createSyntheticBlockForConvertedStatements(ts.addRange(statements, convertedLoopBodyStatements));
                }
                else {
                    var statement = ts.visitNode(node.statement, visitor, ts.isStatement, factory.liftToBlock);
                    if (ts.isBlock(statement)) {
                        return factory.updateBlock(statement, ts.setTextRange(factory.createNodeArray(ts.concatenate(statements, statement.statements)), statement.statements));
                    }
                    else {
                        statements.push(statement);
                        return createSyntheticBlockForConvertedStatements(statements);
                    }
                }
            }
            function createSyntheticBlockForConvertedStatements(statements) {
                return ts.setEmitFlags(factory.createBlock(factory.createNodeArray(statements), 
                /*multiLine*/ true), 48 /* NoSourceMap */ | 384 /* NoTokenSourceMaps */);
            }
            function convertForOfStatementForArray(node, outermostLabeledStatement, convertedLoopBodyStatements) {
                // The following ES6 code:
                //
                //    for (let v of expr) { }
                //
                // should be emitted as
                //
                //    for (var _i = 0, _a = expr; _i < _a.length; _i++) {
                //        var v = _a[_i];
                //    }
                //
                // where _a and _i are temps emitted to capture the RHS and the counter,
                // respectively.
                // When the left hand side is an expression instead of a let declaration,
                // the "let v" is not emitted.
                // When the left hand side is a let/const, the v is renamed if there is
                // another v in scope.
                // Note that all assignments to the LHS are emitted in the body, including
                // all destructuring.
                // Note also that because an extra statement is needed to assign to the LHS,
                // for-of bodies are always emitted as blocks.
                var expression = ts.visitNode(node.expression, visitor, ts.isExpression);
                // In the case where the user wrote an identifier as the RHS, like this:
                //
                //     for (let v of arr) { }
                //
                // we don't want to emit a temporary variable for the RHS, just use it directly.
                var counter = factory.createLoopVariable();
                var rhsReference = ts.isIdentifier(expression) ? factory.getGeneratedNameForNode(expression) : factory.createTempVariable(/*recordTempVariable*/ undefined);
                // The old emitter does not emit source maps for the expression
                ts.setEmitFlags(expression, 48 /* NoSourceMap */ | ts.getEmitFlags(expression));
                var forStatement = ts.setTextRange(factory.createForStatement(
                /*initializer*/ ts.setEmitFlags(ts.setTextRange(factory.createVariableDeclarationList([
                    ts.setTextRange(factory.createVariableDeclaration(counter, /*exclamationToken*/ undefined, /*type*/ undefined, factory.createNumericLiteral(0)), ts.moveRangePos(node.expression, -1)),
                    ts.setTextRange(factory.createVariableDeclaration(rhsReference, /*exclamationToken*/ undefined, /*type*/ undefined, expression), node.expression)
                ]), node.expression), 2097152 /* NoHoisting */), 
                /*condition*/ ts.setTextRange(factory.createLessThan(counter, factory.createPropertyAccessExpression(rhsReference, "length")), node.expression), 
                /*incrementor*/ ts.setTextRange(factory.createPostfixIncrement(counter), node.expression), 
                /*statement*/ convertForOfStatementHead(node, factory.createElementAccessExpression(rhsReference, counter), convertedLoopBodyStatements)), 
                /*location*/ node);
                // Disable trailing source maps for the OpenParenToken to align source map emit with the old emitter.
                ts.setEmitFlags(forStatement, 256 /* NoTokenTrailingSourceMaps */);
                ts.setTextRange(forStatement, node);
                return factory.restoreEnclosingLabel(forStatement, outermostLabeledStatement, convertedLoopState && resetLabel);
            }
            function convertForOfStatementForIterable(node, outermostLabeledStatement, convertedLoopBodyStatements, ancestorFacts) {
                var expression = ts.visitNode(node.expression, visitor, ts.isExpression);
                var iterator = ts.isIdentifier(expression) ? factory.getGeneratedNameForNode(expression) : factory.createTempVariable(/*recordTempVariable*/ undefined);
                var result = ts.isIdentifier(expression) ? factory.getGeneratedNameForNode(iterator) : factory.createTempVariable(/*recordTempVariable*/ undefined);
                var errorRecord = factory.createUniqueName("e");
                var catchVariable = factory.getGeneratedNameForNode(errorRecord);
                var returnMethod = factory.createTempVariable(/*recordTempVariable*/ undefined);
                var values = ts.setTextRange(emitHelpers().createValuesHelper(expression), node.expression);
                var next = factory.createCallExpression(factory.createPropertyAccessExpression(iterator, "next"), /*typeArguments*/ undefined, []);
                hoistVariableDeclaration(errorRecord);
                hoistVariableDeclaration(returnMethod);
                // if we are enclosed in an outer loop ensure we reset 'errorRecord' per each iteration
                var initializer = ancestorFacts & 1024 /* IterationContainer */
                    ? factory.inlineExpressions([factory.createAssignment(errorRecord, factory.createVoidZero()), values])
                    : values;
                var forStatement = ts.setEmitFlags(ts.setTextRange(factory.createForStatement(
                /*initializer*/ ts.setEmitFlags(ts.setTextRange(factory.createVariableDeclarationList([
                    ts.setTextRange(factory.createVariableDeclaration(iterator, /*exclamationToken*/ undefined, /*type*/ undefined, initializer), node.expression),
                    factory.createVariableDeclaration(result, /*exclamationToken*/ undefined, /*type*/ undefined, next)
                ]), node.expression), 2097152 /* NoHoisting */), 
                /*condition*/ factory.createLogicalNot(factory.createPropertyAccessExpression(result, "done")), 
                /*incrementor*/ factory.createAssignment(result, next), 
                /*statement*/ convertForOfStatementHead(node, factory.createPropertyAccessExpression(result, "value"), convertedLoopBodyStatements)), 
                /*location*/ node), 256 /* NoTokenTrailingSourceMaps */);
                return factory.createTryStatement(factory.createBlock([
                    factory.restoreEnclosingLabel(forStatement, outermostLabeledStatement, convertedLoopState && resetLabel)
                ]), factory.createCatchClause(factory.createVariableDeclaration(catchVariable), ts.setEmitFlags(factory.createBlock([
                    factory.createExpressionStatement(factory.createAssignment(errorRecord, factory.createObjectLiteralExpression([
                        factory.createPropertyAssignment("error", catchVariable)
                    ])))
                ]), 1 /* SingleLine */)), factory.createBlock([
                    factory.createTryStatement(
                    /*tryBlock*/ factory.createBlock([
                        ts.setEmitFlags(factory.createIfStatement(factory.createLogicalAnd(factory.createLogicalAnd(result, factory.createLogicalNot(factory.createPropertyAccessExpression(result, "done"))), factory.createAssignment(returnMethod, factory.createPropertyAccessExpression(iterator, "return"))), factory.createExpressionStatement(factory.createFunctionCallCall(returnMethod, iterator, []))), 1 /* SingleLine */),
                    ]), 
                    /*catchClause*/ undefined, 
                    /*finallyBlock*/ ts.setEmitFlags(factory.createBlock([
                        ts.setEmitFlags(factory.createIfStatement(errorRecord, factory.createThrowStatement(factory.createPropertyAccessExpression(errorRecord, "error"))), 1 /* SingleLine */)
                    ]), 1 /* SingleLine */))
                ]));
            }
            /**
             * Visits an ObjectLiteralExpression with computed property names.
             *
             * @param node An ObjectLiteralExpression node.
             */
            function visitObjectLiteralExpression(node) {
                var properties = node.properties;
                // Find the first computed property.
                // Everything until that point can be emitted as part of the initial object literal.
                var numInitialProperties = -1, hasComputed = false;
                for (var i = 0; i < properties.length; i++) {
                    var property = properties[i];
                    if ((property.transformFlags & 524288 /* ContainsYield */ &&
                        hierarchyFacts & 4 /* AsyncFunctionBody */)
                        || (hasComputed = ts.Debug.checkDefined(property.name).kind === 160 /* ComputedPropertyName */)) {
                        numInitialProperties = i;
                        break;
                    }
                }
                if (numInitialProperties < 0) {
                    return ts.visitEachChild(node, visitor, context);
                }
                // For computed properties, we need to create a unique handle to the object
                // literal so we can modify it without risking internal assignments tainting the object.
                var temp = factory.createTempVariable(hoistVariableDeclaration);
                // Write out the first non-computed properties, then emit the rest through indexing on the temp variable.
                var expressions = [];
                var assignment = factory.createAssignment(temp, ts.setEmitFlags(factory.createObjectLiteralExpression(ts.visitNodes(properties, visitor, ts.isObjectLiteralElementLike, 0, numInitialProperties), node.multiLine), hasComputed ? 65536 /* Indented */ : 0));
                if (node.multiLine) {
                    ts.startOnNewLine(assignment);
                }
                expressions.push(assignment);
                addObjectLiteralMembers(expressions, node, temp, numInitialProperties);
                // We need to clone the temporary identifier so that we can write it on a
                // new line
                expressions.push(node.multiLine ? ts.startOnNewLine(ts.setParent(ts.setTextRange(factory.cloneNode(temp), temp), temp.parent)) : temp);
                return factory.inlineExpressions(expressions);
            }
            function shouldConvertPartOfIterationStatement(node) {
                return (resolver.getNodeCheckFlags(node) & 131072 /* ContainsCapturedBlockScopeBinding */) !== 0;
            }
            function shouldConvertInitializerOfForStatement(node) {
                return ts.isForStatement(node) && !!node.initializer && shouldConvertPartOfIterationStatement(node.initializer);
            }
            function shouldConvertConditionOfForStatement(node) {
                return ts.isForStatement(node) && !!node.condition && shouldConvertPartOfIterationStatement(node.condition);
            }
            function shouldConvertIncrementorOfForStatement(node) {
                return ts.isForStatement(node) && !!node.incrementor && shouldConvertPartOfIterationStatement(node.incrementor);
            }
            function shouldConvertIterationStatement(node) {
                return shouldConvertBodyOfIterationStatement(node)
                    || shouldConvertInitializerOfForStatement(node);
            }
            function shouldConvertBodyOfIterationStatement(node) {
                return (resolver.getNodeCheckFlags(node) & 65536 /* LoopWithCapturedBlockScopedBinding */) !== 0;
            }
            /**
             * Records constituents of name for the given variable to be hoisted in the outer scope.
             */
            function hoistVariableDeclarationDeclaredInConvertedLoop(state, node) {
                if (!state.hoistedLocalVariables) {
                    state.hoistedLocalVariables = [];
                }
                visit(node.name);
                function visit(node) {
                    if (node.kind === 79 /* Identifier */) {
                        state.hoistedLocalVariables.push(node);
                    }
                    else {
                        for (var _i = 0, _a = node.elements; _i < _a.length; _i++) {
                            var element = _a[_i];
                            if (!ts.isOmittedExpression(element)) {
                                visit(element.name);
                            }
                        }
                    }
                }
            }
            function convertIterationStatementBodyIfNecessary(node, outermostLabeledStatement, ancestorFacts, convert) {
                if (!shouldConvertIterationStatement(node)) {
                    var saveAllowedNonLabeledJumps = void 0;
                    if (convertedLoopState) {
                        // we get here if we are trying to emit normal loop loop inside converted loop
                        // set allowedNonLabeledJumps to Break | Continue to mark that break\continue inside the loop should be emitted as is
                        saveAllowedNonLabeledJumps = convertedLoopState.allowedNonLabeledJumps;
                        convertedLoopState.allowedNonLabeledJumps = 2 /* Break */ | 4 /* Continue */;
                    }
                    var result = convert
                        ? convert(node, outermostLabeledStatement, /*convertedLoopBodyStatements*/ undefined, ancestorFacts)
                        : factory.restoreEnclosingLabel(ts.isForStatement(node) ? visitEachChildOfForStatement(node) : ts.visitEachChild(node, visitor, context), outermostLabeledStatement, convertedLoopState && resetLabel);
                    if (convertedLoopState) {
                        convertedLoopState.allowedNonLabeledJumps = saveAllowedNonLabeledJumps;
                    }
                    return result;
                }
                var currentState = createConvertedLoopState(node);
                var statements = [];
                var outerConvertedLoopState = convertedLoopState;
                convertedLoopState = currentState;
                var initializerFunction = shouldConvertInitializerOfForStatement(node) ? createFunctionForInitializerOfForStatement(node, currentState) : undefined;
                var bodyFunction = shouldConvertBodyOfIterationStatement(node) ? createFunctionForBodyOfIterationStatement(node, currentState, outerConvertedLoopState) : undefined;
                convertedLoopState = outerConvertedLoopState;
                if (initializerFunction)
                    statements.push(initializerFunction.functionDeclaration);
                if (bodyFunction)
                    statements.push(bodyFunction.functionDeclaration);
                addExtraDeclarationsForConvertedLoop(statements, currentState, outerConvertedLoopState);
                if (initializerFunction) {
                    statements.push(generateCallToConvertedLoopInitializer(initializerFunction.functionName, initializerFunction.containsYield));
                }
                var loop;
                if (bodyFunction) {
                    if (convert) {
                        loop = convert(node, outermostLabeledStatement, bodyFunction.part, ancestorFacts);
                    }
                    else {
                        var clone_4 = convertIterationStatementCore(node, initializerFunction, factory.createBlock(bodyFunction.part, /*multiLine*/ true));
                        loop = factory.restoreEnclosingLabel(clone_4, outermostLabeledStatement, convertedLoopState && resetLabel);
                    }
                }
                else {
                    var clone_5 = convertIterationStatementCore(node, initializerFunction, ts.visitNode(node.statement, visitor, ts.isStatement, factory.liftToBlock));
                    loop = factory.restoreEnclosingLabel(clone_5, outermostLabeledStatement, convertedLoopState && resetLabel);
                }
                statements.push(loop);
                return statements;
            }
            function convertIterationStatementCore(node, initializerFunction, convertedLoopBody) {
                switch (node.kind) {
                    case 240 /* ForStatement */: return convertForStatement(node, initializerFunction, convertedLoopBody);
                    case 241 /* ForInStatement */: return convertForInStatement(node, convertedLoopBody);
                    case 242 /* ForOfStatement */: return convertForOfStatement(node, convertedLoopBody);
                    case 238 /* DoStatement */: return convertDoStatement(node, convertedLoopBody);
                    case 239 /* WhileStatement */: return convertWhileStatement(node, convertedLoopBody);
                    default: return ts.Debug.failBadSyntaxKind(node, "IterationStatement expected");
                }
            }
            function convertForStatement(node, initializerFunction, convertedLoopBody) {
                var shouldConvertCondition = node.condition && shouldConvertPartOfIterationStatement(node.condition);
                var shouldConvertIncrementor = shouldConvertCondition || node.incrementor && shouldConvertPartOfIterationStatement(node.incrementor);
                return factory.updateForStatement(node, ts.visitNode(initializerFunction ? initializerFunction.part : node.initializer, visitorWithUnusedExpressionResult, ts.isForInitializer), ts.visitNode(shouldConvertCondition ? undefined : node.condition, visitor, ts.isExpression), ts.visitNode(shouldConvertIncrementor ? undefined : node.incrementor, visitorWithUnusedExpressionResult, ts.isExpression), convertedLoopBody);
            }
            function convertForOfStatement(node, convertedLoopBody) {
                return factory.updateForOfStatement(node, 
                /*awaitModifier*/ undefined, ts.visitNode(node.initializer, visitor, ts.isForInitializer), ts.visitNode(node.expression, visitor, ts.isExpression), convertedLoopBody);
            }
            function convertForInStatement(node, convertedLoopBody) {
                return factory.updateForInStatement(node, ts.visitNode(node.initializer, visitor, ts.isForInitializer), ts.visitNode(node.expression, visitor, ts.isExpression), convertedLoopBody);
            }
            function convertDoStatement(node, convertedLoopBody) {
                return factory.updateDoStatement(node, convertedLoopBody, ts.visitNode(node.expression, visitor, ts.isExpression));
            }
            function convertWhileStatement(node, convertedLoopBody) {
                return factory.updateWhileStatement(node, ts.visitNode(node.expression, visitor, ts.isExpression), convertedLoopBody);
            }
            function createConvertedLoopState(node) {
                var loopInitializer;
                switch (node.kind) {
                    case 240 /* ForStatement */:
                    case 241 /* ForInStatement */:
                    case 242 /* ForOfStatement */:
                        var initializer = node.initializer;
                        if (initializer && initializer.kind === 253 /* VariableDeclarationList */) {
                            loopInitializer = initializer;
                        }
                        break;
                }
                // variables that will be passed to the loop as parameters
                var loopParameters = [];
                // variables declared in the loop initializer that will be changed inside the loop
                var loopOutParameters = [];
                if (loopInitializer && (ts.getCombinedNodeFlags(loopInitializer) & 3 /* BlockScoped */)) {
                    var hasCapturedBindingsInForInitializer = shouldConvertInitializerOfForStatement(node);
                    for (var _i = 0, _a = loopInitializer.declarations; _i < _a.length; _i++) {
                        var decl = _a[_i];
                        processLoopVariableDeclaration(node, decl, loopParameters, loopOutParameters, hasCapturedBindingsInForInitializer);
                    }
                }
                var currentState = { loopParameters: loopParameters, loopOutParameters: loopOutParameters };
                if (convertedLoopState) {
                    // convertedOuterLoopState !== undefined means that this converted loop is nested in another converted loop.
                    // if outer converted loop has already accumulated some state - pass it through
                    if (convertedLoopState.argumentsName) {
                        // outer loop has already used 'arguments' so we've already have some name to alias it
                        // use the same name in all nested loops
                        currentState.argumentsName = convertedLoopState.argumentsName;
                    }
                    if (convertedLoopState.thisName) {
                        // outer loop has already used 'this' so we've already have some name to alias it
                        // use the same name in all nested loops
                        currentState.thisName = convertedLoopState.thisName;
                    }
                    if (convertedLoopState.hoistedLocalVariables) {
                        // we've already collected some non-block scoped variable declarations in enclosing loop
                        // use the same storage in nested loop
                        currentState.hoistedLocalVariables = convertedLoopState.hoistedLocalVariables;
                    }
                }
                return currentState;
            }
            function addExtraDeclarationsForConvertedLoop(statements, state, outerState) {
                var extraVariableDeclarations;
                // propagate state from the inner loop to the outer loop if necessary
                if (state.argumentsName) {
                    // if alias for arguments is set
                    if (outerState) {
                        // pass it to outer converted loop
                        outerState.argumentsName = state.argumentsName;
                    }
                    else {
                        // this is top level converted loop and we need to create an alias for 'arguments' object
                        (extraVariableDeclarations || (extraVariableDeclarations = [])).push(factory.createVariableDeclaration(state.argumentsName, 
                        /*exclamationToken*/ undefined, 
                        /*type*/ undefined, factory.createIdentifier("arguments")));
                    }
                }
                if (state.thisName) {
                    // if alias for this is set
                    if (outerState) {
                        // pass it to outer converted loop
                        outerState.thisName = state.thisName;
                    }
                    else {
                        // this is top level converted loop so we need to create an alias for 'this' here
                        // NOTE:
                        // if converted loops were all nested in arrow function then we'll always emit '_this' so convertedLoopState.thisName will not be set.
                        // If it is set this means that all nested loops are not nested in arrow function and it is safe to capture 'this'.
                        (extraVariableDeclarations || (extraVariableDeclarations = [])).push(factory.createVariableDeclaration(state.thisName, 
                        /*exclamationToken*/ undefined, 
                        /*type*/ undefined, factory.createIdentifier("this")));
                    }
                }
                if (state.hoistedLocalVariables) {
                    // if hoistedLocalVariables !== undefined this means that we've possibly collected some variable declarations to be hoisted later
                    if (outerState) {
                        // pass them to outer converted loop
                        outerState.hoistedLocalVariables = state.hoistedLocalVariables;
                    }
                    else {
                        if (!extraVariableDeclarations) {
                            extraVariableDeclarations = [];
                        }
                        // hoist collected variable declarations
                        for (var _i = 0, _a = state.hoistedLocalVariables; _i < _a.length; _i++) {
                            var identifier = _a[_i];
                            extraVariableDeclarations.push(factory.createVariableDeclaration(identifier));
                        }
                    }
                }
                // add extra variables to hold out parameters if necessary
                if (state.loopOutParameters.length) {
                    if (!extraVariableDeclarations) {
                        extraVariableDeclarations = [];
                    }
                    for (var _b = 0, _c = state.loopOutParameters; _b < _c.length; _b++) {
                        var outParam = _c[_b];
                        extraVariableDeclarations.push(factory.createVariableDeclaration(outParam.outParamName));
                    }
                }
                if (state.conditionVariable) {
                    if (!extraVariableDeclarations) {
                        extraVariableDeclarations = [];
                    }
                    extraVariableDeclarations.push(factory.createVariableDeclaration(state.conditionVariable, /*exclamationToken*/ undefined, /*type*/ undefined, factory.createFalse()));
                }
                // create variable statement to hold all introduced variable declarations
                if (extraVariableDeclarations) {
                    statements.push(factory.createVariableStatement(
                    /*modifiers*/ undefined, factory.createVariableDeclarationList(extraVariableDeclarations)));
                }
            }
            function createOutVariable(p) {
                return factory.createVariableDeclaration(p.originalName, /*exclamationToken*/ undefined, /*type*/ undefined, p.outParamName);
            }
            /**
             * Creates a `_loop_init` function for a `ForStatement` with a block-scoped initializer
             * that is captured in a closure inside of the initializer. The `_loop_init` function is
             * used to preserve the per-iteration environment semantics of
             * [13.7.4.8 RS: ForBodyEvaluation](https://tc39.github.io/ecma262/#sec-forbodyevaluation).
             */
            function createFunctionForInitializerOfForStatement(node, currentState) {
                var functionName = factory.createUniqueName("_loop_init");
                var containsYield = (node.initializer.transformFlags & 524288 /* ContainsYield */) !== 0;
                var emitFlags = 0 /* None */;
                if (currentState.containsLexicalThis)
                    emitFlags |= 8 /* CapturesThis */;
                if (containsYield && hierarchyFacts & 4 /* AsyncFunctionBody */)
                    emitFlags |= 262144 /* AsyncFunctionBody */;
                var statements = [];
                statements.push(factory.createVariableStatement(/*modifiers*/ undefined, node.initializer));
                copyOutParameters(currentState.loopOutParameters, 2 /* Initializer */, 1 /* ToOutParameter */, statements);
                // This transforms the following ES2015 syntax:
                //
                //  for (let i = (setImmediate(() => console.log(i)), 0); i < 2; i++) {
                //      // loop body
                //  }
                //
                // Into the following ES5 syntax:
                //
                //  var _loop_init_1 = function () {
                //      var i = (setImmediate(() => console.log(i)), 0);
                //      out_i_1 = i;
                //  };
                //  var out_i_1;
                //  _loop_init_1();
                //  for (var i = out_i_1; i < 2; i++) {
                //      // loop body
                //  }
                //
                // Which prevents mutations to `i` in the per-iteration environment of the body
                // from affecting the initial value for `i` outside of the per-iteration environment.
                var functionDeclaration = factory.createVariableStatement(
                /*modifiers*/ undefined, ts.setEmitFlags(factory.createVariableDeclarationList([
                    factory.createVariableDeclaration(functionName, 
                    /*exclamationToken*/ undefined, 
                    /*type*/ undefined, ts.setEmitFlags(factory.createFunctionExpression(
                    /*modifiers*/ undefined, containsYield ? factory.createToken(41 /* AsteriskToken */) : undefined, 
                    /*name*/ undefined, 
                    /*typeParameters*/ undefined, 
                    /*parameters*/ undefined, 
                    /*type*/ undefined, ts.visitNode(factory.createBlock(statements, /*multiLine*/ true), visitor, ts.isBlock)), emitFlags))
                ]), 2097152 /* NoHoisting */));
                var part = factory.createVariableDeclarationList(ts.map(currentState.loopOutParameters, createOutVariable));
                return { functionName: functionName, containsYield: containsYield, functionDeclaration: functionDeclaration, part: part };
            }
            /**
             * Creates a `_loop` function for an `IterationStatement` with a block-scoped initializer
             * that is captured in a closure inside of the loop body. The `_loop` function is used to
             * preserve the per-iteration environment semantics of
             * [13.7.4.8 RS: ForBodyEvaluation](https://tc39.github.io/ecma262/#sec-forbodyevaluation).
             */
            function createFunctionForBodyOfIterationStatement(node, currentState, outerState) {
                var functionName = factory.createUniqueName("_loop");
                startLexicalEnvironment();
                var statement = ts.visitNode(node.statement, visitor, ts.isStatement, factory.liftToBlock);
                var lexicalEnvironment = endLexicalEnvironment();
                var statements = [];
                if (shouldConvertConditionOfForStatement(node) || shouldConvertIncrementorOfForStatement(node)) {
                    // If a block-scoped variable declared in the initializer of `node` is captured in
                    // the condition or incrementor, we must move the condition and incrementor into
                    // the body of the for loop.
                    //
                    // This transforms the following ES2015 syntax:
                    //
                    //  for (let i = 0; setImmediate(() => console.log(i)), i < 2; setImmediate(() => console.log(i)), i++) {
                    //      // loop body
                    //  }
                    //
                    // Into the following ES5 syntax:
                    //
                    //  var _loop_1 = function (i) {
                    //      if (inc_1)
                    //          setImmediate(() => console.log(i)), i++;
                    //      else
                    //          inc_1 = true;
                    //      if (!(setImmediate(() => console.log(i)), i < 2))
                    //          return out_i_1 = i, "break";
                    //      // loop body
                    //      out_i_1 = i;
                    //  }
                    //  var out_i_1, inc_1 = false;
                    //  for (var i = 0;;) {
                    //      var state_1 = _loop_1(i);
                    //      i = out_i_1;
                    //      if (state_1 === "break")
                    //          break;
                    //  }
                    //
                    // Which prevents mutations to `i` in the per-iteration environment of the body
                    // from affecting the value of `i` in the previous per-iteration environment.
                    //
                    // Note that the incrementor of a `for` loop is evaluated in a *new* per-iteration
                    // environment that is carried over to the next iteration of the loop. As a result,
                    // we must indicate whether this is the first evaluation of the loop body so that
                    // we only evaluate the incrementor on subsequent evaluations.
                    currentState.conditionVariable = factory.createUniqueName("inc");
                    if (node.incrementor) {
                        statements.push(factory.createIfStatement(currentState.conditionVariable, factory.createExpressionStatement(ts.visitNode(node.incrementor, visitor, ts.isExpression)), factory.createExpressionStatement(factory.createAssignment(currentState.conditionVariable, factory.createTrue()))));
                    }
                    else {
                        statements.push(factory.createIfStatement(factory.createLogicalNot(currentState.conditionVariable), factory.createExpressionStatement(factory.createAssignment(currentState.conditionVariable, factory.createTrue()))));
                    }
                    if (shouldConvertConditionOfForStatement(node)) {
                        statements.push(factory.createIfStatement(factory.createPrefixUnaryExpression(53 /* ExclamationToken */, ts.visitNode(node.condition, visitor, ts.isExpression)), ts.visitNode(factory.createBreakStatement(), visitor, ts.isStatement)));
                    }
                }
                if (ts.isBlock(statement)) {
                    ts.addRange(statements, statement.statements);
                }
                else {
                    statements.push(statement);
                }
                copyOutParameters(currentState.loopOutParameters, 1 /* Body */, 1 /* ToOutParameter */, statements);
                ts.insertStatementsAfterStandardPrologue(statements, lexicalEnvironment);
                var loopBody = factory.createBlock(statements, /*multiLine*/ true);
                if (ts.isBlock(statement))
                    ts.setOriginalNode(loopBody, statement);
                var containsYield = (node.statement.transformFlags & 524288 /* ContainsYield */) !== 0;
                var emitFlags = 524288 /* ReuseTempVariableScope */;
                if (currentState.containsLexicalThis)
                    emitFlags |= 8 /* CapturesThis */;
                if (containsYield && (hierarchyFacts & 4 /* AsyncFunctionBody */) !== 0)
                    emitFlags |= 262144 /* AsyncFunctionBody */;
                // This transforms the following ES2015 syntax (in addition to other variations):
                //
                //  for (let i = 0; i < 2; i++) {
                //      setImmediate(() => console.log(i));
                //  }
                //
                // Into the following ES5 syntax:
                //
                //  var _loop_1 = function (i) {
                //      setImmediate(() => console.log(i));
                //  };
                //  for (var i = 0; i < 2; i++) {
                //      _loop_1(i);
                //  }
                var functionDeclaration = factory.createVariableStatement(
                /*modifiers*/ undefined, ts.setEmitFlags(factory.createVariableDeclarationList([
                    factory.createVariableDeclaration(functionName, 
                    /*exclamationToken*/ undefined, 
                    /*type*/ undefined, ts.setEmitFlags(factory.createFunctionExpression(
                    /*modifiers*/ undefined, containsYield ? factory.createToken(41 /* AsteriskToken */) : undefined, 
                    /*name*/ undefined, 
                    /*typeParameters*/ undefined, currentState.loopParameters, 
                    /*type*/ undefined, loopBody), emitFlags))
                ]), 2097152 /* NoHoisting */));
                var part = generateCallToConvertedLoop(functionName, currentState, outerState, containsYield);
                return { functionName: functionName, containsYield: containsYield, functionDeclaration: functionDeclaration, part: part };
            }
            function copyOutParameter(outParam, copyDirection) {
                var source = copyDirection === 0 /* ToOriginal */ ? outParam.outParamName : outParam.originalName;
                var target = copyDirection === 0 /* ToOriginal */ ? outParam.originalName : outParam.outParamName;
                return factory.createBinaryExpression(target, 63 /* EqualsToken */, source);
            }
            function copyOutParameters(outParams, partFlags, copyDirection, statements) {
                for (var _i = 0, outParams_1 = outParams; _i < outParams_1.length; _i++) {
                    var outParam = outParams_1[_i];
                    if (outParam.flags & partFlags) {
                        statements.push(factory.createExpressionStatement(copyOutParameter(outParam, copyDirection)));
                    }
                }
            }
            function generateCallToConvertedLoopInitializer(initFunctionExpressionName, containsYield) {
                var call = factory.createCallExpression(initFunctionExpressionName, /*typeArguments*/ undefined, []);
                var callResult = containsYield
                    ? factory.createYieldExpression(factory.createToken(41 /* AsteriskToken */), ts.setEmitFlags(call, 8388608 /* Iterator */))
                    : call;
                return factory.createExpressionStatement(callResult);
            }
            function generateCallToConvertedLoop(loopFunctionExpressionName, state, outerState, containsYield) {
                var statements = [];
                // loop is considered simple if it does not have any return statements or break\continue that transfer control outside of the loop
                // simple loops are emitted as just 'loop()';
                // NOTE: if loop uses only 'continue' it still will be emitted as simple loop
                var isSimpleLoop = !(state.nonLocalJumps & ~4 /* Continue */) &&
                    !state.labeledNonLocalBreaks &&
                    !state.labeledNonLocalContinues;
                var call = factory.createCallExpression(loopFunctionExpressionName, /*typeArguments*/ undefined, ts.map(state.loopParameters, function (p) { return p.name; }));
                var callResult = containsYield
                    ? factory.createYieldExpression(factory.createToken(41 /* AsteriskToken */), ts.setEmitFlags(call, 8388608 /* Iterator */))
                    : call;
                if (isSimpleLoop) {
                    statements.push(factory.createExpressionStatement(callResult));
                    copyOutParameters(state.loopOutParameters, 1 /* Body */, 0 /* ToOriginal */, statements);
                }
                else {
                    var loopResultName = factory.createUniqueName("state");
                    var stateVariable = factory.createVariableStatement(
                    /*modifiers*/ undefined, factory.createVariableDeclarationList([factory.createVariableDeclaration(loopResultName, /*exclamationToken*/ undefined, /*type*/ undefined, callResult)]));
                    statements.push(stateVariable);
                    copyOutParameters(state.loopOutParameters, 1 /* Body */, 0 /* ToOriginal */, statements);
                    if (state.nonLocalJumps & 8 /* Return */) {
                        var returnStatement = void 0;
                        if (outerState) {
                            outerState.nonLocalJumps |= 8 /* Return */;
                            returnStatement = factory.createReturnStatement(loopResultName);
                        }
                        else {
                            returnStatement = factory.createReturnStatement(factory.createPropertyAccessExpression(loopResultName, "value"));
                        }
                        statements.push(factory.createIfStatement(factory.createTypeCheck(loopResultName, "object"), returnStatement));
                    }
                    if (state.nonLocalJumps & 2 /* Break */) {
                        statements.push(factory.createIfStatement(factory.createStrictEquality(loopResultName, factory.createStringLiteral("break")), factory.createBreakStatement()));
                    }
                    if (state.labeledNonLocalBreaks || state.labeledNonLocalContinues) {
                        var caseClauses = [];
                        processLabeledJumps(state.labeledNonLocalBreaks, /*isBreak*/ true, loopResultName, outerState, caseClauses);
                        processLabeledJumps(state.labeledNonLocalContinues, /*isBreak*/ false, loopResultName, outerState, caseClauses);
                        statements.push(factory.createSwitchStatement(loopResultName, factory.createCaseBlock(caseClauses)));
                    }
                }
                return statements;
            }
            function setLabeledJump(state, isBreak, labelText, labelMarker) {
                if (isBreak) {
                    if (!state.labeledNonLocalBreaks) {
                        state.labeledNonLocalBreaks = new ts.Map();
                    }
                    state.labeledNonLocalBreaks.set(labelText, labelMarker);
                }
                else {
                    if (!state.labeledNonLocalContinues) {
                        state.labeledNonLocalContinues = new ts.Map();
                    }
                    state.labeledNonLocalContinues.set(labelText, labelMarker);
                }
            }
            function processLabeledJumps(table, isBreak, loopResultName, outerLoop, caseClauses) {
                if (!table) {
                    return;
                }
                table.forEach(function (labelMarker, labelText) {
                    var statements = [];
                    // if there are no outer converted loop or outer label in question is located inside outer converted loop
                    // then emit labeled break\continue
                    // otherwise propagate pair 'label -> marker' to outer converted loop and emit 'return labelMarker' so outer loop can later decide what to do
                    if (!outerLoop || (outerLoop.labels && outerLoop.labels.get(labelText))) {
                        var label = factory.createIdentifier(labelText);
                        statements.push(isBreak ? factory.createBreakStatement(label) : factory.createContinueStatement(label));
                    }
                    else {
                        setLabeledJump(outerLoop, isBreak, labelText, labelMarker);
                        statements.push(factory.createReturnStatement(loopResultName));
                    }
                    caseClauses.push(factory.createCaseClause(factory.createStringLiteral(labelMarker), statements));
                });
            }
            function processLoopVariableDeclaration(container, decl, loopParameters, loopOutParameters, hasCapturedBindingsInForInitializer) {
                var name = decl.name;
                if (ts.isBindingPattern(name)) {
                    for (var _i = 0, _a = name.elements; _i < _a.length; _i++) {
                        var element = _a[_i];
                        if (!ts.isOmittedExpression(element)) {
                            processLoopVariableDeclaration(container, element, loopParameters, loopOutParameters, hasCapturedBindingsInForInitializer);
                        }
                    }
                }
                else {
                    loopParameters.push(factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, name));
                    var checkFlags = resolver.getNodeCheckFlags(decl);
                    if (checkFlags & 4194304 /* NeedsLoopOutParameter */ || hasCapturedBindingsInForInitializer) {
                        var outParamName = factory.createUniqueName("out_" + ts.idText(name));
                        var flags = 0;
                        if (checkFlags & 4194304 /* NeedsLoopOutParameter */) {
                            flags |= 1 /* Body */;
                        }
                        if (ts.isForStatement(container) && container.initializer && resolver.isBindingCapturedByNode(container.initializer, decl)) {
                            flags |= 2 /* Initializer */;
                        }
                        loopOutParameters.push({ flags: flags, originalName: name, outParamName: outParamName });
                    }
                }
            }
            /**
             * Adds the members of an object literal to an array of expressions.
             *
             * @param expressions An array of expressions.
             * @param node An ObjectLiteralExpression node.
             * @param receiver The receiver for members of the ObjectLiteralExpression.
             * @param numInitialNonComputedProperties The number of initial properties without
             *                                        computed property names.
             */
            function addObjectLiteralMembers(expressions, node, receiver, start) {
                var properties = node.properties;
                var numProperties = properties.length;
                for (var i = start; i < numProperties; i++) {
                    var property = properties[i];
                    switch (property.kind) {
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            var accessors = ts.getAllAccessorDeclarations(node.properties, property);
                            if (property === accessors.firstAccessor) {
                                expressions.push(transformAccessorsToExpression(receiver, accessors, node, !!node.multiLine));
                            }
                            break;
                        case 167 /* MethodDeclaration */:
                            expressions.push(transformObjectLiteralMethodDeclarationToExpression(property, receiver, node, node.multiLine));
                            break;
                        case 291 /* PropertyAssignment */:
                            expressions.push(transformPropertyAssignmentToExpression(property, receiver, node.multiLine));
                            break;
                        case 292 /* ShorthandPropertyAssignment */:
                            expressions.push(transformShorthandPropertyAssignmentToExpression(property, receiver, node.multiLine));
                            break;
                        default:
                            ts.Debug.failBadSyntaxKind(node);
                            break;
                    }
                }
            }
            /**
             * Transforms a PropertyAssignment node into an expression.
             *
             * @param node The ObjectLiteralExpression that contains the PropertyAssignment.
             * @param property The PropertyAssignment node.
             * @param receiver The receiver for the assignment.
             */
            function transformPropertyAssignmentToExpression(property, receiver, startsOnNewLine) {
                var expression = factory.createAssignment(ts.createMemberAccessForPropertyName(factory, receiver, ts.visitNode(property.name, visitor, ts.isPropertyName)), ts.visitNode(property.initializer, visitor, ts.isExpression));
                ts.setTextRange(expression, property);
                if (startsOnNewLine) {
                    ts.startOnNewLine(expression);
                }
                return expression;
            }
            /**
             * Transforms a ShorthandPropertyAssignment node into an expression.
             *
             * @param node The ObjectLiteralExpression that contains the ShorthandPropertyAssignment.
             * @param property The ShorthandPropertyAssignment node.
             * @param receiver The receiver for the assignment.
             */
            function transformShorthandPropertyAssignmentToExpression(property, receiver, startsOnNewLine) {
                var expression = factory.createAssignment(ts.createMemberAccessForPropertyName(factory, receiver, ts.visitNode(property.name, visitor, ts.isPropertyName)), factory.cloneNode(property.name));
                ts.setTextRange(expression, property);
                if (startsOnNewLine) {
                    ts.startOnNewLine(expression);
                }
                return expression;
            }
            /**
             * Transforms a MethodDeclaration of an ObjectLiteralExpression into an expression.
             *
             * @param node The ObjectLiteralExpression that contains the MethodDeclaration.
             * @param method The MethodDeclaration node.
             * @param receiver The receiver for the assignment.
             */
            function transformObjectLiteralMethodDeclarationToExpression(method, receiver, container, startsOnNewLine) {
                var expression = factory.createAssignment(ts.createMemberAccessForPropertyName(factory, receiver, ts.visitNode(method.name, visitor, ts.isPropertyName)), transformFunctionLikeToExpression(method, /*location*/ method, /*name*/ undefined, container));
                ts.setTextRange(expression, method);
                if (startsOnNewLine) {
                    ts.startOnNewLine(expression);
                }
                return expression;
            }
            function visitCatchClause(node) {
                var ancestorFacts = enterSubtree(7104 /* BlockScopeExcludes */, 0 /* BlockScopeIncludes */);
                var updated;
                ts.Debug.assert(!!node.variableDeclaration, "Catch clause variable should always be present when downleveling ES2015.");
                if (ts.isBindingPattern(node.variableDeclaration.name)) {
                    var temp = factory.createTempVariable(/*recordTempVariable*/ undefined);
                    var newVariableDeclaration = factory.createVariableDeclaration(temp);
                    ts.setTextRange(newVariableDeclaration, node.variableDeclaration);
                    var vars = ts.flattenDestructuringBinding(node.variableDeclaration, visitor, context, 0 /* All */, temp);
                    var list = factory.createVariableDeclarationList(vars);
                    ts.setTextRange(list, node.variableDeclaration);
                    var destructure = factory.createVariableStatement(/*modifiers*/ undefined, list);
                    updated = factory.updateCatchClause(node, newVariableDeclaration, addStatementToStartOfBlock(node.block, destructure));
                }
                else {
                    updated = ts.visitEachChild(node, visitor, context);
                }
                exitSubtree(ancestorFacts, 0 /* None */, 0 /* None */);
                return updated;
            }
            function addStatementToStartOfBlock(block, statement) {
                var transformedStatements = ts.visitNodes(block.statements, visitor, ts.isStatement);
                return factory.updateBlock(block, __spreadArray([statement], transformedStatements, true));
            }
            /**
             * Visits a MethodDeclaration of an ObjectLiteralExpression and transforms it into a
             * PropertyAssignment.
             *
             * @param node A MethodDeclaration node.
             */
            function visitMethodDeclaration(node) {
                // We should only get here for methods on an object literal with regular identifier names.
                // Methods on classes are handled in visitClassDeclaration/visitClassExpression.
                // Methods with computed property names are handled in visitObjectLiteralExpression.
                ts.Debug.assert(!ts.isComputedPropertyName(node.name));
                var functionExpression = transformFunctionLikeToExpression(node, /*location*/ ts.moveRangePos(node, -1), /*name*/ undefined, /*container*/ undefined);
                ts.setEmitFlags(functionExpression, 512 /* NoLeadingComments */ | ts.getEmitFlags(functionExpression));
                return ts.setTextRange(factory.createPropertyAssignment(node.name, functionExpression), 
                /*location*/ node);
            }
            /**
             * Visits an AccessorDeclaration of an ObjectLiteralExpression.
             *
             * @param node An AccessorDeclaration node.
             */
            function visitAccessorDeclaration(node) {
                ts.Debug.assert(!ts.isComputedPropertyName(node.name));
                var savedConvertedLoopState = convertedLoopState;
                convertedLoopState = undefined;
                var ancestorFacts = enterSubtree(32670 /* FunctionExcludes */, 65 /* FunctionIncludes */);
                var updated;
                var parameters = ts.visitParameterList(node.parameters, visitor, context);
                var body = transformFunctionBody(node);
                if (node.kind === 170 /* GetAccessor */) {
                    updated = factory.updateGetAccessorDeclaration(node, node.decorators, node.modifiers, node.name, parameters, node.type, body);
                }
                else {
                    updated = factory.updateSetAccessorDeclaration(node, node.decorators, node.modifiers, node.name, parameters, body);
                }
                exitSubtree(ancestorFacts, 98304 /* FunctionSubtreeExcludes */, 0 /* None */);
                convertedLoopState = savedConvertedLoopState;
                return updated;
            }
            /**
             * Visits a ShorthandPropertyAssignment and transforms it into a PropertyAssignment.
             *
             * @param node A ShorthandPropertyAssignment node.
             */
            function visitShorthandPropertyAssignment(node) {
                return ts.setTextRange(factory.createPropertyAssignment(node.name, visitIdentifier(factory.cloneNode(node.name))), 
                /*location*/ node);
            }
            function visitComputedPropertyName(node) {
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visits a YieldExpression node.
             *
             * @param node A YieldExpression node.
             */
            function visitYieldExpression(node) {
                // `yield` expressions are transformed using the generators transformer.
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visits an ArrayLiteralExpression that contains a spread element.
             *
             * @param node An ArrayLiteralExpression node.
             */
            function visitArrayLiteralExpression(node) {
                if (ts.some(node.elements, ts.isSpreadElement)) {
                    // We are here because we contain a SpreadElementExpression.
                    return transformAndSpreadElements(node.elements, /*isArgumentList*/ false, !!node.multiLine, /*hasTrailingComma*/ !!node.elements.hasTrailingComma);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visits a CallExpression that contains either a spread element or `super`.
             *
             * @param node a CallExpression.
             */
            function visitCallExpression(node) {
                if (ts.getEmitFlags(node) & 33554432 /* TypeScriptClassWrapper */) {
                    return visitTypeScriptClassWrapper(node);
                }
                var expression = ts.skipOuterExpressions(node.expression);
                if (expression.kind === 106 /* SuperKeyword */ ||
                    ts.isSuperProperty(expression) ||
                    ts.some(node.arguments, ts.isSpreadElement)) {
                    return visitCallExpressionWithPotentialCapturedThisAssignment(node, /*assignToCapturedThis*/ true);
                }
                return factory.updateCallExpression(node, ts.visitNode(node.expression, callExpressionVisitor, ts.isExpression), 
                /*typeArguments*/ undefined, ts.visitNodes(node.arguments, visitor, ts.isExpression));
            }
            function visitTypeScriptClassWrapper(node) {
                // This is a call to a class wrapper function (an IIFE) created by the 'ts' transformer.
                // The wrapper has a form similar to:
                //
                //  (function() {
                //      class C { // 1
                //      }
                //      C.x = 1; // 2
                //      return C;
                //  }())
                //
                // When we transform the class, we end up with something like this:
                //
                //  (function () {
                //      var C = (function () { // 3
                //          function C() {
                //          }
                //          return C; // 4
                //      }());
                //      C.x = 1;
                //      return C;
                //  }())
                //
                // We want to simplify the two nested IIFEs to end up with something like this:
                //
                //  (function () {
                //      function C() {
                //      }
                //      C.x = 1;
                //      return C;
                //  }())
                // We skip any outer expressions in a number of places to get to the innermost
                // expression, but we will restore them later to preserve comments and source maps.
                var body = ts.cast(ts.cast(ts.skipOuterExpressions(node.expression), ts.isArrowFunction).body, ts.isBlock);
                // The class statements are the statements generated by visiting the first statement with initializer of the
                // body (1), while all other statements are added to remainingStatements (2)
                var isVariableStatementWithInitializer = function (stmt) { return ts.isVariableStatement(stmt) && !!ts.first(stmt.declarationList.declarations).initializer; };
                // visit the class body statements outside of any converted loop body.
                var savedConvertedLoopState = convertedLoopState;
                convertedLoopState = undefined;
                var bodyStatements = ts.visitNodes(body.statements, classWrapperStatementVisitor, ts.isStatement);
                convertedLoopState = savedConvertedLoopState;
                var classStatements = ts.filter(bodyStatements, isVariableStatementWithInitializer);
                var remainingStatements = ts.filter(bodyStatements, function (stmt) { return !isVariableStatementWithInitializer(stmt); });
                var varStatement = ts.cast(ts.first(classStatements), ts.isVariableStatement);
                // We know there is only one variable declaration here as we verified this in an
                // earlier call to isTypeScriptClassWrapper
                var variable = varStatement.declarationList.declarations[0];
                var initializer = ts.skipOuterExpressions(variable.initializer);
                // Under certain conditions, the 'ts' transformer may introduce a class alias, which
                // we see as an assignment, for example:
                //
                //  (function () {
                //      var C_1;
                //      var C = C_1 = (function () {
                //          function C() {
                //          }
                //          C.x = function () { return C_1; }
                //          return C;
                //      }());
                //      C = C_1 = __decorate([dec], C);
                //      return C;
                //  }())
                //
                var aliasAssignment = ts.tryCast(initializer, ts.isAssignmentExpression);
                if (!aliasAssignment && ts.isBinaryExpression(initializer) && initializer.operatorToken.kind === 27 /* CommaToken */) {
                    aliasAssignment = ts.tryCast(initializer.left, ts.isAssignmentExpression);
                }
                // The underlying call (3) is another IIFE that may contain a '_super' argument.
                var call = ts.cast(aliasAssignment ? ts.skipOuterExpressions(aliasAssignment.right) : initializer, ts.isCallExpression);
                var func = ts.cast(ts.skipOuterExpressions(call.expression), ts.isFunctionExpression);
                var funcStatements = func.body.statements;
                var classBodyStart = 0;
                var classBodyEnd = -1;
                var statements = [];
                if (aliasAssignment) {
                    // If we have a class alias assignment, we need to move it to the down-level constructor
                    // function we generated for the class.
                    var extendsCall = ts.tryCast(funcStatements[classBodyStart], ts.isExpressionStatement);
                    if (extendsCall) {
                        statements.push(extendsCall);
                        classBodyStart++;
                    }
                    // The next statement is the function declaration.
                    statements.push(funcStatements[classBodyStart]);
                    classBodyStart++;
                    // Add the class alias following the declaration.
                    statements.push(factory.createExpressionStatement(factory.createAssignment(aliasAssignment.left, ts.cast(variable.name, ts.isIdentifier))));
                }
                // Find the trailing 'return' statement (4)
                while (!ts.isReturnStatement(ts.elementAt(funcStatements, classBodyEnd))) {
                    classBodyEnd--;
                }
                // When we extract the statements of the inner IIFE, we exclude the 'return' statement (4)
                // as we already have one that has been introduced by the 'ts' transformer.
                ts.addRange(statements, funcStatements, classBodyStart, classBodyEnd);
                if (classBodyEnd < -1) {
                    // If there were any hoisted declarations following the return statement, we should
                    // append them.
                    ts.addRange(statements, funcStatements, classBodyEnd + 1);
                }
                // Add the remaining statements of the outer wrapper.
                ts.addRange(statements, remainingStatements);
                // The 'es2015' class transform may add an end-of-declaration marker. If so we will add it
                // after the remaining statements from the 'ts' transformer.
                ts.addRange(statements, classStatements, /*start*/ 1);
                // Recreate any outer parentheses or partially-emitted expressions to preserve source map
                // and comment locations.
                return factory.restoreOuterExpressions(node.expression, factory.restoreOuterExpressions(variable.initializer, factory.restoreOuterExpressions(aliasAssignment && aliasAssignment.right, factory.updateCallExpression(call, factory.restoreOuterExpressions(call.expression, factory.updateFunctionExpression(func, 
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, func.parameters, 
                /*type*/ undefined, factory.updateBlock(func.body, statements))), 
                /*typeArguments*/ undefined, call.arguments))));
            }
            function visitImmediateSuperCallInBody(node) {
                return visitCallExpressionWithPotentialCapturedThisAssignment(node, /*assignToCapturedThis*/ false);
            }
            function visitCallExpressionWithPotentialCapturedThisAssignment(node, assignToCapturedThis) {
                // We are here either because SuperKeyword was used somewhere in the expression, or
                // because we contain a SpreadElementExpression.
                if (node.transformFlags & 16384 /* ContainsRestOrSpread */ ||
                    node.expression.kind === 106 /* SuperKeyword */ ||
                    ts.isSuperProperty(ts.skipOuterExpressions(node.expression))) {
                    var _a = factory.createCallBinding(node.expression, hoistVariableDeclaration), target = _a.target, thisArg = _a.thisArg;
                    if (node.expression.kind === 106 /* SuperKeyword */) {
                        ts.setEmitFlags(thisArg, 4 /* NoSubstitution */);
                    }
                    var resultingCall = void 0;
                    if (node.transformFlags & 16384 /* ContainsRestOrSpread */) {
                        // [source]
                        //      f(...a, b)
                        //      x.m(...a, b)
                        //      super(...a, b)
                        //      super.m(...a, b) // in static
                        //      super.m(...a, b) // in instance
                        //
                        // [output]
                        //      f.apply(void 0, a.concat([b]))
                        //      (_a = x).m.apply(_a, a.concat([b]))
                        //      _super.apply(this, a.concat([b]))
                        //      _super.m.apply(this, a.concat([b]))
                        //      _super.prototype.m.apply(this, a.concat([b]))
                        resultingCall = factory.createFunctionApplyCall(ts.visitNode(target, callExpressionVisitor, ts.isExpression), node.expression.kind === 106 /* SuperKeyword */ ? thisArg : ts.visitNode(thisArg, visitor, ts.isExpression), transformAndSpreadElements(node.arguments, /*isArgumentList*/ true, /*multiLine*/ false, /*hasTrailingComma*/ false));
                    }
                    else {
                        // [source]
                        //      super(a)
                        //      super.m(a) // in static
                        //      super.m(a) // in instance
                        //
                        // [output]
                        //      _super.call(this, a)
                        //      _super.m.call(this, a)
                        //      _super.prototype.m.call(this, a)
                        resultingCall = ts.setTextRange(factory.createFunctionCallCall(ts.visitNode(target, callExpressionVisitor, ts.isExpression), node.expression.kind === 106 /* SuperKeyword */ ? thisArg : ts.visitNode(thisArg, visitor, ts.isExpression), ts.visitNodes(node.arguments, visitor, ts.isExpression)), node);
                    }
                    if (node.expression.kind === 106 /* SuperKeyword */) {
                        var initializer = factory.createLogicalOr(resultingCall, createActualThis());
                        resultingCall = assignToCapturedThis
                            ? factory.createAssignment(factory.createUniqueName("_this", 16 /* Optimistic */ | 32 /* FileLevel */), initializer)
                            : initializer;
                    }
                    return ts.setOriginalNode(resultingCall, node);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visits a NewExpression that contains a spread element.
             *
             * @param node A NewExpression node.
             */
            function visitNewExpression(node) {
                if (ts.some(node.arguments, ts.isSpreadElement)) {
                    // We are here because we contain a SpreadElementExpression.
                    // [source]
                    //      new C(...a)
                    //
                    // [output]
                    //      new ((_a = C).bind.apply(_a, [void 0].concat(a)))()
                    var _a = factory.createCallBinding(factory.createPropertyAccessExpression(node.expression, "bind"), hoistVariableDeclaration), target = _a.target, thisArg = _a.thisArg;
                    return factory.createNewExpression(factory.createFunctionApplyCall(ts.visitNode(target, visitor, ts.isExpression), thisArg, transformAndSpreadElements(factory.createNodeArray(__spreadArray([factory.createVoidZero()], node.arguments, true)), /*isArgumentList*/ true, /*multiLine*/ false, /*hasTrailingComma*/ false)), 
                    /*typeArguments*/ undefined, []);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Transforms an array of Expression nodes that contains a SpreadExpression.
             *
             * @param elements The array of Expression nodes.
             * @param isArgumentList A value indicating whether to ensure that the result is a fresh array.
             * This should be `false` when spreading into an `ArrayLiteral`, and `true` when spreading into an
             * argument list.
             * @param multiLine A value indicating whether the result should be emitted on multiple lines.
             */
            function transformAndSpreadElements(elements, isArgumentList, multiLine, hasTrailingComma) {
                // When there is no leading SpreadElement:
                //
                // [source]
                //      [a, ...b, c]
                //
                // [output (downlevelIteration)]
                //      __spreadArray(__spreadArray([a], __read(b)), [c])
                //
                // [output]
                //      __spreadArray(__spreadArray([a], b), [c])
                //
                // When there *is* a leading SpreadElement:
                //
                // [source]
                //      [...a, b]
                //
                // [output (downlevelIteration)]
                //      __spreadArray(__spreadArray([], __read(a)), [b])
                //
                // [output]
                //      __spreadArray(__spreadArray([], a), [b])
                //
                // NOTE: We use `isPackedArrayLiteral` below rather than just `isArrayLiteral`
                // because ES2015 spread will replace _missing_ array elements with `undefined`,
                // so we cannot just use an array as is. For example:
                //
                // `[1, ...[2, , 3]]` becomes `[1, 2, undefined, 3]`
                //
                // However, for packed array literals (i.e., an array literal with no OmittedExpression
                // elements), we can use the array as-is.
                // Map spans of spread expressions into their expressions and spans of other
                // expressions into an array literal.
                var numElements = elements.length;
                var segments = ts.flatten(
                // As we visit each element, we return one of two functions to use as the "key":
                // - `visitSpanOfSpreads` for one or more contiguous `...` spread expressions, i.e. `...a, ...b` in `[1, 2, ...a, ...b]`
                // - `visitSpanOfNonSpreads` for one or more contiguous non-spread elements, i.e. `1, 2`, in `[1, 2, ...a, ...b]`
                ts.spanMap(elements, partitionSpread, function (partition, visitPartition, _start, end) {
                    return visitPartition(partition, multiLine, hasTrailingComma && end === numElements);
                }));
                if (segments.length === 1) {
                    var firstSegment = segments[0];
                    // If we don't need a unique copy, then we are spreading into an argument list for
                    // a CallExpression or NewExpression. When using `--downlevelIteration`, we need
                    // to coerce this into an array for use with `apply`, so we will use the code path
                    // that follows instead.
                    if (isArgumentList && !compilerOptions.downlevelIteration
                        || ts.isPackedArrayLiteral(firstSegment.expression) // see NOTE (above)
                        || ts.isCallToHelper(firstSegment.expression, "___spreadArray")) {
                        return firstSegment.expression;
                    }
                }
                var helpers = emitHelpers();
                var startsWithSpread = segments[0].kind !== 0 /* None */;
                var expression = startsWithSpread ? factory.createArrayLiteralExpression() :
                    segments[0].expression;
                for (var i = startsWithSpread ? 0 : 1; i < segments.length; i++) {
                    var segment = segments[i];
                    // If this is for an argument list, it doesn't matter if the array is packed or sparse
                    expression = helpers.createSpreadArrayHelper(expression, segment.expression, segment.kind === 1 /* UnpackedSpread */ && !isArgumentList);
                }
                return expression;
            }
            function partitionSpread(node) {
                return ts.isSpreadElement(node)
                    ? visitSpanOfSpreads
                    : visitSpanOfNonSpreads;
            }
            function visitSpanOfSpreads(chunk) {
                return ts.map(chunk, visitExpressionOfSpread);
            }
            function visitExpressionOfSpread(node) {
                var expression = ts.visitNode(node.expression, visitor, ts.isExpression);
                // We don't need to pack already packed array literals, or existing calls to the `__read` helper.
                var isCallToReadHelper = ts.isCallToHelper(expression, "___read");
                var kind = isCallToReadHelper || ts.isPackedArrayLiteral(expression) ? 2 /* PackedSpread */ : 1 /* UnpackedSpread */;
                // We don't need the `__read` helper for array literals. Array packing will be performed by `__spreadArray`.
                if (compilerOptions.downlevelIteration && kind === 1 /* UnpackedSpread */ && !ts.isArrayLiteralExpression(expression) && !isCallToReadHelper) {
                    expression = emitHelpers().createReadHelper(expression, /*count*/ undefined);
                    // the `__read` helper returns a packed array, so we don't need to ensure a packed array
                    kind = 2 /* PackedSpread */;
                }
                return createSpreadSegment(kind, expression);
            }
            function visitSpanOfNonSpreads(chunk, multiLine, hasTrailingComma) {
                var expression = factory.createArrayLiteralExpression(ts.visitNodes(factory.createNodeArray(chunk, hasTrailingComma), visitor, ts.isExpression), multiLine);
                // We do not pack non-spread segments, this is so that `[1, , ...[2, , 3], , 4]` is properly downleveled to
                // `[1, , 2, undefined, 3, , 4]`. See the NOTE in `transformAndSpreadElements`
                return createSpreadSegment(0 /* None */, expression);
            }
            function visitSpreadElement(node) {
                return ts.visitNode(node.expression, visitor, ts.isExpression);
            }
            /**
             * Visits a template literal.
             *
             * @param node A template literal.
             */
            function visitTemplateLiteral(node) {
                return ts.setTextRange(factory.createStringLiteral(node.text), node);
            }
            /**
             * Visits a string literal with an extended unicode escape.
             *
             * @param node A string literal.
             */
            function visitStringLiteral(node) {
                if (node.hasExtendedUnicodeEscape) {
                    return ts.setTextRange(factory.createStringLiteral(node.text), node);
                }
                return node;
            }
            /**
             * Visits a binary or octal (ES6) numeric literal.
             *
             * @param node A string literal.
             */
            function visitNumericLiteral(node) {
                if (node.numericLiteralFlags & 384 /* BinaryOrOctalSpecifier */) {
                    return ts.setTextRange(factory.createNumericLiteral(node.text), node);
                }
                return node;
            }
            /**
             * Visits a TaggedTemplateExpression node.
             *
             * @param node A TaggedTemplateExpression node.
             */
            function visitTaggedTemplateExpression(node) {
                return ts.processTaggedTemplateExpression(context, node, visitor, currentSourceFile, recordTaggedTemplateString, ts.ProcessLevel.All);
            }
            /**
             * Visits a TemplateExpression node.
             *
             * @param node A TemplateExpression node.
             */
            function visitTemplateExpression(node) {
                var expressions = [];
                addTemplateHead(expressions, node);
                addTemplateSpans(expressions, node);
                // createAdd will check if each expression binds less closely than binary '+'.
                // If it does, it wraps the expression in parentheses. Otherwise, something like
                //    `abc${ 1 << 2 }`
                // becomes
                //    "abc" + 1 << 2 + ""
                // which is really
                //    ("abc" + 1) << (2 + "")
                // rather than
                //    "abc" + (1 << 2) + ""
                var expression = ts.reduceLeft(expressions, factory.createAdd);
                if (ts.nodeIsSynthesized(expression)) {
                    ts.setTextRange(expression, node);
                }
                return expression;
            }
            /**
             * Gets a value indicating whether we need to include the head of a TemplateExpression.
             *
             * @param node A TemplateExpression node.
             */
            function shouldAddTemplateHead(node) {
                // If this expression has an empty head literal and the first template span has a non-empty
                // literal, then emitting the empty head literal is not necessary.
                //     `${ foo } and ${ bar }`
                // can be emitted as
                //     foo + " and " + bar
                // This is because it is only required that one of the first two operands in the emit
                // output must be a string literal, so that the other operand and all following operands
                // are forced into strings.
                //
                // If the first template span has an empty literal, then the head must still be emitted.
                //     `${ foo }${ bar }`
                // must still be emitted as
                //     "" + foo + bar
                // There is always atleast one templateSpan in this code path, since
                // NoSubstitutionTemplateLiterals are directly emitted via emitLiteral()
                ts.Debug.assert(node.templateSpans.length !== 0);
                return node.head.text.length !== 0 || node.templateSpans[0].literal.text.length === 0;
            }
            /**
             * Adds the head of a TemplateExpression to an array of expressions.
             *
             * @param expressions An array of expressions.
             * @param node A TemplateExpression node.
             */
            function addTemplateHead(expressions, node) {
                if (!shouldAddTemplateHead(node)) {
                    return;
                }
                expressions.push(factory.createStringLiteral(node.head.text));
            }
            /**
             * Visits and adds the template spans of a TemplateExpression to an array of expressions.
             *
             * @param expressions An array of expressions.
             * @param node A TemplateExpression node.
             */
            function addTemplateSpans(expressions, node) {
                for (var _i = 0, _a = node.templateSpans; _i < _a.length; _i++) {
                    var span = _a[_i];
                    expressions.push(ts.visitNode(span.expression, visitor, ts.isExpression));
                    // Only emit if the literal is non-empty.
                    // The binary '+' operator is left-associative, so the first string concatenation
                    // with the head will force the result up to this point to be a string.
                    // Emitting a '+ ""' has no semantic effect for middles and tails.
                    if (span.literal.text.length !== 0) {
                        expressions.push(factory.createStringLiteral(span.literal.text));
                    }
                }
            }
            /**
             * Visits the `super` keyword
             */
            function visitSuperKeyword(isExpressionOfCall) {
                return hierarchyFacts & 8 /* NonStaticClassElement */
                    && !isExpressionOfCall
                    ? factory.createPropertyAccessExpression(factory.createUniqueName("_super", 16 /* Optimistic */ | 32 /* FileLevel */), "prototype")
                    : factory.createUniqueName("_super", 16 /* Optimistic */ | 32 /* FileLevel */);
            }
            function visitMetaProperty(node) {
                if (node.keywordToken === 103 /* NewKeyword */ && node.name.escapedText === "target") {
                    hierarchyFacts |= 32768 /* NewTarget */;
                    return factory.createUniqueName("_newTarget", 16 /* Optimistic */ | 32 /* FileLevel */);
                }
                return node;
            }
            /**
             * Called by the printer just before a node is printed.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to be printed.
             * @param emitCallback The callback used to emit the node.
             */
            function onEmitNode(hint, node, emitCallback) {
                if (enabledSubstitutions & 1 /* CapturedThis */ && ts.isFunctionLike(node)) {
                    // If we are tracking a captured `this`, keep track of the enclosing function.
                    var ancestorFacts = enterSubtree(32670 /* FunctionExcludes */, ts.getEmitFlags(node) & 8 /* CapturesThis */
                        ? 65 /* FunctionIncludes */ | 16 /* CapturesThis */
                        : 65 /* FunctionIncludes */);
                    previousOnEmitNode(hint, node, emitCallback);
                    exitSubtree(ancestorFacts, 0 /* None */, 0 /* None */);
                    return;
                }
                previousOnEmitNode(hint, node, emitCallback);
            }
            /**
             * Enables a more costly code path for substitutions when we determine a source file
             * contains block-scoped bindings (e.g. `let` or `const`).
             */
            function enableSubstitutionsForBlockScopedBindings() {
                if ((enabledSubstitutions & 2 /* BlockScopedBindings */) === 0) {
                    enabledSubstitutions |= 2 /* BlockScopedBindings */;
                    context.enableSubstitution(79 /* Identifier */);
                }
            }
            /**
             * Enables a more costly code path for substitutions when we determine a source file
             * contains a captured `this`.
             */
            function enableSubstitutionsForCapturedThis() {
                if ((enabledSubstitutions & 1 /* CapturedThis */) === 0) {
                    enabledSubstitutions |= 1 /* CapturedThis */;
                    context.enableSubstitution(108 /* ThisKeyword */);
                    context.enableEmitNotification(169 /* Constructor */);
                    context.enableEmitNotification(167 /* MethodDeclaration */);
                    context.enableEmitNotification(170 /* GetAccessor */);
                    context.enableEmitNotification(171 /* SetAccessor */);
                    context.enableEmitNotification(212 /* ArrowFunction */);
                    context.enableEmitNotification(211 /* FunctionExpression */);
                    context.enableEmitNotification(254 /* FunctionDeclaration */);
                }
            }
            /**
             * Hooks node substitutions.
             *
             * @param hint The context for the emitter.
             * @param node The node to substitute.
             */
            function onSubstituteNode(hint, node) {
                node = previousOnSubstituteNode(hint, node);
                if (hint === 1 /* Expression */) {
                    return substituteExpression(node);
                }
                if (ts.isIdentifier(node)) {
                    return substituteIdentifier(node);
                }
                return node;
            }
            /**
             * Hooks substitutions for non-expression identifiers.
             */
            function substituteIdentifier(node) {
                // Only substitute the identifier if we have enabled substitutions for block-scoped
                // bindings.
                if (enabledSubstitutions & 2 /* BlockScopedBindings */ && !ts.isInternalName(node)) {
                    var original = ts.getParseTreeNode(node, ts.isIdentifier);
                    if (original && isNameOfDeclarationWithCollidingName(original)) {
                        return ts.setTextRange(factory.getGeneratedNameForNode(original), node);
                    }
                }
                return node;
            }
            /**
             * Determines whether a name is the name of a declaration with a colliding name.
             * NOTE: This function expects to be called with an original source tree node.
             *
             * @param node An original source tree node.
             */
            function isNameOfDeclarationWithCollidingName(node) {
                switch (node.parent.kind) {
                    case 201 /* BindingElement */:
                    case 255 /* ClassDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 252 /* VariableDeclaration */:
                        return node.parent.name === node
                            && resolver.isDeclarationWithCollidingName(node.parent);
                }
                return false;
            }
            /**
             * Substitutes an expression.
             *
             * @param node An Expression node.
             */
            function substituteExpression(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        return substituteExpressionIdentifier(node);
                    case 108 /* ThisKeyword */:
                        return substituteThisKeyword(node);
                }
                return node;
            }
            /**
             * Substitutes an expression identifier.
             *
             * @param node An Identifier node.
             */
            function substituteExpressionIdentifier(node) {
                if (enabledSubstitutions & 2 /* BlockScopedBindings */ && !ts.isInternalName(node)) {
                    var declaration = resolver.getReferencedDeclarationWithCollidingName(node);
                    if (declaration && !(ts.isClassLike(declaration) && isPartOfClassBody(declaration, node))) {
                        return ts.setTextRange(factory.getGeneratedNameForNode(ts.getNameOfDeclaration(declaration)), node);
                    }
                }
                return node;
            }
            function isPartOfClassBody(declaration, node) {
                var currentNode = ts.getParseTreeNode(node);
                if (!currentNode || currentNode === declaration || currentNode.end <= declaration.pos || currentNode.pos >= declaration.end) {
                    // if the node has no correlation to a parse tree node, its definitely not
                    // part of the body.
                    // if the node is outside of the document range of the declaration, its
                    // definitely not part of the body.
                    return false;
                }
                var blockScope = ts.getEnclosingBlockScopeContainer(declaration);
                while (currentNode) {
                    if (currentNode === blockScope || currentNode === declaration) {
                        // if we are in the enclosing block scope of the declaration, we are definitely
                        // not inside the class body.
                        return false;
                    }
                    if (ts.isClassElement(currentNode) && currentNode.parent === declaration) {
                        return true;
                    }
                    currentNode = currentNode.parent;
                }
                return false;
            }
            /**
             * Substitutes `this` when contained within an arrow function.
             *
             * @param node The ThisKeyword node.
             */
            function substituteThisKeyword(node) {
                if (enabledSubstitutions & 1 /* CapturedThis */
                    && hierarchyFacts & 16 /* CapturesThis */) {
                    return ts.setTextRange(factory.createUniqueName("_this", 16 /* Optimistic */ | 32 /* FileLevel */), node);
                }
                return node;
            }
            function getClassMemberPrefix(node, member) {
                return ts.isStatic(member)
                    ? factory.getInternalName(node)
                    : factory.createPropertyAccessExpression(factory.getInternalName(node), "prototype");
            }
            function hasSynthesizedDefaultSuperCall(constructor, hasExtendsClause) {
                if (!constructor || !hasExtendsClause) {
                    return false;
                }
                if (ts.some(constructor.parameters)) {
                    return false;
                }
                var statement = ts.firstOrUndefined(constructor.body.statements);
                if (!statement || !ts.nodeIsSynthesized(statement) || statement.kind !== 236 /* ExpressionStatement */) {
                    return false;
                }
                var statementExpression = statement.expression;
                if (!ts.nodeIsSynthesized(statementExpression) || statementExpression.kind !== 206 /* CallExpression */) {
                    return false;
                }
                var callTarget = statementExpression.expression;
                if (!ts.nodeIsSynthesized(callTarget) || callTarget.kind !== 106 /* SuperKeyword */) {
                    return false;
                }
                var callArgument = ts.singleOrUndefined(statementExpression.arguments);
                if (!callArgument || !ts.nodeIsSynthesized(callArgument) || callArgument.kind !== 223 /* SpreadElement */) {
                    return false;
                }
                var expression = callArgument.expression;
                return ts.isIdentifier(expression) && expression.escapedText === "arguments";
            }
        }
        ts.transformES2015 = transformES2015;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        /**
         * Transforms ES5 syntax into ES3 syntax.
         *
         * @param context Context and state information for the transformation.
         */
        function transformES5(context) {
            var factory = context.factory;
            var compilerOptions = context.getCompilerOptions();
            // enable emit notification only if using --jsx preserve or react-native
            var previousOnEmitNode;
            var noSubstitution;
            if (compilerOptions.jsx === 1 /* Preserve */ || compilerOptions.jsx === 3 /* ReactNative */) {
                previousOnEmitNode = context.onEmitNode;
                context.onEmitNode = onEmitNode;
                context.enableEmitNotification(278 /* JsxOpeningElement */);
                context.enableEmitNotification(279 /* JsxClosingElement */);
                context.enableEmitNotification(277 /* JsxSelfClosingElement */);
                noSubstitution = [];
            }
            var previousOnSubstituteNode = context.onSubstituteNode;
            context.onSubstituteNode = onSubstituteNode;
            context.enableSubstitution(204 /* PropertyAccessExpression */);
            context.enableSubstitution(291 /* PropertyAssignment */);
            return ts.chainBundle(context, transformSourceFile);
            /**
             * Transforms an ES5 source file to ES3.
             *
             * @param node A SourceFile
             */
            function transformSourceFile(node) {
                return node;
            }
            /**
             * Called by the printer just before a node is printed.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to emit.
             * @param emitCallback A callback used to emit the node.
             */
            function onEmitNode(hint, node, emitCallback) {
                switch (node.kind) {
                    case 278 /* JsxOpeningElement */:
                    case 279 /* JsxClosingElement */:
                    case 277 /* JsxSelfClosingElement */:
                        var tagName = node.tagName;
                        noSubstitution[ts.getOriginalNodeId(tagName)] = true;
                        break;
                }
                previousOnEmitNode(hint, node, emitCallback);
            }
            /**
             * Hooks node substitutions.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to substitute.
             */
            function onSubstituteNode(hint, node) {
                if (node.id && noSubstitution && noSubstitution[node.id]) {
                    return previousOnSubstituteNode(hint, node);
                }
                node = previousOnSubstituteNode(hint, node);
                if (ts.isPropertyAccessExpression(node)) {
                    return substitutePropertyAccessExpression(node);
                }
                else if (ts.isPropertyAssignment(node)) {
                    return substitutePropertyAssignment(node);
                }
                return node;
            }
            /**
             * Substitutes a PropertyAccessExpression whose name is a reserved word.
             *
             * @param node A PropertyAccessExpression
             */
            function substitutePropertyAccessExpression(node) {
                if (ts.isPrivateIdentifier(node.name)) {
                    return node;
                }
                var literalName = trySubstituteReservedName(node.name);
                if (literalName) {
                    return ts.setTextRange(factory.createElementAccessExpression(node.expression, literalName), node);
                }
                return node;
            }
            /**
             * Substitutes a PropertyAssignment whose name is a reserved word.
             *
             * @param node A PropertyAssignment
             */
            function substitutePropertyAssignment(node) {
                var literalName = ts.isIdentifier(node.name) && trySubstituteReservedName(node.name);
                if (literalName) {
                    return factory.updatePropertyAssignment(node, literalName, node.initializer);
                }
                return node;
            }
            /**
             * If an identifier name is a reserved word, returns a string literal for the name.
             *
             * @param name An Identifier
             */
            function trySubstituteReservedName(name) {
                var token = name.originalKeywordKind || (ts.nodeIsSynthesized(name) ? ts.stringToToken(ts.idText(name)) : undefined);
                if (token !== undefined && token >= 81 /* FirstReservedWord */ && token <= 116 /* LastReservedWord */) {
                    return ts.setTextRange(factory.createStringLiteralFromNode(name), name);
                }
                return undefined;
            }
        }
        ts.transformES5 = transformES5;
    })(ts || (ts = {}));
    // Transforms generator functions into a compatible ES5 representation with similar runtime
    // semantics. This is accomplished by first transforming the body of each generator
    // function into an intermediate representation that is the compiled into a JavaScript
    // switch statement.
    //
    // Many functions in this transformer will contain comments indicating the expected
    // intermediate representation. For illustrative purposes, the following intermediate
    // language is used to define this intermediate representation:
    //
    //  .nop                            - Performs no operation.
    //  .local NAME, ...                - Define local variable declarations.
    //  .mark LABEL                     - Mark the location of a label.
    //  .br LABEL                       - Jump to a label. If jumping out of a protected
    //                                    region, all .finally blocks are executed.
    //  .brtrue LABEL, (x)              - Jump to a label IIF the expression `x` is truthy.
    //                                    If jumping out of a protected region, all .finally
    //                                    blocks are executed.
    //  .brfalse LABEL, (x)             - Jump to a label IIF the expression `x` is falsey.
    //                                    If jumping out of a protected region, all .finally
    //                                    blocks are executed.
    //  .yield (x)                      - Yield the value of the optional expression `x`.
    //                                    Resume at the next label.
    //  .yieldstar (x)                  - Delegate yield to the value of the optional
    //                                    expression `x`. Resume at the next label.
    //                                    NOTE: `x` must be an Iterator, not an Iterable.
    //  .loop CONTINUE, BREAK           - Marks the beginning of a loop. Any "continue" or
    //                                    "break" abrupt completions jump to the CONTINUE or
    //                                    BREAK labels, respectively.
    //  .endloop                        - Marks the end of a loop.
    //  .with (x)                       - Marks the beginning of a WithStatement block, using
    //                                    the supplied expression.
    //  .endwith                        - Marks the end of a WithStatement.
    //  .switch                         - Marks the beginning of a SwitchStatement.
    //  .endswitch                      - Marks the end of a SwitchStatement.
    //  .labeled NAME                   - Marks the beginning of a LabeledStatement with the
    //                                    supplied name.
    //  .endlabeled                     - Marks the end of a LabeledStatement.
    //  .try TRY, CATCH, FINALLY, END   - Marks the beginning of a protected region, and the
    //                                    labels for each block.
    //  .catch (x)                      - Marks the beginning of a catch block.
    //  .finally                        - Marks the beginning of a finally block.
    //  .endfinally                     - Marks the end of a finally block.
    //  .endtry                         - Marks the end of a protected region.
    //  .throw (x)                      - Throws the value of the expression `x`.
    //  .return (x)                     - Returns the value of the expression `x`.
    //
    // In addition, the illustrative intermediate representation introduces some special
    // variables:
    //
    //  %sent%                          - Either returns the next value sent to the generator,
    //                                    returns the result of a delegated yield, or throws
    //                                    the exception sent to the generator.
    //  %error%                         - Returns the value of the current exception in a
    //                                    catch block.
    //
    // This intermediate representation is then compiled into JavaScript syntax. The resulting
    // compilation output looks something like the following:
    //
    //  function f() {
    //      var /*locals*/;
    //      /*functions*/
    //      return __generator(function (state) {
    //          switch (state.label) {
    //              /*cases per label*/
    //          }
    //      });
    //  }
    //
    // Each of the above instructions corresponds to JavaScript emit similar to the following:
    //
    //  .local NAME                   | var NAME;
    // -------------------------------|----------------------------------------------
    //  .mark LABEL                   | case LABEL:
    // -------------------------------|----------------------------------------------
    //  .br LABEL                     |     return [3 /*break*/, LABEL];
    // -------------------------------|----------------------------------------------
    //  .brtrue LABEL, (x)            |     if (x) return [3 /*break*/, LABEL];
    // -------------------------------|----------------------------------------------
    //  .brfalse LABEL, (x)           |     if (!(x)) return [3, /*break*/, LABEL];
    // -------------------------------|----------------------------------------------
    //  .yield (x)                    |     return [4 /*yield*/, x];
    //  .mark RESUME                  | case RESUME:
    //      a = %sent%;               |     a = state.sent();
    // -------------------------------|----------------------------------------------
    //  .yieldstar (x)                |     return [5 /*yield**/, x];
    //  .mark RESUME                  | case RESUME:
    //      a = %sent%;               |     a = state.sent();
    // -------------------------------|----------------------------------------------
    //  .with (_a)                    |     with (_a) {
    //      a();                      |         a();
    //                                |     }
    //                                |     state.label = LABEL;
    //  .mark LABEL                   | case LABEL:
    //                                |     with (_a) {
    //      b();                      |         b();
    //                                |     }
    //  .endwith                      |
    // -------------------------------|----------------------------------------------
    //                                | case 0:
    //                                |     state.trys = [];
    //                                | ...
    //  .try TRY, CATCH, FINALLY, END |
    //  .mark TRY                     | case TRY:
    //                                |     state.trys.push([TRY, CATCH, FINALLY, END]);
    //  .nop                          |
    //      a();                      |     a();
    //  .br END                       |     return [3 /*break*/, END];
    //  .catch (e)                    |
    //  .mark CATCH                   | case CATCH:
    //                                |     e = state.sent();
    //      b();                      |     b();
    //  .br END                       |     return [3 /*break*/, END];
    //  .finally                      |
    //  .mark FINALLY                 | case FINALLY:
    //      c();                      |     c();
    //  .endfinally                   |     return [7 /*endfinally*/];
    //  .endtry                       |
    //  .mark END                     | case END:
    /*@internal*/
    var ts;
    (function (ts) {
        var OpCode;
        (function (OpCode) {
            OpCode[OpCode["Nop"] = 0] = "Nop";
            OpCode[OpCode["Statement"] = 1] = "Statement";
            OpCode[OpCode["Assign"] = 2] = "Assign";
            OpCode[OpCode["Break"] = 3] = "Break";
            OpCode[OpCode["BreakWhenTrue"] = 4] = "BreakWhenTrue";
            OpCode[OpCode["BreakWhenFalse"] = 5] = "BreakWhenFalse";
            OpCode[OpCode["Yield"] = 6] = "Yield";
            OpCode[OpCode["YieldStar"] = 7] = "YieldStar";
            OpCode[OpCode["Return"] = 8] = "Return";
            OpCode[OpCode["Throw"] = 9] = "Throw";
            OpCode[OpCode["Endfinally"] = 10] = "Endfinally"; // Marks the end of a `finally` block
        })(OpCode || (OpCode = {}));
        // whether a generated code block is opening or closing at the current operation for a FunctionBuilder
        var BlockAction;
        (function (BlockAction) {
            BlockAction[BlockAction["Open"] = 0] = "Open";
            BlockAction[BlockAction["Close"] = 1] = "Close";
        })(BlockAction || (BlockAction = {}));
        // the kind for a generated code block in a FunctionBuilder
        var CodeBlockKind;
        (function (CodeBlockKind) {
            CodeBlockKind[CodeBlockKind["Exception"] = 0] = "Exception";
            CodeBlockKind[CodeBlockKind["With"] = 1] = "With";
            CodeBlockKind[CodeBlockKind["Switch"] = 2] = "Switch";
            CodeBlockKind[CodeBlockKind["Loop"] = 3] = "Loop";
            CodeBlockKind[CodeBlockKind["Labeled"] = 4] = "Labeled";
        })(CodeBlockKind || (CodeBlockKind = {}));
        // the state for a generated code exception block
        var ExceptionBlockState;
        (function (ExceptionBlockState) {
            ExceptionBlockState[ExceptionBlockState["Try"] = 0] = "Try";
            ExceptionBlockState[ExceptionBlockState["Catch"] = 1] = "Catch";
            ExceptionBlockState[ExceptionBlockState["Finally"] = 2] = "Finally";
            ExceptionBlockState[ExceptionBlockState["Done"] = 3] = "Done";
        })(ExceptionBlockState || (ExceptionBlockState = {}));
        // NOTE: changes to this enum should be reflected in the __generator helper.
        var Instruction;
        (function (Instruction) {
            Instruction[Instruction["Next"] = 0] = "Next";
            Instruction[Instruction["Throw"] = 1] = "Throw";
            Instruction[Instruction["Return"] = 2] = "Return";
            Instruction[Instruction["Break"] = 3] = "Break";
            Instruction[Instruction["Yield"] = 4] = "Yield";
            Instruction[Instruction["YieldStar"] = 5] = "YieldStar";
            Instruction[Instruction["Catch"] = 6] = "Catch";
            Instruction[Instruction["Endfinally"] = 7] = "Endfinally";
        })(Instruction || (Instruction = {}));
        function getInstructionName(instruction) {
            switch (instruction) {
                case 2 /* Return */: return "return";
                case 3 /* Break */: return "break";
                case 4 /* Yield */: return "yield";
                case 5 /* YieldStar */: return "yield*";
                case 7 /* Endfinally */: return "endfinally";
                default: return undefined; // TODO: GH#18217
            }
        }
        function transformGenerators(context) {
            var factory = context.factory, emitHelpers = context.getEmitHelperFactory, resumeLexicalEnvironment = context.resumeLexicalEnvironment, endLexicalEnvironment = context.endLexicalEnvironment, hoistFunctionDeclaration = context.hoistFunctionDeclaration, hoistVariableDeclaration = context.hoistVariableDeclaration;
            var compilerOptions = context.getCompilerOptions();
            var languageVersion = ts.getEmitScriptTarget(compilerOptions);
            var resolver = context.getEmitResolver();
            var previousOnSubstituteNode = context.onSubstituteNode;
            context.onSubstituteNode = onSubstituteNode;
            var renamedCatchVariables;
            var renamedCatchVariableDeclarations;
            var inGeneratorFunctionBody;
            var inStatementContainingYield;
            // The following three arrays store information about generated code blocks.
            // All three arrays are correlated by their index. This approach is used over allocating
            // objects to store the same information to avoid GC overhead.
            //
            var blocks; // Information about the code block
            var blockOffsets; // The operation offset at which a code block begins or ends
            var blockActions; // Whether the code block is opened or closed
            var blockStack; // A stack of currently open code blocks
            // Labels are used to mark locations in the code that can be the target of a Break (jump)
            // operation. These are translated into case clauses in a switch statement.
            // The following two arrays are correlated by their index. This approach is used over
            // allocating objects to store the same information to avoid GC overhead.
            //
            var labelOffsets; // The operation offset at which the label is defined.
            var labelExpressions; // The NumericLiteral nodes bound to each label.
            var nextLabelId = 1; // The next label id to use.
            // Operations store information about generated code for the function body. This
            // Includes things like statements, assignments, breaks (jumps), and yields.
            // The following three arrays are correlated by their index. This approach is used over
            // allocating objects to store the same information to avoid GC overhead.
            //
            var operations; // The operation to perform.
            var operationArguments; // The arguments to the operation.
            var operationLocations; // The source map location for the operation.
            var state; // The name of the state object used by the generator at runtime.
            // The following variables store information used by the `build` function:
            //
            var blockIndex = 0; // The index of the current block.
            var labelNumber = 0; // The current label number.
            var labelNumbers;
            var lastOperationWasAbrupt; // Indicates whether the last operation was abrupt (break/continue).
            var lastOperationWasCompletion; // Indicates whether the last operation was a completion (return/throw).
            var clauses; // The case clauses generated for labels.
            var statements; // The statements for the current label.
            var exceptionBlockStack; // A stack of containing exception blocks.
            var currentExceptionBlock; // The current exception block.
            var withBlockStack; // A stack containing `with` blocks.
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                if (node.isDeclarationFile || (node.transformFlags & 1024 /* ContainsGenerator */) === 0) {
                    return node;
                }
                var visited = ts.visitEachChild(node, visitor, context);
                ts.addEmitHelpers(visited, context.readEmitHelpers());
                return visited;
            }
            /**
             * Visits a node.
             *
             * @param node The node to visit.
             */
            function visitor(node) {
                var transformFlags = node.transformFlags;
                if (inStatementContainingYield) {
                    return visitJavaScriptInStatementContainingYield(node);
                }
                else if (inGeneratorFunctionBody) {
                    return visitJavaScriptInGeneratorFunctionBody(node);
                }
                else if (ts.isFunctionLikeDeclaration(node) && node.asteriskToken) {
                    return visitGenerator(node);
                }
                else if (transformFlags & 1024 /* ContainsGenerator */) {
                    return ts.visitEachChild(node, visitor, context);
                }
                else {
                    return node;
                }
            }
            /**
             * Visits a node that is contained within a statement that contains yield.
             *
             * @param node The node to visit.
             */
            function visitJavaScriptInStatementContainingYield(node) {
                switch (node.kind) {
                    case 238 /* DoStatement */:
                        return visitDoStatement(node);
                    case 239 /* WhileStatement */:
                        return visitWhileStatement(node);
                    case 247 /* SwitchStatement */:
                        return visitSwitchStatement(node);
                    case 248 /* LabeledStatement */:
                        return visitLabeledStatement(node);
                    default:
                        return visitJavaScriptInGeneratorFunctionBody(node);
                }
            }
            /**
             * Visits a node that is contained within a generator function.
             *
             * @param node The node to visit.
             */
            function visitJavaScriptInGeneratorFunctionBody(node) {
                switch (node.kind) {
                    case 254 /* FunctionDeclaration */:
                        return visitFunctionDeclaration(node);
                    case 211 /* FunctionExpression */:
                        return visitFunctionExpression(node);
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return visitAccessorDeclaration(node);
                    case 235 /* VariableStatement */:
                        return visitVariableStatement(node);
                    case 240 /* ForStatement */:
                        return visitForStatement(node);
                    case 241 /* ForInStatement */:
                        return visitForInStatement(node);
                    case 244 /* BreakStatement */:
                        return visitBreakStatement(node);
                    case 243 /* ContinueStatement */:
                        return visitContinueStatement(node);
                    case 245 /* ReturnStatement */:
                        return visitReturnStatement(node);
                    default:
                        if (node.transformFlags & 524288 /* ContainsYield */) {
                            return visitJavaScriptContainingYield(node);
                        }
                        else if (node.transformFlags & (1024 /* ContainsGenerator */ | 2097152 /* ContainsHoistedDeclarationOrCompletion */)) {
                            return ts.visitEachChild(node, visitor, context);
                        }
                        else {
                            return node;
                        }
                }
            }
            /**
             * Visits a node that contains a YieldExpression.
             *
             * @param node The node to visit.
             */
            function visitJavaScriptContainingYield(node) {
                switch (node.kind) {
                    case 219 /* BinaryExpression */:
                        return visitBinaryExpression(node);
                    case 346 /* CommaListExpression */:
                        return visitCommaListExpression(node);
                    case 220 /* ConditionalExpression */:
                        return visitConditionalExpression(node);
                    case 222 /* YieldExpression */:
                        return visitYieldExpression(node);
                    case 202 /* ArrayLiteralExpression */:
                        return visitArrayLiteralExpression(node);
                    case 203 /* ObjectLiteralExpression */:
                        return visitObjectLiteralExpression(node);
                    case 205 /* ElementAccessExpression */:
                        return visitElementAccessExpression(node);
                    case 206 /* CallExpression */:
                        return visitCallExpression(node);
                    case 207 /* NewExpression */:
                        return visitNewExpression(node);
                    default:
                        return ts.visitEachChild(node, visitor, context);
                }
            }
            /**
             * Visits a generator function.
             *
             * @param node The node to visit.
             */
            function visitGenerator(node) {
                switch (node.kind) {
                    case 254 /* FunctionDeclaration */:
                        return visitFunctionDeclaration(node);
                    case 211 /* FunctionExpression */:
                        return visitFunctionExpression(node);
                    default:
                        return ts.Debug.failBadSyntaxKind(node);
                }
            }
            /**
             * Visits a function declaration.
             *
             * This will be called when one of the following conditions are met:
             * - The function declaration is a generator function.
             * - The function declaration is contained within the body of a generator function.
             *
             * @param node The node to visit.
             */
            function visitFunctionDeclaration(node) {
                // Currently, we only support generators that were originally async functions.
                if (node.asteriskToken) {
                    node = ts.setOriginalNode(ts.setTextRange(factory.createFunctionDeclaration(
                    /*decorators*/ undefined, node.modifiers, 
                    /*asteriskToken*/ undefined, node.name, 
                    /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                    /*type*/ undefined, transformGeneratorFunctionBody(node.body)), 
                    /*location*/ node), node);
                }
                else {
                    var savedInGeneratorFunctionBody = inGeneratorFunctionBody;
                    var savedInStatementContainingYield = inStatementContainingYield;
                    inGeneratorFunctionBody = false;
                    inStatementContainingYield = false;
                    node = ts.visitEachChild(node, visitor, context);
                    inGeneratorFunctionBody = savedInGeneratorFunctionBody;
                    inStatementContainingYield = savedInStatementContainingYield;
                }
                if (inGeneratorFunctionBody) {
                    // Function declarations in a generator function body are hoisted
                    // to the top of the lexical scope and elided from the current statement.
                    hoistFunctionDeclaration(node);
                    return undefined;
                }
                else {
                    return node;
                }
            }
            /**
             * Visits a function expression.
             *
             * This will be called when one of the following conditions are met:
             * - The function expression is a generator function.
             * - The function expression is contained within the body of a generator function.
             *
             * @param node The node to visit.
             */
            function visitFunctionExpression(node) {
                // Currently, we only support generators that were originally async functions.
                if (node.asteriskToken) {
                    node = ts.setOriginalNode(ts.setTextRange(factory.createFunctionExpression(
                    /*modifiers*/ undefined, 
                    /*asteriskToken*/ undefined, node.name, 
                    /*typeParameters*/ undefined, ts.visitParameterList(node.parameters, visitor, context), 
                    /*type*/ undefined, transformGeneratorFunctionBody(node.body)), 
                    /*location*/ node), node);
                }
                else {
                    var savedInGeneratorFunctionBody = inGeneratorFunctionBody;
                    var savedInStatementContainingYield = inStatementContainingYield;
                    inGeneratorFunctionBody = false;
                    inStatementContainingYield = false;
                    node = ts.visitEachChild(node, visitor, context);
                    inGeneratorFunctionBody = savedInGeneratorFunctionBody;
                    inStatementContainingYield = savedInStatementContainingYield;
                }
                return node;
            }
            /**
             * Visits a get or set accessor declaration.
             *
             * This will be called when one of the following conditions are met:
             * - The accessor is contained within the body of a generator function.
             *
             * @param node The node to visit.
             */
            function visitAccessorDeclaration(node) {
                var savedInGeneratorFunctionBody = inGeneratorFunctionBody;
                var savedInStatementContainingYield = inStatementContainingYield;
                inGeneratorFunctionBody = false;
                inStatementContainingYield = false;
                node = ts.visitEachChild(node, visitor, context);
                inGeneratorFunctionBody = savedInGeneratorFunctionBody;
                inStatementContainingYield = savedInStatementContainingYield;
                return node;
            }
            /**
             * Transforms the body of a generator function declaration.
             *
             * @param node The function body to transform.
             */
            function transformGeneratorFunctionBody(body) {
                // Save existing generator state
                var statements = [];
                var savedInGeneratorFunctionBody = inGeneratorFunctionBody;
                var savedInStatementContainingYield = inStatementContainingYield;
                var savedBlocks = blocks;
                var savedBlockOffsets = blockOffsets;
                var savedBlockActions = blockActions;
                var savedBlockStack = blockStack;
                var savedLabelOffsets = labelOffsets;
                var savedLabelExpressions = labelExpressions;
                var savedNextLabelId = nextLabelId;
                var savedOperations = operations;
                var savedOperationArguments = operationArguments;
                var savedOperationLocations = operationLocations;
                var savedState = state;
                // Initialize generator state
                inGeneratorFunctionBody = true;
                inStatementContainingYield = false;
                blocks = undefined;
                blockOffsets = undefined;
                blockActions = undefined;
                blockStack = undefined;
                labelOffsets = undefined;
                labelExpressions = undefined;
                nextLabelId = 1;
                operations = undefined;
                operationArguments = undefined;
                operationLocations = undefined;
                state = factory.createTempVariable(/*recordTempVariable*/ undefined);
                // Build the generator
                resumeLexicalEnvironment();
                var statementOffset = factory.copyPrologue(body.statements, statements, /*ensureUseStrict*/ false, visitor);
                transformAndEmitStatements(body.statements, statementOffset);
                var buildResult = build();
                ts.insertStatementsAfterStandardPrologue(statements, endLexicalEnvironment());
                statements.push(factory.createReturnStatement(buildResult));
                // Restore previous generator state
                inGeneratorFunctionBody = savedInGeneratorFunctionBody;
                inStatementContainingYield = savedInStatementContainingYield;
                blocks = savedBlocks;
                blockOffsets = savedBlockOffsets;
                blockActions = savedBlockActions;
                blockStack = savedBlockStack;
                labelOffsets = savedLabelOffsets;
                labelExpressions = savedLabelExpressions;
                nextLabelId = savedNextLabelId;
                operations = savedOperations;
                operationArguments = savedOperationArguments;
                operationLocations = savedOperationLocations;
                state = savedState;
                return ts.setTextRange(factory.createBlock(statements, body.multiLine), body);
            }
            /**
             * Visits a variable statement.
             *
             * This will be called when one of the following conditions are met:
             * - The variable statement is contained within the body of a generator function.
             *
             * @param node The node to visit.
             */
            function visitVariableStatement(node) {
                if (node.transformFlags & 524288 /* ContainsYield */) {
                    transformAndEmitVariableDeclarationList(node.declarationList);
                    return undefined;
                }
                else {
                    // Do not hoist custom prologues.
                    if (ts.getEmitFlags(node) & 1048576 /* CustomPrologue */) {
                        return node;
                    }
                    for (var _i = 0, _a = node.declarationList.declarations; _i < _a.length; _i++) {
                        var variable = _a[_i];
                        hoistVariableDeclaration(variable.name);
                    }
                    var variables = ts.getInitializedVariables(node.declarationList);
                    if (variables.length === 0) {
                        return undefined;
                    }
                    return ts.setSourceMapRange(factory.createExpressionStatement(factory.inlineExpressions(ts.map(variables, transformInitializedVariable))), node);
                }
            }
            /**
             * Visits a binary expression.
             *
             * This will be called when one of the following conditions are met:
             * - The node contains a YieldExpression.
             *
             * @param node The node to visit.
             */
            function visitBinaryExpression(node) {
                var assoc = ts.getExpressionAssociativity(node);
                switch (assoc) {
                    case 0 /* Left */:
                        return visitLeftAssociativeBinaryExpression(node);
                    case 1 /* Right */:
                        return visitRightAssociativeBinaryExpression(node);
                    default:
                        return ts.Debug.assertNever(assoc);
                }
            }
            /**
             * Visits a right-associative binary expression containing `yield`.
             *
             * @param node The node to visit.
             */
            function visitRightAssociativeBinaryExpression(node) {
                var left = node.left, right = node.right;
                if (containsYield(right)) {
                    var target = void 0;
                    switch (left.kind) {
                        case 204 /* PropertyAccessExpression */:
                            // [source]
                            //      a.b = yield;
                            //
                            // [intermediate]
                            //  .local _a
                            //      _a = a;
                            //  .yield resumeLabel
                            //  .mark resumeLabel
                            //      _a.b = %sent%;
                            target = factory.updatePropertyAccessExpression(left, cacheExpression(ts.visitNode(left.expression, visitor, ts.isLeftHandSideExpression)), left.name);
                            break;
                        case 205 /* ElementAccessExpression */:
                            // [source]
                            //      a[b] = yield;
                            //
                            // [intermediate]
                            //  .local _a, _b
                            //      _a = a;
                            //      _b = b;
                            //  .yield resumeLabel
                            //  .mark resumeLabel
                            //      _a[_b] = %sent%;
                            target = factory.updateElementAccessExpression(left, cacheExpression(ts.visitNode(left.expression, visitor, ts.isLeftHandSideExpression)), cacheExpression(ts.visitNode(left.argumentExpression, visitor, ts.isExpression)));
                            break;
                        default:
                            target = ts.visitNode(left, visitor, ts.isExpression);
                            break;
                    }
                    var operator = node.operatorToken.kind;
                    if (ts.isCompoundAssignment(operator)) {
                        return ts.setTextRange(factory.createAssignment(target, ts.setTextRange(factory.createBinaryExpression(cacheExpression(target), ts.getNonAssignmentOperatorForCompoundAssignment(operator), ts.visitNode(right, visitor, ts.isExpression)), node)), node);
                    }
                    else {
                        return factory.updateBinaryExpression(node, target, node.operatorToken, ts.visitNode(right, visitor, ts.isExpression));
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitLeftAssociativeBinaryExpression(node) {
                if (containsYield(node.right)) {
                    if (ts.isLogicalOperator(node.operatorToken.kind)) {
                        return visitLogicalBinaryExpression(node);
                    }
                    else if (node.operatorToken.kind === 27 /* CommaToken */) {
                        return visitCommaExpression(node);
                    }
                    // [source]
                    //      a() + (yield) + c()
                    //
                    // [intermediate]
                    //  .local _a
                    //      _a = a();
                    //  .yield resumeLabel
                    //      _a + %sent% + c()
                    return factory.updateBinaryExpression(node, cacheExpression(ts.visitNode(node.left, visitor, ts.isExpression)), node.operatorToken, ts.visitNode(node.right, visitor, ts.isExpression));
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visits a comma expression containing `yield`.
             *
             * @param node The node to visit.
             */
            function visitCommaExpression(node) {
                // [source]
                //      x = a(), yield, b();
                //
                // [intermediate]
                //      a();
                //  .yield resumeLabel
                //  .mark resumeLabel
                //      x = %sent%, b();
                var pendingExpressions = [];
                visit(node.left);
                visit(node.right);
                return factory.inlineExpressions(pendingExpressions);
                function visit(node) {
                    if (ts.isBinaryExpression(node) && node.operatorToken.kind === 27 /* CommaToken */) {
                        visit(node.left);
                        visit(node.right);
                    }
                    else {
                        if (containsYield(node) && pendingExpressions.length > 0) {
                            emitWorker(1 /* Statement */, [factory.createExpressionStatement(factory.inlineExpressions(pendingExpressions))]);
                            pendingExpressions = [];
                        }
                        pendingExpressions.push(ts.visitNode(node, visitor, ts.isExpression));
                    }
                }
            }
            /**
             * Visits a comma-list expression.
             *
             * @param node The node to visit.
             */
            function visitCommaListExpression(node) {
                // flattened version of `visitCommaExpression`
                var pendingExpressions = [];
                for (var _i = 0, _a = node.elements; _i < _a.length; _i++) {
                    var elem = _a[_i];
                    if (ts.isBinaryExpression(elem) && elem.operatorToken.kind === 27 /* CommaToken */) {
                        pendingExpressions.push(visitCommaExpression(elem));
                    }
                    else {
                        if (containsYield(elem) && pendingExpressions.length > 0) {
                            emitWorker(1 /* Statement */, [factory.createExpressionStatement(factory.inlineExpressions(pendingExpressions))]);
                            pendingExpressions = [];
                        }
                        pendingExpressions.push(ts.visitNode(elem, visitor, ts.isExpression));
                    }
                }
                return factory.inlineExpressions(pendingExpressions);
            }
            /**
             * Visits a logical binary expression containing `yield`.
             *
             * @param node A node to visit.
             */
            function visitLogicalBinaryExpression(node) {
                // Logical binary expressions (`&&` and `||`) are shortcutting expressions and need
                // to be transformed as such:
                //
                // [source]
                //      x = a() && yield;
                //
                // [intermediate]
                //  .local _a
                //      _a = a();
                //  .brfalse resultLabel, (_a)
                //  .yield resumeLabel
                //  .mark resumeLabel
                //      _a = %sent%;
                //  .mark resultLabel
                //      x = _a;
                //
                // [source]
                //      x = a() || yield;
                //
                // [intermediate]
                //  .local _a
                //      _a = a();
                //  .brtrue resultLabel, (_a)
                //  .yield resumeLabel
                //  .mark resumeLabel
                //      _a = %sent%;
                //  .mark resultLabel
                //      x = _a;
                var resultLabel = defineLabel();
                var resultLocal = declareLocal();
                emitAssignment(resultLocal, ts.visitNode(node.left, visitor, ts.isExpression), /*location*/ node.left);
                if (node.operatorToken.kind === 55 /* AmpersandAmpersandToken */) {
                    // Logical `&&` shortcuts when the left-hand operand is falsey.
                    emitBreakWhenFalse(resultLabel, resultLocal, /*location*/ node.left);
                }
                else {
                    // Logical `||` shortcuts when the left-hand operand is truthy.
                    emitBreakWhenTrue(resultLabel, resultLocal, /*location*/ node.left);
                }
                emitAssignment(resultLocal, ts.visitNode(node.right, visitor, ts.isExpression), /*location*/ node.right);
                markLabel(resultLabel);
                return resultLocal;
            }
            /**
             * Visits a conditional expression containing `yield`.
             *
             * @param node The node to visit.
             */
            function visitConditionalExpression(node) {
                // [source]
                //      x = a() ? yield : b();
                //
                // [intermediate]
                //  .local _a
                //  .brfalse whenFalseLabel, (a())
                //  .yield resumeLabel
                //  .mark resumeLabel
                //      _a = %sent%;
                //  .br resultLabel
                //  .mark whenFalseLabel
                //      _a = b();
                //  .mark resultLabel
                //      x = _a;
                // We only need to perform a specific transformation if a `yield` expression exists
                // in either the `whenTrue` or `whenFalse` branches.
                // A `yield` in the condition will be handled by the normal visitor.
                if (containsYield(node.whenTrue) || containsYield(node.whenFalse)) {
                    var whenFalseLabel = defineLabel();
                    var resultLabel = defineLabel();
                    var resultLocal = declareLocal();
                    emitBreakWhenFalse(whenFalseLabel, ts.visitNode(node.condition, visitor, ts.isExpression), /*location*/ node.condition);
                    emitAssignment(resultLocal, ts.visitNode(node.whenTrue, visitor, ts.isExpression), /*location*/ node.whenTrue);
                    emitBreak(resultLabel);
                    markLabel(whenFalseLabel);
                    emitAssignment(resultLocal, ts.visitNode(node.whenFalse, visitor, ts.isExpression), /*location*/ node.whenFalse);
                    markLabel(resultLabel);
                    return resultLocal;
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visits a `yield` expression.
             *
             * @param node The node to visit.
             */
            function visitYieldExpression(node) {
                // [source]
                //      x = yield a();
                //
                // [intermediate]
                //  .yield resumeLabel, (a())
                //  .mark resumeLabel
                //      x = %sent%;
                var resumeLabel = defineLabel();
                var expression = ts.visitNode(node.expression, visitor, ts.isExpression);
                if (node.asteriskToken) {
                    // NOTE: `expression` must be defined for `yield*`.
                    var iterator = (ts.getEmitFlags(node.expression) & 8388608 /* Iterator */) === 0
                        ? ts.setTextRange(emitHelpers().createValuesHelper(expression), node)
                        : expression;
                    emitYieldStar(iterator, /*location*/ node);
                }
                else {
                    emitYield(expression, /*location*/ node);
                }
                markLabel(resumeLabel);
                return createGeneratorResume(/*location*/ node);
            }
            /**
             * Visits an ArrayLiteralExpression that contains a YieldExpression.
             *
             * @param node The node to visit.
             */
            function visitArrayLiteralExpression(node) {
                return visitElements(node.elements, /*leadingElement*/ undefined, /*location*/ undefined, node.multiLine);
            }
            /**
             * Visits an array of expressions containing one or more YieldExpression nodes
             * and returns an expression for the resulting value.
             *
             * @param elements The elements to visit.
             * @param multiLine Whether array literals created should be emitted on multiple lines.
             */
            function visitElements(elements, leadingElement, location, multiLine) {
                // [source]
                //      ar = [1, yield, 2];
                //
                // [intermediate]
                //  .local _a
                //      _a = [1];
                //  .yield resumeLabel
                //  .mark resumeLabel
                //      ar = _a.concat([%sent%, 2]);
                var numInitialElements = countInitialNodesWithoutYield(elements);
                var temp;
                if (numInitialElements > 0) {
                    temp = declareLocal();
                    var initialElements = ts.visitNodes(elements, visitor, ts.isExpression, 0, numInitialElements);
                    emitAssignment(temp, factory.createArrayLiteralExpression(leadingElement
                        ? __spreadArray([leadingElement], initialElements, true) : initialElements));
                    leadingElement = undefined;
                }
                var expressions = ts.reduceLeft(elements, reduceElement, [], numInitialElements);
                return temp
                    ? factory.createArrayConcatCall(temp, [factory.createArrayLiteralExpression(expressions, multiLine)])
                    : ts.setTextRange(factory.createArrayLiteralExpression(leadingElement ? __spreadArray([leadingElement], expressions, true) : expressions, multiLine), location);
                function reduceElement(expressions, element) {
                    if (containsYield(element) && expressions.length > 0) {
                        var hasAssignedTemp = temp !== undefined;
                        if (!temp) {
                            temp = declareLocal();
                        }
                        emitAssignment(temp, hasAssignedTemp
                            ? factory.createArrayConcatCall(temp, [factory.createArrayLiteralExpression(expressions, multiLine)])
                            : factory.createArrayLiteralExpression(leadingElement ? __spreadArray([leadingElement], expressions, true) : expressions, multiLine));
                        leadingElement = undefined;
                        expressions = [];
                    }
                    expressions.push(ts.visitNode(element, visitor, ts.isExpression));
                    return expressions;
                }
            }
            function visitObjectLiteralExpression(node) {
                // [source]
                //      o = {
                //          a: 1,
                //          b: yield,
                //          c: 2
                //      };
                //
                // [intermediate]
                //  .local _a
                //      _a = {
                //          a: 1
                //      };
                //  .yield resumeLabel
                //  .mark resumeLabel
                //      o = (_a.b = %sent%,
                //          _a.c = 2,
                //          _a);
                var properties = node.properties;
                var multiLine = node.multiLine;
                var numInitialProperties = countInitialNodesWithoutYield(properties);
                var temp = declareLocal();
                emitAssignment(temp, factory.createObjectLiteralExpression(ts.visitNodes(properties, visitor, ts.isObjectLiteralElementLike, 0, numInitialProperties), multiLine));
                var expressions = ts.reduceLeft(properties, reduceProperty, [], numInitialProperties);
                // TODO(rbuckton): Does this need to be parented?
                expressions.push(multiLine ? ts.startOnNewLine(ts.setParent(ts.setTextRange(factory.cloneNode(temp), temp), temp.parent)) : temp);
                return factory.inlineExpressions(expressions);
                function reduceProperty(expressions, property) {
                    if (containsYield(property) && expressions.length > 0) {
                        emitStatement(factory.createExpressionStatement(factory.inlineExpressions(expressions)));
                        expressions = [];
                    }
                    var expression = ts.createExpressionForObjectLiteralElementLike(factory, node, property, temp);
                    var visited = ts.visitNode(expression, visitor, ts.isExpression);
                    if (visited) {
                        if (multiLine) {
                            ts.startOnNewLine(visited);
                        }
                        expressions.push(visited);
                    }
                    return expressions;
                }
            }
            /**
             * Visits an ElementAccessExpression that contains a YieldExpression.
             *
             * @param node The node to visit.
             */
            function visitElementAccessExpression(node) {
                if (containsYield(node.argumentExpression)) {
                    // [source]
                    //      a = x[yield];
                    //
                    // [intermediate]
                    //  .local _a
                    //      _a = x;
                    //  .yield resumeLabel
                    //  .mark resumeLabel
                    //      a = _a[%sent%]
                    return factory.updateElementAccessExpression(node, cacheExpression(ts.visitNode(node.expression, visitor, ts.isLeftHandSideExpression)), ts.visitNode(node.argumentExpression, visitor, ts.isExpression));
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitCallExpression(node) {
                if (!ts.isImportCall(node) && ts.forEach(node.arguments, containsYield)) {
                    // [source]
                    //      a.b(1, yield, 2);
                    //
                    // [intermediate]
                    //  .local _a, _b, _c
                    //      _b = (_a = a).b;
                    //      _c = [1];
                    //  .yield resumeLabel
                    //  .mark resumeLabel
                    //      _b.apply(_a, _c.concat([%sent%, 2]));
                    var _a = factory.createCallBinding(node.expression, hoistVariableDeclaration, languageVersion, /*cacheIdentifiers*/ true), target = _a.target, thisArg = _a.thisArg;
                    return ts.setOriginalNode(ts.setTextRange(factory.createFunctionApplyCall(cacheExpression(ts.visitNode(target, visitor, ts.isLeftHandSideExpression)), thisArg, visitElements(node.arguments)), node), node);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitNewExpression(node) {
                if (ts.forEach(node.arguments, containsYield)) {
                    // [source]
                    //      new a.b(1, yield, 2);
                    //
                    // [intermediate]
                    //  .local _a, _b, _c
                    //      _b = (_a = a.b).bind;
                    //      _c = [1];
                    //  .yield resumeLabel
                    //  .mark resumeLabel
                    //      new (_b.apply(_a, _c.concat([%sent%, 2])));
                    var _a = factory.createCallBinding(factory.createPropertyAccessExpression(node.expression, "bind"), hoistVariableDeclaration), target = _a.target, thisArg = _a.thisArg;
                    return ts.setOriginalNode(ts.setTextRange(factory.createNewExpression(factory.createFunctionApplyCall(cacheExpression(ts.visitNode(target, visitor, ts.isExpression)), thisArg, visitElements(node.arguments, 
                    /*leadingElement*/ factory.createVoidZero())), 
                    /*typeArguments*/ undefined, []), node), node);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function transformAndEmitStatements(statements, start) {
                if (start === void 0) {
                    start = 0;
                }
                var numStatements = statements.length;
                for (var i = start; i < numStatements; i++) {
                    transformAndEmitStatement(statements[i]);
                }
            }
            function transformAndEmitEmbeddedStatement(node) {
                if (ts.isBlock(node)) {
                    transformAndEmitStatements(node.statements);
                }
                else {
                    transformAndEmitStatement(node);
                }
            }
            function transformAndEmitStatement(node) {
                var savedInStatementContainingYield = inStatementContainingYield;
                if (!inStatementContainingYield) {
                    inStatementContainingYield = containsYield(node);
                }
                transformAndEmitStatementWorker(node);
                inStatementContainingYield = savedInStatementContainingYield;
            }
            function transformAndEmitStatementWorker(node) {
                switch (node.kind) {
                    case 233 /* Block */:
                        return transformAndEmitBlock(node);
                    case 236 /* ExpressionStatement */:
                        return transformAndEmitExpressionStatement(node);
                    case 237 /* IfStatement */:
                        return transformAndEmitIfStatement(node);
                    case 238 /* DoStatement */:
                        return transformAndEmitDoStatement(node);
                    case 239 /* WhileStatement */:
                        return transformAndEmitWhileStatement(node);
                    case 240 /* ForStatement */:
                        return transformAndEmitForStatement(node);
                    case 241 /* ForInStatement */:
                        return transformAndEmitForInStatement(node);
                    case 243 /* ContinueStatement */:
                        return transformAndEmitContinueStatement(node);
                    case 244 /* BreakStatement */:
                        return transformAndEmitBreakStatement(node);
                    case 245 /* ReturnStatement */:
                        return transformAndEmitReturnStatement(node);
                    case 246 /* WithStatement */:
                        return transformAndEmitWithStatement(node);
                    case 247 /* SwitchStatement */:
                        return transformAndEmitSwitchStatement(node);
                    case 248 /* LabeledStatement */:
                        return transformAndEmitLabeledStatement(node);
                    case 249 /* ThrowStatement */:
                        return transformAndEmitThrowStatement(node);
                    case 250 /* TryStatement */:
                        return transformAndEmitTryStatement(node);
                    default:
                        return emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function transformAndEmitBlock(node) {
                if (containsYield(node)) {
                    transformAndEmitStatements(node.statements);
                }
                else {
                    emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function transformAndEmitExpressionStatement(node) {
                emitStatement(ts.visitNode(node, visitor, ts.isStatement));
            }
            function transformAndEmitVariableDeclarationList(node) {
                for (var _i = 0, _a = node.declarations; _i < _a.length; _i++) {
                    var variable = _a[_i];
                    var name = factory.cloneNode(variable.name);
                    ts.setCommentRange(name, variable.name);
                    hoistVariableDeclaration(name);
                }
                var variables = ts.getInitializedVariables(node);
                var numVariables = variables.length;
                var variablesWritten = 0;
                var pendingExpressions = [];
                while (variablesWritten < numVariables) {
                    for (var i = variablesWritten; i < numVariables; i++) {
                        var variable = variables[i];
                        if (containsYield(variable.initializer) && pendingExpressions.length > 0) {
                            break;
                        }
                        pendingExpressions.push(transformInitializedVariable(variable));
                    }
                    if (pendingExpressions.length) {
                        emitStatement(factory.createExpressionStatement(factory.inlineExpressions(pendingExpressions)));
                        variablesWritten += pendingExpressions.length;
                        pendingExpressions = [];
                    }
                }
                return undefined;
            }
            function transformInitializedVariable(node) {
                return ts.setSourceMapRange(factory.createAssignment(ts.setSourceMapRange(factory.cloneNode(node.name), node.name), ts.visitNode(node.initializer, visitor, ts.isExpression)), node);
            }
            function transformAndEmitIfStatement(node) {
                if (containsYield(node)) {
                    // [source]
                    //      if (x)
                    //          /*thenStatement*/
                    //      else
                    //          /*elseStatement*/
                    //
                    // [intermediate]
                    //  .brfalse elseLabel, (x)
                    //      /*thenStatement*/
                    //  .br endLabel
                    //  .mark elseLabel
                    //      /*elseStatement*/
                    //  .mark endLabel
                    if (containsYield(node.thenStatement) || containsYield(node.elseStatement)) {
                        var endLabel = defineLabel();
                        var elseLabel = node.elseStatement ? defineLabel() : undefined;
                        emitBreakWhenFalse(node.elseStatement ? elseLabel : endLabel, ts.visitNode(node.expression, visitor, ts.isExpression), /*location*/ node.expression);
                        transformAndEmitEmbeddedStatement(node.thenStatement);
                        if (node.elseStatement) {
                            emitBreak(endLabel);
                            markLabel(elseLabel);
                            transformAndEmitEmbeddedStatement(node.elseStatement);
                        }
                        markLabel(endLabel);
                    }
                    else {
                        emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                    }
                }
                else {
                    emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function transformAndEmitDoStatement(node) {
                if (containsYield(node)) {
                    // [source]
                    //      do {
                    //          /*body*/
                    //      }
                    //      while (i < 10);
                    //
                    // [intermediate]
                    //  .loop conditionLabel, endLabel
                    //  .mark loopLabel
                    //      /*body*/
                    //  .mark conditionLabel
                    //  .brtrue loopLabel, (i < 10)
                    //  .endloop
                    //  .mark endLabel
                    var conditionLabel = defineLabel();
                    var loopLabel = defineLabel();
                    beginLoopBlock(/*continueLabel*/ conditionLabel);
                    markLabel(loopLabel);
                    transformAndEmitEmbeddedStatement(node.statement);
                    markLabel(conditionLabel);
                    emitBreakWhenTrue(loopLabel, ts.visitNode(node.expression, visitor, ts.isExpression));
                    endLoopBlock();
                }
                else {
                    emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function visitDoStatement(node) {
                if (inStatementContainingYield) {
                    beginScriptLoopBlock();
                    node = ts.visitEachChild(node, visitor, context);
                    endLoopBlock();
                    return node;
                }
                else {
                    return ts.visitEachChild(node, visitor, context);
                }
            }
            function transformAndEmitWhileStatement(node) {
                if (containsYield(node)) {
                    // [source]
                    //      while (i < 10) {
                    //          /*body*/
                    //      }
                    //
                    // [intermediate]
                    //  .loop loopLabel, endLabel
                    //  .mark loopLabel
                    //  .brfalse endLabel, (i < 10)
                    //      /*body*/
                    //  .br loopLabel
                    //  .endloop
                    //  .mark endLabel
                    var loopLabel = defineLabel();
                    var endLabel = beginLoopBlock(loopLabel);
                    markLabel(loopLabel);
                    emitBreakWhenFalse(endLabel, ts.visitNode(node.expression, visitor, ts.isExpression));
                    transformAndEmitEmbeddedStatement(node.statement);
                    emitBreak(loopLabel);
                    endLoopBlock();
                }
                else {
                    emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function visitWhileStatement(node) {
                if (inStatementContainingYield) {
                    beginScriptLoopBlock();
                    node = ts.visitEachChild(node, visitor, context);
                    endLoopBlock();
                    return node;
                }
                else {
                    return ts.visitEachChild(node, visitor, context);
                }
            }
            function transformAndEmitForStatement(node) {
                if (containsYield(node)) {
                    // [source]
                    //      for (var i = 0; i < 10; i++) {
                    //          /*body*/
                    //      }
                    //
                    // [intermediate]
                    //  .local i
                    //      i = 0;
                    //  .loop incrementLabel, endLoopLabel
                    //  .mark conditionLabel
                    //  .brfalse endLoopLabel, (i < 10)
                    //      /*body*/
                    //  .mark incrementLabel
                    //      i++;
                    //  .br conditionLabel
                    //  .endloop
                    //  .mark endLoopLabel
                    var conditionLabel = defineLabel();
                    var incrementLabel = defineLabel();
                    var endLabel = beginLoopBlock(incrementLabel);
                    if (node.initializer) {
                        var initializer = node.initializer;
                        if (ts.isVariableDeclarationList(initializer)) {
                            transformAndEmitVariableDeclarationList(initializer);
                        }
                        else {
                            emitStatement(ts.setTextRange(factory.createExpressionStatement(ts.visitNode(initializer, visitor, ts.isExpression)), initializer));
                        }
                    }
                    markLabel(conditionLabel);
                    if (node.condition) {
                        emitBreakWhenFalse(endLabel, ts.visitNode(node.condition, visitor, ts.isExpression));
                    }
                    transformAndEmitEmbeddedStatement(node.statement);
                    markLabel(incrementLabel);
                    if (node.incrementor) {
                        emitStatement(ts.setTextRange(factory.createExpressionStatement(ts.visitNode(node.incrementor, visitor, ts.isExpression)), node.incrementor));
                    }
                    emitBreak(conditionLabel);
                    endLoopBlock();
                }
                else {
                    emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function visitForStatement(node) {
                if (inStatementContainingYield) {
                    beginScriptLoopBlock();
                }
                var initializer = node.initializer;
                if (initializer && ts.isVariableDeclarationList(initializer)) {
                    for (var _i = 0, _a = initializer.declarations; _i < _a.length; _i++) {
                        var variable = _a[_i];
                        hoistVariableDeclaration(variable.name);
                    }
                    var variables = ts.getInitializedVariables(initializer);
                    node = factory.updateForStatement(node, variables.length > 0
                        ? factory.inlineExpressions(ts.map(variables, transformInitializedVariable))
                        : undefined, ts.visitNode(node.condition, visitor, ts.isExpression), ts.visitNode(node.incrementor, visitor, ts.isExpression), ts.visitIterationBody(node.statement, visitor, context));
                }
                else {
                    node = ts.visitEachChild(node, visitor, context);
                }
                if (inStatementContainingYield) {
                    endLoopBlock();
                }
                return node;
            }
            function transformAndEmitForInStatement(node) {
                // TODO(rbuckton): Source map locations
                if (containsYield(node)) {
                    // [source]
                    //      for (var p in o) {
                    //          /*body*/
                    //      }
                    //
                    // [intermediate]
                    //  .local _a, _b, _i
                    //      _a = [];
                    //      for (_b in o) _a.push(_b);
                    //      _i = 0;
                    //  .loop incrementLabel, endLoopLabel
                    //  .mark conditionLabel
                    //  .brfalse endLoopLabel, (_i < _a.length)
                    //      p = _a[_i];
                    //      /*body*/
                    //  .mark incrementLabel
                    //      _b++;
                    //  .br conditionLabel
                    //  .endloop
                    //  .mark endLoopLabel
                    var keysArray = declareLocal(); // _a
                    var key = declareLocal(); // _b
                    var keysIndex = factory.createLoopVariable(); // _i
                    var initializer = node.initializer;
                    hoistVariableDeclaration(keysIndex);
                    emitAssignment(keysArray, factory.createArrayLiteralExpression());
                    emitStatement(factory.createForInStatement(key, ts.visitNode(node.expression, visitor, ts.isExpression), factory.createExpressionStatement(factory.createCallExpression(factory.createPropertyAccessExpression(keysArray, "push"), 
                    /*typeArguments*/ undefined, [key]))));
                    emitAssignment(keysIndex, factory.createNumericLiteral(0));
                    var conditionLabel = defineLabel();
                    var incrementLabel = defineLabel();
                    var endLabel = beginLoopBlock(incrementLabel);
                    markLabel(conditionLabel);
                    emitBreakWhenFalse(endLabel, factory.createLessThan(keysIndex, factory.createPropertyAccessExpression(keysArray, "length")));
                    var variable = void 0;
                    if (ts.isVariableDeclarationList(initializer)) {
                        for (var _i = 0, _a = initializer.declarations; _i < _a.length; _i++) {
                            var variable_1 = _a[_i];
                            hoistVariableDeclaration(variable_1.name);
                        }
                        variable = factory.cloneNode(initializer.declarations[0].name);
                    }
                    else {
                        variable = ts.visitNode(initializer, visitor, ts.isExpression);
                        ts.Debug.assert(ts.isLeftHandSideExpression(variable));
                    }
                    emitAssignment(variable, factory.createElementAccessExpression(keysArray, keysIndex));
                    transformAndEmitEmbeddedStatement(node.statement);
                    markLabel(incrementLabel);
                    emitStatement(factory.createExpressionStatement(factory.createPostfixIncrement(keysIndex)));
                    emitBreak(conditionLabel);
                    endLoopBlock();
                }
                else {
                    emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function visitForInStatement(node) {
                // [source]
                //      for (var x in a) {
                //          /*body*/
                //      }
                //
                // [intermediate]
                //  .local x
                //  .loop
                //      for (x in a) {
                //          /*body*/
                //      }
                //  .endloop
                if (inStatementContainingYield) {
                    beginScriptLoopBlock();
                }
                var initializer = node.initializer;
                if (ts.isVariableDeclarationList(initializer)) {
                    for (var _i = 0, _a = initializer.declarations; _i < _a.length; _i++) {
                        var variable = _a[_i];
                        hoistVariableDeclaration(variable.name);
                    }
                    node = factory.updateForInStatement(node, initializer.declarations[0].name, ts.visitNode(node.expression, visitor, ts.isExpression), ts.visitNode(node.statement, visitor, ts.isStatement, factory.liftToBlock));
                }
                else {
                    node = ts.visitEachChild(node, visitor, context);
                }
                if (inStatementContainingYield) {
                    endLoopBlock();
                }
                return node;
            }
            function transformAndEmitContinueStatement(node) {
                var label = findContinueTarget(node.label ? ts.idText(node.label) : undefined);
                if (label > 0) {
                    emitBreak(label, /*location*/ node);
                }
                else {
                    // invalid continue without a containing loop. Leave the node as is, per #17875.
                    emitStatement(node);
                }
            }
            function visitContinueStatement(node) {
                if (inStatementContainingYield) {
                    var label = findContinueTarget(node.label && ts.idText(node.label));
                    if (label > 0) {
                        return createInlineBreak(label, /*location*/ node);
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function transformAndEmitBreakStatement(node) {
                var label = findBreakTarget(node.label ? ts.idText(node.label) : undefined);
                if (label > 0) {
                    emitBreak(label, /*location*/ node);
                }
                else {
                    // invalid break without a containing loop, switch, or labeled statement. Leave the node as is, per #17875.
                    emitStatement(node);
                }
            }
            function visitBreakStatement(node) {
                if (inStatementContainingYield) {
                    var label = findBreakTarget(node.label && ts.idText(node.label));
                    if (label > 0) {
                        return createInlineBreak(label, /*location*/ node);
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function transformAndEmitReturnStatement(node) {
                emitReturn(ts.visitNode(node.expression, visitor, ts.isExpression), 
                /*location*/ node);
            }
            function visitReturnStatement(node) {
                return createInlineReturn(ts.visitNode(node.expression, visitor, ts.isExpression), 
                /*location*/ node);
            }
            function transformAndEmitWithStatement(node) {
                if (containsYield(node)) {
                    // [source]
                    //      with (x) {
                    //          /*body*/
                    //      }
                    //
                    // [intermediate]
                    //  .with (x)
                    //      /*body*/
                    //  .endwith
                    beginWithBlock(cacheExpression(ts.visitNode(node.expression, visitor, ts.isExpression)));
                    transformAndEmitEmbeddedStatement(node.statement);
                    endWithBlock();
                }
                else {
                    emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function transformAndEmitSwitchStatement(node) {
                if (containsYield(node.caseBlock)) {
                    // [source]
                    //      switch (x) {
                    //          case a:
                    //              /*caseStatements*/
                    //          case b:
                    //              /*caseStatements*/
                    //          default:
                    //              /*defaultStatements*/
                    //      }
                    //
                    // [intermediate]
                    //  .local _a
                    //  .switch endLabel
                    //      _a = x;
                    //      switch (_a) {
                    //          case a:
                    //  .br clauseLabels[0]
                    //      }
                    //      switch (_a) {
                    //          case b:
                    //  .br clauseLabels[1]
                    //      }
                    //  .br clauseLabels[2]
                    //  .mark clauseLabels[0]
                    //      /*caseStatements*/
                    //  .mark clauseLabels[1]
                    //      /*caseStatements*/
                    //  .mark clauseLabels[2]
                    //      /*caseStatements*/
                    //  .endswitch
                    //  .mark endLabel
                    var caseBlock = node.caseBlock;
                    var numClauses = caseBlock.clauses.length;
                    var endLabel = beginSwitchBlock();
                    var expression = cacheExpression(ts.visitNode(node.expression, visitor, ts.isExpression));
                    // Create labels for each clause and find the index of the first default clause.
                    var clauseLabels = [];
                    var defaultClauseIndex = -1;
                    for (var i = 0; i < numClauses; i++) {
                        var clause = caseBlock.clauses[i];
                        clauseLabels.push(defineLabel());
                        if (clause.kind === 288 /* DefaultClause */ && defaultClauseIndex === -1) {
                            defaultClauseIndex = i;
                        }
                    }
                    // Emit switch statements for each run of case clauses either from the first case
                    // clause or the next case clause with a `yield` in its expression, up to the next
                    // case clause with a `yield` in its expression.
                    var clausesWritten = 0;
                    var pendingClauses = [];
                    while (clausesWritten < numClauses) {
                        var defaultClausesSkipped = 0;
                        for (var i = clausesWritten; i < numClauses; i++) {
                            var clause = caseBlock.clauses[i];
                            if (clause.kind === 287 /* CaseClause */) {
                                if (containsYield(clause.expression) && pendingClauses.length > 0) {
                                    break;
                                }
                                pendingClauses.push(factory.createCaseClause(ts.visitNode(clause.expression, visitor, ts.isExpression), [
                                    createInlineBreak(clauseLabels[i], /*location*/ clause.expression)
                                ]));
                            }
                            else {
                                defaultClausesSkipped++;
                            }
                        }
                        if (pendingClauses.length) {
                            emitStatement(factory.createSwitchStatement(expression, factory.createCaseBlock(pendingClauses)));
                            clausesWritten += pendingClauses.length;
                            pendingClauses = [];
                        }
                        if (defaultClausesSkipped > 0) {
                            clausesWritten += defaultClausesSkipped;
                            defaultClausesSkipped = 0;
                        }
                    }
                    if (defaultClauseIndex >= 0) {
                        emitBreak(clauseLabels[defaultClauseIndex]);
                    }
                    else {
                        emitBreak(endLabel);
                    }
                    for (var i = 0; i < numClauses; i++) {
                        markLabel(clauseLabels[i]);
                        transformAndEmitStatements(caseBlock.clauses[i].statements);
                    }
                    endSwitchBlock();
                }
                else {
                    emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function visitSwitchStatement(node) {
                if (inStatementContainingYield) {
                    beginScriptSwitchBlock();
                }
                node = ts.visitEachChild(node, visitor, context);
                if (inStatementContainingYield) {
                    endSwitchBlock();
                }
                return node;
            }
            function transformAndEmitLabeledStatement(node) {
                if (containsYield(node)) {
                    // [source]
                    //      x: {
                    //          /*body*/
                    //      }
                    //
                    // [intermediate]
                    //  .labeled "x", endLabel
                    //      /*body*/
                    //  .endlabeled
                    //  .mark endLabel
                    beginLabeledBlock(ts.idText(node.label));
                    transformAndEmitEmbeddedStatement(node.statement);
                    endLabeledBlock();
                }
                else {
                    emitStatement(ts.visitNode(node, visitor, ts.isStatement));
                }
            }
            function visitLabeledStatement(node) {
                if (inStatementContainingYield) {
                    beginScriptLabeledBlock(ts.idText(node.label));
                }
                node = ts.visitEachChild(node, visitor, context);
                if (inStatementContainingYield) {
                    endLabeledBlock();
                }
                return node;
            }
            function transformAndEmitThrowStatement(node) {
                var _a;
                // TODO(rbuckton): `expression` should be required on `throw`.
                emitThrow(ts.visitNode((_a = node.expression) !== null && _a !== void 0 ? _a : factory.createVoidZero(), visitor, ts.isExpression), 
                /*location*/ node);
            }
            function transformAndEmitTryStatement(node) {
                if (containsYield(node)) {
                    // [source]
                    //      try {
                    //          /*tryBlock*/
                    //      }
                    //      catch (e) {
                    //          /*catchBlock*/
                    //      }
                    //      finally {
                    //          /*finallyBlock*/
                    //      }
                    //
                    // [intermediate]
                    //  .local _a
                    //  .try tryLabel, catchLabel, finallyLabel, endLabel
                    //  .mark tryLabel
                    //  .nop
                    //      /*tryBlock*/
                    //  .br endLabel
                    //  .catch
                    //  .mark catchLabel
                    //      _a = %error%;
                    //      /*catchBlock*/
                    //  .br endLabel
                    //  .finally
                    //  .mark finallyLabel
                    //      /*finallyBlock*/
                    //  .endfinally
                    //  .endtry
                    //  .mark endLabel
                    beginExceptionBlock();
                    transformAndEmitEmbeddedStatement(node.tryBlock);
                    if (node.catchClause) {
                        beginCatchBlock(node.catchClause.variableDeclaration); // TODO: GH#18217
                        transformAndEmitEmbeddedStatement(node.catchClause.block);
                    }
                    if (node.finallyBlock) {
                        beginFinallyBlock();
                        transformAndEmitEmbeddedStatement(node.finallyBlock);
                    }
                    endExceptionBlock();
                }
                else {
                    emitStatement(ts.visitEachChild(node, visitor, context));
                }
            }
            function containsYield(node) {
                return !!node && (node.transformFlags & 524288 /* ContainsYield */) !== 0;
            }
            function countInitialNodesWithoutYield(nodes) {
                var numNodes = nodes.length;
                for (var i = 0; i < numNodes; i++) {
                    if (containsYield(nodes[i])) {
                        return i;
                    }
                }
                return -1;
            }
            function onSubstituteNode(hint, node) {
                node = previousOnSubstituteNode(hint, node);
                if (hint === 1 /* Expression */) {
                    return substituteExpression(node);
                }
                return node;
            }
            function substituteExpression(node) {
                if (ts.isIdentifier(node)) {
                    return substituteExpressionIdentifier(node);
                }
                return node;
            }
            function substituteExpressionIdentifier(node) {
                if (!ts.isGeneratedIdentifier(node) && renamedCatchVariables && renamedCatchVariables.has(ts.idText(node))) {
                    var original = ts.getOriginalNode(node);
                    if (ts.isIdentifier(original) && original.parent) {
                        var declaration = resolver.getReferencedValueDeclaration(original);
                        if (declaration) {
                            var name = renamedCatchVariableDeclarations[ts.getOriginalNodeId(declaration)];
                            if (name) {
                                // TODO(rbuckton): Does this need to be parented?
                                var clone_6 = ts.setParent(ts.setTextRange(factory.cloneNode(name), name), name.parent);
                                ts.setSourceMapRange(clone_6, node);
                                ts.setCommentRange(clone_6, node);
                                return clone_6;
                            }
                        }
                    }
                }
                return node;
            }
            function cacheExpression(node) {
                if (ts.isGeneratedIdentifier(node) || ts.getEmitFlags(node) & 4096 /* HelperName */) {
                    return node;
                }
                var temp = factory.createTempVariable(hoistVariableDeclaration);
                emitAssignment(temp, node, /*location*/ node);
                return temp;
            }
            function declareLocal(name) {
                var temp = name
                    ? factory.createUniqueName(name)
                    : factory.createTempVariable(/*recordTempVariable*/ undefined);
                hoistVariableDeclaration(temp);
                return temp;
            }
            /**
             * Defines a label, uses as the target of a Break operation.
             */
            function defineLabel() {
                if (!labelOffsets) {
                    labelOffsets = [];
                }
                var label = nextLabelId;
                nextLabelId++;
                labelOffsets[label] = -1;
                return label;
            }
            /**
             * Marks the current operation with the specified label.
             */
            function markLabel(label) {
                ts.Debug.assert(labelOffsets !== undefined, "No labels were defined.");
                labelOffsets[label] = operations ? operations.length : 0;
            }
            /**
             * Begins a block operation (With, Break/Continue, Try/Catch/Finally)
             *
             * @param block Information about the block.
             */
            function beginBlock(block) {
                if (!blocks) {
                    blocks = [];
                    blockActions = [];
                    blockOffsets = [];
                    blockStack = [];
                }
                var index = blockActions.length;
                blockActions[index] = 0 /* Open */;
                blockOffsets[index] = operations ? operations.length : 0;
                blocks[index] = block;
                blockStack.push(block);
                return index;
            }
            /**
             * Ends the current block operation.
             */
            function endBlock() {
                var block = peekBlock();
                if (block === undefined)
                    return ts.Debug.fail("beginBlock was never called.");
                var index = blockActions.length;
                blockActions[index] = 1 /* Close */;
                blockOffsets[index] = operations ? operations.length : 0;
                blocks[index] = block;
                blockStack.pop();
                return block;
            }
            /**
             * Gets the current open block.
             */
            function peekBlock() {
                return ts.lastOrUndefined(blockStack);
            }
            /**
             * Gets the kind of the current open block.
             */
            function peekBlockKind() {
                var block = peekBlock();
                return block && block.kind;
            }
            /**
             * Begins a code block for a generated `with` statement.
             *
             * @param expression An identifier representing expression for the `with` block.
             */
            function beginWithBlock(expression) {
                var startLabel = defineLabel();
                var endLabel = defineLabel();
                markLabel(startLabel);
                beginBlock({
                    kind: 1 /* With */,
                    expression: expression,
                    startLabel: startLabel,
                    endLabel: endLabel
                });
            }
            /**
             * Ends a code block for a generated `with` statement.
             */
            function endWithBlock() {
                ts.Debug.assert(peekBlockKind() === 1 /* With */);
                var block = endBlock();
                markLabel(block.endLabel);
            }
            /**
             * Begins a code block for a generated `try` statement.
             */
            function beginExceptionBlock() {
                var startLabel = defineLabel();
                var endLabel = defineLabel();
                markLabel(startLabel);
                beginBlock({
                    kind: 0 /* Exception */,
                    state: 0 /* Try */,
                    startLabel: startLabel,
                    endLabel: endLabel
                });
                emitNop();
                return endLabel;
            }
            /**
             * Enters the `catch` clause of a generated `try` statement.
             *
             * @param variable The catch variable.
             */
            function beginCatchBlock(variable) {
                ts.Debug.assert(peekBlockKind() === 0 /* Exception */);
                // generated identifiers should already be unique within a file
                var name;
                if (ts.isGeneratedIdentifier(variable.name)) {
                    name = variable.name;
                    hoistVariableDeclaration(variable.name);
                }
                else {
                    var text = ts.idText(variable.name);
                    name = declareLocal(text);
                    if (!renamedCatchVariables) {
                        renamedCatchVariables = new ts.Map();
                        renamedCatchVariableDeclarations = [];
                        context.enableSubstitution(79 /* Identifier */);
                    }
                    renamedCatchVariables.set(text, true);
                    renamedCatchVariableDeclarations[ts.getOriginalNodeId(variable)] = name;
                }
                var exception = peekBlock();
                ts.Debug.assert(exception.state < 1 /* Catch */);
                var endLabel = exception.endLabel;
                emitBreak(endLabel);
                var catchLabel = defineLabel();
                markLabel(catchLabel);
                exception.state = 1 /* Catch */;
                exception.catchVariable = name;
                exception.catchLabel = catchLabel;
                emitAssignment(name, factory.createCallExpression(factory.createPropertyAccessExpression(state, "sent"), /*typeArguments*/ undefined, []));
                emitNop();
            }
            /**
             * Enters the `finally` block of a generated `try` statement.
             */
            function beginFinallyBlock() {
                ts.Debug.assert(peekBlockKind() === 0 /* Exception */);
                var exception = peekBlock();
                ts.Debug.assert(exception.state < 2 /* Finally */);
                var endLabel = exception.endLabel;
                emitBreak(endLabel);
                var finallyLabel = defineLabel();
                markLabel(finallyLabel);
                exception.state = 2 /* Finally */;
                exception.finallyLabel = finallyLabel;
            }
            /**
             * Ends the code block for a generated `try` statement.
             */
            function endExceptionBlock() {
                ts.Debug.assert(peekBlockKind() === 0 /* Exception */);
                var exception = endBlock();
                var state = exception.state;
                if (state < 2 /* Finally */) {
                    emitBreak(exception.endLabel);
                }
                else {
                    emitEndfinally();
                }
                markLabel(exception.endLabel);
                emitNop();
                exception.state = 3 /* Done */;
            }
            /**
             * Begins a code block that supports `break` or `continue` statements that are defined in
             * the source tree and not from generated code.
             *
             * @param labelText Names from containing labeled statements.
             */
            function beginScriptLoopBlock() {
                beginBlock({
                    kind: 3 /* Loop */,
                    isScript: true,
                    breakLabel: -1,
                    continueLabel: -1
                });
            }
            /**
             * Begins a code block that supports `break` or `continue` statements that are defined in
             * generated code. Returns a label used to mark the operation to which to jump when a
             * `break` statement targets this block.
             *
             * @param continueLabel A Label used to mark the operation to which to jump when a
             *                      `continue` statement targets this block.
             */
            function beginLoopBlock(continueLabel) {
                var breakLabel = defineLabel();
                beginBlock({
                    kind: 3 /* Loop */,
                    isScript: false,
                    breakLabel: breakLabel,
                    continueLabel: continueLabel,
                });
                return breakLabel;
            }
            /**
             * Ends a code block that supports `break` or `continue` statements that are defined in
             * generated code or in the source tree.
             */
            function endLoopBlock() {
                ts.Debug.assert(peekBlockKind() === 3 /* Loop */);
                var block = endBlock();
                var breakLabel = block.breakLabel;
                if (!block.isScript) {
                    markLabel(breakLabel);
                }
            }
            /**
             * Begins a code block that supports `break` statements that are defined in the source
             * tree and not from generated code.
             *
             */
            function beginScriptSwitchBlock() {
                beginBlock({
                    kind: 2 /* Switch */,
                    isScript: true,
                    breakLabel: -1
                });
            }
            /**
             * Begins a code block that supports `break` statements that are defined in generated code.
             * Returns a label used to mark the operation to which to jump when a `break` statement
             * targets this block.
             */
            function beginSwitchBlock() {
                var breakLabel = defineLabel();
                beginBlock({
                    kind: 2 /* Switch */,
                    isScript: false,
                    breakLabel: breakLabel,
                });
                return breakLabel;
            }
            /**
             * Ends a code block that supports `break` statements that are defined in generated code.
             */
            function endSwitchBlock() {
                ts.Debug.assert(peekBlockKind() === 2 /* Switch */);
                var block = endBlock();
                var breakLabel = block.breakLabel;
                if (!block.isScript) {
                    markLabel(breakLabel);
                }
            }
            function beginScriptLabeledBlock(labelText) {
                beginBlock({
                    kind: 4 /* Labeled */,
                    isScript: true,
                    labelText: labelText,
                    breakLabel: -1
                });
            }
            function beginLabeledBlock(labelText) {
                var breakLabel = defineLabel();
                beginBlock({
                    kind: 4 /* Labeled */,
                    isScript: false,
                    labelText: labelText,
                    breakLabel: breakLabel
                });
            }
            function endLabeledBlock() {
                ts.Debug.assert(peekBlockKind() === 4 /* Labeled */);
                var block = endBlock();
                if (!block.isScript) {
                    markLabel(block.breakLabel);
                }
            }
            /**
             * Indicates whether the provided block supports `break` statements.
             *
             * @param block A code block.
             */
            function supportsUnlabeledBreak(block) {
                return block.kind === 2 /* Switch */
                    || block.kind === 3 /* Loop */;
            }
            /**
             * Indicates whether the provided block supports `break` statements with labels.
             *
             * @param block A code block.
             */
            function supportsLabeledBreakOrContinue(block) {
                return block.kind === 4 /* Labeled */;
            }
            /**
             * Indicates whether the provided block supports `continue` statements.
             *
             * @param block A code block.
             */
            function supportsUnlabeledContinue(block) {
                return block.kind === 3 /* Loop */;
            }
            function hasImmediateContainingLabeledBlock(labelText, start) {
                for (var j = start; j >= 0; j--) {
                    var containingBlock = blockStack[j];
                    if (supportsLabeledBreakOrContinue(containingBlock)) {
                        if (containingBlock.labelText === labelText) {
                            return true;
                        }
                    }
                    else {
                        break;
                    }
                }
                return false;
            }
            /**
             * Finds the label that is the target for a `break` statement.
             *
             * @param labelText An optional name of a containing labeled statement.
             */
            function findBreakTarget(labelText) {
                if (blockStack) {
                    if (labelText) {
                        for (var i = blockStack.length - 1; i >= 0; i--) {
                            var block = blockStack[i];
                            if (supportsLabeledBreakOrContinue(block) && block.labelText === labelText) {
                                return block.breakLabel;
                            }
                            else if (supportsUnlabeledBreak(block) && hasImmediateContainingLabeledBlock(labelText, i - 1)) {
                                return block.breakLabel;
                            }
                        }
                    }
                    else {
                        for (var i = blockStack.length - 1; i >= 0; i--) {
                            var block = blockStack[i];
                            if (supportsUnlabeledBreak(block)) {
                                return block.breakLabel;
                            }
                        }
                    }
                }
                return 0;
            }
            /**
             * Finds the label that is the target for a `continue` statement.
             *
             * @param labelText An optional name of a containing labeled statement.
             */
            function findContinueTarget(labelText) {
                if (blockStack) {
                    if (labelText) {
                        for (var i = blockStack.length - 1; i >= 0; i--) {
                            var block = blockStack[i];
                            if (supportsUnlabeledContinue(block) && hasImmediateContainingLabeledBlock(labelText, i - 1)) {
                                return block.continueLabel;
                            }
                        }
                    }
                    else {
                        for (var i = blockStack.length - 1; i >= 0; i--) {
                            var block = blockStack[i];
                            if (supportsUnlabeledContinue(block)) {
                                return block.continueLabel;
                            }
                        }
                    }
                }
                return 0;
            }
            /**
             * Creates an expression that can be used to indicate the value for a label.
             *
             * @param label A label.
             */
            function createLabel(label) {
                if (label !== undefined && label > 0) {
                    if (labelExpressions === undefined) {
                        labelExpressions = [];
                    }
                    var expression = factory.createNumericLiteral(-1);
                    if (labelExpressions[label] === undefined) {
                        labelExpressions[label] = [expression];
                    }
                    else {
                        labelExpressions[label].push(expression);
                    }
                    return expression;
                }
                return factory.createOmittedExpression();
            }
            /**
             * Creates a numeric literal for the provided instruction.
             */
            function createInstruction(instruction) {
                var literal = factory.createNumericLiteral(instruction);
                ts.addSyntheticTrailingComment(literal, 3 /* MultiLineCommentTrivia */, getInstructionName(instruction));
                return literal;
            }
            /**
             * Creates a statement that can be used indicate a Break operation to the provided label.
             *
             * @param label A label.
             * @param location An optional source map location for the statement.
             */
            function createInlineBreak(label, location) {
                ts.Debug.assertLessThan(0, label, "Invalid label");
                return ts.setTextRange(factory.createReturnStatement(factory.createArrayLiteralExpression([
                    createInstruction(3 /* Break */),
                    createLabel(label)
                ])), location);
            }
            /**
             * Creates a statement that can be used indicate a Return operation.
             *
             * @param expression The expression for the return statement.
             * @param location An optional source map location for the statement.
             */
            function createInlineReturn(expression, location) {
                return ts.setTextRange(factory.createReturnStatement(factory.createArrayLiteralExpression(expression
                    ? [createInstruction(2 /* Return */), expression]
                    : [createInstruction(2 /* Return */)])), location);
            }
            /**
             * Creates an expression that can be used to resume from a Yield operation.
             */
            function createGeneratorResume(location) {
                return ts.setTextRange(factory.createCallExpression(factory.createPropertyAccessExpression(state, "sent"), 
                /*typeArguments*/ undefined, []), location);
            }
            /**
             * Emits an empty instruction.
             */
            function emitNop() {
                emitWorker(0 /* Nop */);
            }
            /**
             * Emits a Statement.
             *
             * @param node A statement.
             */
            function emitStatement(node) {
                if (node) {
                    emitWorker(1 /* Statement */, [node]);
                }
                else {
                    emitNop();
                }
            }
            /**
             * Emits an Assignment operation.
             *
             * @param left The left-hand side of the assignment.
             * @param right The right-hand side of the assignment.
             * @param location An optional source map location for the assignment.
             */
            function emitAssignment(left, right, location) {
                emitWorker(2 /* Assign */, [left, right], location);
            }
            /**
             * Emits a Break operation to the specified label.
             *
             * @param label A label.
             * @param location An optional source map location for the assignment.
             */
            function emitBreak(label, location) {
                emitWorker(3 /* Break */, [label], location);
            }
            /**
             * Emits a Break operation to the specified label when a condition evaluates to a truthy
             * value at runtime.
             *
             * @param label A label.
             * @param condition The condition.
             * @param location An optional source map location for the assignment.
             */
            function emitBreakWhenTrue(label, condition, location) {
                emitWorker(4 /* BreakWhenTrue */, [label, condition], location);
            }
            /**
             * Emits a Break to the specified label when a condition evaluates to a falsey value at
             * runtime.
             *
             * @param label A label.
             * @param condition The condition.
             * @param location An optional source map location for the assignment.
             */
            function emitBreakWhenFalse(label, condition, location) {
                emitWorker(5 /* BreakWhenFalse */, [label, condition], location);
            }
            /**
             * Emits a YieldStar operation for the provided expression.
             *
             * @param expression An optional value for the yield operation.
             * @param location An optional source map location for the assignment.
             */
            function emitYieldStar(expression, location) {
                emitWorker(7 /* YieldStar */, [expression], location);
            }
            /**
             * Emits a Yield operation for the provided expression.
             *
             * @param expression An optional value for the yield operation.
             * @param location An optional source map location for the assignment.
             */
            function emitYield(expression, location) {
                emitWorker(6 /* Yield */, [expression], location);
            }
            /**
             * Emits a Return operation for the provided expression.
             *
             * @param expression An optional value for the operation.
             * @param location An optional source map location for the assignment.
             */
            function emitReturn(expression, location) {
                emitWorker(8 /* Return */, [expression], location);
            }
            /**
             * Emits a Throw operation for the provided expression.
             *
             * @param expression A value for the operation.
             * @param location An optional source map location for the assignment.
             */
            function emitThrow(expression, location) {
                emitWorker(9 /* Throw */, [expression], location);
            }
            /**
             * Emits an Endfinally operation. This is used to handle `finally` block semantics.
             */
            function emitEndfinally() {
                emitWorker(10 /* Endfinally */);
            }
            /**
             * Emits an operation.
             *
             * @param code The OpCode for the operation.
             * @param args The optional arguments for the operation.
             */
            function emitWorker(code, args, location) {
                if (operations === undefined) {
                    operations = [];
                    operationArguments = [];
                    operationLocations = [];
                }
                if (labelOffsets === undefined) {
                    // mark entry point
                    markLabel(defineLabel());
                }
                var operationIndex = operations.length;
                operations[operationIndex] = code;
                operationArguments[operationIndex] = args;
                operationLocations[operationIndex] = location;
            }
            /**
             * Builds the generator function body.
             */
            function build() {
                blockIndex = 0;
                labelNumber = 0;
                labelNumbers = undefined;
                lastOperationWasAbrupt = false;
                lastOperationWasCompletion = false;
                clauses = undefined;
                statements = undefined;
                exceptionBlockStack = undefined;
                currentExceptionBlock = undefined;
                withBlockStack = undefined;
                var buildResult = buildStatements();
                return emitHelpers().createGeneratorHelper(ts.setEmitFlags(factory.createFunctionExpression(
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, [factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, state)], 
                /*type*/ undefined, factory.createBlock(buildResult, 
                /*multiLine*/ buildResult.length > 0)), 524288 /* ReuseTempVariableScope */));
            }
            /**
             * Builds the statements for the generator function body.
             */
            function buildStatements() {
                if (operations) {
                    for (var operationIndex = 0; operationIndex < operations.length; operationIndex++) {
                        writeOperation(operationIndex);
                    }
                    flushFinalLabel(operations.length);
                }
                else {
                    flushFinalLabel(0);
                }
                if (clauses) {
                    var labelExpression = factory.createPropertyAccessExpression(state, "label");
                    var switchStatement = factory.createSwitchStatement(labelExpression, factory.createCaseBlock(clauses));
                    return [ts.startOnNewLine(switchStatement)];
                }
                if (statements) {
                    return statements;
                }
                return [];
            }
            /**
             * Flush the current label and advance to a new label.
             */
            function flushLabel() {
                if (!statements) {
                    return;
                }
                appendLabel(/*markLabelEnd*/ !lastOperationWasAbrupt);
                lastOperationWasAbrupt = false;
                lastOperationWasCompletion = false;
                labelNumber++;
            }
            /**
             * Flush the final label of the generator function body.
             */
            function flushFinalLabel(operationIndex) {
                if (isFinalLabelReachable(operationIndex)) {
                    tryEnterLabel(operationIndex);
                    withBlockStack = undefined;
                    writeReturn(/*expression*/ undefined, /*operationLocation*/ undefined);
                }
                if (statements && clauses) {
                    appendLabel(/*markLabelEnd*/ false);
                }
                updateLabelExpressions();
            }
            /**
             * Tests whether the final label of the generator function body
             * is reachable by user code.
             */
            function isFinalLabelReachable(operationIndex) {
                // if the last operation was *not* a completion (return/throw) then
                // the final label is reachable.
                if (!lastOperationWasCompletion) {
                    return true;
                }
                // if there are no labels defined or referenced, then the final label is
                // not reachable.
                if (!labelOffsets || !labelExpressions) {
                    return false;
                }
                // if the label for this offset is referenced, then the final label
                // is reachable.
                for (var label = 0; label < labelOffsets.length; label++) {
                    if (labelOffsets[label] === operationIndex && labelExpressions[label]) {
                        return true;
                    }
                }
                return false;
            }
            /**
             * Appends a case clause for the last label and sets the new label.
             *
             * @param markLabelEnd Indicates that the transition between labels was a fall-through
             *                     from a previous case clause and the change in labels should be
             *                     reflected on the `state` object.
             */
            function appendLabel(markLabelEnd) {
                if (!clauses) {
                    clauses = [];
                }
                if (statements) {
                    if (withBlockStack) {
                        // The previous label was nested inside one or more `with` blocks, so we
                        // surround the statements in generated `with` blocks to create the same environment.
                        for (var i = withBlockStack.length - 1; i >= 0; i--) {
                            var withBlock = withBlockStack[i];
                            statements = [factory.createWithStatement(withBlock.expression, factory.createBlock(statements))];
                        }
                    }
                    if (currentExceptionBlock) {
                        // The previous label was nested inside of an exception block, so we must
                        // indicate entry into a protected region by pushing the label numbers
                        // for each block in the protected region.
                        var startLabel = currentExceptionBlock.startLabel, catchLabel = currentExceptionBlock.catchLabel, finallyLabel = currentExceptionBlock.finallyLabel, endLabel = currentExceptionBlock.endLabel;
                        statements.unshift(factory.createExpressionStatement(factory.createCallExpression(factory.createPropertyAccessExpression(factory.createPropertyAccessExpression(state, "trys"), "push"), 
                        /*typeArguments*/ undefined, [
                            factory.createArrayLiteralExpression([
                                createLabel(startLabel),
                                createLabel(catchLabel),
                                createLabel(finallyLabel),
                                createLabel(endLabel)
                            ])
                        ])));
                        currentExceptionBlock = undefined;
                    }
                    if (markLabelEnd) {
                        // The case clause for the last label falls through to this label, so we
                        // add an assignment statement to reflect the change in labels.
                        statements.push(factory.createExpressionStatement(factory.createAssignment(factory.createPropertyAccessExpression(state, "label"), factory.createNumericLiteral(labelNumber + 1))));
                    }
                }
                clauses.push(factory.createCaseClause(factory.createNumericLiteral(labelNumber), statements || []));
                statements = undefined;
            }
            /**
             * Tries to enter into a new label at the current operation index.
             */
            function tryEnterLabel(operationIndex) {
                if (!labelOffsets) {
                    return;
                }
                for (var label = 0; label < labelOffsets.length; label++) {
                    if (labelOffsets[label] === operationIndex) {
                        flushLabel();
                        if (labelNumbers === undefined) {
                            labelNumbers = [];
                        }
                        if (labelNumbers[labelNumber] === undefined) {
                            labelNumbers[labelNumber] = [label];
                        }
                        else {
                            labelNumbers[labelNumber].push(label);
                        }
                    }
                }
            }
            /**
             * Updates literal expressions for labels with actual label numbers.
             */
            function updateLabelExpressions() {
                if (labelExpressions !== undefined && labelNumbers !== undefined) {
                    for (var labelNumber_1 = 0; labelNumber_1 < labelNumbers.length; labelNumber_1++) {
                        var labels = labelNumbers[labelNumber_1];
                        if (labels !== undefined) {
                            for (var _i = 0, labels_1 = labels; _i < labels_1.length; _i++) {
                                var label = labels_1[_i];
                                var expressions = labelExpressions[label];
                                if (expressions !== undefined) {
                                    for (var _a = 0, expressions_1 = expressions; _a < expressions_1.length; _a++) {
                                        var expression = expressions_1[_a];
                                        expression.text = String(labelNumber_1);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            /**
             * Tries to enter or leave a code block.
             */
            function tryEnterOrLeaveBlock(operationIndex) {
                if (blocks) {
                    for (; blockIndex < blockActions.length && blockOffsets[blockIndex] <= operationIndex; blockIndex++) {
                        var block = blocks[blockIndex];
                        var blockAction = blockActions[blockIndex];
                        switch (block.kind) {
                            case 0 /* Exception */:
                                if (blockAction === 0 /* Open */) {
                                    if (!exceptionBlockStack) {
                                        exceptionBlockStack = [];
                                    }
                                    if (!statements) {
                                        statements = [];
                                    }
                                    exceptionBlockStack.push(currentExceptionBlock);
                                    currentExceptionBlock = block;
                                }
                                else if (blockAction === 1 /* Close */) {
                                    currentExceptionBlock = exceptionBlockStack.pop();
                                }
                                break;
                            case 1 /* With */:
                                if (blockAction === 0 /* Open */) {
                                    if (!withBlockStack) {
                                        withBlockStack = [];
                                    }
                                    withBlockStack.push(block);
                                }
                                else if (blockAction === 1 /* Close */) {
                                    withBlockStack.pop();
                                }
                                break;
                            // default: do nothing
                        }
                    }
                }
            }
            /**
             * Writes an operation as a statement to the current label's statement list.
             *
             * @param operation The OpCode of the operation
             */
            function writeOperation(operationIndex) {
                tryEnterLabel(operationIndex);
                tryEnterOrLeaveBlock(operationIndex);
                // early termination, nothing else to process in this label
                if (lastOperationWasAbrupt) {
                    return;
                }
                lastOperationWasAbrupt = false;
                lastOperationWasCompletion = false;
                var opcode = operations[operationIndex];
                if (opcode === 0 /* Nop */) {
                    return;
                }
                else if (opcode === 10 /* Endfinally */) {
                    return writeEndfinally();
                }
                var args = operationArguments[operationIndex];
                if (opcode === 1 /* Statement */) {
                    return writeStatement(args[0]);
                }
                var location = operationLocations[operationIndex];
                switch (opcode) {
                    case 2 /* Assign */:
                        return writeAssign(args[0], args[1], location);
                    case 3 /* Break */:
                        return writeBreak(args[0], location);
                    case 4 /* BreakWhenTrue */:
                        return writeBreakWhenTrue(args[0], args[1], location);
                    case 5 /* BreakWhenFalse */:
                        return writeBreakWhenFalse(args[0], args[1], location);
                    case 6 /* Yield */:
                        return writeYield(args[0], location);
                    case 7 /* YieldStar */:
                        return writeYieldStar(args[0], location);
                    case 8 /* Return */:
                        return writeReturn(args[0], location);
                    case 9 /* Throw */:
                        return writeThrow(args[0], location);
                }
            }
            /**
             * Writes a statement to the current label's statement list.
             *
             * @param statement A statement to write.
             */
            function writeStatement(statement) {
                if (statement) {
                    if (!statements) {
                        statements = [statement];
                    }
                    else {
                        statements.push(statement);
                    }
                }
            }
            /**
             * Writes an Assign operation to the current label's statement list.
             *
             * @param left The left-hand side of the assignment.
             * @param right The right-hand side of the assignment.
             * @param operationLocation The source map location for the operation.
             */
            function writeAssign(left, right, operationLocation) {
                writeStatement(ts.setTextRange(factory.createExpressionStatement(factory.createAssignment(left, right)), operationLocation));
            }
            /**
             * Writes a Throw operation to the current label's statement list.
             *
             * @param expression The value to throw.
             * @param operationLocation The source map location for the operation.
             */
            function writeThrow(expression, operationLocation) {
                lastOperationWasAbrupt = true;
                lastOperationWasCompletion = true;
                writeStatement(ts.setTextRange(factory.createThrowStatement(expression), operationLocation));
            }
            /**
             * Writes a Return operation to the current label's statement list.
             *
             * @param expression The value to return.
             * @param operationLocation The source map location for the operation.
             */
            function writeReturn(expression, operationLocation) {
                lastOperationWasAbrupt = true;
                lastOperationWasCompletion = true;
                writeStatement(ts.setEmitFlags(ts.setTextRange(factory.createReturnStatement(factory.createArrayLiteralExpression(expression
                    ? [createInstruction(2 /* Return */), expression]
                    : [createInstruction(2 /* Return */)])), operationLocation), 384 /* NoTokenSourceMaps */));
            }
            /**
             * Writes a Break operation to the current label's statement list.
             *
             * @param label The label for the Break.
             * @param operationLocation The source map location for the operation.
             */
            function writeBreak(label, operationLocation) {
                lastOperationWasAbrupt = true;
                writeStatement(ts.setEmitFlags(ts.setTextRange(factory.createReturnStatement(factory.createArrayLiteralExpression([
                    createInstruction(3 /* Break */),
                    createLabel(label)
                ])), operationLocation), 384 /* NoTokenSourceMaps */));
            }
            /**
             * Writes a BreakWhenTrue operation to the current label's statement list.
             *
             * @param label The label for the Break.
             * @param condition The condition for the Break.
             * @param operationLocation The source map location for the operation.
             */
            function writeBreakWhenTrue(label, condition, operationLocation) {
                writeStatement(ts.setEmitFlags(factory.createIfStatement(condition, ts.setEmitFlags(ts.setTextRange(factory.createReturnStatement(factory.createArrayLiteralExpression([
                    createInstruction(3 /* Break */),
                    createLabel(label)
                ])), operationLocation), 384 /* NoTokenSourceMaps */)), 1 /* SingleLine */));
            }
            /**
             * Writes a BreakWhenFalse operation to the current label's statement list.
             *
             * @param label The label for the Break.
             * @param condition The condition for the Break.
             * @param operationLocation The source map location for the operation.
             */
            function writeBreakWhenFalse(label, condition, operationLocation) {
                writeStatement(ts.setEmitFlags(factory.createIfStatement(factory.createLogicalNot(condition), ts.setEmitFlags(ts.setTextRange(factory.createReturnStatement(factory.createArrayLiteralExpression([
                    createInstruction(3 /* Break */),
                    createLabel(label)
                ])), operationLocation), 384 /* NoTokenSourceMaps */)), 1 /* SingleLine */));
            }
            /**
             * Writes a Yield operation to the current label's statement list.
             *
             * @param expression The expression to yield.
             * @param operationLocation The source map location for the operation.
             */
            function writeYield(expression, operationLocation) {
                lastOperationWasAbrupt = true;
                writeStatement(ts.setEmitFlags(ts.setTextRange(factory.createReturnStatement(factory.createArrayLiteralExpression(expression
                    ? [createInstruction(4 /* Yield */), expression]
                    : [createInstruction(4 /* Yield */)])), operationLocation), 384 /* NoTokenSourceMaps */));
            }
            /**
             * Writes a YieldStar instruction to the current label's statement list.
             *
             * @param expression The expression to yield.
             * @param operationLocation The source map location for the operation.
             */
            function writeYieldStar(expression, operationLocation) {
                lastOperationWasAbrupt = true;
                writeStatement(ts.setEmitFlags(ts.setTextRange(factory.createReturnStatement(factory.createArrayLiteralExpression([
                    createInstruction(5 /* YieldStar */),
                    expression
                ])), operationLocation), 384 /* NoTokenSourceMaps */));
            }
            /**
             * Writes an Endfinally instruction to the current label's statement list.
             */
            function writeEndfinally() {
                lastOperationWasAbrupt = true;
                writeStatement(factory.createReturnStatement(factory.createArrayLiteralExpression([
                    createInstruction(7 /* Endfinally */)
                ])));
            }
        }
        ts.transformGenerators = transformGenerators;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function transformModule(context) {
            function getTransformModuleDelegate(moduleKind) {
                switch (moduleKind) {
                    case ts.ModuleKind.AMD: return transformAMDModule;
                    case ts.ModuleKind.UMD: return transformUMDModule;
                    default: return transformCommonJSModule;
                }
            }
            var factory = context.factory, emitHelpers = context.getEmitHelperFactory, startLexicalEnvironment = context.startLexicalEnvironment, endLexicalEnvironment = context.endLexicalEnvironment, hoistVariableDeclaration = context.hoistVariableDeclaration;
            var compilerOptions = context.getCompilerOptions();
            var resolver = context.getEmitResolver();
            var host = context.getEmitHost();
            var languageVersion = ts.getEmitScriptTarget(compilerOptions);
            var moduleKind = ts.getEmitModuleKind(compilerOptions);
            var previousOnSubstituteNode = context.onSubstituteNode;
            var previousOnEmitNode = context.onEmitNode;
            context.onSubstituteNode = onSubstituteNode;
            context.onEmitNode = onEmitNode;
            context.enableSubstitution(206 /* CallExpression */); // Substitute calls to imported/exported symbols to avoid incorrect `this`.
            context.enableSubstitution(208 /* TaggedTemplateExpression */); // Substitute calls to imported/exported symbols to avoid incorrect `this`.
            context.enableSubstitution(79 /* Identifier */); // Substitutes expression identifiers with imported/exported symbols.
            context.enableSubstitution(219 /* BinaryExpression */); // Substitutes assignments to exported symbols.
            context.enableSubstitution(292 /* ShorthandPropertyAssignment */); // Substitutes shorthand property assignments for imported/exported symbols.
            context.enableEmitNotification(300 /* SourceFile */); // Restore state when substituting nodes in a file.
            var moduleInfoMap = []; // The ExternalModuleInfo for each file.
            var deferredExports = []; // Exports to defer until an EndOfDeclarationMarker is found.
            var currentSourceFile; // The current file.
            var currentModuleInfo; // The ExternalModuleInfo for the current file.
            var noSubstitution = []; // Set of nodes for which substitution rules should be ignored.
            var needUMDDynamicImportHelper;
            return ts.chainBundle(context, transformSourceFile);
            /**
             * Transforms the module aspects of a SourceFile.
             *
             * @param node The SourceFile node.
             */
            function transformSourceFile(node) {
                if (node.isDeclarationFile ||
                    !(ts.isEffectiveExternalModule(node, compilerOptions) ||
                        node.transformFlags & 4194304 /* ContainsDynamicImport */ ||
                        (ts.isJsonSourceFile(node) && ts.hasJsonModuleEmitEnabled(compilerOptions) && ts.outFile(compilerOptions)))) {
                    return node;
                }
                currentSourceFile = node;
                currentModuleInfo = ts.collectExternalModuleInfo(context, node, resolver, compilerOptions);
                moduleInfoMap[ts.getOriginalNodeId(node)] = currentModuleInfo;
                // Perform the transformation.
                var transformModule = getTransformModuleDelegate(moduleKind);
                var updated = transformModule(node);
                currentSourceFile = undefined;
                currentModuleInfo = undefined;
                needUMDDynamicImportHelper = false;
                return updated;
            }
            function shouldEmitUnderscoreUnderscoreESModule() {
                if (!currentModuleInfo.exportEquals && ts.isExternalModule(currentSourceFile)) {
                    return true;
                }
                return false;
            }
            /**
             * Transforms a SourceFile into a CommonJS module.
             *
             * @param node The SourceFile node.
             */
            function transformCommonJSModule(node) {
                startLexicalEnvironment();
                var statements = [];
                var ensureUseStrict = ts.getStrictOptionValue(compilerOptions, "alwaysStrict") || (!compilerOptions.noImplicitUseStrict && ts.isExternalModule(currentSourceFile));
                var statementOffset = factory.copyPrologue(node.statements, statements, ensureUseStrict && !ts.isJsonSourceFile(node), topLevelVisitor);
                if (shouldEmitUnderscoreUnderscoreESModule()) {
                    ts.append(statements, createUnderscoreUnderscoreESModule());
                }
                if (ts.length(currentModuleInfo.exportedNames)) {
                    var chunkSize = 50;
                    for (var i = 0; i < currentModuleInfo.exportedNames.length; i += chunkSize) {
                        ts.append(statements, factory.createExpressionStatement(ts.reduceLeft(currentModuleInfo.exportedNames.slice(i, i + chunkSize), function (prev, nextId) { return factory.createAssignment(factory.createPropertyAccessExpression(factory.createIdentifier("exports"), factory.createIdentifier(ts.idText(nextId))), prev); }, factory.createVoidZero())));
                    }
                }
                ts.append(statements, ts.visitNode(currentModuleInfo.externalHelpersImportDeclaration, topLevelVisitor, ts.isStatement));
                ts.addRange(statements, ts.visitNodes(node.statements, topLevelVisitor, ts.isStatement, statementOffset));
                addExportEqualsIfNeeded(statements, /*emitAsReturn*/ false);
                ts.insertStatementsAfterStandardPrologue(statements, endLexicalEnvironment());
                var updated = factory.updateSourceFile(node, ts.setTextRange(factory.createNodeArray(statements), node.statements));
                ts.addEmitHelpers(updated, context.readEmitHelpers());
                return updated;
            }
            /**
             * Transforms a SourceFile into an AMD module.
             *
             * @param node The SourceFile node.
             */
            function transformAMDModule(node) {
                var define = factory.createIdentifier("define");
                var moduleName = ts.tryGetModuleNameFromFile(factory, node, host, compilerOptions);
                var jsonSourceFile = ts.isJsonSourceFile(node) && node;
                // An AMD define function has the following shape:
                //
                //     define(id?, dependencies?, factory);
                //
                // This has the shape of the following:
                //
                //     define(name, ["module1", "module2"], function (module1Alias) { ... }
                //
                // The location of the alias in the parameter list in the factory function needs to
                // match the position of the module name in the dependency list.
                //
                // To ensure this is true in cases of modules with no aliases, e.g.:
                //
                //     import "module"
                //
                // or
                //
                //     /// <amd-dependency path= "a.css" />
                //
                // we need to add modules without alias names to the end of the dependencies list
                var _a = collectAsynchronousDependencies(node, /*includeNonAmdDependencies*/ true), aliasedModuleNames = _a.aliasedModuleNames, unaliasedModuleNames = _a.unaliasedModuleNames, importAliasNames = _a.importAliasNames;
                // Create an updated SourceFile:
                //
                //     define(mofactory.updateSourceFile", "module2"], function ...
                var updated = factory.updateSourceFile(node, ts.setTextRange(factory.createNodeArray([
                    factory.createExpressionStatement(factory.createCallExpression(define, 
                    /*typeArguments*/ undefined, __spreadArray(__spreadArray([], (moduleName ? [moduleName] : []), true), [
                        // Add the dependency array argument:
                        //
                        //     ["require", "exports", module1", "module2", ...]
                        factory.createArrayLiteralExpression(jsonSourceFile ? ts.emptyArray : __spreadArray(__spreadArray([
                            factory.createStringLiteral("require"),
                            factory.createStringLiteral("exports")
                        ], aliasedModuleNames, true), unaliasedModuleNames, true)),
                        // Add the module body function argument:
                        //
                        //     function (require, exports, module1, module2) ...
                        jsonSourceFile ?
                            jsonSourceFile.statements.length ? jsonSourceFile.statements[0].expression : factory.createObjectLiteralExpression() :
                            factory.createFunctionExpression(
                            /*modifiers*/ undefined, 
                            /*asteriskToken*/ undefined, 
                            /*name*/ undefined, 
                            /*typeParameters*/ undefined, __spreadArray([
                                factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, "require"),
                                factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, "exports")
                            ], importAliasNames, true), 
                            /*type*/ undefined, transformAsynchronousModuleBody(node))
                    ], false)))
                ]), 
                /*location*/ node.statements));
                ts.addEmitHelpers(updated, context.readEmitHelpers());
                return updated;
            }
            /**
             * Transforms a SourceFile into a UMD module.
             *
             * @param node The SourceFile node.
             */
            function transformUMDModule(node) {
                var _a = collectAsynchronousDependencies(node, /*includeNonAmdDependencies*/ false), aliasedModuleNames = _a.aliasedModuleNames, unaliasedModuleNames = _a.unaliasedModuleNames, importAliasNames = _a.importAliasNames;
                var moduleName = ts.tryGetModuleNameFromFile(factory, node, host, compilerOptions);
                var umdHeader = factory.createFunctionExpression(
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, [factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, "factory")], 
                /*type*/ undefined, ts.setTextRange(factory.createBlock([
                    factory.createIfStatement(factory.createLogicalAnd(factory.createTypeCheck(factory.createIdentifier("module"), "object"), factory.createTypeCheck(factory.createPropertyAccessExpression(factory.createIdentifier("module"), "exports"), "object")), factory.createBlock([
                        factory.createVariableStatement(
                        /*modifiers*/ undefined, [
                            factory.createVariableDeclaration("v", 
                            /*exclamationToken*/ undefined, 
                            /*type*/ undefined, factory.createCallExpression(factory.createIdentifier("factory"), 
                            /*typeArguments*/ undefined, [
                                factory.createIdentifier("require"),
                                factory.createIdentifier("exports")
                            ]))
                        ]),
                        ts.setEmitFlags(factory.createIfStatement(factory.createStrictInequality(factory.createIdentifier("v"), factory.createIdentifier("undefined")), factory.createExpressionStatement(factory.createAssignment(factory.createPropertyAccessExpression(factory.createIdentifier("module"), "exports"), factory.createIdentifier("v")))), 1 /* SingleLine */)
                    ]), factory.createIfStatement(factory.createLogicalAnd(factory.createTypeCheck(factory.createIdentifier("define"), "function"), factory.createPropertyAccessExpression(factory.createIdentifier("define"), "amd")), factory.createBlock([
                        factory.createExpressionStatement(factory.createCallExpression(factory.createIdentifier("define"), 
                        /*typeArguments*/ undefined, __spreadArray(__spreadArray([], (moduleName ? [moduleName] : []), true), [
                            factory.createArrayLiteralExpression(__spreadArray(__spreadArray([
                                factory.createStringLiteral("require"),
                                factory.createStringLiteral("exports")
                            ], aliasedModuleNames, true), unaliasedModuleNames, true)),
                            factory.createIdentifier("factory")
                        ], false)))
                    ])))
                ], 
                /*multiLine*/ true), 
                /*location*/ undefined));
                // Create an updated SourceFile:
                //
                //  (function (factory) {
                //      if (typeof module === "object" && typeof module.exports === "object") {
                //          var v = factory(require, exports);
                //          if (v !== undefined) module.exports = v;
                //      }
                //      else if (typeof define === 'function' && define.amd) {
                //          define(["require", "exports"], factory);
                //      }
                //  })(function ...)
                var updated = factory.updateSourceFile(node, ts.setTextRange(factory.createNodeArray([
                    factory.createExpressionStatement(factory.createCallExpression(umdHeader, 
                    /*typeArguments*/ undefined, [
                        // Add the module body function argument:
                        //
                        //     function (require, exports) ...
                        factory.createFunctionExpression(
                        /*modifiers*/ undefined, 
                        /*asteriskToken*/ undefined, 
                        /*name*/ undefined, 
                        /*typeParameters*/ undefined, __spreadArray([
                            factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, "require"),
                            factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, "exports")
                        ], importAliasNames, true), 
                        /*type*/ undefined, transformAsynchronousModuleBody(node))
                    ]))
                ]), 
                /*location*/ node.statements));
                ts.addEmitHelpers(updated, context.readEmitHelpers());
                return updated;
            }
            /**
             * Collect the additional asynchronous dependencies for the module.
             *
             * @param node The source file.
             * @param includeNonAmdDependencies A value indicating whether to include non-AMD dependencies.
             */
            function collectAsynchronousDependencies(node, includeNonAmdDependencies) {
                // names of modules with corresponding parameter in the factory function
                var aliasedModuleNames = [];
                // names of modules with no corresponding parameters in factory function
                var unaliasedModuleNames = [];
                // names of the parameters in the factory function; these
                // parameters need to match the indexes of the corresponding
                // module names in aliasedModuleNames.
                var importAliasNames = [];
                // Fill in amd-dependency tags
                for (var _i = 0, _a = node.amdDependencies; _i < _a.length; _i++) {
                    var amdDependency = _a[_i];
                    if (amdDependency.name) {
                        aliasedModuleNames.push(factory.createStringLiteral(amdDependency.path));
                        importAliasNames.push(factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, amdDependency.name));
                    }
                    else {
                        unaliasedModuleNames.push(factory.createStringLiteral(amdDependency.path));
                    }
                }
                for (var _b = 0, _c = currentModuleInfo.externalImports; _b < _c.length; _b++) {
                    var importNode = _c[_b];
                    // Find the name of the external module
                    var externalModuleName = ts.getExternalModuleNameLiteral(factory, importNode, currentSourceFile, host, resolver, compilerOptions);
                    // Find the name of the module alias, if there is one
                    var importAliasName = ts.getLocalNameForExternalImport(factory, importNode, currentSourceFile);
                    // It is possible that externalModuleName is undefined if it is not string literal.
                    // This can happen in the invalid import syntax.
                    // E.g : "import * from alias from 'someLib';"
                    if (externalModuleName) {
                        if (includeNonAmdDependencies && importAliasName) {
                            // Set emitFlags on the name of the classDeclaration
                            // This is so that when printer will not substitute the identifier
                            ts.setEmitFlags(importAliasName, 4 /* NoSubstitution */);
                            aliasedModuleNames.push(externalModuleName);
                            importAliasNames.push(factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, importAliasName));
                        }
                        else {
                            unaliasedModuleNames.push(externalModuleName);
                        }
                    }
                }
                return { aliasedModuleNames: aliasedModuleNames, unaliasedModuleNames: unaliasedModuleNames, importAliasNames: importAliasNames };
            }
            function getAMDImportExpressionForImport(node) {
                if (ts.isImportEqualsDeclaration(node) || ts.isExportDeclaration(node) || !ts.getExternalModuleNameLiteral(factory, node, currentSourceFile, host, resolver, compilerOptions)) {
                    return undefined;
                }
                var name = ts.getLocalNameForExternalImport(factory, node, currentSourceFile); // TODO: GH#18217
                var expr = getHelperExpressionForImport(node, name);
                if (expr === name) {
                    return undefined;
                }
                return factory.createExpressionStatement(factory.createAssignment(name, expr));
            }
            /**
             * Transforms a SourceFile into an AMD or UMD module body.
             *
             * @param node The SourceFile node.
             */
            function transformAsynchronousModuleBody(node) {
                startLexicalEnvironment();
                var statements = [];
                var statementOffset = factory.copyPrologue(node.statements, statements, /*ensureUseStrict*/ !compilerOptions.noImplicitUseStrict, topLevelVisitor);
                if (shouldEmitUnderscoreUnderscoreESModule()) {
                    ts.append(statements, createUnderscoreUnderscoreESModule());
                }
                if (ts.length(currentModuleInfo.exportedNames)) {
                    ts.append(statements, factory.createExpressionStatement(ts.reduceLeft(currentModuleInfo.exportedNames, function (prev, nextId) { return factory.createAssignment(factory.createPropertyAccessExpression(factory.createIdentifier("exports"), factory.createIdentifier(ts.idText(nextId))), prev); }, factory.createVoidZero())));
                }
                // Visit each statement of the module body.
                ts.append(statements, ts.visitNode(currentModuleInfo.externalHelpersImportDeclaration, topLevelVisitor, ts.isStatement));
                if (moduleKind === ts.ModuleKind.AMD) {
                    ts.addRange(statements, ts.mapDefined(currentModuleInfo.externalImports, getAMDImportExpressionForImport));
                }
                ts.addRange(statements, ts.visitNodes(node.statements, topLevelVisitor, ts.isStatement, statementOffset));
                // Append the 'export =' statement if provided.
                addExportEqualsIfNeeded(statements, /*emitAsReturn*/ true);
                // End the lexical environment for the module body
                // and merge any new lexical declarations.
                ts.insertStatementsAfterStandardPrologue(statements, endLexicalEnvironment());
                var body = factory.createBlock(statements, /*multiLine*/ true);
                if (needUMDDynamicImportHelper) {
                    ts.addEmitHelper(body, dynamicImportUMDHelper);
                }
                return body;
            }
            /**
             * Adds the down-level representation of `export=` to the statement list if one exists
             * in the source file.
             *
             * @param statements The Statement list to modify.
             * @param emitAsReturn A value indicating whether to emit the `export=` statement as a
             * return statement.
             */
            function addExportEqualsIfNeeded(statements, emitAsReturn) {
                if (currentModuleInfo.exportEquals) {
                    var expressionResult = ts.visitNode(currentModuleInfo.exportEquals.expression, visitor);
                    if (expressionResult) {
                        if (emitAsReturn) {
                            var statement = factory.createReturnStatement(expressionResult);
                            ts.setTextRange(statement, currentModuleInfo.exportEquals);
                            ts.setEmitFlags(statement, 384 /* NoTokenSourceMaps */ | 1536 /* NoComments */);
                            statements.push(statement);
                        }
                        else {
                            var statement = factory.createExpressionStatement(factory.createAssignment(factory.createPropertyAccessExpression(factory.createIdentifier("module"), "exports"), expressionResult));
                            ts.setTextRange(statement, currentModuleInfo.exportEquals);
                            ts.setEmitFlags(statement, 1536 /* NoComments */);
                            statements.push(statement);
                        }
                    }
                }
            }
            //
            // Top-Level Source Element Visitors
            //
            /**
             * Visits a node at the top level of the source file.
             *
             * @param node The node to visit.
             */
            function topLevelVisitor(node) {
                switch (node.kind) {
                    case 264 /* ImportDeclaration */:
                        return visitImportDeclaration(node);
                    case 263 /* ImportEqualsDeclaration */:
                        return visitImportEqualsDeclaration(node);
                    case 270 /* ExportDeclaration */:
                        return visitExportDeclaration(node);
                    case 269 /* ExportAssignment */:
                        return visitExportAssignment(node);
                    case 235 /* VariableStatement */:
                        return visitVariableStatement(node);
                    case 254 /* FunctionDeclaration */:
                        return visitFunctionDeclaration(node);
                    case 255 /* ClassDeclaration */:
                        return visitClassDeclaration(node);
                    case 347 /* MergeDeclarationMarker */:
                        return visitMergeDeclarationMarker(node);
                    case 348 /* EndOfDeclarationMarker */:
                        return visitEndOfDeclarationMarker(node);
                    default:
                        return visitor(node);
                }
            }
            function visitorWorker(node, valueIsDiscarded) {
                // This visitor does not need to descend into the tree if there is no dynamic import, destructuring assignment, or update expression
                // as export/import statements are only transformed at the top level of a file.
                if (!(node.transformFlags & (4194304 /* ContainsDynamicImport */ | 2048 /* ContainsDestructuringAssignment */ | 67108864 /* ContainsUpdateExpressionForIdentifier */))) {
                    return node;
                }
                switch (node.kind) {
                    case 240 /* ForStatement */:
                        return visitForStatement(node);
                    case 236 /* ExpressionStatement */:
                        return visitExpressionStatement(node);
                    case 210 /* ParenthesizedExpression */:
                        return visitParenthesizedExpression(node, valueIsDiscarded);
                    case 345 /* PartiallyEmittedExpression */:
                        return visitPartiallyEmittedExpression(node, valueIsDiscarded);
                    case 206 /* CallExpression */:
                        if (ts.isImportCall(node)) {
                            return visitImportCallExpression(node);
                        }
                        break;
                    case 219 /* BinaryExpression */:
                        if (ts.isDestructuringAssignment(node)) {
                            return visitDestructuringAssignment(node, valueIsDiscarded);
                        }
                        break;
                    case 217 /* PrefixUnaryExpression */:
                    case 218 /* PostfixUnaryExpression */:
                        return visitPreOrPostfixUnaryExpression(node, valueIsDiscarded);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitor(node) {
                return visitorWorker(node, /*valueIsDiscarded*/ false);
            }
            function discardedValueVisitor(node) {
                return visitorWorker(node, /*valueIsDiscarded*/ true);
            }
            function destructuringNeedsFlattening(node) {
                if (ts.isObjectLiteralExpression(node)) {
                    for (var _i = 0, _a = node.properties; _i < _a.length; _i++) {
                        var elem = _a[_i];
                        switch (elem.kind) {
                            case 291 /* PropertyAssignment */:
                                if (destructuringNeedsFlattening(elem.initializer)) {
                                    return true;
                                }
                                break;
                            case 292 /* ShorthandPropertyAssignment */:
                                if (destructuringNeedsFlattening(elem.name)) {
                                    return true;
                                }
                                break;
                            case 293 /* SpreadAssignment */:
                                if (destructuringNeedsFlattening(elem.expression)) {
                                    return true;
                                }
                                break;
                            case 167 /* MethodDeclaration */:
                            case 170 /* GetAccessor */:
                            case 171 /* SetAccessor */:
                                return false;
                            default: ts.Debug.assertNever(elem, "Unhandled object member kind");
                        }
                    }
                }
                else if (ts.isArrayLiteralExpression(node)) {
                    for (var _b = 0, _c = node.elements; _b < _c.length; _b++) {
                        var elem = _c[_b];
                        if (ts.isSpreadElement(elem)) {
                            if (destructuringNeedsFlattening(elem.expression)) {
                                return true;
                            }
                        }
                        else if (destructuringNeedsFlattening(elem)) {
                            return true;
                        }
                    }
                }
                else if (ts.isIdentifier(node)) {
                    return ts.length(getExports(node)) > (ts.isExportName(node) ? 1 : 0);
                }
                return false;
            }
            function visitDestructuringAssignment(node, valueIsDiscarded) {
                if (destructuringNeedsFlattening(node.left)) {
                    return ts.flattenDestructuringAssignment(node, visitor, context, 0 /* All */, !valueIsDiscarded, createAllExportExpressions);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitForStatement(node) {
                return factory.updateForStatement(node, ts.visitNode(node.initializer, discardedValueVisitor, ts.isForInitializer), ts.visitNode(node.condition, visitor, ts.isExpression), ts.visitNode(node.incrementor, discardedValueVisitor, ts.isExpression), ts.visitIterationBody(node.statement, visitor, context));
            }
            function visitExpressionStatement(node) {
                return factory.updateExpressionStatement(node, ts.visitNode(node.expression, discardedValueVisitor, ts.isExpression));
            }
            function visitParenthesizedExpression(node, valueIsDiscarded) {
                return factory.updateParenthesizedExpression(node, ts.visitNode(node.expression, valueIsDiscarded ? discardedValueVisitor : visitor, ts.isExpression));
            }
            function visitPartiallyEmittedExpression(node, valueIsDiscarded) {
                return factory.updatePartiallyEmittedExpression(node, ts.visitNode(node.expression, valueIsDiscarded ? discardedValueVisitor : visitor, ts.isExpression));
            }
            function visitPreOrPostfixUnaryExpression(node, valueIsDiscarded) {
                // When we see a prefix or postfix increment expression whose operand is an exported
                // symbol, we should ensure all exports of that symbol are updated with the correct
                // value.
                //
                // - We do not transform generated identifiers for any reason.
                // - We do not transform identifiers tagged with the LocalName flag.
                // - We do not transform identifiers that were originally the name of an enum or
                //   namespace due to how they are transformed in TypeScript.
                // - We only transform identifiers that are exported at the top level.
                if ((node.operator === 45 /* PlusPlusToken */ || node.operator === 46 /* MinusMinusToken */)
                    && ts.isIdentifier(node.operand)
                    && !ts.isGeneratedIdentifier(node.operand)
                    && !ts.isLocalName(node.operand)
                    && !ts.isDeclarationNameOfEnumOrNamespace(node.operand)) {
                    var exportedNames = getExports(node.operand);
                    if (exportedNames) {
                        var temp = void 0;
                        var expression = ts.visitNode(node.operand, visitor, ts.isExpression);
                        if (ts.isPrefixUnaryExpression(node)) {
                            expression = factory.updatePrefixUnaryExpression(node, expression);
                        }
                        else {
                            expression = factory.updatePostfixUnaryExpression(node, expression);
                            if (!valueIsDiscarded) {
                                temp = factory.createTempVariable(hoistVariableDeclaration);
                                expression = factory.createAssignment(temp, expression);
                                ts.setTextRange(expression, node);
                            }
                            expression = factory.createComma(expression, factory.cloneNode(node.operand));
                            ts.setTextRange(expression, node);
                        }
                        for (var _i = 0, exportedNames_1 = exportedNames; _i < exportedNames_1.length; _i++) {
                            var exportName = exportedNames_1[_i];
                            noSubstitution[ts.getNodeId(expression)] = true;
                            expression = createExportExpression(exportName, expression);
                            ts.setTextRange(expression, node);
                        }
                        if (temp) {
                            noSubstitution[ts.getNodeId(expression)] = true;
                            expression = factory.createComma(expression, temp);
                            ts.setTextRange(expression, node);
                        }
                        return expression;
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            function visitImportCallExpression(node) {
                var externalModuleName = ts.getExternalModuleNameLiteral(factory, node, currentSourceFile, host, resolver, compilerOptions);
                var firstArgument = ts.visitNode(ts.firstOrUndefined(node.arguments), visitor);
                // Only use the external module name if it differs from the first argument. This allows us to preserve the quote style of the argument on output.
                var argument = externalModuleName && (!firstArgument || !ts.isStringLiteral(firstArgument) || firstArgument.text !== externalModuleName.text) ? externalModuleName : firstArgument;
                var containsLexicalThis = !!(node.transformFlags & 8192 /* ContainsLexicalThis */);
                switch (compilerOptions.module) {
                    case ts.ModuleKind.AMD:
                        return createImportCallExpressionAMD(argument, containsLexicalThis);
                    case ts.ModuleKind.UMD:
                        return createImportCallExpressionUMD(argument !== null && argument !== void 0 ? argument : factory.createVoidZero(), containsLexicalThis);
                    case ts.ModuleKind.CommonJS:
                    default:
                        return createImportCallExpressionCommonJS(argument, containsLexicalThis);
                }
            }
            function createImportCallExpressionUMD(arg, containsLexicalThis) {
                // (function (factory) {
                //      ... (regular UMD)
                // }
                // })(function (require, exports, useSyncRequire) {
                //      "use strict";
                //      Object.defineProperty(exports, "__esModule", { value: true });
                //      var __syncRequire = typeof module === "object" && typeof module.exports === "object";
                //      var __resolved = new Promise(function (resolve) { resolve(); });
                //      .....
                //      __syncRequire
                //          ? __resolved.then(function () { return require(x); }) /*CommonJs Require*/
                //          : new Promise(function (_a, _b) { require([x], _a, _b); }); /*Amd Require*/
                // });
                needUMDDynamicImportHelper = true;
                if (ts.isSimpleCopiableExpression(arg)) {
                    var argClone = ts.isGeneratedIdentifier(arg) ? arg : ts.isStringLiteral(arg) ? factory.createStringLiteralFromNode(arg) : ts.setEmitFlags(ts.setTextRange(factory.cloneNode(arg), arg), 1536 /* NoComments */);
                    return factory.createConditionalExpression(
                    /*condition*/ factory.createIdentifier("__syncRequire"), 
                    /*questionToken*/ undefined, 
                    /*whenTrue*/ createImportCallExpressionCommonJS(arg, containsLexicalThis), 
                    /*colonToken*/ undefined, 
                    /*whenFalse*/ createImportCallExpressionAMD(argClone, containsLexicalThis));
                }
                else {
                    var temp = factory.createTempVariable(hoistVariableDeclaration);
                    return factory.createComma(factory.createAssignment(temp, arg), factory.createConditionalExpression(
                    /*condition*/ factory.createIdentifier("__syncRequire"), 
                    /*questionToken*/ undefined, 
                    /*whenTrue*/ createImportCallExpressionCommonJS(temp, containsLexicalThis), 
                    /*colonToken*/ undefined, 
                    /*whenFalse*/ createImportCallExpressionAMD(temp, containsLexicalThis)));
                }
            }
            function createImportCallExpressionAMD(arg, containsLexicalThis) {
                // improt("./blah")
                // emit as
                // define(["require", "exports", "blah"], function (require, exports) {
                //     ...
                //     new Promise(function (_a, _b) { require([x], _a, _b); }); /*Amd Require*/
                // });
                var resolve = factory.createUniqueName("resolve");
                var reject = factory.createUniqueName("reject");
                var parameters = [
                    factory.createParameterDeclaration(/*decorator*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, /*name*/ resolve),
                    factory.createParameterDeclaration(/*decorator*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, /*name*/ reject)
                ];
                var body = factory.createBlock([
                    factory.createExpressionStatement(factory.createCallExpression(factory.createIdentifier("require"), 
                    /*typeArguments*/ undefined, [factory.createArrayLiteralExpression([arg || factory.createOmittedExpression()]), resolve, reject]))
                ]);
                var func;
                if (languageVersion >= 2 /* ES2015 */) {
                    func = factory.createArrowFunction(
                    /*modifiers*/ undefined, 
                    /*typeParameters*/ undefined, parameters, 
                    /*type*/ undefined, 
                    /*equalsGreaterThanToken*/ undefined, body);
                }
                else {
                    func = factory.createFunctionExpression(
                    /*modifiers*/ undefined, 
                    /*asteriskToken*/ undefined, 
                    /*name*/ undefined, 
                    /*typeParameters*/ undefined, parameters, 
                    /*type*/ undefined, body);
                    // if there is a lexical 'this' in the import call arguments, ensure we indicate
                    // that this new function expression indicates it captures 'this' so that the
                    // es2015 transformer will properly substitute 'this' with '_this'.
                    if (containsLexicalThis) {
                        ts.setEmitFlags(func, 8 /* CapturesThis */);
                    }
                }
                var promise = factory.createNewExpression(factory.createIdentifier("Promise"), /*typeArguments*/ undefined, [func]);
                if (compilerOptions.esModuleInterop) {
                    return factory.createCallExpression(factory.createPropertyAccessExpression(promise, factory.createIdentifier("then")), /*typeArguments*/ undefined, [emitHelpers().createImportStarCallbackHelper()]);
                }
                return promise;
            }
            function createImportCallExpressionCommonJS(arg, containsLexicalThis) {
                // import("./blah")
                // emit as
                // Promise.resolve().then(function () { return require(x); }) /*CommonJs Require*/
                // We have to wrap require in then callback so that require is done in asynchronously
                // if we simply do require in resolve callback in Promise constructor. We will execute the loading immediately
                var promiseResolveCall = factory.createCallExpression(factory.createPropertyAccessExpression(factory.createIdentifier("Promise"), "resolve"), /*typeArguments*/ undefined, /*argumentsArray*/ []);
                var requireCall = factory.createCallExpression(factory.createIdentifier("require"), /*typeArguments*/ undefined, arg ? [arg] : []);
                if (compilerOptions.esModuleInterop) {
                    requireCall = emitHelpers().createImportStarHelper(requireCall);
                }
                var func;
                if (languageVersion >= 2 /* ES2015 */) {
                    func = factory.createArrowFunction(
                    /*modifiers*/ undefined, 
                    /*typeParameters*/ undefined, 
                    /*parameters*/ [], 
                    /*type*/ undefined, 
                    /*equalsGreaterThanToken*/ undefined, requireCall);
                }
                else {
                    func = factory.createFunctionExpression(
                    /*modifiers*/ undefined, 
                    /*asteriskToken*/ undefined, 
                    /*name*/ undefined, 
                    /*typeParameters*/ undefined, 
                    /*parameters*/ [], 
                    /*type*/ undefined, factory.createBlock([factory.createReturnStatement(requireCall)]));
                    // if there is a lexical 'this' in the import call arguments, ensure we indicate
                    // that this new function expression indicates it captures 'this' so that the
                    // es2015 transformer will properly substitute 'this' with '_this'.
                    if (containsLexicalThis) {
                        ts.setEmitFlags(func, 8 /* CapturesThis */);
                    }
                }
                return factory.createCallExpression(factory.createPropertyAccessExpression(promiseResolveCall, "then"), /*typeArguments*/ undefined, [func]);
            }
            function getHelperExpressionForExport(node, innerExpr) {
                if (!compilerOptions.esModuleInterop || ts.getEmitFlags(node) & 67108864 /* NeverApplyImportHelper */) {
                    return innerExpr;
                }
                if (ts.getExportNeedsImportStarHelper(node)) {
                    return emitHelpers().createImportStarHelper(innerExpr);
                }
                return innerExpr;
            }
            function getHelperExpressionForImport(node, innerExpr) {
                if (!compilerOptions.esModuleInterop || ts.getEmitFlags(node) & 67108864 /* NeverApplyImportHelper */) {
                    return innerExpr;
                }
                if (ts.getImportNeedsImportStarHelper(node)) {
                    return emitHelpers().createImportStarHelper(innerExpr);
                }
                if (ts.getImportNeedsImportDefaultHelper(node)) {
                    return emitHelpers().createImportDefaultHelper(innerExpr);
                }
                return innerExpr;
            }
            /**
             * Visits an ImportDeclaration node.
             *
             * @param node The node to visit.
             */
            function visitImportDeclaration(node) {
                var statements;
                var namespaceDeclaration = ts.getNamespaceDeclarationNode(node);
                if (moduleKind !== ts.ModuleKind.AMD) {
                    if (!node.importClause) {
                        // import "mod";
                        return ts.setOriginalNode(ts.setTextRange(factory.createExpressionStatement(createRequireCall(node)), node), node);
                    }
                    else {
                        var variables = [];
                        if (namespaceDeclaration && !ts.isDefaultImport(node)) {
                            // import * as n from "mod";
                            variables.push(factory.createVariableDeclaration(factory.cloneNode(namespaceDeclaration.name), 
                            /*exclamationToken*/ undefined, 
                            /*type*/ undefined, getHelperExpressionForImport(node, createRequireCall(node))));
                        }
                        else {
                            // import d from "mod";
                            // import { x, y } from "mod";
                            // import d, { x, y } from "mod";
                            // import d, * as n from "mod";
                            variables.push(factory.createVariableDeclaration(factory.getGeneratedNameForNode(node), 
                            /*exclamationToken*/ undefined, 
                            /*type*/ undefined, getHelperExpressionForImport(node, createRequireCall(node))));
                            if (namespaceDeclaration && ts.isDefaultImport(node)) {
                                variables.push(factory.createVariableDeclaration(factory.cloneNode(namespaceDeclaration.name), 
                                /*exclamationToken*/ undefined, 
                                /*type*/ undefined, factory.getGeneratedNameForNode(node)));
                            }
                        }
                        statements = ts.append(statements, ts.setOriginalNode(ts.setTextRange(factory.createVariableStatement(
                        /*modifiers*/ undefined, factory.createVariableDeclarationList(variables, languageVersion >= 2 /* ES2015 */ ? 2 /* Const */ : 0 /* None */)), 
                        /*location*/ node), 
                        /*original*/ node));
                    }
                }
                else if (namespaceDeclaration && ts.isDefaultImport(node)) {
                    // import d, * as n from "mod";
                    statements = ts.append(statements, factory.createVariableStatement(
                    /*modifiers*/ undefined, factory.createVariableDeclarationList([
                        ts.setOriginalNode(ts.setTextRange(factory.createVariableDeclaration(factory.cloneNode(namespaceDeclaration.name), 
                        /*exclamationToken*/ undefined, 
                        /*type*/ undefined, factory.getGeneratedNameForNode(node)), 
                        /*location*/ node), 
                        /*original*/ node)
                    ], languageVersion >= 2 /* ES2015 */ ? 2 /* Const */ : 0 /* None */)));
                }
                if (hasAssociatedEndOfDeclarationMarker(node)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfImportDeclaration(deferredExports[id], node);
                }
                else {
                    statements = appendExportsOfImportDeclaration(statements, node);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Creates a `require()` call to import an external module.
             *
             * @param importNode The declararation to import.
             */
            function createRequireCall(importNode) {
                var moduleName = ts.getExternalModuleNameLiteral(factory, importNode, currentSourceFile, host, resolver, compilerOptions);
                var args = [];
                if (moduleName) {
                    args.push(moduleName);
                }
                return factory.createCallExpression(factory.createIdentifier("require"), /*typeArguments*/ undefined, args);
            }
            /**
             * Visits an ImportEqualsDeclaration node.
             *
             * @param node The node to visit.
             */
            function visitImportEqualsDeclaration(node) {
                ts.Debug.assert(ts.isExternalModuleImportEqualsDeclaration(node), "import= for internal module references should be handled in an earlier transformer.");
                var statements;
                if (moduleKind !== ts.ModuleKind.AMD) {
                    if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                        statements = ts.append(statements, ts.setOriginalNode(ts.setTextRange(factory.createExpressionStatement(createExportExpression(node.name, createRequireCall(node))), node), node));
                    }
                    else {
                        statements = ts.append(statements, ts.setOriginalNode(ts.setTextRange(factory.createVariableStatement(
                        /*modifiers*/ undefined, factory.createVariableDeclarationList([
                            factory.createVariableDeclaration(factory.cloneNode(node.name), 
                            /*exclamationToken*/ undefined, 
                            /*type*/ undefined, createRequireCall(node))
                        ], 
                        /*flags*/ languageVersion >= 2 /* ES2015 */ ? 2 /* Const */ : 0 /* None */)), node), node));
                    }
                }
                else {
                    if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                        statements = ts.append(statements, ts.setOriginalNode(ts.setTextRange(factory.createExpressionStatement(createExportExpression(factory.getExportName(node), factory.getLocalName(node))), node), node));
                    }
                }
                if (hasAssociatedEndOfDeclarationMarker(node)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfImportEqualsDeclaration(deferredExports[id], node);
                }
                else {
                    statements = appendExportsOfImportEqualsDeclaration(statements, node);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Visits an ExportDeclaration node.
             *
             * @param The node to visit.
             */
            function visitExportDeclaration(node) {
                if (!node.moduleSpecifier) {
                    // Elide export declarations with no module specifier as they are handled
                    // elsewhere.
                    return undefined;
                }
                var generatedName = factory.getGeneratedNameForNode(node);
                if (node.exportClause && ts.isNamedExports(node.exportClause)) {
                    var statements = [];
                    // export { x, y } from "mod";
                    if (moduleKind !== ts.ModuleKind.AMD) {
                        statements.push(ts.setOriginalNode(ts.setTextRange(factory.createVariableStatement(
                        /*modifiers*/ undefined, factory.createVariableDeclarationList([
                            factory.createVariableDeclaration(generatedName, 
                            /*exclamationToken*/ undefined, 
                            /*type*/ undefined, createRequireCall(node))
                        ])), 
                        /*location*/ node), 
                        /* original */ node));
                    }
                    for (var _i = 0, _a = node.exportClause.elements; _i < _a.length; _i++) {
                        var specifier = _a[_i];
                        if (languageVersion === 0 /* ES3 */) {
                            statements.push(ts.setOriginalNode(ts.setTextRange(factory.createExpressionStatement(emitHelpers().createCreateBindingHelper(generatedName, factory.createStringLiteralFromNode(specifier.propertyName || specifier.name), specifier.propertyName ? factory.createStringLiteralFromNode(specifier.name) : undefined)), specifier), specifier));
                        }
                        else {
                            var exportNeedsImportDefault = !!compilerOptions.esModuleInterop &&
                                !(ts.getEmitFlags(node) & 67108864 /* NeverApplyImportHelper */) &&
                                ts.idText(specifier.propertyName || specifier.name) === "default";
                            var exportedValue = factory.createPropertyAccessExpression(exportNeedsImportDefault ? emitHelpers().createImportDefaultHelper(generatedName) : generatedName, specifier.propertyName || specifier.name);
                            statements.push(ts.setOriginalNode(ts.setTextRange(factory.createExpressionStatement(createExportExpression(factory.getExportName(specifier), exportedValue, /* location */ undefined, /* liveBinding */ true)), specifier), specifier));
                        }
                    }
                    return ts.singleOrMany(statements);
                }
                else if (node.exportClause) {
                    var statements = [];
                    // export * as ns from "mod";
                    // export * as default from "mod";
                    statements.push(ts.setOriginalNode(ts.setTextRange(factory.createExpressionStatement(createExportExpression(factory.cloneNode(node.exportClause.name), getHelperExpressionForExport(node, moduleKind !== ts.ModuleKind.AMD ?
                        createRequireCall(node) :
                        ts.isExportNamespaceAsDefaultDeclaration(node) ? generatedName :
                            factory.createIdentifier(ts.idText(node.exportClause.name))))), node), node));
                    return ts.singleOrMany(statements);
                }
                else {
                    // export * from "mod";
                    return ts.setOriginalNode(ts.setTextRange(factory.createExpressionStatement(emitHelpers().createExportStarHelper(moduleKind !== ts.ModuleKind.AMD ? createRequireCall(node) : generatedName)), node), node);
                }
            }
            /**
             * Visits an ExportAssignment node.
             *
             * @param node The node to visit.
             */
            function visitExportAssignment(node) {
                if (node.isExportEquals) {
                    return undefined;
                }
                var statements;
                var original = node.original;
                if (original && hasAssociatedEndOfDeclarationMarker(original)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportStatement(deferredExports[id], factory.createIdentifier("default"), ts.visitNode(node.expression, visitor), /*location*/ node, /*allowComments*/ true);
                }
                else {
                    statements = appendExportStatement(statements, factory.createIdentifier("default"), ts.visitNode(node.expression, visitor), /*location*/ node, /*allowComments*/ true);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Visits a FunctionDeclaration node.
             *
             * @param node The node to visit.
             */
            function visitFunctionDeclaration(node) {
                var statements;
                if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                    statements = ts.append(statements, ts.setOriginalNode(ts.setTextRange(factory.createFunctionDeclaration(
                    /*decorators*/ undefined, ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), node.asteriskToken, factory.getDeclarationName(node, /*allowComments*/ true, /*allowSourceMaps*/ true), 
                    /*typeParameters*/ undefined, ts.visitNodes(node.parameters, visitor), 
                    /*type*/ undefined, ts.visitEachChild(node.body, visitor, context)), 
                    /*location*/ node), 
                    /*original*/ node));
                }
                else {
                    statements = ts.append(statements, ts.visitEachChild(node, visitor, context));
                }
                if (hasAssociatedEndOfDeclarationMarker(node)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfHoistedDeclaration(deferredExports[id], node);
                }
                else {
                    statements = appendExportsOfHoistedDeclaration(statements, node);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Visits a ClassDeclaration node.
             *
             * @param node The node to visit.
             */
            function visitClassDeclaration(node) {
                var statements;
                if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                    statements = ts.append(statements, ts.setOriginalNode(ts.setTextRange(factory.createClassDeclaration(
                    /*decorators*/ undefined, ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), factory.getDeclarationName(node, /*allowComments*/ true, /*allowSourceMaps*/ true), 
                    /*typeParameters*/ undefined, ts.visitNodes(node.heritageClauses, visitor), ts.visitNodes(node.members, visitor)), node), node));
                }
                else {
                    statements = ts.append(statements, ts.visitEachChild(node, visitor, context));
                }
                if (hasAssociatedEndOfDeclarationMarker(node)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfHoistedDeclaration(deferredExports[id], node);
                }
                else {
                    statements = appendExportsOfHoistedDeclaration(statements, node);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Visits a VariableStatement node.
             *
             * @param node The node to visit.
             */
            function visitVariableStatement(node) {
                var statements;
                var variables;
                var expressions;
                if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                    var modifiers = void 0;
                    var removeCommentsOnExpressions = false;
                    // If we're exporting these variables, then these just become assignments to 'exports.x'.
                    for (var _i = 0, _a = node.declarationList.declarations; _i < _a.length; _i++) {
                        var variable = _a[_i];
                        if (ts.isIdentifier(variable.name) && ts.isLocalName(variable.name)) {
                            if (!modifiers) {
                                modifiers = ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier);
                            }
                            variables = ts.append(variables, variable);
                        }
                        else if (variable.initializer) {
                            if (!ts.isBindingPattern(variable.name) && (ts.isArrowFunction(variable.initializer) || ts.isFunctionExpression(variable.initializer) || ts.isClassExpression(variable.initializer))) {
                                var expression = factory.createAssignment(ts.setTextRange(factory.createPropertyAccessExpression(factory.createIdentifier("exports"), variable.name), 
                                /*location*/ variable.name), factory.createIdentifier(ts.getTextOfIdentifierOrLiteral(variable.name)));
                                var updatedVariable = factory.createVariableDeclaration(variable.name, variable.exclamationToken, variable.type, ts.visitNode(variable.initializer, visitor));
                                variables = ts.append(variables, updatedVariable);
                                expressions = ts.append(expressions, expression);
                                removeCommentsOnExpressions = true;
                            }
                            else {
                                expressions = ts.append(expressions, transformInitializedVariable(variable));
                            }
                        }
                    }
                    if (variables) {
                        statements = ts.append(statements, factory.updateVariableStatement(node, modifiers, factory.updateVariableDeclarationList(node.declarationList, variables)));
                    }
                    if (expressions) {
                        var statement = ts.setOriginalNode(ts.setTextRange(factory.createExpressionStatement(factory.inlineExpressions(expressions)), node), node);
                        if (removeCommentsOnExpressions) {
                            ts.removeAllComments(statement);
                        }
                        statements = ts.append(statements, statement);
                    }
                }
                else {
                    statements = ts.append(statements, ts.visitEachChild(node, visitor, context));
                }
                if (hasAssociatedEndOfDeclarationMarker(node)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfVariableStatement(deferredExports[id], node);
                }
                else {
                    statements = appendExportsOfVariableStatement(statements, node);
                }
                return ts.singleOrMany(statements);
            }
            function createAllExportExpressions(name, value, location) {
                var exportedNames = getExports(name);
                if (exportedNames) {
                    // For each additional export of the declaration, apply an export assignment.
                    var expression = ts.isExportName(name) ? value : factory.createAssignment(name, value);
                    for (var _i = 0, exportedNames_2 = exportedNames; _i < exportedNames_2.length; _i++) {
                        var exportName = exportedNames_2[_i];
                        // Mark the node to prevent triggering substitution.
                        ts.setEmitFlags(expression, 4 /* NoSubstitution */);
                        expression = createExportExpression(exportName, expression, /*location*/ location);
                    }
                    return expression;
                }
                return factory.createAssignment(name, value);
            }
            /**
             * Transforms an exported variable with an initializer into an expression.
             *
             * @param node The node to transform.
             */
            function transformInitializedVariable(node) {
                if (ts.isBindingPattern(node.name)) {
                    return ts.flattenDestructuringAssignment(ts.visitNode(node, visitor), 
                    /*visitor*/ undefined, context, 0 /* All */, 
                    /*needsValue*/ false, createAllExportExpressions);
                }
                else {
                    return factory.createAssignment(ts.setTextRange(factory.createPropertyAccessExpression(factory.createIdentifier("exports"), node.name), 
                    /*location*/ node.name), node.initializer ? ts.visitNode(node.initializer, visitor) : factory.createVoidZero());
                }
            }
            /**
             * Visits a MergeDeclarationMarker used as a placeholder for the beginning of a merged
             * and transformed declaration.
             *
             * @param node The node to visit.
             */
            function visitMergeDeclarationMarker(node) {
                // For an EnumDeclaration or ModuleDeclaration that merges with a preceeding
                // declaration we do not emit a leading variable declaration. To preserve the
                // begin/end semantics of the declararation and to properly handle exports
                // we wrapped the leading variable declaration in a `MergeDeclarationMarker`.
                //
                // To balance the declaration, add the exports of the elided variable
                // statement.
                if (hasAssociatedEndOfDeclarationMarker(node) && node.original.kind === 235 /* VariableStatement */) {
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfVariableStatement(deferredExports[id], node.original);
                }
                return node;
            }
            /**
             * Determines whether a node has an associated EndOfDeclarationMarker.
             *
             * @param node The node to test.
             */
            function hasAssociatedEndOfDeclarationMarker(node) {
                return (ts.getEmitFlags(node) & 4194304 /* HasEndOfDeclarationMarker */) !== 0;
            }
            /**
             * Visits a DeclarationMarker used as a placeholder for the end of a transformed
             * declaration.
             *
             * @param node The node to visit.
             */
            function visitEndOfDeclarationMarker(node) {
                // For some transformations we emit an `EndOfDeclarationMarker` to mark the actual
                // end of the transformed declaration. We use this marker to emit any deferred exports
                // of the declaration.
                var id = ts.getOriginalNodeId(node);
                var statements = deferredExports[id];
                if (statements) {
                    delete deferredExports[id];
                    return ts.append(statements, node);
                }
                return node;
            }
            /**
             * Appends the exports of an ImportDeclaration to a statement list, returning the
             * statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration whose exports are to be recorded.
             */
            function appendExportsOfImportDeclaration(statements, decl) {
                if (currentModuleInfo.exportEquals) {
                    return statements;
                }
                var importClause = decl.importClause;
                if (!importClause) {
                    return statements;
                }
                if (importClause.name) {
                    statements = appendExportsOfDeclaration(statements, importClause);
                }
                var namedBindings = importClause.namedBindings;
                if (namedBindings) {
                    switch (namedBindings.kind) {
                        case 266 /* NamespaceImport */:
                            statements = appendExportsOfDeclaration(statements, namedBindings);
                            break;
                        case 267 /* NamedImports */:
                            for (var _i = 0, _a = namedBindings.elements; _i < _a.length; _i++) {
                                var importBinding = _a[_i];
                                statements = appendExportsOfDeclaration(statements, importBinding, /* liveBinding */ true);
                            }
                            break;
                    }
                }
                return statements;
            }
            /**
             * Appends the exports of an ImportEqualsDeclaration to a statement list, returning the
             * statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration whose exports are to be recorded.
             */
            function appendExportsOfImportEqualsDeclaration(statements, decl) {
                if (currentModuleInfo.exportEquals) {
                    return statements;
                }
                return appendExportsOfDeclaration(statements, decl);
            }
            /**
             * Appends the exports of a VariableStatement to a statement list, returning the statement
             * list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param node The VariableStatement whose exports are to be recorded.
             */
            function appendExportsOfVariableStatement(statements, node) {
                if (currentModuleInfo.exportEquals) {
                    return statements;
                }
                for (var _i = 0, _a = node.declarationList.declarations; _i < _a.length; _i++) {
                    var decl = _a[_i];
                    statements = appendExportsOfBindingElement(statements, decl);
                }
                return statements;
            }
            /**
             * Appends the exports of a VariableDeclaration or BindingElement to a statement list,
             * returning the statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration whose exports are to be recorded.
             */
            function appendExportsOfBindingElement(statements, decl) {
                if (currentModuleInfo.exportEquals) {
                    return statements;
                }
                if (ts.isBindingPattern(decl.name)) {
                    for (var _i = 0, _a = decl.name.elements; _i < _a.length; _i++) {
                        var element = _a[_i];
                        if (!ts.isOmittedExpression(element)) {
                            statements = appendExportsOfBindingElement(statements, element);
                        }
                    }
                }
                else if (!ts.isGeneratedIdentifier(decl.name)) {
                    statements = appendExportsOfDeclaration(statements, decl);
                }
                return statements;
            }
            /**
             * Appends the exports of a ClassDeclaration or FunctionDeclaration to a statement list,
             * returning the statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration whose exports are to be recorded.
             */
            function appendExportsOfHoistedDeclaration(statements, decl) {
                if (currentModuleInfo.exportEquals) {
                    return statements;
                }
                if (ts.hasSyntacticModifier(decl, 1 /* Export */)) {
                    var exportName = ts.hasSyntacticModifier(decl, 512 /* Default */) ? factory.createIdentifier("default") : factory.getDeclarationName(decl);
                    statements = appendExportStatement(statements, exportName, factory.getLocalName(decl), /*location*/ decl);
                }
                if (decl.name) {
                    statements = appendExportsOfDeclaration(statements, decl);
                }
                return statements;
            }
            /**
             * Appends the exports of a declaration to a statement list, returning the statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration to export.
             */
            function appendExportsOfDeclaration(statements, decl, liveBinding) {
                var name = factory.getDeclarationName(decl);
                var exportSpecifiers = currentModuleInfo.exportSpecifiers.get(ts.idText(name));
                if (exportSpecifiers) {
                    for (var _i = 0, exportSpecifiers_1 = exportSpecifiers; _i < exportSpecifiers_1.length; _i++) {
                        var exportSpecifier = exportSpecifiers_1[_i];
                        statements = appendExportStatement(statements, exportSpecifier.name, name, /*location*/ exportSpecifier.name, /* allowComments */ undefined, liveBinding);
                    }
                }
                return statements;
            }
            /**
             * Appends the down-level representation of an export to a statement list, returning the
             * statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param exportName The name of the export.
             * @param expression The expression to export.
             * @param location The location to use for source maps and comments for the export.
             * @param allowComments Whether to allow comments on the export.
             */
            function appendExportStatement(statements, exportName, expression, location, allowComments, liveBinding) {
                statements = ts.append(statements, createExportStatement(exportName, expression, location, allowComments, liveBinding));
                return statements;
            }
            function createUnderscoreUnderscoreESModule() {
                var statement;
                if (languageVersion === 0 /* ES3 */) {
                    statement = factory.createExpressionStatement(createExportExpression(factory.createIdentifier("__esModule"), factory.createTrue()));
                }
                else {
                    statement = factory.createExpressionStatement(factory.createCallExpression(factory.createPropertyAccessExpression(factory.createIdentifier("Object"), "defineProperty"), 
                    /*typeArguments*/ undefined, [
                        factory.createIdentifier("exports"),
                        factory.createStringLiteral("__esModule"),
                        factory.createObjectLiteralExpression([
                            factory.createPropertyAssignment("value", factory.createTrue())
                        ])
                    ]));
                }
                ts.setEmitFlags(statement, 1048576 /* CustomPrologue */);
                return statement;
            }
            /**
             * Creates a call to the current file's export function to export a value.
             *
             * @param name The bound name of the export.
             * @param value The exported value.
             * @param location The location to use for source maps and comments for the export.
             * @param allowComments An optional value indicating whether to emit comments for the statement.
             */
            function createExportStatement(name, value, location, allowComments, liveBinding) {
                var statement = ts.setTextRange(factory.createExpressionStatement(createExportExpression(name, value, /* location */ undefined, liveBinding)), location);
                ts.startOnNewLine(statement);
                if (!allowComments) {
                    ts.setEmitFlags(statement, 1536 /* NoComments */);
                }
                return statement;
            }
            /**
             * Creates a call to the current file's export function to export a value.
             *
             * @param name The bound name of the export.
             * @param value The exported value.
             * @param location The location to use for source maps and comments for the export.
             */
            function createExportExpression(name, value, location, liveBinding) {
                return ts.setTextRange(liveBinding && languageVersion !== 0 /* ES3 */ ? factory.createCallExpression(factory.createPropertyAccessExpression(factory.createIdentifier("Object"), "defineProperty"), 
                /*typeArguments*/ undefined, [
                    factory.createIdentifier("exports"),
                    factory.createStringLiteralFromNode(name),
                    factory.createObjectLiteralExpression([
                        factory.createPropertyAssignment("enumerable", factory.createTrue()),
                        factory.createPropertyAssignment("get", factory.createFunctionExpression(
                        /*modifiers*/ undefined, 
                        /*asteriskToken*/ undefined, 
                        /*name*/ undefined, 
                        /*typeParameters*/ undefined, 
                        /*parameters*/ [], 
                        /*type*/ undefined, factory.createBlock([factory.createReturnStatement(value)])))
                    ])
                ]) : factory.createAssignment(factory.createPropertyAccessExpression(factory.createIdentifier("exports"), factory.cloneNode(name)), value), location);
            }
            //
            // Modifier Visitors
            //
            /**
             * Visit nodes to elide module-specific modifiers.
             *
             * @param node The node to visit.
             */
            function modifierVisitor(node) {
                // Elide module-specific modifiers.
                switch (node.kind) {
                    case 93 /* ExportKeyword */:
                    case 88 /* DefaultKeyword */:
                        return undefined;
                }
                return node;
            }
            //
            // Emit Notification
            //
            /**
             * Hook for node emit notifications.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to emit.
             * @param emit A callback used to emit the node in the printer.
             */
            function onEmitNode(hint, node, emitCallback) {
                if (node.kind === 300 /* SourceFile */) {
                    currentSourceFile = node;
                    currentModuleInfo = moduleInfoMap[ts.getOriginalNodeId(currentSourceFile)];
                    previousOnEmitNode(hint, node, emitCallback);
                    currentSourceFile = undefined;
                    currentModuleInfo = undefined;
                }
                else {
                    previousOnEmitNode(hint, node, emitCallback);
                }
            }
            //
            // Substitutions
            //
            /**
             * Hooks node substitutions.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to substitute.
             */
            function onSubstituteNode(hint, node) {
                node = previousOnSubstituteNode(hint, node);
                if (node.id && noSubstitution[node.id]) {
                    return node;
                }
                if (hint === 1 /* Expression */) {
                    return substituteExpression(node);
                }
                else if (ts.isShorthandPropertyAssignment(node)) {
                    return substituteShorthandPropertyAssignment(node);
                }
                return node;
            }
            /**
             * Substitution for a ShorthandPropertyAssignment whose declaration name is an imported
             * or exported symbol.
             *
             * @param node The node to substitute.
             */
            function substituteShorthandPropertyAssignment(node) {
                var name = node.name;
                var exportedOrImportedName = substituteExpressionIdentifier(name);
                if (exportedOrImportedName !== name) {
                    // A shorthand property with an assignment initializer is probably part of a
                    // destructuring assignment
                    if (node.objectAssignmentInitializer) {
                        var initializer = factory.createAssignment(exportedOrImportedName, node.objectAssignmentInitializer);
                        return ts.setTextRange(factory.createPropertyAssignment(name, initializer), node);
                    }
                    return ts.setTextRange(factory.createPropertyAssignment(name, exportedOrImportedName), node);
                }
                return node;
            }
            /**
             * Substitution for an Expression that may contain an imported or exported symbol.
             *
             * @param node The node to substitute.
             */
            function substituteExpression(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        return substituteExpressionIdentifier(node);
                    case 206 /* CallExpression */:
                        return substituteCallExpression(node);
                    case 208 /* TaggedTemplateExpression */:
                        return substituteTaggedTemplateExpression(node);
                    case 219 /* BinaryExpression */:
                        return substituteBinaryExpression(node);
                }
                return node;
            }
            function substituteCallExpression(node) {
                if (ts.isIdentifier(node.expression)) {
                    var expression = substituteExpressionIdentifier(node.expression);
                    noSubstitution[ts.getNodeId(expression)] = true;
                    if (!ts.isIdentifier(expression)) {
                        return ts.addEmitFlags(factory.updateCallExpression(node, expression, 
                        /*typeArguments*/ undefined, node.arguments), 536870912 /* IndirectCall */);
                    }
                }
                return node;
            }
            function substituteTaggedTemplateExpression(node) {
                if (ts.isIdentifier(node.tag)) {
                    var tag = substituteExpressionIdentifier(node.tag);
                    noSubstitution[ts.getNodeId(tag)] = true;
                    if (!ts.isIdentifier(tag)) {
                        return ts.addEmitFlags(factory.updateTaggedTemplateExpression(node, tag, 
                        /*typeArguments*/ undefined, node.template), 536870912 /* IndirectCall */);
                    }
                }
                return node;
            }
            /**
             * Substitution for an Identifier expression that may contain an imported or exported
             * symbol.
             *
             * @param node The node to substitute.
             */
            function substituteExpressionIdentifier(node) {
                var _a, _b;
                if (ts.getEmitFlags(node) & 4096 /* HelperName */) {
                    var externalHelpersModuleName = ts.getExternalHelpersModuleName(currentSourceFile);
                    if (externalHelpersModuleName) {
                        return factory.createPropertyAccessExpression(externalHelpersModuleName, node);
                    }
                    return node;
                }
                else if (!(ts.isGeneratedIdentifier(node) && !(node.autoGenerateFlags & 64 /* AllowNameSubstitution */)) && !ts.isLocalName(node)) {
                    var exportContainer = resolver.getReferencedExportContainer(node, ts.isExportName(node));
                    if (exportContainer && exportContainer.kind === 300 /* SourceFile */) {
                        return ts.setTextRange(factory.createPropertyAccessExpression(factory.createIdentifier("exports"), factory.cloneNode(node)), 
                        /*location*/ node);
                    }
                    var importDeclaration = resolver.getReferencedImportDeclaration(node);
                    if (importDeclaration) {
                        if (ts.isImportClause(importDeclaration)) {
                            return ts.setTextRange(factory.createPropertyAccessExpression(factory.getGeneratedNameForNode(importDeclaration.parent), factory.createIdentifier("default")), 
                            /*location*/ node);
                        }
                        else if (ts.isImportSpecifier(importDeclaration)) {
                            var name = importDeclaration.propertyName || importDeclaration.name;
                            return ts.setTextRange(factory.createPropertyAccessExpression(factory.getGeneratedNameForNode(((_b = (_a = importDeclaration.parent) === null || _a === void 0 ? void 0 : _a.parent) === null || _b === void 0 ? void 0 : _b.parent) || importDeclaration), factory.cloneNode(name)), 
                            /*location*/ node);
                        }
                    }
                }
                return node;
            }
            /**
             * Substitution for a BinaryExpression that may contain an imported or exported symbol.
             *
             * @param node The node to substitute.
             */
            function substituteBinaryExpression(node) {
                // When we see an assignment expression whose left-hand side is an exported symbol,
                // we should ensure all exports of that symbol are updated with the correct value.
                //
                // - We do not substitute generated identifiers for any reason.
                // - We do not substitute identifiers tagged with the LocalName flag.
                // - We do not substitute identifiers that were originally the name of an enum or
                //   namespace due to how they are transformed in TypeScript.
                // - We only substitute identifiers that are exported at the top level.
                if (ts.isAssignmentOperator(node.operatorToken.kind)
                    && ts.isIdentifier(node.left)
                    && !ts.isGeneratedIdentifier(node.left)
                    && !ts.isLocalName(node.left)
                    && !ts.isDeclarationNameOfEnumOrNamespace(node.left)) {
                    var exportedNames = getExports(node.left);
                    if (exportedNames) {
                        // For each additional export of the declaration, apply an export assignment.
                        var expression = node;
                        for (var _i = 0, exportedNames_3 = exportedNames; _i < exportedNames_3.length; _i++) {
                            var exportName = exportedNames_3[_i];
                            // Mark the node to prevent triggering this rule again.
                            noSubstitution[ts.getNodeId(expression)] = true;
                            expression = createExportExpression(exportName, expression, /*location*/ node);
                        }
                        return expression;
                    }
                }
                return node;
            }
            /**
             * Gets the additional exports of a name.
             *
             * @param name The name.
             */
            function getExports(name) {
                if (!ts.isGeneratedIdentifier(name)) {
                    var valueDeclaration = resolver.getReferencedImportDeclaration(name)
                        || resolver.getReferencedValueDeclaration(name);
                    if (valueDeclaration) {
                        return currentModuleInfo
                            && currentModuleInfo.exportedBindings[ts.getOriginalNodeId(valueDeclaration)];
                    }
                }
            }
        }
        ts.transformModule = transformModule;
        // emit helper for dynamic import
        var dynamicImportUMDHelper = {
            name: "typescript:dynamicimport-sync-require",
            scoped: true,
            text: "\n            var __syncRequire = typeof module === \"object\" && typeof module.exports === \"object\";"
        };
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function transformSystemModule(context) {
            var factory = context.factory, startLexicalEnvironment = context.startLexicalEnvironment, endLexicalEnvironment = context.endLexicalEnvironment, hoistVariableDeclaration = context.hoistVariableDeclaration;
            var compilerOptions = context.getCompilerOptions();
            var resolver = context.getEmitResolver();
            var host = context.getEmitHost();
            var previousOnSubstituteNode = context.onSubstituteNode;
            var previousOnEmitNode = context.onEmitNode;
            context.onSubstituteNode = onSubstituteNode;
            context.onEmitNode = onEmitNode;
            context.enableSubstitution(79 /* Identifier */); // Substitutes expression identifiers for imported symbols.
            context.enableSubstitution(292 /* ShorthandPropertyAssignment */); // Substitutes expression identifiers for imported symbols
            context.enableSubstitution(219 /* BinaryExpression */); // Substitutes assignments to exported symbols.
            context.enableSubstitution(229 /* MetaProperty */); // Substitutes 'import.meta'
            context.enableEmitNotification(300 /* SourceFile */); // Restore state when substituting nodes in a file.
            var moduleInfoMap = []; // The ExternalModuleInfo for each file.
            var deferredExports = []; // Exports to defer until an EndOfDeclarationMarker is found.
            var exportFunctionsMap = []; // The export function associated with a source file.
            var noSubstitutionMap = []; // Set of nodes for which substitution rules should be ignored for each file.
            var contextObjectMap = []; // The context object associated with a source file.
            var currentSourceFile; // The current file.
            var moduleInfo; // ExternalModuleInfo for the current file.
            var exportFunction; // The export function for the current file.
            var contextObject; // The context object for the current file.
            var hoistedStatements;
            var enclosingBlockScopedContainer;
            var noSubstitution; // Set of nodes for which substitution rules should be ignored.
            return ts.chainBundle(context, transformSourceFile);
            /**
             * Transforms the module aspects of a SourceFile.
             *
             * @param node The SourceFile node.
             */
            function transformSourceFile(node) {
                if (node.isDeclarationFile || !(ts.isEffectiveExternalModule(node, compilerOptions) || node.transformFlags & 4194304 /* ContainsDynamicImport */)) {
                    return node;
                }
                var id = ts.getOriginalNodeId(node);
                currentSourceFile = node;
                enclosingBlockScopedContainer = node;
                // System modules have the following shape:
                //
                //     System.register(['dep-1', ... 'dep-n'], function(exports) {/* module body function */})
                //
                // The parameter 'exports' here is a callback '<T>(name: string, value: T) => T' that
                // is used to publish exported values. 'exports' returns its 'value' argument so in
                // most cases expressions that mutate exported values can be rewritten as:
                //
                //     expr -> exports('name', expr)
                //
                // The only exception in this rule is postfix unary operators,
                // see comment to 'substitutePostfixUnaryExpression' for more details
                // Collect information about the external module and dependency groups.
                moduleInfo = moduleInfoMap[id] = ts.collectExternalModuleInfo(context, node, resolver, compilerOptions);
                // Make sure that the name of the 'exports' function does not conflict with
                // existing identifiers.
                exportFunction = factory.createUniqueName("exports");
                exportFunctionsMap[id] = exportFunction;
                contextObject = contextObjectMap[id] = factory.createUniqueName("context");
                // Add the body of the module.
                var dependencyGroups = collectDependencyGroups(moduleInfo.externalImports);
                var moduleBodyBlock = createSystemModuleBody(node, dependencyGroups);
                var moduleBodyFunction = factory.createFunctionExpression(
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, 
                /*name*/ undefined, 
                /*typeParameters*/ undefined, [
                    factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, exportFunction),
                    factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, contextObject)
                ], 
                /*type*/ undefined, moduleBodyBlock);
                // Write the call to `System.register`
                // Clear the emit-helpers flag for later passes since we'll have already used it in the module body
                // So the helper will be emit at the correct position instead of at the top of the source-file
                var moduleName = ts.tryGetModuleNameFromFile(factory, node, host, compilerOptions);
                var dependencies = factory.createArrayLiteralExpression(ts.map(dependencyGroups, function (dependencyGroup) { return dependencyGroup.name; }));
                var updated = ts.setEmitFlags(factory.updateSourceFile(node, ts.setTextRange(factory.createNodeArray([
                    factory.createExpressionStatement(factory.createCallExpression(factory.createPropertyAccessExpression(factory.createIdentifier("System"), "register"), 
                    /*typeArguments*/ undefined, moduleName
                        ? [moduleName, dependencies, moduleBodyFunction]
                        : [dependencies, moduleBodyFunction]))
                ]), node.statements)), 1024 /* NoTrailingComments */);
                if (!ts.outFile(compilerOptions)) {
                    ts.moveEmitHelpers(updated, moduleBodyBlock, function (helper) { return !helper.scoped; });
                }
                if (noSubstitution) {
                    noSubstitutionMap[id] = noSubstitution;
                    noSubstitution = undefined;
                }
                currentSourceFile = undefined;
                moduleInfo = undefined;
                exportFunction = undefined;
                contextObject = undefined;
                hoistedStatements = undefined;
                enclosingBlockScopedContainer = undefined;
                return updated;
            }
            /**
             * Collects the dependency groups for this files imports.
             *
             * @param externalImports The imports for the file.
             */
            function collectDependencyGroups(externalImports) {
                var groupIndices = new ts.Map();
                var dependencyGroups = [];
                for (var _i = 0, externalImports_1 = externalImports; _i < externalImports_1.length; _i++) {
                    var externalImport = externalImports_1[_i];
                    var externalModuleName = ts.getExternalModuleNameLiteral(factory, externalImport, currentSourceFile, host, resolver, compilerOptions);
                    if (externalModuleName) {
                        var text = externalModuleName.text;
                        var groupIndex = groupIndices.get(text);
                        if (groupIndex !== undefined) {
                            // deduplicate/group entries in dependency list by the dependency name
                            dependencyGroups[groupIndex].externalImports.push(externalImport);
                        }
                        else {
                            groupIndices.set(text, dependencyGroups.length);
                            dependencyGroups.push({
                                name: externalModuleName,
                                externalImports: [externalImport]
                            });
                        }
                    }
                }
                return dependencyGroups;
            }
            /**
             * Adds the statements for the module body function for the source file.
             *
             * @param node The source file for the module.
             * @param dependencyGroups The grouped dependencies of the module.
             */
            function createSystemModuleBody(node, dependencyGroups) {
                // Shape of the body in system modules:
                //
                //  function (exports) {
                //      <list of local aliases for imports>
                //      <hoisted variable declarations>
                //      <hoisted function declarations>
                //      return {
                //          setters: [
                //              <list of setter function for imports>
                //          ],
                //          execute: function() {
                //              <module statements>
                //          }
                //      }
                //      <temp declarations>
                //  }
                //
                // i.e:
                //
                //   import {x} from 'file1'
                //   var y = 1;
                //   export function foo() { return y + x(); }
                //   console.log(y);
                //
                // Will be transformed to:
                //
                //  function(exports) {
                //      function foo() { return y + file_1.x(); }
                //      exports("foo", foo);
                //      var file_1, y;
                //      return {
                //          setters: [
                //              function(v) { file_1 = v }
                //          ],
                //          execute(): function() {
                //              y = 1;
                //              console.log(y);
                //          }
                //      };
                //  }
                var statements = [];
                // We start a new lexical environment in this function body, but *not* in the
                // body of the execute function. This allows us to emit temporary declarations
                // only in the outer module body and not in the inner one.
                startLexicalEnvironment();
                // Add any prologue directives.
                var ensureUseStrict = ts.getStrictOptionValue(compilerOptions, "alwaysStrict") || (!compilerOptions.noImplicitUseStrict && ts.isExternalModule(currentSourceFile));
                var statementOffset = factory.copyPrologue(node.statements, statements, ensureUseStrict, topLevelVisitor);
                // var __moduleName = context_1 && context_1.id;
                statements.push(factory.createVariableStatement(
                /*modifiers*/ undefined, factory.createVariableDeclarationList([
                    factory.createVariableDeclaration("__moduleName", 
                    /*exclamationToken*/ undefined, 
                    /*type*/ undefined, factory.createLogicalAnd(contextObject, factory.createPropertyAccessExpression(contextObject, "id")))
                ])));
                // Visit the synthetic external helpers import declaration if present
                ts.visitNode(moduleInfo.externalHelpersImportDeclaration, topLevelVisitor, ts.isStatement);
                // Visit the statements of the source file, emitting any transformations into
                // the `executeStatements` array. We do this *before* we fill the `setters` array
                // as we both emit transformations as well as aggregate some data used when creating
                // setters. This allows us to reduce the number of times we need to loop through the
                // statements of the source file.
                var executeStatements = ts.visitNodes(node.statements, topLevelVisitor, ts.isStatement, statementOffset);
                // Emit early exports for function declarations.
                ts.addRange(statements, hoistedStatements);
                // We emit hoisted variables early to align roughly with our previous emit output.
                // Two key differences in this approach are:
                // - Temporary variables will appear at the top rather than at the bottom of the file
                ts.insertStatementsAfterStandardPrologue(statements, endLexicalEnvironment());
                var exportStarFunction = addExportStarIfNeeded(statements); // TODO: GH#18217
                var modifiers = node.transformFlags & 1048576 /* ContainsAwait */ ?
                    factory.createModifiersFromModifierFlags(256 /* Async */) :
                    undefined;
                var moduleObject = factory.createObjectLiteralExpression([
                    factory.createPropertyAssignment("setters", createSettersArray(exportStarFunction, dependencyGroups)),
                    factory.createPropertyAssignment("execute", factory.createFunctionExpression(modifiers, 
                    /*asteriskToken*/ undefined, 
                    /*name*/ undefined, 
                    /*typeParameters*/ undefined, 
                    /*parameters*/ [], 
                    /*type*/ undefined, factory.createBlock(executeStatements, /*multiLine*/ true)))
                ], /*multiLine*/ true);
                statements.push(factory.createReturnStatement(moduleObject));
                return factory.createBlock(statements, /*multiLine*/ true);
            }
            /**
             * Adds an exportStar function to a statement list if it is needed for the file.
             *
             * @param statements A statement list.
             */
            function addExportStarIfNeeded(statements) {
                if (!moduleInfo.hasExportStarsToExportValues) {
                    return;
                }
                // when resolving exports local exported entries/indirect exported entries in the module
                // should always win over entries with similar names that were added via star exports
                // to support this we store names of local/indirect exported entries in a set.
                // this set is used to filter names brought by star expors.
                // local names set should only be added if we have anything exported
                if (!moduleInfo.exportedNames && moduleInfo.exportSpecifiers.size === 0) {
                    // no exported declarations (export var ...) or export specifiers (export {x})
                    // check if we have any non star export declarations.
                    var hasExportDeclarationWithExportClause = false;
                    for (var _i = 0, _a = moduleInfo.externalImports; _i < _a.length; _i++) {
                        var externalImport = _a[_i];
                        if (externalImport.kind === 270 /* ExportDeclaration */ && externalImport.exportClause) {
                            hasExportDeclarationWithExportClause = true;
                            break;
                        }
                    }
                    if (!hasExportDeclarationWithExportClause) {
                        // we still need to emit exportStar helper
                        var exportStarFunction_1 = createExportStarFunction(/*localNames*/ undefined);
                        statements.push(exportStarFunction_1);
                        return exportStarFunction_1.name;
                    }
                }
                var exportedNames = [];
                if (moduleInfo.exportedNames) {
                    for (var _b = 0, _c = moduleInfo.exportedNames; _b < _c.length; _b++) {
                        var exportedLocalName = _c[_b];
                        if (exportedLocalName.escapedText === "default") {
                            continue;
                        }
                        // write name of exported declaration, i.e 'export var x...'
                        exportedNames.push(factory.createPropertyAssignment(factory.createStringLiteralFromNode(exportedLocalName), factory.createTrue()));
                    }
                }
                var exportedNamesStorageRef = factory.createUniqueName("exportedNames");
                statements.push(factory.createVariableStatement(
                /*modifiers*/ undefined, factory.createVariableDeclarationList([
                    factory.createVariableDeclaration(exportedNamesStorageRef, 
                    /*exclamationToken*/ undefined, 
                    /*type*/ undefined, factory.createObjectLiteralExpression(exportedNames, /*multiline*/ true))
                ])));
                var exportStarFunction = createExportStarFunction(exportedNamesStorageRef);
                statements.push(exportStarFunction);
                return exportStarFunction.name;
            }
            /**
             * Creates an exportStar function for the file, with an optional set of excluded local
             * names.
             *
             * @param localNames An optional reference to an object containing a set of excluded local
             * names.
             */
            function createExportStarFunction(localNames) {
                var exportStarFunction = factory.createUniqueName("exportStar");
                var m = factory.createIdentifier("m");
                var n = factory.createIdentifier("n");
                var exports = factory.createIdentifier("exports");
                var condition = factory.createStrictInequality(n, factory.createStringLiteral("default"));
                if (localNames) {
                    condition = factory.createLogicalAnd(condition, factory.createLogicalNot(factory.createCallExpression(factory.createPropertyAccessExpression(localNames, "hasOwnProperty"), 
                    /*typeArguments*/ undefined, [n])));
                }
                return factory.createFunctionDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*asteriskToken*/ undefined, exportStarFunction, 
                /*typeParameters*/ undefined, [factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, m)], 
                /*type*/ undefined, factory.createBlock([
                    factory.createVariableStatement(
                    /*modifiers*/ undefined, factory.createVariableDeclarationList([
                        factory.createVariableDeclaration(exports, 
                        /*exclamationToken*/ undefined, 
                        /*type*/ undefined, factory.createObjectLiteralExpression([]))
                    ])),
                    factory.createForInStatement(factory.createVariableDeclarationList([
                        factory.createVariableDeclaration(n)
                    ]), m, factory.createBlock([
                        ts.setEmitFlags(factory.createIfStatement(condition, factory.createExpressionStatement(factory.createAssignment(factory.createElementAccessExpression(exports, n), factory.createElementAccessExpression(m, n)))), 1 /* SingleLine */)
                    ])),
                    factory.createExpressionStatement(factory.createCallExpression(exportFunction, 
                    /*typeArguments*/ undefined, [exports]))
                ], /*multiline*/ true));
            }
            /**
             * Creates an array setter callbacks for each dependency group.
             *
             * @param exportStarFunction A reference to an exportStarFunction for the file.
             * @param dependencyGroups An array of grouped dependencies.
             */
            function createSettersArray(exportStarFunction, dependencyGroups) {
                var setters = [];
                for (var _i = 0, dependencyGroups_1 = dependencyGroups; _i < dependencyGroups_1.length; _i++) {
                    var group_2 = dependencyGroups_1[_i];
                    // derive a unique name for parameter from the first named entry in the group
                    var localName = ts.forEach(group_2.externalImports, function (i) { return ts.getLocalNameForExternalImport(factory, i, currentSourceFile); });
                    var parameterName = localName ? factory.getGeneratedNameForNode(localName) : factory.createUniqueName("");
                    var statements = [];
                    for (var _a = 0, _b = group_2.externalImports; _a < _b.length; _a++) {
                        var entry = _b[_a];
                        var importVariableName = ts.getLocalNameForExternalImport(factory, entry, currentSourceFile); // TODO: GH#18217
                        switch (entry.kind) {
                            case 264 /* ImportDeclaration */:
                                if (!entry.importClause) {
                                    // 'import "..."' case
                                    // module is imported only for side-effects, no emit required
                                    break;
                                }
                            // falls through
                            case 263 /* ImportEqualsDeclaration */:
                                ts.Debug.assert(importVariableName !== undefined);
                                // save import into the local
                                statements.push(factory.createExpressionStatement(factory.createAssignment(importVariableName, parameterName)));
                                break;
                            case 270 /* ExportDeclaration */:
                                ts.Debug.assert(importVariableName !== undefined);
                                if (entry.exportClause) {
                                    if (ts.isNamedExports(entry.exportClause)) {
                                        //  export {a, b as c} from 'foo'
                                        //
                                        // emit as:
                                        //
                                        //  exports_({
                                        //     "a": _["a"],
                                        //     "c": _["b"]
                                        //  });
                                        var properties = [];
                                        for (var _c = 0, _d = entry.exportClause.elements; _c < _d.length; _c++) {
                                            var e = _d[_c];
                                            properties.push(factory.createPropertyAssignment(factory.createStringLiteral(ts.idText(e.name)), factory.createElementAccessExpression(parameterName, factory.createStringLiteral(ts.idText(e.propertyName || e.name)))));
                                        }
                                        statements.push(factory.createExpressionStatement(factory.createCallExpression(exportFunction, 
                                        /*typeArguments*/ undefined, [factory.createObjectLiteralExpression(properties, /*multiline*/ true)])));
                                    }
                                    else {
                                        statements.push(factory.createExpressionStatement(factory.createCallExpression(exportFunction, 
                                        /*typeArguments*/ undefined, [
                                            factory.createStringLiteral(ts.idText(entry.exportClause.name)),
                                            parameterName
                                        ])));
                                    }
                                }
                                else {
                                    //  export * from 'foo'
                                    //
                                    // emit as:
                                    //
                                    //  exportStar(foo_1_1);
                                    statements.push(factory.createExpressionStatement(factory.createCallExpression(exportStarFunction, 
                                    /*typeArguments*/ undefined, [parameterName])));
                                }
                                break;
                        }
                    }
                    setters.push(factory.createFunctionExpression(
                    /*modifiers*/ undefined, 
                    /*asteriskToken*/ undefined, 
                    /*name*/ undefined, 
                    /*typeParameters*/ undefined, [factory.createParameterDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*dotDotDotToken*/ undefined, parameterName)], 
                    /*type*/ undefined, factory.createBlock(statements, /*multiLine*/ true)));
                }
                return factory.createArrayLiteralExpression(setters, /*multiLine*/ true);
            }
            //
            // Top-level Source Element Visitors
            //
            /**
             * Visit source elements at the top-level of a module.
             *
             * @param node The node to visit.
             */
            function topLevelVisitor(node) {
                switch (node.kind) {
                    case 264 /* ImportDeclaration */:
                        return visitImportDeclaration(node);
                    case 263 /* ImportEqualsDeclaration */:
                        return visitImportEqualsDeclaration(node);
                    case 270 /* ExportDeclaration */:
                        return visitExportDeclaration(node);
                    case 269 /* ExportAssignment */:
                        return visitExportAssignment(node);
                    default:
                        return topLevelNestedVisitor(node);
                }
            }
            /**
             * Visits an ImportDeclaration node.
             *
             * @param node The node to visit.
             */
            function visitImportDeclaration(node) {
                var statements;
                if (node.importClause) {
                    hoistVariableDeclaration(ts.getLocalNameForExternalImport(factory, node, currentSourceFile)); // TODO: GH#18217
                }
                if (hasAssociatedEndOfDeclarationMarker(node)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfImportDeclaration(deferredExports[id], node);
                }
                else {
                    statements = appendExportsOfImportDeclaration(statements, node);
                }
                return ts.singleOrMany(statements);
            }
            function visitExportDeclaration(node) {
                ts.Debug.assertIsDefined(node);
                return undefined;
            }
            /**
             * Visits an ImportEqualsDeclaration node.
             *
             * @param node The node to visit.
             */
            function visitImportEqualsDeclaration(node) {
                ts.Debug.assert(ts.isExternalModuleImportEqualsDeclaration(node), "import= for internal module references should be handled in an earlier transformer.");
                var statements;
                hoistVariableDeclaration(ts.getLocalNameForExternalImport(factory, node, currentSourceFile)); // TODO: GH#18217
                if (hasAssociatedEndOfDeclarationMarker(node)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfImportEqualsDeclaration(deferredExports[id], node);
                }
                else {
                    statements = appendExportsOfImportEqualsDeclaration(statements, node);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Visits an ExportAssignment node.
             *
             * @param node The node to visit.
             */
            function visitExportAssignment(node) {
                if (node.isExportEquals) {
                    // Elide `export=` as it is illegal in a SystemJS module.
                    return undefined;
                }
                var expression = ts.visitNode(node.expression, visitor, ts.isExpression);
                var original = node.original;
                if (original && hasAssociatedEndOfDeclarationMarker(original)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportStatement(deferredExports[id], factory.createIdentifier("default"), expression, /*allowComments*/ true);
                }
                else {
                    return createExportStatement(factory.createIdentifier("default"), expression, /*allowComments*/ true);
                }
            }
            /**
             * Visits a FunctionDeclaration, hoisting it to the outer module body function.
             *
             * @param node The node to visit.
             */
            function visitFunctionDeclaration(node) {
                if (ts.hasSyntacticModifier(node, 1 /* Export */)) {
                    hoistedStatements = ts.append(hoistedStatements, factory.updateFunctionDeclaration(node, node.decorators, ts.visitNodes(node.modifiers, modifierVisitor, ts.isModifier), node.asteriskToken, factory.getDeclarationName(node, /*allowComments*/ true, /*allowSourceMaps*/ true), 
                    /*typeParameters*/ undefined, ts.visitNodes(node.parameters, visitor, ts.isParameterDeclaration), 
                    /*type*/ undefined, ts.visitNode(node.body, visitor, ts.isBlock)));
                }
                else {
                    hoistedStatements = ts.append(hoistedStatements, ts.visitEachChild(node, visitor, context));
                }
                if (hasAssociatedEndOfDeclarationMarker(node)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfHoistedDeclaration(deferredExports[id], node);
                }
                else {
                    hoistedStatements = appendExportsOfHoistedDeclaration(hoistedStatements, node);
                }
                return undefined;
            }
            /**
             * Visits a ClassDeclaration, hoisting its name to the outer module body function.
             *
             * @param node The node to visit.
             */
            function visitClassDeclaration(node) {
                var statements;
                // Hoist the name of the class declaration to the outer module body function.
                var name = factory.getLocalName(node);
                hoistVariableDeclaration(name);
                // Rewrite the class declaration into an assignment of a class expression.
                statements = ts.append(statements, ts.setTextRange(factory.createExpressionStatement(factory.createAssignment(name, ts.setTextRange(factory.createClassExpression(ts.visitNodes(node.decorators, visitor, ts.isDecorator), 
                /*modifiers*/ undefined, node.name, 
                /*typeParameters*/ undefined, ts.visitNodes(node.heritageClauses, visitor, ts.isHeritageClause), ts.visitNodes(node.members, visitor, ts.isClassElement)), node))), node));
                if (hasAssociatedEndOfDeclarationMarker(node)) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfHoistedDeclaration(deferredExports[id], node);
                }
                else {
                    statements = appendExportsOfHoistedDeclaration(statements, node);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Visits a variable statement, hoisting declared names to the top-level module body.
             * Each declaration is rewritten into an assignment expression.
             *
             * @param node The node to visit.
             */
            function visitVariableStatement(node) {
                if (!shouldHoistVariableDeclarationList(node.declarationList)) {
                    return ts.visitNode(node, visitor, ts.isStatement);
                }
                var expressions;
                var isExportedDeclaration = ts.hasSyntacticModifier(node, 1 /* Export */);
                var isMarkedDeclaration = hasAssociatedEndOfDeclarationMarker(node);
                for (var _i = 0, _a = node.declarationList.declarations; _i < _a.length; _i++) {
                    var variable = _a[_i];
                    if (variable.initializer) {
                        expressions = ts.append(expressions, transformInitializedVariable(variable, isExportedDeclaration && !isMarkedDeclaration));
                    }
                    else {
                        hoistBindingElement(variable);
                    }
                }
                var statements;
                if (expressions) {
                    statements = ts.append(statements, ts.setTextRange(factory.createExpressionStatement(factory.inlineExpressions(expressions)), node));
                }
                if (isMarkedDeclaration) {
                    // Defer exports until we encounter an EndOfDeclarationMarker node
                    var id = ts.getOriginalNodeId(node);
                    deferredExports[id] = appendExportsOfVariableStatement(deferredExports[id], node, isExportedDeclaration);
                }
                else {
                    statements = appendExportsOfVariableStatement(statements, node, /*exportSelf*/ false);
                }
                return ts.singleOrMany(statements);
            }
            /**
             * Hoists the declared names of a VariableDeclaration or BindingElement.
             *
             * @param node The declaration to hoist.
             */
            function hoistBindingElement(node) {
                if (ts.isBindingPattern(node.name)) {
                    for (var _i = 0, _a = node.name.elements; _i < _a.length; _i++) {
                        var element = _a[_i];
                        if (!ts.isOmittedExpression(element)) {
                            hoistBindingElement(element);
                        }
                    }
                }
                else {
                    hoistVariableDeclaration(factory.cloneNode(node.name));
                }
            }
            /**
             * Determines whether a VariableDeclarationList should be hoisted.
             *
             * @param node The node to test.
             */
            function shouldHoistVariableDeclarationList(node) {
                // hoist only non-block scoped declarations or block scoped declarations parented by source file
                return (ts.getEmitFlags(node) & 2097152 /* NoHoisting */) === 0
                    && (enclosingBlockScopedContainer.kind === 300 /* SourceFile */
                        || (ts.getOriginalNode(node).flags & 3 /* BlockScoped */) === 0);
            }
            /**
             * Transform an initialized variable declaration into an expression.
             *
             * @param node The node to transform.
             * @param isExportedDeclaration A value indicating whether the variable is exported.
             */
            function transformInitializedVariable(node, isExportedDeclaration) {
                var createAssignment = isExportedDeclaration ? createExportedVariableAssignment : createNonExportedVariableAssignment;
                return ts.isBindingPattern(node.name)
                    ? ts.flattenDestructuringAssignment(node, visitor, context, 0 /* All */, 
                    /*needsValue*/ false, createAssignment)
                    : node.initializer ? createAssignment(node.name, ts.visitNode(node.initializer, visitor, ts.isExpression)) : node.name;
            }
            /**
             * Creates an assignment expression for an exported variable declaration.
             *
             * @param name The name of the variable.
             * @param value The value of the variable's initializer.
             * @param location The source map location for the assignment.
             */
            function createExportedVariableAssignment(name, value, location) {
                return createVariableAssignment(name, value, location, /*isExportedDeclaration*/ true);
            }
            /**
             * Creates an assignment expression for a non-exported variable declaration.
             *
             * @param name The name of the variable.
             * @param value The value of the variable's initializer.
             * @param location The source map location for the assignment.
             */
            function createNonExportedVariableAssignment(name, value, location) {
                return createVariableAssignment(name, value, location, /*isExportedDeclaration*/ false);
            }
            /**
             * Creates an assignment expression for a variable declaration.
             *
             * @param name The name of the variable.
             * @param value The value of the variable's initializer.
             * @param location The source map location for the assignment.
             * @param isExportedDeclaration A value indicating whether the variable is exported.
             */
            function createVariableAssignment(name, value, location, isExportedDeclaration) {
                hoistVariableDeclaration(factory.cloneNode(name));
                return isExportedDeclaration
                    ? createExportExpression(name, preventSubstitution(ts.setTextRange(factory.createAssignment(name, value), location)))
                    : preventSubstitution(ts.setTextRange(factory.createAssignment(name, value), location));
            }
            /**
             * Visits a MergeDeclarationMarker used as a placeholder for the beginning of a merged
             * and transformed declaration.
             *
             * @param node The node to visit.
             */
            function visitMergeDeclarationMarker(node) {
                // For an EnumDeclaration or ModuleDeclaration that merges with a preceeding
                // declaration we do not emit a leading variable declaration. To preserve the
                // begin/end semantics of the declararation and to properly handle exports
                // we wrapped the leading variable declaration in a `MergeDeclarationMarker`.
                //
                // To balance the declaration, we defer the exports of the elided variable
                // statement until we visit this declaration's `EndOfDeclarationMarker`.
                if (hasAssociatedEndOfDeclarationMarker(node) && node.original.kind === 235 /* VariableStatement */) {
                    var id = ts.getOriginalNodeId(node);
                    var isExportedDeclaration = ts.hasSyntacticModifier(node.original, 1 /* Export */);
                    deferredExports[id] = appendExportsOfVariableStatement(deferredExports[id], node.original, isExportedDeclaration);
                }
                return node;
            }
            /**
             * Determines whether a node has an associated EndOfDeclarationMarker.
             *
             * @param node The node to test.
             */
            function hasAssociatedEndOfDeclarationMarker(node) {
                return (ts.getEmitFlags(node) & 4194304 /* HasEndOfDeclarationMarker */) !== 0;
            }
            /**
             * Visits a DeclarationMarker used as a placeholder for the end of a transformed
             * declaration.
             *
             * @param node The node to visit.
             */
            function visitEndOfDeclarationMarker(node) {
                // For some transformations we emit an `EndOfDeclarationMarker` to mark the actual
                // end of the transformed declaration. We use this marker to emit any deferred exports
                // of the declaration.
                var id = ts.getOriginalNodeId(node);
                var statements = deferredExports[id];
                if (statements) {
                    delete deferredExports[id];
                    return ts.append(statements, node);
                }
                else {
                    var original = ts.getOriginalNode(node);
                    if (ts.isModuleOrEnumDeclaration(original)) {
                        return ts.append(appendExportsOfDeclaration(statements, original), node);
                    }
                }
                return node;
            }
            /**
             * Appends the exports of an ImportDeclaration to a statement list, returning the
             * statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration whose exports are to be recorded.
             */
            function appendExportsOfImportDeclaration(statements, decl) {
                if (moduleInfo.exportEquals) {
                    return statements;
                }
                var importClause = decl.importClause;
                if (!importClause) {
                    return statements;
                }
                if (importClause.name) {
                    statements = appendExportsOfDeclaration(statements, importClause);
                }
                var namedBindings = importClause.namedBindings;
                if (namedBindings) {
                    switch (namedBindings.kind) {
                        case 266 /* NamespaceImport */:
                            statements = appendExportsOfDeclaration(statements, namedBindings);
                            break;
                        case 267 /* NamedImports */:
                            for (var _i = 0, _a = namedBindings.elements; _i < _a.length; _i++) {
                                var importBinding = _a[_i];
                                statements = appendExportsOfDeclaration(statements, importBinding);
                            }
                            break;
                    }
                }
                return statements;
            }
            /**
             * Appends the export of an ImportEqualsDeclaration to a statement list, returning the
             * statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration whose exports are to be recorded.
             */
            function appendExportsOfImportEqualsDeclaration(statements, decl) {
                if (moduleInfo.exportEquals) {
                    return statements;
                }
                return appendExportsOfDeclaration(statements, decl);
            }
            /**
             * Appends the exports of a VariableStatement to a statement list, returning the statement
             * list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param node The VariableStatement whose exports are to be recorded.
             * @param exportSelf A value indicating whether to also export each VariableDeclaration of
             * `nodes` declaration list.
             */
            function appendExportsOfVariableStatement(statements, node, exportSelf) {
                if (moduleInfo.exportEquals) {
                    return statements;
                }
                for (var _i = 0, _a = node.declarationList.declarations; _i < _a.length; _i++) {
                    var decl = _a[_i];
                    if (decl.initializer || exportSelf) {
                        statements = appendExportsOfBindingElement(statements, decl, exportSelf);
                    }
                }
                return statements;
            }
            /**
             * Appends the exports of a VariableDeclaration or BindingElement to a statement list,
             * returning the statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration whose exports are to be recorded.
             * @param exportSelf A value indicating whether to also export the declaration itself.
             */
            function appendExportsOfBindingElement(statements, decl, exportSelf) {
                if (moduleInfo.exportEquals) {
                    return statements;
                }
                if (ts.isBindingPattern(decl.name)) {
                    for (var _i = 0, _a = decl.name.elements; _i < _a.length; _i++) {
                        var element = _a[_i];
                        if (!ts.isOmittedExpression(element)) {
                            statements = appendExportsOfBindingElement(statements, element, exportSelf);
                        }
                    }
                }
                else if (!ts.isGeneratedIdentifier(decl.name)) {
                    var excludeName = void 0;
                    if (exportSelf) {
                        statements = appendExportStatement(statements, decl.name, factory.getLocalName(decl));
                        excludeName = ts.idText(decl.name);
                    }
                    statements = appendExportsOfDeclaration(statements, decl, excludeName);
                }
                return statements;
            }
            /**
             * Appends the exports of a ClassDeclaration or FunctionDeclaration to a statement list,
             * returning the statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration whose exports are to be recorded.
             */
            function appendExportsOfHoistedDeclaration(statements, decl) {
                if (moduleInfo.exportEquals) {
                    return statements;
                }
                var excludeName;
                if (ts.hasSyntacticModifier(decl, 1 /* Export */)) {
                    var exportName = ts.hasSyntacticModifier(decl, 512 /* Default */) ? factory.createStringLiteral("default") : decl.name;
                    statements = appendExportStatement(statements, exportName, factory.getLocalName(decl));
                    excludeName = ts.getTextOfIdentifierOrLiteral(exportName);
                }
                if (decl.name) {
                    statements = appendExportsOfDeclaration(statements, decl, excludeName);
                }
                return statements;
            }
            /**
             * Appends the exports of a declaration to a statement list, returning the statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param decl The declaration to export.
             * @param excludeName An optional name to exclude from exports.
             */
            function appendExportsOfDeclaration(statements, decl, excludeName) {
                if (moduleInfo.exportEquals) {
                    return statements;
                }
                var name = factory.getDeclarationName(decl);
                var exportSpecifiers = moduleInfo.exportSpecifiers.get(ts.idText(name));
                if (exportSpecifiers) {
                    for (var _i = 0, exportSpecifiers_2 = exportSpecifiers; _i < exportSpecifiers_2.length; _i++) {
                        var exportSpecifier = exportSpecifiers_2[_i];
                        if (exportSpecifier.name.escapedText !== excludeName) {
                            statements = appendExportStatement(statements, exportSpecifier.name, name);
                        }
                    }
                }
                return statements;
            }
            /**
             * Appends the down-level representation of an export to a statement list, returning the
             * statement list.
             *
             * @param statements A statement list to which the down-level export statements are to be
             * appended. If `statements` is `undefined`, a new array is allocated if statements are
             * appended.
             * @param exportName The name of the export.
             * @param expression The expression to export.
             * @param allowComments Whether to allow comments on the export.
             */
            function appendExportStatement(statements, exportName, expression, allowComments) {
                statements = ts.append(statements, createExportStatement(exportName, expression, allowComments));
                return statements;
            }
            /**
             * Creates a call to the current file's export function to export a value.
             *
             * @param name The bound name of the export.
             * @param value The exported value.
             * @param allowComments An optional value indicating whether to emit comments for the statement.
             */
            function createExportStatement(name, value, allowComments) {
                var statement = factory.createExpressionStatement(createExportExpression(name, value));
                ts.startOnNewLine(statement);
                if (!allowComments) {
                    ts.setEmitFlags(statement, 1536 /* NoComments */);
                }
                return statement;
            }
            /**
             * Creates a call to the current file's export function to export a value.
             *
             * @param name The bound name of the export.
             * @param value The exported value.
             */
            function createExportExpression(name, value) {
                var exportName = ts.isIdentifier(name) ? factory.createStringLiteralFromNode(name) : name;
                ts.setEmitFlags(value, ts.getEmitFlags(value) | 1536 /* NoComments */);
                return ts.setCommentRange(factory.createCallExpression(exportFunction, /*typeArguments*/ undefined, [exportName, value]), value);
            }
            //
            // Top-Level or Nested Source Element Visitors
            //
            /**
             * Visit nested elements at the top-level of a module.
             *
             * @param node The node to visit.
             */
            function topLevelNestedVisitor(node) {
                switch (node.kind) {
                    case 235 /* VariableStatement */:
                        return visitVariableStatement(node);
                    case 254 /* FunctionDeclaration */:
                        return visitFunctionDeclaration(node);
                    case 255 /* ClassDeclaration */:
                        return visitClassDeclaration(node);
                    case 240 /* ForStatement */:
                        return visitForStatement(node, /*isTopLevel*/ true);
                    case 241 /* ForInStatement */:
                        return visitForInStatement(node);
                    case 242 /* ForOfStatement */:
                        return visitForOfStatement(node);
                    case 238 /* DoStatement */:
                        return visitDoStatement(node);
                    case 239 /* WhileStatement */:
                        return visitWhileStatement(node);
                    case 248 /* LabeledStatement */:
                        return visitLabeledStatement(node);
                    case 246 /* WithStatement */:
                        return visitWithStatement(node);
                    case 247 /* SwitchStatement */:
                        return visitSwitchStatement(node);
                    case 261 /* CaseBlock */:
                        return visitCaseBlock(node);
                    case 287 /* CaseClause */:
                        return visitCaseClause(node);
                    case 288 /* DefaultClause */:
                        return visitDefaultClause(node);
                    case 250 /* TryStatement */:
                        return visitTryStatement(node);
                    case 290 /* CatchClause */:
                        return visitCatchClause(node);
                    case 233 /* Block */:
                        return visitBlock(node);
                    case 347 /* MergeDeclarationMarker */:
                        return visitMergeDeclarationMarker(node);
                    case 348 /* EndOfDeclarationMarker */:
                        return visitEndOfDeclarationMarker(node);
                    default:
                        return visitor(node);
                }
            }
            /**
             * Visits the body of a ForStatement to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitForStatement(node, isTopLevel) {
                var savedEnclosingBlockScopedContainer = enclosingBlockScopedContainer;
                enclosingBlockScopedContainer = node;
                node = factory.updateForStatement(node, ts.visitNode(node.initializer, isTopLevel ? visitForInitializer : discardedValueVisitor, ts.isForInitializer), ts.visitNode(node.condition, visitor, ts.isExpression), ts.visitNode(node.incrementor, discardedValueVisitor, ts.isExpression), ts.visitIterationBody(node.statement, isTopLevel ? topLevelNestedVisitor : visitor, context));
                enclosingBlockScopedContainer = savedEnclosingBlockScopedContainer;
                return node;
            }
            /**
             * Visits the body of a ForInStatement to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitForInStatement(node) {
                var savedEnclosingBlockScopedContainer = enclosingBlockScopedContainer;
                enclosingBlockScopedContainer = node;
                node = factory.updateForInStatement(node, visitForInitializer(node.initializer), ts.visitNode(node.expression, visitor, ts.isExpression), ts.visitIterationBody(node.statement, topLevelNestedVisitor, context));
                enclosingBlockScopedContainer = savedEnclosingBlockScopedContainer;
                return node;
            }
            /**
             * Visits the body of a ForOfStatement to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitForOfStatement(node) {
                var savedEnclosingBlockScopedContainer = enclosingBlockScopedContainer;
                enclosingBlockScopedContainer = node;
                node = factory.updateForOfStatement(node, node.awaitModifier, visitForInitializer(node.initializer), ts.visitNode(node.expression, visitor, ts.isExpression), ts.visitIterationBody(node.statement, topLevelNestedVisitor, context));
                enclosingBlockScopedContainer = savedEnclosingBlockScopedContainer;
                return node;
            }
            /**
             * Determines whether to hoist the initializer of a ForStatement, ForInStatement, or
             * ForOfStatement.
             *
             * @param node The node to test.
             */
            function shouldHoistForInitializer(node) {
                return ts.isVariableDeclarationList(node)
                    && shouldHoistVariableDeclarationList(node);
            }
            /**
             * Visits the initializer of a ForStatement, ForInStatement, or ForOfStatement
             *
             * @param node The node to visit.
             */
            function visitForInitializer(node) {
                if (shouldHoistForInitializer(node)) {
                    var expressions = void 0;
                    for (var _i = 0, _a = node.declarations; _i < _a.length; _i++) {
                        var variable = _a[_i];
                        expressions = ts.append(expressions, transformInitializedVariable(variable, /*isExportedDeclaration*/ false));
                        if (!variable.initializer) {
                            hoistBindingElement(variable);
                        }
                    }
                    return expressions ? factory.inlineExpressions(expressions) : factory.createOmittedExpression();
                }
                else {
                    return ts.visitNode(node, discardedValueVisitor, ts.isExpression);
                }
            }
            /**
             * Visits the body of a DoStatement to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitDoStatement(node) {
                return factory.updateDoStatement(node, ts.visitIterationBody(node.statement, topLevelNestedVisitor, context), ts.visitNode(node.expression, visitor, ts.isExpression));
            }
            /**
             * Visits the body of a WhileStatement to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitWhileStatement(node) {
                return factory.updateWhileStatement(node, ts.visitNode(node.expression, visitor, ts.isExpression), ts.visitIterationBody(node.statement, topLevelNestedVisitor, context));
            }
            /**
             * Visits the body of a LabeledStatement to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitLabeledStatement(node) {
                return factory.updateLabeledStatement(node, node.label, ts.visitNode(node.statement, topLevelNestedVisitor, ts.isStatement, factory.liftToBlock));
            }
            /**
             * Visits the body of a WithStatement to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitWithStatement(node) {
                return factory.updateWithStatement(node, ts.visitNode(node.expression, visitor, ts.isExpression), ts.visitNode(node.statement, topLevelNestedVisitor, ts.isStatement, factory.liftToBlock));
            }
            /**
             * Visits the body of a SwitchStatement to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitSwitchStatement(node) {
                return factory.updateSwitchStatement(node, ts.visitNode(node.expression, visitor, ts.isExpression), ts.visitNode(node.caseBlock, topLevelNestedVisitor, ts.isCaseBlock));
            }
            /**
             * Visits the body of a CaseBlock to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitCaseBlock(node) {
                var savedEnclosingBlockScopedContainer = enclosingBlockScopedContainer;
                enclosingBlockScopedContainer = node;
                node = factory.updateCaseBlock(node, ts.visitNodes(node.clauses, topLevelNestedVisitor, ts.isCaseOrDefaultClause));
                enclosingBlockScopedContainer = savedEnclosingBlockScopedContainer;
                return node;
            }
            /**
             * Visits the body of a CaseClause to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitCaseClause(node) {
                return factory.updateCaseClause(node, ts.visitNode(node.expression, visitor, ts.isExpression), ts.visitNodes(node.statements, topLevelNestedVisitor, ts.isStatement));
            }
            /**
             * Visits the body of a DefaultClause to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitDefaultClause(node) {
                return ts.visitEachChild(node, topLevelNestedVisitor, context);
            }
            /**
             * Visits the body of a TryStatement to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitTryStatement(node) {
                return ts.visitEachChild(node, topLevelNestedVisitor, context);
            }
            /**
             * Visits the body of a CatchClause to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitCatchClause(node) {
                var savedEnclosingBlockScopedContainer = enclosingBlockScopedContainer;
                enclosingBlockScopedContainer = node;
                node = factory.updateCatchClause(node, node.variableDeclaration, ts.visitNode(node.block, topLevelNestedVisitor, ts.isBlock));
                enclosingBlockScopedContainer = savedEnclosingBlockScopedContainer;
                return node;
            }
            /**
             * Visits the body of a Block to hoist declarations.
             *
             * @param node The node to visit.
             */
            function visitBlock(node) {
                var savedEnclosingBlockScopedContainer = enclosingBlockScopedContainer;
                enclosingBlockScopedContainer = node;
                node = ts.visitEachChild(node, topLevelNestedVisitor, context);
                enclosingBlockScopedContainer = savedEnclosingBlockScopedContainer;
                return node;
            }
            //
            // Destructuring Assignment Visitors
            //
            /**
             * Visit nodes to flatten destructuring assignments to exported symbols.
             *
             * @param node The node to visit.
             */
            function visitorWorker(node, valueIsDiscarded) {
                if (!(node.transformFlags & (2048 /* ContainsDestructuringAssignment */ | 4194304 /* ContainsDynamicImport */ | 67108864 /* ContainsUpdateExpressionForIdentifier */))) {
                    return node;
                }
                switch (node.kind) {
                    case 240 /* ForStatement */:
                        return visitForStatement(node, /*isTopLevel*/ false);
                    case 236 /* ExpressionStatement */:
                        return visitExpressionStatement(node);
                    case 210 /* ParenthesizedExpression */:
                        return visitParenthesizedExpression(node, valueIsDiscarded);
                    case 345 /* PartiallyEmittedExpression */:
                        return visitPartiallyEmittedExpression(node, valueIsDiscarded);
                    case 219 /* BinaryExpression */:
                        if (ts.isDestructuringAssignment(node)) {
                            return visitDestructuringAssignment(node, valueIsDiscarded);
                        }
                        break;
                    case 206 /* CallExpression */:
                        if (ts.isImportCall(node)) {
                            return visitImportCallExpression(node);
                        }
                        break;
                    case 217 /* PrefixUnaryExpression */:
                    case 218 /* PostfixUnaryExpression */:
                        return visitPrefixOrPostfixUnaryExpression(node, valueIsDiscarded);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Visit nodes to flatten destructuring assignments to exported symbols.
             *
             * @param node The node to visit.
             */
            function visitor(node) {
                return visitorWorker(node, /*valueIsDiscarded*/ false);
            }
            function discardedValueVisitor(node) {
                return visitorWorker(node, /*valueIsDiscarded*/ true);
            }
            function visitExpressionStatement(node) {
                return factory.updateExpressionStatement(node, ts.visitNode(node.expression, discardedValueVisitor, ts.isExpression));
            }
            function visitParenthesizedExpression(node, valueIsDiscarded) {
                return factory.updateParenthesizedExpression(node, ts.visitNode(node.expression, valueIsDiscarded ? discardedValueVisitor : visitor, ts.isExpression));
            }
            function visitPartiallyEmittedExpression(node, valueIsDiscarded) {
                return factory.updatePartiallyEmittedExpression(node, ts.visitNode(node.expression, valueIsDiscarded ? discardedValueVisitor : visitor, ts.isExpression));
            }
            function visitImportCallExpression(node) {
                // import("./blah")
                // emit as
                // System.register([], function (_export, _context) {
                //     return {
                //         setters: [],
                //         execute: () => {
                //             _context.import('./blah');
                //         }
                //     };
                // });
                var externalModuleName = ts.getExternalModuleNameLiteral(factory, node, currentSourceFile, host, resolver, compilerOptions);
                var firstArgument = ts.visitNode(ts.firstOrUndefined(node.arguments), visitor);
                // Only use the external module name if it differs from the first argument. This allows us to preserve the quote style of the argument on output.
                var argument = externalModuleName && (!firstArgument || !ts.isStringLiteral(firstArgument) || firstArgument.text !== externalModuleName.text) ? externalModuleName : firstArgument;
                return factory.createCallExpression(factory.createPropertyAccessExpression(contextObject, factory.createIdentifier("import")), 
                /*typeArguments*/ undefined, argument ? [argument] : []);
            }
            /**
             * Visits a DestructuringAssignment to flatten destructuring to exported symbols.
             *
             * @param node The node to visit.
             */
            function visitDestructuringAssignment(node, valueIsDiscarded) {
                if (hasExportedReferenceInDestructuringTarget(node.left)) {
                    return ts.flattenDestructuringAssignment(node, visitor, context, 0 /* All */, !valueIsDiscarded);
                }
                return ts.visitEachChild(node, visitor, context);
            }
            /**
             * Determines whether the target of a destructuring assignment refers to an exported symbol.
             *
             * @param node The destructuring target.
             */
            function hasExportedReferenceInDestructuringTarget(node) {
                if (ts.isAssignmentExpression(node, /*excludeCompoundAssignment*/ true)) {
                    return hasExportedReferenceInDestructuringTarget(node.left);
                }
                else if (ts.isSpreadElement(node)) {
                    return hasExportedReferenceInDestructuringTarget(node.expression);
                }
                else if (ts.isObjectLiteralExpression(node)) {
                    return ts.some(node.properties, hasExportedReferenceInDestructuringTarget);
                }
                else if (ts.isArrayLiteralExpression(node)) {
                    return ts.some(node.elements, hasExportedReferenceInDestructuringTarget);
                }
                else if (ts.isShorthandPropertyAssignment(node)) {
                    return hasExportedReferenceInDestructuringTarget(node.name);
                }
                else if (ts.isPropertyAssignment(node)) {
                    return hasExportedReferenceInDestructuringTarget(node.initializer);
                }
                else if (ts.isIdentifier(node)) {
                    var container = resolver.getReferencedExportContainer(node);
                    return container !== undefined && container.kind === 300 /* SourceFile */;
                }
                else {
                    return false;
                }
            }
            function visitPrefixOrPostfixUnaryExpression(node, valueIsDiscarded) {
                // When we see a prefix or postfix increment expression whose operand is an exported
                // symbol, we should ensure all exports of that symbol are updated with the correct
                // value.
                //
                // - We do not transform generated identifiers for any reason.
                // - We do not transform identifiers tagged with the LocalName flag.
                // - We do not transform identifiers that were originally the name of an enum or
                //   namespace due to how they are transformed in TypeScript.
                // - We only transform identifiers that are exported at the top level.
                if ((node.operator === 45 /* PlusPlusToken */ || node.operator === 46 /* MinusMinusToken */)
                    && ts.isIdentifier(node.operand)
                    && !ts.isGeneratedIdentifier(node.operand)
                    && !ts.isLocalName(node.operand)
                    && !ts.isDeclarationNameOfEnumOrNamespace(node.operand)) {
                    var exportedNames = getExports(node.operand);
                    if (exportedNames) {
                        var temp = void 0;
                        var expression = ts.visitNode(node.operand, visitor, ts.isExpression);
                        if (ts.isPrefixUnaryExpression(node)) {
                            expression = factory.updatePrefixUnaryExpression(node, expression);
                        }
                        else {
                            expression = factory.updatePostfixUnaryExpression(node, expression);
                            if (!valueIsDiscarded) {
                                temp = factory.createTempVariable(hoistVariableDeclaration);
                                expression = factory.createAssignment(temp, expression);
                                ts.setTextRange(expression, node);
                            }
                            expression = factory.createComma(expression, factory.cloneNode(node.operand));
                            ts.setTextRange(expression, node);
                        }
                        for (var _i = 0, exportedNames_4 = exportedNames; _i < exportedNames_4.length; _i++) {
                            var exportName = exportedNames_4[_i];
                            expression = createExportExpression(exportName, preventSubstitution(expression));
                        }
                        if (temp) {
                            expression = factory.createComma(expression, temp);
                            ts.setTextRange(expression, node);
                        }
                        return expression;
                    }
                }
                return ts.visitEachChild(node, visitor, context);
            }
            //
            // Modifier Visitors
            //
            /**
             * Visit nodes to elide module-specific modifiers.
             *
             * @param node The node to visit.
             */
            function modifierVisitor(node) {
                switch (node.kind) {
                    case 93 /* ExportKeyword */:
                    case 88 /* DefaultKeyword */:
                        return undefined;
                }
                return node;
            }
            //
            // Emit Notification
            //
            /**
             * Hook for node emit notifications.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to emit.
             * @param emitCallback A callback used to emit the node in the printer.
             */
            function onEmitNode(hint, node, emitCallback) {
                if (node.kind === 300 /* SourceFile */) {
                    var id = ts.getOriginalNodeId(node);
                    currentSourceFile = node;
                    moduleInfo = moduleInfoMap[id];
                    exportFunction = exportFunctionsMap[id];
                    noSubstitution = noSubstitutionMap[id];
                    contextObject = contextObjectMap[id];
                    if (noSubstitution) {
                        delete noSubstitutionMap[id];
                    }
                    previousOnEmitNode(hint, node, emitCallback);
                    currentSourceFile = undefined;
                    moduleInfo = undefined;
                    exportFunction = undefined;
                    contextObject = undefined;
                    noSubstitution = undefined;
                }
                else {
                    previousOnEmitNode(hint, node, emitCallback);
                }
            }
            //
            // Substitutions
            //
            /**
             * Hooks node substitutions.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to substitute.
             */
            function onSubstituteNode(hint, node) {
                node = previousOnSubstituteNode(hint, node);
                if (isSubstitutionPrevented(node)) {
                    return node;
                }
                if (hint === 1 /* Expression */) {
                    return substituteExpression(node);
                }
                else if (hint === 4 /* Unspecified */) {
                    return substituteUnspecified(node);
                }
                return node;
            }
            /**
             * Substitute the node, if necessary.
             *
             * @param node The node to substitute.
             */
            function substituteUnspecified(node) {
                switch (node.kind) {
                    case 292 /* ShorthandPropertyAssignment */:
                        return substituteShorthandPropertyAssignment(node);
                }
                return node;
            }
            /**
             * Substitution for a ShorthandPropertyAssignment whose name that may contain an imported or exported symbol.
             *
             * @param node The node to substitute.
             */
            function substituteShorthandPropertyAssignment(node) {
                var _a, _b;
                var name = node.name;
                if (!ts.isGeneratedIdentifier(name) && !ts.isLocalName(name)) {
                    var importDeclaration = resolver.getReferencedImportDeclaration(name);
                    if (importDeclaration) {
                        if (ts.isImportClause(importDeclaration)) {
                            return ts.setTextRange(factory.createPropertyAssignment(factory.cloneNode(name), factory.createPropertyAccessExpression(factory.getGeneratedNameForNode(importDeclaration.parent), factory.createIdentifier("default"))), 
                            /*location*/ node);
                        }
                        else if (ts.isImportSpecifier(importDeclaration)) {
                            return ts.setTextRange(factory.createPropertyAssignment(factory.cloneNode(name), factory.createPropertyAccessExpression(factory.getGeneratedNameForNode(((_b = (_a = importDeclaration.parent) === null || _a === void 0 ? void 0 : _a.parent) === null || _b === void 0 ? void 0 : _b.parent) || importDeclaration), factory.cloneNode(importDeclaration.propertyName || importDeclaration.name))), 
                            /*location*/ node);
                        }
                    }
                }
                return node;
            }
            /**
             * Substitute the expression, if necessary.
             *
             * @param node The node to substitute.
             */
            function substituteExpression(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        return substituteExpressionIdentifier(node);
                    case 219 /* BinaryExpression */:
                        return substituteBinaryExpression(node);
                    case 229 /* MetaProperty */:
                        return substituteMetaProperty(node);
                }
                return node;
            }
            /**
             * Substitution for an Identifier expression that may contain an imported or exported symbol.
             *
             * @param node The node to substitute.
             */
            function substituteExpressionIdentifier(node) {
                var _a, _b;
                if (ts.getEmitFlags(node) & 4096 /* HelperName */) {
                    var externalHelpersModuleName = ts.getExternalHelpersModuleName(currentSourceFile);
                    if (externalHelpersModuleName) {
                        return factory.createPropertyAccessExpression(externalHelpersModuleName, node);
                    }
                    return node;
                }
                // When we see an identifier in an expression position that
                // points to an imported symbol, we should substitute a qualified
                // reference to the imported symbol if one is needed.
                //
                // - We do not substitute generated identifiers for any reason.
                // - We do not substitute identifiers tagged with the LocalName flag.
                if (!ts.isGeneratedIdentifier(node) && !ts.isLocalName(node)) {
                    var importDeclaration = resolver.getReferencedImportDeclaration(node);
                    if (importDeclaration) {
                        if (ts.isImportClause(importDeclaration)) {
                            return ts.setTextRange(factory.createPropertyAccessExpression(factory.getGeneratedNameForNode(importDeclaration.parent), factory.createIdentifier("default")), 
                            /*location*/ node);
                        }
                        else if (ts.isImportSpecifier(importDeclaration)) {
                            return ts.setTextRange(factory.createPropertyAccessExpression(factory.getGeneratedNameForNode(((_b = (_a = importDeclaration.parent) === null || _a === void 0 ? void 0 : _a.parent) === null || _b === void 0 ? void 0 : _b.parent) || importDeclaration), factory.cloneNode(importDeclaration.propertyName || importDeclaration.name)), 
                            /*location*/ node);
                        }
                    }
                }
                return node;
            }
            /**
             * Substitution for a BinaryExpression that may contain an imported or exported symbol.
             *
             * @param node The node to substitute.
             */
            function substituteBinaryExpression(node) {
                // When we see an assignment expression whose left-hand side is an exported symbol,
                // we should ensure all exports of that symbol are updated with the correct value.
                //
                // - We do not substitute generated identifiers for any reason.
                // - We do not substitute identifiers tagged with the LocalName flag.
                // - We do not substitute identifiers that were originally the name of an enum or
                //   namespace due to how they are transformed in TypeScript.
                // - We only substitute identifiers that are exported at the top level.
                if (ts.isAssignmentOperator(node.operatorToken.kind)
                    && ts.isIdentifier(node.left)
                    && !ts.isGeneratedIdentifier(node.left)
                    && !ts.isLocalName(node.left)
                    && !ts.isDeclarationNameOfEnumOrNamespace(node.left)) {
                    var exportedNames = getExports(node.left);
                    if (exportedNames) {
                        // For each additional export of the declaration, apply an export assignment.
                        var expression = node;
                        for (var _i = 0, exportedNames_5 = exportedNames; _i < exportedNames_5.length; _i++) {
                            var exportName = exportedNames_5[_i];
                            expression = createExportExpression(exportName, preventSubstitution(expression));
                        }
                        return expression;
                    }
                }
                return node;
            }
            function substituteMetaProperty(node) {
                if (ts.isImportMeta(node)) {
                    return factory.createPropertyAccessExpression(contextObject, factory.createIdentifier("meta"));
                }
                return node;
            }
            /**
             * Gets the exports of a name.
             *
             * @param name The name.
             */
            function getExports(name) {
                var exportedNames;
                if (!ts.isGeneratedIdentifier(name)) {
                    var valueDeclaration = resolver.getReferencedImportDeclaration(name)
                        || resolver.getReferencedValueDeclaration(name);
                    if (valueDeclaration) {
                        var exportContainer = resolver.getReferencedExportContainer(name, /*prefixLocals*/ false);
                        if (exportContainer && exportContainer.kind === 300 /* SourceFile */) {
                            exportedNames = ts.append(exportedNames, factory.getDeclarationName(valueDeclaration));
                        }
                        exportedNames = ts.addRange(exportedNames, moduleInfo && moduleInfo.exportedBindings[ts.getOriginalNodeId(valueDeclaration)]);
                    }
                }
                return exportedNames;
            }
            /**
             * Prevent substitution of a node for this transformer.
             *
             * @param node The node which should not be substituted.
             */
            function preventSubstitution(node) {
                if (noSubstitution === undefined)
                    noSubstitution = [];
                noSubstitution[ts.getNodeId(node)] = true;
                return node;
            }
            /**
             * Determines whether a node should not be substituted.
             *
             * @param node The node to test.
             */
            function isSubstitutionPrevented(node) {
                return noSubstitution && node.id && noSubstitution[node.id];
            }
        }
        ts.transformSystemModule = transformSystemModule;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function transformECMAScriptModule(context) {
            var factory = context.factory, emitHelpers = context.getEmitHelperFactory;
            var compilerOptions = context.getCompilerOptions();
            var previousOnEmitNode = context.onEmitNode;
            var previousOnSubstituteNode = context.onSubstituteNode;
            context.onEmitNode = onEmitNode;
            context.onSubstituteNode = onSubstituteNode;
            context.enableEmitNotification(300 /* SourceFile */);
            context.enableSubstitution(79 /* Identifier */);
            var helperNameSubstitutions;
            return ts.chainBundle(context, transformSourceFile);
            function transformSourceFile(node) {
                if (node.isDeclarationFile) {
                    return node;
                }
                if (ts.isExternalModule(node) || compilerOptions.isolatedModules) {
                    var result = updateExternalModule(node);
                    if (!ts.isExternalModule(node) || ts.some(result.statements, ts.isExternalModuleIndicator)) {
                        return result;
                    }
                    return factory.updateSourceFile(result, ts.setTextRange(factory.createNodeArray(__spreadArray(__spreadArray([], result.statements, true), [ts.createEmptyExports(factory)], false)), result.statements));
                }
                return node;
            }
            function updateExternalModule(node) {
                var externalHelpersImportDeclaration = ts.createExternalHelpersImportDeclarationIfNeeded(factory, emitHelpers(), node, compilerOptions);
                if (externalHelpersImportDeclaration) {
                    var statements = [];
                    var statementOffset = factory.copyPrologue(node.statements, statements);
                    ts.append(statements, externalHelpersImportDeclaration);
                    ts.addRange(statements, ts.visitNodes(node.statements, visitor, ts.isStatement, statementOffset));
                    return factory.updateSourceFile(node, ts.setTextRange(factory.createNodeArray(statements), node.statements));
                }
                else {
                    return ts.visitEachChild(node, visitor, context);
                }
            }
            function visitor(node) {
                switch (node.kind) {
                    case 263 /* ImportEqualsDeclaration */:
                        // Elide `import=` as it is not legal with --module ES6
                        return undefined;
                    case 269 /* ExportAssignment */:
                        return visitExportAssignment(node);
                    case 270 /* ExportDeclaration */:
                        var exportDecl = node;
                        return visitExportDeclaration(exportDecl);
                }
                return node;
            }
            function visitExportAssignment(node) {
                // Elide `export=` as it is not legal with --module ES6
                return node.isExportEquals ? undefined : node;
            }
            function visitExportDeclaration(node) {
                // `export * as ns` only needs to be transformed in ES2015
                if (compilerOptions.module !== undefined && compilerOptions.module > ts.ModuleKind.ES2015) {
                    return node;
                }
                // Either ill-formed or don't need to be tranformed.
                if (!node.exportClause || !ts.isNamespaceExport(node.exportClause) || !node.moduleSpecifier) {
                    return node;
                }
                var oldIdentifier = node.exportClause.name;
                var synthName = factory.getGeneratedNameForNode(oldIdentifier);
                var importDecl = factory.createImportDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, factory.createImportClause(
                /*isTypeOnly*/ false, 
                /*name*/ undefined, factory.createNamespaceImport(synthName)), node.moduleSpecifier);
                ts.setOriginalNode(importDecl, node.exportClause);
                var exportDecl = ts.isExportNamespaceAsDefaultDeclaration(node) ? factory.createExportDefault(synthName) : factory.createExportDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*isTypeOnly*/ false, factory.createNamedExports([factory.createExportSpecifier(synthName, oldIdentifier)]));
                ts.setOriginalNode(exportDecl, node);
                return [importDecl, exportDecl];
            }
            //
            // Emit Notification
            //
            /**
             * Hook for node emit.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to emit.
             * @param emit A callback used to emit the node in the printer.
             */
            function onEmitNode(hint, node, emitCallback) {
                if (ts.isSourceFile(node)) {
                    if ((ts.isExternalModule(node) || compilerOptions.isolatedModules) && compilerOptions.importHelpers) {
                        helperNameSubstitutions = new ts.Map();
                    }
                    previousOnEmitNode(hint, node, emitCallback);
                    helperNameSubstitutions = undefined;
                }
                else {
                    previousOnEmitNode(hint, node, emitCallback);
                }
            }
            //
            // Substitutions
            //
            /**
             * Hooks node substitutions.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to substitute.
             */
            function onSubstituteNode(hint, node) {
                node = previousOnSubstituteNode(hint, node);
                if (helperNameSubstitutions && ts.isIdentifier(node) && ts.getEmitFlags(node) & 4096 /* HelperName */) {
                    return substituteHelperName(node);
                }
                return node;
            }
            function substituteHelperName(node) {
                var name = ts.idText(node);
                var substitution = helperNameSubstitutions.get(name);
                if (!substitution) {
                    helperNameSubstitutions.set(name, substitution = factory.createUniqueName(name, 16 /* Optimistic */ | 32 /* FileLevel */));
                }
                return substitution;
            }
        }
        ts.transformECMAScriptModule = transformECMAScriptModule;
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function canProduceDiagnostics(node) {
            return ts.isVariableDeclaration(node) ||
                ts.isPropertyDeclaration(node) ||
                ts.isPropertySignature(node) ||
                ts.isBindingElement(node) ||
                ts.isSetAccessor(node) ||
                ts.isGetAccessor(node) ||
                ts.isConstructSignatureDeclaration(node) ||
                ts.isCallSignatureDeclaration(node) ||
                ts.isMethodDeclaration(node) ||
                ts.isMethodSignature(node) ||
                ts.isFunctionDeclaration(node) ||
                ts.isParameter(node) ||
                ts.isTypeParameterDeclaration(node) ||
                ts.isExpressionWithTypeArguments(node) ||
                ts.isImportEqualsDeclaration(node) ||
                ts.isTypeAliasDeclaration(node) ||
                ts.isConstructorDeclaration(node) ||
                ts.isIndexSignatureDeclaration(node) ||
                ts.isPropertyAccessExpression(node) ||
                ts.isJSDocTypeAlias(node);
        }
        ts.canProduceDiagnostics = canProduceDiagnostics;
        function createGetSymbolAccessibilityDiagnosticForNodeName(node) {
            if (ts.isSetAccessor(node) || ts.isGetAccessor(node)) {
                return getAccessorNameVisibilityError;
            }
            else if (ts.isMethodSignature(node) || ts.isMethodDeclaration(node)) {
                return getMethodNameVisibilityError;
            }
            else {
                return createGetSymbolAccessibilityDiagnosticForNode(node);
            }
            function getAccessorNameVisibilityError(symbolAccessibilityResult) {
                var diagnosticMessage = getAccessorNameVisibilityDiagnosticMessage(symbolAccessibilityResult);
                return diagnosticMessage !== undefined ? {
                    diagnosticMessage: diagnosticMessage,
                    errorNode: node,
                    typeName: node.name
                } : undefined;
            }
            function getAccessorNameVisibilityDiagnosticMessage(symbolAccessibilityResult) {
                if (ts.isStatic(node)) {
                    return symbolAccessibilityResult.errorModuleName ?
                        symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                            ts.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                            ts.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2 :
                        ts.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_private_name_1;
                }
                else if (node.parent.kind === 255 /* ClassDeclaration */) {
                    return symbolAccessibilityResult.errorModuleName ?
                        symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                            ts.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                            ts.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2 :
                        ts.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_private_name_1;
                }
                else {
                    return symbolAccessibilityResult.errorModuleName ?
                        ts.Diagnostics.Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2 :
                        ts.Diagnostics.Property_0_of_exported_interface_has_or_is_using_private_name_1;
                }
            }
            function getMethodNameVisibilityError(symbolAccessibilityResult) {
                var diagnosticMessage = getMethodNameVisibilityDiagnosticMessage(symbolAccessibilityResult);
                return diagnosticMessage !== undefined ? {
                    diagnosticMessage: diagnosticMessage,
                    errorNode: node,
                    typeName: node.name
                } : undefined;
            }
            function getMethodNameVisibilityDiagnosticMessage(symbolAccessibilityResult) {
                if (ts.isStatic(node)) {
                    return symbolAccessibilityResult.errorModuleName ?
                        symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                            ts.Diagnostics.Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                            ts.Diagnostics.Public_static_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2 :
                        ts.Diagnostics.Public_static_method_0_of_exported_class_has_or_is_using_private_name_1;
                }
                else if (node.parent.kind === 255 /* ClassDeclaration */) {
                    return symbolAccessibilityResult.errorModuleName ?
                        symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                            ts.Diagnostics.Public_method_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                            ts.Diagnostics.Public_method_0_of_exported_class_has_or_is_using_name_1_from_private_module_2 :
                        ts.Diagnostics.Public_method_0_of_exported_class_has_or_is_using_private_name_1;
                }
                else {
                    return symbolAccessibilityResult.errorModuleName ?
                        ts.Diagnostics.Method_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2 :
                        ts.Diagnostics.Method_0_of_exported_interface_has_or_is_using_private_name_1;
                }
            }
        }
        ts.createGetSymbolAccessibilityDiagnosticForNodeName = createGetSymbolAccessibilityDiagnosticForNodeName;
        function createGetSymbolAccessibilityDiagnosticForNode(node) {
            if (ts.isVariableDeclaration(node) || ts.isPropertyDeclaration(node) || ts.isPropertySignature(node) || ts.isPropertyAccessExpression(node) || ts.isBindingElement(node) || ts.isConstructorDeclaration(node)) {
                return getVariableDeclarationTypeVisibilityError;
            }
            else if (ts.isSetAccessor(node) || ts.isGetAccessor(node)) {
                return getAccessorDeclarationTypeVisibilityError;
            }
            else if (ts.isConstructSignatureDeclaration(node) || ts.isCallSignatureDeclaration(node) || ts.isMethodDeclaration(node) || ts.isMethodSignature(node) || ts.isFunctionDeclaration(node) || ts.isIndexSignatureDeclaration(node)) {
                return getReturnTypeVisibilityError;
            }
            else if (ts.isParameter(node)) {
                if (ts.isParameterPropertyDeclaration(node, node.parent) && ts.hasSyntacticModifier(node.parent, 8 /* Private */)) {
                    return getVariableDeclarationTypeVisibilityError;
                }
                return getParameterDeclarationTypeVisibilityError;
            }
            else if (ts.isTypeParameterDeclaration(node)) {
                return getTypeParameterConstraintVisibilityError;
            }
            else if (ts.isExpressionWithTypeArguments(node)) {
                return getHeritageClauseVisibilityError;
            }
            else if (ts.isImportEqualsDeclaration(node)) {
                return getImportEntityNameVisibilityError;
            }
            else if (ts.isTypeAliasDeclaration(node) || ts.isJSDocTypeAlias(node)) {
                return getTypeAliasDeclarationVisibilityError;
            }
            else {
                return ts.Debug.assertNever(node, "Attempted to set a declaration diagnostic context for unhandled node kind: " + ts.SyntaxKind[node.kind]);
            }
            function getVariableDeclarationTypeVisibilityDiagnosticMessage(symbolAccessibilityResult) {
                if (node.kind === 252 /* VariableDeclaration */ || node.kind === 201 /* BindingElement */) {
                    return symbolAccessibilityResult.errorModuleName ?
                        symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                            ts.Diagnostics.Exported_variable_0_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                            ts.Diagnostics.Exported_variable_0_has_or_is_using_name_1_from_private_module_2 :
                        ts.Diagnostics.Exported_variable_0_has_or_is_using_private_name_1;
                }
                // This check is to ensure we don't report error on constructor parameter property as that error would be reported during parameter emit
                // The only exception here is if the constructor was marked as private. we are not emitting the constructor parameters at all.
                else if (node.kind === 165 /* PropertyDeclaration */ || node.kind === 204 /* PropertyAccessExpression */ || node.kind === 164 /* PropertySignature */ ||
                    (node.kind === 162 /* Parameter */ && ts.hasSyntacticModifier(node.parent, 8 /* Private */))) {
                    // TODO(jfreeman): Deal with computed properties in error reporting.
                    if (ts.isStatic(node)) {
                        return symbolAccessibilityResult.errorModuleName ?
                            symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                ts.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                                ts.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Public_static_property_0_of_exported_class_has_or_is_using_private_name_1;
                    }
                    else if (node.parent.kind === 255 /* ClassDeclaration */ || node.kind === 162 /* Parameter */) {
                        return symbolAccessibilityResult.errorModuleName ?
                            symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                ts.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                                ts.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Public_property_0_of_exported_class_has_or_is_using_private_name_1;
                    }
                    else {
                        // Interfaces cannot have types that cannot be named
                        return symbolAccessibilityResult.errorModuleName ?
                            ts.Diagnostics.Property_0_of_exported_interface_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Property_0_of_exported_interface_has_or_is_using_private_name_1;
                    }
                }
            }
            function getVariableDeclarationTypeVisibilityError(symbolAccessibilityResult) {
                var diagnosticMessage = getVariableDeclarationTypeVisibilityDiagnosticMessage(symbolAccessibilityResult);
                return diagnosticMessage !== undefined ? {
                    diagnosticMessage: diagnosticMessage,
                    errorNode: node,
                    typeName: node.name
                } : undefined;
            }
            function getAccessorDeclarationTypeVisibilityError(symbolAccessibilityResult) {
                var diagnosticMessage;
                if (node.kind === 171 /* SetAccessor */) {
                    // Getters can infer the return type from the returned expression, but setters cannot, so the
                    // "_from_external_module_1_but_cannot_be_named" case cannot occur.
                    if (ts.isStatic(node)) {
                        diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                            ts.Diagnostics.Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Parameter_type_of_public_static_setter_0_from_exported_class_has_or_is_using_private_name_1;
                    }
                    else {
                        diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                            ts.Diagnostics.Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Parameter_type_of_public_setter_0_from_exported_class_has_or_is_using_private_name_1;
                    }
                }
                else {
                    if (ts.isStatic(node)) {
                        diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                            symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                ts.Diagnostics.Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                                ts.Diagnostics.Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Return_type_of_public_static_getter_0_from_exported_class_has_or_is_using_private_name_1;
                    }
                    else {
                        diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                            symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                ts.Diagnostics.Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                                ts.Diagnostics.Return_type_of_public_getter_0_from_exported_class_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Return_type_of_public_getter_0_from_exported_class_has_or_is_using_private_name_1;
                    }
                }
                return {
                    diagnosticMessage: diagnosticMessage,
                    errorNode: node.name,
                    typeName: node.name
                };
            }
            function getReturnTypeVisibilityError(symbolAccessibilityResult) {
                var diagnosticMessage;
                switch (node.kind) {
                    case 173 /* ConstructSignature */:
                        // Interfaces cannot have return types that cannot be named
                        diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                            ts.Diagnostics.Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1 :
                            ts.Diagnostics.Return_type_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_0;
                        break;
                    case 172 /* CallSignature */:
                        // Interfaces cannot have return types that cannot be named
                        diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                            ts.Diagnostics.Return_type_of_call_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1 :
                            ts.Diagnostics.Return_type_of_call_signature_from_exported_interface_has_or_is_using_private_name_0;
                        break;
                    case 174 /* IndexSignature */:
                        // Interfaces cannot have return types that cannot be named
                        diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                            ts.Diagnostics.Return_type_of_index_signature_from_exported_interface_has_or_is_using_name_0_from_private_module_1 :
                            ts.Diagnostics.Return_type_of_index_signature_from_exported_interface_has_or_is_using_private_name_0;
                        break;
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                        if (ts.isStatic(node)) {
                            diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                                symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                    ts.Diagnostics.Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named :
                                    ts.Diagnostics.Return_type_of_public_static_method_from_exported_class_has_or_is_using_name_0_from_private_module_1 :
                                ts.Diagnostics.Return_type_of_public_static_method_from_exported_class_has_or_is_using_private_name_0;
                        }
                        else if (node.parent.kind === 255 /* ClassDeclaration */) {
                            diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                                symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                    ts.Diagnostics.Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named :
                                    ts.Diagnostics.Return_type_of_public_method_from_exported_class_has_or_is_using_name_0_from_private_module_1 :
                                ts.Diagnostics.Return_type_of_public_method_from_exported_class_has_or_is_using_private_name_0;
                        }
                        else {
                            // Interfaces cannot have return types that cannot be named
                            diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                                ts.Diagnostics.Return_type_of_method_from_exported_interface_has_or_is_using_name_0_from_private_module_1 :
                                ts.Diagnostics.Return_type_of_method_from_exported_interface_has_or_is_using_private_name_0;
                        }
                        break;
                    case 254 /* FunctionDeclaration */:
                        diagnosticMessage = symbolAccessibilityResult.errorModuleName ?
                            symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                ts.Diagnostics.Return_type_of_exported_function_has_or_is_using_name_0_from_external_module_1_but_cannot_be_named :
                                ts.Diagnostics.Return_type_of_exported_function_has_or_is_using_name_0_from_private_module_1 :
                            ts.Diagnostics.Return_type_of_exported_function_has_or_is_using_private_name_0;
                        break;
                    default:
                        return ts.Debug.fail("This is unknown kind for signature: " + node.kind);
                }
                return {
                    diagnosticMessage: diagnosticMessage,
                    errorNode: node.name || node
                };
            }
            function getParameterDeclarationTypeVisibilityError(symbolAccessibilityResult) {
                var diagnosticMessage = getParameterDeclarationTypeVisibilityDiagnosticMessage(symbolAccessibilityResult);
                return diagnosticMessage !== undefined ? {
                    diagnosticMessage: diagnosticMessage,
                    errorNode: node,
                    typeName: node.name
                } : undefined;
            }
            function getParameterDeclarationTypeVisibilityDiagnosticMessage(symbolAccessibilityResult) {
                switch (node.parent.kind) {
                    case 169 /* Constructor */:
                        return symbolAccessibilityResult.errorModuleName ?
                            symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                ts.Diagnostics.Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                                ts.Diagnostics.Parameter_0_of_constructor_from_exported_class_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Parameter_0_of_constructor_from_exported_class_has_or_is_using_private_name_1;
                    case 173 /* ConstructSignature */:
                    case 178 /* ConstructorType */:
                        // Interfaces cannot have parameter types that cannot be named
                        return symbolAccessibilityResult.errorModuleName ?
                            ts.Diagnostics.Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1;
                    case 172 /* CallSignature */:
                        // Interfaces cannot have parameter types that cannot be named
                        return symbolAccessibilityResult.errorModuleName ?
                            ts.Diagnostics.Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1;
                    case 174 /* IndexSignature */:
                        // Interfaces cannot have parameter types that cannot be named
                        return symbolAccessibilityResult.errorModuleName ?
                            ts.Diagnostics.Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Parameter_0_of_index_signature_from_exported_interface_has_or_is_using_private_name_1;
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                        if (ts.isStatic(node.parent)) {
                            return symbolAccessibilityResult.errorModuleName ?
                                symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                    ts.Diagnostics.Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                                    ts.Diagnostics.Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_name_1_from_private_module_2 :
                                ts.Diagnostics.Parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1;
                        }
                        else if (node.parent.parent.kind === 255 /* ClassDeclaration */) {
                            return symbolAccessibilityResult.errorModuleName ?
                                symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                    ts.Diagnostics.Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                                    ts.Diagnostics.Parameter_0_of_public_method_from_exported_class_has_or_is_using_name_1_from_private_module_2 :
                                ts.Diagnostics.Parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1;
                        }
                        else {
                            // Interfaces cannot have parameter types that cannot be named
                            return symbolAccessibilityResult.errorModuleName ?
                                ts.Diagnostics.Parameter_0_of_method_from_exported_interface_has_or_is_using_name_1_from_private_module_2 :
                                ts.Diagnostics.Parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1;
                        }
                    case 254 /* FunctionDeclaration */:
                    case 177 /* FunctionType */:
                        return symbolAccessibilityResult.errorModuleName ?
                            symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                ts.Diagnostics.Parameter_0_of_exported_function_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                                ts.Diagnostics.Parameter_0_of_exported_function_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Parameter_0_of_exported_function_has_or_is_using_private_name_1;
                    case 171 /* SetAccessor */:
                    case 170 /* GetAccessor */:
                        return symbolAccessibilityResult.errorModuleName ?
                            symbolAccessibilityResult.accessibility === 2 /* CannotBeNamed */ ?
                                ts.Diagnostics.Parameter_0_of_accessor_has_or_is_using_name_1_from_external_module_2_but_cannot_be_named :
                                ts.Diagnostics.Parameter_0_of_accessor_has_or_is_using_name_1_from_private_module_2 :
                            ts.Diagnostics.Parameter_0_of_accessor_has_or_is_using_private_name_1;
                    default:
                        return ts.Debug.fail("Unknown parent for parameter: " + ts.SyntaxKind[node.parent.kind]);
                }
            }
            function getTypeParameterConstraintVisibilityError() {
                // Type parameter constraints are named by user so we should always be able to name it
                var diagnosticMessage;
                switch (node.parent.kind) {
                    case 255 /* ClassDeclaration */:
                        diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_exported_class_has_or_is_using_private_name_1;
                        break;
                    case 256 /* InterfaceDeclaration */:
                        diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_exported_interface_has_or_is_using_private_name_1;
                        break;
                    case 193 /* MappedType */:
                        diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_exported_mapped_object_type_is_using_private_name_1;
                        break;
                    case 178 /* ConstructorType */:
                    case 173 /* ConstructSignature */:
                        diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_constructor_signature_from_exported_interface_has_or_is_using_private_name_1;
                        break;
                    case 172 /* CallSignature */:
                        diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_call_signature_from_exported_interface_has_or_is_using_private_name_1;
                        break;
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                        if (ts.isStatic(node.parent)) {
                            diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_public_static_method_from_exported_class_has_or_is_using_private_name_1;
                        }
                        else if (node.parent.parent.kind === 255 /* ClassDeclaration */) {
                            diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_public_method_from_exported_class_has_or_is_using_private_name_1;
                        }
                        else {
                            diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_method_from_exported_interface_has_or_is_using_private_name_1;
                        }
                        break;
                    case 177 /* FunctionType */:
                    case 254 /* FunctionDeclaration */:
                        diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_exported_function_has_or_is_using_private_name_1;
                        break;
                    case 257 /* TypeAliasDeclaration */:
                        diagnosticMessage = ts.Diagnostics.Type_parameter_0_of_exported_type_alias_has_or_is_using_private_name_1;
                        break;
                    default:
                        return ts.Debug.fail("This is unknown parent for type parameter: " + node.parent.kind);
                }
                return {
                    diagnosticMessage: diagnosticMessage,
                    errorNode: node,
                    typeName: node.name
                };
            }
            function getHeritageClauseVisibilityError() {
                var diagnosticMessage;
                // Heritage clause is written by user so it can always be named
                if (ts.isClassDeclaration(node.parent.parent)) {
                    // Class or Interface implemented/extended is inaccessible
                    diagnosticMessage = ts.isHeritageClause(node.parent) && node.parent.token === 117 /* ImplementsKeyword */ ?
                        ts.Diagnostics.Implements_clause_of_exported_class_0_has_or_is_using_private_name_1 :
                        node.parent.parent.name ? ts.Diagnostics.extends_clause_of_exported_class_0_has_or_is_using_private_name_1 :
                            ts.Diagnostics.extends_clause_of_exported_class_has_or_is_using_private_name_0;
                }
                else {
                    // interface is inaccessible
                    diagnosticMessage = ts.Diagnostics.extends_clause_of_exported_interface_0_has_or_is_using_private_name_1;
                }
                return {
                    diagnosticMessage: diagnosticMessage,
                    errorNode: node,
                    typeName: ts.getNameOfDeclaration(node.parent.parent)
                };
            }
            function getImportEntityNameVisibilityError() {
                return {
                    diagnosticMessage: ts.Diagnostics.Import_declaration_0_is_using_private_name_1,
                    errorNode: node,
                    typeName: node.name
                };
            }
            function getTypeAliasDeclarationVisibilityError(symbolAccessibilityResult) {
                return {
                    diagnosticMessage: symbolAccessibilityResult.errorModuleName
                        ? ts.Diagnostics.Exported_type_alias_0_has_or_is_using_private_name_1_from_module_2
                        : ts.Diagnostics.Exported_type_alias_0_has_or_is_using_private_name_1,
                    errorNode: ts.isJSDocTypeAlias(node) ? ts.Debug.checkDefined(node.typeExpression) : node.type,
                    typeName: ts.isJSDocTypeAlias(node) ? ts.getNameOfDeclaration(node) : node.name,
                };
            }
        }
        ts.createGetSymbolAccessibilityDiagnosticForNode = createGetSymbolAccessibilityDiagnosticForNode;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function getDeclarationDiagnostics(host, resolver, file) {
            var compilerOptions = host.getCompilerOptions();
            var result = ts.transformNodes(resolver, host, ts.factory, compilerOptions, file ? [file] : ts.filter(host.getSourceFiles(), ts.isSourceFileNotJson), [transformDeclarations], /*allowDtsFiles*/ false);
            return result.diagnostics;
        }
        ts.getDeclarationDiagnostics = getDeclarationDiagnostics;
        function hasInternalAnnotation(range, currentSourceFile) {
            var comment = currentSourceFile.text.substring(range.pos, range.end);
            return ts.stringContains(comment, "@internal");
        }
        function isInternalDeclaration(node, currentSourceFile) {
            var parseTreeNode = ts.getParseTreeNode(node);
            if (parseTreeNode && parseTreeNode.kind === 162 /* Parameter */) {
                var paramIdx = parseTreeNode.parent.parameters.indexOf(parseTreeNode);
                var previousSibling = paramIdx > 0 ? parseTreeNode.parent.parameters[paramIdx - 1] : undefined;
                var text = currentSourceFile.text;
                var commentRanges = previousSibling
                    ? ts.concatenate(
                    // to handle
                    // ... parameters, /* @internal */
                    // public param: string
                    ts.getTrailingCommentRanges(text, ts.skipTrivia(text, previousSibling.end + 1, /* stopAfterLineBreak */ false, /* stopAtComments */ true)), ts.getLeadingCommentRanges(text, node.pos))
                    : ts.getTrailingCommentRanges(text, ts.skipTrivia(text, node.pos, /* stopAfterLineBreak */ false, /* stopAtComments */ true));
                return commentRanges && commentRanges.length && hasInternalAnnotation(ts.last(commentRanges), currentSourceFile);
            }
            var leadingCommentRanges = parseTreeNode && ts.getLeadingCommentRangesOfNode(parseTreeNode, currentSourceFile);
            return !!ts.forEach(leadingCommentRanges, function (range) {
                return hasInternalAnnotation(range, currentSourceFile);
            });
        }
        ts.isInternalDeclaration = isInternalDeclaration;
        var declarationEmitNodeBuilderFlags = 1024 /* MultilineObjectLiterals */ |
            2048 /* WriteClassExpressionAsTypeLiteral */ |
            4096 /* UseTypeOfFunction */ |
            8 /* UseStructuralFallback */ |
            524288 /* AllowEmptyTuple */ |
            4 /* GenerateNamesForShadowedTypeParams */ |
            1 /* NoTruncation */;
        /**
         * Transforms a ts file into a .d.ts file
         * This process requires type information, which is retrieved through the emit resolver. Because of this,
         * in many places this transformer assumes it will be operating on parse tree nodes directly.
         * This means that _no transforms should be allowed to occur before this one_.
         */
        function transformDeclarations(context) {
            var throwDiagnostic = function () { return ts.Debug.fail("Diagnostic emitted without context"); };
            var getSymbolAccessibilityDiagnostic = throwDiagnostic;
            var needsDeclare = true;
            var isBundledEmit = false;
            var resultHasExternalModuleIndicator = false;
            var needsScopeFixMarker = false;
            var resultHasScopeMarker = false;
            var enclosingDeclaration;
            var necessaryTypeReferences;
            var lateMarkedStatements;
            var lateStatementReplacementMap;
            var suppressNewDiagnosticContexts;
            var exportedModulesFromDeclarationEmit;
            var factory = context.factory;
            var host = context.getEmitHost();
            var symbolTracker = {
                trackSymbol: trackSymbol,
                reportInaccessibleThisError: reportInaccessibleThisError,
                reportInaccessibleUniqueSymbolError: reportInaccessibleUniqueSymbolError,
                reportCyclicStructureError: reportCyclicStructureError,
                reportPrivateInBaseOfClassExpression: reportPrivateInBaseOfClassExpression,
                reportLikelyUnsafeImportRequiredError: reportLikelyUnsafeImportRequiredError,
                reportTruncationError: reportTruncationError,
                moduleResolverHost: host,
                trackReferencedAmbientModule: trackReferencedAmbientModule,
                trackExternalModuleSymbolOfImportTypeNode: trackExternalModuleSymbolOfImportTypeNode,
                reportNonlocalAugmentation: reportNonlocalAugmentation,
                reportNonSerializableProperty: reportNonSerializableProperty
            };
            var errorNameNode;
            var errorFallbackNode;
            var currentSourceFile;
            var refs;
            var libs;
            var emittedImports; // must be declared in container so it can be `undefined` while transformer's first pass
            var resolver = context.getEmitResolver();
            var options = context.getCompilerOptions();
            var noResolve = options.noResolve, stripInternal = options.stripInternal;
            return transformRoot;
            function recordTypeReferenceDirectivesIfNecessary(typeReferenceDirectives) {
                if (!typeReferenceDirectives) {
                    return;
                }
                necessaryTypeReferences = necessaryTypeReferences || new ts.Set();
                for (var _i = 0, typeReferenceDirectives_2 = typeReferenceDirectives; _i < typeReferenceDirectives_2.length; _i++) {
                    var ref = typeReferenceDirectives_2[_i];
                    necessaryTypeReferences.add(ref);
                }
            }
            function trackReferencedAmbientModule(node, symbol) {
                // If it is visible via `// <reference types="..."/>`, then we should just use that
                var directives = resolver.getTypeReferenceDirectivesForSymbol(symbol, 67108863 /* All */);
                if (ts.length(directives)) {
                    return recordTypeReferenceDirectivesIfNecessary(directives);
                }
                // Otherwise we should emit a path-based reference
                var container = ts.getSourceFileOfNode(node);
                refs.set(ts.getOriginalNodeId(container), container);
            }
            function handleSymbolAccessibilityError(symbolAccessibilityResult) {
                if (symbolAccessibilityResult.accessibility === 0 /* Accessible */) {
                    // Add aliases back onto the possible imports list if they're not there so we can try them again with updated visibility info
                    if (symbolAccessibilityResult && symbolAccessibilityResult.aliasesToMakeVisible) {
                        if (!lateMarkedStatements) {
                            lateMarkedStatements = symbolAccessibilityResult.aliasesToMakeVisible;
                        }
                        else {
                            for (var _i = 0, _a = symbolAccessibilityResult.aliasesToMakeVisible; _i < _a.length; _i++) {
                                var ref = _a[_i];
                                ts.pushIfUnique(lateMarkedStatements, ref);
                            }
                        }
                    }
                    // TODO: Do all these accessibility checks inside/after the first pass in the checker when declarations are enabled, if possible
                }
                else {
                    // Report error
                    var errorInfo = getSymbolAccessibilityDiagnostic(symbolAccessibilityResult);
                    if (errorInfo) {
                        if (errorInfo.typeName) {
                            context.addDiagnostic(ts.createDiagnosticForNode(symbolAccessibilityResult.errorNode || errorInfo.errorNode, errorInfo.diagnosticMessage, ts.getTextOfNode(errorInfo.typeName), symbolAccessibilityResult.errorSymbolName, symbolAccessibilityResult.errorModuleName));
                        }
                        else {
                            context.addDiagnostic(ts.createDiagnosticForNode(symbolAccessibilityResult.errorNode || errorInfo.errorNode, errorInfo.diagnosticMessage, symbolAccessibilityResult.errorSymbolName, symbolAccessibilityResult.errorModuleName));
                        }
                        return true;
                    }
                }
                return false;
            }
            function trackExternalModuleSymbolOfImportTypeNode(symbol) {
                if (!isBundledEmit) {
                    (exportedModulesFromDeclarationEmit || (exportedModulesFromDeclarationEmit = [])).push(symbol);
                }
            }
            function trackSymbol(symbol, enclosingDeclaration, meaning) {
                if (symbol.flags & 262144 /* TypeParameter */)
                    return false;
                var issuedDiagnostic = handleSymbolAccessibilityError(resolver.isSymbolAccessible(symbol, enclosingDeclaration, meaning, /*shouldComputeAliasesToMakeVisible*/ true));
                recordTypeReferenceDirectivesIfNecessary(resolver.getTypeReferenceDirectivesForSymbol(symbol, meaning));
                return issuedDiagnostic;
            }
            function reportPrivateInBaseOfClassExpression(propertyName) {
                if (errorNameNode || errorFallbackNode) {
                    context.addDiagnostic(ts.createDiagnosticForNode((errorNameNode || errorFallbackNode), ts.Diagnostics.Property_0_of_exported_class_expression_may_not_be_private_or_protected, propertyName));
                }
            }
            function errorDeclarationNameWithFallback() {
                return errorNameNode ? ts.declarationNameToString(errorNameNode) :
                    errorFallbackNode && ts.getNameOfDeclaration(errorFallbackNode) ? ts.declarationNameToString(ts.getNameOfDeclaration(errorFallbackNode)) :
                        errorFallbackNode && ts.isExportAssignment(errorFallbackNode) ? errorFallbackNode.isExportEquals ? "export=" : "default" :
                            "(Missing)"; // same fallback declarationNameToString uses when node is zero-width (ie, nameless)
            }
            function reportInaccessibleUniqueSymbolError() {
                if (errorNameNode || errorFallbackNode) {
                    context.addDiagnostic(ts.createDiagnosticForNode((errorNameNode || errorFallbackNode), ts.Diagnostics.The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary, errorDeclarationNameWithFallback(), "unique symbol"));
                }
            }
            function reportCyclicStructureError() {
                if (errorNameNode || errorFallbackNode) {
                    context.addDiagnostic(ts.createDiagnosticForNode((errorNameNode || errorFallbackNode), ts.Diagnostics.The_inferred_type_of_0_references_a_type_with_a_cyclic_structure_which_cannot_be_trivially_serialized_A_type_annotation_is_necessary, errorDeclarationNameWithFallback()));
                }
            }
            function reportInaccessibleThisError() {
                if (errorNameNode || errorFallbackNode) {
                    context.addDiagnostic(ts.createDiagnosticForNode((errorNameNode || errorFallbackNode), ts.Diagnostics.The_inferred_type_of_0_references_an_inaccessible_1_type_A_type_annotation_is_necessary, errorDeclarationNameWithFallback(), "this"));
                }
            }
            function reportLikelyUnsafeImportRequiredError(specifier) {
                if (errorNameNode || errorFallbackNode) {
                    context.addDiagnostic(ts.createDiagnosticForNode((errorNameNode || errorFallbackNode), ts.Diagnostics.The_inferred_type_of_0_cannot_be_named_without_a_reference_to_1_This_is_likely_not_portable_A_type_annotation_is_necessary, errorDeclarationNameWithFallback(), specifier));
                }
            }
            function reportTruncationError() {
                if (errorNameNode || errorFallbackNode) {
                    context.addDiagnostic(ts.createDiagnosticForNode((errorNameNode || errorFallbackNode), ts.Diagnostics.The_inferred_type_of_this_node_exceeds_the_maximum_length_the_compiler_will_serialize_An_explicit_type_annotation_is_needed));
                }
            }
            function reportNonlocalAugmentation(containingFile, parentSymbol, symbol) {
                var _a;
                var primaryDeclaration = (_a = parentSymbol.declarations) === null || _a === void 0 ? void 0 : _a.find(function (d) { return ts.getSourceFileOfNode(d) === containingFile; });
                var augmentingDeclarations = ts.filter(symbol.declarations, function (d) { return ts.getSourceFileOfNode(d) !== containingFile; });
                if (augmentingDeclarations) {
                    for (var _i = 0, augmentingDeclarations_1 = augmentingDeclarations; _i < augmentingDeclarations_1.length; _i++) {
                        var augmentations = augmentingDeclarations_1[_i];
                        context.addDiagnostic(ts.addRelatedInfo(ts.createDiagnosticForNode(augmentations, ts.Diagnostics.Declaration_augments_declaration_in_another_file_This_cannot_be_serialized), ts.createDiagnosticForNode(primaryDeclaration, ts.Diagnostics.This_is_the_declaration_being_augmented_Consider_moving_the_augmenting_declaration_into_the_same_file)));
                    }
                }
            }
            function reportNonSerializableProperty(propertyName) {
                if (errorNameNode || errorFallbackNode) {
                    context.addDiagnostic(ts.createDiagnosticForNode((errorNameNode || errorFallbackNode), ts.Diagnostics.The_type_of_this_node_cannot_be_serialized_because_its_property_0_cannot_be_serialized, propertyName));
                }
            }
            function transformDeclarationsForJS(sourceFile, bundled) {
                var oldDiag = getSymbolAccessibilityDiagnostic;
                getSymbolAccessibilityDiagnostic = function (s) {
                    return (s.errorNode && ts.canProduceDiagnostics(s.errorNode) ? ts.createGetSymbolAccessibilityDiagnosticForNode(s.errorNode)(s) : ({
                        diagnosticMessage: s.errorModuleName
                            ? ts.Diagnostics.Declaration_emit_for_this_file_requires_using_private_name_0_from_module_1_An_explicit_type_annotation_may_unblock_declaration_emit
                            : ts.Diagnostics.Declaration_emit_for_this_file_requires_using_private_name_0_An_explicit_type_annotation_may_unblock_declaration_emit,
                        errorNode: s.errorNode || sourceFile
                    }));
                };
                var result = resolver.getDeclarationStatementsForSourceFile(sourceFile, declarationEmitNodeBuilderFlags, symbolTracker, bundled);
                getSymbolAccessibilityDiagnostic = oldDiag;
                return result;
            }
            function transformRoot(node) {
                if (node.kind === 300 /* SourceFile */ && node.isDeclarationFile) {
                    return node;
                }
                if (node.kind === 301 /* Bundle */) {
                    isBundledEmit = true;
                    refs = new ts.Map();
                    libs = new ts.Map();
                    var hasNoDefaultLib_1 = false;
                    var bundle = factory.createBundle(ts.map(node.sourceFiles, function (sourceFile) {
                        if (sourceFile.isDeclarationFile)
                            return undefined; // Omit declaration files from bundle results, too // TODO: GH#18217
                        hasNoDefaultLib_1 = hasNoDefaultLib_1 || sourceFile.hasNoDefaultLib;
                        currentSourceFile = sourceFile;
                        enclosingDeclaration = sourceFile;
                        lateMarkedStatements = undefined;
                        suppressNewDiagnosticContexts = false;
                        lateStatementReplacementMap = new ts.Map();
                        getSymbolAccessibilityDiagnostic = throwDiagnostic;
                        needsScopeFixMarker = false;
                        resultHasScopeMarker = false;
                        collectReferences(sourceFile, refs);
                        collectLibs(sourceFile, libs);
                        if (ts.isExternalOrCommonJsModule(sourceFile) || ts.isJsonSourceFile(sourceFile)) {
                            resultHasExternalModuleIndicator = false; // unused in external module bundle emit (all external modules are within module blocks, therefore are known to be modules)
                            needsDeclare = false;
                            var statements = ts.isSourceFileJS(sourceFile) ? factory.createNodeArray(transformDeclarationsForJS(sourceFile, /*bundled*/ true)) : ts.visitNodes(sourceFile.statements, visitDeclarationStatements);
                            var newFile = factory.updateSourceFile(sourceFile, [factory.createModuleDeclaration([], [factory.createModifier(134 /* DeclareKeyword */)], factory.createStringLiteral(ts.getResolvedExternalModuleName(context.getEmitHost(), sourceFile)), factory.createModuleBlock(ts.setTextRange(factory.createNodeArray(transformAndReplaceLatePaintedStatements(statements)), sourceFile.statements)))], /*isDeclarationFile*/ true, /*referencedFiles*/ [], /*typeReferences*/ [], /*hasNoDefaultLib*/ false, /*libReferences*/ []);
                            return newFile;
                        }
                        needsDeclare = true;
                        var updated = ts.isSourceFileJS(sourceFile) ? factory.createNodeArray(transformDeclarationsForJS(sourceFile)) : ts.visitNodes(sourceFile.statements, visitDeclarationStatements);
                        return factory.updateSourceFile(sourceFile, transformAndReplaceLatePaintedStatements(updated), /*isDeclarationFile*/ true, /*referencedFiles*/ [], /*typeReferences*/ [], /*hasNoDefaultLib*/ false, /*libReferences*/ []);
                    }), ts.mapDefined(node.prepends, function (prepend) {
                        if (prepend.kind === 303 /* InputFiles */) {
                            var sourceFile = ts.createUnparsedSourceFile(prepend, "dts", stripInternal);
                            hasNoDefaultLib_1 = hasNoDefaultLib_1 || !!sourceFile.hasNoDefaultLib;
                            collectReferences(sourceFile, refs);
                            recordTypeReferenceDirectivesIfNecessary(sourceFile.typeReferenceDirectives);
                            collectLibs(sourceFile, libs);
                            return sourceFile;
                        }
                        return prepend;
                    }));
                    bundle.syntheticFileReferences = [];
                    bundle.syntheticTypeReferences = getFileReferencesForUsedTypeReferences();
                    bundle.syntheticLibReferences = getLibReferences();
                    bundle.hasNoDefaultLib = hasNoDefaultLib_1;
                    var outputFilePath_1 = ts.getDirectoryPath(ts.normalizeSlashes(ts.getOutputPathsFor(node, host, /*forceDtsPaths*/ true).declarationFilePath));
                    var referenceVisitor_1 = mapReferencesIntoArray(bundle.syntheticFileReferences, outputFilePath_1);
                    refs.forEach(referenceVisitor_1);
                    return bundle;
                }
                // Single source file
                needsDeclare = true;
                needsScopeFixMarker = false;
                resultHasScopeMarker = false;
                enclosingDeclaration = node;
                currentSourceFile = node;
                getSymbolAccessibilityDiagnostic = throwDiagnostic;
                isBundledEmit = false;
                resultHasExternalModuleIndicator = false;
                suppressNewDiagnosticContexts = false;
                lateMarkedStatements = undefined;
                lateStatementReplacementMap = new ts.Map();
                necessaryTypeReferences = undefined;
                refs = collectReferences(currentSourceFile, new ts.Map());
                libs = collectLibs(currentSourceFile, new ts.Map());
                var references = [];
                var outputFilePath = ts.getDirectoryPath(ts.normalizeSlashes(ts.getOutputPathsFor(node, host, /*forceDtsPaths*/ true).declarationFilePath));
                var referenceVisitor = mapReferencesIntoArray(references, outputFilePath);
                var combinedStatements;
                if (ts.isSourceFileJS(currentSourceFile)) {
                    combinedStatements = factory.createNodeArray(transformDeclarationsForJS(node));
                    refs.forEach(referenceVisitor);
                    emittedImports = ts.filter(combinedStatements, ts.isAnyImportSyntax);
                }
                else {
                    var statements = ts.visitNodes(node.statements, visitDeclarationStatements);
                    combinedStatements = ts.setTextRange(factory.createNodeArray(transformAndReplaceLatePaintedStatements(statements)), node.statements);
                    refs.forEach(referenceVisitor);
                    emittedImports = ts.filter(combinedStatements, ts.isAnyImportSyntax);
                    if (ts.isExternalModule(node) && (!resultHasExternalModuleIndicator || (needsScopeFixMarker && !resultHasScopeMarker))) {
                        combinedStatements = ts.setTextRange(factory.createNodeArray(__spreadArray(__spreadArray([], combinedStatements, true), [ts.createEmptyExports(factory)], false)), combinedStatements);
                    }
                }
                var updated = factory.updateSourceFile(node, combinedStatements, /*isDeclarationFile*/ true, references, getFileReferencesForUsedTypeReferences(), node.hasNoDefaultLib, getLibReferences());
                updated.exportedModulesFromDeclarationEmit = exportedModulesFromDeclarationEmit;
                return updated;
                function getLibReferences() {
                    return ts.map(ts.arrayFrom(libs.keys()), function (lib) { return ({ fileName: lib, pos: -1, end: -1 }); });
                }
                function getFileReferencesForUsedTypeReferences() {
                    return necessaryTypeReferences ? ts.mapDefined(ts.arrayFrom(necessaryTypeReferences.keys()), getFileReferenceForTypeName) : [];
                }
                function getFileReferenceForTypeName(typeName) {
                    // Elide type references for which we have imports
                    if (emittedImports) {
                        for (var _i = 0, emittedImports_1 = emittedImports; _i < emittedImports_1.length; _i++) {
                            var importStatement = emittedImports_1[_i];
                            if (ts.isImportEqualsDeclaration(importStatement) && ts.isExternalModuleReference(importStatement.moduleReference)) {
                                var expr = importStatement.moduleReference.expression;
                                if (ts.isStringLiteralLike(expr) && expr.text === typeName) {
                                    return undefined;
                                }
                            }
                            else if (ts.isImportDeclaration(importStatement) && ts.isStringLiteral(importStatement.moduleSpecifier) && importStatement.moduleSpecifier.text === typeName) {
                                return undefined;
                            }
                        }
                    }
                    return { fileName: typeName, pos: -1, end: -1 };
                }
                function mapReferencesIntoArray(references, outputFilePath) {
                    return function (file) {
                        var declFileName;
                        if (file.isDeclarationFile) { // Neither decl files or js should have their refs changed
                            declFileName = file.fileName;
                        }
                        else {
                            if (isBundledEmit && ts.contains(node.sourceFiles, file))
                                return; // Omit references to files which are being merged
                            var paths = ts.getOutputPathsFor(file, host, /*forceDtsPaths*/ true);
                            declFileName = paths.declarationFilePath || paths.jsFilePath || file.fileName;
                        }
                        if (declFileName) {
                            var specifier = ts.moduleSpecifiers.getModuleSpecifier(options, currentSourceFile, ts.toPath(outputFilePath, host.getCurrentDirectory(), host.getCanonicalFileName), ts.toPath(declFileName, host.getCurrentDirectory(), host.getCanonicalFileName), host);
                            if (!ts.pathIsRelative(specifier)) {
                                // If some compiler option/symlink/whatever allows access to the file containing the ambient module declaration
                                // via a non-relative name, emit a type reference directive to that non-relative name, rather than
                                // a relative path to the declaration file
                                recordTypeReferenceDirectivesIfNecessary([specifier]);
                                return;
                            }
                            var fileName = ts.getRelativePathToDirectoryOrUrl(outputFilePath, declFileName, host.getCurrentDirectory(), host.getCanonicalFileName, 
                            /*isAbsolutePathAnUrl*/ false);
                            if (ts.startsWith(fileName, "./") && ts.hasExtension(fileName)) {
                                fileName = fileName.substring(2);
                            }
                            // omit references to files from node_modules (npm may disambiguate module
                            // references when installing this package, making the path is unreliable).
                            if (ts.startsWith(fileName, "node_modules/") || ts.pathContainsNodeModules(fileName)) {
                                return;
                            }
                            references.push({ pos: -1, end: -1, fileName: fileName });
                        }
                    };
                }
            }
            function collectReferences(sourceFile, ret) {
                if (noResolve || (!ts.isUnparsedSource(sourceFile) && ts.isSourceFileJS(sourceFile)))
                    return ret;
                ts.forEach(sourceFile.referencedFiles, function (f) {
                    var elem = host.getSourceFileFromReference(sourceFile, f);
                    if (elem) {
                        ret.set(ts.getOriginalNodeId(elem), elem);
                    }
                });
                return ret;
            }
            function collectLibs(sourceFile, ret) {
                ts.forEach(sourceFile.libReferenceDirectives, function (ref) {
                    var lib = host.getLibFileFromReference(ref);
                    if (lib) {
                        ret.set(ts.toFileNameLowerCase(ref.fileName), true);
                    }
                });
                return ret;
            }
            function filterBindingPatternInitializers(name) {
                if (name.kind === 79 /* Identifier */) {
                    return name;
                }
                else {
                    if (name.kind === 200 /* ArrayBindingPattern */) {
                        return factory.updateArrayBindingPattern(name, ts.visitNodes(name.elements, visitBindingElement));
                    }
                    else {
                        return factory.updateObjectBindingPattern(name, ts.visitNodes(name.elements, visitBindingElement));
                    }
                }
                function visitBindingElement(elem) {
                    if (elem.kind === 225 /* OmittedExpression */) {
                        return elem;
                    }
                    return factory.updateBindingElement(elem, elem.dotDotDotToken, elem.propertyName, filterBindingPatternInitializers(elem.name), shouldPrintWithInitializer(elem) ? elem.initializer : undefined);
                }
            }
            function ensureParameter(p, modifierMask, type) {
                var oldDiag;
                if (!suppressNewDiagnosticContexts) {
                    oldDiag = getSymbolAccessibilityDiagnostic;
                    getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(p);
                }
                var newParam = factory.updateParameterDeclaration(p, 
                /*decorators*/ undefined, maskModifiers(p, modifierMask), p.dotDotDotToken, filterBindingPatternInitializers(p.name), resolver.isOptionalParameter(p) ? (p.questionToken || factory.createToken(57 /* QuestionToken */)) : undefined, ensureType(p, type || p.type, /*ignorePrivate*/ true), // Ignore private param props, since this type is going straight back into a param
                ensureNoInitializer(p));
                if (!suppressNewDiagnosticContexts) {
                    getSymbolAccessibilityDiagnostic = oldDiag;
                }
                return newParam;
            }
            function shouldPrintWithInitializer(node) {
                return canHaveLiteralInitializer(node) && resolver.isLiteralConstDeclaration(ts.getParseTreeNode(node)); // TODO: Make safe
            }
            function ensureNoInitializer(node) {
                if (shouldPrintWithInitializer(node)) {
                    return resolver.createLiteralConstValue(ts.getParseTreeNode(node), symbolTracker); // TODO: Make safe
                }
                return undefined;
            }
            function ensureType(node, type, ignorePrivate) {
                if (!ignorePrivate && ts.hasEffectiveModifier(node, 8 /* Private */)) {
                    // Private nodes emit no types (except private parameter properties, whose parameter types are actually visible)
                    return;
                }
                if (shouldPrintWithInitializer(node)) {
                    // Literal const declarations will have an initializer ensured rather than a type
                    return;
                }
                var shouldUseResolverType = node.kind === 162 /* Parameter */ &&
                    (resolver.isRequiredInitializedParameter(node) ||
                        resolver.isOptionalUninitializedParameterProperty(node));
                if (type && !shouldUseResolverType) {
                    return ts.visitNode(type, visitDeclarationSubtree);
                }
                if (!ts.getParseTreeNode(node)) {
                    return type ? ts.visitNode(type, visitDeclarationSubtree) : factory.createKeywordTypeNode(129 /* AnyKeyword */);
                }
                if (node.kind === 171 /* SetAccessor */) {
                    // Set accessors with no associated type node (from it's param or get accessor return) are `any` since they are never contextually typed right now
                    // (The inferred type here will be void, but the old declaration emitter printed `any`, so this replicates that)
                    return factory.createKeywordTypeNode(129 /* AnyKeyword */);
                }
                errorNameNode = node.name;
                var oldDiag;
                if (!suppressNewDiagnosticContexts) {
                    oldDiag = getSymbolAccessibilityDiagnostic;
                    getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(node);
                }
                if (node.kind === 252 /* VariableDeclaration */ || node.kind === 201 /* BindingElement */) {
                    return cleanup(resolver.createTypeOfDeclaration(node, enclosingDeclaration, declarationEmitNodeBuilderFlags, symbolTracker));
                }
                if (node.kind === 162 /* Parameter */
                    || node.kind === 165 /* PropertyDeclaration */
                    || node.kind === 164 /* PropertySignature */) {
                    if (!node.initializer)
                        return cleanup(resolver.createTypeOfDeclaration(node, enclosingDeclaration, declarationEmitNodeBuilderFlags, symbolTracker, shouldUseResolverType));
                    return cleanup(resolver.createTypeOfDeclaration(node, enclosingDeclaration, declarationEmitNodeBuilderFlags, symbolTracker, shouldUseResolverType) || resolver.createTypeOfExpression(node.initializer, enclosingDeclaration, declarationEmitNodeBuilderFlags, symbolTracker));
                }
                return cleanup(resolver.createReturnTypeOfSignatureDeclaration(node, enclosingDeclaration, declarationEmitNodeBuilderFlags, symbolTracker));
                function cleanup(returnValue) {
                    errorNameNode = undefined;
                    if (!suppressNewDiagnosticContexts) {
                        getSymbolAccessibilityDiagnostic = oldDiag;
                    }
                    return returnValue || factory.createKeywordTypeNode(129 /* AnyKeyword */);
                }
            }
            function isDeclarationAndNotVisible(node) {
                node = ts.getParseTreeNode(node);
                switch (node.kind) {
                    case 254 /* FunctionDeclaration */:
                    case 259 /* ModuleDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 255 /* ClassDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                    case 258 /* EnumDeclaration */:
                        return !resolver.isDeclarationVisible(node);
                    // The following should be doing their own visibility checks based on filtering their members
                    case 252 /* VariableDeclaration */:
                        return !getBindingNameVisible(node);
                    case 263 /* ImportEqualsDeclaration */:
                    case 264 /* ImportDeclaration */:
                    case 270 /* ExportDeclaration */:
                    case 269 /* ExportAssignment */:
                        return false;
                    case 168 /* ClassStaticBlockDeclaration */:
                        return true;
                }
                return false;
            }
            // If the ExpandoFunctionDeclaration have multiple overloads, then we only need to emit properties for the last one.
            function shouldEmitFunctionProperties(input) {
                var _a;
                if (input.body) {
                    return true;
                }
                var overloadSignatures = (_a = input.symbol.declarations) === null || _a === void 0 ? void 0 : _a.filter(function (decl) { return ts.isFunctionDeclaration(decl) && !decl.body; });
                return !overloadSignatures || overloadSignatures.indexOf(input) === overloadSignatures.length - 1;
            }
            function getBindingNameVisible(elem) {
                if (ts.isOmittedExpression(elem)) {
                    return false;
                }
                if (ts.isBindingPattern(elem.name)) {
                    // If any child binding pattern element has been marked visible (usually by collect linked aliases), then this is visible
                    return ts.some(elem.name.elements, getBindingNameVisible);
                }
                else {
                    return resolver.isDeclarationVisible(elem);
                }
            }
            function updateParamsList(node, params, modifierMask) {
                if (ts.hasEffectiveModifier(node, 8 /* Private */)) {
                    return undefined; // TODO: GH#18217
                }
                var newParams = ts.map(params, function (p) { return ensureParameter(p, modifierMask); });
                if (!newParams) {
                    return undefined; // TODO: GH#18217
                }
                return factory.createNodeArray(newParams, params.hasTrailingComma);
            }
            function updateAccessorParamsList(input, isPrivate) {
                var newParams;
                if (!isPrivate) {
                    var thisParameter = ts.getThisParameter(input);
                    if (thisParameter) {
                        newParams = [ensureParameter(thisParameter)];
                    }
                }
                if (ts.isSetAccessorDeclaration(input)) {
                    var newValueParameter = void 0;
                    if (!isPrivate) {
                        var valueParameter = ts.getSetAccessorValueParameter(input);
                        if (valueParameter) {
                            var accessorType = getTypeAnnotationFromAllAccessorDeclarations(input, resolver.getAllAccessorDeclarations(input));
                            newValueParameter = ensureParameter(valueParameter, /*modifierMask*/ undefined, accessorType);
                        }
                    }
                    if (!newValueParameter) {
                        newValueParameter = factory.createParameterDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, 
                        /*dotDotDotToken*/ undefined, "value");
                    }
                    newParams = ts.append(newParams, newValueParameter);
                }
                return factory.createNodeArray(newParams || ts.emptyArray);
            }
            function ensureTypeParams(node, params) {
                return ts.hasEffectiveModifier(node, 8 /* Private */) ? undefined : ts.visitNodes(params, visitDeclarationSubtree);
            }
            function isEnclosingDeclaration(node) {
                return ts.isSourceFile(node)
                    || ts.isTypeAliasDeclaration(node)
                    || ts.isModuleDeclaration(node)
                    || ts.isClassDeclaration(node)
                    || ts.isInterfaceDeclaration(node)
                    || ts.isFunctionLike(node)
                    || ts.isIndexSignatureDeclaration(node)
                    || ts.isMappedTypeNode(node);
            }
            function checkEntityNameVisibility(entityName, enclosingDeclaration) {
                var visibilityResult = resolver.isEntityNameVisible(entityName, enclosingDeclaration);
                handleSymbolAccessibilityError(visibilityResult);
                recordTypeReferenceDirectivesIfNecessary(resolver.getTypeReferenceDirectivesForEntityName(entityName));
            }
            function preserveJsDoc(updated, original) {
                if (ts.hasJSDocNodes(updated) && ts.hasJSDocNodes(original)) {
                    updated.jsDoc = original.jsDoc;
                }
                return ts.setCommentRange(updated, ts.getCommentRange(original));
            }
            function rewriteModuleSpecifier(parent, input) {
                if (!input)
                    return undefined; // TODO: GH#18217
                resultHasExternalModuleIndicator = resultHasExternalModuleIndicator || (parent.kind !== 259 /* ModuleDeclaration */ && parent.kind !== 198 /* ImportType */);
                if (ts.isStringLiteralLike(input)) {
                    if (isBundledEmit) {
                        var newName = ts.getExternalModuleNameFromDeclaration(context.getEmitHost(), resolver, parent);
                        if (newName) {
                            return factory.createStringLiteral(newName);
                        }
                    }
                    else {
                        var symbol = resolver.getSymbolOfExternalModuleSpecifier(input);
                        if (symbol) {
                            (exportedModulesFromDeclarationEmit || (exportedModulesFromDeclarationEmit = [])).push(symbol);
                        }
                    }
                }
                return input;
            }
            function transformImportEqualsDeclaration(decl) {
                if (!resolver.isDeclarationVisible(decl))
                    return;
                if (decl.moduleReference.kind === 275 /* ExternalModuleReference */) {
                    // Rewrite external module names if necessary
                    var specifier = ts.getExternalModuleImportEqualsDeclarationExpression(decl);
                    return factory.updateImportEqualsDeclaration(decl, 
                    /*decorators*/ undefined, decl.modifiers, decl.isTypeOnly, decl.name, factory.updateExternalModuleReference(decl.moduleReference, rewriteModuleSpecifier(decl, specifier)));
                }
                else {
                    var oldDiag = getSymbolAccessibilityDiagnostic;
                    getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(decl);
                    checkEntityNameVisibility(decl.moduleReference, enclosingDeclaration);
                    getSymbolAccessibilityDiagnostic = oldDiag;
                    return decl;
                }
            }
            function transformImportDeclaration(decl) {
                if (!decl.importClause) {
                    // import "mod" - possibly needed for side effects? (global interface patches, module augmentations, etc)
                    return factory.updateImportDeclaration(decl, 
                    /*decorators*/ undefined, decl.modifiers, decl.importClause, rewriteModuleSpecifier(decl, decl.moduleSpecifier));
                }
                // The `importClause` visibility corresponds to the default's visibility.
                var visibleDefaultBinding = decl.importClause && decl.importClause.name && resolver.isDeclarationVisible(decl.importClause) ? decl.importClause.name : undefined;
                if (!decl.importClause.namedBindings) {
                    // No named bindings (either namespace or list), meaning the import is just default or should be elided
                    return visibleDefaultBinding && factory.updateImportDeclaration(decl, /*decorators*/ undefined, decl.modifiers, factory.updateImportClause(decl.importClause, decl.importClause.isTypeOnly, visibleDefaultBinding, 
                    /*namedBindings*/ undefined), rewriteModuleSpecifier(decl, decl.moduleSpecifier));
                }
                if (decl.importClause.namedBindings.kind === 266 /* NamespaceImport */) {
                    // Namespace import (optionally with visible default)
                    var namedBindings = resolver.isDeclarationVisible(decl.importClause.namedBindings) ? decl.importClause.namedBindings : /*namedBindings*/ undefined;
                    return visibleDefaultBinding || namedBindings ? factory.updateImportDeclaration(decl, /*decorators*/ undefined, decl.modifiers, factory.updateImportClause(decl.importClause, decl.importClause.isTypeOnly, visibleDefaultBinding, namedBindings), rewriteModuleSpecifier(decl, decl.moduleSpecifier)) : undefined;
                }
                // Named imports (optionally with visible default)
                var bindingList = ts.mapDefined(decl.importClause.namedBindings.elements, function (b) { return resolver.isDeclarationVisible(b) ? b : undefined; });
                if ((bindingList && bindingList.length) || visibleDefaultBinding) {
                    return factory.updateImportDeclaration(decl, 
                    /*decorators*/ undefined, decl.modifiers, factory.updateImportClause(decl.importClause, decl.importClause.isTypeOnly, visibleDefaultBinding, bindingList && bindingList.length ? factory.updateNamedImports(decl.importClause.namedBindings, bindingList) : undefined), rewriteModuleSpecifier(decl, decl.moduleSpecifier));
                }
                // Augmentation of export depends on import
                if (resolver.isImportRequiredByAugmentation(decl)) {
                    return factory.updateImportDeclaration(decl, 
                    /*decorators*/ undefined, decl.modifiers, 
                    /*importClause*/ undefined, rewriteModuleSpecifier(decl, decl.moduleSpecifier));
                }
                // Nothing visible
            }
            function transformAndReplaceLatePaintedStatements(statements) {
                // This is a `while` loop because `handleSymbolAccessibilityError` can see additional import aliases marked as visible during
                // error handling which must now be included in the output and themselves checked for errors.
                // For example:
                // ```
                // module A {
                //   export module Q {}
                //   import B = Q;
                //   import C = B;
                //   export import D = C;
                // }
                // ```
                // In such a scenario, only Q and D are initially visible, but we don't consider imports as private names - instead we say they if they are referenced they must
                // be recorded. So while checking D's visibility we mark C as visible, then we must check C which in turn marks B, completing the chain of
                // dependent imports and allowing a valid declaration file output. Today, this dependent alias marking only happens for internal import aliases.
                while (ts.length(lateMarkedStatements)) {
                    var i = lateMarkedStatements.shift();
                    if (!ts.isLateVisibilityPaintedStatement(i)) {
                        return ts.Debug.fail("Late replaced statement was found which is not handled by the declaration transformer!: " + (ts.SyntaxKind ? ts.SyntaxKind[i.kind] : i.kind));
                    }
                    var priorNeedsDeclare = needsDeclare;
                    needsDeclare = i.parent && ts.isSourceFile(i.parent) && !(ts.isExternalModule(i.parent) && isBundledEmit);
                    var result = transformTopLevelDeclaration(i);
                    needsDeclare = priorNeedsDeclare;
                    lateStatementReplacementMap.set(ts.getOriginalNodeId(i), result);
                }
                // And lastly, we need to get the final form of all those indetermine import declarations from before and add them to the output list
                // (and remove them from the set to examine for outter declarations)
                return ts.visitNodes(statements, visitLateVisibilityMarkedStatements);
                function visitLateVisibilityMarkedStatements(statement) {
                    if (ts.isLateVisibilityPaintedStatement(statement)) {
                        var key = ts.getOriginalNodeId(statement);
                        if (lateStatementReplacementMap.has(key)) {
                            var result = lateStatementReplacementMap.get(key);
                            lateStatementReplacementMap.delete(key);
                            if (result) {
                                if (ts.isArray(result) ? ts.some(result, ts.needsScopeMarker) : ts.needsScopeMarker(result)) {
                                    // Top-level declarations in .d.ts files are always considered exported even without a modifier unless there's an export assignment or specifier
                                    needsScopeFixMarker = true;
                                }
                                if (ts.isSourceFile(statement.parent) && (ts.isArray(result) ? ts.some(result, ts.isExternalModuleIndicator) : ts.isExternalModuleIndicator(result))) {
                                    resultHasExternalModuleIndicator = true;
                                }
                            }
                            return result;
                        }
                    }
                    return statement;
                }
            }
            function visitDeclarationSubtree(input) {
                if (shouldStripInternal(input))
                    return;
                if (ts.isDeclaration(input)) {
                    if (isDeclarationAndNotVisible(input))
                        return;
                    if (ts.hasDynamicName(input) && !resolver.isLateBound(ts.getParseTreeNode(input))) {
                        return;
                    }
                }
                // Elide implementation signatures from overload sets
                if (ts.isFunctionLike(input) && resolver.isImplementationOfOverload(input))
                    return;
                // Elide semicolon class statements
                if (ts.isSemicolonClassElement(input))
                    return;
                var previousEnclosingDeclaration;
                if (isEnclosingDeclaration(input)) {
                    previousEnclosingDeclaration = enclosingDeclaration;
                    enclosingDeclaration = input;
                }
                var oldDiag = getSymbolAccessibilityDiagnostic;
                // Setup diagnostic-related flags before first potential `cleanup` call, otherwise
                // We'd see a TDZ violation at runtime
                var canProduceDiagnostic = ts.canProduceDiagnostics(input);
                var oldWithinObjectLiteralType = suppressNewDiagnosticContexts;
                var shouldEnterSuppressNewDiagnosticsContextContext = ((input.kind === 180 /* TypeLiteral */ || input.kind === 193 /* MappedType */) && input.parent.kind !== 257 /* TypeAliasDeclaration */);
                // Emit methods which are private as properties with no type information
                if (ts.isMethodDeclaration(input) || ts.isMethodSignature(input)) {
                    if (ts.hasEffectiveModifier(input, 8 /* Private */)) {
                        if (input.symbol && input.symbol.declarations && input.symbol.declarations[0] !== input)
                            return; // Elide all but the first overload
                        return cleanup(factory.createPropertyDeclaration(/*decorators*/ undefined, ensureModifiers(input), input.name, /*questionToken*/ undefined, /*type*/ undefined, /*initializer*/ undefined));
                    }
                }
                if (canProduceDiagnostic && !suppressNewDiagnosticContexts) {
                    getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(input);
                }
                if (ts.isTypeQueryNode(input)) {
                    checkEntityNameVisibility(input.exprName, enclosingDeclaration);
                }
                if (shouldEnterSuppressNewDiagnosticsContextContext) {
                    // We stop making new diagnostic contexts within object literal types. Unless it's an object type on the RHS of a type alias declaration. Then we do.
                    suppressNewDiagnosticContexts = true;
                }
                if (isProcessedComponent(input)) {
                    switch (input.kind) {
                        case 226 /* ExpressionWithTypeArguments */: {
                            if ((ts.isEntityName(input.expression) || ts.isEntityNameExpression(input.expression))) {
                                checkEntityNameVisibility(input.expression, enclosingDeclaration);
                            }
                            var node = ts.visitEachChild(input, visitDeclarationSubtree, context);
                            return cleanup(factory.updateExpressionWithTypeArguments(node, node.expression, node.typeArguments));
                        }
                        case 176 /* TypeReference */: {
                            checkEntityNameVisibility(input.typeName, enclosingDeclaration);
                            var node = ts.visitEachChild(input, visitDeclarationSubtree, context);
                            return cleanup(factory.updateTypeReferenceNode(node, node.typeName, node.typeArguments));
                        }
                        case 173 /* ConstructSignature */:
                            return cleanup(factory.updateConstructSignature(input, ensureTypeParams(input, input.typeParameters), updateParamsList(input, input.parameters), ensureType(input, input.type)));
                        case 169 /* Constructor */: {
                            // A constructor declaration may not have a type annotation
                            var ctor = factory.createConstructorDeclaration(
                            /*decorators*/ undefined, 
                            /*modifiers*/ ensureModifiers(input), updateParamsList(input, input.parameters, 0 /* None */), 
                            /*body*/ undefined);
                            return cleanup(ctor);
                        }
                        case 167 /* MethodDeclaration */: {
                            if (ts.isPrivateIdentifier(input.name)) {
                                return cleanup(/*returnValue*/ undefined);
                            }
                            var sig = factory.createMethodDeclaration(
                            /*decorators*/ undefined, ensureModifiers(input), 
                            /*asteriskToken*/ undefined, input.name, input.questionToken, ensureTypeParams(input, input.typeParameters), updateParamsList(input, input.parameters), ensureType(input, input.type), 
                            /*body*/ undefined);
                            return cleanup(sig);
                        }
                        case 170 /* GetAccessor */: {
                            if (ts.isPrivateIdentifier(input.name)) {
                                return cleanup(/*returnValue*/ undefined);
                            }
                            var accessorType = getTypeAnnotationFromAllAccessorDeclarations(input, resolver.getAllAccessorDeclarations(input));
                            return cleanup(factory.updateGetAccessorDeclaration(input, 
                            /*decorators*/ undefined, ensureModifiers(input), input.name, updateAccessorParamsList(input, ts.hasEffectiveModifier(input, 8 /* Private */)), ensureType(input, accessorType), 
                            /*body*/ undefined));
                        }
                        case 171 /* SetAccessor */: {
                            if (ts.isPrivateIdentifier(input.name)) {
                                return cleanup(/*returnValue*/ undefined);
                            }
                            return cleanup(factory.updateSetAccessorDeclaration(input, 
                            /*decorators*/ undefined, ensureModifiers(input), input.name, updateAccessorParamsList(input, ts.hasEffectiveModifier(input, 8 /* Private */)), 
                            /*body*/ undefined));
                        }
                        case 165 /* PropertyDeclaration */:
                            if (ts.isPrivateIdentifier(input.name)) {
                                return cleanup(/*returnValue*/ undefined);
                            }
                            return cleanup(factory.updatePropertyDeclaration(input, 
                            /*decorators*/ undefined, ensureModifiers(input), input.name, input.questionToken, ensureType(input, input.type), ensureNoInitializer(input)));
                        case 164 /* PropertySignature */:
                            if (ts.isPrivateIdentifier(input.name)) {
                                return cleanup(/*returnValue*/ undefined);
                            }
                            return cleanup(factory.updatePropertySignature(input, ensureModifiers(input), input.name, input.questionToken, ensureType(input, input.type)));
                        case 166 /* MethodSignature */: {
                            if (ts.isPrivateIdentifier(input.name)) {
                                return cleanup(/*returnValue*/ undefined);
                            }
                            return cleanup(factory.updateMethodSignature(input, ensureModifiers(input), input.name, input.questionToken, ensureTypeParams(input, input.typeParameters), updateParamsList(input, input.parameters), ensureType(input, input.type)));
                        }
                        case 172 /* CallSignature */: {
                            return cleanup(factory.updateCallSignature(input, ensureTypeParams(input, input.typeParameters), updateParamsList(input, input.parameters), ensureType(input, input.type)));
                        }
                        case 174 /* IndexSignature */: {
                            return cleanup(factory.updateIndexSignature(input, 
                            /*decorators*/ undefined, ensureModifiers(input), updateParamsList(input, input.parameters), ts.visitNode(input.type, visitDeclarationSubtree) || factory.createKeywordTypeNode(129 /* AnyKeyword */)));
                        }
                        case 252 /* VariableDeclaration */: {
                            if (ts.isBindingPattern(input.name)) {
                                return recreateBindingPattern(input.name);
                            }
                            shouldEnterSuppressNewDiagnosticsContextContext = true;
                            suppressNewDiagnosticContexts = true; // Variable declaration types also suppress new diagnostic contexts, provided the contexts wouldn't be made for binding pattern types
                            return cleanup(factory.updateVariableDeclaration(input, input.name, /*exclamationToken*/ undefined, ensureType(input, input.type), ensureNoInitializer(input)));
                        }
                        case 161 /* TypeParameter */: {
                            if (isPrivateMethodTypeParameter(input) && (input.default || input.constraint)) {
                                return cleanup(factory.updateTypeParameterDeclaration(input, input.name, /*constraint*/ undefined, /*defaultType*/ undefined));
                            }
                            return cleanup(ts.visitEachChild(input, visitDeclarationSubtree, context));
                        }
                        case 187 /* ConditionalType */: {
                            // We have to process conditional types in a special way because for visibility purposes we need to push a new enclosingDeclaration
                            // just for the `infer` types in the true branch. It's an implicit declaration scope that only applies to _part_ of the type.
                            var checkType = ts.visitNode(input.checkType, visitDeclarationSubtree);
                            var extendsType = ts.visitNode(input.extendsType, visitDeclarationSubtree);
                            var oldEnclosingDecl = enclosingDeclaration;
                            enclosingDeclaration = input.trueType;
                            var trueType = ts.visitNode(input.trueType, visitDeclarationSubtree);
                            enclosingDeclaration = oldEnclosingDecl;
                            var falseType = ts.visitNode(input.falseType, visitDeclarationSubtree);
                            return cleanup(factory.updateConditionalTypeNode(input, checkType, extendsType, trueType, falseType));
                        }
                        case 177 /* FunctionType */: {
                            return cleanup(factory.updateFunctionTypeNode(input, ts.visitNodes(input.typeParameters, visitDeclarationSubtree), updateParamsList(input, input.parameters), ts.visitNode(input.type, visitDeclarationSubtree)));
                        }
                        case 178 /* ConstructorType */: {
                            return cleanup(factory.updateConstructorTypeNode(input, ensureModifiers(input), ts.visitNodes(input.typeParameters, visitDeclarationSubtree), updateParamsList(input, input.parameters), ts.visitNode(input.type, visitDeclarationSubtree)));
                        }
                        case 198 /* ImportType */: {
                            if (!ts.isLiteralImportTypeNode(input))
                                return cleanup(input);
                            return cleanup(factory.updateImportTypeNode(input, factory.updateLiteralTypeNode(input.argument, rewriteModuleSpecifier(input, input.argument.literal)), input.qualifier, ts.visitNodes(input.typeArguments, visitDeclarationSubtree, ts.isTypeNode), input.isTypeOf));
                        }
                        default: ts.Debug.assertNever(input, "Attempted to process unhandled node kind: " + ts.SyntaxKind[input.kind]);
                    }
                }
                if (ts.isTupleTypeNode(input) && (ts.getLineAndCharacterOfPosition(currentSourceFile, input.pos).line === ts.getLineAndCharacterOfPosition(currentSourceFile, input.end).line)) {
                    ts.setEmitFlags(input, 1 /* SingleLine */);
                }
                return cleanup(ts.visitEachChild(input, visitDeclarationSubtree, context));
                function cleanup(returnValue) {
                    if (returnValue && canProduceDiagnostic && ts.hasDynamicName(input)) {
                        checkName(input);
                    }
                    if (isEnclosingDeclaration(input)) {
                        enclosingDeclaration = previousEnclosingDeclaration;
                    }
                    if (canProduceDiagnostic && !suppressNewDiagnosticContexts) {
                        getSymbolAccessibilityDiagnostic = oldDiag;
                    }
                    if (shouldEnterSuppressNewDiagnosticsContextContext) {
                        suppressNewDiagnosticContexts = oldWithinObjectLiteralType;
                    }
                    if (returnValue === input) {
                        return returnValue;
                    }
                    return returnValue && ts.setOriginalNode(preserveJsDoc(returnValue, input), input);
                }
            }
            function isPrivateMethodTypeParameter(node) {
                return node.parent.kind === 167 /* MethodDeclaration */ && ts.hasEffectiveModifier(node.parent, 8 /* Private */);
            }
            function visitDeclarationStatements(input) {
                if (!isPreservedDeclarationStatement(input)) {
                    // return undefined for unmatched kinds to omit them from the tree
                    return;
                }
                if (shouldStripInternal(input))
                    return;
                switch (input.kind) {
                    case 270 /* ExportDeclaration */: {
                        if (ts.isSourceFile(input.parent)) {
                            resultHasExternalModuleIndicator = true;
                        }
                        resultHasScopeMarker = true;
                        // Always visible if the parent node isn't dropped for being not visible
                        // Rewrite external module names if necessary
                        return factory.updateExportDeclaration(input, 
                        /*decorators*/ undefined, input.modifiers, input.isTypeOnly, input.exportClause, rewriteModuleSpecifier(input, input.moduleSpecifier));
                    }
                    case 269 /* ExportAssignment */: {
                        // Always visible if the parent node isn't dropped for being not visible
                        if (ts.isSourceFile(input.parent)) {
                            resultHasExternalModuleIndicator = true;
                        }
                        resultHasScopeMarker = true;
                        if (input.expression.kind === 79 /* Identifier */) {
                            return input;
                        }
                        else {
                            var newId = factory.createUniqueName("_default", 16 /* Optimistic */);
                            getSymbolAccessibilityDiagnostic = function () {
                                return ({
                                    diagnosticMessage: ts.Diagnostics.Default_export_of_the_module_has_or_is_using_private_name_0,
                                    errorNode: input
                                });
                            };
                            errorFallbackNode = input;
                            var varDecl = factory.createVariableDeclaration(newId, /*exclamationToken*/ undefined, resolver.createTypeOfExpression(input.expression, input, declarationEmitNodeBuilderFlags, symbolTracker), /*initializer*/ undefined);
                            errorFallbackNode = undefined;
                            var statement = factory.createVariableStatement(needsDeclare ? [factory.createModifier(134 /* DeclareKeyword */)] : [], factory.createVariableDeclarationList([varDecl], 2 /* Const */));
                            return [statement, factory.updateExportAssignment(input, input.decorators, input.modifiers, newId)];
                        }
                    }
                }
                var result = transformTopLevelDeclaration(input);
                // Don't actually transform yet; just leave as original node - will be elided/swapped by late pass
                lateStatementReplacementMap.set(ts.getOriginalNodeId(input), result);
                return input;
            }
            function stripExportModifiers(statement) {
                if (ts.isImportEqualsDeclaration(statement) || ts.hasEffectiveModifier(statement, 512 /* Default */) || !ts.canHaveModifiers(statement)) {
                    // `export import` statements should remain as-is, as imports are _not_ implicitly exported in an ambient namespace
                    // Likewise, `export default` classes and the like and just be `default`, so we preserve their `export` modifiers, too
                    return statement;
                }
                var modifiers = factory.createModifiersFromModifierFlags(ts.getEffectiveModifierFlags(statement) & (27647 /* All */ ^ 1 /* Export */));
                return factory.updateModifiers(statement, modifiers);
            }
            function transformTopLevelDeclaration(input) {
                if (shouldStripInternal(input))
                    return;
                switch (input.kind) {
                    case 263 /* ImportEqualsDeclaration */: {
                        return transformImportEqualsDeclaration(input);
                    }
                    case 264 /* ImportDeclaration */: {
                        return transformImportDeclaration(input);
                    }
                }
                if (ts.isDeclaration(input) && isDeclarationAndNotVisible(input))
                    return;
                // Elide implementation signatures from overload sets
                if (ts.isFunctionLike(input) && resolver.isImplementationOfOverload(input))
                    return;
                var previousEnclosingDeclaration;
                if (isEnclosingDeclaration(input)) {
                    previousEnclosingDeclaration = enclosingDeclaration;
                    enclosingDeclaration = input;
                }
                var canProdiceDiagnostic = ts.canProduceDiagnostics(input);
                var oldDiag = getSymbolAccessibilityDiagnostic;
                if (canProdiceDiagnostic) {
                    getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(input);
                }
                var previousNeedsDeclare = needsDeclare;
                switch (input.kind) {
                    case 257 /* TypeAliasDeclaration */: // Type aliases get `declare`d if need be (for legacy support), but that's all
                        return cleanup(factory.updateTypeAliasDeclaration(input, 
                        /*decorators*/ undefined, ensureModifiers(input), input.name, ts.visitNodes(input.typeParameters, visitDeclarationSubtree, ts.isTypeParameterDeclaration), ts.visitNode(input.type, visitDeclarationSubtree, ts.isTypeNode)));
                    case 256 /* InterfaceDeclaration */: {
                        return cleanup(factory.updateInterfaceDeclaration(input, 
                        /*decorators*/ undefined, ensureModifiers(input), input.name, ensureTypeParams(input, input.typeParameters), transformHeritageClauses(input.heritageClauses), ts.visitNodes(input.members, visitDeclarationSubtree)));
                    }
                    case 254 /* FunctionDeclaration */: {
                        // Generators lose their generator-ness, excepting their return type
                        var clean = cleanup(factory.updateFunctionDeclaration(input, 
                        /*decorators*/ undefined, ensureModifiers(input), 
                        /*asteriskToken*/ undefined, input.name, ensureTypeParams(input, input.typeParameters), updateParamsList(input, input.parameters), ensureType(input, input.type), 
                        /*body*/ undefined));
                        if (clean && resolver.isExpandoFunctionDeclaration(input) && shouldEmitFunctionProperties(input)) {
                            var props = resolver.getPropertiesOfContainerFunction(input);
                            // Use parseNodeFactory so it is usable as an enclosing declaration
                            var fakespace_1 = ts.parseNodeFactory.createModuleDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, clean.name || factory.createIdentifier("_default"), factory.createModuleBlock([]), 16 /* Namespace */);
                            ts.setParent(fakespace_1, enclosingDeclaration);
                            fakespace_1.locals = ts.createSymbolTable(props);
                            fakespace_1.symbol = props[0].parent;
                            var exportMappings_1 = [];
                            var declarations = ts.mapDefined(props, function (p) {
                                if (!p.valueDeclaration || !ts.isPropertyAccessExpression(p.valueDeclaration)) {
                                    return undefined; // TODO GH#33569: Handle element access expressions that created late bound names (rather than silently omitting them)
                                }
                                getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(p.valueDeclaration);
                                var type = resolver.createTypeOfDeclaration(p.valueDeclaration, fakespace_1, declarationEmitNodeBuilderFlags, symbolTracker);
                                getSymbolAccessibilityDiagnostic = oldDiag;
                                var nameStr = ts.unescapeLeadingUnderscores(p.escapedName);
                                var isNonContextualKeywordName = ts.isStringANonContextualKeyword(nameStr);
                                var name = isNonContextualKeywordName ? factory.getGeneratedNameForNode(p.valueDeclaration) : factory.createIdentifier(nameStr);
                                if (isNonContextualKeywordName) {
                                    exportMappings_1.push([name, nameStr]);
                                }
                                var varDecl = factory.createVariableDeclaration(name, /*exclamationToken*/ undefined, type, /*initializer*/ undefined);
                                return factory.createVariableStatement(isNonContextualKeywordName ? undefined : [factory.createToken(93 /* ExportKeyword */)], factory.createVariableDeclarationList([varDecl]));
                            });
                            if (!exportMappings_1.length) {
                                declarations = ts.mapDefined(declarations, function (declaration) { return factory.updateModifiers(declaration, 0 /* None */); });
                            }
                            else {
                                declarations.push(factory.createExportDeclaration(
                                /*decorators*/ undefined, 
                                /*modifiers*/ undefined, 
                                /*isTypeOnly*/ false, factory.createNamedExports(ts.map(exportMappings_1, function (_a) {
                                    var gen = _a[0], exp = _a[1];
                                    return factory.createExportSpecifier(gen, exp);
                                }))));
                            }
                            var namespaceDecl = factory.createModuleDeclaration(/*decorators*/ undefined, ensureModifiers(input), input.name, factory.createModuleBlock(declarations), 16 /* Namespace */);
                            if (!ts.hasEffectiveModifier(clean, 512 /* Default */)) {
                                return [clean, namespaceDecl];
                            }
                            var modifiers = factory.createModifiersFromModifierFlags((ts.getEffectiveModifierFlags(clean) & ~513 /* ExportDefault */) | 2 /* Ambient */);
                            var cleanDeclaration = factory.updateFunctionDeclaration(clean, 
                            /*decorators*/ undefined, modifiers, 
                            /*asteriskToken*/ undefined, clean.name, clean.typeParameters, clean.parameters, clean.type, 
                            /*body*/ undefined);
                            var namespaceDeclaration = factory.updateModuleDeclaration(namespaceDecl, 
                            /*decorators*/ undefined, modifiers, namespaceDecl.name, namespaceDecl.body);
                            var exportDefaultDeclaration = factory.createExportAssignment(
                            /*decorators*/ undefined, 
                            /*modifiers*/ undefined, 
                            /*isExportEquals*/ false, namespaceDecl.name);
                            if (ts.isSourceFile(input.parent)) {
                                resultHasExternalModuleIndicator = true;
                            }
                            resultHasScopeMarker = true;
                            return [cleanDeclaration, namespaceDeclaration, exportDefaultDeclaration];
                        }
                        else {
                            return clean;
                        }
                    }
                    case 259 /* ModuleDeclaration */: {
                        needsDeclare = false;
                        var inner = input.body;
                        if (inner && inner.kind === 260 /* ModuleBlock */) {
                            var oldNeedsScopeFix = needsScopeFixMarker;
                            var oldHasScopeFix = resultHasScopeMarker;
                            resultHasScopeMarker = false;
                            needsScopeFixMarker = false;
                            var statements = ts.visitNodes(inner.statements, visitDeclarationStatements);
                            var lateStatements = transformAndReplaceLatePaintedStatements(statements);
                            if (input.flags & 8388608 /* Ambient */) {
                                needsScopeFixMarker = false; // If it was `declare`'d everything is implicitly exported already, ignore late printed "privates"
                            }
                            // With the final list of statements, there are 3 possibilities:
                            // 1. There's an export assignment or export declaration in the namespace - do nothing
                            // 2. Everything is exported and there are no export assignments or export declarations - strip all export modifiers
                            // 3. Some things are exported, some are not, and there's no marker - add an empty marker
                            if (!ts.isGlobalScopeAugmentation(input) && !hasScopeMarker(lateStatements) && !resultHasScopeMarker) {
                                if (needsScopeFixMarker) {
                                    lateStatements = factory.createNodeArray(__spreadArray(__spreadArray([], lateStatements, true), [ts.createEmptyExports(factory)], false));
                                }
                                else {
                                    lateStatements = ts.visitNodes(lateStatements, stripExportModifiers);
                                }
                            }
                            var body = factory.updateModuleBlock(inner, lateStatements);
                            needsDeclare = previousNeedsDeclare;
                            needsScopeFixMarker = oldNeedsScopeFix;
                            resultHasScopeMarker = oldHasScopeFix;
                            var mods = ensureModifiers(input);
                            return cleanup(factory.updateModuleDeclaration(input, 
                            /*decorators*/ undefined, mods, ts.isExternalModuleAugmentation(input) ? rewriteModuleSpecifier(input, input.name) : input.name, body));
                        }
                        else {
                            needsDeclare = previousNeedsDeclare;
                            var mods = ensureModifiers(input);
                            needsDeclare = false;
                            ts.visitNode(inner, visitDeclarationStatements);
                            // eagerly transform nested namespaces (the nesting doesn't need any elision or painting done)
                            var id = ts.getOriginalNodeId(inner); // TODO: GH#18217
                            var body = lateStatementReplacementMap.get(id);
                            lateStatementReplacementMap.delete(id);
                            return cleanup(factory.updateModuleDeclaration(input, 
                            /*decorators*/ undefined, mods, input.name, body));
                        }
                    }
                    case 255 /* ClassDeclaration */: {
                        errorNameNode = input.name;
                        errorFallbackNode = input;
                        var modifiers = factory.createNodeArray(ensureModifiers(input));
                        var typeParameters = ensureTypeParams(input, input.typeParameters);
                        var ctor = ts.getFirstConstructorWithBody(input);
                        var parameterProperties = void 0;
                        if (ctor) {
                            var oldDiag_1 = getSymbolAccessibilityDiagnostic;
                            parameterProperties = ts.compact(ts.flatMap(ctor.parameters, function (param) {
                                if (!ts.hasSyntacticModifier(param, 16476 /* ParameterPropertyModifier */) || shouldStripInternal(param))
                                    return;
                                getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(param);
                                if (param.name.kind === 79 /* Identifier */) {
                                    return preserveJsDoc(factory.createPropertyDeclaration(
                                    /*decorators*/ undefined, ensureModifiers(param), param.name, param.questionToken, ensureType(param, param.type), ensureNoInitializer(param)), param);
                                }
                                else {
                                    // Pattern - this is currently an error, but we emit declarations for it somewhat correctly
                                    return walkBindingPattern(param.name);
                                }
                                function walkBindingPattern(pattern) {
                                    var elems;
                                    for (var _i = 0, _a = pattern.elements; _i < _a.length; _i++) {
                                        var elem = _a[_i];
                                        if (ts.isOmittedExpression(elem))
                                            continue;
                                        if (ts.isBindingPattern(elem.name)) {
                                            elems = ts.concatenate(elems, walkBindingPattern(elem.name));
                                        }
                                        elems = elems || [];
                                        elems.push(factory.createPropertyDeclaration(
                                        /*decorators*/ undefined, ensureModifiers(param), elem.name, 
                                        /*questionToken*/ undefined, ensureType(elem, /*type*/ undefined), 
                                        /*initializer*/ undefined));
                                    }
                                    return elems;
                                }
                            }));
                            getSymbolAccessibilityDiagnostic = oldDiag_1;
                        }
                        var hasPrivateIdentifier = ts.some(input.members, function (member) { return !!member.name && ts.isPrivateIdentifier(member.name); });
                        // When the class has at least one private identifier, create a unique constant identifier to retain the nominal typing behavior
                        // Prevents other classes with the same public members from being used in place of the current class
                        var privateIdentifier = hasPrivateIdentifier ? [
                            factory.createPropertyDeclaration(
                            /*decorators*/ undefined, 
                            /*modifiers*/ undefined, factory.createPrivateIdentifier("#private"), 
                            /*questionToken*/ undefined, 
                            /*type*/ undefined, 
                            /*initializer*/ undefined)
                        ] : undefined;
                        var memberNodes = ts.concatenate(ts.concatenate(privateIdentifier, parameterProperties), ts.visitNodes(input.members, visitDeclarationSubtree));
                        var members = factory.createNodeArray(memberNodes);
                        var extendsClause_1 = ts.getEffectiveBaseTypeNode(input);
                        if (extendsClause_1 && !ts.isEntityNameExpression(extendsClause_1.expression) && extendsClause_1.expression.kind !== 104 /* NullKeyword */) {
                            // We must add a temporary declaration for the extends clause expression
                            var oldId = input.name ? ts.unescapeLeadingUnderscores(input.name.escapedText) : "default";
                            var newId_1 = factory.createUniqueName(oldId + "_base", 16 /* Optimistic */);
                            getSymbolAccessibilityDiagnostic = function () {
                                return ({
                                    diagnosticMessage: ts.Diagnostics.extends_clause_of_exported_class_0_has_or_is_using_private_name_1,
                                    errorNode: extendsClause_1,
                                    typeName: input.name
                                });
                            };
                            var varDecl = factory.createVariableDeclaration(newId_1, /*exclamationToken*/ undefined, resolver.createTypeOfExpression(extendsClause_1.expression, input, declarationEmitNodeBuilderFlags, symbolTracker), /*initializer*/ undefined);
                            var statement = factory.createVariableStatement(needsDeclare ? [factory.createModifier(134 /* DeclareKeyword */)] : [], factory.createVariableDeclarationList([varDecl], 2 /* Const */));
                            var heritageClauses = factory.createNodeArray(ts.map(input.heritageClauses, function (clause) {
                                if (clause.token === 94 /* ExtendsKeyword */) {
                                    var oldDiag_2 = getSymbolAccessibilityDiagnostic;
                                    getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(clause.types[0]);
                                    var newClause = factory.updateHeritageClause(clause, ts.map(clause.types, function (t) { return factory.updateExpressionWithTypeArguments(t, newId_1, ts.visitNodes(t.typeArguments, visitDeclarationSubtree)); }));
                                    getSymbolAccessibilityDiagnostic = oldDiag_2;
                                    return newClause;
                                }
                                return factory.updateHeritageClause(clause, ts.visitNodes(factory.createNodeArray(ts.filter(clause.types, function (t) { return ts.isEntityNameExpression(t.expression) || t.expression.kind === 104 /* NullKeyword */; })), visitDeclarationSubtree));
                            }));
                            return [statement, cleanup(factory.updateClassDeclaration(input, 
                                /*decorators*/ undefined, modifiers, input.name, typeParameters, heritageClauses, members))]; // TODO: GH#18217
                        }
                        else {
                            var heritageClauses = transformHeritageClauses(input.heritageClauses);
                            return cleanup(factory.updateClassDeclaration(input, 
                            /*decorators*/ undefined, modifiers, input.name, typeParameters, heritageClauses, members));
                        }
                    }
                    case 235 /* VariableStatement */: {
                        return cleanup(transformVariableStatement(input));
                    }
                    case 258 /* EnumDeclaration */: {
                        return cleanup(factory.updateEnumDeclaration(input, /*decorators*/ undefined, factory.createNodeArray(ensureModifiers(input)), input.name, factory.createNodeArray(ts.mapDefined(input.members, function (m) {
                            if (shouldStripInternal(m))
                                return;
                            // Rewrite enum values to their constants, if available
                            var constValue = resolver.getConstantValue(m);
                            return preserveJsDoc(factory.updateEnumMember(m, m.name, constValue !== undefined ? typeof constValue === "string" ? factory.createStringLiteral(constValue) : factory.createNumericLiteral(constValue) : undefined), m);
                        }))));
                    }
                }
                // Anything left unhandled is an error, so this should be unreachable
                return ts.Debug.assertNever(input, "Unhandled top-level node in declaration emit: " + ts.SyntaxKind[input.kind]);
                function cleanup(node) {
                    if (isEnclosingDeclaration(input)) {
                        enclosingDeclaration = previousEnclosingDeclaration;
                    }
                    if (canProdiceDiagnostic) {
                        getSymbolAccessibilityDiagnostic = oldDiag;
                    }
                    if (input.kind === 259 /* ModuleDeclaration */) {
                        needsDeclare = previousNeedsDeclare;
                    }
                    if (node === input) {
                        return node;
                    }
                    errorFallbackNode = undefined;
                    errorNameNode = undefined;
                    return node && ts.setOriginalNode(preserveJsDoc(node, input), input);
                }
            }
            function transformVariableStatement(input) {
                if (!ts.forEach(input.declarationList.declarations, getBindingNameVisible))
                    return;
                var nodes = ts.visitNodes(input.declarationList.declarations, visitDeclarationSubtree);
                if (!ts.length(nodes))
                    return;
                return factory.updateVariableStatement(input, factory.createNodeArray(ensureModifiers(input)), factory.updateVariableDeclarationList(input.declarationList, nodes));
            }
            function recreateBindingPattern(d) {
                return ts.flatten(ts.mapDefined(d.elements, function (e) { return recreateBindingElement(e); }));
            }
            function recreateBindingElement(e) {
                if (e.kind === 225 /* OmittedExpression */) {
                    return;
                }
                if (e.name) {
                    if (!getBindingNameVisible(e))
                        return;
                    if (ts.isBindingPattern(e.name)) {
                        return recreateBindingPattern(e.name);
                    }
                    else {
                        return factory.createVariableDeclaration(e.name, /*exclamationToken*/ undefined, ensureType(e, /*type*/ undefined), /*initializer*/ undefined);
                    }
                }
            }
            function checkName(node) {
                var oldDiag;
                if (!suppressNewDiagnosticContexts) {
                    oldDiag = getSymbolAccessibilityDiagnostic;
                    getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNodeName(node);
                }
                errorNameNode = node.name;
                ts.Debug.assert(resolver.isLateBound(ts.getParseTreeNode(node))); // Should only be called with dynamic names
                var decl = node;
                var entityName = decl.name.expression;
                checkEntityNameVisibility(entityName, enclosingDeclaration);
                if (!suppressNewDiagnosticContexts) {
                    getSymbolAccessibilityDiagnostic = oldDiag;
                }
                errorNameNode = undefined;
            }
            function shouldStripInternal(node) {
                return !!stripInternal && !!node && isInternalDeclaration(node, currentSourceFile);
            }
            function isScopeMarker(node) {
                return ts.isExportAssignment(node) || ts.isExportDeclaration(node);
            }
            function hasScopeMarker(statements) {
                return ts.some(statements, isScopeMarker);
            }
            function ensureModifiers(node) {
                var currentFlags = ts.getEffectiveModifierFlags(node);
                var newFlags = ensureModifierFlags(node);
                if (currentFlags === newFlags) {
                    return node.modifiers;
                }
                return factory.createModifiersFromModifierFlags(newFlags);
            }
            function ensureModifierFlags(node) {
                var mask = 27647 /* All */ ^ (4 /* Public */ | 256 /* Async */ | 16384 /* Override */); // No async and override modifiers in declaration files
                var additions = (needsDeclare && !isAlwaysType(node)) ? 2 /* Ambient */ : 0 /* None */;
                var parentIsFile = node.parent.kind === 300 /* SourceFile */;
                if (!parentIsFile || (isBundledEmit && parentIsFile && ts.isExternalModule(node.parent))) {
                    mask ^= 2 /* Ambient */;
                    additions = 0 /* None */;
                }
                return maskModifierFlags(node, mask, additions);
            }
            function getTypeAnnotationFromAllAccessorDeclarations(node, accessors) {
                var accessorType = getTypeAnnotationFromAccessor(node);
                if (!accessorType && node !== accessors.firstAccessor) {
                    accessorType = getTypeAnnotationFromAccessor(accessors.firstAccessor);
                    // If we end up pulling the type from the second accessor, we also need to change the diagnostic context to get the expected error message
                    getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(accessors.firstAccessor);
                }
                if (!accessorType && accessors.secondAccessor && node !== accessors.secondAccessor) {
                    accessorType = getTypeAnnotationFromAccessor(accessors.secondAccessor);
                    // If we end up pulling the type from the second accessor, we also need to change the diagnostic context to get the expected error message
                    getSymbolAccessibilityDiagnostic = ts.createGetSymbolAccessibilityDiagnosticForNode(accessors.secondAccessor);
                }
                return accessorType;
            }
            function transformHeritageClauses(nodes) {
                return factory.createNodeArray(ts.filter(ts.map(nodes, function (clause) {
                    return factory.updateHeritageClause(clause, ts.visitNodes(factory.createNodeArray(ts.filter(clause.types, function (t) {
                        return ts.isEntityNameExpression(t.expression) || (clause.token === 94 /* ExtendsKeyword */ && t.expression.kind === 104 /* NullKeyword */);
                    })), visitDeclarationSubtree));
                }), function (clause) { return clause.types && !!clause.types.length; }));
            }
        }
        ts.transformDeclarations = transformDeclarations;
        function isAlwaysType(node) {
            if (node.kind === 256 /* InterfaceDeclaration */) {
                return true;
            }
            return false;
        }
        // Elide "public" modifier, as it is the default
        function maskModifiers(node, modifierMask, modifierAdditions) {
            return ts.factory.createModifiersFromModifierFlags(maskModifierFlags(node, modifierMask, modifierAdditions));
        }
        function maskModifierFlags(node, modifierMask, modifierAdditions) {
            if (modifierMask === void 0) {
                modifierMask = 27647 /* All */ ^ 4 /* Public */;
            }
            if (modifierAdditions === void 0) {
                modifierAdditions = 0 /* None */;
            }
            var flags = (ts.getEffectiveModifierFlags(node) & modifierMask) | modifierAdditions;
            if (flags & 512 /* Default */ && !(flags & 1 /* Export */)) {
                // A non-exported default is a nonsequitor - we usually try to remove all export modifiers
                // from statements in ambient declarations; but a default export must retain its export modifier to be syntactically valid
                flags ^= 1 /* Export */;
            }
            if (flags & 512 /* Default */ && flags & 2 /* Ambient */) {
                flags ^= 2 /* Ambient */; // `declare` is never required alongside `default` (and would be an error if printed)
            }
            return flags;
        }
        function getTypeAnnotationFromAccessor(accessor) {
            if (accessor) {
                return accessor.kind === 170 /* GetAccessor */
                    ? accessor.type // Getter - return type
                    : accessor.parameters.length > 0
                        ? accessor.parameters[0].type // Setter parameter type
                        : undefined;
            }
        }
        function canHaveLiteralInitializer(node) {
            switch (node.kind) {
                case 165 /* PropertyDeclaration */:
                case 164 /* PropertySignature */:
                    return !ts.hasEffectiveModifier(node, 8 /* Private */);
                case 162 /* Parameter */:
                case 252 /* VariableDeclaration */:
                    return true;
            }
            return false;
        }
        function isPreservedDeclarationStatement(node) {
            switch (node.kind) {
                case 254 /* FunctionDeclaration */:
                case 259 /* ModuleDeclaration */:
                case 263 /* ImportEqualsDeclaration */:
                case 256 /* InterfaceDeclaration */:
                case 255 /* ClassDeclaration */:
                case 257 /* TypeAliasDeclaration */:
                case 258 /* EnumDeclaration */:
                case 235 /* VariableStatement */:
                case 264 /* ImportDeclaration */:
                case 270 /* ExportDeclaration */:
                case 269 /* ExportAssignment */:
                    return true;
            }
            return false;
        }
        function isProcessedComponent(node) {
            switch (node.kind) {
                case 173 /* ConstructSignature */:
                case 169 /* Constructor */:
                case 167 /* MethodDeclaration */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 165 /* PropertyDeclaration */:
                case 164 /* PropertySignature */:
                case 166 /* MethodSignature */:
                case 172 /* CallSignature */:
                case 174 /* IndexSignature */:
                case 252 /* VariableDeclaration */:
                case 161 /* TypeParameter */:
                case 226 /* ExpressionWithTypeArguments */:
                case 176 /* TypeReference */:
                case 187 /* ConditionalType */:
                case 177 /* FunctionType */:
                case 178 /* ConstructorType */:
                case 198 /* ImportType */:
                    return true;
            }
            return false;
        }
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function getModuleTransformer(moduleKind) {
            switch (moduleKind) {
                case ts.ModuleKind.ESNext:
                case ts.ModuleKind.ES2020:
                case ts.ModuleKind.ES2015:
                    return ts.transformECMAScriptModule;
                case ts.ModuleKind.System:
                    return ts.transformSystemModule;
                default:
                    return ts.transformModule;
            }
        }
        var TransformationState;
        (function (TransformationState) {
            TransformationState[TransformationState["Uninitialized"] = 0] = "Uninitialized";
            TransformationState[TransformationState["Initialized"] = 1] = "Initialized";
            TransformationState[TransformationState["Completed"] = 2] = "Completed";
            TransformationState[TransformationState["Disposed"] = 3] = "Disposed";
        })(TransformationState || (TransformationState = {}));
        var SyntaxKindFeatureFlags;
        (function (SyntaxKindFeatureFlags) {
            SyntaxKindFeatureFlags[SyntaxKindFeatureFlags["Substitution"] = 1] = "Substitution";
            SyntaxKindFeatureFlags[SyntaxKindFeatureFlags["EmitNotifications"] = 2] = "EmitNotifications";
        })(SyntaxKindFeatureFlags || (SyntaxKindFeatureFlags = {}));
        ts.noTransformers = { scriptTransformers: ts.emptyArray, declarationTransformers: ts.emptyArray };
        function getTransformers(compilerOptions, customTransformers, emitOnlyDtsFiles) {
            return {
                scriptTransformers: getScriptTransformers(compilerOptions, customTransformers, emitOnlyDtsFiles),
                declarationTransformers: getDeclarationTransformers(customTransformers),
            };
        }
        ts.getTransformers = getTransformers;
        function getScriptTransformers(compilerOptions, customTransformers, emitOnlyDtsFiles) {
            if (emitOnlyDtsFiles)
                return ts.emptyArray;
            var languageVersion = ts.getEmitScriptTarget(compilerOptions);
            var moduleKind = ts.getEmitModuleKind(compilerOptions);
            var transformers = [];
            ts.addRange(transformers, customTransformers && ts.map(customTransformers.before, wrapScriptTransformerFactory));
            transformers.push(ts.transformTypeScript);
            transformers.push(ts.transformClassFields);
            if (ts.getJSXTransformEnabled(compilerOptions)) {
                transformers.push(ts.transformJsx);
            }
            if (languageVersion < 99 /* ESNext */) {
                transformers.push(ts.transformESNext);
            }
            if (languageVersion < 8 /* ES2021 */) {
                transformers.push(ts.transformES2021);
            }
            if (languageVersion < 7 /* ES2020 */) {
                transformers.push(ts.transformES2020);
            }
            if (languageVersion < 6 /* ES2019 */) {
                transformers.push(ts.transformES2019);
            }
            if (languageVersion < 5 /* ES2018 */) {
                transformers.push(ts.transformES2018);
            }
            if (languageVersion < 4 /* ES2017 */) {
                transformers.push(ts.transformES2017);
            }
            if (languageVersion < 3 /* ES2016 */) {
                transformers.push(ts.transformES2016);
            }
            if (languageVersion < 2 /* ES2015 */) {
                transformers.push(ts.transformES2015);
                transformers.push(ts.transformGenerators);
            }
            transformers.push(getModuleTransformer(moduleKind));
            // The ES5 transformer is last so that it can substitute expressions like `exports.default`
            // for ES3.
            if (languageVersion < 1 /* ES5 */) {
                transformers.push(ts.transformES5);
            }
            ts.addRange(transformers, customTransformers && ts.map(customTransformers.after, wrapScriptTransformerFactory));
            return transformers;
        }
        function getDeclarationTransformers(customTransformers) {
            var transformers = [];
            transformers.push(ts.transformDeclarations);
            ts.addRange(transformers, customTransformers && ts.map(customTransformers.afterDeclarations, wrapDeclarationTransformerFactory));
            return transformers;
        }
        /**
         * Wrap a custom script or declaration transformer object in a `Transformer` callback with fallback support for transforming bundles.
         */
        function wrapCustomTransformer(transformer) {
            return function (node) { return ts.isBundle(node) ? transformer.transformBundle(node) : transformer.transformSourceFile(node); };
        }
        /**
         * Wrap a transformer factory that may return a custom script or declaration transformer object.
         */
        function wrapCustomTransformerFactory(transformer, handleDefault) {
            return function (context) {
                var customTransformer = transformer(context);
                return typeof customTransformer === "function"
                    ? handleDefault(context, customTransformer)
                    : wrapCustomTransformer(customTransformer);
            };
        }
        function wrapScriptTransformerFactory(transformer) {
            return wrapCustomTransformerFactory(transformer, ts.chainBundle);
        }
        function wrapDeclarationTransformerFactory(transformer) {
            return wrapCustomTransformerFactory(transformer, function (_, node) { return node; });
        }
        function noEmitSubstitution(_hint, node) {
            return node;
        }
        ts.noEmitSubstitution = noEmitSubstitution;
        function noEmitNotification(hint, node, callback) {
            callback(hint, node);
        }
        ts.noEmitNotification = noEmitNotification;
        /**
         * Transforms an array of SourceFiles by passing them through each transformer.
         *
         * @param resolver The emit resolver provided by the checker.
         * @param host The emit host object used to interact with the file system.
         * @param options Compiler options to surface in the `TransformationContext`.
         * @param nodes An array of nodes to transform.
         * @param transforms An array of `TransformerFactory` callbacks.
         * @param allowDtsFiles A value indicating whether to allow the transformation of .d.ts files.
         */
        function transformNodes(resolver, host, factory, options, nodes, transformers, allowDtsFiles) {
            var enabledSyntaxKindFeatures = new Array(350 /* Count */);
            var lexicalEnvironmentVariableDeclarations;
            var lexicalEnvironmentFunctionDeclarations;
            var lexicalEnvironmentStatements;
            var lexicalEnvironmentFlags = 0 /* None */;
            var lexicalEnvironmentVariableDeclarationsStack = [];
            var lexicalEnvironmentFunctionDeclarationsStack = [];
            var lexicalEnvironmentStatementsStack = [];
            var lexicalEnvironmentFlagsStack = [];
            var lexicalEnvironmentStackOffset = 0;
            var lexicalEnvironmentSuspended = false;
            var blockScopedVariableDeclarationsStack = [];
            var blockScopeStackOffset = 0;
            var blockScopedVariableDeclarations;
            var emitHelpers;
            var onSubstituteNode = noEmitSubstitution;
            var onEmitNode = noEmitNotification;
            var state = 0 /* Uninitialized */;
            var diagnostics = [];
            // The transformation context is provided to each transformer as part of transformer
            // initialization.
            var context = {
                factory: factory,
                getCompilerOptions: function () { return options; },
                getEmitResolver: function () { return resolver; },
                getEmitHost: function () { return host; },
                getEmitHelperFactory: ts.memoize(function () { return ts.createEmitHelperFactory(context); }),
                startLexicalEnvironment: startLexicalEnvironment,
                suspendLexicalEnvironment: suspendLexicalEnvironment,
                resumeLexicalEnvironment: resumeLexicalEnvironment,
                endLexicalEnvironment: endLexicalEnvironment,
                setLexicalEnvironmentFlags: setLexicalEnvironmentFlags,
                getLexicalEnvironmentFlags: getLexicalEnvironmentFlags,
                hoistVariableDeclaration: hoistVariableDeclaration,
                hoistFunctionDeclaration: hoistFunctionDeclaration,
                addInitializationStatement: addInitializationStatement,
                startBlockScope: startBlockScope,
                endBlockScope: endBlockScope,
                addBlockScopedVariable: addBlockScopedVariable,
                requestEmitHelper: requestEmitHelper,
                readEmitHelpers: readEmitHelpers,
                enableSubstitution: enableSubstitution,
                enableEmitNotification: enableEmitNotification,
                isSubstitutionEnabled: isSubstitutionEnabled,
                isEmitNotificationEnabled: isEmitNotificationEnabled,
                get onSubstituteNode() { return onSubstituteNode; },
                set onSubstituteNode(value) {
                    ts.Debug.assert(state < 1 /* Initialized */, "Cannot modify transformation hooks after initialization has completed.");
                    ts.Debug.assert(value !== undefined, "Value must not be 'undefined'");
                    onSubstituteNode = value;
                },
                get onEmitNode() { return onEmitNode; },
                set onEmitNode(value) {
                    ts.Debug.assert(state < 1 /* Initialized */, "Cannot modify transformation hooks after initialization has completed.");
                    ts.Debug.assert(value !== undefined, "Value must not be 'undefined'");
                    onEmitNode = value;
                },
                addDiagnostic: function (diag) {
                    diagnostics.push(diag);
                }
            };
            // Ensure the parse tree is clean before applying transformations
            for (var _i = 0, nodes_2 = nodes; _i < nodes_2.length; _i++) {
                var node = nodes_2[_i];
                ts.disposeEmitNodes(ts.getSourceFileOfNode(ts.getParseTreeNode(node)));
            }
            ts.performance.mark("beforeTransform");
            // Chain together and initialize each transformer.
            var transformersWithContext = transformers.map(function (t) { return t(context); });
            var transformation = function (node) {
                for (var _i = 0, transformersWithContext_1 = transformersWithContext; _i < transformersWithContext_1.length; _i++) {
                    var transform = transformersWithContext_1[_i];
                    node = transform(node);
                }
                return node;
            };
            // prevent modification of transformation hooks.
            state = 1 /* Initialized */;
            // Transform each node.
            var transformed = [];
            for (var _a = 0, nodes_3 = nodes; _a < nodes_3.length; _a++) {
                var node = nodes_3[_a];
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("emit" /* Emit */, "transformNodes", node.kind === 300 /* SourceFile */ ? { path: node.path } : { kind: node.kind, pos: node.pos, end: node.end });
                transformed.push((allowDtsFiles ? transformation : transformRoot)(node));
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
            }
            // prevent modification of the lexical environment.
            state = 2 /* Completed */;
            ts.performance.mark("afterTransform");
            ts.performance.measure("transformTime", "beforeTransform", "afterTransform");
            return {
                transformed: transformed,
                substituteNode: substituteNode,
                emitNodeWithNotification: emitNodeWithNotification,
                isEmitNotificationEnabled: isEmitNotificationEnabled,
                dispose: dispose,
                diagnostics: diagnostics
            };
            function transformRoot(node) {
                return node && (!ts.isSourceFile(node) || !node.isDeclarationFile) ? transformation(node) : node;
            }
            /**
             * Enables expression substitutions in the pretty printer for the provided SyntaxKind.
             */
            function enableSubstitution(kind) {
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the transformation context after transformation has completed.");
                enabledSyntaxKindFeatures[kind] |= 1 /* Substitution */;
            }
            /**
             * Determines whether expression substitutions are enabled for the provided node.
             */
            function isSubstitutionEnabled(node) {
                return (enabledSyntaxKindFeatures[node.kind] & 1 /* Substitution */) !== 0
                    && (ts.getEmitFlags(node) & 4 /* NoSubstitution */) === 0;
            }
            /**
             * Emits a node with possible substitution.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to emit.
             * @param emitCallback The callback used to emit the node or its substitute.
             */
            function substituteNode(hint, node) {
                ts.Debug.assert(state < 3 /* Disposed */, "Cannot substitute a node after the result is disposed.");
                return node && isSubstitutionEnabled(node) && onSubstituteNode(hint, node) || node;
            }
            /**
             * Enables before/after emit notifications in the pretty printer for the provided SyntaxKind.
             */
            function enableEmitNotification(kind) {
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the transformation context after transformation has completed.");
                enabledSyntaxKindFeatures[kind] |= 2 /* EmitNotifications */;
            }
            /**
             * Determines whether before/after emit notifications should be raised in the pretty
             * printer when it emits a node.
             */
            function isEmitNotificationEnabled(node) {
                return (enabledSyntaxKindFeatures[node.kind] & 2 /* EmitNotifications */) !== 0
                    || (ts.getEmitFlags(node) & 2 /* AdviseOnEmitNode */) !== 0;
            }
            /**
             * Emits a node with possible emit notification.
             *
             * @param hint A hint as to the intended usage of the node.
             * @param node The node to emit.
             * @param emitCallback The callback used to emit the node.
             */
            function emitNodeWithNotification(hint, node, emitCallback) {
                ts.Debug.assert(state < 3 /* Disposed */, "Cannot invoke TransformationResult callbacks after the result is disposed.");
                if (node) {
                    // TODO: Remove check and unconditionally use onEmitNode when API is breakingly changed
                    // (see https://github.com/microsoft/TypeScript/pull/36248/files/5062623f39120171b98870c71344b3242eb03d23#r369766739)
                    if (isEmitNotificationEnabled(node)) {
                        onEmitNode(hint, node, emitCallback);
                    }
                    else {
                        emitCallback(hint, node);
                    }
                }
            }
            /**
             * Records a hoisted variable declaration for the provided name within a lexical environment.
             */
            function hoistVariableDeclaration(name) {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot modify the lexical environment during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the lexical environment after transformation has completed.");
                var decl = ts.setEmitFlags(factory.createVariableDeclaration(name), 64 /* NoNestedSourceMaps */);
                if (!lexicalEnvironmentVariableDeclarations) {
                    lexicalEnvironmentVariableDeclarations = [decl];
                }
                else {
                    lexicalEnvironmentVariableDeclarations.push(decl);
                }
                if (lexicalEnvironmentFlags & 1 /* InParameters */) {
                    lexicalEnvironmentFlags |= 2 /* VariablesHoistedInParameters */;
                }
            }
            /**
             * Records a hoisted function declaration within a lexical environment.
             */
            function hoistFunctionDeclaration(func) {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot modify the lexical environment during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the lexical environment after transformation has completed.");
                ts.setEmitFlags(func, 1048576 /* CustomPrologue */);
                if (!lexicalEnvironmentFunctionDeclarations) {
                    lexicalEnvironmentFunctionDeclarations = [func];
                }
                else {
                    lexicalEnvironmentFunctionDeclarations.push(func);
                }
            }
            /**
             * Adds an initialization statement to the top of the lexical environment.
             */
            function addInitializationStatement(node) {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot modify the lexical environment during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the lexical environment after transformation has completed.");
                ts.setEmitFlags(node, 1048576 /* CustomPrologue */);
                if (!lexicalEnvironmentStatements) {
                    lexicalEnvironmentStatements = [node];
                }
                else {
                    lexicalEnvironmentStatements.push(node);
                }
            }
            /**
             * Starts a new lexical environment. Any existing hoisted variable or function declarations
             * are pushed onto a stack, and the related storage variables are reset.
             */
            function startLexicalEnvironment() {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot modify the lexical environment during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the lexical environment after transformation has completed.");
                ts.Debug.assert(!lexicalEnvironmentSuspended, "Lexical environment is suspended.");
                // Save the current lexical environment. Rather than resizing the array we adjust the
                // stack size variable. This allows us to reuse existing array slots we've
                // already allocated between transformations to avoid allocation and GC overhead during
                // transformation.
                lexicalEnvironmentVariableDeclarationsStack[lexicalEnvironmentStackOffset] = lexicalEnvironmentVariableDeclarations;
                lexicalEnvironmentFunctionDeclarationsStack[lexicalEnvironmentStackOffset] = lexicalEnvironmentFunctionDeclarations;
                lexicalEnvironmentStatementsStack[lexicalEnvironmentStackOffset] = lexicalEnvironmentStatements;
                lexicalEnvironmentFlagsStack[lexicalEnvironmentStackOffset] = lexicalEnvironmentFlags;
                lexicalEnvironmentStackOffset++;
                lexicalEnvironmentVariableDeclarations = undefined;
                lexicalEnvironmentFunctionDeclarations = undefined;
                lexicalEnvironmentStatements = undefined;
                lexicalEnvironmentFlags = 0 /* None */;
            }
            /** Suspends the current lexical environment, usually after visiting a parameter list. */
            function suspendLexicalEnvironment() {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot modify the lexical environment during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the lexical environment after transformation has completed.");
                ts.Debug.assert(!lexicalEnvironmentSuspended, "Lexical environment is already suspended.");
                lexicalEnvironmentSuspended = true;
            }
            /** Resumes a suspended lexical environment, usually before visiting a function body. */
            function resumeLexicalEnvironment() {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot modify the lexical environment during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the lexical environment after transformation has completed.");
                ts.Debug.assert(lexicalEnvironmentSuspended, "Lexical environment is not suspended.");
                lexicalEnvironmentSuspended = false;
            }
            /**
             * Ends a lexical environment. The previous set of hoisted declarations are restored and
             * any hoisted declarations added in this environment are returned.
             */
            function endLexicalEnvironment() {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot modify the lexical environment during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the lexical environment after transformation has completed.");
                ts.Debug.assert(!lexicalEnvironmentSuspended, "Lexical environment is suspended.");
                var statements;
                if (lexicalEnvironmentVariableDeclarations ||
                    lexicalEnvironmentFunctionDeclarations ||
                    lexicalEnvironmentStatements) {
                    if (lexicalEnvironmentFunctionDeclarations) {
                        statements = __spreadArray([], lexicalEnvironmentFunctionDeclarations, true);
                    }
                    if (lexicalEnvironmentVariableDeclarations) {
                        var statement = factory.createVariableStatement(
                        /*modifiers*/ undefined, factory.createVariableDeclarationList(lexicalEnvironmentVariableDeclarations));
                        ts.setEmitFlags(statement, 1048576 /* CustomPrologue */);
                        if (!statements) {
                            statements = [statement];
                        }
                        else {
                            statements.push(statement);
                        }
                    }
                    if (lexicalEnvironmentStatements) {
                        if (!statements) {
                            statements = __spreadArray([], lexicalEnvironmentStatements, true);
                        }
                        else {
                            statements = __spreadArray(__spreadArray([], statements, true), lexicalEnvironmentStatements, true);
                        }
                    }
                }
                // Restore the previous lexical environment.
                lexicalEnvironmentStackOffset--;
                lexicalEnvironmentVariableDeclarations = lexicalEnvironmentVariableDeclarationsStack[lexicalEnvironmentStackOffset];
                lexicalEnvironmentFunctionDeclarations = lexicalEnvironmentFunctionDeclarationsStack[lexicalEnvironmentStackOffset];
                lexicalEnvironmentStatements = lexicalEnvironmentStatementsStack[lexicalEnvironmentStackOffset];
                lexicalEnvironmentFlags = lexicalEnvironmentFlagsStack[lexicalEnvironmentStackOffset];
                if (lexicalEnvironmentStackOffset === 0) {
                    lexicalEnvironmentVariableDeclarationsStack = [];
                    lexicalEnvironmentFunctionDeclarationsStack = [];
                    lexicalEnvironmentStatementsStack = [];
                    lexicalEnvironmentFlagsStack = [];
                }
                return statements;
            }
            function setLexicalEnvironmentFlags(flags, value) {
                lexicalEnvironmentFlags = value ?
                    lexicalEnvironmentFlags | flags :
                    lexicalEnvironmentFlags & ~flags;
            }
            function getLexicalEnvironmentFlags() {
                return lexicalEnvironmentFlags;
            }
            /**
             * Starts a block scope. Any existing block hoisted variables are pushed onto the stack and the related storage variables are reset.
             */
            function startBlockScope() {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot start a block scope during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot start a block scope after transformation has completed.");
                blockScopedVariableDeclarationsStack[blockScopeStackOffset] = blockScopedVariableDeclarations;
                blockScopeStackOffset++;
                blockScopedVariableDeclarations = undefined;
            }
            /**
             * Ends a block scope. The previous set of block hoisted variables are restored. Any hoisted declarations are returned.
             */
            function endBlockScope() {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot end a block scope during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot end a block scope after transformation has completed.");
                var statements = ts.some(blockScopedVariableDeclarations) ?
                    [
                        factory.createVariableStatement(
                        /*modifiers*/ undefined, factory.createVariableDeclarationList(blockScopedVariableDeclarations.map(function (identifier) { return factory.createVariableDeclaration(identifier); }), 1 /* Let */))
                    ] : undefined;
                blockScopeStackOffset--;
                blockScopedVariableDeclarations = blockScopedVariableDeclarationsStack[blockScopeStackOffset];
                if (blockScopeStackOffset === 0) {
                    blockScopedVariableDeclarationsStack = [];
                }
                return statements;
            }
            function addBlockScopedVariable(name) {
                ts.Debug.assert(blockScopeStackOffset > 0, "Cannot add a block scoped variable outside of an iteration body.");
                (blockScopedVariableDeclarations || (blockScopedVariableDeclarations = [])).push(name);
            }
            function requestEmitHelper(helper) {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot modify the transformation context during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the transformation context after transformation has completed.");
                ts.Debug.assert(!helper.scoped, "Cannot request a scoped emit helper.");
                if (helper.dependencies) {
                    for (var _i = 0, _a = helper.dependencies; _i < _a.length; _i++) {
                        var h = _a[_i];
                        requestEmitHelper(h);
                    }
                }
                emitHelpers = ts.append(emitHelpers, helper);
            }
            function readEmitHelpers() {
                ts.Debug.assert(state > 0 /* Uninitialized */, "Cannot modify the transformation context during initialization.");
                ts.Debug.assert(state < 2 /* Completed */, "Cannot modify the transformation context after transformation has completed.");
                var helpers = emitHelpers;
                emitHelpers = undefined;
                return helpers;
            }
            function dispose() {
                if (state < 3 /* Disposed */) {
                    // Clean up emit nodes on parse tree
                    for (var _i = 0, nodes_4 = nodes; _i < nodes_4.length; _i++) {
                        var node = nodes_4[_i];
                        ts.disposeEmitNodes(ts.getSourceFileOfNode(ts.getParseTreeNode(node)));
                    }
                    // Release references to external entries for GC purposes.
                    lexicalEnvironmentVariableDeclarations = undefined;
                    lexicalEnvironmentVariableDeclarationsStack = undefined;
                    lexicalEnvironmentFunctionDeclarations = undefined;
                    lexicalEnvironmentFunctionDeclarationsStack = undefined;
                    onSubstituteNode = undefined;
                    onEmitNode = undefined;
                    emitHelpers = undefined;
                    // Prevent further use of the transformation result.
                    state = 3 /* Disposed */;
                }
            }
        }
        ts.transformNodes = transformNodes;
        ts.nullTransformationContext = {
            factory: ts.factory,
            getCompilerOptions: function () { return ({}); },
            getEmitResolver: ts.notImplemented,
            getEmitHost: ts.notImplemented,
            getEmitHelperFactory: ts.notImplemented,
            startLexicalEnvironment: ts.noop,
            resumeLexicalEnvironment: ts.noop,
            suspendLexicalEnvironment: ts.noop,
            endLexicalEnvironment: ts.returnUndefined,
            setLexicalEnvironmentFlags: ts.noop,
            getLexicalEnvironmentFlags: function () { return 0; },
            hoistVariableDeclaration: ts.noop,
            hoistFunctionDeclaration: ts.noop,
            addInitializationStatement: ts.noop,
            startBlockScope: ts.noop,
            endBlockScope: ts.returnUndefined,
            addBlockScopedVariable: ts.noop,
            requestEmitHelper: ts.noop,
            readEmitHelpers: ts.notImplemented,
            enableSubstitution: ts.noop,
            enableEmitNotification: ts.noop,
            isSubstitutionEnabled: ts.notImplemented,
            isEmitNotificationEnabled: ts.notImplemented,
            onSubstituteNode: noEmitSubstitution,
            onEmitNode: noEmitNotification,
            addDiagnostic: ts.noop,
        };
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        var brackets = createBracketsMap();
        /*@internal*/
        function isBuildInfoFile(file) {
            return ts.fileExtensionIs(file, ".tsbuildinfo" /* TsBuildInfo */);
        }
        ts.isBuildInfoFile = isBuildInfoFile;
        /*@internal*/
        /**
         * Iterates over the source files that are expected to have an emit output.
         *
         * @param host An EmitHost.
         * @param action The action to execute.
         * @param sourceFilesOrTargetSourceFile
         *   If an array, the full list of source files to emit.
         *   Else, calls `getSourceFilesToEmit` with the (optional) target source file to determine the list of source files to emit.
         */
        function forEachEmittedFile(host, action, sourceFilesOrTargetSourceFile, forceDtsEmit, onlyBuildInfo, includeBuildInfo) {
            if (forceDtsEmit === void 0) {
                forceDtsEmit = false;
            }
            var sourceFiles = ts.isArray(sourceFilesOrTargetSourceFile) ? sourceFilesOrTargetSourceFile : ts.getSourceFilesToEmit(host, sourceFilesOrTargetSourceFile, forceDtsEmit);
            var options = host.getCompilerOptions();
            if (ts.outFile(options)) {
                var prepends = host.getPrependNodes();
                if (sourceFiles.length || prepends.length) {
                    var bundle = ts.factory.createBundle(sourceFiles, prepends);
                    var result = action(getOutputPathsFor(bundle, host, forceDtsEmit), bundle);
                    if (result) {
                        return result;
                    }
                }
            }
            else {
                if (!onlyBuildInfo) {
                    for (var _a = 0, sourceFiles_1 = sourceFiles; _a < sourceFiles_1.length; _a++) {
                        var sourceFile = sourceFiles_1[_a];
                        var result = action(getOutputPathsFor(sourceFile, host, forceDtsEmit), sourceFile);
                        if (result) {
                            return result;
                        }
                    }
                }
                if (includeBuildInfo) {
                    var buildInfoPath = getTsBuildInfoEmitOutputFilePath(options);
                    if (buildInfoPath)
                        return action({ buildInfoPath: buildInfoPath }, /*sourceFileOrBundle*/ undefined);
                }
            }
        }
        ts.forEachEmittedFile = forEachEmittedFile;
        function getTsBuildInfoEmitOutputFilePath(options) {
            var configFile = options.configFilePath;
            if (!ts.isIncrementalCompilation(options))
                return undefined;
            if (options.tsBuildInfoFile)
                return options.tsBuildInfoFile;
            var outPath = ts.outFile(options);
            var buildInfoExtensionLess;
            if (outPath) {
                buildInfoExtensionLess = ts.removeFileExtension(outPath);
            }
            else {
                if (!configFile)
                    return undefined;
                var configFileExtensionLess = ts.removeFileExtension(configFile);
                buildInfoExtensionLess = options.outDir ?
                    options.rootDir ?
                        ts.resolvePath(options.outDir, ts.getRelativePathFromDirectory(options.rootDir, configFileExtensionLess, /*ignoreCase*/ true)) :
                        ts.combinePaths(options.outDir, ts.getBaseFileName(configFileExtensionLess)) :
                    configFileExtensionLess;
            }
            return buildInfoExtensionLess + ".tsbuildinfo" /* TsBuildInfo */;
        }
        ts.getTsBuildInfoEmitOutputFilePath = getTsBuildInfoEmitOutputFilePath;
        /*@internal*/
        function getOutputPathsForBundle(options, forceDtsPaths) {
            var outPath = ts.outFile(options);
            var jsFilePath = options.emitDeclarationOnly ? undefined : outPath;
            var sourceMapFilePath = jsFilePath && getSourceMapFilePath(jsFilePath, options);
            var declarationFilePath = (forceDtsPaths || ts.getEmitDeclarations(options)) ? ts.removeFileExtension(outPath) + ".d.ts" /* Dts */ : undefined;
            var declarationMapPath = declarationFilePath && ts.getAreDeclarationMapsEnabled(options) ? declarationFilePath + ".map" : undefined;
            var buildInfoPath = getTsBuildInfoEmitOutputFilePath(options);
            return { jsFilePath: jsFilePath, sourceMapFilePath: sourceMapFilePath, declarationFilePath: declarationFilePath, declarationMapPath: declarationMapPath, buildInfoPath: buildInfoPath };
        }
        ts.getOutputPathsForBundle = getOutputPathsForBundle;
        /*@internal*/
        function getOutputPathsFor(sourceFile, host, forceDtsPaths) {
            var options = host.getCompilerOptions();
            if (sourceFile.kind === 301 /* Bundle */) {
                return getOutputPathsForBundle(options, forceDtsPaths);
            }
            else {
                var ownOutputFilePath = ts.getOwnEmitOutputFilePath(sourceFile.fileName, host, getOutputExtension(sourceFile, options));
                var isJsonFile = ts.isJsonSourceFile(sourceFile);
                // If json file emits to the same location skip writing it, if emitDeclarationOnly skip writing it
                var isJsonEmittedToSameLocation = isJsonFile &&
                    ts.comparePaths(sourceFile.fileName, ownOutputFilePath, host.getCurrentDirectory(), !host.useCaseSensitiveFileNames()) === 0 /* EqualTo */;
                var jsFilePath = options.emitDeclarationOnly || isJsonEmittedToSameLocation ? undefined : ownOutputFilePath;
                var sourceMapFilePath = !jsFilePath || ts.isJsonSourceFile(sourceFile) ? undefined : getSourceMapFilePath(jsFilePath, options);
                var declarationFilePath = (forceDtsPaths || (ts.getEmitDeclarations(options) && !isJsonFile)) ? ts.getDeclarationEmitOutputFilePath(sourceFile.fileName, host) : undefined;
                var declarationMapPath = declarationFilePath && ts.getAreDeclarationMapsEnabled(options) ? declarationFilePath + ".map" : undefined;
                return { jsFilePath: jsFilePath, sourceMapFilePath: sourceMapFilePath, declarationFilePath: declarationFilePath, declarationMapPath: declarationMapPath, buildInfoPath: undefined };
            }
        }
        ts.getOutputPathsFor = getOutputPathsFor;
        function getSourceMapFilePath(jsFilePath, options) {
            return (options.sourceMap && !options.inlineSourceMap) ? jsFilePath + ".map" : undefined;
        }
        // JavaScript files are always LanguageVariant.JSX, as JSX syntax is allowed in .js files also.
        // So for JavaScript files, '.jsx' is only emitted if the input was '.jsx', and JsxEmit.Preserve.
        // For TypeScript, the only time to emit with a '.jsx' extension, is on JSX input, and JsxEmit.Preserve
        /* @internal */
        function getOutputExtension(sourceFile, options) {
            if (ts.isJsonSourceFile(sourceFile)) {
                return ".json" /* Json */;
            }
            if (options.jsx === 1 /* Preserve */) {
                if (ts.isSourceFileJS(sourceFile)) {
                    if (ts.fileExtensionIs(sourceFile.fileName, ".jsx" /* Jsx */)) {
                        return ".jsx" /* Jsx */;
                    }
                }
                else if (sourceFile.languageVariant === 1 /* JSX */) {
                    // TypeScript source file preserving JSX syntax
                    return ".jsx" /* Jsx */;
                }
            }
            return ".js" /* Js */;
        }
        ts.getOutputExtension = getOutputExtension;
        function getOutputPathWithoutChangingExt(inputFileName, configFile, ignoreCase, outputDir, getCommonSourceDirectory) {
            return outputDir ?
                ts.resolvePath(outputDir, ts.getRelativePathFromDirectory(getCommonSourceDirectory ? getCommonSourceDirectory() : getCommonSourceDirectoryOfConfig(configFile, ignoreCase), inputFileName, ignoreCase)) :
                inputFileName;
        }
        /* @internal */
        function getOutputDeclarationFileName(inputFileName, configFile, ignoreCase, getCommonSourceDirectory) {
            ts.Debug.assert(!ts.fileExtensionIs(inputFileName, ".d.ts" /* Dts */) && !ts.fileExtensionIs(inputFileName, ".json" /* Json */));
            return ts.changeExtension(getOutputPathWithoutChangingExt(inputFileName, configFile, ignoreCase, configFile.options.declarationDir || configFile.options.outDir, getCommonSourceDirectory), ".d.ts" /* Dts */);
        }
        ts.getOutputDeclarationFileName = getOutputDeclarationFileName;
        function getOutputJSFileName(inputFileName, configFile, ignoreCase, getCommonSourceDirectory) {
            if (configFile.options.emitDeclarationOnly)
                return undefined;
            var isJsonFile = ts.fileExtensionIs(inputFileName, ".json" /* Json */);
            var outputFileName = ts.changeExtension(getOutputPathWithoutChangingExt(inputFileName, configFile, ignoreCase, configFile.options.outDir, getCommonSourceDirectory), isJsonFile ?
                ".json" /* Json */ :
                configFile.options.jsx === 1 /* Preserve */ && (ts.fileExtensionIs(inputFileName, ".tsx" /* Tsx */) || ts.fileExtensionIs(inputFileName, ".jsx" /* Jsx */)) ?
                    ".jsx" /* Jsx */ :
                    ".js" /* Js */);
            return !isJsonFile || ts.comparePaths(inputFileName, outputFileName, ts.Debug.checkDefined(configFile.options.configFilePath), ignoreCase) !== 0 /* EqualTo */ ?
                outputFileName :
                undefined;
        }
        function createAddOutput() {
            var outputs;
            return { addOutput: addOutput, getOutputs: getOutputs };
            function addOutput(path) {
                if (path) {
                    (outputs || (outputs = [])).push(path);
                }
            }
            function getOutputs() {
                return outputs || ts.emptyArray;
            }
        }
        function getSingleOutputFileNames(configFile, addOutput) {
            var _a = getOutputPathsForBundle(configFile.options, /*forceDtsPaths*/ false), jsFilePath = _a.jsFilePath, sourceMapFilePath = _a.sourceMapFilePath, declarationFilePath = _a.declarationFilePath, declarationMapPath = _a.declarationMapPath, buildInfoPath = _a.buildInfoPath;
            addOutput(jsFilePath);
            addOutput(sourceMapFilePath);
            addOutput(declarationFilePath);
            addOutput(declarationMapPath);
            addOutput(buildInfoPath);
        }
        function getOwnOutputFileNames(configFile, inputFileName, ignoreCase, addOutput, getCommonSourceDirectory) {
            if (ts.fileExtensionIs(inputFileName, ".d.ts" /* Dts */))
                return;
            var js = getOutputJSFileName(inputFileName, configFile, ignoreCase, getCommonSourceDirectory);
            addOutput(js);
            if (ts.fileExtensionIs(inputFileName, ".json" /* Json */))
                return;
            if (js && configFile.options.sourceMap) {
                addOutput(js + ".map");
            }
            if (ts.getEmitDeclarations(configFile.options)) {
                var dts = getOutputDeclarationFileName(inputFileName, configFile, ignoreCase, getCommonSourceDirectory);
                addOutput(dts);
                if (configFile.options.declarationMap) {
                    addOutput(dts + ".map");
                }
            }
        }
        /*@internal*/
        function getCommonSourceDirectory(options, emittedFiles, currentDirectory, getCanonicalFileName, checkSourceFilesBelongToPath) {
            var commonSourceDirectory;
            if (options.rootDir) {
                // If a rootDir is specified use it as the commonSourceDirectory
                commonSourceDirectory = ts.getNormalizedAbsolutePath(options.rootDir, currentDirectory);
                checkSourceFilesBelongToPath === null || checkSourceFilesBelongToPath === void 0 ? void 0 : checkSourceFilesBelongToPath(options.rootDir);
            }
            else if (options.composite && options.configFilePath) {
                // Project compilations never infer their root from the input source paths
                commonSourceDirectory = ts.getDirectoryPath(ts.normalizeSlashes(options.configFilePath));
                checkSourceFilesBelongToPath === null || checkSourceFilesBelongToPath === void 0 ? void 0 : checkSourceFilesBelongToPath(commonSourceDirectory);
            }
            else {
                commonSourceDirectory = ts.computeCommonSourceDirectoryOfFilenames(emittedFiles(), currentDirectory, getCanonicalFileName);
            }
            if (commonSourceDirectory && commonSourceDirectory[commonSourceDirectory.length - 1] !== ts.directorySeparator) {
                // Make sure directory path ends with directory separator so this string can directly
                // used to replace with "" to get the relative path of the source file and the relative path doesn't
                // start with / making it rooted path
                commonSourceDirectory += ts.directorySeparator;
            }
            return commonSourceDirectory;
        }
        ts.getCommonSourceDirectory = getCommonSourceDirectory;
        /*@internal*/
        function getCommonSourceDirectoryOfConfig(_a, ignoreCase) {
            var options = _a.options, fileNames = _a.fileNames;
            return getCommonSourceDirectory(options, function () { return ts.filter(fileNames, function (file) { return !(options.noEmitForJsFiles && ts.fileExtensionIsOneOf(file, ts.supportedJSExtensions)) && !ts.fileExtensionIs(file, ".d.ts" /* Dts */); }); }, ts.getDirectoryPath(ts.normalizeSlashes(ts.Debug.checkDefined(options.configFilePath))), ts.createGetCanonicalFileName(!ignoreCase));
        }
        ts.getCommonSourceDirectoryOfConfig = getCommonSourceDirectoryOfConfig;
        /*@internal*/
        function getAllProjectOutputs(configFile, ignoreCase) {
            var _a = createAddOutput(), addOutput = _a.addOutput, getOutputs = _a.getOutputs;
            if (ts.outFile(configFile.options)) {
                getSingleOutputFileNames(configFile, addOutput);
            }
            else {
                var getCommonSourceDirectory_1 = ts.memoize(function () { return getCommonSourceDirectoryOfConfig(configFile, ignoreCase); });
                for (var _b = 0, _c = configFile.fileNames; _b < _c.length; _b++) {
                    var inputFileName = _c[_b];
                    getOwnOutputFileNames(configFile, inputFileName, ignoreCase, addOutput, getCommonSourceDirectory_1);
                }
                addOutput(getTsBuildInfoEmitOutputFilePath(configFile.options));
            }
            return getOutputs();
        }
        ts.getAllProjectOutputs = getAllProjectOutputs;
        function getOutputFileNames(commandLine, inputFileName, ignoreCase) {
            inputFileName = ts.normalizePath(inputFileName);
            ts.Debug.assert(ts.contains(commandLine.fileNames, inputFileName), "Expected fileName to be present in command line");
            var _a = createAddOutput(), addOutput = _a.addOutput, getOutputs = _a.getOutputs;
            if (ts.outFile(commandLine.options)) {
                getSingleOutputFileNames(commandLine, addOutput);
            }
            else {
                getOwnOutputFileNames(commandLine, inputFileName, ignoreCase, addOutput);
            }
            return getOutputs();
        }
        ts.getOutputFileNames = getOutputFileNames;
        /*@internal*/
        function getFirstProjectOutput(configFile, ignoreCase) {
            if (ts.outFile(configFile.options)) {
                var jsFilePath = getOutputPathsForBundle(configFile.options, /*forceDtsPaths*/ false).jsFilePath;
                return ts.Debug.checkDefined(jsFilePath, "project " + configFile.options.configFilePath + " expected to have at least one output");
            }
            var getCommonSourceDirectory = ts.memoize(function () { return getCommonSourceDirectoryOfConfig(configFile, ignoreCase); });
            for (var _a = 0, _b = configFile.fileNames; _a < _b.length; _a++) {
                var inputFileName = _b[_a];
                if (ts.fileExtensionIs(inputFileName, ".d.ts" /* Dts */))
                    continue;
                var jsFilePath = getOutputJSFileName(inputFileName, configFile, ignoreCase, getCommonSourceDirectory);
                if (jsFilePath)
                    return jsFilePath;
                if (ts.fileExtensionIs(inputFileName, ".json" /* Json */))
                    continue;
                if (ts.getEmitDeclarations(configFile.options)) {
                    return getOutputDeclarationFileName(inputFileName, configFile, ignoreCase, getCommonSourceDirectory);
                }
            }
            var buildInfoPath = getTsBuildInfoEmitOutputFilePath(configFile.options);
            if (buildInfoPath)
                return buildInfoPath;
            return ts.Debug.fail("project " + configFile.options.configFilePath + " expected to have at least one output");
        }
        ts.getFirstProjectOutput = getFirstProjectOutput;
        /*@internal*/
        // targetSourceFile is when users only want one file in entire project to be emitted. This is used in compileOnSave feature
        function emitFiles(resolver, host, targetSourceFile, _a, emitOnlyDtsFiles, onlyBuildInfo, forceDtsEmit) {
            var scriptTransformers = _a.scriptTransformers, declarationTransformers = _a.declarationTransformers;
            var compilerOptions = host.getCompilerOptions();
            var sourceMapDataList = (compilerOptions.sourceMap || compilerOptions.inlineSourceMap || ts.getAreDeclarationMapsEnabled(compilerOptions)) ? [] : undefined;
            var emittedFilesList = compilerOptions.listEmittedFiles ? [] : undefined;
            var emitterDiagnostics = ts.createDiagnosticCollection();
            var newLine = ts.getNewLineCharacter(compilerOptions, function () { return host.getNewLine(); });
            var writer = ts.createTextWriter(newLine);
            var _b = ts.performance.createTimer("printTime", "beforePrint", "afterPrint"), enter = _b.enter, exit = _b.exit;
            var bundleBuildInfo;
            var emitSkipped = false;
            var exportedModulesFromDeclarationEmit;
            // Emit each output file
            enter();
            forEachEmittedFile(host, emitSourceFileOrBundle, ts.getSourceFilesToEmit(host, targetSourceFile, forceDtsEmit), forceDtsEmit, onlyBuildInfo, !targetSourceFile);
            exit();
            return {
                emitSkipped: emitSkipped,
                diagnostics: emitterDiagnostics.getDiagnostics(),
                emittedFiles: emittedFilesList,
                sourceMaps: sourceMapDataList,
                exportedModulesFromDeclarationEmit: exportedModulesFromDeclarationEmit
            };
            function emitSourceFileOrBundle(_a, sourceFileOrBundle) {
                var jsFilePath = _a.jsFilePath, sourceMapFilePath = _a.sourceMapFilePath, declarationFilePath = _a.declarationFilePath, declarationMapPath = _a.declarationMapPath, buildInfoPath = _a.buildInfoPath;
                var buildInfoDirectory;
                if (buildInfoPath && sourceFileOrBundle && ts.isBundle(sourceFileOrBundle)) {
                    buildInfoDirectory = ts.getDirectoryPath(ts.getNormalizedAbsolutePath(buildInfoPath, host.getCurrentDirectory()));
                    bundleBuildInfo = {
                        commonSourceDirectory: relativeToBuildInfo(host.getCommonSourceDirectory()),
                        sourceFiles: sourceFileOrBundle.sourceFiles.map(function (file) { return relativeToBuildInfo(ts.getNormalizedAbsolutePath(file.fileName, host.getCurrentDirectory())); })
                    };
                }
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("emit" /* Emit */, "emitJsFileOrBundle", { jsFilePath: jsFilePath });
                emitJsFileOrBundle(sourceFileOrBundle, jsFilePath, sourceMapFilePath, relativeToBuildInfo);
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("emit" /* Emit */, "emitDeclarationFileOrBundle", { declarationFilePath: declarationFilePath });
                emitDeclarationFileOrBundle(sourceFileOrBundle, declarationFilePath, declarationMapPath, relativeToBuildInfo);
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("emit" /* Emit */, "emitBuildInfo", { buildInfoPath: buildInfoPath });
                emitBuildInfo(bundleBuildInfo, buildInfoPath);
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                if (!emitSkipped && emittedFilesList) {
                    if (!emitOnlyDtsFiles) {
                        if (jsFilePath) {
                            emittedFilesList.push(jsFilePath);
                        }
                        if (sourceMapFilePath) {
                            emittedFilesList.push(sourceMapFilePath);
                        }
                        if (buildInfoPath) {
                            emittedFilesList.push(buildInfoPath);
                        }
                    }
                    if (declarationFilePath) {
                        emittedFilesList.push(declarationFilePath);
                    }
                    if (declarationMapPath) {
                        emittedFilesList.push(declarationMapPath);
                    }
                }
                function relativeToBuildInfo(path) {
                    return ts.ensurePathIsNonModuleName(ts.getRelativePathFromDirectory(buildInfoDirectory, path, host.getCanonicalFileName));
                }
            }
            function emitBuildInfo(bundle, buildInfoPath) {
                // Write build information if applicable
                if (!buildInfoPath || targetSourceFile || emitSkipped)
                    return;
                var program = host.getProgramBuildInfo();
                if (host.isEmitBlocked(buildInfoPath)) {
                    emitSkipped = true;
                    return;
                }
                var version = ts.version; // Extracted into a const so the form is stable between namespace and module
                ts.writeFile(host, emitterDiagnostics, buildInfoPath, getBuildInfoText({ bundle: bundle, program: program, version: version }), /*writeByteOrderMark*/ false);
            }
            function emitJsFileOrBundle(sourceFileOrBundle, jsFilePath, sourceMapFilePath, relativeToBuildInfo) {
                if (!sourceFileOrBundle || emitOnlyDtsFiles || !jsFilePath) {
                    return;
                }
                // Make sure not to write js file and source map file if any of them cannot be written
                if ((jsFilePath && host.isEmitBlocked(jsFilePath)) || compilerOptions.noEmit) {
                    emitSkipped = true;
                    return;
                }
                // Transform the source files
                var transform = ts.transformNodes(resolver, host, ts.factory, compilerOptions, [sourceFileOrBundle], scriptTransformers, /*allowDtsFiles*/ false);
                var printerOptions = {
                    removeComments: compilerOptions.removeComments,
                    newLine: compilerOptions.newLine,
                    noEmitHelpers: compilerOptions.noEmitHelpers,
                    module: compilerOptions.module,
                    target: compilerOptions.target,
                    sourceMap: compilerOptions.sourceMap,
                    inlineSourceMap: compilerOptions.inlineSourceMap,
                    inlineSources: compilerOptions.inlineSources,
                    extendedDiagnostics: compilerOptions.extendedDiagnostics,
                    writeBundleFileInfo: !!bundleBuildInfo,
                    relativeToBuildInfo: relativeToBuildInfo
                };
                // Create a printer to print the nodes
                var printer = createPrinter(printerOptions, {
                    // resolver hooks
                    hasGlobalName: resolver.hasGlobalName,
                    // transform hooks
                    onEmitNode: transform.emitNodeWithNotification,
                    isEmitNotificationEnabled: transform.isEmitNotificationEnabled,
                    substituteNode: transform.substituteNode,
                });
                ts.Debug.assert(transform.transformed.length === 1, "Should only see one output from the transform");
                printSourceFileOrBundle(jsFilePath, sourceMapFilePath, transform.transformed[0], printer, compilerOptions);
                // Clean up emit nodes on parse tree
                transform.dispose();
                if (bundleBuildInfo)
                    bundleBuildInfo.js = printer.bundleFileInfo;
            }
            function emitDeclarationFileOrBundle(sourceFileOrBundle, declarationFilePath, declarationMapPath, relativeToBuildInfo) {
                if (!sourceFileOrBundle)
                    return;
                if (!declarationFilePath) {
                    if (emitOnlyDtsFiles || compilerOptions.emitDeclarationOnly)
                        emitSkipped = true;
                    return;
                }
                var sourceFiles = ts.isSourceFile(sourceFileOrBundle) ? [sourceFileOrBundle] : sourceFileOrBundle.sourceFiles;
                var filesForEmit = forceDtsEmit ? sourceFiles : ts.filter(sourceFiles, ts.isSourceFileNotJson);
                // Setup and perform the transformation to retrieve declarations from the input files
                var inputListOrBundle = ts.outFile(compilerOptions) ? [ts.factory.createBundle(filesForEmit, !ts.isSourceFile(sourceFileOrBundle) ? sourceFileOrBundle.prepends : undefined)] : filesForEmit;
                if (emitOnlyDtsFiles && !ts.getEmitDeclarations(compilerOptions)) {
                    // Checker wont collect the linked aliases since thats only done when declaration is enabled.
                    // Do that here when emitting only dts files
                    filesForEmit.forEach(collectLinkedAliases);
                }
                var declarationTransform = ts.transformNodes(resolver, host, ts.factory, compilerOptions, inputListOrBundle, declarationTransformers, /*allowDtsFiles*/ false);
                if (ts.length(declarationTransform.diagnostics)) {
                    for (var _a = 0, _b = declarationTransform.diagnostics; _a < _b.length; _a++) {
                        var diagnostic = _b[_a];
                        emitterDiagnostics.add(diagnostic);
                    }
                }
                var printerOptions = {
                    removeComments: compilerOptions.removeComments,
                    newLine: compilerOptions.newLine,
                    noEmitHelpers: true,
                    module: compilerOptions.module,
                    target: compilerOptions.target,
                    sourceMap: compilerOptions.sourceMap,
                    inlineSourceMap: compilerOptions.inlineSourceMap,
                    extendedDiagnostics: compilerOptions.extendedDiagnostics,
                    onlyPrintJsDocStyle: true,
                    writeBundleFileInfo: !!bundleBuildInfo,
                    recordInternalSection: !!bundleBuildInfo,
                    relativeToBuildInfo: relativeToBuildInfo
                };
                var declarationPrinter = createPrinter(printerOptions, {
                    // resolver hooks
                    hasGlobalName: resolver.hasGlobalName,
                    // transform hooks
                    onEmitNode: declarationTransform.emitNodeWithNotification,
                    isEmitNotificationEnabled: declarationTransform.isEmitNotificationEnabled,
                    substituteNode: declarationTransform.substituteNode,
                });
                var declBlocked = (!!declarationTransform.diagnostics && !!declarationTransform.diagnostics.length) || !!host.isEmitBlocked(declarationFilePath) || !!compilerOptions.noEmit;
                emitSkipped = emitSkipped || declBlocked;
                if (!declBlocked || forceDtsEmit) {
                    ts.Debug.assert(declarationTransform.transformed.length === 1, "Should only see one output from the decl transform");
                    printSourceFileOrBundle(declarationFilePath, declarationMapPath, declarationTransform.transformed[0], declarationPrinter, {
                        sourceMap: !forceDtsEmit && compilerOptions.declarationMap,
                        sourceRoot: compilerOptions.sourceRoot,
                        mapRoot: compilerOptions.mapRoot,
                        extendedDiagnostics: compilerOptions.extendedDiagnostics,
                        // Explicitly do not passthru either `inline` option
                    });
                    if (forceDtsEmit && declarationTransform.transformed[0].kind === 300 /* SourceFile */) {
                        var sourceFile = declarationTransform.transformed[0];
                        exportedModulesFromDeclarationEmit = sourceFile.exportedModulesFromDeclarationEmit;
                    }
                }
                declarationTransform.dispose();
                if (bundleBuildInfo)
                    bundleBuildInfo.dts = declarationPrinter.bundleFileInfo;
            }
            function collectLinkedAliases(node) {
                if (ts.isExportAssignment(node)) {
                    if (node.expression.kind === 79 /* Identifier */) {
                        resolver.collectLinkedAliases(node.expression, /*setVisibility*/ true);
                    }
                    return;
                }
                else if (ts.isExportSpecifier(node)) {
                    resolver.collectLinkedAliases(node.propertyName || node.name, /*setVisibility*/ true);
                    return;
                }
                ts.forEachChild(node, collectLinkedAliases);
            }
            function printSourceFileOrBundle(jsFilePath, sourceMapFilePath, sourceFileOrBundle, printer, mapOptions) {
                var bundle = sourceFileOrBundle.kind === 301 /* Bundle */ ? sourceFileOrBundle : undefined;
                var sourceFile = sourceFileOrBundle.kind === 300 /* SourceFile */ ? sourceFileOrBundle : undefined;
                var sourceFiles = bundle ? bundle.sourceFiles : [sourceFile];
                var sourceMapGenerator;
                if (shouldEmitSourceMaps(mapOptions, sourceFileOrBundle)) {
                    sourceMapGenerator = ts.createSourceMapGenerator(host, ts.getBaseFileName(ts.normalizeSlashes(jsFilePath)), getSourceRoot(mapOptions), getSourceMapDirectory(mapOptions, jsFilePath, sourceFile), mapOptions);
                }
                if (bundle) {
                    printer.writeBundle(bundle, writer, sourceMapGenerator);
                }
                else {
                    printer.writeFile(sourceFile, writer, sourceMapGenerator);
                }
                if (sourceMapGenerator) {
                    if (sourceMapDataList) {
                        sourceMapDataList.push({
                            inputSourceFileNames: sourceMapGenerator.getSources(),
                            sourceMap: sourceMapGenerator.toJSON()
                        });
                    }
                    var sourceMappingURL = getSourceMappingURL(mapOptions, sourceMapGenerator, jsFilePath, sourceMapFilePath, sourceFile);
                    if (sourceMappingURL) {
                        if (!writer.isAtStartOfLine())
                            writer.rawWrite(newLine);
                        writer.writeComment("//# " + "sourceMappingURL" + "=" + sourceMappingURL); // Tools can sometimes see this line as a source mapping url comment
                    }
                    // Write the source map
                    if (sourceMapFilePath) {
                        var sourceMap = sourceMapGenerator.toString();
                        ts.writeFile(host, emitterDiagnostics, sourceMapFilePath, sourceMap, /*writeByteOrderMark*/ false, sourceFiles);
                    }
                }
                else {
                    writer.writeLine();
                }
                // Write the output file
                ts.writeFile(host, emitterDiagnostics, jsFilePath, writer.getText(), !!compilerOptions.emitBOM, sourceFiles);
                // Reset state
                writer.clear();
            }
            function shouldEmitSourceMaps(mapOptions, sourceFileOrBundle) {
                return (mapOptions.sourceMap || mapOptions.inlineSourceMap)
                    && (sourceFileOrBundle.kind !== 300 /* SourceFile */ || !ts.fileExtensionIs(sourceFileOrBundle.fileName, ".json" /* Json */));
            }
            function getSourceRoot(mapOptions) {
                // Normalize source root and make sure it has trailing "/" so that it can be used to combine paths with the
                // relative paths of the sources list in the sourcemap
                var sourceRoot = ts.normalizeSlashes(mapOptions.sourceRoot || "");
                return sourceRoot ? ts.ensureTrailingDirectorySeparator(sourceRoot) : sourceRoot;
            }
            function getSourceMapDirectory(mapOptions, filePath, sourceFile) {
                if (mapOptions.sourceRoot)
                    return host.getCommonSourceDirectory();
                if (mapOptions.mapRoot) {
                    var sourceMapDir = ts.normalizeSlashes(mapOptions.mapRoot);
                    if (sourceFile) {
                        // For modules or multiple emit files the mapRoot will have directory structure like the sources
                        // So if src\a.ts and src\lib\b.ts are compiled together user would be moving the maps into mapRoot\a.js.map and mapRoot\lib\b.js.map
                        sourceMapDir = ts.getDirectoryPath(ts.getSourceFilePathInNewDir(sourceFile.fileName, host, sourceMapDir));
                    }
                    if (ts.getRootLength(sourceMapDir) === 0) {
                        // The relative paths are relative to the common directory
                        sourceMapDir = ts.combinePaths(host.getCommonSourceDirectory(), sourceMapDir);
                    }
                    return sourceMapDir;
                }
                return ts.getDirectoryPath(ts.normalizePath(filePath));
            }
            function getSourceMappingURL(mapOptions, sourceMapGenerator, filePath, sourceMapFilePath, sourceFile) {
                if (mapOptions.inlineSourceMap) {
                    // Encode the sourceMap into the sourceMap url
                    var sourceMapText = sourceMapGenerator.toString();
                    var base64SourceMapText = ts.base64encode(ts.sys, sourceMapText);
                    return "data:application/json;base64," + base64SourceMapText;
                }
                var sourceMapFile = ts.getBaseFileName(ts.normalizeSlashes(ts.Debug.checkDefined(sourceMapFilePath)));
                if (mapOptions.mapRoot) {
                    var sourceMapDir = ts.normalizeSlashes(mapOptions.mapRoot);
                    if (sourceFile) {
                        // For modules or multiple emit files the mapRoot will have directory structure like the sources
                        // So if src\a.ts and src\lib\b.ts are compiled together user would be moving the maps into mapRoot\a.js.map and mapRoot\lib\b.js.map
                        sourceMapDir = ts.getDirectoryPath(ts.getSourceFilePathInNewDir(sourceFile.fileName, host, sourceMapDir));
                    }
                    if (ts.getRootLength(sourceMapDir) === 0) {
                        // The relative paths are relative to the common directory
                        sourceMapDir = ts.combinePaths(host.getCommonSourceDirectory(), sourceMapDir);
                        return encodeURI(ts.getRelativePathToDirectoryOrUrl(ts.getDirectoryPath(ts.normalizePath(filePath)), // get the relative sourceMapDir path based on jsFilePath
                        ts.combinePaths(sourceMapDir, sourceMapFile), // this is where user expects to see sourceMap
                        host.getCurrentDirectory(), host.getCanonicalFileName, 
                        /*isAbsolutePathAnUrl*/ true));
                    }
                    else {
                        return encodeURI(ts.combinePaths(sourceMapDir, sourceMapFile));
                    }
                }
                return encodeURI(sourceMapFile);
            }
        }
        ts.emitFiles = emitFiles;
        /*@internal*/
        function getBuildInfoText(buildInfo) {
            return JSON.stringify(buildInfo);
        }
        ts.getBuildInfoText = getBuildInfoText;
        /*@internal*/
        function getBuildInfo(buildInfoText) {
            return JSON.parse(buildInfoText);
        }
        ts.getBuildInfo = getBuildInfo;
        /*@internal*/
        ts.notImplementedResolver = {
            hasGlobalName: ts.notImplemented,
            getReferencedExportContainer: ts.notImplemented,
            getReferencedImportDeclaration: ts.notImplemented,
            getReferencedDeclarationWithCollidingName: ts.notImplemented,
            isDeclarationWithCollidingName: ts.notImplemented,
            isValueAliasDeclaration: ts.notImplemented,
            isReferencedAliasDeclaration: ts.notImplemented,
            isTopLevelValueImportEqualsWithEntityName: ts.notImplemented,
            getNodeCheckFlags: ts.notImplemented,
            isDeclarationVisible: ts.notImplemented,
            isLateBound: function (_node) { return false; },
            collectLinkedAliases: ts.notImplemented,
            isImplementationOfOverload: ts.notImplemented,
            isRequiredInitializedParameter: ts.notImplemented,
            isOptionalUninitializedParameterProperty: ts.notImplemented,
            isExpandoFunctionDeclaration: ts.notImplemented,
            getPropertiesOfContainerFunction: ts.notImplemented,
            createTypeOfDeclaration: ts.notImplemented,
            createReturnTypeOfSignatureDeclaration: ts.notImplemented,
            createTypeOfExpression: ts.notImplemented,
            createLiteralConstValue: ts.notImplemented,
            isSymbolAccessible: ts.notImplemented,
            isEntityNameVisible: ts.notImplemented,
            // Returns the constant value this property access resolves to: notImplemented, or 'undefined' for a non-constant
            getConstantValue: ts.notImplemented,
            getReferencedValueDeclaration: ts.notImplemented,
            getTypeReferenceSerializationKind: ts.notImplemented,
            isOptionalParameter: ts.notImplemented,
            moduleExportsSomeValue: ts.notImplemented,
            isArgumentsLocalBinding: ts.notImplemented,
            getExternalModuleFileFromDeclaration: ts.notImplemented,
            getTypeReferenceDirectivesForEntityName: ts.notImplemented,
            getTypeReferenceDirectivesForSymbol: ts.notImplemented,
            isLiteralConstDeclaration: ts.notImplemented,
            getJsxFactoryEntity: ts.notImplemented,
            getJsxFragmentFactoryEntity: ts.notImplemented,
            getAllAccessorDeclarations: ts.notImplemented,
            getSymbolOfExternalModuleSpecifier: ts.notImplemented,
            isBindingCapturedByNode: ts.notImplemented,
            getDeclarationStatementsForSourceFile: ts.notImplemented,
            isImportRequiredByAugmentation: ts.notImplemented,
        };
        function createSourceFilesFromBundleBuildInfo(bundle, buildInfoDirectory, host) {
            var _a;
            var jsBundle = ts.Debug.checkDefined(bundle.js);
            var prologueMap = ((_a = jsBundle.sources) === null || _a === void 0 ? void 0 : _a.prologues) && ts.arrayToMap(jsBundle.sources.prologues, function (prologueInfo) { return prologueInfo.file; });
            return bundle.sourceFiles.map(function (fileName, index) {
                var _a, _b;
                var prologueInfo = prologueMap === null || prologueMap === void 0 ? void 0 : prologueMap.get(index);
                var statements = prologueInfo === null || prologueInfo === void 0 ? void 0 : prologueInfo.directives.map(function (directive) {
                    var literal = ts.setTextRange(ts.factory.createStringLiteral(directive.expression.text), directive.expression);
                    var statement = ts.setTextRange(ts.factory.createExpressionStatement(literal), directive);
                    ts.setParent(literal, statement);
                    return statement;
                });
                var eofToken = ts.factory.createToken(1 /* EndOfFileToken */);
                var sourceFile = ts.factory.createSourceFile(statements !== null && statements !== void 0 ? statements : [], eofToken, 0 /* None */);
                sourceFile.fileName = ts.getRelativePathFromDirectory(host.getCurrentDirectory(), ts.getNormalizedAbsolutePath(fileName, buildInfoDirectory), !host.useCaseSensitiveFileNames());
                sourceFile.text = (_a = prologueInfo === null || prologueInfo === void 0 ? void 0 : prologueInfo.text) !== null && _a !== void 0 ? _a : "";
                ts.setTextRangePosWidth(sourceFile, 0, (_b = prologueInfo === null || prologueInfo === void 0 ? void 0 : prologueInfo.text.length) !== null && _b !== void 0 ? _b : 0);
                ts.setEachParent(sourceFile.statements, sourceFile);
                ts.setTextRangePosWidth(eofToken, sourceFile.end, 0);
                ts.setParent(eofToken, sourceFile);
                return sourceFile;
            });
        }
        /*@internal*/
        function emitUsingBuildInfo(config, host, getCommandLine, customTransformers) {
            var _a = getOutputPathsForBundle(config.options, /*forceDtsPaths*/ false), buildInfoPath = _a.buildInfoPath, jsFilePath = _a.jsFilePath, sourceMapFilePath = _a.sourceMapFilePath, declarationFilePath = _a.declarationFilePath, declarationMapPath = _a.declarationMapPath;
            var buildInfoText = host.readFile(ts.Debug.checkDefined(buildInfoPath));
            if (!buildInfoText)
                return buildInfoPath;
            var jsFileText = host.readFile(ts.Debug.checkDefined(jsFilePath));
            if (!jsFileText)
                return jsFilePath;
            var sourceMapText = sourceMapFilePath && host.readFile(sourceMapFilePath);
            // error if no source map or for now if inline sourcemap
            if ((sourceMapFilePath && !sourceMapText) || config.options.inlineSourceMap)
                return sourceMapFilePath || "inline sourcemap decoding";
            // read declaration text
            var declarationText = declarationFilePath && host.readFile(declarationFilePath);
            if (declarationFilePath && !declarationText)
                return declarationFilePath;
            var declarationMapText = declarationMapPath && host.readFile(declarationMapPath);
            // error if no source map or for now if inline sourcemap
            if ((declarationMapPath && !declarationMapText) || config.options.inlineSourceMap)
                return declarationMapPath || "inline sourcemap decoding";
            var buildInfo = getBuildInfo(buildInfoText);
            if (!buildInfo.bundle || !buildInfo.bundle.js || (declarationText && !buildInfo.bundle.dts))
                return buildInfoPath;
            var buildInfoDirectory = ts.getDirectoryPath(ts.getNormalizedAbsolutePath(buildInfoPath, host.getCurrentDirectory()));
            var ownPrependInput = ts.createInputFiles(jsFileText, declarationText, sourceMapFilePath, sourceMapText, declarationMapPath, declarationMapText, jsFilePath, declarationFilePath, buildInfoPath, buildInfo, 
            /*onlyOwnText*/ true);
            var outputFiles = [];
            var prependNodes = ts.createPrependNodes(config.projectReferences, getCommandLine, function (f) { return host.readFile(f); });
            var sourceFilesForJsEmit = createSourceFilesFromBundleBuildInfo(buildInfo.bundle, buildInfoDirectory, host);
            var emitHost = {
                getPrependNodes: ts.memoize(function () { return __spreadArray(__spreadArray([], prependNodes, true), [ownPrependInput], false); }),
                getCanonicalFileName: host.getCanonicalFileName,
                getCommonSourceDirectory: function () { return ts.getNormalizedAbsolutePath(buildInfo.bundle.commonSourceDirectory, buildInfoDirectory); },
                getCompilerOptions: function () { return config.options; },
                getCurrentDirectory: function () { return host.getCurrentDirectory(); },
                getNewLine: function () { return host.getNewLine(); },
                getSourceFile: ts.returnUndefined,
                getSourceFileByPath: ts.returnUndefined,
                getSourceFiles: function () { return sourceFilesForJsEmit; },
                getLibFileFromReference: ts.notImplemented,
                isSourceFileFromExternalLibrary: ts.returnFalse,
                getResolvedProjectReferenceToRedirect: ts.returnUndefined,
                getProjectReferenceRedirect: ts.returnUndefined,
                isSourceOfProjectReferenceRedirect: ts.returnFalse,
                writeFile: function (name, text, writeByteOrderMark) {
                    switch (name) {
                        case jsFilePath:
                            if (jsFileText === text)
                                return;
                            break;
                        case sourceMapFilePath:
                            if (sourceMapText === text)
                                return;
                            break;
                        case buildInfoPath:
                            var newBuildInfo = getBuildInfo(text);
                            newBuildInfo.program = buildInfo.program;
                            // Update sourceFileInfo
                            var _a = buildInfo.bundle, js = _a.js, dts = _a.dts, sourceFiles = _a.sourceFiles;
                            newBuildInfo.bundle.js.sources = js.sources;
                            if (dts) {
                                newBuildInfo.bundle.dts.sources = dts.sources;
                            }
                            newBuildInfo.bundle.sourceFiles = sourceFiles;
                            outputFiles.push({ name: name, text: getBuildInfoText(newBuildInfo), writeByteOrderMark: writeByteOrderMark });
                            return;
                        case declarationFilePath:
                            if (declarationText === text)
                                return;
                            break;
                        case declarationMapPath:
                            if (declarationMapText === text)
                                return;
                            break;
                        default:
                            ts.Debug.fail("Unexpected path: " + name);
                    }
                    outputFiles.push({ name: name, text: text, writeByteOrderMark: writeByteOrderMark });
                },
                isEmitBlocked: ts.returnFalse,
                readFile: function (f) { return host.readFile(f); },
                fileExists: function (f) { return host.fileExists(f); },
                useCaseSensitiveFileNames: function () { return host.useCaseSensitiveFileNames(); },
                getProgramBuildInfo: ts.returnUndefined,
                getSourceFileFromReference: ts.returnUndefined,
                redirectTargetsMap: ts.createMultiMap(),
                getFileIncludeReasons: ts.notImplemented,
            };
            emitFiles(ts.notImplementedResolver, emitHost, 
            /*targetSourceFile*/ undefined, ts.getTransformers(config.options, customTransformers));
            return outputFiles;
        }
        ts.emitUsingBuildInfo = emitUsingBuildInfo;
        var PipelinePhase;
        (function (PipelinePhase) {
            PipelinePhase[PipelinePhase["Notification"] = 0] = "Notification";
            PipelinePhase[PipelinePhase["Substitution"] = 1] = "Substitution";
            PipelinePhase[PipelinePhase["Comments"] = 2] = "Comments";
            PipelinePhase[PipelinePhase["SourceMaps"] = 3] = "SourceMaps";
            PipelinePhase[PipelinePhase["Emit"] = 4] = "Emit";
        })(PipelinePhase || (PipelinePhase = {}));
        function createPrinter(printerOptions, handlers) {
            if (printerOptions === void 0) {
                printerOptions = {};
            }
            if (handlers === void 0) {
                handlers = {};
            }
            var hasGlobalName = handlers.hasGlobalName, _a = handlers.onEmitNode, onEmitNode = _a === void 0 ? ts.noEmitNotification : _a, isEmitNotificationEnabled = handlers.isEmitNotificationEnabled, _b = handlers.substituteNode, substituteNode = _b === void 0 ? ts.noEmitSubstitution : _b, onBeforeEmitNode = handlers.onBeforeEmitNode, onAfterEmitNode = handlers.onAfterEmitNode, onBeforeEmitNodeArray = handlers.onBeforeEmitNodeArray, onAfterEmitNodeArray = handlers.onAfterEmitNodeArray, onBeforeEmitToken = handlers.onBeforeEmitToken, onAfterEmitToken = handlers.onAfterEmitToken;
            var extendedDiagnostics = !!printerOptions.extendedDiagnostics;
            var newLine = ts.getNewLineCharacter(printerOptions);
            var moduleKind = ts.getEmitModuleKind(printerOptions);
            var bundledHelpers = new ts.Map();
            var currentSourceFile;
            var nodeIdToGeneratedName; // Map of generated names for specific nodes.
            var autoGeneratedIdToGeneratedName; // Map of generated names for temp and loop variables.
            var generatedNames; // Set of names generated by the NameGenerator.
            var tempFlagsStack; // Stack of enclosing name generation scopes.
            var tempFlags; // TempFlags for the current name generation scope.
            var reservedNamesStack; // Stack of TempFlags reserved in enclosing name generation scopes.
            var reservedNames; // TempFlags to reserve in nested name generation scopes.
            var preserveSourceNewlines = printerOptions.preserveSourceNewlines; // Can be overridden inside nodes with the `IgnoreSourceNewlines` emit flag.
            var nextListElementPos; // See comment in `getLeadingLineTerminatorCount`.
            var writer;
            var ownWriter; // Reusable `EmitTextWriter` for basic printing.
            var write = writeBase;
            var isOwnFileEmit;
            var bundleFileInfo = printerOptions.writeBundleFileInfo ? { sections: [] } : undefined;
            var relativeToBuildInfo = bundleFileInfo ? ts.Debug.checkDefined(printerOptions.relativeToBuildInfo) : undefined;
            var recordInternalSection = printerOptions.recordInternalSection;
            var sourceFileTextPos = 0;
            var sourceFileTextKind = "text" /* Text */;
            // Source Maps
            var sourceMapsDisabled = true;
            var sourceMapGenerator;
            var sourceMapSource;
            var sourceMapSourceIndex = -1;
            var mostRecentlyAddedSourceMapSource;
            var mostRecentlyAddedSourceMapSourceIndex = -1;
            // Comments
            var containerPos = -1;
            var containerEnd = -1;
            var declarationListContainerEnd = -1;
            var currentLineMap;
            var detachedCommentsInfo;
            var hasWrittenComment = false;
            var commentsDisabled = !!printerOptions.removeComments;
            var lastSubstitution;
            var currentParenthesizerRule;
            var _c = ts.performance.createTimerIf(extendedDiagnostics, "commentTime", "beforeComment", "afterComment"), enterComment = _c.enter, exitComment = _c.exit;
            var parenthesizer = ts.factory.parenthesizer;
            var emitBinaryExpression = createEmitBinaryExpression();
            reset();
            return {
                // public API
                printNode: printNode,
                printList: printList,
                printFile: printFile,
                printBundle: printBundle,
                // internal API
                writeNode: writeNode,
                writeList: writeList,
                writeFile: writeFile,
                writeBundle: writeBundle,
                bundleFileInfo: bundleFileInfo
            };
            function printNode(hint, node, sourceFile) {
                switch (hint) {
                    case 0 /* SourceFile */:
                        ts.Debug.assert(ts.isSourceFile(node), "Expected a SourceFile node.");
                        break;
                    case 2 /* IdentifierName */:
                        ts.Debug.assert(ts.isIdentifier(node), "Expected an Identifier node.");
                        break;
                    case 1 /* Expression */:
                        ts.Debug.assert(ts.isExpression(node), "Expected an Expression node.");
                        break;
                }
                switch (node.kind) {
                    case 300 /* SourceFile */: return printFile(node);
                    case 301 /* Bundle */: return printBundle(node);
                    case 302 /* UnparsedSource */: return printUnparsedSource(node);
                }
                writeNode(hint, node, sourceFile, beginPrint());
                return endPrint();
            }
            function printList(format, nodes, sourceFile) {
                writeList(format, nodes, sourceFile, beginPrint());
                return endPrint();
            }
            function printBundle(bundle) {
                writeBundle(bundle, beginPrint(), /*sourceMapEmitter*/ undefined);
                return endPrint();
            }
            function printFile(sourceFile) {
                writeFile(sourceFile, beginPrint(), /*sourceMapEmitter*/ undefined);
                return endPrint();
            }
            function printUnparsedSource(unparsed) {
                writeUnparsedSource(unparsed, beginPrint());
                return endPrint();
            }
            function writeNode(hint, node, sourceFile, output) {
                var previousWriter = writer;
                setWriter(output, /*_sourceMapGenerator*/ undefined);
                print(hint, node, sourceFile);
                reset();
                writer = previousWriter;
            }
            function writeList(format, nodes, sourceFile, output) {
                var previousWriter = writer;
                setWriter(output, /*_sourceMapGenerator*/ undefined);
                if (sourceFile) {
                    setSourceFile(sourceFile);
                }
                emitList(/*parentNode*/ undefined, nodes, format);
                reset();
                writer = previousWriter;
            }
            function getTextPosWithWriteLine() {
                return writer.getTextPosWithWriteLine ? writer.getTextPosWithWriteLine() : writer.getTextPos();
            }
            function updateOrPushBundleFileTextLike(pos, end, kind) {
                var last = ts.lastOrUndefined(bundleFileInfo.sections);
                if (last && last.kind === kind) {
                    last.end = end;
                }
                else {
                    bundleFileInfo.sections.push({ pos: pos, end: end, kind: kind });
                }
            }
            function recordBundleFileInternalSectionStart(node) {
                if (recordInternalSection &&
                    bundleFileInfo &&
                    currentSourceFile &&
                    (ts.isDeclaration(node) || ts.isVariableStatement(node)) &&
                    ts.isInternalDeclaration(node, currentSourceFile) &&
                    sourceFileTextKind !== "internal" /* Internal */) {
                    var prevSourceFileTextKind = sourceFileTextKind;
                    recordBundleFileTextLikeSection(writer.getTextPos());
                    sourceFileTextPos = getTextPosWithWriteLine();
                    sourceFileTextKind = "internal" /* Internal */;
                    return prevSourceFileTextKind;
                }
                return undefined;
            }
            function recordBundleFileInternalSectionEnd(prevSourceFileTextKind) {
                if (prevSourceFileTextKind) {
                    recordBundleFileTextLikeSection(writer.getTextPos());
                    sourceFileTextPos = getTextPosWithWriteLine();
                    sourceFileTextKind = prevSourceFileTextKind;
                }
            }
            function recordBundleFileTextLikeSection(end) {
                if (sourceFileTextPos < end) {
                    updateOrPushBundleFileTextLike(sourceFileTextPos, end, sourceFileTextKind);
                    return true;
                }
                return false;
            }
            function writeBundle(bundle, output, sourceMapGenerator) {
                var _a;
                isOwnFileEmit = false;
                var previousWriter = writer;
                setWriter(output, sourceMapGenerator);
                emitShebangIfNeeded(bundle);
                emitPrologueDirectivesIfNeeded(bundle);
                emitHelpers(bundle);
                emitSyntheticTripleSlashReferencesIfNeeded(bundle);
                for (var _b = 0, _c = bundle.prepends; _b < _c.length; _b++) {
                    var prepend = _c[_b];
                    writeLine();
                    var pos = writer.getTextPos();
                    var savedSections = bundleFileInfo && bundleFileInfo.sections;
                    if (savedSections)
                        bundleFileInfo.sections = [];
                    print(4 /* Unspecified */, prepend, /*sourceFile*/ undefined);
                    if (bundleFileInfo) {
                        var newSections = bundleFileInfo.sections;
                        bundleFileInfo.sections = savedSections;
                        if (prepend.oldFileOfCurrentEmit)
                            (_a = bundleFileInfo.sections).push.apply(_a, newSections);
                        else {
                            newSections.forEach(function (section) { return ts.Debug.assert(ts.isBundleFileTextLike(section)); });
                            bundleFileInfo.sections.push({
                                pos: pos,
                                end: writer.getTextPos(),
                                kind: "prepend" /* Prepend */,
                                data: relativeToBuildInfo(prepend.fileName),
                                texts: newSections
                            });
                        }
                    }
                }
                sourceFileTextPos = getTextPosWithWriteLine();
                for (var _d = 0, _e = bundle.sourceFiles; _d < _e.length; _d++) {
                    var sourceFile = _e[_d];
                    print(0 /* SourceFile */, sourceFile, sourceFile);
                }
                if (bundleFileInfo && bundle.sourceFiles.length) {
                    var end = writer.getTextPos();
                    if (recordBundleFileTextLikeSection(end)) {
                        // Store prologues
                        var prologues = getPrologueDirectivesFromBundledSourceFiles(bundle);
                        if (prologues) {
                            if (!bundleFileInfo.sources)
                                bundleFileInfo.sources = {};
                            bundleFileInfo.sources.prologues = prologues;
                        }
                        // Store helpes
                        var helpers = getHelpersFromBundledSourceFiles(bundle);
                        if (helpers) {
                            if (!bundleFileInfo.sources)
                                bundleFileInfo.sources = {};
                            bundleFileInfo.sources.helpers = helpers;
                        }
                    }
                }
                reset();
                writer = previousWriter;
            }
            function writeUnparsedSource(unparsed, output) {
                var previousWriter = writer;
                setWriter(output, /*_sourceMapGenerator*/ undefined);
                print(4 /* Unspecified */, unparsed, /*sourceFile*/ undefined);
                reset();
                writer = previousWriter;
            }
            function writeFile(sourceFile, output, sourceMapGenerator) {
                isOwnFileEmit = true;
                var previousWriter = writer;
                setWriter(output, sourceMapGenerator);
                emitShebangIfNeeded(sourceFile);
                emitPrologueDirectivesIfNeeded(sourceFile);
                print(0 /* SourceFile */, sourceFile, sourceFile);
                reset();
                writer = previousWriter;
            }
            function beginPrint() {
                return ownWriter || (ownWriter = ts.createTextWriter(newLine));
            }
            function endPrint() {
                var text = ownWriter.getText();
                ownWriter.clear();
                return text;
            }
            function print(hint, node, sourceFile) {
                if (sourceFile) {
                    setSourceFile(sourceFile);
                }
                pipelineEmit(hint, node, /*parenthesizerRule*/ undefined);
            }
            function setSourceFile(sourceFile) {
                currentSourceFile = sourceFile;
                currentLineMap = undefined;
                detachedCommentsInfo = undefined;
                if (sourceFile) {
                    setSourceMapSource(sourceFile);
                }
            }
            function setWriter(_writer, _sourceMapGenerator) {
                if (_writer && printerOptions.omitTrailingSemicolon) {
                    _writer = ts.getTrailingSemicolonDeferringWriter(_writer);
                }
                writer = _writer; // TODO: GH#18217
                sourceMapGenerator = _sourceMapGenerator;
                sourceMapsDisabled = !writer || !sourceMapGenerator;
            }
            function reset() {
                nodeIdToGeneratedName = [];
                autoGeneratedIdToGeneratedName = [];
                generatedNames = new ts.Set();
                tempFlagsStack = [];
                tempFlags = 0 /* Auto */;
                reservedNamesStack = [];
                currentSourceFile = undefined;
                currentLineMap = undefined;
                detachedCommentsInfo = undefined;
                setWriter(/*output*/ undefined, /*_sourceMapGenerator*/ undefined);
            }
            function getCurrentLineMap() {
                return currentLineMap || (currentLineMap = ts.getLineStarts(currentSourceFile));
            }
            function emit(node, parenthesizerRule) {
                if (node === undefined)
                    return;
                var prevSourceFileTextKind = recordBundleFileInternalSectionStart(node);
                pipelineEmit(4 /* Unspecified */, node, parenthesizerRule);
                recordBundleFileInternalSectionEnd(prevSourceFileTextKind);
            }
            function emitIdentifierName(node) {
                if (node === undefined)
                    return;
                pipelineEmit(2 /* IdentifierName */, node, /*parenthesizerRule*/ undefined);
            }
            function emitExpression(node, parenthesizerRule) {
                if (node === undefined)
                    return;
                pipelineEmit(1 /* Expression */, node, parenthesizerRule);
            }
            function emitJsxAttributeValue(node) {
                pipelineEmit(ts.isStringLiteral(node) ? 6 /* JsxAttributeValue */ : 4 /* Unspecified */, node);
            }
            function beforeEmitNode(node) {
                if (preserveSourceNewlines && (ts.getEmitFlags(node) & 134217728 /* IgnoreSourceNewlines */)) {
                    preserveSourceNewlines = false;
                }
            }
            function afterEmitNode(savedPreserveSourceNewlines) {
                preserveSourceNewlines = savedPreserveSourceNewlines;
            }
            function pipelineEmit(emitHint, node, parenthesizerRule) {
                currentParenthesizerRule = parenthesizerRule;
                var pipelinePhase = getPipelinePhase(0 /* Notification */, emitHint, node);
                pipelinePhase(emitHint, node);
                currentParenthesizerRule = undefined;
            }
            function shouldEmitComments(node) {
                return !commentsDisabled && !ts.isSourceFile(node);
            }
            function shouldEmitSourceMaps(node) {
                return !sourceMapsDisabled &&
                    !ts.isSourceFile(node) &&
                    !ts.isInJsonFile(node) &&
                    !ts.isUnparsedSource(node) &&
                    !ts.isUnparsedPrepend(node);
            }
            function getPipelinePhase(phase, emitHint, node) {
                switch (phase) {
                    case 0 /* Notification */:
                        if (onEmitNode !== ts.noEmitNotification && (!isEmitNotificationEnabled || isEmitNotificationEnabled(node))) {
                            return pipelineEmitWithNotification;
                        }
                    // falls through
                    case 1 /* Substitution */:
                        if (substituteNode !== ts.noEmitSubstitution && (lastSubstitution = substituteNode(emitHint, node) || node) !== node) {
                            if (currentParenthesizerRule) {
                                lastSubstitution = currentParenthesizerRule(lastSubstitution);
                            }
                            return pipelineEmitWithSubstitution;
                        }
                    // falls through
                    case 2 /* Comments */:
                        if (shouldEmitComments(node)) {
                            return pipelineEmitWithComments;
                        }
                    // falls through
                    case 3 /* SourceMaps */:
                        if (shouldEmitSourceMaps(node)) {
                            return pipelineEmitWithSourceMaps;
                        }
                    // falls through
                    case 4 /* Emit */:
                        return pipelineEmitWithHint;
                    default:
                        return ts.Debug.assertNever(phase);
                }
            }
            function getNextPipelinePhase(currentPhase, emitHint, node) {
                return getPipelinePhase(currentPhase + 1, emitHint, node);
            }
            function pipelineEmitWithNotification(hint, node) {
                var pipelinePhase = getNextPipelinePhase(0 /* Notification */, hint, node);
                onEmitNode(hint, node, pipelinePhase);
            }
            function pipelineEmitWithHint(hint, node) {
                onBeforeEmitNode === null || onBeforeEmitNode === void 0 ? void 0 : onBeforeEmitNode(node);
                if (preserveSourceNewlines) {
                    var savedPreserveSourceNewlines = preserveSourceNewlines;
                    beforeEmitNode(node);
                    pipelineEmitWithHintWorker(hint, node);
                    afterEmitNode(savedPreserveSourceNewlines);
                }
                else {
                    pipelineEmitWithHintWorker(hint, node);
                }
                onAfterEmitNode === null || onAfterEmitNode === void 0 ? void 0 : onAfterEmitNode(node);
                // clear the parenthesizer rule as we ascend
                currentParenthesizerRule = undefined;
            }
            function pipelineEmitWithHintWorker(hint, node) {
                if (hint === 0 /* SourceFile */)
                    return emitSourceFile(ts.cast(node, ts.isSourceFile));
                if (hint === 2 /* IdentifierName */)
                    return emitIdentifier(ts.cast(node, ts.isIdentifier));
                if (hint === 6 /* JsxAttributeValue */)
                    return emitLiteral(ts.cast(node, ts.isStringLiteral), /*jsxAttributeEscape*/ true);
                if (hint === 3 /* MappedTypeParameter */)
                    return emitMappedTypeParameter(ts.cast(node, ts.isTypeParameterDeclaration));
                if (hint === 5 /* EmbeddedStatement */) {
                    ts.Debug.assertNode(node, ts.isEmptyStatement);
                    return emitEmptyStatement(/*isEmbeddedStatement*/ true);
                }
                if (hint === 4 /* Unspecified */) {
                    switch (node.kind) {
                        // Pseudo-literals
                        case 15 /* TemplateHead */:
                        case 16 /* TemplateMiddle */:
                        case 17 /* TemplateTail */:
                            return emitLiteral(node, /*jsxAttributeEscape*/ false);
                        // Identifiers
                        case 79 /* Identifier */:
                            return emitIdentifier(node);
                        // PrivateIdentifiers
                        case 80 /* PrivateIdentifier */:
                            return emitPrivateIdentifier(node);
                        // Parse tree nodes
                        // Names
                        case 159 /* QualifiedName */:
                            return emitQualifiedName(node);
                        case 160 /* ComputedPropertyName */:
                            return emitComputedPropertyName(node);
                        // Signature elements
                        case 161 /* TypeParameter */:
                            return emitTypeParameter(node);
                        case 162 /* Parameter */:
                            return emitParameter(node);
                        case 163 /* Decorator */:
                            return emitDecorator(node);
                        // Type members
                        case 164 /* PropertySignature */:
                            return emitPropertySignature(node);
                        case 165 /* PropertyDeclaration */:
                            return emitPropertyDeclaration(node);
                        case 166 /* MethodSignature */:
                            return emitMethodSignature(node);
                        case 167 /* MethodDeclaration */:
                            return emitMethodDeclaration(node);
                        case 168 /* ClassStaticBlockDeclaration */:
                            return emitClassStaticBlockDeclaration(node);
                        case 169 /* Constructor */:
                            return emitConstructor(node);
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            return emitAccessorDeclaration(node);
                        case 172 /* CallSignature */:
                            return emitCallSignature(node);
                        case 173 /* ConstructSignature */:
                            return emitConstructSignature(node);
                        case 174 /* IndexSignature */:
                            return emitIndexSignature(node);
                        // Types
                        case 175 /* TypePredicate */:
                            return emitTypePredicate(node);
                        case 176 /* TypeReference */:
                            return emitTypeReference(node);
                        case 177 /* FunctionType */:
                            return emitFunctionType(node);
                        case 178 /* ConstructorType */:
                            return emitConstructorType(node);
                        case 179 /* TypeQuery */:
                            return emitTypeQuery(node);
                        case 180 /* TypeLiteral */:
                            return emitTypeLiteral(node);
                        case 181 /* ArrayType */:
                            return emitArrayType(node);
                        case 182 /* TupleType */:
                            return emitTupleType(node);
                        case 183 /* OptionalType */:
                            return emitOptionalType(node);
                        // SyntaxKind.RestType is handled below
                        case 185 /* UnionType */:
                            return emitUnionType(node);
                        case 186 /* IntersectionType */:
                            return emitIntersectionType(node);
                        case 187 /* ConditionalType */:
                            return emitConditionalType(node);
                        case 188 /* InferType */:
                            return emitInferType(node);
                        case 189 /* ParenthesizedType */:
                            return emitParenthesizedType(node);
                        case 226 /* ExpressionWithTypeArguments */:
                            return emitExpressionWithTypeArguments(node);
                        case 190 /* ThisType */:
                            return emitThisType();
                        case 191 /* TypeOperator */:
                            return emitTypeOperator(node);
                        case 192 /* IndexedAccessType */:
                            return emitIndexedAccessType(node);
                        case 193 /* MappedType */:
                            return emitMappedType(node);
                        case 194 /* LiteralType */:
                            return emitLiteralType(node);
                        case 195 /* NamedTupleMember */:
                            return emitNamedTupleMember(node);
                        case 196 /* TemplateLiteralType */:
                            return emitTemplateType(node);
                        case 197 /* TemplateLiteralTypeSpan */:
                            return emitTemplateTypeSpan(node);
                        case 198 /* ImportType */:
                            return emitImportTypeNode(node);
                        // Binding patterns
                        case 199 /* ObjectBindingPattern */:
                            return emitObjectBindingPattern(node);
                        case 200 /* ArrayBindingPattern */:
                            return emitArrayBindingPattern(node);
                        case 201 /* BindingElement */:
                            return emitBindingElement(node);
                        // Misc
                        case 231 /* TemplateSpan */:
                            return emitTemplateSpan(node);
                        case 232 /* SemicolonClassElement */:
                            return emitSemicolonClassElement();
                        // Statements
                        case 233 /* Block */:
                            return emitBlock(node);
                        case 235 /* VariableStatement */:
                            return emitVariableStatement(node);
                        case 234 /* EmptyStatement */:
                            return emitEmptyStatement(/*isEmbeddedStatement*/ false);
                        case 236 /* ExpressionStatement */:
                            return emitExpressionStatement(node);
                        case 237 /* IfStatement */:
                            return emitIfStatement(node);
                        case 238 /* DoStatement */:
                            return emitDoStatement(node);
                        case 239 /* WhileStatement */:
                            return emitWhileStatement(node);
                        case 240 /* ForStatement */:
                            return emitForStatement(node);
                        case 241 /* ForInStatement */:
                            return emitForInStatement(node);
                        case 242 /* ForOfStatement */:
                            return emitForOfStatement(node);
                        case 243 /* ContinueStatement */:
                            return emitContinueStatement(node);
                        case 244 /* BreakStatement */:
                            return emitBreakStatement(node);
                        case 245 /* ReturnStatement */:
                            return emitReturnStatement(node);
                        case 246 /* WithStatement */:
                            return emitWithStatement(node);
                        case 247 /* SwitchStatement */:
                            return emitSwitchStatement(node);
                        case 248 /* LabeledStatement */:
                            return emitLabeledStatement(node);
                        case 249 /* ThrowStatement */:
                            return emitThrowStatement(node);
                        case 250 /* TryStatement */:
                            return emitTryStatement(node);
                        case 251 /* DebuggerStatement */:
                            return emitDebuggerStatement(node);
                        // Declarations
                        case 252 /* VariableDeclaration */:
                            return emitVariableDeclaration(node);
                        case 253 /* VariableDeclarationList */:
                            return emitVariableDeclarationList(node);
                        case 254 /* FunctionDeclaration */:
                            return emitFunctionDeclaration(node);
                        case 255 /* ClassDeclaration */:
                            return emitClassDeclaration(node);
                        case 256 /* InterfaceDeclaration */:
                            return emitInterfaceDeclaration(node);
                        case 257 /* TypeAliasDeclaration */:
                            return emitTypeAliasDeclaration(node);
                        case 258 /* EnumDeclaration */:
                            return emitEnumDeclaration(node);
                        case 259 /* ModuleDeclaration */:
                            return emitModuleDeclaration(node);
                        case 260 /* ModuleBlock */:
                            return emitModuleBlock(node);
                        case 261 /* CaseBlock */:
                            return emitCaseBlock(node);
                        case 262 /* NamespaceExportDeclaration */:
                            return emitNamespaceExportDeclaration(node);
                        case 263 /* ImportEqualsDeclaration */:
                            return emitImportEqualsDeclaration(node);
                        case 264 /* ImportDeclaration */:
                            return emitImportDeclaration(node);
                        case 265 /* ImportClause */:
                            return emitImportClause(node);
                        case 266 /* NamespaceImport */:
                            return emitNamespaceImport(node);
                        case 272 /* NamespaceExport */:
                            return emitNamespaceExport(node);
                        case 267 /* NamedImports */:
                            return emitNamedImports(node);
                        case 268 /* ImportSpecifier */:
                            return emitImportSpecifier(node);
                        case 269 /* ExportAssignment */:
                            return emitExportAssignment(node);
                        case 270 /* ExportDeclaration */:
                            return emitExportDeclaration(node);
                        case 271 /* NamedExports */:
                            return emitNamedExports(node);
                        case 273 /* ExportSpecifier */:
                            return emitExportSpecifier(node);
                        case 274 /* MissingDeclaration */:
                            return;
                        // Module references
                        case 275 /* ExternalModuleReference */:
                            return emitExternalModuleReference(node);
                        // JSX (non-expression)
                        case 11 /* JsxText */:
                            return emitJsxText(node);
                        case 278 /* JsxOpeningElement */:
                        case 281 /* JsxOpeningFragment */:
                            return emitJsxOpeningElementOrFragment(node);
                        case 279 /* JsxClosingElement */:
                        case 282 /* JsxClosingFragment */:
                            return emitJsxClosingElementOrFragment(node);
                        case 283 /* JsxAttribute */:
                            return emitJsxAttribute(node);
                        case 284 /* JsxAttributes */:
                            return emitJsxAttributes(node);
                        case 285 /* JsxSpreadAttribute */:
                            return emitJsxSpreadAttribute(node);
                        case 286 /* JsxExpression */:
                            return emitJsxExpression(node);
                        // Clauses
                        case 287 /* CaseClause */:
                            return emitCaseClause(node);
                        case 288 /* DefaultClause */:
                            return emitDefaultClause(node);
                        case 289 /* HeritageClause */:
                            return emitHeritageClause(node);
                        case 290 /* CatchClause */:
                            return emitCatchClause(node);
                        // Property assignments
                        case 291 /* PropertyAssignment */:
                            return emitPropertyAssignment(node);
                        case 292 /* ShorthandPropertyAssignment */:
                            return emitShorthandPropertyAssignment(node);
                        case 293 /* SpreadAssignment */:
                            return emitSpreadAssignment(node);
                        // Enum
                        case 294 /* EnumMember */:
                            return emitEnumMember(node);
                        // Unparsed
                        case 295 /* UnparsedPrologue */:
                            return writeUnparsedNode(node);
                        case 302 /* UnparsedSource */:
                        case 296 /* UnparsedPrepend */:
                            return emitUnparsedSourceOrPrepend(node);
                        case 297 /* UnparsedText */:
                        case 298 /* UnparsedInternalText */:
                            return emitUnparsedTextLike(node);
                        case 299 /* UnparsedSyntheticReference */:
                            return emitUnparsedSyntheticReference(node);
                        // Top-level nodes
                        case 300 /* SourceFile */:
                            return emitSourceFile(node);
                        case 301 /* Bundle */:
                            return ts.Debug.fail("Bundles should be printed using printBundle");
                        // SyntaxKind.UnparsedSource (handled above)
                        case 303 /* InputFiles */:
                            return ts.Debug.fail("InputFiles should not be printed");
                        // JSDoc nodes (only used in codefixes currently)
                        case 304 /* JSDocTypeExpression */:
                            return emitJSDocTypeExpression(node);
                        case 305 /* JSDocNameReference */:
                            return emitJSDocNameReference(node);
                        case 307 /* JSDocAllType */:
                            return writePunctuation("*");
                        case 308 /* JSDocUnknownType */:
                            return writePunctuation("?");
                        case 309 /* JSDocNullableType */:
                            return emitJSDocNullableType(node);
                        case 310 /* JSDocNonNullableType */:
                            return emitJSDocNonNullableType(node);
                        case 311 /* JSDocOptionalType */:
                            return emitJSDocOptionalType(node);
                        case 312 /* JSDocFunctionType */:
                            return emitJSDocFunctionType(node);
                        case 184 /* RestType */:
                        case 313 /* JSDocVariadicType */:
                            return emitRestOrJSDocVariadicType(node);
                        case 314 /* JSDocNamepathType */:
                            return;
                        case 315 /* JSDocComment */:
                            return emitJSDoc(node);
                        case 317 /* JSDocTypeLiteral */:
                            return emitJSDocTypeLiteral(node);
                        case 318 /* JSDocSignature */:
                            return emitJSDocSignature(node);
                        case 322 /* JSDocTag */:
                        case 327 /* JSDocClassTag */:
                            return emitJSDocSimpleTag(node);
                        case 323 /* JSDocAugmentsTag */:
                        case 324 /* JSDocImplementsTag */:
                            return emitJSDocHeritageTag(node);
                        case 325 /* JSDocAuthorTag */:
                        case 326 /* JSDocDeprecatedTag */:
                            return;
                        // SyntaxKind.JSDocClassTag (see JSDocTag, above)
                        case 328 /* JSDocPublicTag */:
                        case 329 /* JSDocPrivateTag */:
                        case 330 /* JSDocProtectedTag */:
                        case 331 /* JSDocReadonlyTag */:
                        case 332 /* JSDocOverrideTag */:
                            return;
                        case 333 /* JSDocCallbackTag */:
                            return emitJSDocCallbackTag(node);
                        // SyntaxKind.JSDocEnumTag (see below)
                        case 335 /* JSDocParameterTag */:
                        case 342 /* JSDocPropertyTag */:
                            return emitJSDocPropertyLikeTag(node);
                        case 334 /* JSDocEnumTag */:
                        case 336 /* JSDocReturnTag */:
                        case 337 /* JSDocThisTag */:
                        case 338 /* JSDocTypeTag */:
                            return emitJSDocSimpleTypedTag(node);
                        case 339 /* JSDocTemplateTag */:
                            return emitJSDocTemplateTag(node);
                        case 340 /* JSDocTypedefTag */:
                            return emitJSDocTypedefTag(node);
                        case 341 /* JSDocSeeTag */:
                            return emitJSDocSeeTag(node);
                        // SyntaxKind.JSDocPropertyTag (see JSDocParameterTag, above)
                        // Transformation nodes
                        case 344 /* NotEmittedStatement */:
                        case 348 /* EndOfDeclarationMarker */:
                        case 347 /* MergeDeclarationMarker */:
                            return;
                    }
                    if (ts.isExpression(node)) {
                        hint = 1 /* Expression */;
                        if (substituteNode !== ts.noEmitSubstitution) {
                            var substitute = substituteNode(hint, node) || node;
                            if (substitute !== node) {
                                node = substitute;
                                if (currentParenthesizerRule) {
                                    node = currentParenthesizerRule(node);
                                }
                            }
                        }
                    }
                }
                if (hint === 1 /* Expression */) {
                    switch (node.kind) {
                        // Literals
                        case 8 /* NumericLiteral */:
                        case 9 /* BigIntLiteral */:
                            return emitNumericOrBigIntLiteral(node);
                        case 10 /* StringLiteral */:
                        case 13 /* RegularExpressionLiteral */:
                        case 14 /* NoSubstitutionTemplateLiteral */:
                            return emitLiteral(node, /*jsxAttributeEscape*/ false);
                        // Identifiers
                        case 79 /* Identifier */:
                            return emitIdentifier(node);
                        // Expressions
                        case 202 /* ArrayLiteralExpression */:
                            return emitArrayLiteralExpression(node);
                        case 203 /* ObjectLiteralExpression */:
                            return emitObjectLiteralExpression(node);
                        case 204 /* PropertyAccessExpression */:
                            return emitPropertyAccessExpression(node);
                        case 205 /* ElementAccessExpression */:
                            return emitElementAccessExpression(node);
                        case 206 /* CallExpression */:
                            return emitCallExpression(node);
                        case 207 /* NewExpression */:
                            return emitNewExpression(node);
                        case 208 /* TaggedTemplateExpression */:
                            return emitTaggedTemplateExpression(node);
                        case 209 /* TypeAssertionExpression */:
                            return emitTypeAssertionExpression(node);
                        case 210 /* ParenthesizedExpression */:
                            return emitParenthesizedExpression(node);
                        case 211 /* FunctionExpression */:
                            return emitFunctionExpression(node);
                        case 212 /* ArrowFunction */:
                            return emitArrowFunction(node);
                        case 213 /* DeleteExpression */:
                            return emitDeleteExpression(node);
                        case 214 /* TypeOfExpression */:
                            return emitTypeOfExpression(node);
                        case 215 /* VoidExpression */:
                            return emitVoidExpression(node);
                        case 216 /* AwaitExpression */:
                            return emitAwaitExpression(node);
                        case 217 /* PrefixUnaryExpression */:
                            return emitPrefixUnaryExpression(node);
                        case 218 /* PostfixUnaryExpression */:
                            return emitPostfixUnaryExpression(node);
                        case 219 /* BinaryExpression */:
                            return emitBinaryExpression(node);
                        case 220 /* ConditionalExpression */:
                            return emitConditionalExpression(node);
                        case 221 /* TemplateExpression */:
                            return emitTemplateExpression(node);
                        case 222 /* YieldExpression */:
                            return emitYieldExpression(node);
                        case 223 /* SpreadElement */:
                            return emitSpreadElement(node);
                        case 224 /* ClassExpression */:
                            return emitClassExpression(node);
                        case 225 /* OmittedExpression */:
                            return;
                        case 227 /* AsExpression */:
                            return emitAsExpression(node);
                        case 228 /* NonNullExpression */:
                            return emitNonNullExpression(node);
                        case 229 /* MetaProperty */:
                            return emitMetaProperty(node);
                        case 230 /* SyntheticExpression */:
                            return ts.Debug.fail("SyntheticExpression should never be printed.");
                        // JSX
                        case 276 /* JsxElement */:
                            return emitJsxElement(node);
                        case 277 /* JsxSelfClosingElement */:
                            return emitJsxSelfClosingElement(node);
                        case 280 /* JsxFragment */:
                            return emitJsxFragment(node);
                        // Synthesized list
                        case 343 /* SyntaxList */:
                            return ts.Debug.fail("SyntaxList should not be printed");
                        // Transformation nodes
                        case 344 /* NotEmittedStatement */:
                            return;
                        case 345 /* PartiallyEmittedExpression */:
                            return emitPartiallyEmittedExpression(node);
                        case 346 /* CommaListExpression */:
                            return emitCommaList(node);
                        case 347 /* MergeDeclarationMarker */:
                        case 348 /* EndOfDeclarationMarker */:
                            return;
                        case 349 /* SyntheticReferenceExpression */:
                            return ts.Debug.fail("SyntheticReferenceExpression should not be printed");
                    }
                }
                if (ts.isKeyword(node.kind))
                    return writeTokenNode(node, writeKeyword);
                if (ts.isTokenKind(node.kind))
                    return writeTokenNode(node, writePunctuation);
                ts.Debug.fail("Unhandled SyntaxKind: " + ts.Debug.formatSyntaxKind(node.kind) + ".");
            }
            function emitMappedTypeParameter(node) {
                emit(node.name);
                writeSpace();
                writeKeyword("in");
                writeSpace();
                emit(node.constraint);
            }
            function pipelineEmitWithSubstitution(hint, node) {
                var pipelinePhase = getNextPipelinePhase(1 /* Substitution */, hint, node);
                ts.Debug.assertIsDefined(lastSubstitution);
                node = lastSubstitution;
                lastSubstitution = undefined;
                pipelinePhase(hint, node);
            }
            function getHelpersFromBundledSourceFiles(bundle) {
                var result;
                if (moduleKind === ts.ModuleKind.None || printerOptions.noEmitHelpers) {
                    return undefined;
                }
                var bundledHelpers = new ts.Map();
                for (var _a = 0, _b = bundle.sourceFiles; _a < _b.length; _a++) {
                    var sourceFile = _b[_a];
                    var shouldSkip = ts.getExternalHelpersModuleName(sourceFile) !== undefined;
                    var helpers = getSortedEmitHelpers(sourceFile);
                    if (!helpers)
                        continue;
                    for (var _c = 0, helpers_5 = helpers; _c < helpers_5.length; _c++) {
                        var helper = helpers_5[_c];
                        if (!helper.scoped && !shouldSkip && !bundledHelpers.get(helper.name)) {
                            bundledHelpers.set(helper.name, true);
                            (result || (result = [])).push(helper.name);
                        }
                    }
                }
                return result;
            }
            function emitHelpers(node) {
                var helpersEmitted = false;
                var bundle = node.kind === 301 /* Bundle */ ? node : undefined;
                if (bundle && moduleKind === ts.ModuleKind.None) {
                    return;
                }
                var numPrepends = bundle ? bundle.prepends.length : 0;
                var numNodes = bundle ? bundle.sourceFiles.length + numPrepends : 1;
                for (var i = 0; i < numNodes; i++) {
                    var currentNode = bundle ? i < numPrepends ? bundle.prepends[i] : bundle.sourceFiles[i - numPrepends] : node;
                    var sourceFile = ts.isSourceFile(currentNode) ? currentNode : ts.isUnparsedSource(currentNode) ? undefined : currentSourceFile;
                    var shouldSkip = printerOptions.noEmitHelpers || (!!sourceFile && ts.hasRecordedExternalHelpers(sourceFile));
                    var shouldBundle = (ts.isSourceFile(currentNode) || ts.isUnparsedSource(currentNode)) && !isOwnFileEmit;
                    var helpers = ts.isUnparsedSource(currentNode) ? currentNode.helpers : getSortedEmitHelpers(currentNode);
                    if (helpers) {
                        for (var _a = 0, helpers_6 = helpers; _a < helpers_6.length; _a++) {
                            var helper = helpers_6[_a];
                            if (!helper.scoped) {
                                // Skip the helper if it can be skipped and the noEmitHelpers compiler
                                // option is set, or if it can be imported and the importHelpers compiler
                                // option is set.
                                if (shouldSkip)
                                    continue;
                                // Skip the helper if it can be bundled but hasn't already been emitted and we
                                // are emitting a bundled module.
                                if (shouldBundle) {
                                    if (bundledHelpers.get(helper.name)) {
                                        continue;
                                    }
                                    bundledHelpers.set(helper.name, true);
                                }
                            }
                            else if (bundle) {
                                // Skip the helper if it is scoped and we are emitting bundled helpers
                                continue;
                            }
                            var pos = getTextPosWithWriteLine();
                            if (typeof helper.text === "string") {
                                writeLines(helper.text);
                            }
                            else {
                                writeLines(helper.text(makeFileLevelOptimisticUniqueName));
                            }
                            if (bundleFileInfo)
                                bundleFileInfo.sections.push({ pos: pos, end: writer.getTextPos(), kind: "emitHelpers" /* EmitHelpers */, data: helper.name });
                            helpersEmitted = true;
                        }
                    }
                }
                return helpersEmitted;
            }
            function getSortedEmitHelpers(node) {
                var helpers = ts.getEmitHelpers(node);
                return helpers && ts.stableSort(helpers, ts.compareEmitHelpers);
            }
            //
            // Literals/Pseudo-literals
            //
            // SyntaxKind.NumericLiteral
            // SyntaxKind.BigIntLiteral
            function emitNumericOrBigIntLiteral(node) {
                emitLiteral(node, /*jsxAttributeEscape*/ false);
            }
            // SyntaxKind.StringLiteral
            // SyntaxKind.RegularExpressionLiteral
            // SyntaxKind.NoSubstitutionTemplateLiteral
            // SyntaxKind.TemplateHead
            // SyntaxKind.TemplateMiddle
            // SyntaxKind.TemplateTail
            function emitLiteral(node, jsxAttributeEscape) {
                var text = getLiteralTextOfNode(node, printerOptions.neverAsciiEscape, jsxAttributeEscape);
                if ((printerOptions.sourceMap || printerOptions.inlineSourceMap)
                    && (node.kind === 10 /* StringLiteral */ || ts.isTemplateLiteralKind(node.kind))) {
                    writeLiteral(text);
                }
                else {
                    // Quick info expects all literals to be called with writeStringLiteral, as there's no specific type for numberLiterals
                    writeStringLiteral(text);
                }
            }
            // SyntaxKind.UnparsedSource
            // SyntaxKind.UnparsedPrepend
            function emitUnparsedSourceOrPrepend(unparsed) {
                for (var _a = 0, _b = unparsed.texts; _a < _b.length; _a++) {
                    var text = _b[_a];
                    writeLine();
                    emit(text);
                }
            }
            // SyntaxKind.UnparsedPrologue
            // SyntaxKind.UnparsedText
            // SyntaxKind.UnparsedInternal
            // SyntaxKind.UnparsedSyntheticReference
            function writeUnparsedNode(unparsed) {
                writer.rawWrite(unparsed.parent.text.substring(unparsed.pos, unparsed.end));
            }
            // SyntaxKind.UnparsedText
            // SyntaxKind.UnparsedInternal
            function emitUnparsedTextLike(unparsed) {
                var pos = getTextPosWithWriteLine();
                writeUnparsedNode(unparsed);
                if (bundleFileInfo) {
                    updateOrPushBundleFileTextLike(pos, writer.getTextPos(), unparsed.kind === 297 /* UnparsedText */ ?
                        "text" /* Text */ :
                        "internal" /* Internal */);
                }
            }
            // SyntaxKind.UnparsedSyntheticReference
            function emitUnparsedSyntheticReference(unparsed) {
                var pos = getTextPosWithWriteLine();
                writeUnparsedNode(unparsed);
                if (bundleFileInfo) {
                    var section = ts.clone(unparsed.section);
                    section.pos = pos;
                    section.end = writer.getTextPos();
                    bundleFileInfo.sections.push(section);
                }
            }
            //
            // Identifiers
            //
            function emitIdentifier(node) {
                var writeText = node.symbol ? writeSymbol : write;
                writeText(getTextOfNode(node, /*includeTrivia*/ false), node.symbol);
                emitList(node, node.typeArguments, 53776 /* TypeParameters */); // Call emitList directly since it could be an array of TypeParameterDeclarations _or_ type arguments
            }
            //
            // Names
            //
            function emitPrivateIdentifier(node) {
                var writeText = node.symbol ? writeSymbol : write;
                writeText(getTextOfNode(node, /*includeTrivia*/ false), node.symbol);
            }
            function emitQualifiedName(node) {
                emitEntityName(node.left);
                writePunctuation(".");
                emit(node.right);
            }
            function emitEntityName(node) {
                if (node.kind === 79 /* Identifier */) {
                    emitExpression(node);
                }
                else {
                    emit(node);
                }
            }
            function emitComputedPropertyName(node) {
                writePunctuation("[");
                emitExpression(node.expression, parenthesizer.parenthesizeExpressionOfComputedPropertyName);
                writePunctuation("]");
            }
            //
            // Signature elements
            //
            function emitTypeParameter(node) {
                emit(node.name);
                if (node.constraint) {
                    writeSpace();
                    writeKeyword("extends");
                    writeSpace();
                    emit(node.constraint);
                }
                if (node.default) {
                    writeSpace();
                    writeOperator("=");
                    writeSpace();
                    emit(node.default);
                }
            }
            function emitParameter(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                emit(node.dotDotDotToken);
                emitNodeWithWriter(node.name, writeParameter);
                emit(node.questionToken);
                if (node.parent && node.parent.kind === 312 /* JSDocFunctionType */ && !node.name) {
                    emit(node.type);
                }
                else {
                    emitTypeAnnotation(node.type);
                }
                // The comment position has to fallback to any present node within the parameterdeclaration because as it turns out, the parser can make parameter declarations with _just_ an initializer.
                emitInitializer(node.initializer, node.type ? node.type.end : node.questionToken ? node.questionToken.end : node.name ? node.name.end : node.modifiers ? node.modifiers.end : node.decorators ? node.decorators.end : node.pos, node, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            function emitDecorator(decorator) {
                writePunctuation("@");
                emitExpression(decorator.expression, parenthesizer.parenthesizeLeftSideOfAccess);
            }
            //
            // Type members
            //
            function emitPropertySignature(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                emitNodeWithWriter(node.name, writeProperty);
                emit(node.questionToken);
                emitTypeAnnotation(node.type);
                writeTrailingSemicolon();
            }
            function emitPropertyDeclaration(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                emit(node.name);
                emit(node.questionToken);
                emit(node.exclamationToken);
                emitTypeAnnotation(node.type);
                emitInitializer(node.initializer, node.type ? node.type.end : node.questionToken ? node.questionToken.end : node.name.end, node);
                writeTrailingSemicolon();
            }
            function emitMethodSignature(node) {
                pushNameGenerationScope(node);
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                emit(node.name);
                emit(node.questionToken);
                emitTypeParameters(node, node.typeParameters);
                emitParameters(node, node.parameters);
                emitTypeAnnotation(node.type);
                writeTrailingSemicolon();
                popNameGenerationScope(node);
            }
            function emitMethodDeclaration(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                emit(node.asteriskToken);
                emit(node.name);
                emit(node.questionToken);
                emitSignatureAndBody(node, emitSignatureHead);
            }
            function emitClassStaticBlockDeclaration(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                writeKeyword("static");
                emitBlockFunctionBody(node.body);
            }
            function emitConstructor(node) {
                emitModifiers(node, node.modifiers);
                writeKeyword("constructor");
                emitSignatureAndBody(node, emitSignatureHead);
            }
            function emitAccessorDeclaration(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                writeKeyword(node.kind === 170 /* GetAccessor */ ? "get" : "set");
                writeSpace();
                emit(node.name);
                emitSignatureAndBody(node, emitSignatureHead);
            }
            function emitCallSignature(node) {
                pushNameGenerationScope(node);
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                emitTypeParameters(node, node.typeParameters);
                emitParameters(node, node.parameters);
                emitTypeAnnotation(node.type);
                writeTrailingSemicolon();
                popNameGenerationScope(node);
            }
            function emitConstructSignature(node) {
                pushNameGenerationScope(node);
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                writeKeyword("new");
                writeSpace();
                emitTypeParameters(node, node.typeParameters);
                emitParameters(node, node.parameters);
                emitTypeAnnotation(node.type);
                writeTrailingSemicolon();
                popNameGenerationScope(node);
            }
            function emitIndexSignature(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                emitParametersForIndexSignature(node, node.parameters);
                emitTypeAnnotation(node.type);
                writeTrailingSemicolon();
            }
            function emitTemplateTypeSpan(node) {
                emit(node.type);
                emit(node.literal);
            }
            function emitSemicolonClassElement() {
                writeTrailingSemicolon();
            }
            //
            // Types
            //
            function emitTypePredicate(node) {
                if (node.assertsModifier) {
                    emit(node.assertsModifier);
                    writeSpace();
                }
                emit(node.parameterName);
                if (node.type) {
                    writeSpace();
                    writeKeyword("is");
                    writeSpace();
                    emit(node.type);
                }
            }
            function emitTypeReference(node) {
                emit(node.typeName);
                emitTypeArguments(node, node.typeArguments);
            }
            function emitFunctionType(node) {
                pushNameGenerationScope(node);
                emitTypeParameters(node, node.typeParameters);
                emitParametersForArrow(node, node.parameters);
                writeSpace();
                writePunctuation("=>");
                writeSpace();
                emit(node.type);
                popNameGenerationScope(node);
            }
            function emitJSDocFunctionType(node) {
                writeKeyword("function");
                emitParameters(node, node.parameters);
                writePunctuation(":");
                emit(node.type);
            }
            function emitJSDocNullableType(node) {
                writePunctuation("?");
                emit(node.type);
            }
            function emitJSDocNonNullableType(node) {
                writePunctuation("!");
                emit(node.type);
            }
            function emitJSDocOptionalType(node) {
                emit(node.type);
                writePunctuation("=");
            }
            function emitConstructorType(node) {
                pushNameGenerationScope(node);
                emitModifiers(node, node.modifiers);
                writeKeyword("new");
                writeSpace();
                emitTypeParameters(node, node.typeParameters);
                emitParameters(node, node.parameters);
                writeSpace();
                writePunctuation("=>");
                writeSpace();
                emit(node.type);
                popNameGenerationScope(node);
            }
            function emitTypeQuery(node) {
                writeKeyword("typeof");
                writeSpace();
                emit(node.exprName);
            }
            function emitTypeLiteral(node) {
                writePunctuation("{");
                var flags = ts.getEmitFlags(node) & 1 /* SingleLine */ ? 768 /* SingleLineTypeLiteralMembers */ : 32897 /* MultiLineTypeLiteralMembers */;
                emitList(node, node.members, flags | 524288 /* NoSpaceIfEmpty */);
                writePunctuation("}");
            }
            function emitArrayType(node) {
                emit(node.elementType, parenthesizer.parenthesizeElementTypeOfArrayType);
                writePunctuation("[");
                writePunctuation("]");
            }
            function emitRestOrJSDocVariadicType(node) {
                writePunctuation("...");
                emit(node.type);
            }
            function emitTupleType(node) {
                emitTokenWithComment(22 /* OpenBracketToken */, node.pos, writePunctuation, node);
                var flags = ts.getEmitFlags(node) & 1 /* SingleLine */ ? 528 /* SingleLineTupleTypeElements */ : 657 /* MultiLineTupleTypeElements */;
                emitList(node, node.elements, flags | 524288 /* NoSpaceIfEmpty */);
                emitTokenWithComment(23 /* CloseBracketToken */, node.elements.end, writePunctuation, node);
            }
            function emitNamedTupleMember(node) {
                emit(node.dotDotDotToken);
                emit(node.name);
                emit(node.questionToken);
                emitTokenWithComment(58 /* ColonToken */, node.name.end, writePunctuation, node);
                writeSpace();
                emit(node.type);
            }
            function emitOptionalType(node) {
                emit(node.type, parenthesizer.parenthesizeElementTypeOfArrayType);
                writePunctuation("?");
            }
            function emitUnionType(node) {
                emitList(node, node.types, 516 /* UnionTypeConstituents */, parenthesizer.parenthesizeMemberOfElementType);
            }
            function emitIntersectionType(node) {
                emitList(node, node.types, 520 /* IntersectionTypeConstituents */, parenthesizer.parenthesizeMemberOfElementType);
            }
            function emitConditionalType(node) {
                emit(node.checkType, parenthesizer.parenthesizeMemberOfConditionalType);
                writeSpace();
                writeKeyword("extends");
                writeSpace();
                emit(node.extendsType, parenthesizer.parenthesizeMemberOfConditionalType);
                writeSpace();
                writePunctuation("?");
                writeSpace();
                emit(node.trueType);
                writeSpace();
                writePunctuation(":");
                writeSpace();
                emit(node.falseType);
            }
            function emitInferType(node) {
                writeKeyword("infer");
                writeSpace();
                emit(node.typeParameter);
            }
            function emitParenthesizedType(node) {
                writePunctuation("(");
                emit(node.type);
                writePunctuation(")");
            }
            function emitThisType() {
                writeKeyword("this");
            }
            function emitTypeOperator(node) {
                writeTokenText(node.operator, writeKeyword);
                writeSpace();
                emit(node.type, parenthesizer.parenthesizeMemberOfElementType);
            }
            function emitIndexedAccessType(node) {
                emit(node.objectType, parenthesizer.parenthesizeMemberOfElementType);
                writePunctuation("[");
                emit(node.indexType);
                writePunctuation("]");
            }
            function emitMappedType(node) {
                var emitFlags = ts.getEmitFlags(node);
                writePunctuation("{");
                if (emitFlags & 1 /* SingleLine */) {
                    writeSpace();
                }
                else {
                    writeLine();
                    increaseIndent();
                }
                if (node.readonlyToken) {
                    emit(node.readonlyToken);
                    if (node.readonlyToken.kind !== 143 /* ReadonlyKeyword */) {
                        writeKeyword("readonly");
                    }
                    writeSpace();
                }
                writePunctuation("[");
                pipelineEmit(3 /* MappedTypeParameter */, node.typeParameter);
                if (node.nameType) {
                    writeSpace();
                    writeKeyword("as");
                    writeSpace();
                    emit(node.nameType);
                }
                writePunctuation("]");
                if (node.questionToken) {
                    emit(node.questionToken);
                    if (node.questionToken.kind !== 57 /* QuestionToken */) {
                        writePunctuation("?");
                    }
                }
                writePunctuation(":");
                writeSpace();
                emit(node.type);
                writeTrailingSemicolon();
                if (emitFlags & 1 /* SingleLine */) {
                    writeSpace();
                }
                else {
                    writeLine();
                    decreaseIndent();
                }
                writePunctuation("}");
            }
            function emitLiteralType(node) {
                emitExpression(node.literal);
            }
            function emitTemplateType(node) {
                emit(node.head);
                emitList(node, node.templateSpans, 262144 /* TemplateExpressionSpans */);
            }
            function emitImportTypeNode(node) {
                if (node.isTypeOf) {
                    writeKeyword("typeof");
                    writeSpace();
                }
                writeKeyword("import");
                writePunctuation("(");
                emit(node.argument);
                writePunctuation(")");
                if (node.qualifier) {
                    writePunctuation(".");
                    emit(node.qualifier);
                }
                emitTypeArguments(node, node.typeArguments);
            }
            //
            // Binding patterns
            //
            function emitObjectBindingPattern(node) {
                writePunctuation("{");
                emitList(node, node.elements, 525136 /* ObjectBindingPatternElements */);
                writePunctuation("}");
            }
            function emitArrayBindingPattern(node) {
                writePunctuation("[");
                emitList(node, node.elements, 524880 /* ArrayBindingPatternElements */);
                writePunctuation("]");
            }
            function emitBindingElement(node) {
                emit(node.dotDotDotToken);
                if (node.propertyName) {
                    emit(node.propertyName);
                    writePunctuation(":");
                    writeSpace();
                }
                emit(node.name);
                emitInitializer(node.initializer, node.name.end, node, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            //
            // Expressions
            //
            function emitArrayLiteralExpression(node) {
                var elements = node.elements;
                var preferNewLine = node.multiLine ? 65536 /* PreferNewLine */ : 0 /* None */;
                emitExpressionList(node, elements, 8914 /* ArrayLiteralExpressionElements */ | preferNewLine, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            function emitObjectLiteralExpression(node) {
                ts.forEach(node.properties, generateMemberNames);
                var indentedFlag = ts.getEmitFlags(node) & 65536 /* Indented */;
                if (indentedFlag) {
                    increaseIndent();
                }
                var preferNewLine = node.multiLine ? 65536 /* PreferNewLine */ : 0 /* None */;
                var allowTrailingComma = currentSourceFile.languageVersion >= 1 /* ES5 */ && !ts.isJsonSourceFile(currentSourceFile) ? 64 /* AllowTrailingComma */ : 0 /* None */;
                emitList(node, node.properties, 526226 /* ObjectLiteralExpressionProperties */ | allowTrailingComma | preferNewLine);
                if (indentedFlag) {
                    decreaseIndent();
                }
            }
            function emitPropertyAccessExpression(node) {
                emitExpression(node.expression, parenthesizer.parenthesizeLeftSideOfAccess);
                var token = node.questionDotToken || ts.setTextRangePosEnd(ts.factory.createToken(24 /* DotToken */), node.expression.end, node.name.pos);
                var linesBeforeDot = getLinesBetweenNodes(node, node.expression, token);
                var linesAfterDot = getLinesBetweenNodes(node, token, node.name);
                writeLinesAndIndent(linesBeforeDot, /*writeSpaceIfNotIndenting*/ false);
                var shouldEmitDotDot = token.kind !== 28 /* QuestionDotToken */ &&
                    mayNeedDotDotForPropertyAccess(node.expression) &&
                    !writer.hasTrailingComment() &&
                    !writer.hasTrailingWhitespace();
                if (shouldEmitDotDot) {
                    writePunctuation(".");
                }
                if (node.questionDotToken) {
                    emit(token);
                }
                else {
                    emitTokenWithComment(token.kind, node.expression.end, writePunctuation, node);
                }
                writeLinesAndIndent(linesAfterDot, /*writeSpaceIfNotIndenting*/ false);
                emit(node.name);
                decreaseIndentIf(linesBeforeDot, linesAfterDot);
            }
            // 1..toString is a valid property access, emit a dot after the literal
            // Also emit a dot if expression is a integer const enum value - it will appear in generated code as numeric literal
            function mayNeedDotDotForPropertyAccess(expression) {
                expression = ts.skipPartiallyEmittedExpressions(expression);
                if (ts.isNumericLiteral(expression)) {
                    // check if numeric literal is a decimal literal that was originally written with a dot
                    var text = getLiteralTextOfNode(expression, /*neverAsciiEscape*/ true, /*jsxAttributeEscape*/ false);
                    // If he number will be printed verbatim and it doesn't already contain a dot, add one
                    // if the expression doesn't have any comments that will be emitted.
                    return !expression.numericLiteralFlags && !ts.stringContains(text, ts.tokenToString(24 /* DotToken */));
                }
                else if (ts.isAccessExpression(expression)) {
                    // check if constant enum value is integer
                    var constantValue = ts.getConstantValue(expression);
                    // isFinite handles cases when constantValue is undefined
                    return typeof constantValue === "number" && isFinite(constantValue)
                        && Math.floor(constantValue) === constantValue;
                }
            }
            function emitElementAccessExpression(node) {
                emitExpression(node.expression, parenthesizer.parenthesizeLeftSideOfAccess);
                emit(node.questionDotToken);
                emitTokenWithComment(22 /* OpenBracketToken */, node.expression.end, writePunctuation, node);
                emitExpression(node.argumentExpression);
                emitTokenWithComment(23 /* CloseBracketToken */, node.argumentExpression.end, writePunctuation, node);
            }
            function emitCallExpression(node) {
                var indirectCall = ts.getEmitFlags(node) & 536870912 /* IndirectCall */;
                if (indirectCall) {
                    writePunctuation("(");
                    writeLiteral("0");
                    writePunctuation(",");
                    writeSpace();
                }
                emitExpression(node.expression, parenthesizer.parenthesizeLeftSideOfAccess);
                if (indirectCall) {
                    writePunctuation(")");
                }
                emit(node.questionDotToken);
                emitTypeArguments(node, node.typeArguments);
                emitExpressionList(node, node.arguments, 2576 /* CallExpressionArguments */, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            function emitNewExpression(node) {
                emitTokenWithComment(103 /* NewKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitExpression(node.expression, parenthesizer.parenthesizeExpressionOfNew);
                emitTypeArguments(node, node.typeArguments);
                emitExpressionList(node, node.arguments, 18960 /* NewExpressionArguments */, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            function emitTaggedTemplateExpression(node) {
                var indirectCall = ts.getEmitFlags(node) & 536870912 /* IndirectCall */;
                if (indirectCall) {
                    writePunctuation("(");
                    writeLiteral("0");
                    writePunctuation(",");
                    writeSpace();
                }
                emitExpression(node.tag, parenthesizer.parenthesizeLeftSideOfAccess);
                if (indirectCall) {
                    writePunctuation(")");
                }
                emitTypeArguments(node, node.typeArguments);
                writeSpace();
                emitExpression(node.template);
            }
            function emitTypeAssertionExpression(node) {
                writePunctuation("<");
                emit(node.type);
                writePunctuation(">");
                emitExpression(node.expression, parenthesizer.parenthesizeOperandOfPrefixUnary);
            }
            function emitParenthesizedExpression(node) {
                var openParenPos = emitTokenWithComment(20 /* OpenParenToken */, node.pos, writePunctuation, node);
                var indented = writeLineSeparatorsAndIndentBefore(node.expression, node);
                emitExpression(node.expression, /*parenthesizerRules*/ undefined);
                writeLineSeparatorsAfter(node.expression, node);
                decreaseIndentIf(indented);
                emitTokenWithComment(21 /* CloseParenToken */, node.expression ? node.expression.end : openParenPos, writePunctuation, node);
            }
            function emitFunctionExpression(node) {
                generateNameIfNeeded(node.name);
                emitFunctionDeclarationOrExpression(node);
            }
            function emitArrowFunction(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                emitSignatureAndBody(node, emitArrowFunctionHead);
            }
            function emitArrowFunctionHead(node) {
                emitTypeParameters(node, node.typeParameters);
                emitParametersForArrow(node, node.parameters);
                emitTypeAnnotation(node.type);
                writeSpace();
                emit(node.equalsGreaterThanToken);
            }
            function emitDeleteExpression(node) {
                emitTokenWithComment(89 /* DeleteKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitExpression(node.expression, parenthesizer.parenthesizeOperandOfPrefixUnary);
            }
            function emitTypeOfExpression(node) {
                emitTokenWithComment(112 /* TypeOfKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitExpression(node.expression, parenthesizer.parenthesizeOperandOfPrefixUnary);
            }
            function emitVoidExpression(node) {
                emitTokenWithComment(114 /* VoidKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitExpression(node.expression, parenthesizer.parenthesizeOperandOfPrefixUnary);
            }
            function emitAwaitExpression(node) {
                emitTokenWithComment(131 /* AwaitKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitExpression(node.expression, parenthesizer.parenthesizeOperandOfPrefixUnary);
            }
            function emitPrefixUnaryExpression(node) {
                writeTokenText(node.operator, writeOperator);
                if (shouldEmitWhitespaceBeforeOperand(node)) {
                    writeSpace();
                }
                emitExpression(node.operand, parenthesizer.parenthesizeOperandOfPrefixUnary);
            }
            function shouldEmitWhitespaceBeforeOperand(node) {
                // In some cases, we need to emit a space between the operator and the operand. One obvious case
                // is when the operator is an identifier, like delete or typeof. We also need to do this for plus
                // and minus expressions in certain cases. Specifically, consider the following two cases (parens
                // are just for clarity of exposition, and not part of the source code):
                //
                //  (+(+1))
                //  (+(++1))
                //
                // We need to emit a space in both cases. In the first case, the absence of a space will make
                // the resulting expression a prefix increment operation. And in the second, it will make the resulting
                // expression a prefix increment whose operand is a plus expression - (++(+x))
                // The same is true of minus of course.
                var operand = node.operand;
                return operand.kind === 217 /* PrefixUnaryExpression */
                    && ((node.operator === 39 /* PlusToken */ && (operand.operator === 39 /* PlusToken */ || operand.operator === 45 /* PlusPlusToken */))
                        || (node.operator === 40 /* MinusToken */ && (operand.operator === 40 /* MinusToken */ || operand.operator === 46 /* MinusMinusToken */)));
            }
            function emitPostfixUnaryExpression(node) {
                emitExpression(node.operand, parenthesizer.parenthesizeOperandOfPostfixUnary);
                writeTokenText(node.operator, writeOperator);
            }
            function createEmitBinaryExpression() {
                return ts.createBinaryExpressionTrampoline(onEnter, onLeft, onOperator, onRight, onExit, /*foldState*/ undefined);
                function onEnter(node, state) {
                    if (state) {
                        state.stackIndex++;
                        state.preserveSourceNewlinesStack[state.stackIndex] = preserveSourceNewlines;
                        state.containerPosStack[state.stackIndex] = containerPos;
                        state.containerEndStack[state.stackIndex] = containerEnd;
                        state.declarationListContainerEndStack[state.stackIndex] = declarationListContainerEnd;
                        var emitComments_1 = state.shouldEmitCommentsStack[state.stackIndex] = shouldEmitComments(node);
                        var emitSourceMaps = state.shouldEmitSourceMapsStack[state.stackIndex] = shouldEmitSourceMaps(node);
                        onBeforeEmitNode === null || onBeforeEmitNode === void 0 ? void 0 : onBeforeEmitNode(node);
                        if (emitComments_1)
                            emitCommentsBeforeNode(node);
                        if (emitSourceMaps)
                            emitSourceMapsBeforeNode(node);
                        beforeEmitNode(node);
                    }
                    else {
                        state = {
                            stackIndex: 0,
                            preserveSourceNewlinesStack: [undefined],
                            containerPosStack: [-1],
                            containerEndStack: [-1],
                            declarationListContainerEndStack: [-1],
                            shouldEmitCommentsStack: [false],
                            shouldEmitSourceMapsStack: [false],
                        };
                    }
                    return state;
                }
                function onLeft(next, _workArea, parent) {
                    return maybeEmitExpression(next, parent, "left");
                }
                function onOperator(operatorToken, _state, node) {
                    var isCommaOperator = operatorToken.kind !== 27 /* CommaToken */;
                    var linesBeforeOperator = getLinesBetweenNodes(node, node.left, operatorToken);
                    var linesAfterOperator = getLinesBetweenNodes(node, operatorToken, node.right);
                    writeLinesAndIndent(linesBeforeOperator, isCommaOperator);
                    emitLeadingCommentsOfPosition(operatorToken.pos);
                    writeTokenNode(operatorToken, operatorToken.kind === 101 /* InKeyword */ ? writeKeyword : writeOperator);
                    emitTrailingCommentsOfPosition(operatorToken.end, /*prefixSpace*/ true); // Binary operators should have a space before the comment starts
                    writeLinesAndIndent(linesAfterOperator, /*writeSpaceIfNotIndenting*/ true);
                }
                function onRight(next, _workArea, parent) {
                    return maybeEmitExpression(next, parent, "right");
                }
                function onExit(node, state) {
                    var linesBeforeOperator = getLinesBetweenNodes(node, node.left, node.operatorToken);
                    var linesAfterOperator = getLinesBetweenNodes(node, node.operatorToken, node.right);
                    decreaseIndentIf(linesBeforeOperator, linesAfterOperator);
                    if (state.stackIndex > 0) {
                        var savedPreserveSourceNewlines = state.preserveSourceNewlinesStack[state.stackIndex];
                        var savedContainerPos = state.containerPosStack[state.stackIndex];
                        var savedContainerEnd = state.containerEndStack[state.stackIndex];
                        var savedDeclarationListContainerEnd = state.declarationListContainerEndStack[state.stackIndex];
                        var shouldEmitComments_1 = state.shouldEmitCommentsStack[state.stackIndex];
                        var shouldEmitSourceMaps_1 = state.shouldEmitSourceMapsStack[state.stackIndex];
                        afterEmitNode(savedPreserveSourceNewlines);
                        if (shouldEmitSourceMaps_1)
                            emitSourceMapsAfterNode(node);
                        if (shouldEmitComments_1)
                            emitCommentsAfterNode(node, savedContainerPos, savedContainerEnd, savedDeclarationListContainerEnd);
                        onAfterEmitNode === null || onAfterEmitNode === void 0 ? void 0 : onAfterEmitNode(node);
                        state.stackIndex--;
                    }
                }
                function maybeEmitExpression(next, parent, side) {
                    var parenthesizerRule = side === "left" ?
                        parenthesizer.getParenthesizeLeftSideOfBinaryForOperator(parent.operatorToken.kind) :
                        parenthesizer.getParenthesizeRightSideOfBinaryForOperator(parent.operatorToken.kind);
                    var pipelinePhase = getPipelinePhase(0 /* Notification */, 1 /* Expression */, next);
                    if (pipelinePhase === pipelineEmitWithSubstitution) {
                        ts.Debug.assertIsDefined(lastSubstitution);
                        next = parenthesizerRule(ts.cast(lastSubstitution, ts.isExpression));
                        pipelinePhase = getNextPipelinePhase(1 /* Substitution */, 1 /* Expression */, next);
                        lastSubstitution = undefined;
                    }
                    if (pipelinePhase === pipelineEmitWithComments ||
                        pipelinePhase === pipelineEmitWithSourceMaps ||
                        pipelinePhase === pipelineEmitWithHint) {
                        if (ts.isBinaryExpression(next)) {
                            return next;
                        }
                    }
                    currentParenthesizerRule = parenthesizerRule;
                    pipelinePhase(1 /* Expression */, next);
                }
            }
            function emitConditionalExpression(node) {
                var linesBeforeQuestion = getLinesBetweenNodes(node, node.condition, node.questionToken);
                var linesAfterQuestion = getLinesBetweenNodes(node, node.questionToken, node.whenTrue);
                var linesBeforeColon = getLinesBetweenNodes(node, node.whenTrue, node.colonToken);
                var linesAfterColon = getLinesBetweenNodes(node, node.colonToken, node.whenFalse);
                emitExpression(node.condition, parenthesizer.parenthesizeConditionOfConditionalExpression);
                writeLinesAndIndent(linesBeforeQuestion, /*writeSpaceIfNotIndenting*/ true);
                emit(node.questionToken);
                writeLinesAndIndent(linesAfterQuestion, /*writeSpaceIfNotIndenting*/ true);
                emitExpression(node.whenTrue, parenthesizer.parenthesizeBranchOfConditionalExpression);
                decreaseIndentIf(linesBeforeQuestion, linesAfterQuestion);
                writeLinesAndIndent(linesBeforeColon, /*writeSpaceIfNotIndenting*/ true);
                emit(node.colonToken);
                writeLinesAndIndent(linesAfterColon, /*writeSpaceIfNotIndenting*/ true);
                emitExpression(node.whenFalse, parenthesizer.parenthesizeBranchOfConditionalExpression);
                decreaseIndentIf(linesBeforeColon, linesAfterColon);
            }
            function emitTemplateExpression(node) {
                emit(node.head);
                emitList(node, node.templateSpans, 262144 /* TemplateExpressionSpans */);
            }
            function emitYieldExpression(node) {
                emitTokenWithComment(125 /* YieldKeyword */, node.pos, writeKeyword, node);
                emit(node.asteriskToken);
                emitExpressionWithLeadingSpace(node.expression, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            function emitSpreadElement(node) {
                emitTokenWithComment(25 /* DotDotDotToken */, node.pos, writePunctuation, node);
                emitExpression(node.expression, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            function emitClassExpression(node) {
                generateNameIfNeeded(node.name);
                emitClassDeclarationOrExpression(node);
            }
            function emitExpressionWithTypeArguments(node) {
                emitExpression(node.expression, parenthesizer.parenthesizeLeftSideOfAccess);
                emitTypeArguments(node, node.typeArguments);
            }
            function emitAsExpression(node) {
                emitExpression(node.expression, /*parenthesizerRules*/ undefined);
                if (node.type) {
                    writeSpace();
                    writeKeyword("as");
                    writeSpace();
                    emit(node.type);
                }
            }
            function emitNonNullExpression(node) {
                emitExpression(node.expression, parenthesizer.parenthesizeLeftSideOfAccess);
                writeOperator("!");
            }
            function emitMetaProperty(node) {
                writeToken(node.keywordToken, node.pos, writePunctuation);
                writePunctuation(".");
                emit(node.name);
            }
            //
            // Misc
            //
            function emitTemplateSpan(node) {
                emitExpression(node.expression);
                emit(node.literal);
            }
            //
            // Statements
            //
            function emitBlock(node) {
                emitBlockStatements(node, /*forceSingleLine*/ !node.multiLine && isEmptyBlock(node));
            }
            function emitBlockStatements(node, forceSingleLine) {
                emitTokenWithComment(18 /* OpenBraceToken */, node.pos, writePunctuation, /*contextNode*/ node);
                var format = forceSingleLine || ts.getEmitFlags(node) & 1 /* SingleLine */ ? 768 /* SingleLineBlockStatements */ : 129 /* MultiLineBlockStatements */;
                emitList(node, node.statements, format);
                emitTokenWithComment(19 /* CloseBraceToken */, node.statements.end, writePunctuation, /*contextNode*/ node, /*indentLeading*/ !!(format & 1 /* MultiLine */));
            }
            function emitVariableStatement(node) {
                emitModifiers(node, node.modifiers);
                emit(node.declarationList);
                writeTrailingSemicolon();
            }
            function emitEmptyStatement(isEmbeddedStatement) {
                // While most trailing semicolons are possibly insignificant, an embedded "empty"
                // statement is significant and cannot be elided by a trailing-semicolon-omitting writer.
                if (isEmbeddedStatement) {
                    writePunctuation(";");
                }
                else {
                    writeTrailingSemicolon();
                }
            }
            function emitExpressionStatement(node) {
                emitExpression(node.expression, parenthesizer.parenthesizeExpressionOfExpressionStatement);
                // Emit semicolon in non json files
                // or if json file that created synthesized expression(eg.define expression statement when --out and amd code generation)
                if (!ts.isJsonSourceFile(currentSourceFile) || ts.nodeIsSynthesized(node.expression)) {
                    writeTrailingSemicolon();
                }
            }
            function emitIfStatement(node) {
                var openParenPos = emitTokenWithComment(99 /* IfKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitTokenWithComment(20 /* OpenParenToken */, openParenPos, writePunctuation, node);
                emitExpression(node.expression);
                emitTokenWithComment(21 /* CloseParenToken */, node.expression.end, writePunctuation, node);
                emitEmbeddedStatement(node, node.thenStatement);
                if (node.elseStatement) {
                    writeLineOrSpace(node, node.thenStatement, node.elseStatement);
                    emitTokenWithComment(91 /* ElseKeyword */, node.thenStatement.end, writeKeyword, node);
                    if (node.elseStatement.kind === 237 /* IfStatement */) {
                        writeSpace();
                        emit(node.elseStatement);
                    }
                    else {
                        emitEmbeddedStatement(node, node.elseStatement);
                    }
                }
            }
            function emitWhileClause(node, startPos) {
                var openParenPos = emitTokenWithComment(115 /* WhileKeyword */, startPos, writeKeyword, node);
                writeSpace();
                emitTokenWithComment(20 /* OpenParenToken */, openParenPos, writePunctuation, node);
                emitExpression(node.expression);
                emitTokenWithComment(21 /* CloseParenToken */, node.expression.end, writePunctuation, node);
            }
            function emitDoStatement(node) {
                emitTokenWithComment(90 /* DoKeyword */, node.pos, writeKeyword, node);
                emitEmbeddedStatement(node, node.statement);
                if (ts.isBlock(node.statement) && !preserveSourceNewlines) {
                    writeSpace();
                }
                else {
                    writeLineOrSpace(node, node.statement, node.expression);
                }
                emitWhileClause(node, node.statement.end);
                writeTrailingSemicolon();
            }
            function emitWhileStatement(node) {
                emitWhileClause(node, node.pos);
                emitEmbeddedStatement(node, node.statement);
            }
            function emitForStatement(node) {
                var openParenPos = emitTokenWithComment(97 /* ForKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                var pos = emitTokenWithComment(20 /* OpenParenToken */, openParenPos, writePunctuation, /*contextNode*/ node);
                emitForBinding(node.initializer);
                pos = emitTokenWithComment(26 /* SemicolonToken */, node.initializer ? node.initializer.end : pos, writePunctuation, node);
                emitExpressionWithLeadingSpace(node.condition);
                pos = emitTokenWithComment(26 /* SemicolonToken */, node.condition ? node.condition.end : pos, writePunctuation, node);
                emitExpressionWithLeadingSpace(node.incrementor);
                emitTokenWithComment(21 /* CloseParenToken */, node.incrementor ? node.incrementor.end : pos, writePunctuation, node);
                emitEmbeddedStatement(node, node.statement);
            }
            function emitForInStatement(node) {
                var openParenPos = emitTokenWithComment(97 /* ForKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitTokenWithComment(20 /* OpenParenToken */, openParenPos, writePunctuation, node);
                emitForBinding(node.initializer);
                writeSpace();
                emitTokenWithComment(101 /* InKeyword */, node.initializer.end, writeKeyword, node);
                writeSpace();
                emitExpression(node.expression);
                emitTokenWithComment(21 /* CloseParenToken */, node.expression.end, writePunctuation, node);
                emitEmbeddedStatement(node, node.statement);
            }
            function emitForOfStatement(node) {
                var openParenPos = emitTokenWithComment(97 /* ForKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitWithTrailingSpace(node.awaitModifier);
                emitTokenWithComment(20 /* OpenParenToken */, openParenPos, writePunctuation, node);
                emitForBinding(node.initializer);
                writeSpace();
                emitTokenWithComment(158 /* OfKeyword */, node.initializer.end, writeKeyword, node);
                writeSpace();
                emitExpression(node.expression);
                emitTokenWithComment(21 /* CloseParenToken */, node.expression.end, writePunctuation, node);
                emitEmbeddedStatement(node, node.statement);
            }
            function emitForBinding(node) {
                if (node !== undefined) {
                    if (node.kind === 253 /* VariableDeclarationList */) {
                        emit(node);
                    }
                    else {
                        emitExpression(node);
                    }
                }
            }
            function emitContinueStatement(node) {
                emitTokenWithComment(86 /* ContinueKeyword */, node.pos, writeKeyword, node);
                emitWithLeadingSpace(node.label);
                writeTrailingSemicolon();
            }
            function emitBreakStatement(node) {
                emitTokenWithComment(81 /* BreakKeyword */, node.pos, writeKeyword, node);
                emitWithLeadingSpace(node.label);
                writeTrailingSemicolon();
            }
            function emitTokenWithComment(token, pos, writer, contextNode, indentLeading) {
                var node = ts.getParseTreeNode(contextNode);
                var isSimilarNode = node && node.kind === contextNode.kind;
                var startPos = pos;
                if (isSimilarNode && currentSourceFile) {
                    pos = ts.skipTrivia(currentSourceFile.text, pos);
                }
                if (isSimilarNode && contextNode.pos !== startPos) {
                    var needsIndent = indentLeading && currentSourceFile && !ts.positionsAreOnSameLine(startPos, pos, currentSourceFile);
                    if (needsIndent) {
                        increaseIndent();
                    }
                    emitLeadingCommentsOfPosition(startPos);
                    if (needsIndent) {
                        decreaseIndent();
                    }
                }
                pos = writeTokenText(token, writer, pos);
                if (isSimilarNode && contextNode.end !== pos) {
                    var isJsxExprContext = contextNode.kind === 286 /* JsxExpression */;
                    emitTrailingCommentsOfPosition(pos, /*prefixSpace*/ !isJsxExprContext, /*forceNoNewline*/ isJsxExprContext);
                }
                return pos;
            }
            function emitReturnStatement(node) {
                emitTokenWithComment(105 /* ReturnKeyword */, node.pos, writeKeyword, /*contextNode*/ node);
                emitExpressionWithLeadingSpace(node.expression);
                writeTrailingSemicolon();
            }
            function emitWithStatement(node) {
                var openParenPos = emitTokenWithComment(116 /* WithKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitTokenWithComment(20 /* OpenParenToken */, openParenPos, writePunctuation, node);
                emitExpression(node.expression);
                emitTokenWithComment(21 /* CloseParenToken */, node.expression.end, writePunctuation, node);
                emitEmbeddedStatement(node, node.statement);
            }
            function emitSwitchStatement(node) {
                var openParenPos = emitTokenWithComment(107 /* SwitchKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitTokenWithComment(20 /* OpenParenToken */, openParenPos, writePunctuation, node);
                emitExpression(node.expression);
                emitTokenWithComment(21 /* CloseParenToken */, node.expression.end, writePunctuation, node);
                writeSpace();
                emit(node.caseBlock);
            }
            function emitLabeledStatement(node) {
                emit(node.label);
                emitTokenWithComment(58 /* ColonToken */, node.label.end, writePunctuation, node);
                writeSpace();
                emit(node.statement);
            }
            function emitThrowStatement(node) {
                emitTokenWithComment(109 /* ThrowKeyword */, node.pos, writeKeyword, node);
                emitExpressionWithLeadingSpace(node.expression);
                writeTrailingSemicolon();
            }
            function emitTryStatement(node) {
                emitTokenWithComment(111 /* TryKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emit(node.tryBlock);
                if (node.catchClause) {
                    writeLineOrSpace(node, node.tryBlock, node.catchClause);
                    emit(node.catchClause);
                }
                if (node.finallyBlock) {
                    writeLineOrSpace(node, node.catchClause || node.tryBlock, node.finallyBlock);
                    emitTokenWithComment(96 /* FinallyKeyword */, (node.catchClause || node.tryBlock).end, writeKeyword, node);
                    writeSpace();
                    emit(node.finallyBlock);
                }
            }
            function emitDebuggerStatement(node) {
                writeToken(87 /* DebuggerKeyword */, node.pos, writeKeyword);
                writeTrailingSemicolon();
            }
            //
            // Declarations
            //
            function emitVariableDeclaration(node) {
                emit(node.name);
                emit(node.exclamationToken);
                emitTypeAnnotation(node.type);
                emitInitializer(node.initializer, node.type ? node.type.end : node.name.end, node, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            function emitVariableDeclarationList(node) {
                writeKeyword(ts.isLet(node) ? "let" : ts.isVarConst(node) ? "const" : "var");
                writeSpace();
                emitList(node, node.declarations, 528 /* VariableDeclarationList */);
            }
            function emitFunctionDeclaration(node) {
                emitFunctionDeclarationOrExpression(node);
            }
            function emitFunctionDeclarationOrExpression(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                writeKeyword("function");
                emit(node.asteriskToken);
                writeSpace();
                emitIdentifierName(node.name);
                emitSignatureAndBody(node, emitSignatureHead);
            }
            function emitSignatureAndBody(node, emitSignatureHead) {
                var body = node.body;
                if (body) {
                    if (ts.isBlock(body)) {
                        var indentedFlag = ts.getEmitFlags(node) & 65536 /* Indented */;
                        if (indentedFlag) {
                            increaseIndent();
                        }
                        pushNameGenerationScope(node);
                        ts.forEach(node.parameters, generateNames);
                        generateNames(node.body);
                        emitSignatureHead(node);
                        emitBlockFunctionBody(body);
                        popNameGenerationScope(node);
                        if (indentedFlag) {
                            decreaseIndent();
                        }
                    }
                    else {
                        emitSignatureHead(node);
                        writeSpace();
                        emitExpression(body, parenthesizer.parenthesizeConciseBodyOfArrowFunction);
                    }
                }
                else {
                    emitSignatureHead(node);
                    writeTrailingSemicolon();
                }
            }
            function emitSignatureHead(node) {
                emitTypeParameters(node, node.typeParameters);
                emitParameters(node, node.parameters);
                emitTypeAnnotation(node.type);
            }
            function shouldEmitBlockFunctionBodyOnSingleLine(body) {
                // We must emit a function body as a single-line body in the following case:
                // * The body has NodeEmitFlags.SingleLine specified.
                // We must emit a function body as a multi-line body in the following cases:
                // * The body is explicitly marked as multi-line.
                // * A non-synthesized body's start and end position are on different lines.
                // * Any statement in the body starts on a new line.
                if (ts.getEmitFlags(body) & 1 /* SingleLine */) {
                    return true;
                }
                if (body.multiLine) {
                    return false;
                }
                if (!ts.nodeIsSynthesized(body) && !ts.rangeIsOnSingleLine(body, currentSourceFile)) {
                    return false;
                }
                if (getLeadingLineTerminatorCount(body, body.statements, 2 /* PreserveLines */)
                    || getClosingLineTerminatorCount(body, body.statements, 2 /* PreserveLines */)) {
                    return false;
                }
                var previousStatement;
                for (var _a = 0, _b = body.statements; _a < _b.length; _a++) {
                    var statement = _b[_a];
                    if (getSeparatingLineTerminatorCount(previousStatement, statement, 2 /* PreserveLines */) > 0) {
                        return false;
                    }
                    previousStatement = statement;
                }
                return true;
            }
            function emitBlockFunctionBody(body) {
                onBeforeEmitNode === null || onBeforeEmitNode === void 0 ? void 0 : onBeforeEmitNode(body);
                writeSpace();
                writePunctuation("{");
                increaseIndent();
                var emitBlockFunctionBody = shouldEmitBlockFunctionBodyOnSingleLine(body)
                    ? emitBlockFunctionBodyOnSingleLine
                    : emitBlockFunctionBodyWorker;
                if (emitBodyWithDetachedComments) {
                    emitBodyWithDetachedComments(body, body.statements, emitBlockFunctionBody);
                }
                else {
                    emitBlockFunctionBody(body);
                }
                decreaseIndent();
                writeToken(19 /* CloseBraceToken */, body.statements.end, writePunctuation, body);
                onAfterEmitNode === null || onAfterEmitNode === void 0 ? void 0 : onAfterEmitNode(body);
            }
            function emitBlockFunctionBodyOnSingleLine(body) {
                emitBlockFunctionBodyWorker(body, /*emitBlockFunctionBodyOnSingleLine*/ true);
            }
            function emitBlockFunctionBodyWorker(body, emitBlockFunctionBodyOnSingleLine) {
                // Emit all the prologue directives (like "use strict").
                var statementOffset = emitPrologueDirectives(body.statements);
                var pos = writer.getTextPos();
                emitHelpers(body);
                if (statementOffset === 0 && pos === writer.getTextPos() && emitBlockFunctionBodyOnSingleLine) {
                    decreaseIndent();
                    emitList(body, body.statements, 768 /* SingleLineFunctionBodyStatements */);
                    increaseIndent();
                }
                else {
                    emitList(body, body.statements, 1 /* MultiLineFunctionBodyStatements */, /*parenthesizerRule*/ undefined, statementOffset);
                }
            }
            function emitClassDeclaration(node) {
                emitClassDeclarationOrExpression(node);
            }
            function emitClassDeclarationOrExpression(node) {
                ts.forEach(node.members, generateMemberNames);
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                writeKeyword("class");
                if (node.name) {
                    writeSpace();
                    emitIdentifierName(node.name);
                }
                var indentedFlag = ts.getEmitFlags(node) & 65536 /* Indented */;
                if (indentedFlag) {
                    increaseIndent();
                }
                emitTypeParameters(node, node.typeParameters);
                emitList(node, node.heritageClauses, 0 /* ClassHeritageClauses */);
                writeSpace();
                writePunctuation("{");
                emitList(node, node.members, 129 /* ClassMembers */);
                writePunctuation("}");
                if (indentedFlag) {
                    decreaseIndent();
                }
            }
            function emitInterfaceDeclaration(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                writeKeyword("interface");
                writeSpace();
                emit(node.name);
                emitTypeParameters(node, node.typeParameters);
                emitList(node, node.heritageClauses, 512 /* HeritageClauses */);
                writeSpace();
                writePunctuation("{");
                emitList(node, node.members, 129 /* InterfaceMembers */);
                writePunctuation("}");
            }
            function emitTypeAliasDeclaration(node) {
                emitDecorators(node, node.decorators);
                emitModifiers(node, node.modifiers);
                writeKeyword("type");
                writeSpace();
                emit(node.name);
                emitTypeParameters(node, node.typeParameters);
                writeSpace();
                writePunctuation("=");
                writeSpace();
                emit(node.type);
                writeTrailingSemicolon();
            }
            function emitEnumDeclaration(node) {
                emitModifiers(node, node.modifiers);
                writeKeyword("enum");
                writeSpace();
                emit(node.name);
                writeSpace();
                writePunctuation("{");
                emitList(node, node.members, 145 /* EnumMembers */);
                writePunctuation("}");
            }
            function emitModuleDeclaration(node) {
                emitModifiers(node, node.modifiers);
                if (~node.flags & 1024 /* GlobalAugmentation */) {
                    writeKeyword(node.flags & 16 /* Namespace */ ? "namespace" : "module");
                    writeSpace();
                }
                emit(node.name);
                var body = node.body;
                if (!body)
                    return writeTrailingSemicolon();
                while (body && ts.isModuleDeclaration(body)) {
                    writePunctuation(".");
                    emit(body.name);
                    body = body.body;
                }
                writeSpace();
                emit(body);
            }
            function emitModuleBlock(node) {
                pushNameGenerationScope(node);
                ts.forEach(node.statements, generateNames);
                emitBlockStatements(node, /*forceSingleLine*/ isEmptyBlock(node));
                popNameGenerationScope(node);
            }
            function emitCaseBlock(node) {
                emitTokenWithComment(18 /* OpenBraceToken */, node.pos, writePunctuation, node);
                emitList(node, node.clauses, 129 /* CaseBlockClauses */);
                emitTokenWithComment(19 /* CloseBraceToken */, node.clauses.end, writePunctuation, node, /*indentLeading*/ true);
            }
            function emitImportEqualsDeclaration(node) {
                emitModifiers(node, node.modifiers);
                emitTokenWithComment(100 /* ImportKeyword */, node.modifiers ? node.modifiers.end : node.pos, writeKeyword, node);
                writeSpace();
                if (node.isTypeOnly) {
                    emitTokenWithComment(150 /* TypeKeyword */, node.pos, writeKeyword, node);
                    writeSpace();
                }
                emit(node.name);
                writeSpace();
                emitTokenWithComment(63 /* EqualsToken */, node.name.end, writePunctuation, node);
                writeSpace();
                emitModuleReference(node.moduleReference);
                writeTrailingSemicolon();
            }
            function emitModuleReference(node) {
                if (node.kind === 79 /* Identifier */) {
                    emitExpression(node);
                }
                else {
                    emit(node);
                }
            }
            function emitImportDeclaration(node) {
                emitModifiers(node, node.modifiers);
                emitTokenWithComment(100 /* ImportKeyword */, node.modifiers ? node.modifiers.end : node.pos, writeKeyword, node);
                writeSpace();
                if (node.importClause) {
                    emit(node.importClause);
                    writeSpace();
                    emitTokenWithComment(154 /* FromKeyword */, node.importClause.end, writeKeyword, node);
                    writeSpace();
                }
                emitExpression(node.moduleSpecifier);
                writeTrailingSemicolon();
            }
            function emitImportClause(node) {
                if (node.isTypeOnly) {
                    emitTokenWithComment(150 /* TypeKeyword */, node.pos, writeKeyword, node);
                    writeSpace();
                }
                emit(node.name);
                if (node.name && node.namedBindings) {
                    emitTokenWithComment(27 /* CommaToken */, node.name.end, writePunctuation, node);
                    writeSpace();
                }
                emit(node.namedBindings);
            }
            function emitNamespaceImport(node) {
                var asPos = emitTokenWithComment(41 /* AsteriskToken */, node.pos, writePunctuation, node);
                writeSpace();
                emitTokenWithComment(127 /* AsKeyword */, asPos, writeKeyword, node);
                writeSpace();
                emit(node.name);
            }
            function emitNamedImports(node) {
                emitNamedImportsOrExports(node);
            }
            function emitImportSpecifier(node) {
                emitImportOrExportSpecifier(node);
            }
            function emitExportAssignment(node) {
                var nextPos = emitTokenWithComment(93 /* ExportKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                if (node.isExportEquals) {
                    emitTokenWithComment(63 /* EqualsToken */, nextPos, writeOperator, node);
                }
                else {
                    emitTokenWithComment(88 /* DefaultKeyword */, nextPos, writeKeyword, node);
                }
                writeSpace();
                emitExpression(node.expression, node.isExportEquals ?
                    parenthesizer.getParenthesizeRightSideOfBinaryForOperator(63 /* EqualsToken */) :
                    parenthesizer.parenthesizeExpressionOfExportDefault);
                writeTrailingSemicolon();
            }
            function emitExportDeclaration(node) {
                var nextPos = emitTokenWithComment(93 /* ExportKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                if (node.isTypeOnly) {
                    nextPos = emitTokenWithComment(150 /* TypeKeyword */, nextPos, writeKeyword, node);
                    writeSpace();
                }
                if (node.exportClause) {
                    emit(node.exportClause);
                }
                else {
                    nextPos = emitTokenWithComment(41 /* AsteriskToken */, nextPos, writePunctuation, node);
                }
                if (node.moduleSpecifier) {
                    writeSpace();
                    var fromPos = node.exportClause ? node.exportClause.end : nextPos;
                    emitTokenWithComment(154 /* FromKeyword */, fromPos, writeKeyword, node);
                    writeSpace();
                    emitExpression(node.moduleSpecifier);
                }
                writeTrailingSemicolon();
            }
            function emitNamespaceExportDeclaration(node) {
                var nextPos = emitTokenWithComment(93 /* ExportKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                nextPos = emitTokenWithComment(127 /* AsKeyword */, nextPos, writeKeyword, node);
                writeSpace();
                nextPos = emitTokenWithComment(141 /* NamespaceKeyword */, nextPos, writeKeyword, node);
                writeSpace();
                emit(node.name);
                writeTrailingSemicolon();
            }
            function emitNamespaceExport(node) {
                var asPos = emitTokenWithComment(41 /* AsteriskToken */, node.pos, writePunctuation, node);
                writeSpace();
                emitTokenWithComment(127 /* AsKeyword */, asPos, writeKeyword, node);
                writeSpace();
                emit(node.name);
            }
            function emitNamedExports(node) {
                emitNamedImportsOrExports(node);
            }
            function emitExportSpecifier(node) {
                emitImportOrExportSpecifier(node);
            }
            function emitNamedImportsOrExports(node) {
                writePunctuation("{");
                emitList(node, node.elements, 525136 /* NamedImportsOrExportsElements */);
                writePunctuation("}");
            }
            function emitImportOrExportSpecifier(node) {
                if (node.propertyName) {
                    emit(node.propertyName);
                    writeSpace();
                    emitTokenWithComment(127 /* AsKeyword */, node.propertyName.end, writeKeyword, node);
                    writeSpace();
                }
                emit(node.name);
            }
            //
            // Module references
            //
            function emitExternalModuleReference(node) {
                writeKeyword("require");
                writePunctuation("(");
                emitExpression(node.expression);
                writePunctuation(")");
            }
            //
            // JSX
            //
            function emitJsxElement(node) {
                emit(node.openingElement);
                emitList(node, node.children, 262144 /* JsxElementOrFragmentChildren */);
                emit(node.closingElement);
            }
            function emitJsxSelfClosingElement(node) {
                writePunctuation("<");
                emitJsxTagName(node.tagName);
                emitTypeArguments(node, node.typeArguments);
                writeSpace();
                emit(node.attributes);
                writePunctuation("/>");
            }
            function emitJsxFragment(node) {
                emit(node.openingFragment);
                emitList(node, node.children, 262144 /* JsxElementOrFragmentChildren */);
                emit(node.closingFragment);
            }
            function emitJsxOpeningElementOrFragment(node) {
                writePunctuation("<");
                if (ts.isJsxOpeningElement(node)) {
                    var indented = writeLineSeparatorsAndIndentBefore(node.tagName, node);
                    emitJsxTagName(node.tagName);
                    emitTypeArguments(node, node.typeArguments);
                    if (node.attributes.properties && node.attributes.properties.length > 0) {
                        writeSpace();
                    }
                    emit(node.attributes);
                    writeLineSeparatorsAfter(node.attributes, node);
                    decreaseIndentIf(indented);
                }
                writePunctuation(">");
            }
            function emitJsxText(node) {
                writer.writeLiteral(node.text);
            }
            function emitJsxClosingElementOrFragment(node) {
                writePunctuation("</");
                if (ts.isJsxClosingElement(node)) {
                    emitJsxTagName(node.tagName);
                }
                writePunctuation(">");
            }
            function emitJsxAttributes(node) {
                emitList(node, node.properties, 262656 /* JsxElementAttributes */);
            }
            function emitJsxAttribute(node) {
                emit(node.name);
                emitNodeWithPrefix("=", writePunctuation, node.initializer, emitJsxAttributeValue);
            }
            function emitJsxSpreadAttribute(node) {
                writePunctuation("{...");
                emitExpression(node.expression);
                writePunctuation("}");
            }
            function hasTrailingCommentsAtPosition(pos) {
                var result = false;
                ts.forEachTrailingCommentRange((currentSourceFile === null || currentSourceFile === void 0 ? void 0 : currentSourceFile.text) || "", pos + 1, function () { return result = true; });
                return result;
            }
            function hasLeadingCommentsAtPosition(pos) {
                var result = false;
                ts.forEachLeadingCommentRange((currentSourceFile === null || currentSourceFile === void 0 ? void 0 : currentSourceFile.text) || "", pos + 1, function () { return result = true; });
                return result;
            }
            function hasCommentsAtPosition(pos) {
                return hasTrailingCommentsAtPosition(pos) || hasLeadingCommentsAtPosition(pos);
            }
            function emitJsxExpression(node) {
                var _a;
                if (node.expression || (!commentsDisabled && !ts.nodeIsSynthesized(node) && hasCommentsAtPosition(node.pos))) { // preserve empty expressions if they contain comments!
                    var isMultiline = currentSourceFile && !ts.nodeIsSynthesized(node) && ts.getLineAndCharacterOfPosition(currentSourceFile, node.pos).line !== ts.getLineAndCharacterOfPosition(currentSourceFile, node.end).line;
                    if (isMultiline) {
                        writer.increaseIndent();
                    }
                    var end = emitTokenWithComment(18 /* OpenBraceToken */, node.pos, writePunctuation, node);
                    emit(node.dotDotDotToken);
                    emitExpression(node.expression);
                    emitTokenWithComment(19 /* CloseBraceToken */, ((_a = node.expression) === null || _a === void 0 ? void 0 : _a.end) || end, writePunctuation, node);
                    if (isMultiline) {
                        writer.decreaseIndent();
                    }
                }
            }
            function emitJsxTagName(node) {
                if (node.kind === 79 /* Identifier */) {
                    emitExpression(node);
                }
                else {
                    emit(node);
                }
            }
            //
            // Clauses
            //
            function emitCaseClause(node) {
                emitTokenWithComment(82 /* CaseKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                emitExpression(node.expression, parenthesizer.parenthesizeExpressionForDisallowedComma);
                emitCaseOrDefaultClauseRest(node, node.statements, node.expression.end);
            }
            function emitDefaultClause(node) {
                var pos = emitTokenWithComment(88 /* DefaultKeyword */, node.pos, writeKeyword, node);
                emitCaseOrDefaultClauseRest(node, node.statements, pos);
            }
            function emitCaseOrDefaultClauseRest(parentNode, statements, colonPos) {
                var emitAsSingleStatement = statements.length === 1 &&
                    (
                    // treat synthesized nodes as located on the same line for emit purposes
                    ts.nodeIsSynthesized(parentNode) ||
                        ts.nodeIsSynthesized(statements[0]) ||
                        ts.rangeStartPositionsAreOnSameLine(parentNode, statements[0], currentSourceFile));
                var format = 163969 /* CaseOrDefaultClauseStatements */;
                if (emitAsSingleStatement) {
                    writeToken(58 /* ColonToken */, colonPos, writePunctuation, parentNode);
                    writeSpace();
                    format &= ~(1 /* MultiLine */ | 128 /* Indented */);
                }
                else {
                    emitTokenWithComment(58 /* ColonToken */, colonPos, writePunctuation, parentNode);
                }
                emitList(parentNode, statements, format);
            }
            function emitHeritageClause(node) {
                writeSpace();
                writeTokenText(node.token, writeKeyword);
                writeSpace();
                emitList(node, node.types, 528 /* HeritageClauseTypes */);
            }
            function emitCatchClause(node) {
                var openParenPos = emitTokenWithComment(83 /* CatchKeyword */, node.pos, writeKeyword, node);
                writeSpace();
                if (node.variableDeclaration) {
                    emitTokenWithComment(20 /* OpenParenToken */, openParenPos, writePunctuation, node);
                    emit(node.variableDeclaration);
                    emitTokenWithComment(21 /* CloseParenToken */, node.variableDeclaration.end, writePunctuation, node);
                    writeSpace();
                }
                emit(node.block);
            }
            //
            // Property assignments
            //
            function emitPropertyAssignment(node) {
                emit(node.name);
                writePunctuation(":");
                writeSpace();
                // This is to ensure that we emit comment in the following case:
                //      For example:
                //          obj = {
                //              id: /*comment1*/ ()=>void
                //          }
                // "comment1" is not considered to be leading comment for node.initializer
                // but rather a trailing comment on the previous node.
                var initializer = node.initializer;
                if ((ts.getEmitFlags(initializer) & 512 /* NoLeadingComments */) === 0) {
                    var commentRange = ts.getCommentRange(initializer);
                    emitTrailingCommentsOfPosition(commentRange.pos);
                }
                emitExpression(initializer, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            function emitShorthandPropertyAssignment(node) {
                emit(node.name);
                if (node.objectAssignmentInitializer) {
                    writeSpace();
                    writePunctuation("=");
                    writeSpace();
                    emitExpression(node.objectAssignmentInitializer, parenthesizer.parenthesizeExpressionForDisallowedComma);
                }
            }
            function emitSpreadAssignment(node) {
                if (node.expression) {
                    emitTokenWithComment(25 /* DotDotDotToken */, node.pos, writePunctuation, node);
                    emitExpression(node.expression, parenthesizer.parenthesizeExpressionForDisallowedComma);
                }
            }
            //
            // Enum
            //
            function emitEnumMember(node) {
                emit(node.name);
                emitInitializer(node.initializer, node.name.end, node, parenthesizer.parenthesizeExpressionForDisallowedComma);
            }
            //
            // JSDoc
            //
            function emitJSDoc(node) {
                write("/**");
                if (node.comment) {
                    var text = ts.getTextOfJSDocComment(node.comment);
                    if (text) {
                        var lines = text.split(/\r\n?|\n/g);
                        for (var _a = 0, lines_2 = lines; _a < lines_2.length; _a++) {
                            var line = lines_2[_a];
                            writeLine();
                            writeSpace();
                            writePunctuation("*");
                            writeSpace();
                            write(line);
                        }
                    }
                }
                if (node.tags) {
                    if (node.tags.length === 1 && node.tags[0].kind === 338 /* JSDocTypeTag */ && !node.comment) {
                        writeSpace();
                        emit(node.tags[0]);
                    }
                    else {
                        emitList(node, node.tags, 33 /* JSDocComment */);
                    }
                }
                writeSpace();
                write("*/");
            }
            function emitJSDocSimpleTypedTag(tag) {
                emitJSDocTagName(tag.tagName);
                emitJSDocTypeExpression(tag.typeExpression);
                emitJSDocComment(tag.comment);
            }
            function emitJSDocSeeTag(tag) {
                emitJSDocTagName(tag.tagName);
                emit(tag.name);
                emitJSDocComment(tag.comment);
            }
            function emitJSDocNameReference(node) {
                writeSpace();
                writePunctuation("{");
                emit(node.name);
                writePunctuation("}");
            }
            function emitJSDocHeritageTag(tag) {
                emitJSDocTagName(tag.tagName);
                writeSpace();
                writePunctuation("{");
                emit(tag.class);
                writePunctuation("}");
                emitJSDocComment(tag.comment);
            }
            function emitJSDocTemplateTag(tag) {
                emitJSDocTagName(tag.tagName);
                emitJSDocTypeExpression(tag.constraint);
                writeSpace();
                emitList(tag, tag.typeParameters, 528 /* CommaListElements */);
                emitJSDocComment(tag.comment);
            }
            function emitJSDocTypedefTag(tag) {
                emitJSDocTagName(tag.tagName);
                if (tag.typeExpression) {
                    if (tag.typeExpression.kind === 304 /* JSDocTypeExpression */) {
                        emitJSDocTypeExpression(tag.typeExpression);
                    }
                    else {
                        writeSpace();
                        writePunctuation("{");
                        write("Object");
                        if (tag.typeExpression.isArrayType) {
                            writePunctuation("[");
                            writePunctuation("]");
                        }
                        writePunctuation("}");
                    }
                }
                if (tag.fullName) {
                    writeSpace();
                    emit(tag.fullName);
                }
                emitJSDocComment(tag.comment);
                if (tag.typeExpression && tag.typeExpression.kind === 317 /* JSDocTypeLiteral */) {
                    emitJSDocTypeLiteral(tag.typeExpression);
                }
            }
            function emitJSDocCallbackTag(tag) {
                emitJSDocTagName(tag.tagName);
                if (tag.name) {
                    writeSpace();
                    emit(tag.name);
                }
                emitJSDocComment(tag.comment);
                emitJSDocSignature(tag.typeExpression);
            }
            function emitJSDocSimpleTag(tag) {
                emitJSDocTagName(tag.tagName);
                emitJSDocComment(tag.comment);
            }
            function emitJSDocTypeLiteral(lit) {
                emitList(lit, ts.factory.createNodeArray(lit.jsDocPropertyTags), 33 /* JSDocComment */);
            }
            function emitJSDocSignature(sig) {
                if (sig.typeParameters) {
                    emitList(sig, ts.factory.createNodeArray(sig.typeParameters), 33 /* JSDocComment */);
                }
                if (sig.parameters) {
                    emitList(sig, ts.factory.createNodeArray(sig.parameters), 33 /* JSDocComment */);
                }
                if (sig.type) {
                    writeLine();
                    writeSpace();
                    writePunctuation("*");
                    writeSpace();
                    emit(sig.type);
                }
            }
            function emitJSDocPropertyLikeTag(param) {
                emitJSDocTagName(param.tagName);
                emitJSDocTypeExpression(param.typeExpression);
                writeSpace();
                if (param.isBracketed) {
                    writePunctuation("[");
                }
                emit(param.name);
                if (param.isBracketed) {
                    writePunctuation("]");
                }
                emitJSDocComment(param.comment);
            }
            function emitJSDocTagName(tagName) {
                writePunctuation("@");
                emit(tagName);
            }
            function emitJSDocComment(comment) {
                var text = ts.getTextOfJSDocComment(comment);
                if (text) {
                    writeSpace();
                    write(text);
                }
            }
            function emitJSDocTypeExpression(typeExpression) {
                if (typeExpression) {
                    writeSpace();
                    writePunctuation("{");
                    emit(typeExpression.type);
                    writePunctuation("}");
                }
            }
            //
            // Top-level nodes
            //
            function emitSourceFile(node) {
                writeLine();
                var statements = node.statements;
                if (emitBodyWithDetachedComments) {
                    // Emit detached comment if there are no prologue directives or if the first node is synthesized.
                    // The synthesized node will have no leading comment so some comments may be missed.
                    var shouldEmitDetachedComment = statements.length === 0 ||
                        !ts.isPrologueDirective(statements[0]) ||
                        ts.nodeIsSynthesized(statements[0]);
                    if (shouldEmitDetachedComment) {
                        emitBodyWithDetachedComments(node, statements, emitSourceFileWorker);
                        return;
                    }
                }
                emitSourceFileWorker(node);
            }
            function emitSyntheticTripleSlashReferencesIfNeeded(node) {
                emitTripleSlashDirectives(!!node.hasNoDefaultLib, node.syntheticFileReferences || [], node.syntheticTypeReferences || [], node.syntheticLibReferences || []);
                for (var _a = 0, _b = node.prepends; _a < _b.length; _a++) {
                    var prepend = _b[_a];
                    if (ts.isUnparsedSource(prepend) && prepend.syntheticReferences) {
                        for (var _c = 0, _d = prepend.syntheticReferences; _c < _d.length; _c++) {
                            var ref = _d[_c];
                            emit(ref);
                            writeLine();
                        }
                    }
                }
            }
            function emitTripleSlashDirectivesIfNeeded(node) {
                if (node.isDeclarationFile)
                    emitTripleSlashDirectives(node.hasNoDefaultLib, node.referencedFiles, node.typeReferenceDirectives, node.libReferenceDirectives);
            }
            function emitTripleSlashDirectives(hasNoDefaultLib, files, types, libs) {
                if (hasNoDefaultLib) {
                    var pos = writer.getTextPos();
                    writeComment("/// <reference no-default-lib=\"true\"/>");
                    if (bundleFileInfo)
                        bundleFileInfo.sections.push({ pos: pos, end: writer.getTextPos(), kind: "no-default-lib" /* NoDefaultLib */ });
                    writeLine();
                }
                if (currentSourceFile && currentSourceFile.moduleName) {
                    writeComment("/// <amd-module name=\"" + currentSourceFile.moduleName + "\" />");
                    writeLine();
                }
                if (currentSourceFile && currentSourceFile.amdDependencies) {
                    for (var _a = 0, _b = currentSourceFile.amdDependencies; _a < _b.length; _a++) {
                        var dep = _b[_a];
                        if (dep.name) {
                            writeComment("/// <amd-dependency name=\"" + dep.name + "\" path=\"" + dep.path + "\" />");
                        }
                        else {
                            writeComment("/// <amd-dependency path=\"" + dep.path + "\" />");
                        }
                        writeLine();
                    }
                }
                for (var _c = 0, files_2 = files; _c < files_2.length; _c++) {
                    var directive = files_2[_c];
                    var pos = writer.getTextPos();
                    writeComment("/// <reference path=\"" + directive.fileName + "\" />");
                    if (bundleFileInfo)
                        bundleFileInfo.sections.push({ pos: pos, end: writer.getTextPos(), kind: "reference" /* Reference */, data: directive.fileName });
                    writeLine();
                }
                for (var _d = 0, types_24 = types; _d < types_24.length; _d++) {
                    var directive = types_24[_d];
                    var pos = writer.getTextPos();
                    writeComment("/// <reference types=\"" + directive.fileName + "\" />");
                    if (bundleFileInfo)
                        bundleFileInfo.sections.push({ pos: pos, end: writer.getTextPos(), kind: "type" /* Type */, data: directive.fileName });
                    writeLine();
                }
                for (var _e = 0, libs_1 = libs; _e < libs_1.length; _e++) {
                    var directive = libs_1[_e];
                    var pos = writer.getTextPos();
                    writeComment("/// <reference lib=\"" + directive.fileName + "\" />");
                    if (bundleFileInfo)
                        bundleFileInfo.sections.push({ pos: pos, end: writer.getTextPos(), kind: "lib" /* Lib */, data: directive.fileName });
                    writeLine();
                }
            }
            function emitSourceFileWorker(node) {
                var statements = node.statements;
                pushNameGenerationScope(node);
                ts.forEach(node.statements, generateNames);
                emitHelpers(node);
                var index = ts.findIndex(statements, function (statement) { return !ts.isPrologueDirective(statement); });
                emitTripleSlashDirectivesIfNeeded(node);
                emitList(node, statements, 1 /* MultiLine */, /*parenthesizerRule*/ undefined, index === -1 ? statements.length : index);
                popNameGenerationScope(node);
            }
            // Transformation nodes
            function emitPartiallyEmittedExpression(node) {
                emitExpression(node.expression);
            }
            function emitCommaList(node) {
                emitExpressionList(node, node.elements, 528 /* CommaListElements */, /*parenthesizerRule*/ undefined);
            }
            /**
             * Emits any prologue directives at the start of a Statement list, returning the
             * number of prologue directives written to the output.
             */
            function emitPrologueDirectives(statements, sourceFile, seenPrologueDirectives, recordBundleFileSection) {
                var needsToSetSourceFile = !!sourceFile;
                for (var i = 0; i < statements.length; i++) {
                    var statement = statements[i];
                    if (ts.isPrologueDirective(statement)) {
                        var shouldEmitPrologueDirective = seenPrologueDirectives ? !seenPrologueDirectives.has(statement.expression.text) : true;
                        if (shouldEmitPrologueDirective) {
                            if (needsToSetSourceFile) {
                                needsToSetSourceFile = false;
                                setSourceFile(sourceFile);
                            }
                            writeLine();
                            var pos = writer.getTextPos();
                            emit(statement);
                            if (recordBundleFileSection && bundleFileInfo)
                                bundleFileInfo.sections.push({ pos: pos, end: writer.getTextPos(), kind: "prologue" /* Prologue */, data: statement.expression.text });
                            if (seenPrologueDirectives) {
                                seenPrologueDirectives.add(statement.expression.text);
                            }
                        }
                    }
                    else {
                        // return index of the first non prologue directive
                        return i;
                    }
                }
                return statements.length;
            }
            function emitUnparsedPrologues(prologues, seenPrologueDirectives) {
                for (var _a = 0, prologues_1 = prologues; _a < prologues_1.length; _a++) {
                    var prologue = prologues_1[_a];
                    if (!seenPrologueDirectives.has(prologue.data)) {
                        writeLine();
                        var pos = writer.getTextPos();
                        emit(prologue);
                        if (bundleFileInfo)
                            bundleFileInfo.sections.push({ pos: pos, end: writer.getTextPos(), kind: "prologue" /* Prologue */, data: prologue.data });
                        if (seenPrologueDirectives) {
                            seenPrologueDirectives.add(prologue.data);
                        }
                    }
                }
            }
            function emitPrologueDirectivesIfNeeded(sourceFileOrBundle) {
                if (ts.isSourceFile(sourceFileOrBundle)) {
                    emitPrologueDirectives(sourceFileOrBundle.statements, sourceFileOrBundle);
                }
                else {
                    var seenPrologueDirectives = new ts.Set();
                    for (var _a = 0, _b = sourceFileOrBundle.prepends; _a < _b.length; _a++) {
                        var prepend = _b[_a];
                        emitUnparsedPrologues(prepend.prologues, seenPrologueDirectives);
                    }
                    for (var _c = 0, _d = sourceFileOrBundle.sourceFiles; _c < _d.length; _c++) {
                        var sourceFile = _d[_c];
                        emitPrologueDirectives(sourceFile.statements, sourceFile, seenPrologueDirectives, /*recordBundleFileSection*/ true);
                    }
                    setSourceFile(undefined);
                }
            }
            function getPrologueDirectivesFromBundledSourceFiles(bundle) {
                var seenPrologueDirectives = new ts.Set();
                var prologues;
                for (var index = 0; index < bundle.sourceFiles.length; index++) {
                    var sourceFile = bundle.sourceFiles[index];
                    var directives = void 0;
                    var end = 0;
                    for (var _a = 0, _b = sourceFile.statements; _a < _b.length; _a++) {
                        var statement = _b[_a];
                        if (!ts.isPrologueDirective(statement))
                            break;
                        if (seenPrologueDirectives.has(statement.expression.text))
                            continue;
                        seenPrologueDirectives.add(statement.expression.text);
                        (directives || (directives = [])).push({
                            pos: statement.pos,
                            end: statement.end,
                            expression: {
                                pos: statement.expression.pos,
                                end: statement.expression.end,
                                text: statement.expression.text
                            }
                        });
                        end = end < statement.end ? statement.end : end;
                    }
                    if (directives)
                        (prologues || (prologues = [])).push({ file: index, text: sourceFile.text.substring(0, end), directives: directives });
                }
                return prologues;
            }
            function emitShebangIfNeeded(sourceFileOrBundle) {
                if (ts.isSourceFile(sourceFileOrBundle) || ts.isUnparsedSource(sourceFileOrBundle)) {
                    var shebang = ts.getShebang(sourceFileOrBundle.text);
                    if (shebang) {
                        writeComment(shebang);
                        writeLine();
                        return true;
                    }
                }
                else {
                    for (var _a = 0, _b = sourceFileOrBundle.prepends; _a < _b.length; _a++) {
                        var prepend = _b[_a];
                        ts.Debug.assertNode(prepend, ts.isUnparsedSource);
                        if (emitShebangIfNeeded(prepend)) {
                            return true;
                        }
                    }
                    for (var _c = 0, _d = sourceFileOrBundle.sourceFiles; _c < _d.length; _c++) {
                        var sourceFile = _d[_c];
                        // Emit only the first encountered shebang
                        if (emitShebangIfNeeded(sourceFile)) {
                            return true;
                        }
                    }
                }
            }
            //
            // Helpers
            //
            function emitNodeWithWriter(node, writer) {
                if (!node)
                    return;
                var savedWrite = write;
                write = writer;
                emit(node);
                write = savedWrite;
            }
            function emitModifiers(node, modifiers) {
                if (modifiers && modifiers.length) {
                    emitList(node, modifiers, 262656 /* Modifiers */);
                    writeSpace();
                }
            }
            function emitTypeAnnotation(node) {
                if (node) {
                    writePunctuation(":");
                    writeSpace();
                    emit(node);
                }
            }
            function emitInitializer(node, equalCommentStartPos, container, parenthesizerRule) {
                if (node) {
                    writeSpace();
                    emitTokenWithComment(63 /* EqualsToken */, equalCommentStartPos, writeOperator, container);
                    writeSpace();
                    emitExpression(node, parenthesizerRule);
                }
            }
            function emitNodeWithPrefix(prefix, prefixWriter, node, emit) {
                if (node) {
                    prefixWriter(prefix);
                    emit(node);
                }
            }
            function emitWithLeadingSpace(node) {
                if (node) {
                    writeSpace();
                    emit(node);
                }
            }
            function emitExpressionWithLeadingSpace(node, parenthesizerRule) {
                if (node) {
                    writeSpace();
                    emitExpression(node, parenthesizerRule);
                }
            }
            function emitWithTrailingSpace(node) {
                if (node) {
                    emit(node);
                    writeSpace();
                }
            }
            function emitEmbeddedStatement(parent, node) {
                if (ts.isBlock(node) || ts.getEmitFlags(parent) & 1 /* SingleLine */) {
                    writeSpace();
                    emit(node);
                }
                else {
                    writeLine();
                    increaseIndent();
                    if (ts.isEmptyStatement(node)) {
                        pipelineEmit(5 /* EmbeddedStatement */, node);
                    }
                    else {
                        emit(node);
                    }
                    decreaseIndent();
                }
            }
            function emitDecorators(parentNode, decorators) {
                emitList(parentNode, decorators, 2146305 /* Decorators */);
            }
            function emitTypeArguments(parentNode, typeArguments) {
                emitList(parentNode, typeArguments, 53776 /* TypeArguments */, parenthesizer.parenthesizeMemberOfElementType);
            }
            function emitTypeParameters(parentNode, typeParameters) {
                if (ts.isFunctionLike(parentNode) && parentNode.typeArguments) { // Quick info uses type arguments in place of type parameters on instantiated signatures
                    return emitTypeArguments(parentNode, parentNode.typeArguments);
                }
                emitList(parentNode, typeParameters, 53776 /* TypeParameters */);
            }
            function emitParameters(parentNode, parameters) {
                emitList(parentNode, parameters, 2576 /* Parameters */);
            }
            function canEmitSimpleArrowHead(parentNode, parameters) {
                var parameter = ts.singleOrUndefined(parameters);
                return parameter
                    && parameter.pos === parentNode.pos // may not have parsed tokens between parent and parameter
                    && ts.isArrowFunction(parentNode) // only arrow functions may have simple arrow head
                    && !parentNode.type // arrow function may not have return type annotation
                    && !ts.some(parentNode.decorators) // parent may not have decorators
                    && !ts.some(parentNode.modifiers) // parent may not have modifiers
                    && !ts.some(parentNode.typeParameters) // parent may not have type parameters
                    && !ts.some(parameter.decorators) // parameter may not have decorators
                    && !ts.some(parameter.modifiers) // parameter may not have modifiers
                    && !parameter.dotDotDotToken // parameter may not be rest
                    && !parameter.questionToken // parameter may not be optional
                    && !parameter.type // parameter may not have a type annotation
                    && !parameter.initializer // parameter may not have an initializer
                    && ts.isIdentifier(parameter.name); // parameter name must be identifier
            }
            function emitParametersForArrow(parentNode, parameters) {
                if (canEmitSimpleArrowHead(parentNode, parameters)) {
                    emitList(parentNode, parameters, 2576 /* Parameters */ & ~2048 /* Parenthesis */);
                }
                else {
                    emitParameters(parentNode, parameters);
                }
            }
            function emitParametersForIndexSignature(parentNode, parameters) {
                emitList(parentNode, parameters, 8848 /* IndexSignatureParameters */);
            }
            function writeDelimiter(format) {
                switch (format & 60 /* DelimitersMask */) {
                    case 0 /* None */:
                        break;
                    case 16 /* CommaDelimited */:
                        writePunctuation(",");
                        break;
                    case 4 /* BarDelimited */:
                        writeSpace();
                        writePunctuation("|");
                        break;
                    case 32 /* AsteriskDelimited */:
                        writeSpace();
                        writePunctuation("*");
                        writeSpace();
                        break;
                    case 8 /* AmpersandDelimited */:
                        writeSpace();
                        writePunctuation("&");
                        break;
                }
            }
            function emitList(parentNode, children, format, parenthesizerRule, start, count) {
                emitNodeList(emit, parentNode, children, format, parenthesizerRule, start, count);
            }
            function emitExpressionList(parentNode, children, format, parenthesizerRule, start, count) {
                emitNodeList(emitExpression, parentNode, children, format, parenthesizerRule, start, count);
            }
            function emitNodeList(emit, parentNode, children, format, parenthesizerRule, start, count) {
                if (start === void 0) {
                    start = 0;
                }
                if (count === void 0) {
                    count = children ? children.length - start : 0;
                }
                var isUndefined = children === undefined;
                if (isUndefined && format & 16384 /* OptionalIfUndefined */) {
                    return;
                }
                var isEmpty = children === undefined || start >= children.length || count === 0;
                if (isEmpty && format & 32768 /* OptionalIfEmpty */) {
                    if (onBeforeEmitNodeArray) {
                        onBeforeEmitNodeArray(children);
                    }
                    if (onAfterEmitNodeArray) {
                        onAfterEmitNodeArray(children);
                    }
                    return;
                }
                if (format & 15360 /* BracketsMask */) {
                    writePunctuation(getOpeningBracket(format));
                    if (isEmpty && children) {
                        emitTrailingCommentsOfPosition(children.pos, /*prefixSpace*/ true); // Emit comments within empty bracketed lists
                    }
                }
                if (onBeforeEmitNodeArray) {
                    onBeforeEmitNodeArray(children);
                }
                if (isEmpty) {
                    // Write a line terminator if the parent node was multi-line
                    if (format & 1 /* MultiLine */ && !(preserveSourceNewlines && (!parentNode || ts.rangeIsOnSingleLine(parentNode, currentSourceFile)))) {
                        writeLine();
                    }
                    else if (format & 256 /* SpaceBetweenBraces */ && !(format & 524288 /* NoSpaceIfEmpty */)) {
                        writeSpace();
                    }
                }
                else {
                    ts.Debug.type(children);
                    // Write the opening line terminator or leading whitespace.
                    var mayEmitInterveningComments = (format & 262144 /* NoInterveningComments */) === 0;
                    var shouldEmitInterveningComments = mayEmitInterveningComments;
                    var leadingLineTerminatorCount = getLeadingLineTerminatorCount(parentNode, children, format); // TODO: GH#18217
                    if (leadingLineTerminatorCount) {
                        writeLine(leadingLineTerminatorCount);
                        shouldEmitInterveningComments = false;
                    }
                    else if (format & 256 /* SpaceBetweenBraces */) {
                        writeSpace();
                    }
                    // Increase the indent, if requested.
                    if (format & 128 /* Indented */) {
                        increaseIndent();
                    }
                    // Emit each child.
                    var previousSibling = void 0;
                    var previousSourceFileTextKind = void 0;
                    var shouldDecreaseIndentAfterEmit = false;
                    for (var i = 0; i < count; i++) {
                        var child = children[start + i];
                        // Write the delimiter if this is not the first node.
                        if (format & 32 /* AsteriskDelimited */) {
                            // always write JSDoc in the format "\n *"
                            writeLine();
                            writeDelimiter(format);
                        }
                        else if (previousSibling) {
                            // i.e
                            //      function commentedParameters(
                            //          /* Parameter a */
                            //          a
                            //          /* End of parameter a */ -> this comment isn't considered to be trailing comment of parameter "a" due to newline
                            //          ,
                            if (format & 60 /* DelimitersMask */ && previousSibling.end !== (parentNode ? parentNode.end : -1)) {
                                emitLeadingCommentsOfPosition(previousSibling.end);
                            }
                            writeDelimiter(format);
                            recordBundleFileInternalSectionEnd(previousSourceFileTextKind);
                            // Write either a line terminator or whitespace to separate the elements.
                            var separatingLineTerminatorCount = getSeparatingLineTerminatorCount(previousSibling, child, format);
                            if (separatingLineTerminatorCount > 0) {
                                // If a synthesized node in a single-line list starts on a new
                                // line, we should increase the indent.
                                if ((format & (3 /* LinesMask */ | 128 /* Indented */)) === 0 /* SingleLine */) {
                                    increaseIndent();
                                    shouldDecreaseIndentAfterEmit = true;
                                }
                                writeLine(separatingLineTerminatorCount);
                                shouldEmitInterveningComments = false;
                            }
                            else if (previousSibling && format & 512 /* SpaceBetweenSiblings */) {
                                writeSpace();
                            }
                        }
                        // Emit this child.
                        previousSourceFileTextKind = recordBundleFileInternalSectionStart(child);
                        if (shouldEmitInterveningComments) {
                            if (emitTrailingCommentsOfPosition) {
                                var commentRange = ts.getCommentRange(child);
                                emitTrailingCommentsOfPosition(commentRange.pos);
                            }
                        }
                        else {
                            shouldEmitInterveningComments = mayEmitInterveningComments;
                        }
                        nextListElementPos = child.pos;
                        if (emit.length === 1) {
                            emit(child);
                        }
                        else {
                            emit(child, parenthesizerRule);
                        }
                        if (shouldDecreaseIndentAfterEmit) {
                            decreaseIndent();
                            shouldDecreaseIndentAfterEmit = false;
                        }
                        previousSibling = child;
                    }
                    // Write a trailing comma, if requested.
                    var emitFlags = previousSibling ? ts.getEmitFlags(previousSibling) : 0;
                    var skipTrailingComments = commentsDisabled || !!(emitFlags & 1024 /* NoTrailingComments */);
                    var hasTrailingComma = (children === null || children === void 0 ? void 0 : children.hasTrailingComma) && (format & 64 /* AllowTrailingComma */) && (format & 16 /* CommaDelimited */);
                    if (hasTrailingComma) {
                        if (previousSibling && !skipTrailingComments) {
                            emitTokenWithComment(27 /* CommaToken */, previousSibling.end, writePunctuation, previousSibling);
                        }
                        else {
                            writePunctuation(",");
                        }
                    }
                    // Emit any trailing comment of the last element in the list
                    // i.e
                    //       var array = [...
                    //          2
                    //          /* end of element 2 */
                    //       ];
                    if (previousSibling && (parentNode ? parentNode.end : -1) !== previousSibling.end && (format & 60 /* DelimitersMask */) && !skipTrailingComments) {
                        emitLeadingCommentsOfPosition(hasTrailingComma && (children === null || children === void 0 ? void 0 : children.end) ? children.end : previousSibling.end);
                    }
                    // Decrease the indent, if requested.
                    if (format & 128 /* Indented */) {
                        decreaseIndent();
                    }
                    recordBundleFileInternalSectionEnd(previousSourceFileTextKind);
                    // Write the closing line terminator or closing whitespace.
                    var closingLineTerminatorCount = getClosingLineTerminatorCount(parentNode, children, format);
                    if (closingLineTerminatorCount) {
                        writeLine(closingLineTerminatorCount);
                    }
                    else if (format & (2097152 /* SpaceAfterList */ | 256 /* SpaceBetweenBraces */)) {
                        writeSpace();
                    }
                }
                if (onAfterEmitNodeArray) {
                    onAfterEmitNodeArray(children);
                }
                if (format & 15360 /* BracketsMask */) {
                    if (isEmpty && children) {
                        emitLeadingCommentsOfPosition(children.end); // Emit leading comments within empty lists
                    }
                    writePunctuation(getClosingBracket(format));
                }
            }
            // Writers
            function writeLiteral(s) {
                writer.writeLiteral(s);
            }
            function writeStringLiteral(s) {
                writer.writeStringLiteral(s);
            }
            function writeBase(s) {
                writer.write(s);
            }
            function writeSymbol(s, sym) {
                writer.writeSymbol(s, sym);
            }
            function writePunctuation(s) {
                writer.writePunctuation(s);
            }
            function writeTrailingSemicolon() {
                writer.writeTrailingSemicolon(";");
            }
            function writeKeyword(s) {
                writer.writeKeyword(s);
            }
            function writeOperator(s) {
                writer.writeOperator(s);
            }
            function writeParameter(s) {
                writer.writeParameter(s);
            }
            function writeComment(s) {
                writer.writeComment(s);
            }
            function writeSpace() {
                writer.writeSpace(" ");
            }
            function writeProperty(s) {
                writer.writeProperty(s);
            }
            function writeLine(count) {
                if (count === void 0) {
                    count = 1;
                }
                for (var i = 0; i < count; i++) {
                    writer.writeLine(i > 0);
                }
            }
            function increaseIndent() {
                writer.increaseIndent();
            }
            function decreaseIndent() {
                writer.decreaseIndent();
            }
            function writeToken(token, pos, writer, contextNode) {
                return !sourceMapsDisabled
                    ? emitTokenWithSourceMap(contextNode, token, writer, pos, writeTokenText)
                    : writeTokenText(token, writer, pos);
            }
            function writeTokenNode(node, writer) {
                if (onBeforeEmitToken) {
                    onBeforeEmitToken(node);
                }
                writer(ts.tokenToString(node.kind));
                if (onAfterEmitToken) {
                    onAfterEmitToken(node);
                }
            }
            function writeTokenText(token, writer, pos) {
                var tokenString = ts.tokenToString(token);
                writer(tokenString);
                return pos < 0 ? pos : pos + tokenString.length;
            }
            function writeLineOrSpace(parentNode, prevChildNode, nextChildNode) {
                if (ts.getEmitFlags(parentNode) & 1 /* SingleLine */) {
                    writeSpace();
                }
                else if (preserveSourceNewlines) {
                    var lines = getLinesBetweenNodes(parentNode, prevChildNode, nextChildNode);
                    if (lines) {
                        writeLine(lines);
                    }
                    else {
                        writeSpace();
                    }
                }
                else {
                    writeLine();
                }
            }
            function writeLines(text) {
                var lines = text.split(/\r\n?|\n/g);
                var indentation = ts.guessIndentation(lines);
                for (var _a = 0, lines_3 = lines; _a < lines_3.length; _a++) {
                    var lineText = lines_3[_a];
                    var line = indentation ? lineText.slice(indentation) : lineText;
                    if (line.length) {
                        writeLine();
                        write(line);
                    }
                }
            }
            function writeLinesAndIndent(lineCount, writeSpaceIfNotIndenting) {
                if (lineCount) {
                    increaseIndent();
                    writeLine(lineCount);
                }
                else if (writeSpaceIfNotIndenting) {
                    writeSpace();
                }
            }
            // Helper function to decrease the indent if we previously indented.  Allows multiple
            // previous indent values to be considered at a time.  This also allows caller to just
            // call this once, passing in all their appropriate indent values, instead of needing
            // to call this helper function multiple times.
            function decreaseIndentIf(value1, value2) {
                if (value1) {
                    decreaseIndent();
                }
                if (value2) {
                    decreaseIndent();
                }
            }
            function getLeadingLineTerminatorCount(parentNode, children, format) {
                if (format & 2 /* PreserveLines */ || preserveSourceNewlines) {
                    if (format & 65536 /* PreferNewLine */) {
                        return 1;
                    }
                    var firstChild_1 = children[0];
                    if (firstChild_1 === undefined) {
                        return !parentNode || ts.rangeIsOnSingleLine(parentNode, currentSourceFile) ? 0 : 1;
                    }
                    if (firstChild_1.pos === nextListElementPos) {
                        // If this child starts at the beginning of a list item in a parent list, its leading
                        // line terminators have already been written as the separating line terminators of the
                        // parent list. Example:
                        //
                        // class Foo {
                        //   constructor() {}
                        //   public foo() {}
                        // }
                        //
                        // The outer list is the list of class members, with one line terminator between the
                        // constructor and the method. The constructor is written, the separating line terminator
                        // is written, and then we start emitting the method. Its modifiers ([public]) constitute an inner
                        // list, so we look for its leading line terminators. If we didn't know that we had already
                        // written a newline as part of the parent list, it would appear that we need to write a
                        // leading newline to start the modifiers.
                        return 0;
                    }
                    if (firstChild_1.kind === 11 /* JsxText */) {
                        // JsxText will be written with its leading whitespace, so don't add more manually.
                        return 0;
                    }
                    if (parentNode &&
                        !ts.positionIsSynthesized(parentNode.pos) &&
                        !ts.nodeIsSynthesized(firstChild_1) &&
                        (!firstChild_1.parent || ts.getOriginalNode(firstChild_1.parent) === ts.getOriginalNode(parentNode))) {
                        if (preserveSourceNewlines) {
                            return getEffectiveLines(function (includeComments) { return ts.getLinesBetweenPositionAndPrecedingNonWhitespaceCharacter(firstChild_1.pos, parentNode.pos, currentSourceFile, includeComments); });
                        }
                        return ts.rangeStartPositionsAreOnSameLine(parentNode, firstChild_1, currentSourceFile) ? 0 : 1;
                    }
                    if (synthesizedNodeStartsOnNewLine(firstChild_1, format)) {
                        return 1;
                    }
                }
                return format & 1 /* MultiLine */ ? 1 : 0;
            }
            function getSeparatingLineTerminatorCount(previousNode, nextNode, format) {
                if (format & 2 /* PreserveLines */ || preserveSourceNewlines) {
                    if (previousNode === undefined || nextNode === undefined) {
                        return 0;
                    }
                    if (nextNode.kind === 11 /* JsxText */) {
                        // JsxText will be written with its leading whitespace, so don't add more manually.
                        return 0;
                    }
                    else if (!ts.nodeIsSynthesized(previousNode) && !ts.nodeIsSynthesized(nextNode)) {
                        if (preserveSourceNewlines && siblingNodePositionsAreComparable(previousNode, nextNode)) {
                            return getEffectiveLines(function (includeComments) { return ts.getLinesBetweenRangeEndAndRangeStart(previousNode, nextNode, currentSourceFile, includeComments); });
                        }
                        // If `preserveSourceNewlines` is `false` we do not intend to preserve the effective lines between the
                        // previous and next node. Instead we naively check whether nodes are on separate lines within the
                        // same node parent. If so, we intend to preserve a single line terminator. This is less precise and
                        // expensive than checking with `preserveSourceNewlines` as above, but the goal is not to preserve the
                        // effective source lines between two sibling nodes.
                        else if (!preserveSourceNewlines && originalNodesHaveSameParent(previousNode, nextNode)) {
                            return ts.rangeEndIsOnSameLineAsRangeStart(previousNode, nextNode, currentSourceFile) ? 0 : 1;
                        }
                        // If the two nodes are not comparable, add a line terminator based on the format that can indicate
                        // whether new lines are preferred or not.
                        return format & 65536 /* PreferNewLine */ ? 1 : 0;
                    }
                    else if (synthesizedNodeStartsOnNewLine(previousNode, format) || synthesizedNodeStartsOnNewLine(nextNode, format)) {
                        return 1;
                    }
                }
                else if (ts.getStartsOnNewLine(nextNode)) {
                    return 1;
                }
                return format & 1 /* MultiLine */ ? 1 : 0;
            }
            function getClosingLineTerminatorCount(parentNode, children, format) {
                if (format & 2 /* PreserveLines */ || preserveSourceNewlines) {
                    if (format & 65536 /* PreferNewLine */) {
                        return 1;
                    }
                    var lastChild = ts.lastOrUndefined(children);
                    if (lastChild === undefined) {
                        return !parentNode || ts.rangeIsOnSingleLine(parentNode, currentSourceFile) ? 0 : 1;
                    }
                    if (parentNode && !ts.positionIsSynthesized(parentNode.pos) && !ts.nodeIsSynthesized(lastChild) && (!lastChild.parent || lastChild.parent === parentNode)) {
                        if (preserveSourceNewlines) {
                            var end_1 = ts.isNodeArray(children) && !ts.positionIsSynthesized(children.end) ? children.end : lastChild.end;
                            return getEffectiveLines(function (includeComments) { return ts.getLinesBetweenPositionAndNextNonWhitespaceCharacter(end_1, parentNode.end, currentSourceFile, includeComments); });
                        }
                        return ts.rangeEndPositionsAreOnSameLine(parentNode, lastChild, currentSourceFile) ? 0 : 1;
                    }
                    if (synthesizedNodeStartsOnNewLine(lastChild, format)) {
                        return 1;
                    }
                }
                if (format & 1 /* MultiLine */ && !(format & 131072 /* NoTrailingNewLine */)) {
                    return 1;
                }
                return 0;
            }
            function getEffectiveLines(getLineDifference) {
                // If 'preserveSourceNewlines' is disabled, we should never call this function
                // because it could be more expensive than alternative approximations.
                ts.Debug.assert(!!preserveSourceNewlines);
                // We start by measuring the line difference from a position to its adjacent comments,
                // so that this is counted as a one-line difference, not two:
                //
                //   node1;
                //   // NODE2 COMMENT
                //   node2;
                var lines = getLineDifference(/*includeComments*/ true);
                if (lines === 0) {
                    // However, if the line difference considering comments was 0, we might have this:
                    //
                    //   node1; // NODE2 COMMENT
                    //   node2;
                    //
                    // in which case we should be ignoring node2's comment, so this too is counted as
                    // a one-line difference, not zero.
                    return getLineDifference(/*includeComments*/ false);
                }
                return lines;
            }
            function writeLineSeparatorsAndIndentBefore(node, parent) {
                var leadingNewlines = preserveSourceNewlines && getLeadingLineTerminatorCount(parent, [node], 0 /* None */);
                if (leadingNewlines) {
                    writeLinesAndIndent(leadingNewlines, /*writeLinesIfNotIndenting*/ false);
                }
                return !!leadingNewlines;
            }
            function writeLineSeparatorsAfter(node, parent) {
                var trailingNewlines = preserveSourceNewlines && getClosingLineTerminatorCount(parent, [node], 0 /* None */);
                if (trailingNewlines) {
                    writeLine(trailingNewlines);
                }
            }
            function synthesizedNodeStartsOnNewLine(node, format) {
                if (ts.nodeIsSynthesized(node)) {
                    var startsOnNewLine = ts.getStartsOnNewLine(node);
                    if (startsOnNewLine === undefined) {
                        return (format & 65536 /* PreferNewLine */) !== 0;
                    }
                    return startsOnNewLine;
                }
                return (format & 65536 /* PreferNewLine */) !== 0;
            }
            function getLinesBetweenNodes(parent, node1, node2) {
                if (ts.getEmitFlags(parent) & 131072 /* NoIndentation */) {
                    return 0;
                }
                parent = skipSynthesizedParentheses(parent);
                node1 = skipSynthesizedParentheses(node1);
                node2 = skipSynthesizedParentheses(node2);
                // Always use a newline for synthesized code if the synthesizer desires it.
                if (ts.getStartsOnNewLine(node2)) {
                    return 1;
                }
                if (!ts.nodeIsSynthesized(parent) && !ts.nodeIsSynthesized(node1) && !ts.nodeIsSynthesized(node2)) {
                    if (preserveSourceNewlines) {
                        return getEffectiveLines(function (includeComments) { return ts.getLinesBetweenRangeEndAndRangeStart(node1, node2, currentSourceFile, includeComments); });
                    }
                    return ts.rangeEndIsOnSameLineAsRangeStart(node1, node2, currentSourceFile) ? 0 : 1;
                }
                return 0;
            }
            function isEmptyBlock(block) {
                return block.statements.length === 0
                    && ts.rangeEndIsOnSameLineAsRangeStart(block, block, currentSourceFile);
            }
            function skipSynthesizedParentheses(node) {
                while (node.kind === 210 /* ParenthesizedExpression */ && ts.nodeIsSynthesized(node)) {
                    node = node.expression;
                }
                return node;
            }
            function getTextOfNode(node, includeTrivia) {
                if (ts.isGeneratedIdentifier(node)) {
                    return generateName(node);
                }
                else if ((ts.isIdentifier(node) || ts.isPrivateIdentifier(node)) && (ts.nodeIsSynthesized(node) || !node.parent || !currentSourceFile || (node.parent && currentSourceFile && ts.getSourceFileOfNode(node) !== ts.getOriginalNode(currentSourceFile)))) {
                    return ts.idText(node);
                }
                else if (node.kind === 10 /* StringLiteral */ && node.textSourceNode) {
                    return getTextOfNode(node.textSourceNode, includeTrivia);
                }
                else if (ts.isLiteralExpression(node) && (ts.nodeIsSynthesized(node) || !node.parent)) {
                    return node.text;
                }
                return ts.getSourceTextOfNodeFromSourceFile(currentSourceFile, node, includeTrivia);
            }
            function getLiteralTextOfNode(node, neverAsciiEscape, jsxAttributeEscape) {
                if (node.kind === 10 /* StringLiteral */ && node.textSourceNode) {
                    var textSourceNode = node.textSourceNode;
                    if (ts.isIdentifier(textSourceNode) || ts.isNumericLiteral(textSourceNode)) {
                        var text = ts.isNumericLiteral(textSourceNode) ? textSourceNode.text : getTextOfNode(textSourceNode);
                        return jsxAttributeEscape ? "\"" + ts.escapeJsxAttributeString(text) + "\"" :
                            neverAsciiEscape || (ts.getEmitFlags(node) & 16777216 /* NoAsciiEscaping */) ? "\"" + ts.escapeString(text) + "\"" :
                                "\"" + ts.escapeNonAsciiString(text) + "\"";
                    }
                    else {
                        return getLiteralTextOfNode(textSourceNode, neverAsciiEscape, jsxAttributeEscape);
                    }
                }
                var flags = (neverAsciiEscape ? 1 /* NeverAsciiEscape */ : 0)
                    | (jsxAttributeEscape ? 2 /* JsxAttributeEscape */ : 0)
                    | (printerOptions.terminateUnterminatedLiterals ? 4 /* TerminateUnterminatedLiterals */ : 0)
                    | (printerOptions.target && printerOptions.target === 99 /* ESNext */ ? 8 /* AllowNumericSeparator */ : 0);
                return ts.getLiteralText(node, currentSourceFile, flags);
            }
            /**
             * Push a new name generation scope.
             */
            function pushNameGenerationScope(node) {
                if (node && ts.getEmitFlags(node) & 524288 /* ReuseTempVariableScope */) {
                    return;
                }
                tempFlagsStack.push(tempFlags);
                tempFlags = 0;
                reservedNamesStack.push(reservedNames);
            }
            /**
             * Pop the current name generation scope.
             */
            function popNameGenerationScope(node) {
                if (node && ts.getEmitFlags(node) & 524288 /* ReuseTempVariableScope */) {
                    return;
                }
                tempFlags = tempFlagsStack.pop();
                reservedNames = reservedNamesStack.pop();
            }
            function reserveNameInNestedScopes(name) {
                if (!reservedNames || reservedNames === ts.lastOrUndefined(reservedNamesStack)) {
                    reservedNames = new ts.Set();
                }
                reservedNames.add(name);
            }
            function generateNames(node) {
                if (!node)
                    return;
                switch (node.kind) {
                    case 233 /* Block */:
                        ts.forEach(node.statements, generateNames);
                        break;
                    case 248 /* LabeledStatement */:
                    case 246 /* WithStatement */:
                    case 238 /* DoStatement */:
                    case 239 /* WhileStatement */:
                        generateNames(node.statement);
                        break;
                    case 237 /* IfStatement */:
                        generateNames(node.thenStatement);
                        generateNames(node.elseStatement);
                        break;
                    case 240 /* ForStatement */:
                    case 242 /* ForOfStatement */:
                    case 241 /* ForInStatement */:
                        generateNames(node.initializer);
                        generateNames(node.statement);
                        break;
                    case 247 /* SwitchStatement */:
                        generateNames(node.caseBlock);
                        break;
                    case 261 /* CaseBlock */:
                        ts.forEach(node.clauses, generateNames);
                        break;
                    case 287 /* CaseClause */:
                    case 288 /* DefaultClause */:
                        ts.forEach(node.statements, generateNames);
                        break;
                    case 250 /* TryStatement */:
                        generateNames(node.tryBlock);
                        generateNames(node.catchClause);
                        generateNames(node.finallyBlock);
                        break;
                    case 290 /* CatchClause */:
                        generateNames(node.variableDeclaration);
                        generateNames(node.block);
                        break;
                    case 235 /* VariableStatement */:
                        generateNames(node.declarationList);
                        break;
                    case 253 /* VariableDeclarationList */:
                        ts.forEach(node.declarations, generateNames);
                        break;
                    case 252 /* VariableDeclaration */:
                    case 162 /* Parameter */:
                    case 201 /* BindingElement */:
                    case 255 /* ClassDeclaration */:
                        generateNameIfNeeded(node.name);
                        break;
                    case 254 /* FunctionDeclaration */:
                        generateNameIfNeeded(node.name);
                        if (ts.getEmitFlags(node) & 524288 /* ReuseTempVariableScope */) {
                            ts.forEach(node.parameters, generateNames);
                            generateNames(node.body);
                        }
                        break;
                    case 199 /* ObjectBindingPattern */:
                    case 200 /* ArrayBindingPattern */:
                        ts.forEach(node.elements, generateNames);
                        break;
                    case 264 /* ImportDeclaration */:
                        generateNames(node.importClause);
                        break;
                    case 265 /* ImportClause */:
                        generateNameIfNeeded(node.name);
                        generateNames(node.namedBindings);
                        break;
                    case 266 /* NamespaceImport */:
                        generateNameIfNeeded(node.name);
                        break;
                    case 272 /* NamespaceExport */:
                        generateNameIfNeeded(node.name);
                        break;
                    case 267 /* NamedImports */:
                        ts.forEach(node.elements, generateNames);
                        break;
                    case 268 /* ImportSpecifier */:
                        generateNameIfNeeded(node.propertyName || node.name);
                        break;
                }
            }
            function generateMemberNames(node) {
                if (!node)
                    return;
                switch (node.kind) {
                    case 291 /* PropertyAssignment */:
                    case 292 /* ShorthandPropertyAssignment */:
                    case 165 /* PropertyDeclaration */:
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        generateNameIfNeeded(node.name);
                        break;
                }
            }
            function generateNameIfNeeded(name) {
                if (name) {
                    if (ts.isGeneratedIdentifier(name)) {
                        generateName(name);
                    }
                    else if (ts.isBindingPattern(name)) {
                        generateNames(name);
                    }
                }
            }
            /**
             * Generate the text for a generated identifier.
             */
            function generateName(name) {
                if ((name.autoGenerateFlags & 7 /* KindMask */) === 4 /* Node */) {
                    // Node names generate unique names based on their original node
                    // and are cached based on that node's id.
                    return generateNameCached(getNodeForGeneratedName(name), name.autoGenerateFlags);
                }
                else {
                    // Auto, Loop, and Unique names are cached based on their unique
                    // autoGenerateId.
                    var autoGenerateId = name.autoGenerateId;
                    return autoGeneratedIdToGeneratedName[autoGenerateId] || (autoGeneratedIdToGeneratedName[autoGenerateId] = makeName(name));
                }
            }
            function generateNameCached(node, flags) {
                var nodeId = ts.getNodeId(node);
                return nodeIdToGeneratedName[nodeId] || (nodeIdToGeneratedName[nodeId] = generateNameForNode(node, flags));
            }
            /**
             * Returns a value indicating whether a name is unique globally, within the current file,
             * or within the NameGenerator.
             */
            function isUniqueName(name) {
                return isFileLevelUniqueName(name)
                    && !generatedNames.has(name)
                    && !(reservedNames && reservedNames.has(name));
            }
            /**
             * Returns a value indicating whether a name is unique globally or within the current file.
             */
            function isFileLevelUniqueName(name) {
                return currentSourceFile ? ts.isFileLevelUniqueName(currentSourceFile, name, hasGlobalName) : true;
            }
            /**
             * Returns a value indicating whether a name is unique within a container.
             */
            function isUniqueLocalName(name, container) {
                for (var node = container; ts.isNodeDescendantOf(node, container); node = node.nextContainer) {
                    if (node.locals) {
                        var local = node.locals.get(ts.escapeLeadingUnderscores(name));
                        // We conservatively include alias symbols to cover cases where they're emitted as locals
                        if (local && local.flags & (111551 /* Value */ | 1048576 /* ExportValue */ | 2097152 /* Alias */)) {
                            return false;
                        }
                    }
                }
                return true;
            }
            /**
             * Return the next available name in the pattern _a ... _z, _0, _1, ...
             * TempFlags._i or TempFlags._n may be used to express a preference for that dedicated name.
             * Note that names generated by makeTempVariableName and makeUniqueName will never conflict.
             */
            function makeTempVariableName(flags, reservedInNestedScopes) {
                if (flags && !(tempFlags & flags)) {
                    var name = flags === 268435456 /* _i */ ? "_i" : "_n";
                    if (isUniqueName(name)) {
                        tempFlags |= flags;
                        if (reservedInNestedScopes) {
                            reserveNameInNestedScopes(name);
                        }
                        return name;
                    }
                }
                while (true) {
                    var count = tempFlags & 268435455 /* CountMask */;
                    tempFlags++;
                    // Skip over 'i' and 'n'
                    if (count !== 8 && count !== 13) {
                        var name = count < 26
                            ? "_" + String.fromCharCode(97 /* a */ + count)
                            : "_" + (count - 26);
                        if (isUniqueName(name)) {
                            if (reservedInNestedScopes) {
                                reserveNameInNestedScopes(name);
                            }
                            return name;
                        }
                    }
                }
            }
            /**
             * Generate a name that is unique within the current file and doesn't conflict with any names
             * in global scope. The name is formed by adding an '_n' suffix to the specified base name,
             * where n is a positive integer. Note that names generated by makeTempVariableName and
             * makeUniqueName are guaranteed to never conflict.
             * If `optimistic` is set, the first instance will use 'baseName' verbatim instead of 'baseName_1'
             */
            function makeUniqueName(baseName, checkFn, optimistic, scoped) {
                if (checkFn === void 0) {
                    checkFn = isUniqueName;
                }
                if (optimistic) {
                    if (checkFn(baseName)) {
                        if (scoped) {
                            reserveNameInNestedScopes(baseName);
                        }
                        else {
                            generatedNames.add(baseName);
                        }
                        return baseName;
                    }
                }
                // Find the first unique 'name_n', where n is a positive number
                if (baseName.charCodeAt(baseName.length - 1) !== 95 /* _ */) {
                    baseName += "_";
                }
                var i = 1;
                while (true) {
                    var generatedName = baseName + i;
                    if (checkFn(generatedName)) {
                        if (scoped) {
                            reserveNameInNestedScopes(generatedName);
                        }
                        else {
                            generatedNames.add(generatedName);
                        }
                        return generatedName;
                    }
                    i++;
                }
            }
            function makeFileLevelOptimisticUniqueName(name) {
                return makeUniqueName(name, isFileLevelUniqueName, /*optimistic*/ true);
            }
            /**
             * Generates a unique name for a ModuleDeclaration or EnumDeclaration.
             */
            function generateNameForModuleOrEnum(node) {
                var name = getTextOfNode(node.name);
                // Use module/enum name itself if it is unique, otherwise make a unique variation
                return isUniqueLocalName(name, node) ? name : makeUniqueName(name);
            }
            /**
             * Generates a unique name for an ImportDeclaration or ExportDeclaration.
             */
            function generateNameForImportOrExportDeclaration(node) {
                var expr = ts.getExternalModuleName(node); // TODO: GH#18217
                var baseName = ts.isStringLiteral(expr) ?
                    ts.makeIdentifierFromModuleName(expr.text) : "module";
                return makeUniqueName(baseName);
            }
            /**
             * Generates a unique name for a default export.
             */
            function generateNameForExportDefault() {
                return makeUniqueName("default");
            }
            /**
             * Generates a unique name for a class expression.
             */
            function generateNameForClassExpression() {
                return makeUniqueName("class");
            }
            function generateNameForMethodOrAccessor(node) {
                if (ts.isIdentifier(node.name)) {
                    return generateNameCached(node.name);
                }
                return makeTempVariableName(0 /* Auto */);
            }
            /**
             * Generates a unique name from a node.
             */
            function generateNameForNode(node, flags) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        return makeUniqueName(getTextOfNode(node), isUniqueName, !!(flags & 16 /* Optimistic */), !!(flags & 8 /* ReservedInNestedScopes */));
                    case 259 /* ModuleDeclaration */:
                    case 258 /* EnumDeclaration */:
                        return generateNameForModuleOrEnum(node);
                    case 264 /* ImportDeclaration */:
                    case 270 /* ExportDeclaration */:
                        return generateNameForImportOrExportDeclaration(node);
                    case 254 /* FunctionDeclaration */:
                    case 255 /* ClassDeclaration */:
                    case 269 /* ExportAssignment */:
                        return generateNameForExportDefault();
                    case 224 /* ClassExpression */:
                        return generateNameForClassExpression();
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return generateNameForMethodOrAccessor(node);
                    case 160 /* ComputedPropertyName */:
                        return makeTempVariableName(0 /* Auto */, /*reserveInNestedScopes*/ true);
                    default:
                        return makeTempVariableName(0 /* Auto */);
                }
            }
            /**
             * Generates a unique identifier for a node.
             */
            function makeName(name) {
                switch (name.autoGenerateFlags & 7 /* KindMask */) {
                    case 1 /* Auto */:
                        return makeTempVariableName(0 /* Auto */, !!(name.autoGenerateFlags & 8 /* ReservedInNestedScopes */));
                    case 2 /* Loop */:
                        return makeTempVariableName(268435456 /* _i */, !!(name.autoGenerateFlags & 8 /* ReservedInNestedScopes */));
                    case 3 /* Unique */:
                        return makeUniqueName(ts.idText(name), (name.autoGenerateFlags & 32 /* FileLevel */) ? isFileLevelUniqueName : isUniqueName, !!(name.autoGenerateFlags & 16 /* Optimistic */), !!(name.autoGenerateFlags & 8 /* ReservedInNestedScopes */));
                }
                return ts.Debug.fail("Unsupported GeneratedIdentifierKind.");
            }
            /**
             * Gets the node from which a name should be generated.
             */
            function getNodeForGeneratedName(name) {
                var autoGenerateId = name.autoGenerateId;
                var node = name;
                var original = node.original;
                while (original) {
                    node = original;
                    // if "node" is a different generated name (having a different
                    // "autoGenerateId"), use it and stop traversing.
                    if (ts.isIdentifier(node)
                        && !!(node.autoGenerateFlags & 4 /* Node */)
                        && node.autoGenerateId !== autoGenerateId) {
                        break;
                    }
                    original = node.original;
                }
                // otherwise, return the original node for the source;
                return node;
            }
            // Comments
            function pipelineEmitWithComments(hint, node) {
                var pipelinePhase = getNextPipelinePhase(2 /* Comments */, hint, node);
                var savedContainerPos = containerPos;
                var savedContainerEnd = containerEnd;
                var savedDeclarationListContainerEnd = declarationListContainerEnd;
                emitCommentsBeforeNode(node);
                pipelinePhase(hint, node);
                emitCommentsAfterNode(node, savedContainerPos, savedContainerEnd, savedDeclarationListContainerEnd);
            }
            function emitCommentsBeforeNode(node) {
                var emitFlags = ts.getEmitFlags(node);
                var commentRange = ts.getCommentRange(node);
                // Emit leading comments
                emitLeadingCommentsOfNode(node, emitFlags, commentRange.pos, commentRange.end);
                if (emitFlags & 2048 /* NoNestedComments */) {
                    commentsDisabled = true;
                }
            }
            function emitCommentsAfterNode(node, savedContainerPos, savedContainerEnd, savedDeclarationListContainerEnd) {
                var emitFlags = ts.getEmitFlags(node);
                var commentRange = ts.getCommentRange(node);
                // Emit trailing comments
                if (emitFlags & 2048 /* NoNestedComments */) {
                    commentsDisabled = false;
                }
                emitTrailingCommentsOfNode(node, emitFlags, commentRange.pos, commentRange.end, savedContainerPos, savedContainerEnd, savedDeclarationListContainerEnd);
            }
            function emitLeadingCommentsOfNode(node, emitFlags, pos, end) {
                enterComment();
                hasWrittenComment = false;
                // We have to explicitly check that the node is JsxText because if the compilerOptions.jsx is "preserve" we will not do any transformation.
                // It is expensive to walk entire tree just to set one kind of node to have no comments.
                var skipLeadingComments = pos < 0 || (emitFlags & 512 /* NoLeadingComments */) !== 0 || node.kind === 11 /* JsxText */;
                var skipTrailingComments = end < 0 || (emitFlags & 1024 /* NoTrailingComments */) !== 0 || node.kind === 11 /* JsxText */;
                // Save current container state on the stack.
                if ((pos > 0 || end > 0) && pos !== end) {
                    // Emit leading comments if the position is not synthesized and the node
                    // has not opted out from emitting leading comments.
                    if (!skipLeadingComments) {
                        emitLeadingComments(pos, /*isEmittedNode*/ node.kind !== 344 /* NotEmittedStatement */);
                    }
                    if (!skipLeadingComments || (pos >= 0 && (emitFlags & 512 /* NoLeadingComments */) !== 0)) {
                        // Advance the container position if comments get emitted or if they've been disabled explicitly using NoLeadingComments.
                        containerPos = pos;
                    }
                    if (!skipTrailingComments || (end >= 0 && (emitFlags & 1024 /* NoTrailingComments */) !== 0)) {
                        // As above.
                        containerEnd = end;
                        // To avoid invalid comment emit in a down-level binding pattern, we
                        // keep track of the last declaration list container's end
                        if (node.kind === 253 /* VariableDeclarationList */) {
                            declarationListContainerEnd = end;
                        }
                    }
                }
                ts.forEach(ts.getSyntheticLeadingComments(node), emitLeadingSynthesizedComment);
                exitComment();
            }
            function emitTrailingCommentsOfNode(node, emitFlags, pos, end, savedContainerPos, savedContainerEnd, savedDeclarationListContainerEnd) {
                enterComment();
                var skipTrailingComments = end < 0 || (emitFlags & 1024 /* NoTrailingComments */) !== 0 || node.kind === 11 /* JsxText */;
                ts.forEach(ts.getSyntheticTrailingComments(node), emitTrailingSynthesizedComment);
                if ((pos > 0 || end > 0) && pos !== end) {
                    // Restore previous container state.
                    containerPos = savedContainerPos;
                    containerEnd = savedContainerEnd;
                    declarationListContainerEnd = savedDeclarationListContainerEnd;
                    // Emit trailing comments if the position is not synthesized and the node
                    // has not opted out from emitting leading comments and is an emitted node.
                    if (!skipTrailingComments && node.kind !== 344 /* NotEmittedStatement */) {
                        emitTrailingComments(end);
                    }
                }
                exitComment();
            }
            function emitLeadingSynthesizedComment(comment) {
                if (comment.hasLeadingNewline || comment.kind === 2 /* SingleLineCommentTrivia */) {
                    writer.writeLine();
                }
                writeSynthesizedComment(comment);
                if (comment.hasTrailingNewLine || comment.kind === 2 /* SingleLineCommentTrivia */) {
                    writer.writeLine();
                }
                else {
                    writer.writeSpace(" ");
                }
            }
            function emitTrailingSynthesizedComment(comment) {
                if (!writer.isAtStartOfLine()) {
                    writer.writeSpace(" ");
                }
                writeSynthesizedComment(comment);
                if (comment.hasTrailingNewLine) {
                    writer.writeLine();
                }
            }
            function writeSynthesizedComment(comment) {
                var text = formatSynthesizedComment(comment);
                var lineMap = comment.kind === 3 /* MultiLineCommentTrivia */ ? ts.computeLineStarts(text) : undefined;
                ts.writeCommentRange(text, lineMap, writer, 0, text.length, newLine);
            }
            function formatSynthesizedComment(comment) {
                return comment.kind === 3 /* MultiLineCommentTrivia */
                    ? "/*" + comment.text + "*/"
                    : "//" + comment.text;
            }
            function emitBodyWithDetachedComments(node, detachedRange, emitCallback) {
                enterComment();
                var pos = detachedRange.pos, end = detachedRange.end;
                var emitFlags = ts.getEmitFlags(node);
                var skipLeadingComments = pos < 0 || (emitFlags & 512 /* NoLeadingComments */) !== 0;
                var skipTrailingComments = commentsDisabled || end < 0 || (emitFlags & 1024 /* NoTrailingComments */) !== 0;
                if (!skipLeadingComments) {
                    emitDetachedCommentsAndUpdateCommentsInfo(detachedRange);
                }
                exitComment();
                if (emitFlags & 2048 /* NoNestedComments */ && !commentsDisabled) {
                    commentsDisabled = true;
                    emitCallback(node);
                    commentsDisabled = false;
                }
                else {
                    emitCallback(node);
                }
                enterComment();
                if (!skipTrailingComments) {
                    emitLeadingComments(detachedRange.end, /*isEmittedNode*/ true);
                    if (hasWrittenComment && !writer.isAtStartOfLine()) {
                        writer.writeLine();
                    }
                }
                exitComment();
            }
            function originalNodesHaveSameParent(nodeA, nodeB) {
                nodeA = ts.getOriginalNode(nodeA);
                // For performance, do not call `getOriginalNode` for `nodeB` if `nodeA` doesn't even
                // have a parent node.
                return nodeA.parent && nodeA.parent === ts.getOriginalNode(nodeB).parent;
            }
            function siblingNodePositionsAreComparable(previousNode, nextNode) {
                if (nextNode.pos < previousNode.end) {
                    return false;
                }
                previousNode = ts.getOriginalNode(previousNode);
                nextNode = ts.getOriginalNode(nextNode);
                var parent = previousNode.parent;
                if (!parent || parent !== nextNode.parent) {
                    return false;
                }
                var parentNodeArray = ts.getContainingNodeArray(previousNode);
                var prevNodeIndex = parentNodeArray === null || parentNodeArray === void 0 ? void 0 : parentNodeArray.indexOf(previousNode);
                return prevNodeIndex !== undefined && prevNodeIndex > -1 && parentNodeArray.indexOf(nextNode) === prevNodeIndex + 1;
            }
            function emitLeadingComments(pos, isEmittedNode) {
                hasWrittenComment = false;
                if (isEmittedNode) {
                    if (pos === 0 && (currentSourceFile === null || currentSourceFile === void 0 ? void 0 : currentSourceFile.isDeclarationFile)) {
                        forEachLeadingCommentToEmit(pos, emitNonTripleSlashLeadingComment);
                    }
                    else {
                        forEachLeadingCommentToEmit(pos, emitLeadingComment);
                    }
                }
                else if (pos === 0) {
                    // If the node will not be emitted in JS, remove all the comments(normal, pinned and ///) associated with the node,
                    // unless it is a triple slash comment at the top of the file.
                    // For Example:
                    //      /// <reference-path ...>
                    //      declare var x;
                    //      /// <reference-path ...>
                    //      interface F {}
                    //  The first /// will NOT be removed while the second one will be removed even though both node will not be emitted
                    forEachLeadingCommentToEmit(pos, emitTripleSlashLeadingComment);
                }
            }
            function emitTripleSlashLeadingComment(commentPos, commentEnd, kind, hasTrailingNewLine, rangePos) {
                if (isTripleSlashComment(commentPos, commentEnd)) {
                    emitLeadingComment(commentPos, commentEnd, kind, hasTrailingNewLine, rangePos);
                }
            }
            function emitNonTripleSlashLeadingComment(commentPos, commentEnd, kind, hasTrailingNewLine, rangePos) {
                if (!isTripleSlashComment(commentPos, commentEnd)) {
                    emitLeadingComment(commentPos, commentEnd, kind, hasTrailingNewLine, rangePos);
                }
            }
            function shouldWriteComment(text, pos) {
                if (printerOptions.onlyPrintJsDocStyle) {
                    return (ts.isJSDocLikeText(text, pos) || ts.isPinnedComment(text, pos));
                }
                return true;
            }
            function emitLeadingComment(commentPos, commentEnd, kind, hasTrailingNewLine, rangePos) {
                if (!shouldWriteComment(currentSourceFile.text, commentPos))
                    return;
                if (!hasWrittenComment) {
                    ts.emitNewLineBeforeLeadingCommentOfPosition(getCurrentLineMap(), writer, rangePos, commentPos);
                    hasWrittenComment = true;
                }
                // Leading comments are emitted at /*leading comment1 */space/*leading comment*/space
                emitPos(commentPos);
                ts.writeCommentRange(currentSourceFile.text, getCurrentLineMap(), writer, commentPos, commentEnd, newLine);
                emitPos(commentEnd);
                if (hasTrailingNewLine) {
                    writer.writeLine();
                }
                else if (kind === 3 /* MultiLineCommentTrivia */) {
                    writer.writeSpace(" ");
                }
            }
            function emitLeadingCommentsOfPosition(pos) {
                if (commentsDisabled || pos === -1) {
                    return;
                }
                emitLeadingComments(pos, /*isEmittedNode*/ true);
            }
            function emitTrailingComments(pos) {
                forEachTrailingCommentToEmit(pos, emitTrailingComment);
            }
            function emitTrailingComment(commentPos, commentEnd, _kind, hasTrailingNewLine) {
                if (!shouldWriteComment(currentSourceFile.text, commentPos))
                    return;
                // trailing comments are emitted at space/*trailing comment1 */space/*trailing comment2*/
                if (!writer.isAtStartOfLine()) {
                    writer.writeSpace(" ");
                }
                emitPos(commentPos);
                ts.writeCommentRange(currentSourceFile.text, getCurrentLineMap(), writer, commentPos, commentEnd, newLine);
                emitPos(commentEnd);
                if (hasTrailingNewLine) {
                    writer.writeLine();
                }
            }
            function emitTrailingCommentsOfPosition(pos, prefixSpace, forceNoNewline) {
                if (commentsDisabled) {
                    return;
                }
                enterComment();
                forEachTrailingCommentToEmit(pos, prefixSpace ? emitTrailingComment : forceNoNewline ? emitTrailingCommentOfPositionNoNewline : emitTrailingCommentOfPosition);
                exitComment();
            }
            function emitTrailingCommentOfPositionNoNewline(commentPos, commentEnd, kind) {
                // trailing comments of a position are emitted at /*trailing comment1 */space/*trailing comment*/space
                emitPos(commentPos);
                ts.writeCommentRange(currentSourceFile.text, getCurrentLineMap(), writer, commentPos, commentEnd, newLine);
                emitPos(commentEnd);
                if (kind === 2 /* SingleLineCommentTrivia */) {
                    writer.writeLine(); // still write a newline for single-line comments, so closing tokens aren't written on the same line
                }
            }
            function emitTrailingCommentOfPosition(commentPos, commentEnd, _kind, hasTrailingNewLine) {
                // trailing comments of a position are emitted at /*trailing comment1 */space/*trailing comment*/space
                emitPos(commentPos);
                ts.writeCommentRange(currentSourceFile.text, getCurrentLineMap(), writer, commentPos, commentEnd, newLine);
                emitPos(commentEnd);
                if (hasTrailingNewLine) {
                    writer.writeLine();
                }
                else {
                    writer.writeSpace(" ");
                }
            }
            function forEachLeadingCommentToEmit(pos, cb) {
                // Emit the leading comments only if the container's pos doesn't match because the container should take care of emitting these comments
                if (currentSourceFile && (containerPos === -1 || pos !== containerPos)) {
                    if (hasDetachedComments(pos)) {
                        forEachLeadingCommentWithoutDetachedComments(cb);
                    }
                    else {
                        ts.forEachLeadingCommentRange(currentSourceFile.text, pos, cb, /*state*/ pos);
                    }
                }
            }
            function forEachTrailingCommentToEmit(end, cb) {
                // Emit the trailing comments only if the container's end doesn't match because the container should take care of emitting these comments
                if (currentSourceFile && (containerEnd === -1 || (end !== containerEnd && end !== declarationListContainerEnd))) {
                    ts.forEachTrailingCommentRange(currentSourceFile.text, end, cb);
                }
            }
            function hasDetachedComments(pos) {
                return detachedCommentsInfo !== undefined && ts.last(detachedCommentsInfo).nodePos === pos;
            }
            function forEachLeadingCommentWithoutDetachedComments(cb) {
                // get the leading comments from detachedPos
                var pos = ts.last(detachedCommentsInfo).detachedCommentEndPos;
                if (detachedCommentsInfo.length - 1) {
                    detachedCommentsInfo.pop();
                }
                else {
                    detachedCommentsInfo = undefined;
                }
                ts.forEachLeadingCommentRange(currentSourceFile.text, pos, cb, /*state*/ pos);
            }
            function emitDetachedCommentsAndUpdateCommentsInfo(range) {
                var currentDetachedCommentInfo = ts.emitDetachedComments(currentSourceFile.text, getCurrentLineMap(), writer, emitComment, range, newLine, commentsDisabled);
                if (currentDetachedCommentInfo) {
                    if (detachedCommentsInfo) {
                        detachedCommentsInfo.push(currentDetachedCommentInfo);
                    }
                    else {
                        detachedCommentsInfo = [currentDetachedCommentInfo];
                    }
                }
            }
            function emitComment(text, lineMap, writer, commentPos, commentEnd, newLine) {
                if (!shouldWriteComment(currentSourceFile.text, commentPos))
                    return;
                emitPos(commentPos);
                ts.writeCommentRange(text, lineMap, writer, commentPos, commentEnd, newLine);
                emitPos(commentEnd);
            }
            /**
             * Determine if the given comment is a triple-slash
             *
             * @return true if the comment is a triple-slash comment else false
             */
            function isTripleSlashComment(commentPos, commentEnd) {
                return ts.isRecognizedTripleSlashComment(currentSourceFile.text, commentPos, commentEnd);
            }
            // Source Maps
            function getParsedSourceMap(node) {
                if (node.parsedSourceMap === undefined && node.sourceMapText !== undefined) {
                    node.parsedSourceMap = ts.tryParseRawSourceMap(node.sourceMapText) || false;
                }
                return node.parsedSourceMap || undefined;
            }
            function pipelineEmitWithSourceMaps(hint, node) {
                var pipelinePhase = getNextPipelinePhase(3 /* SourceMaps */, hint, node);
                emitSourceMapsBeforeNode(node);
                pipelinePhase(hint, node);
                emitSourceMapsAfterNode(node);
            }
            function emitSourceMapsBeforeNode(node) {
                var emitFlags = ts.getEmitFlags(node);
                var sourceMapRange = ts.getSourceMapRange(node);
                // Emit leading sourcemap
                if (ts.isUnparsedNode(node)) {
                    ts.Debug.assertIsDefined(node.parent, "UnparsedNodes must have parent pointers");
                    var parsed = getParsedSourceMap(node.parent);
                    if (parsed && sourceMapGenerator) {
                        sourceMapGenerator.appendSourceMap(writer.getLine(), writer.getColumn(), parsed, node.parent.sourceMapPath, node.parent.getLineAndCharacterOfPosition(node.pos), node.parent.getLineAndCharacterOfPosition(node.end));
                    }
                }
                else {
                    var source = sourceMapRange.source || sourceMapSource;
                    if (node.kind !== 344 /* NotEmittedStatement */
                        && (emitFlags & 16 /* NoLeadingSourceMap */) === 0
                        && sourceMapRange.pos >= 0) {
                        emitSourcePos(sourceMapRange.source || sourceMapSource, skipSourceTrivia(source, sourceMapRange.pos));
                    }
                    if (emitFlags & 64 /* NoNestedSourceMaps */) {
                        sourceMapsDisabled = true;
                    }
                }
            }
            function emitSourceMapsAfterNode(node) {
                var emitFlags = ts.getEmitFlags(node);
                var sourceMapRange = ts.getSourceMapRange(node);
                // Emit trailing sourcemap
                if (!ts.isUnparsedNode(node)) {
                    if (emitFlags & 64 /* NoNestedSourceMaps */) {
                        sourceMapsDisabled = false;
                    }
                    if (node.kind !== 344 /* NotEmittedStatement */
                        && (emitFlags & 32 /* NoTrailingSourceMap */) === 0
                        && sourceMapRange.end >= 0) {
                        emitSourcePos(sourceMapRange.source || sourceMapSource, sourceMapRange.end);
                    }
                }
            }
            /**
             * Skips trivia such as comments and white-space that can be optionally overridden by the source-map source
             */
            function skipSourceTrivia(source, pos) {
                return source.skipTrivia ? source.skipTrivia(pos) : ts.skipTrivia(source.text, pos);
            }
            /**
             * Emits a mapping.
             *
             * If the position is synthetic (undefined or a negative value), no mapping will be
             * created.
             *
             * @param pos The position.
             */
            function emitPos(pos) {
                if (sourceMapsDisabled || ts.positionIsSynthesized(pos) || isJsonSourceMapSource(sourceMapSource)) {
                    return;
                }
                var _a = ts.getLineAndCharacterOfPosition(sourceMapSource, pos), sourceLine = _a.line, sourceCharacter = _a.character;
                sourceMapGenerator.addMapping(writer.getLine(), writer.getColumn(), sourceMapSourceIndex, sourceLine, sourceCharacter, 
                /*nameIndex*/ undefined);
            }
            function emitSourcePos(source, pos) {
                if (source !== sourceMapSource) {
                    var savedSourceMapSource = sourceMapSource;
                    var savedSourceMapSourceIndex = sourceMapSourceIndex;
                    setSourceMapSource(source);
                    emitPos(pos);
                    resetSourceMapSource(savedSourceMapSource, savedSourceMapSourceIndex);
                }
                else {
                    emitPos(pos);
                }
            }
            /**
             * Emits a token of a node with possible leading and trailing source maps.
             *
             * @param node The node containing the token.
             * @param token The token to emit.
             * @param tokenStartPos The start pos of the token.
             * @param emitCallback The callback used to emit the token.
             */
            function emitTokenWithSourceMap(node, token, writer, tokenPos, emitCallback) {
                if (sourceMapsDisabled || node && ts.isInJsonFile(node)) {
                    return emitCallback(token, writer, tokenPos);
                }
                var emitNode = node && node.emitNode;
                var emitFlags = emitNode && emitNode.flags || 0 /* None */;
                var range = emitNode && emitNode.tokenSourceMapRanges && emitNode.tokenSourceMapRanges[token];
                var source = range && range.source || sourceMapSource;
                tokenPos = skipSourceTrivia(source, range ? range.pos : tokenPos);
                if ((emitFlags & 128 /* NoTokenLeadingSourceMaps */) === 0 && tokenPos >= 0) {
                    emitSourcePos(source, tokenPos);
                }
                tokenPos = emitCallback(token, writer, tokenPos);
                if (range)
                    tokenPos = range.end;
                if ((emitFlags & 256 /* NoTokenTrailingSourceMaps */) === 0 && tokenPos >= 0) {
                    emitSourcePos(source, tokenPos);
                }
                return tokenPos;
            }
            function setSourceMapSource(source) {
                if (sourceMapsDisabled) {
                    return;
                }
                sourceMapSource = source;
                if (source === mostRecentlyAddedSourceMapSource) {
                    // Fast path for when the new source map is the most recently added, in which case
                    // we use its captured index without going through the source map generator.
                    sourceMapSourceIndex = mostRecentlyAddedSourceMapSourceIndex;
                    return;
                }
                if (isJsonSourceMapSource(source)) {
                    return;
                }
                sourceMapSourceIndex = sourceMapGenerator.addSource(source.fileName);
                if (printerOptions.inlineSources) {
                    sourceMapGenerator.setSourceContent(sourceMapSourceIndex, source.text);
                }
                mostRecentlyAddedSourceMapSource = source;
                mostRecentlyAddedSourceMapSourceIndex = sourceMapSourceIndex;
            }
            function resetSourceMapSource(source, sourceIndex) {
                sourceMapSource = source;
                sourceMapSourceIndex = sourceIndex;
            }
            function isJsonSourceMapSource(sourceFile) {
                return ts.fileExtensionIs(sourceFile.fileName, ".json" /* Json */);
            }
        }
        ts.createPrinter = createPrinter;
        function createBracketsMap() {
            var brackets = [];
            brackets[1024 /* Braces */] = ["{", "}"];
            brackets[2048 /* Parenthesis */] = ["(", ")"];
            brackets[4096 /* AngleBrackets */] = ["<", ">"];
            brackets[8192 /* SquareBrackets */] = ["[", "]"];
            return brackets;
        }
        function getOpeningBracket(format) {
            return brackets[format & 15360 /* BracketsMask */][0];
        }
        function getClosingBracket(format) {
            return brackets[format & 15360 /* BracketsMask */][1];
        }
        // Flags enum to track count of temp variables and a few dedicated names
        var TempFlags;
        (function (TempFlags) {
            TempFlags[TempFlags["Auto"] = 0] = "Auto";
            TempFlags[TempFlags["CountMask"] = 268435455] = "CountMask";
            TempFlags[TempFlags["_i"] = 268435456] = "_i";
        })(TempFlags || (TempFlags = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function createCachedDirectoryStructureHost(host, currentDirectory, useCaseSensitiveFileNames) {
            if (!host.getDirectories || !host.readDirectory) {
                return undefined;
            }
            var cachedReadDirectoryResult = new ts.Map();
            var getCanonicalFileName = ts.createGetCanonicalFileName(useCaseSensitiveFileNames);
            return {
                useCaseSensitiveFileNames: useCaseSensitiveFileNames,
                fileExists: fileExists,
                readFile: function (path, encoding) { return host.readFile(path, encoding); },
                directoryExists: host.directoryExists && directoryExists,
                getDirectories: getDirectories,
                readDirectory: readDirectory,
                createDirectory: host.createDirectory && createDirectory,
                writeFile: host.writeFile && writeFile,
                addOrDeleteFileOrDirectory: addOrDeleteFileOrDirectory,
                addOrDeleteFile: addOrDeleteFile,
                clearCache: clearCache,
                realpath: host.realpath && realpath
            };
            function toPath(fileName) {
                return ts.toPath(fileName, currentDirectory, getCanonicalFileName);
            }
            function getCachedFileSystemEntries(rootDirPath) {
                return cachedReadDirectoryResult.get(ts.ensureTrailingDirectorySeparator(rootDirPath));
            }
            function getCachedFileSystemEntriesForBaseDir(path) {
                return getCachedFileSystemEntries(ts.getDirectoryPath(path));
            }
            function getBaseNameOfFileName(fileName) {
                return ts.getBaseFileName(ts.normalizePath(fileName));
            }
            function createCachedFileSystemEntries(rootDir, rootDirPath) {
                var _a;
                if (!host.realpath || ts.ensureTrailingDirectorySeparator(toPath(host.realpath(rootDir))) === rootDirPath) {
                    var resultFromHost = {
                        files: ts.map(host.readDirectory(rootDir, /*extensions*/ undefined, /*exclude*/ undefined, /*include*/ ["*.*"]), getBaseNameOfFileName) || [],
                        directories: host.getDirectories(rootDir) || []
                    };
                    cachedReadDirectoryResult.set(ts.ensureTrailingDirectorySeparator(rootDirPath), resultFromHost);
                    return resultFromHost;
                }
                // If the directory is symlink do not cache the result
                if ((_a = host.directoryExists) === null || _a === void 0 ? void 0 : _a.call(host, rootDir)) {
                    cachedReadDirectoryResult.set(rootDirPath, false);
                    return false;
                }
                // Non existing directory
                return undefined;
            }
            /**
             * If the readDirectory result was already cached, it returns that
             * Otherwise gets result from host and caches it.
             * The host request is done under try catch block to avoid caching incorrect result
             */
            function tryReadDirectory(rootDir, rootDirPath) {
                rootDirPath = ts.ensureTrailingDirectorySeparator(rootDirPath);
                var cachedResult = getCachedFileSystemEntries(rootDirPath);
                if (cachedResult) {
                    return cachedResult;
                }
                try {
                    return createCachedFileSystemEntries(rootDir, rootDirPath);
                }
                catch (_e) {
                    // If there is exception to read directories, dont cache the result and direct the calls to host
                    ts.Debug.assert(!cachedReadDirectoryResult.has(ts.ensureTrailingDirectorySeparator(rootDirPath)));
                    return undefined;
                }
            }
            function fileNameEqual(name1, name2) {
                return getCanonicalFileName(name1) === getCanonicalFileName(name2);
            }
            function hasEntry(entries, name) {
                return ts.some(entries, function (file) { return fileNameEqual(file, name); });
            }
            function updateFileSystemEntry(entries, baseName, isValid) {
                if (hasEntry(entries, baseName)) {
                    if (!isValid) {
                        return ts.filterMutate(entries, function (entry) { return !fileNameEqual(entry, baseName); });
                    }
                }
                else if (isValid) {
                    return entries.push(baseName);
                }
            }
            function writeFile(fileName, data, writeByteOrderMark) {
                var path = toPath(fileName);
                var result = getCachedFileSystemEntriesForBaseDir(path);
                if (result) {
                    updateFilesOfFileSystemEntry(result, getBaseNameOfFileName(fileName), /*fileExists*/ true);
                }
                return host.writeFile(fileName, data, writeByteOrderMark);
            }
            function fileExists(fileName) {
                var path = toPath(fileName);
                var result = getCachedFileSystemEntriesForBaseDir(path);
                return result && hasEntry(result.files, getBaseNameOfFileName(fileName)) ||
                    host.fileExists(fileName);
            }
            function directoryExists(dirPath) {
                var path = toPath(dirPath);
                return cachedReadDirectoryResult.has(ts.ensureTrailingDirectorySeparator(path)) || host.directoryExists(dirPath);
            }
            function createDirectory(dirPath) {
                var path = toPath(dirPath);
                var result = getCachedFileSystemEntriesForBaseDir(path);
                var baseFileName = getBaseNameOfFileName(dirPath);
                if (result) {
                    updateFileSystemEntry(result.directories, baseFileName, /*isValid*/ true);
                }
                host.createDirectory(dirPath);
            }
            function getDirectories(rootDir) {
                var rootDirPath = toPath(rootDir);
                var result = tryReadDirectory(rootDir, rootDirPath);
                if (result) {
                    return result.directories.slice();
                }
                return host.getDirectories(rootDir);
            }
            function readDirectory(rootDir, extensions, excludes, includes, depth) {
                var rootDirPath = toPath(rootDir);
                var rootResult = tryReadDirectory(rootDir, rootDirPath);
                var rootSymLinkResult;
                if (rootResult !== undefined) {
                    return ts.matchFiles(rootDir, extensions, excludes, includes, useCaseSensitiveFileNames, currentDirectory, depth, getFileSystemEntries, realpath, directoryExists);
                }
                return host.readDirectory(rootDir, extensions, excludes, includes, depth);
                function getFileSystemEntries(dir) {
                    var path = toPath(dir);
                    if (path === rootDirPath) {
                        return rootResult || getFileSystemEntriesFromHost(dir, path);
                    }
                    var result = tryReadDirectory(dir, path);
                    return result !== undefined ?
                        result || getFileSystemEntriesFromHost(dir, path) :
                        ts.emptyFileSystemEntries;
                }
                function getFileSystemEntriesFromHost(dir, path) {
                    if (rootSymLinkResult && path === rootDirPath)
                        return rootSymLinkResult;
                    var result = {
                        files: ts.map(host.readDirectory(dir, /*extensions*/ undefined, /*exclude*/ undefined, /*include*/ ["*.*"]), getBaseNameOfFileName) || ts.emptyArray,
                        directories: host.getDirectories(dir) || ts.emptyArray
                    };
                    if (path === rootDirPath)
                        rootSymLinkResult = result;
                    return result;
                }
            }
            function realpath(s) {
                return host.realpath ? host.realpath(s) : s;
            }
            function addOrDeleteFileOrDirectory(fileOrDirectory, fileOrDirectoryPath) {
                var existingResult = getCachedFileSystemEntries(fileOrDirectoryPath);
                if (existingResult !== undefined) {
                    // Just clear the cache for now
                    // For now just clear the cache, since this could mean that multiple level entries might need to be re-evaluated
                    clearCache();
                    return undefined;
                }
                var parentResult = getCachedFileSystemEntriesForBaseDir(fileOrDirectoryPath);
                if (!parentResult) {
                    return undefined;
                }
                // This was earlier a file (hence not in cached directory contents)
                // or we never cached the directory containing it
                if (!host.directoryExists) {
                    // Since host doesnt support directory exists, clear the cache as otherwise it might not be same
                    clearCache();
                    return undefined;
                }
                var baseName = getBaseNameOfFileName(fileOrDirectory);
                var fsQueryResult = {
                    fileExists: host.fileExists(fileOrDirectoryPath),
                    directoryExists: host.directoryExists(fileOrDirectoryPath)
                };
                if (fsQueryResult.directoryExists || hasEntry(parentResult.directories, baseName)) {
                    // Folder added or removed, clear the cache instead of updating the folder and its structure
                    clearCache();
                }
                else {
                    // No need to update the directory structure, just files
                    updateFilesOfFileSystemEntry(parentResult, baseName, fsQueryResult.fileExists);
                }
                return fsQueryResult;
            }
            function addOrDeleteFile(fileName, filePath, eventKind) {
                if (eventKind === ts.FileWatcherEventKind.Changed) {
                    return;
                }
                var parentResult = getCachedFileSystemEntriesForBaseDir(filePath);
                if (parentResult) {
                    updateFilesOfFileSystemEntry(parentResult, getBaseNameOfFileName(fileName), eventKind === ts.FileWatcherEventKind.Created);
                }
            }
            function updateFilesOfFileSystemEntry(parentResult, baseName, fileExists) {
                updateFileSystemEntry(parentResult.files, baseName, fileExists);
            }
            function clearCache() {
                cachedReadDirectoryResult.clear();
            }
        }
        ts.createCachedDirectoryStructureHost = createCachedDirectoryStructureHost;
        var ConfigFileProgramReloadLevel;
        (function (ConfigFileProgramReloadLevel) {
            ConfigFileProgramReloadLevel[ConfigFileProgramReloadLevel["None"] = 0] = "None";
            /** Update the file name list from the disk */
            ConfigFileProgramReloadLevel[ConfigFileProgramReloadLevel["Partial"] = 1] = "Partial";
            /** Reload completely by re-reading contents of config file from disk and updating program */
            ConfigFileProgramReloadLevel[ConfigFileProgramReloadLevel["Full"] = 2] = "Full";
        })(ConfigFileProgramReloadLevel = ts.ConfigFileProgramReloadLevel || (ts.ConfigFileProgramReloadLevel = {}));
        /**
         * Updates the map of shared extended config file watches with a new set of extended config files from a base config file of the project
         */
        function updateSharedExtendedConfigFileWatcher(projectPath, options, extendedConfigFilesMap, createExtendedConfigFileWatch, toPath) {
            var _a;
            var extendedConfigs = ts.arrayToMap(((_a = options === null || options === void 0 ? void 0 : options.configFile) === null || _a === void 0 ? void 0 : _a.extendedSourceFiles) || ts.emptyArray, toPath);
            // remove project from all unrelated watchers
            extendedConfigFilesMap.forEach(function (watcher, extendedConfigFilePath) {
                if (!extendedConfigs.has(extendedConfigFilePath)) {
                    watcher.projects.delete(projectPath);
                    watcher.close();
                }
            });
            // Update the extended config files watcher
            extendedConfigs.forEach(function (extendedConfigFileName, extendedConfigFilePath) {
                var existing = extendedConfigFilesMap.get(extendedConfigFilePath);
                if (existing) {
                    existing.projects.add(projectPath);
                }
                else {
                    // start watching previously unseen extended config
                    extendedConfigFilesMap.set(extendedConfigFilePath, {
                        projects: new ts.Set([projectPath]),
                        watcher: createExtendedConfigFileWatch(extendedConfigFileName, extendedConfigFilePath),
                        close: function () {
                            var existing = extendedConfigFilesMap.get(extendedConfigFilePath);
                            if (!existing || existing.projects.size !== 0)
                                return;
                            existing.watcher.close();
                            extendedConfigFilesMap.delete(extendedConfigFilePath);
                        },
                    });
                }
            });
        }
        ts.updateSharedExtendedConfigFileWatcher = updateSharedExtendedConfigFileWatcher;
        /**
         * Remove the project from the extended config file watchers and close not needed watches
         */
        function clearSharedExtendedConfigFileWatcher(projectPath, extendedConfigFilesMap) {
            extendedConfigFilesMap.forEach(function (watcher) {
                if (watcher.projects.delete(projectPath))
                    watcher.close();
            });
        }
        ts.clearSharedExtendedConfigFileWatcher = clearSharedExtendedConfigFileWatcher;
        /**
         * Clean the extendsConfigCache when extended config file has changed
         */
        function cleanExtendedConfigCache(extendedConfigCache, extendedConfigFilePath, toPath) {
            if (!extendedConfigCache.delete(extendedConfigFilePath))
                return;
            extendedConfigCache.forEach(function (_a, key) {
                var _b;
                var extendedResult = _a.extendedResult;
                if ((_b = extendedResult.extendedSourceFiles) === null || _b === void 0 ? void 0 : _b.some(function (extendedFile) { return toPath(extendedFile) === extendedConfigFilePath; })) {
                    cleanExtendedConfigCache(extendedConfigCache, key, toPath);
                }
            });
        }
        ts.cleanExtendedConfigCache = cleanExtendedConfigCache;
        /**
         * Updates watchers based on the package json files used in module resolution
         */
        function updatePackageJsonWatch(lookups, packageJsonWatches, createPackageJsonWatch) {
            var newMap = new ts.Map(lookups);
            ts.mutateMap(packageJsonWatches, newMap, {
                createNewValue: createPackageJsonWatch,
                onDeleteValue: ts.closeFileWatcher
            });
        }
        ts.updatePackageJsonWatch = updatePackageJsonWatch;
        /**
         * Updates the existing missing file watches with the new set of missing files after new program is created
         */
        function updateMissingFilePathsWatch(program, missingFileWatches, createMissingFileWatch) {
            var missingFilePaths = program.getMissingFilePaths();
            // TODO(rbuckton): Should be a `Set` but that requires changing the below code that uses `mutateMap`
            var newMissingFilePathMap = ts.arrayToMap(missingFilePaths, ts.identity, ts.returnTrue);
            // Update the missing file paths watcher
            ts.mutateMap(missingFileWatches, newMissingFilePathMap, {
                // Watch the missing files
                createNewValue: createMissingFileWatch,
                // Files that are no longer missing (e.g. because they are no longer required)
                // should no longer be watched.
                onDeleteValue: ts.closeFileWatcher
            });
        }
        ts.updateMissingFilePathsWatch = updateMissingFilePathsWatch;
        /**
         * Updates the existing wild card directory watches with the new set of wild card directories from the config file
         * after new program is created because the config file was reloaded or program was created first time from the config file
         * Note that there is no need to call this function when the program is updated with additional files without reloading config files,
         * as wildcard directories wont change unless reloading config file
         */
        function updateWatchingWildcardDirectories(existingWatchedForWildcards, wildcardDirectories, watchDirectory) {
            ts.mutateMap(existingWatchedForWildcards, wildcardDirectories, {
                // Create new watch and recursive info
                createNewValue: createWildcardDirectoryWatcher,
                // Close existing watch thats not needed any more
                onDeleteValue: closeFileWatcherOf,
                // Close existing watch that doesnt match in the flags
                onExistingValue: updateWildcardDirectoryWatcher
            });
            function createWildcardDirectoryWatcher(directory, flags) {
                // Create new watch and recursive info
                return {
                    watcher: watchDirectory(directory, flags),
                    flags: flags
                };
            }
            function updateWildcardDirectoryWatcher(existingWatcher, flags, directory) {
                // Watcher needs to be updated if the recursive flags dont match
                if (existingWatcher.flags === flags) {
                    return;
                }
                existingWatcher.watcher.close();
                existingWatchedForWildcards.set(directory, createWildcardDirectoryWatcher(directory, flags));
            }
        }
        ts.updateWatchingWildcardDirectories = updateWatchingWildcardDirectories;
        /* @internal */
        function isIgnoredFileFromWildCardWatching(_a) {
            var watchedDirPath = _a.watchedDirPath, fileOrDirectory = _a.fileOrDirectory, fileOrDirectoryPath = _a.fileOrDirectoryPath, configFileName = _a.configFileName, options = _a.options, program = _a.program, extraFileExtensions = _a.extraFileExtensions, currentDirectory = _a.currentDirectory, useCaseSensitiveFileNames = _a.useCaseSensitiveFileNames, writeLog = _a.writeLog, toPath = _a.toPath;
            var newPath = ts.removeIgnoredPath(fileOrDirectoryPath);
            if (!newPath) {
                writeLog("Project: " + configFileName + " Detected ignored path: " + fileOrDirectory);
                return true;
            }
            fileOrDirectoryPath = newPath;
            if (fileOrDirectoryPath === watchedDirPath)
                return false;
            // If the the added or created file or directory is not supported file name, ignore the file
            // But when watched directory is added/removed, we need to reload the file list
            if (ts.hasExtension(fileOrDirectoryPath) && !ts.isSupportedSourceFileName(fileOrDirectory, options, extraFileExtensions)) {
                writeLog("Project: " + configFileName + " Detected file add/remove of non supported extension: " + fileOrDirectory);
                return true;
            }
            if (ts.isExcludedFile(fileOrDirectory, options.configFile.configFileSpecs, ts.getNormalizedAbsolutePath(ts.getDirectoryPath(configFileName), currentDirectory), useCaseSensitiveFileNames, currentDirectory)) {
                writeLog("Project: " + configFileName + " Detected excluded file: " + fileOrDirectory);
                return true;
            }
            if (!program)
                return false;
            // We want to ignore emit file check if file is not going to be emitted next to source file
            // In that case we follow config file inclusion rules
            if (ts.outFile(options) || options.outDir)
                return false;
            // File if emitted next to input needs to be ignored
            if (ts.fileExtensionIs(fileOrDirectoryPath, ".d.ts" /* Dts */)) {
                // If its declaration directory: its not ignored if not excluded by config
                if (options.declarationDir)
                    return false;
            }
            else if (!ts.fileExtensionIsOneOf(fileOrDirectoryPath, ts.supportedJSExtensions)) {
                return false;
            }
            // just check if sourceFile with the name exists
            var filePathWithoutExtension = ts.removeFileExtension(fileOrDirectoryPath);
            var realProgram = ts.isArray(program) ? undefined : isBuilderProgram(program) ? program.getProgramOrUndefined() : program;
            var builderProgram = !realProgram && !ts.isArray(program) ? program : undefined;
            if (hasSourceFile((filePathWithoutExtension + ".ts" /* Ts */)) ||
                hasSourceFile((filePathWithoutExtension + ".tsx" /* Tsx */))) {
                writeLog("Project: " + configFileName + " Detected output file: " + fileOrDirectory);
                return true;
            }
            return false;
            function hasSourceFile(file) {
                return realProgram ?
                    !!realProgram.getSourceFileByPath(file) :
                    builderProgram ?
                        builderProgram.getState().fileInfos.has(file) :
                        !!ts.find(program, function (rootFile) { return toPath(rootFile) === file; });
            }
        }
        ts.isIgnoredFileFromWildCardWatching = isIgnoredFileFromWildCardWatching;
        function isBuilderProgram(program) {
            return !!program.getState;
        }
        function isEmittedFileOfProgram(program, file) {
            if (!program) {
                return false;
            }
            return program.isEmittedFile(file);
        }
        ts.isEmittedFileOfProgram = isEmittedFileOfProgram;
        var WatchLogLevel;
        (function (WatchLogLevel) {
            WatchLogLevel[WatchLogLevel["None"] = 0] = "None";
            WatchLogLevel[WatchLogLevel["TriggerOnly"] = 1] = "TriggerOnly";
            WatchLogLevel[WatchLogLevel["Verbose"] = 2] = "Verbose";
        })(WatchLogLevel = ts.WatchLogLevel || (ts.WatchLogLevel = {}));
        function getWatchFactory(host, watchLogLevel, log, getDetailWatchInfo) {
            ts.setSysLog(watchLogLevel === WatchLogLevel.Verbose ? log : ts.noop);
            var plainInvokeFactory = {
                watchFile: function (file, callback, pollingInterval, options) { return host.watchFile(file, callback, pollingInterval, options); },
                watchDirectory: function (directory, callback, flags, options) { return host.watchDirectory(directory, callback, (flags & 1 /* Recursive */) !== 0, options); },
            };
            var triggerInvokingFactory = watchLogLevel !== WatchLogLevel.None ?
                {
                    watchFile: createTriggerLoggingAddWatch("watchFile"),
                    watchDirectory: createTriggerLoggingAddWatch("watchDirectory")
                } :
                undefined;
            var factory = watchLogLevel === WatchLogLevel.Verbose ?
                {
                    watchFile: createFileWatcherWithLogging,
                    watchDirectory: createDirectoryWatcherWithLogging
                } :
                triggerInvokingFactory || plainInvokeFactory;
            var excludeWatcherFactory = watchLogLevel === WatchLogLevel.Verbose ?
                createExcludeWatcherWithLogging :
                ts.returnNoopFileWatcher;
            return {
                watchFile: createExcludeHandlingAddWatch("watchFile"),
                watchDirectory: createExcludeHandlingAddWatch("watchDirectory")
            };
            function createExcludeHandlingAddWatch(key) {
                return function (file, cb, flags, options, detailInfo1, detailInfo2) {
                    var _a;
                    return !ts.matchesExclude(file, key === "watchFile" ? options === null || options === void 0 ? void 0 : options.excludeFiles : options === null || options === void 0 ? void 0 : options.excludeDirectories, useCaseSensitiveFileNames(), ((_a = host.getCurrentDirectory) === null || _a === void 0 ? void 0 : _a.call(host)) || "") ?
                        factory[key].call(/*thisArgs*/ undefined, file, cb, flags, options, detailInfo1, detailInfo2) :
                        excludeWatcherFactory(file, flags, options, detailInfo1, detailInfo2);
                };
            }
            function useCaseSensitiveFileNames() {
                return typeof host.useCaseSensitiveFileNames === "boolean" ?
                    host.useCaseSensitiveFileNames :
                    host.useCaseSensitiveFileNames();
            }
            function createExcludeWatcherWithLogging(file, flags, options, detailInfo1, detailInfo2) {
                log("ExcludeWatcher:: Added:: " + getWatchInfo(file, flags, options, detailInfo1, detailInfo2, getDetailWatchInfo));
                return {
                    close: function () { return log("ExcludeWatcher:: Close:: " + getWatchInfo(file, flags, options, detailInfo1, detailInfo2, getDetailWatchInfo)); }
                };
            }
            function createFileWatcherWithLogging(file, cb, flags, options, detailInfo1, detailInfo2) {
                log("FileWatcher:: Added:: " + getWatchInfo(file, flags, options, detailInfo1, detailInfo2, getDetailWatchInfo));
                var watcher = triggerInvokingFactory.watchFile(file, cb, flags, options, detailInfo1, detailInfo2);
                return {
                    close: function () {
                        log("FileWatcher:: Close:: " + getWatchInfo(file, flags, options, detailInfo1, detailInfo2, getDetailWatchInfo));
                        watcher.close();
                    }
                };
            }
            function createDirectoryWatcherWithLogging(file, cb, flags, options, detailInfo1, detailInfo2) {
                var watchInfo = "DirectoryWatcher:: Added:: " + getWatchInfo(file, flags, options, detailInfo1, detailInfo2, getDetailWatchInfo);
                log(watchInfo);
                var start = ts.timestamp();
                var watcher = triggerInvokingFactory.watchDirectory(file, cb, flags, options, detailInfo1, detailInfo2);
                var elapsed = ts.timestamp() - start;
                log("Elapsed:: " + elapsed + "ms " + watchInfo);
                return {
                    close: function () {
                        var watchInfo = "DirectoryWatcher:: Close:: " + getWatchInfo(file, flags, options, detailInfo1, detailInfo2, getDetailWatchInfo);
                        log(watchInfo);
                        var start = ts.timestamp();
                        watcher.close();
                        var elapsed = ts.timestamp() - start;
                        log("Elapsed:: " + elapsed + "ms " + watchInfo);
                    }
                };
            }
            function createTriggerLoggingAddWatch(key) {
                return function (file, cb, flags, options, detailInfo1, detailInfo2) {
                    return plainInvokeFactory[key].call(/*thisArgs*/ undefined, file, function () {
                        var args = [];
                        for (var _i = 0; _i < arguments.length; _i++) {
                            args[_i] = arguments[_i];
                        }
                        var triggerredInfo = (key === "watchFile" ? "FileWatcher" : "DirectoryWatcher") + ":: Triggered with " + args[0] + " " + (args[1] !== undefined ? args[1] : "") + ":: " + getWatchInfo(file, flags, options, detailInfo1, detailInfo2, getDetailWatchInfo);
                        log(triggerredInfo);
                        var start = ts.timestamp();
                        cb.call.apply(cb, __spreadArray([/*thisArg*/ undefined], args, false));
                        var elapsed = ts.timestamp() - start;
                        log("Elapsed:: " + elapsed + "ms " + triggerredInfo);
                    }, flags, options, detailInfo1, detailInfo2);
                };
            }
            function getWatchInfo(file, flags, options, detailInfo1, detailInfo2, getDetailWatchInfo) {
                return "WatchInfo: " + file + " " + flags + " " + JSON.stringify(options) + " " + (getDetailWatchInfo ? getDetailWatchInfo(detailInfo1, detailInfo2) : detailInfo2 === undefined ? detailInfo1 : detailInfo1 + " " + detailInfo2);
            }
        }
        ts.getWatchFactory = getWatchFactory;
        function getFallbackOptions(options) {
            var fallbackPolling = options === null || options === void 0 ? void 0 : options.fallbackPolling;
            return {
                watchFile: fallbackPolling !== undefined ?
                    fallbackPolling :
                    ts.WatchFileKind.PriorityPollingInterval
            };
        }
        ts.getFallbackOptions = getFallbackOptions;
        function closeFileWatcherOf(objWithWatcher) {
            objWithWatcher.watcher.close();
        }
        ts.closeFileWatcherOf = closeFileWatcherOf;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        function findConfigFile(searchPath, fileExists, configName) {
            if (configName === void 0) {
                configName = "tsconfig.json";
            }
            return ts.forEachAncestorDirectory(searchPath, function (ancestor) {
                var fileName = ts.combinePaths(ancestor, configName);
                return fileExists(fileName) ? fileName : undefined;
            });
        }
        ts.findConfigFile = findConfigFile;
        function resolveTripleslashReference(moduleName, containingFile) {
            var basePath = ts.getDirectoryPath(containingFile);
            var referencedFileName = ts.isRootedDiskPath(moduleName) ? moduleName : ts.combinePaths(basePath, moduleName);
            return ts.normalizePath(referencedFileName);
        }
        ts.resolveTripleslashReference = resolveTripleslashReference;
        /* @internal */
        function computeCommonSourceDirectoryOfFilenames(fileNames, currentDirectory, getCanonicalFileName) {
            var commonPathComponents;
            var failed = ts.forEach(fileNames, function (sourceFile) {
                // Each file contributes into common source file path
                var sourcePathComponents = ts.getNormalizedPathComponents(sourceFile, currentDirectory);
                sourcePathComponents.pop(); // The base file name is not part of the common directory path
                if (!commonPathComponents) {
                    // first file
                    commonPathComponents = sourcePathComponents;
                    return;
                }
                var n = Math.min(commonPathComponents.length, sourcePathComponents.length);
                for (var i = 0; i < n; i++) {
                    if (getCanonicalFileName(commonPathComponents[i]) !== getCanonicalFileName(sourcePathComponents[i])) {
                        if (i === 0) {
                            // Failed to find any common path component
                            return true;
                        }
                        // New common path found that is 0 -> i-1
                        commonPathComponents.length = i;
                        break;
                    }
                }
                // If the sourcePathComponents was shorter than the commonPathComponents, truncate to the sourcePathComponents
                if (sourcePathComponents.length < commonPathComponents.length) {
                    commonPathComponents.length = sourcePathComponents.length;
                }
            });
            // A common path can not be found when paths span multiple drives on windows, for example
            if (failed) {
                return "";
            }
            if (!commonPathComponents) { // Can happen when all input files are .d.ts files
                return currentDirectory;
            }
            return ts.getPathFromPathComponents(commonPathComponents);
        }
        ts.computeCommonSourceDirectoryOfFilenames = computeCommonSourceDirectoryOfFilenames;
        function createCompilerHost(options, setParentNodes) {
            return createCompilerHostWorker(options, setParentNodes);
        }
        ts.createCompilerHost = createCompilerHost;
        /*@internal*/
        // TODO(shkamat): update this after reworking ts build API
        function createCompilerHostWorker(options, setParentNodes, system) {
            if (system === void 0) {
                system = ts.sys;
            }
            var existingDirectories = new ts.Map();
            var getCanonicalFileName = ts.createGetCanonicalFileName(system.useCaseSensitiveFileNames);
            var computeHash = ts.maybeBind(system, system.createHash) || ts.generateDjb2Hash;
            function getSourceFile(fileName, languageVersion, onError) {
                var text;
                try {
                    ts.performance.mark("beforeIORead");
                    text = compilerHost.readFile(fileName);
                    ts.performance.mark("afterIORead");
                    ts.performance.measure("I/O Read", "beforeIORead", "afterIORead");
                }
                catch (e) {
                    if (onError) {
                        onError(e.message);
                    }
                    text = "";
                }
                return text !== undefined ? ts.createSourceFile(fileName, text, languageVersion, setParentNodes) : undefined;
            }
            function directoryExists(directoryPath) {
                if (existingDirectories.has(directoryPath)) {
                    return true;
                }
                if ((compilerHost.directoryExists || system.directoryExists)(directoryPath)) {
                    existingDirectories.set(directoryPath, true);
                    return true;
                }
                return false;
            }
            function writeFile(fileName, data, writeByteOrderMark, onError) {
                try {
                    ts.performance.mark("beforeIOWrite");
                    // NOTE: If patchWriteFileEnsuringDirectory has been called,
                    // the system.writeFile will do its own directory creation and
                    // the ensureDirectoriesExist call will always be redundant.
                    ts.writeFileEnsuringDirectories(fileName, data, writeByteOrderMark, function (path, data, writeByteOrderMark) { return writeFileWorker(path, data, writeByteOrderMark); }, function (path) { return (compilerHost.createDirectory || system.createDirectory)(path); }, function (path) { return directoryExists(path); });
                    ts.performance.mark("afterIOWrite");
                    ts.performance.measure("I/O Write", "beforeIOWrite", "afterIOWrite");
                }
                catch (e) {
                    if (onError) {
                        onError(e.message);
                    }
                }
            }
            var outputFingerprints;
            function writeFileWorker(fileName, data, writeByteOrderMark) {
                if (!ts.isWatchSet(options) || !system.getModifiedTime) {
                    system.writeFile(fileName, data, writeByteOrderMark);
                    return;
                }
                if (!outputFingerprints) {
                    outputFingerprints = new ts.Map();
                }
                var hash = computeHash(data);
                var mtimeBefore = system.getModifiedTime(fileName);
                if (mtimeBefore) {
                    var fingerprint = outputFingerprints.get(fileName);
                    // If output has not been changed, and the file has no external modification
                    if (fingerprint &&
                        fingerprint.byteOrderMark === writeByteOrderMark &&
                        fingerprint.hash === hash &&
                        fingerprint.mtime.getTime() === mtimeBefore.getTime()) {
                        return;
                    }
                }
                system.writeFile(fileName, data, writeByteOrderMark);
                var mtimeAfter = system.getModifiedTime(fileName) || ts.missingFileModifiedTime;
                outputFingerprints.set(fileName, {
                    hash: hash,
                    byteOrderMark: writeByteOrderMark,
                    mtime: mtimeAfter
                });
            }
            function getDefaultLibLocation() {
                return ts.getDirectoryPath(ts.normalizePath(system.getExecutingFilePath()));
            }
            var newLine = ts.getNewLineCharacter(options, function () { return system.newLine; });
            var realpath = system.realpath && (function (path) { return system.realpath(path); });
            var compilerHost = {
                getSourceFile: getSourceFile,
                getDefaultLibLocation: getDefaultLibLocation,
                getDefaultLibFileName: function (options) { return ts.combinePaths(getDefaultLibLocation(), ts.getDefaultLibFileName(options)); },
                writeFile: writeFile,
                getCurrentDirectory: ts.memoize(function () { return system.getCurrentDirectory(); }),
                useCaseSensitiveFileNames: function () { return system.useCaseSensitiveFileNames; },
                getCanonicalFileName: getCanonicalFileName,
                getNewLine: function () { return newLine; },
                fileExists: function (fileName) { return system.fileExists(fileName); },
                readFile: function (fileName) { return system.readFile(fileName); },
                trace: function (s) { return system.write(s + newLine); },
                directoryExists: function (directoryName) { return system.directoryExists(directoryName); },
                getEnvironmentVariable: function (name) { return system.getEnvironmentVariable ? system.getEnvironmentVariable(name) : ""; },
                getDirectories: function (path) { return system.getDirectories(path); },
                realpath: realpath,
                readDirectory: function (path, extensions, include, exclude, depth) { return system.readDirectory(path, extensions, include, exclude, depth); },
                createDirectory: function (d) { return system.createDirectory(d); },
                createHash: ts.maybeBind(system, system.createHash)
            };
            return compilerHost;
        }
        ts.createCompilerHostWorker = createCompilerHostWorker;
        /*@internal*/
        function changeCompilerHostLikeToUseCache(host, toPath, getSourceFile) {
            var originalReadFile = host.readFile;
            var originalFileExists = host.fileExists;
            var originalDirectoryExists = host.directoryExists;
            var originalCreateDirectory = host.createDirectory;
            var originalWriteFile = host.writeFile;
            var readFileCache = new ts.Map();
            var fileExistsCache = new ts.Map();
            var directoryExistsCache = new ts.Map();
            var sourceFileCache = new ts.Map();
            var readFileWithCache = function (fileName) {
                var key = toPath(fileName);
                var value = readFileCache.get(key);
                if (value !== undefined)
                    return value !== false ? value : undefined;
                return setReadFileCache(key, fileName);
            };
            var setReadFileCache = function (key, fileName) {
                var newValue = originalReadFile.call(host, fileName);
                readFileCache.set(key, newValue !== undefined ? newValue : false);
                return newValue;
            };
            host.readFile = function (fileName) {
                var key = toPath(fileName);
                var value = readFileCache.get(key);
                if (value !== undefined)
                    return value !== false ? value : undefined; // could be .d.ts from output
                // Cache json or buildInfo
                if (!ts.fileExtensionIs(fileName, ".json" /* Json */) && !ts.isBuildInfoFile(fileName)) {
                    return originalReadFile.call(host, fileName);
                }
                return setReadFileCache(key, fileName);
            };
            var getSourceFileWithCache = getSourceFile ? function (fileName, languageVersion, onError, shouldCreateNewSourceFile) {
                var key = toPath(fileName);
                var value = sourceFileCache.get(key);
                if (value)
                    return value;
                var sourceFile = getSourceFile(fileName, languageVersion, onError, shouldCreateNewSourceFile);
                if (sourceFile && (ts.isDeclarationFileName(fileName) || ts.fileExtensionIs(fileName, ".json" /* Json */))) {
                    sourceFileCache.set(key, sourceFile);
                }
                return sourceFile;
            } : undefined;
            // fileExists for any kind of extension
            host.fileExists = function (fileName) {
                var key = toPath(fileName);
                var value = fileExistsCache.get(key);
                if (value !== undefined)
                    return value;
                var newValue = originalFileExists.call(host, fileName);
                fileExistsCache.set(key, !!newValue);
                return newValue;
            };
            if (originalWriteFile) {
                host.writeFile = function (fileName, data, writeByteOrderMark, onError, sourceFiles) {
                    var key = toPath(fileName);
                    fileExistsCache.delete(key);
                    var value = readFileCache.get(key);
                    if (value !== undefined && value !== data) {
                        readFileCache.delete(key);
                        sourceFileCache.delete(key);
                    }
                    else if (getSourceFileWithCache) {
                        var sourceFile = sourceFileCache.get(key);
                        if (sourceFile && sourceFile.text !== data) {
                            sourceFileCache.delete(key);
                        }
                    }
                    originalWriteFile.call(host, fileName, data, writeByteOrderMark, onError, sourceFiles);
                };
            }
            // directoryExists
            if (originalDirectoryExists && originalCreateDirectory) {
                host.directoryExists = function (directory) {
                    var key = toPath(directory);
                    var value = directoryExistsCache.get(key);
                    if (value !== undefined)
                        return value;
                    var newValue = originalDirectoryExists.call(host, directory);
                    directoryExistsCache.set(key, !!newValue);
                    return newValue;
                };
                host.createDirectory = function (directory) {
                    var key = toPath(directory);
                    directoryExistsCache.delete(key);
                    originalCreateDirectory.call(host, directory);
                };
            }
            return {
                originalReadFile: originalReadFile,
                originalFileExists: originalFileExists,
                originalDirectoryExists: originalDirectoryExists,
                originalCreateDirectory: originalCreateDirectory,
                originalWriteFile: originalWriteFile,
                getSourceFileWithCache: getSourceFileWithCache,
                readFileWithCache: readFileWithCache
            };
        }
        ts.changeCompilerHostLikeToUseCache = changeCompilerHostLikeToUseCache;
        function getPreEmitDiagnostics(program, sourceFile, cancellationToken) {
            var diagnostics;
            diagnostics = ts.addRange(diagnostics, program.getConfigFileParsingDiagnostics());
            diagnostics = ts.addRange(diagnostics, program.getOptionsDiagnostics(cancellationToken));
            diagnostics = ts.addRange(diagnostics, program.getSyntacticDiagnostics(sourceFile, cancellationToken));
            diagnostics = ts.addRange(diagnostics, program.getGlobalDiagnostics(cancellationToken));
            diagnostics = ts.addRange(diagnostics, program.getSemanticDiagnostics(sourceFile, cancellationToken));
            if (ts.getEmitDeclarations(program.getCompilerOptions())) {
                diagnostics = ts.addRange(diagnostics, program.getDeclarationDiagnostics(sourceFile, cancellationToken));
            }
            return ts.sortAndDeduplicateDiagnostics(diagnostics || ts.emptyArray);
        }
        ts.getPreEmitDiagnostics = getPreEmitDiagnostics;
        function formatDiagnostics(diagnostics, host) {
            var output = "";
            for (var _i = 0, diagnostics_3 = diagnostics; _i < diagnostics_3.length; _i++) {
                var diagnostic = diagnostics_3[_i];
                output += formatDiagnostic(diagnostic, host);
            }
            return output;
        }
        ts.formatDiagnostics = formatDiagnostics;
        function formatDiagnostic(diagnostic, host) {
            var errorMessage = ts.diagnosticCategoryName(diagnostic) + " TS" + diagnostic.code + ": " + flattenDiagnosticMessageText(diagnostic.messageText, host.getNewLine()) + host.getNewLine();
            if (diagnostic.file) {
                var _a = ts.getLineAndCharacterOfPosition(diagnostic.file, diagnostic.start), line = _a.line, character = _a.character; // TODO: GH#18217
                var fileName = diagnostic.file.fileName;
                var relativeFileName = ts.convertToRelativePath(fileName, host.getCurrentDirectory(), function (fileName) { return host.getCanonicalFileName(fileName); });
                return relativeFileName + "(" + (line + 1) + "," + (character + 1) + "): " + errorMessage;
            }
            return errorMessage;
        }
        ts.formatDiagnostic = formatDiagnostic;
        /** @internal */
        var ForegroundColorEscapeSequences;
        (function (ForegroundColorEscapeSequences) {
            ForegroundColorEscapeSequences["Grey"] = "\u001B[90m";
            ForegroundColorEscapeSequences["Red"] = "\u001B[91m";
            ForegroundColorEscapeSequences["Yellow"] = "\u001B[93m";
            ForegroundColorEscapeSequences["Blue"] = "\u001B[94m";
            ForegroundColorEscapeSequences["Cyan"] = "\u001B[96m";
        })(ForegroundColorEscapeSequences = ts.ForegroundColorEscapeSequences || (ts.ForegroundColorEscapeSequences = {}));
        var gutterStyleSequence = "\u001b[7m";
        var gutterSeparator = " ";
        var resetEscapeSequence = "\u001b[0m";
        var ellipsis = "...";
        var halfIndent = "  ";
        var indent = "    ";
        function getCategoryFormat(category) {
            switch (category) {
                case ts.DiagnosticCategory.Error: return ForegroundColorEscapeSequences.Red;
                case ts.DiagnosticCategory.Warning: return ForegroundColorEscapeSequences.Yellow;
                case ts.DiagnosticCategory.Suggestion: return ts.Debug.fail("Should never get an Info diagnostic on the command line.");
                case ts.DiagnosticCategory.Message: return ForegroundColorEscapeSequences.Blue;
            }
        }
        /** @internal */
        function formatColorAndReset(text, formatStyle) {
            return formatStyle + text + resetEscapeSequence;
        }
        ts.formatColorAndReset = formatColorAndReset;
        function formatCodeSpan(file, start, length, indent, squiggleColor, host) {
            var _a = ts.getLineAndCharacterOfPosition(file, start), firstLine = _a.line, firstLineChar = _a.character;
            var _b = ts.getLineAndCharacterOfPosition(file, start + length), lastLine = _b.line, lastLineChar = _b.character;
            var lastLineInFile = ts.getLineAndCharacterOfPosition(file, file.text.length).line;
            var hasMoreThanFiveLines = (lastLine - firstLine) >= 4;
            var gutterWidth = (lastLine + 1 + "").length;
            if (hasMoreThanFiveLines) {
                gutterWidth = Math.max(ellipsis.length, gutterWidth);
            }
            var context = "";
            for (var i = firstLine; i <= lastLine; i++) {
                context += host.getNewLine();
                // If the error spans over 5 lines, we'll only show the first 2 and last 2 lines,
                // so we'll skip ahead to the second-to-last line.
                if (hasMoreThanFiveLines && firstLine + 1 < i && i < lastLine - 1) {
                    context += indent + formatColorAndReset(ts.padLeft(ellipsis, gutterWidth), gutterStyleSequence) + gutterSeparator + host.getNewLine();
                    i = lastLine - 1;
                }
                var lineStart = ts.getPositionOfLineAndCharacter(file, i, 0);
                var lineEnd = i < lastLineInFile ? ts.getPositionOfLineAndCharacter(file, i + 1, 0) : file.text.length;
                var lineContent = file.text.slice(lineStart, lineEnd);
                lineContent = ts.trimStringEnd(lineContent); // trim from end
                lineContent = lineContent.replace(/\t/g, " "); // convert tabs to single spaces
                // Output the gutter and the actual contents of the line.
                context += indent + formatColorAndReset(ts.padLeft(i + 1 + "", gutterWidth), gutterStyleSequence) + gutterSeparator;
                context += lineContent + host.getNewLine();
                // Output the gutter and the error span for the line using tildes.
                context += indent + formatColorAndReset(ts.padLeft("", gutterWidth), gutterStyleSequence) + gutterSeparator;
                context += squiggleColor;
                if (i === firstLine) {
                    // If we're on the last line, then limit it to the last character of the last line.
                    // Otherwise, we'll just squiggle the rest of the line, giving 'slice' no end position.
                    var lastCharForLine = i === lastLine ? lastLineChar : undefined;
                    context += lineContent.slice(0, firstLineChar).replace(/\S/g, " ");
                    context += lineContent.slice(firstLineChar, lastCharForLine).replace(/./g, "~");
                }
                else if (i === lastLine) {
                    context += lineContent.slice(0, lastLineChar).replace(/./g, "~");
                }
                else {
                    // Squiggle the entire line.
                    context += lineContent.replace(/./g, "~");
                }
                context += resetEscapeSequence;
            }
            return context;
        }
        /* @internal */
        function formatLocation(file, start, host, color) {
            if (color === void 0) {
                color = formatColorAndReset;
            }
            var _a = ts.getLineAndCharacterOfPosition(file, start), firstLine = _a.line, firstLineChar = _a.character; // TODO: GH#18217
            var relativeFileName = host ? ts.convertToRelativePath(file.fileName, host.getCurrentDirectory(), function (fileName) { return host.getCanonicalFileName(fileName); }) : file.fileName;
            var output = "";
            output += color(relativeFileName, ForegroundColorEscapeSequences.Cyan);
            output += ":";
            output += color("" + (firstLine + 1), ForegroundColorEscapeSequences.Yellow);
            output += ":";
            output += color("" + (firstLineChar + 1), ForegroundColorEscapeSequences.Yellow);
            return output;
        }
        ts.formatLocation = formatLocation;
        function formatDiagnosticsWithColorAndContext(diagnostics, host) {
            var output = "";
            for (var _i = 0, diagnostics_4 = diagnostics; _i < diagnostics_4.length; _i++) {
                var diagnostic = diagnostics_4[_i];
                if (diagnostic.file) {
                    var file = diagnostic.file, start = diagnostic.start;
                    output += formatLocation(file, start, host); // TODO: GH#18217
                    output += " - ";
                }
                output += formatColorAndReset(ts.diagnosticCategoryName(diagnostic), getCategoryFormat(diagnostic.category));
                output += formatColorAndReset(" TS" + diagnostic.code + ": ", ForegroundColorEscapeSequences.Grey);
                output += flattenDiagnosticMessageText(diagnostic.messageText, host.getNewLine());
                if (diagnostic.file) {
                    output += host.getNewLine();
                    output += formatCodeSpan(diagnostic.file, diagnostic.start, diagnostic.length, "", getCategoryFormat(diagnostic.category), host); // TODO: GH#18217
                }
                if (diagnostic.relatedInformation) {
                    output += host.getNewLine();
                    for (var _a = 0, _b = diagnostic.relatedInformation; _a < _b.length; _a++) {
                        var _c = _b[_a], file = _c.file, start = _c.start, length_9 = _c.length, messageText = _c.messageText;
                        if (file) {
                            output += host.getNewLine();
                            output += halfIndent + formatLocation(file, start, host); // TODO: GH#18217
                            output += formatCodeSpan(file, start, length_9, indent, ForegroundColorEscapeSequences.Cyan, host); // TODO: GH#18217
                        }
                        output += host.getNewLine();
                        output += indent + flattenDiagnosticMessageText(messageText, host.getNewLine());
                    }
                }
                output += host.getNewLine();
            }
            return output;
        }
        ts.formatDiagnosticsWithColorAndContext = formatDiagnosticsWithColorAndContext;
        function flattenDiagnosticMessageText(diag, newLine, indent) {
            if (indent === void 0) {
                indent = 0;
            }
            if (ts.isString(diag)) {
                return diag;
            }
            else if (diag === undefined) {
                return "";
            }
            var result = "";
            if (indent) {
                result += newLine;
                for (var i = 0; i < indent; i++) {
                    result += "  ";
                }
            }
            result += diag.messageText;
            indent++;
            if (diag.next) {
                for (var _i = 0, _a = diag.next; _i < _a.length; _i++) {
                    var kid = _a[_i];
                    result += flattenDiagnosticMessageText(kid, newLine, indent);
                }
            }
            return result;
        }
        ts.flattenDiagnosticMessageText = flattenDiagnosticMessageText;
        /* @internal */
        function loadWithLocalCache(names, containingFile, redirectedReference, loader) {
            if (names.length === 0) {
                return [];
            }
            var resolutions = [];
            var cache = new ts.Map();
            for (var _i = 0, names_2 = names; _i < names_2.length; _i++) {
                var name = names_2[_i];
                var result = void 0;
                if (cache.has(name)) {
                    result = cache.get(name);
                }
                else {
                    cache.set(name, result = loader(name, containingFile, redirectedReference));
                }
                resolutions.push(result);
            }
            return resolutions;
        }
        ts.loadWithLocalCache = loadWithLocalCache;
        /* @internal */
        function forEachResolvedProjectReference(resolvedProjectReferences, cb) {
            return forEachProjectReference(/*projectReferences*/ undefined, resolvedProjectReferences, function (resolvedRef, parent) { return resolvedRef && cb(resolvedRef, parent); });
        }
        ts.forEachResolvedProjectReference = forEachResolvedProjectReference;
        function forEachProjectReference(projectReferences, resolvedProjectReferences, cbResolvedRef, cbRef) {
            var seenResolvedRefs;
            return worker(projectReferences, resolvedProjectReferences, /*parent*/ undefined);
            function worker(projectReferences, resolvedProjectReferences, parent) {
                // Visit project references first
                if (cbRef) {
                    var result = cbRef(projectReferences, parent);
                    if (result) {
                        return result;
                    }
                }
                return ts.forEach(resolvedProjectReferences, function (resolvedRef, index) {
                    if (resolvedRef && (seenResolvedRefs === null || seenResolvedRefs === void 0 ? void 0 : seenResolvedRefs.has(resolvedRef.sourceFile.path))) {
                        // ignore recursives
                        return undefined;
                    }
                    var result = cbResolvedRef(resolvedRef, parent, index);
                    if (result || !resolvedRef)
                        return result;
                    (seenResolvedRefs || (seenResolvedRefs = new ts.Set())).add(resolvedRef.sourceFile.path);
                    return worker(resolvedRef.commandLine.projectReferences, resolvedRef.references, resolvedRef);
                });
            }
        }
        /* @internal */
        ts.inferredTypesContainingFile = "__inferred type names__.ts";
        /*@internal*/
        function isReferencedFile(reason) {
            switch (reason === null || reason === void 0 ? void 0 : reason.kind) {
                case ts.FileIncludeKind.Import:
                case ts.FileIncludeKind.ReferenceFile:
                case ts.FileIncludeKind.TypeReferenceDirective:
                case ts.FileIncludeKind.LibReferenceDirective:
                    return true;
                default:
                    return false;
            }
        }
        ts.isReferencedFile = isReferencedFile;
        /*@internal*/
        function isReferenceFileLocation(location) {
            return location.pos !== undefined;
        }
        ts.isReferenceFileLocation = isReferenceFileLocation;
        /*@internal*/
        function getReferencedFileLocation(getSourceFileByPath, ref) {
            var _a, _b, _c;
            var _d, _e, _f, _g;
            var file = ts.Debug.checkDefined(getSourceFileByPath(ref.file));
            var kind = ref.kind, index = ref.index;
            var pos, end, packageId;
            switch (kind) {
                case ts.FileIncludeKind.Import:
                    var importLiteral = getModuleNameStringLiteralAt(file, index);
                    packageId = (_e = (_d = file.resolvedModules) === null || _d === void 0 ? void 0 : _d.get(importLiteral.text)) === null || _e === void 0 ? void 0 : _e.packageId;
                    if (importLiteral.pos === -1)
                        return { file: file, packageId: packageId, text: importLiteral.text };
                    pos = ts.skipTrivia(file.text, importLiteral.pos);
                    end = importLiteral.end;
                    break;
                case ts.FileIncludeKind.ReferenceFile:
                    (_a = file.referencedFiles[index], pos = _a.pos, end = _a.end);
                    break;
                case ts.FileIncludeKind.TypeReferenceDirective:
                    (_b = file.typeReferenceDirectives[index], pos = _b.pos, end = _b.end);
                    packageId = (_g = (_f = file.resolvedTypeReferenceDirectiveNames) === null || _f === void 0 ? void 0 : _f.get(ts.toFileNameLowerCase(file.typeReferenceDirectives[index].fileName))) === null || _g === void 0 ? void 0 : _g.packageId;
                    break;
                case ts.FileIncludeKind.LibReferenceDirective:
                    (_c = file.libReferenceDirectives[index], pos = _c.pos, end = _c.end);
                    break;
                default:
                    return ts.Debug.assertNever(kind);
            }
            return { file: file, pos: pos, end: end, packageId: packageId };
        }
        ts.getReferencedFileLocation = getReferencedFileLocation;
        /**
         * Determines if program structure is upto date or needs to be recreated
         */
        /* @internal */
        function isProgramUptoDate(program, rootFileNames, newOptions, getSourceVersion, fileExists, hasInvalidatedResolution, hasChangedAutomaticTypeDirectiveNames, getParsedCommandLine, projectReferences) {
            // If we haven't created a program yet or have changed automatic type directives, then it is not up-to-date
            if (!program || (hasChangedAutomaticTypeDirectiveNames === null || hasChangedAutomaticTypeDirectiveNames === void 0 ? void 0 : hasChangedAutomaticTypeDirectiveNames()))
                return false;
            // If root file names don't match
            if (!ts.arrayIsEqualTo(program.getRootFileNames(), rootFileNames))
                return false;
            var seenResolvedRefs;
            // If project references don't match
            if (!ts.arrayIsEqualTo(program.getProjectReferences(), projectReferences, projectReferenceUptoDate))
                return false;
            // If any file is not up-to-date, then the whole program is not up-to-date
            if (program.getSourceFiles().some(sourceFileNotUptoDate))
                return false;
            // If any of the missing file paths are now created
            if (program.getMissingFilePaths().some(fileExists))
                return false;
            var currentOptions = program.getCompilerOptions();
            // If the compilation settings do no match, then the program is not up-to-date
            if (!ts.compareDataObjects(currentOptions, newOptions))
                return false;
            // If everything matches but the text of config file is changed,
            // error locations can change for program options, so update the program
            if (currentOptions.configFile && newOptions.configFile)
                return currentOptions.configFile.text === newOptions.configFile.text;
            return true;
            function sourceFileNotUptoDate(sourceFile) {
                return !sourceFileVersionUptoDate(sourceFile) ||
                    hasInvalidatedResolution(sourceFile.path);
            }
            function sourceFileVersionUptoDate(sourceFile) {
                return sourceFile.version === getSourceVersion(sourceFile.resolvedPath, sourceFile.fileName);
            }
            function projectReferenceUptoDate(oldRef, newRef, index) {
                return ts.projectReferenceIsEqualTo(oldRef, newRef) &&
                    resolvedProjectReferenceUptoDate(program.getResolvedProjectReferences()[index], oldRef);
            }
            function resolvedProjectReferenceUptoDate(oldResolvedRef, oldRef) {
                if (oldResolvedRef) {
                    // Assume true
                    if (ts.contains(seenResolvedRefs, oldResolvedRef))
                        return true;
                    var refPath_1 = resolveProjectReferencePath(oldRef);
                    var newParsedCommandLine = getParsedCommandLine(refPath_1);
                    // Check if config file exists
                    if (!newParsedCommandLine)
                        return false;
                    // If change in source file
                    if (oldResolvedRef.commandLine.options.configFile !== newParsedCommandLine.options.configFile)
                        return false;
                    // check file names
                    if (!ts.arrayIsEqualTo(oldResolvedRef.commandLine.fileNames, newParsedCommandLine.fileNames))
                        return false;
                    // Add to seen before checking the referenced paths of this config file
                    (seenResolvedRefs || (seenResolvedRefs = [])).push(oldResolvedRef);
                    // If child project references are upto date, this project reference is uptodate
                    return !ts.forEach(oldResolvedRef.references, function (childResolvedRef, index) {
                        return !resolvedProjectReferenceUptoDate(childResolvedRef, oldResolvedRef.commandLine.projectReferences[index]);
                    });
                }
                // In old program, not able to resolve project reference path,
                // so if config file doesnt exist, it is uptodate.
                var refPath = resolveProjectReferencePath(oldRef);
                return !getParsedCommandLine(refPath);
            }
        }
        ts.isProgramUptoDate = isProgramUptoDate;
        function getConfigFileParsingDiagnostics(configFileParseResult) {
            return configFileParseResult.options.configFile ? __spreadArray(__spreadArray([], configFileParseResult.options.configFile.parseDiagnostics, true), configFileParseResult.errors, true) :
                configFileParseResult.errors;
        }
        ts.getConfigFileParsingDiagnostics = getConfigFileParsingDiagnostics;
        /**
         * Determine if source file needs to be re-created even if its text hasn't changed
         */
        function shouldProgramCreateNewSourceFiles(program, newOptions) {
            if (!program)
                return false;
            // If any compiler options change, we can't reuse old source file even if version match
            // The change in options like these could result in change in syntax tree or `sourceFile.bindDiagnostics`.
            return ts.optionsHaveChanges(program.getCompilerOptions(), newOptions, ts.sourceFileAffectingCompilerOptions);
        }
        function createCreateProgramOptions(rootNames, options, host, oldProgram, configFileParsingDiagnostics) {
            return {
                rootNames: rootNames,
                options: options,
                host: host,
                oldProgram: oldProgram,
                configFileParsingDiagnostics: configFileParsingDiagnostics
            };
        }
        function createProgram(rootNamesOrOptions, _options, _host, _oldProgram, _configFileParsingDiagnostics) {
            var _a, _b, _c;
            var createProgramOptions = ts.isArray(rootNamesOrOptions) ? createCreateProgramOptions(rootNamesOrOptions, _options, _host, _oldProgram, _configFileParsingDiagnostics) : rootNamesOrOptions; // TODO: GH#18217
            var rootNames = createProgramOptions.rootNames, options = createProgramOptions.options, configFileParsingDiagnostics = createProgramOptions.configFileParsingDiagnostics, projectReferences = createProgramOptions.projectReferences;
            var oldProgram = createProgramOptions.oldProgram;
            var processingDefaultLibFiles;
            var processingOtherFiles;
            var files;
            var symlinks;
            var commonSourceDirectory;
            var diagnosticsProducingTypeChecker;
            var noDiagnosticsTypeChecker;
            var classifiableNames;
            var ambientModuleNameToUnmodifiedFileName = new ts.Map();
            var fileReasons = ts.createMultiMap();
            var cachedBindAndCheckDiagnosticsForFile = {};
            var cachedDeclarationDiagnosticsForFile = {};
            var resolvedTypeReferenceDirectives = new ts.Map();
            var fileProcessingDiagnostics;
            // The below settings are to track if a .js file should be add to the program if loaded via searching under node_modules.
            // This works as imported modules are discovered recursively in a depth first manner, specifically:
            // - For each root file, findSourceFile is called.
            // - This calls processImportedModules for each module imported in the source file.
            // - This calls resolveModuleNames, and then calls findSourceFile for each resolved module.
            // As all these operations happen - and are nested - within the createProgram call, they close over the below variables.
            // The current resolution depth is tracked by incrementing/decrementing as the depth first search progresses.
            var maxNodeModuleJsDepth = typeof options.maxNodeModuleJsDepth === "number" ? options.maxNodeModuleJsDepth : 0;
            var currentNodeModulesDepth = 0;
            // If a module has some of its imports skipped due to being at the depth limit under node_modules, then track
            // this, as it may be imported at a shallower depth later, and then it will need its skipped imports processed.
            var modulesWithElidedImports = new ts.Map();
            // Track source files that are source files found by searching under node_modules, as these shouldn't be compiled.
            var sourceFilesFoundSearchingNodeModules = new ts.Map();
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("program" /* Program */, "createProgram", { configFilePath: options.configFilePath, rootDir: options.rootDir }, /*separateBeginAndEnd*/ true);
            ts.performance.mark("beforeProgram");
            var host = createProgramOptions.host || createCompilerHost(options);
            var configParsingHost = parseConfigHostFromCompilerHostLike(host);
            var skipDefaultLib = options.noLib;
            var getDefaultLibraryFileName = ts.memoize(function () { return host.getDefaultLibFileName(options); });
            var defaultLibraryPath = host.getDefaultLibLocation ? host.getDefaultLibLocation() : ts.getDirectoryPath(getDefaultLibraryFileName());
            var programDiagnostics = ts.createDiagnosticCollection();
            var currentDirectory = host.getCurrentDirectory();
            var supportedExtensions = ts.getSupportedExtensions(options);
            var supportedExtensionsWithJsonIfResolveJsonModule = ts.getSuppoertedExtensionsWithJsonIfResolveJsonModule(options, supportedExtensions);
            // Map storing if there is emit blocking diagnostics for given input
            var hasEmitBlockingDiagnostics = new ts.Map();
            var _compilerOptionsObjectLiteralSyntax;
            var moduleResolutionCache;
            var typeReferenceDirectiveResolutionCache;
            var actualResolveModuleNamesWorker;
            var hasInvalidatedResolution = host.hasInvalidatedResolution || ts.returnFalse;
            if (host.resolveModuleNames) {
                actualResolveModuleNamesWorker = function (moduleNames, containingFile, reusedNames, redirectedReference) {
                    return host.resolveModuleNames(ts.Debug.checkEachDefined(moduleNames), containingFile, reusedNames, redirectedReference, options).map(function (resolved) {
                        // An older host may have omitted extension, in which case we should infer it from the file extension of resolvedFileName.
                        if (!resolved || resolved.extension !== undefined) {
                            return resolved;
                        }
                        var withExtension = ts.clone(resolved);
                        withExtension.extension = ts.extensionFromPath(resolved.resolvedFileName);
                        return withExtension;
                    });
                };
            }
            else {
                moduleResolutionCache = ts.createModuleResolutionCache(currentDirectory, getCanonicalFileName, options);
                var loader_1 = function (moduleName, containingFile, redirectedReference) { return ts.resolveModuleName(moduleName, containingFile, options, host, moduleResolutionCache, redirectedReference).resolvedModule; }; // TODO: GH#18217
                actualResolveModuleNamesWorker = function (moduleNames, containingFile, _reusedNames, redirectedReference) { return loadWithLocalCache(ts.Debug.checkEachDefined(moduleNames), containingFile, redirectedReference, loader_1); };
            }
            var actualResolveTypeReferenceDirectiveNamesWorker;
            if (host.resolveTypeReferenceDirectives) {
                actualResolveTypeReferenceDirectiveNamesWorker = function (typeDirectiveNames, containingFile, redirectedReference) { return host.resolveTypeReferenceDirectives(ts.Debug.checkEachDefined(typeDirectiveNames), containingFile, redirectedReference, options); };
            }
            else {
                typeReferenceDirectiveResolutionCache = ts.createTypeReferenceDirectiveResolutionCache(currentDirectory, getCanonicalFileName, /*options*/ undefined, moduleResolutionCache === null || moduleResolutionCache === void 0 ? void 0 : moduleResolutionCache.getPackageJsonInfoCache());
                var loader_2 = function (typesRef, containingFile, redirectedReference) { return ts.resolveTypeReferenceDirective(typesRef, containingFile, options, host, redirectedReference, typeReferenceDirectiveResolutionCache).resolvedTypeReferenceDirective; }; // TODO: GH#18217
                actualResolveTypeReferenceDirectiveNamesWorker = function (typeReferenceDirectiveNames, containingFile, redirectedReference) { return loadWithLocalCache(ts.Debug.checkEachDefined(typeReferenceDirectiveNames), containingFile, redirectedReference, loader_2); };
            }
            // Map from a stringified PackageId to the source file with that id.
            // Only one source file may have a given packageId. Others become redirects (see createRedirectSourceFile).
            // `packageIdToSourceFile` is only used while building the program, while `sourceFileToPackageName` and `isSourceFileTargetOfRedirect` are kept around.
            var packageIdToSourceFile = new ts.Map();
            // Maps from a SourceFile's `.path` to the name of the package it was imported with.
            var sourceFileToPackageName = new ts.Map();
            // Key is a file name. Value is the (non-empty, or undefined) list of files that redirect to it.
            var redirectTargetsMap = ts.createMultiMap();
            var usesUriStyleNodeCoreModules = false;
            /**
             * map with
             * - SourceFile if present
             * - false if sourceFile missing for source of project reference redirect
             * - undefined otherwise
             */
            var filesByName = new ts.Map();
            var missingFilePaths;
            // stores 'filename -> file association' ignoring case
            // used to track cases when two file names differ only in casing
            var filesByNameIgnoreCase = host.useCaseSensitiveFileNames() ? new ts.Map() : undefined;
            // A parallel array to projectReferences storing the results of reading in the referenced tsconfig files
            var resolvedProjectReferences;
            var projectReferenceRedirects;
            var mapFromFileToProjectReferenceRedirects;
            var mapFromToProjectReferenceRedirectSource;
            var useSourceOfProjectReferenceRedirect = !!((_a = host.useSourceOfProjectReferenceRedirect) === null || _a === void 0 ? void 0 : _a.call(host)) &&
                !options.disableSourceOfProjectReferenceRedirect;
            var _d = updateHostForUseSourceOfProjectReferenceRedirect({
                compilerHost: host,
                getSymlinkCache: getSymlinkCache,
                useSourceOfProjectReferenceRedirect: useSourceOfProjectReferenceRedirect,
                toPath: toPath,
                getResolvedProjectReferences: getResolvedProjectReferences,
                getSourceOfProjectReferenceRedirect: getSourceOfProjectReferenceRedirect,
                forEachResolvedProjectReference: forEachResolvedProjectReference
            }), onProgramCreateComplete = _d.onProgramCreateComplete, fileExists = _d.fileExists, directoryExists = _d.directoryExists;
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("program" /* Program */, "shouldProgramCreateNewSourceFiles", { hasOldProgram: !!oldProgram });
            var shouldCreateNewSourceFile = shouldProgramCreateNewSourceFiles(oldProgram, options);
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
            // We set `structuralIsReused` to `undefined` because `tryReuseStructureFromOldProgram` calls `tryReuseStructureFromOldProgram` which checks
            // `structuralIsReused`, which would be a TDZ violation if it was not set in advance to `undefined`.
            var structureIsReused;
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("program" /* Program */, "tryReuseStructureFromOldProgram", {});
            structureIsReused = tryReuseStructureFromOldProgram(); // eslint-disable-line prefer-const
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
            if (structureIsReused !== 2 /* Completely */) {
                processingDefaultLibFiles = [];
                processingOtherFiles = [];
                if (projectReferences) {
                    if (!resolvedProjectReferences) {
                        resolvedProjectReferences = projectReferences.map(parseProjectReferenceConfigFile);
                    }
                    if (rootNames.length) {
                        resolvedProjectReferences === null || resolvedProjectReferences === void 0 ? void 0 : resolvedProjectReferences.forEach(function (parsedRef, index) {
                            if (!parsedRef)
                                return;
                            var out = ts.outFile(parsedRef.commandLine.options);
                            if (useSourceOfProjectReferenceRedirect) {
                                if (out || ts.getEmitModuleKind(parsedRef.commandLine.options) === ts.ModuleKind.None) {
                                    for (var _i = 0, _a = parsedRef.commandLine.fileNames; _i < _a.length; _i++) {
                                        var fileName = _a[_i];
                                        processProjectReferenceFile(fileName, { kind: ts.FileIncludeKind.SourceFromProjectReference, index: index });
                                    }
                                }
                            }
                            else {
                                if (out) {
                                    processProjectReferenceFile(ts.changeExtension(out, ".d.ts"), { kind: ts.FileIncludeKind.OutputFromProjectReference, index: index });
                                }
                                else if (ts.getEmitModuleKind(parsedRef.commandLine.options) === ts.ModuleKind.None) {
                                    var getCommonSourceDirectory_2 = ts.memoize(function () { return ts.getCommonSourceDirectoryOfConfig(parsedRef.commandLine, !host.useCaseSensitiveFileNames()); });
                                    for (var _b = 0, _c = parsedRef.commandLine.fileNames; _b < _c.length; _b++) {
                                        var fileName = _c[_b];
                                        if (!ts.fileExtensionIs(fileName, ".d.ts" /* Dts */) && !ts.fileExtensionIs(fileName, ".json" /* Json */)) {
                                            processProjectReferenceFile(ts.getOutputDeclarationFileName(fileName, parsedRef.commandLine, !host.useCaseSensitiveFileNames(), getCommonSourceDirectory_2), { kind: ts.FileIncludeKind.OutputFromProjectReference, index: index });
                                        }
                                    }
                                }
                            }
                        });
                    }
                }
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("program" /* Program */, "processRootFiles", { count: rootNames.length });
                ts.forEach(rootNames, function (name, index) { return processRootFile(name, /*isDefaultLib*/ false, /*ignoreNoDefaultLib*/ false, { kind: ts.FileIncludeKind.RootFile, index: index }); });
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                // load type declarations specified via 'types' argument or implicitly from types/ and node_modules/@types folders
                var typeReferences = rootNames.length ? ts.getAutomaticTypeDirectiveNames(options, host) : ts.emptyArray;
                if (typeReferences.length) {
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("program" /* Program */, "processTypeReferences", { count: typeReferences.length });
                    // This containingFilename needs to match with the one used in managed-side
                    var containingDirectory = options.configFilePath ? ts.getDirectoryPath(options.configFilePath) : host.getCurrentDirectory();
                    var containingFilename = ts.combinePaths(containingDirectory, ts.inferredTypesContainingFile);
                    var resolutions = resolveTypeReferenceDirectiveNamesWorker(typeReferences, containingFilename);
                    for (var i = 0; i < typeReferences.length; i++) {
                        processTypeReferenceDirective(typeReferences[i], resolutions[i], { kind: ts.FileIncludeKind.AutomaticTypeDirectiveFile, typeReference: typeReferences[i], packageId: (_b = resolutions[i]) === null || _b === void 0 ? void 0 : _b.packageId });
                    }
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                }
                // Do not process the default library if:
                //  - The '--noLib' flag is used.
                //  - A 'no-default-lib' reference comment is encountered in
                //      processing the root files.
                if (rootNames.length && !skipDefaultLib) {
                    // If '--lib' is not specified, include default library file according to '--target'
                    // otherwise, using options specified in '--lib' instead of '--target' default library file
                    var defaultLibraryFileName = getDefaultLibraryFileName();
                    if (!options.lib && defaultLibraryFileName) {
                        processRootFile(defaultLibraryFileName, /*isDefaultLib*/ true, /*ignoreNoDefaultLib*/ false, { kind: ts.FileIncludeKind.LibFile });
                    }
                    else {
                        ts.forEach(options.lib, function (libFileName, index) {
                            processRootFile(ts.combinePaths(defaultLibraryPath, libFileName), /*isDefaultLib*/ true, /*ignoreNoDefaultLib*/ false, { kind: ts.FileIncludeKind.LibFile, index: index });
                        });
                    }
                }
                missingFilePaths = ts.arrayFrom(ts.mapDefinedIterator(filesByName.entries(), function (_a) {
                    var path = _a[0], file = _a[1];
                    return file === undefined ? path : undefined;
                }));
                files = ts.stableSort(processingDefaultLibFiles, compareDefaultLibFiles).concat(processingOtherFiles);
                processingDefaultLibFiles = undefined;
                processingOtherFiles = undefined;
            }
            ts.Debug.assert(!!missingFilePaths);
            // Release any files we have acquired in the old program but are
            // not part of the new program.
            if (oldProgram && host.onReleaseOldSourceFile) {
                var oldSourceFiles = oldProgram.getSourceFiles();
                for (var _i = 0, oldSourceFiles_1 = oldSourceFiles; _i < oldSourceFiles_1.length; _i++) {
                    var oldSourceFile = oldSourceFiles_1[_i];
                    var newFile = getSourceFileByPath(oldSourceFile.resolvedPath);
                    if (shouldCreateNewSourceFile || !newFile ||
                        // old file wasn't redirect but new file is
                        (oldSourceFile.resolvedPath === oldSourceFile.path && newFile.resolvedPath !== oldSourceFile.path)) {
                        host.onReleaseOldSourceFile(oldSourceFile, oldProgram.getCompilerOptions(), !!getSourceFileByPath(oldSourceFile.path));
                    }
                }
                if (!host.getParsedCommandLine) {
                    oldProgram.forEachResolvedProjectReference(function (resolvedProjectReference) {
                        if (!getResolvedProjectReferenceByPath(resolvedProjectReference.sourceFile.path)) {
                            host.onReleaseOldSourceFile(resolvedProjectReference.sourceFile, oldProgram.getCompilerOptions(), /*hasSourceFileByPath*/ false);
                        }
                    });
                }
            }
            // Release commandlines that new program does not use
            if (oldProgram && host.onReleaseParsedCommandLine) {
                forEachProjectReference(oldProgram.getProjectReferences(), oldProgram.getResolvedProjectReferences(), function (oldResolvedRef, parent, index) {
                    var oldReference = (parent === null || parent === void 0 ? void 0 : parent.commandLine.projectReferences[index]) || oldProgram.getProjectReferences()[index];
                    var oldRefPath = resolveProjectReferencePath(oldReference);
                    if (!(projectReferenceRedirects === null || projectReferenceRedirects === void 0 ? void 0 : projectReferenceRedirects.has(toPath(oldRefPath)))) {
                        host.onReleaseParsedCommandLine(oldRefPath, oldResolvedRef, oldProgram.getCompilerOptions());
                    }
                });
            }
            typeReferenceDirectiveResolutionCache = undefined;
            // unconditionally set oldProgram to undefined to prevent it from being captured in closure
            oldProgram = undefined;
            var program = {
                getRootFileNames: function () { return rootNames; },
                getSourceFile: getSourceFile,
                getSourceFileByPath: getSourceFileByPath,
                getSourceFiles: function () { return files; },
                getMissingFilePaths: function () { return missingFilePaths; },
                getModuleResolutionCache: function () { return moduleResolutionCache; },
                getFilesByNameMap: function () { return filesByName; },
                getCompilerOptions: function () { return options; },
                getSyntacticDiagnostics: getSyntacticDiagnostics,
                getOptionsDiagnostics: getOptionsDiagnostics,
                getGlobalDiagnostics: getGlobalDiagnostics,
                getSemanticDiagnostics: getSemanticDiagnostics,
                getCachedSemanticDiagnostics: getCachedSemanticDiagnostics,
                getSuggestionDiagnostics: getSuggestionDiagnostics,
                getDeclarationDiagnostics: getDeclarationDiagnostics,
                getBindAndCheckDiagnostics: getBindAndCheckDiagnostics,
                getProgramDiagnostics: getProgramDiagnostics,
                getTypeChecker: getTypeChecker,
                getClassifiableNames: getClassifiableNames,
                getDiagnosticsProducingTypeChecker: getDiagnosticsProducingTypeChecker,
                getCommonSourceDirectory: getCommonSourceDirectory,
                emit: emit,
                getCurrentDirectory: function () { return currentDirectory; },
                getNodeCount: function () { return getDiagnosticsProducingTypeChecker().getNodeCount(); },
                getIdentifierCount: function () { return getDiagnosticsProducingTypeChecker().getIdentifierCount(); },
                getSymbolCount: function () { return getDiagnosticsProducingTypeChecker().getSymbolCount(); },
                getTypeCount: function () { return getDiagnosticsProducingTypeChecker().getTypeCount(); },
                getInstantiationCount: function () { return getDiagnosticsProducingTypeChecker().getInstantiationCount(); },
                getRelationCacheSizes: function () { return getDiagnosticsProducingTypeChecker().getRelationCacheSizes(); },
                getFileProcessingDiagnostics: function () { return fileProcessingDiagnostics; },
                getResolvedTypeReferenceDirectives: function () { return resolvedTypeReferenceDirectives; },
                isSourceFileFromExternalLibrary: isSourceFileFromExternalLibrary,
                isSourceFileDefaultLibrary: isSourceFileDefaultLibrary,
                dropDiagnosticsProducingTypeChecker: dropDiagnosticsProducingTypeChecker,
                getSourceFileFromReference: getSourceFileFromReference,
                getLibFileFromReference: getLibFileFromReference,
                sourceFileToPackageName: sourceFileToPackageName,
                redirectTargetsMap: redirectTargetsMap,
                usesUriStyleNodeCoreModules: usesUriStyleNodeCoreModules,
                isEmittedFile: isEmittedFile,
                getConfigFileParsingDiagnostics: getConfigFileParsingDiagnostics,
                getResolvedModuleWithFailedLookupLocationsFromCache: getResolvedModuleWithFailedLookupLocationsFromCache,
                getProjectReferences: getProjectReferences,
                getResolvedProjectReferences: getResolvedProjectReferences,
                getProjectReferenceRedirect: getProjectReferenceRedirect,
                getResolvedProjectReferenceToRedirect: getResolvedProjectReferenceToRedirect,
                getResolvedProjectReferenceByPath: getResolvedProjectReferenceByPath,
                forEachResolvedProjectReference: forEachResolvedProjectReference,
                isSourceOfProjectReferenceRedirect: isSourceOfProjectReferenceRedirect,
                emitBuildInfo: emitBuildInfo,
                fileExists: fileExists,
                directoryExists: directoryExists,
                getSymlinkCache: getSymlinkCache,
                realpath: (_c = host.realpath) === null || _c === void 0 ? void 0 : _c.bind(host),
                useCaseSensitiveFileNames: function () { return host.useCaseSensitiveFileNames(); },
                getFileIncludeReasons: function () { return fileReasons; },
                structureIsReused: structureIsReused,
            };
            onProgramCreateComplete();
            // Add file processingDiagnostics
            fileProcessingDiagnostics === null || fileProcessingDiagnostics === void 0 ? void 0 : fileProcessingDiagnostics.forEach(function (diagnostic) {
                switch (diagnostic.kind) {
                    case 1 /* FilePreprocessingFileExplainingDiagnostic */:
                        return programDiagnostics.add(createDiagnosticExplainingFile(diagnostic.file && getSourceFileByPath(diagnostic.file), diagnostic.fileProcessingReason, diagnostic.diagnostic, diagnostic.args || ts.emptyArray));
                    case 0 /* FilePreprocessingReferencedDiagnostic */:
                        var _a = getReferencedFileLocation(getSourceFileByPath, diagnostic.reason), file = _a.file, pos = _a.pos, end = _a.end;
                        return programDiagnostics.add(ts.createFileDiagnostic.apply(void 0, __spreadArray([file, ts.Debug.checkDefined(pos), ts.Debug.checkDefined(end) - pos, diagnostic.diagnostic], diagnostic.args || ts.emptyArray, false)));
                    default:
                        ts.Debug.assertNever(diagnostic);
                }
            });
            verifyCompilerOptions();
            ts.performance.mark("afterProgram");
            ts.performance.measure("Program", "beforeProgram", "afterProgram");
            ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
            return program;
            function resolveModuleNamesWorker(moduleNames, containingFile, reusedNames) {
                if (!moduleNames.length)
                    return ts.emptyArray;
                var containingFileName = ts.getNormalizedAbsolutePath(containingFile.originalFileName, currentDirectory);
                var redirectedReference = getRedirectReferenceForResolution(containingFile);
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("program" /* Program */, "resolveModuleNamesWorker", { containingFileName: containingFileName });
                ts.performance.mark("beforeResolveModule");
                var result = actualResolveModuleNamesWorker(moduleNames, containingFileName, reusedNames, redirectedReference);
                ts.performance.mark("afterResolveModule");
                ts.performance.measure("ResolveModule", "beforeResolveModule", "afterResolveModule");
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                return result;
            }
            function resolveTypeReferenceDirectiveNamesWorker(typeDirectiveNames, containingFile) {
                if (!typeDirectiveNames.length)
                    return [];
                var containingFileName = !ts.isString(containingFile) ? ts.getNormalizedAbsolutePath(containingFile.originalFileName, currentDirectory) : containingFile;
                var redirectedReference = !ts.isString(containingFile) ? getRedirectReferenceForResolution(containingFile) : undefined;
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("program" /* Program */, "resolveTypeReferenceDirectiveNamesWorker", { containingFileName: containingFileName });
                ts.performance.mark("beforeResolveTypeReference");
                var result = actualResolveTypeReferenceDirectiveNamesWorker(typeDirectiveNames, containingFileName, redirectedReference);
                ts.performance.mark("afterResolveTypeReference");
                ts.performance.measure("ResolveTypeReference", "beforeResolveTypeReference", "afterResolveTypeReference");
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                return result;
            }
            function getRedirectReferenceForResolution(file) {
                var redirect = getResolvedProjectReferenceToRedirect(file.originalFileName);
                if (redirect || !ts.fileExtensionIs(file.originalFileName, ".d.ts" /* Dts */))
                    return redirect;
                // The originalFileName could not be actual source file name if file found was d.ts from referecned project
                // So in this case try to look up if this is output from referenced project, if it is use the redirected project in that case
                var resultFromDts = getRedirectReferenceForResolutionFromSourceOfProject(file.originalFileName, file.path);
                if (resultFromDts)
                    return resultFromDts;
                // If preserveSymlinks is true, module resolution wont jump the symlink
                // but the resolved real path may be the .d.ts from project reference
                // Note:: Currently we try the real path only if the
                // file is from node_modules to avoid having to run real path on all file paths
                if (!host.realpath || !options.preserveSymlinks || !ts.stringContains(file.originalFileName, ts.nodeModulesPathPart))
                    return undefined;
                var realDeclarationFileName = host.realpath(file.originalFileName);
                var realDeclarationPath = toPath(realDeclarationFileName);
                return realDeclarationPath === file.path ? undefined : getRedirectReferenceForResolutionFromSourceOfProject(realDeclarationFileName, realDeclarationPath);
            }
            function getRedirectReferenceForResolutionFromSourceOfProject(fileName, filePath) {
                var source = getSourceOfProjectReferenceRedirect(fileName);
                if (ts.isString(source))
                    return getResolvedProjectReferenceToRedirect(source);
                if (!source)
                    return undefined;
                // Output of .d.ts file so return resolved ref that matches the out file name
                return forEachResolvedProjectReference(function (resolvedRef) {
                    var out = ts.outFile(resolvedRef.commandLine.options);
                    if (!out)
                        return undefined;
                    return toPath(out) === filePath ? resolvedRef : undefined;
                });
            }
            function compareDefaultLibFiles(a, b) {
                return ts.compareValues(getDefaultLibFilePriority(a), getDefaultLibFilePriority(b));
            }
            function getDefaultLibFilePriority(a) {
                if (ts.containsPath(defaultLibraryPath, a.fileName, /*ignoreCase*/ false)) {
                    var basename = ts.getBaseFileName(a.fileName);
                    if (basename === "lib.d.ts" || basename === "lib.es6.d.ts")
                        return 0;
                    var name = ts.removeSuffix(ts.removePrefix(basename, "lib."), ".d.ts");
                    var index = ts.libs.indexOf(name);
                    if (index !== -1)
                        return index + 1;
                }
                return ts.libs.length + 2;
            }
            function getResolvedModuleWithFailedLookupLocationsFromCache(moduleName, containingFile) {
                return moduleResolutionCache && ts.resolveModuleNameFromCache(moduleName, containingFile, moduleResolutionCache);
            }
            function toPath(fileName) {
                return ts.toPath(fileName, currentDirectory, getCanonicalFileName);
            }
            function getCommonSourceDirectory() {
                if (commonSourceDirectory === undefined) {
                    var emittedFiles_1 = ts.filter(files, function (file) { return ts.sourceFileMayBeEmitted(file, program); });
                    commonSourceDirectory = ts.getCommonSourceDirectory(options, function () { return ts.mapDefined(emittedFiles_1, function (file) { return file.isDeclarationFile ? undefined : file.fileName; }); }, currentDirectory, getCanonicalFileName, function (commonSourceDirectory) { return checkSourceFilesBelongToPath(emittedFiles_1, commonSourceDirectory); });
                }
                return commonSourceDirectory;
            }
            function getClassifiableNames() {
                var _a;
                if (!classifiableNames) {
                    // Initialize a checker so that all our files are bound.
                    getTypeChecker();
                    classifiableNames = new ts.Set();
                    for (var _i = 0, files_3 = files; _i < files_3.length; _i++) {
                        var sourceFile = files_3[_i];
                        (_a = sourceFile.classifiableNames) === null || _a === void 0 ? void 0 : _a.forEach(function (value) { return classifiableNames.add(value); });
                    }
                }
                return classifiableNames;
            }
            function resolveModuleNamesReusingOldState(moduleNames, file) {
                if (structureIsReused === 0 /* Not */ && !file.ambientModuleNames.length) {
                    // If the old program state does not permit reusing resolutions and `file` does not contain locally defined ambient modules,
                    // the best we can do is fallback to the default logic.
                    return resolveModuleNamesWorker(moduleNames, file, /*reusedNames*/ undefined);
                }
                var oldSourceFile = oldProgram && oldProgram.getSourceFile(file.fileName);
                if (oldSourceFile !== file && file.resolvedModules) {
                    // `file` was created for the new program.
                    //
                    // We only set `file.resolvedModules` via work from the current function,
                    // so it is defined iff we already called the current function on `file`.
                    // That call happened no later than the creation of the `file` object,
                    // which per above occurred during the current program creation.
                    // Since we assume the filesystem does not change during program creation,
                    // it is safe to reuse resolutions from the earlier call.
                    var result_14 = [];
                    for (var _i = 0, moduleNames_1 = moduleNames; _i < moduleNames_1.length; _i++) {
                        var moduleName = moduleNames_1[_i];
                        var resolvedModule = file.resolvedModules.get(moduleName);
                        result_14.push(resolvedModule);
                    }
                    return result_14;
                }
                // At this point, we know at least one of the following hold:
                // - file has local declarations for ambient modules
                // - old program state is available
                // With this information, we can infer some module resolutions without performing resolution.
                /** An ordered list of module names for which we cannot recover the resolution. */
                var unknownModuleNames;
                /**
                 * The indexing of elements in this list matches that of `moduleNames`.
                 *
                 * Before combining results, result[i] is in one of the following states:
                 * * undefined: needs to be recomputed,
                 * * predictedToResolveToAmbientModuleMarker: known to be an ambient module.
                 * Needs to be reset to undefined before returning,
                 * * ResolvedModuleFull instance: can be reused.
                 */
                var result;
                var reusedNames;
                /** A transient placeholder used to mark predicted resolution in the result list. */
                var predictedToResolveToAmbientModuleMarker = {};
                for (var i = 0; i < moduleNames.length; i++) {
                    var moduleName = moduleNames[i];
                    // If the source file is unchanged and doesnt have invalidated resolution, reuse the module resolutions
                    if (file === oldSourceFile && !hasInvalidatedResolution(oldSourceFile.path)) {
                        var oldResolvedModule = ts.getResolvedModule(oldSourceFile, moduleName);
                        if (oldResolvedModule) {
                            if (ts.isTraceEnabled(options, host)) {
                                ts.trace(host, oldResolvedModule.packageId ?
                                    ts.Diagnostics.Reusing_resolution_of_module_0_from_1_of_old_program_it_was_successfully_resolved_to_2_with_Package_ID_3 :
                                    ts.Diagnostics.Reusing_resolution_of_module_0_from_1_of_old_program_it_was_successfully_resolved_to_2, moduleName, ts.getNormalizedAbsolutePath(file.originalFileName, currentDirectory), oldResolvedModule.resolvedFileName, oldResolvedModule.packageId && ts.packageIdToString(oldResolvedModule.packageId));
                            }
                            (result || (result = new Array(moduleNames.length)))[i] = oldResolvedModule;
                            (reusedNames || (reusedNames = [])).push(moduleName);
                            continue;
                        }
                    }
                    // We know moduleName resolves to an ambient module provided that moduleName:
                    // - is in the list of ambient modules locally declared in the current source file.
                    // - resolved to an ambient module in the old program whose declaration is in an unmodified file
                    //   (so the same module declaration will land in the new program)
                    var resolvesToAmbientModuleInNonModifiedFile = false;
                    if (ts.contains(file.ambientModuleNames, moduleName)) {
                        resolvesToAmbientModuleInNonModifiedFile = true;
                        if (ts.isTraceEnabled(options, host)) {
                            ts.trace(host, ts.Diagnostics.Module_0_was_resolved_as_locally_declared_ambient_module_in_file_1, moduleName, ts.getNormalizedAbsolutePath(file.originalFileName, currentDirectory));
                        }
                    }
                    else {
                        resolvesToAmbientModuleInNonModifiedFile = moduleNameResolvesToAmbientModuleInNonModifiedFile(moduleName);
                    }
                    if (resolvesToAmbientModuleInNonModifiedFile) {
                        (result || (result = new Array(moduleNames.length)))[i] = predictedToResolveToAmbientModuleMarker;
                    }
                    else {
                        // Resolution failed in the old program, or resolved to an ambient module for which we can't reuse the result.
                        (unknownModuleNames || (unknownModuleNames = [])).push(moduleName);
                    }
                }
                var resolutions = unknownModuleNames && unknownModuleNames.length
                    ? resolveModuleNamesWorker(unknownModuleNames, file, reusedNames)
                    : ts.emptyArray;
                // Combine results of resolutions and predicted results
                if (!result) {
                    // There were no unresolved/ambient resolutions.
                    ts.Debug.assert(resolutions.length === moduleNames.length);
                    return resolutions;
                }
                var j = 0;
                for (var i = 0; i < result.length; i++) {
                    if (result[i]) {
                        // `result[i]` is either a `ResolvedModuleFull` or a marker.
                        // If it is the former, we can leave it as is.
                        if (result[i] === predictedToResolveToAmbientModuleMarker) {
                            result[i] = undefined; // TODO: GH#18217
                        }
                    }
                    else {
                        result[i] = resolutions[j];
                        j++;
                    }
                }
                ts.Debug.assert(j === resolutions.length);
                return result;
                // If we change our policy of rechecking failed lookups on each program create,
                // we should adjust the value returned here.
                function moduleNameResolvesToAmbientModuleInNonModifiedFile(moduleName) {
                    var resolutionToFile = ts.getResolvedModule(oldSourceFile, moduleName);
                    var resolvedFile = resolutionToFile && oldProgram.getSourceFile(resolutionToFile.resolvedFileName);
                    if (resolutionToFile && resolvedFile) {
                        // In the old program, we resolved to an ambient module that was in the same
                        //   place as we expected to find an actual module file.
                        // We actually need to return 'false' here even though this seems like a 'true' case
                        //   because the normal module resolution algorithm will find this anyway.
                        return false;
                    }
                    // at least one of declarations should come from non-modified source file
                    var unmodifiedFile = ambientModuleNameToUnmodifiedFileName.get(moduleName);
                    if (!unmodifiedFile) {
                        return false;
                    }
                    if (ts.isTraceEnabled(options, host)) {
                        ts.trace(host, ts.Diagnostics.Module_0_was_resolved_as_ambient_module_declared_in_1_since_this_file_was_not_modified, moduleName, unmodifiedFile);
                    }
                    return true;
                }
            }
            function canReuseProjectReferences() {
                return !forEachProjectReference(oldProgram.getProjectReferences(), oldProgram.getResolvedProjectReferences(), function (oldResolvedRef, parent, index) {
                    var newRef = (parent ? parent.commandLine.projectReferences : projectReferences)[index];
                    var newResolvedRef = parseProjectReferenceConfigFile(newRef);
                    if (oldResolvedRef) {
                        // Resolved project reference has gone missing or changed
                        return !newResolvedRef ||
                            newResolvedRef.sourceFile !== oldResolvedRef.sourceFile ||
                            !ts.arrayIsEqualTo(oldResolvedRef.commandLine.fileNames, newResolvedRef.commandLine.fileNames);
                    }
                    else {
                        // A previously-unresolved reference may be resolved now
                        return newResolvedRef !== undefined;
                    }
                }, function (oldProjectReferences, parent) {
                    // If array of references is changed, we cant resue old program
                    var newReferences = parent ? getResolvedProjectReferenceByPath(parent.sourceFile.path).commandLine.projectReferences : projectReferences;
                    return !ts.arrayIsEqualTo(oldProjectReferences, newReferences, ts.projectReferenceIsEqualTo);
                });
            }
            function tryReuseStructureFromOldProgram() {
                var _a;
                if (!oldProgram) {
                    return 0 /* Not */;
                }
                // check properties that can affect structure of the program or module resolution strategy
                // if any of these properties has changed - structure cannot be reused
                var oldOptions = oldProgram.getCompilerOptions();
                if (ts.changesAffectModuleResolution(oldOptions, options)) {
                    return 0 /* Not */;
                }
                // there is an old program, check if we can reuse its structure
                var oldRootNames = oldProgram.getRootFileNames();
                if (!ts.arrayIsEqualTo(oldRootNames, rootNames)) {
                    return 0 /* Not */;
                }
                // Check if any referenced project tsconfig files are different
                if (!canReuseProjectReferences()) {
                    return 0 /* Not */;
                }
                if (projectReferences) {
                    resolvedProjectReferences = projectReferences.map(parseProjectReferenceConfigFile);
                }
                // check if program source files has changed in the way that can affect structure of the program
                var newSourceFiles = [];
                var modifiedSourceFiles = [];
                structureIsReused = 2 /* Completely */;
                // If the missing file paths are now present, it can change the progam structure,
                // and hence cant reuse the structure.
                // This is same as how we dont reuse the structure if one of the file from old program is now missing
                if (oldProgram.getMissingFilePaths().some(function (missingFilePath) { return host.fileExists(missingFilePath); })) {
                    return 0 /* Not */;
                }
                var oldSourceFiles = oldProgram.getSourceFiles();
                var SeenPackageName;
                (function (SeenPackageName) {
                    SeenPackageName[SeenPackageName["Exists"] = 0] = "Exists";
                    SeenPackageName[SeenPackageName["Modified"] = 1] = "Modified";
                })(SeenPackageName || (SeenPackageName = {}));
                var seenPackageNames = new ts.Map();
                for (var _i = 0, oldSourceFiles_2 = oldSourceFiles; _i < oldSourceFiles_2.length; _i++) {
                    var oldSourceFile = oldSourceFiles_2[_i];
                    var newSourceFile = host.getSourceFileByPath
                        ? host.getSourceFileByPath(oldSourceFile.fileName, oldSourceFile.resolvedPath, options.target, /*onError*/ undefined, shouldCreateNewSourceFile)
                        : host.getSourceFile(oldSourceFile.fileName, options.target, /*onError*/ undefined, shouldCreateNewSourceFile); // TODO: GH#18217
                    if (!newSourceFile) {
                        return 0 /* Not */;
                    }
                    ts.Debug.assert(!newSourceFile.redirectInfo, "Host should not return a redirect source file from `getSourceFile`");
                    var fileChanged = void 0;
                    if (oldSourceFile.redirectInfo) {
                        // We got `newSourceFile` by path, so it is actually for the unredirected file.
                        // This lets us know if the unredirected file has changed. If it has we should break the redirect.
                        if (newSourceFile !== oldSourceFile.redirectInfo.unredirected) {
                            // Underlying file has changed. Might not redirect anymore. Must rebuild program.
                            return 0 /* Not */;
                        }
                        fileChanged = false;
                        newSourceFile = oldSourceFile; // Use the redirect.
                    }
                    else if (oldProgram.redirectTargetsMap.has(oldSourceFile.path)) {
                        // If a redirected-to source file changes, the redirect may be broken.
                        if (newSourceFile !== oldSourceFile) {
                            return 0 /* Not */;
                        }
                        fileChanged = false;
                    }
                    else {
                        fileChanged = newSourceFile !== oldSourceFile;
                    }
                    // Since the project references havent changed, its right to set originalFileName and resolvedPath here
                    newSourceFile.path = oldSourceFile.path;
                    newSourceFile.originalFileName = oldSourceFile.originalFileName;
                    newSourceFile.resolvedPath = oldSourceFile.resolvedPath;
                    newSourceFile.fileName = oldSourceFile.fileName;
                    var packageName = oldProgram.sourceFileToPackageName.get(oldSourceFile.path);
                    if (packageName !== undefined) {
                        // If there are 2 different source files for the same package name and at least one of them changes,
                        // they might become redirects. So we must rebuild the program.
                        var prevKind = seenPackageNames.get(packageName);
                        var newKind = fileChanged ? 1 /* Modified */ : 0 /* Exists */;
                        if ((prevKind !== undefined && newKind === 1 /* Modified */) || prevKind === 1 /* Modified */) {
                            return 0 /* Not */;
                        }
                        seenPackageNames.set(packageName, newKind);
                    }
                    if (fileChanged) {
                        // The `newSourceFile` object was created for the new program.
                        if (!ts.arrayIsEqualTo(oldSourceFile.libReferenceDirectives, newSourceFile.libReferenceDirectives, fileReferenceIsEqualTo)) {
                            // 'lib' references has changed. Matches behavior in changesAffectModuleResolution
                            structureIsReused = 1 /* SafeModules */;
                        }
                        if (oldSourceFile.hasNoDefaultLib !== newSourceFile.hasNoDefaultLib) {
                            // value of no-default-lib has changed
                            // this will affect if default library is injected into the list of files
                            structureIsReused = 1 /* SafeModules */;
                        }
                        // check tripleslash references
                        if (!ts.arrayIsEqualTo(oldSourceFile.referencedFiles, newSourceFile.referencedFiles, fileReferenceIsEqualTo)) {
                            // tripleslash references has changed
                            structureIsReused = 1 /* SafeModules */;
                        }
                        // check imports and module augmentations
                        collectExternalModuleReferences(newSourceFile);
                        if (!ts.arrayIsEqualTo(oldSourceFile.imports, newSourceFile.imports, moduleNameIsEqualTo)) {
                            // imports has changed
                            structureIsReused = 1 /* SafeModules */;
                        }
                        if (!ts.arrayIsEqualTo(oldSourceFile.moduleAugmentations, newSourceFile.moduleAugmentations, moduleNameIsEqualTo)) {
                            // moduleAugmentations has changed
                            structureIsReused = 1 /* SafeModules */;
                        }
                        if ((oldSourceFile.flags & 3145728 /* PermanentlySetIncrementalFlags */) !== (newSourceFile.flags & 3145728 /* PermanentlySetIncrementalFlags */)) {
                            // dynamicImport has changed
                            structureIsReused = 1 /* SafeModules */;
                        }
                        if (!ts.arrayIsEqualTo(oldSourceFile.typeReferenceDirectives, newSourceFile.typeReferenceDirectives, fileReferenceIsEqualTo)) {
                            // 'types' references has changed
                            structureIsReused = 1 /* SafeModules */;
                        }
                        // tentatively approve the file
                        modifiedSourceFiles.push({ oldFile: oldSourceFile, newFile: newSourceFile });
                    }
                    else if (hasInvalidatedResolution(oldSourceFile.path)) {
                        // 'module/types' references could have changed
                        structureIsReused = 1 /* SafeModules */;
                        // add file to the modified list so that we will resolve it later
                        modifiedSourceFiles.push({ oldFile: oldSourceFile, newFile: newSourceFile });
                    }
                    // if file has passed all checks it should be safe to reuse it
                    newSourceFiles.push(newSourceFile);
                }
                if (structureIsReused !== 2 /* Completely */) {
                    return structureIsReused;
                }
                var modifiedFiles = modifiedSourceFiles.map(function (f) { return f.oldFile; });
                for (var _b = 0, oldSourceFiles_3 = oldSourceFiles; _b < oldSourceFiles_3.length; _b++) {
                    var oldFile = oldSourceFiles_3[_b];
                    if (!ts.contains(modifiedFiles, oldFile)) {
                        for (var _c = 0, _d = oldFile.ambientModuleNames; _c < _d.length; _c++) {
                            var moduleName = _d[_c];
                            ambientModuleNameToUnmodifiedFileName.set(moduleName, oldFile.fileName);
                        }
                    }
                }
                // try to verify results of module resolution
                for (var _e = 0, modifiedSourceFiles_1 = modifiedSourceFiles; _e < modifiedSourceFiles_1.length; _e++) {
                    var _f = modifiedSourceFiles_1[_e], oldSourceFile = _f.oldFile, newSourceFile = _f.newFile;
                    var moduleNames = getModuleNames(newSourceFile);
                    var resolutions = resolveModuleNamesReusingOldState(moduleNames, newSourceFile);
                    // ensure that module resolution results are still correct
                    var resolutionsChanged = ts.hasChangesInResolutions(moduleNames, resolutions, oldSourceFile.resolvedModules, ts.moduleResolutionIsEqualTo);
                    if (resolutionsChanged) {
                        structureIsReused = 1 /* SafeModules */;
                        newSourceFile.resolvedModules = ts.zipToMap(moduleNames, resolutions);
                    }
                    else {
                        newSourceFile.resolvedModules = oldSourceFile.resolvedModules;
                    }
                    // We lower-case all type references because npm automatically lowercases all packages. See GH#9824.
                    var typesReferenceDirectives = ts.map(newSourceFile.typeReferenceDirectives, function (ref) { return ts.toFileNameLowerCase(ref.fileName); });
                    var typeReferenceResolutions = resolveTypeReferenceDirectiveNamesWorker(typesReferenceDirectives, newSourceFile);
                    // ensure that types resolutions are still correct
                    var typeReferenceEesolutionsChanged = ts.hasChangesInResolutions(typesReferenceDirectives, typeReferenceResolutions, oldSourceFile.resolvedTypeReferenceDirectiveNames, ts.typeDirectiveIsEqualTo);
                    if (typeReferenceEesolutionsChanged) {
                        structureIsReused = 1 /* SafeModules */;
                        newSourceFile.resolvedTypeReferenceDirectiveNames = ts.zipToMap(typesReferenceDirectives, typeReferenceResolutions);
                    }
                    else {
                        newSourceFile.resolvedTypeReferenceDirectiveNames = oldSourceFile.resolvedTypeReferenceDirectiveNames;
                    }
                }
                if (structureIsReused !== 2 /* Completely */) {
                    return structureIsReused;
                }
                if (ts.changesAffectingProgramStructure(oldOptions, options) || ((_a = host.hasChangedAutomaticTypeDirectiveNames) === null || _a === void 0 ? void 0 : _a.call(host))) {
                    return 1 /* SafeModules */;
                }
                missingFilePaths = oldProgram.getMissingFilePaths();
                // update fileName -> file mapping
                ts.Debug.assert(newSourceFiles.length === oldProgram.getSourceFiles().length);
                for (var _g = 0, newSourceFiles_1 = newSourceFiles; _g < newSourceFiles_1.length; _g++) {
                    var newSourceFile = newSourceFiles_1[_g];
                    filesByName.set(newSourceFile.path, newSourceFile);
                }
                var oldFilesByNameMap = oldProgram.getFilesByNameMap();
                oldFilesByNameMap.forEach(function (oldFile, path) {
                    if (!oldFile) {
                        filesByName.set(path, oldFile);
                        return;
                    }
                    if (oldFile.path === path) {
                        // Set the file as found during node modules search if it was found that way in old progra,
                        if (oldProgram.isSourceFileFromExternalLibrary(oldFile)) {
                            sourceFilesFoundSearchingNodeModules.set(oldFile.path, true);
                        }
                        return;
                    }
                    filesByName.set(path, filesByName.get(oldFile.path));
                });
                files = newSourceFiles;
                fileReasons = oldProgram.getFileIncludeReasons();
                fileProcessingDiagnostics = oldProgram.getFileProcessingDiagnostics();
                resolvedTypeReferenceDirectives = oldProgram.getResolvedTypeReferenceDirectives();
                sourceFileToPackageName = oldProgram.sourceFileToPackageName;
                redirectTargetsMap = oldProgram.redirectTargetsMap;
                usesUriStyleNodeCoreModules = oldProgram.usesUriStyleNodeCoreModules;
                return 2 /* Completely */;
            }
            function getEmitHost(writeFileCallback) {
                return {
                    getPrependNodes: getPrependNodes,
                    getCanonicalFileName: getCanonicalFileName,
                    getCommonSourceDirectory: program.getCommonSourceDirectory,
                    getCompilerOptions: program.getCompilerOptions,
                    getCurrentDirectory: function () { return currentDirectory; },
                    getNewLine: function () { return host.getNewLine(); },
                    getSourceFile: program.getSourceFile,
                    getSourceFileByPath: program.getSourceFileByPath,
                    getSourceFiles: program.getSourceFiles,
                    getLibFileFromReference: program.getLibFileFromReference,
                    isSourceFileFromExternalLibrary: isSourceFileFromExternalLibrary,
                    getResolvedProjectReferenceToRedirect: getResolvedProjectReferenceToRedirect,
                    getProjectReferenceRedirect: getProjectReferenceRedirect,
                    isSourceOfProjectReferenceRedirect: isSourceOfProjectReferenceRedirect,
                    getSymlinkCache: getSymlinkCache,
                    writeFile: writeFileCallback || (function (fileName, data, writeByteOrderMark, onError, sourceFiles) { return host.writeFile(fileName, data, writeByteOrderMark, onError, sourceFiles); }),
                    isEmitBlocked: isEmitBlocked,
                    readFile: function (f) { return host.readFile(f); },
                    fileExists: function (f) {
                        // Use local caches
                        var path = toPath(f);
                        if (getSourceFileByPath(path))
                            return true;
                        if (ts.contains(missingFilePaths, path))
                            return false;
                        // Before falling back to the host
                        return host.fileExists(f);
                    },
                    useCaseSensitiveFileNames: function () { return host.useCaseSensitiveFileNames(); },
                    getProgramBuildInfo: function () { return program.getProgramBuildInfo && program.getProgramBuildInfo(); },
                    getSourceFileFromReference: function (file, ref) { return program.getSourceFileFromReference(file, ref); },
                    redirectTargetsMap: redirectTargetsMap,
                    getFileIncludeReasons: program.getFileIncludeReasons,
                };
            }
            function emitBuildInfo(writeFileCallback) {
                ts.Debug.assert(!ts.outFile(options));
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("emit" /* Emit */, "emitBuildInfo", {}, /*separateBeginAndEnd*/ true);
                ts.performance.mark("beforeEmit");
                var emitResult = ts.emitFiles(ts.notImplementedResolver, getEmitHost(writeFileCallback), 
                /*targetSourceFile*/ undefined, 
                /*transformers*/ ts.noTransformers, 
                /*emitOnlyDtsFiles*/ false, 
                /*onlyBuildInfo*/ true);
                ts.performance.mark("afterEmit");
                ts.performance.measure("Emit", "beforeEmit", "afterEmit");
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                return emitResult;
            }
            function getResolvedProjectReferences() {
                return resolvedProjectReferences;
            }
            function getProjectReferences() {
                return projectReferences;
            }
            function getPrependNodes() {
                return createPrependNodes(projectReferences, function (_ref, index) { var _a; return (_a = resolvedProjectReferences[index]) === null || _a === void 0 ? void 0 : _a.commandLine; }, function (fileName) {
                    var path = toPath(fileName);
                    var sourceFile = getSourceFileByPath(path);
                    return sourceFile ? sourceFile.text : filesByName.has(path) ? undefined : host.readFile(path);
                });
            }
            function isSourceFileFromExternalLibrary(file) {
                return !!sourceFilesFoundSearchingNodeModules.get(file.path);
            }
            function isSourceFileDefaultLibrary(file) {
                if (file.hasNoDefaultLib) {
                    return true;
                }
                if (!options.noLib) {
                    return false;
                }
                // If '--lib' is not specified, include default library file according to '--target'
                // otherwise, using options specified in '--lib' instead of '--target' default library file
                var equalityComparer = host.useCaseSensitiveFileNames() ? ts.equateStringsCaseSensitive : ts.equateStringsCaseInsensitive;
                if (!options.lib) {
                    return equalityComparer(file.fileName, getDefaultLibraryFileName());
                }
                else {
                    return ts.some(options.lib, function (libFileName) { return equalityComparer(file.fileName, ts.combinePaths(defaultLibraryPath, libFileName)); });
                }
            }
            function getDiagnosticsProducingTypeChecker() {
                return diagnosticsProducingTypeChecker || (diagnosticsProducingTypeChecker = ts.createTypeChecker(program, /*produceDiagnostics:*/ true));
            }
            function dropDiagnosticsProducingTypeChecker() {
                diagnosticsProducingTypeChecker = undefined;
            }
            function getTypeChecker() {
                return noDiagnosticsTypeChecker || (noDiagnosticsTypeChecker = ts.createTypeChecker(program, /*produceDiagnostics:*/ false));
            }
            function emit(sourceFile, writeFileCallback, cancellationToken, emitOnlyDtsFiles, transformers, forceDtsEmit) {
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("emit" /* Emit */, "emit", { path: sourceFile === null || sourceFile === void 0 ? void 0 : sourceFile.path }, /*separateBeginAndEnd*/ true);
                var result = runWithCancellationToken(function () { return emitWorker(program, sourceFile, writeFileCallback, cancellationToken, emitOnlyDtsFiles, transformers, forceDtsEmit); });
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                return result;
            }
            function isEmitBlocked(emitFileName) {
                return hasEmitBlockingDiagnostics.has(toPath(emitFileName));
            }
            function emitWorker(program, sourceFile, writeFileCallback, cancellationToken, emitOnlyDtsFiles, customTransformers, forceDtsEmit) {
                if (!forceDtsEmit) {
                    var result = handleNoEmitOptions(program, sourceFile, writeFileCallback, cancellationToken);
                    if (result)
                        return result;
                }
                // Create the emit resolver outside of the "emitTime" tracking code below.  That way
                // any cost associated with it (like type checking) are appropriate associated with
                // the type-checking counter.
                //
                // If the -out option is specified, we should not pass the source file to getEmitResolver.
                // This is because in the -out scenario all files need to be emitted, and therefore all
                // files need to be type checked. And the way to specify that all files need to be type
                // checked is to not pass the file to getEmitResolver.
                var emitResolver = getDiagnosticsProducingTypeChecker().getEmitResolver(ts.outFile(options) ? undefined : sourceFile, cancellationToken);
                ts.performance.mark("beforeEmit");
                var emitResult = ts.emitFiles(emitResolver, getEmitHost(writeFileCallback), sourceFile, ts.getTransformers(options, customTransformers, emitOnlyDtsFiles), emitOnlyDtsFiles, 
                /*onlyBuildInfo*/ false, forceDtsEmit);
                ts.performance.mark("afterEmit");
                ts.performance.measure("Emit", "beforeEmit", "afterEmit");
                return emitResult;
            }
            function getSourceFile(fileName) {
                return getSourceFileByPath(toPath(fileName));
            }
            function getSourceFileByPath(path) {
                return filesByName.get(path) || undefined;
            }
            function getDiagnosticsHelper(sourceFile, getDiagnostics, cancellationToken) {
                if (sourceFile) {
                    return getDiagnostics(sourceFile, cancellationToken);
                }
                return ts.sortAndDeduplicateDiagnostics(ts.flatMap(program.getSourceFiles(), function (sourceFile) {
                    if (cancellationToken) {
                        cancellationToken.throwIfCancellationRequested();
                    }
                    return getDiagnostics(sourceFile, cancellationToken);
                }));
            }
            function getSyntacticDiagnostics(sourceFile, cancellationToken) {
                return getDiagnosticsHelper(sourceFile, getSyntacticDiagnosticsForFile, cancellationToken);
            }
            function getSemanticDiagnostics(sourceFile, cancellationToken) {
                return getDiagnosticsHelper(sourceFile, getSemanticDiagnosticsForFile, cancellationToken);
            }
            function getCachedSemanticDiagnostics(sourceFile) {
                var _a;
                return sourceFile
                    ? (_a = cachedBindAndCheckDiagnosticsForFile.perFile) === null || _a === void 0 ? void 0 : _a.get(sourceFile.path)
                    : cachedBindAndCheckDiagnosticsForFile.allDiagnostics;
            }
            function getBindAndCheckDiagnostics(sourceFile, cancellationToken) {
                return getBindAndCheckDiagnosticsForFile(sourceFile, cancellationToken);
            }
            function getProgramDiagnostics(sourceFile) {
                var _a;
                if (ts.skipTypeChecking(sourceFile, options, program)) {
                    return ts.emptyArray;
                }
                var programDiagnosticsInFile = programDiagnostics.getDiagnostics(sourceFile.fileName);
                if (!((_a = sourceFile.commentDirectives) === null || _a === void 0 ? void 0 : _a.length)) {
                    return programDiagnosticsInFile;
                }
                return getDiagnosticsWithPrecedingDirectives(sourceFile, sourceFile.commentDirectives, programDiagnosticsInFile).diagnostics;
            }
            function getDeclarationDiagnostics(sourceFile, cancellationToken) {
                var options = program.getCompilerOptions();
                // collect diagnostics from the program only once if either no source file was specified or out/outFile is set (bundled emit)
                if (!sourceFile || ts.outFile(options)) {
                    return getDeclarationDiagnosticsWorker(sourceFile, cancellationToken);
                }
                else {
                    return getDiagnosticsHelper(sourceFile, getDeclarationDiagnosticsForFile, cancellationToken);
                }
            }
            function getSyntacticDiagnosticsForFile(sourceFile) {
                // For JavaScript files, we report semantic errors for using TypeScript-only
                // constructs from within a JavaScript file as syntactic errors.
                if (ts.isSourceFileJS(sourceFile)) {
                    if (!sourceFile.additionalSyntacticDiagnostics) {
                        sourceFile.additionalSyntacticDiagnostics = getJSSyntacticDiagnosticsForFile(sourceFile);
                    }
                    return ts.concatenate(sourceFile.additionalSyntacticDiagnostics, sourceFile.parseDiagnostics);
                }
                return sourceFile.parseDiagnostics;
            }
            function runWithCancellationToken(func) {
                try {
                    return func();
                }
                catch (e) {
                    if (e instanceof ts.OperationCanceledException) {
                        // We were canceled while performing the operation.  Because our type checker
                        // might be a bad state, we need to throw it away.
                        //
                        // Note: we are overly aggressive here.  We do not actually *have* to throw away
                        // the "noDiagnosticsTypeChecker".  However, for simplicity, i'd like to keep
                        // the lifetimes of these two TypeCheckers the same.  Also, we generally only
                        // cancel when the user has made a change anyways.  And, in that case, we (the
                        // program instance) will get thrown away anyways.  So trying to keep one of
                        // these type checkers alive doesn't serve much purpose.
                        noDiagnosticsTypeChecker = undefined;
                        diagnosticsProducingTypeChecker = undefined;
                    }
                    throw e;
                }
            }
            function getSemanticDiagnosticsForFile(sourceFile, cancellationToken) {
                return ts.concatenate(filterSemanticDiagnostics(getBindAndCheckDiagnosticsForFile(sourceFile, cancellationToken), options), getProgramDiagnostics(sourceFile));
            }
            function getBindAndCheckDiagnosticsForFile(sourceFile, cancellationToken) {
                return getAndCacheDiagnostics(sourceFile, cancellationToken, cachedBindAndCheckDiagnosticsForFile, getBindAndCheckDiagnosticsForFileNoCache);
            }
            function getBindAndCheckDiagnosticsForFileNoCache(sourceFile, cancellationToken) {
                return runWithCancellationToken(function () {
                    if (ts.skipTypeChecking(sourceFile, options, program)) {
                        return ts.emptyArray;
                    }
                    var typeChecker = getDiagnosticsProducingTypeChecker();
                    ts.Debug.assert(!!sourceFile.bindDiagnostics);
                    var isCheckJs = ts.isCheckJsEnabledForFile(sourceFile, options);
                    var isTsNoCheck = !!sourceFile.checkJsDirective && sourceFile.checkJsDirective.enabled === false;
                    // By default, only type-check .ts, .tsx, 'Deferred' and 'External' files (external files are added by plugins)
                    var includeBindAndCheckDiagnostics = !isTsNoCheck && (sourceFile.scriptKind === 3 /* TS */ || sourceFile.scriptKind === 4 /* TSX */
                        || sourceFile.scriptKind === 5 /* External */ || isCheckJs || sourceFile.scriptKind === 7 /* Deferred */);
                    var bindDiagnostics = includeBindAndCheckDiagnostics ? sourceFile.bindDiagnostics : ts.emptyArray;
                    var checkDiagnostics = includeBindAndCheckDiagnostics ? typeChecker.getDiagnostics(sourceFile, cancellationToken) : ts.emptyArray;
                    return getMergedBindAndCheckDiagnostics(sourceFile, includeBindAndCheckDiagnostics, bindDiagnostics, checkDiagnostics, isCheckJs ? sourceFile.jsDocDiagnostics : undefined);
                });
            }
            function getMergedBindAndCheckDiagnostics(sourceFile, includeBindAndCheckDiagnostics) {
                var _a;
                var allDiagnostics = [];
                for (var _i = 2; _i < arguments.length; _i++) {
                    allDiagnostics[_i - 2] = arguments[_i];
                }
                var flatDiagnostics = ts.flatten(allDiagnostics);
                if (!includeBindAndCheckDiagnostics || !((_a = sourceFile.commentDirectives) === null || _a === void 0 ? void 0 : _a.length)) {
                    return flatDiagnostics;
                }
                var _b = getDiagnosticsWithPrecedingDirectives(sourceFile, sourceFile.commentDirectives, flatDiagnostics), diagnostics = _b.diagnostics, directives = _b.directives;
                for (var _c = 0, _d = directives.getUnusedExpectations(); _c < _d.length; _c++) {
                    var errorExpectation = _d[_c];
                    diagnostics.push(ts.createDiagnosticForRange(sourceFile, errorExpectation.range, ts.Diagnostics.Unused_ts_expect_error_directive));
                }
                return diagnostics;
            }
            /**
             * Creates a map of comment directives along with the diagnostics immediately preceded by one of them.
             * Comments that match to any of those diagnostics are marked as used.
             */
            function getDiagnosticsWithPrecedingDirectives(sourceFile, commentDirectives, flatDiagnostics) {
                // Diagnostics are only reported if there is no comment directive preceding them
                // This will modify the directives map by marking "used" ones with a corresponding diagnostic
                var directives = ts.createCommentDirectivesMap(sourceFile, commentDirectives);
                var diagnostics = flatDiagnostics.filter(function (diagnostic) { return markPrecedingCommentDirectiveLine(diagnostic, directives) === -1; });
                return { diagnostics: diagnostics, directives: directives };
            }
            function getSuggestionDiagnostics(sourceFile, cancellationToken) {
                return runWithCancellationToken(function () {
                    return getDiagnosticsProducingTypeChecker().getSuggestionDiagnostics(sourceFile, cancellationToken);
                });
            }
            /**
             * @returns The line index marked as preceding the diagnostic, or -1 if none was.
             */
            function markPrecedingCommentDirectiveLine(diagnostic, directives) {
                var file = diagnostic.file, start = diagnostic.start;
                if (!file) {
                    return -1;
                }
                // Start out with the line just before the text
                var lineStarts = ts.getLineStarts(file);
                var line = ts.computeLineAndCharacterOfPosition(lineStarts, start).line - 1; // TODO: GH#18217
                while (line >= 0) {
                    // As soon as that line is known to have a comment directive, use that
                    if (directives.markUsed(line)) {
                        return line;
                    }
                    // Stop searching if the line is not empty and not a comment
                    var lineText = file.text.slice(lineStarts[line], lineStarts[line + 1]).trim();
                    if (lineText !== "" && !/^(\s*)\/\/(.*)$/.test(lineText)) {
                        return -1;
                    }
                    line--;
                }
                return -1;
            }
            function getJSSyntacticDiagnosticsForFile(sourceFile) {
                return runWithCancellationToken(function () {
                    var diagnostics = [];
                    walk(sourceFile, sourceFile);
                    ts.forEachChildRecursively(sourceFile, walk, walkArray);
                    return diagnostics;
                    function walk(node, parent) {
                        // Return directly from the case if the given node doesnt want to visit each child
                        // Otherwise break to visit each child
                        switch (parent.kind) {
                            case 162 /* Parameter */:
                            case 165 /* PropertyDeclaration */:
                            case 167 /* MethodDeclaration */:
                                if (parent.questionToken === node) {
                                    diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.The_0_modifier_can_only_be_used_in_TypeScript_files, "?"));
                                    return "skip";
                                }
                            // falls through
                            case 166 /* MethodSignature */:
                            case 169 /* Constructor */:
                            case 170 /* GetAccessor */:
                            case 171 /* SetAccessor */:
                            case 211 /* FunctionExpression */:
                            case 254 /* FunctionDeclaration */:
                            case 212 /* ArrowFunction */:
                            case 252 /* VariableDeclaration */:
                                // type annotation
                                if (parent.type === node) {
                                    diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.Type_annotations_can_only_be_used_in_TypeScript_files));
                                    return "skip";
                                }
                        }
                        switch (node.kind) {
                            case 265 /* ImportClause */:
                                if (node.isTypeOnly) {
                                    diagnostics.push(createDiagnosticForNode(parent, ts.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files, "import type"));
                                    return "skip";
                                }
                                break;
                            case 270 /* ExportDeclaration */:
                                if (node.isTypeOnly) {
                                    diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files, "export type"));
                                    return "skip";
                                }
                                break;
                            case 263 /* ImportEqualsDeclaration */:
                                diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.import_can_only_be_used_in_TypeScript_files));
                                return "skip";
                            case 269 /* ExportAssignment */:
                                if (node.isExportEquals) {
                                    diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.export_can_only_be_used_in_TypeScript_files));
                                    return "skip";
                                }
                                break;
                            case 289 /* HeritageClause */:
                                var heritageClause = node;
                                if (heritageClause.token === 117 /* ImplementsKeyword */) {
                                    diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.implements_clauses_can_only_be_used_in_TypeScript_files));
                                    return "skip";
                                }
                                break;
                            case 256 /* InterfaceDeclaration */:
                                var interfaceKeyword = ts.tokenToString(118 /* InterfaceKeyword */);
                                ts.Debug.assertIsDefined(interfaceKeyword);
                                diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files, interfaceKeyword));
                                return "skip";
                            case 259 /* ModuleDeclaration */:
                                var moduleKeyword = node.flags & 16 /* Namespace */ ? ts.tokenToString(141 /* NamespaceKeyword */) : ts.tokenToString(140 /* ModuleKeyword */);
                                ts.Debug.assertIsDefined(moduleKeyword);
                                diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files, moduleKeyword));
                                return "skip";
                            case 257 /* TypeAliasDeclaration */:
                                diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.Type_aliases_can_only_be_used_in_TypeScript_files));
                                return "skip";
                            case 258 /* EnumDeclaration */:
                                var enumKeyword = ts.Debug.checkDefined(ts.tokenToString(92 /* EnumKeyword */));
                                diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics._0_declarations_can_only_be_used_in_TypeScript_files, enumKeyword));
                                return "skip";
                            case 228 /* NonNullExpression */:
                                diagnostics.push(createDiagnosticForNode(node, ts.Diagnostics.Non_null_assertions_can_only_be_used_in_TypeScript_files));
                                return "skip";
                            case 227 /* AsExpression */:
                                diagnostics.push(createDiagnosticForNode(node.type, ts.Diagnostics.Type_assertion_expressions_can_only_be_used_in_TypeScript_files));
                                return "skip";
                            case 209 /* TypeAssertionExpression */:
                                ts.Debug.fail(); // Won't parse these in a JS file anyway, as they are interpreted as JSX.
                        }
                    }
                    function walkArray(nodes, parent) {
                        if (parent.decorators === nodes && !options.experimentalDecorators) {
                            diagnostics.push(createDiagnosticForNode(parent, ts.Diagnostics.Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_in_your_tsconfig_or_jsconfig_to_remove_this_warning));
                        }
                        switch (parent.kind) {
                            case 255 /* ClassDeclaration */:
                            case 224 /* ClassExpression */:
                            case 167 /* MethodDeclaration */:
                            case 169 /* Constructor */:
                            case 170 /* GetAccessor */:
                            case 171 /* SetAccessor */:
                            case 211 /* FunctionExpression */:
                            case 254 /* FunctionDeclaration */:
                            case 212 /* ArrowFunction */:
                                // Check type parameters
                                if (nodes === parent.typeParameters) {
                                    diagnostics.push(createDiagnosticForNodeArray(nodes, ts.Diagnostics.Type_parameter_declarations_can_only_be_used_in_TypeScript_files));
                                    return "skip";
                                }
                            // falls through
                            case 235 /* VariableStatement */:
                                // Check modifiers
                                if (nodes === parent.modifiers) {
                                    checkModifiers(parent.modifiers, parent.kind === 235 /* VariableStatement */);
                                    return "skip";
                                }
                                break;
                            case 165 /* PropertyDeclaration */:
                                // Check modifiers of property declaration
                                if (nodes === parent.modifiers) {
                                    for (var _i = 0, _a = nodes; _i < _a.length; _i++) {
                                        var modifier = _a[_i];
                                        if (modifier.kind !== 124 /* StaticKeyword */) {
                                            diagnostics.push(createDiagnosticForNode(modifier, ts.Diagnostics.The_0_modifier_can_only_be_used_in_TypeScript_files, ts.tokenToString(modifier.kind)));
                                        }
                                    }
                                    return "skip";
                                }
                                break;
                            case 162 /* Parameter */:
                                // Check modifiers of parameter declaration
                                if (nodes === parent.modifiers) {
                                    diagnostics.push(createDiagnosticForNodeArray(nodes, ts.Diagnostics.Parameter_modifiers_can_only_be_used_in_TypeScript_files));
                                    return "skip";
                                }
                                break;
                            case 206 /* CallExpression */:
                            case 207 /* NewExpression */:
                            case 226 /* ExpressionWithTypeArguments */:
                            case 277 /* JsxSelfClosingElement */:
                            case 278 /* JsxOpeningElement */:
                            case 208 /* TaggedTemplateExpression */:
                                // Check type arguments
                                if (nodes === parent.typeArguments) {
                                    diagnostics.push(createDiagnosticForNodeArray(nodes, ts.Diagnostics.Type_arguments_can_only_be_used_in_TypeScript_files));
                                    return "skip";
                                }
                                break;
                        }
                    }
                    function checkModifiers(modifiers, isConstValid) {
                        for (var _i = 0, modifiers_2 = modifiers; _i < modifiers_2.length; _i++) {
                            var modifier = modifiers_2[_i];
                            switch (modifier.kind) {
                                case 85 /* ConstKeyword */:
                                    if (isConstValid) {
                                        continue;
                                    }
                                // to report error,
                                // falls through
                                case 123 /* PublicKeyword */:
                                case 121 /* PrivateKeyword */:
                                case 122 /* ProtectedKeyword */:
                                case 143 /* ReadonlyKeyword */:
                                case 134 /* DeclareKeyword */:
                                case 126 /* AbstractKeyword */:
                                case 157 /* OverrideKeyword */:
                                    diagnostics.push(createDiagnosticForNode(modifier, ts.Diagnostics.The_0_modifier_can_only_be_used_in_TypeScript_files, ts.tokenToString(modifier.kind)));
                                    break;
                                // These are all legal modifiers.
                                case 124 /* StaticKeyword */:
                                case 93 /* ExportKeyword */:
                                case 88 /* DefaultKeyword */:
                            }
                        }
                    }
                    function createDiagnosticForNodeArray(nodes, message, arg0, arg1, arg2) {
                        var start = nodes.pos;
                        return ts.createFileDiagnostic(sourceFile, start, nodes.end - start, message, arg0, arg1, arg2);
                    }
                    // Since these are syntactic diagnostics, parent might not have been set
                    // this means the sourceFile cannot be infered from the node
                    function createDiagnosticForNode(node, message, arg0, arg1, arg2) {
                        return ts.createDiagnosticForNodeInSourceFile(sourceFile, node, message, arg0, arg1, arg2);
                    }
                });
            }
            function getDeclarationDiagnosticsWorker(sourceFile, cancellationToken) {
                return getAndCacheDiagnostics(sourceFile, cancellationToken, cachedDeclarationDiagnosticsForFile, getDeclarationDiagnosticsForFileNoCache);
            }
            function getDeclarationDiagnosticsForFileNoCache(sourceFile, cancellationToken) {
                return runWithCancellationToken(function () {
                    var resolver = getDiagnosticsProducingTypeChecker().getEmitResolver(sourceFile, cancellationToken);
                    // Don't actually write any files since we're just getting diagnostics.
                    return ts.getDeclarationDiagnostics(getEmitHost(ts.noop), resolver, sourceFile) || ts.emptyArray;
                });
            }
            function getAndCacheDiagnostics(sourceFile, cancellationToken, cache, getDiagnostics) {
                var _a;
                var cachedResult = sourceFile
                    ? (_a = cache.perFile) === null || _a === void 0 ? void 0 : _a.get(sourceFile.path)
                    : cache.allDiagnostics;
                if (cachedResult) {
                    return cachedResult;
                }
                var result = getDiagnostics(sourceFile, cancellationToken);
                if (sourceFile) {
                    (cache.perFile || (cache.perFile = new ts.Map())).set(sourceFile.path, result);
                }
                else {
                    cache.allDiagnostics = result;
                }
                return result;
            }
            function getDeclarationDiagnosticsForFile(sourceFile, cancellationToken) {
                return sourceFile.isDeclarationFile ? [] : getDeclarationDiagnosticsWorker(sourceFile, cancellationToken);
            }
            function getOptionsDiagnostics() {
                return ts.sortAndDeduplicateDiagnostics(ts.concatenate(programDiagnostics.getGlobalDiagnostics(), getOptionsDiagnosticsOfConfigFile()));
            }
            function getOptionsDiagnosticsOfConfigFile() {
                if (!options.configFile) {
                    return ts.emptyArray;
                }
                var diagnostics = programDiagnostics.getDiagnostics(options.configFile.fileName);
                forEachResolvedProjectReference(function (resolvedRef) {
                    diagnostics = ts.concatenate(diagnostics, programDiagnostics.getDiagnostics(resolvedRef.sourceFile.fileName));
                });
                return diagnostics;
            }
            function getGlobalDiagnostics() {
                return rootNames.length ? ts.sortAndDeduplicateDiagnostics(getDiagnosticsProducingTypeChecker().getGlobalDiagnostics().slice()) : ts.emptyArray;
            }
            function getConfigFileParsingDiagnostics() {
                return configFileParsingDiagnostics || ts.emptyArray;
            }
            function processRootFile(fileName, isDefaultLib, ignoreNoDefaultLib, reason) {
                processSourceFile(ts.normalizePath(fileName), isDefaultLib, ignoreNoDefaultLib, /*packageId*/ undefined, reason);
            }
            function fileReferenceIsEqualTo(a, b) {
                return a.fileName === b.fileName;
            }
            function moduleNameIsEqualTo(a, b) {
                return a.kind === 79 /* Identifier */
                    ? b.kind === 79 /* Identifier */ && a.escapedText === b.escapedText
                    : b.kind === 10 /* StringLiteral */ && a.text === b.text;
            }
            function createSyntheticImport(text, file) {
                var externalHelpersModuleReference = ts.factory.createStringLiteral(text);
                var importDecl = ts.factory.createImportDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*importClause*/ undefined, externalHelpersModuleReference);
                ts.addEmitFlags(importDecl, 67108864 /* NeverApplyImportHelper */);
                ts.setParent(externalHelpersModuleReference, importDecl);
                ts.setParent(importDecl, file);
                // explicitly unset the synthesized flag on these declarations so the checker API will answer questions about them
                // (which is required to build the dependency graph for incremental emit)
                externalHelpersModuleReference.flags &= ~8 /* Synthesized */;
                importDecl.flags &= ~8 /* Synthesized */;
                return externalHelpersModuleReference;
            }
            function collectExternalModuleReferences(file) {
                if (file.imports) {
                    return;
                }
                var isJavaScriptFile = ts.isSourceFileJS(file);
                var isExternalModuleFile = ts.isExternalModule(file);
                // file.imports may not be undefined if there exists dynamic import
                var imports;
                var moduleAugmentations;
                var ambientModules;
                // If we are importing helpers, we need to add a synthetic reference to resolve the
                // helpers library.
                if ((options.isolatedModules || isExternalModuleFile)
                    && !file.isDeclarationFile) {
                    if (options.importHelpers) {
                        // synthesize 'import "tslib"' declaration
                        imports = [createSyntheticImport(ts.externalHelpersModuleNameText, file)];
                    }
                    var jsxImport = ts.getJSXRuntimeImport(ts.getJSXImplicitImportBase(options, file), options);
                    if (jsxImport) {
                        // synthesize `import "base/jsx-runtime"` declaration
                        (imports || (imports = [])).push(createSyntheticImport(jsxImport, file));
                    }
                }
                for (var _i = 0, _a = file.statements; _i < _a.length; _i++) {
                    var node = _a[_i];
                    collectModuleReferences(node, /*inAmbientModule*/ false);
                }
                if ((file.flags & 1048576 /* PossiblyContainsDynamicImport */) || isJavaScriptFile) {
                    collectDynamicImportOrRequireCalls(file);
                }
                file.imports = imports || ts.emptyArray;
                file.moduleAugmentations = moduleAugmentations || ts.emptyArray;
                file.ambientModuleNames = ambientModules || ts.emptyArray;
                return;
                function collectModuleReferences(node, inAmbientModule) {
                    if (ts.isAnyImportOrReExport(node)) {
                        var moduleNameExpr = ts.getExternalModuleName(node);
                        // TypeScript 1.0 spec (April 2014): 12.1.6
                        // An ExternalImportDeclaration in an AmbientExternalModuleDeclaration may reference other external modules
                        // only through top - level external module names. Relative external module names are not permitted.
                        if (moduleNameExpr && ts.isStringLiteral(moduleNameExpr) && moduleNameExpr.text && (!inAmbientModule || !ts.isExternalModuleNameRelative(moduleNameExpr.text))) {
                            imports = ts.append(imports, moduleNameExpr);
                            if (!usesUriStyleNodeCoreModules && currentNodeModulesDepth === 0 && !file.isDeclarationFile) {
                                usesUriStyleNodeCoreModules = ts.startsWith(moduleNameExpr.text, "node:");
                            }
                        }
                    }
                    else if (ts.isModuleDeclaration(node)) {
                        if (ts.isAmbientModule(node) && (inAmbientModule || ts.hasSyntacticModifier(node, 2 /* Ambient */) || file.isDeclarationFile)) {
                            var nameText = ts.getTextOfIdentifierOrLiteral(node.name);
                            // Ambient module declarations can be interpreted as augmentations for some existing external modules.
                            // This will happen in two cases:
                            // - if current file is external module then module augmentation is a ambient module declaration defined in the top level scope
                            // - if current file is not external module then module augmentation is an ambient module declaration with non-relative module name
                            //   immediately nested in top level ambient module declaration .
                            if (isExternalModuleFile || (inAmbientModule && !ts.isExternalModuleNameRelative(nameText))) {
                                (moduleAugmentations || (moduleAugmentations = [])).push(node.name);
                            }
                            else if (!inAmbientModule) {
                                if (file.isDeclarationFile) {
                                    // for global .d.ts files record name of ambient module
                                    (ambientModules || (ambientModules = [])).push(nameText);
                                }
                                // An AmbientExternalModuleDeclaration declares an external module.
                                // This type of declaration is permitted only in the global module.
                                // The StringLiteral must specify a top - level external module name.
                                // Relative external module names are not permitted
                                // NOTE: body of ambient module is always a module block, if it exists
                                var body = node.body;
                                if (body) {
                                    for (var _i = 0, _a = body.statements; _i < _a.length; _i++) {
                                        var statement = _a[_i];
                                        collectModuleReferences(statement, /*inAmbientModule*/ true);
                                    }
                                }
                            }
                        }
                    }
                }
                function collectDynamicImportOrRequireCalls(file) {
                    var r = /import|require/g;
                    while (r.exec(file.text) !== null) { // eslint-disable-line no-null/no-null
                        var node = getNodeAtPosition(file, r.lastIndex);
                        if (isJavaScriptFile && ts.isRequireCall(node, /*checkArgumentIsStringLiteralLike*/ true)) {
                            imports = ts.append(imports, node.arguments[0]);
                        }
                        // we have to check the argument list has length of 1. We will still have to process these even though we have parsing error.
                        else if (ts.isImportCall(node) && node.arguments.length === 1 && ts.isStringLiteralLike(node.arguments[0])) {
                            imports = ts.append(imports, node.arguments[0]);
                        }
                        else if (ts.isLiteralImportTypeNode(node)) {
                            imports = ts.append(imports, node.argument.literal);
                        }
                    }
                }
                /** Returns a token if position is in [start-of-leading-trivia, end), includes JSDoc only in JS files */
                function getNodeAtPosition(sourceFile, position) {
                    var current = sourceFile;
                    var getContainingChild = function (child) {
                        if (child.pos <= position && (position < child.end || (position === child.end && (child.kind === 1 /* EndOfFileToken */)))) {
                            return child;
                        }
                    };
                    while (true) {
                        var child = isJavaScriptFile && ts.hasJSDocNodes(current) && ts.forEach(current.jsDoc, getContainingChild) || ts.forEachChild(current, getContainingChild);
                        if (!child) {
                            return current;
                        }
                        current = child;
                    }
                }
            }
            function getLibFileFromReference(ref) {
                var libName = ts.toFileNameLowerCase(ref.fileName);
                var libFileName = ts.libMap.get(libName);
                if (libFileName) {
                    return getSourceFile(ts.combinePaths(defaultLibraryPath, libFileName));
                }
            }
            /** This should have similar behavior to 'processSourceFile' without diagnostics or mutation. */
            function getSourceFileFromReference(referencingFile, ref) {
                return getSourceFileFromReferenceWorker(resolveTripleslashReference(ref.fileName, referencingFile.fileName), getSourceFile);
            }
            function getSourceFileFromReferenceWorker(fileName, getSourceFile, fail, reason) {
                if (ts.hasExtension(fileName)) {
                    var canonicalFileName_1 = host.getCanonicalFileName(fileName);
                    if (!options.allowNonTsExtensions && !ts.forEach(supportedExtensionsWithJsonIfResolveJsonModule, function (extension) { return ts.fileExtensionIs(canonicalFileName_1, extension); })) {
                        if (fail) {
                            if (ts.hasJSFileExtension(canonicalFileName_1)) {
                                fail(ts.Diagnostics.File_0_is_a_JavaScript_file_Did_you_mean_to_enable_the_allowJs_option, fileName);
                            }
                            else {
                                fail(ts.Diagnostics.File_0_has_an_unsupported_extension_The_only_supported_extensions_are_1, fileName, "'" + supportedExtensions.join("', '") + "'");
                            }
                        }
                        return undefined;
                    }
                    var sourceFile = getSourceFile(fileName);
                    if (fail) {
                        if (!sourceFile) {
                            var redirect = getProjectReferenceRedirect(fileName);
                            if (redirect) {
                                fail(ts.Diagnostics.Output_file_0_has_not_been_built_from_source_file_1, redirect, fileName);
                            }
                            else {
                                fail(ts.Diagnostics.File_0_not_found, fileName);
                            }
                        }
                        else if (isReferencedFile(reason) && canonicalFileName_1 === host.getCanonicalFileName(getSourceFileByPath(reason.file).fileName)) {
                            fail(ts.Diagnostics.A_file_cannot_have_a_reference_to_itself);
                        }
                    }
                    return sourceFile;
                }
                else {
                    var sourceFileNoExtension = options.allowNonTsExtensions && getSourceFile(fileName);
                    if (sourceFileNoExtension)
                        return sourceFileNoExtension;
                    if (fail && options.allowNonTsExtensions) {
                        fail(ts.Diagnostics.File_0_not_found, fileName);
                        return undefined;
                    }
                    var sourceFileWithAddedExtension = ts.forEach(supportedExtensions, function (extension) { return getSourceFile(fileName + extension); });
                    if (fail && !sourceFileWithAddedExtension)
                        fail(ts.Diagnostics.Could_not_resolve_the_path_0_with_the_extensions_Colon_1, fileName, "'" + supportedExtensions.join("', '") + "'");
                    return sourceFileWithAddedExtension;
                }
            }
            /** This has side effects through `findSourceFile`. */
            function processSourceFile(fileName, isDefaultLib, ignoreNoDefaultLib, packageId, reason) {
                getSourceFileFromReferenceWorker(fileName, function (fileName) { return findSourceFile(fileName, toPath(fileName), isDefaultLib, ignoreNoDefaultLib, reason, packageId); }, // TODO: GH#18217
                function (diagnostic) {
                    var args = [];
                    for (var _i = 1; _i < arguments.length; _i++) {
                        args[_i - 1] = arguments[_i];
                    }
                    return addFilePreprocessingFileExplainingDiagnostic(/*file*/ undefined, reason, diagnostic, args);
                }, reason);
            }
            function processProjectReferenceFile(fileName, reason) {
                return processSourceFile(fileName, /*isDefaultLib*/ false, /*ignoreNoDefaultLib*/ false, /*packageId*/ undefined, reason);
            }
            function reportFileNamesDifferOnlyInCasingError(fileName, existingFile, reason) {
                var hasExistingReasonToReportErrorOn = !isReferencedFile(reason) && ts.some(fileReasons.get(existingFile.path), isReferencedFile);
                if (hasExistingReasonToReportErrorOn) {
                    addFilePreprocessingFileExplainingDiagnostic(existingFile, reason, ts.Diagnostics.Already_included_file_name_0_differs_from_file_name_1_only_in_casing, [existingFile.fileName, fileName]);
                }
                else {
                    addFilePreprocessingFileExplainingDiagnostic(existingFile, reason, ts.Diagnostics.File_name_0_differs_from_already_included_file_name_1_only_in_casing, [fileName, existingFile.fileName]);
                }
            }
            function createRedirectSourceFile(redirectTarget, unredirected, fileName, path, resolvedPath, originalFileName) {
                var redirect = Object.create(redirectTarget);
                redirect.fileName = fileName;
                redirect.path = path;
                redirect.resolvedPath = resolvedPath;
                redirect.originalFileName = originalFileName;
                redirect.redirectInfo = { redirectTarget: redirectTarget, unredirected: unredirected };
                sourceFilesFoundSearchingNodeModules.set(path, currentNodeModulesDepth > 0);
                Object.defineProperties(redirect, {
                    id: {
                        get: function () { return this.redirectInfo.redirectTarget.id; },
                        set: function (value) { this.redirectInfo.redirectTarget.id = value; },
                    },
                    symbol: {
                        get: function () { return this.redirectInfo.redirectTarget.symbol; },
                        set: function (value) { this.redirectInfo.redirectTarget.symbol = value; },
                    },
                });
                return redirect;
            }
            // Get source file from normalized fileName
            function findSourceFile(fileName, path, isDefaultLib, ignoreNoDefaultLib, reason, packageId) {
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("program" /* Program */, "findSourceFile", {
                    fileName: fileName,
                    isDefaultLib: isDefaultLib || undefined,
                    fileIncludeKind: ts.FileIncludeKind[reason.kind],
                });
                var result = findSourceFileWorker(fileName, path, isDefaultLib, ignoreNoDefaultLib, reason, packageId);
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
                return result;
            }
            function findSourceFileWorker(fileName, path, isDefaultLib, ignoreNoDefaultLib, reason, packageId) {
                if (useSourceOfProjectReferenceRedirect) {
                    var source = getSourceOfProjectReferenceRedirect(fileName);
                    // If preserveSymlinks is true, module resolution wont jump the symlink
                    // but the resolved real path may be the .d.ts from project reference
                    // Note:: Currently we try the real path only if the
                    // file is from node_modules to avoid having to run real path on all file paths
                    if (!source &&
                        host.realpath &&
                        options.preserveSymlinks &&
                        ts.isDeclarationFileName(fileName) &&
                        ts.stringContains(fileName, ts.nodeModulesPathPart)) {
                        var realPath = host.realpath(fileName);
                        if (realPath !== fileName)
                            source = getSourceOfProjectReferenceRedirect(realPath);
                    }
                    if (source) {
                        var file_1 = ts.isString(source) ?
                            findSourceFile(source, toPath(source), isDefaultLib, ignoreNoDefaultLib, reason, packageId) :
                            undefined;
                        if (file_1)
                            addFileToFilesByName(file_1, path, /*redirectedPath*/ undefined);
                        return file_1;
                    }
                }
                var originalFileName = fileName;
                if (filesByName.has(path)) {
                    var file_2 = filesByName.get(path);
                    addFileIncludeReason(file_2 || undefined, reason);
                    // try to check if we've already seen this file but with a different casing in path
                    // NOTE: this only makes sense for case-insensitive file systems, and only on files which are not redirected
                    if (file_2 && options.forceConsistentCasingInFileNames) {
                        var checkedName = file_2.fileName;
                        var isRedirect = toPath(checkedName) !== toPath(fileName);
                        if (isRedirect) {
                            fileName = getProjectReferenceRedirect(fileName) || fileName;
                        }
                        // Check if it differs only in drive letters its ok to ignore that error:
                        var checkedAbsolutePath = ts.getNormalizedAbsolutePathWithoutRoot(checkedName, currentDirectory);
                        var inputAbsolutePath = ts.getNormalizedAbsolutePathWithoutRoot(fileName, currentDirectory);
                        if (checkedAbsolutePath !== inputAbsolutePath) {
                            reportFileNamesDifferOnlyInCasingError(fileName, file_2, reason);
                        }
                    }
                    // If the file was previously found via a node_modules search, but is now being processed as a root file,
                    // then everything it sucks in may also be marked incorrectly, and needs to be checked again.
                    if (file_2 && sourceFilesFoundSearchingNodeModules.get(file_2.path) && currentNodeModulesDepth === 0) {
                        sourceFilesFoundSearchingNodeModules.set(file_2.path, false);
                        if (!options.noResolve) {
                            processReferencedFiles(file_2, isDefaultLib);
                            processTypeReferenceDirectives(file_2);
                        }
                        if (!options.noLib) {
                            processLibReferenceDirectives(file_2);
                        }
                        modulesWithElidedImports.set(file_2.path, false);
                        processImportedModules(file_2);
                    }
                    // See if we need to reprocess the imports due to prior skipped imports
                    else if (file_2 && modulesWithElidedImports.get(file_2.path)) {
                        if (currentNodeModulesDepth < maxNodeModuleJsDepth) {
                            modulesWithElidedImports.set(file_2.path, false);
                            processImportedModules(file_2);
                        }
                    }
                    return file_2 || undefined;
                }
                var redirectedPath;
                if (isReferencedFile(reason) && !useSourceOfProjectReferenceRedirect) {
                    var redirectProject = getProjectReferenceRedirectProject(fileName);
                    if (redirectProject) {
                        if (ts.outFile(redirectProject.commandLine.options)) {
                            // Shouldnt create many to 1 mapping file in --out scenario
                            return undefined;
                        }
                        var redirect = getProjectReferenceOutputName(redirectProject, fileName);
                        fileName = redirect;
                        // Once we start redirecting to a file, we can potentially come back to it
                        // via a back-reference from another file in the .d.ts folder. If that happens we'll
                        // end up trying to add it to the program *again* because we were tracking it via its
                        // original (un-redirected) name. So we have to map both the original path and the redirected path
                        // to the source file we're about to find/create
                        redirectedPath = toPath(redirect);
                    }
                }
                // We haven't looked for this file, do so now and cache result
                var file = host.getSourceFile(fileName, options.target, function (hostErrorMessage) { return addFilePreprocessingFileExplainingDiagnostic(/*file*/ undefined, reason, ts.Diagnostics.Cannot_read_file_0_Colon_1, [fileName, hostErrorMessage]); }, shouldCreateNewSourceFile);
                if (packageId) {
                    var packageIdKey = ts.packageIdToString(packageId);
                    var fileFromPackageId = packageIdToSourceFile.get(packageIdKey);
                    if (fileFromPackageId) {
                        // Some other SourceFile already exists with this package name and version.
                        // Instead of creating a duplicate, just redirect to the existing one.
                        var dupFile = createRedirectSourceFile(fileFromPackageId, file, fileName, path, toPath(fileName), originalFileName); // TODO: GH#18217
                        redirectTargetsMap.add(fileFromPackageId.path, fileName);
                        addFileToFilesByName(dupFile, path, redirectedPath);
                        addFileIncludeReason(dupFile, reason);
                        sourceFileToPackageName.set(path, packageId.name);
                        processingOtherFiles.push(dupFile);
                        return dupFile;
                    }
                    else if (file) {
                        // This is the first source file to have this packageId.
                        packageIdToSourceFile.set(packageIdKey, file);
                        sourceFileToPackageName.set(path, packageId.name);
                    }
                }
                addFileToFilesByName(file, path, redirectedPath);
                if (file) {
                    sourceFilesFoundSearchingNodeModules.set(path, currentNodeModulesDepth > 0);
                    file.fileName = fileName; // Ensure that source file has same name as what we were looking for
                    file.path = path;
                    file.resolvedPath = toPath(fileName);
                    file.originalFileName = originalFileName;
                    addFileIncludeReason(file, reason);
                    if (host.useCaseSensitiveFileNames()) {
                        var pathLowerCase = ts.toFileNameLowerCase(path);
                        // for case-sensitive file systems check if we've already seen some file with similar filename ignoring case
                        var existingFile = filesByNameIgnoreCase.get(pathLowerCase);
                        if (existingFile) {
                            reportFileNamesDifferOnlyInCasingError(fileName, existingFile, reason);
                        }
                        else {
                            filesByNameIgnoreCase.set(pathLowerCase, file);
                        }
                    }
                    skipDefaultLib = skipDefaultLib || (file.hasNoDefaultLib && !ignoreNoDefaultLib);
                    if (!options.noResolve) {
                        processReferencedFiles(file, isDefaultLib);
                        processTypeReferenceDirectives(file);
                    }
                    if (!options.noLib) {
                        processLibReferenceDirectives(file);
                    }
                    // always process imported modules to record module name resolutions
                    processImportedModules(file);
                    if (isDefaultLib) {
                        processingDefaultLibFiles.push(file);
                    }
                    else {
                        processingOtherFiles.push(file);
                    }
                }
                return file;
            }
            function addFileIncludeReason(file, reason) {
                if (file)
                    fileReasons.add(file.path, reason);
            }
            function addFileToFilesByName(file, path, redirectedPath) {
                if (redirectedPath) {
                    filesByName.set(redirectedPath, file);
                    filesByName.set(path, file || false);
                }
                else {
                    filesByName.set(path, file);
                }
            }
            function getProjectReferenceRedirect(fileName) {
                var referencedProject = getProjectReferenceRedirectProject(fileName);
                return referencedProject && getProjectReferenceOutputName(referencedProject, fileName);
            }
            function getProjectReferenceRedirectProject(fileName) {
                // Ignore dts or any json files
                if (!resolvedProjectReferences || !resolvedProjectReferences.length || ts.fileExtensionIs(fileName, ".d.ts" /* Dts */) || ts.fileExtensionIs(fileName, ".json" /* Json */)) {
                    return undefined;
                }
                // If this file is produced by a referenced project, we need to rewrite it to
                // look in the output folder of the referenced project rather than the input
                return getResolvedProjectReferenceToRedirect(fileName);
            }
            function getProjectReferenceOutputName(referencedProject, fileName) {
                var out = ts.outFile(referencedProject.commandLine.options);
                return out ?
                    ts.changeExtension(out, ".d.ts" /* Dts */) :
                    ts.getOutputDeclarationFileName(fileName, referencedProject.commandLine, !host.useCaseSensitiveFileNames());
            }
            /**
             * Get the referenced project if the file is input file from that reference project
             */
            function getResolvedProjectReferenceToRedirect(fileName) {
                if (mapFromFileToProjectReferenceRedirects === undefined) {
                    mapFromFileToProjectReferenceRedirects = new ts.Map();
                    forEachResolvedProjectReference(function (referencedProject) {
                        // not input file from the referenced project, ignore
                        if (toPath(options.configFilePath) !== referencedProject.sourceFile.path) {
                            referencedProject.commandLine.fileNames.forEach(function (f) {
                                return mapFromFileToProjectReferenceRedirects.set(toPath(f), referencedProject.sourceFile.path);
                            });
                        }
                    });
                }
                var referencedProjectPath = mapFromFileToProjectReferenceRedirects.get(toPath(fileName));
                return referencedProjectPath && getResolvedProjectReferenceByPath(referencedProjectPath);
            }
            function forEachResolvedProjectReference(cb) {
                return ts.forEachResolvedProjectReference(resolvedProjectReferences, cb);
            }
            function getSourceOfProjectReferenceRedirect(file) {
                if (!ts.isDeclarationFileName(file))
                    return undefined;
                if (mapFromToProjectReferenceRedirectSource === undefined) {
                    mapFromToProjectReferenceRedirectSource = new ts.Map();
                    forEachResolvedProjectReference(function (resolvedRef) {
                        var out = ts.outFile(resolvedRef.commandLine.options);
                        if (out) {
                            // Dont know which source file it means so return true?
                            var outputDts = ts.changeExtension(out, ".d.ts" /* Dts */);
                            mapFromToProjectReferenceRedirectSource.set(toPath(outputDts), true);
                        }
                        else {
                            var getCommonSourceDirectory_3 = ts.memoize(function () { return ts.getCommonSourceDirectoryOfConfig(resolvedRef.commandLine, !host.useCaseSensitiveFileNames()); });
                            ts.forEach(resolvedRef.commandLine.fileNames, function (fileName) {
                                if (!ts.fileExtensionIs(fileName, ".d.ts" /* Dts */) && !ts.fileExtensionIs(fileName, ".json" /* Json */)) {
                                    var outputDts = ts.getOutputDeclarationFileName(fileName, resolvedRef.commandLine, !host.useCaseSensitiveFileNames(), getCommonSourceDirectory_3);
                                    mapFromToProjectReferenceRedirectSource.set(toPath(outputDts), fileName);
                                }
                            });
                        }
                    });
                }
                return mapFromToProjectReferenceRedirectSource.get(toPath(file));
            }
            function isSourceOfProjectReferenceRedirect(fileName) {
                return useSourceOfProjectReferenceRedirect && !!getResolvedProjectReferenceToRedirect(fileName);
            }
            function getResolvedProjectReferenceByPath(projectReferencePath) {
                if (!projectReferenceRedirects) {
                    return undefined;
                }
                return projectReferenceRedirects.get(projectReferencePath) || undefined;
            }
            function processReferencedFiles(file, isDefaultLib) {
                ts.forEach(file.referencedFiles, function (ref, index) {
                    processSourceFile(resolveTripleslashReference(ref.fileName, file.fileName), isDefaultLib, 
                    /*ignoreNoDefaultLib*/ false, 
                    /*packageId*/ undefined, { kind: ts.FileIncludeKind.ReferenceFile, file: file.path, index: index, });
                });
            }
            function processTypeReferenceDirectives(file) {
                // We lower-case all type references because npm automatically lowercases all packages. See GH#9824.
                var typeDirectives = ts.map(file.typeReferenceDirectives, function (ref) { return ts.toFileNameLowerCase(ref.fileName); });
                if (!typeDirectives) {
                    return;
                }
                var resolutions = resolveTypeReferenceDirectiveNamesWorker(typeDirectives, file);
                for (var index = 0; index < typeDirectives.length; index++) {
                    var ref = file.typeReferenceDirectives[index];
                    var resolvedTypeReferenceDirective = resolutions[index];
                    // store resolved type directive on the file
                    var fileName = ts.toFileNameLowerCase(ref.fileName);
                    ts.setResolvedTypeReferenceDirective(file, fileName, resolvedTypeReferenceDirective);
                    processTypeReferenceDirective(fileName, resolvedTypeReferenceDirective, { kind: ts.FileIncludeKind.TypeReferenceDirective, file: file.path, index: index, });
                }
            }
            function processTypeReferenceDirective(typeReferenceDirective, resolvedTypeReferenceDirective, reason) {
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.push("program" /* Program */, "processTypeReferenceDirective", { directive: typeReferenceDirective, hasResolved: !!resolveModuleNamesReusingOldState, refKind: reason.kind, refPath: isReferencedFile(reason) ? reason.file : undefined });
                processTypeReferenceDirectiveWorker(typeReferenceDirective, resolvedTypeReferenceDirective, reason);
                ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.pop();
            }
            function processTypeReferenceDirectiveWorker(typeReferenceDirective, resolvedTypeReferenceDirective, reason) {
                // If we already found this library as a primary reference - nothing to do
                var previousResolution = resolvedTypeReferenceDirectives.get(typeReferenceDirective);
                if (previousResolution && previousResolution.primary) {
                    return;
                }
                var saveResolution = true;
                if (resolvedTypeReferenceDirective) {
                    if (resolvedTypeReferenceDirective.isExternalLibraryImport)
                        currentNodeModulesDepth++;
                    if (resolvedTypeReferenceDirective.primary) {
                        // resolved from the primary path
                        processSourceFile(resolvedTypeReferenceDirective.resolvedFileName, /*isDefaultLib*/ false, /*ignoreNoDefaultLib*/ false, resolvedTypeReferenceDirective.packageId, reason); // TODO: GH#18217
                    }
                    else {
                        // If we already resolved to this file, it must have been a secondary reference. Check file contents
                        // for sameness and possibly issue an error
                        if (previousResolution) {
                            // Don't bother reading the file again if it's the same file.
                            if (resolvedTypeReferenceDirective.resolvedFileName !== previousResolution.resolvedFileName) {
                                var otherFileText = host.readFile(resolvedTypeReferenceDirective.resolvedFileName);
                                var existingFile = getSourceFile(previousResolution.resolvedFileName);
                                if (otherFileText !== existingFile.text) {
                                    addFilePreprocessingFileExplainingDiagnostic(existingFile, reason, ts.Diagnostics.Conflicting_definitions_for_0_found_at_1_and_2_Consider_installing_a_specific_version_of_this_library_to_resolve_the_conflict, [typeReferenceDirective, resolvedTypeReferenceDirective.resolvedFileName, previousResolution.resolvedFileName]);
                                }
                            }
                            // don't overwrite previous resolution result
                            saveResolution = false;
                        }
                        else {
                            // First resolution of this library
                            processSourceFile(resolvedTypeReferenceDirective.resolvedFileName, /*isDefaultLib*/ false, /*ignoreNoDefaultLib*/ false, resolvedTypeReferenceDirective.packageId, reason);
                        }
                    }
                    if (resolvedTypeReferenceDirective.isExternalLibraryImport)
                        currentNodeModulesDepth--;
                }
                else {
                    addFilePreprocessingFileExplainingDiagnostic(/*file*/ undefined, reason, ts.Diagnostics.Cannot_find_type_definition_file_for_0, [typeReferenceDirective]);
                }
                if (saveResolution) {
                    resolvedTypeReferenceDirectives.set(typeReferenceDirective, resolvedTypeReferenceDirective);
                }
            }
            function processLibReferenceDirectives(file) {
                ts.forEach(file.libReferenceDirectives, function (libReference, index) {
                    var libName = ts.toFileNameLowerCase(libReference.fileName);
                    var libFileName = ts.libMap.get(libName);
                    if (libFileName) {
                        // we ignore any 'no-default-lib' reference set on this file.
                        processRootFile(ts.combinePaths(defaultLibraryPath, libFileName), /*isDefaultLib*/ true, /*ignoreNoDefaultLib*/ true, { kind: ts.FileIncludeKind.LibReferenceDirective, file: file.path, index: index, });
                    }
                    else {
                        var unqualifiedLibName = ts.removeSuffix(ts.removePrefix(libName, "lib."), ".d.ts");
                        var suggestion = ts.getSpellingSuggestion(unqualifiedLibName, ts.libs, ts.identity);
                        var diagnostic = suggestion ? ts.Diagnostics.Cannot_find_lib_definition_for_0_Did_you_mean_1 : ts.Diagnostics.Cannot_find_lib_definition_for_0;
                        (fileProcessingDiagnostics || (fileProcessingDiagnostics = [])).push({
                            kind: 0 /* FilePreprocessingReferencedDiagnostic */,
                            reason: { kind: ts.FileIncludeKind.LibReferenceDirective, file: file.path, index: index, },
                            diagnostic: diagnostic,
                            args: [libName, suggestion]
                        });
                    }
                });
            }
            function getCanonicalFileName(fileName) {
                return host.getCanonicalFileName(fileName);
            }
            function processImportedModules(file) {
                var _a;
                collectExternalModuleReferences(file);
                if (file.imports.length || file.moduleAugmentations.length) {
                    // Because global augmentation doesn't have string literal name, we can check for global augmentation as such.
                    var moduleNames = getModuleNames(file);
                    var resolutions = resolveModuleNamesReusingOldState(moduleNames, file);
                    ts.Debug.assert(resolutions.length === moduleNames.length);
                    var optionsForFile = (useSourceOfProjectReferenceRedirect ? (_a = getRedirectReferenceForResolution(file)) === null || _a === void 0 ? void 0 : _a.commandLine.options : undefined) || options;
                    for (var index = 0; index < moduleNames.length; index++) {
                        var resolution = resolutions[index];
                        ts.setResolvedModule(file, moduleNames[index], resolution);
                        if (!resolution) {
                            continue;
                        }
                        var isFromNodeModulesSearch = resolution.isExternalLibraryImport;
                        var isJsFile = !ts.resolutionExtensionIsTSOrJson(resolution.extension);
                        var isJsFileFromNodeModules = isFromNodeModulesSearch && isJsFile;
                        var resolvedFileName = resolution.resolvedFileName;
                        if (isFromNodeModulesSearch) {
                            currentNodeModulesDepth++;
                        }
                        // add file to program only if:
                        // - resolution was successful
                        // - noResolve is falsy
                        // - module name comes from the list of imports
                        // - it's not a top level JavaScript module that exceeded the search max
                        var elideImport = isJsFileFromNodeModules && currentNodeModulesDepth > maxNodeModuleJsDepth;
                        // Don't add the file if it has a bad extension (e.g. 'tsx' if we don't have '--allowJs')
                        // This may still end up being an untyped module -- the file won't be included but imports will be allowed.
                        var shouldAddFile = resolvedFileName
                            && !getResolutionDiagnostic(optionsForFile, resolution)
                            && !optionsForFile.noResolve
                            && index < file.imports.length
                            && !elideImport
                            && !(isJsFile && !ts.getAllowJSCompilerOption(optionsForFile))
                            && (ts.isInJSFile(file.imports[index]) || !(file.imports[index].flags & 4194304 /* JSDoc */));
                        if (elideImport) {
                            modulesWithElidedImports.set(file.path, true);
                        }
                        else if (shouldAddFile) {
                            var path = toPath(resolvedFileName);
                            findSourceFile(resolvedFileName, path, 
                            /*isDefaultLib*/ false, 
                            /*ignoreNoDefaultLib*/ false, { kind: ts.FileIncludeKind.Import, file: file.path, index: index, }, resolution.packageId);
                        }
                        if (isFromNodeModulesSearch) {
                            currentNodeModulesDepth--;
                        }
                    }
                }
                else {
                    // no imports - drop cached module resolutions
                    file.resolvedModules = undefined;
                }
            }
            function checkSourceFilesBelongToPath(sourceFiles, rootDirectory) {
                var allFilesBelongToPath = true;
                var absoluteRootDirectoryPath = host.getCanonicalFileName(ts.getNormalizedAbsolutePath(rootDirectory, currentDirectory));
                for (var _i = 0, sourceFiles_2 = sourceFiles; _i < sourceFiles_2.length; _i++) {
                    var sourceFile = sourceFiles_2[_i];
                    if (!sourceFile.isDeclarationFile) {
                        var absoluteSourceFilePath = host.getCanonicalFileName(ts.getNormalizedAbsolutePath(sourceFile.fileName, currentDirectory));
                        if (absoluteSourceFilePath.indexOf(absoluteRootDirectoryPath) !== 0) {
                            addProgramDiagnosticExplainingFile(sourceFile, ts.Diagnostics.File_0_is_not_under_rootDir_1_rootDir_is_expected_to_contain_all_source_files, [sourceFile.fileName, rootDirectory]);
                            allFilesBelongToPath = false;
                        }
                    }
                }
                return allFilesBelongToPath;
            }
            function parseProjectReferenceConfigFile(ref) {
                if (!projectReferenceRedirects) {
                    projectReferenceRedirects = new ts.Map();
                }
                // The actual filename (i.e. add "/tsconfig.json" if necessary)
                var refPath = resolveProjectReferencePath(ref);
                var sourceFilePath = toPath(refPath);
                var fromCache = projectReferenceRedirects.get(sourceFilePath);
                if (fromCache !== undefined) {
                    return fromCache || undefined;
                }
                var commandLine;
                var sourceFile;
                if (host.getParsedCommandLine) {
                    commandLine = host.getParsedCommandLine(refPath);
                    if (!commandLine) {
                        addFileToFilesByName(/*sourceFile*/ undefined, sourceFilePath, /*redirectedPath*/ undefined);
                        projectReferenceRedirects.set(sourceFilePath, false);
                        return undefined;
                    }
                    sourceFile = ts.Debug.checkDefined(commandLine.options.configFile);
                    ts.Debug.assert(!sourceFile.path || sourceFile.path === sourceFilePath);
                    addFileToFilesByName(sourceFile, sourceFilePath, /*redirectedPath*/ undefined);
                }
                else {
                    // An absolute path pointing to the containing directory of the config file
                    var basePath = ts.getNormalizedAbsolutePath(ts.getDirectoryPath(refPath), host.getCurrentDirectory());
                    sourceFile = host.getSourceFile(refPath, 100 /* JSON */);
                    addFileToFilesByName(sourceFile, sourceFilePath, /*redirectedPath*/ undefined);
                    if (sourceFile === undefined) {
                        projectReferenceRedirects.set(sourceFilePath, false);
                        return undefined;
                    }
                    commandLine = ts.parseJsonSourceFileConfigFileContent(sourceFile, configParsingHost, basePath, /*existingOptions*/ undefined, refPath);
                }
                sourceFile.fileName = refPath;
                sourceFile.path = sourceFilePath;
                sourceFile.resolvedPath = sourceFilePath;
                sourceFile.originalFileName = refPath;
                var resolvedRef = { commandLine: commandLine, sourceFile: sourceFile };
                projectReferenceRedirects.set(sourceFilePath, resolvedRef);
                if (commandLine.projectReferences) {
                    resolvedRef.references = commandLine.projectReferences.map(parseProjectReferenceConfigFile);
                }
                return resolvedRef;
            }
            function verifyCompilerOptions() {
                if (options.strictPropertyInitialization && !ts.getStrictOptionValue(options, "strictNullChecks")) {
                    createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1, "strictPropertyInitialization", "strictNullChecks");
                }
                if (options.exactOptionalPropertyTypes && !ts.getStrictOptionValue(options, "strictNullChecks")) {
                    createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1, "exactOptionalPropertyTypes", "strictNullChecks");
                }
                if (options.isolatedModules) {
                    if (options.out) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "out", "isolatedModules");
                    }
                    if (options.outFile) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "outFile", "isolatedModules");
                    }
                }
                if (options.inlineSourceMap) {
                    if (options.sourceMap) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "sourceMap", "inlineSourceMap");
                    }
                    if (options.mapRoot) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "mapRoot", "inlineSourceMap");
                    }
                }
                if (options.composite) {
                    if (options.declaration === false) {
                        createDiagnosticForOptionName(ts.Diagnostics.Composite_projects_may_not_disable_declaration_emit, "declaration");
                    }
                    if (options.incremental === false) {
                        createDiagnosticForOptionName(ts.Diagnostics.Composite_projects_may_not_disable_incremental_compilation, "declaration");
                    }
                }
                var outputFile = ts.outFile(options);
                if (options.tsBuildInfoFile) {
                    if (!ts.isIncrementalCompilation(options)) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2, "tsBuildInfoFile", "incremental", "composite");
                    }
                }
                else if (options.incremental && !outputFile && !options.configFilePath) {
                    programDiagnostics.add(ts.createCompilerDiagnostic(ts.Diagnostics.Option_incremental_can_only_be_specified_using_tsconfig_emitting_to_single_file_or_when_option_tsBuildInfoFile_is_specified));
                }
                verifyProjectReferences();
                // List of collected files is complete; validate exhautiveness if this is a project with a file list
                if (options.composite) {
                    var rootPaths = new ts.Set(rootNames.map(toPath));
                    for (var _i = 0, files_4 = files; _i < files_4.length; _i++) {
                        var file = files_4[_i];
                        // Ignore file that is not emitted
                        if (ts.sourceFileMayBeEmitted(file, program) && !rootPaths.has(file.path)) {
                            addProgramDiagnosticExplainingFile(file, ts.Diagnostics.File_0_is_not_listed_within_the_file_list_of_project_1_Projects_must_list_all_files_or_use_an_include_pattern, [file.fileName, options.configFilePath || ""]);
                        }
                    }
                }
                if (options.paths) {
                    for (var key in options.paths) {
                        if (!ts.hasProperty(options.paths, key)) {
                            continue;
                        }
                        if (!ts.hasZeroOrOneAsteriskCharacter(key)) {
                            createDiagnosticForOptionPaths(/*onKey*/ true, key, ts.Diagnostics.Pattern_0_can_have_at_most_one_Asterisk_character, key);
                        }
                        if (ts.isArray(options.paths[key])) {
                            var len = options.paths[key].length;
                            if (len === 0) {
                                createDiagnosticForOptionPaths(/*onKey*/ false, key, ts.Diagnostics.Substitutions_for_pattern_0_shouldn_t_be_an_empty_array, key);
                            }
                            for (var i = 0; i < len; i++) {
                                var subst = options.paths[key][i];
                                var typeOfSubst = typeof subst;
                                if (typeOfSubst === "string") {
                                    if (!ts.hasZeroOrOneAsteriskCharacter(subst)) {
                                        createDiagnosticForOptionPathKeyValue(key, i, ts.Diagnostics.Substitution_0_in_pattern_1_can_have_at_most_one_Asterisk_character, subst, key);
                                    }
                                    if (!options.baseUrl && !ts.pathIsRelative(subst) && !ts.pathIsAbsolute(subst)) {
                                        createDiagnosticForOptionPathKeyValue(key, i, ts.Diagnostics.Non_relative_paths_are_not_allowed_when_baseUrl_is_not_set_Did_you_forget_a_leading_Slash);
                                    }
                                }
                                else {
                                    createDiagnosticForOptionPathKeyValue(key, i, ts.Diagnostics.Substitution_0_for_pattern_1_has_incorrect_type_expected_string_got_2, subst, key, typeOfSubst);
                                }
                            }
                        }
                        else {
                            createDiagnosticForOptionPaths(/*onKey*/ false, key, ts.Diagnostics.Substitutions_for_pattern_0_should_be_an_array, key);
                        }
                    }
                }
                if (!options.sourceMap && !options.inlineSourceMap) {
                    if (options.inlineSources) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided, "inlineSources");
                    }
                    if (options.sourceRoot) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_can_only_be_used_when_either_option_inlineSourceMap_or_option_sourceMap_is_provided, "sourceRoot");
                    }
                }
                if (options.out && options.outFile) {
                    createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "out", "outFile");
                }
                if (options.mapRoot && !(options.sourceMap || options.declarationMap)) {
                    // Error to specify --mapRoot without --sourcemap
                    createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2, "mapRoot", "sourceMap", "declarationMap");
                }
                if (options.declarationDir) {
                    if (!ts.getEmitDeclarations(options)) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2, "declarationDir", "declaration", "composite");
                    }
                    if (outputFile) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "declarationDir", options.out ? "out" : "outFile");
                    }
                }
                if (options.declarationMap && !ts.getEmitDeclarations(options)) {
                    createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2, "declarationMap", "declaration", "composite");
                }
                if (options.lib && options.noLib) {
                    createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "lib", "noLib");
                }
                if (options.noImplicitUseStrict && ts.getStrictOptionValue(options, "alwaysStrict")) {
                    createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "noImplicitUseStrict", "alwaysStrict");
                }
                var languageVersion = options.target || 0 /* ES3 */;
                var firstNonAmbientExternalModuleSourceFile = ts.find(files, function (f) { return ts.isExternalModule(f) && !f.isDeclarationFile; });
                if (options.isolatedModules) {
                    if (options.module === ts.ModuleKind.None && languageVersion < 2 /* ES2015 */) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_isolatedModules_can_only_be_used_when_either_option_module_is_provided_or_option_target_is_ES2015_or_higher, "isolatedModules", "target");
                    }
                    if (options.preserveConstEnums === false) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_preserveConstEnums_cannot_be_disabled_when_isolatedModules_is_enabled, "preserveConstEnums", "isolatedModules");
                    }
                    var firstNonExternalModuleSourceFile = ts.find(files, function (f) { return !ts.isExternalModule(f) && !ts.isSourceFileJS(f) && !f.isDeclarationFile && f.scriptKind !== 6 /* JSON */; });
                    if (firstNonExternalModuleSourceFile) {
                        var span = ts.getErrorSpanForNode(firstNonExternalModuleSourceFile, firstNonExternalModuleSourceFile);
                        programDiagnostics.add(ts.createFileDiagnostic(firstNonExternalModuleSourceFile, span.start, span.length, ts.Diagnostics._0_cannot_be_compiled_under_isolatedModules_because_it_is_considered_a_global_script_file_Add_an_import_export_or_an_empty_export_statement_to_make_it_a_module, ts.getBaseFileName(firstNonExternalModuleSourceFile.fileName)));
                    }
                }
                else if (firstNonAmbientExternalModuleSourceFile && languageVersion < 2 /* ES2015 */ && options.module === ts.ModuleKind.None) {
                    // We cannot use createDiagnosticFromNode because nodes do not have parents yet
                    var span = ts.getErrorSpanForNode(firstNonAmbientExternalModuleSourceFile, firstNonAmbientExternalModuleSourceFile.externalModuleIndicator);
                    programDiagnostics.add(ts.createFileDiagnostic(firstNonAmbientExternalModuleSourceFile, span.start, span.length, ts.Diagnostics.Cannot_use_imports_exports_or_module_augmentations_when_module_is_none));
                }
                // Cannot specify module gen that isn't amd or system with --out
                if (outputFile && !options.emitDeclarationOnly) {
                    if (options.module && !(options.module === ts.ModuleKind.AMD || options.module === ts.ModuleKind.System)) {
                        createDiagnosticForOptionName(ts.Diagnostics.Only_amd_and_system_modules_are_supported_alongside_0, options.out ? "out" : "outFile", "module");
                    }
                    else if (options.module === undefined && firstNonAmbientExternalModuleSourceFile) {
                        var span = ts.getErrorSpanForNode(firstNonAmbientExternalModuleSourceFile, firstNonAmbientExternalModuleSourceFile.externalModuleIndicator);
                        programDiagnostics.add(ts.createFileDiagnostic(firstNonAmbientExternalModuleSourceFile, span.start, span.length, ts.Diagnostics.Cannot_compile_modules_using_option_0_unless_the_module_flag_is_amd_or_system, options.out ? "out" : "outFile"));
                    }
                }
                if (options.resolveJsonModule) {
                    if (ts.getEmitModuleResolutionKind(options) !== ts.ModuleResolutionKind.NodeJs) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_resolveJsonModule_cannot_be_specified_without_node_module_resolution_strategy, "resolveJsonModule");
                    }
                    // Any emit other than common js, amd, es2015 or esnext is error
                    else if (!ts.hasJsonModuleEmitEnabled(options)) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_resolveJsonModule_can_only_be_specified_when_module_code_generation_is_commonjs_amd_es2015_or_esNext, "resolveJsonModule", "module");
                    }
                }
                // there has to be common source directory if user specified --outdir || --rootDir || --sourceRoot
                // if user specified --mapRoot, there needs to be common source directory if there would be multiple files being emitted
                if (options.outDir || // there is --outDir specified
                    options.rootDir || // there is --rootDir specified
                    options.sourceRoot || // there is --sourceRoot specified
                    options.mapRoot) { // there is --mapRoot specified
                    // Precalculate and cache the common source directory
                    var dir = getCommonSourceDirectory();
                    // If we failed to find a good common directory, but outDir is specified and at least one of our files is on a windows drive/URL/other resource, add a failure
                    if (options.outDir && dir === "" && files.some(function (file) { return ts.getRootLength(file.fileName) > 1; })) {
                        createDiagnosticForOptionName(ts.Diagnostics.Cannot_find_the_common_subdirectory_path_for_the_input_files, "outDir");
                    }
                }
                if (options.useDefineForClassFields && languageVersion === 0 /* ES3 */) {
                    createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_when_option_target_is_ES3, "useDefineForClassFields");
                }
                if (options.checkJs && !ts.getAllowJSCompilerOption(options)) {
                    programDiagnostics.add(ts.createCompilerDiagnostic(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1, "checkJs", "allowJs"));
                }
                if (options.emitDeclarationOnly) {
                    if (!ts.getEmitDeclarations(options)) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1_or_option_2, "emitDeclarationOnly", "declaration", "composite");
                    }
                    if (options.noEmit) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "emitDeclarationOnly", "noEmit");
                    }
                }
                if (options.emitDecoratorMetadata &&
                    !options.experimentalDecorators) {
                    createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1, "emitDecoratorMetadata", "experimentalDecorators");
                }
                if (options.jsxFactory) {
                    if (options.reactNamespace) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_with_option_1, "reactNamespace", "jsxFactory");
                    }
                    if (options.jsx === 4 /* ReactJSX */ || options.jsx === 5 /* ReactJSXDev */) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_when_option_jsx_is_1, "jsxFactory", ts.inverseJsxOptionMap.get("" + options.jsx));
                    }
                    if (!ts.parseIsolatedEntityName(options.jsxFactory, languageVersion)) {
                        createOptionValueDiagnostic("jsxFactory", ts.Diagnostics.Invalid_value_for_jsxFactory_0_is_not_a_valid_identifier_or_qualified_name, options.jsxFactory);
                    }
                }
                else if (options.reactNamespace && !ts.isIdentifierText(options.reactNamespace, languageVersion)) {
                    createOptionValueDiagnostic("reactNamespace", ts.Diagnostics.Invalid_value_for_reactNamespace_0_is_not_a_valid_identifier, options.reactNamespace);
                }
                if (options.jsxFragmentFactory) {
                    if (!options.jsxFactory) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_without_specifying_option_1, "jsxFragmentFactory", "jsxFactory");
                    }
                    if (options.jsx === 4 /* ReactJSX */ || options.jsx === 5 /* ReactJSXDev */) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_when_option_jsx_is_1, "jsxFragmentFactory", ts.inverseJsxOptionMap.get("" + options.jsx));
                    }
                    if (!ts.parseIsolatedEntityName(options.jsxFragmentFactory, languageVersion)) {
                        createOptionValueDiagnostic("jsxFragmentFactory", ts.Diagnostics.Invalid_value_for_jsxFragmentFactory_0_is_not_a_valid_identifier_or_qualified_name, options.jsxFragmentFactory);
                    }
                }
                if (options.reactNamespace) {
                    if (options.jsx === 4 /* ReactJSX */ || options.jsx === 5 /* ReactJSXDev */) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_when_option_jsx_is_1, "reactNamespace", ts.inverseJsxOptionMap.get("" + options.jsx));
                    }
                }
                if (options.jsxImportSource) {
                    if (options.jsx === 2 /* React */) {
                        createDiagnosticForOptionName(ts.Diagnostics.Option_0_cannot_be_specified_when_option_jsx_is_1, "jsxImportSource", ts.inverseJsxOptionMap.get("" + options.jsx));
                    }
                }
                // If the emit is enabled make sure that every output file is unique and not overwriting any of the input files
                if (!options.noEmit && !options.suppressOutputPathCheck) {
                    var emitHost = getEmitHost();
                    var emitFilesSeen_1 = new ts.Set();
                    ts.forEachEmittedFile(emitHost, function (emitFileNames) {
                        if (!options.emitDeclarationOnly) {
                            verifyEmitFilePath(emitFileNames.jsFilePath, emitFilesSeen_1);
                        }
                        verifyEmitFilePath(emitFileNames.declarationFilePath, emitFilesSeen_1);
                    });
                }
                // Verify that all the emit files are unique and don't overwrite input files
                function verifyEmitFilePath(emitFileName, emitFilesSeen) {
                    if (emitFileName) {
                        var emitFilePath = toPath(emitFileName);
                        // Report error if the output overwrites input file
                        if (filesByName.has(emitFilePath)) {
                            var chain = void 0;
                            if (!options.configFilePath) {
                                // The program is from either an inferred project or an external project
                                chain = ts.chainDiagnosticMessages(/*details*/ undefined, ts.Diagnostics.Adding_a_tsconfig_json_file_will_help_organize_projects_that_contain_both_TypeScript_and_JavaScript_files_Learn_more_at_https_Colon_Slash_Slashaka_ms_Slashtsconfig);
                            }
                            chain = ts.chainDiagnosticMessages(chain, ts.Diagnostics.Cannot_write_file_0_because_it_would_overwrite_input_file, emitFileName);
                            blockEmittingOfFile(emitFileName, ts.createCompilerDiagnosticFromMessageChain(chain));
                        }
                        var emitFileKey = !host.useCaseSensitiveFileNames() ? ts.toFileNameLowerCase(emitFilePath) : emitFilePath;
                        // Report error if multiple files write into same file
                        if (emitFilesSeen.has(emitFileKey)) {
                            // Already seen the same emit file - report error
                            blockEmittingOfFile(emitFileName, ts.createCompilerDiagnostic(ts.Diagnostics.Cannot_write_file_0_because_it_would_be_overwritten_by_multiple_input_files, emitFileName));
                        }
                        else {
                            emitFilesSeen.add(emitFileKey);
                        }
                    }
                }
            }
            function createDiagnosticExplainingFile(file, fileProcessingReason, diagnostic, args) {
                var _a;
                var fileIncludeReasons;
                var relatedInfo;
                var locationReason = isReferencedFile(fileProcessingReason) ? fileProcessingReason : undefined;
                if (file)
                    (_a = fileReasons.get(file.path)) === null || _a === void 0 ? void 0 : _a.forEach(processReason);
                if (fileProcessingReason)
                    processReason(fileProcessingReason);
                // If we have location and there is only one reason file is in which is the location, dont add details for file include
                if (locationReason && (fileIncludeReasons === null || fileIncludeReasons === void 0 ? void 0 : fileIncludeReasons.length) === 1)
                    fileIncludeReasons = undefined;
                var location = locationReason && getReferencedFileLocation(getSourceFileByPath, locationReason);
                var fileIncludeReasonDetails = fileIncludeReasons && ts.chainDiagnosticMessages(fileIncludeReasons, ts.Diagnostics.The_file_is_in_the_program_because_Colon);
                var redirectInfo = file && ts.explainIfFileIsRedirect(file);
                var chain = ts.chainDiagnosticMessages.apply(void 0, __spreadArray([redirectInfo ? fileIncludeReasonDetails ? __spreadArray([fileIncludeReasonDetails], redirectInfo, true) : redirectInfo : fileIncludeReasonDetails, diagnostic], args || ts.emptyArray, false));
                return location && isReferenceFileLocation(location) ?
                    ts.createFileDiagnosticFromMessageChain(location.file, location.pos, location.end - location.pos, chain, relatedInfo) :
                    ts.createCompilerDiagnosticFromMessageChain(chain, relatedInfo);
                function processReason(reason) {
                    (fileIncludeReasons || (fileIncludeReasons = [])).push(ts.fileIncludeReasonToDiagnostics(program, reason));
                    if (!locationReason && isReferencedFile(reason)) {
                        // Report error at first reference file or file currently in processing and dont report in related information
                        locationReason = reason;
                    }
                    else if (locationReason !== reason) {
                        relatedInfo = ts.append(relatedInfo, fileIncludeReasonToRelatedInformation(reason));
                    }
                    // Remove fileProcessingReason if its already included in fileReasons of the program
                    if (reason === fileProcessingReason)
                        fileProcessingReason = undefined;
                }
            }
            function addFilePreprocessingFileExplainingDiagnostic(file, fileProcessingReason, diagnostic, args) {
                (fileProcessingDiagnostics || (fileProcessingDiagnostics = [])).push({
                    kind: 1 /* FilePreprocessingFileExplainingDiagnostic */,
                    file: file && file.path,
                    fileProcessingReason: fileProcessingReason,
                    diagnostic: diagnostic,
                    args: args
                });
            }
            function addProgramDiagnosticExplainingFile(file, diagnostic, args) {
                programDiagnostics.add(createDiagnosticExplainingFile(file, /*fileProcessingReason*/ undefined, diagnostic, args));
            }
            function fileIncludeReasonToRelatedInformation(reason) {
                if (isReferencedFile(reason)) {
                    var referenceLocation = getReferencedFileLocation(getSourceFileByPath, reason);
                    var message_2;
                    switch (reason.kind) {
                        case ts.FileIncludeKind.Import:
                            message_2 = ts.Diagnostics.File_is_included_via_import_here;
                            break;
                        case ts.FileIncludeKind.ReferenceFile:
                            message_2 = ts.Diagnostics.File_is_included_via_reference_here;
                            break;
                        case ts.FileIncludeKind.TypeReferenceDirective:
                            message_2 = ts.Diagnostics.File_is_included_via_type_library_reference_here;
                            break;
                        case ts.FileIncludeKind.LibReferenceDirective:
                            message_2 = ts.Diagnostics.File_is_included_via_library_reference_here;
                            break;
                        default:
                            ts.Debug.assertNever(reason);
                    }
                    return isReferenceFileLocation(referenceLocation) ? ts.createFileDiagnostic(referenceLocation.file, referenceLocation.pos, referenceLocation.end - referenceLocation.pos, message_2) : undefined;
                }
                if (!options.configFile)
                    return undefined;
                var configFileNode;
                var message;
                switch (reason.kind) {
                    case ts.FileIncludeKind.RootFile:
                        if (!options.configFile.configFileSpecs)
                            return undefined;
                        var fileName = ts.getNormalizedAbsolutePath(rootNames[reason.index], currentDirectory);
                        var matchedByFiles = ts.getMatchedFileSpec(program, fileName);
                        if (matchedByFiles) {
                            configFileNode = ts.getTsConfigPropArrayElementValue(options.configFile, "files", matchedByFiles);
                            message = ts.Diagnostics.File_is_matched_by_files_list_specified_here;
                            break;
                        }
                        var matchedByInclude = ts.getMatchedIncludeSpec(program, fileName);
                        // Could be additional files specified as roots
                        if (!matchedByInclude)
                            return undefined;
                        configFileNode = ts.getTsConfigPropArrayElementValue(options.configFile, "include", matchedByInclude);
                        message = ts.Diagnostics.File_is_matched_by_include_pattern_specified_here;
                        break;
                    case ts.FileIncludeKind.SourceFromProjectReference:
                    case ts.FileIncludeKind.OutputFromProjectReference:
                        var referencedResolvedRef_1 = ts.Debug.checkDefined(resolvedProjectReferences === null || resolvedProjectReferences === void 0 ? void 0 : resolvedProjectReferences[reason.index]);
                        var referenceInfo = forEachProjectReference(projectReferences, resolvedProjectReferences, function (resolvedRef, parent, index) {
                            return resolvedRef === referencedResolvedRef_1 ? { sourceFile: (parent === null || parent === void 0 ? void 0 : parent.sourceFile) || options.configFile, index: index } : undefined;
                        });
                        if (!referenceInfo)
                            return undefined;
                        var sourceFile = referenceInfo.sourceFile, index = referenceInfo.index;
                        var referencesSyntax = ts.firstDefined(ts.getTsConfigPropArray(sourceFile, "references"), function (property) { return ts.isArrayLiteralExpression(property.initializer) ? property.initializer : undefined; });
                        return referencesSyntax && referencesSyntax.elements.length > index ?
                            ts.createDiagnosticForNodeInSourceFile(sourceFile, referencesSyntax.elements[index], reason.kind === ts.FileIncludeKind.OutputFromProjectReference ?
                                ts.Diagnostics.File_is_output_from_referenced_project_specified_here :
                                ts.Diagnostics.File_is_source_from_referenced_project_specified_here) :
                            undefined;
                    case ts.FileIncludeKind.AutomaticTypeDirectiveFile:
                        if (!options.types)
                            return undefined;
                        configFileNode = getOptionsSyntaxByArrayElementValue("types", reason.typeReference);
                        message = ts.Diagnostics.File_is_entry_point_of_type_library_specified_here;
                        break;
                    case ts.FileIncludeKind.LibFile:
                        if (reason.index !== undefined) {
                            configFileNode = getOptionsSyntaxByArrayElementValue("lib", options.lib[reason.index]);
                            message = ts.Diagnostics.File_is_library_specified_here;
                            break;
                        }
                        var target = ts.forEachEntry(ts.targetOptionDeclaration.type, function (value, key) { return value === options.target ? key : undefined; });
                        configFileNode = target ? getOptionsSyntaxByValue("target", target) : undefined;
                        message = ts.Diagnostics.File_is_default_library_for_target_specified_here;
                        break;
                    default:
                        ts.Debug.assertNever(reason);
                }
                return configFileNode && ts.createDiagnosticForNodeInSourceFile(options.configFile, configFileNode, message);
            }
            function verifyProjectReferences() {
                var buildInfoPath = !options.suppressOutputPathCheck ? ts.getTsBuildInfoEmitOutputFilePath(options) : undefined;
                forEachProjectReference(projectReferences, resolvedProjectReferences, function (resolvedRef, parent, index) {
                    var ref = (parent ? parent.commandLine.projectReferences : projectReferences)[index];
                    var parentFile = parent && parent.sourceFile;
                    if (!resolvedRef) {
                        createDiagnosticForReference(parentFile, index, ts.Diagnostics.File_0_not_found, ref.path);
                        return;
                    }
                    var options = resolvedRef.commandLine.options;
                    if (!options.composite || options.noEmit) {
                        // ok to not have composite if the current program is container only
                        var inputs = parent ? parent.commandLine.fileNames : rootNames;
                        if (inputs.length) {
                            if (!options.composite)
                                createDiagnosticForReference(parentFile, index, ts.Diagnostics.Referenced_project_0_must_have_setting_composite_Colon_true, ref.path);
                            if (options.noEmit)
                                createDiagnosticForReference(parentFile, index, ts.Diagnostics.Referenced_project_0_may_not_disable_emit, ref.path);
                        }
                    }
                    if (ref.prepend) {
                        var out = ts.outFile(options);
                        if (out) {
                            if (!host.fileExists(out)) {
                                createDiagnosticForReference(parentFile, index, ts.Diagnostics.Output_file_0_from_project_1_does_not_exist, out, ref.path);
                            }
                        }
                        else {
                            createDiagnosticForReference(parentFile, index, ts.Diagnostics.Cannot_prepend_project_0_because_it_does_not_have_outFile_set, ref.path);
                        }
                    }
                    if (!parent && buildInfoPath && buildInfoPath === ts.getTsBuildInfoEmitOutputFilePath(options)) {
                        createDiagnosticForReference(parentFile, index, ts.Diagnostics.Cannot_write_file_0_because_it_will_overwrite_tsbuildinfo_file_generated_by_referenced_project_1, buildInfoPath, ref.path);
                        hasEmitBlockingDiagnostics.set(toPath(buildInfoPath), true);
                    }
                });
            }
            function createDiagnosticForOptionPathKeyValue(key, valueIndex, message, arg0, arg1, arg2) {
                var needCompilerDiagnostic = true;
                var pathsSyntax = getOptionPathsSyntax();
                for (var _i = 0, pathsSyntax_1 = pathsSyntax; _i < pathsSyntax_1.length; _i++) {
                    var pathProp = pathsSyntax_1[_i];
                    if (ts.isObjectLiteralExpression(pathProp.initializer)) {
                        for (var _a = 0, _b = ts.getPropertyAssignment(pathProp.initializer, key); _a < _b.length; _a++) {
                            var keyProps = _b[_a];
                            var initializer = keyProps.initializer;
                            if (ts.isArrayLiteralExpression(initializer) && initializer.elements.length > valueIndex) {
                                programDiagnostics.add(ts.createDiagnosticForNodeInSourceFile(options.configFile, initializer.elements[valueIndex], message, arg0, arg1, arg2));
                                needCompilerDiagnostic = false;
                            }
                        }
                    }
                }
                if (needCompilerDiagnostic) {
                    programDiagnostics.add(ts.createCompilerDiagnostic(message, arg0, arg1, arg2));
                }
            }
            function createDiagnosticForOptionPaths(onKey, key, message, arg0) {
                var needCompilerDiagnostic = true;
                var pathsSyntax = getOptionPathsSyntax();
                for (var _i = 0, pathsSyntax_2 = pathsSyntax; _i < pathsSyntax_2.length; _i++) {
                    var pathProp = pathsSyntax_2[_i];
                    if (ts.isObjectLiteralExpression(pathProp.initializer) &&
                        createOptionDiagnosticInObjectLiteralSyntax(pathProp.initializer, onKey, key, /*key2*/ undefined, message, arg0)) {
                        needCompilerDiagnostic = false;
                    }
                }
                if (needCompilerDiagnostic) {
                    programDiagnostics.add(ts.createCompilerDiagnostic(message, arg0));
                }
            }
            function getOptionsSyntaxByName(name) {
                var compilerOptionsObjectLiteralSyntax = getCompilerOptionsObjectLiteralSyntax();
                return compilerOptionsObjectLiteralSyntax && ts.getPropertyAssignment(compilerOptionsObjectLiteralSyntax, name);
            }
            function getOptionPathsSyntax() {
                return getOptionsSyntaxByName("paths") || ts.emptyArray;
            }
            function getOptionsSyntaxByValue(name, value) {
                var syntaxByName = getOptionsSyntaxByName(name);
                return syntaxByName && ts.firstDefined(syntaxByName, function (property) { return ts.isStringLiteral(property.initializer) && property.initializer.text === value ? property.initializer : undefined; });
            }
            function getOptionsSyntaxByArrayElementValue(name, value) {
                var compilerOptionsObjectLiteralSyntax = getCompilerOptionsObjectLiteralSyntax();
                return compilerOptionsObjectLiteralSyntax && ts.getPropertyArrayElementValue(compilerOptionsObjectLiteralSyntax, name, value);
            }
            function createDiagnosticForOptionName(message, option1, option2, option3) {
                createDiagnosticForOption(/*onKey*/ true, option1, option2, message, option1, option2, option3);
            }
            function createOptionValueDiagnostic(option1, message, arg0) {
                createDiagnosticForOption(/*onKey*/ false, option1, /*option2*/ undefined, message, arg0);
            }
            function createDiagnosticForReference(sourceFile, index, message, arg0, arg1) {
                var referencesSyntax = ts.firstDefined(ts.getTsConfigPropArray(sourceFile || options.configFile, "references"), function (property) { return ts.isArrayLiteralExpression(property.initializer) ? property.initializer : undefined; });
                if (referencesSyntax && referencesSyntax.elements.length > index) {
                    programDiagnostics.add(ts.createDiagnosticForNodeInSourceFile(sourceFile || options.configFile, referencesSyntax.elements[index], message, arg0, arg1));
                }
                else {
                    programDiagnostics.add(ts.createCompilerDiagnostic(message, arg0, arg1));
                }
            }
            function createDiagnosticForOption(onKey, option1, option2, message, arg0, arg1, arg2) {
                var compilerOptionsObjectLiteralSyntax = getCompilerOptionsObjectLiteralSyntax();
                var needCompilerDiagnostic = !compilerOptionsObjectLiteralSyntax ||
                    !createOptionDiagnosticInObjectLiteralSyntax(compilerOptionsObjectLiteralSyntax, onKey, option1, option2, message, arg0, arg1, arg2);
                if (needCompilerDiagnostic) {
                    programDiagnostics.add(ts.createCompilerDiagnostic(message, arg0, arg1, arg2));
                }
            }
            function getCompilerOptionsObjectLiteralSyntax() {
                if (_compilerOptionsObjectLiteralSyntax === undefined) {
                    _compilerOptionsObjectLiteralSyntax = false;
                    var jsonObjectLiteral = ts.getTsConfigObjectLiteralExpression(options.configFile);
                    if (jsonObjectLiteral) {
                        for (var _i = 0, _a = ts.getPropertyAssignment(jsonObjectLiteral, "compilerOptions"); _i < _a.length; _i++) {
                            var prop = _a[_i];
                            if (ts.isObjectLiteralExpression(prop.initializer)) {
                                _compilerOptionsObjectLiteralSyntax = prop.initializer;
                                break;
                            }
                        }
                    }
                }
                return _compilerOptionsObjectLiteralSyntax || undefined;
            }
            function createOptionDiagnosticInObjectLiteralSyntax(objectLiteral, onKey, key1, key2, message, arg0, arg1, arg2) {
                var props = ts.getPropertyAssignment(objectLiteral, key1, key2);
                for (var _i = 0, props_3 = props; _i < props_3.length; _i++) {
                    var prop = props_3[_i];
                    programDiagnostics.add(ts.createDiagnosticForNodeInSourceFile(options.configFile, onKey ? prop.name : prop.initializer, message, arg0, arg1, arg2));
                }
                return !!props.length;
            }
            function blockEmittingOfFile(emitFileName, diag) {
                hasEmitBlockingDiagnostics.set(toPath(emitFileName), true);
                programDiagnostics.add(diag);
            }
            function isEmittedFile(file) {
                if (options.noEmit) {
                    return false;
                }
                // If this is source file, its not emitted file
                var filePath = toPath(file);
                if (getSourceFileByPath(filePath)) {
                    return false;
                }
                // If options have --outFile or --out just check that
                var out = ts.outFile(options);
                if (out) {
                    return isSameFile(filePath, out) || isSameFile(filePath, ts.removeFileExtension(out) + ".d.ts" /* Dts */);
                }
                // If declarationDir is specified, return if its a file in that directory
                if (options.declarationDir && ts.containsPath(options.declarationDir, filePath, currentDirectory, !host.useCaseSensitiveFileNames())) {
                    return true;
                }
                // If --outDir, check if file is in that directory
                if (options.outDir) {
                    return ts.containsPath(options.outDir, filePath, currentDirectory, !host.useCaseSensitiveFileNames());
                }
                if (ts.fileExtensionIsOneOf(filePath, ts.supportedJSExtensions) || ts.fileExtensionIs(filePath, ".d.ts" /* Dts */)) {
                    // Otherwise just check if sourceFile with the name exists
                    var filePathWithoutExtension = ts.removeFileExtension(filePath);
                    return !!getSourceFileByPath((filePathWithoutExtension + ".ts" /* Ts */)) ||
                        !!getSourceFileByPath((filePathWithoutExtension + ".tsx" /* Tsx */));
                }
                return false;
            }
            function isSameFile(file1, file2) {
                return ts.comparePaths(file1, file2, currentDirectory, !host.useCaseSensitiveFileNames()) === 0 /* EqualTo */;
            }
            function getSymlinkCache() {
                if (host.getSymlinkCache) {
                    return host.getSymlinkCache();
                }
                if (!symlinks) {
                    symlinks = ts.createSymlinkCache(currentDirectory, getCanonicalFileName);
                }
                if (files && resolvedTypeReferenceDirectives && !symlinks.hasProcessedResolutions()) {
                    symlinks.setSymlinksFromResolutions(files, resolvedTypeReferenceDirectives);
                }
                return symlinks;
            }
        }
        ts.createProgram = createProgram;
        function updateHostForUseSourceOfProjectReferenceRedirect(host) {
            var setOfDeclarationDirectories;
            var originalFileExists = host.compilerHost.fileExists;
            var originalDirectoryExists = host.compilerHost.directoryExists;
            var originalGetDirectories = host.compilerHost.getDirectories;
            var originalRealpath = host.compilerHost.realpath;
            if (!host.useSourceOfProjectReferenceRedirect)
                return { onProgramCreateComplete: ts.noop, fileExists: fileExists };
            host.compilerHost.fileExists = fileExists;
            var directoryExists;
            if (originalDirectoryExists) {
                // This implementation of directoryExists checks if the directory being requested is
                // directory of .d.ts file for the referenced Project.
                // If it is it returns true irrespective of whether that directory exists on host
                directoryExists = host.compilerHost.directoryExists = function (path) {
                    if (originalDirectoryExists.call(host.compilerHost, path)) {
                        handleDirectoryCouldBeSymlink(path);
                        return true;
                    }
                    if (!host.getResolvedProjectReferences())
                        return false;
                    if (!setOfDeclarationDirectories) {
                        setOfDeclarationDirectories = new ts.Set();
                        host.forEachResolvedProjectReference(function (ref) {
                            var out = ts.outFile(ref.commandLine.options);
                            if (out) {
                                setOfDeclarationDirectories.add(ts.getDirectoryPath(host.toPath(out)));
                            }
                            else {
                                // Set declaration's in different locations only, if they are next to source the directory present doesnt change
                                var declarationDir = ref.commandLine.options.declarationDir || ref.commandLine.options.outDir;
                                if (declarationDir) {
                                    setOfDeclarationDirectories.add(host.toPath(declarationDir));
                                }
                            }
                        });
                    }
                    return fileOrDirectoryExistsUsingSource(path, /*isFile*/ false);
                };
            }
            if (originalGetDirectories) {
                // Call getDirectories only if directory actually present on the host
                // This is needed to ensure that we arent getting directories that we fake about presence for
                host.compilerHost.getDirectories = function (path) {
                    return !host.getResolvedProjectReferences() || (originalDirectoryExists && originalDirectoryExists.call(host.compilerHost, path)) ?
                        originalGetDirectories.call(host.compilerHost, path) :
                        [];
                };
            }
            // This is something we keep for life time of the host
            if (originalRealpath) {
                host.compilerHost.realpath = function (s) {
                    var _a;
                    return ((_a = host.getSymlinkCache().getSymlinkedFiles()) === null || _a === void 0 ? void 0 : _a.get(host.toPath(s))) ||
                        originalRealpath.call(host.compilerHost, s);
                };
            }
            return { onProgramCreateComplete: onProgramCreateComplete, fileExists: fileExists, directoryExists: directoryExists };
            function onProgramCreateComplete() {
                host.compilerHost.fileExists = originalFileExists;
                host.compilerHost.directoryExists = originalDirectoryExists;
                host.compilerHost.getDirectories = originalGetDirectories;
                // DO not revert realpath as it could be used later
            }
            // This implementation of fileExists checks if the file being requested is
            // .d.ts file for the referenced Project.
            // If it is it returns true irrespective of whether that file exists on host
            function fileExists(file) {
                if (originalFileExists.call(host.compilerHost, file))
                    return true;
                if (!host.getResolvedProjectReferences())
                    return false;
                if (!ts.isDeclarationFileName(file))
                    return false;
                // Project references go to source file instead of .d.ts file
                return fileOrDirectoryExistsUsingSource(file, /*isFile*/ true);
            }
            function fileExistsIfProjectReferenceDts(file) {
                var source = host.getSourceOfProjectReferenceRedirect(file);
                return source !== undefined ?
                    ts.isString(source) ? originalFileExists.call(host.compilerHost, source) : true :
                    undefined;
            }
            function directoryExistsIfProjectReferenceDeclDir(dir) {
                var dirPath = host.toPath(dir);
                var dirPathWithTrailingDirectorySeparator = "" + dirPath + ts.directorySeparator;
                return ts.forEachKey(setOfDeclarationDirectories, function (declDirPath) {
                    return dirPath === declDirPath ||
                        // Any parent directory of declaration dir
                        ts.startsWith(declDirPath, dirPathWithTrailingDirectorySeparator) ||
                        // Any directory inside declaration dir
                        ts.startsWith(dirPath, declDirPath + "/");
                });
            }
            function handleDirectoryCouldBeSymlink(directory) {
                var _a;
                if (!host.getResolvedProjectReferences() || ts.containsIgnoredPath(directory))
                    return;
                // Because we already watch node_modules, handle symlinks in there
                if (!originalRealpath || !ts.stringContains(directory, ts.nodeModulesPathPart))
                    return;
                var symlinkCache = host.getSymlinkCache();
                var directoryPath = ts.ensureTrailingDirectorySeparator(host.toPath(directory));
                if ((_a = symlinkCache.getSymlinkedDirectories()) === null || _a === void 0 ? void 0 : _a.has(directoryPath))
                    return;
                var real = ts.normalizePath(originalRealpath.call(host.compilerHost, directory));
                var realPath;
                if (real === directory ||
                    (realPath = ts.ensureTrailingDirectorySeparator(host.toPath(real))) === directoryPath) {
                    // not symlinked
                    symlinkCache.setSymlinkedDirectory(directoryPath, false);
                    return;
                }
                symlinkCache.setSymlinkedDirectory(directory, {
                    real: ts.ensureTrailingDirectorySeparator(real),
                    realPath: realPath
                });
            }
            function fileOrDirectoryExistsUsingSource(fileOrDirectory, isFile) {
                var _a;
                var fileOrDirectoryExistsUsingSource = isFile ?
                    function (file) { return fileExistsIfProjectReferenceDts(file); } :
                    function (dir) { return directoryExistsIfProjectReferenceDeclDir(dir); };
                // Check current directory or file
                var result = fileOrDirectoryExistsUsingSource(fileOrDirectory);
                if (result !== undefined)
                    return result;
                var symlinkCache = host.getSymlinkCache();
                var symlinkedDirectories = symlinkCache.getSymlinkedDirectories();
                if (!symlinkedDirectories)
                    return false;
                var fileOrDirectoryPath = host.toPath(fileOrDirectory);
                if (!ts.stringContains(fileOrDirectoryPath, ts.nodeModulesPathPart))
                    return false;
                if (isFile && ((_a = symlinkCache.getSymlinkedFiles()) === null || _a === void 0 ? void 0 : _a.has(fileOrDirectoryPath)))
                    return true;
                // If it contains node_modules check if its one of the symlinked path we know of
                return ts.firstDefinedIterator(symlinkedDirectories.entries(), function (_a) {
                    var directoryPath = _a[0], symlinkedDirectory = _a[1];
                    if (!symlinkedDirectory || !ts.startsWith(fileOrDirectoryPath, directoryPath))
                        return undefined;
                    var result = fileOrDirectoryExistsUsingSource(fileOrDirectoryPath.replace(directoryPath, symlinkedDirectory.realPath));
                    if (isFile && result) {
                        // Store the real path for the file'
                        var absolutePath = ts.getNormalizedAbsolutePath(fileOrDirectory, host.compilerHost.getCurrentDirectory());
                        symlinkCache.setSymlinkedFile(fileOrDirectoryPath, "" + symlinkedDirectory.real + absolutePath.replace(new RegExp(directoryPath, "i"), ""));
                    }
                    return result;
                }) || false;
            }
        }
        /*@internal*/
        ts.emitSkippedWithNoDiagnostics = { diagnostics: ts.emptyArray, sourceMaps: undefined, emittedFiles: undefined, emitSkipped: true };
        /*@internal*/
        function handleNoEmitOptions(program, sourceFile, writeFile, cancellationToken) {
            var options = program.getCompilerOptions();
            if (options.noEmit) {
                // Cache the semantic diagnostics
                program.getSemanticDiagnostics(sourceFile, cancellationToken);
                return sourceFile || ts.outFile(options) ?
                    ts.emitSkippedWithNoDiagnostics :
                    program.emitBuildInfo(writeFile, cancellationToken);
            }
            // If the noEmitOnError flag is set, then check if we have any errors so far.  If so,
            // immediately bail out.  Note that we pass 'undefined' for 'sourceFile' so that we
            // get any preEmit diagnostics, not just the ones
            if (!options.noEmitOnError)
                return undefined;
            var diagnostics = __spreadArray(__spreadArray(__spreadArray(__spreadArray([], program.getOptionsDiagnostics(cancellationToken), true), program.getSyntacticDiagnostics(sourceFile, cancellationToken), true), program.getGlobalDiagnostics(cancellationToken), true), program.getSemanticDiagnostics(sourceFile, cancellationToken), true);
            if (diagnostics.length === 0 && ts.getEmitDeclarations(program.getCompilerOptions())) {
                diagnostics = program.getDeclarationDiagnostics(/*sourceFile*/ undefined, cancellationToken);
            }
            if (!diagnostics.length)
                return undefined;
            var emittedFiles;
            if (!sourceFile && !ts.outFile(options)) {
                var emitResult = program.emitBuildInfo(writeFile, cancellationToken);
                if (emitResult.diagnostics)
                    diagnostics = __spreadArray(__spreadArray([], diagnostics, true), emitResult.diagnostics, true);
                emittedFiles = emitResult.emittedFiles;
            }
            return { diagnostics: diagnostics, sourceMaps: undefined, emittedFiles: emittedFiles, emitSkipped: true };
        }
        ts.handleNoEmitOptions = handleNoEmitOptions;
        /*@internal*/
        function filterSemanticDiagnostics(diagnostic, option) {
            return ts.filter(diagnostic, function (d) { return !d.skippedOn || !option[d.skippedOn]; });
        }
        ts.filterSemanticDiagnostics = filterSemanticDiagnostics;
        /* @internal */
        function parseConfigHostFromCompilerHostLike(host, directoryStructureHost) {
            if (directoryStructureHost === void 0) {
                directoryStructureHost = host;
            }
            return {
                fileExists: function (f) { return directoryStructureHost.fileExists(f); },
                readDirectory: function (root, extensions, excludes, includes, depth) {
                    ts.Debug.assertIsDefined(directoryStructureHost.readDirectory, "'CompilerHost.readDirectory' must be implemented to correctly process 'projectReferences'");
                    return directoryStructureHost.readDirectory(root, extensions, excludes, includes, depth);
                },
                readFile: function (f) { return directoryStructureHost.readFile(f); },
                useCaseSensitiveFileNames: host.useCaseSensitiveFileNames(),
                getCurrentDirectory: function () { return host.getCurrentDirectory(); },
                onUnRecoverableConfigFileDiagnostic: host.onUnRecoverableConfigFileDiagnostic || ts.returnUndefined,
                trace: host.trace ? function (s) { return host.trace(s); } : undefined
            };
        }
        ts.parseConfigHostFromCompilerHostLike = parseConfigHostFromCompilerHostLike;
        /* @internal */
        function createPrependNodes(projectReferences, getCommandLine, readFile) {
            if (!projectReferences)
                return ts.emptyArray;
            var nodes;
            for (var i = 0; i < projectReferences.length; i++) {
                var ref = projectReferences[i];
                var resolvedRefOpts = getCommandLine(ref, i);
                if (ref.prepend && resolvedRefOpts && resolvedRefOpts.options) {
                    var out = ts.outFile(resolvedRefOpts.options);
                    // Upstream project didn't have outFile set -- skip (error will have been issued earlier)
                    if (!out)
                        continue;
                    var _a = ts.getOutputPathsForBundle(resolvedRefOpts.options, /*forceDtsPaths*/ true), jsFilePath = _a.jsFilePath, sourceMapFilePath = _a.sourceMapFilePath, declarationFilePath = _a.declarationFilePath, declarationMapPath = _a.declarationMapPath, buildInfoPath = _a.buildInfoPath;
                    var node = ts.createInputFiles(readFile, jsFilePath, sourceMapFilePath, declarationFilePath, declarationMapPath, buildInfoPath);
                    (nodes || (nodes = [])).push(node);
                }
            }
            return nodes || ts.emptyArray;
        }
        ts.createPrependNodes = createPrependNodes;
        function resolveProjectReferencePath(hostOrRef, ref) {
            var passedInRef = ref ? ref : hostOrRef;
            return ts.resolveConfigFileProjectName(passedInRef.path);
        }
        ts.resolveProjectReferencePath = resolveProjectReferencePath;
        /* @internal */
        /**
         * Returns a DiagnosticMessage if we won't include a resolved module due to its extension.
         * The DiagnosticMessage's parameters are the imported module name, and the filename it resolved to.
         * This returns a diagnostic even if the module will be an untyped module.
         */
        function getResolutionDiagnostic(options, _a) {
            var extension = _a.extension;
            switch (extension) {
                case ".ts" /* Ts */:
                case ".d.ts" /* Dts */:
                    // These are always allowed.
                    return undefined;
                case ".tsx" /* Tsx */:
                    return needJsx();
                case ".jsx" /* Jsx */:
                    return needJsx() || needAllowJs();
                case ".js" /* Js */:
                    return needAllowJs();
                case ".json" /* Json */:
                    return needResolveJsonModule();
            }
            function needJsx() {
                return options.jsx ? undefined : ts.Diagnostics.Module_0_was_resolved_to_1_but_jsx_is_not_set;
            }
            function needAllowJs() {
                return ts.getAllowJSCompilerOption(options) || !ts.getStrictOptionValue(options, "noImplicitAny") ? undefined : ts.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type;
            }
            function needResolveJsonModule() {
                return options.resolveJsonModule ? undefined : ts.Diagnostics.Module_0_was_resolved_to_1_but_resolveJsonModule_is_not_used;
            }
        }
        ts.getResolutionDiagnostic = getResolutionDiagnostic;
        function getModuleNames(_a) {
            var imports = _a.imports, moduleAugmentations = _a.moduleAugmentations;
            var res = imports.map(function (i) { return i.text; });
            for (var _i = 0, moduleAugmentations_1 = moduleAugmentations; _i < moduleAugmentations_1.length; _i++) {
                var aug = moduleAugmentations_1[_i];
                if (aug.kind === 10 /* StringLiteral */) {
                    res.push(aug.text);
                }
                // Do nothing if it's an Identifier; we don't need to do module resolution for `declare global`.
            }
            return res;
        }
        /* @internal */
        function getModuleNameStringLiteralAt(_a, index) {
            var imports = _a.imports, moduleAugmentations = _a.moduleAugmentations;
            if (index < imports.length)
                return imports[index];
            var augIndex = imports.length;
            for (var _i = 0, moduleAugmentations_2 = moduleAugmentations; _i < moduleAugmentations_2.length; _i++) {
                var aug = moduleAugmentations_2[_i];
                if (aug.kind === 10 /* StringLiteral */) {
                    if (index === augIndex)
                        return aug;
                    augIndex++;
                }
                // Do nothing if it's an Identifier; we don't need to do module resolution for `declare global`.
            }
            ts.Debug.fail("should never ask for module name at index higher than possible module name");
        }
        ts.getModuleNameStringLiteralAt = getModuleNameStringLiteralAt;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function getFileEmitOutput(program, sourceFile, emitOnlyDtsFiles, cancellationToken, customTransformers, forceDtsEmit) {
            var outputFiles = [];
            var _a = program.emit(sourceFile, writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers, forceDtsEmit), emitSkipped = _a.emitSkipped, diagnostics = _a.diagnostics, exportedModulesFromDeclarationEmit = _a.exportedModulesFromDeclarationEmit;
            return { outputFiles: outputFiles, emitSkipped: emitSkipped, diagnostics: diagnostics, exportedModulesFromDeclarationEmit: exportedModulesFromDeclarationEmit };
            function writeFile(fileName, text, writeByteOrderMark) {
                outputFiles.push({ name: fileName, writeByteOrderMark: writeByteOrderMark, text: text });
            }
        }
        ts.getFileEmitOutput = getFileEmitOutput;
        var BuilderState;
        (function (BuilderState) {
            var manyToManyPathMapCount = 0;
            function createManyToManyPathMap() {
                function create(forward, reverse, deleted) {
                    var version = 0;
                    var map = {
                        id: manyToManyPathMapCount++,
                        version: function () { return version; },
                        clone: function () { return create(new ts.Map(forward), new ts.Map(reverse), deleted && new ts.Set(deleted)); },
                        forEach: function (fn) { return forward.forEach(fn); },
                        getKeys: function (v) { return reverse.get(v); },
                        getValues: function (k) { return forward.get(k); },
                        hasKey: function (k) { return forward.has(k); },
                        keys: function () { return forward.keys(); },
                        deletedKeys: function () { return deleted; },
                        deleteKey: function (k) {
                            (deleted || (deleted = new ts.Set())).add(k);
                            var set = forward.get(k);
                            if (!set) {
                                return false;
                            }
                            set.forEach(function (v) { return deleteFromMultimap(reverse, v, k); });
                            forward.delete(k);
                            version++;
                            return true;
                        },
                        set: function (k, vSet) {
                            var changed = !!(deleted === null || deleted === void 0 ? void 0 : deleted.delete(k));
                            var existingVSet = forward.get(k);
                            forward.set(k, vSet);
                            existingVSet === null || existingVSet === void 0 ? void 0 : existingVSet.forEach(function (v) {
                                if (!vSet.has(v)) {
                                    changed = true;
                                    deleteFromMultimap(reverse, v, k);
                                }
                            });
                            vSet.forEach(function (v) {
                                if (!(existingVSet === null || existingVSet === void 0 ? void 0 : existingVSet.has(v))) {
                                    changed = true;
                                    addToMultimap(reverse, v, k);
                                }
                            });
                            if (changed) {
                                version++;
                            }
                            return map;
                        },
                    };
                    return map;
                }
                return create(new ts.Map(), new ts.Map(), /*deleted*/ undefined);
            }
            BuilderState.createManyToManyPathMap = createManyToManyPathMap;
            function addToMultimap(map, k, v) {
                var set = map.get(k);
                if (!set) {
                    set = new ts.Set();
                    map.set(k, set);
                }
                set.add(v);
            }
            function deleteFromMultimap(map, k, v, removeEmpty) {
                if (removeEmpty === void 0) {
                    removeEmpty = true;
                }
                var set = map.get(k);
                if (set === null || set === void 0 ? void 0 : set.delete(v)) {
                    if (removeEmpty && !set.size) {
                        map.delete(k);
                    }
                    return true;
                }
                return false;
            }
            function getReferencedFilesFromImportedModuleSymbol(symbol) {
                return ts.mapDefined(symbol.declarations, function (declaration) { var _a; return (_a = ts.getSourceFileOfNode(declaration)) === null || _a === void 0 ? void 0 : _a.resolvedPath; });
            }
            /**
             * Get the module source file and all augmenting files from the import name node from file
             */
            function getReferencedFilesFromImportLiteral(checker, importName) {
                var symbol = checker.getSymbolAtLocation(importName);
                return symbol && getReferencedFilesFromImportedModuleSymbol(symbol);
            }
            /**
             * Gets the path to reference file from file name, it could be resolvedPath if present otherwise path
             */
            function getReferencedFileFromFileName(program, fileName, sourceFileDirectory, getCanonicalFileName) {
                return ts.toPath(program.getProjectReferenceRedirect(fileName) || fileName, sourceFileDirectory, getCanonicalFileName);
            }
            /**
             * Gets the referenced files for a file from the program with values for the keys as referenced file's path to be true
             */
            function getReferencedFiles(program, sourceFile, getCanonicalFileName) {
                var referencedFiles;
                // We need to use a set here since the code can contain the same import twice,
                // but that will only be one dependency.
                // To avoid invernal conversion, the key of the referencedFiles map must be of type Path
                if (sourceFile.imports && sourceFile.imports.length > 0) {
                    var checker = program.getTypeChecker();
                    for (var _i = 0, _a = sourceFile.imports; _i < _a.length; _i++) {
                        var importName = _a[_i];
                        var declarationSourceFilePaths = getReferencedFilesFromImportLiteral(checker, importName);
                        declarationSourceFilePaths === null || declarationSourceFilePaths === void 0 ? void 0 : declarationSourceFilePaths.forEach(addReferencedFile);
                    }
                }
                var sourceFileDirectory = ts.getDirectoryPath(sourceFile.resolvedPath);
                // Handle triple slash references
                if (sourceFile.referencedFiles && sourceFile.referencedFiles.length > 0) {
                    for (var _b = 0, _c = sourceFile.referencedFiles; _b < _c.length; _b++) {
                        var referencedFile = _c[_b];
                        var referencedPath = getReferencedFileFromFileName(program, referencedFile.fileName, sourceFileDirectory, getCanonicalFileName);
                        addReferencedFile(referencedPath);
                    }
                }
                // Handle type reference directives
                if (sourceFile.resolvedTypeReferenceDirectiveNames) {
                    sourceFile.resolvedTypeReferenceDirectiveNames.forEach(function (resolvedTypeReferenceDirective) {
                        if (!resolvedTypeReferenceDirective) {
                            return;
                        }
                        var fileName = resolvedTypeReferenceDirective.resolvedFileName; // TODO: GH#18217
                        var typeFilePath = getReferencedFileFromFileName(program, fileName, sourceFileDirectory, getCanonicalFileName);
                        addReferencedFile(typeFilePath);
                    });
                }
                // Add module augmentation as references
                if (sourceFile.moduleAugmentations.length) {
                    var checker = program.getTypeChecker();
                    for (var _d = 0, _e = sourceFile.moduleAugmentations; _d < _e.length; _d++) {
                        var moduleName = _e[_d];
                        if (!ts.isStringLiteral(moduleName)) {
                            continue;
                        }
                        var symbol = checker.getSymbolAtLocation(moduleName);
                        if (!symbol) {
                            continue;
                        }
                        // Add any file other than our own as reference
                        addReferenceFromAmbientModule(symbol);
                    }
                }
                // From ambient modules
                for (var _f = 0, _g = program.getTypeChecker().getAmbientModules(); _f < _g.length; _f++) {
                    var ambientModule = _g[_f];
                    if (ambientModule.declarations && ambientModule.declarations.length > 1) {
                        addReferenceFromAmbientModule(ambientModule);
                    }
                }
                return referencedFiles;
                function addReferenceFromAmbientModule(symbol) {
                    if (!symbol.declarations) {
                        return;
                    }
                    // Add any file other than our own as reference
                    for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                        var declaration = _a[_i];
                        var declarationSourceFile = ts.getSourceFileOfNode(declaration);
                        if (declarationSourceFile &&
                            declarationSourceFile !== sourceFile) {
                            addReferencedFile(declarationSourceFile.resolvedPath);
                        }
                    }
                }
                function addReferencedFile(referencedPath) {
                    (referencedFiles || (referencedFiles = new ts.Set())).add(referencedPath);
                }
            }
            /**
             * Returns true if oldState is reusable, that is the emitKind = module/non module has not changed
             */
            function canReuseOldState(newReferencedMap, oldState) {
                return oldState && !oldState.referencedMap === !newReferencedMap;
            }
            BuilderState.canReuseOldState = canReuseOldState;
            /**
             * Creates the state of file references and signature for the new program from oldState if it is safe
             */
            function create(newProgram, getCanonicalFileName, oldState, disableUseFileVersionAsSignature) {
                var fileInfos = new ts.Map();
                var referencedMap = newProgram.getCompilerOptions().module !== ts.ModuleKind.None ? createManyToManyPathMap() : undefined;
                var exportedModulesMap = referencedMap ? createManyToManyPathMap() : undefined;
                var hasCalledUpdateShapeSignature = new ts.Set();
                var useOldState = canReuseOldState(referencedMap, oldState);
                // Ensure source files have parent pointers set
                newProgram.getTypeChecker();
                // Create the reference map, and set the file infos
                for (var _i = 0, _a = newProgram.getSourceFiles(); _i < _a.length; _i++) {
                    var sourceFile = _a[_i];
                    var version_2 = ts.Debug.checkDefined(sourceFile.version, "Program intended to be used with Builder should have source files with versions set");
                    var oldInfo = useOldState ? oldState.fileInfos.get(sourceFile.resolvedPath) : undefined;
                    if (referencedMap) {
                        var newReferences = getReferencedFiles(newProgram, sourceFile, getCanonicalFileName);
                        if (newReferences) {
                            referencedMap.set(sourceFile.resolvedPath, newReferences);
                        }
                        // Copy old visible to outside files map
                        if (useOldState) {
                            var exportedModules = oldState.exportedModulesMap.getValues(sourceFile.resolvedPath);
                            if (exportedModules) {
                                exportedModulesMap.set(sourceFile.resolvedPath, exportedModules);
                            }
                        }
                    }
                    fileInfos.set(sourceFile.resolvedPath, { version: version_2, signature: oldInfo && oldInfo.signature, affectsGlobalScope: isFileAffectingGlobalScope(sourceFile) || undefined });
                }
                return {
                    fileInfos: fileInfos,
                    referencedMap: referencedMap,
                    exportedModulesMap: exportedModulesMap,
                    hasCalledUpdateShapeSignature: hasCalledUpdateShapeSignature,
                    useFileVersionAsSignature: !disableUseFileVersionAsSignature && !useOldState
                };
            }
            BuilderState.create = create;
            /**
             * Releases needed properties
             */
            function releaseCache(state) {
                state.allFilesExcludingDefaultLibraryFile = undefined;
                state.allFileNames = undefined;
            }
            BuilderState.releaseCache = releaseCache;
            /**
             * Creates a clone of the state
             */
            function clone(state) {
                var _a, _b;
                // Dont need to backup allFiles info since its cache anyway
                return {
                    fileInfos: new ts.Map(state.fileInfos),
                    referencedMap: (_a = state.referencedMap) === null || _a === void 0 ? void 0 : _a.clone(),
                    exportedModulesMap: (_b = state.exportedModulesMap) === null || _b === void 0 ? void 0 : _b.clone(),
                    hasCalledUpdateShapeSignature: new ts.Set(state.hasCalledUpdateShapeSignature),
                    useFileVersionAsSignature: state.useFileVersionAsSignature,
                };
            }
            BuilderState.clone = clone;
            /**
             * Gets the files affected by the path from the program
             */
            function getFilesAffectedBy(state, programOfThisState, path, cancellationToken, computeHash, cacheToUpdateSignature, exportedModulesMapCache) {
                // Since the operation could be cancelled, the signatures are always stored in the cache
                // They will be committed once it is safe to use them
                // eg when calling this api from tsserver, if there is no cancellation of the operation
                // In the other cases the affected files signatures are committed only after the iteration through the result is complete
                var signatureCache = cacheToUpdateSignature || new ts.Map();
                var sourceFile = programOfThisState.getSourceFileByPath(path);
                if (!sourceFile) {
                    return ts.emptyArray;
                }
                if (!updateShapeSignature(state, programOfThisState, sourceFile, signatureCache, cancellationToken, computeHash, exportedModulesMapCache)) {
                    return [sourceFile];
                }
                var result = (state.referencedMap ? getFilesAffectedByUpdatedShapeWhenModuleEmit : getFilesAffectedByUpdatedShapeWhenNonModuleEmit)(state, programOfThisState, sourceFile, signatureCache, cancellationToken, computeHash, exportedModulesMapCache);
                if (!cacheToUpdateSignature) {
                    // Commit all the signatures in the signature cache
                    updateSignaturesFromCache(state, signatureCache);
                }
                return result;
            }
            BuilderState.getFilesAffectedBy = getFilesAffectedBy;
            /**
             * Updates the signatures from the cache into state's fileinfo signatures
             * This should be called whenever it is safe to commit the state of the builder
             */
            function updateSignaturesFromCache(state, signatureCache) {
                signatureCache.forEach(function (signature, path) { return updateSignatureOfFile(state, signature, path); });
            }
            BuilderState.updateSignaturesFromCache = updateSignaturesFromCache;
            function updateSignatureOfFile(state, signature, path) {
                state.fileInfos.get(path).signature = signature;
                state.hasCalledUpdateShapeSignature.add(path);
            }
            BuilderState.updateSignatureOfFile = updateSignatureOfFile;
            /**
             * Returns if the shape of the signature has changed since last emit
             */
            function updateShapeSignature(state, programOfThisState, sourceFile, cacheToUpdateSignature, cancellationToken, computeHash, exportedModulesMapCache, useFileVersionAsSignature) {
                if (useFileVersionAsSignature === void 0) {
                    useFileVersionAsSignature = state.useFileVersionAsSignature;
                }
                ts.Debug.assert(!!sourceFile);
                ts.Debug.assert(!exportedModulesMapCache || !!state.exportedModulesMap, "Compute visible to outside map only if visibleToOutsideReferencedMap present in the state");
                // If we have cached the result for this file, that means hence forth we should assume file shape is uptodate
                if (state.hasCalledUpdateShapeSignature.has(sourceFile.resolvedPath) || cacheToUpdateSignature.has(sourceFile.resolvedPath)) {
                    return false;
                }
                var info = state.fileInfos.get(sourceFile.resolvedPath);
                if (!info)
                    return ts.Debug.fail();
                var prevSignature = info.signature;
                var latestSignature;
                if (!sourceFile.isDeclarationFile && !useFileVersionAsSignature) {
                    var emitOutput_1 = getFileEmitOutput(programOfThisState, sourceFile, 
                    /*emitOnlyDtsFiles*/ true, cancellationToken, 
                    /*customTransformers*/ undefined, 
                    /*forceDtsEmit*/ true);
                    var firstDts_1 = ts.firstOrUndefined(emitOutput_1.outputFiles);
                    if (firstDts_1) {
                        ts.Debug.assert(ts.fileExtensionIs(firstDts_1.name, ".d.ts" /* Dts */), "File extension for signature expected to be dts", function () { return "Found: " + ts.getAnyExtensionFromPath(firstDts_1.name) + " for " + firstDts_1.name + ":: All output files: " + JSON.stringify(emitOutput_1.outputFiles.map(function (f) { return f.name; })); });
                        latestSignature = (computeHash || ts.generateDjb2Hash)(firstDts_1.text);
                        if (exportedModulesMapCache && latestSignature !== prevSignature) {
                            updateExportedModules(sourceFile, emitOutput_1.exportedModulesFromDeclarationEmit, exportedModulesMapCache);
                        }
                    }
                }
                // Default is to use file version as signature
                if (latestSignature === undefined) {
                    latestSignature = sourceFile.version;
                    if (exportedModulesMapCache && latestSignature !== prevSignature) {
                        // All the references in this file are exported
                        var references = state.referencedMap ? state.referencedMap.getValues(sourceFile.resolvedPath) : undefined;
                        if (references) {
                            exportedModulesMapCache.set(sourceFile.resolvedPath, references);
                        }
                        else {
                            exportedModulesMapCache.deleteKey(sourceFile.resolvedPath);
                        }
                    }
                }
                cacheToUpdateSignature.set(sourceFile.resolvedPath, latestSignature);
                return latestSignature !== prevSignature;
            }
            BuilderState.updateShapeSignature = updateShapeSignature;
            /**
             * Coverts the declaration emit result into exported modules map
             */
            function updateExportedModules(sourceFile, exportedModulesFromDeclarationEmit, exportedModulesMapCache) {
                if (!exportedModulesFromDeclarationEmit) {
                    exportedModulesMapCache.deleteKey(sourceFile.resolvedPath);
                    return;
                }
                var exportedModules;
                exportedModulesFromDeclarationEmit.forEach(function (symbol) { return addExportedModule(getReferencedFilesFromImportedModuleSymbol(symbol)); });
                if (exportedModules) {
                    exportedModulesMapCache.set(sourceFile.resolvedPath, exportedModules);
                }
                else {
                    exportedModulesMapCache.deleteKey(sourceFile.resolvedPath);
                }
                function addExportedModule(exportedModulePaths) {
                    if (exportedModulePaths === null || exportedModulePaths === void 0 ? void 0 : exportedModulePaths.length) {
                        if (!exportedModules) {
                            exportedModules = new ts.Set();
                        }
                        exportedModulePaths.forEach(function (path) { return exportedModules.add(path); });
                    }
                }
            }
            /**
             * Updates the exported modules from cache into state's exported modules map
             * This should be called whenever it is safe to commit the state of the builder
             */
            function updateExportedFilesMapFromCache(state, exportedModulesMapCache) {
                var _a;
                if (exportedModulesMapCache) {
                    ts.Debug.assert(!!state.exportedModulesMap);
                    var cacheId = exportedModulesMapCache.id;
                    var cacheVersion = exportedModulesMapCache.version();
                    if (state.previousCache) {
                        if (state.previousCache.id === cacheId && state.previousCache.version === cacheVersion) {
                            // If this is the same cache at the same version as last time this BuilderState
                            // was updated, there's no need to update again
                            return;
                        }
                        state.previousCache.id = cacheId;
                        state.previousCache.version = cacheVersion;
                    }
                    else {
                        state.previousCache = { id: cacheId, version: cacheVersion };
                    }
                    (_a = exportedModulesMapCache.deletedKeys()) === null || _a === void 0 ? void 0 : _a.forEach(function (path) { return state.exportedModulesMap.deleteKey(path); });
                    exportedModulesMapCache.forEach(function (exportedModules, path) { return state.exportedModulesMap.set(path, exportedModules); });
                }
            }
            BuilderState.updateExportedFilesMapFromCache = updateExportedFilesMapFromCache;
            /**
             * Get all the dependencies of the sourceFile
             */
            function getAllDependencies(state, programOfThisState, sourceFile) {
                var compilerOptions = programOfThisState.getCompilerOptions();
                // With --out or --outFile all outputs go into single file, all files depend on each other
                if (ts.outFile(compilerOptions)) {
                    return getAllFileNames(state, programOfThisState);
                }
                // If this is non module emit, or its a global file, it depends on all the source files
                if (!state.referencedMap || isFileAffectingGlobalScope(sourceFile)) {
                    return getAllFileNames(state, programOfThisState);
                }
                // Get the references, traversing deep from the referenceMap
                var seenMap = new ts.Set();
                var queue = [sourceFile.resolvedPath];
                while (queue.length) {
                    var path = queue.pop();
                    if (!seenMap.has(path)) {
                        seenMap.add(path);
                        var references = state.referencedMap.getValues(path);
                        if (references) {
                            var iterator = references.keys();
                            for (var iterResult = iterator.next(); !iterResult.done; iterResult = iterator.next()) {
                                queue.push(iterResult.value);
                            }
                        }
                    }
                }
                return ts.arrayFrom(ts.mapDefinedIterator(seenMap.keys(), function (path) { var _a, _b; return (_b = (_a = programOfThisState.getSourceFileByPath(path)) === null || _a === void 0 ? void 0 : _a.fileName) !== null && _b !== void 0 ? _b : path; }));
            }
            BuilderState.getAllDependencies = getAllDependencies;
            /**
             * Gets the names of all files from the program
             */
            function getAllFileNames(state, programOfThisState) {
                if (!state.allFileNames) {
                    var sourceFiles = programOfThisState.getSourceFiles();
                    state.allFileNames = sourceFiles === ts.emptyArray ? ts.emptyArray : sourceFiles.map(function (file) { return file.fileName; });
                }
                return state.allFileNames;
            }
            /**
             * Gets the files referenced by the the file path
             */
            function getReferencedByPaths(state, referencedFilePath) {
                var keys = state.referencedMap.getKeys(referencedFilePath);
                return keys ? ts.arrayFrom(keys.keys()) : [];
            }
            BuilderState.getReferencedByPaths = getReferencedByPaths;
            /**
             * For script files that contains only ambient external modules, although they are not actually external module files,
             * they can only be consumed via importing elements from them. Regular script files cannot consume them. Therefore,
             * there are no point to rebuild all script files if these special files have changed. However, if any statement
             * in the file is not ambient external module, we treat it as a regular script file.
             */
            function containsOnlyAmbientModules(sourceFile) {
                for (var _i = 0, _a = sourceFile.statements; _i < _a.length; _i++) {
                    var statement = _a[_i];
                    if (!ts.isModuleWithStringLiteralName(statement)) {
                        return false;
                    }
                }
                return true;
            }
            /**
             * Return true if file contains anything that augments to global scope we need to build them as if
             * they are global files as well as module
             */
            function containsGlobalScopeAugmentation(sourceFile) {
                return ts.some(sourceFile.moduleAugmentations, function (augmentation) { return ts.isGlobalScopeAugmentation(augmentation.parent); });
            }
            /**
             * Return true if the file will invalidate all files because it affectes global scope
             */
            function isFileAffectingGlobalScope(sourceFile) {
                return containsGlobalScopeAugmentation(sourceFile) ||
                    !ts.isExternalOrCommonJsModule(sourceFile) && !ts.isJsonSourceFile(sourceFile) && !containsOnlyAmbientModules(sourceFile);
            }
            /**
             * Gets all files of the program excluding the default library file
             */
            function getAllFilesExcludingDefaultLibraryFile(state, programOfThisState, firstSourceFile) {
                // Use cached result
                if (state.allFilesExcludingDefaultLibraryFile) {
                    return state.allFilesExcludingDefaultLibraryFile;
                }
                var result;
                if (firstSourceFile)
                    addSourceFile(firstSourceFile);
                for (var _i = 0, _a = programOfThisState.getSourceFiles(); _i < _a.length; _i++) {
                    var sourceFile = _a[_i];
                    if (sourceFile !== firstSourceFile) {
                        addSourceFile(sourceFile);
                    }
                }
                state.allFilesExcludingDefaultLibraryFile = result || ts.emptyArray;
                return state.allFilesExcludingDefaultLibraryFile;
                function addSourceFile(sourceFile) {
                    if (!programOfThisState.isSourceFileDefaultLibrary(sourceFile)) {
                        (result || (result = [])).push(sourceFile);
                    }
                }
            }
            BuilderState.getAllFilesExcludingDefaultLibraryFile = getAllFilesExcludingDefaultLibraryFile;
            /**
             * When program emits non modular code, gets the files affected by the sourceFile whose shape has changed
             */
            function getFilesAffectedByUpdatedShapeWhenNonModuleEmit(state, programOfThisState, sourceFileWithUpdatedShape) {
                var compilerOptions = programOfThisState.getCompilerOptions();
                // If `--out` or `--outFile` is specified, any new emit will result in re-emitting the entire project,
                // so returning the file itself is good enough.
                if (compilerOptions && ts.outFile(compilerOptions)) {
                    return [sourceFileWithUpdatedShape];
                }
                return getAllFilesExcludingDefaultLibraryFile(state, programOfThisState, sourceFileWithUpdatedShape);
            }
            /**
             * When program emits modular code, gets the files affected by the sourceFile whose shape has changed
             */
            function getFilesAffectedByUpdatedShapeWhenModuleEmit(state, programOfThisState, sourceFileWithUpdatedShape, cacheToUpdateSignature, cancellationToken, computeHash, exportedModulesMapCache) {
                if (isFileAffectingGlobalScope(sourceFileWithUpdatedShape)) {
                    return getAllFilesExcludingDefaultLibraryFile(state, programOfThisState, sourceFileWithUpdatedShape);
                }
                var compilerOptions = programOfThisState.getCompilerOptions();
                if (compilerOptions && (compilerOptions.isolatedModules || ts.outFile(compilerOptions))) {
                    return [sourceFileWithUpdatedShape];
                }
                // Now we need to if each file in the referencedBy list has a shape change as well.
                // Because if so, its own referencedBy files need to be saved as well to make the
                // emitting result consistent with files on disk.
                var seenFileNamesMap = new ts.Map();
                // Start with the paths this file was referenced by
                seenFileNamesMap.set(sourceFileWithUpdatedShape.resolvedPath, sourceFileWithUpdatedShape);
                var queue = getReferencedByPaths(state, sourceFileWithUpdatedShape.resolvedPath);
                while (queue.length > 0) {
                    var currentPath = queue.pop();
                    if (!seenFileNamesMap.has(currentPath)) {
                        var currentSourceFile = programOfThisState.getSourceFileByPath(currentPath);
                        seenFileNamesMap.set(currentPath, currentSourceFile);
                        if (currentSourceFile && updateShapeSignature(state, programOfThisState, currentSourceFile, cacheToUpdateSignature, cancellationToken, computeHash, exportedModulesMapCache)) {
                            queue.push.apply(queue, getReferencedByPaths(state, currentSourceFile.resolvedPath));
                        }
                    }
                }
                // Return array of values that needs emit
                return ts.arrayFrom(ts.mapDefinedIterator(seenFileNamesMap.values(), function (value) { return value; }));
            }
        })(BuilderState = ts.BuilderState || (ts.BuilderState = {}));
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var BuilderFileEmit;
        (function (BuilderFileEmit) {
            BuilderFileEmit[BuilderFileEmit["DtsOnly"] = 0] = "DtsOnly";
            BuilderFileEmit[BuilderFileEmit["Full"] = 1] = "Full";
        })(BuilderFileEmit = ts.BuilderFileEmit || (ts.BuilderFileEmit = {}));
        function hasSameKeys(map1, map2) {
            // Has same size and every key is present in both maps
            return map1 === map2 || map1 !== undefined && map2 !== undefined && map1.size === map2.size && !ts.forEachKey(map1, function (key) { return !map2.has(key); });
        }
        /**
         * Create the state so that we can iterate on changedFiles/affected files
         */
        function createBuilderProgramState(newProgram, getCanonicalFileName, oldState, disableUseFileVersionAsSignature) {
            var state = ts.BuilderState.create(newProgram, getCanonicalFileName, oldState, disableUseFileVersionAsSignature);
            state.program = newProgram;
            var compilerOptions = newProgram.getCompilerOptions();
            state.compilerOptions = compilerOptions;
            // With --out or --outFile, any change affects all semantic diagnostics so no need to cache them
            if (!ts.outFile(compilerOptions)) {
                state.semanticDiagnosticsPerFile = new ts.Map();
            }
            state.changedFilesSet = new ts.Set();
            var useOldState = ts.BuilderState.canReuseOldState(state.referencedMap, oldState);
            var oldCompilerOptions = useOldState ? oldState.compilerOptions : undefined;
            var canCopySemanticDiagnostics = useOldState && oldState.semanticDiagnosticsPerFile && !!state.semanticDiagnosticsPerFile &&
                !ts.compilerOptionsAffectSemanticDiagnostics(compilerOptions, oldCompilerOptions);
            if (useOldState) {
                // Verify the sanity of old state
                if (!oldState.currentChangedFilePath) {
                    var affectedSignatures = oldState.currentAffectedFilesSignatures;
                    ts.Debug.assert(!oldState.affectedFiles && (!affectedSignatures || !affectedSignatures.size), "Cannot reuse if only few affected files of currentChangedFile were iterated");
                }
                var changedFilesSet = oldState.changedFilesSet;
                if (canCopySemanticDiagnostics) {
                    ts.Debug.assert(!changedFilesSet || !ts.forEachKey(changedFilesSet, function (path) { return oldState.semanticDiagnosticsPerFile.has(path); }), "Semantic diagnostics shouldnt be available for changed files");
                }
                // Copy old state's changed files set
                changedFilesSet === null || changedFilesSet === void 0 ? void 0 : changedFilesSet.forEach(function (value) { return state.changedFilesSet.add(value); });
                if (!ts.outFile(compilerOptions) && oldState.affectedFilesPendingEmit) {
                    state.affectedFilesPendingEmit = oldState.affectedFilesPendingEmit.slice();
                    state.affectedFilesPendingEmitKind = oldState.affectedFilesPendingEmitKind && new ts.Map(oldState.affectedFilesPendingEmitKind);
                    state.affectedFilesPendingEmitIndex = oldState.affectedFilesPendingEmitIndex;
                    state.seenAffectedFiles = new ts.Set();
                }
            }
            // Update changed files and copy semantic diagnostics if we can
            var referencedMap = state.referencedMap;
            var oldReferencedMap = useOldState ? oldState.referencedMap : undefined;
            var copyDeclarationFileDiagnostics = canCopySemanticDiagnostics && !compilerOptions.skipLibCheck === !oldCompilerOptions.skipLibCheck;
            var copyLibFileDiagnostics = copyDeclarationFileDiagnostics && !compilerOptions.skipDefaultLibCheck === !oldCompilerOptions.skipDefaultLibCheck;
            state.fileInfos.forEach(function (info, sourceFilePath) {
                var oldInfo;
                var newReferences;
                // if not using old state, every file is changed
                if (!useOldState ||
                    // File wasn't present in old state
                    !(oldInfo = oldState.fileInfos.get(sourceFilePath)) ||
                    // versions dont match
                    oldInfo.version !== info.version ||
                    // Referenced files changed
                    !hasSameKeys(newReferences = referencedMap && referencedMap.getValues(sourceFilePath), oldReferencedMap && oldReferencedMap.getValues(sourceFilePath)) ||
                    // Referenced file was deleted in the new program
                    newReferences && ts.forEachKey(newReferences, function (path) { return !state.fileInfos.has(path) && oldState.fileInfos.has(path); })) {
                    // Register file as changed file and do not copy semantic diagnostics, since all changed files need to be re-evaluated
                    state.changedFilesSet.add(sourceFilePath);
                }
                else if (canCopySemanticDiagnostics) {
                    var sourceFile = newProgram.getSourceFileByPath(sourceFilePath);
                    if (sourceFile.isDeclarationFile && !copyDeclarationFileDiagnostics) {
                        return;
                    }
                    if (sourceFile.hasNoDefaultLib && !copyLibFileDiagnostics) {
                        return;
                    }
                    // Unchanged file copy diagnostics
                    var diagnostics = oldState.semanticDiagnosticsPerFile.get(sourceFilePath);
                    if (diagnostics) {
                        state.semanticDiagnosticsPerFile.set(sourceFilePath, oldState.hasReusableDiagnostic ? convertToDiagnostics(diagnostics, newProgram, getCanonicalFileName) : diagnostics);
                        if (!state.semanticDiagnosticsFromOldState) {
                            state.semanticDiagnosticsFromOldState = new ts.Set();
                        }
                        state.semanticDiagnosticsFromOldState.add(sourceFilePath);
                    }
                }
            });
            // If the global file is removed, add all files as changed
            if (useOldState && ts.forEachEntry(oldState.fileInfos, function (info, sourceFilePath) { return info.affectsGlobalScope && !state.fileInfos.has(sourceFilePath); })) {
                ts.BuilderState.getAllFilesExcludingDefaultLibraryFile(state, newProgram, /*firstSourceFile*/ undefined)
                    .forEach(function (file) { return state.changedFilesSet.add(file.resolvedPath); });
            }
            else if (oldCompilerOptions && !ts.outFile(compilerOptions) && ts.compilerOptionsAffectEmit(compilerOptions, oldCompilerOptions)) {
                // Add all files to affectedFilesPendingEmit since emit changed
                newProgram.getSourceFiles().forEach(function (f) { return addToAffectedFilesPendingEmit(state, f.resolvedPath, 1 /* Full */); });
                ts.Debug.assert(!state.seenAffectedFiles || !state.seenAffectedFiles.size);
                state.seenAffectedFiles = state.seenAffectedFiles || new ts.Set();
            }
            state.buildInfoEmitPending = !!state.changedFilesSet.size;
            return state;
        }
        function convertToDiagnostics(diagnostics, newProgram, getCanonicalFileName) {
            if (!diagnostics.length)
                return ts.emptyArray;
            var buildInfoDirectory = ts.getDirectoryPath(ts.getNormalizedAbsolutePath(ts.getTsBuildInfoEmitOutputFilePath(newProgram.getCompilerOptions()), newProgram.getCurrentDirectory()));
            return diagnostics.map(function (diagnostic) {
                var result = convertToDiagnosticRelatedInformation(diagnostic, newProgram, toPath);
                result.reportsUnnecessary = diagnostic.reportsUnnecessary;
                result.reportsDeprecated = diagnostic.reportDeprecated;
                result.source = diagnostic.source;
                result.skippedOn = diagnostic.skippedOn;
                var relatedInformation = diagnostic.relatedInformation;
                result.relatedInformation = relatedInformation ?
                    relatedInformation.length ?
                        relatedInformation.map(function (r) { return convertToDiagnosticRelatedInformation(r, newProgram, toPath); }) :
                        [] :
                    undefined;
                return result;
            });
            function toPath(path) {
                return ts.toPath(path, buildInfoDirectory, getCanonicalFileName);
            }
        }
        function convertToDiagnosticRelatedInformation(diagnostic, newProgram, toPath) {
            var file = diagnostic.file;
            return __assign(__assign({}, diagnostic), { file: file ? newProgram.getSourceFileByPath(toPath(file)) : undefined });
        }
        /**
         * Releases program and other related not needed properties
         */
        function releaseCache(state) {
            ts.BuilderState.releaseCache(state);
            state.program = undefined;
        }
        /**
         * Creates a clone of the state
         */
        function cloneBuilderProgramState(state) {
            var _a;
            var newState = ts.BuilderState.clone(state);
            newState.semanticDiagnosticsPerFile = state.semanticDiagnosticsPerFile && new ts.Map(state.semanticDiagnosticsPerFile);
            newState.changedFilesSet = new ts.Set(state.changedFilesSet);
            newState.affectedFiles = state.affectedFiles;
            newState.affectedFilesIndex = state.affectedFilesIndex;
            newState.currentChangedFilePath = state.currentChangedFilePath;
            newState.currentAffectedFilesSignatures = state.currentAffectedFilesSignatures && new ts.Map(state.currentAffectedFilesSignatures);
            newState.currentAffectedFilesExportedModulesMap = (_a = state.currentAffectedFilesExportedModulesMap) === null || _a === void 0 ? void 0 : _a.clone();
            newState.seenAffectedFiles = state.seenAffectedFiles && new ts.Set(state.seenAffectedFiles);
            newState.cleanedDiagnosticsOfLibFiles = state.cleanedDiagnosticsOfLibFiles;
            newState.semanticDiagnosticsFromOldState = state.semanticDiagnosticsFromOldState && new ts.Set(state.semanticDiagnosticsFromOldState);
            newState.program = state.program;
            newState.compilerOptions = state.compilerOptions;
            newState.affectedFilesPendingEmit = state.affectedFilesPendingEmit && state.affectedFilesPendingEmit.slice();
            newState.affectedFilesPendingEmitKind = state.affectedFilesPendingEmitKind && new ts.Map(state.affectedFilesPendingEmitKind);
            newState.affectedFilesPendingEmitIndex = state.affectedFilesPendingEmitIndex;
            newState.seenEmittedFiles = state.seenEmittedFiles && new ts.Map(state.seenEmittedFiles);
            newState.programEmitComplete = state.programEmitComplete;
            return newState;
        }
        /**
         * Verifies that source file is ok to be used in calls that arent handled by next
         */
        function assertSourceFileOkWithoutNextAffectedCall(state, sourceFile) {
            ts.Debug.assert(!sourceFile || !state.affectedFiles || state.affectedFiles[state.affectedFilesIndex - 1] !== sourceFile || !state.semanticDiagnosticsPerFile.has(sourceFile.resolvedPath));
        }
        /**
         * This function returns the next affected file to be processed.
         * Note that until doneAffected is called it would keep reporting same result
         * This is to allow the callers to be able to actually remove affected file only when the operation is complete
         * eg. if during diagnostics check cancellation token ends up cancelling the request, the affected file should be retained
         */
        function getNextAffectedFile(state, cancellationToken, computeHash) {
            while (true) {
                var affectedFiles = state.affectedFiles;
                if (affectedFiles) {
                    var seenAffectedFiles = state.seenAffectedFiles;
                    var affectedFilesIndex = state.affectedFilesIndex; // TODO: GH#18217
                    while (affectedFilesIndex < affectedFiles.length) {
                        var affectedFile = affectedFiles[affectedFilesIndex];
                        if (!seenAffectedFiles.has(affectedFile.resolvedPath)) {
                            // Set the next affected file as seen and remove the cached semantic diagnostics
                            state.affectedFilesIndex = affectedFilesIndex;
                            handleDtsMayChangeOfAffectedFile(state, affectedFile, cancellationToken, computeHash);
                            return affectedFile;
                        }
                        affectedFilesIndex++;
                    }
                    // Remove the changed file from the change set
                    state.changedFilesSet.delete(state.currentChangedFilePath);
                    state.currentChangedFilePath = undefined;
                    // Commit the changes in file signature
                    ts.BuilderState.updateSignaturesFromCache(state, state.currentAffectedFilesSignatures);
                    state.currentAffectedFilesSignatures.clear();
                    ts.BuilderState.updateExportedFilesMapFromCache(state, state.currentAffectedFilesExportedModulesMap);
                    state.affectedFiles = undefined;
                }
                // Get next changed file
                var nextKey = state.changedFilesSet.keys().next();
                if (nextKey.done) {
                    // Done
                    return undefined;
                }
                // With --out or --outFile all outputs go into single file
                // so operations are performed directly on program, return program
                var program = ts.Debug.checkDefined(state.program);
                var compilerOptions = program.getCompilerOptions();
                if (ts.outFile(compilerOptions)) {
                    ts.Debug.assert(!state.semanticDiagnosticsPerFile);
                    return program;
                }
                // Get next batch of affected files
                if (!state.currentAffectedFilesSignatures)
                    state.currentAffectedFilesSignatures = new ts.Map();
                if (state.exportedModulesMap) {
                    state.currentAffectedFilesExportedModulesMap || (state.currentAffectedFilesExportedModulesMap = ts.BuilderState.createManyToManyPathMap());
                }
                state.affectedFiles = ts.BuilderState.getFilesAffectedBy(state, program, nextKey.value, cancellationToken, computeHash, state.currentAffectedFilesSignatures, state.currentAffectedFilesExportedModulesMap);
                state.currentChangedFilePath = nextKey.value;
                state.affectedFilesIndex = 0;
                if (!state.seenAffectedFiles)
                    state.seenAffectedFiles = new ts.Set();
            }
        }
        /**
         * Returns next file to be emitted from files that retrieved semantic diagnostics but did not emit yet
         */
        function getNextAffectedFilePendingEmit(state) {
            var affectedFilesPendingEmit = state.affectedFilesPendingEmit;
            if (affectedFilesPendingEmit) {
                var seenEmittedFiles = (state.seenEmittedFiles || (state.seenEmittedFiles = new ts.Map()));
                for (var i = state.affectedFilesPendingEmitIndex; i < affectedFilesPendingEmit.length; i++) {
                    var affectedFile = ts.Debug.checkDefined(state.program).getSourceFileByPath(affectedFilesPendingEmit[i]);
                    if (affectedFile) {
                        var seenKind = seenEmittedFiles.get(affectedFile.resolvedPath);
                        var emitKind = ts.Debug.checkDefined(ts.Debug.checkDefined(state.affectedFilesPendingEmitKind).get(affectedFile.resolvedPath));
                        if (seenKind === undefined || seenKind < emitKind) {
                            // emit this file
                            state.affectedFilesPendingEmitIndex = i;
                            return { affectedFile: affectedFile, emitKind: emitKind };
                        }
                    }
                }
                state.affectedFilesPendingEmit = undefined;
                state.affectedFilesPendingEmitKind = undefined;
                state.affectedFilesPendingEmitIndex = undefined;
            }
            return undefined;
        }
        /**
         *  Handles semantic diagnostics and dts emit for affectedFile and files, that are referencing modules that export entities from affected file
         *  This is because even though js emit doesnt change, dts emit / type used can change resulting in need for dts emit and js change
         */
        function handleDtsMayChangeOfAffectedFile(state, affectedFile, cancellationToken, computeHash) {
            var _a;
            removeSemanticDiagnosticsOf(state, affectedFile.resolvedPath);
            // If affected files is everything except default library, then nothing more to do
            if (state.allFilesExcludingDefaultLibraryFile === state.affectedFiles) {
                if (!state.cleanedDiagnosticsOfLibFiles) {
                    state.cleanedDiagnosticsOfLibFiles = true;
                    var program_1 = ts.Debug.checkDefined(state.program);
                    var options_2 = program_1.getCompilerOptions();
                    ts.forEach(program_1.getSourceFiles(), function (f) {
                        return program_1.isSourceFileDefaultLibrary(f) &&
                            !ts.skipTypeChecking(f, options_2, program_1) &&
                            removeSemanticDiagnosticsOf(state, f.resolvedPath);
                    });
                }
                // When a change affects the global scope, all files are considered to be affected without updating their signature
                // That means when affected file is handled, its signature can be out of date
                // To avoid this, ensure that we update the signature for any affected file in this scenario.
                ts.BuilderState.updateShapeSignature(state, ts.Debug.checkDefined(state.program), affectedFile, ts.Debug.checkDefined(state.currentAffectedFilesSignatures), cancellationToken, computeHash, state.currentAffectedFilesExportedModulesMap);
                return;
            }
            else {
                ts.Debug.assert(state.hasCalledUpdateShapeSignature.has(affectedFile.resolvedPath) || ((_a = state.currentAffectedFilesSignatures) === null || _a === void 0 ? void 0 : _a.has(affectedFile.resolvedPath)), "Signature not updated for affected file: " + affectedFile.fileName);
            }
            if (!state.compilerOptions.assumeChangesOnlyAffectDirectDependencies) {
                forEachReferencingModulesOfExportOfAffectedFile(state, affectedFile, function (state, path) { return handleDtsMayChangeOf(state, path, cancellationToken, computeHash); });
            }
        }
        /**
         * Handle the dts may change, so they need to be added to pending emit if dts emit is enabled,
         * Also we need to make sure signature is updated for these files
         */
        function handleDtsMayChangeOf(state, path, cancellationToken, computeHash) {
            removeSemanticDiagnosticsOf(state, path);
            if (!state.changedFilesSet.has(path)) {
                var program = ts.Debug.checkDefined(state.program);
                var sourceFile = program.getSourceFileByPath(path);
                if (sourceFile) {
                    // Even though the js emit doesnt change and we are already handling dts emit and semantic diagnostics
                    // we need to update the signature to reflect correctness of the signature(which is output d.ts emit) of this file
                    // This ensures that we dont later during incremental builds considering wrong signature.
                    // Eg where this also is needed to ensure that .tsbuildinfo generated by incremental build should be same as if it was first fresh build
                    // But we avoid expensive full shape computation, as using file version as shape is enough for correctness.
                    ts.BuilderState.updateShapeSignature(state, program, sourceFile, ts.Debug.checkDefined(state.currentAffectedFilesSignatures), cancellationToken, computeHash, state.currentAffectedFilesExportedModulesMap, 
                    /* useFileVersionAsSignature */ true);
                    // If not dts emit, nothing more to do
                    if (ts.getEmitDeclarations(state.compilerOptions)) {
                        addToAffectedFilesPendingEmit(state, path, 0 /* DtsOnly */);
                    }
                }
            }
        }
        /**
         * Removes semantic diagnostics for path and
         * returns true if there are no more semantic diagnostics from the old state
         */
        function removeSemanticDiagnosticsOf(state, path) {
            if (!state.semanticDiagnosticsFromOldState) {
                return true;
            }
            state.semanticDiagnosticsFromOldState.delete(path);
            state.semanticDiagnosticsPerFile.delete(path);
            return !state.semanticDiagnosticsFromOldState.size;
        }
        function isChangedSignature(state, path) {
            var newSignature = ts.Debug.checkDefined(state.currentAffectedFilesSignatures).get(path);
            var oldSignature = ts.Debug.checkDefined(state.fileInfos.get(path)).signature;
            return newSignature !== oldSignature;
        }
        /**
         * Iterate on referencing modules that export entities from affected file
         */
        function forEachReferencingModulesOfExportOfAffectedFile(state, affectedFile, fn) {
            var _a, _b;
            // If there was change in signature (dts output) for the changed file,
            // then only we need to handle pending file emit
            if (!state.exportedModulesMap || !state.changedFilesSet.has(affectedFile.resolvedPath)) {
                return;
            }
            if (!isChangedSignature(state, affectedFile.resolvedPath))
                return;
            // Since isolated modules dont change js files, files affected by change in signature is itself
            // But we need to cleanup semantic diagnostics and queue dts emit for affected files
            if (state.compilerOptions.isolatedModules) {
                var seenFileNamesMap = new ts.Map();
                seenFileNamesMap.set(affectedFile.resolvedPath, true);
                var queue = ts.BuilderState.getReferencedByPaths(state, affectedFile.resolvedPath);
                while (queue.length > 0) {
                    var currentPath = queue.pop();
                    if (!seenFileNamesMap.has(currentPath)) {
                        seenFileNamesMap.set(currentPath, true);
                        fn(state, currentPath);
                        if (isChangedSignature(state, currentPath)) {
                            var currentSourceFile = ts.Debug.checkDefined(state.program).getSourceFileByPath(currentPath);
                            queue.push.apply(queue, ts.BuilderState.getReferencedByPaths(state, currentSourceFile.resolvedPath));
                        }
                    }
                }
            }
            ts.Debug.assert(!!state.currentAffectedFilesExportedModulesMap);
            var seenFileAndExportsOfFile = new ts.Set();
            // Go through exported modules from cache first
            // If exported modules has path, all files referencing file exported from are affected
            (_a = state.currentAffectedFilesExportedModulesMap.getKeys(affectedFile.resolvedPath)) === null || _a === void 0 ? void 0 : _a.forEach(function (exportedFromPath) {
                return forEachFilesReferencingPath(state, exportedFromPath, seenFileAndExportsOfFile, fn);
            });
            // If exported from path is not from cache and exported modules has path, all files referencing file exported from are affected
            (_b = state.exportedModulesMap.getKeys(affectedFile.resolvedPath)) === null || _b === void 0 ? void 0 : _b.forEach(function (exportedFromPath) {
                var _a;
                // If the cache had an updated value, skip
                return !state.currentAffectedFilesExportedModulesMap.hasKey(exportedFromPath) &&
                    !((_a = state.currentAffectedFilesExportedModulesMap.deletedKeys()) === null || _a === void 0 ? void 0 : _a.has(exportedFromPath)) &&
                    forEachFilesReferencingPath(state, exportedFromPath, seenFileAndExportsOfFile, fn);
            });
        }
        /**
         * Iterate on files referencing referencedPath
         */
        function forEachFilesReferencingPath(state, referencedPath, seenFileAndExportsOfFile, fn) {
            var _a;
            (_a = state.referencedMap.getKeys(referencedPath)) === null || _a === void 0 ? void 0 : _a.forEach(function (filePath) {
                return forEachFileAndExportsOfFile(state, filePath, seenFileAndExportsOfFile, fn);
            });
        }
        /**
         * fn on file and iterate on anything that exports this file
         */
        function forEachFileAndExportsOfFile(state, filePath, seenFileAndExportsOfFile, fn) {
            var _a, _b, _c;
            if (!ts.tryAddToSet(seenFileAndExportsOfFile, filePath)) {
                return;
            }
            fn(state, filePath);
            ts.Debug.assert(!!state.currentAffectedFilesExportedModulesMap);
            // Go through exported modules from cache first
            // If exported modules has path, all files referencing file exported from are affected
            (_a = state.currentAffectedFilesExportedModulesMap.getKeys(filePath)) === null || _a === void 0 ? void 0 : _a.forEach(function (exportedFromPath) {
                return forEachFileAndExportsOfFile(state, exportedFromPath, seenFileAndExportsOfFile, fn);
            });
            // If exported from path is not from cache and exported modules has path, all files referencing file exported from are affected
            (_b = state.exportedModulesMap.getKeys(filePath)) === null || _b === void 0 ? void 0 : _b.forEach(function (exportedFromPath) {
                var _a;
                // If the cache had an updated value, skip
                return !state.currentAffectedFilesExportedModulesMap.hasKey(exportedFromPath) &&
                    !((_a = state.currentAffectedFilesExportedModulesMap.deletedKeys()) === null || _a === void 0 ? void 0 : _a.has(exportedFromPath)) &&
                    forEachFileAndExportsOfFile(state, exportedFromPath, seenFileAndExportsOfFile, fn);
            });
            // Remove diagnostics of files that import this file (without going to exports of referencing files)
            (_c = state.referencedMap.getKeys(filePath)) === null || _c === void 0 ? void 0 : _c.forEach(function (referencingFilePath) {
                return !seenFileAndExportsOfFile.has(referencingFilePath) && // Not already removed diagnostic file
                    fn(state, referencingFilePath);
            } // Dont add to seen since this is not yet done with the export removal
            );
        }
        /**
         * This is called after completing operation on the next affected file.
         * The operations here are postponed to ensure that cancellation during the iteration is handled correctly
         */
        function doneWithAffectedFile(state, affected, emitKind, isPendingEmit, isBuildInfoEmit) {
            if (isBuildInfoEmit) {
                state.buildInfoEmitPending = false;
            }
            else if (affected === state.program) {
                state.changedFilesSet.clear();
                state.programEmitComplete = true;
            }
            else {
                state.seenAffectedFiles.add(affected.resolvedPath);
                if (emitKind !== undefined) {
                    (state.seenEmittedFiles || (state.seenEmittedFiles = new ts.Map())).set(affected.resolvedPath, emitKind);
                }
                if (isPendingEmit) {
                    state.affectedFilesPendingEmitIndex++;
                    state.buildInfoEmitPending = true;
                }
                else {
                    state.affectedFilesIndex++;
                }
            }
        }
        /**
         * Returns the result with affected file
         */
        function toAffectedFileResult(state, result, affected) {
            doneWithAffectedFile(state, affected);
            return { result: result, affected: affected };
        }
        /**
         * Returns the result with affected file
         */
        function toAffectedFileEmitResult(state, result, affected, emitKind, isPendingEmit, isBuildInfoEmit) {
            doneWithAffectedFile(state, affected, emitKind, isPendingEmit, isBuildInfoEmit);
            return { result: result, affected: affected };
        }
        /**
         * Gets semantic diagnostics for the file which are
         * bindAndCheckDiagnostics (from cache) and program diagnostics
         */
        function getSemanticDiagnosticsOfFile(state, sourceFile, cancellationToken) {
            return ts.concatenate(getBinderAndCheckerDiagnosticsOfFile(state, sourceFile, cancellationToken), ts.Debug.checkDefined(state.program).getProgramDiagnostics(sourceFile));
        }
        /**
         * Gets the binder and checker diagnostics either from cache if present, or otherwise from program and caches it
         * Note that it is assumed that when asked about binder and checker diagnostics, the file has been taken out of affected files/changed file set
         */
        function getBinderAndCheckerDiagnosticsOfFile(state, sourceFile, cancellationToken) {
            var path = sourceFile.resolvedPath;
            if (state.semanticDiagnosticsPerFile) {
                var cachedDiagnostics = state.semanticDiagnosticsPerFile.get(path);
                // Report the bind and check diagnostics from the cache if we already have those diagnostics present
                if (cachedDiagnostics) {
                    return ts.filterSemanticDiagnostics(cachedDiagnostics, state.compilerOptions);
                }
            }
            // Diagnostics werent cached, get them from program, and cache the result
            var diagnostics = ts.Debug.checkDefined(state.program).getBindAndCheckDiagnostics(sourceFile, cancellationToken);
            if (state.semanticDiagnosticsPerFile) {
                state.semanticDiagnosticsPerFile.set(path, diagnostics);
            }
            return ts.filterSemanticDiagnostics(diagnostics, state.compilerOptions);
        }
        /**
         * Gets the program information to be emitted in buildInfo so that we can use it to create new program
         */
        function getProgramBuildInfo(state, getCanonicalFileName) {
            if (ts.outFile(state.compilerOptions))
                return undefined;
            var currentDirectory = ts.Debug.checkDefined(state.program).getCurrentDirectory();
            var buildInfoDirectory = ts.getDirectoryPath(ts.getNormalizedAbsolutePath(ts.getTsBuildInfoEmitOutputFilePath(state.compilerOptions), currentDirectory));
            var fileNames = [];
            var fileNameToFileId = new ts.Map();
            var fileIdsList;
            var fileNamesToFileIdListId;
            var fileInfos = ts.arrayFrom(state.fileInfos.entries(), function (_a) {
                var key = _a[0], value = _a[1];
                // Ensure fileId
                var fileId = toFileId(key);
                ts.Debug.assert(fileNames[fileId - 1] === relativeToBuildInfo(key));
                var signature = state.currentAffectedFilesSignatures && state.currentAffectedFilesSignatures.get(key);
                var actualSignature = signature !== null && signature !== void 0 ? signature : value.signature;
                return value.version === actualSignature ?
                    value.affectsGlobalScope ?
                        { version: value.version, signature: undefined, affectsGlobalScope: true } :
                        value.version :
                    actualSignature !== undefined ?
                        signature === undefined ?
                            value :
                            { version: value.version, signature: signature, affectsGlobalScope: value.affectsGlobalScope } :
                        { version: value.version, signature: false, affectsGlobalScope: value.affectsGlobalScope };
            });
            var referencedMap;
            if (state.referencedMap) {
                referencedMap = ts.arrayFrom(state.referencedMap.keys()).sort(ts.compareStringsCaseSensitive).map(function (key) {
                    return [
                        toFileId(key),
                        toFileIdListId(state.referencedMap.getValues(key))
                    ];
                });
            }
            var exportedModulesMap;
            if (state.exportedModulesMap) {
                exportedModulesMap = ts.mapDefined(ts.arrayFrom(state.exportedModulesMap.keys()).sort(ts.compareStringsCaseSensitive), function (key) {
                    var _a;
                    if (state.currentAffectedFilesExportedModulesMap) {
                        if ((_a = state.currentAffectedFilesExportedModulesMap.deletedKeys()) === null || _a === void 0 ? void 0 : _a.has(key)) {
                            return undefined;
                        }
                        var newValue = state.currentAffectedFilesExportedModulesMap.getValues(key);
                        if (newValue) {
                            return [toFileId(key), toFileIdListId(newValue)];
                        }
                    }
                    // Not in temporary cache, use existing value
                    return [toFileId(key), toFileIdListId(state.exportedModulesMap.getValues(key))];
                });
            }
            var semanticDiagnosticsPerFile;
            if (state.semanticDiagnosticsPerFile) {
                for (var _i = 0, _a = ts.arrayFrom(state.semanticDiagnosticsPerFile.keys()).sort(ts.compareStringsCaseSensitive); _i < _a.length; _i++) {
                    var key = _a[_i];
                    var value = state.semanticDiagnosticsPerFile.get(key);
                    (semanticDiagnosticsPerFile || (semanticDiagnosticsPerFile = [])).push(value.length ?
                        [
                            toFileId(key),
                            state.hasReusableDiagnostic ?
                                value :
                                convertToReusableDiagnostics(value, relativeToBuildInfo)
                        ] :
                        toFileId(key));
                }
            }
            var affectedFilesPendingEmit;
            if (state.affectedFilesPendingEmit) {
                var seenFiles = new ts.Set();
                for (var _b = 0, _c = state.affectedFilesPendingEmit.slice(state.affectedFilesPendingEmitIndex).sort(ts.compareStringsCaseSensitive); _b < _c.length; _b++) {
                    var path = _c[_b];
                    if (ts.tryAddToSet(seenFiles, path)) {
                        (affectedFilesPendingEmit || (affectedFilesPendingEmit = [])).push([toFileId(path), state.affectedFilesPendingEmitKind.get(path)]);
                    }
                }
            }
            return {
                fileNames: fileNames,
                fileInfos: fileInfos,
                options: convertToProgramBuildInfoCompilerOptions(state.compilerOptions, relativeToBuildInfoEnsuringAbsolutePath),
                fileIdsList: fileIdsList,
                referencedMap: referencedMap,
                exportedModulesMap: exportedModulesMap,
                semanticDiagnosticsPerFile: semanticDiagnosticsPerFile,
                affectedFilesPendingEmit: affectedFilesPendingEmit,
            };
            function relativeToBuildInfoEnsuringAbsolutePath(path) {
                return relativeToBuildInfo(ts.getNormalizedAbsolutePath(path, currentDirectory));
            }
            function relativeToBuildInfo(path) {
                return ts.ensurePathIsNonModuleName(ts.getRelativePathFromDirectory(buildInfoDirectory, path, getCanonicalFileName));
            }
            function toFileId(path) {
                var fileId = fileNameToFileId.get(path);
                if (fileId === undefined) {
                    fileNames.push(relativeToBuildInfo(path));
                    fileNameToFileId.set(path, fileId = fileNames.length);
                }
                return fileId;
            }
            function toFileIdListId(set) {
                var fileIds = ts.arrayFrom(set.keys(), toFileId).sort(ts.compareValues);
                var key = fileIds.join();
                var fileIdListId = fileNamesToFileIdListId === null || fileNamesToFileIdListId === void 0 ? void 0 : fileNamesToFileIdListId.get(key);
                if (fileIdListId === undefined) {
                    (fileIdsList || (fileIdsList = [])).push(fileIds);
                    (fileNamesToFileIdListId || (fileNamesToFileIdListId = new ts.Map())).set(key, fileIdListId = fileIdsList.length);
                }
                return fileIdListId;
            }
        }
        function convertToProgramBuildInfoCompilerOptions(options, relativeToBuildInfo) {
            var result;
            var optionsNameMap = ts.getOptionsNameMap().optionsNameMap;
            for (var _i = 0, _a = ts.getOwnKeys(options).sort(ts.compareStringsCaseSensitive); _i < _a.length; _i++) {
                var name = _a[_i];
                var optionKey = name.toLowerCase();
                var optionInfo = optionsNameMap.get(optionKey);
                if ((optionInfo === null || optionInfo === void 0 ? void 0 : optionInfo.affectsEmit) || (optionInfo === null || optionInfo === void 0 ? void 0 : optionInfo.affectsSemanticDiagnostics) ||
                    // We need to store `strict`, even though it won't be examined directly, so that the
                    // flags it controls (e.g. `strictNullChecks`) will be retrieved correctly from the buildinfo
                    optionKey === "strict" ||
                    // We need to store these to determine whether `lib` files need to be rechecked.
                    optionKey === "skiplibcheck" || optionKey === "skipdefaultlibcheck") {
                    (result || (result = {}))[name] = convertToReusableCompilerOptionValue(optionInfo, options[name], relativeToBuildInfo);
                }
            }
            return result;
        }
        function convertToReusableCompilerOptionValue(option, value, relativeToBuildInfo) {
            if (option) {
                if (option.type === "list") {
                    var values = value;
                    if (option.element.isFilePath && values.length) {
                        return values.map(relativeToBuildInfo);
                    }
                }
                else if (option.isFilePath) {
                    return relativeToBuildInfo(value);
                }
            }
            return value;
        }
        function convertToReusableDiagnostics(diagnostics, relativeToBuildInfo) {
            ts.Debug.assert(!!diagnostics.length);
            return diagnostics.map(function (diagnostic) {
                var result = convertToReusableDiagnosticRelatedInformation(diagnostic, relativeToBuildInfo);
                result.reportsUnnecessary = diagnostic.reportsUnnecessary;
                result.reportDeprecated = diagnostic.reportsDeprecated;
                result.source = diagnostic.source;
                result.skippedOn = diagnostic.skippedOn;
                var relatedInformation = diagnostic.relatedInformation;
                result.relatedInformation = relatedInformation ?
                    relatedInformation.length ?
                        relatedInformation.map(function (r) { return convertToReusableDiagnosticRelatedInformation(r, relativeToBuildInfo); }) :
                        [] :
                    undefined;
                return result;
            });
        }
        function convertToReusableDiagnosticRelatedInformation(diagnostic, relativeToBuildInfo) {
            var file = diagnostic.file;
            return __assign(__assign({}, diagnostic), { file: file ? relativeToBuildInfo(file.resolvedPath) : undefined });
        }
        var BuilderProgramKind;
        (function (BuilderProgramKind) {
            BuilderProgramKind[BuilderProgramKind["SemanticDiagnosticsBuilderProgram"] = 0] = "SemanticDiagnosticsBuilderProgram";
            BuilderProgramKind[BuilderProgramKind["EmitAndSemanticDiagnosticsBuilderProgram"] = 1] = "EmitAndSemanticDiagnosticsBuilderProgram";
        })(BuilderProgramKind = ts.BuilderProgramKind || (ts.BuilderProgramKind = {}));
        function getBuilderCreationParameters(newProgramOrRootNames, hostOrOptions, oldProgramOrHost, configFileParsingDiagnosticsOrOldProgram, configFileParsingDiagnostics, projectReferences) {
            var host;
            var newProgram;
            var oldProgram;
            if (newProgramOrRootNames === undefined) {
                ts.Debug.assert(hostOrOptions === undefined);
                host = oldProgramOrHost;
                oldProgram = configFileParsingDiagnosticsOrOldProgram;
                ts.Debug.assert(!!oldProgram);
                newProgram = oldProgram.getProgram();
            }
            else if (ts.isArray(newProgramOrRootNames)) {
                oldProgram = configFileParsingDiagnosticsOrOldProgram;
                newProgram = ts.createProgram({
                    rootNames: newProgramOrRootNames,
                    options: hostOrOptions,
                    host: oldProgramOrHost,
                    oldProgram: oldProgram && oldProgram.getProgramOrUndefined(),
                    configFileParsingDiagnostics: configFileParsingDiagnostics,
                    projectReferences: projectReferences
                });
                host = oldProgramOrHost;
            }
            else {
                newProgram = newProgramOrRootNames;
                host = hostOrOptions;
                oldProgram = oldProgramOrHost;
                configFileParsingDiagnostics = configFileParsingDiagnosticsOrOldProgram;
            }
            return { host: host, newProgram: newProgram, oldProgram: oldProgram, configFileParsingDiagnostics: configFileParsingDiagnostics || ts.emptyArray };
        }
        ts.getBuilderCreationParameters = getBuilderCreationParameters;
        function createBuilderProgram(kind, _a) {
            var newProgram = _a.newProgram, host = _a.host, oldProgram = _a.oldProgram, configFileParsingDiagnostics = _a.configFileParsingDiagnostics;
            // Return same program if underlying program doesnt change
            var oldState = oldProgram && oldProgram.getState();
            if (oldState && newProgram === oldState.program && configFileParsingDiagnostics === newProgram.getConfigFileParsingDiagnostics()) {
                newProgram = undefined; // TODO: GH#18217
                oldState = undefined;
                return oldProgram;
            }
            /**
             * Create the canonical file name for identity
             */
            var getCanonicalFileName = ts.createGetCanonicalFileName(host.useCaseSensitiveFileNames());
            /**
             * Computing hash to for signature verification
             */
            var computeHash = ts.maybeBind(host, host.createHash);
            var state = createBuilderProgramState(newProgram, getCanonicalFileName, oldState, host.disableUseFileVersionAsSignature);
            var backupState;
            newProgram.getProgramBuildInfo = function () { return getProgramBuildInfo(state, getCanonicalFileName); };
            // To ensure that we arent storing any references to old program or new program without state
            newProgram = undefined; // TODO: GH#18217
            oldProgram = undefined;
            oldState = undefined;
            var getState = function () { return state; };
            var builderProgram = createRedirectedBuilderProgram(getState, configFileParsingDiagnostics);
            builderProgram.getState = getState;
            builderProgram.backupState = function () {
                ts.Debug.assert(backupState === undefined);
                backupState = cloneBuilderProgramState(state);
            };
            builderProgram.restoreState = function () {
                state = ts.Debug.checkDefined(backupState);
                backupState = undefined;
            };
            builderProgram.getAllDependencies = function (sourceFile) { return ts.BuilderState.getAllDependencies(state, ts.Debug.checkDefined(state.program), sourceFile); };
            builderProgram.getSemanticDiagnostics = getSemanticDiagnostics;
            builderProgram.emit = emit;
            builderProgram.releaseProgram = function () {
                releaseCache(state);
                backupState = undefined;
            };
            if (kind === BuilderProgramKind.SemanticDiagnosticsBuilderProgram) {
                builderProgram.getSemanticDiagnosticsOfNextAffectedFile = getSemanticDiagnosticsOfNextAffectedFile;
            }
            else if (kind === BuilderProgramKind.EmitAndSemanticDiagnosticsBuilderProgram) {
                builderProgram.getSemanticDiagnosticsOfNextAffectedFile = getSemanticDiagnosticsOfNextAffectedFile;
                builderProgram.emitNextAffectedFile = emitNextAffectedFile;
                builderProgram.emitBuildInfo = emitBuildInfo;
            }
            else {
                ts.notImplemented();
            }
            return builderProgram;
            function emitBuildInfo(writeFile, cancellationToken) {
                if (state.buildInfoEmitPending) {
                    var result = ts.Debug.checkDefined(state.program).emitBuildInfo(writeFile || ts.maybeBind(host, host.writeFile), cancellationToken);
                    state.buildInfoEmitPending = false;
                    return result;
                }
                return ts.emitSkippedWithNoDiagnostics;
            }
            /**
             * Emits the next affected file's emit result (EmitResult and sourceFiles emitted) or returns undefined if iteration is complete
             * The first of writeFile if provided, writeFile of BuilderProgramHost if provided, writeFile of compiler host
             * in that order would be used to write the files
             */
            function emitNextAffectedFile(writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers) {
                var affected = getNextAffectedFile(state, cancellationToken, computeHash);
                var emitKind = 1 /* Full */;
                var isPendingEmitFile = false;
                if (!affected) {
                    if (!ts.outFile(state.compilerOptions)) {
                        var pendingAffectedFile = getNextAffectedFilePendingEmit(state);
                        if (!pendingAffectedFile) {
                            if (!state.buildInfoEmitPending) {
                                return undefined;
                            }
                            var affected_1 = ts.Debug.checkDefined(state.program);
                            return toAffectedFileEmitResult(state, 
                            // When whole program is affected, do emit only once (eg when --out or --outFile is specified)
                            // Otherwise just affected file
                            affected_1.emitBuildInfo(writeFile || ts.maybeBind(host, host.writeFile), cancellationToken), affected_1, 1 /* Full */, 
                            /*isPendingEmitFile*/ false, 
                            /*isBuildInfoEmit*/ true);
                        }
                        (affected = pendingAffectedFile.affectedFile, emitKind = pendingAffectedFile.emitKind);
                        isPendingEmitFile = true;
                    }
                    else {
                        var program = ts.Debug.checkDefined(state.program);
                        if (state.programEmitComplete)
                            return undefined;
                        affected = program;
                    }
                }
                return toAffectedFileEmitResult(state, 
                // When whole program is affected, do emit only once (eg when --out or --outFile is specified)
                // Otherwise just affected file
                ts.Debug.checkDefined(state.program).emit(affected === state.program ? undefined : affected, writeFile || ts.maybeBind(host, host.writeFile), cancellationToken, emitOnlyDtsFiles || emitKind === 0 /* DtsOnly */, customTransformers), affected, emitKind, isPendingEmitFile);
            }
            /**
             * Emits the JavaScript and declaration files.
             * When targetSource file is specified, emits the files corresponding to that source file,
             * otherwise for the whole program.
             * In case of EmitAndSemanticDiagnosticsBuilderProgram, when targetSourceFile is specified,
             * it is assumed that that file is handled from affected file list. If targetSourceFile is not specified,
             * it will only emit all the affected files instead of whole program
             *
             * The first of writeFile if provided, writeFile of BuilderProgramHost if provided, writeFile of compiler host
             * in that order would be used to write the files
             */
            function emit(targetSourceFile, writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers) {
                var restorePendingEmitOnHandlingNoEmitSuccess = false;
                var savedAffectedFilesPendingEmit;
                var savedAffectedFilesPendingEmitKind;
                var savedAffectedFilesPendingEmitIndex;
                // Backup and restore affected pendings emit state for non emit Builder if noEmitOnError is enabled and emitBuildInfo could be written in case there are errors
                // This ensures pending files to emit is updated in tsbuildinfo
                // Note that when there are no errors, emit proceeds as if everything is emitted as it is callers reponsibility to write the files to disk if at all (because its builder that doesnt track files to emit)
                if (kind !== BuilderProgramKind.EmitAndSemanticDiagnosticsBuilderProgram &&
                    !targetSourceFile &&
                    !ts.outFile(state.compilerOptions) &&
                    !state.compilerOptions.noEmit &&
                    state.compilerOptions.noEmitOnError) {
                    restorePendingEmitOnHandlingNoEmitSuccess = true;
                    savedAffectedFilesPendingEmit = state.affectedFilesPendingEmit && state.affectedFilesPendingEmit.slice();
                    savedAffectedFilesPendingEmitKind = state.affectedFilesPendingEmitKind && new ts.Map(state.affectedFilesPendingEmitKind);
                    savedAffectedFilesPendingEmitIndex = state.affectedFilesPendingEmitIndex;
                }
                if (kind === BuilderProgramKind.EmitAndSemanticDiagnosticsBuilderProgram) {
                    assertSourceFileOkWithoutNextAffectedCall(state, targetSourceFile);
                }
                var result = ts.handleNoEmitOptions(builderProgram, targetSourceFile, writeFile, cancellationToken);
                if (result)
                    return result;
                if (restorePendingEmitOnHandlingNoEmitSuccess) {
                    state.affectedFilesPendingEmit = savedAffectedFilesPendingEmit;
                    state.affectedFilesPendingEmitKind = savedAffectedFilesPendingEmitKind;
                    state.affectedFilesPendingEmitIndex = savedAffectedFilesPendingEmitIndex;
                }
                // Emit only affected files if using builder for emit
                if (!targetSourceFile && kind === BuilderProgramKind.EmitAndSemanticDiagnosticsBuilderProgram) {
                    // Emit and report any errors we ran into.
                    var sourceMaps = [];
                    var emitSkipped = false;
                    var diagnostics = void 0;
                    var emittedFiles = [];
                    var affectedEmitResult = void 0;
                    while (affectedEmitResult = emitNextAffectedFile(writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers)) {
                        emitSkipped = emitSkipped || affectedEmitResult.result.emitSkipped;
                        diagnostics = ts.addRange(diagnostics, affectedEmitResult.result.diagnostics);
                        emittedFiles = ts.addRange(emittedFiles, affectedEmitResult.result.emittedFiles);
                        sourceMaps = ts.addRange(sourceMaps, affectedEmitResult.result.sourceMaps);
                    }
                    return {
                        emitSkipped: emitSkipped,
                        diagnostics: diagnostics || ts.emptyArray,
                        emittedFiles: emittedFiles,
                        sourceMaps: sourceMaps
                    };
                }
                return ts.Debug.checkDefined(state.program).emit(targetSourceFile, writeFile || ts.maybeBind(host, host.writeFile), cancellationToken, emitOnlyDtsFiles, customTransformers);
            }
            /**
             * Return the semantic diagnostics for the next affected file or undefined if iteration is complete
             * If provided ignoreSourceFile would be called before getting the diagnostics and would ignore the sourceFile if the returned value was true
             */
            function getSemanticDiagnosticsOfNextAffectedFile(cancellationToken, ignoreSourceFile) {
                while (true) {
                    var affected = getNextAffectedFile(state, cancellationToken, computeHash);
                    if (!affected) {
                        // Done
                        return undefined;
                    }
                    else if (affected === state.program) {
                        // When whole program is affected, get all semantic diagnostics (eg when --out or --outFile is specified)
                        return toAffectedFileResult(state, state.program.getSemanticDiagnostics(/*targetSourceFile*/ undefined, cancellationToken), affected);
                    }
                    // Add file to affected file pending emit to handle for later emit time
                    // Apart for emit builder do this for tsbuildinfo, do this for non emit builder when noEmit is set as tsbuildinfo is written and reused between emitters
                    if (kind === BuilderProgramKind.EmitAndSemanticDiagnosticsBuilderProgram || state.compilerOptions.noEmit || state.compilerOptions.noEmitOnError) {
                        addToAffectedFilesPendingEmit(state, affected.resolvedPath, 1 /* Full */);
                    }
                    // Get diagnostics for the affected file if its not ignored
                    if (ignoreSourceFile && ignoreSourceFile(affected)) {
                        // Get next affected file
                        doneWithAffectedFile(state, affected);
                        continue;
                    }
                    return toAffectedFileResult(state, getSemanticDiagnosticsOfFile(state, affected, cancellationToken), affected);
                }
            }
            /**
             * Gets the semantic diagnostics from the program corresponding to this state of file (if provided) or whole program
             * The semantic diagnostics are cached and managed here
             * Note that it is assumed that when asked about semantic diagnostics through this API,
             * the file has been taken out of affected files so it is safe to use cache or get from program and cache the diagnostics
             * In case of SemanticDiagnosticsBuilderProgram if the source file is not provided,
             * it will iterate through all the affected files, to ensure that cache stays valid and yet provide a way to get all semantic diagnostics
             */
            function getSemanticDiagnostics(sourceFile, cancellationToken) {
                assertSourceFileOkWithoutNextAffectedCall(state, sourceFile);
                var compilerOptions = ts.Debug.checkDefined(state.program).getCompilerOptions();
                if (ts.outFile(compilerOptions)) {
                    ts.Debug.assert(!state.semanticDiagnosticsPerFile);
                    // We dont need to cache the diagnostics just return them from program
                    return ts.Debug.checkDefined(state.program).getSemanticDiagnostics(sourceFile, cancellationToken);
                }
                if (sourceFile) {
                    return getSemanticDiagnosticsOfFile(state, sourceFile, cancellationToken);
                }
                // When semantic builder asks for diagnostics of the whole program,
                // ensure that all the affected files are handled
                // eslint-disable-next-line no-empty
                while (getSemanticDiagnosticsOfNextAffectedFile(cancellationToken)) {
                }
                var diagnostics;
                for (var _i = 0, _a = ts.Debug.checkDefined(state.program).getSourceFiles(); _i < _a.length; _i++) {
                    var sourceFile_1 = _a[_i];
                    diagnostics = ts.addRange(diagnostics, getSemanticDiagnosticsOfFile(state, sourceFile_1, cancellationToken));
                }
                return diagnostics || ts.emptyArray;
            }
        }
        ts.createBuilderProgram = createBuilderProgram;
        function addToAffectedFilesPendingEmit(state, affectedFilePendingEmit, kind) {
            if (!state.affectedFilesPendingEmit)
                state.affectedFilesPendingEmit = [];
            if (!state.affectedFilesPendingEmitKind)
                state.affectedFilesPendingEmitKind = new ts.Map();
            var existingKind = state.affectedFilesPendingEmitKind.get(affectedFilePendingEmit);
            state.affectedFilesPendingEmit.push(affectedFilePendingEmit);
            state.affectedFilesPendingEmitKind.set(affectedFilePendingEmit, existingKind || kind);
            // affectedFilesPendingEmitIndex === undefined
            // - means the emit state.affectedFilesPendingEmit was undefined before adding current affected files
            //   so start from 0 as array would be affectedFilesPendingEmit
            // else, continue to iterate from existing index, the current set is appended to existing files
            if (state.affectedFilesPendingEmitIndex === undefined) {
                state.affectedFilesPendingEmitIndex = 0;
            }
        }
        function toBuilderStateFileInfo(fileInfo) {
            return ts.isString(fileInfo) ?
                { version: fileInfo, signature: fileInfo, affectsGlobalScope: undefined } :
                ts.isString(fileInfo.signature) ?
                    fileInfo :
                    { version: fileInfo.version, signature: fileInfo.signature === false ? undefined : fileInfo.version, affectsGlobalScope: fileInfo.affectsGlobalScope };
        }
        ts.toBuilderStateFileInfo = toBuilderStateFileInfo;
        function createBuildProgramUsingProgramBuildInfo(program, buildInfoPath, host) {
            var _a;
            var buildInfoDirectory = ts.getDirectoryPath(ts.getNormalizedAbsolutePath(buildInfoPath, host.getCurrentDirectory()));
            var getCanonicalFileName = ts.createGetCanonicalFileName(host.useCaseSensitiveFileNames());
            var filePaths = program.fileNames.map(toPath);
            var filePathsSetList = (_a = program.fileIdsList) === null || _a === void 0 ? void 0 : _a.map(function (fileIds) { return new ts.Set(fileIds.map(toFilePath)); });
            var fileInfos = new ts.Map();
            program.fileInfos.forEach(function (fileInfo, index) { return fileInfos.set(toFilePath(index + 1), toBuilderStateFileInfo(fileInfo)); });
            var state = {
                fileInfos: fileInfos,
                compilerOptions: program.options ? ts.convertToOptionsWithAbsolutePaths(program.options, toAbsolutePath) : {},
                referencedMap: toManyToManyPathMap(program.referencedMap),
                exportedModulesMap: toManyToManyPathMap(program.exportedModulesMap),
                semanticDiagnosticsPerFile: program.semanticDiagnosticsPerFile && ts.arrayToMap(program.semanticDiagnosticsPerFile, function (value) { return toFilePath(ts.isNumber(value) ? value : value[0]); }, function (value) { return ts.isNumber(value) ? ts.emptyArray : value[1]; }),
                hasReusableDiagnostic: true,
                affectedFilesPendingEmit: ts.map(program.affectedFilesPendingEmit, function (value) { return toFilePath(value[0]); }),
                affectedFilesPendingEmitKind: program.affectedFilesPendingEmit && ts.arrayToMap(program.affectedFilesPendingEmit, function (value) { return toFilePath(value[0]); }, function (value) { return value[1]; }),
                affectedFilesPendingEmitIndex: program.affectedFilesPendingEmit && 0,
            };
            return {
                getState: function () { return state; },
                backupState: ts.noop,
                restoreState: ts.noop,
                getProgram: ts.notImplemented,
                getProgramOrUndefined: ts.returnUndefined,
                releaseProgram: ts.noop,
                getCompilerOptions: function () { return state.compilerOptions; },
                getSourceFile: ts.notImplemented,
                getSourceFiles: ts.notImplemented,
                getOptionsDiagnostics: ts.notImplemented,
                getGlobalDiagnostics: ts.notImplemented,
                getConfigFileParsingDiagnostics: ts.notImplemented,
                getSyntacticDiagnostics: ts.notImplemented,
                getDeclarationDiagnostics: ts.notImplemented,
                getSemanticDiagnostics: ts.notImplemented,
                emit: ts.notImplemented,
                getAllDependencies: ts.notImplemented,
                getCurrentDirectory: ts.notImplemented,
                emitNextAffectedFile: ts.notImplemented,
                getSemanticDiagnosticsOfNextAffectedFile: ts.notImplemented,
                emitBuildInfo: ts.notImplemented,
                close: ts.noop,
            };
            function toPath(path) {
                return ts.toPath(path, buildInfoDirectory, getCanonicalFileName);
            }
            function toAbsolutePath(path) {
                return ts.getNormalizedAbsolutePath(path, buildInfoDirectory);
            }
            function toFilePath(fileId) {
                return filePaths[fileId - 1];
            }
            function toFilePathsSet(fileIdsListId) {
                return filePathsSetList[fileIdsListId - 1];
            }
            function toManyToManyPathMap(referenceMap) {
                if (!referenceMap) {
                    return undefined;
                }
                var map = ts.BuilderState.createManyToManyPathMap();
                referenceMap.forEach(function (_a) {
                    var fileId = _a[0], fileIdListId = _a[1];
                    return map.set(toFilePath(fileId), toFilePathsSet(fileIdListId));
                });
                return map;
            }
        }
        ts.createBuildProgramUsingProgramBuildInfo = createBuildProgramUsingProgramBuildInfo;
        function createRedirectedBuilderProgram(getState, configFileParsingDiagnostics) {
            return {
                getState: ts.notImplemented,
                backupState: ts.noop,
                restoreState: ts.noop,
                getProgram: getProgram,
                getProgramOrUndefined: function () { return getState().program; },
                releaseProgram: function () { return getState().program = undefined; },
                getCompilerOptions: function () { return getState().compilerOptions; },
                getSourceFile: function (fileName) { return getProgram().getSourceFile(fileName); },
                getSourceFiles: function () { return getProgram().getSourceFiles(); },
                getOptionsDiagnostics: function (cancellationToken) { return getProgram().getOptionsDiagnostics(cancellationToken); },
                getGlobalDiagnostics: function (cancellationToken) { return getProgram().getGlobalDiagnostics(cancellationToken); },
                getConfigFileParsingDiagnostics: function () { return configFileParsingDiagnostics; },
                getSyntacticDiagnostics: function (sourceFile, cancellationToken) { return getProgram().getSyntacticDiagnostics(sourceFile, cancellationToken); },
                getDeclarationDiagnostics: function (sourceFile, cancellationToken) { return getProgram().getDeclarationDiagnostics(sourceFile, cancellationToken); },
                getSemanticDiagnostics: function (sourceFile, cancellationToken) { return getProgram().getSemanticDiagnostics(sourceFile, cancellationToken); },
                emit: function (sourceFile, writeFile, cancellationToken, emitOnlyDts, customTransformers) { return getProgram().emit(sourceFile, writeFile, cancellationToken, emitOnlyDts, customTransformers); },
                emitBuildInfo: function (writeFile, cancellationToken) { return getProgram().emitBuildInfo(writeFile, cancellationToken); },
                getAllDependencies: ts.notImplemented,
                getCurrentDirectory: function () { return getProgram().getCurrentDirectory(); },
                close: ts.noop,
            };
            function getProgram() {
                return ts.Debug.checkDefined(getState().program);
            }
        }
        ts.createRedirectedBuilderProgram = createRedirectedBuilderProgram;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        function createSemanticDiagnosticsBuilderProgram(newProgramOrRootNames, hostOrOptions, oldProgramOrHost, configFileParsingDiagnosticsOrOldProgram, configFileParsingDiagnostics, projectReferences) {
            return ts.createBuilderProgram(ts.BuilderProgramKind.SemanticDiagnosticsBuilderProgram, ts.getBuilderCreationParameters(newProgramOrRootNames, hostOrOptions, oldProgramOrHost, configFileParsingDiagnosticsOrOldProgram, configFileParsingDiagnostics, projectReferences));
        }
        ts.createSemanticDiagnosticsBuilderProgram = createSemanticDiagnosticsBuilderProgram;
        function createEmitAndSemanticDiagnosticsBuilderProgram(newProgramOrRootNames, hostOrOptions, oldProgramOrHost, configFileParsingDiagnosticsOrOldProgram, configFileParsingDiagnostics, projectReferences) {
            return ts.createBuilderProgram(ts.BuilderProgramKind.EmitAndSemanticDiagnosticsBuilderProgram, ts.getBuilderCreationParameters(newProgramOrRootNames, hostOrOptions, oldProgramOrHost, configFileParsingDiagnosticsOrOldProgram, configFileParsingDiagnostics, projectReferences));
        }
        ts.createEmitAndSemanticDiagnosticsBuilderProgram = createEmitAndSemanticDiagnosticsBuilderProgram;
        function createAbstractBuilder(newProgramOrRootNames, hostOrOptions, oldProgramOrHost, configFileParsingDiagnosticsOrOldProgram, configFileParsingDiagnostics, projectReferences) {
            var _a = ts.getBuilderCreationParameters(newProgramOrRootNames, hostOrOptions, oldProgramOrHost, configFileParsingDiagnosticsOrOldProgram, configFileParsingDiagnostics, projectReferences), newProgram = _a.newProgram, newConfigFileParsingDiagnostics = _a.configFileParsingDiagnostics;
            return ts.createRedirectedBuilderProgram(function () { return ({ program: newProgram, compilerOptions: newProgram.getCompilerOptions() }); }, newConfigFileParsingDiagnostics);
        }
        ts.createAbstractBuilder = createAbstractBuilder;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        function removeIgnoredPath(path) {
            // Consider whole staging folder as if node_modules changed.
            if (ts.endsWith(path, "/node_modules/.staging")) {
                return ts.removeSuffix(path, "/.staging");
            }
            return ts.some(ts.ignoredPaths, function (searchPath) { return ts.stringContains(path, searchPath); }) ?
                undefined :
                path;
        }
        ts.removeIgnoredPath = removeIgnoredPath;
        /**
         * Filter out paths like
         * "/", "/user", "/user/username", "/user/username/folderAtRoot",
         * "c:/", "c:/users", "c:/users/username", "c:/users/username/folderAtRoot", "c:/folderAtRoot"
         * @param dirPath
         */
        function canWatchDirectory(dirPath) {
            var rootLength = ts.getRootLength(dirPath);
            if (dirPath.length === rootLength) {
                // Ignore "/", "c:/"
                return false;
            }
            var nextDirectorySeparator = dirPath.indexOf(ts.directorySeparator, rootLength);
            if (nextDirectorySeparator === -1) {
                // ignore "/user", "c:/users" or "c:/folderAtRoot"
                return false;
            }
            var pathPartForUserCheck = dirPath.substring(rootLength, nextDirectorySeparator + 1);
            var isNonDirectorySeparatorRoot = rootLength > 1 || dirPath.charCodeAt(0) !== 47 /* slash */;
            if (isNonDirectorySeparatorRoot &&
                dirPath.search(/[a-zA-Z]:/) !== 0 && // Non dos style paths
                pathPartForUserCheck.search(/[a-zA-z]\$\//) === 0) { // Dos style nextPart
                nextDirectorySeparator = dirPath.indexOf(ts.directorySeparator, nextDirectorySeparator + 1);
                if (nextDirectorySeparator === -1) {
                    // ignore "//vda1cs4850/c$/folderAtRoot"
                    return false;
                }
                pathPartForUserCheck = dirPath.substring(rootLength + pathPartForUserCheck.length, nextDirectorySeparator + 1);
            }
            if (isNonDirectorySeparatorRoot &&
                pathPartForUserCheck.search(/users\//i) !== 0) {
                // Paths like c:/folderAtRoot/subFolder are allowed
                return true;
            }
            for (var searchIndex = nextDirectorySeparator + 1, searchLevels = 2; searchLevels > 0; searchLevels--) {
                searchIndex = dirPath.indexOf(ts.directorySeparator, searchIndex) + 1;
                if (searchIndex === 0) {
                    // Folder isnt at expected minimum levels
                    return false;
                }
            }
            return true;
        }
        ts.canWatchDirectory = canWatchDirectory;
        function createResolutionCache(resolutionHost, rootDirForResolution, logChangesWhenResolvingModule) {
            var filesWithChangedSetOfUnresolvedImports;
            var filesWithInvalidatedResolutions;
            var filesWithInvalidatedNonRelativeUnresolvedImports;
            var nonRelativeExternalModuleResolutions = ts.createMultiMap();
            var resolutionsWithFailedLookups = [];
            var resolvedFileToResolution = ts.createMultiMap();
            var hasChangedAutomaticTypeDirectiveNames = false;
            var failedLookupChecks;
            var startsWithPathChecks;
            var isInDirectoryChecks;
            var getCurrentDirectory = ts.memoize(function () { return resolutionHost.getCurrentDirectory(); }); // TODO: GH#18217
            var cachedDirectoryStructureHost = resolutionHost.getCachedDirectoryStructureHost();
            // The resolvedModuleNames and resolvedTypeReferenceDirectives are the cache of resolutions per file.
            // The key in the map is source file's path.
            // The values are Map of resolutions with key being name lookedup.
            var resolvedModuleNames = new ts.Map();
            var perDirectoryResolvedModuleNames = ts.createCacheWithRedirects();
            var nonRelativeModuleNameCache = ts.createCacheWithRedirects();
            var moduleResolutionCache = ts.createModuleResolutionCache(getCurrentDirectory(), resolutionHost.getCanonicalFileName, 
            /*options*/ undefined, perDirectoryResolvedModuleNames, nonRelativeModuleNameCache);
            var resolvedTypeReferenceDirectives = new ts.Map();
            var perDirectoryResolvedTypeReferenceDirectives = ts.createCacheWithRedirects();
            var typeReferenceDirectiveResolutionCache = ts.createTypeReferenceDirectiveResolutionCache(getCurrentDirectory(), resolutionHost.getCanonicalFileName, 
            /*options*/ undefined, moduleResolutionCache.getPackageJsonInfoCache(), perDirectoryResolvedTypeReferenceDirectives);
            /**
             * These are the extensions that failed lookup files will have by default,
             * any other extension of failed lookup will be store that path in custom failed lookup path
             * This helps in not having to comb through all resolutions when files are added/removed
             * Note that .d.ts file also has .d.ts extension hence will be part of default extensions
             */
            var failedLookupDefaultExtensions = [".ts" /* Ts */, ".tsx" /* Tsx */, ".js" /* Js */, ".jsx" /* Jsx */, ".json" /* Json */];
            var customFailedLookupPaths = new ts.Map();
            var directoryWatchesOfFailedLookups = new ts.Map();
            var rootDir = rootDirForResolution && ts.removeTrailingDirectorySeparator(ts.getNormalizedAbsolutePath(rootDirForResolution, getCurrentDirectory()));
            var rootPath = (rootDir && resolutionHost.toPath(rootDir)); // TODO: GH#18217
            var rootSplitLength = rootPath !== undefined ? rootPath.split(ts.directorySeparator).length : 0;
            // TypeRoot watches for the types that get added as part of getAutomaticTypeDirectiveNames
            var typeRootsWatches = new ts.Map();
            return {
                getModuleResolutionCache: function () { return moduleResolutionCache; },
                startRecordingFilesWithChangedResolutions: startRecordingFilesWithChangedResolutions,
                finishRecordingFilesWithChangedResolutions: finishRecordingFilesWithChangedResolutions,
                // perDirectoryResolvedModuleNames and perDirectoryResolvedTypeReferenceDirectives could be non empty if there was exception during program update
                // (between startCachingPerDirectoryResolution and finishCachingPerDirectoryResolution)
                startCachingPerDirectoryResolution: clearPerDirectoryResolutions,
                finishCachingPerDirectoryResolution: finishCachingPerDirectoryResolution,
                resolveModuleNames: resolveModuleNames,
                getResolvedModuleWithFailedLookupLocationsFromCache: getResolvedModuleWithFailedLookupLocationsFromCache,
                resolveTypeReferenceDirectives: resolveTypeReferenceDirectives,
                removeResolutionsFromProjectReferenceRedirects: removeResolutionsFromProjectReferenceRedirects,
                removeResolutionsOfFile: removeResolutionsOfFile,
                hasChangedAutomaticTypeDirectiveNames: function () { return hasChangedAutomaticTypeDirectiveNames; },
                invalidateResolutionOfFile: invalidateResolutionOfFile,
                invalidateResolutionsOfFailedLookupLocations: invalidateResolutionsOfFailedLookupLocations,
                setFilesWithInvalidatedNonRelativeUnresolvedImports: setFilesWithInvalidatedNonRelativeUnresolvedImports,
                createHasInvalidatedResolution: createHasInvalidatedResolution,
                isFileWithInvalidatedNonRelativeUnresolvedImports: isFileWithInvalidatedNonRelativeUnresolvedImports,
                updateTypeRootsWatch: updateTypeRootsWatch,
                closeTypeRootsWatch: closeTypeRootsWatch,
                clear: clear
            };
            function getResolvedModule(resolution) {
                return resolution.resolvedModule;
            }
            function getResolvedTypeReferenceDirective(resolution) {
                return resolution.resolvedTypeReferenceDirective;
            }
            function isInDirectoryPath(dir, file) {
                if (dir === undefined || file.length <= dir.length) {
                    return false;
                }
                return ts.startsWith(file, dir) && file[dir.length] === ts.directorySeparator;
            }
            function clear() {
                ts.clearMap(directoryWatchesOfFailedLookups, ts.closeFileWatcherOf);
                customFailedLookupPaths.clear();
                nonRelativeExternalModuleResolutions.clear();
                closeTypeRootsWatch();
                resolvedModuleNames.clear();
                resolvedTypeReferenceDirectives.clear();
                resolvedFileToResolution.clear();
                resolutionsWithFailedLookups.length = 0;
                failedLookupChecks = undefined;
                startsWithPathChecks = undefined;
                isInDirectoryChecks = undefined;
                // perDirectoryResolvedModuleNames and perDirectoryResolvedTypeReferenceDirectives could be non empty if there was exception during program update
                // (between startCachingPerDirectoryResolution and finishCachingPerDirectoryResolution)
                clearPerDirectoryResolutions();
                hasChangedAutomaticTypeDirectiveNames = false;
            }
            function startRecordingFilesWithChangedResolutions() {
                filesWithChangedSetOfUnresolvedImports = [];
            }
            function finishRecordingFilesWithChangedResolutions() {
                var collected = filesWithChangedSetOfUnresolvedImports;
                filesWithChangedSetOfUnresolvedImports = undefined;
                return collected;
            }
            function isFileWithInvalidatedNonRelativeUnresolvedImports(path) {
                if (!filesWithInvalidatedNonRelativeUnresolvedImports) {
                    return false;
                }
                // Invalidated if file has unresolved imports
                var value = filesWithInvalidatedNonRelativeUnresolvedImports.get(path);
                return !!value && !!value.length;
            }
            function createHasInvalidatedResolution(forceAllFilesAsInvalidated) {
                // Ensure pending resolutions are applied
                invalidateResolutionsOfFailedLookupLocations();
                if (forceAllFilesAsInvalidated) {
                    // Any file asked would have invalidated resolution
                    filesWithInvalidatedResolutions = undefined;
                    return ts.returnTrue;
                }
                var collected = filesWithInvalidatedResolutions;
                filesWithInvalidatedResolutions = undefined;
                return function (path) {
                    return (!!collected && collected.has(path)) ||
                        isFileWithInvalidatedNonRelativeUnresolvedImports(path);
                };
            }
            function clearPerDirectoryResolutions() {
                moduleResolutionCache.clear();
                typeReferenceDirectiveResolutionCache.clear();
                nonRelativeExternalModuleResolutions.forEach(watchFailedLookupLocationOfNonRelativeModuleResolutions);
                nonRelativeExternalModuleResolutions.clear();
            }
            function finishCachingPerDirectoryResolution() {
                filesWithInvalidatedNonRelativeUnresolvedImports = undefined;
                clearPerDirectoryResolutions();
                directoryWatchesOfFailedLookups.forEach(function (watcher, path) {
                    if (watcher.refCount === 0) {
                        directoryWatchesOfFailedLookups.delete(path);
                        watcher.watcher.close();
                    }
                });
                hasChangedAutomaticTypeDirectiveNames = false;
            }
            function resolveModuleName(moduleName, containingFile, compilerOptions, host, redirectedReference) {
                var _a;
                var primaryResult = ts.resolveModuleName(moduleName, containingFile, compilerOptions, host, moduleResolutionCache, redirectedReference);
                // return result immediately only if global cache support is not enabled or if it is .ts, .tsx or .d.ts
                if (!resolutionHost.getGlobalCache) {
                    return primaryResult;
                }
                // otherwise try to load typings from @types
                var globalCache = resolutionHost.getGlobalCache();
                if (globalCache !== undefined && !ts.isExternalModuleNameRelative(moduleName) && !(primaryResult.resolvedModule && ts.extensionIsTS(primaryResult.resolvedModule.extension))) {
                    // create different collection of failed lookup locations for second pass
                    // if it will fail and we've already found something during the first pass - we don't want to pollute its results
                    var _b = ts.loadModuleFromGlobalCache(ts.Debug.checkDefined(resolutionHost.globalCacheResolutionModuleName)(moduleName), resolutionHost.projectName, compilerOptions, host, globalCache, moduleResolutionCache), resolvedModule = _b.resolvedModule, failedLookupLocations = _b.failedLookupLocations;
                    if (resolvedModule) {
                        // Modify existing resolution so its saved in the directory cache as well
                        primaryResult.resolvedModule = resolvedModule;
                        (_a = primaryResult.failedLookupLocations).push.apply(_a, failedLookupLocations);
                        return primaryResult;
                    }
                }
                // Default return the result from the first pass
                return primaryResult;
            }
            function resolveTypeReferenceDirective(typeReferenceDirectiveName, containingFile, options, host, redirectedReference) {
                return ts.resolveTypeReferenceDirective(typeReferenceDirectiveName, containingFile, options, host, redirectedReference, typeReferenceDirectiveResolutionCache);
            }
            function resolveNamesWithLocalCache(_a) {
                var _b, _c, _d;
                var names = _a.names, containingFile = _a.containingFile, redirectedReference = _a.redirectedReference, cache = _a.cache, perDirectoryCacheWithRedirects = _a.perDirectoryCacheWithRedirects, loader = _a.loader, getResolutionWithResolvedFileName = _a.getResolutionWithResolvedFileName, shouldRetryResolution = _a.shouldRetryResolution, reusedNames = _a.reusedNames, logChanges = _a.logChanges;
                var path = resolutionHost.toPath(containingFile);
                var resolutionsInFile = cache.get(path) || cache.set(path, new ts.Map()).get(path);
                var dirPath = ts.getDirectoryPath(path);
                var perDirectoryCache = perDirectoryCacheWithRedirects.getOrCreateMapOfCacheRedirects(redirectedReference);
                var perDirectoryResolution = perDirectoryCache.get(dirPath);
                if (!perDirectoryResolution) {
                    perDirectoryResolution = new ts.Map();
                    perDirectoryCache.set(dirPath, perDirectoryResolution);
                }
                var resolvedModules = [];
                var compilerOptions = resolutionHost.getCompilationSettings();
                var hasInvalidatedNonRelativeUnresolvedImport = logChanges && isFileWithInvalidatedNonRelativeUnresolvedImports(path);
                // All the resolutions in this file are invalidated if this file wasn't resolved using same redirect
                var program = resolutionHost.getCurrentProgram();
                var oldRedirect = program && program.getResolvedProjectReferenceToRedirect(containingFile);
                var unmatchedRedirects = oldRedirect ?
                    !redirectedReference || redirectedReference.sourceFile.path !== oldRedirect.sourceFile.path :
                    !!redirectedReference;
                var seenNamesInFile = new ts.Map();
                for (var _i = 0, names_3 = names; _i < names_3.length; _i++) {
                    var name = names_3[_i];
                    var resolution = resolutionsInFile.get(name);
                    // Resolution is valid if it is present and not invalidated
                    if (!seenNamesInFile.has(name) &&
                        unmatchedRedirects || !resolution || resolution.isInvalidated ||
                        // If the name is unresolved import that was invalidated, recalculate
                        (hasInvalidatedNonRelativeUnresolvedImport && !ts.isExternalModuleNameRelative(name) && shouldRetryResolution(resolution))) {
                        var existingResolution = resolution;
                        var resolutionInDirectory = perDirectoryResolution.get(name);
                        if (resolutionInDirectory) {
                            resolution = resolutionInDirectory;
                            var host = ((_b = resolutionHost.getCompilerHost) === null || _b === void 0 ? void 0 : _b.call(resolutionHost)) || resolutionHost;
                            if (ts.isTraceEnabled(compilerOptions, host)) {
                                var resolved = getResolutionWithResolvedFileName(resolution);
                                ts.trace(host, loader === resolveModuleName ?
                                    (resolved === null || resolved === void 0 ? void 0 : resolved.resolvedFileName) ?
                                        resolved.packagetId ?
                                            ts.Diagnostics.Reusing_resolution_of_module_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_3_with_Package_ID_4 :
                                            ts.Diagnostics.Reusing_resolution_of_module_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_3 :
                                        ts.Diagnostics.Reusing_resolution_of_module_0_from_1_found_in_cache_from_location_2_it_was_not_resolved :
                                    (resolved === null || resolved === void 0 ? void 0 : resolved.resolvedFileName) ?
                                        resolved.packagetId ?
                                            ts.Diagnostics.Reusing_resolution_of_type_reference_directive_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_3_with_Package_ID_4 :
                                            ts.Diagnostics.Reusing_resolution_of_type_reference_directive_0_from_1_found_in_cache_from_location_2_it_was_successfully_resolved_to_3 :
                                        ts.Diagnostics.Reusing_resolution_of_type_reference_directive_0_from_1_found_in_cache_from_location_2_it_was_not_resolved, name, containingFile, ts.getDirectoryPath(containingFile), resolved === null || resolved === void 0 ? void 0 : resolved.resolvedFileName, (resolved === null || resolved === void 0 ? void 0 : resolved.packagetId) && ts.packageIdToString(resolved.packagetId));
                            }
                        }
                        else {
                            resolution = loader(name, containingFile, compilerOptions, ((_c = resolutionHost.getCompilerHost) === null || _c === void 0 ? void 0 : _c.call(resolutionHost)) || resolutionHost, redirectedReference);
                            perDirectoryResolution.set(name, resolution);
                        }
                        resolutionsInFile.set(name, resolution);
                        watchFailedLookupLocationsOfExternalModuleResolutions(name, resolution, path, getResolutionWithResolvedFileName);
                        if (existingResolution) {
                            stopWatchFailedLookupLocationOfResolution(existingResolution, path, getResolutionWithResolvedFileName);
                        }
                        if (logChanges && filesWithChangedSetOfUnresolvedImports && !resolutionIsEqualTo(existingResolution, resolution)) {
                            filesWithChangedSetOfUnresolvedImports.push(path);
                            // reset log changes to avoid recording the same file multiple times
                            logChanges = false;
                        }
                    }
                    else {
                        var host = ((_d = resolutionHost.getCompilerHost) === null || _d === void 0 ? void 0 : _d.call(resolutionHost)) || resolutionHost;
                        if (ts.isTraceEnabled(compilerOptions, host) && !seenNamesInFile.has(name)) {
                            var resolved = getResolutionWithResolvedFileName(resolution);
                            ts.trace(host, loader === resolveModuleName ?
                                (resolved === null || resolved === void 0 ? void 0 : resolved.resolvedFileName) ?
                                    resolved.packagetId ?
                                        ts.Diagnostics.Reusing_resolution_of_module_0_from_1_of_old_program_it_was_successfully_resolved_to_2_with_Package_ID_3 :
                                        ts.Diagnostics.Reusing_resolution_of_module_0_from_1_of_old_program_it_was_successfully_resolved_to_2 :
                                    ts.Diagnostics.Reusing_resolution_of_module_0_from_1_of_old_program_it_was_not_resolved :
                                (resolved === null || resolved === void 0 ? void 0 : resolved.resolvedFileName) ?
                                    resolved.packagetId ?
                                        ts.Diagnostics.Reusing_resolution_of_type_reference_directive_0_from_1_of_old_program_it_was_successfully_resolved_to_2_with_Package_ID_3 :
                                        ts.Diagnostics.Reusing_resolution_of_type_reference_directive_0_from_1_of_old_program_it_was_successfully_resolved_to_2 :
                                    ts.Diagnostics.Reusing_resolution_of_type_reference_directive_0_from_1_of_old_program_it_was_not_resolved, name, containingFile, resolved === null || resolved === void 0 ? void 0 : resolved.resolvedFileName, (resolved === null || resolved === void 0 ? void 0 : resolved.packagetId) && ts.packageIdToString(resolved.packagetId));
                        }
                    }
                    ts.Debug.assert(resolution !== undefined && !resolution.isInvalidated);
                    seenNamesInFile.set(name, true);
                    resolvedModules.push(getResolutionWithResolvedFileName(resolution));
                }
                // Stop watching and remove the unused name
                resolutionsInFile.forEach(function (resolution, name) {
                    if (!seenNamesInFile.has(name) && !ts.contains(reusedNames, name)) {
                        stopWatchFailedLookupLocationOfResolution(resolution, path, getResolutionWithResolvedFileName);
                        resolutionsInFile.delete(name);
                    }
                });
                return resolvedModules;
                function resolutionIsEqualTo(oldResolution, newResolution) {
                    if (oldResolution === newResolution) {
                        return true;
                    }
                    if (!oldResolution || !newResolution) {
                        return false;
                    }
                    var oldResult = getResolutionWithResolvedFileName(oldResolution);
                    var newResult = getResolutionWithResolvedFileName(newResolution);
                    if (oldResult === newResult) {
                        return true;
                    }
                    if (!oldResult || !newResult) {
                        return false;
                    }
                    return oldResult.resolvedFileName === newResult.resolvedFileName;
                }
            }
            function resolveTypeReferenceDirectives(typeDirectiveNames, containingFile, redirectedReference) {
                return resolveNamesWithLocalCache({
                    names: typeDirectiveNames,
                    containingFile: containingFile,
                    redirectedReference: redirectedReference,
                    cache: resolvedTypeReferenceDirectives,
                    perDirectoryCacheWithRedirects: perDirectoryResolvedTypeReferenceDirectives,
                    loader: resolveTypeReferenceDirective,
                    getResolutionWithResolvedFileName: getResolvedTypeReferenceDirective,
                    shouldRetryResolution: function (resolution) { return resolution.resolvedTypeReferenceDirective === undefined; },
                });
            }
            function resolveModuleNames(moduleNames, containingFile, reusedNames, redirectedReference) {
                return resolveNamesWithLocalCache({
                    names: moduleNames,
                    containingFile: containingFile,
                    redirectedReference: redirectedReference,
                    cache: resolvedModuleNames,
                    perDirectoryCacheWithRedirects: perDirectoryResolvedModuleNames,
                    loader: resolveModuleName,
                    getResolutionWithResolvedFileName: getResolvedModule,
                    shouldRetryResolution: function (resolution) { return !resolution.resolvedModule || !ts.resolutionExtensionIsTSOrJson(resolution.resolvedModule.extension); },
                    reusedNames: reusedNames,
                    logChanges: logChangesWhenResolvingModule,
                });
            }
            function getResolvedModuleWithFailedLookupLocationsFromCache(moduleName, containingFile) {
                var cache = resolvedModuleNames.get(resolutionHost.toPath(containingFile));
                return cache && cache.get(moduleName);
            }
            function isNodeModulesAtTypesDirectory(dirPath) {
                return ts.endsWith(dirPath, "/node_modules/@types");
            }
            function getDirectoryToWatchFailedLookupLocation(failedLookupLocation, failedLookupLocationPath) {
                if (isInDirectoryPath(rootPath, failedLookupLocationPath)) {
                    // Ensure failed look up is normalized path
                    failedLookupLocation = ts.isRootedDiskPath(failedLookupLocation) ? ts.normalizePath(failedLookupLocation) : ts.getNormalizedAbsolutePath(failedLookupLocation, getCurrentDirectory());
                    var failedLookupPathSplit = failedLookupLocationPath.split(ts.directorySeparator);
                    var failedLookupSplit = failedLookupLocation.split(ts.directorySeparator);
                    ts.Debug.assert(failedLookupSplit.length === failedLookupPathSplit.length, "FailedLookup: " + failedLookupLocation + " failedLookupLocationPath: " + failedLookupLocationPath);
                    if (failedLookupPathSplit.length > rootSplitLength + 1) {
                        // Instead of watching root, watch directory in root to avoid watching excluded directories not needed for module resolution
                        return {
                            dir: failedLookupSplit.slice(0, rootSplitLength + 1).join(ts.directorySeparator),
                            dirPath: failedLookupPathSplit.slice(0, rootSplitLength + 1).join(ts.directorySeparator)
                        };
                    }
                    else {
                        // Always watch root directory non recursively
                        return {
                            dir: rootDir,
                            dirPath: rootPath,
                            nonRecursive: false
                        };
                    }
                }
                return getDirectoryToWatchFromFailedLookupLocationDirectory(ts.getDirectoryPath(ts.getNormalizedAbsolutePath(failedLookupLocation, getCurrentDirectory())), ts.getDirectoryPath(failedLookupLocationPath));
            }
            function getDirectoryToWatchFromFailedLookupLocationDirectory(dir, dirPath) {
                // If directory path contains node module, get the most parent node_modules directory for watching
                while (ts.pathContainsNodeModules(dirPath)) {
                    dir = ts.getDirectoryPath(dir);
                    dirPath = ts.getDirectoryPath(dirPath);
                }
                // If the directory is node_modules use it to watch, always watch it recursively
                if (ts.isNodeModulesDirectory(dirPath)) {
                    return canWatchDirectory(ts.getDirectoryPath(dirPath)) ? { dir: dir, dirPath: dirPath } : undefined;
                }
                var nonRecursive = true;
                // Use some ancestor of the root directory
                var subDirectoryPath, subDirectory;
                if (rootPath !== undefined) {
                    while (!isInDirectoryPath(dirPath, rootPath)) {
                        var parentPath = ts.getDirectoryPath(dirPath);
                        if (parentPath === dirPath) {
                            break;
                        }
                        nonRecursive = false;
                        subDirectoryPath = dirPath;
                        subDirectory = dir;
                        dirPath = parentPath;
                        dir = ts.getDirectoryPath(dir);
                    }
                }
                return canWatchDirectory(dirPath) ? { dir: subDirectory || dir, dirPath: subDirectoryPath || dirPath, nonRecursive: nonRecursive } : undefined;
            }
            function isPathWithDefaultFailedLookupExtension(path) {
                return ts.fileExtensionIsOneOf(path, failedLookupDefaultExtensions);
            }
            function watchFailedLookupLocationsOfExternalModuleResolutions(name, resolution, filePath, getResolutionWithResolvedFileName) {
                if (resolution.refCount) {
                    resolution.refCount++;
                    ts.Debug.assertDefined(resolution.files);
                }
                else {
                    resolution.refCount = 1;
                    ts.Debug.assert(ts.length(resolution.files) === 0); // This resolution shouldnt be referenced by any file yet
                    if (ts.isExternalModuleNameRelative(name)) {
                        watchFailedLookupLocationOfResolution(resolution);
                    }
                    else {
                        nonRelativeExternalModuleResolutions.add(name, resolution);
                    }
                    var resolved = getResolutionWithResolvedFileName(resolution);
                    if (resolved && resolved.resolvedFileName) {
                        resolvedFileToResolution.add(resolutionHost.toPath(resolved.resolvedFileName), resolution);
                    }
                }
                (resolution.files || (resolution.files = [])).push(filePath);
            }
            function watchFailedLookupLocationOfResolution(resolution) {
                ts.Debug.assert(!!resolution.refCount);
                var failedLookupLocations = resolution.failedLookupLocations;
                if (!failedLookupLocations.length)
                    return;
                resolutionsWithFailedLookups.push(resolution);
                var setAtRoot = false;
                for (var _i = 0, failedLookupLocations_1 = failedLookupLocations; _i < failedLookupLocations_1.length; _i++) {
                    var failedLookupLocation = failedLookupLocations_1[_i];
                    var failedLookupLocationPath = resolutionHost.toPath(failedLookupLocation);
                    var toWatch = getDirectoryToWatchFailedLookupLocation(failedLookupLocation, failedLookupLocationPath);
                    if (toWatch) {
                        var dir = toWatch.dir, dirPath = toWatch.dirPath, nonRecursive = toWatch.nonRecursive;
                        // If the failed lookup location path is not one of the supported extensions,
                        // store it in the custom path
                        if (!isPathWithDefaultFailedLookupExtension(failedLookupLocationPath)) {
                            var refCount = customFailedLookupPaths.get(failedLookupLocationPath) || 0;
                            customFailedLookupPaths.set(failedLookupLocationPath, refCount + 1);
                        }
                        if (dirPath === rootPath) {
                            ts.Debug.assert(!nonRecursive);
                            setAtRoot = true;
                        }
                        else {
                            setDirectoryWatcher(dir, dirPath, nonRecursive);
                        }
                    }
                }
                if (setAtRoot) {
                    // This is always non recursive
                    setDirectoryWatcher(rootDir, rootPath, /*nonRecursive*/ true); // TODO: GH#18217
                }
            }
            function watchFailedLookupLocationOfNonRelativeModuleResolutions(resolutions, name) {
                var program = resolutionHost.getCurrentProgram();
                if (!program || !program.getTypeChecker().tryFindAmbientModuleWithoutAugmentations(name)) {
                    resolutions.forEach(watchFailedLookupLocationOfResolution);
                }
            }
            function setDirectoryWatcher(dir, dirPath, nonRecursive) {
                var dirWatcher = directoryWatchesOfFailedLookups.get(dirPath);
                if (dirWatcher) {
                    ts.Debug.assert(!!nonRecursive === !!dirWatcher.nonRecursive);
                    dirWatcher.refCount++;
                }
                else {
                    directoryWatchesOfFailedLookups.set(dirPath, { watcher: createDirectoryWatcher(dir, dirPath, nonRecursive), refCount: 1, nonRecursive: nonRecursive });
                }
            }
            function stopWatchFailedLookupLocationOfResolution(resolution, filePath, getResolutionWithResolvedFileName) {
                ts.unorderedRemoveItem(ts.Debug.assertDefined(resolution.files), filePath);
                resolution.refCount--;
                if (resolution.refCount) {
                    return;
                }
                var resolved = getResolutionWithResolvedFileName(resolution);
                if (resolved && resolved.resolvedFileName) {
                    resolvedFileToResolution.remove(resolutionHost.toPath(resolved.resolvedFileName), resolution);
                }
                if (!ts.unorderedRemoveItem(resolutionsWithFailedLookups, resolution)) {
                    // If not watching failed lookups, it wont be there in resolutionsWithFailedLookups
                    return;
                }
                var failedLookupLocations = resolution.failedLookupLocations;
                var removeAtRoot = false;
                for (var _i = 0, failedLookupLocations_2 = failedLookupLocations; _i < failedLookupLocations_2.length; _i++) {
                    var failedLookupLocation = failedLookupLocations_2[_i];
                    var failedLookupLocationPath = resolutionHost.toPath(failedLookupLocation);
                    var toWatch = getDirectoryToWatchFailedLookupLocation(failedLookupLocation, failedLookupLocationPath);
                    if (toWatch) {
                        var dirPath = toWatch.dirPath;
                        var refCount = customFailedLookupPaths.get(failedLookupLocationPath);
                        if (refCount) {
                            if (refCount === 1) {
                                customFailedLookupPaths.delete(failedLookupLocationPath);
                            }
                            else {
                                ts.Debug.assert(refCount > 1);
                                customFailedLookupPaths.set(failedLookupLocationPath, refCount - 1);
                            }
                        }
                        if (dirPath === rootPath) {
                            removeAtRoot = true;
                        }
                        else {
                            removeDirectoryWatcher(dirPath);
                        }
                    }
                }
                if (removeAtRoot) {
                    removeDirectoryWatcher(rootPath);
                }
            }
            function removeDirectoryWatcher(dirPath) {
                var dirWatcher = directoryWatchesOfFailedLookups.get(dirPath);
                // Do not close the watcher yet since it might be needed by other failed lookup locations.
                dirWatcher.refCount--;
            }
            function createDirectoryWatcher(directory, dirPath, nonRecursive) {
                return resolutionHost.watchDirectoryOfFailedLookupLocation(directory, function (fileOrDirectory) {
                    var fileOrDirectoryPath = resolutionHost.toPath(fileOrDirectory);
                    if (cachedDirectoryStructureHost) {
                        // Since the file existence changed, update the sourceFiles cache
                        cachedDirectoryStructureHost.addOrDeleteFileOrDirectory(fileOrDirectory, fileOrDirectoryPath);
                    }
                    scheduleInvalidateResolutionOfFailedLookupLocation(fileOrDirectoryPath, dirPath === fileOrDirectoryPath);
                }, nonRecursive ? 0 /* None */ : 1 /* Recursive */);
            }
            function removeResolutionsOfFileFromCache(cache, filePath, getResolutionWithResolvedFileName) {
                // Deleted file, stop watching failed lookups for all the resolutions in the file
                var resolutions = cache.get(filePath);
                if (resolutions) {
                    resolutions.forEach(function (resolution) { return stopWatchFailedLookupLocationOfResolution(resolution, filePath, getResolutionWithResolvedFileName); });
                    cache.delete(filePath);
                }
            }
            function removeResolutionsFromProjectReferenceRedirects(filePath) {
                if (!ts.fileExtensionIs(filePath, ".json" /* Json */)) {
                    return;
                }
                var program = resolutionHost.getCurrentProgram();
                if (!program) {
                    return;
                }
                // If this file is input file for the referenced project, get it
                var resolvedProjectReference = program.getResolvedProjectReferenceByPath(filePath);
                if (!resolvedProjectReference) {
                    return;
                }
                // filePath is for the projectReference and the containing file is from this project reference, invalidate the resolution
                resolvedProjectReference.commandLine.fileNames.forEach(function (f) { return removeResolutionsOfFile(resolutionHost.toPath(f)); });
            }
            function removeResolutionsOfFile(filePath) {
                removeResolutionsOfFileFromCache(resolvedModuleNames, filePath, getResolvedModule);
                removeResolutionsOfFileFromCache(resolvedTypeReferenceDirectives, filePath, getResolvedTypeReferenceDirective);
            }
            function invalidateResolutions(resolutions, canInvalidate) {
                if (!resolutions)
                    return false;
                var invalidated = false;
                for (var _i = 0, resolutions_1 = resolutions; _i < resolutions_1.length; _i++) {
                    var resolution = resolutions_1[_i];
                    if (resolution.isInvalidated || !canInvalidate(resolution))
                        continue;
                    resolution.isInvalidated = invalidated = true;
                    for (var _a = 0, _b = ts.Debug.assertDefined(resolution.files); _a < _b.length; _a++) {
                        var containingFilePath = _b[_a];
                        (filesWithInvalidatedResolutions || (filesWithInvalidatedResolutions = new ts.Set())).add(containingFilePath);
                        // When its a file with inferred types resolution, invalidate type reference directive resolution
                        hasChangedAutomaticTypeDirectiveNames = hasChangedAutomaticTypeDirectiveNames || ts.endsWith(containingFilePath, ts.inferredTypesContainingFile);
                    }
                }
                return invalidated;
            }
            function invalidateResolutionOfFile(filePath) {
                removeResolutionsOfFile(filePath);
                // Resolution is invalidated if the resulting file name is same as the deleted file path
                var prevHasChangedAutomaticTypeDirectiveNames = hasChangedAutomaticTypeDirectiveNames;
                if (invalidateResolutions(resolvedFileToResolution.get(filePath), ts.returnTrue) &&
                    hasChangedAutomaticTypeDirectiveNames &&
                    !prevHasChangedAutomaticTypeDirectiveNames) {
                    resolutionHost.onChangedAutomaticTypeDirectiveNames();
                }
            }
            function setFilesWithInvalidatedNonRelativeUnresolvedImports(filesMap) {
                ts.Debug.assert(filesWithInvalidatedNonRelativeUnresolvedImports === filesMap || filesWithInvalidatedNonRelativeUnresolvedImports === undefined);
                filesWithInvalidatedNonRelativeUnresolvedImports = filesMap;
            }
            function scheduleInvalidateResolutionOfFailedLookupLocation(fileOrDirectoryPath, isCreatingWatchedDirectory) {
                if (isCreatingWatchedDirectory) {
                    // Watching directory is created
                    // Invalidate any resolution has failed lookup in this directory
                    (isInDirectoryChecks || (isInDirectoryChecks = [])).push(fileOrDirectoryPath);
                }
                else {
                    // If something to do with folder/file starting with "." in node_modules folder, skip it
                    var updatedPath = removeIgnoredPath(fileOrDirectoryPath);
                    if (!updatedPath)
                        return false;
                    fileOrDirectoryPath = updatedPath;
                    // prevent saving an open file from over-eagerly triggering invalidation
                    if (resolutionHost.fileIsOpen(fileOrDirectoryPath)) {
                        return false;
                    }
                    // Some file or directory in the watching directory is created
                    // Return early if it does not have any of the watching extension or not the custom failed lookup path
                    var dirOfFileOrDirectory = ts.getDirectoryPath(fileOrDirectoryPath);
                    if (isNodeModulesAtTypesDirectory(fileOrDirectoryPath) || ts.isNodeModulesDirectory(fileOrDirectoryPath) ||
                        isNodeModulesAtTypesDirectory(dirOfFileOrDirectory) || ts.isNodeModulesDirectory(dirOfFileOrDirectory)) {
                        // Invalidate any resolution from this directory
                        (failedLookupChecks || (failedLookupChecks = [])).push(fileOrDirectoryPath);
                        (startsWithPathChecks || (startsWithPathChecks = new ts.Set())).add(fileOrDirectoryPath);
                    }
                    else {
                        if (!isPathWithDefaultFailedLookupExtension(fileOrDirectoryPath) && !customFailedLookupPaths.has(fileOrDirectoryPath)) {
                            return false;
                        }
                        // Ignore emits from the program
                        if (ts.isEmittedFileOfProgram(resolutionHost.getCurrentProgram(), fileOrDirectoryPath)) {
                            return false;
                        }
                        // Resolution need to be invalidated if failed lookup location is same as the file or directory getting created
                        (failedLookupChecks || (failedLookupChecks = [])).push(fileOrDirectoryPath);
                        // If the invalidated file is from a node_modules package, invalidate everything else
                        // in the package since we might not get notifications for other files in the package.
                        // This hardens our logic against unreliable file watchers.
                        var packagePath = ts.parseNodeModuleFromPath(fileOrDirectoryPath);
                        if (packagePath)
                            (startsWithPathChecks || (startsWithPathChecks = new ts.Set())).add(packagePath);
                    }
                }
                resolutionHost.scheduleInvalidateResolutionsOfFailedLookupLocations();
            }
            function invalidateResolutionsOfFailedLookupLocations() {
                if (!failedLookupChecks && !startsWithPathChecks && !isInDirectoryChecks) {
                    return false;
                }
                var invalidated = invalidateResolutions(resolutionsWithFailedLookups, canInvalidateFailedLookupResolution);
                failedLookupChecks = undefined;
                startsWithPathChecks = undefined;
                isInDirectoryChecks = undefined;
                return invalidated;
            }
            function canInvalidateFailedLookupResolution(resolution) {
                return resolution.failedLookupLocations.some(function (location) {
                    var locationPath = resolutionHost.toPath(location);
                    return ts.contains(failedLookupChecks, locationPath) ||
                        ts.firstDefinedIterator((startsWithPathChecks === null || startsWithPathChecks === void 0 ? void 0 : startsWithPathChecks.keys()) || ts.emptyIterator, function (fileOrDirectoryPath) { return ts.startsWith(locationPath, fileOrDirectoryPath) ? true : undefined; }) ||
                        (isInDirectoryChecks === null || isInDirectoryChecks === void 0 ? void 0 : isInDirectoryChecks.some(function (fileOrDirectoryPath) { return isInDirectoryPath(fileOrDirectoryPath, locationPath); }));
                });
            }
            function closeTypeRootsWatch() {
                ts.clearMap(typeRootsWatches, ts.closeFileWatcher);
            }
            function getDirectoryToWatchFailedLookupLocationFromTypeRoot(typeRoot, typeRootPath) {
                if (isInDirectoryPath(rootPath, typeRootPath)) {
                    return rootPath;
                }
                var toWatch = getDirectoryToWatchFromFailedLookupLocationDirectory(typeRoot, typeRootPath);
                return toWatch && directoryWatchesOfFailedLookups.has(toWatch.dirPath) ? toWatch.dirPath : undefined;
            }
            function createTypeRootsWatch(typeRootPath, typeRoot) {
                // Create new watch and recursive info
                return resolutionHost.watchTypeRootsDirectory(typeRoot, function (fileOrDirectory) {
                    var fileOrDirectoryPath = resolutionHost.toPath(fileOrDirectory);
                    if (cachedDirectoryStructureHost) {
                        // Since the file existence changed, update the sourceFiles cache
                        cachedDirectoryStructureHost.addOrDeleteFileOrDirectory(fileOrDirectory, fileOrDirectoryPath);
                    }
                    // For now just recompile
                    // We could potentially store more data here about whether it was/would be really be used or not
                    // and with that determine to trigger compilation but for now this is enough
                    hasChangedAutomaticTypeDirectiveNames = true;
                    resolutionHost.onChangedAutomaticTypeDirectiveNames();
                    // Since directory watchers invoked are flaky, the failed lookup location events might not be triggered
                    // So handle to failed lookup locations here as well to ensure we are invalidating resolutions
                    var dirPath = getDirectoryToWatchFailedLookupLocationFromTypeRoot(typeRoot, typeRootPath);
                    if (dirPath) {
                        scheduleInvalidateResolutionOfFailedLookupLocation(fileOrDirectoryPath, dirPath === fileOrDirectoryPath);
                    }
                }, 1 /* Recursive */);
            }
            /**
             * Watches the types that would get added as part of getAutomaticTypeDirectiveNames
             * To be called when compiler options change
             */
            function updateTypeRootsWatch() {
                var options = resolutionHost.getCompilationSettings();
                if (options.types) {
                    // No need to do any watch since resolution cache is going to handle the failed lookups
                    // for the types added by this
                    closeTypeRootsWatch();
                    return;
                }
                // we need to assume the directories exist to ensure that we can get all the type root directories that get included
                // But filter directories that are at root level to say directory doesnt exist, so that we arent watching them
                var typeRoots = ts.getEffectiveTypeRoots(options, { directoryExists: directoryExistsForTypeRootWatch, getCurrentDirectory: getCurrentDirectory });
                if (typeRoots) {
                    ts.mutateMap(typeRootsWatches, ts.arrayToMap(typeRoots, function (tr) { return resolutionHost.toPath(tr); }), {
                        createNewValue: createTypeRootsWatch,
                        onDeleteValue: ts.closeFileWatcher
                    });
                }
                else {
                    closeTypeRootsWatch();
                }
            }
            /**
             * Use this function to return if directory exists to get type roots to watch
             * If we return directory exists then only the paths will be added to type roots
             * Hence return true for all directories except root directories which are filtered from watching
             */
            function directoryExistsForTypeRootWatch(nodeTypesDirectory) {
                var dir = ts.getDirectoryPath(ts.getDirectoryPath(nodeTypesDirectory));
                var dirPath = resolutionHost.toPath(dir);
                return dirPath === rootPath || canWatchDirectory(dirPath);
            }
        }
        ts.createResolutionCache = createResolutionCache;
    })(ts || (ts = {}));
    // Used by importFixes, getEditsForFileRename, and declaration emit to synthesize import module specifiers.
    /* @internal */
    var ts;
    (function (ts) {
        var moduleSpecifiers;
        (function (moduleSpecifiers_1) {
            var RelativePreference;
            (function (RelativePreference) {
                RelativePreference[RelativePreference["Relative"] = 0] = "Relative";
                RelativePreference[RelativePreference["NonRelative"] = 1] = "NonRelative";
                RelativePreference[RelativePreference["Shortest"] = 2] = "Shortest";
                RelativePreference[RelativePreference["ExternalNonRelative"] = 3] = "ExternalNonRelative";
            })(RelativePreference || (RelativePreference = {}));
            // See UserPreferences#importPathEnding
            var Ending;
            (function (Ending) {
                Ending[Ending["Minimal"] = 0] = "Minimal";
                Ending[Ending["Index"] = 1] = "Index";
                Ending[Ending["JsExtension"] = 2] = "JsExtension";
            })(Ending || (Ending = {}));
            function getPreferences(_a, compilerOptions, importingSourceFile) {
                var importModuleSpecifierPreference = _a.importModuleSpecifierPreference, importModuleSpecifierEnding = _a.importModuleSpecifierEnding;
                return {
                    relativePreference: importModuleSpecifierPreference === "relative" ? 0 /* Relative */ :
                        importModuleSpecifierPreference === "non-relative" ? 1 /* NonRelative */ :
                            importModuleSpecifierPreference === "project-relative" ? 3 /* ExternalNonRelative */ :
                                2 /* Shortest */,
                    ending: getEnding(),
                };
                function getEnding() {
                    switch (importModuleSpecifierEnding) {
                        case "minimal": return 0 /* Minimal */;
                        case "index": return 1 /* Index */;
                        case "js": return 2 /* JsExtension */;
                        default: return usesJsExtensionOnImports(importingSourceFile) ? 2 /* JsExtension */
                            : ts.getEmitModuleResolutionKind(compilerOptions) !== ts.ModuleResolutionKind.NodeJs ? 1 /* Index */ : 0 /* Minimal */;
                    }
                }
            }
            function getPreferencesForUpdate(compilerOptions, oldImportSpecifier) {
                return {
                    relativePreference: ts.isExternalModuleNameRelative(oldImportSpecifier) ? 0 /* Relative */ : 1 /* NonRelative */,
                    ending: ts.hasJSFileExtension(oldImportSpecifier) ?
                        2 /* JsExtension */ :
                        ts.getEmitModuleResolutionKind(compilerOptions) !== ts.ModuleResolutionKind.NodeJs || ts.endsWith(oldImportSpecifier, "index") ? 1 /* Index */ : 0 /* Minimal */,
                };
            }
            function updateModuleSpecifier(compilerOptions, importingSourceFileName, toFileName, host, oldImportSpecifier) {
                var res = getModuleSpecifierWorker(compilerOptions, importingSourceFileName, toFileName, host, getPreferencesForUpdate(compilerOptions, oldImportSpecifier), {});
                if (res === oldImportSpecifier)
                    return undefined;
                return res;
            }
            moduleSpecifiers_1.updateModuleSpecifier = updateModuleSpecifier;
            // Note: importingSourceFile is just for usesJsExtensionOnImports
            function getModuleSpecifier(compilerOptions, importingSourceFile, importingSourceFileName, toFileName, host) {
                return getModuleSpecifierWorker(compilerOptions, importingSourceFileName, toFileName, host, getPreferences({}, compilerOptions, importingSourceFile), {});
            }
            moduleSpecifiers_1.getModuleSpecifier = getModuleSpecifier;
            function getNodeModulesPackageName(compilerOptions, importingSourceFileName, nodeModulesFileName, host, preferences) {
                var info = getInfo(importingSourceFileName, host);
                var modulePaths = getAllModulePaths(importingSourceFileName, nodeModulesFileName, host, preferences);
                return ts.firstDefined(modulePaths, function (modulePath) { return tryGetModuleNameAsNodeModule(modulePath, info, host, compilerOptions, /*packageNameOnly*/ true); });
            }
            moduleSpecifiers_1.getNodeModulesPackageName = getNodeModulesPackageName;
            function getModuleSpecifierWorker(compilerOptions, importingSourceFileName, toFileName, host, preferences, userPreferences) {
                var info = getInfo(importingSourceFileName, host);
                var modulePaths = getAllModulePaths(importingSourceFileName, toFileName, host, userPreferences);
                return ts.firstDefined(modulePaths, function (modulePath) { return tryGetModuleNameAsNodeModule(modulePath, info, host, compilerOptions); }) ||
                    getLocalModuleSpecifier(toFileName, info, compilerOptions, host, preferences);
            }
            function tryGetModuleSpecifiersFromCache(moduleSymbol, importingSourceFile, host, userPreferences) {
                return tryGetModuleSpecifiersFromCacheWorker(moduleSymbol, importingSourceFile, host, userPreferences)[0];
            }
            moduleSpecifiers_1.tryGetModuleSpecifiersFromCache = tryGetModuleSpecifiersFromCache;
            function tryGetModuleSpecifiersFromCacheWorker(moduleSymbol, importingSourceFile, host, userPreferences) {
                var _a;
                var moduleSourceFile = ts.getSourceFileOfModule(moduleSymbol);
                if (!moduleSourceFile) {
                    return ts.emptyArray;
                }
                var cache = (_a = host.getModuleSpecifierCache) === null || _a === void 0 ? void 0 : _a.call(host);
                var cached = cache === null || cache === void 0 ? void 0 : cache.get(importingSourceFile.path, moduleSourceFile.path, userPreferences);
                return [cached === null || cached === void 0 ? void 0 : cached.moduleSpecifiers, moduleSourceFile, cached === null || cached === void 0 ? void 0 : cached.modulePaths, cache];
            }
            /** Returns an import for each symlink and for the realpath. */
            function getModuleSpecifiers(moduleSymbol, checker, compilerOptions, importingSourceFile, host, userPreferences) {
                return getModuleSpecifiersWithCacheInfo(moduleSymbol, checker, compilerOptions, importingSourceFile, host, userPreferences).moduleSpecifiers;
            }
            moduleSpecifiers_1.getModuleSpecifiers = getModuleSpecifiers;
            function getModuleSpecifiersWithCacheInfo(moduleSymbol, checker, compilerOptions, importingSourceFile, host, userPreferences) {
                var computedWithoutCache = false;
                var ambient = tryGetModuleNameFromAmbientModule(moduleSymbol, checker);
                if (ambient)
                    return { moduleSpecifiers: [ambient], computedWithoutCache: computedWithoutCache };
                // eslint-disable-next-line prefer-const
                var _a = tryGetModuleSpecifiersFromCacheWorker(moduleSymbol, importingSourceFile, host, userPreferences), specifiers = _a[0], moduleSourceFile = _a[1], modulePaths = _a[2], cache = _a[3];
                if (specifiers)
                    return { moduleSpecifiers: specifiers, computedWithoutCache: computedWithoutCache };
                if (!moduleSourceFile)
                    return { moduleSpecifiers: ts.emptyArray, computedWithoutCache: computedWithoutCache };
                computedWithoutCache = true;
                modulePaths || (modulePaths = getAllModulePathsWorker(importingSourceFile.path, moduleSourceFile.originalFileName, host));
                var result = computeModuleSpecifiers(modulePaths, compilerOptions, importingSourceFile, host, userPreferences);
                cache === null || cache === void 0 ? void 0 : cache.set(importingSourceFile.path, moduleSourceFile.path, userPreferences, modulePaths, result);
                return { moduleSpecifiers: result, computedWithoutCache: computedWithoutCache };
            }
            moduleSpecifiers_1.getModuleSpecifiersWithCacheInfo = getModuleSpecifiersWithCacheInfo;
            function computeModuleSpecifiers(modulePaths, compilerOptions, importingSourceFile, host, userPreferences) {
                var info = getInfo(importingSourceFile.path, host);
                var preferences = getPreferences(userPreferences, compilerOptions, importingSourceFile);
                var existingSpecifier = ts.forEach(modulePaths, function (modulePath) {
                    return ts.forEach(host.getFileIncludeReasons().get(ts.toPath(modulePath.path, host.getCurrentDirectory(), info.getCanonicalFileName)), function (reason) {
                        if (reason.kind !== ts.FileIncludeKind.Import || reason.file !== importingSourceFile.path)
                            return undefined;
                        var specifier = ts.getModuleNameStringLiteralAt(importingSourceFile, reason.index).text;
                        // If the preference is for non relative and the module specifier is relative, ignore it
                        return preferences.relativePreference !== 1 /* NonRelative */ || !ts.pathIsRelative(specifier) ?
                            specifier :
                            undefined;
                    });
                });
                if (existingSpecifier) {
                    var moduleSpecifiers_2 = [existingSpecifier];
                    return moduleSpecifiers_2;
                }
                var importedFileIsInNodeModules = ts.some(modulePaths, function (p) { return p.isInNodeModules; });
                // Module specifier priority:
                //   1. "Bare package specifiers" (e.g. "@foo/bar") resulting from a path through node_modules to a package.json's "types" entry
                //   2. Specifiers generated using "paths" from tsconfig
                //   3. Non-relative specfiers resulting from a path through node_modules (e.g. "@foo/bar/path/to/file")
                //   4. Relative paths
                var nodeModulesSpecifiers;
                var pathsSpecifiers;
                var relativeSpecifiers;
                for (var _i = 0, modulePaths_1 = modulePaths; _i < modulePaths_1.length; _i++) {
                    var modulePath = modulePaths_1[_i];
                    var specifier = tryGetModuleNameAsNodeModule(modulePath, info, host, compilerOptions);
                    nodeModulesSpecifiers = ts.append(nodeModulesSpecifiers, specifier);
                    if (specifier && modulePath.isRedirect) {
                        // If we got a specifier for a redirect, it was a bare package specifier (e.g. "@foo/bar",
                        // not "@foo/bar/path/to/file"). No other specifier will be this good, so stop looking.
                        return nodeModulesSpecifiers;
                    }
                    if (!specifier && !modulePath.isRedirect) {
                        var local = getLocalModuleSpecifier(modulePath.path, info, compilerOptions, host, preferences);
                        if (ts.pathIsBareSpecifier(local)) {
                            pathsSpecifiers = ts.append(pathsSpecifiers, local);
                        }
                        else if (!importedFileIsInNodeModules || modulePath.isInNodeModules) {
                            // Why this extra conditional, not just an `else`? If some path to the file contained
                            // 'node_modules', but we can't create a non-relative specifier (e.g. "@foo/bar/path/to/file"),
                            // that means we had to go through a *sibling's* node_modules, not one we can access directly.
                            // If some path to the file was in node_modules but another was not, this likely indicates that
                            // we have a monorepo structure with symlinks. In this case, the non-node_modules path is
                            // probably the realpath, e.g. "../bar/path/to/file", but a relative path to another package
                            // in a monorepo is probably not portable. So, the module specifier we actually go with will be
                            // the relative path through node_modules, so that the declaration emitter can produce a
                            // portability error. (See declarationEmitReexportedSymlinkReference3)
                            relativeSpecifiers = ts.append(relativeSpecifiers, local);
                        }
                    }
                }
                return (pathsSpecifiers === null || pathsSpecifiers === void 0 ? void 0 : pathsSpecifiers.length) ? pathsSpecifiers :
                    (nodeModulesSpecifiers === null || nodeModulesSpecifiers === void 0 ? void 0 : nodeModulesSpecifiers.length) ? nodeModulesSpecifiers :
                        ts.Debug.checkDefined(relativeSpecifiers);
            }
            // importingSourceFileName is separate because getEditsForFileRename may need to specify an updated path
            function getInfo(importingSourceFileName, host) {
                var getCanonicalFileName = ts.createGetCanonicalFileName(host.useCaseSensitiveFileNames ? host.useCaseSensitiveFileNames() : true);
                var sourceDirectory = ts.getDirectoryPath(importingSourceFileName);
                return { getCanonicalFileName: getCanonicalFileName, importingSourceFileName: importingSourceFileName, sourceDirectory: sourceDirectory };
            }
            function getLocalModuleSpecifier(moduleFileName, info, compilerOptions, host, _a) {
                var ending = _a.ending, relativePreference = _a.relativePreference;
                var baseUrl = compilerOptions.baseUrl, paths = compilerOptions.paths, rootDirs = compilerOptions.rootDirs;
                var sourceDirectory = info.sourceDirectory, getCanonicalFileName = info.getCanonicalFileName;
                var relativePath = rootDirs && tryGetModuleNameFromRootDirs(rootDirs, moduleFileName, sourceDirectory, getCanonicalFileName, ending, compilerOptions) ||
                    removeExtensionAndIndexPostFix(ts.ensurePathIsNonModuleName(ts.getRelativePathFromDirectory(sourceDirectory, moduleFileName, getCanonicalFileName)), ending, compilerOptions);
                if (!baseUrl && !paths || relativePreference === 0 /* Relative */) {
                    return relativePath;
                }
                var baseDirectory = ts.getNormalizedAbsolutePath(ts.getPathsBasePath(compilerOptions, host) || baseUrl, host.getCurrentDirectory());
                var relativeToBaseUrl = getRelativePathIfInDirectory(moduleFileName, baseDirectory, getCanonicalFileName);
                if (!relativeToBaseUrl) {
                    return relativePath;
                }
                var importRelativeToBaseUrl = removeExtensionAndIndexPostFix(relativeToBaseUrl, ending, compilerOptions);
                var fromPaths = paths && tryGetModuleNameFromPaths(ts.removeFileExtension(relativeToBaseUrl), importRelativeToBaseUrl, paths);
                var nonRelative = fromPaths === undefined && baseUrl !== undefined ? importRelativeToBaseUrl : fromPaths;
                if (!nonRelative) {
                    return relativePath;
                }
                if (relativePreference === 1 /* NonRelative */) {
                    return nonRelative;
                }
                if (relativePreference === 3 /* ExternalNonRelative */) {
                    var projectDirectory = compilerOptions.configFilePath ?
                        ts.toPath(ts.getDirectoryPath(compilerOptions.configFilePath), host.getCurrentDirectory(), info.getCanonicalFileName) :
                        info.getCanonicalFileName(host.getCurrentDirectory());
                    var modulePath = ts.toPath(moduleFileName, projectDirectory, getCanonicalFileName);
                    var sourceIsInternal = ts.startsWith(sourceDirectory, projectDirectory);
                    var targetIsInternal = ts.startsWith(modulePath, projectDirectory);
                    if (sourceIsInternal && !targetIsInternal || !sourceIsInternal && targetIsInternal) {
                        // 1. The import path crosses the boundary of the tsconfig.json-containing directory.
                        //
                        //      src/
                        //        tsconfig.json
                        //        index.ts -------
                        //      lib/              | (path crosses tsconfig.json)
                        //        imported.ts <---
                        //
                        return nonRelative;
                    }
                    var nearestTargetPackageJson = getNearestAncestorDirectoryWithPackageJson(host, ts.getDirectoryPath(modulePath));
                    var nearestSourcePackageJson = getNearestAncestorDirectoryWithPackageJson(host, sourceDirectory);
                    if (nearestSourcePackageJson !== nearestTargetPackageJson) {
                        // 2. The importing and imported files are part of different packages.
                        //
                        //      packages/a/
                        //        package.json
                        //        index.ts --------
                        //      packages/b/        | (path crosses package.json)
                        //        package.json     |
                        //        component.ts <---
                        //
                        return nonRelative;
                    }
                    return relativePath;
                }
                if (relativePreference !== 2 /* Shortest */)
                    ts.Debug.assertNever(relativePreference);
                // Prefer a relative import over a baseUrl import if it has fewer components.
                return isPathRelativeToParent(nonRelative) || countPathComponents(relativePath) < countPathComponents(nonRelative) ? relativePath : nonRelative;
            }
            function countPathComponents(path) {
                var count = 0;
                for (var i = ts.startsWith(path, "./") ? 2 : 0; i < path.length; i++) {
                    if (path.charCodeAt(i) === 47 /* slash */)
                        count++;
                }
                return count;
            }
            moduleSpecifiers_1.countPathComponents = countPathComponents;
            function usesJsExtensionOnImports(_a) {
                var imports = _a.imports;
                return ts.firstDefined(imports, function (_a) {
                    var text = _a.text;
                    return ts.pathIsRelative(text) ? ts.hasJSFileExtension(text) : undefined;
                }) || false;
            }
            function comparePathsByRedirectAndNumberOfDirectorySeparators(a, b) {
                return ts.compareBooleans(b.isRedirect, a.isRedirect) || ts.compareNumberOfDirectorySeparators(a.path, b.path);
            }
            function getNearestAncestorDirectoryWithPackageJson(host, fileName) {
                if (host.getNearestAncestorDirectoryWithPackageJson) {
                    return host.getNearestAncestorDirectoryWithPackageJson(fileName);
                }
                return !!ts.forEachAncestorDirectory(fileName, function (directory) {
                    return host.fileExists(ts.combinePaths(directory, "package.json")) ? true : undefined;
                });
            }
            function forEachFileNameOfModule(importingFileName, importedFileName, host, preferSymlinks, cb) {
                var _a;
                var getCanonicalFileName = ts.hostGetCanonicalFileName(host);
                var cwd = host.getCurrentDirectory();
                var referenceRedirect = host.isSourceOfProjectReferenceRedirect(importedFileName) ? host.getProjectReferenceRedirect(importedFileName) : undefined;
                var importedPath = ts.toPath(importedFileName, cwd, getCanonicalFileName);
                var redirects = host.redirectTargetsMap.get(importedPath) || ts.emptyArray;
                var importedFileNames = __spreadArray(__spreadArray(__spreadArray([], (referenceRedirect ? [referenceRedirect] : ts.emptyArray), true), [importedFileName], false), redirects, true);
                var targets = importedFileNames.map(function (f) { return ts.getNormalizedAbsolutePath(f, cwd); });
                var shouldFilterIgnoredPaths = !ts.every(targets, ts.containsIgnoredPath);
                if (!preferSymlinks) {
                    // Symlinks inside ignored paths are already filtered out of the symlink cache,
                    // so we only need to remove them from the realpath filenames.
                    var result_15 = ts.forEach(targets, function (p) { return !(shouldFilterIgnoredPaths && ts.containsIgnoredPath(p)) && cb(p, referenceRedirect === p); });
                    if (result_15)
                        return result_15;
                }
                var symlinkedDirectories = (_a = host.getSymlinkCache) === null || _a === void 0 ? void 0 : _a.call(host).getSymlinkedDirectoriesByRealpath();
                var fullImportedFileName = ts.getNormalizedAbsolutePath(importedFileName, cwd);
                var result = symlinkedDirectories && ts.forEachAncestorDirectory(ts.getDirectoryPath(fullImportedFileName), function (realPathDirectory) {
                    var symlinkDirectories = symlinkedDirectories.get(ts.ensureTrailingDirectorySeparator(ts.toPath(realPathDirectory, cwd, getCanonicalFileName)));
                    if (!symlinkDirectories)
                        return undefined; // Continue to ancestor directory
                    // Don't want to a package to globally import from itself (importNameCodeFix_symlink_own_package.ts)
                    if (ts.startsWithDirectory(importingFileName, realPathDirectory, getCanonicalFileName)) {
                        return false; // Stop search, each ancestor directory will also hit this condition
                    }
                    return ts.forEach(targets, function (target) {
                        if (!ts.startsWithDirectory(target, realPathDirectory, getCanonicalFileName)) {
                            return;
                        }
                        var relative = ts.getRelativePathFromDirectory(realPathDirectory, target, getCanonicalFileName);
                        for (var _i = 0, symlinkDirectories_1 = symlinkDirectories; _i < symlinkDirectories_1.length; _i++) {
                            var symlinkDirectory = symlinkDirectories_1[_i];
                            var option = ts.resolvePath(symlinkDirectory, relative);
                            var result_16 = cb(option, target === referenceRedirect);
                            shouldFilterIgnoredPaths = true; // We found a non-ignored path in symlinks, so we can reject ignored-path realpaths
                            if (result_16)
                                return result_16;
                        }
                    });
                });
                return result || (preferSymlinks
                    ? ts.forEach(targets, function (p) { return shouldFilterIgnoredPaths && ts.containsIgnoredPath(p) ? undefined : cb(p, p === referenceRedirect); })
                    : undefined);
            }
            moduleSpecifiers_1.forEachFileNameOfModule = forEachFileNameOfModule;
            /**
             * Looks for existing imports that use symlinks to this module.
             * Symlinks will be returned first so they are preferred over the real path.
             */
            function getAllModulePaths(importingFilePath, importedFileName, host, preferences, importedFilePath) {
                var _a;
                if (importedFilePath === void 0) {
                    importedFilePath = ts.toPath(importedFileName, host.getCurrentDirectory(), ts.hostGetCanonicalFileName(host));
                }
                var cache = (_a = host.getModuleSpecifierCache) === null || _a === void 0 ? void 0 : _a.call(host);
                if (cache) {
                    var cached = cache.get(importingFilePath, importedFilePath, preferences);
                    if (cached === null || cached === void 0 ? void 0 : cached.modulePaths)
                        return cached.modulePaths;
                }
                var modulePaths = getAllModulePathsWorker(importingFilePath, importedFileName, host);
                if (cache) {
                    cache.setModulePaths(importingFilePath, importedFilePath, preferences, modulePaths);
                }
                return modulePaths;
            }
            function getAllModulePathsWorker(importingFileName, importedFileName, host) {
                var getCanonicalFileName = ts.hostGetCanonicalFileName(host);
                var allFileNames = new ts.Map();
                var importedFileFromNodeModules = false;
                forEachFileNameOfModule(importingFileName, importedFileName, host, 
                /*preferSymlinks*/ true, function (path, isRedirect) {
                    var isInNodeModules = ts.pathContainsNodeModules(path);
                    allFileNames.set(path, { path: getCanonicalFileName(path), isRedirect: isRedirect, isInNodeModules: isInNodeModules });
                    importedFileFromNodeModules = importedFileFromNodeModules || isInNodeModules;
                    // don't return value, so we collect everything
                });
                // Sort by paths closest to importing file Name directory
                var sortedPaths = [];
                var _loop_31 = function (directory) {
                    var directoryStart = ts.ensureTrailingDirectorySeparator(directory);
                    var pathsInDirectory;
                    allFileNames.forEach(function (_a, fileName) {
                        var path = _a.path, isRedirect = _a.isRedirect, isInNodeModules = _a.isInNodeModules;
                        if (ts.startsWith(path, directoryStart)) {
                            (pathsInDirectory || (pathsInDirectory = [])).push({ path: fileName, isRedirect: isRedirect, isInNodeModules: isInNodeModules });
                            allFileNames.delete(fileName);
                        }
                    });
                    if (pathsInDirectory) {
                        if (pathsInDirectory.length > 1) {
                            pathsInDirectory.sort(comparePathsByRedirectAndNumberOfDirectorySeparators);
                        }
                        sortedPaths.push.apply(sortedPaths, pathsInDirectory);
                    }
                    var newDirectory = ts.getDirectoryPath(directory);
                    if (newDirectory === directory)
                        return out_directory_1 = directory, "break";
                    directory = newDirectory;
                    out_directory_1 = directory;
                };
                var out_directory_1;
                for (var directory = ts.getDirectoryPath(importingFileName); allFileNames.size !== 0;) {
                    var state_9 = _loop_31(directory);
                    directory = out_directory_1;
                    if (state_9 === "break")
                        break;
                }
                if (allFileNames.size) {
                    var remainingPaths = ts.arrayFrom(allFileNames.values());
                    if (remainingPaths.length > 1)
                        remainingPaths.sort(comparePathsByRedirectAndNumberOfDirectorySeparators);
                    sortedPaths.push.apply(sortedPaths, remainingPaths);
                }
                return sortedPaths;
            }
            function tryGetModuleNameFromAmbientModule(moduleSymbol, checker) {
                var _a;
                var decl = (_a = moduleSymbol.declarations) === null || _a === void 0 ? void 0 : _a.find(function (d) { return ts.isNonGlobalAmbientModule(d) && (!ts.isExternalModuleAugmentation(d) || !ts.isExternalModuleNameRelative(ts.getTextOfIdentifierOrLiteral(d.name))); });
                if (decl) {
                    return decl.name.text;
                }
                // the module could be a namespace, which is export through "export=" from an ambient module.
                /**
                 * declare module "m" {
                 *     namespace ns {
                 *         class c {}
                 *     }
                 *     export = ns;
                 * }
                 */
                // `import {c} from "m";` is valid, in which case, `moduleSymbol` is "ns", but the module name should be "m"
                var ambientModuleDeclareCandidates = ts.mapDefined(moduleSymbol.declarations, function (d) {
                    var _a, _b, _c, _d;
                    if (!ts.isModuleDeclaration(d))
                        return;
                    var topNamespace = getTopNamespace(d);
                    if (!(((_a = topNamespace === null || topNamespace === void 0 ? void 0 : topNamespace.parent) === null || _a === void 0 ? void 0 : _a.parent)
                        && ts.isModuleBlock(topNamespace.parent) && ts.isAmbientModule(topNamespace.parent.parent) && ts.isSourceFile(topNamespace.parent.parent.parent)))
                        return;
                    var exportAssignment = (_d = (_c = (_b = topNamespace.parent.parent.symbol.exports) === null || _b === void 0 ? void 0 : _b.get("export=")) === null || _c === void 0 ? void 0 : _c.valueDeclaration) === null || _d === void 0 ? void 0 : _d.expression;
                    if (!exportAssignment)
                        return;
                    var exportSymbol = checker.getSymbolAtLocation(exportAssignment);
                    if (!exportSymbol)
                        return;
                    var originalExportSymbol = (exportSymbol === null || exportSymbol === void 0 ? void 0 : exportSymbol.flags) & 2097152 /* Alias */ ? checker.getAliasedSymbol(exportSymbol) : exportSymbol;
                    if (originalExportSymbol === d.symbol)
                        return topNamespace.parent.parent;
                    function getTopNamespace(namespaceDeclaration) {
                        while (namespaceDeclaration.flags & 4 /* NestedNamespace */) {
                            namespaceDeclaration = namespaceDeclaration.parent;
                        }
                        return namespaceDeclaration;
                    }
                });
                var ambientModuleDeclare = ambientModuleDeclareCandidates[0];
                if (ambientModuleDeclare) {
                    return ambientModuleDeclare.name.text;
                }
            }
            function tryGetModuleNameFromPaths(relativeToBaseUrlWithIndex, relativeToBaseUrl, paths) {
                for (var key in paths) {
                    for (var _i = 0, _a = paths[key]; _i < _a.length; _i++) {
                        var patternText_1 = _a[_i];
                        var pattern = ts.removeFileExtension(ts.normalizePath(patternText_1));
                        var indexOfStar = pattern.indexOf("*");
                        if (indexOfStar !== -1) {
                            var prefix = pattern.substr(0, indexOfStar);
                            var suffix = pattern.substr(indexOfStar + 1);
                            if (relativeToBaseUrl.length >= prefix.length + suffix.length &&
                                ts.startsWith(relativeToBaseUrl, prefix) &&
                                ts.endsWith(relativeToBaseUrl, suffix) ||
                                !suffix && relativeToBaseUrl === ts.removeTrailingDirectorySeparator(prefix)) {
                                var matchedStar = relativeToBaseUrl.substr(prefix.length, relativeToBaseUrl.length - suffix.length - prefix.length);
                                return key.replace("*", matchedStar);
                            }
                        }
                        else if (pattern === relativeToBaseUrl || pattern === relativeToBaseUrlWithIndex) {
                            return key;
                        }
                    }
                }
            }
            function tryGetModuleNameFromRootDirs(rootDirs, moduleFileName, sourceDirectory, getCanonicalFileName, ending, compilerOptions) {
                var normalizedTargetPath = getPathRelativeToRootDirs(moduleFileName, rootDirs, getCanonicalFileName);
                if (normalizedTargetPath === undefined) {
                    return undefined;
                }
                var normalizedSourcePath = getPathRelativeToRootDirs(sourceDirectory, rootDirs, getCanonicalFileName);
                var relativePath = normalizedSourcePath !== undefined ? ts.ensurePathIsNonModuleName(ts.getRelativePathFromDirectory(normalizedSourcePath, normalizedTargetPath, getCanonicalFileName)) : normalizedTargetPath;
                return ts.getEmitModuleResolutionKind(compilerOptions) === ts.ModuleResolutionKind.NodeJs
                    ? removeExtensionAndIndexPostFix(relativePath, ending, compilerOptions)
                    : ts.removeFileExtension(relativePath);
            }
            function tryGetModuleNameAsNodeModule(_a, _b, host, options, packageNameOnly) {
                var path = _a.path, isRedirect = _a.isRedirect;
                var getCanonicalFileName = _b.getCanonicalFileName, sourceDirectory = _b.sourceDirectory;
                if (!host.fileExists || !host.readFile) {
                    return undefined;
                }
                var parts = getNodeModulePathParts(path);
                if (!parts) {
                    return undefined;
                }
                // Simplify the full file path to something that can be resolved by Node.
                var moduleSpecifier = path;
                var isPackageRootPath = false;
                if (!packageNameOnly) {
                    var packageRootIndex = parts.packageRootIndex;
                    var moduleFileNameForExtensionless = void 0;
                    while (true) {
                        // If the module could be imported by a directory name, use that directory's name
                        var _c = tryDirectoryWithPackageJson(packageRootIndex), moduleFileToTry = _c.moduleFileToTry, packageRootPath = _c.packageRootPath;
                        if (packageRootPath) {
                            moduleSpecifier = packageRootPath;
                            isPackageRootPath = true;
                            break;
                        }
                        if (!moduleFileNameForExtensionless)
                            moduleFileNameForExtensionless = moduleFileToTry;
                        // try with next level of directory
                        packageRootIndex = path.indexOf(ts.directorySeparator, packageRootIndex + 1);
                        if (packageRootIndex === -1) {
                            moduleSpecifier = getExtensionlessFileName(moduleFileNameForExtensionless);
                            break;
                        }
                    }
                }
                if (isRedirect && !isPackageRootPath) {
                    return undefined;
                }
                var globalTypingsCacheLocation = host.getGlobalTypingsCacheLocation && host.getGlobalTypingsCacheLocation();
                // Get a path that's relative to node_modules or the importing file's path
                // if node_modules folder is in this folder or any of its parent folders, no need to keep it.
                var pathToTopLevelNodeModules = getCanonicalFileName(moduleSpecifier.substring(0, parts.topLevelNodeModulesIndex));
                if (!(ts.startsWith(sourceDirectory, pathToTopLevelNodeModules) || globalTypingsCacheLocation && ts.startsWith(getCanonicalFileName(globalTypingsCacheLocation), pathToTopLevelNodeModules))) {
                    return undefined;
                }
                // If the module was found in @types, get the actual Node package name
                var nodeModulesDirectoryName = moduleSpecifier.substring(parts.topLevelPackageNameIndex + 1);
                var packageName = ts.getPackageNameFromTypesPackageName(nodeModulesDirectoryName);
                // For classic resolution, only allow importing from node_modules/@types, not other node_modules
                return ts.getEmitModuleResolutionKind(options) !== ts.ModuleResolutionKind.NodeJs && packageName === nodeModulesDirectoryName ? undefined : packageName;
                function tryDirectoryWithPackageJson(packageRootIndex) {
                    var packageRootPath = path.substring(0, packageRootIndex);
                    var packageJsonPath = ts.combinePaths(packageRootPath, "package.json");
                    var moduleFileToTry = path;
                    if (host.fileExists(packageJsonPath)) {
                        var packageJsonContent = JSON.parse(host.readFile(packageJsonPath));
                        var versionPaths = packageJsonContent.typesVersions
                            ? ts.getPackageJsonTypesVersionsPaths(packageJsonContent.typesVersions)
                            : undefined;
                        if (versionPaths) {
                            var subModuleName = path.slice(packageRootPath.length + 1);
                            var fromPaths = tryGetModuleNameFromPaths(ts.removeFileExtension(subModuleName), removeExtensionAndIndexPostFix(subModuleName, 0 /* Minimal */, options), versionPaths.paths);
                            if (fromPaths !== undefined) {
                                moduleFileToTry = ts.combinePaths(packageRootPath, fromPaths);
                            }
                        }
                        // If the file is the main module, it can be imported by the package name
                        var mainFileRelative = packageJsonContent.typings || packageJsonContent.types || packageJsonContent.main;
                        if (ts.isString(mainFileRelative)) {
                            var mainExportFile = ts.toPath(mainFileRelative, packageRootPath, getCanonicalFileName);
                            if (ts.removeFileExtension(mainExportFile) === ts.removeFileExtension(getCanonicalFileName(moduleFileToTry))) {
                                return { packageRootPath: packageRootPath, moduleFileToTry: moduleFileToTry };
                            }
                        }
                    }
                    return { moduleFileToTry: moduleFileToTry };
                }
                function getExtensionlessFileName(path) {
                    // We still have a file name - remove the extension
                    var fullModulePathWithoutExtension = ts.removeFileExtension(path);
                    // If the file is /index, it can be imported by its directory name
                    // IFF there is not _also_ a file by the same name
                    if (getCanonicalFileName(fullModulePathWithoutExtension.substring(parts.fileNameIndex)) === "/index" && !tryGetAnyFileFromPath(host, fullModulePathWithoutExtension.substring(0, parts.fileNameIndex))) {
                        return fullModulePathWithoutExtension.substring(0, parts.fileNameIndex);
                    }
                    return fullModulePathWithoutExtension;
                }
            }
            function tryGetAnyFileFromPath(host, path) {
                if (!host.fileExists)
                    return;
                // We check all js, `node` and `json` extensions in addition to TS, since node module resolution would also choose those over the directory
                var extensions = ts.getSupportedExtensions({ allowJs: true }, [{ extension: "node", isMixedContent: false }, { extension: "json", isMixedContent: false, scriptKind: 6 /* JSON */ }]);
                for (var _i = 0, extensions_3 = extensions; _i < extensions_3.length; _i++) {
                    var e = extensions_3[_i];
                    var fullPath = path + e;
                    if (host.fileExists(fullPath)) {
                        return fullPath;
                    }
                }
            }
            function getNodeModulePathParts(fullPath) {
                // If fullPath can't be valid module file within node_modules, returns undefined.
                // Example of expected pattern: /base/path/node_modules/[@scope/otherpackage/@otherscope/node_modules/]package/[subdirectory/]file.js
                // Returns indices:                       ^            ^                                                      ^             ^
                var topLevelNodeModulesIndex = 0;
                var topLevelPackageNameIndex = 0;
                var packageRootIndex = 0;
                var fileNameIndex = 0;
                var States;
                (function (States) {
                    States[States["BeforeNodeModules"] = 0] = "BeforeNodeModules";
                    States[States["NodeModules"] = 1] = "NodeModules";
                    States[States["Scope"] = 2] = "Scope";
                    States[States["PackageContent"] = 3] = "PackageContent";
                })(States || (States = {}));
                var partStart = 0;
                var partEnd = 0;
                var state = 0 /* BeforeNodeModules */;
                while (partEnd >= 0) {
                    partStart = partEnd;
                    partEnd = fullPath.indexOf("/", partStart + 1);
                    switch (state) {
                        case 0 /* BeforeNodeModules */:
                            if (fullPath.indexOf(ts.nodeModulesPathPart, partStart) === partStart) {
                                topLevelNodeModulesIndex = partStart;
                                topLevelPackageNameIndex = partEnd;
                                state = 1 /* NodeModules */;
                            }
                            break;
                        case 1 /* NodeModules */:
                        case 2 /* Scope */:
                            if (state === 1 /* NodeModules */ && fullPath.charAt(partStart + 1) === "@") {
                                state = 2 /* Scope */;
                            }
                            else {
                                packageRootIndex = partEnd;
                                state = 3 /* PackageContent */;
                            }
                            break;
                        case 3 /* PackageContent */:
                            if (fullPath.indexOf(ts.nodeModulesPathPart, partStart) === partStart) {
                                state = 1 /* NodeModules */;
                            }
                            else {
                                state = 3 /* PackageContent */;
                            }
                            break;
                    }
                }
                fileNameIndex = partStart;
                return state > 1 /* NodeModules */ ? { topLevelNodeModulesIndex: topLevelNodeModulesIndex, topLevelPackageNameIndex: topLevelPackageNameIndex, packageRootIndex: packageRootIndex, fileNameIndex: fileNameIndex } : undefined;
            }
            function getPathRelativeToRootDirs(path, rootDirs, getCanonicalFileName) {
                return ts.firstDefined(rootDirs, function (rootDir) {
                    var relativePath = getRelativePathIfInDirectory(path, rootDir, getCanonicalFileName); // TODO: GH#18217
                    return isPathRelativeToParent(relativePath) ? undefined : relativePath;
                });
            }
            function removeExtensionAndIndexPostFix(fileName, ending, options) {
                if (ts.fileExtensionIs(fileName, ".json" /* Json */))
                    return fileName;
                var noExtension = ts.removeFileExtension(fileName);
                switch (ending) {
                    case 0 /* Minimal */:
                        return ts.removeSuffix(noExtension, "/index");
                    case 1 /* Index */:
                        return noExtension;
                    case 2 /* JsExtension */:
                        return noExtension + getJSExtensionForFile(fileName, options);
                    default:
                        return ts.Debug.assertNever(ending);
                }
            }
            function getJSExtensionForFile(fileName, options) {
                var _a;
                return (_a = tryGetJSExtensionForFile(fileName, options)) !== null && _a !== void 0 ? _a : ts.Debug.fail("Extension " + ts.extensionFromPath(fileName) + " is unsupported:: FileName:: " + fileName);
            }
            function tryGetJSExtensionForFile(fileName, options) {
                var ext = ts.tryGetExtensionFromPath(fileName);
                switch (ext) {
                    case ".ts" /* Ts */:
                    case ".d.ts" /* Dts */:
                        return ".js" /* Js */;
                    case ".tsx" /* Tsx */:
                        return options.jsx === 1 /* Preserve */ ? ".jsx" /* Jsx */ : ".js" /* Js */;
                    case ".js" /* Js */:
                    case ".jsx" /* Jsx */:
                    case ".json" /* Json */:
                        return ext;
                    default:
                        return undefined;
                }
            }
            moduleSpecifiers_1.tryGetJSExtensionForFile = tryGetJSExtensionForFile;
            function getRelativePathIfInDirectory(path, directoryPath, getCanonicalFileName) {
                var relativePath = ts.getRelativePathToDirectoryOrUrl(directoryPath, path, directoryPath, getCanonicalFileName, /*isAbsolutePathAnUrl*/ false);
                return ts.isRootedDiskPath(relativePath) ? undefined : relativePath;
            }
            function isPathRelativeToParent(path) {
                return ts.startsWith(path, "..");
            }
        })(moduleSpecifiers = ts.moduleSpecifiers || (ts.moduleSpecifiers = {}));
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var sysFormatDiagnosticsHost = ts.sys ? {
            getCurrentDirectory: function () { return ts.sys.getCurrentDirectory(); },
            getNewLine: function () { return ts.sys.newLine; },
            getCanonicalFileName: ts.createGetCanonicalFileName(ts.sys.useCaseSensitiveFileNames)
        } : undefined;
        /**
         * Create a function that reports error by writing to the system and handles the formating of the diagnostic
         */
        function createDiagnosticReporter(system, pretty) {
            var host = system === ts.sys && sysFormatDiagnosticsHost ? sysFormatDiagnosticsHost : {
                getCurrentDirectory: function () { return system.getCurrentDirectory(); },
                getNewLine: function () { return system.newLine; },
                getCanonicalFileName: ts.createGetCanonicalFileName(system.useCaseSensitiveFileNames),
            };
            if (!pretty) {
                return function (diagnostic) { return system.write(ts.formatDiagnostic(diagnostic, host)); };
            }
            var diagnostics = new Array(1);
            return function (diagnostic) {
                diagnostics[0] = diagnostic;
                system.write(ts.formatDiagnosticsWithColorAndContext(diagnostics, host) + host.getNewLine());
                diagnostics[0] = undefined; // TODO: GH#18217
            };
        }
        ts.createDiagnosticReporter = createDiagnosticReporter;
        /**
         * @returns Whether the screen was cleared.
         */
        function clearScreenIfNotWatchingForFileChanges(system, diagnostic, options) {
            if (system.clearScreen &&
                !options.preserveWatchOutput &&
                !options.extendedDiagnostics &&
                !options.diagnostics &&
                ts.contains(ts.screenStartingMessageCodes, diagnostic.code)) {
                system.clearScreen();
                return true;
            }
            return false;
        }
        ts.screenStartingMessageCodes = [
            ts.Diagnostics.Starting_compilation_in_watch_mode.code,
            ts.Diagnostics.File_change_detected_Starting_incremental_compilation.code,
        ];
        function getPlainDiagnosticFollowingNewLines(diagnostic, newLine) {
            return ts.contains(ts.screenStartingMessageCodes, diagnostic.code)
                ? newLine + newLine
                : newLine;
        }
        /**
         * Get locale specific time based on whether we are in test mode
         */
        function getLocaleTimeString(system) {
            return !system.now ?
                new Date().toLocaleTimeString() :
                system.now().toLocaleTimeString("en-US", { timeZone: "UTC" });
        }
        ts.getLocaleTimeString = getLocaleTimeString;
        /**
         * Create a function that reports watch status by writing to the system and handles the formating of the diagnostic
         */
        function createWatchStatusReporter(system, pretty) {
            return pretty ?
                function (diagnostic, newLine, options) {
                    clearScreenIfNotWatchingForFileChanges(system, diagnostic, options);
                    var output = "[" + ts.formatColorAndReset(getLocaleTimeString(system), ts.ForegroundColorEscapeSequences.Grey) + "] ";
                    output += "" + ts.flattenDiagnosticMessageText(diagnostic.messageText, system.newLine) + (newLine + newLine);
                    system.write(output);
                } :
                function (diagnostic, newLine, options) {
                    var output = "";
                    if (!clearScreenIfNotWatchingForFileChanges(system, diagnostic, options)) {
                        output += newLine;
                    }
                    output += getLocaleTimeString(system) + " - ";
                    output += "" + ts.flattenDiagnosticMessageText(diagnostic.messageText, system.newLine) + getPlainDiagnosticFollowingNewLines(diagnostic, newLine);
                    system.write(output);
                };
        }
        ts.createWatchStatusReporter = createWatchStatusReporter;
        /** Parses config file using System interface */
        function parseConfigFileWithSystem(configFileName, optionsToExtend, extendedConfigCache, watchOptionsToExtend, system, reportDiagnostic) {
            var host = system;
            host.onUnRecoverableConfigFileDiagnostic = function (diagnostic) { return reportUnrecoverableDiagnostic(system, reportDiagnostic, diagnostic); };
            var result = ts.getParsedCommandLineOfConfigFile(configFileName, optionsToExtend, host, extendedConfigCache, watchOptionsToExtend);
            host.onUnRecoverableConfigFileDiagnostic = undefined; // TODO: GH#18217
            return result;
        }
        ts.parseConfigFileWithSystem = parseConfigFileWithSystem;
        function getErrorCountForSummary(diagnostics) {
            return ts.countWhere(diagnostics, function (diagnostic) { return diagnostic.category === ts.DiagnosticCategory.Error; });
        }
        ts.getErrorCountForSummary = getErrorCountForSummary;
        function getWatchErrorSummaryDiagnosticMessage(errorCount) {
            return errorCount === 1 ?
                ts.Diagnostics.Found_1_error_Watching_for_file_changes :
                ts.Diagnostics.Found_0_errors_Watching_for_file_changes;
        }
        ts.getWatchErrorSummaryDiagnosticMessage = getWatchErrorSummaryDiagnosticMessage;
        function getErrorSummaryText(errorCount, newLine) {
            if (errorCount === 0)
                return "";
            var d = ts.createCompilerDiagnostic(errorCount === 1 ? ts.Diagnostics.Found_1_error : ts.Diagnostics.Found_0_errors, errorCount);
            return "" + newLine + ts.flattenDiagnosticMessageText(d.messageText, newLine) + newLine + newLine;
        }
        ts.getErrorSummaryText = getErrorSummaryText;
        function isBuilderProgram(program) {
            return !!program.getState;
        }
        ts.isBuilderProgram = isBuilderProgram;
        function listFiles(program, write) {
            var options = program.getCompilerOptions();
            if (options.explainFiles) {
                explainFiles(isBuilderProgram(program) ? program.getProgram() : program, write);
            }
            else if (options.listFiles || options.listFilesOnly) {
                ts.forEach(program.getSourceFiles(), function (file) {
                    write(file.fileName);
                });
            }
        }
        ts.listFiles = listFiles;
        function explainFiles(program, write) {
            var _a, _b;
            var reasons = program.getFileIncludeReasons();
            var getCanonicalFileName = ts.createGetCanonicalFileName(program.useCaseSensitiveFileNames());
            var relativeFileName = function (fileName) { return ts.convertToRelativePath(fileName, program.getCurrentDirectory(), getCanonicalFileName); };
            for (var _i = 0, _c = program.getSourceFiles(); _i < _c.length; _i++) {
                var file = _c[_i];
                write("" + toFileName(file, relativeFileName));
                (_a = reasons.get(file.path)) === null || _a === void 0 ? void 0 : _a.forEach(function (reason) { return write("  " + fileIncludeReasonToDiagnostics(program, reason, relativeFileName).messageText); });
                (_b = explainIfFileIsRedirect(file, relativeFileName)) === null || _b === void 0 ? void 0 : _b.forEach(function (d) { return write("  " + d.messageText); });
            }
        }
        ts.explainFiles = explainFiles;
        function explainIfFileIsRedirect(file, fileNameConvertor) {
            var result;
            if (file.path !== file.resolvedPath) {
                (result || (result = [])).push(ts.chainDiagnosticMessages(
                /*details*/ undefined, ts.Diagnostics.File_is_output_of_project_reference_source_0, toFileName(file.originalFileName, fileNameConvertor)));
            }
            if (file.redirectInfo) {
                (result || (result = [])).push(ts.chainDiagnosticMessages(
                /*details*/ undefined, ts.Diagnostics.File_redirects_to_file_0, toFileName(file.redirectInfo.redirectTarget, fileNameConvertor)));
            }
            return result;
        }
        ts.explainIfFileIsRedirect = explainIfFileIsRedirect;
        function getMatchedFileSpec(program, fileName) {
            var _a;
            var configFile = program.getCompilerOptions().configFile;
            if (!((_a = configFile === null || configFile === void 0 ? void 0 : configFile.configFileSpecs) === null || _a === void 0 ? void 0 : _a.validatedFilesSpec))
                return undefined;
            var getCanonicalFileName = ts.createGetCanonicalFileName(program.useCaseSensitiveFileNames());
            var filePath = getCanonicalFileName(fileName);
            var basePath = ts.getDirectoryPath(ts.getNormalizedAbsolutePath(configFile.fileName, program.getCurrentDirectory()));
            return ts.find(configFile.configFileSpecs.validatedFilesSpec, function (fileSpec) { return getCanonicalFileName(ts.getNormalizedAbsolutePath(fileSpec, basePath)) === filePath; });
        }
        ts.getMatchedFileSpec = getMatchedFileSpec;
        function getMatchedIncludeSpec(program, fileName) {
            var _a, _b;
            var configFile = program.getCompilerOptions().configFile;
            if (!((_a = configFile === null || configFile === void 0 ? void 0 : configFile.configFileSpecs) === null || _a === void 0 ? void 0 : _a.validatedIncludeSpecs))
                return undefined;
            var isJsonFile = ts.fileExtensionIs(fileName, ".json" /* Json */);
            var basePath = ts.getDirectoryPath(ts.getNormalizedAbsolutePath(configFile.fileName, program.getCurrentDirectory()));
            var useCaseSensitiveFileNames = program.useCaseSensitiveFileNames();
            return ts.find((_b = configFile === null || configFile === void 0 ? void 0 : configFile.configFileSpecs) === null || _b === void 0 ? void 0 : _b.validatedIncludeSpecs, function (includeSpec) {
                if (isJsonFile && !ts.endsWith(includeSpec, ".json" /* Json */))
                    return false;
                var pattern = ts.getPatternFromSpec(includeSpec, basePath, "files");
                return !!pattern && ts.getRegexFromPattern("(" + pattern + ")$", useCaseSensitiveFileNames).test(fileName);
            });
        }
        ts.getMatchedIncludeSpec = getMatchedIncludeSpec;
        function fileIncludeReasonToDiagnostics(program, reason, fileNameConvertor) {
            var _a, _b;
            var options = program.getCompilerOptions();
            if (ts.isReferencedFile(reason)) {
                var referenceLocation = ts.getReferencedFileLocation(function (path) { return program.getSourceFileByPath(path); }, reason);
                var referenceText = ts.isReferenceFileLocation(referenceLocation) ? referenceLocation.file.text.substring(referenceLocation.pos, referenceLocation.end) : "\"" + referenceLocation.text + "\"";
                var message = void 0;
                ts.Debug.assert(ts.isReferenceFileLocation(referenceLocation) || reason.kind === ts.FileIncludeKind.Import, "Only synthetic references are imports");
                switch (reason.kind) {
                    case ts.FileIncludeKind.Import:
                        if (ts.isReferenceFileLocation(referenceLocation)) {
                            message = referenceLocation.packageId ?
                                ts.Diagnostics.Imported_via_0_from_file_1_with_packageId_2 :
                                ts.Diagnostics.Imported_via_0_from_file_1;
                        }
                        else if (referenceLocation.text === ts.externalHelpersModuleNameText) {
                            message = referenceLocation.packageId ?
                                ts.Diagnostics.Imported_via_0_from_file_1_with_packageId_2_to_import_importHelpers_as_specified_in_compilerOptions :
                                ts.Diagnostics.Imported_via_0_from_file_1_to_import_importHelpers_as_specified_in_compilerOptions;
                        }
                        else {
                            message = referenceLocation.packageId ?
                                ts.Diagnostics.Imported_via_0_from_file_1_with_packageId_2_to_import_jsx_and_jsxs_factory_functions :
                                ts.Diagnostics.Imported_via_0_from_file_1_to_import_jsx_and_jsxs_factory_functions;
                        }
                        break;
                    case ts.FileIncludeKind.ReferenceFile:
                        ts.Debug.assert(!referenceLocation.packageId);
                        message = ts.Diagnostics.Referenced_via_0_from_file_1;
                        break;
                    case ts.FileIncludeKind.TypeReferenceDirective:
                        message = referenceLocation.packageId ?
                            ts.Diagnostics.Type_library_referenced_via_0_from_file_1_with_packageId_2 :
                            ts.Diagnostics.Type_library_referenced_via_0_from_file_1;
                        break;
                    case ts.FileIncludeKind.LibReferenceDirective:
                        ts.Debug.assert(!referenceLocation.packageId);
                        message = ts.Diagnostics.Library_referenced_via_0_from_file_1;
                        break;
                    default:
                        ts.Debug.assertNever(reason);
                }
                return ts.chainDiagnosticMessages(
                /*details*/ undefined, message, referenceText, toFileName(referenceLocation.file, fileNameConvertor), referenceLocation.packageId && ts.packageIdToString(referenceLocation.packageId));
            }
            switch (reason.kind) {
                case ts.FileIncludeKind.RootFile:
                    if (!((_a = options.configFile) === null || _a === void 0 ? void 0 : _a.configFileSpecs))
                        return ts.chainDiagnosticMessages(/*details*/ undefined, ts.Diagnostics.Root_file_specified_for_compilation);
                    var fileName = ts.getNormalizedAbsolutePath(program.getRootFileNames()[reason.index], program.getCurrentDirectory());
                    var matchedByFiles = getMatchedFileSpec(program, fileName);
                    if (matchedByFiles)
                        return ts.chainDiagnosticMessages(/*details*/ undefined, ts.Diagnostics.Part_of_files_list_in_tsconfig_json);
                    var matchedByInclude = getMatchedIncludeSpec(program, fileName);
                    return matchedByInclude ?
                        ts.chainDiagnosticMessages(
                        /*details*/ undefined, ts.Diagnostics.Matched_by_include_pattern_0_in_1, matchedByInclude, toFileName(options.configFile, fileNameConvertor)) :
                        // Could be additional files specified as roots
                        ts.chainDiagnosticMessages(/*details*/ undefined, ts.Diagnostics.Root_file_specified_for_compilation);
                case ts.FileIncludeKind.SourceFromProjectReference:
                case ts.FileIncludeKind.OutputFromProjectReference:
                    var isOutput = reason.kind === ts.FileIncludeKind.OutputFromProjectReference;
                    var referencedResolvedRef = ts.Debug.checkDefined((_b = program.getResolvedProjectReferences()) === null || _b === void 0 ? void 0 : _b[reason.index]);
                    return ts.chainDiagnosticMessages(
                    /*details*/ undefined, ts.outFile(options) ?
                        isOutput ?
                            ts.Diagnostics.Output_from_referenced_project_0_included_because_1_specified :
                            ts.Diagnostics.Source_from_referenced_project_0_included_because_1_specified :
                        isOutput ?
                            ts.Diagnostics.Output_from_referenced_project_0_included_because_module_is_specified_as_none :
                            ts.Diagnostics.Source_from_referenced_project_0_included_because_module_is_specified_as_none, toFileName(referencedResolvedRef.sourceFile.fileName, fileNameConvertor), options.outFile ? "--outFile" : "--out");
                case ts.FileIncludeKind.AutomaticTypeDirectiveFile:
                    return ts.chainDiagnosticMessages(
                    /*details*/ undefined, options.types ?
                        reason.packageId ?
                            ts.Diagnostics.Entry_point_of_type_library_0_specified_in_compilerOptions_with_packageId_1 :
                            ts.Diagnostics.Entry_point_of_type_library_0_specified_in_compilerOptions :
                        reason.packageId ?
                            ts.Diagnostics.Entry_point_for_implicit_type_library_0_with_packageId_1 :
                            ts.Diagnostics.Entry_point_for_implicit_type_library_0, reason.typeReference, reason.packageId && ts.packageIdToString(reason.packageId));
                case ts.FileIncludeKind.LibFile:
                    if (reason.index !== undefined)
                        return ts.chainDiagnosticMessages(/*details*/ undefined, ts.Diagnostics.Library_0_specified_in_compilerOptions, options.lib[reason.index]);
                    var target = ts.forEachEntry(ts.targetOptionDeclaration.type, function (value, key) { return value === options.target ? key : undefined; });
                    return ts.chainDiagnosticMessages(
                    /*details*/ undefined, target ?
                        ts.Diagnostics.Default_library_for_target_0 :
                        ts.Diagnostics.Default_library, target);
                default:
                    ts.Debug.assertNever(reason);
            }
        }
        ts.fileIncludeReasonToDiagnostics = fileIncludeReasonToDiagnostics;
        function toFileName(file, fileNameConvertor) {
            var fileName = ts.isString(file) ? file : file.fileName;
            return fileNameConvertor ? fileNameConvertor(fileName) : fileName;
        }
        /**
         * Helper that emit files, report diagnostics and lists emitted and/or source files depending on compiler options
         */
        function emitFilesAndReportErrors(program, reportDiagnostic, write, reportSummary, writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers) {
            var isListFilesOnly = !!program.getCompilerOptions().listFilesOnly;
            // First get and report any syntactic errors.
            var allDiagnostics = program.getConfigFileParsingDiagnostics().slice();
            var configFileParsingDiagnosticsLength = allDiagnostics.length;
            ts.addRange(allDiagnostics, program.getSyntacticDiagnostics(/*sourceFile*/ undefined, cancellationToken));
            // If we didn't have any syntactic errors, then also try getting the global and
            // semantic errors.
            if (allDiagnostics.length === configFileParsingDiagnosticsLength) {
                ts.addRange(allDiagnostics, program.getOptionsDiagnostics(cancellationToken));
                if (!isListFilesOnly) {
                    ts.addRange(allDiagnostics, program.getGlobalDiagnostics(cancellationToken));
                    if (allDiagnostics.length === configFileParsingDiagnosticsLength) {
                        ts.addRange(allDiagnostics, program.getSemanticDiagnostics(/*sourceFile*/ undefined, cancellationToken));
                    }
                }
            }
            // Emit and report any errors we ran into.
            var emitResult = isListFilesOnly
                ? { emitSkipped: true, diagnostics: ts.emptyArray }
                : program.emit(/*targetSourceFile*/ undefined, writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers);
            var emittedFiles = emitResult.emittedFiles, emitDiagnostics = emitResult.diagnostics;
            ts.addRange(allDiagnostics, emitDiagnostics);
            var diagnostics = ts.sortAndDeduplicateDiagnostics(allDiagnostics);
            diagnostics.forEach(reportDiagnostic);
            if (write) {
                var currentDir_1 = program.getCurrentDirectory();
                ts.forEach(emittedFiles, function (file) {
                    var filepath = ts.getNormalizedAbsolutePath(file, currentDir_1);
                    write("TSFILE: " + filepath);
                });
                listFiles(program, write);
            }
            if (reportSummary) {
                reportSummary(getErrorCountForSummary(diagnostics));
            }
            return {
                emitResult: emitResult,
                diagnostics: diagnostics,
            };
        }
        ts.emitFilesAndReportErrors = emitFilesAndReportErrors;
        function emitFilesAndReportErrorsAndGetExitStatus(program, reportDiagnostic, write, reportSummary, writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers) {
            var _a = emitFilesAndReportErrors(program, reportDiagnostic, write, reportSummary, writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers), emitResult = _a.emitResult, diagnostics = _a.diagnostics;
            if (emitResult.emitSkipped && diagnostics.length > 0) {
                // If the emitter didn't emit anything, then pass that value along.
                return ts.ExitStatus.DiagnosticsPresent_OutputsSkipped;
            }
            else if (diagnostics.length > 0) {
                // The emitter emitted something, inform the caller if that happened in the presence
                // of diagnostics or not.
                return ts.ExitStatus.DiagnosticsPresent_OutputsGenerated;
            }
            return ts.ExitStatus.Success;
        }
        ts.emitFilesAndReportErrorsAndGetExitStatus = emitFilesAndReportErrorsAndGetExitStatus;
        ts.noopFileWatcher = { close: ts.noop };
        ts.returnNoopFileWatcher = function () { return ts.noopFileWatcher; };
        function createWatchHost(system, reportWatchStatus) {
            if (system === void 0) {
                system = ts.sys;
            }
            var onWatchStatusChange = reportWatchStatus || createWatchStatusReporter(system);
            return {
                onWatchStatusChange: onWatchStatusChange,
                watchFile: ts.maybeBind(system, system.watchFile) || ts.returnNoopFileWatcher,
                watchDirectory: ts.maybeBind(system, system.watchDirectory) || ts.returnNoopFileWatcher,
                setTimeout: ts.maybeBind(system, system.setTimeout) || ts.noop,
                clearTimeout: ts.maybeBind(system, system.clearTimeout) || ts.noop
            };
        }
        ts.createWatchHost = createWatchHost;
        ts.WatchType = {
            ConfigFile: "Config file",
            ExtendedConfigFile: "Extended config file",
            SourceFile: "Source file",
            MissingFile: "Missing file",
            WildcardDirectory: "Wild card directory",
            FailedLookupLocations: "Failed Lookup Locations",
            TypeRoots: "Type roots",
            ConfigFileOfReferencedProject: "Config file of referened project",
            ExtendedConfigOfReferencedProject: "Extended config file of referenced project",
            WildcardDirectoryOfReferencedProject: "Wild card directory of referenced project",
            PackageJson: "package.json file",
        };
        function createWatchFactory(host, options) {
            var watchLogLevel = host.trace ? options.extendedDiagnostics ? ts.WatchLogLevel.Verbose : options.diagnostics ? ts.WatchLogLevel.TriggerOnly : ts.WatchLogLevel.None : ts.WatchLogLevel.None;
            var writeLog = watchLogLevel !== ts.WatchLogLevel.None ? (function (s) { return host.trace(s); }) : ts.noop;
            var result = ts.getWatchFactory(host, watchLogLevel, writeLog);
            result.writeLog = writeLog;
            return result;
        }
        ts.createWatchFactory = createWatchFactory;
        function createCompilerHostFromProgramHost(host, getCompilerOptions, directoryStructureHost) {
            if (directoryStructureHost === void 0) {
                directoryStructureHost = host;
            }
            var useCaseSensitiveFileNames = host.useCaseSensitiveFileNames();
            var hostGetNewLine = ts.memoize(function () { return host.getNewLine(); });
            return {
                getSourceFile: function (fileName, languageVersion, onError) {
                    var text;
                    try {
                        ts.performance.mark("beforeIORead");
                        text = host.readFile(fileName, getCompilerOptions().charset);
                        ts.performance.mark("afterIORead");
                        ts.performance.measure("I/O Read", "beforeIORead", "afterIORead");
                    }
                    catch (e) {
                        if (onError) {
                            onError(e.message);
                        }
                        text = "";
                    }
                    return text !== undefined ? ts.createSourceFile(fileName, text, languageVersion) : undefined;
                },
                getDefaultLibLocation: ts.maybeBind(host, host.getDefaultLibLocation),
                getDefaultLibFileName: function (options) { return host.getDefaultLibFileName(options); },
                writeFile: writeFile,
                getCurrentDirectory: ts.memoize(function () { return host.getCurrentDirectory(); }),
                useCaseSensitiveFileNames: function () { return useCaseSensitiveFileNames; },
                getCanonicalFileName: ts.createGetCanonicalFileName(useCaseSensitiveFileNames),
                getNewLine: function () { return ts.getNewLineCharacter(getCompilerOptions(), hostGetNewLine); },
                fileExists: function (f) { return host.fileExists(f); },
                readFile: function (f) { return host.readFile(f); },
                trace: ts.maybeBind(host, host.trace),
                directoryExists: ts.maybeBind(directoryStructureHost, directoryStructureHost.directoryExists),
                getDirectories: ts.maybeBind(directoryStructureHost, directoryStructureHost.getDirectories),
                realpath: ts.maybeBind(host, host.realpath),
                getEnvironmentVariable: ts.maybeBind(host, host.getEnvironmentVariable) || (function () { return ""; }),
                createHash: ts.maybeBind(host, host.createHash),
                readDirectory: ts.maybeBind(host, host.readDirectory),
                disableUseFileVersionAsSignature: host.disableUseFileVersionAsSignature,
            };
            function writeFile(fileName, text, writeByteOrderMark, onError) {
                try {
                    ts.performance.mark("beforeIOWrite");
                    // NOTE: If patchWriteFileEnsuringDirectory has been called,
                    // the host.writeFile will do its own directory creation and
                    // the ensureDirectoriesExist call will always be redundant.
                    ts.writeFileEnsuringDirectories(fileName, text, writeByteOrderMark, function (path, data, writeByteOrderMark) { return host.writeFile(path, data, writeByteOrderMark); }, function (path) { return host.createDirectory(path); }, function (path) { return host.directoryExists(path); });
                    ts.performance.mark("afterIOWrite");
                    ts.performance.measure("I/O Write", "beforeIOWrite", "afterIOWrite");
                }
                catch (e) {
                    if (onError) {
                        onError(e.message);
                    }
                }
            }
        }
        ts.createCompilerHostFromProgramHost = createCompilerHostFromProgramHost;
        function setGetSourceFileAsHashVersioned(compilerHost, host) {
            var originalGetSourceFile = compilerHost.getSourceFile;
            var computeHash = ts.maybeBind(host, host.createHash) || ts.generateDjb2Hash;
            compilerHost.getSourceFile = function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                var result = originalGetSourceFile.call.apply(originalGetSourceFile, __spreadArray([compilerHost], args, false));
                if (result) {
                    result.version = computeHash(result.text);
                }
                return result;
            };
        }
        ts.setGetSourceFileAsHashVersioned = setGetSourceFileAsHashVersioned;
        /**
         * Creates the watch compiler host that can be extended with config file or root file names and options host
         */
        function createProgramHost(system, createProgram) {
            var getDefaultLibLocation = ts.memoize(function () { return ts.getDirectoryPath(ts.normalizePath(system.getExecutingFilePath())); });
            return {
                useCaseSensitiveFileNames: function () { return system.useCaseSensitiveFileNames; },
                getNewLine: function () { return system.newLine; },
                getCurrentDirectory: ts.memoize(function () { return system.getCurrentDirectory(); }),
                getDefaultLibLocation: getDefaultLibLocation,
                getDefaultLibFileName: function (options) { return ts.combinePaths(getDefaultLibLocation(), ts.getDefaultLibFileName(options)); },
                fileExists: function (path) { return system.fileExists(path); },
                readFile: function (path, encoding) { return system.readFile(path, encoding); },
                directoryExists: function (path) { return system.directoryExists(path); },
                getDirectories: function (path) { return system.getDirectories(path); },
                readDirectory: function (path, extensions, exclude, include, depth) { return system.readDirectory(path, extensions, exclude, include, depth); },
                realpath: ts.maybeBind(system, system.realpath),
                getEnvironmentVariable: ts.maybeBind(system, system.getEnvironmentVariable),
                trace: function (s) { return system.write(s + system.newLine); },
                createDirectory: function (path) { return system.createDirectory(path); },
                writeFile: function (path, data, writeByteOrderMark) { return system.writeFile(path, data, writeByteOrderMark); },
                createHash: ts.maybeBind(system, system.createHash),
                createProgram: createProgram || ts.createEmitAndSemanticDiagnosticsBuilderProgram,
                disableUseFileVersionAsSignature: system.disableUseFileVersionAsSignature,
            };
        }
        ts.createProgramHost = createProgramHost;
        /**
         * Creates the watch compiler host that can be extended with config file or root file names and options host
         */
        function createWatchCompilerHost(system, createProgram, reportDiagnostic, reportWatchStatus) {
            if (system === void 0) {
                system = ts.sys;
            }
            var write = function (s) { return system.write(s + system.newLine); };
            var result = createProgramHost(system, createProgram);
            ts.copyProperties(result, createWatchHost(system, reportWatchStatus));
            result.afterProgramCreate = function (builderProgram) {
                var compilerOptions = builderProgram.getCompilerOptions();
                var newLine = ts.getNewLineCharacter(compilerOptions, function () { return system.newLine; });
                emitFilesAndReportErrors(builderProgram, reportDiagnostic, write, function (errorCount) { return result.onWatchStatusChange(ts.createCompilerDiagnostic(getWatchErrorSummaryDiagnosticMessage(errorCount), errorCount), newLine, compilerOptions, errorCount); });
            };
            return result;
        }
        /**
         * Report error and exit
         */
        function reportUnrecoverableDiagnostic(system, reportDiagnostic, diagnostic) {
            reportDiagnostic(diagnostic);
            system.exit(ts.ExitStatus.DiagnosticsPresent_OutputsSkipped);
        }
        /**
         * Creates the watch compiler host from system for config file in watch mode
         */
        function createWatchCompilerHostOfConfigFile(_a) {
            var configFileName = _a.configFileName, optionsToExtend = _a.optionsToExtend, watchOptionsToExtend = _a.watchOptionsToExtend, extraFileExtensions = _a.extraFileExtensions, system = _a.system, createProgram = _a.createProgram, reportDiagnostic = _a.reportDiagnostic, reportWatchStatus = _a.reportWatchStatus;
            var diagnosticReporter = reportDiagnostic || createDiagnosticReporter(system);
            var host = createWatchCompilerHost(system, createProgram, diagnosticReporter, reportWatchStatus);
            host.onUnRecoverableConfigFileDiagnostic = function (diagnostic) { return reportUnrecoverableDiagnostic(system, diagnosticReporter, diagnostic); };
            host.configFileName = configFileName;
            host.optionsToExtend = optionsToExtend;
            host.watchOptionsToExtend = watchOptionsToExtend;
            host.extraFileExtensions = extraFileExtensions;
            return host;
        }
        ts.createWatchCompilerHostOfConfigFile = createWatchCompilerHostOfConfigFile;
        /**
         * Creates the watch compiler host from system for compiling root files and options in watch mode
         */
        function createWatchCompilerHostOfFilesAndCompilerOptions(_a) {
            var rootFiles = _a.rootFiles, options = _a.options, watchOptions = _a.watchOptions, projectReferences = _a.projectReferences, system = _a.system, createProgram = _a.createProgram, reportDiagnostic = _a.reportDiagnostic, reportWatchStatus = _a.reportWatchStatus;
            var host = createWatchCompilerHost(system, createProgram, reportDiagnostic || createDiagnosticReporter(system), reportWatchStatus);
            host.rootFiles = rootFiles;
            host.options = options;
            host.watchOptions = watchOptions;
            host.projectReferences = projectReferences;
            return host;
        }
        ts.createWatchCompilerHostOfFilesAndCompilerOptions = createWatchCompilerHostOfFilesAndCompilerOptions;
        function performIncrementalCompilation(input) {
            var system = input.system || ts.sys;
            var host = input.host || (input.host = ts.createIncrementalCompilerHost(input.options, system));
            var builderProgram = ts.createIncrementalProgram(input);
            var exitStatus = emitFilesAndReportErrorsAndGetExitStatus(builderProgram, input.reportDiagnostic || createDiagnosticReporter(system), function (s) { return host.trace && host.trace(s); }, input.reportErrorSummary || input.options.pretty ? function (errorCount) { return system.write(getErrorSummaryText(errorCount, system.newLine)); } : undefined);
            if (input.afterProgramEmitAndDiagnostics)
                input.afterProgramEmitAndDiagnostics(builderProgram);
            return exitStatus;
        }
        ts.performIncrementalCompilation = performIncrementalCompilation;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        function readBuilderProgram(compilerOptions, host) {
            if (ts.outFile(compilerOptions))
                return undefined;
            var buildInfoPath = ts.getTsBuildInfoEmitOutputFilePath(compilerOptions);
            if (!buildInfoPath)
                return undefined;
            var content = host.readFile(buildInfoPath);
            if (!content)
                return undefined;
            var buildInfo = ts.getBuildInfo(content);
            if (buildInfo.version !== ts.version)
                return undefined;
            if (!buildInfo.program)
                return undefined;
            return ts.createBuildProgramUsingProgramBuildInfo(buildInfo.program, buildInfoPath, host);
        }
        ts.readBuilderProgram = readBuilderProgram;
        function createIncrementalCompilerHost(options, system) {
            if (system === void 0) {
                system = ts.sys;
            }
            var host = ts.createCompilerHostWorker(options, /*setParentNodes*/ undefined, system);
            host.createHash = ts.maybeBind(system, system.createHash);
            host.disableUseFileVersionAsSignature = system.disableUseFileVersionAsSignature;
            ts.setGetSourceFileAsHashVersioned(host, system);
            ts.changeCompilerHostLikeToUseCache(host, function (fileName) { return ts.toPath(fileName, host.getCurrentDirectory(), host.getCanonicalFileName); });
            return host;
        }
        ts.createIncrementalCompilerHost = createIncrementalCompilerHost;
        function createIncrementalProgram(_a) {
            var rootNames = _a.rootNames, options = _a.options, configFileParsingDiagnostics = _a.configFileParsingDiagnostics, projectReferences = _a.projectReferences, host = _a.host, createProgram = _a.createProgram;
            host = host || createIncrementalCompilerHost(options);
            createProgram = createProgram || ts.createEmitAndSemanticDiagnosticsBuilderProgram;
            var oldProgram = readBuilderProgram(options, host);
            return createProgram(rootNames, options, host, oldProgram, configFileParsingDiagnostics, projectReferences);
        }
        ts.createIncrementalProgram = createIncrementalProgram;
        function createWatchCompilerHost(rootFilesOrConfigFileName, options, system, createProgram, reportDiagnostic, reportWatchStatus, projectReferencesOrWatchOptionsToExtend, watchOptionsOrExtraFileExtensions) {
            if (ts.isArray(rootFilesOrConfigFileName)) {
                return ts.createWatchCompilerHostOfFilesAndCompilerOptions({
                    rootFiles: rootFilesOrConfigFileName,
                    options: options,
                    watchOptions: watchOptionsOrExtraFileExtensions,
                    projectReferences: projectReferencesOrWatchOptionsToExtend,
                    system: system,
                    createProgram: createProgram,
                    reportDiagnostic: reportDiagnostic,
                    reportWatchStatus: reportWatchStatus,
                });
            }
            else {
                return ts.createWatchCompilerHostOfConfigFile({
                    configFileName: rootFilesOrConfigFileName,
                    optionsToExtend: options,
                    watchOptionsToExtend: projectReferencesOrWatchOptionsToExtend,
                    extraFileExtensions: watchOptionsOrExtraFileExtensions,
                    system: system,
                    createProgram: createProgram,
                    reportDiagnostic: reportDiagnostic,
                    reportWatchStatus: reportWatchStatus,
                });
            }
        }
        ts.createWatchCompilerHost = createWatchCompilerHost;
        function createWatchProgram(host) {
            var builderProgram;
            var reloadLevel; // level to indicate if the program needs to be reloaded from config file/just filenames etc
            var missingFilesMap; // Map of file watchers for the missing files
            var packageJsonMap; // map of watchers for package json files used in module resolution
            var watchedWildcardDirectories; // map of watchers for the wild card directories in the config file
            var timerToUpdateProgram; // timer callback to recompile the program
            var timerToInvalidateFailedLookupResolutions; // timer callback to invalidate resolutions for changes in failed lookup locations
            var parsedConfigs; // Parsed commandline and watching cached for referenced projects
            var sharedExtendedConfigFileWatchers; // Map of file watchers for extended files, shared between different referenced projects
            var extendedConfigCache = host.extendedConfigCache; // Cache for extended config evaluation
            var changesAffectResolution = false; // Flag for indicating non-config changes affect module resolution
            var sourceFilesCache = new ts.Map(); // Cache that stores the source file and version info
            var missingFilePathsRequestedForRelease; // These paths are held temporarily so that we can remove the entry from source file cache if the file is not tracked by missing files
            var hasChangedCompilerOptions = false; // True if the compiler options have changed between compilations
            var useCaseSensitiveFileNames = host.useCaseSensitiveFileNames();
            var currentDirectory = host.getCurrentDirectory();
            var configFileName = host.configFileName, _a = host.optionsToExtend, optionsToExtendForConfigFile = _a === void 0 ? {} : _a, watchOptionsToExtend = host.watchOptionsToExtend, extraFileExtensions = host.extraFileExtensions, createProgram = host.createProgram;
            var rootFileNames = host.rootFiles, compilerOptions = host.options, watchOptions = host.watchOptions, projectReferences = host.projectReferences;
            var wildcardDirectories;
            var configFileParsingDiagnostics;
            var canConfigFileJsonReportNoInputFiles = false;
            var hasChangedConfigFileParsingErrors = false;
            var cachedDirectoryStructureHost = configFileName === undefined ? undefined : ts.createCachedDirectoryStructureHost(host, currentDirectory, useCaseSensitiveFileNames);
            var directoryStructureHost = cachedDirectoryStructureHost || host;
            var parseConfigFileHost = ts.parseConfigHostFromCompilerHostLike(host, directoryStructureHost);
            // From tsc we want to get already parsed result and hence check for rootFileNames
            var newLine = updateNewLine();
            if (configFileName && host.configFileParsingResult) {
                setConfigFileParsingResult(host.configFileParsingResult);
                newLine = updateNewLine();
            }
            reportWatchDiagnostic(ts.Diagnostics.Starting_compilation_in_watch_mode);
            if (configFileName && !host.configFileParsingResult) {
                newLine = ts.getNewLineCharacter(optionsToExtendForConfigFile, function () { return host.getNewLine(); });
                ts.Debug.assert(!rootFileNames);
                parseConfigFile();
                newLine = updateNewLine();
            }
            var _b = ts.createWatchFactory(host, compilerOptions), watchFile = _b.watchFile, watchDirectory = _b.watchDirectory, writeLog = _b.writeLog;
            var getCanonicalFileName = ts.createGetCanonicalFileName(useCaseSensitiveFileNames);
            writeLog("Current directory: " + currentDirectory + " CaseSensitiveFileNames: " + useCaseSensitiveFileNames);
            var configFileWatcher;
            if (configFileName) {
                configFileWatcher = watchFile(configFileName, scheduleProgramReload, ts.PollingInterval.High, watchOptions, ts.WatchType.ConfigFile);
            }
            var compilerHost = ts.createCompilerHostFromProgramHost(host, function () { return compilerOptions; }, directoryStructureHost);
            ts.setGetSourceFileAsHashVersioned(compilerHost, host);
            // Members for CompilerHost
            var getNewSourceFile = compilerHost.getSourceFile;
            compilerHost.getSourceFile = function (fileName) {
                var args = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    args[_i - 1] = arguments[_i];
                }
                return getVersionedSourceFileByPath.apply(void 0, __spreadArray([fileName, toPath(fileName)], args, false));
            };
            compilerHost.getSourceFileByPath = getVersionedSourceFileByPath;
            compilerHost.getNewLine = function () { return newLine; };
            compilerHost.fileExists = fileExists;
            compilerHost.onReleaseOldSourceFile = onReleaseOldSourceFile;
            compilerHost.onReleaseParsedCommandLine = onReleaseParsedCommandLine;
            // Members for ResolutionCacheHost
            compilerHost.toPath = toPath;
            compilerHost.getCompilationSettings = function () { return compilerOptions; };
            compilerHost.useSourceOfProjectReferenceRedirect = ts.maybeBind(host, host.useSourceOfProjectReferenceRedirect);
            compilerHost.watchDirectoryOfFailedLookupLocation = function (dir, cb, flags) { return watchDirectory(dir, cb, flags, watchOptions, ts.WatchType.FailedLookupLocations); };
            compilerHost.watchTypeRootsDirectory = function (dir, cb, flags) { return watchDirectory(dir, cb, flags, watchOptions, ts.WatchType.TypeRoots); };
            compilerHost.getCachedDirectoryStructureHost = function () { return cachedDirectoryStructureHost; };
            compilerHost.scheduleInvalidateResolutionsOfFailedLookupLocations = scheduleInvalidateResolutionsOfFailedLookupLocations;
            compilerHost.onInvalidatedResolution = scheduleProgramUpdate;
            compilerHost.onChangedAutomaticTypeDirectiveNames = scheduleProgramUpdate;
            compilerHost.fileIsOpen = ts.returnFalse;
            compilerHost.getCurrentProgram = getCurrentProgram;
            compilerHost.writeLog = writeLog;
            compilerHost.getParsedCommandLine = getParsedCommandLine;
            // Cache for the module resolution
            var resolutionCache = ts.createResolutionCache(compilerHost, configFileName ?
                ts.getDirectoryPath(ts.getNormalizedAbsolutePath(configFileName, currentDirectory)) :
                currentDirectory, 
            /*logChangesWhenResolvingModule*/ false);
            // Resolve module using host module resolution strategy if provided otherwise use resolution cache to resolve module names
            compilerHost.resolveModuleNames = host.resolveModuleNames ?
                (function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    return host.resolveModuleNames.apply(host, args);
                }) :
                (function (moduleNames, containingFile, reusedNames, redirectedReference) { return resolutionCache.resolveModuleNames(moduleNames, containingFile, reusedNames, redirectedReference); });
            compilerHost.resolveTypeReferenceDirectives = host.resolveTypeReferenceDirectives ?
                (function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    return host.resolveTypeReferenceDirectives.apply(host, args);
                }) :
                (function (typeDirectiveNames, containingFile, redirectedReference) { return resolutionCache.resolveTypeReferenceDirectives(typeDirectiveNames, containingFile, redirectedReference); });
            var userProvidedResolution = !!host.resolveModuleNames || !!host.resolveTypeReferenceDirectives;
            builderProgram = readBuilderProgram(compilerOptions, compilerHost);
            synchronizeProgram();
            // Update the wild card directory watch
            watchConfigFileWildCardDirectories();
            // Update extended config file watch
            if (configFileName)
                updateExtendedConfigFilesWatches(toPath(configFileName), compilerOptions, watchOptions, ts.WatchType.ExtendedConfigFile);
            return configFileName ?
                { getCurrentProgram: getCurrentBuilderProgram, getProgram: updateProgram, close: close } :
                { getCurrentProgram: getCurrentBuilderProgram, getProgram: updateProgram, updateRootFileNames: updateRootFileNames, close: close };
            function close() {
                clearInvalidateResolutionsOfFailedLookupLocations();
                resolutionCache.clear();
                ts.clearMap(sourceFilesCache, function (value) {
                    if (value && value.fileWatcher) {
                        value.fileWatcher.close();
                        value.fileWatcher = undefined;
                    }
                });
                if (configFileWatcher) {
                    configFileWatcher.close();
                    configFileWatcher = undefined;
                }
                extendedConfigCache === null || extendedConfigCache === void 0 ? void 0 : extendedConfigCache.clear();
                extendedConfigCache = undefined;
                if (sharedExtendedConfigFileWatchers) {
                    ts.clearMap(sharedExtendedConfigFileWatchers, ts.closeFileWatcherOf);
                    sharedExtendedConfigFileWatchers = undefined;
                }
                if (watchedWildcardDirectories) {
                    ts.clearMap(watchedWildcardDirectories, ts.closeFileWatcherOf);
                    watchedWildcardDirectories = undefined;
                }
                if (missingFilesMap) {
                    ts.clearMap(missingFilesMap, ts.closeFileWatcher);
                    missingFilesMap = undefined;
                }
                if (parsedConfigs) {
                    ts.clearMap(parsedConfigs, function (config) {
                        var _a;
                        (_a = config.watcher) === null || _a === void 0 ? void 0 : _a.close();
                        config.watcher = undefined;
                        if (config.watchedDirectories)
                            ts.clearMap(config.watchedDirectories, ts.closeFileWatcherOf);
                        config.watchedDirectories = undefined;
                    });
                    parsedConfigs = undefined;
                }
            }
            function getCurrentBuilderProgram() {
                return builderProgram;
            }
            function getCurrentProgram() {
                return builderProgram && builderProgram.getProgramOrUndefined();
            }
            function synchronizeProgram() {
                writeLog("Synchronizing program");
                clearInvalidateResolutionsOfFailedLookupLocations();
                var program = getCurrentBuilderProgram();
                if (hasChangedCompilerOptions) {
                    newLine = updateNewLine();
                    if (program && (changesAffectResolution || ts.changesAffectModuleResolution(program.getCompilerOptions(), compilerOptions))) {
                        resolutionCache.clear();
                    }
                }
                // All resolutions are invalid if user provided resolutions
                var hasInvalidatedResolution = resolutionCache.createHasInvalidatedResolution(userProvidedResolution || changesAffectResolution);
                if (ts.isProgramUptoDate(getCurrentProgram(), rootFileNames, compilerOptions, getSourceVersion, fileExists, hasInvalidatedResolution, hasChangedAutomaticTypeDirectiveNames, getParsedCommandLine, projectReferences)) {
                    if (hasChangedConfigFileParsingErrors) {
                        builderProgram = createProgram(/*rootNames*/ undefined, /*options*/ undefined, compilerHost, builderProgram, configFileParsingDiagnostics, projectReferences);
                        hasChangedConfigFileParsingErrors = false;
                    }
                }
                else {
                    createNewProgram(hasInvalidatedResolution);
                }
                changesAffectResolution = false; // reset for next sync
                if (host.afterProgramCreate && program !== builderProgram) {
                    host.afterProgramCreate(builderProgram);
                }
                return builderProgram;
            }
            function createNewProgram(hasInvalidatedResolution) {
                // Compile the program
                writeLog("CreatingProgramWith::");
                writeLog("  roots: " + JSON.stringify(rootFileNames));
                writeLog("  options: " + JSON.stringify(compilerOptions));
                if (projectReferences)
                    writeLog("  projectReferences: " + JSON.stringify(projectReferences));
                var needsUpdateInTypeRootWatch = hasChangedCompilerOptions || !getCurrentProgram();
                hasChangedCompilerOptions = false;
                hasChangedConfigFileParsingErrors = false;
                resolutionCache.startCachingPerDirectoryResolution();
                compilerHost.hasInvalidatedResolution = hasInvalidatedResolution;
                compilerHost.hasChangedAutomaticTypeDirectiveNames = hasChangedAutomaticTypeDirectiveNames;
                builderProgram = createProgram(rootFileNames, compilerOptions, compilerHost, builderProgram, configFileParsingDiagnostics, projectReferences);
                // map package json cache entries to their realpaths so we don't try to watch across symlinks
                var packageCacheEntries = ts.map(resolutionCache.getModuleResolutionCache().getPackageJsonInfoCache().entries(), function (_a) {
                    var path = _a[0], data = _a[1];
                    return [compilerHost.realpath ? toPath(compilerHost.realpath(path)) : path, data];
                });
                resolutionCache.finishCachingPerDirectoryResolution();
                // Update watches
                ts.updateMissingFilePathsWatch(builderProgram.getProgram(), missingFilesMap || (missingFilesMap = new ts.Map()), watchMissingFilePath);
                ts.updatePackageJsonWatch(packageCacheEntries, packageJsonMap || (packageJsonMap = new ts.Map()), watchPackageJsonLookupPath);
                if (needsUpdateInTypeRootWatch) {
                    resolutionCache.updateTypeRootsWatch();
                }
                if (missingFilePathsRequestedForRelease) {
                    // These are the paths that program creater told us as not in use any more but were missing on the disk.
                    // We didnt remove the entry for them from sourceFiles cache so that we dont have to do File IO,
                    // if there is already watcher for it (for missing files)
                    // At this point our watches were updated, hence now we know that these paths are not tracked and need to be removed
                    // so that at later time we have correct result of their presence
                    for (var _i = 0, missingFilePathsRequestedForRelease_1 = missingFilePathsRequestedForRelease; _i < missingFilePathsRequestedForRelease_1.length; _i++) {
                        var missingFilePath = missingFilePathsRequestedForRelease_1[_i];
                        if (!missingFilesMap.has(missingFilePath)) {
                            sourceFilesCache.delete(missingFilePath);
                        }
                    }
                    missingFilePathsRequestedForRelease = undefined;
                }
            }
            function updateRootFileNames(files) {
                ts.Debug.assert(!configFileName, "Cannot update root file names with config file watch mode");
                rootFileNames = files;
                scheduleProgramUpdate();
            }
            function updateNewLine() {
                return ts.getNewLineCharacter(compilerOptions || optionsToExtendForConfigFile, function () { return host.getNewLine(); });
            }
            function toPath(fileName) {
                return ts.toPath(fileName, currentDirectory, getCanonicalFileName);
            }
            function isFileMissingOnHost(hostSourceFile) {
                return typeof hostSourceFile === "boolean";
            }
            function isFilePresenceUnknownOnHost(hostSourceFile) {
                return typeof hostSourceFile.version === "boolean";
            }
            function fileExists(fileName) {
                var path = toPath(fileName);
                // If file is missing on host from cache, we can definitely say file doesnt exist
                // otherwise we need to ensure from the disk
                if (isFileMissingOnHost(sourceFilesCache.get(path))) {
                    return false;
                }
                return directoryStructureHost.fileExists(fileName);
            }
            function getVersionedSourceFileByPath(fileName, path, languageVersion, onError, shouldCreateNewSourceFile) {
                var hostSourceFile = sourceFilesCache.get(path);
                // No source file on the host
                if (isFileMissingOnHost(hostSourceFile)) {
                    return undefined;
                }
                // Create new source file if requested or the versions dont match
                if (hostSourceFile === undefined || shouldCreateNewSourceFile || isFilePresenceUnknownOnHost(hostSourceFile)) {
                    var sourceFile = getNewSourceFile(fileName, languageVersion, onError);
                    if (hostSourceFile) {
                        if (sourceFile) {
                            // Set the source file and create file watcher now that file was present on the disk
                            hostSourceFile.sourceFile = sourceFile;
                            hostSourceFile.version = sourceFile.version;
                            if (!hostSourceFile.fileWatcher) {
                                hostSourceFile.fileWatcher = watchFilePath(path, fileName, onSourceFileChange, ts.PollingInterval.Low, watchOptions, ts.WatchType.SourceFile);
                            }
                        }
                        else {
                            // There is no source file on host any more, close the watch, missing file paths will track it
                            if (hostSourceFile.fileWatcher) {
                                hostSourceFile.fileWatcher.close();
                            }
                            sourceFilesCache.set(path, false);
                        }
                    }
                    else {
                        if (sourceFile) {
                            var fileWatcher = watchFilePath(path, fileName, onSourceFileChange, ts.PollingInterval.Low, watchOptions, ts.WatchType.SourceFile);
                            sourceFilesCache.set(path, { sourceFile: sourceFile, version: sourceFile.version, fileWatcher: fileWatcher });
                        }
                        else {
                            sourceFilesCache.set(path, false);
                        }
                    }
                    return sourceFile;
                }
                return hostSourceFile.sourceFile;
            }
            function nextSourceFileVersion(path) {
                var hostSourceFile = sourceFilesCache.get(path);
                if (hostSourceFile !== undefined) {
                    if (isFileMissingOnHost(hostSourceFile)) {
                        // The next version, lets set it as presence unknown file
                        sourceFilesCache.set(path, { version: false });
                    }
                    else {
                        hostSourceFile.version = false;
                    }
                }
            }
            function getSourceVersion(path) {
                var hostSourceFile = sourceFilesCache.get(path);
                return !hostSourceFile || !hostSourceFile.version ? undefined : hostSourceFile.version;
            }
            function onReleaseOldSourceFile(oldSourceFile, _oldOptions, hasSourceFileByPath) {
                var hostSourceFileInfo = sourceFilesCache.get(oldSourceFile.resolvedPath);
                // If this is the source file thats in the cache and new program doesnt need it,
                // remove the cached entry.
                // Note we arent deleting entry if file became missing in new program or
                // there was version update and new source file was created.
                if (hostSourceFileInfo !== undefined) {
                    // record the missing file paths so they can be removed later if watchers arent tracking them
                    if (isFileMissingOnHost(hostSourceFileInfo)) {
                        (missingFilePathsRequestedForRelease || (missingFilePathsRequestedForRelease = [])).push(oldSourceFile.path);
                    }
                    else if (hostSourceFileInfo.sourceFile === oldSourceFile) {
                        if (hostSourceFileInfo.fileWatcher) {
                            hostSourceFileInfo.fileWatcher.close();
                        }
                        sourceFilesCache.delete(oldSourceFile.resolvedPath);
                        if (!hasSourceFileByPath) {
                            resolutionCache.removeResolutionsOfFile(oldSourceFile.path);
                        }
                    }
                }
            }
            function reportWatchDiagnostic(message) {
                if (host.onWatchStatusChange) {
                    host.onWatchStatusChange(ts.createCompilerDiagnostic(message), newLine, compilerOptions || optionsToExtendForConfigFile);
                }
            }
            function hasChangedAutomaticTypeDirectiveNames() {
                return resolutionCache.hasChangedAutomaticTypeDirectiveNames();
            }
            function clearInvalidateResolutionsOfFailedLookupLocations() {
                if (!timerToInvalidateFailedLookupResolutions)
                    return false;
                host.clearTimeout(timerToInvalidateFailedLookupResolutions);
                timerToInvalidateFailedLookupResolutions = undefined;
                return true;
            }
            function scheduleInvalidateResolutionsOfFailedLookupLocations() {
                if (!host.setTimeout || !host.clearTimeout) {
                    return resolutionCache.invalidateResolutionsOfFailedLookupLocations();
                }
                var pending = clearInvalidateResolutionsOfFailedLookupLocations();
                writeLog("Scheduling invalidateFailedLookup" + (pending ? ", Cancelled earlier one" : ""));
                timerToInvalidateFailedLookupResolutions = host.setTimeout(invalidateResolutionsOfFailedLookup, 250);
            }
            function invalidateResolutionsOfFailedLookup() {
                timerToInvalidateFailedLookupResolutions = undefined;
                if (resolutionCache.invalidateResolutionsOfFailedLookupLocations()) {
                    scheduleProgramUpdate();
                }
            }
            // Upon detecting a file change, wait for 250ms and then perform a recompilation. This gives batch
            // operations (such as saving all modified files in an editor) a chance to complete before we kick
            // off a new compilation.
            function scheduleProgramUpdate() {
                if (!host.setTimeout || !host.clearTimeout) {
                    return;
                }
                if (timerToUpdateProgram) {
                    host.clearTimeout(timerToUpdateProgram);
                }
                writeLog("Scheduling update");
                timerToUpdateProgram = host.setTimeout(updateProgramWithWatchStatus, 250);
            }
            function scheduleProgramReload() {
                ts.Debug.assert(!!configFileName);
                reloadLevel = ts.ConfigFileProgramReloadLevel.Full;
                scheduleProgramUpdate();
            }
            function updateProgramWithWatchStatus() {
                timerToUpdateProgram = undefined;
                reportWatchDiagnostic(ts.Diagnostics.File_change_detected_Starting_incremental_compilation);
                updateProgram();
            }
            function updateProgram() {
                switch (reloadLevel) {
                    case ts.ConfigFileProgramReloadLevel.Partial:
                        ts.perfLogger.logStartUpdateProgram("PartialConfigReload");
                        reloadFileNamesFromConfigFile();
                        break;
                    case ts.ConfigFileProgramReloadLevel.Full:
                        ts.perfLogger.logStartUpdateProgram("FullConfigReload");
                        reloadConfigFile();
                        break;
                    default:
                        ts.perfLogger.logStartUpdateProgram("SynchronizeProgram");
                        synchronizeProgram();
                        break;
                }
                ts.perfLogger.logStopUpdateProgram("Done");
                return getCurrentBuilderProgram();
            }
            function reloadFileNamesFromConfigFile() {
                writeLog("Reloading new file names and options");
                rootFileNames = ts.getFileNamesFromConfigSpecs(compilerOptions.configFile.configFileSpecs, ts.getNormalizedAbsolutePath(ts.getDirectoryPath(configFileName), currentDirectory), compilerOptions, parseConfigFileHost, extraFileExtensions);
                if (ts.updateErrorForNoInputFiles(rootFileNames, ts.getNormalizedAbsolutePath(configFileName, currentDirectory), compilerOptions.configFile.configFileSpecs, configFileParsingDiagnostics, canConfigFileJsonReportNoInputFiles)) {
                    hasChangedConfigFileParsingErrors = true;
                }
                // Update the program
                synchronizeProgram();
            }
            function reloadConfigFile() {
                writeLog("Reloading config file: " + configFileName);
                reloadLevel = ts.ConfigFileProgramReloadLevel.None;
                if (cachedDirectoryStructureHost) {
                    cachedDirectoryStructureHost.clearCache();
                }
                parseConfigFile();
                hasChangedCompilerOptions = true;
                synchronizeProgram();
                // Update the wild card directory watch
                watchConfigFileWildCardDirectories();
                // Update extended config file watch
                updateExtendedConfigFilesWatches(toPath(configFileName), compilerOptions, watchOptions, ts.WatchType.ExtendedConfigFile);
            }
            function parseConfigFile() {
                setConfigFileParsingResult(ts.getParsedCommandLineOfConfigFile(configFileName, optionsToExtendForConfigFile, parseConfigFileHost, extendedConfigCache || (extendedConfigCache = new ts.Map()), watchOptionsToExtend, extraFileExtensions)); // TODO: GH#18217
            }
            function setConfigFileParsingResult(configFileParseResult) {
                rootFileNames = configFileParseResult.fileNames;
                compilerOptions = configFileParseResult.options;
                watchOptions = configFileParseResult.watchOptions;
                projectReferences = configFileParseResult.projectReferences;
                wildcardDirectories = configFileParseResult.wildcardDirectories;
                configFileParsingDiagnostics = ts.getConfigFileParsingDiagnostics(configFileParseResult).slice();
                canConfigFileJsonReportNoInputFiles = ts.canJsonReportNoInputFiles(configFileParseResult.raw);
                hasChangedConfigFileParsingErrors = true;
            }
            function getParsedCommandLine(configFileName) {
                var configPath = toPath(configFileName);
                var config = parsedConfigs === null || parsedConfigs === void 0 ? void 0 : parsedConfigs.get(configPath);
                if (config) {
                    if (!config.reloadLevel)
                        return config.parsedCommandLine;
                    // With host implementing getParsedCommandLine we cant just update file names
                    if (config.parsedCommandLine && config.reloadLevel === ts.ConfigFileProgramReloadLevel.Partial && !host.getParsedCommandLine) {
                        writeLog("Reloading new file names and options");
                        var fileNames = ts.getFileNamesFromConfigSpecs(config.parsedCommandLine.options.configFile.configFileSpecs, ts.getNormalizedAbsolutePath(ts.getDirectoryPath(configFileName), currentDirectory), compilerOptions, parseConfigFileHost);
                        config.parsedCommandLine = __assign(__assign({}, config.parsedCommandLine), { fileNames: fileNames });
                        config.reloadLevel = undefined;
                        return config.parsedCommandLine;
                    }
                }
                writeLog("Loading config file: " + configFileName);
                var parsedCommandLine = host.getParsedCommandLine ?
                    host.getParsedCommandLine(configFileName) :
                    getParsedCommandLineFromConfigFileHost(configFileName);
                if (config) {
                    config.parsedCommandLine = parsedCommandLine;
                    config.reloadLevel = undefined;
                }
                else {
                    (parsedConfigs || (parsedConfigs = new ts.Map())).set(configPath, config = { parsedCommandLine: parsedCommandLine });
                }
                watchReferencedProject(configFileName, configPath, config);
                return parsedCommandLine;
            }
            function getParsedCommandLineFromConfigFileHost(configFileName) {
                // Ignore the file absent errors
                var onUnRecoverableConfigFileDiagnostic = parseConfigFileHost.onUnRecoverableConfigFileDiagnostic;
                parseConfigFileHost.onUnRecoverableConfigFileDiagnostic = ts.noop;
                var parsedCommandLine = ts.getParsedCommandLineOfConfigFile(configFileName, 
                /*optionsToExtend*/ undefined, parseConfigFileHost, extendedConfigCache || (extendedConfigCache = new ts.Map()), watchOptionsToExtend);
                parseConfigFileHost.onUnRecoverableConfigFileDiagnostic = onUnRecoverableConfigFileDiagnostic;
                return parsedCommandLine;
            }
            function onReleaseParsedCommandLine(fileName) {
                var _a;
                var path = toPath(fileName);
                var config = parsedConfigs === null || parsedConfigs === void 0 ? void 0 : parsedConfigs.get(path);
                if (!config)
                    return;
                parsedConfigs.delete(path);
                if (config.watchedDirectories)
                    ts.clearMap(config.watchedDirectories, ts.closeFileWatcherOf);
                (_a = config.watcher) === null || _a === void 0 ? void 0 : _a.close();
                ts.clearSharedExtendedConfigFileWatcher(path, sharedExtendedConfigFileWatchers);
            }
            function watchFilePath(path, file, callback, pollingInterval, options, watchType) {
                return watchFile(file, function (fileName, eventKind) { return callback(fileName, eventKind, path); }, pollingInterval, options, watchType);
            }
            function onSourceFileChange(fileName, eventKind, path) {
                updateCachedSystemWithFile(fileName, path, eventKind);
                // Update the source file cache
                if (eventKind === ts.FileWatcherEventKind.Deleted && sourceFilesCache.has(path)) {
                    resolutionCache.invalidateResolutionOfFile(path);
                }
                nextSourceFileVersion(path);
                // Update the program
                scheduleProgramUpdate();
            }
            function updateCachedSystemWithFile(fileName, path, eventKind) {
                if (cachedDirectoryStructureHost) {
                    cachedDirectoryStructureHost.addOrDeleteFile(fileName, path, eventKind);
                }
            }
            function watchMissingFilePath(missingFilePath) {
                // If watching missing referenced config file, we are already watching it so no need for separate watcher
                return (parsedConfigs === null || parsedConfigs === void 0 ? void 0 : parsedConfigs.has(missingFilePath)) ?
                    ts.noopFileWatcher :
                    watchFilePath(missingFilePath, missingFilePath, onMissingFileChange, ts.PollingInterval.Medium, watchOptions, ts.WatchType.MissingFile);
            }
            function watchPackageJsonLookupPath(packageJsonPath) {
                // If the package.json is pulled into the compilation itself (eg, via json imports), don't add a second watcher here
                return sourceFilesCache.has(packageJsonPath) ?
                    ts.noopFileWatcher :
                    watchFilePath(packageJsonPath, packageJsonPath, onPackageJsonChange, ts.PollingInterval.High, watchOptions, ts.WatchType.PackageJson);
            }
            function onPackageJsonChange(fileName, eventKind, path) {
                updateCachedSystemWithFile(fileName, path, eventKind);
                // package.json changes invalidate module resolution and can change the set of loaded files
                // so if we witness a change to one, we have to do a full reload
                reloadLevel = ts.ConfigFileProgramReloadLevel.Full;
                changesAffectResolution = true;
                // Update the program
                scheduleProgramUpdate();
            }
            function onMissingFileChange(fileName, eventKind, missingFilePath) {
                updateCachedSystemWithFile(fileName, missingFilePath, eventKind);
                if (eventKind === ts.FileWatcherEventKind.Created && missingFilesMap.has(missingFilePath)) {
                    missingFilesMap.get(missingFilePath).close();
                    missingFilesMap.delete(missingFilePath);
                    // Delete the entry in the source files cache so that new source file is created
                    nextSourceFileVersion(missingFilePath);
                    // When a missing file is created, we should update the graph.
                    scheduleProgramUpdate();
                }
            }
            function watchConfigFileWildCardDirectories() {
                if (wildcardDirectories) {
                    ts.updateWatchingWildcardDirectories(watchedWildcardDirectories || (watchedWildcardDirectories = new ts.Map()), new ts.Map(ts.getEntries(wildcardDirectories)), watchWildcardDirectory);
                }
                else if (watchedWildcardDirectories) {
                    ts.clearMap(watchedWildcardDirectories, ts.closeFileWatcherOf);
                }
            }
            function watchWildcardDirectory(directory, flags) {
                return watchDirectory(directory, function (fileOrDirectory) {
                    ts.Debug.assert(!!configFileName);
                    var fileOrDirectoryPath = toPath(fileOrDirectory);
                    // Since the file existence changed, update the sourceFiles cache
                    if (cachedDirectoryStructureHost) {
                        cachedDirectoryStructureHost.addOrDeleteFileOrDirectory(fileOrDirectory, fileOrDirectoryPath);
                    }
                    nextSourceFileVersion(fileOrDirectoryPath);
                    if (ts.isIgnoredFileFromWildCardWatching({
                        watchedDirPath: toPath(directory),
                        fileOrDirectory: fileOrDirectory,
                        fileOrDirectoryPath: fileOrDirectoryPath,
                        configFileName: configFileName,
                        extraFileExtensions: extraFileExtensions,
                        options: compilerOptions,
                        program: getCurrentBuilderProgram() || rootFileNames,
                        currentDirectory: currentDirectory,
                        useCaseSensitiveFileNames: useCaseSensitiveFileNames,
                        writeLog: writeLog,
                        toPath: toPath,
                    }))
                        return;
                    // Reload is pending, do the reload
                    if (reloadLevel !== ts.ConfigFileProgramReloadLevel.Full) {
                        reloadLevel = ts.ConfigFileProgramReloadLevel.Partial;
                        // Schedule Update the program
                        scheduleProgramUpdate();
                    }
                }, flags, watchOptions, ts.WatchType.WildcardDirectory);
            }
            function updateExtendedConfigFilesWatches(forProjectPath, options, watchOptions, watchType) {
                ts.updateSharedExtendedConfigFileWatcher(forProjectPath, options, sharedExtendedConfigFileWatchers || (sharedExtendedConfigFileWatchers = new ts.Map()), function (extendedConfigFileName, extendedConfigFilePath) {
                    return watchFile(extendedConfigFileName, function (_fileName, eventKind) {
                        var _a;
                        updateCachedSystemWithFile(extendedConfigFileName, extendedConfigFilePath, eventKind);
                        // Update extended config cache
                        if (extendedConfigCache)
                            ts.cleanExtendedConfigCache(extendedConfigCache, extendedConfigFilePath, toPath);
                        // Update projects
                        var projects = (_a = sharedExtendedConfigFileWatchers.get(extendedConfigFilePath)) === null || _a === void 0 ? void 0 : _a.projects;
                        // If there are no referenced projects this extended config file watcher depend on ignore
                        if (!(projects === null || projects === void 0 ? void 0 : projects.size))
                            return;
                        projects.forEach(function (projectPath) {
                            if (toPath(configFileName) === projectPath) {
                                // If this is the config file of the project, reload completely
                                reloadLevel = ts.ConfigFileProgramReloadLevel.Full;
                            }
                            else {
                                // Reload config for the referenced projects and remove the resolutions from referenced projects since the config file changed
                                var config = parsedConfigs === null || parsedConfigs === void 0 ? void 0 : parsedConfigs.get(projectPath);
                                if (config)
                                    config.reloadLevel = ts.ConfigFileProgramReloadLevel.Full;
                                resolutionCache.removeResolutionsFromProjectReferenceRedirects(projectPath);
                            }
                            scheduleProgramUpdate();
                        });
                    }, ts.PollingInterval.High, watchOptions, watchType);
                }, toPath);
            }
            function watchReferencedProject(configFileName, configPath, commandLine) {
                var _a, _b, _c, _d, _e;
                // Watch file
                commandLine.watcher || (commandLine.watcher = watchFile(configFileName, function (_fileName, eventKind) {
                    updateCachedSystemWithFile(configFileName, configPath, eventKind);
                    var config = parsedConfigs === null || parsedConfigs === void 0 ? void 0 : parsedConfigs.get(configPath);
                    if (config)
                        config.reloadLevel = ts.ConfigFileProgramReloadLevel.Full;
                    resolutionCache.removeResolutionsFromProjectReferenceRedirects(configPath);
                    scheduleProgramUpdate();
                }, ts.PollingInterval.High, ((_a = commandLine.parsedCommandLine) === null || _a === void 0 ? void 0 : _a.watchOptions) || watchOptions, ts.WatchType.ConfigFileOfReferencedProject));
                // Watch Wild card
                if ((_b = commandLine.parsedCommandLine) === null || _b === void 0 ? void 0 : _b.wildcardDirectories) {
                    ts.updateWatchingWildcardDirectories(commandLine.watchedDirectories || (commandLine.watchedDirectories = new ts.Map()), new ts.Map(ts.getEntries((_c = commandLine.parsedCommandLine) === null || _c === void 0 ? void 0 : _c.wildcardDirectories)), function (directory, flags) {
                        var _a;
                        return watchDirectory(directory, function (fileOrDirectory) {
                            var fileOrDirectoryPath = toPath(fileOrDirectory);
                            // Since the file existence changed, update the sourceFiles cache
                            if (cachedDirectoryStructureHost) {
                                cachedDirectoryStructureHost.addOrDeleteFileOrDirectory(fileOrDirectory, fileOrDirectoryPath);
                            }
                            nextSourceFileVersion(fileOrDirectoryPath);
                            var config = parsedConfigs === null || parsedConfigs === void 0 ? void 0 : parsedConfigs.get(configPath);
                            if (!(config === null || config === void 0 ? void 0 : config.parsedCommandLine))
                                return;
                            if (ts.isIgnoredFileFromWildCardWatching({
                                watchedDirPath: toPath(directory),
                                fileOrDirectory: fileOrDirectory,
                                fileOrDirectoryPath: fileOrDirectoryPath,
                                configFileName: configFileName,
                                options: config.parsedCommandLine.options,
                                program: config.parsedCommandLine.fileNames,
                                currentDirectory: currentDirectory,
                                useCaseSensitiveFileNames: useCaseSensitiveFileNames,
                                writeLog: writeLog,
                                toPath: toPath,
                            }))
                                return;
                            // Reload is pending, do the reload
                            if (config.reloadLevel !== ts.ConfigFileProgramReloadLevel.Full) {
                                config.reloadLevel = ts.ConfigFileProgramReloadLevel.Partial;
                                // Schedule Update the program
                                scheduleProgramUpdate();
                            }
                        }, flags, ((_a = commandLine.parsedCommandLine) === null || _a === void 0 ? void 0 : _a.watchOptions) || watchOptions, ts.WatchType.WildcardDirectoryOfReferencedProject);
                    });
                }
                else if (commandLine.watchedDirectories) {
                    ts.clearMap(commandLine.watchedDirectories, ts.closeFileWatcherOf);
                    commandLine.watchedDirectories = undefined;
                }
                // Watch extended config files
                updateExtendedConfigFilesWatches(configPath, (_d = commandLine.parsedCommandLine) === null || _d === void 0 ? void 0 : _d.options, ((_e = commandLine.parsedCommandLine) === null || _e === void 0 ? void 0 : _e.watchOptions) || watchOptions, ts.WatchType.ExtendedConfigOfReferencedProject);
            }
        }
        ts.createWatchProgram = createWatchProgram;
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var UpToDateStatusType;
        (function (UpToDateStatusType) {
            UpToDateStatusType[UpToDateStatusType["Unbuildable"] = 0] = "Unbuildable";
            UpToDateStatusType[UpToDateStatusType["UpToDate"] = 1] = "UpToDate";
            /**
             * The project appears out of date because its upstream inputs are newer than its outputs,
             * but all of its outputs are actually newer than the previous identical outputs of its (.d.ts) inputs.
             * This means we can Pseudo-build (just touch timestamps), as if we had actually built this project.
             */
            UpToDateStatusType[UpToDateStatusType["UpToDateWithUpstreamTypes"] = 2] = "UpToDateWithUpstreamTypes";
            /**
             * The project appears out of date because its upstream inputs are newer than its outputs,
             * but all of its outputs are actually newer than the previous identical outputs of its (.d.ts) inputs.
             * This means we can Pseudo-build (just manipulate outputs), as if we had actually built this project.
             */
            UpToDateStatusType[UpToDateStatusType["OutOfDateWithPrepend"] = 3] = "OutOfDateWithPrepend";
            UpToDateStatusType[UpToDateStatusType["OutputMissing"] = 4] = "OutputMissing";
            UpToDateStatusType[UpToDateStatusType["OutOfDateWithSelf"] = 5] = "OutOfDateWithSelf";
            UpToDateStatusType[UpToDateStatusType["OutOfDateWithUpstream"] = 6] = "OutOfDateWithUpstream";
            UpToDateStatusType[UpToDateStatusType["UpstreamOutOfDate"] = 7] = "UpstreamOutOfDate";
            UpToDateStatusType[UpToDateStatusType["UpstreamBlocked"] = 8] = "UpstreamBlocked";
            UpToDateStatusType[UpToDateStatusType["ComputingUpstream"] = 9] = "ComputingUpstream";
            UpToDateStatusType[UpToDateStatusType["TsVersionOutputOfDate"] = 10] = "TsVersionOutputOfDate";
            /**
             * Projects with no outputs (i.e. "solution" files)
             */
            UpToDateStatusType[UpToDateStatusType["ContainerOnly"] = 11] = "ContainerOnly";
        })(UpToDateStatusType = ts.UpToDateStatusType || (ts.UpToDateStatusType = {}));
        function resolveConfigFileProjectName(project) {
            if (ts.fileExtensionIs(project, ".json" /* Json */)) {
                return project;
            }
            return ts.combinePaths(project, "tsconfig.json");
        }
        ts.resolveConfigFileProjectName = resolveConfigFileProjectName;
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        var minimumDate = new Date(-8640000000000000);
        var maximumDate = new Date(8640000000000000);
        var BuildResultFlags;
        (function (BuildResultFlags) {
            BuildResultFlags[BuildResultFlags["None"] = 0] = "None";
            /**
             * No errors of any kind occurred during build
             */
            BuildResultFlags[BuildResultFlags["Success"] = 1] = "Success";
            /**
             * None of the .d.ts files emitted by this build were
             * different from the existing files on disk
             */
            BuildResultFlags[BuildResultFlags["DeclarationOutputUnchanged"] = 2] = "DeclarationOutputUnchanged";
            BuildResultFlags[BuildResultFlags["ConfigFileErrors"] = 4] = "ConfigFileErrors";
            BuildResultFlags[BuildResultFlags["SyntaxErrors"] = 8] = "SyntaxErrors";
            BuildResultFlags[BuildResultFlags["TypeErrors"] = 16] = "TypeErrors";
            BuildResultFlags[BuildResultFlags["DeclarationEmitErrors"] = 32] = "DeclarationEmitErrors";
            BuildResultFlags[BuildResultFlags["EmitErrors"] = 64] = "EmitErrors";
            BuildResultFlags[BuildResultFlags["AnyErrors"] = 124] = "AnyErrors";
        })(BuildResultFlags || (BuildResultFlags = {}));
        function getOrCreateValueFromConfigFileMap(configFileMap, resolved, createT) {
            var existingValue = configFileMap.get(resolved);
            var newValue;
            if (!existingValue) {
                newValue = createT();
                configFileMap.set(resolved, newValue);
            }
            return existingValue || newValue;
        }
        function getOrCreateValueMapFromConfigFileMap(configFileMap, resolved) {
            return getOrCreateValueFromConfigFileMap(configFileMap, resolved, function () { return new ts.Map(); });
        }
        function newer(date1, date2) {
            return date2 > date1 ? date2 : date1;
        }
        function isDeclarationFile(fileName) {
            return ts.fileExtensionIs(fileName, ".d.ts" /* Dts */);
        }
        /*@internal*/
        function isCircularBuildOrder(buildOrder) {
            return !!buildOrder && !!buildOrder.buildOrder;
        }
        ts.isCircularBuildOrder = isCircularBuildOrder;
        /*@internal*/
        function getBuildOrderFromAnyBuildOrder(anyBuildOrder) {
            return isCircularBuildOrder(anyBuildOrder) ? anyBuildOrder.buildOrder : anyBuildOrder;
        }
        ts.getBuildOrderFromAnyBuildOrder = getBuildOrderFromAnyBuildOrder;
        /**
         * Create a function that reports watch status by writing to the system and handles the formating of the diagnostic
         */
        function createBuilderStatusReporter(system, pretty) {
            return function (diagnostic) {
                var output = pretty ? "[" + ts.formatColorAndReset(ts.getLocaleTimeString(system), ts.ForegroundColorEscapeSequences.Grey) + "] " : ts.getLocaleTimeString(system) + " - ";
                output += "" + ts.flattenDiagnosticMessageText(diagnostic.messageText, system.newLine) + (system.newLine + system.newLine);
                system.write(output);
            };
        }
        ts.createBuilderStatusReporter = createBuilderStatusReporter;
        function createSolutionBuilderHostBase(system, createProgram, reportDiagnostic, reportSolutionBuilderStatus) {
            var host = ts.createProgramHost(system, createProgram);
            host.getModifiedTime = system.getModifiedTime ? function (path) { return system.getModifiedTime(path); } : ts.returnUndefined;
            host.setModifiedTime = system.setModifiedTime ? function (path, date) { return system.setModifiedTime(path, date); } : ts.noop;
            host.deleteFile = system.deleteFile ? function (path) { return system.deleteFile(path); } : ts.noop;
            host.reportDiagnostic = reportDiagnostic || ts.createDiagnosticReporter(system);
            host.reportSolutionBuilderStatus = reportSolutionBuilderStatus || createBuilderStatusReporter(system);
            host.now = ts.maybeBind(system, system.now); // For testing
            return host;
        }
        function createSolutionBuilderHost(system, createProgram, reportDiagnostic, reportSolutionBuilderStatus, reportErrorSummary) {
            if (system === void 0) {
                system = ts.sys;
            }
            var host = createSolutionBuilderHostBase(system, createProgram, reportDiagnostic, reportSolutionBuilderStatus);
            host.reportErrorSummary = reportErrorSummary;
            return host;
        }
        ts.createSolutionBuilderHost = createSolutionBuilderHost;
        function createSolutionBuilderWithWatchHost(system, createProgram, reportDiagnostic, reportSolutionBuilderStatus, reportWatchStatus) {
            if (system === void 0) {
                system = ts.sys;
            }
            var host = createSolutionBuilderHostBase(system, createProgram, reportDiagnostic, reportSolutionBuilderStatus);
            var watchHost = ts.createWatchHost(system, reportWatchStatus);
            ts.copyProperties(host, watchHost);
            return host;
        }
        ts.createSolutionBuilderWithWatchHost = createSolutionBuilderWithWatchHost;
        function getCompilerOptionsOfBuildOptions(buildOptions) {
            var result = {};
            ts.commonOptionsWithBuild.forEach(function (option) {
                if (ts.hasProperty(buildOptions, option.name))
                    result[option.name] = buildOptions[option.name];
            });
            return result;
        }
        function createSolutionBuilder(host, rootNames, defaultOptions) {
            return createSolutionBuilderWorker(/*watch*/ false, host, rootNames, defaultOptions);
        }
        ts.createSolutionBuilder = createSolutionBuilder;
        function createSolutionBuilderWithWatch(host, rootNames, defaultOptions, baseWatchOptions) {
            return createSolutionBuilderWorker(/*watch*/ true, host, rootNames, defaultOptions, baseWatchOptions);
        }
        ts.createSolutionBuilderWithWatch = createSolutionBuilderWithWatch;
        function createSolutionBuilderState(watch, hostOrHostWithWatch, rootNames, options, baseWatchOptions) {
            var host = hostOrHostWithWatch;
            var hostWithWatch = hostOrHostWithWatch;
            var currentDirectory = host.getCurrentDirectory();
            var getCanonicalFileName = ts.createGetCanonicalFileName(host.useCaseSensitiveFileNames());
            // State of the solution
            var baseCompilerOptions = getCompilerOptionsOfBuildOptions(options);
            var compilerHost = ts.createCompilerHostFromProgramHost(host, function () { return state.projectCompilerOptions; });
            ts.setGetSourceFileAsHashVersioned(compilerHost, host);
            compilerHost.getParsedCommandLine = function (fileName) { return parseConfigFile(state, fileName, toResolvedConfigFilePath(state, fileName)); };
            compilerHost.resolveModuleNames = ts.maybeBind(host, host.resolveModuleNames);
            compilerHost.resolveTypeReferenceDirectives = ts.maybeBind(host, host.resolveTypeReferenceDirectives);
            var moduleResolutionCache = !compilerHost.resolveModuleNames ? ts.createModuleResolutionCache(currentDirectory, getCanonicalFileName) : undefined;
            var typeReferenceDirectiveResolutionCache = !compilerHost.resolveTypeReferenceDirectives ? ts.createTypeReferenceDirectiveResolutionCache(currentDirectory, getCanonicalFileName, /*options*/ undefined, moduleResolutionCache === null || moduleResolutionCache === void 0 ? void 0 : moduleResolutionCache.getPackageJsonInfoCache()) : undefined;
            if (!compilerHost.resolveModuleNames) {
                var loader_3 = function (moduleName, containingFile, redirectedReference) { return ts.resolveModuleName(moduleName, containingFile, state.projectCompilerOptions, compilerHost, moduleResolutionCache, redirectedReference).resolvedModule; };
                compilerHost.resolveModuleNames = function (moduleNames, containingFile, _reusedNames, redirectedReference) {
                    return ts.loadWithLocalCache(ts.Debug.checkEachDefined(moduleNames), containingFile, redirectedReference, loader_3);
                };
            }
            if (!compilerHost.resolveTypeReferenceDirectives) {
                var loader_4 = function (moduleName, containingFile, redirectedReference) { return ts.resolveTypeReferenceDirective(moduleName, containingFile, state.projectCompilerOptions, compilerHost, redirectedReference, state.typeReferenceDirectiveResolutionCache).resolvedTypeReferenceDirective; };
                compilerHost.resolveTypeReferenceDirectives = function (typeReferenceDirectiveNames, containingFile, redirectedReference) {
                    return ts.loadWithLocalCache(ts.Debug.checkEachDefined(typeReferenceDirectiveNames), containingFile, redirectedReference, loader_4);
                };
            }
            var _a = ts.createWatchFactory(hostWithWatch, options), watchFile = _a.watchFile, watchDirectory = _a.watchDirectory, writeLog = _a.writeLog;
            var state = {
                host: host,
                hostWithWatch: hostWithWatch,
                currentDirectory: currentDirectory,
                getCanonicalFileName: getCanonicalFileName,
                parseConfigFileHost: ts.parseConfigHostFromCompilerHostLike(host),
                write: ts.maybeBind(host, host.trace),
                // State of solution
                options: options,
                baseCompilerOptions: baseCompilerOptions,
                rootNames: rootNames,
                baseWatchOptions: baseWatchOptions,
                resolvedConfigFilePaths: new ts.Map(),
                configFileCache: new ts.Map(),
                projectStatus: new ts.Map(),
                buildInfoChecked: new ts.Map(),
                extendedConfigCache: new ts.Map(),
                builderPrograms: new ts.Map(),
                diagnostics: new ts.Map(),
                projectPendingBuild: new ts.Map(),
                projectErrorsReported: new ts.Map(),
                compilerHost: compilerHost,
                moduleResolutionCache: moduleResolutionCache,
                typeReferenceDirectiveResolutionCache: typeReferenceDirectiveResolutionCache,
                // Mutable state
                buildOrder: undefined,
                readFileWithCache: function (f) { return host.readFile(f); },
                projectCompilerOptions: baseCompilerOptions,
                cache: undefined,
                allProjectBuildPending: true,
                needsSummary: true,
                watchAllProjectsPending: watch,
                currentInvalidatedProject: undefined,
                // Watch state
                watch: watch,
                allWatchedWildcardDirectories: new ts.Map(),
                allWatchedInputFiles: new ts.Map(),
                allWatchedConfigFiles: new ts.Map(),
                allWatchedExtendedConfigFiles: new ts.Map(),
                allWatchedPackageJsonFiles: new ts.Map(),
                lastCachedPackageJsonLookups: new ts.Map(),
                timerToBuildInvalidatedProject: undefined,
                reportFileChangeDetected: false,
                watchFile: watchFile,
                watchDirectory: watchDirectory,
                writeLog: writeLog,
            };
            return state;
        }
        function toPath(state, fileName) {
            return ts.toPath(fileName, state.currentDirectory, state.getCanonicalFileName);
        }
        function toResolvedConfigFilePath(state, fileName) {
            var resolvedConfigFilePaths = state.resolvedConfigFilePaths;
            var path = resolvedConfigFilePaths.get(fileName);
            if (path !== undefined)
                return path;
            var resolvedPath = toPath(state, fileName);
            resolvedConfigFilePaths.set(fileName, resolvedPath);
            return resolvedPath;
        }
        function isParsedCommandLine(entry) {
            return !!entry.options;
        }
        function getCachedParsedConfigFile(state, configFilePath) {
            var value = state.configFileCache.get(configFilePath);
            return value && isParsedCommandLine(value) ? value : undefined;
        }
        function parseConfigFile(state, configFileName, configFilePath) {
            var configFileCache = state.configFileCache;
            var value = configFileCache.get(configFilePath);
            if (value) {
                return isParsedCommandLine(value) ? value : undefined;
            }
            var diagnostic;
            var parseConfigFileHost = state.parseConfigFileHost, baseCompilerOptions = state.baseCompilerOptions, baseWatchOptions = state.baseWatchOptions, extendedConfigCache = state.extendedConfigCache, host = state.host;
            var parsed;
            if (host.getParsedCommandLine) {
                parsed = host.getParsedCommandLine(configFileName);
                if (!parsed)
                    diagnostic = ts.createCompilerDiagnostic(ts.Diagnostics.File_0_not_found, configFileName);
            }
            else {
                parseConfigFileHost.onUnRecoverableConfigFileDiagnostic = function (d) { return diagnostic = d; };
                parsed = ts.getParsedCommandLineOfConfigFile(configFileName, baseCompilerOptions, parseConfigFileHost, extendedConfigCache, baseWatchOptions);
                parseConfigFileHost.onUnRecoverableConfigFileDiagnostic = ts.noop;
            }
            configFileCache.set(configFilePath, parsed || diagnostic);
            return parsed;
        }
        function resolveProjectName(state, name) {
            return ts.resolveConfigFileProjectName(ts.resolvePath(state.currentDirectory, name));
        }
        function createBuildOrder(state, roots) {
            var temporaryMarks = new ts.Map();
            var permanentMarks = new ts.Map();
            var circularityReportStack = [];
            var buildOrder;
            var circularDiagnostics;
            for (var _i = 0, roots_1 = roots; _i < roots_1.length; _i++) {
                var root = roots_1[_i];
                visit(root);
            }
            return circularDiagnostics ?
                { buildOrder: buildOrder || ts.emptyArray, circularDiagnostics: circularDiagnostics } :
                buildOrder || ts.emptyArray;
            function visit(configFileName, inCircularContext) {
                var projPath = toResolvedConfigFilePath(state, configFileName);
                // Already visited
                if (permanentMarks.has(projPath))
                    return;
                // Circular
                if (temporaryMarks.has(projPath)) {
                    if (!inCircularContext) {
                        (circularDiagnostics || (circularDiagnostics = [])).push(ts.createCompilerDiagnostic(ts.Diagnostics.Project_references_may_not_form_a_circular_graph_Cycle_detected_Colon_0, circularityReportStack.join("\r\n")));
                    }
                    return;
                }
                temporaryMarks.set(projPath, true);
                circularityReportStack.push(configFileName);
                var parsed = parseConfigFile(state, configFileName, projPath);
                if (parsed && parsed.projectReferences) {
                    for (var _i = 0, _a = parsed.projectReferences; _i < _a.length; _i++) {
                        var ref = _a[_i];
                        var resolvedRefPath = resolveProjectName(state, ref.path);
                        visit(resolvedRefPath, inCircularContext || ref.circular);
                    }
                }
                circularityReportStack.pop();
                permanentMarks.set(projPath, true);
                (buildOrder || (buildOrder = [])).push(configFileName);
            }
        }
        function getBuildOrder(state) {
            return state.buildOrder || createStateBuildOrder(state);
        }
        function createStateBuildOrder(state) {
            var buildOrder = createBuildOrder(state, state.rootNames.map(function (f) { return resolveProjectName(state, f); }));
            // Clear all to ResolvedConfigFilePaths cache to start fresh
            state.resolvedConfigFilePaths.clear();
            // TODO(rbuckton): Should be a `Set`, but that requires changing the code below that uses `mutateMapSkippingNewValues`
            var currentProjects = new ts.Map(getBuildOrderFromAnyBuildOrder(buildOrder).map(function (resolved) { return [toResolvedConfigFilePath(state, resolved), true]; }));
            var noopOnDelete = { onDeleteValue: ts.noop };
            // Config file cache
            ts.mutateMapSkippingNewValues(state.configFileCache, currentProjects, noopOnDelete);
            ts.mutateMapSkippingNewValues(state.projectStatus, currentProjects, noopOnDelete);
            ts.mutateMapSkippingNewValues(state.buildInfoChecked, currentProjects, noopOnDelete);
            ts.mutateMapSkippingNewValues(state.builderPrograms, currentProjects, noopOnDelete);
            ts.mutateMapSkippingNewValues(state.diagnostics, currentProjects, noopOnDelete);
            ts.mutateMapSkippingNewValues(state.projectPendingBuild, currentProjects, noopOnDelete);
            ts.mutateMapSkippingNewValues(state.projectErrorsReported, currentProjects, noopOnDelete);
            // Remove watches for the program no longer in the solution
            if (state.watch) {
                ts.mutateMapSkippingNewValues(state.allWatchedConfigFiles, currentProjects, { onDeleteValue: ts.closeFileWatcher });
                state.allWatchedExtendedConfigFiles.forEach(function (watcher) {
                    watcher.projects.forEach(function (project) {
                        if (!currentProjects.has(project)) {
                            watcher.projects.delete(project);
                        }
                    });
                    watcher.close();
                });
                ts.mutateMapSkippingNewValues(state.allWatchedWildcardDirectories, currentProjects, { onDeleteValue: function (existingMap) { return existingMap.forEach(ts.closeFileWatcherOf); } });
                ts.mutateMapSkippingNewValues(state.allWatchedInputFiles, currentProjects, { onDeleteValue: function (existingMap) { return existingMap.forEach(ts.closeFileWatcher); } });
                ts.mutateMapSkippingNewValues(state.allWatchedPackageJsonFiles, currentProjects, { onDeleteValue: function (existingMap) { return existingMap.forEach(ts.closeFileWatcher); } });
            }
            return state.buildOrder = buildOrder;
        }
        function getBuildOrderFor(state, project, onlyReferences) {
            var resolvedProject = project && resolveProjectName(state, project);
            var buildOrderFromState = getBuildOrder(state);
            if (isCircularBuildOrder(buildOrderFromState))
                return buildOrderFromState;
            if (resolvedProject) {
                var projectPath_1 = toResolvedConfigFilePath(state, resolvedProject);
                var projectIndex = ts.findIndex(buildOrderFromState, function (configFileName) { return toResolvedConfigFilePath(state, configFileName) === projectPath_1; });
                if (projectIndex === -1)
                    return undefined;
            }
            var buildOrder = resolvedProject ? createBuildOrder(state, [resolvedProject]) : buildOrderFromState;
            ts.Debug.assert(!isCircularBuildOrder(buildOrder));
            ts.Debug.assert(!onlyReferences || resolvedProject !== undefined);
            ts.Debug.assert(!onlyReferences || buildOrder[buildOrder.length - 1] === resolvedProject);
            return onlyReferences ? buildOrder.slice(0, buildOrder.length - 1) : buildOrder;
        }
        function enableCache(state) {
            if (state.cache) {
                disableCache(state);
            }
            var compilerHost = state.compilerHost, host = state.host;
            var originalReadFileWithCache = state.readFileWithCache;
            var originalGetSourceFile = compilerHost.getSourceFile;
            var _a = ts.changeCompilerHostLikeToUseCache(host, function (fileName) { return toPath(state, fileName); }, function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return originalGetSourceFile.call.apply(originalGetSourceFile, __spreadArray([compilerHost], args, false));
            }), originalReadFile = _a.originalReadFile, originalFileExists = _a.originalFileExists, originalDirectoryExists = _a.originalDirectoryExists, originalCreateDirectory = _a.originalCreateDirectory, originalWriteFile = _a.originalWriteFile, getSourceFileWithCache = _a.getSourceFileWithCache, readFileWithCache = _a.readFileWithCache;
            state.readFileWithCache = readFileWithCache;
            compilerHost.getSourceFile = getSourceFileWithCache;
            state.cache = {
                originalReadFile: originalReadFile,
                originalFileExists: originalFileExists,
                originalDirectoryExists: originalDirectoryExists,
                originalCreateDirectory: originalCreateDirectory,
                originalWriteFile: originalWriteFile,
                originalReadFileWithCache: originalReadFileWithCache,
                originalGetSourceFile: originalGetSourceFile,
            };
        }
        function disableCache(state) {
            if (!state.cache)
                return;
            var cache = state.cache, host = state.host, compilerHost = state.compilerHost, extendedConfigCache = state.extendedConfigCache, moduleResolutionCache = state.moduleResolutionCache, typeReferenceDirectiveResolutionCache = state.typeReferenceDirectiveResolutionCache;
            host.readFile = cache.originalReadFile;
            host.fileExists = cache.originalFileExists;
            host.directoryExists = cache.originalDirectoryExists;
            host.createDirectory = cache.originalCreateDirectory;
            host.writeFile = cache.originalWriteFile;
            compilerHost.getSourceFile = cache.originalGetSourceFile;
            state.readFileWithCache = cache.originalReadFileWithCache;
            extendedConfigCache.clear();
            moduleResolutionCache === null || moduleResolutionCache === void 0 ? void 0 : moduleResolutionCache.clear();
            typeReferenceDirectiveResolutionCache === null || typeReferenceDirectiveResolutionCache === void 0 ? void 0 : typeReferenceDirectiveResolutionCache.clear();
            state.cache = undefined;
        }
        function clearProjectStatus(state, resolved) {
            state.projectStatus.delete(resolved);
            state.diagnostics.delete(resolved);
        }
        function addProjToQueue(_a, proj, reloadLevel) {
            var projectPendingBuild = _a.projectPendingBuild;
            var value = projectPendingBuild.get(proj);
            if (value === undefined) {
                projectPendingBuild.set(proj, reloadLevel);
            }
            else if (value < reloadLevel) {
                projectPendingBuild.set(proj, reloadLevel);
            }
        }
        function setupInitialBuild(state, cancellationToken) {
            // Set initial build if not already built
            if (!state.allProjectBuildPending)
                return;
            state.allProjectBuildPending = false;
            if (state.options.watch) {
                reportWatchStatus(state, ts.Diagnostics.Starting_compilation_in_watch_mode);
            }
            enableCache(state);
            var buildOrder = getBuildOrderFromAnyBuildOrder(getBuildOrder(state));
            buildOrder.forEach(function (configFileName) {
                return state.projectPendingBuild.set(toResolvedConfigFilePath(state, configFileName), ts.ConfigFileProgramReloadLevel.None);
            });
            if (cancellationToken) {
                cancellationToken.throwIfCancellationRequested();
            }
        }
        var InvalidatedProjectKind;
        (function (InvalidatedProjectKind) {
            InvalidatedProjectKind[InvalidatedProjectKind["Build"] = 0] = "Build";
            InvalidatedProjectKind[InvalidatedProjectKind["UpdateBundle"] = 1] = "UpdateBundle";
            InvalidatedProjectKind[InvalidatedProjectKind["UpdateOutputFileStamps"] = 2] = "UpdateOutputFileStamps";
        })(InvalidatedProjectKind = ts.InvalidatedProjectKind || (ts.InvalidatedProjectKind = {}));
        function doneInvalidatedProject(state, projectPath) {
            state.projectPendingBuild.delete(projectPath);
            state.currentInvalidatedProject = undefined;
            return state.diagnostics.has(projectPath) ?
                ts.ExitStatus.DiagnosticsPresent_OutputsSkipped :
                ts.ExitStatus.Success;
        }
        function createUpdateOutputFileStampsProject(state, project, projectPath, config, buildOrder) {
            var updateOutputFileStampsPending = true;
            return {
                kind: InvalidatedProjectKind.UpdateOutputFileStamps,
                project: project,
                projectPath: projectPath,
                buildOrder: buildOrder,
                getCompilerOptions: function () { return config.options; },
                getCurrentDirectory: function () { return state.currentDirectory; },
                updateOutputFileStatmps: function () {
                    updateOutputTimestamps(state, config, projectPath);
                    updateOutputFileStampsPending = false;
                },
                done: function () {
                    if (updateOutputFileStampsPending) {
                        updateOutputTimestamps(state, config, projectPath);
                    }
                    return doneInvalidatedProject(state, projectPath);
                }
            };
        }
        var BuildStep;
        (function (BuildStep) {
            BuildStep[BuildStep["CreateProgram"] = 0] = "CreateProgram";
            BuildStep[BuildStep["SyntaxDiagnostics"] = 1] = "SyntaxDiagnostics";
            BuildStep[BuildStep["SemanticDiagnostics"] = 2] = "SemanticDiagnostics";
            BuildStep[BuildStep["Emit"] = 3] = "Emit";
            BuildStep[BuildStep["EmitBundle"] = 4] = "EmitBundle";
            BuildStep[BuildStep["EmitBuildInfo"] = 5] = "EmitBuildInfo";
            BuildStep[BuildStep["BuildInvalidatedProjectOfBundle"] = 6] = "BuildInvalidatedProjectOfBundle";
            BuildStep[BuildStep["QueueReferencingProjects"] = 7] = "QueueReferencingProjects";
            BuildStep[BuildStep["Done"] = 8] = "Done";
        })(BuildStep || (BuildStep = {}));
        function createBuildOrUpdateInvalidedProject(kind, state, project, projectPath, projectIndex, config, buildOrder) {
            var step = kind === InvalidatedProjectKind.Build ? BuildStep.CreateProgram : BuildStep.EmitBundle;
            var program;
            var buildResult;
            var invalidatedProjectOfBundle;
            return kind === InvalidatedProjectKind.Build ?
                {
                    kind: kind,
                    project: project,
                    projectPath: projectPath,
                    buildOrder: buildOrder,
                    getCompilerOptions: function () { return config.options; },
                    getCurrentDirectory: function () { return state.currentDirectory; },
                    getBuilderProgram: function () { return withProgramOrUndefined(ts.identity); },
                    getProgram: function () {
                        return withProgramOrUndefined(function (program) { return program.getProgramOrUndefined(); });
                    },
                    getSourceFile: function (fileName) {
                        return withProgramOrUndefined(function (program) { return program.getSourceFile(fileName); });
                    },
                    getSourceFiles: function () {
                        return withProgramOrEmptyArray(function (program) { return program.getSourceFiles(); });
                    },
                    getOptionsDiagnostics: function (cancellationToken) {
                        return withProgramOrEmptyArray(function (program) { return program.getOptionsDiagnostics(cancellationToken); });
                    },
                    getGlobalDiagnostics: function (cancellationToken) {
                        return withProgramOrEmptyArray(function (program) { return program.getGlobalDiagnostics(cancellationToken); });
                    },
                    getConfigFileParsingDiagnostics: function () {
                        return withProgramOrEmptyArray(function (program) { return program.getConfigFileParsingDiagnostics(); });
                    },
                    getSyntacticDiagnostics: function (sourceFile, cancellationToken) {
                        return withProgramOrEmptyArray(function (program) { return program.getSyntacticDiagnostics(sourceFile, cancellationToken); });
                    },
                    getAllDependencies: function (sourceFile) {
                        return withProgramOrEmptyArray(function (program) { return program.getAllDependencies(sourceFile); });
                    },
                    getSemanticDiagnostics: function (sourceFile, cancellationToken) {
                        return withProgramOrEmptyArray(function (program) { return program.getSemanticDiagnostics(sourceFile, cancellationToken); });
                    },
                    getSemanticDiagnosticsOfNextAffectedFile: function (cancellationToken, ignoreSourceFile) {
                        return withProgramOrUndefined(function (program) {
                            return (program.getSemanticDiagnosticsOfNextAffectedFile) &&
                                program.getSemanticDiagnosticsOfNextAffectedFile(cancellationToken, ignoreSourceFile);
                        });
                    },
                    emit: function (targetSourceFile, writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers) {
                        if (targetSourceFile || emitOnlyDtsFiles) {
                            return withProgramOrUndefined(function (program) { var _a, _b; return program.emit(targetSourceFile, writeFile, cancellationToken, emitOnlyDtsFiles, customTransformers || ((_b = (_a = state.host).getCustomTransformers) === null || _b === void 0 ? void 0 : _b.call(_a, project))); });
                        }
                        executeSteps(BuildStep.SemanticDiagnostics, cancellationToken);
                        if (step === BuildStep.EmitBuildInfo) {
                            return emitBuildInfo(writeFile, cancellationToken);
                        }
                        if (step !== BuildStep.Emit)
                            return undefined;
                        return emit(writeFile, cancellationToken, customTransformers);
                    },
                    done: done
                } :
                {
                    kind: kind,
                    project: project,
                    projectPath: projectPath,
                    buildOrder: buildOrder,
                    getCompilerOptions: function () { return config.options; },
                    getCurrentDirectory: function () { return state.currentDirectory; },
                    emit: function (writeFile, customTransformers) {
                        if (step !== BuildStep.EmitBundle)
                            return invalidatedProjectOfBundle;
                        return emitBundle(writeFile, customTransformers);
                    },
                    done: done,
                };
            function done(cancellationToken, writeFile, customTransformers) {
                executeSteps(BuildStep.Done, cancellationToken, writeFile, customTransformers);
                return doneInvalidatedProject(state, projectPath);
            }
            function withProgramOrUndefined(action) {
                executeSteps(BuildStep.CreateProgram);
                return program && action(program);
            }
            function withProgramOrEmptyArray(action) {
                return withProgramOrUndefined(action) || ts.emptyArray;
            }
            function createProgram() {
                var _a, _b;
                ts.Debug.assert(program === undefined);
                if (state.options.dry) {
                    reportStatus(state, ts.Diagnostics.A_non_dry_build_would_build_project_0, project);
                    buildResult = BuildResultFlags.Success;
                    step = BuildStep.QueueReferencingProjects;
                    return;
                }
                if (state.options.verbose)
                    reportStatus(state, ts.Diagnostics.Building_project_0, project);
                if (config.fileNames.length === 0) {
                    reportAndStoreErrors(state, projectPath, ts.getConfigFileParsingDiagnostics(config));
                    // Nothing to build - must be a solution file, basically
                    buildResult = BuildResultFlags.None;
                    step = BuildStep.QueueReferencingProjects;
                    return;
                }
                var host = state.host, compilerHost = state.compilerHost;
                state.projectCompilerOptions = config.options;
                // Update module resolution cache if needed
                (_a = state.moduleResolutionCache) === null || _a === void 0 ? void 0 : _a.update(config.options);
                (_b = state.typeReferenceDirectiveResolutionCache) === null || _b === void 0 ? void 0 : _b.update(config.options);
                // Create program
                program = host.createProgram(config.fileNames, config.options, compilerHost, getOldProgram(state, projectPath, config), ts.getConfigFileParsingDiagnostics(config), config.projectReferences);
                state.lastCachedPackageJsonLookups.set(projectPath, state.moduleResolutionCache && ts.map(state.moduleResolutionCache.getPackageJsonInfoCache().entries(), function (_a) {
                    var path = _a[0], data = _a[1];
                    return [state.host.realpath ? toPath(state, state.host.realpath(path)) : path, data];
                }));
                if (state.watch) {
                    state.builderPrograms.set(projectPath, program);
                }
                step++;
            }
            function handleDiagnostics(diagnostics, errorFlags, errorType) {
                var _a;
                if (diagnostics.length) {
                    (_a = buildErrors(state, projectPath, program, config, diagnostics, errorFlags, errorType), buildResult = _a.buildResult, step = _a.step);
                }
                else {
                    step++;
                }
            }
            function getSyntaxDiagnostics(cancellationToken) {
                ts.Debug.assertIsDefined(program);
                handleDiagnostics(__spreadArray(__spreadArray(__spreadArray(__spreadArray([], program.getConfigFileParsingDiagnostics(), true), program.getOptionsDiagnostics(cancellationToken), true), program.getGlobalDiagnostics(cancellationToken), true), program.getSyntacticDiagnostics(/*sourceFile*/ undefined, cancellationToken), true), BuildResultFlags.SyntaxErrors, "Syntactic");
            }
            function getSemanticDiagnostics(cancellationToken) {
                handleDiagnostics(ts.Debug.checkDefined(program).getSemanticDiagnostics(/*sourceFile*/ undefined, cancellationToken), BuildResultFlags.TypeErrors, "Semantic");
            }
            function emit(writeFileCallback, cancellationToken, customTransformers) {
                var _a;
                var _b, _c;
                ts.Debug.assertIsDefined(program);
                ts.Debug.assert(step === BuildStep.Emit);
                // Before emitting lets backup state, so we can revert it back if there are declaration errors to handle emit and declaration errors correctly
                program.backupState();
                var declDiagnostics;
                var reportDeclarationDiagnostics = function (d) { return (declDiagnostics || (declDiagnostics = [])).push(d); };
                var outputFiles = [];
                var emitResult = ts.emitFilesAndReportErrors(program, reportDeclarationDiagnostics, 
                /*write*/ undefined, 
                /*reportSummary*/ undefined, function (name, text, writeByteOrderMark) { return outputFiles.push({ name: name, text: text, writeByteOrderMark: writeByteOrderMark }); }, cancellationToken, 
                /*emitOnlyDts*/ false, customTransformers || ((_c = (_b = state.host).getCustomTransformers) === null || _c === void 0 ? void 0 : _c.call(_b, project))).emitResult;
                // Don't emit .d.ts if there are decl file errors
                if (declDiagnostics) {
                    program.restoreState();
                    (_a = buildErrors(state, projectPath, program, config, declDiagnostics, BuildResultFlags.DeclarationEmitErrors, "Declaration file"), buildResult = _a.buildResult, step = _a.step);
                    return {
                        emitSkipped: true,
                        diagnostics: emitResult.diagnostics
                    };
                }
                // Actual Emit
                var host = state.host, compilerHost = state.compilerHost;
                var resultFlags = BuildResultFlags.DeclarationOutputUnchanged;
                var newestDeclarationFileContentChangedTime = minimumDate;
                var anyDtsChanged = false;
                var emitterDiagnostics = ts.createDiagnosticCollection();
                var emittedOutputs = new ts.Map();
                outputFiles.forEach(function (_a) {
                    var name = _a.name, text = _a.text, writeByteOrderMark = _a.writeByteOrderMark;
                    var priorChangeTime;
                    if (!anyDtsChanged && isDeclarationFile(name)) {
                        // Check for unchanged .d.ts files
                        if (host.fileExists(name) && state.readFileWithCache(name) === text) {
                            priorChangeTime = host.getModifiedTime(name);
                        }
                        else {
                            resultFlags &= ~BuildResultFlags.DeclarationOutputUnchanged;
                            anyDtsChanged = true;
                        }
                    }
                    emittedOutputs.set(toPath(state, name), name);
                    ts.writeFile(writeFileCallback ? { writeFile: writeFileCallback } : compilerHost, emitterDiagnostics, name, text, writeByteOrderMark);
                    if (priorChangeTime !== undefined) {
                        newestDeclarationFileContentChangedTime = newer(priorChangeTime, newestDeclarationFileContentChangedTime);
                    }
                });
                finishEmit(emitterDiagnostics, emittedOutputs, newestDeclarationFileContentChangedTime, 
                /*newestDeclarationFileContentChangedTimeIsMaximumDate*/ anyDtsChanged, outputFiles.length ? outputFiles[0].name : ts.getFirstProjectOutput(config, !host.useCaseSensitiveFileNames()), resultFlags);
                return emitResult;
            }
            function emitBuildInfo(writeFileCallback, cancellationToken) {
                ts.Debug.assertIsDefined(program);
                ts.Debug.assert(step === BuildStep.EmitBuildInfo);
                var emitResult = program.emitBuildInfo(writeFileCallback, cancellationToken);
                if (emitResult.diagnostics.length) {
                    reportErrors(state, emitResult.diagnostics);
                    state.diagnostics.set(projectPath, __spreadArray(__spreadArray([], state.diagnostics.get(projectPath), true), emitResult.diagnostics, true));
                    buildResult = BuildResultFlags.EmitErrors & buildResult;
                }
                if (emitResult.emittedFiles && state.write) {
                    emitResult.emittedFiles.forEach(function (name) { return listEmittedFile(state, config, name); });
                }
                afterProgramDone(state, program, config);
                step = BuildStep.QueueReferencingProjects;
                return emitResult;
            }
            function finishEmit(emitterDiagnostics, emittedOutputs, priorNewestUpdateTime, newestDeclarationFileContentChangedTimeIsMaximumDate, oldestOutputFileName, resultFlags) {
                var _a;
                var emitDiagnostics = emitterDiagnostics.getDiagnostics();
                if (emitDiagnostics.length) {
                    (_a = buildErrors(state, projectPath, program, config, emitDiagnostics, BuildResultFlags.EmitErrors, "Emit"), buildResult = _a.buildResult, step = _a.step);
                    return emitDiagnostics;
                }
                if (state.write) {
                    emittedOutputs.forEach(function (name) { return listEmittedFile(state, config, name); });
                }
                // Update time stamps for rest of the outputs
                var newestDeclarationFileContentChangedTime = updateOutputTimestampsWorker(state, config, priorNewestUpdateTime, ts.Diagnostics.Updating_unchanged_output_timestamps_of_project_0, emittedOutputs);
                state.diagnostics.delete(projectPath);
                state.projectStatus.set(projectPath, {
                    type: ts.UpToDateStatusType.UpToDate,
                    newestDeclarationFileContentChangedTime: newestDeclarationFileContentChangedTimeIsMaximumDate ?
                        maximumDate :
                        newestDeclarationFileContentChangedTime,
                    oldestOutputFileName: oldestOutputFileName
                });
                afterProgramDone(state, program, config);
                step = BuildStep.QueueReferencingProjects;
                buildResult = resultFlags;
                return emitDiagnostics;
            }
            function emitBundle(writeFileCallback, customTransformers) {
                var _a, _b;
                ts.Debug.assert(kind === InvalidatedProjectKind.UpdateBundle);
                if (state.options.dry) {
                    reportStatus(state, ts.Diagnostics.A_non_dry_build_would_update_output_of_project_0, project);
                    buildResult = BuildResultFlags.Success;
                    step = BuildStep.QueueReferencingProjects;
                    return undefined;
                }
                if (state.options.verbose)
                    reportStatus(state, ts.Diagnostics.Updating_output_of_project_0, project);
                // Update js, and source map
                var compilerHost = state.compilerHost;
                state.projectCompilerOptions = config.options;
                var outputFiles = ts.emitUsingBuildInfo(config, compilerHost, function (ref) {
                    var refName = resolveProjectName(state, ref.path);
                    return parseConfigFile(state, refName, toResolvedConfigFilePath(state, refName));
                }, customTransformers || ((_b = (_a = state.host).getCustomTransformers) === null || _b === void 0 ? void 0 : _b.call(_a, project)));
                if (ts.isString(outputFiles)) {
                    reportStatus(state, ts.Diagnostics.Cannot_update_output_of_project_0_because_there_was_error_reading_file_1, project, relName(state, outputFiles));
                    step = BuildStep.BuildInvalidatedProjectOfBundle;
                    return invalidatedProjectOfBundle = createBuildOrUpdateInvalidedProject(InvalidatedProjectKind.Build, state, project, projectPath, projectIndex, config, buildOrder);
                }
                // Actual Emit
                ts.Debug.assert(!!outputFiles.length);
                var emitterDiagnostics = ts.createDiagnosticCollection();
                var emittedOutputs = new ts.Map();
                outputFiles.forEach(function (_a) {
                    var name = _a.name, text = _a.text, writeByteOrderMark = _a.writeByteOrderMark;
                    emittedOutputs.set(toPath(state, name), name);
                    ts.writeFile(writeFileCallback ? { writeFile: writeFileCallback } : compilerHost, emitterDiagnostics, name, text, writeByteOrderMark);
                });
                var emitDiagnostics = finishEmit(emitterDiagnostics, emittedOutputs, minimumDate, 
                /*newestDeclarationFileContentChangedTimeIsMaximumDate*/ false, outputFiles[0].name, BuildResultFlags.DeclarationOutputUnchanged);
                return { emitSkipped: false, diagnostics: emitDiagnostics };
            }
            function executeSteps(till, cancellationToken, writeFile, customTransformers) {
                while (step <= till && step < BuildStep.Done) {
                    var currentStep = step;
                    switch (step) {
                        case BuildStep.CreateProgram:
                            createProgram();
                            break;
                        case BuildStep.SyntaxDiagnostics:
                            getSyntaxDiagnostics(cancellationToken);
                            break;
                        case BuildStep.SemanticDiagnostics:
                            getSemanticDiagnostics(cancellationToken);
                            break;
                        case BuildStep.Emit:
                            emit(writeFile, cancellationToken, customTransformers);
                            break;
                        case BuildStep.EmitBuildInfo:
                            emitBuildInfo(writeFile, cancellationToken);
                            break;
                        case BuildStep.EmitBundle:
                            emitBundle(writeFile, customTransformers);
                            break;
                        case BuildStep.BuildInvalidatedProjectOfBundle:
                            ts.Debug.checkDefined(invalidatedProjectOfBundle).done(cancellationToken, writeFile, customTransformers);
                            step = BuildStep.Done;
                            break;
                        case BuildStep.QueueReferencingProjects:
                            queueReferencingProjects(state, project, projectPath, projectIndex, config, buildOrder, ts.Debug.checkDefined(buildResult));
                            step++;
                            break;
                        // Should never be done
                        case BuildStep.Done:
                        default:
                            ts.assertType(step);
                    }
                    ts.Debug.assert(step > currentStep);
                }
            }
        }
        function needsBuild(_a, status, config) {
            var options = _a.options;
            if (status.type !== ts.UpToDateStatusType.OutOfDateWithPrepend || options.force)
                return true;
            return config.fileNames.length === 0 ||
                !!ts.getConfigFileParsingDiagnostics(config).length ||
                !ts.isIncrementalCompilation(config.options);
        }
        function getNextInvalidatedProject(state, buildOrder, reportQueue) {
            if (!state.projectPendingBuild.size)
                return undefined;
            if (isCircularBuildOrder(buildOrder))
                return undefined;
            if (state.currentInvalidatedProject) {
                // Only if same buildOrder the currentInvalidated project can be sent again
                return ts.arrayIsEqualTo(state.currentInvalidatedProject.buildOrder, buildOrder) ?
                    state.currentInvalidatedProject :
                    undefined;
            }
            var options = state.options, projectPendingBuild = state.projectPendingBuild;
            for (var projectIndex = 0; projectIndex < buildOrder.length; projectIndex++) {
                var project = buildOrder[projectIndex];
                var projectPath = toResolvedConfigFilePath(state, project);
                var reloadLevel = state.projectPendingBuild.get(projectPath);
                if (reloadLevel === undefined)
                    continue;
                if (reportQueue) {
                    reportQueue = false;
                    reportBuildQueue(state, buildOrder);
                }
                var config = parseConfigFile(state, project, projectPath);
                if (!config) {
                    reportParseConfigFileDiagnostic(state, projectPath);
                    projectPendingBuild.delete(projectPath);
                    continue;
                }
                if (reloadLevel === ts.ConfigFileProgramReloadLevel.Full) {
                    watchConfigFile(state, project, projectPath, config);
                    watchExtendedConfigFiles(state, projectPath, config);
                    watchWildCardDirectories(state, project, projectPath, config);
                    watchInputFiles(state, project, projectPath, config);
                    watchPackageJsonFiles(state, project, projectPath, config);
                }
                else if (reloadLevel === ts.ConfigFileProgramReloadLevel.Partial) {
                    // Update file names
                    config.fileNames = ts.getFileNamesFromConfigSpecs(config.options.configFile.configFileSpecs, ts.getDirectoryPath(project), config.options, state.parseConfigFileHost);
                    ts.updateErrorForNoInputFiles(config.fileNames, project, config.options.configFile.configFileSpecs, config.errors, ts.canJsonReportNoInputFiles(config.raw));
                    watchInputFiles(state, project, projectPath, config);
                    watchPackageJsonFiles(state, project, projectPath, config);
                }
                var status = getUpToDateStatus(state, config, projectPath);
                verboseReportProjectStatus(state, project, status);
                if (!options.force) {
                    if (status.type === ts.UpToDateStatusType.UpToDate) {
                        reportAndStoreErrors(state, projectPath, ts.getConfigFileParsingDiagnostics(config));
                        projectPendingBuild.delete(projectPath);
                        // Up to date, skip
                        if (options.dry) {
                            // In a dry build, inform the user of this fact
                            reportStatus(state, ts.Diagnostics.Project_0_is_up_to_date, project);
                        }
                        continue;
                    }
                    if (status.type === ts.UpToDateStatusType.UpToDateWithUpstreamTypes) {
                        reportAndStoreErrors(state, projectPath, ts.getConfigFileParsingDiagnostics(config));
                        return createUpdateOutputFileStampsProject(state, project, projectPath, config, buildOrder);
                    }
                }
                if (status.type === ts.UpToDateStatusType.UpstreamBlocked) {
                    reportAndStoreErrors(state, projectPath, ts.getConfigFileParsingDiagnostics(config));
                    projectPendingBuild.delete(projectPath);
                    if (options.verbose) {
                        reportStatus(state, status.upstreamProjectBlocked ?
                            ts.Diagnostics.Skipping_build_of_project_0_because_its_dependency_1_was_not_built :
                            ts.Diagnostics.Skipping_build_of_project_0_because_its_dependency_1_has_errors, project, status.upstreamProjectName);
                    }
                    continue;
                }
                if (status.type === ts.UpToDateStatusType.ContainerOnly) {
                    reportAndStoreErrors(state, projectPath, ts.getConfigFileParsingDiagnostics(config));
                    projectPendingBuild.delete(projectPath);
                    // Do nothing
                    continue;
                }
                return createBuildOrUpdateInvalidedProject(needsBuild(state, status, config) ?
                    InvalidatedProjectKind.Build :
                    InvalidatedProjectKind.UpdateBundle, state, project, projectPath, projectIndex, config, buildOrder);
            }
            return undefined;
        }
        function listEmittedFile(_a, proj, file) {
            var write = _a.write;
            if (write && proj.options.listEmittedFiles) {
                write("TSFILE: " + file);
            }
        }
        function getOldProgram(_a, proj, parsed) {
            var options = _a.options, builderPrograms = _a.builderPrograms, compilerHost = _a.compilerHost;
            if (options.force)
                return undefined;
            var value = builderPrograms.get(proj);
            if (value)
                return value;
            return ts.readBuilderProgram(parsed.options, compilerHost);
        }
        function afterProgramDone(state, program, config) {
            if (program) {
                if (program && state.write)
                    ts.listFiles(program, state.write);
                if (state.host.afterProgramEmitAndDiagnostics) {
                    state.host.afterProgramEmitAndDiagnostics(program);
                }
                program.releaseProgram();
            }
            else if (state.host.afterEmitBundle) {
                state.host.afterEmitBundle(config);
            }
            state.projectCompilerOptions = state.baseCompilerOptions;
        }
        function buildErrors(state, resolvedPath, program, config, diagnostics, buildResult, errorType) {
            var canEmitBuildInfo = !(buildResult & BuildResultFlags.SyntaxErrors) && program && !ts.outFile(program.getCompilerOptions());
            reportAndStoreErrors(state, resolvedPath, diagnostics);
            state.projectStatus.set(resolvedPath, { type: ts.UpToDateStatusType.Unbuildable, reason: errorType + " errors" });
            if (canEmitBuildInfo)
                return { buildResult: buildResult, step: BuildStep.EmitBuildInfo };
            afterProgramDone(state, program, config);
            return { buildResult: buildResult, step: BuildStep.QueueReferencingProjects };
        }
        function checkConfigFileUpToDateStatus(state, configFile, oldestOutputFileTime, oldestOutputFileName) {
            // Check tsconfig time
            var tsconfigTime = ts.getModifiedTime(state.host, configFile);
            if (oldestOutputFileTime < tsconfigTime) {
                return {
                    type: ts.UpToDateStatusType.OutOfDateWithSelf,
                    outOfDateOutputFileName: oldestOutputFileName,
                    newerInputFileName: configFile
                };
            }
        }
        function getUpToDateStatusWorker(state, project, resolvedPath) {
            var force = !!state.options.force;
            var newestInputFileName = undefined;
            var newestInputFileTime = minimumDate;
            var host = state.host;
            // Get timestamps of input files
            for (var _i = 0, _a = project.fileNames; _i < _a.length; _i++) {
                var inputFile = _a[_i];
                if (!host.fileExists(inputFile)) {
                    return {
                        type: ts.UpToDateStatusType.Unbuildable,
                        reason: inputFile + " does not exist"
                    };
                }
                if (!force) {
                    var inputTime = ts.getModifiedTime(host, inputFile);
                    host.getModifiedTime(inputFile);
                    if (inputTime > newestInputFileTime) {
                        newestInputFileName = inputFile;
                        newestInputFileTime = inputTime;
                    }
                }
            }
            // Container if no files are specified in the project
            if (!project.fileNames.length && !ts.canJsonReportNoInputFiles(project.raw)) {
                return {
                    type: ts.UpToDateStatusType.ContainerOnly
                };
            }
            // Collect the expected outputs of this project
            var outputs = ts.getAllProjectOutputs(project, !host.useCaseSensitiveFileNames());
            // Now see if all outputs are newer than the newest input
            var oldestOutputFileName = "(none)";
            var oldestOutputFileTime = maximumDate;
            var newestOutputFileName = "(none)";
            var newestOutputFileTime = minimumDate;
            var missingOutputFileName;
            var newestDeclarationFileContentChangedTime = minimumDate;
            var isOutOfDateWithInputs = false;
            if (!force) {
                for (var _b = 0, outputs_1 = outputs; _b < outputs_1.length; _b++) {
                    var output = outputs_1[_b];
                    // Output is missing; can stop checking
                    // Don't immediately return because we can still be upstream-blocked, which is a higher-priority status
                    if (!host.fileExists(output)) {
                        missingOutputFileName = output;
                        break;
                    }
                    var outputTime = ts.getModifiedTime(host, output);
                    if (outputTime < oldestOutputFileTime) {
                        oldestOutputFileTime = outputTime;
                        oldestOutputFileName = output;
                    }
                    // If an output is older than the newest input, we can stop checking
                    // Don't immediately return because we can still be upstream-blocked, which is a higher-priority status
                    if (outputTime < newestInputFileTime) {
                        isOutOfDateWithInputs = true;
                        break;
                    }
                    if (outputTime > newestOutputFileTime) {
                        newestOutputFileTime = outputTime;
                        newestOutputFileName = output;
                    }
                    // Keep track of when the most recent time a .d.ts file was changed.
                    // In addition to file timestamps, we also keep track of when a .d.ts file
                    // had its file touched but not had its contents changed - this allows us
                    // to skip a downstream typecheck
                    if (isDeclarationFile(output)) {
                        var outputModifiedTime = ts.getModifiedTime(host, output);
                        newestDeclarationFileContentChangedTime = newer(newestDeclarationFileContentChangedTime, outputModifiedTime);
                    }
                }
            }
            var pseudoUpToDate = false;
            var usesPrepend = false;
            var upstreamChangedProject;
            if (project.projectReferences) {
                state.projectStatus.set(resolvedPath, { type: ts.UpToDateStatusType.ComputingUpstream });
                for (var _c = 0, _d = project.projectReferences; _c < _d.length; _c++) {
                    var ref = _d[_c];
                    usesPrepend = usesPrepend || !!(ref.prepend);
                    var resolvedRef = ts.resolveProjectReferencePath(ref);
                    var resolvedRefPath = toResolvedConfigFilePath(state, resolvedRef);
                    var refStatus = getUpToDateStatus(state, parseConfigFile(state, resolvedRef, resolvedRefPath), resolvedRefPath);
                    // Its a circular reference ignore the status of this project
                    if (refStatus.type === ts.UpToDateStatusType.ComputingUpstream ||
                        refStatus.type === ts.UpToDateStatusType.ContainerOnly) { // Container only ignore this project
                        continue;
                    }
                    // An upstream project is blocked
                    if (refStatus.type === ts.UpToDateStatusType.Unbuildable ||
                        refStatus.type === ts.UpToDateStatusType.UpstreamBlocked) {
                        return {
                            type: ts.UpToDateStatusType.UpstreamBlocked,
                            upstreamProjectName: ref.path,
                            upstreamProjectBlocked: refStatus.type === ts.UpToDateStatusType.UpstreamBlocked
                        };
                    }
                    // If the upstream project is out of date, then so are we (someone shouldn't have asked, though?)
                    if (refStatus.type !== ts.UpToDateStatusType.UpToDate) {
                        return {
                            type: ts.UpToDateStatusType.UpstreamOutOfDate,
                            upstreamProjectName: ref.path
                        };
                    }
                    // Check oldest output file name only if there is no missing output file name
                    // (a check we will have skipped if this is a forced build)
                    if (!force && !missingOutputFileName) {
                        // If the upstream project's newest file is older than our oldest output, we
                        // can't be out of date because of it
                        if (refStatus.newestInputFileTime && refStatus.newestInputFileTime <= oldestOutputFileTime) {
                            continue;
                        }
                        // If the upstream project has only change .d.ts files, and we've built
                        // *after* those files, then we're "psuedo up to date" and eligible for a fast rebuild
                        if (refStatus.newestDeclarationFileContentChangedTime && refStatus.newestDeclarationFileContentChangedTime <= oldestOutputFileTime) {
                            pseudoUpToDate = true;
                            upstreamChangedProject = ref.path;
                            continue;
                        }
                        // We have an output older than an upstream output - we are out of date
                        ts.Debug.assert(oldestOutputFileName !== undefined, "Should have an oldest output filename here");
                        return {
                            type: ts.UpToDateStatusType.OutOfDateWithUpstream,
                            outOfDateOutputFileName: oldestOutputFileName,
                            newerProjectName: ref.path
                        };
                    }
                }
            }
            if (missingOutputFileName !== undefined) {
                return {
                    type: ts.UpToDateStatusType.OutputMissing,
                    missingOutputFileName: missingOutputFileName
                };
            }
            if (isOutOfDateWithInputs) {
                return {
                    type: ts.UpToDateStatusType.OutOfDateWithSelf,
                    outOfDateOutputFileName: oldestOutputFileName,
                    newerInputFileName: newestInputFileName
                };
            }
            else {
                // Check tsconfig time
                var configStatus = checkConfigFileUpToDateStatus(state, project.options.configFilePath, oldestOutputFileTime, oldestOutputFileName);
                if (configStatus)
                    return configStatus;
                // Check extended config time
                var extendedConfigStatus = ts.forEach(project.options.configFile.extendedSourceFiles || ts.emptyArray, function (configFile) { return checkConfigFileUpToDateStatus(state, configFile, oldestOutputFileTime, oldestOutputFileName); });
                if (extendedConfigStatus)
                    return extendedConfigStatus;
                // Check package file time
                var dependentPackageFileStatus = ts.forEach(state.lastCachedPackageJsonLookups.get(resolvedPath) || ts.emptyArray, function (_a) {
                    var path = _a[0];
                    return checkConfigFileUpToDateStatus(state, path, oldestOutputFileTime, oldestOutputFileName);
                });
                if (dependentPackageFileStatus)
                    return dependentPackageFileStatus;
            }
            if (!force && !state.buildInfoChecked.has(resolvedPath)) {
                state.buildInfoChecked.set(resolvedPath, true);
                var buildInfoPath = ts.getTsBuildInfoEmitOutputFilePath(project.options);
                if (buildInfoPath) {
                    var value = state.readFileWithCache(buildInfoPath);
                    var buildInfo = value && ts.getBuildInfo(value);
                    if (buildInfo && (buildInfo.bundle || buildInfo.program) && buildInfo.version !== ts.version) {
                        return {
                            type: ts.UpToDateStatusType.TsVersionOutputOfDate,
                            version: buildInfo.version
                        };
                    }
                }
            }
            if (usesPrepend && pseudoUpToDate) {
                return {
                    type: ts.UpToDateStatusType.OutOfDateWithPrepend,
                    outOfDateOutputFileName: oldestOutputFileName,
                    newerProjectName: upstreamChangedProject
                };
            }
            // Up to date
            return {
                type: pseudoUpToDate ? ts.UpToDateStatusType.UpToDateWithUpstreamTypes : ts.UpToDateStatusType.UpToDate,
                newestDeclarationFileContentChangedTime: newestDeclarationFileContentChangedTime,
                newestInputFileTime: newestInputFileTime,
                newestOutputFileTime: newestOutputFileTime,
                newestInputFileName: newestInputFileName,
                newestOutputFileName: newestOutputFileName,
                oldestOutputFileName: oldestOutputFileName
            };
        }
        function getUpToDateStatus(state, project, resolvedPath) {
            if (project === undefined) {
                return { type: ts.UpToDateStatusType.Unbuildable, reason: "File deleted mid-build" };
            }
            var prior = state.projectStatus.get(resolvedPath);
            if (prior !== undefined) {
                return prior;
            }
            var actual = getUpToDateStatusWorker(state, project, resolvedPath);
            state.projectStatus.set(resolvedPath, actual);
            return actual;
        }
        function updateOutputTimestampsWorker(state, proj, priorNewestUpdateTime, verboseMessage, skipOutputs) {
            if (proj.options.noEmit)
                return priorNewestUpdateTime;
            var host = state.host;
            var outputs = ts.getAllProjectOutputs(proj, !host.useCaseSensitiveFileNames());
            if (!skipOutputs || outputs.length !== skipOutputs.size) {
                var reportVerbose = !!state.options.verbose;
                var now = host.now ? host.now() : new Date();
                for (var _i = 0, outputs_2 = outputs; _i < outputs_2.length; _i++) {
                    var file = outputs_2[_i];
                    if (skipOutputs && skipOutputs.has(toPath(state, file))) {
                        continue;
                    }
                    if (reportVerbose) {
                        reportVerbose = false;
                        reportStatus(state, verboseMessage, proj.options.configFilePath);
                    }
                    if (isDeclarationFile(file)) {
                        priorNewestUpdateTime = newer(priorNewestUpdateTime, ts.getModifiedTime(host, file));
                    }
                    host.setModifiedTime(file, now);
                }
            }
            return priorNewestUpdateTime;
        }
        function updateOutputTimestamps(state, proj, resolvedPath) {
            if (state.options.dry) {
                return reportStatus(state, ts.Diagnostics.A_non_dry_build_would_update_timestamps_for_output_of_project_0, proj.options.configFilePath);
            }
            var priorNewestUpdateTime = updateOutputTimestampsWorker(state, proj, minimumDate, ts.Diagnostics.Updating_output_timestamps_of_project_0);
            state.projectStatus.set(resolvedPath, {
                type: ts.UpToDateStatusType.UpToDate,
                newestDeclarationFileContentChangedTime: priorNewestUpdateTime,
                oldestOutputFileName: ts.getFirstProjectOutput(proj, !state.host.useCaseSensitiveFileNames())
            });
        }
        function queueReferencingProjects(state, project, projectPath, projectIndex, config, buildOrder, buildResult) {
            // Queue only if there are no errors
            if (buildResult & BuildResultFlags.AnyErrors)
                return;
            // Only composite projects can be referenced by other projects
            if (!config.options.composite)
                return;
            // Always use build order to queue projects
            for (var index = projectIndex + 1; index < buildOrder.length; index++) {
                var nextProject = buildOrder[index];
                var nextProjectPath = toResolvedConfigFilePath(state, nextProject);
                if (state.projectPendingBuild.has(nextProjectPath))
                    continue;
                var nextProjectConfig = parseConfigFile(state, nextProject, nextProjectPath);
                if (!nextProjectConfig || !nextProjectConfig.projectReferences)
                    continue;
                for (var _i = 0, _a = nextProjectConfig.projectReferences; _i < _a.length; _i++) {
                    var ref = _a[_i];
                    var resolvedRefPath = resolveProjectName(state, ref.path);
                    if (toResolvedConfigFilePath(state, resolvedRefPath) !== projectPath)
                        continue;
                    // If the project is referenced with prepend, always build downstream projects,
                    // If declaration output is changed, build the project
                    // otherwise mark the project UpToDateWithUpstreamTypes so it updates output time stamps
                    var status = state.projectStatus.get(nextProjectPath);
                    if (status) {
                        switch (status.type) {
                            case ts.UpToDateStatusType.UpToDate:
                                if (buildResult & BuildResultFlags.DeclarationOutputUnchanged) {
                                    if (ref.prepend) {
                                        state.projectStatus.set(nextProjectPath, {
                                            type: ts.UpToDateStatusType.OutOfDateWithPrepend,
                                            outOfDateOutputFileName: status.oldestOutputFileName,
                                            newerProjectName: project
                                        });
                                    }
                                    else {
                                        status.type = ts.UpToDateStatusType.UpToDateWithUpstreamTypes;
                                    }
                                    break;
                                }
                            // falls through
                            case ts.UpToDateStatusType.UpToDateWithUpstreamTypes:
                            case ts.UpToDateStatusType.OutOfDateWithPrepend:
                                if (!(buildResult & BuildResultFlags.DeclarationOutputUnchanged)) {
                                    state.projectStatus.set(nextProjectPath, {
                                        type: ts.UpToDateStatusType.OutOfDateWithUpstream,
                                        outOfDateOutputFileName: status.type === ts.UpToDateStatusType.OutOfDateWithPrepend ? status.outOfDateOutputFileName : status.oldestOutputFileName,
                                        newerProjectName: project
                                    });
                                }
                                break;
                            case ts.UpToDateStatusType.UpstreamBlocked:
                                if (toResolvedConfigFilePath(state, resolveProjectName(state, status.upstreamProjectName)) === projectPath) {
                                    clearProjectStatus(state, nextProjectPath);
                                }
                                break;
                        }
                    }
                    addProjToQueue(state, nextProjectPath, ts.ConfigFileProgramReloadLevel.None);
                    break;
                }
            }
        }
        function build(state, project, cancellationToken, writeFile, getCustomTransformers, onlyReferences) {
            var buildOrder = getBuildOrderFor(state, project, onlyReferences);
            if (!buildOrder)
                return ts.ExitStatus.InvalidProject_OutputsSkipped;
            setupInitialBuild(state, cancellationToken);
            var reportQueue = true;
            var successfulProjects = 0;
            while (true) {
                var invalidatedProject = getNextInvalidatedProject(state, buildOrder, reportQueue);
                if (!invalidatedProject)
                    break;
                reportQueue = false;
                invalidatedProject.done(cancellationToken, writeFile, getCustomTransformers === null || getCustomTransformers === void 0 ? void 0 : getCustomTransformers(invalidatedProject.project));
                if (!state.diagnostics.has(invalidatedProject.projectPath))
                    successfulProjects++;
            }
            disableCache(state);
            reportErrorSummary(state, buildOrder);
            startWatching(state, buildOrder);
            return isCircularBuildOrder(buildOrder)
                ? ts.ExitStatus.ProjectReferenceCycle_OutputsSkipped
                : !buildOrder.some(function (p) { return state.diagnostics.has(toResolvedConfigFilePath(state, p)); })
                    ? ts.ExitStatus.Success
                    : successfulProjects
                        ? ts.ExitStatus.DiagnosticsPresent_OutputsGenerated
                        : ts.ExitStatus.DiagnosticsPresent_OutputsSkipped;
        }
        function clean(state, project, onlyReferences) {
            var buildOrder = getBuildOrderFor(state, project, onlyReferences);
            if (!buildOrder)
                return ts.ExitStatus.InvalidProject_OutputsSkipped;
            if (isCircularBuildOrder(buildOrder)) {
                reportErrors(state, buildOrder.circularDiagnostics);
                return ts.ExitStatus.ProjectReferenceCycle_OutputsSkipped;
            }
            var options = state.options, host = state.host;
            var filesToDelete = options.dry ? [] : undefined;
            for (var _i = 0, buildOrder_1 = buildOrder; _i < buildOrder_1.length; _i++) {
                var proj = buildOrder_1[_i];
                var resolvedPath = toResolvedConfigFilePath(state, proj);
                var parsed = parseConfigFile(state, proj, resolvedPath);
                if (parsed === undefined) {
                    // File has gone missing; fine to ignore here
                    reportParseConfigFileDiagnostic(state, resolvedPath);
                    continue;
                }
                var outputs = ts.getAllProjectOutputs(parsed, !host.useCaseSensitiveFileNames());
                if (!outputs.length)
                    continue;
                var inputFileNames = new ts.Set(parsed.fileNames.map(function (f) { return toPath(state, f); }));
                for (var _a = 0, outputs_3 = outputs; _a < outputs_3.length; _a++) {
                    var output = outputs_3[_a];
                    // If output name is same as input file name, do not delete and ignore the error
                    if (inputFileNames.has(toPath(state, output)))
                        continue;
                    if (host.fileExists(output)) {
                        if (filesToDelete) {
                            filesToDelete.push(output);
                        }
                        else {
                            host.deleteFile(output);
                            invalidateProject(state, resolvedPath, ts.ConfigFileProgramReloadLevel.None);
                        }
                    }
                }
            }
            if (filesToDelete) {
                reportStatus(state, ts.Diagnostics.A_non_dry_build_would_delete_the_following_files_Colon_0, filesToDelete.map(function (f) { return "\r\n * " + f; }).join(""));
            }
            return ts.ExitStatus.Success;
        }
        function invalidateProject(state, resolved, reloadLevel) {
            // If host implements getParsedCommandLine, we cant get list of files from parseConfigFileHost
            if (state.host.getParsedCommandLine && reloadLevel === ts.ConfigFileProgramReloadLevel.Partial) {
                reloadLevel = ts.ConfigFileProgramReloadLevel.Full;
            }
            if (reloadLevel === ts.ConfigFileProgramReloadLevel.Full) {
                state.configFileCache.delete(resolved);
                state.buildOrder = undefined;
            }
            state.needsSummary = true;
            clearProjectStatus(state, resolved);
            addProjToQueue(state, resolved, reloadLevel);
            enableCache(state);
        }
        function invalidateProjectAndScheduleBuilds(state, resolvedPath, reloadLevel) {
            state.reportFileChangeDetected = true;
            invalidateProject(state, resolvedPath, reloadLevel);
            scheduleBuildInvalidatedProject(state);
        }
        function scheduleBuildInvalidatedProject(state) {
            var hostWithWatch = state.hostWithWatch;
            if (!hostWithWatch.setTimeout || !hostWithWatch.clearTimeout) {
                return;
            }
            if (state.timerToBuildInvalidatedProject) {
                hostWithWatch.clearTimeout(state.timerToBuildInvalidatedProject);
            }
            state.timerToBuildInvalidatedProject = hostWithWatch.setTimeout(buildNextInvalidatedProject, 250, state);
        }
        function buildNextInvalidatedProject(state) {
            state.timerToBuildInvalidatedProject = undefined;
            if (state.reportFileChangeDetected) {
                state.reportFileChangeDetected = false;
                state.projectErrorsReported.clear();
                reportWatchStatus(state, ts.Diagnostics.File_change_detected_Starting_incremental_compilation);
            }
            var buildOrder = getBuildOrder(state);
            var invalidatedProject = getNextInvalidatedProject(state, buildOrder, /*reportQueue*/ false);
            if (invalidatedProject) {
                invalidatedProject.done();
                if (state.projectPendingBuild.size) {
                    // Schedule next project for build
                    if (state.watch && !state.timerToBuildInvalidatedProject) {
                        scheduleBuildInvalidatedProject(state);
                    }
                    return;
                }
            }
            disableCache(state);
            reportErrorSummary(state, buildOrder);
        }
        function watchConfigFile(state, resolved, resolvedPath, parsed) {
            if (!state.watch || state.allWatchedConfigFiles.has(resolvedPath))
                return;
            state.allWatchedConfigFiles.set(resolvedPath, state.watchFile(resolved, function () {
                invalidateProjectAndScheduleBuilds(state, resolvedPath, ts.ConfigFileProgramReloadLevel.Full);
            }, ts.PollingInterval.High, parsed === null || parsed === void 0 ? void 0 : parsed.watchOptions, ts.WatchType.ConfigFile, resolved));
        }
        function watchExtendedConfigFiles(state, resolvedPath, parsed) {
            ts.updateSharedExtendedConfigFileWatcher(resolvedPath, parsed === null || parsed === void 0 ? void 0 : parsed.options, state.allWatchedExtendedConfigFiles, function (extendedConfigFileName, extendedConfigFilePath) {
                return state.watchFile(extendedConfigFileName, function () {
                    var _a;
                    return (_a = state.allWatchedExtendedConfigFiles.get(extendedConfigFilePath)) === null || _a === void 0 ? void 0 : _a.projects.forEach(function (projectConfigFilePath) {
                        return invalidateProjectAndScheduleBuilds(state, projectConfigFilePath, ts.ConfigFileProgramReloadLevel.Full);
                    });
                }, ts.PollingInterval.High, parsed === null || parsed === void 0 ? void 0 : parsed.watchOptions, ts.WatchType.ExtendedConfigFile);
            }, function (fileName) { return toPath(state, fileName); });
        }
        function watchWildCardDirectories(state, resolved, resolvedPath, parsed) {
            if (!state.watch)
                return;
            ts.updateWatchingWildcardDirectories(getOrCreateValueMapFromConfigFileMap(state.allWatchedWildcardDirectories, resolvedPath), new ts.Map(ts.getEntries(parsed.wildcardDirectories)), function (dir, flags) {
                return state.watchDirectory(dir, function (fileOrDirectory) {
                    var _a;
                    if (ts.isIgnoredFileFromWildCardWatching({
                        watchedDirPath: toPath(state, dir),
                        fileOrDirectory: fileOrDirectory,
                        fileOrDirectoryPath: toPath(state, fileOrDirectory),
                        configFileName: resolved,
                        currentDirectory: state.currentDirectory,
                        options: parsed.options,
                        program: state.builderPrograms.get(resolvedPath) || ((_a = getCachedParsedConfigFile(state, resolvedPath)) === null || _a === void 0 ? void 0 : _a.fileNames),
                        useCaseSensitiveFileNames: state.parseConfigFileHost.useCaseSensitiveFileNames,
                        writeLog: function (s) { return state.writeLog(s); },
                        toPath: function (fileName) { return toPath(state, fileName); }
                    }))
                        return;
                    invalidateProjectAndScheduleBuilds(state, resolvedPath, ts.ConfigFileProgramReloadLevel.Partial);
                }, flags, parsed === null || parsed === void 0 ? void 0 : parsed.watchOptions, ts.WatchType.WildcardDirectory, resolved);
            });
        }
        function watchInputFiles(state, resolved, resolvedPath, parsed) {
            if (!state.watch)
                return;
            ts.mutateMap(getOrCreateValueMapFromConfigFileMap(state.allWatchedInputFiles, resolvedPath), ts.arrayToMap(parsed.fileNames, function (fileName) { return toPath(state, fileName); }), {
                createNewValue: function (_path, input) { return state.watchFile(input, function () { return invalidateProjectAndScheduleBuilds(state, resolvedPath, ts.ConfigFileProgramReloadLevel.None); }, ts.PollingInterval.Low, parsed === null || parsed === void 0 ? void 0 : parsed.watchOptions, ts.WatchType.SourceFile, resolved); },
                onDeleteValue: ts.closeFileWatcher,
            });
        }
        function watchPackageJsonFiles(state, resolved, resolvedPath, parsed) {
            if (!state.watch || !state.lastCachedPackageJsonLookups)
                return;
            ts.mutateMap(getOrCreateValueMapFromConfigFileMap(state.allWatchedPackageJsonFiles, resolvedPath), new ts.Map(state.lastCachedPackageJsonLookups.get(resolvedPath)), {
                createNewValue: function (path, _input) { return state.watchFile(path, function () { return invalidateProjectAndScheduleBuilds(state, resolvedPath, ts.ConfigFileProgramReloadLevel.Full); }, ts.PollingInterval.High, parsed === null || parsed === void 0 ? void 0 : parsed.watchOptions, ts.WatchType.PackageJson, resolved); },
                onDeleteValue: ts.closeFileWatcher,
            });
        }
        function startWatching(state, buildOrder) {
            if (!state.watchAllProjectsPending)
                return;
            state.watchAllProjectsPending = false;
            for (var _i = 0, _a = getBuildOrderFromAnyBuildOrder(buildOrder); _i < _a.length; _i++) {
                var resolved = _a[_i];
                var resolvedPath = toResolvedConfigFilePath(state, resolved);
                var cfg = parseConfigFile(state, resolved, resolvedPath);
                // Watch this file
                watchConfigFile(state, resolved, resolvedPath, cfg);
                watchExtendedConfigFiles(state, resolvedPath, cfg);
                if (cfg) {
                    // Update watchers for wildcard directories
                    watchWildCardDirectories(state, resolved, resolvedPath, cfg);
                    // Watch input files
                    watchInputFiles(state, resolved, resolvedPath, cfg);
                    // Watch package json files
                    watchPackageJsonFiles(state, resolved, resolvedPath, cfg);
                }
            }
        }
        function stopWatching(state) {
            ts.clearMap(state.allWatchedConfigFiles, ts.closeFileWatcher);
            ts.clearMap(state.allWatchedExtendedConfigFiles, ts.closeFileWatcherOf);
            ts.clearMap(state.allWatchedWildcardDirectories, function (watchedWildcardDirectories) { return ts.clearMap(watchedWildcardDirectories, ts.closeFileWatcherOf); });
            ts.clearMap(state.allWatchedInputFiles, function (watchedWildcardDirectories) { return ts.clearMap(watchedWildcardDirectories, ts.closeFileWatcher); });
            ts.clearMap(state.allWatchedPackageJsonFiles, function (watchedPacageJsonFiles) { return ts.clearMap(watchedPacageJsonFiles, ts.closeFileWatcher); });
        }
        function createSolutionBuilderWorker(watch, hostOrHostWithWatch, rootNames, options, baseWatchOptions) {
            var state = createSolutionBuilderState(watch, hostOrHostWithWatch, rootNames, options, baseWatchOptions);
            return {
                build: function (project, cancellationToken, writeFile, getCustomTransformers) { return build(state, project, cancellationToken, writeFile, getCustomTransformers); },
                clean: function (project) { return clean(state, project); },
                buildReferences: function (project, cancellationToken, writeFile, getCustomTransformers) { return build(state, project, cancellationToken, writeFile, getCustomTransformers, /*onlyReferences*/ true); },
                cleanReferences: function (project) { return clean(state, project, /*onlyReferences*/ true); },
                getNextInvalidatedProject: function (cancellationToken) {
                    setupInitialBuild(state, cancellationToken);
                    return getNextInvalidatedProject(state, getBuildOrder(state), /*reportQueue*/ false);
                },
                getBuildOrder: function () { return getBuildOrder(state); },
                getUpToDateStatusOfProject: function (project) {
                    var configFileName = resolveProjectName(state, project);
                    var configFilePath = toResolvedConfigFilePath(state, configFileName);
                    return getUpToDateStatus(state, parseConfigFile(state, configFileName, configFilePath), configFilePath);
                },
                invalidateProject: function (configFilePath, reloadLevel) { return invalidateProject(state, configFilePath, reloadLevel || ts.ConfigFileProgramReloadLevel.None); },
                buildNextInvalidatedProject: function () { return buildNextInvalidatedProject(state); },
                getAllParsedConfigs: function () { return ts.arrayFrom(ts.mapDefinedIterator(state.configFileCache.values(), function (config) { return isParsedCommandLine(config) ? config : undefined; })); },
                close: function () { return stopWatching(state); },
            };
        }
        function relName(state, path) {
            return ts.convertToRelativePath(path, state.currentDirectory, function (f) { return state.getCanonicalFileName(f); });
        }
        function reportStatus(state, message) {
            var args = [];
            for (var _i = 2; _i < arguments.length; _i++) {
                args[_i - 2] = arguments[_i];
            }
            state.host.reportSolutionBuilderStatus(ts.createCompilerDiagnostic.apply(void 0, __spreadArray([message], args, false)));
        }
        function reportWatchStatus(state, message) {
            var _a, _b;
            var args = [];
            for (var _i = 2; _i < arguments.length; _i++) {
                args[_i - 2] = arguments[_i];
            }
            (_b = (_a = state.hostWithWatch).onWatchStatusChange) === null || _b === void 0 ? void 0 : _b.call(_a, ts.createCompilerDiagnostic.apply(void 0, __spreadArray([message], args, false)), state.host.getNewLine(), state.baseCompilerOptions);
        }
        function reportErrors(_a, errors) {
            var host = _a.host;
            errors.forEach(function (err) { return host.reportDiagnostic(err); });
        }
        function reportAndStoreErrors(state, proj, errors) {
            reportErrors(state, errors);
            state.projectErrorsReported.set(proj, true);
            if (errors.length) {
                state.diagnostics.set(proj, errors);
            }
        }
        function reportParseConfigFileDiagnostic(state, proj) {
            reportAndStoreErrors(state, proj, [state.configFileCache.get(proj)]);
        }
        function reportErrorSummary(state, buildOrder) {
            if (!state.needsSummary)
                return;
            state.needsSummary = false;
            var canReportSummary = state.watch || !!state.host.reportErrorSummary;
            var diagnostics = state.diagnostics;
            var totalErrors = 0;
            if (isCircularBuildOrder(buildOrder)) {
                reportBuildQueue(state, buildOrder.buildOrder);
                reportErrors(state, buildOrder.circularDiagnostics);
                if (canReportSummary)
                    totalErrors += ts.getErrorCountForSummary(buildOrder.circularDiagnostics);
            }
            else {
                // Report errors from the other projects
                buildOrder.forEach(function (project) {
                    var projectPath = toResolvedConfigFilePath(state, project);
                    if (!state.projectErrorsReported.has(projectPath)) {
                        reportErrors(state, diagnostics.get(projectPath) || ts.emptyArray);
                    }
                });
                if (canReportSummary)
                    diagnostics.forEach(function (singleProjectErrors) { return totalErrors += ts.getErrorCountForSummary(singleProjectErrors); });
            }
            if (state.watch) {
                reportWatchStatus(state, ts.getWatchErrorSummaryDiagnosticMessage(totalErrors), totalErrors);
            }
            else if (state.host.reportErrorSummary) {
                state.host.reportErrorSummary(totalErrors);
            }
        }
        /**
         * Report the build ordering inferred from the current project graph if we're in verbose mode
         */
        function reportBuildQueue(state, buildQueue) {
            if (state.options.verbose) {
                reportStatus(state, ts.Diagnostics.Projects_in_this_build_Colon_0, buildQueue.map(function (s) { return "\r\n    * " + relName(state, s); }).join(""));
            }
        }
        function reportUpToDateStatus(state, configFileName, status) {
            if (state.options.force && (status.type === ts.UpToDateStatusType.UpToDate || status.type === ts.UpToDateStatusType.UpToDateWithUpstreamTypes)) {
                return reportStatus(state, ts.Diagnostics.Project_0_is_being_forcibly_rebuilt, relName(state, configFileName));
            }
            switch (status.type) {
                case ts.UpToDateStatusType.OutOfDateWithSelf:
                    return reportStatus(state, ts.Diagnostics.Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2, relName(state, configFileName), relName(state, status.outOfDateOutputFileName), relName(state, status.newerInputFileName));
                case ts.UpToDateStatusType.OutOfDateWithUpstream:
                    return reportStatus(state, ts.Diagnostics.Project_0_is_out_of_date_because_oldest_output_1_is_older_than_newest_input_2, relName(state, configFileName), relName(state, status.outOfDateOutputFileName), relName(state, status.newerProjectName));
                case ts.UpToDateStatusType.OutputMissing:
                    return reportStatus(state, ts.Diagnostics.Project_0_is_out_of_date_because_output_file_1_does_not_exist, relName(state, configFileName), relName(state, status.missingOutputFileName));
                case ts.UpToDateStatusType.UpToDate:
                    if (status.newestInputFileTime !== undefined) {
                        return reportStatus(state, ts.Diagnostics.Project_0_is_up_to_date_because_newest_input_1_is_older_than_oldest_output_2, relName(state, configFileName), relName(state, status.newestInputFileName || ""), relName(state, status.oldestOutputFileName || ""));
                    }
                    // Don't report anything for "up to date because it was already built" -- too verbose
                    break;
                case ts.UpToDateStatusType.OutOfDateWithPrepend:
                    return reportStatus(state, ts.Diagnostics.Project_0_is_out_of_date_because_output_of_its_dependency_1_has_changed, relName(state, configFileName), relName(state, status.newerProjectName));
                case ts.UpToDateStatusType.UpToDateWithUpstreamTypes:
                    return reportStatus(state, ts.Diagnostics.Project_0_is_up_to_date_with_d_ts_files_from_its_dependencies, relName(state, configFileName));
                case ts.UpToDateStatusType.UpstreamOutOfDate:
                    return reportStatus(state, ts.Diagnostics.Project_0_is_out_of_date_because_its_dependency_1_is_out_of_date, relName(state, configFileName), relName(state, status.upstreamProjectName));
                case ts.UpToDateStatusType.UpstreamBlocked:
                    return reportStatus(state, status.upstreamProjectBlocked ?
                        ts.Diagnostics.Project_0_can_t_be_built_because_its_dependency_1_was_not_built :
                        ts.Diagnostics.Project_0_can_t_be_built_because_its_dependency_1_has_errors, relName(state, configFileName), relName(state, status.upstreamProjectName));
                case ts.UpToDateStatusType.Unbuildable:
                    return reportStatus(state, ts.Diagnostics.Failed_to_parse_file_0_Colon_1, relName(state, configFileName), status.reason);
                case ts.UpToDateStatusType.TsVersionOutputOfDate:
                    return reportStatus(state, ts.Diagnostics.Project_0_is_out_of_date_because_output_for_it_was_generated_with_version_1_that_differs_with_current_version_2, relName(state, configFileName), status.version, ts.version);
                case ts.UpToDateStatusType.ContainerOnly:
                // Don't report status on "solution" projects
                // falls through
                case ts.UpToDateStatusType.ComputingUpstream:
                    // Should never leak from getUptoDateStatusWorker
                    break;
                default:
                    ts.assertType(status);
            }
        }
        /**
         * Report the up-to-date status of a project if we're in verbose mode
         */
        function verboseReportProjectStatus(state, configFileName, status) {
            if (state.options.verbose) {
                reportUpToDateStatus(state, configFileName, status);
            }
        }
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        var server;
        (function (server) {
            /* @internal */
            server.ActionSet = "action::set";
            /* @internal */
            server.ActionInvalidate = "action::invalidate";
            /* @internal */
            server.ActionPackageInstalled = "action::packageInstalled";
            /* @internal */
            server.EventTypesRegistry = "event::typesRegistry";
            /* @internal */
            server.EventBeginInstallTypes = "event::beginInstallTypes";
            /* @internal */
            server.EventEndInstallTypes = "event::endInstallTypes";
            /* @internal */
            server.EventInitializationFailed = "event::initializationFailed";
            /* @internal */
            var Arguments;
            (function (Arguments) {
                Arguments.GlobalCacheLocation = "--globalTypingsCacheLocation";
                Arguments.LogFile = "--logFile";
                Arguments.EnableTelemetry = "--enableTelemetry";
                Arguments.TypingSafeListLocation = "--typingSafeListLocation";
                Arguments.TypesMapLocation = "--typesMapLocation";
                /**
                 * This argument specifies the location of the NPM executable.
                 * typingsInstaller will run the command with `${npmLocation} install ...`.
                 */
                Arguments.NpmLocation = "--npmLocation";
                /**
                 * Flag indicating that the typings installer should try to validate the default npm location.
                 * If the default npm is not found when this flag is enabled, fallback to `npm install`
                 */
                Arguments.ValidateDefaultNpmLocation = "--validateDefaultNpmLocation";
            })(Arguments = server.Arguments || (server.Arguments = {}));
            /* @internal */
            function hasArgument(argumentName) {
                return ts.sys.args.indexOf(argumentName) >= 0;
            }
            server.hasArgument = hasArgument;
            /* @internal */
            function findArgument(argumentName) {
                var index = ts.sys.args.indexOf(argumentName);
                return index >= 0 && index < ts.sys.args.length - 1
                    ? ts.sys.args[index + 1]
                    : undefined;
            }
            server.findArgument = findArgument;
            /* @internal */
            function nowString() {
                // E.g. "12:34:56.789"
                var d = new Date();
                return ts.padLeft(d.getHours().toString(), 2, "0") + ":" + ts.padLeft(d.getMinutes().toString(), 2, "0") + ":" + ts.padLeft(d.getSeconds().toString(), 2, "0") + "." + ts.padLeft(d.getMilliseconds().toString(), 3, "0");
            }
            server.nowString = nowString;
        })(server = ts.server || (ts.server = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var JsTyping;
        (function (JsTyping) {
            function isTypingUpToDate(cachedTyping, availableTypingVersions) {
                var availableVersion = new ts.Version(ts.getProperty(availableTypingVersions, "ts" + ts.versionMajorMinor) || ts.getProperty(availableTypingVersions, "latest"));
                return availableVersion.compareTo(cachedTyping.version) <= 0;
            }
            JsTyping.isTypingUpToDate = isTypingUpToDate;
            var unprefixedNodeCoreModuleList = [
                "assert",
                "assert/strict",
                "async_hooks",
                "buffer",
                "child_process",
                "cluster",
                "console",
                "constants",
                "crypto",
                "dgram",
                "diagnostics_channel",
                "dns",
                "dns/promises",
                "domain",
                "events",
                "fs",
                "fs/promises",
                "http",
                "https",
                "http2",
                "inspector",
                "module",
                "net",
                "os",
                "path",
                "perf_hooks",
                "process",
                "punycode",
                "querystring",
                "readline",
                "repl",
                "stream",
                "stream/promises",
                "string_decoder",
                "timers",
                "timers/promises",
                "tls",
                "trace_events",
                "tty",
                "url",
                "util",
                "util/types",
                "v8",
                "vm",
                "wasi",
                "worker_threads",
                "zlib"
            ];
            JsTyping.prefixedNodeCoreModuleList = unprefixedNodeCoreModuleList.map(function (name) { return "node:" + name; });
            JsTyping.nodeCoreModuleList = __spreadArray(__spreadArray([], unprefixedNodeCoreModuleList, true), JsTyping.prefixedNodeCoreModuleList, true);
            JsTyping.nodeCoreModules = new ts.Set(JsTyping.nodeCoreModuleList);
            function nonRelativeModuleNameForTypingCache(moduleName) {
                return JsTyping.nodeCoreModules.has(moduleName) ? "node" : moduleName;
            }
            JsTyping.nonRelativeModuleNameForTypingCache = nonRelativeModuleNameForTypingCache;
            function loadSafeList(host, safeListPath) {
                var result = ts.readConfigFile(safeListPath, function (path) { return host.readFile(path); });
                return new ts.Map(ts.getEntries(result.config));
            }
            JsTyping.loadSafeList = loadSafeList;
            function loadTypesMap(host, typesMapPath) {
                var result = ts.readConfigFile(typesMapPath, function (path) { return host.readFile(path); });
                if (result.config) {
                    return new ts.Map(ts.getEntries(result.config.simpleMap));
                }
                return undefined;
            }
            JsTyping.loadTypesMap = loadTypesMap;
            /**
             * @param host is the object providing I/O related operations.
             * @param fileNames are the file names that belong to the same project
             * @param projectRootPath is the path to the project root directory
             * @param safeListPath is the path used to retrieve the safe list
             * @param packageNameToTypingLocation is the map of package names to their cached typing locations and installed versions
             * @param typeAcquisition is used to customize the typing acquisition process
             * @param compilerOptions are used as a source for typing inference
             */
            function discoverTypings(host, log, fileNames, projectRootPath, safeList, packageNameToTypingLocation, typeAcquisition, unresolvedImports, typesRegistry) {
                if (!typeAcquisition || !typeAcquisition.enable) {
                    return { cachedTypingPaths: [], newTypingNames: [], filesToWatch: [] };
                }
                // A typing name to typing file path mapping
                var inferredTypings = new ts.Map();
                // Only infer typings for .js and .jsx files
                fileNames = ts.mapDefined(fileNames, function (fileName) {
                    var path = ts.normalizePath(fileName);
                    if (ts.hasJSFileExtension(path)) {
                        return path;
                    }
                });
                var filesToWatch = [];
                if (typeAcquisition.include)
                    addInferredTypings(typeAcquisition.include, "Explicitly included types");
                var exclude = typeAcquisition.exclude || [];
                // Directories to search for package.json, bower.json and other typing information
                var possibleSearchDirs = new ts.Set(fileNames.map(ts.getDirectoryPath));
                possibleSearchDirs.add(projectRootPath);
                possibleSearchDirs.forEach(function (searchDir) {
                    var packageJsonPath = ts.combinePaths(searchDir, "package.json");
                    getTypingNamesFromJson(packageJsonPath, filesToWatch);
                    var bowerJsonPath = ts.combinePaths(searchDir, "bower.json");
                    getTypingNamesFromJson(bowerJsonPath, filesToWatch);
                    var bowerComponentsPath = ts.combinePaths(searchDir, "bower_components");
                    getTypingNamesFromPackagesFolder(bowerComponentsPath, filesToWatch);
                    var nodeModulesPath = ts.combinePaths(searchDir, "node_modules");
                    getTypingNamesFromPackagesFolder(nodeModulesPath, filesToWatch);
                });
                if (!typeAcquisition.disableFilenameBasedTypeAcquisition) {
                    getTypingNamesFromSourceFileNames(fileNames);
                }
                // add typings for unresolved imports
                if (unresolvedImports) {
                    var module_1 = ts.deduplicate(unresolvedImports.map(nonRelativeModuleNameForTypingCache), ts.equateStringsCaseSensitive, ts.compareStringsCaseSensitive);
                    addInferredTypings(module_1, "Inferred typings from unresolved imports");
                }
                // Add the cached typing locations for inferred typings that are already installed
                packageNameToTypingLocation.forEach(function (typing, name) {
                    var registryEntry = typesRegistry.get(name);
                    if (inferredTypings.has(name) && inferredTypings.get(name) === undefined && registryEntry !== undefined && isTypingUpToDate(typing, registryEntry)) {
                        inferredTypings.set(name, typing.typingLocation);
                    }
                });
                // Remove typings that the user has added to the exclude list
                for (var _i = 0, exclude_1 = exclude; _i < exclude_1.length; _i++) {
                    var excludeTypingName = exclude_1[_i];
                    var didDelete = inferredTypings.delete(excludeTypingName);
                    if (didDelete && log)
                        log("Typing for " + excludeTypingName + " is in exclude list, will be ignored.");
                }
                var newTypingNames = [];
                var cachedTypingPaths = [];
                inferredTypings.forEach(function (inferred, typing) {
                    if (inferred !== undefined) {
                        cachedTypingPaths.push(inferred);
                    }
                    else {
                        newTypingNames.push(typing);
                    }
                });
                var result = { cachedTypingPaths: cachedTypingPaths, newTypingNames: newTypingNames, filesToWatch: filesToWatch };
                if (log)
                    log("Result: " + JSON.stringify(result));
                return result;
                function addInferredTyping(typingName) {
                    if (!inferredTypings.has(typingName)) {
                        inferredTypings.set(typingName, undefined); // TODO: GH#18217
                    }
                }
                function addInferredTypings(typingNames, message) {
                    if (log)
                        log(message + ": " + JSON.stringify(typingNames));
                    ts.forEach(typingNames, addInferredTyping);
                }
                /**
                 * Get the typing info from common package manager json files like package.json or bower.json
                 */
                function getTypingNamesFromJson(jsonPath, filesToWatch) {
                    if (!host.fileExists(jsonPath)) {
                        return;
                    }
                    filesToWatch.push(jsonPath);
                    var jsonConfig = ts.readConfigFile(jsonPath, function (path) { return host.readFile(path); }).config;
                    var jsonTypingNames = ts.flatMap([jsonConfig.dependencies, jsonConfig.devDependencies, jsonConfig.optionalDependencies, jsonConfig.peerDependencies], ts.getOwnKeys);
                    addInferredTypings(jsonTypingNames, "Typing names in '" + jsonPath + "' dependencies");
                }
                /**
                 * Infer typing names from given file names. For example, the file name "jquery-min.2.3.4.js"
                 * should be inferred to the 'jquery' typing name; and "angular-route.1.2.3.js" should be inferred
                 * to the 'angular-route' typing name.
                 * @param fileNames are the names for source files in the project
                 */
                function getTypingNamesFromSourceFileNames(fileNames) {
                    var fromFileNames = ts.mapDefined(fileNames, function (j) {
                        if (!ts.hasJSFileExtension(j))
                            return undefined;
                        var inferredTypingName = ts.removeFileExtension(ts.getBaseFileName(j.toLowerCase()));
                        var cleanedTypingName = ts.removeMinAndVersionNumbers(inferredTypingName);
                        return safeList.get(cleanedTypingName);
                    });
                    if (fromFileNames.length) {
                        addInferredTypings(fromFileNames, "Inferred typings from file names");
                    }
                    var hasJsxFile = ts.some(fileNames, function (f) { return ts.fileExtensionIs(f, ".jsx" /* Jsx */); });
                    if (hasJsxFile) {
                        if (log)
                            log("Inferred 'react' typings due to presence of '.jsx' extension");
                        addInferredTyping("react");
                    }
                }
                /**
                 * Infer typing names from packages folder (ex: node_module, bower_components)
                 * @param packagesFolderPath is the path to the packages folder
                 */
                function getTypingNamesFromPackagesFolder(packagesFolderPath, filesToWatch) {
                    filesToWatch.push(packagesFolderPath);
                    // Todo: add support for ModuleResolutionHost too
                    if (!host.directoryExists(packagesFolderPath)) {
                        return;
                    }
                    // depth of 2, so we access `node_modules/foo` but not `node_modules/foo/bar`
                    var fileNames = host.readDirectory(packagesFolderPath, [".json" /* Json */], /*excludes*/ undefined, /*includes*/ undefined, /*depth*/ 2);
                    if (log)
                        log("Searching for typing names in " + packagesFolderPath + "; all files: " + JSON.stringify(fileNames));
                    var packageNames = [];
                    for (var _i = 0, fileNames_1 = fileNames; _i < fileNames_1.length; _i++) {
                        var fileName = fileNames_1[_i];
                        var normalizedFileName = ts.normalizePath(fileName);
                        var baseFileName = ts.getBaseFileName(normalizedFileName);
                        if (baseFileName !== "package.json" && baseFileName !== "bower.json") {
                            continue;
                        }
                        var result_1 = ts.readConfigFile(normalizedFileName, function (path) { return host.readFile(path); });
                        var packageJson = result_1.config;
                        // npm 3's package.json contains a "_requiredBy" field
                        // we should include all the top level module names for npm 2, and only module names whose
                        // "_requiredBy" field starts with "#" or equals "/" for npm 3.
                        if (baseFileName === "package.json" && packageJson._requiredBy &&
                            ts.filter(packageJson._requiredBy, function (r) { return r[0] === "#" || r === "/"; }).length === 0) {
                            continue;
                        }
                        // If the package has its own d.ts typings, those will take precedence. Otherwise the package name will be used
                        // to download d.ts files from DefinitelyTyped
                        if (!packageJson.name) {
                            continue;
                        }
                        var ownTypes = packageJson.types || packageJson.typings;
                        if (ownTypes) {
                            var absolutePath = ts.getNormalizedAbsolutePath(ownTypes, ts.getDirectoryPath(normalizedFileName));
                            if (log)
                                log("    Package '" + packageJson.name + "' provides its own types.");
                            inferredTypings.set(packageJson.name, absolutePath);
                        }
                        else {
                            packageNames.push(packageJson.name);
                        }
                    }
                    addInferredTypings(packageNames, "    Found package names");
                }
            }
            JsTyping.discoverTypings = discoverTypings;
            var NameValidationResult;
            (function (NameValidationResult) {
                NameValidationResult[NameValidationResult["Ok"] = 0] = "Ok";
                NameValidationResult[NameValidationResult["EmptyName"] = 1] = "EmptyName";
                NameValidationResult[NameValidationResult["NameTooLong"] = 2] = "NameTooLong";
                NameValidationResult[NameValidationResult["NameStartsWithDot"] = 3] = "NameStartsWithDot";
                NameValidationResult[NameValidationResult["NameStartsWithUnderscore"] = 4] = "NameStartsWithUnderscore";
                NameValidationResult[NameValidationResult["NameContainsNonURISafeCharacters"] = 5] = "NameContainsNonURISafeCharacters";
            })(NameValidationResult = JsTyping.NameValidationResult || (JsTyping.NameValidationResult = {}));
            var maxPackageNameLength = 214;
            /**
             * Validates package name using rules defined at https://docs.npmjs.com/files/package.json
             */
            function validatePackageName(packageName) {
                return validatePackageNameWorker(packageName, /*supportScopedPackage*/ true);
            }
            JsTyping.validatePackageName = validatePackageName;
            function validatePackageNameWorker(packageName, supportScopedPackage) {
                if (!packageName) {
                    return 1 /* EmptyName */;
                }
                if (packageName.length > maxPackageNameLength) {
                    return 2 /* NameTooLong */;
                }
                if (packageName.charCodeAt(0) === 46 /* dot */) {
                    return 3 /* NameStartsWithDot */;
                }
                if (packageName.charCodeAt(0) === 95 /* _ */) {
                    return 4 /* NameStartsWithUnderscore */;
                }
                // check if name is scope package like: starts with @ and has one '/' in the middle
                // scoped packages are not currently supported
                if (supportScopedPackage) {
                    var matches = /^@([^/]+)\/([^/]+)$/.exec(packageName);
                    if (matches) {
                        var scopeResult = validatePackageNameWorker(matches[1], /*supportScopedPackage*/ false);
                        if (scopeResult !== 0 /* Ok */) {
                            return { name: matches[1], isScopeName: true, result: scopeResult };
                        }
                        var packageResult = validatePackageNameWorker(matches[2], /*supportScopedPackage*/ false);
                        if (packageResult !== 0 /* Ok */) {
                            return { name: matches[2], isScopeName: false, result: packageResult };
                        }
                        return 0 /* Ok */;
                    }
                }
                if (encodeURIComponent(packageName) !== packageName) {
                    return 5 /* NameContainsNonURISafeCharacters */;
                }
                return 0 /* Ok */;
            }
            function renderPackageNameValidationFailure(result, typing) {
                return typeof result === "object" ?
                    renderPackageNameValidationFailureWorker(typing, result.result, result.name, result.isScopeName) :
                    renderPackageNameValidationFailureWorker(typing, result, typing, /*isScopeName*/ false);
            }
            JsTyping.renderPackageNameValidationFailure = renderPackageNameValidationFailure;
            function renderPackageNameValidationFailureWorker(typing, result, name, isScopeName) {
                var kind = isScopeName ? "Scope" : "Package";
                switch (result) {
                    case 1 /* EmptyName */:
                        return "'" + typing + "':: " + kind + " name '" + name + "' cannot be empty";
                    case 2 /* NameTooLong */:
                        return "'" + typing + "':: " + kind + " name '" + name + "' should be less than " + maxPackageNameLength + " characters";
                    case 3 /* NameStartsWithDot */:
                        return "'" + typing + "':: " + kind + " name '" + name + "' cannot start with '.'";
                    case 4 /* NameStartsWithUnderscore */:
                        return "'" + typing + "':: " + kind + " name '" + name + "' cannot start with '_'";
                    case 5 /* NameContainsNonURISafeCharacters */:
                        return "'" + typing + "':: " + kind + " name '" + name + "' contains non URI safe characters";
                    case 0 /* Ok */:
                        return ts.Debug.fail(); // Shouldn't have called this.
                    default:
                        throw ts.Debug.assertNever(result);
                }
            }
        })(JsTyping = ts.JsTyping || (ts.JsTyping = {}));
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        var ScriptSnapshot;
        (function (ScriptSnapshot) {
            var StringScriptSnapshot = /** @class */ (function () {
                function StringScriptSnapshot(text) {
                    this.text = text;
                }
                StringScriptSnapshot.prototype.getText = function (start, end) {
                    return start === 0 && end === this.text.length
                        ? this.text
                        : this.text.substring(start, end);
                };
                StringScriptSnapshot.prototype.getLength = function () {
                    return this.text.length;
                };
                StringScriptSnapshot.prototype.getChangeRange = function () {
                    // Text-based snapshots do not support incremental parsing. Return undefined
                    // to signal that to the caller.
                    return undefined;
                };
                return StringScriptSnapshot;
            }());
            function fromString(text) {
                return new StringScriptSnapshot(text);
            }
            ScriptSnapshot.fromString = fromString;
        })(ScriptSnapshot = ts.ScriptSnapshot || (ts.ScriptSnapshot = {}));
        /* @internal */
        var PackageJsonDependencyGroup;
        (function (PackageJsonDependencyGroup) {
            PackageJsonDependencyGroup[PackageJsonDependencyGroup["Dependencies"] = 1] = "Dependencies";
            PackageJsonDependencyGroup[PackageJsonDependencyGroup["DevDependencies"] = 2] = "DevDependencies";
            PackageJsonDependencyGroup[PackageJsonDependencyGroup["PeerDependencies"] = 4] = "PeerDependencies";
            PackageJsonDependencyGroup[PackageJsonDependencyGroup["OptionalDependencies"] = 8] = "OptionalDependencies";
            PackageJsonDependencyGroup[PackageJsonDependencyGroup["All"] = 15] = "All";
        })(PackageJsonDependencyGroup = ts.PackageJsonDependencyGroup || (ts.PackageJsonDependencyGroup = {}));
        /* @internal */
        var PackageJsonAutoImportPreference;
        (function (PackageJsonAutoImportPreference) {
            PackageJsonAutoImportPreference[PackageJsonAutoImportPreference["Off"] = 0] = "Off";
            PackageJsonAutoImportPreference[PackageJsonAutoImportPreference["On"] = 1] = "On";
            PackageJsonAutoImportPreference[PackageJsonAutoImportPreference["Auto"] = 2] = "Auto";
        })(PackageJsonAutoImportPreference = ts.PackageJsonAutoImportPreference || (ts.PackageJsonAutoImportPreference = {}));
        var LanguageServiceMode;
        (function (LanguageServiceMode) {
            LanguageServiceMode[LanguageServiceMode["Semantic"] = 0] = "Semantic";
            LanguageServiceMode[LanguageServiceMode["PartialSemantic"] = 1] = "PartialSemantic";
            LanguageServiceMode[LanguageServiceMode["Syntactic"] = 2] = "Syntactic";
        })(LanguageServiceMode = ts.LanguageServiceMode || (ts.LanguageServiceMode = {}));
        /* @internal */
        ts.emptyOptions = {};
        var SemanticClassificationFormat;
        (function (SemanticClassificationFormat) {
            SemanticClassificationFormat["Original"] = "original";
            SemanticClassificationFormat["TwentyTwenty"] = "2020";
        })(SemanticClassificationFormat = ts.SemanticClassificationFormat || (ts.SemanticClassificationFormat = {}));
        var CompletionTriggerKind;
        (function (CompletionTriggerKind) {
            /** Completion was triggered by typing an identifier, manual invocation (e.g Ctrl+Space) or via API. */
            CompletionTriggerKind[CompletionTriggerKind["Invoked"] = 1] = "Invoked";
            /** Completion was triggered by a trigger character. */
            CompletionTriggerKind[CompletionTriggerKind["TriggerCharacter"] = 2] = "TriggerCharacter";
            /** Completion was re-triggered as the current completion list is incomplete. */
            CompletionTriggerKind[CompletionTriggerKind["TriggerForIncompleteCompletions"] = 3] = "TriggerForIncompleteCompletions";
        })(CompletionTriggerKind = ts.CompletionTriggerKind || (ts.CompletionTriggerKind = {}));
        var InlayHintKind;
        (function (InlayHintKind) {
            InlayHintKind["Type"] = "Type";
            InlayHintKind["Parameter"] = "Parameter";
            InlayHintKind["Enum"] = "Enum";
        })(InlayHintKind = ts.InlayHintKind || (ts.InlayHintKind = {}));
        var HighlightSpanKind;
        (function (HighlightSpanKind) {
            HighlightSpanKind["none"] = "none";
            HighlightSpanKind["definition"] = "definition";
            HighlightSpanKind["reference"] = "reference";
            HighlightSpanKind["writtenReference"] = "writtenReference";
        })(HighlightSpanKind = ts.HighlightSpanKind || (ts.HighlightSpanKind = {}));
        var IndentStyle;
        (function (IndentStyle) {
            IndentStyle[IndentStyle["None"] = 0] = "None";
            IndentStyle[IndentStyle["Block"] = 1] = "Block";
            IndentStyle[IndentStyle["Smart"] = 2] = "Smart";
        })(IndentStyle = ts.IndentStyle || (ts.IndentStyle = {}));
        var SemicolonPreference;
        (function (SemicolonPreference) {
            SemicolonPreference["Ignore"] = "ignore";
            SemicolonPreference["Insert"] = "insert";
            SemicolonPreference["Remove"] = "remove";
        })(SemicolonPreference = ts.SemicolonPreference || (ts.SemicolonPreference = {}));
        function getDefaultFormatCodeSettings(newLineCharacter) {
            return {
                indentSize: 4,
                tabSize: 4,
                newLineCharacter: newLineCharacter || "\n",
                convertTabsToSpaces: true,
                indentStyle: IndentStyle.Smart,
                insertSpaceAfterConstructor: false,
                insertSpaceAfterCommaDelimiter: true,
                insertSpaceAfterSemicolonInForStatements: true,
                insertSpaceBeforeAndAfterBinaryOperators: true,
                insertSpaceAfterKeywordsInControlFlowStatements: true,
                insertSpaceAfterFunctionKeywordForAnonymousFunctions: false,
                insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis: false,
                insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets: false,
                insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces: true,
                insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces: false,
                insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces: false,
                insertSpaceBeforeFunctionParenthesis: false,
                placeOpenBraceOnNewLineForFunctions: false,
                placeOpenBraceOnNewLineForControlBlocks: false,
                semicolons: SemicolonPreference.Ignore,
                trimTrailingWhitespace: true
            };
        }
        ts.getDefaultFormatCodeSettings = getDefaultFormatCodeSettings;
        /* @internal */
        ts.testFormatSettings = getDefaultFormatCodeSettings("\n");
        var SymbolDisplayPartKind;
        (function (SymbolDisplayPartKind) {
            SymbolDisplayPartKind[SymbolDisplayPartKind["aliasName"] = 0] = "aliasName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["className"] = 1] = "className";
            SymbolDisplayPartKind[SymbolDisplayPartKind["enumName"] = 2] = "enumName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["fieldName"] = 3] = "fieldName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["interfaceName"] = 4] = "interfaceName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["keyword"] = 5] = "keyword";
            SymbolDisplayPartKind[SymbolDisplayPartKind["lineBreak"] = 6] = "lineBreak";
            SymbolDisplayPartKind[SymbolDisplayPartKind["numericLiteral"] = 7] = "numericLiteral";
            SymbolDisplayPartKind[SymbolDisplayPartKind["stringLiteral"] = 8] = "stringLiteral";
            SymbolDisplayPartKind[SymbolDisplayPartKind["localName"] = 9] = "localName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["methodName"] = 10] = "methodName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["moduleName"] = 11] = "moduleName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["operator"] = 12] = "operator";
            SymbolDisplayPartKind[SymbolDisplayPartKind["parameterName"] = 13] = "parameterName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["propertyName"] = 14] = "propertyName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["punctuation"] = 15] = "punctuation";
            SymbolDisplayPartKind[SymbolDisplayPartKind["space"] = 16] = "space";
            SymbolDisplayPartKind[SymbolDisplayPartKind["text"] = 17] = "text";
            SymbolDisplayPartKind[SymbolDisplayPartKind["typeParameterName"] = 18] = "typeParameterName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["enumMemberName"] = 19] = "enumMemberName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["functionName"] = 20] = "functionName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["regularExpressionLiteral"] = 21] = "regularExpressionLiteral";
            SymbolDisplayPartKind[SymbolDisplayPartKind["link"] = 22] = "link";
            SymbolDisplayPartKind[SymbolDisplayPartKind["linkName"] = 23] = "linkName";
            SymbolDisplayPartKind[SymbolDisplayPartKind["linkText"] = 24] = "linkText";
        })(SymbolDisplayPartKind = ts.SymbolDisplayPartKind || (ts.SymbolDisplayPartKind = {}));
        var OutliningSpanKind;
        (function (OutliningSpanKind) {
            /** Single or multi-line comments */
            OutliningSpanKind["Comment"] = "comment";
            /** Sections marked by '// #region' and '// #endregion' comments */
            OutliningSpanKind["Region"] = "region";
            /** Declarations and expressions */
            OutliningSpanKind["Code"] = "code";
            /** Contiguous blocks of import declarations */
            OutliningSpanKind["Imports"] = "imports";
        })(OutliningSpanKind = ts.OutliningSpanKind || (ts.OutliningSpanKind = {}));
        var OutputFileType;
        (function (OutputFileType) {
            OutputFileType[OutputFileType["JavaScript"] = 0] = "JavaScript";
            OutputFileType[OutputFileType["SourceMap"] = 1] = "SourceMap";
            OutputFileType[OutputFileType["Declaration"] = 2] = "Declaration";
        })(OutputFileType = ts.OutputFileType || (ts.OutputFileType = {}));
        var EndOfLineState;
        (function (EndOfLineState) {
            EndOfLineState[EndOfLineState["None"] = 0] = "None";
            EndOfLineState[EndOfLineState["InMultiLineCommentTrivia"] = 1] = "InMultiLineCommentTrivia";
            EndOfLineState[EndOfLineState["InSingleQuoteStringLiteral"] = 2] = "InSingleQuoteStringLiteral";
            EndOfLineState[EndOfLineState["InDoubleQuoteStringLiteral"] = 3] = "InDoubleQuoteStringLiteral";
            EndOfLineState[EndOfLineState["InTemplateHeadOrNoSubstitutionTemplate"] = 4] = "InTemplateHeadOrNoSubstitutionTemplate";
            EndOfLineState[EndOfLineState["InTemplateMiddleOrTail"] = 5] = "InTemplateMiddleOrTail";
            EndOfLineState[EndOfLineState["InTemplateSubstitutionPosition"] = 6] = "InTemplateSubstitutionPosition";
        })(EndOfLineState = ts.EndOfLineState || (ts.EndOfLineState = {}));
        var TokenClass;
        (function (TokenClass) {
            TokenClass[TokenClass["Punctuation"] = 0] = "Punctuation";
            TokenClass[TokenClass["Keyword"] = 1] = "Keyword";
            TokenClass[TokenClass["Operator"] = 2] = "Operator";
            TokenClass[TokenClass["Comment"] = 3] = "Comment";
            TokenClass[TokenClass["Whitespace"] = 4] = "Whitespace";
            TokenClass[TokenClass["Identifier"] = 5] = "Identifier";
            TokenClass[TokenClass["NumberLiteral"] = 6] = "NumberLiteral";
            TokenClass[TokenClass["BigIntLiteral"] = 7] = "BigIntLiteral";
            TokenClass[TokenClass["StringLiteral"] = 8] = "StringLiteral";
            TokenClass[TokenClass["RegExpLiteral"] = 9] = "RegExpLiteral";
        })(TokenClass = ts.TokenClass || (ts.TokenClass = {}));
        var ScriptElementKind;
        (function (ScriptElementKind) {
            ScriptElementKind["unknown"] = "";
            ScriptElementKind["warning"] = "warning";
            /** predefined type (void) or keyword (class) */
            ScriptElementKind["keyword"] = "keyword";
            /** top level script node */
            ScriptElementKind["scriptElement"] = "script";
            /** module foo {} */
            ScriptElementKind["moduleElement"] = "module";
            /** class X {} */
            ScriptElementKind["classElement"] = "class";
            /** var x = class X {} */
            ScriptElementKind["localClassElement"] = "local class";
            /** interface Y {} */
            ScriptElementKind["interfaceElement"] = "interface";
            /** type T = ... */
            ScriptElementKind["typeElement"] = "type";
            /** enum E */
            ScriptElementKind["enumElement"] = "enum";
            ScriptElementKind["enumMemberElement"] = "enum member";
            /**
             * Inside module and script only
             * const v = ..
             */
            ScriptElementKind["variableElement"] = "var";
            /** Inside function */
            ScriptElementKind["localVariableElement"] = "local var";
            /**
             * Inside module and script only
             * function f() { }
             */
            ScriptElementKind["functionElement"] = "function";
            /** Inside function */
            ScriptElementKind["localFunctionElement"] = "local function";
            /** class X { [public|private]* foo() {} } */
            ScriptElementKind["memberFunctionElement"] = "method";
            /** class X { [public|private]* [get|set] foo:number; } */
            ScriptElementKind["memberGetAccessorElement"] = "getter";
            ScriptElementKind["memberSetAccessorElement"] = "setter";
            /**
             * class X { [public|private]* foo:number; }
             * interface Y { foo:number; }
             */
            ScriptElementKind["memberVariableElement"] = "property";
            /**
             * class X { constructor() { } }
             * class X { static { } }
             */
            ScriptElementKind["constructorImplementationElement"] = "constructor";
            /** interface Y { ():number; } */
            ScriptElementKind["callSignatureElement"] = "call";
            /** interface Y { []:number; } */
            ScriptElementKind["indexSignatureElement"] = "index";
            /** interface Y { new():Y; } */
            ScriptElementKind["constructSignatureElement"] = "construct";
            /** function foo(*Y*: string) */
            ScriptElementKind["parameterElement"] = "parameter";
            ScriptElementKind["typeParameterElement"] = "type parameter";
            ScriptElementKind["primitiveType"] = "primitive type";
            ScriptElementKind["label"] = "label";
            ScriptElementKind["alias"] = "alias";
            ScriptElementKind["constElement"] = "const";
            ScriptElementKind["letElement"] = "let";
            ScriptElementKind["directory"] = "directory";
            ScriptElementKind["externalModuleName"] = "external module name";
            /**
             * <JsxTagName attribute1 attribute2={0} />
             */
            ScriptElementKind["jsxAttribute"] = "JSX attribute";
            /** String literal */
            ScriptElementKind["string"] = "string";
            /** Jsdoc @link: in `{@link C link text}`, the before and after text "{@link " and "}" */
            ScriptElementKind["link"] = "link";
            /** Jsdoc @link: in `{@link C link text}`, the entity name "C" */
            ScriptElementKind["linkName"] = "link name";
            /** Jsdoc @link: in `{@link C link text}`, the link text "link text" */
            ScriptElementKind["linkText"] = "link text";
        })(ScriptElementKind = ts.ScriptElementKind || (ts.ScriptElementKind = {}));
        var ScriptElementKindModifier;
        (function (ScriptElementKindModifier) {
            ScriptElementKindModifier["none"] = "";
            ScriptElementKindModifier["publicMemberModifier"] = "public";
            ScriptElementKindModifier["privateMemberModifier"] = "private";
            ScriptElementKindModifier["protectedMemberModifier"] = "protected";
            ScriptElementKindModifier["exportedModifier"] = "export";
            ScriptElementKindModifier["ambientModifier"] = "declare";
            ScriptElementKindModifier["staticModifier"] = "static";
            ScriptElementKindModifier["abstractModifier"] = "abstract";
            ScriptElementKindModifier["optionalModifier"] = "optional";
            ScriptElementKindModifier["deprecatedModifier"] = "deprecated";
            ScriptElementKindModifier["dtsModifier"] = ".d.ts";
            ScriptElementKindModifier["tsModifier"] = ".ts";
            ScriptElementKindModifier["tsxModifier"] = ".tsx";
            ScriptElementKindModifier["jsModifier"] = ".js";
            ScriptElementKindModifier["jsxModifier"] = ".jsx";
            ScriptElementKindModifier["jsonModifier"] = ".json";
        })(ScriptElementKindModifier = ts.ScriptElementKindModifier || (ts.ScriptElementKindModifier = {}));
        var ClassificationTypeNames;
        (function (ClassificationTypeNames) {
            ClassificationTypeNames["comment"] = "comment";
            ClassificationTypeNames["identifier"] = "identifier";
            ClassificationTypeNames["keyword"] = "keyword";
            ClassificationTypeNames["numericLiteral"] = "number";
            ClassificationTypeNames["bigintLiteral"] = "bigint";
            ClassificationTypeNames["operator"] = "operator";
            ClassificationTypeNames["stringLiteral"] = "string";
            ClassificationTypeNames["whiteSpace"] = "whitespace";
            ClassificationTypeNames["text"] = "text";
            ClassificationTypeNames["punctuation"] = "punctuation";
            ClassificationTypeNames["className"] = "class name";
            ClassificationTypeNames["enumName"] = "enum name";
            ClassificationTypeNames["interfaceName"] = "interface name";
            ClassificationTypeNames["moduleName"] = "module name";
            ClassificationTypeNames["typeParameterName"] = "type parameter name";
            ClassificationTypeNames["typeAliasName"] = "type alias name";
            ClassificationTypeNames["parameterName"] = "parameter name";
            ClassificationTypeNames["docCommentTagName"] = "doc comment tag name";
            ClassificationTypeNames["jsxOpenTagName"] = "jsx open tag name";
            ClassificationTypeNames["jsxCloseTagName"] = "jsx close tag name";
            ClassificationTypeNames["jsxSelfClosingTagName"] = "jsx self closing tag name";
            ClassificationTypeNames["jsxAttribute"] = "jsx attribute";
            ClassificationTypeNames["jsxText"] = "jsx text";
            ClassificationTypeNames["jsxAttributeStringLiteralValue"] = "jsx attribute string literal value";
        })(ClassificationTypeNames = ts.ClassificationTypeNames || (ts.ClassificationTypeNames = {}));
        var ClassificationType;
        (function (ClassificationType) {
            ClassificationType[ClassificationType["comment"] = 1] = "comment";
            ClassificationType[ClassificationType["identifier"] = 2] = "identifier";
            ClassificationType[ClassificationType["keyword"] = 3] = "keyword";
            ClassificationType[ClassificationType["numericLiteral"] = 4] = "numericLiteral";
            ClassificationType[ClassificationType["operator"] = 5] = "operator";
            ClassificationType[ClassificationType["stringLiteral"] = 6] = "stringLiteral";
            ClassificationType[ClassificationType["regularExpressionLiteral"] = 7] = "regularExpressionLiteral";
            ClassificationType[ClassificationType["whiteSpace"] = 8] = "whiteSpace";
            ClassificationType[ClassificationType["text"] = 9] = "text";
            ClassificationType[ClassificationType["punctuation"] = 10] = "punctuation";
            ClassificationType[ClassificationType["className"] = 11] = "className";
            ClassificationType[ClassificationType["enumName"] = 12] = "enumName";
            ClassificationType[ClassificationType["interfaceName"] = 13] = "interfaceName";
            ClassificationType[ClassificationType["moduleName"] = 14] = "moduleName";
            ClassificationType[ClassificationType["typeParameterName"] = 15] = "typeParameterName";
            ClassificationType[ClassificationType["typeAliasName"] = 16] = "typeAliasName";
            ClassificationType[ClassificationType["parameterName"] = 17] = "parameterName";
            ClassificationType[ClassificationType["docCommentTagName"] = 18] = "docCommentTagName";
            ClassificationType[ClassificationType["jsxOpenTagName"] = 19] = "jsxOpenTagName";
            ClassificationType[ClassificationType["jsxCloseTagName"] = 20] = "jsxCloseTagName";
            ClassificationType[ClassificationType["jsxSelfClosingTagName"] = 21] = "jsxSelfClosingTagName";
            ClassificationType[ClassificationType["jsxAttribute"] = 22] = "jsxAttribute";
            ClassificationType[ClassificationType["jsxText"] = 23] = "jsxText";
            ClassificationType[ClassificationType["jsxAttributeStringLiteralValue"] = 24] = "jsxAttributeStringLiteralValue";
            ClassificationType[ClassificationType["bigintLiteral"] = 25] = "bigintLiteral";
        })(ClassificationType = ts.ClassificationType || (ts.ClassificationType = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        // These utilities are common to multiple language service features.
        //#region
        ts.scanner = ts.createScanner(99 /* Latest */, /*skipTrivia*/ true);
        var SemanticMeaning;
        (function (SemanticMeaning) {
            SemanticMeaning[SemanticMeaning["None"] = 0] = "None";
            SemanticMeaning[SemanticMeaning["Value"] = 1] = "Value";
            SemanticMeaning[SemanticMeaning["Type"] = 2] = "Type";
            SemanticMeaning[SemanticMeaning["Namespace"] = 4] = "Namespace";
            SemanticMeaning[SemanticMeaning["All"] = 7] = "All";
        })(SemanticMeaning = ts.SemanticMeaning || (ts.SemanticMeaning = {}));
        function getMeaningFromDeclaration(node) {
            switch (node.kind) {
                case 252 /* VariableDeclaration */:
                    return ts.isInJSFile(node) && ts.getJSDocEnumTag(node) ? 7 /* All */ : 1 /* Value */;
                case 162 /* Parameter */:
                case 201 /* BindingElement */:
                case 165 /* PropertyDeclaration */:
                case 164 /* PropertySignature */:
                case 291 /* PropertyAssignment */:
                case 292 /* ShorthandPropertyAssignment */:
                case 167 /* MethodDeclaration */:
                case 166 /* MethodSignature */:
                case 169 /* Constructor */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 254 /* FunctionDeclaration */:
                case 211 /* FunctionExpression */:
                case 212 /* ArrowFunction */:
                case 290 /* CatchClause */:
                case 283 /* JsxAttribute */:
                    return 1 /* Value */;
                case 161 /* TypeParameter */:
                case 256 /* InterfaceDeclaration */:
                case 257 /* TypeAliasDeclaration */:
                case 180 /* TypeLiteral */:
                    return 2 /* Type */;
                case 340 /* JSDocTypedefTag */:
                    // If it has no name node, it shares the name with the value declaration below it.
                    return node.name === undefined ? 1 /* Value */ | 2 /* Type */ : 2 /* Type */;
                case 294 /* EnumMember */:
                case 255 /* ClassDeclaration */:
                    return 1 /* Value */ | 2 /* Type */;
                case 259 /* ModuleDeclaration */:
                    if (ts.isAmbientModule(node)) {
                        return 4 /* Namespace */ | 1 /* Value */;
                    }
                    else if (ts.getModuleInstanceState(node) === 1 /* Instantiated */) {
                        return 4 /* Namespace */ | 1 /* Value */;
                    }
                    else {
                        return 4 /* Namespace */;
                    }
                case 258 /* EnumDeclaration */:
                case 267 /* NamedImports */:
                case 268 /* ImportSpecifier */:
                case 263 /* ImportEqualsDeclaration */:
                case 264 /* ImportDeclaration */:
                case 269 /* ExportAssignment */:
                case 270 /* ExportDeclaration */:
                    return 7 /* All */;
                // An external module can be a Value
                case 300 /* SourceFile */:
                    return 4 /* Namespace */ | 1 /* Value */;
            }
            return 7 /* All */;
        }
        ts.getMeaningFromDeclaration = getMeaningFromDeclaration;
        function getMeaningFromLocation(node) {
            node = getAdjustedReferenceLocation(node);
            var parent = node.parent;
            if (node.kind === 300 /* SourceFile */) {
                return 1 /* Value */;
            }
            else if (ts.isExportAssignment(parent)
                || ts.isExportSpecifier(parent)
                || ts.isExternalModuleReference(parent)
                || ts.isImportSpecifier(parent)
                || ts.isImportClause(parent)
                || ts.isImportEqualsDeclaration(parent) && node === parent.name) {
                var decl = parent;
                while (decl) {
                    if (ts.isImportEqualsDeclaration(decl) || ts.isImportClause(decl) || ts.isExportDeclaration(decl)) {
                        return decl.isTypeOnly ? 2 /* Type */ : 7 /* All */;
                    }
                    decl = decl.parent;
                }
                return 7 /* All */;
            }
            else if (isInRightSideOfInternalImportEqualsDeclaration(node)) {
                return getMeaningFromRightHandSideOfImportEquals(node);
            }
            else if (ts.isDeclarationName(node)) {
                return getMeaningFromDeclaration(parent);
            }
            else if (ts.isEntityName(node) && ts.findAncestor(node, ts.or(ts.isJSDocNameReference, ts.isJSDocLinkLike, ts.isJSDocMemberName))) {
                return 7 /* All */;
            }
            else if (isTypeReference(node)) {
                return 2 /* Type */;
            }
            else if (isNamespaceReference(node)) {
                return 4 /* Namespace */;
            }
            else if (ts.isTypeParameterDeclaration(parent)) {
                ts.Debug.assert(ts.isJSDocTemplateTag(parent.parent)); // Else would be handled by isDeclarationName
                return 2 /* Type */;
            }
            else if (ts.isLiteralTypeNode(parent)) {
                // This might be T["name"], which is actually referencing a property and not a type. So allow both meanings.
                return 2 /* Type */ | 1 /* Value */;
            }
            else {
                return 1 /* Value */;
            }
        }
        ts.getMeaningFromLocation = getMeaningFromLocation;
        function getMeaningFromRightHandSideOfImportEquals(node) {
            //     import a = |b|; // Namespace
            //     import a = |b.c|; // Value, type, namespace
            //     import a = |b.c|.d; // Namespace
            var name = node.kind === 159 /* QualifiedName */ ? node : ts.isQualifiedName(node.parent) && node.parent.right === node ? node.parent : undefined;
            return name && name.parent.kind === 263 /* ImportEqualsDeclaration */ ? 7 /* All */ : 4 /* Namespace */;
        }
        function isInRightSideOfInternalImportEqualsDeclaration(node) {
            while (node.parent.kind === 159 /* QualifiedName */) {
                node = node.parent;
            }
            return ts.isInternalModuleImportEqualsDeclaration(node.parent) && node.parent.moduleReference === node;
        }
        ts.isInRightSideOfInternalImportEqualsDeclaration = isInRightSideOfInternalImportEqualsDeclaration;
        function isNamespaceReference(node) {
            return isQualifiedNameNamespaceReference(node) || isPropertyAccessNamespaceReference(node);
        }
        function isQualifiedNameNamespaceReference(node) {
            var root = node;
            var isLastClause = true;
            if (root.parent.kind === 159 /* QualifiedName */) {
                while (root.parent && root.parent.kind === 159 /* QualifiedName */) {
                    root = root.parent;
                }
                isLastClause = root.right === node;
            }
            return root.parent.kind === 176 /* TypeReference */ && !isLastClause;
        }
        function isPropertyAccessNamespaceReference(node) {
            var root = node;
            var isLastClause = true;
            if (root.parent.kind === 204 /* PropertyAccessExpression */) {
                while (root.parent && root.parent.kind === 204 /* PropertyAccessExpression */) {
                    root = root.parent;
                }
                isLastClause = root.name === node;
            }
            if (!isLastClause && root.parent.kind === 226 /* ExpressionWithTypeArguments */ && root.parent.parent.kind === 289 /* HeritageClause */) {
                var decl = root.parent.parent.parent;
                return (decl.kind === 255 /* ClassDeclaration */ && root.parent.parent.token === 117 /* ImplementsKeyword */) ||
                    (decl.kind === 256 /* InterfaceDeclaration */ && root.parent.parent.token === 94 /* ExtendsKeyword */);
            }
            return false;
        }
        function isTypeReference(node) {
            if (ts.isRightSideOfQualifiedNameOrPropertyAccess(node)) {
                node = node.parent;
            }
            switch (node.kind) {
                case 108 /* ThisKeyword */:
                    return !ts.isExpressionNode(node);
                case 190 /* ThisType */:
                    return true;
            }
            switch (node.parent.kind) {
                case 176 /* TypeReference */:
                    return true;
                case 198 /* ImportType */:
                    return !node.parent.isTypeOf;
                case 226 /* ExpressionWithTypeArguments */:
                    return !ts.isExpressionWithTypeArgumentsInClassExtendsClause(node.parent);
            }
            return false;
        }
        function isCallExpressionTarget(node, includeElementAccess, skipPastOuterExpressions) {
            if (includeElementAccess === void 0) {
                includeElementAccess = false;
            }
            if (skipPastOuterExpressions === void 0) {
                skipPastOuterExpressions = false;
            }
            return isCalleeWorker(node, ts.isCallExpression, selectExpressionOfCallOrNewExpressionOrDecorator, includeElementAccess, skipPastOuterExpressions);
        }
        ts.isCallExpressionTarget = isCallExpressionTarget;
        function isNewExpressionTarget(node, includeElementAccess, skipPastOuterExpressions) {
            if (includeElementAccess === void 0) {
                includeElementAccess = false;
            }
            if (skipPastOuterExpressions === void 0) {
                skipPastOuterExpressions = false;
            }
            return isCalleeWorker(node, ts.isNewExpression, selectExpressionOfCallOrNewExpressionOrDecorator, includeElementAccess, skipPastOuterExpressions);
        }
        ts.isNewExpressionTarget = isNewExpressionTarget;
        function isCallOrNewExpressionTarget(node, includeElementAccess, skipPastOuterExpressions) {
            if (includeElementAccess === void 0) {
                includeElementAccess = false;
            }
            if (skipPastOuterExpressions === void 0) {
                skipPastOuterExpressions = false;
            }
            return isCalleeWorker(node, ts.isCallOrNewExpression, selectExpressionOfCallOrNewExpressionOrDecorator, includeElementAccess, skipPastOuterExpressions);
        }
        ts.isCallOrNewExpressionTarget = isCallOrNewExpressionTarget;
        function isTaggedTemplateTag(node, includeElementAccess, skipPastOuterExpressions) {
            if (includeElementAccess === void 0) {
                includeElementAccess = false;
            }
            if (skipPastOuterExpressions === void 0) {
                skipPastOuterExpressions = false;
            }
            return isCalleeWorker(node, ts.isTaggedTemplateExpression, selectTagOfTaggedTemplateExpression, includeElementAccess, skipPastOuterExpressions);
        }
        ts.isTaggedTemplateTag = isTaggedTemplateTag;
        function isDecoratorTarget(node, includeElementAccess, skipPastOuterExpressions) {
            if (includeElementAccess === void 0) {
                includeElementAccess = false;
            }
            if (skipPastOuterExpressions === void 0) {
                skipPastOuterExpressions = false;
            }
            return isCalleeWorker(node, ts.isDecorator, selectExpressionOfCallOrNewExpressionOrDecorator, includeElementAccess, skipPastOuterExpressions);
        }
        ts.isDecoratorTarget = isDecoratorTarget;
        function isJsxOpeningLikeElementTagName(node, includeElementAccess, skipPastOuterExpressions) {
            if (includeElementAccess === void 0) {
                includeElementAccess = false;
            }
            if (skipPastOuterExpressions === void 0) {
                skipPastOuterExpressions = false;
            }
            return isCalleeWorker(node, ts.isJsxOpeningLikeElement, selectTagNameOfJsxOpeningLikeElement, includeElementAccess, skipPastOuterExpressions);
        }
        ts.isJsxOpeningLikeElementTagName = isJsxOpeningLikeElementTagName;
        function selectExpressionOfCallOrNewExpressionOrDecorator(node) {
            return node.expression;
        }
        function selectTagOfTaggedTemplateExpression(node) {
            return node.tag;
        }
        function selectTagNameOfJsxOpeningLikeElement(node) {
            return node.tagName;
        }
        function isCalleeWorker(node, pred, calleeSelector, includeElementAccess, skipPastOuterExpressions) {
            var target = includeElementAccess ? climbPastPropertyOrElementAccess(node) : climbPastPropertyAccess(node);
            if (skipPastOuterExpressions) {
                target = ts.skipOuterExpressions(target);
            }
            return !!target && !!target.parent && pred(target.parent) && calleeSelector(target.parent) === target;
        }
        function climbPastPropertyAccess(node) {
            return isRightSideOfPropertyAccess(node) ? node.parent : node;
        }
        ts.climbPastPropertyAccess = climbPastPropertyAccess;
        function climbPastPropertyOrElementAccess(node) {
            return isRightSideOfPropertyAccess(node) || isArgumentExpressionOfElementAccess(node) ? node.parent : node;
        }
        ts.climbPastPropertyOrElementAccess = climbPastPropertyOrElementAccess;
        function getTargetLabel(referenceNode, labelName) {
            while (referenceNode) {
                if (referenceNode.kind === 248 /* LabeledStatement */ && referenceNode.label.escapedText === labelName) {
                    return referenceNode.label;
                }
                referenceNode = referenceNode.parent;
            }
            return undefined;
        }
        ts.getTargetLabel = getTargetLabel;
        function hasPropertyAccessExpressionWithName(node, funcName) {
            if (!ts.isPropertyAccessExpression(node.expression)) {
                return false;
            }
            return node.expression.name.text === funcName;
        }
        ts.hasPropertyAccessExpressionWithName = hasPropertyAccessExpressionWithName;
        function isJumpStatementTarget(node) {
            var _a;
            return ts.isIdentifier(node) && ((_a = ts.tryCast(node.parent, ts.isBreakOrContinueStatement)) === null || _a === void 0 ? void 0 : _a.label) === node;
        }
        ts.isJumpStatementTarget = isJumpStatementTarget;
        function isLabelOfLabeledStatement(node) {
            var _a;
            return ts.isIdentifier(node) && ((_a = ts.tryCast(node.parent, ts.isLabeledStatement)) === null || _a === void 0 ? void 0 : _a.label) === node;
        }
        ts.isLabelOfLabeledStatement = isLabelOfLabeledStatement;
        function isLabelName(node) {
            return isLabelOfLabeledStatement(node) || isJumpStatementTarget(node);
        }
        ts.isLabelName = isLabelName;
        function isTagName(node) {
            var _a;
            return ((_a = ts.tryCast(node.parent, ts.isJSDocTag)) === null || _a === void 0 ? void 0 : _a.tagName) === node;
        }
        ts.isTagName = isTagName;
        function isRightSideOfQualifiedName(node) {
            var _a;
            return ((_a = ts.tryCast(node.parent, ts.isQualifiedName)) === null || _a === void 0 ? void 0 : _a.right) === node;
        }
        ts.isRightSideOfQualifiedName = isRightSideOfQualifiedName;
        function isRightSideOfPropertyAccess(node) {
            var _a;
            return ((_a = ts.tryCast(node.parent, ts.isPropertyAccessExpression)) === null || _a === void 0 ? void 0 : _a.name) === node;
        }
        ts.isRightSideOfPropertyAccess = isRightSideOfPropertyAccess;
        function isArgumentExpressionOfElementAccess(node) {
            var _a;
            return ((_a = ts.tryCast(node.parent, ts.isElementAccessExpression)) === null || _a === void 0 ? void 0 : _a.argumentExpression) === node;
        }
        ts.isArgumentExpressionOfElementAccess = isArgumentExpressionOfElementAccess;
        function isNameOfModuleDeclaration(node) {
            var _a;
            return ((_a = ts.tryCast(node.parent, ts.isModuleDeclaration)) === null || _a === void 0 ? void 0 : _a.name) === node;
        }
        ts.isNameOfModuleDeclaration = isNameOfModuleDeclaration;
        function isNameOfFunctionDeclaration(node) {
            var _a;
            return ts.isIdentifier(node) && ((_a = ts.tryCast(node.parent, ts.isFunctionLike)) === null || _a === void 0 ? void 0 : _a.name) === node;
        }
        ts.isNameOfFunctionDeclaration = isNameOfFunctionDeclaration;
        function isLiteralNameOfPropertyDeclarationOrIndexAccess(node) {
            switch (node.parent.kind) {
                case 165 /* PropertyDeclaration */:
                case 164 /* PropertySignature */:
                case 291 /* PropertyAssignment */:
                case 294 /* EnumMember */:
                case 167 /* MethodDeclaration */:
                case 166 /* MethodSignature */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 259 /* ModuleDeclaration */:
                    return ts.getNameOfDeclaration(node.parent) === node;
                case 205 /* ElementAccessExpression */:
                    return node.parent.argumentExpression === node;
                case 160 /* ComputedPropertyName */:
                    return true;
                case 194 /* LiteralType */:
                    return node.parent.parent.kind === 192 /* IndexedAccessType */;
                default:
                    return false;
            }
        }
        ts.isLiteralNameOfPropertyDeclarationOrIndexAccess = isLiteralNameOfPropertyDeclarationOrIndexAccess;
        function isExpressionOfExternalModuleImportEqualsDeclaration(node) {
            return ts.isExternalModuleImportEqualsDeclaration(node.parent.parent) &&
                ts.getExternalModuleImportEqualsDeclarationExpression(node.parent.parent) === node;
        }
        ts.isExpressionOfExternalModuleImportEqualsDeclaration = isExpressionOfExternalModuleImportEqualsDeclaration;
        function getContainerNode(node) {
            if (ts.isJSDocTypeAlias(node)) {
                // This doesn't just apply to the node immediately under the comment, but to everything in its parent's scope.
                // node.parent = the JSDoc comment, node.parent.parent = the node having the comment.
                // Then we get parent again in the loop.
                node = node.parent.parent;
            }
            while (true) {
                node = node.parent;
                if (!node) {
                    return undefined;
                }
                switch (node.kind) {
                    case 300 /* SourceFile */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 255 /* ClassDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 259 /* ModuleDeclaration */:
                        return node;
                }
            }
        }
        ts.getContainerNode = getContainerNode;
        function getNodeKind(node) {
            switch (node.kind) {
                case 300 /* SourceFile */:
                    return ts.isExternalModule(node) ? "module" /* moduleElement */ : "script" /* scriptElement */;
                case 259 /* ModuleDeclaration */:
                    return "module" /* moduleElement */;
                case 255 /* ClassDeclaration */:
                case 224 /* ClassExpression */:
                    return "class" /* classElement */;
                case 256 /* InterfaceDeclaration */: return "interface" /* interfaceElement */;
                case 257 /* TypeAliasDeclaration */:
                case 333 /* JSDocCallbackTag */:
                case 340 /* JSDocTypedefTag */:
                    return "type" /* typeElement */;
                case 258 /* EnumDeclaration */: return "enum" /* enumElement */;
                case 252 /* VariableDeclaration */:
                    return getKindOfVariableDeclaration(node);
                case 201 /* BindingElement */:
                    return getKindOfVariableDeclaration(ts.getRootDeclaration(node));
                case 212 /* ArrowFunction */:
                case 254 /* FunctionDeclaration */:
                case 211 /* FunctionExpression */:
                    return "function" /* functionElement */;
                case 170 /* GetAccessor */: return "getter" /* memberGetAccessorElement */;
                case 171 /* SetAccessor */: return "setter" /* memberSetAccessorElement */;
                case 167 /* MethodDeclaration */:
                case 166 /* MethodSignature */:
                    return "method" /* memberFunctionElement */;
                case 291 /* PropertyAssignment */:
                    var initializer = node.initializer;
                    return ts.isFunctionLike(initializer) ? "method" /* memberFunctionElement */ : "property" /* memberVariableElement */;
                case 165 /* PropertyDeclaration */:
                case 164 /* PropertySignature */:
                case 292 /* ShorthandPropertyAssignment */:
                case 293 /* SpreadAssignment */:
                    return "property" /* memberVariableElement */;
                case 174 /* IndexSignature */: return "index" /* indexSignatureElement */;
                case 173 /* ConstructSignature */: return "construct" /* constructSignatureElement */;
                case 172 /* CallSignature */: return "call" /* callSignatureElement */;
                case 169 /* Constructor */:
                case 168 /* ClassStaticBlockDeclaration */:
                    return "constructor" /* constructorImplementationElement */;
                case 161 /* TypeParameter */: return "type parameter" /* typeParameterElement */;
                case 294 /* EnumMember */: return "enum member" /* enumMemberElement */;
                case 162 /* Parameter */: return ts.hasSyntacticModifier(node, 16476 /* ParameterPropertyModifier */) ? "property" /* memberVariableElement */ : "parameter" /* parameterElement */;
                case 263 /* ImportEqualsDeclaration */:
                case 268 /* ImportSpecifier */:
                case 273 /* ExportSpecifier */:
                case 266 /* NamespaceImport */:
                case 272 /* NamespaceExport */:
                    return "alias" /* alias */;
                case 219 /* BinaryExpression */:
                    var kind = ts.getAssignmentDeclarationKind(node);
                    var right = node.right;
                    switch (kind) {
                        case 7 /* ObjectDefinePropertyValue */:
                        case 8 /* ObjectDefinePropertyExports */:
                        case 9 /* ObjectDefinePrototypeProperty */:
                        case 0 /* None */:
                            return "" /* unknown */;
                        case 1 /* ExportsProperty */:
                        case 2 /* ModuleExports */:
                            var rightKind = getNodeKind(right);
                            return rightKind === "" /* unknown */ ? "const" /* constElement */ : rightKind;
                        case 3 /* PrototypeProperty */:
                            return ts.isFunctionExpression(right) ? "method" /* memberFunctionElement */ : "property" /* memberVariableElement */;
                        case 4 /* ThisProperty */:
                            return "property" /* memberVariableElement */; // property
                        case 5 /* Property */:
                            // static method / property
                            return ts.isFunctionExpression(right) ? "method" /* memberFunctionElement */ : "property" /* memberVariableElement */;
                        case 6 /* Prototype */:
                            return "local class" /* localClassElement */;
                        default: {
                            ts.assertType(kind);
                            return "" /* unknown */;
                        }
                    }
                case 79 /* Identifier */:
                    return ts.isImportClause(node.parent) ? "alias" /* alias */ : "" /* unknown */;
                case 269 /* ExportAssignment */:
                    var scriptKind = getNodeKind(node.expression);
                    // If the expression didn't come back with something (like it does for an identifiers)
                    return scriptKind === "" /* unknown */ ? "const" /* constElement */ : scriptKind;
                default:
                    return "" /* unknown */;
            }
            function getKindOfVariableDeclaration(v) {
                return ts.isVarConst(v)
                    ? "const" /* constElement */
                    : ts.isLet(v)
                        ? "let" /* letElement */
                        : "var" /* variableElement */;
            }
        }
        ts.getNodeKind = getNodeKind;
        function isThis(node) {
            switch (node.kind) {
                case 108 /* ThisKeyword */:
                    // case SyntaxKind.ThisType: TODO: GH#9267
                    return true;
                case 79 /* Identifier */:
                    // 'this' as a parameter
                    return ts.identifierIsThisKeyword(node) && node.parent.kind === 162 /* Parameter */;
                default:
                    return false;
            }
        }
        ts.isThis = isThis;
        // Matches the beginning of a triple slash directive
        var tripleSlashDirectivePrefixRegex = /^\/\/\/\s*</;
        function getLineStartPositionForPosition(position, sourceFile) {
            var lineStarts = ts.getLineStarts(sourceFile);
            var line = sourceFile.getLineAndCharacterOfPosition(position).line;
            return lineStarts[line];
        }
        ts.getLineStartPositionForPosition = getLineStartPositionForPosition;
        function rangeContainsRange(r1, r2) {
            return startEndContainsRange(r1.pos, r1.end, r2);
        }
        ts.rangeContainsRange = rangeContainsRange;
        function rangeContainsRangeExclusive(r1, r2) {
            return rangeContainsPositionExclusive(r1, r2.pos) && rangeContainsPositionExclusive(r1, r2.end);
        }
        ts.rangeContainsRangeExclusive = rangeContainsRangeExclusive;
        function rangeContainsPosition(r, pos) {
            return r.pos <= pos && pos <= r.end;
        }
        ts.rangeContainsPosition = rangeContainsPosition;
        function rangeContainsPositionExclusive(r, pos) {
            return r.pos < pos && pos < r.end;
        }
        ts.rangeContainsPositionExclusive = rangeContainsPositionExclusive;
        function startEndContainsRange(start, end, range) {
            return start <= range.pos && end >= range.end;
        }
        ts.startEndContainsRange = startEndContainsRange;
        function rangeContainsStartEnd(range, start, end) {
            return range.pos <= start && range.end >= end;
        }
        ts.rangeContainsStartEnd = rangeContainsStartEnd;
        function rangeOverlapsWithStartEnd(r1, start, end) {
            return startEndOverlapsWithStartEnd(r1.pos, r1.end, start, end);
        }
        ts.rangeOverlapsWithStartEnd = rangeOverlapsWithStartEnd;
        function nodeOverlapsWithStartEnd(node, sourceFile, start, end) {
            return startEndOverlapsWithStartEnd(node.getStart(sourceFile), node.end, start, end);
        }
        ts.nodeOverlapsWithStartEnd = nodeOverlapsWithStartEnd;
        function startEndOverlapsWithStartEnd(start1, end1, start2, end2) {
            var start = Math.max(start1, start2);
            var end = Math.min(end1, end2);
            return start < end;
        }
        ts.startEndOverlapsWithStartEnd = startEndOverlapsWithStartEnd;
        /**
         * Assumes `candidate.start <= position` holds.
         */
        function positionBelongsToNode(candidate, position, sourceFile) {
            ts.Debug.assert(candidate.pos <= position);
            return position < candidate.end || !isCompletedNode(candidate, sourceFile);
        }
        ts.positionBelongsToNode = positionBelongsToNode;
        function isCompletedNode(n, sourceFile) {
            if (n === undefined || ts.nodeIsMissing(n)) {
                return false;
            }
            switch (n.kind) {
                case 255 /* ClassDeclaration */:
                case 256 /* InterfaceDeclaration */:
                case 258 /* EnumDeclaration */:
                case 203 /* ObjectLiteralExpression */:
                case 199 /* ObjectBindingPattern */:
                case 180 /* TypeLiteral */:
                case 233 /* Block */:
                case 260 /* ModuleBlock */:
                case 261 /* CaseBlock */:
                case 267 /* NamedImports */:
                case 271 /* NamedExports */:
                    return nodeEndsWith(n, 19 /* CloseBraceToken */, sourceFile);
                case 290 /* CatchClause */:
                    return isCompletedNode(n.block, sourceFile);
                case 207 /* NewExpression */:
                    if (!n.arguments) {
                        return true;
                    }
                // falls through
                case 206 /* CallExpression */:
                case 210 /* ParenthesizedExpression */:
                case 189 /* ParenthesizedType */:
                    return nodeEndsWith(n, 21 /* CloseParenToken */, sourceFile);
                case 177 /* FunctionType */:
                case 178 /* ConstructorType */:
                    return isCompletedNode(n.type, sourceFile);
                case 169 /* Constructor */:
                case 170 /* GetAccessor */:
                case 171 /* SetAccessor */:
                case 254 /* FunctionDeclaration */:
                case 211 /* FunctionExpression */:
                case 167 /* MethodDeclaration */:
                case 166 /* MethodSignature */:
                case 173 /* ConstructSignature */:
                case 172 /* CallSignature */:
                case 212 /* ArrowFunction */:
                    if (n.body) {
                        return isCompletedNode(n.body, sourceFile);
                    }
                    if (n.type) {
                        return isCompletedNode(n.type, sourceFile);
                    }
                    // Even though type parameters can be unclosed, we can get away with
                    // having at least a closing paren.
                    return hasChildOfKind(n, 21 /* CloseParenToken */, sourceFile);
                case 259 /* ModuleDeclaration */:
                    return !!n.body && isCompletedNode(n.body, sourceFile);
                case 237 /* IfStatement */:
                    if (n.elseStatement) {
                        return isCompletedNode(n.elseStatement, sourceFile);
                    }
                    return isCompletedNode(n.thenStatement, sourceFile);
                case 236 /* ExpressionStatement */:
                    return isCompletedNode(n.expression, sourceFile) ||
                        hasChildOfKind(n, 26 /* SemicolonToken */, sourceFile);
                case 202 /* ArrayLiteralExpression */:
                case 200 /* ArrayBindingPattern */:
                case 205 /* ElementAccessExpression */:
                case 160 /* ComputedPropertyName */:
                case 182 /* TupleType */:
                    return nodeEndsWith(n, 23 /* CloseBracketToken */, sourceFile);
                case 174 /* IndexSignature */:
                    if (n.type) {
                        return isCompletedNode(n.type, sourceFile);
                    }
                    return hasChildOfKind(n, 23 /* CloseBracketToken */, sourceFile);
                case 287 /* CaseClause */:
                case 288 /* DefaultClause */:
                    // there is no such thing as terminator token for CaseClause/DefaultClause so for simplicity always consider them non-completed
                    return false;
                case 240 /* ForStatement */:
                case 241 /* ForInStatement */:
                case 242 /* ForOfStatement */:
                case 239 /* WhileStatement */:
                    return isCompletedNode(n.statement, sourceFile);
                case 238 /* DoStatement */:
                    // rough approximation: if DoStatement has While keyword - then if node is completed is checking the presence of ')';
                    return hasChildOfKind(n, 115 /* WhileKeyword */, sourceFile)
                        ? nodeEndsWith(n, 21 /* CloseParenToken */, sourceFile)
                        : isCompletedNode(n.statement, sourceFile);
                case 179 /* TypeQuery */:
                    return isCompletedNode(n.exprName, sourceFile);
                case 214 /* TypeOfExpression */:
                case 213 /* DeleteExpression */:
                case 215 /* VoidExpression */:
                case 222 /* YieldExpression */:
                case 223 /* SpreadElement */:
                    var unaryWordExpression = n;
                    return isCompletedNode(unaryWordExpression.expression, sourceFile);
                case 208 /* TaggedTemplateExpression */:
                    return isCompletedNode(n.template, sourceFile);
                case 221 /* TemplateExpression */:
                    var lastSpan = ts.lastOrUndefined(n.templateSpans);
                    return isCompletedNode(lastSpan, sourceFile);
                case 231 /* TemplateSpan */:
                    return ts.nodeIsPresent(n.literal);
                case 270 /* ExportDeclaration */:
                case 264 /* ImportDeclaration */:
                    return ts.nodeIsPresent(n.moduleSpecifier);
                case 217 /* PrefixUnaryExpression */:
                    return isCompletedNode(n.operand, sourceFile);
                case 219 /* BinaryExpression */:
                    return isCompletedNode(n.right, sourceFile);
                case 220 /* ConditionalExpression */:
                    return isCompletedNode(n.whenFalse, sourceFile);
                default:
                    return true;
            }
        }
        /*
         * Checks if node ends with 'expectedLastToken'.
         * If child at position 'length - 1' is 'SemicolonToken' it is skipped and 'expectedLastToken' is compared with child at position 'length - 2'.
         */
        function nodeEndsWith(n, expectedLastToken, sourceFile) {
            var children = n.getChildren(sourceFile);
            if (children.length) {
                var lastChild = ts.last(children);
                if (lastChild.kind === expectedLastToken) {
                    return true;
                }
                else if (lastChild.kind === 26 /* SemicolonToken */ && children.length !== 1) {
                    return children[children.length - 2].kind === expectedLastToken;
                }
            }
            return false;
        }
        function findListItemInfo(node) {
            var list = findContainingList(node);
            // It is possible at this point for syntaxList to be undefined, either if
            // node.parent had no list child, or if none of its list children contained
            // the span of node. If this happens, return undefined. The caller should
            // handle this case.
            if (!list) {
                return undefined;
            }
            var children = list.getChildren();
            var listItemIndex = ts.indexOfNode(children, node);
            return {
                listItemIndex: listItemIndex,
                list: list
            };
        }
        ts.findListItemInfo = findListItemInfo;
        function hasChildOfKind(n, kind, sourceFile) {
            return !!findChildOfKind(n, kind, sourceFile);
        }
        ts.hasChildOfKind = hasChildOfKind;
        function findChildOfKind(n, kind, sourceFile) {
            return ts.find(n.getChildren(sourceFile), function (c) { return c.kind === kind; });
        }
        ts.findChildOfKind = findChildOfKind;
        function findContainingList(node) {
            // The node might be a list element (nonsynthetic) or a comma (synthetic). Either way, it will
            // be parented by the container of the SyntaxList, not the SyntaxList itself.
            // In order to find the list item index, we first need to locate SyntaxList itself and then search
            // for the position of the relevant node (or comma).
            var syntaxList = ts.find(node.parent.getChildren(), function (c) { return ts.isSyntaxList(c) && rangeContainsRange(c, node); });
            // Either we didn't find an appropriate list, or the list must contain us.
            ts.Debug.assert(!syntaxList || ts.contains(syntaxList.getChildren(), node));
            return syntaxList;
        }
        ts.findContainingList = findContainingList;
        function isDefaultModifier(node) {
            return node.kind === 88 /* DefaultKeyword */;
        }
        function isClassKeyword(node) {
            return node.kind === 84 /* ClassKeyword */;
        }
        function isFunctionKeyword(node) {
            return node.kind === 98 /* FunctionKeyword */;
        }
        function getAdjustedLocationForClass(node) {
            if (ts.isNamedDeclaration(node)) {
                return node.name;
            }
            if (ts.isClassDeclaration(node)) {
                // for class and function declarations, use the `default` modifier
                // when the declaration is unnamed.
                var defaultModifier = node.modifiers && ts.find(node.modifiers, isDefaultModifier);
                if (defaultModifier)
                    return defaultModifier;
            }
            if (ts.isClassExpression(node)) {
                // for class expressions, use the `class` keyword when the class is unnamed
                var classKeyword = ts.find(node.getChildren(), isClassKeyword);
                if (classKeyword)
                    return classKeyword;
            }
        }
        function getAdjustedLocationForFunction(node) {
            if (ts.isNamedDeclaration(node)) {
                return node.name;
            }
            if (ts.isFunctionDeclaration(node)) {
                // for class and function declarations, use the `default` modifier
                // when the declaration is unnamed.
                var defaultModifier = ts.find(node.modifiers, isDefaultModifier);
                if (defaultModifier)
                    return defaultModifier;
            }
            if (ts.isFunctionExpression(node)) {
                // for function expressions, use the `function` keyword when the function is unnamed
                var functionKeyword = ts.find(node.getChildren(), isFunctionKeyword);
                if (functionKeyword)
                    return functionKeyword;
            }
        }
        function getAncestorTypeNode(node) {
            var lastTypeNode;
            ts.findAncestor(node, function (a) {
                if (ts.isTypeNode(a)) {
                    lastTypeNode = a;
                }
                return !ts.isQualifiedName(a.parent) && !ts.isTypeNode(a.parent) && !ts.isTypeElement(a.parent);
            });
            return lastTypeNode;
        }
        function getContextualTypeOrAncestorTypeNodeType(node, checker) {
            var contextualType = checker.getContextualType(node);
            if (contextualType) {
                return contextualType;
            }
            var parent = node.parent;
            if (parent && ts.isBinaryExpression(parent) && isEqualityOperatorKind(parent.operatorToken.kind)) {
                return checker.getTypeAtLocation(node === parent.left ? parent.right : parent.left);
            }
            var ancestorTypeNode = getAncestorTypeNode(node);
            return ancestorTypeNode && checker.getTypeAtLocation(ancestorTypeNode);
        }
        ts.getContextualTypeOrAncestorTypeNodeType = getContextualTypeOrAncestorTypeNodeType;
        function getAdjustedLocationForDeclaration(node, forRename) {
            if (!forRename) {
                switch (node.kind) {
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                        return getAdjustedLocationForClass(node);
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                        return getAdjustedLocationForFunction(node);
                }
            }
            if (ts.isNamedDeclaration(node)) {
                return node.name;
            }
        }
        function getAdjustedLocationForImportDeclaration(node, forRename) {
            if (node.importClause) {
                if (node.importClause.name && node.importClause.namedBindings) {
                    // do not adjust if we have both a name and named bindings
                    return;
                }
                // /**/import [|name|] from ...;
                // import /**/type [|name|] from ...;
                if (node.importClause.name) {
                    return node.importClause.name;
                }
                // /**/import { [|name|] } from ...;
                // /**/import { propertyName as [|name|] } from ...;
                // /**/import * as [|name|] from ...;
                // import /**/type { [|name|] } from ...;
                // import /**/type { propertyName as [|name|] } from ...;
                // import /**/type * as [|name|] from ...;
                if (node.importClause.namedBindings) {
                    if (ts.isNamedImports(node.importClause.namedBindings)) {
                        // do nothing if there is more than one binding
                        var onlyBinding = ts.singleOrUndefined(node.importClause.namedBindings.elements);
                        if (!onlyBinding) {
                            return;
                        }
                        return onlyBinding.name;
                    }
                    else if (ts.isNamespaceImport(node.importClause.namedBindings)) {
                        return node.importClause.namedBindings.name;
                    }
                }
            }
            if (!forRename) {
                // /**/import "[|module|]";
                // /**/import ... from "[|module|]";
                // import /**/type ... from "[|module|]";
                return node.moduleSpecifier;
            }
        }
        function getAdjustedLocationForExportDeclaration(node, forRename) {
            if (node.exportClause) {
                // /**/export { [|name|] } ...
                // /**/export { propertyName as [|name|] } ...
                // /**/export * as [|name|] ...
                // export /**/type { [|name|] } from ...
                // export /**/type { propertyName as [|name|] } from ...
                // export /**/type * as [|name|] ...
                if (ts.isNamedExports(node.exportClause)) {
                    // do nothing if there is more than one binding
                    var onlyBinding = ts.singleOrUndefined(node.exportClause.elements);
                    if (!onlyBinding) {
                        return;
                    }
                    return node.exportClause.elements[0].name;
                }
                else if (ts.isNamespaceExport(node.exportClause)) {
                    return node.exportClause.name;
                }
            }
            if (!forRename) {
                // /**/export * from "[|module|]";
                // export /**/type * from "[|module|]";
                return node.moduleSpecifier;
            }
        }
        function getAdjustedLocationForHeritageClause(node) {
            // /**/extends [|name|]
            // /**/implements [|name|]
            if (node.types.length === 1) {
                return node.types[0].expression;
            }
            // /**/extends name1, name2 ...
            // /**/implements name1, name2 ...
        }
        function getAdjustedLocation(node, forRename) {
            var parent = node.parent;
            // /**/<modifier> [|name|] ...
            // /**/<modifier> <class|interface|type|enum|module|namespace|function|get|set> [|name|] ...
            // /**/<class|interface|type|enum|module|namespace|function|get|set> [|name|] ...
            // /**/import [|name|] = ...
            //
            // NOTE: If the node is a modifier, we don't adjust its location if it is the `default` modifier as that is handled
            // specially by `getSymbolAtLocation`.
            if (ts.isModifier(node) && (forRename || node.kind !== 88 /* DefaultKeyword */) ? ts.contains(parent.modifiers, node) :
                node.kind === 84 /* ClassKeyword */ ? ts.isClassDeclaration(parent) || ts.isClassExpression(node) :
                    node.kind === 98 /* FunctionKeyword */ ? ts.isFunctionDeclaration(parent) || ts.isFunctionExpression(node) :
                        node.kind === 118 /* InterfaceKeyword */ ? ts.isInterfaceDeclaration(parent) :
                            node.kind === 92 /* EnumKeyword */ ? ts.isEnumDeclaration(parent) :
                                node.kind === 150 /* TypeKeyword */ ? ts.isTypeAliasDeclaration(parent) :
                                    node.kind === 141 /* NamespaceKeyword */ || node.kind === 140 /* ModuleKeyword */ ? ts.isModuleDeclaration(parent) :
                                        node.kind === 100 /* ImportKeyword */ ? ts.isImportEqualsDeclaration(parent) :
                                            node.kind === 135 /* GetKeyword */ ? ts.isGetAccessorDeclaration(parent) :
                                                node.kind === 147 /* SetKeyword */ && ts.isSetAccessorDeclaration(parent)) {
                var location = getAdjustedLocationForDeclaration(parent, forRename);
                if (location) {
                    return location;
                }
            }
            // /**/<var|let|const> [|name|] ...
            if ((node.kind === 113 /* VarKeyword */ || node.kind === 85 /* ConstKeyword */ || node.kind === 119 /* LetKeyword */) &&
                ts.isVariableDeclarationList(parent) && parent.declarations.length === 1) {
                var decl = parent.declarations[0];
                if (ts.isIdentifier(decl.name)) {
                    return decl.name;
                }
            }
            if (node.kind === 150 /* TypeKeyword */) {
                // import /**/type [|name|] from ...;
                // import /**/type { [|name|] } from ...;
                // import /**/type { propertyName as [|name|] } from ...;
                // import /**/type ... from "[|module|]";
                if (ts.isImportClause(parent) && parent.isTypeOnly) {
                    var location = getAdjustedLocationForImportDeclaration(parent.parent, forRename);
                    if (location) {
                        return location;
                    }
                }
                // export /**/type { [|name|] } from ...;
                // export /**/type { propertyName as [|name|] } from ...;
                // export /**/type * from "[|module|]";
                // export /**/type * as ... from "[|module|]";
                if (ts.isExportDeclaration(parent) && parent.isTypeOnly) {
                    var location = getAdjustedLocationForExportDeclaration(parent, forRename);
                    if (location) {
                        return location;
                    }
                }
            }
            // import { propertyName /**/as [|name|] } ...
            // import * /**/as [|name|] ...
            // export { propertyName /**/as [|name|] } ...
            // export * /**/as [|name|] ...
            if (node.kind === 127 /* AsKeyword */) {
                if (ts.isImportSpecifier(parent) && parent.propertyName ||
                    ts.isExportSpecifier(parent) && parent.propertyName ||
                    ts.isNamespaceImport(parent) ||
                    ts.isNamespaceExport(parent)) {
                    return parent.name;
                }
                if (ts.isExportDeclaration(parent) && parent.exportClause && ts.isNamespaceExport(parent.exportClause)) {
                    return parent.exportClause.name;
                }
            }
            // /**/import [|name|] from ...;
            // /**/import { [|name|] } from ...;
            // /**/import { propertyName as [|name|] } from ...;
            // /**/import ... from "[|module|]";
            // /**/import "[|module|]";
            if (node.kind === 100 /* ImportKeyword */ && ts.isImportDeclaration(parent)) {
                var location = getAdjustedLocationForImportDeclaration(parent, forRename);
                if (location) {
                    return location;
                }
            }
            if (node.kind === 93 /* ExportKeyword */) {
                // /**/export { [|name|] } ...;
                // /**/export { propertyName as [|name|] } ...;
                // /**/export * from "[|module|]";
                // /**/export * as ... from "[|module|]";
                if (ts.isExportDeclaration(parent)) {
                    var location = getAdjustedLocationForExportDeclaration(parent, forRename);
                    if (location) {
                        return location;
                    }
                }
                // NOTE: We don't adjust the location of the `default` keyword as that is handled specially by `getSymbolAtLocation`.
                // /**/export default [|name|];
                // /**/export = [|name|];
                if (ts.isExportAssignment(parent)) {
                    return ts.skipOuterExpressions(parent.expression);
                }
            }
            // import name = /**/require("[|module|]");
            if (node.kind === 144 /* RequireKeyword */ && ts.isExternalModuleReference(parent)) {
                return parent.expression;
            }
            // import ... /**/from "[|module|]";
            // export ... /**/from "[|module|]";
            if (node.kind === 154 /* FromKeyword */ && (ts.isImportDeclaration(parent) || ts.isExportDeclaration(parent)) && parent.moduleSpecifier) {
                return parent.moduleSpecifier;
            }
            // class ... /**/extends [|name|] ...
            // class ... /**/implements [|name|] ...
            // class ... /**/implements name1, name2 ...
            // interface ... /**/extends [|name|] ...
            // interface ... /**/extends name1, name2 ...
            if ((node.kind === 94 /* ExtendsKeyword */ || node.kind === 117 /* ImplementsKeyword */) && ts.isHeritageClause(parent) && parent.token === node.kind) {
                var location = getAdjustedLocationForHeritageClause(parent);
                if (location) {
                    return location;
                }
            }
            if (node.kind === 94 /* ExtendsKeyword */) {
                // ... <T /**/extends [|U|]> ...
                if (ts.isTypeParameterDeclaration(parent) && parent.constraint && ts.isTypeReferenceNode(parent.constraint)) {
                    return parent.constraint.typeName;
                }
                // ... T /**/extends [|U|] ? ...
                if (ts.isConditionalTypeNode(parent) && ts.isTypeReferenceNode(parent.extendsType)) {
                    return parent.extendsType.typeName;
                }
            }
            // ... T extends /**/infer [|U|] ? ...
            if (node.kind === 136 /* InferKeyword */ && ts.isInferTypeNode(parent)) {
                return parent.typeParameter.name;
            }
            // { [ [|K|] /**/in keyof T]: ... }
            if (node.kind === 101 /* InKeyword */ && ts.isTypeParameterDeclaration(parent) && ts.isMappedTypeNode(parent.parent)) {
                return parent.name;
            }
            // /**/keyof [|T|]
            if (node.kind === 139 /* KeyOfKeyword */ && ts.isTypeOperatorNode(parent) && parent.operator === 139 /* KeyOfKeyword */ &&
                ts.isTypeReferenceNode(parent.type)) {
                return parent.type.typeName;
            }
            // /**/readonly [|name|][]
            if (node.kind === 143 /* ReadonlyKeyword */ && ts.isTypeOperatorNode(parent) && parent.operator === 143 /* ReadonlyKeyword */ &&
                ts.isArrayTypeNode(parent.type) && ts.isTypeReferenceNode(parent.type.elementType)) {
                return parent.type.elementType.typeName;
            }
            if (!forRename) {
                // /**/new [|name|]
                // /**/void [|name|]
                // /**/void obj.[|name|]
                // /**/typeof [|name|]
                // /**/typeof obj.[|name|]
                // /**/await [|name|]
                // /**/await obj.[|name|]
                // /**/yield [|name|]
                // /**/yield obj.[|name|]
                // /**/delete obj.[|name|]
                if (node.kind === 103 /* NewKeyword */ && ts.isNewExpression(parent) ||
                    node.kind === 114 /* VoidKeyword */ && ts.isVoidExpression(parent) ||
                    node.kind === 112 /* TypeOfKeyword */ && ts.isTypeOfExpression(parent) ||
                    node.kind === 131 /* AwaitKeyword */ && ts.isAwaitExpression(parent) ||
                    node.kind === 125 /* YieldKeyword */ && ts.isYieldExpression(parent) ||
                    node.kind === 89 /* DeleteKeyword */ && ts.isDeleteExpression(parent)) {
                    if (parent.expression) {
                        return ts.skipOuterExpressions(parent.expression);
                    }
                }
                // left /**/in [|name|]
                // left /**/instanceof [|name|]
                if ((node.kind === 101 /* InKeyword */ || node.kind === 102 /* InstanceOfKeyword */) && ts.isBinaryExpression(parent) && parent.operatorToken === node) {
                    return ts.skipOuterExpressions(parent.right);
                }
                // left /**/as [|name|]
                if (node.kind === 127 /* AsKeyword */ && ts.isAsExpression(parent) && ts.isTypeReferenceNode(parent.type)) {
                    return parent.type.typeName;
                }
                // for (... /**/in [|name|])
                // for (... /**/of [|name|])
                if (node.kind === 101 /* InKeyword */ && ts.isForInStatement(parent) ||
                    node.kind === 158 /* OfKeyword */ && ts.isForOfStatement(parent)) {
                    return ts.skipOuterExpressions(parent.expression);
                }
            }
            return node;
        }
        /**
         * Adjusts the location used for "find references" and "go to definition" when the cursor was not
         * on a property name.
         */
        function getAdjustedReferenceLocation(node) {
            return getAdjustedLocation(node, /*forRename*/ false);
        }
        ts.getAdjustedReferenceLocation = getAdjustedReferenceLocation;
        /**
         * Adjusts the location used for "rename" when the cursor was not on a property name.
         */
        function getAdjustedRenameLocation(node) {
            return getAdjustedLocation(node, /*forRename*/ true);
        }
        ts.getAdjustedRenameLocation = getAdjustedRenameLocation;
        /**
         * Gets the token whose text has range [start, end) and
         * position >= start and (position < end or (position === end && token is literal or keyword or identifier))
         */
        function getTouchingPropertyName(sourceFile, position) {
            return getTouchingToken(sourceFile, position, function (n) { return ts.isPropertyNameLiteral(n) || ts.isKeyword(n.kind) || ts.isPrivateIdentifier(n); });
        }
        ts.getTouchingPropertyName = getTouchingPropertyName;
        /**
         * Returns the token if position is in [start, end).
         * If position === end, returns the preceding token if includeItemAtEndPosition(previousToken) === true
         */
        function getTouchingToken(sourceFile, position, includePrecedingTokenAtEndPosition) {
            return getTokenAtPositionWorker(sourceFile, position, /*allowPositionInLeadingTrivia*/ false, includePrecedingTokenAtEndPosition, /*includeEndPosition*/ false);
        }
        ts.getTouchingToken = getTouchingToken;
        /** Returns a token if position is in [start-of-leading-trivia, end) */
        function getTokenAtPosition(sourceFile, position) {
            return getTokenAtPositionWorker(sourceFile, position, /*allowPositionInLeadingTrivia*/ true, /*includePrecedingTokenAtEndPosition*/ undefined, /*includeEndPosition*/ false);
        }
        ts.getTokenAtPosition = getTokenAtPosition;
        /** Get the token whose text contains the position */
        function getTokenAtPositionWorker(sourceFile, position, allowPositionInLeadingTrivia, includePrecedingTokenAtEndPosition, includeEndPosition) {
            var current = sourceFile;
            outer: while (true) {
                // find the child that contains 'position'
                for (var _i = 0, _a = current.getChildren(sourceFile); _i < _a.length; _i++) {
                    var child = _a[_i];
                    var start = allowPositionInLeadingTrivia ? child.getFullStart() : child.getStart(sourceFile, /*includeJsDoc*/ true);
                    if (start > position) {
                        // If this child begins after position, then all subsequent children will as well.
                        break;
                    }
                    var end = child.getEnd();
                    if (position < end || (position === end && (child.kind === 1 /* EndOfFileToken */ || includeEndPosition))) {
                        current = child;
                        continue outer;
                    }
                    else if (includePrecedingTokenAtEndPosition && end === position) {
                        var previousToken = findPrecedingToken(position, sourceFile, child);
                        if (previousToken && includePrecedingTokenAtEndPosition(previousToken)) {
                            return previousToken;
                        }
                    }
                }
                return current;
            }
        }
        /**
         * Returns the first token where position is in [start, end),
         * excluding `JsxText` tokens containing only whitespace.
         */
        function findFirstNonJsxWhitespaceToken(sourceFile, position) {
            var tokenAtPosition = getTokenAtPosition(sourceFile, position);
            while (isWhiteSpaceOnlyJsxText(tokenAtPosition)) {
                var nextToken = findNextToken(tokenAtPosition, tokenAtPosition.parent, sourceFile);
                if (!nextToken)
                    return;
                tokenAtPosition = nextToken;
            }
            return tokenAtPosition;
        }
        ts.findFirstNonJsxWhitespaceToken = findFirstNonJsxWhitespaceToken;
        /**
         * The token on the left of the position is the token that strictly includes the position
         * or sits to the left of the cursor if it is on a boundary. For example
         *
         *   fo|o               -> will return foo
         *   foo <comment> |bar -> will return foo
         *
         */
        function findTokenOnLeftOfPosition(file, position) {
            // Ideally, getTokenAtPosition should return a token. However, it is currently
            // broken, so we do a check to make sure the result was indeed a token.
            var tokenAtPosition = getTokenAtPosition(file, position);
            if (ts.isToken(tokenAtPosition) && position > tokenAtPosition.getStart(file) && position < tokenAtPosition.getEnd()) {
                return tokenAtPosition;
            }
            return findPrecedingToken(position, file);
        }
        ts.findTokenOnLeftOfPosition = findTokenOnLeftOfPosition;
        function findNextToken(previousToken, parent, sourceFile) {
            return find(parent);
            function find(n) {
                if (ts.isToken(n) && n.pos === previousToken.end) {
                    // this is token that starts at the end of previous token - return it
                    return n;
                }
                return ts.firstDefined(n.getChildren(sourceFile), function (child) {
                    var shouldDiveInChildNode = 
                    // previous token is enclosed somewhere in the child
                    (child.pos <= previousToken.pos && child.end > previousToken.end) ||
                        // previous token ends exactly at the beginning of child
                        (child.pos === previousToken.end);
                    return shouldDiveInChildNode && nodeHasTokens(child, sourceFile) ? find(child) : undefined;
                });
            }
        }
        ts.findNextToken = findNextToken;
        /**
         * Finds the rightmost token satisfying `token.end <= position`,
         * excluding `JsxText` tokens containing only whitespace.
         */
        function findPrecedingToken(position, sourceFile, startNode, excludeJsdoc) {
            var result = find(startNode || sourceFile);
            ts.Debug.assert(!(result && isWhiteSpaceOnlyJsxText(result)));
            return result;
            function find(n) {
                if (isNonWhitespaceToken(n) && n.kind !== 1 /* EndOfFileToken */) {
                    return n;
                }
                var children = n.getChildren(sourceFile);
                var i = ts.binarySearchKey(children, position, function (_, i) { return i; }, function (middle, _) {
                    // This last callback is more of a selector than a comparator -
                    // `EqualTo` causes the `middle` result to be returned
                    // `GreaterThan` causes recursion on the left of the middle
                    // `LessThan` causes recursion on the right of the middle
                    if (position < children[middle].end) {
                        // first element whose end position is greater than the input position
                        if (!children[middle - 1] || position >= children[middle - 1].end) {
                            return 0 /* EqualTo */;
                        }
                        return 1 /* GreaterThan */;
                    }
                    return -1 /* LessThan */;
                });
                if (i >= 0 && children[i]) {
                    var child = children[i];
                    // Note that the span of a node's tokens is [node.getStart(...), node.end).
                    // Given that `position < child.end` and child has constituent tokens, we distinguish these cases:
                    // 1) `position` precedes `child`'s tokens or `child` has no tokens (ie: in a comment or whitespace preceding `child`):
                    // we need to find the last token in a previous child.
                    // 2) `position` is within the same span: we recurse on `child`.
                    if (position < child.end) {
                        var start = child.getStart(sourceFile, /*includeJsDoc*/ !excludeJsdoc);
                        var lookInPreviousChild = (start >= position) || // cursor in the leading trivia
                            !nodeHasTokens(child, sourceFile) ||
                            isWhiteSpaceOnlyJsxText(child);
                        if (lookInPreviousChild) {
                            // actual start of the node is past the position - previous token should be at the end of previous child
                            var candidate_1 = findRightmostChildNodeWithTokens(children, /*exclusiveStartPosition*/ i, sourceFile);
                            return candidate_1 && findRightmostToken(candidate_1, sourceFile);
                        }
                        else {
                            // candidate should be in this node
                            return find(child);
                        }
                    }
                }
                ts.Debug.assert(startNode !== undefined || n.kind === 300 /* SourceFile */ || n.kind === 1 /* EndOfFileToken */ || ts.isJSDocCommentContainingNode(n));
                // Here we know that none of child token nodes embrace the position,
                // the only known case is when position is at the end of the file.
                // Try to find the rightmost token in the file without filtering.
                // Namely we are skipping the check: 'position < node.end'
                var candidate = findRightmostChildNodeWithTokens(children, /*exclusiveStartPosition*/ children.length, sourceFile);
                return candidate && findRightmostToken(candidate, sourceFile);
            }
        }
        ts.findPrecedingToken = findPrecedingToken;
        function isNonWhitespaceToken(n) {
            return ts.isToken(n) && !isWhiteSpaceOnlyJsxText(n);
        }
        function findRightmostToken(n, sourceFile) {
            if (isNonWhitespaceToken(n)) {
                return n;
            }
            var children = n.getChildren(sourceFile);
            if (children.length === 0) {
                return n;
            }
            var candidate = findRightmostChildNodeWithTokens(children, /*exclusiveStartPosition*/ children.length, sourceFile);
            return candidate && findRightmostToken(candidate, sourceFile);
        }
        /**
         * Finds the rightmost child to the left of `children[exclusiveStartPosition]` which is a non-all-whitespace token or has constituent tokens.
         */
        function findRightmostChildNodeWithTokens(children, exclusiveStartPosition, sourceFile) {
            for (var i = exclusiveStartPosition - 1; i >= 0; i--) {
                var child = children[i];
                if (isWhiteSpaceOnlyJsxText(child)) {
                    ts.Debug.assert(i > 0, "`JsxText` tokens should not be the first child of `JsxElement | JsxSelfClosingElement`");
                }
                else if (nodeHasTokens(children[i], sourceFile)) {
                    return children[i];
                }
            }
        }
        function isInString(sourceFile, position, previousToken) {
            if (previousToken === void 0) {
                previousToken = findPrecedingToken(position, sourceFile);
            }
            if (previousToken && ts.isStringTextContainingNode(previousToken)) {
                var start = previousToken.getStart(sourceFile);
                var end = previousToken.getEnd();
                // To be "in" one of these literals, the position has to be:
                //   1. entirely within the token text.
                //   2. at the end position of an unterminated token.
                //   3. at the end of a regular expression (due to trailing flags like '/foo/g').
                if (start < position && position < end) {
                    return true;
                }
                if (position === end) {
                    return !!previousToken.isUnterminated;
                }
            }
            return false;
        }
        ts.isInString = isInString;
        /**
         * returns true if the position is in between the open and close elements of an JSX expression.
         */
        function isInsideJsxElementOrAttribute(sourceFile, position) {
            var token = getTokenAtPosition(sourceFile, position);
            if (!token) {
                return false;
            }
            if (token.kind === 11 /* JsxText */) {
                return true;
            }
            // <div>Hello |</div>
            if (token.kind === 29 /* LessThanToken */ && token.parent.kind === 11 /* JsxText */) {
                return true;
            }
            // <div> { | </div> or <div a={| </div>
            if (token.kind === 29 /* LessThanToken */ && token.parent.kind === 286 /* JsxExpression */) {
                return true;
            }
            // <div> {
            // |
            // } < /div>
            if (token && token.kind === 19 /* CloseBraceToken */ && token.parent.kind === 286 /* JsxExpression */) {
                return true;
            }
            // <div>|</div>
            if (token.kind === 29 /* LessThanToken */ && token.parent.kind === 279 /* JsxClosingElement */) {
                return true;
            }
            return false;
        }
        ts.isInsideJsxElementOrAttribute = isInsideJsxElementOrAttribute;
        function isWhiteSpaceOnlyJsxText(node) {
            return ts.isJsxText(node) && node.containsOnlyTriviaWhiteSpaces;
        }
        function isInTemplateString(sourceFile, position) {
            var token = getTokenAtPosition(sourceFile, position);
            return ts.isTemplateLiteralKind(token.kind) && position > token.getStart(sourceFile);
        }
        ts.isInTemplateString = isInTemplateString;
        function isInJSXText(sourceFile, position) {
            var token = getTokenAtPosition(sourceFile, position);
            if (ts.isJsxText(token)) {
                return true;
            }
            if (token.kind === 18 /* OpenBraceToken */ && ts.isJsxExpression(token.parent) && ts.isJsxElement(token.parent.parent)) {
                return true;
            }
            if (token.kind === 29 /* LessThanToken */ && ts.isJsxOpeningLikeElement(token.parent) && ts.isJsxElement(token.parent.parent)) {
                return true;
            }
            return false;
        }
        ts.isInJSXText = isInJSXText;
        function isInsideJsxElement(sourceFile, position) {
            function isInsideJsxElementTraversal(node) {
                while (node) {
                    if (node.kind >= 277 /* JsxSelfClosingElement */ && node.kind <= 286 /* JsxExpression */
                        || node.kind === 11 /* JsxText */
                        || node.kind === 29 /* LessThanToken */
                        || node.kind === 31 /* GreaterThanToken */
                        || node.kind === 79 /* Identifier */
                        || node.kind === 19 /* CloseBraceToken */
                        || node.kind === 18 /* OpenBraceToken */
                        || node.kind === 43 /* SlashToken */) {
                        node = node.parent;
                    }
                    else if (node.kind === 276 /* JsxElement */) {
                        if (position > node.getStart(sourceFile))
                            return true;
                        node = node.parent;
                    }
                    else {
                        return false;
                    }
                }
                return false;
            }
            return isInsideJsxElementTraversal(getTokenAtPosition(sourceFile, position));
        }
        ts.isInsideJsxElement = isInsideJsxElement;
        function findPrecedingMatchingToken(token, matchingTokenKind, sourceFile) {
            var closeTokenText = ts.tokenToString(token.kind);
            var matchingTokenText = ts.tokenToString(matchingTokenKind);
            var tokenFullStart = token.getFullStart();
            // Text-scan based fast path - can be bamboozled by comments and other trivia, but often provides
            // a good, fast approximation without too much extra work in the cases where it fails.
            var bestGuessIndex = sourceFile.text.lastIndexOf(matchingTokenText, tokenFullStart);
            if (bestGuessIndex === -1) {
                return undefined; // if the token text doesn't appear in the file, there can't be a match - super fast bail
            }
            // we can only use the textual result directly if we didn't have to count any close tokens within the range
            if (sourceFile.text.lastIndexOf(closeTokenText, tokenFullStart - 1) < bestGuessIndex) {
                var nodeAtGuess = findPrecedingToken(bestGuessIndex + 1, sourceFile);
                if (nodeAtGuess && nodeAtGuess.kind === matchingTokenKind) {
                    return nodeAtGuess;
                }
            }
            var tokenKind = token.kind;
            var remainingMatchingTokens = 0;
            while (true) {
                var preceding = findPrecedingToken(token.getFullStart(), sourceFile);
                if (!preceding) {
                    return undefined;
                }
                token = preceding;
                if (token.kind === matchingTokenKind) {
                    if (remainingMatchingTokens === 0) {
                        return token;
                    }
                    remainingMatchingTokens--;
                }
                else if (token.kind === tokenKind) {
                    remainingMatchingTokens++;
                }
            }
        }
        ts.findPrecedingMatchingToken = findPrecedingMatchingToken;
        function removeOptionality(type, isOptionalExpression, isOptionalChain) {
            return isOptionalExpression ? type.getNonNullableType() :
                isOptionalChain ? type.getNonOptionalType() :
                    type;
        }
        ts.removeOptionality = removeOptionality;
        function isPossiblyTypeArgumentPosition(token, sourceFile, checker) {
            var info = getPossibleTypeArgumentsInfo(token, sourceFile);
            return info !== undefined && (ts.isPartOfTypeNode(info.called) ||
                getPossibleGenericSignatures(info.called, info.nTypeArguments, checker).length !== 0 ||
                isPossiblyTypeArgumentPosition(info.called, sourceFile, checker));
        }
        ts.isPossiblyTypeArgumentPosition = isPossiblyTypeArgumentPosition;
        function getPossibleGenericSignatures(called, typeArgumentCount, checker) {
            var type = checker.getTypeAtLocation(called);
            if (ts.isOptionalChain(called.parent)) {
                type = removeOptionality(type, ts.isOptionalChainRoot(called.parent), /*isOptionalChain*/ true);
            }
            var signatures = ts.isNewExpression(called.parent) ? type.getConstructSignatures() : type.getCallSignatures();
            return signatures.filter(function (candidate) { return !!candidate.typeParameters && candidate.typeParameters.length >= typeArgumentCount; });
        }
        ts.getPossibleGenericSignatures = getPossibleGenericSignatures;
        // Get info for an expression like `f <` that may be the start of type arguments.
        function getPossibleTypeArgumentsInfo(tokenIn, sourceFile) {
            // This is a rare case, but one that saves on a _lot_ of work if true - if the source file has _no_ `<` character,
            // then there obviously can't be any type arguments - no expensive brace-matching backwards scanning required
            if (sourceFile.text.lastIndexOf("<", tokenIn ? tokenIn.pos : sourceFile.text.length) === -1) {
                return undefined;
            }
            var token = tokenIn;
            // This function determines if the node could be type argument position
            // Since during editing, when type argument list is not complete,
            // the tree could be of any shape depending on the tokens parsed before current node,
            // scanning of the previous identifier followed by "<" before current node would give us better result
            // Note that we also balance out the already provided type arguments, arrays, object literals while doing so
            var remainingLessThanTokens = 0;
            var nTypeArguments = 0;
            while (token) {
                switch (token.kind) {
                    case 29 /* LessThanToken */:
                        // Found the beginning of the generic argument expression
                        token = findPrecedingToken(token.getFullStart(), sourceFile);
                        if (token && token.kind === 28 /* QuestionDotToken */) {
                            token = findPrecedingToken(token.getFullStart(), sourceFile);
                        }
                        if (!token || !ts.isIdentifier(token))
                            return undefined;
                        if (!remainingLessThanTokens) {
                            return ts.isDeclarationName(token) ? undefined : { called: token, nTypeArguments: nTypeArguments };
                        }
                        remainingLessThanTokens--;
                        break;
                    case 49 /* GreaterThanGreaterThanGreaterThanToken */:
                        remainingLessThanTokens = +3;
                        break;
                    case 48 /* GreaterThanGreaterThanToken */:
                        remainingLessThanTokens = +2;
                        break;
                    case 31 /* GreaterThanToken */:
                        remainingLessThanTokens++;
                        break;
                    case 19 /* CloseBraceToken */:
                        // This can be object type, skip until we find the matching open brace token
                        // Skip until the matching open brace token
                        token = findPrecedingMatchingToken(token, 18 /* OpenBraceToken */, sourceFile);
                        if (!token)
                            return undefined;
                        break;
                    case 21 /* CloseParenToken */:
                        // This can be object type, skip until we find the matching open brace token
                        // Skip until the matching open brace token
                        token = findPrecedingMatchingToken(token, 20 /* OpenParenToken */, sourceFile);
                        if (!token)
                            return undefined;
                        break;
                    case 23 /* CloseBracketToken */:
                        // This can be object type, skip until we find the matching open brace token
                        // Skip until the matching open brace token
                        token = findPrecedingMatchingToken(token, 22 /* OpenBracketToken */, sourceFile);
                        if (!token)
                            return undefined;
                        break;
                    // Valid tokens in a type name. Skip.
                    case 27 /* CommaToken */:
                        nTypeArguments++;
                        break;
                    case 38 /* EqualsGreaterThanToken */:
                    // falls through
                    case 79 /* Identifier */:
                    case 10 /* StringLiteral */:
                    case 8 /* NumericLiteral */:
                    case 9 /* BigIntLiteral */:
                    case 110 /* TrueKeyword */:
                    case 95 /* FalseKeyword */:
                    // falls through
                    case 112 /* TypeOfKeyword */:
                    case 94 /* ExtendsKeyword */:
                    case 139 /* KeyOfKeyword */:
                    case 24 /* DotToken */:
                    case 51 /* BarToken */:
                    case 57 /* QuestionToken */:
                    case 58 /* ColonToken */:
                        break;
                    default:
                        if (ts.isTypeNode(token)) {
                            break;
                        }
                        // Invalid token in type
                        return undefined;
                }
                token = findPrecedingToken(token.getFullStart(), sourceFile);
            }
            return undefined;
        }
        ts.getPossibleTypeArgumentsInfo = getPossibleTypeArgumentsInfo;
        /**
         * Returns true if the cursor at position in sourceFile is within a comment.
         *
         * @param tokenAtPosition Must equal `getTokenAtPosition(sourceFile, position)
         * @param predicate Additional predicate to test on the comment range.
         */
        function isInComment(sourceFile, position, tokenAtPosition) {
            return ts.formatting.getRangeOfEnclosingComment(sourceFile, position, /*precedingToken*/ undefined, tokenAtPosition);
        }
        ts.isInComment = isInComment;
        function hasDocComment(sourceFile, position) {
            var token = getTokenAtPosition(sourceFile, position);
            return !!ts.findAncestor(token, ts.isJSDoc);
        }
        ts.hasDocComment = hasDocComment;
        function nodeHasTokens(n, sourceFile) {
            // If we have a token or node that has a non-zero width, it must have tokens.
            // Note: getWidth() does not take trivia into account.
            return n.kind === 1 /* EndOfFileToken */ ? !!n.jsDoc : n.getWidth(sourceFile) !== 0;
        }
        function getNodeModifiers(node, excludeFlags) {
            if (excludeFlags === void 0) {
                excludeFlags = 0 /* None */;
            }
            var result = [];
            var flags = ts.isDeclaration(node)
                ? ts.getCombinedNodeFlagsAlwaysIncludeJSDoc(node) & ~excludeFlags
                : 0 /* None */;
            if (flags & 8 /* Private */)
                result.push("private" /* privateMemberModifier */);
            if (flags & 16 /* Protected */)
                result.push("protected" /* protectedMemberModifier */);
            if (flags & 4 /* Public */)
                result.push("public" /* publicMemberModifier */);
            if (flags & 32 /* Static */ || ts.isClassStaticBlockDeclaration(node))
                result.push("static" /* staticModifier */);
            if (flags & 128 /* Abstract */)
                result.push("abstract" /* abstractModifier */);
            if (flags & 1 /* Export */)
                result.push("export" /* exportedModifier */);
            if (flags & 8192 /* Deprecated */)
                result.push("deprecated" /* deprecatedModifier */);
            if (node.flags & 8388608 /* Ambient */)
                result.push("declare" /* ambientModifier */);
            if (node.kind === 269 /* ExportAssignment */)
                result.push("export" /* exportedModifier */);
            return result.length > 0 ? result.join(",") : "" /* none */;
        }
        ts.getNodeModifiers = getNodeModifiers;
        function getTypeArgumentOrTypeParameterList(node) {
            if (node.kind === 176 /* TypeReference */ || node.kind === 206 /* CallExpression */) {
                return node.typeArguments;
            }
            if (ts.isFunctionLike(node) || node.kind === 255 /* ClassDeclaration */ || node.kind === 256 /* InterfaceDeclaration */) {
                return node.typeParameters;
            }
            return undefined;
        }
        ts.getTypeArgumentOrTypeParameterList = getTypeArgumentOrTypeParameterList;
        function isComment(kind) {
            return kind === 2 /* SingleLineCommentTrivia */ || kind === 3 /* MultiLineCommentTrivia */;
        }
        ts.isComment = isComment;
        function isStringOrRegularExpressionOrTemplateLiteral(kind) {
            if (kind === 10 /* StringLiteral */
                || kind === 13 /* RegularExpressionLiteral */
                || ts.isTemplateLiteralKind(kind)) {
                return true;
            }
            return false;
        }
        ts.isStringOrRegularExpressionOrTemplateLiteral = isStringOrRegularExpressionOrTemplateLiteral;
        function isPunctuation(kind) {
            return 18 /* FirstPunctuation */ <= kind && kind <= 78 /* LastPunctuation */;
        }
        ts.isPunctuation = isPunctuation;
        function isInsideTemplateLiteral(node, position, sourceFile) {
            return ts.isTemplateLiteralKind(node.kind)
                && (node.getStart(sourceFile) < position && position < node.end) || (!!node.isUnterminated && position === node.end);
        }
        ts.isInsideTemplateLiteral = isInsideTemplateLiteral;
        function isAccessibilityModifier(kind) {
            switch (kind) {
                case 123 /* PublicKeyword */:
                case 121 /* PrivateKeyword */:
                case 122 /* ProtectedKeyword */:
                    return true;
            }
            return false;
        }
        ts.isAccessibilityModifier = isAccessibilityModifier;
        function cloneCompilerOptions(options) {
            var result = ts.clone(options);
            ts.setConfigFileInOptions(result, options && options.configFile);
            return result;
        }
        ts.cloneCompilerOptions = cloneCompilerOptions;
        function isArrayLiteralOrObjectLiteralDestructuringPattern(node) {
            if (node.kind === 202 /* ArrayLiteralExpression */ ||
                node.kind === 203 /* ObjectLiteralExpression */) {
                // [a,b,c] from:
                // [a, b, c] = someExpression;
                if (node.parent.kind === 219 /* BinaryExpression */ &&
                    node.parent.left === node &&
                    node.parent.operatorToken.kind === 63 /* EqualsToken */) {
                    return true;
                }
                // [a, b, c] from:
                // for([a, b, c] of expression)
                if (node.parent.kind === 242 /* ForOfStatement */ &&
                    node.parent.initializer === node) {
                    return true;
                }
                // [a, b, c] of
                // [x, [a, b, c] ] = someExpression
                // or
                // {x, a: {a, b, c} } = someExpression
                if (isArrayLiteralOrObjectLiteralDestructuringPattern(node.parent.kind === 291 /* PropertyAssignment */ ? node.parent.parent : node.parent)) {
                    return true;
                }
            }
            return false;
        }
        ts.isArrayLiteralOrObjectLiteralDestructuringPattern = isArrayLiteralOrObjectLiteralDestructuringPattern;
        function isInReferenceComment(sourceFile, position) {
            return isInReferenceCommentWorker(sourceFile, position, /*shouldBeReference*/ true);
        }
        ts.isInReferenceComment = isInReferenceComment;
        function isInNonReferenceComment(sourceFile, position) {
            return isInReferenceCommentWorker(sourceFile, position, /*shouldBeReference*/ false);
        }
        ts.isInNonReferenceComment = isInNonReferenceComment;
        function isInReferenceCommentWorker(sourceFile, position, shouldBeReference) {
            var range = isInComment(sourceFile, position, /*tokenAtPosition*/ undefined);
            return !!range && shouldBeReference === tripleSlashDirectivePrefixRegex.test(sourceFile.text.substring(range.pos, range.end));
        }
        function getReplacementSpanForContextToken(contextToken) {
            if (!contextToken)
                return undefined;
            switch (contextToken.kind) {
                case 10 /* StringLiteral */:
                case 14 /* NoSubstitutionTemplateLiteral */:
                    return createTextSpanFromStringLiteralLikeContent(contextToken);
                default:
                    return createTextSpanFromNode(contextToken);
            }
        }
        ts.getReplacementSpanForContextToken = getReplacementSpanForContextToken;
        function createTextSpanFromNode(node, sourceFile, endNode) {
            return ts.createTextSpanFromBounds(node.getStart(sourceFile), (endNode || node).getEnd());
        }
        ts.createTextSpanFromNode = createTextSpanFromNode;
        function createTextSpanFromStringLiteralLikeContent(node) {
            if (node.isUnterminated)
                return undefined;
            return ts.createTextSpanFromBounds(node.getStart() + 1, node.getEnd() - 1);
        }
        ts.createTextSpanFromStringLiteralLikeContent = createTextSpanFromStringLiteralLikeContent;
        function createTextRangeFromNode(node, sourceFile) {
            return ts.createRange(node.getStart(sourceFile), node.end);
        }
        ts.createTextRangeFromNode = createTextRangeFromNode;
        function createTextSpanFromRange(range) {
            return ts.createTextSpanFromBounds(range.pos, range.end);
        }
        ts.createTextSpanFromRange = createTextSpanFromRange;
        function createTextRangeFromSpan(span) {
            return ts.createRange(span.start, span.start + span.length);
        }
        ts.createTextRangeFromSpan = createTextRangeFromSpan;
        function createTextChangeFromStartLength(start, length, newText) {
            return createTextChange(ts.createTextSpan(start, length), newText);
        }
        ts.createTextChangeFromStartLength = createTextChangeFromStartLength;
        function createTextChange(span, newText) {
            return { span: span, newText: newText };
        }
        ts.createTextChange = createTextChange;
        ts.typeKeywords = [
            129 /* AnyKeyword */,
            128 /* AssertsKeyword */,
            156 /* BigIntKeyword */,
            132 /* BooleanKeyword */,
            95 /* FalseKeyword */,
            136 /* InferKeyword */,
            139 /* KeyOfKeyword */,
            142 /* NeverKeyword */,
            104 /* NullKeyword */,
            145 /* NumberKeyword */,
            146 /* ObjectKeyword */,
            143 /* ReadonlyKeyword */,
            148 /* StringKeyword */,
            149 /* SymbolKeyword */,
            110 /* TrueKeyword */,
            114 /* VoidKeyword */,
            151 /* UndefinedKeyword */,
            152 /* UniqueKeyword */,
            153 /* UnknownKeyword */,
        ];
        function isTypeKeyword(kind) {
            return ts.contains(ts.typeKeywords, kind);
        }
        ts.isTypeKeyword = isTypeKeyword;
        function isTypeKeywordToken(node) {
            return node.kind === 150 /* TypeKeyword */;
        }
        ts.isTypeKeywordToken = isTypeKeywordToken;
        /** True if the symbol is for an external module, as opposed to a namespace. */
        function isExternalModuleSymbol(moduleSymbol) {
            return !!(moduleSymbol.flags & 1536 /* Module */) && moduleSymbol.name.charCodeAt(0) === 34 /* doubleQuote */;
        }
        ts.isExternalModuleSymbol = isExternalModuleSymbol;
        function nodeSeenTracker() {
            var seen = [];
            return function (node) {
                var id = ts.getNodeId(node);
                return !seen[id] && (seen[id] = true);
            };
        }
        ts.nodeSeenTracker = nodeSeenTracker;
        function getSnapshotText(snap) {
            return snap.getText(0, snap.getLength());
        }
        ts.getSnapshotText = getSnapshotText;
        function repeatString(str, count) {
            var result = "";
            for (var i = 0; i < count; i++) {
                result += str;
            }
            return result;
        }
        ts.repeatString = repeatString;
        function skipConstraint(type) {
            return type.isTypeParameter() ? type.getConstraint() || type : type;
        }
        ts.skipConstraint = skipConstraint;
        function getNameFromPropertyName(name) {
            return name.kind === 160 /* ComputedPropertyName */
                // treat computed property names where expression is string/numeric literal as just string/numeric literal
                ? ts.isStringOrNumericLiteralLike(name.expression) ? name.expression.text : undefined
                : ts.isPrivateIdentifier(name) ? ts.idText(name) : ts.getTextOfIdentifierOrLiteral(name);
        }
        ts.getNameFromPropertyName = getNameFromPropertyName;
        function programContainsModules(program) {
            return program.getSourceFiles().some(function (s) { return !s.isDeclarationFile && !program.isSourceFileFromExternalLibrary(s) && !!(s.externalModuleIndicator || s.commonJsModuleIndicator); });
        }
        ts.programContainsModules = programContainsModules;
        function programContainsEs6Modules(program) {
            return program.getSourceFiles().some(function (s) { return !s.isDeclarationFile && !program.isSourceFileFromExternalLibrary(s) && !!s.externalModuleIndicator; });
        }
        ts.programContainsEs6Modules = programContainsEs6Modules;
        function compilerOptionsIndicateEs6Modules(compilerOptions) {
            return !!compilerOptions.module || compilerOptions.target >= 2 /* ES2015 */ || !!compilerOptions.noEmit;
        }
        ts.compilerOptionsIndicateEs6Modules = compilerOptionsIndicateEs6Modules;
        function createModuleSpecifierResolutionHost(program, host) {
            // Mix in `getSymlinkCache` from Program when host doesn't have it
            // in order for non-Project hosts to have a symlinks cache.
            return {
                fileExists: function (fileName) { return program.fileExists(fileName); },
                getCurrentDirectory: function () { return host.getCurrentDirectory(); },
                readFile: ts.maybeBind(host, host.readFile),
                useCaseSensitiveFileNames: ts.maybeBind(host, host.useCaseSensitiveFileNames),
                getSymlinkCache: ts.maybeBind(host, host.getSymlinkCache) || program.getSymlinkCache,
                getModuleSpecifierCache: ts.maybeBind(host, host.getModuleSpecifierCache),
                getGlobalTypingsCacheLocation: ts.maybeBind(host, host.getGlobalTypingsCacheLocation),
                redirectTargetsMap: program.redirectTargetsMap,
                getProjectReferenceRedirect: function (fileName) { return program.getProjectReferenceRedirect(fileName); },
                isSourceOfProjectReferenceRedirect: function (fileName) { return program.isSourceOfProjectReferenceRedirect(fileName); },
                getNearestAncestorDirectoryWithPackageJson: ts.maybeBind(host, host.getNearestAncestorDirectoryWithPackageJson),
                getFileIncludeReasons: function () { return program.getFileIncludeReasons(); },
            };
        }
        ts.createModuleSpecifierResolutionHost = createModuleSpecifierResolutionHost;
        function getModuleSpecifierResolverHost(program, host) {
            return __assign(__assign({}, createModuleSpecifierResolutionHost(program, host)), { getCommonSourceDirectory: function () { return program.getCommonSourceDirectory(); } });
        }
        ts.getModuleSpecifierResolverHost = getModuleSpecifierResolverHost;
        function makeImportIfNecessary(defaultImport, namedImports, moduleSpecifier, quotePreference) {
            return defaultImport || namedImports && namedImports.length ? makeImport(defaultImport, namedImports, moduleSpecifier, quotePreference) : undefined;
        }
        ts.makeImportIfNecessary = makeImportIfNecessary;
        function makeImport(defaultImport, namedImports, moduleSpecifier, quotePreference, isTypeOnly) {
            return ts.factory.createImportDeclaration(
            /*decorators*/ undefined, 
            /*modifiers*/ undefined, defaultImport || namedImports
                ? ts.factory.createImportClause(!!isTypeOnly, defaultImport, namedImports && namedImports.length ? ts.factory.createNamedImports(namedImports) : undefined)
                : undefined, typeof moduleSpecifier === "string" ? makeStringLiteral(moduleSpecifier, quotePreference) : moduleSpecifier);
        }
        ts.makeImport = makeImport;
        function makeStringLiteral(text, quotePreference) {
            return ts.factory.createStringLiteral(text, quotePreference === 0 /* Single */);
        }
        ts.makeStringLiteral = makeStringLiteral;
        var QuotePreference;
        (function (QuotePreference) {
            QuotePreference[QuotePreference["Single"] = 0] = "Single";
            QuotePreference[QuotePreference["Double"] = 1] = "Double";
        })(QuotePreference = ts.QuotePreference || (ts.QuotePreference = {}));
        function quotePreferenceFromString(str, sourceFile) {
            return ts.isStringDoubleQuoted(str, sourceFile) ? 1 /* Double */ : 0 /* Single */;
        }
        ts.quotePreferenceFromString = quotePreferenceFromString;
        function getQuotePreference(sourceFile, preferences) {
            if (preferences.quotePreference && preferences.quotePreference !== "auto") {
                return preferences.quotePreference === "single" ? 0 /* Single */ : 1 /* Double */;
            }
            else {
                // ignore synthetic import added when importHelpers: true
                var firstModuleSpecifier = sourceFile.imports &&
                    ts.find(sourceFile.imports, function (n) { return ts.isStringLiteral(n) && !ts.nodeIsSynthesized(n.parent); });
                return firstModuleSpecifier ? quotePreferenceFromString(firstModuleSpecifier, sourceFile) : 1 /* Double */;
            }
        }
        ts.getQuotePreference = getQuotePreference;
        function getQuoteFromPreference(qp) {
            switch (qp) {
                case 0 /* Single */: return "'";
                case 1 /* Double */: return '"';
                default: return ts.Debug.assertNever(qp);
            }
        }
        ts.getQuoteFromPreference = getQuoteFromPreference;
        function symbolNameNoDefault(symbol) {
            var escaped = symbolEscapedNameNoDefault(symbol);
            return escaped === undefined ? undefined : ts.unescapeLeadingUnderscores(escaped);
        }
        ts.symbolNameNoDefault = symbolNameNoDefault;
        function symbolEscapedNameNoDefault(symbol) {
            if (symbol.escapedName !== "default" /* Default */) {
                return symbol.escapedName;
            }
            return ts.firstDefined(symbol.declarations, function (decl) {
                var name = ts.getNameOfDeclaration(decl);
                return name && name.kind === 79 /* Identifier */ ? name.escapedText : undefined;
            });
        }
        ts.symbolEscapedNameNoDefault = symbolEscapedNameNoDefault;
        function isModuleSpecifierLike(node) {
            return ts.isStringLiteralLike(node) && (ts.isExternalModuleReference(node.parent) ||
                ts.isImportDeclaration(node.parent) ||
                ts.isRequireCall(node.parent, /*requireStringLiteralLikeArgument*/ false) && node.parent.arguments[0] === node ||
                ts.isImportCall(node.parent) && node.parent.arguments[0] === node);
        }
        ts.isModuleSpecifierLike = isModuleSpecifierLike;
        function isObjectBindingElementWithoutPropertyName(bindingElement) {
            return ts.isBindingElement(bindingElement) &&
                ts.isObjectBindingPattern(bindingElement.parent) &&
                ts.isIdentifier(bindingElement.name) &&
                !bindingElement.propertyName;
        }
        ts.isObjectBindingElementWithoutPropertyName = isObjectBindingElementWithoutPropertyName;
        function getPropertySymbolFromBindingElement(checker, bindingElement) {
            var typeOfPattern = checker.getTypeAtLocation(bindingElement.parent);
            return typeOfPattern && checker.getPropertyOfType(typeOfPattern, bindingElement.name.text);
        }
        ts.getPropertySymbolFromBindingElement = getPropertySymbolFromBindingElement;
        function getParentNodeInSpan(node, file, span) {
            if (!node)
                return undefined;
            while (node.parent) {
                if (ts.isSourceFile(node.parent) || !spanContainsNode(span, node.parent, file)) {
                    return node;
                }
                node = node.parent;
            }
        }
        ts.getParentNodeInSpan = getParentNodeInSpan;
        function spanContainsNode(span, node, file) {
            return ts.textSpanContainsPosition(span, node.getStart(file)) &&
                node.getEnd() <= ts.textSpanEnd(span);
        }
        function findModifier(node, kind) {
            return node.modifiers && ts.find(node.modifiers, function (m) { return m.kind === kind; });
        }
        ts.findModifier = findModifier;
        function insertImports(changes, sourceFile, imports, blankLineBetween) {
            var decl = ts.isArray(imports) ? imports[0] : imports;
            var importKindPredicate = decl.kind === 235 /* VariableStatement */ ? ts.isRequireVariableStatement : ts.isAnyImportSyntax;
            var existingImportStatements = ts.filter(sourceFile.statements, importKindPredicate);
            var sortedNewImports = ts.isArray(imports) ? ts.stableSort(imports, ts.OrganizeImports.compareImportsOrRequireStatements) : [imports];
            if (!existingImportStatements.length) {
                changes.insertNodesAtTopOfFile(sourceFile, sortedNewImports, blankLineBetween);
            }
            else if (existingImportStatements && ts.OrganizeImports.importsAreSorted(existingImportStatements)) {
                for (var _i = 0, sortedNewImports_1 = sortedNewImports; _i < sortedNewImports_1.length; _i++) {
                    var newImport = sortedNewImports_1[_i];
                    var insertionIndex = ts.OrganizeImports.getImportDeclarationInsertionIndex(existingImportStatements, newImport);
                    if (insertionIndex === 0) {
                        // If the first import is top-of-file, insert after the leading comment which is likely the header.
                        var options = existingImportStatements[0] === sourceFile.statements[0] ?
                            { leadingTriviaOption: ts.textChanges.LeadingTriviaOption.Exclude } : {};
                        changes.insertNodeBefore(sourceFile, existingImportStatements[0], newImport, /*blankLineBetween*/ false, options);
                    }
                    else {
                        var prevImport = existingImportStatements[insertionIndex - 1];
                        changes.insertNodeAfter(sourceFile, prevImport, newImport);
                    }
                }
            }
            else {
                var lastExistingImport = ts.lastOrUndefined(existingImportStatements);
                if (lastExistingImport) {
                    changes.insertNodesAfter(sourceFile, lastExistingImport, sortedNewImports);
                }
                else {
                    changes.insertNodesAtTopOfFile(sourceFile, sortedNewImports, blankLineBetween);
                }
            }
        }
        ts.insertImports = insertImports;
        function getTypeKeywordOfTypeOnlyImport(importClause, sourceFile) {
            ts.Debug.assert(importClause.isTypeOnly);
            return ts.cast(importClause.getChildAt(0, sourceFile), isTypeKeywordToken);
        }
        ts.getTypeKeywordOfTypeOnlyImport = getTypeKeywordOfTypeOnlyImport;
        function textSpansEqual(a, b) {
            return !!a && !!b && a.start === b.start && a.length === b.length;
        }
        ts.textSpansEqual = textSpansEqual;
        function documentSpansEqual(a, b) {
            return a.fileName === b.fileName && textSpansEqual(a.textSpan, b.textSpan);
        }
        ts.documentSpansEqual = documentSpansEqual;
        /**
         * Iterates through 'array' by index and performs the callback on each element of array until the callback
         * returns a truthy value, then returns that value.
         * If no such value is found, the callback is applied to each element of array and undefined is returned.
         */
        function forEachUnique(array, callback) {
            if (array) {
                for (var i = 0; i < array.length; i++) {
                    if (array.indexOf(array[i]) === i) {
                        var result = callback(array[i], i);
                        if (result) {
                            return result;
                        }
                    }
                }
            }
            return undefined;
        }
        ts.forEachUnique = forEachUnique;
        function isTextWhiteSpaceLike(text, startPos, endPos) {
            for (var i = startPos; i < endPos; i++) {
                if (!ts.isWhiteSpaceLike(text.charCodeAt(i))) {
                    return false;
                }
            }
            return true;
        }
        ts.isTextWhiteSpaceLike = isTextWhiteSpaceLike;
        // #endregion
        // Display-part writer helpers
        // #region
        function isFirstDeclarationOfSymbolParameter(symbol) {
            return symbol.declarations && symbol.declarations.length > 0 && symbol.declarations[0].kind === 162 /* Parameter */;
        }
        ts.isFirstDeclarationOfSymbolParameter = isFirstDeclarationOfSymbolParameter;
        var displayPartWriter = getDisplayPartWriter();
        function getDisplayPartWriter() {
            var absoluteMaximumLength = ts.defaultMaximumTruncationLength * 10; // A hard cutoff to avoid overloading the messaging channel in worst-case scenarios
            var displayParts;
            var lineStart;
            var indent;
            var length;
            resetWriter();
            var unknownWrite = function (text) { return writeKind(text, ts.SymbolDisplayPartKind.text); };
            return {
                displayParts: function () {
                    var finalText = displayParts.length && displayParts[displayParts.length - 1].text;
                    if (length > absoluteMaximumLength && finalText && finalText !== "...") {
                        if (!ts.isWhiteSpaceLike(finalText.charCodeAt(finalText.length - 1))) {
                            displayParts.push(displayPart(" ", ts.SymbolDisplayPartKind.space));
                        }
                        displayParts.push(displayPart("...", ts.SymbolDisplayPartKind.punctuation));
                    }
                    return displayParts;
                },
                writeKeyword: function (text) { return writeKind(text, ts.SymbolDisplayPartKind.keyword); },
                writeOperator: function (text) { return writeKind(text, ts.SymbolDisplayPartKind.operator); },
                writePunctuation: function (text) { return writeKind(text, ts.SymbolDisplayPartKind.punctuation); },
                writeTrailingSemicolon: function (text) { return writeKind(text, ts.SymbolDisplayPartKind.punctuation); },
                writeSpace: function (text) { return writeKind(text, ts.SymbolDisplayPartKind.space); },
                writeStringLiteral: function (text) { return writeKind(text, ts.SymbolDisplayPartKind.stringLiteral); },
                writeParameter: function (text) { return writeKind(text, ts.SymbolDisplayPartKind.parameterName); },
                writeProperty: function (text) { return writeKind(text, ts.SymbolDisplayPartKind.propertyName); },
                writeLiteral: function (text) { return writeKind(text, ts.SymbolDisplayPartKind.stringLiteral); },
                writeSymbol: writeSymbol,
                writeLine: writeLine,
                write: unknownWrite,
                writeComment: unknownWrite,
                getText: function () { return ""; },
                getTextPos: function () { return 0; },
                getColumn: function () { return 0; },
                getLine: function () { return 0; },
                isAtStartOfLine: function () { return false; },
                hasTrailingWhitespace: function () { return false; },
                hasTrailingComment: function () { return false; },
                rawWrite: ts.notImplemented,
                getIndent: function () { return indent; },
                increaseIndent: function () { indent++; },
                decreaseIndent: function () { indent--; },
                clear: resetWriter,
                trackSymbol: function () { return false; },
                reportInaccessibleThisError: ts.noop,
                reportInaccessibleUniqueSymbolError: ts.noop,
                reportPrivateInBaseOfClassExpression: ts.noop,
            };
            function writeIndent() {
                if (length > absoluteMaximumLength)
                    return;
                if (lineStart) {
                    var indentString = ts.getIndentString(indent);
                    if (indentString) {
                        length += indentString.length;
                        displayParts.push(displayPart(indentString, ts.SymbolDisplayPartKind.space));
                    }
                    lineStart = false;
                }
            }
            function writeKind(text, kind) {
                if (length > absoluteMaximumLength)
                    return;
                writeIndent();
                length += text.length;
                displayParts.push(displayPart(text, kind));
            }
            function writeSymbol(text, symbol) {
                if (length > absoluteMaximumLength)
                    return;
                writeIndent();
                length += text.length;
                displayParts.push(symbolPart(text, symbol));
            }
            function writeLine() {
                if (length > absoluteMaximumLength)
                    return;
                length += 1;
                displayParts.push(lineBreakPart());
                lineStart = true;
            }
            function resetWriter() {
                displayParts = [];
                lineStart = true;
                indent = 0;
                length = 0;
            }
        }
        function symbolPart(text, symbol) {
            return displayPart(text, displayPartKind(symbol));
            function displayPartKind(symbol) {
                var flags = symbol.flags;
                if (flags & 3 /* Variable */) {
                    return isFirstDeclarationOfSymbolParameter(symbol) ? ts.SymbolDisplayPartKind.parameterName : ts.SymbolDisplayPartKind.localName;
                }
                else if (flags & 4 /* Property */) {
                    return ts.SymbolDisplayPartKind.propertyName;
                }
                else if (flags & 32768 /* GetAccessor */) {
                    return ts.SymbolDisplayPartKind.propertyName;
                }
                else if (flags & 65536 /* SetAccessor */) {
                    return ts.SymbolDisplayPartKind.propertyName;
                }
                else if (flags & 8 /* EnumMember */) {
                    return ts.SymbolDisplayPartKind.enumMemberName;
                }
                else if (flags & 16 /* Function */) {
                    return ts.SymbolDisplayPartKind.functionName;
                }
                else if (flags & 32 /* Class */) {
                    return ts.SymbolDisplayPartKind.className;
                }
                else if (flags & 64 /* Interface */) {
                    return ts.SymbolDisplayPartKind.interfaceName;
                }
                else if (flags & 384 /* Enum */) {
                    return ts.SymbolDisplayPartKind.enumName;
                }
                else if (flags & 1536 /* Module */) {
                    return ts.SymbolDisplayPartKind.moduleName;
                }
                else if (flags & 8192 /* Method */) {
                    return ts.SymbolDisplayPartKind.methodName;
                }
                else if (flags & 262144 /* TypeParameter */) {
                    return ts.SymbolDisplayPartKind.typeParameterName;
                }
                else if (flags & 524288 /* TypeAlias */) {
                    return ts.SymbolDisplayPartKind.aliasName;
                }
                else if (flags & 2097152 /* Alias */) {
                    return ts.SymbolDisplayPartKind.aliasName;
                }
                return ts.SymbolDisplayPartKind.text;
            }
        }
        ts.symbolPart = symbolPart;
        function displayPart(text, kind) {
            return { text: text, kind: ts.SymbolDisplayPartKind[kind] };
        }
        ts.displayPart = displayPart;
        function spacePart() {
            return displayPart(" ", ts.SymbolDisplayPartKind.space);
        }
        ts.spacePart = spacePart;
        function keywordPart(kind) {
            return displayPart(ts.tokenToString(kind), ts.SymbolDisplayPartKind.keyword);
        }
        ts.keywordPart = keywordPart;
        function punctuationPart(kind) {
            return displayPart(ts.tokenToString(kind), ts.SymbolDisplayPartKind.punctuation);
        }
        ts.punctuationPart = punctuationPart;
        function operatorPart(kind) {
            return displayPart(ts.tokenToString(kind), ts.SymbolDisplayPartKind.operator);
        }
        ts.operatorPart = operatorPart;
        function parameterNamePart(text) {
            return displayPart(text, ts.SymbolDisplayPartKind.parameterName);
        }
        ts.parameterNamePart = parameterNamePart;
        function propertyNamePart(text) {
            return displayPart(text, ts.SymbolDisplayPartKind.propertyName);
        }
        ts.propertyNamePart = propertyNamePart;
        function textOrKeywordPart(text) {
            var kind = ts.stringToToken(text);
            return kind === undefined
                ? textPart(text)
                : keywordPart(kind);
        }
        ts.textOrKeywordPart = textOrKeywordPart;
        function textPart(text) {
            return displayPart(text, ts.SymbolDisplayPartKind.text);
        }
        ts.textPart = textPart;
        function typeAliasNamePart(text) {
            return displayPart(text, ts.SymbolDisplayPartKind.aliasName);
        }
        ts.typeAliasNamePart = typeAliasNamePart;
        function typeParameterNamePart(text) {
            return displayPart(text, ts.SymbolDisplayPartKind.typeParameterName);
        }
        ts.typeParameterNamePart = typeParameterNamePart;
        function linkTextPart(text) {
            return displayPart(text, ts.SymbolDisplayPartKind.linkText);
        }
        ts.linkTextPart = linkTextPart;
        function linkNamePart(name, target) {
            return {
                text: ts.getTextOfNode(name),
                kind: ts.SymbolDisplayPartKind[ts.SymbolDisplayPartKind.linkName],
                target: {
                    fileName: ts.getSourceFileOfNode(target).fileName,
                    textSpan: createTextSpanFromNode(target),
                },
            };
        }
        ts.linkNamePart = linkNamePart;
        function linkPart(text) {
            return displayPart(text, ts.SymbolDisplayPartKind.link);
        }
        ts.linkPart = linkPart;
        function buildLinkParts(link, checker) {
            var _a;
            var prefix = ts.isJSDocLink(link) ? "link"
                : ts.isJSDocLinkCode(link) ? "linkcode"
                    : "linkplain";
            var parts = [linkPart("{@" + prefix + " ")];
            if (!link.name) {
                if (link.text) {
                    parts.push(linkTextPart(link.text));
                }
            }
            else {
                var symbol = checker === null || checker === void 0 ? void 0 : checker.getSymbolAtLocation(link.name);
                var decl = (symbol === null || symbol === void 0 ? void 0 : symbol.valueDeclaration) || ((_a = symbol === null || symbol === void 0 ? void 0 : symbol.declarations) === null || _a === void 0 ? void 0 : _a[0]);
                if (decl) {
                    parts.push(linkNamePart(link.name, decl));
                    if (link.text) {
                        parts.push(linkTextPart(link.text));
                    }
                }
                else {
                    parts.push(linkTextPart(ts.getTextOfNode(link.name) + " " + link.text));
                }
            }
            parts.push(linkPart("}"));
            return parts;
        }
        ts.buildLinkParts = buildLinkParts;
        var carriageReturnLineFeed = "\r\n";
        /**
         * The default is CRLF.
         */
        function getNewLineOrDefaultFromHost(host, formatSettings) {
            var _a;
            return (formatSettings === null || formatSettings === void 0 ? void 0 : formatSettings.newLineCharacter) ||
                ((_a = host.getNewLine) === null || _a === void 0 ? void 0 : _a.call(host)) ||
                carriageReturnLineFeed;
        }
        ts.getNewLineOrDefaultFromHost = getNewLineOrDefaultFromHost;
        function lineBreakPart() {
            return displayPart("\n", ts.SymbolDisplayPartKind.lineBreak);
        }
        ts.lineBreakPart = lineBreakPart;
        function mapToDisplayParts(writeDisplayParts) {
            try {
                writeDisplayParts(displayPartWriter);
                return displayPartWriter.displayParts();
            }
            finally {
                displayPartWriter.clear();
            }
        }
        ts.mapToDisplayParts = mapToDisplayParts;
        function typeToDisplayParts(typechecker, type, enclosingDeclaration, flags) {
            if (flags === void 0) {
                flags = 0 /* None */;
            }
            return mapToDisplayParts(function (writer) {
                typechecker.writeType(type, enclosingDeclaration, flags | 1024 /* MultilineObjectLiterals */ | 16384 /* UseAliasDefinedOutsideCurrentScope */, writer);
            });
        }
        ts.typeToDisplayParts = typeToDisplayParts;
        function symbolToDisplayParts(typeChecker, symbol, enclosingDeclaration, meaning, flags) {
            if (flags === void 0) {
                flags = 0 /* None */;
            }
            return mapToDisplayParts(function (writer) {
                typeChecker.writeSymbol(symbol, enclosingDeclaration, meaning, flags | 8 /* UseAliasDefinedOutsideCurrentScope */, writer);
            });
        }
        ts.symbolToDisplayParts = symbolToDisplayParts;
        function signatureToDisplayParts(typechecker, signature, enclosingDeclaration, flags) {
            if (flags === void 0) {
                flags = 0 /* None */;
            }
            flags |= 16384 /* UseAliasDefinedOutsideCurrentScope */ | 1024 /* MultilineObjectLiterals */ | 32 /* WriteTypeArgumentsOfSignature */ | 8192 /* OmitParameterModifiers */;
            return mapToDisplayParts(function (writer) {
                typechecker.writeSignature(signature, enclosingDeclaration, flags, /*signatureKind*/ undefined, writer);
            });
        }
        ts.signatureToDisplayParts = signatureToDisplayParts;
        function isImportOrExportSpecifierName(location) {
            return !!location.parent && ts.isImportOrExportSpecifier(location.parent) && location.parent.propertyName === location;
        }
        ts.isImportOrExportSpecifierName = isImportOrExportSpecifierName;
        function getScriptKind(fileName, host) {
            // First check to see if the script kind was specified by the host. Chances are the host
            // may override the default script kind for the file extension.
            return ts.ensureScriptKind(fileName, host.getScriptKind && host.getScriptKind(fileName));
        }
        ts.getScriptKind = getScriptKind;
        function getSymbolTarget(symbol, checker) {
            var next = symbol;
            while (isAliasSymbol(next) || (isTransientSymbol(next) && next.target)) {
                if (isTransientSymbol(next) && next.target) {
                    next = next.target;
                }
                else {
                    next = ts.skipAlias(next, checker);
                }
            }
            return next;
        }
        ts.getSymbolTarget = getSymbolTarget;
        function isTransientSymbol(symbol) {
            return (symbol.flags & 33554432 /* Transient */) !== 0;
        }
        function isAliasSymbol(symbol) {
            return (symbol.flags & 2097152 /* Alias */) !== 0;
        }
        function getUniqueSymbolId(symbol, checker) {
            return ts.getSymbolId(ts.skipAlias(symbol, checker));
        }
        ts.getUniqueSymbolId = getUniqueSymbolId;
        function getFirstNonSpaceCharacterPosition(text, position) {
            while (ts.isWhiteSpaceLike(text.charCodeAt(position))) {
                position += 1;
            }
            return position;
        }
        ts.getFirstNonSpaceCharacterPosition = getFirstNonSpaceCharacterPosition;
        function getPrecedingNonSpaceCharacterPosition(text, position) {
            while (position > -1 && ts.isWhiteSpaceSingleLine(text.charCodeAt(position))) {
                position -= 1;
            }
            return position + 1;
        }
        ts.getPrecedingNonSpaceCharacterPosition = getPrecedingNonSpaceCharacterPosition;
        /**
         * Creates a deep, memberwise clone of a node with no source map location.
         *
         * WARNING: This is an expensive operation and is only intended to be used in refactorings
         * and code fixes (because those are triggered by explicit user actions).
         */
        function getSynthesizedDeepClone(node, includeTrivia) {
            if (includeTrivia === void 0) {
                includeTrivia = true;
            }
            var clone = node && getSynthesizedDeepCloneWorker(node);
            if (clone && !includeTrivia)
                suppressLeadingAndTrailingTrivia(clone);
            return clone;
        }
        ts.getSynthesizedDeepClone = getSynthesizedDeepClone;
        function getSynthesizedDeepCloneWithReplacements(node, includeTrivia, replaceNode) {
            var clone = replaceNode(node);
            if (clone) {
                ts.setOriginalNode(clone, node);
            }
            else {
                clone = getSynthesizedDeepCloneWorker(node, replaceNode);
            }
            if (clone && !includeTrivia)
                suppressLeadingAndTrailingTrivia(clone);
            return clone;
        }
        ts.getSynthesizedDeepCloneWithReplacements = getSynthesizedDeepCloneWithReplacements;
        function getSynthesizedDeepCloneWorker(node, replaceNode) {
            var nodeClone = replaceNode
                ? function (n) { return getSynthesizedDeepCloneWithReplacements(n, /*includeTrivia*/ true, replaceNode); }
                : getSynthesizedDeepClone;
            var nodesClone = replaceNode
                ? function (ns) { return ns && getSynthesizedDeepClonesWithReplacements(ns, /*includeTrivia*/ true, replaceNode); }
                : function (ns) { return ns && getSynthesizedDeepClones(ns); };
            var visited = ts.visitEachChild(node, nodeClone, ts.nullTransformationContext, nodesClone, nodeClone);
            if (visited === node) {
                // This only happens for leaf nodes - internal nodes always see their children change.
                var clone_1 = ts.isStringLiteral(node) ? ts.setOriginalNode(ts.factory.createStringLiteralFromNode(node), node) :
                    ts.isNumericLiteral(node) ? ts.setOriginalNode(ts.factory.createNumericLiteral(node.text, node.numericLiteralFlags), node) :
                        ts.factory.cloneNode(node);
                return ts.setTextRange(clone_1, node);
            }
            // PERF: As an optimization, rather than calling factory.cloneNode, we'll update
            // the new node created by visitEachChild with the extra changes factory.cloneNode
            // would have made.
            visited.parent = undefined;
            return visited;
        }
        function getSynthesizedDeepClones(nodes, includeTrivia) {
            if (includeTrivia === void 0) {
                includeTrivia = true;
            }
            return nodes && ts.factory.createNodeArray(nodes.map(function (n) { return getSynthesizedDeepClone(n, includeTrivia); }), nodes.hasTrailingComma);
        }
        ts.getSynthesizedDeepClones = getSynthesizedDeepClones;
        function getSynthesizedDeepClonesWithReplacements(nodes, includeTrivia, replaceNode) {
            return ts.factory.createNodeArray(nodes.map(function (n) { return getSynthesizedDeepCloneWithReplacements(n, includeTrivia, replaceNode); }), nodes.hasTrailingComma);
        }
        ts.getSynthesizedDeepClonesWithReplacements = getSynthesizedDeepClonesWithReplacements;
        /**
         * Sets EmitFlags to suppress leading and trailing trivia on the node.
         */
        function suppressLeadingAndTrailingTrivia(node) {
            suppressLeadingTrivia(node);
            suppressTrailingTrivia(node);
        }
        ts.suppressLeadingAndTrailingTrivia = suppressLeadingAndTrailingTrivia;
        /**
         * Sets EmitFlags to suppress leading trivia on the node.
         */
        function suppressLeadingTrivia(node) {
            addEmitFlagsRecursively(node, 512 /* NoLeadingComments */, getFirstChild);
        }
        ts.suppressLeadingTrivia = suppressLeadingTrivia;
        /**
         * Sets EmitFlags to suppress trailing trivia on the node.
         */
        function suppressTrailingTrivia(node) {
            addEmitFlagsRecursively(node, 1024 /* NoTrailingComments */, ts.getLastChild);
        }
        ts.suppressTrailingTrivia = suppressTrailingTrivia;
        function copyComments(sourceNode, targetNode) {
            var sourceFile = sourceNode.getSourceFile();
            var text = sourceFile.text;
            if (hasLeadingLineBreak(sourceNode, text)) {
                copyLeadingComments(sourceNode, targetNode, sourceFile);
            }
            else {
                copyTrailingAsLeadingComments(sourceNode, targetNode, sourceFile);
            }
            copyTrailingComments(sourceNode, targetNode, sourceFile);
        }
        ts.copyComments = copyComments;
        function hasLeadingLineBreak(node, text) {
            var start = node.getFullStart();
            var end = node.getStart();
            for (var i = start; i < end; i++) {
                if (text.charCodeAt(i) === 10 /* lineFeed */)
                    return true;
            }
            return false;
        }
        function addEmitFlagsRecursively(node, flag, getChild) {
            ts.addEmitFlags(node, flag);
            var child = getChild(node);
            if (child)
                addEmitFlagsRecursively(child, flag, getChild);
        }
        function getFirstChild(node) {
            return node.forEachChild(function (child) { return child; });
        }
        function getUniqueName(baseName, sourceFile) {
            var nameText = baseName;
            for (var i = 1; !ts.isFileLevelUniqueName(sourceFile, nameText); i++) {
                nameText = baseName + "_" + i;
            }
            return nameText;
        }
        ts.getUniqueName = getUniqueName;
        /**
         * @return The index of the (only) reference to the extracted symbol.  We want the cursor
         * to be on the reference, rather than the declaration, because it's closer to where the
         * user was before extracting it.
         */
        function getRenameLocation(edits, renameFilename, name, preferLastLocation) {
            var delta = 0;
            var lastPos = -1;
            for (var _i = 0, edits_1 = edits; _i < edits_1.length; _i++) {
                var _a = edits_1[_i], fileName = _a.fileName, textChanges_2 = _a.textChanges;
                ts.Debug.assert(fileName === renameFilename);
                for (var _b = 0, textChanges_1 = textChanges_2; _b < textChanges_1.length; _b++) {
                    var change = textChanges_1[_b];
                    var span = change.span, newText = change.newText;
                    var index = indexInTextChange(newText, name);
                    if (index !== -1) {
                        lastPos = span.start + delta + index;
                        // If the reference comes first, return immediately.
                        if (!preferLastLocation) {
                            return lastPos;
                        }
                    }
                    delta += newText.length - span.length;
                }
            }
            // If the declaration comes first, return the position of the last occurrence.
            ts.Debug.assert(preferLastLocation);
            ts.Debug.assert(lastPos >= 0);
            return lastPos;
        }
        ts.getRenameLocation = getRenameLocation;
        function copyLeadingComments(sourceNode, targetNode, sourceFile, commentKind, hasTrailingNewLine) {
            ts.forEachLeadingCommentRange(sourceFile.text, sourceNode.pos, getAddCommentsFunction(targetNode, sourceFile, commentKind, hasTrailingNewLine, ts.addSyntheticLeadingComment));
        }
        ts.copyLeadingComments = copyLeadingComments;
        function copyTrailingComments(sourceNode, targetNode, sourceFile, commentKind, hasTrailingNewLine) {
            ts.forEachTrailingCommentRange(sourceFile.text, sourceNode.end, getAddCommentsFunction(targetNode, sourceFile, commentKind, hasTrailingNewLine, ts.addSyntheticTrailingComment));
        }
        ts.copyTrailingComments = copyTrailingComments;
        /**
         * This function copies the trailing comments for the token that comes before `sourceNode`, as leading comments of `targetNode`.
         * This is useful because sometimes a comment that refers to `sourceNode` will be a leading comment for `sourceNode`, according to the
         * notion of trivia ownership, and instead will be a trailing comment for the token before `sourceNode`, e.g.:
         * `function foo(\* not leading comment for a *\ a: string) {}`
         * The comment refers to `a` but belongs to the `(` token, but we might want to copy it.
         */
        function copyTrailingAsLeadingComments(sourceNode, targetNode, sourceFile, commentKind, hasTrailingNewLine) {
            ts.forEachTrailingCommentRange(sourceFile.text, sourceNode.pos, getAddCommentsFunction(targetNode, sourceFile, commentKind, hasTrailingNewLine, ts.addSyntheticLeadingComment));
        }
        ts.copyTrailingAsLeadingComments = copyTrailingAsLeadingComments;
        function getAddCommentsFunction(targetNode, sourceFile, commentKind, hasTrailingNewLine, cb) {
            return function (pos, end, kind, htnl) {
                if (kind === 3 /* MultiLineCommentTrivia */) {
                    // Remove leading /*
                    pos += 2;
                    // Remove trailing */
                    end -= 2;
                }
                else {
                    // Remove leading //
                    pos += 2;
                }
                cb(targetNode, commentKind || kind, sourceFile.text.slice(pos, end), hasTrailingNewLine !== undefined ? hasTrailingNewLine : htnl);
            };
        }
        function indexInTextChange(change, name) {
            if (ts.startsWith(change, name))
                return 0;
            // Add a " " to avoid references inside words
            var idx = change.indexOf(" " + name);
            if (idx === -1)
                idx = change.indexOf("." + name);
            if (idx === -1)
                idx = change.indexOf('"' + name);
            return idx === -1 ? -1 : idx + 1;
        }
        /* @internal */
        function needsParentheses(expression) {
            return ts.isBinaryExpression(expression) && expression.operatorToken.kind === 27 /* CommaToken */
                || ts.isObjectLiteralExpression(expression)
                || ts.isAsExpression(expression) && ts.isObjectLiteralExpression(expression.expression);
        }
        ts.needsParentheses = needsParentheses;
        function getContextualTypeFromParent(node, checker) {
            var parent = node.parent;
            switch (parent.kind) {
                case 207 /* NewExpression */:
                    return checker.getContextualType(parent);
                case 219 /* BinaryExpression */: {
                    var _a = parent, left = _a.left, operatorToken = _a.operatorToken, right = _a.right;
                    return isEqualityOperatorKind(operatorToken.kind)
                        ? checker.getTypeAtLocation(node === right ? left : right)
                        : checker.getContextualType(node);
                }
                case 287 /* CaseClause */:
                    return parent.expression === node ? getSwitchedType(parent, checker) : undefined;
                default:
                    return checker.getContextualType(node);
            }
        }
        ts.getContextualTypeFromParent = getContextualTypeFromParent;
        function quote(sourceFile, preferences, text) {
            // Editors can pass in undefined or empty string - we want to infer the preference in those cases.
            var quotePreference = getQuotePreference(sourceFile, preferences);
            var quoted = JSON.stringify(text);
            return quotePreference === 0 /* Single */ ? "'" + ts.stripQuotes(quoted).replace(/'/g, "\\'").replace(/\\"/g, '"') + "'" : quoted;
        }
        ts.quote = quote;
        function isEqualityOperatorKind(kind) {
            switch (kind) {
                case 36 /* EqualsEqualsEqualsToken */:
                case 34 /* EqualsEqualsToken */:
                case 37 /* ExclamationEqualsEqualsToken */:
                case 35 /* ExclamationEqualsToken */:
                    return true;
                default:
                    return false;
            }
        }
        ts.isEqualityOperatorKind = isEqualityOperatorKind;
        function isStringLiteralOrTemplate(node) {
            switch (node.kind) {
                case 10 /* StringLiteral */:
                case 14 /* NoSubstitutionTemplateLiteral */:
                case 221 /* TemplateExpression */:
                case 208 /* TaggedTemplateExpression */:
                    return true;
                default:
                    return false;
            }
        }
        ts.isStringLiteralOrTemplate = isStringLiteralOrTemplate;
        function hasIndexSignature(type) {
            return !!type.getStringIndexType() || !!type.getNumberIndexType();
        }
        ts.hasIndexSignature = hasIndexSignature;
        function getSwitchedType(caseClause, checker) {
            return checker.getTypeAtLocation(caseClause.parent.parent.expression);
        }
        ts.getSwitchedType = getSwitchedType;
        ts.ANONYMOUS = "anonymous function";
        function getTypeNodeIfAccessible(type, enclosingScope, program, host) {
            var checker = program.getTypeChecker();
            var typeIsAccessible = true;
            var notAccessible = function () { typeIsAccessible = false; };
            var res = checker.typeToTypeNode(type, enclosingScope, 1 /* NoTruncation */, {
                trackSymbol: function (symbol, declaration, meaning) {
                    typeIsAccessible = typeIsAccessible && checker.isSymbolAccessible(symbol, declaration, meaning, /*shouldComputeAliasToMarkVisible*/ false).accessibility === 0 /* Accessible */;
                    return !typeIsAccessible;
                },
                reportInaccessibleThisError: notAccessible,
                reportPrivateInBaseOfClassExpression: notAccessible,
                reportInaccessibleUniqueSymbolError: notAccessible,
                moduleResolverHost: getModuleSpecifierResolverHost(program, host)
            });
            return typeIsAccessible ? res : undefined;
        }
        ts.getTypeNodeIfAccessible = getTypeNodeIfAccessible;
        function syntaxRequiresTrailingCommaOrSemicolonOrASI(kind) {
            return kind === 172 /* CallSignature */
                || kind === 173 /* ConstructSignature */
                || kind === 174 /* IndexSignature */
                || kind === 164 /* PropertySignature */
                || kind === 166 /* MethodSignature */;
        }
        ts.syntaxRequiresTrailingCommaOrSemicolonOrASI = syntaxRequiresTrailingCommaOrSemicolonOrASI;
        function syntaxRequiresTrailingFunctionBlockOrSemicolonOrASI(kind) {
            return kind === 254 /* FunctionDeclaration */
                || kind === 169 /* Constructor */
                || kind === 167 /* MethodDeclaration */
                || kind === 170 /* GetAccessor */
                || kind === 171 /* SetAccessor */;
        }
        ts.syntaxRequiresTrailingFunctionBlockOrSemicolonOrASI = syntaxRequiresTrailingFunctionBlockOrSemicolonOrASI;
        function syntaxRequiresTrailingModuleBlockOrSemicolonOrASI(kind) {
            return kind === 259 /* ModuleDeclaration */;
        }
        ts.syntaxRequiresTrailingModuleBlockOrSemicolonOrASI = syntaxRequiresTrailingModuleBlockOrSemicolonOrASI;
        function syntaxRequiresTrailingSemicolonOrASI(kind) {
            return kind === 235 /* VariableStatement */
                || kind === 236 /* ExpressionStatement */
                || kind === 238 /* DoStatement */
                || kind === 243 /* ContinueStatement */
                || kind === 244 /* BreakStatement */
                || kind === 245 /* ReturnStatement */
                || kind === 249 /* ThrowStatement */
                || kind === 251 /* DebuggerStatement */
                || kind === 165 /* PropertyDeclaration */
                || kind === 257 /* TypeAliasDeclaration */
                || kind === 264 /* ImportDeclaration */
                || kind === 263 /* ImportEqualsDeclaration */
                || kind === 270 /* ExportDeclaration */
                || kind === 262 /* NamespaceExportDeclaration */
                || kind === 269 /* ExportAssignment */;
        }
        ts.syntaxRequiresTrailingSemicolonOrASI = syntaxRequiresTrailingSemicolonOrASI;
        ts.syntaxMayBeASICandidate = ts.or(syntaxRequiresTrailingCommaOrSemicolonOrASI, syntaxRequiresTrailingFunctionBlockOrSemicolonOrASI, syntaxRequiresTrailingModuleBlockOrSemicolonOrASI, syntaxRequiresTrailingSemicolonOrASI);
        function nodeIsASICandidate(node, sourceFile) {
            var lastToken = node.getLastToken(sourceFile);
            if (lastToken && lastToken.kind === 26 /* SemicolonToken */) {
                return false;
            }
            if (syntaxRequiresTrailingCommaOrSemicolonOrASI(node.kind)) {
                if (lastToken && lastToken.kind === 27 /* CommaToken */) {
                    return false;
                }
            }
            else if (syntaxRequiresTrailingModuleBlockOrSemicolonOrASI(node.kind)) {
                var lastChild = ts.last(node.getChildren(sourceFile));
                if (lastChild && ts.isModuleBlock(lastChild)) {
                    return false;
                }
            }
            else if (syntaxRequiresTrailingFunctionBlockOrSemicolonOrASI(node.kind)) {
                var lastChild = ts.last(node.getChildren(sourceFile));
                if (lastChild && ts.isFunctionBlock(lastChild)) {
                    return false;
                }
            }
            else if (!syntaxRequiresTrailingSemicolonOrASI(node.kind)) {
                return false;
            }
            // See comment in parser’s `parseDoStatement`
            if (node.kind === 238 /* DoStatement */) {
                return true;
            }
            var topNode = ts.findAncestor(node, function (ancestor) { return !ancestor.parent; });
            var nextToken = findNextToken(node, topNode, sourceFile);
            if (!nextToken || nextToken.kind === 19 /* CloseBraceToken */) {
                return true;
            }
            var startLine = sourceFile.getLineAndCharacterOfPosition(node.getEnd()).line;
            var endLine = sourceFile.getLineAndCharacterOfPosition(nextToken.getStart(sourceFile)).line;
            return startLine !== endLine;
        }
        function positionIsASICandidate(pos, context, sourceFile) {
            var contextAncestor = ts.findAncestor(context, function (ancestor) {
                if (ancestor.end !== pos) {
                    return "quit";
                }
                return ts.syntaxMayBeASICandidate(ancestor.kind);
            });
            return !!contextAncestor && nodeIsASICandidate(contextAncestor, sourceFile);
        }
        ts.positionIsASICandidate = positionIsASICandidate;
        function probablyUsesSemicolons(sourceFile) {
            var withSemicolon = 0;
            var withoutSemicolon = 0;
            var nStatementsToObserve = 5;
            ts.forEachChild(sourceFile, function visit(node) {
                if (syntaxRequiresTrailingSemicolonOrASI(node.kind)) {
                    var lastToken = node.getLastToken(sourceFile);
                    if (lastToken && lastToken.kind === 26 /* SemicolonToken */) {
                        withSemicolon++;
                    }
                    else {
                        withoutSemicolon++;
                    }
                }
                if (withSemicolon + withoutSemicolon >= nStatementsToObserve) {
                    return true;
                }
                return ts.forEachChild(node, visit);
            });
            // One statement missing a semicolon isn’t sufficient evidence to say the user
            // doesn’t want semicolons, because they may not even be done writing that statement.
            if (withSemicolon === 0 && withoutSemicolon <= 1) {
                return true;
            }
            // If even 2/5 places have a semicolon, the user probably wants semicolons
            return withSemicolon / withoutSemicolon > 1 / nStatementsToObserve;
        }
        ts.probablyUsesSemicolons = probablyUsesSemicolons;
        function tryGetDirectories(host, directoryName) {
            return tryIOAndConsumeErrors(host, host.getDirectories, directoryName) || [];
        }
        ts.tryGetDirectories = tryGetDirectories;
        function tryReadDirectory(host, path, extensions, exclude, include) {
            return tryIOAndConsumeErrors(host, host.readDirectory, path, extensions, exclude, include) || ts.emptyArray;
        }
        ts.tryReadDirectory = tryReadDirectory;
        function tryFileExists(host, path) {
            return tryIOAndConsumeErrors(host, host.fileExists, path);
        }
        ts.tryFileExists = tryFileExists;
        function tryDirectoryExists(host, path) {
            return tryAndIgnoreErrors(function () { return ts.directoryProbablyExists(path, host); }) || false;
        }
        ts.tryDirectoryExists = tryDirectoryExists;
        function tryAndIgnoreErrors(cb) {
            try {
                return cb();
            }
            catch (_a) {
                return undefined;
            }
        }
        ts.tryAndIgnoreErrors = tryAndIgnoreErrors;
        function tryIOAndConsumeErrors(host, toApply) {
            var args = [];
            for (var _i = 2; _i < arguments.length; _i++) {
                args[_i - 2] = arguments[_i];
            }
            return tryAndIgnoreErrors(function () { return toApply && toApply.apply(host, args); });
        }
        ts.tryIOAndConsumeErrors = tryIOAndConsumeErrors;
        function findPackageJsons(startDirectory, host, stopDirectory) {
            var paths = [];
            ts.forEachAncestorDirectory(startDirectory, function (ancestor) {
                if (ancestor === stopDirectory) {
                    return true;
                }
                var currentConfigPath = ts.combinePaths(ancestor, "package.json");
                if (tryFileExists(host, currentConfigPath)) {
                    paths.push(currentConfigPath);
                }
            });
            return paths;
        }
        ts.findPackageJsons = findPackageJsons;
        function findPackageJson(directory, host) {
            var packageJson;
            ts.forEachAncestorDirectory(directory, function (ancestor) {
                if (ancestor === "node_modules")
                    return true;
                packageJson = ts.findConfigFile(ancestor, function (f) { return tryFileExists(host, f); }, "package.json");
                if (packageJson) {
                    return true; // break out
                }
            });
            return packageJson;
        }
        ts.findPackageJson = findPackageJson;
        function getPackageJsonsVisibleToFile(fileName, host) {
            if (!host.fileExists) {
                return [];
            }
            var packageJsons = [];
            ts.forEachAncestorDirectory(ts.getDirectoryPath(fileName), function (ancestor) {
                var packageJsonFileName = ts.combinePaths(ancestor, "package.json");
                if (host.fileExists(packageJsonFileName)) {
                    var info = createPackageJsonInfo(packageJsonFileName, host);
                    if (info) {
                        packageJsons.push(info);
                    }
                }
            });
            return packageJsons;
        }
        ts.getPackageJsonsVisibleToFile = getPackageJsonsVisibleToFile;
        function createPackageJsonInfo(fileName, host) {
            if (!host.readFile) {
                return undefined;
            }
            var dependencyKeys = ["dependencies", "devDependencies", "optionalDependencies", "peerDependencies"];
            var stringContent = host.readFile(fileName) || "";
            var content = tryParseJson(stringContent);
            var info = {};
            if (content) {
                for (var _i = 0, dependencyKeys_1 = dependencyKeys; _i < dependencyKeys_1.length; _i++) {
                    var key = dependencyKeys_1[_i];
                    var dependencies = content[key];
                    if (!dependencies) {
                        continue;
                    }
                    var dependencyMap = new ts.Map();
                    for (var packageName in dependencies) {
                        dependencyMap.set(packageName, dependencies[packageName]);
                    }
                    info[key] = dependencyMap;
                }
            }
            var dependencyGroups = [
                [1 /* Dependencies */, info.dependencies],
                [2 /* DevDependencies */, info.devDependencies],
                [8 /* OptionalDependencies */, info.optionalDependencies],
                [4 /* PeerDependencies */, info.peerDependencies],
            ];
            return __assign(__assign({}, info), { parseable: !!content, fileName: fileName, get: get, has: function (dependencyName, inGroups) {
                    return !!get(dependencyName, inGroups);
                } });
            function get(dependencyName, inGroups) {
                if (inGroups === void 0) {
                    inGroups = 15 /* All */;
                }
                for (var _i = 0, dependencyGroups_1 = dependencyGroups; _i < dependencyGroups_1.length; _i++) {
                    var _a = dependencyGroups_1[_i], group_1 = _a[0], deps = _a[1];
                    if (deps && (inGroups & group_1)) {
                        var dep = deps.get(dependencyName);
                        if (dep !== undefined) {
                            return dep;
                        }
                    }
                }
            }
        }
        ts.createPackageJsonInfo = createPackageJsonInfo;
        function createPackageJsonImportFilter(fromFile, preferences, host) {
            var packageJsons = ((host.getPackageJsonsVisibleToFile && host.getPackageJsonsVisibleToFile(fromFile.fileName)) || getPackageJsonsVisibleToFile(fromFile.fileName, host)).filter(function (p) { return p.parseable; });
            var usesNodeCoreModules;
            return { allowsImportingAmbientModule: allowsImportingAmbientModule, allowsImportingSourceFile: allowsImportingSourceFile, allowsImportingSpecifier: allowsImportingSpecifier };
            function moduleSpecifierIsCoveredByPackageJson(specifier) {
                var packageName = getNodeModuleRootSpecifier(specifier);
                for (var _i = 0, packageJsons_1 = packageJsons; _i < packageJsons_1.length; _i++) {
                    var packageJson = packageJsons_1[_i];
                    if (packageJson.has(packageName) || packageJson.has(ts.getTypesPackageName(packageName))) {
                        return true;
                    }
                }
                return false;
            }
            function allowsImportingAmbientModule(moduleSymbol, moduleSpecifierResolutionHost) {
                if (!packageJsons.length || !moduleSymbol.valueDeclaration) {
                    return true;
                }
                var declaringSourceFile = moduleSymbol.valueDeclaration.getSourceFile();
                var declaringNodeModuleName = getNodeModulesPackageNameFromFileName(declaringSourceFile.fileName, moduleSpecifierResolutionHost);
                if (typeof declaringNodeModuleName === "undefined") {
                    return true;
                }
                var declaredModuleSpecifier = ts.stripQuotes(moduleSymbol.getName());
                if (isAllowedCoreNodeModulesImport(declaredModuleSpecifier)) {
                    return true;
                }
                return moduleSpecifierIsCoveredByPackageJson(declaringNodeModuleName)
                    || moduleSpecifierIsCoveredByPackageJson(declaredModuleSpecifier);
            }
            function allowsImportingSourceFile(sourceFile, moduleSpecifierResolutionHost) {
                if (!packageJsons.length) {
                    return true;
                }
                var moduleSpecifier = getNodeModulesPackageNameFromFileName(sourceFile.fileName, moduleSpecifierResolutionHost);
                if (!moduleSpecifier) {
                    return true;
                }
                return moduleSpecifierIsCoveredByPackageJson(moduleSpecifier);
            }
            function allowsImportingSpecifier(moduleSpecifier) {
                if (!packageJsons.length || isAllowedCoreNodeModulesImport(moduleSpecifier)) {
                    return true;
                }
                if (ts.pathIsRelative(moduleSpecifier) || ts.isRootedDiskPath(moduleSpecifier)) {
                    return true;
                }
                return moduleSpecifierIsCoveredByPackageJson(moduleSpecifier);
            }
            function isAllowedCoreNodeModulesImport(moduleSpecifier) {
                // If we’re in JavaScript, it can be difficult to tell whether the user wants to import
                // from Node core modules or not. We can start by seeing if the user is actually using
                // any node core modules, as opposed to simply having @types/node accidentally as a
                // dependency of a dependency.
                if (ts.isSourceFileJS(fromFile) && ts.JsTyping.nodeCoreModules.has(moduleSpecifier)) {
                    if (usesNodeCoreModules === undefined) {
                        usesNodeCoreModules = consumesNodeCoreModules(fromFile);
                    }
                    if (usesNodeCoreModules) {
                        return true;
                    }
                }
                return false;
            }
            function getNodeModulesPackageNameFromFileName(importedFileName, moduleSpecifierResolutionHost) {
                if (!ts.stringContains(importedFileName, "node_modules")) {
                    return undefined;
                }
                var specifier = ts.moduleSpecifiers.getNodeModulesPackageName(host.getCompilationSettings(), fromFile.path, importedFileName, moduleSpecifierResolutionHost, preferences);
                if (!specifier) {
                    return undefined;
                }
                // Paths here are not node_modules, so we don’t care about them;
                // returning anything will trigger a lookup in package.json.
                if (!ts.pathIsRelative(specifier) && !ts.isRootedDiskPath(specifier)) {
                    return getNodeModuleRootSpecifier(specifier);
                }
            }
            function getNodeModuleRootSpecifier(fullSpecifier) {
                var components = ts.getPathComponents(ts.getPackageNameFromTypesPackageName(fullSpecifier)).slice(1);
                // Scoped packages
                if (ts.startsWith(components[0], "@")) {
                    return components[0] + "/" + components[1];
                }
                return components[0];
            }
        }
        ts.createPackageJsonImportFilter = createPackageJsonImportFilter;
        function tryParseJson(text) {
            try {
                return JSON.parse(text);
            }
            catch (_a) {
                return undefined;
            }
        }
        function consumesNodeCoreModules(sourceFile) {
            return ts.some(sourceFile.imports, function (_a) {
                var text = _a.text;
                return ts.JsTyping.nodeCoreModules.has(text);
            });
        }
        ts.consumesNodeCoreModules = consumesNodeCoreModules;
        function isInsideNodeModules(fileOrDirectory) {
            return ts.contains(ts.getPathComponents(fileOrDirectory), "node_modules");
        }
        ts.isInsideNodeModules = isInsideNodeModules;
        function isDiagnosticWithLocation(diagnostic) {
            return diagnostic.file !== undefined && diagnostic.start !== undefined && diagnostic.length !== undefined;
        }
        ts.isDiagnosticWithLocation = isDiagnosticWithLocation;
        function findDiagnosticForNode(node, sortedFileDiagnostics) {
            var span = createTextSpanFromNode(node);
            var index = ts.binarySearchKey(sortedFileDiagnostics, span, ts.identity, ts.compareTextSpans);
            if (index >= 0) {
                var diagnostic = sortedFileDiagnostics[index];
                ts.Debug.assertEqual(diagnostic.file, node.getSourceFile(), "Diagnostics proided to 'findDiagnosticForNode' must be from a single SourceFile");
                return ts.cast(diagnostic, isDiagnosticWithLocation);
            }
        }
        ts.findDiagnosticForNode = findDiagnosticForNode;
        function getDiagnosticsWithinSpan(span, sortedFileDiagnostics) {
            var _a;
            var index = ts.binarySearchKey(sortedFileDiagnostics, span.start, function (diag) { return diag.start; }, ts.compareValues);
            if (index < 0) {
                index = ~index;
            }
            while (((_a = sortedFileDiagnostics[index - 1]) === null || _a === void 0 ? void 0 : _a.start) === span.start) {
                index--;
            }
            var result = [];
            var end = ts.textSpanEnd(span);
            while (true) {
                var diagnostic = ts.tryCast(sortedFileDiagnostics[index], isDiagnosticWithLocation);
                if (!diagnostic || diagnostic.start > end) {
                    break;
                }
                if (ts.textSpanContainsTextSpan(span, diagnostic)) {
                    result.push(diagnostic);
                }
                index++;
            }
            return result;
        }
        ts.getDiagnosticsWithinSpan = getDiagnosticsWithinSpan;
        /* @internal */
        function getRefactorContextSpan(_a) {
            var startPosition = _a.startPosition, endPosition = _a.endPosition;
            return ts.createTextSpanFromBounds(startPosition, endPosition === undefined ? startPosition : endPosition);
        }
        ts.getRefactorContextSpan = getRefactorContextSpan;
        function mapOneOrMany(valueOrArray, f, resultSelector) {
            if (resultSelector === void 0) {
                resultSelector = ts.identity;
            }
            return valueOrArray ? ts.isArray(valueOrArray) ? resultSelector(ts.map(valueOrArray, f)) : f(valueOrArray, 0) : undefined;
        }
        ts.mapOneOrMany = mapOneOrMany;
        /**
         * If the provided value is an array, the first element of the array is returned; otherwise, the provided value is returned instead.
         */
        function firstOrOnly(valueOrArray) {
            return ts.isArray(valueOrArray) ? ts.first(valueOrArray) : valueOrArray;
        }
        ts.firstOrOnly = firstOrOnly;
        function getNameForExportedSymbol(symbol, scriptTarget) {
            if (!(symbol.flags & 33554432 /* Transient */) && (symbol.escapedName === "export=" /* ExportEquals */ || symbol.escapedName === "default" /* Default */)) {
                // Name of "export default foo;" is "foo". Name of "export default 0" is the filename converted to camelCase.
                return ts.firstDefined(symbol.declarations, function (d) { var _a; return ts.isExportAssignment(d) ? (_a = ts.tryCast(ts.skipOuterExpressions(d.expression), ts.isIdentifier)) === null || _a === void 0 ? void 0 : _a.text : undefined; })
                    || ts.codefix.moduleSymbolToValidIdentifier(getSymbolParentOrFail(symbol), scriptTarget);
            }
            return symbol.name;
        }
        ts.getNameForExportedSymbol = getNameForExportedSymbol;
        function getSymbolParentOrFail(symbol) {
            var _a;
            return ts.Debug.checkDefined(symbol.parent, "Symbol parent was undefined. Flags: " + ts.Debug.formatSymbolFlags(symbol.flags) + ". " +
                ("Declarations: " + ((_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.map(function (d) {
                    var kind = ts.Debug.formatSyntaxKind(d.kind);
                    var inJS = ts.isInJSFile(d);
                    var expression = d.expression;
                    return (inJS ? "[JS]" : "") + kind + (expression ? " (expression: " + ts.Debug.formatSyntaxKind(expression.kind) + ")" : "");
                }).join(", ")) + "."));
        }
        /**
         * Useful to check whether a string contains another string at a specific index
         * without allocating another string or traversing the entire contents of the outer string.
         *
         * This function is useful in place of either of the following:
         *
         * ```ts
         * // Allocates
         * haystack.substr(startIndex, needle.length) === needle
         *
         * // Full traversal
         * haystack.indexOf(needle, startIndex) === startIndex
         * ```
         *
         * @param haystack The string that potentially contains `needle`.
         * @param needle The string whose content might sit within `haystack`.
         * @param startIndex The index within `haystack` to start searching for `needle`.
         */
        function stringContainsAt(haystack, needle, startIndex) {
            var needleLength = needle.length;
            if (needleLength + startIndex > haystack.length) {
                return false;
            }
            for (var i = 0; i < needleLength; i++) {
                if (needle.charCodeAt(i) !== haystack.charCodeAt(i + startIndex))
                    return false;
            }
            return true;
        }
        ts.stringContainsAt = stringContainsAt;
        function startsWithUnderscore(name) {
            return name.charCodeAt(0) === 95 /* _ */;
        }
        ts.startsWithUnderscore = startsWithUnderscore;
        function isGlobalDeclaration(declaration) {
            return !isNonGlobalDeclaration(declaration);
        }
        ts.isGlobalDeclaration = isGlobalDeclaration;
        function isNonGlobalDeclaration(declaration) {
            var sourceFile = declaration.getSourceFile();
            // If the file is not a module, the declaration is global
            if (!sourceFile.externalModuleIndicator && !sourceFile.commonJsModuleIndicator) {
                return false;
            }
            // If the file is a module written in TypeScript, it still might be in a `declare global` augmentation
            return ts.isInJSFile(declaration) || !ts.findAncestor(declaration, ts.isGlobalScopeAugmentation);
        }
        ts.isNonGlobalDeclaration = isNonGlobalDeclaration;
        function isDeprecatedDeclaration(decl) {
            return !!(ts.getCombinedNodeFlagsAlwaysIncludeJSDoc(decl) & 8192 /* Deprecated */);
        }
        ts.isDeprecatedDeclaration = isDeprecatedDeclaration;
        function shouldUseUriStyleNodeCoreModules(file, program) {
            var decisionFromFile = ts.firstDefined(file.imports, function (node) {
                if (ts.JsTyping.nodeCoreModules.has(node.text)) {
                    return ts.startsWith(node.text, "node:");
                }
            });
            return decisionFromFile !== null && decisionFromFile !== void 0 ? decisionFromFile : program.usesUriStyleNodeCoreModules;
        }
        ts.shouldUseUriStyleNodeCoreModules = shouldUseUriStyleNodeCoreModules;
        // #endregion
    })(ts || (ts = {}));
    /*@internal*/
    var ts;
    (function (ts) {
        var ImportKind;
        (function (ImportKind) {
            ImportKind[ImportKind["Named"] = 0] = "Named";
            ImportKind[ImportKind["Default"] = 1] = "Default";
            ImportKind[ImportKind["Namespace"] = 2] = "Namespace";
            ImportKind[ImportKind["CommonJS"] = 3] = "CommonJS";
        })(ImportKind = ts.ImportKind || (ts.ImportKind = {}));
        var ExportKind;
        (function (ExportKind) {
            ExportKind[ExportKind["Named"] = 0] = "Named";
            ExportKind[ExportKind["Default"] = 1] = "Default";
            ExportKind[ExportKind["ExportEquals"] = 2] = "ExportEquals";
            ExportKind[ExportKind["UMD"] = 3] = "UMD";
        })(ExportKind = ts.ExportKind || (ts.ExportKind = {}));
        function createCacheableExportInfoMap(host) {
            var exportInfoId = 1;
            var exportInfo = ts.createMultiMap();
            var symbols = new ts.Map();
            var usableByFileName;
            var cache = {
                isUsableByFile: function (importingFile) { return importingFile === usableByFileName; },
                isEmpty: function () { return !exportInfo.size; },
                clear: function () {
                    exportInfo.clear();
                    symbols.clear();
                    usableByFileName = undefined;
                },
                add: function (importingFile, symbol, symbolTableKey, moduleSymbol, moduleFile, exportKind, isFromPackageJson, scriptTarget, checker) {
                    if (importingFile !== usableByFileName) {
                        cache.clear();
                        usableByFileName = importingFile;
                    }
                    var isDefault = exportKind === 1 /* Default */;
                    var namedSymbol = isDefault && ts.getLocalSymbolForExportDefault(symbol) || symbol;
                    // A re-export merged with an export from a module augmentation can result in `symbol`
                    // being an external module symbol; the name it is re-exported by will be `symbolTableKey`
                    // (which comes from the keys of `moduleSymbol.exports`.)
                    var importedName = ts.isExternalModuleSymbol(namedSymbol)
                        ? ts.unescapeLeadingUnderscores(symbolTableKey)
                        : ts.getNameForExportedSymbol(namedSymbol, scriptTarget);
                    var moduleName = ts.stripQuotes(moduleSymbol.name);
                    var id = exportInfoId++;
                    var storedSymbol = symbol.flags & 33554432 /* Transient */ ? undefined : symbol;
                    var storedModuleSymbol = moduleSymbol.flags & 33554432 /* Transient */ ? undefined : moduleSymbol;
                    if (!storedSymbol || !storedModuleSymbol)
                        symbols.set(id, [symbol, moduleSymbol]);
                    exportInfo.add(key(importedName, symbol, moduleName, checker), {
                        id: id,
                        symbolTableKey: symbolTableKey,
                        symbolName: importedName,
                        moduleName: moduleName,
                        moduleFile: moduleFile,
                        moduleFileName: moduleFile === null || moduleFile === void 0 ? void 0 : moduleFile.fileName,
                        exportKind: exportKind,
                        targetFlags: ts.skipAlias(symbol, checker).flags,
                        isFromPackageJson: isFromPackageJson,
                        symbol: storedSymbol,
                        moduleSymbol: storedModuleSymbol,
                    });
                },
                get: function (importingFile, importedName, symbol, moduleName, checker) {
                    if (importingFile !== usableByFileName)
                        return;
                    var result = exportInfo.get(key(importedName, symbol, moduleName, checker));
                    return result === null || result === void 0 ? void 0 : result.map(rehydrateCachedInfo);
                },
                forEach: function (importingFile, action) {
                    if (importingFile !== usableByFileName)
                        return;
                    exportInfo.forEach(function (info, key) {
                        var _a = parseKey(key), symbolName = _a.symbolName, ambientModuleName = _a.ambientModuleName;
                        action(info.map(rehydrateCachedInfo), symbolName, !!ambientModuleName);
                    });
                },
                releaseSymbols: function () {
                    symbols.clear();
                },
                onFileChanged: function (oldSourceFile, newSourceFile, typeAcquisitionEnabled) {
                    if (fileIsGlobalOnly(oldSourceFile) && fileIsGlobalOnly(newSourceFile)) {
                        // File is purely global; doesn't affect export map
                        return false;
                    }
                    if (usableByFileName && usableByFileName !== newSourceFile.path ||
                        // If ATA is enabled, auto-imports uses existing imports to guess whether you want auto-imports from node.
                        // Adding or removing imports from node could change the outcome of that guess, so could change the suggestions list.
                        typeAcquisitionEnabled && ts.consumesNodeCoreModules(oldSourceFile) !== ts.consumesNodeCoreModules(newSourceFile) ||
                        // Module agumentation and ambient module changes can add or remove exports available to be auto-imported.
                        // Changes elsewhere in the file can change the *type* of an export in a module augmentation,
                        // but type info is gathered in getCompletionEntryDetails, which doesn’t use the cache.
                        !ts.arrayIsEqualTo(oldSourceFile.moduleAugmentations, newSourceFile.moduleAugmentations) ||
                        !ambientModuleDeclarationsAreEqual(oldSourceFile, newSourceFile)) {
                        cache.clear();
                        return true;
                    }
                    usableByFileName = newSourceFile.path;
                    return false;
                },
            };
            if (ts.Debug.isDebugging) {
                Object.defineProperty(cache, "__cache", { get: function () { return exportInfo; } });
            }
            return cache;
            function rehydrateCachedInfo(info) {
                if (info.symbol && info.moduleSymbol)
                    return info;
                var id = info.id, exportKind = info.exportKind, targetFlags = info.targetFlags, isFromPackageJson = info.isFromPackageJson, moduleFileName = info.moduleFileName;
                var _a = symbols.get(id) || ts.emptyArray, cachedSymbol = _a[0], cachedModuleSymbol = _a[1];
                if (cachedSymbol && cachedModuleSymbol) {
                    return {
                        symbol: cachedSymbol,
                        moduleSymbol: cachedModuleSymbol,
                        moduleFileName: moduleFileName,
                        exportKind: exportKind,
                        targetFlags: targetFlags,
                        isFromPackageJson: isFromPackageJson,
                    };
                }
                var checker = (isFromPackageJson
                    ? host.getPackageJsonAutoImportProvider()
                    : host.getCurrentProgram()).getTypeChecker();
                var moduleSymbol = info.moduleSymbol || cachedModuleSymbol || ts.Debug.checkDefined(info.moduleFile
                    ? checker.getMergedSymbol(info.moduleFile.symbol)
                    : checker.tryFindAmbientModule(info.moduleName));
                var symbol = info.symbol || cachedSymbol || ts.Debug.checkDefined(exportKind === 2 /* ExportEquals */
                    ? checker.resolveExternalModuleSymbol(moduleSymbol)
                    : checker.tryGetMemberInModuleExportsAndProperties(ts.unescapeLeadingUnderscores(info.symbolTableKey), moduleSymbol), "Could not find symbol '" + info.symbolName + "' by key '" + info.symbolTableKey + "' in module " + moduleSymbol.name);
                symbols.set(id, [symbol, moduleSymbol]);
                return {
                    symbol: symbol,
                    moduleSymbol: moduleSymbol,
                    moduleFileName: moduleFileName,
                    exportKind: exportKind,
                    targetFlags: targetFlags,
                    isFromPackageJson: isFromPackageJson,
                };
            }
            function key(importedName, symbol, moduleName, checker) {
                var unquoted = ts.stripQuotes(moduleName);
                var moduleKey = ts.isExternalModuleNameRelative(unquoted) ? "/" : unquoted;
                var target = ts.skipAlias(symbol, checker);
                return importedName + "|" + createSymbolKey(target) + "|" + moduleKey;
            }
            function parseKey(key) {
                var symbolName = key.substring(0, key.indexOf("|"));
                var moduleKey = key.substring(key.lastIndexOf("|") + 1);
                var ambientModuleName = moduleKey === "/" ? undefined : moduleKey;
                return { symbolName: symbolName, ambientModuleName: ambientModuleName };
            }
            function createSymbolKey(symbol) {
                var key = symbol.name;
                while (symbol.parent) {
                    key += "," + symbol.parent.name;
                    symbol = symbol.parent;
                }
                return key;
            }
            function fileIsGlobalOnly(file) {
                return !file.commonJsModuleIndicator && !file.externalModuleIndicator && !file.moduleAugmentations && !file.ambientModuleNames;
            }
            function ambientModuleDeclarationsAreEqual(oldSourceFile, newSourceFile) {
                if (!ts.arrayIsEqualTo(oldSourceFile.ambientModuleNames, newSourceFile.ambientModuleNames)) {
                    return false;
                }
                var oldFileStatementIndex = -1;
                var newFileStatementIndex = -1;
                var _loop_1 = function (ambientModuleName) {
                    var isMatchingModuleDeclaration = function (node) { return ts.isNonGlobalAmbientModule(node) && node.name.text === ambientModuleName; };
                    oldFileStatementIndex = ts.findIndex(oldSourceFile.statements, isMatchingModuleDeclaration, oldFileStatementIndex + 1);
                    newFileStatementIndex = ts.findIndex(newSourceFile.statements, isMatchingModuleDeclaration, newFileStatementIndex + 1);
                    if (oldSourceFile.statements[oldFileStatementIndex] !== newSourceFile.statements[newFileStatementIndex]) {
                        return { value: false };
                    }
                };
                for (var _i = 0, _a = newSourceFile.ambientModuleNames; _i < _a.length; _i++) {
                    var ambientModuleName = _a[_i];
                    var state_1 = _loop_1(ambientModuleName);
                    if (typeof state_1 === "object")
                        return state_1.value;
                }
                return true;
            }
        }
        ts.createCacheableExportInfoMap = createCacheableExportInfoMap;
        function isImportableFile(program, from, to, preferences, packageJsonFilter, moduleSpecifierResolutionHost, moduleSpecifierCache) {
            var _a;
            if (from === to)
                return false;
            var cachedResult = moduleSpecifierCache === null || moduleSpecifierCache === void 0 ? void 0 : moduleSpecifierCache.get(from.path, to.path, preferences);
            if ((cachedResult === null || cachedResult === void 0 ? void 0 : cachedResult.isAutoImportable) !== undefined) {
                return cachedResult.isAutoImportable;
            }
            var getCanonicalFileName = ts.hostGetCanonicalFileName(moduleSpecifierResolutionHost);
            var globalTypingsCache = (_a = moduleSpecifierResolutionHost.getGlobalTypingsCacheLocation) === null || _a === void 0 ? void 0 : _a.call(moduleSpecifierResolutionHost);
            var hasImportablePath = !!ts.moduleSpecifiers.forEachFileNameOfModule(from.fileName, to.fileName, moduleSpecifierResolutionHost, 
            /*preferSymlinks*/ false, function (toPath) {
                var toFile = program.getSourceFile(toPath);
                // Determine to import using toPath only if toPath is what we were looking at
                // or there doesnt exist the file in the program by the symlink
                return (toFile === to || !toFile) &&
                    isImportablePath(from.fileName, toPath, getCanonicalFileName, globalTypingsCache);
            });
            if (packageJsonFilter) {
                var isAutoImportable = hasImportablePath && packageJsonFilter.allowsImportingSourceFile(to, moduleSpecifierResolutionHost);
                moduleSpecifierCache === null || moduleSpecifierCache === void 0 ? void 0 : moduleSpecifierCache.setIsAutoImportable(from.path, to.path, preferences, isAutoImportable);
                return isAutoImportable;
            }
            return hasImportablePath;
        }
        ts.isImportableFile = isImportableFile;
        /**
         * Don't include something from a `node_modules` that isn't actually reachable by a global import.
         * A relative import to node_modules is usually a bad idea.
         */
        function isImportablePath(fromPath, toPath, getCanonicalFileName, globalCachePath) {
            // If it's in a `node_modules` but is not reachable from here via a global import, don't bother.
            var toNodeModules = ts.forEachAncestorDirectory(toPath, function (ancestor) { return ts.getBaseFileName(ancestor) === "node_modules" ? ancestor : undefined; });
            var toNodeModulesParent = toNodeModules && ts.getDirectoryPath(getCanonicalFileName(toNodeModules));
            return toNodeModulesParent === undefined
                || ts.startsWith(getCanonicalFileName(fromPath), toNodeModulesParent)
                || (!!globalCachePath && ts.startsWith(getCanonicalFileName(globalCachePath), toNodeModulesParent));
        }
        function forEachExternalModuleToImportFrom(program, host, useAutoImportProvider, cb) {
            var _a, _b;
            forEachExternalModule(program.getTypeChecker(), program.getSourceFiles(), function (module, file) { return cb(module, file, program, /*isFromPackageJson*/ false); });
            var autoImportProvider = useAutoImportProvider && ((_a = host.getPackageJsonAutoImportProvider) === null || _a === void 0 ? void 0 : _a.call(host));
            if (autoImportProvider) {
                var start = ts.timestamp();
                forEachExternalModule(autoImportProvider.getTypeChecker(), autoImportProvider.getSourceFiles(), function (module, file) { return cb(module, file, autoImportProvider, /*isFromPackageJson*/ true); });
                (_b = host.log) === null || _b === void 0 ? void 0 : _b.call(host, "forEachExternalModuleToImportFrom autoImportProvider: " + (ts.timestamp() - start));
            }
        }
        ts.forEachExternalModuleToImportFrom = forEachExternalModuleToImportFrom;
        function forEachExternalModule(checker, allSourceFiles, cb) {
            for (var _i = 0, _a = checker.getAmbientModules(); _i < _a.length; _i++) {
                var ambient = _a[_i];
                if (!ts.stringContains(ambient.name, "*")) {
                    cb(ambient, /*sourceFile*/ undefined);
                }
            }
            for (var _b = 0, allSourceFiles_1 = allSourceFiles; _b < allSourceFiles_1.length; _b++) {
                var sourceFile = allSourceFiles_1[_b];
                if (ts.isExternalOrCommonJsModule(sourceFile)) {
                    cb(checker.getMergedSymbol(sourceFile.symbol), sourceFile);
                }
            }
        }
        function getExportInfoMap(importingFile, host, program, cancellationToken) {
            var _a, _b, _c, _d, _e;
            var start = ts.timestamp();
            // Pulling the AutoImportProvider project will trigger its updateGraph if pending,
            // which will invalidate the export map cache if things change, so pull it before
            // checking the cache.
            (_a = host.getPackageJsonAutoImportProvider) === null || _a === void 0 ? void 0 : _a.call(host);
            var cache = ((_b = host.getCachedExportInfoMap) === null || _b === void 0 ? void 0 : _b.call(host)) || createCacheableExportInfoMap({
                getCurrentProgram: function () { return program; },
                getPackageJsonAutoImportProvider: function () { var _a; return (_a = host.getPackageJsonAutoImportProvider) === null || _a === void 0 ? void 0 : _a.call(host); },
            });
            if (cache.isUsableByFile(importingFile.path)) {
                (_c = host.log) === null || _c === void 0 ? void 0 : _c.call(host, "getExportInfoMap: cache hit");
                return cache;
            }
            (_d = host.log) === null || _d === void 0 ? void 0 : _d.call(host, "getExportInfoMap: cache miss or empty; calculating new results");
            var compilerOptions = program.getCompilerOptions();
            var scriptTarget = ts.getEmitScriptTarget(compilerOptions);
            var moduleCount = 0;
            forEachExternalModuleToImportFrom(program, host, /*useAutoImportProvider*/ true, function (moduleSymbol, moduleFile, program, isFromPackageJson) {
                if (++moduleCount % 100 === 0)
                    cancellationToken === null || cancellationToken === void 0 ? void 0 : cancellationToken.throwIfCancellationRequested();
                var seenExports = new ts.Map();
                var checker = program.getTypeChecker();
                var defaultInfo = getDefaultLikeExportInfo(moduleSymbol, checker, compilerOptions);
                // Note: I think we shouldn't actually see resolved module symbols here, but weird merges
                // can cause it to happen: see 'completionsImport_mergedReExport.ts'
                if (defaultInfo && isImportableSymbol(defaultInfo.symbol, checker)) {
                    cache.add(importingFile.path, defaultInfo.symbol, defaultInfo.exportKind === 1 /* Default */ ? "default" /* Default */ : "export=" /* ExportEquals */, moduleSymbol, moduleFile, defaultInfo.exportKind, isFromPackageJson, scriptTarget, checker);
                }
                checker.forEachExportAndPropertyOfModule(moduleSymbol, function (exported, key) {
                    if (exported !== (defaultInfo === null || defaultInfo === void 0 ? void 0 : defaultInfo.symbol) && isImportableSymbol(exported, checker) && ts.addToSeen(seenExports, exported)) {
                        cache.add(importingFile.path, exported, key, moduleSymbol, moduleFile, 0 /* Named */, isFromPackageJson, scriptTarget, checker);
                    }
                });
            });
            (_e = host.log) === null || _e === void 0 ? void 0 : _e.call(host, "getExportInfoMap: done in " + (ts.timestamp() - start) + " ms");
            return cache;
        }
        ts.getExportInfoMap = getExportInfoMap;
        function getDefaultLikeExportInfo(moduleSymbol, checker, compilerOptions) {
            var exported = getDefaultLikeExportWorker(moduleSymbol, checker);
            if (!exported)
                return undefined;
            var symbol = exported.symbol, exportKind = exported.exportKind;
            var info = getDefaultExportInfoWorker(symbol, checker, compilerOptions);
            return info && __assign({ symbol: symbol, exportKind: exportKind }, info);
        }
        ts.getDefaultLikeExportInfo = getDefaultLikeExportInfo;
        function isImportableSymbol(symbol, checker) {
            return !checker.isUndefinedSymbol(symbol) && !checker.isUnknownSymbol(symbol) && !ts.isKnownSymbol(symbol) && !ts.isPrivateIdentifierSymbol(symbol);
        }
        function getDefaultLikeExportWorker(moduleSymbol, checker) {
            var exportEquals = checker.resolveExternalModuleSymbol(moduleSymbol);
            if (exportEquals !== moduleSymbol)
                return { symbol: exportEquals, exportKind: 2 /* ExportEquals */ };
            var defaultExport = checker.tryGetMemberInModuleExports("default" /* Default */, moduleSymbol);
            if (defaultExport)
                return { symbol: defaultExport, exportKind: 1 /* Default */ };
        }
        function getDefaultExportInfoWorker(defaultExport, checker, compilerOptions) {
            var localSymbol = ts.getLocalSymbolForExportDefault(defaultExport);
            if (localSymbol)
                return { symbolForMeaning: localSymbol, name: localSymbol.name };
            var name = getNameForExportDefault(defaultExport);
            if (name !== undefined)
                return { symbolForMeaning: defaultExport, name: name };
            if (defaultExport.flags & 2097152 /* Alias */) {
                var aliased = checker.getImmediateAliasedSymbol(defaultExport);
                if (aliased && aliased.parent) {
                    // - `aliased` will be undefined if the module is exporting an unresolvable name,
                    //    but we can still offer completions for it.
                    // - `aliased.parent` will be undefined if the module is exporting `globalThis.something`,
                    //    or another expression that resolves to a global.
                    return getDefaultExportInfoWorker(aliased, checker, compilerOptions);
                }
            }
            if (defaultExport.escapedName !== "default" /* Default */ &&
                defaultExport.escapedName !== "export=" /* ExportEquals */) {
                return { symbolForMeaning: defaultExport, name: defaultExport.getName() };
            }
            return { symbolForMeaning: defaultExport, name: ts.getNameForExportedSymbol(defaultExport, compilerOptions.target) };
        }
        function getNameForExportDefault(symbol) {
            return symbol.declarations && ts.firstDefined(symbol.declarations, function (declaration) {
                var _a;
                if (ts.isExportAssignment(declaration)) {
                    return (_a = ts.tryCast(ts.skipOuterExpressions(declaration.expression), ts.isIdentifier)) === null || _a === void 0 ? void 0 : _a.text;
                }
                else if (ts.isExportSpecifier(declaration)) {
                    ts.Debug.assert(declaration.name.text === "default" /* Default */, "Expected the specifier to be a default export");
                    return declaration.propertyName && declaration.propertyName.text;
                }
            });
        }
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        /** The classifier is used for syntactic highlighting in editors via the TSServer */
        function createClassifier() {
            var scanner = ts.createScanner(99 /* Latest */, /*skipTrivia*/ false);
            function getClassificationsForLine(text, lexState, syntacticClassifierAbsent) {
                return convertClassificationsToResult(getEncodedLexicalClassifications(text, lexState, syntacticClassifierAbsent), text);
            }
            // If there is a syntactic classifier ('syntacticClassifierAbsent' is false),
            // we will be more conservative in order to avoid conflicting with the syntactic classifier.
            function getEncodedLexicalClassifications(text, lexState, syntacticClassifierAbsent) {
                var token = 0 /* Unknown */;
                var lastNonTriviaToken = 0 /* Unknown */;
                // Just a stack of TemplateHeads and OpenCurlyBraces, used to perform rudimentary (inexact)
                // classification on template strings. Because of the context free nature of templates,
                // the only precise way to classify a template portion would be by propagating the stack across
                // lines, just as we do with the end-of-line state. However, this is a burden for implementers,
                // and the behavior is entirely subsumed by the syntactic classifier anyway, so we instead
                // flatten any nesting when the template stack is non-empty and encode it in the end-of-line state.
                // Situations in which this fails are
                //  1) When template strings are nested across different lines:
                //          `hello ${ `world
                //          ` }`
                //
                //     Where on the second line, you will get the closing of a template,
                //     a closing curly, and a new template.
                //
                //  2) When substitution expressions have curly braces and the curly brace falls on the next line:
                //          `hello ${ () => {
                //          return "world" } } `
                //
                //     Where on the second line, you will get the 'return' keyword,
                //     a string literal, and a template end consisting of '} } `'.
                var templateStack = [];
                var _a = getPrefixFromLexState(lexState), prefix = _a.prefix, pushTemplate = _a.pushTemplate;
                text = prefix + text;
                var offset = prefix.length;
                if (pushTemplate) {
                    templateStack.push(15 /* TemplateHead */);
                }
                scanner.setText(text);
                var endOfLineState = 0 /* None */;
                var spans = [];
                // We can run into an unfortunate interaction between the lexical and syntactic classifier
                // when the user is typing something generic.  Consider the case where the user types:
                //
                //      Foo<number
                //
                // From the lexical classifier's perspective, 'number' is a keyword, and so the word will
                // be classified as such.  However, from the syntactic classifier's tree-based perspective
                // this is simply an expression with the identifier 'number' on the RHS of the less than
                // token.  So the classification will go back to being an identifier.  The moment the user
                // types again, number will become a keyword, then an identifier, etc. etc.
                //
                // To try to avoid this problem, we avoid classifying contextual keywords as keywords
                // when the user is potentially typing something generic.  We just can't do a good enough
                // job at the lexical level, and so well leave it up to the syntactic classifier to make
                // the determination.
                //
                // In order to determine if the user is potentially typing something generic, we use a
                // weak heuristic where we track < and > tokens.  It's a weak heuristic, but should
                // work well enough in practice.
                var angleBracketStack = 0;
                do {
                    token = scanner.scan();
                    if (!ts.isTrivia(token)) {
                        handleToken();
                        lastNonTriviaToken = token;
                    }
                    var end = scanner.getTextPos();
                    pushEncodedClassification(scanner.getTokenPos(), end, offset, classFromKind(token), spans);
                    if (end >= text.length) {
                        var end_1 = getNewEndOfLineState(scanner, token, ts.lastOrUndefined(templateStack));
                        if (end_1 !== undefined) {
                            endOfLineState = end_1;
                        }
                    }
                } while (token !== 1 /* EndOfFileToken */);
                function handleToken() {
                    switch (token) {
                        case 43 /* SlashToken */:
                        case 68 /* SlashEqualsToken */:
                            if (!noRegexTable[lastNonTriviaToken] && scanner.reScanSlashToken() === 13 /* RegularExpressionLiteral */) {
                                token = 13 /* RegularExpressionLiteral */;
                            }
                            break;
                        case 29 /* LessThanToken */:
                            if (lastNonTriviaToken === 79 /* Identifier */) {
                                // Could be the start of something generic.  Keep track of that by bumping
                                // up the current count of generic contexts we may be in.
                                angleBracketStack++;
                            }
                            break;
                        case 31 /* GreaterThanToken */:
                            if (angleBracketStack > 0) {
                                // If we think we're currently in something generic, then mark that that
                                // generic entity is complete.
                                angleBracketStack--;
                            }
                            break;
                        case 129 /* AnyKeyword */:
                        case 148 /* StringKeyword */:
                        case 145 /* NumberKeyword */:
                        case 132 /* BooleanKeyword */:
                        case 149 /* SymbolKeyword */:
                            if (angleBracketStack > 0 && !syntacticClassifierAbsent) {
                                // If it looks like we're could be in something generic, don't classify this
                                // as a keyword.  We may just get overwritten by the syntactic classifier,
                                // causing a noisy experience for the user.
                                token = 79 /* Identifier */;
                            }
                            break;
                        case 15 /* TemplateHead */:
                            templateStack.push(token);
                            break;
                        case 18 /* OpenBraceToken */:
                            // If we don't have anything on the template stack,
                            // then we aren't trying to keep track of a previously scanned template head.
                            if (templateStack.length > 0) {
                                templateStack.push(token);
                            }
                            break;
                        case 19 /* CloseBraceToken */:
                            // If we don't have anything on the template stack,
                            // then we aren't trying to keep track of a previously scanned template head.
                            if (templateStack.length > 0) {
                                var lastTemplateStackToken = ts.lastOrUndefined(templateStack);
                                if (lastTemplateStackToken === 15 /* TemplateHead */) {
                                    token = scanner.reScanTemplateToken(/* isTaggedTemplate */ false);
                                    // Only pop on a TemplateTail; a TemplateMiddle indicates there is more for us.
                                    if (token === 17 /* TemplateTail */) {
                                        templateStack.pop();
                                    }
                                    else {
                                        ts.Debug.assertEqual(token, 16 /* TemplateMiddle */, "Should have been a template middle.");
                                    }
                                }
                                else {
                                    ts.Debug.assertEqual(lastTemplateStackToken, 18 /* OpenBraceToken */, "Should have been an open brace");
                                    templateStack.pop();
                                }
                            }
                            break;
                        default:
                            if (!ts.isKeyword(token)) {
                                break;
                            }
                            if (lastNonTriviaToken === 24 /* DotToken */) {
                                token = 79 /* Identifier */;
                            }
                            else if (ts.isKeyword(lastNonTriviaToken) && ts.isKeyword(token) && !canFollow(lastNonTriviaToken, token)) {
                                // We have two keywords in a row.  Only treat the second as a keyword if
                                // it's a sequence that could legally occur in the language.  Otherwise
                                // treat it as an identifier.  This way, if someone writes "private var"
                                // we recognize that 'var' is actually an identifier here.
                                token = 79 /* Identifier */;
                            }
                    }
                }
                return { endOfLineState: endOfLineState, spans: spans };
            }
            return { getClassificationsForLine: getClassificationsForLine, getEncodedLexicalClassifications: getEncodedLexicalClassifications };
        }
        ts.createClassifier = createClassifier;
        /// We do not have a full parser support to know when we should parse a regex or not
        /// If we consider every slash token to be a regex, we could be missing cases like "1/2/3", where
        /// we have a series of divide operator. this list allows us to be more accurate by ruling out
        /// locations where a regexp cannot exist.
        var noRegexTable = ts.arrayToNumericMap([
            79 /* Identifier */,
            10 /* StringLiteral */,
            8 /* NumericLiteral */,
            9 /* BigIntLiteral */,
            13 /* RegularExpressionLiteral */,
            108 /* ThisKeyword */,
            45 /* PlusPlusToken */,
            46 /* MinusMinusToken */,
            21 /* CloseParenToken */,
            23 /* CloseBracketToken */,
            19 /* CloseBraceToken */,
            110 /* TrueKeyword */,
            95 /* FalseKeyword */,
        ], function (token) { return token; }, function () { return true; });
        function getNewEndOfLineState(scanner, token, lastOnTemplateStack) {
            switch (token) {
                case 10 /* StringLiteral */: {
                    // Check to see if we finished up on a multiline string literal.
                    if (!scanner.isUnterminated())
                        return undefined;
                    var tokenText = scanner.getTokenText();
                    var lastCharIndex = tokenText.length - 1;
                    var numBackslashes = 0;
                    while (tokenText.charCodeAt(lastCharIndex - numBackslashes) === 92 /* backslash */) {
                        numBackslashes++;
                    }
                    // If we have an odd number of backslashes, then the multiline string is unclosed
                    if ((numBackslashes & 1) === 0)
                        return undefined;
                    return tokenText.charCodeAt(0) === 34 /* doubleQuote */ ? 3 /* InDoubleQuoteStringLiteral */ : 2 /* InSingleQuoteStringLiteral */;
                }
                case 3 /* MultiLineCommentTrivia */:
                    // Check to see if the multiline comment was unclosed.
                    return scanner.isUnterminated() ? 1 /* InMultiLineCommentTrivia */ : undefined;
                default:
                    if (ts.isTemplateLiteralKind(token)) {
                        if (!scanner.isUnterminated()) {
                            return undefined;
                        }
                        switch (token) {
                            case 17 /* TemplateTail */:
                                return 5 /* InTemplateMiddleOrTail */;
                            case 14 /* NoSubstitutionTemplateLiteral */:
                                return 4 /* InTemplateHeadOrNoSubstitutionTemplate */;
                            default:
                                return ts.Debug.fail("Only 'NoSubstitutionTemplateLiteral's and 'TemplateTail's can be unterminated; got SyntaxKind #" + token);
                        }
                    }
                    return lastOnTemplateStack === 15 /* TemplateHead */ ? 6 /* InTemplateSubstitutionPosition */ : undefined;
            }
        }
        function pushEncodedClassification(start, end, offset, classification, result) {
            if (classification === 8 /* whiteSpace */) {
                // Don't bother with whitespace classifications.  They're not needed.
                return;
            }
            if (start === 0 && offset > 0) {
                // We're classifying the first token, and this was a case where we prepended text.
                // We should consider the start of this token to be at the start of the original text.
                start += offset;
            }
            var length = end - start;
            if (length > 0) {
                // All our tokens are in relation to the augmented text.  Move them back to be
                // relative to the original text.
                result.push(start - offset, length, classification);
            }
        }
        function convertClassificationsToResult(classifications, text) {
            var entries = [];
            var dense = classifications.spans;
            var lastEnd = 0;
            for (var i = 0; i < dense.length; i += 3) {
                var start = dense[i];
                var length_1 = dense[i + 1];
                var type = dense[i + 2];
                // Make a whitespace entry between the last item and this one.
                if (lastEnd >= 0) {
                    var whitespaceLength_1 = start - lastEnd;
                    if (whitespaceLength_1 > 0) {
                        entries.push({ length: whitespaceLength_1, classification: ts.TokenClass.Whitespace });
                    }
                }
                entries.push({ length: length_1, classification: convertClassification(type) });
                lastEnd = start + length_1;
            }
            var whitespaceLength = text.length - lastEnd;
            if (whitespaceLength > 0) {
                entries.push({ length: whitespaceLength, classification: ts.TokenClass.Whitespace });
            }
            return { entries: entries, finalLexState: classifications.endOfLineState };
        }
        function convertClassification(type) {
            switch (type) {
                case 1 /* comment */: return ts.TokenClass.Comment;
                case 3 /* keyword */: return ts.TokenClass.Keyword;
                case 4 /* numericLiteral */: return ts.TokenClass.NumberLiteral;
                case 25 /* bigintLiteral */: return ts.TokenClass.BigIntLiteral;
                case 5 /* operator */: return ts.TokenClass.Operator;
                case 6 /* stringLiteral */: return ts.TokenClass.StringLiteral;
                case 8 /* whiteSpace */: return ts.TokenClass.Whitespace;
                case 10 /* punctuation */: return ts.TokenClass.Punctuation;
                case 2 /* identifier */:
                case 11 /* className */:
                case 12 /* enumName */:
                case 13 /* interfaceName */:
                case 14 /* moduleName */:
                case 15 /* typeParameterName */:
                case 16 /* typeAliasName */:
                case 9 /* text */:
                case 17 /* parameterName */:
                    return ts.TokenClass.Identifier;
                default:
                    return undefined; // TODO: GH#18217 Debug.assertNever(type);
            }
        }
        /** Returns true if 'keyword2' can legally follow 'keyword1' in any language construct. */
        function canFollow(keyword1, keyword2) {
            if (!ts.isAccessibilityModifier(keyword1)) {
                // Assume any other keyword combination is legal.
                // This can be refined in the future if there are more cases we want the classifier to be better at.
                return true;
            }
            switch (keyword2) {
                case 135 /* GetKeyword */:
                case 147 /* SetKeyword */:
                case 133 /* ConstructorKeyword */:
                case 124 /* StaticKeyword */:
                    return true; // Allow things like "public get", "public constructor" and "public static".
                default:
                    return false; // Any other keyword following "public" is actually an identifier, not a real keyword.
            }
        }
        function getPrefixFromLexState(lexState) {
            // If we're in a string literal, then prepend: "\
            // (and a newline).  That way when we lex we'll think we're still in a string literal.
            //
            // If we're in a multiline comment, then prepend: /*
            // (and a newline).  That way when we lex we'll think we're still in a multiline comment.
            switch (lexState) {
                case 3 /* InDoubleQuoteStringLiteral */:
                    return { prefix: "\"\\\n" };
                case 2 /* InSingleQuoteStringLiteral */:
                    return { prefix: "'\\\n" };
                case 1 /* InMultiLineCommentTrivia */:
                    return { prefix: "/*\n" };
                case 4 /* InTemplateHeadOrNoSubstitutionTemplate */:
                    return { prefix: "`\n" };
                case 5 /* InTemplateMiddleOrTail */:
                    return { prefix: "}\n", pushTemplate: true };
                case 6 /* InTemplateSubstitutionPosition */:
                    return { prefix: "", pushTemplate: true };
                case 0 /* None */:
                    return { prefix: "" };
                default:
                    return ts.Debug.assertNever(lexState);
            }
        }
        function isBinaryExpressionOperatorToken(token) {
            switch (token) {
                case 41 /* AsteriskToken */:
                case 43 /* SlashToken */:
                case 44 /* PercentToken */:
                case 39 /* PlusToken */:
                case 40 /* MinusToken */:
                case 47 /* LessThanLessThanToken */:
                case 48 /* GreaterThanGreaterThanToken */:
                case 49 /* GreaterThanGreaterThanGreaterThanToken */:
                case 29 /* LessThanToken */:
                case 31 /* GreaterThanToken */:
                case 32 /* LessThanEqualsToken */:
                case 33 /* GreaterThanEqualsToken */:
                case 102 /* InstanceOfKeyword */:
                case 101 /* InKeyword */:
                case 127 /* AsKeyword */:
                case 34 /* EqualsEqualsToken */:
                case 35 /* ExclamationEqualsToken */:
                case 36 /* EqualsEqualsEqualsToken */:
                case 37 /* ExclamationEqualsEqualsToken */:
                case 50 /* AmpersandToken */:
                case 52 /* CaretToken */:
                case 51 /* BarToken */:
                case 55 /* AmpersandAmpersandToken */:
                case 56 /* BarBarToken */:
                case 74 /* BarEqualsToken */:
                case 73 /* AmpersandEqualsToken */:
                case 78 /* CaretEqualsToken */:
                case 70 /* LessThanLessThanEqualsToken */:
                case 71 /* GreaterThanGreaterThanEqualsToken */:
                case 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */:
                case 64 /* PlusEqualsToken */:
                case 65 /* MinusEqualsToken */:
                case 66 /* AsteriskEqualsToken */:
                case 68 /* SlashEqualsToken */:
                case 69 /* PercentEqualsToken */:
                case 63 /* EqualsToken */:
                case 27 /* CommaToken */:
                case 60 /* QuestionQuestionToken */:
                case 75 /* BarBarEqualsToken */:
                case 76 /* AmpersandAmpersandEqualsToken */:
                case 77 /* QuestionQuestionEqualsToken */:
                    return true;
                default:
                    return false;
            }
        }
        function isPrefixUnaryExpressionOperatorToken(token) {
            switch (token) {
                case 39 /* PlusToken */:
                case 40 /* MinusToken */:
                case 54 /* TildeToken */:
                case 53 /* ExclamationToken */:
                case 45 /* PlusPlusToken */:
                case 46 /* MinusMinusToken */:
                    return true;
                default:
                    return false;
            }
        }
        function classFromKind(token) {
            if (ts.isKeyword(token)) {
                return 3 /* keyword */;
            }
            else if (isBinaryExpressionOperatorToken(token) || isPrefixUnaryExpressionOperatorToken(token)) {
                return 5 /* operator */;
            }
            else if (token >= 18 /* FirstPunctuation */ && token <= 78 /* LastPunctuation */) {
                return 10 /* punctuation */;
            }
            switch (token) {
                case 8 /* NumericLiteral */:
                    return 4 /* numericLiteral */;
                case 9 /* BigIntLiteral */:
                    return 25 /* bigintLiteral */;
                case 10 /* StringLiteral */:
                    return 6 /* stringLiteral */;
                case 13 /* RegularExpressionLiteral */:
                    return 7 /* regularExpressionLiteral */;
                case 7 /* ConflictMarkerTrivia */:
                case 3 /* MultiLineCommentTrivia */:
                case 2 /* SingleLineCommentTrivia */:
                    return 1 /* comment */;
                case 5 /* WhitespaceTrivia */:
                case 4 /* NewLineTrivia */:
                    return 8 /* whiteSpace */;
                case 79 /* Identifier */:
                default:
                    if (ts.isTemplateLiteralKind(token)) {
                        return 6 /* stringLiteral */;
                    }
                    return 2 /* identifier */;
            }
        }
        /* @internal */
        function getSemanticClassifications(typeChecker, cancellationToken, sourceFile, classifiableNames, span) {
            return convertClassificationsToSpans(getEncodedSemanticClassifications(typeChecker, cancellationToken, sourceFile, classifiableNames, span));
        }
        ts.getSemanticClassifications = getSemanticClassifications;
        function checkForClassificationCancellation(cancellationToken, kind) {
            // We don't want to actually call back into our host on every node to find out if we've
            // been canceled.  That would be an enormous amount of chattyness, along with the all
            // the overhead of marshalling the data to/from the host.  So instead we pick a few
            // reasonable node kinds to bother checking on.  These node kinds represent high level
            // constructs that we would expect to see commonly, but just at a far less frequent
            // interval.
            //
            // For example, in checker.ts (around 750k) we only have around 600 of these constructs.
            // That means we're calling back into the host around every 1.2k of the file we process.
            // Lib.d.ts has similar numbers.
            switch (kind) {
                case 259 /* ModuleDeclaration */:
                case 255 /* ClassDeclaration */:
                case 256 /* InterfaceDeclaration */:
                case 254 /* FunctionDeclaration */:
                case 224 /* ClassExpression */:
                case 211 /* FunctionExpression */:
                case 212 /* ArrowFunction */:
                    cancellationToken.throwIfCancellationRequested();
            }
        }
        /* @internal */
        function getEncodedSemanticClassifications(typeChecker, cancellationToken, sourceFile, classifiableNames, span) {
            var spans = [];
            sourceFile.forEachChild(function cb(node) {
                // Only walk into nodes that intersect the requested span.
                if (!node || !ts.textSpanIntersectsWith(span, node.pos, node.getFullWidth())) {
                    return;
                }
                checkForClassificationCancellation(cancellationToken, node.kind);
                // Only bother calling into the typechecker if this is an identifier that
                // could possibly resolve to a type name.  This makes classification run
                // in a third of the time it would normally take.
                if (ts.isIdentifier(node) && !ts.nodeIsMissing(node) && classifiableNames.has(node.escapedText)) {
                    var symbol = typeChecker.getSymbolAtLocation(node);
                    var type = symbol && classifySymbol(symbol, ts.getMeaningFromLocation(node), typeChecker);
                    if (type) {
                        pushClassification(node.getStart(sourceFile), node.getEnd(), type);
                    }
                }
                node.forEachChild(cb);
            });
            return { spans: spans, endOfLineState: 0 /* None */ };
            function pushClassification(start, end, type) {
                var length = end - start;
                ts.Debug.assert(length > 0, "Classification had non-positive length of " + length);
                spans.push(start);
                spans.push(length);
                spans.push(type);
            }
        }
        ts.getEncodedSemanticClassifications = getEncodedSemanticClassifications;
        function classifySymbol(symbol, meaningAtPosition, checker) {
            var flags = symbol.getFlags();
            if ((flags & 2885600 /* Classifiable */) === 0 /* None */) {
                return undefined;
            }
            else if (flags & 32 /* Class */) {
                return 11 /* className */;
            }
            else if (flags & 384 /* Enum */) {
                return 12 /* enumName */;
            }
            else if (flags & 524288 /* TypeAlias */) {
                return 16 /* typeAliasName */;
            }
            else if (flags & 1536 /* Module */) {
                // Only classify a module as such if
                //  - It appears in a namespace context.
                //  - There exists a module declaration which actually impacts the value side.
                return meaningAtPosition & 4 /* Namespace */ || meaningAtPosition & 1 /* Value */ && hasValueSideModule(symbol) ? 14 /* moduleName */ : undefined;
            }
            else if (flags & 2097152 /* Alias */) {
                return classifySymbol(checker.getAliasedSymbol(symbol), meaningAtPosition, checker);
            }
            else if (meaningAtPosition & 2 /* Type */) {
                return flags & 64 /* Interface */ ? 13 /* interfaceName */ : flags & 262144 /* TypeParameter */ ? 15 /* typeParameterName */ : undefined;
            }
            else {
                return undefined;
            }
        }
        /** Returns true if there exists a module that introduces entities on the value side. */
        function hasValueSideModule(symbol) {
            return ts.some(symbol.declarations, function (declaration) {
                return ts.isModuleDeclaration(declaration) && ts.getModuleInstanceState(declaration) === 1 /* Instantiated */;
            });
        }
        function getClassificationTypeName(type) {
            switch (type) {
                case 1 /* comment */: return "comment" /* comment */;
                case 2 /* identifier */: return "identifier" /* identifier */;
                case 3 /* keyword */: return "keyword" /* keyword */;
                case 4 /* numericLiteral */: return "number" /* numericLiteral */;
                case 25 /* bigintLiteral */: return "bigint" /* bigintLiteral */;
                case 5 /* operator */: return "operator" /* operator */;
                case 6 /* stringLiteral */: return "string" /* stringLiteral */;
                case 8 /* whiteSpace */: return "whitespace" /* whiteSpace */;
                case 9 /* text */: return "text" /* text */;
                case 10 /* punctuation */: return "punctuation" /* punctuation */;
                case 11 /* className */: return "class name" /* className */;
                case 12 /* enumName */: return "enum name" /* enumName */;
                case 13 /* interfaceName */: return "interface name" /* interfaceName */;
                case 14 /* moduleName */: return "module name" /* moduleName */;
                case 15 /* typeParameterName */: return "type parameter name" /* typeParameterName */;
                case 16 /* typeAliasName */: return "type alias name" /* typeAliasName */;
                case 17 /* parameterName */: return "parameter name" /* parameterName */;
                case 18 /* docCommentTagName */: return "doc comment tag name" /* docCommentTagName */;
                case 19 /* jsxOpenTagName */: return "jsx open tag name" /* jsxOpenTagName */;
                case 20 /* jsxCloseTagName */: return "jsx close tag name" /* jsxCloseTagName */;
                case 21 /* jsxSelfClosingTagName */: return "jsx self closing tag name" /* jsxSelfClosingTagName */;
                case 22 /* jsxAttribute */: return "jsx attribute" /* jsxAttribute */;
                case 23 /* jsxText */: return "jsx text" /* jsxText */;
                case 24 /* jsxAttributeStringLiteralValue */: return "jsx attribute string literal value" /* jsxAttributeStringLiteralValue */;
                default: return undefined; // TODO: GH#18217 throw Debug.assertNever(type);
            }
        }
        function convertClassificationsToSpans(classifications) {
            ts.Debug.assert(classifications.spans.length % 3 === 0);
            var dense = classifications.spans;
            var result = [];
            for (var i = 0; i < dense.length; i += 3) {
                result.push({
                    textSpan: ts.createTextSpan(dense[i], dense[i + 1]),
                    classificationType: getClassificationTypeName(dense[i + 2])
                });
            }
            return result;
        }
        /* @internal */
        function getSyntacticClassifications(cancellationToken, sourceFile, span) {
            return convertClassificationsToSpans(getEncodedSyntacticClassifications(cancellationToken, sourceFile, span));
        }
        ts.getSyntacticClassifications = getSyntacticClassifications;
        /* @internal */
        function getEncodedSyntacticClassifications(cancellationToken, sourceFile, span) {
            var spanStart = span.start;
            var spanLength = span.length;
            // Make a scanner we can get trivia from.
            var triviaScanner = ts.createScanner(99 /* Latest */, /*skipTrivia*/ false, sourceFile.languageVariant, sourceFile.text);
            var mergeConflictScanner = ts.createScanner(99 /* Latest */, /*skipTrivia*/ false, sourceFile.languageVariant, sourceFile.text);
            var result = [];
            processElement(sourceFile);
            return { spans: result, endOfLineState: 0 /* None */ };
            function pushClassification(start, length, type) {
                result.push(start);
                result.push(length);
                result.push(type);
            }
            function classifyLeadingTriviaAndGetTokenStart(token) {
                triviaScanner.setTextPos(token.pos);
                while (true) {
                    var start = triviaScanner.getTextPos();
                    // only bother scanning if we have something that could be trivia.
                    if (!ts.couldStartTrivia(sourceFile.text, start)) {
                        return start;
                    }
                    var kind = triviaScanner.scan();
                    var end = triviaScanner.getTextPos();
                    var width = end - start;
                    // The moment we get something that isn't trivia, then stop processing.
                    if (!ts.isTrivia(kind)) {
                        return start;
                    }
                    switch (kind) {
                        case 4 /* NewLineTrivia */:
                        case 5 /* WhitespaceTrivia */:
                            // Don't bother with newlines/whitespace.
                            continue;
                        case 2 /* SingleLineCommentTrivia */:
                        case 3 /* MultiLineCommentTrivia */:
                            // Only bother with the trivia if it at least intersects the span of interest.
                            classifyComment(token, kind, start, width);
                            // Classifying a comment might cause us to reuse the trivia scanner
                            // (because of jsdoc comments).  So after we classify the comment make
                            // sure we set the scanner position back to where it needs to be.
                            triviaScanner.setTextPos(end);
                            continue;
                        case 7 /* ConflictMarkerTrivia */:
                            var text = sourceFile.text;
                            var ch = text.charCodeAt(start);
                            // for the <<<<<<< and >>>>>>> markers, we just add them in as comments
                            // in the classification stream.
                            if (ch === 60 /* lessThan */ || ch === 62 /* greaterThan */) {
                                pushClassification(start, width, 1 /* comment */);
                                continue;
                            }
                            // for the ||||||| and ======== markers, add a comment for the first line,
                            // and then lex all subsequent lines up until the end of the conflict marker.
                            ts.Debug.assert(ch === 124 /* bar */ || ch === 61 /* equals */);
                            classifyDisabledMergeCode(text, start, end);
                            break;
                        case 6 /* ShebangTrivia */:
                            // TODO: Maybe we should classify these.
                            break;
                        default:
                            ts.Debug.assertNever(kind);
                    }
                }
            }
            function classifyComment(token, kind, start, width) {
                if (kind === 3 /* MultiLineCommentTrivia */) {
                    // See if this is a doc comment.  If so, we'll classify certain portions of it
                    // specially.
                    var docCommentAndDiagnostics = ts.parseIsolatedJSDocComment(sourceFile.text, start, width);
                    if (docCommentAndDiagnostics && docCommentAndDiagnostics.jsDoc) {
                        // TODO: This should be predicated on `token["kind"]` being compatible with `HasJSDoc["kind"]`
                        ts.setParent(docCommentAndDiagnostics.jsDoc, token);
                        classifyJSDocComment(docCommentAndDiagnostics.jsDoc);
                        return;
                    }
                }
                else if (kind === 2 /* SingleLineCommentTrivia */) {
                    if (tryClassifyTripleSlashComment(start, width)) {
                        return;
                    }
                }
                // Simple comment.  Just add as is.
                pushCommentRange(start, width);
            }
            function pushCommentRange(start, width) {
                pushClassification(start, width, 1 /* comment */);
            }
            function classifyJSDocComment(docComment) {
                var _a, _b, _c, _d, _e, _f, _g;
                var pos = docComment.pos;
                if (docComment.tags) {
                    for (var _i = 0, _h = docComment.tags; _i < _h.length; _i++) {
                        var tag = _h[_i];
                        // As we walk through each tag, classify the portion of text from the end of
                        // the last tag (or the start of the entire doc comment) as 'comment'.
                        if (tag.pos !== pos) {
                            pushCommentRange(pos, tag.pos - pos);
                        }
                        pushClassification(tag.pos, 1, 10 /* punctuation */); // "@"
                        pushClassification(tag.tagName.pos, tag.tagName.end - tag.tagName.pos, 18 /* docCommentTagName */); // e.g. "param"
                        pos = tag.tagName.end;
                        var commentStart = tag.tagName.end;
                        switch (tag.kind) {
                            case 335 /* JSDocParameterTag */:
                                var param = tag;
                                processJSDocParameterTag(param);
                                commentStart = param.isNameFirst && ((_a = param.typeExpression) === null || _a === void 0 ? void 0 : _a.end) || param.name.end;
                                break;
                            case 342 /* JSDocPropertyTag */:
                                var prop = tag;
                                commentStart = prop.isNameFirst && ((_b = prop.typeExpression) === null || _b === void 0 ? void 0 : _b.end) || prop.name.end;
                                break;
                            case 339 /* JSDocTemplateTag */:
                                processJSDocTemplateTag(tag);
                                pos = tag.end;
                                commentStart = tag.typeParameters.end;
                                break;
                            case 340 /* JSDocTypedefTag */:
                                var type = tag;
                                commentStart = ((_c = type.typeExpression) === null || _c === void 0 ? void 0 : _c.kind) === 304 /* JSDocTypeExpression */ && ((_d = type.fullName) === null || _d === void 0 ? void 0 : _d.end) || ((_e = type.typeExpression) === null || _e === void 0 ? void 0 : _e.end) || commentStart;
                                break;
                            case 333 /* JSDocCallbackTag */:
                                commentStart = tag.typeExpression.end;
                                break;
                            case 338 /* JSDocTypeTag */:
                                processElement(tag.typeExpression);
                                pos = tag.end;
                                commentStart = tag.typeExpression.end;
                                break;
                            case 337 /* JSDocThisTag */:
                            case 334 /* JSDocEnumTag */:
                                commentStart = tag.typeExpression.end;
                                break;
                            case 336 /* JSDocReturnTag */:
                                processElement(tag.typeExpression);
                                pos = tag.end;
                                commentStart = ((_f = tag.typeExpression) === null || _f === void 0 ? void 0 : _f.end) || commentStart;
                                break;
                            case 341 /* JSDocSeeTag */:
                                commentStart = ((_g = tag.name) === null || _g === void 0 ? void 0 : _g.end) || commentStart;
                                break;
                            case 323 /* JSDocAugmentsTag */:
                            case 324 /* JSDocImplementsTag */:
                                commentStart = tag.class.end;
                                break;
                        }
                        if (typeof tag.comment === "object") {
                            pushCommentRange(tag.comment.pos, tag.comment.end - tag.comment.pos);
                        }
                        else if (typeof tag.comment === "string") {
                            pushCommentRange(commentStart, tag.end - commentStart);
                        }
                    }
                }
                if (pos !== docComment.end) {
                    pushCommentRange(pos, docComment.end - pos);
                }
                return;
                function processJSDocParameterTag(tag) {
                    if (tag.isNameFirst) {
                        pushCommentRange(pos, tag.name.pos - pos);
                        pushClassification(tag.name.pos, tag.name.end - tag.name.pos, 17 /* parameterName */);
                        pos = tag.name.end;
                    }
                    if (tag.typeExpression) {
                        pushCommentRange(pos, tag.typeExpression.pos - pos);
                        processElement(tag.typeExpression);
                        pos = tag.typeExpression.end;
                    }
                    if (!tag.isNameFirst) {
                        pushCommentRange(pos, tag.name.pos - pos);
                        pushClassification(tag.name.pos, tag.name.end - tag.name.pos, 17 /* parameterName */);
                        pos = tag.name.end;
                    }
                }
            }
            function tryClassifyTripleSlashComment(start, width) {
                var tripleSlashXMLCommentRegEx = /^(\/\/\/\s*)(<)(?:(\S+)((?:[^/]|\/[^>])*)(\/>)?)?/im;
                // Require a leading whitespace character (the parser already does) to prevent terrible backtracking performance
                var attributeRegex = /(\s)(\S+)(\s*)(=)(\s*)('[^']+'|"[^"]+")/img;
                var text = sourceFile.text.substr(start, width);
                var match = tripleSlashXMLCommentRegEx.exec(text);
                if (!match) {
                    return false;
                }
                // Limiting classification to exactly the elements and attributes
                // defined in `ts.commentPragmas` would be excessive, but we can avoid
                // some obvious false positives (e.g. in XML-like doc comments) by
                // checking the element name.
                // eslint-disable-next-line no-in-operator
                if (!match[3] || !(match[3] in ts.commentPragmas)) {
                    return false;
                }
                var pos = start;
                pushCommentRange(pos, match[1].length); // ///
                pos += match[1].length;
                pushClassification(pos, match[2].length, 10 /* punctuation */); // <
                pos += match[2].length;
                pushClassification(pos, match[3].length, 21 /* jsxSelfClosingTagName */); // element name
                pos += match[3].length;
                var attrText = match[4];
                var attrPos = pos;
                while (true) {
                    var attrMatch = attributeRegex.exec(attrText);
                    if (!attrMatch) {
                        break;
                    }
                    var newAttrPos = pos + attrMatch.index + attrMatch[1].length; // whitespace
                    if (newAttrPos > attrPos) {
                        pushCommentRange(attrPos, newAttrPos - attrPos);
                        attrPos = newAttrPos;
                    }
                    pushClassification(attrPos, attrMatch[2].length, 22 /* jsxAttribute */); // attribute name
                    attrPos += attrMatch[2].length;
                    if (attrMatch[3].length) {
                        pushCommentRange(attrPos, attrMatch[3].length); // whitespace
                        attrPos += attrMatch[3].length;
                    }
                    pushClassification(attrPos, attrMatch[4].length, 5 /* operator */); // =
                    attrPos += attrMatch[4].length;
                    if (attrMatch[5].length) {
                        pushCommentRange(attrPos, attrMatch[5].length); // whitespace
                        attrPos += attrMatch[5].length;
                    }
                    pushClassification(attrPos, attrMatch[6].length, 24 /* jsxAttributeStringLiteralValue */); // attribute value
                    attrPos += attrMatch[6].length;
                }
                pos += match[4].length;
                if (pos > attrPos) {
                    pushCommentRange(attrPos, pos - attrPos);
                }
                if (match[5]) {
                    pushClassification(pos, match[5].length, 10 /* punctuation */); // />
                    pos += match[5].length;
                }
                var end = start + width;
                if (pos < end) {
                    pushCommentRange(pos, end - pos);
                }
                return true;
            }
            function processJSDocTemplateTag(tag) {
                for (var _i = 0, _a = tag.getChildren(); _i < _a.length; _i++) {
                    var child = _a[_i];
                    processElement(child);
                }
            }
            function classifyDisabledMergeCode(text, start, end) {
                // Classify the line that the ||||||| or ======= marker is on as a comment.
                // Then just lex all further tokens and add them to the result.
                var i;
                for (i = start; i < end; i++) {
                    if (ts.isLineBreak(text.charCodeAt(i))) {
                        break;
                    }
                }
                pushClassification(start, i - start, 1 /* comment */);
                mergeConflictScanner.setTextPos(i);
                while (mergeConflictScanner.getTextPos() < end) {
                    classifyDisabledCodeToken();
                }
            }
            function classifyDisabledCodeToken() {
                var start = mergeConflictScanner.getTextPos();
                var tokenKind = mergeConflictScanner.scan();
                var end = mergeConflictScanner.getTextPos();
                var type = classifyTokenType(tokenKind);
                if (type) {
                    pushClassification(start, end - start, type);
                }
            }
            /**
             * Returns true if node should be treated as classified and no further processing is required.
             * False will mean that node is not classified and traverse routine should recurse into node contents.
             */
            function tryClassifyNode(node) {
                if (ts.isJSDoc(node)) {
                    return true;
                }
                if (ts.nodeIsMissing(node)) {
                    return true;
                }
                var classifiedElementName = tryClassifyJsxElementName(node);
                if (!ts.isToken(node) && node.kind !== 11 /* JsxText */ && classifiedElementName === undefined) {
                    return false;
                }
                var tokenStart = node.kind === 11 /* JsxText */ ? node.pos : classifyLeadingTriviaAndGetTokenStart(node);
                var tokenWidth = node.end - tokenStart;
                ts.Debug.assert(tokenWidth >= 0);
                if (tokenWidth > 0) {
                    var type = classifiedElementName || classifyTokenType(node.kind, node);
                    if (type) {
                        pushClassification(tokenStart, tokenWidth, type);
                    }
                }
                return true;
            }
            function tryClassifyJsxElementName(token) {
                switch (token.parent && token.parent.kind) {
                    case 278 /* JsxOpeningElement */:
                        if (token.parent.tagName === token) {
                            return 19 /* jsxOpenTagName */;
                        }
                        break;
                    case 279 /* JsxClosingElement */:
                        if (token.parent.tagName === token) {
                            return 20 /* jsxCloseTagName */;
                        }
                        break;
                    case 277 /* JsxSelfClosingElement */:
                        if (token.parent.tagName === token) {
                            return 21 /* jsxSelfClosingTagName */;
                        }
                        break;
                    case 283 /* JsxAttribute */:
                        if (token.parent.name === token) {
                            return 22 /* jsxAttribute */;
                        }
                        break;
                }
                return undefined;
            }
            // for accurate classification, the actual token should be passed in.  however, for
            // cases like 'disabled merge code' classification, we just get the token kind and
            // classify based on that instead.
            function classifyTokenType(tokenKind, token) {
                if (ts.isKeyword(tokenKind)) {
                    return 3 /* keyword */;
                }
                // Special case `<` and `>`: If they appear in a generic context they are punctuation,
                // not operators.
                if (tokenKind === 29 /* LessThanToken */ || tokenKind === 31 /* GreaterThanToken */) {
                    // If the node owning the token has a type argument list or type parameter list, then
                    // we can effectively assume that a '<' and '>' belong to those lists.
                    if (token && ts.getTypeArgumentOrTypeParameterList(token.parent)) {
                        return 10 /* punctuation */;
                    }
                }
                if (ts.isPunctuation(tokenKind)) {
                    if (token) {
                        var parent = token.parent;
                        if (tokenKind === 63 /* EqualsToken */) {
                            // the '=' in a variable declaration is special cased here.
                            if (parent.kind === 252 /* VariableDeclaration */ ||
                                parent.kind === 165 /* PropertyDeclaration */ ||
                                parent.kind === 162 /* Parameter */ ||
                                parent.kind === 283 /* JsxAttribute */) {
                                return 5 /* operator */;
                            }
                        }
                        if (parent.kind === 219 /* BinaryExpression */ ||
                            parent.kind === 217 /* PrefixUnaryExpression */ ||
                            parent.kind === 218 /* PostfixUnaryExpression */ ||
                            parent.kind === 220 /* ConditionalExpression */) {
                            return 5 /* operator */;
                        }
                    }
                    return 10 /* punctuation */;
                }
                else if (tokenKind === 8 /* NumericLiteral */) {
                    return 4 /* numericLiteral */;
                }
                else if (tokenKind === 9 /* BigIntLiteral */) {
                    return 25 /* bigintLiteral */;
                }
                else if (tokenKind === 10 /* StringLiteral */) {
                    return token && token.parent.kind === 283 /* JsxAttribute */ ? 24 /* jsxAttributeStringLiteralValue */ : 6 /* stringLiteral */;
                }
                else if (tokenKind === 13 /* RegularExpressionLiteral */) {
                    // TODO: we should get another classification type for these literals.
                    return 6 /* stringLiteral */;
                }
                else if (ts.isTemplateLiteralKind(tokenKind)) {
                    // TODO (drosen): we should *also* get another classification type for these literals.
                    return 6 /* stringLiteral */;
                }
                else if (tokenKind === 11 /* JsxText */) {
                    return 23 /* jsxText */;
                }
                else if (tokenKind === 79 /* Identifier */) {
                    if (token) {
                        switch (token.parent.kind) {
                            case 255 /* ClassDeclaration */:
                                if (token.parent.name === token) {
                                    return 11 /* className */;
                                }
                                return;
                            case 161 /* TypeParameter */:
                                if (token.parent.name === token) {
                                    return 15 /* typeParameterName */;
                                }
                                return;
                            case 256 /* InterfaceDeclaration */:
                                if (token.parent.name === token) {
                                    return 13 /* interfaceName */;
                                }
                                return;
                            case 258 /* EnumDeclaration */:
                                if (token.parent.name === token) {
                                    return 12 /* enumName */;
                                }
                                return;
                            case 259 /* ModuleDeclaration */:
                                if (token.parent.name === token) {
                                    return 14 /* moduleName */;
                                }
                                return;
                            case 162 /* Parameter */:
                                if (token.parent.name === token) {
                                    return ts.isThisIdentifier(token) ? 3 /* keyword */ : 17 /* parameterName */;
                                }
                                return;
                        }
                    }
                    return 2 /* identifier */;
                }
            }
            function processElement(element) {
                if (!element) {
                    return;
                }
                // Ignore nodes that don't intersect the original span to classify.
                if (ts.decodedTextSpanIntersectsWith(spanStart, spanLength, element.pos, element.getFullWidth())) {
                    checkForClassificationCancellation(cancellationToken, element.kind);
                    for (var _i = 0, _a = element.getChildren(sourceFile); _i < _a.length; _i++) {
                        var child = _a[_i];
                        if (!tryClassifyNode(child)) {
                            // Recurse into our child nodes.
                            processElement(child);
                        }
                    }
                }
            }
        }
        ts.getEncodedSyntacticClassifications = getEncodedSyntacticClassifications;
    })(ts || (ts = {}));
    /** @internal */
    var ts;
    (function (ts) {
        var classifier;
        (function (classifier) {
            var v2020;
            (function (v2020) {
                var TokenEncodingConsts;
                (function (TokenEncodingConsts) {
                    TokenEncodingConsts[TokenEncodingConsts["typeOffset"] = 8] = "typeOffset";
                    TokenEncodingConsts[TokenEncodingConsts["modifierMask"] = 255] = "modifierMask";
                })(TokenEncodingConsts = v2020.TokenEncodingConsts || (v2020.TokenEncodingConsts = {}));
                var TokenType;
                (function (TokenType) {
                    TokenType[TokenType["class"] = 0] = "class";
                    TokenType[TokenType["enum"] = 1] = "enum";
                    TokenType[TokenType["interface"] = 2] = "interface";
                    TokenType[TokenType["namespace"] = 3] = "namespace";
                    TokenType[TokenType["typeParameter"] = 4] = "typeParameter";
                    TokenType[TokenType["type"] = 5] = "type";
                    TokenType[TokenType["parameter"] = 6] = "parameter";
                    TokenType[TokenType["variable"] = 7] = "variable";
                    TokenType[TokenType["enumMember"] = 8] = "enumMember";
                    TokenType[TokenType["property"] = 9] = "property";
                    TokenType[TokenType["function"] = 10] = "function";
                    TokenType[TokenType["member"] = 11] = "member";
                })(TokenType = v2020.TokenType || (v2020.TokenType = {}));
                var TokenModifier;
                (function (TokenModifier) {
                    TokenModifier[TokenModifier["declaration"] = 0] = "declaration";
                    TokenModifier[TokenModifier["static"] = 1] = "static";
                    TokenModifier[TokenModifier["async"] = 2] = "async";
                    TokenModifier[TokenModifier["readonly"] = 3] = "readonly";
                    TokenModifier[TokenModifier["defaultLibrary"] = 4] = "defaultLibrary";
                    TokenModifier[TokenModifier["local"] = 5] = "local";
                })(TokenModifier = v2020.TokenModifier || (v2020.TokenModifier = {}));
                /** This is mainly used internally for testing */
                function getSemanticClassifications(program, cancellationToken, sourceFile, span) {
                    var classifications = getEncodedSemanticClassifications(program, cancellationToken, sourceFile, span);
                    ts.Debug.assert(classifications.spans.length % 3 === 0);
                    var dense = classifications.spans;
                    var result = [];
                    for (var i = 0; i < dense.length; i += 3) {
                        result.push({
                            textSpan: ts.createTextSpan(dense[i], dense[i + 1]),
                            classificationType: dense[i + 2]
                        });
                    }
                    return result;
                }
                v2020.getSemanticClassifications = getSemanticClassifications;
                function getEncodedSemanticClassifications(program, cancellationToken, sourceFile, span) {
                    return {
                        spans: getSemanticTokens(program, sourceFile, span, cancellationToken),
                        endOfLineState: 0 /* None */
                    };
                }
                v2020.getEncodedSemanticClassifications = getEncodedSemanticClassifications;
                function getSemanticTokens(program, sourceFile, span, cancellationToken) {
                    var resultTokens = [];
                    var collector = function (node, typeIdx, modifierSet) {
                        resultTokens.push(node.getStart(sourceFile), node.getWidth(sourceFile), ((typeIdx + 1) << 8 /* typeOffset */) + modifierSet);
                    };
                    if (program && sourceFile) {
                        collectTokens(program, sourceFile, span, collector, cancellationToken);
                    }
                    return resultTokens;
                }
                function collectTokens(program, sourceFile, span, collector, cancellationToken) {
                    var typeChecker = program.getTypeChecker();
                    var inJSXElement = false;
                    function visit(node) {
                        switch (node.kind) {
                            case 259 /* ModuleDeclaration */:
                            case 255 /* ClassDeclaration */:
                            case 256 /* InterfaceDeclaration */:
                            case 254 /* FunctionDeclaration */:
                            case 224 /* ClassExpression */:
                            case 211 /* FunctionExpression */:
                            case 212 /* ArrowFunction */:
                                cancellationToken.throwIfCancellationRequested();
                        }
                        if (!node || !ts.textSpanIntersectsWith(span, node.pos, node.getFullWidth()) || node.getFullWidth() === 0) {
                            return;
                        }
                        var prevInJSXElement = inJSXElement;
                        if (ts.isJsxElement(node) || ts.isJsxSelfClosingElement(node)) {
                            inJSXElement = true;
                        }
                        if (ts.isJsxExpression(node)) {
                            inJSXElement = false;
                        }
                        if (ts.isIdentifier(node) && !inJSXElement && !inImportClause(node) && !isInfinityOrNaNString(node.escapedText)) {
                            var symbol = typeChecker.getSymbolAtLocation(node);
                            if (symbol) {
                                if (symbol.flags & 2097152 /* Alias */) {
                                    symbol = typeChecker.getAliasedSymbol(symbol);
                                }
                                var typeIdx = classifySymbol(symbol, ts.getMeaningFromLocation(node));
                                if (typeIdx !== undefined) {
                                    var modifierSet = 0;
                                    if (node.parent) {
                                        var parentIsDeclaration = (ts.isBindingElement(node.parent) || tokenFromDeclarationMapping.get(node.parent.kind) === typeIdx);
                                        if (parentIsDeclaration && node.parent.name === node) {
                                            modifierSet = 1 << 0 /* declaration */;
                                        }
                                    }
                                    // property declaration in constructor
                                    if (typeIdx === 6 /* parameter */ && isRightSideOfQualifiedNameOrPropertyAccess(node)) {
                                        typeIdx = 9 /* property */;
                                    }
                                    typeIdx = reclassifyByType(typeChecker, node, typeIdx);
                                    var decl = symbol.valueDeclaration;
                                    if (decl) {
                                        var modifiers = ts.getCombinedModifierFlags(decl);
                                        var nodeFlags = ts.getCombinedNodeFlags(decl);
                                        if (modifiers & 32 /* Static */) {
                                            modifierSet |= 1 << 1 /* static */;
                                        }
                                        if (modifiers & 256 /* Async */) {
                                            modifierSet |= 1 << 2 /* async */;
                                        }
                                        if (typeIdx !== 0 /* class */ && typeIdx !== 2 /* interface */) {
                                            if ((modifiers & 64 /* Readonly */) || (nodeFlags & 2 /* Const */) || (symbol.getFlags() & 8 /* EnumMember */)) {
                                                modifierSet |= 1 << 3 /* readonly */;
                                            }
                                        }
                                        if ((typeIdx === 7 /* variable */ || typeIdx === 10 /* function */) && isLocalDeclaration(decl, sourceFile)) {
                                            modifierSet |= 1 << 5 /* local */;
                                        }
                                        if (program.isSourceFileDefaultLibrary(decl.getSourceFile())) {
                                            modifierSet |= 1 << 4 /* defaultLibrary */;
                                        }
                                    }
                                    else if (symbol.declarations && symbol.declarations.some(function (d) { return program.isSourceFileDefaultLibrary(d.getSourceFile()); })) {
                                        modifierSet |= 1 << 4 /* defaultLibrary */;
                                    }
                                    collector(node, typeIdx, modifierSet);
                                }
                            }
                        }
                        ts.forEachChild(node, visit);
                        inJSXElement = prevInJSXElement;
                    }
                    visit(sourceFile);
                }
                function classifySymbol(symbol, meaning) {
                    var flags = symbol.getFlags();
                    if (flags & 32 /* Class */) {
                        return 0 /* class */;
                    }
                    else if (flags & 384 /* Enum */) {
                        return 1 /* enum */;
                    }
                    else if (flags & 524288 /* TypeAlias */) {
                        return 5 /* type */;
                    }
                    else if (flags & 64 /* Interface */) {
                        if (meaning & 2 /* Type */) {
                            return 2 /* interface */;
                        }
                    }
                    else if (flags & 262144 /* TypeParameter */) {
                        return 4 /* typeParameter */;
                    }
                    var decl = symbol.valueDeclaration || symbol.declarations && symbol.declarations[0];
                    if (decl && ts.isBindingElement(decl)) {
                        decl = getDeclarationForBindingElement(decl);
                    }
                    return decl && tokenFromDeclarationMapping.get(decl.kind);
                }
                function reclassifyByType(typeChecker, node, typeIdx) {
                    // type based classifications
                    if (typeIdx === 7 /* variable */ || typeIdx === 9 /* property */ || typeIdx === 6 /* parameter */) {
                        var type_1 = typeChecker.getTypeAtLocation(node);
                        if (type_1) {
                            var test = function (condition) {
                                return condition(type_1) || type_1.isUnion() && type_1.types.some(condition);
                            };
                            if (typeIdx !== 6 /* parameter */ && test(function (t) { return t.getConstructSignatures().length > 0; })) {
                                return 0 /* class */;
                            }
                            if (test(function (t) { return t.getCallSignatures().length > 0; }) && !test(function (t) { return t.getProperties().length > 0; }) || isExpressionInCallExpression(node)) {
                                return typeIdx === 9 /* property */ ? 11 /* member */ : 10 /* function */;
                            }
                        }
                    }
                    return typeIdx;
                }
                function isLocalDeclaration(decl, sourceFile) {
                    if (ts.isBindingElement(decl)) {
                        decl = getDeclarationForBindingElement(decl);
                    }
                    if (ts.isVariableDeclaration(decl)) {
                        return (!ts.isSourceFile(decl.parent.parent.parent) || ts.isCatchClause(decl.parent)) && decl.getSourceFile() === sourceFile;
                    }
                    else if (ts.isFunctionDeclaration(decl)) {
                        return !ts.isSourceFile(decl.parent) && decl.getSourceFile() === sourceFile;
                    }
                    return false;
                }
                function getDeclarationForBindingElement(element) {
                    while (true) {
                        if (ts.isBindingElement(element.parent.parent)) {
                            element = element.parent.parent;
                        }
                        else {
                            return element.parent.parent;
                        }
                    }
                }
                function inImportClause(node) {
                    var parent = node.parent;
                    return parent && (ts.isImportClause(parent) || ts.isImportSpecifier(parent) || ts.isNamespaceImport(parent));
                }
                function isExpressionInCallExpression(node) {
                    while (isRightSideOfQualifiedNameOrPropertyAccess(node)) {
                        node = node.parent;
                    }
                    return ts.isCallExpression(node.parent) && node.parent.expression === node;
                }
                function isRightSideOfQualifiedNameOrPropertyAccess(node) {
                    return (ts.isQualifiedName(node.parent) && node.parent.right === node) || (ts.isPropertyAccessExpression(node.parent) && node.parent.name === node);
                }
                function isInfinityOrNaNString(name) {
                    return name === "Infinity" || name === "NaN";
                }
                var tokenFromDeclarationMapping = new ts.Map([
                    [252 /* VariableDeclaration */, 7 /* variable */],
                    [162 /* Parameter */, 6 /* parameter */],
                    [165 /* PropertyDeclaration */, 9 /* property */],
                    [259 /* ModuleDeclaration */, 3 /* namespace */],
                    [258 /* EnumDeclaration */, 1 /* enum */],
                    [294 /* EnumMember */, 8 /* enumMember */],
                    [255 /* ClassDeclaration */, 0 /* class */],
                    [167 /* MethodDeclaration */, 11 /* member */],
                    [254 /* FunctionDeclaration */, 10 /* function */],
                    [211 /* FunctionExpression */, 10 /* function */],
                    [166 /* MethodSignature */, 11 /* member */],
                    [170 /* GetAccessor */, 9 /* property */],
                    [171 /* SetAccessor */, 9 /* property */],
                    [164 /* PropertySignature */, 9 /* property */],
                    [256 /* InterfaceDeclaration */, 2 /* interface */],
                    [257 /* TypeAliasDeclaration */, 5 /* type */],
                    [161 /* TypeParameter */, 4 /* typeParameter */],
                    [291 /* PropertyAssignment */, 9 /* property */],
                    [292 /* ShorthandPropertyAssignment */, 9 /* property */]
                ]);
            })(v2020 = classifier.v2020 || (classifier.v2020 = {}));
        })(classifier = ts.classifier || (ts.classifier = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var Completions;
        (function (Completions) {
            var StringCompletions;
            (function (StringCompletions) {
                function getStringLiteralCompletions(sourceFile, position, contextToken, checker, options, host, log, preferences) {
                    if (ts.isInReferenceComment(sourceFile, position)) {
                        var entries = getTripleSlashReferenceCompletion(sourceFile, position, options, host);
                        return entries && convertPathCompletions(entries);
                    }
                    if (ts.isInString(sourceFile, position, contextToken)) {
                        if (!contextToken || !ts.isStringLiteralLike(contextToken))
                            return undefined;
                        var entries = getStringLiteralCompletionEntries(sourceFile, contextToken, position, checker, options, host, preferences);
                        return convertStringLiteralCompletions(entries, contextToken, sourceFile, checker, log, options, preferences);
                    }
                }
                StringCompletions.getStringLiteralCompletions = getStringLiteralCompletions;
                function convertStringLiteralCompletions(completion, contextToken, sourceFile, checker, log, options, preferences) {
                    if (completion === undefined) {
                        return undefined;
                    }
                    var optionalReplacementSpan = ts.createTextSpanFromStringLiteralLikeContent(contextToken);
                    switch (completion.kind) {
                        case 0 /* Paths */:
                            return convertPathCompletions(completion.paths);
                        case 1 /* Properties */: {
                            var entries = [];
                            Completions.getCompletionEntriesFromSymbols(completion.symbols, entries, contextToken, sourceFile, sourceFile, checker, 99 /* ESNext */, log, 4 /* String */, preferences, options); // Target will not be used, so arbitrary
                            return { isGlobalCompletion: false, isMemberCompletion: true, isNewIdentifierLocation: completion.hasIndexSignature, optionalReplacementSpan: optionalReplacementSpan, entries: entries };
                        }
                        case 2 /* Types */: {
                            var entries = completion.types.map(function (type) {
                                return ({
                                    name: type.value,
                                    kindModifiers: "" /* none */,
                                    kind: "string" /* string */,
                                    sortText: Completions.SortText.LocationPriority,
                                    replacementSpan: ts.getReplacementSpanForContextToken(contextToken)
                                });
                            });
                            return { isGlobalCompletion: false, isMemberCompletion: false, isNewIdentifierLocation: completion.isNewIdentifier, optionalReplacementSpan: optionalReplacementSpan, entries: entries };
                        }
                        default:
                            return ts.Debug.assertNever(completion);
                    }
                }
                function getStringLiteralCompletionDetails(name, sourceFile, position, contextToken, checker, options, host, cancellationToken, preferences) {
                    if (!contextToken || !ts.isStringLiteralLike(contextToken))
                        return undefined;
                    var completions = getStringLiteralCompletionEntries(sourceFile, contextToken, position, checker, options, host, preferences);
                    return completions && stringLiteralCompletionDetails(name, contextToken, completions, sourceFile, checker, cancellationToken);
                }
                StringCompletions.getStringLiteralCompletionDetails = getStringLiteralCompletionDetails;
                function stringLiteralCompletionDetails(name, location, completion, sourceFile, checker, cancellationToken) {
                    switch (completion.kind) {
                        case 0 /* Paths */: {
                            var match = ts.find(completion.paths, function (p) { return p.name === name; });
                            return match && Completions.createCompletionDetails(name, kindModifiersFromExtension(match.extension), match.kind, [ts.textPart(name)]);
                        }
                        case 1 /* Properties */: {
                            var match = ts.find(completion.symbols, function (s) { return s.name === name; });
                            return match && Completions.createCompletionDetailsForSymbol(match, checker, sourceFile, location, cancellationToken);
                        }
                        case 2 /* Types */:
                            return ts.find(completion.types, function (t) { return t.value === name; }) ? Completions.createCompletionDetails(name, "" /* none */, "type" /* typeElement */, [ts.textPart(name)]) : undefined;
                        default:
                            return ts.Debug.assertNever(completion);
                    }
                }
                function convertPathCompletions(pathCompletions) {
                    var isGlobalCompletion = false; // We don't want the editor to offer any other completions, such as snippets, inside a comment.
                    var isNewIdentifierLocation = true; // The user may type in a path that doesn't yet exist, creating a "new identifier" with respect to the collection of identifiers the server is aware of.
                    var entries = pathCompletions.map(function (_a) {
                        var name = _a.name, kind = _a.kind, span = _a.span, extension = _a.extension;
                        return ({ name: name, kind: kind, kindModifiers: kindModifiersFromExtension(extension), sortText: Completions.SortText.LocationPriority, replacementSpan: span });
                    });
                    return { isGlobalCompletion: isGlobalCompletion, isMemberCompletion: false, isNewIdentifierLocation: isNewIdentifierLocation, entries: entries };
                }
                function kindModifiersFromExtension(extension) {
                    switch (extension) {
                        case ".d.ts" /* Dts */: return ".d.ts" /* dtsModifier */;
                        case ".js" /* Js */: return ".js" /* jsModifier */;
                        case ".json" /* Json */: return ".json" /* jsonModifier */;
                        case ".jsx" /* Jsx */: return ".jsx" /* jsxModifier */;
                        case ".ts" /* Ts */: return ".ts" /* tsModifier */;
                        case ".tsx" /* Tsx */: return ".tsx" /* tsxModifier */;
                        case ".tsbuildinfo" /* TsBuildInfo */: return ts.Debug.fail("Extension " + ".tsbuildinfo" /* TsBuildInfo */ + " is unsupported.");
                        case undefined: return "" /* none */;
                        default:
                            return ts.Debug.assertNever(extension);
                    }
                }
                var StringLiteralCompletionKind;
                (function (StringLiteralCompletionKind) {
                    StringLiteralCompletionKind[StringLiteralCompletionKind["Paths"] = 0] = "Paths";
                    StringLiteralCompletionKind[StringLiteralCompletionKind["Properties"] = 1] = "Properties";
                    StringLiteralCompletionKind[StringLiteralCompletionKind["Types"] = 2] = "Types";
                })(StringLiteralCompletionKind || (StringLiteralCompletionKind = {}));
                function getStringLiteralCompletionEntries(sourceFile, node, position, typeChecker, compilerOptions, host, preferences) {
                    var parent = walkUpParentheses(node.parent);
                    switch (parent.kind) {
                        case 194 /* LiteralType */: {
                            var grandParent = walkUpParentheses(parent.parent);
                            switch (grandParent.kind) {
                                case 176 /* TypeReference */: {
                                    var typeReference_1 = grandParent;
                                    var typeArgument = ts.findAncestor(parent, function (n) { return n.parent === typeReference_1; });
                                    if (typeArgument) {
                                        return { kind: 2 /* Types */, types: getStringLiteralTypes(typeChecker.getTypeArgumentConstraint(typeArgument)), isNewIdentifier: false };
                                    }
                                    return undefined;
                                }
                                case 192 /* IndexedAccessType */:
                                    // Get all apparent property names
                                    // i.e. interface Foo {
                                    //          foo: string;
                                    //          bar: string;
                                    //      }
                                    //      let x: Foo["/*completion position*/"]
                                    var _a = grandParent, indexType = _a.indexType, objectType = _a.objectType;
                                    if (!ts.rangeContainsPosition(indexType, position)) {
                                        return undefined;
                                    }
                                    return stringLiteralCompletionsFromProperties(typeChecker.getTypeFromTypeNode(objectType));
                                case 198 /* ImportType */:
                                    return { kind: 0 /* Paths */, paths: getStringLiteralCompletionsFromModuleNames(sourceFile, node, compilerOptions, host, typeChecker, preferences) };
                                case 185 /* UnionType */: {
                                    if (!ts.isTypeReferenceNode(grandParent.parent)) {
                                        return undefined;
                                    }
                                    var alreadyUsedTypes_1 = getAlreadyUsedTypesInStringLiteralUnion(grandParent, parent);
                                    var types = getStringLiteralTypes(typeChecker.getTypeArgumentConstraint(grandParent)).filter(function (t) { return !ts.contains(alreadyUsedTypes_1, t.value); });
                                    return { kind: 2 /* Types */, types: types, isNewIdentifier: false };
                                }
                                default:
                                    return undefined;
                            }
                        }
                        case 291 /* PropertyAssignment */:
                            if (ts.isObjectLiteralExpression(parent.parent) && parent.name === node) {
                                // Get quoted name of properties of the object literal expression
                                // i.e. interface ConfigFiles {
                                //          'jspm:dev': string
                                //      }
                                //      let files: ConfigFiles = {
                                //          '/*completion position*/'
                                //      }
                                //
                                //      function foo(c: ConfigFiles) {}
                                //      foo({
                                //          '/*completion position*/'
                                //      });
                                return stringLiteralCompletionsForObjectLiteral(typeChecker, parent.parent);
                            }
                            return fromContextualType();
                        case 205 /* ElementAccessExpression */: {
                            var _b = parent, expression = _b.expression, argumentExpression = _b.argumentExpression;
                            if (node === ts.skipParentheses(argumentExpression)) {
                                // Get all names of properties on the expression
                                // i.e. interface A {
                                //      'prop1': string
                                // }
                                // let a: A;
                                // a['/*completion position*/']
                                return stringLiteralCompletionsFromProperties(typeChecker.getTypeAtLocation(expression));
                            }
                            return undefined;
                        }
                        case 206 /* CallExpression */:
                        case 207 /* NewExpression */:
                            if (!isRequireCallArgument(node) && !ts.isImportCall(parent)) {
                                var argumentInfo = ts.SignatureHelp.getArgumentInfoForCompletions(node, position, sourceFile);
                                // Get string literal completions from specialized signatures of the target
                                // i.e. declare function f(a: 'A');
                                // f("/*completion position*/")
                                return argumentInfo ? getStringLiteralCompletionsFromSignature(argumentInfo, typeChecker) : fromContextualType();
                            }
                        // falls through (is `require("")` or `require(""` or `import("")`)
                        case 264 /* ImportDeclaration */:
                        case 270 /* ExportDeclaration */:
                        case 275 /* ExternalModuleReference */:
                            // Get all known external module names or complete a path to a module
                            // i.e. import * as ns from "/*completion position*/";
                            //      var y = import("/*completion position*/");
                            //      import x = require("/*completion position*/");
                            //      var y = require("/*completion position*/");
                            //      export * from "/*completion position*/";
                            return { kind: 0 /* Paths */, paths: getStringLiteralCompletionsFromModuleNames(sourceFile, node, compilerOptions, host, typeChecker, preferences) };
                        default:
                            return fromContextualType();
                    }
                    function fromContextualType() {
                        // Get completion for string literal from string literal type
                        // i.e. var x: "hi" | "hello" = "/*completion position*/"
                        return { kind: 2 /* Types */, types: getStringLiteralTypes(ts.getContextualTypeFromParent(node, typeChecker)), isNewIdentifier: false };
                    }
                }
                function walkUpParentheses(node) {
                    switch (node.kind) {
                        case 189 /* ParenthesizedType */:
                            return ts.walkUpParenthesizedTypes(node);
                        case 210 /* ParenthesizedExpression */:
                            return ts.walkUpParenthesizedExpressions(node);
                        default:
                            return node;
                    }
                }
                function getAlreadyUsedTypesInStringLiteralUnion(union, current) {
                    return ts.mapDefined(union.types, function (type) {
                        return type !== current && ts.isLiteralTypeNode(type) && ts.isStringLiteral(type.literal) ? type.literal.text : undefined;
                    });
                }
                function getStringLiteralCompletionsFromSignature(argumentInfo, checker) {
                    var isNewIdentifier = false;
                    var uniques = new ts.Map();
                    var candidates = [];
                    checker.getResolvedSignature(argumentInfo.invocation, candidates, argumentInfo.argumentCount);
                    var types = ts.flatMap(candidates, function (candidate) {
                        if (!ts.signatureHasRestParameter(candidate) && argumentInfo.argumentCount > candidate.parameters.length)
                            return;
                        var type = checker.getParameterType(candidate, argumentInfo.argumentIndex);
                        isNewIdentifier = isNewIdentifier || !!(type.flags & 4 /* String */);
                        return getStringLiteralTypes(type, uniques);
                    });
                    return { kind: 2 /* Types */, types: types, isNewIdentifier: isNewIdentifier };
                }
                function stringLiteralCompletionsFromProperties(type) {
                    return type && {
                        kind: 1 /* Properties */,
                        symbols: ts.filter(type.getApparentProperties(), function (prop) { return !(prop.valueDeclaration && ts.isPrivateIdentifierClassElementDeclaration(prop.valueDeclaration)); }),
                        hasIndexSignature: ts.hasIndexSignature(type)
                    };
                }
                function stringLiteralCompletionsForObjectLiteral(checker, objectLiteralExpression) {
                    var contextualType = checker.getContextualType(objectLiteralExpression);
                    if (!contextualType)
                        return undefined;
                    var completionsType = checker.getContextualType(objectLiteralExpression, 4 /* Completions */);
                    var symbols = Completions.getPropertiesForObjectExpression(contextualType, completionsType, objectLiteralExpression, checker);
                    return {
                        kind: 1 /* Properties */,
                        symbols: symbols,
                        hasIndexSignature: ts.hasIndexSignature(contextualType)
                    };
                }
                function getStringLiteralTypes(type, uniques) {
                    if (uniques === void 0) {
                        uniques = new ts.Map();
                    }
                    if (!type)
                        return ts.emptyArray;
                    type = ts.skipConstraint(type);
                    return type.isUnion() ? ts.flatMap(type.types, function (t) { return getStringLiteralTypes(t, uniques); }) :
                        type.isStringLiteral() && !(type.flags & 1024 /* EnumLiteral */) && ts.addToSeen(uniques, type.value) ? [type] : ts.emptyArray;
                }
                function nameAndKind(name, kind, extension) {
                    return { name: name, kind: kind, extension: extension };
                }
                function directoryResult(name) {
                    return nameAndKind(name, "directory" /* directory */, /*extension*/ undefined);
                }
                function addReplacementSpans(text, textStart, names) {
                    var span = getDirectoryFragmentTextSpan(text, textStart);
                    var wholeSpan = text.length === 0 ? undefined : ts.createTextSpan(textStart, text.length);
                    return names.map(function (_a) {
                        var name = _a.name, kind = _a.kind, extension = _a.extension;
                        return Math.max(name.indexOf(ts.directorySeparator), name.indexOf(ts.altDirectorySeparator)) !== -1 ? { name: name, kind: kind, extension: extension, span: wholeSpan } : { name: name, kind: kind, extension: extension, span: span };
                    });
                }
                function getStringLiteralCompletionsFromModuleNames(sourceFile, node, compilerOptions, host, typeChecker, preferences) {
                    return addReplacementSpans(node.text, node.getStart(sourceFile) + 1, getStringLiteralCompletionsFromModuleNamesWorker(sourceFile, node, compilerOptions, host, typeChecker, preferences));
                }
                function getStringLiteralCompletionsFromModuleNamesWorker(sourceFile, node, compilerOptions, host, typeChecker, preferences) {
                    var literalValue = ts.normalizeSlashes(node.text);
                    var scriptPath = sourceFile.path;
                    var scriptDirectory = ts.getDirectoryPath(scriptPath);
                    return isPathRelativeToScript(literalValue) || !compilerOptions.baseUrl && (ts.isRootedDiskPath(literalValue) || ts.isUrl(literalValue))
                        ? getCompletionEntriesForRelativeModules(literalValue, scriptDirectory, compilerOptions, host, scriptPath, preferences)
                        : getCompletionEntriesForNonRelativeModules(literalValue, scriptDirectory, compilerOptions, host, typeChecker);
                }
                function getExtensionOptions(compilerOptions, includeExtensionsOption) {
                    if (includeExtensionsOption === void 0) {
                        includeExtensionsOption = 0 /* Exclude */;
                    }
                    return { extensions: getSupportedExtensionsForModuleResolution(compilerOptions), includeExtensionsOption: includeExtensionsOption };
                }
                function getCompletionEntriesForRelativeModules(literalValue, scriptDirectory, compilerOptions, host, scriptPath, preferences) {
                    var includeExtensions = preferences.importModuleSpecifierEnding === "js" ? 2 /* ModuleSpecifierCompletion */ : 0 /* Exclude */;
                    var extensionOptions = getExtensionOptions(compilerOptions, includeExtensions);
                    if (compilerOptions.rootDirs) {
                        return getCompletionEntriesForDirectoryFragmentWithRootDirs(compilerOptions.rootDirs, literalValue, scriptDirectory, extensionOptions, compilerOptions, host, scriptPath);
                    }
                    else {
                        return getCompletionEntriesForDirectoryFragment(literalValue, scriptDirectory, extensionOptions, host, scriptPath);
                    }
                }
                function getSupportedExtensionsForModuleResolution(compilerOptions) {
                    var extensions = ts.getSupportedExtensions(compilerOptions);
                    return compilerOptions.resolveJsonModule && ts.getEmitModuleResolutionKind(compilerOptions) === ts.ModuleResolutionKind.NodeJs ?
                        extensions.concat(".json" /* Json */) :
                        extensions;
                }
                /**
                 * Takes a script path and returns paths for all potential folders that could be merged with its
                 * containing folder via the "rootDirs" compiler option
                 */
                function getBaseDirectoriesFromRootDirs(rootDirs, basePath, scriptDirectory, ignoreCase) {
                    // Make all paths absolute/normalized if they are not already
                    rootDirs = rootDirs.map(function (rootDirectory) { return ts.normalizePath(ts.isRootedDiskPath(rootDirectory) ? rootDirectory : ts.combinePaths(basePath, rootDirectory)); });
                    // Determine the path to the directory containing the script relative to the root directory it is contained within
                    var relativeDirectory = ts.firstDefined(rootDirs, function (rootDirectory) {
                        return ts.containsPath(rootDirectory, scriptDirectory, basePath, ignoreCase) ? scriptDirectory.substr(rootDirectory.length) : undefined;
                    }); // TODO: GH#18217
                    // Now find a path for each potential directory that is to be merged with the one containing the script
                    return ts.deduplicate(__spreadArray(__spreadArray([], rootDirs.map(function (rootDirectory) { return ts.combinePaths(rootDirectory, relativeDirectory); }), true), [scriptDirectory], false), ts.equateStringsCaseSensitive, ts.compareStringsCaseSensitive);
                }
                function getCompletionEntriesForDirectoryFragmentWithRootDirs(rootDirs, fragment, scriptDirectory, extensionOptions, compilerOptions, host, exclude) {
                    var basePath = compilerOptions.project || host.getCurrentDirectory();
                    var ignoreCase = !(host.useCaseSensitiveFileNames && host.useCaseSensitiveFileNames());
                    var baseDirectories = getBaseDirectoriesFromRootDirs(rootDirs, basePath, scriptDirectory, ignoreCase);
                    return ts.flatMap(baseDirectories, function (baseDirectory) { return getCompletionEntriesForDirectoryFragment(fragment, baseDirectory, extensionOptions, host, exclude); });
                }
                var IncludeExtensionsOption;
                (function (IncludeExtensionsOption) {
                    IncludeExtensionsOption[IncludeExtensionsOption["Exclude"] = 0] = "Exclude";
                    IncludeExtensionsOption[IncludeExtensionsOption["Include"] = 1] = "Include";
                    IncludeExtensionsOption[IncludeExtensionsOption["ModuleSpecifierCompletion"] = 2] = "ModuleSpecifierCompletion";
                })(IncludeExtensionsOption || (IncludeExtensionsOption = {}));
                /**
                 * Given a path ending at a directory, gets the completions for the path, and filters for those entries containing the basename.
                 */
                function getCompletionEntriesForDirectoryFragment(fragment, scriptPath, _a, host, exclude, result) {
                    var extensions = _a.extensions, includeExtensionsOption = _a.includeExtensionsOption;
                    if (result === void 0) {
                        result = [];
                    }
                    if (fragment === undefined) {
                        fragment = "";
                    }
                    fragment = ts.normalizeSlashes(fragment);
                    /**
                     * Remove the basename from the path. Note that we don't use the basename to filter completions;
                     * the client is responsible for refining completions.
                     */
                    if (!ts.hasTrailingDirectorySeparator(fragment)) {
                        fragment = ts.getDirectoryPath(fragment);
                    }
                    if (fragment === "") {
                        fragment = "." + ts.directorySeparator;
                    }
                    fragment = ts.ensureTrailingDirectorySeparator(fragment);
                    // const absolutePath = normalizeAndPreserveTrailingSlash(isRootedDiskPath(fragment) ? fragment : combinePaths(scriptPath, fragment)); // TODO(rbuckton): should use resolvePaths
                    var absolutePath = ts.resolvePath(scriptPath, fragment);
                    var baseDirectory = ts.hasTrailingDirectorySeparator(absolutePath) ? absolutePath : ts.getDirectoryPath(absolutePath);
                    var ignoreCase = !(host.useCaseSensitiveFileNames && host.useCaseSensitiveFileNames());
                    if (!ts.tryDirectoryExists(host, baseDirectory))
                        return result;
                    // Enumerate the available files if possible
                    var files = ts.tryReadDirectory(host, baseDirectory, extensions, /*exclude*/ undefined, /*include*/ ["./*"]);
                    if (files) {
                        /**
                         * Multiple file entries might map to the same truncated name once we remove extensions
                         * (happens iff includeExtensionsOption === includeExtensionsOption.Exclude) so we use a set-like data structure. Eg:
                         *
                         * both foo.ts and foo.tsx become foo
                         */
                        var foundFiles = new ts.Map(); // maps file to its extension
                        for (var _i = 0, files_1 = files; _i < files_1.length; _i++) {
                            var filePath = files_1[_i];
                            filePath = ts.normalizePath(filePath);
                            if (exclude && ts.comparePaths(filePath, exclude, scriptPath, ignoreCase) === 0 /* EqualTo */) {
                                continue;
                            }
                            var foundFileName = void 0;
                            var outputExtension = ts.moduleSpecifiers.tryGetJSExtensionForFile(filePath, host.getCompilationSettings());
                            if (includeExtensionsOption === 0 /* Exclude */ && !ts.fileExtensionIs(filePath, ".json" /* Json */)) {
                                foundFileName = ts.removeFileExtension(ts.getBaseFileName(filePath));
                                foundFiles.set(foundFileName, ts.tryGetExtensionFromPath(filePath));
                            }
                            else if (includeExtensionsOption === 2 /* ModuleSpecifierCompletion */ && outputExtension) {
                                foundFileName = ts.changeExtension(ts.getBaseFileName(filePath), outputExtension);
                                foundFiles.set(foundFileName, outputExtension);
                            }
                            else {
                                foundFileName = ts.getBaseFileName(filePath);
                                foundFiles.set(foundFileName, ts.tryGetExtensionFromPath(filePath));
                            }
                        }
                        foundFiles.forEach(function (ext, foundFile) {
                            result.push(nameAndKind(foundFile, "script" /* scriptElement */, ext));
                        });
                    }
                    // If possible, get folder completion as well
                    var directories = ts.tryGetDirectories(host, baseDirectory);
                    if (directories) {
                        for (var _b = 0, directories_1 = directories; _b < directories_1.length; _b++) {
                            var directory = directories_1[_b];
                            var directoryName = ts.getBaseFileName(ts.normalizePath(directory));
                            if (directoryName !== "@types") {
                                result.push(directoryResult(directoryName));
                            }
                        }
                    }
                    // check for a version redirect
                    var packageJsonPath = ts.findPackageJson(baseDirectory, host);
                    if (packageJsonPath) {
                        var packageJson = ts.readJson(packageJsonPath, host);
                        var typesVersions = packageJson.typesVersions;
                        if (typeof typesVersions === "object") {
                            var versionResult = ts.getPackageJsonTypesVersionsPaths(typesVersions);
                            var versionPaths = versionResult && versionResult.paths;
                            var rest = absolutePath.slice(ts.ensureTrailingDirectorySeparator(baseDirectory).length);
                            if (versionPaths) {
                                addCompletionEntriesFromPaths(result, rest, baseDirectory, extensions, versionPaths, host);
                            }
                        }
                    }
                    return result;
                }
                function addCompletionEntriesFromPaths(result, fragment, baseDirectory, fileExtensions, paths, host) {
                    for (var path in paths) {
                        if (!ts.hasProperty(paths, path))
                            continue;
                        var patterns = paths[path];
                        if (patterns) {
                            var _loop_2 = function (name, kind, extension) {
                                // Path mappings may provide a duplicate way to get to something we've already added, so don't add again.
                                if (!result.some(function (entry) { return entry.name === name; })) {
                                    result.push(nameAndKind(name, kind, extension));
                                }
                            };
                            for (var _i = 0, _a = getCompletionsForPathMapping(path, patterns, fragment, baseDirectory, fileExtensions, host); _i < _a.length; _i++) {
                                var _b = _a[_i], name = _b.name, kind = _b.kind, extension = _b.extension;
                                _loop_2(name, kind, extension);
                            }
                        }
                    }
                }
                /**
                 * Check all of the declared modules and those in node modules. Possible sources of modules:
                 *      Modules that are found by the type checker
                 *      Modules found relative to "baseUrl" compliler options (including patterns from "paths" compiler option)
                 *      Modules from node_modules (i.e. those listed in package.json)
                 *          This includes all files that are found in node_modules/moduleName/ with acceptable file extensions
                 */
                function getCompletionEntriesForNonRelativeModules(fragment, scriptPath, compilerOptions, host, typeChecker) {
                    var baseUrl = compilerOptions.baseUrl, paths = compilerOptions.paths;
                    var result = [];
                    var extensionOptions = getExtensionOptions(compilerOptions);
                    if (baseUrl) {
                        var projectDir = compilerOptions.project || host.getCurrentDirectory();
                        var absolute = ts.normalizePath(ts.combinePaths(projectDir, baseUrl));
                        getCompletionEntriesForDirectoryFragment(fragment, absolute, extensionOptions, host, /*exclude*/ undefined, result);
                        if (paths) {
                            addCompletionEntriesFromPaths(result, fragment, absolute, extensionOptions.extensions, paths, host);
                        }
                    }
                    var fragmentDirectory = getFragmentDirectory(fragment);
                    for (var _i = 0, _a = getAmbientModuleCompletions(fragment, fragmentDirectory, typeChecker); _i < _a.length; _i++) {
                        var ambientName = _a[_i];
                        result.push(nameAndKind(ambientName, "external module name" /* externalModuleName */, /*extension*/ undefined));
                    }
                    getCompletionEntriesFromTypings(host, compilerOptions, scriptPath, fragmentDirectory, extensionOptions, result);
                    if (ts.getEmitModuleResolutionKind(compilerOptions) === ts.ModuleResolutionKind.NodeJs) {
                        // If looking for a global package name, don't just include everything in `node_modules` because that includes dependencies' own dependencies.
                        // (But do if we didn't find anything, e.g. 'package.json' missing.)
                        var foundGlobal = false;
                        if (fragmentDirectory === undefined) {
                            var _loop_3 = function (moduleName) {
                                if (!result.some(function (entry) { return entry.name === moduleName; })) {
                                    foundGlobal = true;
                                    result.push(nameAndKind(moduleName, "external module name" /* externalModuleName */, /*extension*/ undefined));
                                }
                            };
                            for (var _b = 0, _c = enumerateNodeModulesVisibleToScript(host, scriptPath); _b < _c.length; _b++) {
                                var moduleName = _c[_b];
                                _loop_3(moduleName);
                            }
                        }
                        if (!foundGlobal) {
                            ts.forEachAncestorDirectory(scriptPath, function (ancestor) {
                                var nodeModules = ts.combinePaths(ancestor, "node_modules");
                                if (ts.tryDirectoryExists(host, nodeModules)) {
                                    getCompletionEntriesForDirectoryFragment(fragment, nodeModules, extensionOptions, host, /*exclude*/ undefined, result);
                                }
                            });
                        }
                    }
                    return result;
                }
                function getFragmentDirectory(fragment) {
                    return containsSlash(fragment) ? ts.hasTrailingDirectorySeparator(fragment) ? fragment : ts.getDirectoryPath(fragment) : undefined;
                }
                function getCompletionsForPathMapping(path, patterns, fragment, baseUrl, fileExtensions, host) {
                    if (!ts.endsWith(path, "*")) {
                        // For a path mapping "foo": ["/x/y/z.ts"], add "foo" itself as a completion.
                        return !ts.stringContains(path, "*") ? justPathMappingName(path) : ts.emptyArray;
                    }
                    var pathPrefix = path.slice(0, path.length - 1);
                    var remainingFragment = ts.tryRemovePrefix(fragment, pathPrefix);
                    return remainingFragment === undefined ? justPathMappingName(pathPrefix) : ts.flatMap(patterns, function (pattern) {
                        return getModulesForPathsPattern(remainingFragment, baseUrl, pattern, fileExtensions, host);
                    });
                    function justPathMappingName(name) {
                        return ts.startsWith(name, fragment) ? [directoryResult(name)] : ts.emptyArray;
                    }
                }
                function getModulesForPathsPattern(fragment, baseUrl, pattern, fileExtensions, host) {
                    if (!host.readDirectory) {
                        return undefined;
                    }
                    var parsed = ts.tryParsePattern(pattern);
                    if (parsed === undefined || ts.isString(parsed)) {
                        return undefined;
                    }
                    // The prefix has two effective parts: the directory path and the base component after the filepath that is not a
                    // full directory component. For example: directory/path/of/prefix/base*
                    var normalizedPrefix = ts.resolvePath(parsed.prefix);
                    var normalizedPrefixDirectory = ts.hasTrailingDirectorySeparator(parsed.prefix) ? normalizedPrefix : ts.getDirectoryPath(normalizedPrefix);
                    var normalizedPrefixBase = ts.hasTrailingDirectorySeparator(parsed.prefix) ? "" : ts.getBaseFileName(normalizedPrefix);
                    var fragmentHasPath = containsSlash(fragment);
                    var fragmentDirectory = fragmentHasPath ? ts.hasTrailingDirectorySeparator(fragment) ? fragment : ts.getDirectoryPath(fragment) : undefined;
                    // Try and expand the prefix to include any path from the fragment so that we can limit the readDirectory call
                    var expandedPrefixDirectory = fragmentHasPath ? ts.combinePaths(normalizedPrefixDirectory, normalizedPrefixBase + fragmentDirectory) : normalizedPrefixDirectory;
                    var normalizedSuffix = ts.normalizePath(parsed.suffix);
                    // Need to normalize after combining: If we combinePaths("a", "../b"), we want "b" and not "a/../b".
                    var baseDirectory = ts.normalizePath(ts.combinePaths(baseUrl, expandedPrefixDirectory));
                    var completePrefix = fragmentHasPath ? baseDirectory : ts.ensureTrailingDirectorySeparator(baseDirectory) + normalizedPrefixBase;
                    // If we have a suffix, then we need to read the directory all the way down. We could create a glob
                    // that encodes the suffix, but we would have to escape the character "?" which readDirectory
                    // doesn't support. For now, this is safer but slower
                    var includeGlob = normalizedSuffix ? "**/*" : "./*";
                    var matches = ts.mapDefined(ts.tryReadDirectory(host, baseDirectory, fileExtensions, /*exclude*/ undefined, [includeGlob]), function (match) {
                        var extension = ts.tryGetExtensionFromPath(match);
                        var name = trimPrefixAndSuffix(match);
                        return name === undefined ? undefined : nameAndKind(ts.removeFileExtension(name), "script" /* scriptElement */, extension);
                    });
                    var directories = ts.mapDefined(ts.tryGetDirectories(host, baseDirectory).map(function (d) { return ts.combinePaths(baseDirectory, d); }), function (dir) {
                        var name = trimPrefixAndSuffix(dir);
                        return name === undefined ? undefined : directoryResult(name);
                    });
                    return __spreadArray(__spreadArray([], matches, true), directories, true);
                    function trimPrefixAndSuffix(path) {
                        var inner = withoutStartAndEnd(ts.normalizePath(path), completePrefix, normalizedSuffix);
                        return inner === undefined ? undefined : removeLeadingDirectorySeparator(inner);
                    }
                }
                function withoutStartAndEnd(s, start, end) {
                    return ts.startsWith(s, start) && ts.endsWith(s, end) ? s.slice(start.length, s.length - end.length) : undefined;
                }
                function removeLeadingDirectorySeparator(path) {
                    return path[0] === ts.directorySeparator ? path.slice(1) : path;
                }
                function getAmbientModuleCompletions(fragment, fragmentDirectory, checker) {
                    // Get modules that the type checker picked up
                    var ambientModules = checker.getAmbientModules().map(function (sym) { return ts.stripQuotes(sym.name); });
                    var nonRelativeModuleNames = ambientModules.filter(function (moduleName) { return ts.startsWith(moduleName, fragment); });
                    // Nested modules of the form "module-name/sub" need to be adjusted to only return the string
                    // after the last '/' that appears in the fragment because that's where the replacement span
                    // starts
                    if (fragmentDirectory !== undefined) {
                        var moduleNameWithSeparator_1 = ts.ensureTrailingDirectorySeparator(fragmentDirectory);
                        return nonRelativeModuleNames.map(function (nonRelativeModuleName) { return ts.removePrefix(nonRelativeModuleName, moduleNameWithSeparator_1); });
                    }
                    return nonRelativeModuleNames;
                }
                function getTripleSlashReferenceCompletion(sourceFile, position, compilerOptions, host) {
                    var token = ts.getTokenAtPosition(sourceFile, position);
                    var commentRanges = ts.getLeadingCommentRanges(sourceFile.text, token.pos);
                    var range = commentRanges && ts.find(commentRanges, function (commentRange) { return position >= commentRange.pos && position <= commentRange.end; });
                    if (!range) {
                        return undefined;
                    }
                    var text = sourceFile.text.slice(range.pos, position);
                    var match = tripleSlashDirectiveFragmentRegex.exec(text);
                    if (!match) {
                        return undefined;
                    }
                    var prefix = match[1], kind = match[2], toComplete = match[3];
                    var scriptPath = ts.getDirectoryPath(sourceFile.path);
                    var names = kind === "path" ? getCompletionEntriesForDirectoryFragment(toComplete, scriptPath, getExtensionOptions(compilerOptions, 1 /* Include */), host, sourceFile.path)
                        : kind === "types" ? getCompletionEntriesFromTypings(host, compilerOptions, scriptPath, getFragmentDirectory(toComplete), getExtensionOptions(compilerOptions))
                            : ts.Debug.fail();
                    return addReplacementSpans(toComplete, range.pos + prefix.length, names);
                }
                function getCompletionEntriesFromTypings(host, options, scriptPath, fragmentDirectory, extensionOptions, result) {
                    if (result === void 0) {
                        result = [];
                    }
                    // Check for typings specified in compiler options
                    var seen = new ts.Map();
                    var typeRoots = ts.tryAndIgnoreErrors(function () { return ts.getEffectiveTypeRoots(options, host); }) || ts.emptyArray;
                    for (var _i = 0, typeRoots_1 = typeRoots; _i < typeRoots_1.length; _i++) {
                        var root = typeRoots_1[_i];
                        getCompletionEntriesFromDirectories(root);
                    }
                    // Also get all @types typings installed in visible node_modules directories
                    for (var _a = 0, _b = ts.findPackageJsons(scriptPath, host); _a < _b.length; _a++) {
                        var packageJson = _b[_a];
                        var typesDir = ts.combinePaths(ts.getDirectoryPath(packageJson), "node_modules/@types");
                        getCompletionEntriesFromDirectories(typesDir);
                    }
                    return result;
                    function getCompletionEntriesFromDirectories(directory) {
                        if (!ts.tryDirectoryExists(host, directory))
                            return;
                        for (var _i = 0, _a = ts.tryGetDirectories(host, directory); _i < _a.length; _i++) {
                            var typeDirectoryName = _a[_i];
                            var packageName = ts.unmangleScopedPackageName(typeDirectoryName);
                            if (options.types && !ts.contains(options.types, packageName))
                                continue;
                            if (fragmentDirectory === undefined) {
                                if (!seen.has(packageName)) {
                                    result.push(nameAndKind(packageName, "external module name" /* externalModuleName */, /*extension*/ undefined));
                                    seen.set(packageName, true);
                                }
                            }
                            else {
                                var baseDirectory = ts.combinePaths(directory, typeDirectoryName);
                                var remainingFragment = ts.tryRemoveDirectoryPrefix(fragmentDirectory, packageName, ts.hostGetCanonicalFileName(host));
                                if (remainingFragment !== undefined) {
                                    getCompletionEntriesForDirectoryFragment(remainingFragment, baseDirectory, extensionOptions, host, /*exclude*/ undefined, result);
                                }
                            }
                        }
                    }
                }
                function enumerateNodeModulesVisibleToScript(host, scriptPath) {
                    if (!host.readFile || !host.fileExists)
                        return ts.emptyArray;
                    var result = [];
                    for (var _i = 0, _a = ts.findPackageJsons(scriptPath, host); _i < _a.length; _i++) {
                        var packageJson = _a[_i];
                        var contents = ts.readJson(packageJson, host); // Cast to assert that readFile is defined
                        // Provide completions for all non @types dependencies
                        for (var _b = 0, nodeModulesDependencyKeys_1 = nodeModulesDependencyKeys; _b < nodeModulesDependencyKeys_1.length; _b++) {
                            var key = nodeModulesDependencyKeys_1[_b];
                            var dependencies = contents[key];
                            if (!dependencies)
                                continue;
                            for (var dep in dependencies) {
                                if (dependencies.hasOwnProperty(dep) && !ts.startsWith(dep, "@types/")) {
                                    result.push(dep);
                                }
                            }
                        }
                    }
                    return result;
                }
                // Replace everything after the last directory separator that appears
                function getDirectoryFragmentTextSpan(text, textStart) {
                    var index = Math.max(text.lastIndexOf(ts.directorySeparator), text.lastIndexOf(ts.altDirectorySeparator));
                    var offset = index !== -1 ? index + 1 : 0;
                    // If the range is an identifier, span is unnecessary.
                    var length = text.length - offset;
                    return length === 0 || ts.isIdentifierText(text.substr(offset, length), 99 /* ESNext */) ? undefined : ts.createTextSpan(textStart + offset, length);
                }
                // Returns true if the path is explicitly relative to the script (i.e. relative to . or ..)
                function isPathRelativeToScript(path) {
                    if (path && path.length >= 2 && path.charCodeAt(0) === 46 /* dot */) {
                        var slashIndex = path.length >= 3 && path.charCodeAt(1) === 46 /* dot */ ? 2 : 1;
                        var slashCharCode = path.charCodeAt(slashIndex);
                        return slashCharCode === 47 /* slash */ || slashCharCode === 92 /* backslash */;
                    }
                    return false;
                }
                /**
                 * Matches a triple slash reference directive with an incomplete string literal for its path. Used
                 * to determine if the caret is currently within the string literal and capture the literal fragment
                 * for completions.
                 * For example, this matches
                 *
                 * /// <reference path="fragment
                 *
                 * but not
                 *
                 * /// <reference path="fragment"
                 */
                var tripleSlashDirectiveFragmentRegex = /^(\/\/\/\s*<reference\s+(path|types)\s*=\s*(?:'|"))([^\3"]*)$/;
                var nodeModulesDependencyKeys = ["dependencies", "devDependencies", "peerDependencies", "optionalDependencies"];
                function containsSlash(fragment) {
                    return ts.stringContains(fragment, ts.directorySeparator);
                }
                /**
                 * Matches
                 *   require(""
                 *   require("")
                 */
                function isRequireCallArgument(node) {
                    return ts.isCallExpression(node.parent) && ts.firstOrUndefined(node.parent.arguments) === node
                        && ts.isIdentifier(node.parent.expression) && node.parent.expression.escapedText === "require";
                }
            })(StringCompletions = Completions.StringCompletions || (Completions.StringCompletions = {}));
        })(Completions = ts.Completions || (ts.Completions = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var Completions;
        (function (Completions) {
            // Exported only for tests
            Completions.moduleSpecifierResolutionLimit = 100;
            Completions.moduleSpecifierResolutionCacheAttemptLimit = 1000;
            // NOTE: Make sure that each entry has the exact same number of digits
            //       since many implementations will sort by string contents,
            //       where "10" is considered less than "2".
            var SortText;
            (function (SortText) {
                SortText["LocalDeclarationPriority"] = "10";
                SortText["LocationPriority"] = "11";
                SortText["OptionalMember"] = "12";
                SortText["MemberDeclaredBySpreadAssignment"] = "13";
                SortText["SuggestedClassMembers"] = "14";
                SortText["GlobalsOrKeywords"] = "15";
                SortText["AutoImportSuggestions"] = "16";
                SortText["JavascriptIdentifiers"] = "17";
                SortText["DeprecatedLocalDeclarationPriority"] = "18";
                SortText["DeprecatedLocationPriority"] = "19";
                SortText["DeprecatedOptionalMember"] = "20";
                SortText["DeprecatedMemberDeclaredBySpreadAssignment"] = "21";
                SortText["DeprecatedSuggestedClassMembers"] = "22";
                SortText["DeprecatedGlobalsOrKeywords"] = "23";
                SortText["DeprecatedAutoImportSuggestions"] = "24";
            })(SortText = Completions.SortText || (Completions.SortText = {}));
            var SortTextId;
            (function (SortTextId) {
                SortTextId[SortTextId["LocalDeclarationPriority"] = 10] = "LocalDeclarationPriority";
                SortTextId[SortTextId["LocationPriority"] = 11] = "LocationPriority";
                SortTextId[SortTextId["OptionalMember"] = 12] = "OptionalMember";
                SortTextId[SortTextId["MemberDeclaredBySpreadAssignment"] = 13] = "MemberDeclaredBySpreadAssignment";
                SortTextId[SortTextId["SuggestedClassMembers"] = 14] = "SuggestedClassMembers";
                SortTextId[SortTextId["GlobalsOrKeywords"] = 15] = "GlobalsOrKeywords";
                SortTextId[SortTextId["AutoImportSuggestions"] = 16] = "AutoImportSuggestions";
                // Don't use these directly.
                SortTextId[SortTextId["_JavaScriptIdentifiers"] = 17] = "_JavaScriptIdentifiers";
                SortTextId[SortTextId["_DeprecatedStart"] = 18] = "_DeprecatedStart";
                SortTextId[SortTextId["_First"] = 10] = "_First";
                SortTextId[SortTextId["DeprecatedOffset"] = 8] = "DeprecatedOffset";
            })(SortTextId || (SortTextId = {}));
            /**
             * Special values for `CompletionInfo['source']` used to disambiguate
             * completion items with the same `name`. (Each completion item must
             * have a unique name/source combination, because those two fields
             * comprise `CompletionEntryIdentifier` in `getCompletionEntryDetails`.
             *
             * When the completion item is an auto-import suggestion, the source
             * is the module specifier of the suggestion. To avoid collisions,
             * the values here should not be a module specifier we would ever
             * generate for an auto-import.
             */
            var CompletionSource;
            (function (CompletionSource) {
                /** Completions that require `this.` insertion text */
                CompletionSource["ThisProperty"] = "ThisProperty/";
            })(CompletionSource = Completions.CompletionSource || (Completions.CompletionSource = {}));
            var SymbolOriginInfoKind;
            (function (SymbolOriginInfoKind) {
                SymbolOriginInfoKind[SymbolOriginInfoKind["ThisType"] = 1] = "ThisType";
                SymbolOriginInfoKind[SymbolOriginInfoKind["SymbolMember"] = 2] = "SymbolMember";
                SymbolOriginInfoKind[SymbolOriginInfoKind["Export"] = 4] = "Export";
                SymbolOriginInfoKind[SymbolOriginInfoKind["Promise"] = 8] = "Promise";
                SymbolOriginInfoKind[SymbolOriginInfoKind["Nullable"] = 16] = "Nullable";
                SymbolOriginInfoKind[SymbolOriginInfoKind["ResolvedExport"] = 32] = "ResolvedExport";
                SymbolOriginInfoKind[SymbolOriginInfoKind["SymbolMemberNoExport"] = 2] = "SymbolMemberNoExport";
                SymbolOriginInfoKind[SymbolOriginInfoKind["SymbolMemberExport"] = 6] = "SymbolMemberExport";
            })(SymbolOriginInfoKind || (SymbolOriginInfoKind = {}));
            function originIsThisType(origin) {
                return !!(origin.kind & 1 /* ThisType */);
            }
            function originIsSymbolMember(origin) {
                return !!(origin.kind & 2 /* SymbolMember */);
            }
            function originIsExport(origin) {
                return !!(origin && origin.kind & 4 /* Export */);
            }
            function originIsResolvedExport(origin) {
                return !!(origin && origin.kind === 32 /* ResolvedExport */);
            }
            function originIncludesSymbolName(origin) {
                return originIsExport(origin) || originIsResolvedExport(origin);
            }
            function originIsPackageJsonImport(origin) {
                return (originIsExport(origin) || originIsResolvedExport(origin)) && !!origin.isFromPackageJson;
            }
            function originIsPromise(origin) {
                return !!(origin.kind & 8 /* Promise */);
            }
            function originIsNullableMember(origin) {
                return !!(origin.kind & 16 /* Nullable */);
            }
            var KeywordCompletionFilters;
            (function (KeywordCompletionFilters) {
                KeywordCompletionFilters[KeywordCompletionFilters["None"] = 0] = "None";
                KeywordCompletionFilters[KeywordCompletionFilters["All"] = 1] = "All";
                KeywordCompletionFilters[KeywordCompletionFilters["ClassElementKeywords"] = 2] = "ClassElementKeywords";
                KeywordCompletionFilters[KeywordCompletionFilters["InterfaceElementKeywords"] = 3] = "InterfaceElementKeywords";
                KeywordCompletionFilters[KeywordCompletionFilters["ConstructorParameterKeywords"] = 4] = "ConstructorParameterKeywords";
                KeywordCompletionFilters[KeywordCompletionFilters["FunctionLikeBodyKeywords"] = 5] = "FunctionLikeBodyKeywords";
                KeywordCompletionFilters[KeywordCompletionFilters["TypeAssertionKeywords"] = 6] = "TypeAssertionKeywords";
                KeywordCompletionFilters[KeywordCompletionFilters["TypeKeywords"] = 7] = "TypeKeywords";
                KeywordCompletionFilters[KeywordCompletionFilters["Last"] = 7] = "Last";
            })(KeywordCompletionFilters || (KeywordCompletionFilters = {}));
            var GlobalsSearch;
            (function (GlobalsSearch) {
                GlobalsSearch[GlobalsSearch["Continue"] = 0] = "Continue";
                GlobalsSearch[GlobalsSearch["Success"] = 1] = "Success";
                GlobalsSearch[GlobalsSearch["Fail"] = 2] = "Fail";
            })(GlobalsSearch || (GlobalsSearch = {}));
            function resolvingModuleSpecifiers(logPrefix, host, program, sourceFile, preferences, isForImportStatementCompletion, cb) {
                var _a, _b, _c;
                var start = ts.timestamp();
                var resolutionLimitExceeded = false;
                var ambientCount = 0;
                var resolvedCount = 0;
                var resolvedFromCacheCount = 0;
                var cacheAttemptCount = 0;
                var result = cb({ tryResolve: tryResolve, resolutionLimitExceeded: function () { return resolutionLimitExceeded; } });
                var hitRateMessage = cacheAttemptCount ? " (" + (resolvedFromCacheCount / cacheAttemptCount * 100).toFixed(1) + "% hit rate)" : "";
                (_a = host.log) === null || _a === void 0 ? void 0 : _a.call(host, logPrefix + ": resolved " + resolvedCount + " module specifiers, plus " + ambientCount + " ambient and " + resolvedFromCacheCount + " from cache" + hitRateMessage);
                (_b = host.log) === null || _b === void 0 ? void 0 : _b.call(host, logPrefix + ": response is " + (resolutionLimitExceeded ? "incomplete" : "complete"));
                (_c = host.log) === null || _c === void 0 ? void 0 : _c.call(host, logPrefix + ": " + (ts.timestamp() - start));
                return result;
                function tryResolve(exportInfo, isFromAmbientModule) {
                    if (isFromAmbientModule) {
                        var result_1 = ts.codefix.getModuleSpecifierForBestExportInfo(exportInfo, sourceFile, program, host, preferences);
                        if (result_1) {
                            ambientCount++;
                        }
                        return result_1;
                    }
                    var shouldResolveModuleSpecifier = isForImportStatementCompletion || preferences.allowIncompleteCompletions && resolvedCount < Completions.moduleSpecifierResolutionLimit;
                    var shouldGetModuleSpecifierFromCache = !shouldResolveModuleSpecifier && preferences.allowIncompleteCompletions && cacheAttemptCount < Completions.moduleSpecifierResolutionCacheAttemptLimit;
                    var result = (shouldResolveModuleSpecifier || shouldGetModuleSpecifierFromCache)
                        ? ts.codefix.getModuleSpecifierForBestExportInfo(exportInfo, sourceFile, program, host, preferences, shouldGetModuleSpecifierFromCache)
                        : undefined;
                    if (!shouldResolveModuleSpecifier && !shouldGetModuleSpecifierFromCache || shouldGetModuleSpecifierFromCache && !result) {
                        resolutionLimitExceeded = true;
                    }
                    resolvedCount += (result === null || result === void 0 ? void 0 : result.computedWithoutCacheCount) || 0;
                    resolvedFromCacheCount += exportInfo.length - resolvedCount;
                    if (shouldGetModuleSpecifierFromCache) {
                        cacheAttemptCount++;
                    }
                    return result;
                }
            }
            function getCompletionsAtPosition(host, program, log, sourceFile, position, preferences, triggerCharacter, completionKind, cancellationToken) {
                var _a;
                var previousToken = getRelevantTokens(position, sourceFile).previousToken;
                if (triggerCharacter && !ts.isInString(sourceFile, position, previousToken) && !isValidTrigger(sourceFile, triggerCharacter, previousToken, position)) {
                    return undefined;
                }
                if (triggerCharacter === " ") {
                    // `isValidTrigger` ensures we are at `import |`
                    if (preferences.includeCompletionsForImportStatements && preferences.includeCompletionsWithInsertText) {
                        return { isGlobalCompletion: true, isMemberCompletion: false, isNewIdentifierLocation: true, isIncomplete: true, entries: [] };
                    }
                    return undefined;
                }
                // If the request is a continuation of an earlier `isIncomplete` response,
                // we can continue it from the cached previous response.
                var typeChecker = program.getTypeChecker();
                var compilerOptions = program.getCompilerOptions();
                var incompleteCompletionsCache = preferences.allowIncompleteCompletions ? (_a = host.getIncompleteCompletionsCache) === null || _a === void 0 ? void 0 : _a.call(host) : undefined;
                if (incompleteCompletionsCache && completionKind === 3 /* TriggerForIncompleteCompletions */ && previousToken && ts.isIdentifier(previousToken)) {
                    var incompleteContinuation = continuePreviousIncompleteResponse(incompleteCompletionsCache, sourceFile, previousToken, program, host, preferences, cancellationToken);
                    if (incompleteContinuation) {
                        return incompleteContinuation;
                    }
                }
                else {
                    incompleteCompletionsCache === null || incompleteCompletionsCache === void 0 ? void 0 : incompleteCompletionsCache.clear();
                }
                var stringCompletions = Completions.StringCompletions.getStringLiteralCompletions(sourceFile, position, previousToken, typeChecker, compilerOptions, host, log, preferences);
                if (stringCompletions) {
                    return stringCompletions;
                }
                if (previousToken && ts.isBreakOrContinueStatement(previousToken.parent)
                    && (previousToken.kind === 81 /* BreakKeyword */ || previousToken.kind === 86 /* ContinueKeyword */ || previousToken.kind === 79 /* Identifier */)) {
                    return getLabelCompletionAtPosition(previousToken.parent);
                }
                var completionData = getCompletionData(program, log, sourceFile, isUncheckedFile(sourceFile, compilerOptions), position, preferences, /*detailsEntryId*/ undefined, host, cancellationToken);
                if (!completionData) {
                    return undefined;
                }
                switch (completionData.kind) {
                    case 0 /* Data */:
                        var response = completionInfoFromData(sourceFile, typeChecker, compilerOptions, log, completionData, preferences);
                        if (response === null || response === void 0 ? void 0 : response.isIncomplete) {
                            incompleteCompletionsCache === null || incompleteCompletionsCache === void 0 ? void 0 : incompleteCompletionsCache.set(response);
                        }
                        return response;
                    case 1 /* JsDocTagName */:
                        // If the current position is a jsDoc tag name, only tag names should be provided for completion
                        return jsdocCompletionInfo(ts.JsDoc.getJSDocTagNameCompletions());
                    case 2 /* JsDocTag */:
                        // If the current position is a jsDoc tag, only tags should be provided for completion
                        return jsdocCompletionInfo(ts.JsDoc.getJSDocTagCompletions());
                    case 3 /* JsDocParameterName */:
                        return jsdocCompletionInfo(ts.JsDoc.getJSDocParameterNameCompletions(completionData.tag));
                    case 4 /* Keywords */:
                        return specificKeywordCompletionInfo(completionData.keywords);
                    default:
                        return ts.Debug.assertNever(completionData);
                }
            }
            Completions.getCompletionsAtPosition = getCompletionsAtPosition;
            function continuePreviousIncompleteResponse(cache, file, location, program, host, preferences, cancellationToken) {
                var _a;
                var previousResponse = cache.get();
                if (!previousResponse)
                    return undefined;
                var lowerCaseTokenText = location.text.toLowerCase();
                var exportMap = ts.getExportInfoMap(file, host, program, cancellationToken);
                var checker = program.getTypeChecker();
                var autoImportProvider = (_a = host.getPackageJsonAutoImportProvider) === null || _a === void 0 ? void 0 : _a.call(host);
                var autoImportProviderChecker = autoImportProvider === null || autoImportProvider === void 0 ? void 0 : autoImportProvider.getTypeChecker();
                var newEntries = resolvingModuleSpecifiers("continuePreviousIncompleteResponse", host, program, file, preferences, 
                /*isForImportStatementCompletion*/ false, function (context) {
                    var entries = ts.mapDefined(previousResponse.entries, function (entry) {
                        if (!entry.hasAction || !entry.source || !entry.data || entry.data.moduleSpecifier) {
                            // Not an auto import or already resolved; keep as is
                            return entry;
                        }
                        if (!charactersFuzzyMatchInString(entry.name, lowerCaseTokenText)) {
                            // No longer matches typed characters; filter out
                            return undefined;
                        }
                        var _a = ts.Debug.checkDefined(getAutoImportSymbolFromCompletionEntryData(entry.name, entry.data, program, host)), symbol = _a.symbol, origin = _a.origin;
                        var info = exportMap.get(file.path, entry.name, symbol, origin.moduleSymbol.name, origin.isFromPackageJson ? autoImportProviderChecker : checker);
                        var result = info && context.tryResolve(info, !ts.isExternalModuleNameRelative(ts.stripQuotes(origin.moduleSymbol.name)));
                        if (!result)
                            return entry;
                        var newOrigin = __assign(__assign({}, origin), { kind: 32 /* ResolvedExport */, moduleSpecifier: result.moduleSpecifier });
                        // Mutating for performance... feels sketchy but nobody else uses the cache,
                        // so why bother allocating a bunch of new objects?
                        entry.data = originToCompletionEntryData(newOrigin);
                        entry.source = getSourceFromOrigin(newOrigin);
                        entry.sourceDisplay = [ts.textPart(newOrigin.moduleSpecifier)];
                        return entry;
                    });
                    if (!context.resolutionLimitExceeded()) {
                        previousResponse.isIncomplete = undefined;
                    }
                    return entries;
                });
                previousResponse.entries = newEntries;
                return previousResponse;
            }
            function jsdocCompletionInfo(entries) {
                return { isGlobalCompletion: false, isMemberCompletion: false, isNewIdentifierLocation: false, entries: entries };
            }
            function specificKeywordCompletionInfo(keywords) {
                return {
                    isGlobalCompletion: false,
                    isMemberCompletion: false,
                    isNewIdentifierLocation: false,
                    entries: keywords.map(function (k) {
                        return ({
                            name: ts.tokenToString(k),
                            kind: "keyword" /* keyword */,
                            kindModifiers: "" /* none */,
                            sortText: SortText.GlobalsOrKeywords,
                        });
                    }),
                };
            }
            function getOptionalReplacementSpan(location) {
                // StringLiteralLike locations are handled separately in stringCompletions.ts
                return (location === null || location === void 0 ? void 0 : location.kind) === 79 /* Identifier */ ? ts.createTextSpanFromNode(location) : undefined;
            }
            function completionInfoFromData(sourceFile, typeChecker, compilerOptions, log, completionData, preferences) {
                var symbols = completionData.symbols, completionKind = completionData.completionKind, isInSnippetScope = completionData.isInSnippetScope, isNewIdentifierLocation = completionData.isNewIdentifierLocation, location = completionData.location, propertyAccessToConvert = completionData.propertyAccessToConvert, keywordFilters = completionData.keywordFilters, literals = completionData.literals, symbolToOriginInfoMap = completionData.symbolToOriginInfoMap, recommendedCompletion = completionData.recommendedCompletion, isJsxInitializer = completionData.isJsxInitializer, isTypeOnlyLocation = completionData.isTypeOnlyLocation, isJsxIdentifierExpected = completionData.isJsxIdentifierExpected, importCompletionNode = completionData.importCompletionNode, insideJsDocTagTypeExpression = completionData.insideJsDocTagTypeExpression, symbolToSortTextIdMap = completionData.symbolToSortTextIdMap, hasUnresolvedAutoImports = completionData.hasUnresolvedAutoImports;
                // Verify if the file is JSX language variant
                if (ts.getLanguageVariant(sourceFile.scriptKind) === 1 /* JSX */) {
                    var completionInfo = getJsxClosingTagCompletion(location, sourceFile);
                    if (completionInfo) {
                        return completionInfo;
                    }
                }
                var entries = [];
                if (isUncheckedFile(sourceFile, compilerOptions)) {
                    var uniqueNames = getCompletionEntriesFromSymbols(symbols, entries, 
                    /* contextToken */ undefined, location, sourceFile, typeChecker, compilerOptions.target, log, completionKind, preferences, compilerOptions, isTypeOnlyLocation, propertyAccessToConvert, isJsxIdentifierExpected, isJsxInitializer, importCompletionNode, recommendedCompletion, symbolToOriginInfoMap, symbolToSortTextIdMap);
                    getJSCompletionEntries(sourceFile, location.pos, uniqueNames, compilerOptions.target, entries); // TODO: GH#18217
                }
                else {
                    if (!isNewIdentifierLocation && (!symbols || symbols.length === 0) && keywordFilters === 0 /* None */) {
                        return undefined;
                    }
                    getCompletionEntriesFromSymbols(symbols, entries, 
                    /* contextToken */ undefined, location, sourceFile, typeChecker, compilerOptions.target, log, completionKind, preferences, compilerOptions, isTypeOnlyLocation, propertyAccessToConvert, isJsxIdentifierExpected, isJsxInitializer, importCompletionNode, recommendedCompletion, symbolToOriginInfoMap, symbolToSortTextIdMap);
                }
                if (keywordFilters !== 0 /* None */) {
                    var entryNames = new ts.Set(entries.map(function (e) { return e.name; }));
                    for (var _i = 0, _a = getKeywordCompletions(keywordFilters, !insideJsDocTagTypeExpression && ts.isSourceFileJS(sourceFile)); _i < _a.length; _i++) {
                        var keywordEntry = _a[_i];
                        if (!entryNames.has(keywordEntry.name)) {
                            entries.push(keywordEntry);
                        }
                    }
                }
                for (var _b = 0, literals_1 = literals; _b < literals_1.length; _b++) {
                    var literal = literals_1[_b];
                    entries.push(createCompletionEntryForLiteral(sourceFile, preferences, literal));
                }
                return {
                    isGlobalCompletion: isInSnippetScope,
                    isIncomplete: preferences.allowIncompleteCompletions && hasUnresolvedAutoImports ? true : undefined,
                    isMemberCompletion: isMemberCompletionKind(completionKind),
                    isNewIdentifierLocation: isNewIdentifierLocation,
                    optionalReplacementSpan: getOptionalReplacementSpan(location),
                    entries: entries
                };
            }
            function isUncheckedFile(sourceFile, compilerOptions) {
                return ts.isSourceFileJS(sourceFile) && !ts.isCheckJsEnabledForFile(sourceFile, compilerOptions);
            }
            function isMemberCompletionKind(kind) {
                switch (kind) {
                    case 0 /* ObjectPropertyDeclaration */:
                    case 3 /* MemberLike */:
                    case 2 /* PropertyAccess */:
                        return true;
                    default:
                        return false;
                }
            }
            function getJsxClosingTagCompletion(location, sourceFile) {
                // We wanna walk up the tree till we find a JSX closing element
                var jsxClosingElement = ts.findAncestor(location, function (node) {
                    switch (node.kind) {
                        case 279 /* JsxClosingElement */:
                            return true;
                        case 43 /* SlashToken */:
                        case 31 /* GreaterThanToken */:
                        case 79 /* Identifier */:
                        case 204 /* PropertyAccessExpression */:
                            return false;
                        default:
                            return "quit";
                    }
                });
                if (jsxClosingElement) {
                    // In the TypeScript JSX element, if such element is not defined. When users query for completion at closing tag,
                    // instead of simply giving unknown value, the completion will return the tag-name of an associated opening-element.
                    // For example:
                    //     var x = <div> </ /*1*/
                    // The completion list at "1" will contain "div>" with type any
                    // And at `<div> </ /*1*/ >` (with a closing `>`), the completion list will contain "div".
                    // And at property access expressions `<MainComponent.Child> </MainComponent. /*1*/ >` the completion will
                    // return full closing tag with an optional replacement span
                    // For example:
                    //     var x = <MainComponent.Child> </     MainComponent /*1*/  >
                    //     var y = <MainComponent.Child> </   /*2*/   MainComponent >
                    // the completion list at "1" and "2" will contain "MainComponent.Child" with a replacement span of closing tag name
                    var hasClosingAngleBracket = !!ts.findChildOfKind(jsxClosingElement, 31 /* GreaterThanToken */, sourceFile);
                    var tagName = jsxClosingElement.parent.openingElement.tagName;
                    var closingTag = tagName.getText(sourceFile);
                    var fullClosingTag = closingTag + (hasClosingAngleBracket ? "" : ">");
                    var replacementSpan = ts.createTextSpanFromNode(jsxClosingElement.tagName);
                    var entry = {
                        name: fullClosingTag,
                        kind: "class" /* classElement */,
                        kindModifiers: undefined,
                        sortText: SortText.LocationPriority,
                    };
                    return { isGlobalCompletion: false, isMemberCompletion: true, isNewIdentifierLocation: false, optionalReplacementSpan: replacementSpan, entries: [entry] };
                }
                return;
            }
            function getJSCompletionEntries(sourceFile, position, uniqueNames, target, entries) {
                ts.getNameTable(sourceFile).forEach(function (pos, name) {
                    // Skip identifiers produced only from the current location
                    if (pos === position) {
                        return;
                    }
                    var realName = ts.unescapeLeadingUnderscores(name);
                    if (!uniqueNames.has(realName) && ts.isIdentifierText(realName, target)) {
                        uniqueNames.add(realName);
                        entries.push({
                            name: realName,
                            kind: "warning" /* warning */,
                            kindModifiers: "",
                            sortText: SortText.JavascriptIdentifiers,
                            isFromUncheckedFile: true
                        });
                    }
                });
            }
            function completionNameForLiteral(sourceFile, preferences, literal) {
                return typeof literal === "object" ? ts.pseudoBigIntToString(literal) + "n" :
                    ts.isString(literal) ? ts.quote(sourceFile, preferences, literal) : JSON.stringify(literal);
            }
            function createCompletionEntryForLiteral(sourceFile, preferences, literal) {
                return { name: completionNameForLiteral(sourceFile, preferences, literal), kind: "string" /* string */, kindModifiers: "" /* none */, sortText: SortText.LocationPriority };
            }
            function createCompletionEntry(symbol, sortText, contextToken, location, sourceFile, typeChecker, name, needsConvertPropertyAccess, origin, recommendedCompletion, propertyAccessToConvert, isJsxInitializer, importCompletionNode, useSemicolons, options, preferences) {
                var _a;
                var insertText;
                var replacementSpan = ts.getReplacementSpanForContextToken(contextToken);
                var data;
                var isSnippet;
                var sourceDisplay;
                var hasAction;
                var insertQuestionDot = origin && originIsNullableMember(origin);
                var useBraces = origin && originIsSymbolMember(origin) || needsConvertPropertyAccess;
                if (origin && originIsThisType(origin)) {
                    insertText = needsConvertPropertyAccess
                        ? "this" + (insertQuestionDot ? "?." : "") + "[" + quotePropertyName(sourceFile, preferences, name) + "]"
                        : "this" + (insertQuestionDot ? "?." : ".") + name;
                }
                // We should only have needsConvertPropertyAccess if there's a property access to convert. But see #21790.
                // Somehow there was a global with a non-identifier name. Hopefully someone will complain about getting a "foo bar" global completion and provide a repro.
                else if ((useBraces || insertQuestionDot) && propertyAccessToConvert) {
                    insertText = useBraces ? needsConvertPropertyAccess ? "[" + quotePropertyName(sourceFile, preferences, name) + "]" : "[" + name + "]" : name;
                    if (insertQuestionDot || propertyAccessToConvert.questionDotToken) {
                        insertText = "?." + insertText;
                    }
                    var dot = ts.findChildOfKind(propertyAccessToConvert, 24 /* DotToken */, sourceFile) ||
                        ts.findChildOfKind(propertyAccessToConvert, 28 /* QuestionDotToken */, sourceFile);
                    if (!dot) {
                        return undefined;
                    }
                    // If the text after the '.' starts with this name, write over it. Else, add new text.
                    var end = ts.startsWith(name, propertyAccessToConvert.name.text) ? propertyAccessToConvert.name.end : dot.end;
                    replacementSpan = ts.createTextSpanFromBounds(dot.getStart(sourceFile), end);
                }
                if (isJsxInitializer) {
                    if (insertText === undefined)
                        insertText = name;
                    insertText = "{" + insertText + "}";
                    if (typeof isJsxInitializer !== "boolean") {
                        replacementSpan = ts.createTextSpanFromNode(isJsxInitializer, sourceFile);
                    }
                }
                if (origin && originIsPromise(origin) && propertyAccessToConvert) {
                    if (insertText === undefined)
                        insertText = name;
                    var precedingToken = ts.findPrecedingToken(propertyAccessToConvert.pos, sourceFile);
                    var awaitText = "";
                    if (precedingToken && ts.positionIsASICandidate(precedingToken.end, precedingToken.parent, sourceFile)) {
                        awaitText = ";";
                    }
                    awaitText += "(await " + propertyAccessToConvert.expression.getText() + ")";
                    insertText = needsConvertPropertyAccess ? "" + awaitText + insertText : "" + awaitText + (insertQuestionDot ? "?." : ".") + insertText;
                    replacementSpan = ts.createTextSpanFromBounds(propertyAccessToConvert.getStart(sourceFile), propertyAccessToConvert.end);
                }
                if (originIsResolvedExport(origin)) {
                    sourceDisplay = [ts.textPart(origin.moduleSpecifier)];
                    if (importCompletionNode) {
                        (_a = getInsertTextAndReplacementSpanForImportCompletion(name, importCompletionNode, origin, useSemicolons, options, preferences), insertText = _a.insertText, replacementSpan = _a.replacementSpan);
                        isSnippet = preferences.includeCompletionsWithSnippetText ? true : undefined;
                    }
                }
                if (insertText !== undefined && !preferences.includeCompletionsWithInsertText) {
                    return undefined;
                }
                if (originIsExport(origin) || originIsResolvedExport(origin)) {
                    data = originToCompletionEntryData(origin);
                    hasAction = !importCompletionNode;
                }
                // TODO(drosen): Right now we just permit *all* semantic meanings when calling
                // 'getSymbolKind' which is permissible given that it is backwards compatible; but
                // really we should consider passing the meaning for the node so that we don't report
                // that a suggestion for a value is an interface.  We COULD also just do what
                // 'getSymbolModifiers' does, which is to use the first declaration.
                // Use a 'sortText' of 0' so that all symbol completion entries come before any other
                // entries (like JavaScript identifier entries).
                return {
                    name: name,
                    kind: ts.SymbolDisplay.getSymbolKind(typeChecker, symbol, location),
                    kindModifiers: ts.SymbolDisplay.getSymbolModifiers(typeChecker, symbol),
                    sortText: sortText,
                    source: getSourceFromOrigin(origin),
                    hasAction: hasAction ? true : undefined,
                    isRecommended: isRecommendedCompletionMatch(symbol, recommendedCompletion, typeChecker) || undefined,
                    insertText: insertText,
                    replacementSpan: replacementSpan,
                    sourceDisplay: sourceDisplay,
                    isSnippet: isSnippet,
                    isPackageJsonImport: originIsPackageJsonImport(origin) || undefined,
                    isImportStatementCompletion: !!importCompletionNode || undefined,
                    data: data,
                };
            }
            function originToCompletionEntryData(origin) {
                return {
                    exportName: origin.exportName,
                    fileName: origin.fileName,
                    ambientModuleName: origin.fileName ? undefined : ts.stripQuotes(origin.moduleSymbol.name),
                    isPackageJsonImport: origin.isFromPackageJson ? true : undefined,
                    moduleSpecifier: originIsResolvedExport(origin) ? origin.moduleSpecifier : undefined,
                };
            }
            function getInsertTextAndReplacementSpanForImportCompletion(name, importCompletionNode, origin, useSemicolons, options, preferences) {
                var sourceFile = importCompletionNode.getSourceFile();
                var replacementSpan = ts.createTextSpanFromNode(importCompletionNode, sourceFile);
                var quotedModuleSpecifier = ts.quote(sourceFile, preferences, origin.moduleSpecifier);
                var exportKind = origin.isDefaultExport ? 1 /* Default */ :
                    origin.exportName === "export=" /* ExportEquals */ ? 2 /* ExportEquals */ :
                        0 /* Named */;
                var tabStop = preferences.includeCompletionsWithSnippetText ? "$1" : "";
                var importKind = ts.codefix.getImportKind(sourceFile, exportKind, options, /*forceImportKeyword*/ true);
                var suffix = useSemicolons ? ";" : "";
                switch (importKind) {
                    case 3 /* CommonJS */: return { replacementSpan: replacementSpan, insertText: "import " + name + tabStop + " = require(" + quotedModuleSpecifier + ")" + suffix };
                    case 1 /* Default */: return { replacementSpan: replacementSpan, insertText: "import " + name + tabStop + " from " + quotedModuleSpecifier + suffix };
                    case 2 /* Namespace */: return { replacementSpan: replacementSpan, insertText: "import * as " + name + " from " + quotedModuleSpecifier + suffix };
                    case 0 /* Named */: return { replacementSpan: replacementSpan, insertText: "import { " + name + tabStop + " } from " + quotedModuleSpecifier + suffix };
                }
            }
            function quotePropertyName(sourceFile, preferences, name) {
                if (/^\d+$/.test(name)) {
                    return name;
                }
                return ts.quote(sourceFile, preferences, name);
            }
            function isRecommendedCompletionMatch(localSymbol, recommendedCompletion, checker) {
                return localSymbol === recommendedCompletion ||
                    !!(localSymbol.flags & 1048576 /* ExportValue */) && checker.getExportSymbolOfSymbol(localSymbol) === recommendedCompletion;
            }
            function getSourceFromOrigin(origin) {
                if (originIsExport(origin)) {
                    return ts.stripQuotes(origin.moduleSymbol.name);
                }
                if (originIsResolvedExport(origin)) {
                    return origin.moduleSpecifier;
                }
                if ((origin === null || origin === void 0 ? void 0 : origin.kind) === 1 /* ThisType */) {
                    return CompletionSource.ThisProperty;
                }
            }
            function getCompletionEntriesFromSymbols(symbols, entries, contextToken, location, sourceFile, typeChecker, target, log, kind, preferences, compilerOptions, isTypeOnlyLocation, propertyAccessToConvert, jsxIdentifierExpected, isJsxInitializer, importCompletionNode, recommendedCompletion, symbolToOriginInfoMap, symbolToSortTextIdMap) {
                var _a;
                var start = ts.timestamp();
                var variableDeclaration = getVariableDeclaration(location);
                var useSemicolons = ts.probablyUsesSemicolons(sourceFile);
                // Tracks unique names.
                // Value is set to false for global variables or completions from external module exports, because we can have multiple of those;
                // true otherwise. Based on the order we add things we will always see locals first, then globals, then module exports.
                // So adding a completion for a local will prevent us from adding completions for external module exports sharing the same name.
                var uniques = new ts.Map();
                for (var i = 0; i < symbols.length; i++) {
                    var symbol = symbols[i];
                    var origin = symbolToOriginInfoMap === null || symbolToOriginInfoMap === void 0 ? void 0 : symbolToOriginInfoMap[i];
                    var info = getCompletionEntryDisplayNameForSymbol(symbol, target, origin, kind, !!jsxIdentifierExpected);
                    if (!info || uniques.get(info.name) || kind === 1 /* Global */ && symbolToSortTextIdMap && !shouldIncludeSymbol(symbol, symbolToSortTextIdMap)) {
                        continue;
                    }
                    var name = info.name, needsConvertPropertyAccess = info.needsConvertPropertyAccess;
                    var sortTextId = (_a = symbolToSortTextIdMap === null || symbolToSortTextIdMap === void 0 ? void 0 : symbolToSortTextIdMap[ts.getSymbolId(symbol)]) !== null && _a !== void 0 ? _a : 11 /* LocationPriority */;
                    var sortText = (isDeprecated(symbol, typeChecker) ? 8 /* DeprecatedOffset */ + sortTextId : sortTextId).toString();
                    var entry = createCompletionEntry(symbol, sortText, contextToken, location, sourceFile, typeChecker, name, needsConvertPropertyAccess, origin, recommendedCompletion, propertyAccessToConvert, isJsxInitializer, importCompletionNode, useSemicolons, compilerOptions, preferences);
                    if (!entry) {
                        continue;
                    }
                    /** True for locals; false for globals, module exports from other files, `this.` completions. */
                    var shouldShadowLaterSymbols = !origin && !(symbol.parent === undefined && !ts.some(symbol.declarations, function (d) { return d.getSourceFile() === location.getSourceFile(); }));
                    uniques.set(name, shouldShadowLaterSymbols);
                    entries.push(entry);
                }
                log("getCompletionsAtPosition: getCompletionEntriesFromSymbols: " + (ts.timestamp() - start));
                // Prevent consumers of this map from having to worry about
                // the boolean value. Externally, it should be seen as the
                // set of all names.
                return {
                    has: function (name) { return uniques.has(name); },
                    add: function (name) { return uniques.set(name, true); },
                };
                function shouldIncludeSymbol(symbol, symbolToSortTextIdMap) {
                    if (!ts.isSourceFile(location)) {
                        // export = /**/ here we want to get all meanings, so any symbol is ok
                        if (ts.isExportAssignment(location.parent)) {
                            return true;
                        }
                        // Filter out variables from their own initializers
                        // `const a = /* no 'a' here */`
                        if (variableDeclaration && symbol.valueDeclaration === variableDeclaration) {
                            return false;
                        }
                        // External modules can have global export declarations that will be
                        // available as global keywords in all scopes. But if the external module
                        // already has an explicit export and user only wants to user explicit
                        // module imports then the global keywords will be filtered out so auto
                        // import suggestions will win in the completion
                        var symbolOrigin = ts.skipAlias(symbol, typeChecker);
                        // We only want to filter out the global keywords
                        // Auto Imports are not available for scripts so this conditional is always false
                        if (!!sourceFile.externalModuleIndicator
                            && !compilerOptions.allowUmdGlobalAccess
                            && symbolToSortTextIdMap[ts.getSymbolId(symbol)] === 15 /* GlobalsOrKeywords */
                            && (symbolToSortTextIdMap[ts.getSymbolId(symbolOrigin)] === 16 /* AutoImportSuggestions */
                                || symbolToSortTextIdMap[ts.getSymbolId(symbolOrigin)] === 11 /* LocationPriority */)) {
                            return false;
                        }
                        // Continue with origin symbol
                        symbol = symbolOrigin;
                        // import m = /**/ <-- It can only access namespace (if typing import = x. this would get member symbols and not namespace)
                        if (ts.isInRightSideOfInternalImportEqualsDeclaration(location)) {
                            return !!(symbol.flags & 1920 /* Namespace */);
                        }
                        if (isTypeOnlyLocation) {
                            // It's a type, but you can reach it by namespace.type as well
                            return symbolCanBeReferencedAtTypeLocation(symbol, typeChecker);
                        }
                    }
                    // expressions are value space (which includes the value namespaces)
                    return !!(ts.getCombinedLocalAndExportSymbolFlags(symbol) & 111551 /* Value */);
                }
            }
            Completions.getCompletionEntriesFromSymbols = getCompletionEntriesFromSymbols;
            function getLabelCompletionAtPosition(node) {
                var entries = getLabelStatementCompletions(node);
                if (entries.length) {
                    return { isGlobalCompletion: false, isMemberCompletion: false, isNewIdentifierLocation: false, entries: entries };
                }
            }
            function getLabelStatementCompletions(node) {
                var entries = [];
                var uniques = new ts.Map();
                var current = node;
                while (current) {
                    if (ts.isFunctionLike(current)) {
                        break;
                    }
                    if (ts.isLabeledStatement(current)) {
                        var name = current.label.text;
                        if (!uniques.has(name)) {
                            uniques.set(name, true);
                            entries.push({
                                name: name,
                                kindModifiers: "" /* none */,
                                kind: "label" /* label */,
                                sortText: SortText.LocationPriority
                            });
                        }
                    }
                    current = current.parent;
                }
                return entries;
            }
            function getSymbolCompletionFromEntryId(program, log, sourceFile, position, entryId, host, preferences) {
                if (entryId.data) {
                    var autoImport = getAutoImportSymbolFromCompletionEntryData(entryId.name, entryId.data, program, host);
                    if (autoImport) {
                        return {
                            type: "symbol",
                            symbol: autoImport.symbol,
                            location: ts.getTouchingPropertyName(sourceFile, position),
                            previousToken: ts.findPrecedingToken(position, sourceFile, /*startNode*/ undefined),
                            isJsxInitializer: false,
                            isTypeOnlyLocation: false,
                            origin: autoImport.origin,
                        };
                    }
                }
                var compilerOptions = program.getCompilerOptions();
                var completionData = getCompletionData(program, log, sourceFile, isUncheckedFile(sourceFile, compilerOptions), position, { includeCompletionsForModuleExports: true, includeCompletionsWithInsertText: true }, entryId, host);
                if (!completionData) {
                    return { type: "none" };
                }
                if (completionData.kind !== 0 /* Data */) {
                    return { type: "request", request: completionData };
                }
                var symbols = completionData.symbols, literals = completionData.literals, location = completionData.location, completionKind = completionData.completionKind, symbolToOriginInfoMap = completionData.symbolToOriginInfoMap, previousToken = completionData.previousToken, isJsxInitializer = completionData.isJsxInitializer, isTypeOnlyLocation = completionData.isTypeOnlyLocation;
                var literal = ts.find(literals, function (l) { return completionNameForLiteral(sourceFile, preferences, l) === entryId.name; });
                if (literal !== undefined)
                    return { type: "literal", literal: literal };
                // Find the symbol with the matching entry name.
                // We don't need to perform character checks here because we're only comparing the
                // name against 'entryName' (which is known to be good), not building a new
                // completion entry.
                return ts.firstDefined(symbols, function (symbol, index) {
                    var origin = symbolToOriginInfoMap[index];
                    var info = getCompletionEntryDisplayNameForSymbol(symbol, compilerOptions.target, origin, completionKind, completionData.isJsxIdentifierExpected);
                    return info && info.name === entryId.name && getSourceFromOrigin(origin) === entryId.source
                        ? { type: "symbol", symbol: symbol, location: location, origin: origin, previousToken: previousToken, isJsxInitializer: isJsxInitializer, isTypeOnlyLocation: isTypeOnlyLocation }
                        : undefined;
                }) || { type: "none" };
            }
            function getCompletionEntryDetails(program, log, sourceFile, position, entryId, host, formatContext, preferences, cancellationToken) {
                var typeChecker = program.getTypeChecker();
                var compilerOptions = program.getCompilerOptions();
                var name = entryId.name;
                var contextToken = ts.findPrecedingToken(position, sourceFile);
                if (ts.isInString(sourceFile, position, contextToken)) {
                    return Completions.StringCompletions.getStringLiteralCompletionDetails(name, sourceFile, position, contextToken, typeChecker, compilerOptions, host, cancellationToken, preferences);
                }
                // Compute all the completion symbols again.
                var symbolCompletion = getSymbolCompletionFromEntryId(program, log, sourceFile, position, entryId, host, preferences);
                switch (symbolCompletion.type) {
                    case "request": {
                        var request = symbolCompletion.request;
                        switch (request.kind) {
                            case 1 /* JsDocTagName */:
                                return ts.JsDoc.getJSDocTagNameCompletionDetails(name);
                            case 2 /* JsDocTag */:
                                return ts.JsDoc.getJSDocTagCompletionDetails(name);
                            case 3 /* JsDocParameterName */:
                                return ts.JsDoc.getJSDocParameterNameCompletionDetails(name);
                            case 4 /* Keywords */:
                                return request.keywords.indexOf(ts.stringToToken(name)) > -1 ? createSimpleDetails(name, "keyword" /* keyword */, ts.SymbolDisplayPartKind.keyword) : undefined;
                            default:
                                return ts.Debug.assertNever(request);
                        }
                    }
                    case "symbol": {
                        var symbol = symbolCompletion.symbol, location = symbolCompletion.location, origin = symbolCompletion.origin, previousToken = symbolCompletion.previousToken;
                        var _a = getCompletionEntryCodeActionsAndSourceDisplay(origin, symbol, program, host, compilerOptions, sourceFile, position, previousToken, formatContext, preferences, entryId.data), codeActions = _a.codeActions, sourceDisplay = _a.sourceDisplay;
                        return createCompletionDetailsForSymbol(symbol, typeChecker, sourceFile, location, cancellationToken, codeActions, sourceDisplay); // TODO: GH#18217
                    }
                    case "literal": {
                        var literal = symbolCompletion.literal;
                        return createSimpleDetails(completionNameForLiteral(sourceFile, preferences, literal), "string" /* string */, typeof literal === "string" ? ts.SymbolDisplayPartKind.stringLiteral : ts.SymbolDisplayPartKind.numericLiteral);
                    }
                    case "none":
                        // Didn't find a symbol with this name.  See if we can find a keyword instead.
                        return allKeywordsCompletions().some(function (c) { return c.name === name; }) ? createSimpleDetails(name, "keyword" /* keyword */, ts.SymbolDisplayPartKind.keyword) : undefined;
                    default:
                        ts.Debug.assertNever(symbolCompletion);
                }
            }
            Completions.getCompletionEntryDetails = getCompletionEntryDetails;
            function createSimpleDetails(name, kind, kind2) {
                return createCompletionDetails(name, "" /* none */, kind, [ts.displayPart(name, kind2)]);
            }
            function createCompletionDetailsForSymbol(symbol, checker, sourceFile, location, cancellationToken, codeActions, sourceDisplay) {
                var _a = checker.runWithCancellationToken(cancellationToken, function (checker) {
                    return ts.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(checker, symbol, sourceFile, location, location, 7 /* All */);
                }), displayParts = _a.displayParts, documentation = _a.documentation, symbolKind = _a.symbolKind, tags = _a.tags;
                return createCompletionDetails(symbol.name, ts.SymbolDisplay.getSymbolModifiers(checker, symbol), symbolKind, displayParts, documentation, tags, codeActions, sourceDisplay);
            }
            Completions.createCompletionDetailsForSymbol = createCompletionDetailsForSymbol;
            function createCompletionDetails(name, kindModifiers, kind, displayParts, documentation, tags, codeActions, source) {
                return { name: name, kindModifiers: kindModifiers, kind: kind, displayParts: displayParts, documentation: documentation, tags: tags, codeActions: codeActions, source: source, sourceDisplay: source };
            }
            Completions.createCompletionDetails = createCompletionDetails;
            function getCompletionEntryCodeActionsAndSourceDisplay(origin, symbol, program, host, compilerOptions, sourceFile, position, previousToken, formatContext, preferences, data) {
                if (data === null || data === void 0 ? void 0 : data.moduleSpecifier) {
                    var _a = getRelevantTokens(position, sourceFile), contextToken = _a.contextToken, previousToken_1 = _a.previousToken;
                    if (previousToken_1 && getImportCompletionNode(contextToken || previousToken_1)) {
                        // Import statement completion: 'import c|'
                        return { codeActions: undefined, sourceDisplay: [ts.textPart(data.moduleSpecifier)] };
                    }
                }
                if (!origin || !(originIsExport(origin) || originIsResolvedExport(origin))) {
                    return { codeActions: undefined, sourceDisplay: undefined };
                }
                var checker = origin.isFromPackageJson ? host.getPackageJsonAutoImportProvider().getTypeChecker() : program.getTypeChecker();
                var moduleSymbol = origin.moduleSymbol;
                var exportedSymbol = checker.getMergedSymbol(ts.skipAlias(symbol.exportSymbol || symbol, checker));
                var _b = ts.codefix.getImportCompletionAction(exportedSymbol, moduleSymbol, sourceFile, ts.getNameForExportedSymbol(symbol, compilerOptions.target), host, program, formatContext, previousToken && ts.isIdentifier(previousToken) ? previousToken.getStart(sourceFile) : position, preferences), moduleSpecifier = _b.moduleSpecifier, codeAction = _b.codeAction;
                ts.Debug.assert(!(data === null || data === void 0 ? void 0 : data.moduleSpecifier) || moduleSpecifier === data.moduleSpecifier);
                return { sourceDisplay: [ts.textPart(moduleSpecifier)], codeActions: [codeAction] };
            }
            function getCompletionEntrySymbol(program, log, sourceFile, position, entryId, host, preferences) {
                var completion = getSymbolCompletionFromEntryId(program, log, sourceFile, position, entryId, host, preferences);
                return completion.type === "symbol" ? completion.symbol : undefined;
            }
            Completions.getCompletionEntrySymbol = getCompletionEntrySymbol;
            var CompletionDataKind;
            (function (CompletionDataKind) {
                CompletionDataKind[CompletionDataKind["Data"] = 0] = "Data";
                CompletionDataKind[CompletionDataKind["JsDocTagName"] = 1] = "JsDocTagName";
                CompletionDataKind[CompletionDataKind["JsDocTag"] = 2] = "JsDocTag";
                CompletionDataKind[CompletionDataKind["JsDocParameterName"] = 3] = "JsDocParameterName";
                CompletionDataKind[CompletionDataKind["Keywords"] = 4] = "Keywords";
            })(CompletionDataKind || (CompletionDataKind = {}));
            var CompletionKind;
            (function (CompletionKind) {
                CompletionKind[CompletionKind["ObjectPropertyDeclaration"] = 0] = "ObjectPropertyDeclaration";
                CompletionKind[CompletionKind["Global"] = 1] = "Global";
                CompletionKind[CompletionKind["PropertyAccess"] = 2] = "PropertyAccess";
                CompletionKind[CompletionKind["MemberLike"] = 3] = "MemberLike";
                CompletionKind[CompletionKind["String"] = 4] = "String";
                CompletionKind[CompletionKind["None"] = 5] = "None";
            })(CompletionKind = Completions.CompletionKind || (Completions.CompletionKind = {}));
            function getRecommendedCompletion(previousToken, contextualType, checker) {
                // For a union, return the first one with a recommended completion.
                return ts.firstDefined(contextualType && (contextualType.isUnion() ? contextualType.types : [contextualType]), function (type) {
                    var symbol = type && type.symbol;
                    // Don't include make a recommended completion for an abstract class
                    return symbol && (symbol.flags & (8 /* EnumMember */ | 384 /* Enum */ | 32 /* Class */) && !ts.isAbstractConstructorSymbol(symbol))
                        ? getFirstSymbolInChain(symbol, previousToken, checker)
                        : undefined;
                });
            }
            function getContextualType(previousToken, position, sourceFile, checker) {
                var parent = previousToken.parent;
                switch (previousToken.kind) {
                    case 79 /* Identifier */:
                        return ts.getContextualTypeFromParent(previousToken, checker);
                    case 63 /* EqualsToken */:
                        switch (parent.kind) {
                            case 252 /* VariableDeclaration */:
                                return checker.getContextualType(parent.initializer); // TODO: GH#18217
                            case 219 /* BinaryExpression */:
                                return checker.getTypeAtLocation(parent.left);
                            case 283 /* JsxAttribute */:
                                return checker.getContextualTypeForJsxAttribute(parent);
                            default:
                                return undefined;
                        }
                    case 103 /* NewKeyword */:
                        return checker.getContextualType(parent);
                    case 82 /* CaseKeyword */:
                        return ts.getSwitchedType(ts.cast(parent, ts.isCaseClause), checker);
                    case 18 /* OpenBraceToken */:
                        return ts.isJsxExpression(parent) && parent.parent.kind !== 276 /* JsxElement */ ? checker.getContextualTypeForJsxAttribute(parent.parent) : undefined;
                    default:
                        var argInfo = ts.SignatureHelp.getArgumentInfoForCompletions(previousToken, position, sourceFile);
                        return argInfo ?
                            // At `,`, treat this as the next argument after the comma.
                            checker.getContextualTypeForArgumentAtIndex(argInfo.invocation, argInfo.argumentIndex + (previousToken.kind === 27 /* CommaToken */ ? 1 : 0)) :
                            ts.isEqualityOperatorKind(previousToken.kind) && ts.isBinaryExpression(parent) && ts.isEqualityOperatorKind(parent.operatorToken.kind) ?
                                // completion at `x ===/**/` should be for the right side
                                checker.getTypeAtLocation(parent.left) :
                                checker.getContextualType(previousToken);
                }
            }
            function getFirstSymbolInChain(symbol, enclosingDeclaration, checker) {
                var chain = checker.getAccessibleSymbolChain(symbol, enclosingDeclaration, /*meaning*/ 67108863 /* All */, /*useOnlyExternalAliasing*/ false);
                if (chain)
                    return ts.first(chain);
                return symbol.parent && (isModuleSymbol(symbol.parent) ? symbol : getFirstSymbolInChain(symbol.parent, enclosingDeclaration, checker));
            }
            function isModuleSymbol(symbol) {
                var _a;
                return !!((_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.some(function (d) { return d.kind === 300 /* SourceFile */; }));
            }
            function getCompletionData(program, log, sourceFile, isUncheckedFile, position, preferences, detailsEntryId, host, cancellationToken) {
                var typeChecker = program.getTypeChecker();
                var start = ts.timestamp();
                var currentToken = ts.getTokenAtPosition(sourceFile, position); // TODO: GH#15853
                // We will check for jsdoc comments with insideComment and getJsDocTagAtPosition. (TODO: that seems rather inefficient to check the same thing so many times.)
                log("getCompletionData: Get current token: " + (ts.timestamp() - start));
                start = ts.timestamp();
                var insideComment = ts.isInComment(sourceFile, position, currentToken);
                log("getCompletionData: Is inside comment: " + (ts.timestamp() - start));
                var insideJsDocTagTypeExpression = false;
                var isInSnippetScope = false;
                if (insideComment) {
                    if (ts.hasDocComment(sourceFile, position)) {
                        if (sourceFile.text.charCodeAt(position - 1) === 64 /* at */) {
                            // The current position is next to the '@' sign, when no tag name being provided yet.
                            // Provide a full list of tag names
                            return { kind: 1 /* JsDocTagName */ };
                        }
                        else {
                            // When completion is requested without "@", we will have check to make sure that
                            // there are no comments prefix the request position. We will only allow "*" and space.
                            // e.g
                            //   /** |c| /*
                            //
                            //   /**
                            //     |c|
                            //    */
                            //
                            //   /**
                            //    * |c|
                            //    */
                            //
                            //   /**
                            //    *         |c|
                            //    */
                            var lineStart = ts.getLineStartPositionForPosition(position, sourceFile);
                            if (!/[^\*|\s(/)]/.test(sourceFile.text.substring(lineStart, position))) {
                                return { kind: 2 /* JsDocTag */ };
                            }
                        }
                    }
                    // Completion should work inside certain JsDoc tags. For example:
                    //     /** @type {number | string} */
                    // Completion should work in the brackets
                    var tag = getJsDocTagAtPosition(currentToken, position);
                    if (tag) {
                        if (tag.tagName.pos <= position && position <= tag.tagName.end) {
                            return { kind: 1 /* JsDocTagName */ };
                        }
                        if (isTagWithTypeExpression(tag) && tag.typeExpression && tag.typeExpression.kind === 304 /* JSDocTypeExpression */) {
                            currentToken = ts.getTokenAtPosition(sourceFile, position);
                            if (!currentToken ||
                                (!ts.isDeclarationName(currentToken) &&
                                    (currentToken.parent.kind !== 342 /* JSDocPropertyTag */ ||
                                        currentToken.parent.name !== currentToken))) {
                                // Use as type location if inside tag's type expression
                                insideJsDocTagTypeExpression = isCurrentlyEditingNode(tag.typeExpression);
                            }
                        }
                        if (!insideJsDocTagTypeExpression && ts.isJSDocParameterTag(tag) && (ts.nodeIsMissing(tag.name) || tag.name.pos <= position && position <= tag.name.end)) {
                            return { kind: 3 /* JsDocParameterName */, tag: tag };
                        }
                    }
                    if (!insideJsDocTagTypeExpression) {
                        // Proceed if the current position is in jsDoc tag expression; otherwise it is a normal
                        // comment or the plain text part of a jsDoc comment, so no completion should be available
                        log("Returning an empty list because completion was inside a regular comment or plain text part of a JsDoc comment.");
                        return undefined;
                    }
                }
                start = ts.timestamp();
                // The decision to provide completion depends on the contextToken, which is determined through the previousToken.
                // Note: 'previousToken' (and thus 'contextToken') can be undefined if we are the beginning of the file
                var tokens = getRelevantTokens(position, sourceFile);
                var previousToken = tokens.previousToken;
                var contextToken = tokens.contextToken;
                log("getCompletionData: Get previous token: " + (ts.timestamp() - start));
                // Find the node where completion is requested on.
                // Also determine whether we are trying to complete with members of that node
                // or attributes of a JSX tag.
                var node = currentToken;
                var propertyAccessToConvert;
                var isRightOfDot = false;
                var isRightOfQuestionDot = false;
                var isRightOfOpenTag = false;
                var isStartingCloseTag = false;
                var isJsxInitializer = false;
                var isJsxIdentifierExpected = false;
                var importCompletionNode;
                var location = ts.getTouchingPropertyName(sourceFile, position);
                if (contextToken) {
                    var importCompletionCandidate = getImportCompletionNode(contextToken);
                    if (importCompletionCandidate === 154 /* FromKeyword */) {
                        return { kind: 4 /* Keywords */, keywords: [154 /* FromKeyword */] };
                    }
                    // Import statement completions use `insertText`, and also require the `data` property of `CompletionEntryIdentifier`
                    // added in TypeScript 4.3 to be sent back from the client during `getCompletionEntryDetails`. Since this feature
                    // is not backward compatible with older clients, the language service defaults to disabling it, allowing newer clients
                    // to opt in with the `includeCompletionsForImportStatements` user preference.
                    if (importCompletionCandidate && preferences.includeCompletionsForImportStatements && preferences.includeCompletionsWithInsertText) {
                        importCompletionNode = importCompletionCandidate;
                    }
                    // Bail out if this is a known invalid completion location
                    if (!importCompletionNode && isCompletionListBlocker(contextToken)) {
                        log("Returning an empty list because completion was requested in an invalid position.");
                        return undefined;
                    }
                    var parent = contextToken.parent;
                    if (contextToken.kind === 24 /* DotToken */ || contextToken.kind === 28 /* QuestionDotToken */) {
                        isRightOfDot = contextToken.kind === 24 /* DotToken */;
                        isRightOfQuestionDot = contextToken.kind === 28 /* QuestionDotToken */;
                        switch (parent.kind) {
                            case 204 /* PropertyAccessExpression */:
                                propertyAccessToConvert = parent;
                                node = propertyAccessToConvert.expression;
                                if ((ts.isCallExpression(node) || ts.isFunctionLike(node)) &&
                                    node.end === contextToken.pos &&
                                    node.getChildCount(sourceFile) &&
                                    ts.last(node.getChildren(sourceFile)).kind !== 21 /* CloseParenToken */) {
                                    // This is likely dot from incorrectly parsed expression and user is starting to write spread
                                    // eg: Math.min(./**/)
                                    // const x = function (./**/) {}
                                    return undefined;
                                }
                                break;
                            case 159 /* QualifiedName */:
                                node = parent.left;
                                break;
                            case 259 /* ModuleDeclaration */:
                                node = parent.name;
                                break;
                            case 198 /* ImportType */:
                                node = parent;
                                break;
                            case 229 /* MetaProperty */:
                                node = parent.getFirstToken(sourceFile);
                                ts.Debug.assert(node.kind === 100 /* ImportKeyword */ || node.kind === 103 /* NewKeyword */);
                                break;
                            default:
                                // There is nothing that precedes the dot, so this likely just a stray character
                                // or leading into a '...' token. Just bail out instead.
                                return undefined;
                        }
                    }
                    else if (!importCompletionNode && sourceFile.languageVariant === 1 /* JSX */) {
                        // <UI.Test /* completion position */ />
                        // If the tagname is a property access expression, we will then walk up to the top most of property access expression.
                        // Then, try to get a JSX container and its associated attributes type.
                        if (parent && parent.kind === 204 /* PropertyAccessExpression */) {
                            contextToken = parent;
                            parent = parent.parent;
                        }
                        // Fix location
                        if (currentToken.parent === location) {
                            switch (currentToken.kind) {
                                case 31 /* GreaterThanToken */:
                                    if (currentToken.parent.kind === 276 /* JsxElement */ || currentToken.parent.kind === 278 /* JsxOpeningElement */) {
                                        location = currentToken;
                                    }
                                    break;
                                case 43 /* SlashToken */:
                                    if (currentToken.parent.kind === 277 /* JsxSelfClosingElement */) {
                                        location = currentToken;
                                    }
                                    break;
                            }
                        }
                        switch (parent.kind) {
                            case 279 /* JsxClosingElement */:
                                if (contextToken.kind === 43 /* SlashToken */) {
                                    isStartingCloseTag = true;
                                    location = contextToken;
                                }
                                break;
                            case 219 /* BinaryExpression */:
                                if (!binaryExpressionMayBeOpenTag(parent)) {
                                    break;
                                }
                            // falls through
                            case 277 /* JsxSelfClosingElement */:
                            case 276 /* JsxElement */:
                            case 278 /* JsxOpeningElement */:
                                isJsxIdentifierExpected = true;
                                if (contextToken.kind === 29 /* LessThanToken */) {
                                    isRightOfOpenTag = true;
                                    location = contextToken;
                                }
                                break;
                            case 286 /* JsxExpression */:
                            case 285 /* JsxSpreadAttribute */:
                                // For `<div foo={true} [||] ></div>`, `parent` will be `{true}` and `previousToken` will be `}`
                                if (previousToken.kind === 19 /* CloseBraceToken */ && currentToken.kind === 31 /* GreaterThanToken */) {
                                    isJsxIdentifierExpected = true;
                                }
                                break;
                            case 283 /* JsxAttribute */:
                                // For `<div className="x" [||] ></div>`, `parent` will be JsxAttribute and `previousToken` will be its initializer
                                if (parent.initializer === previousToken &&
                                    previousToken.end < position) {
                                    isJsxIdentifierExpected = true;
                                    break;
                                }
                                switch (previousToken.kind) {
                                    case 63 /* EqualsToken */:
                                        isJsxInitializer = true;
                                        break;
                                    case 79 /* Identifier */:
                                        isJsxIdentifierExpected = true;
                                        // For `<div x=[|f/**/|]`, `parent` will be `x` and `previousToken.parent` will be `f` (which is its own JsxAttribute)
                                        // Note for `<div someBool f>` we don't want to treat this as a jsx inializer, instead it's the attribute name.
                                        if (parent !== previousToken.parent &&
                                            !parent.initializer &&
                                            ts.findChildOfKind(parent, 63 /* EqualsToken */, sourceFile)) {
                                            isJsxInitializer = previousToken;
                                        }
                                }
                                break;
                        }
                    }
                }
                var semanticStart = ts.timestamp();
                var completionKind = 5 /* None */;
                var isNewIdentifierLocation = false;
                var isNonContextualObjectLiteral = false;
                var hasUnresolvedAutoImports = false;
                var keywordFilters = 0 /* None */;
                // This also gets mutated in nested-functions after the return
                var symbols = [];
                var symbolToOriginInfoMap = [];
                var symbolToSortTextIdMap = [];
                var seenPropertySymbols = new ts.Map();
                var isTypeOnlyLocation = isTypeOnlyCompletion();
                var getModuleSpecifierResolutionHost = ts.memoizeOne(function (isFromPackageJson) {
                    return ts.createModuleSpecifierResolutionHost(isFromPackageJson ? host.getPackageJsonAutoImportProvider() : program, host);
                });
                if (isRightOfDot || isRightOfQuestionDot) {
                    getTypeScriptMemberSymbols();
                }
                else if (isRightOfOpenTag) {
                    symbols = typeChecker.getJsxIntrinsicTagNamesAt(location);
                    ts.Debug.assertEachIsDefined(symbols, "getJsxIntrinsicTagNames() should all be defined");
                    tryGetGlobalSymbols();
                    completionKind = 1 /* Global */;
                    keywordFilters = 0 /* None */;
                }
                else if (isStartingCloseTag) {
                    var tagName = contextToken.parent.parent.openingElement.tagName;
                    var tagSymbol = typeChecker.getSymbolAtLocation(tagName);
                    if (tagSymbol) {
                        symbols = [tagSymbol];
                    }
                    completionKind = 1 /* Global */;
                    keywordFilters = 0 /* None */;
                }
                else {
                    // For JavaScript or TypeScript, if we're not after a dot, then just try to get the
                    // global symbols in scope.  These results should be valid for either language as
                    // the set of symbols that can be referenced from this location.
                    if (!tryGetGlobalSymbols()) {
                        return undefined;
                    }
                }
                log("getCompletionData: Semantic work: " + (ts.timestamp() - semanticStart));
                var contextualType = previousToken && getContextualType(previousToken, position, sourceFile, typeChecker);
                var literals = ts.mapDefined(contextualType && (contextualType.isUnion() ? contextualType.types : [contextualType]), function (t) { return t.isLiteral() ? t.value : undefined; });
                var recommendedCompletion = previousToken && contextualType && getRecommendedCompletion(previousToken, contextualType, typeChecker);
                return {
                    kind: 0 /* Data */,
                    symbols: symbols,
                    completionKind: completionKind,
                    isInSnippetScope: isInSnippetScope,
                    propertyAccessToConvert: propertyAccessToConvert,
                    isNewIdentifierLocation: isNewIdentifierLocation,
                    location: location,
                    keywordFilters: keywordFilters,
                    literals: literals,
                    symbolToOriginInfoMap: symbolToOriginInfoMap,
                    recommendedCompletion: recommendedCompletion,
                    previousToken: previousToken,
                    isJsxInitializer: isJsxInitializer,
                    insideJsDocTagTypeExpression: insideJsDocTagTypeExpression,
                    symbolToSortTextIdMap: symbolToSortTextIdMap,
                    isTypeOnlyLocation: isTypeOnlyLocation,
                    isJsxIdentifierExpected: isJsxIdentifierExpected,
                    importCompletionNode: importCompletionNode,
                    hasUnresolvedAutoImports: hasUnresolvedAutoImports,
                };
                function isTagWithTypeExpression(tag) {
                    switch (tag.kind) {
                        case 335 /* JSDocParameterTag */:
                        case 342 /* JSDocPropertyTag */:
                        case 336 /* JSDocReturnTag */:
                        case 338 /* JSDocTypeTag */:
                        case 340 /* JSDocTypedefTag */:
                            return true;
                        default:
                            return false;
                    }
                }
                function getTypeScriptMemberSymbols() {
                    // Right of dot member completion list
                    completionKind = 2 /* PropertyAccess */;
                    // Since this is qualified name check it's a type node location
                    var isImportType = ts.isLiteralImportTypeNode(node);
                    var isTypeLocation = insideJsDocTagTypeExpression
                        || (isImportType && !node.isTypeOf)
                        || ts.isPartOfTypeNode(node.parent)
                        || ts.isPossiblyTypeArgumentPosition(contextToken, sourceFile, typeChecker);
                    var isRhsOfImportDeclaration = ts.isInRightSideOfInternalImportEqualsDeclaration(node);
                    if (ts.isEntityName(node) || isImportType || ts.isPropertyAccessExpression(node)) {
                        var isNamespaceName = ts.isModuleDeclaration(node.parent);
                        if (isNamespaceName)
                            isNewIdentifierLocation = true;
                        var symbol = typeChecker.getSymbolAtLocation(node);
                        if (symbol) {
                            symbol = ts.skipAlias(symbol, typeChecker);
                            if (symbol.flags & (1536 /* Module */ | 384 /* Enum */)) {
                                // Extract module or enum members
                                var exportedSymbols = typeChecker.getExportsOfModule(symbol);
                                ts.Debug.assertEachIsDefined(exportedSymbols, "getExportsOfModule() should all be defined");
                                var isValidValueAccess_1 = function (symbol) { return typeChecker.isValidPropertyAccess(isImportType ? node : node.parent, symbol.name); };
                                var isValidTypeAccess_1 = function (symbol) { return symbolCanBeReferencedAtTypeLocation(symbol, typeChecker); };
                                var isValidAccess = isNamespaceName
                                    // At `namespace N.M/**/`, if this is the only declaration of `M`, don't include `M` as a completion.
                                    ? function (symbol) { var _a; return !!(symbol.flags & 1920 /* Namespace */) && !((_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.every(function (d) { return d.parent === node.parent; })); }
                                    : isRhsOfImportDeclaration ?
                                        // Any kind is allowed when dotting off namespace in internal import equals declaration
                                        function (symbol) { return isValidTypeAccess_1(symbol) || isValidValueAccess_1(symbol); } :
                                        isTypeLocation ? isValidTypeAccess_1 : isValidValueAccess_1;
                                for (var _i = 0, exportedSymbols_1 = exportedSymbols; _i < exportedSymbols_1.length; _i++) {
                                    var exportedSymbol = exportedSymbols_1[_i];
                                    if (isValidAccess(exportedSymbol)) {
                                        symbols.push(exportedSymbol);
                                    }
                                }
                                // If the module is merged with a value, we must get the type of the class and add its propertes (for inherited static methods).
                                if (!isTypeLocation &&
                                    symbol.declarations &&
                                    symbol.declarations.some(function (d) { return d.kind !== 300 /* SourceFile */ && d.kind !== 259 /* ModuleDeclaration */ && d.kind !== 258 /* EnumDeclaration */; })) {
                                    var type = typeChecker.getTypeOfSymbolAtLocation(symbol, node).getNonOptionalType();
                                    var insertQuestionDot = false;
                                    if (type.isNullableType()) {
                                        var canCorrectToQuestionDot = isRightOfDot &&
                                            !isRightOfQuestionDot &&
                                            preferences.includeAutomaticOptionalChainCompletions !== false;
                                        if (canCorrectToQuestionDot || isRightOfQuestionDot) {
                                            type = type.getNonNullableType();
                                            if (canCorrectToQuestionDot) {
                                                insertQuestionDot = true;
                                            }
                                        }
                                    }
                                    addTypeProperties(type, !!(node.flags & 32768 /* AwaitContext */), insertQuestionDot);
                                }
                                return;
                            }
                        }
                    }
                    if (!isTypeLocation) {
                        // GH#39946. Pulling on the type of a node inside of a function with a contextual `this` parameter can result in a circularity
                        // if the `node` is part of the exprssion of a `yield` or `return`. This circularity doesn't exist at compile time because
                        // we will check (and cache) the type of `this` *before* checking the type of the node.
                        var container = ts.getThisContainer(node, /*includeArrowFunctions*/ false);
                        if (!ts.isSourceFile(container) && container.parent)
                            typeChecker.getTypeAtLocation(container);
                        var type = typeChecker.getTypeAtLocation(node).getNonOptionalType();
                        var insertQuestionDot = false;
                        if (type.isNullableType()) {
                            var canCorrectToQuestionDot = isRightOfDot &&
                                !isRightOfQuestionDot &&
                                preferences.includeAutomaticOptionalChainCompletions !== false;
                            if (canCorrectToQuestionDot || isRightOfQuestionDot) {
                                type = type.getNonNullableType();
                                if (canCorrectToQuestionDot) {
                                    insertQuestionDot = true;
                                }
                            }
                        }
                        addTypeProperties(type, !!(node.flags & 32768 /* AwaitContext */), insertQuestionDot);
                    }
                }
                function addTypeProperties(type, insertAwait, insertQuestionDot) {
                    isNewIdentifierLocation = !!type.getStringIndexType();
                    if (isRightOfQuestionDot && ts.some(type.getCallSignatures())) {
                        isNewIdentifierLocation = true;
                    }
                    var propertyAccess = node.kind === 198 /* ImportType */ ? node : node.parent;
                    if (isUncheckedFile) {
                        // In javascript files, for union types, we don't just get the members that
                        // the individual types have in common, we also include all the members that
                        // each individual type has. This is because we're going to add all identifiers
                        // anyways. So we might as well elevate the members that were at least part
                        // of the individual types to a higher status since we know what they are.
                        symbols.push.apply(symbols, ts.filter(getPropertiesForCompletion(type, typeChecker), function (s) { return typeChecker.isValidPropertyAccessForCompletions(propertyAccess, type, s); }));
                    }
                    else {
                        for (var _i = 0, _a = type.getApparentProperties(); _i < _a.length; _i++) {
                            var symbol = _a[_i];
                            if (typeChecker.isValidPropertyAccessForCompletions(propertyAccess, type, symbol)) {
                                addPropertySymbol(symbol, /* insertAwait */ false, insertQuestionDot);
                            }
                        }
                    }
                    if (insertAwait && preferences.includeCompletionsWithInsertText) {
                        var promiseType = typeChecker.getPromisedTypeOfPromise(type);
                        if (promiseType) {
                            for (var _b = 0, _c = promiseType.getApparentProperties(); _b < _c.length; _b++) {
                                var symbol = _c[_b];
                                if (typeChecker.isValidPropertyAccessForCompletions(propertyAccess, promiseType, symbol)) {
                                    addPropertySymbol(symbol, /* insertAwait */ true, insertQuestionDot);
                                }
                            }
                        }
                    }
                }
                function addPropertySymbol(symbol, insertAwait, insertQuestionDot) {
                    // For a computed property with an accessible name like `Symbol.iterator`,
                    // we'll add a completion for the *name* `Symbol` instead of for the property.
                    // If this is e.g. [Symbol.iterator], add a completion for `Symbol`.
                    var computedPropertyName = ts.firstDefined(symbol.declarations, function (decl) { return ts.tryCast(ts.getNameOfDeclaration(decl), ts.isComputedPropertyName); });
                    if (computedPropertyName) {
                        var leftMostName = getLeftMostName(computedPropertyName.expression); // The completion is for `Symbol`, not `iterator`.
                        var nameSymbol = leftMostName && typeChecker.getSymbolAtLocation(leftMostName);
                        // If this is nested like for `namespace N { export const sym = Symbol(); }`, we'll add the completion for `N`.
                        var firstAccessibleSymbol = nameSymbol && getFirstSymbolInChain(nameSymbol, contextToken, typeChecker);
                        if (firstAccessibleSymbol && ts.addToSeen(seenPropertySymbols, ts.getSymbolId(firstAccessibleSymbol))) {
                            var index = symbols.length;
                            symbols.push(firstAccessibleSymbol);
                            var moduleSymbol = firstAccessibleSymbol.parent;
                            if (!moduleSymbol || !ts.isExternalModuleSymbol(moduleSymbol)) {
                                symbolToOriginInfoMap[index] = { kind: getNullableSymbolOriginInfoKind(2 /* SymbolMemberNoExport */) };
                            }
                            else {
                                var origin = {
                                    kind: getNullableSymbolOriginInfoKind(6 /* SymbolMemberExport */),
                                    moduleSymbol: moduleSymbol,
                                    isDefaultExport: false,
                                    symbolName: firstAccessibleSymbol.name,
                                    exportName: firstAccessibleSymbol.name,
                                    fileName: ts.isExternalModuleNameRelative(ts.stripQuotes(moduleSymbol.name)) ? ts.cast(moduleSymbol.valueDeclaration, ts.isSourceFile).fileName : undefined,
                                };
                                symbolToOriginInfoMap[index] = origin;
                            }
                        }
                        else if (preferences.includeCompletionsWithInsertText) {
                            addSymbolOriginInfo(symbol);
                            addSymbolSortInfo(symbol);
                            symbols.push(symbol);
                        }
                    }
                    else {
                        addSymbolOriginInfo(symbol);
                        addSymbolSortInfo(symbol);
                        symbols.push(symbol);
                    }
                    function addSymbolSortInfo(symbol) {
                        if (isStaticProperty(symbol)) {
                            symbolToSortTextIdMap[ts.getSymbolId(symbol)] = 10 /* LocalDeclarationPriority */;
                        }
                    }
                    function addSymbolOriginInfo(symbol) {
                        if (preferences.includeCompletionsWithInsertText) {
                            if (insertAwait && ts.addToSeen(seenPropertySymbols, ts.getSymbolId(symbol))) {
                                symbolToOriginInfoMap[symbols.length] = { kind: getNullableSymbolOriginInfoKind(8 /* Promise */) };
                            }
                            else if (insertQuestionDot) {
                                symbolToOriginInfoMap[symbols.length] = { kind: 16 /* Nullable */ };
                            }
                        }
                    }
                    function getNullableSymbolOriginInfoKind(kind) {
                        return insertQuestionDot ? kind | 16 /* Nullable */ : kind;
                    }
                }
                /** Given 'a.b.c', returns 'a'. */
                function getLeftMostName(e) {
                    return ts.isIdentifier(e) ? e : ts.isPropertyAccessExpression(e) ? getLeftMostName(e.expression) : undefined;
                }
                function tryGetGlobalSymbols() {
                    var result = tryGetObjectTypeLiteralInTypeArgumentCompletionSymbols()
                        || tryGetObjectLikeCompletionSymbols()
                        || tryGetImportCompletionSymbols()
                        || tryGetImportOrExportClauseCompletionSymbols()
                        || tryGetLocalNamedExportCompletionSymbols()
                        || tryGetConstructorCompletion()
                        || tryGetClassLikeCompletionSymbols()
                        || tryGetJsxCompletionSymbols()
                        || (getGlobalCompletions(), 1 /* Success */);
                    return result === 1 /* Success */;
                }
                function tryGetConstructorCompletion() {
                    if (!tryGetConstructorLikeCompletionContainer(contextToken))
                        return 0 /* Continue */;
                    // no members, only keywords
                    completionKind = 5 /* None */;
                    // Declaring new property/method/accessor
                    isNewIdentifierLocation = true;
                    // Has keywords for constructor parameter
                    keywordFilters = 4 /* ConstructorParameterKeywords */;
                    return 1 /* Success */;
                }
                function tryGetJsxCompletionSymbols() {
                    var jsxContainer = tryGetContainingJsxElement(contextToken);
                    // Cursor is inside a JSX self-closing element or opening element
                    var attrsType = jsxContainer && typeChecker.getContextualType(jsxContainer.attributes);
                    if (!attrsType)
                        return 0 /* Continue */;
                    var completionsType = jsxContainer && typeChecker.getContextualType(jsxContainer.attributes, 4 /* Completions */);
                    symbols = ts.concatenate(symbols, filterJsxAttributes(getPropertiesForObjectExpression(attrsType, completionsType, jsxContainer.attributes, typeChecker), jsxContainer.attributes.properties));
                    setSortTextToOptionalMember();
                    completionKind = 3 /* MemberLike */;
                    isNewIdentifierLocation = false;
                    return 1 /* Success */;
                }
                function tryGetImportCompletionSymbols() {
                    if (!importCompletionNode)
                        return 0 /* Continue */;
                    isNewIdentifierLocation = true;
                    collectAutoImports();
                    return 1 /* Success */;
                }
                function getGlobalCompletions() {
                    keywordFilters = tryGetFunctionLikeBodyCompletionContainer(contextToken) ? 5 /* FunctionLikeBodyKeywords */ : 1 /* All */;
                    // Get all entities in the current scope.
                    completionKind = 1 /* Global */;
                    isNewIdentifierLocation = isNewIdentifierDefinitionLocation();
                    if (previousToken !== contextToken) {
                        ts.Debug.assert(!!previousToken, "Expected 'contextToken' to be defined when different from 'previousToken'.");
                    }
                    // We need to find the node that will give us an appropriate scope to begin
                    // aggregating completion candidates. This is achieved in 'getScopeNode'
                    // by finding the first node that encompasses a position, accounting for whether a node
                    // is "complete" to decide whether a position belongs to the node.
                    //
                    // However, at the end of an identifier, we are interested in the scope of the identifier
                    // itself, but fall outside of the identifier. For instance:
                    //
                    //      xyz => x$
                    //
                    // the cursor is outside of both the 'x' and the arrow function 'xyz => x',
                    // so 'xyz' is not returned in our results.
                    //
                    // We define 'adjustedPosition' so that we may appropriately account for
                    // being at the end of an identifier. The intention is that if requesting completion
                    // at the end of an identifier, it should be effectively equivalent to requesting completion
                    // anywhere inside/at the beginning of the identifier. So in the previous case, the
                    // 'adjustedPosition' will work as if requesting completion in the following:
                    //
                    //      xyz => $x
                    //
                    // If previousToken !== contextToken, then
                    //   - 'contextToken' was adjusted to the token prior to 'previousToken'
                    //      because we were at the end of an identifier.
                    //   - 'previousToken' is defined.
                    var adjustedPosition = previousToken !== contextToken ?
                        previousToken.getStart() :
                        position;
                    var scopeNode = getScopeNode(contextToken, adjustedPosition, sourceFile) || sourceFile;
                    isInSnippetScope = isSnippetScope(scopeNode);
                    var symbolMeanings = (isTypeOnlyLocation ? 0 /* None */ : 111551 /* Value */) | 788968 /* Type */ | 1920 /* Namespace */ | 2097152 /* Alias */;
                    symbols = ts.concatenate(symbols, typeChecker.getSymbolsInScope(scopeNode, symbolMeanings));
                    ts.Debug.assertEachIsDefined(symbols, "getSymbolsInScope() should all be defined");
                    for (var _i = 0, symbols_1 = symbols; _i < symbols_1.length; _i++) {
                        var symbol = symbols_1[_i];
                        if (!typeChecker.isArgumentsSymbol(symbol) &&
                            !ts.some(symbol.declarations, function (d) { return d.getSourceFile() === sourceFile; })) {
                            symbolToSortTextIdMap[ts.getSymbolId(symbol)] = 15 /* GlobalsOrKeywords */;
                        }
                    }
                    // Need to insert 'this.' before properties of `this` type, so only do that if `includeInsertTextCompletions`
                    if (preferences.includeCompletionsWithInsertText && scopeNode.kind !== 300 /* SourceFile */) {
                        var thisType = typeChecker.tryGetThisTypeAt(scopeNode, /*includeGlobalThis*/ false);
                        if (thisType && !isProbablyGlobalType(thisType, sourceFile, typeChecker)) {
                            for (var _a = 0, _b = getPropertiesForCompletion(thisType, typeChecker); _a < _b.length; _a++) {
                                var symbol = _b[_a];
                                symbolToOriginInfoMap[symbols.length] = { kind: 1 /* ThisType */ };
                                symbols.push(symbol);
                                symbolToSortTextIdMap[ts.getSymbolId(symbol)] = 14 /* SuggestedClassMembers */;
                            }
                        }
                    }
                    collectAutoImports();
                    if (isTypeOnlyLocation) {
                        keywordFilters = contextToken && ts.isAssertionExpression(contextToken.parent)
                            ? 6 /* TypeAssertionKeywords */
                            : 7 /* TypeKeywords */;
                    }
                }
                function shouldOfferImportCompletions() {
                    // If already typing an import statement, provide completions for it.
                    if (importCompletionNode)
                        return true;
                    // If current completion is for non-contextual Object literal shortahands, ignore auto-import symbols
                    if (isNonContextualObjectLiteral)
                        return false;
                    // If not already a module, must have modules enabled.
                    if (!preferences.includeCompletionsForModuleExports)
                        return false;
                    // If already using ES6 modules, OK to continue using them.
                    if (sourceFile.externalModuleIndicator || sourceFile.commonJsModuleIndicator)
                        return true;
                    // If module transpilation is enabled or we're targeting es6 or above, or not emitting, OK.
                    if (ts.compilerOptionsIndicateEs6Modules(program.getCompilerOptions()))
                        return true;
                    // If some file is using ES6 modules, assume that it's OK to add more.
                    return ts.programContainsModules(program);
                }
                function isSnippetScope(scopeNode) {
                    switch (scopeNode.kind) {
                        case 300 /* SourceFile */:
                        case 221 /* TemplateExpression */:
                        case 286 /* JsxExpression */:
                        case 233 /* Block */:
                            return true;
                        default:
                            return ts.isStatement(scopeNode);
                    }
                }
                function isTypeOnlyCompletion() {
                    return insideJsDocTagTypeExpression
                        || !isContextTokenValueLocation(contextToken) &&
                            (ts.isPossiblyTypeArgumentPosition(contextToken, sourceFile, typeChecker)
                                || ts.isPartOfTypeNode(location)
                                || isContextTokenTypeLocation(contextToken));
                }
                function isContextTokenValueLocation(contextToken) {
                    return contextToken &&
                        ((contextToken.kind === 112 /* TypeOfKeyword */ &&
                            (contextToken.parent.kind === 179 /* TypeQuery */ || ts.isTypeOfExpression(contextToken.parent))) ||
                            (contextToken.kind === 128 /* AssertsKeyword */ && contextToken.parent.kind === 175 /* TypePredicate */));
                }
                function isContextTokenTypeLocation(contextToken) {
                    if (contextToken) {
                        var parentKind = contextToken.parent.kind;
                        switch (contextToken.kind) {
                            case 58 /* ColonToken */:
                                return parentKind === 165 /* PropertyDeclaration */ ||
                                    parentKind === 164 /* PropertySignature */ ||
                                    parentKind === 162 /* Parameter */ ||
                                    parentKind === 252 /* VariableDeclaration */ ||
                                    ts.isFunctionLikeKind(parentKind);
                            case 63 /* EqualsToken */:
                                return parentKind === 257 /* TypeAliasDeclaration */;
                            case 127 /* AsKeyword */:
                                return parentKind === 227 /* AsExpression */;
                            case 29 /* LessThanToken */:
                                return parentKind === 176 /* TypeReference */ ||
                                    parentKind === 209 /* TypeAssertionExpression */;
                            case 94 /* ExtendsKeyword */:
                                return parentKind === 161 /* TypeParameter */;
                        }
                    }
                    return false;
                }
                /** Mutates `symbols`, `symbolToOriginInfoMap`, and `symbolToSortTextIdMap` */
                function collectAutoImports() {
                    var _a, _b;
                    if (!shouldOfferImportCompletions())
                        return;
                    ts.Debug.assert(!(detailsEntryId === null || detailsEntryId === void 0 ? void 0 : detailsEntryId.data), "Should not run 'collectAutoImports' when faster path is available via `data`");
                    if (detailsEntryId && !detailsEntryId.source) {
                        // Asking for completion details for an item that is not an auto-import
                        return;
                    }
                    var moduleSpecifierCache = (_a = host.getModuleSpecifierCache) === null || _a === void 0 ? void 0 : _a.call(host);
                    var lowerCaseTokenText = previousToken && ts.isIdentifier(previousToken) ? previousToken.text.toLowerCase() : "";
                    var exportInfo = ts.getExportInfoMap(sourceFile, host, program, cancellationToken);
                    var packageJsonAutoImportProvider = (_b = host.getPackageJsonAutoImportProvider) === null || _b === void 0 ? void 0 : _b.call(host);
                    var packageJsonFilter = detailsEntryId ? undefined : ts.createPackageJsonImportFilter(sourceFile, preferences, host);
                    resolvingModuleSpecifiers("collectAutoImports", host, program, sourceFile, preferences, !!importCompletionNode, function (context) {
                        exportInfo.forEach(sourceFile.path, function (info, symbolName, isFromAmbientModule) {
                            if (!ts.isIdentifierText(symbolName, ts.getEmitScriptTarget(host.getCompilationSettings())))
                                return;
                            if (!detailsEntryId && ts.isStringANonContextualKeyword(symbolName))
                                return;
                            // `targetFlags` should be the same for each `info`
                            if (!isTypeOnlyLocation && !importCompletionNode && !(info[0].targetFlags & 111551 /* Value */))
                                return;
                            if (isTypeOnlyLocation && !(info[0].targetFlags & (1536 /* Module */ | 788968 /* Type */)))
                                return;
                            var isCompletionDetailsMatch = detailsEntryId && ts.some(info, function (i) { return detailsEntryId.source === ts.stripQuotes(i.moduleSymbol.name); });
                            if (isCompletionDetailsMatch || !detailsEntryId && charactersFuzzyMatchInString(symbolName, lowerCaseTokenText)) {
                                var defaultExportInfo = ts.find(info, isImportableExportInfo);
                                if (!defaultExportInfo) {
                                    return;
                                }
                                // If we don't need to resolve module specifiers, we can use any re-export that is importable at all
                                // (We need to ensure that at least one is importable to show a completion.)
                                var _a = context.tryResolve(info, isFromAmbientModule) || {}, _b = _a.exportInfo, exportInfo_1 = _b === void 0 ? defaultExportInfo : _b, moduleSpecifier = _a.moduleSpecifier;
                                var isDefaultExport = exportInfo_1.exportKind === 1 /* Default */;
                                var symbol = isDefaultExport && ts.getLocalSymbolForExportDefault(exportInfo_1.symbol) || exportInfo_1.symbol;
                                pushAutoImportSymbol(symbol, {
                                    kind: moduleSpecifier ? 32 /* ResolvedExport */ : 4 /* Export */,
                                    moduleSpecifier: moduleSpecifier,
                                    symbolName: symbolName,
                                    exportName: exportInfo_1.exportKind === 2 /* ExportEquals */ ? "export=" /* ExportEquals */ : exportInfo_1.symbol.name,
                                    fileName: exportInfo_1.moduleFileName,
                                    isDefaultExport: isDefaultExport,
                                    moduleSymbol: exportInfo_1.moduleSymbol,
                                    isFromPackageJson: exportInfo_1.isFromPackageJson,
                                });
                            }
                        });
                        hasUnresolvedAutoImports = context.resolutionLimitExceeded();
                    });
                    function isImportableExportInfo(info) {
                        var moduleFile = ts.tryCast(info.moduleSymbol.valueDeclaration, ts.isSourceFile);
                        if (!moduleFile) {
                            var moduleName = ts.stripQuotes(info.moduleSymbol.name);
                            if (ts.JsTyping.nodeCoreModules.has(moduleName) && ts.startsWith(moduleName, "node:") !== ts.shouldUseUriStyleNodeCoreModules(sourceFile, program)) {
                                return false;
                            }
                            return packageJsonFilter
                                ? packageJsonFilter.allowsImportingAmbientModule(info.moduleSymbol, getModuleSpecifierResolutionHost(info.isFromPackageJson))
                                : true;
                        }
                        return ts.isImportableFile(info.isFromPackageJson ? packageJsonAutoImportProvider : program, sourceFile, moduleFile, preferences, packageJsonFilter, getModuleSpecifierResolutionHost(info.isFromPackageJson), moduleSpecifierCache);
                    }
                }
                function pushAutoImportSymbol(symbol, origin) {
                    var symbolId = ts.getSymbolId(symbol);
                    if (symbolToSortTextIdMap[symbolId] === 15 /* GlobalsOrKeywords */) {
                        // If an auto-importable symbol is available as a global, don't add the auto import
                        return;
                    }
                    symbolToOriginInfoMap[symbols.length] = origin;
                    symbolToSortTextIdMap[symbolId] = importCompletionNode ? 11 /* LocationPriority */ : 16 /* AutoImportSuggestions */;
                    symbols.push(symbol);
                }
                /**
                 * Finds the first node that "embraces" the position, so that one may
                 * accurately aggregate locals from the closest containing scope.
                 */
                function getScopeNode(initialToken, position, sourceFile) {
                    var scope = initialToken;
                    while (scope && !ts.positionBelongsToNode(scope, position, sourceFile)) {
                        scope = scope.parent;
                    }
                    return scope;
                }
                function isCompletionListBlocker(contextToken) {
                    var start = ts.timestamp();
                    var result = isInStringOrRegularExpressionOrTemplateLiteral(contextToken) ||
                        isSolelyIdentifierDefinitionLocation(contextToken) ||
                        isDotOfNumericLiteral(contextToken) ||
                        isInJsxText(contextToken) ||
                        ts.isBigIntLiteral(contextToken);
                    log("getCompletionsAtPosition: isCompletionListBlocker: " + (ts.timestamp() - start));
                    return result;
                }
                function isInJsxText(contextToken) {
                    if (contextToken.kind === 11 /* JsxText */) {
                        return true;
                    }
                    if (contextToken.kind === 31 /* GreaterThanToken */ && contextToken.parent) {
                        if (contextToken.parent.kind === 278 /* JsxOpeningElement */) {
                            // Two possibilities:
                            //   1. <div>/**/
                            //      - contextToken: GreaterThanToken (before cursor)
                            //      - location: JSXElement
                            //      - different parents (JSXOpeningElement, JSXElement)
                            //   2. <Component<string> /**/>
                            //      - contextToken: GreaterThanToken (before cursor)
                            //      - location: GreaterThanToken (after cursor)
                            //      - same parent (JSXOpeningElement)
                            return location.parent.kind !== 278 /* JsxOpeningElement */;
                        }
                        if (contextToken.parent.kind === 279 /* JsxClosingElement */ || contextToken.parent.kind === 277 /* JsxSelfClosingElement */) {
                            return !!contextToken.parent.parent && contextToken.parent.parent.kind === 276 /* JsxElement */;
                        }
                    }
                    return false;
                }
                function isNewIdentifierDefinitionLocation() {
                    if (contextToken) {
                        var containingNodeKind = contextToken.parent.kind;
                        // Previous token may have been a keyword that was converted to an identifier.
                        switch (keywordForNode(contextToken)) {
                            case 27 /* CommaToken */:
                                return containingNodeKind === 206 /* CallExpression */ // func( a, |
                                    || containingNodeKind === 169 /* Constructor */ // constructor( a, |   /* public, protected, private keywords are allowed here, so show completion */
                                    || containingNodeKind === 207 /* NewExpression */ // new C(a, |
                                    || containingNodeKind === 202 /* ArrayLiteralExpression */ // [a, |
                                    || containingNodeKind === 219 /* BinaryExpression */ // const x = (a, |
                                    || containingNodeKind === 177 /* FunctionType */ // var x: (s: string, list|
                                    || containingNodeKind === 203 /* ObjectLiteralExpression */; // const obj = { x, |
                            case 20 /* OpenParenToken */:
                                return containingNodeKind === 206 /* CallExpression */ // func( |
                                    || containingNodeKind === 169 /* Constructor */ // constructor( |
                                    || containingNodeKind === 207 /* NewExpression */ // new C(a|
                                    || containingNodeKind === 210 /* ParenthesizedExpression */ // const x = (a|
                                    || containingNodeKind === 189 /* ParenthesizedType */; // function F(pred: (a| /* this can become an arrow function, where 'a' is the argument */
                            case 22 /* OpenBracketToken */:
                                return containingNodeKind === 202 /* ArrayLiteralExpression */ // [ |
                                    || containingNodeKind === 174 /* IndexSignature */ // [ | : string ]
                                    || containingNodeKind === 160 /* ComputedPropertyName */; // [ |    /* this can become an index signature */
                            case 140 /* ModuleKeyword */: // module |
                            case 141 /* NamespaceKeyword */: // namespace |
                                return true;
                            case 24 /* DotToken */:
                                return containingNodeKind === 259 /* ModuleDeclaration */; // module A.|
                            case 18 /* OpenBraceToken */:
                                return containingNodeKind === 255 /* ClassDeclaration */ // class A { |
                                    || containingNodeKind === 203 /* ObjectLiteralExpression */; // const obj = { |
                            case 63 /* EqualsToken */:
                                return containingNodeKind === 252 /* VariableDeclaration */ // const x = a|
                                    || containingNodeKind === 219 /* BinaryExpression */; // x = a|
                            case 15 /* TemplateHead */:
                                return containingNodeKind === 221 /* TemplateExpression */; // `aa ${|
                            case 16 /* TemplateMiddle */:
                                return containingNodeKind === 231 /* TemplateSpan */; // `aa ${10} dd ${|
                            case 123 /* PublicKeyword */:
                            case 121 /* PrivateKeyword */:
                            case 122 /* ProtectedKeyword */:
                                return containingNodeKind === 165 /* PropertyDeclaration */; // class A{ public |
                        }
                    }
                    return false;
                }
                function isInStringOrRegularExpressionOrTemplateLiteral(contextToken) {
                    // To be "in" one of these literals, the position has to be:
                    //   1. entirely within the token text.
                    //   2. at the end position of an unterminated token.
                    //   3. at the end of a regular expression (due to trailing flags like '/foo/g').
                    return (ts.isRegularExpressionLiteral(contextToken) || ts.isStringTextContainingNode(contextToken)) && (ts.rangeContainsPositionExclusive(ts.createTextRangeFromSpan(ts.createTextSpanFromNode(contextToken)), position) ||
                        position === contextToken.end && (!!contextToken.isUnterminated || ts.isRegularExpressionLiteral(contextToken)));
                }
                function tryGetObjectTypeLiteralInTypeArgumentCompletionSymbols() {
                    var typeLiteralNode = tryGetTypeLiteralNode(contextToken);
                    if (!typeLiteralNode)
                        return 0 /* Continue */;
                    var intersectionTypeNode = ts.isIntersectionTypeNode(typeLiteralNode.parent) ? typeLiteralNode.parent : undefined;
                    var containerTypeNode = intersectionTypeNode || typeLiteralNode;
                    var containerExpectedType = getConstraintOfTypeArgumentProperty(containerTypeNode, typeChecker);
                    if (!containerExpectedType)
                        return 0 /* Continue */;
                    var containerActualType = typeChecker.getTypeFromTypeNode(containerTypeNode);
                    var members = getPropertiesForCompletion(containerExpectedType, typeChecker);
                    var existingMembers = getPropertiesForCompletion(containerActualType, typeChecker);
                    var existingMemberEscapedNames = new ts.Set();
                    existingMembers.forEach(function (s) { return existingMemberEscapedNames.add(s.escapedName); });
                    symbols = ts.concatenate(symbols, ts.filter(members, function (s) { return !existingMemberEscapedNames.has(s.escapedName); }));
                    completionKind = 0 /* ObjectPropertyDeclaration */;
                    isNewIdentifierLocation = true;
                    return 1 /* Success */;
                }
                /**
                 * Aggregates relevant symbols for completion in object literals and object binding patterns.
                 * Relevant symbols are stored in the captured 'symbols' variable.
                 *
                 * @returns true if 'symbols' was successfully populated; false otherwise.
                 */
                function tryGetObjectLikeCompletionSymbols() {
                    var objectLikeContainer = tryGetObjectLikeCompletionContainer(contextToken);
                    if (!objectLikeContainer)
                        return 0 /* Continue */;
                    // We're looking up possible property names from contextual/inferred/declared type.
                    completionKind = 0 /* ObjectPropertyDeclaration */;
                    var typeMembers;
                    var existingMembers;
                    if (objectLikeContainer.kind === 203 /* ObjectLiteralExpression */) {
                        var instantiatedType = tryGetObjectLiteralContextualType(objectLikeContainer, typeChecker);
                        // Check completions for Object property value shorthand
                        if (instantiatedType === undefined) {
                            if (objectLikeContainer.flags & 16777216 /* InWithStatement */) {
                                return 2 /* Fail */;
                            }
                            isNonContextualObjectLiteral = true;
                            return 0 /* Continue */;
                        }
                        var completionsType = typeChecker.getContextualType(objectLikeContainer, 4 /* Completions */);
                        var hasStringIndexType = (completionsType || instantiatedType).getStringIndexType();
                        var hasNumberIndextype = (completionsType || instantiatedType).getNumberIndexType();
                        isNewIdentifierLocation = !!hasStringIndexType || !!hasNumberIndextype;
                        typeMembers = getPropertiesForObjectExpression(instantiatedType, completionsType, objectLikeContainer, typeChecker);
                        existingMembers = objectLikeContainer.properties;
                        if (typeMembers.length === 0) {
                            // Edge case: If NumberIndexType exists
                            if (!hasNumberIndextype) {
                                isNonContextualObjectLiteral = true;
                                return 0 /* Continue */;
                            }
                        }
                    }
                    else {
                        ts.Debug.assert(objectLikeContainer.kind === 199 /* ObjectBindingPattern */);
                        // We are *only* completing on properties from the type being destructured.
                        isNewIdentifierLocation = false;
                        var rootDeclaration = ts.getRootDeclaration(objectLikeContainer.parent);
                        if (!ts.isVariableLike(rootDeclaration))
                            return ts.Debug.fail("Root declaration is not variable-like.");
                        // We don't want to complete using the type acquired by the shape
                        // of the binding pattern; we are only interested in types acquired
                        // through type declaration or inference.
                        // Also proceed if rootDeclaration is a parameter and if its containing function expression/arrow function is contextually typed -
                        // type of parameter will flow in from the contextual type of the function
                        var canGetType = ts.hasInitializer(rootDeclaration) || ts.hasType(rootDeclaration) || rootDeclaration.parent.parent.kind === 242 /* ForOfStatement */;
                        if (!canGetType && rootDeclaration.kind === 162 /* Parameter */) {
                            if (ts.isExpression(rootDeclaration.parent)) {
                                canGetType = !!typeChecker.getContextualType(rootDeclaration.parent);
                            }
                            else if (rootDeclaration.parent.kind === 167 /* MethodDeclaration */ || rootDeclaration.parent.kind === 171 /* SetAccessor */) {
                                canGetType = ts.isExpression(rootDeclaration.parent.parent) && !!typeChecker.getContextualType(rootDeclaration.parent.parent);
                            }
                        }
                        if (canGetType) {
                            var typeForObject_1 = typeChecker.getTypeAtLocation(objectLikeContainer);
                            if (!typeForObject_1)
                                return 2 /* Fail */;
                            // In a binding pattern, get only known properties (unless in the same scope).
                            // Everywhere else we will get all possible properties.
                            var containerClass_1 = ts.getContainingClass(objectLikeContainer);
                            typeMembers = typeChecker.getPropertiesOfType(typeForObject_1).filter(function (symbol) {
                                // either public
                                return !(ts.getDeclarationModifierFlagsFromSymbol(symbol) & 24 /* NonPublicAccessibilityModifier */)
                                    // or we're in it
                                    || containerClass_1 && ts.contains(typeForObject_1.symbol.declarations, containerClass_1);
                            });
                            existingMembers = objectLikeContainer.elements;
                        }
                    }
                    if (typeMembers && typeMembers.length > 0) {
                        // Add filtered items to the completion list
                        symbols = ts.concatenate(symbols, filterObjectMembersList(typeMembers, ts.Debug.checkDefined(existingMembers)));
                    }
                    setSortTextToOptionalMember();
                    return 1 /* Success */;
                }
                /**
                 * Aggregates relevant symbols for completion in import clauses and export clauses
                 * whose declarations have a module specifier; for instance, symbols will be aggregated for
                 *
                 *      import { | } from "moduleName";
                 *      export { a as foo, | } from "moduleName";
                 *
                 * but not for
                 *
                 *      export { | };
                 *
                 * Relevant symbols are stored in the captured 'symbols' variable.
                 */
                function tryGetImportOrExportClauseCompletionSymbols() {
                    // `import { |` or `import { a as 0, | }`
                    var namedImportsOrExports = contextToken && (contextToken.kind === 18 /* OpenBraceToken */ || contextToken.kind === 27 /* CommaToken */)
                        ? ts.tryCast(contextToken.parent, ts.isNamedImportsOrExports) : undefined;
                    if (!namedImportsOrExports)
                        return 0 /* Continue */;
                    // try to show exported member for imported/re-exported module
                    var moduleSpecifier = (namedImportsOrExports.kind === 267 /* NamedImports */ ? namedImportsOrExports.parent.parent : namedImportsOrExports.parent).moduleSpecifier;
                    if (!moduleSpecifier)
                        return namedImportsOrExports.kind === 267 /* NamedImports */ ? 2 /* Fail */ : 0 /* Continue */;
                    var moduleSpecifierSymbol = typeChecker.getSymbolAtLocation(moduleSpecifier); // TODO: GH#18217
                    if (!moduleSpecifierSymbol)
                        return 2 /* Fail */;
                    completionKind = 3 /* MemberLike */;
                    isNewIdentifierLocation = false;
                    var exports = typeChecker.getExportsAndPropertiesOfModule(moduleSpecifierSymbol);
                    var existing = new ts.Set(namedImportsOrExports.elements.filter(function (n) { return !isCurrentlyEditingNode(n); }).map(function (n) { return (n.propertyName || n.name).escapedText; }));
                    symbols = ts.concatenate(symbols, exports.filter(function (e) { return e.escapedName !== "default" /* Default */ && !existing.has(e.escapedName); }));
                    return 1 /* Success */;
                }
                /**
                 * Adds local declarations for completions in named exports:
                 *
                 *   export { | };
                 *
                 * Does not check for the absence of a module specifier (`export {} from "./other"`)
                 * because `tryGetImportOrExportClauseCompletionSymbols` runs first and handles that,
                 * preventing this function from running.
                 */
                function tryGetLocalNamedExportCompletionSymbols() {
                    var _a;
                    var namedExports = contextToken && (contextToken.kind === 18 /* OpenBraceToken */ || contextToken.kind === 27 /* CommaToken */)
                        ? ts.tryCast(contextToken.parent, ts.isNamedExports)
                        : undefined;
                    if (!namedExports) {
                        return 0 /* Continue */;
                    }
                    var localsContainer = ts.findAncestor(namedExports, ts.or(ts.isSourceFile, ts.isModuleDeclaration));
                    completionKind = 5 /* None */;
                    isNewIdentifierLocation = false;
                    (_a = localsContainer.locals) === null || _a === void 0 ? void 0 : _a.forEach(function (symbol, name) {
                        var _a, _b;
                        symbols.push(symbol);
                        if ((_b = (_a = localsContainer.symbol) === null || _a === void 0 ? void 0 : _a.exports) === null || _b === void 0 ? void 0 : _b.has(name)) {
                            symbolToSortTextIdMap[ts.getSymbolId(symbol)] = 12 /* OptionalMember */;
                        }
                    });
                    return 1 /* Success */;
                }
                /**
                 * Aggregates relevant symbols for completion in class declaration
                 * Relevant symbols are stored in the captured 'symbols' variable.
                 */
                function tryGetClassLikeCompletionSymbols() {
                    var decl = tryGetObjectTypeDeclarationCompletionContainer(sourceFile, contextToken, location, position);
                    if (!decl)
                        return 0 /* Continue */;
                    // We're looking up possible property names from parent type.
                    completionKind = 3 /* MemberLike */;
                    // Declaring new property/method/accessor
                    isNewIdentifierLocation = true;
                    keywordFilters = contextToken.kind === 41 /* AsteriskToken */ ? 0 /* None */ :
                        ts.isClassLike(decl) ? 2 /* ClassElementKeywords */ : 3 /* InterfaceElementKeywords */;
                    // If you're in an interface you don't want to repeat things from super-interface. So just stop here.
                    if (!ts.isClassLike(decl))
                        return 1 /* Success */;
                    var classElement = contextToken.kind === 26 /* SemicolonToken */ ? contextToken.parent.parent : contextToken.parent;
                    var classElementModifierFlags = ts.isClassElement(classElement) ? ts.getEffectiveModifierFlags(classElement) : 0 /* None */;
                    // If this is context token is not something we are editing now, consider if this would lead to be modifier
                    if (contextToken.kind === 79 /* Identifier */ && !isCurrentlyEditingNode(contextToken)) {
                        switch (contextToken.getText()) {
                            case "private":
                                classElementModifierFlags = classElementModifierFlags | 8 /* Private */;
                                break;
                            case "static":
                                classElementModifierFlags = classElementModifierFlags | 32 /* Static */;
                                break;
                            case "override":
                                classElementModifierFlags = classElementModifierFlags | 16384 /* Override */;
                                break;
                        }
                    }
                    if (ts.isClassStaticBlockDeclaration(classElement)) {
                        classElementModifierFlags |= 32 /* Static */;
                    }
                    // No member list for private methods
                    if (!(classElementModifierFlags & 8 /* Private */)) {
                        // List of property symbols of base type that are not private and already implemented
                        var baseTypeNodes = ts.isClassLike(decl) && classElementModifierFlags & 16384 /* Override */ ? ts.singleElementArray(ts.getEffectiveBaseTypeNode(decl)) : ts.getAllSuperTypeNodes(decl);
                        var baseSymbols = ts.flatMap(baseTypeNodes, function (baseTypeNode) {
                            var type = typeChecker.getTypeAtLocation(baseTypeNode);
                            return classElementModifierFlags & 32 /* Static */ ?
                                (type === null || type === void 0 ? void 0 : type.symbol) && typeChecker.getPropertiesOfType(typeChecker.getTypeOfSymbolAtLocation(type.symbol, decl)) :
                                type && typeChecker.getPropertiesOfType(type);
                        });
                        symbols = ts.concatenate(symbols, filterClassMembersList(baseSymbols, decl.members, classElementModifierFlags));
                    }
                    return 1 /* Success */;
                }
                /**
                 * Returns the immediate owning object literal or binding pattern of a context token,
                 * on the condition that one exists and that the context implies completion should be given.
                 */
                function tryGetObjectLikeCompletionContainer(contextToken) {
                    if (contextToken) {
                        var parent = contextToken.parent;
                        switch (contextToken.kind) {
                            case 18 /* OpenBraceToken */: // const x = { |
                            case 27 /* CommaToken */: // const x = { a: 0, |
                                if (ts.isObjectLiteralExpression(parent) || ts.isObjectBindingPattern(parent)) {
                                    return parent;
                                }
                                break;
                            case 41 /* AsteriskToken */:
                                return ts.isMethodDeclaration(parent) ? ts.tryCast(parent.parent, ts.isObjectLiteralExpression) : undefined;
                            case 79 /* Identifier */:
                                return contextToken.text === "async" && ts.isShorthandPropertyAssignment(contextToken.parent)
                                    ? contextToken.parent.parent : undefined;
                        }
                    }
                    return undefined;
                }
                function isConstructorParameterCompletion(node) {
                    return !!node.parent && ts.isParameter(node.parent) && ts.isConstructorDeclaration(node.parent.parent)
                        && (ts.isParameterPropertyModifier(node.kind) || ts.isDeclarationName(node));
                }
                /**
                 * Returns the immediate owning class declaration of a context token,
                 * on the condition that one exists and that the context implies completion should be given.
                 */
                function tryGetConstructorLikeCompletionContainer(contextToken) {
                    if (contextToken) {
                        var parent = contextToken.parent;
                        switch (contextToken.kind) {
                            case 20 /* OpenParenToken */:
                            case 27 /* CommaToken */:
                                return ts.isConstructorDeclaration(contextToken.parent) ? contextToken.parent : undefined;
                            default:
                                if (isConstructorParameterCompletion(contextToken)) {
                                    return parent.parent;
                                }
                        }
                    }
                    return undefined;
                }
                function tryGetFunctionLikeBodyCompletionContainer(contextToken) {
                    if (contextToken) {
                        var prev_1;
                        var container = ts.findAncestor(contextToken.parent, function (node) {
                            if (ts.isClassLike(node)) {
                                return "quit";
                            }
                            if (ts.isFunctionLikeDeclaration(node) && prev_1 === node.body) {
                                return true;
                            }
                            prev_1 = node;
                            return false;
                        });
                        return container && container;
                    }
                }
                function tryGetContainingJsxElement(contextToken) {
                    if (contextToken) {
                        var parent = contextToken.parent;
                        switch (contextToken.kind) {
                            case 31 /* GreaterThanToken */: // End of a type argument list
                            case 30 /* LessThanSlashToken */:
                            case 43 /* SlashToken */:
                            case 79 /* Identifier */:
                            case 204 /* PropertyAccessExpression */:
                            case 284 /* JsxAttributes */:
                            case 283 /* JsxAttribute */:
                            case 285 /* JsxSpreadAttribute */:
                                if (parent && (parent.kind === 277 /* JsxSelfClosingElement */ || parent.kind === 278 /* JsxOpeningElement */)) {
                                    if (contextToken.kind === 31 /* GreaterThanToken */) {
                                        var precedingToken = ts.findPrecedingToken(contextToken.pos, sourceFile, /*startNode*/ undefined);
                                        if (!parent.typeArguments || (precedingToken && precedingToken.kind === 43 /* SlashToken */))
                                            break;
                                    }
                                    return parent;
                                }
                                else if (parent.kind === 283 /* JsxAttribute */) {
                                    // Currently we parse JsxOpeningLikeElement as:
                                    //      JsxOpeningLikeElement
                                    //          attributes: JsxAttributes
                                    //             properties: NodeArray<JsxAttributeLike>
                                    return parent.parent.parent;
                                }
                                break;
                            // The context token is the closing } or " of an attribute, which means
                            // its parent is a JsxExpression, whose parent is a JsxAttribute,
                            // whose parent is a JsxOpeningLikeElement
                            case 10 /* StringLiteral */:
                                if (parent && ((parent.kind === 283 /* JsxAttribute */) || (parent.kind === 285 /* JsxSpreadAttribute */))) {
                                    // Currently we parse JsxOpeningLikeElement as:
                                    //      JsxOpeningLikeElement
                                    //          attributes: JsxAttributes
                                    //             properties: NodeArray<JsxAttributeLike>
                                    return parent.parent.parent;
                                }
                                break;
                            case 19 /* CloseBraceToken */:
                                if (parent &&
                                    parent.kind === 286 /* JsxExpression */ &&
                                    parent.parent && parent.parent.kind === 283 /* JsxAttribute */) {
                                    // Currently we parse JsxOpeningLikeElement as:
                                    //      JsxOpeningLikeElement
                                    //          attributes: JsxAttributes
                                    //             properties: NodeArray<JsxAttributeLike>
                                    //                  each JsxAttribute can have initializer as JsxExpression
                                    return parent.parent.parent.parent;
                                }
                                if (parent && parent.kind === 285 /* JsxSpreadAttribute */) {
                                    // Currently we parse JsxOpeningLikeElement as:
                                    //      JsxOpeningLikeElement
                                    //          attributes: JsxAttributes
                                    //             properties: NodeArray<JsxAttributeLike>
                                    return parent.parent.parent;
                                }
                                break;
                        }
                    }
                    return undefined;
                }
                /**
                 * @returns true if we are certain that the currently edited location must define a new location; false otherwise.
                 */
                function isSolelyIdentifierDefinitionLocation(contextToken) {
                    var parent = contextToken.parent;
                    var containingNodeKind = parent.kind;
                    switch (contextToken.kind) {
                        case 27 /* CommaToken */:
                            return containingNodeKind === 252 /* VariableDeclaration */ ||
                                isVariableDeclarationListButNotTypeArgument(contextToken) ||
                                containingNodeKind === 235 /* VariableStatement */ ||
                                containingNodeKind === 258 /* EnumDeclaration */ || // enum a { foo, |
                                isFunctionLikeButNotConstructor(containingNodeKind) ||
                                containingNodeKind === 256 /* InterfaceDeclaration */ || // interface A<T, |
                                containingNodeKind === 200 /* ArrayBindingPattern */ || // var [x, y|
                                containingNodeKind === 257 /* TypeAliasDeclaration */ || // type Map, K, |
                                // class A<T, |
                                // var C = class D<T, |
                                (ts.isClassLike(parent) &&
                                    !!parent.typeParameters &&
                                    parent.typeParameters.end >= contextToken.pos);
                        case 24 /* DotToken */:
                            return containingNodeKind === 200 /* ArrayBindingPattern */; // var [.|
                        case 58 /* ColonToken */:
                            return containingNodeKind === 201 /* BindingElement */; // var {x :html|
                        case 22 /* OpenBracketToken */:
                            return containingNodeKind === 200 /* ArrayBindingPattern */; // var [x|
                        case 20 /* OpenParenToken */:
                            return containingNodeKind === 290 /* CatchClause */ ||
                                isFunctionLikeButNotConstructor(containingNodeKind);
                        case 18 /* OpenBraceToken */:
                            return containingNodeKind === 258 /* EnumDeclaration */; // enum a { |
                        case 29 /* LessThanToken */:
                            return containingNodeKind === 255 /* ClassDeclaration */ || // class A< |
                                containingNodeKind === 224 /* ClassExpression */ || // var C = class D< |
                                containingNodeKind === 256 /* InterfaceDeclaration */ || // interface A< |
                                containingNodeKind === 257 /* TypeAliasDeclaration */ || // type List< |
                                ts.isFunctionLikeKind(containingNodeKind);
                        case 124 /* StaticKeyword */:
                            return containingNodeKind === 165 /* PropertyDeclaration */ && !ts.isClassLike(parent.parent);
                        case 25 /* DotDotDotToken */:
                            return containingNodeKind === 162 /* Parameter */ ||
                                (!!parent.parent && parent.parent.kind === 200 /* ArrayBindingPattern */); // var [...z|
                        case 123 /* PublicKeyword */:
                        case 121 /* PrivateKeyword */:
                        case 122 /* ProtectedKeyword */:
                            return containingNodeKind === 162 /* Parameter */ && !ts.isConstructorDeclaration(parent.parent);
                        case 127 /* AsKeyword */:
                            return containingNodeKind === 268 /* ImportSpecifier */ ||
                                containingNodeKind === 273 /* ExportSpecifier */ ||
                                containingNodeKind === 266 /* NamespaceImport */;
                        case 135 /* GetKeyword */:
                        case 147 /* SetKeyword */:
                            return !isFromObjectTypeDeclaration(contextToken);
                        case 84 /* ClassKeyword */:
                        case 92 /* EnumKeyword */:
                        case 118 /* InterfaceKeyword */:
                        case 98 /* FunctionKeyword */:
                        case 113 /* VarKeyword */:
                        case 100 /* ImportKeyword */:
                        case 119 /* LetKeyword */:
                        case 85 /* ConstKeyword */:
                        case 136 /* InferKeyword */:
                        case 150 /* TypeKeyword */: // type htm|
                            return true;
                        case 41 /* AsteriskToken */:
                            return ts.isFunctionLike(contextToken.parent) && !ts.isMethodDeclaration(contextToken.parent);
                    }
                    // If the previous token is keyword corresponding to class member completion keyword
                    // there will be completion available here
                    if (isClassMemberCompletionKeyword(keywordForNode(contextToken)) && isFromObjectTypeDeclaration(contextToken)) {
                        return false;
                    }
                    if (isConstructorParameterCompletion(contextToken)) {
                        // constructor parameter completion is available only if
                        // - its modifier of the constructor parameter or
                        // - its name of the parameter and not being edited
                        // eg. constructor(a |<- this shouldnt show completion
                        if (!ts.isIdentifier(contextToken) ||
                            ts.isParameterPropertyModifier(keywordForNode(contextToken)) ||
                            isCurrentlyEditingNode(contextToken)) {
                            return false;
                        }
                    }
                    // Previous token may have been a keyword that was converted to an identifier.
                    switch (keywordForNode(contextToken)) {
                        case 126 /* AbstractKeyword */:
                        case 84 /* ClassKeyword */:
                        case 85 /* ConstKeyword */:
                        case 134 /* DeclareKeyword */:
                        case 92 /* EnumKeyword */:
                        case 98 /* FunctionKeyword */:
                        case 118 /* InterfaceKeyword */:
                        case 119 /* LetKeyword */:
                        case 121 /* PrivateKeyword */:
                        case 122 /* ProtectedKeyword */:
                        case 123 /* PublicKeyword */:
                        case 124 /* StaticKeyword */:
                        case 113 /* VarKeyword */:
                            return true;
                        case 130 /* AsyncKeyword */:
                            return ts.isPropertyDeclaration(contextToken.parent);
                    }
                    // If we are inside a class declaration, and `constructor` is totally not present,
                    // but we request a completion manually at a whitespace...
                    var ancestorClassLike = ts.findAncestor(contextToken.parent, ts.isClassLike);
                    if (ancestorClassLike && contextToken === previousToken && isPreviousPropertyDeclarationTerminated(contextToken, position)) {
                        return false; // Don't block completions.
                    }
                    var ancestorPropertyDeclaraion = ts.getAncestor(contextToken.parent, 165 /* PropertyDeclaration */);
                    // If we are inside a class declaration and typing `constructor` after property declaration...
                    if (ancestorPropertyDeclaraion
                        && contextToken !== previousToken
                        && ts.isClassLike(previousToken.parent.parent)
                        // And the cursor is at the token...
                        && position <= previousToken.end) {
                        // If we are sure that the previous property declaration is terminated according to newline or semicolon...
                        if (isPreviousPropertyDeclarationTerminated(contextToken, previousToken.end)) {
                            return false; // Don't block completions.
                        }
                        else if (contextToken.kind !== 63 /* EqualsToken */
                            // Should not block: `class C { blah = c/**/ }`
                            // But should block: `class C { blah = somewhat c/**/ }` and `class C { blah: SomeType c/**/ }`
                            && (ts.isInitializedProperty(ancestorPropertyDeclaraion)
                                || ts.hasType(ancestorPropertyDeclaraion))) {
                            return true;
                        }
                    }
                    return ts.isDeclarationName(contextToken)
                        && !ts.isShorthandPropertyAssignment(contextToken.parent)
                        && !ts.isJsxAttribute(contextToken.parent)
                        // Don't block completions if we're in `class C /**/`, because we're *past* the end of the identifier and might want to complete `extends`.
                        // If `contextToken !== previousToken`, this is `class C ex/**/`.
                        && !(ts.isClassLike(contextToken.parent) && (contextToken !== previousToken || position > previousToken.end));
                }
                function isPreviousPropertyDeclarationTerminated(contextToken, position) {
                    return contextToken.kind !== 63 /* EqualsToken */ &&
                        (contextToken.kind === 26 /* SemicolonToken */
                            || !ts.positionsAreOnSameLine(contextToken.end, position, sourceFile));
                }
                function isFunctionLikeButNotConstructor(kind) {
                    return ts.isFunctionLikeKind(kind) && kind !== 169 /* Constructor */;
                }
                function isDotOfNumericLiteral(contextToken) {
                    if (contextToken.kind === 8 /* NumericLiteral */) {
                        var text = contextToken.getFullText();
                        return text.charAt(text.length - 1) === ".";
                    }
                    return false;
                }
                function isVariableDeclarationListButNotTypeArgument(node) {
                    return node.parent.kind === 253 /* VariableDeclarationList */
                        && !ts.isPossiblyTypeArgumentPosition(node, sourceFile, typeChecker);
                }
                /**
                 * Filters out completion suggestions for named imports or exports.
                 *
                 * @returns Symbols to be suggested in an object binding pattern or object literal expression, barring those whose declarations
                 *          do not occur at the current position and have not otherwise been typed.
                 */
                function filterObjectMembersList(contextualMemberSymbols, existingMembers) {
                    if (existingMembers.length === 0) {
                        return contextualMemberSymbols;
                    }
                    var membersDeclaredBySpreadAssignment = new ts.Set();
                    var existingMemberNames = new ts.Set();
                    for (var _i = 0, existingMembers_1 = existingMembers; _i < existingMembers_1.length; _i++) {
                        var m = existingMembers_1[_i];
                        // Ignore omitted expressions for missing members
                        if (m.kind !== 291 /* PropertyAssignment */ &&
                            m.kind !== 292 /* ShorthandPropertyAssignment */ &&
                            m.kind !== 201 /* BindingElement */ &&
                            m.kind !== 167 /* MethodDeclaration */ &&
                            m.kind !== 170 /* GetAccessor */ &&
                            m.kind !== 171 /* SetAccessor */ &&
                            m.kind !== 293 /* SpreadAssignment */) {
                            continue;
                        }
                        // If this is the current item we are editing right now, do not filter it out
                        if (isCurrentlyEditingNode(m)) {
                            continue;
                        }
                        var existingName = void 0;
                        if (ts.isSpreadAssignment(m)) {
                            setMembersDeclaredBySpreadAssignment(m, membersDeclaredBySpreadAssignment);
                        }
                        else if (ts.isBindingElement(m) && m.propertyName) {
                            // include only identifiers in completion list
                            if (m.propertyName.kind === 79 /* Identifier */) {
                                existingName = m.propertyName.escapedText;
                            }
                        }
                        else {
                            // TODO: Account for computed property name
                            // NOTE: if one only performs this step when m.name is an identifier,
                            // things like '__proto__' are not filtered out.
                            var name = ts.getNameOfDeclaration(m);
                            existingName = name && ts.isPropertyNameLiteral(name) ? ts.getEscapedTextOfIdentifierOrLiteral(name) : undefined;
                        }
                        if (existingName !== undefined) {
                            existingMemberNames.add(existingName);
                        }
                    }
                    var filteredSymbols = contextualMemberSymbols.filter(function (m) { return !existingMemberNames.has(m.escapedName); });
                    setSortTextToMemberDeclaredBySpreadAssignment(membersDeclaredBySpreadAssignment, filteredSymbols);
                    return filteredSymbols;
                }
                function setMembersDeclaredBySpreadAssignment(declaration, membersDeclaredBySpreadAssignment) {
                    var expression = declaration.expression;
                    var symbol = typeChecker.getSymbolAtLocation(expression);
                    var type = symbol && typeChecker.getTypeOfSymbolAtLocation(symbol, expression);
                    var properties = type && type.properties;
                    if (properties) {
                        properties.forEach(function (property) {
                            membersDeclaredBySpreadAssignment.add(property.name);
                        });
                    }
                }
                // Set SortText to OptionalMember if it is an optional member
                function setSortTextToOptionalMember() {
                    symbols.forEach(function (m) {
                        var _a;
                        if (m.flags & 16777216 /* Optional */) {
                            var symbolId = ts.getSymbolId(m);
                            symbolToSortTextIdMap[symbolId] = (_a = symbolToSortTextIdMap[symbolId]) !== null && _a !== void 0 ? _a : 12 /* OptionalMember */;
                        }
                    });
                }
                // Set SortText to MemberDeclaredBySpreadAssignment if it is fulfilled by spread assignment
                function setSortTextToMemberDeclaredBySpreadAssignment(membersDeclaredBySpreadAssignment, contextualMemberSymbols) {
                    if (membersDeclaredBySpreadAssignment.size === 0) {
                        return;
                    }
                    for (var _i = 0, contextualMemberSymbols_1 = contextualMemberSymbols; _i < contextualMemberSymbols_1.length; _i++) {
                        var contextualMemberSymbol = contextualMemberSymbols_1[_i];
                        if (membersDeclaredBySpreadAssignment.has(contextualMemberSymbol.name)) {
                            symbolToSortTextIdMap[ts.getSymbolId(contextualMemberSymbol)] = 13 /* MemberDeclaredBySpreadAssignment */;
                        }
                    }
                }
                /**
                 * Filters out completion suggestions for class elements.
                 *
                 * @returns Symbols to be suggested in an class element depending on existing memebers and symbol flags
                 */
                function filterClassMembersList(baseSymbols, existingMembers, currentClassElementModifierFlags) {
                    var existingMemberNames = new ts.Set();
                    for (var _i = 0, existingMembers_2 = existingMembers; _i < existingMembers_2.length; _i++) {
                        var m = existingMembers_2[_i];
                        // Ignore omitted expressions for missing members
                        if (m.kind !== 165 /* PropertyDeclaration */ &&
                            m.kind !== 167 /* MethodDeclaration */ &&
                            m.kind !== 170 /* GetAccessor */ &&
                            m.kind !== 171 /* SetAccessor */) {
                            continue;
                        }
                        // If this is the current item we are editing right now, do not filter it out
                        if (isCurrentlyEditingNode(m)) {
                            continue;
                        }
                        // Dont filter member even if the name matches if it is declared private in the list
                        if (ts.hasEffectiveModifier(m, 8 /* Private */)) {
                            continue;
                        }
                        // do not filter it out if the static presence doesnt match
                        if (ts.isStatic(m) !== !!(currentClassElementModifierFlags & 32 /* Static */)) {
                            continue;
                        }
                        var existingName = ts.getPropertyNameForPropertyNameNode(m.name);
                        if (existingName) {
                            existingMemberNames.add(existingName);
                        }
                    }
                    return baseSymbols.filter(function (propertySymbol) {
                        return !existingMemberNames.has(propertySymbol.escapedName) &&
                            !!propertySymbol.declarations &&
                            !(ts.getDeclarationModifierFlagsFromSymbol(propertySymbol) & 8 /* Private */) &&
                            !(propertySymbol.valueDeclaration && ts.isPrivateIdentifierClassElementDeclaration(propertySymbol.valueDeclaration));
                    });
                }
                /**
                 * Filters out completion suggestions from 'symbols' according to existing JSX attributes.
                 *
                 * @returns Symbols to be suggested in a JSX element, barring those whose attributes
                 *          do not occur at the current position and have not otherwise been typed.
                 */
                function filterJsxAttributes(symbols, attributes) {
                    var seenNames = new ts.Set();
                    var membersDeclaredBySpreadAssignment = new ts.Set();
                    for (var _i = 0, attributes_1 = attributes; _i < attributes_1.length; _i++) {
                        var attr = attributes_1[_i];
                        // If this is the current item we are editing right now, do not filter it out
                        if (isCurrentlyEditingNode(attr)) {
                            continue;
                        }
                        if (attr.kind === 283 /* JsxAttribute */) {
                            seenNames.add(attr.name.escapedText);
                        }
                        else if (ts.isJsxSpreadAttribute(attr)) {
                            setMembersDeclaredBySpreadAssignment(attr, membersDeclaredBySpreadAssignment);
                        }
                    }
                    var filteredSymbols = symbols.filter(function (a) { return !seenNames.has(a.escapedName); });
                    setSortTextToMemberDeclaredBySpreadAssignment(membersDeclaredBySpreadAssignment, filteredSymbols);
                    return filteredSymbols;
                }
                function isCurrentlyEditingNode(node) {
                    return node.getStart(sourceFile) <= position && position <= node.getEnd();
                }
            }
            function getRelevantTokens(position, sourceFile) {
                var previousToken = ts.findPrecedingToken(position, sourceFile);
                if (previousToken && position <= previousToken.end && (ts.isMemberName(previousToken) || ts.isKeyword(previousToken.kind))) {
                    var contextToken = ts.findPrecedingToken(previousToken.getFullStart(), sourceFile, /*startNode*/ undefined); // TODO: GH#18217
                    return { contextToken: contextToken, previousToken: previousToken };
                }
                return { contextToken: previousToken, previousToken: previousToken };
            }
            function getAutoImportSymbolFromCompletionEntryData(name, data, program, host) {
                var containingProgram = data.isPackageJsonImport ? host.getPackageJsonAutoImportProvider() : program;
                var checker = containingProgram.getTypeChecker();
                var moduleSymbol = data.ambientModuleName ? checker.tryFindAmbientModule(data.ambientModuleName) :
                    data.fileName ? checker.getMergedSymbol(ts.Debug.checkDefined(containingProgram.getSourceFile(data.fileName)).symbol) :
                        undefined;
                if (!moduleSymbol)
                    return undefined;
                var symbol = data.exportName === "export=" /* ExportEquals */
                    ? checker.resolveExternalModuleSymbol(moduleSymbol)
                    : checker.tryGetMemberInModuleExportsAndProperties(data.exportName, moduleSymbol);
                if (!symbol)
                    return undefined;
                var isDefaultExport = data.exportName === "default" /* Default */;
                symbol = isDefaultExport && ts.getLocalSymbolForExportDefault(symbol) || symbol;
                return {
                    symbol: symbol,
                    origin: {
                        kind: data.moduleSpecifier ? 32 /* ResolvedExport */ : 4 /* Export */,
                        moduleSymbol: moduleSymbol,
                        symbolName: name,
                        isDefaultExport: isDefaultExport,
                        exportName: data.exportName,
                        fileName: data.fileName,
                        isFromPackageJson: !!data.isPackageJsonImport,
                    }
                };
            }
            function getCompletionEntryDisplayNameForSymbol(symbol, target, origin, kind, jsxIdentifierExpected) {
                var name = originIncludesSymbolName(origin) ? origin.symbolName : symbol.name;
                if (name === undefined
                    // If the symbol is external module, don't show it in the completion list
                    // (i.e declare module "http" { const x; } | // <= request completion here, "http" should not be there)
                    || symbol.flags & 1536 /* Module */ && ts.isSingleOrDoubleQuote(name.charCodeAt(0))
                    // If the symbol is the internal name of an ES symbol, it is not a valid entry. Internal names for ES symbols start with "__@"
                    || ts.isKnownSymbol(symbol)) {
                    return undefined;
                }
                var validNameResult = { name: name, needsConvertPropertyAccess: false };
                if (ts.isIdentifierText(name, target, jsxIdentifierExpected ? 1 /* JSX */ : 0 /* Standard */) || symbol.valueDeclaration && ts.isPrivateIdentifierClassElementDeclaration(symbol.valueDeclaration)) {
                    return validNameResult;
                }
                switch (kind) {
                    case 3 /* MemberLike */:
                        return undefined;
                    case 0 /* ObjectPropertyDeclaration */:
                        // TODO: GH#18169
                        return { name: JSON.stringify(name), needsConvertPropertyAccess: false };
                    case 2 /* PropertyAccess */:
                    case 1 /* Global */: // For a 'this.' completion it will be in a global context, but may have a non-identifier name.
                        // Don't add a completion for a name starting with a space. See https://github.com/Microsoft/TypeScript/pull/20547
                        return name.charCodeAt(0) === 32 /* space */ ? undefined : { name: name, needsConvertPropertyAccess: true };
                    case 5 /* None */:
                    case 4 /* String */:
                        return validNameResult;
                    default:
                        ts.Debug.assertNever(kind);
                }
            }
            // A cache of completion entries for keywords, these do not change between sessions
            var _keywordCompletions = [];
            var allKeywordsCompletions = ts.memoize(function () {
                var res = [];
                for (var i = 81 /* FirstKeyword */; i <= 158 /* LastKeyword */; i++) {
                    res.push({
                        name: ts.tokenToString(i),
                        kind: "keyword" /* keyword */,
                        kindModifiers: "" /* none */,
                        sortText: SortText.GlobalsOrKeywords
                    });
                }
                return res;
            });
            function getKeywordCompletions(keywordFilter, filterOutTsOnlyKeywords) {
                if (!filterOutTsOnlyKeywords)
                    return getTypescriptKeywordCompletions(keywordFilter);
                var index = keywordFilter + 7 /* Last */ + 1;
                return _keywordCompletions[index] ||
                    (_keywordCompletions[index] = getTypescriptKeywordCompletions(keywordFilter)
                        .filter(function (entry) { return !isTypeScriptOnlyKeyword(ts.stringToToken(entry.name)); }));
            }
            function getTypescriptKeywordCompletions(keywordFilter) {
                return _keywordCompletions[keywordFilter] || (_keywordCompletions[keywordFilter] = allKeywordsCompletions().filter(function (entry) {
                    var kind = ts.stringToToken(entry.name);
                    switch (keywordFilter) {
                        case 0 /* None */:
                            return false;
                        case 1 /* All */:
                            return isFunctionLikeBodyKeyword(kind)
                                || kind === 134 /* DeclareKeyword */
                                || kind === 140 /* ModuleKeyword */
                                || kind === 150 /* TypeKeyword */
                                || kind === 141 /* NamespaceKeyword */
                                || ts.isTypeKeyword(kind) && kind !== 151 /* UndefinedKeyword */;
                        case 5 /* FunctionLikeBodyKeywords */:
                            return isFunctionLikeBodyKeyword(kind);
                        case 2 /* ClassElementKeywords */:
                            return isClassMemberCompletionKeyword(kind);
                        case 3 /* InterfaceElementKeywords */:
                            return isInterfaceOrTypeLiteralCompletionKeyword(kind);
                        case 4 /* ConstructorParameterKeywords */:
                            return ts.isParameterPropertyModifier(kind);
                        case 6 /* TypeAssertionKeywords */:
                            return ts.isTypeKeyword(kind) || kind === 85 /* ConstKeyword */;
                        case 7 /* TypeKeywords */:
                            return ts.isTypeKeyword(kind);
                        default:
                            return ts.Debug.assertNever(keywordFilter);
                    }
                }));
            }
            function isTypeScriptOnlyKeyword(kind) {
                switch (kind) {
                    case 126 /* AbstractKeyword */:
                    case 129 /* AnyKeyword */:
                    case 156 /* BigIntKeyword */:
                    case 132 /* BooleanKeyword */:
                    case 134 /* DeclareKeyword */:
                    case 92 /* EnumKeyword */:
                    case 155 /* GlobalKeyword */:
                    case 117 /* ImplementsKeyword */:
                    case 136 /* InferKeyword */:
                    case 118 /* InterfaceKeyword */:
                    case 138 /* IsKeyword */:
                    case 139 /* KeyOfKeyword */:
                    case 140 /* ModuleKeyword */:
                    case 141 /* NamespaceKeyword */:
                    case 142 /* NeverKeyword */:
                    case 145 /* NumberKeyword */:
                    case 146 /* ObjectKeyword */:
                    case 157 /* OverrideKeyword */:
                    case 121 /* PrivateKeyword */:
                    case 122 /* ProtectedKeyword */:
                    case 123 /* PublicKeyword */:
                    case 143 /* ReadonlyKeyword */:
                    case 148 /* StringKeyword */:
                    case 149 /* SymbolKeyword */:
                    case 150 /* TypeKeyword */:
                    case 152 /* UniqueKeyword */:
                    case 153 /* UnknownKeyword */:
                        return true;
                    default:
                        return false;
                }
            }
            function isInterfaceOrTypeLiteralCompletionKeyword(kind) {
                return kind === 143 /* ReadonlyKeyword */;
            }
            function isClassMemberCompletionKeyword(kind) {
                switch (kind) {
                    case 126 /* AbstractKeyword */:
                    case 133 /* ConstructorKeyword */:
                    case 135 /* GetKeyword */:
                    case 147 /* SetKeyword */:
                    case 130 /* AsyncKeyword */:
                    case 134 /* DeclareKeyword */:
                    case 157 /* OverrideKeyword */:
                        return true;
                    default:
                        return ts.isClassMemberModifier(kind);
                }
            }
            function isFunctionLikeBodyKeyword(kind) {
                return kind === 130 /* AsyncKeyword */
                    || kind === 131 /* AwaitKeyword */
                    || kind === 127 /* AsKeyword */
                    || !ts.isContextualKeyword(kind) && !isClassMemberCompletionKeyword(kind);
            }
            function keywordForNode(node) {
                return ts.isIdentifier(node) ? node.originalKeywordKind || 0 /* Unknown */ : node.kind;
            }
            /** Get the corresponding JSDocTag node if the position is in a jsDoc comment */
            function getJsDocTagAtPosition(node, position) {
                var jsdoc = ts.findAncestor(node, ts.isJSDoc);
                return jsdoc && jsdoc.tags && (ts.rangeContainsPosition(jsdoc, position) ? ts.findLast(jsdoc.tags, function (tag) { return tag.pos < position; }) : undefined);
            }
            function getPropertiesForObjectExpression(contextualType, completionsType, obj, checker) {
                var hasCompletionsType = completionsType && completionsType !== contextualType;
                var type = hasCompletionsType && !(completionsType.flags & 3 /* AnyOrUnknown */)
                    ? checker.getUnionType([contextualType, completionsType])
                    : contextualType;
                var properties = type.isUnion()
                    ? checker.getAllPossiblePropertiesOfTypes(type.types.filter(function (memberType) {
                        // If we're providing completions for an object literal, skip primitive, array-like, or callable types since those shouldn't be implemented by object literals.
                        return !(memberType.flags & 131068 /* Primitive */ ||
                            checker.isArrayLikeType(memberType) ||
                            ts.typeHasCallOrConstructSignatures(memberType, checker) ||
                            checker.isTypeInvalidDueToUnionDiscriminant(memberType, obj));
                    }))
                    : type.getApparentProperties();
                return hasCompletionsType ? properties.filter(hasDeclarationOtherThanSelf) : properties;
                // Filter out members whose only declaration is the object literal itself to avoid
                // self-fulfilling completions like:
                //
                // function f<T>(x: T) {}
                // f({ abc/**/: "" }) // `abc` is a member of `T` but only because it declares itself
                function hasDeclarationOtherThanSelf(member) {
                    return ts.some(member.declarations, function (decl) { return decl.parent !== obj; });
                }
            }
            Completions.getPropertiesForObjectExpression = getPropertiesForObjectExpression;
            /**
             * Gets all properties on a type, but if that type is a union of several types,
             * excludes array-like types or callable/constructable types.
             */
            function getPropertiesForCompletion(type, checker) {
                return type.isUnion()
                    ? ts.Debug.checkEachDefined(checker.getAllPossiblePropertiesOfTypes(type.types), "getAllPossiblePropertiesOfTypes() should all be defined")
                    : ts.Debug.checkEachDefined(type.getApparentProperties(), "getApparentProperties() should all be defined");
            }
            /**
             * Returns the immediate owning class declaration of a context token,
             * on the condition that one exists and that the context implies completion should be given.
             */
            function tryGetObjectTypeDeclarationCompletionContainer(sourceFile, contextToken, location, position) {
                // class c { method() { } | method2() { } }
                switch (location.kind) {
                    case 343 /* SyntaxList */:
                        return ts.tryCast(location.parent, ts.isObjectTypeDeclaration);
                    case 1 /* EndOfFileToken */:
                        var cls = ts.tryCast(ts.lastOrUndefined(ts.cast(location.parent, ts.isSourceFile).statements), ts.isObjectTypeDeclaration);
                        if (cls && !ts.findChildOfKind(cls, 19 /* CloseBraceToken */, sourceFile)) {
                            return cls;
                        }
                        break;
                    case 79 /* Identifier */: {
                        // class c { public prop = c| }
                        if (ts.isPropertyDeclaration(location.parent) && location.parent.initializer === location) {
                            return undefined;
                        }
                        // class c extends React.Component { a: () => 1\n compon| }
                        if (isFromObjectTypeDeclaration(location)) {
                            return ts.findAncestor(location, ts.isObjectTypeDeclaration);
                        }
                    }
                }
                if (!contextToken)
                    return undefined;
                // class C { blah; constructor/**/ } and so on
                if (location.kind === 133 /* ConstructorKeyword */
                    // class C { blah \n constructor/**/ }
                    || (ts.isIdentifier(contextToken) && ts.isPropertyDeclaration(contextToken.parent) && ts.isClassLike(location))) {
                    return ts.findAncestor(contextToken, ts.isClassLike);
                }
                switch (contextToken.kind) {
                    case 63 /* EqualsToken */: // class c { public prop = | /* global completions */ }
                        return undefined;
                    case 26 /* SemicolonToken */: // class c {getValue(): number; | }
                    case 19 /* CloseBraceToken */: // class c { method() { } | }
                        // class c { method() { } b| }
                        return isFromObjectTypeDeclaration(location) && location.parent.name === location
                            ? location.parent.parent
                            : ts.tryCast(location, ts.isObjectTypeDeclaration);
                    case 18 /* OpenBraceToken */: // class c { |
                    case 27 /* CommaToken */: // class c {getValue(): number, | }
                        return ts.tryCast(contextToken.parent, ts.isObjectTypeDeclaration);
                    default:
                        if (!isFromObjectTypeDeclaration(contextToken)) {
                            // class c extends React.Component { a: () => 1\n| }
                            if (ts.getLineAndCharacterOfPosition(sourceFile, contextToken.getEnd()).line !== ts.getLineAndCharacterOfPosition(sourceFile, position).line && ts.isObjectTypeDeclaration(location)) {
                                return location;
                            }
                            return undefined;
                        }
                        var isValidKeyword = ts.isClassLike(contextToken.parent.parent) ? isClassMemberCompletionKeyword : isInterfaceOrTypeLiteralCompletionKeyword;
                        return (isValidKeyword(contextToken.kind) || contextToken.kind === 41 /* AsteriskToken */ || ts.isIdentifier(contextToken) && isValidKeyword(ts.stringToToken(contextToken.text))) // TODO: GH#18217
                            ? contextToken.parent.parent : undefined;
                }
            }
            function tryGetTypeLiteralNode(node) {
                if (!node)
                    return undefined;
                var parent = node.parent;
                switch (node.kind) {
                    case 18 /* OpenBraceToken */:
                        if (ts.isTypeLiteralNode(parent)) {
                            return parent;
                        }
                        break;
                    case 26 /* SemicolonToken */:
                    case 27 /* CommaToken */:
                    case 79 /* Identifier */:
                        if (parent.kind === 164 /* PropertySignature */ && ts.isTypeLiteralNode(parent.parent)) {
                            return parent.parent;
                        }
                        break;
                }
                return undefined;
            }
            function getConstraintOfTypeArgumentProperty(node, checker) {
                if (!node)
                    return undefined;
                if (ts.isTypeNode(node) && ts.isTypeReferenceType(node.parent)) {
                    return checker.getTypeArgumentConstraint(node);
                }
                var t = getConstraintOfTypeArgumentProperty(node.parent, checker);
                if (!t)
                    return undefined;
                switch (node.kind) {
                    case 164 /* PropertySignature */:
                        return checker.getTypeOfPropertyOfContextualType(t, node.symbol.escapedName);
                    case 186 /* IntersectionType */:
                    case 180 /* TypeLiteral */:
                    case 185 /* UnionType */:
                        return t;
                }
            }
            // TODO: GH#19856 Would like to return `node is Node & { parent: (ClassElement | TypeElement) & { parent: ObjectTypeDeclaration } }` but then compilation takes > 10 minutes
            function isFromObjectTypeDeclaration(node) {
                return node.parent && ts.isClassOrTypeElement(node.parent) && ts.isObjectTypeDeclaration(node.parent.parent);
            }
            function isValidTrigger(sourceFile, triggerCharacter, contextToken, position) {
                switch (triggerCharacter) {
                    case ".":
                    case "@":
                        return true;
                    case '"':
                    case "'":
                    case "`":
                        // Only automatically bring up completions if this is an opening quote.
                        return !!contextToken && ts.isStringLiteralOrTemplate(contextToken) && position === contextToken.getStart(sourceFile) + 1;
                    case "#":
                        return !!contextToken && ts.isPrivateIdentifier(contextToken) && !!ts.getContainingClass(contextToken);
                    case "<":
                        // Opening JSX tag
                        return !!contextToken && contextToken.kind === 29 /* LessThanToken */ && (!ts.isBinaryExpression(contextToken.parent) || binaryExpressionMayBeOpenTag(contextToken.parent));
                    case "/":
                        return !!contextToken && (ts.isStringLiteralLike(contextToken)
                            ? !!ts.tryGetImportFromModuleSpecifier(contextToken)
                            : contextToken.kind === 43 /* SlashToken */ && ts.isJsxClosingElement(contextToken.parent));
                    case " ":
                        return !!contextToken && ts.isImportKeyword(contextToken) && contextToken.parent.kind === 300 /* SourceFile */;
                    default:
                        return ts.Debug.assertNever(triggerCharacter);
                }
            }
            function binaryExpressionMayBeOpenTag(_a) {
                var left = _a.left;
                return ts.nodeIsMissing(left);
            }
            /** Determines if a type is exactly the same type resolved by the global 'self', 'global', or 'globalThis'. */
            function isProbablyGlobalType(type, sourceFile, checker) {
                // The type of `self` and `window` is the same in lib.dom.d.ts, but `window` does not exist in
                // lib.webworker.d.ts, so checking against `self` is also a check against `window` when it exists.
                var selfSymbol = checker.resolveName("self", /*location*/ undefined, 111551 /* Value */, /*excludeGlobals*/ false);
                if (selfSymbol && checker.getTypeOfSymbolAtLocation(selfSymbol, sourceFile) === type) {
                    return true;
                }
                var globalSymbol = checker.resolveName("global", /*location*/ undefined, 111551 /* Value */, /*excludeGlobals*/ false);
                if (globalSymbol && checker.getTypeOfSymbolAtLocation(globalSymbol, sourceFile) === type) {
                    return true;
                }
                var globalThisSymbol = checker.resolveName("globalThis", /*location*/ undefined, 111551 /* Value */, /*excludeGlobals*/ false);
                if (globalThisSymbol && checker.getTypeOfSymbolAtLocation(globalThisSymbol, sourceFile) === type) {
                    return true;
                }
                return false;
            }
            function isStaticProperty(symbol) {
                return !!(symbol.valueDeclaration && ts.getEffectiveModifierFlags(symbol.valueDeclaration) & 32 /* Static */ && ts.isClassLike(symbol.valueDeclaration.parent));
            }
            function tryGetObjectLiteralContextualType(node, typeChecker) {
                var type = typeChecker.getContextualType(node);
                if (type) {
                    return type;
                }
                if (ts.isBinaryExpression(node.parent) && node.parent.operatorToken.kind === 63 /* EqualsToken */ && node === node.parent.left) {
                    // Object literal is assignment pattern: ({ | } = x)
                    return typeChecker.getTypeAtLocation(node.parent);
                }
                return undefined;
            }
            function getImportCompletionNode(contextToken) {
                var candidate = getCandidate();
                return candidate === 154 /* FromKeyword */ || candidate && ts.rangeIsOnSingleLine(candidate, candidate.getSourceFile()) ? candidate : undefined;
                function getCandidate() {
                    var parent = contextToken.parent;
                    if (ts.isImportEqualsDeclaration(parent)) {
                        return isModuleSpecifierMissingOrEmpty(parent.moduleReference) ? parent : undefined;
                    }
                    if (ts.isNamedImports(parent) || ts.isNamespaceImport(parent)) {
                        if (isModuleSpecifierMissingOrEmpty(parent.parent.parent.moduleSpecifier) && (ts.isNamespaceImport(parent) || parent.elements.length < 2) && !parent.parent.name) {
                            // At `import { ... } |` or `import * as Foo |`, the only possible completion is `from`
                            return contextToken.kind === 19 /* CloseBraceToken */ || contextToken.kind === 79 /* Identifier */
                                ? 154 /* FromKeyword */
                                : parent.parent.parent;
                        }
                        return undefined;
                    }
                    if (ts.isImportKeyword(contextToken) && ts.isSourceFile(parent)) {
                        // A lone import keyword with nothing following it does not parse as a statement at all
                        return contextToken;
                    }
                    if (ts.isImportKeyword(contextToken) && ts.isImportDeclaration(parent)) {
                        // `import s| from`
                        return isModuleSpecifierMissingOrEmpty(parent.moduleSpecifier) ? parent : undefined;
                    }
                    return undefined;
                }
            }
            function isModuleSpecifierMissingOrEmpty(specifier) {
                var _a;
                if (ts.nodeIsMissing(specifier))
                    return true;
                return !((_a = ts.tryCast(ts.isExternalModuleReference(specifier) ? specifier.expression : specifier, ts.isStringLiteralLike)) === null || _a === void 0 ? void 0 : _a.text);
            }
            function getVariableDeclaration(property) {
                var variableDeclaration = ts.findAncestor(property, function (node) {
                    return ts.isFunctionBlock(node) || isArrowFunctionBody(node) || ts.isBindingPattern(node)
                        ? "quit"
                        : ts.isVariableDeclaration(node);
                });
                return variableDeclaration;
            }
            function isArrowFunctionBody(node) {
                return node.parent && ts.isArrowFunction(node.parent) && node.parent.body === node;
            }
            ;
            /** True if symbol is a type or a module containing at least one type. */
            function symbolCanBeReferencedAtTypeLocation(symbol, checker, seenModules) {
                if (seenModules === void 0) {
                    seenModules = new ts.Map();
                }
                var sym = ts.skipAlias(symbol.exportSymbol || symbol, checker);
                return !!(sym.flags & 788968 /* Type */) || checker.isUnknownSymbol(sym) ||
                    !!(sym.flags & 1536 /* Module */) && ts.addToSeen(seenModules, ts.getSymbolId(sym)) &&
                        checker.getExportsOfModule(sym).some(function (e) { return symbolCanBeReferencedAtTypeLocation(e, checker, seenModules); });
            }
            function isDeprecated(symbol, checker) {
                var declarations = ts.skipAlias(symbol, checker).declarations;
                return !!ts.length(declarations) && ts.every(declarations, ts.isDeprecatedDeclaration);
            }
            /**
             * True if the first character of `lowercaseCharacters` is the first character
             * of some "word" in `identiferString` (where the string is split into "words"
             * by camelCase and snake_case segments), then if the remaining characters of
             * `lowercaseCharacters` appear, in order, in the rest of `identifierString`.
             *
             * True:
             * 'state' in 'useState'
             * 'sae' in 'useState'
             * 'viable' in 'ENVIRONMENT_VARIABLE'
             *
             * False:
             * 'staet' in 'useState'
             * 'tate' in 'useState'
             * 'ment' in 'ENVIRONMENT_VARIABLE'
             */
            function charactersFuzzyMatchInString(identifierString, lowercaseCharacters) {
                if (lowercaseCharacters.length === 0) {
                    return true;
                }
                var matchedFirstCharacter = false;
                var prevChar;
                var characterIndex = 0;
                var len = identifierString.length;
                for (var strIndex = 0; strIndex < len; strIndex++) {
                    var strChar = identifierString.charCodeAt(strIndex);
                    var testChar = lowercaseCharacters.charCodeAt(characterIndex);
                    if (strChar === testChar || strChar === toUpperCharCode(testChar)) {
                        matchedFirstCharacter || (matchedFirstCharacter = prevChar === undefined || // Beginning of word
                            97 /* a */ <= prevChar && prevChar <= 122 /* z */ && 65 /* A */ <= strChar && strChar <= 90 /* Z */ || // camelCase transition
                            prevChar === 95 /* _ */ && strChar !== 95 /* _ */); // snake_case transition
                        if (matchedFirstCharacter) {
                            characterIndex++;
                        }
                        if (characterIndex === lowercaseCharacters.length) {
                            return true;
                        }
                    }
                    prevChar = strChar;
                }
                // Did not find all characters
                return false;
            }
            function toUpperCharCode(charCode) {
                if (97 /* a */ <= charCode && charCode <= 122 /* z */) {
                    return charCode - 32;
                }
                return charCode;
            }
        })(Completions = ts.Completions || (ts.Completions = {}));
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        /* @internal */
        var DocumentHighlights;
        (function (DocumentHighlights) {
            function getDocumentHighlights(program, cancellationToken, sourceFile, position, sourceFilesToSearch) {
                var node = ts.getTouchingPropertyName(sourceFile, position);
                if (node.parent && (ts.isJsxOpeningElement(node.parent) && node.parent.tagName === node || ts.isJsxClosingElement(node.parent))) {
                    // For a JSX element, just highlight the matching tag, not all references.
                    var _a = node.parent.parent, openingElement = _a.openingElement, closingElement = _a.closingElement;
                    var highlightSpans = [openingElement, closingElement].map(function (_a) {
                        var tagName = _a.tagName;
                        return getHighlightSpanForNode(tagName, sourceFile);
                    });
                    return [{ fileName: sourceFile.fileName, highlightSpans: highlightSpans }];
                }
                return getSemanticDocumentHighlights(position, node, program, cancellationToken, sourceFilesToSearch) || getSyntacticDocumentHighlights(node, sourceFile);
            }
            DocumentHighlights.getDocumentHighlights = getDocumentHighlights;
            function getHighlightSpanForNode(node, sourceFile) {
                return {
                    fileName: sourceFile.fileName,
                    textSpan: ts.createTextSpanFromNode(node, sourceFile),
                    kind: "none" /* none */
                };
            }
            function getSemanticDocumentHighlights(position, node, program, cancellationToken, sourceFilesToSearch) {
                var sourceFilesSet = new ts.Set(sourceFilesToSearch.map(function (f) { return f.fileName; }));
                var referenceEntries = ts.FindAllReferences.getReferenceEntriesForNode(position, node, program, sourceFilesToSearch, cancellationToken, /*options*/ undefined, sourceFilesSet);
                if (!referenceEntries)
                    return undefined;
                var map = ts.arrayToMultiMap(referenceEntries.map(ts.FindAllReferences.toHighlightSpan), function (e) { return e.fileName; }, function (e) { return e.span; });
                var getCanonicalFileName = ts.createGetCanonicalFileName(program.useCaseSensitiveFileNames());
                return ts.mapDefined(ts.arrayFrom(map.entries()), function (_a) {
                    var fileName = _a[0], highlightSpans = _a[1];
                    if (!sourceFilesSet.has(fileName)) {
                        if (!program.redirectTargetsMap.has(ts.toPath(fileName, program.getCurrentDirectory(), getCanonicalFileName))) {
                            return undefined;
                        }
                        var redirectTarget_1 = program.getSourceFile(fileName);
                        var redirect = ts.find(sourceFilesToSearch, function (f) { return !!f.redirectInfo && f.redirectInfo.redirectTarget === redirectTarget_1; });
                        fileName = redirect.fileName;
                        ts.Debug.assert(sourceFilesSet.has(fileName));
                    }
                    return { fileName: fileName, highlightSpans: highlightSpans };
                });
            }
            function getSyntacticDocumentHighlights(node, sourceFile) {
                var highlightSpans = getHighlightSpans(node, sourceFile);
                return highlightSpans && [{ fileName: sourceFile.fileName, highlightSpans: highlightSpans }];
            }
            function getHighlightSpans(node, sourceFile) {
                switch (node.kind) {
                    case 99 /* IfKeyword */:
                    case 91 /* ElseKeyword */:
                        return ts.isIfStatement(node.parent) ? getIfElseOccurrences(node.parent, sourceFile) : undefined;
                    case 105 /* ReturnKeyword */:
                        return useParent(node.parent, ts.isReturnStatement, getReturnOccurrences);
                    case 109 /* ThrowKeyword */:
                        return useParent(node.parent, ts.isThrowStatement, getThrowOccurrences);
                    case 111 /* TryKeyword */:
                    case 83 /* CatchKeyword */:
                    case 96 /* FinallyKeyword */:
                        var tryStatement = node.kind === 83 /* CatchKeyword */ ? node.parent.parent : node.parent;
                        return useParent(tryStatement, ts.isTryStatement, getTryCatchFinallyOccurrences);
                    case 107 /* SwitchKeyword */:
                        return useParent(node.parent, ts.isSwitchStatement, getSwitchCaseDefaultOccurrences);
                    case 82 /* CaseKeyword */:
                    case 88 /* DefaultKeyword */: {
                        if (ts.isDefaultClause(node.parent) || ts.isCaseClause(node.parent)) {
                            return useParent(node.parent.parent.parent, ts.isSwitchStatement, getSwitchCaseDefaultOccurrences);
                        }
                        return undefined;
                    }
                    case 81 /* BreakKeyword */:
                    case 86 /* ContinueKeyword */:
                        return useParent(node.parent, ts.isBreakOrContinueStatement, getBreakOrContinueStatementOccurrences);
                    case 97 /* ForKeyword */:
                    case 115 /* WhileKeyword */:
                    case 90 /* DoKeyword */:
                        return useParent(node.parent, function (n) { return ts.isIterationStatement(n, /*lookInLabeledStatements*/ true); }, getLoopBreakContinueOccurrences);
                    case 133 /* ConstructorKeyword */:
                        return getFromAllDeclarations(ts.isConstructorDeclaration, [133 /* ConstructorKeyword */]);
                    case 135 /* GetKeyword */:
                    case 147 /* SetKeyword */:
                        return getFromAllDeclarations(ts.isAccessor, [135 /* GetKeyword */, 147 /* SetKeyword */]);
                    case 131 /* AwaitKeyword */:
                        return useParent(node.parent, ts.isAwaitExpression, getAsyncAndAwaitOccurrences);
                    case 130 /* AsyncKeyword */:
                        return highlightSpans(getAsyncAndAwaitOccurrences(node));
                    case 125 /* YieldKeyword */:
                        return highlightSpans(getYieldOccurrences(node));
                    default:
                        return ts.isModifierKind(node.kind) && (ts.isDeclaration(node.parent) || ts.isVariableStatement(node.parent))
                            ? highlightSpans(getModifierOccurrences(node.kind, node.parent))
                            : undefined;
                }
                function getFromAllDeclarations(nodeTest, keywords) {
                    return useParent(node.parent, nodeTest, function (decl) {
                        return ts.mapDefined(decl.symbol.declarations, function (d) {
                            return nodeTest(d) ? ts.find(d.getChildren(sourceFile), function (c) { return ts.contains(keywords, c.kind); }) : undefined;
                        });
                    });
                }
                function useParent(node, nodeTest, getNodes) {
                    return nodeTest(node) ? highlightSpans(getNodes(node, sourceFile)) : undefined;
                }
                function highlightSpans(nodes) {
                    return nodes && nodes.map(function (node) { return getHighlightSpanForNode(node, sourceFile); });
                }
            }
            /**
             * Aggregates all throw-statements within this node *without* crossing
             * into function boundaries and try-blocks with catch-clauses.
             */
            function aggregateOwnedThrowStatements(node) {
                if (ts.isThrowStatement(node)) {
                    return [node];
                }
                else if (ts.isTryStatement(node)) {
                    // Exceptions thrown within a try block lacking a catch clause are "owned" in the current context.
                    return ts.concatenate(node.catchClause ? aggregateOwnedThrowStatements(node.catchClause) : node.tryBlock && aggregateOwnedThrowStatements(node.tryBlock), node.finallyBlock && aggregateOwnedThrowStatements(node.finallyBlock));
                }
                // Do not cross function boundaries.
                return ts.isFunctionLike(node) ? undefined : flatMapChildren(node, aggregateOwnedThrowStatements);
            }
            /**
             * For lack of a better name, this function takes a throw statement and returns the
             * nearest ancestor that is a try-block (whose try statement has a catch clause),
             * function-block, or source file.
             */
            function getThrowStatementOwner(throwStatement) {
                var child = throwStatement;
                while (child.parent) {
                    var parent = child.parent;
                    if (ts.isFunctionBlock(parent) || parent.kind === 300 /* SourceFile */) {
                        return parent;
                    }
                    // A throw-statement is only owned by a try-statement if the try-statement has
                    // a catch clause, and if the throw-statement occurs within the try block.
                    if (ts.isTryStatement(parent) && parent.tryBlock === child && parent.catchClause) {
                        return child;
                    }
                    child = parent;
                }
                return undefined;
            }
            function aggregateAllBreakAndContinueStatements(node) {
                return ts.isBreakOrContinueStatement(node) ? [node] : ts.isFunctionLike(node) ? undefined : flatMapChildren(node, aggregateAllBreakAndContinueStatements);
            }
            function flatMapChildren(node, cb) {
                var result = [];
                node.forEachChild(function (child) {
                    var value = cb(child);
                    if (value !== undefined) {
                        result.push.apply(result, ts.toArray(value));
                    }
                });
                return result;
            }
            function ownsBreakOrContinueStatement(owner, statement) {
                var actualOwner = getBreakOrContinueOwner(statement);
                return !!actualOwner && actualOwner === owner;
            }
            function getBreakOrContinueOwner(statement) {
                return ts.findAncestor(statement, function (node) {
                    switch (node.kind) {
                        case 247 /* SwitchStatement */:
                            if (statement.kind === 243 /* ContinueStatement */) {
                                return false;
                            }
                        // falls through
                        case 240 /* ForStatement */:
                        case 241 /* ForInStatement */:
                        case 242 /* ForOfStatement */:
                        case 239 /* WhileStatement */:
                        case 238 /* DoStatement */:
                            return !statement.label || isLabeledBy(node, statement.label.escapedText);
                        default:
                            // Don't cross function boundaries.
                            // TODO: GH#20090
                            return ts.isFunctionLike(node) && "quit";
                    }
                });
            }
            function getModifierOccurrences(modifier, declaration) {
                return ts.mapDefined(getNodesToSearchForModifier(declaration, ts.modifierToFlag(modifier)), function (node) { return ts.findModifier(node, modifier); });
            }
            function getNodesToSearchForModifier(declaration, modifierFlag) {
                // Types of node whose children might have modifiers.
                var container = declaration.parent;
                switch (container.kind) {
                    case 260 /* ModuleBlock */:
                    case 300 /* SourceFile */:
                    case 233 /* Block */:
                    case 287 /* CaseClause */:
                    case 288 /* DefaultClause */:
                        // Container is either a class declaration or the declaration is a classDeclaration
                        if (modifierFlag & 128 /* Abstract */ && ts.isClassDeclaration(declaration)) {
                            return __spreadArray(__spreadArray([], declaration.members, true), [declaration], false);
                        }
                        else {
                            return container.statements;
                        }
                    case 169 /* Constructor */:
                    case 167 /* MethodDeclaration */:
                    case 254 /* FunctionDeclaration */:
                        return __spreadArray(__spreadArray([], container.parameters, true), (ts.isClassLike(container.parent) ? container.parent.members : []), true);
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                    case 256 /* InterfaceDeclaration */:
                    case 180 /* TypeLiteral */:
                        var nodes = container.members;
                        // If we're an accessibility modifier, we're in an instance member and should search
                        // the constructor's parameter list for instance members as well.
                        if (modifierFlag & (28 /* AccessibilityModifier */ | 64 /* Readonly */)) {
                            var constructor = ts.find(container.members, ts.isConstructorDeclaration);
                            if (constructor) {
                                return __spreadArray(__spreadArray([], nodes, true), constructor.parameters, true);
                            }
                        }
                        else if (modifierFlag & 128 /* Abstract */) {
                            return __spreadArray(__spreadArray([], nodes, true), [container], false);
                        }
                        return nodes;
                    // Syntactically invalid positions that the parser might produce anyway
                    case 203 /* ObjectLiteralExpression */:
                        return undefined;
                    default:
                        ts.Debug.assertNever(container, "Invalid container kind.");
                }
            }
            function pushKeywordIf(keywordList, token) {
                var expected = [];
                for (var _i = 2; _i < arguments.length; _i++) {
                    expected[_i - 2] = arguments[_i];
                }
                if (token && ts.contains(expected, token.kind)) {
                    keywordList.push(token);
                    return true;
                }
                return false;
            }
            function getLoopBreakContinueOccurrences(loopNode) {
                var keywords = [];
                if (pushKeywordIf(keywords, loopNode.getFirstToken(), 97 /* ForKeyword */, 115 /* WhileKeyword */, 90 /* DoKeyword */)) {
                    // If we succeeded and got a do-while loop, then start looking for a 'while' keyword.
                    if (loopNode.kind === 238 /* DoStatement */) {
                        var loopTokens = loopNode.getChildren();
                        for (var i = loopTokens.length - 1; i >= 0; i--) {
                            if (pushKeywordIf(keywords, loopTokens[i], 115 /* WhileKeyword */)) {
                                break;
                            }
                        }
                    }
                }
                ts.forEach(aggregateAllBreakAndContinueStatements(loopNode.statement), function (statement) {
                    if (ownsBreakOrContinueStatement(loopNode, statement)) {
                        pushKeywordIf(keywords, statement.getFirstToken(), 81 /* BreakKeyword */, 86 /* ContinueKeyword */);
                    }
                });
                return keywords;
            }
            function getBreakOrContinueStatementOccurrences(breakOrContinueStatement) {
                var owner = getBreakOrContinueOwner(breakOrContinueStatement);
                if (owner) {
                    switch (owner.kind) {
                        case 240 /* ForStatement */:
                        case 241 /* ForInStatement */:
                        case 242 /* ForOfStatement */:
                        case 238 /* DoStatement */:
                        case 239 /* WhileStatement */:
                            return getLoopBreakContinueOccurrences(owner);
                        case 247 /* SwitchStatement */:
                            return getSwitchCaseDefaultOccurrences(owner);
                    }
                }
                return undefined;
            }
            function getSwitchCaseDefaultOccurrences(switchStatement) {
                var keywords = [];
                pushKeywordIf(keywords, switchStatement.getFirstToken(), 107 /* SwitchKeyword */);
                // Go through each clause in the switch statement, collecting the 'case'/'default' keywords.
                ts.forEach(switchStatement.caseBlock.clauses, function (clause) {
                    pushKeywordIf(keywords, clause.getFirstToken(), 82 /* CaseKeyword */, 88 /* DefaultKeyword */);
                    ts.forEach(aggregateAllBreakAndContinueStatements(clause), function (statement) {
                        if (ownsBreakOrContinueStatement(switchStatement, statement)) {
                            pushKeywordIf(keywords, statement.getFirstToken(), 81 /* BreakKeyword */);
                        }
                    });
                });
                return keywords;
            }
            function getTryCatchFinallyOccurrences(tryStatement, sourceFile) {
                var keywords = [];
                pushKeywordIf(keywords, tryStatement.getFirstToken(), 111 /* TryKeyword */);
                if (tryStatement.catchClause) {
                    pushKeywordIf(keywords, tryStatement.catchClause.getFirstToken(), 83 /* CatchKeyword */);
                }
                if (tryStatement.finallyBlock) {
                    var finallyKeyword = ts.findChildOfKind(tryStatement, 96 /* FinallyKeyword */, sourceFile);
                    pushKeywordIf(keywords, finallyKeyword, 96 /* FinallyKeyword */);
                }
                return keywords;
            }
            function getThrowOccurrences(throwStatement, sourceFile) {
                var owner = getThrowStatementOwner(throwStatement);
                if (!owner) {
                    return undefined;
                }
                var keywords = [];
                ts.forEach(aggregateOwnedThrowStatements(owner), function (throwStatement) {
                    keywords.push(ts.findChildOfKind(throwStatement, 109 /* ThrowKeyword */, sourceFile));
                });
                // If the "owner" is a function, then we equate 'return' and 'throw' statements in their
                // ability to "jump out" of the function, and include occurrences for both.
                if (ts.isFunctionBlock(owner)) {
                    ts.forEachReturnStatement(owner, function (returnStatement) {
                        keywords.push(ts.findChildOfKind(returnStatement, 105 /* ReturnKeyword */, sourceFile));
                    });
                }
                return keywords;
            }
            function getReturnOccurrences(returnStatement, sourceFile) {
                var func = ts.getContainingFunction(returnStatement);
                if (!func) {
                    return undefined;
                }
                var keywords = [];
                ts.forEachReturnStatement(ts.cast(func.body, ts.isBlock), function (returnStatement) {
                    keywords.push(ts.findChildOfKind(returnStatement, 105 /* ReturnKeyword */, sourceFile));
                });
                // Include 'throw' statements that do not occur within a try block.
                ts.forEach(aggregateOwnedThrowStatements(func.body), function (throwStatement) {
                    keywords.push(ts.findChildOfKind(throwStatement, 109 /* ThrowKeyword */, sourceFile));
                });
                return keywords;
            }
            function getAsyncAndAwaitOccurrences(node) {
                var func = ts.getContainingFunction(node);
                if (!func) {
                    return undefined;
                }
                var keywords = [];
                if (func.modifiers) {
                    func.modifiers.forEach(function (modifier) {
                        pushKeywordIf(keywords, modifier, 130 /* AsyncKeyword */);
                    });
                }
                ts.forEachChild(func, function (child) {
                    traverseWithoutCrossingFunction(child, function (node) {
                        if (ts.isAwaitExpression(node)) {
                            pushKeywordIf(keywords, node.getFirstToken(), 131 /* AwaitKeyword */);
                        }
                    });
                });
                return keywords;
            }
            function getYieldOccurrences(node) {
                var func = ts.getContainingFunction(node);
                if (!func) {
                    return undefined;
                }
                var keywords = [];
                ts.forEachChild(func, function (child) {
                    traverseWithoutCrossingFunction(child, function (node) {
                        if (ts.isYieldExpression(node)) {
                            pushKeywordIf(keywords, node.getFirstToken(), 125 /* YieldKeyword */);
                        }
                    });
                });
                return keywords;
            }
            // Do not cross function/class/interface/module/type boundaries.
            function traverseWithoutCrossingFunction(node, cb) {
                cb(node);
                if (!ts.isFunctionLike(node) && !ts.isClassLike(node) && !ts.isInterfaceDeclaration(node) && !ts.isModuleDeclaration(node) && !ts.isTypeAliasDeclaration(node) && !ts.isTypeNode(node)) {
                    ts.forEachChild(node, function (child) { return traverseWithoutCrossingFunction(child, cb); });
                }
            }
            function getIfElseOccurrences(ifStatement, sourceFile) {
                var keywords = getIfElseKeywords(ifStatement, sourceFile);
                var result = [];
                // We'd like to highlight else/ifs together if they are only separated by whitespace
                // (i.e. the keywords are separated by no comments, no newlines).
                for (var i = 0; i < keywords.length; i++) {
                    if (keywords[i].kind === 91 /* ElseKeyword */ && i < keywords.length - 1) {
                        var elseKeyword = keywords[i];
                        var ifKeyword = keywords[i + 1]; // this *should* always be an 'if' keyword.
                        var shouldCombineElseAndIf = true;
                        // Avoid recalculating getStart() by iterating backwards.
                        for (var j = ifKeyword.getStart(sourceFile) - 1; j >= elseKeyword.end; j--) {
                            if (!ts.isWhiteSpaceSingleLine(sourceFile.text.charCodeAt(j))) {
                                shouldCombineElseAndIf = false;
                                break;
                            }
                        }
                        if (shouldCombineElseAndIf) {
                            result.push({
                                fileName: sourceFile.fileName,
                                textSpan: ts.createTextSpanFromBounds(elseKeyword.getStart(), ifKeyword.end),
                                kind: "reference" /* reference */
                            });
                            i++; // skip the next keyword
                            continue;
                        }
                    }
                    // Ordinary case: just highlight the keyword.
                    result.push(getHighlightSpanForNode(keywords[i], sourceFile));
                }
                return result;
            }
            function getIfElseKeywords(ifStatement, sourceFile) {
                var keywords = [];
                // Traverse upwards through all parent if-statements linked by their else-branches.
                while (ts.isIfStatement(ifStatement.parent) && ifStatement.parent.elseStatement === ifStatement) {
                    ifStatement = ifStatement.parent;
                }
                // Now traverse back down through the else branches, aggregating if/else keywords of if-statements.
                while (true) {
                    var children = ifStatement.getChildren(sourceFile);
                    pushKeywordIf(keywords, children[0], 99 /* IfKeyword */);
                    // Generally the 'else' keyword is second-to-last, so we traverse backwards.
                    for (var i = children.length - 1; i >= 0; i--) {
                        if (pushKeywordIf(keywords, children[i], 91 /* ElseKeyword */)) {
                            break;
                        }
                    }
                    if (!ifStatement.elseStatement || !ts.isIfStatement(ifStatement.elseStatement)) {
                        break;
                    }
                    ifStatement = ifStatement.elseStatement;
                }
                return keywords;
            }
            /**
             * Whether or not a 'node' is preceded by a label of the given string.
             * Note: 'node' cannot be a SourceFile.
             */
            function isLabeledBy(node, labelName) {
                return !!ts.findAncestor(node.parent, function (owner) { return !ts.isLabeledStatement(owner) ? "quit" : owner.label.escapedText === labelName; });
            }
        })(DocumentHighlights = ts.DocumentHighlights || (ts.DocumentHighlights = {}));
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        function isDocumentRegistryEntry(entry) {
            return !!entry.sourceFile;
        }
        function createDocumentRegistry(useCaseSensitiveFileNames, currentDirectory) {
            return createDocumentRegistryInternal(useCaseSensitiveFileNames, currentDirectory);
        }
        ts.createDocumentRegistry = createDocumentRegistry;
        /*@internal*/
        function createDocumentRegistryInternal(useCaseSensitiveFileNames, currentDirectory, externalCache) {
            if (currentDirectory === void 0) {
                currentDirectory = "";
            }
            // Maps from compiler setting target (ES3, ES5, etc.) to all the cached documents we have
            // for those settings.
            var buckets = new ts.Map();
            var getCanonicalFileName = ts.createGetCanonicalFileName(!!useCaseSensitiveFileNames);
            function reportStats() {
                var bucketInfoArray = ts.arrayFrom(buckets.keys()).filter(function (name) { return name && name.charAt(0) === "_"; }).map(function (name) {
                    var entries = buckets.get(name);
                    var sourceFiles = [];
                    entries.forEach(function (entry, name) {
                        if (isDocumentRegistryEntry(entry)) {
                            sourceFiles.push({
                                name: name,
                                scriptKind: entry.sourceFile.scriptKind,
                                refCount: entry.languageServiceRefCount
                            });
                        }
                        else {
                            entry.forEach(function (value, scriptKind) { return sourceFiles.push({ name: name, scriptKind: scriptKind, refCount: value.languageServiceRefCount }); });
                        }
                    });
                    sourceFiles.sort(function (x, y) { return y.refCount - x.refCount; });
                    return {
                        bucket: name,
                        sourceFiles: sourceFiles
                    };
                });
                return JSON.stringify(bucketInfoArray, undefined, 2);
            }
            function acquireDocument(fileName, compilationSettings, scriptSnapshot, version, scriptKind) {
                var path = ts.toPath(fileName, currentDirectory, getCanonicalFileName);
                var key = getKeyForCompilationSettings(compilationSettings);
                return acquireDocumentWithKey(fileName, path, compilationSettings, key, scriptSnapshot, version, scriptKind);
            }
            function acquireDocumentWithKey(fileName, path, compilationSettings, key, scriptSnapshot, version, scriptKind) {
                return acquireOrUpdateDocument(fileName, path, compilationSettings, key, scriptSnapshot, version, /*acquiring*/ true, scriptKind);
            }
            function updateDocument(fileName, compilationSettings, scriptSnapshot, version, scriptKind) {
                var path = ts.toPath(fileName, currentDirectory, getCanonicalFileName);
                var key = getKeyForCompilationSettings(compilationSettings);
                return updateDocumentWithKey(fileName, path, compilationSettings, key, scriptSnapshot, version, scriptKind);
            }
            function updateDocumentWithKey(fileName, path, compilationSettings, key, scriptSnapshot, version, scriptKind) {
                return acquireOrUpdateDocument(fileName, path, compilationSettings, key, scriptSnapshot, version, /*acquiring*/ false, scriptKind);
            }
            function getDocumentRegistryEntry(bucketEntry, scriptKind) {
                var entry = isDocumentRegistryEntry(bucketEntry) ? bucketEntry : bucketEntry.get(ts.Debug.checkDefined(scriptKind, "If there are more than one scriptKind's for same document the scriptKind should be provided"));
                ts.Debug.assert(scriptKind === undefined || !entry || entry.sourceFile.scriptKind === scriptKind, "Script kind should match provided ScriptKind:" + scriptKind + " and sourceFile.scriptKind: " + (entry === null || entry === void 0 ? void 0 : entry.sourceFile.scriptKind) + ", !entry: " + !entry);
                return entry;
            }
            function acquireOrUpdateDocument(fileName, path, compilationSettings, key, scriptSnapshot, version, acquiring, scriptKind) {
                scriptKind = ts.ensureScriptKind(fileName, scriptKind);
                var scriptTarget = scriptKind === 6 /* JSON */ ? 100 /* JSON */ : compilationSettings.target || 1 /* ES5 */;
                var bucket = ts.getOrUpdate(buckets, key, function () { return new ts.Map(); });
                var bucketEntry = bucket.get(path);
                var entry = bucketEntry && getDocumentRegistryEntry(bucketEntry, scriptKind);
                if (!entry && externalCache) {
                    var sourceFile = externalCache.getDocument(key, path);
                    if (sourceFile) {
                        ts.Debug.assert(acquiring);
                        entry = {
                            sourceFile: sourceFile,
                            languageServiceRefCount: 0
                        };
                        setBucketEntry();
                    }
                }
                if (!entry) {
                    // Have never seen this file with these settings.  Create a new source file for it.
                    var sourceFile = ts.createLanguageServiceSourceFile(fileName, scriptSnapshot, scriptTarget, version, /*setNodeParents*/ false, scriptKind);
                    if (externalCache) {
                        externalCache.setDocument(key, path, sourceFile);
                    }
                    entry = {
                        sourceFile: sourceFile,
                        languageServiceRefCount: 1,
                    };
                    setBucketEntry();
                }
                else {
                    // We have an entry for this file.  However, it may be for a different version of
                    // the script snapshot.  If so, update it appropriately.  Otherwise, we can just
                    // return it as is.
                    if (entry.sourceFile.version !== version) {
                        entry.sourceFile = ts.updateLanguageServiceSourceFile(entry.sourceFile, scriptSnapshot, version, scriptSnapshot.getChangeRange(entry.sourceFile.scriptSnapshot)); // TODO: GH#18217
                        if (externalCache) {
                            externalCache.setDocument(key, path, entry.sourceFile);
                        }
                    }
                    // If we're acquiring, then this is the first time this LS is asking for this document.
                    // Increase our ref count so we know there's another LS using the document.  If we're
                    // not acquiring, then that means the LS is 'updating' the file instead, and that means
                    // it has already acquired the document previously.  As such, we do not need to increase
                    // the ref count.
                    if (acquiring) {
                        entry.languageServiceRefCount++;
                    }
                }
                ts.Debug.assert(entry.languageServiceRefCount !== 0);
                return entry.sourceFile;
                function setBucketEntry() {
                    if (!bucketEntry) {
                        bucket.set(path, entry);
                    }
                    else if (isDocumentRegistryEntry(bucketEntry)) {
                        var scriptKindMap = new ts.Map();
                        scriptKindMap.set(bucketEntry.sourceFile.scriptKind, bucketEntry);
                        scriptKindMap.set(scriptKind, entry);
                        bucket.set(path, scriptKindMap);
                    }
                    else {
                        bucketEntry.set(scriptKind, entry);
                    }
                }
            }
            function releaseDocument(fileName, compilationSettings, scriptKind) {
                var path = ts.toPath(fileName, currentDirectory, getCanonicalFileName);
                var key = getKeyForCompilationSettings(compilationSettings);
                return releaseDocumentWithKey(path, key, scriptKind);
            }
            function releaseDocumentWithKey(path, key, scriptKind) {
                var bucket = ts.Debug.checkDefined(buckets.get(key));
                var bucketEntry = bucket.get(path);
                var entry = getDocumentRegistryEntry(bucketEntry, scriptKind);
                entry.languageServiceRefCount--;
                ts.Debug.assert(entry.languageServiceRefCount >= 0);
                if (entry.languageServiceRefCount === 0) {
                    if (isDocumentRegistryEntry(bucketEntry)) {
                        bucket.delete(path);
                    }
                    else {
                        bucketEntry.delete(scriptKind);
                        if (bucketEntry.size === 1) {
                            bucket.set(path, ts.firstDefinedIterator(bucketEntry.values(), ts.identity));
                        }
                    }
                }
            }
            function getLanguageServiceRefCounts(path, scriptKind) {
                return ts.arrayFrom(buckets.entries(), function (_a) {
                    var key = _a[0], bucket = _a[1];
                    var bucketEntry = bucket.get(path);
                    var entry = bucketEntry && getDocumentRegistryEntry(bucketEntry, scriptKind);
                    return [key, entry && entry.languageServiceRefCount];
                });
            }
            return {
                acquireDocument: acquireDocument,
                acquireDocumentWithKey: acquireDocumentWithKey,
                updateDocument: updateDocument,
                updateDocumentWithKey: updateDocumentWithKey,
                releaseDocument: releaseDocument,
                releaseDocumentWithKey: releaseDocumentWithKey,
                getLanguageServiceRefCounts: getLanguageServiceRefCounts,
                reportStats: reportStats,
                getKeyForCompilationSettings: getKeyForCompilationSettings
            };
        }
        ts.createDocumentRegistryInternal = createDocumentRegistryInternal;
        function getKeyForCompilationSettings(settings) {
            return ts.sourceFileAffectingCompilerOptions.map(function (option) { return ts.getCompilerOptionValue(settings, option); }).join("|");
        }
    })(ts || (ts = {}));
    /* Code for finding imports of an exported symbol. Used only by FindAllReferences. */
    /* @internal */
    var ts;
    (function (ts) {
        var FindAllReferences;
        (function (FindAllReferences) {
            /** Creates the imports map and returns an ImportTracker that uses it. Call this lazily to avoid calling `getDirectImportsMap` unnecessarily.  */
            function createImportTracker(sourceFiles, sourceFilesSet, checker, cancellationToken) {
                var allDirectImports = getDirectImportsMap(sourceFiles, checker, cancellationToken);
                return function (exportSymbol, exportInfo, isForRename) {
                    var _a = getImportersForExport(sourceFiles, sourceFilesSet, allDirectImports, exportInfo, checker, cancellationToken), directImports = _a.directImports, indirectUsers = _a.indirectUsers;
                    return __assign({ indirectUsers: indirectUsers }, getSearchesFromDirectImports(directImports, exportSymbol, exportInfo.exportKind, checker, isForRename));
                };
            }
            FindAllReferences.createImportTracker = createImportTracker;
            var ExportKind;
            (function (ExportKind) {
                ExportKind[ExportKind["Named"] = 0] = "Named";
                ExportKind[ExportKind["Default"] = 1] = "Default";
                ExportKind[ExportKind["ExportEquals"] = 2] = "ExportEquals";
            })(ExportKind = FindAllReferences.ExportKind || (FindAllReferences.ExportKind = {}));
            var ImportExport;
            (function (ImportExport) {
                ImportExport[ImportExport["Import"] = 0] = "Import";
                ImportExport[ImportExport["Export"] = 1] = "Export";
            })(ImportExport = FindAllReferences.ImportExport || (FindAllReferences.ImportExport = {}));
            /** Returns import statements that directly reference the exporting module, and a list of files that may access the module through a namespace. */
            function getImportersForExport(sourceFiles, sourceFilesSet, allDirectImports, _a, checker, cancellationToken) {
                var exportingModuleSymbol = _a.exportingModuleSymbol, exportKind = _a.exportKind;
                var markSeenDirectImport = ts.nodeSeenTracker();
                var markSeenIndirectUser = ts.nodeSeenTracker();
                var directImports = [];
                var isAvailableThroughGlobal = !!exportingModuleSymbol.globalExports;
                var indirectUserDeclarations = isAvailableThroughGlobal ? undefined : [];
                handleDirectImports(exportingModuleSymbol);
                return { directImports: directImports, indirectUsers: getIndirectUsers() };
                function getIndirectUsers() {
                    if (isAvailableThroughGlobal) {
                        // It has `export as namespace`, so anything could potentially use it.
                        return sourceFiles;
                    }
                    // Module augmentations may use this module's exports without importing it.
                    if (exportingModuleSymbol.declarations) {
                        for (var _i = 0, _a = exportingModuleSymbol.declarations; _i < _a.length; _i++) {
                            var decl = _a[_i];
                            if (ts.isExternalModuleAugmentation(decl) && sourceFilesSet.has(decl.getSourceFile().fileName)) {
                                addIndirectUser(decl);
                            }
                        }
                    }
                    // This may return duplicates (if there are multiple module declarations in a single source file, all importing the same thing as a namespace), but `State.markSearchedSymbol` will handle that.
                    return indirectUserDeclarations.map(ts.getSourceFileOfNode);
                }
                function handleDirectImports(exportingModuleSymbol) {
                    var theseDirectImports = getDirectImports(exportingModuleSymbol);
                    if (theseDirectImports) {
                        for (var _i = 0, theseDirectImports_1 = theseDirectImports; _i < theseDirectImports_1.length; _i++) {
                            var direct = theseDirectImports_1[_i];
                            if (!markSeenDirectImport(direct)) {
                                continue;
                            }
                            if (cancellationToken)
                                cancellationToken.throwIfCancellationRequested();
                            switch (direct.kind) {
                                case 206 /* CallExpression */:
                                    if (ts.isImportCall(direct)) {
                                        handleImportCall(direct);
                                        break;
                                    }
                                    if (!isAvailableThroughGlobal) {
                                        var parent = direct.parent;
                                        if (exportKind === 2 /* ExportEquals */ && parent.kind === 252 /* VariableDeclaration */) {
                                            var name = parent.name;
                                            if (name.kind === 79 /* Identifier */) {
                                                directImports.push(name);
                                                break;
                                            }
                                        }
                                    }
                                    break;
                                case 79 /* Identifier */: // for 'const x = require("y");
                                    break; // TODO: GH#23879
                                case 263 /* ImportEqualsDeclaration */:
                                    handleNamespaceImport(direct, direct.name, ts.hasSyntacticModifier(direct, 1 /* Export */), /*alreadyAddedDirect*/ false);
                                    break;
                                case 264 /* ImportDeclaration */:
                                    directImports.push(direct);
                                    var namedBindings = direct.importClause && direct.importClause.namedBindings;
                                    if (namedBindings && namedBindings.kind === 266 /* NamespaceImport */) {
                                        handleNamespaceImport(direct, namedBindings.name, /*isReExport*/ false, /*alreadyAddedDirect*/ true);
                                    }
                                    else if (!isAvailableThroughGlobal && ts.isDefaultImport(direct)) {
                                        addIndirectUser(getSourceFileLikeForImportDeclaration(direct)); // Add a check for indirect uses to handle synthetic default imports
                                    }
                                    break;
                                case 270 /* ExportDeclaration */:
                                    if (!direct.exportClause) {
                                        // This is `export * from "foo"`, so imports of this module may import the export too.
                                        handleDirectImports(getContainingModuleSymbol(direct, checker));
                                    }
                                    else if (direct.exportClause.kind === 272 /* NamespaceExport */) {
                                        // `export * as foo from "foo"` add to indirect uses
                                        addIndirectUser(getSourceFileLikeForImportDeclaration(direct), /** addTransitiveDependencies */ true);
                                    }
                                    else {
                                        // This is `export { foo } from "foo"` and creates an alias symbol, so recursive search will get handle re-exports.
                                        directImports.push(direct);
                                    }
                                    break;
                                case 198 /* ImportType */:
                                    // Only check for typeof import('xyz')
                                    if (direct.isTypeOf && !direct.qualifier && isExported(direct)) {
                                        addIndirectUser(direct.getSourceFile(), /** addTransitiveDependencies */ true);
                                    }
                                    directImports.push(direct);
                                    break;
                                default:
                                    ts.Debug.failBadSyntaxKind(direct, "Unexpected import kind.");
                            }
                        }
                    }
                }
                function handleImportCall(importCall) {
                    var top = ts.findAncestor(importCall, isAmbientModuleDeclaration) || importCall.getSourceFile();
                    addIndirectUser(top, /** addTransitiveDependencies */ !!isExported(importCall, /** stopAtAmbientModule */ true));
                }
                function isExported(node, stopAtAmbientModule) {
                    if (stopAtAmbientModule === void 0) {
                        stopAtAmbientModule = false;
                    }
                    return ts.findAncestor(node, function (node) {
                        if (stopAtAmbientModule && isAmbientModuleDeclaration(node))
                            return "quit";
                        return ts.some(node.modifiers, function (mod) { return mod.kind === 93 /* ExportKeyword */; });
                    });
                }
                function handleNamespaceImport(importDeclaration, name, isReExport, alreadyAddedDirect) {
                    if (exportKind === 2 /* ExportEquals */) {
                        // This is a direct import, not import-as-namespace.
                        if (!alreadyAddedDirect)
                            directImports.push(importDeclaration);
                    }
                    else if (!isAvailableThroughGlobal) {
                        var sourceFileLike = getSourceFileLikeForImportDeclaration(importDeclaration);
                        ts.Debug.assert(sourceFileLike.kind === 300 /* SourceFile */ || sourceFileLike.kind === 259 /* ModuleDeclaration */);
                        if (isReExport || findNamespaceReExports(sourceFileLike, name, checker)) {
                            addIndirectUser(sourceFileLike, /** addTransitiveDependencies */ true);
                        }
                        else {
                            addIndirectUser(sourceFileLike);
                        }
                    }
                }
                /** Adds a module and all of its transitive dependencies as possible indirect users. */
                function addIndirectUser(sourceFileLike, addTransitiveDependencies) {
                    if (addTransitiveDependencies === void 0) {
                        addTransitiveDependencies = false;
                    }
                    ts.Debug.assert(!isAvailableThroughGlobal);
                    var isNew = markSeenIndirectUser(sourceFileLike);
                    if (!isNew)
                        return;
                    indirectUserDeclarations.push(sourceFileLike); // TODO: GH#18217
                    if (!addTransitiveDependencies)
                        return;
                    var moduleSymbol = checker.getMergedSymbol(sourceFileLike.symbol);
                    if (!moduleSymbol)
                        return;
                    ts.Debug.assert(!!(moduleSymbol.flags & 1536 /* Module */));
                    var directImports = getDirectImports(moduleSymbol);
                    if (directImports) {
                        for (var _i = 0, directImports_1 = directImports; _i < directImports_1.length; _i++) {
                            var directImport = directImports_1[_i];
                            if (!ts.isImportTypeNode(directImport)) {
                                addIndirectUser(getSourceFileLikeForImportDeclaration(directImport), /** addTransitiveDependencies */ true);
                            }
                        }
                    }
                }
                function getDirectImports(moduleSymbol) {
                    return allDirectImports.get(ts.getSymbolId(moduleSymbol).toString());
                }
            }
            /**
             * Given the set of direct imports of a module, we need to find which ones import the particular exported symbol.
             * The returned `importSearches` will result in the entire source file being searched.
             * But re-exports will be placed in 'singleReferences' since they cannot be locally referenced.
             */
            function getSearchesFromDirectImports(directImports, exportSymbol, exportKind, checker, isForRename) {
                var importSearches = [];
                var singleReferences = [];
                function addSearch(location, symbol) {
                    importSearches.push([location, symbol]);
                }
                if (directImports) {
                    for (var _i = 0, directImports_2 = directImports; _i < directImports_2.length; _i++) {
                        var decl = directImports_2[_i];
                        handleImport(decl);
                    }
                }
                return { importSearches: importSearches, singleReferences: singleReferences };
                function handleImport(decl) {
                    if (decl.kind === 263 /* ImportEqualsDeclaration */) {
                        if (isExternalModuleImportEquals(decl)) {
                            handleNamespaceImportLike(decl.name);
                        }
                        return;
                    }
                    if (decl.kind === 79 /* Identifier */) {
                        handleNamespaceImportLike(decl);
                        return;
                    }
                    if (decl.kind === 198 /* ImportType */) {
                        if (decl.qualifier) {
                            var firstIdentifier = ts.getFirstIdentifier(decl.qualifier);
                            if (firstIdentifier.escapedText === ts.symbolName(exportSymbol)) {
                                singleReferences.push(firstIdentifier);
                            }
                        }
                        else if (exportKind === 2 /* ExportEquals */) {
                            singleReferences.push(decl.argument.literal);
                        }
                        return;
                    }
                    // Ignore if there's a grammar error
                    if (decl.moduleSpecifier.kind !== 10 /* StringLiteral */) {
                        return;
                    }
                    if (decl.kind === 270 /* ExportDeclaration */) {
                        if (decl.exportClause && ts.isNamedExports(decl.exportClause)) {
                            searchForNamedImport(decl.exportClause);
                        }
                        return;
                    }
                    var _a = decl.importClause || { name: undefined, namedBindings: undefined }, name = _a.name, namedBindings = _a.namedBindings;
                    if (namedBindings) {
                        switch (namedBindings.kind) {
                            case 266 /* NamespaceImport */:
                                handleNamespaceImportLike(namedBindings.name);
                                break;
                            case 267 /* NamedImports */:
                                // 'default' might be accessed as a named import `{ default as foo }`.
                                if (exportKind === 0 /* Named */ || exportKind === 1 /* Default */) {
                                    searchForNamedImport(namedBindings);
                                }
                                break;
                            default:
                                ts.Debug.assertNever(namedBindings);
                        }
                    }
                    // `export =` might be imported by a default import if `--allowSyntheticDefaultImports` is on, so this handles both ExportKind.Default and ExportKind.ExportEquals.
                    // If a default import has the same name as the default export, allow to rename it.
                    // Given `import f` and `export default function f`, we will rename both, but for `import g` we will rename just that.
                    if (name && (exportKind === 1 /* Default */ || exportKind === 2 /* ExportEquals */) && (!isForRename || name.escapedText === ts.symbolEscapedNameNoDefault(exportSymbol))) {
                        var defaultImportAlias = checker.getSymbolAtLocation(name);
                        addSearch(name, defaultImportAlias);
                    }
                }
                /**
                 * `import x = require("./x")` or `import * as x from "./x"`.
                 * An `export =` may be imported by this syntax, so it may be a direct import.
                 * If it's not a direct import, it will be in `indirectUsers`, so we don't have to do anything here.
                 */
                function handleNamespaceImportLike(importName) {
                    // Don't rename an import that already has a different name than the export.
                    if (exportKind === 2 /* ExportEquals */ && (!isForRename || isNameMatch(importName.escapedText))) {
                        addSearch(importName, checker.getSymbolAtLocation(importName));
                    }
                }
                function searchForNamedImport(namedBindings) {
                    if (!namedBindings) {
                        return;
                    }
                    for (var _i = 0, _a = namedBindings.elements; _i < _a.length; _i++) {
                        var element = _a[_i];
                        var name = element.name, propertyName = element.propertyName;
                        if (!isNameMatch((propertyName || name).escapedText)) {
                            continue;
                        }
                        if (propertyName) {
                            // This is `import { foo as bar } from "./a"` or `export { foo as bar } from "./a"`. `foo` isn't a local in the file, so just add it as a single reference.
                            singleReferences.push(propertyName);
                            // If renaming `{ foo as bar }`, don't touch `bar`, just `foo`.
                            // But do rename `foo` in ` { default as foo }` if that's the original export name.
                            if (!isForRename || name.escapedText === exportSymbol.escapedName) {
                                // Search locally for `bar`.
                                addSearch(name, checker.getSymbolAtLocation(name));
                            }
                        }
                        else {
                            var localSymbol = element.kind === 273 /* ExportSpecifier */ && element.propertyName
                                ? checker.getExportSpecifierLocalTargetSymbol(element) // For re-exporting under a different name, we want to get the re-exported symbol.
                                : checker.getSymbolAtLocation(name);
                            addSearch(name, localSymbol);
                        }
                    }
                }
                function isNameMatch(name) {
                    // Use name of "default" even in `export =` case because we may have allowSyntheticDefaultImports
                    return name === exportSymbol.escapedName || exportKind !== 0 /* Named */ && name === "default" /* Default */;
                }
            }
            /** Returns 'true' is the namespace 'name' is re-exported from this module, and 'false' if it is only used locally. */
            function findNamespaceReExports(sourceFileLike, name, checker) {
                var namespaceImportSymbol = checker.getSymbolAtLocation(name);
                return !!forEachPossibleImportOrExportStatement(sourceFileLike, function (statement) {
                    if (!ts.isExportDeclaration(statement))
                        return;
                    var exportClause = statement.exportClause, moduleSpecifier = statement.moduleSpecifier;
                    return !moduleSpecifier && exportClause && ts.isNamedExports(exportClause) &&
                        exportClause.elements.some(function (element) { return checker.getExportSpecifierLocalTargetSymbol(element) === namespaceImportSymbol; });
                });
            }
            function findModuleReferences(program, sourceFiles, searchModuleSymbol) {
                var refs = [];
                var checker = program.getTypeChecker();
                for (var _i = 0, sourceFiles_1 = sourceFiles; _i < sourceFiles_1.length; _i++) {
                    var referencingFile = sourceFiles_1[_i];
                    var searchSourceFile = searchModuleSymbol.valueDeclaration;
                    if ((searchSourceFile === null || searchSourceFile === void 0 ? void 0 : searchSourceFile.kind) === 300 /* SourceFile */) {
                        for (var _a = 0, _b = referencingFile.referencedFiles; _a < _b.length; _a++) {
                            var ref = _b[_a];
                            if (program.getSourceFileFromReference(referencingFile, ref) === searchSourceFile) {
                                refs.push({ kind: "reference", referencingFile: referencingFile, ref: ref });
                            }
                        }
                        for (var _c = 0, _d = referencingFile.typeReferenceDirectives; _c < _d.length; _c++) {
                            var ref = _d[_c];
                            var referenced = program.getResolvedTypeReferenceDirectives().get(ref.fileName);
                            if (referenced !== undefined && referenced.resolvedFileName === searchSourceFile.fileName) {
                                refs.push({ kind: "reference", referencingFile: referencingFile, ref: ref });
                            }
                        }
                    }
                    forEachImport(referencingFile, function (_importDecl, moduleSpecifier) {
                        var moduleSymbol = checker.getSymbolAtLocation(moduleSpecifier);
                        if (moduleSymbol === searchModuleSymbol) {
                            refs.push({ kind: "import", literal: moduleSpecifier });
                        }
                    });
                }
                return refs;
            }
            FindAllReferences.findModuleReferences = findModuleReferences;
            /** Returns a map from a module symbol Id to all import statements that directly reference the module. */
            function getDirectImportsMap(sourceFiles, checker, cancellationToken) {
                var map = new ts.Map();
                for (var _i = 0, sourceFiles_2 = sourceFiles; _i < sourceFiles_2.length; _i++) {
                    var sourceFile = sourceFiles_2[_i];
                    if (cancellationToken)
                        cancellationToken.throwIfCancellationRequested();
                    forEachImport(sourceFile, function (importDecl, moduleSpecifier) {
                        var moduleSymbol = checker.getSymbolAtLocation(moduleSpecifier);
                        if (moduleSymbol) {
                            var id = ts.getSymbolId(moduleSymbol).toString();
                            var imports = map.get(id);
                            if (!imports) {
                                map.set(id, imports = []);
                            }
                            imports.push(importDecl);
                        }
                    });
                }
                return map;
            }
            /** Iterates over all statements at the top level or in module declarations. Returns the first truthy result. */
            function forEachPossibleImportOrExportStatement(sourceFileLike, action) {
                return ts.forEach(sourceFileLike.kind === 300 /* SourceFile */ ? sourceFileLike.statements : sourceFileLike.body.statements, function (statement) {
                    return action(statement) || (isAmbientModuleDeclaration(statement) && ts.forEach(statement.body && statement.body.statements, action));
                });
            }
            /** Calls `action` for each import, re-export, or require() in a file. */
            function forEachImport(sourceFile, action) {
                if (sourceFile.externalModuleIndicator || sourceFile.imports !== undefined) {
                    for (var _i = 0, _a = sourceFile.imports; _i < _a.length; _i++) {
                        var i = _a[_i];
                        action(ts.importFromModuleSpecifier(i), i);
                    }
                }
                else {
                    forEachPossibleImportOrExportStatement(sourceFile, function (statement) {
                        switch (statement.kind) {
                            case 270 /* ExportDeclaration */:
                            case 264 /* ImportDeclaration */: {
                                var decl = statement;
                                if (decl.moduleSpecifier && ts.isStringLiteral(decl.moduleSpecifier)) {
                                    action(decl, decl.moduleSpecifier);
                                }
                                break;
                            }
                            case 263 /* ImportEqualsDeclaration */: {
                                var decl = statement;
                                if (isExternalModuleImportEquals(decl)) {
                                    action(decl, decl.moduleReference.expression);
                                }
                                break;
                            }
                        }
                    });
                }
            }
            /**
             * Given a local reference, we might notice that it's an import/export and recursively search for references of that.
             * If at an import, look locally for the symbol it imports.
             * If at an export, look for all imports of it.
             * This doesn't handle export specifiers; that is done in `getReferencesAtExportSpecifier`.
             * @param comingFromExport If we are doing a search for all exports, don't bother looking backwards for the imported symbol, since that's the reason we're here.
             */
            function getImportOrExportSymbol(node, symbol, checker, comingFromExport) {
                return comingFromExport ? getExport() : getExport() || getImport();
                function getExport() {
                    var _a;
                    var parent = node.parent;
                    var grandparent = parent.parent;
                    if (symbol.exportSymbol) {
                        if (parent.kind === 204 /* PropertyAccessExpression */) {
                            // When accessing an export of a JS module, there's no alias. The symbol will still be flagged as an export even though we're at the use.
                            // So check that we are at the declaration.
                            return ((_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.some(function (d) { return d === parent; })) && ts.isBinaryExpression(grandparent)
                                ? getSpecialPropertyExport(grandparent, /*useLhsSymbol*/ false)
                                : undefined;
                        }
                        else {
                            return exportInfo(symbol.exportSymbol, getExportKindForDeclaration(parent));
                        }
                    }
                    else {
                        var exportNode = getExportNode(parent, node);
                        if (exportNode && ts.hasSyntacticModifier(exportNode, 1 /* Export */)) {
                            if (ts.isImportEqualsDeclaration(exportNode) && exportNode.moduleReference === node) {
                                // We're at `Y` in `export import X = Y`. This is not the exported symbol, the left-hand-side is. So treat this as an import statement.
                                if (comingFromExport) {
                                    return undefined;
                                }
                                var lhsSymbol = checker.getSymbolAtLocation(exportNode.name);
                                return { kind: 0 /* Import */, symbol: lhsSymbol };
                            }
                            else {
                                return exportInfo(symbol, getExportKindForDeclaration(exportNode));
                            }
                        }
                        else if (ts.isNamespaceExport(parent)) {
                            return exportInfo(symbol, 0 /* Named */);
                        }
                        // If we are in `export = a;` or `export default a;`, `parent` is the export assignment.
                        else if (ts.isExportAssignment(parent)) {
                            return getExportAssignmentExport(parent);
                        }
                        // If we are in `export = class A {};` (or `export = class A {};`) at `A`, `parent.parent` is the export assignment.
                        else if (ts.isExportAssignment(grandparent)) {
                            return getExportAssignmentExport(grandparent);
                        }
                        // Similar for `module.exports =` and `exports.A =`.
                        else if (ts.isBinaryExpression(parent)) {
                            return getSpecialPropertyExport(parent, /*useLhsSymbol*/ true);
                        }
                        else if (ts.isBinaryExpression(grandparent)) {
                            return getSpecialPropertyExport(grandparent, /*useLhsSymbol*/ true);
                        }
                        else if (ts.isJSDocTypedefTag(parent)) {
                            return exportInfo(symbol, 0 /* Named */);
                        }
                    }
                    function getExportAssignmentExport(ex) {
                        // Get the symbol for the `export =` node; its parent is the module it's the export of.
                        if (!ex.symbol.parent)
                            return undefined;
                        var exportKind = ex.isExportEquals ? 2 /* ExportEquals */ : 1 /* Default */;
                        return { kind: 1 /* Export */, symbol: symbol, exportInfo: { exportingModuleSymbol: ex.symbol.parent, exportKind: exportKind } };
                    }
                    function getSpecialPropertyExport(node, useLhsSymbol) {
                        var kind;
                        switch (ts.getAssignmentDeclarationKind(node)) {
                            case 1 /* ExportsProperty */:
                                kind = 0 /* Named */;
                                break;
                            case 2 /* ModuleExports */:
                                kind = 2 /* ExportEquals */;
                                break;
                            default:
                                return undefined;
                        }
                        var sym = useLhsSymbol ? checker.getSymbolAtLocation(ts.getNameOfAccessExpression(ts.cast(node.left, ts.isAccessExpression))) : symbol;
                        return sym && exportInfo(sym, kind);
                    }
                }
                function getImport() {
                    var isImport = isNodeImport(node);
                    if (!isImport)
                        return undefined;
                    // A symbol being imported is always an alias. So get what that aliases to find the local symbol.
                    var importedSymbol = checker.getImmediateAliasedSymbol(symbol);
                    if (!importedSymbol)
                        return undefined;
                    // Search on the local symbol in the exporting module, not the exported symbol.
                    importedSymbol = skipExportSpecifierSymbol(importedSymbol, checker);
                    // Similarly, skip past the symbol for 'export ='
                    if (importedSymbol.escapedName === "export=") {
                        importedSymbol = getExportEqualsLocalSymbol(importedSymbol, checker);
                    }
                    // If the import has a different name than the export, do not continue searching.
                    // If `importedName` is undefined, do continue searching as the export is anonymous.
                    // (All imports returned from this function will be ignored anyway if we are in rename and this is a not a named export.)
                    var importedName = ts.symbolEscapedNameNoDefault(importedSymbol);
                    if (importedName === undefined || importedName === "default" /* Default */ || importedName === symbol.escapedName) {
                        return { kind: 0 /* Import */, symbol: importedSymbol };
                    }
                }
                function exportInfo(symbol, kind) {
                    var exportInfo = getExportInfo(symbol, kind, checker);
                    return exportInfo && { kind: 1 /* Export */, symbol: symbol, exportInfo: exportInfo };
                }
                // Not meant for use with export specifiers or export assignment.
                function getExportKindForDeclaration(node) {
                    return ts.hasSyntacticModifier(node, 512 /* Default */) ? 1 /* Default */ : 0 /* Named */;
                }
            }
            FindAllReferences.getImportOrExportSymbol = getImportOrExportSymbol;
            function getExportEqualsLocalSymbol(importedSymbol, checker) {
                if (importedSymbol.flags & 2097152 /* Alias */) {
                    return ts.Debug.checkDefined(checker.getImmediateAliasedSymbol(importedSymbol));
                }
                var decl = ts.Debug.checkDefined(importedSymbol.valueDeclaration);
                if (ts.isExportAssignment(decl)) { // `export = class {}`
                    return ts.Debug.checkDefined(decl.expression.symbol);
                }
                else if (ts.isBinaryExpression(decl)) { // `module.exports = class {}`
                    return ts.Debug.checkDefined(decl.right.symbol);
                }
                else if (ts.isSourceFile(decl)) { // json module
                    return ts.Debug.checkDefined(decl.symbol);
                }
                return ts.Debug.fail();
            }
            // If a reference is a class expression, the exported node would be its parent.
            // If a reference is a variable declaration, the exported node would be the variable statement.
            function getExportNode(parent, node) {
                var declaration = ts.isVariableDeclaration(parent) ? parent : ts.isBindingElement(parent) ? ts.walkUpBindingElementsAndPatterns(parent) : undefined;
                if (declaration) {
                    return parent.name !== node ? undefined :
                        ts.isCatchClause(declaration.parent) ? undefined : ts.isVariableStatement(declaration.parent.parent) ? declaration.parent.parent : undefined;
                }
                else {
                    return parent;
                }
            }
            function isNodeImport(node) {
                var parent = node.parent;
                switch (parent.kind) {
                    case 263 /* ImportEqualsDeclaration */:
                        return parent.name === node && isExternalModuleImportEquals(parent);
                    case 268 /* ImportSpecifier */:
                        // For a rename import `{ foo as bar }`, don't search for the imported symbol. Just find local uses of `bar`.
                        return !parent.propertyName;
                    case 265 /* ImportClause */:
                    case 266 /* NamespaceImport */:
                        ts.Debug.assert(parent.name === node);
                        return true;
                    case 201 /* BindingElement */:
                        return ts.isInJSFile(node) && ts.isRequireVariableDeclaration(parent);
                    default:
                        return false;
                }
            }
            function getExportInfo(exportSymbol, exportKind, checker) {
                var moduleSymbol = exportSymbol.parent;
                if (!moduleSymbol)
                    return undefined; // This can happen if an `export` is not at the top-level (which is a compile error).
                var exportingModuleSymbol = checker.getMergedSymbol(moduleSymbol); // Need to get merged symbol in case there's an augmentation.
                // `export` may appear in a namespace. In that case, just rely on global search.
                return ts.isExternalModuleSymbol(exportingModuleSymbol) ? { exportingModuleSymbol: exportingModuleSymbol, exportKind: exportKind } : undefined;
            }
            FindAllReferences.getExportInfo = getExportInfo;
            /** If at an export specifier, go to the symbol it refers to. */
            function skipExportSpecifierSymbol(symbol, checker) {
                // For `export { foo } from './bar", there's nothing to skip, because it does not create a new alias. But `export { foo } does.
                if (symbol.declarations) {
                    for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                        var declaration = _a[_i];
                        if (ts.isExportSpecifier(declaration) && !declaration.propertyName && !declaration.parent.parent.moduleSpecifier) {
                            return checker.getExportSpecifierLocalTargetSymbol(declaration);
                        }
                        else if (ts.isPropertyAccessExpression(declaration) && ts.isModuleExportsAccessExpression(declaration.expression) && !ts.isPrivateIdentifier(declaration.name)) {
                            // Export of form 'module.exports.propName = expr';
                            return checker.getSymbolAtLocation(declaration);
                        }
                        else if (ts.isShorthandPropertyAssignment(declaration)
                            && ts.isBinaryExpression(declaration.parent.parent)
                            && ts.getAssignmentDeclarationKind(declaration.parent.parent) === 2 /* ModuleExports */) {
                            return checker.getExportSpecifierLocalTargetSymbol(declaration.name);
                        }
                    }
                }
                return symbol;
            }
            function getContainingModuleSymbol(importer, checker) {
                return checker.getMergedSymbol(getSourceFileLikeForImportDeclaration(importer).symbol);
            }
            function getSourceFileLikeForImportDeclaration(node) {
                if (node.kind === 206 /* CallExpression */) {
                    return node.getSourceFile();
                }
                var parent = node.parent;
                if (parent.kind === 300 /* SourceFile */) {
                    return parent;
                }
                ts.Debug.assert(parent.kind === 260 /* ModuleBlock */);
                return ts.cast(parent.parent, isAmbientModuleDeclaration);
            }
            function isAmbientModuleDeclaration(node) {
                return node.kind === 259 /* ModuleDeclaration */ && node.name.kind === 10 /* StringLiteral */;
            }
            function isExternalModuleImportEquals(eq) {
                return eq.moduleReference.kind === 275 /* ExternalModuleReference */ && eq.moduleReference.expression.kind === 10 /* StringLiteral */;
            }
        })(FindAllReferences = ts.FindAllReferences || (ts.FindAllReferences = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var FindAllReferences;
        (function (FindAllReferences) {
            var DefinitionKind;
            (function (DefinitionKind) {
                DefinitionKind[DefinitionKind["Symbol"] = 0] = "Symbol";
                DefinitionKind[DefinitionKind["Label"] = 1] = "Label";
                DefinitionKind[DefinitionKind["Keyword"] = 2] = "Keyword";
                DefinitionKind[DefinitionKind["This"] = 3] = "This";
                DefinitionKind[DefinitionKind["String"] = 4] = "String";
                DefinitionKind[DefinitionKind["TripleSlashReference"] = 5] = "TripleSlashReference";
            })(DefinitionKind = FindAllReferences.DefinitionKind || (FindAllReferences.DefinitionKind = {}));
            var EntryKind;
            (function (EntryKind) {
                EntryKind[EntryKind["Span"] = 0] = "Span";
                EntryKind[EntryKind["Node"] = 1] = "Node";
                EntryKind[EntryKind["StringLiteral"] = 2] = "StringLiteral";
                EntryKind[EntryKind["SearchedLocalFoundProperty"] = 3] = "SearchedLocalFoundProperty";
                EntryKind[EntryKind["SearchedPropertyFoundLocal"] = 4] = "SearchedPropertyFoundLocal";
            })(EntryKind = FindAllReferences.EntryKind || (FindAllReferences.EntryKind = {}));
            function nodeEntry(node, kind) {
                if (kind === void 0) {
                    kind = 1 /* Node */;
                }
                return {
                    kind: kind,
                    node: node.name || node,
                    context: getContextNodeForNodeEntry(node)
                };
            }
            FindAllReferences.nodeEntry = nodeEntry;
            function isContextWithStartAndEndNode(node) {
                return node && node.kind === undefined;
            }
            FindAllReferences.isContextWithStartAndEndNode = isContextWithStartAndEndNode;
            function getContextNodeForNodeEntry(node) {
                if (ts.isDeclaration(node)) {
                    return getContextNode(node);
                }
                if (!node.parent)
                    return undefined;
                if (!ts.isDeclaration(node.parent) && !ts.isExportAssignment(node.parent)) {
                    // Special property assignment in javascript
                    if (ts.isInJSFile(node)) {
                        var binaryExpression = ts.isBinaryExpression(node.parent) ?
                            node.parent :
                            ts.isAccessExpression(node.parent) &&
                                ts.isBinaryExpression(node.parent.parent) &&
                                node.parent.parent.left === node.parent ?
                                node.parent.parent :
                                undefined;
                        if (binaryExpression && ts.getAssignmentDeclarationKind(binaryExpression) !== 0 /* None */) {
                            return getContextNode(binaryExpression);
                        }
                    }
                    // Jsx Tags
                    if (ts.isJsxOpeningElement(node.parent) || ts.isJsxClosingElement(node.parent)) {
                        return node.parent.parent;
                    }
                    else if (ts.isJsxSelfClosingElement(node.parent) ||
                        ts.isLabeledStatement(node.parent) ||
                        ts.isBreakOrContinueStatement(node.parent)) {
                        return node.parent;
                    }
                    else if (ts.isStringLiteralLike(node)) {
                        var validImport = ts.tryGetImportFromModuleSpecifier(node);
                        if (validImport) {
                            var declOrStatement = ts.findAncestor(validImport, function (node) {
                                return ts.isDeclaration(node) ||
                                    ts.isStatement(node) ||
                                    ts.isJSDocTag(node);
                            });
                            return ts.isDeclaration(declOrStatement) ?
                                getContextNode(declOrStatement) :
                                declOrStatement;
                        }
                    }
                    // Handle computed property name
                    var propertyName = ts.findAncestor(node, ts.isComputedPropertyName);
                    return propertyName ?
                        getContextNode(propertyName.parent) :
                        undefined;
                }
                if (node.parent.name === node || // node is name of declaration, use parent
                    ts.isConstructorDeclaration(node.parent) ||
                    ts.isExportAssignment(node.parent) ||
                    // Property name of the import export specifier or binding pattern, use parent
                    ((ts.isImportOrExportSpecifier(node.parent) || ts.isBindingElement(node.parent))
                        && node.parent.propertyName === node) ||
                    // Is default export
                    (node.kind === 88 /* DefaultKeyword */ && ts.hasSyntacticModifier(node.parent, 513 /* ExportDefault */))) {
                    return getContextNode(node.parent);
                }
                return undefined;
            }
            function getContextNode(node) {
                if (!node)
                    return undefined;
                switch (node.kind) {
                    case 252 /* VariableDeclaration */:
                        return !ts.isVariableDeclarationList(node.parent) || node.parent.declarations.length !== 1 ?
                            node :
                            ts.isVariableStatement(node.parent.parent) ?
                                node.parent.parent :
                                ts.isForInOrOfStatement(node.parent.parent) ?
                                    getContextNode(node.parent.parent) :
                                    node.parent;
                    case 201 /* BindingElement */:
                        return getContextNode(node.parent.parent);
                    case 268 /* ImportSpecifier */:
                        return node.parent.parent.parent;
                    case 273 /* ExportSpecifier */:
                    case 266 /* NamespaceImport */:
                        return node.parent.parent;
                    case 265 /* ImportClause */:
                    case 272 /* NamespaceExport */:
                        return node.parent;
                    case 219 /* BinaryExpression */:
                        return ts.isExpressionStatement(node.parent) ?
                            node.parent :
                            node;
                    case 242 /* ForOfStatement */:
                    case 241 /* ForInStatement */:
                        return {
                            start: node.initializer,
                            end: node.expression
                        };
                    case 291 /* PropertyAssignment */:
                    case 292 /* ShorthandPropertyAssignment */:
                        return ts.isArrayLiteralOrObjectLiteralDestructuringPattern(node.parent) ?
                            getContextNode(ts.findAncestor(node.parent, function (node) {
                                return ts.isBinaryExpression(node) || ts.isForInOrOfStatement(node);
                            })) :
                            node;
                    default:
                        return node;
                }
            }
            FindAllReferences.getContextNode = getContextNode;
            function toContextSpan(textSpan, sourceFile, context) {
                if (!context)
                    return undefined;
                var contextSpan = isContextWithStartAndEndNode(context) ?
                    getTextSpan(context.start, sourceFile, context.end) :
                    getTextSpan(context, sourceFile);
                return contextSpan.start !== textSpan.start || contextSpan.length !== textSpan.length ?
                    { contextSpan: contextSpan } :
                    undefined;
            }
            FindAllReferences.toContextSpan = toContextSpan;
            var FindReferencesUse;
            (function (FindReferencesUse) {
                /**
                 * When searching for references to a symbol, the location will not be adjusted (this is the default behavior when not specified).
                 */
                FindReferencesUse[FindReferencesUse["Other"] = 0] = "Other";
                /**
                 * When searching for references to a symbol, the location will be adjusted if the cursor was on a keyword.
                 */
                FindReferencesUse[FindReferencesUse["References"] = 1] = "References";
                /**
                 * When searching for references to a symbol, the location will be adjusted if the cursor was on a keyword.
                 * Unlike `References`, the location will only be adjusted keyword belonged to a declaration with a valid name.
                 * If set, we will find fewer references -- if it is referenced by several different names, we still only find references for the original name.
                 */
                FindReferencesUse[FindReferencesUse["Rename"] = 2] = "Rename";
            })(FindReferencesUse = FindAllReferences.FindReferencesUse || (FindAllReferences.FindReferencesUse = {}));
            function findReferencedSymbols(program, cancellationToken, sourceFiles, sourceFile, position) {
                var node = ts.getTouchingPropertyName(sourceFile, position);
                var referencedSymbols = Core.getReferencedSymbolsForNode(position, node, program, sourceFiles, cancellationToken, { use: 1 /* References */ });
                var checker = program.getTypeChecker();
                return !referencedSymbols || !referencedSymbols.length ? undefined : ts.mapDefined(referencedSymbols, function (_a) {
                    var definition = _a.definition, references = _a.references;
                    // Only include referenced symbols that have a valid definition.
                    return definition && {
                        definition: checker.runWithCancellationToken(cancellationToken, function (checker) { return definitionToReferencedSymbolDefinitionInfo(definition, checker, node); }),
                        references: references.map(toReferenceEntry)
                    };
                });
            }
            FindAllReferences.findReferencedSymbols = findReferencedSymbols;
            function getImplementationsAtPosition(program, cancellationToken, sourceFiles, sourceFile, position) {
                var node = ts.getTouchingPropertyName(sourceFile, position);
                var referenceEntries;
                var entries = getImplementationReferenceEntries(program, cancellationToken, sourceFiles, node, position);
                if (node.parent.kind === 204 /* PropertyAccessExpression */
                    || node.parent.kind === 201 /* BindingElement */
                    || node.parent.kind === 205 /* ElementAccessExpression */
                    || node.kind === 106 /* SuperKeyword */) {
                    referenceEntries = entries && __spreadArray([], entries, true);
                }
                else {
                    var queue = entries && __spreadArray([], entries, true);
                    var seenNodes = new ts.Map();
                    while (queue && queue.length) {
                        var entry = queue.shift();
                        if (!ts.addToSeen(seenNodes, ts.getNodeId(entry.node))) {
                            continue;
                        }
                        referenceEntries = ts.append(referenceEntries, entry);
                        var entries_1 = getImplementationReferenceEntries(program, cancellationToken, sourceFiles, entry.node, entry.node.pos);
                        if (entries_1) {
                            queue.push.apply(queue, entries_1);
                        }
                    }
                }
                var checker = program.getTypeChecker();
                return ts.map(referenceEntries, function (entry) { return toImplementationLocation(entry, checker); });
            }
            FindAllReferences.getImplementationsAtPosition = getImplementationsAtPosition;
            function getImplementationReferenceEntries(program, cancellationToken, sourceFiles, node, position) {
                if (node.kind === 300 /* SourceFile */) {
                    return undefined;
                }
                var checker = program.getTypeChecker();
                // If invoked directly on a shorthand property assignment, then return
                // the declaration of the symbol being assigned (not the symbol being assigned to).
                if (node.parent.kind === 292 /* ShorthandPropertyAssignment */) {
                    var result_2 = [];
                    Core.getReferenceEntriesForShorthandPropertyAssignment(node, checker, function (node) { return result_2.push(nodeEntry(node)); });
                    return result_2;
                }
                else if (node.kind === 106 /* SuperKeyword */ || ts.isSuperProperty(node.parent)) {
                    // References to and accesses on the super keyword only have one possible implementation, so no
                    // need to "Find all References"
                    var symbol = checker.getSymbolAtLocation(node);
                    return symbol.valueDeclaration && [nodeEntry(symbol.valueDeclaration)];
                }
                else {
                    // Perform "Find all References" and retrieve only those that are implementations
                    return getReferenceEntriesForNode(position, node, program, sourceFiles, cancellationToken, { implementations: true, use: 1 /* References */ });
                }
            }
            function findReferenceOrRenameEntries(program, cancellationToken, sourceFiles, node, position, options, convertEntry) {
                return ts.map(flattenEntries(Core.getReferencedSymbolsForNode(position, node, program, sourceFiles, cancellationToken, options)), function (entry) { return convertEntry(entry, node, program.getTypeChecker()); });
            }
            FindAllReferences.findReferenceOrRenameEntries = findReferenceOrRenameEntries;
            function getReferenceEntriesForNode(position, node, program, sourceFiles, cancellationToken, options, sourceFilesSet) {
                if (options === void 0) {
                    options = {};
                }
                if (sourceFilesSet === void 0) {
                    sourceFilesSet = new ts.Set(sourceFiles.map(function (f) { return f.fileName; }));
                }
                return flattenEntries(Core.getReferencedSymbolsForNode(position, node, program, sourceFiles, cancellationToken, options, sourceFilesSet));
            }
            FindAllReferences.getReferenceEntriesForNode = getReferenceEntriesForNode;
            function flattenEntries(referenceSymbols) {
                return referenceSymbols && ts.flatMap(referenceSymbols, function (r) { return r.references; });
            }
            function definitionToReferencedSymbolDefinitionInfo(def, checker, originalNode) {
                var info = (function () {
                    switch (def.type) {
                        case 0 /* Symbol */: {
                            var symbol = def.symbol;
                            var _a = getDefinitionKindAndDisplayParts(symbol, checker, originalNode), displayParts_1 = _a.displayParts, kind_1 = _a.kind;
                            var name_1 = displayParts_1.map(function (p) { return p.text; }).join("");
                            var declaration = symbol.declarations && ts.firstOrUndefined(symbol.declarations);
                            var node = declaration ? (ts.getNameOfDeclaration(declaration) || declaration) : originalNode;
                            return __assign(__assign({}, getFileAndTextSpanFromNode(node)), { name: name_1, kind: kind_1, displayParts: displayParts_1, context: getContextNode(declaration) });
                        }
                        case 1 /* Label */: {
                            var node = def.node;
                            return __assign(__assign({}, getFileAndTextSpanFromNode(node)), { name: node.text, kind: "label" /* label */, displayParts: [ts.displayPart(node.text, ts.SymbolDisplayPartKind.text)] });
                        }
                        case 2 /* Keyword */: {
                            var node = def.node;
                            var name_2 = ts.tokenToString(node.kind);
                            return __assign(__assign({}, getFileAndTextSpanFromNode(node)), { name: name_2, kind: "keyword" /* keyword */, displayParts: [{ text: name_2, kind: "keyword" /* keyword */ }] });
                        }
                        case 3 /* This */: {
                            var node = def.node;
                            var symbol = checker.getSymbolAtLocation(node);
                            var displayParts_2 = symbol && ts.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(checker, symbol, node.getSourceFile(), ts.getContainerNode(node), node).displayParts || [ts.textPart("this")];
                            return __assign(__assign({}, getFileAndTextSpanFromNode(node)), { name: "this", kind: "var" /* variableElement */, displayParts: displayParts_2 });
                        }
                        case 4 /* String */: {
                            var node = def.node;
                            return __assign(__assign({}, getFileAndTextSpanFromNode(node)), { name: node.text, kind: "var" /* variableElement */, displayParts: [ts.displayPart(ts.getTextOfNode(node), ts.SymbolDisplayPartKind.stringLiteral)] });
                        }
                        case 5 /* TripleSlashReference */: {
                            return {
                                textSpan: ts.createTextSpanFromRange(def.reference),
                                sourceFile: def.file,
                                name: def.reference.fileName,
                                kind: "string" /* string */,
                                displayParts: [ts.displayPart("\"" + def.reference.fileName + "\"", ts.SymbolDisplayPartKind.stringLiteral)]
                            };
                        }
                        default:
                            return ts.Debug.assertNever(def);
                    }
                })();
                var sourceFile = info.sourceFile, textSpan = info.textSpan, name = info.name, kind = info.kind, displayParts = info.displayParts, context = info.context;
                return __assign({ containerKind: "" /* unknown */, containerName: "", fileName: sourceFile.fileName, kind: kind, name: name, textSpan: textSpan, displayParts: displayParts }, toContextSpan(textSpan, sourceFile, context));
            }
            function getFileAndTextSpanFromNode(node) {
                var sourceFile = node.getSourceFile();
                return {
                    sourceFile: sourceFile,
                    textSpan: getTextSpan(ts.isComputedPropertyName(node) ? node.expression : node, sourceFile)
                };
            }
            function getDefinitionKindAndDisplayParts(symbol, checker, node) {
                var meaning = Core.getIntersectingMeaningFromDeclarations(node, symbol);
                var enclosingDeclaration = symbol.declarations && ts.firstOrUndefined(symbol.declarations) || node;
                var _a = ts.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(checker, symbol, enclosingDeclaration.getSourceFile(), enclosingDeclaration, enclosingDeclaration, meaning), displayParts = _a.displayParts, symbolKind = _a.symbolKind;
                return { displayParts: displayParts, kind: symbolKind };
            }
            function toRenameLocation(entry, originalNode, checker, providePrefixAndSuffixText) {
                return __assign(__assign({}, entryToDocumentSpan(entry)), (providePrefixAndSuffixText && getPrefixAndSuffixText(entry, originalNode, checker)));
            }
            FindAllReferences.toRenameLocation = toRenameLocation;
            function toReferenceEntry(entry) {
                var documentSpan = entryToDocumentSpan(entry);
                if (entry.kind === 0 /* Span */) {
                    return __assign(__assign({}, documentSpan), { isWriteAccess: false, isDefinition: false });
                }
                var kind = entry.kind, node = entry.node;
                return __assign(__assign({}, documentSpan), { isWriteAccess: isWriteAccessForReference(node), isDefinition: isDefinitionForReference(node), isInString: kind === 2 /* StringLiteral */ ? true : undefined });
            }
            FindAllReferences.toReferenceEntry = toReferenceEntry;
            function entryToDocumentSpan(entry) {
                if (entry.kind === 0 /* Span */) {
                    return { textSpan: entry.textSpan, fileName: entry.fileName };
                }
                else {
                    var sourceFile = entry.node.getSourceFile();
                    var textSpan = getTextSpan(entry.node, sourceFile);
                    return __assign({ textSpan: textSpan, fileName: sourceFile.fileName }, toContextSpan(textSpan, sourceFile, entry.context));
                }
            }
            function getPrefixAndSuffixText(entry, originalNode, checker) {
                if (entry.kind !== 0 /* Span */ && ts.isIdentifier(originalNode)) {
                    var node = entry.node, kind = entry.kind;
                    var parent = node.parent;
                    var name = originalNode.text;
                    var isShorthandAssignment = ts.isShorthandPropertyAssignment(parent);
                    if (isShorthandAssignment || (ts.isObjectBindingElementWithoutPropertyName(parent) && parent.name === node && parent.dotDotDotToken === undefined)) {
                        var prefixColon = { prefixText: name + ": " };
                        var suffixColon = { suffixText: ": " + name };
                        if (kind === 3 /* SearchedLocalFoundProperty */) {
                            return prefixColon;
                        }
                        if (kind === 4 /* SearchedPropertyFoundLocal */) {
                            return suffixColon;
                        }
                        // In `const o = { x }; o.x`, symbolAtLocation at `x` in `{ x }` is the property symbol.
                        // For a binding element `const { x } = o;`, symbolAtLocation at `x` is the property symbol.
                        if (isShorthandAssignment) {
                            var grandParent = parent.parent;
                            if (ts.isObjectLiteralExpression(grandParent) &&
                                ts.isBinaryExpression(grandParent.parent) &&
                                ts.isModuleExportsAccessExpression(grandParent.parent.left)) {
                                return prefixColon;
                            }
                            return suffixColon;
                        }
                        else {
                            return prefixColon;
                        }
                    }
                    else if (ts.isImportSpecifier(parent) && !parent.propertyName) {
                        // If the original symbol was using this alias, just rename the alias.
                        var originalSymbol = ts.isExportSpecifier(originalNode.parent) ? checker.getExportSpecifierLocalTargetSymbol(originalNode.parent) : checker.getSymbolAtLocation(originalNode);
                        return ts.contains(originalSymbol.declarations, parent) ? { prefixText: name + " as " } : ts.emptyOptions;
                    }
                    else if (ts.isExportSpecifier(parent) && !parent.propertyName) {
                        // If the symbol for the node is same as declared node symbol use prefix text
                        return originalNode === entry.node || checker.getSymbolAtLocation(originalNode) === checker.getSymbolAtLocation(entry.node) ?
                            { prefixText: name + " as " } :
                            { suffixText: " as " + name };
                    }
                }
                return ts.emptyOptions;
            }
            function toImplementationLocation(entry, checker) {
                var documentSpan = entryToDocumentSpan(entry);
                if (entry.kind !== 0 /* Span */) {
                    var node = entry.node;
                    return __assign(__assign({}, documentSpan), implementationKindDisplayParts(node, checker));
                }
                else {
                    return __assign(__assign({}, documentSpan), { kind: "" /* unknown */, displayParts: [] });
                }
            }
            function implementationKindDisplayParts(node, checker) {
                var symbol = checker.getSymbolAtLocation(ts.isDeclaration(node) && node.name ? node.name : node);
                if (symbol) {
                    return getDefinitionKindAndDisplayParts(symbol, checker, node);
                }
                else if (node.kind === 203 /* ObjectLiteralExpression */) {
                    return {
                        kind: "interface" /* interfaceElement */,
                        displayParts: [ts.punctuationPart(20 /* OpenParenToken */), ts.textPart("object literal"), ts.punctuationPart(21 /* CloseParenToken */)]
                    };
                }
                else if (node.kind === 224 /* ClassExpression */) {
                    return {
                        kind: "local class" /* localClassElement */,
                        displayParts: [ts.punctuationPart(20 /* OpenParenToken */), ts.textPart("anonymous local class"), ts.punctuationPart(21 /* CloseParenToken */)]
                    };
                }
                else {
                    return { kind: ts.getNodeKind(node), displayParts: [] };
                }
            }
            function toHighlightSpan(entry) {
                var documentSpan = entryToDocumentSpan(entry);
                if (entry.kind === 0 /* Span */) {
                    return {
                        fileName: documentSpan.fileName,
                        span: {
                            textSpan: documentSpan.textSpan,
                            kind: "reference" /* reference */
                        }
                    };
                }
                var writeAccess = isWriteAccessForReference(entry.node);
                var span = __assign({ textSpan: documentSpan.textSpan, kind: writeAccess ? "writtenReference" /* writtenReference */ : "reference" /* reference */, isInString: entry.kind === 2 /* StringLiteral */ ? true : undefined }, documentSpan.contextSpan && { contextSpan: documentSpan.contextSpan });
                return { fileName: documentSpan.fileName, span: span };
            }
            FindAllReferences.toHighlightSpan = toHighlightSpan;
            function getTextSpan(node, sourceFile, endNode) {
                var start = node.getStart(sourceFile);
                var end = (endNode || node).getEnd();
                if (ts.isStringLiteralLike(node)) {
                    ts.Debug.assert(endNode === undefined);
                    start += 1;
                    end -= 1;
                }
                return ts.createTextSpanFromBounds(start, end);
            }
            function getTextSpanOfEntry(entry) {
                return entry.kind === 0 /* Span */ ? entry.textSpan :
                    getTextSpan(entry.node, entry.node.getSourceFile());
            }
            FindAllReferences.getTextSpanOfEntry = getTextSpanOfEntry;
            /** A node is considered a writeAccess iff it is a name of a declaration or a target of an assignment */
            function isWriteAccessForReference(node) {
                var decl = ts.getDeclarationFromName(node);
                return !!decl && declarationIsWriteAccess(decl) || node.kind === 88 /* DefaultKeyword */ || ts.isWriteAccess(node);
            }
            function isDefinitionForReference(node) {
                return node.kind === 88 /* DefaultKeyword */
                    || !!ts.getDeclarationFromName(node)
                    || ts.isLiteralComputedPropertyDeclarationName(node)
                    || (node.kind === 133 /* ConstructorKeyword */ && ts.isConstructorDeclaration(node.parent));
            }
            /**
             * True if 'decl' provides a value, as in `function f() {}`;
             * false if 'decl' is just a location for a future write, as in 'let x;'
             */
            function declarationIsWriteAccess(decl) {
                // Consider anything in an ambient declaration to be a write access since it may be coming from JS.
                if (!!(decl.flags & 8388608 /* Ambient */))
                    return true;
                switch (decl.kind) {
                    case 219 /* BinaryExpression */:
                    case 201 /* BindingElement */:
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                    case 88 /* DefaultKeyword */:
                    case 258 /* EnumDeclaration */:
                    case 294 /* EnumMember */:
                    case 273 /* ExportSpecifier */:
                    case 265 /* ImportClause */: // default import
                    case 263 /* ImportEqualsDeclaration */:
                    case 268 /* ImportSpecifier */:
                    case 256 /* InterfaceDeclaration */:
                    case 333 /* JSDocCallbackTag */:
                    case 340 /* JSDocTypedefTag */:
                    case 283 /* JsxAttribute */:
                    case 259 /* ModuleDeclaration */:
                    case 262 /* NamespaceExportDeclaration */:
                    case 266 /* NamespaceImport */:
                    case 272 /* NamespaceExport */:
                    case 162 /* Parameter */:
                    case 292 /* ShorthandPropertyAssignment */:
                    case 257 /* TypeAliasDeclaration */:
                    case 161 /* TypeParameter */:
                        return true;
                    case 291 /* PropertyAssignment */:
                        // In `({ x: y } = 0);`, `x` is not a write access. (Won't call this function for `y`.)
                        return !ts.isArrayLiteralOrObjectLiteralDestructuringPattern(decl.parent);
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 169 /* Constructor */:
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return !!decl.body;
                    case 252 /* VariableDeclaration */:
                    case 165 /* PropertyDeclaration */:
                        return !!decl.initializer || ts.isCatchClause(decl.parent);
                    case 166 /* MethodSignature */:
                    case 164 /* PropertySignature */:
                    case 342 /* JSDocPropertyTag */:
                    case 335 /* JSDocParameterTag */:
                        return false;
                    default:
                        return ts.Debug.failBadSyntaxKind(decl);
                }
            }
            /** Encapsulates the core find-all-references algorithm. */
            var Core;
            (function (Core) {
                /** Core find-all-references algorithm. Handles special cases before delegating to `getReferencedSymbolsForSymbol`. */
                function getReferencedSymbolsForNode(position, node, program, sourceFiles, cancellationToken, options, sourceFilesSet) {
                    var _a, _b;
                    if (options === void 0) {
                        options = {};
                    }
                    if (sourceFilesSet === void 0) {
                        sourceFilesSet = new ts.Set(sourceFiles.map(function (f) { return f.fileName; }));
                    }
                    if (options.use === 1 /* References */) {
                        node = ts.getAdjustedReferenceLocation(node);
                    }
                    else if (options.use === 2 /* Rename */) {
                        node = ts.getAdjustedRenameLocation(node);
                    }
                    if (ts.isSourceFile(node)) {
                        var resolvedRef = ts.GoToDefinition.getReferenceAtPosition(node, position, program);
                        if (!(resolvedRef === null || resolvedRef === void 0 ? void 0 : resolvedRef.file)) {
                            return undefined;
                        }
                        var moduleSymbol = program.getTypeChecker().getMergedSymbol(resolvedRef.file.symbol);
                        if (moduleSymbol) {
                            return getReferencedSymbolsForModule(program, moduleSymbol, /*excludeImportTypeOfExportEquals*/ false, sourceFiles, sourceFilesSet);
                        }
                        var fileIncludeReasons = program.getFileIncludeReasons();
                        if (!fileIncludeReasons) {
                            return undefined;
                        }
                        return [{
                                definition: { type: 5 /* TripleSlashReference */, reference: resolvedRef.reference, file: node },
                                references: getReferencesForNonModule(resolvedRef.file, fileIncludeReasons, program) || ts.emptyArray
                            }];
                    }
                    if (!options.implementations) {
                        var special = getReferencedSymbolsSpecial(node, sourceFiles, cancellationToken);
                        if (special) {
                            return special;
                        }
                    }
                    var checker = program.getTypeChecker();
                    // constructors should use the class symbol, detected by name, if present
                    var symbol = checker.getSymbolAtLocation(ts.isConstructorDeclaration(node) && node.parent.name || node);
                    // Could not find a symbol e.g. unknown identifier
                    if (!symbol) {
                        // String literal might be a property (and thus have a symbol), so do this here rather than in getReferencedSymbolsSpecial.
                        if (!options.implementations && ts.isStringLiteralLike(node)) {
                            if (ts.isModuleSpecifierLike(node)) {
                                var fileIncludeReasons = program.getFileIncludeReasons();
                                var referencedFileName = (_b = (_a = node.getSourceFile().resolvedModules) === null || _a === void 0 ? void 0 : _a.get(node.text)) === null || _b === void 0 ? void 0 : _b.resolvedFileName;
                                var referencedFile = referencedFileName ? program.getSourceFile(referencedFileName) : undefined;
                                if (referencedFile) {
                                    return [{ definition: { type: 4 /* String */, node: node }, references: getReferencesForNonModule(referencedFile, fileIncludeReasons, program) || ts.emptyArray }];
                                }
                                // Fall through to string literal references. This is not very likely to return
                                // anything useful, but I guess it's better than nothing, and there's an existing
                                // test that expects this to happen (fourslash/cases/untypedModuleImport.ts).
                            }
                            return getReferencesForStringLiteral(node, sourceFiles, checker, cancellationToken);
                        }
                        return undefined;
                    }
                    if (symbol.escapedName === "export=" /* ExportEquals */) {
                        return getReferencedSymbolsForModule(program, symbol.parent, /*excludeImportTypeOfExportEquals*/ false, sourceFiles, sourceFilesSet);
                    }
                    var moduleReferences = getReferencedSymbolsForModuleIfDeclaredBySourceFile(symbol, program, sourceFiles, cancellationToken, options, sourceFilesSet);
                    if (moduleReferences && !(symbol.flags & 33554432 /* Transient */)) {
                        return moduleReferences;
                    }
                    var aliasedSymbol = getMergedAliasedSymbolOfNamespaceExportDeclaration(node, symbol, checker);
                    var moduleReferencesOfExportTarget = aliasedSymbol &&
                        getReferencedSymbolsForModuleIfDeclaredBySourceFile(aliasedSymbol, program, sourceFiles, cancellationToken, options, sourceFilesSet);
                    var references = getReferencedSymbolsForSymbol(symbol, node, sourceFiles, sourceFilesSet, checker, cancellationToken, options);
                    return mergeReferences(program, moduleReferences, references, moduleReferencesOfExportTarget);
                }
                Core.getReferencedSymbolsForNode = getReferencedSymbolsForNode;
                function getReferencesForFileName(fileName, program, sourceFiles, sourceFilesSet) {
                    var _a, _b;
                    if (sourceFilesSet === void 0) {
                        sourceFilesSet = new ts.Set(sourceFiles.map(function (f) { return f.fileName; }));
                    }
                    var moduleSymbol = (_a = program.getSourceFile(fileName)) === null || _a === void 0 ? void 0 : _a.symbol;
                    if (moduleSymbol) {
                        return ((_b = getReferencedSymbolsForModule(program, moduleSymbol, /*excludeImportTypeOfExportEquals*/ false, sourceFiles, sourceFilesSet)[0]) === null || _b === void 0 ? void 0 : _b.references) || ts.emptyArray;
                    }
                    var fileIncludeReasons = program.getFileIncludeReasons();
                    var referencedFile = program.getSourceFile(fileName);
                    return referencedFile && fileIncludeReasons && getReferencesForNonModule(referencedFile, fileIncludeReasons, program) || ts.emptyArray;
                }
                Core.getReferencesForFileName = getReferencesForFileName;
                function getReferencesForNonModule(referencedFile, refFileMap, program) {
                    var entries;
                    var references = refFileMap.get(referencedFile.path) || ts.emptyArray;
                    for (var _i = 0, references_1 = references; _i < references_1.length; _i++) {
                        var ref = references_1[_i];
                        if (ts.isReferencedFile(ref)) {
                            var referencingFile = program.getSourceFileByPath(ref.file);
                            var location = ts.getReferencedFileLocation(program.getSourceFileByPath, ref);
                            if (ts.isReferenceFileLocation(location)) {
                                entries = ts.append(entries, {
                                    kind: 0 /* Span */,
                                    fileName: referencingFile.fileName,
                                    textSpan: ts.createTextSpanFromRange(location)
                                });
                            }
                        }
                    }
                    return entries;
                }
                function getMergedAliasedSymbolOfNamespaceExportDeclaration(node, symbol, checker) {
                    if (node.parent && ts.isNamespaceExportDeclaration(node.parent)) {
                        var aliasedSymbol = checker.getAliasedSymbol(symbol);
                        var targetSymbol = checker.getMergedSymbol(aliasedSymbol);
                        if (aliasedSymbol !== targetSymbol) {
                            return targetSymbol;
                        }
                    }
                    return undefined;
                }
                function getReferencedSymbolsForModuleIfDeclaredBySourceFile(symbol, program, sourceFiles, cancellationToken, options, sourceFilesSet) {
                    var moduleSourceFile = (symbol.flags & 1536 /* Module */) && symbol.declarations && ts.find(symbol.declarations, ts.isSourceFile);
                    if (!moduleSourceFile)
                        return undefined;
                    var exportEquals = symbol.exports.get("export=" /* ExportEquals */);
                    // If !!exportEquals, we're about to add references to `import("mod")` anyway, so don't double-count them.
                    var moduleReferences = getReferencedSymbolsForModule(program, symbol, !!exportEquals, sourceFiles, sourceFilesSet);
                    if (!exportEquals || !sourceFilesSet.has(moduleSourceFile.fileName))
                        return moduleReferences;
                    // Continue to get references to 'export ='.
                    var checker = program.getTypeChecker();
                    symbol = ts.skipAlias(exportEquals, checker);
                    return mergeReferences(program, moduleReferences, getReferencedSymbolsForSymbol(symbol, /*node*/ undefined, sourceFiles, sourceFilesSet, checker, cancellationToken, options));
                }
                /**
                 * Merges the references by sorting them (by file index in sourceFiles and their location in it) that point to same definition symbol
                 */
                function mergeReferences(program) {
                    var referencesToMerge = [];
                    for (var _i = 1; _i < arguments.length; _i++) {
                        referencesToMerge[_i - 1] = arguments[_i];
                    }
                    var result;
                    for (var _a = 0, referencesToMerge_1 = referencesToMerge; _a < referencesToMerge_1.length; _a++) {
                        var references = referencesToMerge_1[_a];
                        if (!references || !references.length)
                            continue;
                        if (!result) {
                            result = references;
                            continue;
                        }
                        var _loop_4 = function (entry) {
                            if (!entry.definition || entry.definition.type !== 0 /* Symbol */) {
                                result.push(entry);
                                return "continue";
                            }
                            var symbol = entry.definition.symbol;
                            var refIndex = ts.findIndex(result, function (ref) {
                                return !!ref.definition &&
                                    ref.definition.type === 0 /* Symbol */ &&
                                    ref.definition.symbol === symbol;
                            });
                            if (refIndex === -1) {
                                result.push(entry);
                                return "continue";
                            }
                            var reference = result[refIndex];
                            result[refIndex] = {
                                definition: reference.definition,
                                references: reference.references.concat(entry.references).sort(function (entry1, entry2) {
                                    var entry1File = getSourceFileIndexOfEntry(program, entry1);
                                    var entry2File = getSourceFileIndexOfEntry(program, entry2);
                                    if (entry1File !== entry2File) {
                                        return ts.compareValues(entry1File, entry2File);
                                    }
                                    var entry1Span = getTextSpanOfEntry(entry1);
                                    var entry2Span = getTextSpanOfEntry(entry2);
                                    return entry1Span.start !== entry2Span.start ?
                                        ts.compareValues(entry1Span.start, entry2Span.start) :
                                        ts.compareValues(entry1Span.length, entry2Span.length);
                                })
                            };
                        };
                        for (var _b = 0, references_2 = references; _b < references_2.length; _b++) {
                            var entry = references_2[_b];
                            _loop_4(entry);
                        }
                    }
                    return result;
                }
                function getSourceFileIndexOfEntry(program, entry) {
                    var sourceFile = entry.kind === 0 /* Span */ ?
                        program.getSourceFile(entry.fileName) :
                        entry.node.getSourceFile();
                    return program.getSourceFiles().indexOf(sourceFile);
                }
                function getReferencedSymbolsForModule(program, symbol, excludeImportTypeOfExportEquals, sourceFiles, sourceFilesSet) {
                    ts.Debug.assert(!!symbol.valueDeclaration);
                    var references = ts.mapDefined(FindAllReferences.findModuleReferences(program, sourceFiles, symbol), function (reference) {
                        if (reference.kind === "import") {
                            var parent = reference.literal.parent;
                            if (ts.isLiteralTypeNode(parent)) {
                                var importType = ts.cast(parent.parent, ts.isImportTypeNode);
                                if (excludeImportTypeOfExportEquals && !importType.qualifier) {
                                    return undefined;
                                }
                            }
                            // import("foo") with no qualifier will reference the `export =` of the module, which may be referenced anyway.
                            return nodeEntry(reference.literal);
                        }
                        else {
                            return {
                                kind: 0 /* Span */,
                                fileName: reference.referencingFile.fileName,
                                textSpan: ts.createTextSpanFromRange(reference.ref),
                            };
                        }
                    });
                    if (symbol.declarations) {
                        for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                            var decl = _a[_i];
                            switch (decl.kind) {
                                case 300 /* SourceFile */:
                                    // Don't include the source file itself. (This may not be ideal behavior, but awkward to include an entire file as a reference.)
                                    break;
                                case 259 /* ModuleDeclaration */:
                                    if (sourceFilesSet.has(decl.getSourceFile().fileName)) {
                                        references.push(nodeEntry(decl.name));
                                    }
                                    break;
                                default:
                                    // This may be merged with something.
                                    ts.Debug.assert(!!(symbol.flags & 33554432 /* Transient */), "Expected a module symbol to be declared by a SourceFile or ModuleDeclaration.");
                            }
                        }
                    }
                    var exported = symbol.exports.get("export=" /* ExportEquals */);
                    if (exported === null || exported === void 0 ? void 0 : exported.declarations) {
                        for (var _b = 0, _c = exported.declarations; _b < _c.length; _b++) {
                            var decl = _c[_b];
                            var sourceFile = decl.getSourceFile();
                            if (sourceFilesSet.has(sourceFile.fileName)) {
                                // At `module.exports = ...`, reference node is `module`
                                var node = ts.isBinaryExpression(decl) && ts.isPropertyAccessExpression(decl.left) ? decl.left.expression :
                                    ts.isExportAssignment(decl) ? ts.Debug.checkDefined(ts.findChildOfKind(decl, 93 /* ExportKeyword */, sourceFile)) :
                                        ts.getNameOfDeclaration(decl) || decl;
                                references.push(nodeEntry(node));
                            }
                        }
                    }
                    return references.length ? [{ definition: { type: 0 /* Symbol */, symbol: symbol }, references: references }] : ts.emptyArray;
                }
                /** As in a `readonly prop: any` or `constructor(readonly prop: any)`, not a `readonly any[]`. */
                function isReadonlyTypeOperator(node) {
                    return node.kind === 143 /* ReadonlyKeyword */
                        && ts.isTypeOperatorNode(node.parent)
                        && node.parent.operator === 143 /* ReadonlyKeyword */;
                }
                /** getReferencedSymbols for special node kinds. */
                function getReferencedSymbolsSpecial(node, sourceFiles, cancellationToken) {
                    if (ts.isTypeKeyword(node.kind)) {
                        // A void expression (i.e., `void foo()`) is not special, but the `void` type is.
                        if (node.kind === 114 /* VoidKeyword */ && ts.isVoidExpression(node.parent)) {
                            return undefined;
                        }
                        // A modifier readonly (like on a property declaration) is not special;
                        // a readonly type keyword (like `readonly string[]`) is.
                        if (node.kind === 143 /* ReadonlyKeyword */ && !isReadonlyTypeOperator(node)) {
                            return undefined;
                        }
                        // Likewise, when we *are* looking for a special keyword, make sure we
                        // *don’t* include readonly member modifiers.
                        return getAllReferencesForKeyword(sourceFiles, node.kind, cancellationToken, node.kind === 143 /* ReadonlyKeyword */ ? isReadonlyTypeOperator : undefined);
                    }
                    if (ts.isStaticModifier(node) && ts.isClassStaticBlockDeclaration(node.parent)) {
                        return [{ definition: { type: 2 /* Keyword */, node: node }, references: [nodeEntry(node)] }];
                    }
                    // Labels
                    if (ts.isJumpStatementTarget(node)) {
                        var labelDefinition = ts.getTargetLabel(node.parent, node.text);
                        // if we have a label definition, look within its statement for references, if not, then
                        // the label is undefined and we have no results..
                        return labelDefinition && getLabelReferencesInNode(labelDefinition.parent, labelDefinition);
                    }
                    else if (ts.isLabelOfLabeledStatement(node)) {
                        // it is a label definition and not a target, search within the parent labeledStatement
                        return getLabelReferencesInNode(node.parent, node);
                    }
                    if (ts.isThis(node)) {
                        return getReferencesForThisKeyword(node, sourceFiles, cancellationToken);
                    }
                    if (node.kind === 106 /* SuperKeyword */) {
                        return getReferencesForSuperKeyword(node);
                    }
                    return undefined;
                }
                /** Core find-all-references algorithm for a normal symbol. */
                function getReferencedSymbolsForSymbol(originalSymbol, node, sourceFiles, sourceFilesSet, checker, cancellationToken, options) {
                    var symbol = node && skipPastExportOrImportSpecifierOrUnion(originalSymbol, node, checker, /*useLocalSymbolForExportSpecifier*/ !isForRenameWithPrefixAndSuffixText(options)) || originalSymbol;
                    // Compute the meaning from the location and the symbol it references
                    var searchMeaning = node ? getIntersectingMeaningFromDeclarations(node, symbol) : 7 /* All */;
                    var result = [];
                    var state = new State(sourceFiles, sourceFilesSet, node ? getSpecialSearchKind(node) : 0 /* None */, checker, cancellationToken, searchMeaning, options, result);
                    var exportSpecifier = !isForRenameWithPrefixAndSuffixText(options) || !symbol.declarations ? undefined : ts.find(symbol.declarations, ts.isExportSpecifier);
                    if (exportSpecifier) {
                        // When renaming at an export specifier, rename the export and not the thing being exported.
                        getReferencesAtExportSpecifier(exportSpecifier.name, symbol, exportSpecifier, state.createSearch(node, originalSymbol, /*comingFrom*/ undefined), state, /*addReferencesHere*/ true, /*alwaysGetReferences*/ true);
                    }
                    else if (node && node.kind === 88 /* DefaultKeyword */ && symbol.escapedName === "default" /* Default */ && symbol.parent) {
                        addReference(node, symbol, state);
                        searchForImportsOfExport(node, symbol, { exportingModuleSymbol: symbol.parent, exportKind: 1 /* Default */ }, state);
                    }
                    else {
                        var search = state.createSearch(node, symbol, /*comingFrom*/ undefined, { allSearchSymbols: node ? populateSearchSymbolSet(symbol, node, checker, options.use === 2 /* Rename */, !!options.providePrefixAndSuffixTextForRename, !!options.implementations) : [symbol] });
                        getReferencesInContainerOrFiles(symbol, state, search);
                    }
                    return result;
                }
                function getReferencesInContainerOrFiles(symbol, state, search) {
                    // Try to get the smallest valid scope that we can limit our search to;
                    // otherwise we'll need to search globally (i.e. include each file).
                    var scope = getSymbolScope(symbol);
                    if (scope) {
                        getReferencesInContainer(scope, scope.getSourceFile(), search, state, /*addReferencesHere*/ !(ts.isSourceFile(scope) && !ts.contains(state.sourceFiles, scope)));
                    }
                    else {
                        // Global search
                        for (var _i = 0, _a = state.sourceFiles; _i < _a.length; _i++) {
                            var sourceFile = _a[_i];
                            state.cancellationToken.throwIfCancellationRequested();
                            searchForName(sourceFile, search, state);
                        }
                    }
                }
                function getSpecialSearchKind(node) {
                    switch (node.kind) {
                        case 169 /* Constructor */:
                        case 133 /* ConstructorKeyword */:
                            return 1 /* Constructor */;
                        case 79 /* Identifier */:
                            if (ts.isClassLike(node.parent)) {
                                ts.Debug.assert(node.parent.name === node);
                                return 2 /* Class */;
                            }
                        // falls through
                        default:
                            return 0 /* None */;
                    }
                }
                /** Handle a few special cases relating to export/import specifiers. */
                function skipPastExportOrImportSpecifierOrUnion(symbol, node, checker, useLocalSymbolForExportSpecifier) {
                    var parent = node.parent;
                    if (ts.isExportSpecifier(parent) && useLocalSymbolForExportSpecifier) {
                        return getLocalSymbolForExportSpecifier(node, symbol, parent, checker);
                    }
                    // If the symbol is declared as part of a declaration like `{ type: "a" } | { type: "b" }`, use the property on the union type to get more references.
                    return ts.firstDefined(symbol.declarations, function (decl) {
                        if (!decl.parent) {
                            // Ignore UMD module and global merge
                            if (symbol.flags & 33554432 /* Transient */)
                                return undefined;
                            // Assertions for GH#21814. We should be handling SourceFile symbols in `getReferencedSymbolsForModule` instead of getting here.
                            ts.Debug.fail("Unexpected symbol at " + ts.Debug.formatSyntaxKind(node.kind) + ": " + ts.Debug.formatSymbol(symbol));
                        }
                        return ts.isTypeLiteralNode(decl.parent) && ts.isUnionTypeNode(decl.parent.parent)
                            ? checker.getPropertyOfType(checker.getTypeFromTypeNode(decl.parent.parent), symbol.name)
                            : undefined;
                    });
                }
                var SpecialSearchKind;
                (function (SpecialSearchKind) {
                    SpecialSearchKind[SpecialSearchKind["None"] = 0] = "None";
                    SpecialSearchKind[SpecialSearchKind["Constructor"] = 1] = "Constructor";
                    SpecialSearchKind[SpecialSearchKind["Class"] = 2] = "Class";
                })(SpecialSearchKind || (SpecialSearchKind = {}));
                function getNonModuleSymbolOfMergedModuleSymbol(symbol) {
                    if (!(symbol.flags & (1536 /* Module */ | 33554432 /* Transient */)))
                        return undefined;
                    var decl = symbol.declarations && ts.find(symbol.declarations, function (d) { return !ts.isSourceFile(d) && !ts.isModuleDeclaration(d); });
                    return decl && decl.symbol;
                }
                /**
                 * Holds all state needed for the finding references.
                 * Unlike `Search`, there is only one `State`.
                 */
                var State = /** @class */ (function () {
                    function State(sourceFiles, sourceFilesSet, specialSearchKind, checker, cancellationToken, searchMeaning, options, result) {
                        this.sourceFiles = sourceFiles;
                        this.sourceFilesSet = sourceFilesSet;
                        this.specialSearchKind = specialSearchKind;
                        this.checker = checker;
                        this.cancellationToken = cancellationToken;
                        this.searchMeaning = searchMeaning;
                        this.options = options;
                        this.result = result;
                        /** Cache for `explicitlyinheritsFrom`. */
                        this.inheritsFromCache = new ts.Map();
                        /**
                         * Type nodes can contain multiple references to the same type. For example:
                         *      let x: Foo & (Foo & Bar) = ...
                         * Because we are returning the implementation locations and not the identifier locations,
                         * duplicate entries would be returned here as each of the type references is part of
                         * the same implementation. For that reason, check before we add a new entry.
                         */
                        this.markSeenContainingTypeReference = ts.nodeSeenTracker();
                        /**
                         * It's possible that we will encounter the right side of `export { foo as bar } from "x";` more than once.
                         * For example:
                         *     // b.ts
                         *     export { foo as bar } from "./a";
                         *     import { bar } from "./b";
                         *
                         * Normally at `foo as bar` we directly add `foo` and do not locally search for it (since it doesn't declare a local).
                         * But another reference to it may appear in the same source file.
                         * See `tests/cases/fourslash/transitiveExportImports3.ts`.
                         */
                        this.markSeenReExportRHS = ts.nodeSeenTracker();
                        this.symbolIdToReferences = [];
                        // Source file ID → symbol ID → Whether the symbol has been searched for in the source file.
                        this.sourceFileToSeenSymbols = [];
                    }
                    State.prototype.includesSourceFile = function (sourceFile) {
                        return this.sourceFilesSet.has(sourceFile.fileName);
                    };
                    /** Gets every place to look for references of an exported symbols. See `ImportsResult` in `importTracker.ts` for more documentation. */
                    State.prototype.getImportSearches = function (exportSymbol, exportInfo) {
                        if (!this.importTracker)
                            this.importTracker = FindAllReferences.createImportTracker(this.sourceFiles, this.sourceFilesSet, this.checker, this.cancellationToken);
                        return this.importTracker(exportSymbol, exportInfo, this.options.use === 2 /* Rename */);
                    };
                    /** @param allSearchSymbols set of additional symbols for use by `includes`. */
                    State.prototype.createSearch = function (location, symbol, comingFrom, searchOptions) {
                        if (searchOptions === void 0) {
                            searchOptions = {};
                        }
                        // Note: if this is an external module symbol, the name doesn't include quotes.
                        // Note: getLocalSymbolForExportDefault handles `export default class C {}`, but not `export default C` or `export { C as default }`.
                        // The other two forms seem to be handled downstream (e.g. in `skipPastExportOrImportSpecifier`), so special-casing the first form
                        // here appears to be intentional).
                        var _a = searchOptions.text, text = _a === void 0 ? ts.stripQuotes(ts.symbolName(ts.getLocalSymbolForExportDefault(symbol) || getNonModuleSymbolOfMergedModuleSymbol(symbol) || symbol)) : _a, _b = searchOptions.allSearchSymbols, allSearchSymbols = _b === void 0 ? [symbol] : _b;
                        var escapedText = ts.escapeLeadingUnderscores(text);
                        var parents = this.options.implementations && location ? getParentSymbolsOfPropertyAccess(location, symbol, this.checker) : undefined;
                        return { symbol: symbol, comingFrom: comingFrom, text: text, escapedText: escapedText, parents: parents, allSearchSymbols: allSearchSymbols, includes: function (sym) { return ts.contains(allSearchSymbols, sym); } };
                    };
                    /**
                     * Callback to add references for a particular searched symbol.
                     * This initializes a reference group, so only call this if you will add at least one reference.
                     */
                    State.prototype.referenceAdder = function (searchSymbol) {
                        var symbolId = ts.getSymbolId(searchSymbol);
                        var references = this.symbolIdToReferences[symbolId];
                        if (!references) {
                            references = this.symbolIdToReferences[symbolId] = [];
                            this.result.push({ definition: { type: 0 /* Symbol */, symbol: searchSymbol }, references: references });
                        }
                        return function (node, kind) { return references.push(nodeEntry(node, kind)); };
                    };
                    /** Add a reference with no associated definition. */
                    State.prototype.addStringOrCommentReference = function (fileName, textSpan) {
                        this.result.push({
                            definition: undefined,
                            references: [{ kind: 0 /* Span */, fileName: fileName, textSpan: textSpan }]
                        });
                    };
                    /** Returns `true` the first time we search for a symbol in a file and `false` afterwards. */
                    State.prototype.markSearchedSymbols = function (sourceFile, symbols) {
                        var sourceId = ts.getNodeId(sourceFile);
                        var seenSymbols = this.sourceFileToSeenSymbols[sourceId] || (this.sourceFileToSeenSymbols[sourceId] = new ts.Set());
                        var anyNewSymbols = false;
                        for (var _i = 0, symbols_2 = symbols; _i < symbols_2.length; _i++) {
                            var sym = symbols_2[_i];
                            anyNewSymbols = ts.tryAddToSet(seenSymbols, ts.getSymbolId(sym)) || anyNewSymbols;
                        }
                        return anyNewSymbols;
                    };
                    return State;
                }());
                /** Search for all imports of a given exported symbol using `State.getImportSearches`. */
                function searchForImportsOfExport(exportLocation, exportSymbol, exportInfo, state) {
                    var _a = state.getImportSearches(exportSymbol, exportInfo), importSearches = _a.importSearches, singleReferences = _a.singleReferences, indirectUsers = _a.indirectUsers;
                    // For `import { foo as bar }` just add the reference to `foo`, and don't otherwise search in the file.
                    if (singleReferences.length) {
                        var addRef = state.referenceAdder(exportSymbol);
                        for (var _i = 0, singleReferences_1 = singleReferences; _i < singleReferences_1.length; _i++) {
                            var singleRef = singleReferences_1[_i];
                            if (shouldAddSingleReference(singleRef, state))
                                addRef(singleRef);
                        }
                    }
                    // For each import, find all references to that import in its source file.
                    for (var _b = 0, importSearches_1 = importSearches; _b < importSearches_1.length; _b++) {
                        var _c = importSearches_1[_b], importLocation = _c[0], importSymbol = _c[1];
                        getReferencesInSourceFile(importLocation.getSourceFile(), state.createSearch(importLocation, importSymbol, 1 /* Export */), state);
                    }
                    if (indirectUsers.length) {
                        var indirectSearch = void 0;
                        switch (exportInfo.exportKind) {
                            case 0 /* Named */:
                                indirectSearch = state.createSearch(exportLocation, exportSymbol, 1 /* Export */);
                                break;
                            case 1 /* Default */:
                                // Search for a property access to '.default'. This can't be renamed.
                                indirectSearch = state.options.use === 2 /* Rename */ ? undefined : state.createSearch(exportLocation, exportSymbol, 1 /* Export */, { text: "default" });
                                break;
                            case 2 /* ExportEquals */:
                                break;
                        }
                        if (indirectSearch) {
                            for (var _d = 0, indirectUsers_1 = indirectUsers; _d < indirectUsers_1.length; _d++) {
                                var indirectUser = indirectUsers_1[_d];
                                searchForName(indirectUser, indirectSearch, state);
                            }
                        }
                    }
                }
                function eachExportReference(sourceFiles, checker, cancellationToken, exportSymbol, exportingModuleSymbol, exportName, isDefaultExport, cb) {
                    var importTracker = FindAllReferences.createImportTracker(sourceFiles, new ts.Set(sourceFiles.map(function (f) { return f.fileName; })), checker, cancellationToken);
                    var _a = importTracker(exportSymbol, { exportKind: isDefaultExport ? 1 /* Default */ : 0 /* Named */, exportingModuleSymbol: exportingModuleSymbol }, /*isForRename*/ false), importSearches = _a.importSearches, indirectUsers = _a.indirectUsers;
                    for (var _i = 0, importSearches_2 = importSearches; _i < importSearches_2.length; _i++) {
                        var importLocation = importSearches_2[_i][0];
                        cb(importLocation);
                    }
                    for (var _b = 0, indirectUsers_2 = indirectUsers; _b < indirectUsers_2.length; _b++) {
                        var indirectUser = indirectUsers_2[_b];
                        for (var _c = 0, _d = getPossibleSymbolReferenceNodes(indirectUser, isDefaultExport ? "default" : exportName); _c < _d.length; _c++) {
                            var node = _d[_c];
                            // Import specifiers should be handled by importSearches
                            var symbol = checker.getSymbolAtLocation(node);
                            var hasExportAssignmentDeclaration = ts.some(symbol === null || symbol === void 0 ? void 0 : symbol.declarations, function (d) { return ts.tryCast(d, ts.isExportAssignment) ? true : false; });
                            if (ts.isIdentifier(node) && !ts.isImportOrExportSpecifier(node.parent) && (symbol === exportSymbol || hasExportAssignmentDeclaration)) {
                                cb(node);
                            }
                        }
                    }
                }
                Core.eachExportReference = eachExportReference;
                function shouldAddSingleReference(singleRef, state) {
                    if (!hasMatchingMeaning(singleRef, state))
                        return false;
                    if (state.options.use !== 2 /* Rename */)
                        return true;
                    // Don't rename an import type `import("./module-name")` when renaming `name` in `export = name;`
                    if (!ts.isIdentifier(singleRef))
                        return false;
                    // At `default` in `import { default as x }` or `export { default as x }`, do add a reference, but do not rename.
                    return !(ts.isImportOrExportSpecifier(singleRef.parent) && singleRef.escapedText === "default" /* Default */);
                }
                // Go to the symbol we imported from and find references for it.
                function searchForImportedSymbol(symbol, state) {
                    if (!symbol.declarations)
                        return;
                    for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                        var declaration = _a[_i];
                        var exportingFile = declaration.getSourceFile();
                        // Need to search in the file even if it's not in the search-file set, because it might export the symbol.
                        getReferencesInSourceFile(exportingFile, state.createSearch(declaration, symbol, 0 /* Import */), state, state.includesSourceFile(exportingFile));
                    }
                }
                /** Search for all occurences of an identifier in a source file (and filter out the ones that match). */
                function searchForName(sourceFile, search, state) {
                    if (ts.getNameTable(sourceFile).get(search.escapedText) !== undefined) {
                        getReferencesInSourceFile(sourceFile, search, state);
                    }
                }
                function getPropertySymbolOfDestructuringAssignment(location, checker) {
                    return ts.isArrayLiteralOrObjectLiteralDestructuringPattern(location.parent.parent)
                        ? checker.getPropertySymbolOfDestructuringAssignment(location)
                        : undefined;
                }
                /**
                 * Determines the smallest scope in which a symbol may have named references.
                 * Note that not every construct has been accounted for. This function can
                 * probably be improved.
                 *
                 * @returns undefined if the scope cannot be determined, implying that
                 * a reference to a symbol can occur anywhere.
                 */
                function getSymbolScope(symbol) {
                    // If this is the symbol of a named function expression or named class expression,
                    // then named references are limited to its own scope.
                    var declarations = symbol.declarations, flags = symbol.flags, parent = symbol.parent, valueDeclaration = symbol.valueDeclaration;
                    if (valueDeclaration && (valueDeclaration.kind === 211 /* FunctionExpression */ || valueDeclaration.kind === 224 /* ClassExpression */)) {
                        return valueDeclaration;
                    }
                    if (!declarations) {
                        return undefined;
                    }
                    // If this is private property or method, the scope is the containing class
                    if (flags & (4 /* Property */ | 8192 /* Method */)) {
                        var privateDeclaration = ts.find(declarations, function (d) { return ts.hasEffectiveModifier(d, 8 /* Private */) || ts.isPrivateIdentifierClassElementDeclaration(d); });
                        if (privateDeclaration) {
                            return ts.getAncestor(privateDeclaration, 255 /* ClassDeclaration */);
                        }
                        // Else this is a public property and could be accessed from anywhere.
                        return undefined;
                    }
                    // If symbol is of object binding pattern element without property name we would want to
                    // look for property too and that could be anywhere
                    if (declarations.some(ts.isObjectBindingElementWithoutPropertyName)) {
                        return undefined;
                    }
                    /*
                    If the symbol has a parent, it's globally visible unless:
                    - It's a private property (handled above).
                    - It's a type parameter.
                    - The parent is an external module: then we should only search in the module (and recurse on the export later).
                    - But if the parent has `export as namespace`, the symbol is globally visible through that namespace.
                    */
                    var exposedByParent = parent && !(symbol.flags & 262144 /* TypeParameter */);
                    if (exposedByParent && !(ts.isExternalModuleSymbol(parent) && !parent.globalExports)) {
                        return undefined;
                    }
                    var scope;
                    for (var _i = 0, declarations_1 = declarations; _i < declarations_1.length; _i++) {
                        var declaration = declarations_1[_i];
                        var container = ts.getContainerNode(declaration);
                        if (scope && scope !== container) {
                            // Different declarations have different containers, bail out
                            return undefined;
                        }
                        if (!container || container.kind === 300 /* SourceFile */ && !ts.isExternalOrCommonJsModule(container)) {
                            // This is a global variable and not an external module, any declaration defined
                            // within this scope is visible outside the file
                            return undefined;
                        }
                        scope = container;
                        if (ts.isFunctionExpression(scope)) {
                            var next = void 0;
                            while (next = ts.getNextJSDocCommentLocation(scope)) {
                                scope = next;
                            }
                        }
                    }
                    // If symbol.parent, this means we are in an export of an external module. (Otherwise we would have returned `undefined` above.)
                    // For an export of a module, we may be in a declaration file, and it may be accessed elsewhere. E.g.:
                    //     declare module "a" { export type T = number; }
                    //     declare module "b" { import { T } from "a"; export const x: T; }
                    // So we must search the whole source file. (Because we will mark the source file as seen, we we won't return to it when searching for imports.)
                    return exposedByParent ? scope.getSourceFile() : scope; // TODO: GH#18217
                }
                /** Used as a quick check for whether a symbol is used at all in a file (besides its definition). */
                function isSymbolReferencedInFile(definition, checker, sourceFile, searchContainer) {
                    if (searchContainer === void 0) {
                        searchContainer = sourceFile;
                    }
                    return eachSymbolReferenceInFile(definition, checker, sourceFile, function () { return true; }, searchContainer) || false;
                }
                Core.isSymbolReferencedInFile = isSymbolReferencedInFile;
                function eachSymbolReferenceInFile(definition, checker, sourceFile, cb, searchContainer) {
                    if (searchContainer === void 0) {
                        searchContainer = sourceFile;
                    }
                    var symbol = ts.isParameterPropertyDeclaration(definition.parent, definition.parent.parent)
                        ? ts.first(checker.getSymbolsOfParameterPropertyDeclaration(definition.parent, definition.text))
                        : checker.getSymbolAtLocation(definition);
                    if (!symbol)
                        return undefined;
                    for (var _i = 0, _a = getPossibleSymbolReferenceNodes(sourceFile, symbol.name, searchContainer); _i < _a.length; _i++) {
                        var token = _a[_i];
                        if (!ts.isIdentifier(token) || token === definition || token.escapedText !== definition.escapedText)
                            continue;
                        var referenceSymbol = checker.getSymbolAtLocation(token); // See GH#19955 for why the type annotation is necessary
                        if (referenceSymbol === symbol
                            || checker.getShorthandAssignmentValueSymbol(token.parent) === symbol
                            || ts.isExportSpecifier(token.parent) && getLocalSymbolForExportSpecifier(token, referenceSymbol, token.parent, checker) === symbol) {
                            var res = cb(token);
                            if (res)
                                return res;
                        }
                    }
                }
                Core.eachSymbolReferenceInFile = eachSymbolReferenceInFile;
                function someSignatureUsage(signature, sourceFiles, checker, cb) {
                    if (!signature.name || !ts.isIdentifier(signature.name))
                        return false;
                    var symbol = ts.Debug.checkDefined(checker.getSymbolAtLocation(signature.name));
                    for (var _i = 0, sourceFiles_3 = sourceFiles; _i < sourceFiles_3.length; _i++) {
                        var sourceFile = sourceFiles_3[_i];
                        for (var _a = 0, _b = getPossibleSymbolReferenceNodes(sourceFile, symbol.name); _a < _b.length; _a++) {
                            var name = _b[_a];
                            if (!ts.isIdentifier(name) || name === signature.name || name.escapedText !== signature.name.escapedText)
                                continue;
                            var called = ts.climbPastPropertyAccess(name);
                            var call = ts.isCallExpression(called.parent) && called.parent.expression === called ? called.parent : undefined;
                            var referenceSymbol = checker.getSymbolAtLocation(name);
                            if (referenceSymbol && checker.getRootSymbols(referenceSymbol).some(function (s) { return s === symbol; })) {
                                if (cb(name, call)) {
                                    return true;
                                }
                            }
                        }
                    }
                    return false;
                }
                Core.someSignatureUsage = someSignatureUsage;
                function getPossibleSymbolReferenceNodes(sourceFile, symbolName, container) {
                    if (container === void 0) {
                        container = sourceFile;
                    }
                    return getPossibleSymbolReferencePositions(sourceFile, symbolName, container).map(function (pos) { return ts.getTouchingPropertyName(sourceFile, pos); });
                }
                function getPossibleSymbolReferencePositions(sourceFile, symbolName, container) {
                    if (container === void 0) {
                        container = sourceFile;
                    }
                    var positions = [];
                    /// TODO: Cache symbol existence for files to save text search
                    // Also, need to make this work for unicode escapes.
                    // Be resilient in the face of a symbol with no name or zero length name
                    if (!symbolName || !symbolName.length) {
                        return positions;
                    }
                    var text = sourceFile.text;
                    var sourceLength = text.length;
                    var symbolNameLength = symbolName.length;
                    var position = text.indexOf(symbolName, container.pos);
                    while (position >= 0) {
                        // If we are past the end, stop looking
                        if (position > container.end)
                            break;
                        // We found a match.  Make sure it's not part of a larger word (i.e. the char
                        // before and after it have to be a non-identifier char).
                        var endPosition = position + symbolNameLength;
                        if ((position === 0 || !ts.isIdentifierPart(text.charCodeAt(position - 1), 99 /* Latest */)) &&
                            (endPosition === sourceLength || !ts.isIdentifierPart(text.charCodeAt(endPosition), 99 /* Latest */))) {
                            // Found a real match.  Keep searching.
                            positions.push(position);
                        }
                        position = text.indexOf(symbolName, position + symbolNameLength + 1);
                    }
                    return positions;
                }
                function getLabelReferencesInNode(container, targetLabel) {
                    var sourceFile = container.getSourceFile();
                    var labelName = targetLabel.text;
                    var references = ts.mapDefined(getPossibleSymbolReferenceNodes(sourceFile, labelName, container), function (node) {
                        // Only pick labels that are either the target label, or have a target that is the target label
                        return node === targetLabel || (ts.isJumpStatementTarget(node) && ts.getTargetLabel(node, labelName) === targetLabel) ? nodeEntry(node) : undefined;
                    });
                    return [{ definition: { type: 1 /* Label */, node: targetLabel }, references: references }];
                }
                function isValidReferencePosition(node, searchSymbolName) {
                    // Compare the length so we filter out strict superstrings of the symbol we are looking for
                    switch (node.kind) {
                        case 80 /* PrivateIdentifier */:
                            if (ts.isJSDocMemberName(node.parent)) {
                                return true;
                            }
                        // falls through I guess
                        case 79 /* Identifier */:
                            return node.text.length === searchSymbolName.length;
                        case 14 /* NoSubstitutionTemplateLiteral */:
                        case 10 /* StringLiteral */: {
                            var str = node;
                            return (ts.isLiteralNameOfPropertyDeclarationOrIndexAccess(str) || ts.isNameOfModuleDeclaration(node) || ts.isExpressionOfExternalModuleImportEqualsDeclaration(node) || (ts.isCallExpression(node.parent) && ts.isBindableObjectDefinePropertyCall(node.parent) && node.parent.arguments[1] === node)) &&
                                str.text.length === searchSymbolName.length;
                        }
                        case 8 /* NumericLiteral */:
                            return ts.isLiteralNameOfPropertyDeclarationOrIndexAccess(node) && node.text.length === searchSymbolName.length;
                        case 88 /* DefaultKeyword */:
                            return "default".length === searchSymbolName.length;
                        default:
                            return false;
                    }
                }
                function getAllReferencesForKeyword(sourceFiles, keywordKind, cancellationToken, filter) {
                    var references = ts.flatMap(sourceFiles, function (sourceFile) {
                        cancellationToken.throwIfCancellationRequested();
                        return ts.mapDefined(getPossibleSymbolReferenceNodes(sourceFile, ts.tokenToString(keywordKind), sourceFile), function (referenceLocation) {
                            if (referenceLocation.kind === keywordKind && (!filter || filter(referenceLocation))) {
                                return nodeEntry(referenceLocation);
                            }
                        });
                    });
                    return references.length ? [{ definition: { type: 2 /* Keyword */, node: references[0].node }, references: references }] : undefined;
                }
                function getReferencesInSourceFile(sourceFile, search, state, addReferencesHere) {
                    if (addReferencesHere === void 0) {
                        addReferencesHere = true;
                    }
                    state.cancellationToken.throwIfCancellationRequested();
                    return getReferencesInContainer(sourceFile, sourceFile, search, state, addReferencesHere);
                }
                /**
                 * Search within node "container" for references for a search value, where the search value is defined as a
                 * tuple of(searchSymbol, searchText, searchLocation, and searchMeaning).
                 * searchLocation: a node where the search value
                 */
                function getReferencesInContainer(container, sourceFile, search, state, addReferencesHere) {
                    if (!state.markSearchedSymbols(sourceFile, search.allSearchSymbols)) {
                        return;
                    }
                    for (var _i = 0, _a = getPossibleSymbolReferencePositions(sourceFile, search.text, container); _i < _a.length; _i++) {
                        var position = _a[_i];
                        getReferencesAtLocation(sourceFile, position, search, state, addReferencesHere);
                    }
                }
                function hasMatchingMeaning(referenceLocation, state) {
                    return !!(ts.getMeaningFromLocation(referenceLocation) & state.searchMeaning);
                }
                function getReferencesAtLocation(sourceFile, position, search, state, addReferencesHere) {
                    var referenceLocation = ts.getTouchingPropertyName(sourceFile, position);
                    if (!isValidReferencePosition(referenceLocation, search.text)) {
                        // This wasn't the start of a token.  Check to see if it might be a
                        // match in a comment or string if that's what the caller is asking
                        // for.
                        if (!state.options.implementations && (state.options.findInStrings && ts.isInString(sourceFile, position) || state.options.findInComments && ts.isInNonReferenceComment(sourceFile, position))) {
                            // In the case where we're looking inside comments/strings, we don't have
                            // an actual definition.  So just use 'undefined' here.  Features like
                            // 'Rename' won't care (as they ignore the definitions), and features like
                            // 'FindReferences' will just filter out these results.
                            state.addStringOrCommentReference(sourceFile.fileName, ts.createTextSpan(position, search.text.length));
                        }
                        return;
                    }
                    if (!hasMatchingMeaning(referenceLocation, state))
                        return;
                    var referenceSymbol = state.checker.getSymbolAtLocation(referenceLocation);
                    if (!referenceSymbol) {
                        return;
                    }
                    var parent = referenceLocation.parent;
                    if (ts.isImportSpecifier(parent) && parent.propertyName === referenceLocation) {
                        // This is added through `singleReferences` in ImportsResult. If we happen to see it again, don't add it again.
                        return;
                    }
                    if (ts.isExportSpecifier(parent)) {
                        ts.Debug.assert(referenceLocation.kind === 79 /* Identifier */);
                        getReferencesAtExportSpecifier(referenceLocation, referenceSymbol, parent, search, state, addReferencesHere);
                        return;
                    }
                    var relatedSymbol = getRelatedSymbol(search, referenceSymbol, referenceLocation, state);
                    if (!relatedSymbol) {
                        getReferenceForShorthandProperty(referenceSymbol, search, state);
                        return;
                    }
                    switch (state.specialSearchKind) {
                        case 0 /* None */:
                            if (addReferencesHere)
                                addReference(referenceLocation, relatedSymbol, state);
                            break;
                        case 1 /* Constructor */:
                            addConstructorReferences(referenceLocation, sourceFile, search, state);
                            break;
                        case 2 /* Class */:
                            addClassStaticThisReferences(referenceLocation, search, state);
                            break;
                        default:
                            ts.Debug.assertNever(state.specialSearchKind);
                    }
                    // Use the parent symbol if the location is commonjs require syntax on javascript files only.
                    referenceSymbol = ts.isInJSFile(referenceLocation) && referenceLocation.parent.kind === 201 /* BindingElement */ && ts.isRequireVariableDeclaration(referenceLocation.parent)
                        ? referenceLocation.parent.symbol
                        : referenceSymbol;
                    getImportOrExportReferences(referenceLocation, referenceSymbol, search, state);
                }
                function getReferencesAtExportSpecifier(referenceLocation, referenceSymbol, exportSpecifier, search, state, addReferencesHere, alwaysGetReferences) {
                    ts.Debug.assert(!alwaysGetReferences || !!state.options.providePrefixAndSuffixTextForRename, "If alwaysGetReferences is true, then prefix/suffix text must be enabled");
                    var parent = exportSpecifier.parent, propertyName = exportSpecifier.propertyName, name = exportSpecifier.name;
                    var exportDeclaration = parent.parent;
                    var localSymbol = getLocalSymbolForExportSpecifier(referenceLocation, referenceSymbol, exportSpecifier, state.checker);
                    if (!alwaysGetReferences && !search.includes(localSymbol)) {
                        return;
                    }
                    if (!propertyName) {
                        // Don't rename at `export { default } from "m";`. (but do continue to search for imports of the re-export)
                        if (!(state.options.use === 2 /* Rename */ && (name.escapedText === "default" /* Default */))) {
                            addRef();
                        }
                    }
                    else if (referenceLocation === propertyName) {
                        // For `export { foo as bar } from "baz"`, "`foo`" will be added from the singleReferences for import searches of the original export.
                        // For `export { foo as bar };`, where `foo` is a local, so add it now.
                        if (!exportDeclaration.moduleSpecifier) {
                            addRef();
                        }
                        if (addReferencesHere && state.options.use !== 2 /* Rename */ && state.markSeenReExportRHS(name)) {
                            addReference(name, ts.Debug.checkDefined(exportSpecifier.symbol), state);
                        }
                    }
                    else {
                        if (state.markSeenReExportRHS(referenceLocation)) {
                            addRef();
                        }
                    }
                    // For `export { foo as bar }`, rename `foo`, but not `bar`.
                    if (!isForRenameWithPrefixAndSuffixText(state.options) || alwaysGetReferences) {
                        var isDefaultExport = referenceLocation.originalKeywordKind === 88 /* DefaultKeyword */
                            || exportSpecifier.name.originalKeywordKind === 88 /* DefaultKeyword */;
                        var exportKind = isDefaultExport ? 1 /* Default */ : 0 /* Named */;
                        var exportSymbol = ts.Debug.checkDefined(exportSpecifier.symbol);
                        var exportInfo = FindAllReferences.getExportInfo(exportSymbol, exportKind, state.checker);
                        if (exportInfo) {
                            searchForImportsOfExport(referenceLocation, exportSymbol, exportInfo, state);
                        }
                    }
                    // At `export { x } from "foo"`, also search for the imported symbol `"foo".x`.
                    if (search.comingFrom !== 1 /* Export */ && exportDeclaration.moduleSpecifier && !propertyName && !isForRenameWithPrefixAndSuffixText(state.options)) {
                        var imported = state.checker.getExportSpecifierLocalTargetSymbol(exportSpecifier);
                        if (imported)
                            searchForImportedSymbol(imported, state);
                    }
                    function addRef() {
                        if (addReferencesHere)
                            addReference(referenceLocation, localSymbol, state);
                    }
                }
                function getLocalSymbolForExportSpecifier(referenceLocation, referenceSymbol, exportSpecifier, checker) {
                    return isExportSpecifierAlias(referenceLocation, exportSpecifier) && checker.getExportSpecifierLocalTargetSymbol(exportSpecifier) || referenceSymbol;
                }
                function isExportSpecifierAlias(referenceLocation, exportSpecifier) {
                    var parent = exportSpecifier.parent, propertyName = exportSpecifier.propertyName, name = exportSpecifier.name;
                    ts.Debug.assert(propertyName === referenceLocation || name === referenceLocation);
                    if (propertyName) {
                        // Given `export { foo as bar } [from "someModule"]`: It's an alias at `foo`, but at `bar` it's a new symbol.
                        return propertyName === referenceLocation;
                    }
                    else {
                        // `export { foo } from "foo"` is a re-export.
                        // `export { foo };` is not a re-export, it creates an alias for the local variable `foo`.
                        return !parent.parent.moduleSpecifier;
                    }
                }
                function getImportOrExportReferences(referenceLocation, referenceSymbol, search, state) {
                    var importOrExport = FindAllReferences.getImportOrExportSymbol(referenceLocation, referenceSymbol, state.checker, search.comingFrom === 1 /* Export */);
                    if (!importOrExport)
                        return;
                    var symbol = importOrExport.symbol;
                    if (importOrExport.kind === 0 /* Import */) {
                        if (!(isForRenameWithPrefixAndSuffixText(state.options))) {
                            searchForImportedSymbol(symbol, state);
                        }
                    }
                    else {
                        searchForImportsOfExport(referenceLocation, symbol, importOrExport.exportInfo, state);
                    }
                }
                function getReferenceForShorthandProperty(_a, search, state) {
                    var flags = _a.flags, valueDeclaration = _a.valueDeclaration;
                    var shorthandValueSymbol = state.checker.getShorthandAssignmentValueSymbol(valueDeclaration);
                    var name = valueDeclaration && ts.getNameOfDeclaration(valueDeclaration);
                    /*
                    * Because in short-hand property assignment, an identifier which stored as name of the short-hand property assignment
                    * has two meanings: property name and property value. Therefore when we do findAllReference at the position where
                    * an identifier is declared, the language service should return the position of the variable declaration as well as
                    * the position in short-hand property assignment excluding property accessing. However, if we do findAllReference at the
                    * position of property accessing, the referenceEntry of such position will be handled in the first case.
                    */
                    if (!(flags & 33554432 /* Transient */) && name && search.includes(shorthandValueSymbol)) {
                        addReference(name, shorthandValueSymbol, state);
                    }
                }
                function addReference(referenceLocation, relatedSymbol, state) {
                    var _a = "kind" in relatedSymbol ? relatedSymbol : { kind: undefined, symbol: relatedSymbol }, kind = _a.kind, symbol = _a.symbol; // eslint-disable-line no-in-operator
                    var addRef = state.referenceAdder(symbol);
                    if (state.options.implementations) {
                        addImplementationReferences(referenceLocation, addRef, state);
                    }
                    else {
                        addRef(referenceLocation, kind);
                    }
                }
                /** Adds references when a constructor is used with `new this()` in its own class and `super()` calls in subclasses.  */
                function addConstructorReferences(referenceLocation, sourceFile, search, state) {
                    if (ts.isNewExpressionTarget(referenceLocation)) {
                        addReference(referenceLocation, search.symbol, state);
                    }
                    var pusher = function () { return state.referenceAdder(search.symbol); };
                    if (ts.isClassLike(referenceLocation.parent)) {
                        ts.Debug.assert(referenceLocation.kind === 88 /* DefaultKeyword */ || referenceLocation.parent.name === referenceLocation);
                        // This is the class declaration containing the constructor.
                        findOwnConstructorReferences(search.symbol, sourceFile, pusher());
                    }
                    else {
                        // If this class appears in `extends C`, then the extending class' "super" calls are references.
                        var classExtending = tryGetClassByExtendingIdentifier(referenceLocation);
                        if (classExtending) {
                            findSuperConstructorAccesses(classExtending, pusher());
                            findInheritedConstructorReferences(classExtending, state);
                        }
                    }
                }
                function addClassStaticThisReferences(referenceLocation, search, state) {
                    addReference(referenceLocation, search.symbol, state);
                    var classLike = referenceLocation.parent;
                    if (state.options.use === 2 /* Rename */ || !ts.isClassLike(classLike))
                        return;
                    ts.Debug.assert(classLike.name === referenceLocation);
                    var addRef = state.referenceAdder(search.symbol);
                    for (var _i = 0, _a = classLike.members; _i < _a.length; _i++) {
                        var member = _a[_i];
                        if (!(ts.isMethodOrAccessor(member) && ts.isStatic(member))) {
                            continue;
                        }
                        if (member.body) {
                            member.body.forEachChild(function cb(node) {
                                if (node.kind === 108 /* ThisKeyword */) {
                                    addRef(node);
                                }
                                else if (!ts.isFunctionLike(node) && !ts.isClassLike(node)) {
                                    node.forEachChild(cb);
                                }
                            });
                        }
                    }
                }
                /**
                 * `classSymbol` is the class where the constructor was defined.
                 * Reference the constructor and all calls to `new this()`.
                 */
                function findOwnConstructorReferences(classSymbol, sourceFile, addNode) {
                    var constructorSymbol = getClassConstructorSymbol(classSymbol);
                    if (constructorSymbol && constructorSymbol.declarations) {
                        for (var _i = 0, _a = constructorSymbol.declarations; _i < _a.length; _i++) {
                            var decl = _a[_i];
                            var ctrKeyword = ts.findChildOfKind(decl, 133 /* ConstructorKeyword */, sourceFile);
                            ts.Debug.assert(decl.kind === 169 /* Constructor */ && !!ctrKeyword);
                            addNode(ctrKeyword);
                        }
                    }
                    if (classSymbol.exports) {
                        classSymbol.exports.forEach(function (member) {
                            var decl = member.valueDeclaration;
                            if (decl && decl.kind === 167 /* MethodDeclaration */) {
                                var body = decl.body;
                                if (body) {
                                    forEachDescendantOfKind(body, 108 /* ThisKeyword */, function (thisKeyword) {
                                        if (ts.isNewExpressionTarget(thisKeyword)) {
                                            addNode(thisKeyword);
                                        }
                                    });
                                }
                            }
                        });
                    }
                }
                function getClassConstructorSymbol(classSymbol) {
                    return classSymbol.members && classSymbol.members.get("__constructor" /* Constructor */);
                }
                /** Find references to `super` in the constructor of an extending class.  */
                function findSuperConstructorAccesses(classDeclaration, addNode) {
                    var constructor = getClassConstructorSymbol(classDeclaration.symbol);
                    if (!(constructor && constructor.declarations)) {
                        return;
                    }
                    for (var _i = 0, _a = constructor.declarations; _i < _a.length; _i++) {
                        var decl = _a[_i];
                        ts.Debug.assert(decl.kind === 169 /* Constructor */);
                        var body = decl.body;
                        if (body) {
                            forEachDescendantOfKind(body, 106 /* SuperKeyword */, function (node) {
                                if (ts.isCallExpressionTarget(node)) {
                                    addNode(node);
                                }
                            });
                        }
                    }
                }
                function hasOwnConstructor(classDeclaration) {
                    return !!getClassConstructorSymbol(classDeclaration.symbol);
                }
                function findInheritedConstructorReferences(classDeclaration, state) {
                    if (hasOwnConstructor(classDeclaration))
                        return;
                    var classSymbol = classDeclaration.symbol;
                    var search = state.createSearch(/*location*/ undefined, classSymbol, /*comingFrom*/ undefined);
                    getReferencesInContainerOrFiles(classSymbol, state, search);
                }
                function addImplementationReferences(refNode, addReference, state) {
                    // Check if we found a function/propertyAssignment/method with an implementation or initializer
                    if (ts.isDeclarationName(refNode) && isImplementation(refNode.parent)) {
                        addReference(refNode);
                        return;
                    }
                    if (refNode.kind !== 79 /* Identifier */) {
                        return;
                    }
                    if (refNode.parent.kind === 292 /* ShorthandPropertyAssignment */) {
                        // Go ahead and dereference the shorthand assignment by going to its definition
                        getReferenceEntriesForShorthandPropertyAssignment(refNode, state.checker, addReference);
                    }
                    // Check if the node is within an extends or implements clause
                    var containingClass = getContainingClassIfInHeritageClause(refNode);
                    if (containingClass) {
                        addReference(containingClass);
                        return;
                    }
                    // If we got a type reference, try and see if the reference applies to any expressions that can implement an interface
                    // Find the first node whose parent isn't a type node -- i.e., the highest type node.
                    var typeNode = ts.findAncestor(refNode, function (a) { return !ts.isQualifiedName(a.parent) && !ts.isTypeNode(a.parent) && !ts.isTypeElement(a.parent); });
                    var typeHavingNode = typeNode.parent;
                    if (ts.hasType(typeHavingNode) && typeHavingNode.type === typeNode && state.markSeenContainingTypeReference(typeHavingNode)) {
                        if (ts.hasInitializer(typeHavingNode)) {
                            addIfImplementation(typeHavingNode.initializer);
                        }
                        else if (ts.isFunctionLike(typeHavingNode) && typeHavingNode.body) {
                            var body = typeHavingNode.body;
                            if (body.kind === 233 /* Block */) {
                                ts.forEachReturnStatement(body, function (returnStatement) {
                                    if (returnStatement.expression)
                                        addIfImplementation(returnStatement.expression);
                                });
                            }
                            else {
                                addIfImplementation(body);
                            }
                        }
                        else if (ts.isAssertionExpression(typeHavingNode)) {
                            addIfImplementation(typeHavingNode.expression);
                        }
                    }
                    function addIfImplementation(e) {
                        if (isImplementationExpression(e))
                            addReference(e);
                    }
                }
                function getContainingClassIfInHeritageClause(node) {
                    return ts.isIdentifier(node) || ts.isPropertyAccessExpression(node) ? getContainingClassIfInHeritageClause(node.parent)
                        : ts.isExpressionWithTypeArguments(node) ? ts.tryCast(node.parent.parent, ts.isClassLike) : undefined;
                }
                /**
                 * Returns true if this is an expression that can be considered an implementation
                 */
                function isImplementationExpression(node) {
                    switch (node.kind) {
                        case 210 /* ParenthesizedExpression */:
                            return isImplementationExpression(node.expression);
                        case 212 /* ArrowFunction */:
                        case 211 /* FunctionExpression */:
                        case 203 /* ObjectLiteralExpression */:
                        case 224 /* ClassExpression */:
                        case 202 /* ArrayLiteralExpression */:
                            return true;
                        default:
                            return false;
                    }
                }
                /**
                 * Determines if the parent symbol occurs somewhere in the child's ancestry. If the parent symbol
                 * is an interface, determines if some ancestor of the child symbol extends or inherits from it.
                 * Also takes in a cache of previous results which makes this slightly more efficient and is
                 * necessary to avoid potential loops like so:
                 *     class A extends B { }
                 *     class B extends A { }
                 *
                 * We traverse the AST rather than using the type checker because users are typically only interested
                 * in explicit implementations of an interface/class when calling "Go to Implementation". Sibling
                 * implementations of types that share a common ancestor with the type whose implementation we are
                 * searching for need to be filtered out of the results. The type checker doesn't let us make the
                 * distinction between structurally compatible implementations and explicit implementations, so we
                 * must use the AST.
                 *
                 * @param symbol         A class or interface Symbol
                 * @param parent        Another class or interface Symbol
                 * @param cachedResults A map of symbol id pairs (i.e. "child,parent") to booleans indicating previous results
                 */
                function explicitlyInheritsFrom(symbol, parent, cachedResults, checker) {
                    if (symbol === parent) {
                        return true;
                    }
                    var key = ts.getSymbolId(symbol) + "," + ts.getSymbolId(parent);
                    var cached = cachedResults.get(key);
                    if (cached !== undefined) {
                        return cached;
                    }
                    // Set the key so that we don't infinitely recurse
                    cachedResults.set(key, false);
                    var inherits = !!symbol.declarations && symbol.declarations.some(function (declaration) {
                        return ts.getAllSuperTypeNodes(declaration).some(function (typeReference) {
                            var type = checker.getTypeAtLocation(typeReference);
                            return !!type && !!type.symbol && explicitlyInheritsFrom(type.symbol, parent, cachedResults, checker);
                        });
                    });
                    cachedResults.set(key, inherits);
                    return inherits;
                }
                function getReferencesForSuperKeyword(superKeyword) {
                    var searchSpaceNode = ts.getSuperContainer(superKeyword, /*stopOnFunctions*/ false);
                    if (!searchSpaceNode) {
                        return undefined;
                    }
                    // Whether 'super' occurs in a static context within a class.
                    var staticFlag = 32 /* Static */;
                    switch (searchSpaceNode.kind) {
                        case 165 /* PropertyDeclaration */:
                        case 164 /* PropertySignature */:
                        case 167 /* MethodDeclaration */:
                        case 166 /* MethodSignature */:
                        case 169 /* Constructor */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            staticFlag &= ts.getSyntacticModifierFlags(searchSpaceNode);
                            searchSpaceNode = searchSpaceNode.parent; // re-assign to be the owning class
                            break;
                        default:
                            return undefined;
                    }
                    var sourceFile = searchSpaceNode.getSourceFile();
                    var references = ts.mapDefined(getPossibleSymbolReferenceNodes(sourceFile, "super", searchSpaceNode), function (node) {
                        if (node.kind !== 106 /* SuperKeyword */) {
                            return;
                        }
                        var container = ts.getSuperContainer(node, /*stopOnFunctions*/ false);
                        // If we have a 'super' container, we must have an enclosing class.
                        // Now make sure the owning class is the same as the search-space
                        // and has the same static qualifier as the original 'super's owner.
                        return container && ts.isStatic(container) === !!staticFlag && container.parent.symbol === searchSpaceNode.symbol ? nodeEntry(node) : undefined;
                    });
                    return [{ definition: { type: 0 /* Symbol */, symbol: searchSpaceNode.symbol }, references: references }];
                }
                function isParameterName(node) {
                    return node.kind === 79 /* Identifier */ && node.parent.kind === 162 /* Parameter */ && node.parent.name === node;
                }
                function getReferencesForThisKeyword(thisOrSuperKeyword, sourceFiles, cancellationToken) {
                    var searchSpaceNode = ts.getThisContainer(thisOrSuperKeyword, /* includeArrowFunctions */ false);
                    // Whether 'this' occurs in a static context within a class.
                    var staticFlag = 32 /* Static */;
                    switch (searchSpaceNode.kind) {
                        case 167 /* MethodDeclaration */:
                        case 166 /* MethodSignature */:
                            if (ts.isObjectLiteralMethod(searchSpaceNode)) {
                                staticFlag &= ts.getSyntacticModifierFlags(searchSpaceNode);
                                searchSpaceNode = searchSpaceNode.parent; // re-assign to be the owning object literals
                                break;
                            }
                        // falls through
                        case 165 /* PropertyDeclaration */:
                        case 164 /* PropertySignature */:
                        case 169 /* Constructor */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            staticFlag &= ts.getSyntacticModifierFlags(searchSpaceNode);
                            searchSpaceNode = searchSpaceNode.parent; // re-assign to be the owning class
                            break;
                        case 300 /* SourceFile */:
                            if (ts.isExternalModule(searchSpaceNode) || isParameterName(thisOrSuperKeyword)) {
                                return undefined;
                            }
                        // falls through
                        case 254 /* FunctionDeclaration */:
                        case 211 /* FunctionExpression */:
                            break;
                        // Computed properties in classes are not handled here because references to this are illegal,
                        // so there is no point finding references to them.
                        default:
                            return undefined;
                    }
                    var references = ts.flatMap(searchSpaceNode.kind === 300 /* SourceFile */ ? sourceFiles : [searchSpaceNode.getSourceFile()], function (sourceFile) {
                        cancellationToken.throwIfCancellationRequested();
                        return getPossibleSymbolReferenceNodes(sourceFile, "this", ts.isSourceFile(searchSpaceNode) ? sourceFile : searchSpaceNode).filter(function (node) {
                            if (!ts.isThis(node)) {
                                return false;
                            }
                            var container = ts.getThisContainer(node, /* includeArrowFunctions */ false);
                            switch (searchSpaceNode.kind) {
                                case 211 /* FunctionExpression */:
                                case 254 /* FunctionDeclaration */:
                                    return searchSpaceNode.symbol === container.symbol;
                                case 167 /* MethodDeclaration */:
                                case 166 /* MethodSignature */:
                                    return ts.isObjectLiteralMethod(searchSpaceNode) && searchSpaceNode.symbol === container.symbol;
                                case 224 /* ClassExpression */:
                                case 255 /* ClassDeclaration */:
                                case 203 /* ObjectLiteralExpression */:
                                    // Make sure the container belongs to the same class/object literals
                                    // and has the appropriate static modifier from the original container.
                                    return container.parent && searchSpaceNode.symbol === container.parent.symbol && ts.isStatic(container) === !!staticFlag;
                                case 300 /* SourceFile */:
                                    return container.kind === 300 /* SourceFile */ && !ts.isExternalModule(container) && !isParameterName(node);
                            }
                        });
                    }).map(function (n) { return nodeEntry(n); });
                    var thisParameter = ts.firstDefined(references, function (r) { return ts.isParameter(r.node.parent) ? r.node : undefined; });
                    return [{
                            definition: { type: 3 /* This */, node: thisParameter || thisOrSuperKeyword },
                            references: references
                        }];
                }
                function getReferencesForStringLiteral(node, sourceFiles, checker, cancellationToken) {
                    var type = ts.getContextualTypeOrAncestorTypeNodeType(node, checker);
                    var references = ts.flatMap(sourceFiles, function (sourceFile) {
                        cancellationToken.throwIfCancellationRequested();
                        return ts.mapDefined(getPossibleSymbolReferenceNodes(sourceFile, node.text), function (ref) {
                            if (ts.isStringLiteralLike(ref) && ref.text === node.text) {
                                if (type) {
                                    var refType = ts.getContextualTypeOrAncestorTypeNodeType(ref, checker);
                                    if (type !== checker.getStringType() && type === refType) {
                                        return nodeEntry(ref, 2 /* StringLiteral */);
                                    }
                                }
                                else {
                                    return nodeEntry(ref, 2 /* StringLiteral */);
                                }
                            }
                        });
                    });
                    return [{
                            definition: { type: 4 /* String */, node: node },
                            references: references
                        }];
                }
                // For certain symbol kinds, we need to include other symbols in the search set.
                // This is not needed when searching for re-exports.
                function populateSearchSymbolSet(symbol, location, checker, isForRename, providePrefixAndSuffixText, implementations) {
                    var result = [];
                    forEachRelatedSymbol(symbol, location, checker, isForRename, !(isForRename && providePrefixAndSuffixText), function (sym, root, base) {
                        // static method/property and instance method/property might have the same name. Only include static or only include instance.
                        if (base) {
                            if (isStaticSymbol(symbol) !== isStaticSymbol(base)) {
                                base = undefined;
                            }
                        }
                        result.push(base || root || sym);
                    }, 
                    // when try to find implementation, implementations is true, and not allowed to find base class
                    /*allowBaseTypes*/ function () { return !implementations; });
                    return result;
                }
                /**
                 * @param allowBaseTypes return true means it would try to find in base class or interface.
                 */
                function forEachRelatedSymbol(symbol, location, checker, isForRenamePopulateSearchSymbolSet, onlyIncludeBindingElementAtReferenceLocation, 
                /**
                 * @param baseSymbol This symbol means one property/mehtod from base class or interface when it is not null or undefined,
                 */
                cbSymbol, allowBaseTypes) {
                    var containingObjectLiteralElement = ts.getContainingObjectLiteralElement(location);
                    if (containingObjectLiteralElement) {
                        /* Because in short-hand property assignment, location has two meaning : property name and as value of the property
                        * When we do findAllReference at the position of the short-hand property assignment, we would want to have references to position of
                        * property name and variable declaration of the identifier.
                        * Like in below example, when querying for all references for an identifier 'name', of the property assignment, the language service
                        * should show both 'name' in 'obj' and 'name' in variable declaration
                        *      const name = "Foo";
                        *      const obj = { name };
                        * In order to do that, we will populate the search set with the value symbol of the identifier as a value of the property assignment
                        * so that when matching with potential reference symbol, both symbols from property declaration and variable declaration
                        * will be included correctly.
                        */
                        var shorthandValueSymbol = checker.getShorthandAssignmentValueSymbol(location.parent); // gets the local symbol
                        if (shorthandValueSymbol && isForRenamePopulateSearchSymbolSet) {
                            // When renaming 'x' in `const o = { x }`, just rename the local variable, not the property.
                            return cbSymbol(shorthandValueSymbol, /*rootSymbol*/ undefined, /*baseSymbol*/ undefined, 3 /* SearchedLocalFoundProperty */);
                        }
                        // If the location is in a context sensitive location (i.e. in an object literal) try
                        // to get a contextual type for it, and add the property symbol from the contextual
                        // type to the search set
                        var contextualType = checker.getContextualType(containingObjectLiteralElement.parent);
                        var res_1 = contextualType && ts.firstDefined(ts.getPropertySymbolsFromContextualType(containingObjectLiteralElement, checker, contextualType, /*unionSymbolOk*/ true), function (sym) { return fromRoot(sym, 4 /* SearchedPropertyFoundLocal */); });
                        if (res_1)
                            return res_1;
                        // If the location is name of property symbol from object literal destructuring pattern
                        // Search the property symbol
                        //      for ( { property: p2 } of elems) { }
                        var propertySymbol = getPropertySymbolOfDestructuringAssignment(location, checker);
                        var res1 = propertySymbol && cbSymbol(propertySymbol, /*rootSymbol*/ undefined, /*baseSymbol*/ undefined, 4 /* SearchedPropertyFoundLocal */);
                        if (res1)
                            return res1;
                        var res2 = shorthandValueSymbol && cbSymbol(shorthandValueSymbol, /*rootSymbol*/ undefined, /*baseSymbol*/ undefined, 3 /* SearchedLocalFoundProperty */);
                        if (res2)
                            return res2;
                    }
                    var aliasedSymbol = getMergedAliasedSymbolOfNamespaceExportDeclaration(location, symbol, checker);
                    if (aliasedSymbol) {
                        // In case of UMD module and global merging, search for global as well
                        var res_2 = cbSymbol(aliasedSymbol, /*rootSymbol*/ undefined, /*baseSymbol*/ undefined, 1 /* Node */);
                        if (res_2)
                            return res_2;
                    }
                    var res = fromRoot(symbol);
                    if (res)
                        return res;
                    if (symbol.valueDeclaration && ts.isParameterPropertyDeclaration(symbol.valueDeclaration, symbol.valueDeclaration.parent)) {
                        // For a parameter property, now try on the other symbol (property if this was a parameter, parameter if this was a property).
                        var paramProps = checker.getSymbolsOfParameterPropertyDeclaration(ts.cast(symbol.valueDeclaration, ts.isParameter), symbol.name);
                        ts.Debug.assert(paramProps.length === 2 && !!(paramProps[0].flags & 1 /* FunctionScopedVariable */) && !!(paramProps[1].flags & 4 /* Property */)); // is [parameter, property]
                        return fromRoot(symbol.flags & 1 /* FunctionScopedVariable */ ? paramProps[1] : paramProps[0]);
                    }
                    var exportSpecifier = ts.getDeclarationOfKind(symbol, 273 /* ExportSpecifier */);
                    if (!isForRenamePopulateSearchSymbolSet || exportSpecifier && !exportSpecifier.propertyName) {
                        var localSymbol = exportSpecifier && checker.getExportSpecifierLocalTargetSymbol(exportSpecifier);
                        if (localSymbol) {
                            var res_3 = cbSymbol(localSymbol, /*rootSymbol*/ undefined, /*baseSymbol*/ undefined, 1 /* Node */);
                            if (res_3)
                                return res_3;
                        }
                    }
                    // symbolAtLocation for a binding element is the local symbol. See if the search symbol is the property.
                    // Don't do this when populating search set for a rename when prefix and suffix text will be provided -- just rename the local.
                    if (!isForRenamePopulateSearchSymbolSet) {
                        var bindingElementPropertySymbol = void 0;
                        if (onlyIncludeBindingElementAtReferenceLocation) {
                            bindingElementPropertySymbol = ts.isObjectBindingElementWithoutPropertyName(location.parent) ? ts.getPropertySymbolFromBindingElement(checker, location.parent) : undefined;
                        }
                        else {
                            bindingElementPropertySymbol = getPropertySymbolOfObjectBindingPatternWithoutPropertyName(symbol, checker);
                        }
                        return bindingElementPropertySymbol && fromRoot(bindingElementPropertySymbol, 4 /* SearchedPropertyFoundLocal */);
                    }
                    ts.Debug.assert(isForRenamePopulateSearchSymbolSet);
                    // due to the above assert and the arguments at the uses of this function,
                    // (onlyIncludeBindingElementAtReferenceLocation <=> !providePrefixAndSuffixTextForRename) holds
                    var includeOriginalSymbolOfBindingElement = onlyIncludeBindingElementAtReferenceLocation;
                    if (includeOriginalSymbolOfBindingElement) {
                        var bindingElementPropertySymbol = getPropertySymbolOfObjectBindingPatternWithoutPropertyName(symbol, checker);
                        return bindingElementPropertySymbol && fromRoot(bindingElementPropertySymbol, 4 /* SearchedPropertyFoundLocal */);
                    }
                    function fromRoot(sym, kind) {
                        // If this is a union property:
                        //   - In populateSearchSymbolsSet we will add all the symbols from all its source symbols in all unioned types.
                        //   - In findRelatedSymbol, we will just use the union symbol if any source symbol is included in the search.
                        // If the symbol is an instantiation from a another symbol (e.g. widened symbol):
                        //   - In populateSearchSymbolsSet, add the root the list
                        //   - In findRelatedSymbol, return the source symbol if that is in the search. (Do not return the instantiation symbol.)
                        return ts.firstDefined(checker.getRootSymbols(sym), function (rootSymbol) {
                            return cbSymbol(sym, rootSymbol, /*baseSymbol*/ undefined, kind)
                                // Add symbol of properties/methods of the same name in base classes and implemented interfaces definitions
                                || (rootSymbol.parent && rootSymbol.parent.flags & (32 /* Class */ | 64 /* Interface */) && allowBaseTypes(rootSymbol)
                                    ? getPropertySymbolsFromBaseTypes(rootSymbol.parent, rootSymbol.name, checker, function (base) { return cbSymbol(sym, rootSymbol, base, kind); })
                                    : undefined);
                        });
                    }
                    function getPropertySymbolOfObjectBindingPatternWithoutPropertyName(symbol, checker) {
                        var bindingElement = ts.getDeclarationOfKind(symbol, 201 /* BindingElement */);
                        if (bindingElement && ts.isObjectBindingElementWithoutPropertyName(bindingElement)) {
                            return ts.getPropertySymbolFromBindingElement(checker, bindingElement);
                        }
                    }
                }
                /**
                 * Find symbol of the given property-name and add the symbol to the given result array
                 * @param symbol a symbol to start searching for the given propertyName
                 * @param propertyName a name of property to search for
                 * @param result an array of symbol of found property symbols
                 * @param previousIterationSymbolsCache a cache of symbol from previous iterations of calling this function to prevent infinite revisiting of the same symbol.
                 *                                The value of previousIterationSymbol is undefined when the function is first called.
                 */
                function getPropertySymbolsFromBaseTypes(symbol, propertyName, checker, cb) {
                    var seen = new ts.Map();
                    return recur(symbol);
                    function recur(symbol) {
                        // Use `addToSeen` to ensure we don't infinitely recurse in this situation:
                        //      interface C extends C {
                        //          /*findRef*/propName: string;
                        //      }
                        if (!(symbol.flags & (32 /* Class */ | 64 /* Interface */)) || !ts.addToSeen(seen, ts.getSymbolId(symbol)))
                            return;
                        return ts.firstDefined(symbol.declarations, function (declaration) {
                            return ts.firstDefined(ts.getAllSuperTypeNodes(declaration), function (typeReference) {
                                var type = checker.getTypeAtLocation(typeReference);
                                var propertySymbol = type && type.symbol && checker.getPropertyOfType(type, propertyName);
                                // Visit the typeReference as well to see if it directly or indirectly uses that property
                                return type && propertySymbol && (ts.firstDefined(checker.getRootSymbols(propertySymbol), cb) || recur(type.symbol));
                            });
                        });
                    }
                }
                function isStaticSymbol(symbol) {
                    if (!symbol.valueDeclaration) {
                        return false;
                    }
                    var modifierFlags = ts.getEffectiveModifierFlags(symbol.valueDeclaration);
                    return !!(modifierFlags & 32 /* Static */);
                }
                function getRelatedSymbol(search, referenceSymbol, referenceLocation, state) {
                    var checker = state.checker;
                    return forEachRelatedSymbol(referenceSymbol, referenceLocation, checker, /*isForRenamePopulateSearchSymbolSet*/ false, 
                    /*onlyIncludeBindingElementAtReferenceLocation*/ state.options.use !== 2 /* Rename */ || !!state.options.providePrefixAndSuffixTextForRename, function (sym, rootSymbol, baseSymbol, kind) {
                        // check whether the symbol used to search itself is just the searched one.
                        if (baseSymbol) {
                            // static method/property and instance method/property might have the same name. Only check static or only check instance.
                            if (isStaticSymbol(referenceSymbol) !== isStaticSymbol(baseSymbol)) {
                                baseSymbol = undefined;
                            }
                        }
                        return search.includes(baseSymbol || rootSymbol || sym)
                            // For a base type, use the symbol for the derived type. For a synthetic (e.g. union) property, use the union symbol.
                            ? { symbol: rootSymbol && !(ts.getCheckFlags(sym) & 6 /* Synthetic */) ? rootSymbol : sym, kind: kind }
                            : undefined;
                    }, 
                    /*allowBaseTypes*/ function (rootSymbol) {
                        return !(search.parents && !search.parents.some(function (parent) { return explicitlyInheritsFrom(rootSymbol.parent, parent, state.inheritsFromCache, checker); }));
                    });
                }
                /**
                 * Given an initial searchMeaning, extracted from a location, widen the search scope based on the declarations
                 * of the corresponding symbol. e.g. if we are searching for "Foo" in value position, but "Foo" references a class
                 * then we need to widen the search to include type positions as well.
                 * On the contrary, if we are searching for "Bar" in type position and we trace bar to an interface, and an uninstantiated
                 * module, we want to keep the search limited to only types, as the two declarations (interface and uninstantiated module)
                 * do not intersect in any of the three spaces.
                 */
                function getIntersectingMeaningFromDeclarations(node, symbol) {
                    var meaning = ts.getMeaningFromLocation(node);
                    var declarations = symbol.declarations;
                    if (declarations) {
                        var lastIterationMeaning = void 0;
                        do {
                            // The result is order-sensitive, for instance if initialMeaning === Namespace, and declarations = [class, instantiated module]
                            // we need to consider both as they initialMeaning intersects with the module in the namespace space, and the module
                            // intersects with the class in the value space.
                            // To achieve that we will keep iterating until the result stabilizes.
                            // Remember the last meaning
                            lastIterationMeaning = meaning;
                            for (var _i = 0, declarations_2 = declarations; _i < declarations_2.length; _i++) {
                                var declaration = declarations_2[_i];
                                var declarationMeaning = ts.getMeaningFromDeclaration(declaration);
                                if (declarationMeaning & meaning) {
                                    meaning |= declarationMeaning;
                                }
                            }
                        } while (meaning !== lastIterationMeaning);
                    }
                    return meaning;
                }
                Core.getIntersectingMeaningFromDeclarations = getIntersectingMeaningFromDeclarations;
                function isImplementation(node) {
                    return !!(node.flags & 8388608 /* Ambient */) ? !(ts.isInterfaceDeclaration(node) || ts.isTypeAliasDeclaration(node)) :
                        (ts.isVariableLike(node) ? ts.hasInitializer(node) :
                            ts.isFunctionLikeDeclaration(node) ? !!node.body :
                                ts.isClassLike(node) || ts.isModuleOrEnumDeclaration(node));
                }
                function getReferenceEntriesForShorthandPropertyAssignment(node, checker, addReference) {
                    var refSymbol = checker.getSymbolAtLocation(node);
                    var shorthandSymbol = checker.getShorthandAssignmentValueSymbol(refSymbol.valueDeclaration);
                    if (shorthandSymbol) {
                        for (var _i = 0, _a = shorthandSymbol.getDeclarations(); _i < _a.length; _i++) {
                            var declaration = _a[_i];
                            if (ts.getMeaningFromDeclaration(declaration) & 1 /* Value */) {
                                addReference(declaration);
                            }
                        }
                    }
                }
                Core.getReferenceEntriesForShorthandPropertyAssignment = getReferenceEntriesForShorthandPropertyAssignment;
                function forEachDescendantOfKind(node, kind, action) {
                    ts.forEachChild(node, function (child) {
                        if (child.kind === kind) {
                            action(child);
                        }
                        forEachDescendantOfKind(child, kind, action);
                    });
                }
                /** Get `C` given `N` if `N` is in the position `class C extends N` or `class C extends foo.N` where `N` is an identifier. */
                function tryGetClassByExtendingIdentifier(node) {
                    return ts.tryGetClassExtendingExpressionWithTypeArguments(ts.climbPastPropertyAccess(node).parent);
                }
                /**
                 * If we are just looking for implementations and this is a property access expression, we need to get the
                 * symbol of the local type of the symbol the property is being accessed on. This is because our search
                 * symbol may have a different parent symbol if the local type's symbol does not declare the property
                 * being accessed (i.e. it is declared in some parent class or interface)
                 */
                function getParentSymbolsOfPropertyAccess(location, symbol, checker) {
                    var propertyAccessExpression = ts.isRightSideOfPropertyAccess(location) ? location.parent : undefined;
                    var lhsType = propertyAccessExpression && checker.getTypeAtLocation(propertyAccessExpression.expression);
                    var res = ts.mapDefined(lhsType && (lhsType.isUnionOrIntersection() ? lhsType.types : lhsType.symbol === symbol.parent ? undefined : [lhsType]), function (t) {
                        return t.symbol && t.symbol.flags & (32 /* Class */ | 64 /* Interface */) ? t.symbol : undefined;
                    });
                    return res.length === 0 ? undefined : res;
                }
                function isForRenameWithPrefixAndSuffixText(options) {
                    return options.use === 2 /* Rename */ && options.providePrefixAndSuffixTextForRename;
                }
            })(Core = FindAllReferences.Core || (FindAllReferences.Core = {}));
        })(FindAllReferences = ts.FindAllReferences || (ts.FindAllReferences = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var CallHierarchy;
        (function (CallHierarchy) {
            /** Indictates whether a node is named function or class expression. */
            function isNamedExpression(node) {
                return (ts.isFunctionExpression(node) || ts.isClassExpression(node)) && ts.isNamedDeclaration(node);
            }
            /** Indicates whether a node is a function, arrow, or class expression assigned to a constant variable. */
            function isConstNamedExpression(node) {
                return (ts.isFunctionExpression(node) || ts.isArrowFunction(node) || ts.isClassExpression(node))
                    && ts.isVariableDeclaration(node.parent)
                    && node === node.parent.initializer
                    && ts.isIdentifier(node.parent.name)
                    && !!(ts.getCombinedNodeFlags(node.parent) & 2 /* Const */);
            }
            /**
             * Indicates whether a node could possibly be a call hierarchy declaration.
             *
             * See `resolveCallHierarchyDeclaration` for the specific rules.
             */
            function isPossibleCallHierarchyDeclaration(node) {
                return ts.isSourceFile(node)
                    || ts.isModuleDeclaration(node)
                    || ts.isFunctionDeclaration(node)
                    || ts.isFunctionExpression(node)
                    || ts.isClassDeclaration(node)
                    || ts.isClassExpression(node)
                    || ts.isClassStaticBlockDeclaration(node)
                    || ts.isMethodDeclaration(node)
                    || ts.isMethodSignature(node)
                    || ts.isGetAccessorDeclaration(node)
                    || ts.isSetAccessorDeclaration(node);
            }
            /**
             * Indicates whether a node is a valid a call hierarchy declaration.
             *
             * See `resolveCallHierarchyDeclaration` for the specific rules.
             */
            function isValidCallHierarchyDeclaration(node) {
                return ts.isSourceFile(node)
                    || ts.isModuleDeclaration(node) && ts.isIdentifier(node.name)
                    || ts.isFunctionDeclaration(node)
                    || ts.isClassDeclaration(node)
                    || ts.isClassStaticBlockDeclaration(node)
                    || ts.isMethodDeclaration(node)
                    || ts.isMethodSignature(node)
                    || ts.isGetAccessorDeclaration(node)
                    || ts.isSetAccessorDeclaration(node)
                    || isNamedExpression(node)
                    || isConstNamedExpression(node);
            }
            /** Gets the node that can be used as a reference to a call hierarchy declaration. */
            function getCallHierarchyDeclarationReferenceNode(node) {
                if (ts.isSourceFile(node))
                    return node;
                if (ts.isNamedDeclaration(node))
                    return node.name;
                if (isConstNamedExpression(node))
                    return node.parent.name;
                return ts.Debug.checkDefined(node.modifiers && ts.find(node.modifiers, isDefaultModifier));
            }
            function isDefaultModifier(node) {
                return node.kind === 88 /* DefaultKeyword */;
            }
            /** Gets the symbol for a call hierarchy declaration. */
            function getSymbolOfCallHierarchyDeclaration(typeChecker, node) {
                var location = getCallHierarchyDeclarationReferenceNode(node);
                return location && typeChecker.getSymbolAtLocation(location);
            }
            /** Gets the text and range for the name of a call hierarchy declaration. */
            function getCallHierarchyItemName(program, node) {
                if (ts.isSourceFile(node)) {
                    return { text: node.fileName, pos: 0, end: 0 };
                }
                if ((ts.isFunctionDeclaration(node) || ts.isClassDeclaration(node)) && !ts.isNamedDeclaration(node)) {
                    var defaultModifier = node.modifiers && ts.find(node.modifiers, isDefaultModifier);
                    if (defaultModifier) {
                        return { text: "default", pos: defaultModifier.getStart(), end: defaultModifier.getEnd() };
                    }
                }
                if (ts.isClassStaticBlockDeclaration(node)) {
                    var sourceFile = node.getSourceFile();
                    var pos = ts.skipTrivia(sourceFile.text, ts.moveRangePastModifiers(node).pos);
                    var end = pos + 6; /* "static".length */
                    var typeChecker = program.getTypeChecker();
                    var symbol = typeChecker.getSymbolAtLocation(node.parent);
                    var prefix = symbol ? typeChecker.symbolToString(symbol, node.parent) + " " : "";
                    return { text: prefix + "static {}", pos: pos, end: end };
                }
                var declName = isConstNamedExpression(node) ? node.parent.name :
                    ts.Debug.checkDefined(ts.getNameOfDeclaration(node), "Expected call hierarchy item to have a name");
                var text = ts.isIdentifier(declName) ? ts.idText(declName) :
                    ts.isStringOrNumericLiteralLike(declName) ? declName.text :
                        ts.isComputedPropertyName(declName) ?
                            ts.isStringOrNumericLiteralLike(declName.expression) ? declName.expression.text :
                                undefined :
                            undefined;
                if (text === undefined) {
                    var typeChecker = program.getTypeChecker();
                    var symbol = typeChecker.getSymbolAtLocation(declName);
                    if (symbol) {
                        text = typeChecker.symbolToString(symbol, node);
                    }
                }
                if (text === undefined) {
                    // get the text from printing the node on a single line without comments...
                    var printer_1 = ts.createPrinter({ removeComments: true, omitTrailingSemicolon: true });
                    text = ts.usingSingleLineStringWriter(function (writer) { return printer_1.writeNode(4 /* Unspecified */, node, node.getSourceFile(), writer); });
                }
                return { text: text, pos: declName.getStart(), end: declName.getEnd() };
            }
            function getCallHierarchItemContainerName(node) {
                var _a, _b;
                if (isConstNamedExpression(node)) {
                    if (ts.isModuleBlock(node.parent.parent.parent.parent) && ts.isIdentifier(node.parent.parent.parent.parent.parent.name)) {
                        return node.parent.parent.parent.parent.parent.name.getText();
                    }
                    return;
                }
                switch (node.kind) {
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 167 /* MethodDeclaration */:
                        if (node.parent.kind === 203 /* ObjectLiteralExpression */) {
                            return (_a = ts.getAssignedName(node.parent)) === null || _a === void 0 ? void 0 : _a.getText();
                        }
                        return (_b = ts.getNameOfDeclaration(node.parent)) === null || _b === void 0 ? void 0 : _b.getText();
                    case 254 /* FunctionDeclaration */:
                    case 255 /* ClassDeclaration */:
                    case 259 /* ModuleDeclaration */:
                        if (ts.isModuleBlock(node.parent) && ts.isIdentifier(node.parent.parent.name)) {
                            return node.parent.parent.name.getText();
                        }
                }
            }
            function findImplementation(typeChecker, node) {
                if (node.body) {
                    return node;
                }
                if (ts.isConstructorDeclaration(node)) {
                    return ts.getFirstConstructorWithBody(node.parent);
                }
                if (ts.isFunctionDeclaration(node) || ts.isMethodDeclaration(node)) {
                    var symbol = getSymbolOfCallHierarchyDeclaration(typeChecker, node);
                    if (symbol && symbol.valueDeclaration && ts.isFunctionLikeDeclaration(symbol.valueDeclaration) && symbol.valueDeclaration.body) {
                        return symbol.valueDeclaration;
                    }
                    return undefined;
                }
                return node;
            }
            function findAllInitialDeclarations(typeChecker, node) {
                var symbol = getSymbolOfCallHierarchyDeclaration(typeChecker, node);
                var declarations;
                if (symbol && symbol.declarations) {
                    var indices = ts.indicesOf(symbol.declarations);
                    var keys_1 = ts.map(symbol.declarations, function (decl) { return ({ file: decl.getSourceFile().fileName, pos: decl.pos }); });
                    indices.sort(function (a, b) { return ts.compareStringsCaseSensitive(keys_1[a].file, keys_1[b].file) || keys_1[a].pos - keys_1[b].pos; });
                    var sortedDeclarations = ts.map(indices, function (i) { return symbol.declarations[i]; });
                    var lastDecl = void 0;
                    for (var _i = 0, sortedDeclarations_1 = sortedDeclarations; _i < sortedDeclarations_1.length; _i++) {
                        var decl = sortedDeclarations_1[_i];
                        if (isValidCallHierarchyDeclaration(decl)) {
                            if (!lastDecl || lastDecl.parent !== decl.parent || lastDecl.end !== decl.pos) {
                                declarations = ts.append(declarations, decl);
                            }
                            lastDecl = decl;
                        }
                    }
                }
                return declarations;
            }
            /** Find the implementation or the first declaration for a call hierarchy declaration. */
            function findImplementationOrAllInitialDeclarations(typeChecker, node) {
                var _a, _b, _c;
                if (ts.isClassStaticBlockDeclaration(node)) {
                    return node;
                }
                if (ts.isFunctionLikeDeclaration(node)) {
                    return (_b = (_a = findImplementation(typeChecker, node)) !== null && _a !== void 0 ? _a : findAllInitialDeclarations(typeChecker, node)) !== null && _b !== void 0 ? _b : node;
                }
                return (_c = findAllInitialDeclarations(typeChecker, node)) !== null && _c !== void 0 ? _c : node;
            }
            /** Resolves the call hierarchy declaration for a node. */
            function resolveCallHierarchyDeclaration(program, location) {
                // A call hierarchy item must refer to either a SourceFile, Module Declaration, Class Static Block, or something intrinsically callable that has a name:
                // - Class Declarations
                // - Class Expressions (with a name)
                // - Function Declarations
                // - Function Expressions (with a name or assigned to a const variable)
                // - Arrow Functions (assigned to a const variable)
                // - Constructors
                // - Class `static {}` initializer blocks
                // - Methods
                // - Accessors
                //
                // If a call is contained in a non-named callable Node (function expression, arrow function, etc.), then
                // its containing `CallHierarchyItem` is a containing function or SourceFile that matches the above list.
                var typeChecker = program.getTypeChecker();
                var followingSymbol = false;
                while (true) {
                    if (isValidCallHierarchyDeclaration(location)) {
                        return findImplementationOrAllInitialDeclarations(typeChecker, location);
                    }
                    if (isPossibleCallHierarchyDeclaration(location)) {
                        var ancestor = ts.findAncestor(location, isValidCallHierarchyDeclaration);
                        return ancestor && findImplementationOrAllInitialDeclarations(typeChecker, ancestor);
                    }
                    if (ts.isDeclarationName(location)) {
                        if (isValidCallHierarchyDeclaration(location.parent)) {
                            return findImplementationOrAllInitialDeclarations(typeChecker, location.parent);
                        }
                        if (isPossibleCallHierarchyDeclaration(location.parent)) {
                            var ancestor = ts.findAncestor(location.parent, isValidCallHierarchyDeclaration);
                            return ancestor && findImplementationOrAllInitialDeclarations(typeChecker, ancestor);
                        }
                        if (ts.isVariableDeclaration(location.parent) && location.parent.initializer && isConstNamedExpression(location.parent.initializer)) {
                            return location.parent.initializer;
                        }
                        return undefined;
                    }
                    if (ts.isConstructorDeclaration(location)) {
                        if (isValidCallHierarchyDeclaration(location.parent)) {
                            return location.parent;
                        }
                        return undefined;
                    }
                    if (location.kind === 124 /* StaticKeyword */ && ts.isClassStaticBlockDeclaration(location.parent)) {
                        location = location.parent;
                        continue;
                    }
                    // #39453
                    if (ts.isVariableDeclaration(location) && location.initializer && isConstNamedExpression(location.initializer)) {
                        return location.initializer;
                    }
                    if (!followingSymbol) {
                        var symbol = typeChecker.getSymbolAtLocation(location);
                        if (symbol) {
                            if (symbol.flags & 2097152 /* Alias */) {
                                symbol = typeChecker.getAliasedSymbol(symbol);
                            }
                            if (symbol.valueDeclaration) {
                                followingSymbol = true;
                                location = symbol.valueDeclaration;
                                continue;
                            }
                        }
                    }
                    return undefined;
                }
            }
            CallHierarchy.resolveCallHierarchyDeclaration = resolveCallHierarchyDeclaration;
            /** Creates a `CallHierarchyItem` for a call hierarchy declaration. */
            function createCallHierarchyItem(program, node) {
                var sourceFile = node.getSourceFile();
                var name = getCallHierarchyItemName(program, node);
                var containerName = getCallHierarchItemContainerName(node);
                var kind = ts.getNodeKind(node);
                var kindModifiers = ts.getNodeModifiers(node);
                var span = ts.createTextSpanFromBounds(ts.skipTrivia(sourceFile.text, node.getFullStart(), /*stopAfterLineBreak*/ false, /*stopAtComments*/ true), node.getEnd());
                var selectionSpan = ts.createTextSpanFromBounds(name.pos, name.end);
                return { file: sourceFile.fileName, kind: kind, kindModifiers: kindModifiers, name: name.text, containerName: containerName, span: span, selectionSpan: selectionSpan };
            }
            CallHierarchy.createCallHierarchyItem = createCallHierarchyItem;
            function isDefined(x) {
                return x !== undefined;
            }
            function convertEntryToCallSite(entry) {
                if (entry.kind === 1 /* Node */) {
                    var node = entry.node;
                    if (ts.isCallOrNewExpressionTarget(node, /*includeElementAccess*/ true, /*skipPastOuterExpressions*/ true)
                        || ts.isTaggedTemplateTag(node, /*includeElementAccess*/ true, /*skipPastOuterExpressions*/ true)
                        || ts.isDecoratorTarget(node, /*includeElementAccess*/ true, /*skipPastOuterExpressions*/ true)
                        || ts.isJsxOpeningLikeElementTagName(node, /*includeElementAccess*/ true, /*skipPastOuterExpressions*/ true)
                        || ts.isRightSideOfPropertyAccess(node)
                        || ts.isArgumentExpressionOfElementAccess(node)) {
                        var sourceFile = node.getSourceFile();
                        var ancestor = ts.findAncestor(node, isValidCallHierarchyDeclaration) || sourceFile;
                        return { declaration: ancestor, range: ts.createTextRangeFromNode(node, sourceFile) };
                    }
                }
            }
            function getCallSiteGroupKey(entry) {
                return ts.getNodeId(entry.declaration);
            }
            function createCallHierarchyIncomingCall(from, fromSpans) {
                return { from: from, fromSpans: fromSpans };
            }
            function convertCallSiteGroupToIncomingCall(program, entries) {
                return createCallHierarchyIncomingCall(createCallHierarchyItem(program, entries[0].declaration), ts.map(entries, function (entry) { return ts.createTextSpanFromRange(entry.range); }));
            }
            /** Gets the call sites that call into the provided call hierarchy declaration. */
            function getIncomingCalls(program, declaration, cancellationToken) {
                // Source files and modules have no incoming calls.
                if (ts.isSourceFile(declaration) || ts.isModuleDeclaration(declaration) || ts.isClassStaticBlockDeclaration(declaration)) {
                    return [];
                }
                var location = getCallHierarchyDeclarationReferenceNode(declaration);
                var calls = ts.filter(ts.FindAllReferences.findReferenceOrRenameEntries(program, cancellationToken, program.getSourceFiles(), location, /*position*/ 0, { use: 1 /* References */ }, convertEntryToCallSite), isDefined);
                return calls ? ts.group(calls, getCallSiteGroupKey, function (entries) { return convertCallSiteGroupToIncomingCall(program, entries); }) : [];
            }
            CallHierarchy.getIncomingCalls = getIncomingCalls;
            function createCallSiteCollector(program, callSites) {
                function recordCallSite(node) {
                    var target = ts.isTaggedTemplateExpression(node) ? node.tag :
                        ts.isJsxOpeningLikeElement(node) ? node.tagName :
                            ts.isAccessExpression(node) ? node :
                                ts.isClassStaticBlockDeclaration(node) ? node :
                                    node.expression;
                    var declaration = resolveCallHierarchyDeclaration(program, target);
                    if (declaration) {
                        var range = ts.createTextRangeFromNode(target, node.getSourceFile());
                        if (ts.isArray(declaration)) {
                            for (var _i = 0, declaration_1 = declaration; _i < declaration_1.length; _i++) {
                                var decl = declaration_1[_i];
                                callSites.push({ declaration: decl, range: range });
                            }
                        }
                        else {
                            callSites.push({ declaration: declaration, range: range });
                        }
                    }
                }
                function collect(node) {
                    if (!node)
                        return;
                    if (node.flags & 8388608 /* Ambient */) {
                        // do not descend into ambient nodes.
                        return;
                    }
                    if (isValidCallHierarchyDeclaration(node)) {
                        // do not descend into other call site declarations, other than class member names
                        if (ts.isClassLike(node)) {
                            for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                                var member = _a[_i];
                                if (member.name && ts.isComputedPropertyName(member.name)) {
                                    collect(member.name.expression);
                                }
                            }
                        }
                        return;
                    }
                    switch (node.kind) {
                        case 79 /* Identifier */:
                        case 263 /* ImportEqualsDeclaration */:
                        case 264 /* ImportDeclaration */:
                        case 270 /* ExportDeclaration */:
                        case 256 /* InterfaceDeclaration */:
                        case 257 /* TypeAliasDeclaration */:
                            // do not descend into nodes that cannot contain callable nodes
                            return;
                        case 168 /* ClassStaticBlockDeclaration */:
                            recordCallSite(node);
                            return;
                        case 209 /* TypeAssertionExpression */:
                        case 227 /* AsExpression */:
                            // do not descend into the type side of an assertion
                            collect(node.expression);
                            return;
                        case 252 /* VariableDeclaration */:
                        case 162 /* Parameter */:
                            // do not descend into the type of a variable or parameter declaration
                            collect(node.name);
                            collect(node.initializer);
                            return;
                        case 206 /* CallExpression */:
                            // do not descend into the type arguments of a call expression
                            recordCallSite(node);
                            collect(node.expression);
                            ts.forEach(node.arguments, collect);
                            return;
                        case 207 /* NewExpression */:
                            // do not descend into the type arguments of a new expression
                            recordCallSite(node);
                            collect(node.expression);
                            ts.forEach(node.arguments, collect);
                            return;
                        case 208 /* TaggedTemplateExpression */:
                            // do not descend into the type arguments of a tagged template expression
                            recordCallSite(node);
                            collect(node.tag);
                            collect(node.template);
                            return;
                        case 278 /* JsxOpeningElement */:
                        case 277 /* JsxSelfClosingElement */:
                            // do not descend into the type arguments of a JsxOpeningLikeElement
                            recordCallSite(node);
                            collect(node.tagName);
                            collect(node.attributes);
                            return;
                        case 163 /* Decorator */:
                            recordCallSite(node);
                            collect(node.expression);
                            return;
                        case 204 /* PropertyAccessExpression */:
                        case 205 /* ElementAccessExpression */:
                            recordCallSite(node);
                            ts.forEachChild(node, collect);
                            break;
                    }
                    if (ts.isPartOfTypeNode(node)) {
                        // do not descend into types
                        return;
                    }
                    ts.forEachChild(node, collect);
                }
                return collect;
            }
            function collectCallSitesOfSourceFile(node, collect) {
                ts.forEach(node.statements, collect);
            }
            function collectCallSitesOfModuleDeclaration(node, collect) {
                if (!ts.hasSyntacticModifier(node, 2 /* Ambient */) && node.body && ts.isModuleBlock(node.body)) {
                    ts.forEach(node.body.statements, collect);
                }
            }
            function collectCallSitesOfFunctionLikeDeclaration(typeChecker, node, collect) {
                var implementation = findImplementation(typeChecker, node);
                if (implementation) {
                    ts.forEach(implementation.parameters, collect);
                    collect(implementation.body);
                }
            }
            function collectCallSitesOfClassStaticBlockDeclaration(node, collect) {
                collect(node.body);
            }
            function collectCallSitesOfClassLikeDeclaration(node, collect) {
                ts.forEach(node.decorators, collect);
                var heritage = ts.getClassExtendsHeritageElement(node);
                if (heritage) {
                    collect(heritage.expression);
                }
                for (var _i = 0, _a = node.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    ts.forEach(member.decorators, collect);
                    if (ts.isPropertyDeclaration(member)) {
                        collect(member.initializer);
                    }
                    else if (ts.isConstructorDeclaration(member) && member.body) {
                        ts.forEach(member.parameters, collect);
                        collect(member.body);
                    }
                    else if (ts.isClassStaticBlockDeclaration(member)) {
                        collect(member);
                    }
                }
            }
            function collectCallSites(program, node) {
                var callSites = [];
                var collect = createCallSiteCollector(program, callSites);
                switch (node.kind) {
                    case 300 /* SourceFile */:
                        collectCallSitesOfSourceFile(node, collect);
                        break;
                    case 259 /* ModuleDeclaration */:
                        collectCallSitesOfModuleDeclaration(node, collect);
                        break;
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        collectCallSitesOfFunctionLikeDeclaration(program.getTypeChecker(), node, collect);
                        break;
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                        collectCallSitesOfClassLikeDeclaration(node, collect);
                        break;
                    case 168 /* ClassStaticBlockDeclaration */:
                        collectCallSitesOfClassStaticBlockDeclaration(node, collect);
                        break;
                    default:
                        ts.Debug.assertNever(node);
                }
                return callSites;
            }
            function createCallHierarchyOutgoingCall(to, fromSpans) {
                return { to: to, fromSpans: fromSpans };
            }
            function convertCallSiteGroupToOutgoingCall(program, entries) {
                return createCallHierarchyOutgoingCall(createCallHierarchyItem(program, entries[0].declaration), ts.map(entries, function (entry) { return ts.createTextSpanFromRange(entry.range); }));
            }
            /** Gets the call sites that call out of the provided call hierarchy declaration. */
            function getOutgoingCalls(program, declaration) {
                if (declaration.flags & 8388608 /* Ambient */ || ts.isMethodSignature(declaration)) {
                    return [];
                }
                return ts.group(collectCallSites(program, declaration), getCallSiteGroupKey, function (entries) { return convertCallSiteGroupToOutgoingCall(program, entries); });
            }
            CallHierarchy.getOutgoingCalls = getOutgoingCalls;
        })(CallHierarchy = ts.CallHierarchy || (ts.CallHierarchy = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        function getEditsForFileRename(program, oldFileOrDirPath, newFileOrDirPath, host, formatContext, preferences, sourceMapper) {
            var useCaseSensitiveFileNames = ts.hostUsesCaseSensitiveFileNames(host);
            var getCanonicalFileName = ts.createGetCanonicalFileName(useCaseSensitiveFileNames);
            var oldToNew = getPathUpdater(oldFileOrDirPath, newFileOrDirPath, getCanonicalFileName, sourceMapper);
            var newToOld = getPathUpdater(newFileOrDirPath, oldFileOrDirPath, getCanonicalFileName, sourceMapper);
            return ts.textChanges.ChangeTracker.with({ host: host, formatContext: formatContext, preferences: preferences }, function (changeTracker) {
                updateTsconfigFiles(program, changeTracker, oldToNew, oldFileOrDirPath, newFileOrDirPath, host.getCurrentDirectory(), useCaseSensitiveFileNames);
                updateImports(program, changeTracker, oldToNew, newToOld, host, getCanonicalFileName);
            });
        }
        ts.getEditsForFileRename = getEditsForFileRename;
        // exported for tests
        function getPathUpdater(oldFileOrDirPath, newFileOrDirPath, getCanonicalFileName, sourceMapper) {
            var canonicalOldPath = getCanonicalFileName(oldFileOrDirPath);
            return function (path) {
                var originalPath = sourceMapper && sourceMapper.tryGetSourcePosition({ fileName: path, pos: 0 });
                var updatedPath = getUpdatedPath(originalPath ? originalPath.fileName : path);
                return originalPath
                    ? updatedPath === undefined ? undefined : makeCorrespondingRelativeChange(originalPath.fileName, updatedPath, path, getCanonicalFileName)
                    : updatedPath;
            };
            function getUpdatedPath(pathToUpdate) {
                if (getCanonicalFileName(pathToUpdate) === canonicalOldPath)
                    return newFileOrDirPath;
                var suffix = ts.tryRemoveDirectoryPrefix(pathToUpdate, canonicalOldPath, getCanonicalFileName);
                return suffix === undefined ? undefined : newFileOrDirPath + "/" + suffix;
            }
        }
        ts.getPathUpdater = getPathUpdater;
        // Relative path from a0 to b0 should be same as relative path from a1 to b1. Returns b1.
        function makeCorrespondingRelativeChange(a0, b0, a1, getCanonicalFileName) {
            var rel = ts.getRelativePathFromFile(a0, b0, getCanonicalFileName);
            return combinePathsSafe(ts.getDirectoryPath(a1), rel);
        }
        function updateTsconfigFiles(program, changeTracker, oldToNew, oldFileOrDirPath, newFileOrDirPath, currentDirectory, useCaseSensitiveFileNames) {
            var configFile = program.getCompilerOptions().configFile;
            if (!configFile)
                return;
            var configDir = ts.getDirectoryPath(configFile.fileName);
            var jsonObjectLiteral = ts.getTsConfigObjectLiteralExpression(configFile);
            if (!jsonObjectLiteral)
                return;
            forEachProperty(jsonObjectLiteral, function (property, propertyName) {
                switch (propertyName) {
                    case "files":
                    case "include":
                    case "exclude": {
                        var foundExactMatch = updatePaths(property);
                        if (foundExactMatch || propertyName !== "include" || !ts.isArrayLiteralExpression(property.initializer))
                            return;
                        var includes = ts.mapDefined(property.initializer.elements, function (e) { return ts.isStringLiteral(e) ? e.text : undefined; });
                        if (includes.length === 0)
                            return;
                        var matchers = ts.getFileMatcherPatterns(configDir, /*excludes*/ [], includes, useCaseSensitiveFileNames, currentDirectory);
                        // If there isn't some include for this, add a new one.
                        if (ts.getRegexFromPattern(ts.Debug.checkDefined(matchers.includeFilePattern), useCaseSensitiveFileNames).test(oldFileOrDirPath) &&
                            !ts.getRegexFromPattern(ts.Debug.checkDefined(matchers.includeFilePattern), useCaseSensitiveFileNames).test(newFileOrDirPath)) {
                            changeTracker.insertNodeAfter(configFile, ts.last(property.initializer.elements), ts.factory.createStringLiteral(relativePath(newFileOrDirPath)));
                        }
                        return;
                    }
                    case "compilerOptions":
                        forEachProperty(property.initializer, function (property, propertyName) {
                            var option = ts.getOptionFromName(propertyName);
                            if (option && (option.isFilePath || option.type === "list" && option.element.isFilePath)) {
                                updatePaths(property);
                            }
                            else if (propertyName === "paths") {
                                forEachProperty(property.initializer, function (pathsProperty) {
                                    if (!ts.isArrayLiteralExpression(pathsProperty.initializer))
                                        return;
                                    for (var _i = 0, _a = pathsProperty.initializer.elements; _i < _a.length; _i++) {
                                        var e = _a[_i];
                                        tryUpdateString(e);
                                    }
                                });
                            }
                        });
                        return;
                }
            });
            function updatePaths(property) {
                var elements = ts.isArrayLiteralExpression(property.initializer) ? property.initializer.elements : [property.initializer];
                var foundExactMatch = false;
                for (var _i = 0, elements_1 = elements; _i < elements_1.length; _i++) {
                    var element = elements_1[_i];
                    foundExactMatch = tryUpdateString(element) || foundExactMatch;
                }
                return foundExactMatch;
            }
            function tryUpdateString(element) {
                if (!ts.isStringLiteral(element))
                    return false;
                var elementFileName = combinePathsSafe(configDir, element.text);
                var updated = oldToNew(elementFileName);
                if (updated !== undefined) {
                    changeTracker.replaceRangeWithText(configFile, createStringRange(element, configFile), relativePath(updated));
                    return true;
                }
                return false;
            }
            function relativePath(path) {
                return ts.getRelativePathFromDirectory(configDir, path, /*ignoreCase*/ !useCaseSensitiveFileNames);
            }
        }
        function updateImports(program, changeTracker, oldToNew, newToOld, host, getCanonicalFileName) {
            var allFiles = program.getSourceFiles();
            var _loop_5 = function (sourceFile) {
                var newFromOld = oldToNew(sourceFile.fileName);
                var newImportFromPath = newFromOld !== null && newFromOld !== void 0 ? newFromOld : sourceFile.fileName;
                var newImportFromDirectory = ts.getDirectoryPath(newImportFromPath);
                var oldFromNew = newToOld(sourceFile.fileName);
                var oldImportFromPath = oldFromNew || sourceFile.fileName;
                var oldImportFromDirectory = ts.getDirectoryPath(oldImportFromPath);
                var importingSourceFileMoved = newFromOld !== undefined || oldFromNew !== undefined;
                updateImportsWorker(sourceFile, changeTracker, function (referenceText) {
                    if (!ts.pathIsRelative(referenceText))
                        return undefined;
                    var oldAbsolute = combinePathsSafe(oldImportFromDirectory, referenceText);
                    var newAbsolute = oldToNew(oldAbsolute);
                    return newAbsolute === undefined ? undefined : ts.ensurePathIsNonModuleName(ts.getRelativePathFromDirectory(newImportFromDirectory, newAbsolute, getCanonicalFileName));
                }, function (importLiteral) {
                    var importedModuleSymbol = program.getTypeChecker().getSymbolAtLocation(importLiteral);
                    // No need to update if it's an ambient module^M
                    if ((importedModuleSymbol === null || importedModuleSymbol === void 0 ? void 0 : importedModuleSymbol.declarations) && importedModuleSymbol.declarations.some(function (d) { return ts.isAmbientModule(d); }))
                        return undefined;
                    var toImport = oldFromNew !== undefined
                        // If we're at the new location (file was already renamed), need to redo module resolution starting from the old location.
                        // TODO:GH#18217
                        ? getSourceFileToImportFromResolved(importLiteral, ts.resolveModuleName(importLiteral.text, oldImportFromPath, program.getCompilerOptions(), host), oldToNew, allFiles)
                        : getSourceFileToImport(importedModuleSymbol, importLiteral, sourceFile, program, host, oldToNew);
                    // Need an update if the imported file moved, or the importing file moved and was using a relative path.
                    return toImport !== undefined && (toImport.updated || (importingSourceFileMoved && ts.pathIsRelative(importLiteral.text)))
                        ? ts.moduleSpecifiers.updateModuleSpecifier(program.getCompilerOptions(), getCanonicalFileName(newImportFromPath), toImport.newFileName, ts.createModuleSpecifierResolutionHost(program, host), importLiteral.text)
                        : undefined;
                });
            };
            for (var _i = 0, allFiles_1 = allFiles; _i < allFiles_1.length; _i++) {
                var sourceFile = allFiles_1[_i];
                _loop_5(sourceFile);
            }
        }
        function combineNormal(pathA, pathB) {
            return ts.normalizePath(ts.combinePaths(pathA, pathB));
        }
        function combinePathsSafe(pathA, pathB) {
            return ts.ensurePathIsNonModuleName(combineNormal(pathA, pathB));
        }
        function getSourceFileToImport(importedModuleSymbol, importLiteral, importingSourceFile, program, host, oldToNew) {
            if (importedModuleSymbol) {
                // `find` should succeed because we checked for ambient modules before calling this function.
                var oldFileName = ts.find(importedModuleSymbol.declarations, ts.isSourceFile).fileName;
                var newFileName = oldToNew(oldFileName);
                return newFileName === undefined ? { newFileName: oldFileName, updated: false } : { newFileName: newFileName, updated: true };
            }
            else {
                var resolved = host.resolveModuleNames
                    ? host.getResolvedModuleWithFailedLookupLocationsFromCache && host.getResolvedModuleWithFailedLookupLocationsFromCache(importLiteral.text, importingSourceFile.fileName)
                    : program.getResolvedModuleWithFailedLookupLocationsFromCache(importLiteral.text, importingSourceFile.fileName);
                return getSourceFileToImportFromResolved(importLiteral, resolved, oldToNew, program.getSourceFiles());
            }
        }
        function getSourceFileToImportFromResolved(importLiteral, resolved, oldToNew, sourceFiles) {
            // Search through all locations looking for a moved file, and only then test already existing files.
            // This is because if `a.ts` is compiled to `a.js` and `a.ts` is moved, we don't want to resolve anything to `a.js`, but to `a.ts`'s new location.
            if (!resolved)
                return undefined;
            // First try resolved module
            if (resolved.resolvedModule) {
                var result_3 = tryChange(resolved.resolvedModule.resolvedFileName);
                if (result_3)
                    return result_3;
            }
            // Then failed lookups that are in the list of sources
            var result = ts.forEach(resolved.failedLookupLocations, tryChangeWithIgnoringPackageJsonExisting)
                // Then failed lookups except package.json since we dont want to touch them (only included ts/js files).
                // At this point, the confidence level of this fix being correct is too low to change bare specifiers or absolute paths.
                || ts.pathIsRelative(importLiteral.text) && ts.forEach(resolved.failedLookupLocations, tryChangeWithIgnoringPackageJson);
            if (result)
                return result;
            // If nothing changed, then result is resolved module file thats not updated
            return resolved.resolvedModule && { newFileName: resolved.resolvedModule.resolvedFileName, updated: false };
            function tryChangeWithIgnoringPackageJsonExisting(oldFileName) {
                var newFileName = oldToNew(oldFileName);
                return newFileName && ts.find(sourceFiles, function (src) { return src.fileName === newFileName; })
                    ? tryChangeWithIgnoringPackageJson(oldFileName) : undefined;
            }
            function tryChangeWithIgnoringPackageJson(oldFileName) {
                return !ts.endsWith(oldFileName, "/package.json") ? tryChange(oldFileName) : undefined;
            }
            function tryChange(oldFileName) {
                var newFileName = oldToNew(oldFileName);
                return newFileName && { newFileName: newFileName, updated: true };
            }
        }
        function updateImportsWorker(sourceFile, changeTracker, updateRef, updateImport) {
            for (var _i = 0, _a = sourceFile.referencedFiles || ts.emptyArray; _i < _a.length; _i++) { // TODO: GH#26162
                var ref = _a[_i];
                var updated = updateRef(ref.fileName);
                if (updated !== undefined && updated !== sourceFile.text.slice(ref.pos, ref.end))
                    changeTracker.replaceRangeWithText(sourceFile, ref, updated);
            }
            for (var _b = 0, _c = sourceFile.imports; _b < _c.length; _b++) {
                var importStringLiteral = _c[_b];
                var updated = updateImport(importStringLiteral);
                if (updated !== undefined && updated !== importStringLiteral.text)
                    changeTracker.replaceRangeWithText(sourceFile, createStringRange(importStringLiteral, sourceFile), updated);
            }
        }
        function createStringRange(node, sourceFile) {
            return ts.createRange(node.getStart(sourceFile) + 1, node.end - 1);
        }
        function forEachProperty(objectLiteral, cb) {
            if (!ts.isObjectLiteralExpression(objectLiteral))
                return;
            for (var _i = 0, _a = objectLiteral.properties; _i < _a.length; _i++) {
                var property = _a[_i];
                if (ts.isPropertyAssignment(property) && ts.isStringLiteral(property.name)) {
                    cb(property, property.name.text);
                }
            }
        }
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var GoToDefinition;
        (function (GoToDefinition) {
            function getDefinitionAtPosition(program, sourceFile, position) {
                var resolvedRef = getReferenceAtPosition(sourceFile, position, program);
                var fileReferenceDefinition = resolvedRef && [getDefinitionInfoForFileReference(resolvedRef.reference.fileName, resolvedRef.fileName, resolvedRef.unverified)] || ts.emptyArray;
                if (resolvedRef === null || resolvedRef === void 0 ? void 0 : resolvedRef.file) {
                    // If `file` is missing, do a symbol-based lookup as well
                    return fileReferenceDefinition;
                }
                var node = ts.getTouchingPropertyName(sourceFile, position);
                if (node === sourceFile) {
                    return undefined;
                }
                var parent = node.parent;
                var typeChecker = program.getTypeChecker();
                if (node.kind === 157 /* OverrideKeyword */ || (ts.isJSDocOverrideTag(node) && ts.rangeContainsPosition(node.tagName, position))) {
                    return getDefinitionFromOverriddenMember(typeChecker, node) || ts.emptyArray;
                }
                // Labels
                if (ts.isJumpStatementTarget(node)) {
                    var label = ts.getTargetLabel(node.parent, node.text);
                    return label ? [createDefinitionInfoFromName(typeChecker, label, "label" /* label */, node.text, /*containerName*/ undefined)] : undefined; // TODO: GH#18217
                }
                if (ts.isStaticModifier(node) && ts.isClassStaticBlockDeclaration(node.parent)) {
                    var classDecl = node.parent.parent;
                    var symbol_1 = getSymbol(classDecl, typeChecker);
                    var staticBlocks = ts.filter(classDecl.members, ts.isClassStaticBlockDeclaration);
                    var containerName_1 = symbol_1 ? typeChecker.symbolToString(symbol_1, classDecl) : "";
                    var sourceFile_1 = node.getSourceFile();
                    return ts.map(staticBlocks, function (staticBlock) {
                        var pos = ts.moveRangePastModifiers(staticBlock).pos;
                        pos = ts.skipTrivia(sourceFile_1.text, pos);
                        return createDefinitionInfoFromName(typeChecker, staticBlock, "constructor" /* constructorImplementationElement */, "static {}", containerName_1, { start: pos, length: "static".length });
                    });
                }
                var symbol = getSymbol(node, typeChecker);
                // Could not find a symbol e.g. node is string or number keyword,
                // or the symbol was an internal symbol and does not have a declaration e.g. undefined symbol
                if (!symbol) {
                    return ts.concatenate(fileReferenceDefinition, getDefinitionInfoForIndexSignatures(node, typeChecker));
                }
                var calledDeclaration = tryGetSignatureDeclaration(typeChecker, node);
                // Don't go to the component constructor definition for a JSX element, just go to the component definition.
                if (calledDeclaration && !(ts.isJsxOpeningLikeElement(node.parent) && isConstructorLike(calledDeclaration))) {
                    var sigInfo = createDefinitionFromSignatureDeclaration(typeChecker, calledDeclaration);
                    // For a function, if this is the original function definition, return just sigInfo.
                    // If this is the original constructor definition, parent is the class.
                    if (typeChecker.getRootSymbols(symbol).some(function (s) { return symbolMatchesSignature(s, calledDeclaration); })) {
                        return [sigInfo];
                    }
                    else {
                        var defs = getDefinitionFromSymbol(typeChecker, symbol, node, calledDeclaration) || ts.emptyArray;
                        // For a 'super()' call, put the signature first, else put the variable first.
                        return node.kind === 106 /* SuperKeyword */ ? __spreadArray([sigInfo], defs, true) : __spreadArray(__spreadArray([], defs, true), [sigInfo], false);
                    }
                }
                // Because name in short-hand property assignment has two different meanings: property name and property value,
                // using go-to-definition at such position should go to the variable declaration of the property value rather than
                // go to the declaration of the property name (in this case stay at the same position). However, if go-to-definition
                // is performed at the location of property access, we would like to go to definition of the property in the short-hand
                // assignment. This case and others are handled by the following code.
                if (node.parent.kind === 292 /* ShorthandPropertyAssignment */) {
                    var shorthandSymbol_1 = typeChecker.getShorthandAssignmentValueSymbol(symbol.valueDeclaration);
                    var definitions = (shorthandSymbol_1 === null || shorthandSymbol_1 === void 0 ? void 0 : shorthandSymbol_1.declarations) ? shorthandSymbol_1.declarations.map(function (decl) { return createDefinitionInfo(decl, typeChecker, shorthandSymbol_1, node); }) : ts.emptyArray;
                    return ts.concatenate(definitions, getDefinitionFromObjectLiteralElement(typeChecker, node) || ts.emptyArray);
                }
                // If the node is the name of a BindingElement within an ObjectBindingPattern instead of just returning the
                // declaration the symbol (which is itself), we should try to get to the original type of the ObjectBindingPattern
                // and return the property declaration for the referenced property.
                // For example:
                //      import('./foo').then(({ b/*goto*/ar }) => undefined); => should get use to the declaration in file "./foo"
                //
                //      function bar<T>(onfulfilled: (value: T) => void) { //....}
                //      interface Test {
                //          pr/*destination*/op1: number
                //      }
                //      bar<Test>(({pr/*goto*/op1})=>{});
                if (ts.isPropertyName(node) && ts.isBindingElement(parent) && ts.isObjectBindingPattern(parent.parent) &&
                    (node === (parent.propertyName || parent.name))) {
                    var name_3 = ts.getNameFromPropertyName(node);
                    var type = typeChecker.getTypeAtLocation(parent.parent);
                    return name_3 === undefined ? ts.emptyArray : ts.flatMap(type.isUnion() ? type.types : [type], function (t) {
                        var prop = t.getProperty(name_3);
                        return prop && getDefinitionFromSymbol(typeChecker, prop, node);
                    });
                }
                return ts.concatenate(fileReferenceDefinition, getDefinitionFromObjectLiteralElement(typeChecker, node) || getDefinitionFromSymbol(typeChecker, symbol, node));
            }
            GoToDefinition.getDefinitionAtPosition = getDefinitionAtPosition;
            /**
             * True if we should not add definitions for both the signature symbol and the definition symbol.
             * True for `const |f = |() => 0`, false for `function |f() {} const |g = f;`.
             * Also true for any assignment RHS.
             */
            function symbolMatchesSignature(s, calledDeclaration) {
                return s === calledDeclaration.symbol
                    || s === calledDeclaration.symbol.parent
                    || ts.isAssignmentExpression(calledDeclaration.parent)
                    || (!ts.isCallLikeExpression(calledDeclaration.parent) && s === calledDeclaration.parent.symbol);
            }
            // If the current location we want to find its definition is in an object literal, try to get the contextual type for the
            // object literal, lookup the property symbol in the contextual type, and use this for goto-definition.
            // For example
            //      interface Props{
            //          /*first*/prop1: number
            //          prop2: boolean
            //      }
            //      function Foo(arg: Props) {}
            //      Foo( { pr/*1*/op1: 10, prop2: true })
            function getDefinitionFromObjectLiteralElement(typeChecker, node) {
                var element = ts.getContainingObjectLiteralElement(node);
                if (element) {
                    var contextualType = element && typeChecker.getContextualType(element.parent);
                    if (contextualType) {
                        return ts.flatMap(ts.getPropertySymbolsFromContextualType(element, typeChecker, contextualType, /*unionSymbolOk*/ false), function (propertySymbol) {
                            return getDefinitionFromSymbol(typeChecker, propertySymbol, node);
                        });
                    }
                }
            }
            function getDefinitionFromOverriddenMember(typeChecker, node) {
                var classElement = ts.findAncestor(node, ts.isClassElement);
                if (!(classElement && classElement.name))
                    return;
                var baseDeclaration = ts.findAncestor(classElement, ts.isClassLike);
                if (!baseDeclaration)
                    return;
                var baseTypeNode = ts.getEffectiveBaseTypeNode(baseDeclaration);
                var baseType = baseTypeNode ? typeChecker.getTypeAtLocation(baseTypeNode) : undefined;
                if (!baseType)
                    return;
                var name = ts.unescapeLeadingUnderscores(ts.getTextOfPropertyName(classElement.name));
                var symbol = ts.hasStaticModifier(classElement)
                    ? typeChecker.getPropertyOfType(typeChecker.getTypeOfSymbolAtLocation(baseType.symbol, baseDeclaration), name)
                    : typeChecker.getPropertyOfType(baseType, name);
                if (!symbol)
                    return;
                return getDefinitionFromSymbol(typeChecker, symbol, node);
            }
            function getReferenceAtPosition(sourceFile, position, program) {
                var _a, _b;
                var referencePath = findReferenceInPosition(sourceFile.referencedFiles, position);
                if (referencePath) {
                    var file = program.getSourceFileFromReference(sourceFile, referencePath);
                    return file && { reference: referencePath, fileName: file.fileName, file: file, unverified: false };
                }
                var typeReferenceDirective = findReferenceInPosition(sourceFile.typeReferenceDirectives, position);
                if (typeReferenceDirective) {
                    var reference = program.getResolvedTypeReferenceDirectives().get(typeReferenceDirective.fileName);
                    var file = reference && program.getSourceFile(reference.resolvedFileName); // TODO:GH#18217
                    return file && { reference: typeReferenceDirective, fileName: file.fileName, file: file, unverified: false };
                }
                var libReferenceDirective = findReferenceInPosition(sourceFile.libReferenceDirectives, position);
                if (libReferenceDirective) {
                    var file = program.getLibFileFromReference(libReferenceDirective);
                    return file && { reference: libReferenceDirective, fileName: file.fileName, file: file, unverified: false };
                }
                if ((_a = sourceFile.resolvedModules) === null || _a === void 0 ? void 0 : _a.size) {
                    var node = ts.getTokenAtPosition(sourceFile, position);
                    if (ts.isModuleSpecifierLike(node) && ts.isExternalModuleNameRelative(node.text) && sourceFile.resolvedModules.has(node.text)) {
                        var verifiedFileName = (_b = sourceFile.resolvedModules.get(node.text)) === null || _b === void 0 ? void 0 : _b.resolvedFileName;
                        var fileName = verifiedFileName || ts.resolvePath(ts.getDirectoryPath(sourceFile.fileName), node.text);
                        return {
                            file: program.getSourceFile(fileName),
                            fileName: fileName,
                            reference: {
                                pos: node.getStart(),
                                end: node.getEnd(),
                                fileName: node.text
                            },
                            unverified: !verifiedFileName,
                        };
                    }
                }
                return undefined;
            }
            GoToDefinition.getReferenceAtPosition = getReferenceAtPosition;
            /// Goto type
            function getTypeDefinitionAtPosition(typeChecker, sourceFile, position) {
                var node = ts.getTouchingPropertyName(sourceFile, position);
                if (node === sourceFile) {
                    return undefined;
                }
                var symbol = typeChecker.getSymbolAtLocation(node);
                if (!symbol)
                    return undefined;
                var typeAtLocation = typeChecker.getTypeOfSymbolAtLocation(symbol, node);
                var returnType = tryGetReturnTypeOfFunction(symbol, typeAtLocation, typeChecker);
                var fromReturnType = returnType && definitionFromType(returnType, typeChecker, node);
                // If a function returns 'void' or some other type with no definition, just return the function definition.
                return fromReturnType && fromReturnType.length !== 0 ? fromReturnType : definitionFromType(typeAtLocation, typeChecker, node);
            }
            GoToDefinition.getTypeDefinitionAtPosition = getTypeDefinitionAtPosition;
            function definitionFromType(type, checker, node) {
                return ts.flatMap(type.isUnion() && !(type.flags & 32 /* Enum */) ? type.types : [type], function (t) {
                    return t.symbol && getDefinitionFromSymbol(checker, t.symbol, node);
                });
            }
            function tryGetReturnTypeOfFunction(symbol, type, checker) {
                // If the type is just a function's inferred type,
                // go-to-type should go to the return type instead, since go-to-definition takes you to the function anyway.
                if (type.symbol === symbol ||
                    // At `const f = () => {}`, the symbol is `f` and the type symbol is at `() => {}`
                    symbol.valueDeclaration && type.symbol && ts.isVariableDeclaration(symbol.valueDeclaration) && symbol.valueDeclaration.initializer === type.symbol.valueDeclaration) {
                    var sigs = type.getCallSignatures();
                    if (sigs.length === 1)
                        return checker.getReturnTypeOfSignature(ts.first(sigs));
                }
                return undefined;
            }
            function getDefinitionAndBoundSpan(program, sourceFile, position) {
                var definitions = getDefinitionAtPosition(program, sourceFile, position);
                if (!definitions || definitions.length === 0) {
                    return undefined;
                }
                // Check if position is on triple slash reference.
                var comment = findReferenceInPosition(sourceFile.referencedFiles, position) ||
                    findReferenceInPosition(sourceFile.typeReferenceDirectives, position) ||
                    findReferenceInPosition(sourceFile.libReferenceDirectives, position);
                if (comment) {
                    return { definitions: definitions, textSpan: ts.createTextSpanFromRange(comment) };
                }
                var node = ts.getTouchingPropertyName(sourceFile, position);
                var textSpan = ts.createTextSpan(node.getStart(), node.getWidth());
                return { definitions: definitions, textSpan: textSpan };
            }
            GoToDefinition.getDefinitionAndBoundSpan = getDefinitionAndBoundSpan;
            // At 'x.foo', see if the type of 'x' has an index signature, and if so find its declarations.
            function getDefinitionInfoForIndexSignatures(node, checker) {
                return ts.mapDefined(checker.getIndexInfosAtLocation(node), function (info) { return info.declaration && createDefinitionFromSignatureDeclaration(checker, info.declaration); });
            }
            function getSymbol(node, checker) {
                var symbol = checker.getSymbolAtLocation(node);
                // If this is an alias, and the request came at the declaration location
                // get the aliased symbol instead. This allows for goto def on an import e.g.
                //   import {A, B} from "mod";
                // to jump to the implementation directly.
                if ((symbol === null || symbol === void 0 ? void 0 : symbol.declarations) && symbol.flags & 2097152 /* Alias */ && shouldSkipAlias(node, symbol.declarations[0])) {
                    var aliased = checker.getAliasedSymbol(symbol);
                    if (aliased.declarations) {
                        return aliased;
                    }
                }
                return symbol;
            }
            // Go to the original declaration for cases:
            //
            //   (1) when the aliased symbol was declared in the location(parent).
            //   (2) when the aliased symbol is originating from an import.
            //
            function shouldSkipAlias(node, declaration) {
                if (node.kind !== 79 /* Identifier */) {
                    return false;
                }
                if (node.parent === declaration) {
                    return true;
                }
                switch (declaration.kind) {
                    case 265 /* ImportClause */:
                    case 263 /* ImportEqualsDeclaration */:
                        return true;
                    case 268 /* ImportSpecifier */:
                        return declaration.parent.kind === 267 /* NamedImports */;
                    case 201 /* BindingElement */:
                    case 252 /* VariableDeclaration */:
                        return ts.isInJSFile(declaration) && ts.isRequireVariableDeclaration(declaration);
                    default:
                        return false;
                }
            }
            function getDefinitionFromSymbol(typeChecker, symbol, node, declarationNode) {
                // There are cases when you extend a function by adding properties to it afterwards,
                // we want to strip those extra properties.
                // For deduping purposes, we also want to exclude any declarationNodes if provided.
                var filteredDeclarations = ts.filter(symbol.declarations, function (d) { return d !== declarationNode && (!ts.isAssignmentDeclaration(d) || d === symbol.valueDeclaration); })
                    || undefined;
                return getConstructSignatureDefinition() || getCallSignatureDefinition() || ts.map(filteredDeclarations, function (declaration) { return createDefinitionInfo(declaration, typeChecker, symbol, node); });
                function getConstructSignatureDefinition() {
                    // Applicable only if we are in a new expression, or we are on a constructor declaration
                    // and in either case the symbol has a construct signature definition, i.e. class
                    if (symbol.flags & 32 /* Class */ && !(symbol.flags & (16 /* Function */ | 3 /* Variable */)) && (ts.isNewExpressionTarget(node) || node.kind === 133 /* ConstructorKeyword */)) {
                        var cls = ts.find(filteredDeclarations, ts.isClassLike) || ts.Debug.fail("Expected declaration to have at least one class-like declaration");
                        return getSignatureDefinition(cls.members, /*selectConstructors*/ true);
                    }
                }
                function getCallSignatureDefinition() {
                    return ts.isCallOrNewExpressionTarget(node) || ts.isNameOfFunctionDeclaration(node)
                        ? getSignatureDefinition(filteredDeclarations, /*selectConstructors*/ false)
                        : undefined;
                }
                function getSignatureDefinition(signatureDeclarations, selectConstructors) {
                    if (!signatureDeclarations) {
                        return undefined;
                    }
                    var declarations = signatureDeclarations.filter(selectConstructors ? ts.isConstructorDeclaration : ts.isFunctionLike);
                    var declarationsWithBody = declarations.filter(function (d) { return !!d.body; });
                    // declarations defined on the global scope can be defined on multiple files. Get all of them.
                    return declarations.length
                        ? declarationsWithBody.length !== 0
                            ? declarationsWithBody.map(function (x) { return createDefinitionInfo(x, typeChecker, symbol, node); })
                            : [createDefinitionInfo(ts.last(declarations), typeChecker, symbol, node)]
                        : undefined;
                }
            }
            /** Creates a DefinitionInfo from a Declaration, using the declaration's name if possible. */
            function createDefinitionInfo(declaration, checker, symbol, node) {
                var symbolName = checker.symbolToString(symbol); // Do not get scoped name, just the name of the symbol
                var symbolKind = ts.SymbolDisplay.getSymbolKind(checker, symbol, node);
                var containerName = symbol.parent ? checker.symbolToString(symbol.parent, node) : "";
                return createDefinitionInfoFromName(checker, declaration, symbolKind, symbolName, containerName);
            }
            /** Creates a DefinitionInfo directly from the name of a declaration. */
            function createDefinitionInfoFromName(checker, declaration, symbolKind, symbolName, containerName, textSpan) {
                var sourceFile = declaration.getSourceFile();
                if (!textSpan) {
                    var name = ts.getNameOfDeclaration(declaration) || declaration;
                    textSpan = ts.createTextSpanFromNode(name, sourceFile);
                }
                return __assign(__assign({ fileName: sourceFile.fileName, textSpan: textSpan, kind: symbolKind, name: symbolName, containerKind: undefined,
                    containerName: containerName }, ts.FindAllReferences.toContextSpan(textSpan, sourceFile, ts.FindAllReferences.getContextNode(declaration))), { isLocal: !isDefinitionVisible(checker, declaration) });
            }
            function isDefinitionVisible(checker, declaration) {
                if (checker.isDeclarationVisible(declaration))
                    return true;
                if (!declaration.parent)
                    return false;
                // Variable initializers are visible if variable is visible
                if (ts.hasInitializer(declaration.parent) && declaration.parent.initializer === declaration)
                    return isDefinitionVisible(checker, declaration.parent);
                // Handle some exceptions here like arrow function, members of class and object literal expression which are technically not visible but we want the definition to be determined by its parent
                switch (declaration.kind) {
                    case 165 /* PropertyDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 167 /* MethodDeclaration */:
                        // Private/protected properties/methods are not visible
                        if (ts.hasEffectiveModifier(declaration, 8 /* Private */))
                            return false;
                    // Public properties/methods are visible if its parents are visible, so:
                    // falls through
                    case 169 /* Constructor */:
                    case 291 /* PropertyAssignment */:
                    case 292 /* ShorthandPropertyAssignment */:
                    case 203 /* ObjectLiteralExpression */:
                    case 224 /* ClassExpression */:
                    case 212 /* ArrowFunction */:
                    case 211 /* FunctionExpression */:
                        return isDefinitionVisible(checker, declaration.parent);
                    default:
                        return false;
                }
            }
            function createDefinitionFromSignatureDeclaration(typeChecker, decl) {
                return createDefinitionInfo(decl, typeChecker, decl.symbol, decl);
            }
            function findReferenceInPosition(refs, pos) {
                return ts.find(refs, function (ref) { return ts.textRangeContainsPositionInclusive(ref, pos); });
            }
            GoToDefinition.findReferenceInPosition = findReferenceInPosition;
            function getDefinitionInfoForFileReference(name, targetFileName, unverified) {
                return {
                    fileName: targetFileName,
                    textSpan: ts.createTextSpanFromBounds(0, 0),
                    kind: "script" /* scriptElement */,
                    name: name,
                    containerName: undefined,
                    containerKind: undefined,
                    unverified: unverified,
                };
            }
            /** Returns a CallLikeExpression where `node` is the target being invoked. */
            function getAncestorCallLikeExpression(node) {
                var target = ts.findAncestor(node, function (n) { return !ts.isRightSideOfPropertyAccess(n); });
                var callLike = target === null || target === void 0 ? void 0 : target.parent;
                return callLike && ts.isCallLikeExpression(callLike) && ts.getInvokedExpression(callLike) === target ? callLike : undefined;
            }
            function tryGetSignatureDeclaration(typeChecker, node) {
                var callLike = getAncestorCallLikeExpression(node);
                var signature = callLike && typeChecker.getResolvedSignature(callLike);
                // Don't go to a function type, go to the value having that type.
                return ts.tryCast(signature && signature.declaration, function (d) { return ts.isFunctionLike(d) && !ts.isFunctionTypeNode(d); });
            }
            function isConstructorLike(node) {
                switch (node.kind) {
                    case 169 /* Constructor */:
                    case 178 /* ConstructorType */:
                    case 173 /* ConstructSignature */:
                        return true;
                    default:
                        return false;
                }
            }
        })(GoToDefinition = ts.GoToDefinition || (ts.GoToDefinition = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var JsDoc;
        (function (JsDoc) {
            var jsDocTagNames = [
                "abstract",
                "access",
                "alias",
                "argument",
                "async",
                "augments",
                "author",
                "borrows",
                "callback",
                "class",
                "classdesc",
                "constant",
                "constructor",
                "constructs",
                "copyright",
                "default",
                "deprecated",
                "description",
                "emits",
                "enum",
                "event",
                "example",
                "exports",
                "extends",
                "external",
                "field",
                "file",
                "fileoverview",
                "fires",
                "function",
                "generator",
                "global",
                "hideconstructor",
                "host",
                "ignore",
                "implements",
                "inheritdoc",
                "inner",
                "instance",
                "interface",
                "kind",
                "lends",
                "license",
                "link",
                "listens",
                "member",
                "memberof",
                "method",
                "mixes",
                "module",
                "name",
                "namespace",
                "override",
                "package",
                "param",
                "private",
                "property",
                "protected",
                "public",
                "readonly",
                "requires",
                "returns",
                "see",
                "since",
                "static",
                "summary",
                "template",
                "this",
                "throws",
                "todo",
                "tutorial",
                "type",
                "typedef",
                "var",
                "variation",
                "version",
                "virtual",
                "yields"
            ];
            var jsDocTagNameCompletionEntries;
            var jsDocTagCompletionEntries;
            function getJsDocCommentsFromDeclarations(declarations, checker) {
                // Only collect doc comments from duplicate declarations once:
                // In case of a union property there might be same declaration multiple times
                // which only varies in type parameter
                // Eg. const a: Array<string> | Array<number>; a.length
                // The property length will have two declarations of property length coming
                // from Array<T> - Array<string> and Array<number>
                var parts = [];
                ts.forEachUnique(declarations, function (declaration) {
                    for (var _i = 0, _a = getCommentHavingNodes(declaration); _i < _a.length; _i++) {
                        var jsdoc = _a[_i];
                        // skip comments containing @typedefs since they're not associated with particular declarations
                        // Exceptions:
                        // - @typedefs are themselves declarations with associated comments
                        // - @param or @return indicate that the author thinks of it as a 'local' @typedef that's part of the function documentation
                        if (jsdoc.comment === undefined
                            || ts.isJSDoc(jsdoc)
                                && declaration.kind !== 340 /* JSDocTypedefTag */ && declaration.kind !== 333 /* JSDocCallbackTag */
                                && jsdoc.tags
                                && jsdoc.tags.some(function (t) { return t.kind === 340 /* JSDocTypedefTag */ || t.kind === 333 /* JSDocCallbackTag */; })
                                && !jsdoc.tags.some(function (t) { return t.kind === 335 /* JSDocParameterTag */ || t.kind === 336 /* JSDocReturnTag */; })) {
                            continue;
                        }
                        var newparts = getDisplayPartsFromComment(jsdoc.comment, checker);
                        if (!ts.contains(parts, newparts, isIdenticalListOfDisplayParts)) {
                            parts.push(newparts);
                        }
                    }
                });
                return ts.flatten(ts.intersperse(parts, [ts.lineBreakPart()]));
            }
            JsDoc.getJsDocCommentsFromDeclarations = getJsDocCommentsFromDeclarations;
            function isIdenticalListOfDisplayParts(parts1, parts2) {
                return ts.arraysEqual(parts1, parts2, function (p1, p2) { return p1.kind === p2.kind && p1.text === p2.text; });
            }
            function getCommentHavingNodes(declaration) {
                switch (declaration.kind) {
                    case 335 /* JSDocParameterTag */:
                    case 342 /* JSDocPropertyTag */:
                        return [declaration];
                    case 333 /* JSDocCallbackTag */:
                    case 340 /* JSDocTypedefTag */:
                        return [declaration, declaration.parent];
                    default:
                        return ts.getJSDocCommentsAndTags(declaration);
                }
            }
            function getJsDocTagsFromDeclarations(declarations, checker) {
                // Only collect doc comments from duplicate declarations once.
                var infos = [];
                ts.forEachUnique(declarations, function (declaration) {
                    var tags = ts.getJSDocTags(declaration);
                    // skip comments containing @typedefs since they're not associated with particular declarations
                    // Exceptions:
                    // - @param or @return indicate that the author thinks of it as a 'local' @typedef that's part of the function documentation
                    if (tags.some(function (t) { return t.kind === 340 /* JSDocTypedefTag */ || t.kind === 333 /* JSDocCallbackTag */; })
                        && !tags.some(function (t) { return t.kind === 335 /* JSDocParameterTag */ || t.kind === 336 /* JSDocReturnTag */; })) {
                        return;
                    }
                    for (var _i = 0, tags_1 = tags; _i < tags_1.length; _i++) {
                        var tag = tags_1[_i];
                        infos.push({ name: tag.tagName.text, text: getCommentDisplayParts(tag, checker) });
                    }
                });
                return infos;
            }
            JsDoc.getJsDocTagsFromDeclarations = getJsDocTagsFromDeclarations;
            function getDisplayPartsFromComment(comment, checker) {
                if (typeof comment === "string") {
                    return [ts.textPart(comment)];
                }
                return ts.flatMap(comment, function (node) { return node.kind === 316 /* JSDocText */ ? [ts.textPart(node.text)] : ts.buildLinkParts(node, checker); });
            }
            function getCommentDisplayParts(tag, checker) {
                var comment = tag.comment, kind = tag.kind;
                var namePart = getTagNameDisplayPart(kind);
                switch (kind) {
                    case 324 /* JSDocImplementsTag */:
                        return withNode(tag.class);
                    case 323 /* JSDocAugmentsTag */:
                        return withNode(tag.class);
                    case 339 /* JSDocTemplateTag */:
                        return addComment(tag.typeParameters.map(function (tp) { return tp.getText(); }).join(", "));
                    case 338 /* JSDocTypeTag */:
                        return withNode(tag.typeExpression);
                    case 340 /* JSDocTypedefTag */:
                    case 333 /* JSDocCallbackTag */:
                    case 342 /* JSDocPropertyTag */:
                    case 335 /* JSDocParameterTag */:
                    case 341 /* JSDocSeeTag */:
                        var name = tag.name;
                        return name ? withNode(name)
                            : comment === undefined ? undefined
                                : getDisplayPartsFromComment(comment, checker);
                    default:
                        return comment === undefined ? undefined : getDisplayPartsFromComment(comment, checker);
                }
                function withNode(node) {
                    return addComment(node.getText());
                }
                function addComment(s) {
                    if (comment) {
                        if (s.match(/^https?$/)) {
                            return __spreadArray([ts.textPart(s)], getDisplayPartsFromComment(comment, checker), true);
                        }
                        else {
                            return __spreadArray([namePart(s), ts.spacePart()], getDisplayPartsFromComment(comment, checker), true);
                        }
                    }
                    else {
                        return [ts.textPart(s)];
                    }
                }
            }
            function getTagNameDisplayPart(kind) {
                switch (kind) {
                    case 335 /* JSDocParameterTag */:
                        return ts.parameterNamePart;
                    case 342 /* JSDocPropertyTag */:
                        return ts.propertyNamePart;
                    case 339 /* JSDocTemplateTag */:
                        return ts.typeParameterNamePart;
                    case 340 /* JSDocTypedefTag */:
                    case 333 /* JSDocCallbackTag */:
                        return ts.typeAliasNamePart;
                    default:
                        return ts.textPart;
                }
            }
            function getJSDocTagNameCompletions() {
                return jsDocTagNameCompletionEntries || (jsDocTagNameCompletionEntries = ts.map(jsDocTagNames, function (tagName) {
                    return {
                        name: tagName,
                        kind: "keyword" /* keyword */,
                        kindModifiers: "",
                        sortText: ts.Completions.SortText.LocationPriority,
                    };
                }));
            }
            JsDoc.getJSDocTagNameCompletions = getJSDocTagNameCompletions;
            JsDoc.getJSDocTagNameCompletionDetails = getJSDocTagCompletionDetails;
            function getJSDocTagCompletions() {
                return jsDocTagCompletionEntries || (jsDocTagCompletionEntries = ts.map(jsDocTagNames, function (tagName) {
                    return {
                        name: "@" + tagName,
                        kind: "keyword" /* keyword */,
                        kindModifiers: "",
                        sortText: ts.Completions.SortText.LocationPriority
                    };
                }));
            }
            JsDoc.getJSDocTagCompletions = getJSDocTagCompletions;
            function getJSDocTagCompletionDetails(name) {
                return {
                    name: name,
                    kind: "" /* unknown */,
                    kindModifiers: "",
                    displayParts: [ts.textPart(name)],
                    documentation: ts.emptyArray,
                    tags: undefined,
                    codeActions: undefined,
                };
            }
            JsDoc.getJSDocTagCompletionDetails = getJSDocTagCompletionDetails;
            function getJSDocParameterNameCompletions(tag) {
                if (!ts.isIdentifier(tag.name)) {
                    return ts.emptyArray;
                }
                var nameThusFar = tag.name.text;
                var jsdoc = tag.parent;
                var fn = jsdoc.parent;
                if (!ts.isFunctionLike(fn))
                    return [];
                return ts.mapDefined(fn.parameters, function (param) {
                    if (!ts.isIdentifier(param.name))
                        return undefined;
                    var name = param.name.text;
                    if (jsdoc.tags.some(function (t) { return t !== tag && ts.isJSDocParameterTag(t) && ts.isIdentifier(t.name) && t.name.escapedText === name; }) // TODO: GH#18217
                        || nameThusFar !== undefined && !ts.startsWith(name, nameThusFar)) {
                        return undefined;
                    }
                    return { name: name, kind: "parameter" /* parameterElement */, kindModifiers: "", sortText: ts.Completions.SortText.LocationPriority };
                });
            }
            JsDoc.getJSDocParameterNameCompletions = getJSDocParameterNameCompletions;
            function getJSDocParameterNameCompletionDetails(name) {
                return {
                    name: name,
                    kind: "parameter" /* parameterElement */,
                    kindModifiers: "",
                    displayParts: [ts.textPart(name)],
                    documentation: ts.emptyArray,
                    tags: undefined,
                    codeActions: undefined,
                };
            }
            JsDoc.getJSDocParameterNameCompletionDetails = getJSDocParameterNameCompletionDetails;
            /**
             * Checks if position points to a valid position to add JSDoc comments, and if so,
             * returns the appropriate template. Otherwise returns an empty string.
             * Valid positions are
             *      - outside of comments, statements, and expressions, and
             *      - preceding a:
             *          - function/constructor/method declaration
             *          - class declarations
             *          - variable statements
             *          - namespace declarations
             *          - interface declarations
             *          - method signatures
             *          - type alias declarations
             *
             * Hosts should ideally check that:
             * - The line is all whitespace up to 'position' before performing the insertion.
             * - If the keystroke sequence "/\*\*" induced the call, we also check that the next
             * non-whitespace character is '*', which (approximately) indicates whether we added
             * the second '*' to complete an existing (JSDoc) comment.
             * @param fileName The file in which to perform the check.
             * @param position The (character-indexed) position in the file where the check should
             * be performed.
             */
            function getDocCommentTemplateAtPosition(newLine, sourceFile, position, options) {
                var tokenAtPos = ts.getTokenAtPosition(sourceFile, position);
                var existingDocComment = ts.findAncestor(tokenAtPos, ts.isJSDoc);
                if (existingDocComment && (existingDocComment.comment !== undefined || ts.length(existingDocComment.tags))) {
                    // Non-empty comment already exists.
                    return undefined;
                }
                var tokenStart = tokenAtPos.getStart(sourceFile);
                // Don't provide a doc comment template based on a *previous* node. (But an existing empty jsdoc comment will likely start before `position`.)
                if (!existingDocComment && tokenStart < position) {
                    return undefined;
                }
                var commentOwnerInfo = getCommentOwnerInfo(tokenAtPos, options);
                if (!commentOwnerInfo) {
                    return undefined;
                }
                var commentOwner = commentOwnerInfo.commentOwner, parameters = commentOwnerInfo.parameters, hasReturn = commentOwnerInfo.hasReturn;
                if (commentOwner.getStart(sourceFile) < position) {
                    return undefined;
                }
                var indentationStr = getIndentationStringAtPosition(sourceFile, position);
                var isJavaScriptFile = ts.hasJSFileExtension(sourceFile.fileName);
                var tags = (parameters ? parameterDocComments(parameters || [], isJavaScriptFile, indentationStr, newLine) : "") +
                    (hasReturn ? returnsDocComment(indentationStr, newLine) : "");
                // A doc comment consists of the following
                // * The opening comment line
                // * the first line (without a param) for the object's untagged info (this is also where the caret ends up)
                // * the '@param'-tagged lines
                // * the '@returns'-tag
                // * TODO: other tags.
                // * the closing comment line
                // * if the caret was directly in front of the object, then we add an extra line and indentation.
                var openComment = "/**";
                var closeComment = " */";
                if (tags) {
                    var preamble = openComment + newLine + indentationStr + " * ";
                    var endLine = tokenStart === position ? newLine + indentationStr : "";
                    var result = preamble + newLine + tags + indentationStr + closeComment + endLine;
                    return { newText: result, caretOffset: preamble.length };
                }
                return { newText: openComment + closeComment, caretOffset: 3 };
            }
            JsDoc.getDocCommentTemplateAtPosition = getDocCommentTemplateAtPosition;
            function getIndentationStringAtPosition(sourceFile, position) {
                var text = sourceFile.text;
                var lineStart = ts.getLineStartPositionForPosition(position, sourceFile);
                var pos = lineStart;
                for (; pos <= position && ts.isWhiteSpaceSingleLine(text.charCodeAt(pos)); pos++)
                    ;
                return text.slice(lineStart, pos);
            }
            function parameterDocComments(parameters, isJavaScriptFile, indentationStr, newLine) {
                return parameters.map(function (_a, i) {
                    var name = _a.name, dotDotDotToken = _a.dotDotDotToken;
                    var paramName = name.kind === 79 /* Identifier */ ? name.text : "param" + i;
                    var type = isJavaScriptFile ? (dotDotDotToken ? "{...any} " : "{any} ") : "";
                    return indentationStr + " * @param " + type + paramName + newLine;
                }).join("");
            }
            function returnsDocComment(indentationStr, newLine) {
                return indentationStr + " * @returns" + newLine;
            }
            function getCommentOwnerInfo(tokenAtPos, options) {
                return ts.forEachAncestor(tokenAtPos, function (n) { return getCommentOwnerInfoWorker(n, options); });
            }
            function getCommentOwnerInfoWorker(commentOwner, options) {
                switch (commentOwner.kind) {
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 167 /* MethodDeclaration */:
                    case 169 /* Constructor */:
                    case 166 /* MethodSignature */:
                    case 212 /* ArrowFunction */:
                        var host = commentOwner;
                        return { commentOwner: commentOwner, parameters: host.parameters, hasReturn: hasReturn(host, options) };
                    case 291 /* PropertyAssignment */:
                        return getCommentOwnerInfoWorker(commentOwner.initializer, options);
                    case 255 /* ClassDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 164 /* PropertySignature */:
                    case 258 /* EnumDeclaration */:
                    case 294 /* EnumMember */:
                    case 257 /* TypeAliasDeclaration */:
                        return { commentOwner: commentOwner };
                    case 235 /* VariableStatement */: {
                        var varStatement = commentOwner;
                        var varDeclarations = varStatement.declarationList.declarations;
                        var host_1 = varDeclarations.length === 1 && varDeclarations[0].initializer
                            ? getRightHandSideOfAssignment(varDeclarations[0].initializer)
                            : undefined;
                        return host_1
                            ? { commentOwner: commentOwner, parameters: host_1.parameters, hasReturn: hasReturn(host_1, options) }
                            : { commentOwner: commentOwner };
                    }
                    case 300 /* SourceFile */:
                        return "quit";
                    case 259 /* ModuleDeclaration */:
                        // If in walking up the tree, we hit a a nested namespace declaration,
                        // then we must be somewhere within a dotted namespace name; however we don't
                        // want to give back a JSDoc template for the 'b' or 'c' in 'namespace a.b.c { }'.
                        return commentOwner.parent.kind === 259 /* ModuleDeclaration */ ? undefined : { commentOwner: commentOwner };
                    case 236 /* ExpressionStatement */:
                        return getCommentOwnerInfoWorker(commentOwner.expression, options);
                    case 219 /* BinaryExpression */: {
                        var be = commentOwner;
                        if (ts.getAssignmentDeclarationKind(be) === 0 /* None */) {
                            return "quit";
                        }
                        return ts.isFunctionLike(be.right)
                            ? { commentOwner: commentOwner, parameters: be.right.parameters, hasReturn: hasReturn(be.right, options) }
                            : { commentOwner: commentOwner };
                    }
                    case 165 /* PropertyDeclaration */:
                        var init = commentOwner.initializer;
                        if (init && (ts.isFunctionExpression(init) || ts.isArrowFunction(init))) {
                            return { commentOwner: commentOwner, parameters: init.parameters, hasReturn: hasReturn(init, options) };
                        }
                }
            }
            function hasReturn(node, options) {
                return !!(options === null || options === void 0 ? void 0 : options.generateReturnInDocTemplate) &&
                    (ts.isArrowFunction(node) && ts.isExpression(node.body)
                        || ts.isFunctionLikeDeclaration(node) && node.body && ts.isBlock(node.body) && !!ts.forEachReturnStatement(node.body, function (n) { return n; }));
            }
            function getRightHandSideOfAssignment(rightHandSide) {
                while (rightHandSide.kind === 210 /* ParenthesizedExpression */) {
                    rightHandSide = rightHandSide.expression;
                }
                switch (rightHandSide.kind) {
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        return rightHandSide;
                    case 224 /* ClassExpression */:
                        return ts.find(rightHandSide.members, ts.isConstructorDeclaration);
                }
            }
        })(JsDoc = ts.JsDoc || (ts.JsDoc = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var NavigateTo;
        (function (NavigateTo) {
            function getNavigateToItems(sourceFiles, checker, cancellationToken, searchValue, maxResultCount, excludeDtsFiles) {
                var patternMatcher = ts.createPatternMatcher(searchValue);
                if (!patternMatcher)
                    return ts.emptyArray;
                var rawItems = [];
                var _loop_6 = function (sourceFile) {
                    cancellationToken.throwIfCancellationRequested();
                    if (excludeDtsFiles && sourceFile.isDeclarationFile) {
                        return "continue";
                    }
                    sourceFile.getNamedDeclarations().forEach(function (declarations, name) {
                        getItemsFromNamedDeclaration(patternMatcher, name, declarations, checker, sourceFile.fileName, rawItems);
                    });
                };
                // Search the declarations in all files and output matched NavigateToItem into array of NavigateToItem[]
                for (var _i = 0, sourceFiles_4 = sourceFiles; _i < sourceFiles_4.length; _i++) {
                    var sourceFile = sourceFiles_4[_i];
                    _loop_6(sourceFile);
                }
                rawItems.sort(compareNavigateToItems);
                return (maxResultCount === undefined ? rawItems : rawItems.slice(0, maxResultCount)).map(createNavigateToItem);
            }
            NavigateTo.getNavigateToItems = getNavigateToItems;
            function getItemsFromNamedDeclaration(patternMatcher, name, declarations, checker, fileName, rawItems) {
                // First do a quick check to see if the name of the declaration matches the
                // last portion of the (possibly) dotted name they're searching for.
                var match = patternMatcher.getMatchForLastSegmentOfPattern(name);
                if (!match) {
                    return; // continue to next named declarations
                }
                for (var _i = 0, declarations_3 = declarations; _i < declarations_3.length; _i++) {
                    var declaration = declarations_3[_i];
                    if (!shouldKeepItem(declaration, checker))
                        continue;
                    if (patternMatcher.patternContainsDots) {
                        // If the pattern has dots in it, then also see if the declaration container matches as well.
                        var fullMatch = patternMatcher.getFullMatch(getContainers(declaration), name);
                        if (fullMatch) {
                            rawItems.push({ name: name, fileName: fileName, matchKind: fullMatch.kind, isCaseSensitive: fullMatch.isCaseSensitive, declaration: declaration });
                        }
                    }
                    else {
                        rawItems.push({ name: name, fileName: fileName, matchKind: match.kind, isCaseSensitive: match.isCaseSensitive, declaration: declaration });
                    }
                }
            }
            function shouldKeepItem(declaration, checker) {
                switch (declaration.kind) {
                    case 265 /* ImportClause */:
                    case 268 /* ImportSpecifier */:
                    case 263 /* ImportEqualsDeclaration */:
                        var importer = checker.getSymbolAtLocation(declaration.name); // TODO: GH#18217
                        var imported = checker.getAliasedSymbol(importer);
                        return importer.escapedName !== imported.escapedName;
                    default:
                        return true;
                }
            }
            function tryAddSingleDeclarationName(declaration, containers) {
                var name = ts.getNameOfDeclaration(declaration);
                return !!name && (pushLiteral(name, containers) || name.kind === 160 /* ComputedPropertyName */ && tryAddComputedPropertyName(name.expression, containers));
            }
            // Only added the names of computed properties if they're simple dotted expressions, like:
            //
            //      [X.Y.Z]() { }
            function tryAddComputedPropertyName(expression, containers) {
                return pushLiteral(expression, containers)
                    || ts.isPropertyAccessExpression(expression) && (containers.push(expression.name.text), true) && tryAddComputedPropertyName(expression.expression, containers);
            }
            function pushLiteral(node, containers) {
                return ts.isPropertyNameLiteral(node) && (containers.push(ts.getTextOfIdentifierOrLiteral(node)), true);
            }
            function getContainers(declaration) {
                var containers = [];
                // First, if we started with a computed property name, then add all but the last
                // portion into the container array.
                var name = ts.getNameOfDeclaration(declaration);
                if (name && name.kind === 160 /* ComputedPropertyName */ && !tryAddComputedPropertyName(name.expression, containers)) {
                    return ts.emptyArray;
                }
                // Don't include the last portion.
                containers.shift();
                // Now, walk up our containers, adding all their names to the container array.
                var container = ts.getContainerNode(declaration);
                while (container) {
                    if (!tryAddSingleDeclarationName(container, containers)) {
                        return ts.emptyArray;
                    }
                    container = ts.getContainerNode(container);
                }
                return containers.reverse();
            }
            function compareNavigateToItems(i1, i2) {
                // TODO(cyrusn): get the gamut of comparisons that VS already uses here.
                return ts.compareValues(i1.matchKind, i2.matchKind)
                    || ts.compareStringsCaseSensitiveUI(i1.name, i2.name);
            }
            function createNavigateToItem(rawItem) {
                var declaration = rawItem.declaration;
                var container = ts.getContainerNode(declaration);
                var containerName = container && ts.getNameOfDeclaration(container);
                return {
                    name: rawItem.name,
                    kind: ts.getNodeKind(declaration),
                    kindModifiers: ts.getNodeModifiers(declaration),
                    matchKind: ts.PatternMatchKind[rawItem.matchKind],
                    isCaseSensitive: rawItem.isCaseSensitive,
                    fileName: rawItem.fileName,
                    textSpan: ts.createTextSpanFromNode(declaration),
                    // TODO(jfreeman): What should be the containerName when the container has a computed name?
                    containerName: containerName ? containerName.text : "",
                    containerKind: containerName ? ts.getNodeKind(container) : "" /* unknown */, // TODO: GH#18217 Just use `container ? ...`
                };
            }
        })(NavigateTo = ts.NavigateTo || (ts.NavigateTo = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var NavigationBar;
        (function (NavigationBar) {
            var _a;
            /**
             * Matches all whitespace characters in a string. Eg:
             *
             * "app.
             *
             * onactivated"
             *
             * matches because of the newline, whereas
             *
             * "app.onactivated"
             *
             * does not match.
             */
            var whiteSpaceRegex = /\s+/g;
            /**
             * Maximum amount of characters to return
             * The amount was chosen arbitrarily.
             */
            var maxLength = 150;
            // Keep sourceFile handy so we don't have to search for it every time we need to call `getText`.
            var curCancellationToken;
            var curSourceFile;
            /**
             * For performance, we keep navigation bar parents on a stack rather than passing them through each recursion.
             * `parent` is the current parent and is *not* stored in parentsStack.
             * `startNode` sets a new parent and `endNode` returns to the previous parent.
             */
            var parentsStack = [];
            var parent;
            var trackedEs5ClassesStack = [];
            var trackedEs5Classes;
            // NavigationBarItem requires an array, but will not mutate it, so just give it this for performance.
            var emptyChildItemArray = [];
            function getNavigationBarItems(sourceFile, cancellationToken) {
                curCancellationToken = cancellationToken;
                curSourceFile = sourceFile;
                try {
                    return ts.map(primaryNavBarMenuItems(rootNavigationBarNode(sourceFile)), convertToPrimaryNavBarMenuItem);
                }
                finally {
                    reset();
                }
            }
            NavigationBar.getNavigationBarItems = getNavigationBarItems;
            function getNavigationTree(sourceFile, cancellationToken) {
                curCancellationToken = cancellationToken;
                curSourceFile = sourceFile;
                try {
                    return convertToTree(rootNavigationBarNode(sourceFile));
                }
                finally {
                    reset();
                }
            }
            NavigationBar.getNavigationTree = getNavigationTree;
            function reset() {
                curSourceFile = undefined;
                curCancellationToken = undefined;
                parentsStack = [];
                parent = undefined;
                emptyChildItemArray = [];
            }
            function nodeText(node) {
                return cleanText(node.getText(curSourceFile));
            }
            function navigationBarNodeKind(n) {
                return n.node.kind;
            }
            function pushChild(parent, child) {
                if (parent.children) {
                    parent.children.push(child);
                }
                else {
                    parent.children = [child];
                }
            }
            function rootNavigationBarNode(sourceFile) {
                ts.Debug.assert(!parentsStack.length);
                var root = { node: sourceFile, name: undefined, additionalNodes: undefined, parent: undefined, children: undefined, indent: 0 };
                parent = root;
                for (var _i = 0, _a = sourceFile.statements; _i < _a.length; _i++) {
                    var statement = _a[_i];
                    addChildrenRecursively(statement);
                }
                endNode();
                ts.Debug.assert(!parent && !parentsStack.length);
                return root;
            }
            function addLeafNode(node, name) {
                pushChild(parent, emptyNavigationBarNode(node, name));
            }
            function emptyNavigationBarNode(node, name) {
                return {
                    node: node,
                    name: name || (ts.isDeclaration(node) || ts.isExpression(node) ? ts.getNameOfDeclaration(node) : undefined),
                    additionalNodes: undefined,
                    parent: parent,
                    children: undefined,
                    indent: parent.indent + 1
                };
            }
            function addTrackedEs5Class(name) {
                if (!trackedEs5Classes) {
                    trackedEs5Classes = new ts.Map();
                }
                trackedEs5Classes.set(name, true);
            }
            function endNestedNodes(depth) {
                for (var i = 0; i < depth; i++)
                    endNode();
            }
            function startNestedNodes(targetNode, entityName) {
                var names = [];
                while (!ts.isPropertyNameLiteral(entityName)) {
                    var name = ts.getNameOrArgument(entityName);
                    var nameText = ts.getElementOrPropertyAccessName(entityName);
                    entityName = entityName.expression;
                    if (nameText === "prototype" || ts.isPrivateIdentifier(name))
                        continue;
                    names.push(name);
                }
                names.push(entityName);
                for (var i = names.length - 1; i > 0; i--) {
                    var name = names[i];
                    startNode(targetNode, name);
                }
                return [names.length - 1, names[0]];
            }
            /**
             * Add a new level of NavigationBarNodes.
             * This pushes to the stack, so you must call `endNode` when you are done adding to this node.
             */
            function startNode(node, name) {
                var navNode = emptyNavigationBarNode(node, name);
                pushChild(parent, navNode);
                // Save the old parent
                parentsStack.push(parent);
                trackedEs5ClassesStack.push(trackedEs5Classes);
                trackedEs5Classes = undefined;
                parent = navNode;
            }
            /** Call after calling `startNode` and adding children to it. */
            function endNode() {
                if (parent.children) {
                    mergeChildren(parent.children, parent);
                    sortChildren(parent.children);
                }
                parent = parentsStack.pop();
                trackedEs5Classes = trackedEs5ClassesStack.pop();
            }
            function addNodeWithRecursiveChild(node, child, name) {
                startNode(node, name);
                addChildrenRecursively(child);
                endNode();
            }
            function addNodeWithRecursiveInitializer(node) {
                if (node.initializer && isFunctionOrClassExpression(node.initializer)) {
                    startNode(node);
                    ts.forEachChild(node.initializer, addChildrenRecursively);
                    endNode();
                }
                else {
                    addNodeWithRecursiveChild(node, node.initializer);
                }
            }
            /**
             * Historically, we've elided dynamic names from the nav tree (including late bound names),
             * but included certain "well known" symbol names. While we no longer distinguish those well-known
             * symbols from other unique symbols, we do the below to retain those members in the nav tree.
             */
            function hasNavigationBarName(node) {
                return !ts.hasDynamicName(node) ||
                    (node.kind !== 219 /* BinaryExpression */ &&
                        ts.isPropertyAccessExpression(node.name.expression) &&
                        ts.isIdentifier(node.name.expression.expression) &&
                        ts.idText(node.name.expression.expression) === "Symbol");
            }
            /** Look for navigation bar items in node's subtree, adding them to the current `parent`. */
            function addChildrenRecursively(node) {
                var _a;
                curCancellationToken.throwIfCancellationRequested();
                if (!node || ts.isToken(node)) {
                    return;
                }
                switch (node.kind) {
                    case 169 /* Constructor */:
                        // Get parameter properties, and treat them as being on the *same* level as the constructor, not under it.
                        var ctr = node;
                        addNodeWithRecursiveChild(ctr, ctr.body);
                        // Parameter properties are children of the class, not the constructor.
                        for (var _i = 0, _b = ctr.parameters; _i < _b.length; _i++) {
                            var param = _b[_i];
                            if (ts.isParameterPropertyDeclaration(param, ctr)) {
                                addLeafNode(param);
                            }
                        }
                        break;
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    case 166 /* MethodSignature */:
                        if (hasNavigationBarName(node)) {
                            addNodeWithRecursiveChild(node, node.body);
                        }
                        break;
                    case 165 /* PropertyDeclaration */:
                        if (hasNavigationBarName(node)) {
                            addNodeWithRecursiveInitializer(node);
                        }
                        break;
                    case 164 /* PropertySignature */:
                        if (hasNavigationBarName(node)) {
                            addLeafNode(node);
                        }
                        break;
                    case 265 /* ImportClause */:
                        var importClause = node;
                        // Handle default import case e.g.:
                        //    import d from "mod";
                        if (importClause.name) {
                            addLeafNode(importClause.name);
                        }
                        // Handle named bindings in imports e.g.:
                        //    import * as NS from "mod";
                        //    import {a, b as B} from "mod";
                        var namedBindings = importClause.namedBindings;
                        if (namedBindings) {
                            if (namedBindings.kind === 266 /* NamespaceImport */) {
                                addLeafNode(namedBindings);
                            }
                            else {
                                for (var _c = 0, _d = namedBindings.elements; _c < _d.length; _c++) {
                                    var element = _d[_c];
                                    addLeafNode(element);
                                }
                            }
                        }
                        break;
                    case 292 /* ShorthandPropertyAssignment */:
                        addNodeWithRecursiveChild(node, node.name);
                        break;
                    case 293 /* SpreadAssignment */:
                        var expression = node.expression;
                        // Use the expression as the name of the SpreadAssignment, otherwise show as <unknown>.
                        ts.isIdentifier(expression) ? addLeafNode(node, expression) : addLeafNode(node);
                        break;
                    case 201 /* BindingElement */:
                    case 291 /* PropertyAssignment */:
                    case 252 /* VariableDeclaration */: {
                        var child = node;
                        if (ts.isBindingPattern(child.name)) {
                            addChildrenRecursively(child.name);
                        }
                        else {
                            addNodeWithRecursiveInitializer(child);
                        }
                        break;
                    }
                    case 254 /* FunctionDeclaration */:
                        var nameNode = node.name;
                        // If we see a function declaration track as a possible ES5 class
                        if (nameNode && ts.isIdentifier(nameNode)) {
                            addTrackedEs5Class(nameNode.text);
                        }
                        addNodeWithRecursiveChild(node, node.body);
                        break;
                    case 212 /* ArrowFunction */:
                    case 211 /* FunctionExpression */:
                        addNodeWithRecursiveChild(node, node.body);
                        break;
                    case 258 /* EnumDeclaration */:
                        startNode(node);
                        for (var _e = 0, _f = node.members; _e < _f.length; _e++) {
                            var member = _f[_e];
                            if (!isComputedProperty(member)) {
                                addLeafNode(member);
                            }
                        }
                        endNode();
                        break;
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                    case 256 /* InterfaceDeclaration */:
                        startNode(node);
                        for (var _g = 0, _h = node.members; _g < _h.length; _g++) {
                            var member = _h[_g];
                            addChildrenRecursively(member);
                        }
                        endNode();
                        break;
                    case 259 /* ModuleDeclaration */:
                        addNodeWithRecursiveChild(node, getInteriorModule(node).body);
                        break;
                    case 269 /* ExportAssignment */: {
                        var expression_1 = node.expression;
                        var child = ts.isObjectLiteralExpression(expression_1) || ts.isCallExpression(expression_1) ? expression_1 :
                            ts.isArrowFunction(expression_1) || ts.isFunctionExpression(expression_1) ? expression_1.body : undefined;
                        if (child) {
                            startNode(node);
                            addChildrenRecursively(child);
                            endNode();
                        }
                        else {
                            addLeafNode(node);
                        }
                        break;
                    }
                    case 273 /* ExportSpecifier */:
                    case 263 /* ImportEqualsDeclaration */:
                    case 174 /* IndexSignature */:
                    case 172 /* CallSignature */:
                    case 173 /* ConstructSignature */:
                    case 257 /* TypeAliasDeclaration */:
                        addLeafNode(node);
                        break;
                    case 206 /* CallExpression */:
                    case 219 /* BinaryExpression */: {
                        var special = ts.getAssignmentDeclarationKind(node);
                        switch (special) {
                            case 1 /* ExportsProperty */:
                            case 2 /* ModuleExports */:
                                addNodeWithRecursiveChild(node, node.right);
                                return;
                            case 6 /* Prototype */:
                            case 3 /* PrototypeProperty */: {
                                var binaryExpression = node;
                                var assignmentTarget = binaryExpression.left;
                                var prototypeAccess = special === 3 /* PrototypeProperty */ ?
                                    assignmentTarget.expression :
                                    assignmentTarget;
                                var depth = 0;
                                var className = void 0;
                                // If we see a prototype assignment, start tracking the target as a class
                                // This is only done for simple classes not nested assignments.
                                if (ts.isIdentifier(prototypeAccess.expression)) {
                                    addTrackedEs5Class(prototypeAccess.expression.text);
                                    className = prototypeAccess.expression;
                                }
                                else {
                                    _a = startNestedNodes(binaryExpression, prototypeAccess.expression), depth = _a[0], className = _a[1];
                                }
                                if (special === 6 /* Prototype */) {
                                    if (ts.isObjectLiteralExpression(binaryExpression.right)) {
                                        if (binaryExpression.right.properties.length > 0) {
                                            startNode(binaryExpression, className);
                                            ts.forEachChild(binaryExpression.right, addChildrenRecursively);
                                            endNode();
                                        }
                                    }
                                }
                                else if (ts.isFunctionExpression(binaryExpression.right) || ts.isArrowFunction(binaryExpression.right)) {
                                    addNodeWithRecursiveChild(node, binaryExpression.right, className);
                                }
                                else {
                                    startNode(binaryExpression, className);
                                    addNodeWithRecursiveChild(node, binaryExpression.right, assignmentTarget.name);
                                    endNode();
                                }
                                endNestedNodes(depth);
                                return;
                            }
                            case 7 /* ObjectDefinePropertyValue */:
                            case 9 /* ObjectDefinePrototypeProperty */: {
                                var defineCall = node;
                                var className = special === 7 /* ObjectDefinePropertyValue */ ?
                                    defineCall.arguments[0] :
                                    defineCall.arguments[0].expression;
                                var memberName = defineCall.arguments[1];
                                var _j = startNestedNodes(node, className), depth = _j[0], classNameIdentifier = _j[1];
                                startNode(node, classNameIdentifier);
                                startNode(node, ts.setTextRange(ts.factory.createIdentifier(memberName.text), memberName));
                                addChildrenRecursively(node.arguments[2]);
                                endNode();
                                endNode();
                                endNestedNodes(depth);
                                return;
                            }
                            case 5 /* Property */: {
                                var binaryExpression = node;
                                var assignmentTarget = binaryExpression.left;
                                var targetFunction = assignmentTarget.expression;
                                if (ts.isIdentifier(targetFunction) && ts.getElementOrPropertyAccessName(assignmentTarget) !== "prototype" &&
                                    trackedEs5Classes && trackedEs5Classes.has(targetFunction.text)) {
                                    if (ts.isFunctionExpression(binaryExpression.right) || ts.isArrowFunction(binaryExpression.right)) {
                                        addNodeWithRecursiveChild(node, binaryExpression.right, targetFunction);
                                    }
                                    else if (ts.isBindableStaticAccessExpression(assignmentTarget)) {
                                        startNode(binaryExpression, targetFunction);
                                        addNodeWithRecursiveChild(binaryExpression.left, binaryExpression.right, ts.getNameOrArgument(assignmentTarget));
                                        endNode();
                                    }
                                    return;
                                }
                                break;
                            }
                            case 4 /* ThisProperty */:
                            case 0 /* None */:
                            case 8 /* ObjectDefinePropertyExports */:
                                break;
                            default:
                                ts.Debug.assertNever(special);
                        }
                    }
                    // falls through
                    default:
                        if (ts.hasJSDocNodes(node)) {
                            ts.forEach(node.jsDoc, function (jsDoc) {
                                ts.forEach(jsDoc.tags, function (tag) {
                                    if (ts.isJSDocTypeAlias(tag)) {
                                        addLeafNode(tag);
                                    }
                                });
                            });
                        }
                        ts.forEachChild(node, addChildrenRecursively);
                }
            }
            /** Merge declarations of the same kind. */
            function mergeChildren(children, node) {
                var nameToItems = new ts.Map();
                ts.filterMutate(children, function (child, index) {
                    var declName = child.name || ts.getNameOfDeclaration(child.node);
                    var name = declName && nodeText(declName);
                    if (!name) {
                        // Anonymous items are never merged.
                        return true;
                    }
                    var itemsWithSameName = nameToItems.get(name);
                    if (!itemsWithSameName) {
                        nameToItems.set(name, child);
                        return true;
                    }
                    if (itemsWithSameName instanceof Array) {
                        for (var _i = 0, itemsWithSameName_1 = itemsWithSameName; _i < itemsWithSameName_1.length; _i++) {
                            var itemWithSameName = itemsWithSameName_1[_i];
                            if (tryMerge(itemWithSameName, child, index, node)) {
                                return false;
                            }
                        }
                        itemsWithSameName.push(child);
                        return true;
                    }
                    else {
                        var itemWithSameName = itemsWithSameName;
                        if (tryMerge(itemWithSameName, child, index, node)) {
                            return false;
                        }
                        nameToItems.set(name, [itemWithSameName, child]);
                        return true;
                    }
                });
            }
            var isEs5ClassMember = (_a = {},
                _a[5 /* Property */] = true,
                _a[3 /* PrototypeProperty */] = true,
                _a[7 /* ObjectDefinePropertyValue */] = true,
                _a[9 /* ObjectDefinePrototypeProperty */] = true,
                _a[0 /* None */] = false,
                _a[1 /* ExportsProperty */] = false,
                _a[2 /* ModuleExports */] = false,
                _a[8 /* ObjectDefinePropertyExports */] = false,
                _a[6 /* Prototype */] = true,
                _a[4 /* ThisProperty */] = false,
                _a);
            function tryMergeEs5Class(a, b, bIndex, parent) {
                function isPossibleConstructor(node) {
                    return ts.isFunctionExpression(node) || ts.isFunctionDeclaration(node) || ts.isVariableDeclaration(node);
                }
                var bAssignmentDeclarationKind = ts.isBinaryExpression(b.node) || ts.isCallExpression(b.node) ?
                    ts.getAssignmentDeclarationKind(b.node) :
                    0 /* None */;
                var aAssignmentDeclarationKind = ts.isBinaryExpression(a.node) || ts.isCallExpression(a.node) ?
                    ts.getAssignmentDeclarationKind(a.node) :
                    0 /* None */;
                // We treat this as an es5 class and merge the nodes in in one of several cases
                if ((isEs5ClassMember[bAssignmentDeclarationKind] && isEs5ClassMember[aAssignmentDeclarationKind]) // merge two class elements
                    || (isPossibleConstructor(a.node) && isEs5ClassMember[bAssignmentDeclarationKind]) // ctor function & member
                    || (isPossibleConstructor(b.node) && isEs5ClassMember[aAssignmentDeclarationKind]) // member & ctor function
                    || (ts.isClassDeclaration(a.node) && isSynthesized(a.node) && isEs5ClassMember[bAssignmentDeclarationKind]) // class (generated) & member
                    || (ts.isClassDeclaration(b.node) && isEs5ClassMember[aAssignmentDeclarationKind]) // member & class (generated)
                    || (ts.isClassDeclaration(a.node) && isSynthesized(a.node) && isPossibleConstructor(b.node)) // class (generated) & ctor
                    || (ts.isClassDeclaration(b.node) && isPossibleConstructor(a.node) && isSynthesized(a.node)) // ctor & class (generated)
                ) {
                    var lastANode = a.additionalNodes && ts.lastOrUndefined(a.additionalNodes) || a.node;
                    if ((!ts.isClassDeclaration(a.node) && !ts.isClassDeclaration(b.node)) // If neither outline node is a class
                        || isPossibleConstructor(a.node) || isPossibleConstructor(b.node) // If either function is a constructor function
                    ) {
                        var ctorFunction = isPossibleConstructor(a.node) ? a.node :
                            isPossibleConstructor(b.node) ? b.node :
                                undefined;
                        if (ctorFunction !== undefined) {
                            var ctorNode = ts.setTextRange(ts.factory.createConstructorDeclaration(/* decorators */ undefined, /* modifiers */ undefined, [], /* body */ undefined), ctorFunction);
                            var ctor = emptyNavigationBarNode(ctorNode);
                            ctor.indent = a.indent + 1;
                            ctor.children = a.node === ctorFunction ? a.children : b.children;
                            a.children = a.node === ctorFunction ? ts.concatenate([ctor], b.children || [b]) : ts.concatenate(a.children || [__assign({}, a)], [ctor]);
                        }
                        else {
                            if (a.children || b.children) {
                                a.children = ts.concatenate(a.children || [__assign({}, a)], b.children || [b]);
                                if (a.children) {
                                    mergeChildren(a.children, a);
                                    sortChildren(a.children);
                                }
                            }
                        }
                        lastANode = a.node = ts.setTextRange(ts.factory.createClassDeclaration(
                        /* decorators */ undefined, 
                        /* modifiers */ undefined, a.name || ts.factory.createIdentifier("__class__"), 
                        /* typeParameters */ undefined, 
                        /* heritageClauses */ undefined, []), a.node);
                    }
                    else {
                        a.children = ts.concatenate(a.children, b.children);
                        if (a.children) {
                            mergeChildren(a.children, a);
                        }
                    }
                    var bNode = b.node;
                    // We merge if the outline node previous to b (bIndex - 1) is already part of the current class
                    // We do this so that statements between class members that do not generate outline nodes do not split up the class outline:
                    // Ex This should produce one outline node C:
                    //    function C() {}; a = 1; C.prototype.m = function () {}
                    // Ex This will produce 3 outline nodes: C, a, C
                    //    function C() {}; let a = 1; C.prototype.m = function () {}
                    if (parent.children[bIndex - 1].node.end === lastANode.end) {
                        ts.setTextRange(lastANode, { pos: lastANode.pos, end: bNode.end });
                    }
                    else {
                        if (!a.additionalNodes)
                            a.additionalNodes = [];
                        a.additionalNodes.push(ts.setTextRange(ts.factory.createClassDeclaration(
                        /* decorators */ undefined, 
                        /* modifiers */ undefined, a.name || ts.factory.createIdentifier("__class__"), 
                        /* typeParameters */ undefined, 
                        /* heritageClauses */ undefined, []), b.node));
                    }
                    return true;
                }
                return bAssignmentDeclarationKind === 0 /* None */ ? false : true;
            }
            function tryMerge(a, b, bIndex, parent) {
                // const v = false as boolean;
                if (tryMergeEs5Class(a, b, bIndex, parent)) {
                    return true;
                }
                if (shouldReallyMerge(a.node, b.node, parent)) {
                    merge(a, b);
                    return true;
                }
                return false;
            }
            /** a and b have the same name, but they may not be mergeable. */
            function shouldReallyMerge(a, b, parent) {
                if (a.kind !== b.kind || a.parent !== b.parent && !(isOwnChild(a, parent) && isOwnChild(b, parent))) {
                    return false;
                }
                switch (a.kind) {
                    case 165 /* PropertyDeclaration */:
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return ts.isStatic(a) === ts.isStatic(b);
                    case 259 /* ModuleDeclaration */:
                        return areSameModule(a, b)
                            && getFullyQualifiedModuleName(a) === getFullyQualifiedModuleName(b);
                    default:
                        return true;
                }
            }
            function isSynthesized(node) {
                return !!(node.flags & 8 /* Synthesized */);
            }
            // We want to merge own children like `I` in in `module A { interface I {} } module A { interface I {} }`
            // We don't want to merge unrelated children like `m` in `const o = { a: { m() {} }, b: { m() {} } };`
            function isOwnChild(n, parent) {
                var par = ts.isModuleBlock(n.parent) ? n.parent.parent : n.parent;
                return par === parent.node || ts.contains(parent.additionalNodes, par);
            }
            // We use 1 NavNode to represent 'A.B.C', but there are multiple source nodes.
            // Only merge module nodes that have the same chain. Don't merge 'A.B.C' with 'A'!
            function areSameModule(a, b) {
                return a.body.kind === b.body.kind && (a.body.kind !== 259 /* ModuleDeclaration */ || areSameModule(a.body, b.body));
            }
            /** Merge source into target. Source should be thrown away after this is called. */
            function merge(target, source) {
                var _a;
                target.additionalNodes = target.additionalNodes || [];
                target.additionalNodes.push(source.node);
                if (source.additionalNodes) {
                    (_a = target.additionalNodes).push.apply(_a, source.additionalNodes);
                }
                target.children = ts.concatenate(target.children, source.children);
                if (target.children) {
                    mergeChildren(target.children, target);
                    sortChildren(target.children);
                }
            }
            /** Recursively ensure that each NavNode's children are in sorted order. */
            function sortChildren(children) {
                children.sort(compareChildren);
            }
            function compareChildren(child1, child2) {
                return ts.compareStringsCaseSensitiveUI(tryGetName(child1.node), tryGetName(child2.node)) // TODO: GH#18217
                    || ts.compareValues(navigationBarNodeKind(child1), navigationBarNodeKind(child2));
            }
            /**
             * This differs from getItemName because this is just used for sorting.
             * We only sort nodes by name that have a more-or-less "direct" name, as opposed to `new()` and the like.
             * So `new()` can still come before an `aardvark` method.
             */
            function tryGetName(node) {
                if (node.kind === 259 /* ModuleDeclaration */) {
                    return getModuleName(node);
                }
                var declName = ts.getNameOfDeclaration(node);
                if (declName && ts.isPropertyName(declName)) {
                    var propertyName = ts.getPropertyNameForPropertyNameNode(declName);
                    return propertyName && ts.unescapeLeadingUnderscores(propertyName);
                }
                switch (node.kind) {
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                    case 224 /* ClassExpression */:
                        return getFunctionOrClassName(node);
                    default:
                        return undefined;
                }
            }
            function getItemName(node, name) {
                if (node.kind === 259 /* ModuleDeclaration */) {
                    return cleanText(getModuleName(node));
                }
                if (name) {
                    var text = ts.isIdentifier(name) ? name.text
                        : ts.isElementAccessExpression(name) ? "[" + nodeText(name.argumentExpression) + "]"
                            : nodeText(name);
                    if (text.length > 0) {
                        return cleanText(text);
                    }
                }
                switch (node.kind) {
                    case 300 /* SourceFile */:
                        var sourceFile = node;
                        return ts.isExternalModule(sourceFile)
                            ? "\"" + ts.escapeString(ts.getBaseFileName(ts.removeFileExtension(ts.normalizePath(sourceFile.fileName)))) + "\""
                            : "<global>";
                    case 269 /* ExportAssignment */:
                        return ts.isExportAssignment(node) && node.isExportEquals ? "export=" /* ExportEquals */ : "default" /* Default */;
                    case 212 /* ArrowFunction */:
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                        if (ts.getSyntacticModifierFlags(node) & 512 /* Default */) {
                            return "default";
                        }
                        // We may get a string with newlines or other whitespace in the case of an object dereference
                        // (eg: "app\n.onactivated"), so we should remove the whitespace for readability in the
                        // navigation bar.
                        return getFunctionOrClassName(node);
                    case 169 /* Constructor */:
                        return "constructor";
                    case 173 /* ConstructSignature */:
                        return "new()";
                    case 172 /* CallSignature */:
                        return "()";
                    case 174 /* IndexSignature */:
                        return "[]";
                    default:
                        return "<unknown>";
                }
            }
            /** Flattens the NavNode tree to a list of items to appear in the primary navbar menu. */
            function primaryNavBarMenuItems(root) {
                // The primary (middle) navbar menu displays the general code navigation hierarchy, similar to the navtree.
                // The secondary (right) navbar menu displays the child items of whichever primary item is selected.
                // Some less interesting items without their own child navigation items (e.g. a local variable declaration) only show up in the secondary menu.
                var primaryNavBarMenuItems = [];
                function recur(item) {
                    if (shouldAppearInPrimaryNavBarMenu(item)) {
                        primaryNavBarMenuItems.push(item);
                        if (item.children) {
                            for (var _i = 0, _a = item.children; _i < _a.length; _i++) {
                                var child = _a[_i];
                                recur(child);
                            }
                        }
                    }
                }
                recur(root);
                return primaryNavBarMenuItems;
                /** Determines if a node should appear in the primary navbar menu. */
                function shouldAppearInPrimaryNavBarMenu(item) {
                    // Items with children should always appear in the primary navbar menu.
                    if (item.children) {
                        return true;
                    }
                    // Some nodes are otherwise important enough to always include in the primary navigation menu.
                    switch (navigationBarNodeKind(item)) {
                        case 255 /* ClassDeclaration */:
                        case 224 /* ClassExpression */:
                        case 258 /* EnumDeclaration */:
                        case 256 /* InterfaceDeclaration */:
                        case 259 /* ModuleDeclaration */:
                        case 300 /* SourceFile */:
                        case 257 /* TypeAliasDeclaration */:
                        case 340 /* JSDocTypedefTag */:
                        case 333 /* JSDocCallbackTag */:
                            return true;
                        case 212 /* ArrowFunction */:
                        case 254 /* FunctionDeclaration */:
                        case 211 /* FunctionExpression */:
                            return isTopLevelFunctionDeclaration(item);
                        default:
                            return false;
                    }
                    function isTopLevelFunctionDeclaration(item) {
                        if (!item.node.body) {
                            return false;
                        }
                        switch (navigationBarNodeKind(item.parent)) {
                            case 260 /* ModuleBlock */:
                            case 300 /* SourceFile */:
                            case 167 /* MethodDeclaration */:
                            case 169 /* Constructor */:
                                return true;
                            default:
                                return false;
                        }
                    }
                }
            }
            function convertToTree(n) {
                return {
                    text: getItemName(n.node, n.name),
                    kind: ts.getNodeKind(n.node),
                    kindModifiers: getModifiers(n.node),
                    spans: getSpans(n),
                    nameSpan: n.name && getNodeSpan(n.name),
                    childItems: ts.map(n.children, convertToTree)
                };
            }
            function convertToPrimaryNavBarMenuItem(n) {
                return {
                    text: getItemName(n.node, n.name),
                    kind: ts.getNodeKind(n.node),
                    kindModifiers: getModifiers(n.node),
                    spans: getSpans(n),
                    childItems: ts.map(n.children, convertToSecondaryNavBarMenuItem) || emptyChildItemArray,
                    indent: n.indent,
                    bolded: false,
                    grayed: false
                };
                function convertToSecondaryNavBarMenuItem(n) {
                    return {
                        text: getItemName(n.node, n.name),
                        kind: ts.getNodeKind(n.node),
                        kindModifiers: ts.getNodeModifiers(n.node),
                        spans: getSpans(n),
                        childItems: emptyChildItemArray,
                        indent: 0,
                        bolded: false,
                        grayed: false
                    };
                }
            }
            function getSpans(n) {
                var spans = [getNodeSpan(n.node)];
                if (n.additionalNodes) {
                    for (var _i = 0, _a = n.additionalNodes; _i < _a.length; _i++) {
                        var node = _a[_i];
                        spans.push(getNodeSpan(node));
                    }
                }
                return spans;
            }
            function getModuleName(moduleDeclaration) {
                // We want to maintain quotation marks.
                if (ts.isAmbientModule(moduleDeclaration)) {
                    return ts.getTextOfNode(moduleDeclaration.name);
                }
                return getFullyQualifiedModuleName(moduleDeclaration);
            }
            function getFullyQualifiedModuleName(moduleDeclaration) {
                // Otherwise, we need to aggregate each identifier to build up the qualified name.
                var result = [ts.getTextOfIdentifierOrLiteral(moduleDeclaration.name)];
                while (moduleDeclaration.body && moduleDeclaration.body.kind === 259 /* ModuleDeclaration */) {
                    moduleDeclaration = moduleDeclaration.body;
                    result.push(ts.getTextOfIdentifierOrLiteral(moduleDeclaration.name));
                }
                return result.join(".");
            }
            /**
             * For 'module A.B.C', we want to get the node for 'C'.
             * We store 'A' as associated with a NavNode, and use getModuleName to traverse down again.
             */
            function getInteriorModule(decl) {
                return decl.body && ts.isModuleDeclaration(decl.body) ? getInteriorModule(decl.body) : decl;
            }
            function isComputedProperty(member) {
                return !member.name || member.name.kind === 160 /* ComputedPropertyName */;
            }
            function getNodeSpan(node) {
                return node.kind === 300 /* SourceFile */ ? ts.createTextSpanFromRange(node) : ts.createTextSpanFromNode(node, curSourceFile);
            }
            function getModifiers(node) {
                if (node.parent && node.parent.kind === 252 /* VariableDeclaration */) {
                    node = node.parent;
                }
                return ts.getNodeModifiers(node);
            }
            function getFunctionOrClassName(node) {
                var parent = node.parent;
                if (node.name && ts.getFullWidth(node.name) > 0) {
                    return cleanText(ts.declarationNameToString(node.name));
                }
                // See if it is a var initializer. If so, use the var name.
                else if (ts.isVariableDeclaration(parent)) {
                    return cleanText(ts.declarationNameToString(parent.name));
                }
                // See if it is of the form "<expr> = function(){...}". If so, use the text from the left-hand side.
                else if (ts.isBinaryExpression(parent) && parent.operatorToken.kind === 63 /* EqualsToken */) {
                    return nodeText(parent.left).replace(whiteSpaceRegex, "");
                }
                // See if it is a property assignment, and if so use the property name
                else if (ts.isPropertyAssignment(parent)) {
                    return nodeText(parent.name);
                }
                // Default exports are named "default"
                else if (ts.getSyntacticModifierFlags(node) & 512 /* Default */) {
                    return "default";
                }
                else if (ts.isClassLike(node)) {
                    return "<class>";
                }
                else if (ts.isCallExpression(parent)) {
                    var name = getCalledExpressionName(parent.expression);
                    if (name !== undefined) {
                        name = cleanText(name);
                        if (name.length > maxLength) {
                            return name + " callback";
                        }
                        var args = cleanText(ts.mapDefined(parent.arguments, function (a) { return ts.isStringLiteralLike(a) ? a.getText(curSourceFile) : undefined; }).join(", "));
                        return name + "(" + args + ") callback";
                    }
                }
                return "<function>";
            }
            // See also 'tryGetPropertyAccessOrIdentifierToString'
            function getCalledExpressionName(expr) {
                if (ts.isIdentifier(expr)) {
                    return expr.text;
                }
                else if (ts.isPropertyAccessExpression(expr)) {
                    var left = getCalledExpressionName(expr.expression);
                    var right = expr.name.text;
                    return left === undefined ? right : left + "." + right;
                }
                else {
                    return undefined;
                }
            }
            function isFunctionOrClassExpression(node) {
                switch (node.kind) {
                    case 212 /* ArrowFunction */:
                    case 211 /* FunctionExpression */:
                    case 224 /* ClassExpression */:
                        return true;
                    default:
                        return false;
                }
            }
            function cleanText(text) {
                // Truncate to maximum amount of characters as we don't want to do a big replace operation.
                text = text.length > maxLength ? text.substring(0, maxLength) + "..." : text;
                // Replaces ECMAScript line terminators and removes the trailing `\` from each line:
                // \n - Line Feed
                // \r - Carriage Return
                // \u2028 - Line separator
                // \u2029 - Paragraph separator
                return text.replace(/\\?(\r?\n|\r|\u2028|\u2029)/g, "");
            }
        })(NavigationBar = ts.NavigationBar || (ts.NavigationBar = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var OrganizeImports;
        (function (OrganizeImports) {
            /**
             * Organize imports by:
             *   1) Removing unused imports
             *   2) Coalescing imports from the same module
             *   3) Sorting imports
             */
            function organizeImports(sourceFile, formatContext, host, program, preferences, skipDestructiveCodeActions) {
                var changeTracker = ts.textChanges.ChangeTracker.fromContext({ host: host, formatContext: formatContext, preferences: preferences });
                var coalesceAndOrganizeImports = function (importGroup) { return ts.stableSort(coalesceImports(removeUnusedImports(importGroup, sourceFile, program, skipDestructiveCodeActions)), function (s1, s2) { return compareImportsOrRequireStatements(s1, s2); }); };
                // All of the old ImportDeclarations in the file, in syntactic order.
                var topLevelImportDecls = sourceFile.statements.filter(ts.isImportDeclaration);
                organizeImportsWorker(topLevelImportDecls, coalesceAndOrganizeImports);
                // All of the old ExportDeclarations in the file, in syntactic order.
                var topLevelExportDecls = sourceFile.statements.filter(ts.isExportDeclaration);
                organizeImportsWorker(topLevelExportDecls, coalesceExports);
                for (var _i = 0, _a = sourceFile.statements.filter(ts.isAmbientModule); _i < _a.length; _i++) {
                    var ambientModule = _a[_i];
                    if (!ambientModule.body) {
                        continue;
                    }
                    var ambientModuleImportDecls = ambientModule.body.statements.filter(ts.isImportDeclaration);
                    organizeImportsWorker(ambientModuleImportDecls, coalesceAndOrganizeImports);
                    var ambientModuleExportDecls = ambientModule.body.statements.filter(ts.isExportDeclaration);
                    organizeImportsWorker(ambientModuleExportDecls, coalesceExports);
                }
                return changeTracker.getChanges();
                function organizeImportsWorker(oldImportDecls, coalesce) {
                    if (ts.length(oldImportDecls) === 0) {
                        return;
                    }
                    // Special case: normally, we'd expect leading and trailing trivia to follow each import
                    // around as it's sorted.  However, we do not want this to happen for leading trivia
                    // on the first import because it is probably the header comment for the file.
                    // Consider: we could do a more careful check that this trivia is actually a header,
                    // but the consequences of being wrong are very minor.
                    ts.suppressLeadingTrivia(oldImportDecls[0]);
                    var oldImportGroups = ts.group(oldImportDecls, function (importDecl) { return getExternalModuleName(importDecl.moduleSpecifier); });
                    var sortedImportGroups = ts.stableSort(oldImportGroups, function (group1, group2) { return compareModuleSpecifiers(group1[0].moduleSpecifier, group2[0].moduleSpecifier); });
                    var newImportDecls = ts.flatMap(sortedImportGroups, function (importGroup) {
                        return getExternalModuleName(importGroup[0].moduleSpecifier)
                            ? coalesce(importGroup)
                            : importGroup;
                    });
                    // Delete all nodes if there are no imports.
                    if (newImportDecls.length === 0) {
                        // Consider the first node to have trailingTrivia as we want to exclude the
                        // "header" comment.
                        changeTracker.deleteNodes(sourceFile, oldImportDecls, {
                            trailingTriviaOption: ts.textChanges.TrailingTriviaOption.Include,
                        }, /*hasTrailingComment*/ true);
                    }
                    else {
                        // Note: Delete the surrounding trivia because it will have been retained in newImportDecls.
                        var replaceOptions = {
                            leadingTriviaOption: ts.textChanges.LeadingTriviaOption.Exclude,
                            trailingTriviaOption: ts.textChanges.TrailingTriviaOption.Include,
                            suffix: ts.getNewLineOrDefaultFromHost(host, formatContext.options),
                        };
                        changeTracker.replaceNodeWithNodes(sourceFile, oldImportDecls[0], newImportDecls, replaceOptions);
                        var hasTrailingComment = changeTracker.nodeHasTrailingComment(sourceFile, oldImportDecls[0], replaceOptions);
                        changeTracker.deleteNodes(sourceFile, oldImportDecls.slice(1), {
                            trailingTriviaOption: ts.textChanges.TrailingTriviaOption.Include,
                        }, hasTrailingComment);
                    }
                }
            }
            OrganizeImports.organizeImports = organizeImports;
            function removeUnusedImports(oldImports, sourceFile, program, skipDestructiveCodeActions) {
                // As a precaution, consider unused import detection to be destructive (GH #43051)
                if (skipDestructiveCodeActions) {
                    return oldImports;
                }
                var typeChecker = program.getTypeChecker();
                var jsxNamespace = typeChecker.getJsxNamespace(sourceFile);
                var jsxFragmentFactory = typeChecker.getJsxFragmentFactory(sourceFile);
                var jsxElementsPresent = !!(sourceFile.transformFlags & 2 /* ContainsJsx */);
                var usedImports = [];
                for (var _i = 0, oldImports_1 = oldImports; _i < oldImports_1.length; _i++) {
                    var importDecl = oldImports_1[_i];
                    var importClause = importDecl.importClause, moduleSpecifier = importDecl.moduleSpecifier;
                    if (!importClause) {
                        // Imports without import clauses are assumed to be included for their side effects and are not removed.
                        usedImports.push(importDecl);
                        continue;
                    }
                    var name = importClause.name, namedBindings = importClause.namedBindings;
                    // Default import
                    if (name && !isDeclarationUsed(name)) {
                        name = undefined;
                    }
                    if (namedBindings) {
                        if (ts.isNamespaceImport(namedBindings)) {
                            // Namespace import
                            if (!isDeclarationUsed(namedBindings.name)) {
                                namedBindings = undefined;
                            }
                        }
                        else {
                            // List of named imports
                            var newElements = namedBindings.elements.filter(function (e) { return isDeclarationUsed(e.name); });
                            if (newElements.length < namedBindings.elements.length) {
                                namedBindings = newElements.length
                                    ? ts.factory.updateNamedImports(namedBindings, newElements)
                                    : undefined;
                            }
                        }
                    }
                    if (name || namedBindings) {
                        usedImports.push(updateImportDeclarationAndClause(importDecl, name, namedBindings));
                    }
                    // If a module is imported to be augmented, it’s used
                    else if (hasModuleDeclarationMatchingSpecifier(sourceFile, moduleSpecifier)) {
                        // If we’re in a declaration file, it’s safe to remove the import clause from it
                        if (sourceFile.isDeclarationFile) {
                            usedImports.push(ts.factory.createImportDeclaration(importDecl.decorators, importDecl.modifiers, 
                            /*importClause*/ undefined, moduleSpecifier));
                        }
                        // If we’re not in a declaration file, we can’t remove the import clause even though
                        // the imported symbols are unused, because removing them makes it look like the import
                        // declaration has side effects, which will cause it to be preserved in the JS emit.
                        else {
                            usedImports.push(importDecl);
                        }
                    }
                }
                return usedImports;
                function isDeclarationUsed(identifier) {
                    // The JSX factory symbol is always used if JSX elements are present - even if they are not allowed.
                    return jsxElementsPresent && (identifier.text === jsxNamespace || jsxFragmentFactory && identifier.text === jsxFragmentFactory) ||
                        ts.FindAllReferences.Core.isSymbolReferencedInFile(identifier, typeChecker, sourceFile);
                }
            }
            function hasModuleDeclarationMatchingSpecifier(sourceFile, moduleSpecifier) {
                var moduleSpecifierText = ts.isStringLiteral(moduleSpecifier) && moduleSpecifier.text;
                return ts.isString(moduleSpecifierText) && ts.some(sourceFile.moduleAugmentations, function (moduleName) {
                    return ts.isStringLiteral(moduleName)
                        && moduleName.text === moduleSpecifierText;
                });
            }
            function getExternalModuleName(specifier) {
                return specifier !== undefined && ts.isStringLiteralLike(specifier)
                    ? specifier.text
                    : undefined;
            }
            // Internal for testing
            /**
             * @param importGroup a list of ImportDeclarations, all with the same module name.
             */
            function coalesceImports(importGroup) {
                var _a;
                if (importGroup.length === 0) {
                    return importGroup;
                }
                var _b = getCategorizedImports(importGroup), importWithoutClause = _b.importWithoutClause, typeOnlyImports = _b.typeOnlyImports, regularImports = _b.regularImports;
                var coalescedImports = [];
                if (importWithoutClause) {
                    coalescedImports.push(importWithoutClause);
                }
                for (var _i = 0, _c = [regularImports, typeOnlyImports]; _i < _c.length; _i++) {
                    var group_2 = _c[_i];
                    var isTypeOnly = group_2 === typeOnlyImports;
                    var defaultImports = group_2.defaultImports, namespaceImports = group_2.namespaceImports, namedImports = group_2.namedImports;
                    // Normally, we don't combine default and namespace imports, but it would be silly to
                    // produce two import declarations in this special case.
                    if (!isTypeOnly && defaultImports.length === 1 && namespaceImports.length === 1 && namedImports.length === 0) {
                        // Add the namespace import to the existing default ImportDeclaration.
                        var defaultImport = defaultImports[0];
                        coalescedImports.push(updateImportDeclarationAndClause(defaultImport, defaultImport.importClause.name, namespaceImports[0].importClause.namedBindings)); // TODO: GH#18217
                        continue;
                    }
                    var sortedNamespaceImports = ts.stableSort(namespaceImports, function (i1, i2) {
                        return compareIdentifiers(i1.importClause.namedBindings.name, i2.importClause.namedBindings.name);
                    }); // TODO: GH#18217
                    for (var _d = 0, sortedNamespaceImports_1 = sortedNamespaceImports; _d < sortedNamespaceImports_1.length; _d++) {
                        var namespaceImport = sortedNamespaceImports_1[_d];
                        // Drop the name, if any
                        coalescedImports.push(updateImportDeclarationAndClause(namespaceImport, /*name*/ undefined, namespaceImport.importClause.namedBindings)); // TODO: GH#18217
                    }
                    if (defaultImports.length === 0 && namedImports.length === 0) {
                        continue;
                    }
                    var newDefaultImport = void 0;
                    var newImportSpecifiers = [];
                    if (defaultImports.length === 1) {
                        newDefaultImport = defaultImports[0].importClause.name;
                    }
                    else {
                        for (var _e = 0, defaultImports_1 = defaultImports; _e < defaultImports_1.length; _e++) {
                            var defaultImport = defaultImports_1[_e];
                            newImportSpecifiers.push(ts.factory.createImportSpecifier(ts.factory.createIdentifier("default"), defaultImport.importClause.name)); // TODO: GH#18217
                        }
                    }
                    newImportSpecifiers.push.apply(newImportSpecifiers, ts.flatMap(namedImports, function (i) { return i.importClause.namedBindings.elements; })); // TODO: GH#18217
                    var sortedImportSpecifiers = sortSpecifiers(newImportSpecifiers);
                    var importDecl = defaultImports.length > 0
                        ? defaultImports[0]
                        : namedImports[0];
                    var newNamedImports = sortedImportSpecifiers.length === 0
                        ? newDefaultImport
                            ? undefined
                            : ts.factory.createNamedImports(ts.emptyArray)
                        : namedImports.length === 0
                            ? ts.factory.createNamedImports(sortedImportSpecifiers)
                            : ts.factory.updateNamedImports(namedImports[0].importClause.namedBindings, sortedImportSpecifiers); // TODO: GH#18217
                    // Type-only imports are not allowed to mix default, namespace, and named imports in any combination.
                    // We could rewrite a default import as a named import (`import { default as name }`), but we currently
                    // choose not to as a stylistic preference.
                    if (isTypeOnly && newDefaultImport && newNamedImports) {
                        coalescedImports.push(updateImportDeclarationAndClause(importDecl, newDefaultImport, /*namedBindings*/ undefined));
                        coalescedImports.push(updateImportDeclarationAndClause((_a = namedImports[0]) !== null && _a !== void 0 ? _a : importDecl, /*name*/ undefined, newNamedImports));
                    }
                    else {
                        coalescedImports.push(updateImportDeclarationAndClause(importDecl, newDefaultImport, newNamedImports));
                    }
                }
                return coalescedImports;
            }
            OrganizeImports.coalesceImports = coalesceImports;
            /*
             * Returns entire import declarations because they may already have been rewritten and
             * may lack parent pointers.  The desired parts can easily be recovered based on the
             * categorization.
             *
             * NB: There may be overlap between `defaultImports` and `namespaceImports`/`namedImports`.
             */
            function getCategorizedImports(importGroup) {
                var importWithoutClause;
                var typeOnlyImports = { defaultImports: [], namespaceImports: [], namedImports: [] };
                var regularImports = { defaultImports: [], namespaceImports: [], namedImports: [] };
                for (var _i = 0, importGroup_1 = importGroup; _i < importGroup_1.length; _i++) {
                    var importDeclaration = importGroup_1[_i];
                    if (importDeclaration.importClause === undefined) {
                        // Only the first such import is interesting - the others are redundant.
                        // Note: Unfortunately, we will lose trivia that was on this node.
                        importWithoutClause = importWithoutClause || importDeclaration;
                        continue;
                    }
                    var group_3 = importDeclaration.importClause.isTypeOnly ? typeOnlyImports : regularImports;
                    var _a = importDeclaration.importClause, name = _a.name, namedBindings = _a.namedBindings;
                    if (name) {
                        group_3.defaultImports.push(importDeclaration);
                    }
                    if (namedBindings) {
                        if (ts.isNamespaceImport(namedBindings)) {
                            group_3.namespaceImports.push(importDeclaration);
                        }
                        else {
                            group_3.namedImports.push(importDeclaration);
                        }
                    }
                }
                return {
                    importWithoutClause: importWithoutClause,
                    typeOnlyImports: typeOnlyImports,
                    regularImports: regularImports,
                };
            }
            // Internal for testing
            /**
             * @param exportGroup a list of ExportDeclarations, all with the same module name.
             */
            function coalesceExports(exportGroup) {
                if (exportGroup.length === 0) {
                    return exportGroup;
                }
                var _a = getCategorizedExports(exportGroup), exportWithoutClause = _a.exportWithoutClause, namedExports = _a.namedExports, typeOnlyExports = _a.typeOnlyExports;
                var coalescedExports = [];
                if (exportWithoutClause) {
                    coalescedExports.push(exportWithoutClause);
                }
                for (var _i = 0, _b = [namedExports, typeOnlyExports]; _i < _b.length; _i++) {
                    var exportGroup_1 = _b[_i];
                    if (exportGroup_1.length === 0) {
                        continue;
                    }
                    var newExportSpecifiers = [];
                    newExportSpecifiers.push.apply(newExportSpecifiers, ts.flatMap(exportGroup_1, function (i) { return i.exportClause && ts.isNamedExports(i.exportClause) ? i.exportClause.elements : ts.emptyArray; }));
                    var sortedExportSpecifiers = sortSpecifiers(newExportSpecifiers);
                    var exportDecl = exportGroup_1[0];
                    coalescedExports.push(ts.factory.updateExportDeclaration(exportDecl, exportDecl.decorators, exportDecl.modifiers, exportDecl.isTypeOnly, exportDecl.exportClause && (ts.isNamedExports(exportDecl.exportClause) ?
                        ts.factory.updateNamedExports(exportDecl.exportClause, sortedExportSpecifiers) :
                        ts.factory.updateNamespaceExport(exportDecl.exportClause, exportDecl.exportClause.name)), exportDecl.moduleSpecifier));
                }
                return coalescedExports;
                /*
                 * Returns entire export declarations because they may already have been rewritten and
                 * may lack parent pointers.  The desired parts can easily be recovered based on the
                 * categorization.
                 */
                function getCategorizedExports(exportGroup) {
                    var exportWithoutClause;
                    var namedExports = [];
                    var typeOnlyExports = [];
                    for (var _i = 0, exportGroup_2 = exportGroup; _i < exportGroup_2.length; _i++) {
                        var exportDeclaration = exportGroup_2[_i];
                        if (exportDeclaration.exportClause === undefined) {
                            // Only the first such export is interesting - the others are redundant.
                            // Note: Unfortunately, we will lose trivia that was on this node.
                            exportWithoutClause = exportWithoutClause || exportDeclaration;
                        }
                        else if (exportDeclaration.isTypeOnly) {
                            typeOnlyExports.push(exportDeclaration);
                        }
                        else {
                            namedExports.push(exportDeclaration);
                        }
                    }
                    return {
                        exportWithoutClause: exportWithoutClause,
                        namedExports: namedExports,
                        typeOnlyExports: typeOnlyExports,
                    };
                }
            }
            OrganizeImports.coalesceExports = coalesceExports;
            function updateImportDeclarationAndClause(importDeclaration, name, namedBindings) {
                return ts.factory.updateImportDeclaration(importDeclaration, importDeclaration.decorators, importDeclaration.modifiers, ts.factory.updateImportClause(importDeclaration.importClause, importDeclaration.importClause.isTypeOnly, name, namedBindings), // TODO: GH#18217
                importDeclaration.moduleSpecifier);
            }
            function sortSpecifiers(specifiers) {
                return ts.stableSort(specifiers, compareImportOrExportSpecifiers);
            }
            function compareImportOrExportSpecifiers(s1, s2) {
                return compareIdentifiers(s1.propertyName || s1.name, s2.propertyName || s2.name)
                    || compareIdentifiers(s1.name, s2.name);
            }
            OrganizeImports.compareImportOrExportSpecifiers = compareImportOrExportSpecifiers;
            /* internal */ // Exported for testing
            function compareModuleSpecifiers(m1, m2) {
                var name1 = m1 === undefined ? undefined : getExternalModuleName(m1);
                var name2 = m2 === undefined ? undefined : getExternalModuleName(m2);
                return ts.compareBooleans(name1 === undefined, name2 === undefined) ||
                    ts.compareBooleans(ts.isExternalModuleNameRelative(name1), ts.isExternalModuleNameRelative(name2)) ||
                    ts.compareStringsCaseInsensitive(name1, name2);
            }
            OrganizeImports.compareModuleSpecifiers = compareModuleSpecifiers;
            function compareIdentifiers(s1, s2) {
                return ts.compareStringsCaseInsensitive(s1.text, s2.text);
            }
            function getModuleSpecifierExpression(declaration) {
                var _a;
                switch (declaration.kind) {
                    case 263 /* ImportEqualsDeclaration */:
                        return (_a = ts.tryCast(declaration.moduleReference, ts.isExternalModuleReference)) === null || _a === void 0 ? void 0 : _a.expression;
                    case 264 /* ImportDeclaration */:
                        return declaration.moduleSpecifier;
                    case 235 /* VariableStatement */:
                        return declaration.declarationList.declarations[0].initializer.arguments[0];
                }
            }
            function importsAreSorted(imports) {
                return ts.arrayIsSorted(imports, compareImportsOrRequireStatements);
            }
            OrganizeImports.importsAreSorted = importsAreSorted;
            function importSpecifiersAreSorted(imports) {
                return ts.arrayIsSorted(imports, compareImportOrExportSpecifiers);
            }
            OrganizeImports.importSpecifiersAreSorted = importSpecifiersAreSorted;
            function getImportDeclarationInsertionIndex(sortedImports, newImport) {
                var index = ts.binarySearch(sortedImports, newImport, ts.identity, compareImportsOrRequireStatements);
                return index < 0 ? ~index : index;
            }
            OrganizeImports.getImportDeclarationInsertionIndex = getImportDeclarationInsertionIndex;
            function getImportSpecifierInsertionIndex(sortedImports, newImport) {
                var index = ts.binarySearch(sortedImports, newImport, ts.identity, compareImportOrExportSpecifiers);
                return index < 0 ? ~index : index;
            }
            OrganizeImports.getImportSpecifierInsertionIndex = getImportSpecifierInsertionIndex;
            function compareImportsOrRequireStatements(s1, s2) {
                return compareModuleSpecifiers(getModuleSpecifierExpression(s1), getModuleSpecifierExpression(s2)) || compareImportKind(s1, s2);
            }
            OrganizeImports.compareImportsOrRequireStatements = compareImportsOrRequireStatements;
            function compareImportKind(s1, s2) {
                return ts.compareValues(getImportKindOrder(s1), getImportKindOrder(s2));
            }
            // 1. Side-effect imports
            // 2. Type-only imports
            // 3. Namespace imports
            // 4. Default imports
            // 5. Named imports
            // 6. ImportEqualsDeclarations
            // 7. Require variable statements
            function getImportKindOrder(s1) {
                var _a;
                switch (s1.kind) {
                    case 264 /* ImportDeclaration */:
                        if (!s1.importClause)
                            return 0;
                        if (s1.importClause.isTypeOnly)
                            return 1;
                        if (((_a = s1.importClause.namedBindings) === null || _a === void 0 ? void 0 : _a.kind) === 266 /* NamespaceImport */)
                            return 2;
                        if (s1.importClause.name)
                            return 3;
                        return 4;
                    case 263 /* ImportEqualsDeclaration */:
                        return 5;
                    case 235 /* VariableStatement */:
                        return 6;
                }
            }
        })(OrganizeImports = ts.OrganizeImports || (ts.OrganizeImports = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var OutliningElementsCollector;
        (function (OutliningElementsCollector) {
            function collectElements(sourceFile, cancellationToken) {
                var res = [];
                addNodeOutliningSpans(sourceFile, cancellationToken, res);
                addRegionOutliningSpans(sourceFile, res);
                return res.sort(function (span1, span2) { return span1.textSpan.start - span2.textSpan.start; });
            }
            OutliningElementsCollector.collectElements = collectElements;
            function addNodeOutliningSpans(sourceFile, cancellationToken, out) {
                var depthRemaining = 40;
                var current = 0;
                // Includes the EOF Token so that comments which aren't attached to statements are included
                var statements = __spreadArray(__spreadArray([], sourceFile.statements, true), [sourceFile.endOfFileToken], false);
                var n = statements.length;
                while (current < n) {
                    while (current < n && !ts.isAnyImportSyntax(statements[current])) {
                        visitNonImportNode(statements[current]);
                        current++;
                    }
                    if (current === n)
                        break;
                    var firstImport = current;
                    while (current < n && ts.isAnyImportSyntax(statements[current])) {
                        addOutliningForLeadingCommentsForNode(statements[current], sourceFile, cancellationToken, out);
                        current++;
                    }
                    var lastImport = current - 1;
                    if (lastImport !== firstImport) {
                        out.push(createOutliningSpanFromBounds(ts.findChildOfKind(statements[firstImport], 100 /* ImportKeyword */, sourceFile).getStart(sourceFile), statements[lastImport].getEnd(), "imports" /* Imports */));
                    }
                }
                function visitNonImportNode(n) {
                    var _a;
                    if (depthRemaining === 0)
                        return;
                    cancellationToken.throwIfCancellationRequested();
                    if (ts.isDeclaration(n) || ts.isVariableStatement(n) || ts.isReturnStatement(n) || n.kind === 1 /* EndOfFileToken */) {
                        addOutliningForLeadingCommentsForNode(n, sourceFile, cancellationToken, out);
                    }
                    if (ts.isFunctionLike(n) && ts.isBinaryExpression(n.parent) && ts.isPropertyAccessExpression(n.parent.left)) {
                        addOutliningForLeadingCommentsForNode(n.parent.left, sourceFile, cancellationToken, out);
                    }
                    if (ts.isBlock(n) || ts.isModuleBlock(n)) {
                        addOutliningForLeadingCommentsForPos(n.statements.end, sourceFile, cancellationToken, out);
                    }
                    if (ts.isClassLike(n) || ts.isInterfaceDeclaration(n)) {
                        addOutliningForLeadingCommentsForPos(n.members.end, sourceFile, cancellationToken, out);
                    }
                    var span = getOutliningSpanForNode(n, sourceFile);
                    if (span)
                        out.push(span);
                    depthRemaining--;
                    if (ts.isCallExpression(n)) {
                        depthRemaining++;
                        visitNonImportNode(n.expression);
                        depthRemaining--;
                        n.arguments.forEach(visitNonImportNode);
                        (_a = n.typeArguments) === null || _a === void 0 ? void 0 : _a.forEach(visitNonImportNode);
                    }
                    else if (ts.isIfStatement(n) && n.elseStatement && ts.isIfStatement(n.elseStatement)) {
                        // Consider an 'else if' to be on the same depth as the 'if'.
                        visitNonImportNode(n.expression);
                        visitNonImportNode(n.thenStatement);
                        depthRemaining++;
                        visitNonImportNode(n.elseStatement);
                        depthRemaining--;
                    }
                    else {
                        n.forEachChild(visitNonImportNode);
                    }
                    depthRemaining++;
                }
            }
            function addRegionOutliningSpans(sourceFile, out) {
                var regions = [];
                var lineStarts = sourceFile.getLineStarts();
                for (var _i = 0, lineStarts_1 = lineStarts; _i < lineStarts_1.length; _i++) {
                    var currentLineStart = lineStarts_1[_i];
                    var lineEnd = sourceFile.getLineEndOfPosition(currentLineStart);
                    var lineText = sourceFile.text.substring(currentLineStart, lineEnd);
                    var result = isRegionDelimiter(lineText);
                    if (!result || ts.isInComment(sourceFile, currentLineStart)) {
                        continue;
                    }
                    if (!result[1]) {
                        var span = ts.createTextSpanFromBounds(sourceFile.text.indexOf("//", currentLineStart), lineEnd);
                        regions.push(createOutliningSpan(span, "region" /* Region */, span, /*autoCollapse*/ false, result[2] || "#region"));
                    }
                    else {
                        var region = regions.pop();
                        if (region) {
                            region.textSpan.length = lineEnd - region.textSpan.start;
                            region.hintSpan.length = lineEnd - region.textSpan.start;
                            out.push(region);
                        }
                    }
                }
            }
            var regionDelimiterRegExp = /^#(end)?region(?:\s+(.*))?(?:\r)?$/;
            function isRegionDelimiter(lineText) {
                // We trim the leading whitespace and // without the regex since the
                // multiple potential whitespace matches can make for some gnarly backtracking behavior
                lineText = ts.trimStringStart(lineText);
                if (!ts.startsWith(lineText, "\/\/")) {
                    return null; // eslint-disable-line no-null/no-null
                }
                lineText = ts.trimString(lineText.slice(2));
                return regionDelimiterRegExp.exec(lineText);
            }
            function addOutliningForLeadingCommentsForPos(pos, sourceFile, cancellationToken, out) {
                var comments = ts.getLeadingCommentRanges(sourceFile.text, pos);
                if (!comments)
                    return;
                var firstSingleLineCommentStart = -1;
                var lastSingleLineCommentEnd = -1;
                var singleLineCommentCount = 0;
                var sourceText = sourceFile.getFullText();
                for (var _i = 0, comments_1 = comments; _i < comments_1.length; _i++) {
                    var _a = comments_1[_i], kind = _a.kind, pos_1 = _a.pos, end = _a.end;
                    cancellationToken.throwIfCancellationRequested();
                    switch (kind) {
                        case 2 /* SingleLineCommentTrivia */:
                            // never fold region delimiters into single-line comment regions
                            var commentText = sourceText.slice(pos_1, end);
                            if (isRegionDelimiter(commentText)) {
                                combineAndAddMultipleSingleLineComments();
                                singleLineCommentCount = 0;
                                break;
                            }
                            // For single line comments, combine consecutive ones (2 or more) into
                            // a single span from the start of the first till the end of the last
                            if (singleLineCommentCount === 0) {
                                firstSingleLineCommentStart = pos_1;
                            }
                            lastSingleLineCommentEnd = end;
                            singleLineCommentCount++;
                            break;
                        case 3 /* MultiLineCommentTrivia */:
                            combineAndAddMultipleSingleLineComments();
                            out.push(createOutliningSpanFromBounds(pos_1, end, "comment" /* Comment */));
                            singleLineCommentCount = 0;
                            break;
                        default:
                            ts.Debug.assertNever(kind);
                    }
                }
                combineAndAddMultipleSingleLineComments();
                function combineAndAddMultipleSingleLineComments() {
                    // Only outline spans of two or more consecutive single line comments
                    if (singleLineCommentCount > 1) {
                        out.push(createOutliningSpanFromBounds(firstSingleLineCommentStart, lastSingleLineCommentEnd, "comment" /* Comment */));
                    }
                }
            }
            function addOutliningForLeadingCommentsForNode(n, sourceFile, cancellationToken, out) {
                if (ts.isJsxText(n))
                    return;
                addOutliningForLeadingCommentsForPos(n.pos, sourceFile, cancellationToken, out);
            }
            function createOutliningSpanFromBounds(pos, end, kind) {
                return createOutliningSpan(ts.createTextSpanFromBounds(pos, end), kind);
            }
            function getOutliningSpanForNode(n, sourceFile) {
                switch (n.kind) {
                    case 233 /* Block */:
                        if (ts.isFunctionLike(n.parent)) {
                            return functionSpan(n.parent, n, sourceFile);
                        }
                        // Check if the block is standalone, or 'attached' to some parent statement.
                        // If the latter, we want to collapse the block, but consider its hint span
                        // to be the entire span of the parent.
                        switch (n.parent.kind) {
                            case 238 /* DoStatement */:
                            case 241 /* ForInStatement */:
                            case 242 /* ForOfStatement */:
                            case 240 /* ForStatement */:
                            case 237 /* IfStatement */:
                            case 239 /* WhileStatement */:
                            case 246 /* WithStatement */:
                            case 290 /* CatchClause */:
                                return spanForNode(n.parent);
                            case 250 /* TryStatement */:
                                // Could be the try-block, or the finally-block.
                                var tryStatement = n.parent;
                                if (tryStatement.tryBlock === n) {
                                    return spanForNode(n.parent);
                                }
                                else if (tryStatement.finallyBlock === n) {
                                    var node = ts.findChildOfKind(tryStatement, 96 /* FinallyKeyword */, sourceFile);
                                    if (node)
                                        return spanForNode(node);
                                }
                            // falls through
                            default:
                                // Block was a standalone block.  In this case we want to only collapse
                                // the span of the block, independent of any parent span.
                                return createOutliningSpan(ts.createTextSpanFromNode(n, sourceFile), "code" /* Code */);
                        }
                    case 260 /* ModuleBlock */:
                        return spanForNode(n.parent);
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                    case 256 /* InterfaceDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 261 /* CaseBlock */:
                    case 180 /* TypeLiteral */:
                    case 199 /* ObjectBindingPattern */:
                        return spanForNode(n);
                    case 182 /* TupleType */:
                        return spanForNode(n, /*autoCollapse*/ false, /*useFullStart*/ !ts.isTupleTypeNode(n.parent), 22 /* OpenBracketToken */);
                    case 287 /* CaseClause */:
                    case 288 /* DefaultClause */:
                        return spanForNodeArray(n.statements);
                    case 203 /* ObjectLiteralExpression */:
                        return spanForObjectOrArrayLiteral(n);
                    case 202 /* ArrayLiteralExpression */:
                        return spanForObjectOrArrayLiteral(n, 22 /* OpenBracketToken */);
                    case 276 /* JsxElement */:
                        return spanForJSXElement(n);
                    case 280 /* JsxFragment */:
                        return spanForJSXFragment(n);
                    case 277 /* JsxSelfClosingElement */:
                    case 278 /* JsxOpeningElement */:
                        return spanForJSXAttributes(n.attributes);
                    case 221 /* TemplateExpression */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                        return spanForTemplateLiteral(n);
                    case 200 /* ArrayBindingPattern */:
                        return spanForNode(n, /*autoCollapse*/ false, /*useFullStart*/ !ts.isBindingElement(n.parent), 22 /* OpenBracketToken */);
                    case 212 /* ArrowFunction */:
                        return spanForArrowFunction(n);
                    case 206 /* CallExpression */:
                        return spanForCallExpression(n);
                }
                function spanForCallExpression(node) {
                    if (!node.arguments.length) {
                        return undefined;
                    }
                    var openToken = ts.findChildOfKind(node, 20 /* OpenParenToken */, sourceFile);
                    var closeToken = ts.findChildOfKind(node, 21 /* CloseParenToken */, sourceFile);
                    if (!openToken || !closeToken || ts.positionsAreOnSameLine(openToken.pos, closeToken.pos, sourceFile)) {
                        return undefined;
                    }
                    return spanBetweenTokens(openToken, closeToken, node, sourceFile, /*autoCollapse*/ false, /*useFullStart*/ true);
                }
                function spanForArrowFunction(node) {
                    if (ts.isBlock(node.body) || ts.positionsAreOnSameLine(node.body.getFullStart(), node.body.getEnd(), sourceFile)) {
                        return undefined;
                    }
                    var textSpan = ts.createTextSpanFromBounds(node.body.getFullStart(), node.body.getEnd());
                    return createOutliningSpan(textSpan, "code" /* Code */, ts.createTextSpanFromNode(node));
                }
                function spanForJSXElement(node) {
                    var textSpan = ts.createTextSpanFromBounds(node.openingElement.getStart(sourceFile), node.closingElement.getEnd());
                    var tagName = node.openingElement.tagName.getText(sourceFile);
                    var bannerText = "<" + tagName + ">...</" + tagName + ">";
                    return createOutliningSpan(textSpan, "code" /* Code */, textSpan, /*autoCollapse*/ false, bannerText);
                }
                function spanForJSXFragment(node) {
                    var textSpan = ts.createTextSpanFromBounds(node.openingFragment.getStart(sourceFile), node.closingFragment.getEnd());
                    var bannerText = "<>...</>";
                    return createOutliningSpan(textSpan, "code" /* Code */, textSpan, /*autoCollapse*/ false, bannerText);
                }
                function spanForJSXAttributes(node) {
                    if (node.properties.length === 0) {
                        return undefined;
                    }
                    return createOutliningSpanFromBounds(node.getStart(sourceFile), node.getEnd(), "code" /* Code */);
                }
                function spanForTemplateLiteral(node) {
                    if (node.kind === 14 /* NoSubstitutionTemplateLiteral */ && node.text.length === 0) {
                        return undefined;
                    }
                    return createOutliningSpanFromBounds(node.getStart(sourceFile), node.getEnd(), "code" /* Code */);
                }
                function spanForObjectOrArrayLiteral(node, open) {
                    if (open === void 0) {
                        open = 18 /* OpenBraceToken */;
                    }
                    // If the block has no leading keywords and is inside an array literal or call expression,
                    // we only want to collapse the span of the block.
                    // Otherwise, the collapsed section will include the end of the previous line.
                    return spanForNode(node, /*autoCollapse*/ false, /*useFullStart*/ !ts.isArrayLiteralExpression(node.parent) && !ts.isCallExpression(node.parent), open);
                }
                function spanForNode(hintSpanNode, autoCollapse, useFullStart, open, close) {
                    if (autoCollapse === void 0) {
                        autoCollapse = false;
                    }
                    if (useFullStart === void 0) {
                        useFullStart = true;
                    }
                    if (open === void 0) {
                        open = 18 /* OpenBraceToken */;
                    }
                    if (close === void 0) {
                        close = open === 18 /* OpenBraceToken */ ? 19 /* CloseBraceToken */ : 23 /* CloseBracketToken */;
                    }
                    var openToken = ts.findChildOfKind(n, open, sourceFile);
                    var closeToken = ts.findChildOfKind(n, close, sourceFile);
                    return openToken && closeToken && spanBetweenTokens(openToken, closeToken, hintSpanNode, sourceFile, autoCollapse, useFullStart);
                }
                function spanForNodeArray(nodeArray) {
                    return nodeArray.length ? createOutliningSpan(ts.createTextSpanFromRange(nodeArray), "code" /* Code */) : undefined;
                }
            }
            function functionSpan(node, body, sourceFile) {
                var openToken = tryGetFunctionOpenToken(node, body, sourceFile);
                var closeToken = ts.findChildOfKind(body, 19 /* CloseBraceToken */, sourceFile);
                return openToken && closeToken && spanBetweenTokens(openToken, closeToken, node, sourceFile, /*autoCollapse*/ node.kind !== 212 /* ArrowFunction */);
            }
            function spanBetweenTokens(openToken, closeToken, hintSpanNode, sourceFile, autoCollapse, useFullStart) {
                if (autoCollapse === void 0) {
                    autoCollapse = false;
                }
                if (useFullStart === void 0) {
                    useFullStart = true;
                }
                var textSpan = ts.createTextSpanFromBounds(useFullStart ? openToken.getFullStart() : openToken.getStart(sourceFile), closeToken.getEnd());
                return createOutliningSpan(textSpan, "code" /* Code */, ts.createTextSpanFromNode(hintSpanNode, sourceFile), autoCollapse);
            }
            function createOutliningSpan(textSpan, kind, hintSpan, autoCollapse, bannerText) {
                if (hintSpan === void 0) {
                    hintSpan = textSpan;
                }
                if (autoCollapse === void 0) {
                    autoCollapse = false;
                }
                if (bannerText === void 0) {
                    bannerText = "...";
                }
                return { textSpan: textSpan, kind: kind, hintSpan: hintSpan, bannerText: bannerText, autoCollapse: autoCollapse };
            }
            function tryGetFunctionOpenToken(node, body, sourceFile) {
                if (ts.isNodeArrayMultiLine(node.parameters, sourceFile)) {
                    var openParenToken = ts.findChildOfKind(node, 20 /* OpenParenToken */, sourceFile);
                    if (openParenToken) {
                        return openParenToken;
                    }
                }
                return ts.findChildOfKind(body, 18 /* OpenBraceToken */, sourceFile);
            }
        })(OutliningElementsCollector = ts.OutliningElementsCollector || (ts.OutliningElementsCollector = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        // Note(cyrusn): this enum is ordered from strongest match type to weakest match type.
        var PatternMatchKind;
        (function (PatternMatchKind) {
            PatternMatchKind[PatternMatchKind["exact"] = 0] = "exact";
            PatternMatchKind[PatternMatchKind["prefix"] = 1] = "prefix";
            PatternMatchKind[PatternMatchKind["substring"] = 2] = "substring";
            PatternMatchKind[PatternMatchKind["camelCase"] = 3] = "camelCase";
        })(PatternMatchKind = ts.PatternMatchKind || (ts.PatternMatchKind = {}));
        function createPatternMatch(kind, isCaseSensitive) {
            return {
                kind: kind,
                isCaseSensitive: isCaseSensitive
            };
        }
        function createPatternMatcher(pattern) {
            // We'll often see the same candidate string many times when searching (For example, when
            // we see the name of a module that is used everywhere, or the name of an overload).  As
            // such, we cache the information we compute about the candidate for the life of this
            // pattern matcher so we don't have to compute it multiple times.
            var stringToWordSpans = new ts.Map();
            var dotSeparatedSegments = pattern.trim().split(".").map(function (p) { return createSegment(p.trim()); });
            // A segment is considered invalid if we couldn't find any words in it.
            if (dotSeparatedSegments.some(function (segment) { return !segment.subWordTextChunks.length; }))
                return undefined;
            return {
                getFullMatch: function (containers, candidate) { return getFullMatch(containers, candidate, dotSeparatedSegments, stringToWordSpans); },
                getMatchForLastSegmentOfPattern: function (candidate) { return matchSegment(candidate, ts.last(dotSeparatedSegments), stringToWordSpans); },
                patternContainsDots: dotSeparatedSegments.length > 1
            };
        }
        ts.createPatternMatcher = createPatternMatcher;
        function getFullMatch(candidateContainers, candidate, dotSeparatedSegments, stringToWordSpans) {
            // First, check that the last part of the dot separated pattern matches the name of the
            // candidate.  If not, then there's no point in proceeding and doing the more
            // expensive work.
            var candidateMatch = matchSegment(candidate, ts.last(dotSeparatedSegments), stringToWordSpans);
            if (!candidateMatch) {
                return undefined;
            }
            // -1 because the last part was checked against the name, and only the rest
            // of the parts are checked against the container.
            if (dotSeparatedSegments.length - 1 > candidateContainers.length) {
                // There weren't enough container parts to match against the pattern parts.
                // So this definitely doesn't match.
                return undefined;
            }
            var bestMatch;
            for (var i = dotSeparatedSegments.length - 2, j = candidateContainers.length - 1; i >= 0; i -= 1, j -= 1) {
                bestMatch = betterMatch(bestMatch, matchSegment(candidateContainers[j], dotSeparatedSegments[i], stringToWordSpans));
            }
            return bestMatch;
        }
        function getWordSpans(word, stringToWordSpans) {
            var spans = stringToWordSpans.get(word);
            if (!spans) {
                stringToWordSpans.set(word, spans = breakIntoWordSpans(word));
            }
            return spans;
        }
        function matchTextChunk(candidate, chunk, stringToWordSpans) {
            var index = indexOfIgnoringCase(candidate, chunk.textLowerCase);
            if (index === 0) {
                // a) Check if the word is a prefix of the candidate, in a case insensitive or
                //    sensitive manner. If it does, return that there was an exact match if the word and candidate are the same length, else a prefix match.
                return createPatternMatch(chunk.text.length === candidate.length ? PatternMatchKind.exact : PatternMatchKind.prefix, /*isCaseSensitive:*/ ts.startsWith(candidate, chunk.text));
            }
            if (chunk.isLowerCase) {
                if (index === -1)
                    return undefined;
                // b) If the part is entirely lowercase, then check if it is contained anywhere in the
                //    candidate in a case insensitive manner.  If so, return that there was a substring
                //    match.
                //
                //    Note: We only have a substring match if the lowercase part is prefix match of some
                //    word part. That way we don't match something like 'Class' when the user types 'a'.
                //    But we would match 'FooAttribute' (since 'Attribute' starts with 'a').
                var wordSpans = getWordSpans(candidate, stringToWordSpans);
                for (var _i = 0, wordSpans_1 = wordSpans; _i < wordSpans_1.length; _i++) {
                    var span = wordSpans_1[_i];
                    if (partStartsWith(candidate, span, chunk.text, /*ignoreCase:*/ true)) {
                        return createPatternMatch(PatternMatchKind.substring, /*isCaseSensitive:*/ partStartsWith(candidate, span, chunk.text, /*ignoreCase:*/ false));
                    }
                }
                // c) Is the pattern a substring of the candidate starting on one of the candidate's word boundaries?
                // We could check every character boundary start of the candidate for the pattern. However, that's
                // an m * n operation in the wost case. Instead, find the first instance of the pattern
                // substring, and see if it starts on a capital letter. It seems unlikely that the user will try to
                // filter the list based on a substring that starts on a capital letter and also with a lowercase one.
                // (Pattern: fogbar, Candidate: quuxfogbarFogBar).
                if (chunk.text.length < candidate.length && isUpperCaseLetter(candidate.charCodeAt(index))) {
                    return createPatternMatch(PatternMatchKind.substring, /*isCaseSensitive:*/ false);
                }
            }
            else {
                // d) If the part was not entirely lowercase, then check if it is contained in the
                //    candidate in a case *sensitive* manner. If so, return that there was a substring
                //    match.
                if (candidate.indexOf(chunk.text) > 0) {
                    return createPatternMatch(PatternMatchKind.substring, /*isCaseSensitive:*/ true);
                }
                // e) If the part was not entirely lowercase, then attempt a camel cased match as well.
                if (chunk.characterSpans.length > 0) {
                    var candidateParts = getWordSpans(candidate, stringToWordSpans);
                    var isCaseSensitive = tryCamelCaseMatch(candidate, candidateParts, chunk, /*ignoreCase:*/ false) ? true
                        : tryCamelCaseMatch(candidate, candidateParts, chunk, /*ignoreCase:*/ true) ? false : undefined;
                    if (isCaseSensitive !== undefined) {
                        return createPatternMatch(PatternMatchKind.camelCase, isCaseSensitive);
                    }
                }
            }
        }
        function matchSegment(candidate, segment, stringToWordSpans) {
            // First check if the segment matches as is.  This is also useful if the segment contains
            // characters we would normally strip when splitting into parts that we also may want to
            // match in the candidate.  For example if the segment is "@int" and the candidate is
            // "@int", then that will show up as an exact match here.
            //
            // Note: if the segment contains a space or an asterisk then we must assume that it's a
            // multi-word segment.
            if (every(segment.totalTextChunk.text, function (ch) { return ch !== 32 /* space */ && ch !== 42 /* asterisk */; })) {
                var match = matchTextChunk(candidate, segment.totalTextChunk, stringToWordSpans);
                if (match)
                    return match;
            }
            // The logic for pattern matching is now as follows:
            //
            // 1) Break the segment passed in into words.  Breaking is rather simple and a
            //    good way to think about it that if gives you all the individual alphanumeric words
            //    of the pattern.
            //
            // 2) For each word try to match the word against the candidate value.
            //
            // 3) Matching is as follows:
            //
            //   a) Check if the word is a prefix of the candidate, in a case insensitive or
            //      sensitive manner. If it does, return that there was an exact match if the word and candidate are the same length, else a prefix match.
            //
            //   If the word is entirely lowercase:
            //      b) Then check if it is contained anywhere in the
            //          candidate in a case insensitive manner.  If so, return that there was a substring
            //          match.
            //
            //          Note: We only have a substring match if the lowercase part is prefix match of
            //          some word part. That way we don't match something like 'Class' when the user
            //          types 'a'. But we would match 'FooAttribute' (since 'Attribute' starts with
            //          'a').
            //
            //       c) The word is all lower case. Is it a case insensitive substring of the candidate starting
            //          on a part boundary of the candidate?
            //
            //   Else:
            //       d) If the word was not entirely lowercase, then check if it is contained in the
            //          candidate in a case *sensitive* manner. If so, return that there was a substring
            //          match.
            //
            //       e) If the word was not entirely lowercase, then attempt a camel cased match as
            //          well.
            //
            // Only if all words have some sort of match is the pattern considered matched.
            var subWordTextChunks = segment.subWordTextChunks;
            var bestMatch;
            for (var _i = 0, subWordTextChunks_1 = subWordTextChunks; _i < subWordTextChunks_1.length; _i++) {
                var subWordTextChunk = subWordTextChunks_1[_i];
                bestMatch = betterMatch(bestMatch, matchTextChunk(candidate, subWordTextChunk, stringToWordSpans));
            }
            return bestMatch;
        }
        function betterMatch(a, b) {
            return ts.min(a, b, compareMatches);
        }
        function compareMatches(a, b) {
            return a === undefined ? 1 /* GreaterThan */ : b === undefined ? -1 /* LessThan */
                : ts.compareValues(a.kind, b.kind) || ts.compareBooleans(!a.isCaseSensitive, !b.isCaseSensitive);
        }
        function partStartsWith(candidate, candidateSpan, pattern, ignoreCase, patternSpan) {
            if (patternSpan === void 0) {
                patternSpan = { start: 0, length: pattern.length };
            }
            return patternSpan.length <= candidateSpan.length // If pattern part is longer than the candidate part there can never be a match.
                && everyInRange(0, patternSpan.length, function (i) { return equalChars(pattern.charCodeAt(patternSpan.start + i), candidate.charCodeAt(candidateSpan.start + i), ignoreCase); });
        }
        function equalChars(ch1, ch2, ignoreCase) {
            return ignoreCase ? toLowerCase(ch1) === toLowerCase(ch2) : ch1 === ch2;
        }
        function tryCamelCaseMatch(candidate, candidateParts, chunk, ignoreCase) {
            var chunkCharacterSpans = chunk.characterSpans;
            // Note: we may have more pattern parts than candidate parts.  This is because multiple
            // pattern parts may match a candidate part.  For example "SiUI" against "SimpleUI".
            // We'll have 3 pattern parts Si/U/I against two candidate parts Simple/UI.  However, U
            // and I will both match in UI.
            var currentCandidate = 0;
            var currentChunkSpan = 0;
            var firstMatch;
            var contiguous;
            while (true) {
                // Let's consider our termination cases
                if (currentChunkSpan === chunkCharacterSpans.length) {
                    return true;
                }
                else if (currentCandidate === candidateParts.length) {
                    // No match, since we still have more of the pattern to hit
                    return false;
                }
                var candidatePart = candidateParts[currentCandidate];
                var gotOneMatchThisCandidate = false;
                // Consider the case of matching SiUI against SimpleUIElement. The candidate parts
                // will be Simple/UI/Element, and the pattern parts will be Si/U/I.  We'll match 'Si'
                // against 'Simple' first.  Then we'll match 'U' against 'UI'. However, we want to
                // still keep matching pattern parts against that candidate part.
                for (; currentChunkSpan < chunkCharacterSpans.length; currentChunkSpan++) {
                    var chunkCharacterSpan = chunkCharacterSpans[currentChunkSpan];
                    if (gotOneMatchThisCandidate) {
                        // We've already gotten one pattern part match in this candidate.  We will
                        // only continue trying to consumer pattern parts if the last part and this
                        // part are both upper case.
                        if (!isUpperCaseLetter(chunk.text.charCodeAt(chunkCharacterSpans[currentChunkSpan - 1].start)) ||
                            !isUpperCaseLetter(chunk.text.charCodeAt(chunkCharacterSpans[currentChunkSpan].start))) {
                            break;
                        }
                    }
                    if (!partStartsWith(candidate, candidatePart, chunk.text, ignoreCase, chunkCharacterSpan)) {
                        break;
                    }
                    gotOneMatchThisCandidate = true;
                    firstMatch = firstMatch === undefined ? currentCandidate : firstMatch;
                    // If we were contiguous, then keep that value.  If we weren't, then keep that
                    // value.  If we don't know, then set the value to 'true' as an initial match is
                    // obviously contiguous.
                    contiguous = contiguous === undefined ? true : contiguous;
                    candidatePart = ts.createTextSpan(candidatePart.start + chunkCharacterSpan.length, candidatePart.length - chunkCharacterSpan.length);
                }
                // Check if we matched anything at all.  If we didn't, then we need to unset the
                // contiguous bit if we currently had it set.
                // If we haven't set the bit yet, then that means we haven't matched anything so
                // far, and we don't want to change that.
                if (!gotOneMatchThisCandidate && contiguous !== undefined) {
                    contiguous = false;
                }
                // Move onto the next candidate.
                currentCandidate++;
            }
        }
        function createSegment(text) {
            return {
                totalTextChunk: createTextChunk(text),
                subWordTextChunks: breakPatternIntoTextChunks(text)
            };
        }
        function isUpperCaseLetter(ch) {
            // Fast check for the ascii range.
            if (ch >= 65 /* A */ && ch <= 90 /* Z */) {
                return true;
            }
            if (ch < 127 /* maxAsciiCharacter */ || !ts.isUnicodeIdentifierStart(ch, 99 /* Latest */)) {
                return false;
            }
            // TODO: find a way to determine this for any unicode characters in a
            // non-allocating manner.
            var str = String.fromCharCode(ch);
            return str === str.toUpperCase();
        }
        function isLowerCaseLetter(ch) {
            // Fast check for the ascii range.
            if (ch >= 97 /* a */ && ch <= 122 /* z */) {
                return true;
            }
            if (ch < 127 /* maxAsciiCharacter */ || !ts.isUnicodeIdentifierStart(ch, 99 /* Latest */)) {
                return false;
            }
            // TODO: find a way to determine this for any unicode characters in a
            // non-allocating manner.
            var str = String.fromCharCode(ch);
            return str === str.toLowerCase();
        }
        // Assumes 'value' is already lowercase.
        function indexOfIgnoringCase(str, value) {
            var n = str.length - value.length;
            var _loop_7 = function (start) {
                if (every(value, function (valueChar, i) { return toLowerCase(str.charCodeAt(i + start)) === valueChar; })) {
                    return { value: start };
                }
            };
            for (var start = 0; start <= n; start++) {
                var state_2 = _loop_7(start);
                if (typeof state_2 === "object")
                    return state_2.value;
            }
            return -1;
        }
        function toLowerCase(ch) {
            // Fast convert for the ascii range.
            if (ch >= 65 /* A */ && ch <= 90 /* Z */) {
                return 97 /* a */ + (ch - 65 /* A */);
            }
            if (ch < 127 /* maxAsciiCharacter */) {
                return ch;
            }
            // TODO: find a way to compute this for any unicode characters in a
            // non-allocating manner.
            return String.fromCharCode(ch).toLowerCase().charCodeAt(0);
        }
        function isDigit(ch) {
            // TODO(cyrusn): Find a way to support this for unicode digits.
            return ch >= 48 /* _0 */ && ch <= 57 /* _9 */;
        }
        function isWordChar(ch) {
            return isUpperCaseLetter(ch) || isLowerCaseLetter(ch) || isDigit(ch) || ch === 95 /* _ */ || ch === 36 /* $ */;
        }
        function breakPatternIntoTextChunks(pattern) {
            var result = [];
            var wordStart = 0;
            var wordLength = 0;
            for (var i = 0; i < pattern.length; i++) {
                var ch = pattern.charCodeAt(i);
                if (isWordChar(ch)) {
                    if (wordLength === 0) {
                        wordStart = i;
                    }
                    wordLength++;
                }
                else {
                    if (wordLength > 0) {
                        result.push(createTextChunk(pattern.substr(wordStart, wordLength)));
                        wordLength = 0;
                    }
                }
            }
            if (wordLength > 0) {
                result.push(createTextChunk(pattern.substr(wordStart, wordLength)));
            }
            return result;
        }
        function createTextChunk(text) {
            var textLowerCase = text.toLowerCase();
            return {
                text: text,
                textLowerCase: textLowerCase,
                isLowerCase: text === textLowerCase,
                characterSpans: breakIntoCharacterSpans(text)
            };
        }
        function breakIntoCharacterSpans(identifier) {
            return breakIntoSpans(identifier, /*word:*/ false);
        }
        ts.breakIntoCharacterSpans = breakIntoCharacterSpans;
        function breakIntoWordSpans(identifier) {
            return breakIntoSpans(identifier, /*word:*/ true);
        }
        ts.breakIntoWordSpans = breakIntoWordSpans;
        function breakIntoSpans(identifier, word) {
            var result = [];
            var wordStart = 0;
            for (var i = 1; i < identifier.length; i++) {
                var lastIsDigit = isDigit(identifier.charCodeAt(i - 1));
                var currentIsDigit = isDigit(identifier.charCodeAt(i));
                var hasTransitionFromLowerToUpper = transitionFromLowerToUpper(identifier, word, i);
                var hasTransitionFromUpperToLower = word && transitionFromUpperToLower(identifier, i, wordStart);
                if (charIsPunctuation(identifier.charCodeAt(i - 1)) ||
                    charIsPunctuation(identifier.charCodeAt(i)) ||
                    lastIsDigit !== currentIsDigit ||
                    hasTransitionFromLowerToUpper ||
                    hasTransitionFromUpperToLower) {
                    if (!isAllPunctuation(identifier, wordStart, i)) {
                        result.push(ts.createTextSpan(wordStart, i - wordStart));
                    }
                    wordStart = i;
                }
            }
            if (!isAllPunctuation(identifier, wordStart, identifier.length)) {
                result.push(ts.createTextSpan(wordStart, identifier.length - wordStart));
            }
            return result;
        }
        function charIsPunctuation(ch) {
            switch (ch) {
                case 33 /* exclamation */:
                case 34 /* doubleQuote */:
                case 35 /* hash */:
                case 37 /* percent */:
                case 38 /* ampersand */:
                case 39 /* singleQuote */:
                case 40 /* openParen */:
                case 41 /* closeParen */:
                case 42 /* asterisk */:
                case 44 /* comma */:
                case 45 /* minus */:
                case 46 /* dot */:
                case 47 /* slash */:
                case 58 /* colon */:
                case 59 /* semicolon */:
                case 63 /* question */:
                case 64 /* at */:
                case 91 /* openBracket */:
                case 92 /* backslash */:
                case 93 /* closeBracket */:
                case 95 /* _ */:
                case 123 /* openBrace */:
                case 125 /* closeBrace */:
                    return true;
            }
            return false;
        }
        function isAllPunctuation(identifier, start, end) {
            return every(identifier, function (ch) { return charIsPunctuation(ch) && ch !== 95 /* _ */; }, start, end);
        }
        function transitionFromUpperToLower(identifier, index, wordStart) {
            // Cases this supports:
            // 1) IDisposable -> I, Disposable
            // 2) UIElement -> UI, Element
            // 3) HTMLDocument -> HTML, Document
            //
            // etc.
            // We have a transition from an upper to a lower letter here.  But we only
            // want to break if all the letters that preceded are uppercase.  i.e. if we
            // have "Foo" we don't want to break that into "F, oo".  But if we have
            // "IFoo" or "UIFoo", then we want to break that into "I, Foo" and "UI,
            // Foo".  i.e. the last uppercase letter belongs to the lowercase letters
            // that follows.  Note: this will make the following not split properly:
            // "HELLOthere".  However, these sorts of names do not show up in .Net
            // programs.
            return index !== wordStart
                && index + 1 < identifier.length
                && isUpperCaseLetter(identifier.charCodeAt(index))
                && isLowerCaseLetter(identifier.charCodeAt(index + 1))
                && every(identifier, isUpperCaseLetter, wordStart, index);
        }
        function transitionFromLowerToUpper(identifier, word, index) {
            var lastIsUpper = isUpperCaseLetter(identifier.charCodeAt(index - 1));
            var currentIsUpper = isUpperCaseLetter(identifier.charCodeAt(index));
            // See if the casing indicates we're starting a new word. Note: if we're breaking on
            // words, then just seeing an upper case character isn't enough.  Instead, it has to
            // be uppercase and the previous character can't be uppercase.
            //
            // For example, breaking "AddMetadata" on words would make: Add Metadata
            //
            // on characters would be: A dd M etadata
            //
            // Break "AM" on words would be: AM
            //
            // on characters would be: A M
            //
            // We break the search string on characters.  But we break the symbol name on words.
            return currentIsUpper && (!word || !lastIsUpper);
        }
        function everyInRange(start, end, pred) {
            for (var i = start; i < end; i++) {
                if (!pred(i)) {
                    return false;
                }
            }
            return true;
        }
        function every(s, pred, start, end) {
            if (start === void 0) {
                start = 0;
            }
            if (end === void 0) {
                end = s.length;
            }
            return everyInRange(start, end, function (i) { return pred(s.charCodeAt(i), i); });
        }
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        function preProcessFile(sourceText, readImportFiles, detectJavaScriptImports) {
            if (readImportFiles === void 0) {
                readImportFiles = true;
            }
            if (detectJavaScriptImports === void 0) {
                detectJavaScriptImports = false;
            }
            var pragmaContext = {
                languageVersion: 1 /* ES5 */,
                pragmas: undefined,
                checkJsDirective: undefined,
                referencedFiles: [],
                typeReferenceDirectives: [],
                libReferenceDirectives: [],
                amdDependencies: [],
                hasNoDefaultLib: undefined,
                moduleName: undefined
            };
            var importedFiles = [];
            var ambientExternalModules;
            var lastToken;
            var currentToken;
            var braceNesting = 0;
            // assume that text represent an external module if it contains at least one top level import/export
            // ambient modules that are found inside external modules are interpreted as module augmentations
            var externalModule = false;
            function nextToken() {
                lastToken = currentToken;
                currentToken = ts.scanner.scan();
                if (currentToken === 18 /* OpenBraceToken */) {
                    braceNesting++;
                }
                else if (currentToken === 19 /* CloseBraceToken */) {
                    braceNesting--;
                }
                return currentToken;
            }
            function getFileReference() {
                var fileName = ts.scanner.getTokenValue();
                var pos = ts.scanner.getTokenPos();
                return { fileName: fileName, pos: pos, end: pos + fileName.length };
            }
            function recordAmbientExternalModule() {
                if (!ambientExternalModules) {
                    ambientExternalModules = [];
                }
                ambientExternalModules.push({ ref: getFileReference(), depth: braceNesting });
            }
            function recordModuleName() {
                importedFiles.push(getFileReference());
                markAsExternalModuleIfTopLevel();
            }
            function markAsExternalModuleIfTopLevel() {
                if (braceNesting === 0) {
                    externalModule = true;
                }
            }
            /**
             * Returns true if at least one token was consumed from the stream
             */
            function tryConsumeDeclare() {
                var token = ts.scanner.getToken();
                if (token === 134 /* DeclareKeyword */) {
                    // declare module "mod"
                    token = nextToken();
                    if (token === 140 /* ModuleKeyword */) {
                        token = nextToken();
                        if (token === 10 /* StringLiteral */) {
                            recordAmbientExternalModule();
                        }
                    }
                    return true;
                }
                return false;
            }
            /**
             * Returns true if at least one token was consumed from the stream
             */
            function tryConsumeImport() {
                if (lastToken === 24 /* DotToken */) {
                    return false;
                }
                var token = ts.scanner.getToken();
                if (token === 100 /* ImportKeyword */) {
                    token = nextToken();
                    if (token === 20 /* OpenParenToken */) {
                        token = nextToken();
                        if (token === 10 /* StringLiteral */ || token === 14 /* NoSubstitutionTemplateLiteral */) {
                            // import("mod");
                            recordModuleName();
                            return true;
                        }
                    }
                    else if (token === 10 /* StringLiteral */) {
                        // import "mod";
                        recordModuleName();
                        return true;
                    }
                    else {
                        if (token === 150 /* TypeKeyword */) {
                            var skipTypeKeyword = ts.scanner.lookAhead(function () {
                                var token = ts.scanner.scan();
                                return token !== 154 /* FromKeyword */ && (token === 41 /* AsteriskToken */ ||
                                    token === 18 /* OpenBraceToken */ ||
                                    token === 79 /* Identifier */ ||
                                    ts.isKeyword(token));
                            });
                            if (skipTypeKeyword) {
                                token = nextToken();
                            }
                        }
                        if (token === 79 /* Identifier */ || ts.isKeyword(token)) {
                            token = nextToken();
                            if (token === 154 /* FromKeyword */) {
                                token = nextToken();
                                if (token === 10 /* StringLiteral */) {
                                    // import d from "mod";
                                    recordModuleName();
                                    return true;
                                }
                            }
                            else if (token === 63 /* EqualsToken */) {
                                if (tryConsumeRequireCall(/*skipCurrentToken*/ true)) {
                                    return true;
                                }
                            }
                            else if (token === 27 /* CommaToken */) {
                                // consume comma and keep going
                                token = nextToken();
                            }
                            else {
                                // unknown syntax
                                return true;
                            }
                        }
                        if (token === 18 /* OpenBraceToken */) {
                            token = nextToken();
                            // consume "{ a as B, c, d as D}" clauses
                            // make sure that it stops on EOF
                            while (token !== 19 /* CloseBraceToken */ && token !== 1 /* EndOfFileToken */) {
                                token = nextToken();
                            }
                            if (token === 19 /* CloseBraceToken */) {
                                token = nextToken();
                                if (token === 154 /* FromKeyword */) {
                                    token = nextToken();
                                    if (token === 10 /* StringLiteral */) {
                                        // import {a as A} from "mod";
                                        // import d, {a, b as B} from "mod"
                                        recordModuleName();
                                    }
                                }
                            }
                        }
                        else if (token === 41 /* AsteriskToken */) {
                            token = nextToken();
                            if (token === 127 /* AsKeyword */) {
                                token = nextToken();
                                if (token === 79 /* Identifier */ || ts.isKeyword(token)) {
                                    token = nextToken();
                                    if (token === 154 /* FromKeyword */) {
                                        token = nextToken();
                                        if (token === 10 /* StringLiteral */) {
                                            // import * as NS from "mod"
                                            // import d, * as NS from "mod"
                                            recordModuleName();
                                        }
                                    }
                                }
                            }
                        }
                    }
                    return true;
                }
                return false;
            }
            function tryConsumeExport() {
                var token = ts.scanner.getToken();
                if (token === 93 /* ExportKeyword */) {
                    markAsExternalModuleIfTopLevel();
                    token = nextToken();
                    if (token === 150 /* TypeKeyword */) {
                        var skipTypeKeyword = ts.scanner.lookAhead(function () {
                            var token = ts.scanner.scan();
                            return token === 41 /* AsteriskToken */ ||
                                token === 18 /* OpenBraceToken */;
                        });
                        if (skipTypeKeyword) {
                            token = nextToken();
                        }
                    }
                    if (token === 18 /* OpenBraceToken */) {
                        token = nextToken();
                        // consume "{ a as B, c, d as D}" clauses
                        // make sure it stops on EOF
                        while (token !== 19 /* CloseBraceToken */ && token !== 1 /* EndOfFileToken */) {
                            token = nextToken();
                        }
                        if (token === 19 /* CloseBraceToken */) {
                            token = nextToken();
                            if (token === 154 /* FromKeyword */) {
                                token = nextToken();
                                if (token === 10 /* StringLiteral */) {
                                    // export {a as A} from "mod";
                                    // export {a, b as B} from "mod"
                                    recordModuleName();
                                }
                            }
                        }
                    }
                    else if (token === 41 /* AsteriskToken */) {
                        token = nextToken();
                        if (token === 154 /* FromKeyword */) {
                            token = nextToken();
                            if (token === 10 /* StringLiteral */) {
                                // export * from "mod"
                                recordModuleName();
                            }
                        }
                    }
                    else if (token === 100 /* ImportKeyword */) {
                        token = nextToken();
                        if (token === 150 /* TypeKeyword */) {
                            var skipTypeKeyword = ts.scanner.lookAhead(function () {
                                var token = ts.scanner.scan();
                                return token === 79 /* Identifier */ ||
                                    ts.isKeyword(token);
                            });
                            if (skipTypeKeyword) {
                                token = nextToken();
                            }
                        }
                        if (token === 79 /* Identifier */ || ts.isKeyword(token)) {
                            token = nextToken();
                            if (token === 63 /* EqualsToken */) {
                                if (tryConsumeRequireCall(/*skipCurrentToken*/ true)) {
                                    return true;
                                }
                            }
                        }
                    }
                    return true;
                }
                return false;
            }
            function tryConsumeRequireCall(skipCurrentToken, allowTemplateLiterals) {
                if (allowTemplateLiterals === void 0) {
                    allowTemplateLiterals = false;
                }
                var token = skipCurrentToken ? nextToken() : ts.scanner.getToken();
                if (token === 144 /* RequireKeyword */) {
                    token = nextToken();
                    if (token === 20 /* OpenParenToken */) {
                        token = nextToken();
                        if (token === 10 /* StringLiteral */ ||
                            allowTemplateLiterals && token === 14 /* NoSubstitutionTemplateLiteral */) {
                            //  require("mod");
                            recordModuleName();
                        }
                    }
                    return true;
                }
                return false;
            }
            function tryConsumeDefine() {
                var token = ts.scanner.getToken();
                if (token === 79 /* Identifier */ && ts.scanner.getTokenValue() === "define") {
                    token = nextToken();
                    if (token !== 20 /* OpenParenToken */) {
                        return true;
                    }
                    token = nextToken();
                    if (token === 10 /* StringLiteral */ || token === 14 /* NoSubstitutionTemplateLiteral */) {
                        // looks like define ("modname", ... - skip string literal and comma
                        token = nextToken();
                        if (token === 27 /* CommaToken */) {
                            token = nextToken();
                        }
                        else {
                            // unexpected token
                            return true;
                        }
                    }
                    // should be start of dependency list
                    if (token !== 22 /* OpenBracketToken */) {
                        return true;
                    }
                    // skip open bracket
                    token = nextToken();
                    // scan until ']' or EOF
                    while (token !== 23 /* CloseBracketToken */ && token !== 1 /* EndOfFileToken */) {
                        // record string literals as module names
                        if (token === 10 /* StringLiteral */ || token === 14 /* NoSubstitutionTemplateLiteral */) {
                            recordModuleName();
                        }
                        token = nextToken();
                    }
                    return true;
                }
                return false;
            }
            function processImports() {
                ts.scanner.setText(sourceText);
                nextToken();
                // Look for:
                //    import "mod";
                //    import d from "mod"
                //    import {a as A } from "mod";
                //    import * as NS from "mod"
                //    import d, {a, b as B} from "mod"
                //    import i = require("mod");
                //    import("mod");
                //    export * from "mod"
                //    export {a as b} from "mod"
                //    export import i = require("mod")
                //    (for JavaScript files) require("mod")
                // Do not look for:
                //    AnySymbol.import("mod")
                //    AnySymbol.nested.import("mod")
                while (true) {
                    if (ts.scanner.getToken() === 1 /* EndOfFileToken */) {
                        break;
                    }
                    // check if at least one of alternative have moved scanner forward
                    if (tryConsumeDeclare() ||
                        tryConsumeImport() ||
                        tryConsumeExport() ||
                        (detectJavaScriptImports && (tryConsumeRequireCall(/*skipCurrentToken*/ false, /*allowTemplateLiterals*/ true) ||
                            tryConsumeDefine()))) {
                        continue;
                    }
                    else {
                        nextToken();
                    }
                }
                ts.scanner.setText(undefined);
            }
            if (readImportFiles) {
                processImports();
            }
            ts.processCommentPragmas(pragmaContext, sourceText);
            ts.processPragmasIntoFields(pragmaContext, ts.noop);
            if (externalModule) {
                // for external modules module all nested ambient modules are augmentations
                if (ambientExternalModules) {
                    // move all detected ambient modules to imported files since they need to be resolved
                    for (var _i = 0, ambientExternalModules_1 = ambientExternalModules; _i < ambientExternalModules_1.length; _i++) {
                        var decl = ambientExternalModules_1[_i];
                        importedFiles.push(decl.ref);
                    }
                }
                return { referencedFiles: pragmaContext.referencedFiles, typeReferenceDirectives: pragmaContext.typeReferenceDirectives, libReferenceDirectives: pragmaContext.libReferenceDirectives, importedFiles: importedFiles, isLibFile: !!pragmaContext.hasNoDefaultLib, ambientExternalModules: undefined };
            }
            else {
                // for global scripts ambient modules still can have augmentations - look for ambient modules with depth > 0
                var ambientModuleNames = void 0;
                if (ambientExternalModules) {
                    for (var _a = 0, ambientExternalModules_2 = ambientExternalModules; _a < ambientExternalModules_2.length; _a++) {
                        var decl = ambientExternalModules_2[_a];
                        if (decl.depth === 0) {
                            if (!ambientModuleNames) {
                                ambientModuleNames = [];
                            }
                            ambientModuleNames.push(decl.ref.fileName);
                        }
                        else {
                            importedFiles.push(decl.ref);
                        }
                    }
                }
                return { referencedFiles: pragmaContext.referencedFiles, typeReferenceDirectives: pragmaContext.typeReferenceDirectives, libReferenceDirectives: pragmaContext.libReferenceDirectives, importedFiles: importedFiles, isLibFile: !!pragmaContext.hasNoDefaultLib, ambientExternalModules: ambientModuleNames };
            }
        }
        ts.preProcessFile = preProcessFile;
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var Rename;
        (function (Rename) {
            function getRenameInfo(program, sourceFile, position, options) {
                var node = ts.getAdjustedRenameLocation(ts.getTouchingPropertyName(sourceFile, position));
                if (nodeIsEligibleForRename(node)) {
                    var renameInfo = getRenameInfoForNode(node, program.getTypeChecker(), sourceFile, program, options);
                    if (renameInfo) {
                        return renameInfo;
                    }
                }
                return getRenameInfoError(ts.Diagnostics.You_cannot_rename_this_element);
            }
            Rename.getRenameInfo = getRenameInfo;
            function getRenameInfoForNode(node, typeChecker, sourceFile, program, options) {
                var symbol = typeChecker.getSymbolAtLocation(node);
                if (!symbol) {
                    if (ts.isStringLiteralLike(node)) {
                        var type = ts.getContextualTypeOrAncestorTypeNodeType(node, typeChecker);
                        if (type && ((type.flags & 128 /* StringLiteral */) || ((type.flags & 1048576 /* Union */) && ts.every(type.types, function (type) { return !!(type.flags & 128 /* StringLiteral */); })))) {
                            return getRenameInfoSuccess(node.text, node.text, "string" /* string */, "", node, sourceFile);
                        }
                    }
                    else if (ts.isLabelName(node)) {
                        var name = ts.getTextOfNode(node);
                        return getRenameInfoSuccess(name, name, "label" /* label */, "" /* none */, node, sourceFile);
                    }
                    return undefined;
                }
                // Only allow a symbol to be renamed if it actually has at least one declaration.
                var declarations = symbol.declarations;
                if (!declarations || declarations.length === 0)
                    return;
                // Disallow rename for elements that are defined in the standard TypeScript library.
                if (declarations.some(function (declaration) { return isDefinedInLibraryFile(program, declaration); })) {
                    return getRenameInfoError(ts.Diagnostics.You_cannot_rename_elements_that_are_defined_in_the_standard_TypeScript_library);
                }
                // Cannot rename `default` as in `import { default as foo } from "./someModule";
                if (ts.isIdentifier(node) && node.originalKeywordKind === 88 /* DefaultKeyword */ && symbol.parent && symbol.parent.flags & 1536 /* Module */) {
                    return undefined;
                }
                if (ts.isStringLiteralLike(node) && ts.tryGetImportFromModuleSpecifier(node)) {
                    return options && options.allowRenameOfImportPath ? getRenameInfoForModule(node, sourceFile, symbol) : undefined;
                }
                var kind = ts.SymbolDisplay.getSymbolKind(typeChecker, symbol, node);
                var specifierName = (ts.isImportOrExportSpecifierName(node) || ts.isStringOrNumericLiteralLike(node) && node.parent.kind === 160 /* ComputedPropertyName */)
                    ? ts.stripQuotes(ts.getTextOfIdentifierOrLiteral(node))
                    : undefined;
                var displayName = specifierName || typeChecker.symbolToString(symbol);
                var fullDisplayName = specifierName || typeChecker.getFullyQualifiedName(symbol);
                return getRenameInfoSuccess(displayName, fullDisplayName, kind, ts.SymbolDisplay.getSymbolModifiers(typeChecker, symbol), node, sourceFile);
            }
            function isDefinedInLibraryFile(program, declaration) {
                var sourceFile = declaration.getSourceFile();
                return program.isSourceFileDefaultLibrary(sourceFile) && ts.fileExtensionIs(sourceFile.fileName, ".d.ts" /* Dts */);
            }
            function getRenameInfoForModule(node, sourceFile, moduleSymbol) {
                if (!ts.isExternalModuleNameRelative(node.text)) {
                    return getRenameInfoError(ts.Diagnostics.You_cannot_rename_a_module_via_a_global_import);
                }
                var moduleSourceFile = moduleSymbol.declarations && ts.find(moduleSymbol.declarations, ts.isSourceFile);
                if (!moduleSourceFile)
                    return undefined;
                var withoutIndex = ts.endsWith(node.text, "/index") || ts.endsWith(node.text, "/index.js") ? undefined : ts.tryRemoveSuffix(ts.removeFileExtension(moduleSourceFile.fileName), "/index");
                var name = withoutIndex === undefined ? moduleSourceFile.fileName : withoutIndex;
                var kind = withoutIndex === undefined ? "module" /* moduleElement */ : "directory" /* directory */;
                var indexAfterLastSlash = node.text.lastIndexOf("/") + 1;
                // Span should only be the last component of the path. + 1 to account for the quote character.
                var triggerSpan = ts.createTextSpan(node.getStart(sourceFile) + 1 + indexAfterLastSlash, node.text.length - indexAfterLastSlash);
                return {
                    canRename: true,
                    fileToRename: name,
                    kind: kind,
                    displayName: name,
                    fullDisplayName: name,
                    kindModifiers: "" /* none */,
                    triggerSpan: triggerSpan,
                };
            }
            function getRenameInfoSuccess(displayName, fullDisplayName, kind, kindModifiers, node, sourceFile) {
                return {
                    canRename: true,
                    fileToRename: undefined,
                    kind: kind,
                    displayName: displayName,
                    fullDisplayName: fullDisplayName,
                    kindModifiers: kindModifiers,
                    triggerSpan: createTriggerSpanForNode(node, sourceFile)
                };
            }
            function getRenameInfoError(diagnostic) {
                return { canRename: false, localizedErrorMessage: ts.getLocaleSpecificMessage(diagnostic) };
            }
            function createTriggerSpanForNode(node, sourceFile) {
                var start = node.getStart(sourceFile);
                var width = node.getWidth(sourceFile);
                if (ts.isStringLiteralLike(node)) {
                    // Exclude the quotes
                    start += 1;
                    width -= 2;
                }
                return ts.createTextSpan(start, width);
            }
            function nodeIsEligibleForRename(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                    case 80 /* PrivateIdentifier */:
                    case 10 /* StringLiteral */:
                    case 14 /* NoSubstitutionTemplateLiteral */:
                    case 108 /* ThisKeyword */:
                        return true;
                    case 8 /* NumericLiteral */:
                        return ts.isLiteralNameOfPropertyDeclarationOrIndexAccess(node);
                    default:
                        return false;
                }
            }
        })(Rename = ts.Rename || (ts.Rename = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var SmartSelectionRange;
        (function (SmartSelectionRange) {
            function getSmartSelectionRange(pos, sourceFile) {
                var _a;
                var selectionRange = {
                    textSpan: ts.createTextSpanFromBounds(sourceFile.getFullStart(), sourceFile.getEnd())
                };
                var parentNode = sourceFile;
                outer: while (true) {
                    var children = getSelectionChildren(parentNode);
                    if (!children.length)
                        break;
                    for (var i = 0; i < children.length; i++) {
                        var prevNode = children[i - 1];
                        var node = children[i];
                        var nextNode = children[i + 1];
                        if (ts.getTokenPosOfNode(node, sourceFile, /*includeJsDoc*/ true) > pos) {
                            break outer;
                        }
                        var comment = ts.singleOrUndefined(ts.getTrailingCommentRanges(sourceFile.text, node.end));
                        if (comment && comment.kind === 2 /* SingleLineCommentTrivia */) {
                            pushSelectionCommentRange(comment.pos, comment.end);
                        }
                        if (positionShouldSnapToNode(sourceFile, pos, node)) {
                            // 1. Blocks are effectively redundant with SyntaxLists.
                            // 2. TemplateSpans, along with the SyntaxLists containing them, are a somewhat unintuitive grouping
                            //    of things that should be considered independently.
                            // 3. A VariableStatement’s children are just a VaraiableDeclarationList and a semicolon.
                            // 4. A lone VariableDeclaration in a VaraibleDeclaration feels redundant with the VariableStatement.
                            // Dive in without pushing a selection range.
                            if (ts.isBlock(node)
                                || ts.isTemplateSpan(node) || ts.isTemplateHead(node) || ts.isTemplateTail(node)
                                || prevNode && ts.isTemplateHead(prevNode)
                                || ts.isVariableDeclarationList(node) && ts.isVariableStatement(parentNode)
                                || ts.isSyntaxList(node) && ts.isVariableDeclarationList(parentNode)
                                || ts.isVariableDeclaration(node) && ts.isSyntaxList(parentNode) && children.length === 1
                                || ts.isJSDocTypeExpression(node) || ts.isJSDocSignature(node) || ts.isJSDocTypeLiteral(node)) {
                                parentNode = node;
                                break;
                            }
                            // Synthesize a stop for '${ ... }' since '${' and '}' actually belong to siblings.
                            if (ts.isTemplateSpan(parentNode) && nextNode && ts.isTemplateMiddleOrTemplateTail(nextNode)) {
                                var start_1 = node.getFullStart() - "${".length;
                                var end_2 = nextNode.getStart() + "}".length;
                                pushSelectionRange(start_1, end_2);
                            }
                            // Blocks with braces, brackets, parens, or JSX tags on separate lines should be
                            // selected from open to close, including whitespace but not including the braces/etc. themselves.
                            var isBetweenMultiLineBookends = ts.isSyntaxList(node) && isListOpener(prevNode) && isListCloser(nextNode)
                                && !ts.positionsAreOnSameLine(prevNode.getStart(), nextNode.getStart(), sourceFile);
                            var start = isBetweenMultiLineBookends ? prevNode.getEnd() : node.getStart();
                            var end = isBetweenMultiLineBookends ? nextNode.getStart() : getEndPos(sourceFile, node);
                            if (ts.hasJSDocNodes(node) && ((_a = node.jsDoc) === null || _a === void 0 ? void 0 : _a.length)) {
                                pushSelectionRange(ts.first(node.jsDoc).getStart(), end);
                            }
                            pushSelectionRange(start, end);
                            // String literals should have a stop both inside and outside their quotes.
                            if (ts.isStringLiteral(node) || ts.isTemplateLiteral(node)) {
                                pushSelectionRange(start + 1, end - 1);
                            }
                            parentNode = node;
                            break;
                        }
                        // If we made it to the end of the for loop, we’re done.
                        // In practice, I’ve only seen this happen at the very end
                        // of a SourceFile.
                        if (i === children.length - 1) {
                            break outer;
                        }
                    }
                }
                return selectionRange;
                function pushSelectionRange(start, end) {
                    // Skip empty ranges
                    if (start !== end) {
                        var textSpan = ts.createTextSpanFromBounds(start, end);
                        if (!selectionRange || (
                        // Skip ranges that are identical to the parent
                        !ts.textSpansEqual(textSpan, selectionRange.textSpan) &&
                            // Skip ranges that don’t contain the original position
                            ts.textSpanIntersectsWithPosition(textSpan, pos))) {
                            selectionRange = __assign({ textSpan: textSpan }, selectionRange && { parent: selectionRange });
                        }
                    }
                }
                function pushSelectionCommentRange(start, end) {
                    pushSelectionRange(start, end);
                    var pos = start;
                    while (sourceFile.text.charCodeAt(pos) === 47 /* slash */) {
                        pos++;
                    }
                    pushSelectionRange(pos, end);
                }
            }
            SmartSelectionRange.getSmartSelectionRange = getSmartSelectionRange;
            /**
             * Like `ts.positionBelongsToNode`, except positions immediately after nodes
             * count too, unless that position belongs to the next node. In effect, makes
             * selections able to snap to preceding tokens when the cursor is on the tail
             * end of them with only whitespace ahead.
             * @param sourceFile The source file containing the nodes.
             * @param pos The position to check.
             * @param node The candidate node to snap to.
             */
            function positionShouldSnapToNode(sourceFile, pos, node) {
                // Can’t use 'ts.positionBelongsToNode()' here because it cleverly accounts
                // for missing nodes, which can’t really be considered when deciding what
                // to select.
                ts.Debug.assert(node.pos <= pos);
                if (pos < node.end) {
                    return true;
                }
                var nodeEnd = node.getEnd();
                if (nodeEnd === pos) {
                    return ts.getTouchingPropertyName(sourceFile, pos).pos < node.end;
                }
                return false;
            }
            var isImport = ts.or(ts.isImportDeclaration, ts.isImportEqualsDeclaration);
            /**
             * Gets the children of a node to be considered for selection ranging,
             * transforming them into an artificial tree according to their intuitive
             * grouping where no grouping actually exists in the parse tree. For example,
             * top-level imports are grouped into their own SyntaxList so they can be
             * selected all together, even though in the AST they’re just siblings of each
             * other as well as of other top-level statements and declarations.
             */
            function getSelectionChildren(node) {
                // Group top-level imports
                if (ts.isSourceFile(node)) {
                    return groupChildren(node.getChildAt(0).getChildren(), isImport);
                }
                // Mapped types _look_ like ObjectTypes with a single member,
                // but in fact don’t contain a SyntaxList or a node containing
                // the “key/value” pair like ObjectTypes do, but it seems intuitive
                // that the selection would snap to those points. The philosophy
                // of choosing a selection range is not so much about what the
                // syntax currently _is_ as what the syntax might easily become
                // if the user is making a selection; e.g., we synthesize a selection
                // around the “key/value” pair not because there’s a node there, but
                // because it allows the mapped type to become an object type with a
                // few keystrokes.
                if (ts.isMappedTypeNode(node)) {
                    var _a = node.getChildren(), openBraceToken = _a[0], children = _a.slice(1);
                    var closeBraceToken = ts.Debug.checkDefined(children.pop());
                    ts.Debug.assertEqual(openBraceToken.kind, 18 /* OpenBraceToken */);
                    ts.Debug.assertEqual(closeBraceToken.kind, 19 /* CloseBraceToken */);
                    // Group `-/+readonly` and `-/+?`
                    var groupedWithPlusMinusTokens = groupChildren(children, function (child) {
                        return child === node.readonlyToken || child.kind === 143 /* ReadonlyKeyword */ ||
                            child === node.questionToken || child.kind === 57 /* QuestionToken */;
                    });
                    // Group type parameter with surrounding brackets
                    var groupedWithBrackets = groupChildren(groupedWithPlusMinusTokens, function (_a) {
                        var kind = _a.kind;
                        return kind === 22 /* OpenBracketToken */ ||
                            kind === 161 /* TypeParameter */ ||
                            kind === 23 /* CloseBracketToken */;
                    });
                    return [
                        openBraceToken,
                        // Pivot on `:`
                        createSyntaxList(splitChildren(groupedWithBrackets, function (_a) {
                            var kind = _a.kind;
                            return kind === 58 /* ColonToken */;
                        })),
                        closeBraceToken,
                    ];
                }
                // Group modifiers and property name, then pivot on `:`.
                if (ts.isPropertySignature(node)) {
                    var children = groupChildren(node.getChildren(), function (child) {
                        return child === node.name || ts.contains(node.modifiers, child);
                    });
                    return splitChildren(children, function (_a) {
                        var kind = _a.kind;
                        return kind === 58 /* ColonToken */;
                    });
                }
                // Group the parameter name with its `...`, then that group with its `?`, then pivot on `=`.
                if (ts.isParameter(node)) {
                    var groupedDotDotDotAndName_1 = groupChildren(node.getChildren(), function (child) {
                        return child === node.dotDotDotToken || child === node.name;
                    });
                    var groupedWithQuestionToken = groupChildren(groupedDotDotDotAndName_1, function (child) {
                        return child === groupedDotDotDotAndName_1[0] || child === node.questionToken;
                    });
                    return splitChildren(groupedWithQuestionToken, function (_a) {
                        var kind = _a.kind;
                        return kind === 63 /* EqualsToken */;
                    });
                }
                // Pivot on '='
                if (ts.isBindingElement(node)) {
                    return splitChildren(node.getChildren(), function (_a) {
                        var kind = _a.kind;
                        return kind === 63 /* EqualsToken */;
                    });
                }
                return node.getChildren();
            }
            /**
             * Groups sibling nodes together into their own SyntaxList if they
             * a) are adjacent, AND b) match a predicate function.
             */
            function groupChildren(children, groupOn) {
                var result = [];
                var group;
                for (var _i = 0, children_1 = children; _i < children_1.length; _i++) {
                    var child = children_1[_i];
                    if (groupOn(child)) {
                        group = group || [];
                        group.push(child);
                    }
                    else {
                        if (group) {
                            result.push(createSyntaxList(group));
                            group = undefined;
                        }
                        result.push(child);
                    }
                }
                if (group) {
                    result.push(createSyntaxList(group));
                }
                return result;
            }
            /**
             * Splits sibling nodes into up to four partitions:
             * 1) everything left of the first node matched by `pivotOn`,
             * 2) the first node matched by `pivotOn`,
             * 3) everything right of the first node matched by `pivotOn`,
             * 4) a trailing semicolon, if `separateTrailingSemicolon` is enabled.
             * The left and right groups, if not empty, will each be grouped into their own containing SyntaxList.
             * @param children The sibling nodes to split.
             * @param pivotOn The predicate function to match the node to be the pivot. The first node that matches
             * the predicate will be used; any others that may match will be included into the right-hand group.
             * @param separateTrailingSemicolon If the last token is a semicolon, it will be returned as a separate
             * child rather than be included in the right-hand group.
             */
            function splitChildren(children, pivotOn, separateTrailingSemicolon) {
                if (separateTrailingSemicolon === void 0) {
                    separateTrailingSemicolon = true;
                }
                if (children.length < 2) {
                    return children;
                }
                var splitTokenIndex = ts.findIndex(children, pivotOn);
                if (splitTokenIndex === -1) {
                    return children;
                }
                var leftChildren = children.slice(0, splitTokenIndex);
                var splitToken = children[splitTokenIndex];
                var lastToken = ts.last(children);
                var separateLastToken = separateTrailingSemicolon && lastToken.kind === 26 /* SemicolonToken */;
                var rightChildren = children.slice(splitTokenIndex + 1, separateLastToken ? children.length - 1 : undefined);
                var result = ts.compact([
                    leftChildren.length ? createSyntaxList(leftChildren) : undefined,
                    splitToken,
                    rightChildren.length ? createSyntaxList(rightChildren) : undefined,
                ]);
                return separateLastToken ? result.concat(lastToken) : result;
            }
            function createSyntaxList(children) {
                ts.Debug.assertGreaterThanOrEqual(children.length, 1);
                return ts.setTextRangePosEnd(ts.parseNodeFactory.createSyntaxList(children), children[0].pos, ts.last(children).end);
            }
            function isListOpener(token) {
                var kind = token && token.kind;
                return kind === 18 /* OpenBraceToken */
                    || kind === 22 /* OpenBracketToken */
                    || kind === 20 /* OpenParenToken */
                    || kind === 278 /* JsxOpeningElement */;
            }
            function isListCloser(token) {
                var kind = token && token.kind;
                return kind === 19 /* CloseBraceToken */
                    || kind === 23 /* CloseBracketToken */
                    || kind === 21 /* CloseParenToken */
                    || kind === 279 /* JsxClosingElement */;
            }
            function getEndPos(sourceFile, node) {
                switch (node.kind) {
                    case 335 /* JSDocParameterTag */:
                    case 333 /* JSDocCallbackTag */:
                    case 342 /* JSDocPropertyTag */:
                    case 340 /* JSDocTypedefTag */:
                    case 337 /* JSDocThisTag */:
                        return sourceFile.getLineEndOfPosition(node.getStart());
                    default:
                        return node.getEnd();
                }
            }
        })(SmartSelectionRange = ts.SmartSelectionRange || (ts.SmartSelectionRange = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var SignatureHelp;
        (function (SignatureHelp) {
            var InvocationKind;
            (function (InvocationKind) {
                InvocationKind[InvocationKind["Call"] = 0] = "Call";
                InvocationKind[InvocationKind["TypeArgs"] = 1] = "TypeArgs";
                InvocationKind[InvocationKind["Contextual"] = 2] = "Contextual";
            })(InvocationKind || (InvocationKind = {}));
            function getSignatureHelpItems(program, sourceFile, position, triggerReason, cancellationToken) {
                var typeChecker = program.getTypeChecker();
                // Decide whether to show signature help
                var startingToken = ts.findTokenOnLeftOfPosition(sourceFile, position);
                if (!startingToken) {
                    // We are at the beginning of the file
                    return undefined;
                }
                // Only need to be careful if the user typed a character and signature help wasn't showing.
                var onlyUseSyntacticOwners = !!triggerReason && triggerReason.kind === "characterTyped";
                // Bail out quickly in the middle of a string or comment, don't provide signature help unless the user explicitly requested it.
                if (onlyUseSyntacticOwners && (ts.isInString(sourceFile, position, startingToken) || ts.isInComment(sourceFile, position))) {
                    return undefined;
                }
                var isManuallyInvoked = !!triggerReason && triggerReason.kind === "invoked";
                var argumentInfo = getContainingArgumentInfo(startingToken, position, sourceFile, typeChecker, isManuallyInvoked);
                if (!argumentInfo)
                    return undefined;
                cancellationToken.throwIfCancellationRequested();
                // Extra syntactic and semantic filtering of signature help
                var candidateInfo = getCandidateOrTypeInfo(argumentInfo, typeChecker, sourceFile, startingToken, onlyUseSyntacticOwners);
                cancellationToken.throwIfCancellationRequested();
                if (!candidateInfo) {
                    // We didn't have any sig help items produced by the TS compiler.  If this is a JS
                    // file, then see if we can figure out anything better.
                    return ts.isSourceFileJS(sourceFile) ? createJSSignatureHelpItems(argumentInfo, program, cancellationToken) : undefined;
                }
                return typeChecker.runWithCancellationToken(cancellationToken, function (typeChecker) {
                    return candidateInfo.kind === 0 /* Candidate */
                        ? createSignatureHelpItems(candidateInfo.candidates, candidateInfo.resolvedSignature, argumentInfo, sourceFile, typeChecker)
                        : createTypeHelpItems(candidateInfo.symbol, argumentInfo, sourceFile, typeChecker);
                });
            }
            SignatureHelp.getSignatureHelpItems = getSignatureHelpItems;
            var CandidateOrTypeKind;
            (function (CandidateOrTypeKind) {
                CandidateOrTypeKind[CandidateOrTypeKind["Candidate"] = 0] = "Candidate";
                CandidateOrTypeKind[CandidateOrTypeKind["Type"] = 1] = "Type";
            })(CandidateOrTypeKind || (CandidateOrTypeKind = {}));
            function getCandidateOrTypeInfo(_a, checker, sourceFile, startingToken, onlyUseSyntacticOwners) {
                var invocation = _a.invocation, argumentCount = _a.argumentCount;
                switch (invocation.kind) {
                    case 0 /* Call */: {
                        if (onlyUseSyntacticOwners && !isSyntacticOwner(startingToken, invocation.node, sourceFile)) {
                            return undefined;
                        }
                        var candidates = [];
                        var resolvedSignature = checker.getResolvedSignatureForSignatureHelp(invocation.node, candidates, argumentCount); // TODO: GH#18217
                        return candidates.length === 0 ? undefined : { kind: 0 /* Candidate */, candidates: candidates, resolvedSignature: resolvedSignature };
                    }
                    case 1 /* TypeArgs */: {
                        var called = invocation.called;
                        if (onlyUseSyntacticOwners && !containsPrecedingToken(startingToken, sourceFile, ts.isIdentifier(called) ? called.parent : called)) {
                            return undefined;
                        }
                        var candidates = ts.getPossibleGenericSignatures(called, argumentCount, checker);
                        if (candidates.length !== 0)
                            return { kind: 0 /* Candidate */, candidates: candidates, resolvedSignature: ts.first(candidates) };
                        var symbol = checker.getSymbolAtLocation(called);
                        return symbol && { kind: 1 /* Type */, symbol: symbol };
                    }
                    case 2 /* Contextual */:
                        return { kind: 0 /* Candidate */, candidates: [invocation.signature], resolvedSignature: invocation.signature };
                    default:
                        return ts.Debug.assertNever(invocation);
                }
            }
            function isSyntacticOwner(startingToken, node, sourceFile) {
                if (!ts.isCallOrNewExpression(node))
                    return false;
                var invocationChildren = node.getChildren(sourceFile);
                switch (startingToken.kind) {
                    case 20 /* OpenParenToken */:
                        return ts.contains(invocationChildren, startingToken);
                    case 27 /* CommaToken */: {
                        var containingList = ts.findContainingList(startingToken);
                        return !!containingList && ts.contains(invocationChildren, containingList);
                    }
                    case 29 /* LessThanToken */:
                        return containsPrecedingToken(startingToken, sourceFile, node.expression);
                    default:
                        return false;
                }
            }
            function createJSSignatureHelpItems(argumentInfo, program, cancellationToken) {
                if (argumentInfo.invocation.kind === 2 /* Contextual */)
                    return undefined;
                // See if we can find some symbol with the call expression name that has call signatures.
                var expression = getExpressionFromInvocation(argumentInfo.invocation);
                var name = ts.isPropertyAccessExpression(expression) ? expression.name.text : undefined;
                var typeChecker = program.getTypeChecker();
                return name === undefined ? undefined : ts.firstDefined(program.getSourceFiles(), function (sourceFile) {
                    return ts.firstDefined(sourceFile.getNamedDeclarations().get(name), function (declaration) {
                        var type = declaration.symbol && typeChecker.getTypeOfSymbolAtLocation(declaration.symbol, declaration);
                        var callSignatures = type && type.getCallSignatures();
                        if (callSignatures && callSignatures.length) {
                            return typeChecker.runWithCancellationToken(cancellationToken, function (typeChecker) {
                                return createSignatureHelpItems(callSignatures, callSignatures[0], argumentInfo, sourceFile, typeChecker, 
                                /*useFullPrefix*/ true);
                            });
                        }
                    });
                });
            }
            function containsPrecedingToken(startingToken, sourceFile, container) {
                var pos = startingToken.getFullStart();
                // There’s a possibility that `startingToken.parent` contains only `startingToken` and
                // missing nodes, none of which are valid to be returned by `findPrecedingToken`. In that
                // case, the preceding token we want is actually higher up the tree—almost definitely the
                // next parent, but theoretically the situation with missing nodes might be happening on
                // multiple nested levels.
                var currentParent = startingToken.parent;
                while (currentParent) {
                    var precedingToken = ts.findPrecedingToken(pos, sourceFile, currentParent, /*excludeJsdoc*/ true);
                    if (precedingToken) {
                        return ts.rangeContainsRange(container, precedingToken);
                    }
                    currentParent = currentParent.parent;
                }
                return ts.Debug.fail("Could not find preceding token");
            }
            function getArgumentInfoForCompletions(node, position, sourceFile) {
                var info = getImmediatelyContainingArgumentInfo(node, position, sourceFile);
                return !info || info.isTypeParameterList || info.invocation.kind !== 0 /* Call */ ? undefined
                    : { invocation: info.invocation.node, argumentCount: info.argumentCount, argumentIndex: info.argumentIndex };
            }
            SignatureHelp.getArgumentInfoForCompletions = getArgumentInfoForCompletions;
            function getArgumentOrParameterListInfo(node, sourceFile) {
                var info = getArgumentOrParameterListAndIndex(node, sourceFile);
                if (!info)
                    return undefined;
                var list = info.list, argumentIndex = info.argumentIndex;
                var argumentCount = getArgumentCount(list);
                if (argumentIndex !== 0) {
                    ts.Debug.assertLessThan(argumentIndex, argumentCount);
                }
                var argumentsSpan = getApplicableSpanForArguments(list, sourceFile);
                return { list: list, argumentIndex: argumentIndex, argumentCount: argumentCount, argumentsSpan: argumentsSpan };
            }
            function getArgumentOrParameterListAndIndex(node, sourceFile) {
                if (node.kind === 29 /* LessThanToken */ || node.kind === 20 /* OpenParenToken */) {
                    // Find the list that starts right *after* the < or ( token.
                    // If the user has just opened a list, consider this item 0.
                    return { list: getChildListThatStartsWithOpenerToken(node.parent, node, sourceFile), argumentIndex: 0 };
                }
                else {
                    // findListItemInfo can return undefined if we are not in parent's argument list
                    // or type argument list. This includes cases where the cursor is:
                    //   - To the right of the closing parenthesis, non-substitution template, or template tail.
                    //   - Between the type arguments and the arguments (greater than token)
                    //   - On the target of the call (parent.func)
                    //   - On the 'new' keyword in a 'new' expression
                    var list = ts.findContainingList(node);
                    return list && { list: list, argumentIndex: getArgumentIndex(list, node) };
                }
            }
            /**
             * Returns relevant information for the argument list and the current argument if we are
             * in the argument of an invocation; returns undefined otherwise.
             */
            function getImmediatelyContainingArgumentInfo(node, position, sourceFile) {
                var parent = node.parent;
                if (ts.isCallOrNewExpression(parent)) {
                    var invocation = parent;
                    // There are 3 cases to handle:
                    //   1. The token introduces a list, and should begin a signature help session
                    //   2. The token is either not associated with a list, or ends a list, so the session should end
                    //   3. The token is buried inside a list, and should give signature help
                    //
                    // The following are examples of each:
                    //
                    //    Case 1:
                    //          foo<#T, U>(#a, b)    -> The token introduces a list, and should begin a signature help session
                    //    Case 2:
                    //          fo#o<T, U>#(a, b)#   -> The token is either not associated with a list, or ends a list, so the session should end
                    //    Case 3:
                    //          foo<T#, U#>(a#, #b#) -> The token is buried inside a list, and should give signature help
                    // Find out if 'node' is an argument, a type argument, or neither
                    var info = getArgumentOrParameterListInfo(node, sourceFile);
                    if (!info)
                        return undefined;
                    var list = info.list, argumentIndex = info.argumentIndex, argumentCount = info.argumentCount, argumentsSpan = info.argumentsSpan;
                    var isTypeParameterList = !!parent.typeArguments && parent.typeArguments.pos === list.pos;
                    return { isTypeParameterList: isTypeParameterList, invocation: { kind: 0 /* Call */, node: invocation }, argumentsSpan: argumentsSpan, argumentIndex: argumentIndex, argumentCount: argumentCount };
                }
                else if (ts.isNoSubstitutionTemplateLiteral(node) && ts.isTaggedTemplateExpression(parent)) {
                    // Check if we're actually inside the template;
                    // otherwise we'll fall out and return undefined.
                    if (ts.isInsideTemplateLiteral(node, position, sourceFile)) {
                        return getArgumentListInfoForTemplate(parent, /*argumentIndex*/ 0, sourceFile);
                    }
                    return undefined;
                }
                else if (ts.isTemplateHead(node) && parent.parent.kind === 208 /* TaggedTemplateExpression */) {
                    var templateExpression = parent;
                    var tagExpression = templateExpression.parent;
                    ts.Debug.assert(templateExpression.kind === 221 /* TemplateExpression */);
                    var argumentIndex = ts.isInsideTemplateLiteral(node, position, sourceFile) ? 0 : 1;
                    return getArgumentListInfoForTemplate(tagExpression, argumentIndex, sourceFile);
                }
                else if (ts.isTemplateSpan(parent) && ts.isTaggedTemplateExpression(parent.parent.parent)) {
                    var templateSpan = parent;
                    var tagExpression = parent.parent.parent;
                    // If we're just after a template tail, don't show signature help.
                    if (ts.isTemplateTail(node) && !ts.isInsideTemplateLiteral(node, position, sourceFile)) {
                        return undefined;
                    }
                    var spanIndex = templateSpan.parent.templateSpans.indexOf(templateSpan);
                    var argumentIndex = getArgumentIndexForTemplatePiece(spanIndex, node, position, sourceFile);
                    return getArgumentListInfoForTemplate(tagExpression, argumentIndex, sourceFile);
                }
                else if (ts.isJsxOpeningLikeElement(parent)) {
                    // Provide a signature help for JSX opening element or JSX self-closing element.
                    // This is not guarantee that JSX tag-name is resolved into stateless function component. (that is done in "getSignatureHelpItems")
                    // i.e
                    //      export function MainButton(props: ButtonProps, context: any): JSX.Element { ... }
                    //      <MainButton /*signatureHelp*/
                    var attributeSpanStart = parent.attributes.pos;
                    var attributeSpanEnd = ts.skipTrivia(sourceFile.text, parent.attributes.end, /*stopAfterLineBreak*/ false);
                    return {
                        isTypeParameterList: false,
                        invocation: { kind: 0 /* Call */, node: parent },
                        argumentsSpan: ts.createTextSpan(attributeSpanStart, attributeSpanEnd - attributeSpanStart),
                        argumentIndex: 0,
                        argumentCount: 1
                    };
                }
                else {
                    var typeArgInfo = ts.getPossibleTypeArgumentsInfo(node, sourceFile);
                    if (typeArgInfo) {
                        var called = typeArgInfo.called, nTypeArguments = typeArgInfo.nTypeArguments;
                        var invocation = { kind: 1 /* TypeArgs */, called: called };
                        var argumentsSpan = ts.createTextSpanFromBounds(called.getStart(sourceFile), node.end);
                        return { isTypeParameterList: true, invocation: invocation, argumentsSpan: argumentsSpan, argumentIndex: nTypeArguments, argumentCount: nTypeArguments + 1 };
                    }
                    return undefined;
                }
            }
            function getImmediatelyContainingArgumentOrContextualParameterInfo(node, position, sourceFile, checker) {
                return tryGetParameterInfo(node, position, sourceFile, checker) || getImmediatelyContainingArgumentInfo(node, position, sourceFile);
            }
            function getHighestBinary(b) {
                return ts.isBinaryExpression(b.parent) ? getHighestBinary(b.parent) : b;
            }
            function countBinaryExpressionParameters(b) {
                return ts.isBinaryExpression(b.left) ? countBinaryExpressionParameters(b.left) + 1 : 2;
            }
            function tryGetParameterInfo(startingToken, _position, sourceFile, checker) {
                var info = getContextualSignatureLocationInfo(startingToken, sourceFile, checker);
                if (!info)
                    return undefined;
                var contextualType = info.contextualType, argumentIndex = info.argumentIndex, argumentCount = info.argumentCount, argumentsSpan = info.argumentsSpan;
                // for optional function condition.
                var nonNullableContextualType = contextualType.getNonNullableType();
                var signatures = nonNullableContextualType.getCallSignatures();
                if (signatures.length !== 1)
                    return undefined;
                var invocation = { kind: 2 /* Contextual */, signature: ts.first(signatures), node: startingToken, symbol: chooseBetterSymbol(nonNullableContextualType.symbol) };
                return { isTypeParameterList: false, invocation: invocation, argumentsSpan: argumentsSpan, argumentIndex: argumentIndex, argumentCount: argumentCount };
            }
            function getContextualSignatureLocationInfo(startingToken, sourceFile, checker) {
                if (startingToken.kind !== 20 /* OpenParenToken */ && startingToken.kind !== 27 /* CommaToken */)
                    return undefined;
                var parent = startingToken.parent;
                switch (parent.kind) {
                    case 210 /* ParenthesizedExpression */:
                    case 167 /* MethodDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        var info = getArgumentOrParameterListInfo(startingToken, sourceFile);
                        if (!info)
                            return undefined;
                        var argumentIndex = info.argumentIndex, argumentCount = info.argumentCount, argumentsSpan = info.argumentsSpan;
                        var contextualType = ts.isMethodDeclaration(parent) ? checker.getContextualTypeForObjectLiteralElement(parent) : checker.getContextualType(parent);
                        return contextualType && { contextualType: contextualType, argumentIndex: argumentIndex, argumentCount: argumentCount, argumentsSpan: argumentsSpan };
                    case 219 /* BinaryExpression */: {
                        var highestBinary = getHighestBinary(parent);
                        var contextualType_1 = checker.getContextualType(highestBinary);
                        var argumentIndex_1 = startingToken.kind === 20 /* OpenParenToken */ ? 0 : countBinaryExpressionParameters(parent) - 1;
                        var argumentCount_1 = countBinaryExpressionParameters(highestBinary);
                        return contextualType_1 && { contextualType: contextualType_1, argumentIndex: argumentIndex_1, argumentCount: argumentCount_1, argumentsSpan: ts.createTextSpanFromNode(parent) };
                    }
                    default:
                        return undefined;
                }
            }
            // The type of a function type node has a symbol at that node, but it's better to use the symbol for a parameter or type alias.
            function chooseBetterSymbol(s) {
                return s.name === "__type" /* Type */
                    ? ts.firstDefined(s.declarations, function (d) { return ts.isFunctionTypeNode(d) ? d.parent.symbol : undefined; }) || s
                    : s;
            }
            function getArgumentIndex(argumentsList, node) {
                // The list we got back can include commas.  In the presence of errors it may
                // also just have nodes without commas.  For example "Foo(a b c)" will have 3
                // args without commas. We want to find what index we're at.  So we count
                // forward until we hit ourselves, only incrementing the index if it isn't a
                // comma.
                //
                // Note: the subtlety around trailing commas (in getArgumentCount) does not apply
                // here.  That's because we're only walking forward until we hit the node we're
                // on.  In that case, even if we're after the trailing comma, we'll still see
                // that trailing comma in the list, and we'll have generated the appropriate
                // arg index.
                var argumentIndex = 0;
                for (var _i = 0, _a = argumentsList.getChildren(); _i < _a.length; _i++) {
                    var child = _a[_i];
                    if (child === node) {
                        break;
                    }
                    if (child.kind !== 27 /* CommaToken */) {
                        argumentIndex++;
                    }
                }
                return argumentIndex;
            }
            function getArgumentCount(argumentsList) {
                // The argument count for a list is normally the number of non-comma children it has.
                // For example, if you have "Foo(a,b)" then there will be three children of the arg
                // list 'a' '<comma>' 'b'.  So, in this case the arg count will be 2.  However, there
                // is a small subtlety.  If you have "Foo(a,)", then the child list will just have
                // 'a' '<comma>'.  So, in the case where the last child is a comma, we increase the
                // arg count by one to compensate.
                //
                // Note: this subtlety only applies to the last comma.  If you had "Foo(a,," then
                // we'll have: 'a' '<comma>' '<missing>'
                // That will give us 2 non-commas.  We then add one for the last comma, giving us an
                // arg count of 3.
                var listChildren = argumentsList.getChildren();
                var argumentCount = ts.countWhere(listChildren, function (arg) { return arg.kind !== 27 /* CommaToken */; });
                if (listChildren.length > 0 && ts.last(listChildren).kind === 27 /* CommaToken */) {
                    argumentCount++;
                }
                return argumentCount;
            }
            // spanIndex is either the index for a given template span.
            // This does not give appropriate results for a NoSubstitutionTemplateLiteral
            function getArgumentIndexForTemplatePiece(spanIndex, node, position, sourceFile) {
                // Because the TemplateStringsArray is the first argument, we have to offset each substitution expression by 1.
                // There are three cases we can encounter:
                //      1. We are precisely in the template literal (argIndex = 0).
                //      2. We are in or to the right of the substitution expression (argIndex = spanIndex + 1).
                //      3. We are directly to the right of the template literal, but because we look for the token on the left,
                //          not enough to put us in the substitution expression; we should consider ourselves part of
                //          the *next* span's expression by offsetting the index (argIndex = (spanIndex + 1) + 1).
                //
                /* eslint-disable no-double-space */
                // Example: f  `# abcd $#{#  1 + 1#  }# efghi ${ #"#hello"#  }  #  `
                //              ^       ^ ^       ^   ^          ^ ^      ^     ^
                // Case:        1       1 3       2   1          3 2      2     1
                /* eslint-enable no-double-space */
                ts.Debug.assert(position >= node.getStart(), "Assumed 'position' could not occur before node.");
                if (ts.isTemplateLiteralToken(node)) {
                    if (ts.isInsideTemplateLiteral(node, position, sourceFile)) {
                        return 0;
                    }
                    return spanIndex + 2;
                }
                return spanIndex + 1;
            }
            function getArgumentListInfoForTemplate(tagExpression, argumentIndex, sourceFile) {
                // argumentCount is either 1 or (numSpans + 1) to account for the template strings array argument.
                var argumentCount = ts.isNoSubstitutionTemplateLiteral(tagExpression.template) ? 1 : tagExpression.template.templateSpans.length + 1;
                if (argumentIndex !== 0) {
                    ts.Debug.assertLessThan(argumentIndex, argumentCount);
                }
                return {
                    isTypeParameterList: false,
                    invocation: { kind: 0 /* Call */, node: tagExpression },
                    argumentsSpan: getApplicableSpanForTaggedTemplate(tagExpression, sourceFile),
                    argumentIndex: argumentIndex,
                    argumentCount: argumentCount
                };
            }
            function getApplicableSpanForArguments(argumentsList, sourceFile) {
                // We use full start and skip trivia on the end because we want to include trivia on
                // both sides. For example,
                //
                //    foo(   /*comment */     a, b, c      /*comment*/     )
                //        |                                               |
                //
                // The applicable span is from the first bar to the second bar (inclusive,
                // but not including parentheses)
                var applicableSpanStart = argumentsList.getFullStart();
                var applicableSpanEnd = ts.skipTrivia(sourceFile.text, argumentsList.getEnd(), /*stopAfterLineBreak*/ false);
                return ts.createTextSpan(applicableSpanStart, applicableSpanEnd - applicableSpanStart);
            }
            function getApplicableSpanForTaggedTemplate(taggedTemplate, sourceFile) {
                var template = taggedTemplate.template;
                var applicableSpanStart = template.getStart();
                var applicableSpanEnd = template.getEnd();
                // We need to adjust the end position for the case where the template does not have a tail.
                // Otherwise, we will not show signature help past the expression.
                // For example,
                //
                //      ` ${ 1 + 1 foo(10)
                //       |       |
                // This is because a Missing node has no width. However, what we actually want is to include trivia
                // leading up to the next token in case the user is about to type in a TemplateMiddle or TemplateTail.
                if (template.kind === 221 /* TemplateExpression */) {
                    var lastSpan = ts.last(template.templateSpans);
                    if (lastSpan.literal.getFullWidth() === 0) {
                        applicableSpanEnd = ts.skipTrivia(sourceFile.text, applicableSpanEnd, /*stopAfterLineBreak*/ false);
                    }
                }
                return ts.createTextSpan(applicableSpanStart, applicableSpanEnd - applicableSpanStart);
            }
            function getContainingArgumentInfo(node, position, sourceFile, checker, isManuallyInvoked) {
                var _loop_8 = function (n) {
                    // If the node is not a subspan of its parent, this is a big problem.
                    // There have been crashes that might be caused by this violation.
                    ts.Debug.assert(ts.rangeContainsRange(n.parent, n), "Not a subspan", function () { return "Child: " + ts.Debug.formatSyntaxKind(n.kind) + ", parent: " + ts.Debug.formatSyntaxKind(n.parent.kind); });
                    var argumentInfo = getImmediatelyContainingArgumentOrContextualParameterInfo(n, position, sourceFile, checker);
                    if (argumentInfo) {
                        return { value: argumentInfo };
                    }
                };
                for (var n = node; !ts.isSourceFile(n) && (isManuallyInvoked || !ts.isBlock(n)); n = n.parent) {
                    var state_3 = _loop_8(n);
                    if (typeof state_3 === "object")
                        return state_3.value;
                }
                return undefined;
            }
            function getChildListThatStartsWithOpenerToken(parent, openerToken, sourceFile) {
                var children = parent.getChildren(sourceFile);
                var indexOfOpenerToken = children.indexOf(openerToken);
                ts.Debug.assert(indexOfOpenerToken >= 0 && children.length > indexOfOpenerToken + 1);
                return children[indexOfOpenerToken + 1];
            }
            function getExpressionFromInvocation(invocation) {
                return invocation.kind === 0 /* Call */ ? ts.getInvokedExpression(invocation.node) : invocation.called;
            }
            function getEnclosingDeclarationFromInvocation(invocation) {
                return invocation.kind === 0 /* Call */ ? invocation.node : invocation.kind === 1 /* TypeArgs */ ? invocation.called : invocation.node;
            }
            var signatureHelpNodeBuilderFlags = 8192 /* OmitParameterModifiers */ | 70221824 /* IgnoreErrors */ | 16384 /* UseAliasDefinedOutsideCurrentScope */;
            function createSignatureHelpItems(candidates, resolvedSignature, _a, sourceFile, typeChecker, useFullPrefix) {
                var _b;
                var isTypeParameterList = _a.isTypeParameterList, argumentCount = _a.argumentCount, applicableSpan = _a.argumentsSpan, invocation = _a.invocation, argumentIndex = _a.argumentIndex;
                var enclosingDeclaration = getEnclosingDeclarationFromInvocation(invocation);
                var callTargetSymbol = invocation.kind === 2 /* Contextual */ ? invocation.symbol : (typeChecker.getSymbolAtLocation(getExpressionFromInvocation(invocation)) || useFullPrefix && ((_b = resolvedSignature.declaration) === null || _b === void 0 ? void 0 : _b.symbol));
                var callTargetDisplayParts = callTargetSymbol ? ts.symbolToDisplayParts(typeChecker, callTargetSymbol, useFullPrefix ? sourceFile : undefined, /*meaning*/ undefined) : ts.emptyArray;
                var items = ts.map(candidates, function (candidateSignature) { return getSignatureHelpItem(candidateSignature, callTargetDisplayParts, isTypeParameterList, typeChecker, enclosingDeclaration, sourceFile); });
                if (argumentIndex !== 0) {
                    ts.Debug.assertLessThan(argumentIndex, argumentCount);
                }
                var selectedItemIndex = 0;
                var itemsSeen = 0;
                for (var i = 0; i < items.length; i++) {
                    var item = items[i];
                    if (candidates[i] === resolvedSignature) {
                        selectedItemIndex = itemsSeen;
                        if (item.length > 1) {
                            // check to see if any items in the list better match than the first one, as the checker isn't filtering the nested lists
                            // (those come from tuple parameter expansion)
                            var count = 0;
                            for (var _i = 0, item_1 = item; _i < item_1.length; _i++) {
                                var i_1 = item_1[_i];
                                if (i_1.isVariadic || i_1.parameters.length >= argumentCount) {
                                    selectedItemIndex = itemsSeen + count;
                                    break;
                                }
                                count++;
                            }
                        }
                    }
                    itemsSeen += item.length;
                }
                ts.Debug.assert(selectedItemIndex !== -1); // If candidates is non-empty it should always include bestSignature. We check for an empty candidates before calling this function.
                var help = { items: ts.flatMapToMutable(items, ts.identity), applicableSpan: applicableSpan, selectedItemIndex: selectedItemIndex, argumentIndex: argumentIndex, argumentCount: argumentCount };
                var selected = help.items[selectedItemIndex];
                if (selected.isVariadic) {
                    var firstRest = ts.findIndex(selected.parameters, function (p) { return !!p.isRest; });
                    if (-1 < firstRest && firstRest < selected.parameters.length - 1) {
                        // We don't have any code to get this correct; instead, don't highlight a current parameter AT ALL
                        help.argumentIndex = selected.parameters.length;
                    }
                    else {
                        help.argumentIndex = Math.min(help.argumentIndex, selected.parameters.length - 1);
                    }
                }
                return help;
            }
            function createTypeHelpItems(symbol, _a, sourceFile, checker) {
                var argumentCount = _a.argumentCount, applicableSpan = _a.argumentsSpan, invocation = _a.invocation, argumentIndex = _a.argumentIndex;
                var typeParameters = checker.getLocalTypeParametersOfClassOrInterfaceOrTypeAlias(symbol);
                if (!typeParameters)
                    return undefined;
                var items = [getTypeHelpItem(symbol, typeParameters, checker, getEnclosingDeclarationFromInvocation(invocation), sourceFile)];
                return { items: items, applicableSpan: applicableSpan, selectedItemIndex: 0, argumentIndex: argumentIndex, argumentCount: argumentCount };
            }
            function getTypeHelpItem(symbol, typeParameters, checker, enclosingDeclaration, sourceFile) {
                var typeSymbolDisplay = ts.symbolToDisplayParts(checker, symbol);
                var printer = ts.createPrinter({ removeComments: true });
                var parameters = typeParameters.map(function (t) { return createSignatureHelpParameterForTypeParameter(t, checker, enclosingDeclaration, sourceFile, printer); });
                var documentation = symbol.getDocumentationComment(checker);
                var tags = symbol.getJsDocTags(checker);
                var prefixDisplayParts = __spreadArray(__spreadArray([], typeSymbolDisplay, true), [ts.punctuationPart(29 /* LessThanToken */)], false);
                return { isVariadic: false, prefixDisplayParts: prefixDisplayParts, suffixDisplayParts: [ts.punctuationPart(31 /* GreaterThanToken */)], separatorDisplayParts: separatorDisplayParts, parameters: parameters, documentation: documentation, tags: tags };
            }
            var separatorDisplayParts = [ts.punctuationPart(27 /* CommaToken */), ts.spacePart()];
            function getSignatureHelpItem(candidateSignature, callTargetDisplayParts, isTypeParameterList, checker, enclosingDeclaration, sourceFile) {
                var infos = (isTypeParameterList ? itemInfoForTypeParameters : itemInfoForParameters)(candidateSignature, checker, enclosingDeclaration, sourceFile);
                return ts.map(infos, function (_a) {
                    var isVariadic = _a.isVariadic, parameters = _a.parameters, prefix = _a.prefix, suffix = _a.suffix;
                    var prefixDisplayParts = __spreadArray(__spreadArray([], callTargetDisplayParts, true), prefix, true);
                    var suffixDisplayParts = __spreadArray(__spreadArray([], suffix, true), returnTypeToDisplayParts(candidateSignature, enclosingDeclaration, checker), true);
                    var documentation = candidateSignature.getDocumentationComment(checker);
                    var tags = candidateSignature.getJsDocTags();
                    return { isVariadic: isVariadic, prefixDisplayParts: prefixDisplayParts, suffixDisplayParts: suffixDisplayParts, separatorDisplayParts: separatorDisplayParts, parameters: parameters, documentation: documentation, tags: tags };
                });
            }
            function returnTypeToDisplayParts(candidateSignature, enclosingDeclaration, checker) {
                return ts.mapToDisplayParts(function (writer) {
                    writer.writePunctuation(":");
                    writer.writeSpace(" ");
                    var predicate = checker.getTypePredicateOfSignature(candidateSignature);
                    if (predicate) {
                        checker.writeTypePredicate(predicate, enclosingDeclaration, /*flags*/ undefined, writer);
                    }
                    else {
                        checker.writeType(checker.getReturnTypeOfSignature(candidateSignature), enclosingDeclaration, /*flags*/ undefined, writer);
                    }
                });
            }
            function itemInfoForTypeParameters(candidateSignature, checker, enclosingDeclaration, sourceFile) {
                var typeParameters = (candidateSignature.target || candidateSignature).typeParameters;
                var printer = ts.createPrinter({ removeComments: true });
                var parameters = (typeParameters || ts.emptyArray).map(function (t) { return createSignatureHelpParameterForTypeParameter(t, checker, enclosingDeclaration, sourceFile, printer); });
                var thisParameter = candidateSignature.thisParameter ? [checker.symbolToParameterDeclaration(candidateSignature.thisParameter, enclosingDeclaration, signatureHelpNodeBuilderFlags)] : [];
                return checker.getExpandedParameters(candidateSignature).map(function (paramList) {
                    var params = ts.factory.createNodeArray(__spreadArray(__spreadArray([], thisParameter, true), ts.map(paramList, function (param) { return checker.symbolToParameterDeclaration(param, enclosingDeclaration, signatureHelpNodeBuilderFlags); }), true));
                    var parameterParts = ts.mapToDisplayParts(function (writer) {
                        printer.writeList(2576 /* CallExpressionArguments */, params, sourceFile, writer);
                    });
                    return { isVariadic: false, parameters: parameters, prefix: [ts.punctuationPart(29 /* LessThanToken */)], suffix: __spreadArray([ts.punctuationPart(31 /* GreaterThanToken */)], parameterParts, true) };
                });
            }
            function itemInfoForParameters(candidateSignature, checker, enclosingDeclaration, sourceFile) {
                var printer = ts.createPrinter({ removeComments: true });
                var typeParameterParts = ts.mapToDisplayParts(function (writer) {
                    if (candidateSignature.typeParameters && candidateSignature.typeParameters.length) {
                        var args = ts.factory.createNodeArray(candidateSignature.typeParameters.map(function (p) { return checker.typeParameterToDeclaration(p, enclosingDeclaration, signatureHelpNodeBuilderFlags); }));
                        printer.writeList(53776 /* TypeParameters */, args, sourceFile, writer);
                    }
                });
                var lists = checker.getExpandedParameters(candidateSignature);
                var isVariadic = !checker.hasEffectiveRestParameter(candidateSignature) ? function (_) { return false; }
                    : lists.length === 1 ? function (_) { return true; }
                        : function (pList) { return !!(pList.length && pList[pList.length - 1].checkFlags & 32768 /* RestParameter */); };
                return lists.map(function (parameterList) {
                    return ({
                        isVariadic: isVariadic(parameterList),
                        parameters: parameterList.map(function (p) { return createSignatureHelpParameterForParameter(p, checker, enclosingDeclaration, sourceFile, printer); }),
                        prefix: __spreadArray(__spreadArray([], typeParameterParts, true), [ts.punctuationPart(20 /* OpenParenToken */)], false),
                        suffix: [ts.punctuationPart(21 /* CloseParenToken */)]
                    });
                });
            }
            function createSignatureHelpParameterForParameter(parameter, checker, enclosingDeclaration, sourceFile, printer) {
                var displayParts = ts.mapToDisplayParts(function (writer) {
                    var param = checker.symbolToParameterDeclaration(parameter, enclosingDeclaration, signatureHelpNodeBuilderFlags);
                    printer.writeNode(4 /* Unspecified */, param, sourceFile, writer);
                });
                var isOptional = checker.isOptionalParameter(parameter.valueDeclaration);
                var isRest = !!(parameter.checkFlags & 32768 /* RestParameter */);
                return { name: parameter.name, documentation: parameter.getDocumentationComment(checker), displayParts: displayParts, isOptional: isOptional, isRest: isRest };
            }
            function createSignatureHelpParameterForTypeParameter(typeParameter, checker, enclosingDeclaration, sourceFile, printer) {
                var displayParts = ts.mapToDisplayParts(function (writer) {
                    var param = checker.typeParameterToDeclaration(typeParameter, enclosingDeclaration, signatureHelpNodeBuilderFlags);
                    printer.writeNode(4 /* Unspecified */, param, sourceFile, writer);
                });
                return { name: typeParameter.symbol.name, documentation: typeParameter.symbol.getDocumentationComment(checker), displayParts: displayParts, isOptional: false, isRest: false };
            }
        })(SignatureHelp = ts.SignatureHelp || (ts.SignatureHelp = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var InlayHints;
        (function (InlayHints) {
            var maxHintsLength = 30;
            var leadingParameterNameCommentRegexFactory = function (name) {
                return new RegExp("^\\s?/\\*\\*?\\s?" + name + "\\s?\\*\\/\\s?$");
            };
            function shouldShowParameterNameHints(preferences) {
                return preferences.includeInlayParameterNameHints === "literals" || preferences.includeInlayParameterNameHints === "all";
            }
            function shouldShowLiteralParameterNameHintsOnly(preferences) {
                return preferences.includeInlayParameterNameHints === "literals";
            }
            function provideInlayHints(context) {
                var file = context.file, program = context.program, span = context.span, cancellationToken = context.cancellationToken, preferences = context.preferences;
                var sourceFileText = file.text;
                var compilerOptions = program.getCompilerOptions();
                var checker = program.getTypeChecker();
                var result = [];
                visitor(file);
                return result;
                function visitor(node) {
                    if (!node || node.getFullWidth() === 0) {
                        return;
                    }
                    switch (node.kind) {
                        case 259 /* ModuleDeclaration */:
                        case 255 /* ClassDeclaration */:
                        case 256 /* InterfaceDeclaration */:
                        case 254 /* FunctionDeclaration */:
                        case 224 /* ClassExpression */:
                        case 211 /* FunctionExpression */:
                        case 167 /* MethodDeclaration */:
                        case 212 /* ArrowFunction */:
                            cancellationToken.throwIfCancellationRequested();
                    }
                    if (!ts.textSpanIntersectsWith(span, node.pos, node.getFullWidth())) {
                        return;
                    }
                    if (ts.isTypeNode(node)) {
                        return;
                    }
                    if (preferences.includeInlayVariableTypeHints && ts.isVariableDeclaration(node)) {
                        visitVariableLikeDeclaration(node);
                    }
                    else if (preferences.includeInlayPropertyDeclarationTypeHints && ts.isPropertyDeclaration(node)) {
                        visitVariableLikeDeclaration(node);
                    }
                    else if (preferences.includeInlayEnumMemberValueHints && ts.isEnumMember(node)) {
                        visitEnumMember(node);
                    }
                    else if (shouldShowParameterNameHints(preferences) && (ts.isCallExpression(node) || ts.isNewExpression(node))) {
                        visitCallOrNewExpression(node);
                    }
                    else {
                        if (preferences.includeInlayFunctionParameterTypeHints && ts.isFunctionLikeDeclaration(node) && ts.hasContextSensitiveParameters(node)) {
                            visitFunctionLikeForParameterType(node);
                        }
                        if (preferences.includeInlayFunctionLikeReturnTypeHints && isSignatureSupportingReturnAnnotation(node)) {
                            visitFunctionDeclarationLikeForReturnType(node);
                        }
                    }
                    return ts.forEachChild(node, visitor);
                }
                function isSignatureSupportingReturnAnnotation(node) {
                    return ts.isArrowFunction(node) || ts.isFunctionExpression(node) || ts.isFunctionDeclaration(node) || ts.isMethodDeclaration(node) || ts.isGetAccessorDeclaration(node);
                }
                function addParameterHints(text, position, isFirstVariadicArgument) {
                    result.push({
                        text: "" + (isFirstVariadicArgument ? "..." : "") + truncation(text, maxHintsLength) + ":",
                        position: position,
                        kind: "Parameter" /* Parameter */,
                        whitespaceAfter: true,
                    });
                }
                function addTypeHints(text, position) {
                    result.push({
                        text: ": " + truncation(text, maxHintsLength),
                        position: position,
                        kind: "Type" /* Type */,
                        whitespaceBefore: true,
                    });
                }
                function addEnumMemberValueHints(text, position) {
                    result.push({
                        text: "= " + truncation(text, maxHintsLength),
                        position: position,
                        kind: "Enum" /* Enum */,
                        whitespaceBefore: true,
                    });
                }
                function visitEnumMember(member) {
                    if (member.initializer) {
                        return;
                    }
                    var enumValue = checker.getConstantValue(member);
                    if (enumValue !== undefined) {
                        addEnumMemberValueHints(enumValue.toString(), member.end);
                    }
                }
                function isModuleReferenceType(type) {
                    return type.symbol && (type.symbol.flags & 1536 /* Module */);
                }
                function visitVariableLikeDeclaration(decl) {
                    if (!decl.initializer || ts.isBindingPattern(decl.name)) {
                        return;
                    }
                    var effectiveTypeAnnotation = ts.getEffectiveTypeAnnotationNode(decl);
                    if (effectiveTypeAnnotation) {
                        return;
                    }
                    var declarationType = checker.getTypeAtLocation(decl);
                    if (isModuleReferenceType(declarationType)) {
                        return;
                    }
                    var typeDisplayString = printTypeInSingleLine(declarationType);
                    if (typeDisplayString) {
                        addTypeHints(typeDisplayString, decl.name.end);
                    }
                }
                function visitCallOrNewExpression(expr) {
                    var args = expr.arguments;
                    if (!args || !args.length) {
                        return;
                    }
                    var candidates = [];
                    var signature = checker.getResolvedSignatureForSignatureHelp(expr, candidates);
                    if (!signature || !candidates.length) {
                        return;
                    }
                    for (var i = 0; i < args.length; ++i) {
                        var originalArg = args[i];
                        var arg = ts.skipParentheses(originalArg);
                        if (shouldShowLiteralParameterNameHintsOnly(preferences) && !isHintableExpression(arg)) {
                            continue;
                        }
                        var identifierNameInfo = checker.getParameterIdentifierNameAtPosition(signature, i);
                        if (identifierNameInfo) {
                            var parameterName = identifierNameInfo[0], isFirstVariadicArgument = identifierNameInfo[1];
                            var isParameterNameNotSameAsArgument = preferences.includeInlayParameterNameHintsWhenArgumentMatchesName || !identifierOrAccessExpressionPostfixMatchesParameterName(arg, parameterName);
                            if (!isParameterNameNotSameAsArgument && !isFirstVariadicArgument) {
                                continue;
                            }
                            var name = ts.unescapeLeadingUnderscores(parameterName);
                            if (leadingCommentsContainsParameterName(arg, name)) {
                                continue;
                            }
                            addParameterHints(name, originalArg.getStart(), isFirstVariadicArgument);
                        }
                    }
                }
                function identifierOrAccessExpressionPostfixMatchesParameterName(expr, parameterName) {
                    if (ts.isIdentifier(expr)) {
                        return expr.text === parameterName;
                    }
                    if (ts.isPropertyAccessExpression(expr)) {
                        return expr.name.text === parameterName;
                    }
                    return false;
                }
                function leadingCommentsContainsParameterName(node, name) {
                    if (!ts.isIdentifierText(name, compilerOptions.target, ts.getLanguageVariant(file.scriptKind))) {
                        return false;
                    }
                    var ranges = ts.getLeadingCommentRanges(sourceFileText, node.pos);
                    if (!(ranges === null || ranges === void 0 ? void 0 : ranges.length)) {
                        return false;
                    }
                    var regex = leadingParameterNameCommentRegexFactory(name);
                    return ts.some(ranges, function (range) { return regex.test(sourceFileText.substring(range.pos, range.end)); });
                }
                function isHintableExpression(node) {
                    switch (node.kind) {
                        case 217 /* PrefixUnaryExpression */:
                            return ts.isLiteralExpression(node.operand);
                        case 110 /* TrueKeyword */:
                        case 95 /* FalseKeyword */:
                        case 212 /* ArrowFunction */:
                        case 211 /* FunctionExpression */:
                        case 203 /* ObjectLiteralExpression */:
                        case 202 /* ArrayLiteralExpression */:
                            return true;
                    }
                    return ts.isLiteralExpression(node);
                }
                function visitFunctionDeclarationLikeForReturnType(decl) {
                    if (ts.isArrowFunction(decl)) {
                        if (!ts.findChildOfKind(decl, 20 /* OpenParenToken */, file)) {
                            return;
                        }
                    }
                    var effectiveTypeAnnotation = ts.getEffectiveReturnTypeNode(decl);
                    if (effectiveTypeAnnotation || !decl.body) {
                        return;
                    }
                    var signature = checker.getSignatureFromDeclaration(decl);
                    if (!signature) {
                        return;
                    }
                    var returnType = checker.getReturnTypeOfSignature(signature);
                    if (isModuleReferenceType(returnType)) {
                        return;
                    }
                    var typeDisplayString = printTypeInSingleLine(returnType);
                    if (!typeDisplayString) {
                        return;
                    }
                    addTypeHints(typeDisplayString, getTypeAnnotationPosition(decl));
                }
                function getTypeAnnotationPosition(decl) {
                    var closeParenToken = ts.findChildOfKind(decl, 21 /* CloseParenToken */, file);
                    if (closeParenToken) {
                        return closeParenToken.end;
                    }
                    return decl.parameters.end;
                }
                function visitFunctionLikeForParameterType(node) {
                    var signature = checker.getSignatureFromDeclaration(node);
                    if (!signature) {
                        return;
                    }
                    for (var i = 0; i < node.parameters.length && i < signature.parameters.length; ++i) {
                        var param = node.parameters[i];
                        var effectiveTypeAnnotation = ts.getEffectiveTypeAnnotationNode(param);
                        if (effectiveTypeAnnotation) {
                            continue;
                        }
                        var typeDisplayString = getParameterDeclarationTypeDisplayString(signature.parameters[i]);
                        if (!typeDisplayString) {
                            continue;
                        }
                        addTypeHints(typeDisplayString, param.end);
                    }
                }
                function getParameterDeclarationTypeDisplayString(symbol) {
                    var valueDeclaration = symbol.valueDeclaration;
                    if (!valueDeclaration || !ts.isParameter(valueDeclaration)) {
                        return undefined;
                    }
                    var signatureParamType = checker.getTypeOfSymbolAtLocation(symbol, valueDeclaration);
                    if (isModuleReferenceType(signatureParamType)) {
                        return undefined;
                    }
                    return printTypeInSingleLine(signatureParamType);
                }
                function truncation(text, maxLength) {
                    if (text.length > maxLength) {
                        return text.substr(0, maxLength - "...".length) + "...";
                    }
                    return text;
                }
                function printTypeInSingleLine(type) {
                    var flags = 70221824 /* IgnoreErrors */ | 1048576 /* AllowUniqueESSymbolType */ | 16384 /* UseAliasDefinedOutsideCurrentScope */;
                    var options = { removeComments: true };
                    var printer = ts.createPrinter(options);
                    return ts.usingSingleLineStringWriter(function (writer) {
                        var typeNode = checker.typeToTypeNode(type, /*enclosingDeclaration*/ undefined, flags, writer);
                        ts.Debug.assertIsDefined(typeNode, "should always get typenode");
                        printer.writeNode(4 /* Unspecified */, typeNode, /*sourceFile*/ file, writer);
                    });
                }
            }
            InlayHints.provideInlayHints = provideInlayHints;
        })(InlayHints = ts.InlayHints || (ts.InlayHints = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var base64UrlRegExp = /^data:(?:application\/json(?:;charset=[uU][tT][fF]-8);base64,([A-Za-z0-9+\/=]+)$)?/;
        function getSourceMapper(host) {
            var getCanonicalFileName = ts.createGetCanonicalFileName(host.useCaseSensitiveFileNames());
            var currentDirectory = host.getCurrentDirectory();
            var sourceFileLike = new ts.Map();
            var documentPositionMappers = new ts.Map();
            return { tryGetSourcePosition: tryGetSourcePosition, tryGetGeneratedPosition: tryGetGeneratedPosition, toLineColumnOffset: toLineColumnOffset, clearCache: clearCache };
            function toPath(fileName) {
                return ts.toPath(fileName, currentDirectory, getCanonicalFileName);
            }
            function getDocumentPositionMapper(generatedFileName, sourceFileName) {
                var path = toPath(generatedFileName);
                var value = documentPositionMappers.get(path);
                if (value)
                    return value;
                var mapper;
                if (host.getDocumentPositionMapper) {
                    mapper = host.getDocumentPositionMapper(generatedFileName, sourceFileName);
                }
                else if (host.readFile) {
                    var file = getSourceFileLike(generatedFileName);
                    mapper = file && ts.getDocumentPositionMapper({ getSourceFileLike: getSourceFileLike, getCanonicalFileName: getCanonicalFileName, log: function (s) { return host.log(s); } }, generatedFileName, ts.getLineInfo(file.text, ts.getLineStarts(file)), function (f) { return !host.fileExists || host.fileExists(f) ? host.readFile(f) : undefined; });
                }
                documentPositionMappers.set(path, mapper || ts.identitySourceMapConsumer);
                return mapper || ts.identitySourceMapConsumer;
            }
            function tryGetSourcePosition(info) {
                if (!ts.isDeclarationFileName(info.fileName))
                    return undefined;
                var file = getSourceFile(info.fileName);
                if (!file)
                    return undefined;
                var newLoc = getDocumentPositionMapper(info.fileName).getSourcePosition(info);
                return !newLoc || newLoc === info ? undefined : tryGetSourcePosition(newLoc) || newLoc;
            }
            function tryGetGeneratedPosition(info) {
                if (ts.isDeclarationFileName(info.fileName))
                    return undefined;
                var sourceFile = getSourceFile(info.fileName);
                if (!sourceFile)
                    return undefined;
                var program = host.getProgram();
                // If this is source file of project reference source (instead of redirect) there is no generated position
                if (program.isSourceOfProjectReferenceRedirect(sourceFile.fileName)) {
                    return undefined;
                }
                var options = program.getCompilerOptions();
                var outPath = ts.outFile(options);
                var declarationPath = outPath ?
                    ts.removeFileExtension(outPath) + ".d.ts" /* Dts */ :
                    ts.getDeclarationEmitOutputFilePathWorker(info.fileName, program.getCompilerOptions(), currentDirectory, program.getCommonSourceDirectory(), getCanonicalFileName);
                if (declarationPath === undefined)
                    return undefined;
                var newLoc = getDocumentPositionMapper(declarationPath, info.fileName).getGeneratedPosition(info);
                return newLoc === info ? undefined : newLoc;
            }
            function getSourceFile(fileName) {
                var program = host.getProgram();
                if (!program)
                    return undefined;
                var path = toPath(fileName);
                // file returned here could be .d.ts when asked for .ts file if projectReferences and module resolution created this source file
                var file = program.getSourceFileByPath(path);
                return file && file.resolvedPath === path ? file : undefined;
            }
            function getOrCreateSourceFileLike(fileName) {
                var path = toPath(fileName);
                var fileFromCache = sourceFileLike.get(path);
                if (fileFromCache !== undefined)
                    return fileFromCache ? fileFromCache : undefined;
                if (!host.readFile || host.fileExists && !host.fileExists(path)) {
                    sourceFileLike.set(path, false);
                    return undefined;
                }
                // And failing that, check the disk
                var text = host.readFile(path);
                var file = text ? createSourceFileLike(text) : false;
                sourceFileLike.set(path, file);
                return file ? file : undefined;
            }
            // This can be called from source mapper in either source program or program that includes generated file
            function getSourceFileLike(fileName) {
                return !host.getSourceFileLike ?
                    getSourceFile(fileName) || getOrCreateSourceFileLike(fileName) :
                    host.getSourceFileLike(fileName);
            }
            function toLineColumnOffset(fileName, position) {
                var file = getSourceFileLike(fileName); // TODO: GH#18217
                return file.getLineAndCharacterOfPosition(position);
            }
            function clearCache() {
                sourceFileLike.clear();
                documentPositionMappers.clear();
            }
        }
        ts.getSourceMapper = getSourceMapper;
        function getDocumentPositionMapper(host, generatedFileName, generatedFileLineInfo, readMapFile) {
            var mapFileName = ts.tryGetSourceMappingURL(generatedFileLineInfo);
            if (mapFileName) {
                var match = base64UrlRegExp.exec(mapFileName);
                if (match) {
                    if (match[1]) {
                        var base64Object = match[1];
                        return convertDocumentToSourceMapper(host, ts.base64decode(ts.sys, base64Object), generatedFileName);
                    }
                    // Not a data URL we can parse, skip it
                    mapFileName = undefined;
                }
            }
            var possibleMapLocations = [];
            if (mapFileName) {
                possibleMapLocations.push(mapFileName);
            }
            possibleMapLocations.push(generatedFileName + ".map");
            var originalMapFileName = mapFileName && ts.getNormalizedAbsolutePath(mapFileName, ts.getDirectoryPath(generatedFileName));
            for (var _i = 0, possibleMapLocations_1 = possibleMapLocations; _i < possibleMapLocations_1.length; _i++) {
                var location = possibleMapLocations_1[_i];
                var mapFileName_1 = ts.getNormalizedAbsolutePath(location, ts.getDirectoryPath(generatedFileName));
                var mapFileContents = readMapFile(mapFileName_1, originalMapFileName);
                if (ts.isString(mapFileContents)) {
                    return convertDocumentToSourceMapper(host, mapFileContents, mapFileName_1);
                }
                if (mapFileContents !== undefined) {
                    return mapFileContents || undefined;
                }
            }
            return undefined;
        }
        ts.getDocumentPositionMapper = getDocumentPositionMapper;
        function convertDocumentToSourceMapper(host, contents, mapFileName) {
            var map = ts.tryParseRawSourceMap(contents);
            if (!map || !map.sources || !map.file || !map.mappings) {
                // obviously invalid map
                return undefined;
            }
            // Dont support sourcemaps that contain inlined sources
            if (map.sourcesContent && map.sourcesContent.some(ts.isString))
                return undefined;
            return ts.createDocumentPositionMapper(host, map, mapFileName);
        }
        function createSourceFileLike(text, lineMap) {
            return {
                text: text,
                lineMap: lineMap,
                getLineAndCharacterOfPosition: function (pos) {
                    return ts.computeLineAndCharacterOfPosition(ts.getLineStarts(this), pos);
                }
            };
        }
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var visitedNestedConvertibleFunctions = new ts.Map();
        function computeSuggestionDiagnostics(sourceFile, program, cancellationToken) {
            program.getSemanticDiagnostics(sourceFile, cancellationToken);
            var diags = [];
            var checker = program.getTypeChecker();
            if (sourceFile.commonJsModuleIndicator &&
                (ts.programContainsEs6Modules(program) || ts.compilerOptionsIndicateEs6Modules(program.getCompilerOptions())) &&
                containsTopLevelCommonjs(sourceFile)) {
                diags.push(ts.createDiagnosticForNode(getErrorNodeFromCommonJsIndicator(sourceFile.commonJsModuleIndicator), ts.Diagnostics.File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module));
            }
            var isJsFile = ts.isSourceFileJS(sourceFile);
            visitedNestedConvertibleFunctions.clear();
            check(sourceFile);
            if (ts.getAllowSyntheticDefaultImports(program.getCompilerOptions())) {
                for (var _i = 0, _a = sourceFile.imports; _i < _a.length; _i++) {
                    var moduleSpecifier = _a[_i];
                    var importNode = ts.importFromModuleSpecifier(moduleSpecifier);
                    var name = importNameForConvertToDefaultImport(importNode);
                    if (!name)
                        continue;
                    var module_1 = ts.getResolvedModule(sourceFile, moduleSpecifier.text);
                    var resolvedFile = module_1 && program.getSourceFile(module_1.resolvedFileName);
                    if (resolvedFile && resolvedFile.externalModuleIndicator && ts.isExportAssignment(resolvedFile.externalModuleIndicator) && resolvedFile.externalModuleIndicator.isExportEquals) {
                        diags.push(ts.createDiagnosticForNode(name, ts.Diagnostics.Import_may_be_converted_to_a_default_import));
                    }
                }
            }
            ts.addRange(diags, sourceFile.bindSuggestionDiagnostics);
            ts.addRange(diags, program.getSuggestionDiagnostics(sourceFile, cancellationToken));
            return diags.sort(function (d1, d2) { return d1.start - d2.start; });
            function check(node) {
                if (isJsFile) {
                    if (canBeConvertedToClass(node, checker)) {
                        diags.push(ts.createDiagnosticForNode(ts.isVariableDeclaration(node.parent) ? node.parent.name : node, ts.Diagnostics.This_constructor_function_may_be_converted_to_a_class_declaration));
                    }
                }
                else {
                    if (ts.isVariableStatement(node) &&
                        node.parent === sourceFile &&
                        node.declarationList.flags & 2 /* Const */ &&
                        node.declarationList.declarations.length === 1) {
                        var init = node.declarationList.declarations[0].initializer;
                        if (init && ts.isRequireCall(init, /*checkArgumentIsStringLiteralLike*/ true)) {
                            diags.push(ts.createDiagnosticForNode(init, ts.Diagnostics.require_call_may_be_converted_to_an_import));
                        }
                    }
                    if (ts.codefix.parameterShouldGetTypeFromJSDoc(node)) {
                        diags.push(ts.createDiagnosticForNode(node.name || node, ts.Diagnostics.JSDoc_types_may_be_moved_to_TypeScript_types));
                    }
                }
                if (canBeConvertedToAsync(node)) {
                    addConvertToAsyncFunctionDiagnostics(node, checker, diags);
                }
                node.forEachChild(check);
            }
        }
        ts.computeSuggestionDiagnostics = computeSuggestionDiagnostics;
        // convertToEs6Module only works on top-level, so don't trigger it if commonjs code only appears in nested scopes.
        function containsTopLevelCommonjs(sourceFile) {
            return sourceFile.statements.some(function (statement) {
                switch (statement.kind) {
                    case 235 /* VariableStatement */:
                        return statement.declarationList.declarations.some(function (decl) {
                            return !!decl.initializer && ts.isRequireCall(propertyAccessLeftHandSide(decl.initializer), /*checkArgumentIsStringLiteralLike*/ true);
                        });
                    case 236 /* ExpressionStatement */: {
                        var expression = statement.expression;
                        if (!ts.isBinaryExpression(expression))
                            return ts.isRequireCall(expression, /*checkArgumentIsStringLiteralLike*/ true);
                        var kind = ts.getAssignmentDeclarationKind(expression);
                        return kind === 1 /* ExportsProperty */ || kind === 2 /* ModuleExports */;
                    }
                    default:
                        return false;
                }
            });
        }
        function propertyAccessLeftHandSide(node) {
            return ts.isPropertyAccessExpression(node) ? propertyAccessLeftHandSide(node.expression) : node;
        }
        function importNameForConvertToDefaultImport(node) {
            switch (node.kind) {
                case 264 /* ImportDeclaration */:
                    var importClause = node.importClause, moduleSpecifier = node.moduleSpecifier;
                    return importClause && !importClause.name && importClause.namedBindings && importClause.namedBindings.kind === 266 /* NamespaceImport */ && ts.isStringLiteral(moduleSpecifier)
                        ? importClause.namedBindings.name
                        : undefined;
                case 263 /* ImportEqualsDeclaration */:
                    return node.name;
                default:
                    return undefined;
            }
        }
        function addConvertToAsyncFunctionDiagnostics(node, checker, diags) {
            // need to check function before checking map so that deeper levels of nested callbacks are checked
            if (isConvertibleFunction(node, checker) && !visitedNestedConvertibleFunctions.has(getKeyFromNode(node))) {
                diags.push(ts.createDiagnosticForNode(!node.name && ts.isVariableDeclaration(node.parent) && ts.isIdentifier(node.parent.name) ? node.parent.name : node, ts.Diagnostics.This_may_be_converted_to_an_async_function));
            }
        }
        function isConvertibleFunction(node, checker) {
            return !ts.isAsyncFunction(node) &&
                node.body &&
                ts.isBlock(node.body) &&
                hasReturnStatementWithPromiseHandler(node.body, checker) &&
                returnsPromise(node, checker);
        }
        function returnsPromise(node, checker) {
            var signature = checker.getSignatureFromDeclaration(node);
            var returnType = signature ? checker.getReturnTypeOfSignature(signature) : undefined;
            return !!returnType && !!checker.getPromisedTypeOfPromise(returnType);
        }
        ts.returnsPromise = returnsPromise;
        function getErrorNodeFromCommonJsIndicator(commonJsModuleIndicator) {
            return ts.isBinaryExpression(commonJsModuleIndicator) ? commonJsModuleIndicator.left : commonJsModuleIndicator;
        }
        function hasReturnStatementWithPromiseHandler(body, checker) {
            return !!ts.forEachReturnStatement(body, function (statement) { return isReturnStatementWithFixablePromiseHandler(statement, checker); });
        }
        function isReturnStatementWithFixablePromiseHandler(node, checker) {
            return ts.isReturnStatement(node) && !!node.expression && isFixablePromiseHandler(node.expression, checker);
        }
        ts.isReturnStatementWithFixablePromiseHandler = isReturnStatementWithFixablePromiseHandler;
        // Should be kept up to date with transformExpression in convertToAsyncFunction.ts
        function isFixablePromiseHandler(node, checker) {
            // ensure outermost call exists and is a promise handler
            if (!isPromiseHandler(node) || !node.arguments.every(function (arg) { return isFixablePromiseArgument(arg, checker); })) {
                return false;
            }
            // ensure all chained calls are valid
            var currentNode = node.expression;
            while (isPromiseHandler(currentNode) || ts.isPropertyAccessExpression(currentNode)) {
                if (ts.isCallExpression(currentNode) && !currentNode.arguments.every(function (arg) { return isFixablePromiseArgument(arg, checker); })) {
                    return false;
                }
                currentNode = currentNode.expression;
            }
            return true;
        }
        ts.isFixablePromiseHandler = isFixablePromiseHandler;
        function isPromiseHandler(node) {
            return ts.isCallExpression(node) && (ts.hasPropertyAccessExpressionWithName(node, "then") && hasSupportedNumberOfArguments(node) ||
                ts.hasPropertyAccessExpressionWithName(node, "catch"));
        }
        function hasSupportedNumberOfArguments(node) {
            if (node.arguments.length > 2)
                return false;
            if (node.arguments.length < 2)
                return true;
            return ts.some(node.arguments, function (arg) {
                return arg.kind === 104 /* NullKeyword */ ||
                    ts.isIdentifier(arg) && arg.text === "undefined";
            });
        }
        // should be kept up to date with getTransformationBody in convertToAsyncFunction.ts
        function isFixablePromiseArgument(arg, checker) {
            switch (arg.kind) {
                case 254 /* FunctionDeclaration */:
                case 211 /* FunctionExpression */:
                    var functionFlags = ts.getFunctionFlags(arg);
                    if (functionFlags & 1 /* Generator */) {
                        return false;
                    }
                // falls through
                case 212 /* ArrowFunction */:
                    visitedNestedConvertibleFunctions.set(getKeyFromNode(arg), true);
                // falls through
                case 104 /* NullKeyword */:
                    return true;
                case 79 /* Identifier */:
                case 204 /* PropertyAccessExpression */: {
                    var symbol = checker.getSymbolAtLocation(arg);
                    if (!symbol) {
                        return false;
                    }
                    return checker.isUndefinedSymbol(symbol) ||
                        ts.some(ts.skipAlias(symbol, checker).declarations, function (d) { return ts.isFunctionLike(d) || ts.hasInitializer(d) && !!d.initializer && ts.isFunctionLike(d.initializer); });
                }
                default:
                    return false;
            }
        }
        function getKeyFromNode(exp) {
            return exp.pos.toString() + ":" + exp.end.toString();
        }
        function canBeConvertedToClass(node, checker) {
            var _a, _b, _c, _d;
            if (node.kind === 211 /* FunctionExpression */) {
                if (ts.isVariableDeclaration(node.parent) && ((_a = node.symbol.members) === null || _a === void 0 ? void 0 : _a.size)) {
                    return true;
                }
                var symbol = checker.getSymbolOfExpando(node, /*allowDeclaration*/ false);
                return !!(symbol && (((_b = symbol.exports) === null || _b === void 0 ? void 0 : _b.size) || ((_c = symbol.members) === null || _c === void 0 ? void 0 : _c.size)));
            }
            if (node.kind === 254 /* FunctionDeclaration */) {
                return !!((_d = node.symbol.members) === null || _d === void 0 ? void 0 : _d.size);
            }
            return false;
        }
        function canBeConvertedToAsync(node) {
            switch (node.kind) {
                case 254 /* FunctionDeclaration */:
                case 167 /* MethodDeclaration */:
                case 211 /* FunctionExpression */:
                case 212 /* ArrowFunction */:
                    return true;
                default:
                    return false;
            }
        }
        ts.canBeConvertedToAsync = canBeConvertedToAsync;
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var SymbolDisplay;
        (function (SymbolDisplay) {
            var symbolDisplayNodeBuilderFlags = 8192 /* OmitParameterModifiers */ | 70221824 /* IgnoreErrors */ | 16384 /* UseAliasDefinedOutsideCurrentScope */;
            // TODO(drosen): use contextual SemanticMeaning.
            function getSymbolKind(typeChecker, symbol, location) {
                var result = getSymbolKindOfConstructorPropertyMethodAccessorFunctionOrVar(typeChecker, symbol, location);
                if (result !== "" /* unknown */) {
                    return result;
                }
                var flags = ts.getCombinedLocalAndExportSymbolFlags(symbol);
                if (flags & 32 /* Class */) {
                    return ts.getDeclarationOfKind(symbol, 224 /* ClassExpression */) ?
                        "local class" /* localClassElement */ : "class" /* classElement */;
                }
                if (flags & 384 /* Enum */)
                    return "enum" /* enumElement */;
                if (flags & 524288 /* TypeAlias */)
                    return "type" /* typeElement */;
                if (flags & 64 /* Interface */)
                    return "interface" /* interfaceElement */;
                if (flags & 262144 /* TypeParameter */)
                    return "type parameter" /* typeParameterElement */;
                if (flags & 8 /* EnumMember */)
                    return "enum member" /* enumMemberElement */;
                if (flags & 2097152 /* Alias */)
                    return "alias" /* alias */;
                if (flags & 1536 /* Module */)
                    return "module" /* moduleElement */;
                return result;
            }
            SymbolDisplay.getSymbolKind = getSymbolKind;
            function getSymbolKindOfConstructorPropertyMethodAccessorFunctionOrVar(typeChecker, symbol, location) {
                var roots = typeChecker.getRootSymbols(symbol);
                // If this is a method from a mapped type, leave as a method so long as it still has a call signature.
                if (roots.length === 1
                    && ts.first(roots).flags & 8192 /* Method */
                    // Ensure the mapped version is still a method, as opposed to `{ [K in keyof I]: number }`.
                    && typeChecker.getTypeOfSymbolAtLocation(symbol, location).getNonNullableType().getCallSignatures().length !== 0) {
                    return "method" /* memberFunctionElement */;
                }
                if (typeChecker.isUndefinedSymbol(symbol)) {
                    return "var" /* variableElement */;
                }
                if (typeChecker.isArgumentsSymbol(symbol)) {
                    return "local var" /* localVariableElement */;
                }
                if (location.kind === 108 /* ThisKeyword */ && ts.isExpression(location)) {
                    return "parameter" /* parameterElement */;
                }
                var flags = ts.getCombinedLocalAndExportSymbolFlags(symbol);
                if (flags & 3 /* Variable */) {
                    if (ts.isFirstDeclarationOfSymbolParameter(symbol)) {
                        return "parameter" /* parameterElement */;
                    }
                    else if (symbol.valueDeclaration && ts.isVarConst(symbol.valueDeclaration)) {
                        return "const" /* constElement */;
                    }
                    else if (ts.forEach(symbol.declarations, ts.isLet)) {
                        return "let" /* letElement */;
                    }
                    return isLocalVariableOrFunction(symbol) ? "local var" /* localVariableElement */ : "var" /* variableElement */;
                }
                if (flags & 16 /* Function */)
                    return isLocalVariableOrFunction(symbol) ? "local function" /* localFunctionElement */ : "function" /* functionElement */;
                if (flags & 32768 /* GetAccessor */)
                    return "getter" /* memberGetAccessorElement */;
                if (flags & 65536 /* SetAccessor */)
                    return "setter" /* memberSetAccessorElement */;
                if (flags & 8192 /* Method */)
                    return "method" /* memberFunctionElement */;
                if (flags & 16384 /* Constructor */)
                    return "constructor" /* constructorImplementationElement */;
                if (flags & 4 /* Property */) {
                    if (flags & 33554432 /* Transient */ && symbol.checkFlags & 6 /* Synthetic */) {
                        // If union property is result of union of non method (property/accessors/variables), it is labeled as property
                        var unionPropertyKind = ts.forEach(typeChecker.getRootSymbols(symbol), function (rootSymbol) {
                            var rootSymbolFlags = rootSymbol.getFlags();
                            if (rootSymbolFlags & (98308 /* PropertyOrAccessor */ | 3 /* Variable */)) {
                                return "property" /* memberVariableElement */;
                            }
                        });
                        if (!unionPropertyKind) {
                            // If this was union of all methods,
                            // make sure it has call signatures before we can label it as method
                            var typeOfUnionProperty = typeChecker.getTypeOfSymbolAtLocation(symbol, location);
                            if (typeOfUnionProperty.getCallSignatures().length) {
                                return "method" /* memberFunctionElement */;
                            }
                            return "property" /* memberVariableElement */;
                        }
                        return unionPropertyKind;
                    }
                    // If we requested completions after `x.` at the top-level, we may be at a source file location.
                    switch (location.parent && location.parent.kind) {
                        // If we've typed a character of the attribute name, will be 'JsxAttribute', else will be 'JsxOpeningElement'.
                        case 278 /* JsxOpeningElement */:
                        case 276 /* JsxElement */:
                        case 277 /* JsxSelfClosingElement */:
                            return location.kind === 79 /* Identifier */ ? "property" /* memberVariableElement */ : "JSX attribute" /* jsxAttribute */;
                        case 283 /* JsxAttribute */:
                            return "JSX attribute" /* jsxAttribute */;
                        default:
                            return "property" /* memberVariableElement */;
                    }
                }
                return "" /* unknown */;
            }
            function getNormalizedSymbolModifiers(symbol) {
                if (symbol.declarations && symbol.declarations.length) {
                    var _a = symbol.declarations, declaration = _a[0], declarations = _a.slice(1);
                    // omit deprecated flag if some declarations are not deprecated
                    var excludeFlags = ts.length(declarations) && ts.isDeprecatedDeclaration(declaration) && ts.some(declarations, function (d) { return !ts.isDeprecatedDeclaration(d); })
                        ? 8192 /* Deprecated */
                        : 0 /* None */;
                    var modifiers = ts.getNodeModifiers(declaration, excludeFlags);
                    if (modifiers) {
                        return modifiers.split(",");
                    }
                }
                return [];
            }
            function getSymbolModifiers(typeChecker, symbol) {
                if (!symbol) {
                    return "" /* none */;
                }
                var modifiers = new ts.Set(getNormalizedSymbolModifiers(symbol));
                if (symbol.flags & 2097152 /* Alias */) {
                    var resolvedSymbol = typeChecker.getAliasedSymbol(symbol);
                    if (resolvedSymbol !== symbol) {
                        ts.forEach(getNormalizedSymbolModifiers(resolvedSymbol), function (modifier) {
                            modifiers.add(modifier);
                        });
                    }
                }
                if (symbol.flags & 16777216 /* Optional */) {
                    modifiers.add("optional" /* optionalModifier */);
                }
                return modifiers.size > 0 ? ts.arrayFrom(modifiers.values()).join(",") : "" /* none */;
            }
            SymbolDisplay.getSymbolModifiers = getSymbolModifiers;
            // TODO(drosen): Currently completion entry details passes the SemanticMeaning.All instead of using semanticMeaning of location
            function getSymbolDisplayPartsDocumentationAndSymbolKind(typeChecker, symbol, sourceFile, enclosingDeclaration, location, semanticMeaning, alias) {
                var _a;
                if (semanticMeaning === void 0) {
                    semanticMeaning = ts.getMeaningFromLocation(location);
                }
                var displayParts = [];
                var documentation = [];
                var tags = [];
                var symbolFlags = ts.getCombinedLocalAndExportSymbolFlags(symbol);
                var symbolKind = semanticMeaning & 1 /* Value */ ? getSymbolKindOfConstructorPropertyMethodAccessorFunctionOrVar(typeChecker, symbol, location) : "" /* unknown */;
                var hasAddedSymbolInfo = false;
                var isThisExpression = location.kind === 108 /* ThisKeyword */ && ts.isInExpressionContext(location);
                var type;
                var printer;
                var documentationFromAlias;
                var tagsFromAlias;
                var hasMultipleSignatures = false;
                if (location.kind === 108 /* ThisKeyword */ && !isThisExpression) {
                    return { displayParts: [ts.keywordPart(108 /* ThisKeyword */)], documentation: [], symbolKind: "primitive type" /* primitiveType */, tags: undefined };
                }
                // Class at constructor site need to be shown as constructor apart from property,method, vars
                if (symbolKind !== "" /* unknown */ || symbolFlags & 32 /* Class */ || symbolFlags & 2097152 /* Alias */) {
                    // If it is accessor they are allowed only if location is at name of the accessor
                    if (symbolKind === "getter" /* memberGetAccessorElement */ || symbolKind === "setter" /* memberSetAccessorElement */) {
                        symbolKind = "property" /* memberVariableElement */;
                    }
                    var signature = void 0;
                    type = isThisExpression ? typeChecker.getTypeAtLocation(location) : typeChecker.getTypeOfSymbolAtLocation(symbol, location);
                    if (location.parent && location.parent.kind === 204 /* PropertyAccessExpression */) {
                        var right = location.parent.name;
                        // Either the location is on the right of a property access, or on the left and the right is missing
                        if (right === location || (right && right.getFullWidth() === 0)) {
                            location = location.parent;
                        }
                    }
                    // try get the call/construct signature from the type if it matches
                    var callExpressionLike = void 0;
                    if (ts.isCallOrNewExpression(location)) {
                        callExpressionLike = location;
                    }
                    else if (ts.isCallExpressionTarget(location) || ts.isNewExpressionTarget(location)) {
                        callExpressionLike = location.parent;
                    }
                    else if (location.parent && (ts.isJsxOpeningLikeElement(location.parent) || ts.isTaggedTemplateExpression(location.parent)) && ts.isFunctionLike(symbol.valueDeclaration)) {
                        callExpressionLike = location.parent;
                    }
                    if (callExpressionLike) {
                        signature = typeChecker.getResolvedSignature(callExpressionLike); // TODO: GH#18217
                        var useConstructSignatures = callExpressionLike.kind === 207 /* NewExpression */ || (ts.isCallExpression(callExpressionLike) && callExpressionLike.expression.kind === 106 /* SuperKeyword */);
                        var allSignatures = useConstructSignatures ? type.getConstructSignatures() : type.getCallSignatures();
                        if (signature && !ts.contains(allSignatures, signature.target) && !ts.contains(allSignatures, signature)) {
                            // Get the first signature if there is one -- allSignatures may contain
                            // either the original signature or its target, so check for either
                            signature = allSignatures.length ? allSignatures[0] : undefined;
                        }
                        if (signature) {
                            if (useConstructSignatures && (symbolFlags & 32 /* Class */)) {
                                // Constructor
                                symbolKind = "constructor" /* constructorImplementationElement */;
                                addPrefixForAnyFunctionOrVar(type.symbol, symbolKind);
                            }
                            else if (symbolFlags & 2097152 /* Alias */) {
                                symbolKind = "alias" /* alias */;
                                pushSymbolKind(symbolKind);
                                displayParts.push(ts.spacePart());
                                if (useConstructSignatures) {
                                    if (signature.flags & 4 /* Abstract */) {
                                        displayParts.push(ts.keywordPart(126 /* AbstractKeyword */));
                                        displayParts.push(ts.spacePart());
                                    }
                                    displayParts.push(ts.keywordPart(103 /* NewKeyword */));
                                    displayParts.push(ts.spacePart());
                                }
                                addFullSymbolName(symbol);
                            }
                            else {
                                addPrefixForAnyFunctionOrVar(symbol, symbolKind);
                            }
                            switch (symbolKind) {
                                case "JSX attribute" /* jsxAttribute */:
                                case "property" /* memberVariableElement */:
                                case "var" /* variableElement */:
                                case "const" /* constElement */:
                                case "let" /* letElement */:
                                case "parameter" /* parameterElement */:
                                case "local var" /* localVariableElement */:
                                    // If it is call or construct signature of lambda's write type name
                                    displayParts.push(ts.punctuationPart(58 /* ColonToken */));
                                    displayParts.push(ts.spacePart());
                                    if (!(ts.getObjectFlags(type) & 16 /* Anonymous */) && type.symbol) {
                                        ts.addRange(displayParts, ts.symbolToDisplayParts(typeChecker, type.symbol, enclosingDeclaration, /*meaning*/ undefined, 4 /* AllowAnyNodeKind */ | 1 /* WriteTypeParametersOrArguments */));
                                        displayParts.push(ts.lineBreakPart());
                                    }
                                    if (useConstructSignatures) {
                                        if (signature.flags & 4 /* Abstract */) {
                                            displayParts.push(ts.keywordPart(126 /* AbstractKeyword */));
                                            displayParts.push(ts.spacePart());
                                        }
                                        displayParts.push(ts.keywordPart(103 /* NewKeyword */));
                                        displayParts.push(ts.spacePart());
                                    }
                                    addSignatureDisplayParts(signature, allSignatures, 262144 /* WriteArrowStyleSignature */);
                                    break;
                                default:
                                    // Just signature
                                    addSignatureDisplayParts(signature, allSignatures);
                            }
                            hasAddedSymbolInfo = true;
                            hasMultipleSignatures = allSignatures.length > 1;
                        }
                    }
                    else if ((ts.isNameOfFunctionDeclaration(location) && !(symbolFlags & 98304 /* Accessor */)) || // name of function declaration
                        (location.kind === 133 /* ConstructorKeyword */ && location.parent.kind === 169 /* Constructor */)) { // At constructor keyword of constructor declaration
                        // get the signature from the declaration and write it
                        var functionDeclaration_1 = location.parent;
                        // Use function declaration to write the signatures only if the symbol corresponding to this declaration
                        var locationIsSymbolDeclaration = symbol.declarations && ts.find(symbol.declarations, function (declaration) {
                            return declaration === (location.kind === 133 /* ConstructorKeyword */ ? functionDeclaration_1.parent : functionDeclaration_1);
                        });
                        if (locationIsSymbolDeclaration) {
                            var allSignatures = functionDeclaration_1.kind === 169 /* Constructor */ ? type.getNonNullableType().getConstructSignatures() : type.getNonNullableType().getCallSignatures();
                            if (!typeChecker.isImplementationOfOverload(functionDeclaration_1)) {
                                signature = typeChecker.getSignatureFromDeclaration(functionDeclaration_1); // TODO: GH#18217
                            }
                            else {
                                signature = allSignatures[0];
                            }
                            if (functionDeclaration_1.kind === 169 /* Constructor */) {
                                // show (constructor) Type(...) signature
                                symbolKind = "constructor" /* constructorImplementationElement */;
                                addPrefixForAnyFunctionOrVar(type.symbol, symbolKind);
                            }
                            else {
                                // (function/method) symbol(..signature)
                                addPrefixForAnyFunctionOrVar(functionDeclaration_1.kind === 172 /* CallSignature */ &&
                                    !(type.symbol.flags & 2048 /* TypeLiteral */ || type.symbol.flags & 4096 /* ObjectLiteral */) ? type.symbol : symbol, symbolKind);
                            }
                            if (signature) {
                                addSignatureDisplayParts(signature, allSignatures);
                            }
                            hasAddedSymbolInfo = true;
                            hasMultipleSignatures = allSignatures.length > 1;
                        }
                    }
                }
                if (symbolFlags & 32 /* Class */ && !hasAddedSymbolInfo && !isThisExpression) {
                    addAliasPrefixIfNecessary();
                    if (ts.getDeclarationOfKind(symbol, 224 /* ClassExpression */)) {
                        // Special case for class expressions because we would like to indicate that
                        // the class name is local to the class body (similar to function expression)
                        //      (local class) class <className>
                        pushSymbolKind("local class" /* localClassElement */);
                    }
                    else {
                        // Class declaration has name which is not local.
                        displayParts.push(ts.keywordPart(84 /* ClassKeyword */));
                    }
                    displayParts.push(ts.spacePart());
                    addFullSymbolName(symbol);
                    writeTypeParametersOfSymbol(symbol, sourceFile);
                }
                if ((symbolFlags & 64 /* Interface */) && (semanticMeaning & 2 /* Type */)) {
                    prefixNextMeaning();
                    displayParts.push(ts.keywordPart(118 /* InterfaceKeyword */));
                    displayParts.push(ts.spacePart());
                    addFullSymbolName(symbol);
                    writeTypeParametersOfSymbol(symbol, sourceFile);
                }
                if ((symbolFlags & 524288 /* TypeAlias */) && (semanticMeaning & 2 /* Type */)) {
                    prefixNextMeaning();
                    displayParts.push(ts.keywordPart(150 /* TypeKeyword */));
                    displayParts.push(ts.spacePart());
                    addFullSymbolName(symbol);
                    writeTypeParametersOfSymbol(symbol, sourceFile);
                    displayParts.push(ts.spacePart());
                    displayParts.push(ts.operatorPart(63 /* EqualsToken */));
                    displayParts.push(ts.spacePart());
                    ts.addRange(displayParts, ts.typeToDisplayParts(typeChecker, typeChecker.getDeclaredTypeOfSymbol(symbol), enclosingDeclaration, 8388608 /* InTypeAlias */));
                }
                if (symbolFlags & 384 /* Enum */) {
                    prefixNextMeaning();
                    if (ts.some(symbol.declarations, function (d) { return ts.isEnumDeclaration(d) && ts.isEnumConst(d); })) {
                        displayParts.push(ts.keywordPart(85 /* ConstKeyword */));
                        displayParts.push(ts.spacePart());
                    }
                    displayParts.push(ts.keywordPart(92 /* EnumKeyword */));
                    displayParts.push(ts.spacePart());
                    addFullSymbolName(symbol);
                }
                if (symbolFlags & 1536 /* Module */ && !isThisExpression) {
                    prefixNextMeaning();
                    var declaration = ts.getDeclarationOfKind(symbol, 259 /* ModuleDeclaration */);
                    var isNamespace = declaration && declaration.name && declaration.name.kind === 79 /* Identifier */;
                    displayParts.push(ts.keywordPart(isNamespace ? 141 /* NamespaceKeyword */ : 140 /* ModuleKeyword */));
                    displayParts.push(ts.spacePart());
                    addFullSymbolName(symbol);
                }
                if ((symbolFlags & 262144 /* TypeParameter */) && (semanticMeaning & 2 /* Type */)) {
                    prefixNextMeaning();
                    displayParts.push(ts.punctuationPart(20 /* OpenParenToken */));
                    displayParts.push(ts.textPart("type parameter"));
                    displayParts.push(ts.punctuationPart(21 /* CloseParenToken */));
                    displayParts.push(ts.spacePart());
                    addFullSymbolName(symbol);
                    if (symbol.parent) {
                        // Class/Interface type parameter
                        addInPrefix();
                        addFullSymbolName(symbol.parent, enclosingDeclaration);
                        writeTypeParametersOfSymbol(symbol.parent, enclosingDeclaration);
                    }
                    else {
                        // Method/function type parameter
                        var decl = ts.getDeclarationOfKind(symbol, 161 /* TypeParameter */);
                        if (decl === undefined)
                            return ts.Debug.fail();
                        var declaration = decl.parent;
                        if (declaration) {
                            if (ts.isFunctionLikeKind(declaration.kind)) {
                                addInPrefix();
                                var signature = typeChecker.getSignatureFromDeclaration(declaration); // TODO: GH#18217
                                if (declaration.kind === 173 /* ConstructSignature */) {
                                    displayParts.push(ts.keywordPart(103 /* NewKeyword */));
                                    displayParts.push(ts.spacePart());
                                }
                                else if (declaration.kind !== 172 /* CallSignature */ && declaration.name) {
                                    addFullSymbolName(declaration.symbol);
                                }
                                ts.addRange(displayParts, ts.signatureToDisplayParts(typeChecker, signature, sourceFile, 32 /* WriteTypeArgumentsOfSignature */));
                            }
                            else if (declaration.kind === 257 /* TypeAliasDeclaration */) {
                                // Type alias type parameter
                                // For example
                                //      type list<T> = T[]; // Both T will go through same code path
                                addInPrefix();
                                displayParts.push(ts.keywordPart(150 /* TypeKeyword */));
                                displayParts.push(ts.spacePart());
                                addFullSymbolName(declaration.symbol);
                                writeTypeParametersOfSymbol(declaration.symbol, sourceFile);
                            }
                        }
                    }
                }
                if (symbolFlags & 8 /* EnumMember */) {
                    symbolKind = "enum member" /* enumMemberElement */;
                    addPrefixForAnyFunctionOrVar(symbol, "enum member");
                    var declaration = (_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a[0];
                    if ((declaration === null || declaration === void 0 ? void 0 : declaration.kind) === 294 /* EnumMember */) {
                        var constantValue = typeChecker.getConstantValue(declaration);
                        if (constantValue !== undefined) {
                            displayParts.push(ts.spacePart());
                            displayParts.push(ts.operatorPart(63 /* EqualsToken */));
                            displayParts.push(ts.spacePart());
                            displayParts.push(ts.displayPart(ts.getTextOfConstantValue(constantValue), typeof constantValue === "number" ? ts.SymbolDisplayPartKind.numericLiteral : ts.SymbolDisplayPartKind.stringLiteral));
                        }
                    }
                }
                // don't use symbolFlags since getAliasedSymbol requires the flag on the symbol itself
                if (symbol.flags & 2097152 /* Alias */) {
                    prefixNextMeaning();
                    if (!hasAddedSymbolInfo) {
                        var resolvedSymbol = typeChecker.getAliasedSymbol(symbol);
                        if (resolvedSymbol !== symbol && resolvedSymbol.declarations && resolvedSymbol.declarations.length > 0) {
                            var resolvedNode = resolvedSymbol.declarations[0];
                            var declarationName = ts.getNameOfDeclaration(resolvedNode);
                            if (declarationName) {
                                var isExternalModuleDeclaration = ts.isModuleWithStringLiteralName(resolvedNode) &&
                                    ts.hasSyntacticModifier(resolvedNode, 2 /* Ambient */);
                                var shouldUseAliasName = symbol.name !== "default" && !isExternalModuleDeclaration;
                                var resolvedInfo = getSymbolDisplayPartsDocumentationAndSymbolKind(typeChecker, resolvedSymbol, ts.getSourceFileOfNode(resolvedNode), resolvedNode, declarationName, semanticMeaning, shouldUseAliasName ? symbol : resolvedSymbol);
                                displayParts.push.apply(displayParts, resolvedInfo.displayParts);
                                displayParts.push(ts.lineBreakPart());
                                documentationFromAlias = resolvedInfo.documentation;
                                tagsFromAlias = resolvedInfo.tags;
                            }
                            else {
                                documentationFromAlias = resolvedSymbol.getContextualDocumentationComment(resolvedNode, typeChecker);
                                tagsFromAlias = resolvedSymbol.getJsDocTags(typeChecker);
                            }
                        }
                    }
                    if (symbol.declarations) {
                        switch (symbol.declarations[0].kind) {
                            case 262 /* NamespaceExportDeclaration */:
                                displayParts.push(ts.keywordPart(93 /* ExportKeyword */));
                                displayParts.push(ts.spacePart());
                                displayParts.push(ts.keywordPart(141 /* NamespaceKeyword */));
                                break;
                            case 269 /* ExportAssignment */:
                                displayParts.push(ts.keywordPart(93 /* ExportKeyword */));
                                displayParts.push(ts.spacePart());
                                displayParts.push(ts.keywordPart(symbol.declarations[0].isExportEquals ? 63 /* EqualsToken */ : 88 /* DefaultKeyword */));
                                break;
                            case 273 /* ExportSpecifier */:
                                displayParts.push(ts.keywordPart(93 /* ExportKeyword */));
                                break;
                            default:
                                displayParts.push(ts.keywordPart(100 /* ImportKeyword */));
                        }
                    }
                    displayParts.push(ts.spacePart());
                    addFullSymbolName(symbol);
                    ts.forEach(symbol.declarations, function (declaration) {
                        if (declaration.kind === 263 /* ImportEqualsDeclaration */) {
                            var importEqualsDeclaration = declaration;
                            if (ts.isExternalModuleImportEqualsDeclaration(importEqualsDeclaration)) {
                                displayParts.push(ts.spacePart());
                                displayParts.push(ts.operatorPart(63 /* EqualsToken */));
                                displayParts.push(ts.spacePart());
                                displayParts.push(ts.keywordPart(144 /* RequireKeyword */));
                                displayParts.push(ts.punctuationPart(20 /* OpenParenToken */));
                                displayParts.push(ts.displayPart(ts.getTextOfNode(ts.getExternalModuleImportEqualsDeclarationExpression(importEqualsDeclaration)), ts.SymbolDisplayPartKind.stringLiteral));
                                displayParts.push(ts.punctuationPart(21 /* CloseParenToken */));
                            }
                            else {
                                var internalAliasSymbol = typeChecker.getSymbolAtLocation(importEqualsDeclaration.moduleReference);
                                if (internalAliasSymbol) {
                                    displayParts.push(ts.spacePart());
                                    displayParts.push(ts.operatorPart(63 /* EqualsToken */));
                                    displayParts.push(ts.spacePart());
                                    addFullSymbolName(internalAliasSymbol, enclosingDeclaration);
                                }
                            }
                            return true;
                        }
                    });
                }
                if (!hasAddedSymbolInfo) {
                    if (symbolKind !== "" /* unknown */) {
                        if (type) {
                            if (isThisExpression) {
                                prefixNextMeaning();
                                displayParts.push(ts.keywordPart(108 /* ThisKeyword */));
                            }
                            else {
                                addPrefixForAnyFunctionOrVar(symbol, symbolKind);
                            }
                            // For properties, variables and local vars: show the type
                            if (symbolKind === "property" /* memberVariableElement */ ||
                                symbolKind === "JSX attribute" /* jsxAttribute */ ||
                                symbolFlags & 3 /* Variable */ ||
                                symbolKind === "local var" /* localVariableElement */ ||
                                isThisExpression) {
                                displayParts.push(ts.punctuationPart(58 /* ColonToken */));
                                displayParts.push(ts.spacePart());
                                // If the type is type parameter, format it specially
                                if (type.symbol && type.symbol.flags & 262144 /* TypeParameter */) {
                                    var typeParameterParts = ts.mapToDisplayParts(function (writer) {
                                        var param = typeChecker.typeParameterToDeclaration(type, enclosingDeclaration, symbolDisplayNodeBuilderFlags);
                                        getPrinter().writeNode(4 /* Unspecified */, param, ts.getSourceFileOfNode(ts.getParseTreeNode(enclosingDeclaration)), writer);
                                    });
                                    ts.addRange(displayParts, typeParameterParts);
                                }
                                else {
                                    ts.addRange(displayParts, ts.typeToDisplayParts(typeChecker, type, enclosingDeclaration));
                                }
                                if (symbol.target && symbol.target.tupleLabelDeclaration) {
                                    var labelDecl = symbol.target.tupleLabelDeclaration;
                                    ts.Debug.assertNode(labelDecl.name, ts.isIdentifier);
                                    displayParts.push(ts.spacePart());
                                    displayParts.push(ts.punctuationPart(20 /* OpenParenToken */));
                                    displayParts.push(ts.textPart(ts.idText(labelDecl.name)));
                                    displayParts.push(ts.punctuationPart(21 /* CloseParenToken */));
                                }
                            }
                            else if (symbolFlags & 16 /* Function */ ||
                                symbolFlags & 8192 /* Method */ ||
                                symbolFlags & 16384 /* Constructor */ ||
                                symbolFlags & 131072 /* Signature */ ||
                                symbolFlags & 98304 /* Accessor */ ||
                                symbolKind === "method" /* memberFunctionElement */) {
                                var allSignatures = type.getNonNullableType().getCallSignatures();
                                if (allSignatures.length) {
                                    addSignatureDisplayParts(allSignatures[0], allSignatures);
                                    hasMultipleSignatures = allSignatures.length > 1;
                                }
                            }
                        }
                    }
                    else {
                        symbolKind = getSymbolKind(typeChecker, symbol, location);
                    }
                }
                if (documentation.length === 0 && !hasMultipleSignatures) {
                    documentation = symbol.getContextualDocumentationComment(enclosingDeclaration, typeChecker);
                }
                if (documentation.length === 0 && symbolFlags & 4 /* Property */) {
                    // For some special property access expressions like `exports.foo = foo` or `module.exports.foo = foo`
                    // there documentation comments might be attached to the right hand side symbol of their declarations.
                    // The pattern of such special property access is that the parent symbol is the symbol of the file.
                    if (symbol.parent && symbol.declarations && ts.forEach(symbol.parent.declarations, function (declaration) { return declaration.kind === 300 /* SourceFile */; })) {
                        for (var _i = 0, _b = symbol.declarations; _i < _b.length; _i++) {
                            var declaration = _b[_i];
                            if (!declaration.parent || declaration.parent.kind !== 219 /* BinaryExpression */) {
                                continue;
                            }
                            var rhsSymbol = typeChecker.getSymbolAtLocation(declaration.parent.right);
                            if (!rhsSymbol) {
                                continue;
                            }
                            documentation = rhsSymbol.getDocumentationComment(typeChecker);
                            tags = rhsSymbol.getJsDocTags(typeChecker);
                            if (documentation.length > 0) {
                                break;
                            }
                        }
                    }
                }
                if (tags.length === 0 && !hasMultipleSignatures) {
                    tags = symbol.getJsDocTags(typeChecker);
                }
                if (documentation.length === 0 && documentationFromAlias) {
                    documentation = documentationFromAlias;
                }
                if (tags.length === 0 && tagsFromAlias) {
                    tags = tagsFromAlias;
                }
                return { displayParts: displayParts, documentation: documentation, symbolKind: symbolKind, tags: tags.length === 0 ? undefined : tags };
                function getPrinter() {
                    if (!printer) {
                        printer = ts.createPrinter({ removeComments: true });
                    }
                    return printer;
                }
                function prefixNextMeaning() {
                    if (displayParts.length) {
                        displayParts.push(ts.lineBreakPart());
                    }
                    addAliasPrefixIfNecessary();
                }
                function addAliasPrefixIfNecessary() {
                    if (alias) {
                        pushSymbolKind("alias" /* alias */);
                        displayParts.push(ts.spacePart());
                    }
                }
                function addInPrefix() {
                    displayParts.push(ts.spacePart());
                    displayParts.push(ts.keywordPart(101 /* InKeyword */));
                    displayParts.push(ts.spacePart());
                }
                function addFullSymbolName(symbolToDisplay, enclosingDeclaration) {
                    if (alias && symbolToDisplay === symbol) {
                        symbolToDisplay = alias;
                    }
                    var fullSymbolDisplayParts = ts.symbolToDisplayParts(typeChecker, symbolToDisplay, enclosingDeclaration || sourceFile, /*meaning*/ undefined, 1 /* WriteTypeParametersOrArguments */ | 2 /* UseOnlyExternalAliasing */ | 4 /* AllowAnyNodeKind */);
                    ts.addRange(displayParts, fullSymbolDisplayParts);
                    if (symbol.flags & 16777216 /* Optional */) {
                        displayParts.push(ts.punctuationPart(57 /* QuestionToken */));
                    }
                }
                function addPrefixForAnyFunctionOrVar(symbol, symbolKind) {
                    prefixNextMeaning();
                    if (symbolKind) {
                        pushSymbolKind(symbolKind);
                        if (symbol && !ts.some(symbol.declarations, function (d) { return ts.isArrowFunction(d) || (ts.isFunctionExpression(d) || ts.isClassExpression(d)) && !d.name; })) {
                            displayParts.push(ts.spacePart());
                            addFullSymbolName(symbol);
                        }
                    }
                }
                function pushSymbolKind(symbolKind) {
                    switch (symbolKind) {
                        case "var" /* variableElement */:
                        case "function" /* functionElement */:
                        case "let" /* letElement */:
                        case "const" /* constElement */:
                        case "constructor" /* constructorImplementationElement */:
                            displayParts.push(ts.textOrKeywordPart(symbolKind));
                            return;
                        default:
                            displayParts.push(ts.punctuationPart(20 /* OpenParenToken */));
                            displayParts.push(ts.textOrKeywordPart(symbolKind));
                            displayParts.push(ts.punctuationPart(21 /* CloseParenToken */));
                            return;
                    }
                }
                function addSignatureDisplayParts(signature, allSignatures, flags) {
                    if (flags === void 0) {
                        flags = 0 /* None */;
                    }
                    ts.addRange(displayParts, ts.signatureToDisplayParts(typeChecker, signature, enclosingDeclaration, flags | 32 /* WriteTypeArgumentsOfSignature */));
                    if (allSignatures.length > 1) {
                        displayParts.push(ts.spacePart());
                        displayParts.push(ts.punctuationPart(20 /* OpenParenToken */));
                        displayParts.push(ts.operatorPart(39 /* PlusToken */));
                        displayParts.push(ts.displayPart((allSignatures.length - 1).toString(), ts.SymbolDisplayPartKind.numericLiteral));
                        displayParts.push(ts.spacePart());
                        displayParts.push(ts.textPart(allSignatures.length === 2 ? "overload" : "overloads"));
                        displayParts.push(ts.punctuationPart(21 /* CloseParenToken */));
                    }
                    documentation = signature.getDocumentationComment(typeChecker);
                    tags = signature.getJsDocTags();
                    if (allSignatures.length > 1 && documentation.length === 0 && tags.length === 0) {
                        documentation = allSignatures[0].getDocumentationComment(typeChecker);
                        tags = allSignatures[0].getJsDocTags();
                    }
                }
                function writeTypeParametersOfSymbol(symbol, enclosingDeclaration) {
                    var typeParameterParts = ts.mapToDisplayParts(function (writer) {
                        var params = typeChecker.symbolToTypeParameterDeclarations(symbol, enclosingDeclaration, symbolDisplayNodeBuilderFlags);
                        getPrinter().writeList(53776 /* TypeParameters */, params, ts.getSourceFileOfNode(ts.getParseTreeNode(enclosingDeclaration)), writer);
                    });
                    ts.addRange(displayParts, typeParameterParts);
                }
            }
            SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind = getSymbolDisplayPartsDocumentationAndSymbolKind;
            function isLocalVariableOrFunction(symbol) {
                if (symbol.parent) {
                    return false; // This is exported symbol
                }
                return ts.forEach(symbol.declarations, function (declaration) {
                    // Function expressions are local
                    if (declaration.kind === 211 /* FunctionExpression */) {
                        return true;
                    }
                    if (declaration.kind !== 252 /* VariableDeclaration */ && declaration.kind !== 254 /* FunctionDeclaration */) {
                        return false;
                    }
                    // If the parent is not sourceFile or module block it is local variable
                    for (var parent = declaration.parent; !ts.isFunctionBlock(parent); parent = parent.parent) {
                        // Reached source file or module block
                        if (parent.kind === 300 /* SourceFile */ || parent.kind === 260 /* ModuleBlock */) {
                            return false;
                        }
                    }
                    // parent is in function block
                    return true;
                });
            }
        })(SymbolDisplay = ts.SymbolDisplay || (ts.SymbolDisplay = {}));
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        /*
         * This function will compile source text from 'input' argument using specified compiler options.
         * If not options are provided - it will use a set of default compiler options.
         * Extra compiler options that will unconditionally be used by this function are:
         * - isolatedModules = true
         * - allowNonTsExtensions = true
         * - noLib = true
         * - noResolve = true
         */
        function transpileModule(input, transpileOptions) {
            var diagnostics = [];
            var options = transpileOptions.compilerOptions ? fixupCompilerOptions(transpileOptions.compilerOptions, diagnostics) : {};
            // mix in default options
            var defaultOptions = ts.getDefaultCompilerOptions();
            for (var key in defaultOptions) {
                if (ts.hasProperty(defaultOptions, key) && options[key] === undefined) {
                    options[key] = defaultOptions[key];
                }
            }
            for (var _i = 0, transpileOptionValueCompilerOptions_1 = ts.transpileOptionValueCompilerOptions; _i < transpileOptionValueCompilerOptions_1.length; _i++) {
                var option = transpileOptionValueCompilerOptions_1[_i];
                options[option.name] = option.transpileOptionValue;
            }
            // transpileModule does not write anything to disk so there is no need to verify that there are no conflicts between input and output paths.
            options.suppressOutputPathCheck = true;
            // Filename can be non-ts file.
            options.allowNonTsExtensions = true;
            // if jsx is specified then treat file as .tsx
            var inputFileName = transpileOptions.fileName || (transpileOptions.compilerOptions && transpileOptions.compilerOptions.jsx ? "module.tsx" : "module.ts");
            var sourceFile = ts.createSourceFile(inputFileName, input, options.target); // TODO: GH#18217
            if (transpileOptions.moduleName) {
                sourceFile.moduleName = transpileOptions.moduleName;
            }
            if (transpileOptions.renamedDependencies) {
                sourceFile.renamedDependencies = new ts.Map(ts.getEntries(transpileOptions.renamedDependencies));
            }
            var newLine = ts.getNewLineCharacter(options);
            // Output
            var outputText;
            var sourceMapText;
            // Create a compilerHost object to allow the compiler to read and write files
            var compilerHost = {
                getSourceFile: function (fileName) { return fileName === ts.normalizePath(inputFileName) ? sourceFile : undefined; },
                writeFile: function (name, text) {
                    if (ts.fileExtensionIs(name, ".map")) {
                        ts.Debug.assertEqual(sourceMapText, undefined, "Unexpected multiple source map outputs, file:", name);
                        sourceMapText = text;
                    }
                    else {
                        ts.Debug.assertEqual(outputText, undefined, "Unexpected multiple outputs, file:", name);
                        outputText = text;
                    }
                },
                getDefaultLibFileName: function () { return "lib.d.ts"; },
                useCaseSensitiveFileNames: function () { return false; },
                getCanonicalFileName: function (fileName) { return fileName; },
                getCurrentDirectory: function () { return ""; },
                getNewLine: function () { return newLine; },
                fileExists: function (fileName) { return fileName === inputFileName; },
                readFile: function () { return ""; },
                directoryExists: function () { return true; },
                getDirectories: function () { return []; }
            };
            var program = ts.createProgram([inputFileName], options, compilerHost);
            if (transpileOptions.reportDiagnostics) {
                ts.addRange(/*to*/ diagnostics, /*from*/ program.getSyntacticDiagnostics(sourceFile));
                ts.addRange(/*to*/ diagnostics, /*from*/ program.getOptionsDiagnostics());
            }
            // Emit
            program.emit(/*targetSourceFile*/ undefined, /*writeFile*/ undefined, /*cancellationToken*/ undefined, /*emitOnlyDtsFiles*/ undefined, transpileOptions.transformers);
            if (outputText === undefined)
                return ts.Debug.fail("Output generation failed");
            return { outputText: outputText, diagnostics: diagnostics, sourceMapText: sourceMapText };
        }
        ts.transpileModule = transpileModule;
        /*
         * This is a shortcut function for transpileModule - it accepts transpileOptions as parameters and returns only outputText part of the result.
         */
        function transpile(input, compilerOptions, fileName, diagnostics, moduleName) {
            var output = transpileModule(input, { compilerOptions: compilerOptions, fileName: fileName, reportDiagnostics: !!diagnostics, moduleName: moduleName });
            // addRange correctly handles cases when wither 'from' or 'to' argument is missing
            ts.addRange(diagnostics, output.diagnostics);
            return output.outputText;
        }
        ts.transpile = transpile;
        var commandLineOptionsStringToEnum;
        /** JS users may pass in string values for enum compiler options (such as ModuleKind), so convert. */
        /*@internal*/
        function fixupCompilerOptions(options, diagnostics) {
            // Lazily create this value to fix module loading errors.
            commandLineOptionsStringToEnum = commandLineOptionsStringToEnum ||
                ts.filter(ts.optionDeclarations, function (o) { return typeof o.type === "object" && !ts.forEachEntry(o.type, function (v) { return typeof v !== "number"; }); });
            options = ts.cloneCompilerOptions(options);
            var _loop_9 = function (opt) {
                if (!ts.hasProperty(options, opt.name)) {
                    return "continue";
                }
                var value = options[opt.name];
                // Value should be a key of opt.type
                if (ts.isString(value)) {
                    // If value is not a string, this will fail
                    options[opt.name] = ts.parseCustomTypeOption(opt, value, diagnostics);
                }
                else {
                    if (!ts.forEachEntry(opt.type, function (v) { return v === value; })) {
                        // Supplied value isn't a valid enum value.
                        diagnostics.push(ts.createCompilerDiagnosticForInvalidCustomType(opt));
                    }
                }
            };
            for (var _i = 0, commandLineOptionsStringToEnum_1 = commandLineOptionsStringToEnum; _i < commandLineOptionsStringToEnum_1.length; _i++) {
                var opt = commandLineOptionsStringToEnum_1[_i];
                _loop_9(opt);
            }
            return options;
        }
        ts.fixupCompilerOptions = fixupCompilerOptions;
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var formatting;
        (function (formatting) {
            var FormattingRequestKind;
            (function (FormattingRequestKind) {
                FormattingRequestKind[FormattingRequestKind["FormatDocument"] = 0] = "FormatDocument";
                FormattingRequestKind[FormattingRequestKind["FormatSelection"] = 1] = "FormatSelection";
                FormattingRequestKind[FormattingRequestKind["FormatOnEnter"] = 2] = "FormatOnEnter";
                FormattingRequestKind[FormattingRequestKind["FormatOnSemicolon"] = 3] = "FormatOnSemicolon";
                FormattingRequestKind[FormattingRequestKind["FormatOnOpeningCurlyBrace"] = 4] = "FormatOnOpeningCurlyBrace";
                FormattingRequestKind[FormattingRequestKind["FormatOnClosingCurlyBrace"] = 5] = "FormatOnClosingCurlyBrace";
            })(FormattingRequestKind = formatting.FormattingRequestKind || (formatting.FormattingRequestKind = {}));
            var FormattingContext = /** @class */ (function () {
                function FormattingContext(sourceFile, formattingRequestKind, options) {
                    this.sourceFile = sourceFile;
                    this.formattingRequestKind = formattingRequestKind;
                    this.options = options;
                }
                FormattingContext.prototype.updateContext = function (currentRange, currentTokenParent, nextRange, nextTokenParent, commonParent) {
                    this.currentTokenSpan = ts.Debug.checkDefined(currentRange);
                    this.currentTokenParent = ts.Debug.checkDefined(currentTokenParent);
                    this.nextTokenSpan = ts.Debug.checkDefined(nextRange);
                    this.nextTokenParent = ts.Debug.checkDefined(nextTokenParent);
                    this.contextNode = ts.Debug.checkDefined(commonParent);
                    // drop cached results
                    this.contextNodeAllOnSameLine = undefined;
                    this.nextNodeAllOnSameLine = undefined;
                    this.tokensAreOnSameLine = undefined;
                    this.contextNodeBlockIsOnOneLine = undefined;
                    this.nextNodeBlockIsOnOneLine = undefined;
                };
                FormattingContext.prototype.ContextNodeAllOnSameLine = function () {
                    if (this.contextNodeAllOnSameLine === undefined) {
                        this.contextNodeAllOnSameLine = this.NodeIsOnOneLine(this.contextNode);
                    }
                    return this.contextNodeAllOnSameLine;
                };
                FormattingContext.prototype.NextNodeAllOnSameLine = function () {
                    if (this.nextNodeAllOnSameLine === undefined) {
                        this.nextNodeAllOnSameLine = this.NodeIsOnOneLine(this.nextTokenParent);
                    }
                    return this.nextNodeAllOnSameLine;
                };
                FormattingContext.prototype.TokensAreOnSameLine = function () {
                    if (this.tokensAreOnSameLine === undefined) {
                        var startLine = this.sourceFile.getLineAndCharacterOfPosition(this.currentTokenSpan.pos).line;
                        var endLine = this.sourceFile.getLineAndCharacterOfPosition(this.nextTokenSpan.pos).line;
                        this.tokensAreOnSameLine = (startLine === endLine);
                    }
                    return this.tokensAreOnSameLine;
                };
                FormattingContext.prototype.ContextNodeBlockIsOnOneLine = function () {
                    if (this.contextNodeBlockIsOnOneLine === undefined) {
                        this.contextNodeBlockIsOnOneLine = this.BlockIsOnOneLine(this.contextNode);
                    }
                    return this.contextNodeBlockIsOnOneLine;
                };
                FormattingContext.prototype.NextNodeBlockIsOnOneLine = function () {
                    if (this.nextNodeBlockIsOnOneLine === undefined) {
                        this.nextNodeBlockIsOnOneLine = this.BlockIsOnOneLine(this.nextTokenParent);
                    }
                    return this.nextNodeBlockIsOnOneLine;
                };
                FormattingContext.prototype.NodeIsOnOneLine = function (node) {
                    var startLine = this.sourceFile.getLineAndCharacterOfPosition(node.getStart(this.sourceFile)).line;
                    var endLine = this.sourceFile.getLineAndCharacterOfPosition(node.getEnd()).line;
                    return startLine === endLine;
                };
                FormattingContext.prototype.BlockIsOnOneLine = function (node) {
                    var openBrace = ts.findChildOfKind(node, 18 /* OpenBraceToken */, this.sourceFile);
                    var closeBrace = ts.findChildOfKind(node, 19 /* CloseBraceToken */, this.sourceFile);
                    if (openBrace && closeBrace) {
                        var startLine = this.sourceFile.getLineAndCharacterOfPosition(openBrace.getEnd()).line;
                        var endLine = this.sourceFile.getLineAndCharacterOfPosition(closeBrace.getStart(this.sourceFile)).line;
                        return startLine === endLine;
                    }
                    return false;
                };
                return FormattingContext;
            }());
            formatting.FormattingContext = FormattingContext;
        })(formatting = ts.formatting || (ts.formatting = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var formatting;
        (function (formatting) {
            var standardScanner = ts.createScanner(99 /* Latest */, /*skipTrivia*/ false, 0 /* Standard */);
            var jsxScanner = ts.createScanner(99 /* Latest */, /*skipTrivia*/ false, 1 /* JSX */);
            var ScanAction;
            (function (ScanAction) {
                ScanAction[ScanAction["Scan"] = 0] = "Scan";
                ScanAction[ScanAction["RescanGreaterThanToken"] = 1] = "RescanGreaterThanToken";
                ScanAction[ScanAction["RescanSlashToken"] = 2] = "RescanSlashToken";
                ScanAction[ScanAction["RescanTemplateToken"] = 3] = "RescanTemplateToken";
                ScanAction[ScanAction["RescanJsxIdentifier"] = 4] = "RescanJsxIdentifier";
                ScanAction[ScanAction["RescanJsxText"] = 5] = "RescanJsxText";
                ScanAction[ScanAction["RescanJsxAttributeValue"] = 6] = "RescanJsxAttributeValue";
            })(ScanAction || (ScanAction = {}));
            function getFormattingScanner(text, languageVariant, startPos, endPos, cb) {
                var scanner = languageVariant === 1 /* JSX */ ? jsxScanner : standardScanner;
                scanner.setText(text);
                scanner.setTextPos(startPos);
                var wasNewLine = true;
                var leadingTrivia;
                var trailingTrivia;
                var savedPos;
                var lastScanAction;
                var lastTokenInfo;
                var res = cb({
                    advance: advance,
                    readTokenInfo: readTokenInfo,
                    readEOFTokenRange: readEOFTokenRange,
                    isOnToken: isOnToken,
                    isOnEOF: isOnEOF,
                    getCurrentLeadingTrivia: function () { return leadingTrivia; },
                    lastTrailingTriviaWasNewLine: function () { return wasNewLine; },
                    skipToEndOf: skipToEndOf,
                    skipToStartOf: skipToStartOf,
                });
                lastTokenInfo = undefined;
                scanner.setText(undefined);
                return res;
                function advance() {
                    lastTokenInfo = undefined;
                    var isStarted = scanner.getStartPos() !== startPos;
                    if (isStarted) {
                        wasNewLine = !!trailingTrivia && ts.last(trailingTrivia).kind === 4 /* NewLineTrivia */;
                    }
                    else {
                        scanner.scan();
                    }
                    leadingTrivia = undefined;
                    trailingTrivia = undefined;
                    var pos = scanner.getStartPos();
                    // Read leading trivia and token
                    while (pos < endPos) {
                        var t = scanner.getToken();
                        if (!ts.isTrivia(t)) {
                            break;
                        }
                        // consume leading trivia
                        scanner.scan();
                        var item = {
                            pos: pos,
                            end: scanner.getStartPos(),
                            kind: t
                        };
                        pos = scanner.getStartPos();
                        leadingTrivia = ts.append(leadingTrivia, item);
                    }
                    savedPos = scanner.getStartPos();
                }
                function shouldRescanGreaterThanToken(node) {
                    switch (node.kind) {
                        case 33 /* GreaterThanEqualsToken */:
                        case 71 /* GreaterThanGreaterThanEqualsToken */:
                        case 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */:
                        case 49 /* GreaterThanGreaterThanGreaterThanToken */:
                        case 48 /* GreaterThanGreaterThanToken */:
                            return true;
                    }
                    return false;
                }
                function shouldRescanJsxIdentifier(node) {
                    if (node.parent) {
                        switch (node.parent.kind) {
                            case 283 /* JsxAttribute */:
                            case 278 /* JsxOpeningElement */:
                            case 279 /* JsxClosingElement */:
                            case 277 /* JsxSelfClosingElement */:
                                // May parse an identifier like `module-layout`; that will be scanned as a keyword at first, but we should parse the whole thing to get an identifier.
                                return ts.isKeyword(node.kind) || node.kind === 79 /* Identifier */;
                        }
                    }
                    return false;
                }
                function shouldRescanJsxText(node) {
                    return ts.isJsxText(node);
                }
                function shouldRescanSlashToken(container) {
                    return container.kind === 13 /* RegularExpressionLiteral */;
                }
                function shouldRescanTemplateToken(container) {
                    return container.kind === 16 /* TemplateMiddle */ ||
                        container.kind === 17 /* TemplateTail */;
                }
                function shouldRescanJsxAttributeValue(node) {
                    return node.parent && ts.isJsxAttribute(node.parent) && node.parent.initializer === node;
                }
                function startsWithSlashToken(t) {
                    return t === 43 /* SlashToken */ || t === 68 /* SlashEqualsToken */;
                }
                function readTokenInfo(n) {
                    ts.Debug.assert(isOnToken());
                    // normally scanner returns the smallest available token
                    // check the kind of context node to determine if scanner should have more greedy behavior and consume more text.
                    var expectedScanAction = shouldRescanGreaterThanToken(n) ? 1 /* RescanGreaterThanToken */ :
                        shouldRescanSlashToken(n) ? 2 /* RescanSlashToken */ :
                            shouldRescanTemplateToken(n) ? 3 /* RescanTemplateToken */ :
                                shouldRescanJsxIdentifier(n) ? 4 /* RescanJsxIdentifier */ :
                                    shouldRescanJsxText(n) ? 5 /* RescanJsxText */ :
                                        shouldRescanJsxAttributeValue(n) ? 6 /* RescanJsxAttributeValue */ :
                                            0 /* Scan */;
                    if (lastTokenInfo && expectedScanAction === lastScanAction) {
                        // readTokenInfo was called before with the same expected scan action.
                        // No need to re-scan text, return existing 'lastTokenInfo'
                        // it is ok to call fixTokenKind here since it does not affect
                        // what portion of text is consumed. In contrast rescanning can change it,
                        // i.e. for '>=' when originally scanner eats just one character
                        // and rescanning forces it to consume more.
                        return fixTokenKind(lastTokenInfo, n);
                    }
                    if (scanner.getStartPos() !== savedPos) {
                        ts.Debug.assert(lastTokenInfo !== undefined);
                        // readTokenInfo was called before but scan action differs - rescan text
                        scanner.setTextPos(savedPos);
                        scanner.scan();
                    }
                    var currentToken = getNextToken(n, expectedScanAction);
                    var token = formatting.createTextRangeWithKind(scanner.getStartPos(), scanner.getTextPos(), currentToken);
                    // consume trailing trivia
                    if (trailingTrivia) {
                        trailingTrivia = undefined;
                    }
                    while (scanner.getStartPos() < endPos) {
                        currentToken = scanner.scan();
                        if (!ts.isTrivia(currentToken)) {
                            break;
                        }
                        var trivia = formatting.createTextRangeWithKind(scanner.getStartPos(), scanner.getTextPos(), currentToken);
                        if (!trailingTrivia) {
                            trailingTrivia = [];
                        }
                        trailingTrivia.push(trivia);
                        if (currentToken === 4 /* NewLineTrivia */) {
                            // move past new line
                            scanner.scan();
                            break;
                        }
                    }
                    lastTokenInfo = { leadingTrivia: leadingTrivia, trailingTrivia: trailingTrivia, token: token };
                    return fixTokenKind(lastTokenInfo, n);
                }
                function getNextToken(n, expectedScanAction) {
                    var token = scanner.getToken();
                    lastScanAction = 0 /* Scan */;
                    switch (expectedScanAction) {
                        case 1 /* RescanGreaterThanToken */:
                            if (token === 31 /* GreaterThanToken */) {
                                lastScanAction = 1 /* RescanGreaterThanToken */;
                                var newToken = scanner.reScanGreaterToken();
                                ts.Debug.assert(n.kind === newToken);
                                return newToken;
                            }
                            break;
                        case 2 /* RescanSlashToken */:
                            if (startsWithSlashToken(token)) {
                                lastScanAction = 2 /* RescanSlashToken */;
                                var newToken = scanner.reScanSlashToken();
                                ts.Debug.assert(n.kind === newToken);
                                return newToken;
                            }
                            break;
                        case 3 /* RescanTemplateToken */:
                            if (token === 19 /* CloseBraceToken */) {
                                lastScanAction = 3 /* RescanTemplateToken */;
                                return scanner.reScanTemplateToken(/* isTaggedTemplate */ false);
                            }
                            break;
                        case 4 /* RescanJsxIdentifier */:
                            lastScanAction = 4 /* RescanJsxIdentifier */;
                            return scanner.scanJsxIdentifier();
                        case 5 /* RescanJsxText */:
                            lastScanAction = 5 /* RescanJsxText */;
                            return scanner.reScanJsxToken(/* allowMultilineJsxText */ false);
                        case 6 /* RescanJsxAttributeValue */:
                            lastScanAction = 6 /* RescanJsxAttributeValue */;
                            return scanner.reScanJsxAttributeValue();
                        case 0 /* Scan */:
                            break;
                        default:
                            ts.Debug.assertNever(expectedScanAction);
                    }
                    return token;
                }
                function readEOFTokenRange() {
                    ts.Debug.assert(isOnEOF());
                    return formatting.createTextRangeWithKind(scanner.getStartPos(), scanner.getTextPos(), 1 /* EndOfFileToken */);
                }
                function isOnToken() {
                    var current = lastTokenInfo ? lastTokenInfo.token.kind : scanner.getToken();
                    var startPos = lastTokenInfo ? lastTokenInfo.token.pos : scanner.getStartPos();
                    return startPos < endPos && current !== 1 /* EndOfFileToken */ && !ts.isTrivia(current);
                }
                function isOnEOF() {
                    var current = lastTokenInfo ? lastTokenInfo.token.kind : scanner.getToken();
                    return current === 1 /* EndOfFileToken */;
                }
                // when containing node in the tree is token
                // but its kind differs from the kind that was returned by the scanner,
                // then kind needs to be fixed. This might happen in cases
                // when parser interprets token differently, i.e keyword treated as identifier
                function fixTokenKind(tokenInfo, container) {
                    if (ts.isToken(container) && tokenInfo.token.kind !== container.kind) {
                        tokenInfo.token.kind = container.kind;
                    }
                    return tokenInfo;
                }
                function skipToEndOf(node) {
                    scanner.setTextPos(node.end);
                    savedPos = scanner.getStartPos();
                    lastScanAction = undefined;
                    lastTokenInfo = undefined;
                    wasNewLine = false;
                    leadingTrivia = undefined;
                    trailingTrivia = undefined;
                }
                function skipToStartOf(node) {
                    scanner.setTextPos(node.pos);
                    savedPos = scanner.getStartPos();
                    lastScanAction = undefined;
                    lastTokenInfo = undefined;
                    wasNewLine = false;
                    leadingTrivia = undefined;
                    trailingTrivia = undefined;
                }
            }
            formatting.getFormattingScanner = getFormattingScanner;
        })(formatting = ts.formatting || (ts.formatting = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var formatting;
        (function (formatting) {
            formatting.anyContext = ts.emptyArray;
            var RuleAction;
            (function (RuleAction) {
                RuleAction[RuleAction["StopProcessingSpaceActions"] = 1] = "StopProcessingSpaceActions";
                RuleAction[RuleAction["StopProcessingTokenActions"] = 2] = "StopProcessingTokenActions";
                RuleAction[RuleAction["InsertSpace"] = 4] = "InsertSpace";
                RuleAction[RuleAction["InsertNewLine"] = 8] = "InsertNewLine";
                RuleAction[RuleAction["DeleteSpace"] = 16] = "DeleteSpace";
                RuleAction[RuleAction["DeleteToken"] = 32] = "DeleteToken";
                RuleAction[RuleAction["InsertTrailingSemicolon"] = 64] = "InsertTrailingSemicolon";
                RuleAction[RuleAction["StopAction"] = 3] = "StopAction";
                RuleAction[RuleAction["ModifySpaceAction"] = 28] = "ModifySpaceAction";
                RuleAction[RuleAction["ModifyTokenAction"] = 96] = "ModifyTokenAction";
            })(RuleAction = formatting.RuleAction || (formatting.RuleAction = {}));
            var RuleFlags;
            (function (RuleFlags) {
                RuleFlags[RuleFlags["None"] = 0] = "None";
                RuleFlags[RuleFlags["CanDeleteNewLines"] = 1] = "CanDeleteNewLines";
            })(RuleFlags = formatting.RuleFlags || (formatting.RuleFlags = {}));
        })(formatting = ts.formatting || (ts.formatting = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var formatting;
        (function (formatting) {
            function getAllRules() {
                var allTokens = [];
                for (var token = 0 /* FirstToken */; token <= 158 /* LastToken */; token++) {
                    if (token !== 1 /* EndOfFileToken */) {
                        allTokens.push(token);
                    }
                }
                function anyTokenExcept() {
                    var tokens = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        tokens[_i] = arguments[_i];
                    }
                    return { tokens: allTokens.filter(function (t) { return !tokens.some(function (t2) { return t2 === t; }); }), isSpecific: false };
                }
                var anyToken = { tokens: allTokens, isSpecific: false };
                var anyTokenIncludingMultilineComments = tokenRangeFrom(__spreadArray(__spreadArray([], allTokens, true), [3 /* MultiLineCommentTrivia */], false));
                var anyTokenIncludingEOF = tokenRangeFrom(__spreadArray(__spreadArray([], allTokens, true), [1 /* EndOfFileToken */], false));
                var keywords = tokenRangeFromRange(81 /* FirstKeyword */, 158 /* LastKeyword */);
                var binaryOperators = tokenRangeFromRange(29 /* FirstBinaryOperator */, 78 /* LastBinaryOperator */);
                var binaryKeywordOperators = [101 /* InKeyword */, 102 /* InstanceOfKeyword */, 158 /* OfKeyword */, 127 /* AsKeyword */, 138 /* IsKeyword */];
                var unaryPrefixOperators = [45 /* PlusPlusToken */, 46 /* MinusMinusToken */, 54 /* TildeToken */, 53 /* ExclamationToken */];
                var unaryPrefixExpressions = [
                    8 /* NumericLiteral */, 9 /* BigIntLiteral */, 79 /* Identifier */, 20 /* OpenParenToken */,
                    22 /* OpenBracketToken */, 18 /* OpenBraceToken */, 108 /* ThisKeyword */, 103 /* NewKeyword */
                ];
                var unaryPreincrementExpressions = [79 /* Identifier */, 20 /* OpenParenToken */, 108 /* ThisKeyword */, 103 /* NewKeyword */];
                var unaryPostincrementExpressions = [79 /* Identifier */, 21 /* CloseParenToken */, 23 /* CloseBracketToken */, 103 /* NewKeyword */];
                var unaryPredecrementExpressions = [79 /* Identifier */, 20 /* OpenParenToken */, 108 /* ThisKeyword */, 103 /* NewKeyword */];
                var unaryPostdecrementExpressions = [79 /* Identifier */, 21 /* CloseParenToken */, 23 /* CloseBracketToken */, 103 /* NewKeyword */];
                var comments = [2 /* SingleLineCommentTrivia */, 3 /* MultiLineCommentTrivia */];
                var typeNames = __spreadArray([79 /* Identifier */], ts.typeKeywords, true);
                // Place a space before open brace in a function declaration
                // TypeScript: Function can have return types, which can be made of tons of different token kinds
                var functionOpenBraceLeftTokenRange = anyTokenIncludingMultilineComments;
                // Place a space before open brace in a TypeScript declaration that has braces as children (class, module, enum, etc)
                var typeScriptOpenBraceLeftTokenRange = tokenRangeFrom([79 /* Identifier */, 3 /* MultiLineCommentTrivia */, 84 /* ClassKeyword */, 93 /* ExportKeyword */, 100 /* ImportKeyword */]);
                // Place a space before open brace in a control flow construct
                var controlOpenBraceLeftTokenRange = tokenRangeFrom([21 /* CloseParenToken */, 3 /* MultiLineCommentTrivia */, 90 /* DoKeyword */, 111 /* TryKeyword */, 96 /* FinallyKeyword */, 91 /* ElseKeyword */]);
                // These rules are higher in priority than user-configurable
                var highPriorityCommonRules = [
                    // Leave comments alone
                    rule("IgnoreBeforeComment", anyToken, comments, formatting.anyContext, 1 /* StopProcessingSpaceActions */),
                    rule("IgnoreAfterLineComment", 2 /* SingleLineCommentTrivia */, anyToken, formatting.anyContext, 1 /* StopProcessingSpaceActions */),
                    rule("NotSpaceBeforeColon", anyToken, 58 /* ColonToken */, [isNonJsxSameLineTokenContext, isNotBinaryOpContext, isNotTypeAnnotationContext], 16 /* DeleteSpace */),
                    rule("SpaceAfterColon", 58 /* ColonToken */, anyToken, [isNonJsxSameLineTokenContext, isNotBinaryOpContext], 4 /* InsertSpace */),
                    rule("NoSpaceBeforeQuestionMark", anyToken, 57 /* QuestionToken */, [isNonJsxSameLineTokenContext, isNotBinaryOpContext, isNotTypeAnnotationContext], 16 /* DeleteSpace */),
                    // insert space after '?' only when it is used in conditional operator
                    rule("SpaceAfterQuestionMarkInConditionalOperator", 57 /* QuestionToken */, anyToken, [isNonJsxSameLineTokenContext, isConditionalOperatorContext], 4 /* InsertSpace */),
                    // in other cases there should be no space between '?' and next token
                    rule("NoSpaceAfterQuestionMark", 57 /* QuestionToken */, anyToken, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBeforeDot", anyToken, [24 /* DotToken */, 28 /* QuestionDotToken */], [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterDot", [24 /* DotToken */, 28 /* QuestionDotToken */], anyToken, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBetweenImportParenInImportType", 100 /* ImportKeyword */, 20 /* OpenParenToken */, [isNonJsxSameLineTokenContext, isImportTypeContext], 16 /* DeleteSpace */),
                    // Special handling of unary operators.
                    // Prefix operators generally shouldn't have a space between
                    // them and their target unary expression.
                    rule("NoSpaceAfterUnaryPrefixOperator", unaryPrefixOperators, unaryPrefixExpressions, [isNonJsxSameLineTokenContext, isNotBinaryOpContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterUnaryPreincrementOperator", 45 /* PlusPlusToken */, unaryPreincrementExpressions, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterUnaryPredecrementOperator", 46 /* MinusMinusToken */, unaryPredecrementExpressions, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBeforeUnaryPostincrementOperator", unaryPostincrementExpressions, 45 /* PlusPlusToken */, [isNonJsxSameLineTokenContext, isNotStatementConditionContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBeforeUnaryPostdecrementOperator", unaryPostdecrementExpressions, 46 /* MinusMinusToken */, [isNonJsxSameLineTokenContext, isNotStatementConditionContext], 16 /* DeleteSpace */),
                    // More unary operator special-casing.
                    // DevDiv 181814: Be careful when removing leading whitespace
                    // around unary operators.  Examples:
                    //      1 - -2  --X--> 1--2
                    //      a + ++b --X--> a+++b
                    rule("SpaceAfterPostincrementWhenFollowedByAdd", 45 /* PlusPlusToken */, 39 /* PlusToken */, [isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("SpaceAfterAddWhenFollowedByUnaryPlus", 39 /* PlusToken */, 39 /* PlusToken */, [isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("SpaceAfterAddWhenFollowedByPreincrement", 39 /* PlusToken */, 45 /* PlusPlusToken */, [isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("SpaceAfterPostdecrementWhenFollowedBySubtract", 46 /* MinusMinusToken */, 40 /* MinusToken */, [isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("SpaceAfterSubtractWhenFollowedByUnaryMinus", 40 /* MinusToken */, 40 /* MinusToken */, [isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("SpaceAfterSubtractWhenFollowedByPredecrement", 40 /* MinusToken */, 46 /* MinusMinusToken */, [isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("NoSpaceAfterCloseBrace", 19 /* CloseBraceToken */, [27 /* CommaToken */, 26 /* SemicolonToken */], [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // For functions and control block place } on a new line [multi-line rule]
                    rule("NewLineBeforeCloseBraceInBlockContext", anyTokenIncludingMultilineComments, 19 /* CloseBraceToken */, [isMultilineBlockContext], 8 /* InsertNewLine */),
                    // Space/new line after }.
                    rule("SpaceAfterCloseBrace", 19 /* CloseBraceToken */, anyTokenExcept(21 /* CloseParenToken */), [isNonJsxSameLineTokenContext, isAfterCodeBlockContext], 4 /* InsertSpace */),
                    // Special case for (}, else) and (}, while) since else & while tokens are not part of the tree which makes SpaceAfterCloseBrace rule not applied
                    // Also should not apply to })
                    rule("SpaceBetweenCloseBraceAndElse", 19 /* CloseBraceToken */, 91 /* ElseKeyword */, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("SpaceBetweenCloseBraceAndWhile", 19 /* CloseBraceToken */, 115 /* WhileKeyword */, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("NoSpaceBetweenEmptyBraceBrackets", 18 /* OpenBraceToken */, 19 /* CloseBraceToken */, [isNonJsxSameLineTokenContext, isObjectContext], 16 /* DeleteSpace */),
                    // Add a space after control dec context if the next character is an open bracket ex: 'if (false)[a, b] = [1, 2];' -> 'if (false) [a, b] = [1, 2];'
                    rule("SpaceAfterConditionalClosingParen", 21 /* CloseParenToken */, 22 /* OpenBracketToken */, [isControlDeclContext], 4 /* InsertSpace */),
                    rule("NoSpaceBetweenFunctionKeywordAndStar", 98 /* FunctionKeyword */, 41 /* AsteriskToken */, [isFunctionDeclarationOrFunctionExpressionContext], 16 /* DeleteSpace */),
                    rule("SpaceAfterStarInGeneratorDeclaration", 41 /* AsteriskToken */, 79 /* Identifier */, [isFunctionDeclarationOrFunctionExpressionContext], 4 /* InsertSpace */),
                    rule("SpaceAfterFunctionInFuncDecl", 98 /* FunctionKeyword */, anyToken, [isFunctionDeclContext], 4 /* InsertSpace */),
                    // Insert new line after { and before } in multi-line contexts.
                    rule("NewLineAfterOpenBraceInBlockContext", 18 /* OpenBraceToken */, anyToken, [isMultilineBlockContext], 8 /* InsertNewLine */),
                    // For get/set members, we check for (identifier,identifier) since get/set don't have tokens and they are represented as just an identifier token.
                    // Though, we do extra check on the context to make sure we are dealing with get/set node. Example:
                    //      get x() {}
                    //      set x(val) {}
                    rule("SpaceAfterGetSetInMember", [135 /* GetKeyword */, 147 /* SetKeyword */], 79 /* Identifier */, [isFunctionDeclContext], 4 /* InsertSpace */),
                    rule("NoSpaceBetweenYieldKeywordAndStar", 125 /* YieldKeyword */, 41 /* AsteriskToken */, [isNonJsxSameLineTokenContext, isYieldOrYieldStarWithOperand], 16 /* DeleteSpace */),
                    rule("SpaceBetweenYieldOrYieldStarAndOperand", [125 /* YieldKeyword */, 41 /* AsteriskToken */], anyToken, [isNonJsxSameLineTokenContext, isYieldOrYieldStarWithOperand], 4 /* InsertSpace */),
                    rule("NoSpaceBetweenReturnAndSemicolon", 105 /* ReturnKeyword */, 26 /* SemicolonToken */, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("SpaceAfterCertainKeywords", [113 /* VarKeyword */, 109 /* ThrowKeyword */, 103 /* NewKeyword */, 89 /* DeleteKeyword */, 105 /* ReturnKeyword */, 112 /* TypeOfKeyword */, 131 /* AwaitKeyword */], anyToken, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("SpaceAfterLetConstInVariableDeclaration", [119 /* LetKeyword */, 85 /* ConstKeyword */], anyToken, [isNonJsxSameLineTokenContext, isStartOfVariableDeclarationList], 4 /* InsertSpace */),
                    rule("NoSpaceBeforeOpenParenInFuncCall", anyToken, 20 /* OpenParenToken */, [isNonJsxSameLineTokenContext, isFunctionCallOrNewContext, isPreviousTokenNotComma], 16 /* DeleteSpace */),
                    // Special case for binary operators (that are keywords). For these we have to add a space and shouldn't follow any user options.
                    rule("SpaceBeforeBinaryKeywordOperator", anyToken, binaryKeywordOperators, [isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("SpaceAfterBinaryKeywordOperator", binaryKeywordOperators, anyToken, [isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("SpaceAfterVoidOperator", 114 /* VoidKeyword */, anyToken, [isNonJsxSameLineTokenContext, isVoidOpContext], 4 /* InsertSpace */),
                    // Async-await
                    rule("SpaceBetweenAsyncAndOpenParen", 130 /* AsyncKeyword */, 20 /* OpenParenToken */, [isArrowFunctionContext, isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("SpaceBetweenAsyncAndFunctionKeyword", 130 /* AsyncKeyword */, [98 /* FunctionKeyword */, 79 /* Identifier */], [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    // Template string
                    rule("NoSpaceBetweenTagAndTemplateString", [79 /* Identifier */, 21 /* CloseParenToken */], [14 /* NoSubstitutionTemplateLiteral */, 15 /* TemplateHead */], [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // JSX opening elements
                    rule("SpaceBeforeJsxAttribute", anyToken, 79 /* Identifier */, [isNextTokenParentJsxAttribute, isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("SpaceBeforeSlashInJsxOpeningElement", anyToken, 43 /* SlashToken */, [isJsxSelfClosingElementContext, isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("NoSpaceBeforeGreaterThanTokenInJsxOpeningElement", 43 /* SlashToken */, 31 /* GreaterThanToken */, [isJsxSelfClosingElementContext, isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBeforeEqualInJsxAttribute", anyToken, 63 /* EqualsToken */, [isJsxAttributeContext, isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterEqualInJsxAttribute", 63 /* EqualsToken */, anyToken, [isJsxAttributeContext, isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // TypeScript-specific rules
                    // Use of module as a function call. e.g.: import m2 = module("m2");
                    rule("NoSpaceAfterModuleImport", [140 /* ModuleKeyword */, 144 /* RequireKeyword */], 20 /* OpenParenToken */, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // Add a space around certain TypeScript keywords
                    rule("SpaceAfterCertainTypeScriptKeywords", [
                        126 /* AbstractKeyword */,
                        84 /* ClassKeyword */,
                        134 /* DeclareKeyword */,
                        88 /* DefaultKeyword */,
                        92 /* EnumKeyword */,
                        93 /* ExportKeyword */,
                        94 /* ExtendsKeyword */,
                        135 /* GetKeyword */,
                        117 /* ImplementsKeyword */,
                        100 /* ImportKeyword */,
                        118 /* InterfaceKeyword */,
                        140 /* ModuleKeyword */,
                        141 /* NamespaceKeyword */,
                        121 /* PrivateKeyword */,
                        123 /* PublicKeyword */,
                        122 /* ProtectedKeyword */,
                        143 /* ReadonlyKeyword */,
                        147 /* SetKeyword */,
                        124 /* StaticKeyword */,
                        150 /* TypeKeyword */,
                        154 /* FromKeyword */,
                        139 /* KeyOfKeyword */,
                        136 /* InferKeyword */,
                    ], anyToken, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("SpaceBeforeCertainTypeScriptKeywords", anyToken, [94 /* ExtendsKeyword */, 117 /* ImplementsKeyword */, 154 /* FromKeyword */], [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    // Treat string literals in module names as identifiers, and add a space between the literal and the opening Brace braces, e.g.: module "m2" {
                    rule("SpaceAfterModuleName", 10 /* StringLiteral */, 18 /* OpenBraceToken */, [isModuleDeclContext], 4 /* InsertSpace */),
                    // Lambda expressions
                    rule("SpaceBeforeArrow", anyToken, 38 /* EqualsGreaterThanToken */, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("SpaceAfterArrow", 38 /* EqualsGreaterThanToken */, anyToken, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    // Optional parameters and let args
                    rule("NoSpaceAfterEllipsis", 25 /* DotDotDotToken */, 79 /* Identifier */, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterOptionalParameters", 57 /* QuestionToken */, [21 /* CloseParenToken */, 27 /* CommaToken */], [isNonJsxSameLineTokenContext, isNotBinaryOpContext], 16 /* DeleteSpace */),
                    // Remove spaces in empty interface literals. e.g.: x: {}
                    rule("NoSpaceBetweenEmptyInterfaceBraceBrackets", 18 /* OpenBraceToken */, 19 /* CloseBraceToken */, [isNonJsxSameLineTokenContext, isObjectTypeContext], 16 /* DeleteSpace */),
                    // generics and type assertions
                    rule("NoSpaceBeforeOpenAngularBracket", typeNames, 29 /* LessThanToken */, [isNonJsxSameLineTokenContext, isTypeArgumentOrParameterOrAssertionContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBetweenCloseParenAndAngularBracket", 21 /* CloseParenToken */, 29 /* LessThanToken */, [isNonJsxSameLineTokenContext, isTypeArgumentOrParameterOrAssertionContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterOpenAngularBracket", 29 /* LessThanToken */, anyToken, [isNonJsxSameLineTokenContext, isTypeArgumentOrParameterOrAssertionContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBeforeCloseAngularBracket", anyToken, 31 /* GreaterThanToken */, [isNonJsxSameLineTokenContext, isTypeArgumentOrParameterOrAssertionContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterCloseAngularBracket", 31 /* GreaterThanToken */, [20 /* OpenParenToken */, 22 /* OpenBracketToken */, 31 /* GreaterThanToken */, 27 /* CommaToken */], [isNonJsxSameLineTokenContext, isTypeArgumentOrParameterOrAssertionContext, isNotFunctionDeclContext /*To prevent an interference with the SpaceBeforeOpenParenInFuncDecl rule*/], 16 /* DeleteSpace */),
                    // decorators
                    rule("SpaceBeforeAt", [21 /* CloseParenToken */, 79 /* Identifier */], 59 /* AtToken */, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("NoSpaceAfterAt", 59 /* AtToken */, anyToken, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // Insert space after @ in decorator
                    rule("SpaceAfterDecorator", anyToken, [
                        126 /* AbstractKeyword */,
                        79 /* Identifier */,
                        93 /* ExportKeyword */,
                        88 /* DefaultKeyword */,
                        84 /* ClassKeyword */,
                        124 /* StaticKeyword */,
                        123 /* PublicKeyword */,
                        121 /* PrivateKeyword */,
                        122 /* ProtectedKeyword */,
                        135 /* GetKeyword */,
                        147 /* SetKeyword */,
                        22 /* OpenBracketToken */,
                        41 /* AsteriskToken */,
                    ], [isEndOfDecoratorContextOnSameLine], 4 /* InsertSpace */),
                    rule("NoSpaceBeforeNonNullAssertionOperator", anyToken, 53 /* ExclamationToken */, [isNonJsxSameLineTokenContext, isNonNullAssertionContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterNewKeywordOnConstructorSignature", 103 /* NewKeyword */, 20 /* OpenParenToken */, [isNonJsxSameLineTokenContext, isConstructorSignatureContext], 16 /* DeleteSpace */),
                    rule("SpaceLessThanAndNonJSXTypeAnnotation", 29 /* LessThanToken */, 29 /* LessThanToken */, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                ];
                // These rules are applied after high priority
                var userConfigurableRules = [
                    // Treat constructor as an identifier in a function declaration, and remove spaces between constructor and following left parentheses
                    rule("SpaceAfterConstructor", 133 /* ConstructorKeyword */, 20 /* OpenParenToken */, [isOptionEnabled("insertSpaceAfterConstructor"), isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("NoSpaceAfterConstructor", 133 /* ConstructorKeyword */, 20 /* OpenParenToken */, [isOptionDisabledOrUndefined("insertSpaceAfterConstructor"), isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("SpaceAfterComma", 27 /* CommaToken */, anyToken, [isOptionEnabled("insertSpaceAfterCommaDelimiter"), isNonJsxSameLineTokenContext, isNonJsxElementOrFragmentContext, isNextTokenNotCloseBracket, isNextTokenNotCloseParen], 4 /* InsertSpace */),
                    rule("NoSpaceAfterComma", 27 /* CommaToken */, anyToken, [isOptionDisabledOrUndefined("insertSpaceAfterCommaDelimiter"), isNonJsxSameLineTokenContext, isNonJsxElementOrFragmentContext], 16 /* DeleteSpace */),
                    // Insert space after function keyword for anonymous functions
                    rule("SpaceAfterAnonymousFunctionKeyword", [98 /* FunctionKeyword */, 41 /* AsteriskToken */], 20 /* OpenParenToken */, [isOptionEnabled("insertSpaceAfterFunctionKeywordForAnonymousFunctions"), isFunctionDeclContext], 4 /* InsertSpace */),
                    rule("NoSpaceAfterAnonymousFunctionKeyword", [98 /* FunctionKeyword */, 41 /* AsteriskToken */], 20 /* OpenParenToken */, [isOptionDisabledOrUndefined("insertSpaceAfterFunctionKeywordForAnonymousFunctions"), isFunctionDeclContext], 16 /* DeleteSpace */),
                    // Insert space after keywords in control flow statements
                    rule("SpaceAfterKeywordInControl", keywords, 20 /* OpenParenToken */, [isOptionEnabled("insertSpaceAfterKeywordsInControlFlowStatements"), isControlDeclContext], 4 /* InsertSpace */),
                    rule("NoSpaceAfterKeywordInControl", keywords, 20 /* OpenParenToken */, [isOptionDisabledOrUndefined("insertSpaceAfterKeywordsInControlFlowStatements"), isControlDeclContext], 16 /* DeleteSpace */),
                    // Insert space after opening and before closing nonempty parenthesis
                    rule("SpaceAfterOpenParen", 20 /* OpenParenToken */, anyToken, [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"), isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("SpaceBeforeCloseParen", anyToken, 21 /* CloseParenToken */, [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"), isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("SpaceBetweenOpenParens", 20 /* OpenParenToken */, 20 /* OpenParenToken */, [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"), isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("NoSpaceBetweenParens", 20 /* OpenParenToken */, 21 /* CloseParenToken */, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterOpenParen", 20 /* OpenParenToken */, anyToken, [isOptionDisabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"), isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBeforeCloseParen", anyToken, 21 /* CloseParenToken */, [isOptionDisabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis"), isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // Insert space after opening and before closing nonempty brackets
                    rule("SpaceAfterOpenBracket", 22 /* OpenBracketToken */, anyToken, [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"), isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("SpaceBeforeCloseBracket", anyToken, 23 /* CloseBracketToken */, [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"), isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("NoSpaceBetweenBrackets", 22 /* OpenBracketToken */, 23 /* CloseBracketToken */, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterOpenBracket", 22 /* OpenBracketToken */, anyToken, [isOptionDisabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"), isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBeforeCloseBracket", anyToken, 23 /* CloseBracketToken */, [isOptionDisabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets"), isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // Insert a space after { and before } in single-line contexts, but remove space from empty object literals {}.
                    rule("SpaceAfterOpenBrace", 18 /* OpenBraceToken */, anyToken, [isOptionEnabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"), isBraceWrappedContext], 4 /* InsertSpace */),
                    rule("SpaceBeforeCloseBrace", anyToken, 19 /* CloseBraceToken */, [isOptionEnabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"), isBraceWrappedContext], 4 /* InsertSpace */),
                    rule("NoSpaceBetweenEmptyBraceBrackets", 18 /* OpenBraceToken */, 19 /* CloseBraceToken */, [isNonJsxSameLineTokenContext, isObjectContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterOpenBrace", 18 /* OpenBraceToken */, anyToken, [isOptionDisabled("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"), isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBeforeCloseBrace", anyToken, 19 /* CloseBraceToken */, [isOptionDisabled("insertSpaceAfterOpeningAndBeforeClosingNonemptyBraces"), isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // Insert a space after opening and before closing empty brace brackets
                    rule("SpaceBetweenEmptyBraceBrackets", 18 /* OpenBraceToken */, 19 /* CloseBraceToken */, [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingEmptyBraces")], 4 /* InsertSpace */),
                    rule("NoSpaceBetweenEmptyBraceBrackets", 18 /* OpenBraceToken */, 19 /* CloseBraceToken */, [isOptionDisabled("insertSpaceAfterOpeningAndBeforeClosingEmptyBraces"), isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // Insert space after opening and before closing template string braces
                    rule("SpaceAfterTemplateHeadAndMiddle", [15 /* TemplateHead */, 16 /* TemplateMiddle */], anyToken, [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"), isNonJsxTextContext], 4 /* InsertSpace */, 1 /* CanDeleteNewLines */),
                    rule("SpaceBeforeTemplateMiddleAndTail", anyToken, [16 /* TemplateMiddle */, 17 /* TemplateTail */], [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"), isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    rule("NoSpaceAfterTemplateHeadAndMiddle", [15 /* TemplateHead */, 16 /* TemplateMiddle */], anyToken, [isOptionDisabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"), isNonJsxTextContext], 16 /* DeleteSpace */, 1 /* CanDeleteNewLines */),
                    rule("NoSpaceBeforeTemplateMiddleAndTail", anyToken, [16 /* TemplateMiddle */, 17 /* TemplateTail */], [isOptionDisabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces"), isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // No space after { and before } in JSX expression
                    rule("SpaceAfterOpenBraceInJsxExpression", 18 /* OpenBraceToken */, anyToken, [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"), isNonJsxSameLineTokenContext, isJsxExpressionContext], 4 /* InsertSpace */),
                    rule("SpaceBeforeCloseBraceInJsxExpression", anyToken, 19 /* CloseBraceToken */, [isOptionEnabled("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"), isNonJsxSameLineTokenContext, isJsxExpressionContext], 4 /* InsertSpace */),
                    rule("NoSpaceAfterOpenBraceInJsxExpression", 18 /* OpenBraceToken */, anyToken, [isOptionDisabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"), isNonJsxSameLineTokenContext, isJsxExpressionContext], 16 /* DeleteSpace */),
                    rule("NoSpaceBeforeCloseBraceInJsxExpression", anyToken, 19 /* CloseBraceToken */, [isOptionDisabledOrUndefined("insertSpaceAfterOpeningAndBeforeClosingJsxExpressionBraces"), isNonJsxSameLineTokenContext, isJsxExpressionContext], 16 /* DeleteSpace */),
                    // Insert space after semicolon in for statement
                    rule("SpaceAfterSemicolonInFor", 26 /* SemicolonToken */, anyToken, [isOptionEnabled("insertSpaceAfterSemicolonInForStatements"), isNonJsxSameLineTokenContext, isForContext], 4 /* InsertSpace */),
                    rule("NoSpaceAfterSemicolonInFor", 26 /* SemicolonToken */, anyToken, [isOptionDisabledOrUndefined("insertSpaceAfterSemicolonInForStatements"), isNonJsxSameLineTokenContext, isForContext], 16 /* DeleteSpace */),
                    // Insert space before and after binary operators
                    rule("SpaceBeforeBinaryOperator", anyToken, binaryOperators, [isOptionEnabled("insertSpaceBeforeAndAfterBinaryOperators"), isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("SpaceAfterBinaryOperator", binaryOperators, anyToken, [isOptionEnabled("insertSpaceBeforeAndAfterBinaryOperators"), isNonJsxSameLineTokenContext, isBinaryOpContext], 4 /* InsertSpace */),
                    rule("NoSpaceBeforeBinaryOperator", anyToken, binaryOperators, [isOptionDisabledOrUndefined("insertSpaceBeforeAndAfterBinaryOperators"), isNonJsxSameLineTokenContext, isBinaryOpContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterBinaryOperator", binaryOperators, anyToken, [isOptionDisabledOrUndefined("insertSpaceBeforeAndAfterBinaryOperators"), isNonJsxSameLineTokenContext, isBinaryOpContext], 16 /* DeleteSpace */),
                    rule("SpaceBeforeOpenParenInFuncDecl", anyToken, 20 /* OpenParenToken */, [isOptionEnabled("insertSpaceBeforeFunctionParenthesis"), isNonJsxSameLineTokenContext, isFunctionDeclContext], 4 /* InsertSpace */),
                    rule("NoSpaceBeforeOpenParenInFuncDecl", anyToken, 20 /* OpenParenToken */, [isOptionDisabledOrUndefined("insertSpaceBeforeFunctionParenthesis"), isNonJsxSameLineTokenContext, isFunctionDeclContext], 16 /* DeleteSpace */),
                    // Open Brace braces after control block
                    rule("NewLineBeforeOpenBraceInControl", controlOpenBraceLeftTokenRange, 18 /* OpenBraceToken */, [isOptionEnabled("placeOpenBraceOnNewLineForControlBlocks"), isControlDeclContext, isBeforeMultilineBlockContext], 8 /* InsertNewLine */, 1 /* CanDeleteNewLines */),
                    // Open Brace braces after function
                    // TypeScript: Function can have return types, which can be made of tons of different token kinds
                    rule("NewLineBeforeOpenBraceInFunction", functionOpenBraceLeftTokenRange, 18 /* OpenBraceToken */, [isOptionEnabled("placeOpenBraceOnNewLineForFunctions"), isFunctionDeclContext, isBeforeMultilineBlockContext], 8 /* InsertNewLine */, 1 /* CanDeleteNewLines */),
                    // Open Brace braces after TypeScript module/class/interface
                    rule("NewLineBeforeOpenBraceInTypeScriptDeclWithBlock", typeScriptOpenBraceLeftTokenRange, 18 /* OpenBraceToken */, [isOptionEnabled("placeOpenBraceOnNewLineForFunctions"), isTypeScriptDeclWithBlockContext, isBeforeMultilineBlockContext], 8 /* InsertNewLine */, 1 /* CanDeleteNewLines */),
                    rule("SpaceAfterTypeAssertion", 31 /* GreaterThanToken */, anyToken, [isOptionEnabled("insertSpaceAfterTypeAssertion"), isNonJsxSameLineTokenContext, isTypeAssertionContext], 4 /* InsertSpace */),
                    rule("NoSpaceAfterTypeAssertion", 31 /* GreaterThanToken */, anyToken, [isOptionDisabledOrUndefined("insertSpaceAfterTypeAssertion"), isNonJsxSameLineTokenContext, isTypeAssertionContext], 16 /* DeleteSpace */),
                    rule("SpaceBeforeTypeAnnotation", anyToken, [57 /* QuestionToken */, 58 /* ColonToken */], [isOptionEnabled("insertSpaceBeforeTypeAnnotation"), isNonJsxSameLineTokenContext, isTypeAnnotationContext], 4 /* InsertSpace */),
                    rule("NoSpaceBeforeTypeAnnotation", anyToken, [57 /* QuestionToken */, 58 /* ColonToken */], [isOptionDisabledOrUndefined("insertSpaceBeforeTypeAnnotation"), isNonJsxSameLineTokenContext, isTypeAnnotationContext], 16 /* DeleteSpace */),
                    rule("NoOptionalSemicolon", 26 /* SemicolonToken */, anyTokenIncludingEOF, [optionEquals("semicolons", ts.SemicolonPreference.Remove), isSemicolonDeletionContext], 32 /* DeleteToken */),
                    rule("OptionalSemicolon", anyToken, anyTokenIncludingEOF, [optionEquals("semicolons", ts.SemicolonPreference.Insert), isSemicolonInsertionContext], 64 /* InsertTrailingSemicolon */),
                ];
                // These rules are lower in priority than user-configurable. Rules earlier in this list have priority over rules later in the list.
                var lowPriorityCommonRules = [
                    // Space after keyword but not before ; or : or ?
                    rule("NoSpaceBeforeSemicolon", anyToken, 26 /* SemicolonToken */, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("SpaceBeforeOpenBraceInControl", controlOpenBraceLeftTokenRange, 18 /* OpenBraceToken */, [isOptionDisabledOrUndefinedOrTokensOnSameLine("placeOpenBraceOnNewLineForControlBlocks"), isControlDeclContext, isNotFormatOnEnter, isSameLineTokenOrBeforeBlockContext], 4 /* InsertSpace */, 1 /* CanDeleteNewLines */),
                    rule("SpaceBeforeOpenBraceInFunction", functionOpenBraceLeftTokenRange, 18 /* OpenBraceToken */, [isOptionDisabledOrUndefinedOrTokensOnSameLine("placeOpenBraceOnNewLineForFunctions"), isFunctionDeclContext, isBeforeBlockContext, isNotFormatOnEnter, isSameLineTokenOrBeforeBlockContext], 4 /* InsertSpace */, 1 /* CanDeleteNewLines */),
                    rule("SpaceBeforeOpenBraceInTypeScriptDeclWithBlock", typeScriptOpenBraceLeftTokenRange, 18 /* OpenBraceToken */, [isOptionDisabledOrUndefinedOrTokensOnSameLine("placeOpenBraceOnNewLineForFunctions"), isTypeScriptDeclWithBlockContext, isNotFormatOnEnter, isSameLineTokenOrBeforeBlockContext], 4 /* InsertSpace */, 1 /* CanDeleteNewLines */),
                    rule("NoSpaceBeforeComma", anyToken, 27 /* CommaToken */, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    // No space before and after indexer `x[]`
                    rule("NoSpaceBeforeOpenBracket", anyTokenExcept(130 /* AsyncKeyword */, 82 /* CaseKeyword */), 22 /* OpenBracketToken */, [isNonJsxSameLineTokenContext], 16 /* DeleteSpace */),
                    rule("NoSpaceAfterCloseBracket", 23 /* CloseBracketToken */, anyToken, [isNonJsxSameLineTokenContext, isNotBeforeBlockInFunctionDeclarationContext], 16 /* DeleteSpace */),
                    rule("SpaceAfterSemicolon", 26 /* SemicolonToken */, anyToken, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    // Remove extra space between for and await
                    rule("SpaceBetweenForAndAwaitKeyword", 97 /* ForKeyword */, 131 /* AwaitKeyword */, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                    // Add a space between statements. All keywords except (do,else,case) has open/close parens after them.
                    // So, we have a rule to add a space for [),Any], [do,Any], [else,Any], and [case,Any]
                    rule("SpaceBetweenStatements", [21 /* CloseParenToken */, 90 /* DoKeyword */, 91 /* ElseKeyword */, 82 /* CaseKeyword */], anyToken, [isNonJsxSameLineTokenContext, isNonJsxElementOrFragmentContext, isNotForContext], 4 /* InsertSpace */),
                    // This low-pri rule takes care of "try {", "catch {" and "finally {" in case the rule SpaceBeforeOpenBraceInControl didn't execute on FormatOnEnter.
                    rule("SpaceAfterTryCatchFinally", [111 /* TryKeyword */, 83 /* CatchKeyword */, 96 /* FinallyKeyword */], 18 /* OpenBraceToken */, [isNonJsxSameLineTokenContext], 4 /* InsertSpace */),
                ];
                return __spreadArray(__spreadArray(__spreadArray([], highPriorityCommonRules, true), userConfigurableRules, true), lowPriorityCommonRules, true);
            }
            formatting.getAllRules = getAllRules;
            /**
             * A rule takes a two tokens (left/right) and a particular context
             * for which you're meant to look at them. You then declare what should the
             * whitespace annotation be between these tokens via the action param.
             *
             * @param debugName Name to print
             * @param left The left side of the comparison
             * @param right The right side of the comparison
             * @param context A set of filters to narrow down the space in which this formatter rule applies
             * @param action a declaration of the expected whitespace
             * @param flags whether the rule deletes a line or not, defaults to no-op
             */
            function rule(debugName, left, right, context, action, flags) {
                if (flags === void 0) {
                    flags = 0 /* None */;
                }
                return { leftTokenRange: toTokenRange(left), rightTokenRange: toTokenRange(right), rule: { debugName: debugName, context: context, action: action, flags: flags } };
            }
            function tokenRangeFrom(tokens) {
                return { tokens: tokens, isSpecific: true };
            }
            function toTokenRange(arg) {
                return typeof arg === "number" ? tokenRangeFrom([arg]) : ts.isArray(arg) ? tokenRangeFrom(arg) : arg;
            }
            function tokenRangeFromRange(from, to, except) {
                if (except === void 0) {
                    except = [];
                }
                var tokens = [];
                for (var token = from; token <= to; token++) {
                    if (!ts.contains(except, token)) {
                        tokens.push(token);
                    }
                }
                return tokenRangeFrom(tokens);
            }
            ///
            /// Contexts
            ///
            function optionEquals(optionName, optionValue) {
                return function (context) { return context.options && context.options[optionName] === optionValue; };
            }
            function isOptionEnabled(optionName) {
                return function (context) { return context.options && context.options.hasOwnProperty(optionName) && !!context.options[optionName]; };
            }
            function isOptionDisabled(optionName) {
                return function (context) { return context.options && context.options.hasOwnProperty(optionName) && !context.options[optionName]; };
            }
            function isOptionDisabledOrUndefined(optionName) {
                return function (context) { return !context.options || !context.options.hasOwnProperty(optionName) || !context.options[optionName]; };
            }
            function isOptionDisabledOrUndefinedOrTokensOnSameLine(optionName) {
                return function (context) { return !context.options || !context.options.hasOwnProperty(optionName) || !context.options[optionName] || context.TokensAreOnSameLine(); };
            }
            function isOptionEnabledOrUndefined(optionName) {
                return function (context) { return !context.options || !context.options.hasOwnProperty(optionName) || !!context.options[optionName]; };
            }
            function isForContext(context) {
                return context.contextNode.kind === 240 /* ForStatement */;
            }
            function isNotForContext(context) {
                return !isForContext(context);
            }
            function isBinaryOpContext(context) {
                switch (context.contextNode.kind) {
                    case 219 /* BinaryExpression */:
                        return context.contextNode.operatorToken.kind !== 27 /* CommaToken */;
                    case 220 /* ConditionalExpression */:
                    case 187 /* ConditionalType */:
                    case 227 /* AsExpression */:
                    case 273 /* ExportSpecifier */:
                    case 268 /* ImportSpecifier */:
                    case 175 /* TypePredicate */:
                    case 185 /* UnionType */:
                    case 186 /* IntersectionType */:
                        return true;
                    // equals in binding elements: function foo([[x, y] = [1, 2]])
                    case 201 /* BindingElement */:
                    // equals in type X = ...
                    // falls through
                    case 257 /* TypeAliasDeclaration */:
                    // equal in import a = module('a');
                    // falls through
                    case 263 /* ImportEqualsDeclaration */:
                    // equal in export = 1
                    // falls through
                    case 269 /* ExportAssignment */:
                    // equal in let a = 0
                    // falls through
                    case 252 /* VariableDeclaration */:
                    // equal in p = 0
                    // falls through
                    case 162 /* Parameter */:
                    case 294 /* EnumMember */:
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                        return context.currentTokenSpan.kind === 63 /* EqualsToken */ || context.nextTokenSpan.kind === 63 /* EqualsToken */;
                    // "in" keyword in for (let x in []) { }
                    case 241 /* ForInStatement */:
                    // "in" keyword in [P in keyof T]: T[P]
                    // falls through
                    case 161 /* TypeParameter */:
                        return context.currentTokenSpan.kind === 101 /* InKeyword */ || context.nextTokenSpan.kind === 101 /* InKeyword */ || context.currentTokenSpan.kind === 63 /* EqualsToken */ || context.nextTokenSpan.kind === 63 /* EqualsToken */;
                    // Technically, "of" is not a binary operator, but format it the same way as "in"
                    case 242 /* ForOfStatement */:
                        return context.currentTokenSpan.kind === 158 /* OfKeyword */ || context.nextTokenSpan.kind === 158 /* OfKeyword */;
                }
                return false;
            }
            function isNotBinaryOpContext(context) {
                return !isBinaryOpContext(context);
            }
            function isNotTypeAnnotationContext(context) {
                return !isTypeAnnotationContext(context);
            }
            function isTypeAnnotationContext(context) {
                var contextKind = context.contextNode.kind;
                return contextKind === 165 /* PropertyDeclaration */ ||
                    contextKind === 164 /* PropertySignature */ ||
                    contextKind === 162 /* Parameter */ ||
                    contextKind === 252 /* VariableDeclaration */ ||
                    ts.isFunctionLikeKind(contextKind);
            }
            function isConditionalOperatorContext(context) {
                return context.contextNode.kind === 220 /* ConditionalExpression */ ||
                    context.contextNode.kind === 187 /* ConditionalType */;
            }
            function isSameLineTokenOrBeforeBlockContext(context) {
                return context.TokensAreOnSameLine() || isBeforeBlockContext(context);
            }
            function isBraceWrappedContext(context) {
                return context.contextNode.kind === 199 /* ObjectBindingPattern */ ||
                    context.contextNode.kind === 193 /* MappedType */ ||
                    isSingleLineBlockContext(context);
            }
            // This check is done before an open brace in a control construct, a function, or a typescript block declaration
            function isBeforeMultilineBlockContext(context) {
                return isBeforeBlockContext(context) && !(context.NextNodeAllOnSameLine() || context.NextNodeBlockIsOnOneLine());
            }
            function isMultilineBlockContext(context) {
                return isBlockContext(context) && !(context.ContextNodeAllOnSameLine() || context.ContextNodeBlockIsOnOneLine());
            }
            function isSingleLineBlockContext(context) {
                return isBlockContext(context) && (context.ContextNodeAllOnSameLine() || context.ContextNodeBlockIsOnOneLine());
            }
            function isBlockContext(context) {
                return nodeIsBlockContext(context.contextNode);
            }
            function isBeforeBlockContext(context) {
                return nodeIsBlockContext(context.nextTokenParent);
            }
            // IMPORTANT!!! This method must return true ONLY for nodes with open and close braces as immediate children
            function nodeIsBlockContext(node) {
                if (nodeIsTypeScriptDeclWithBlockContext(node)) {
                    // This means we are in a context that looks like a block to the user, but in the grammar is actually not a node (it's a class, module, enum, object type literal, etc).
                    return true;
                }
                switch (node.kind) {
                    case 233 /* Block */:
                    case 261 /* CaseBlock */:
                    case 203 /* ObjectLiteralExpression */:
                    case 260 /* ModuleBlock */:
                        return true;
                }
                return false;
            }
            function isFunctionDeclContext(context) {
                switch (context.contextNode.kind) {
                    case 254 /* FunctionDeclaration */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    // case SyntaxKind.MemberFunctionDeclaration:
                    // falls through
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                    // case SyntaxKind.MethodSignature:
                    // falls through
                    case 172 /* CallSignature */:
                    case 211 /* FunctionExpression */:
                    case 169 /* Constructor */:
                    case 212 /* ArrowFunction */:
                    // case SyntaxKind.ConstructorDeclaration:
                    // case SyntaxKind.SimpleArrowFunctionExpression:
                    // case SyntaxKind.ParenthesizedArrowFunctionExpression:
                    // falls through
                    case 256 /* InterfaceDeclaration */: // This one is not truly a function, but for formatting purposes, it acts just like one
                        return true;
                }
                return false;
            }
            function isNotFunctionDeclContext(context) {
                return !isFunctionDeclContext(context);
            }
            function isFunctionDeclarationOrFunctionExpressionContext(context) {
                return context.contextNode.kind === 254 /* FunctionDeclaration */ || context.contextNode.kind === 211 /* FunctionExpression */;
            }
            function isTypeScriptDeclWithBlockContext(context) {
                return nodeIsTypeScriptDeclWithBlockContext(context.contextNode);
            }
            function nodeIsTypeScriptDeclWithBlockContext(node) {
                switch (node.kind) {
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                    case 256 /* InterfaceDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 180 /* TypeLiteral */:
                    case 259 /* ModuleDeclaration */:
                    case 270 /* ExportDeclaration */:
                    case 271 /* NamedExports */:
                    case 264 /* ImportDeclaration */:
                    case 267 /* NamedImports */:
                        return true;
                }
                return false;
            }
            function isAfterCodeBlockContext(context) {
                switch (context.currentTokenParent.kind) {
                    case 255 /* ClassDeclaration */:
                    case 259 /* ModuleDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 290 /* CatchClause */:
                    case 260 /* ModuleBlock */:
                    case 247 /* SwitchStatement */:
                        return true;
                    case 233 /* Block */: {
                        var blockParent = context.currentTokenParent.parent;
                        // In a codefix scenario, we can't rely on parents being set. So just always return true.
                        if (!blockParent || blockParent.kind !== 212 /* ArrowFunction */ && blockParent.kind !== 211 /* FunctionExpression */) {
                            return true;
                        }
                    }
                }
                return false;
            }
            function isControlDeclContext(context) {
                switch (context.contextNode.kind) {
                    case 237 /* IfStatement */:
                    case 247 /* SwitchStatement */:
                    case 240 /* ForStatement */:
                    case 241 /* ForInStatement */:
                    case 242 /* ForOfStatement */:
                    case 239 /* WhileStatement */:
                    case 250 /* TryStatement */:
                    case 238 /* DoStatement */:
                    case 246 /* WithStatement */:
                    // TODO
                    // case SyntaxKind.ElseClause:
                    // falls through
                    case 290 /* CatchClause */:
                        return true;
                    default:
                        return false;
                }
            }
            function isObjectContext(context) {
                return context.contextNode.kind === 203 /* ObjectLiteralExpression */;
            }
            function isFunctionCallContext(context) {
                return context.contextNode.kind === 206 /* CallExpression */;
            }
            function isNewContext(context) {
                return context.contextNode.kind === 207 /* NewExpression */;
            }
            function isFunctionCallOrNewContext(context) {
                return isFunctionCallContext(context) || isNewContext(context);
            }
            function isPreviousTokenNotComma(context) {
                return context.currentTokenSpan.kind !== 27 /* CommaToken */;
            }
            function isNextTokenNotCloseBracket(context) {
                return context.nextTokenSpan.kind !== 23 /* CloseBracketToken */;
            }
            function isNextTokenNotCloseParen(context) {
                return context.nextTokenSpan.kind !== 21 /* CloseParenToken */;
            }
            function isArrowFunctionContext(context) {
                return context.contextNode.kind === 212 /* ArrowFunction */;
            }
            function isImportTypeContext(context) {
                return context.contextNode.kind === 198 /* ImportType */;
            }
            function isNonJsxSameLineTokenContext(context) {
                return context.TokensAreOnSameLine() && context.contextNode.kind !== 11 /* JsxText */;
            }
            function isNonJsxTextContext(context) {
                return context.contextNode.kind !== 11 /* JsxText */;
            }
            function isNonJsxElementOrFragmentContext(context) {
                return context.contextNode.kind !== 276 /* JsxElement */ && context.contextNode.kind !== 280 /* JsxFragment */;
            }
            function isJsxExpressionContext(context) {
                return context.contextNode.kind === 286 /* JsxExpression */ || context.contextNode.kind === 285 /* JsxSpreadAttribute */;
            }
            function isNextTokenParentJsxAttribute(context) {
                return context.nextTokenParent.kind === 283 /* JsxAttribute */;
            }
            function isJsxAttributeContext(context) {
                return context.contextNode.kind === 283 /* JsxAttribute */;
            }
            function isJsxSelfClosingElementContext(context) {
                return context.contextNode.kind === 277 /* JsxSelfClosingElement */;
            }
            function isNotBeforeBlockInFunctionDeclarationContext(context) {
                return !isFunctionDeclContext(context) && !isBeforeBlockContext(context);
            }
            function isEndOfDecoratorContextOnSameLine(context) {
                return context.TokensAreOnSameLine() &&
                    !!context.contextNode.decorators &&
                    nodeIsInDecoratorContext(context.currentTokenParent) &&
                    !nodeIsInDecoratorContext(context.nextTokenParent);
            }
            function nodeIsInDecoratorContext(node) {
                while (ts.isExpressionNode(node)) {
                    node = node.parent;
                }
                return node.kind === 163 /* Decorator */;
            }
            function isStartOfVariableDeclarationList(context) {
                return context.currentTokenParent.kind === 253 /* VariableDeclarationList */ &&
                    context.currentTokenParent.getStart(context.sourceFile) === context.currentTokenSpan.pos;
            }
            function isNotFormatOnEnter(context) {
                return context.formattingRequestKind !== 2 /* FormatOnEnter */;
            }
            function isModuleDeclContext(context) {
                return context.contextNode.kind === 259 /* ModuleDeclaration */;
            }
            function isObjectTypeContext(context) {
                return context.contextNode.kind === 180 /* TypeLiteral */; // && context.contextNode.parent.kind !== SyntaxKind.InterfaceDeclaration;
            }
            function isConstructorSignatureContext(context) {
                return context.contextNode.kind === 173 /* ConstructSignature */;
            }
            function isTypeArgumentOrParameterOrAssertion(token, parent) {
                if (token.kind !== 29 /* LessThanToken */ && token.kind !== 31 /* GreaterThanToken */) {
                    return false;
                }
                switch (parent.kind) {
                    case 176 /* TypeReference */:
                    case 209 /* TypeAssertionExpression */:
                    case 257 /* TypeAliasDeclaration */:
                    case 255 /* ClassDeclaration */:
                    case 224 /* ClassExpression */:
                    case 256 /* InterfaceDeclaration */:
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 172 /* CallSignature */:
                    case 173 /* ConstructSignature */:
                    case 206 /* CallExpression */:
                    case 207 /* NewExpression */:
                    case 226 /* ExpressionWithTypeArguments */:
                        return true;
                    default:
                        return false;
                }
            }
            function isTypeArgumentOrParameterOrAssertionContext(context) {
                return isTypeArgumentOrParameterOrAssertion(context.currentTokenSpan, context.currentTokenParent) ||
                    isTypeArgumentOrParameterOrAssertion(context.nextTokenSpan, context.nextTokenParent);
            }
            function isTypeAssertionContext(context) {
                return context.contextNode.kind === 209 /* TypeAssertionExpression */;
            }
            function isVoidOpContext(context) {
                return context.currentTokenSpan.kind === 114 /* VoidKeyword */ && context.currentTokenParent.kind === 215 /* VoidExpression */;
            }
            function isYieldOrYieldStarWithOperand(context) {
                return context.contextNode.kind === 222 /* YieldExpression */ && context.contextNode.expression !== undefined;
            }
            function isNonNullAssertionContext(context) {
                return context.contextNode.kind === 228 /* NonNullExpression */;
            }
            function isNotStatementConditionContext(context) {
                return !isStatementConditionContext(context);
            }
            function isStatementConditionContext(context) {
                switch (context.contextNode.kind) {
                    case 237 /* IfStatement */:
                    case 240 /* ForStatement */:
                    case 241 /* ForInStatement */:
                    case 242 /* ForOfStatement */:
                    case 238 /* DoStatement */:
                    case 239 /* WhileStatement */:
                        return true;
                    default:
                        return false;
                }
            }
            function isSemicolonDeletionContext(context) {
                var nextTokenKind = context.nextTokenSpan.kind;
                var nextTokenStart = context.nextTokenSpan.pos;
                if (ts.isTrivia(nextTokenKind)) {
                    var nextRealToken = context.nextTokenParent === context.currentTokenParent
                        ? ts.findNextToken(context.currentTokenParent, ts.findAncestor(context.currentTokenParent, function (a) { return !a.parent; }), context.sourceFile)
                        : context.nextTokenParent.getFirstToken(context.sourceFile);
                    if (!nextRealToken) {
                        return true;
                    }
                    nextTokenKind = nextRealToken.kind;
                    nextTokenStart = nextRealToken.getStart(context.sourceFile);
                }
                var startLine = context.sourceFile.getLineAndCharacterOfPosition(context.currentTokenSpan.pos).line;
                var endLine = context.sourceFile.getLineAndCharacterOfPosition(nextTokenStart).line;
                if (startLine === endLine) {
                    return nextTokenKind === 19 /* CloseBraceToken */
                        || nextTokenKind === 1 /* EndOfFileToken */;
                }
                if (nextTokenKind === 232 /* SemicolonClassElement */ ||
                    nextTokenKind === 26 /* SemicolonToken */) {
                    return false;
                }
                if (context.contextNode.kind === 256 /* InterfaceDeclaration */ ||
                    context.contextNode.kind === 257 /* TypeAliasDeclaration */) {
                    // Can’t remove semicolon after `foo`; it would parse as a method declaration:
                    //
                    // interface I {
                    //   foo;
                    //   (): void
                    // }
                    return !ts.isPropertySignature(context.currentTokenParent)
                        || !!context.currentTokenParent.type
                        || nextTokenKind !== 20 /* OpenParenToken */;
                }
                if (ts.isPropertyDeclaration(context.currentTokenParent)) {
                    return !context.currentTokenParent.initializer;
                }
                return context.currentTokenParent.kind !== 240 /* ForStatement */
                    && context.currentTokenParent.kind !== 234 /* EmptyStatement */
                    && context.currentTokenParent.kind !== 232 /* SemicolonClassElement */
                    && nextTokenKind !== 22 /* OpenBracketToken */
                    && nextTokenKind !== 20 /* OpenParenToken */
                    && nextTokenKind !== 39 /* PlusToken */
                    && nextTokenKind !== 40 /* MinusToken */
                    && nextTokenKind !== 43 /* SlashToken */
                    && nextTokenKind !== 13 /* RegularExpressionLiteral */
                    && nextTokenKind !== 27 /* CommaToken */
                    && nextTokenKind !== 221 /* TemplateExpression */
                    && nextTokenKind !== 15 /* TemplateHead */
                    && nextTokenKind !== 14 /* NoSubstitutionTemplateLiteral */
                    && nextTokenKind !== 24 /* DotToken */;
            }
            function isSemicolonInsertionContext(context) {
                return ts.positionIsASICandidate(context.currentTokenSpan.end, context.currentTokenParent, context.sourceFile);
            }
        })(formatting = ts.formatting || (ts.formatting = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var formatting;
        (function (formatting) {
            function getFormatContext(options, host) {
                return { options: options, getRules: getRulesMap(), host: host };
            }
            formatting.getFormatContext = getFormatContext;
            var rulesMapCache;
            function getRulesMap() {
                if (rulesMapCache === undefined) {
                    rulesMapCache = createRulesMap(formatting.getAllRules());
                }
                return rulesMapCache;
            }
            /**
             * For a given rule action, gets a mask of other rule actions that
             * cannot be applied at the same position.
             */
            function getRuleActionExclusion(ruleAction) {
                var mask = 0;
                if (ruleAction & 1 /* StopProcessingSpaceActions */) {
                    mask |= 28 /* ModifySpaceAction */;
                }
                if (ruleAction & 2 /* StopProcessingTokenActions */) {
                    mask |= 96 /* ModifyTokenAction */;
                }
                if (ruleAction & 28 /* ModifySpaceAction */) {
                    mask |= 28 /* ModifySpaceAction */;
                }
                if (ruleAction & 96 /* ModifyTokenAction */) {
                    mask |= 96 /* ModifyTokenAction */;
                }
                return mask;
            }
            function createRulesMap(rules) {
                var map = buildMap(rules);
                return function (context) {
                    var bucket = map[getRuleBucketIndex(context.currentTokenSpan.kind, context.nextTokenSpan.kind)];
                    if (bucket) {
                        var rules_1 = [];
                        var ruleActionMask = 0;
                        for (var _i = 0, bucket_1 = bucket; _i < bucket_1.length; _i++) {
                            var rule = bucket_1[_i];
                            var acceptRuleActions = ~getRuleActionExclusion(ruleActionMask);
                            if (rule.action & acceptRuleActions && ts.every(rule.context, function (c) { return c(context); })) {
                                rules_1.push(rule);
                                ruleActionMask |= rule.action;
                            }
                        }
                        if (rules_1.length) {
                            return rules_1;
                        }
                    }
                };
            }
            function buildMap(rules) {
                // Map from bucket index to array of rules
                var map = new Array(mapRowLength * mapRowLength);
                // This array is used only during construction of the rulesbucket in the map
                var rulesBucketConstructionStateList = new Array(map.length);
                for (var _i = 0, rules_2 = rules; _i < rules_2.length; _i++) {
                    var rule = rules_2[_i];
                    var specificRule = rule.leftTokenRange.isSpecific && rule.rightTokenRange.isSpecific;
                    for (var _a = 0, _b = rule.leftTokenRange.tokens; _a < _b.length; _a++) {
                        var left = _b[_a];
                        for (var _c = 0, _d = rule.rightTokenRange.tokens; _c < _d.length; _c++) {
                            var right = _d[_c];
                            var index = getRuleBucketIndex(left, right);
                            var rulesBucket = map[index];
                            if (rulesBucket === undefined) {
                                rulesBucket = map[index] = [];
                            }
                            addRule(rulesBucket, rule.rule, specificRule, rulesBucketConstructionStateList, index);
                        }
                    }
                }
                return map;
            }
            function getRuleBucketIndex(row, column) {
                ts.Debug.assert(row <= 158 /* LastKeyword */ && column <= 158 /* LastKeyword */, "Must compute formatting context from tokens");
                return (row * mapRowLength) + column;
            }
            var maskBitSize = 5;
            var mask = 31; // MaskBitSize bits
            var mapRowLength = 158 /* LastToken */ + 1;
            var RulesPosition;
            (function (RulesPosition) {
                RulesPosition[RulesPosition["StopRulesSpecific"] = 0] = "StopRulesSpecific";
                RulesPosition[RulesPosition["StopRulesAny"] = maskBitSize * 1] = "StopRulesAny";
                RulesPosition[RulesPosition["ContextRulesSpecific"] = maskBitSize * 2] = "ContextRulesSpecific";
                RulesPosition[RulesPosition["ContextRulesAny"] = maskBitSize * 3] = "ContextRulesAny";
                RulesPosition[RulesPosition["NoContextRulesSpecific"] = maskBitSize * 4] = "NoContextRulesSpecific";
                RulesPosition[RulesPosition["NoContextRulesAny"] = maskBitSize * 5] = "NoContextRulesAny";
            })(RulesPosition || (RulesPosition = {}));
            // The Rules list contains all the inserted rules into a rulebucket in the following order:
            //    1- Ignore rules with specific token combination
            //    2- Ignore rules with any token combination
            //    3- Context rules with specific token combination
            //    4- Context rules with any token combination
            //    5- Non-context rules with specific token combination
            //    6- Non-context rules with any token combination
            //
            // The member rulesInsertionIndexBitmap is used to describe the number of rules
            // in each sub-bucket (above) hence can be used to know the index of where to insert
            // the next rule. It's a bitmap which contains 6 different sections each is given 5 bits.
            //
            // Example:
            // In order to insert a rule to the end of sub-bucket (3), we get the index by adding
            // the values in the bitmap segments 3rd, 2nd, and 1st.
            function addRule(rules, rule, specificTokens, constructionState, rulesBucketIndex) {
                var position = rule.action & 3 /* StopAction */ ?
                    specificTokens ? RulesPosition.StopRulesSpecific : RulesPosition.StopRulesAny :
                    rule.context !== formatting.anyContext ?
                        specificTokens ? RulesPosition.ContextRulesSpecific : RulesPosition.ContextRulesAny :
                        specificTokens ? RulesPosition.NoContextRulesSpecific : RulesPosition.NoContextRulesAny;
                var state = constructionState[rulesBucketIndex] || 0;
                rules.splice(getInsertionIndex(state, position), 0, rule);
                constructionState[rulesBucketIndex] = increaseInsertionIndex(state, position);
            }
            function getInsertionIndex(indexBitmap, maskPosition) {
                var index = 0;
                for (var pos = 0; pos <= maskPosition; pos += maskBitSize) {
                    index += indexBitmap & mask;
                    indexBitmap >>= maskBitSize;
                }
                return index;
            }
            function increaseInsertionIndex(indexBitmap, maskPosition) {
                var value = ((indexBitmap >> maskPosition) & mask) + 1;
                ts.Debug.assert((value & mask) === value, "Adding more rules into the sub-bucket than allowed. Maximum allowed is 32 rules.");
                return (indexBitmap & ~(mask << maskPosition)) | (value << maskPosition);
            }
        })(formatting = ts.formatting || (ts.formatting = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var formatting;
        (function (formatting) {
            function createTextRangeWithKind(pos, end, kind) {
                var textRangeWithKind = { pos: pos, end: end, kind: kind };
                if (ts.Debug.isDebugging) {
                    Object.defineProperty(textRangeWithKind, "__debugKind", {
                        get: function () { return ts.Debug.formatSyntaxKind(kind); },
                    });
                }
                return textRangeWithKind;
            }
            formatting.createTextRangeWithKind = createTextRangeWithKind;
            var Constants;
            (function (Constants) {
                Constants[Constants["Unknown"] = -1] = "Unknown";
            })(Constants || (Constants = {}));
            function formatOnEnter(position, sourceFile, formatContext) {
                var line = sourceFile.getLineAndCharacterOfPosition(position).line;
                if (line === 0) {
                    return [];
                }
                // After the enter key, the cursor is now at a new line. The new line may or may not contain non-whitespace characters.
                // If the new line has only whitespaces, we won't want to format this line, because that would remove the indentation as
                // trailing whitespaces. So the end of the formatting span should be the later one between:
                //  1. the end of the previous line
                //  2. the last non-whitespace character in the current line
                var endOfFormatSpan = ts.getEndLinePosition(line, sourceFile);
                while (ts.isWhiteSpaceSingleLine(sourceFile.text.charCodeAt(endOfFormatSpan))) {
                    endOfFormatSpan--;
                }
                // if the character at the end of the span is a line break, we shouldn't include it, because it indicates we don't want to
                // touch the current line at all. Also, on some OSes the line break consists of two characters (\r\n), we should test if the
                // previous character before the end of format span is line break character as well.
                if (ts.isLineBreak(sourceFile.text.charCodeAt(endOfFormatSpan))) {
                    endOfFormatSpan--;
                }
                var span = {
                    // get start position for the previous line
                    pos: ts.getStartPositionOfLine(line - 1, sourceFile),
                    // end value is exclusive so add 1 to the result
                    end: endOfFormatSpan + 1
                };
                return formatSpan(span, sourceFile, formatContext, 2 /* FormatOnEnter */);
            }
            formatting.formatOnEnter = formatOnEnter;
            function formatOnSemicolon(position, sourceFile, formatContext) {
                var semicolon = findImmediatelyPrecedingTokenOfKind(position, 26 /* SemicolonToken */, sourceFile);
                return formatNodeLines(findOutermostNodeWithinListLevel(semicolon), sourceFile, formatContext, 3 /* FormatOnSemicolon */);
            }
            formatting.formatOnSemicolon = formatOnSemicolon;
            function formatOnOpeningCurly(position, sourceFile, formatContext) {
                var openingCurly = findImmediatelyPrecedingTokenOfKind(position, 18 /* OpenBraceToken */, sourceFile);
                if (!openingCurly) {
                    return [];
                }
                var curlyBraceRange = openingCurly.parent;
                var outermostNode = findOutermostNodeWithinListLevel(curlyBraceRange);
                /**
                 * We limit the span to end at the opening curly to handle the case where
                 * the brace matched to that just typed will be incorrect after further edits.
                 * For example, we could type the opening curly for the following method
                 * body without brace-matching activated:
                 * ```
                 * class C {
                 *     foo()
                 * }
                 * ```
                 * and we wouldn't want to move the closing brace.
                 */
                var textRange = {
                    pos: ts.getLineStartPositionForPosition(outermostNode.getStart(sourceFile), sourceFile),
                    end: position
                };
                return formatSpan(textRange, sourceFile, formatContext, 4 /* FormatOnOpeningCurlyBrace */);
            }
            formatting.formatOnOpeningCurly = formatOnOpeningCurly;
            function formatOnClosingCurly(position, sourceFile, formatContext) {
                var precedingToken = findImmediatelyPrecedingTokenOfKind(position, 19 /* CloseBraceToken */, sourceFile);
                return formatNodeLines(findOutermostNodeWithinListLevel(precedingToken), sourceFile, formatContext, 5 /* FormatOnClosingCurlyBrace */);
            }
            formatting.formatOnClosingCurly = formatOnClosingCurly;
            function formatDocument(sourceFile, formatContext) {
                var span = {
                    pos: 0,
                    end: sourceFile.text.length
                };
                return formatSpan(span, sourceFile, formatContext, 0 /* FormatDocument */);
            }
            formatting.formatDocument = formatDocument;
            function formatSelection(start, end, sourceFile, formatContext) {
                // format from the beginning of the line
                var span = {
                    pos: ts.getLineStartPositionForPosition(start, sourceFile),
                    end: end,
                };
                return formatSpan(span, sourceFile, formatContext, 1 /* FormatSelection */);
            }
            formatting.formatSelection = formatSelection;
            /**
             * Validating `expectedTokenKind` ensures the token was typed in the context we expect (eg: not a comment).
             * @param expectedTokenKind The kind of the last token constituting the desired parent node.
             */
            function findImmediatelyPrecedingTokenOfKind(end, expectedTokenKind, sourceFile) {
                var precedingToken = ts.findPrecedingToken(end, sourceFile);
                return precedingToken && precedingToken.kind === expectedTokenKind && end === precedingToken.getEnd() ?
                    precedingToken :
                    undefined;
            }
            /**
             * Finds the highest node enclosing `node` at the same list level as `node`
             * and whose end does not exceed `node.end`.
             *
             * Consider typing the following
             * ```
             * let x = 1;
             * while (true) {
             * }
             * ```
             * Upon typing the closing curly, we want to format the entire `while`-statement, but not the preceding
             * variable declaration.
             */
            function findOutermostNodeWithinListLevel(node) {
                var current = node;
                while (current &&
                    current.parent &&
                    current.parent.end === node.end &&
                    !isListElement(current.parent, current)) {
                    current = current.parent;
                }
                return current;
            }
            // Returns true if node is a element in some list in parent
            // i.e. parent is class declaration with the list of members and node is one of members.
            function isListElement(parent, node) {
                switch (parent.kind) {
                    case 255 /* ClassDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                        return ts.rangeContainsRange(parent.members, node);
                    case 259 /* ModuleDeclaration */:
                        var body = parent.body;
                        return !!body && body.kind === 260 /* ModuleBlock */ && ts.rangeContainsRange(body.statements, node);
                    case 300 /* SourceFile */:
                    case 233 /* Block */:
                    case 260 /* ModuleBlock */:
                        return ts.rangeContainsRange(parent.statements, node);
                    case 290 /* CatchClause */:
                        return ts.rangeContainsRange(parent.block.statements, node);
                }
                return false;
            }
            /** find node that fully contains given text range */
            function findEnclosingNode(range, sourceFile) {
                return find(sourceFile);
                function find(n) {
                    var candidate = ts.forEachChild(n, function (c) { return ts.startEndContainsRange(c.getStart(sourceFile), c.end, range) && c; });
                    if (candidate) {
                        var result = find(candidate);
                        if (result) {
                            return result;
                        }
                    }
                    return n;
                }
            }
            /** formatting is not applied to ranges that contain parse errors.
             * This function will return a predicate that for a given text range will tell
             * if there are any parse errors that overlap with the range.
             */
            function prepareRangeContainsErrorFunction(errors, originalRange) {
                if (!errors.length) {
                    return rangeHasNoErrors;
                }
                // pick only errors that fall in range
                var sorted = errors
                    .filter(function (d) { return ts.rangeOverlapsWithStartEnd(originalRange, d.start, d.start + d.length); }) // TODO: GH#18217
                    .sort(function (e1, e2) { return e1.start - e2.start; });
                if (!sorted.length) {
                    return rangeHasNoErrors;
                }
                var index = 0;
                return function (r) {
                    // in current implementation sequence of arguments [r1, r2...] is monotonically increasing.
                    // 'index' tracks the index of the most recent error that was checked.
                    while (true) {
                        if (index >= sorted.length) {
                            // all errors in the range were already checked -> no error in specified range
                            return false;
                        }
                        var error = sorted[index];
                        if (r.end <= error.start) {
                            // specified range ends before the error referred by 'index' - no error in range
                            return false;
                        }
                        if (ts.startEndOverlapsWithStartEnd(r.pos, r.end, error.start, error.start + error.length)) {
                            // specified range overlaps with error range
                            return true;
                        }
                        index++;
                    }
                };
                function rangeHasNoErrors() {
                    return false;
                }
            }
            /**
             * Start of the original range might fall inside the comment - scanner will not yield appropriate results
             * This function will look for token that is located before the start of target range
             * and return its end as start position for the scanner.
             */
            function getScanStartPosition(enclosingNode, originalRange, sourceFile) {
                var start = enclosingNode.getStart(sourceFile);
                if (start === originalRange.pos && enclosingNode.end === originalRange.end) {
                    return start;
                }
                var precedingToken = ts.findPrecedingToken(originalRange.pos, sourceFile);
                if (!precedingToken) {
                    // no preceding token found - start from the beginning of enclosing node
                    return enclosingNode.pos;
                }
                // preceding token ends after the start of original range (i.e when originalRange.pos falls in the middle of literal)
                // start from the beginning of enclosingNode to handle the entire 'originalRange'
                if (precedingToken.end >= originalRange.pos) {
                    return enclosingNode.pos;
                }
                return precedingToken.end;
            }
            /*
             * For cases like
             * if (a ||
             *     b ||$
             *     c) {...}
             * If we hit Enter at $ we want line '    b ||' to be indented.
             * Formatting will be applied to the last two lines.
             * Node that fully encloses these lines is binary expression 'a ||...'.
             * Initial indentation for this node will be 0.
             * Binary expressions don't introduce new indentation scopes, however it is possible
             * that some parent node on the same line does - like if statement in this case.
             * Note that we are considering parents only from the same line with initial node -
             * if parent is on the different line - its delta was already contributed
             * to the initial indentation.
             */
            function getOwnOrInheritedDelta(n, options, sourceFile) {
                var previousLine = -1 /* Unknown */;
                var child;
                while (n) {
                    var line = sourceFile.getLineAndCharacterOfPosition(n.getStart(sourceFile)).line;
                    if (previousLine !== -1 /* Unknown */ && line !== previousLine) {
                        break;
                    }
                    if (formatting.SmartIndenter.shouldIndentChildNode(options, n, child, sourceFile)) {
                        return options.indentSize;
                    }
                    previousLine = line;
                    child = n;
                    n = n.parent;
                }
                return 0;
            }
            function formatNodeGivenIndentation(node, sourceFileLike, languageVariant, initialIndentation, delta, formatContext) {
                var range = { pos: 0, end: sourceFileLike.text.length };
                return formatting.getFormattingScanner(sourceFileLike.text, languageVariant, range.pos, range.end, function (scanner) {
                    return formatSpanWorker(range, node, initialIndentation, delta, scanner, formatContext, 1 /* FormatSelection */, function (_) { return false; }, // assume that node does not have any errors
                    sourceFileLike);
                });
            }
            formatting.formatNodeGivenIndentation = formatNodeGivenIndentation;
            function formatNodeLines(node, sourceFile, formatContext, requestKind) {
                if (!node) {
                    return [];
                }
                var span = {
                    pos: ts.getLineStartPositionForPosition(node.getStart(sourceFile), sourceFile),
                    end: node.end
                };
                return formatSpan(span, sourceFile, formatContext, requestKind);
            }
            function formatSpan(originalRange, sourceFile, formatContext, requestKind) {
                // find the smallest node that fully wraps the range and compute the initial indentation for the node
                var enclosingNode = findEnclosingNode(originalRange, sourceFile);
                return formatting.getFormattingScanner(sourceFile.text, sourceFile.languageVariant, getScanStartPosition(enclosingNode, originalRange, sourceFile), originalRange.end, function (scanner) { return formatSpanWorker(originalRange, enclosingNode, formatting.SmartIndenter.getIndentationForNode(enclosingNode, originalRange, sourceFile, formatContext.options), getOwnOrInheritedDelta(enclosingNode, formatContext.options, sourceFile), scanner, formatContext, requestKind, prepareRangeContainsErrorFunction(sourceFile.parseDiagnostics, originalRange), sourceFile); });
            }
            function formatSpanWorker(originalRange, enclosingNode, initialIndentation, delta, formattingScanner, _a, requestKind, rangeContainsError, sourceFile) {
                var options = _a.options, getRules = _a.getRules, host = _a.host;
                // formatting context is used by rules provider
                var formattingContext = new formatting.FormattingContext(sourceFile, requestKind, options);
                var previousRange;
                var previousParent;
                var previousRangeStartLine;
                var lastIndentedLine;
                var indentationOnLastIndentedLine = -1 /* Unknown */;
                var edits = [];
                formattingScanner.advance();
                if (formattingScanner.isOnToken()) {
                    var startLine = sourceFile.getLineAndCharacterOfPosition(enclosingNode.getStart(sourceFile)).line;
                    var undecoratedStartLine = startLine;
                    if (enclosingNode.decorators) {
                        undecoratedStartLine = sourceFile.getLineAndCharacterOfPosition(ts.getNonDecoratorTokenPosOfNode(enclosingNode, sourceFile)).line;
                    }
                    processNode(enclosingNode, enclosingNode, startLine, undecoratedStartLine, initialIndentation, delta);
                }
                if (!formattingScanner.isOnToken()) {
                    var indentation = formatting.SmartIndenter.nodeWillIndentChild(options, enclosingNode, /*child*/ undefined, sourceFile, /*indentByDefault*/ false)
                        ? initialIndentation + options.indentSize
                        : initialIndentation;
                    var leadingTrivia = formattingScanner.getCurrentLeadingTrivia();
                    if (leadingTrivia) {
                        indentTriviaItems(leadingTrivia, indentation, /*indentNextTokenOrTrivia*/ false, function (item) { return processRange(item, sourceFile.getLineAndCharacterOfPosition(item.pos), enclosingNode, enclosingNode, /*dynamicIndentation*/ undefined); });
                    }
                }
                if (options.trimTrailingWhitespace !== false) {
                    trimTrailingWhitespacesForRemainingRange();
                }
                return edits;
                // local functions
                /** Tries to compute the indentation for a list element.
                 * If list element is not in range then
                 * function will pick its actual indentation
                 * so it can be pushed downstream as inherited indentation.
                 * If list element is in the range - its indentation will be equal
                 * to inherited indentation from its predecessors.
                 */
                function tryComputeIndentationForListItem(startPos, endPos, parentStartLine, range, inheritedIndentation) {
                    if (ts.rangeOverlapsWithStartEnd(range, startPos, endPos) ||
                        ts.rangeContainsStartEnd(range, startPos, endPos) /* Not to miss zero-range nodes e.g. JsxText */) {
                        if (inheritedIndentation !== -1 /* Unknown */) {
                            return inheritedIndentation;
                        }
                    }
                    else {
                        var startLine = sourceFile.getLineAndCharacterOfPosition(startPos).line;
                        var startLinePosition = ts.getLineStartPositionForPosition(startPos, sourceFile);
                        var column = formatting.SmartIndenter.findFirstNonWhitespaceColumn(startLinePosition, startPos, sourceFile, options);
                        if (startLine !== parentStartLine || startPos === column) {
                            // Use the base indent size if it is greater than
                            // the indentation of the inherited predecessor.
                            var baseIndentSize = formatting.SmartIndenter.getBaseIndentation(options);
                            return baseIndentSize > column ? baseIndentSize : column;
                        }
                    }
                    return -1 /* Unknown */;
                }
                function computeIndentation(node, startLine, inheritedIndentation, parent, parentDynamicIndentation, effectiveParentStartLine) {
                    var delta = formatting.SmartIndenter.shouldIndentChildNode(options, node) ? options.indentSize : 0;
                    if (effectiveParentStartLine === startLine) {
                        // if node is located on the same line with the parent
                        // - inherit indentation from the parent
                        // - push children if either parent of node itself has non-zero delta
                        return {
                            indentation: startLine === lastIndentedLine ? indentationOnLastIndentedLine : parentDynamicIndentation.getIndentation(),
                            delta: Math.min(options.indentSize, parentDynamicIndentation.getDelta(node) + delta)
                        };
                    }
                    else if (inheritedIndentation === -1 /* Unknown */) {
                        if (node.kind === 20 /* OpenParenToken */ && startLine === lastIndentedLine) {
                            // the is used for chaining methods formatting
                            // - we need to get the indentation on last line and the delta of parent
                            return { indentation: indentationOnLastIndentedLine, delta: parentDynamicIndentation.getDelta(node) };
                        }
                        else if (formatting.SmartIndenter.childStartsOnTheSameLineWithElseInIfStatement(parent, node, startLine, sourceFile) ||
                            formatting.SmartIndenter.childIsUnindentedBranchOfConditionalExpression(parent, node, startLine, sourceFile) ||
                            formatting.SmartIndenter.argumentStartsOnSameLineAsPreviousArgument(parent, node, startLine, sourceFile)) {
                            return { indentation: parentDynamicIndentation.getIndentation(), delta: delta };
                        }
                        else {
                            return { indentation: parentDynamicIndentation.getIndentation() + parentDynamicIndentation.getDelta(node), delta: delta };
                        }
                    }
                    else {
                        return { indentation: inheritedIndentation, delta: delta };
                    }
                }
                function getFirstNonDecoratorTokenOfNode(node) {
                    if (node.modifiers && node.modifiers.length) {
                        return node.modifiers[0].kind;
                    }
                    switch (node.kind) {
                        case 255 /* ClassDeclaration */: return 84 /* ClassKeyword */;
                        case 256 /* InterfaceDeclaration */: return 118 /* InterfaceKeyword */;
                        case 254 /* FunctionDeclaration */: return 98 /* FunctionKeyword */;
                        case 258 /* EnumDeclaration */: return 258 /* EnumDeclaration */;
                        case 170 /* GetAccessor */: return 135 /* GetKeyword */;
                        case 171 /* SetAccessor */: return 147 /* SetKeyword */;
                        case 167 /* MethodDeclaration */:
                            if (node.asteriskToken) {
                                return 41 /* AsteriskToken */;
                            }
                        // falls through
                        case 165 /* PropertyDeclaration */:
                        case 162 /* Parameter */:
                            var name = ts.getNameOfDeclaration(node);
                            if (name) {
                                return name.kind;
                            }
                    }
                }
                function getDynamicIndentation(node, nodeStartLine, indentation, delta) {
                    return {
                        getIndentationForComment: function (kind, tokenIndentation, container) {
                            switch (kind) {
                                // preceding comment to the token that closes the indentation scope inherits the indentation from the scope
                                // ..  {
                                //     // comment
                                // }
                                case 19 /* CloseBraceToken */:
                                case 23 /* CloseBracketToken */:
                                case 21 /* CloseParenToken */:
                                    return indentation + getDelta(container);
                            }
                            return tokenIndentation !== -1 /* Unknown */ ? tokenIndentation : indentation;
                        },
                        // if list end token is LessThanToken '>' then its delta should be explicitly suppressed
                        // so that LessThanToken as a binary operator can still be indented.
                        // foo.then
                        //     <
                        //         number,
                        //         string,
                        //     >();
                        // vs
                        // var a = xValue
                        //     > yValue;
                        getIndentationForToken: function (line, kind, container, suppressDelta) {
                            return !suppressDelta && shouldAddDelta(line, kind, container) ? indentation + getDelta(container) : indentation;
                        },
                        getIndentation: function () { return indentation; },
                        getDelta: getDelta,
                        recomputeIndentation: function (lineAdded, parent) {
                            if (formatting.SmartIndenter.shouldIndentChildNode(options, parent, node, sourceFile)) {
                                indentation += lineAdded ? options.indentSize : -options.indentSize;
                                delta = formatting.SmartIndenter.shouldIndentChildNode(options, node) ? options.indentSize : 0;
                            }
                        }
                    };
                    function shouldAddDelta(line, kind, container) {
                        switch (kind) {
                            // open and close brace, 'else' and 'while' (in do statement) tokens has indentation of the parent
                            case 18 /* OpenBraceToken */:
                            case 19 /* CloseBraceToken */:
                            case 21 /* CloseParenToken */:
                            case 91 /* ElseKeyword */:
                            case 115 /* WhileKeyword */:
                            case 59 /* AtToken */:
                                return false;
                            case 43 /* SlashToken */:
                            case 31 /* GreaterThanToken */:
                                switch (container.kind) {
                                    case 278 /* JsxOpeningElement */:
                                    case 279 /* JsxClosingElement */:
                                    case 277 /* JsxSelfClosingElement */:
                                    case 226 /* ExpressionWithTypeArguments */:
                                        return false;
                                }
                                break;
                            case 22 /* OpenBracketToken */:
                            case 23 /* CloseBracketToken */:
                                if (container.kind !== 193 /* MappedType */) {
                                    return false;
                                }
                                break;
                        }
                        // if token line equals to the line of containing node (this is a first token in the node) - use node indentation
                        return nodeStartLine !== line
                            // if this token is the first token following the list of decorators, we do not need to indent
                            && !(node.decorators && kind === getFirstNonDecoratorTokenOfNode(node));
                    }
                    function getDelta(child) {
                        // Delta value should be zero when the node explicitly prevents indentation of the child node
                        return formatting.SmartIndenter.nodeWillIndentChild(options, node, child, sourceFile, /*indentByDefault*/ true) ? delta : 0;
                    }
                }
                function processNode(node, contextNode, nodeStartLine, undecoratedNodeStartLine, indentation, delta) {
                    if (!ts.rangeOverlapsWithStartEnd(originalRange, node.getStart(sourceFile), node.getEnd())) {
                        return;
                    }
                    var nodeDynamicIndentation = getDynamicIndentation(node, nodeStartLine, indentation, delta);
                    // a useful observations when tracking context node
                    //        /
                    //      [a]
                    //   /   |   \
                    //  [b] [c] [d]
                    // node 'a' is a context node for nodes 'b', 'c', 'd'
                    // except for the leftmost leaf token in [b] - in this case context node ('e') is located somewhere above 'a'
                    // this rule can be applied recursively to child nodes of 'a'.
                    //
                    // context node is set to parent node value after processing every child node
                    // context node is set to parent of the token after processing every token
                    var childContextNode = contextNode;
                    // if there are any tokens that logically belong to node and interleave child nodes
                    // such tokens will be consumed in processChildNode for the child that follows them
                    ts.forEachChild(node, function (child) {
                        processChildNode(child, /*inheritedIndentation*/ -1 /* Unknown */, node, nodeDynamicIndentation, nodeStartLine, undecoratedNodeStartLine, /*isListItem*/ false);
                    }, function (nodes) {
                        processChildNodes(nodes, node, nodeStartLine, nodeDynamicIndentation);
                    });
                    // proceed any tokens in the node that are located after child nodes
                    while (formattingScanner.isOnToken()) {
                        var tokenInfo = formattingScanner.readTokenInfo(node);
                        if (tokenInfo.token.end > node.end) {
                            break;
                        }
                        consumeTokenAndAdvanceScanner(tokenInfo, node, nodeDynamicIndentation, node);
                    }
                    if (!node.parent && formattingScanner.isOnEOF()) {
                        var token = formattingScanner.readEOFTokenRange();
                        if (token.end <= node.end && previousRange) {
                            processPair(token, sourceFile.getLineAndCharacterOfPosition(token.pos).line, node, previousRange, previousRangeStartLine, previousParent, contextNode, nodeDynamicIndentation);
                        }
                    }
                    function processChildNode(child, inheritedIndentation, parent, parentDynamicIndentation, parentStartLine, undecoratedParentStartLine, isListItem, isFirstListItem) {
                        var childStartPos = child.getStart(sourceFile);
                        var childStartLine = sourceFile.getLineAndCharacterOfPosition(childStartPos).line;
                        var undecoratedChildStartLine = childStartLine;
                        if (child.decorators) {
                            undecoratedChildStartLine = sourceFile.getLineAndCharacterOfPosition(ts.getNonDecoratorTokenPosOfNode(child, sourceFile)).line;
                        }
                        // if child is a list item - try to get its indentation, only if parent is within the original range.
                        var childIndentationAmount = -1 /* Unknown */;
                        if (isListItem && ts.rangeContainsRange(originalRange, parent)) {
                            childIndentationAmount = tryComputeIndentationForListItem(childStartPos, child.end, parentStartLine, originalRange, inheritedIndentation);
                            if (childIndentationAmount !== -1 /* Unknown */) {
                                inheritedIndentation = childIndentationAmount;
                            }
                        }
                        // child node is outside the target range - do not dive inside
                        if (!ts.rangeOverlapsWithStartEnd(originalRange, child.pos, child.end)) {
                            if (child.end < originalRange.pos) {
                                formattingScanner.skipToEndOf(child);
                            }
                            return inheritedIndentation;
                        }
                        if (child.getFullWidth() === 0) {
                            return inheritedIndentation;
                        }
                        while (formattingScanner.isOnToken()) {
                            // proceed any parent tokens that are located prior to child.getStart()
                            var tokenInfo = formattingScanner.readTokenInfo(node);
                            if (tokenInfo.token.end > childStartPos) {
                                if (tokenInfo.token.pos > childStartPos) {
                                    formattingScanner.skipToStartOf(child);
                                }
                                // stop when formatting scanner advances past the beginning of the child
                                break;
                            }
                            consumeTokenAndAdvanceScanner(tokenInfo, node, parentDynamicIndentation, node);
                        }
                        if (!formattingScanner.isOnToken()) {
                            return inheritedIndentation;
                        }
                        if (ts.isToken(child)) {
                            // if child node is a token, it does not impact indentation, proceed it using parent indentation scope rules
                            var tokenInfo = formattingScanner.readTokenInfo(child);
                            // JSX text shouldn't affect indenting
                            if (child.kind !== 11 /* JsxText */) {
                                ts.Debug.assert(tokenInfo.token.end === child.end, "Token end is child end");
                                consumeTokenAndAdvanceScanner(tokenInfo, node, parentDynamicIndentation, child);
                                return inheritedIndentation;
                            }
                        }
                        var effectiveParentStartLine = child.kind === 163 /* Decorator */ ? childStartLine : undecoratedParentStartLine;
                        var childIndentation = computeIndentation(child, childStartLine, childIndentationAmount, node, parentDynamicIndentation, effectiveParentStartLine);
                        processNode(child, childContextNode, childStartLine, undecoratedChildStartLine, childIndentation.indentation, childIndentation.delta);
                        childContextNode = node;
                        if (isFirstListItem && parent.kind === 202 /* ArrayLiteralExpression */ && inheritedIndentation === -1 /* Unknown */) {
                            inheritedIndentation = childIndentation.indentation;
                        }
                        return inheritedIndentation;
                    }
                    function processChildNodes(nodes, parent, parentStartLine, parentDynamicIndentation) {
                        ts.Debug.assert(ts.isNodeArray(nodes));
                        var listStartToken = getOpenTokenForList(parent, nodes);
                        var listDynamicIndentation = parentDynamicIndentation;
                        var startLine = parentStartLine;
                        if (listStartToken !== 0 /* Unknown */) {
                            // introduce a new indentation scope for lists (including list start and end tokens)
                            while (formattingScanner.isOnToken()) {
                                var tokenInfo = formattingScanner.readTokenInfo(parent);
                                if (tokenInfo.token.end > nodes.pos) {
                                    // stop when formatting scanner moves past the beginning of node list
                                    break;
                                }
                                else if (tokenInfo.token.kind === listStartToken) {
                                    // consume list start token
                                    startLine = sourceFile.getLineAndCharacterOfPosition(tokenInfo.token.pos).line;
                                    consumeTokenAndAdvanceScanner(tokenInfo, parent, parentDynamicIndentation, parent);
                                    var indentationOnListStartToken = void 0;
                                    if (indentationOnLastIndentedLine !== -1 /* Unknown */) {
                                        // scanner just processed list start token so consider last indentation as list indentation
                                        // function foo(): { // last indentation was 0, list item will be indented based on this value
                                        //   foo: number;
                                        // }: {};
                                        indentationOnListStartToken = indentationOnLastIndentedLine;
                                    }
                                    else {
                                        var startLinePosition = ts.getLineStartPositionForPosition(tokenInfo.token.pos, sourceFile);
                                        indentationOnListStartToken = formatting.SmartIndenter.findFirstNonWhitespaceColumn(startLinePosition, tokenInfo.token.pos, sourceFile, options);
                                    }
                                    listDynamicIndentation = getDynamicIndentation(parent, parentStartLine, indentationOnListStartToken, options.indentSize); // TODO: GH#18217
                                }
                                else {
                                    // consume any tokens that precede the list as child elements of 'node' using its indentation scope
                                    consumeTokenAndAdvanceScanner(tokenInfo, parent, parentDynamicIndentation, parent);
                                }
                            }
                        }
                        var inheritedIndentation = -1 /* Unknown */;
                        for (var i = 0; i < nodes.length; i++) {
                            var child = nodes[i];
                            inheritedIndentation = processChildNode(child, inheritedIndentation, node, listDynamicIndentation, startLine, startLine, /*isListItem*/ true, /*isFirstListItem*/ i === 0);
                        }
                        var listEndToken = getCloseTokenForOpenToken(listStartToken);
                        if (listEndToken !== 0 /* Unknown */ && formattingScanner.isOnToken()) {
                            var tokenInfo = formattingScanner.readTokenInfo(parent);
                            if (tokenInfo.token.kind === 27 /* CommaToken */ && ts.isCallLikeExpression(parent)) {
                                var commaTokenLine = sourceFile.getLineAndCharacterOfPosition(tokenInfo.token.pos).line;
                                if (startLine !== commaTokenLine) {
                                    formattingScanner.advance();
                                    tokenInfo = formattingScanner.isOnToken() ? formattingScanner.readTokenInfo(parent) : undefined;
                                }
                            }
                            // consume the list end token only if it is still belong to the parent
                            // there might be the case when current token matches end token but does not considered as one
                            // function (x: function) <--
                            // without this check close paren will be interpreted as list end token for function expression which is wrong
                            if (tokenInfo && tokenInfo.token.kind === listEndToken && ts.rangeContainsRange(parent, tokenInfo.token)) {
                                // consume list end token
                                consumeTokenAndAdvanceScanner(tokenInfo, parent, listDynamicIndentation, parent, /*isListEndToken*/ true);
                            }
                        }
                    }
                    function consumeTokenAndAdvanceScanner(currentTokenInfo, parent, dynamicIndentation, container, isListEndToken) {
                        ts.Debug.assert(ts.rangeContainsRange(parent, currentTokenInfo.token));
                        var lastTriviaWasNewLine = formattingScanner.lastTrailingTriviaWasNewLine();
                        var indentToken = false;
                        if (currentTokenInfo.leadingTrivia) {
                            processTrivia(currentTokenInfo.leadingTrivia, parent, childContextNode, dynamicIndentation);
                        }
                        var lineAction = 0 /* None */;
                        var isTokenInRange = ts.rangeContainsRange(originalRange, currentTokenInfo.token);
                        var tokenStart = sourceFile.getLineAndCharacterOfPosition(currentTokenInfo.token.pos);
                        if (isTokenInRange) {
                            var rangeHasError = rangeContainsError(currentTokenInfo.token);
                            // save previousRange since processRange will overwrite this value with current one
                            var savePreviousRange = previousRange;
                            lineAction = processRange(currentTokenInfo.token, tokenStart, parent, childContextNode, dynamicIndentation);
                            // do not indent comments\token if token range overlaps with some error
                            if (!rangeHasError) {
                                if (lineAction === 0 /* None */) {
                                    // indent token only if end line of previous range does not match start line of the token
                                    var prevEndLine = savePreviousRange && sourceFile.getLineAndCharacterOfPosition(savePreviousRange.end).line;
                                    indentToken = lastTriviaWasNewLine && tokenStart.line !== prevEndLine;
                                }
                                else {
                                    indentToken = lineAction === 1 /* LineAdded */;
                                }
                            }
                        }
                        if (currentTokenInfo.trailingTrivia) {
                            processTrivia(currentTokenInfo.trailingTrivia, parent, childContextNode, dynamicIndentation);
                        }
                        if (indentToken) {
                            var tokenIndentation = (isTokenInRange && !rangeContainsError(currentTokenInfo.token)) ?
                                dynamicIndentation.getIndentationForToken(tokenStart.line, currentTokenInfo.token.kind, container, !!isListEndToken) :
                                -1 /* Unknown */;
                            var indentNextTokenOrTrivia = true;
                            if (currentTokenInfo.leadingTrivia) {
                                var commentIndentation_1 = dynamicIndentation.getIndentationForComment(currentTokenInfo.token.kind, tokenIndentation, container);
                                indentNextTokenOrTrivia = indentTriviaItems(currentTokenInfo.leadingTrivia, commentIndentation_1, indentNextTokenOrTrivia, function (item) { return insertIndentation(item.pos, commentIndentation_1, /*lineAdded*/ false); });
                            }
                            // indent token only if is it is in target range and does not overlap with any error ranges
                            if (tokenIndentation !== -1 /* Unknown */ && indentNextTokenOrTrivia) {
                                insertIndentation(currentTokenInfo.token.pos, tokenIndentation, lineAction === 1 /* LineAdded */);
                                lastIndentedLine = tokenStart.line;
                                indentationOnLastIndentedLine = tokenIndentation;
                            }
                        }
                        formattingScanner.advance();
                        childContextNode = parent;
                    }
                }
                function indentTriviaItems(trivia, commentIndentation, indentNextTokenOrTrivia, indentSingleLine) {
                    for (var _i = 0, trivia_1 = trivia; _i < trivia_1.length; _i++) {
                        var triviaItem = trivia_1[_i];
                        var triviaInRange = ts.rangeContainsRange(originalRange, triviaItem);
                        switch (triviaItem.kind) {
                            case 3 /* MultiLineCommentTrivia */:
                                if (triviaInRange) {
                                    indentMultilineComment(triviaItem, commentIndentation, /*firstLineIsIndented*/ !indentNextTokenOrTrivia);
                                }
                                indentNextTokenOrTrivia = false;
                                break;
                            case 2 /* SingleLineCommentTrivia */:
                                if (indentNextTokenOrTrivia && triviaInRange) {
                                    indentSingleLine(triviaItem);
                                }
                                indentNextTokenOrTrivia = false;
                                break;
                            case 4 /* NewLineTrivia */:
                                indentNextTokenOrTrivia = true;
                                break;
                        }
                    }
                    return indentNextTokenOrTrivia;
                }
                function processTrivia(trivia, parent, contextNode, dynamicIndentation) {
                    for (var _i = 0, trivia_2 = trivia; _i < trivia_2.length; _i++) {
                        var triviaItem = trivia_2[_i];
                        if (ts.isComment(triviaItem.kind) && ts.rangeContainsRange(originalRange, triviaItem)) {
                            var triviaItemStart = sourceFile.getLineAndCharacterOfPosition(triviaItem.pos);
                            processRange(triviaItem, triviaItemStart, parent, contextNode, dynamicIndentation);
                        }
                    }
                }
                function processRange(range, rangeStart, parent, contextNode, dynamicIndentation) {
                    var rangeHasError = rangeContainsError(range);
                    var lineAction = 0 /* None */;
                    if (!rangeHasError) {
                        if (!previousRange) {
                            // trim whitespaces starting from the beginning of the span up to the current line
                            var originalStart = sourceFile.getLineAndCharacterOfPosition(originalRange.pos);
                            trimTrailingWhitespacesForLines(originalStart.line, rangeStart.line);
                        }
                        else {
                            lineAction =
                                processPair(range, rangeStart.line, parent, previousRange, previousRangeStartLine, previousParent, contextNode, dynamicIndentation);
                        }
                    }
                    previousRange = range;
                    previousParent = parent;
                    previousRangeStartLine = rangeStart.line;
                    return lineAction;
                }
                function processPair(currentItem, currentStartLine, currentParent, previousItem, previousStartLine, previousParent, contextNode, dynamicIndentation) {
                    formattingContext.updateContext(previousItem, previousParent, currentItem, currentParent, contextNode);
                    var rules = getRules(formattingContext);
                    var trimTrailingWhitespaces = formattingContext.options.trimTrailingWhitespace !== false;
                    var lineAction = 0 /* None */;
                    if (rules) {
                        // Apply rules in reverse order so that higher priority rules (which are first in the array)
                        // win in a conflict with lower priority rules.
                        ts.forEachRight(rules, function (rule) {
                            lineAction = applyRuleEdits(rule, previousItem, previousStartLine, currentItem, currentStartLine);
                            switch (lineAction) {
                                case 2 /* LineRemoved */:
                                    // Handle the case where the next line is moved to be the end of this line.
                                    // In this case we don't indent the next line in the next pass.
                                    if (currentParent.getStart(sourceFile) === currentItem.pos) {
                                        dynamicIndentation.recomputeIndentation(/*lineAddedByFormatting*/ false, contextNode);
                                    }
                                    break;
                                case 1 /* LineAdded */:
                                    // Handle the case where token2 is moved to the new line.
                                    // In this case we indent token2 in the next pass but we set
                                    // sameLineIndent flag to notify the indenter that the indentation is within the line.
                                    if (currentParent.getStart(sourceFile) === currentItem.pos) {
                                        dynamicIndentation.recomputeIndentation(/*lineAddedByFormatting*/ true, contextNode);
                                    }
                                    break;
                                default:
                                    ts.Debug.assert(lineAction === 0 /* None */);
                            }
                            // We need to trim trailing whitespace between the tokens if they were on different lines, and no rule was applied to put them on the same line
                            trimTrailingWhitespaces = trimTrailingWhitespaces && !(rule.action & 16 /* DeleteSpace */) && rule.flags !== 1 /* CanDeleteNewLines */;
                        });
                    }
                    else {
                        trimTrailingWhitespaces = trimTrailingWhitespaces && currentItem.kind !== 1 /* EndOfFileToken */;
                    }
                    if (currentStartLine !== previousStartLine && trimTrailingWhitespaces) {
                        // We need to trim trailing whitespace between the tokens if they were on different lines, and no rule was applied to put them on the same line
                        trimTrailingWhitespacesForLines(previousStartLine, currentStartLine, previousItem);
                    }
                    return lineAction;
                }
                function insertIndentation(pos, indentation, lineAdded) {
                    var indentationString = getIndentationString(indentation, options);
                    if (lineAdded) {
                        // new line is added before the token by the formatting rules
                        // insert indentation string at the very beginning of the token
                        recordReplace(pos, 0, indentationString);
                    }
                    else {
                        var tokenStart = sourceFile.getLineAndCharacterOfPosition(pos);
                        var startLinePosition = ts.getStartPositionOfLine(tokenStart.line, sourceFile);
                        if (indentation !== characterToColumn(startLinePosition, tokenStart.character) || indentationIsDifferent(indentationString, startLinePosition)) {
                            recordReplace(startLinePosition, tokenStart.character, indentationString);
                        }
                    }
                }
                function characterToColumn(startLinePosition, characterInLine) {
                    var column = 0;
                    for (var i = 0; i < characterInLine; i++) {
                        if (sourceFile.text.charCodeAt(startLinePosition + i) === 9 /* tab */) {
                            column += options.tabSize - column % options.tabSize;
                        }
                        else {
                            column++;
                        }
                    }
                    return column;
                }
                function indentationIsDifferent(indentationString, startLinePosition) {
                    return indentationString !== sourceFile.text.substr(startLinePosition, indentationString.length);
                }
                function indentMultilineComment(commentRange, indentation, firstLineIsIndented, indentFinalLine) {
                    if (indentFinalLine === void 0) {
                        indentFinalLine = true;
                    }
                    // split comment in lines
                    var startLine = sourceFile.getLineAndCharacterOfPosition(commentRange.pos).line;
                    var endLine = sourceFile.getLineAndCharacterOfPosition(commentRange.end).line;
                    if (startLine === endLine) {
                        if (!firstLineIsIndented) {
                            // treat as single line comment
                            insertIndentation(commentRange.pos, indentation, /*lineAdded*/ false);
                        }
                        return;
                    }
                    var parts = [];
                    var startPos = commentRange.pos;
                    for (var line = startLine; line < endLine; line++) {
                        var endOfLine = ts.getEndLinePosition(line, sourceFile);
                        parts.push({ pos: startPos, end: endOfLine });
                        startPos = ts.getStartPositionOfLine(line + 1, sourceFile);
                    }
                    if (indentFinalLine) {
                        parts.push({ pos: startPos, end: commentRange.end });
                    }
                    if (parts.length === 0)
                        return;
                    var startLinePos = ts.getStartPositionOfLine(startLine, sourceFile);
                    var nonWhitespaceColumnInFirstPart = formatting.SmartIndenter.findFirstNonWhitespaceCharacterAndColumn(startLinePos, parts[0].pos, sourceFile, options);
                    var startIndex = 0;
                    if (firstLineIsIndented) {
                        startIndex = 1;
                        startLine++;
                    }
                    // shift all parts on the delta size
                    var delta = indentation - nonWhitespaceColumnInFirstPart.column;
                    for (var i = startIndex; i < parts.length; i++, startLine++) {
                        var startLinePos_1 = ts.getStartPositionOfLine(startLine, sourceFile);
                        var nonWhitespaceCharacterAndColumn = i === 0
                            ? nonWhitespaceColumnInFirstPart
                            : formatting.SmartIndenter.findFirstNonWhitespaceCharacterAndColumn(parts[i].pos, parts[i].end, sourceFile, options);
                        var newIndentation = nonWhitespaceCharacterAndColumn.column + delta;
                        if (newIndentation > 0) {
                            var indentationString = getIndentationString(newIndentation, options);
                            recordReplace(startLinePos_1, nonWhitespaceCharacterAndColumn.character, indentationString);
                        }
                        else {
                            recordDelete(startLinePos_1, nonWhitespaceCharacterAndColumn.character);
                        }
                    }
                }
                function trimTrailingWhitespacesForLines(line1, line2, range) {
                    for (var line = line1; line < line2; line++) {
                        var lineStartPosition = ts.getStartPositionOfLine(line, sourceFile);
                        var lineEndPosition = ts.getEndLinePosition(line, sourceFile);
                        // do not trim whitespaces in comments or template expression
                        if (range && (ts.isComment(range.kind) || ts.isStringOrRegularExpressionOrTemplateLiteral(range.kind)) && range.pos <= lineEndPosition && range.end > lineEndPosition) {
                            continue;
                        }
                        var whitespaceStart = getTrailingWhitespaceStartPosition(lineStartPosition, lineEndPosition);
                        if (whitespaceStart !== -1) {
                            ts.Debug.assert(whitespaceStart === lineStartPosition || !ts.isWhiteSpaceSingleLine(sourceFile.text.charCodeAt(whitespaceStart - 1)));
                            recordDelete(whitespaceStart, lineEndPosition + 1 - whitespaceStart);
                        }
                    }
                }
                /**
                 * @param start The position of the first character in range
                 * @param end The position of the last character in range
                 */
                function getTrailingWhitespaceStartPosition(start, end) {
                    var pos = end;
                    while (pos >= start && ts.isWhiteSpaceSingleLine(sourceFile.text.charCodeAt(pos))) {
                        pos--;
                    }
                    if (pos !== end) {
                        return pos + 1;
                    }
                    return -1;
                }
                /**
                 * Trimming will be done for lines after the previous range
                 */
                function trimTrailingWhitespacesForRemainingRange() {
                    var startPosition = previousRange ? previousRange.end : originalRange.pos;
                    var startLine = sourceFile.getLineAndCharacterOfPosition(startPosition).line;
                    var endLine = sourceFile.getLineAndCharacterOfPosition(originalRange.end).line;
                    trimTrailingWhitespacesForLines(startLine, endLine + 1, previousRange);
                }
                function recordDelete(start, len) {
                    if (len) {
                        edits.push(ts.createTextChangeFromStartLength(start, len, ""));
                    }
                }
                function recordReplace(start, len, newText) {
                    if (len || newText) {
                        edits.push(ts.createTextChangeFromStartLength(start, len, newText));
                    }
                }
                function recordInsert(start, text) {
                    if (text) {
                        edits.push(ts.createTextChangeFromStartLength(start, 0, text));
                    }
                }
                function applyRuleEdits(rule, previousRange, previousStartLine, currentRange, currentStartLine) {
                    var onLaterLine = currentStartLine !== previousStartLine;
                    switch (rule.action) {
                        case 1 /* StopProcessingSpaceActions */:
                            // no action required
                            return 0 /* None */;
                        case 16 /* DeleteSpace */:
                            if (previousRange.end !== currentRange.pos) {
                                // delete characters starting from t1.end up to t2.pos exclusive
                                recordDelete(previousRange.end, currentRange.pos - previousRange.end);
                                return onLaterLine ? 2 /* LineRemoved */ : 0 /* None */;
                            }
                            break;
                        case 32 /* DeleteToken */:
                            recordDelete(previousRange.pos, previousRange.end - previousRange.pos);
                            break;
                        case 8 /* InsertNewLine */:
                            // exit early if we on different lines and rule cannot change number of newlines
                            // if line1 and line2 are on subsequent lines then no edits are required - ok to exit
                            // if line1 and line2 are separated with more than one newline - ok to exit since we cannot delete extra new lines
                            if (rule.flags !== 1 /* CanDeleteNewLines */ && previousStartLine !== currentStartLine) {
                                return 0 /* None */;
                            }
                            // edit should not be applied if we have one line feed between elements
                            var lineDelta = currentStartLine - previousStartLine;
                            if (lineDelta !== 1) {
                                recordReplace(previousRange.end, currentRange.pos - previousRange.end, ts.getNewLineOrDefaultFromHost(host, options));
                                return onLaterLine ? 0 /* None */ : 1 /* LineAdded */;
                            }
                            break;
                        case 4 /* InsertSpace */:
                            // exit early if we on different lines and rule cannot change number of newlines
                            if (rule.flags !== 1 /* CanDeleteNewLines */ && previousStartLine !== currentStartLine) {
                                return 0 /* None */;
                            }
                            var posDelta = currentRange.pos - previousRange.end;
                            if (posDelta !== 1 || sourceFile.text.charCodeAt(previousRange.end) !== 32 /* space */) {
                                recordReplace(previousRange.end, currentRange.pos - previousRange.end, " ");
                                return onLaterLine ? 2 /* LineRemoved */ : 0 /* None */;
                            }
                            break;
                        case 64 /* InsertTrailingSemicolon */:
                            recordInsert(previousRange.end, ";");
                    }
                    return 0 /* None */;
                }
            }
            var LineAction;
            (function (LineAction) {
                LineAction[LineAction["None"] = 0] = "None";
                LineAction[LineAction["LineAdded"] = 1] = "LineAdded";
                LineAction[LineAction["LineRemoved"] = 2] = "LineRemoved";
            })(LineAction || (LineAction = {}));
            /**
             * @param precedingToken pass `null` if preceding token was already computed and result was `undefined`.
             */
            function getRangeOfEnclosingComment(sourceFile, position, precedingToken, tokenAtPosition) {
                if (tokenAtPosition === void 0) {
                    tokenAtPosition = ts.getTokenAtPosition(sourceFile, position);
                }
                var jsdoc = ts.findAncestor(tokenAtPosition, ts.isJSDoc);
                if (jsdoc)
                    tokenAtPosition = jsdoc.parent;
                var tokenStart = tokenAtPosition.getStart(sourceFile);
                if (tokenStart <= position && position < tokenAtPosition.getEnd()) {
                    return undefined;
                }
                // eslint-disable-next-line no-null/no-null
                precedingToken = precedingToken === null ? undefined : precedingToken === undefined ? ts.findPrecedingToken(position, sourceFile) : precedingToken;
                // Between two consecutive tokens, all comments are either trailing on the former
                // or leading on the latter (and none are in both lists).
                var trailingRangesOfPreviousToken = precedingToken && ts.getTrailingCommentRanges(sourceFile.text, precedingToken.end);
                var leadingCommentRangesOfNextToken = ts.getLeadingCommentRangesOfNode(tokenAtPosition, sourceFile);
                var commentRanges = ts.concatenate(trailingRangesOfPreviousToken, leadingCommentRangesOfNextToken);
                return commentRanges && ts.find(commentRanges, function (range) {
                    return ts.rangeContainsPositionExclusive(range, position) ||
                        // The end marker of a single-line comment does not include the newline character.
                        // With caret at `^`, in the following case, we are inside a comment (^ denotes the cursor position):
                        //
                        //    // asdf   ^\n
                        //
                        // But for closed multi-line comments, we don't want to be inside the comment in the following case:
                        //
                        //    /* asdf */^
                        //
                        // However, unterminated multi-line comments *do* contain their end.
                        //
                        // Internally, we represent the end of the comment at the newline and closing '/', respectively.
                        //
                        position === range.end && (range.kind === 2 /* SingleLineCommentTrivia */ || position === sourceFile.getFullWidth());
                });
            }
            formatting.getRangeOfEnclosingComment = getRangeOfEnclosingComment;
            function getOpenTokenForList(node, list) {
                switch (node.kind) {
                    case 169 /* Constructor */:
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 212 /* ArrowFunction */:
                        if (node.typeParameters === list) {
                            return 29 /* LessThanToken */;
                        }
                        else if (node.parameters === list) {
                            return 20 /* OpenParenToken */;
                        }
                        break;
                    case 206 /* CallExpression */:
                    case 207 /* NewExpression */:
                        if (node.typeArguments === list) {
                            return 29 /* LessThanToken */;
                        }
                        else if (node.arguments === list) {
                            return 20 /* OpenParenToken */;
                        }
                        break;
                    case 176 /* TypeReference */:
                        if (node.typeArguments === list) {
                            return 29 /* LessThanToken */;
                        }
                        break;
                    case 180 /* TypeLiteral */:
                        return 18 /* OpenBraceToken */;
                }
                return 0 /* Unknown */;
            }
            function getCloseTokenForOpenToken(kind) {
                switch (kind) {
                    case 20 /* OpenParenToken */:
                        return 21 /* CloseParenToken */;
                    case 29 /* LessThanToken */:
                        return 31 /* GreaterThanToken */;
                    case 18 /* OpenBraceToken */:
                        return 19 /* CloseBraceToken */;
                }
                return 0 /* Unknown */;
            }
            var internedSizes;
            var internedTabsIndentation;
            var internedSpacesIndentation;
            function getIndentationString(indentation, options) {
                // reset interned strings if FormatCodeOptions were changed
                var resetInternedStrings = !internedSizes || (internedSizes.tabSize !== options.tabSize || internedSizes.indentSize !== options.indentSize);
                if (resetInternedStrings) {
                    internedSizes = { tabSize: options.tabSize, indentSize: options.indentSize };
                    internedTabsIndentation = internedSpacesIndentation = undefined;
                }
                if (!options.convertTabsToSpaces) {
                    var tabs = Math.floor(indentation / options.tabSize);
                    var spaces = indentation - tabs * options.tabSize;
                    var tabString = void 0;
                    if (!internedTabsIndentation) {
                        internedTabsIndentation = [];
                    }
                    if (internedTabsIndentation[tabs] === undefined) {
                        internedTabsIndentation[tabs] = tabString = ts.repeatString("\t", tabs);
                    }
                    else {
                        tabString = internedTabsIndentation[tabs];
                    }
                    return spaces ? tabString + ts.repeatString(" ", spaces) : tabString;
                }
                else {
                    var spacesString = void 0;
                    var quotient = Math.floor(indentation / options.indentSize);
                    var remainder = indentation % options.indentSize;
                    if (!internedSpacesIndentation) {
                        internedSpacesIndentation = [];
                    }
                    if (internedSpacesIndentation[quotient] === undefined) {
                        spacesString = ts.repeatString(" ", options.indentSize * quotient);
                        internedSpacesIndentation[quotient] = spacesString;
                    }
                    else {
                        spacesString = internedSpacesIndentation[quotient];
                    }
                    return remainder ? spacesString + ts.repeatString(" ", remainder) : spacesString;
                }
            }
            formatting.getIndentationString = getIndentationString;
        })(formatting = ts.formatting || (ts.formatting = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var formatting;
        (function (formatting) {
            var SmartIndenter;
            (function (SmartIndenter) {
                var Value;
                (function (Value) {
                    Value[Value["Unknown"] = -1] = "Unknown";
                })(Value || (Value = {}));
                /**
                 * @param assumeNewLineBeforeCloseBrace
                 * `false` when called on text from a real source file.
                 * `true` when we need to assume `position` is on a newline.
                 *
                 * This is useful for codefixes. Consider
                 * ```
                 * function f() {
                 * |}
                 * ```
                 * with `position` at `|`.
                 *
                 * When inserting some text after an open brace, we would like to get indentation as if a newline was already there.
                 * By default indentation at `position` will be 0 so 'assumeNewLineBeforeCloseBrace' overrides this behavior.
                 */
                function getIndentation(position, sourceFile, options, assumeNewLineBeforeCloseBrace) {
                    if (assumeNewLineBeforeCloseBrace === void 0) {
                        assumeNewLineBeforeCloseBrace = false;
                    }
                    if (position > sourceFile.text.length) {
                        return getBaseIndentation(options); // past EOF
                    }
                    // no indentation when the indent style is set to none,
                    // so we can return fast
                    if (options.indentStyle === ts.IndentStyle.None) {
                        return 0;
                    }
                    var precedingToken = ts.findPrecedingToken(position, sourceFile, /*startNode*/ undefined, /*excludeJsdoc*/ true);
                    // eslint-disable-next-line no-null/no-null
                    var enclosingCommentRange = formatting.getRangeOfEnclosingComment(sourceFile, position, precedingToken || null);
                    if (enclosingCommentRange && enclosingCommentRange.kind === 3 /* MultiLineCommentTrivia */) {
                        return getCommentIndent(sourceFile, position, options, enclosingCommentRange);
                    }
                    if (!precedingToken) {
                        return getBaseIndentation(options);
                    }
                    // no indentation in string \regex\template literals
                    var precedingTokenIsLiteral = ts.isStringOrRegularExpressionOrTemplateLiteral(precedingToken.kind);
                    if (precedingTokenIsLiteral && precedingToken.getStart(sourceFile) <= position && position < precedingToken.end) {
                        return 0;
                    }
                    var lineAtPosition = sourceFile.getLineAndCharacterOfPosition(position).line;
                    // indentation is first non-whitespace character in a previous line
                    // for block indentation, we should look for a line which contains something that's not
                    // whitespace.
                    if (options.indentStyle === ts.IndentStyle.Block) {
                        return getBlockIndent(sourceFile, position, options);
                    }
                    if (precedingToken.kind === 27 /* CommaToken */ && precedingToken.parent.kind !== 219 /* BinaryExpression */) {
                        // previous token is comma that separates items in list - find the previous item and try to derive indentation from it
                        var actualIndentation = getActualIndentationForListItemBeforeComma(precedingToken, sourceFile, options);
                        if (actualIndentation !== -1 /* Unknown */) {
                            return actualIndentation;
                        }
                    }
                    var containerList = getListByPosition(position, precedingToken.parent, sourceFile);
                    // use list position if the preceding token is before any list items
                    if (containerList && !ts.rangeContainsRange(containerList, precedingToken)) {
                        return getActualIndentationForListStartLine(containerList, sourceFile, options) + options.indentSize; // TODO: GH#18217
                    }
                    return getSmartIndent(sourceFile, position, precedingToken, lineAtPosition, assumeNewLineBeforeCloseBrace, options);
                }
                SmartIndenter.getIndentation = getIndentation;
                function getCommentIndent(sourceFile, position, options, enclosingCommentRange) {
                    var previousLine = ts.getLineAndCharacterOfPosition(sourceFile, position).line - 1;
                    var commentStartLine = ts.getLineAndCharacterOfPosition(sourceFile, enclosingCommentRange.pos).line;
                    ts.Debug.assert(commentStartLine >= 0);
                    if (previousLine <= commentStartLine) {
                        return findFirstNonWhitespaceColumn(ts.getStartPositionOfLine(commentStartLine, sourceFile), position, sourceFile, options);
                    }
                    var startPositionOfLine = ts.getStartPositionOfLine(previousLine, sourceFile);
                    var _a = findFirstNonWhitespaceCharacterAndColumn(startPositionOfLine, position, sourceFile, options), column = _a.column, character = _a.character;
                    if (column === 0) {
                        return column;
                    }
                    var firstNonWhitespaceCharacterCode = sourceFile.text.charCodeAt(startPositionOfLine + character);
                    return firstNonWhitespaceCharacterCode === 42 /* asterisk */ ? column - 1 : column;
                }
                function getBlockIndent(sourceFile, position, options) {
                    // move backwards until we find a line with a non-whitespace character,
                    // then find the first non-whitespace character for that line.
                    var current = position;
                    while (current > 0) {
                        var char = sourceFile.text.charCodeAt(current);
                        if (!ts.isWhiteSpaceLike(char)) {
                            break;
                        }
                        current--;
                    }
                    var lineStart = ts.getLineStartPositionForPosition(current, sourceFile);
                    return findFirstNonWhitespaceColumn(lineStart, current, sourceFile, options);
                }
                function getSmartIndent(sourceFile, position, precedingToken, lineAtPosition, assumeNewLineBeforeCloseBrace, options) {
                    // try to find node that can contribute to indentation and includes 'position' starting from 'precedingToken'
                    // if such node is found - compute initial indentation for 'position' inside this node
                    var previous;
                    var current = precedingToken;
                    while (current) {
                        if (ts.positionBelongsToNode(current, position, sourceFile) && shouldIndentChildNode(options, current, previous, sourceFile, /*isNextChild*/ true)) {
                            var currentStart = getStartLineAndCharacterForNode(current, sourceFile);
                            var nextTokenKind = nextTokenIsCurlyBraceOnSameLineAsCursor(precedingToken, current, lineAtPosition, sourceFile);
                            var indentationDelta = nextTokenKind !== 0 /* Unknown */
                                // handle cases when codefix is about to be inserted before the close brace
                                ? assumeNewLineBeforeCloseBrace && nextTokenKind === 2 /* CloseBrace */ ? options.indentSize : 0
                                : lineAtPosition !== currentStart.line ? options.indentSize : 0;
                            return getIndentationForNodeWorker(current, currentStart, /*ignoreActualIndentationRange*/ undefined, indentationDelta, sourceFile, /*isNextChild*/ true, options); // TODO: GH#18217
                        }
                        // check if current node is a list item - if yes, take indentation from it
                        // do not consider parent-child line sharing yet:
                        // function foo(a
                        //    | preceding node 'a' does share line with its parent but indentation is expected
                        var actualIndentation = getActualIndentationForListItem(current, sourceFile, options, /*listIndentsChild*/ true);
                        if (actualIndentation !== -1 /* Unknown */) {
                            return actualIndentation;
                        }
                        previous = current;
                        current = current.parent;
                    }
                    // no parent was found - return the base indentation of the SourceFile
                    return getBaseIndentation(options);
                }
                function getIndentationForNode(n, ignoreActualIndentationRange, sourceFile, options) {
                    var start = sourceFile.getLineAndCharacterOfPosition(n.getStart(sourceFile));
                    return getIndentationForNodeWorker(n, start, ignoreActualIndentationRange, /*indentationDelta*/ 0, sourceFile, /*isNextChild*/ false, options);
                }
                SmartIndenter.getIndentationForNode = getIndentationForNode;
                function getBaseIndentation(options) {
                    return options.baseIndentSize || 0;
                }
                SmartIndenter.getBaseIndentation = getBaseIndentation;
                function getIndentationForNodeWorker(current, currentStart, ignoreActualIndentationRange, indentationDelta, sourceFile, isNextChild, options) {
                    var _a;
                    var parent = current.parent;
                    // Walk up the tree and collect indentation for parent-child node pairs. Indentation is not added if
                    // * parent and child nodes start on the same line, or
                    // * parent is an IfStatement and child starts on the same line as an 'else clause'.
                    while (parent) {
                        var useActualIndentation = true;
                        if (ignoreActualIndentationRange) {
                            var start = current.getStart(sourceFile);
                            useActualIndentation = start < ignoreActualIndentationRange.pos || start > ignoreActualIndentationRange.end;
                        }
                        var containingListOrParentStart = getContainingListOrParentStart(parent, current, sourceFile);
                        var parentAndChildShareLine = containingListOrParentStart.line === currentStart.line ||
                            childStartsOnTheSameLineWithElseInIfStatement(parent, current, currentStart.line, sourceFile);
                        if (useActualIndentation) {
                            // check if current node is a list item - if yes, take indentation from it
                            var firstListChild = (_a = getContainingList(current, sourceFile)) === null || _a === void 0 ? void 0 : _a[0];
                            // A list indents its children if the children begin on a later line than the list itself:
                            //
                            // f1(               L0 - List start
                            //   {               L1 - First child start: indented, along with all other children
                            //     prop: 0
                            //   },
                            //   {
                            //     prop: 1
                            //   }
                            // )
                            //
                            // f2({             L0 - List start and first child start: children are not indented.
                            //   prop: 0             Object properties are indented only one level, because the list
                            // }, {                  itself contributes nothing.
                            //   prop: 1        L3 - The indentation of the second object literal is best understood by
                            // })                    looking at the relationship between the list and *first* list item.
                            var listIndentsChild = !!firstListChild && getStartLineAndCharacterForNode(firstListChild, sourceFile).line > containingListOrParentStart.line;
                            var actualIndentation = getActualIndentationForListItem(current, sourceFile, options, listIndentsChild);
                            if (actualIndentation !== -1 /* Unknown */) {
                                return actualIndentation + indentationDelta;
                            }
                            // try to fetch actual indentation for current node from source text
                            actualIndentation = getActualIndentationForNode(current, parent, currentStart, parentAndChildShareLine, sourceFile, options);
                            if (actualIndentation !== -1 /* Unknown */) {
                                return actualIndentation + indentationDelta;
                            }
                        }
                        // increase indentation if parent node wants its content to be indented and parent and child nodes don't start on the same line
                        if (shouldIndentChildNode(options, parent, current, sourceFile, isNextChild) && !parentAndChildShareLine) {
                            indentationDelta += options.indentSize;
                        }
                        // In our AST, a call argument's `parent` is the call-expression, not the argument list.
                        // We would like to increase indentation based on the relationship between an argument and its argument-list,
                        // so we spoof the starting position of the (parent) call-expression to match the (non-parent) argument-list.
                        // But, the spoofed start-value could then cause a problem when comparing the start position of the call-expression
                        // to *its* parent (in the case of an iife, an expression statement), adding an extra level of indentation.
                        //
                        // Instead, when at an argument, we unspoof the starting position of the enclosing call expression
                        // *after* applying indentation for the argument.
                        var useTrueStart = isArgumentAndStartLineOverlapsExpressionBeingCalled(parent, current, currentStart.line, sourceFile);
                        current = parent;
                        parent = current.parent;
                        currentStart = useTrueStart ? sourceFile.getLineAndCharacterOfPosition(current.getStart(sourceFile)) : containingListOrParentStart;
                    }
                    return indentationDelta + getBaseIndentation(options);
                }
                function getContainingListOrParentStart(parent, child, sourceFile) {
                    var containingList = getContainingList(child, sourceFile);
                    var startPos = containingList ? containingList.pos : parent.getStart(sourceFile);
                    return sourceFile.getLineAndCharacterOfPosition(startPos);
                }
                /*
                 * Function returns Value.Unknown if indentation cannot be determined
                 */
                function getActualIndentationForListItemBeforeComma(commaToken, sourceFile, options) {
                    // previous token is comma that separates items in list - find the previous item and try to derive indentation from it
                    var commaItemInfo = ts.findListItemInfo(commaToken);
                    if (commaItemInfo && commaItemInfo.listItemIndex > 0) {
                        return deriveActualIndentationFromList(commaItemInfo.list.getChildren(), commaItemInfo.listItemIndex - 1, sourceFile, options);
                    }
                    else {
                        // handle broken code gracefully
                        return -1 /* Unknown */;
                    }
                }
                /*
                 * Function returns Value.Unknown if actual indentation for node should not be used (i.e because node is nested expression)
                 */
                function getActualIndentationForNode(current, parent, currentLineAndChar, parentAndChildShareLine, sourceFile, options) {
                    // actual indentation is used for statements\declarations if one of cases below is true:
                    // - parent is SourceFile - by default immediate children of SourceFile are not indented except when user indents them manually
                    // - parent and child are not on the same line
                    var useActualIndentation = (ts.isDeclaration(current) || ts.isStatementButNotDeclaration(current)) &&
                        (parent.kind === 300 /* SourceFile */ || !parentAndChildShareLine);
                    if (!useActualIndentation) {
                        return -1 /* Unknown */;
                    }
                    return findColumnForFirstNonWhitespaceCharacterInLine(currentLineAndChar, sourceFile, options);
                }
                var NextTokenKind;
                (function (NextTokenKind) {
                    NextTokenKind[NextTokenKind["Unknown"] = 0] = "Unknown";
                    NextTokenKind[NextTokenKind["OpenBrace"] = 1] = "OpenBrace";
                    NextTokenKind[NextTokenKind["CloseBrace"] = 2] = "CloseBrace";
                })(NextTokenKind || (NextTokenKind = {}));
                function nextTokenIsCurlyBraceOnSameLineAsCursor(precedingToken, current, lineAtPosition, sourceFile) {
                    var nextToken = ts.findNextToken(precedingToken, current, sourceFile);
                    if (!nextToken) {
                        return 0 /* Unknown */;
                    }
                    if (nextToken.kind === 18 /* OpenBraceToken */) {
                        // open braces are always indented at the parent level
                        return 1 /* OpenBrace */;
                    }
                    else if (nextToken.kind === 19 /* CloseBraceToken */) {
                        // close braces are indented at the parent level if they are located on the same line with cursor
                        // this means that if new line will be added at $ position, this case will be indented
                        // class A {
                        //    $
                        // }
                        /// and this one - not
                        // class A {
                        // $}
                        var nextTokenStartLine = getStartLineAndCharacterForNode(nextToken, sourceFile).line;
                        return lineAtPosition === nextTokenStartLine ? 2 /* CloseBrace */ : 0 /* Unknown */;
                    }
                    return 0 /* Unknown */;
                }
                function getStartLineAndCharacterForNode(n, sourceFile) {
                    return sourceFile.getLineAndCharacterOfPosition(n.getStart(sourceFile));
                }
                function isArgumentAndStartLineOverlapsExpressionBeingCalled(parent, child, childStartLine, sourceFile) {
                    if (!(ts.isCallExpression(parent) && ts.contains(parent.arguments, child))) {
                        return false;
                    }
                    var expressionOfCallExpressionEnd = parent.expression.getEnd();
                    var expressionOfCallExpressionEndLine = ts.getLineAndCharacterOfPosition(sourceFile, expressionOfCallExpressionEnd).line;
                    return expressionOfCallExpressionEndLine === childStartLine;
                }
                SmartIndenter.isArgumentAndStartLineOverlapsExpressionBeingCalled = isArgumentAndStartLineOverlapsExpressionBeingCalled;
                function childStartsOnTheSameLineWithElseInIfStatement(parent, child, childStartLine, sourceFile) {
                    if (parent.kind === 237 /* IfStatement */ && parent.elseStatement === child) {
                        var elseKeyword = ts.findChildOfKind(parent, 91 /* ElseKeyword */, sourceFile);
                        ts.Debug.assert(elseKeyword !== undefined);
                        var elseKeywordStartLine = getStartLineAndCharacterForNode(elseKeyword, sourceFile).line;
                        return elseKeywordStartLine === childStartLine;
                    }
                    return false;
                }
                SmartIndenter.childStartsOnTheSameLineWithElseInIfStatement = childStartsOnTheSameLineWithElseInIfStatement;
                // A multiline conditional typically increases the indentation of its whenTrue and whenFalse children:
                //
                // condition
                //   ? whenTrue
                //   : whenFalse;
                //
                // However, that indentation does not apply if the subexpressions themselves span multiple lines,
                // applying their own indentation:
                //
                // (() => {
                //   return complexCalculationForCondition();
                // })() ? {
                //   whenTrue: 'multiline object literal'
                // } : (
                //   whenFalse('multiline parenthesized expression')
                // );
                //
                // In these cases, we must discard the indentation increase that would otherwise be applied to the
                // whenTrue and whenFalse children to avoid double-indenting their contents. To identify this scenario,
                // we check for the whenTrue branch beginning on the line that the condition ends, and the whenFalse
                // branch beginning on the line that the whenTrue branch ends.
                function childIsUnindentedBranchOfConditionalExpression(parent, child, childStartLine, sourceFile) {
                    if (ts.isConditionalExpression(parent) && (child === parent.whenTrue || child === parent.whenFalse)) {
                        var conditionEndLine = ts.getLineAndCharacterOfPosition(sourceFile, parent.condition.end).line;
                        if (child === parent.whenTrue) {
                            return childStartLine === conditionEndLine;
                        }
                        else {
                            // On the whenFalse side, we have to look at the whenTrue side, because if that one was
                            // indented, whenFalse must also be indented:
                            //
                            // const y = true
                            //   ? 1 : (          L1: whenTrue indented because it's on a new line
                            //     0              L2: indented two stops, one because whenTrue was indented
                            //   );                   and one because of the parentheses spanning multiple lines
                            var trueStartLine = getStartLineAndCharacterForNode(parent.whenTrue, sourceFile).line;
                            var trueEndLine = ts.getLineAndCharacterOfPosition(sourceFile, parent.whenTrue.end).line;
                            return conditionEndLine === trueStartLine && trueEndLine === childStartLine;
                        }
                    }
                    return false;
                }
                SmartIndenter.childIsUnindentedBranchOfConditionalExpression = childIsUnindentedBranchOfConditionalExpression;
                function argumentStartsOnSameLineAsPreviousArgument(parent, child, childStartLine, sourceFile) {
                    if (ts.isCallOrNewExpression(parent)) {
                        if (!parent.arguments)
                            return false;
                        var currentNode = ts.find(parent.arguments, function (arg) { return arg.pos === child.pos; });
                        // If it's not one of the arguments, don't look past this
                        if (!currentNode)
                            return false;
                        var currentIndex = parent.arguments.indexOf(currentNode);
                        if (currentIndex === 0)
                            return false; // Can't look at previous node if first
                        var previousNode = parent.arguments[currentIndex - 1];
                        var lineOfPreviousNode = ts.getLineAndCharacterOfPosition(sourceFile, previousNode.getEnd()).line;
                        if (childStartLine === lineOfPreviousNode) {
                            return true;
                        }
                    }
                    return false;
                }
                SmartIndenter.argumentStartsOnSameLineAsPreviousArgument = argumentStartsOnSameLineAsPreviousArgument;
                function getContainingList(node, sourceFile) {
                    return node.parent && getListByRange(node.getStart(sourceFile), node.getEnd(), node.parent, sourceFile);
                }
                SmartIndenter.getContainingList = getContainingList;
                function getListByPosition(pos, node, sourceFile) {
                    return node && getListByRange(pos, pos, node, sourceFile);
                }
                function getListByRange(start, end, node, sourceFile) {
                    switch (node.kind) {
                        case 176 /* TypeReference */:
                            return getList(node.typeArguments);
                        case 203 /* ObjectLiteralExpression */:
                            return getList(node.properties);
                        case 202 /* ArrayLiteralExpression */:
                            return getList(node.elements);
                        case 180 /* TypeLiteral */:
                            return getList(node.members);
                        case 254 /* FunctionDeclaration */:
                        case 211 /* FunctionExpression */:
                        case 212 /* ArrowFunction */:
                        case 167 /* MethodDeclaration */:
                        case 166 /* MethodSignature */:
                        case 172 /* CallSignature */:
                        case 169 /* Constructor */:
                        case 178 /* ConstructorType */:
                        case 173 /* ConstructSignature */:
                            return getList(node.typeParameters) || getList(node.parameters);
                        case 255 /* ClassDeclaration */:
                        case 224 /* ClassExpression */:
                        case 256 /* InterfaceDeclaration */:
                        case 257 /* TypeAliasDeclaration */:
                        case 339 /* JSDocTemplateTag */:
                            return getList(node.typeParameters);
                        case 207 /* NewExpression */:
                        case 206 /* CallExpression */:
                            return getList(node.typeArguments) || getList(node.arguments);
                        case 253 /* VariableDeclarationList */:
                            return getList(node.declarations);
                        case 267 /* NamedImports */:
                        case 271 /* NamedExports */:
                            return getList(node.elements);
                        case 199 /* ObjectBindingPattern */:
                        case 200 /* ArrayBindingPattern */:
                            return getList(node.elements);
                    }
                    function getList(list) {
                        return list && ts.rangeContainsStartEnd(getVisualListRange(node, list, sourceFile), start, end) ? list : undefined;
                    }
                }
                function getVisualListRange(node, list, sourceFile) {
                    var children = node.getChildren(sourceFile);
                    for (var i = 1; i < children.length - 1; i++) {
                        if (children[i].pos === list.pos && children[i].end === list.end) {
                            return { pos: children[i - 1].end, end: children[i + 1].getStart(sourceFile) };
                        }
                    }
                    return list;
                }
                function getActualIndentationForListStartLine(list, sourceFile, options) {
                    if (!list) {
                        return -1 /* Unknown */;
                    }
                    return findColumnForFirstNonWhitespaceCharacterInLine(sourceFile.getLineAndCharacterOfPosition(list.pos), sourceFile, options);
                }
                function getActualIndentationForListItem(node, sourceFile, options, listIndentsChild) {
                    if (node.parent && node.parent.kind === 253 /* VariableDeclarationList */) {
                        // VariableDeclarationList has no wrapping tokens
                        return -1 /* Unknown */;
                    }
                    var containingList = getContainingList(node, sourceFile);
                    if (containingList) {
                        var index = containingList.indexOf(node);
                        if (index !== -1) {
                            var result = deriveActualIndentationFromList(containingList, index, sourceFile, options);
                            if (result !== -1 /* Unknown */) {
                                return result;
                            }
                        }
                        return getActualIndentationForListStartLine(containingList, sourceFile, options) + (listIndentsChild ? options.indentSize : 0); // TODO: GH#18217
                    }
                    return -1 /* Unknown */;
                }
                function deriveActualIndentationFromList(list, index, sourceFile, options) {
                    ts.Debug.assert(index >= 0 && index < list.length);
                    var node = list[index];
                    // walk toward the start of the list starting from current node and check if the line is the same for all items.
                    // if end line for item [i - 1] differs from the start line for item [i] - find column of the first non-whitespace character on the line of item [i]
                    var lineAndCharacter = getStartLineAndCharacterForNode(node, sourceFile);
                    for (var i = index - 1; i >= 0; i--) {
                        if (list[i].kind === 27 /* CommaToken */) {
                            continue;
                        }
                        // skip list items that ends on the same line with the current list element
                        var prevEndLine = sourceFile.getLineAndCharacterOfPosition(list[i].end).line;
                        if (prevEndLine !== lineAndCharacter.line) {
                            return findColumnForFirstNonWhitespaceCharacterInLine(lineAndCharacter, sourceFile, options);
                        }
                        lineAndCharacter = getStartLineAndCharacterForNode(list[i], sourceFile);
                    }
                    return -1 /* Unknown */;
                }
                function findColumnForFirstNonWhitespaceCharacterInLine(lineAndCharacter, sourceFile, options) {
                    var lineStart = sourceFile.getPositionOfLineAndCharacter(lineAndCharacter.line, 0);
                    return findFirstNonWhitespaceColumn(lineStart, lineStart + lineAndCharacter.character, sourceFile, options);
                }
                /**
                 * Character is the actual index of the character since the beginning of the line.
                 * Column - position of the character after expanding tabs to spaces.
                 * "0\t2$"
                 * value of 'character' for '$' is 3
                 * value of 'column' for '$' is 6 (assuming that tab size is 4)
                 */
                function findFirstNonWhitespaceCharacterAndColumn(startPos, endPos, sourceFile, options) {
                    var character = 0;
                    var column = 0;
                    for (var pos = startPos; pos < endPos; pos++) {
                        var ch = sourceFile.text.charCodeAt(pos);
                        if (!ts.isWhiteSpaceSingleLine(ch)) {
                            break;
                        }
                        if (ch === 9 /* tab */) {
                            column += options.tabSize + (column % options.tabSize);
                        }
                        else {
                            column++;
                        }
                        character++;
                    }
                    return { column: column, character: character };
                }
                SmartIndenter.findFirstNonWhitespaceCharacterAndColumn = findFirstNonWhitespaceCharacterAndColumn;
                function findFirstNonWhitespaceColumn(startPos, endPos, sourceFile, options) {
                    return findFirstNonWhitespaceCharacterAndColumn(startPos, endPos, sourceFile, options).column;
                }
                SmartIndenter.findFirstNonWhitespaceColumn = findFirstNonWhitespaceColumn;
                function nodeWillIndentChild(settings, parent, child, sourceFile, indentByDefault) {
                    var childKind = child ? child.kind : 0 /* Unknown */;
                    switch (parent.kind) {
                        case 236 /* ExpressionStatement */:
                        case 255 /* ClassDeclaration */:
                        case 224 /* ClassExpression */:
                        case 256 /* InterfaceDeclaration */:
                        case 258 /* EnumDeclaration */:
                        case 257 /* TypeAliasDeclaration */:
                        case 202 /* ArrayLiteralExpression */:
                        case 233 /* Block */:
                        case 260 /* ModuleBlock */:
                        case 203 /* ObjectLiteralExpression */:
                        case 180 /* TypeLiteral */:
                        case 193 /* MappedType */:
                        case 182 /* TupleType */:
                        case 261 /* CaseBlock */:
                        case 288 /* DefaultClause */:
                        case 287 /* CaseClause */:
                        case 210 /* ParenthesizedExpression */:
                        case 204 /* PropertyAccessExpression */:
                        case 206 /* CallExpression */:
                        case 207 /* NewExpression */:
                        case 235 /* VariableStatement */:
                        case 269 /* ExportAssignment */:
                        case 245 /* ReturnStatement */:
                        case 220 /* ConditionalExpression */:
                        case 200 /* ArrayBindingPattern */:
                        case 199 /* ObjectBindingPattern */:
                        case 278 /* JsxOpeningElement */:
                        case 281 /* JsxOpeningFragment */:
                        case 277 /* JsxSelfClosingElement */:
                        case 286 /* JsxExpression */:
                        case 166 /* MethodSignature */:
                        case 172 /* CallSignature */:
                        case 173 /* ConstructSignature */:
                        case 162 /* Parameter */:
                        case 177 /* FunctionType */:
                        case 178 /* ConstructorType */:
                        case 189 /* ParenthesizedType */:
                        case 208 /* TaggedTemplateExpression */:
                        case 216 /* AwaitExpression */:
                        case 271 /* NamedExports */:
                        case 267 /* NamedImports */:
                        case 273 /* ExportSpecifier */:
                        case 268 /* ImportSpecifier */:
                        case 165 /* PropertyDeclaration */:
                            return true;
                        case 252 /* VariableDeclaration */:
                        case 291 /* PropertyAssignment */:
                        case 219 /* BinaryExpression */:
                            if (!settings.indentMultiLineObjectLiteralBeginningOnBlankLine && sourceFile && childKind === 203 /* ObjectLiteralExpression */) { // TODO: GH#18217
                                return rangeIsOnOneLine(sourceFile, child);
                            }
                            if (parent.kind !== 219 /* BinaryExpression */) {
                                return true;
                            }
                            break;
                        case 238 /* DoStatement */:
                        case 239 /* WhileStatement */:
                        case 241 /* ForInStatement */:
                        case 242 /* ForOfStatement */:
                        case 240 /* ForStatement */:
                        case 237 /* IfStatement */:
                        case 254 /* FunctionDeclaration */:
                        case 211 /* FunctionExpression */:
                        case 167 /* MethodDeclaration */:
                        case 169 /* Constructor */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                            return childKind !== 233 /* Block */;
                        case 212 /* ArrowFunction */:
                            if (sourceFile && childKind === 210 /* ParenthesizedExpression */) {
                                return rangeIsOnOneLine(sourceFile, child);
                            }
                            return childKind !== 233 /* Block */;
                        case 270 /* ExportDeclaration */:
                            return childKind !== 271 /* NamedExports */;
                        case 264 /* ImportDeclaration */:
                            return childKind !== 265 /* ImportClause */ ||
                                (!!child.namedBindings && child.namedBindings.kind !== 267 /* NamedImports */);
                        case 276 /* JsxElement */:
                            return childKind !== 279 /* JsxClosingElement */;
                        case 280 /* JsxFragment */:
                            return childKind !== 282 /* JsxClosingFragment */;
                        case 186 /* IntersectionType */:
                        case 185 /* UnionType */:
                            if (childKind === 180 /* TypeLiteral */ || childKind === 182 /* TupleType */) {
                                return false;
                            }
                            break;
                    }
                    // No explicit rule for given nodes so the result will follow the default value argument
                    return indentByDefault;
                }
                SmartIndenter.nodeWillIndentChild = nodeWillIndentChild;
                function isControlFlowEndingStatement(kind, parent) {
                    switch (kind) {
                        case 245 /* ReturnStatement */:
                        case 249 /* ThrowStatement */:
                        case 243 /* ContinueStatement */:
                        case 244 /* BreakStatement */:
                            return parent.kind !== 233 /* Block */;
                        default:
                            return false;
                    }
                }
                /**
                 * True when the parent node should indent the given child by an explicit rule.
                 * @param isNextChild If true, we are judging indent of a hypothetical child *after* this one, not the current child.
                 */
                function shouldIndentChildNode(settings, parent, child, sourceFile, isNextChild) {
                    if (isNextChild === void 0) {
                        isNextChild = false;
                    }
                    return nodeWillIndentChild(settings, parent, child, sourceFile, /*indentByDefault*/ false)
                        && !(isNextChild && child && isControlFlowEndingStatement(child.kind, parent));
                }
                SmartIndenter.shouldIndentChildNode = shouldIndentChildNode;
                function rangeIsOnOneLine(sourceFile, range) {
                    var rangeStart = ts.skipTrivia(sourceFile.text, range.pos);
                    var startLine = sourceFile.getLineAndCharacterOfPosition(rangeStart).line;
                    var endLine = sourceFile.getLineAndCharacterOfPosition(range.end).line;
                    return startLine === endLine;
                }
            })(SmartIndenter = formatting.SmartIndenter || (formatting.SmartIndenter = {}));
        })(formatting = ts.formatting || (ts.formatting = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var textChanges;
        (function (textChanges_3) {
            /**
             * Currently for simplicity we store recovered positions on the node itself.
             * It can be changed to side-table later if we decide that current design is too invasive.
             */
            function getPos(n) {
                var result = n.__pos;
                ts.Debug.assert(typeof result === "number");
                return result;
            }
            function setPos(n, pos) {
                ts.Debug.assert(typeof pos === "number");
                n.__pos = pos;
            }
            function getEnd(n) {
                var result = n.__end;
                ts.Debug.assert(typeof result === "number");
                return result;
            }
            function setEnd(n, end) {
                ts.Debug.assert(typeof end === "number");
                n.__end = end;
            }
            var LeadingTriviaOption;
            (function (LeadingTriviaOption) {
                /** Exclude all leading trivia (use getStart()) */
                LeadingTriviaOption[LeadingTriviaOption["Exclude"] = 0] = "Exclude";
                /** Include leading trivia and,
                 * if there are no line breaks between the node and the previous token,
                 * include all trivia between the node and the previous token
                 */
                LeadingTriviaOption[LeadingTriviaOption["IncludeAll"] = 1] = "IncludeAll";
                /**
                 * Include attached JSDoc comments
                 */
                LeadingTriviaOption[LeadingTriviaOption["JSDoc"] = 2] = "JSDoc";
                /**
                 * Only delete trivia on the same line as getStart().
                 * Used to avoid deleting leading comments
                 */
                LeadingTriviaOption[LeadingTriviaOption["StartLine"] = 3] = "StartLine";
            })(LeadingTriviaOption = textChanges_3.LeadingTriviaOption || (textChanges_3.LeadingTriviaOption = {}));
            var TrailingTriviaOption;
            (function (TrailingTriviaOption) {
                /** Exclude all trailing trivia (use getEnd()) */
                TrailingTriviaOption[TrailingTriviaOption["Exclude"] = 0] = "Exclude";
                /** Doesn't include whitespace, but does strip comments */
                TrailingTriviaOption[TrailingTriviaOption["ExcludeWhitespace"] = 1] = "ExcludeWhitespace";
                /** Include trailing trivia */
                TrailingTriviaOption[TrailingTriviaOption["Include"] = 2] = "Include";
            })(TrailingTriviaOption = textChanges_3.TrailingTriviaOption || (textChanges_3.TrailingTriviaOption = {}));
            function skipWhitespacesAndLineBreaks(text, start) {
                return ts.skipTrivia(text, start, /*stopAfterLineBreak*/ false, /*stopAtComments*/ true);
            }
            function hasCommentsBeforeLineBreak(text, start) {
                var i = start;
                while (i < text.length) {
                    var ch = text.charCodeAt(i);
                    if (ts.isWhiteSpaceSingleLine(ch)) {
                        i++;
                        continue;
                    }
                    return ch === 47 /* slash */;
                }
                return false;
            }
            var useNonAdjustedPositions = {
                leadingTriviaOption: LeadingTriviaOption.Exclude,
                trailingTriviaOption: TrailingTriviaOption.Exclude,
            };
            var ChangeKind;
            (function (ChangeKind) {
                ChangeKind[ChangeKind["Remove"] = 0] = "Remove";
                ChangeKind[ChangeKind["ReplaceWithSingleNode"] = 1] = "ReplaceWithSingleNode";
                ChangeKind[ChangeKind["ReplaceWithMultipleNodes"] = 2] = "ReplaceWithMultipleNodes";
                ChangeKind[ChangeKind["Text"] = 3] = "Text";
            })(ChangeKind || (ChangeKind = {}));
            function getAdjustedRange(sourceFile, startNode, endNode, options) {
                return { pos: getAdjustedStartPosition(sourceFile, startNode, options), end: getAdjustedEndPosition(sourceFile, endNode, options) };
            }
            function getAdjustedStartPosition(sourceFile, node, options, hasTrailingComment) {
                var _a, _b;
                if (hasTrailingComment === void 0) {
                    hasTrailingComment = false;
                }
                var leadingTriviaOption = options.leadingTriviaOption;
                if (leadingTriviaOption === LeadingTriviaOption.Exclude) {
                    return node.getStart(sourceFile);
                }
                if (leadingTriviaOption === LeadingTriviaOption.StartLine) {
                    var startPos = node.getStart(sourceFile);
                    var pos = ts.getLineStartPositionForPosition(startPos, sourceFile);
                    return ts.rangeContainsPosition(node, pos) ? pos : startPos;
                }
                if (leadingTriviaOption === LeadingTriviaOption.JSDoc) {
                    var JSDocComments = ts.getJSDocCommentRanges(node, sourceFile.text);
                    if (JSDocComments === null || JSDocComments === void 0 ? void 0 : JSDocComments.length) {
                        return ts.getLineStartPositionForPosition(JSDocComments[0].pos, sourceFile);
                    }
                }
                var fullStart = node.getFullStart();
                var start = node.getStart(sourceFile);
                if (fullStart === start) {
                    return start;
                }
                var fullStartLine = ts.getLineStartPositionForPosition(fullStart, sourceFile);
                var startLine = ts.getLineStartPositionForPosition(start, sourceFile);
                if (startLine === fullStartLine) {
                    // full start and start of the node are on the same line
                    //   a,     b;
                    //    ^     ^
                    //    |   start
                    // fullstart
                    // when b is replaced - we usually want to keep the leading trvia
                    // when b is deleted - we delete it
                    return leadingTriviaOption === LeadingTriviaOption.IncludeAll ? fullStart : start;
                }
                // if node has a trailing comments, use comment end position as the text has already been included.
                if (hasTrailingComment) {
                    // Check first for leading comments as if the node is the first import, we want to exclude the trivia;
                    // otherwise we get the trailing comments.
                    var comment = ((_a = ts.getLeadingCommentRanges(sourceFile.text, fullStart)) === null || _a === void 0 ? void 0 : _a[0]) || ((_b = ts.getTrailingCommentRanges(sourceFile.text, fullStart)) === null || _b === void 0 ? void 0 : _b[0]);
                    if (comment) {
                        return ts.skipTrivia(sourceFile.text, comment.end, /*stopAfterLineBreak*/ true, /*stopAtComments*/ true);
                    }
                }
                // get start position of the line following the line that contains fullstart position
                // (but only if the fullstart isn't the very beginning of the file)
                var nextLineStart = fullStart > 0 ? 1 : 0;
                var adjustedStartPosition = ts.getStartPositionOfLine(ts.getLineOfLocalPosition(sourceFile, fullStartLine) + nextLineStart, sourceFile);
                // skip whitespaces/newlines
                adjustedStartPosition = skipWhitespacesAndLineBreaks(sourceFile.text, adjustedStartPosition);
                return ts.getStartPositionOfLine(ts.getLineOfLocalPosition(sourceFile, adjustedStartPosition), sourceFile);
            }
            /** Return the end position of a multiline comment of it is on another line; otherwise returns `undefined`; */
            function getEndPositionOfMultilineTrailingComment(sourceFile, node, options) {
                var end = node.end;
                var trailingTriviaOption = options.trailingTriviaOption;
                if (trailingTriviaOption === TrailingTriviaOption.Include) {
                    // If the trailing comment is a multiline comment that extends to the next lines,
                    // return the end of the comment and track it for the next nodes to adjust.
                    var comments = ts.getTrailingCommentRanges(sourceFile.text, end);
                    if (comments) {
                        var nodeEndLine = ts.getLineOfLocalPosition(sourceFile, node.end);
                        for (var _i = 0, comments_2 = comments; _i < comments_2.length; _i++) {
                            var comment = comments_2[_i];
                            // Single line can break the loop as trivia will only be this line.
                            // Comments on subsequest lines are also ignored.
                            if (comment.kind === 2 /* SingleLineCommentTrivia */ || ts.getLineOfLocalPosition(sourceFile, comment.pos) > nodeEndLine) {
                                break;
                            }
                            // Get the end line of the comment and compare against the end line of the node.
                            // If the comment end line position and the multiline comment extends to multiple lines,
                            // then is safe to return the end position.
                            var commentEndLine = ts.getLineOfLocalPosition(sourceFile, comment.end);
                            if (commentEndLine > nodeEndLine) {
                                return ts.skipTrivia(sourceFile.text, comment.end, /*stopAfterLineBreak*/ true, /*stopAtComments*/ true);
                            }
                        }
                    }
                }
                return undefined;
            }
            function getAdjustedEndPosition(sourceFile, node, options) {
                var _a;
                var end = node.end;
                var trailingTriviaOption = options.trailingTriviaOption;
                if (trailingTriviaOption === TrailingTriviaOption.Exclude) {
                    return end;
                }
                if (trailingTriviaOption === TrailingTriviaOption.ExcludeWhitespace) {
                    var comments = ts.concatenate(ts.getTrailingCommentRanges(sourceFile.text, end), ts.getLeadingCommentRanges(sourceFile.text, end));
                    var realEnd = (_a = comments === null || comments === void 0 ? void 0 : comments[comments.length - 1]) === null || _a === void 0 ? void 0 : _a.end;
                    if (realEnd) {
                        return realEnd;
                    }
                    return end;
                }
                var multilineEndPosition = getEndPositionOfMultilineTrailingComment(sourceFile, node, options);
                if (multilineEndPosition) {
                    return multilineEndPosition;
                }
                var newEnd = ts.skipTrivia(sourceFile.text, end, /*stopAfterLineBreak*/ true);
                return newEnd !== end && (trailingTriviaOption === TrailingTriviaOption.Include || ts.isLineBreak(sourceFile.text.charCodeAt(newEnd - 1)))
                    ? newEnd
                    : end;
            }
            /**
             * Checks if 'candidate' argument is a legal separator in the list that contains 'node' as an element
             */
            function isSeparator(node, candidate) {
                return !!candidate && !!node.parent && (candidate.kind === 27 /* CommaToken */ || (candidate.kind === 26 /* SemicolonToken */ && node.parent.kind === 203 /* ObjectLiteralExpression */));
            }
            function isThisTypeAnnotatable(containingFunction) {
                return ts.isFunctionExpression(containingFunction) || ts.isFunctionDeclaration(containingFunction);
            }
            textChanges_3.isThisTypeAnnotatable = isThisTypeAnnotatable;
            var ChangeTracker = /** @class */ (function () {
                /** Public for tests only. Other callers should use `ChangeTracker.with`. */
                function ChangeTracker(newLineCharacter, formatContext) {
                    this.newLineCharacter = newLineCharacter;
                    this.formatContext = formatContext;
                    this.changes = [];
                    this.newFiles = [];
                    this.classesWithNodesInsertedAtStart = new ts.Map(); // Set<ClassDeclaration> implemented as Map<node id, ClassDeclaration>
                    this.deletedNodes = [];
                }
                ChangeTracker.fromContext = function (context) {
                    return new ChangeTracker(ts.getNewLineOrDefaultFromHost(context.host, context.formatContext.options), context.formatContext);
                };
                ChangeTracker.with = function (context, cb) {
                    var tracker = ChangeTracker.fromContext(context);
                    cb(tracker);
                    return tracker.getChanges();
                };
                ChangeTracker.prototype.pushRaw = function (sourceFile, change) {
                    ts.Debug.assertEqual(sourceFile.fileName, change.fileName);
                    for (var _i = 0, _a = change.textChanges; _i < _a.length; _i++) {
                        var c = _a[_i];
                        this.changes.push({
                            kind: ChangeKind.Text,
                            sourceFile: sourceFile,
                            text: c.newText,
                            range: ts.createTextRangeFromSpan(c.span),
                        });
                    }
                };
                ChangeTracker.prototype.deleteRange = function (sourceFile, range) {
                    this.changes.push({ kind: ChangeKind.Remove, sourceFile: sourceFile, range: range });
                };
                ChangeTracker.prototype.delete = function (sourceFile, node) {
                    this.deletedNodes.push({ sourceFile: sourceFile, node: node });
                };
                ChangeTracker.prototype.deleteNode = function (sourceFile, node, options) {
                    if (options === void 0) {
                        options = { leadingTriviaOption: LeadingTriviaOption.IncludeAll };
                    }
                    this.deleteRange(sourceFile, getAdjustedRange(sourceFile, node, node, options));
                };
                ChangeTracker.prototype.deleteNodes = function (sourceFile, nodes, options, hasTrailingComment) {
                    if (options === void 0) {
                        options = { leadingTriviaOption: LeadingTriviaOption.IncludeAll };
                    }
                    // When deleting multiple nodes we need to track if the end position is including multiline trailing comments.
                    for (var _i = 0, nodes_1 = nodes; _i < nodes_1.length; _i++) {
                        var node = nodes_1[_i];
                        var pos = getAdjustedStartPosition(sourceFile, node, options, hasTrailingComment);
                        var end = getAdjustedEndPosition(sourceFile, node, options);
                        this.deleteRange(sourceFile, { pos: pos, end: end });
                        hasTrailingComment = !!getEndPositionOfMultilineTrailingComment(sourceFile, node, options);
                    }
                };
                ChangeTracker.prototype.deleteModifier = function (sourceFile, modifier) {
                    this.deleteRange(sourceFile, { pos: modifier.getStart(sourceFile), end: ts.skipTrivia(sourceFile.text, modifier.end, /*stopAfterLineBreak*/ true) });
                };
                ChangeTracker.prototype.deleteNodeRange = function (sourceFile, startNode, endNode, options) {
                    if (options === void 0) {
                        options = { leadingTriviaOption: LeadingTriviaOption.IncludeAll };
                    }
                    var startPosition = getAdjustedStartPosition(sourceFile, startNode, options);
                    var endPosition = getAdjustedEndPosition(sourceFile, endNode, options);
                    this.deleteRange(sourceFile, { pos: startPosition, end: endPosition });
                };
                ChangeTracker.prototype.deleteNodeRangeExcludingEnd = function (sourceFile, startNode, afterEndNode, options) {
                    if (options === void 0) {
                        options = { leadingTriviaOption: LeadingTriviaOption.IncludeAll };
                    }
                    var startPosition = getAdjustedStartPosition(sourceFile, startNode, options);
                    var endPosition = afterEndNode === undefined ? sourceFile.text.length : getAdjustedStartPosition(sourceFile, afterEndNode, options);
                    this.deleteRange(sourceFile, { pos: startPosition, end: endPosition });
                };
                ChangeTracker.prototype.replaceRange = function (sourceFile, range, newNode, options) {
                    if (options === void 0) {
                        options = {};
                    }
                    this.changes.push({ kind: ChangeKind.ReplaceWithSingleNode, sourceFile: sourceFile, range: range, options: options, node: newNode });
                };
                ChangeTracker.prototype.replaceNode = function (sourceFile, oldNode, newNode, options) {
                    if (options === void 0) {
                        options = useNonAdjustedPositions;
                    }
                    this.replaceRange(sourceFile, getAdjustedRange(sourceFile, oldNode, oldNode, options), newNode, options);
                };
                ChangeTracker.prototype.replaceNodeRange = function (sourceFile, startNode, endNode, newNode, options) {
                    if (options === void 0) {
                        options = useNonAdjustedPositions;
                    }
                    this.replaceRange(sourceFile, getAdjustedRange(sourceFile, startNode, endNode, options), newNode, options);
                };
                ChangeTracker.prototype.replaceRangeWithNodes = function (sourceFile, range, newNodes, options) {
                    if (options === void 0) {
                        options = {};
                    }
                    this.changes.push({ kind: ChangeKind.ReplaceWithMultipleNodes, sourceFile: sourceFile, range: range, options: options, nodes: newNodes });
                };
                ChangeTracker.prototype.replaceNodeWithNodes = function (sourceFile, oldNode, newNodes, options) {
                    if (options === void 0) {
                        options = useNonAdjustedPositions;
                    }
                    this.replaceRangeWithNodes(sourceFile, getAdjustedRange(sourceFile, oldNode, oldNode, options), newNodes, options);
                };
                ChangeTracker.prototype.replaceNodeWithText = function (sourceFile, oldNode, text) {
                    this.replaceRangeWithText(sourceFile, getAdjustedRange(sourceFile, oldNode, oldNode, useNonAdjustedPositions), text);
                };
                ChangeTracker.prototype.replaceNodeRangeWithNodes = function (sourceFile, startNode, endNode, newNodes, options) {
                    if (options === void 0) {
                        options = useNonAdjustedPositions;
                    }
                    this.replaceRangeWithNodes(sourceFile, getAdjustedRange(sourceFile, startNode, endNode, options), newNodes, options);
                };
                ChangeTracker.prototype.nodeHasTrailingComment = function (sourceFile, oldNode, configurableEnd) {
                    if (configurableEnd === void 0) {
                        configurableEnd = useNonAdjustedPositions;
                    }
                    return !!getEndPositionOfMultilineTrailingComment(sourceFile, oldNode, configurableEnd);
                };
                ChangeTracker.prototype.nextCommaToken = function (sourceFile, node) {
                    var next = ts.findNextToken(node, node.parent, sourceFile);
                    return next && next.kind === 27 /* CommaToken */ ? next : undefined;
                };
                ChangeTracker.prototype.replacePropertyAssignment = function (sourceFile, oldNode, newNode) {
                    var suffix = this.nextCommaToken(sourceFile, oldNode) ? "" : ("," + this.newLineCharacter);
                    this.replaceNode(sourceFile, oldNode, newNode, { suffix: suffix });
                };
                ChangeTracker.prototype.insertNodeAt = function (sourceFile, pos, newNode, options) {
                    if (options === void 0) {
                        options = {};
                    }
                    this.replaceRange(sourceFile, ts.createRange(pos), newNode, options);
                };
                ChangeTracker.prototype.insertNodesAt = function (sourceFile, pos, newNodes, options) {
                    if (options === void 0) {
                        options = {};
                    }
                    this.replaceRangeWithNodes(sourceFile, ts.createRange(pos), newNodes, options);
                };
                ChangeTracker.prototype.insertNodeAtTopOfFile = function (sourceFile, newNode, blankLineBetween) {
                    this.insertAtTopOfFile(sourceFile, newNode, blankLineBetween);
                };
                ChangeTracker.prototype.insertNodesAtTopOfFile = function (sourceFile, newNodes, blankLineBetween) {
                    this.insertAtTopOfFile(sourceFile, newNodes, blankLineBetween);
                };
                ChangeTracker.prototype.insertAtTopOfFile = function (sourceFile, insert, blankLineBetween) {
                    var pos = getInsertionPositionAtSourceFileTop(sourceFile);
                    var options = {
                        prefix: pos === 0 ? undefined : this.newLineCharacter,
                        suffix: (ts.isLineBreak(sourceFile.text.charCodeAt(pos)) ? "" : this.newLineCharacter) + (blankLineBetween ? this.newLineCharacter : ""),
                    };
                    if (ts.isArray(insert)) {
                        this.insertNodesAt(sourceFile, pos, insert, options);
                    }
                    else {
                        this.insertNodeAt(sourceFile, pos, insert, options);
                    }
                };
                ChangeTracker.prototype.insertFirstParameter = function (sourceFile, parameters, newParam) {
                    var p0 = ts.firstOrUndefined(parameters);
                    if (p0) {
                        this.insertNodeBefore(sourceFile, p0, newParam);
                    }
                    else {
                        this.insertNodeAt(sourceFile, parameters.pos, newParam);
                    }
                };
                ChangeTracker.prototype.insertNodeBefore = function (sourceFile, before, newNode, blankLineBetween, options) {
                    if (blankLineBetween === void 0) {
                        blankLineBetween = false;
                    }
                    if (options === void 0) {
                        options = {};
                    }
                    this.insertNodeAt(sourceFile, getAdjustedStartPosition(sourceFile, before, options), newNode, this.getOptionsForInsertNodeBefore(before, newNode, blankLineBetween));
                };
                ChangeTracker.prototype.insertModifierAt = function (sourceFile, pos, modifier, options) {
                    if (options === void 0) {
                        options = {};
                    }
                    this.insertNodeAt(sourceFile, pos, ts.factory.createToken(modifier), options);
                };
                ChangeTracker.prototype.insertModifierBefore = function (sourceFile, modifier, before) {
                    return this.insertModifierAt(sourceFile, before.getStart(sourceFile), modifier, { suffix: " " });
                };
                ChangeTracker.prototype.insertCommentBeforeLine = function (sourceFile, lineNumber, position, commentText) {
                    var lineStartPosition = ts.getStartPositionOfLine(lineNumber, sourceFile);
                    var startPosition = ts.getFirstNonSpaceCharacterPosition(sourceFile.text, lineStartPosition);
                    // First try to see if we can put the comment on the previous line.
                    // We need to make sure that we are not in the middle of a string literal or a comment.
                    // If so, we do not want to separate the node from its comment if we can.
                    // Otherwise, add an extra new line immediately before the error span.
                    var insertAtLineStart = isValidLocationToAddComment(sourceFile, startPosition);
                    var token = ts.getTouchingToken(sourceFile, insertAtLineStart ? startPosition : position);
                    var indent = sourceFile.text.slice(lineStartPosition, startPosition);
                    var text = (insertAtLineStart ? "" : this.newLineCharacter) + "//" + commentText + this.newLineCharacter + indent;
                    this.insertText(sourceFile, token.getStart(sourceFile), text);
                };
                ChangeTracker.prototype.insertJsdocCommentBefore = function (sourceFile, node, tag) {
                    var fnStart = node.getStart(sourceFile);
                    if (node.jsDoc) {
                        for (var _i = 0, _a = node.jsDoc; _i < _a.length; _i++) {
                            var jsdoc = _a[_i];
                            this.deleteRange(sourceFile, {
                                pos: ts.getLineStartPositionForPosition(jsdoc.getStart(sourceFile), sourceFile),
                                end: getAdjustedEndPosition(sourceFile, jsdoc, /*options*/ {})
                            });
                        }
                    }
                    var startPosition = ts.getPrecedingNonSpaceCharacterPosition(sourceFile.text, fnStart - 1);
                    var indent = sourceFile.text.slice(startPosition, fnStart);
                    this.insertNodeAt(sourceFile, fnStart, tag, { preserveLeadingWhitespace: false, suffix: this.newLineCharacter + indent });
                };
                ChangeTracker.prototype.replaceRangeWithText = function (sourceFile, range, text) {
                    this.changes.push({ kind: ChangeKind.Text, sourceFile: sourceFile, range: range, text: text });
                };
                ChangeTracker.prototype.insertText = function (sourceFile, pos, text) {
                    this.replaceRangeWithText(sourceFile, ts.createRange(pos), text);
                };
                /** Prefer this over replacing a node with another that has a type annotation, as it avoids reformatting the other parts of the node. */
                ChangeTracker.prototype.tryInsertTypeAnnotation = function (sourceFile, node, type) {
                    var _a;
                    var endNode;
                    if (ts.isFunctionLike(node)) {
                        endNode = ts.findChildOfKind(node, 21 /* CloseParenToken */, sourceFile);
                        if (!endNode) {
                            if (!ts.isArrowFunction(node))
                                return false; // Function missing parentheses, give up
                            // If no `)`, is an arrow function `x => x`, so use the end of the first parameter
                            endNode = ts.first(node.parameters);
                        }
                    }
                    else {
                        endNode = (_a = (node.kind === 252 /* VariableDeclaration */ ? node.exclamationToken : node.questionToken)) !== null && _a !== void 0 ? _a : node.name;
                    }
                    this.insertNodeAt(sourceFile, endNode.end, type, { prefix: ": " });
                    return true;
                };
                ChangeTracker.prototype.tryInsertThisTypeAnnotation = function (sourceFile, node, type) {
                    var start = ts.findChildOfKind(node, 20 /* OpenParenToken */, sourceFile).getStart(sourceFile) + 1;
                    var suffix = node.parameters.length ? ", " : "";
                    this.insertNodeAt(sourceFile, start, type, { prefix: "this: ", suffix: suffix });
                };
                ChangeTracker.prototype.insertTypeParameters = function (sourceFile, node, typeParameters) {
                    // If no `(`, is an arrow function `x => x`, so use the pos of the first parameter
                    var start = (ts.findChildOfKind(node, 20 /* OpenParenToken */, sourceFile) || ts.first(node.parameters)).getStart(sourceFile);
                    this.insertNodesAt(sourceFile, start, typeParameters, { prefix: "<", suffix: ">", joiner: ", " });
                };
                ChangeTracker.prototype.getOptionsForInsertNodeBefore = function (before, inserted, blankLineBetween) {
                    if (ts.isStatement(before) || ts.isClassElement(before)) {
                        return { suffix: blankLineBetween ? this.newLineCharacter + this.newLineCharacter : this.newLineCharacter };
                    }
                    else if (ts.isVariableDeclaration(before)) { // insert `x = 1, ` into `const x = 1, y = 2;
                        return { suffix: ", " };
                    }
                    else if (ts.isParameter(before)) {
                        return ts.isParameter(inserted) ? { suffix: ", " } : {};
                    }
                    else if (ts.isStringLiteral(before) && ts.isImportDeclaration(before.parent) || ts.isNamedImports(before)) {
                        return { suffix: ", " };
                    }
                    else if (ts.isImportSpecifier(before)) {
                        return { suffix: "," + (blankLineBetween ? this.newLineCharacter : " ") };
                    }
                    return ts.Debug.failBadSyntaxKind(before); // We haven't handled this kind of node yet -- add it
                };
                ChangeTracker.prototype.insertNodeAtConstructorStart = function (sourceFile, ctr, newStatement) {
                    var firstStatement = ts.firstOrUndefined(ctr.body.statements);
                    if (!firstStatement || !ctr.body.multiLine) {
                        this.replaceConstructorBody(sourceFile, ctr, __spreadArray([newStatement], ctr.body.statements, true));
                    }
                    else {
                        this.insertNodeBefore(sourceFile, firstStatement, newStatement);
                    }
                };
                ChangeTracker.prototype.insertNodeAtConstructorStartAfterSuperCall = function (sourceFile, ctr, newStatement) {
                    var superCallStatement = ts.find(ctr.body.statements, function (stmt) { return ts.isExpressionStatement(stmt) && ts.isSuperCall(stmt.expression); });
                    if (!superCallStatement || !ctr.body.multiLine) {
                        this.replaceConstructorBody(sourceFile, ctr, __spreadArray(__spreadArray([], ctr.body.statements, true), [newStatement], false));
                    }
                    else {
                        this.insertNodeAfter(sourceFile, superCallStatement, newStatement);
                    }
                };
                ChangeTracker.prototype.insertNodeAtConstructorEnd = function (sourceFile, ctr, newStatement) {
                    var lastStatement = ts.lastOrUndefined(ctr.body.statements);
                    if (!lastStatement || !ctr.body.multiLine) {
                        this.replaceConstructorBody(sourceFile, ctr, __spreadArray(__spreadArray([], ctr.body.statements, true), [newStatement], false));
                    }
                    else {
                        this.insertNodeAfter(sourceFile, lastStatement, newStatement);
                    }
                };
                ChangeTracker.prototype.replaceConstructorBody = function (sourceFile, ctr, statements) {
                    this.replaceNode(sourceFile, ctr.body, ts.factory.createBlock(statements, /*multiLine*/ true));
                };
                ChangeTracker.prototype.insertNodeAtEndOfScope = function (sourceFile, scope, newNode) {
                    var pos = getAdjustedStartPosition(sourceFile, scope.getLastToken(), {});
                    this.insertNodeAt(sourceFile, pos, newNode, {
                        prefix: ts.isLineBreak(sourceFile.text.charCodeAt(scope.getLastToken().pos)) ? this.newLineCharacter : this.newLineCharacter + this.newLineCharacter,
                        suffix: this.newLineCharacter
                    });
                };
                ChangeTracker.prototype.insertNodeAtClassStart = function (sourceFile, cls, newElement) {
                    this.insertNodeAtStartWorker(sourceFile, cls, newElement);
                };
                ChangeTracker.prototype.insertNodeAtObjectStart = function (sourceFile, obj, newElement) {
                    this.insertNodeAtStartWorker(sourceFile, obj, newElement);
                };
                ChangeTracker.prototype.insertNodeAtStartWorker = function (sourceFile, cls, newElement) {
                    var _a;
                    var indentation = (_a = this.guessIndentationFromExistingMembers(sourceFile, cls)) !== null && _a !== void 0 ? _a : this.computeIndentationForNewMember(sourceFile, cls);
                    this.insertNodeAt(sourceFile, getMembersOrProperties(cls).pos, newElement, this.getInsertNodeAtStartInsertOptions(sourceFile, cls, indentation));
                };
                /**
                 * Tries to guess the indentation from the existing members of a class/interface/object. All members must be on
                 * new lines and must share the same indentation.
                 */
                ChangeTracker.prototype.guessIndentationFromExistingMembers = function (sourceFile, cls) {
                    var indentation;
                    var lastRange = cls;
                    for (var _i = 0, _a = getMembersOrProperties(cls); _i < _a.length; _i++) {
                        var member = _a[_i];
                        if (ts.rangeStartPositionsAreOnSameLine(lastRange, member, sourceFile)) {
                            // each indented member must be on a new line
                            return undefined;
                        }
                        var memberStart = member.getStart(sourceFile);
                        var memberIndentation = ts.formatting.SmartIndenter.findFirstNonWhitespaceColumn(ts.getLineStartPositionForPosition(memberStart, sourceFile), memberStart, sourceFile, this.formatContext.options);
                        if (indentation === undefined) {
                            indentation = memberIndentation;
                        }
                        else if (memberIndentation !== indentation) {
                            // indentation of multiple members is not consistent
                            return undefined;
                        }
                        lastRange = member;
                    }
                    return indentation;
                };
                ChangeTracker.prototype.computeIndentationForNewMember = function (sourceFile, cls) {
                    var _a;
                    var clsStart = cls.getStart(sourceFile);
                    return ts.formatting.SmartIndenter.findFirstNonWhitespaceColumn(ts.getLineStartPositionForPosition(clsStart, sourceFile), clsStart, sourceFile, this.formatContext.options)
                        + ((_a = this.formatContext.options.indentSize) !== null && _a !== void 0 ? _a : 4);
                };
                ChangeTracker.prototype.getInsertNodeAtStartInsertOptions = function (sourceFile, cls, indentation) {
                    // Rules:
                    // - Always insert leading newline.
                    // - For object literals:
                    //   - Add a trailing comma if there are existing members in the node, or the source file is not a JSON file
                    //     (because trailing commas are generally illegal in a JSON file).
                    //   - Add a leading comma if the source file is not a JSON file, there are existing insertions,
                    //     and the node is empty (because we didn't add a trailing comma per the previous rule).
                    // - Only insert a trailing newline if body is single-line and there are no other insertions for the node.
                    //   NOTE: This is handled in `finishClassesWithNodesInsertedAtStart`.
                    var members = getMembersOrProperties(cls);
                    var isEmpty = members.length === 0;
                    var isFirstInsertion = ts.addToSeen(this.classesWithNodesInsertedAtStart, ts.getNodeId(cls), { node: cls, sourceFile: sourceFile });
                    var insertTrailingComma = ts.isObjectLiteralExpression(cls) && (!ts.isJsonSourceFile(sourceFile) || !isEmpty);
                    var insertLeadingComma = ts.isObjectLiteralExpression(cls) && ts.isJsonSourceFile(sourceFile) && isEmpty && !isFirstInsertion;
                    return {
                        indentation: indentation,
                        prefix: (insertLeadingComma ? "," : "") + this.newLineCharacter,
                        suffix: insertTrailingComma ? "," : ""
                    };
                };
                ChangeTracker.prototype.insertNodeAfterComma = function (sourceFile, after, newNode) {
                    var endPosition = this.insertNodeAfterWorker(sourceFile, this.nextCommaToken(sourceFile, after) || after, newNode);
                    this.insertNodeAt(sourceFile, endPosition, newNode, this.getInsertNodeAfterOptions(sourceFile, after));
                };
                ChangeTracker.prototype.insertNodeAfter = function (sourceFile, after, newNode) {
                    var endPosition = this.insertNodeAfterWorker(sourceFile, after, newNode);
                    this.insertNodeAt(sourceFile, endPosition, newNode, this.getInsertNodeAfterOptions(sourceFile, after));
                };
                ChangeTracker.prototype.insertNodeAtEndOfList = function (sourceFile, list, newNode) {
                    this.insertNodeAt(sourceFile, list.end, newNode, { prefix: ", " });
                };
                ChangeTracker.prototype.insertNodesAfter = function (sourceFile, after, newNodes) {
                    var endPosition = this.insertNodeAfterWorker(sourceFile, after, ts.first(newNodes));
                    this.insertNodesAt(sourceFile, endPosition, newNodes, this.getInsertNodeAfterOptions(sourceFile, after));
                };
                ChangeTracker.prototype.insertNodeAfterWorker = function (sourceFile, after, newNode) {
                    if (needSemicolonBetween(after, newNode)) {
                        // check if previous statement ends with semicolon
                        // if not - insert semicolon to preserve the code from changing the meaning due to ASI
                        if (sourceFile.text.charCodeAt(after.end - 1) !== 59 /* semicolon */) {
                            this.replaceRange(sourceFile, ts.createRange(after.end), ts.factory.createToken(26 /* SemicolonToken */));
                        }
                    }
                    var endPosition = getAdjustedEndPosition(sourceFile, after, {});
                    return endPosition;
                };
                ChangeTracker.prototype.getInsertNodeAfterOptions = function (sourceFile, after) {
                    var options = this.getInsertNodeAfterOptionsWorker(after);
                    return __assign(__assign({}, options), { prefix: after.end === sourceFile.end && ts.isStatement(after) ? (options.prefix ? "\n" + options.prefix : "\n") : options.prefix });
                };
                ChangeTracker.prototype.getInsertNodeAfterOptionsWorker = function (node) {
                    switch (node.kind) {
                        case 255 /* ClassDeclaration */:
                        case 259 /* ModuleDeclaration */:
                            return { prefix: this.newLineCharacter, suffix: this.newLineCharacter };
                        case 252 /* VariableDeclaration */:
                        case 10 /* StringLiteral */:
                        case 79 /* Identifier */:
                            return { prefix: ", " };
                        case 291 /* PropertyAssignment */:
                            return { suffix: "," + this.newLineCharacter };
                        case 93 /* ExportKeyword */:
                            return { prefix: " " };
                        case 162 /* Parameter */:
                            return {};
                        default:
                            ts.Debug.assert(ts.isStatement(node) || ts.isClassOrTypeElement(node)); // Else we haven't handled this kind of node yet -- add it
                            return { suffix: this.newLineCharacter };
                    }
                };
                ChangeTracker.prototype.insertName = function (sourceFile, node, name) {
                    ts.Debug.assert(!node.name);
                    if (node.kind === 212 /* ArrowFunction */) {
                        var arrow = ts.findChildOfKind(node, 38 /* EqualsGreaterThanToken */, sourceFile);
                        var lparen = ts.findChildOfKind(node, 20 /* OpenParenToken */, sourceFile);
                        if (lparen) {
                            // `() => {}` --> `function f() {}`
                            this.insertNodesAt(sourceFile, lparen.getStart(sourceFile), [ts.factory.createToken(98 /* FunctionKeyword */), ts.factory.createIdentifier(name)], { joiner: " " });
                            deleteNode(this, sourceFile, arrow);
                        }
                        else {
                            // `x => {}` -> `function f(x) {}`
                            this.insertText(sourceFile, ts.first(node.parameters).getStart(sourceFile), "function " + name + "(");
                            // Replacing full range of arrow to get rid of the leading space -- replace ` =>` with `)`
                            this.replaceRange(sourceFile, arrow, ts.factory.createToken(21 /* CloseParenToken */));
                        }
                        if (node.body.kind !== 233 /* Block */) {
                            // `() => 0` => `function f() { return 0; }`
                            this.insertNodesAt(sourceFile, node.body.getStart(sourceFile), [ts.factory.createToken(18 /* OpenBraceToken */), ts.factory.createToken(105 /* ReturnKeyword */)], { joiner: " ", suffix: " " });
                            this.insertNodesAt(sourceFile, node.body.end, [ts.factory.createToken(26 /* SemicolonToken */), ts.factory.createToken(19 /* CloseBraceToken */)], { joiner: " " });
                        }
                    }
                    else {
                        var pos = ts.findChildOfKind(node, node.kind === 211 /* FunctionExpression */ ? 98 /* FunctionKeyword */ : 84 /* ClassKeyword */, sourceFile).end;
                        this.insertNodeAt(sourceFile, pos, ts.factory.createIdentifier(name), { prefix: " " });
                    }
                };
                ChangeTracker.prototype.insertExportModifier = function (sourceFile, node) {
                    this.insertText(sourceFile, node.getStart(sourceFile), "export ");
                };
                /**
                 * This function should be used to insert nodes in lists when nodes don't carry separators as the part of the node range,
                 * i.e. arguments in arguments lists, parameters in parameter lists etc.
                 * Note that separators are part of the node in statements and class elements.
                 */
                ChangeTracker.prototype.insertNodeInListAfter = function (sourceFile, after, newNode, containingList) {
                    if (containingList === void 0) {
                        containingList = ts.formatting.SmartIndenter.getContainingList(after, sourceFile);
                    }
                    if (!containingList) {
                        ts.Debug.fail("node is not a list element");
                        return;
                    }
                    var index = ts.indexOfNode(containingList, after);
                    if (index < 0) {
                        return;
                    }
                    var end = after.getEnd();
                    if (index !== containingList.length - 1) {
                        // any element except the last one
                        // use next sibling as an anchor
                        var nextToken = ts.getTokenAtPosition(sourceFile, after.end);
                        if (nextToken && isSeparator(after, nextToken)) {
                            // for list
                            // a, b, c
                            // create change for adding 'e' after 'a' as
                            // - find start of next element after a (it is b)
                            // - use next element start as start and end position in final change
                            // - build text of change by formatting the text of node + whitespace trivia of b
                            // in multiline case it will work as
                            //   a,
                            //   b,
                            //   c,
                            // result - '*' denotes leading trivia that will be inserted after new text (displayed as '#')
                            //   a,
                            //   insertedtext<separator>#
                            // ###b,
                            //   c,
                            var nextNode = containingList[index + 1];
                            var startPos = skipWhitespacesAndLineBreaks(sourceFile.text, nextNode.getFullStart());
                            // write separator and leading trivia of the next element as suffix
                            var suffix = "" + ts.tokenToString(nextToken.kind) + sourceFile.text.substring(nextToken.end, startPos);
                            this.insertNodesAt(sourceFile, startPos, [newNode], { suffix: suffix });
                        }
                    }
                    else {
                        var afterStart = after.getStart(sourceFile);
                        var afterStartLinePosition = ts.getLineStartPositionForPosition(afterStart, sourceFile);
                        var separator = void 0;
                        var multilineList = false;
                        // insert element after the last element in the list that has more than one item
                        // pick the element preceding the after element to:
                        // - pick the separator
                        // - determine if list is a multiline
                        if (containingList.length === 1) {
                            // if list has only one element then we'll format is as multiline if node has comment in trailing trivia, or as singleline otherwise
                            // i.e. var x = 1 // this is x
                            //     | new element will be inserted at this position
                            separator = 27 /* CommaToken */;
                        }
                        else {
                            // element has more than one element, pick separator from the list
                            var tokenBeforeInsertPosition = ts.findPrecedingToken(after.pos, sourceFile);
                            separator = isSeparator(after, tokenBeforeInsertPosition) ? tokenBeforeInsertPosition.kind : 27 /* CommaToken */;
                            // determine if list is multiline by checking lines of after element and element that precedes it.
                            var afterMinusOneStartLinePosition = ts.getLineStartPositionForPosition(containingList[index - 1].getStart(sourceFile), sourceFile);
                            multilineList = afterMinusOneStartLinePosition !== afterStartLinePosition;
                        }
                        if (hasCommentsBeforeLineBreak(sourceFile.text, after.end)) {
                            // in this case we'll always treat containing list as multiline
                            multilineList = true;
                        }
                        if (multilineList) {
                            // insert separator immediately following the 'after' node to preserve comments in trailing trivia
                            this.replaceRange(sourceFile, ts.createRange(end), ts.factory.createToken(separator));
                            // use the same indentation as 'after' item
                            var indentation = ts.formatting.SmartIndenter.findFirstNonWhitespaceColumn(afterStartLinePosition, afterStart, sourceFile, this.formatContext.options);
                            // insert element before the line break on the line that contains 'after' element
                            var insertPos = ts.skipTrivia(sourceFile.text, end, /*stopAfterLineBreak*/ true, /*stopAtComments*/ false);
                            // find position before "\n" or "\r\n"
                            while (insertPos !== end && ts.isLineBreak(sourceFile.text.charCodeAt(insertPos - 1))) {
                                insertPos--;
                            }
                            this.replaceRange(sourceFile, ts.createRange(insertPos), newNode, { indentation: indentation, prefix: this.newLineCharacter });
                        }
                        else {
                            this.replaceRange(sourceFile, ts.createRange(end), newNode, { prefix: ts.tokenToString(separator) + " " });
                        }
                    }
                };
                ChangeTracker.prototype.parenthesizeExpression = function (sourceFile, expression) {
                    this.replaceRange(sourceFile, ts.rangeOfNode(expression), ts.factory.createParenthesizedExpression(expression));
                };
                ChangeTracker.prototype.finishClassesWithNodesInsertedAtStart = function () {
                    var _this = this;
                    this.classesWithNodesInsertedAtStart.forEach(function (_a) {
                        var node = _a.node, sourceFile = _a.sourceFile;
                        var _b = getClassOrObjectBraceEnds(node, sourceFile), openBraceEnd = _b[0], closeBraceEnd = _b[1];
                        if (openBraceEnd !== undefined && closeBraceEnd !== undefined) {
                            var isEmpty = getMembersOrProperties(node).length === 0;
                            var isSingleLine = ts.positionsAreOnSameLine(openBraceEnd, closeBraceEnd, sourceFile);
                            if (isEmpty && isSingleLine && openBraceEnd !== closeBraceEnd - 1) {
                                // For `class C { }` remove the whitespace inside the braces.
                                _this.deleteRange(sourceFile, ts.createRange(openBraceEnd, closeBraceEnd - 1));
                            }
                            if (isSingleLine) {
                                _this.insertText(sourceFile, closeBraceEnd - 1, _this.newLineCharacter);
                            }
                        }
                    });
                };
                ChangeTracker.prototype.finishDeleteDeclarations = function () {
                    var _this = this;
                    var deletedNodesInLists = new ts.Set(); // Stores nodes in lists that we already deleted. Used to avoid deleting `, ` twice in `a, b`.
                    var _loop_10 = function (sourceFile, node) {
                        if (!this_1.deletedNodes.some(function (d) { return d.sourceFile === sourceFile && ts.rangeContainsRangeExclusive(d.node, node); })) {
                            if (ts.isArray(node)) {
                                this_1.deleteRange(sourceFile, ts.rangeOfTypeParameters(sourceFile, node));
                            }
                            else {
                                deleteDeclaration.deleteDeclaration(this_1, deletedNodesInLists, sourceFile, node);
                            }
                        }
                    };
                    var this_1 = this;
                    for (var _i = 0, _a = this.deletedNodes; _i < _a.length; _i++) {
                        var _b = _a[_i], sourceFile = _b.sourceFile, node = _b.node;
                        _loop_10(sourceFile, node);
                    }
                    deletedNodesInLists.forEach(function (node) {
                        var sourceFile = node.getSourceFile();
                        var list = ts.formatting.SmartIndenter.getContainingList(node, sourceFile);
                        if (node !== ts.last(list))
                            return;
                        var lastNonDeletedIndex = ts.findLastIndex(list, function (n) { return !deletedNodesInLists.has(n); }, list.length - 2);
                        if (lastNonDeletedIndex !== -1) {
                            _this.deleteRange(sourceFile, { pos: list[lastNonDeletedIndex].end, end: startPositionToDeleteNodeInList(sourceFile, list[lastNonDeletedIndex + 1]) });
                        }
                    });
                };
                /**
                 * Note: after calling this, the TextChanges object must be discarded!
                 * @param validate only for tests
                 *    The reason we must validate as part of this method is that `getNonFormattedText` changes the node's positions,
                 *    so we can only call this once and can't get the non-formatted text separately.
                 */
                ChangeTracker.prototype.getChanges = function (validate) {
                    this.finishDeleteDeclarations();
                    this.finishClassesWithNodesInsertedAtStart();
                    var changes = changesToText.getTextChangesFromChanges(this.changes, this.newLineCharacter, this.formatContext, validate);
                    for (var _i = 0, _a = this.newFiles; _i < _a.length; _i++) {
                        var _b = _a[_i], oldFile = _b.oldFile, fileName = _b.fileName, statements = _b.statements;
                        changes.push(changesToText.newFileChanges(oldFile, fileName, statements, this.newLineCharacter, this.formatContext));
                    }
                    return changes;
                };
                ChangeTracker.prototype.createNewFile = function (oldFile, fileName, statements) {
                    this.newFiles.push({ oldFile: oldFile, fileName: fileName, statements: statements });
                };
                return ChangeTracker;
            }());
            textChanges_3.ChangeTracker = ChangeTracker;
            // find first non-whitespace position in the leading trivia of the node
            function startPositionToDeleteNodeInList(sourceFile, node) {
                return ts.skipTrivia(sourceFile.text, getAdjustedStartPosition(sourceFile, node, { leadingTriviaOption: LeadingTriviaOption.IncludeAll }), /*stopAfterLineBreak*/ false, /*stopAtComments*/ true);
            }
            function getClassOrObjectBraceEnds(cls, sourceFile) {
                var open = ts.findChildOfKind(cls, 18 /* OpenBraceToken */, sourceFile);
                var close = ts.findChildOfKind(cls, 19 /* CloseBraceToken */, sourceFile);
                return [open === null || open === void 0 ? void 0 : open.end, close === null || close === void 0 ? void 0 : close.end];
            }
            function getMembersOrProperties(cls) {
                return ts.isObjectLiteralExpression(cls) ? cls.properties : cls.members;
            }
            function getNewFileText(statements, scriptKind, newLineCharacter, formatContext) {
                return changesToText.newFileChangesWorker(/*oldFile*/ undefined, scriptKind, statements, newLineCharacter, formatContext);
            }
            textChanges_3.getNewFileText = getNewFileText;
            var changesToText;
            (function (changesToText) {
                function getTextChangesFromChanges(changes, newLineCharacter, formatContext, validate) {
                    return ts.mapDefined(ts.group(changes, function (c) { return c.sourceFile.path; }), function (changesInFile) {
                        var sourceFile = changesInFile[0].sourceFile;
                        // order changes by start position
                        // If the start position is the same, put the shorter range first, since an empty range (x, x) may precede (x, y) but not vice-versa.
                        var normalized = ts.stableSort(changesInFile, function (a, b) { return (a.range.pos - b.range.pos) || (a.range.end - b.range.end); });
                        var _loop_11 = function (i) {
                            ts.Debug.assert(normalized[i].range.end <= normalized[i + 1].range.pos, "Changes overlap", function () {
                                return JSON.stringify(normalized[i].range) + " and " + JSON.stringify(normalized[i + 1].range);
                            });
                        };
                        // verify that change intervals do not overlap, except possibly at end points.
                        for (var i = 0; i < normalized.length - 1; i++) {
                            _loop_11(i);
                        }
                        var textChanges = ts.mapDefined(normalized, function (c) {
                            var span = ts.createTextSpanFromRange(c.range);
                            var newText = computeNewText(c, sourceFile, newLineCharacter, formatContext, validate);
                            // Filter out redundant changes.
                            if (span.length === newText.length && ts.stringContainsAt(sourceFile.text, newText, span.start)) {
                                return undefined;
                            }
                            return ts.createTextChange(span, newText);
                        });
                        return textChanges.length > 0 ? { fileName: sourceFile.fileName, textChanges: textChanges } : undefined;
                    });
                }
                changesToText.getTextChangesFromChanges = getTextChangesFromChanges;
                function newFileChanges(oldFile, fileName, statements, newLineCharacter, formatContext) {
                    var text = newFileChangesWorker(oldFile, ts.getScriptKindFromFileName(fileName), statements, newLineCharacter, formatContext);
                    return { fileName: fileName, textChanges: [ts.createTextChange(ts.createTextSpan(0, 0), text)], isNewFile: true };
                }
                changesToText.newFileChanges = newFileChanges;
                function newFileChangesWorker(oldFile, scriptKind, statements, newLineCharacter, formatContext) {
                    // TODO: this emits the file, parses it back, then formats it that -- may be a less roundabout way to do this
                    var nonFormattedText = statements.map(function (s) { return s === 4 /* NewLineTrivia */ ? "" : getNonformattedText(s, oldFile, newLineCharacter).text; }).join(newLineCharacter);
                    var sourceFile = ts.createSourceFile("any file name", nonFormattedText, 99 /* ESNext */, /*setParentNodes*/ true, scriptKind);
                    var changes = ts.formatting.formatDocument(sourceFile, formatContext);
                    return applyChanges(nonFormattedText, changes) + newLineCharacter;
                }
                changesToText.newFileChangesWorker = newFileChangesWorker;
                function computeNewText(change, sourceFile, newLineCharacter, formatContext, validate) {
                    var _a;
                    if (change.kind === ChangeKind.Remove) {
                        return "";
                    }
                    if (change.kind === ChangeKind.Text) {
                        return change.text;
                    }
                    var _b = change.options, options = _b === void 0 ? {} : _b, pos = change.range.pos;
                    var format = function (n) { return getFormattedTextOfNode(n, sourceFile, pos, options, newLineCharacter, formatContext, validate); };
                    var text = change.kind === ChangeKind.ReplaceWithMultipleNodes
                        ? change.nodes.map(function (n) { return ts.removeSuffix(format(n), newLineCharacter); }).join(((_a = change.options) === null || _a === void 0 ? void 0 : _a.joiner) || newLineCharacter)
                        : format(change.node);
                    // strip initial indentation (spaces or tabs) if text will be inserted in the middle of the line
                    var noIndent = (options.preserveLeadingWhitespace || options.indentation !== undefined || ts.getLineStartPositionForPosition(pos, sourceFile) === pos) ? text : text.replace(/^\s+/, "");
                    return (options.prefix || "") + noIndent
                        + ((!options.suffix || ts.endsWith(noIndent, options.suffix))
                            ? "" : options.suffix);
                }
                function getFormatCodeSettingsForWriting(_a, sourceFile) {
                    var options = _a.options;
                    var shouldAutoDetectSemicolonPreference = !options.semicolons || options.semicolons === ts.SemicolonPreference.Ignore;
                    var shouldRemoveSemicolons = options.semicolons === ts.SemicolonPreference.Remove || shouldAutoDetectSemicolonPreference && !ts.probablyUsesSemicolons(sourceFile);
                    return __assign(__assign({}, options), { semicolons: shouldRemoveSemicolons ? ts.SemicolonPreference.Remove : ts.SemicolonPreference.Ignore });
                }
                /** Note: this may mutate `nodeIn`. */
                function getFormattedTextOfNode(nodeIn, sourceFile, pos, _a, newLineCharacter, formatContext, validate) {
                    var indentation = _a.indentation, prefix = _a.prefix, delta = _a.delta;
                    var _b = getNonformattedText(nodeIn, sourceFile, newLineCharacter), node = _b.node, text = _b.text;
                    if (validate)
                        validate(node, text);
                    var formatOptions = getFormatCodeSettingsForWriting(formatContext, sourceFile);
                    var initialIndentation = indentation !== undefined
                        ? indentation
                        : ts.formatting.SmartIndenter.getIndentation(pos, sourceFile, formatOptions, prefix === newLineCharacter || ts.getLineStartPositionForPosition(pos, sourceFile) === pos);
                    if (delta === undefined) {
                        delta = ts.formatting.SmartIndenter.shouldIndentChildNode(formatOptions, nodeIn) ? (formatOptions.indentSize || 0) : 0;
                    }
                    var file = { text: text, getLineAndCharacterOfPosition: function (pos) { return ts.getLineAndCharacterOfPosition(this, pos); } };
                    var changes = ts.formatting.formatNodeGivenIndentation(node, file, sourceFile.languageVariant, initialIndentation, delta, __assign(__assign({}, formatContext), { options: formatOptions }));
                    return applyChanges(text, changes);
                }
                /** Note: output node may be mutated input node. */
                function getNonformattedText(node, sourceFile, newLineCharacter) {
                    var writer = createWriter(newLineCharacter);
                    var newLine = newLineCharacter === "\n" ? 1 /* LineFeed */ : 0 /* CarriageReturnLineFeed */;
                    ts.createPrinter({
                        newLine: newLine,
                        neverAsciiEscape: true,
                        preserveSourceNewlines: true,
                        terminateUnterminatedLiterals: true
                    }, writer).writeNode(4 /* Unspecified */, node, sourceFile, writer);
                    return { text: writer.getText(), node: assignPositionsToNode(node) };
                }
                changesToText.getNonformattedText = getNonformattedText;
            })(changesToText || (changesToText = {}));
            function applyChanges(text, changes) {
                for (var i = changes.length - 1; i >= 0; i--) {
                    var _a = changes[i], span = _a.span, newText = _a.newText;
                    text = "" + text.substring(0, span.start) + newText + text.substring(ts.textSpanEnd(span));
                }
                return text;
            }
            textChanges_3.applyChanges = applyChanges;
            function isTrivia(s) {
                return ts.skipTrivia(s, 0) === s.length;
            }
            function assignPositionsToNode(node) {
                var visited = ts.visitEachChild(node, assignPositionsToNode, ts.nullTransformationContext, assignPositionsToNodeArray, assignPositionsToNode);
                // create proxy node for non synthesized nodes
                var newNode = ts.nodeIsSynthesized(visited) ? visited : Object.create(visited);
                ts.setTextRangePosEnd(newNode, getPos(node), getEnd(node));
                return newNode;
            }
            function assignPositionsToNodeArray(nodes, visitor, test, start, count) {
                var visited = ts.visitNodes(nodes, visitor, test, start, count);
                if (!visited) {
                    return visited;
                }
                // clone nodearray if necessary
                var nodeArray = visited === nodes ? ts.factory.createNodeArray(visited.slice(0)) : visited;
                ts.setTextRangePosEnd(nodeArray, getPos(nodes), getEnd(nodes));
                return nodeArray;
            }
            function createWriter(newLine) {
                var lastNonTriviaPosition = 0;
                var writer = ts.createTextWriter(newLine);
                var onBeforeEmitNode = function (node) {
                    if (node) {
                        setPos(node, lastNonTriviaPosition);
                    }
                };
                var onAfterEmitNode = function (node) {
                    if (node) {
                        setEnd(node, lastNonTriviaPosition);
                    }
                };
                var onBeforeEmitNodeArray = function (nodes) {
                    if (nodes) {
                        setPos(nodes, lastNonTriviaPosition);
                    }
                };
                var onAfterEmitNodeArray = function (nodes) {
                    if (nodes) {
                        setEnd(nodes, lastNonTriviaPosition);
                    }
                };
                var onBeforeEmitToken = function (node) {
                    if (node) {
                        setPos(node, lastNonTriviaPosition);
                    }
                };
                var onAfterEmitToken = function (node) {
                    if (node) {
                        setEnd(node, lastNonTriviaPosition);
                    }
                };
                function setLastNonTriviaPosition(s, force) {
                    if (force || !isTrivia(s)) {
                        lastNonTriviaPosition = writer.getTextPos();
                        var i = 0;
                        while (ts.isWhiteSpaceLike(s.charCodeAt(s.length - i - 1))) {
                            i++;
                        }
                        // trim trailing whitespaces
                        lastNonTriviaPosition -= i;
                    }
                }
                function write(s) {
                    writer.write(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeComment(s) {
                    writer.writeComment(s);
                }
                function writeKeyword(s) {
                    writer.writeKeyword(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeOperator(s) {
                    writer.writeOperator(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writePunctuation(s) {
                    writer.writePunctuation(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeTrailingSemicolon(s) {
                    writer.writeTrailingSemicolon(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeParameter(s) {
                    writer.writeParameter(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeProperty(s) {
                    writer.writeProperty(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeSpace(s) {
                    writer.writeSpace(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeStringLiteral(s) {
                    writer.writeStringLiteral(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeSymbol(s, sym) {
                    writer.writeSymbol(s, sym);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeLine(force) {
                    writer.writeLine(force);
                }
                function increaseIndent() {
                    writer.increaseIndent();
                }
                function decreaseIndent() {
                    writer.decreaseIndent();
                }
                function getText() {
                    return writer.getText();
                }
                function rawWrite(s) {
                    writer.rawWrite(s);
                    setLastNonTriviaPosition(s, /*force*/ false);
                }
                function writeLiteral(s) {
                    writer.writeLiteral(s);
                    setLastNonTriviaPosition(s, /*force*/ true);
                }
                function getTextPos() {
                    return writer.getTextPos();
                }
                function getLine() {
                    return writer.getLine();
                }
                function getColumn() {
                    return writer.getColumn();
                }
                function getIndent() {
                    return writer.getIndent();
                }
                function isAtStartOfLine() {
                    return writer.isAtStartOfLine();
                }
                function clear() {
                    writer.clear();
                    lastNonTriviaPosition = 0;
                }
                return {
                    onBeforeEmitNode: onBeforeEmitNode,
                    onAfterEmitNode: onAfterEmitNode,
                    onBeforeEmitNodeArray: onBeforeEmitNodeArray,
                    onAfterEmitNodeArray: onAfterEmitNodeArray,
                    onBeforeEmitToken: onBeforeEmitToken,
                    onAfterEmitToken: onAfterEmitToken,
                    write: write,
                    writeComment: writeComment,
                    writeKeyword: writeKeyword,
                    writeOperator: writeOperator,
                    writePunctuation: writePunctuation,
                    writeTrailingSemicolon: writeTrailingSemicolon,
                    writeParameter: writeParameter,
                    writeProperty: writeProperty,
                    writeSpace: writeSpace,
                    writeStringLiteral: writeStringLiteral,
                    writeSymbol: writeSymbol,
                    writeLine: writeLine,
                    increaseIndent: increaseIndent,
                    decreaseIndent: decreaseIndent,
                    getText: getText,
                    rawWrite: rawWrite,
                    writeLiteral: writeLiteral,
                    getTextPos: getTextPos,
                    getLine: getLine,
                    getColumn: getColumn,
                    getIndent: getIndent,
                    isAtStartOfLine: isAtStartOfLine,
                    hasTrailingComment: function () { return writer.hasTrailingComment(); },
                    hasTrailingWhitespace: function () { return writer.hasTrailingWhitespace(); },
                    clear: clear
                };
            }
            function getInsertionPositionAtSourceFileTop(sourceFile) {
                var lastPrologue;
                for (var _i = 0, _a = sourceFile.statements; _i < _a.length; _i++) {
                    var node = _a[_i];
                    if (ts.isPrologueDirective(node)) {
                        lastPrologue = node;
                    }
                    else {
                        break;
                    }
                }
                var position = 0;
                var text = sourceFile.text;
                if (lastPrologue) {
                    position = lastPrologue.end;
                    advancePastLineBreak();
                    return position;
                }
                var shebang = ts.getShebang(text);
                if (shebang !== undefined) {
                    position = shebang.length;
                    advancePastLineBreak();
                }
                var ranges = ts.getLeadingCommentRanges(text, position);
                if (!ranges)
                    return position;
                // Find the first attached comment to the first node and add before it
                var lastComment;
                var firstNodeLine;
                for (var _b = 0, ranges_1 = ranges; _b < ranges_1.length; _b++) {
                    var range = ranges_1[_b];
                    if (range.kind === 3 /* MultiLineCommentTrivia */) {
                        if (ts.isPinnedComment(text, range.pos)) {
                            lastComment = { range: range, pinnedOrTripleSlash: true };
                            continue;
                        }
                    }
                    else if (ts.isRecognizedTripleSlashComment(text, range.pos, range.end)) {
                        lastComment = { range: range, pinnedOrTripleSlash: true };
                        continue;
                    }
                    if (lastComment) {
                        // Always insert after pinned or triple slash comments
                        if (lastComment.pinnedOrTripleSlash)
                            break;
                        // There was a blank line between the last comment and this comment.
                        // This comment is not part of the copyright comments
                        var commentLine = sourceFile.getLineAndCharacterOfPosition(range.pos).line;
                        var lastCommentEndLine = sourceFile.getLineAndCharacterOfPosition(lastComment.range.end).line;
                        if (commentLine >= lastCommentEndLine + 2)
                            break;
                    }
                    if (sourceFile.statements.length) {
                        if (firstNodeLine === undefined)
                            firstNodeLine = sourceFile.getLineAndCharacterOfPosition(sourceFile.statements[0].getStart()).line;
                        var commentEndLine = sourceFile.getLineAndCharacterOfPosition(range.end).line;
                        if (firstNodeLine < commentEndLine + 2)
                            break;
                    }
                    lastComment = { range: range, pinnedOrTripleSlash: false };
                }
                if (lastComment) {
                    position = lastComment.range.end;
                    advancePastLineBreak();
                }
                return position;
                function advancePastLineBreak() {
                    if (position < text.length) {
                        var charCode = text.charCodeAt(position);
                        if (ts.isLineBreak(charCode)) {
                            position++;
                            if (position < text.length && charCode === 13 /* carriageReturn */ && text.charCodeAt(position) === 10 /* lineFeed */) {
                                position++;
                            }
                        }
                    }
                }
            }
            function isValidLocationToAddComment(sourceFile, position) {
                return !ts.isInComment(sourceFile, position) && !ts.isInString(sourceFile, position) && !ts.isInTemplateString(sourceFile, position) && !ts.isInJSXText(sourceFile, position);
            }
            textChanges_3.isValidLocationToAddComment = isValidLocationToAddComment;
            function needSemicolonBetween(a, b) {
                return (ts.isPropertySignature(a) || ts.isPropertyDeclaration(a)) && ts.isClassOrTypeElement(b) && b.name.kind === 160 /* ComputedPropertyName */
                    || ts.isStatementButNotDeclaration(a) && ts.isStatementButNotDeclaration(b); // TODO: only if b would start with a `(` or `[`
            }
            var deleteDeclaration;
            (function (deleteDeclaration_1) {
                function deleteDeclaration(changes, deletedNodesInLists, sourceFile, node) {
                    switch (node.kind) {
                        case 162 /* Parameter */: {
                            var oldFunction = node.parent;
                            if (ts.isArrowFunction(oldFunction) &&
                                oldFunction.parameters.length === 1 &&
                                !ts.findChildOfKind(oldFunction, 20 /* OpenParenToken */, sourceFile)) {
                                // Lambdas with exactly one parameter are special because, after removal, there
                                // must be an empty parameter list (i.e. `()`) and this won't necessarily be the
                                // case if the parameter is simply removed (e.g. in `x => 1`).
                                changes.replaceNodeWithText(sourceFile, node, "()");
                            }
                            else {
                                deleteNodeInList(changes, deletedNodesInLists, sourceFile, node);
                            }
                            break;
                        }
                        case 264 /* ImportDeclaration */:
                        case 263 /* ImportEqualsDeclaration */:
                            var isFirstImport = sourceFile.imports.length && node === ts.first(sourceFile.imports).parent || node === ts.find(sourceFile.statements, ts.isAnyImportSyntax);
                            // For first import, leave header comment in place, otherwise only delete JSDoc comments
                            deleteNode(changes, sourceFile, node, {
                                leadingTriviaOption: isFirstImport ? LeadingTriviaOption.Exclude : ts.hasJSDocNodes(node) ? LeadingTriviaOption.JSDoc : LeadingTriviaOption.StartLine,
                            });
                            break;
                        case 201 /* BindingElement */:
                            var pattern = node.parent;
                            var preserveComma = pattern.kind === 200 /* ArrayBindingPattern */ && node !== ts.last(pattern.elements);
                            if (preserveComma) {
                                deleteNode(changes, sourceFile, node);
                            }
                            else {
                                deleteNodeInList(changes, deletedNodesInLists, sourceFile, node);
                            }
                            break;
                        case 252 /* VariableDeclaration */:
                            deleteVariableDeclaration(changes, deletedNodesInLists, sourceFile, node);
                            break;
                        case 161 /* TypeParameter */:
                            deleteNodeInList(changes, deletedNodesInLists, sourceFile, node);
                            break;
                        case 268 /* ImportSpecifier */:
                            var namedImports = node.parent;
                            if (namedImports.elements.length === 1) {
                                deleteImportBinding(changes, sourceFile, namedImports);
                            }
                            else {
                                deleteNodeInList(changes, deletedNodesInLists, sourceFile, node);
                            }
                            break;
                        case 266 /* NamespaceImport */:
                            deleteImportBinding(changes, sourceFile, node);
                            break;
                        case 26 /* SemicolonToken */:
                            deleteNode(changes, sourceFile, node, { trailingTriviaOption: TrailingTriviaOption.Exclude });
                            break;
                        case 98 /* FunctionKeyword */:
                            deleteNode(changes, sourceFile, node, { leadingTriviaOption: LeadingTriviaOption.Exclude });
                            break;
                        case 255 /* ClassDeclaration */:
                        case 254 /* FunctionDeclaration */:
                            deleteNode(changes, sourceFile, node, { leadingTriviaOption: ts.hasJSDocNodes(node) ? LeadingTriviaOption.JSDoc : LeadingTriviaOption.StartLine });
                            break;
                        default:
                            if (!node.parent) {
                                // a misbehaving client can reach here with the SourceFile node
                                deleteNode(changes, sourceFile, node);
                            }
                            else if (ts.isImportClause(node.parent) && node.parent.name === node) {
                                deleteDefaultImport(changes, sourceFile, node.parent);
                            }
                            else if (ts.isCallExpression(node.parent) && ts.contains(node.parent.arguments, node)) {
                                deleteNodeInList(changes, deletedNodesInLists, sourceFile, node);
                            }
                            else {
                                deleteNode(changes, sourceFile, node);
                            }
                    }
                }
                deleteDeclaration_1.deleteDeclaration = deleteDeclaration;
                function deleteDefaultImport(changes, sourceFile, importClause) {
                    if (!importClause.namedBindings) {
                        // Delete the whole import
                        deleteNode(changes, sourceFile, importClause.parent);
                    }
                    else {
                        // import |d,| * as ns from './file'
                        var start = importClause.name.getStart(sourceFile);
                        var nextToken = ts.getTokenAtPosition(sourceFile, importClause.name.end);
                        if (nextToken && nextToken.kind === 27 /* CommaToken */) {
                            // shift first non-whitespace position after comma to the start position of the node
                            var end = ts.skipTrivia(sourceFile.text, nextToken.end, /*stopAfterLineBreaks*/ false, /*stopAtComments*/ true);
                            changes.deleteRange(sourceFile, { pos: start, end: end });
                        }
                        else {
                            deleteNode(changes, sourceFile, importClause.name);
                        }
                    }
                }
                function deleteImportBinding(changes, sourceFile, node) {
                    if (node.parent.name) {
                        // Delete named imports while preserving the default import
                        // import d|, * as ns| from './file'
                        // import d|, { a }| from './file'
                        var previousToken = ts.Debug.checkDefined(ts.getTokenAtPosition(sourceFile, node.pos - 1));
                        changes.deleteRange(sourceFile, { pos: previousToken.getStart(sourceFile), end: node.end });
                    }
                    else {
                        // Delete the entire import declaration
                        // |import * as ns from './file'|
                        // |import { a } from './file'|
                        var importDecl = ts.getAncestor(node, 264 /* ImportDeclaration */);
                        deleteNode(changes, sourceFile, importDecl);
                    }
                }
                function deleteVariableDeclaration(changes, deletedNodesInLists, sourceFile, node) {
                    var parent = node.parent;
                    if (parent.kind === 290 /* CatchClause */) {
                        // TODO: There's currently no unused diagnostic for this, could be a suggestion
                        changes.deleteNodeRange(sourceFile, ts.findChildOfKind(parent, 20 /* OpenParenToken */, sourceFile), ts.findChildOfKind(parent, 21 /* CloseParenToken */, sourceFile));
                        return;
                    }
                    if (parent.declarations.length !== 1) {
                        deleteNodeInList(changes, deletedNodesInLists, sourceFile, node);
                        return;
                    }
                    var gp = parent.parent;
                    switch (gp.kind) {
                        case 242 /* ForOfStatement */:
                        case 241 /* ForInStatement */:
                            changes.replaceNode(sourceFile, node, ts.factory.createObjectLiteralExpression());
                            break;
                        case 240 /* ForStatement */:
                            deleteNode(changes, sourceFile, parent);
                            break;
                        case 235 /* VariableStatement */:
                            deleteNode(changes, sourceFile, gp, { leadingTriviaOption: ts.hasJSDocNodes(gp) ? LeadingTriviaOption.JSDoc : LeadingTriviaOption.StartLine });
                            break;
                        default:
                            ts.Debug.assertNever(gp);
                    }
                }
            })(deleteDeclaration || (deleteDeclaration = {}));
            /** Warning: This deletes comments too. See `copyComments` in `convertFunctionToEs6Class`. */
            // Exported for tests only! (TODO: improve tests to not need this)
            function deleteNode(changes, sourceFile, node, options) {
                if (options === void 0) {
                    options = { leadingTriviaOption: LeadingTriviaOption.IncludeAll };
                }
                var startPosition = getAdjustedStartPosition(sourceFile, node, options);
                var endPosition = getAdjustedEndPosition(sourceFile, node, options);
                changes.deleteRange(sourceFile, { pos: startPosition, end: endPosition });
            }
            textChanges_3.deleteNode = deleteNode;
            function deleteNodeInList(changes, deletedNodesInLists, sourceFile, node) {
                var containingList = ts.Debug.checkDefined(ts.formatting.SmartIndenter.getContainingList(node, sourceFile));
                var index = ts.indexOfNode(containingList, node);
                ts.Debug.assert(index !== -1);
                if (containingList.length === 1) {
                    deleteNode(changes, sourceFile, node);
                    return;
                }
                // Note: We will only delete a comma *after* a node. This will leave a trailing comma if we delete the last node.
                // That's handled in the end by `finishTrailingCommaAfterDeletingNodesInList`.
                ts.Debug.assert(!deletedNodesInLists.has(node), "Deleting a node twice");
                deletedNodesInLists.add(node);
                changes.deleteRange(sourceFile, {
                    pos: startPositionToDeleteNodeInList(sourceFile, node),
                    end: index === containingList.length - 1 ? getAdjustedEndPosition(sourceFile, node, {}) : startPositionToDeleteNodeInList(sourceFile, containingList[index + 1]),
                });
            }
        })(textChanges = ts.textChanges || (ts.textChanges = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var errorCodeToFixes = ts.createMultiMap();
            var fixIdToRegistration = new ts.Map();
            function diagnosticToString(diag) {
                return ts.isArray(diag)
                    ? ts.formatStringFromArgs(ts.getLocaleSpecificMessage(diag[0]), diag.slice(1))
                    : ts.getLocaleSpecificMessage(diag);
            }
            function createCodeFixActionWithoutFixAll(fixName, changes, description) {
                return createCodeFixActionWorker(fixName, diagnosticToString(description), changes, /*fixId*/ undefined, /*fixAllDescription*/ undefined);
            }
            codefix.createCodeFixActionWithoutFixAll = createCodeFixActionWithoutFixAll;
            function createCodeFixAction(fixName, changes, description, fixId, fixAllDescription, command) {
                return createCodeFixActionWorker(fixName, diagnosticToString(description), changes, fixId, diagnosticToString(fixAllDescription), command);
            }
            codefix.createCodeFixAction = createCodeFixAction;
            function createCodeFixActionMaybeFixAll(fixName, changes, description, fixId, fixAllDescription, command) {
                return createCodeFixActionWorker(fixName, diagnosticToString(description), changes, fixId, fixAllDescription && diagnosticToString(fixAllDescription), command);
            }
            codefix.createCodeFixActionMaybeFixAll = createCodeFixActionMaybeFixAll;
            function createCodeFixActionWorker(fixName, description, changes, fixId, fixAllDescription, command) {
                return { fixName: fixName, description: description, changes: changes, fixId: fixId, fixAllDescription: fixAllDescription, commands: command ? [command] : undefined };
            }
            function registerCodeFix(reg) {
                for (var _i = 0, _a = reg.errorCodes; _i < _a.length; _i++) {
                    var error = _a[_i];
                    errorCodeToFixes.add(String(error), reg);
                }
                if (reg.fixIds) {
                    for (var _b = 0, _c = reg.fixIds; _b < _c.length; _b++) {
                        var fixId = _c[_b];
                        ts.Debug.assert(!fixIdToRegistration.has(fixId));
                        fixIdToRegistration.set(fixId, reg);
                    }
                }
            }
            codefix.registerCodeFix = registerCodeFix;
            function getSupportedErrorCodes() {
                return ts.arrayFrom(errorCodeToFixes.keys());
            }
            codefix.getSupportedErrorCodes = getSupportedErrorCodes;
            function removeFixIdIfFixAllUnavailable(registration, diagnostics) {
                var errorCodes = registration.errorCodes;
                var maybeFixableDiagnostics = 0;
                for (var _i = 0, diagnostics_1 = diagnostics; _i < diagnostics_1.length; _i++) {
                    var diag = diagnostics_1[_i];
                    if (ts.contains(errorCodes, diag.code))
                        maybeFixableDiagnostics++;
                    if (maybeFixableDiagnostics > 1)
                        break;
                }
                var fixAllUnavailable = maybeFixableDiagnostics < 2;
                return function (_a) {
                    var fixId = _a.fixId, fixAllDescription = _a.fixAllDescription, action = __rest(_a, ["fixId", "fixAllDescription"]);
                    return fixAllUnavailable ? action : __assign(__assign({}, action), { fixId: fixId, fixAllDescription: fixAllDescription });
                };
            }
            function getFixes(context) {
                var diagnostics = getDiagnostics(context);
                var registrations = errorCodeToFixes.get(String(context.errorCode));
                return ts.flatMap(registrations, function (f) { return ts.map(f.getCodeActions(context), removeFixIdIfFixAllUnavailable(f, diagnostics)); });
            }
            codefix.getFixes = getFixes;
            function getAllFixes(context) {
                // Currently fixId is always a string.
                return fixIdToRegistration.get(ts.cast(context.fixId, ts.isString)).getAllCodeActions(context);
            }
            codefix.getAllFixes = getAllFixes;
            function createCombinedCodeActions(changes, commands) {
                return { changes: changes, commands: commands };
            }
            codefix.createCombinedCodeActions = createCombinedCodeActions;
            function createFileTextChanges(fileName, textChanges) {
                return { fileName: fileName, textChanges: textChanges };
            }
            codefix.createFileTextChanges = createFileTextChanges;
            function codeFixAll(context, errorCodes, use) {
                var commands = [];
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return eachDiagnostic(context, errorCodes, function (diag) { return use(t, diag, commands); }); });
                return createCombinedCodeActions(changes, commands.length === 0 ? undefined : commands);
            }
            codefix.codeFixAll = codeFixAll;
            function eachDiagnostic(context, errorCodes, cb) {
                for (var _i = 0, _a = getDiagnostics(context); _i < _a.length; _i++) {
                    var diag = _a[_i];
                    if (ts.contains(errorCodes, diag.code)) {
                        cb(diag);
                    }
                }
            }
            codefix.eachDiagnostic = eachDiagnostic;
            function getDiagnostics(_a) {
                var program = _a.program, sourceFile = _a.sourceFile, cancellationToken = _a.cancellationToken;
                return __spreadArray(__spreadArray(__spreadArray([], program.getSemanticDiagnostics(sourceFile, cancellationToken), true), program.getSyntacticDiagnostics(sourceFile, cancellationToken), true), ts.computeSuggestionDiagnostics(sourceFile, program, cancellationToken), true);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor_1) {
            // A map with the refactor code as key, the refactor itself as value
            // e.g.  nonSuggestableRefactors[refactorCode] -> the refactor you want
            var refactors = new ts.Map();
            /** @param name An unique code associated with each refactor. Does not have to be human-readable. */
            function registerRefactor(name, refactor) {
                refactors.set(name, refactor);
            }
            refactor_1.registerRefactor = registerRefactor;
            function getApplicableRefactors(context) {
                return ts.arrayFrom(ts.flatMapIterator(refactors.values(), function (refactor) {
                    var _a;
                    return context.cancellationToken && context.cancellationToken.isCancellationRequested() ||
                        !((_a = refactor.kinds) === null || _a === void 0 ? void 0 : _a.some(function (kind) { return refactor_1.refactorKindBeginsWith(kind, context.kind); })) ? undefined :
                        refactor.getAvailableActions(context);
                }));
            }
            refactor_1.getApplicableRefactors = getApplicableRefactors;
            function getEditsForRefactor(context, refactorName, actionName) {
                var refactor = refactors.get(refactorName);
                return refactor && refactor.getEditsForAction(context, actionName);
            }
            refactor_1.getEditsForRefactor = getEditsForRefactor;
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "addConvertToUnknownForNonOverlappingTypes";
            var errorCodes = [ts.Diagnostics.Conversion_of_type_0_to_type_1_may_be_a_mistake_because_neither_type_sufficiently_overlaps_with_the_other_If_this_was_intentional_convert_the_expression_to_unknown_first.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return makeChange(t, context.sourceFile, context.span.start); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_unknown_conversion_for_non_overlapping_types, fixId, ts.Diagnostics.Add_unknown_to_all_conversions_of_non_overlapping_types)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) { return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return makeChange(changes, diag.file, diag.start); }); },
            });
            function makeChange(changeTracker, sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                var assertion = ts.Debug.checkDefined(ts.findAncestor(token, function (n) { return ts.isAsExpression(n) || ts.isTypeAssertionExpression(n); }), "Expected to find an assertion expression");
                var replacement = ts.isAsExpression(assertion)
                    ? ts.factory.createAsExpression(assertion.expression, ts.factory.createKeywordTypeNode(153 /* UnknownKeyword */))
                    : ts.factory.createTypeAssertion(ts.factory.createKeywordTypeNode(153 /* UnknownKeyword */), assertion.expression);
                changeTracker.replaceNode(sourceFile, assertion.expression, replacement);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            codefix.registerCodeFix({
                errorCodes: [
                    ts.Diagnostics.await_expressions_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_has_no_imports_or_exports_Consider_adding_an_empty_export_to_make_this_file_a_module.code,
                    ts.Diagnostics.for_await_loops_are_only_allowed_at_the_top_level_of_a_file_when_that_file_is_a_module_but_this_file_has_no_imports_or_exports_Consider_adding_an_empty_export_to_make_this_file_a_module.code,
                ],
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (changes) {
                        var exportDeclaration = ts.factory.createExportDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, 
                        /*isTypeOnly*/ false, ts.factory.createNamedExports([]), 
                        /*moduleSpecifier*/ undefined);
                        changes.insertNodeAtEndOfScope(sourceFile, sourceFile, exportDeclaration);
                    });
                    return [codefix.createCodeFixActionWithoutFixAll("addEmptyExportDeclaration", changes, ts.Diagnostics.Add_export_to_make_this_file_into_a_module)];
                },
            });
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "addMissingAsync";
            var errorCodes = [
                ts.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code,
                ts.Diagnostics.Type_0_is_not_assignable_to_type_1.code,
                ts.Diagnostics.Type_0_is_not_comparable_to_type_1.code
            ];
            codefix.registerCodeFix({
                fixIds: [fixId],
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, errorCode = context.errorCode, cancellationToken = context.cancellationToken, program = context.program, span = context.span;
                    var diagnostic = ts.find(program.getDiagnosticsProducingTypeChecker().getDiagnostics(sourceFile, cancellationToken), getIsMatchingAsyncError(span, errorCode));
                    var directSpan = diagnostic && diagnostic.relatedInformation && ts.find(diagnostic.relatedInformation, function (r) { return r.code === ts.Diagnostics.Did_you_mean_to_mark_this_function_as_async.code; });
                    var decl = getFixableErrorSpanDeclaration(sourceFile, directSpan);
                    if (!decl) {
                        return;
                    }
                    var trackChanges = function (cb) { return ts.textChanges.ChangeTracker.with(context, cb); };
                    return [getFix(context, decl, trackChanges)];
                },
                getAllCodeActions: function (context) {
                    var sourceFile = context.sourceFile;
                    var fixedDeclarations = new ts.Set();
                    return codefix.codeFixAll(context, errorCodes, function (t, diagnostic) {
                        var span = diagnostic.relatedInformation && ts.find(diagnostic.relatedInformation, function (r) { return r.code === ts.Diagnostics.Did_you_mean_to_mark_this_function_as_async.code; });
                        var decl = getFixableErrorSpanDeclaration(sourceFile, span);
                        if (!decl) {
                            return;
                        }
                        var trackChanges = function (cb) { return (cb(t), []); };
                        return getFix(context, decl, trackChanges, fixedDeclarations);
                    });
                },
            });
            function getFix(context, decl, trackChanges, fixedDeclarations) {
                var changes = trackChanges(function (t) { return makeChange(t, context.sourceFile, decl, fixedDeclarations); });
                return codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_async_modifier_to_containing_function, fixId, ts.Diagnostics.Add_all_missing_async_modifiers);
            }
            function makeChange(changeTracker, sourceFile, insertionSite, fixedDeclarations) {
                if (fixedDeclarations) {
                    if (fixedDeclarations.has(ts.getNodeId(insertionSite))) {
                        return;
                    }
                }
                fixedDeclarations === null || fixedDeclarations === void 0 ? void 0 : fixedDeclarations.add(ts.getNodeId(insertionSite));
                var cloneWithModifier = ts.factory.updateModifiers(ts.getSynthesizedDeepClone(insertionSite, /*includeTrivia*/ true), ts.factory.createNodeArray(ts.factory.createModifiersFromModifierFlags(ts.getSyntacticModifierFlags(insertionSite) | 256 /* Async */)));
                changeTracker.replaceNode(sourceFile, insertionSite, cloneWithModifier);
            }
            function getFixableErrorSpanDeclaration(sourceFile, span) {
                if (!span)
                    return undefined;
                var token = ts.getTokenAtPosition(sourceFile, span.start);
                // Checker has already done work to determine that async might be possible, and has attached
                // related info to the node, so start by finding the signature that exactly matches up
                // with the diagnostic range.
                var decl = ts.findAncestor(token, function (node) {
                    if (node.getStart(sourceFile) < span.start || node.getEnd() > ts.textSpanEnd(span)) {
                        return "quit";
                    }
                    return (ts.isArrowFunction(node) || ts.isMethodDeclaration(node) || ts.isFunctionExpression(node) || ts.isFunctionDeclaration(node)) && ts.textSpansEqual(span, ts.createTextSpanFromNode(node, sourceFile));
                });
                return decl;
            }
            function getIsMatchingAsyncError(span, errorCode) {
                return function (_a) {
                    var start = _a.start, length = _a.length, relatedInformation = _a.relatedInformation, code = _a.code;
                    return ts.isNumber(start) && ts.isNumber(length) && ts.textSpansEqual({ start: start, length: length }, span) &&
                        code === errorCode &&
                        !!relatedInformation &&
                        ts.some(relatedInformation, function (related) { return related.code === ts.Diagnostics.Did_you_mean_to_mark_this_function_as_async.code; });
                };
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "addMissingAwait";
            var propertyAccessCode = ts.Diagnostics.Property_0_does_not_exist_on_type_1.code;
            var callableConstructableErrorCodes = [
                ts.Diagnostics.This_expression_is_not_callable.code,
                ts.Diagnostics.This_expression_is_not_constructable.code,
            ];
            var errorCodes = __spreadArray([
                ts.Diagnostics.An_arithmetic_operand_must_be_of_type_any_number_bigint_or_an_enum_type.code,
                ts.Diagnostics.The_left_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type.code,
                ts.Diagnostics.The_right_hand_side_of_an_arithmetic_operation_must_be_of_type_any_number_bigint_or_an_enum_type.code,
                ts.Diagnostics.Operator_0_cannot_be_applied_to_type_1.code,
                ts.Diagnostics.Operator_0_cannot_be_applied_to_types_1_and_2.code,
                ts.Diagnostics.This_condition_will_always_return_0_since_the_types_1_and_2_have_no_overlap.code,
                ts.Diagnostics.This_condition_will_always_return_true_since_this_0_is_always_defined.code,
                ts.Diagnostics.Type_0_is_not_an_array_type.code,
                ts.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type.code,
                ts.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_Use_compiler_option_downlevelIteration_to_allow_iterating_of_iterators.code,
                ts.Diagnostics.Type_0_is_not_an_array_type_or_a_string_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator.code,
                ts.Diagnostics.Type_0_is_not_an_array_type_or_does_not_have_a_Symbol_iterator_method_that_returns_an_iterator.code,
                ts.Diagnostics.Type_0_must_have_a_Symbol_iterator_method_that_returns_an_iterator.code,
                ts.Diagnostics.Type_0_must_have_a_Symbol_asyncIterator_method_that_returns_an_async_iterator.code,
                ts.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code,
                propertyAccessCode
            ], callableConstructableErrorCodes, true);
            codefix.registerCodeFix({
                fixIds: [fixId],
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, errorCode = context.errorCode, span = context.span, cancellationToken = context.cancellationToken, program = context.program;
                    var expression = getFixableErrorSpanExpression(sourceFile, errorCode, span, cancellationToken, program);
                    if (!expression) {
                        return;
                    }
                    var checker = context.program.getTypeChecker();
                    var trackChanges = function (cb) { return ts.textChanges.ChangeTracker.with(context, cb); };
                    return ts.compact([
                        getDeclarationSiteFix(context, expression, errorCode, checker, trackChanges),
                        getUseSiteFix(context, expression, errorCode, checker, trackChanges)
                    ]);
                },
                getAllCodeActions: function (context) {
                    var sourceFile = context.sourceFile, program = context.program, cancellationToken = context.cancellationToken;
                    var checker = context.program.getTypeChecker();
                    var fixedDeclarations = new ts.Set();
                    return codefix.codeFixAll(context, errorCodes, function (t, diagnostic) {
                        var expression = getFixableErrorSpanExpression(sourceFile, diagnostic.code, diagnostic, cancellationToken, program);
                        if (!expression) {
                            return;
                        }
                        var trackChanges = function (cb) { return (cb(t), []); };
                        return getDeclarationSiteFix(context, expression, diagnostic.code, checker, trackChanges, fixedDeclarations)
                            || getUseSiteFix(context, expression, diagnostic.code, checker, trackChanges, fixedDeclarations);
                    });
                },
            });
            function getDeclarationSiteFix(context, expression, errorCode, checker, trackChanges, fixedDeclarations) {
                var sourceFile = context.sourceFile, program = context.program, cancellationToken = context.cancellationToken;
                var awaitableInitializers = findAwaitableInitializers(expression, sourceFile, cancellationToken, program, checker);
                if (awaitableInitializers) {
                    var initializerChanges = trackChanges(function (t) {
                        ts.forEach(awaitableInitializers.initializers, function (_a) {
                            var expression = _a.expression;
                            return makeChange(t, errorCode, sourceFile, checker, expression, fixedDeclarations);
                        });
                        if (fixedDeclarations && awaitableInitializers.needsSecondPassForFixAll) {
                            makeChange(t, errorCode, sourceFile, checker, expression, fixedDeclarations);
                        }
                    });
                    // No fix-all because it will already be included once with the use site fix,
                    // and for simplicity the fix-all doesn‘t let the user choose between use-site and declaration-site fixes.
                    return codefix.createCodeFixActionWithoutFixAll("addMissingAwaitToInitializer", initializerChanges, awaitableInitializers.initializers.length === 1
                        ? [ts.Diagnostics.Add_await_to_initializer_for_0, awaitableInitializers.initializers[0].declarationSymbol.name]
                        : ts.Diagnostics.Add_await_to_initializers);
                }
            }
            function getUseSiteFix(context, expression, errorCode, checker, trackChanges, fixedDeclarations) {
                var changes = trackChanges(function (t) { return makeChange(t, errorCode, context.sourceFile, checker, expression, fixedDeclarations); });
                return codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_await, fixId, ts.Diagnostics.Fix_all_expressions_possibly_missing_await);
            }
            function isMissingAwaitError(sourceFile, errorCode, span, cancellationToken, program) {
                var checker = program.getDiagnosticsProducingTypeChecker();
                var diagnostics = checker.getDiagnostics(sourceFile, cancellationToken);
                return ts.some(diagnostics, function (_a) {
                    var start = _a.start, length = _a.length, relatedInformation = _a.relatedInformation, code = _a.code;
                    return ts.isNumber(start) && ts.isNumber(length) && ts.textSpansEqual({ start: start, length: length }, span) &&
                        code === errorCode &&
                        !!relatedInformation &&
                        ts.some(relatedInformation, function (related) { return related.code === ts.Diagnostics.Did_you_forget_to_use_await.code; });
                });
            }
            function getFixableErrorSpanExpression(sourceFile, errorCode, span, cancellationToken, program) {
                var token = ts.getTokenAtPosition(sourceFile, span.start);
                // Checker has already done work to determine that await might be possible, and has attached
                // related info to the node, so start by finding the expression that exactly matches up
                // with the diagnostic range.
                var expression = ts.findAncestor(token, function (node) {
                    if (node.getStart(sourceFile) < span.start || node.getEnd() > ts.textSpanEnd(span)) {
                        return "quit";
                    }
                    return ts.isExpression(node) && ts.textSpansEqual(span, ts.createTextSpanFromNode(node, sourceFile));
                });
                return expression
                    && isMissingAwaitError(sourceFile, errorCode, span, cancellationToken, program)
                    && isInsideAwaitableBody(expression) ? expression : undefined;
            }
            function findAwaitableInitializers(expression, sourceFile, cancellationToken, program, checker) {
                var identifiers = getIdentifiersFromErrorSpanExpression(expression, checker);
                if (!identifiers) {
                    return;
                }
                var isCompleteFix = identifiers.isCompleteFix;
                var initializers;
                var _loop_12 = function (identifier) {
                    var symbol = checker.getSymbolAtLocation(identifier);
                    if (!symbol) {
                        return "continue";
                    }
                    var declaration = ts.tryCast(symbol.valueDeclaration, ts.isVariableDeclaration);
                    var variableName = declaration && ts.tryCast(declaration.name, ts.isIdentifier);
                    var variableStatement = ts.getAncestor(declaration, 235 /* VariableStatement */);
                    if (!declaration || !variableStatement ||
                        declaration.type ||
                        !declaration.initializer ||
                        variableStatement.getSourceFile() !== sourceFile ||
                        ts.hasSyntacticModifier(variableStatement, 1 /* Export */) ||
                        !variableName ||
                        !isInsideAwaitableBody(declaration.initializer)) {
                        isCompleteFix = false;
                        return "continue";
                    }
                    var diagnostics = program.getSemanticDiagnostics(sourceFile, cancellationToken);
                    var isUsedElsewhere = ts.FindAllReferences.Core.eachSymbolReferenceInFile(variableName, checker, sourceFile, function (reference) {
                        return identifier !== reference && !symbolReferenceIsAlsoMissingAwait(reference, diagnostics, sourceFile, checker);
                    });
                    if (isUsedElsewhere) {
                        isCompleteFix = false;
                        return "continue";
                    }
                    (initializers || (initializers = [])).push({
                        expression: declaration.initializer,
                        declarationSymbol: symbol,
                    });
                };
                for (var _i = 0, _a = identifiers.identifiers; _i < _a.length; _i++) {
                    var identifier = _a[_i];
                    _loop_12(identifier);
                }
                return initializers && {
                    initializers: initializers,
                    needsSecondPassForFixAll: !isCompleteFix,
                };
            }
            function getIdentifiersFromErrorSpanExpression(expression, checker) {
                if (ts.isPropertyAccessExpression(expression.parent) && ts.isIdentifier(expression.parent.expression)) {
                    return { identifiers: [expression.parent.expression], isCompleteFix: true };
                }
                if (ts.isIdentifier(expression)) {
                    return { identifiers: [expression], isCompleteFix: true };
                }
                if (ts.isBinaryExpression(expression)) {
                    var sides = void 0;
                    var isCompleteFix = true;
                    for (var _i = 0, _a = [expression.left, expression.right]; _i < _a.length; _i++) {
                        var side = _a[_i];
                        var type = checker.getTypeAtLocation(side);
                        if (checker.getPromisedTypeOfPromise(type)) {
                            if (!ts.isIdentifier(side)) {
                                isCompleteFix = false;
                                continue;
                            }
                            (sides || (sides = [])).push(side);
                        }
                    }
                    return sides && { identifiers: sides, isCompleteFix: isCompleteFix };
                }
            }
            function symbolReferenceIsAlsoMissingAwait(reference, diagnostics, sourceFile, checker) {
                var errorNode = ts.isPropertyAccessExpression(reference.parent) ? reference.parent.name :
                    ts.isBinaryExpression(reference.parent) ? reference.parent :
                        reference;
                var diagnostic = ts.find(diagnostics, function (diagnostic) {
                    return diagnostic.start === errorNode.getStart(sourceFile) &&
                        (diagnostic.start + diagnostic.length) === errorNode.getEnd();
                });
                return diagnostic && ts.contains(errorCodes, diagnostic.code) ||
                    // A Promise is usually not correct in a binary expression (it’s not valid
                    // in an arithmetic expression and an equality comparison seems unusual),
                    // but if the other side of the binary expression has an error, the side
                    // is typed `any` which will squash the error that would identify this
                    // Promise as an invalid operand. So if the whole binary expression is
                    // typed `any` as a result, there is a strong likelihood that this Promise
                    // is accidentally missing `await`.
                    checker.getTypeAtLocation(errorNode).flags & 1 /* Any */;
            }
            function isInsideAwaitableBody(node) {
                return node.kind & 32768 /* AwaitContext */ || !!ts.findAncestor(node, function (ancestor) {
                    return ancestor.parent && ts.isArrowFunction(ancestor.parent) && ancestor.parent.body === ancestor ||
                        ts.isBlock(ancestor) && (ancestor.parent.kind === 254 /* FunctionDeclaration */ ||
                            ancestor.parent.kind === 211 /* FunctionExpression */ ||
                            ancestor.parent.kind === 212 /* ArrowFunction */ ||
                            ancestor.parent.kind === 167 /* MethodDeclaration */);
                });
            }
            function makeChange(changeTracker, errorCode, sourceFile, checker, insertionSite, fixedDeclarations) {
                if (ts.isBinaryExpression(insertionSite)) {
                    for (var _i = 0, _a = [insertionSite.left, insertionSite.right]; _i < _a.length; _i++) {
                        var side = _a[_i];
                        if (fixedDeclarations && ts.isIdentifier(side)) {
                            var symbol = checker.getSymbolAtLocation(side);
                            if (symbol && fixedDeclarations.has(ts.getSymbolId(symbol))) {
                                continue;
                            }
                        }
                        var type = checker.getTypeAtLocation(side);
                        var newNode = checker.getPromisedTypeOfPromise(type) ? ts.factory.createAwaitExpression(side) : side;
                        changeTracker.replaceNode(sourceFile, side, newNode);
                    }
                }
                else if (errorCode === propertyAccessCode && ts.isPropertyAccessExpression(insertionSite.parent)) {
                    if (fixedDeclarations && ts.isIdentifier(insertionSite.parent.expression)) {
                        var symbol = checker.getSymbolAtLocation(insertionSite.parent.expression);
                        if (symbol && fixedDeclarations.has(ts.getSymbolId(symbol))) {
                            return;
                        }
                    }
                    changeTracker.replaceNode(sourceFile, insertionSite.parent.expression, ts.factory.createParenthesizedExpression(ts.factory.createAwaitExpression(insertionSite.parent.expression)));
                    insertLeadingSemicolonIfNeeded(changeTracker, insertionSite.parent.expression, sourceFile);
                }
                else if (ts.contains(callableConstructableErrorCodes, errorCode) && ts.isCallOrNewExpression(insertionSite.parent)) {
                    if (fixedDeclarations && ts.isIdentifier(insertionSite)) {
                        var symbol = checker.getSymbolAtLocation(insertionSite);
                        if (symbol && fixedDeclarations.has(ts.getSymbolId(symbol))) {
                            return;
                        }
                    }
                    changeTracker.replaceNode(sourceFile, insertionSite, ts.factory.createParenthesizedExpression(ts.factory.createAwaitExpression(insertionSite)));
                    insertLeadingSemicolonIfNeeded(changeTracker, insertionSite, sourceFile);
                }
                else {
                    if (fixedDeclarations && ts.isVariableDeclaration(insertionSite.parent) && ts.isIdentifier(insertionSite.parent.name)) {
                        var symbol = checker.getSymbolAtLocation(insertionSite.parent.name);
                        if (symbol && !ts.tryAddToSet(fixedDeclarations, ts.getSymbolId(symbol))) {
                            return;
                        }
                    }
                    changeTracker.replaceNode(sourceFile, insertionSite, ts.factory.createAwaitExpression(insertionSite));
                }
            }
            function insertLeadingSemicolonIfNeeded(changeTracker, beforeNode, sourceFile) {
                var precedingToken = ts.findPrecedingToken(beforeNode.pos, sourceFile);
                if (precedingToken && ts.positionIsASICandidate(precedingToken.end, precedingToken.parent, sourceFile)) {
                    changeTracker.insertText(sourceFile, beforeNode.getStart(sourceFile), ";");
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "addMissingConst";
            var errorCodes = [
                ts.Diagnostics.Cannot_find_name_0.code,
                ts.Diagnostics.No_value_exists_in_scope_for_the_shorthand_property_0_Either_declare_one_or_provide_an_initializer.code
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return makeChange(t, context.sourceFile, context.span.start, context.program); });
                    if (changes.length > 0) {
                        return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_const_to_unresolved_variable, fixId, ts.Diagnostics.Add_const_to_all_unresolved_variables)];
                    }
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    var fixedNodes = new ts.Set();
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return makeChange(changes, diag.file, diag.start, context.program, fixedNodes); });
                },
            });
            function makeChange(changeTracker, sourceFile, pos, program, fixedNodes) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                var forInitializer = ts.findAncestor(token, function (node) {
                    return ts.isForInOrOfStatement(node.parent) ? node.parent.initializer === node :
                        isPossiblyPartOfDestructuring(node) ? false : "quit";
                });
                if (forInitializer)
                    return applyChange(changeTracker, forInitializer, sourceFile, fixedNodes);
                var parent = token.parent;
                if (ts.isBinaryExpression(parent) && parent.operatorToken.kind === 63 /* EqualsToken */ && ts.isExpressionStatement(parent.parent)) {
                    return applyChange(changeTracker, token, sourceFile, fixedNodes);
                }
                if (ts.isArrayLiteralExpression(parent)) {
                    var checker_1 = program.getTypeChecker();
                    if (!ts.every(parent.elements, function (element) { return arrayElementCouldBeVariableDeclaration(element, checker_1); })) {
                        return;
                    }
                    return applyChange(changeTracker, parent, sourceFile, fixedNodes);
                }
                var commaExpression = ts.findAncestor(token, function (node) {
                    return ts.isExpressionStatement(node.parent) ? true :
                        isPossiblyPartOfCommaSeperatedInitializer(node) ? false : "quit";
                });
                if (commaExpression) {
                    var checker = program.getTypeChecker();
                    if (!expressionCouldBeVariableDeclaration(commaExpression, checker)) {
                        return;
                    }
                    return applyChange(changeTracker, commaExpression, sourceFile, fixedNodes);
                }
            }
            function applyChange(changeTracker, initializer, sourceFile, fixedNodes) {
                if (!fixedNodes || ts.tryAddToSet(fixedNodes, initializer)) {
                    changeTracker.insertModifierBefore(sourceFile, 85 /* ConstKeyword */, initializer);
                }
            }
            function isPossiblyPartOfDestructuring(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                    case 202 /* ArrayLiteralExpression */:
                    case 203 /* ObjectLiteralExpression */:
                    case 291 /* PropertyAssignment */:
                    case 292 /* ShorthandPropertyAssignment */:
                        return true;
                    default:
                        return false;
                }
            }
            function arrayElementCouldBeVariableDeclaration(expression, checker) {
                var identifier = ts.isIdentifier(expression) ? expression :
                    ts.isAssignmentExpression(expression, /*excludeCompoundAssignment*/ true) && ts.isIdentifier(expression.left) ? expression.left :
                        undefined;
                return !!identifier && !checker.getSymbolAtLocation(identifier);
            }
            function isPossiblyPartOfCommaSeperatedInitializer(node) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                    case 219 /* BinaryExpression */:
                    case 27 /* CommaToken */:
                        return true;
                    default:
                        return false;
                }
            }
            function expressionCouldBeVariableDeclaration(expression, checker) {
                if (!ts.isBinaryExpression(expression)) {
                    return false;
                }
                if (expression.operatorToken.kind === 27 /* CommaToken */) {
                    return ts.every([expression.left, expression.right], function (expression) { return expressionCouldBeVariableDeclaration(expression, checker); });
                }
                return expression.operatorToken.kind === 63 /* EqualsToken */
                    && ts.isIdentifier(expression.left)
                    && !checker.getSymbolAtLocation(expression.left);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "addMissingDeclareProperty";
            var errorCodes = [
                ts.Diagnostics.Property_0_will_overwrite_the_base_property_in_1_If_this_is_intentional_add_an_initializer_Otherwise_add_a_declare_modifier_or_remove_the_redundant_declaration.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return makeChange(t, context.sourceFile, context.span.start); });
                    if (changes.length > 0) {
                        return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Prefix_with_declare, fixId, ts.Diagnostics.Prefix_all_incorrect_property_declarations_with_declare)];
                    }
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    var fixedNodes = new ts.Set();
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return makeChange(changes, diag.file, diag.start, fixedNodes); });
                },
            });
            function makeChange(changeTracker, sourceFile, pos, fixedNodes) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                if (!ts.isIdentifier(token)) {
                    return;
                }
                var declaration = token.parent;
                if (declaration.kind === 165 /* PropertyDeclaration */ &&
                    (!fixedNodes || ts.tryAddToSet(fixedNodes, declaration))) {
                    changeTracker.insertModifierBefore(sourceFile, 134 /* DeclareKeyword */, declaration);
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "addMissingInvocationForDecorator";
            var errorCodes = [ts.Diagnostics._0_accepts_too_few_arguments_to_be_used_as_a_decorator_here_Did_you_mean_to_call_it_first_and_write_0.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return makeChange(t, context.sourceFile, context.span.start); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Call_decorator_expression, fixId, ts.Diagnostics.Add_to_all_uncalled_decorators)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) { return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return makeChange(changes, diag.file, diag.start); }); },
            });
            function makeChange(changeTracker, sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                var decorator = ts.findAncestor(token, ts.isDecorator);
                ts.Debug.assert(!!decorator, "Expected position to be owned by a decorator.");
                var replacement = ts.factory.createCallExpression(decorator.expression, /*typeArguments*/ undefined, /*argumentsArray*/ undefined);
                changeTracker.replaceNode(sourceFile, decorator.expression, replacement);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "addNameToNamelessParameter";
            var errorCodes = [ts.Diagnostics.Parameter_has_a_name_but_no_type_Did_you_mean_0_Colon_1.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return makeChange(t, context.sourceFile, context.span.start); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_parameter_name, fixId, ts.Diagnostics.Add_names_to_all_parameters_without_names)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) { return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return makeChange(changes, diag.file, diag.start); }); },
            });
            function makeChange(changeTracker, sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                if (!ts.isIdentifier(token)) {
                    return ts.Debug.fail("add-name-to-nameless-parameter operates on identifiers, but got a " + ts.Debug.formatSyntaxKind(token.kind));
                }
                var param = token.parent;
                if (!ts.isParameter(param)) {
                    return ts.Debug.fail("Tried to add a parameter name to a non-parameter: " + ts.Debug.formatSyntaxKind(token.kind));
                }
                var i = param.parent.parameters.indexOf(param);
                ts.Debug.assert(!param.type, "Tried to add a parameter name to a parameter that already had one.");
                ts.Debug.assert(i > -1, "Parameter not found in parent parameter list.");
                var replacement = ts.factory.createParameterDeclaration(
                /*decorators*/ undefined, param.modifiers, param.dotDotDotToken, "arg" + i, param.questionToken, ts.factory.createTypeReferenceNode(token, /*typeArguments*/ undefined), param.initializer);
                changeTracker.replaceNode(sourceFile, token, replacement);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "annotateWithTypeFromJSDoc";
            var errorCodes = [ts.Diagnostics.JSDoc_types_may_be_moved_to_TypeScript_types.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var decl = getDeclaration(context.sourceFile, context.span.start);
                    if (!decl)
                        return;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, context.sourceFile, decl); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Annotate_with_type_from_JSDoc, fixId, ts.Diagnostics.Annotate_everything_with_types_from_JSDoc)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var decl = getDeclaration(diag.file, diag.start);
                        if (decl)
                            doChange(changes, diag.file, decl);
                    });
                },
            });
            function getDeclaration(file, pos) {
                var name = ts.getTokenAtPosition(file, pos);
                // For an arrow function with no name, 'name' lands on the first parameter.
                return ts.tryCast(ts.isParameter(name.parent) ? name.parent.parent : name.parent, parameterShouldGetTypeFromJSDoc);
            }
            function parameterShouldGetTypeFromJSDoc(node) {
                return isDeclarationWithType(node) && hasUsableJSDoc(node);
            }
            codefix.parameterShouldGetTypeFromJSDoc = parameterShouldGetTypeFromJSDoc;
            function hasUsableJSDoc(decl) {
                return ts.isFunctionLikeDeclaration(decl)
                    ? decl.parameters.some(hasUsableJSDoc) || (!decl.type && !!ts.getJSDocReturnType(decl))
                    : !decl.type && !!ts.getJSDocType(decl);
            }
            function doChange(changes, sourceFile, decl) {
                if (ts.isFunctionLikeDeclaration(decl) && (ts.getJSDocReturnType(decl) || decl.parameters.some(function (p) { return !!ts.getJSDocType(p); }))) {
                    if (!decl.typeParameters) {
                        var typeParameters = ts.getJSDocTypeParameterDeclarations(decl);
                        if (typeParameters.length)
                            changes.insertTypeParameters(sourceFile, decl, typeParameters);
                    }
                    var needParens = ts.isArrowFunction(decl) && !ts.findChildOfKind(decl, 20 /* OpenParenToken */, sourceFile);
                    if (needParens)
                        changes.insertNodeBefore(sourceFile, ts.first(decl.parameters), ts.factory.createToken(20 /* OpenParenToken */));
                    for (var _i = 0, _a = decl.parameters; _i < _a.length; _i++) {
                        var param = _a[_i];
                        if (!param.type) {
                            var paramType = ts.getJSDocType(param);
                            if (paramType)
                                changes.tryInsertTypeAnnotation(sourceFile, param, transformJSDocType(paramType));
                        }
                    }
                    if (needParens)
                        changes.insertNodeAfter(sourceFile, ts.last(decl.parameters), ts.factory.createToken(21 /* CloseParenToken */));
                    if (!decl.type) {
                        var returnType = ts.getJSDocReturnType(decl);
                        if (returnType)
                            changes.tryInsertTypeAnnotation(sourceFile, decl, transformJSDocType(returnType));
                    }
                }
                else {
                    var jsdocType = ts.Debug.checkDefined(ts.getJSDocType(decl), "A JSDocType for this declaration should exist"); // If not defined, shouldn't have been an error to fix
                    ts.Debug.assert(!decl.type, "The JSDocType decl should have a type"); // If defined, shouldn't have been an error to fix.
                    changes.tryInsertTypeAnnotation(sourceFile, decl, transformJSDocType(jsdocType));
                }
            }
            function isDeclarationWithType(node) {
                return ts.isFunctionLikeDeclaration(node) ||
                    node.kind === 252 /* VariableDeclaration */ ||
                    node.kind === 164 /* PropertySignature */ ||
                    node.kind === 165 /* PropertyDeclaration */;
            }
            function transformJSDocType(node) {
                switch (node.kind) {
                    case 307 /* JSDocAllType */:
                    case 308 /* JSDocUnknownType */:
                        return ts.factory.createTypeReferenceNode("any", ts.emptyArray);
                    case 311 /* JSDocOptionalType */:
                        return transformJSDocOptionalType(node);
                    case 310 /* JSDocNonNullableType */:
                        return transformJSDocType(node.type);
                    case 309 /* JSDocNullableType */:
                        return transformJSDocNullableType(node);
                    case 313 /* JSDocVariadicType */:
                        return transformJSDocVariadicType(node);
                    case 312 /* JSDocFunctionType */:
                        return transformJSDocFunctionType(node);
                    case 176 /* TypeReference */:
                        return transformJSDocTypeReference(node);
                    default:
                        var visited = ts.visitEachChild(node, transformJSDocType, ts.nullTransformationContext);
                        ts.setEmitFlags(visited, 1 /* SingleLine */);
                        return visited;
                }
            }
            function transformJSDocOptionalType(node) {
                return ts.factory.createUnionTypeNode([ts.visitNode(node.type, transformJSDocType), ts.factory.createTypeReferenceNode("undefined", ts.emptyArray)]);
            }
            function transformJSDocNullableType(node) {
                return ts.factory.createUnionTypeNode([ts.visitNode(node.type, transformJSDocType), ts.factory.createTypeReferenceNode("null", ts.emptyArray)]);
            }
            function transformJSDocVariadicType(node) {
                return ts.factory.createArrayTypeNode(ts.visitNode(node.type, transformJSDocType));
            }
            function transformJSDocFunctionType(node) {
                var _a;
                // TODO: This does not properly handle `function(new:C, string)` per https://github.com/google/closure-compiler/wiki/Types-in-the-Closure-Type-System#the-javascript-type-language
                //       however we do handle it correctly in `serializeTypeForDeclaration` in checker.ts
                return ts.factory.createFunctionTypeNode(ts.emptyArray, node.parameters.map(transformJSDocParameter), (_a = node.type) !== null && _a !== void 0 ? _a : ts.factory.createKeywordTypeNode(129 /* AnyKeyword */));
            }
            function transformJSDocParameter(node) {
                var index = node.parent.parameters.indexOf(node);
                var isRest = node.type.kind === 313 /* JSDocVariadicType */ && index === node.parent.parameters.length - 1; // TODO: GH#18217
                var name = node.name || (isRest ? "rest" : "arg" + index);
                var dotdotdot = isRest ? ts.factory.createToken(25 /* DotDotDotToken */) : node.dotDotDotToken;
                return ts.factory.createParameterDeclaration(node.decorators, node.modifiers, dotdotdot, name, node.questionToken, ts.visitNode(node.type, transformJSDocType), node.initializer);
            }
            function transformJSDocTypeReference(node) {
                var name = node.typeName;
                var args = node.typeArguments;
                if (ts.isIdentifier(node.typeName)) {
                    if (ts.isJSDocIndexSignature(node)) {
                        return transformJSDocIndexSignature(node);
                    }
                    var text = node.typeName.text;
                    switch (node.typeName.text) {
                        case "String":
                        case "Boolean":
                        case "Object":
                        case "Number":
                            text = text.toLowerCase();
                            break;
                        case "array":
                        case "date":
                        case "promise":
                            text = text[0].toUpperCase() + text.slice(1);
                            break;
                    }
                    name = ts.factory.createIdentifier(text);
                    if ((text === "Array" || text === "Promise") && !node.typeArguments) {
                        args = ts.factory.createNodeArray([ts.factory.createTypeReferenceNode("any", ts.emptyArray)]);
                    }
                    else {
                        args = ts.visitNodes(node.typeArguments, transformJSDocType);
                    }
                }
                return ts.factory.createTypeReferenceNode(name, args);
            }
            function transformJSDocIndexSignature(node) {
                var index = ts.factory.createParameterDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*dotDotDotToken*/ undefined, node.typeArguments[0].kind === 145 /* NumberKeyword */ ? "n" : "s", 
                /*questionToken*/ undefined, ts.factory.createTypeReferenceNode(node.typeArguments[0].kind === 145 /* NumberKeyword */ ? "number" : "string", []), 
                /*initializer*/ undefined);
                var indexSignature = ts.factory.createTypeLiteralNode([ts.factory.createIndexSignature(/*decorators*/ undefined, /*modifiers*/ undefined, [index], node.typeArguments[1])]);
                ts.setEmitFlags(indexSignature, 1 /* SingleLine */);
                return indexSignature;
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "convertFunctionToEs6Class";
            var errorCodes = [ts.Diagnostics.This_constructor_function_may_be_converted_to_a_class_declaration.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) {
                        return doChange(t, context.sourceFile, context.span.start, context.program.getTypeChecker(), context.preferences, context.program.getCompilerOptions());
                    });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Convert_function_to_an_ES2015_class, fixId, ts.Diagnostics.Convert_all_constructor_functions_to_classes)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, err) {
                        return doChange(changes, err.file, err.start, context.program.getTypeChecker(), context.preferences, context.program.getCompilerOptions());
                    });
                },
            });
            function doChange(changes, sourceFile, position, checker, preferences, compilerOptions) {
                var ctorSymbol = checker.getSymbolAtLocation(ts.getTokenAtPosition(sourceFile, position));
                if (!ctorSymbol || !ctorSymbol.valueDeclaration || !(ctorSymbol.flags & (16 /* Function */ | 3 /* Variable */))) {
                    // Bad input
                    return undefined;
                }
                var ctorDeclaration = ctorSymbol.valueDeclaration;
                if (ts.isFunctionDeclaration(ctorDeclaration)) {
                    changes.replaceNode(sourceFile, ctorDeclaration, createClassFromFunctionDeclaration(ctorDeclaration));
                }
                else if (ts.isVariableDeclaration(ctorDeclaration)) {
                    var classDeclaration = createClassFromVariableDeclaration(ctorDeclaration);
                    if (!classDeclaration) {
                        return undefined;
                    }
                    var ancestor = ctorDeclaration.parent.parent;
                    if (ts.isVariableDeclarationList(ctorDeclaration.parent) && ctorDeclaration.parent.declarations.length > 1) {
                        changes.delete(sourceFile, ctorDeclaration);
                        changes.insertNodeAfter(sourceFile, ancestor, classDeclaration);
                    }
                    else {
                        changes.replaceNode(sourceFile, ancestor, classDeclaration);
                    }
                }
                function createClassElementsFromSymbol(symbol) {
                    var memberElements = [];
                    // all instance members are stored in the "member" array of symbol
                    if (symbol.members) {
                        symbol.members.forEach(function (member, key) {
                            if (key === "constructor" && member.valueDeclaration) {
                                // fn.prototype.constructor = fn
                                changes.delete(sourceFile, member.valueDeclaration.parent);
                                return;
                            }
                            var memberElement = createClassElement(member, /*modifiers*/ undefined);
                            if (memberElement) {
                                memberElements.push.apply(memberElements, memberElement);
                            }
                        });
                    }
                    // all static members are stored in the "exports" array of symbol
                    if (symbol.exports) {
                        symbol.exports.forEach(function (member) {
                            if (member.name === "prototype" && member.declarations) {
                                var firstDeclaration = member.declarations[0];
                                // only one "x.prototype = { ... }" will pass
                                if (member.declarations.length === 1 &&
                                    ts.isPropertyAccessExpression(firstDeclaration) &&
                                    ts.isBinaryExpression(firstDeclaration.parent) &&
                                    firstDeclaration.parent.operatorToken.kind === 63 /* EqualsToken */ &&
                                    ts.isObjectLiteralExpression(firstDeclaration.parent.right)) {
                                    var prototypes = firstDeclaration.parent.right;
                                    var memberElement = createClassElement(prototypes.symbol, /** modifiers */ undefined);
                                    if (memberElement) {
                                        memberElements.push.apply(memberElements, memberElement);
                                    }
                                }
                            }
                            else {
                                var memberElement = createClassElement(member, [ts.factory.createToken(124 /* StaticKeyword */)]);
                                if (memberElement) {
                                    memberElements.push.apply(memberElements, memberElement);
                                }
                            }
                        });
                    }
                    return memberElements;
                    function shouldConvertDeclaration(_target, source) {
                        // Right now the only thing we can convert are function expressions, get/set accessors and methods
                        // other values like normal value fields ({a: 1}) shouldn't get transformed.
                        // We can update this once ES public class properties are available.
                        if (ts.isAccessExpression(_target)) {
                            if (ts.isPropertyAccessExpression(_target) && isConstructorAssignment(_target))
                                return true;
                            return ts.isFunctionLike(source);
                        }
                        else {
                            return ts.every(_target.properties, function (property) {
                                // a() {}
                                if (ts.isMethodDeclaration(property) || ts.isGetOrSetAccessorDeclaration(property))
                                    return true;
                                // a: function() {}
                                if (ts.isPropertyAssignment(property) && ts.isFunctionExpression(property.initializer) && !!property.name)
                                    return true;
                                // x.prototype.constructor = fn
                                if (isConstructorAssignment(property))
                                    return true;
                                return false;
                            });
                        }
                    }
                    function createClassElement(symbol, modifiers) {
                        // Right now the only thing we can convert are function expressions, which are marked as methods
                        // or { x: y } type prototype assignments, which are marked as ObjectLiteral
                        var members = [];
                        if (!(symbol.flags & 8192 /* Method */) && !(symbol.flags & 4096 /* ObjectLiteral */)) {
                            return members;
                        }
                        var memberDeclaration = symbol.valueDeclaration;
                        var assignmentBinaryExpression = memberDeclaration.parent;
                        var assignmentExpr = assignmentBinaryExpression.right;
                        if (!shouldConvertDeclaration(memberDeclaration, assignmentExpr)) {
                            return members;
                        }
                        // delete the entire statement if this expression is the sole expression to take care of the semicolon at the end
                        var nodeToDelete = assignmentBinaryExpression.parent && assignmentBinaryExpression.parent.kind === 236 /* ExpressionStatement */
                            ? assignmentBinaryExpression.parent : assignmentBinaryExpression;
                        changes.delete(sourceFile, nodeToDelete);
                        if (!assignmentExpr) {
                            members.push(ts.factory.createPropertyDeclaration([], modifiers, symbol.name, /*questionToken*/ undefined, 
                            /*type*/ undefined, /*initializer*/ undefined));
                            return members;
                        }
                        // f.x = expr
                        if (ts.isAccessExpression(memberDeclaration) && (ts.isFunctionExpression(assignmentExpr) || ts.isArrowFunction(assignmentExpr))) {
                            var quotePreference = ts.getQuotePreference(sourceFile, preferences);
                            var name = tryGetPropertyName(memberDeclaration, compilerOptions, quotePreference);
                            if (name) {
                                return createFunctionLikeExpressionMember(members, assignmentExpr, name);
                            }
                            return members;
                        }
                        // f.prototype = { ... }
                        else if (ts.isObjectLiteralExpression(assignmentExpr)) {
                            return ts.flatMap(assignmentExpr.properties, function (property) {
                                if (ts.isMethodDeclaration(property) || ts.isGetOrSetAccessorDeclaration(property)) {
                                    // MethodDeclaration and AccessorDeclaration can appear in a class directly
                                    return members.concat(property);
                                }
                                if (ts.isPropertyAssignment(property) && ts.isFunctionExpression(property.initializer)) {
                                    return createFunctionLikeExpressionMember(members, property.initializer, property.name);
                                }
                                // Drop constructor assignments
                                if (isConstructorAssignment(property))
                                    return members;
                                return [];
                            });
                        }
                        else {
                            // Don't try to declare members in JavaScript files
                            if (ts.isSourceFileJS(sourceFile))
                                return members;
                            if (!ts.isPropertyAccessExpression(memberDeclaration))
                                return members;
                            var prop = ts.factory.createPropertyDeclaration(/*decorators*/ undefined, modifiers, memberDeclaration.name, /*questionToken*/ undefined, /*type*/ undefined, assignmentExpr);
                            ts.copyLeadingComments(assignmentBinaryExpression.parent, prop, sourceFile);
                            members.push(prop);
                            return members;
                        }
                        function createFunctionLikeExpressionMember(members, expression, name) {
                            if (ts.isFunctionExpression(expression))
                                return createFunctionExpressionMember(members, expression, name);
                            else
                                return createArrowFunctionExpressionMember(members, expression, name);
                        }
                        function createFunctionExpressionMember(members, functionExpression, name) {
                            var fullModifiers = ts.concatenate(modifiers, getModifierKindFromSource(functionExpression, 130 /* AsyncKeyword */));
                            var method = ts.factory.createMethodDeclaration(/*decorators*/ undefined, fullModifiers, /*asteriskToken*/ undefined, name, /*questionToken*/ undefined, 
                            /*typeParameters*/ undefined, functionExpression.parameters, /*type*/ undefined, functionExpression.body);
                            ts.copyLeadingComments(assignmentBinaryExpression, method, sourceFile);
                            return members.concat(method);
                        }
                        function createArrowFunctionExpressionMember(members, arrowFunction, name) {
                            var arrowFunctionBody = arrowFunction.body;
                            var bodyBlock;
                            // case 1: () => { return [1,2,3] }
                            if (arrowFunctionBody.kind === 233 /* Block */) {
                                bodyBlock = arrowFunctionBody;
                            }
                            // case 2: () => [1,2,3]
                            else {
                                bodyBlock = ts.factory.createBlock([ts.factory.createReturnStatement(arrowFunctionBody)]);
                            }
                            var fullModifiers = ts.concatenate(modifiers, getModifierKindFromSource(arrowFunction, 130 /* AsyncKeyword */));
                            var method = ts.factory.createMethodDeclaration(/*decorators*/ undefined, fullModifiers, /*asteriskToken*/ undefined, name, /*questionToken*/ undefined, 
                            /*typeParameters*/ undefined, arrowFunction.parameters, /*type*/ undefined, bodyBlock);
                            ts.copyLeadingComments(assignmentBinaryExpression, method, sourceFile);
                            return members.concat(method);
                        }
                    }
                }
                function createClassFromVariableDeclaration(node) {
                    var initializer = node.initializer;
                    if (!initializer || !ts.isFunctionExpression(initializer) || !ts.isIdentifier(node.name)) {
                        return undefined;
                    }
                    var memberElements = createClassElementsFromSymbol(node.symbol);
                    if (initializer.body) {
                        memberElements.unshift(ts.factory.createConstructorDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, initializer.parameters, initializer.body));
                    }
                    var modifiers = getModifierKindFromSource(node.parent.parent, 93 /* ExportKeyword */);
                    var cls = ts.factory.createClassDeclaration(/*decorators*/ undefined, modifiers, node.name, 
                    /*typeParameters*/ undefined, /*heritageClauses*/ undefined, memberElements);
                    // Don't call copyComments here because we'll already leave them in place
                    return cls;
                }
                function createClassFromFunctionDeclaration(node) {
                    var memberElements = createClassElementsFromSymbol(ctorSymbol);
                    if (node.body) {
                        memberElements.unshift(ts.factory.createConstructorDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, node.parameters, node.body));
                    }
                    var modifiers = getModifierKindFromSource(node, 93 /* ExportKeyword */);
                    var cls = ts.factory.createClassDeclaration(/*decorators*/ undefined, modifiers, node.name, 
                    /*typeParameters*/ undefined, /*heritageClauses*/ undefined, memberElements);
                    // Don't call copyComments here because we'll already leave them in place
                    return cls;
                }
            }
            function getModifierKindFromSource(source, kind) {
                return ts.filter(source.modifiers, function (modifier) { return modifier.kind === kind; });
            }
            function isConstructorAssignment(x) {
                if (!x.name)
                    return false;
                if (ts.isIdentifier(x.name) && x.name.text === "constructor")
                    return true;
                return false;
            }
            function tryGetPropertyName(node, compilerOptions, quotePreference) {
                if (ts.isPropertyAccessExpression(node)) {
                    return node.name;
                }
                var propName = node.argumentExpression;
                if (ts.isNumericLiteral(propName)) {
                    return propName;
                }
                if (ts.isStringLiteralLike(propName)) {
                    return ts.isIdentifierText(propName.text, compilerOptions.target) ? ts.factory.createIdentifier(propName.text)
                        : ts.isNoSubstitutionTemplateLiteral(propName) ? ts.factory.createStringLiteral(propName.text, quotePreference === 0 /* Single */)
                            : propName;
                }
                return undefined;
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "convertToAsyncFunction";
            var errorCodes = [ts.Diagnostics.This_may_be_converted_to_an_async_function.code];
            var codeActionSucceeded = true;
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    codeActionSucceeded = true;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return convertToAsyncFunction(t, context.sourceFile, context.span.start, context.program.getTypeChecker()); });
                    return codeActionSucceeded ? [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Convert_to_async_function, fixId, ts.Diagnostics.Convert_all_to_async_functions)] : [];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) { return codefix.codeFixAll(context, errorCodes, function (changes, err) { return convertToAsyncFunction(changes, err.file, err.start, context.program.getTypeChecker()); }); },
            });
            var SynthBindingNameKind;
            (function (SynthBindingNameKind) {
                SynthBindingNameKind[SynthBindingNameKind["Identifier"] = 0] = "Identifier";
                SynthBindingNameKind[SynthBindingNameKind["BindingPattern"] = 1] = "BindingPattern";
            })(SynthBindingNameKind || (SynthBindingNameKind = {}));
            function convertToAsyncFunction(changes, sourceFile, position, checker) {
                // get the function declaration - returns a promise
                var tokenAtPosition = ts.getTokenAtPosition(sourceFile, position);
                var functionToConvert;
                // if the parent of a FunctionLikeDeclaration is a variable declaration, the convertToAsync diagnostic will be reported on the variable name
                if (ts.isIdentifier(tokenAtPosition) && ts.isVariableDeclaration(tokenAtPosition.parent) &&
                    tokenAtPosition.parent.initializer && ts.isFunctionLikeDeclaration(tokenAtPosition.parent.initializer)) {
                    functionToConvert = tokenAtPosition.parent.initializer;
                }
                else {
                    functionToConvert = ts.tryCast(ts.getContainingFunction(ts.getTokenAtPosition(sourceFile, position)), ts.canBeConvertedToAsync);
                }
                if (!functionToConvert) {
                    return;
                }
                var synthNamesMap = new ts.Map();
                var isInJavascript = ts.isInJSFile(functionToConvert);
                var setOfExpressionsToReturn = getAllPromiseExpressionsToReturn(functionToConvert, checker);
                var functionToConvertRenamed = renameCollidingVarNames(functionToConvert, checker, synthNamesMap);
                if (!ts.returnsPromise(functionToConvertRenamed, checker)) {
                    return;
                }
                var returnStatements = functionToConvertRenamed.body && ts.isBlock(functionToConvertRenamed.body) ? getReturnStatementsWithPromiseHandlers(functionToConvertRenamed.body, checker) : ts.emptyArray;
                var transformer = { checker: checker, synthNamesMap: synthNamesMap, setOfExpressionsToReturn: setOfExpressionsToReturn, isInJSFile: isInJavascript };
                if (!returnStatements.length) {
                    return;
                }
                var pos = functionToConvert.modifiers ? functionToConvert.modifiers.end :
                    functionToConvert.decorators ? ts.skipTrivia(sourceFile.text, functionToConvert.decorators.end) :
                        functionToConvert.getStart(sourceFile);
                var options = functionToConvert.modifiers ? { prefix: " " } : { suffix: " " };
                changes.insertModifierAt(sourceFile, pos, 130 /* AsyncKeyword */, options);
                var _loop_13 = function (returnStatement) {
                    ts.forEachChild(returnStatement, function visit(node) {
                        if (ts.isCallExpression(node)) {
                            var newNodes = transformExpression(node, transformer);
                            changes.replaceNodeWithNodes(sourceFile, returnStatement, newNodes);
                        }
                        else if (!ts.isFunctionLike(node)) {
                            ts.forEachChild(node, visit);
                        }
                    });
                };
                for (var _i = 0, returnStatements_1 = returnStatements; _i < returnStatements_1.length; _i++) {
                    var returnStatement = returnStatements_1[_i];
                    _loop_13(returnStatement);
                }
            }
            function getReturnStatementsWithPromiseHandlers(body, checker) {
                var res = [];
                ts.forEachReturnStatement(body, function (ret) {
                    if (ts.isReturnStatementWithFixablePromiseHandler(ret, checker))
                        res.push(ret);
                });
                return res;
            }
            /*
                Finds all of the expressions of promise type that should not be saved in a variable during the refactor
            */
            function getAllPromiseExpressionsToReturn(func, checker) {
                if (!func.body) {
                    return new ts.Set();
                }
                var setOfExpressionsToReturn = new ts.Set();
                ts.forEachChild(func.body, function visit(node) {
                    if (isPromiseReturningCallExpression(node, checker, "then")) {
                        setOfExpressionsToReturn.add(ts.getNodeId(node));
                        ts.forEach(node.arguments, visit);
                    }
                    else if (isPromiseReturningCallExpression(node, checker, "catch")) {
                        setOfExpressionsToReturn.add(ts.getNodeId(node));
                        // if .catch() is the last call in the chain, move leftward in the chain until we hit something else that should be returned
                        ts.forEachChild(node, visit);
                    }
                    else if (isPromiseTypedExpression(node, checker)) {
                        setOfExpressionsToReturn.add(ts.getNodeId(node));
                        // don't recurse here, since we won't refactor any children or arguments of the expression
                    }
                    else {
                        ts.forEachChild(node, visit);
                    }
                });
                return setOfExpressionsToReturn;
            }
            function isPromiseReturningCallExpression(node, checker, name) {
                if (!ts.isCallExpression(node))
                    return false;
                var isExpressionOfName = ts.hasPropertyAccessExpressionWithName(node, name);
                var nodeType = isExpressionOfName && checker.getTypeAtLocation(node);
                return !!(nodeType && checker.getPromisedTypeOfPromise(nodeType));
            }
            function isPromiseTypedExpression(node, checker) {
                if (!ts.isExpression(node))
                    return false;
                return !!checker.getPromisedTypeOfPromise(checker.getTypeAtLocation(node));
            }
            /*
                Renaming of identifiers may be necessary as the refactor changes scopes -
                This function collects all existing identifier names and names of identifiers that will be created in the refactor.
                It then checks for any collisions and renames them through getSynthesizedDeepClone
            */
            function renameCollidingVarNames(nodeToRename, checker, synthNamesMap) {
                var identsToRenameMap = new ts.Map(); // key is the symbol id
                var collidingSymbolMap = ts.createMultiMap();
                ts.forEachChild(nodeToRename, function visit(node) {
                    if (!ts.isIdentifier(node)) {
                        ts.forEachChild(node, visit);
                        return;
                    }
                    var symbol = checker.getSymbolAtLocation(node);
                    if (symbol) {
                        var type = checker.getTypeAtLocation(node);
                        // Note - the choice of the last call signature is arbitrary
                        var lastCallSignature = getLastCallSignature(type, checker);
                        var symbolIdString = ts.getSymbolId(symbol).toString();
                        // If the identifier refers to a function, we want to add the new synthesized variable for the declaration. Example:
                        //   fetch('...').then(response => { ... })
                        // will eventually become
                        //   const response = await fetch('...')
                        // so we push an entry for 'response'.
                        if (lastCallSignature && !ts.isParameter(node.parent) && !ts.isFunctionLikeDeclaration(node.parent) && !synthNamesMap.has(symbolIdString)) {
                            var firstParameter = ts.firstOrUndefined(lastCallSignature.parameters);
                            var ident = (firstParameter === null || firstParameter === void 0 ? void 0 : firstParameter.valueDeclaration)
                                && ts.isParameter(firstParameter.valueDeclaration)
                                && ts.tryCast(firstParameter.valueDeclaration.name, ts.isIdentifier)
                                || ts.factory.createUniqueName("result", 16 /* Optimistic */);
                            var synthName = getNewNameIfConflict(ident, collidingSymbolMap);
                            synthNamesMap.set(symbolIdString, synthName);
                            collidingSymbolMap.add(ident.text, symbol);
                        }
                        // We only care about identifiers that are parameters, variable declarations, or binding elements
                        else if (node.parent && (ts.isParameter(node.parent) || ts.isVariableDeclaration(node.parent) || ts.isBindingElement(node.parent))) {
                            var originalName = node.text;
                            var collidingSymbols = collidingSymbolMap.get(originalName);
                            // if the identifier name conflicts with a different identifier that we've already seen
                            if (collidingSymbols && collidingSymbols.some(function (prevSymbol) { return prevSymbol !== symbol; })) {
                                var newName = getNewNameIfConflict(node, collidingSymbolMap);
                                identsToRenameMap.set(symbolIdString, newName.identifier);
                                synthNamesMap.set(symbolIdString, newName);
                                collidingSymbolMap.add(originalName, symbol);
                            }
                            else {
                                var identifier = ts.getSynthesizedDeepClone(node);
                                synthNamesMap.set(symbolIdString, createSynthIdentifier(identifier));
                                collidingSymbolMap.add(originalName, symbol);
                            }
                        }
                    }
                });
                return ts.getSynthesizedDeepCloneWithReplacements(nodeToRename, /*includeTrivia*/ true, function (original) {
                    if (ts.isBindingElement(original) && ts.isIdentifier(original.name) && ts.isObjectBindingPattern(original.parent)) {
                        var symbol = checker.getSymbolAtLocation(original.name);
                        var renameInfo = symbol && identsToRenameMap.get(String(ts.getSymbolId(symbol)));
                        if (renameInfo && renameInfo.text !== (original.name || original.propertyName).getText()) {
                            return ts.factory.createBindingElement(original.dotDotDotToken, original.propertyName || original.name, renameInfo, original.initializer);
                        }
                    }
                    else if (ts.isIdentifier(original)) {
                        var symbol = checker.getSymbolAtLocation(original);
                        var renameInfo = symbol && identsToRenameMap.get(String(ts.getSymbolId(symbol)));
                        if (renameInfo) {
                            return ts.factory.createIdentifier(renameInfo.text);
                        }
                    }
                });
            }
            function getNewNameIfConflict(name, originalNames) {
                var numVarsSameName = (originalNames.get(name.text) || ts.emptyArray).length;
                var identifier = numVarsSameName === 0 ? name : ts.factory.createIdentifier(name.text + "_" + numVarsSameName);
                return createSynthIdentifier(identifier);
            }
            function silentFail() {
                codeActionSucceeded = false;
                return ts.emptyArray;
            }
            // dispatch function to recursively build the refactoring
            // should be kept up to date with isFixablePromiseHandler in suggestionDiagnostics.ts
            function transformExpression(node, transformer, prevArgName) {
                if (isPromiseReturningCallExpression(node, transformer.checker, "then")) {
                    if (node.arguments.length === 0)
                        return silentFail();
                    return transformThen(node, transformer, prevArgName);
                }
                if (isPromiseReturningCallExpression(node, transformer.checker, "catch")) {
                    return transformCatch(node, transformer, prevArgName);
                }
                if (ts.isPropertyAccessExpression(node)) {
                    return transformExpression(node.expression, transformer, prevArgName);
                }
                var nodeType = transformer.checker.getTypeAtLocation(node);
                if (nodeType && transformer.checker.getPromisedTypeOfPromise(nodeType)) {
                    ts.Debug.assertNode(node.original.parent, ts.isPropertyAccessExpression);
                    return transformPromiseExpressionOfPropertyAccess(node, transformer, prevArgName);
                }
                return silentFail();
            }
            function transformCatch(node, transformer, prevArgName) {
                var func = ts.singleOrUndefined(node.arguments);
                var argName = func ? getArgBindingName(func, transformer) : undefined;
                var possibleNameForVarDecl;
                // If there is another call in the chain after the .catch() we are transforming, we will need to save the result of both paths (try block and catch block)
                // To do this, we will need to synthesize a variable that we were not aware of while we were adding identifiers to the synthNamesMap
                // We will use the prevArgName and then update the synthNamesMap with a new variable name for the next transformation step
                if (prevArgName && !shouldReturn(node, transformer)) {
                    if (isSynthIdentifier(prevArgName)) {
                        possibleNameForVarDecl = prevArgName;
                        transformer.synthNamesMap.forEach(function (val, key) {
                            if (val.identifier.text === prevArgName.identifier.text) {
                                var newSynthName = createUniqueSynthName(prevArgName);
                                transformer.synthNamesMap.set(key, newSynthName);
                            }
                        });
                    }
                    else {
                        possibleNameForVarDecl = createSynthIdentifier(ts.factory.createUniqueName("result", 16 /* Optimistic */), prevArgName.types);
                    }
                    // We are about to write a 'let' variable declaration, but `transformExpression` for both
                    // the try block and catch block will assign to this name. Setting this flag indicates
                    // that future assignments should be written as `name = value` instead of `const name = value`.
                    possibleNameForVarDecl.hasBeenDeclared = true;
                }
                var tryBlock = ts.factory.createBlock(transformExpression(node.expression, transformer, possibleNameForVarDecl));
                var transformationBody = func ? getTransformationBody(func, possibleNameForVarDecl, argName, node, transformer) : ts.emptyArray;
                var catchArg = argName ? isSynthIdentifier(argName) ? argName.identifier.text : argName.bindingPattern : "e";
                var catchVariableDeclaration = ts.factory.createVariableDeclaration(catchArg);
                var catchClause = ts.factory.createCatchClause(catchVariableDeclaration, ts.factory.createBlock(transformationBody));
                // In order to avoid an implicit any, we will synthesize a type for the declaration using the unions of the types of both paths (try block and catch block)
                var varDeclList;
                var varDeclIdentifier;
                if (possibleNameForVarDecl && !shouldReturn(node, transformer)) {
                    varDeclIdentifier = ts.getSynthesizedDeepClone(possibleNameForVarDecl.identifier);
                    var typeArray = possibleNameForVarDecl.types;
                    var unionType = transformer.checker.getUnionType(typeArray, 2 /* Subtype */);
                    var unionTypeNode = transformer.isInJSFile ? undefined : transformer.checker.typeToTypeNode(unionType, /*enclosingDeclaration*/ undefined, /*flags*/ undefined);
                    var varDecl = [ts.factory.createVariableDeclaration(varDeclIdentifier, /*exclamationToken*/ undefined, unionTypeNode)];
                    varDeclList = ts.factory.createVariableStatement(/*modifiers*/ undefined, ts.factory.createVariableDeclarationList(varDecl, 1 /* Let */));
                }
                var tryStatement = ts.factory.createTryStatement(tryBlock, catchClause, /*finallyBlock*/ undefined);
                var destructuredResult = prevArgName && varDeclIdentifier && isSynthBindingPattern(prevArgName)
                    && ts.factory.createVariableStatement(/*modifiers*/ undefined, ts.factory.createVariableDeclarationList([ts.factory.createVariableDeclaration(ts.getSynthesizedDeepClone(prevArgName.bindingPattern), /*exclamationToken*/ undefined, /*type*/ undefined, varDeclIdentifier)], 2 /* Const */));
                return ts.compact([varDeclList, tryStatement, destructuredResult]);
            }
            function createUniqueSynthName(prevArgName) {
                var renamedPrevArg = ts.factory.createUniqueName(prevArgName.identifier.text, 16 /* Optimistic */);
                return createSynthIdentifier(renamedPrevArg);
            }
            function transformThen(node, transformer, prevArgName) {
                var _a = node.arguments, onFulfilled = _a[0], onRejected = _a[1];
                var onFulfilledArgumentName = getArgBindingName(onFulfilled, transformer);
                var transformationBody = getTransformationBody(onFulfilled, prevArgName, onFulfilledArgumentName, node, transformer);
                if (onRejected) {
                    var onRejectedArgumentName = getArgBindingName(onRejected, transformer);
                    var tryBlock = ts.factory.createBlock(transformExpression(node.expression, transformer, onFulfilledArgumentName).concat(transformationBody));
                    var transformationBody2 = getTransformationBody(onRejected, prevArgName, onRejectedArgumentName, node, transformer);
                    var catchArg = onRejectedArgumentName ? isSynthIdentifier(onRejectedArgumentName) ? onRejectedArgumentName.identifier.text : onRejectedArgumentName.bindingPattern : "e";
                    var catchVariableDeclaration = ts.factory.createVariableDeclaration(catchArg);
                    var catchClause = ts.factory.createCatchClause(catchVariableDeclaration, ts.factory.createBlock(transformationBody2));
                    return [ts.factory.createTryStatement(tryBlock, catchClause, /* finallyBlock */ undefined)];
                }
                return transformExpression(node.expression, transformer, onFulfilledArgumentName).concat(transformationBody);
            }
            /**
             * Transforms the 'x' part of `x.then(...)`, or the 'y()' part of `y().catch(...)`, where 'x' and 'y()' are Promises.
             */
            function transformPromiseExpressionOfPropertyAccess(node, transformer, prevArgName) {
                if (shouldReturn(node, transformer)) {
                    return [ts.factory.createReturnStatement(ts.getSynthesizedDeepClone(node))];
                }
                return createVariableOrAssignmentOrExpressionStatement(prevArgName, ts.factory.createAwaitExpression(node), /*typeAnnotation*/ undefined);
            }
            function createVariableOrAssignmentOrExpressionStatement(variableName, rightHandSide, typeAnnotation) {
                if (!variableName || isEmptyBindingName(variableName)) {
                    // if there's no argName to assign to, there still might be side effects
                    return [ts.factory.createExpressionStatement(rightHandSide)];
                }
                if (isSynthIdentifier(variableName) && variableName.hasBeenDeclared) {
                    // if the variable has already been declared, we don't need "let" or "const"
                    return [ts.factory.createExpressionStatement(ts.factory.createAssignment(ts.getSynthesizedDeepClone(variableName.identifier), rightHandSide))];
                }
                return [
                    ts.factory.createVariableStatement(
                    /*modifiers*/ undefined, ts.factory.createVariableDeclarationList([
                        ts.factory.createVariableDeclaration(ts.getSynthesizedDeepClone(getNode(variableName)), 
                        /*exclamationToken*/ undefined, typeAnnotation, rightHandSide)
                    ], 2 /* Const */))
                ];
            }
            function maybeAnnotateAndReturn(expressionToReturn, typeAnnotation) {
                if (typeAnnotation && expressionToReturn) {
                    var name = ts.factory.createUniqueName("result", 16 /* Optimistic */);
                    return __spreadArray(__spreadArray([], createVariableOrAssignmentOrExpressionStatement(createSynthIdentifier(name), expressionToReturn, typeAnnotation), true), [
                        ts.factory.createReturnStatement(name)
                    ], false);
                }
                return [ts.factory.createReturnStatement(expressionToReturn)];
            }
            // should be kept up to date with isFixablePromiseArgument in suggestionDiagnostics.ts
            function getTransformationBody(func, prevArgName, argName, parent, transformer) {
                var _a, _b, _c, _d, _e;
                switch (func.kind) {
                    case 104 /* NullKeyword */:
                        // do not produce a transformed statement for a null argument
                        break;
                    case 204 /* PropertyAccessExpression */:
                    case 79 /* Identifier */: // identifier includes undefined
                        if (!argName) {
                            // undefined was argument passed to promise handler
                            break;
                        }
                        var synthCall = ts.factory.createCallExpression(ts.getSynthesizedDeepClone(func), /*typeArguments*/ undefined, isSynthIdentifier(argName) ? [argName.identifier] : []);
                        if (shouldReturn(parent, transformer)) {
                            return maybeAnnotateAndReturn(synthCall, (_a = parent.typeArguments) === null || _a === void 0 ? void 0 : _a[0]);
                        }
                        var type = transformer.checker.getTypeAtLocation(func);
                        var callSignatures = transformer.checker.getSignaturesOfType(type, 0 /* Call */);
                        if (!callSignatures.length) {
                            // if identifier in handler has no call signatures, it's invalid
                            return silentFail();
                        }
                        var returnType = callSignatures[0].getReturnType();
                        var varDeclOrAssignment = createVariableOrAssignmentOrExpressionStatement(prevArgName, ts.factory.createAwaitExpression(synthCall), (_b = parent.typeArguments) === null || _b === void 0 ? void 0 : _b[0]);
                        if (prevArgName) {
                            prevArgName.types.push(returnType);
                        }
                        return varDeclOrAssignment;
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */: {
                        var funcBody = func.body;
                        var returnType_1 = (_c = getLastCallSignature(transformer.checker.getTypeAtLocation(func), transformer.checker)) === null || _c === void 0 ? void 0 : _c.getReturnType();
                        // Arrow functions with block bodies { } will enter this control flow
                        if (ts.isBlock(funcBody)) {
                            var refactoredStmts = [];
                            var seenReturnStatement = false;
                            for (var _i = 0, _f = funcBody.statements; _i < _f.length; _i++) {
                                var statement = _f[_i];
                                if (ts.isReturnStatement(statement)) {
                                    seenReturnStatement = true;
                                    if (ts.isReturnStatementWithFixablePromiseHandler(statement, transformer.checker)) {
                                        refactoredStmts = refactoredStmts.concat(getInnerTransformationBody(transformer, [statement], prevArgName));
                                    }
                                    else {
                                        var possiblyAwaitedRightHandSide = returnType_1 && statement.expression ? getPossiblyAwaitedRightHandSide(transformer.checker, returnType_1, statement.expression) : statement.expression;
                                        refactoredStmts.push.apply(refactoredStmts, maybeAnnotateAndReturn(possiblyAwaitedRightHandSide, (_d = parent.typeArguments) === null || _d === void 0 ? void 0 : _d[0]));
                                    }
                                }
                                else {
                                    refactoredStmts.push(statement);
                                }
                            }
                            return shouldReturn(parent, transformer)
                                ? refactoredStmts.map(function (s) { return ts.getSynthesizedDeepClone(s); })
                                : removeReturns(refactoredStmts, prevArgName, transformer, seenReturnStatement);
                        }
                        else {
                            var innerRetStmts = ts.isFixablePromiseHandler(funcBody, transformer.checker) ? [ts.factory.createReturnStatement(funcBody)] : ts.emptyArray;
                            var innerCbBody = getInnerTransformationBody(transformer, innerRetStmts, prevArgName);
                            if (innerCbBody.length > 0) {
                                return innerCbBody;
                            }
                            if (returnType_1) {
                                var possiblyAwaitedRightHandSide = getPossiblyAwaitedRightHandSide(transformer.checker, returnType_1, funcBody);
                                if (!shouldReturn(parent, transformer)) {
                                    var transformedStatement = createVariableOrAssignmentOrExpressionStatement(prevArgName, possiblyAwaitedRightHandSide, /*typeAnnotation*/ undefined);
                                    if (prevArgName) {
                                        prevArgName.types.push(returnType_1);
                                    }
                                    return transformedStatement;
                                }
                                else {
                                    return maybeAnnotateAndReturn(possiblyAwaitedRightHandSide, (_e = parent.typeArguments) === null || _e === void 0 ? void 0 : _e[0]);
                                }
                            }
                            else {
                                return silentFail();
                            }
                        }
                    }
                    default:
                        // If no cases apply, we've found a transformation body we don't know how to handle, so the refactoring should no-op to avoid deleting code.
                        return silentFail();
                }
                return ts.emptyArray;
            }
            function getPossiblyAwaitedRightHandSide(checker, type, expr) {
                var rightHandSide = ts.getSynthesizedDeepClone(expr);
                return !!checker.getPromisedTypeOfPromise(type) ? ts.factory.createAwaitExpression(rightHandSide) : rightHandSide;
            }
            function getLastCallSignature(type, checker) {
                var callSignatures = checker.getSignaturesOfType(type, 0 /* Call */);
                return ts.lastOrUndefined(callSignatures);
            }
            function removeReturns(stmts, prevArgName, transformer, seenReturnStatement) {
                var ret = [];
                for (var _i = 0, stmts_1 = stmts; _i < stmts_1.length; _i++) {
                    var stmt = stmts_1[_i];
                    if (ts.isReturnStatement(stmt)) {
                        if (stmt.expression) {
                            var possiblyAwaitedExpression = isPromiseTypedExpression(stmt.expression, transformer.checker) ? ts.factory.createAwaitExpression(stmt.expression) : stmt.expression;
                            if (prevArgName === undefined) {
                                ret.push(ts.factory.createExpressionStatement(possiblyAwaitedExpression));
                            }
                            else {
                                ret.push(ts.factory.createVariableStatement(/*modifiers*/ undefined, (ts.factory.createVariableDeclarationList([ts.factory.createVariableDeclaration(getNode(prevArgName), /*exclamationToken*/ undefined, /*type*/ undefined, possiblyAwaitedExpression)], 2 /* Const */))));
                            }
                        }
                    }
                    else {
                        ret.push(ts.getSynthesizedDeepClone(stmt));
                    }
                }
                // if block has no return statement, need to define prevArgName as undefined to prevent undeclared variables
                if (!seenReturnStatement && prevArgName !== undefined) {
                    ret.push(ts.factory.createVariableStatement(/*modifiers*/ undefined, (ts.factory.createVariableDeclarationList([ts.factory.createVariableDeclaration(getNode(prevArgName), /*exclamationToken*/ undefined, /*type*/ undefined, ts.factory.createIdentifier("undefined"))], 2 /* Const */))));
                }
                return ret;
            }
            function getInnerTransformationBody(transformer, innerRetStmts, prevArgName) {
                var innerCbBody = [];
                for (var _i = 0, innerRetStmts_1 = innerRetStmts; _i < innerRetStmts_1.length; _i++) {
                    var stmt = innerRetStmts_1[_i];
                    ts.forEachChild(stmt, function visit(node) {
                        if (ts.isCallExpression(node)) {
                            var temp = transformExpression(node, transformer, prevArgName);
                            innerCbBody = innerCbBody.concat(temp);
                            if (innerCbBody.length > 0) {
                                return;
                            }
                        }
                        else if (!ts.isFunctionLike(node)) {
                            ts.forEachChild(node, visit);
                        }
                    });
                }
                return innerCbBody;
            }
            function getArgBindingName(funcNode, transformer) {
                var types = [];
                var name;
                if (ts.isFunctionLikeDeclaration(funcNode)) {
                    if (funcNode.parameters.length > 0) {
                        var param = funcNode.parameters[0].name;
                        name = getMappedBindingNameOrDefault(param);
                    }
                }
                else if (ts.isIdentifier(funcNode)) {
                    name = getMapEntryOrDefault(funcNode);
                }
                else if (ts.isPropertyAccessExpression(funcNode) && ts.isIdentifier(funcNode.name)) {
                    name = getMapEntryOrDefault(funcNode.name);
                }
                // return undefined argName when arg is null or undefined
                // eslint-disable-next-line no-in-operator
                if (!name || "identifier" in name && name.identifier.text === "undefined") {
                    return undefined;
                }
                return name;
                function getMappedBindingNameOrDefault(bindingName) {
                    if (ts.isIdentifier(bindingName))
                        return getMapEntryOrDefault(bindingName);
                    var elements = ts.flatMap(bindingName.elements, function (element) {
                        if (ts.isOmittedExpression(element))
                            return [];
                        return [getMappedBindingNameOrDefault(element.name)];
                    });
                    return createSynthBindingPattern(bindingName, elements);
                }
                function getMapEntryOrDefault(identifier) {
                    var originalNode = getOriginalNode(identifier);
                    var symbol = getSymbol(originalNode);
                    if (!symbol) {
                        return createSynthIdentifier(identifier, types);
                    }
                    var mapEntry = transformer.synthNamesMap.get(ts.getSymbolId(symbol).toString());
                    return mapEntry || createSynthIdentifier(identifier, types);
                }
                function getSymbol(node) {
                    return node.symbol ? node.symbol : transformer.checker.getSymbolAtLocation(node);
                }
                function getOriginalNode(node) {
                    return node.original ? node.original : node;
                }
            }
            function isEmptyBindingName(bindingName) {
                if (!bindingName) {
                    return true;
                }
                if (isSynthIdentifier(bindingName)) {
                    return !bindingName.identifier.text;
                }
                return ts.every(bindingName.elements, isEmptyBindingName);
            }
            function getNode(bindingName) {
                return isSynthIdentifier(bindingName) ? bindingName.identifier : bindingName.bindingPattern;
            }
            function createSynthIdentifier(identifier, types) {
                if (types === void 0) {
                    types = [];
                }
                return { kind: 0 /* Identifier */, identifier: identifier, types: types, hasBeenDeclared: false };
            }
            function createSynthBindingPattern(bindingPattern, elements, types) {
                if (elements === void 0) {
                    elements = ts.emptyArray;
                }
                if (types === void 0) {
                    types = [];
                }
                return { kind: 1 /* BindingPattern */, bindingPattern: bindingPattern, elements: elements, types: types };
            }
            function isSynthIdentifier(bindingName) {
                return bindingName.kind === 0 /* Identifier */;
            }
            function isSynthBindingPattern(bindingName) {
                return bindingName.kind === 1 /* BindingPattern */;
            }
            function shouldReturn(expression, transformer) {
                return !!expression.original && transformer.setOfExpressionsToReturn.has(ts.getNodeId(expression.original));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            codefix.registerCodeFix({
                errorCodes: [ts.Diagnostics.File_is_a_CommonJS_module_it_may_be_converted_to_an_ES6_module.code],
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, program = context.program, preferences = context.preferences;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (changes) {
                        var moduleExportsChangedToDefault = convertFileToEs6Module(sourceFile, program.getTypeChecker(), changes, program.getCompilerOptions().target, ts.getQuotePreference(sourceFile, preferences));
                        if (moduleExportsChangedToDefault) {
                            for (var _i = 0, _a = program.getSourceFiles(); _i < _a.length; _i++) {
                                var importingFile = _a[_i];
                                fixImportOfModuleExports(importingFile, sourceFile, changes, ts.getQuotePreference(importingFile, preferences));
                            }
                        }
                    });
                    // No support for fix-all since this applies to the whole file at once anyway.
                    return [codefix.createCodeFixActionWithoutFixAll("convertToEs6Module", changes, ts.Diagnostics.Convert_to_ES6_module)];
                },
            });
            function fixImportOfModuleExports(importingFile, exportingFile, changes, quotePreference) {
                for (var _i = 0, _a = importingFile.imports; _i < _a.length; _i++) {
                    var moduleSpecifier = _a[_i];
                    var imported = ts.getResolvedModule(importingFile, moduleSpecifier.text);
                    if (!imported || imported.resolvedFileName !== exportingFile.fileName) {
                        continue;
                    }
                    var importNode = ts.importFromModuleSpecifier(moduleSpecifier);
                    switch (importNode.kind) {
                        case 263 /* ImportEqualsDeclaration */:
                            changes.replaceNode(importingFile, importNode, ts.makeImport(importNode.name, /*namedImports*/ undefined, moduleSpecifier, quotePreference));
                            break;
                        case 206 /* CallExpression */:
                            if (ts.isRequireCall(importNode, /*checkArgumentIsStringLiteralLike*/ false)) {
                                changes.replaceNode(importingFile, importNode, ts.factory.createPropertyAccessExpression(ts.getSynthesizedDeepClone(importNode), "default"));
                            }
                            break;
                    }
                }
            }
            /** @returns Whether we converted a `module.exports =` to a default export. */
            function convertFileToEs6Module(sourceFile, checker, changes, target, quotePreference) {
                var identifiers = { original: collectFreeIdentifiers(sourceFile), additional: new ts.Set() };
                var exports = collectExportRenames(sourceFile, checker, identifiers);
                convertExportsAccesses(sourceFile, exports, changes);
                var moduleExportsChangedToDefault = false;
                var useSitesToUnqualify;
                // Process variable statements first to collect use sites that need to be updated inside other transformations
                for (var _i = 0, _a = ts.filter(sourceFile.statements, ts.isVariableStatement); _i < _a.length; _i++) {
                    var statement = _a[_i];
                    var newUseSites = convertVariableStatement(sourceFile, statement, changes, checker, identifiers, target, quotePreference);
                    if (newUseSites) {
                        ts.copyEntries(newUseSites, useSitesToUnqualify !== null && useSitesToUnqualify !== void 0 ? useSitesToUnqualify : (useSitesToUnqualify = new ts.Map()));
                    }
                }
                // `convertStatement` will delete entries from `useSitesToUnqualify` when containing statements are replaced
                for (var _b = 0, _c = ts.filter(sourceFile.statements, function (s) { return !ts.isVariableStatement(s); }); _b < _c.length; _b++) {
                    var statement = _c[_b];
                    var moduleExportsChanged = convertStatement(sourceFile, statement, checker, changes, identifiers, target, exports, useSitesToUnqualify, quotePreference);
                    moduleExportsChangedToDefault = moduleExportsChangedToDefault || moduleExportsChanged;
                }
                // Remaining use sites can be changed directly
                useSitesToUnqualify === null || useSitesToUnqualify === void 0 ? void 0 : useSitesToUnqualify.forEach(function (replacement, original) {
                    changes.replaceNode(sourceFile, original, replacement);
                });
                return moduleExportsChangedToDefault;
            }
            function collectExportRenames(sourceFile, checker, identifiers) {
                var res = new ts.Map();
                forEachExportReference(sourceFile, function (node) {
                    var _a = node.name, text = _a.text, originalKeywordKind = _a.originalKeywordKind;
                    if (!res.has(text) && (originalKeywordKind !== undefined && ts.isNonContextualKeyword(originalKeywordKind)
                        || checker.resolveName(text, node, 111551 /* Value */, /*excludeGlobals*/ true))) {
                        // Unconditionally add an underscore in case `text` is a keyword.
                        res.set(text, makeUniqueName("_" + text, identifiers));
                    }
                });
                return res;
            }
            function convertExportsAccesses(sourceFile, exports, changes) {
                forEachExportReference(sourceFile, function (node, isAssignmentLhs) {
                    if (isAssignmentLhs) {
                        return;
                    }
                    var text = node.name.text;
                    changes.replaceNode(sourceFile, node, ts.factory.createIdentifier(exports.get(text) || text));
                });
            }
            function forEachExportReference(sourceFile, cb) {
                sourceFile.forEachChild(function recur(node) {
                    if (ts.isPropertyAccessExpression(node) && ts.isExportsOrModuleExportsOrAlias(sourceFile, node.expression) && ts.isIdentifier(node.name)) {
                        var parent = node.parent;
                        cb(node, ts.isBinaryExpression(parent) && parent.left === node && parent.operatorToken.kind === 63 /* EqualsToken */);
                    }
                    node.forEachChild(recur);
                });
            }
            function convertStatement(sourceFile, statement, checker, changes, identifiers, target, exports, useSitesToUnqualify, quotePreference) {
                switch (statement.kind) {
                    case 235 /* VariableStatement */:
                        convertVariableStatement(sourceFile, statement, changes, checker, identifiers, target, quotePreference);
                        return false;
                    case 236 /* ExpressionStatement */: {
                        var expression = statement.expression;
                        switch (expression.kind) {
                            case 206 /* CallExpression */: {
                                if (ts.isRequireCall(expression, /*checkArgumentIsStringLiteralLike*/ true)) {
                                    // For side-effecting require() call, just make a side-effecting import.
                                    changes.replaceNode(sourceFile, statement, ts.makeImport(/*name*/ undefined, /*namedImports*/ undefined, expression.arguments[0], quotePreference));
                                }
                                return false;
                            }
                            case 219 /* BinaryExpression */: {
                                var operatorToken = expression.operatorToken;
                                return operatorToken.kind === 63 /* EqualsToken */ && convertAssignment(sourceFile, checker, expression, changes, exports, useSitesToUnqualify);
                            }
                        }
                    }
                    // falls through
                    default:
                        return false;
                }
            }
            function convertVariableStatement(sourceFile, statement, changes, checker, identifiers, target, quotePreference) {
                var declarationList = statement.declarationList;
                var foundImport = false;
                var converted = ts.map(declarationList.declarations, function (decl) {
                    var name = decl.name, initializer = decl.initializer;
                    if (initializer) {
                        if (ts.isExportsOrModuleExportsOrAlias(sourceFile, initializer)) {
                            // `const alias = module.exports;` can be removed.
                            foundImport = true;
                            return convertedImports([]);
                        }
                        else if (ts.isRequireCall(initializer, /*checkArgumentIsStringLiteralLike*/ true)) {
                            foundImport = true;
                            return convertSingleImport(name, initializer.arguments[0], checker, identifiers, target, quotePreference);
                        }
                        else if (ts.isPropertyAccessExpression(initializer) && ts.isRequireCall(initializer.expression, /*checkArgumentIsStringLiteralLike*/ true)) {
                            foundImport = true;
                            return convertPropertyAccessImport(name, initializer.name.text, initializer.expression.arguments[0], identifiers, quotePreference);
                        }
                    }
                    // Move it out to its own variable statement. (This will not be used if `!foundImport`)
                    return convertedImports([ts.factory.createVariableStatement(/*modifiers*/ undefined, ts.factory.createVariableDeclarationList([decl], declarationList.flags))]);
                });
                if (foundImport) {
                    // useNonAdjustedEndPosition to ensure we don't eat the newline after the statement.
                    changes.replaceNodeWithNodes(sourceFile, statement, ts.flatMap(converted, function (c) { return c.newImports; }));
                    var combinedUseSites_1;
                    ts.forEach(converted, function (c) {
                        if (c.useSitesToUnqualify) {
                            ts.copyEntries(c.useSitesToUnqualify, combinedUseSites_1 !== null && combinedUseSites_1 !== void 0 ? combinedUseSites_1 : (combinedUseSites_1 = new ts.Map()));
                        }
                    });
                    return combinedUseSites_1;
                }
            }
            /** Converts `const name = require("moduleSpecifier").propertyName` */
            function convertPropertyAccessImport(name, propertyName, moduleSpecifier, identifiers, quotePreference) {
                switch (name.kind) {
                    case 199 /* ObjectBindingPattern */:
                    case 200 /* ArrayBindingPattern */: {
                        // `const [a, b] = require("c").d` --> `import { d } from "c"; const [a, b] = d;`
                        var tmp = makeUniqueName(propertyName, identifiers);
                        return convertedImports([
                            makeSingleImport(tmp, propertyName, moduleSpecifier, quotePreference),
                            makeConst(/*modifiers*/ undefined, name, ts.factory.createIdentifier(tmp)),
                        ]);
                    }
                    case 79 /* Identifier */:
                        // `const a = require("b").c` --> `import { c as a } from "./b";
                        return convertedImports([makeSingleImport(name.text, propertyName, moduleSpecifier, quotePreference)]);
                    default:
                        return ts.Debug.assertNever(name, "Convert to ES6 module got invalid syntax form " + name.kind);
                }
            }
            function convertAssignment(sourceFile, checker, assignment, changes, exports, useSitesToUnqualify) {
                var left = assignment.left, right = assignment.right;
                if (!ts.isPropertyAccessExpression(left)) {
                    return false;
                }
                if (ts.isExportsOrModuleExportsOrAlias(sourceFile, left)) {
                    if (ts.isExportsOrModuleExportsOrAlias(sourceFile, right)) {
                        // `const alias = module.exports;` or `module.exports = alias;` can be removed.
                        changes.delete(sourceFile, assignment.parent);
                    }
                    else {
                        var replacement = ts.isObjectLiteralExpression(right) ? tryChangeModuleExportsObject(right, useSitesToUnqualify)
                            : ts.isRequireCall(right, /*checkArgumentIsStringLiteralLike*/ true) ? convertReExportAll(right.arguments[0], checker)
                                : undefined;
                        if (replacement) {
                            changes.replaceNodeWithNodes(sourceFile, assignment.parent, replacement[0]);
                            return replacement[1];
                        }
                        else {
                            changes.replaceRangeWithText(sourceFile, ts.createRange(left.getStart(sourceFile), right.pos), "export default");
                            return true;
                        }
                    }
                }
                else if (ts.isExportsOrModuleExportsOrAlias(sourceFile, left.expression)) {
                    convertNamedExport(sourceFile, assignment, changes, exports);
                }
                return false;
            }
            /**
             * Convert `module.exports = { ... }` to individual exports..
             * We can't always do this if the module has interesting members -- then it will be a default export instead.
             */
            function tryChangeModuleExportsObject(object, useSitesToUnqualify) {
                var statements = ts.mapAllOrFail(object.properties, function (prop) {
                    switch (prop.kind) {
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                        // TODO: Maybe we should handle this? See fourslash test `refactorConvertToEs6Module_export_object_shorthand.ts`.
                        // falls through
                        case 292 /* ShorthandPropertyAssignment */:
                        case 293 /* SpreadAssignment */:
                            return undefined;
                        case 291 /* PropertyAssignment */:
                            return !ts.isIdentifier(prop.name) ? undefined : convertExportsDotXEquals_replaceNode(prop.name.text, prop.initializer, useSitesToUnqualify);
                        case 167 /* MethodDeclaration */:
                            return !ts.isIdentifier(prop.name) ? undefined : functionExpressionToDeclaration(prop.name.text, [ts.factory.createToken(93 /* ExportKeyword */)], prop, useSitesToUnqualify);
                        default:
                            ts.Debug.assertNever(prop, "Convert to ES6 got invalid prop kind " + prop.kind);
                    }
                });
                return statements && [statements, false];
            }
            function convertNamedExport(sourceFile, assignment, changes, exports) {
                // If "originalKeywordKind" was set, this is e.g. `exports.
                var text = assignment.left.name.text;
                var rename = exports.get(text);
                if (rename !== undefined) {
                    /*
                    const _class = 0;
                    export { _class as class };
                    */
                    var newNodes = [
                        makeConst(/*modifiers*/ undefined, rename, assignment.right),
                        makeExportDeclaration([ts.factory.createExportSpecifier(rename, text)]),
                    ];
                    changes.replaceNodeWithNodes(sourceFile, assignment.parent, newNodes);
                }
                else {
                    convertExportsPropertyAssignment(assignment, sourceFile, changes);
                }
            }
            function convertReExportAll(reExported, checker) {
                // `module.exports = require("x");` ==> `export * from "x"; export { default } from "x";`
                var moduleSpecifier = reExported.text;
                var moduleSymbol = checker.getSymbolAtLocation(reExported);
                var exports = moduleSymbol ? moduleSymbol.exports : ts.emptyMap;
                return exports.has("export=" /* ExportEquals */) ? [[reExportDefault(moduleSpecifier)], true] :
                    !exports.has("default" /* Default */) ? [[reExportStar(moduleSpecifier)], false] :
                        // If there's some non-default export, must include both `export *` and `export default`.
                        exports.size > 1 ? [[reExportStar(moduleSpecifier), reExportDefault(moduleSpecifier)], true] : [[reExportDefault(moduleSpecifier)], true];
            }
            function reExportStar(moduleSpecifier) {
                return makeExportDeclaration(/*exportClause*/ undefined, moduleSpecifier);
            }
            function reExportDefault(moduleSpecifier) {
                return makeExportDeclaration([ts.factory.createExportSpecifier(/*propertyName*/ undefined, "default")], moduleSpecifier);
            }
            function convertExportsPropertyAssignment(_a, sourceFile, changes) {
                var left = _a.left, right = _a.right, parent = _a.parent;
                var name = left.name.text;
                if ((ts.isFunctionExpression(right) || ts.isArrowFunction(right) || ts.isClassExpression(right)) && (!right.name || right.name.text === name)) {
                    // `exports.f = function() {}` -> `export function f() {}` -- Replace `exports.f = ` with `export `, and insert the name after `function`.
                    changes.replaceRange(sourceFile, { pos: left.getStart(sourceFile), end: right.getStart(sourceFile) }, ts.factory.createToken(93 /* ExportKeyword */), { suffix: " " });
                    if (!right.name)
                        changes.insertName(sourceFile, right, name);
                    var semi = ts.findChildOfKind(parent, 26 /* SemicolonToken */, sourceFile);
                    if (semi)
                        changes.delete(sourceFile, semi);
                }
                else {
                    // `exports.f = function g() {}` -> `export const f = function g() {}` -- just replace `exports.` with `export const `
                    changes.replaceNodeRangeWithNodes(sourceFile, left.expression, ts.findChildOfKind(left, 24 /* DotToken */, sourceFile), [ts.factory.createToken(93 /* ExportKeyword */), ts.factory.createToken(85 /* ConstKeyword */)], { joiner: " ", suffix: " " });
                }
            }
            // TODO: GH#22492 this will cause an error if a change has been made inside the body of the node.
            function convertExportsDotXEquals_replaceNode(name, exported, useSitesToUnqualify) {
                var modifiers = [ts.factory.createToken(93 /* ExportKeyword */)];
                switch (exported.kind) {
                    case 211 /* FunctionExpression */: {
                        var expressionName = exported.name;
                        if (expressionName && expressionName.text !== name) {
                            // `exports.f = function g() {}` -> `export const f = function g() {}`
                            return exportConst();
                        }
                    }
                    // falls through
                    case 212 /* ArrowFunction */:
                        // `exports.f = function() {}` --> `export function f() {}`
                        return functionExpressionToDeclaration(name, modifiers, exported, useSitesToUnqualify);
                    case 224 /* ClassExpression */:
                        // `exports.C = class {}` --> `export class C {}`
                        return classExpressionToDeclaration(name, modifiers, exported, useSitesToUnqualify);
                    default:
                        return exportConst();
                }
                function exportConst() {
                    // `exports.x = 0;` --> `export const x = 0;`
                    return makeConst(modifiers, ts.factory.createIdentifier(name), replaceImportUseSites(exported, useSitesToUnqualify)); // TODO: GH#18217
                }
            }
            function replaceImportUseSites(nodeOrNodes, useSitesToUnqualify) {
                if (!useSitesToUnqualify || !ts.some(ts.arrayFrom(useSitesToUnqualify.keys()), function (original) { return ts.rangeContainsRange(nodeOrNodes, original); })) {
                    return nodeOrNodes;
                }
                return ts.isArray(nodeOrNodes)
                    ? ts.getSynthesizedDeepClonesWithReplacements(nodeOrNodes, /*includeTrivia*/ true, replaceNode)
                    : ts.getSynthesizedDeepCloneWithReplacements(nodeOrNodes, /*includeTrivia*/ true, replaceNode);
                function replaceNode(original) {
                    // We are replacing `mod.SomeExport` wih `SomeExport`, so we only need to look at PropertyAccessExpressions
                    if (original.kind === 204 /* PropertyAccessExpression */) {
                        var replacement = useSitesToUnqualify.get(original);
                        // Remove entry from `useSitesToUnqualify` so the refactor knows it's taken care of by the parent statement we're replacing
                        useSitesToUnqualify.delete(original);
                        return replacement;
                    }
                }
            }
            /**
             * Converts `const <<name>> = require("x");`.
             * Returns nodes that will replace the variable declaration for the commonjs import.
             * May also make use `changes` to remove qualifiers at the use sites of imports, to change `mod.x` to `x`.
             */
            function convertSingleImport(name, moduleSpecifier, checker, identifiers, target, quotePreference) {
                switch (name.kind) {
                    case 199 /* ObjectBindingPattern */: {
                        var importSpecifiers = ts.mapAllOrFail(name.elements, function (e) {
                            return e.dotDotDotToken || e.initializer || e.propertyName && !ts.isIdentifier(e.propertyName) || !ts.isIdentifier(e.name)
                                ? undefined
                                // (TODO: GH#18217)
                                // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-assertion
                                : makeImportSpecifier(e.propertyName && e.propertyName.text, e.name.text);
                        });
                        if (importSpecifiers) {
                            return convertedImports([ts.makeImport(/*name*/ undefined, importSpecifiers, moduleSpecifier, quotePreference)]);
                        }
                    }
                    // falls through -- object destructuring has an interesting pattern and must be a variable declaration
                    case 200 /* ArrayBindingPattern */: {
                        /*
                        import x from "x";
                        const [a, b, c] = x;
                        */
                        var tmp = makeUniqueName(codefix.moduleSpecifierToValidIdentifier(moduleSpecifier.text, target), identifiers);
                        return convertedImports([
                            ts.makeImport(ts.factory.createIdentifier(tmp), /*namedImports*/ undefined, moduleSpecifier, quotePreference),
                            makeConst(/*modifiers*/ undefined, ts.getSynthesizedDeepClone(name), ts.factory.createIdentifier(tmp)),
                        ]);
                    }
                    case 79 /* Identifier */:
                        return convertSingleIdentifierImport(name, moduleSpecifier, checker, identifiers, quotePreference);
                    default:
                        return ts.Debug.assertNever(name, "Convert to ES6 module got invalid name kind " + name.kind);
                }
            }
            /**
             * Convert `import x = require("x").`
             * Also:
             * - Convert `x.default()` to `x()` to handle ES6 default export
             * - Converts uses like `x.y()` to `y()` and uses a named import.
             */
            function convertSingleIdentifierImport(name, moduleSpecifier, checker, identifiers, quotePreference) {
                var nameSymbol = checker.getSymbolAtLocation(name);
                // Maps from module property name to name actually used. (The same if there isn't shadowing.)
                var namedBindingsNames = new ts.Map();
                // True if there is some non-property use like `x()` or `f(x)`.
                var needDefaultImport = false;
                var useSitesToUnqualify;
                for (var _i = 0, _a = identifiers.original.get(name.text); _i < _a.length; _i++) {
                    var use = _a[_i];
                    if (checker.getSymbolAtLocation(use) !== nameSymbol || use === name) {
                        // This was a use of a different symbol with the same name, due to shadowing. Ignore.
                        continue;
                    }
                    var parent = use.parent;
                    if (ts.isPropertyAccessExpression(parent)) {
                        var propertyName = parent.name.text;
                        if (propertyName === "default") {
                            needDefaultImport = true;
                            var importDefaultName = use.getText();
                            (useSitesToUnqualify !== null && useSitesToUnqualify !== void 0 ? useSitesToUnqualify : (useSitesToUnqualify = new ts.Map())).set(parent, ts.factory.createIdentifier(importDefaultName));
                        }
                        else {
                            ts.Debug.assert(parent.expression === use, "Didn't expect expression === use"); // Else shouldn't have been in `collectIdentifiers`
                            var idName = namedBindingsNames.get(propertyName);
                            if (idName === undefined) {
                                idName = makeUniqueName(propertyName, identifiers);
                                namedBindingsNames.set(propertyName, idName);
                            }
                            (useSitesToUnqualify !== null && useSitesToUnqualify !== void 0 ? useSitesToUnqualify : (useSitesToUnqualify = new ts.Map())).set(parent, ts.factory.createIdentifier(idName));
                        }
                    }
                    else {
                        needDefaultImport = true;
                    }
                }
                var namedBindings = namedBindingsNames.size === 0 ? undefined : ts.arrayFrom(ts.mapIterator(namedBindingsNames.entries(), function (_a) {
                    var propertyName = _a[0], idName = _a[1];
                    return ts.factory.createImportSpecifier(propertyName === idName ? undefined : ts.factory.createIdentifier(propertyName), ts.factory.createIdentifier(idName));
                }));
                if (!namedBindings) {
                    // If it was unused, ensure that we at least import *something*.
                    needDefaultImport = true;
                }
                return convertedImports([ts.makeImport(needDefaultImport ? ts.getSynthesizedDeepClone(name) : undefined, namedBindings, moduleSpecifier, quotePreference)], useSitesToUnqualify);
            }
            // Identifiers helpers
            function makeUniqueName(name, identifiers) {
                while (identifiers.original.has(name) || identifiers.additional.has(name)) {
                    name = "_" + name;
                }
                identifiers.additional.add(name);
                return name;
            }
            function collectFreeIdentifiers(file) {
                var map = ts.createMultiMap();
                forEachFreeIdentifier(file, function (id) { return map.add(id.text, id); });
                return map;
            }
            /**
             * A free identifier is an identifier that can be accessed through name lookup as a local variable.
             * In the expression `x.y`, `x` is a free identifier, but `y` is not.
             */
            function forEachFreeIdentifier(node, cb) {
                if (ts.isIdentifier(node) && isFreeIdentifier(node))
                    cb(node);
                node.forEachChild(function (child) { return forEachFreeIdentifier(child, cb); });
            }
            function isFreeIdentifier(node) {
                var parent = node.parent;
                switch (parent.kind) {
                    case 204 /* PropertyAccessExpression */:
                        return parent.name !== node;
                    case 201 /* BindingElement */:
                        return parent.propertyName !== node;
                    case 268 /* ImportSpecifier */:
                        return parent.propertyName !== node;
                    default:
                        return true;
                }
            }
            // Node helpers
            function functionExpressionToDeclaration(name, additionalModifiers, fn, useSitesToUnqualify) {
                return ts.factory.createFunctionDeclaration(ts.getSynthesizedDeepClones(fn.decorators), // TODO: GH#19915 Don't think this is even legal.
                ts.concatenate(additionalModifiers, ts.getSynthesizedDeepClones(fn.modifiers)), ts.getSynthesizedDeepClone(fn.asteriskToken), name, ts.getSynthesizedDeepClones(fn.typeParameters), ts.getSynthesizedDeepClones(fn.parameters), ts.getSynthesizedDeepClone(fn.type), ts.factory.converters.convertToFunctionBlock(replaceImportUseSites(fn.body, useSitesToUnqualify)));
            }
            function classExpressionToDeclaration(name, additionalModifiers, cls, useSitesToUnqualify) {
                return ts.factory.createClassDeclaration(ts.getSynthesizedDeepClones(cls.decorators), // TODO: GH#19915 Don't think this is even legal.
                ts.concatenate(additionalModifiers, ts.getSynthesizedDeepClones(cls.modifiers)), name, ts.getSynthesizedDeepClones(cls.typeParameters), ts.getSynthesizedDeepClones(cls.heritageClauses), replaceImportUseSites(cls.members, useSitesToUnqualify));
            }
            function makeSingleImport(localName, propertyName, moduleSpecifier, quotePreference) {
                return propertyName === "default"
                    ? ts.makeImport(ts.factory.createIdentifier(localName), /*namedImports*/ undefined, moduleSpecifier, quotePreference)
                    : ts.makeImport(/*name*/ undefined, [makeImportSpecifier(propertyName, localName)], moduleSpecifier, quotePreference);
            }
            function makeImportSpecifier(propertyName, name) {
                return ts.factory.createImportSpecifier(propertyName !== undefined && propertyName !== name ? ts.factory.createIdentifier(propertyName) : undefined, ts.factory.createIdentifier(name));
            }
            function makeConst(modifiers, name, init) {
                return ts.factory.createVariableStatement(modifiers, ts.factory.createVariableDeclarationList([ts.factory.createVariableDeclaration(name, /*exclamationToken*/ undefined, /*type*/ undefined, init)], 2 /* Const */));
            }
            function makeExportDeclaration(exportSpecifiers, moduleSpecifier) {
                return ts.factory.createExportDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*isTypeOnly*/ false, exportSpecifiers && ts.factory.createNamedExports(exportSpecifiers), moduleSpecifier === undefined ? undefined : ts.factory.createStringLiteral(moduleSpecifier));
            }
            function convertedImports(newImports, useSitesToUnqualify) {
                return {
                    newImports: newImports,
                    useSitesToUnqualify: useSitesToUnqualify
                };
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "correctQualifiedNameToIndexedAccessType";
            var errorCodes = [ts.Diagnostics.Cannot_access_0_1_because_0_is_a_type_but_not_a_namespace_Did_you_mean_to_retrieve_the_type_of_the_property_1_in_0_with_0_1.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var qualifiedName = getQualifiedName(context.sourceFile, context.span.start);
                    if (!qualifiedName)
                        return undefined;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, context.sourceFile, qualifiedName); });
                    var newText = qualifiedName.left.text + "[\"" + qualifiedName.right.text + "\"]";
                    return [codefix.createCodeFixAction(fixId, changes, [ts.Diagnostics.Rewrite_as_the_indexed_access_type_0, newText], fixId, ts.Diagnostics.Rewrite_all_as_indexed_access_types)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var q = getQualifiedName(diag.file, diag.start);
                        if (q) {
                            doChange(changes, diag.file, q);
                        }
                    });
                },
            });
            function getQualifiedName(sourceFile, pos) {
                var qualifiedName = ts.findAncestor(ts.getTokenAtPosition(sourceFile, pos), ts.isQualifiedName);
                ts.Debug.assert(!!qualifiedName, "Expected position to be owned by a qualified name.");
                return ts.isIdentifier(qualifiedName.left) ? qualifiedName : undefined;
            }
            function doChange(changeTracker, sourceFile, qualifiedName) {
                var rightText = qualifiedName.right.text;
                var replacement = ts.factory.createIndexedAccessTypeNode(ts.factory.createTypeReferenceNode(qualifiedName.left, /*typeArguments*/ undefined), ts.factory.createLiteralTypeNode(ts.factory.createStringLiteral(rightText)));
                changeTracker.replaceNode(sourceFile, qualifiedName, replacement);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var errorCodes = [ts.Diagnostics.Re_exporting_a_type_when_the_isolatedModules_flag_is_provided_requires_using_export_type.code];
            var fixId = "convertToTypeOnlyExport";
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return fixSingleExportDeclaration(t, getExportSpecifierForDiagnosticSpan(context.span, context.sourceFile), context); });
                    if (changes.length) {
                        return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Convert_to_type_only_export, fixId, ts.Diagnostics.Convert_all_re_exported_types_to_type_only_exports)];
                    }
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    var fixedExportDeclarations = new ts.Map();
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var exportSpecifier = getExportSpecifierForDiagnosticSpan(diag, context.sourceFile);
                        if (exportSpecifier && ts.addToSeen(fixedExportDeclarations, ts.getNodeId(exportSpecifier.parent.parent))) {
                            fixSingleExportDeclaration(changes, exportSpecifier, context);
                        }
                    });
                }
            });
            function getExportSpecifierForDiagnosticSpan(span, sourceFile) {
                return ts.tryCast(ts.getTokenAtPosition(sourceFile, span.start).parent, ts.isExportSpecifier);
            }
            function fixSingleExportDeclaration(changes, exportSpecifier, context) {
                if (!exportSpecifier) {
                    return;
                }
                var exportClause = exportSpecifier.parent;
                var exportDeclaration = exportClause.parent;
                var typeExportSpecifiers = getTypeExportSpecifiers(exportSpecifier, context);
                if (typeExportSpecifiers.length === exportClause.elements.length) {
                    changes.insertModifierBefore(context.sourceFile, 150 /* TypeKeyword */, exportClause);
                }
                else {
                    var valueExportDeclaration = ts.factory.updateExportDeclaration(exportDeclaration, exportDeclaration.decorators, exportDeclaration.modifiers, 
                    /*isTypeOnly*/ false, ts.factory.updateNamedExports(exportClause, ts.filter(exportClause.elements, function (e) { return !ts.contains(typeExportSpecifiers, e); })), exportDeclaration.moduleSpecifier);
                    var typeExportDeclaration = ts.factory.createExportDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, 
                    /*isTypeOnly*/ true, ts.factory.createNamedExports(typeExportSpecifiers), exportDeclaration.moduleSpecifier);
                    changes.replaceNode(context.sourceFile, exportDeclaration, valueExportDeclaration, {
                        leadingTriviaOption: ts.textChanges.LeadingTriviaOption.IncludeAll,
                        trailingTriviaOption: ts.textChanges.TrailingTriviaOption.Exclude
                    });
                    changes.insertNodeAfter(context.sourceFile, exportDeclaration, typeExportDeclaration);
                }
            }
            function getTypeExportSpecifiers(originExportSpecifier, context) {
                var exportClause = originExportSpecifier.parent;
                if (exportClause.elements.length === 1) {
                    return exportClause.elements;
                }
                var diagnostics = ts.getDiagnosticsWithinSpan(ts.createTextSpanFromNode(exportClause), context.program.getSemanticDiagnostics(context.sourceFile, context.cancellationToken));
                return ts.filter(exportClause.elements, function (element) {
                    var _a;
                    return element === originExportSpecifier || ((_a = ts.findDiagnosticForNode(element, diagnostics)) === null || _a === void 0 ? void 0 : _a.code) === errorCodes[0];
                });
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var errorCodes = [ts.Diagnostics.This_import_is_never_used_as_a_value_and_must_use_import_type_because_importsNotUsedAsValues_is_set_to_error.code];
            var fixId = "convertToTypeOnlyImport";
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) {
                        var importDeclaration = getImportDeclarationForDiagnosticSpan(context.span, context.sourceFile);
                        fixSingleImportDeclaration(t, importDeclaration, context);
                    });
                    if (changes.length) {
                        return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Convert_to_type_only_import, fixId, ts.Diagnostics.Convert_all_imports_not_used_as_a_value_to_type_only_imports)];
                    }
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var importDeclaration = getImportDeclarationForDiagnosticSpan(diag, context.sourceFile);
                        fixSingleImportDeclaration(changes, importDeclaration, context);
                    });
                }
            });
            function getImportDeclarationForDiagnosticSpan(span, sourceFile) {
                return ts.tryCast(ts.getTokenAtPosition(sourceFile, span.start).parent, ts.isImportDeclaration);
            }
            function fixSingleImportDeclaration(changes, importDeclaration, context) {
                if (!(importDeclaration === null || importDeclaration === void 0 ? void 0 : importDeclaration.importClause)) {
                    return;
                }
                var importClause = importDeclaration.importClause;
                // `changes.insertModifierBefore` produces a range that might overlap further changes
                changes.insertText(context.sourceFile, importDeclaration.getStart() + "import".length, " type");
                // `import type foo, { Bar }` is not allowed, so move `foo` to new declaration
                if (importClause.name && importClause.namedBindings) {
                    changes.deleteNodeRangeExcludingEnd(context.sourceFile, importClause.name, importDeclaration.importClause.namedBindings);
                    changes.insertNodeBefore(context.sourceFile, importDeclaration, ts.factory.updateImportDeclaration(importDeclaration, 
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, ts.factory.createImportClause(
                    /*isTypeOnly*/ true, importClause.name, 
                    /*namedBindings*/ undefined), importDeclaration.moduleSpecifier));
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "convertLiteralTypeToMappedType";
            var errorCodes = [ts.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here_Did_you_mean_to_use_1_in_0.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var info = getInfo(sourceFile, span.start);
                    if (!info) {
                        return undefined;
                    }
                    var name = info.name, constraint = info.constraint;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, info); });
                    return [codefix.createCodeFixAction(fixId, changes, [ts.Diagnostics.Convert_0_to_1_in_0, constraint, name], fixId, ts.Diagnostics.Convert_all_type_literals_to_mapped_type)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var info = getInfo(diag.file, diag.start);
                        if (info) {
                            doChange(changes, diag.file, info);
                        }
                    });
                }
            });
            function getInfo(sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                if (ts.isIdentifier(token)) {
                    var propertySignature = ts.cast(token.parent.parent, ts.isPropertySignature);
                    var propertyName = token.getText(sourceFile);
                    return {
                        container: ts.cast(propertySignature.parent, ts.isTypeLiteralNode),
                        typeNode: propertySignature.type,
                        constraint: propertyName,
                        name: propertyName === "K" ? "P" : "K",
                    };
                }
                return undefined;
            }
            function doChange(changes, sourceFile, _a) {
                var container = _a.container, typeNode = _a.typeNode, constraint = _a.constraint, name = _a.name;
                changes.replaceNode(sourceFile, container, ts.factory.createMappedTypeNode(/*readonlyToken*/ undefined, ts.factory.createTypeParameterDeclaration(name, ts.factory.createTypeReferenceNode(constraint)), /*nameType*/ undefined, /*questionToken*/ undefined, typeNode));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var errorCodes = [
                ts.Diagnostics.Class_0_incorrectly_implements_interface_1.code,
                ts.Diagnostics.Class_0_incorrectly_implements_class_1_Did_you_mean_to_extend_1_and_inherit_its_members_as_a_subclass.code
            ];
            var fixId = "fixClassIncorrectlyImplementsInterface"; // TODO: share a group with fixClassDoesntImplementInheritedAbstractMember?
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var classDeclaration = getClass(sourceFile, span.start);
                    return ts.mapDefined(ts.getEffectiveImplementsTypeNodes(classDeclaration), function (implementedTypeNode) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addMissingDeclarations(context, implementedTypeNode, sourceFile, classDeclaration, t, context.preferences); });
                        return changes.length === 0 ? undefined : codefix.createCodeFixAction(fixId, changes, [ts.Diagnostics.Implement_interface_0, implementedTypeNode.getText(sourceFile)], fixId, ts.Diagnostics.Implement_all_unimplemented_interfaces);
                    });
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    var seenClassDeclarations = new ts.Map();
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var classDeclaration = getClass(diag.file, diag.start);
                        if (ts.addToSeen(seenClassDeclarations, ts.getNodeId(classDeclaration))) {
                            for (var _i = 0, _a = ts.getEffectiveImplementsTypeNodes(classDeclaration); _i < _a.length; _i++) {
                                var implementedTypeNode = _a[_i];
                                addMissingDeclarations(context, implementedTypeNode, diag.file, classDeclaration, changes, context.preferences);
                            }
                        }
                    });
                },
            });
            function getClass(sourceFile, pos) {
                return ts.Debug.checkDefined(ts.getContainingClass(ts.getTokenAtPosition(sourceFile, pos)), "There should be a containing class");
            }
            function symbolPointsToNonPrivateMember(symbol) {
                return !symbol.valueDeclaration || !(ts.getEffectiveModifierFlags(symbol.valueDeclaration) & 8 /* Private */);
            }
            function addMissingDeclarations(context, implementedTypeNode, sourceFile, classDeclaration, changeTracker, preferences) {
                var checker = context.program.getTypeChecker();
                var maybeHeritageClauseSymbol = getHeritageClauseSymbolTable(classDeclaration, checker);
                // Note that this is ultimately derived from a map indexed by symbol names,
                // so duplicates cannot occur.
                var implementedType = checker.getTypeAtLocation(implementedTypeNode);
                var implementedTypeSymbols = checker.getPropertiesOfType(implementedType);
                var nonPrivateAndNotExistedInHeritageClauseMembers = implementedTypeSymbols.filter(ts.and(symbolPointsToNonPrivateMember, function (symbol) { return !maybeHeritageClauseSymbol.has(symbol.escapedName); }));
                var classType = checker.getTypeAtLocation(classDeclaration);
                var constructor = ts.find(classDeclaration.members, function (m) { return ts.isConstructorDeclaration(m); });
                if (!classType.getNumberIndexType()) {
                    createMissingIndexSignatureDeclaration(implementedType, 1 /* Number */);
                }
                if (!classType.getStringIndexType()) {
                    createMissingIndexSignatureDeclaration(implementedType, 0 /* String */);
                }
                var importAdder = codefix.createImportAdder(sourceFile, context.program, preferences, context.host);
                codefix.createMissingMemberNodes(classDeclaration, nonPrivateAndNotExistedInHeritageClauseMembers, sourceFile, context, preferences, importAdder, function (member) { return insertInterfaceMemberNode(sourceFile, classDeclaration, member); });
                importAdder.writeFixes(changeTracker);
                function createMissingIndexSignatureDeclaration(type, kind) {
                    var indexInfoOfKind = checker.getIndexInfoOfType(type, kind);
                    if (indexInfoOfKind) {
                        insertInterfaceMemberNode(sourceFile, classDeclaration, checker.indexInfoToIndexSignatureDeclaration(indexInfoOfKind, classDeclaration, /*flags*/ undefined, codefix.getNoopSymbolTrackerWithResolver(context)));
                    }
                }
                // Either adds the node at the top of the class, or if there's a constructor right after that
                function insertInterfaceMemberNode(sourceFile, cls, newElement) {
                    if (constructor) {
                        changeTracker.insertNodeAfter(sourceFile, constructor, newElement);
                    }
                    else {
                        changeTracker.insertNodeAtClassStart(sourceFile, cls, newElement);
                    }
                }
            }
            function getHeritageClauseSymbolTable(classDeclaration, checker) {
                var heritageClauseNode = ts.getEffectiveBaseTypeNode(classDeclaration);
                if (!heritageClauseNode)
                    return ts.createSymbolTable();
                var heritageClauseType = checker.getTypeAtLocation(heritageClauseNode);
                var heritageClauseTypeSymbols = checker.getPropertiesOfType(heritageClauseType);
                return ts.createSymbolTable(heritageClauseTypeSymbols.filter(symbolPointsToNonPrivateMember));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            codefix.importFixName = "import";
            var importFixId = "fixMissingImport";
            var errorCodes = [
                ts.Diagnostics.Cannot_find_name_0.code,
                ts.Diagnostics.Cannot_find_name_0_Did_you_mean_1.code,
                ts.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0.code,
                ts.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0.code,
                ts.Diagnostics.Cannot_find_namespace_0.code,
                ts.Diagnostics._0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead.code,
                ts.Diagnostics._0_only_refers_to_a_type_but_is_being_used_as_a_value_here.code,
                ts.Diagnostics.No_value_exists_in_scope_for_the_shorthand_property_0_Either_declare_one_or_provide_an_initializer.code
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var errorCode = context.errorCode, preferences = context.preferences, sourceFile = context.sourceFile, span = context.span;
                    var info = getFixesInfo(context, errorCode, span.start, /*useAutoImportProvider*/ true);
                    if (!info)
                        return undefined;
                    var fixes = info.fixes, symbolName = info.symbolName;
                    var quotePreference = ts.getQuotePreference(sourceFile, preferences);
                    return fixes.map(function (fix) { return codeActionForFix(context, sourceFile, symbolName, fix, quotePreference); });
                },
                fixIds: [importFixId],
                getAllCodeActions: function (context) {
                    var sourceFile = context.sourceFile, program = context.program, preferences = context.preferences, host = context.host;
                    var importAdder = createImportAdderWorker(sourceFile, program, /*useAutoImportProvider*/ true, preferences, host);
                    codefix.eachDiagnostic(context, errorCodes, function (diag) { return importAdder.addImportFromDiagnostic(diag, context); });
                    return codefix.createCombinedCodeActions(ts.textChanges.ChangeTracker.with(context, importAdder.writeFixes));
                },
            });
            function createImportAdder(sourceFile, program, preferences, host) {
                return createImportAdderWorker(sourceFile, program, /*useAutoImportProvider*/ false, preferences, host);
            }
            codefix.createImportAdder = createImportAdder;
            function createImportAdderWorker(sourceFile, program, useAutoImportProvider, preferences, host) {
                var compilerOptions = program.getCompilerOptions();
                // Namespace fixes don't conflict, so just build a list.
                var addToNamespace = [];
                var importType = [];
                // Keys are import clause node IDs.
                var addToExisting = new ts.Map();
                var newImports = new ts.Map();
                return { addImportFromDiagnostic: addImportFromDiagnostic, addImportFromExportedSymbol: addImportFromExportedSymbol, writeFixes: writeFixes };
                function addImportFromDiagnostic(diagnostic, context) {
                    var info = getFixesInfo(context, diagnostic.code, diagnostic.start, useAutoImportProvider);
                    if (!info || !info.fixes.length)
                        return;
                    addImport(info);
                }
                function addImportFromExportedSymbol(exportedSymbol, usageIsTypeOnly) {
                    var moduleSymbol = ts.Debug.checkDefined(exportedSymbol.parent);
                    var symbolName = ts.getNameForExportedSymbol(exportedSymbol, ts.getEmitScriptTarget(compilerOptions));
                    var checker = program.getTypeChecker();
                    var symbol = checker.getMergedSymbol(ts.skipAlias(exportedSymbol, checker));
                    var exportInfos = getAllReExportingModules(sourceFile, symbol, moduleSymbol, symbolName, host, program, preferences, useAutoImportProvider);
                    var preferTypeOnlyImport = !!usageIsTypeOnly && compilerOptions.importsNotUsedAsValues === 2 /* Error */;
                    var useRequire = shouldUseRequire(sourceFile, program);
                    var fix = getImportFixForSymbol(sourceFile, exportInfos, moduleSymbol, symbolName, program, /*position*/ undefined, preferTypeOnlyImport, useRequire, host, preferences);
                    if (fix) {
                        addImport({ fixes: [fix], symbolName: symbolName });
                    }
                }
                function addImport(info) {
                    var fixes = info.fixes, symbolName = info.symbolName;
                    var fix = ts.first(fixes);
                    switch (fix.kind) {
                        case 0 /* UseNamespace */:
                            addToNamespace.push(fix);
                            break;
                        case 1 /* ImportType */:
                            importType.push(fix);
                            break;
                        case 2 /* AddToExisting */: {
                            var importClauseOrBindingPattern = fix.importClauseOrBindingPattern, importKind = fix.importKind, canUseTypeOnlyImport = fix.canUseTypeOnlyImport;
                            var key = String(ts.getNodeId(importClauseOrBindingPattern));
                            var entry = addToExisting.get(key);
                            if (!entry) {
                                addToExisting.set(key, entry = { importClauseOrBindingPattern: importClauseOrBindingPattern, defaultImport: undefined, namedImports: [], canUseTypeOnlyImport: canUseTypeOnlyImport });
                            }
                            if (importKind === 0 /* Named */) {
                                ts.pushIfUnique(entry.namedImports, symbolName);
                            }
                            else {
                                ts.Debug.assert(entry.defaultImport === undefined || entry.defaultImport === symbolName, "(Add to Existing) Default import should be missing or match symbolName");
                                entry.defaultImport = symbolName;
                            }
                            break;
                        }
                        case 3 /* AddNew */: {
                            var moduleSpecifier = fix.moduleSpecifier, importKind = fix.importKind, useRequire = fix.useRequire, typeOnly = fix.typeOnly;
                            var entry = newImports.get(moduleSpecifier);
                            if (!entry) {
                                newImports.set(moduleSpecifier, entry = { namedImports: [], namespaceLikeImport: undefined, typeOnly: typeOnly, useRequire: useRequire });
                            }
                            else {
                                // An import clause can only be type-only if every import fix contributing to it can be type-only.
                                entry.typeOnly = entry.typeOnly && typeOnly;
                            }
                            switch (importKind) {
                                case 1 /* Default */:
                                    ts.Debug.assert(entry.defaultImport === undefined || entry.defaultImport === symbolName, "(Add new) Default import should be missing or match symbolName");
                                    entry.defaultImport = symbolName;
                                    break;
                                case 0 /* Named */:
                                    ts.pushIfUnique(entry.namedImports || (entry.namedImports = []), symbolName);
                                    break;
                                case 3 /* CommonJS */:
                                case 2 /* Namespace */:
                                    ts.Debug.assert(entry.namespaceLikeImport === undefined || entry.namespaceLikeImport.name === symbolName, "Namespacelike import shoudl be missing or match symbolName");
                                    entry.namespaceLikeImport = { importKind: importKind, name: symbolName };
                                    break;
                            }
                            break;
                        }
                        default:
                            ts.Debug.assertNever(fix, "fix wasn't never - got kind " + fix.kind);
                    }
                }
                function writeFixes(changeTracker) {
                    var quotePreference = ts.getQuotePreference(sourceFile, preferences);
                    for (var _i = 0, addToNamespace_1 = addToNamespace; _i < addToNamespace_1.length; _i++) {
                        var fix = addToNamespace_1[_i];
                        addNamespaceQualifier(changeTracker, sourceFile, fix);
                    }
                    for (var _a = 0, importType_1 = importType; _a < importType_1.length; _a++) {
                        var fix = importType_1[_a];
                        addImportType(changeTracker, sourceFile, fix, quotePreference);
                    }
                    addToExisting.forEach(function (_a) {
                        var importClauseOrBindingPattern = _a.importClauseOrBindingPattern, defaultImport = _a.defaultImport, namedImports = _a.namedImports, canUseTypeOnlyImport = _a.canUseTypeOnlyImport;
                        doAddExistingFix(changeTracker, sourceFile, importClauseOrBindingPattern, defaultImport, namedImports, canUseTypeOnlyImport);
                    });
                    var newDeclarations;
                    newImports.forEach(function (_a, moduleSpecifier) {
                        var useRequire = _a.useRequire, imports = __rest(_a, ["useRequire"]);
                        var getDeclarations = useRequire ? getNewRequires : getNewImports;
                        newDeclarations = ts.combine(newDeclarations, getDeclarations(moduleSpecifier, quotePreference, imports));
                    });
                    if (newDeclarations) {
                        ts.insertImports(changeTracker, sourceFile, newDeclarations, /*blankLineBetween*/ true);
                    }
                }
            }
            // Sorted with the preferred fix coming first.
            var ImportFixKind;
            (function (ImportFixKind) {
                ImportFixKind[ImportFixKind["UseNamespace"] = 0] = "UseNamespace";
                ImportFixKind[ImportFixKind["ImportType"] = 1] = "ImportType";
                ImportFixKind[ImportFixKind["AddToExisting"] = 2] = "AddToExisting";
                ImportFixKind[ImportFixKind["AddNew"] = 3] = "AddNew";
            })(ImportFixKind || (ImportFixKind = {}));
            function getImportCompletionAction(exportedSymbol, moduleSymbol, sourceFile, symbolName, host, program, formatContext, position, preferences) {
                var compilerOptions = program.getCompilerOptions();
                var exportInfos = ts.pathIsBareSpecifier(ts.stripQuotes(moduleSymbol.name))
                    ? [getSymbolExportInfoForSymbol(exportedSymbol, moduleSymbol, program, host)]
                    : getAllReExportingModules(sourceFile, exportedSymbol, moduleSymbol, symbolName, host, program, preferences, /*useAutoImportProvider*/ true);
                var useRequire = shouldUseRequire(sourceFile, program);
                var preferTypeOnlyImport = compilerOptions.importsNotUsedAsValues === 2 /* Error */ && !ts.isSourceFileJS(sourceFile) && ts.isValidTypeOnlyAliasUseSite(ts.getTokenAtPosition(sourceFile, position));
                var fix = ts.Debug.checkDefined(getImportFixForSymbol(sourceFile, exportInfos, moduleSymbol, symbolName, program, position, preferTypeOnlyImport, useRequire, host, preferences));
                return { moduleSpecifier: fix.moduleSpecifier, codeAction: codeFixActionToCodeAction(codeActionForFix({ host: host, formatContext: formatContext, preferences: preferences }, sourceFile, symbolName, fix, ts.getQuotePreference(sourceFile, preferences))) };
            }
            codefix.getImportCompletionAction = getImportCompletionAction;
            function getImportFixForSymbol(sourceFile, exportInfos, moduleSymbol, symbolName, program, position, preferTypeOnlyImport, useRequire, host, preferences) {
                ts.Debug.assert(exportInfos.some(function (info) { return info.moduleSymbol === moduleSymbol || info.symbol.parent === moduleSymbol; }), "Some exportInfo should match the specified moduleSymbol");
                return getBestFix(getImportFixes(exportInfos, symbolName, position, preferTypeOnlyImport, useRequire, program, sourceFile, host, preferences), sourceFile, program, host, preferences);
            }
            function codeFixActionToCodeAction(_a) {
                var description = _a.description, changes = _a.changes, commands = _a.commands;
                return { description: description, changes: changes, commands: commands };
            }
            function getSymbolExportInfoForSymbol(symbol, moduleSymbol, program, host) {
                var _a, _b;
                var compilerOptions = program.getCompilerOptions();
                var mainProgramInfo = getInfoWithChecker(program.getTypeChecker(), /*isFromPackageJson*/ false);
                if (mainProgramInfo) {
                    return mainProgramInfo;
                }
                var autoImportProvider = (_b = (_a = host.getPackageJsonAutoImportProvider) === null || _a === void 0 ? void 0 : _a.call(host)) === null || _b === void 0 ? void 0 : _b.getTypeChecker();
                return ts.Debug.checkDefined(autoImportProvider && getInfoWithChecker(autoImportProvider, /*isFromPackageJson*/ true), "Could not find symbol in specified module for code actions");
                function getInfoWithChecker(checker, isFromPackageJson) {
                    var defaultInfo = ts.getDefaultLikeExportInfo(moduleSymbol, checker, compilerOptions);
                    if (defaultInfo && ts.skipAlias(defaultInfo.symbol, checker) === symbol) {
                        return { symbol: defaultInfo.symbol, moduleSymbol: moduleSymbol, moduleFileName: undefined, exportKind: defaultInfo.exportKind, targetFlags: ts.skipAlias(symbol, checker).flags, isFromPackageJson: isFromPackageJson };
                    }
                    var named = checker.tryGetMemberInModuleExportsAndProperties(symbol.name, moduleSymbol);
                    if (named && ts.skipAlias(named, checker) === symbol) {
                        return { symbol: named, moduleSymbol: moduleSymbol, moduleFileName: undefined, exportKind: 0 /* Named */, targetFlags: ts.skipAlias(symbol, checker).flags, isFromPackageJson: isFromPackageJson };
                    }
                }
            }
            function getAllReExportingModules(importingFile, exportedSymbol, exportingModuleSymbol, symbolName, host, program, preferences, useAutoImportProvider) {
                var result = [];
                var compilerOptions = program.getCompilerOptions();
                var getModuleSpecifierResolutionHost = ts.memoizeOne(function (isFromPackageJson) {
                    return ts.createModuleSpecifierResolutionHost(isFromPackageJson ? host.getPackageJsonAutoImportProvider() : program, host);
                });
                ts.forEachExternalModuleToImportFrom(program, host, useAutoImportProvider, function (moduleSymbol, moduleFile, program, isFromPackageJson) {
                    var checker = program.getTypeChecker();
                    // Don't import from a re-export when looking "up" like to `./index` or `../index`.
                    if (moduleFile && moduleSymbol !== exportingModuleSymbol && ts.startsWith(importingFile.fileName, ts.getDirectoryPath(moduleFile.fileName))) {
                        return;
                    }
                    var defaultInfo = ts.getDefaultLikeExportInfo(moduleSymbol, checker, compilerOptions);
                    if (defaultInfo && (defaultInfo.name === symbolName || moduleSymbolToValidIdentifier(moduleSymbol, compilerOptions.target) === symbolName) && ts.skipAlias(defaultInfo.symbol, checker) === exportedSymbol && isImportable(program, moduleFile, isFromPackageJson)) {
                        result.push({ symbol: defaultInfo.symbol, moduleSymbol: moduleSymbol, moduleFileName: moduleFile === null || moduleFile === void 0 ? void 0 : moduleFile.fileName, exportKind: defaultInfo.exportKind, targetFlags: ts.skipAlias(defaultInfo.symbol, checker).flags, isFromPackageJson: isFromPackageJson });
                    }
                    for (var _i = 0, _a = checker.getExportsAndPropertiesOfModule(moduleSymbol); _i < _a.length; _i++) {
                        var exported = _a[_i];
                        if (exported.name === symbolName && ts.skipAlias(exported, checker) === exportedSymbol && isImportable(program, moduleFile, isFromPackageJson)) {
                            result.push({ symbol: exported, moduleSymbol: moduleSymbol, moduleFileName: moduleFile === null || moduleFile === void 0 ? void 0 : moduleFile.fileName, exportKind: 0 /* Named */, targetFlags: ts.skipAlias(exported, checker).flags, isFromPackageJson: isFromPackageJson });
                        }
                    }
                });
                return result;
                function isImportable(program, moduleFile, isFromPackageJson) {
                    var _a;
                    return !moduleFile || ts.isImportableFile(program, importingFile, moduleFile, preferences, /*packageJsonFilter*/ undefined, getModuleSpecifierResolutionHost(isFromPackageJson), (_a = host.getModuleSpecifierCache) === null || _a === void 0 ? void 0 : _a.call(host));
                }
            }
            function getModuleSpecifierForBestExportInfo(exportInfo, importingFile, program, host, preferences, fromCacheOnly) {
                var _a = getNewImportFixes(program, importingFile, 
                /*position*/ undefined, 
                /*preferTypeOnlyImport*/ false, 
                /*useRequire*/ false, exportInfo, host, preferences, fromCacheOnly), fixes = _a.fixes, computedWithoutCacheCount = _a.computedWithoutCacheCount;
                var result = getBestFix(fixes, importingFile, program, host, preferences);
                return result && __assign(__assign({}, result), { computedWithoutCacheCount: computedWithoutCacheCount });
            }
            codefix.getModuleSpecifierForBestExportInfo = getModuleSpecifierForBestExportInfo;
            function isTypeOnlyPosition(sourceFile, position) {
                return ts.isValidTypeOnlyAliasUseSite(ts.getTokenAtPosition(sourceFile, position));
            }
            function getImportFixes(exportInfos, symbolName, 
            /** undefined only for missing JSX namespace */
            position, preferTypeOnlyImport, useRequire, program, sourceFile, host, preferences) {
                var checker = program.getTypeChecker();
                var existingImports = ts.flatMap(exportInfos, function (info) { return getExistingImportDeclarations(info, checker, sourceFile, program.getCompilerOptions()); });
                var useNamespace = position === undefined ? undefined : tryUseExistingNamespaceImport(existingImports, symbolName, position, checker);
                var addToExisting = tryAddToExistingImport(existingImports, position !== undefined && isTypeOnlyPosition(sourceFile, position));
                // Don't bother providing an action to add a new import if we can add to an existing one.
                var addImport = addToExisting ? [addToExisting] : getFixesForAddImport(exportInfos, existingImports, program, sourceFile, position, preferTypeOnlyImport, useRequire, host, preferences);
                return __spreadArray(__spreadArray([], (useNamespace ? [useNamespace] : ts.emptyArray), true), addImport, true);
            }
            function tryUseExistingNamespaceImport(existingImports, symbolName, position, checker) {
                // It is possible that multiple import statements with the same specifier exist in the file.
                // e.g.
                //
                //     import * as ns from "foo";
                //     import { member1, member2 } from "foo";
                //
                //     member3/**/ <-- cusor here
                //
                // in this case we should provie 2 actions:
                //     1. change "member3" to "ns.member3"
                //     2. add "member3" to the second import statement's import list
                // and it is up to the user to decide which one fits best.
                return ts.firstDefined(existingImports, function (_a) {
                    var declaration = _a.declaration;
                    var namespacePrefix = getNamespaceLikeImportText(declaration);
                    var moduleSpecifier = ts.tryGetModuleSpecifierFromDeclaration(declaration);
                    if (namespacePrefix && moduleSpecifier) {
                        var moduleSymbol = getTargetModuleFromNamespaceLikeImport(declaration, checker);
                        if (moduleSymbol && moduleSymbol.exports.has(ts.escapeLeadingUnderscores(symbolName))) {
                            return { kind: 0 /* UseNamespace */, namespacePrefix: namespacePrefix, position: position, moduleSpecifier: moduleSpecifier };
                        }
                    }
                });
            }
            function getTargetModuleFromNamespaceLikeImport(declaration, checker) {
                var _a;
                switch (declaration.kind) {
                    case 252 /* VariableDeclaration */:
                        return checker.resolveExternalModuleName(declaration.initializer.arguments[0]);
                    case 263 /* ImportEqualsDeclaration */:
                        return checker.getAliasedSymbol(declaration.symbol);
                    case 264 /* ImportDeclaration */:
                        var namespaceImport = ts.tryCast((_a = declaration.importClause) === null || _a === void 0 ? void 0 : _a.namedBindings, ts.isNamespaceImport);
                        return namespaceImport && checker.getAliasedSymbol(namespaceImport.symbol);
                    default:
                        return ts.Debug.assertNever(declaration);
                }
            }
            function getNamespaceLikeImportText(declaration) {
                var _a, _b, _c;
                switch (declaration.kind) {
                    case 252 /* VariableDeclaration */:
                        return (_a = ts.tryCast(declaration.name, ts.isIdentifier)) === null || _a === void 0 ? void 0 : _a.text;
                    case 263 /* ImportEqualsDeclaration */:
                        return declaration.name.text;
                    case 264 /* ImportDeclaration */:
                        return (_c = ts.tryCast((_b = declaration.importClause) === null || _b === void 0 ? void 0 : _b.namedBindings, ts.isNamespaceImport)) === null || _c === void 0 ? void 0 : _c.name.text;
                    default:
                        return ts.Debug.assertNever(declaration);
                }
            }
            function tryAddToExistingImport(existingImports, canUseTypeOnlyImport) {
                return ts.firstDefined(existingImports, function (_a) {
                    var declaration = _a.declaration, importKind = _a.importKind;
                    if (declaration.kind === 263 /* ImportEqualsDeclaration */)
                        return undefined;
                    if (declaration.kind === 252 /* VariableDeclaration */) {
                        return (importKind === 0 /* Named */ || importKind === 1 /* Default */) && declaration.name.kind === 199 /* ObjectBindingPattern */
                            ? { kind: 2 /* AddToExisting */, importClauseOrBindingPattern: declaration.name, importKind: importKind, moduleSpecifier: declaration.initializer.arguments[0].text, canUseTypeOnlyImport: false }
                            : undefined;
                    }
                    var importClause = declaration.importClause;
                    if (!importClause || !ts.isStringLiteralLike(declaration.moduleSpecifier))
                        return undefined;
                    var name = importClause.name, namedBindings = importClause.namedBindings;
                    // A type-only import may not have both a default and named imports, so the only way a name can
                    // be added to an existing type-only import is adding a named import to existing named bindings.
                    if (importClause.isTypeOnly && !(importKind === 0 /* Named */ && namedBindings))
                        return undefined;
                    return importKind === 1 /* Default */ && !name || importKind === 0 /* Named */ && (!namedBindings || namedBindings.kind === 267 /* NamedImports */)
                        ? { kind: 2 /* AddToExisting */, importClauseOrBindingPattern: importClause, importKind: importKind, moduleSpecifier: declaration.moduleSpecifier.text, canUseTypeOnlyImport: canUseTypeOnlyImport }
                        : undefined;
                });
            }
            function getExistingImportDeclarations(_a, checker, importingFile, compilerOptions) {
                var moduleSymbol = _a.moduleSymbol, exportKind = _a.exportKind, targetFlags = _a.targetFlags;
                // Can't use an es6 import for a type in JS.
                if (!(targetFlags & 111551 /* Value */) && ts.isSourceFileJS(importingFile))
                    return ts.emptyArray;
                var importKind = getImportKind(importingFile, exportKind, compilerOptions);
                return ts.mapDefined(importingFile.imports, function (moduleSpecifier) {
                    var i = ts.importFromModuleSpecifier(moduleSpecifier);
                    if (ts.isRequireVariableDeclaration(i.parent)) {
                        return checker.resolveExternalModuleName(moduleSpecifier) === moduleSymbol ? { declaration: i.parent, importKind: importKind } : undefined;
                    }
                    if (i.kind === 264 /* ImportDeclaration */ || i.kind === 263 /* ImportEqualsDeclaration */) {
                        return checker.getSymbolAtLocation(moduleSpecifier) === moduleSymbol ? { declaration: i, importKind: importKind } : undefined;
                    }
                });
            }
            function shouldUseRequire(sourceFile, program) {
                // 1. TypeScript files don't use require variable declarations
                if (!ts.isSourceFileJS(sourceFile)) {
                    return false;
                }
                // 2. If the current source file is unambiguously CJS or ESM, go with that
                if (sourceFile.commonJsModuleIndicator && !sourceFile.externalModuleIndicator)
                    return true;
                if (sourceFile.externalModuleIndicator && !sourceFile.commonJsModuleIndicator)
                    return false;
                // 3. If there's a tsconfig/jsconfig, use its module setting
                var compilerOptions = program.getCompilerOptions();
                if (compilerOptions.configFile) {
                    return ts.getEmitModuleKind(compilerOptions) < ts.ModuleKind.ES2015;
                }
                // 4. Match the first other JS file in the program that's unambiguously CJS or ESM
                for (var _i = 0, _a = program.getSourceFiles(); _i < _a.length; _i++) {
                    var otherFile = _a[_i];
                    if (otherFile === sourceFile || !ts.isSourceFileJS(otherFile) || program.isSourceFileFromExternalLibrary(otherFile))
                        continue;
                    if (otherFile.commonJsModuleIndicator && !otherFile.externalModuleIndicator)
                        return true;
                    if (otherFile.externalModuleIndicator && !otherFile.commonJsModuleIndicator)
                        return false;
                }
                // 5. Literally nothing to go on
                return true;
            }
            function getNewImportFixes(program, sourceFile, position, preferTypeOnlyImport, useRequire, moduleSymbols, host, preferences, fromCacheOnly) {
                var isJs = ts.isSourceFileJS(sourceFile);
                var compilerOptions = program.getCompilerOptions();
                var moduleSpecifierResolutionHost = ts.createModuleSpecifierResolutionHost(program, host);
                var checker = program.getTypeChecker();
                var getModuleSpecifiers = fromCacheOnly
                    ? function (moduleSymbol) { return ({ moduleSpecifiers: ts.moduleSpecifiers.tryGetModuleSpecifiersFromCache(moduleSymbol, sourceFile, moduleSpecifierResolutionHost, preferences), computedWithoutCache: false }); }
                    : function (moduleSymbol) { return ts.moduleSpecifiers.getModuleSpecifiersWithCacheInfo(moduleSymbol, checker, compilerOptions, sourceFile, moduleSpecifierResolutionHost, preferences); };
                var computedWithoutCacheCount = 0;
                var fixes = ts.flatMap(moduleSymbols, function (exportInfo) {
                    var _a = getModuleSpecifiers(exportInfo.moduleSymbol), computedWithoutCache = _a.computedWithoutCache, moduleSpecifiers = _a.moduleSpecifiers;
                    computedWithoutCacheCount += computedWithoutCache ? 1 : 0;
                    return moduleSpecifiers === null || moduleSpecifiers === void 0 ? void 0 : moduleSpecifiers.map(function (moduleSpecifier) {
                        // `position` should only be undefined at a missing jsx namespace, in which case we shouldn't be looking for pure types.
                        return !(exportInfo.targetFlags & 111551 /* Value */) && isJs && position !== undefined
                            ? { kind: 1 /* ImportType */, moduleSpecifier: moduleSpecifier, position: position, exportInfo: exportInfo }
                            : {
                                kind: 3 /* AddNew */,
                                moduleSpecifier: moduleSpecifier,
                                importKind: getImportKind(sourceFile, exportInfo.exportKind, compilerOptions),
                                useRequire: useRequire,
                                typeOnly: preferTypeOnlyImport,
                                exportInfo: exportInfo,
                            };
                    });
                });
                return { computedWithoutCacheCount: computedWithoutCacheCount, fixes: fixes };
            }
            function getFixesForAddImport(exportInfos, existingImports, program, sourceFile, position, preferTypeOnlyImport, useRequire, host, preferences) {
                var existingDeclaration = ts.firstDefined(existingImports, function (info) { return newImportInfoFromExistingSpecifier(info, preferTypeOnlyImport, useRequire); });
                return existingDeclaration ? [existingDeclaration] : getNewImportFixes(program, sourceFile, position, preferTypeOnlyImport, useRequire, exportInfos, host, preferences).fixes;
            }
            function newImportInfoFromExistingSpecifier(_a, preferTypeOnlyImport, useRequire) {
                var declaration = _a.declaration, importKind = _a.importKind;
                var moduleSpecifier = ts.tryGetModuleSpecifierFromDeclaration(declaration);
                return moduleSpecifier
                    ? { kind: 3 /* AddNew */, moduleSpecifier: moduleSpecifier, importKind: importKind, typeOnly: preferTypeOnlyImport, useRequire: useRequire }
                    : undefined;
            }
            function getFixesInfo(context, errorCode, pos, useAutoImportProvider) {
                var symbolToken = ts.getTokenAtPosition(context.sourceFile, pos);
                var info = errorCode === ts.Diagnostics._0_refers_to_a_UMD_global_but_the_current_file_is_a_module_Consider_adding_an_import_instead.code
                    ? getFixesInfoForUMDImport(context, symbolToken)
                    : ts.isIdentifier(symbolToken) ? getFixesInfoForNonUMDImport(context, symbolToken, useAutoImportProvider) : undefined;
                return info && __assign(__assign({}, info), { fixes: sortFixes(info.fixes, context.sourceFile, context.program, context.host, context.preferences) });
            }
            function sortFixes(fixes, sourceFile, program, host, preferences) {
                var allowsImportingSpecifier = ts.createPackageJsonImportFilter(sourceFile, preferences, host).allowsImportingSpecifier;
                return ts.sort(fixes, function (a, b) { return ts.compareValues(a.kind, b.kind) || compareModuleSpecifiers(a, b, sourceFile, program, allowsImportingSpecifier); });
            }
            function getBestFix(fixes, sourceFile, program, host, preferences) {
                if (!ts.some(fixes))
                    return;
                // These will always be placed first if available, and are better than other kinds
                if (fixes[0].kind === 0 /* UseNamespace */ || fixes[0].kind === 2 /* AddToExisting */) {
                    return fixes[0];
                }
                var allowsImportingSpecifier = ts.createPackageJsonImportFilter(sourceFile, preferences, host).allowsImportingSpecifier;
                return fixes.reduce(function (best, fix) {
                    return compareModuleSpecifiers(fix, best, sourceFile, program, allowsImportingSpecifier) === -1 /* LessThan */ ? fix : best;
                });
            }
            function compareModuleSpecifiers(a, b, importingFile, program, allowsImportingSpecifier) {
                if (a.kind !== 0 /* UseNamespace */ && b.kind !== 0 /* UseNamespace */) {
                    return ts.compareBooleans(allowsImportingSpecifier(a.moduleSpecifier), allowsImportingSpecifier(b.moduleSpecifier))
                        || compareNodeCoreModuleSpecifiers(a.moduleSpecifier, b.moduleSpecifier, importingFile, program)
                        || ts.compareNumberOfDirectorySeparators(a.moduleSpecifier, b.moduleSpecifier);
                }
                return 0 /* EqualTo */;
            }
            function compareNodeCoreModuleSpecifiers(a, b, importingFile, program) {
                if (ts.startsWith(a, "node:") && !ts.startsWith(b, "node:"))
                    return ts.shouldUseUriStyleNodeCoreModules(importingFile, program) ? -1 /* LessThan */ : 1 /* GreaterThan */;
                if (ts.startsWith(b, "node:") && !ts.startsWith(a, "node:"))
                    return ts.shouldUseUriStyleNodeCoreModules(importingFile, program) ? 1 /* GreaterThan */ : -1 /* LessThan */;
                return 0 /* EqualTo */;
            }
            function getFixesInfoForUMDImport(_a, token) {
                var sourceFile = _a.sourceFile, program = _a.program, host = _a.host, preferences = _a.preferences;
                var checker = program.getTypeChecker();
                var umdSymbol = getUmdSymbol(token, checker);
                if (!umdSymbol)
                    return undefined;
                var symbol = checker.getAliasedSymbol(umdSymbol);
                var symbolName = umdSymbol.name;
                var exportInfos = [{ symbol: umdSymbol, moduleSymbol: symbol, moduleFileName: undefined, exportKind: 3 /* UMD */, targetFlags: symbol.flags, isFromPackageJson: false }];
                var useRequire = shouldUseRequire(sourceFile, program);
                var fixes = getImportFixes(exportInfos, symbolName, ts.isIdentifier(token) ? token.getStart(sourceFile) : undefined, /*preferTypeOnlyImport*/ false, useRequire, program, sourceFile, host, preferences);
                return { fixes: fixes, symbolName: symbolName };
            }
            function getUmdSymbol(token, checker) {
                // try the identifier to see if it is the umd symbol
                var umdSymbol = ts.isIdentifier(token) ? checker.getSymbolAtLocation(token) : undefined;
                if (ts.isUMDExportSymbol(umdSymbol))
                    return umdSymbol;
                // The error wasn't for the symbolAtLocation, it was for the JSX tag itself, which needs access to e.g. `React`.
                var parent = token.parent;
                return (ts.isJsxOpeningLikeElement(parent) && parent.tagName === token) || ts.isJsxOpeningFragment(parent)
                    ? ts.tryCast(checker.resolveName(checker.getJsxNamespace(parent), ts.isJsxOpeningLikeElement(parent) ? token : parent, 111551 /* Value */, /*excludeGlobals*/ false), ts.isUMDExportSymbol)
                    : undefined;
            }
            /**
             * @param forceImportKeyword Indicates that the user has already typed `import`, so the result must start with `import`.
             * (In other words, do not allow `const x = require("...")` for JS files.)
             */
            function getImportKind(importingFile, exportKind, compilerOptions, forceImportKeyword) {
                switch (exportKind) {
                    case 0 /* Named */: return 0 /* Named */;
                    case 1 /* Default */: return 1 /* Default */;
                    case 2 /* ExportEquals */: return getExportEqualsImportKind(importingFile, compilerOptions, !!forceImportKeyword);
                    case 3 /* UMD */: return getUmdImportKind(importingFile, compilerOptions, !!forceImportKeyword);
                    default: return ts.Debug.assertNever(exportKind);
                }
            }
            codefix.getImportKind = getImportKind;
            function getUmdImportKind(importingFile, compilerOptions, forceImportKeyword) {
                // Import a synthetic `default` if enabled.
                if (ts.getAllowSyntheticDefaultImports(compilerOptions)) {
                    return 1 /* Default */;
                }
                // When a synthetic `default` is unavailable, use `import..require` if the module kind supports it.
                var moduleKind = ts.getEmitModuleKind(compilerOptions);
                switch (moduleKind) {
                    case ts.ModuleKind.AMD:
                    case ts.ModuleKind.CommonJS:
                    case ts.ModuleKind.UMD:
                        if (ts.isInJSFile(importingFile)) {
                            return ts.isExternalModule(importingFile) || forceImportKeyword ? 2 /* Namespace */ : 3 /* CommonJS */;
                        }
                        return 3 /* CommonJS */;
                    case ts.ModuleKind.System:
                    case ts.ModuleKind.ES2015:
                    case ts.ModuleKind.ES2020:
                    case ts.ModuleKind.ESNext:
                    case ts.ModuleKind.None:
                        // Fall back to the `import * as ns` style import.
                        return 2 /* Namespace */;
                    default:
                        return ts.Debug.assertNever(moduleKind, "Unexpected moduleKind " + moduleKind);
                }
            }
            function getFixesInfoForNonUMDImport(_a, symbolToken, useAutoImportProvider) {
                var sourceFile = _a.sourceFile, program = _a.program, cancellationToken = _a.cancellationToken, host = _a.host, preferences = _a.preferences;
                var checker = program.getTypeChecker();
                var compilerOptions = program.getCompilerOptions();
                var symbolName = getSymbolName(sourceFile, checker, symbolToken, compilerOptions);
                // "default" is a keyword and not a legal identifier for the import, so we don't expect it here
                ts.Debug.assert(symbolName !== "default" /* Default */, "'default' isn't a legal identifier and couldn't occur here");
                var preferTypeOnlyImport = compilerOptions.importsNotUsedAsValues === 2 /* Error */ && ts.isValidTypeOnlyAliasUseSite(symbolToken);
                var useRequire = shouldUseRequire(sourceFile, program);
                var exportInfos = getExportInfos(symbolName, ts.getMeaningFromLocation(symbolToken), cancellationToken, sourceFile, program, useAutoImportProvider, host, preferences);
                var fixes = ts.arrayFrom(ts.flatMapIterator(exportInfos.entries(), function (_a) {
                    var _ = _a[0], exportInfos = _a[1];
                    return getImportFixes(exportInfos, symbolName, symbolToken.getStart(sourceFile), preferTypeOnlyImport, useRequire, program, sourceFile, host, preferences);
                }));
                return { fixes: fixes, symbolName: symbolName };
            }
            function getSymbolName(sourceFile, checker, symbolToken, compilerOptions) {
                var parent = symbolToken.parent;
                if ((ts.isJsxOpeningLikeElement(parent) || ts.isJsxClosingElement(parent)) && parent.tagName === symbolToken && compilerOptions.jsx !== 4 /* ReactJSX */ && compilerOptions.jsx !== 5 /* ReactJSXDev */) {
                    var jsxNamespace = checker.getJsxNamespace(sourceFile);
                    if (ts.isIntrinsicJsxName(symbolToken.text) || !checker.resolveName(jsxNamespace, parent, 111551 /* Value */, /*excludeGlobals*/ true)) {
                        return jsxNamespace;
                    }
                }
                return symbolToken.text;
            }
            // Returns a map from an exported symbol's ID to a list of every way it's (re-)exported.
            function getExportInfos(symbolName, currentTokenMeaning, cancellationToken, fromFile, program, useAutoImportProvider, host, preferences) {
                var _a;
                // For each original symbol, keep all re-exports of that symbol together so we can call `getCodeActionsForImport` on the whole group at once.
                // Maps symbol id to info for modules providing that symbol (original export + re-exports).
                var originalSymbolToExportInfos = ts.createMultiMap();
                var packageJsonFilter = ts.createPackageJsonImportFilter(fromFile, preferences, host);
                var moduleSpecifierCache = (_a = host.getModuleSpecifierCache) === null || _a === void 0 ? void 0 : _a.call(host);
                var getModuleSpecifierResolutionHost = ts.memoizeOne(function (isFromPackageJson) {
                    return ts.createModuleSpecifierResolutionHost(isFromPackageJson ? host.getPackageJsonAutoImportProvider() : program, host);
                });
                function addSymbol(moduleSymbol, toFile, exportedSymbol, exportKind, program, isFromPackageJson) {
                    var moduleSpecifierResolutionHost = getModuleSpecifierResolutionHost(isFromPackageJson);
                    if (toFile && ts.isImportableFile(program, fromFile, toFile, preferences, packageJsonFilter, moduleSpecifierResolutionHost, moduleSpecifierCache) ||
                        !toFile && packageJsonFilter.allowsImportingAmbientModule(moduleSymbol, moduleSpecifierResolutionHost)) {
                        var checker = program.getTypeChecker();
                        originalSymbolToExportInfos.add(ts.getUniqueSymbolId(exportedSymbol, checker).toString(), { symbol: exportedSymbol, moduleSymbol: moduleSymbol, moduleFileName: toFile === null || toFile === void 0 ? void 0 : toFile.fileName, exportKind: exportKind, targetFlags: ts.skipAlias(exportedSymbol, checker).flags, isFromPackageJson: isFromPackageJson });
                    }
                }
                ts.forEachExternalModuleToImportFrom(program, host, useAutoImportProvider, function (moduleSymbol, sourceFile, program, isFromPackageJson) {
                    var checker = program.getTypeChecker();
                    cancellationToken.throwIfCancellationRequested();
                    var compilerOptions = program.getCompilerOptions();
                    var defaultInfo = ts.getDefaultLikeExportInfo(moduleSymbol, checker, compilerOptions);
                    if (defaultInfo && (defaultInfo.name === symbolName || moduleSymbolToValidIdentifier(moduleSymbol, compilerOptions.target) === symbolName) && symbolHasMeaning(defaultInfo.symbolForMeaning, currentTokenMeaning)) {
                        addSymbol(moduleSymbol, sourceFile, defaultInfo.symbol, defaultInfo.exportKind, program, isFromPackageJson);
                    }
                    // check exports with the same name
                    var exportSymbolWithIdenticalName = checker.tryGetMemberInModuleExportsAndProperties(symbolName, moduleSymbol);
                    if (exportSymbolWithIdenticalName && symbolHasMeaning(exportSymbolWithIdenticalName, currentTokenMeaning)) {
                        addSymbol(moduleSymbol, sourceFile, exportSymbolWithIdenticalName, 0 /* Named */, program, isFromPackageJson);
                    }
                });
                return originalSymbolToExportInfos;
            }
            function getExportEqualsImportKind(importingFile, compilerOptions, forceImportKeyword) {
                var allowSyntheticDefaults = ts.getAllowSyntheticDefaultImports(compilerOptions);
                var isJS = ts.isInJSFile(importingFile);
                // 1. 'import =' will not work in es2015+ TS files, so the decision is between a default
                //    and a namespace import, based on allowSyntheticDefaultImports/esModuleInterop.
                if (!isJS && ts.getEmitModuleKind(compilerOptions) >= ts.ModuleKind.ES2015) {
                    return allowSyntheticDefaults ? 1 /* Default */ : 2 /* Namespace */;
                }
                // 2. 'import =' will not work in JavaScript, so the decision is between a default import,
                //    a namespace import, and const/require.
                if (isJS) {
                    return ts.isExternalModule(importingFile) || forceImportKeyword
                        ? allowSyntheticDefaults ? 1 /* Default */ : 2 /* Namespace */
                        : 3 /* CommonJS */;
                }
                // 3. At this point the most correct choice is probably 'import =', but people
                //    really hate that, so look to see if the importing file has any precedent
                //    on how to handle it.
                for (var _i = 0, _a = importingFile.statements; _i < _a.length; _i++) {
                    var statement = _a[_i];
                    // `import foo` parses as an ImportEqualsDeclaration even though it could be an ImportDeclaration
                    if (ts.isImportEqualsDeclaration(statement) && !ts.nodeIsMissing(statement.moduleReference)) {
                        return 3 /* CommonJS */;
                    }
                }
                // 4. We have no precedent to go on, so just use a default import if
                //    allowSyntheticDefaultImports/esModuleInterop is enabled.
                return allowSyntheticDefaults ? 1 /* Default */ : 3 /* CommonJS */;
            }
            function codeActionForFix(context, sourceFile, symbolName, fix, quotePreference) {
                var diag;
                var changes = ts.textChanges.ChangeTracker.with(context, function (tracker) {
                    diag = codeActionForFixWorker(tracker, sourceFile, symbolName, fix, quotePreference);
                });
                return codefix.createCodeFixAction(codefix.importFixName, changes, diag, importFixId, ts.Diagnostics.Add_all_missing_imports);
            }
            function codeActionForFixWorker(changes, sourceFile, symbolName, fix, quotePreference) {
                switch (fix.kind) {
                    case 0 /* UseNamespace */:
                        addNamespaceQualifier(changes, sourceFile, fix);
                        return [ts.Diagnostics.Change_0_to_1, symbolName, fix.namespacePrefix + "." + symbolName];
                    case 1 /* ImportType */:
                        addImportType(changes, sourceFile, fix, quotePreference);
                        return [ts.Diagnostics.Change_0_to_1, symbolName, getImportTypePrefix(fix.moduleSpecifier, quotePreference) + symbolName];
                    case 2 /* AddToExisting */: {
                        var importClauseOrBindingPattern = fix.importClauseOrBindingPattern, importKind = fix.importKind, canUseTypeOnlyImport = fix.canUseTypeOnlyImport, moduleSpecifier = fix.moduleSpecifier;
                        doAddExistingFix(changes, sourceFile, importClauseOrBindingPattern, importKind === 1 /* Default */ ? symbolName : undefined, importKind === 0 /* Named */ ? [symbolName] : ts.emptyArray, canUseTypeOnlyImport);
                        var moduleSpecifierWithoutQuotes = ts.stripQuotes(moduleSpecifier);
                        return [importKind === 1 /* Default */ ? ts.Diagnostics.Add_default_import_0_to_existing_import_declaration_from_1 : ts.Diagnostics.Add_0_to_existing_import_declaration_from_1, symbolName, moduleSpecifierWithoutQuotes]; // you too!
                    }
                    case 3 /* AddNew */: {
                        var importKind = fix.importKind, moduleSpecifier = fix.moduleSpecifier, typeOnly = fix.typeOnly, useRequire = fix.useRequire;
                        var getDeclarations = useRequire ? getNewRequires : getNewImports;
                        var importsCollection = importKind === 1 /* Default */ ? { defaultImport: symbolName, typeOnly: typeOnly } :
                            importKind === 0 /* Named */ ? { namedImports: [symbolName], typeOnly: typeOnly } :
                                { namespaceLikeImport: { importKind: importKind, name: symbolName }, typeOnly: typeOnly };
                        ts.insertImports(changes, sourceFile, getDeclarations(moduleSpecifier, quotePreference, importsCollection), /*blankLineBetween*/ true);
                        return [importKind === 1 /* Default */ ? ts.Diagnostics.Import_default_0_from_module_1 : ts.Diagnostics.Import_0_from_module_1, symbolName, moduleSpecifier];
                    }
                    default:
                        return ts.Debug.assertNever(fix, "Unexpected fix kind " + fix.kind);
                }
            }
            function doAddExistingFix(changes, sourceFile, clause, defaultImport, namedImports, canUseTypeOnlyImport) {
                if (clause.kind === 199 /* ObjectBindingPattern */) {
                    if (defaultImport) {
                        addElementToBindingPattern(clause, defaultImport, "default");
                    }
                    for (var _i = 0, namedImports_1 = namedImports; _i < namedImports_1.length; _i++) {
                        var specifier = namedImports_1[_i];
                        addElementToBindingPattern(clause, specifier, /*propertyName*/ undefined);
                    }
                    return;
                }
                var convertTypeOnlyToRegular = !canUseTypeOnlyImport && clause.isTypeOnly;
                if (defaultImport) {
                    ts.Debug.assert(!clause.name, "Cannot add a default import to an import clause that already has one");
                    changes.insertNodeAt(sourceFile, clause.getStart(sourceFile), ts.factory.createIdentifier(defaultImport), { suffix: ", " });
                }
                if (namedImports.length) {
                    var existingSpecifiers = clause.namedBindings && ts.cast(clause.namedBindings, ts.isNamedImports).elements;
                    var newSpecifiers = ts.stableSort(namedImports.map(function (name) { return ts.factory.createImportSpecifier(/*propertyName*/ undefined, ts.factory.createIdentifier(name)); }), ts.OrganizeImports.compareImportOrExportSpecifiers);
                    if ((existingSpecifiers === null || existingSpecifiers === void 0 ? void 0 : existingSpecifiers.length) && ts.OrganizeImports.importSpecifiersAreSorted(existingSpecifiers)) {
                        for (var _a = 0, newSpecifiers_1 = newSpecifiers; _a < newSpecifiers_1.length; _a++) {
                            var spec = newSpecifiers_1[_a];
                            var insertionIndex = ts.OrganizeImports.getImportSpecifierInsertionIndex(existingSpecifiers, spec);
                            var prevSpecifier = clause.namedBindings.elements[insertionIndex - 1];
                            if (prevSpecifier) {
                                changes.insertNodeInListAfter(sourceFile, prevSpecifier, spec);
                            }
                            else {
                                changes.insertNodeBefore(sourceFile, existingSpecifiers[0], spec, !ts.positionsAreOnSameLine(existingSpecifiers[0].getStart(), clause.parent.getStart(), sourceFile));
                            }
                        }
                    }
                    else if (existingSpecifiers === null || existingSpecifiers === void 0 ? void 0 : existingSpecifiers.length) {
                        for (var _b = 0, newSpecifiers_2 = newSpecifiers; _b < newSpecifiers_2.length; _b++) {
                            var spec = newSpecifiers_2[_b];
                            changes.insertNodeInListAfter(sourceFile, ts.last(existingSpecifiers), spec, existingSpecifiers);
                        }
                    }
                    else {
                        if (newSpecifiers.length) {
                            var namedImports_2 = ts.factory.createNamedImports(newSpecifiers);
                            if (clause.namedBindings) {
                                changes.replaceNode(sourceFile, clause.namedBindings, namedImports_2);
                            }
                            else {
                                changes.insertNodeAfter(sourceFile, ts.Debug.checkDefined(clause.name, "Import clause must have either named imports or a default import"), namedImports_2);
                            }
                        }
                    }
                }
                if (convertTypeOnlyToRegular) {
                    changes.delete(sourceFile, ts.getTypeKeywordOfTypeOnlyImport(clause, sourceFile));
                }
                function addElementToBindingPattern(bindingPattern, name, propertyName) {
                    var element = ts.factory.createBindingElement(/*dotDotDotToken*/ undefined, propertyName, name);
                    if (bindingPattern.elements.length) {
                        changes.insertNodeInListAfter(sourceFile, ts.last(bindingPattern.elements), element);
                    }
                    else {
                        changes.replaceNode(sourceFile, bindingPattern, ts.factory.createObjectBindingPattern([element]));
                    }
                }
            }
            function addNamespaceQualifier(changes, sourceFile, _a) {
                var namespacePrefix = _a.namespacePrefix, position = _a.position;
                changes.insertText(sourceFile, position, namespacePrefix + ".");
            }
            function addImportType(changes, sourceFile, _a, quotePreference) {
                var moduleSpecifier = _a.moduleSpecifier, position = _a.position;
                changes.insertText(sourceFile, position, getImportTypePrefix(moduleSpecifier, quotePreference));
            }
            function getImportTypePrefix(moduleSpecifier, quotePreference) {
                var quote = ts.getQuoteFromPreference(quotePreference);
                return "import(" + quote + moduleSpecifier + quote + ").";
            }
            function getNewImports(moduleSpecifier, quotePreference, imports) {
                var _a, _b;
                var quotedModuleSpecifier = ts.makeStringLiteral(moduleSpecifier, quotePreference);
                var statements;
                if (imports.defaultImport !== undefined || ((_a = imports.namedImports) === null || _a === void 0 ? void 0 : _a.length)) {
                    statements = ts.combine(statements, ts.makeImport(imports.defaultImport === undefined ? undefined : ts.factory.createIdentifier(imports.defaultImport), (_b = imports.namedImports) === null || _b === void 0 ? void 0 : _b.map(function (n) { return ts.factory.createImportSpecifier(/*propertyName*/ undefined, ts.factory.createIdentifier(n)); }), moduleSpecifier, quotePreference, imports.typeOnly));
                }
                var namespaceLikeImport = imports.namespaceLikeImport, typeOnly = imports.typeOnly;
                if (namespaceLikeImport) {
                    var declaration = namespaceLikeImport.importKind === 3 /* CommonJS */
                        ? ts.factory.createImportEqualsDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, typeOnly, ts.factory.createIdentifier(namespaceLikeImport.name), ts.factory.createExternalModuleReference(quotedModuleSpecifier))
                        : ts.factory.createImportDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, ts.factory.createImportClause(typeOnly, 
                        /*name*/ undefined, ts.factory.createNamespaceImport(ts.factory.createIdentifier(namespaceLikeImport.name))), quotedModuleSpecifier);
                    statements = ts.combine(statements, declaration);
                }
                return ts.Debug.checkDefined(statements);
            }
            function getNewRequires(moduleSpecifier, quotePreference, imports) {
                var _a, _b;
                var quotedModuleSpecifier = ts.makeStringLiteral(moduleSpecifier, quotePreference);
                var statements;
                // const { default: foo, bar, etc } = require('./mod');
                if (imports.defaultImport || ((_a = imports.namedImports) === null || _a === void 0 ? void 0 : _a.length)) {
                    var bindingElements = ((_b = imports.namedImports) === null || _b === void 0 ? void 0 : _b.map(function (name) { return ts.factory.createBindingElement(/*dotDotDotToken*/ undefined, /*propertyName*/ undefined, name); })) || [];
                    if (imports.defaultImport) {
                        bindingElements.unshift(ts.factory.createBindingElement(/*dotDotDotToken*/ undefined, "default", imports.defaultImport));
                    }
                    var declaration = createConstEqualsRequireDeclaration(ts.factory.createObjectBindingPattern(bindingElements), quotedModuleSpecifier);
                    statements = ts.combine(statements, declaration);
                }
                // const foo = require('./mod');
                if (imports.namespaceLikeImport) {
                    var declaration = createConstEqualsRequireDeclaration(imports.namespaceLikeImport.name, quotedModuleSpecifier);
                    statements = ts.combine(statements, declaration);
                }
                return ts.Debug.checkDefined(statements);
            }
            function createConstEqualsRequireDeclaration(name, quotedModuleSpecifier) {
                return ts.factory.createVariableStatement(
                /*modifiers*/ undefined, ts.factory.createVariableDeclarationList([
                    ts.factory.createVariableDeclaration(typeof name === "string" ? ts.factory.createIdentifier(name) : name, 
                    /*exclamationToken*/ undefined, 
                    /*type*/ undefined, ts.factory.createCallExpression(ts.factory.createIdentifier("require"), /*typeArguments*/ undefined, [quotedModuleSpecifier]))
                ], 2 /* Const */));
            }
            function symbolHasMeaning(_a, meaning) {
                var declarations = _a.declarations;
                return ts.some(declarations, function (decl) { return !!(ts.getMeaningFromDeclaration(decl) & meaning); });
            }
            function moduleSymbolToValidIdentifier(moduleSymbol, target) {
                return moduleSpecifierToValidIdentifier(ts.removeFileExtension(ts.stripQuotes(moduleSymbol.name)), target);
            }
            codefix.moduleSymbolToValidIdentifier = moduleSymbolToValidIdentifier;
            function moduleSpecifierToValidIdentifier(moduleSpecifier, target) {
                var baseName = ts.getBaseFileName(ts.removeSuffix(moduleSpecifier, "/index"));
                var res = "";
                var lastCharWasValid = true;
                var firstCharCode = baseName.charCodeAt(0);
                if (ts.isIdentifierStart(firstCharCode, target)) {
                    res += String.fromCharCode(firstCharCode);
                }
                else {
                    lastCharWasValid = false;
                }
                for (var i = 1; i < baseName.length; i++) {
                    var ch = baseName.charCodeAt(i);
                    var isValid = ts.isIdentifierPart(ch, target);
                    if (isValid) {
                        var char = String.fromCharCode(ch);
                        if (!lastCharWasValid) {
                            char = char.toUpperCase();
                        }
                        res += char;
                    }
                    lastCharWasValid = isValid;
                }
                // Need `|| "_"` to ensure result isn't empty.
                return !ts.isStringANonContextualKeyword(res) ? res || "_" : "_" + res;
            }
            codefix.moduleSpecifierToValidIdentifier = moduleSpecifierToValidIdentifier;
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var _a;
            var fixName = "fixOverrideModifier";
            var fixAddOverrideId = "fixAddOverrideModifier";
            var fixRemoveOverrideId = "fixRemoveOverrideModifier";
            var errorCodes = [
                ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0.code,
                ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_its_containing_class_0_does_not_extend_another_class.code,
                ts.Diagnostics.This_member_must_have_an_override_modifier_because_it_overrides_an_abstract_method_that_is_declared_in_the_base_class_0.code,
                ts.Diagnostics.This_member_must_have_an_override_modifier_because_it_overrides_a_member_in_the_base_class_0.code,
                ts.Diagnostics.This_parameter_property_must_have_an_override_modifier_because_it_overrides_a_member_in_base_class_0.code
            ];
            var errorCodeFixIdMap = (_a = {},
                _a[ts.Diagnostics.This_member_must_have_an_override_modifier_because_it_overrides_a_member_in_the_base_class_0.code] = [
                    ts.Diagnostics.Add_override_modifier, fixAddOverrideId, ts.Diagnostics.Add_all_missing_override_modifiers,
                ],
                _a[ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_its_containing_class_0_does_not_extend_another_class.code] = [
                    ts.Diagnostics.Remove_override_modifier, fixRemoveOverrideId, ts.Diagnostics.Remove_all_unnecessary_override_modifiers
                ],
                _a[ts.Diagnostics.This_parameter_property_must_have_an_override_modifier_because_it_overrides_a_member_in_base_class_0.code] = [
                    ts.Diagnostics.Add_override_modifier, fixAddOverrideId, ts.Diagnostics.Add_all_missing_override_modifiers,
                ],
                _a[ts.Diagnostics.This_member_must_have_an_override_modifier_because_it_overrides_an_abstract_method_that_is_declared_in_the_base_class_0.code] = [
                    ts.Diagnostics.Add_override_modifier, fixAddOverrideId, ts.Diagnostics.Remove_all_unnecessary_override_modifiers
                ],
                _a[ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0.code] = [
                    ts.Diagnostics.Remove_override_modifier, fixRemoveOverrideId, ts.Diagnostics.Remove_all_unnecessary_override_modifiers
                ],
                _a);
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var errorCode = context.errorCode, span = context.span, sourceFile = context.sourceFile;
                    var info = errorCodeFixIdMap[errorCode];
                    if (!info)
                        return ts.emptyArray;
                    var descriptions = info[0], fixId = info[1], fixAllDescriptions = info[2];
                    if (ts.isSourceFileJS(sourceFile))
                        return ts.emptyArray;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (changes) { return dispatchChanges(changes, context, errorCode, span.start); });
                    return [
                        codefix.createCodeFixActionMaybeFixAll(fixName, changes, descriptions, fixId, fixAllDescriptions)
                    ];
                },
                fixIds: [fixName, fixAddOverrideId, fixRemoveOverrideId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var code = diag.code, start = diag.start, file = diag.file;
                        var info = errorCodeFixIdMap[code];
                        if (!info || info[1] !== context.fixId || ts.isSourceFileJS(file)) {
                            return;
                        }
                        dispatchChanges(changes, context, code, start);
                    });
                }
            });
            function dispatchChanges(changeTracker, context, errorCode, pos) {
                switch (errorCode) {
                    case ts.Diagnostics.This_member_must_have_an_override_modifier_because_it_overrides_a_member_in_the_base_class_0.code:
                    case ts.Diagnostics.This_member_must_have_an_override_modifier_because_it_overrides_an_abstract_method_that_is_declared_in_the_base_class_0.code:
                    case ts.Diagnostics.This_parameter_property_must_have_an_override_modifier_because_it_overrides_a_member_in_base_class_0.code:
                        return doAddOverrideModifierChange(changeTracker, context.sourceFile, pos);
                    case ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0.code:
                    case ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_its_containing_class_0_does_not_extend_another_class.code:
                        return doRemoveOverrideModifierChange(changeTracker, context.sourceFile, pos);
                    default:
                        ts.Debug.fail("Unexpected error code: " + errorCode);
                }
            }
            function doAddOverrideModifierChange(changeTracker, sourceFile, pos) {
                var classElement = findContainerClassElementLike(sourceFile, pos);
                var modifiers = classElement.modifiers || ts.emptyArray;
                var staticModifier = ts.find(modifiers, ts.isStaticModifier);
                var abstractModifier = ts.find(modifiers, ts.isAbstractModifier);
                var accessibilityModifier = ts.find(modifiers, function (m) { return ts.isAccessibilityModifier(m.kind); });
                var modifierPos = abstractModifier ? abstractModifier.end :
                    staticModifier ? staticModifier.end :
                        accessibilityModifier ? accessibilityModifier.end :
                            classElement.decorators ? ts.skipTrivia(sourceFile.text, classElement.decorators.end) : classElement.getStart(sourceFile);
                var options = accessibilityModifier || staticModifier || abstractModifier ? { prefix: " " } : { suffix: " " };
                changeTracker.insertModifierAt(sourceFile, modifierPos, 157 /* OverrideKeyword */, options);
            }
            function doRemoveOverrideModifierChange(changeTracker, sourceFile, pos) {
                var classElement = findContainerClassElementLike(sourceFile, pos);
                var overrideModifier = classElement.modifiers && ts.find(classElement.modifiers, function (modifier) { return modifier.kind === 157 /* OverrideKeyword */; });
                ts.Debug.assertIsDefined(overrideModifier);
                changeTracker.deleteModifier(sourceFile, overrideModifier);
            }
            function isClassElementLikeHasJSDoc(node) {
                switch (node.kind) {
                    case 169 /* Constructor */:
                    case 165 /* PropertyDeclaration */:
                    case 167 /* MethodDeclaration */:
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */:
                        return true;
                    case 162 /* Parameter */:
                        return ts.isParameterPropertyDeclaration(node, node.parent);
                    default:
                        return false;
                }
            }
            function findContainerClassElementLike(sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                var classElement = ts.findAncestor(token, function (node) {
                    if (ts.isClassLike(node))
                        return "quit";
                    return isClassElementLikeHasJSDoc(node);
                });
                ts.Debug.assert(classElement && isClassElementLikeHasJSDoc(classElement));
                return classElement;
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixNoPropertyAccessFromIndexSignature";
            var errorCodes = [
                ts.Diagnostics.Property_0_comes_from_an_index_signature_so_it_must_be_accessed_with_0.code
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                fixIds: [fixId],
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span, preferences = context.preferences;
                    var property = getPropertyAccessExpression(sourceFile, span.start);
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, context.sourceFile, property, preferences); });
                    return [codefix.createCodeFixAction(fixId, changes, [ts.Diagnostics.Use_element_access_for_0, property.name.text], fixId, ts.Diagnostics.Use_element_access_for_all_undeclared_properties)];
                },
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return doChange(changes, diag.file, getPropertyAccessExpression(diag.file, diag.start), context.preferences); });
                }
            });
            function doChange(changes, sourceFile, node, preferences) {
                var quotePreference = ts.getQuotePreference(sourceFile, preferences);
                var argumentsExpression = ts.factory.createStringLiteral(node.name.text, quotePreference === 0 /* Single */);
                changes.replaceNode(sourceFile, node, ts.isPropertyAccessChain(node) ?
                    ts.factory.createElementAccessChain(node.expression, node.questionDotToken, argumentsExpression) :
                    ts.factory.createElementAccessExpression(node.expression, argumentsExpression));
            }
            function getPropertyAccessExpression(sourceFile, pos) {
                return ts.cast(ts.getTokenAtPosition(sourceFile, pos).parent, ts.isPropertyAccessExpression);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixImplicitThis";
            var errorCodes = [ts.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, program = context.program, span = context.span;
                    var diagnostic;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) {
                        diagnostic = doChange(t, sourceFile, span.start, program.getTypeChecker());
                    });
                    return diagnostic ? [codefix.createCodeFixAction(fixId, changes, diagnostic, fixId, ts.Diagnostics.Fix_all_implicit_this_errors)] : ts.emptyArray;
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        doChange(changes, diag.file, diag.start, context.program.getTypeChecker());
                    });
                },
            });
            function doChange(changes, sourceFile, pos, checker) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                ts.Debug.assert(token.kind === 108 /* ThisKeyword */);
                var fn = ts.getThisContainer(token, /*includeArrowFunctions*/ false);
                if (!ts.isFunctionDeclaration(fn) && !ts.isFunctionExpression(fn))
                    return undefined;
                if (!ts.isSourceFile(ts.getThisContainer(fn, /*includeArrowFunctions*/ false))) { // 'this' is defined outside, convert to arrow function
                    var fnKeyword = ts.Debug.assertDefined(ts.findChildOfKind(fn, 98 /* FunctionKeyword */, sourceFile));
                    var name = fn.name;
                    var body = ts.Debug.assertDefined(fn.body); // Should be defined because the function contained a 'this' expression
                    if (ts.isFunctionExpression(fn)) {
                        if (name && ts.FindAllReferences.Core.isSymbolReferencedInFile(name, checker, sourceFile, body)) {
                            // Function expression references itself. To fix we would have to extract it to a const.
                            return undefined;
                        }
                        // `function() {}` --> `() => {}`
                        changes.delete(sourceFile, fnKeyword);
                        if (name) {
                            changes.delete(sourceFile, name);
                        }
                        changes.insertText(sourceFile, body.pos, " =>");
                        return [ts.Diagnostics.Convert_function_expression_0_to_arrow_function, name ? name.text : ts.ANONYMOUS];
                    }
                    else {
                        // `function f() {}` => `const f = () => {}`
                        // `name` should be defined because we only do this in inner contexts, and name is only undefined for `export default function() {}`.
                        changes.replaceNode(sourceFile, fnKeyword, ts.factory.createToken(85 /* ConstKeyword */));
                        changes.insertText(sourceFile, name.end, " = ");
                        changes.insertText(sourceFile, body.pos, " =>");
                        return [ts.Diagnostics.Convert_function_declaration_0_to_arrow_function, name.text];
                    }
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixIncorrectNamedTupleSyntax";
            var errorCodes = [
                ts.Diagnostics.A_labeled_tuple_element_is_declared_as_optional_with_a_question_mark_after_the_name_and_before_the_colon_rather_than_after_the_type.code,
                ts.Diagnostics.A_labeled_tuple_element_is_declared_as_rest_with_a_before_the_name_rather_than_before_the_type.code
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var namedTupleMember = getNamedTupleMember(sourceFile, span.start);
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, namedTupleMember); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Move_labeled_tuple_element_modifiers_to_labels, fixId, ts.Diagnostics.Move_labeled_tuple_element_modifiers_to_labels)];
                },
                fixIds: [fixId]
            });
            function getNamedTupleMember(sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                return ts.findAncestor(token, function (t) { return t.kind === 195 /* NamedTupleMember */; });
            }
            function doChange(changes, sourceFile, namedTupleMember) {
                if (!namedTupleMember) {
                    return;
                }
                var unwrappedType = namedTupleMember.type;
                var sawOptional = false;
                var sawRest = false;
                while (unwrappedType.kind === 183 /* OptionalType */ || unwrappedType.kind === 184 /* RestType */ || unwrappedType.kind === 189 /* ParenthesizedType */) {
                    if (unwrappedType.kind === 183 /* OptionalType */) {
                        sawOptional = true;
                    }
                    else if (unwrappedType.kind === 184 /* RestType */) {
                        sawRest = true;
                    }
                    unwrappedType = unwrappedType.type;
                }
                var updated = ts.factory.updateNamedTupleMember(namedTupleMember, namedTupleMember.dotDotDotToken || (sawRest ? ts.factory.createToken(25 /* DotDotDotToken */) : undefined), namedTupleMember.name, namedTupleMember.questionToken || (sawOptional ? ts.factory.createToken(57 /* QuestionToken */) : undefined), unwrappedType);
                if (updated === namedTupleMember) {
                    return;
                }
                changes.replaceNode(sourceFile, namedTupleMember, updated);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixSpelling";
            var errorCodes = [
                ts.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2.code,
                ts.Diagnostics.Property_0_may_not_exist_on_type_1_Did_you_mean_2.code,
                ts.Diagnostics.Cannot_find_name_0_Did_you_mean_1.code,
                ts.Diagnostics.Could_not_find_name_0_Did_you_mean_1.code,
                ts.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0.code,
                ts.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0.code,
                ts.Diagnostics._0_has_no_exported_member_named_1_Did_you_mean_2.code,
                ts.Diagnostics.This_member_cannot_have_an_override_modifier_because_it_is_not_declared_in_the_base_class_0_Did_you_mean_1.code,
                // for JSX class components
                ts.Diagnostics.No_overload_matches_this_call.code,
                // for JSX FC
                ts.Diagnostics.Type_0_is_not_assignable_to_type_1.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, errorCode = context.errorCode;
                    var info = getInfo(sourceFile, context.span.start, context, errorCode);
                    if (!info)
                        return undefined;
                    var node = info.node, suggestedSymbol = info.suggestedSymbol;
                    var target = context.host.getCompilationSettings().target;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, node, suggestedSymbol, target); });
                    return [codefix.createCodeFixAction("spelling", changes, [ts.Diagnostics.Change_spelling_to_0, ts.symbolName(suggestedSymbol)], fixId, ts.Diagnostics.Fix_all_detected_spelling_errors)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var info = getInfo(diag.file, diag.start, context, diag.code);
                        var target = context.host.getCompilationSettings().target;
                        if (info)
                            doChange(changes, context.sourceFile, info.node, info.suggestedSymbol, target);
                    });
                },
            });
            function getInfo(sourceFile, pos, context, errorCode) {
                // This is the identifier of the misspelled word. eg:
                // this.speling = 1;
                //      ^^^^^^^
                var node = ts.getTokenAtPosition(sourceFile, pos);
                var parent = node.parent;
                // Only fix spelling for No_overload_matches_this_call emitted on the React class component
                if ((errorCode === ts.Diagnostics.No_overload_matches_this_call.code ||
                    errorCode === ts.Diagnostics.Type_0_is_not_assignable_to_type_1.code) &&
                    !ts.isJsxAttribute(parent))
                    return undefined;
                var checker = context.program.getTypeChecker();
                var suggestedSymbol;
                if (ts.isPropertyAccessExpression(parent) && parent.name === node) {
                    ts.Debug.assert(ts.isMemberName(node), "Expected an identifier for spelling (property access)");
                    var containingType = checker.getTypeAtLocation(parent.expression);
                    if (parent.flags & 32 /* OptionalChain */) {
                        containingType = checker.getNonNullableType(containingType);
                    }
                    suggestedSymbol = checker.getSuggestedSymbolForNonexistentProperty(node, containingType);
                }
                else if (ts.isQualifiedName(parent) && parent.right === node) {
                    var symbol = checker.getSymbolAtLocation(parent.left);
                    if (symbol && symbol.flags & 1536 /* Module */) {
                        suggestedSymbol = checker.getSuggestedSymbolForNonexistentModule(parent.right, symbol);
                    }
                }
                else if (ts.isImportSpecifier(parent) && parent.name === node) {
                    ts.Debug.assertNode(node, ts.isIdentifier, "Expected an identifier for spelling (import)");
                    var importDeclaration = ts.findAncestor(node, ts.isImportDeclaration);
                    var resolvedSourceFile = getResolvedSourceFileFromImportDeclaration(sourceFile, context, importDeclaration);
                    if (resolvedSourceFile && resolvedSourceFile.symbol) {
                        suggestedSymbol = checker.getSuggestedSymbolForNonexistentModule(node, resolvedSourceFile.symbol);
                    }
                }
                else if (ts.isJsxAttribute(parent) && parent.name === node) {
                    ts.Debug.assertNode(node, ts.isIdentifier, "Expected an identifier for JSX attribute");
                    var tag = ts.findAncestor(node, ts.isJsxOpeningLikeElement);
                    var props = checker.getContextualTypeForArgumentAtIndex(tag, 0);
                    suggestedSymbol = checker.getSuggestedSymbolForNonexistentJSXAttribute(node, props);
                }
                else if (ts.hasSyntacticModifier(parent, 16384 /* Override */) && ts.isClassElement(parent) && parent.name === node) {
                    var baseDeclaration = ts.findAncestor(node, ts.isClassLike);
                    var baseTypeNode = baseDeclaration ? ts.getEffectiveBaseTypeNode(baseDeclaration) : undefined;
                    var baseType = baseTypeNode ? checker.getTypeAtLocation(baseTypeNode) : undefined;
                    if (baseType) {
                        suggestedSymbol = checker.getSuggestedSymbolForNonexistentClassMember(ts.getTextOfNode(node), baseType);
                    }
                }
                else {
                    var meaning = ts.getMeaningFromLocation(node);
                    var name = ts.getTextOfNode(node);
                    ts.Debug.assert(name !== undefined, "name should be defined");
                    suggestedSymbol = checker.getSuggestedSymbolForNonexistentSymbol(node, name, convertSemanticMeaningToSymbolFlags(meaning));
                }
                return suggestedSymbol === undefined ? undefined : { node: node, suggestedSymbol: suggestedSymbol };
            }
            function doChange(changes, sourceFile, node, suggestedSymbol, target) {
                var suggestion = ts.symbolName(suggestedSymbol);
                if (!ts.isIdentifierText(suggestion, target) && ts.isPropertyAccessExpression(node.parent)) {
                    var valDecl = suggestedSymbol.valueDeclaration;
                    if (valDecl && ts.isNamedDeclaration(valDecl) && ts.isPrivateIdentifier(valDecl.name)) {
                        changes.replaceNode(sourceFile, node, ts.factory.createIdentifier(suggestion));
                    }
                    else {
                        changes.replaceNode(sourceFile, node.parent, ts.factory.createElementAccessExpression(node.parent.expression, ts.factory.createStringLiteral(suggestion)));
                    }
                }
                else {
                    changes.replaceNode(sourceFile, node, ts.factory.createIdentifier(suggestion));
                }
            }
            function convertSemanticMeaningToSymbolFlags(meaning) {
                var flags = 0;
                if (meaning & 4 /* Namespace */) {
                    flags |= 1920 /* Namespace */;
                }
                if (meaning & 2 /* Type */) {
                    flags |= 788968 /* Type */;
                }
                if (meaning & 1 /* Value */) {
                    flags |= 111551 /* Value */;
                }
                return flags;
            }
            function getResolvedSourceFileFromImportDeclaration(sourceFile, context, importDeclaration) {
                if (!importDeclaration || !ts.isStringLiteralLike(importDeclaration.moduleSpecifier))
                    return undefined;
                var resolvedModule = ts.getResolvedModule(sourceFile, importDeclaration.moduleSpecifier.text);
                if (!resolvedModule)
                    return undefined;
                return context.program.getSourceFile(resolvedModule.resolvedFileName);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "returnValueCorrect";
            var fixIdAddReturnStatement = "fixAddReturnStatement";
            var fixRemoveBracesFromArrowFunctionBody = "fixRemoveBracesFromArrowFunctionBody";
            var fixIdWrapTheBlockWithParen = "fixWrapTheBlockWithParen";
            var errorCodes = [
                ts.Diagnostics.A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value.code,
                ts.Diagnostics.Type_0_is_not_assignable_to_type_1.code,
                ts.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code
            ];
            var ProblemKind;
            (function (ProblemKind) {
                ProblemKind[ProblemKind["MissingReturnStatement"] = 0] = "MissingReturnStatement";
                ProblemKind[ProblemKind["MissingParentheses"] = 1] = "MissingParentheses";
            })(ProblemKind || (ProblemKind = {}));
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                fixIds: [fixIdAddReturnStatement, fixRemoveBracesFromArrowFunctionBody, fixIdWrapTheBlockWithParen],
                getCodeActions: function (context) {
                    var program = context.program, sourceFile = context.sourceFile, start = context.span.start, errorCode = context.errorCode;
                    var info = getInfo(program.getTypeChecker(), sourceFile, start, errorCode);
                    if (!info)
                        return undefined;
                    if (info.kind === ProblemKind.MissingReturnStatement) {
                        return ts.append([getActionForfixAddReturnStatement(context, info.expression, info.statement)], ts.isArrowFunction(info.declaration) ? getActionForFixRemoveBracesFromArrowFunctionBody(context, info.declaration, info.expression, info.commentSource) : undefined);
                    }
                    else {
                        return [getActionForfixWrapTheBlockWithParen(context, info.declaration, info.expression)];
                    }
                },
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var info = getInfo(context.program.getTypeChecker(), diag.file, diag.start, diag.code);
                        if (!info)
                            return undefined;
                        switch (context.fixId) {
                            case fixIdAddReturnStatement:
                                addReturnStatement(changes, diag.file, info.expression, info.statement);
                                break;
                            case fixRemoveBracesFromArrowFunctionBody:
                                if (!ts.isArrowFunction(info.declaration))
                                    return undefined;
                                removeBlockBodyBrace(changes, diag.file, info.declaration, info.expression, info.commentSource, /* withParen */ false);
                                break;
                            case fixIdWrapTheBlockWithParen:
                                if (!ts.isArrowFunction(info.declaration))
                                    return undefined;
                                wrapBlockWithParen(changes, diag.file, info.declaration, info.expression);
                                break;
                            default:
                                ts.Debug.fail(JSON.stringify(context.fixId));
                        }
                    });
                },
            });
            function createObjectTypeFromLabeledExpression(checker, label, expression) {
                var member = checker.createSymbol(4 /* Property */, label.escapedText);
                member.type = checker.getTypeAtLocation(expression);
                var members = ts.createSymbolTable([member]);
                return checker.createAnonymousType(/*symbol*/ undefined, members, [], [], []);
            }
            function getFixInfo(checker, declaration, expectType, isFunctionType) {
                if (!declaration.body || !ts.isBlock(declaration.body) || ts.length(declaration.body.statements) !== 1)
                    return undefined;
                var firstStatement = ts.first(declaration.body.statements);
                if (ts.isExpressionStatement(firstStatement) && checkFixedAssignableTo(checker, declaration, checker.getTypeAtLocation(firstStatement.expression), expectType, isFunctionType)) {
                    return {
                        declaration: declaration,
                        kind: ProblemKind.MissingReturnStatement,
                        expression: firstStatement.expression,
                        statement: firstStatement,
                        commentSource: firstStatement.expression
                    };
                }
                else if (ts.isLabeledStatement(firstStatement) && ts.isExpressionStatement(firstStatement.statement)) {
                    var node = ts.factory.createObjectLiteralExpression([ts.factory.createPropertyAssignment(firstStatement.label, firstStatement.statement.expression)]);
                    var nodeType = createObjectTypeFromLabeledExpression(checker, firstStatement.label, firstStatement.statement.expression);
                    if (checkFixedAssignableTo(checker, declaration, nodeType, expectType, isFunctionType)) {
                        return ts.isArrowFunction(declaration) ? {
                            declaration: declaration,
                            kind: ProblemKind.MissingParentheses,
                            expression: node,
                            statement: firstStatement,
                            commentSource: firstStatement.statement.expression
                        } : {
                            declaration: declaration,
                            kind: ProblemKind.MissingReturnStatement,
                            expression: node,
                            statement: firstStatement,
                            commentSource: firstStatement.statement.expression
                        };
                    }
                }
                else if (ts.isBlock(firstStatement) && ts.length(firstStatement.statements) === 1) {
                    var firstBlockStatement = ts.first(firstStatement.statements);
                    if (ts.isLabeledStatement(firstBlockStatement) && ts.isExpressionStatement(firstBlockStatement.statement)) {
                        var node = ts.factory.createObjectLiteralExpression([ts.factory.createPropertyAssignment(firstBlockStatement.label, firstBlockStatement.statement.expression)]);
                        var nodeType = createObjectTypeFromLabeledExpression(checker, firstBlockStatement.label, firstBlockStatement.statement.expression);
                        if (checkFixedAssignableTo(checker, declaration, nodeType, expectType, isFunctionType)) {
                            return {
                                declaration: declaration,
                                kind: ProblemKind.MissingReturnStatement,
                                expression: node,
                                statement: firstStatement,
                                commentSource: firstBlockStatement
                            };
                        }
                    }
                }
                return undefined;
            }
            function checkFixedAssignableTo(checker, declaration, exprType, type, isFunctionType) {
                if (isFunctionType) {
                    var sig = checker.getSignatureFromDeclaration(declaration);
                    if (sig) {
                        if (ts.hasSyntacticModifier(declaration, 256 /* Async */)) {
                            exprType = checker.createPromiseType(exprType);
                        }
                        var newSig = checker.createSignature(declaration, sig.typeParameters, sig.thisParameter, sig.parameters, exprType, 
                        /*typePredicate*/ undefined, sig.minArgumentCount, sig.flags);
                        exprType = checker.createAnonymousType(
                        /*symbol*/ undefined, ts.createSymbolTable(), [newSig], [], []);
                    }
                    else {
                        exprType = checker.getAnyType();
                    }
                }
                return checker.isTypeAssignableTo(exprType, type);
            }
            function getInfo(checker, sourceFile, position, errorCode) {
                var node = ts.getTokenAtPosition(sourceFile, position);
                if (!node.parent)
                    return undefined;
                var declaration = ts.findAncestor(node.parent, ts.isFunctionLikeDeclaration);
                switch (errorCode) {
                    case ts.Diagnostics.A_function_whose_declared_type_is_neither_void_nor_any_must_return_a_value.code:
                        if (!declaration || !declaration.body || !declaration.type || !ts.rangeContainsRange(declaration.type, node))
                            return undefined;
                        return getFixInfo(checker, declaration, checker.getTypeFromTypeNode(declaration.type), /* isFunctionType */ false);
                    case ts.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code:
                        if (!declaration || !ts.isCallExpression(declaration.parent) || !declaration.body)
                            return undefined;
                        var pos = declaration.parent.arguments.indexOf(declaration);
                        var type = checker.getContextualTypeForArgumentAtIndex(declaration.parent, pos);
                        if (!type)
                            return undefined;
                        return getFixInfo(checker, declaration, type, /* isFunctionType */ true);
                    case ts.Diagnostics.Type_0_is_not_assignable_to_type_1.code:
                        if (!ts.isDeclarationName(node) || !ts.isVariableLike(node.parent) && !ts.isJsxAttribute(node.parent))
                            return undefined;
                        var initializer = getVariableLikeInitializer(node.parent);
                        if (!initializer || !ts.isFunctionLikeDeclaration(initializer) || !initializer.body)
                            return undefined;
                        return getFixInfo(checker, initializer, checker.getTypeAtLocation(node.parent), /* isFunctionType */ true);
                }
                return undefined;
            }
            function getVariableLikeInitializer(declaration) {
                switch (declaration.kind) {
                    case 252 /* VariableDeclaration */:
                    case 162 /* Parameter */:
                    case 201 /* BindingElement */:
                    case 165 /* PropertyDeclaration */:
                    case 291 /* PropertyAssignment */:
                        return declaration.initializer;
                    case 283 /* JsxAttribute */:
                        return declaration.initializer && (ts.isJsxExpression(declaration.initializer) ? declaration.initializer.expression : undefined);
                    case 292 /* ShorthandPropertyAssignment */:
                    case 164 /* PropertySignature */:
                    case 294 /* EnumMember */:
                    case 342 /* JSDocPropertyTag */:
                    case 335 /* JSDocParameterTag */:
                        return undefined;
                }
            }
            function addReturnStatement(changes, sourceFile, expression, statement) {
                ts.suppressLeadingAndTrailingTrivia(expression);
                var probablyNeedSemi = ts.probablyUsesSemicolons(sourceFile);
                changes.replaceNode(sourceFile, statement, ts.factory.createReturnStatement(expression), {
                    leadingTriviaOption: ts.textChanges.LeadingTriviaOption.Exclude,
                    trailingTriviaOption: ts.textChanges.TrailingTriviaOption.Exclude,
                    suffix: probablyNeedSemi ? ";" : undefined
                });
            }
            function removeBlockBodyBrace(changes, sourceFile, declaration, expression, commentSource, withParen) {
                var newBody = (withParen || ts.needsParentheses(expression)) ? ts.factory.createParenthesizedExpression(expression) : expression;
                ts.suppressLeadingAndTrailingTrivia(commentSource);
                ts.copyComments(commentSource, newBody);
                changes.replaceNode(sourceFile, declaration.body, newBody);
            }
            function wrapBlockWithParen(changes, sourceFile, declaration, expression) {
                changes.replaceNode(sourceFile, declaration.body, ts.factory.createParenthesizedExpression(expression));
            }
            function getActionForfixAddReturnStatement(context, expression, statement) {
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addReturnStatement(t, context.sourceFile, expression, statement); });
                return codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_a_return_statement, fixIdAddReturnStatement, ts.Diagnostics.Add_all_missing_return_statement);
            }
            function getActionForFixRemoveBracesFromArrowFunctionBody(context, declaration, expression, commentSource) {
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return removeBlockBodyBrace(t, context.sourceFile, declaration, expression, commentSource, /* withParen */ false); });
                return codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Remove_braces_from_arrow_function_body, fixRemoveBracesFromArrowFunctionBody, ts.Diagnostics.Remove_braces_from_all_arrow_function_bodies_with_relevant_issues);
            }
            function getActionForfixWrapTheBlockWithParen(context, declaration, expression) {
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return wrapBlockWithParen(t, context.sourceFile, declaration, expression); });
                return codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Wrap_the_following_body_with_parentheses_which_should_be_an_object_literal, fixIdWrapTheBlockWithParen, ts.Diagnostics.Wrap_all_object_literal_with_parentheses);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixMissingMember = "fixMissingMember";
            var fixMissingProperties = "fixMissingProperties";
            var fixMissingAttributes = "fixMissingAttributes";
            var fixMissingFunctionDeclaration = "fixMissingFunctionDeclaration";
            var errorCodes = [
                ts.Diagnostics.Property_0_does_not_exist_on_type_1.code,
                ts.Diagnostics.Property_0_does_not_exist_on_type_1_Did_you_mean_2.code,
                ts.Diagnostics.Property_0_is_missing_in_type_1_but_required_in_type_2.code,
                ts.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2.code,
                ts.Diagnostics.Type_0_is_missing_the_following_properties_from_type_1_Colon_2_and_3_more.code,
                ts.Diagnostics.Cannot_find_name_0.code
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var typeChecker = context.program.getTypeChecker();
                    var info = getInfo(context.sourceFile, context.span.start, typeChecker, context.program);
                    if (!info) {
                        return undefined;
                    }
                    if (info.kind === 3 /* ObjectLiteral */) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addObjectLiteralProperties(t, context, info); });
                        return [codefix.createCodeFixAction(fixMissingProperties, changes, ts.Diagnostics.Add_missing_properties, fixMissingProperties, ts.Diagnostics.Add_all_missing_properties)];
                    }
                    if (info.kind === 4 /* JsxAttributes */) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addJsxAttributes(t, context, info); });
                        return [codefix.createCodeFixAction(fixMissingAttributes, changes, ts.Diagnostics.Add_missing_attributes, fixMissingAttributes, ts.Diagnostics.Add_all_missing_attributes)];
                    }
                    if (info.kind === 2 /* Function */) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addFunctionDeclaration(t, context, info); });
                        return [codefix.createCodeFixAction(fixMissingFunctionDeclaration, changes, [ts.Diagnostics.Add_missing_function_declaration_0, info.token.text], fixMissingFunctionDeclaration, ts.Diagnostics.Add_all_missing_function_declarations)];
                    }
                    if (info.kind === 0 /* Enum */) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addEnumMemberDeclaration(t, context.program.getTypeChecker(), info); });
                        return [codefix.createCodeFixAction(fixMissingMember, changes, [ts.Diagnostics.Add_missing_enum_member_0, info.token.text], fixMissingMember, ts.Diagnostics.Add_all_missing_members)];
                    }
                    return ts.concatenate(getActionsForMissingMethodDeclaration(context, info), getActionsForMissingMemberDeclaration(context, info));
                },
                fixIds: [fixMissingMember, fixMissingFunctionDeclaration, fixMissingProperties, fixMissingAttributes],
                getAllCodeActions: function (context) {
                    var program = context.program, fixId = context.fixId;
                    var checker = program.getTypeChecker();
                    var seen = new ts.Map();
                    var typeDeclToMembers = new ts.Map();
                    return codefix.createCombinedCodeActions(ts.textChanges.ChangeTracker.with(context, function (changes) {
                        codefix.eachDiagnostic(context, errorCodes, function (diag) {
                            var info = getInfo(diag.file, diag.start, checker, context.program);
                            if (!info || !ts.addToSeen(seen, ts.getNodeId(info.parentDeclaration) + "#" + info.token.text)) {
                                return;
                            }
                            if (fixId === fixMissingFunctionDeclaration && info.kind === 2 /* Function */) {
                                addFunctionDeclaration(changes, context, info);
                            }
                            else if (fixId === fixMissingProperties && info.kind === 3 /* ObjectLiteral */) {
                                addObjectLiteralProperties(changes, context, info);
                            }
                            else if (fixId === fixMissingAttributes && info.kind === 4 /* JsxAttributes */) {
                                addJsxAttributes(changes, context, info);
                            }
                            else {
                                if (info.kind === 0 /* Enum */) {
                                    addEnumMemberDeclaration(changes, checker, info);
                                }
                                if (info.kind === 1 /* ClassOrInterface */) {
                                    var parentDeclaration = info.parentDeclaration, token_1 = info.token;
                                    var infos = ts.getOrUpdate(typeDeclToMembers, parentDeclaration, function () { return []; });
                                    if (!infos.some(function (i) { return i.token.text === token_1.text; })) {
                                        infos.push(info);
                                    }
                                }
                            }
                        });
                        typeDeclToMembers.forEach(function (infos, classDeclaration) {
                            var supers = codefix.getAllSupers(classDeclaration, checker);
                            var _loop_14 = function (info) {
                                // If some superclass added this property, don't add it again.
                                if (supers.some(function (superClassOrInterface) {
                                    var superInfos = typeDeclToMembers.get(superClassOrInterface);
                                    return !!superInfos && superInfos.some(function (_a) {
                                        var token = _a.token;
                                        return token.text === info.token.text;
                                    });
                                }))
                                    return "continue";
                                var parentDeclaration = info.parentDeclaration, declSourceFile = info.declSourceFile, modifierFlags = info.modifierFlags, token = info.token, call = info.call, isJSFile = info.isJSFile;
                                // Always prefer to add a method declaration if possible.
                                if (call && !ts.isPrivateIdentifier(token)) {
                                    addMethodDeclaration(context, changes, call, token, modifierFlags & 32 /* Static */, parentDeclaration, declSourceFile);
                                }
                                else {
                                    if (isJSFile && !ts.isInterfaceDeclaration(parentDeclaration)) {
                                        addMissingMemberInJs(changes, declSourceFile, parentDeclaration, token, !!(modifierFlags & 32 /* Static */));
                                    }
                                    else {
                                        var typeNode = getTypeNode(program.getTypeChecker(), parentDeclaration, token);
                                        addPropertyDeclaration(changes, declSourceFile, parentDeclaration, token.text, typeNode, modifierFlags & 32 /* Static */);
                                    }
                                }
                            };
                            for (var _i = 0, infos_1 = infos; _i < infos_1.length; _i++) {
                                var info = infos_1[_i];
                                _loop_14(info);
                            }
                        });
                    }));
                },
            });
            var InfoKind;
            (function (InfoKind) {
                InfoKind[InfoKind["Enum"] = 0] = "Enum";
                InfoKind[InfoKind["ClassOrInterface"] = 1] = "ClassOrInterface";
                InfoKind[InfoKind["Function"] = 2] = "Function";
                InfoKind[InfoKind["ObjectLiteral"] = 3] = "ObjectLiteral";
                InfoKind[InfoKind["JsxAttributes"] = 4] = "JsxAttributes";
            })(InfoKind || (InfoKind = {}));
            function getInfo(sourceFile, tokenPos, checker, program) {
                // The identifier of the missing property. eg:
                // this.missing = 1;
                //      ^^^^^^^
                var token = ts.getTokenAtPosition(sourceFile, tokenPos);
                if (!ts.isIdentifier(token) && !ts.isPrivateIdentifier(token)) {
                    return undefined;
                }
                var parent = token.parent;
                if (ts.isIdentifier(token) && ts.hasInitializer(parent) && parent.initializer && ts.isObjectLiteralExpression(parent.initializer)) {
                    var properties = ts.arrayFrom(checker.getUnmatchedProperties(checker.getTypeAtLocation(parent.initializer), checker.getTypeAtLocation(token), /* requireOptionalProperties */ false, /* matchDiscriminantProperties */ false));
                    if (ts.length(properties)) {
                        return { kind: 3 /* ObjectLiteral */, token: token, properties: properties, parentDeclaration: parent.initializer };
                    }
                }
                if (ts.isIdentifier(token) && ts.isJsxOpeningLikeElement(token.parent)) {
                    var attributes = getUnmatchedAttributes(checker, token.parent);
                    if (ts.length(attributes)) {
                        return { kind: 4 /* JsxAttributes */, token: token, attributes: attributes, parentDeclaration: token.parent };
                    }
                }
                if (ts.isIdentifier(token) && ts.isCallExpression(parent)) {
                    return { kind: 2 /* Function */, token: token, call: parent, sourceFile: sourceFile, modifierFlags: 0 /* None */, parentDeclaration: sourceFile };
                }
                if (!ts.isPropertyAccessExpression(parent)) {
                    return undefined;
                }
                var leftExpressionType = ts.skipConstraint(checker.getTypeAtLocation(parent.expression));
                var symbol = leftExpressionType.symbol;
                if (!symbol || !symbol.declarations) {
                    return undefined;
                }
                if (ts.isIdentifier(token) && ts.isCallExpression(parent.parent)) {
                    var moduleDeclaration = ts.find(symbol.declarations, ts.isModuleDeclaration);
                    var moduleDeclarationSourceFile = moduleDeclaration === null || moduleDeclaration === void 0 ? void 0 : moduleDeclaration.getSourceFile();
                    if (moduleDeclaration && moduleDeclarationSourceFile && !isSourceFileFromLibrary(program, moduleDeclarationSourceFile)) {
                        return { kind: 2 /* Function */, token: token, call: parent.parent, sourceFile: sourceFile, modifierFlags: 1 /* Export */, parentDeclaration: moduleDeclaration };
                    }
                    var moduleSourceFile = ts.find(symbol.declarations, ts.isSourceFile);
                    if (sourceFile.commonJsModuleIndicator) {
                        return;
                    }
                    if (moduleSourceFile && !isSourceFileFromLibrary(program, moduleSourceFile)) {
                        return { kind: 2 /* Function */, token: token, call: parent.parent, sourceFile: moduleSourceFile, modifierFlags: 1 /* Export */, parentDeclaration: moduleSourceFile };
                    }
                }
                var classDeclaration = ts.find(symbol.declarations, ts.isClassLike);
                // Don't suggest adding private identifiers to anything other than a class.
                if (!classDeclaration && ts.isPrivateIdentifier(token)) {
                    return undefined;
                }
                // Prefer to change the class instead of the interface if they are merged
                var classOrInterface = classDeclaration || ts.find(symbol.declarations, ts.isInterfaceDeclaration);
                if (classOrInterface && !isSourceFileFromLibrary(program, classOrInterface.getSourceFile())) {
                    var makeStatic = (leftExpressionType.target || leftExpressionType) !== checker.getDeclaredTypeOfSymbol(symbol);
                    if (makeStatic && (ts.isPrivateIdentifier(token) || ts.isInterfaceDeclaration(classOrInterface))) {
                        return undefined;
                    }
                    var declSourceFile = classOrInterface.getSourceFile();
                    var modifierFlags = (makeStatic ? 32 /* Static */ : 0) | (ts.startsWithUnderscore(token.text) ? 8 /* Private */ : 0);
                    var isJSFile = ts.isSourceFileJS(declSourceFile);
                    var call = ts.tryCast(parent.parent, ts.isCallExpression);
                    return { kind: 1 /* ClassOrInterface */, token: token, call: call, modifierFlags: modifierFlags, parentDeclaration: classOrInterface, declSourceFile: declSourceFile, isJSFile: isJSFile };
                }
                var enumDeclaration = ts.find(symbol.declarations, ts.isEnumDeclaration);
                if (enumDeclaration && !ts.isPrivateIdentifier(token) && !isSourceFileFromLibrary(program, enumDeclaration.getSourceFile())) {
                    return { kind: 0 /* Enum */, token: token, parentDeclaration: enumDeclaration };
                }
                return undefined;
            }
            function isSourceFileFromLibrary(program, node) {
                return program.isSourceFileFromExternalLibrary(node) || program.isSourceFileDefaultLibrary(node);
            }
            function getActionsForMissingMemberDeclaration(context, info) {
                return info.isJSFile ? ts.singleElementArray(createActionForAddMissingMemberInJavascriptFile(context, info)) :
                    createActionsForAddMissingMemberInTypeScriptFile(context, info);
            }
            function createActionForAddMissingMemberInJavascriptFile(context, _a) {
                var parentDeclaration = _a.parentDeclaration, declSourceFile = _a.declSourceFile, modifierFlags = _a.modifierFlags, token = _a.token;
                if (ts.isInterfaceDeclaration(parentDeclaration)) {
                    return undefined;
                }
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addMissingMemberInJs(t, declSourceFile, parentDeclaration, token, !!(modifierFlags & 32 /* Static */)); });
                if (changes.length === 0) {
                    return undefined;
                }
                var diagnostic = modifierFlags & 32 /* Static */ ? ts.Diagnostics.Initialize_static_property_0 :
                    ts.isPrivateIdentifier(token) ? ts.Diagnostics.Declare_a_private_field_named_0 : ts.Diagnostics.Initialize_property_0_in_the_constructor;
                return codefix.createCodeFixAction(fixMissingMember, changes, [diagnostic, token.text], fixMissingMember, ts.Diagnostics.Add_all_missing_members);
            }
            function addMissingMemberInJs(changeTracker, declSourceFile, classDeclaration, token, makeStatic) {
                var tokenName = token.text;
                if (makeStatic) {
                    if (classDeclaration.kind === 224 /* ClassExpression */) {
                        return;
                    }
                    var className = classDeclaration.name.getText();
                    var staticInitialization = initializePropertyToUndefined(ts.factory.createIdentifier(className), tokenName);
                    changeTracker.insertNodeAfter(declSourceFile, classDeclaration, staticInitialization);
                }
                else if (ts.isPrivateIdentifier(token)) {
                    var property = ts.factory.createPropertyDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, tokenName, 
                    /*questionToken*/ undefined, 
                    /*type*/ undefined, 
                    /*initializer*/ undefined);
                    var lastProp = getNodeToInsertPropertyAfter(classDeclaration);
                    if (lastProp) {
                        changeTracker.insertNodeAfter(declSourceFile, lastProp, property);
                    }
                    else {
                        changeTracker.insertNodeAtClassStart(declSourceFile, classDeclaration, property);
                    }
                }
                else {
                    var classConstructor = ts.getFirstConstructorWithBody(classDeclaration);
                    if (!classConstructor) {
                        return;
                    }
                    var propertyInitialization = initializePropertyToUndefined(ts.factory.createThis(), tokenName);
                    changeTracker.insertNodeAtConstructorEnd(declSourceFile, classConstructor, propertyInitialization);
                }
            }
            function initializePropertyToUndefined(obj, propertyName) {
                return ts.factory.createExpressionStatement(ts.factory.createAssignment(ts.factory.createPropertyAccessExpression(obj, propertyName), createUndefined()));
            }
            function createActionsForAddMissingMemberInTypeScriptFile(context, _a) {
                var parentDeclaration = _a.parentDeclaration, declSourceFile = _a.declSourceFile, modifierFlags = _a.modifierFlags, token = _a.token;
                var memberName = token.text;
                var isStatic = modifierFlags & 32 /* Static */;
                var typeNode = getTypeNode(context.program.getTypeChecker(), parentDeclaration, token);
                var addPropertyDeclarationChanges = function (modifierFlags) { return ts.textChanges.ChangeTracker.with(context, function (t) { return addPropertyDeclaration(t, declSourceFile, parentDeclaration, memberName, typeNode, modifierFlags); }); };
                var actions = [codefix.createCodeFixAction(fixMissingMember, addPropertyDeclarationChanges(modifierFlags & 32 /* Static */), [isStatic ? ts.Diagnostics.Declare_static_property_0 : ts.Diagnostics.Declare_property_0, memberName], fixMissingMember, ts.Diagnostics.Add_all_missing_members)];
                if (isStatic || ts.isPrivateIdentifier(token)) {
                    return actions;
                }
                if (modifierFlags & 8 /* Private */) {
                    actions.unshift(codefix.createCodeFixActionWithoutFixAll(fixMissingMember, addPropertyDeclarationChanges(8 /* Private */), [ts.Diagnostics.Declare_private_property_0, memberName]));
                }
                actions.push(createAddIndexSignatureAction(context, declSourceFile, parentDeclaration, token.text, typeNode));
                return actions;
            }
            function getTypeNode(checker, classDeclaration, token) {
                var typeNode;
                if (token.parent.parent.kind === 219 /* BinaryExpression */) {
                    var binaryExpression = token.parent.parent;
                    var otherExpression = token.parent === binaryExpression.left ? binaryExpression.right : binaryExpression.left;
                    var widenedType = checker.getWidenedType(checker.getBaseTypeOfLiteralType(checker.getTypeAtLocation(otherExpression)));
                    typeNode = checker.typeToTypeNode(widenedType, classDeclaration, 1 /* NoTruncation */);
                }
                else {
                    var contextualType = checker.getContextualType(token.parent);
                    typeNode = contextualType ? checker.typeToTypeNode(contextualType, /*enclosingDeclaration*/ undefined, 1 /* NoTruncation */) : undefined;
                }
                return typeNode || ts.factory.createKeywordTypeNode(129 /* AnyKeyword */);
            }
            function addPropertyDeclaration(changeTracker, declSourceFile, classDeclaration, tokenName, typeNode, modifierFlags) {
                var property = ts.factory.createPropertyDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ modifierFlags ? ts.factory.createNodeArray(ts.factory.createModifiersFromModifierFlags(modifierFlags)) : undefined, tokenName, 
                /*questionToken*/ undefined, typeNode, 
                /*initializer*/ undefined);
                var lastProp = getNodeToInsertPropertyAfter(classDeclaration);
                if (lastProp) {
                    changeTracker.insertNodeAfter(declSourceFile, lastProp, property);
                }
                else {
                    changeTracker.insertNodeAtClassStart(declSourceFile, classDeclaration, property);
                }
            }
            // Gets the last of the first run of PropertyDeclarations, or undefined if the class does not start with a PropertyDeclaration.
            function getNodeToInsertPropertyAfter(cls) {
                var res;
                for (var _i = 0, _a = cls.members; _i < _a.length; _i++) {
                    var member = _a[_i];
                    if (!ts.isPropertyDeclaration(member))
                        break;
                    res = member;
                }
                return res;
            }
            function createAddIndexSignatureAction(context, declSourceFile, classDeclaration, tokenName, typeNode) {
                // Index signatures cannot have the static modifier.
                var stringTypeNode = ts.factory.createKeywordTypeNode(148 /* StringKeyword */);
                var indexingParameter = ts.factory.createParameterDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*dotDotDotToken*/ undefined, "x", 
                /*questionToken*/ undefined, stringTypeNode, 
                /*initializer*/ undefined);
                var indexSignature = ts.factory.createIndexSignature(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, [indexingParameter], typeNode);
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return t.insertNodeAtClassStart(declSourceFile, classDeclaration, indexSignature); });
                // No fixId here because code-fix-all currently only works on adding individual named properties.
                return codefix.createCodeFixActionWithoutFixAll(fixMissingMember, changes, [ts.Diagnostics.Add_index_signature_for_property_0, tokenName]);
            }
            function getActionsForMissingMethodDeclaration(context, info) {
                var parentDeclaration = info.parentDeclaration, declSourceFile = info.declSourceFile, modifierFlags = info.modifierFlags, token = info.token, call = info.call;
                if (call === undefined) {
                    return undefined;
                }
                // Private methods are not implemented yet.
                if (ts.isPrivateIdentifier(token)) {
                    return undefined;
                }
                var methodName = token.text;
                var addMethodDeclarationChanges = function (modifierFlags) { return ts.textChanges.ChangeTracker.with(context, function (t) { return addMethodDeclaration(context, t, call, token, modifierFlags, parentDeclaration, declSourceFile); }); };
                var actions = [codefix.createCodeFixAction(fixMissingMember, addMethodDeclarationChanges(modifierFlags & 32 /* Static */), [modifierFlags & 32 /* Static */ ? ts.Diagnostics.Declare_static_method_0 : ts.Diagnostics.Declare_method_0, methodName], fixMissingMember, ts.Diagnostics.Add_all_missing_members)];
                if (modifierFlags & 8 /* Private */) {
                    actions.unshift(codefix.createCodeFixActionWithoutFixAll(fixMissingMember, addMethodDeclarationChanges(8 /* Private */), [ts.Diagnostics.Declare_private_method_0, methodName]));
                }
                return actions;
            }
            function addMethodDeclaration(context, changes, callExpression, name, modifierFlags, parentDeclaration, sourceFile) {
                var importAdder = codefix.createImportAdder(sourceFile, context.program, context.preferences, context.host);
                var methodDeclaration = codefix.createSignatureDeclarationFromCallExpression(167 /* MethodDeclaration */, context, importAdder, callExpression, name, modifierFlags, parentDeclaration);
                var containingMethodDeclaration = ts.findAncestor(callExpression, function (n) { return ts.isMethodDeclaration(n) || ts.isConstructorDeclaration(n); });
                if (containingMethodDeclaration && containingMethodDeclaration.parent === parentDeclaration) {
                    changes.insertNodeAfter(sourceFile, containingMethodDeclaration, methodDeclaration);
                }
                else {
                    changes.insertNodeAtClassStart(sourceFile, parentDeclaration, methodDeclaration);
                }
                importAdder.writeFixes(changes);
            }
            function addEnumMemberDeclaration(changes, checker, _a) {
                var token = _a.token, parentDeclaration = _a.parentDeclaration;
                /**
                 * create initializer only literal enum that has string initializer.
                 * value of initializer is a string literal that equal to name of enum member.
                 * numeric enum or empty enum will not create initializer.
                 */
                var hasStringInitializer = ts.some(parentDeclaration.members, function (member) {
                    var type = checker.getTypeAtLocation(member);
                    return !!(type && type.flags & 402653316 /* StringLike */);
                });
                var enumMember = ts.factory.createEnumMember(token, hasStringInitializer ? ts.factory.createStringLiteral(token.text) : undefined);
                changes.replaceNode(parentDeclaration.getSourceFile(), parentDeclaration, ts.factory.updateEnumDeclaration(parentDeclaration, parentDeclaration.decorators, parentDeclaration.modifiers, parentDeclaration.name, ts.concatenate(parentDeclaration.members, ts.singleElementArray(enumMember))), {
                    leadingTriviaOption: ts.textChanges.LeadingTriviaOption.IncludeAll,
                    trailingTriviaOption: ts.textChanges.TrailingTriviaOption.Exclude
                });
            }
            function addFunctionDeclaration(changes, context, info) {
                var importAdder = codefix.createImportAdder(context.sourceFile, context.program, context.preferences, context.host);
                var functionDeclaration = codefix.createSignatureDeclarationFromCallExpression(254 /* FunctionDeclaration */, context, importAdder, info.call, ts.idText(info.token), info.modifierFlags, info.parentDeclaration);
                changes.insertNodeAtEndOfScope(info.sourceFile, info.parentDeclaration, functionDeclaration);
            }
            function addJsxAttributes(changes, context, info) {
                var importAdder = codefix.createImportAdder(context.sourceFile, context.program, context.preferences, context.host);
                var quotePreference = ts.getQuotePreference(context.sourceFile, context.preferences);
                var checker = context.program.getTypeChecker();
                var jsxAttributesNode = info.parentDeclaration.attributes;
                var hasSpreadAttribute = ts.some(jsxAttributesNode.properties, ts.isJsxSpreadAttribute);
                var attrs = ts.map(info.attributes, function (attr) {
                    var value = attr.valueDeclaration ? tryGetValueFromType(context, checker, importAdder, quotePreference, checker.getTypeAtLocation(attr.valueDeclaration)) : createUndefined();
                    return ts.factory.createJsxAttribute(ts.factory.createIdentifier(attr.name), ts.factory.createJsxExpression(/*dotDotDotToken*/ undefined, value));
                });
                var jsxAttributes = ts.factory.createJsxAttributes(hasSpreadAttribute ? __spreadArray(__spreadArray([], attrs, true), jsxAttributesNode.properties, true) : __spreadArray(__spreadArray([], jsxAttributesNode.properties, true), attrs, true));
                var options = { prefix: jsxAttributesNode.pos === jsxAttributesNode.end ? " " : undefined };
                changes.replaceNode(context.sourceFile, jsxAttributesNode, jsxAttributes, options);
            }
            function addObjectLiteralProperties(changes, context, info) {
                var importAdder = codefix.createImportAdder(context.sourceFile, context.program, context.preferences, context.host);
                var quotePreference = ts.getQuotePreference(context.sourceFile, context.preferences);
                var checker = context.program.getTypeChecker();
                var props = ts.map(info.properties, function (prop) {
                    var initializer = prop.valueDeclaration ? tryGetValueFromType(context, checker, importAdder, quotePreference, checker.getTypeAtLocation(prop.valueDeclaration)) : createUndefined();
                    return ts.factory.createPropertyAssignment(prop.name, initializer);
                });
                changes.replaceNode(context.sourceFile, info.parentDeclaration, ts.factory.createObjectLiteralExpression(__spreadArray(__spreadArray([], info.parentDeclaration.properties, true), props, true), /*multiLine*/ true));
            }
            function tryGetValueFromType(context, checker, importAdder, quotePreference, type) {
                if (type.flags & 3 /* AnyOrUnknown */) {
                    return createUndefined();
                }
                if (type.flags & (4 /* String */ | 134217728 /* TemplateLiteral */)) {
                    return ts.factory.createStringLiteral("", /* isSingleQuote */ quotePreference === 0 /* Single */);
                }
                if (type.flags & 8 /* Number */) {
                    return ts.factory.createNumericLiteral(0);
                }
                if (type.flags & 64 /* BigInt */) {
                    return ts.factory.createBigIntLiteral("0n");
                }
                if (type.flags & 16 /* Boolean */) {
                    return ts.factory.createFalse();
                }
                if (type.flags & 1056 /* EnumLike */) {
                    var enumMember = type.symbol.exports ? ts.firstOrUndefined(ts.arrayFrom(type.symbol.exports.values())) : type.symbol;
                    var name = checker.symbolToExpression(type.symbol.parent ? type.symbol.parent : type.symbol, 111551 /* Value */, /*enclosingDeclaration*/ undefined, /*flags*/ undefined);
                    return enumMember === undefined || name === undefined ? ts.factory.createNumericLiteral(0) : ts.factory.createPropertyAccessExpression(name, checker.symbolToString(enumMember));
                }
                if (type.flags & 256 /* NumberLiteral */) {
                    return ts.factory.createNumericLiteral(type.value);
                }
                if (type.flags & 2048 /* BigIntLiteral */) {
                    return ts.factory.createBigIntLiteral(type.value);
                }
                if (type.flags & 128 /* StringLiteral */) {
                    return ts.factory.createStringLiteral(type.value, /* isSingleQuote */ quotePreference === 0 /* Single */);
                }
                if (type.flags & 512 /* BooleanLiteral */) {
                    return (type === checker.getFalseType() || type === checker.getFalseType(/*fresh*/ true)) ? ts.factory.createFalse() : ts.factory.createTrue();
                }
                if (type.flags & 65536 /* Null */) {
                    return ts.factory.createNull();
                }
                if (type.flags & 1048576 /* Union */) {
                    var expression = ts.firstDefined(type.types, function (t) { return tryGetValueFromType(context, checker, importAdder, quotePreference, t); });
                    return expression !== null && expression !== void 0 ? expression : createUndefined();
                }
                if (checker.isArrayLikeType(type)) {
                    return ts.factory.createArrayLiteralExpression();
                }
                if (isObjectLiteralType(type)) {
                    var props = ts.map(checker.getPropertiesOfType(type), function (prop) {
                        var initializer = prop.valueDeclaration ? tryGetValueFromType(context, checker, importAdder, quotePreference, checker.getTypeAtLocation(prop.valueDeclaration)) : createUndefined();
                        return ts.factory.createPropertyAssignment(prop.name, initializer);
                    });
                    return ts.factory.createObjectLiteralExpression(props, /*multiLine*/ true);
                }
                if (ts.getObjectFlags(type) & 16 /* Anonymous */) {
                    var decl = ts.find(type.symbol.declarations || ts.emptyArray, ts.or(ts.isFunctionTypeNode, ts.isMethodSignature, ts.isMethodDeclaration));
                    if (decl === undefined)
                        return createUndefined();
                    var signature = checker.getSignaturesOfType(type, 0 /* Call */);
                    if (signature === undefined)
                        return createUndefined();
                    var func = codefix.createSignatureDeclarationFromSignature(211 /* FunctionExpression */, context, quotePreference, signature[0], codefix.createStubbedBody(ts.Diagnostics.Function_not_implemented.message, quotePreference), /*name*/ undefined, /*modifiers*/ undefined, /*optional*/ undefined, /*enclosingDeclaration*/ undefined, importAdder);
                    return func !== null && func !== void 0 ? func : createUndefined();
                }
                if (ts.getObjectFlags(type) & 1 /* Class */) {
                    var classDeclaration = ts.getClassLikeDeclarationOfSymbol(type.symbol);
                    if (classDeclaration === undefined || ts.hasAbstractModifier(classDeclaration))
                        return createUndefined();
                    var constructorDeclaration = ts.getFirstConstructorWithBody(classDeclaration);
                    if (constructorDeclaration && ts.length(constructorDeclaration.parameters))
                        return createUndefined();
                    return ts.factory.createNewExpression(ts.factory.createIdentifier(type.symbol.name), /*typeArguments*/ undefined, /*argumentsArray*/ undefined);
                }
                return createUndefined();
            }
            function createUndefined() {
                return ts.factory.createIdentifier("undefined");
            }
            function isObjectLiteralType(type) {
                return (type.flags & 524288 /* Object */) &&
                    ((ts.getObjectFlags(type) & 128 /* ObjectLiteral */) || (type.symbol && ts.tryCast(ts.singleOrUndefined(type.symbol.declarations), ts.isTypeLiteralNode)));
            }
            function getUnmatchedAttributes(checker, source) {
                var attrsType = checker.getContextualType(source.attributes);
                if (attrsType === undefined)
                    return ts.emptyArray;
                var targetProps = attrsType.getProperties();
                if (!ts.length(targetProps))
                    return ts.emptyArray;
                var seenNames = new ts.Set();
                for (var _i = 0, _a = source.attributes.properties; _i < _a.length; _i++) {
                    var sourceProp = _a[_i];
                    if (ts.isJsxAttribute(sourceProp)) {
                        seenNames.add(sourceProp.name.escapedText);
                    }
                    if (ts.isJsxSpreadAttribute(sourceProp)) {
                        var type = checker.getTypeAtLocation(sourceProp.expression);
                        for (var _b = 0, _c = type.getProperties(); _b < _c.length; _b++) {
                            var prop = _c[_b];
                            seenNames.add(prop.escapedName);
                        }
                    }
                }
                return ts.filter(targetProps, function (targetProp) {
                    return !((targetProp.flags & 16777216 /* Optional */ || ts.getCheckFlags(targetProp) & 48 /* Partial */) || seenNames.has(targetProp.escapedName));
                });
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "addMissingNewOperator";
            var errorCodes = [ts.Diagnostics.Value_of_type_0_is_not_callable_Did_you_mean_to_include_new.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addMissingNewOperator(t, sourceFile, span); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_missing_new_operator_to_call, fixId, ts.Diagnostics.Add_missing_new_operator_to_all_calls)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        return addMissingNewOperator(changes, context.sourceFile, diag);
                    });
                },
            });
            function addMissingNewOperator(changes, sourceFile, span) {
                var call = ts.cast(findAncestorMatchingSpan(sourceFile, span), ts.isCallExpression);
                var newExpression = ts.factory.createNewExpression(call.expression, call.typeArguments, call.arguments);
                changes.replaceNode(sourceFile, call, newExpression);
            }
            function findAncestorMatchingSpan(sourceFile, span) {
                var token = ts.getTokenAtPosition(sourceFile, span.start);
                var end = ts.textSpanEnd(span);
                while (token.end < end) {
                    token = token.parent;
                }
                return token;
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixName = "fixCannotFindModule";
            var fixIdInstallTypesPackage = "installTypesPackage";
            var errorCodeCannotFindModule = ts.Diagnostics.Cannot_find_module_0_or_its_corresponding_type_declarations.code;
            var errorCodes = [
                errorCodeCannotFindModule,
                ts.Diagnostics.Could_not_find_a_declaration_file_for_module_0_1_implicitly_has_an_any_type.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var host = context.host, sourceFile = context.sourceFile, start = context.span.start;
                    var packageName = tryGetImportedPackageName(sourceFile, start);
                    if (packageName === undefined)
                        return undefined;
                    var typesPackageName = getTypesPackageNameToInstall(packageName, host, context.errorCode);
                    return typesPackageName === undefined
                        ? []
                        : [codefix.createCodeFixAction(fixName, /*changes*/ [], [ts.Diagnostics.Install_0, typesPackageName], fixIdInstallTypesPackage, ts.Diagnostics.Install_all_missing_types_packages, getInstallCommand(sourceFile.fileName, typesPackageName))];
                },
                fixIds: [fixIdInstallTypesPackage],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (_changes, diag, commands) {
                        var packageName = tryGetImportedPackageName(diag.file, diag.start);
                        if (packageName === undefined)
                            return undefined;
                        switch (context.fixId) {
                            case fixIdInstallTypesPackage: {
                                var pkg = getTypesPackageNameToInstall(packageName, context.host, diag.code);
                                if (pkg) {
                                    commands.push(getInstallCommand(diag.file.fileName, pkg));
                                }
                                break;
                            }
                            default:
                                ts.Debug.fail("Bad fixId: " + context.fixId);
                        }
                    });
                },
            });
            function getInstallCommand(fileName, packageName) {
                return { type: "install package", file: fileName, packageName: packageName };
            }
            function tryGetImportedPackageName(sourceFile, pos) {
                var moduleSpecifierText = ts.tryCast(ts.getTokenAtPosition(sourceFile, pos), ts.isStringLiteral);
                if (!moduleSpecifierText)
                    return undefined;
                var moduleName = moduleSpecifierText.text;
                var packageName = ts.parsePackageName(moduleName).packageName;
                return ts.isExternalModuleNameRelative(packageName) ? undefined : packageName;
            }
            function getTypesPackageNameToInstall(packageName, host, diagCode) {
                var _a;
                return diagCode === errorCodeCannotFindModule
                    ? (ts.JsTyping.nodeCoreModules.has(packageName) ? "@types/node" : undefined)
                    : (((_a = host.isKnownTypesPackageName) === null || _a === void 0 ? void 0 : _a.call(host, packageName)) ? ts.getTypesPackageName(packageName) : undefined);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var errorCodes = [
                ts.Diagnostics.Non_abstract_class_0_does_not_implement_inherited_abstract_member_1_from_class_2.code,
                ts.Diagnostics.Non_abstract_class_expression_does_not_implement_inherited_abstract_member_0_from_class_1.code,
            ];
            var fixId = "fixClassDoesntImplementInheritedAbstractMember";
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) {
                        return addMissingMembers(getClass(sourceFile, span.start), sourceFile, context, t, context.preferences);
                    });
                    return changes.length === 0 ? undefined : [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Implement_inherited_abstract_class, fixId, ts.Diagnostics.Implement_all_inherited_abstract_classes)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    var seenClassDeclarations = new ts.Map();
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var classDeclaration = getClass(diag.file, diag.start);
                        if (ts.addToSeen(seenClassDeclarations, ts.getNodeId(classDeclaration))) {
                            addMissingMembers(classDeclaration, context.sourceFile, context, changes, context.preferences);
                        }
                    });
                },
            });
            function getClass(sourceFile, pos) {
                // Token is the identifier in the case of a class declaration
                // or the class keyword token in the case of a class expression.
                var token = ts.getTokenAtPosition(sourceFile, pos);
                return ts.cast(token.parent, ts.isClassLike);
            }
            function addMissingMembers(classDeclaration, sourceFile, context, changeTracker, preferences) {
                var extendsNode = ts.getEffectiveBaseTypeNode(classDeclaration);
                var checker = context.program.getTypeChecker();
                var instantiatedExtendsType = checker.getTypeAtLocation(extendsNode);
                // Note that this is ultimately derived from a map indexed by symbol names,
                // so duplicates cannot occur.
                var abstractAndNonPrivateExtendsSymbols = checker.getPropertiesOfType(instantiatedExtendsType).filter(symbolPointsToNonPrivateAndAbstractMember);
                var importAdder = codefix.createImportAdder(sourceFile, context.program, preferences, context.host);
                codefix.createMissingMemberNodes(classDeclaration, abstractAndNonPrivateExtendsSymbols, sourceFile, context, preferences, importAdder, function (member) { return changeTracker.insertNodeAtClassStart(sourceFile, classDeclaration, member); });
                importAdder.writeFixes(changeTracker);
            }
            function symbolPointsToNonPrivateAndAbstractMember(symbol) {
                // See `codeFixClassExtendAbstractProtectedProperty.ts` in https://github.com/Microsoft/TypeScript/pull/11547/files
                // (now named `codeFixClassExtendAbstractPrivateProperty.ts`)
                var flags = ts.getSyntacticModifierFlags(ts.first(symbol.getDeclarations()));
                return !(flags & 8 /* Private */) && !!(flags & 128 /* Abstract */);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "classSuperMustPrecedeThisAccess";
            var errorCodes = [ts.Diagnostics.super_must_be_called_before_accessing_this_in_the_constructor_of_a_derived_class.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var nodes = getNodes(sourceFile, span.start);
                    if (!nodes)
                        return undefined;
                    var constructor = nodes.constructor, superCall = nodes.superCall;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, constructor, superCall); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Make_super_call_the_first_statement_in_the_constructor, fixId, ts.Diagnostics.Make_all_super_calls_the_first_statement_in_their_constructor)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    var sourceFile = context.sourceFile;
                    var seenClasses = new ts.Map(); // Ensure we only do this once per class.
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var nodes = getNodes(diag.file, diag.start);
                        if (!nodes)
                            return;
                        var constructor = nodes.constructor, superCall = nodes.superCall;
                        if (ts.addToSeen(seenClasses, ts.getNodeId(constructor.parent))) {
                            doChange(changes, sourceFile, constructor, superCall);
                        }
                    });
                },
            });
            function doChange(changes, sourceFile, constructor, superCall) {
                changes.insertNodeAtConstructorStart(sourceFile, constructor, superCall);
                changes.delete(sourceFile, superCall);
            }
            function getNodes(sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                if (token.kind !== 108 /* ThisKeyword */)
                    return undefined;
                var constructor = ts.getContainingFunction(token);
                var superCall = findSuperCall(constructor.body);
                // figure out if the `this` access is actually inside the supercall
                // i.e. super(this.a), since in that case we won't suggest a fix
                return superCall && !superCall.expression.arguments.some(function (arg) { return ts.isPropertyAccessExpression(arg) && arg.expression === token; }) ? { constructor: constructor, superCall: superCall } : undefined;
            }
            function findSuperCall(n) {
                return ts.isExpressionStatement(n) && ts.isSuperCall(n.expression)
                    ? n
                    : ts.isFunctionLike(n)
                        ? undefined
                        : ts.forEachChild(n, findSuperCall);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "constructorForDerivedNeedSuperCall";
            var errorCodes = [ts.Diagnostics.Constructors_for_derived_classes_must_contain_a_super_call.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var ctr = getNode(sourceFile, span.start);
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, ctr); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_missing_super_call, fixId, ts.Diagnostics.Add_all_missing_super_calls)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        return doChange(changes, context.sourceFile, getNode(diag.file, diag.start));
                    });
                },
            });
            function getNode(sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                ts.Debug.assert(ts.isConstructorDeclaration(token.parent), "token should be at the constructor declaration");
                return token.parent;
            }
            function doChange(changes, sourceFile, ctr) {
                var superCall = ts.factory.createExpressionStatement(ts.factory.createCallExpression(ts.factory.createSuper(), /*typeArguments*/ undefined, /*argumentsArray*/ ts.emptyArray));
                changes.insertNodeAtConstructorStart(sourceFile, ctr, superCall);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "enableExperimentalDecorators";
            var errorCodes = [
                ts.Diagnostics.Experimental_support_for_decorators_is_a_feature_that_is_subject_to_change_in_a_future_release_Set_the_experimentalDecorators_option_in_your_tsconfig_or_jsconfig_to_remove_this_warning.code
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var configFile = context.program.getCompilerOptions().configFile;
                    if (configFile === undefined) {
                        return undefined;
                    }
                    var changes = ts.textChanges.ChangeTracker.with(context, function (changeTracker) { return doChange(changeTracker, configFile); });
                    return [codefix.createCodeFixActionWithoutFixAll(fixId, changes, ts.Diagnostics.Enable_the_experimentalDecorators_option_in_your_configuration_file)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes) {
                        var configFile = context.program.getCompilerOptions().configFile;
                        if (configFile === undefined) {
                            return undefined;
                        }
                        doChange(changes, configFile);
                    });
                },
            });
            function doChange(changeTracker, configFile) {
                codefix.setJsonCompilerOptionValue(changeTracker, configFile, "experimentalDecorators", ts.factory.createTrue());
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixID = "fixEnableJsxFlag";
            var errorCodes = [ts.Diagnostics.Cannot_use_JSX_unless_the_jsx_flag_is_provided.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var configFile = context.program.getCompilerOptions().configFile;
                    if (configFile === undefined) {
                        return undefined;
                    }
                    var changes = ts.textChanges.ChangeTracker.with(context, function (changeTracker) {
                        return doChange(changeTracker, configFile);
                    });
                    return [
                        codefix.createCodeFixActionWithoutFixAll(fixID, changes, ts.Diagnostics.Enable_the_jsx_flag_in_your_configuration_file)
                    ];
                },
                fixIds: [fixID],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes) {
                        var configFile = context.program.getCompilerOptions().configFile;
                        if (configFile === undefined) {
                            return undefined;
                        }
                        doChange(changes, configFile);
                    });
                }
            });
            function doChange(changeTracker, configFile) {
                codefix.setJsonCompilerOptionValue(changeTracker, configFile, "jsx", ts.factory.createStringLiteral("react"));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            codefix.registerCodeFix({
                errorCodes: [
                    ts.Diagnostics.Top_level_await_expressions_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_target_option_is_set_to_es2017_or_higher.code,
                    ts.Diagnostics.Top_level_for_await_loops_are_only_allowed_when_the_module_option_is_set_to_esnext_or_system_and_the_target_option_is_set_to_es2017_or_higher.code,
                ],
                getCodeActions: function (context) {
                    var compilerOptions = context.program.getCompilerOptions();
                    var configFile = compilerOptions.configFile;
                    if (configFile === undefined) {
                        return undefined;
                    }
                    var codeFixes = [];
                    var moduleKind = ts.getEmitModuleKind(compilerOptions);
                    var moduleOutOfRange = moduleKind >= ts.ModuleKind.ES2015 && moduleKind < ts.ModuleKind.ESNext;
                    if (moduleOutOfRange) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (changes) {
                            codefix.setJsonCompilerOptionValue(changes, configFile, "module", ts.factory.createStringLiteral("esnext"));
                        });
                        codeFixes.push(codefix.createCodeFixActionWithoutFixAll("fixModuleOption", changes, [ts.Diagnostics.Set_the_module_option_in_your_configuration_file_to_0, "esnext"]));
                    }
                    var target = ts.getEmitScriptTarget(compilerOptions);
                    var targetOutOfRange = target < 4 /* ES2017 */ || target > 99 /* ESNext */;
                    if (targetOutOfRange) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (tracker) {
                            var configObject = ts.getTsConfigObjectLiteralExpression(configFile);
                            if (!configObject)
                                return;
                            var options = [["target", ts.factory.createStringLiteral("es2017")]];
                            if (moduleKind === ts.ModuleKind.CommonJS) {
                                // Ensure we preserve the default module kind (commonjs), as targets >= ES2015 have a default module kind of es2015.
                                options.push(["module", ts.factory.createStringLiteral("commonjs")]);
                            }
                            codefix.setJsonCompilerOptionValues(tracker, configFile, options);
                        });
                        codeFixes.push(codefix.createCodeFixActionWithoutFixAll("fixTargetOption", changes, [ts.Diagnostics.Set_the_target_option_in_your_configuration_file_to_0, "es2017"]));
                    }
                    return codeFixes.length ? codeFixes : undefined;
                }
            });
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixPropertyAssignment";
            var errorCodes = [
                ts.Diagnostics.Did_you_mean_to_use_a_Colon_An_can_only_follow_a_property_name_when_the_containing_object_literal_is_part_of_a_destructuring_pattern.code
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                fixIds: [fixId],
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var property = getProperty(sourceFile, span.start);
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, context.sourceFile, property); });
                    return [codefix.createCodeFixAction(fixId, changes, [ts.Diagnostics.Change_0_to_1, "=", ":"], fixId, [ts.Diagnostics.Switch_each_misused_0_to_1, "=", ":"])];
                },
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return doChange(changes, diag.file, getProperty(diag.file, diag.start)); });
                }
            });
            function doChange(changes, sourceFile, node) {
                changes.replaceNode(sourceFile, node, ts.factory.createPropertyAssignment(node.name, node.objectAssignmentInitializer));
            }
            function getProperty(sourceFile, pos) {
                return ts.cast(ts.getTokenAtPosition(sourceFile, pos).parent, ts.isShorthandPropertyAssignment);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "extendsInterfaceBecomesImplements";
            var errorCodes = [ts.Diagnostics.Cannot_extend_an_interface_0_Did_you_mean_implements.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile;
                    var nodes = getNodes(sourceFile, context.span.start);
                    if (!nodes)
                        return undefined;
                    var extendsToken = nodes.extendsToken, heritageClauses = nodes.heritageClauses;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChanges(t, sourceFile, extendsToken, heritageClauses); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Change_extends_to_implements, fixId, ts.Diagnostics.Change_all_extended_interfaces_to_implements)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var nodes = getNodes(diag.file, diag.start);
                        if (nodes)
                            doChanges(changes, diag.file, nodes.extendsToken, nodes.heritageClauses);
                    });
                },
            });
            function getNodes(sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                var heritageClauses = ts.getContainingClass(token).heritageClauses;
                var extendsToken = heritageClauses[0].getFirstToken();
                return extendsToken.kind === 94 /* ExtendsKeyword */ ? { extendsToken: extendsToken, heritageClauses: heritageClauses } : undefined;
            }
            function doChanges(changes, sourceFile, extendsToken, heritageClauses) {
                changes.replaceNode(sourceFile, extendsToken, ts.factory.createToken(117 /* ImplementsKeyword */));
                // If there is already an implements clause, replace the implements keyword with a comma.
                if (heritageClauses.length === 2 &&
                    heritageClauses[0].token === 94 /* ExtendsKeyword */ &&
                    heritageClauses[1].token === 117 /* ImplementsKeyword */) {
                    var implementsToken = heritageClauses[1].getFirstToken();
                    var implementsFullStart = implementsToken.getFullStart();
                    changes.replaceRange(sourceFile, { pos: implementsFullStart, end: implementsFullStart }, ts.factory.createToken(27 /* CommaToken */));
                    // Rough heuristic: delete trailing whitespace after keyword so that it's not excessive.
                    // (Trailing because leading might be indentation, which is more sensitive.)
                    var text = sourceFile.text;
                    var end = implementsToken.end;
                    while (end < text.length && ts.isWhiteSpaceSingleLine(text.charCodeAt(end))) {
                        end++;
                    }
                    changes.deleteRange(sourceFile, { pos: implementsToken.getStart(), end: end });
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "forgottenThisPropertyAccess";
            var didYouMeanStaticMemberCode = ts.Diagnostics.Cannot_find_name_0_Did_you_mean_the_static_member_1_0.code;
            var errorCodes = [
                ts.Diagnostics.Cannot_find_name_0_Did_you_mean_the_instance_member_this_0.code,
                ts.Diagnostics.Private_identifiers_are_not_allowed_outside_class_bodies.code,
                didYouMeanStaticMemberCode,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile;
                    var info = getInfo(sourceFile, context.span.start, context.errorCode);
                    if (!info) {
                        return undefined;
                    }
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, info); });
                    return [codefix.createCodeFixAction(fixId, changes, [ts.Diagnostics.Add_0_to_unresolved_variable, info.className || "this"], fixId, ts.Diagnostics.Add_qualifier_to_all_unresolved_variables_matching_a_member_name)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var info = getInfo(diag.file, diag.start, diag.code);
                        if (info)
                            doChange(changes, context.sourceFile, info);
                    });
                },
            });
            function getInfo(sourceFile, pos, diagCode) {
                var node = ts.getTokenAtPosition(sourceFile, pos);
                if (ts.isIdentifier(node)) {
                    return { node: node, className: diagCode === didYouMeanStaticMemberCode ? ts.getContainingClass(node).name.text : undefined };
                }
            }
            function doChange(changes, sourceFile, _a) {
                var node = _a.node, className = _a.className;
                // TODO (https://github.com/Microsoft/TypeScript/issues/21246): use shared helper
                ts.suppressLeadingAndTrailingTrivia(node);
                changes.replaceNode(sourceFile, node, ts.factory.createPropertyAccessExpression(className ? ts.factory.createIdentifier(className) : ts.factory.createThis(), node));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixIdExpression = "fixInvalidJsxCharacters_expression";
            var fixIdHtmlEntity = "fixInvalidJsxCharacters_htmlEntity";
            var errorCodes = [
                ts.Diagnostics.Unexpected_token_Did_you_mean_or_gt.code,
                ts.Diagnostics.Unexpected_token_Did_you_mean_or_rbrace.code
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                fixIds: [fixIdExpression, fixIdHtmlEntity],
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, preferences = context.preferences, span = context.span;
                    var changeToExpression = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, preferences, sourceFile, span.start, /* useHtmlEntity */ false); });
                    var changeToHtmlEntity = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, preferences, sourceFile, span.start, /* useHtmlEntity */ true); });
                    return [
                        codefix.createCodeFixAction(fixIdExpression, changeToExpression, ts.Diagnostics.Wrap_invalid_character_in_an_expression_container, fixIdExpression, ts.Diagnostics.Wrap_all_invalid_characters_in_an_expression_container),
                        codefix.createCodeFixAction(fixIdHtmlEntity, changeToHtmlEntity, ts.Diagnostics.Convert_invalid_character_to_its_html_entity_code, fixIdHtmlEntity, ts.Diagnostics.Convert_all_invalid_characters_to_HTML_entity_code)
                    ];
                },
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diagnostic) { return doChange(changes, context.preferences, diagnostic.file, diagnostic.start, context.fixId === fixIdHtmlEntity); });
                }
            });
            var htmlEntity = {
                ">": "&gt;",
                "}": "&rbrace;",
            };
            function isValidCharacter(character) {
                return ts.hasProperty(htmlEntity, character);
            }
            function doChange(changes, preferences, sourceFile, start, useHtmlEntity) {
                var character = sourceFile.getText()[start];
                // sanity check
                if (!isValidCharacter(character)) {
                    return;
                }
                var replacement = useHtmlEntity ? htmlEntity[character] : "{" + ts.quote(sourceFile, preferences, character) + "}";
                changes.replaceRangeWithText(sourceFile, { pos: start, end: start + 1 }, replacement);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixName = "unusedIdentifier";
            var fixIdPrefix = "unusedIdentifier_prefix";
            var fixIdDelete = "unusedIdentifier_delete";
            var fixIdDeleteImports = "unusedIdentifier_deleteImports";
            var fixIdInfer = "unusedIdentifier_infer";
            var errorCodes = [
                ts.Diagnostics._0_is_declared_but_its_value_is_never_read.code,
                ts.Diagnostics._0_is_declared_but_never_used.code,
                ts.Diagnostics.Property_0_is_declared_but_its_value_is_never_read.code,
                ts.Diagnostics.All_imports_in_import_declaration_are_unused.code,
                ts.Diagnostics.All_destructured_elements_are_unused.code,
                ts.Diagnostics.All_variables_are_unused.code,
                ts.Diagnostics.All_type_parameters_are_unused.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var errorCode = context.errorCode, sourceFile = context.sourceFile, program = context.program, cancellationToken = context.cancellationToken;
                    var checker = program.getTypeChecker();
                    var sourceFiles = program.getSourceFiles();
                    var token = ts.getTokenAtPosition(sourceFile, context.span.start);
                    if (ts.isJSDocTemplateTag(token)) {
                        return [createDeleteFix(ts.textChanges.ChangeTracker.with(context, function (t) { return t.delete(sourceFile, token); }), ts.Diagnostics.Remove_template_tag)];
                    }
                    if (token.kind === 29 /* LessThanToken */) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return deleteTypeParameters(t, sourceFile, token); });
                        return [createDeleteFix(changes, ts.Diagnostics.Remove_type_parameters)];
                    }
                    var importDecl = tryGetFullImport(token);
                    if (importDecl) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return t.delete(sourceFile, importDecl); });
                        return [codefix.createCodeFixAction(fixName, changes, [ts.Diagnostics.Remove_import_from_0, ts.showModuleSpecifier(importDecl)], fixIdDeleteImports, ts.Diagnostics.Delete_all_unused_imports)];
                    }
                    else if (isImport(token)) {
                        var deletion = ts.textChanges.ChangeTracker.with(context, function (t) { return tryDeleteDeclaration(sourceFile, token, t, checker, sourceFiles, program, cancellationToken, /*isFixAll*/ false); });
                        if (deletion.length) {
                            return [codefix.createCodeFixAction(fixName, deletion, [ts.Diagnostics.Remove_unused_declaration_for_Colon_0, token.getText(sourceFile)], fixIdDeleteImports, ts.Diagnostics.Delete_all_unused_imports)];
                        }
                    }
                    if (ts.isObjectBindingPattern(token.parent) || ts.isArrayBindingPattern(token.parent)) {
                        if (ts.isParameter(token.parent.parent)) {
                            var elements = token.parent.elements;
                            var diagnostic = [
                                elements.length > 1 ? ts.Diagnostics.Remove_unused_declarations_for_Colon_0 : ts.Diagnostics.Remove_unused_declaration_for_Colon_0,
                                ts.map(elements, function (e) { return e.getText(sourceFile); }).join(", ")
                            ];
                            return [
                                createDeleteFix(ts.textChanges.ChangeTracker.with(context, function (t) {
                                    return deleteDestructuringElements(t, sourceFile, token.parent);
                                }), diagnostic)
                            ];
                        }
                        return [
                            createDeleteFix(ts.textChanges.ChangeTracker.with(context, function (t) {
                                return t.delete(sourceFile, token.parent.parent);
                            }), ts.Diagnostics.Remove_unused_destructuring_declaration)
                        ];
                    }
                    if (canDeleteEntireVariableStatement(sourceFile, token)) {
                        return [
                            createDeleteFix(ts.textChanges.ChangeTracker.with(context, function (t) {
                                return deleteEntireVariableStatement(t, sourceFile, token.parent);
                            }), ts.Diagnostics.Remove_variable_statement)
                        ];
                    }
                    var result = [];
                    if (token.kind === 136 /* InferKeyword */) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return changeInferToUnknown(t, sourceFile, token); });
                        var name = ts.cast(token.parent, ts.isInferTypeNode).typeParameter.name.text;
                        result.push(codefix.createCodeFixAction(fixName, changes, [ts.Diagnostics.Replace_infer_0_with_unknown, name], fixIdInfer, ts.Diagnostics.Replace_all_unused_infer_with_unknown));
                    }
                    else {
                        var deletion = ts.textChanges.ChangeTracker.with(context, function (t) {
                            return tryDeleteDeclaration(sourceFile, token, t, checker, sourceFiles, program, cancellationToken, /*isFixAll*/ false);
                        });
                        if (deletion.length) {
                            var name = ts.isComputedPropertyName(token.parent) ? token.parent : token;
                            result.push(createDeleteFix(deletion, [ts.Diagnostics.Remove_unused_declaration_for_Colon_0, name.getText(sourceFile)]));
                        }
                    }
                    var prefix = ts.textChanges.ChangeTracker.with(context, function (t) { return tryPrefixDeclaration(t, errorCode, sourceFile, token); });
                    if (prefix.length) {
                        result.push(codefix.createCodeFixAction(fixName, prefix, [ts.Diagnostics.Prefix_0_with_an_underscore, token.getText(sourceFile)], fixIdPrefix, ts.Diagnostics.Prefix_all_unused_declarations_with_where_possible));
                    }
                    return result;
                },
                fixIds: [fixIdPrefix, fixIdDelete, fixIdDeleteImports, fixIdInfer],
                getAllCodeActions: function (context) {
                    var sourceFile = context.sourceFile, program = context.program, cancellationToken = context.cancellationToken;
                    var checker = program.getTypeChecker();
                    var sourceFiles = program.getSourceFiles();
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var token = ts.getTokenAtPosition(sourceFile, diag.start);
                        switch (context.fixId) {
                            case fixIdPrefix:
                                tryPrefixDeclaration(changes, diag.code, sourceFile, token);
                                break;
                            case fixIdDeleteImports: {
                                var importDecl = tryGetFullImport(token);
                                if (importDecl) {
                                    changes.delete(sourceFile, importDecl);
                                }
                                else if (isImport(token)) {
                                    tryDeleteDeclaration(sourceFile, token, changes, checker, sourceFiles, program, cancellationToken, /*isFixAll*/ true);
                                }
                                break;
                            }
                            case fixIdDelete: {
                                if (token.kind === 136 /* InferKeyword */ || isImport(token)) {
                                    break; // Can't delete
                                }
                                else if (ts.isJSDocTemplateTag(token)) {
                                    changes.delete(sourceFile, token);
                                }
                                else if (token.kind === 29 /* LessThanToken */) {
                                    deleteTypeParameters(changes, sourceFile, token);
                                }
                                else if (ts.isObjectBindingPattern(token.parent)) {
                                    if (token.parent.parent.initializer) {
                                        break;
                                    }
                                    else if (!ts.isParameter(token.parent.parent) || isNotProvidedArguments(token.parent.parent, checker, sourceFiles)) {
                                        changes.delete(sourceFile, token.parent.parent);
                                    }
                                }
                                else if (ts.isArrayBindingPattern(token.parent.parent) && token.parent.parent.parent.initializer) {
                                    break;
                                }
                                else if (canDeleteEntireVariableStatement(sourceFile, token)) {
                                    deleteEntireVariableStatement(changes, sourceFile, token.parent);
                                }
                                else {
                                    tryDeleteDeclaration(sourceFile, token, changes, checker, sourceFiles, program, cancellationToken, /*isFixAll*/ true);
                                }
                                break;
                            }
                            case fixIdInfer:
                                if (token.kind === 136 /* InferKeyword */) {
                                    changeInferToUnknown(changes, sourceFile, token);
                                }
                                break;
                            default:
                                ts.Debug.fail(JSON.stringify(context.fixId));
                        }
                    });
                },
            });
            function changeInferToUnknown(changes, sourceFile, token) {
                changes.replaceNode(sourceFile, token.parent, ts.factory.createKeywordTypeNode(153 /* UnknownKeyword */));
            }
            function createDeleteFix(changes, diag) {
                return codefix.createCodeFixAction(fixName, changes, diag, fixIdDelete, ts.Diagnostics.Delete_all_unused_declarations);
            }
            function deleteTypeParameters(changes, sourceFile, token) {
                changes.delete(sourceFile, ts.Debug.checkDefined(ts.cast(token.parent, ts.isDeclarationWithTypeParameterChildren).typeParameters, "The type parameter to delete should exist"));
            }
            function isImport(token) {
                return token.kind === 100 /* ImportKeyword */
                    || token.kind === 79 /* Identifier */ && (token.parent.kind === 268 /* ImportSpecifier */ || token.parent.kind === 265 /* ImportClause */);
            }
            /** Sometimes the diagnostic span is an entire ImportDeclaration, so we should remove the whole thing. */
            function tryGetFullImport(token) {
                return token.kind === 100 /* ImportKeyword */ ? ts.tryCast(token.parent, ts.isImportDeclaration) : undefined;
            }
            function canDeleteEntireVariableStatement(sourceFile, token) {
                return ts.isVariableDeclarationList(token.parent) && ts.first(token.parent.getChildren(sourceFile)) === token;
            }
            function deleteEntireVariableStatement(changes, sourceFile, node) {
                changes.delete(sourceFile, node.parent.kind === 235 /* VariableStatement */ ? node.parent : node);
            }
            function deleteDestructuringElements(changes, sourceFile, node) {
                ts.forEach(node.elements, function (n) { return changes.delete(sourceFile, n); });
            }
            function tryPrefixDeclaration(changes, errorCode, sourceFile, token) {
                // Don't offer to prefix a property.
                if (errorCode === ts.Diagnostics.Property_0_is_declared_but_its_value_is_never_read.code)
                    return;
                if (token.kind === 136 /* InferKeyword */) {
                    token = ts.cast(token.parent, ts.isInferTypeNode).typeParameter.name;
                }
                if (ts.isIdentifier(token) && canPrefix(token)) {
                    changes.replaceNode(sourceFile, token, ts.factory.createIdentifier("_" + token.text));
                    if (ts.isParameter(token.parent)) {
                        ts.getJSDocParameterTags(token.parent).forEach(function (tag) {
                            if (ts.isIdentifier(tag.name)) {
                                changes.replaceNode(sourceFile, tag.name, ts.factory.createIdentifier("_" + tag.name.text));
                            }
                        });
                    }
                }
            }
            function canPrefix(token) {
                switch (token.parent.kind) {
                    case 162 /* Parameter */:
                    case 161 /* TypeParameter */:
                        return true;
                    case 252 /* VariableDeclaration */: {
                        var varDecl = token.parent;
                        switch (varDecl.parent.parent.kind) {
                            case 242 /* ForOfStatement */:
                            case 241 /* ForInStatement */:
                                return true;
                        }
                    }
                }
                return false;
            }
            function tryDeleteDeclaration(sourceFile, token, changes, checker, sourceFiles, program, cancellationToken, isFixAll) {
                tryDeleteDeclarationWorker(token, changes, sourceFile, checker, sourceFiles, program, cancellationToken, isFixAll);
                if (ts.isIdentifier(token)) {
                    ts.FindAllReferences.Core.eachSymbolReferenceInFile(token, checker, sourceFile, function (ref) {
                        if (ts.isPropertyAccessExpression(ref.parent) && ref.parent.name === ref)
                            ref = ref.parent;
                        if (!isFixAll && mayDeleteExpression(ref)) {
                            changes.delete(sourceFile, ref.parent.parent);
                        }
                    });
                }
            }
            function tryDeleteDeclarationWorker(token, changes, sourceFile, checker, sourceFiles, program, cancellationToken, isFixAll) {
                var parent = token.parent;
                if (ts.isParameter(parent)) {
                    tryDeleteParameter(changes, sourceFile, parent, checker, sourceFiles, program, cancellationToken, isFixAll);
                }
                else if (!(isFixAll && ts.isIdentifier(token) && ts.FindAllReferences.Core.isSymbolReferencedInFile(token, checker, sourceFile))) {
                    var node = ts.isImportClause(parent) ? token : ts.isComputedPropertyName(parent) ? parent.parent : parent;
                    ts.Debug.assert(node !== sourceFile, "should not delete whole source file");
                    changes.delete(sourceFile, node);
                }
            }
            function tryDeleteParameter(changes, sourceFile, parameter, checker, sourceFiles, program, cancellationToken, isFixAll) {
                if (isFixAll === void 0) {
                    isFixAll = false;
                }
                if (mayDeleteParameter(checker, sourceFile, parameter, sourceFiles, program, cancellationToken, isFixAll)) {
                    if (parameter.modifiers && parameter.modifiers.length > 0 &&
                        (!ts.isIdentifier(parameter.name) || ts.FindAllReferences.Core.isSymbolReferencedInFile(parameter.name, checker, sourceFile))) {
                        parameter.modifiers.forEach(function (modifier) { return changes.deleteModifier(sourceFile, modifier); });
                    }
                    else if (!parameter.initializer && isNotProvidedArguments(parameter, checker, sourceFiles)) {
                        changes.delete(sourceFile, parameter);
                    }
                }
            }
            function isNotProvidedArguments(parameter, checker, sourceFiles) {
                var index = parameter.parent.parameters.indexOf(parameter);
                // Just in case the call didn't provide enough arguments.
                return !ts.FindAllReferences.Core.someSignatureUsage(parameter.parent, sourceFiles, checker, function (_, call) { return !call || call.arguments.length > index; });
            }
            function mayDeleteParameter(checker, sourceFile, parameter, sourceFiles, program, cancellationToken, isFixAll) {
                var parent = parameter.parent;
                switch (parent.kind) {
                    case 167 /* MethodDeclaration */:
                    case 169 /* Constructor */:
                        var index = parent.parameters.indexOf(parameter);
                        var referent = ts.isMethodDeclaration(parent) ? parent.name : parent;
                        var entries = ts.FindAllReferences.Core.getReferencedSymbolsForNode(parent.pos, referent, program, sourceFiles, cancellationToken);
                        if (entries) {
                            for (var _i = 0, entries_2 = entries; _i < entries_2.length; _i++) {
                                var entry = entries_2[_i];
                                for (var _a = 0, _b = entry.references; _a < _b.length; _a++) {
                                    var reference = _b[_a];
                                    if (reference.kind === 1 /* Node */) {
                                        // argument in super(...)
                                        var isSuperCall_1 = ts.isSuperKeyword(reference.node)
                                            && ts.isCallExpression(reference.node.parent)
                                            && reference.node.parent.arguments.length > index;
                                        // argument in super.m(...)
                                        var isSuperMethodCall = ts.isPropertyAccessExpression(reference.node.parent)
                                            && ts.isSuperKeyword(reference.node.parent.expression)
                                            && ts.isCallExpression(reference.node.parent.parent)
                                            && reference.node.parent.parent.arguments.length > index;
                                        // parameter in overridden or overriding method
                                        var isOverriddenMethod = (ts.isMethodDeclaration(reference.node.parent) || ts.isMethodSignature(reference.node.parent))
                                            && reference.node.parent !== parameter.parent
                                            && reference.node.parent.parameters.length > index;
                                        if (isSuperCall_1 || isSuperMethodCall || isOverriddenMethod)
                                            return false;
                                    }
                                }
                            }
                        }
                        return true;
                    case 254 /* FunctionDeclaration */: {
                        if (parent.name && isCallbackLike(checker, sourceFile, parent.name)) {
                            return isLastParameter(parent, parameter, isFixAll);
                        }
                        return true;
                    }
                    case 211 /* FunctionExpression */:
                    case 212 /* ArrowFunction */:
                        // Can't remove a non-last parameter in a callback. Can remove a parameter in code-fix-all if future parameters are also unused.
                        return isLastParameter(parent, parameter, isFixAll);
                    case 171 /* SetAccessor */:
                        // Setter must have a parameter
                        return false;
                    default:
                        return ts.Debug.failBadSyntaxKind(parent);
                }
            }
            function isCallbackLike(checker, sourceFile, name) {
                return !!ts.FindAllReferences.Core.eachSymbolReferenceInFile(name, checker, sourceFile, function (reference) {
                    return ts.isIdentifier(reference) && ts.isCallExpression(reference.parent) && reference.parent.arguments.indexOf(reference) >= 0;
                });
            }
            function isLastParameter(func, parameter, isFixAll) {
                var parameters = func.parameters;
                var index = parameters.indexOf(parameter);
                ts.Debug.assert(index !== -1, "The parameter should already be in the list");
                return isFixAll ?
                    parameters.slice(index + 1).every(function (p) { return ts.isIdentifier(p.name) && !p.symbol.isReferenced; }) :
                    index === parameters.length - 1;
            }
            function mayDeleteExpression(node) {
                return ((ts.isBinaryExpression(node.parent) && node.parent.left === node) ||
                    ((ts.isPostfixUnaryExpression(node.parent) || ts.isPrefixUnaryExpression(node.parent)) && node.parent.operand === node)) && ts.isExpressionStatement(node.parent.parent);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixUnreachableCode";
            var errorCodes = [ts.Diagnostics.Unreachable_code_detected.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, context.sourceFile, context.span.start, context.span.length, context.errorCode); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Remove_unreachable_code, fixId, ts.Diagnostics.Remove_all_unreachable_code)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) { return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return doChange(changes, diag.file, diag.start, diag.length, diag.code); }); },
            });
            function doChange(changes, sourceFile, start, length, errorCode) {
                var token = ts.getTokenAtPosition(sourceFile, start);
                var statement = ts.findAncestor(token, ts.isStatement);
                if (statement.getStart(sourceFile) !== token.getStart(sourceFile)) {
                    var logData = JSON.stringify({
                        statementKind: ts.Debug.formatSyntaxKind(statement.kind),
                        tokenKind: ts.Debug.formatSyntaxKind(token.kind),
                        errorCode: errorCode,
                        start: start,
                        length: length
                    });
                    ts.Debug.fail("Token and statement should start at the same point. " + logData);
                }
                var container = (ts.isBlock(statement.parent) ? statement.parent : statement).parent;
                if (!ts.isBlock(statement.parent) || statement === ts.first(statement.parent.statements)) {
                    switch (container.kind) {
                        case 237 /* IfStatement */:
                            if (container.elseStatement) {
                                if (ts.isBlock(statement.parent)) {
                                    break;
                                }
                                else {
                                    changes.replaceNode(sourceFile, statement, ts.factory.createBlock(ts.emptyArray));
                                }
                                return;
                            }
                        // falls through
                        case 239 /* WhileStatement */:
                        case 240 /* ForStatement */:
                            changes.delete(sourceFile, container);
                            return;
                    }
                }
                if (ts.isBlock(statement.parent)) {
                    var end_3 = start + length;
                    var lastStatement = ts.Debug.checkDefined(lastWhere(ts.sliceAfter(statement.parent.statements, statement), function (s) { return s.pos < end_3; }), "Some statement should be last");
                    changes.deleteNodeRange(sourceFile, statement, lastStatement);
                }
                else {
                    changes.delete(sourceFile, statement);
                }
            }
            function lastWhere(a, pred) {
                var last;
                for (var _i = 0, a_1 = a; _i < a_1.length; _i++) {
                    var value = a_1[_i];
                    if (!pred(value))
                        break;
                    last = value;
                }
                return last;
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixUnusedLabel";
            var errorCodes = [ts.Diagnostics.Unused_label.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, context.sourceFile, context.span.start); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Remove_unused_label, fixId, ts.Diagnostics.Remove_all_unused_labels)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) { return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return doChange(changes, diag.file, diag.start); }); },
            });
            function doChange(changes, sourceFile, start) {
                var token = ts.getTokenAtPosition(sourceFile, start);
                var labeledStatement = ts.cast(token.parent, ts.isLabeledStatement);
                var pos = token.getStart(sourceFile);
                var statementPos = labeledStatement.statement.getStart(sourceFile);
                // If label is on a separate line, just delete the rest of that line, but not the indentation of the labeled statement.
                var end = ts.positionsAreOnSameLine(pos, statementPos, sourceFile) ? statementPos
                    : ts.skipTrivia(sourceFile.text, ts.findChildOfKind(labeledStatement, 58 /* ColonToken */, sourceFile).end, /*stopAfterLineBreak*/ true);
                changes.deleteRange(sourceFile, { pos: pos, end: end });
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixIdPlain = "fixJSDocTypes_plain";
            var fixIdNullable = "fixJSDocTypes_nullable";
            var errorCodes = [ts.Diagnostics.JSDoc_types_can_only_be_used_inside_documentation_comments.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile;
                    var checker = context.program.getTypeChecker();
                    var info = getInfo(sourceFile, context.span.start, checker);
                    if (!info)
                        return undefined;
                    var typeNode = info.typeNode, type = info.type;
                    var original = typeNode.getText(sourceFile);
                    var actions = [fix(type, fixIdPlain, ts.Diagnostics.Change_all_jsdoc_style_types_to_TypeScript)];
                    if (typeNode.kind === 309 /* JSDocNullableType */) {
                        // for nullable types, suggest the flow-compatible `T | null | undefined`
                        // in addition to the jsdoc/closure-compatible `T | null`
                        actions.push(fix(checker.getNullableType(type, 32768 /* Undefined */), fixIdNullable, ts.Diagnostics.Change_all_jsdoc_style_types_to_TypeScript_and_add_undefined_to_nullable_types));
                    }
                    return actions;
                    function fix(type, fixId, fixAllDescription) {
                        var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, typeNode, type, checker); });
                        return codefix.createCodeFixAction("jdocTypes", changes, [ts.Diagnostics.Change_0_to_1, original, checker.typeToString(type)], fixId, fixAllDescription);
                    }
                },
                fixIds: [fixIdPlain, fixIdNullable],
                getAllCodeActions: function (context) {
                    var fixId = context.fixId, program = context.program, sourceFile = context.sourceFile;
                    var checker = program.getTypeChecker();
                    return codefix.codeFixAll(context, errorCodes, function (changes, err) {
                        var info = getInfo(err.file, err.start, checker);
                        if (!info)
                            return;
                        var typeNode = info.typeNode, type = info.type;
                        var fixedType = typeNode.kind === 309 /* JSDocNullableType */ && fixId === fixIdNullable ? checker.getNullableType(type, 32768 /* Undefined */) : type;
                        doChange(changes, sourceFile, typeNode, fixedType, checker);
                    });
                }
            });
            function doChange(changes, sourceFile, oldTypeNode, newType, checker) {
                changes.replaceNode(sourceFile, oldTypeNode, checker.typeToTypeNode(newType, /*enclosingDeclaration*/ oldTypeNode, /*flags*/ undefined)); // TODO: GH#18217
            }
            function getInfo(sourceFile, pos, checker) {
                var decl = ts.findAncestor(ts.getTokenAtPosition(sourceFile, pos), isTypeContainer);
                var typeNode = decl && decl.type;
                return typeNode && { typeNode: typeNode, type: checker.getTypeFromTypeNode(typeNode) };
            }
            function isTypeContainer(node) {
                // NOTE: Some locations are not handled yet:
                // MappedTypeNode.typeParameters and SignatureDeclaration.typeParameters, as well as CallExpression.typeArguments
                switch (node.kind) {
                    case 227 /* AsExpression */:
                    case 172 /* CallSignature */:
                    case 173 /* ConstructSignature */:
                    case 254 /* FunctionDeclaration */:
                    case 170 /* GetAccessor */:
                    case 174 /* IndexSignature */:
                    case 193 /* MappedType */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                    case 162 /* Parameter */:
                    case 165 /* PropertyDeclaration */:
                    case 164 /* PropertySignature */:
                    case 171 /* SetAccessor */:
                    case 257 /* TypeAliasDeclaration */:
                    case 209 /* TypeAssertionExpression */:
                    case 252 /* VariableDeclaration */:
                        return true;
                    default:
                        return false;
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixMissingCallParentheses";
            var errorCodes = [
                ts.Diagnostics.This_condition_will_always_return_true_since_this_function_is_always_defined_Did_you_mean_to_call_it_instead.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                fixIds: [fixId],
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var callName = getCallName(sourceFile, span.start);
                    if (!callName)
                        return;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, context.sourceFile, callName); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_missing_call_parentheses, fixId, ts.Diagnostics.Add_all_missing_call_parentheses)];
                },
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var callName = getCallName(diag.file, diag.start);
                        if (callName)
                            doChange(changes, diag.file, callName);
                    });
                }
            });
            function doChange(changes, sourceFile, name) {
                changes.replaceNodeWithText(sourceFile, name, name.text + "()");
            }
            function getCallName(sourceFile, start) {
                var token = ts.getTokenAtPosition(sourceFile, start);
                if (ts.isPropertyAccessExpression(token.parent)) {
                    var current = token.parent;
                    while (ts.isPropertyAccessExpression(current.parent)) {
                        current = current.parent;
                    }
                    return current.name;
                }
                if (ts.isIdentifier(token)) {
                    return token;
                }
                return undefined;
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixAwaitInSyncFunction";
            var errorCodes = [
                ts.Diagnostics.await_expressions_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules.code,
                ts.Diagnostics.for_await_loops_are_only_allowed_within_async_functions_and_at_the_top_levels_of_modules.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var nodes = getNodes(sourceFile, span.start);
                    if (!nodes)
                        return undefined;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, nodes); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_async_modifier_to_containing_function, fixId, ts.Diagnostics.Add_all_missing_async_modifiers)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    var seen = new ts.Map();
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var nodes = getNodes(diag.file, diag.start);
                        if (!nodes || !ts.addToSeen(seen, ts.getNodeId(nodes.insertBefore)))
                            return;
                        doChange(changes, context.sourceFile, nodes);
                    });
                },
            });
            function getReturnType(expr) {
                if (expr.type) {
                    return expr.type;
                }
                if (ts.isVariableDeclaration(expr.parent) &&
                    expr.parent.type &&
                    ts.isFunctionTypeNode(expr.parent.type)) {
                    return expr.parent.type.type;
                }
            }
            function getNodes(sourceFile, start) {
                var token = ts.getTokenAtPosition(sourceFile, start);
                var containingFunction = ts.getContainingFunction(token);
                if (!containingFunction) {
                    return;
                }
                var insertBefore;
                switch (containingFunction.kind) {
                    case 167 /* MethodDeclaration */:
                        insertBefore = containingFunction.name;
                        break;
                    case 254 /* FunctionDeclaration */:
                    case 211 /* FunctionExpression */:
                        insertBefore = ts.findChildOfKind(containingFunction, 98 /* FunctionKeyword */, sourceFile);
                        break;
                    case 212 /* ArrowFunction */:
                        insertBefore = ts.findChildOfKind(containingFunction, 20 /* OpenParenToken */, sourceFile) || ts.first(containingFunction.parameters);
                        break;
                    default:
                        return;
                }
                return insertBefore && {
                    insertBefore: insertBefore,
                    returnType: getReturnType(containingFunction)
                };
            }
            function doChange(changes, sourceFile, _a) {
                var insertBefore = _a.insertBefore, returnType = _a.returnType;
                if (returnType) {
                    var entityName = ts.getEntityNameFromTypeNode(returnType);
                    if (!entityName || entityName.kind !== 79 /* Identifier */ || entityName.text !== "Promise") {
                        changes.replaceNode(sourceFile, returnType, ts.factory.createTypeReferenceNode("Promise", ts.factory.createNodeArray([returnType])));
                    }
                }
                changes.insertModifierBefore(sourceFile, 130 /* AsyncKeyword */, insertBefore);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var errorCodes = [
                ts.Diagnostics._0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property.code,
                ts.Diagnostics._0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor.code,
            ];
            var fixId = "fixPropertyOverrideAccessor";
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var edits = doChange(context.sourceFile, context.span.start, context.span.length, context.errorCode, context);
                    if (edits) {
                        return [codefix.createCodeFixAction(fixId, edits, ts.Diagnostics.Generate_get_and_set_accessors, fixId, ts.Diagnostics.Generate_get_and_set_accessors_for_all_overriding_properties)];
                    }
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var edits = doChange(diag.file, diag.start, diag.length, diag.code, context);
                        if (edits) {
                            for (var _i = 0, edits_2 = edits; _i < edits_2.length; _i++) {
                                var edit = edits_2[_i];
                                changes.pushRaw(context.sourceFile, edit);
                            }
                        }
                    });
                },
            });
            function doChange(file, start, length, code, context) {
                var startPosition;
                var endPosition;
                if (code === ts.Diagnostics._0_is_defined_as_an_accessor_in_class_1_but_is_overridden_here_in_2_as_an_instance_property.code) {
                    startPosition = start;
                    endPosition = start + length;
                }
                else if (code === ts.Diagnostics._0_is_defined_as_a_property_in_class_1_but_is_overridden_here_in_2_as_an_accessor.code) {
                    var checker = context.program.getTypeChecker();
                    var node = ts.getTokenAtPosition(file, start).parent;
                    ts.Debug.assert(ts.isAccessor(node), "error span of fixPropertyOverrideAccessor should only be on an accessor");
                    var containingClass = node.parent;
                    ts.Debug.assert(ts.isClassLike(containingClass), "erroneous accessors should only be inside classes");
                    var base = ts.singleOrUndefined(codefix.getAllSupers(containingClass, checker));
                    if (!base)
                        return [];
                    var name = ts.unescapeLeadingUnderscores(ts.getTextOfPropertyName(node.name));
                    var baseProp = checker.getPropertyOfType(checker.getTypeAtLocation(base), name);
                    if (!baseProp || !baseProp.valueDeclaration)
                        return [];
                    startPosition = baseProp.valueDeclaration.pos;
                    endPosition = baseProp.valueDeclaration.end;
                    file = ts.getSourceFileOfNode(baseProp.valueDeclaration);
                }
                else {
                    ts.Debug.fail("fixPropertyOverrideAccessor codefix got unexpected error code " + code);
                }
                return codefix.generateAccessorFromProperty(file, context.program, startPosition, endPosition, context, ts.Diagnostics.Generate_get_and_set_accessors.message);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "inferFromUsage";
            var errorCodes = [
                // Variable declarations
                ts.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined.code,
                // Variable uses
                ts.Diagnostics.Variable_0_implicitly_has_an_1_type.code,
                // Parameter declarations
                ts.Diagnostics.Parameter_0_implicitly_has_an_1_type.code,
                ts.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code,
                // Get Accessor declarations
                ts.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation.code,
                ts.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type.code,
                // Set Accessor declarations
                ts.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation.code,
                // Property declarations
                ts.Diagnostics.Member_0_implicitly_has_an_1_type.code,
                //// Suggestions
                // Variable declarations
                ts.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage.code,
                // Variable uses
                ts.Diagnostics.Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code,
                // Parameter declarations
                ts.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code,
                ts.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage.code,
                // Get Accessor declarations
                ts.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage.code,
                ts.Diagnostics._0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage.code,
                // Set Accessor declarations
                ts.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage.code,
                // Property declarations
                ts.Diagnostics.Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code,
                // Function expressions and declarations
                ts.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, program = context.program, start = context.span.start, errorCode = context.errorCode, cancellationToken = context.cancellationToken, host = context.host, preferences = context.preferences;
                    var token = ts.getTokenAtPosition(sourceFile, start);
                    var declaration;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (changes) { declaration = doChange(changes, sourceFile, token, errorCode, program, cancellationToken, /*markSeen*/ ts.returnTrue, host, preferences); });
                    var name = declaration && ts.getNameOfDeclaration(declaration);
                    return !name || changes.length === 0 ? undefined
                        : [codefix.createCodeFixAction(fixId, changes, [getDiagnostic(errorCode, token), name.getText(sourceFile)], fixId, ts.Diagnostics.Infer_all_types_from_usage)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    var sourceFile = context.sourceFile, program = context.program, cancellationToken = context.cancellationToken, host = context.host, preferences = context.preferences;
                    var markSeen = ts.nodeSeenTracker();
                    return codefix.codeFixAll(context, errorCodes, function (changes, err) {
                        doChange(changes, sourceFile, ts.getTokenAtPosition(err.file, err.start), err.code, program, cancellationToken, markSeen, host, preferences);
                    });
                },
            });
            function getDiagnostic(errorCode, token) {
                switch (errorCode) {
                    case ts.Diagnostics.Parameter_0_implicitly_has_an_1_type.code:
                    case ts.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:
                        return ts.isSetAccessorDeclaration(ts.getContainingFunction(token)) ? ts.Diagnostics.Infer_type_of_0_from_usage : ts.Diagnostics.Infer_parameter_types_from_usage; // TODO: GH#18217
                    case ts.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code:
                    case ts.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage.code:
                        return ts.Diagnostics.Infer_parameter_types_from_usage;
                    case ts.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation.code:
                        return ts.Diagnostics.Infer_this_type_of_0_from_usage;
                    default:
                        return ts.Diagnostics.Infer_type_of_0_from_usage;
                }
            }
            /** Map suggestion code to error code */
            function mapSuggestionDiagnostic(errorCode) {
                switch (errorCode) {
                    case ts.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_but_a_better_type_may_be_inferred_from_usage.code:
                        return ts.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined.code;
                    case ts.Diagnostics.Variable_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:
                        return ts.Diagnostics.Variable_0_implicitly_has_an_1_type.code;
                    case ts.Diagnostics.Parameter_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:
                        return ts.Diagnostics.Parameter_0_implicitly_has_an_1_type.code;
                    case ts.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type_but_a_better_type_may_be_inferred_from_usage.code:
                        return ts.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code;
                    case ts.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_get_accessor_may_be_inferred_from_usage.code:
                        return ts.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation.code;
                    case ts.Diagnostics._0_implicitly_has_an_1_return_type_but_a_better_type_may_be_inferred_from_usage.code:
                        return ts.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type.code;
                    case ts.Diagnostics.Property_0_implicitly_has_type_any_but_a_better_type_for_its_set_accessor_may_be_inferred_from_usage.code:
                        return ts.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation.code;
                    case ts.Diagnostics.Member_0_implicitly_has_an_1_type_but_a_better_type_may_be_inferred_from_usage.code:
                        return ts.Diagnostics.Member_0_implicitly_has_an_1_type.code;
                }
                return errorCode;
            }
            function doChange(changes, sourceFile, token, errorCode, program, cancellationToken, markSeen, host, preferences) {
                if (!ts.isParameterPropertyModifier(token.kind) && token.kind !== 79 /* Identifier */ && token.kind !== 25 /* DotDotDotToken */ && token.kind !== 108 /* ThisKeyword */) {
                    return undefined;
                }
                var parent = token.parent;
                var importAdder = codefix.createImportAdder(sourceFile, program, preferences, host);
                errorCode = mapSuggestionDiagnostic(errorCode);
                switch (errorCode) {
                    // Variable and Property declarations
                    case ts.Diagnostics.Member_0_implicitly_has_an_1_type.code:
                    case ts.Diagnostics.Variable_0_implicitly_has_type_1_in_some_locations_where_its_type_cannot_be_determined.code:
                        if ((ts.isVariableDeclaration(parent) && markSeen(parent)) || ts.isPropertyDeclaration(parent) || ts.isPropertySignature(parent)) { // handle bad location
                            annotateVariableDeclaration(changes, importAdder, sourceFile, parent, program, host, cancellationToken);
                            importAdder.writeFixes(changes);
                            return parent;
                        }
                        if (ts.isPropertyAccessExpression(parent)) {
                            var type = inferTypeForVariableFromUsage(parent.name, program, cancellationToken);
                            var typeNode = ts.getTypeNodeIfAccessible(type, parent, program, host);
                            if (typeNode) {
                                // Note that the codefix will never fire with an existing `@type` tag, so there is no need to merge tags
                                var typeTag = ts.factory.createJSDocTypeTag(/*tagName*/ undefined, ts.factory.createJSDocTypeExpression(typeNode), /*comment*/ undefined);
                                addJSDocTags(changes, sourceFile, ts.cast(parent.parent.parent, ts.isExpressionStatement), [typeTag]);
                            }
                            importAdder.writeFixes(changes);
                            return parent;
                        }
                        return undefined;
                    case ts.Diagnostics.Variable_0_implicitly_has_an_1_type.code: {
                        var symbol = program.getTypeChecker().getSymbolAtLocation(token);
                        if (symbol && symbol.valueDeclaration && ts.isVariableDeclaration(symbol.valueDeclaration) && markSeen(symbol.valueDeclaration)) {
                            annotateVariableDeclaration(changes, importAdder, sourceFile, symbol.valueDeclaration, program, host, cancellationToken);
                            importAdder.writeFixes(changes);
                            return symbol.valueDeclaration;
                        }
                        return undefined;
                    }
                }
                var containingFunction = ts.getContainingFunction(token);
                if (containingFunction === undefined) {
                    return undefined;
                }
                var declaration;
                switch (errorCode) {
                    // Parameter declarations
                    case ts.Diagnostics.Parameter_0_implicitly_has_an_1_type.code:
                        if (ts.isSetAccessorDeclaration(containingFunction)) {
                            annotateSetAccessor(changes, importAdder, sourceFile, containingFunction, program, host, cancellationToken);
                            declaration = containingFunction;
                            break;
                        }
                    // falls through
                    case ts.Diagnostics.Rest_parameter_0_implicitly_has_an_any_type.code:
                        if (markSeen(containingFunction)) {
                            var param = ts.cast(parent, ts.isParameter);
                            annotateParameters(changes, importAdder, sourceFile, param, containingFunction, program, host, cancellationToken);
                            declaration = param;
                        }
                        break;
                    // Get Accessor declarations
                    case ts.Diagnostics.Property_0_implicitly_has_type_any_because_its_get_accessor_lacks_a_return_type_annotation.code:
                    case ts.Diagnostics._0_which_lacks_return_type_annotation_implicitly_has_an_1_return_type.code:
                        if (ts.isGetAccessorDeclaration(containingFunction) && ts.isIdentifier(containingFunction.name)) {
                            annotate(changes, importAdder, sourceFile, containingFunction, inferTypeForVariableFromUsage(containingFunction.name, program, cancellationToken), program, host);
                            declaration = containingFunction;
                        }
                        break;
                    // Set Accessor declarations
                    case ts.Diagnostics.Property_0_implicitly_has_type_any_because_its_set_accessor_lacks_a_parameter_type_annotation.code:
                        if (ts.isSetAccessorDeclaration(containingFunction)) {
                            annotateSetAccessor(changes, importAdder, sourceFile, containingFunction, program, host, cancellationToken);
                            declaration = containingFunction;
                        }
                        break;
                    // Function 'this'
                    case ts.Diagnostics.this_implicitly_has_type_any_because_it_does_not_have_a_type_annotation.code:
                        if (ts.textChanges.isThisTypeAnnotatable(containingFunction) && markSeen(containingFunction)) {
                            annotateThis(changes, sourceFile, containingFunction, program, host, cancellationToken);
                            declaration = containingFunction;
                        }
                        break;
                    default:
                        return ts.Debug.fail(String(errorCode));
                }
                importAdder.writeFixes(changes);
                return declaration;
            }
            function annotateVariableDeclaration(changes, importAdder, sourceFile, declaration, program, host, cancellationToken) {
                if (ts.isIdentifier(declaration.name)) {
                    annotate(changes, importAdder, sourceFile, declaration, inferTypeForVariableFromUsage(declaration.name, program, cancellationToken), program, host);
                }
            }
            function annotateParameters(changes, importAdder, sourceFile, parameterDeclaration, containingFunction, program, host, cancellationToken) {
                if (!ts.isIdentifier(parameterDeclaration.name)) {
                    return;
                }
                var parameterInferences = inferTypeForParametersFromUsage(containingFunction, sourceFile, program, cancellationToken);
                ts.Debug.assert(containingFunction.parameters.length === parameterInferences.length, "Parameter count and inference count should match");
                if (ts.isInJSFile(containingFunction)) {
                    annotateJSDocParameters(changes, sourceFile, parameterInferences, program, host);
                }
                else {
                    var needParens = ts.isArrowFunction(containingFunction) && !ts.findChildOfKind(containingFunction, 20 /* OpenParenToken */, sourceFile);
                    if (needParens)
                        changes.insertNodeBefore(sourceFile, ts.first(containingFunction.parameters), ts.factory.createToken(20 /* OpenParenToken */));
                    for (var _i = 0, parameterInferences_1 = parameterInferences; _i < parameterInferences_1.length; _i++) {
                        var _a = parameterInferences_1[_i], declaration = _a.declaration, type = _a.type;
                        if (declaration && !declaration.type && !declaration.initializer) {
                            annotate(changes, importAdder, sourceFile, declaration, type, program, host);
                        }
                    }
                    if (needParens)
                        changes.insertNodeAfter(sourceFile, ts.last(containingFunction.parameters), ts.factory.createToken(21 /* CloseParenToken */));
                }
            }
            function annotateThis(changes, sourceFile, containingFunction, program, host, cancellationToken) {
                var references = getFunctionReferences(containingFunction, sourceFile, program, cancellationToken);
                if (!references || !references.length) {
                    return;
                }
                var thisInference = inferTypeFromReferences(program, references, cancellationToken).thisParameter();
                var typeNode = ts.getTypeNodeIfAccessible(thisInference, containingFunction, program, host);
                if (!typeNode) {
                    return;
                }
                if (ts.isInJSFile(containingFunction)) {
                    annotateJSDocThis(changes, sourceFile, containingFunction, typeNode);
                }
                else {
                    changes.tryInsertThisTypeAnnotation(sourceFile, containingFunction, typeNode);
                }
            }
            function annotateJSDocThis(changes, sourceFile, containingFunction, typeNode) {
                addJSDocTags(changes, sourceFile, containingFunction, [
                    ts.factory.createJSDocThisTag(/*tagName*/ undefined, ts.factory.createJSDocTypeExpression(typeNode)),
                ]);
            }
            function annotateSetAccessor(changes, importAdder, sourceFile, setAccessorDeclaration, program, host, cancellationToken) {
                var param = ts.firstOrUndefined(setAccessorDeclaration.parameters);
                if (param && ts.isIdentifier(setAccessorDeclaration.name) && ts.isIdentifier(param.name)) {
                    var type = inferTypeForVariableFromUsage(setAccessorDeclaration.name, program, cancellationToken);
                    if (type === program.getTypeChecker().getAnyType()) {
                        type = inferTypeForVariableFromUsage(param.name, program, cancellationToken);
                    }
                    if (ts.isInJSFile(setAccessorDeclaration)) {
                        annotateJSDocParameters(changes, sourceFile, [{ declaration: param, type: type }], program, host);
                    }
                    else {
                        annotate(changes, importAdder, sourceFile, param, type, program, host);
                    }
                }
            }
            function annotate(changes, importAdder, sourceFile, declaration, type, program, host) {
                var typeNode = ts.getTypeNodeIfAccessible(type, declaration, program, host);
                if (typeNode) {
                    if (ts.isInJSFile(sourceFile) && declaration.kind !== 164 /* PropertySignature */) {
                        var parent = ts.isVariableDeclaration(declaration) ? ts.tryCast(declaration.parent.parent, ts.isVariableStatement) : declaration;
                        if (!parent) {
                            return;
                        }
                        var typeExpression = ts.factory.createJSDocTypeExpression(typeNode);
                        var typeTag = ts.isGetAccessorDeclaration(declaration) ? ts.factory.createJSDocReturnTag(/*tagName*/ undefined, typeExpression, /*comment*/ undefined) : ts.factory.createJSDocTypeTag(/*tagName*/ undefined, typeExpression, /*comment*/ undefined);
                        addJSDocTags(changes, sourceFile, parent, [typeTag]);
                    }
                    else if (!tryReplaceImportTypeNodeWithAutoImport(typeNode, declaration, sourceFile, changes, importAdder, ts.getEmitScriptTarget(program.getCompilerOptions()))) {
                        changes.tryInsertTypeAnnotation(sourceFile, declaration, typeNode);
                    }
                }
            }
            function tryReplaceImportTypeNodeWithAutoImport(typeNode, declaration, sourceFile, changes, importAdder, scriptTarget) {
                var importableReference = codefix.tryGetAutoImportableReferenceFromTypeNode(typeNode, scriptTarget);
                if (importableReference && changes.tryInsertTypeAnnotation(sourceFile, declaration, importableReference.typeNode)) {
                    ts.forEach(importableReference.symbols, function (s) { return importAdder.addImportFromExportedSymbol(s, /*usageIsTypeOnly*/ true); });
                    return true;
                }
                return false;
            }
            function annotateJSDocParameters(changes, sourceFile, parameterInferences, program, host) {
                var signature = parameterInferences.length && parameterInferences[0].declaration.parent;
                if (!signature) {
                    return;
                }
                var inferences = ts.mapDefined(parameterInferences, function (inference) {
                    var param = inference.declaration;
                    // only infer parameters that have (1) no type and (2) an accessible inferred type
                    if (param.initializer || ts.getJSDocType(param) || !ts.isIdentifier(param.name)) {
                        return;
                    }
                    var typeNode = inference.type && ts.getTypeNodeIfAccessible(inference.type, param, program, host);
                    if (typeNode) {
                        var name = ts.factory.cloneNode(param.name);
                        ts.setEmitFlags(name, 1536 /* NoComments */ | 2048 /* NoNestedComments */);
                        return { name: ts.factory.cloneNode(param.name), param: param, isOptional: !!inference.isOptional, typeNode: typeNode };
                    }
                });
                if (!inferences.length) {
                    return;
                }
                if (ts.isArrowFunction(signature) || ts.isFunctionExpression(signature)) {
                    var needParens = ts.isArrowFunction(signature) && !ts.findChildOfKind(signature, 20 /* OpenParenToken */, sourceFile);
                    if (needParens) {
                        changes.insertNodeBefore(sourceFile, ts.first(signature.parameters), ts.factory.createToken(20 /* OpenParenToken */));
                    }
                    ts.forEach(inferences, function (_a) {
                        var typeNode = _a.typeNode, param = _a.param;
                        var typeTag = ts.factory.createJSDocTypeTag(/*tagName*/ undefined, ts.factory.createJSDocTypeExpression(typeNode));
                        var jsDoc = ts.factory.createJSDocComment(/*comment*/ undefined, [typeTag]);
                        changes.insertNodeAt(sourceFile, param.getStart(sourceFile), jsDoc, { suffix: " " });
                    });
                    if (needParens) {
                        changes.insertNodeAfter(sourceFile, ts.last(signature.parameters), ts.factory.createToken(21 /* CloseParenToken */));
                    }
                }
                else {
                    var paramTags = ts.map(inferences, function (_a) {
                        var name = _a.name, typeNode = _a.typeNode, isOptional = _a.isOptional;
                        return ts.factory.createJSDocParameterTag(/*tagName*/ undefined, name, /*isBracketed*/ !!isOptional, ts.factory.createJSDocTypeExpression(typeNode), /* isNameFirst */ false, /*comment*/ undefined);
                    });
                    addJSDocTags(changes, sourceFile, signature, paramTags);
                }
            }
            function addJSDocTags(changes, sourceFile, parent, newTags) {
                var comments = ts.flatMap(parent.jsDoc, function (j) { return typeof j.comment === "string" ? ts.factory.createJSDocText(j.comment) : j.comment; });
                var oldTags = ts.flatMapToMutable(parent.jsDoc, function (j) { return j.tags; });
                var unmergedNewTags = newTags.filter(function (newTag) {
                    return !oldTags || !oldTags.some(function (tag, i) {
                        var merged = tryMergeJsdocTags(tag, newTag);
                        if (merged)
                            oldTags[i] = merged;
                        return !!merged;
                    });
                });
                var tag = ts.factory.createJSDocComment(ts.factory.createNodeArray(ts.intersperse(comments, ts.factory.createJSDocText("\n"))), ts.factory.createNodeArray(__spreadArray(__spreadArray([], (oldTags || ts.emptyArray), true), unmergedNewTags, true)));
                var jsDocNode = parent.kind === 212 /* ArrowFunction */ ? getJsDocNodeForArrowFunction(parent) : parent;
                jsDocNode.jsDoc = parent.jsDoc;
                jsDocNode.jsDocCache = parent.jsDocCache;
                changes.insertJsdocCommentBefore(sourceFile, jsDocNode, tag);
            }
            codefix.addJSDocTags = addJSDocTags;
            function getJsDocNodeForArrowFunction(signature) {
                if (signature.parent.kind === 165 /* PropertyDeclaration */) {
                    return signature.parent;
                }
                return signature.parent.parent;
            }
            function tryMergeJsdocTags(oldTag, newTag) {
                if (oldTag.kind !== newTag.kind) {
                    return undefined;
                }
                switch (oldTag.kind) {
                    case 335 /* JSDocParameterTag */: {
                        var oldParam = oldTag;
                        var newParam = newTag;
                        return ts.isIdentifier(oldParam.name) && ts.isIdentifier(newParam.name) && oldParam.name.escapedText === newParam.name.escapedText
                            ? ts.factory.createJSDocParameterTag(/*tagName*/ undefined, newParam.name, /*isBracketed*/ false, newParam.typeExpression, newParam.isNameFirst, oldParam.comment)
                            : undefined;
                    }
                    case 336 /* JSDocReturnTag */:
                        return ts.factory.createJSDocReturnTag(/*tagName*/ undefined, newTag.typeExpression, oldTag.comment);
                }
            }
            function getReferences(token, program, cancellationToken) {
                // Position shouldn't matter since token is not a SourceFile.
                return ts.mapDefined(ts.FindAllReferences.getReferenceEntriesForNode(-1, token, program, program.getSourceFiles(), cancellationToken), function (entry) {
                    return entry.kind !== 0 /* Span */ ? ts.tryCast(entry.node, ts.isIdentifier) : undefined;
                });
            }
            function inferTypeForVariableFromUsage(token, program, cancellationToken) {
                var references = getReferences(token, program, cancellationToken);
                return inferTypeFromReferences(program, references, cancellationToken).single();
            }
            function inferTypeForParametersFromUsage(func, sourceFile, program, cancellationToken) {
                var references = getFunctionReferences(func, sourceFile, program, cancellationToken);
                return references && inferTypeFromReferences(program, references, cancellationToken).parameters(func) ||
                    func.parameters.map(function (p) {
                        return ({
                            declaration: p,
                            type: ts.isIdentifier(p.name) ? inferTypeForVariableFromUsage(p.name, program, cancellationToken) : program.getTypeChecker().getAnyType()
                        });
                    });
            }
            function getFunctionReferences(containingFunction, sourceFile, program, cancellationToken) {
                var searchToken;
                switch (containingFunction.kind) {
                    case 169 /* Constructor */:
                        searchToken = ts.findChildOfKind(containingFunction, 133 /* ConstructorKeyword */, sourceFile);
                        break;
                    case 212 /* ArrowFunction */:
                    case 211 /* FunctionExpression */:
                        var parent = containingFunction.parent;
                        searchToken = (ts.isVariableDeclaration(parent) || ts.isPropertyDeclaration(parent)) && ts.isIdentifier(parent.name) ?
                            parent.name :
                            containingFunction.name;
                        break;
                    case 254 /* FunctionDeclaration */:
                    case 167 /* MethodDeclaration */:
                    case 166 /* MethodSignature */:
                        searchToken = containingFunction.name;
                        break;
                }
                if (!searchToken) {
                    return undefined;
                }
                return getReferences(searchToken, program, cancellationToken);
            }
            function inferTypeFromReferences(program, references, cancellationToken) {
                var checker = program.getTypeChecker();
                var builtinConstructors = {
                    string: function () { return checker.getStringType(); },
                    number: function () { return checker.getNumberType(); },
                    Array: function (t) { return checker.createArrayType(t); },
                    Promise: function (t) { return checker.createPromiseType(t); },
                };
                var builtins = [
                    checker.getStringType(),
                    checker.getNumberType(),
                    checker.createArrayType(checker.getAnyType()),
                    checker.createPromiseType(checker.getAnyType()),
                ];
                return {
                    single: single,
                    parameters: parameters,
                    thisParameter: thisParameter,
                };
                function createEmptyUsage() {
                    return {
                        isNumber: undefined,
                        isString: undefined,
                        isNumberOrString: undefined,
                        candidateTypes: undefined,
                        properties: undefined,
                        calls: undefined,
                        constructs: undefined,
                        numberIndex: undefined,
                        stringIndex: undefined,
                        candidateThisTypes: undefined,
                        inferredTypes: undefined,
                    };
                }
                function combineUsages(usages) {
                    var combinedProperties = new ts.Map();
                    for (var _i = 0, usages_1 = usages; _i < usages_1.length; _i++) {
                        var u = usages_1[_i];
                        if (u.properties) {
                            u.properties.forEach(function (p, name) {
                                if (!combinedProperties.has(name)) {
                                    combinedProperties.set(name, []);
                                }
                                combinedProperties.get(name).push(p);
                            });
                        }
                    }
                    var properties = new ts.Map();
                    combinedProperties.forEach(function (ps, name) {
                        properties.set(name, combineUsages(ps));
                    });
                    return {
                        isNumber: usages.some(function (u) { return u.isNumber; }),
                        isString: usages.some(function (u) { return u.isString; }),
                        isNumberOrString: usages.some(function (u) { return u.isNumberOrString; }),
                        candidateTypes: ts.flatMap(usages, function (u) { return u.candidateTypes; }),
                        properties: properties,
                        calls: ts.flatMap(usages, function (u) { return u.calls; }),
                        constructs: ts.flatMap(usages, function (u) { return u.constructs; }),
                        numberIndex: ts.forEach(usages, function (u) { return u.numberIndex; }),
                        stringIndex: ts.forEach(usages, function (u) { return u.stringIndex; }),
                        candidateThisTypes: ts.flatMap(usages, function (u) { return u.candidateThisTypes; }),
                        inferredTypes: undefined, // clear type cache
                    };
                }
                function single() {
                    return combineTypes(inferTypesFromReferencesSingle(references));
                }
                function parameters(declaration) {
                    if (references.length === 0 || !declaration.parameters) {
                        return undefined;
                    }
                    var usage = createEmptyUsage();
                    for (var _i = 0, references_3 = references; _i < references_3.length; _i++) {
                        var reference = references_3[_i];
                        cancellationToken.throwIfCancellationRequested();
                        calculateUsageOfNode(reference, usage);
                    }
                    var calls = __spreadArray(__spreadArray([], usage.constructs || [], true), usage.calls || [], true);
                    return declaration.parameters.map(function (parameter, parameterIndex) {
                        var types = [];
                        var isRest = ts.isRestParameter(parameter);
                        var isOptional = false;
                        for (var _i = 0, calls_1 = calls; _i < calls_1.length; _i++) {
                            var call = calls_1[_i];
                            if (call.argumentTypes.length <= parameterIndex) {
                                isOptional = ts.isInJSFile(declaration);
                                types.push(checker.getUndefinedType());
                            }
                            else if (isRest) {
                                for (var i = parameterIndex; i < call.argumentTypes.length; i++) {
                                    types.push(checker.getBaseTypeOfLiteralType(call.argumentTypes[i]));
                                }
                            }
                            else {
                                types.push(checker.getBaseTypeOfLiteralType(call.argumentTypes[parameterIndex]));
                            }
                        }
                        if (ts.isIdentifier(parameter.name)) {
                            var inferred = inferTypesFromReferencesSingle(getReferences(parameter.name, program, cancellationToken));
                            types.push.apply(types, (isRest ? ts.mapDefined(inferred, checker.getElementTypeOfArrayType) : inferred));
                        }
                        var type = combineTypes(types);
                        return {
                            type: isRest ? checker.createArrayType(type) : type,
                            isOptional: isOptional && !isRest,
                            declaration: parameter
                        };
                    });
                }
                function thisParameter() {
                    var usage = createEmptyUsage();
                    for (var _i = 0, references_4 = references; _i < references_4.length; _i++) {
                        var reference = references_4[_i];
                        cancellationToken.throwIfCancellationRequested();
                        calculateUsageOfNode(reference, usage);
                    }
                    return combineTypes(usage.candidateThisTypes || ts.emptyArray);
                }
                function inferTypesFromReferencesSingle(references) {
                    var usage = createEmptyUsage();
                    for (var _i = 0, references_5 = references; _i < references_5.length; _i++) {
                        var reference = references_5[_i];
                        cancellationToken.throwIfCancellationRequested();
                        calculateUsageOfNode(reference, usage);
                    }
                    return inferTypes(usage);
                }
                function calculateUsageOfNode(node, usage) {
                    while (ts.isRightSideOfQualifiedNameOrPropertyAccess(node)) {
                        node = node.parent;
                    }
                    switch (node.parent.kind) {
                        case 236 /* ExpressionStatement */:
                            inferTypeFromExpressionStatement(node, usage);
                            break;
                        case 218 /* PostfixUnaryExpression */:
                            usage.isNumber = true;
                            break;
                        case 217 /* PrefixUnaryExpression */:
                            inferTypeFromPrefixUnaryExpression(node.parent, usage);
                            break;
                        case 219 /* BinaryExpression */:
                            inferTypeFromBinaryExpression(node, node.parent, usage);
                            break;
                        case 287 /* CaseClause */:
                        case 288 /* DefaultClause */:
                            inferTypeFromSwitchStatementLabel(node.parent, usage);
                            break;
                        case 206 /* CallExpression */:
                        case 207 /* NewExpression */:
                            if (node.parent.expression === node) {
                                inferTypeFromCallExpression(node.parent, usage);
                            }
                            else {
                                inferTypeFromContextualType(node, usage);
                            }
                            break;
                        case 204 /* PropertyAccessExpression */:
                            inferTypeFromPropertyAccessExpression(node.parent, usage);
                            break;
                        case 205 /* ElementAccessExpression */:
                            inferTypeFromPropertyElementExpression(node.parent, node, usage);
                            break;
                        case 291 /* PropertyAssignment */:
                        case 292 /* ShorthandPropertyAssignment */:
                            inferTypeFromPropertyAssignment(node.parent, usage);
                            break;
                        case 165 /* PropertyDeclaration */:
                            inferTypeFromPropertyDeclaration(node.parent, usage);
                            break;
                        case 252 /* VariableDeclaration */: {
                            var _a = node.parent, name = _a.name, initializer = _a.initializer;
                            if (node === name) {
                                if (initializer) { // This can happen for `let x = null;` which still has an implicit-any error.
                                    addCandidateType(usage, checker.getTypeAtLocation(initializer));
                                }
                                break;
                            }
                        }
                        // falls through
                        default:
                            return inferTypeFromContextualType(node, usage);
                    }
                }
                function inferTypeFromContextualType(node, usage) {
                    if (ts.isExpressionNode(node)) {
                        addCandidateType(usage, checker.getContextualType(node));
                    }
                }
                function inferTypeFromExpressionStatement(node, usage) {
                    addCandidateType(usage, ts.isCallExpression(node) ? checker.getVoidType() : checker.getAnyType());
                }
                function inferTypeFromPrefixUnaryExpression(node, usage) {
                    switch (node.operator) {
                        case 45 /* PlusPlusToken */:
                        case 46 /* MinusMinusToken */:
                        case 40 /* MinusToken */:
                        case 54 /* TildeToken */:
                            usage.isNumber = true;
                            break;
                        case 39 /* PlusToken */:
                            usage.isNumberOrString = true;
                            break;
                        // case SyntaxKind.ExclamationToken:
                        // no inferences here;
                    }
                }
                function inferTypeFromBinaryExpression(node, parent, usage) {
                    switch (parent.operatorToken.kind) {
                        // ExponentiationOperator
                        case 42 /* AsteriskAsteriskToken */:
                        // MultiplicativeOperator
                        // falls through
                        case 41 /* AsteriskToken */:
                        case 43 /* SlashToken */:
                        case 44 /* PercentToken */:
                        // ShiftOperator
                        // falls through
                        case 47 /* LessThanLessThanToken */:
                        case 48 /* GreaterThanGreaterThanToken */:
                        case 49 /* GreaterThanGreaterThanGreaterThanToken */:
                        // BitwiseOperator
                        // falls through
                        case 50 /* AmpersandToken */:
                        case 51 /* BarToken */:
                        case 52 /* CaretToken */:
                        // CompoundAssignmentOperator
                        // falls through
                        case 65 /* MinusEqualsToken */:
                        case 67 /* AsteriskAsteriskEqualsToken */:
                        case 66 /* AsteriskEqualsToken */:
                        case 68 /* SlashEqualsToken */:
                        case 69 /* PercentEqualsToken */:
                        case 73 /* AmpersandEqualsToken */:
                        case 74 /* BarEqualsToken */:
                        case 78 /* CaretEqualsToken */:
                        case 70 /* LessThanLessThanEqualsToken */:
                        case 72 /* GreaterThanGreaterThanGreaterThanEqualsToken */:
                        case 71 /* GreaterThanGreaterThanEqualsToken */:
                        // AdditiveOperator
                        // falls through
                        case 40 /* MinusToken */:
                        // RelationalOperator
                        // falls through
                        case 29 /* LessThanToken */:
                        case 32 /* LessThanEqualsToken */:
                        case 31 /* GreaterThanToken */:
                        case 33 /* GreaterThanEqualsToken */:
                            var operandType = checker.getTypeAtLocation(parent.left === node ? parent.right : parent.left);
                            if (operandType.flags & 1056 /* EnumLike */) {
                                addCandidateType(usage, operandType);
                            }
                            else {
                                usage.isNumber = true;
                            }
                            break;
                        case 64 /* PlusEqualsToken */:
                        case 39 /* PlusToken */:
                            var otherOperandType = checker.getTypeAtLocation(parent.left === node ? parent.right : parent.left);
                            if (otherOperandType.flags & 1056 /* EnumLike */) {
                                addCandidateType(usage, otherOperandType);
                            }
                            else if (otherOperandType.flags & 296 /* NumberLike */) {
                                usage.isNumber = true;
                            }
                            else if (otherOperandType.flags & 402653316 /* StringLike */) {
                                usage.isString = true;
                            }
                            else if (otherOperandType.flags & 1 /* Any */) {
                                // do nothing, maybe we'll learn something elsewhere
                            }
                            else {
                                usage.isNumberOrString = true;
                            }
                            break;
                        //  AssignmentOperators
                        case 63 /* EqualsToken */:
                        case 34 /* EqualsEqualsToken */:
                        case 36 /* EqualsEqualsEqualsToken */:
                        case 37 /* ExclamationEqualsEqualsToken */:
                        case 35 /* ExclamationEqualsToken */:
                            addCandidateType(usage, checker.getTypeAtLocation(parent.left === node ? parent.right : parent.left));
                            break;
                        case 101 /* InKeyword */:
                            if (node === parent.left) {
                                usage.isString = true;
                            }
                            break;
                        // LogicalOperator Or NullishCoalescing
                        case 56 /* BarBarToken */:
                        case 60 /* QuestionQuestionToken */:
                            if (node === parent.left &&
                                (node.parent.parent.kind === 252 /* VariableDeclaration */ || ts.isAssignmentExpression(node.parent.parent, /*excludeCompoundAssignment*/ true))) {
                                // var x = x || {};
                                // TODO: use getFalsyflagsOfType
                                addCandidateType(usage, checker.getTypeAtLocation(parent.right));
                            }
                            break;
                        case 55 /* AmpersandAmpersandToken */:
                        case 27 /* CommaToken */:
                        case 102 /* InstanceOfKeyword */:
                            // nothing to infer here
                            break;
                    }
                }
                function inferTypeFromSwitchStatementLabel(parent, usage) {
                    addCandidateType(usage, checker.getTypeAtLocation(parent.parent.parent.expression));
                }
                function inferTypeFromCallExpression(parent, usage) {
                    var call = {
                        argumentTypes: [],
                        return_: createEmptyUsage()
                    };
                    if (parent.arguments) {
                        for (var _i = 0, _a = parent.arguments; _i < _a.length; _i++) {
                            var argument = _a[_i];
                            call.argumentTypes.push(checker.getTypeAtLocation(argument));
                        }
                    }
                    calculateUsageOfNode(parent, call.return_);
                    if (parent.kind === 206 /* CallExpression */) {
                        (usage.calls || (usage.calls = [])).push(call);
                    }
                    else {
                        (usage.constructs || (usage.constructs = [])).push(call);
                    }
                }
                function inferTypeFromPropertyAccessExpression(parent, usage) {
                    var name = ts.escapeLeadingUnderscores(parent.name.text);
                    if (!usage.properties) {
                        usage.properties = new ts.Map();
                    }
                    var propertyUsage = usage.properties.get(name) || createEmptyUsage();
                    calculateUsageOfNode(parent, propertyUsage);
                    usage.properties.set(name, propertyUsage);
                }
                function inferTypeFromPropertyElementExpression(parent, node, usage) {
                    if (node === parent.argumentExpression) {
                        usage.isNumberOrString = true;
                        return;
                    }
                    else {
                        var indexType = checker.getTypeAtLocation(parent.argumentExpression);
                        var indexUsage = createEmptyUsage();
                        calculateUsageOfNode(parent, indexUsage);
                        if (indexType.flags & 296 /* NumberLike */) {
                            usage.numberIndex = indexUsage;
                        }
                        else {
                            usage.stringIndex = indexUsage;
                        }
                    }
                }
                function inferTypeFromPropertyAssignment(assignment, usage) {
                    var nodeWithRealType = ts.isVariableDeclaration(assignment.parent.parent) ?
                        assignment.parent.parent :
                        assignment.parent;
                    addCandidateThisType(usage, checker.getTypeAtLocation(nodeWithRealType));
                }
                function inferTypeFromPropertyDeclaration(declaration, usage) {
                    addCandidateThisType(usage, checker.getTypeAtLocation(declaration.parent));
                }
                function removeLowPriorityInferences(inferences, priorities) {
                    var toRemove = [];
                    for (var _i = 0, inferences_1 = inferences; _i < inferences_1.length; _i++) {
                        var i = inferences_1[_i];
                        for (var _a = 0, priorities_1 = priorities; _a < priorities_1.length; _a++) {
                            var _b = priorities_1[_a], high = _b.high, low = _b.low;
                            if (high(i)) {
                                ts.Debug.assert(!low(i), "Priority can't have both low and high");
                                toRemove.push(low);
                            }
                        }
                    }
                    return inferences.filter(function (i) { return toRemove.every(function (f) { return !f(i); }); });
                }
                function combineFromUsage(usage) {
                    return combineTypes(inferTypes(usage));
                }
                function combineTypes(inferences) {
                    if (!inferences.length)
                        return checker.getAnyType();
                    // 1. string or number individually override string | number
                    // 2. non-any, non-void overrides any or void
                    // 3. non-nullable, non-any, non-void, non-anonymous overrides anonymous types
                    var stringNumber = checker.getUnionType([checker.getStringType(), checker.getNumberType()]);
                    var priorities = [
                        {
                            high: function (t) { return t === checker.getStringType() || t === checker.getNumberType(); },
                            low: function (t) { return t === stringNumber; }
                        },
                        {
                            high: function (t) { return !(t.flags & (1 /* Any */ | 16384 /* Void */)); },
                            low: function (t) { return !!(t.flags & (1 /* Any */ | 16384 /* Void */)); }
                        },
                        {
                            high: function (t) { return !(t.flags & (98304 /* Nullable */ | 1 /* Any */ | 16384 /* Void */)) && !(ts.getObjectFlags(t) & 16 /* Anonymous */); },
                            low: function (t) { return !!(ts.getObjectFlags(t) & 16 /* Anonymous */); }
                        }
                    ];
                    var good = removeLowPriorityInferences(inferences, priorities);
                    var anons = good.filter(function (i) { return ts.getObjectFlags(i) & 16 /* Anonymous */; });
                    if (anons.length) {
                        good = good.filter(function (i) { return !(ts.getObjectFlags(i) & 16 /* Anonymous */); });
                        good.push(combineAnonymousTypes(anons));
                    }
                    return checker.getWidenedType(checker.getUnionType(good.map(checker.getBaseTypeOfLiteralType), 2 /* Subtype */));
                }
                function combineAnonymousTypes(anons) {
                    if (anons.length === 1) {
                        return anons[0];
                    }
                    var calls = [];
                    var constructs = [];
                    var stringIndices = [];
                    var numberIndices = [];
                    var stringIndexReadonly = false;
                    var numberIndexReadonly = false;
                    var props = ts.createMultiMap();
                    for (var _i = 0, anons_1 = anons; _i < anons_1.length; _i++) {
                        var anon = anons_1[_i];
                        for (var _a = 0, _b = checker.getPropertiesOfType(anon); _a < _b.length; _a++) {
                            var p = _b[_a];
                            props.add(p.name, p.valueDeclaration ? checker.getTypeOfSymbolAtLocation(p, p.valueDeclaration) : checker.getAnyType());
                        }
                        calls.push.apply(calls, checker.getSignaturesOfType(anon, 0 /* Call */));
                        constructs.push.apply(constructs, checker.getSignaturesOfType(anon, 1 /* Construct */));
                        var stringIndexInfo = checker.getIndexInfoOfType(anon, 0 /* String */);
                        if (stringIndexInfo) {
                            stringIndices.push(stringIndexInfo.type);
                            stringIndexReadonly = stringIndexReadonly || stringIndexInfo.isReadonly;
                        }
                        var numberIndexInfo = checker.getIndexInfoOfType(anon, 1 /* Number */);
                        if (numberIndexInfo) {
                            numberIndices.push(numberIndexInfo.type);
                            numberIndexReadonly = numberIndexReadonly || numberIndexInfo.isReadonly;
                        }
                    }
                    var members = ts.mapEntries(props, function (name, types) {
                        var isOptional = types.length < anons.length ? 16777216 /* Optional */ : 0;
                        var s = checker.createSymbol(4 /* Property */ | isOptional, name);
                        s.type = checker.getUnionType(types);
                        return [name, s];
                    });
                    var indexInfos = [];
                    if (stringIndices.length)
                        indexInfos.push(checker.createIndexInfo(checker.getStringType(), checker.getUnionType(stringIndices), stringIndexReadonly));
                    if (numberIndices.length)
                        indexInfos.push(checker.createIndexInfo(checker.getNumberType(), checker.getUnionType(numberIndices), numberIndexReadonly));
                    return checker.createAnonymousType(anons[0].symbol, members, calls, constructs, indexInfos);
                }
                function inferTypes(usage) {
                    var _a, _b, _c;
                    var types = [];
                    if (usage.isNumber) {
                        types.push(checker.getNumberType());
                    }
                    if (usage.isString) {
                        types.push(checker.getStringType());
                    }
                    if (usage.isNumberOrString) {
                        types.push(checker.getUnionType([checker.getStringType(), checker.getNumberType()]));
                    }
                    if (usage.numberIndex) {
                        types.push(checker.createArrayType(combineFromUsage(usage.numberIndex)));
                    }
                    if (((_a = usage.properties) === null || _a === void 0 ? void 0 : _a.size) || ((_b = usage.calls) === null || _b === void 0 ? void 0 : _b.length) || ((_c = usage.constructs) === null || _c === void 0 ? void 0 : _c.length) || usage.stringIndex) {
                        types.push(inferStructuralType(usage));
                    }
                    types.push.apply(types, (usage.candidateTypes || []).map(function (t) { return checker.getBaseTypeOfLiteralType(t); }));
                    types.push.apply(types, inferNamedTypesFromProperties(usage));
                    return types;
                }
                function inferStructuralType(usage) {
                    var members = new ts.Map();
                    if (usage.properties) {
                        usage.properties.forEach(function (u, name) {
                            var symbol = checker.createSymbol(4 /* Property */, name);
                            symbol.type = combineFromUsage(u);
                            members.set(name, symbol);
                        });
                    }
                    var callSignatures = usage.calls ? [getSignatureFromCalls(usage.calls)] : [];
                    var constructSignatures = usage.constructs ? [getSignatureFromCalls(usage.constructs)] : [];
                    var indexInfos = usage.stringIndex ? [checker.createIndexInfo(checker.getStringType(), combineFromUsage(usage.stringIndex), /*isReadonly*/ false)] : [];
                    return checker.createAnonymousType(/*symbol*/ undefined, members, callSignatures, constructSignatures, indexInfos);
                }
                function inferNamedTypesFromProperties(usage) {
                    if (!usage.properties || !usage.properties.size)
                        return [];
                    var types = builtins.filter(function (t) { return allPropertiesAreAssignableToUsage(t, usage); });
                    if (0 < types.length && types.length < 3) {
                        return types.map(function (t) { return inferInstantiationFromUsage(t, usage); });
                    }
                    return [];
                }
                function allPropertiesAreAssignableToUsage(type, usage) {
                    if (!usage.properties)
                        return false;
                    return !ts.forEachEntry(usage.properties, function (propUsage, name) {
                        var source = checker.getTypeOfPropertyOfType(type, name);
                        if (!source) {
                            return true;
                        }
                        if (propUsage.calls) {
                            var sigs = checker.getSignaturesOfType(source, 0 /* Call */);
                            return !sigs.length || !checker.isTypeAssignableTo(source, getFunctionFromCalls(propUsage.calls));
                        }
                        else {
                            return !checker.isTypeAssignableTo(source, combineFromUsage(propUsage));
                        }
                    });
                }
                /**
                 * inference is limited to
                 * 1. generic types with a single parameter
                 * 2. inference to/from calls with a single signature
                 */
                function inferInstantiationFromUsage(type, usage) {
                    if (!(ts.getObjectFlags(type) & 4 /* Reference */) || !usage.properties) {
                        return type;
                    }
                    var generic = type.target;
                    var singleTypeParameter = ts.singleOrUndefined(generic.typeParameters);
                    if (!singleTypeParameter)
                        return type;
                    var types = [];
                    usage.properties.forEach(function (propUsage, name) {
                        var genericPropertyType = checker.getTypeOfPropertyOfType(generic, name);
                        ts.Debug.assert(!!genericPropertyType, "generic should have all the properties of its reference.");
                        types.push.apply(types, inferTypeParameters(genericPropertyType, combineFromUsage(propUsage), singleTypeParameter));
                    });
                    return builtinConstructors[type.symbol.escapedName](combineTypes(types));
                }
                function inferTypeParameters(genericType, usageType, typeParameter) {
                    if (genericType === typeParameter) {
                        return [usageType];
                    }
                    else if (genericType.flags & 3145728 /* UnionOrIntersection */) {
                        return ts.flatMap(genericType.types, function (t) { return inferTypeParameters(t, usageType, typeParameter); });
                    }
                    else if (ts.getObjectFlags(genericType) & 4 /* Reference */ && ts.getObjectFlags(usageType) & 4 /* Reference */) {
                        // this is wrong because we need a reference to the targetType to, so we can check that it's also a reference
                        var genericArgs = checker.getTypeArguments(genericType);
                        var usageArgs = checker.getTypeArguments(usageType);
                        var types = [];
                        if (genericArgs && usageArgs) {
                            for (var i = 0; i < genericArgs.length; i++) {
                                if (usageArgs[i]) {
                                    types.push.apply(types, inferTypeParameters(genericArgs[i], usageArgs[i], typeParameter));
                                }
                            }
                        }
                        return types;
                    }
                    var genericSigs = checker.getSignaturesOfType(genericType, 0 /* Call */);
                    var usageSigs = checker.getSignaturesOfType(usageType, 0 /* Call */);
                    if (genericSigs.length === 1 && usageSigs.length === 1) {
                        return inferFromSignatures(genericSigs[0], usageSigs[0], typeParameter);
                    }
                    return [];
                }
                function inferFromSignatures(genericSig, usageSig, typeParameter) {
                    var types = [];
                    for (var i = 0; i < genericSig.parameters.length; i++) {
                        var genericParam = genericSig.parameters[i];
                        var usageParam = usageSig.parameters[i];
                        var isRest = genericSig.declaration && ts.isRestParameter(genericSig.declaration.parameters[i]);
                        if (!usageParam) {
                            break;
                        }
                        var genericParamType = genericParam.valueDeclaration ? checker.getTypeOfSymbolAtLocation(genericParam, genericParam.valueDeclaration) : checker.getAnyType();
                        var elementType = isRest && checker.getElementTypeOfArrayType(genericParamType);
                        if (elementType) {
                            genericParamType = elementType;
                        }
                        var targetType = usageParam.type
                            || (usageParam.valueDeclaration ? checker.getTypeOfSymbolAtLocation(usageParam, usageParam.valueDeclaration) : checker.getAnyType());
                        types.push.apply(types, inferTypeParameters(genericParamType, targetType, typeParameter));
                    }
                    var genericReturn = checker.getReturnTypeOfSignature(genericSig);
                    var usageReturn = checker.getReturnTypeOfSignature(usageSig);
                    types.push.apply(types, inferTypeParameters(genericReturn, usageReturn, typeParameter));
                    return types;
                }
                function getFunctionFromCalls(calls) {
                    return checker.createAnonymousType(/*symbol*/ undefined, ts.createSymbolTable(), [getSignatureFromCalls(calls)], ts.emptyArray, ts.emptyArray);
                }
                function getSignatureFromCalls(calls) {
                    var parameters = [];
                    var length = Math.max.apply(Math, calls.map(function (c) { return c.argumentTypes.length; }));
                    var _loop_15 = function (i) {
                        var symbol = checker.createSymbol(1 /* FunctionScopedVariable */, ts.escapeLeadingUnderscores("arg" + i));
                        symbol.type = combineTypes(calls.map(function (call) { return call.argumentTypes[i] || checker.getUndefinedType(); }));
                        if (calls.some(function (call) { return call.argumentTypes[i] === undefined; })) {
                            symbol.flags |= 16777216 /* Optional */;
                        }
                        parameters.push(symbol);
                    };
                    for (var i = 0; i < length; i++) {
                        _loop_15(i);
                    }
                    var returnType = combineFromUsage(combineUsages(calls.map(function (call) { return call.return_; })));
                    return checker.createSignature(/*declaration*/ undefined, /*typeParameters*/ undefined, /*thisParameter*/ undefined, parameters, returnType, /*typePredicate*/ undefined, length, 0 /* None */);
                }
                function addCandidateType(usage, type) {
                    if (type && !(type.flags & 1 /* Any */) && !(type.flags & 131072 /* Never */)) {
                        (usage.candidateTypes || (usage.candidateTypes = [])).push(type);
                    }
                }
                function addCandidateThisType(usage, type) {
                    if (type && !(type.flags & 1 /* Any */) && !(type.flags & 131072 /* Never */)) {
                        (usage.candidateThisTypes || (usage.candidateThisTypes = [])).push(type);
                    }
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixReturnTypeInAsyncFunction";
            var errorCodes = [
                ts.Diagnostics.The_return_type_of_an_async_function_or_method_must_be_the_global_Promise_T_type_Did_you_mean_to_write_Promise_0.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                fixIds: [fixId],
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, program = context.program, span = context.span;
                    var checker = program.getTypeChecker();
                    var info = getInfo(sourceFile, program.getTypeChecker(), span.start);
                    if (!info) {
                        return undefined;
                    }
                    var returnTypeNode = info.returnTypeNode, returnType = info.returnType, promisedTypeNode = info.promisedTypeNode, promisedType = info.promisedType;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, returnTypeNode, promisedTypeNode); });
                    return [codefix.createCodeFixAction(fixId, changes, [ts.Diagnostics.Replace_0_with_Promise_1,
                            checker.typeToString(returnType), checker.typeToString(promisedType)], fixId, ts.Diagnostics.Fix_all_incorrect_return_type_of_an_async_functions)];
                },
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var info = getInfo(diag.file, context.program.getTypeChecker(), diag.start);
                        if (info) {
                            doChange(changes, diag.file, info.returnTypeNode, info.promisedTypeNode);
                        }
                    });
                }
            });
            function getInfo(sourceFile, checker, pos) {
                if (ts.isInJSFile(sourceFile)) {
                    return undefined;
                }
                var token = ts.getTokenAtPosition(sourceFile, pos);
                var func = ts.findAncestor(token, ts.isFunctionLikeDeclaration);
                var returnTypeNode = func === null || func === void 0 ? void 0 : func.type;
                if (!returnTypeNode) {
                    return undefined;
                }
                var returnType = checker.getTypeFromTypeNode(returnTypeNode);
                var promisedType = checker.getAwaitedType(returnType) || checker.getVoidType();
                var promisedTypeNode = checker.typeToTypeNode(promisedType, /*enclosingDeclaration*/ returnTypeNode, /*flags*/ undefined);
                if (promisedTypeNode) {
                    return { returnTypeNode: returnTypeNode, returnType: returnType, promisedTypeNode: promisedTypeNode, promisedType: promisedType };
                }
            }
            function doChange(changes, sourceFile, returnTypeNode, promisedTypeNode) {
                changes.replaceNode(sourceFile, returnTypeNode, ts.factory.createTypeReferenceNode("Promise", [promisedTypeNode]));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixName = "disableJsDiagnostics";
            var fixId = "disableJsDiagnostics";
            var errorCodes = ts.mapDefined(Object.keys(ts.Diagnostics), function (key) {
                var diag = ts.Diagnostics[key];
                return diag.category === ts.DiagnosticCategory.Error ? diag.code : undefined;
            });
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, program = context.program, span = context.span, host = context.host, formatContext = context.formatContext;
                    if (!ts.isInJSFile(sourceFile) || !ts.isCheckJsEnabledForFile(sourceFile, program.getCompilerOptions())) {
                        return undefined;
                    }
                    var newLineCharacter = sourceFile.checkJsDirective ? "" : ts.getNewLineOrDefaultFromHost(host, formatContext.options);
                    var fixes = [
                        // fixId unnecessary because adding `// @ts-nocheck` even once will ignore every error in the file.
                        codefix.createCodeFixActionWithoutFixAll(fixName, [codefix.createFileTextChanges(sourceFile.fileName, [
                                ts.createTextChange(sourceFile.checkJsDirective
                                    ? ts.createTextSpanFromBounds(sourceFile.checkJsDirective.pos, sourceFile.checkJsDirective.end)
                                    : ts.createTextSpan(0, 0), "// @ts-nocheck" + newLineCharacter),
                            ])], ts.Diagnostics.Disable_checking_for_this_file),
                    ];
                    if (ts.textChanges.isValidLocationToAddComment(sourceFile, span.start)) {
                        fixes.unshift(codefix.createCodeFixAction(fixName, ts.textChanges.ChangeTracker.with(context, function (t) { return makeChange(t, sourceFile, span.start); }), ts.Diagnostics.Ignore_this_error_message, fixId, ts.Diagnostics.Add_ts_ignore_to_all_error_messages));
                    }
                    return fixes;
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    var seenLines = new ts.Set();
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        if (ts.textChanges.isValidLocationToAddComment(diag.file, diag.start)) {
                            makeChange(changes, diag.file, diag.start, seenLines);
                        }
                    });
                },
            });
            function makeChange(changes, sourceFile, position, seenLines) {
                var lineNumber = ts.getLineAndCharacterOfPosition(sourceFile, position).line;
                // Only need to add `// @ts-ignore` for a line once.
                if (!seenLines || ts.tryAddToSet(seenLines, lineNumber)) {
                    changes.insertCommentBeforeLine(sourceFile, lineNumber, position, " @ts-ignore");
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            /**
             * Finds members of the resolved type that are missing in the class pointed to by class decl
             * and generates source code for the missing members.
             * @param possiblyMissingSymbols The collection of symbols to filter and then get insertions for.
             * @param importAdder If provided, type annotations will use identifier type references instead of ImportTypeNodes, and the missing imports will be added to the importAdder.
             * @returns Empty string iff there are no member insertions.
             */
            function createMissingMemberNodes(classDeclaration, possiblyMissingSymbols, sourceFile, context, preferences, importAdder, addClassElement) {
                var classMembers = classDeclaration.symbol.members;
                for (var _i = 0, possiblyMissingSymbols_1 = possiblyMissingSymbols; _i < possiblyMissingSymbols_1.length; _i++) {
                    var symbol = possiblyMissingSymbols_1[_i];
                    if (!classMembers.has(symbol.escapedName)) {
                        addNewNodeForMemberSymbol(symbol, classDeclaration, sourceFile, context, preferences, importAdder, addClassElement);
                    }
                }
            }
            codefix.createMissingMemberNodes = createMissingMemberNodes;
            function getNoopSymbolTrackerWithResolver(context) {
                return {
                    trackSymbol: function () { return false; },
                    moduleResolverHost: ts.getModuleSpecifierResolverHost(context.program, context.host),
                };
            }
            codefix.getNoopSymbolTrackerWithResolver = getNoopSymbolTrackerWithResolver;
            /**
             * @returns Empty string iff there we can't figure out a representation for `symbol` in `enclosingDeclaration`.
             */
            function addNewNodeForMemberSymbol(symbol, enclosingDeclaration, sourceFile, context, preferences, importAdder, addClassElement) {
                var declarations = symbol.getDeclarations();
                if (!(declarations && declarations.length)) {
                    return undefined;
                }
                var checker = context.program.getTypeChecker();
                var scriptTarget = ts.getEmitScriptTarget(context.program.getCompilerOptions());
                var declaration = declarations[0];
                var name = ts.getSynthesizedDeepClone(ts.getNameOfDeclaration(declaration), /*includeTrivia*/ false);
                var visibilityModifier = createVisibilityModifier(ts.getEffectiveModifierFlags(declaration));
                var modifiers = visibilityModifier ? ts.factory.createNodeArray([visibilityModifier]) : undefined;
                var type = checker.getWidenedType(checker.getTypeOfSymbolAtLocation(symbol, enclosingDeclaration));
                var optional = !!(symbol.flags & 16777216 /* Optional */);
                var ambient = !!(enclosingDeclaration.flags & 8388608 /* Ambient */);
                var quotePreference = ts.getQuotePreference(sourceFile, preferences);
                switch (declaration.kind) {
                    case 164 /* PropertySignature */:
                    case 165 /* PropertyDeclaration */:
                        var flags = quotePreference === 0 /* Single */ ? 268435456 /* UseSingleQuotesForStringLiteralType */ : undefined;
                        var typeNode = checker.typeToTypeNode(type, enclosingDeclaration, flags, getNoopSymbolTrackerWithResolver(context));
                        if (importAdder) {
                            var importableReference = tryGetAutoImportableReferenceFromTypeNode(typeNode, scriptTarget);
                            if (importableReference) {
                                typeNode = importableReference.typeNode;
                                importSymbols(importAdder, importableReference.symbols);
                            }
                        }
                        addClassElement(ts.factory.createPropertyDeclaration(
                        /*decorators*/ undefined, modifiers, name, optional ? ts.factory.createToken(57 /* QuestionToken */) : undefined, typeNode, 
                        /*initializer*/ undefined));
                        break;
                    case 170 /* GetAccessor */:
                    case 171 /* SetAccessor */: {
                        var typeNode_1 = checker.typeToTypeNode(type, enclosingDeclaration, /*flags*/ undefined, getNoopSymbolTrackerWithResolver(context));
                        var allAccessors = ts.getAllAccessorDeclarations(declarations, declaration);
                        var orderedAccessors = allAccessors.secondAccessor
                            ? [allAccessors.firstAccessor, allAccessors.secondAccessor]
                            : [allAccessors.firstAccessor];
                        if (importAdder) {
                            var importableReference = tryGetAutoImportableReferenceFromTypeNode(typeNode_1, scriptTarget);
                            if (importableReference) {
                                typeNode_1 = importableReference.typeNode;
                                importSymbols(importAdder, importableReference.symbols);
                            }
                        }
                        for (var _i = 0, orderedAccessors_1 = orderedAccessors; _i < orderedAccessors_1.length; _i++) {
                            var accessor = orderedAccessors_1[_i];
                            if (ts.isGetAccessorDeclaration(accessor)) {
                                addClassElement(ts.factory.createGetAccessorDeclaration(
                                /*decorators*/ undefined, modifiers, name, ts.emptyArray, typeNode_1, ambient ? undefined : createStubbedMethodBody(quotePreference)));
                            }
                            else {
                                ts.Debug.assertNode(accessor, ts.isSetAccessorDeclaration, "The counterpart to a getter should be a setter");
                                var parameter = ts.getSetAccessorValueParameter(accessor);
                                var parameterName = parameter && ts.isIdentifier(parameter.name) ? ts.idText(parameter.name) : undefined;
                                addClassElement(ts.factory.createSetAccessorDeclaration(
                                /*decorators*/ undefined, modifiers, name, createDummyParameters(1, [parameterName], [typeNode_1], 1, /*inJs*/ false), ambient ? undefined : createStubbedMethodBody(quotePreference)));
                            }
                        }
                        break;
                    }
                    case 166 /* MethodSignature */:
                    case 167 /* MethodDeclaration */:
                        // The signature for the implementation appears as an entry in `signatures` iff
                        // there is only one signature.
                        // If there are overloads and an implementation signature, it appears as an
                        // extra declaration that isn't a signature for `type`.
                        // If there is more than one overload but no implementation signature
                        // (eg: an abstract method or interface declaration), there is a 1-1
                        // correspondence of declarations and signatures.
                        var signatures = checker.getSignaturesOfType(type, 0 /* Call */);
                        if (!ts.some(signatures)) {
                            break;
                        }
                        if (declarations.length === 1) {
                            ts.Debug.assert(signatures.length === 1, "One declaration implies one signature");
                            var signature = signatures[0];
                            outputMethod(quotePreference, signature, modifiers, name, ambient ? undefined : createStubbedMethodBody(quotePreference));
                            break;
                        }
                        for (var _a = 0, signatures_1 = signatures; _a < signatures_1.length; _a++) {
                            var signature = signatures_1[_a];
                            // Ensure nodes are fresh so they can have different positions when going through formatting.
                            outputMethod(quotePreference, signature, ts.getSynthesizedDeepClones(modifiers, /*includeTrivia*/ false), ts.getSynthesizedDeepClone(name, /*includeTrivia*/ false));
                        }
                        if (!ambient) {
                            if (declarations.length > signatures.length) {
                                var signature = checker.getSignatureFromDeclaration(declarations[declarations.length - 1]);
                                outputMethod(quotePreference, signature, modifiers, name, createStubbedMethodBody(quotePreference));
                            }
                            else {
                                ts.Debug.assert(declarations.length === signatures.length, "Declarations and signatures should match count");
                                addClassElement(createMethodImplementingSignatures(checker, context, enclosingDeclaration, signatures, name, optional, modifiers, quotePreference));
                            }
                        }
                        break;
                }
                function outputMethod(quotePreference, signature, modifiers, name, body) {
                    var method = createSignatureDeclarationFromSignature(167 /* MethodDeclaration */, context, quotePreference, signature, body, name, modifiers, optional, enclosingDeclaration, importAdder);
                    if (method)
                        addClassElement(method);
                }
            }
            function createSignatureDeclarationFromSignature(kind, context, quotePreference, signature, body, name, modifiers, optional, enclosingDeclaration, importAdder) {
                var program = context.program;
                var checker = program.getTypeChecker();
                var scriptTarget = ts.getEmitScriptTarget(program.getCompilerOptions());
                var flags = 1 /* NoTruncation */ | 1073741824 /* NoUndefinedOptionalParameterType */ | 256 /* SuppressAnyReturnType */ | (quotePreference === 0 /* Single */ ? 268435456 /* UseSingleQuotesForStringLiteralType */ : 0);
                var signatureDeclaration = checker.signatureToSignatureDeclaration(signature, kind, enclosingDeclaration, flags, getNoopSymbolTrackerWithResolver(context));
                if (!signatureDeclaration) {
                    return undefined;
                }
                var typeParameters = signatureDeclaration.typeParameters;
                var parameters = signatureDeclaration.parameters;
                var type = signatureDeclaration.type;
                if (importAdder) {
                    if (typeParameters) {
                        var newTypeParameters = ts.sameMap(typeParameters, function (typeParameterDecl) {
                            var constraint = typeParameterDecl.constraint;
                            var defaultType = typeParameterDecl.default;
                            if (constraint) {
                                var importableReference = tryGetAutoImportableReferenceFromTypeNode(constraint, scriptTarget);
                                if (importableReference) {
                                    constraint = importableReference.typeNode;
                                    importSymbols(importAdder, importableReference.symbols);
                                }
                            }
                            if (defaultType) {
                                var importableReference = tryGetAutoImportableReferenceFromTypeNode(defaultType, scriptTarget);
                                if (importableReference) {
                                    defaultType = importableReference.typeNode;
                                    importSymbols(importAdder, importableReference.symbols);
                                }
                            }
                            return ts.factory.updateTypeParameterDeclaration(typeParameterDecl, typeParameterDecl.name, constraint, defaultType);
                        });
                        if (typeParameters !== newTypeParameters) {
                            typeParameters = ts.setTextRange(ts.factory.createNodeArray(newTypeParameters, typeParameters.hasTrailingComma), typeParameters);
                        }
                    }
                    var newParameters = ts.sameMap(parameters, function (parameterDecl) {
                        var importableReference = tryGetAutoImportableReferenceFromTypeNode(parameterDecl.type, scriptTarget);
                        var type = parameterDecl.type;
                        if (importableReference) {
                            type = importableReference.typeNode;
                            importSymbols(importAdder, importableReference.symbols);
                        }
                        return ts.factory.updateParameterDeclaration(parameterDecl, parameterDecl.decorators, parameterDecl.modifiers, parameterDecl.dotDotDotToken, parameterDecl.name, parameterDecl.questionToken, type, parameterDecl.initializer);
                    });
                    if (parameters !== newParameters) {
                        parameters = ts.setTextRange(ts.factory.createNodeArray(newParameters, parameters.hasTrailingComma), parameters);
                    }
                    if (type) {
                        var importableReference = tryGetAutoImportableReferenceFromTypeNode(type, scriptTarget);
                        if (importableReference) {
                            type = importableReference.typeNode;
                            importSymbols(importAdder, importableReference.symbols);
                        }
                    }
                }
                var questionToken = optional ? ts.factory.createToken(57 /* QuestionToken */) : undefined;
                var asteriskToken = signatureDeclaration.asteriskToken;
                if (ts.isFunctionExpression(signatureDeclaration)) {
                    return ts.factory.updateFunctionExpression(signatureDeclaration, modifiers, signatureDeclaration.asteriskToken, ts.tryCast(name, ts.isIdentifier), typeParameters, parameters, type, body !== null && body !== void 0 ? body : signatureDeclaration.body);
                }
                if (ts.isArrowFunction(signatureDeclaration)) {
                    return ts.factory.updateArrowFunction(signatureDeclaration, modifiers, typeParameters, parameters, type, signatureDeclaration.equalsGreaterThanToken, body !== null && body !== void 0 ? body : signatureDeclaration.body);
                }
                if (ts.isMethodDeclaration(signatureDeclaration)) {
                    return ts.factory.updateMethodDeclaration(signatureDeclaration, /* decorators */ undefined, modifiers, asteriskToken, name !== null && name !== void 0 ? name : ts.factory.createIdentifier(""), questionToken, typeParameters, parameters, type, body);
                }
                return undefined;
            }
            codefix.createSignatureDeclarationFromSignature = createSignatureDeclarationFromSignature;
            function createSignatureDeclarationFromCallExpression(kind, context, importAdder, call, name, modifierFlags, contextNode) {
                var quotePreference = ts.getQuotePreference(context.sourceFile, context.preferences);
                var scriptTarget = ts.getEmitScriptTarget(context.program.getCompilerOptions());
                var tracker = getNoopSymbolTrackerWithResolver(context);
                var checker = context.program.getTypeChecker();
                var isJs = ts.isInJSFile(contextNode);
                var typeArguments = call.typeArguments, args = call.arguments, parent = call.parent;
                var contextualType = isJs ? undefined : checker.getContextualType(call);
                var names = ts.map(args, function (arg) {
                    return ts.isIdentifier(arg) ? arg.text : ts.isPropertyAccessExpression(arg) && ts.isIdentifier(arg.name) ? arg.name.text : undefined;
                });
                var types = isJs ? [] : ts.map(args, function (arg) {
                    return typeToAutoImportableTypeNode(checker, importAdder, checker.getBaseTypeOfLiteralType(checker.getTypeAtLocation(arg)), contextNode, scriptTarget, /*flags*/ undefined, tracker);
                });
                var modifiers = modifierFlags
                    ? ts.factory.createNodeArray(ts.factory.createModifiersFromModifierFlags(modifierFlags))
                    : undefined;
                var asteriskToken = ts.isYieldExpression(parent)
                    ? ts.factory.createToken(41 /* AsteriskToken */)
                    : undefined;
                var typeParameters = isJs || typeArguments === undefined
                    ? undefined
                    : ts.map(typeArguments, function (_, i) {
                        return ts.factory.createTypeParameterDeclaration(84 /* T */ + typeArguments.length - 1 <= 90 /* Z */ ? String.fromCharCode(84 /* T */ + i) : "T" + i);
                    });
                var parameters = createDummyParameters(args.length, names, types, /*minArgumentCount*/ undefined, isJs);
                var type = isJs || contextualType === undefined
                    ? undefined
                    : checker.typeToTypeNode(contextualType, contextNode, /*flags*/ undefined, tracker);
                if (kind === 167 /* MethodDeclaration */) {
                    return ts.factory.createMethodDeclaration(
                    /*decorators*/ undefined, modifiers, asteriskToken, name, 
                    /*questionToken*/ undefined, typeParameters, parameters, type, ts.isInterfaceDeclaration(contextNode) ? undefined : createStubbedMethodBody(quotePreference));
                }
                return ts.factory.createFunctionDeclaration(
                /*decorators*/ undefined, modifiers, asteriskToken, name, typeParameters, parameters, type, createStubbedBody(ts.Diagnostics.Function_not_implemented.message, quotePreference));
            }
            codefix.createSignatureDeclarationFromCallExpression = createSignatureDeclarationFromCallExpression;
            function typeToAutoImportableTypeNode(checker, importAdder, type, contextNode, scriptTarget, flags, tracker) {
                var typeNode = checker.typeToTypeNode(type, contextNode, flags, tracker);
                if (typeNode && ts.isImportTypeNode(typeNode)) {
                    var importableReference = tryGetAutoImportableReferenceFromTypeNode(typeNode, scriptTarget);
                    if (importableReference) {
                        importSymbols(importAdder, importableReference.symbols);
                        typeNode = importableReference.typeNode;
                    }
                }
                // Ensure nodes are fresh so they can have different positions when going through formatting.
                return ts.getSynthesizedDeepClone(typeNode);
            }
            codefix.typeToAutoImportableTypeNode = typeToAutoImportableTypeNode;
            function createDummyParameters(argCount, names, types, minArgumentCount, inJs) {
                var parameters = [];
                for (var i = 0; i < argCount; i++) {
                    var newParameter = ts.factory.createParameterDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, 
                    /*dotDotDotToken*/ undefined, 
                    /*name*/ names && names[i] || "arg" + i, 
                    /*questionToken*/ minArgumentCount !== undefined && i >= minArgumentCount ? ts.factory.createToken(57 /* QuestionToken */) : undefined, 
                    /*type*/ inJs ? undefined : types && types[i] || ts.factory.createKeywordTypeNode(129 /* AnyKeyword */), 
                    /*initializer*/ undefined);
                    parameters.push(newParameter);
                }
                return parameters;
            }
            function createMethodImplementingSignatures(checker, context, enclosingDeclaration, signatures, name, optional, modifiers, quotePreference) {
                /** This is *a* signature with the maximal number of arguments,
                 * such that if there is a "maximal" signature without rest arguments,
                 * this is one of them.
                 */
                var maxArgsSignature = signatures[0];
                var minArgumentCount = signatures[0].minArgumentCount;
                var someSigHasRestParameter = false;
                for (var _i = 0, signatures_2 = signatures; _i < signatures_2.length; _i++) {
                    var sig = signatures_2[_i];
                    minArgumentCount = Math.min(sig.minArgumentCount, minArgumentCount);
                    if (ts.signatureHasRestParameter(sig)) {
                        someSigHasRestParameter = true;
                    }
                    if (sig.parameters.length >= maxArgsSignature.parameters.length && (!ts.signatureHasRestParameter(sig) || ts.signatureHasRestParameter(maxArgsSignature))) {
                        maxArgsSignature = sig;
                    }
                }
                var maxNonRestArgs = maxArgsSignature.parameters.length - (ts.signatureHasRestParameter(maxArgsSignature) ? 1 : 0);
                var maxArgsParameterSymbolNames = maxArgsSignature.parameters.map(function (symbol) { return symbol.name; });
                var parameters = createDummyParameters(maxNonRestArgs, maxArgsParameterSymbolNames, /* types */ undefined, minArgumentCount, /*inJs*/ false);
                if (someSigHasRestParameter) {
                    var anyArrayType = ts.factory.createArrayTypeNode(ts.factory.createKeywordTypeNode(129 /* AnyKeyword */));
                    var restParameter = ts.factory.createParameterDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, ts.factory.createToken(25 /* DotDotDotToken */), maxArgsParameterSymbolNames[maxNonRestArgs] || "rest", 
                    /*questionToken*/ maxNonRestArgs >= minArgumentCount ? ts.factory.createToken(57 /* QuestionToken */) : undefined, anyArrayType, 
                    /*initializer*/ undefined);
                    parameters.push(restParameter);
                }
                return createStubbedMethod(modifiers, name, optional, 
                /*typeParameters*/ undefined, parameters, getReturnTypeFromSignatures(signatures, checker, context, enclosingDeclaration), quotePreference);
            }
            function getReturnTypeFromSignatures(signatures, checker, context, enclosingDeclaration) {
                if (ts.length(signatures)) {
                    var type = checker.getUnionType(ts.map(signatures, checker.getReturnTypeOfSignature));
                    return checker.typeToTypeNode(type, enclosingDeclaration, /*flags*/ undefined, getNoopSymbolTrackerWithResolver(context));
                }
            }
            function createStubbedMethod(modifiers, name, optional, typeParameters, parameters, returnType, quotePreference) {
                return ts.factory.createMethodDeclaration(
                /*decorators*/ undefined, modifiers, 
                /*asteriskToken*/ undefined, name, optional ? ts.factory.createToken(57 /* QuestionToken */) : undefined, typeParameters, parameters, returnType, createStubbedMethodBody(quotePreference));
            }
            function createStubbedMethodBody(quotePreference) {
                return createStubbedBody(ts.Diagnostics.Method_not_implemented.message, quotePreference);
            }
            function createStubbedBody(text, quotePreference) {
                return ts.factory.createBlock([ts.factory.createThrowStatement(ts.factory.createNewExpression(ts.factory.createIdentifier("Error"), 
                    /*typeArguments*/ undefined, 
                    // TODO Handle auto quote preference.
                    [ts.factory.createStringLiteral(text, /*isSingleQuote*/ quotePreference === 0 /* Single */)]))], 
                /*multiline*/ true);
            }
            codefix.createStubbedBody = createStubbedBody;
            function createVisibilityModifier(flags) {
                if (flags & 4 /* Public */) {
                    return ts.factory.createToken(123 /* PublicKeyword */);
                }
                else if (flags & 16 /* Protected */) {
                    return ts.factory.createToken(122 /* ProtectedKeyword */);
                }
                return undefined;
            }
            function setJsonCompilerOptionValues(changeTracker, configFile, options) {
                var tsconfigObjectLiteral = ts.getTsConfigObjectLiteralExpression(configFile);
                if (!tsconfigObjectLiteral)
                    return undefined;
                var compilerOptionsProperty = findJsonProperty(tsconfigObjectLiteral, "compilerOptions");
                if (compilerOptionsProperty === undefined) {
                    changeTracker.insertNodeAtObjectStart(configFile, tsconfigObjectLiteral, createJsonPropertyAssignment("compilerOptions", ts.factory.createObjectLiteralExpression(options.map(function (_a) {
                        var optionName = _a[0], optionValue = _a[1];
                        return createJsonPropertyAssignment(optionName, optionValue);
                    }), /*multiLine*/ true)));
                    return;
                }
                var compilerOptions = compilerOptionsProperty.initializer;
                if (!ts.isObjectLiteralExpression(compilerOptions)) {
                    return;
                }
                for (var _i = 0, options_1 = options; _i < options_1.length; _i++) {
                    var _a = options_1[_i], optionName = _a[0], optionValue = _a[1];
                    var optionProperty = findJsonProperty(compilerOptions, optionName);
                    if (optionProperty === undefined) {
                        changeTracker.insertNodeAtObjectStart(configFile, compilerOptions, createJsonPropertyAssignment(optionName, optionValue));
                    }
                    else {
                        changeTracker.replaceNode(configFile, optionProperty.initializer, optionValue);
                    }
                }
            }
            codefix.setJsonCompilerOptionValues = setJsonCompilerOptionValues;
            function setJsonCompilerOptionValue(changeTracker, configFile, optionName, optionValue) {
                setJsonCompilerOptionValues(changeTracker, configFile, [[optionName, optionValue]]);
            }
            codefix.setJsonCompilerOptionValue = setJsonCompilerOptionValue;
            function createJsonPropertyAssignment(name, initializer) {
                return ts.factory.createPropertyAssignment(ts.factory.createStringLiteral(name), initializer);
            }
            codefix.createJsonPropertyAssignment = createJsonPropertyAssignment;
            function findJsonProperty(obj, name) {
                return ts.find(obj.properties, function (p) { return ts.isPropertyAssignment(p) && !!p.name && ts.isStringLiteral(p.name) && p.name.text === name; });
            }
            codefix.findJsonProperty = findJsonProperty;
            /**
             * Given a type node containing 'import("./a").SomeType<import("./b").OtherType<...>>',
             * returns an equivalent type reference node with any nested ImportTypeNodes also replaced
             * with type references, and a list of symbols that must be imported to use the type reference.
             */
            function tryGetAutoImportableReferenceFromTypeNode(importTypeNode, scriptTarget) {
                var symbols;
                var typeNode = ts.visitNode(importTypeNode, visit);
                if (symbols && typeNode) {
                    return { typeNode: typeNode, symbols: symbols };
                }
                function visit(node) {
                    var _a;
                    if (ts.isLiteralImportTypeNode(node) && node.qualifier) {
                        // Symbol for the left-most thing after the dot
                        var firstIdentifier = ts.getFirstIdentifier(node.qualifier);
                        var name = ts.getNameForExportedSymbol(firstIdentifier.symbol, scriptTarget);
                        var qualifier = name !== firstIdentifier.text
                            ? replaceFirstIdentifierOfEntityName(node.qualifier, ts.factory.createIdentifier(name))
                            : node.qualifier;
                        symbols = ts.append(symbols, firstIdentifier.symbol);
                        var typeArguments = (_a = node.typeArguments) === null || _a === void 0 ? void 0 : _a.map(visit);
                        return ts.factory.createTypeReferenceNode(qualifier, typeArguments);
                    }
                    return ts.visitEachChild(node, visit, ts.nullTransformationContext);
                }
            }
            codefix.tryGetAutoImportableReferenceFromTypeNode = tryGetAutoImportableReferenceFromTypeNode;
            function replaceFirstIdentifierOfEntityName(name, newIdentifier) {
                if (name.kind === 79 /* Identifier */) {
                    return newIdentifier;
                }
                return ts.factory.createQualifiedName(replaceFirstIdentifierOfEntityName(name.left, newIdentifier), name.right);
            }
            function importSymbols(importAdder, symbols) {
                symbols.forEach(function (s) { return importAdder.addImportFromExportedSymbol(s, /*usageIsTypeOnly*/ true); });
            }
            codefix.importSymbols = importSymbols;
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            function generateAccessorFromProperty(file, program, start, end, context, _actionName) {
                var fieldInfo = getAccessorConvertiblePropertyAtPosition(file, program, start, end);
                if (!fieldInfo || ts.refactor.isRefactorErrorInfo(fieldInfo))
                    return undefined;
                var changeTracker = ts.textChanges.ChangeTracker.fromContext(context);
                var isStatic = fieldInfo.isStatic, isReadonly = fieldInfo.isReadonly, fieldName = fieldInfo.fieldName, accessorName = fieldInfo.accessorName, originalName = fieldInfo.originalName, type = fieldInfo.type, container = fieldInfo.container, declaration = fieldInfo.declaration;
                ts.suppressLeadingAndTrailingTrivia(fieldName);
                ts.suppressLeadingAndTrailingTrivia(accessorName);
                ts.suppressLeadingAndTrailingTrivia(declaration);
                ts.suppressLeadingAndTrailingTrivia(container);
                var accessorModifiers;
                var fieldModifiers;
                if (ts.isClassLike(container)) {
                    var modifierFlags = ts.getEffectiveModifierFlags(declaration);
                    if (ts.isSourceFileJS(file)) {
                        var modifiers = ts.createModifiers(modifierFlags);
                        accessorModifiers = modifiers;
                        fieldModifiers = modifiers;
                    }
                    else {
                        accessorModifiers = ts.createModifiers(prepareModifierFlagsForAccessor(modifierFlags));
                        fieldModifiers = ts.createModifiers(prepareModifierFlagsForField(modifierFlags));
                    }
                }
                updateFieldDeclaration(changeTracker, file, declaration, type, fieldName, fieldModifiers);
                var getAccessor = generateGetAccessor(fieldName, accessorName, type, accessorModifiers, isStatic, container);
                ts.suppressLeadingAndTrailingTrivia(getAccessor);
                insertAccessor(changeTracker, file, getAccessor, declaration, container);
                if (isReadonly) {
                    // readonly modifier only existed in classLikeDeclaration
                    var constructor = ts.getFirstConstructorWithBody(container);
                    if (constructor) {
                        updateReadonlyPropertyInitializerStatementConstructor(changeTracker, file, constructor, fieldName.text, originalName);
                    }
                }
                else {
                    var setAccessor = generateSetAccessor(fieldName, accessorName, type, accessorModifiers, isStatic, container);
                    ts.suppressLeadingAndTrailingTrivia(setAccessor);
                    insertAccessor(changeTracker, file, setAccessor, declaration, container);
                }
                return changeTracker.getChanges();
            }
            codefix.generateAccessorFromProperty = generateAccessorFromProperty;
            function isConvertibleName(name) {
                return ts.isIdentifier(name) || ts.isStringLiteral(name);
            }
            function isAcceptedDeclaration(node) {
                return ts.isParameterPropertyDeclaration(node, node.parent) || ts.isPropertyDeclaration(node) || ts.isPropertyAssignment(node);
            }
            function createPropertyName(name, originalName) {
                return ts.isIdentifier(originalName) ? ts.factory.createIdentifier(name) : ts.factory.createStringLiteral(name);
            }
            function createAccessorAccessExpression(fieldName, isStatic, container) {
                var leftHead = isStatic ? container.name : ts.factory.createThis(); // TODO: GH#18217
                return ts.isIdentifier(fieldName) ? ts.factory.createPropertyAccessExpression(leftHead, fieldName) : ts.factory.createElementAccessExpression(leftHead, ts.factory.createStringLiteralFromNode(fieldName));
            }
            function prepareModifierFlagsForAccessor(modifierFlags) {
                modifierFlags &= ~64 /* Readonly */; // avoid Readonly modifier because it will convert to get accessor
                modifierFlags &= ~8 /* Private */;
                if (!(modifierFlags & 16 /* Protected */)) {
                    modifierFlags |= 4 /* Public */;
                }
                return modifierFlags;
            }
            function prepareModifierFlagsForField(modifierFlags) {
                modifierFlags &= ~4 /* Public */;
                modifierFlags &= ~16 /* Protected */;
                modifierFlags |= 8 /* Private */;
                return modifierFlags;
            }
            function getAccessorConvertiblePropertyAtPosition(file, program, start, end, considerEmptySpans) {
                if (considerEmptySpans === void 0) {
                    considerEmptySpans = true;
                }
                var node = ts.getTokenAtPosition(file, start);
                var cursorRequest = start === end && considerEmptySpans;
                var declaration = ts.findAncestor(node.parent, isAcceptedDeclaration);
                // make sure declaration have AccessibilityModifier or Static Modifier or Readonly Modifier
                var meaning = 28 /* AccessibilityModifier */ | 32 /* Static */ | 64 /* Readonly */;
                if (!declaration || (!(ts.nodeOverlapsWithStartEnd(declaration.name, file, start, end) || cursorRequest))) {
                    return {
                        error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_property_for_which_to_generate_accessor)
                    };
                }
                if (!isConvertibleName(declaration.name)) {
                    return {
                        error: ts.getLocaleSpecificMessage(ts.Diagnostics.Name_is_not_valid)
                    };
                }
                if ((ts.getEffectiveModifierFlags(declaration) | meaning) !== meaning) {
                    return {
                        error: ts.getLocaleSpecificMessage(ts.Diagnostics.Can_only_convert_property_with_modifier)
                    };
                }
                var name = declaration.name.text;
                var startWithUnderscore = ts.startsWithUnderscore(name);
                var fieldName = createPropertyName(startWithUnderscore ? name : ts.getUniqueName("_" + name, file), declaration.name);
                var accessorName = createPropertyName(startWithUnderscore ? ts.getUniqueName(name.substring(1), file) : name, declaration.name);
                return {
                    isStatic: ts.hasStaticModifier(declaration),
                    isReadonly: ts.hasEffectiveReadonlyModifier(declaration),
                    type: getDeclarationType(declaration, program),
                    container: declaration.kind === 162 /* Parameter */ ? declaration.parent.parent : declaration.parent,
                    originalName: declaration.name.text,
                    declaration: declaration,
                    fieldName: fieldName,
                    accessorName: accessorName,
                    renameAccessor: startWithUnderscore
                };
            }
            codefix.getAccessorConvertiblePropertyAtPosition = getAccessorConvertiblePropertyAtPosition;
            function generateGetAccessor(fieldName, accessorName, type, modifiers, isStatic, container) {
                return ts.factory.createGetAccessorDeclaration(
                /*decorators*/ undefined, modifiers, accessorName, 
                /*parameters*/ undefined, // TODO: GH#18217
                type, ts.factory.createBlock([
                    ts.factory.createReturnStatement(createAccessorAccessExpression(fieldName, isStatic, container))
                ], /*multiLine*/ true));
            }
            function generateSetAccessor(fieldName, accessorName, type, modifiers, isStatic, container) {
                return ts.factory.createSetAccessorDeclaration(
                /*decorators*/ undefined, modifiers, accessorName, [ts.factory.createParameterDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, 
                    /*dotDotDotToken*/ undefined, ts.factory.createIdentifier("value"), 
                    /*questionToken*/ undefined, type)], ts.factory.createBlock([
                    ts.factory.createExpressionStatement(ts.factory.createAssignment(createAccessorAccessExpression(fieldName, isStatic, container), ts.factory.createIdentifier("value")))
                ], /*multiLine*/ true));
            }
            function updatePropertyDeclaration(changeTracker, file, declaration, type, fieldName, modifiers) {
                var property = ts.factory.updatePropertyDeclaration(declaration, declaration.decorators, modifiers, fieldName, declaration.questionToken || declaration.exclamationToken, type, declaration.initializer);
                changeTracker.replaceNode(file, declaration, property);
            }
            function updatePropertyAssignmentDeclaration(changeTracker, file, declaration, fieldName) {
                var assignment = ts.factory.updatePropertyAssignment(declaration, fieldName, declaration.initializer);
                changeTracker.replacePropertyAssignment(file, declaration, assignment);
            }
            function updateFieldDeclaration(changeTracker, file, declaration, type, fieldName, modifiers) {
                if (ts.isPropertyDeclaration(declaration)) {
                    updatePropertyDeclaration(changeTracker, file, declaration, type, fieldName, modifiers);
                }
                else if (ts.isPropertyAssignment(declaration)) {
                    updatePropertyAssignmentDeclaration(changeTracker, file, declaration, fieldName);
                }
                else {
                    changeTracker.replaceNode(file, declaration, ts.factory.updateParameterDeclaration(declaration, declaration.decorators, modifiers, declaration.dotDotDotToken, ts.cast(fieldName, ts.isIdentifier), declaration.questionToken, declaration.type, declaration.initializer));
                }
            }
            function insertAccessor(changeTracker, file, accessor, declaration, container) {
                ts.isParameterPropertyDeclaration(declaration, declaration.parent) ? changeTracker.insertNodeAtClassStart(file, container, accessor) :
                    ts.isPropertyAssignment(declaration) ? changeTracker.insertNodeAfterComma(file, declaration, accessor) :
                        changeTracker.insertNodeAfter(file, declaration, accessor);
            }
            function updateReadonlyPropertyInitializerStatementConstructor(changeTracker, file, constructor, fieldName, originalName) {
                if (!constructor.body)
                    return;
                constructor.body.forEachChild(function recur(node) {
                    if (ts.isElementAccessExpression(node) &&
                        node.expression.kind === 108 /* ThisKeyword */ &&
                        ts.isStringLiteral(node.argumentExpression) &&
                        node.argumentExpression.text === originalName &&
                        ts.isWriteAccess(node)) {
                        changeTracker.replaceNode(file, node.argumentExpression, ts.factory.createStringLiteral(fieldName));
                    }
                    if (ts.isPropertyAccessExpression(node) && node.expression.kind === 108 /* ThisKeyword */ && node.name.text === originalName && ts.isWriteAccess(node)) {
                        changeTracker.replaceNode(file, node.name, ts.factory.createIdentifier(fieldName));
                    }
                    if (!ts.isFunctionLike(node) && !ts.isClassLike(node)) {
                        node.forEachChild(recur);
                    }
                });
            }
            function getDeclarationType(declaration, program) {
                var typeNode = ts.getTypeAnnotationNode(declaration);
                if (ts.isPropertyDeclaration(declaration) && typeNode && declaration.questionToken) {
                    var typeChecker = program.getTypeChecker();
                    var type = typeChecker.getTypeFromTypeNode(typeNode);
                    if (!typeChecker.isTypeAssignableTo(typeChecker.getUndefinedType(), type)) {
                        var types = ts.isUnionTypeNode(typeNode) ? typeNode.types : [typeNode];
                        return ts.factory.createUnionTypeNode(__spreadArray(__spreadArray([], types, true), [ts.factory.createKeywordTypeNode(151 /* UndefinedKeyword */)], false));
                    }
                }
                return typeNode;
            }
            function getAllSupers(decl, checker) {
                var res = [];
                while (decl) {
                    var superElement = ts.getClassExtendsHeritageElement(decl);
                    var superSymbol = superElement && checker.getSymbolAtLocation(superElement.expression);
                    if (!superSymbol)
                        break;
                    var symbol = superSymbol.flags & 2097152 /* Alias */ ? checker.getAliasedSymbol(superSymbol) : superSymbol;
                    var superDecl = symbol.declarations && ts.find(symbol.declarations, ts.isClassLike);
                    if (!superDecl)
                        break;
                    res.push(superDecl);
                    decl = superDecl;
                }
                return res;
            }
            codefix.getAllSupers = getAllSupers;
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixName = "invalidImportSyntax";
            function getCodeFixesForImportDeclaration(context, node) {
                var sourceFile = ts.getSourceFileOfNode(node);
                var namespace = ts.getNamespaceDeclarationNode(node);
                var opts = context.program.getCompilerOptions();
                var variations = [];
                // import Bluebird from "bluebird";
                variations.push(createAction(context, sourceFile, node, ts.makeImport(namespace.name, /*namedImports*/ undefined, node.moduleSpecifier, ts.getQuotePreference(sourceFile, context.preferences))));
                if (ts.getEmitModuleKind(opts) === ts.ModuleKind.CommonJS) {
                    // import Bluebird = require("bluebird");
                    variations.push(createAction(context, sourceFile, node, ts.factory.createImportEqualsDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, 
                    /*isTypeOnly*/ false, namespace.name, ts.factory.createExternalModuleReference(node.moduleSpecifier))));
                }
                return variations;
            }
            function createAction(context, sourceFile, node, replacement) {
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return t.replaceNode(sourceFile, node, replacement); });
                return codefix.createCodeFixActionWithoutFixAll(fixName, changes, [ts.Diagnostics.Replace_import_with_0, changes[0].textChanges[0].newText]);
            }
            codefix.registerCodeFix({
                errorCodes: [
                    ts.Diagnostics.This_expression_is_not_callable.code,
                    ts.Diagnostics.This_expression_is_not_constructable.code,
                ],
                getCodeActions: getActionsForUsageOfInvalidImport
            });
            function getActionsForUsageOfInvalidImport(context) {
                var sourceFile = context.sourceFile;
                var targetKind = ts.Diagnostics.This_expression_is_not_callable.code === context.errorCode ? 206 /* CallExpression */ : 207 /* NewExpression */;
                var node = ts.findAncestor(ts.getTokenAtPosition(sourceFile, context.span.start), function (a) { return a.kind === targetKind; });
                if (!node) {
                    return [];
                }
                var expr = node.expression;
                return getImportCodeFixesForExpression(context, expr);
            }
            codefix.registerCodeFix({
                errorCodes: [
                    // The following error codes cover pretty much all assignability errors that could involve an expression
                    ts.Diagnostics.Argument_of_type_0_is_not_assignable_to_parameter_of_type_1.code,
                    ts.Diagnostics.Type_0_does_not_satisfy_the_constraint_1.code,
                    ts.Diagnostics.Type_0_is_not_assignable_to_type_1.code,
                    ts.Diagnostics.Type_0_is_not_assignable_to_type_1_Two_different_types_with_this_name_exist_but_they_are_unrelated.code,
                    ts.Diagnostics.Type_predicate_0_is_not_assignable_to_1.code,
                    ts.Diagnostics.Property_0_of_type_1_is_not_assignable_to_2_index_type_3.code,
                    ts.Diagnostics._0_index_type_1_is_not_assignable_to_2_index_type_3.code,
                    ts.Diagnostics.Property_0_in_type_1_is_not_assignable_to_the_same_property_in_base_type_2.code,
                    ts.Diagnostics.Property_0_in_type_1_is_not_assignable_to_type_2.code,
                    ts.Diagnostics.Property_0_of_JSX_spread_attribute_is_not_assignable_to_target_property.code,
                    ts.Diagnostics.The_this_context_of_type_0_is_not_assignable_to_method_s_this_of_type_1.code,
                ],
                getCodeActions: getActionsForInvalidImportLocation
            });
            function getActionsForInvalidImportLocation(context) {
                var sourceFile = context.sourceFile;
                var node = ts.findAncestor(ts.getTokenAtPosition(sourceFile, context.span.start), function (a) { return a.getStart() === context.span.start && a.getEnd() === (context.span.start + context.span.length); });
                if (!node) {
                    return [];
                }
                return getImportCodeFixesForExpression(context, node);
            }
            function getImportCodeFixesForExpression(context, expr) {
                var type = context.program.getTypeChecker().getTypeAtLocation(expr);
                if (!(type.symbol && type.symbol.originatingImport)) {
                    return [];
                }
                var fixes = [];
                var relatedImport = type.symbol.originatingImport; // TODO: GH#18217
                if (!ts.isImportCall(relatedImport)) {
                    ts.addRange(fixes, getCodeFixesForImportDeclaration(context, relatedImport));
                }
                if (ts.isExpression(expr) && !(ts.isNamedDeclaration(expr.parent) && expr.parent.name === expr)) {
                    var sourceFile_2 = context.sourceFile;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return t.replaceNode(sourceFile_2, expr, ts.factory.createPropertyAccessExpression(expr, "default"), {}); });
                    fixes.push(codefix.createCodeFixActionWithoutFixAll(fixName, changes, ts.Diagnostics.Use_synthetic_default_member));
                }
                return fixes;
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixName = "strictClassInitialization";
            var fixIdAddDefiniteAssignmentAssertions = "addMissingPropertyDefiniteAssignmentAssertions";
            var fixIdAddUndefinedType = "addMissingPropertyUndefinedType";
            var fixIdAddInitializer = "addMissingPropertyInitializer";
            var errorCodes = [ts.Diagnostics.Property_0_has_no_initializer_and_is_not_definitely_assigned_in_the_constructor.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var propertyDeclaration = getPropertyDeclaration(context.sourceFile, context.span.start);
                    if (!propertyDeclaration)
                        return;
                    var result = [
                        getActionForAddMissingUndefinedType(context, propertyDeclaration),
                        getActionForAddMissingDefiniteAssignmentAssertion(context, propertyDeclaration)
                    ];
                    ts.append(result, getActionForAddMissingInitializer(context, propertyDeclaration));
                    return result;
                },
                fixIds: [fixIdAddDefiniteAssignmentAssertions, fixIdAddUndefinedType, fixIdAddInitializer],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var propertyDeclaration = getPropertyDeclaration(diag.file, diag.start);
                        if (!propertyDeclaration)
                            return;
                        switch (context.fixId) {
                            case fixIdAddDefiniteAssignmentAssertions:
                                addDefiniteAssignmentAssertion(changes, diag.file, propertyDeclaration);
                                break;
                            case fixIdAddUndefinedType:
                                addUndefinedType(changes, diag.file, propertyDeclaration);
                                break;
                            case fixIdAddInitializer:
                                var checker = context.program.getTypeChecker();
                                var initializer = getInitializer(checker, propertyDeclaration);
                                if (!initializer)
                                    return;
                                addInitializer(changes, diag.file, propertyDeclaration, initializer);
                                break;
                            default:
                                ts.Debug.fail(JSON.stringify(context.fixId));
                        }
                    });
                },
            });
            function getPropertyDeclaration(sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                return ts.isIdentifier(token) ? ts.cast(token.parent, ts.isPropertyDeclaration) : undefined;
            }
            function getActionForAddMissingDefiniteAssignmentAssertion(context, propertyDeclaration) {
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addDefiniteAssignmentAssertion(t, context.sourceFile, propertyDeclaration); });
                return codefix.createCodeFixAction(fixName, changes, [ts.Diagnostics.Add_definite_assignment_assertion_to_property_0, propertyDeclaration.getText()], fixIdAddDefiniteAssignmentAssertions, ts.Diagnostics.Add_definite_assignment_assertions_to_all_uninitialized_properties);
            }
            function addDefiniteAssignmentAssertion(changeTracker, propertyDeclarationSourceFile, propertyDeclaration) {
                var property = ts.factory.updatePropertyDeclaration(propertyDeclaration, propertyDeclaration.decorators, propertyDeclaration.modifiers, propertyDeclaration.name, ts.factory.createToken(53 /* ExclamationToken */), propertyDeclaration.type, propertyDeclaration.initializer);
                changeTracker.replaceNode(propertyDeclarationSourceFile, propertyDeclaration, property);
            }
            function getActionForAddMissingUndefinedType(context, propertyDeclaration) {
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addUndefinedType(t, context.sourceFile, propertyDeclaration); });
                return codefix.createCodeFixAction(fixName, changes, [ts.Diagnostics.Add_undefined_type_to_property_0, propertyDeclaration.name.getText()], fixIdAddUndefinedType, ts.Diagnostics.Add_undefined_type_to_all_uninitialized_properties);
            }
            function addUndefinedType(changeTracker, propertyDeclarationSourceFile, propertyDeclaration) {
                var undefinedTypeNode = ts.factory.createKeywordTypeNode(151 /* UndefinedKeyword */);
                var type = propertyDeclaration.type; // TODO: GH#18217
                var types = ts.isUnionTypeNode(type) ? type.types.concat(undefinedTypeNode) : [type, undefinedTypeNode];
                changeTracker.replaceNode(propertyDeclarationSourceFile, type, ts.factory.createUnionTypeNode(types));
            }
            function getActionForAddMissingInitializer(context, propertyDeclaration) {
                var checker = context.program.getTypeChecker();
                var initializer = getInitializer(checker, propertyDeclaration);
                if (!initializer)
                    return undefined;
                var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return addInitializer(t, context.sourceFile, propertyDeclaration, initializer); });
                return codefix.createCodeFixAction(fixName, changes, [ts.Diagnostics.Add_initializer_to_property_0, propertyDeclaration.name.getText()], fixIdAddInitializer, ts.Diagnostics.Add_initializers_to_all_uninitialized_properties);
            }
            function addInitializer(changeTracker, propertyDeclarationSourceFile, propertyDeclaration, initializer) {
                var property = ts.factory.updatePropertyDeclaration(propertyDeclaration, propertyDeclaration.decorators, propertyDeclaration.modifiers, propertyDeclaration.name, propertyDeclaration.questionToken, propertyDeclaration.type, initializer);
                changeTracker.replaceNode(propertyDeclarationSourceFile, propertyDeclaration, property);
            }
            function getInitializer(checker, propertyDeclaration) {
                return getDefaultValueFromType(checker, checker.getTypeFromTypeNode(propertyDeclaration.type)); // TODO: GH#18217
            }
            function getDefaultValueFromType(checker, type) {
                if (type.flags & 512 /* BooleanLiteral */) {
                    return (type === checker.getFalseType() || type === checker.getFalseType(/*fresh*/ true)) ? ts.factory.createFalse() : ts.factory.createTrue();
                }
                else if (type.isStringLiteral()) {
                    return ts.factory.createStringLiteral(type.value);
                }
                else if (type.isNumberLiteral()) {
                    return ts.factory.createNumericLiteral(type.value);
                }
                else if (type.flags & 2048 /* BigIntLiteral */) {
                    return ts.factory.createBigIntLiteral(type.value);
                }
                else if (type.isUnion()) {
                    return ts.firstDefined(type.types, function (t) { return getDefaultValueFromType(checker, t); });
                }
                else if (type.isClass()) {
                    var classDeclaration = ts.getClassLikeDeclarationOfSymbol(type.symbol);
                    if (!classDeclaration || ts.hasSyntacticModifier(classDeclaration, 128 /* Abstract */))
                        return undefined;
                    var constructorDeclaration = ts.getFirstConstructorWithBody(classDeclaration);
                    if (constructorDeclaration && constructorDeclaration.parameters.length)
                        return undefined;
                    return ts.factory.createNewExpression(ts.factory.createIdentifier(type.symbol.name), /*typeArguments*/ undefined, /*argumentsArray*/ undefined);
                }
                else if (checker.isArrayLikeType(type)) {
                    return ts.factory.createArrayLiteralExpression();
                }
                return undefined;
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "requireInTs";
            var errorCodes = [ts.Diagnostics.require_call_may_be_converted_to_an_import.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var info = getInfo(context.sourceFile, context.program, context.span.start);
                    if (!info) {
                        return undefined;
                    }
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, context.sourceFile, info); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Convert_require_to_import, fixId, ts.Diagnostics.Convert_all_require_to_import)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var info = getInfo(diag.file, context.program, diag.start);
                        if (info) {
                            doChange(changes, context.sourceFile, info);
                        }
                    });
                },
            });
            function doChange(changes, sourceFile, info) {
                var allowSyntheticDefaults = info.allowSyntheticDefaults, defaultImportName = info.defaultImportName, namedImports = info.namedImports, statement = info.statement, required = info.required;
                changes.replaceNode(sourceFile, statement, defaultImportName && !allowSyntheticDefaults
                    ? ts.factory.createImportEqualsDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*isTypeOnly*/ false, defaultImportName, ts.factory.createExternalModuleReference(required))
                    : ts.factory.createImportDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, ts.factory.createImportClause(/*isTypeOnly*/ false, defaultImportName, namedImports), required));
            }
            function getInfo(sourceFile, program, pos) {
                var parent = ts.getTokenAtPosition(sourceFile, pos).parent;
                if (!ts.isRequireCall(parent, /*checkArgumentIsStringLiteralLike*/ true)) {
                    throw ts.Debug.failBadSyntaxKind(parent);
                }
                var decl = ts.cast(parent.parent, ts.isVariableDeclaration);
                var defaultImportName = ts.tryCast(decl.name, ts.isIdentifier);
                var namedImports = ts.isObjectBindingPattern(decl.name) ? tryCreateNamedImportsFromObjectBindingPattern(decl.name) : undefined;
                if (defaultImportName || namedImports) {
                    return {
                        allowSyntheticDefaults: ts.getAllowSyntheticDefaultImports(program.getCompilerOptions()),
                        defaultImportName: defaultImportName,
                        namedImports: namedImports,
                        statement: ts.cast(decl.parent.parent, ts.isVariableStatement),
                        required: ts.first(parent.arguments)
                    };
                }
            }
            function tryCreateNamedImportsFromObjectBindingPattern(node) {
                var importSpecifiers = [];
                for (var _i = 0, _a = node.elements; _i < _a.length; _i++) {
                    var element = _a[_i];
                    if (!ts.isIdentifier(element.name) || element.initializer) {
                        return undefined;
                    }
                    importSpecifiers.push(ts.factory.createImportSpecifier(ts.tryCast(element.propertyName, ts.isIdentifier), element.name));
                }
                if (importSpecifiers.length) {
                    return ts.factory.createNamedImports(importSpecifiers);
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "useDefaultImport";
            var errorCodes = [ts.Diagnostics.Import_may_be_converted_to_a_default_import.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, start = context.span.start;
                    var info = getInfo(sourceFile, start);
                    if (!info)
                        return undefined;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, info, context.preferences); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Convert_to_default_import, fixId, ts.Diagnostics.Convert_all_to_default_imports)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var info = getInfo(diag.file, diag.start);
                        if (info)
                            doChange(changes, diag.file, info, context.preferences);
                    });
                },
            });
            function getInfo(sourceFile, pos) {
                var name = ts.getTokenAtPosition(sourceFile, pos);
                if (!ts.isIdentifier(name))
                    return undefined; // bad input
                var parent = name.parent;
                if (ts.isImportEqualsDeclaration(parent) && ts.isExternalModuleReference(parent.moduleReference)) {
                    return { importNode: parent, name: name, moduleSpecifier: parent.moduleReference.expression };
                }
                else if (ts.isNamespaceImport(parent)) {
                    var importNode = parent.parent.parent;
                    return { importNode: importNode, name: name, moduleSpecifier: importNode.moduleSpecifier };
                }
            }
            function doChange(changes, sourceFile, info, preferences) {
                changes.replaceNode(sourceFile, info.importNode, ts.makeImport(info.name, /*namedImports*/ undefined, info.moduleSpecifier, ts.getQuotePreference(sourceFile, preferences)));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "useBigintLiteral";
            var errorCodes = [
                ts.Diagnostics.Numeric_literals_with_absolute_values_equal_to_2_53_or_greater_are_too_large_to_be_represented_accurately_as_integers.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return makeChange(t, context.sourceFile, context.span); });
                    if (changes.length > 0) {
                        return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Convert_to_a_bigint_numeric_literal, fixId, ts.Diagnostics.Convert_all_to_bigint_numeric_literals)];
                    }
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return makeChange(changes, diag.file, diag); });
                },
            });
            function makeChange(changeTracker, sourceFile, span) {
                var numericLiteral = ts.tryCast(ts.getTokenAtPosition(sourceFile, span.start), ts.isNumericLiteral);
                if (!numericLiteral) {
                    return;
                }
                // We use .getText to overcome parser inaccuracies: https://github.com/microsoft/TypeScript/issues/33298
                var newText = numericLiteral.getText(sourceFile) + "n";
                changeTracker.replaceNode(sourceFile, numericLiteral, ts.factory.createBigIntLiteral(newText));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixIdAddMissingTypeof = "fixAddModuleReferTypeMissingTypeof";
            var fixId = fixIdAddMissingTypeof;
            var errorCodes = [ts.Diagnostics.Module_0_does_not_refer_to_a_type_but_is_used_as_a_type_here_Did_you_mean_typeof_import_0.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var importType = getImportTypeNode(sourceFile, span.start);
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, importType); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Add_missing_typeof, fixId, ts.Diagnostics.Add_missing_typeof)];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        return doChange(changes, context.sourceFile, getImportTypeNode(diag.file, diag.start));
                    });
                },
            });
            function getImportTypeNode(sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                ts.Debug.assert(token.kind === 100 /* ImportKeyword */, "This token should be an ImportKeyword");
                ts.Debug.assert(token.parent.kind === 198 /* ImportType */, "Token parent should be an ImportType");
                return token.parent;
            }
            function doChange(changes, sourceFile, importType) {
                var newTypeNode = ts.factory.updateImportTypeNode(importType, importType.argument, importType.qualifier, importType.typeArguments, /* isTypeOf */ true);
                changes.replaceNode(sourceFile, importType, newTypeNode);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixID = "wrapJsxInFragment";
            var errorCodes = [ts.Diagnostics.JSX_expressions_must_have_one_parent_element.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var node = findNodeToFix(sourceFile, span.start);
                    if (!node)
                        return undefined;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, node); });
                    return [codefix.createCodeFixAction(fixID, changes, ts.Diagnostics.Wrap_in_JSX_fragment, fixID, ts.Diagnostics.Wrap_all_unparented_JSX_in_JSX_fragment)];
                },
                fixIds: [fixID],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var node = findNodeToFix(context.sourceFile, diag.start);
                        if (!node)
                            return undefined;
                        doChange(changes, context.sourceFile, node);
                    });
                },
            });
            function findNodeToFix(sourceFile, pos) {
                // The error always at 1st token that is "<" in "<a /><a />"
                var lessThanToken = ts.getTokenAtPosition(sourceFile, pos);
                var firstJsxElementOrOpenElement = lessThanToken.parent;
                var binaryExpr = firstJsxElementOrOpenElement.parent;
                if (!ts.isBinaryExpression(binaryExpr)) {
                    // In case the start element is a JsxSelfClosingElement, it the end.
                    // For JsxOpenElement, find one more parent
                    binaryExpr = binaryExpr.parent;
                    if (!ts.isBinaryExpression(binaryExpr))
                        return undefined;
                }
                if (!ts.nodeIsMissing(binaryExpr.operatorToken))
                    return undefined;
                return binaryExpr;
            }
            function doChange(changeTracker, sf, node) {
                var jsx = flattenInvalidBinaryExpr(node);
                if (jsx)
                    changeTracker.replaceNode(sf, node, ts.factory.createJsxFragment(ts.factory.createJsxOpeningFragment(), jsx, ts.factory.createJsxJsxClosingFragment()));
            }
            // The invalid syntax is constructed as
            // InvalidJsxTree :: One of
            //     JsxElement CommaToken InvalidJsxTree
            //     JsxElement CommaToken JsxElement
            function flattenInvalidBinaryExpr(node) {
                var children = [];
                var current = node;
                while (true) {
                    if (ts.isBinaryExpression(current) && ts.nodeIsMissing(current.operatorToken) && current.operatorToken.kind === 27 /* CommaToken */) {
                        children.push(current.left);
                        if (ts.isJsxChild(current.right)) {
                            children.push(current.right);
                            // Indicates the tree has go to the bottom
                            return children;
                        }
                        else if (ts.isBinaryExpression(current.right)) {
                            current = current.right;
                            continue;
                        }
                        // Unreachable case
                        else
                            return undefined;
                    }
                    // Unreachable case
                    else
                        return undefined;
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixIdAddMissingTypeof = "fixConvertToMappedObjectType";
            var fixId = fixIdAddMissingTypeof;
            var errorCodes = [ts.Diagnostics.An_index_signature_parameter_type_cannot_be_a_literal_type_or_generic_type_Consider_using_a_mapped_object_type_instead.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span;
                    var info = getInfo(sourceFile, span.start);
                    if (!info)
                        return undefined;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, info); });
                    var name = ts.idText(info.container.name);
                    return [codefix.createCodeFixAction(fixId, changes, [ts.Diagnostics.Convert_0_to_mapped_object_type, name], fixId, [ts.Diagnostics.Convert_0_to_mapped_object_type, name])];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var info = getInfo(diag.file, diag.start);
                        if (info)
                            doChange(changes, diag.file, info);
                    });
                }
            });
            function getInfo(sourceFile, pos) {
                var token = ts.getTokenAtPosition(sourceFile, pos);
                var indexSignature = ts.cast(token.parent.parent, ts.isIndexSignatureDeclaration);
                if (ts.isClassDeclaration(indexSignature.parent))
                    return undefined;
                var container = ts.isInterfaceDeclaration(indexSignature.parent) ? indexSignature.parent : ts.cast(indexSignature.parent.parent, ts.isTypeAliasDeclaration);
                return { indexSignature: indexSignature, container: container };
            }
            function createTypeAliasFromInterface(declaration, type) {
                return ts.factory.createTypeAliasDeclaration(declaration.decorators, declaration.modifiers, declaration.name, declaration.typeParameters, type);
            }
            function doChange(changes, sourceFile, _a) {
                var indexSignature = _a.indexSignature, container = _a.container;
                var members = ts.isInterfaceDeclaration(container) ? container.members : container.type.members;
                var otherMembers = members.filter(function (member) { return !ts.isIndexSignatureDeclaration(member); });
                var parameter = ts.first(indexSignature.parameters);
                var mappedTypeParameter = ts.factory.createTypeParameterDeclaration(ts.cast(parameter.name, ts.isIdentifier), parameter.type);
                var mappedIntersectionType = ts.factory.createMappedTypeNode(ts.hasEffectiveReadonlyModifier(indexSignature) ? ts.factory.createModifier(143 /* ReadonlyKeyword */) : undefined, mappedTypeParameter, 
                /*nameType*/ undefined, indexSignature.questionToken, indexSignature.type);
                var intersectionType = ts.factory.createIntersectionTypeNode(__spreadArray(__spreadArray(__spreadArray([], ts.getAllSuperTypeNodes(container), true), [
                    mappedIntersectionType
                ], false), (otherMembers.length ? [ts.factory.createTypeLiteralNode(otherMembers)] : ts.emptyArray), true));
                changes.replaceNode(sourceFile, container, createTypeAliasFromInterface(container, intersectionType));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "removeAccidentalCallParentheses";
            var errorCodes = [
                ts.Diagnostics.This_expression_is_not_callable_because_it_is_a_get_accessor_Did_you_mean_to_use_it_without.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var callExpression = ts.findAncestor(ts.getTokenAtPosition(context.sourceFile, context.span.start), ts.isCallExpression);
                    if (!callExpression) {
                        return undefined;
                    }
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) {
                        t.deleteRange(context.sourceFile, { pos: callExpression.expression.end, end: callExpression.end });
                    });
                    return [codefix.createCodeFixActionWithoutFixAll(fixId, changes, ts.Diagnostics.Remove_parentheses)];
                },
                fixIds: [fixId],
            });
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "removeUnnecessaryAwait";
            var errorCodes = [
                ts.Diagnostics.await_has_no_effect_on_the_type_of_this_expression.code,
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return makeChange(t, context.sourceFile, context.span); });
                    if (changes.length > 0) {
                        return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Remove_unnecessary_await, fixId, ts.Diagnostics.Remove_all_unnecessary_uses_of_await)];
                    }
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return makeChange(changes, diag.file, diag); });
                },
            });
            function makeChange(changeTracker, sourceFile, span) {
                var awaitKeyword = ts.tryCast(ts.getTokenAtPosition(sourceFile, span.start), function (node) { return node.kind === 131 /* AwaitKeyword */; });
                var awaitExpression = awaitKeyword && ts.tryCast(awaitKeyword.parent, ts.isAwaitExpression);
                if (!awaitExpression) {
                    return;
                }
                var expressionToReplace = awaitExpression;
                var hasSurroundingParens = ts.isParenthesizedExpression(awaitExpression.parent);
                if (hasSurroundingParens) {
                    var leftMostExpression = ts.getLeftmostExpression(awaitExpression.expression, /*stopAtCallExpressions*/ false);
                    if (ts.isIdentifier(leftMostExpression)) {
                        var precedingToken = ts.findPrecedingToken(awaitExpression.parent.pos, sourceFile);
                        if (precedingToken && precedingToken.kind !== 103 /* NewKeyword */) {
                            expressionToReplace = awaitExpression.parent;
                        }
                    }
                }
                changeTracker.replaceNode(sourceFile, expressionToReplace, awaitExpression.expression);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var errorCodes = [ts.Diagnostics.A_type_only_import_can_specify_a_default_import_or_named_bindings_but_not_both.code];
            var fixId = "splitTypeOnlyImport";
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                fixIds: [fixId],
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) {
                        return splitTypeOnlyImport(t, getImportDeclaration(context.sourceFile, context.span), context);
                    });
                    if (changes.length) {
                        return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Split_into_two_separate_import_declarations, fixId, ts.Diagnostics.Split_all_invalid_type_only_imports)];
                    }
                },
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, error) {
                        splitTypeOnlyImport(changes, getImportDeclaration(context.sourceFile, error), context);
                    });
                },
            });
            function getImportDeclaration(sourceFile, span) {
                return ts.findAncestor(ts.getTokenAtPosition(sourceFile, span.start), ts.isImportDeclaration);
            }
            function splitTypeOnlyImport(changes, importDeclaration, context) {
                if (!importDeclaration) {
                    return;
                }
                var importClause = ts.Debug.checkDefined(importDeclaration.importClause);
                changes.replaceNode(context.sourceFile, importDeclaration, ts.factory.updateImportDeclaration(importDeclaration, importDeclaration.decorators, importDeclaration.modifiers, ts.factory.updateImportClause(importClause, importClause.isTypeOnly, importClause.name, /*namedBindings*/ undefined), importDeclaration.moduleSpecifier));
                changes.insertNodeAfter(context.sourceFile, importDeclaration, ts.factory.createImportDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, ts.factory.updateImportClause(importClause, importClause.isTypeOnly, /*name*/ undefined, importClause.namedBindings), importDeclaration.moduleSpecifier));
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixConvertConstToLet";
            var errorCodes = [ts.Diagnostics.Cannot_assign_to_0_because_it_is_a_constant.code];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile, span = context.span, program = context.program;
                    var range = getConstTokenRange(sourceFile, span.start, program);
                    if (range === undefined)
                        return;
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, range); });
                    return [codefix.createCodeFixAction(fixId, changes, ts.Diagnostics.Convert_const_to_let, fixId, ts.Diagnostics.Convert_const_to_let)];
                },
                fixIds: [fixId]
            });
            function getConstTokenRange(sourceFile, pos, program) {
                var _a;
                var checker = program.getTypeChecker();
                var symbol = checker.getSymbolAtLocation(ts.getTokenAtPosition(sourceFile, pos));
                var declaration = ts.tryCast((_a = symbol === null || symbol === void 0 ? void 0 : symbol.valueDeclaration) === null || _a === void 0 ? void 0 : _a.parent, ts.isVariableDeclarationList);
                if (declaration === undefined)
                    return;
                var constToken = ts.findChildOfKind(declaration, 85 /* ConstKeyword */, sourceFile);
                if (constToken === undefined)
                    return;
                return ts.createRange(constToken.pos, constToken.end);
            }
            function doChange(changes, sourceFile, range) {
                changes.replaceRangeWithText(sourceFile, range, "let");
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixId = "fixExpectedComma";
            var expectedErrorCode = ts.Diagnostics._0_expected.code;
            var errorCodes = [expectedErrorCode];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                getCodeActions: function (context) {
                    var sourceFile = context.sourceFile;
                    var info = getInfo(sourceFile, context.span.start, context.errorCode);
                    if (!info) {
                        return undefined;
                    }
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(t, sourceFile, info); });
                    return [codefix.createCodeFixAction(fixId, changes, [ts.Diagnostics.Change_0_to_1, ";", ","], fixId, [ts.Diagnostics.Change_0_to_1, ";", ","])];
                },
                fixIds: [fixId],
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) {
                        var info = getInfo(diag.file, diag.start, diag.code);
                        if (info)
                            doChange(changes, context.sourceFile, info);
                    });
                },
            });
            function getInfo(sourceFile, pos, _) {
                var node = ts.getTokenAtPosition(sourceFile, pos);
                return (node.kind === 26 /* SemicolonToken */ &&
                    node.parent &&
                    (ts.isObjectLiteralExpression(node.parent) ||
                        ts.isArrayLiteralExpression(node.parent))) ? { node: node } : undefined;
            }
            function doChange(changes, sourceFile, _a) {
                var node = _a.node;
                var newNode = ts.factory.createToken(27 /* CommaToken */);
                changes.replaceNode(sourceFile, node, newNode);
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var codefix;
        (function (codefix) {
            var fixName = "addVoidToPromise";
            var fixId = "addVoidToPromise";
            var errorCodes = [
                ts.Diagnostics.Expected_0_arguments_but_got_1_Did_you_forget_to_include_void_in_your_type_argument_to_Promise.code
            ];
            codefix.registerCodeFix({
                errorCodes: errorCodes,
                fixIds: [fixId],
                getCodeActions: function (context) {
                    var changes = ts.textChanges.ChangeTracker.with(context, function (t) { return makeChange(t, context.sourceFile, context.span, context.program); });
                    if (changes.length > 0) {
                        return [codefix.createCodeFixAction(fixName, changes, ts.Diagnostics.Add_void_to_Promise_resolved_without_a_value, fixId, ts.Diagnostics.Add_void_to_all_Promises_resolved_without_a_value)];
                    }
                },
                getAllCodeActions: function (context) {
                    return codefix.codeFixAll(context, errorCodes, function (changes, diag) { return makeChange(changes, diag.file, diag, context.program, new ts.Set()); });
                }
            });
            function makeChange(changes, sourceFile, span, program, seen) {
                var node = ts.getTokenAtPosition(sourceFile, span.start);
                if (!ts.isIdentifier(node) || !ts.isCallExpression(node.parent) || node.parent.expression !== node || node.parent.arguments.length !== 0)
                    return;
                var checker = program.getTypeChecker();
                var symbol = checker.getSymbolAtLocation(node);
                // decl should be `new Promise((<decl>) => {})`
                var decl = symbol === null || symbol === void 0 ? void 0 : symbol.valueDeclaration;
                if (!decl || !ts.isParameter(decl) || !ts.isNewExpression(decl.parent.parent))
                    return;
                // no need to make this change if we have already seen this parameter.
                if (seen === null || seen === void 0 ? void 0 : seen.has(decl))
                    return;
                seen === null || seen === void 0 ? void 0 : seen.add(decl);
                var typeArguments = getEffectiveTypeArguments(decl.parent.parent);
                if (ts.some(typeArguments)) {
                    // append ` | void` to type argument
                    var typeArgument = typeArguments[0];
                    var needsParens = !ts.isUnionTypeNode(typeArgument) && !ts.isParenthesizedTypeNode(typeArgument) &&
                        ts.isParenthesizedTypeNode(ts.factory.createUnionTypeNode([typeArgument, ts.factory.createKeywordTypeNode(114 /* VoidKeyword */)]).types[0]);
                    if (needsParens) {
                        changes.insertText(sourceFile, typeArgument.pos, "(");
                    }
                    changes.insertText(sourceFile, typeArgument.end, needsParens ? ") | void" : " | void");
                }
                else {
                    // make sure the Promise is type is untyped (i.e., `unknown`)
                    var signature = checker.getResolvedSignature(node.parent);
                    var parameter = signature === null || signature === void 0 ? void 0 : signature.parameters[0];
                    var parameterType = parameter && checker.getTypeOfSymbolAtLocation(parameter, decl.parent.parent);
                    if (ts.isInJSFile(decl)) {
                        if (!parameterType || parameterType.flags & 3 /* AnyOrUnknown */) {
                            // give the expression a type
                            changes.insertText(sourceFile, decl.parent.parent.end, ")");
                            changes.insertText(sourceFile, ts.skipTrivia(sourceFile.text, decl.parent.parent.pos), "/** @type {Promise<void>} */(");
                        }
                    }
                    else {
                        if (!parameterType || parameterType.flags & 2 /* Unknown */) {
                            // add `void` type argument
                            changes.insertText(sourceFile, decl.parent.parent.expression.end, "<void>");
                        }
                    }
                }
            }
            function getEffectiveTypeArguments(node) {
                var _a;
                if (ts.isInJSFile(node)) {
                    if (ts.isParenthesizedExpression(node.parent)) {
                        var jsDocType = (_a = ts.getJSDocTypeTag(node.parent)) === null || _a === void 0 ? void 0 : _a.typeExpression.type;
                        if (jsDocType && ts.isTypeReferenceNode(jsDocType) && ts.isIdentifier(jsDocType.typeName) && ts.idText(jsDocType.typeName) === "Promise") {
                            return jsDocType.typeArguments;
                        }
                    }
                }
                else {
                    return node.typeArguments;
                }
            }
        })(codefix = ts.codefix || (ts.codefix = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var refactorName = "Convert export";
            var defaultToNamedAction = {
                name: "Convert default export to named export",
                description: ts.Diagnostics.Convert_default_export_to_named_export.message,
                kind: "refactor.rewrite.export.named"
            };
            var namedToDefaultAction = {
                name: "Convert named export to default export",
                description: ts.Diagnostics.Convert_named_export_to_default_export.message,
                kind: "refactor.rewrite.export.default"
            };
            refactor.registerRefactor(refactorName, {
                kinds: [
                    defaultToNamedAction.kind,
                    namedToDefaultAction.kind
                ],
                getAvailableActions: function (context) {
                    var info = getInfo(context, context.triggerReason === "invoked");
                    if (!info)
                        return ts.emptyArray;
                    if (!refactor.isRefactorErrorInfo(info)) {
                        var action = info.wasDefault ? defaultToNamedAction : namedToDefaultAction;
                        return [{ name: refactorName, description: action.description, actions: [action] }];
                    }
                    if (context.preferences.provideRefactorNotApplicableReason) {
                        return [
                            { name: refactorName, description: ts.Diagnostics.Convert_default_export_to_named_export.message, actions: [
                                    __assign(__assign({}, defaultToNamedAction), { notApplicableReason: info.error }),
                                    __assign(__assign({}, namedToDefaultAction), { notApplicableReason: info.error }),
                                ] }
                        ];
                    }
                    return ts.emptyArray;
                },
                getEditsForAction: function (context, actionName) {
                    ts.Debug.assert(actionName === defaultToNamedAction.name || actionName === namedToDefaultAction.name, "Unexpected action name");
                    var info = getInfo(context);
                    ts.Debug.assert(info && !refactor.isRefactorErrorInfo(info), "Expected applicable refactor info");
                    var edits = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(context.file, context.program, info, t, context.cancellationToken); });
                    return { edits: edits, renameFilename: undefined, renameLocation: undefined };
                },
            });
            ;
            function getInfo(context, considerPartialSpans) {
                if (considerPartialSpans === void 0) {
                    considerPartialSpans = true;
                }
                var file = context.file, program = context.program;
                var span = ts.getRefactorContextSpan(context);
                var token = ts.getTokenAtPosition(file, span.start);
                var exportNode = !!(token.parent && ts.getSyntacticModifierFlags(token.parent) & 1 /* Export */) && considerPartialSpans ? token.parent : ts.getParentNodeInSpan(token, file, span);
                if (!exportNode || (!ts.isSourceFile(exportNode.parent) && !(ts.isModuleBlock(exportNode.parent) && ts.isAmbientModule(exportNode.parent.parent)))) {
                    return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_export_statement) };
                }
                var exportingModuleSymbol = ts.isSourceFile(exportNode.parent) ? exportNode.parent.symbol : exportNode.parent.parent.symbol;
                var flags = ts.getSyntacticModifierFlags(exportNode) || ((ts.isExportAssignment(exportNode) && !exportNode.isExportEquals) ? 513 /* ExportDefault */ : 0 /* None */);
                var wasDefault = !!(flags & 512 /* Default */);
                // If source file already has a default export, don't offer refactor.
                if (!(flags & 1 /* Export */) || !wasDefault && exportingModuleSymbol.exports.has("default" /* Default */)) {
                    return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.This_file_already_has_a_default_export) };
                }
                var checker = program.getTypeChecker();
                var noSymbolError = function (id) {
                    return (ts.isIdentifier(id) && checker.getSymbolAtLocation(id)) ? undefined
                        : { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Can_only_convert_named_export) };
                };
                switch (exportNode.kind) {
                    case 254 /* FunctionDeclaration */:
                    case 255 /* ClassDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                    case 259 /* ModuleDeclaration */: {
                        var node = exportNode;
                        if (!node.name)
                            return undefined;
                        return noSymbolError(node.name)
                            || { exportNode: node, exportName: node.name, wasDefault: wasDefault, exportingModuleSymbol: exportingModuleSymbol };
                    }
                    case 235 /* VariableStatement */: {
                        var vs = exportNode;
                        // Must be `export const x = something;`.
                        if (!(vs.declarationList.flags & 2 /* Const */) || vs.declarationList.declarations.length !== 1) {
                            return undefined;
                        }
                        var decl = ts.first(vs.declarationList.declarations);
                        if (!decl.initializer)
                            return undefined;
                        ts.Debug.assert(!wasDefault, "Can't have a default flag here");
                        return noSymbolError(decl.name)
                            || { exportNode: vs, exportName: decl.name, wasDefault: wasDefault, exportingModuleSymbol: exportingModuleSymbol };
                    }
                    case 269 /* ExportAssignment */: {
                        var node = exportNode;
                        if (node.isExportEquals)
                            return undefined;
                        return noSymbolError(node.expression)
                            || { exportNode: node, exportName: node.expression, wasDefault: wasDefault, exportingModuleSymbol: exportingModuleSymbol };
                    }
                    default:
                        return undefined;
                }
            }
            function doChange(exportingSourceFile, program, info, changes, cancellationToken) {
                changeExport(exportingSourceFile, info, changes, program.getTypeChecker());
                changeImports(program, info, changes, cancellationToken);
            }
            function changeExport(exportingSourceFile, _a, changes, checker) {
                var wasDefault = _a.wasDefault, exportNode = _a.exportNode, exportName = _a.exportName;
                if (wasDefault) {
                    if (ts.isExportAssignment(exportNode) && !exportNode.isExportEquals) {
                        var exp = exportNode.expression;
                        var spec = makeExportSpecifier(exp.text, exp.text);
                        changes.replaceNode(exportingSourceFile, exportNode, ts.factory.createExportDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*isTypeOnly*/ false, ts.factory.createNamedExports([spec])));
                    }
                    else {
                        changes.delete(exportingSourceFile, ts.Debug.checkDefined(ts.findModifier(exportNode, 88 /* DefaultKeyword */), "Should find a default keyword in modifier list"));
                    }
                }
                else {
                    var exportKeyword = ts.Debug.checkDefined(ts.findModifier(exportNode, 93 /* ExportKeyword */), "Should find an export keyword in modifier list");
                    switch (exportNode.kind) {
                        case 254 /* FunctionDeclaration */:
                        case 255 /* ClassDeclaration */:
                        case 256 /* InterfaceDeclaration */:
                            changes.insertNodeAfter(exportingSourceFile, exportKeyword, ts.factory.createToken(88 /* DefaultKeyword */));
                            break;
                        case 235 /* VariableStatement */:
                            // If 'x' isn't used in this file and doesn't have type definition, `export const x = 0;` --> `export default 0;`
                            var decl = ts.first(exportNode.declarationList.declarations);
                            if (!ts.FindAllReferences.Core.isSymbolReferencedInFile(exportName, checker, exportingSourceFile) && !decl.type) {
                                // We checked in `getInfo` that an initializer exists.
                                changes.replaceNode(exportingSourceFile, exportNode, ts.factory.createExportDefault(ts.Debug.checkDefined(decl.initializer, "Initializer was previously known to be present")));
                                break;
                            }
                        // falls through
                        case 258 /* EnumDeclaration */:
                        case 257 /* TypeAliasDeclaration */:
                        case 259 /* ModuleDeclaration */:
                            // `export type T = number;` -> `type T = number; export default T;`
                            changes.deleteModifier(exportingSourceFile, exportKeyword);
                            changes.insertNodeAfter(exportingSourceFile, exportNode, ts.factory.createExportDefault(ts.factory.createIdentifier(exportName.text)));
                            break;
                        default:
                            ts.Debug.fail("Unexpected exportNode kind " + exportNode.kind);
                    }
                }
            }
            function changeImports(program, _a, changes, cancellationToken) {
                var wasDefault = _a.wasDefault, exportName = _a.exportName, exportingModuleSymbol = _a.exportingModuleSymbol;
                var checker = program.getTypeChecker();
                var exportSymbol = ts.Debug.checkDefined(checker.getSymbolAtLocation(exportName), "Export name should resolve to a symbol");
                ts.FindAllReferences.Core.eachExportReference(program.getSourceFiles(), checker, cancellationToken, exportSymbol, exportingModuleSymbol, exportName.text, wasDefault, function (ref) {
                    var importingSourceFile = ref.getSourceFile();
                    if (wasDefault) {
                        changeDefaultToNamedImport(importingSourceFile, ref, changes, exportName.text);
                    }
                    else {
                        changeNamedToDefaultImport(importingSourceFile, ref, changes);
                    }
                });
            }
            function changeDefaultToNamedImport(importingSourceFile, ref, changes, exportName) {
                var parent = ref.parent;
                switch (parent.kind) {
                    case 204 /* PropertyAccessExpression */:
                        // `a.default` --> `a.foo`
                        changes.replaceNode(importingSourceFile, ref, ts.factory.createIdentifier(exportName));
                        break;
                    case 268 /* ImportSpecifier */:
                    case 273 /* ExportSpecifier */: {
                        var spec = parent;
                        // `default as foo` --> `foo`, `default as bar` --> `foo as bar`
                        changes.replaceNode(importingSourceFile, spec, makeImportSpecifier(exportName, spec.name.text));
                        break;
                    }
                    case 265 /* ImportClause */: {
                        var clause = parent;
                        ts.Debug.assert(clause.name === ref, "Import clause name should match provided ref");
                        var spec = makeImportSpecifier(exportName, ref.text);
                        var namedBindings = clause.namedBindings;
                        if (!namedBindings) {
                            // `import foo from "./a";` --> `import { foo } from "./a";`
                            changes.replaceNode(importingSourceFile, ref, ts.factory.createNamedImports([spec]));
                        }
                        else if (namedBindings.kind === 266 /* NamespaceImport */) {
                            // `import foo, * as a from "./a";` --> `import * as a from ".a/"; import { foo } from "./a";`
                            changes.deleteRange(importingSourceFile, { pos: ref.getStart(importingSourceFile), end: namedBindings.getStart(importingSourceFile) });
                            var quotePreference = ts.isStringLiteral(clause.parent.moduleSpecifier) ? ts.quotePreferenceFromString(clause.parent.moduleSpecifier, importingSourceFile) : 1 /* Double */;
                            var newImport = ts.makeImport(/*default*/ undefined, [makeImportSpecifier(exportName, ref.text)], clause.parent.moduleSpecifier, quotePreference);
                            changes.insertNodeAfter(importingSourceFile, clause.parent, newImport);
                        }
                        else {
                            // `import foo, { bar } from "./a"` --> `import { bar, foo } from "./a";`
                            changes.delete(importingSourceFile, ref);
                            changes.insertNodeAtEndOfList(importingSourceFile, namedBindings.elements, spec);
                        }
                        break;
                    }
                    default:
                        ts.Debug.failBadSyntaxKind(parent);
                }
            }
            function changeNamedToDefaultImport(importingSourceFile, ref, changes) {
                var parent = ref.parent;
                switch (parent.kind) {
                    case 204 /* PropertyAccessExpression */:
                        // `a.foo` --> `a.default`
                        changes.replaceNode(importingSourceFile, ref, ts.factory.createIdentifier("default"));
                        break;
                    case 268 /* ImportSpecifier */: {
                        // `import { foo } from "./a";` --> `import foo from "./a";`
                        // `import { foo as bar } from "./a";` --> `import bar from "./a";`
                        var defaultImport = ts.factory.createIdentifier(parent.name.text);
                        if (parent.parent.elements.length === 1) {
                            changes.replaceNode(importingSourceFile, parent.parent, defaultImport);
                        }
                        else {
                            changes.delete(importingSourceFile, parent);
                            changes.insertNodeBefore(importingSourceFile, parent.parent, defaultImport);
                        }
                        break;
                    }
                    case 273 /* ExportSpecifier */: {
                        // `export { foo } from "./a";` --> `export { default as foo } from "./a";`
                        // `export { foo as bar } from "./a";` --> `export { default as bar } from "./a";`
                        // `export { foo as default } from "./a";` --> `export { default } from "./a";`
                        // (Because `export foo from "./a";` isn't valid syntax.)
                        changes.replaceNode(importingSourceFile, parent, makeExportSpecifier("default", parent.name.text));
                        break;
                    }
                    default:
                        ts.Debug.assertNever(parent, "Unexpected parent kind " + parent.kind);
                }
            }
            function makeImportSpecifier(propertyName, name) {
                return ts.factory.createImportSpecifier(propertyName === name ? undefined : ts.factory.createIdentifier(propertyName), ts.factory.createIdentifier(name));
            }
            function makeExportSpecifier(propertyName, name) {
                return ts.factory.createExportSpecifier(propertyName === name ? undefined : ts.factory.createIdentifier(propertyName), ts.factory.createIdentifier(name));
            }
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var refactorName = "Convert import";
            var namespaceToNamedAction = {
                name: "Convert namespace import to named imports",
                description: ts.Diagnostics.Convert_namespace_import_to_named_imports.message,
                kind: "refactor.rewrite.import.named",
            };
            var namedToNamespaceAction = {
                name: "Convert named imports to namespace import",
                description: ts.Diagnostics.Convert_named_imports_to_namespace_import.message,
                kind: "refactor.rewrite.import.namespace",
            };
            refactor.registerRefactor(refactorName, {
                kinds: [
                    namespaceToNamedAction.kind,
                    namedToNamespaceAction.kind
                ],
                getAvailableActions: function (context) {
                    var info = getImportToConvert(context, context.triggerReason === "invoked");
                    if (!info)
                        return ts.emptyArray;
                    if (!refactor.isRefactorErrorInfo(info)) {
                        var namespaceImport = info.kind === 266 /* NamespaceImport */;
                        var action = namespaceImport ? namespaceToNamedAction : namedToNamespaceAction;
                        return [{ name: refactorName, description: action.description, actions: [action] }];
                    }
                    if (context.preferences.provideRefactorNotApplicableReason) {
                        return [
                            { name: refactorName, description: namespaceToNamedAction.description,
                                actions: [__assign(__assign({}, namespaceToNamedAction), { notApplicableReason: info.error })] },
                            { name: refactorName, description: namedToNamespaceAction.description,
                                actions: [__assign(__assign({}, namedToNamespaceAction), { notApplicableReason: info.error })] }
                        ];
                    }
                    return ts.emptyArray;
                },
                getEditsForAction: function (context, actionName) {
                    ts.Debug.assert(actionName === namespaceToNamedAction.name || actionName === namedToNamespaceAction.name, "Unexpected action name");
                    var info = getImportToConvert(context);
                    ts.Debug.assert(info && !refactor.isRefactorErrorInfo(info), "Expected applicable refactor info");
                    var edits = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(context.file, context.program, t, info); });
                    return { edits: edits, renameFilename: undefined, renameLocation: undefined };
                }
            });
            // Can convert imports of the form `import * as m from "m";` or `import d, { x, y } from "m";`.
            function getImportToConvert(context, considerPartialSpans) {
                if (considerPartialSpans === void 0) {
                    considerPartialSpans = true;
                }
                var file = context.file;
                var span = ts.getRefactorContextSpan(context);
                var token = ts.getTokenAtPosition(file, span.start);
                var importDecl = considerPartialSpans ? ts.findAncestor(token, ts.isImportDeclaration) : ts.getParentNodeInSpan(token, file, span);
                if (!importDecl || !ts.isImportDeclaration(importDecl))
                    return { error: "Selection is not an import declaration." };
                if (importDecl.getEnd() < span.start + span.length)
                    return undefined;
                var importClause = importDecl.importClause;
                if (!importClause) {
                    return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_import_clause) };
                }
                if (!importClause.namedBindings) {
                    return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_namespace_import_or_named_imports) };
                }
                return importClause.namedBindings;
            }
            function doChange(sourceFile, program, changes, toConvert) {
                var checker = program.getTypeChecker();
                if (toConvert.kind === 266 /* NamespaceImport */) {
                    doChangeNamespaceToNamed(sourceFile, checker, changes, toConvert, ts.getAllowSyntheticDefaultImports(program.getCompilerOptions()));
                }
                else {
                    doChangeNamedToNamespace(sourceFile, checker, changes, toConvert);
                }
            }
            function doChangeNamespaceToNamed(sourceFile, checker, changes, toConvert, allowSyntheticDefaultImports) {
                var usedAsNamespaceOrDefault = false;
                var nodesToReplace = [];
                var conflictingNames = new ts.Map();
                ts.FindAllReferences.Core.eachSymbolReferenceInFile(toConvert.name, checker, sourceFile, function (id) {
                    if (!ts.isPropertyAccessOrQualifiedName(id.parent)) {
                        usedAsNamespaceOrDefault = true;
                    }
                    else {
                        var exportName = getRightOfPropertyAccessOrQualifiedName(id.parent).text;
                        if (checker.resolveName(exportName, id, 67108863 /* All */, /*excludeGlobals*/ true)) {
                            conflictingNames.set(exportName, true);
                        }
                        ts.Debug.assert(getLeftOfPropertyAccessOrQualifiedName(id.parent) === id, "Parent expression should match id");
                        nodesToReplace.push(id.parent);
                    }
                });
                // We may need to change `mod.x` to `_x` to avoid a name conflict.
                var exportNameToImportName = new ts.Map();
                for (var _i = 0, nodesToReplace_1 = nodesToReplace; _i < nodesToReplace_1.length; _i++) {
                    var propertyAccessOrQualifiedName = nodesToReplace_1[_i];
                    var exportName = getRightOfPropertyAccessOrQualifiedName(propertyAccessOrQualifiedName).text;
                    var importName = exportNameToImportName.get(exportName);
                    if (importName === undefined) {
                        exportNameToImportName.set(exportName, importName = conflictingNames.has(exportName) ? ts.getUniqueName(exportName, sourceFile) : exportName);
                    }
                    changes.replaceNode(sourceFile, propertyAccessOrQualifiedName, ts.factory.createIdentifier(importName));
                }
                var importSpecifiers = [];
                exportNameToImportName.forEach(function (name, propertyName) {
                    importSpecifiers.push(ts.factory.createImportSpecifier(name === propertyName ? undefined : ts.factory.createIdentifier(propertyName), ts.factory.createIdentifier(name)));
                });
                var importDecl = toConvert.parent.parent;
                if (usedAsNamespaceOrDefault && !allowSyntheticDefaultImports) {
                    // Need to leave the namespace import alone
                    changes.insertNodeAfter(sourceFile, importDecl, updateImport(importDecl, /*defaultImportName*/ undefined, importSpecifiers));
                }
                else {
                    changes.replaceNode(sourceFile, importDecl, updateImport(importDecl, usedAsNamespaceOrDefault ? ts.factory.createIdentifier(toConvert.name.text) : undefined, importSpecifiers));
                }
            }
            function getRightOfPropertyAccessOrQualifiedName(propertyAccessOrQualifiedName) {
                return ts.isPropertyAccessExpression(propertyAccessOrQualifiedName) ? propertyAccessOrQualifiedName.name : propertyAccessOrQualifiedName.right;
            }
            function getLeftOfPropertyAccessOrQualifiedName(propertyAccessOrQualifiedName) {
                return ts.isPropertyAccessExpression(propertyAccessOrQualifiedName) ? propertyAccessOrQualifiedName.expression : propertyAccessOrQualifiedName.left;
            }
            function doChangeNamedToNamespace(sourceFile, checker, changes, toConvert) {
                var importDecl = toConvert.parent.parent;
                var moduleSpecifier = importDecl.moduleSpecifier;
                var toConvertSymbols = new ts.Set();
                toConvert.elements.forEach(function (namedImport) {
                    var symbol = checker.getSymbolAtLocation(namedImport.name);
                    if (symbol) {
                        toConvertSymbols.add(symbol);
                    }
                });
                var preferredName = moduleSpecifier && ts.isStringLiteral(moduleSpecifier) ? ts.codefix.moduleSpecifierToValidIdentifier(moduleSpecifier.text, 99 /* ESNext */) : "module";
                function hasNamespaceNameConflict(namedImport) {
                    // We need to check if the preferred namespace name (`preferredName`) we'd like to use in the refactored code will present a name conflict.
                    // A name conflict means that, in a scope where we would like to use the preferred namespace name, there already exists a symbol with that name in that scope.
                    // We are going to use the namespace name in the scopes the named imports being refactored are referenced,
                    // so we look for conflicts by looking at every reference to those named imports.
                    return !!ts.FindAllReferences.Core.eachSymbolReferenceInFile(namedImport.name, checker, sourceFile, function (id) {
                        var symbol = checker.resolveName(preferredName, id, 67108863 /* All */, /*excludeGlobals*/ true);
                        if (symbol) { // There already is a symbol with the same name as the preferred namespace name.
                            if (toConvertSymbols.has(symbol)) { // `preferredName` resolves to a symbol for one of the named import references we are going to transform into namespace import references...
                                return ts.isExportSpecifier(id.parent); // ...but if this reference is an export specifier, it will not be transformed, so it is a conflict; otherwise, it will be renamed and is not a conflict.
                            }
                            return true; // `preferredName` resolves to any other symbol, which will be present in the refactored code and so poses a name conflict.
                        }
                        return false; // There is no symbol with the same name as the preferred namespace name, so no conflict.
                    });
                }
                var namespaceNameConflicts = toConvert.elements.some(hasNamespaceNameConflict);
                var namespaceImportName = namespaceNameConflicts ? ts.getUniqueName(preferredName, sourceFile) : preferredName;
                // Imports that need to be kept as named imports in the refactored code, to avoid changing the semantics.
                // More specifically, those are named imports that appear in named exports in the original code, e.g. `a` in `import { a } from "m"; export { a }`.
                var neededNamedImports = new ts.Set();
                var _loop_16 = function (element) {
                    var propertyName = (element.propertyName || element.name).text;
                    ts.FindAllReferences.Core.eachSymbolReferenceInFile(element.name, checker, sourceFile, function (id) {
                        var access = ts.factory.createPropertyAccessExpression(ts.factory.createIdentifier(namespaceImportName), propertyName);
                        if (ts.isShorthandPropertyAssignment(id.parent)) {
                            changes.replaceNode(sourceFile, id.parent, ts.factory.createPropertyAssignment(id.text, access));
                        }
                        else if (ts.isExportSpecifier(id.parent)) {
                            neededNamedImports.add(element);
                        }
                        else {
                            changes.replaceNode(sourceFile, id, access);
                        }
                    });
                };
                for (var _i = 0, _a = toConvert.elements; _i < _a.length; _i++) {
                    var element = _a[_i];
                    _loop_16(element);
                }
                changes.replaceNode(sourceFile, toConvert, ts.factory.createNamespaceImport(ts.factory.createIdentifier(namespaceImportName)));
                if (neededNamedImports.size) {
                    var newNamedImports = ts.arrayFrom(neededNamedImports.values()).map(function (element) {
                        return ts.factory.createImportSpecifier(element.propertyName && ts.factory.createIdentifier(element.propertyName.text), ts.factory.createIdentifier(element.name.text));
                    });
                    changes.insertNodeAfter(sourceFile, toConvert.parent.parent, updateImport(importDecl, /*defaultImportName*/ undefined, newNamedImports));
                }
            }
            function updateImport(old, defaultImportName, elements) {
                return ts.factory.createImportDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, ts.factory.createImportClause(/*isTypeOnly*/ false, defaultImportName, elements && elements.length ? ts.factory.createNamedImports(elements) : undefined), old.moduleSpecifier);
            }
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var convertToOptionalChainExpression;
            (function (convertToOptionalChainExpression) {
                var refactorName = "Convert to optional chain expression";
                var convertToOptionalChainExpressionMessage = ts.getLocaleSpecificMessage(ts.Diagnostics.Convert_to_optional_chain_expression);
                var toOptionalChainAction = {
                    name: refactorName,
                    description: convertToOptionalChainExpressionMessage,
                    kind: "refactor.rewrite.expression.optionalChain",
                };
                refactor.registerRefactor(refactorName, {
                    kinds: [toOptionalChainAction.kind],
                    getAvailableActions: getAvailableActions,
                    getEditsForAction: getEditsForAction
                });
                function getAvailableActions(context) {
                    var info = getInfo(context, context.triggerReason === "invoked");
                    if (!info)
                        return ts.emptyArray;
                    if (!refactor.isRefactorErrorInfo(info)) {
                        return [{
                                name: refactorName,
                                description: convertToOptionalChainExpressionMessage,
                                actions: [toOptionalChainAction],
                            }];
                    }
                    if (context.preferences.provideRefactorNotApplicableReason) {
                        return [{
                                name: refactorName,
                                description: convertToOptionalChainExpressionMessage,
                                actions: [__assign(__assign({}, toOptionalChainAction), { notApplicableReason: info.error })],
                            }];
                    }
                    return ts.emptyArray;
                }
                function getEditsForAction(context, actionName) {
                    var info = getInfo(context);
                    ts.Debug.assert(info && !refactor.isRefactorErrorInfo(info), "Expected applicable refactor info");
                    var edits = ts.textChanges.ChangeTracker.with(context, function (t) {
                        return doChange(context.file, context.program.getTypeChecker(), t, info, actionName);
                    });
                    return { edits: edits, renameFilename: undefined, renameLocation: undefined };
                }
                ;
                function isValidExpression(node) {
                    return ts.isBinaryExpression(node) || ts.isConditionalExpression(node);
                }
                function isValidStatement(node) {
                    return ts.isExpressionStatement(node) || ts.isReturnStatement(node) || ts.isVariableStatement(node);
                }
                function isValidExpressionOrStatement(node) {
                    return isValidExpression(node) || isValidStatement(node);
                }
                function getInfo(context, considerEmptySpans) {
                    if (considerEmptySpans === void 0) {
                        considerEmptySpans = true;
                    }
                    var file = context.file, program = context.program;
                    var span = ts.getRefactorContextSpan(context);
                    var forEmptySpan = span.length === 0;
                    if (forEmptySpan && !considerEmptySpans)
                        return undefined;
                    // selecting fo[|o && foo.ba|]r should be valid, so adjust span to fit start and end tokens
                    var startToken = ts.getTokenAtPosition(file, span.start);
                    var endToken = ts.findTokenOnLeftOfPosition(file, span.start + span.length);
                    var adjustedSpan = ts.createTextSpanFromBounds(startToken.pos, endToken && endToken.end >= startToken.pos ? endToken.getEnd() : startToken.getEnd());
                    var parent = forEmptySpan ? getValidParentNodeOfEmptySpan(startToken) : getValidParentNodeContainingSpan(startToken, adjustedSpan);
                    var expression = parent && isValidExpressionOrStatement(parent) ? getExpression(parent) : undefined;
                    if (!expression)
                        return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_convertible_access_expression) };
                    var checker = program.getTypeChecker();
                    return ts.isConditionalExpression(expression) ? getConditionalInfo(expression, checker) : getBinaryInfo(expression);
                }
                function getConditionalInfo(expression, checker) {
                    var condition = expression.condition;
                    var finalExpression = getFinalExpressionInChain(expression.whenTrue);
                    if (!finalExpression || checker.isNullableType(checker.getTypeAtLocation(finalExpression))) {
                        return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_convertible_access_expression) };
                    }
                    if ((ts.isPropertyAccessExpression(condition) || ts.isIdentifier(condition))
                        && getMatchingStart(condition, finalExpression.expression)) {
                        return { finalExpression: finalExpression, occurrences: [condition], expression: expression };
                    }
                    else if (ts.isBinaryExpression(condition)) {
                        var occurrences = getOccurrencesInExpression(finalExpression.expression, condition);
                        return occurrences ? { finalExpression: finalExpression, occurrences: occurrences, expression: expression } :
                            { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_matching_access_expressions) };
                    }
                }
                function getBinaryInfo(expression) {
                    if (expression.operatorToken.kind !== 55 /* AmpersandAmpersandToken */) {
                        return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Can_only_convert_logical_AND_access_chains) };
                    }
                    ;
                    var finalExpression = getFinalExpressionInChain(expression.right);
                    if (!finalExpression)
                        return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_convertible_access_expression) };
                    var occurrences = getOccurrencesInExpression(finalExpression.expression, expression.left);
                    return occurrences ? { finalExpression: finalExpression, occurrences: occurrences, expression: expression } :
                        { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_matching_access_expressions) };
                }
                /**
                 * Gets a list of property accesses that appear in matchTo and occur in sequence in expression.
                 */
                function getOccurrencesInExpression(matchTo, expression) {
                    var occurrences = [];
                    while (ts.isBinaryExpression(expression) && expression.operatorToken.kind === 55 /* AmpersandAmpersandToken */) {
                        var match = getMatchingStart(ts.skipParentheses(matchTo), ts.skipParentheses(expression.right));
                        if (!match) {
                            break;
                        }
                        occurrences.push(match);
                        matchTo = match;
                        expression = expression.left;
                    }
                    var finalMatch = getMatchingStart(matchTo, expression);
                    if (finalMatch) {
                        occurrences.push(finalMatch);
                    }
                    return occurrences.length > 0 ? occurrences : undefined;
                }
                /**
                 * Returns subchain if chain begins with subchain syntactically.
                 */
                function getMatchingStart(chain, subchain) {
                    if (!ts.isIdentifier(subchain) && !ts.isPropertyAccessExpression(subchain) && !ts.isElementAccessExpression(subchain)) {
                        return undefined;
                    }
                    return chainStartsWith(chain, subchain) ? subchain : undefined;
                }
                /**
                 * Returns true if chain begins with subchain syntactically.
                 */
                function chainStartsWith(chain, subchain) {
                    // skip until we find a matching identifier.
                    while (ts.isCallExpression(chain) || ts.isPropertyAccessExpression(chain) || ts.isElementAccessExpression(chain)) {
                        if (getTextOfChainNode(chain) === getTextOfChainNode(subchain))
                            break;
                        chain = chain.expression;
                    }
                    // check that the chains match at each access. Call chains in subchain are not valid.
                    while ((ts.isPropertyAccessExpression(chain) && ts.isPropertyAccessExpression(subchain)) ||
                        (ts.isElementAccessExpression(chain) && ts.isElementAccessExpression(subchain))) {
                        if (getTextOfChainNode(chain) !== getTextOfChainNode(subchain))
                            return false;
                        chain = chain.expression;
                        subchain = subchain.expression;
                    }
                    // check if we have reached a final identifier.
                    return ts.isIdentifier(chain) && ts.isIdentifier(subchain) && chain.getText() === subchain.getText();
                }
                function getTextOfChainNode(node) {
                    if (ts.isIdentifier(node) || ts.isStringOrNumericLiteralLike(node)) {
                        return node.getText();
                    }
                    if (ts.isPropertyAccessExpression(node)) {
                        return getTextOfChainNode(node.name);
                    }
                    if (ts.isElementAccessExpression(node)) {
                        return getTextOfChainNode(node.argumentExpression);
                    }
                    return undefined;
                }
                /**
                 * Find the least ancestor of the input node that is a valid type for extraction and contains the input span.
                 */
                function getValidParentNodeContainingSpan(node, span) {
                    while (node.parent) {
                        if (isValidExpressionOrStatement(node) && span.length !== 0 && node.end >= span.start + span.length) {
                            return node;
                        }
                        node = node.parent;
                    }
                    return undefined;
                }
                /**
                 * Finds an ancestor of the input node that is a valid type for extraction, skipping subexpressions.
                 */
                function getValidParentNodeOfEmptySpan(node) {
                    while (node.parent) {
                        if (isValidExpressionOrStatement(node) && !isValidExpressionOrStatement(node.parent)) {
                            return node;
                        }
                        node = node.parent;
                    }
                    return undefined;
                }
                /**
                 * Gets an expression of valid extraction type from a valid statement or expression.
                 */
                function getExpression(node) {
                    if (isValidExpression(node)) {
                        return node;
                    }
                    if (ts.isVariableStatement(node)) {
                        var variable = ts.getSingleVariableOfVariableStatement(node);
                        var initializer = variable === null || variable === void 0 ? void 0 : variable.initializer;
                        return initializer && isValidExpression(initializer) ? initializer : undefined;
                    }
                    return node.expression && isValidExpression(node.expression) ? node.expression : undefined;
                }
                /**
                 * Gets a property access expression which may be nested inside of a binary expression. The final
                 * expression in an && chain will occur as the right child of the parent binary expression, unless
                 * it is followed by a different binary operator.
                 * @param node the right child of a binary expression or a call expression.
                 */
                function getFinalExpressionInChain(node) {
                    // foo && |foo.bar === 1|; - here the right child of the && binary expression is another binary expression.
                    // the rightmost member of the && chain should be the leftmost child of that expression.
                    node = ts.skipParentheses(node);
                    if (ts.isBinaryExpression(node)) {
                        return getFinalExpressionInChain(node.left);
                    }
                    // foo && |foo.bar()()| - nested calls are treated like further accesses.
                    else if ((ts.isPropertyAccessExpression(node) || ts.isElementAccessExpression(node) || ts.isCallExpression(node)) && !ts.isOptionalChain(node)) {
                        return node;
                    }
                    return undefined;
                }
                /**
                 * Creates an access chain from toConvert with '?.' accesses at expressions appearing in occurrences.
                 */
                function convertOccurrences(checker, toConvert, occurrences) {
                    if (ts.isPropertyAccessExpression(toConvert) || ts.isElementAccessExpression(toConvert) || ts.isCallExpression(toConvert)) {
                        var chain = convertOccurrences(checker, toConvert.expression, occurrences);
                        var lastOccurrence = occurrences.length > 0 ? occurrences[occurrences.length - 1] : undefined;
                        var isOccurrence = (lastOccurrence === null || lastOccurrence === void 0 ? void 0 : lastOccurrence.getText()) === toConvert.expression.getText();
                        if (isOccurrence)
                            occurrences.pop();
                        if (ts.isCallExpression(toConvert)) {
                            return isOccurrence ?
                                ts.factory.createCallChain(chain, ts.factory.createToken(28 /* QuestionDotToken */), toConvert.typeArguments, toConvert.arguments) :
                                ts.factory.createCallChain(chain, toConvert.questionDotToken, toConvert.typeArguments, toConvert.arguments);
                        }
                        else if (ts.isPropertyAccessExpression(toConvert)) {
                            return isOccurrence ?
                                ts.factory.createPropertyAccessChain(chain, ts.factory.createToken(28 /* QuestionDotToken */), toConvert.name) :
                                ts.factory.createPropertyAccessChain(chain, toConvert.questionDotToken, toConvert.name);
                        }
                        else if (ts.isElementAccessExpression(toConvert)) {
                            return isOccurrence ?
                                ts.factory.createElementAccessChain(chain, ts.factory.createToken(28 /* QuestionDotToken */), toConvert.argumentExpression) :
                                ts.factory.createElementAccessChain(chain, toConvert.questionDotToken, toConvert.argumentExpression);
                        }
                    }
                    return toConvert;
                }
                function doChange(sourceFile, checker, changes, info, _actionName) {
                    var finalExpression = info.finalExpression, occurrences = info.occurrences, expression = info.expression;
                    var firstOccurrence = occurrences[occurrences.length - 1];
                    var convertedChain = convertOccurrences(checker, finalExpression, occurrences);
                    if (convertedChain && (ts.isPropertyAccessExpression(convertedChain) || ts.isElementAccessExpression(convertedChain) || ts.isCallExpression(convertedChain))) {
                        if (ts.isBinaryExpression(expression)) {
                            changes.replaceNodeRange(sourceFile, firstOccurrence, finalExpression, convertedChain);
                        }
                        else if (ts.isConditionalExpression(expression)) {
                            changes.replaceNode(sourceFile, expression, ts.factory.createBinaryExpression(convertedChain, ts.factory.createToken(60 /* QuestionQuestionToken */), expression.whenFalse));
                        }
                    }
                }
            })(convertToOptionalChainExpression = refactor.convertToOptionalChainExpression || (refactor.convertToOptionalChainExpression = {}));
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var addOrRemoveBracesToArrowFunction;
            (function (addOrRemoveBracesToArrowFunction) {
                var refactorName = "Convert overload list to single signature";
                var refactorDescription = ts.Diagnostics.Convert_overload_list_to_single_signature.message;
                var functionOverloadAction = {
                    name: refactorName,
                    description: refactorDescription,
                    kind: "refactor.rewrite.function.overloadList",
                };
                refactor.registerRefactor(refactorName, {
                    kinds: [functionOverloadAction.kind],
                    getEditsForAction: getEditsForAction,
                    getAvailableActions: getAvailableActions
                });
                function getAvailableActions(context) {
                    var file = context.file, startPosition = context.startPosition, program = context.program;
                    var info = getConvertableOverloadListAtPosition(file, startPosition, program);
                    if (!info)
                        return ts.emptyArray;
                    return [{
                            name: refactorName,
                            description: refactorDescription,
                            actions: [functionOverloadAction]
                        }];
                }
                function getEditsForAction(context) {
                    var file = context.file, startPosition = context.startPosition, program = context.program;
                    var signatureDecls = getConvertableOverloadListAtPosition(file, startPosition, program);
                    if (!signatureDecls)
                        return undefined;
                    var checker = program.getTypeChecker();
                    var lastDeclaration = signatureDecls[signatureDecls.length - 1];
                    var updated = lastDeclaration;
                    switch (lastDeclaration.kind) {
                        case 166 /* MethodSignature */: {
                            updated = ts.factory.updateMethodSignature(lastDeclaration, lastDeclaration.modifiers, lastDeclaration.name, lastDeclaration.questionToken, lastDeclaration.typeParameters, getNewParametersForCombinedSignature(signatureDecls), lastDeclaration.type);
                            break;
                        }
                        case 167 /* MethodDeclaration */: {
                            updated = ts.factory.updateMethodDeclaration(lastDeclaration, lastDeclaration.decorators, lastDeclaration.modifiers, lastDeclaration.asteriskToken, lastDeclaration.name, lastDeclaration.questionToken, lastDeclaration.typeParameters, getNewParametersForCombinedSignature(signatureDecls), lastDeclaration.type, lastDeclaration.body);
                            break;
                        }
                        case 172 /* CallSignature */: {
                            updated = ts.factory.updateCallSignature(lastDeclaration, lastDeclaration.typeParameters, getNewParametersForCombinedSignature(signatureDecls), lastDeclaration.type);
                            break;
                        }
                        case 169 /* Constructor */: {
                            updated = ts.factory.updateConstructorDeclaration(lastDeclaration, lastDeclaration.decorators, lastDeclaration.modifiers, getNewParametersForCombinedSignature(signatureDecls), lastDeclaration.body);
                            break;
                        }
                        case 173 /* ConstructSignature */: {
                            updated = ts.factory.updateConstructSignature(lastDeclaration, lastDeclaration.typeParameters, getNewParametersForCombinedSignature(signatureDecls), lastDeclaration.type);
                            break;
                        }
                        case 254 /* FunctionDeclaration */: {
                            updated = ts.factory.updateFunctionDeclaration(lastDeclaration, lastDeclaration.decorators, lastDeclaration.modifiers, lastDeclaration.asteriskToken, lastDeclaration.name, lastDeclaration.typeParameters, getNewParametersForCombinedSignature(signatureDecls), lastDeclaration.type, lastDeclaration.body);
                            break;
                        }
                        default: return ts.Debug.failBadSyntaxKind(lastDeclaration, "Unhandled signature kind in overload list conversion refactoring");
                    }
                    if (updated === lastDeclaration) {
                        return; // No edits to apply, do nothing
                    }
                    var edits = ts.textChanges.ChangeTracker.with(context, function (t) {
                        t.replaceNodeRange(file, signatureDecls[0], signatureDecls[signatureDecls.length - 1], updated);
                    });
                    return { renameFilename: undefined, renameLocation: undefined, edits: edits };
                    function getNewParametersForCombinedSignature(signatureDeclarations) {
                        var lastSig = signatureDeclarations[signatureDeclarations.length - 1];
                        if (ts.isFunctionLikeDeclaration(lastSig) && lastSig.body) {
                            // Trim away implementation signature arguments (they should already be compatible with overloads, but are likely less precise to guarantee compatability with the overloads)
                            signatureDeclarations = signatureDeclarations.slice(0, signatureDeclarations.length - 1);
                        }
                        return ts.factory.createNodeArray([
                            ts.factory.createParameterDeclaration(
                            /*decorators*/ undefined, 
                            /*modifiers*/ undefined, ts.factory.createToken(25 /* DotDotDotToken */), "args", 
                            /*questionToken*/ undefined, ts.factory.createUnionTypeNode(ts.map(signatureDeclarations, convertSignatureParametersToTuple)))
                        ]);
                    }
                    function convertSignatureParametersToTuple(decl) {
                        var members = ts.map(decl.parameters, convertParameterToNamedTupleMember);
                        return ts.setEmitFlags(ts.factory.createTupleTypeNode(members), ts.some(members, function (m) { return !!ts.length(ts.getSyntheticLeadingComments(m)); }) ? 0 /* None */ : 1 /* SingleLine */);
                    }
                    function convertParameterToNamedTupleMember(p) {
                        ts.Debug.assert(ts.isIdentifier(p.name)); // This is checked during refactoring applicability checking
                        var result = ts.setTextRange(ts.factory.createNamedTupleMember(p.dotDotDotToken, p.name, p.questionToken, p.type || ts.factory.createKeywordTypeNode(129 /* AnyKeyword */)), p);
                        var parameterDocComment = p.symbol && p.symbol.getDocumentationComment(checker);
                        if (parameterDocComment) {
                            var newComment = ts.displayPartsToString(parameterDocComment);
                            if (newComment.length) {
                                ts.setSyntheticLeadingComments(result, [{
                                        text: "*\n" + newComment.split("\n").map(function (c) { return " * " + c; }).join("\n") + "\n ",
                                        kind: 3 /* MultiLineCommentTrivia */,
                                        pos: -1,
                                        end: -1,
                                        hasTrailingNewLine: true,
                                        hasLeadingNewline: true,
                                    }]);
                            }
                        }
                        return result;
                    }
                }
                function isConvertableSignatureDeclaration(d) {
                    switch (d.kind) {
                        case 166 /* MethodSignature */:
                        case 167 /* MethodDeclaration */:
                        case 172 /* CallSignature */:
                        case 169 /* Constructor */:
                        case 173 /* ConstructSignature */:
                        case 254 /* FunctionDeclaration */:
                            return true;
                    }
                    return false;
                }
                function getConvertableOverloadListAtPosition(file, startPosition, program) {
                    var node = ts.getTokenAtPosition(file, startPosition);
                    var containingDecl = ts.findAncestor(node, isConvertableSignatureDeclaration);
                    if (!containingDecl) {
                        return;
                    }
                    var checker = program.getTypeChecker();
                    var signatureSymbol = containingDecl.symbol;
                    if (!signatureSymbol) {
                        return;
                    }
                    var decls = signatureSymbol.declarations;
                    if (ts.length(decls) <= 1) {
                        return;
                    }
                    if (!ts.every(decls, function (d) { return ts.getSourceFileOfNode(d) === file; })) {
                        return;
                    }
                    if (!isConvertableSignatureDeclaration(decls[0])) {
                        return;
                    }
                    var kindOne = decls[0].kind;
                    if (!ts.every(decls, function (d) { return d.kind === kindOne; })) {
                        return;
                    }
                    var signatureDecls = decls;
                    if (ts.some(signatureDecls, function (d) { return !!d.typeParameters || ts.some(d.parameters, function (p) { return !!p.decorators || !!p.modifiers || !ts.isIdentifier(p.name); }); })) {
                        return;
                    }
                    var signatures = ts.mapDefined(signatureDecls, function (d) { return checker.getSignatureFromDeclaration(d); });
                    if (ts.length(signatures) !== ts.length(decls)) {
                        return;
                    }
                    var returnOne = checker.getReturnTypeOfSignature(signatures[0]);
                    if (!ts.every(signatures, function (s) { return checker.getReturnTypeOfSignature(s) === returnOne; })) {
                        return;
                    }
                    return signatureDecls;
                }
            })(addOrRemoveBracesToArrowFunction = refactor.addOrRemoveBracesToArrowFunction || (refactor.addOrRemoveBracesToArrowFunction = {}));
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var extractSymbol;
            (function (extractSymbol) {
                var refactorName = "Extract Symbol";
                var extractConstantAction = {
                    name: "Extract Constant",
                    description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_constant),
                    kind: "refactor.extract.constant",
                };
                var extractFunctionAction = {
                    name: "Extract Function",
                    description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_function),
                    kind: "refactor.extract.function",
                };
                refactor.registerRefactor(refactorName, {
                    kinds: [
                        extractConstantAction.kind,
                        extractFunctionAction.kind
                    ],
                    getAvailableActions: getAvailableActions,
                    getEditsForAction: getEditsForAction
                });
                /**
                 * Compute the associated code actions
                 * Exported for tests.
                 */
                function getAvailableActions(context) {
                    var requestedRefactor = context.kind;
                    var rangeToExtract = getRangeToExtract(context.file, ts.getRefactorContextSpan(context), context.triggerReason === "invoked");
                    var targetRange = rangeToExtract.targetRange;
                    if (targetRange === undefined) {
                        if (!rangeToExtract.errors || rangeToExtract.errors.length === 0 || !context.preferences.provideRefactorNotApplicableReason) {
                            return ts.emptyArray;
                        }
                        var errors = [];
                        if (refactor.refactorKindBeginsWith(extractFunctionAction.kind, requestedRefactor)) {
                            errors.push({
                                name: refactorName,
                                description: extractFunctionAction.description,
                                actions: [__assign(__assign({}, extractFunctionAction), { notApplicableReason: getStringError(rangeToExtract.errors) })]
                            });
                        }
                        if (refactor.refactorKindBeginsWith(extractConstantAction.kind, requestedRefactor)) {
                            errors.push({
                                name: refactorName,
                                description: extractConstantAction.description,
                                actions: [__assign(__assign({}, extractConstantAction), { notApplicableReason: getStringError(rangeToExtract.errors) })]
                            });
                        }
                        return errors;
                    }
                    var extractions = getPossibleExtractions(targetRange, context);
                    if (extractions === undefined) {
                        // No extractions possible
                        return ts.emptyArray;
                    }
                    var functionActions = [];
                    var usedFunctionNames = new ts.Map();
                    var innermostErrorFunctionAction;
                    var constantActions = [];
                    var usedConstantNames = new ts.Map();
                    var innermostErrorConstantAction;
                    var i = 0;
                    for (var _i = 0, extractions_1 = extractions; _i < extractions_1.length; _i++) {
                        var _a = extractions_1[_i], functionExtraction = _a.functionExtraction, constantExtraction = _a.constantExtraction;
                        var description = functionExtraction.description;
                        if (refactor.refactorKindBeginsWith(extractFunctionAction.kind, requestedRefactor)) {
                            if (functionExtraction.errors.length === 0) {
                                // Don't issue refactorings with duplicated names.
                                // Scopes come back in "innermost first" order, so extractions will
                                // preferentially go into nearer scopes
                                if (!usedFunctionNames.has(description)) {
                                    usedFunctionNames.set(description, true);
                                    functionActions.push({
                                        description: description,
                                        name: "function_scope_" + i,
                                        kind: extractFunctionAction.kind
                                    });
                                }
                            }
                            else if (!innermostErrorFunctionAction) {
                                innermostErrorFunctionAction = {
                                    description: description,
                                    name: "function_scope_" + i,
                                    notApplicableReason: getStringError(functionExtraction.errors),
                                    kind: extractFunctionAction.kind
                                };
                            }
                        }
                        if (refactor.refactorKindBeginsWith(extractConstantAction.kind, requestedRefactor)) {
                            if (constantExtraction.errors.length === 0) {
                                // Don't issue refactorings with duplicated names.
                                // Scopes come back in "innermost first" order, so extractions will
                                // preferentially go into nearer scopes
                                var description_1 = constantExtraction.description;
                                if (!usedConstantNames.has(description_1)) {
                                    usedConstantNames.set(description_1, true);
                                    constantActions.push({
                                        description: description_1,
                                        name: "constant_scope_" + i,
                                        kind: extractConstantAction.kind
                                    });
                                }
                            }
                            else if (!innermostErrorConstantAction) {
                                innermostErrorConstantAction = {
                                    description: description,
                                    name: "constant_scope_" + i,
                                    notApplicableReason: getStringError(constantExtraction.errors),
                                    kind: extractConstantAction.kind
                                };
                            }
                        }
                        // *do* increment i anyway because we'll look for the i-th scope
                        // later when actually doing the refactoring if the user requests it
                        i++;
                    }
                    var infos = [];
                    if (functionActions.length) {
                        infos.push({
                            name: refactorName,
                            description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_function),
                            actions: functionActions,
                        });
                    }
                    else if (context.preferences.provideRefactorNotApplicableReason && innermostErrorFunctionAction) {
                        infos.push({
                            name: refactorName,
                            description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_function),
                            actions: [innermostErrorFunctionAction]
                        });
                    }
                    if (constantActions.length) {
                        infos.push({
                            name: refactorName,
                            description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_constant),
                            actions: constantActions
                        });
                    }
                    else if (context.preferences.provideRefactorNotApplicableReason && innermostErrorConstantAction) {
                        infos.push({
                            name: refactorName,
                            description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_constant),
                            actions: [innermostErrorConstantAction]
                        });
                    }
                    return infos.length ? infos : ts.emptyArray;
                    function getStringError(errors) {
                        var error = errors[0].messageText;
                        if (typeof error !== "string") {
                            error = error.messageText;
                        }
                        return error;
                    }
                }
                extractSymbol.getAvailableActions = getAvailableActions;
                /* Exported for tests */
                function getEditsForAction(context, actionName) {
                    var rangeToExtract = getRangeToExtract(context.file, ts.getRefactorContextSpan(context));
                    var targetRange = rangeToExtract.targetRange; // TODO:GH#18217
                    var parsedFunctionIndexMatch = /^function_scope_(\d+)$/.exec(actionName);
                    if (parsedFunctionIndexMatch) {
                        var index = +parsedFunctionIndexMatch[1];
                        ts.Debug.assert(isFinite(index), "Expected to parse a finite number from the function scope index");
                        return getFunctionExtractionAtIndex(targetRange, context, index);
                    }
                    var parsedConstantIndexMatch = /^constant_scope_(\d+)$/.exec(actionName);
                    if (parsedConstantIndexMatch) {
                        var index = +parsedConstantIndexMatch[1];
                        ts.Debug.assert(isFinite(index), "Expected to parse a finite number from the constant scope index");
                        return getConstantExtractionAtIndex(targetRange, context, index);
                    }
                    ts.Debug.fail("Unrecognized action name");
                }
                extractSymbol.getEditsForAction = getEditsForAction;
                // Move these into diagnostic messages if they become user-facing
                var Messages;
                (function (Messages) {
                    function createMessage(message) {
                        return { message: message, code: 0, category: ts.DiagnosticCategory.Message, key: message };
                    }
                    Messages.cannotExtractRange = createMessage("Cannot extract range.");
                    Messages.cannotExtractImport = createMessage("Cannot extract import statement.");
                    Messages.cannotExtractSuper = createMessage("Cannot extract super call.");
                    Messages.cannotExtractJSDoc = createMessage("Cannot extract JSDoc.");
                    Messages.cannotExtractEmpty = createMessage("Cannot extract empty range.");
                    Messages.expressionExpected = createMessage("expression expected.");
                    Messages.uselessConstantType = createMessage("No reason to extract constant of type.");
                    Messages.statementOrExpressionExpected = createMessage("Statement or expression expected.");
                    Messages.cannotExtractRangeContainingConditionalBreakOrContinueStatements = createMessage("Cannot extract range containing conditional break or continue statements.");
                    Messages.cannotExtractRangeContainingConditionalReturnStatement = createMessage("Cannot extract range containing conditional return statement.");
                    Messages.cannotExtractRangeContainingLabeledBreakOrContinueStatementWithTargetOutsideOfTheRange = createMessage("Cannot extract range containing labeled break or continue with target outside of the range.");
                    Messages.cannotExtractRangeThatContainsWritesToReferencesLocatedOutsideOfTheTargetRangeInGenerators = createMessage("Cannot extract range containing writes to references located outside of the target range in generators.");
                    Messages.typeWillNotBeVisibleInTheNewScope = createMessage("Type will not visible in the new scope.");
                    Messages.functionWillNotBeVisibleInTheNewScope = createMessage("Function will not visible in the new scope.");
                    Messages.cannotExtractIdentifier = createMessage("Select more than a single identifier.");
                    Messages.cannotExtractExportedEntity = createMessage("Cannot extract exported declaration");
                    Messages.cannotWriteInExpression = createMessage("Cannot write back side-effects when extracting an expression");
                    Messages.cannotExtractReadonlyPropertyInitializerOutsideConstructor = createMessage("Cannot move initialization of read-only class property outside of the constructor");
                    Messages.cannotExtractAmbientBlock = createMessage("Cannot extract code from ambient contexts");
                    Messages.cannotAccessVariablesFromNestedScopes = createMessage("Cannot access variables from nested scopes");
                    Messages.cannotExtractToJSClass = createMessage("Cannot extract constant to a class scope in JS");
                    Messages.cannotExtractToExpressionArrowFunction = createMessage("Cannot extract constant to an arrow function without a block");
                })(Messages = extractSymbol.Messages || (extractSymbol.Messages = {}));
                var RangeFacts;
                (function (RangeFacts) {
                    RangeFacts[RangeFacts["None"] = 0] = "None";
                    RangeFacts[RangeFacts["HasReturn"] = 1] = "HasReturn";
                    RangeFacts[RangeFacts["IsGenerator"] = 2] = "IsGenerator";
                    RangeFacts[RangeFacts["IsAsyncFunction"] = 4] = "IsAsyncFunction";
                    RangeFacts[RangeFacts["UsesThis"] = 8] = "UsesThis";
                    /**
                     * The range is in a function which needs the 'static' modifier in a class
                     */
                    RangeFacts[RangeFacts["InStaticRegion"] = 16] = "InStaticRegion";
                })(RangeFacts || (RangeFacts = {}));
                /**
                 * getRangeToExtract takes a span inside a text file and returns either an expression or an array
                 * of statements representing the minimum set of nodes needed to extract the entire span. This
                 * process may fail, in which case a set of errors is returned instead. These errors are shown to
                 * users if they have the provideRefactorNotApplicableReason option set.
                 */
                // exported only for tests
                function getRangeToExtract(sourceFile, span, invoked) {
                    if (invoked === void 0) {
                        invoked = true;
                    }
                    var length = span.length;
                    if (length === 0 && !invoked) {
                        return { errors: [ts.createFileDiagnostic(sourceFile, span.start, length, Messages.cannotExtractEmpty)] };
                    }
                    var cursorRequest = length === 0 && invoked;
                    var startToken = ts.findFirstNonJsxWhitespaceToken(sourceFile, span.start);
                    var endToken = ts.findTokenOnLeftOfPosition(sourceFile, ts.textSpanEnd(span));
                    /* If the refactoring command is invoked through a keyboard action it's safe to assume that the user is actively looking for
                    refactoring actions at the span location. As they may not know the exact range that will trigger a refactoring, we expand the
                    searched span to cover a real node range making it more likely that something useful will show up. */
                    var adjustedSpan = startToken && endToken && invoked ? getAdjustedSpanFromNodes(startToken, endToken, sourceFile) : span;
                    // Walk up starting from the the start position until we find a non-SourceFile node that subsumes the selected span.
                    // This may fail (e.g. you select two statements in the root of a source file)
                    var start = cursorRequest ? getExtractableParent(startToken) : ts.getParentNodeInSpan(startToken, sourceFile, adjustedSpan);
                    // Do the same for the ending position
                    var end = cursorRequest ? start : ts.getParentNodeInSpan(endToken, sourceFile, adjustedSpan);
                    var declarations = [];
                    // We'll modify these flags as we walk the tree to collect data
                    // about what things need to be done as part of the extraction.
                    var rangeFacts = RangeFacts.None;
                    if (!start || !end) {
                        // cannot find either start or end node
                        return { errors: [ts.createFileDiagnostic(sourceFile, span.start, length, Messages.cannotExtractRange)] };
                    }
                    if (ts.isJSDoc(start)) {
                        return { errors: [ts.createFileDiagnostic(sourceFile, span.start, length, Messages.cannotExtractJSDoc)] };
                    }
                    if (start.parent !== end.parent) {
                        // start and end nodes belong to different subtrees
                        return { errors: [ts.createFileDiagnostic(sourceFile, span.start, length, Messages.cannotExtractRange)] };
                    }
                    if (start !== end) {
                        // start and end should be statements and parent should be either block or a source file
                        if (!isBlockLike(start.parent)) {
                            return { errors: [ts.createFileDiagnostic(sourceFile, span.start, length, Messages.cannotExtractRange)] };
                        }
                        var statements = [];
                        var start2 = start; // TODO: GH#18217 Need to alias `start` to get this to compile. See https://github.com/Microsoft/TypeScript/issues/19955#issuecomment-344118248
                        for (var _i = 0, _a = start2.parent.statements; _i < _a.length; _i++) {
                            var statement = _a[_i];
                            if (statement === start || statements.length) {
                                var errors_1 = checkNode(statement);
                                if (errors_1) {
                                    return { errors: errors_1 };
                                }
                                statements.push(statement);
                            }
                            if (statement === end) {
                                break;
                            }
                        }
                        if (!statements.length) {
                            // https://github.com/Microsoft/TypeScript/issues/20559
                            // Ranges like [|case 1: break;|] will fail to populate `statements` because
                            // they will never find `start` in `start.parent.statements`.
                            // Consider: We could support ranges like [|case 1:|] by refining them to just
                            // the expression.
                            return { errors: [ts.createFileDiagnostic(sourceFile, span.start, length, Messages.cannotExtractRange)] };
                        }
                        return { targetRange: { range: statements, facts: rangeFacts, declarations: declarations } };
                    }
                    if (ts.isReturnStatement(start) && !start.expression) {
                        // Makes no sense to extract an expression-less return statement.
                        return { errors: [ts.createFileDiagnostic(sourceFile, span.start, length, Messages.cannotExtractRange)] };
                    }
                    // We have a single node (start)
                    var node = refineNode(start);
                    var errors = checkRootNode(node) || checkNode(node);
                    if (errors) {
                        return { errors: errors };
                    }
                    return { targetRange: { range: getStatementOrExpressionRange(node), facts: rangeFacts, declarations: declarations } }; // TODO: GH#18217
                    /**
                     * Attempt to refine the extraction node (generally, by shrinking it) to produce better results.
                     * @param node The unrefined extraction node.
                     */
                    function refineNode(node) {
                        if (ts.isReturnStatement(node)) {
                            if (node.expression) {
                                return node.expression;
                            }
                        }
                        else if (ts.isVariableStatement(node)) {
                            var numInitializers = 0;
                            var lastInitializer = void 0;
                            for (var _i = 0, _a = node.declarationList.declarations; _i < _a.length; _i++) {
                                var declaration = _a[_i];
                                if (declaration.initializer) {
                                    numInitializers++;
                                    lastInitializer = declaration.initializer;
                                }
                            }
                            if (numInitializers === 1) {
                                return lastInitializer;
                            }
                            // No special handling if there are multiple initializers.
                        }
                        else if (ts.isVariableDeclaration(node)) {
                            if (node.initializer) {
                                return node.initializer;
                            }
                        }
                        return node;
                    }
                    function checkRootNode(node) {
                        if (ts.isIdentifier(ts.isExpressionStatement(node) ? node.expression : node)) {
                            return [ts.createDiagnosticForNode(node, Messages.cannotExtractIdentifier)];
                        }
                        return undefined;
                    }
                    function checkForStaticContext(nodeToCheck, containingClass) {
                        var current = nodeToCheck;
                        while (current !== containingClass) {
                            if (current.kind === 165 /* PropertyDeclaration */) {
                                if (ts.isStatic(current)) {
                                    rangeFacts |= RangeFacts.InStaticRegion;
                                }
                                break;
                            }
                            else if (current.kind === 162 /* Parameter */) {
                                var ctorOrMethod = ts.getContainingFunction(current);
                                if (ctorOrMethod.kind === 169 /* Constructor */) {
                                    rangeFacts |= RangeFacts.InStaticRegion;
                                }
                                break;
                            }
                            else if (current.kind === 167 /* MethodDeclaration */) {
                                if (ts.isStatic(current)) {
                                    rangeFacts |= RangeFacts.InStaticRegion;
                                }
                            }
                            current = current.parent;
                        }
                    }
                    // Verifies whether we can actually extract this node or not.
                    function checkNode(nodeToCheck) {
                        var PermittedJumps;
                        (function (PermittedJumps) {
                            PermittedJumps[PermittedJumps["None"] = 0] = "None";
                            PermittedJumps[PermittedJumps["Break"] = 1] = "Break";
                            PermittedJumps[PermittedJumps["Continue"] = 2] = "Continue";
                            PermittedJumps[PermittedJumps["Return"] = 4] = "Return";
                        })(PermittedJumps || (PermittedJumps = {}));
                        // We believe it's true because the node is from the (unmodified) tree.
                        ts.Debug.assert(nodeToCheck.pos <= nodeToCheck.end, "This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809 (1)");
                        // For understanding how skipTrivia functioned:
                        ts.Debug.assert(!ts.positionIsSynthesized(nodeToCheck.pos), "This failure could trigger https://github.com/Microsoft/TypeScript/issues/20809 (2)");
                        if (!ts.isStatement(nodeToCheck) && !(ts.isExpressionNode(nodeToCheck) && isExtractableExpression(nodeToCheck))) {
                            return [ts.createDiagnosticForNode(nodeToCheck, Messages.statementOrExpressionExpected)];
                        }
                        if (nodeToCheck.flags & 8388608 /* Ambient */) {
                            return [ts.createDiagnosticForNode(nodeToCheck, Messages.cannotExtractAmbientBlock)];
                        }
                        // If we're in a class, see whether we're in a static region (static property initializer, static method, class constructor parameter default)
                        var containingClass = ts.getContainingClass(nodeToCheck);
                        if (containingClass) {
                            checkForStaticContext(nodeToCheck, containingClass);
                        }
                        var errors;
                        var permittedJumps = 4 /* Return */;
                        var seenLabels;
                        visit(nodeToCheck);
                        return errors;
                        function visit(node) {
                            if (errors) {
                                // already found an error - can stop now
                                return true;
                            }
                            if (ts.isDeclaration(node)) {
                                var declaringNode = (node.kind === 252 /* VariableDeclaration */) ? node.parent.parent : node;
                                if (ts.hasSyntacticModifier(declaringNode, 1 /* Export */)) {
                                    // TODO: GH#18217 Silly to use `errors ||` since it's definitely not defined (see top of `visit`)
                                    // Also, if we're only pushing one error, just use `let error: Diagnostic | undefined`!
                                    // Also TODO: GH#19956
                                    (errors || (errors = [])).push(ts.createDiagnosticForNode(node, Messages.cannotExtractExportedEntity));
                                    return true;
                                }
                                declarations.push(node.symbol);
                            }
                            // Some things can't be extracted in certain situations
                            switch (node.kind) {
                                case 264 /* ImportDeclaration */:
                                    (errors || (errors = [])).push(ts.createDiagnosticForNode(node, Messages.cannotExtractImport));
                                    return true;
                                case 269 /* ExportAssignment */:
                                    (errors || (errors = [])).push(ts.createDiagnosticForNode(node, Messages.cannotExtractExportedEntity));
                                    return true;
                                case 106 /* SuperKeyword */:
                                    // For a super *constructor call*, we have to be extracting the entire class,
                                    // but a super *method call* simply implies a 'this' reference
                                    if (node.parent.kind === 206 /* CallExpression */) {
                                        // Super constructor call
                                        var containingClass_1 = ts.getContainingClass(node); // TODO:GH#18217
                                        if (containingClass_1.pos < span.start || containingClass_1.end >= (span.start + span.length)) {
                                            (errors || (errors = [])).push(ts.createDiagnosticForNode(node, Messages.cannotExtractSuper));
                                            return true;
                                        }
                                    }
                                    else {
                                        rangeFacts |= RangeFacts.UsesThis;
                                    }
                                    break;
                                case 212 /* ArrowFunction */:
                                    // check if arrow function uses this
                                    ts.forEachChild(node, function check(n) {
                                        if (ts.isThis(n)) {
                                            rangeFacts |= RangeFacts.UsesThis;
                                        }
                                        else if (ts.isClassLike(n) || (ts.isFunctionLike(n) && !ts.isArrowFunction(n))) {
                                            return false;
                                        }
                                        else {
                                            ts.forEachChild(n, check);
                                        }
                                    });
                                // falls through
                                case 255 /* ClassDeclaration */:
                                case 254 /* FunctionDeclaration */:
                                    if (ts.isSourceFile(node.parent) && node.parent.externalModuleIndicator === undefined) {
                                        // You cannot extract global declarations
                                        (errors || (errors = [])).push(ts.createDiagnosticForNode(node, Messages.functionWillNotBeVisibleInTheNewScope));
                                    }
                                // falls through
                                case 224 /* ClassExpression */:
                                case 211 /* FunctionExpression */:
                                case 167 /* MethodDeclaration */:
                                case 169 /* Constructor */:
                                case 170 /* GetAccessor */:
                                case 171 /* SetAccessor */:
                                    // do not dive into functions or classes
                                    return false;
                            }
                            var savedPermittedJumps = permittedJumps;
                            switch (node.kind) {
                                case 237 /* IfStatement */:
                                    permittedJumps = 0 /* None */;
                                    break;
                                case 250 /* TryStatement */:
                                    // forbid all jumps inside try blocks
                                    permittedJumps = 0 /* None */;
                                    break;
                                case 233 /* Block */:
                                    if (node.parent && node.parent.kind === 250 /* TryStatement */ && node.parent.finallyBlock === node) {
                                        // allow unconditional returns from finally blocks
                                        permittedJumps = 4 /* Return */;
                                    }
                                    break;
                                case 288 /* DefaultClause */:
                                case 287 /* CaseClause */:
                                    // allow unlabeled break inside case clauses
                                    permittedJumps |= 1 /* Break */;
                                    break;
                                default:
                                    if (ts.isIterationStatement(node, /*lookInLabeledStatements*/ false)) {
                                        // allow unlabeled break/continue inside loops
                                        permittedJumps |= 1 /* Break */ | 2 /* Continue */;
                                    }
                                    break;
                            }
                            switch (node.kind) {
                                case 190 /* ThisType */:
                                case 108 /* ThisKeyword */:
                                    rangeFacts |= RangeFacts.UsesThis;
                                    break;
                                case 248 /* LabeledStatement */: {
                                    var label = node.label;
                                    (seenLabels || (seenLabels = [])).push(label.escapedText);
                                    ts.forEachChild(node, visit);
                                    seenLabels.pop();
                                    break;
                                }
                                case 244 /* BreakStatement */:
                                case 243 /* ContinueStatement */: {
                                    var label = node.label;
                                    if (label) {
                                        if (!ts.contains(seenLabels, label.escapedText)) {
                                            // attempts to jump to label that is not in range to be extracted
                                            (errors || (errors = [])).push(ts.createDiagnosticForNode(node, Messages.cannotExtractRangeContainingLabeledBreakOrContinueStatementWithTargetOutsideOfTheRange));
                                        }
                                    }
                                    else {
                                        if (!(permittedJumps & (node.kind === 244 /* BreakStatement */ ? 1 /* Break */ : 2 /* Continue */))) {
                                            // attempt to break or continue in a forbidden context
                                            (errors || (errors = [])).push(ts.createDiagnosticForNode(node, Messages.cannotExtractRangeContainingConditionalBreakOrContinueStatements));
                                        }
                                    }
                                    break;
                                }
                                case 216 /* AwaitExpression */:
                                    rangeFacts |= RangeFacts.IsAsyncFunction;
                                    break;
                                case 222 /* YieldExpression */:
                                    rangeFacts |= RangeFacts.IsGenerator;
                                    break;
                                case 245 /* ReturnStatement */:
                                    if (permittedJumps & 4 /* Return */) {
                                        rangeFacts |= RangeFacts.HasReturn;
                                    }
                                    else {
                                        (errors || (errors = [])).push(ts.createDiagnosticForNode(node, Messages.cannotExtractRangeContainingConditionalReturnStatement));
                                    }
                                    break;
                                default:
                                    ts.forEachChild(node, visit);
                                    break;
                            }
                            permittedJumps = savedPermittedJumps;
                        }
                    }
                }
                extractSymbol.getRangeToExtract = getRangeToExtract;
                /**
                 * Includes the final semicolon so that the span covers statements in cases where it would otherwise
                 * only cover the declaration list.
                 */
                function getAdjustedSpanFromNodes(startNode, endNode, sourceFile) {
                    var start = startNode.getStart(sourceFile);
                    var end = endNode.getEnd();
                    if (sourceFile.text.charCodeAt(end) === 59 /* semicolon */) {
                        end++;
                    }
                    return { start: start, length: end - start };
                }
                function getStatementOrExpressionRange(node) {
                    if (ts.isStatement(node)) {
                        return [node];
                    }
                    else if (ts.isExpressionNode(node)) {
                        // If our selection is the expression in an ExpressionStatement, expand
                        // the selection to include the enclosing Statement (this stops us
                        // from trying to care about the return value of the extracted function
                        // and eliminates double semicolon insertion in certain scenarios)
                        return ts.isExpressionStatement(node.parent) ? [node.parent] : node;
                    }
                    return undefined;
                }
                function isScope(node) {
                    return ts.isFunctionLikeDeclaration(node) || ts.isSourceFile(node) || ts.isModuleBlock(node) || ts.isClassLike(node);
                }
                /**
                 * Computes possible places we could extract the function into. For example,
                 * you may be able to extract into a class method *or* local closure *or* namespace function,
                 * depending on what's in the extracted body.
                 */
                function collectEnclosingScopes(range) {
                    var current = isReadonlyArray(range.range) ? ts.first(range.range) : range.range;
                    if (range.facts & RangeFacts.UsesThis) {
                        // if range uses this as keyword or as type inside the class then it can only be extracted to a method of the containing class
                        var containingClass = ts.getContainingClass(current);
                        if (containingClass) {
                            var containingFunction = ts.findAncestor(current, ts.isFunctionLikeDeclaration);
                            return containingFunction
                                ? [containingFunction, containingClass]
                                : [containingClass];
                        }
                    }
                    var scopes = [];
                    while (true) {
                        current = current.parent;
                        // A function parameter's initializer is actually in the outer scope, not the function declaration
                        if (current.kind === 162 /* Parameter */) {
                            // Skip all the way to the outer scope of the function that declared this parameter
                            current = ts.findAncestor(current, function (parent) { return ts.isFunctionLikeDeclaration(parent); }).parent;
                        }
                        // We want to find the nearest parent where we can place an "equivalent" sibling to the node we're extracting out of.
                        // Walk up to the closest parent of a place where we can logically put a sibling:
                        //  * Function declaration
                        //  * Class declaration or expression
                        //  * Module/namespace or source file
                        if (isScope(current)) {
                            scopes.push(current);
                            if (current.kind === 300 /* SourceFile */) {
                                return scopes;
                            }
                        }
                    }
                }
                function getFunctionExtractionAtIndex(targetRange, context, requestedChangesIndex) {
                    var _a = getPossibleExtractionsWorker(targetRange, context), scopes = _a.scopes, _b = _a.readsAndWrites, target = _b.target, usagesPerScope = _b.usagesPerScope, functionErrorsPerScope = _b.functionErrorsPerScope, exposedVariableDeclarations = _b.exposedVariableDeclarations;
                    ts.Debug.assert(!functionErrorsPerScope[requestedChangesIndex].length, "The extraction went missing? How?");
                    context.cancellationToken.throwIfCancellationRequested(); // TODO: GH#18217
                    return extractFunctionInScope(target, scopes[requestedChangesIndex], usagesPerScope[requestedChangesIndex], exposedVariableDeclarations, targetRange, context);
                }
                function getConstantExtractionAtIndex(targetRange, context, requestedChangesIndex) {
                    var _a = getPossibleExtractionsWorker(targetRange, context), scopes = _a.scopes, _b = _a.readsAndWrites, target = _b.target, usagesPerScope = _b.usagesPerScope, constantErrorsPerScope = _b.constantErrorsPerScope, exposedVariableDeclarations = _b.exposedVariableDeclarations;
                    ts.Debug.assert(!constantErrorsPerScope[requestedChangesIndex].length, "The extraction went missing? How?");
                    ts.Debug.assert(exposedVariableDeclarations.length === 0, "Extract constant accepted a range containing a variable declaration?");
                    context.cancellationToken.throwIfCancellationRequested();
                    var expression = ts.isExpression(target)
                        ? target
                        : target.statements[0].expression;
                    return extractConstantInScope(expression, scopes[requestedChangesIndex], usagesPerScope[requestedChangesIndex], targetRange.facts, context);
                }
                /**
                 * Given a piece of text to extract ('targetRange'), computes a list of possible extractions.
                 * Each returned ExtractResultForScope corresponds to a possible target scope and is either a set of changes
                 * or an error explaining why we can't extract into that scope.
                 */
                function getPossibleExtractions(targetRange, context) {
                    var _a = getPossibleExtractionsWorker(targetRange, context), scopes = _a.scopes, _b = _a.readsAndWrites, functionErrorsPerScope = _b.functionErrorsPerScope, constantErrorsPerScope = _b.constantErrorsPerScope;
                    // Need the inner type annotation to avoid https://github.com/Microsoft/TypeScript/issues/7547
                    var extractions = scopes.map(function (scope, i) {
                        var functionDescriptionPart = getDescriptionForFunctionInScope(scope);
                        var constantDescriptionPart = getDescriptionForConstantInScope(scope);
                        var scopeDescription = ts.isFunctionLikeDeclaration(scope)
                            ? getDescriptionForFunctionLikeDeclaration(scope)
                            : ts.isClassLike(scope)
                                ? getDescriptionForClassLikeDeclaration(scope)
                                : getDescriptionForModuleLikeDeclaration(scope);
                        var functionDescription;
                        var constantDescription;
                        if (scopeDescription === 1 /* Global */) {
                            functionDescription = ts.formatStringFromArgs(ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_0_in_1_scope), [functionDescriptionPart, "global"]);
                            constantDescription = ts.formatStringFromArgs(ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_0_in_1_scope), [constantDescriptionPart, "global"]);
                        }
                        else if (scopeDescription === 0 /* Module */) {
                            functionDescription = ts.formatStringFromArgs(ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_0_in_1_scope), [functionDescriptionPart, "module"]);
                            constantDescription = ts.formatStringFromArgs(ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_0_in_1_scope), [constantDescriptionPart, "module"]);
                        }
                        else {
                            functionDescription = ts.formatStringFromArgs(ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_0_in_1), [functionDescriptionPart, scopeDescription]);
                            constantDescription = ts.formatStringFromArgs(ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_0_in_1), [constantDescriptionPart, scopeDescription]);
                        }
                        // Customize the phrasing for the innermost scope to increase clarity.
                        if (i === 0 && !ts.isClassLike(scope)) {
                            constantDescription = ts.formatStringFromArgs(ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_0_in_enclosing_scope), [constantDescriptionPart]);
                        }
                        return {
                            functionExtraction: {
                                description: functionDescription,
                                errors: functionErrorsPerScope[i],
                            },
                            constantExtraction: {
                                description: constantDescription,
                                errors: constantErrorsPerScope[i],
                            },
                        };
                    });
                    return extractions;
                }
                function getPossibleExtractionsWorker(targetRange, context) {
                    var sourceFile = context.file;
                    var scopes = collectEnclosingScopes(targetRange);
                    var enclosingTextRange = getEnclosingTextRange(targetRange, sourceFile);
                    var readsAndWrites = collectReadsAndWrites(targetRange, scopes, enclosingTextRange, sourceFile, context.program.getTypeChecker(), context.cancellationToken);
                    return { scopes: scopes, readsAndWrites: readsAndWrites };
                }
                function getDescriptionForFunctionInScope(scope) {
                    return ts.isFunctionLikeDeclaration(scope)
                        ? "inner function"
                        : ts.isClassLike(scope)
                            ? "method"
                            : "function";
                }
                function getDescriptionForConstantInScope(scope) {
                    return ts.isClassLike(scope)
                        ? "readonly field"
                        : "constant";
                }
                function getDescriptionForFunctionLikeDeclaration(scope) {
                    switch (scope.kind) {
                        case 169 /* Constructor */:
                            return "constructor";
                        case 211 /* FunctionExpression */:
                        case 254 /* FunctionDeclaration */:
                            return scope.name
                                ? "function '" + scope.name.text + "'"
                                : ts.ANONYMOUS;
                        case 212 /* ArrowFunction */:
                            return "arrow function";
                        case 167 /* MethodDeclaration */:
                            return "method '" + scope.name.getText() + "'";
                        case 170 /* GetAccessor */:
                            return "'get " + scope.name.getText() + "'";
                        case 171 /* SetAccessor */:
                            return "'set " + scope.name.getText() + "'";
                        default:
                            throw ts.Debug.assertNever(scope, "Unexpected scope kind " + scope.kind);
                    }
                }
                function getDescriptionForClassLikeDeclaration(scope) {
                    return scope.kind === 255 /* ClassDeclaration */
                        ? scope.name ? "class '" + scope.name.text + "'" : "anonymous class declaration"
                        : scope.name ? "class expression '" + scope.name.text + "'" : "anonymous class expression";
                }
                function getDescriptionForModuleLikeDeclaration(scope) {
                    return scope.kind === 260 /* ModuleBlock */
                        ? "namespace '" + scope.parent.name.getText() + "'"
                        : scope.externalModuleIndicator ? 0 /* Module */ : 1 /* Global */;
                }
                var SpecialScope;
                (function (SpecialScope) {
                    SpecialScope[SpecialScope["Module"] = 0] = "Module";
                    SpecialScope[SpecialScope["Global"] = 1] = "Global";
                })(SpecialScope || (SpecialScope = {}));
                /**
                 * Result of 'extractRange' operation for a specific scope.
                 * Stores either a list of changes that should be applied to extract a range or a list of errors
                 */
                function extractFunctionInScope(node, scope, _a, exposedVariableDeclarations, range, context) {
                    var usagesInScope = _a.usages, typeParameterUsages = _a.typeParameterUsages, substitutions = _a.substitutions;
                    var checker = context.program.getTypeChecker();
                    var scriptTarget = ts.getEmitScriptTarget(context.program.getCompilerOptions());
                    var importAdder = ts.codefix.createImportAdder(context.file, context.program, context.preferences, context.host);
                    // Make a unique name for the extracted function
                    var file = scope.getSourceFile();
                    var functionNameText = ts.getUniqueName(ts.isClassLike(scope) ? "newMethod" : "newFunction", file);
                    var isJS = ts.isInJSFile(scope);
                    var functionName = ts.factory.createIdentifier(functionNameText);
                    var returnType;
                    var parameters = [];
                    var callArguments = [];
                    var writes;
                    usagesInScope.forEach(function (usage, name) {
                        var typeNode;
                        if (!isJS) {
                            var type = checker.getTypeOfSymbolAtLocation(usage.symbol, usage.node);
                            // Widen the type so we don't emit nonsense annotations like "function fn(x: 3) {"
                            type = checker.getBaseTypeOfLiteralType(type);
                            typeNode = ts.codefix.typeToAutoImportableTypeNode(checker, importAdder, type, scope, scriptTarget, 1 /* NoTruncation */);
                        }
                        var paramDecl = ts.factory.createParameterDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, 
                        /*dotDotDotToken*/ undefined, 
                        /*name*/ name, 
                        /*questionToken*/ undefined, typeNode);
                        parameters.push(paramDecl);
                        if (usage.usage === 2 /* Write */) {
                            (writes || (writes = [])).push(usage);
                        }
                        callArguments.push(ts.factory.createIdentifier(name));
                    });
                    var typeParametersAndDeclarations = ts.arrayFrom(typeParameterUsages.values()).map(function (type) { return ({ type: type, declaration: getFirstDeclaration(type) }); });
                    var sortedTypeParametersAndDeclarations = typeParametersAndDeclarations.sort(compareTypesByDeclarationOrder);
                    var typeParameters = sortedTypeParametersAndDeclarations.length === 0
                        ? undefined
                        : sortedTypeParametersAndDeclarations.map(function (t) { return t.declaration; });
                    // Strictly speaking, we should check whether each name actually binds to the appropriate type
                    // parameter.  In cases of shadowing, they may not.
                    var callTypeArguments = typeParameters !== undefined
                        ? typeParameters.map(function (decl) { return ts.factory.createTypeReferenceNode(decl.name, /*typeArguments*/ undefined); })
                        : undefined;
                    // Provide explicit return types for contextually-typed functions
                    // to avoid problems when there are literal types present
                    if (ts.isExpression(node) && !isJS) {
                        var contextualType = checker.getContextualType(node);
                        returnType = checker.typeToTypeNode(contextualType, scope, 1 /* NoTruncation */); // TODO: GH#18217
                    }
                    var _b = transformFunctionBody(node, exposedVariableDeclarations, writes, substitutions, !!(range.facts & RangeFacts.HasReturn)), body = _b.body, returnValueProperty = _b.returnValueProperty;
                    ts.suppressLeadingAndTrailingTrivia(body);
                    var newFunction;
                    if (ts.isClassLike(scope)) {
                        // always create private method in TypeScript files
                        var modifiers = isJS ? [] : [ts.factory.createModifier(121 /* PrivateKeyword */)];
                        if (range.facts & RangeFacts.InStaticRegion) {
                            modifiers.push(ts.factory.createModifier(124 /* StaticKeyword */));
                        }
                        if (range.facts & RangeFacts.IsAsyncFunction) {
                            modifiers.push(ts.factory.createModifier(130 /* AsyncKeyword */));
                        }
                        newFunction = ts.factory.createMethodDeclaration(
                        /*decorators*/ undefined, modifiers.length ? modifiers : undefined, range.facts & RangeFacts.IsGenerator ? ts.factory.createToken(41 /* AsteriskToken */) : undefined, functionName, 
                        /*questionToken*/ undefined, typeParameters, parameters, returnType, body);
                    }
                    else {
                        newFunction = ts.factory.createFunctionDeclaration(
                        /*decorators*/ undefined, range.facts & RangeFacts.IsAsyncFunction ? [ts.factory.createToken(130 /* AsyncKeyword */)] : undefined, range.facts & RangeFacts.IsGenerator ? ts.factory.createToken(41 /* AsteriskToken */) : undefined, functionName, typeParameters, parameters, returnType, body);
                    }
                    var changeTracker = ts.textChanges.ChangeTracker.fromContext(context);
                    var minInsertionPos = (isReadonlyArray(range.range) ? ts.last(range.range) : range.range).end;
                    var nodeToInsertBefore = getNodeToInsertFunctionBefore(minInsertionPos, scope);
                    if (nodeToInsertBefore) {
                        changeTracker.insertNodeBefore(context.file, nodeToInsertBefore, newFunction, /*blankLineBetween*/ true);
                    }
                    else {
                        changeTracker.insertNodeAtEndOfScope(context.file, scope, newFunction);
                    }
                    importAdder.writeFixes(changeTracker);
                    var newNodes = [];
                    // replace range with function call
                    var called = getCalledExpression(scope, range, functionNameText);
                    var call = ts.factory.createCallExpression(called, callTypeArguments, // Note that no attempt is made to take advantage of type argument inference
                    callArguments);
                    if (range.facts & RangeFacts.IsGenerator) {
                        call = ts.factory.createYieldExpression(ts.factory.createToken(41 /* AsteriskToken */), call);
                    }
                    if (range.facts & RangeFacts.IsAsyncFunction) {
                        call = ts.factory.createAwaitExpression(call);
                    }
                    if (isInJSXContent(node)) {
                        call = ts.factory.createJsxExpression(/*dotDotDotToken*/ undefined, call);
                    }
                    if (exposedVariableDeclarations.length && !writes) {
                        // No need to mix declarations and writes.
                        // How could any variables be exposed if there's a return statement?
                        ts.Debug.assert(!returnValueProperty, "Expected no returnValueProperty");
                        ts.Debug.assert(!(range.facts & RangeFacts.HasReturn), "Expected RangeFacts.HasReturn flag to be unset");
                        if (exposedVariableDeclarations.length === 1) {
                            // Declaring exactly one variable: let x = newFunction();
                            var variableDeclaration = exposedVariableDeclarations[0];
                            newNodes.push(ts.factory.createVariableStatement(
                            /*modifiers*/ undefined, ts.factory.createVariableDeclarationList([ts.factory.createVariableDeclaration(ts.getSynthesizedDeepClone(variableDeclaration.name), /*exclamationToken*/ undefined, /*type*/ ts.getSynthesizedDeepClone(variableDeclaration.type), /*initializer*/ call)], // TODO (acasey): test binding patterns
                            variableDeclaration.parent.flags)));
                        }
                        else {
                            // Declaring multiple variables / return properties:
                            //   let {x, y} = newFunction();
                            var bindingElements = [];
                            var typeElements = [];
                            var commonNodeFlags = exposedVariableDeclarations[0].parent.flags;
                            var sawExplicitType = false;
                            for (var _i = 0, exposedVariableDeclarations_1 = exposedVariableDeclarations; _i < exposedVariableDeclarations_1.length; _i++) {
                                var variableDeclaration = exposedVariableDeclarations_1[_i];
                                bindingElements.push(ts.factory.createBindingElement(
                                /*dotDotDotToken*/ undefined, 
                                /*propertyName*/ undefined, 
                                /*name*/ ts.getSynthesizedDeepClone(variableDeclaration.name)));
                                // Being returned through an object literal will have widened the type.
                                var variableType = checker.typeToTypeNode(checker.getBaseTypeOfLiteralType(checker.getTypeAtLocation(variableDeclaration)), scope, 1 /* NoTruncation */);
                                typeElements.push(ts.factory.createPropertySignature(
                                /*modifiers*/ undefined, 
                                /*name*/ variableDeclaration.symbol.name, 
                                /*questionToken*/ undefined, 
                                /*type*/ variableType));
                                sawExplicitType = sawExplicitType || variableDeclaration.type !== undefined;
                                commonNodeFlags = commonNodeFlags & variableDeclaration.parent.flags;
                            }
                            var typeLiteral = sawExplicitType ? ts.factory.createTypeLiteralNode(typeElements) : undefined;
                            if (typeLiteral) {
                                ts.setEmitFlags(typeLiteral, 1 /* SingleLine */);
                            }
                            newNodes.push(ts.factory.createVariableStatement(
                            /*modifiers*/ undefined, ts.factory.createVariableDeclarationList([ts.factory.createVariableDeclaration(ts.factory.createObjectBindingPattern(bindingElements), 
                                /*exclamationToken*/ undefined, 
                                /*type*/ typeLiteral, 
                                /*initializer*/ call)], commonNodeFlags)));
                        }
                    }
                    else if (exposedVariableDeclarations.length || writes) {
                        if (exposedVariableDeclarations.length) {
                            // CONSIDER: we're going to create one statement per variable, but we could actually preserve their original grouping.
                            for (var _c = 0, exposedVariableDeclarations_2 = exposedVariableDeclarations; _c < exposedVariableDeclarations_2.length; _c++) {
                                var variableDeclaration = exposedVariableDeclarations_2[_c];
                                var flags = variableDeclaration.parent.flags;
                                if (flags & 2 /* Const */) {
                                    flags = (flags & ~2 /* Const */) | 1 /* Let */;
                                }
                                newNodes.push(ts.factory.createVariableStatement(
                                /*modifiers*/ undefined, ts.factory.createVariableDeclarationList([ts.factory.createVariableDeclaration(variableDeclaration.symbol.name, /*exclamationToken*/ undefined, getTypeDeepCloneUnionUndefined(variableDeclaration.type))], flags)));
                            }
                        }
                        if (returnValueProperty) {
                            // has both writes and return, need to create variable declaration to hold return value;
                            newNodes.push(ts.factory.createVariableStatement(
                            /*modifiers*/ undefined, ts.factory.createVariableDeclarationList([ts.factory.createVariableDeclaration(returnValueProperty, /*exclamationToken*/ undefined, getTypeDeepCloneUnionUndefined(returnType))], 1 /* Let */)));
                        }
                        var assignments = getPropertyAssignmentsForWritesAndVariableDeclarations(exposedVariableDeclarations, writes);
                        if (returnValueProperty) {
                            assignments.unshift(ts.factory.createShorthandPropertyAssignment(returnValueProperty));
                        }
                        // propagate writes back
                        if (assignments.length === 1) {
                            // We would only have introduced a return value property if there had been
                            // other assignments to make.
                            ts.Debug.assert(!returnValueProperty, "Shouldn't have returnValueProperty here");
                            newNodes.push(ts.factory.createExpressionStatement(ts.factory.createAssignment(assignments[0].name, call)));
                            if (range.facts & RangeFacts.HasReturn) {
                                newNodes.push(ts.factory.createReturnStatement());
                            }
                        }
                        else {
                            // emit e.g.
                            //   { a, b, __return } = newFunction(a, b);
                            //   return __return;
                            newNodes.push(ts.factory.createExpressionStatement(ts.factory.createAssignment(ts.factory.createObjectLiteralExpression(assignments), call)));
                            if (returnValueProperty) {
                                newNodes.push(ts.factory.createReturnStatement(ts.factory.createIdentifier(returnValueProperty)));
                            }
                        }
                    }
                    else {
                        if (range.facts & RangeFacts.HasReturn) {
                            newNodes.push(ts.factory.createReturnStatement(call));
                        }
                        else if (isReadonlyArray(range.range)) {
                            newNodes.push(ts.factory.createExpressionStatement(call));
                        }
                        else {
                            newNodes.push(call);
                        }
                    }
                    if (isReadonlyArray(range.range)) {
                        changeTracker.replaceNodeRangeWithNodes(context.file, ts.first(range.range), ts.last(range.range), newNodes);
                    }
                    else {
                        changeTracker.replaceNodeWithNodes(context.file, range.range, newNodes);
                    }
                    var edits = changeTracker.getChanges();
                    var renameRange = isReadonlyArray(range.range) ? ts.first(range.range) : range.range;
                    var renameFilename = renameRange.getSourceFile().fileName;
                    var renameLocation = ts.getRenameLocation(edits, renameFilename, functionNameText, /*isDeclaredBeforeUse*/ false);
                    return { renameFilename: renameFilename, renameLocation: renameLocation, edits: edits };
                    function getTypeDeepCloneUnionUndefined(typeNode) {
                        if (typeNode === undefined) {
                            return undefined;
                        }
                        var clone = ts.getSynthesizedDeepClone(typeNode);
                        var withoutParens = clone;
                        while (ts.isParenthesizedTypeNode(withoutParens)) {
                            withoutParens = withoutParens.type;
                        }
                        return ts.isUnionTypeNode(withoutParens) && ts.find(withoutParens.types, function (t) { return t.kind === 151 /* UndefinedKeyword */; })
                            ? clone
                            : ts.factory.createUnionTypeNode([clone, ts.factory.createKeywordTypeNode(151 /* UndefinedKeyword */)]);
                    }
                }
                /**
                 * Result of 'extractRange' operation for a specific scope.
                 * Stores either a list of changes that should be applied to extract a range or a list of errors
                 */
                function extractConstantInScope(node, scope, _a, rangeFacts, context) {
                    var _b;
                    var substitutions = _a.substitutions;
                    var checker = context.program.getTypeChecker();
                    // Make a unique name for the extracted variable
                    var file = scope.getSourceFile();
                    var localNameText = ts.getUniqueName(ts.isClassLike(scope) ? "newProperty" : "newLocal", file);
                    var isJS = ts.isInJSFile(scope);
                    var variableType = isJS || !checker.isContextSensitive(node)
                        ? undefined
                        : checker.typeToTypeNode(checker.getContextualType(node), scope, 1 /* NoTruncation */); // TODO: GH#18217
                    var initializer = transformConstantInitializer(node, substitutions);
                    (_b = transformFunctionInitializerAndType(variableType, initializer), variableType = _b.variableType, initializer = _b.initializer);
                    ts.suppressLeadingAndTrailingTrivia(initializer);
                    var changeTracker = ts.textChanges.ChangeTracker.fromContext(context);
                    if (ts.isClassLike(scope)) {
                        ts.Debug.assert(!isJS, "Cannot extract to a JS class"); // See CannotExtractToJSClass
                        var modifiers = [];
                        modifiers.push(ts.factory.createModifier(121 /* PrivateKeyword */));
                        if (rangeFacts & RangeFacts.InStaticRegion) {
                            modifiers.push(ts.factory.createModifier(124 /* StaticKeyword */));
                        }
                        modifiers.push(ts.factory.createModifier(143 /* ReadonlyKeyword */));
                        var newVariable = ts.factory.createPropertyDeclaration(
                        /*decorators*/ undefined, modifiers, localNameText, 
                        /*questionToken*/ undefined, variableType, initializer);
                        var localReference = ts.factory.createPropertyAccessExpression(rangeFacts & RangeFacts.InStaticRegion
                            ? ts.factory.createIdentifier(scope.name.getText()) // TODO: GH#18217
                            : ts.factory.createThis(), ts.factory.createIdentifier(localNameText));
                        if (isInJSXContent(node)) {
                            localReference = ts.factory.createJsxExpression(/*dotDotDotToken*/ undefined, localReference);
                        }
                        // Declare
                        var maxInsertionPos = node.pos;
                        var nodeToInsertBefore = getNodeToInsertPropertyBefore(maxInsertionPos, scope);
                        changeTracker.insertNodeBefore(context.file, nodeToInsertBefore, newVariable, /*blankLineBetween*/ true);
                        // Consume
                        changeTracker.replaceNode(context.file, node, localReference);
                    }
                    else {
                        var newVariableDeclaration = ts.factory.createVariableDeclaration(localNameText, /*exclamationToken*/ undefined, variableType, initializer);
                        // If the node is part of an initializer in a list of variable declarations, insert a new
                        // variable declaration into the list (in case it depends on earlier ones).
                        // CONSIDER: If the declaration list isn't const, we might want to split it into multiple
                        // lists so that the newly extracted one can be const.
                        var oldVariableDeclaration = getContainingVariableDeclarationIfInList(node, scope);
                        if (oldVariableDeclaration) {
                            // Declare
                            // CONSIDER: could detect that each is on a separate line (See `extractConstant_VariableList_MultipleLines` in `extractConstants.ts`)
                            changeTracker.insertNodeBefore(context.file, oldVariableDeclaration, newVariableDeclaration);
                            // Consume
                            var localReference = ts.factory.createIdentifier(localNameText);
                            changeTracker.replaceNode(context.file, node, localReference);
                        }
                        else if (node.parent.kind === 236 /* ExpressionStatement */ && scope === ts.findAncestor(node, isScope)) {
                            // If the parent is an expression statement and the target scope is the immediately enclosing one,
                            // replace the statement with the declaration.
                            var newVariableStatement = ts.factory.createVariableStatement(
                            /*modifiers*/ undefined, ts.factory.createVariableDeclarationList([newVariableDeclaration], 2 /* Const */));
                            changeTracker.replaceNode(context.file, node.parent, newVariableStatement);
                        }
                        else {
                            var newVariableStatement = ts.factory.createVariableStatement(
                            /*modifiers*/ undefined, ts.factory.createVariableDeclarationList([newVariableDeclaration], 2 /* Const */));
                            // Declare
                            var nodeToInsertBefore = getNodeToInsertConstantBefore(node, scope);
                            if (nodeToInsertBefore.pos === 0) {
                                changeTracker.insertNodeAtTopOfFile(context.file, newVariableStatement, /*blankLineBetween*/ false);
                            }
                            else {
                                changeTracker.insertNodeBefore(context.file, nodeToInsertBefore, newVariableStatement, /*blankLineBetween*/ false);
                            }
                            // Consume
                            if (node.parent.kind === 236 /* ExpressionStatement */) {
                                // If the parent is an expression statement, delete it.
                                changeTracker.delete(context.file, node.parent);
                            }
                            else {
                                var localReference = ts.factory.createIdentifier(localNameText);
                                // When extract to a new variable in JSX content, need to wrap a {} out of the new variable
                                // or it will become a plain text
                                if (isInJSXContent(node)) {
                                    localReference = ts.factory.createJsxExpression(/*dotDotDotToken*/ undefined, localReference);
                                }
                                changeTracker.replaceNode(context.file, node, localReference);
                            }
                        }
                    }
                    var edits = changeTracker.getChanges();
                    var renameFilename = node.getSourceFile().fileName;
                    var renameLocation = ts.getRenameLocation(edits, renameFilename, localNameText, /*isDeclaredBeforeUse*/ true);
                    return { renameFilename: renameFilename, renameLocation: renameLocation, edits: edits };
                    function transformFunctionInitializerAndType(variableType, initializer) {
                        // If no contextual type exists there is nothing to transfer to the function signature
                        if (variableType === undefined)
                            return { variableType: variableType, initializer: initializer };
                        // Only do this for function expressions and arrow functions that are not generic
                        if (!ts.isFunctionExpression(initializer) && !ts.isArrowFunction(initializer) || !!initializer.typeParameters)
                            return { variableType: variableType, initializer: initializer };
                        var functionType = checker.getTypeAtLocation(node);
                        var functionSignature = ts.singleOrUndefined(checker.getSignaturesOfType(functionType, 0 /* Call */));
                        // If no function signature, maybe there was an error, do nothing
                        if (!functionSignature)
                            return { variableType: variableType, initializer: initializer };
                        // If the function signature has generic type parameters we don't attempt to move the parameters
                        if (!!functionSignature.getTypeParameters())
                            return { variableType: variableType, initializer: initializer };
                        // We add parameter types if needed
                        var parameters = [];
                        var hasAny = false;
                        for (var _i = 0, _a = initializer.parameters; _i < _a.length; _i++) {
                            var p = _a[_i];
                            if (p.type) {
                                parameters.push(p);
                            }
                            else {
                                var paramType = checker.getTypeAtLocation(p);
                                if (paramType === checker.getAnyType())
                                    hasAny = true;
                                parameters.push(ts.factory.updateParameterDeclaration(p, p.decorators, p.modifiers, p.dotDotDotToken, p.name, p.questionToken, p.type || checker.typeToTypeNode(paramType, scope, 1 /* NoTruncation */), p.initializer));
                            }
                        }
                        // If a parameter was inferred as any we skip adding function parameters at all.
                        // Turning an implicit any (which under common settings is a error) to an explicit
                        // is probably actually a worse refactor outcome.
                        if (hasAny)
                            return { variableType: variableType, initializer: initializer };
                        variableType = undefined;
                        if (ts.isArrowFunction(initializer)) {
                            initializer = ts.factory.updateArrowFunction(initializer, node.modifiers, initializer.typeParameters, parameters, initializer.type || checker.typeToTypeNode(functionSignature.getReturnType(), scope, 1 /* NoTruncation */), initializer.equalsGreaterThanToken, initializer.body);
                        }
                        else {
                            if (functionSignature && !!functionSignature.thisParameter) {
                                var firstParameter = ts.firstOrUndefined(parameters);
                                // If the function signature has a this parameter and if the first defined parameter is not the this parameter, we must add it
                                // Note: If this parameter was already there, it would have been previously updated with the type if not type was present
                                if ((!firstParameter || (ts.isIdentifier(firstParameter.name) && firstParameter.name.escapedText !== "this"))) {
                                    var thisType = checker.getTypeOfSymbolAtLocation(functionSignature.thisParameter, node);
                                    parameters.splice(0, 0, ts.factory.createParameterDeclaration(
                                    /* decorators */ undefined, 
                                    /* modifiers */ undefined, 
                                    /* dotDotDotToken */ undefined, "this", 
                                    /* questionToken */ undefined, checker.typeToTypeNode(thisType, scope, 1 /* NoTruncation */)));
                                }
                            }
                            initializer = ts.factory.updateFunctionExpression(initializer, node.modifiers, initializer.asteriskToken, initializer.name, initializer.typeParameters, parameters, initializer.type || checker.typeToTypeNode(functionSignature.getReturnType(), scope, 1 /* NoTruncation */), initializer.body);
                        }
                        return { variableType: variableType, initializer: initializer };
                    }
                }
                function getContainingVariableDeclarationIfInList(node, scope) {
                    var prevNode;
                    while (node !== undefined && node !== scope) {
                        if (ts.isVariableDeclaration(node) &&
                            node.initializer === prevNode &&
                            ts.isVariableDeclarationList(node.parent) &&
                            node.parent.declarations.length > 1) {
                            return node;
                        }
                        prevNode = node;
                        node = node.parent;
                    }
                }
                function getFirstDeclaration(type) {
                    var firstDeclaration;
                    var symbol = type.symbol;
                    if (symbol && symbol.declarations) {
                        for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                            var declaration = _a[_i];
                            if (firstDeclaration === undefined || declaration.pos < firstDeclaration.pos) {
                                firstDeclaration = declaration;
                            }
                        }
                    }
                    return firstDeclaration;
                }
                function compareTypesByDeclarationOrder(_a, _b) {
                    var type1 = _a.type, declaration1 = _a.declaration;
                    var type2 = _b.type, declaration2 = _b.declaration;
                    return ts.compareProperties(declaration1, declaration2, "pos", ts.compareValues)
                        || ts.compareStringsCaseSensitive(type1.symbol ? type1.symbol.getName() : "", type2.symbol ? type2.symbol.getName() : "")
                        || ts.compareValues(type1.id, type2.id);
                }
                function getCalledExpression(scope, range, functionNameText) {
                    var functionReference = ts.factory.createIdentifier(functionNameText);
                    if (ts.isClassLike(scope)) {
                        var lhs = range.facts & RangeFacts.InStaticRegion ? ts.factory.createIdentifier(scope.name.text) : ts.factory.createThis(); // TODO: GH#18217
                        return ts.factory.createPropertyAccessExpression(lhs, functionReference);
                    }
                    else {
                        return functionReference;
                    }
                }
                function transformFunctionBody(body, exposedVariableDeclarations, writes, substitutions, hasReturn) {
                    var hasWritesOrVariableDeclarations = writes !== undefined || exposedVariableDeclarations.length > 0;
                    if (ts.isBlock(body) && !hasWritesOrVariableDeclarations && substitutions.size === 0) {
                        // already block, no declarations or writes to propagate back, no substitutions - can use node as is
                        return { body: ts.factory.createBlock(body.statements, /*multLine*/ true), returnValueProperty: undefined };
                    }
                    var returnValueProperty;
                    var ignoreReturns = false;
                    var statements = ts.factory.createNodeArray(ts.isBlock(body) ? body.statements.slice(0) : [ts.isStatement(body) ? body : ts.factory.createReturnStatement(body)]);
                    // rewrite body if either there are writes that should be propagated back via return statements or there are substitutions
                    if (hasWritesOrVariableDeclarations || substitutions.size) {
                        var rewrittenStatements = ts.visitNodes(statements, visitor).slice();
                        if (hasWritesOrVariableDeclarations && !hasReturn && ts.isStatement(body)) {
                            // add return at the end to propagate writes back in case if control flow falls out of the function body
                            // it is ok to know that range has at least one return since it we only allow unconditional returns
                            var assignments = getPropertyAssignmentsForWritesAndVariableDeclarations(exposedVariableDeclarations, writes);
                            if (assignments.length === 1) {
                                rewrittenStatements.push(ts.factory.createReturnStatement(assignments[0].name));
                            }
                            else {
                                rewrittenStatements.push(ts.factory.createReturnStatement(ts.factory.createObjectLiteralExpression(assignments)));
                            }
                        }
                        return { body: ts.factory.createBlock(rewrittenStatements, /*multiLine*/ true), returnValueProperty: returnValueProperty };
                    }
                    else {
                        return { body: ts.factory.createBlock(statements, /*multiLine*/ true), returnValueProperty: undefined };
                    }
                    function visitor(node) {
                        if (!ignoreReturns && ts.isReturnStatement(node) && hasWritesOrVariableDeclarations) {
                            var assignments = getPropertyAssignmentsForWritesAndVariableDeclarations(exposedVariableDeclarations, writes);
                            if (node.expression) {
                                if (!returnValueProperty) {
                                    returnValueProperty = "__return";
                                }
                                assignments.unshift(ts.factory.createPropertyAssignment(returnValueProperty, ts.visitNode(node.expression, visitor)));
                            }
                            if (assignments.length === 1) {
                                return ts.factory.createReturnStatement(assignments[0].name);
                            }
                            else {
                                return ts.factory.createReturnStatement(ts.factory.createObjectLiteralExpression(assignments));
                            }
                        }
                        else {
                            var oldIgnoreReturns = ignoreReturns;
                            ignoreReturns = ignoreReturns || ts.isFunctionLikeDeclaration(node) || ts.isClassLike(node);
                            var substitution = substitutions.get(ts.getNodeId(node).toString());
                            var result = substitution ? ts.getSynthesizedDeepClone(substitution) : ts.visitEachChild(node, visitor, ts.nullTransformationContext);
                            ignoreReturns = oldIgnoreReturns;
                            return result;
                        }
                    }
                }
                function transformConstantInitializer(initializer, substitutions) {
                    return substitutions.size
                        ? visitor(initializer)
                        : initializer;
                    function visitor(node) {
                        var substitution = substitutions.get(ts.getNodeId(node).toString());
                        return substitution ? ts.getSynthesizedDeepClone(substitution) : ts.visitEachChild(node, visitor, ts.nullTransformationContext);
                    }
                }
                function getStatementsOrClassElements(scope) {
                    if (ts.isFunctionLikeDeclaration(scope)) {
                        var body = scope.body; // TODO: GH#18217
                        if (ts.isBlock(body)) {
                            return body.statements;
                        }
                    }
                    else if (ts.isModuleBlock(scope) || ts.isSourceFile(scope)) {
                        return scope.statements;
                    }
                    else if (ts.isClassLike(scope)) {
                        return scope.members;
                    }
                    else {
                        ts.assertType(scope);
                    }
                    return ts.emptyArray;
                }
                /**
                 * If `scope` contains a function after `minPos`, then return the first such function.
                 * Otherwise, return `undefined`.
                 */
                function getNodeToInsertFunctionBefore(minPos, scope) {
                    return ts.find(getStatementsOrClassElements(scope), function (child) {
                        return child.pos >= minPos && ts.isFunctionLikeDeclaration(child) && !ts.isConstructorDeclaration(child);
                    });
                }
                function getNodeToInsertPropertyBefore(maxPos, scope) {
                    var members = scope.members;
                    ts.Debug.assert(members.length > 0, "Found no members"); // There must be at least one child, since we extracted from one.
                    var prevMember;
                    var allProperties = true;
                    for (var _i = 0, members_1 = members; _i < members_1.length; _i++) {
                        var member = members_1[_i];
                        if (member.pos > maxPos) {
                            return prevMember || members[0];
                        }
                        if (allProperties && !ts.isPropertyDeclaration(member)) {
                            // If it is non-vacuously true that all preceding members are properties,
                            // insert before the current member (i.e. at the end of the list of properties).
                            if (prevMember !== undefined) {
                                return member;
                            }
                            allProperties = false;
                        }
                        prevMember = member;
                    }
                    if (prevMember === undefined)
                        return ts.Debug.fail(); // If the loop didn't return, then it did set prevMember.
                    return prevMember;
                }
                function getNodeToInsertConstantBefore(node, scope) {
                    ts.Debug.assert(!ts.isClassLike(scope));
                    var prevScope;
                    for (var curr = node; curr !== scope; curr = curr.parent) {
                        if (isScope(curr)) {
                            prevScope = curr;
                        }
                    }
                    for (var curr = (prevScope || node).parent;; curr = curr.parent) {
                        if (isBlockLike(curr)) {
                            var prevStatement = void 0;
                            for (var _i = 0, _a = curr.statements; _i < _a.length; _i++) {
                                var statement = _a[_i];
                                if (statement.pos > node.pos) {
                                    break;
                                }
                                prevStatement = statement;
                            }
                            if (!prevStatement && ts.isCaseClause(curr)) {
                                // We must have been in the expression of the case clause.
                                ts.Debug.assert(ts.isSwitchStatement(curr.parent.parent), "Grandparent isn't a switch statement");
                                return curr.parent.parent;
                            }
                            // There must be at least one statement since we started in one.
                            return ts.Debug.checkDefined(prevStatement, "prevStatement failed to get set");
                        }
                        ts.Debug.assert(curr !== scope, "Didn't encounter a block-like before encountering scope");
                    }
                }
                function getPropertyAssignmentsForWritesAndVariableDeclarations(exposedVariableDeclarations, writes) {
                    var variableAssignments = ts.map(exposedVariableDeclarations, function (v) { return ts.factory.createShorthandPropertyAssignment(v.symbol.name); });
                    var writeAssignments = ts.map(writes, function (w) { return ts.factory.createShorthandPropertyAssignment(w.symbol.name); });
                    // TODO: GH#18217 `variableAssignments` not possibly undefined!
                    return variableAssignments === undefined
                        ? writeAssignments
                        : writeAssignments === undefined
                            ? variableAssignments
                            : variableAssignments.concat(writeAssignments);
                }
                function isReadonlyArray(v) {
                    return ts.isArray(v);
                }
                /**
                 * Produces a range that spans the entirety of nodes, given a selection
                 * that might start/end in the middle of nodes.
                 *
                 * For example, when the user makes a selection like this
                 *                     v---v
                 *   var someThing = foo + bar;
                 *  this returns     ^-------^
                 */
                function getEnclosingTextRange(targetRange, sourceFile) {
                    return isReadonlyArray(targetRange.range)
                        ? { pos: ts.first(targetRange.range).getStart(sourceFile), end: ts.last(targetRange.range).getEnd() }
                        : targetRange.range;
                }
                var Usage;
                (function (Usage) {
                    // value should be passed to extracted method
                    Usage[Usage["Read"] = 1] = "Read";
                    // value should be passed to extracted method and propagated back
                    Usage[Usage["Write"] = 2] = "Write";
                })(Usage || (Usage = {}));
                function collectReadsAndWrites(targetRange, scopes, enclosingTextRange, sourceFile, checker, cancellationToken) {
                    var allTypeParameterUsages = new ts.Map(); // Key is type ID
                    var usagesPerScope = [];
                    var substitutionsPerScope = [];
                    var functionErrorsPerScope = [];
                    var constantErrorsPerScope = [];
                    var visibleDeclarationsInExtractedRange = [];
                    var exposedVariableSymbolSet = new ts.Map(); // Key is symbol ID
                    var exposedVariableDeclarations = [];
                    var firstExposedNonVariableDeclaration;
                    var expression = !isReadonlyArray(targetRange.range)
                        ? targetRange.range
                        : targetRange.range.length === 1 && ts.isExpressionStatement(targetRange.range[0])
                            ? targetRange.range[0].expression
                            : undefined;
                    var expressionDiagnostic;
                    if (expression === undefined) {
                        var statements = targetRange.range;
                        var start = ts.first(statements).getStart();
                        var end = ts.last(statements).end;
                        expressionDiagnostic = ts.createFileDiagnostic(sourceFile, start, end - start, Messages.expressionExpected);
                    }
                    else if (checker.getTypeAtLocation(expression).flags & (16384 /* Void */ | 131072 /* Never */)) {
                        expressionDiagnostic = ts.createDiagnosticForNode(expression, Messages.uselessConstantType);
                    }
                    // initialize results
                    for (var _i = 0, scopes_1 = scopes; _i < scopes_1.length; _i++) {
                        var scope = scopes_1[_i];
                        usagesPerScope.push({ usages: new ts.Map(), typeParameterUsages: new ts.Map(), substitutions: new ts.Map() });
                        substitutionsPerScope.push(new ts.Map());
                        functionErrorsPerScope.push([]);
                        var constantErrors = [];
                        if (expressionDiagnostic) {
                            constantErrors.push(expressionDiagnostic);
                        }
                        if (ts.isClassLike(scope) && ts.isInJSFile(scope)) {
                            constantErrors.push(ts.createDiagnosticForNode(scope, Messages.cannotExtractToJSClass));
                        }
                        if (ts.isArrowFunction(scope) && !ts.isBlock(scope.body)) {
                            // TODO (https://github.com/Microsoft/TypeScript/issues/18924): allow this
                            constantErrors.push(ts.createDiagnosticForNode(scope, Messages.cannotExtractToExpressionArrowFunction));
                        }
                        constantErrorsPerScope.push(constantErrors);
                    }
                    var seenUsages = new ts.Map();
                    var target = isReadonlyArray(targetRange.range) ? ts.factory.createBlock(targetRange.range) : targetRange.range;
                    var unmodifiedNode = isReadonlyArray(targetRange.range) ? ts.first(targetRange.range) : targetRange.range;
                    var inGenericContext = isInGenericContext(unmodifiedNode);
                    collectUsages(target);
                    // Unfortunately, this code takes advantage of the knowledge that the generated method
                    // will use the contextual type of an expression as the return type of the extracted
                    // method (and will therefore "use" all the types involved).
                    if (inGenericContext && !isReadonlyArray(targetRange.range)) {
                        var contextualType = checker.getContextualType(targetRange.range); // TODO: GH#18217
                        recordTypeParameterUsages(contextualType);
                    }
                    if (allTypeParameterUsages.size > 0) {
                        var seenTypeParameterUsages = new ts.Map(); // Key is type ID
                        var i_2 = 0;
                        for (var curr = unmodifiedNode; curr !== undefined && i_2 < scopes.length; curr = curr.parent) {
                            if (curr === scopes[i_2]) {
                                // Copy current contents of seenTypeParameterUsages into scope.
                                seenTypeParameterUsages.forEach(function (typeParameter, id) {
                                    usagesPerScope[i_2].typeParameterUsages.set(id, typeParameter);
                                });
                                i_2++;
                            }
                            // Note that we add the current node's type parameters *after* updating the corresponding scope.
                            if (ts.isDeclarationWithTypeParameters(curr)) {
                                for (var _a = 0, _b = ts.getEffectiveTypeParameterDeclarations(curr); _a < _b.length; _a++) {
                                    var typeParameterDecl = _b[_a];
                                    var typeParameter = checker.getTypeAtLocation(typeParameterDecl);
                                    if (allTypeParameterUsages.has(typeParameter.id.toString())) {
                                        seenTypeParameterUsages.set(typeParameter.id.toString(), typeParameter);
                                    }
                                }
                            }
                        }
                        // If we didn't get through all the scopes, then there were some that weren't in our
                        // parent chain (impossible at time of writing).  A conservative solution would be to
                        // copy allTypeParameterUsages into all remaining scopes.
                        ts.Debug.assert(i_2 === scopes.length, "Should have iterated all scopes");
                    }
                    // If there are any declarations in the extracted block that are used in the same enclosing
                    // lexical scope, we can't move the extraction "up" as those declarations will become unreachable
                    if (visibleDeclarationsInExtractedRange.length) {
                        var containingLexicalScopeOfExtraction = ts.isBlockScope(scopes[0], scopes[0].parent)
                            ? scopes[0]
                            : ts.getEnclosingBlockScopeContainer(scopes[0]);
                        ts.forEachChild(containingLexicalScopeOfExtraction, checkForUsedDeclarations);
                    }
                    var _loop_17 = function (i) {
                        var scopeUsages = usagesPerScope[i];
                        // Special case: in the innermost scope, all usages are available.
                        // (The computed value reflects the value at the top-level of the scope, but the
                        // local will actually be declared at the same level as the extracted expression).
                        if (i > 0 && (scopeUsages.usages.size > 0 || scopeUsages.typeParameterUsages.size > 0)) {
                            var errorNode = isReadonlyArray(targetRange.range) ? targetRange.range[0] : targetRange.range;
                            constantErrorsPerScope[i].push(ts.createDiagnosticForNode(errorNode, Messages.cannotAccessVariablesFromNestedScopes));
                        }
                        var hasWrite = false;
                        var readonlyClassPropertyWrite;
                        usagesPerScope[i].usages.forEach(function (value) {
                            if (value.usage === 2 /* Write */) {
                                hasWrite = true;
                                if (value.symbol.flags & 106500 /* ClassMember */ &&
                                    value.symbol.valueDeclaration &&
                                    ts.hasEffectiveModifier(value.symbol.valueDeclaration, 64 /* Readonly */)) {
                                    readonlyClassPropertyWrite = value.symbol.valueDeclaration;
                                }
                            }
                        });
                        // If an expression was extracted, then there shouldn't have been any variable declarations.
                        ts.Debug.assert(isReadonlyArray(targetRange.range) || exposedVariableDeclarations.length === 0, "No variable declarations expected if something was extracted");
                        if (hasWrite && !isReadonlyArray(targetRange.range)) {
                            var diag = ts.createDiagnosticForNode(targetRange.range, Messages.cannotWriteInExpression);
                            functionErrorsPerScope[i].push(diag);
                            constantErrorsPerScope[i].push(diag);
                        }
                        else if (readonlyClassPropertyWrite && i > 0) {
                            var diag = ts.createDiagnosticForNode(readonlyClassPropertyWrite, Messages.cannotExtractReadonlyPropertyInitializerOutsideConstructor);
                            functionErrorsPerScope[i].push(diag);
                            constantErrorsPerScope[i].push(diag);
                        }
                        else if (firstExposedNonVariableDeclaration) {
                            var diag = ts.createDiagnosticForNode(firstExposedNonVariableDeclaration, Messages.cannotExtractExportedEntity);
                            functionErrorsPerScope[i].push(diag);
                            constantErrorsPerScope[i].push(diag);
                        }
                    };
                    for (var i = 0; i < scopes.length; i++) {
                        _loop_17(i);
                    }
                    return { target: target, usagesPerScope: usagesPerScope, functionErrorsPerScope: functionErrorsPerScope, constantErrorsPerScope: constantErrorsPerScope, exposedVariableDeclarations: exposedVariableDeclarations };
                    function isInGenericContext(node) {
                        return !!ts.findAncestor(node, function (n) { return ts.isDeclarationWithTypeParameters(n) && ts.getEffectiveTypeParameterDeclarations(n).length !== 0; });
                    }
                    function recordTypeParameterUsages(type) {
                        // PERF: This is potentially very expensive.  `type` could be a library type with
                        // a lot of properties, each of which the walker will visit.  Unfortunately, the
                        // solution isn't as trivial as filtering to user types because of (e.g.) Array.
                        var symbolWalker = checker.getSymbolWalker(function () { return (cancellationToken.throwIfCancellationRequested(), true); });
                        var visitedTypes = symbolWalker.walkType(type).visitedTypes;
                        for (var _i = 0, visitedTypes_1 = visitedTypes; _i < visitedTypes_1.length; _i++) {
                            var visitedType = visitedTypes_1[_i];
                            if (visitedType.isTypeParameter()) {
                                allTypeParameterUsages.set(visitedType.id.toString(), visitedType);
                            }
                        }
                    }
                    function collectUsages(node, valueUsage) {
                        if (valueUsage === void 0) {
                            valueUsage = 1 /* Read */;
                        }
                        if (inGenericContext) {
                            var type = checker.getTypeAtLocation(node);
                            recordTypeParameterUsages(type);
                        }
                        if (ts.isDeclaration(node) && node.symbol) {
                            visibleDeclarationsInExtractedRange.push(node);
                        }
                        if (ts.isAssignmentExpression(node)) {
                            // use 'write' as default usage for values
                            collectUsages(node.left, 2 /* Write */);
                            collectUsages(node.right);
                        }
                        else if (ts.isUnaryExpressionWithWrite(node)) {
                            collectUsages(node.operand, 2 /* Write */);
                        }
                        else if (ts.isPropertyAccessExpression(node) || ts.isElementAccessExpression(node)) {
                            // use 'write' as default usage for values
                            ts.forEachChild(node, collectUsages);
                        }
                        else if (ts.isIdentifier(node)) {
                            if (!node.parent) {
                                return;
                            }
                            if (ts.isQualifiedName(node.parent) && node !== node.parent.left) {
                                return;
                            }
                            if (ts.isPropertyAccessExpression(node.parent) && node !== node.parent.expression) {
                                return;
                            }
                            recordUsage(node, valueUsage, /*isTypeNode*/ ts.isPartOfTypeNode(node));
                        }
                        else {
                            ts.forEachChild(node, collectUsages);
                        }
                    }
                    function recordUsage(n, usage, isTypeNode) {
                        var symbolId = recordUsagebySymbol(n, usage, isTypeNode);
                        if (symbolId) {
                            for (var i = 0; i < scopes.length; i++) {
                                // push substitution from map<symbolId, subst> to map<nodeId, subst> to simplify rewriting
                                var substitution = substitutionsPerScope[i].get(symbolId);
                                if (substitution) {
                                    usagesPerScope[i].substitutions.set(ts.getNodeId(n).toString(), substitution);
                                }
                            }
                        }
                    }
                    function recordUsagebySymbol(identifier, usage, isTypeName) {
                        var symbol = getSymbolReferencedByIdentifier(identifier);
                        if (!symbol) {
                            // cannot find symbol - do nothing
                            return undefined;
                        }
                        var symbolId = ts.getSymbolId(symbol).toString();
                        var lastUsage = seenUsages.get(symbolId);
                        // there are two kinds of value usages
                        // - reads - if range contains a read from the value located outside of the range then value should be passed as a parameter
                        // - writes - if range contains a write to a value located outside the range the value should be passed as a parameter and
                        //   returned as a return value
                        // 'write' case is a superset of 'read' so if we already have processed 'write' of some symbol there is not need to handle 'read'
                        // since all information is already recorded
                        if (lastUsage && lastUsage >= usage) {
                            return symbolId;
                        }
                        seenUsages.set(symbolId, usage);
                        if (lastUsage) {
                            // if we get here this means that we are trying to handle 'write' and 'read' was already processed
                            // walk scopes and update existing records.
                            for (var _i = 0, usagesPerScope_1 = usagesPerScope; _i < usagesPerScope_1.length; _i++) {
                                var perScope = usagesPerScope_1[_i];
                                var prevEntry = perScope.usages.get(identifier.text);
                                if (prevEntry) {
                                    perScope.usages.set(identifier.text, { usage: usage, symbol: symbol, node: identifier });
                                }
                            }
                            return symbolId;
                        }
                        // find first declaration in this file
                        var decls = symbol.getDeclarations();
                        var declInFile = decls && ts.find(decls, function (d) { return d.getSourceFile() === sourceFile; });
                        if (!declInFile) {
                            return undefined;
                        }
                        if (ts.rangeContainsStartEnd(enclosingTextRange, declInFile.getStart(), declInFile.end)) {
                            // declaration is located in range to be extracted - do nothing
                            return undefined;
                        }
                        if (targetRange.facts & RangeFacts.IsGenerator && usage === 2 /* Write */) {
                            // this is write to a reference located outside of the target scope and range is extracted into generator
                            // currently this is unsupported scenario
                            var diag = ts.createDiagnosticForNode(identifier, Messages.cannotExtractRangeThatContainsWritesToReferencesLocatedOutsideOfTheTargetRangeInGenerators);
                            for (var _a = 0, functionErrorsPerScope_1 = functionErrorsPerScope; _a < functionErrorsPerScope_1.length; _a++) {
                                var errors = functionErrorsPerScope_1[_a];
                                errors.push(diag);
                            }
                            for (var _b = 0, constantErrorsPerScope_1 = constantErrorsPerScope; _b < constantErrorsPerScope_1.length; _b++) {
                                var errors = constantErrorsPerScope_1[_b];
                                errors.push(diag);
                            }
                        }
                        for (var i = 0; i < scopes.length; i++) {
                            var scope = scopes[i];
                            var resolvedSymbol = checker.resolveName(symbol.name, scope, symbol.flags, /*excludeGlobals*/ false);
                            if (resolvedSymbol === symbol) {
                                continue;
                            }
                            if (!substitutionsPerScope[i].has(symbolId)) {
                                var substitution = tryReplaceWithQualifiedNameOrPropertyAccess(symbol.exportSymbol || symbol, scope, isTypeName);
                                if (substitution) {
                                    substitutionsPerScope[i].set(symbolId, substitution);
                                }
                                else if (isTypeName) {
                                    // If the symbol is a type parameter that won't be in scope, we'll pass it as a type argument
                                    // so there's no problem.
                                    if (!(symbol.flags & 262144 /* TypeParameter */)) {
                                        var diag = ts.createDiagnosticForNode(identifier, Messages.typeWillNotBeVisibleInTheNewScope);
                                        functionErrorsPerScope[i].push(diag);
                                        constantErrorsPerScope[i].push(diag);
                                    }
                                }
                                else {
                                    usagesPerScope[i].usages.set(identifier.text, { usage: usage, symbol: symbol, node: identifier });
                                }
                            }
                        }
                        return symbolId;
                    }
                    function checkForUsedDeclarations(node) {
                        // If this node is entirely within the original extraction range, we don't need to do anything.
                        if (node === targetRange.range || (isReadonlyArray(targetRange.range) && targetRange.range.indexOf(node) >= 0)) {
                            return;
                        }
                        // Otherwise check and recurse.
                        var sym = ts.isIdentifier(node)
                            ? getSymbolReferencedByIdentifier(node)
                            : checker.getSymbolAtLocation(node);
                        if (sym) {
                            var decl = ts.find(visibleDeclarationsInExtractedRange, function (d) { return d.symbol === sym; });
                            if (decl) {
                                if (ts.isVariableDeclaration(decl)) {
                                    var idString = decl.symbol.id.toString();
                                    if (!exposedVariableSymbolSet.has(idString)) {
                                        exposedVariableDeclarations.push(decl);
                                        exposedVariableSymbolSet.set(idString, true);
                                    }
                                }
                                else {
                                    // CONSIDER: this includes binding elements, which we could
                                    // expose in the same way as variables.
                                    firstExposedNonVariableDeclaration = firstExposedNonVariableDeclaration || decl;
                                }
                            }
                        }
                        ts.forEachChild(node, checkForUsedDeclarations);
                    }
                    /**
                     * Return the symbol referenced by an identifier (even if it declares a different symbol).
                     */
                    function getSymbolReferencedByIdentifier(identifier) {
                        // If the identifier is both a property name and its value, we're only interested in its value
                        // (since the name is a declaration and will be included in the extracted range).
                        return identifier.parent && ts.isShorthandPropertyAssignment(identifier.parent) && identifier.parent.name === identifier
                            ? checker.getShorthandAssignmentValueSymbol(identifier.parent)
                            : checker.getSymbolAtLocation(identifier);
                    }
                    function tryReplaceWithQualifiedNameOrPropertyAccess(symbol, scopeDecl, isTypeNode) {
                        if (!symbol) {
                            return undefined;
                        }
                        var decls = symbol.getDeclarations();
                        if (decls && decls.some(function (d) { return d.parent === scopeDecl; })) {
                            return ts.factory.createIdentifier(symbol.name);
                        }
                        var prefix = tryReplaceWithQualifiedNameOrPropertyAccess(symbol.parent, scopeDecl, isTypeNode);
                        if (prefix === undefined) {
                            return undefined;
                        }
                        return isTypeNode
                            ? ts.factory.createQualifiedName(prefix, ts.factory.createIdentifier(symbol.name))
                            : ts.factory.createPropertyAccessExpression(prefix, symbol.name);
                    }
                }
                function getExtractableParent(node) {
                    return ts.findAncestor(node, function (node) { return node.parent && isExtractableExpression(node) && !ts.isBinaryExpression(node.parent); });
                }
                /**
                 * Computes whether or not a node represents an expression in a position where it could
                 * be extracted.
                 * The isExpression() in utilities.ts returns some false positives we need to handle,
                 * such as `import x from 'y'` -- the 'y' is a StringLiteral but is *not* an expression
                 * in the sense of something that you could extract on
                 */
                function isExtractableExpression(node) {
                    var parent = node.parent;
                    switch (parent.kind) {
                        case 294 /* EnumMember */:
                            return false;
                    }
                    switch (node.kind) {
                        case 10 /* StringLiteral */:
                            return parent.kind !== 264 /* ImportDeclaration */ &&
                                parent.kind !== 268 /* ImportSpecifier */;
                        case 223 /* SpreadElement */:
                        case 199 /* ObjectBindingPattern */:
                        case 201 /* BindingElement */:
                            return false;
                        case 79 /* Identifier */:
                            return parent.kind !== 201 /* BindingElement */ &&
                                parent.kind !== 268 /* ImportSpecifier */ &&
                                parent.kind !== 273 /* ExportSpecifier */;
                    }
                    return true;
                }
                function isBlockLike(node) {
                    switch (node.kind) {
                        case 233 /* Block */:
                        case 300 /* SourceFile */:
                        case 260 /* ModuleBlock */:
                        case 287 /* CaseClause */:
                            return true;
                        default:
                            return false;
                    }
                }
                function isInJSXContent(node) {
                    return (ts.isJsxElement(node) || ts.isJsxSelfClosingElement(node) || ts.isJsxFragment(node)) && (ts.isJsxElement(node.parent) || ts.isJsxFragment(node.parent));
                }
            })(extractSymbol = refactor.extractSymbol || (refactor.extractSymbol = {}));
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var refactorName = "Extract type";
            var extractToTypeAliasAction = {
                name: "Extract to type alias",
                description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_type_alias),
                kind: "refactor.extract.type",
            };
            var extractToInterfaceAction = {
                name: "Extract to interface",
                description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_interface),
                kind: "refactor.extract.interface",
            };
            var extractToTypeDefAction = {
                name: "Extract to typedef",
                description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_to_typedef),
                kind: "refactor.extract.typedef"
            };
            refactor.registerRefactor(refactorName, {
                kinds: [
                    extractToTypeAliasAction.kind,
                    extractToInterfaceAction.kind,
                    extractToTypeDefAction.kind
                ],
                getAvailableActions: function (context) {
                    var info = getRangeToExtract(context, context.triggerReason === "invoked");
                    if (!info)
                        return ts.emptyArray;
                    if (!refactor.isRefactorErrorInfo(info)) {
                        return [{
                                name: refactorName,
                                description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_type),
                                actions: info.isJS ?
                                    [extractToTypeDefAction] : ts.append([extractToTypeAliasAction], info.typeElements && extractToInterfaceAction)
                            }];
                    }
                    if (context.preferences.provideRefactorNotApplicableReason) {
                        return [{
                                name: refactorName,
                                description: ts.getLocaleSpecificMessage(ts.Diagnostics.Extract_type),
                                actions: [
                                    __assign(__assign({}, extractToTypeDefAction), { notApplicableReason: info.error }),
                                    __assign(__assign({}, extractToTypeAliasAction), { notApplicableReason: info.error }),
                                    __assign(__assign({}, extractToInterfaceAction), { notApplicableReason: info.error }),
                                ]
                            }];
                    }
                    return ts.emptyArray;
                },
                getEditsForAction: function (context, actionName) {
                    var file = context.file;
                    var info = getRangeToExtract(context);
                    ts.Debug.assert(info && !refactor.isRefactorErrorInfo(info), "Expected to find a range to extract");
                    var name = ts.getUniqueName("NewType", file);
                    var edits = ts.textChanges.ChangeTracker.with(context, function (changes) {
                        switch (actionName) {
                            case extractToTypeAliasAction.name:
                                ts.Debug.assert(!info.isJS, "Invalid actionName/JS combo");
                                return doTypeAliasChange(changes, file, name, info);
                            case extractToTypeDefAction.name:
                                ts.Debug.assert(info.isJS, "Invalid actionName/JS combo");
                                return doTypedefChange(changes, file, name, info);
                            case extractToInterfaceAction.name:
                                ts.Debug.assert(!info.isJS && !!info.typeElements, "Invalid actionName/JS combo");
                                return doInterfaceChange(changes, file, name, info);
                            default:
                                ts.Debug.fail("Unexpected action name");
                        }
                    });
                    var renameFilename = file.fileName;
                    var renameLocation = ts.getRenameLocation(edits, renameFilename, name, /*preferLastLocation*/ false);
                    return { edits: edits, renameFilename: renameFilename, renameLocation: renameLocation };
                }
            });
            function getRangeToExtract(context, considerEmptySpans) {
                if (considerEmptySpans === void 0) {
                    considerEmptySpans = true;
                }
                var file = context.file, startPosition = context.startPosition;
                var isJS = ts.isSourceFileJS(file);
                var current = ts.getTokenAtPosition(file, startPosition);
                var range = ts.createTextRangeFromSpan(ts.getRefactorContextSpan(context));
                var cursorRequest = range.pos === range.end && considerEmptySpans;
                var selection = ts.findAncestor(current, (function (node) {
                    return node.parent && ts.isTypeNode(node) && !rangeContainsSkipTrivia(range, node.parent, file) &&
                        (cursorRequest || ts.nodeOverlapsWithStartEnd(current, file, range.pos, range.end));
                }));
                if (!selection || !ts.isTypeNode(selection))
                    return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Selection_is_not_a_valid_type_node) };
                var checker = context.program.getTypeChecker();
                var firstStatement = ts.Debug.checkDefined(ts.findAncestor(selection, ts.isStatement), "Should find a statement");
                var typeParameters = collectTypeParameters(checker, selection, firstStatement, file);
                if (!typeParameters)
                    return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.No_type_could_be_extracted_from_this_type_node) };
                var typeElements = flattenTypeLiteralNodeReference(checker, selection);
                return { isJS: isJS, selection: selection, firstStatement: firstStatement, typeParameters: typeParameters, typeElements: typeElements };
            }
            function flattenTypeLiteralNodeReference(checker, node) {
                if (!node)
                    return undefined;
                if (ts.isIntersectionTypeNode(node)) {
                    var result = [];
                    var seen_1 = new ts.Map();
                    for (var _i = 0, _a = node.types; _i < _a.length; _i++) {
                        var type = _a[_i];
                        var flattenedTypeMembers = flattenTypeLiteralNodeReference(checker, type);
                        if (!flattenedTypeMembers || !flattenedTypeMembers.every(function (type) { return type.name && ts.addToSeen(seen_1, ts.getNameFromPropertyName(type.name)); })) {
                            return undefined;
                        }
                        ts.addRange(result, flattenedTypeMembers);
                    }
                    return result;
                }
                else if (ts.isParenthesizedTypeNode(node)) {
                    return flattenTypeLiteralNodeReference(checker, node.type);
                }
                else if (ts.isTypeLiteralNode(node)) {
                    return node.members;
                }
                return undefined;
            }
            function rangeContainsSkipTrivia(r1, node, file) {
                return ts.rangeContainsStartEnd(r1, ts.skipTrivia(file.text, node.pos), node.end);
            }
            function collectTypeParameters(checker, selection, statement, file) {
                var result = [];
                return visitor(selection) ? undefined : result;
                function visitor(node) {
                    if (ts.isTypeReferenceNode(node)) {
                        if (ts.isIdentifier(node.typeName)) {
                            var symbol = checker.resolveName(node.typeName.text, node.typeName, 262144 /* TypeParameter */, /* excludeGlobals */ true);
                            if (symbol === null || symbol === void 0 ? void 0 : symbol.declarations) {
                                var declaration = ts.cast(ts.first(symbol.declarations), ts.isTypeParameterDeclaration);
                                if (rangeContainsSkipTrivia(statement, declaration, file) && !rangeContainsSkipTrivia(selection, declaration, file)) {
                                    ts.pushIfUnique(result, declaration);
                                }
                            }
                        }
                    }
                    else if (ts.isInferTypeNode(node)) {
                        var conditionalTypeNode = ts.findAncestor(node, function (n) { return ts.isConditionalTypeNode(n) && rangeContainsSkipTrivia(n.extendsType, node, file); });
                        if (!conditionalTypeNode || !rangeContainsSkipTrivia(selection, conditionalTypeNode, file)) {
                            return true;
                        }
                    }
                    else if ((ts.isTypePredicateNode(node) || ts.isThisTypeNode(node))) {
                        var functionLikeNode = ts.findAncestor(node.parent, ts.isFunctionLike);
                        if (functionLikeNode && functionLikeNode.type && rangeContainsSkipTrivia(functionLikeNode.type, node, file) && !rangeContainsSkipTrivia(selection, functionLikeNode, file)) {
                            return true;
                        }
                    }
                    else if (ts.isTypeQueryNode(node)) {
                        if (ts.isIdentifier(node.exprName)) {
                            var symbol = checker.resolveName(node.exprName.text, node.exprName, 111551 /* Value */, /* excludeGlobals */ false);
                            if ((symbol === null || symbol === void 0 ? void 0 : symbol.valueDeclaration) && rangeContainsSkipTrivia(statement, symbol.valueDeclaration, file) && !rangeContainsSkipTrivia(selection, symbol.valueDeclaration, file)) {
                                return true;
                            }
                        }
                        else {
                            if (ts.isThisIdentifier(node.exprName.left) && !rangeContainsSkipTrivia(selection, node.parent, file)) {
                                return true;
                            }
                        }
                    }
                    if (file && ts.isTupleTypeNode(node) && (ts.getLineAndCharacterOfPosition(file, node.pos).line === ts.getLineAndCharacterOfPosition(file, node.end).line)) {
                        ts.setEmitFlags(node, 1 /* SingleLine */);
                    }
                    return ts.forEachChild(node, visitor);
                }
            }
            function doTypeAliasChange(changes, file, name, info) {
                var firstStatement = info.firstStatement, selection = info.selection, typeParameters = info.typeParameters;
                var newTypeNode = ts.factory.createTypeAliasDeclaration(
                /* decorators */ undefined, 
                /* modifiers */ undefined, name, typeParameters.map(function (id) { return ts.factory.updateTypeParameterDeclaration(id, id.name, id.constraint, /* defaultType */ undefined); }), selection);
                changes.insertNodeBefore(file, firstStatement, ts.ignoreSourceNewlines(newTypeNode), /* blankLineBetween */ true);
                changes.replaceNode(file, selection, ts.factory.createTypeReferenceNode(name, typeParameters.map(function (id) { return ts.factory.createTypeReferenceNode(id.name, /* typeArguments */ undefined); })), { leadingTriviaOption: ts.textChanges.LeadingTriviaOption.Exclude, trailingTriviaOption: ts.textChanges.TrailingTriviaOption.ExcludeWhitespace });
            }
            function doInterfaceChange(changes, file, name, info) {
                var _a;
                var firstStatement = info.firstStatement, selection = info.selection, typeParameters = info.typeParameters, typeElements = info.typeElements;
                var newTypeNode = ts.factory.createInterfaceDeclaration(
                /* decorators */ undefined, 
                /* modifiers */ undefined, name, typeParameters, 
                /* heritageClauses */ undefined, typeElements);
                ts.setTextRange(newTypeNode, (_a = typeElements[0]) === null || _a === void 0 ? void 0 : _a.parent);
                changes.insertNodeBefore(file, firstStatement, ts.ignoreSourceNewlines(newTypeNode), /* blankLineBetween */ true);
                changes.replaceNode(file, selection, ts.factory.createTypeReferenceNode(name, typeParameters.map(function (id) { return ts.factory.createTypeReferenceNode(id.name, /* typeArguments */ undefined); })), { leadingTriviaOption: ts.textChanges.LeadingTriviaOption.Exclude, trailingTriviaOption: ts.textChanges.TrailingTriviaOption.ExcludeWhitespace });
            }
            function doTypedefChange(changes, file, name, info) {
                var firstStatement = info.firstStatement, selection = info.selection, typeParameters = info.typeParameters;
                var node = ts.factory.createJSDocTypedefTag(ts.factory.createIdentifier("typedef"), ts.factory.createJSDocTypeExpression(selection), ts.factory.createIdentifier(name));
                var templates = [];
                ts.forEach(typeParameters, function (typeParameter) {
                    var constraint = ts.getEffectiveConstraintOfTypeParameter(typeParameter);
                    var parameter = ts.factory.createTypeParameterDeclaration(typeParameter.name);
                    var template = ts.factory.createJSDocTemplateTag(ts.factory.createIdentifier("template"), constraint && ts.cast(constraint, ts.isJSDocTypeExpression), [parameter]);
                    templates.push(template);
                });
                changes.insertNodeBefore(file, firstStatement, ts.factory.createJSDocComment(/* comment */ undefined, ts.factory.createNodeArray(ts.concatenate(templates, [node]))), /* blankLineBetween */ true);
                changes.replaceNode(file, selection, ts.factory.createTypeReferenceNode(name, typeParameters.map(function (id) { return ts.factory.createTypeReferenceNode(id.name, /* typeArguments */ undefined); })));
            }
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var generateGetAccessorAndSetAccessor;
            (function (generateGetAccessorAndSetAccessor) {
                var actionName = "Generate 'get' and 'set' accessors";
                var actionDescription = ts.Diagnostics.Generate_get_and_set_accessors.message;
                var generateGetSetAction = {
                    name: actionName,
                    description: actionDescription,
                    kind: "refactor.rewrite.property.generateAccessors",
                };
                refactor.registerRefactor(actionName, {
                    kinds: [generateGetSetAction.kind],
                    getEditsForAction: function (context, actionName) {
                        if (!context.endPosition)
                            return undefined;
                        var info = ts.codefix.getAccessorConvertiblePropertyAtPosition(context.file, context.program, context.startPosition, context.endPosition);
                        ts.Debug.assert(info && !refactor.isRefactorErrorInfo(info), "Expected applicable refactor info");
                        var edits = ts.codefix.generateAccessorFromProperty(context.file, context.program, context.startPosition, context.endPosition, context, actionName);
                        if (!edits)
                            return undefined;
                        var renameFilename = context.file.fileName;
                        var nameNeedRename = info.renameAccessor ? info.accessorName : info.fieldName;
                        var renameLocationOffset = ts.isIdentifier(nameNeedRename) ? 0 : -1;
                        var renameLocation = renameLocationOffset + ts.getRenameLocation(edits, renameFilename, nameNeedRename.text, /*preferLastLocation*/ ts.isParameter(info.declaration));
                        return { renameFilename: renameFilename, renameLocation: renameLocation, edits: edits };
                    },
                    getAvailableActions: function (context) {
                        if (!context.endPosition)
                            return ts.emptyArray;
                        var info = ts.codefix.getAccessorConvertiblePropertyAtPosition(context.file, context.program, context.startPosition, context.endPosition, context.triggerReason === "invoked");
                        if (!info)
                            return ts.emptyArray;
                        if (!refactor.isRefactorErrorInfo(info)) {
                            return [{
                                    name: actionName,
                                    description: actionDescription,
                                    actions: [generateGetSetAction],
                                }];
                        }
                        if (context.preferences.provideRefactorNotApplicableReason) {
                            return [{
                                    name: actionName,
                                    description: actionDescription,
                                    actions: [__assign(__assign({}, generateGetSetAction), { notApplicableReason: info.error })],
                                }];
                        }
                        return ts.emptyArray;
                    }
                });
            })(generateGetAccessorAndSetAccessor = refactor.generateGetAccessorAndSetAccessor || (refactor.generateGetAccessorAndSetAccessor = {}));
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            ;
            /**
             * Checks if some refactor info has refactor error info.
             */
            function isRefactorErrorInfo(info) {
                return info.error !== undefined;
            }
            refactor.isRefactorErrorInfo = isRefactorErrorInfo;
            /**
             * Checks if string "known" begins with string "requested".
             * Used to match requested kinds with a known kind.
             */
            function refactorKindBeginsWith(known, requested) {
                if (!requested)
                    return true;
                return known.substr(0, requested.length) === requested;
            }
            refactor.refactorKindBeginsWith = refactorKindBeginsWith;
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var refactorName = "Move to a new file";
            var description = ts.getLocaleSpecificMessage(ts.Diagnostics.Move_to_a_new_file);
            var moveToNewFileAction = {
                name: refactorName,
                description: description,
                kind: "refactor.move.newFile",
            };
            refactor.registerRefactor(refactorName, {
                kinds: [moveToNewFileAction.kind],
                getAvailableActions: function (context) {
                    var statements = getStatementsToMove(context);
                    if (context.preferences.allowTextChangesInNewFiles && statements) {
                        return [{ name: refactorName, description: description, actions: [moveToNewFileAction] }];
                    }
                    if (context.preferences.provideRefactorNotApplicableReason) {
                        return [{ name: refactorName, description: description, actions: [__assign(__assign({}, moveToNewFileAction), { notApplicableReason: ts.getLocaleSpecificMessage(ts.Diagnostics.Selection_is_not_a_valid_statement_or_statements) })]
                            }];
                    }
                    return ts.emptyArray;
                },
                getEditsForAction: function (context, actionName) {
                    ts.Debug.assert(actionName === refactorName, "Wrong refactor invoked");
                    var statements = ts.Debug.checkDefined(getStatementsToMove(context));
                    var edits = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(context.file, context.program, statements, t, context.host, context.preferences); });
                    return { edits: edits, renameFilename: undefined, renameLocation: undefined };
                }
            });
            function getRangeToMove(context) {
                var file = context.file;
                var range = ts.createTextRangeFromSpan(ts.getRefactorContextSpan(context));
                var statements = file.statements;
                var startNodeIndex = ts.findIndex(statements, function (s) { return s.end > range.pos; });
                if (startNodeIndex === -1)
                    return undefined;
                var startStatement = statements[startNodeIndex];
                if (ts.isNamedDeclaration(startStatement) && startStatement.name && ts.rangeContainsRange(startStatement.name, range)) {
                    return { toMove: [statements[startNodeIndex]], afterLast: statements[startNodeIndex + 1] };
                }
                // Can't only partially include the start node or be partially into the next node
                if (range.pos > startStatement.getStart(file))
                    return undefined;
                var afterEndNodeIndex = ts.findIndex(statements, function (s) { return s.end > range.end; }, startNodeIndex);
                // Can't be partially into the next node
                if (afterEndNodeIndex !== -1 && (afterEndNodeIndex === 0 || statements[afterEndNodeIndex].getStart(file) < range.end))
                    return undefined;
                return {
                    toMove: statements.slice(startNodeIndex, afterEndNodeIndex === -1 ? statements.length : afterEndNodeIndex),
                    afterLast: afterEndNodeIndex === -1 ? undefined : statements[afterEndNodeIndex],
                };
            }
            function doChange(oldFile, program, toMove, changes, host, preferences) {
                var checker = program.getTypeChecker();
                var usage = getUsageInfo(oldFile, toMove.all, checker);
                var currentDirectory = ts.getDirectoryPath(oldFile.fileName);
                var extension = ts.extensionFromPath(oldFile.fileName);
                var newModuleName = makeUniqueModuleName(getNewModuleName(usage.movedSymbols), extension, currentDirectory, host);
                var newFileNameWithExtension = newModuleName + extension;
                // If previous file was global, this is easy.
                changes.createNewFile(oldFile, ts.combinePaths(currentDirectory, newFileNameWithExtension), getNewStatementsAndRemoveFromOldFile(oldFile, usage, changes, toMove, program, newModuleName, preferences));
                addNewFileToTsconfig(program, changes, oldFile.fileName, newFileNameWithExtension, ts.hostGetCanonicalFileName(host));
            }
            function getStatementsToMove(context) {
                var rangeToMove = getRangeToMove(context);
                if (rangeToMove === undefined)
                    return undefined;
                var all = [];
                var ranges = [];
                var toMove = rangeToMove.toMove, afterLast = rangeToMove.afterLast;
                ts.getRangesWhere(toMove, isAllowedStatementToMove, function (start, afterEndIndex) {
                    for (var i = start; i < afterEndIndex; i++)
                        all.push(toMove[i]);
                    ranges.push({ first: toMove[start], afterLast: afterLast });
                });
                return all.length === 0 ? undefined : { all: all, ranges: ranges };
            }
            function isAllowedStatementToMove(statement) {
                // Filters imports and prologue directives out of the range of statements to move.
                // Imports will be copied to the new file anyway, and may still be needed in the old file.
                // Prologue directives will be copied to the new file and should be left in the old file.
                return !isPureImport(statement) && !ts.isPrologueDirective(statement);
                ;
            }
            function isPureImport(node) {
                switch (node.kind) {
                    case 264 /* ImportDeclaration */:
                        return true;
                    case 263 /* ImportEqualsDeclaration */:
                        return !ts.hasSyntacticModifier(node, 1 /* Export */);
                    case 235 /* VariableStatement */:
                        return node.declarationList.declarations.every(function (d) { return !!d.initializer && ts.isRequireCall(d.initializer, /*checkArgumentIsStringLiteralLike*/ true); });
                    default:
                        return false;
                }
            }
            function addNewFileToTsconfig(program, changes, oldFileName, newFileNameWithExtension, getCanonicalFileName) {
                var cfg = program.getCompilerOptions().configFile;
                if (!cfg)
                    return;
                var newFileAbsolutePath = ts.normalizePath(ts.combinePaths(oldFileName, "..", newFileNameWithExtension));
                var newFilePath = ts.getRelativePathFromFile(cfg.fileName, newFileAbsolutePath, getCanonicalFileName);
                var cfgObject = cfg.statements[0] && ts.tryCast(cfg.statements[0].expression, ts.isObjectLiteralExpression);
                var filesProp = cfgObject && ts.find(cfgObject.properties, function (prop) {
                    return ts.isPropertyAssignment(prop) && ts.isStringLiteral(prop.name) && prop.name.text === "files";
                });
                if (filesProp && ts.isArrayLiteralExpression(filesProp.initializer)) {
                    changes.insertNodeInListAfter(cfg, ts.last(filesProp.initializer.elements), ts.factory.createStringLiteral(newFilePath), filesProp.initializer.elements);
                }
            }
            function getNewStatementsAndRemoveFromOldFile(oldFile, usage, changes, toMove, program, newModuleName, preferences) {
                var checker = program.getTypeChecker();
                var prologueDirectives = ts.takeWhile(oldFile.statements, ts.isPrologueDirective);
                if (!oldFile.externalModuleIndicator && !oldFile.commonJsModuleIndicator) {
                    deleteMovedStatements(oldFile, toMove.ranges, changes);
                    return __spreadArray(__spreadArray([], prologueDirectives, true), toMove.all, true);
                }
                var useEs6ModuleSyntax = !!oldFile.externalModuleIndicator;
                var quotePreference = ts.getQuotePreference(oldFile, preferences);
                var importsFromNewFile = createOldFileImportsFromNewFile(usage.oldFileImportsFromNewFile, newModuleName, useEs6ModuleSyntax, quotePreference);
                if (importsFromNewFile) {
                    ts.insertImports(changes, oldFile, importsFromNewFile, /*blankLineBetween*/ true);
                }
                deleteUnusedOldImports(oldFile, toMove.all, changes, usage.unusedImportsFromOldFile, checker);
                deleteMovedStatements(oldFile, toMove.ranges, changes);
                updateImportsInOtherFiles(changes, program, oldFile, usage.movedSymbols, newModuleName);
                var imports = getNewFileImportsAndAddExportInOldFile(oldFile, usage.oldImportsNeededByNewFile, usage.newFileImportsFromOldFile, changes, checker, useEs6ModuleSyntax, quotePreference);
                var body = addExports(oldFile, toMove.all, usage.oldFileImportsFromNewFile, useEs6ModuleSyntax);
                if (imports.length && body.length) {
                    return __spreadArray(__spreadArray(__spreadArray(__spreadArray([], prologueDirectives, true), imports, true), [
                        4 /* NewLineTrivia */
                    ], false), body, true);
                }
                return __spreadArray(__spreadArray(__spreadArray([], prologueDirectives, true), imports, true), body, true);
            }
            function deleteMovedStatements(sourceFile, moved, changes) {
                for (var _i = 0, moved_1 = moved; _i < moved_1.length; _i++) {
                    var _a = moved_1[_i], first_1 = _a.first, afterLast = _a.afterLast;
                    changes.deleteNodeRangeExcludingEnd(sourceFile, first_1, afterLast);
                }
            }
            function deleteUnusedOldImports(oldFile, toMove, changes, toDelete, checker) {
                for (var _i = 0, _a = oldFile.statements; _i < _a.length; _i++) {
                    var statement = _a[_i];
                    if (ts.contains(toMove, statement))
                        continue;
                    forEachImportInStatement(statement, function (i) { return deleteUnusedImports(oldFile, i, changes, function (name) { return toDelete.has(checker.getSymbolAtLocation(name)); }); });
                }
            }
            function updateImportsInOtherFiles(changes, program, oldFile, movedSymbols, newModuleName) {
                var checker = program.getTypeChecker();
                var _loop_18 = function (sourceFile) {
                    if (sourceFile === oldFile)
                        return "continue";
                    var _loop_19 = function (statement) {
                        forEachImportInStatement(statement, function (importNode) {
                            if (checker.getSymbolAtLocation(moduleSpecifierFromImport(importNode)) !== oldFile.symbol)
                                return;
                            var shouldMove = function (name) {
                                var symbol = ts.isBindingElement(name.parent)
                                    ? ts.getPropertySymbolFromBindingElement(checker, name.parent)
                                    : ts.skipAlias(checker.getSymbolAtLocation(name), checker); // TODO: GH#18217
                                return !!symbol && movedSymbols.has(symbol);
                            };
                            deleteUnusedImports(sourceFile, importNode, changes, shouldMove); // These will be changed to imports from the new file
                            var newModuleSpecifier = ts.combinePaths(ts.getDirectoryPath(moduleSpecifierFromImport(importNode).text), newModuleName);
                            var newImportDeclaration = filterImport(importNode, ts.factory.createStringLiteral(newModuleSpecifier), shouldMove);
                            if (newImportDeclaration)
                                changes.insertNodeAfter(sourceFile, statement, newImportDeclaration);
                            var ns = getNamespaceLikeImport(importNode);
                            if (ns)
                                updateNamespaceLikeImport(changes, sourceFile, checker, movedSymbols, newModuleName, newModuleSpecifier, ns, importNode);
                        });
                    };
                    for (var _b = 0, _c = sourceFile.statements; _b < _c.length; _b++) {
                        var statement = _c[_b];
                        _loop_19(statement);
                    }
                };
                for (var _i = 0, _a = program.getSourceFiles(); _i < _a.length; _i++) {
                    var sourceFile = _a[_i];
                    _loop_18(sourceFile);
                }
            }
            function getNamespaceLikeImport(node) {
                switch (node.kind) {
                    case 264 /* ImportDeclaration */:
                        return node.importClause && node.importClause.namedBindings && node.importClause.namedBindings.kind === 266 /* NamespaceImport */ ?
                            node.importClause.namedBindings.name : undefined;
                    case 263 /* ImportEqualsDeclaration */:
                        return node.name;
                    case 252 /* VariableDeclaration */:
                        return ts.tryCast(node.name, ts.isIdentifier);
                    default:
                        return ts.Debug.assertNever(node, "Unexpected node kind " + node.kind);
                }
            }
            function updateNamespaceLikeImport(changes, sourceFile, checker, movedSymbols, newModuleName, newModuleSpecifier, oldImportId, oldImportNode) {
                var preferredNewNamespaceName = ts.codefix.moduleSpecifierToValidIdentifier(newModuleName, 99 /* ESNext */);
                var needUniqueName = false;
                var toChange = [];
                ts.FindAllReferences.Core.eachSymbolReferenceInFile(oldImportId, checker, sourceFile, function (ref) {
                    if (!ts.isPropertyAccessExpression(ref.parent))
                        return;
                    needUniqueName = needUniqueName || !!checker.resolveName(preferredNewNamespaceName, ref, 67108863 /* All */, /*excludeGlobals*/ true);
                    if (movedSymbols.has(checker.getSymbolAtLocation(ref.parent.name))) {
                        toChange.push(ref);
                    }
                });
                if (toChange.length) {
                    var newNamespaceName = needUniqueName ? ts.getUniqueName(preferredNewNamespaceName, sourceFile) : preferredNewNamespaceName;
                    for (var _i = 0, toChange_1 = toChange; _i < toChange_1.length; _i++) {
                        var ref = toChange_1[_i];
                        changes.replaceNode(sourceFile, ref, ts.factory.createIdentifier(newNamespaceName));
                    }
                    changes.insertNodeAfter(sourceFile, oldImportNode, updateNamespaceLikeImportNode(oldImportNode, newModuleName, newModuleSpecifier));
                }
            }
            function updateNamespaceLikeImportNode(node, newNamespaceName, newModuleSpecifier) {
                var newNamespaceId = ts.factory.createIdentifier(newNamespaceName);
                var newModuleString = ts.factory.createStringLiteral(newModuleSpecifier);
                switch (node.kind) {
                    case 264 /* ImportDeclaration */:
                        return ts.factory.createImportDeclaration(
                        /*decorators*/ undefined, /*modifiers*/ undefined, ts.factory.createImportClause(/*isTypeOnly*/ false, /*name*/ undefined, ts.factory.createNamespaceImport(newNamespaceId)), newModuleString);
                    case 263 /* ImportEqualsDeclaration */:
                        return ts.factory.createImportEqualsDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, /*isTypeOnly*/ false, newNamespaceId, ts.factory.createExternalModuleReference(newModuleString));
                    case 252 /* VariableDeclaration */:
                        return ts.factory.createVariableDeclaration(newNamespaceId, /*exclamationToken*/ undefined, /*type*/ undefined, createRequireCall(newModuleString));
                    default:
                        return ts.Debug.assertNever(node, "Unexpected node kind " + node.kind);
                }
            }
            function moduleSpecifierFromImport(i) {
                return (i.kind === 264 /* ImportDeclaration */ ? i.moduleSpecifier
                    : i.kind === 263 /* ImportEqualsDeclaration */ ? i.moduleReference.expression
                        : i.initializer.arguments[0]);
            }
            function forEachImportInStatement(statement, cb) {
                if (ts.isImportDeclaration(statement)) {
                    if (ts.isStringLiteral(statement.moduleSpecifier))
                        cb(statement);
                }
                else if (ts.isImportEqualsDeclaration(statement)) {
                    if (ts.isExternalModuleReference(statement.moduleReference) && ts.isStringLiteralLike(statement.moduleReference.expression)) {
                        cb(statement);
                    }
                }
                else if (ts.isVariableStatement(statement)) {
                    for (var _i = 0, _a = statement.declarationList.declarations; _i < _a.length; _i++) {
                        var decl = _a[_i];
                        if (decl.initializer && ts.isRequireCall(decl.initializer, /*checkArgumentIsStringLiteralLike*/ true)) {
                            cb(decl);
                        }
                    }
                }
            }
            function createOldFileImportsFromNewFile(newFileNeedExport, newFileNameWithExtension, useEs6Imports, quotePreference) {
                var defaultImport;
                var imports = [];
                newFileNeedExport.forEach(function (symbol) {
                    if (symbol.escapedName === "default" /* Default */) {
                        defaultImport = ts.factory.createIdentifier(ts.symbolNameNoDefault(symbol)); // TODO: GH#18217
                    }
                    else {
                        imports.push(symbol.name);
                    }
                });
                return makeImportOrRequire(defaultImport, imports, newFileNameWithExtension, useEs6Imports, quotePreference);
            }
            function makeImportOrRequire(defaultImport, imports, path, useEs6Imports, quotePreference) {
                path = ts.ensurePathIsNonModuleName(path);
                if (useEs6Imports) {
                    var specifiers = imports.map(function (i) { return ts.factory.createImportSpecifier(/*propertyName*/ undefined, ts.factory.createIdentifier(i)); });
                    return ts.makeImportIfNecessary(defaultImport, specifiers, path, quotePreference);
                }
                else {
                    ts.Debug.assert(!defaultImport, "No default import should exist"); // If there's a default export, it should have been an es6 module.
                    var bindingElements = imports.map(function (i) { return ts.factory.createBindingElement(/*dotDotDotToken*/ undefined, /*propertyName*/ undefined, i); });
                    return bindingElements.length
                        ? makeVariableStatement(ts.factory.createObjectBindingPattern(bindingElements), /*type*/ undefined, createRequireCall(ts.factory.createStringLiteral(path)))
                        : undefined;
                }
            }
            function makeVariableStatement(name, type, initializer, flags) {
                if (flags === void 0) {
                    flags = 2 /* Const */;
                }
                return ts.factory.createVariableStatement(/*modifiers*/ undefined, ts.factory.createVariableDeclarationList([ts.factory.createVariableDeclaration(name, /*exclamationToken*/ undefined, type, initializer)], flags));
            }
            function createRequireCall(moduleSpecifier) {
                return ts.factory.createCallExpression(ts.factory.createIdentifier("require"), /*typeArguments*/ undefined, [moduleSpecifier]);
            }
            function addExports(sourceFile, toMove, needExport, useEs6Exports) {
                return ts.flatMap(toMove, function (statement) {
                    if (isTopLevelDeclarationStatement(statement) &&
                        !isExported(sourceFile, statement, useEs6Exports) &&
                        forEachTopLevelDeclaration(statement, function (d) { return needExport.has(ts.Debug.checkDefined(d.symbol)); })) {
                        var exports = addExport(statement, useEs6Exports);
                        if (exports)
                            return exports;
                    }
                    return statement;
                });
            }
            function deleteUnusedImports(sourceFile, importDecl, changes, isUnused) {
                switch (importDecl.kind) {
                    case 264 /* ImportDeclaration */:
                        deleteUnusedImportsInDeclaration(sourceFile, importDecl, changes, isUnused);
                        break;
                    case 263 /* ImportEqualsDeclaration */:
                        if (isUnused(importDecl.name)) {
                            changes.delete(sourceFile, importDecl);
                        }
                        break;
                    case 252 /* VariableDeclaration */:
                        deleteUnusedImportsInVariableDeclaration(sourceFile, importDecl, changes, isUnused);
                        break;
                    default:
                        ts.Debug.assertNever(importDecl, "Unexpected import decl kind " + importDecl.kind);
                }
            }
            function deleteUnusedImportsInDeclaration(sourceFile, importDecl, changes, isUnused) {
                if (!importDecl.importClause)
                    return;
                var _a = importDecl.importClause, name = _a.name, namedBindings = _a.namedBindings;
                var defaultUnused = !name || isUnused(name);
                var namedBindingsUnused = !namedBindings ||
                    (namedBindings.kind === 266 /* NamespaceImport */ ? isUnused(namedBindings.name) : namedBindings.elements.length !== 0 && namedBindings.elements.every(function (e) { return isUnused(e.name); }));
                if (defaultUnused && namedBindingsUnused) {
                    changes.delete(sourceFile, importDecl);
                }
                else {
                    if (name && defaultUnused) {
                        changes.delete(sourceFile, name);
                    }
                    if (namedBindings) {
                        if (namedBindingsUnused) {
                            changes.replaceNode(sourceFile, importDecl.importClause, ts.factory.updateImportClause(importDecl.importClause, importDecl.importClause.isTypeOnly, name, /*namedBindings*/ undefined));
                        }
                        else if (namedBindings.kind === 267 /* NamedImports */) {
                            for (var _i = 0, _b = namedBindings.elements; _i < _b.length; _i++) {
                                var element = _b[_i];
                                if (isUnused(element.name))
                                    changes.delete(sourceFile, element);
                            }
                        }
                    }
                }
            }
            function deleteUnusedImportsInVariableDeclaration(sourceFile, varDecl, changes, isUnused) {
                var name = varDecl.name;
                switch (name.kind) {
                    case 79 /* Identifier */:
                        if (isUnused(name)) {
                            changes.delete(sourceFile, name);
                        }
                        break;
                    case 200 /* ArrayBindingPattern */:
                        break;
                    case 199 /* ObjectBindingPattern */:
                        if (name.elements.every(function (e) { return ts.isIdentifier(e.name) && isUnused(e.name); })) {
                            changes.delete(sourceFile, ts.isVariableDeclarationList(varDecl.parent) && varDecl.parent.declarations.length === 1 ? varDecl.parent.parent : varDecl);
                        }
                        else {
                            for (var _i = 0, _a = name.elements; _i < _a.length; _i++) {
                                var element = _a[_i];
                                if (ts.isIdentifier(element.name) && isUnused(element.name)) {
                                    changes.delete(sourceFile, element.name);
                                }
                            }
                        }
                        break;
                }
            }
            function getNewFileImportsAndAddExportInOldFile(oldFile, importsToCopy, newFileImportsFromOldFile, changes, checker, useEs6ModuleSyntax, quotePreference) {
                var copiedOldImports = [];
                for (var _i = 0, _a = oldFile.statements; _i < _a.length; _i++) {
                    var oldStatement = _a[_i];
                    forEachImportInStatement(oldStatement, function (i) {
                        ts.append(copiedOldImports, filterImport(i, moduleSpecifierFromImport(i), function (name) { return importsToCopy.has(checker.getSymbolAtLocation(name)); }));
                    });
                }
                // Also, import things used from the old file, and insert 'export' modifiers as necessary in the old file.
                var oldFileDefault;
                var oldFileNamedImports = [];
                var markSeenTop = ts.nodeSeenTracker(); // Needed because multiple declarations may appear in `const x = 0, y = 1;`.
                newFileImportsFromOldFile.forEach(function (symbol) {
                    if (!symbol.declarations) {
                        return;
                    }
                    for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                        var decl = _a[_i];
                        if (!isTopLevelDeclaration(decl))
                            continue;
                        var name = nameOfTopLevelDeclaration(decl);
                        if (!name)
                            continue;
                        var top = getTopLevelDeclarationStatement(decl);
                        if (markSeenTop(top)) {
                            addExportToChanges(oldFile, top, name, changes, useEs6ModuleSyntax);
                        }
                        if (ts.hasSyntacticModifier(decl, 512 /* Default */)) {
                            oldFileDefault = name;
                        }
                        else {
                            oldFileNamedImports.push(name.text);
                        }
                    }
                });
                ts.append(copiedOldImports, makeImportOrRequire(oldFileDefault, oldFileNamedImports, ts.removeFileExtension(ts.getBaseFileName(oldFile.fileName)), useEs6ModuleSyntax, quotePreference));
                return copiedOldImports;
            }
            function makeUniqueModuleName(moduleName, extension, inDirectory, host) {
                var newModuleName = moduleName;
                for (var i = 1;; i++) {
                    var name = ts.combinePaths(inDirectory, newModuleName + extension);
                    if (!host.fileExists(name))
                        return newModuleName; // TODO: GH#18217
                    newModuleName = moduleName + "." + i;
                }
            }
            function getNewModuleName(movedSymbols) {
                return movedSymbols.forEachEntry(ts.symbolNameNoDefault) || "newFile";
            }
            function getUsageInfo(oldFile, toMove, checker) {
                var movedSymbols = new SymbolSet();
                var oldImportsNeededByNewFile = new SymbolSet();
                var newFileImportsFromOldFile = new SymbolSet();
                var containsJsx = ts.find(toMove, function (statement) { return !!(statement.transformFlags & 2 /* ContainsJsx */); });
                var jsxNamespaceSymbol = getJsxNamespaceSymbol(containsJsx);
                if (jsxNamespaceSymbol) { // Might not exist (e.g. in non-compiling code)
                    oldImportsNeededByNewFile.add(jsxNamespaceSymbol);
                }
                for (var _i = 0, toMove_1 = toMove; _i < toMove_1.length; _i++) {
                    var statement = toMove_1[_i];
                    forEachTopLevelDeclaration(statement, function (decl) {
                        movedSymbols.add(ts.Debug.checkDefined(ts.isExpressionStatement(decl) ? checker.getSymbolAtLocation(decl.expression.left) : decl.symbol, "Need a symbol here"));
                    });
                }
                for (var _a = 0, toMove_2 = toMove; _a < toMove_2.length; _a++) {
                    var statement = toMove_2[_a];
                    forEachReference(statement, checker, function (symbol) {
                        if (!symbol.declarations)
                            return;
                        for (var _i = 0, _a = symbol.declarations; _i < _a.length; _i++) {
                            var decl = _a[_i];
                            if (isInImport(decl)) {
                                oldImportsNeededByNewFile.add(symbol);
                            }
                            else if (isTopLevelDeclaration(decl) && sourceFileOfTopLevelDeclaration(decl) === oldFile && !movedSymbols.has(symbol)) {
                                newFileImportsFromOldFile.add(symbol);
                            }
                        }
                    });
                }
                var unusedImportsFromOldFile = oldImportsNeededByNewFile.clone();
                var oldFileImportsFromNewFile = new SymbolSet();
                for (var _b = 0, _c = oldFile.statements; _b < _c.length; _b++) {
                    var statement = _c[_b];
                    if (ts.contains(toMove, statement))
                        continue;
                    // jsxNamespaceSymbol will only be set iff it is in oldImportsNeededByNewFile.
                    if (jsxNamespaceSymbol && !!(statement.transformFlags & 2 /* ContainsJsx */)) {
                        unusedImportsFromOldFile.delete(jsxNamespaceSymbol);
                    }
                    forEachReference(statement, checker, function (symbol) {
                        if (movedSymbols.has(symbol))
                            oldFileImportsFromNewFile.add(symbol);
                        unusedImportsFromOldFile.delete(symbol);
                    });
                }
                return { movedSymbols: movedSymbols, newFileImportsFromOldFile: newFileImportsFromOldFile, oldFileImportsFromNewFile: oldFileImportsFromNewFile, oldImportsNeededByNewFile: oldImportsNeededByNewFile, unusedImportsFromOldFile: unusedImportsFromOldFile };
                function getJsxNamespaceSymbol(containsJsx) {
                    if (containsJsx === undefined) {
                        return undefined;
                    }
                    var jsxNamespace = checker.getJsxNamespace(containsJsx);
                    // Strictly speaking, this could resolve to a symbol other than the JSX namespace.
                    // This will produce erroneous output (probably, an incorrectly copied import) but
                    // is expected to be very rare and easily reversible.
                    var jsxNamespaceSymbol = checker.resolveName(jsxNamespace, containsJsx, 1920 /* Namespace */, /*excludeGlobals*/ true);
                    return !!jsxNamespaceSymbol && ts.some(jsxNamespaceSymbol.declarations, isInImport)
                        ? jsxNamespaceSymbol
                        : undefined;
                }
            }
            // Below should all be utilities
            function isInImport(decl) {
                switch (decl.kind) {
                    case 263 /* ImportEqualsDeclaration */:
                    case 268 /* ImportSpecifier */:
                    case 265 /* ImportClause */:
                    case 266 /* NamespaceImport */:
                        return true;
                    case 252 /* VariableDeclaration */:
                        return isVariableDeclarationInImport(decl);
                    case 201 /* BindingElement */:
                        return ts.isVariableDeclaration(decl.parent.parent) && isVariableDeclarationInImport(decl.parent.parent);
                    default:
                        return false;
                }
            }
            function isVariableDeclarationInImport(decl) {
                return ts.isSourceFile(decl.parent.parent.parent) &&
                    !!decl.initializer && ts.isRequireCall(decl.initializer, /*checkArgumentIsStringLiteralLike*/ true);
            }
            function filterImport(i, moduleSpecifier, keep) {
                switch (i.kind) {
                    case 264 /* ImportDeclaration */: {
                        var clause = i.importClause;
                        if (!clause)
                            return undefined;
                        var defaultImport = clause.name && keep(clause.name) ? clause.name : undefined;
                        var namedBindings = clause.namedBindings && filterNamedBindings(clause.namedBindings, keep);
                        return defaultImport || namedBindings
                            ? ts.factory.createImportDeclaration(/*decorators*/ undefined, /*modifiers*/ undefined, ts.factory.createImportClause(/*isTypeOnly*/ false, defaultImport, namedBindings), moduleSpecifier)
                            : undefined;
                    }
                    case 263 /* ImportEqualsDeclaration */:
                        return keep(i.name) ? i : undefined;
                    case 252 /* VariableDeclaration */: {
                        var name = filterBindingName(i.name, keep);
                        return name ? makeVariableStatement(name, i.type, createRequireCall(moduleSpecifier), i.parent.flags) : undefined;
                    }
                    default:
                        return ts.Debug.assertNever(i, "Unexpected import kind " + i.kind);
                }
            }
            function filterNamedBindings(namedBindings, keep) {
                if (namedBindings.kind === 266 /* NamespaceImport */) {
                    return keep(namedBindings.name) ? namedBindings : undefined;
                }
                else {
                    var newElements = namedBindings.elements.filter(function (e) { return keep(e.name); });
                    return newElements.length ? ts.factory.createNamedImports(newElements) : undefined;
                }
            }
            function filterBindingName(name, keep) {
                switch (name.kind) {
                    case 79 /* Identifier */:
                        return keep(name) ? name : undefined;
                    case 200 /* ArrayBindingPattern */:
                        return name;
                    case 199 /* ObjectBindingPattern */: {
                        // We can't handle nested destructurings or property names well here, so just copy them all.
                        var newElements = name.elements.filter(function (prop) { return prop.propertyName || !ts.isIdentifier(prop.name) || keep(prop.name); });
                        return newElements.length ? ts.factory.createObjectBindingPattern(newElements) : undefined;
                    }
                }
            }
            function forEachReference(node, checker, onReference) {
                node.forEachChild(function cb(node) {
                    if (ts.isIdentifier(node) && !ts.isDeclarationName(node)) {
                        var sym = checker.getSymbolAtLocation(node);
                        if (sym)
                            onReference(sym);
                    }
                    else {
                        node.forEachChild(cb);
                    }
                });
            }
            var SymbolSet = /** @class */ (function () {
                function SymbolSet() {
                    this.map = new ts.Map();
                }
                SymbolSet.prototype.add = function (symbol) {
                    this.map.set(String(ts.getSymbolId(symbol)), symbol);
                };
                SymbolSet.prototype.has = function (symbol) {
                    return this.map.has(String(ts.getSymbolId(symbol)));
                };
                SymbolSet.prototype.delete = function (symbol) {
                    this.map.delete(String(ts.getSymbolId(symbol)));
                };
                SymbolSet.prototype.forEach = function (cb) {
                    this.map.forEach(cb);
                };
                SymbolSet.prototype.forEachEntry = function (cb) {
                    return ts.forEachEntry(this.map, cb);
                };
                SymbolSet.prototype.clone = function () {
                    var clone = new SymbolSet();
                    ts.copyEntries(this.map, clone.map);
                    return clone;
                };
                return SymbolSet;
            }());
            function isTopLevelDeclaration(node) {
                return isNonVariableTopLevelDeclaration(node) && ts.isSourceFile(node.parent) || ts.isVariableDeclaration(node) && ts.isSourceFile(node.parent.parent.parent);
            }
            function sourceFileOfTopLevelDeclaration(node) {
                return ts.isVariableDeclaration(node) ? node.parent.parent.parent : node.parent;
            }
            function isTopLevelDeclarationStatement(node) {
                ts.Debug.assert(ts.isSourceFile(node.parent), "Node parent should be a SourceFile");
                return isNonVariableTopLevelDeclaration(node) || ts.isVariableStatement(node);
            }
            function isNonVariableTopLevelDeclaration(node) {
                switch (node.kind) {
                    case 254 /* FunctionDeclaration */:
                    case 255 /* ClassDeclaration */:
                    case 259 /* ModuleDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 263 /* ImportEqualsDeclaration */:
                        return true;
                    default:
                        return false;
                }
            }
            function forEachTopLevelDeclaration(statement, cb) {
                switch (statement.kind) {
                    case 254 /* FunctionDeclaration */:
                    case 255 /* ClassDeclaration */:
                    case 259 /* ModuleDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 263 /* ImportEqualsDeclaration */:
                        return cb(statement);
                    case 235 /* VariableStatement */:
                        return ts.firstDefined(statement.declarationList.declarations, function (decl) { return forEachTopLevelDeclarationInBindingName(decl.name, cb); });
                    case 236 /* ExpressionStatement */: {
                        var expression = statement.expression;
                        return ts.isBinaryExpression(expression) && ts.getAssignmentDeclarationKind(expression) === 1 /* ExportsProperty */
                            ? cb(statement)
                            : undefined;
                    }
                }
            }
            function forEachTopLevelDeclarationInBindingName(name, cb) {
                switch (name.kind) {
                    case 79 /* Identifier */:
                        return cb(ts.cast(name.parent, function (x) { return ts.isVariableDeclaration(x) || ts.isBindingElement(x); }));
                    case 200 /* ArrayBindingPattern */:
                    case 199 /* ObjectBindingPattern */:
                        return ts.firstDefined(name.elements, function (em) { return ts.isOmittedExpression(em) ? undefined : forEachTopLevelDeclarationInBindingName(em.name, cb); });
                    default:
                        return ts.Debug.assertNever(name, "Unexpected name kind " + name.kind);
                }
            }
            function nameOfTopLevelDeclaration(d) {
                return ts.isExpressionStatement(d) ? ts.tryCast(d.expression.left.name, ts.isIdentifier) : ts.tryCast(d.name, ts.isIdentifier);
            }
            function getTopLevelDeclarationStatement(d) {
                switch (d.kind) {
                    case 252 /* VariableDeclaration */:
                        return d.parent.parent;
                    case 201 /* BindingElement */:
                        return getTopLevelDeclarationStatement(ts.cast(d.parent.parent, function (p) { return ts.isVariableDeclaration(p) || ts.isBindingElement(p); }));
                    default:
                        return d;
                }
            }
            function addExportToChanges(sourceFile, decl, name, changes, useEs6Exports) {
                if (isExported(sourceFile, decl, useEs6Exports, name))
                    return;
                if (useEs6Exports) {
                    if (!ts.isExpressionStatement(decl))
                        changes.insertExportModifier(sourceFile, decl);
                }
                else {
                    var names = getNamesToExportInCommonJS(decl);
                    if (names.length !== 0)
                        changes.insertNodesAfter(sourceFile, decl, names.map(createExportAssignment));
                }
            }
            function isExported(sourceFile, decl, useEs6Exports, name) {
                var _a;
                if (useEs6Exports) {
                    return !ts.isExpressionStatement(decl) && ts.hasSyntacticModifier(decl, 1 /* Export */) || !!(name && ((_a = sourceFile.symbol.exports) === null || _a === void 0 ? void 0 : _a.has(name.escapedText)));
                }
                return getNamesToExportInCommonJS(decl).some(function (name) { return sourceFile.symbol.exports.has(ts.escapeLeadingUnderscores(name)); });
            }
            function addExport(decl, useEs6Exports) {
                return useEs6Exports ? [addEs6Export(decl)] : addCommonjsExport(decl);
            }
            function addEs6Export(d) {
                var modifiers = ts.concatenate([ts.factory.createModifier(93 /* ExportKeyword */)], d.modifiers);
                switch (d.kind) {
                    case 254 /* FunctionDeclaration */:
                        return ts.factory.updateFunctionDeclaration(d, d.decorators, modifiers, d.asteriskToken, d.name, d.typeParameters, d.parameters, d.type, d.body);
                    case 255 /* ClassDeclaration */:
                        return ts.factory.updateClassDeclaration(d, d.decorators, modifiers, d.name, d.typeParameters, d.heritageClauses, d.members);
                    case 235 /* VariableStatement */:
                        return ts.factory.updateVariableStatement(d, modifiers, d.declarationList);
                    case 259 /* ModuleDeclaration */:
                        return ts.factory.updateModuleDeclaration(d, d.decorators, modifiers, d.name, d.body);
                    case 258 /* EnumDeclaration */:
                        return ts.factory.updateEnumDeclaration(d, d.decorators, modifiers, d.name, d.members);
                    case 257 /* TypeAliasDeclaration */:
                        return ts.factory.updateTypeAliasDeclaration(d, d.decorators, modifiers, d.name, d.typeParameters, d.type);
                    case 256 /* InterfaceDeclaration */:
                        return ts.factory.updateInterfaceDeclaration(d, d.decorators, modifiers, d.name, d.typeParameters, d.heritageClauses, d.members);
                    case 263 /* ImportEqualsDeclaration */:
                        return ts.factory.updateImportEqualsDeclaration(d, d.decorators, modifiers, d.isTypeOnly, d.name, d.moduleReference);
                    case 236 /* ExpressionStatement */:
                        return ts.Debug.fail(); // Shouldn't try to add 'export' keyword to `exports.x = ...`
                    default:
                        return ts.Debug.assertNever(d, "Unexpected declaration kind " + d.kind);
                }
            }
            function addCommonjsExport(decl) {
                return __spreadArray([decl], getNamesToExportInCommonJS(decl).map(createExportAssignment), true);
            }
            function getNamesToExportInCommonJS(decl) {
                switch (decl.kind) {
                    case 254 /* FunctionDeclaration */:
                    case 255 /* ClassDeclaration */:
                        return [decl.name.text]; // TODO: GH#18217
                    case 235 /* VariableStatement */:
                        return ts.mapDefined(decl.declarationList.declarations, function (d) { return ts.isIdentifier(d.name) ? d.name.text : undefined; });
                    case 259 /* ModuleDeclaration */:
                    case 258 /* EnumDeclaration */:
                    case 257 /* TypeAliasDeclaration */:
                    case 256 /* InterfaceDeclaration */:
                    case 263 /* ImportEqualsDeclaration */:
                        return ts.emptyArray;
                    case 236 /* ExpressionStatement */:
                        return ts.Debug.fail("Can't export an ExpressionStatement"); // Shouldn't try to add 'export' keyword to `exports.x = ...`
                    default:
                        return ts.Debug.assertNever(decl, "Unexpected decl kind " + decl.kind);
                }
            }
            /** Creates `exports.x = x;` */
            function createExportAssignment(name) {
                return ts.factory.createExpressionStatement(ts.factory.createBinaryExpression(ts.factory.createPropertyAccessExpression(ts.factory.createIdentifier("exports"), ts.factory.createIdentifier(name)), 63 /* EqualsToken */, ts.factory.createIdentifier(name)));
            }
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var addOrRemoveBracesToArrowFunction;
            (function (addOrRemoveBracesToArrowFunction) {
                var refactorName = "Add or remove braces in an arrow function";
                var refactorDescription = ts.Diagnostics.Add_or_remove_braces_in_an_arrow_function.message;
                var addBracesAction = {
                    name: "Add braces to arrow function",
                    description: ts.Diagnostics.Add_braces_to_arrow_function.message,
                    kind: "refactor.rewrite.arrow.braces.add",
                };
                var removeBracesAction = {
                    name: "Remove braces from arrow function",
                    description: ts.Diagnostics.Remove_braces_from_arrow_function.message,
                    kind: "refactor.rewrite.arrow.braces.remove"
                };
                refactor.registerRefactor(refactorName, {
                    kinds: [removeBracesAction.kind],
                    getEditsForAction: getEditsForAction,
                    getAvailableActions: getAvailableActions
                });
                function getAvailableActions(context) {
                    var file = context.file, startPosition = context.startPosition, triggerReason = context.triggerReason;
                    var info = getConvertibleArrowFunctionAtPosition(file, startPosition, triggerReason === "invoked");
                    if (!info)
                        return ts.emptyArray;
                    if (!refactor.isRefactorErrorInfo(info)) {
                        return [{
                                name: refactorName,
                                description: refactorDescription,
                                actions: [
                                    info.addBraces ? addBracesAction : removeBracesAction
                                ]
                            }];
                    }
                    if (context.preferences.provideRefactorNotApplicableReason) {
                        return [{
                                name: refactorName,
                                description: refactorDescription,
                                actions: [
                                    __assign(__assign({}, addBracesAction), { notApplicableReason: info.error }),
                                    __assign(__assign({}, removeBracesAction), { notApplicableReason: info.error }),
                                ]
                            }];
                    }
                    return ts.emptyArray;
                }
                function getEditsForAction(context, actionName) {
                    var file = context.file, startPosition = context.startPosition;
                    var info = getConvertibleArrowFunctionAtPosition(file, startPosition);
                    ts.Debug.assert(info && !refactor.isRefactorErrorInfo(info), "Expected applicable refactor info");
                    var expression = info.expression, returnStatement = info.returnStatement, func = info.func;
                    var body;
                    if (actionName === addBracesAction.name) {
                        var returnStatement_1 = ts.factory.createReturnStatement(expression);
                        body = ts.factory.createBlock([returnStatement_1], /* multiLine */ true);
                        ts.suppressLeadingAndTrailingTrivia(body);
                        ts.copyLeadingComments(expression, returnStatement_1, file, 3 /* MultiLineCommentTrivia */, /* hasTrailingNewLine */ true);
                    }
                    else if (actionName === removeBracesAction.name && returnStatement) {
                        var actualExpression = expression || ts.factory.createVoidZero();
                        body = ts.needsParentheses(actualExpression) ? ts.factory.createParenthesizedExpression(actualExpression) : actualExpression;
                        ts.suppressLeadingAndTrailingTrivia(body);
                        ts.copyTrailingAsLeadingComments(returnStatement, body, file, 3 /* MultiLineCommentTrivia */, /* hasTrailingNewLine */ false);
                        ts.copyLeadingComments(returnStatement, body, file, 3 /* MultiLineCommentTrivia */, /* hasTrailingNewLine */ false);
                        ts.copyTrailingComments(returnStatement, body, file, 3 /* MultiLineCommentTrivia */, /* hasTrailingNewLine */ false);
                    }
                    else {
                        ts.Debug.fail("invalid action");
                    }
                    var edits = ts.textChanges.ChangeTracker.with(context, function (t) {
                        t.replaceNode(file, func.body, body);
                    });
                    return { renameFilename: undefined, renameLocation: undefined, edits: edits };
                }
                function getConvertibleArrowFunctionAtPosition(file, startPosition, considerFunctionBodies, kind) {
                    if (considerFunctionBodies === void 0) {
                        considerFunctionBodies = true;
                    }
                    var node = ts.getTokenAtPosition(file, startPosition);
                    var func = ts.getContainingFunction(node);
                    if (!func) {
                        return {
                            error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_find_a_containing_arrow_function)
                        };
                    }
                    if (!ts.isArrowFunction(func)) {
                        return {
                            error: ts.getLocaleSpecificMessage(ts.Diagnostics.Containing_function_is_not_an_arrow_function)
                        };
                    }
                    if ((!ts.rangeContainsRange(func, node) || ts.rangeContainsRange(func.body, node) && !considerFunctionBodies)) {
                        return undefined;
                    }
                    if (refactor.refactorKindBeginsWith(addBracesAction.kind, kind) && ts.isExpression(func.body)) {
                        return { func: func, addBraces: true, expression: func.body };
                    }
                    else if (refactor.refactorKindBeginsWith(removeBracesAction.kind, kind) && ts.isBlock(func.body) && func.body.statements.length === 1) {
                        var firstStatement = ts.first(func.body.statements);
                        if (ts.isReturnStatement(firstStatement)) {
                            return { func: func, addBraces: false, expression: firstStatement.expression, returnStatement: firstStatement };
                        }
                    }
                    return undefined;
                }
            })(addOrRemoveBracesToArrowFunction = refactor.addOrRemoveBracesToArrowFunction || (refactor.addOrRemoveBracesToArrowFunction = {}));
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var convertParamsToDestructuredObject;
            (function (convertParamsToDestructuredObject) {
                var refactorName = "Convert parameters to destructured object";
                var minimumParameterLength = 2;
                var refactorDescription = ts.getLocaleSpecificMessage(ts.Diagnostics.Convert_parameters_to_destructured_object);
                var toDestructuredAction = {
                    name: refactorName,
                    description: refactorDescription,
                    kind: "refactor.rewrite.parameters.toDestructured"
                };
                refactor.registerRefactor(refactorName, {
                    kinds: [toDestructuredAction.kind],
                    getEditsForAction: getEditsForAction,
                    getAvailableActions: getAvailableActions
                });
                function getAvailableActions(context) {
                    var file = context.file, startPosition = context.startPosition;
                    var isJSFile = ts.isSourceFileJS(file);
                    if (isJSFile)
                        return ts.emptyArray; // TODO: GH#30113
                    var functionDeclaration = getFunctionDeclarationAtPosition(file, startPosition, context.program.getTypeChecker());
                    if (!functionDeclaration)
                        return ts.emptyArray;
                    return [{
                            name: refactorName,
                            description: refactorDescription,
                            actions: [toDestructuredAction]
                        }];
                }
                function getEditsForAction(context, actionName) {
                    ts.Debug.assert(actionName === refactorName, "Unexpected action name");
                    var file = context.file, startPosition = context.startPosition, program = context.program, cancellationToken = context.cancellationToken, host = context.host;
                    var functionDeclaration = getFunctionDeclarationAtPosition(file, startPosition, program.getTypeChecker());
                    if (!functionDeclaration || !cancellationToken)
                        return undefined;
                    var groupedReferences = getGroupedReferences(functionDeclaration, program, cancellationToken);
                    if (groupedReferences.valid) {
                        var edits = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(file, program, host, t, functionDeclaration, groupedReferences); });
                        return { renameFilename: undefined, renameLocation: undefined, edits: edits };
                    }
                    return { edits: [] }; // TODO: GH#30113
                }
                function doChange(sourceFile, program, host, changes, functionDeclaration, groupedReferences) {
                    var signature = groupedReferences.signature;
                    var newFunctionDeclarationParams = ts.map(createNewParameters(functionDeclaration, program, host), function (param) { return ts.getSynthesizedDeepClone(param); });
                    if (signature) {
                        var newSignatureParams = ts.map(createNewParameters(signature, program, host), function (param) { return ts.getSynthesizedDeepClone(param); });
                        replaceParameters(signature, newSignatureParams);
                    }
                    replaceParameters(functionDeclaration, newFunctionDeclarationParams);
                    var functionCalls = ts.sortAndDeduplicate(groupedReferences.functionCalls, /*comparer*/ function (a, b) { return ts.compareValues(a.pos, b.pos); });
                    for (var _i = 0, functionCalls_1 = functionCalls; _i < functionCalls_1.length; _i++) {
                        var call = functionCalls_1[_i];
                        if (call.arguments && call.arguments.length) {
                            var newArgument = ts.getSynthesizedDeepClone(createNewArgument(functionDeclaration, call.arguments), /*includeTrivia*/ true);
                            changes.replaceNodeRange(ts.getSourceFileOfNode(call), ts.first(call.arguments), ts.last(call.arguments), newArgument, { leadingTriviaOption: ts.textChanges.LeadingTriviaOption.IncludeAll, trailingTriviaOption: ts.textChanges.TrailingTriviaOption.Include });
                        }
                    }
                    function replaceParameters(declarationOrSignature, parameterDeclarations) {
                        changes.replaceNodeRangeWithNodes(sourceFile, ts.first(declarationOrSignature.parameters), ts.last(declarationOrSignature.parameters), parameterDeclarations, {
                            joiner: ", ",
                            // indentation is set to 0 because otherwise the object parameter will be indented if there is a `this` parameter
                            indentation: 0,
                            leadingTriviaOption: ts.textChanges.LeadingTriviaOption.IncludeAll,
                            trailingTriviaOption: ts.textChanges.TrailingTriviaOption.Include
                        });
                    }
                }
                function getGroupedReferences(functionDeclaration, program, cancellationToken) {
                    var functionNames = getFunctionNames(functionDeclaration);
                    var classNames = ts.isConstructorDeclaration(functionDeclaration) ? getClassNames(functionDeclaration) : [];
                    var names = ts.deduplicate(__spreadArray(__spreadArray([], functionNames, true), classNames, true), ts.equateValues);
                    var checker = program.getTypeChecker();
                    var references = ts.flatMap(names, /*mapfn*/ function (/*mapfn*/ name) { return ts.FindAllReferences.getReferenceEntriesForNode(-1, name, program, program.getSourceFiles(), cancellationToken); });
                    var groupedReferences = groupReferences(references);
                    if (!ts.every(groupedReferences.declarations, /*callback*/ function (/*callback*/ decl) { return ts.contains(names, decl); })) {
                        groupedReferences.valid = false;
                    }
                    return groupedReferences;
                    function groupReferences(referenceEntries) {
                        var classReferences = { accessExpressions: [], typeUsages: [] };
                        var groupedReferences = { functionCalls: [], declarations: [], classReferences: classReferences, valid: true };
                        var functionSymbols = ts.map(functionNames, getSymbolTargetAtLocation);
                        var classSymbols = ts.map(classNames, getSymbolTargetAtLocation);
                        var isConstructor = ts.isConstructorDeclaration(functionDeclaration);
                        var contextualSymbols = ts.map(functionNames, function (name) { return getSymbolForContextualType(name, checker); });
                        for (var _i = 0, referenceEntries_1 = referenceEntries; _i < referenceEntries_1.length; _i++) {
                            var entry = referenceEntries_1[_i];
                            if (entry.kind === 0 /* Span */) {
                                groupedReferences.valid = false;
                                continue;
                            }
                            /* Declarations in object literals may be implementations of method signatures which have a different symbol from the declaration
                            For example:
                                interface IFoo { m(a: number): void }
                                const foo: IFoo = { m(a: number): void {} }
                            In these cases we get the symbol for the signature from the contextual type.
                            */
                            if (ts.contains(contextualSymbols, getSymbolTargetAtLocation(entry.node))) {
                                if (isValidMethodSignature(entry.node.parent)) {
                                    groupedReferences.signature = entry.node.parent;
                                    continue;
                                }
                                var call = entryToFunctionCall(entry);
                                if (call) {
                                    groupedReferences.functionCalls.push(call);
                                    continue;
                                }
                            }
                            var contextualSymbol = getSymbolForContextualType(entry.node, checker);
                            if (contextualSymbol && ts.contains(contextualSymbols, contextualSymbol)) {
                                var decl = entryToDeclaration(entry);
                                if (decl) {
                                    groupedReferences.declarations.push(decl);
                                    continue;
                                }
                            }
                            /* We compare symbols because in some cases find all references wil return a reference that may or may not be to the refactored function.
                            Example from the refactorConvertParamsToDestructuredObject_methodCallUnion.ts test:
                                class A { foo(a: number, b: number) { return a + b; } }
                                class B { foo(c: number, d: number) { return c + d; } }
                                declare const ab: A | B;
                                ab.foo(1, 2);
                            Find all references will return `ab.foo(1, 2)` as a reference to A's `foo` but we could be calling B's `foo`.
                            When looking for constructor calls, however, the symbol on the constructor call reference is going to be the corresponding class symbol.
                            So we need to add a special case for this because when calling a constructor of a class through one of its subclasses,
                            the symbols are going to be different.
                            */
                            if (ts.contains(functionSymbols, getSymbolTargetAtLocation(entry.node)) || ts.isNewExpressionTarget(entry.node)) {
                                var importOrExportReference = entryToImportOrExport(entry);
                                if (importOrExportReference) {
                                    continue;
                                }
                                var decl = entryToDeclaration(entry);
                                if (decl) {
                                    groupedReferences.declarations.push(decl);
                                    continue;
                                }
                                var call = entryToFunctionCall(entry);
                                if (call) {
                                    groupedReferences.functionCalls.push(call);
                                    continue;
                                }
                            }
                            // if the refactored function is a constructor, we must also check if the references to its class are valid
                            if (isConstructor && ts.contains(classSymbols, getSymbolTargetAtLocation(entry.node))) {
                                var importOrExportReference = entryToImportOrExport(entry);
                                if (importOrExportReference) {
                                    continue;
                                }
                                var decl = entryToDeclaration(entry);
                                if (decl) {
                                    groupedReferences.declarations.push(decl);
                                    continue;
                                }
                                var accessExpression = entryToAccessExpression(entry);
                                if (accessExpression) {
                                    classReferences.accessExpressions.push(accessExpression);
                                    continue;
                                }
                                // Only class declarations are allowed to be used as a type (in a heritage clause),
                                // otherwise `findAllReferences` might not be able to track constructor calls.
                                if (ts.isClassDeclaration(functionDeclaration.parent)) {
                                    var type = entryToType(entry);
                                    if (type) {
                                        classReferences.typeUsages.push(type);
                                        continue;
                                    }
                                }
                            }
                            groupedReferences.valid = false;
                        }
                        return groupedReferences;
                    }
                    function getSymbolTargetAtLocation(node) {
                        var symbol = checker.getSymbolAtLocation(node);
                        return symbol && ts.getSymbolTarget(symbol, checker);
                    }
                }
                /**
                 * Gets the symbol for the contextual type of the node if it is not a union or intersection.
                 */
                function getSymbolForContextualType(node, checker) {
                    var element = ts.getContainingObjectLiteralElement(node);
                    if (element) {
                        var contextualType = checker.getContextualTypeForObjectLiteralElement(element);
                        var symbol = contextualType === null || contextualType === void 0 ? void 0 : contextualType.getSymbol();
                        if (symbol && !(ts.getCheckFlags(symbol) & 6 /* Synthetic */)) {
                            return symbol;
                        }
                    }
                }
                function entryToImportOrExport(entry) {
                    var node = entry.node;
                    if (ts.isImportSpecifier(node.parent)
                        || ts.isImportClause(node.parent)
                        || ts.isImportEqualsDeclaration(node.parent)
                        || ts.isNamespaceImport(node.parent)) {
                        return node;
                    }
                    if (ts.isExportSpecifier(node.parent) || ts.isExportAssignment(node.parent)) {
                        return node;
                    }
                    return undefined;
                }
                function entryToDeclaration(entry) {
                    if (ts.isDeclaration(entry.node.parent)) {
                        return entry.node;
                    }
                    return undefined;
                }
                function entryToFunctionCall(entry) {
                    if (entry.node.parent) {
                        var functionReference = entry.node;
                        var parent = functionReference.parent;
                        switch (parent.kind) {
                            // foo(...) or super(...) or new Foo(...)
                            case 206 /* CallExpression */:
                            case 207 /* NewExpression */:
                                var callOrNewExpression = ts.tryCast(parent, ts.isCallOrNewExpression);
                                if (callOrNewExpression && callOrNewExpression.expression === functionReference) {
                                    return callOrNewExpression;
                                }
                                break;
                            // x.foo(...)
                            case 204 /* PropertyAccessExpression */:
                                var propertyAccessExpression = ts.tryCast(parent, ts.isPropertyAccessExpression);
                                if (propertyAccessExpression && propertyAccessExpression.parent && propertyAccessExpression.name === functionReference) {
                                    var callOrNewExpression_1 = ts.tryCast(propertyAccessExpression.parent, ts.isCallOrNewExpression);
                                    if (callOrNewExpression_1 && callOrNewExpression_1.expression === propertyAccessExpression) {
                                        return callOrNewExpression_1;
                                    }
                                }
                                break;
                            // x["foo"](...)
                            case 205 /* ElementAccessExpression */:
                                var elementAccessExpression = ts.tryCast(parent, ts.isElementAccessExpression);
                                if (elementAccessExpression && elementAccessExpression.parent && elementAccessExpression.argumentExpression === functionReference) {
                                    var callOrNewExpression_2 = ts.tryCast(elementAccessExpression.parent, ts.isCallOrNewExpression);
                                    if (callOrNewExpression_2 && callOrNewExpression_2.expression === elementAccessExpression) {
                                        return callOrNewExpression_2;
                                    }
                                }
                                break;
                        }
                    }
                    return undefined;
                }
                function entryToAccessExpression(entry) {
                    if (entry.node.parent) {
                        var reference = entry.node;
                        var parent = reference.parent;
                        switch (parent.kind) {
                            // `C.foo`
                            case 204 /* PropertyAccessExpression */:
                                var propertyAccessExpression = ts.tryCast(parent, ts.isPropertyAccessExpression);
                                if (propertyAccessExpression && propertyAccessExpression.expression === reference) {
                                    return propertyAccessExpression;
                                }
                                break;
                            // `C["foo"]`
                            case 205 /* ElementAccessExpression */:
                                var elementAccessExpression = ts.tryCast(parent, ts.isElementAccessExpression);
                                if (elementAccessExpression && elementAccessExpression.expression === reference) {
                                    return elementAccessExpression;
                                }
                                break;
                        }
                    }
                    return undefined;
                }
                function entryToType(entry) {
                    var reference = entry.node;
                    if (ts.getMeaningFromLocation(reference) === 2 /* Type */ || ts.isExpressionWithTypeArgumentsInClassExtendsClause(reference.parent)) {
                        return reference;
                    }
                    return undefined;
                }
                function getFunctionDeclarationAtPosition(file, startPosition, checker) {
                    var node = ts.getTouchingToken(file, startPosition);
                    var functionDeclaration = ts.getContainingFunctionDeclaration(node);
                    // don't offer refactor on top-level JSDoc
                    if (isTopLevelJSDoc(node))
                        return undefined;
                    if (functionDeclaration
                        && isValidFunctionDeclaration(functionDeclaration, checker)
                        && ts.rangeContainsRange(functionDeclaration, node)
                        && !(functionDeclaration.body && ts.rangeContainsRange(functionDeclaration.body, node)))
                        return functionDeclaration;
                    return undefined;
                }
                function isTopLevelJSDoc(node) {
                    var containingJSDoc = ts.findAncestor(node, ts.isJSDocNode);
                    if (containingJSDoc) {
                        var containingNonJSDoc = ts.findAncestor(containingJSDoc, function (n) { return !ts.isJSDocNode(n); });
                        return !!containingNonJSDoc && ts.isFunctionLikeDeclaration(containingNonJSDoc);
                    }
                    return false;
                }
                function isValidMethodSignature(node) {
                    return ts.isMethodSignature(node) && (ts.isInterfaceDeclaration(node.parent) || ts.isTypeLiteralNode(node.parent));
                }
                function isValidFunctionDeclaration(functionDeclaration, checker) {
                    var _a;
                    if (!isValidParameterNodeArray(functionDeclaration.parameters, checker))
                        return false;
                    switch (functionDeclaration.kind) {
                        case 254 /* FunctionDeclaration */:
                            return hasNameOrDefault(functionDeclaration) && isSingleImplementation(functionDeclaration, checker);
                        case 167 /* MethodDeclaration */:
                            if (ts.isObjectLiteralExpression(functionDeclaration.parent)) {
                                var contextualSymbol = getSymbolForContextualType(functionDeclaration.name, checker);
                                // don't offer the refactor when there are multiple signatures since we won't know which ones the user wants to change
                                return ((_a = contextualSymbol === null || contextualSymbol === void 0 ? void 0 : contextualSymbol.declarations) === null || _a === void 0 ? void 0 : _a.length) === 1 && isSingleImplementation(functionDeclaration, checker);
                            }
                            return isSingleImplementation(functionDeclaration, checker);
                        case 169 /* Constructor */:
                            if (ts.isClassDeclaration(functionDeclaration.parent)) {
                                return hasNameOrDefault(functionDeclaration.parent) && isSingleImplementation(functionDeclaration, checker);
                            }
                            else {
                                return isValidVariableDeclaration(functionDeclaration.parent.parent)
                                    && isSingleImplementation(functionDeclaration, checker);
                            }
                        case 211 /* FunctionExpression */:
                        case 212 /* ArrowFunction */:
                            return isValidVariableDeclaration(functionDeclaration.parent);
                    }
                    return false;
                }
                function isSingleImplementation(functionDeclaration, checker) {
                    return !!functionDeclaration.body && !checker.isImplementationOfOverload(functionDeclaration);
                }
                function hasNameOrDefault(functionOrClassDeclaration) {
                    if (!functionOrClassDeclaration.name) {
                        var defaultKeyword = ts.findModifier(functionOrClassDeclaration, 88 /* DefaultKeyword */);
                        return !!defaultKeyword;
                    }
                    return true;
                }
                function isValidParameterNodeArray(parameters, checker) {
                    return getRefactorableParametersLength(parameters) >= minimumParameterLength
                        && ts.every(parameters, /*callback*/ function (/*callback*/ paramDecl) { return isValidParameterDeclaration(paramDecl, checker); });
                }
                function isValidParameterDeclaration(parameterDeclaration, checker) {
                    if (ts.isRestParameter(parameterDeclaration)) {
                        var type = checker.getTypeAtLocation(parameterDeclaration);
                        if (!checker.isArrayType(type) && !checker.isTupleType(type))
                            return false;
                    }
                    return !parameterDeclaration.modifiers && !parameterDeclaration.decorators && ts.isIdentifier(parameterDeclaration.name);
                }
                function isValidVariableDeclaration(node) {
                    return ts.isVariableDeclaration(node) && ts.isVarConst(node) && ts.isIdentifier(node.name) && !node.type; // TODO: GH#30113
                }
                function hasThisParameter(parameters) {
                    return parameters.length > 0 && ts.isThis(parameters[0].name);
                }
                function getRefactorableParametersLength(parameters) {
                    if (hasThisParameter(parameters)) {
                        return parameters.length - 1;
                    }
                    return parameters.length;
                }
                function getRefactorableParameters(parameters) {
                    if (hasThisParameter(parameters)) {
                        parameters = ts.factory.createNodeArray(parameters.slice(1), parameters.hasTrailingComma);
                    }
                    return parameters;
                }
                function createPropertyOrShorthandAssignment(name, initializer) {
                    if (ts.isIdentifier(initializer) && ts.getTextOfIdentifierOrLiteral(initializer) === name) {
                        return ts.factory.createShorthandPropertyAssignment(name);
                    }
                    return ts.factory.createPropertyAssignment(name, initializer);
                }
                function createNewArgument(functionDeclaration, functionArguments) {
                    var parameters = getRefactorableParameters(functionDeclaration.parameters);
                    var hasRestParameter = ts.isRestParameter(ts.last(parameters));
                    var nonRestArguments = hasRestParameter ? functionArguments.slice(0, parameters.length - 1) : functionArguments;
                    var properties = ts.map(nonRestArguments, function (arg, i) {
                        var parameterName = getParameterName(parameters[i]);
                        var property = createPropertyOrShorthandAssignment(parameterName, arg);
                        ts.suppressLeadingAndTrailingTrivia(property.name);
                        if (ts.isPropertyAssignment(property))
                            ts.suppressLeadingAndTrailingTrivia(property.initializer);
                        ts.copyComments(arg, property);
                        return property;
                    });
                    if (hasRestParameter && functionArguments.length >= parameters.length) {
                        var restArguments = functionArguments.slice(parameters.length - 1);
                        var restProperty = ts.factory.createPropertyAssignment(getParameterName(ts.last(parameters)), ts.factory.createArrayLiteralExpression(restArguments));
                        properties.push(restProperty);
                    }
                    var objectLiteral = ts.factory.createObjectLiteralExpression(properties, /*multiLine*/ false);
                    return objectLiteral;
                }
                function createNewParameters(functionDeclaration, program, host) {
                    var checker = program.getTypeChecker();
                    var refactorableParameters = getRefactorableParameters(functionDeclaration.parameters);
                    var bindingElements = ts.map(refactorableParameters, createBindingElementFromParameterDeclaration);
                    var objectParameterName = ts.factory.createObjectBindingPattern(bindingElements);
                    var objectParameterType = createParameterTypeNode(refactorableParameters);
                    var objectInitializer;
                    // If every parameter in the original function was optional, add an empty object initializer to the new object parameter
                    if (ts.every(refactorableParameters, isOptionalParameter)) {
                        objectInitializer = ts.factory.createObjectLiteralExpression();
                    }
                    var objectParameter = ts.factory.createParameterDeclaration(
                    /*decorators*/ undefined, 
                    /*modifiers*/ undefined, 
                    /*dotDotDotToken*/ undefined, objectParameterName, 
                    /*questionToken*/ undefined, objectParameterType, objectInitializer);
                    if (hasThisParameter(functionDeclaration.parameters)) {
                        var thisParameter = functionDeclaration.parameters[0];
                        var newThisParameter = ts.factory.createParameterDeclaration(
                        /*decorators*/ undefined, 
                        /*modifiers*/ undefined, 
                        /*dotDotDotToken*/ undefined, thisParameter.name, 
                        /*questionToken*/ undefined, thisParameter.type);
                        ts.suppressLeadingAndTrailingTrivia(newThisParameter.name);
                        ts.copyComments(thisParameter.name, newThisParameter.name);
                        if (thisParameter.type) {
                            ts.suppressLeadingAndTrailingTrivia(newThisParameter.type);
                            ts.copyComments(thisParameter.type, newThisParameter.type);
                        }
                        return ts.factory.createNodeArray([newThisParameter, objectParameter]);
                    }
                    return ts.factory.createNodeArray([objectParameter]);
                    function createBindingElementFromParameterDeclaration(parameterDeclaration) {
                        var element = ts.factory.createBindingElement(
                        /*dotDotDotToken*/ undefined, 
                        /*propertyName*/ undefined, getParameterName(parameterDeclaration), ts.isRestParameter(parameterDeclaration) && isOptionalParameter(parameterDeclaration) ? ts.factory.createArrayLiteralExpression() : parameterDeclaration.initializer);
                        ts.suppressLeadingAndTrailingTrivia(element);
                        if (parameterDeclaration.initializer && element.initializer) {
                            ts.copyComments(parameterDeclaration.initializer, element.initializer);
                        }
                        return element;
                    }
                    function createParameterTypeNode(parameters) {
                        var members = ts.map(parameters, createPropertySignatureFromParameterDeclaration);
                        var typeNode = ts.addEmitFlags(ts.factory.createTypeLiteralNode(members), 1 /* SingleLine */);
                        return typeNode;
                    }
                    function createPropertySignatureFromParameterDeclaration(parameterDeclaration) {
                        var parameterType = parameterDeclaration.type;
                        if (!parameterType && (parameterDeclaration.initializer || ts.isRestParameter(parameterDeclaration))) {
                            parameterType = getTypeNode(parameterDeclaration);
                        }
                        var propertySignature = ts.factory.createPropertySignature(
                        /*modifiers*/ undefined, getParameterName(parameterDeclaration), isOptionalParameter(parameterDeclaration) ? ts.factory.createToken(57 /* QuestionToken */) : parameterDeclaration.questionToken, parameterType);
                        ts.suppressLeadingAndTrailingTrivia(propertySignature);
                        ts.copyComments(parameterDeclaration.name, propertySignature.name);
                        if (parameterDeclaration.type && propertySignature.type) {
                            ts.copyComments(parameterDeclaration.type, propertySignature.type);
                        }
                        return propertySignature;
                    }
                    function getTypeNode(node) {
                        var type = checker.getTypeAtLocation(node);
                        return ts.getTypeNodeIfAccessible(type, node, program, host);
                    }
                    function isOptionalParameter(parameterDeclaration) {
                        if (ts.isRestParameter(parameterDeclaration)) {
                            var type = checker.getTypeAtLocation(parameterDeclaration);
                            return !checker.isTupleType(type);
                        }
                        return checker.isOptionalParameter(parameterDeclaration);
                    }
                }
                function getParameterName(paramDeclaration) {
                    return ts.getTextOfIdentifierOrLiteral(paramDeclaration.name);
                }
                function getClassNames(constructorDeclaration) {
                    switch (constructorDeclaration.parent.kind) {
                        case 255 /* ClassDeclaration */:
                            var classDeclaration = constructorDeclaration.parent;
                            if (classDeclaration.name)
                                return [classDeclaration.name];
                            // If the class declaration doesn't have a name, it should have a default modifier.
                            // We validated this in `isValidFunctionDeclaration` through `hasNameOrDefault`
                            var defaultModifier = ts.Debug.checkDefined(ts.findModifier(classDeclaration, 88 /* DefaultKeyword */), "Nameless class declaration should be a default export");
                            return [defaultModifier];
                        case 224 /* ClassExpression */:
                            var classExpression = constructorDeclaration.parent;
                            var variableDeclaration = constructorDeclaration.parent.parent;
                            var className = classExpression.name;
                            if (className)
                                return [className, variableDeclaration.name];
                            return [variableDeclaration.name];
                    }
                }
                function getFunctionNames(functionDeclaration) {
                    switch (functionDeclaration.kind) {
                        case 254 /* FunctionDeclaration */:
                            if (functionDeclaration.name)
                                return [functionDeclaration.name];
                            // If the function declaration doesn't have a name, it should have a default modifier.
                            // We validated this in `isValidFunctionDeclaration` through `hasNameOrDefault`
                            var defaultModifier = ts.Debug.checkDefined(ts.findModifier(functionDeclaration, 88 /* DefaultKeyword */), "Nameless function declaration should be a default export");
                            return [defaultModifier];
                        case 167 /* MethodDeclaration */:
                            return [functionDeclaration.name];
                        case 169 /* Constructor */:
                            var ctrKeyword = ts.Debug.checkDefined(ts.findChildOfKind(functionDeclaration, 133 /* ConstructorKeyword */, functionDeclaration.getSourceFile()), "Constructor declaration should have constructor keyword");
                            if (functionDeclaration.parent.kind === 224 /* ClassExpression */) {
                                var variableDeclaration = functionDeclaration.parent.parent;
                                return [variableDeclaration.name, ctrKeyword];
                            }
                            return [ctrKeyword];
                        case 212 /* ArrowFunction */:
                            return [functionDeclaration.parent.name];
                        case 211 /* FunctionExpression */:
                            if (functionDeclaration.name)
                                return [functionDeclaration.name, functionDeclaration.parent.name];
                            return [functionDeclaration.parent.name];
                        default:
                            return ts.Debug.assertNever(functionDeclaration, "Unexpected function declaration kind " + functionDeclaration.kind);
                    }
                }
            })(convertParamsToDestructuredObject = refactor.convertParamsToDestructuredObject || (refactor.convertParamsToDestructuredObject = {}));
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var convertStringOrTemplateLiteral;
            (function (convertStringOrTemplateLiteral) {
                var refactorName = "Convert to template string";
                var refactorDescription = ts.getLocaleSpecificMessage(ts.Diagnostics.Convert_to_template_string);
                var convertStringAction = {
                    name: refactorName,
                    description: refactorDescription,
                    kind: "refactor.rewrite.string"
                };
                refactor.registerRefactor(refactorName, {
                    kinds: [convertStringAction.kind],
                    getEditsForAction: getEditsForAction,
                    getAvailableActions: getAvailableActions
                });
                function getAvailableActions(context) {
                    var file = context.file, startPosition = context.startPosition;
                    var node = getNodeOrParentOfParentheses(file, startPosition);
                    var maybeBinary = getParentBinaryExpression(node);
                    var refactorInfo = { name: refactorName, description: refactorDescription, actions: [] };
                    if (ts.isBinaryExpression(maybeBinary) && treeToArray(maybeBinary).isValidConcatenation) {
                        refactorInfo.actions.push(convertStringAction);
                        return [refactorInfo];
                    }
                    else if (context.preferences.provideRefactorNotApplicableReason) {
                        refactorInfo.actions.push(__assign(__assign({}, convertStringAction), { notApplicableReason: ts.getLocaleSpecificMessage(ts.Diagnostics.Can_only_convert_string_concatenation) }));
                        return [refactorInfo];
                    }
                    return ts.emptyArray;
                }
                function getNodeOrParentOfParentheses(file, startPosition) {
                    var node = ts.getTokenAtPosition(file, startPosition);
                    var nestedBinary = getParentBinaryExpression(node);
                    var isNonStringBinary = !treeToArray(nestedBinary).isValidConcatenation;
                    if (isNonStringBinary &&
                        ts.isParenthesizedExpression(nestedBinary.parent) &&
                        ts.isBinaryExpression(nestedBinary.parent.parent)) {
                        return nestedBinary.parent.parent;
                    }
                    return node;
                }
                function getEditsForAction(context, actionName) {
                    var file = context.file, startPosition = context.startPosition;
                    var node = getNodeOrParentOfParentheses(file, startPosition);
                    switch (actionName) {
                        case refactorDescription:
                            return { edits: getEditsForToTemplateLiteral(context, node) };
                        default:
                            return ts.Debug.fail("invalid action");
                    }
                }
                function getEditsForToTemplateLiteral(context, node) {
                    var maybeBinary = getParentBinaryExpression(node);
                    var file = context.file;
                    var templateLiteral = nodesToTemplate(treeToArray(maybeBinary), file);
                    var trailingCommentRanges = ts.getTrailingCommentRanges(file.text, maybeBinary.end);
                    if (trailingCommentRanges) {
                        var lastComment = trailingCommentRanges[trailingCommentRanges.length - 1];
                        var trailingRange_1 = { pos: trailingCommentRanges[0].pos, end: lastComment.end };
                        // since suppressTrailingTrivia(maybeBinary) does not work, the trailing comment is removed manually
                        // otherwise it would have the trailing comment twice
                        return ts.textChanges.ChangeTracker.with(context, function (t) {
                            t.deleteRange(file, trailingRange_1);
                            t.replaceNode(file, maybeBinary, templateLiteral);
                        });
                    }
                    else {
                        return ts.textChanges.ChangeTracker.with(context, function (t) { return t.replaceNode(file, maybeBinary, templateLiteral); });
                    }
                }
                function isNotEqualsOperator(node) {
                    return node.operatorToken.kind !== 63 /* EqualsToken */;
                }
                function getParentBinaryExpression(expr) {
                    var container = ts.findAncestor(expr.parent, function (n) {
                        switch (n.kind) {
                            case 204 /* PropertyAccessExpression */:
                            case 205 /* ElementAccessExpression */:
                                return false;
                            case 221 /* TemplateExpression */:
                            case 219 /* BinaryExpression */:
                                return !(ts.isBinaryExpression(n.parent) && isNotEqualsOperator(n.parent));
                            default:
                                return "quit";
                        }
                    });
                    return (container || expr);
                }
                function treeToArray(current) {
                    var loop = function (current) {
                        if (!ts.isBinaryExpression(current)) {
                            return { nodes: [current], operators: [], validOperators: true,
                                hasString: ts.isStringLiteral(current) || ts.isNoSubstitutionTemplateLiteral(current) };
                        }
                        var _a = loop(current.left), nodes = _a.nodes, operators = _a.operators, leftHasString = _a.hasString, leftOperatorValid = _a.validOperators;
                        if (!(leftHasString || ts.isStringLiteral(current.right) || ts.isTemplateExpression(current.right))) {
                            return { nodes: [current], operators: [], hasString: false, validOperators: true };
                        }
                        var currentOperatorValid = current.operatorToken.kind === 39 /* PlusToken */;
                        var validOperators = leftOperatorValid && currentOperatorValid;
                        nodes.push(current.right);
                        operators.push(current.operatorToken);
                        return { nodes: nodes, operators: operators, hasString: true, validOperators: validOperators };
                    };
                    var _a = loop(current), nodes = _a.nodes, operators = _a.operators, validOperators = _a.validOperators, hasString = _a.hasString;
                    return { nodes: nodes, operators: operators, isValidConcatenation: validOperators && hasString };
                }
                // to copy comments following the operator
                // "foo" + /* comment */ "bar"
                var copyTrailingOperatorComments = function (operators, file) {
                    return function (index, targetNode) {
                        if (index < operators.length) {
                            ts.copyTrailingComments(operators[index], targetNode, file, 3 /* MultiLineCommentTrivia */, /* hasTrailingNewLine */ false);
                        }
                    };
                };
                // to copy comments following the string
                // "foo" /* comment */ + "bar" /* comment */ + "bar2"
                var copyCommentFromMultiNode = function (nodes, file, copyOperatorComments) {
                    return function (indexes, targetNode) {
                        while (indexes.length > 0) {
                            var index = indexes.shift();
                            ts.copyTrailingComments(nodes[index], targetNode, file, 3 /* MultiLineCommentTrivia */, /* hasTrailingNewLine */ false);
                            copyOperatorComments(index, targetNode);
                        }
                    };
                };
                function escapeRawStringForTemplate(s) {
                    // Escaping for $s in strings that are to be used in template strings
                    // Naive implementation: replace \x by itself and otherwise $ and ` by \$ and \`.
                    // But to complicate it a bit, this should work for raw strings too.
                    return s.replace(/\\.|[$`]/g, function (m) { return m[0] === "\\" ? m : "\\" + m; });
                    // Finally, a less-backslash-happy version can work too, doing only ${ instead of all $s:
                    //     s.replace(/\\.|\${|`/g, m => m[0] === "\\" ? m : "\\" + m);
                    // but `\$${foo}` is likely more clear than the more-confusing-but-still-working `$${foo}`.
                }
                function getRawTextOfTemplate(node) {
                    // in these cases the right side is ${
                    var rightShaving = ts.isTemplateHead(node) || ts.isTemplateMiddle(node) ? -2 : -1;
                    return ts.getTextOfNode(node).slice(1, rightShaving);
                }
                function concatConsecutiveString(index, nodes) {
                    var indexes = [];
                    var text = "", rawText = "";
                    while (index < nodes.length) {
                        var node = nodes[index];
                        if (ts.isStringLiteralLike(node)) { // includes isNoSubstitutionTemplateLiteral(node)
                            text += node.text;
                            rawText += escapeRawStringForTemplate(ts.getTextOfNode(node).slice(1, -1));
                            indexes.push(index);
                            index++;
                        }
                        else if (ts.isTemplateExpression(node)) {
                            text += node.head.text;
                            rawText += getRawTextOfTemplate(node.head);
                            break;
                        }
                        else {
                            break;
                        }
                    }
                    return [index, text, rawText, indexes];
                }
                function nodesToTemplate(_a, file) {
                    var nodes = _a.nodes, operators = _a.operators;
                    var copyOperatorComments = copyTrailingOperatorComments(operators, file);
                    var copyCommentFromStringLiterals = copyCommentFromMultiNode(nodes, file, copyOperatorComments);
                    var _b = concatConsecutiveString(0, nodes), begin = _b[0], headText = _b[1], rawHeadText = _b[2], headIndexes = _b[3];
                    if (begin === nodes.length) {
                        var noSubstitutionTemplateLiteral = ts.factory.createNoSubstitutionTemplateLiteral(headText, rawHeadText);
                        copyCommentFromStringLiterals(headIndexes, noSubstitutionTemplateLiteral);
                        return noSubstitutionTemplateLiteral;
                    }
                    var templateSpans = [];
                    var templateHead = ts.factory.createTemplateHead(headText, rawHeadText);
                    copyCommentFromStringLiterals(headIndexes, templateHead);
                    var _loop_20 = function (i) {
                        var currentNode = getExpressionFromParenthesesOrExpression(nodes[i]);
                        copyOperatorComments(i, currentNode);
                        var _c = concatConsecutiveString(i + 1, nodes), newIndex = _c[0], subsequentText = _c[1], rawSubsequentText = _c[2], stringIndexes = _c[3];
                        i = newIndex - 1;
                        var isLast = i === nodes.length - 1;
                        if (ts.isTemplateExpression(currentNode)) {
                            var spans = ts.map(currentNode.templateSpans, function (span, index) {
                                copyExpressionComments(span);
                                var isLastSpan = index === currentNode.templateSpans.length - 1;
                                var text = span.literal.text + (isLastSpan ? subsequentText : "");
                                var rawText = getRawTextOfTemplate(span.literal) + (isLastSpan ? rawSubsequentText : "");
                                return ts.factory.createTemplateSpan(span.expression, isLast
                                    ? ts.factory.createTemplateTail(text, rawText)
                                    : ts.factory.createTemplateMiddle(text, rawText));
                            });
                            templateSpans.push.apply(templateSpans, spans);
                        }
                        else {
                            var templatePart = isLast
                                ? ts.factory.createTemplateTail(subsequentText, rawSubsequentText)
                                : ts.factory.createTemplateMiddle(subsequentText, rawSubsequentText);
                            copyCommentFromStringLiterals(stringIndexes, templatePart);
                            templateSpans.push(ts.factory.createTemplateSpan(currentNode, templatePart));
                        }
                        out_i_1 = i;
                    };
                    var out_i_1;
                    for (var i = begin; i < nodes.length; i++) {
                        _loop_20(i);
                        i = out_i_1;
                    }
                    return ts.factory.createTemplateExpression(templateHead, templateSpans);
                }
                // to copy comments following the opening & closing parentheses
                // "foo" + ( /* comment */ 5 + 5 ) /* comment */ + "bar"
                function copyExpressionComments(node) {
                    var file = node.getSourceFile();
                    ts.copyTrailingComments(node, node.expression, file, 3 /* MultiLineCommentTrivia */, /* hasTrailingNewLine */ false);
                    ts.copyTrailingAsLeadingComments(node.expression, node.expression, file, 3 /* MultiLineCommentTrivia */, /* hasTrailingNewLine */ false);
                }
                function getExpressionFromParenthesesOrExpression(node) {
                    if (ts.isParenthesizedExpression(node)) {
                        copyExpressionComments(node);
                        node = node.expression;
                    }
                    return node;
                }
            })(convertStringOrTemplateLiteral = refactor.convertStringOrTemplateLiteral || (refactor.convertStringOrTemplateLiteral = {}));
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var convertArrowFunctionOrFunctionExpression;
            (function (convertArrowFunctionOrFunctionExpression) {
                var refactorName = "Convert arrow function or function expression";
                var refactorDescription = ts.getLocaleSpecificMessage(ts.Diagnostics.Convert_arrow_function_or_function_expression);
                var toAnonymousFunctionAction = {
                    name: "Convert to anonymous function",
                    description: ts.getLocaleSpecificMessage(ts.Diagnostics.Convert_to_anonymous_function),
                    kind: "refactor.rewrite.function.anonymous",
                };
                var toNamedFunctionAction = {
                    name: "Convert to named function",
                    description: ts.getLocaleSpecificMessage(ts.Diagnostics.Convert_to_named_function),
                    kind: "refactor.rewrite.function.named",
                };
                var toArrowFunctionAction = {
                    name: "Convert to arrow function",
                    description: ts.getLocaleSpecificMessage(ts.Diagnostics.Convert_to_arrow_function),
                    kind: "refactor.rewrite.function.arrow",
                };
                refactor.registerRefactor(refactorName, {
                    kinds: [
                        toAnonymousFunctionAction.kind,
                        toNamedFunctionAction.kind,
                        toArrowFunctionAction.kind
                    ],
                    getEditsForAction: getEditsForAction,
                    getAvailableActions: getAvailableActions
                });
                function getAvailableActions(context) {
                    var file = context.file, startPosition = context.startPosition, program = context.program, kind = context.kind;
                    var info = getFunctionInfo(file, startPosition, program);
                    if (!info)
                        return ts.emptyArray;
                    var selectedVariableDeclaration = info.selectedVariableDeclaration, func = info.func;
                    var possibleActions = [];
                    var errors = [];
                    if (refactor.refactorKindBeginsWith(toNamedFunctionAction.kind, kind)) {
                        var error = selectedVariableDeclaration || (ts.isArrowFunction(func) && ts.isVariableDeclaration(func.parent)) ?
                            undefined : ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_convert_to_named_function);
                        if (error) {
                            errors.push(__assign(__assign({}, toNamedFunctionAction), { notApplicableReason: error }));
                        }
                        else {
                            possibleActions.push(toNamedFunctionAction);
                        }
                    }
                    if (refactor.refactorKindBeginsWith(toAnonymousFunctionAction.kind, kind)) {
                        var error = !selectedVariableDeclaration && ts.isArrowFunction(func) ?
                            undefined : ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_convert_to_anonymous_function);
                        if (error) {
                            errors.push(__assign(__assign({}, toAnonymousFunctionAction), { notApplicableReason: error }));
                        }
                        else {
                            possibleActions.push(toAnonymousFunctionAction);
                        }
                    }
                    if (refactor.refactorKindBeginsWith(toArrowFunctionAction.kind, kind)) {
                        var error = ts.isFunctionExpression(func) ? undefined : ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_convert_to_arrow_function);
                        if (error) {
                            errors.push(__assign(__assign({}, toArrowFunctionAction), { notApplicableReason: error }));
                        }
                        else {
                            possibleActions.push(toArrowFunctionAction);
                        }
                    }
                    return [{
                            name: refactorName,
                            description: refactorDescription,
                            actions: possibleActions.length === 0 && context.preferences.provideRefactorNotApplicableReason ?
                                errors : possibleActions
                        }];
                }
                function getEditsForAction(context, actionName) {
                    var file = context.file, startPosition = context.startPosition, program = context.program;
                    var info = getFunctionInfo(file, startPosition, program);
                    if (!info)
                        return undefined;
                    var func = info.func;
                    var edits = [];
                    switch (actionName) {
                        case toAnonymousFunctionAction.name:
                            edits.push.apply(edits, getEditInfoForConvertToAnonymousFunction(context, func));
                            break;
                        case toNamedFunctionAction.name:
                            var variableInfo = getVariableInfo(func);
                            if (!variableInfo)
                                return undefined;
                            edits.push.apply(edits, getEditInfoForConvertToNamedFunction(context, func, variableInfo));
                            break;
                        case toArrowFunctionAction.name:
                            if (!ts.isFunctionExpression(func))
                                return undefined;
                            edits.push.apply(edits, getEditInfoForConvertToArrowFunction(context, func));
                            break;
                        default:
                            return ts.Debug.fail("invalid action");
                    }
                    return { renameFilename: undefined, renameLocation: undefined, edits: edits };
                }
                function containingThis(node) {
                    var containsThis = false;
                    node.forEachChild(function checkThis(child) {
                        if (ts.isThis(child)) {
                            containsThis = true;
                            return;
                        }
                        if (!ts.isClassLike(child) && !ts.isFunctionDeclaration(child) && !ts.isFunctionExpression(child)) {
                            ts.forEachChild(child, checkThis);
                        }
                    });
                    return containsThis;
                }
                function getFunctionInfo(file, startPosition, program) {
                    var token = ts.getTokenAtPosition(file, startPosition);
                    var typeChecker = program.getTypeChecker();
                    var func = tryGetFunctionFromVariableDeclaration(file, typeChecker, token.parent);
                    if (func && !containingThis(func.body) && !typeChecker.containsArgumentsReference(func)) {
                        return { selectedVariableDeclaration: true, func: func };
                    }
                    var maybeFunc = ts.getContainingFunction(token);
                    if (maybeFunc &&
                        (ts.isFunctionExpression(maybeFunc) || ts.isArrowFunction(maybeFunc)) &&
                        !ts.rangeContainsRange(maybeFunc.body, token) &&
                        !containingThis(maybeFunc.body) &&
                        !typeChecker.containsArgumentsReference(maybeFunc)) {
                        if (ts.isFunctionExpression(maybeFunc) && isFunctionReferencedInFile(file, typeChecker, maybeFunc))
                            return undefined;
                        return { selectedVariableDeclaration: false, func: maybeFunc };
                    }
                    return undefined;
                }
                function isSingleVariableDeclaration(parent) {
                    return ts.isVariableDeclaration(parent) || (ts.isVariableDeclarationList(parent) && parent.declarations.length === 1);
                }
                function tryGetFunctionFromVariableDeclaration(sourceFile, typeChecker, parent) {
                    if (!isSingleVariableDeclaration(parent)) {
                        return undefined;
                    }
                    var variableDeclaration = ts.isVariableDeclaration(parent) ? parent : ts.first(parent.declarations);
                    var initializer = variableDeclaration.initializer;
                    if (initializer && (ts.isArrowFunction(initializer) || ts.isFunctionExpression(initializer) && !isFunctionReferencedInFile(sourceFile, typeChecker, initializer))) {
                        return initializer;
                    }
                    return undefined;
                }
                function convertToBlock(body) {
                    if (ts.isExpression(body)) {
                        var returnStatement = ts.factory.createReturnStatement(body);
                        var file = body.getSourceFile();
                        ts.suppressLeadingAndTrailingTrivia(returnStatement);
                        ts.copyTrailingAsLeadingComments(body, returnStatement, file, /* commentKind */ undefined, /* hasTrailingNewLine */ true);
                        return ts.factory.createBlock([returnStatement], /* multiLine */ true);
                    }
                    else {
                        return body;
                    }
                }
                function getVariableInfo(func) {
                    var variableDeclaration = func.parent;
                    if (!ts.isVariableDeclaration(variableDeclaration) || !ts.isVariableDeclarationInVariableStatement(variableDeclaration))
                        return undefined;
                    var variableDeclarationList = variableDeclaration.parent;
                    var statement = variableDeclarationList.parent;
                    if (!ts.isVariableDeclarationList(variableDeclarationList) || !ts.isVariableStatement(statement) || !ts.isIdentifier(variableDeclaration.name))
                        return undefined;
                    return { variableDeclaration: variableDeclaration, variableDeclarationList: variableDeclarationList, statement: statement, name: variableDeclaration.name };
                }
                function getEditInfoForConvertToAnonymousFunction(context, func) {
                    var file = context.file;
                    var body = convertToBlock(func.body);
                    var newNode = ts.factory.createFunctionExpression(func.modifiers, func.asteriskToken, /* name */ undefined, func.typeParameters, func.parameters, func.type, body);
                    return ts.textChanges.ChangeTracker.with(context, function (t) { return t.replaceNode(file, func, newNode); });
                }
                function getEditInfoForConvertToNamedFunction(context, func, variableInfo) {
                    var file = context.file;
                    var body = convertToBlock(func.body);
                    var variableDeclaration = variableInfo.variableDeclaration, variableDeclarationList = variableInfo.variableDeclarationList, statement = variableInfo.statement, name = variableInfo.name;
                    ts.suppressLeadingTrivia(statement);
                    var modifiersFlags = (ts.getCombinedModifierFlags(variableDeclaration) & 1 /* Export */) | ts.getEffectiveModifierFlags(func);
                    var modifiers = ts.factory.createModifiersFromModifierFlags(modifiersFlags);
                    var newNode = ts.factory.createFunctionDeclaration(func.decorators, ts.length(modifiers) ? modifiers : undefined, func.asteriskToken, name, func.typeParameters, func.parameters, func.type, body);
                    if (variableDeclarationList.declarations.length === 1) {
                        return ts.textChanges.ChangeTracker.with(context, function (t) { return t.replaceNode(file, statement, newNode); });
                    }
                    else {
                        return ts.textChanges.ChangeTracker.with(context, function (t) {
                            t.delete(file, variableDeclaration);
                            t.insertNodeAfter(file, statement, newNode);
                        });
                    }
                }
                function getEditInfoForConvertToArrowFunction(context, func) {
                    var file = context.file;
                    var statements = func.body.statements;
                    var head = statements[0];
                    var body;
                    if (canBeConvertedToExpression(func.body, head)) {
                        body = head.expression;
                        ts.suppressLeadingAndTrailingTrivia(body);
                        ts.copyComments(head, body);
                    }
                    else {
                        body = func.body;
                    }
                    var newNode = ts.factory.createArrowFunction(func.modifiers, func.typeParameters, func.parameters, func.type, ts.factory.createToken(38 /* EqualsGreaterThanToken */), body);
                    return ts.textChanges.ChangeTracker.with(context, function (t) { return t.replaceNode(file, func, newNode); });
                }
                function canBeConvertedToExpression(body, head) {
                    return body.statements.length === 1 && ((ts.isReturnStatement(head) && !!head.expression));
                }
                function isFunctionReferencedInFile(sourceFile, typeChecker, node) {
                    return !!node.name && ts.FindAllReferences.Core.isSymbolReferencedInFile(node.name, typeChecker, sourceFile);
                }
            })(convertArrowFunctionOrFunctionExpression = refactor.convertArrowFunctionOrFunctionExpression || (refactor.convertArrowFunctionOrFunctionExpression = {}));
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var refactor;
        (function (refactor) {
            var inferFunctionReturnType;
            (function (inferFunctionReturnType) {
                var refactorName = "Infer function return type";
                var refactorDescription = ts.Diagnostics.Infer_function_return_type.message;
                var inferReturnTypeAction = {
                    name: refactorName,
                    description: refactorDescription,
                    kind: "refactor.rewrite.function.returnType"
                };
                refactor.registerRefactor(refactorName, {
                    kinds: [inferReturnTypeAction.kind],
                    getEditsForAction: getEditsForAction,
                    getAvailableActions: getAvailableActions
                });
                function getEditsForAction(context) {
                    var info = getInfo(context);
                    if (info && !refactor.isRefactorErrorInfo(info)) {
                        var edits = ts.textChanges.ChangeTracker.with(context, function (t) { return doChange(context.file, t, info.declaration, info.returnTypeNode); });
                        return { renameFilename: undefined, renameLocation: undefined, edits: edits };
                    }
                    return undefined;
                }
                function getAvailableActions(context) {
                    var info = getInfo(context);
                    if (!info)
                        return ts.emptyArray;
                    if (!refactor.isRefactorErrorInfo(info)) {
                        return [{
                                name: refactorName,
                                description: refactorDescription,
                                actions: [inferReturnTypeAction]
                            }];
                    }
                    if (context.preferences.provideRefactorNotApplicableReason) {
                        return [{
                                name: refactorName,
                                description: refactorDescription,
                                actions: [__assign(__assign({}, inferReturnTypeAction), { notApplicableReason: info.error })]
                            }];
                    }
                    return ts.emptyArray;
                }
                function doChange(sourceFile, changes, declaration, typeNode) {
                    var closeParen = ts.findChildOfKind(declaration, 21 /* CloseParenToken */, sourceFile);
                    var needParens = ts.isArrowFunction(declaration) && closeParen === undefined;
                    var endNode = needParens ? ts.first(declaration.parameters) : closeParen;
                    if (endNode) {
                        if (needParens) {
                            changes.insertNodeBefore(sourceFile, endNode, ts.factory.createToken(20 /* OpenParenToken */));
                            changes.insertNodeAfter(sourceFile, endNode, ts.factory.createToken(21 /* CloseParenToken */));
                        }
                        changes.insertNodeAt(sourceFile, endNode.end, typeNode, { prefix: ": " });
                    }
                }
                function getInfo(context) {
                    if (ts.isInJSFile(context.file) || !refactor.refactorKindBeginsWith(inferReturnTypeAction.kind, context.kind))
                        return;
                    var token = ts.getTokenAtPosition(context.file, context.startPosition);
                    var declaration = ts.findAncestor(token, function (n) {
                        return ts.isBlock(n) || n.parent && ts.isArrowFunction(n.parent) && (n.kind === 38 /* EqualsGreaterThanToken */ || n.parent.body === n) ? "quit" :
                            isConvertibleDeclaration(n);
                    });
                    if (!declaration || !declaration.body || declaration.type) {
                        return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Return_type_must_be_inferred_from_a_function) };
                    }
                    var typeChecker = context.program.getTypeChecker();
                    var returnType = tryGetReturnType(typeChecker, declaration);
                    if (!returnType) {
                        return { error: ts.getLocaleSpecificMessage(ts.Diagnostics.Could_not_determine_function_return_type) };
                    }
                    var returnTypeNode = typeChecker.typeToTypeNode(returnType, declaration, 1 /* NoTruncation */);
                    if (returnTypeNode) {
                        return { declaration: declaration, returnTypeNode: returnTypeNode };
                    }
                }
                function isConvertibleDeclaration(node) {
                    switch (node.kind) {
                        case 254 /* FunctionDeclaration */:
                        case 211 /* FunctionExpression */:
                        case 212 /* ArrowFunction */:
                        case 167 /* MethodDeclaration */:
                            return true;
                        default:
                            return false;
                    }
                }
                function tryGetReturnType(typeChecker, node) {
                    if (typeChecker.isImplementationOfOverload(node)) {
                        var signatures = typeChecker.getTypeAtLocation(node).getCallSignatures();
                        if (signatures.length > 1) {
                            return typeChecker.getUnionType(ts.mapDefined(signatures, function (s) { return s.getReturnType(); }));
                        }
                    }
                    var signature = typeChecker.getSignatureFromDeclaration(node);
                    if (signature) {
                        return typeChecker.getReturnTypeOfSignature(signature);
                    }
                }
            })(inferFunctionReturnType = refactor.inferFunctionReturnType || (refactor.inferFunctionReturnType = {}));
        })(refactor = ts.refactor || (ts.refactor = {}));
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        /** The version of the language service API */
        ts.servicesVersion = "0.8";
        function createNode(kind, pos, end, parent) {
            var node = ts.isNodeKind(kind) ? new NodeObject(kind, pos, end) :
                kind === 79 /* Identifier */ ? new IdentifierObject(79 /* Identifier */, pos, end) :
                    kind === 80 /* PrivateIdentifier */ ? new PrivateIdentifierObject(80 /* PrivateIdentifier */, pos, end) :
                        new TokenObject(kind, pos, end);
            node.parent = parent;
            node.flags = parent.flags & 25358336 /* ContextFlags */;
            return node;
        }
        var NodeObject = /** @class */ (function () {
            function NodeObject(kind, pos, end) {
                this.pos = pos;
                this.end = end;
                this.flags = 0 /* None */;
                this.modifierFlagsCache = 0 /* None */;
                this.transformFlags = 0 /* None */;
                this.parent = undefined;
                this.kind = kind;
            }
            NodeObject.prototype.assertHasRealPosition = function (message) {
                // eslint-disable-next-line debug-assert
                ts.Debug.assert(!ts.positionIsSynthesized(this.pos) && !ts.positionIsSynthesized(this.end), message || "Node must have a real position for this operation");
            };
            NodeObject.prototype.getSourceFile = function () {
                return ts.getSourceFileOfNode(this);
            };
            NodeObject.prototype.getStart = function (sourceFile, includeJsDocComment) {
                this.assertHasRealPosition();
                return ts.getTokenPosOfNode(this, sourceFile, includeJsDocComment);
            };
            NodeObject.prototype.getFullStart = function () {
                this.assertHasRealPosition();
                return this.pos;
            };
            NodeObject.prototype.getEnd = function () {
                this.assertHasRealPosition();
                return this.end;
            };
            NodeObject.prototype.getWidth = function (sourceFile) {
                this.assertHasRealPosition();
                return this.getEnd() - this.getStart(sourceFile);
            };
            NodeObject.prototype.getFullWidth = function () {
                this.assertHasRealPosition();
                return this.end - this.pos;
            };
            NodeObject.prototype.getLeadingTriviaWidth = function (sourceFile) {
                this.assertHasRealPosition();
                return this.getStart(sourceFile) - this.pos;
            };
            NodeObject.prototype.getFullText = function (sourceFile) {
                this.assertHasRealPosition();
                return (sourceFile || this.getSourceFile()).text.substring(this.pos, this.end);
            };
            NodeObject.prototype.getText = function (sourceFile) {
                this.assertHasRealPosition();
                if (!sourceFile) {
                    sourceFile = this.getSourceFile();
                }
                return sourceFile.text.substring(this.getStart(sourceFile), this.getEnd());
            };
            NodeObject.prototype.getChildCount = function (sourceFile) {
                return this.getChildren(sourceFile).length;
            };
            NodeObject.prototype.getChildAt = function (index, sourceFile) {
                return this.getChildren(sourceFile)[index];
            };
            NodeObject.prototype.getChildren = function (sourceFile) {
                this.assertHasRealPosition("Node without a real position cannot be scanned and thus has no token nodes - use forEachChild and collect the result if that's fine");
                return this._children || (this._children = createChildren(this, sourceFile));
            };
            NodeObject.prototype.getFirstToken = function (sourceFile) {
                this.assertHasRealPosition();
                var children = this.getChildren(sourceFile);
                if (!children.length) {
                    return undefined;
                }
                var child = ts.find(children, function (kid) { return kid.kind < 304 /* FirstJSDocNode */ || kid.kind > 342 /* LastJSDocNode */; });
                return child.kind < 159 /* FirstNode */ ?
                    child :
                    child.getFirstToken(sourceFile);
            };
            NodeObject.prototype.getLastToken = function (sourceFile) {
                this.assertHasRealPosition();
                var children = this.getChildren(sourceFile);
                var child = ts.lastOrUndefined(children);
                if (!child) {
                    return undefined;
                }
                return child.kind < 159 /* FirstNode */ ? child : child.getLastToken(sourceFile);
            };
            NodeObject.prototype.forEachChild = function (cbNode, cbNodeArray) {
                return ts.forEachChild(this, cbNode, cbNodeArray);
            };
            return NodeObject;
        }());
        function createChildren(node, sourceFile) {
            if (!ts.isNodeKind(node.kind)) {
                return ts.emptyArray;
            }
            var children = [];
            if (ts.isJSDocCommentContainingNode(node)) {
                /** Don't add trivia for "tokens" since this is in a comment. */
                node.forEachChild(function (child) { children.push(child); });
                return children;
            }
            ts.scanner.setText((sourceFile || node.getSourceFile()).text);
            var pos = node.pos;
            var processNode = function (child) {
                addSyntheticNodes(children, pos, child.pos, node);
                children.push(child);
                pos = child.end;
            };
            var processNodes = function (nodes) {
                addSyntheticNodes(children, pos, nodes.pos, node);
                children.push(createSyntaxList(nodes, node));
                pos = nodes.end;
            };
            // jsDocComments need to be the first children
            ts.forEach(node.jsDoc, processNode);
            // For syntactic classifications, all trivia are classified together, including jsdoc comments.
            // For that to work, the jsdoc comments should still be the leading trivia of the first child.
            // Restoring the scanner position ensures that.
            pos = node.pos;
            node.forEachChild(processNode, processNodes);
            addSyntheticNodes(children, pos, node.end, node);
            ts.scanner.setText(undefined);
            return children;
        }
        function addSyntheticNodes(nodes, pos, end, parent) {
            ts.scanner.setTextPos(pos);
            while (pos < end) {
                var token = ts.scanner.scan();
                var textPos = ts.scanner.getTextPos();
                if (textPos <= end) {
                    if (token === 79 /* Identifier */) {
                        ts.Debug.fail("Did not expect " + ts.Debug.formatSyntaxKind(parent.kind) + " to have an Identifier in its trivia");
                    }
                    nodes.push(createNode(token, pos, textPos, parent));
                }
                pos = textPos;
                if (token === 1 /* EndOfFileToken */) {
                    break;
                }
            }
        }
        function createSyntaxList(nodes, parent) {
            var list = createNode(343 /* SyntaxList */, nodes.pos, nodes.end, parent);
            list._children = [];
            var pos = nodes.pos;
            for (var _i = 0, nodes_2 = nodes; _i < nodes_2.length; _i++) {
                var node = nodes_2[_i];
                addSyntheticNodes(list._children, pos, node.pos, parent);
                list._children.push(node);
                pos = node.end;
            }
            addSyntheticNodes(list._children, pos, nodes.end, parent);
            return list;
        }
        var TokenOrIdentifierObject = /** @class */ (function () {
            function TokenOrIdentifierObject(pos, end) {
                // Set properties in same order as NodeObject
                this.pos = pos;
                this.end = end;
                this.flags = 0 /* None */;
                this.modifierFlagsCache = 0 /* None */;
                this.transformFlags = 0 /* None */;
                this.parent = undefined;
            }
            TokenOrIdentifierObject.prototype.getSourceFile = function () {
                return ts.getSourceFileOfNode(this);
            };
            TokenOrIdentifierObject.prototype.getStart = function (sourceFile, includeJsDocComment) {
                return ts.getTokenPosOfNode(this, sourceFile, includeJsDocComment);
            };
            TokenOrIdentifierObject.prototype.getFullStart = function () {
                return this.pos;
            };
            TokenOrIdentifierObject.prototype.getEnd = function () {
                return this.end;
            };
            TokenOrIdentifierObject.prototype.getWidth = function (sourceFile) {
                return this.getEnd() - this.getStart(sourceFile);
            };
            TokenOrIdentifierObject.prototype.getFullWidth = function () {
                return this.end - this.pos;
            };
            TokenOrIdentifierObject.prototype.getLeadingTriviaWidth = function (sourceFile) {
                return this.getStart(sourceFile) - this.pos;
            };
            TokenOrIdentifierObject.prototype.getFullText = function (sourceFile) {
                return (sourceFile || this.getSourceFile()).text.substring(this.pos, this.end);
            };
            TokenOrIdentifierObject.prototype.getText = function (sourceFile) {
                if (!sourceFile) {
                    sourceFile = this.getSourceFile();
                }
                return sourceFile.text.substring(this.getStart(sourceFile), this.getEnd());
            };
            TokenOrIdentifierObject.prototype.getChildCount = function () {
                return this.getChildren().length;
            };
            TokenOrIdentifierObject.prototype.getChildAt = function (index) {
                return this.getChildren()[index];
            };
            TokenOrIdentifierObject.prototype.getChildren = function () {
                return this.kind === 1 /* EndOfFileToken */ ? this.jsDoc || ts.emptyArray : ts.emptyArray;
            };
            TokenOrIdentifierObject.prototype.getFirstToken = function () {
                return undefined;
            };
            TokenOrIdentifierObject.prototype.getLastToken = function () {
                return undefined;
            };
            TokenOrIdentifierObject.prototype.forEachChild = function () {
                return undefined;
            };
            return TokenOrIdentifierObject;
        }());
        var SymbolObject = /** @class */ (function () {
            function SymbolObject(flags, name) {
                this.flags = flags;
                this.escapedName = name;
            }
            SymbolObject.prototype.getFlags = function () {
                return this.flags;
            };
            Object.defineProperty(SymbolObject.prototype, "name", {
                get: function () {
                    return ts.symbolName(this);
                },
                enumerable: false,
                configurable: true
            });
            SymbolObject.prototype.getEscapedName = function () {
                return this.escapedName;
            };
            SymbolObject.prototype.getName = function () {
                return this.name;
            };
            SymbolObject.prototype.getDeclarations = function () {
                return this.declarations;
            };
            SymbolObject.prototype.getDocumentationComment = function (checker) {
                if (!this.documentationComment) {
                    this.documentationComment = ts.emptyArray; // Set temporarily to avoid an infinite loop finding inherited docs
                    if (!this.declarations && this.target && this.target.tupleLabelDeclaration) {
                        var labelDecl = this.target.tupleLabelDeclaration;
                        this.documentationComment = getDocumentationComment([labelDecl], checker);
                    }
                    else {
                        this.documentationComment = getDocumentationComment(this.declarations, checker);
                    }
                }
                return this.documentationComment;
            };
            SymbolObject.prototype.getContextualDocumentationComment = function (context, checker) {
                switch (context === null || context === void 0 ? void 0 : context.kind) {
                    case 170 /* GetAccessor */:
                        if (!this.contextualGetAccessorDocumentationComment) {
                            this.contextualGetAccessorDocumentationComment = ts.emptyArray;
                            this.contextualGetAccessorDocumentationComment = getDocumentationComment(ts.filter(this.declarations, ts.isGetAccessor), checker);
                        }
                        return this.contextualGetAccessorDocumentationComment;
                    case 171 /* SetAccessor */:
                        if (!this.contextualSetAccessorDocumentationComment) {
                            this.contextualSetAccessorDocumentationComment = ts.emptyArray;
                            this.contextualSetAccessorDocumentationComment = getDocumentationComment(ts.filter(this.declarations, ts.isSetAccessor), checker);
                        }
                        return this.contextualSetAccessorDocumentationComment;
                    default:
                        return this.getDocumentationComment(checker);
                }
            };
            SymbolObject.prototype.getJsDocTags = function (checker) {
                if (this.tags === undefined) {
                    this.tags = ts.JsDoc.getJsDocTagsFromDeclarations(this.declarations, checker);
                }
                return this.tags;
            };
            return SymbolObject;
        }());
        var TokenObject = /** @class */ (function (_super) {
            __extends(TokenObject, _super);
            function TokenObject(kind, pos, end) {
                var _this = _super.call(this, pos, end) || this;
                _this.kind = kind;
                return _this;
            }
            return TokenObject;
        }(TokenOrIdentifierObject));
        var IdentifierObject = /** @class */ (function (_super) {
            __extends(IdentifierObject, _super);
            function IdentifierObject(_kind, pos, end) {
                var _this = _super.call(this, pos, end) || this;
                _this.kind = 79 /* Identifier */;
                return _this;
            }
            Object.defineProperty(IdentifierObject.prototype, "text", {
                get: function () {
                    return ts.idText(this);
                },
                enumerable: false,
                configurable: true
            });
            return IdentifierObject;
        }(TokenOrIdentifierObject));
        IdentifierObject.prototype.kind = 79 /* Identifier */;
        var PrivateIdentifierObject = /** @class */ (function (_super) {
            __extends(PrivateIdentifierObject, _super);
            function PrivateIdentifierObject(_kind, pos, end) {
                return _super.call(this, pos, end) || this;
            }
            Object.defineProperty(PrivateIdentifierObject.prototype, "text", {
                get: function () {
                    return ts.idText(this);
                },
                enumerable: false,
                configurable: true
            });
            return PrivateIdentifierObject;
        }(TokenOrIdentifierObject));
        PrivateIdentifierObject.prototype.kind = 80 /* PrivateIdentifier */;
        var TypeObject = /** @class */ (function () {
            function TypeObject(checker, flags) {
                this.checker = checker;
                this.flags = flags;
            }
            TypeObject.prototype.getFlags = function () {
                return this.flags;
            };
            TypeObject.prototype.getSymbol = function () {
                return this.symbol;
            };
            TypeObject.prototype.getProperties = function () {
                return this.checker.getPropertiesOfType(this);
            };
            TypeObject.prototype.getProperty = function (propertyName) {
                return this.checker.getPropertyOfType(this, propertyName);
            };
            TypeObject.prototype.getApparentProperties = function () {
                return this.checker.getAugmentedPropertiesOfType(this);
            };
            TypeObject.prototype.getCallSignatures = function () {
                return this.checker.getSignaturesOfType(this, 0 /* Call */);
            };
            TypeObject.prototype.getConstructSignatures = function () {
                return this.checker.getSignaturesOfType(this, 1 /* Construct */);
            };
            TypeObject.prototype.getStringIndexType = function () {
                return this.checker.getIndexTypeOfType(this, 0 /* String */);
            };
            TypeObject.prototype.getNumberIndexType = function () {
                return this.checker.getIndexTypeOfType(this, 1 /* Number */);
            };
            TypeObject.prototype.getBaseTypes = function () {
                return this.isClassOrInterface() ? this.checker.getBaseTypes(this) : undefined;
            };
            TypeObject.prototype.isNullableType = function () {
                return this.checker.isNullableType(this);
            };
            TypeObject.prototype.getNonNullableType = function () {
                return this.checker.getNonNullableType(this);
            };
            TypeObject.prototype.getNonOptionalType = function () {
                return this.checker.getNonOptionalType(this);
            };
            TypeObject.prototype.getConstraint = function () {
                return this.checker.getBaseConstraintOfType(this);
            };
            TypeObject.prototype.getDefault = function () {
                return this.checker.getDefaultFromTypeParameter(this);
            };
            TypeObject.prototype.isUnion = function () {
                return !!(this.flags & 1048576 /* Union */);
            };
            TypeObject.prototype.isIntersection = function () {
                return !!(this.flags & 2097152 /* Intersection */);
            };
            TypeObject.prototype.isUnionOrIntersection = function () {
                return !!(this.flags & 3145728 /* UnionOrIntersection */);
            };
            TypeObject.prototype.isLiteral = function () {
                return !!(this.flags & 384 /* StringOrNumberLiteral */);
            };
            TypeObject.prototype.isStringLiteral = function () {
                return !!(this.flags & 128 /* StringLiteral */);
            };
            TypeObject.prototype.isNumberLiteral = function () {
                return !!(this.flags & 256 /* NumberLiteral */);
            };
            TypeObject.prototype.isTypeParameter = function () {
                return !!(this.flags & 262144 /* TypeParameter */);
            };
            TypeObject.prototype.isClassOrInterface = function () {
                return !!(ts.getObjectFlags(this) & 3 /* ClassOrInterface */);
            };
            TypeObject.prototype.isClass = function () {
                return !!(ts.getObjectFlags(this) & 1 /* Class */);
            };
            Object.defineProperty(TypeObject.prototype, "typeArguments", {
                /**
                 * This polyfills `referenceType.typeArguments` for API consumers
                 */
                get: function () {
                    if (ts.getObjectFlags(this) & 4 /* Reference */) {
                        return this.checker.getTypeArguments(this);
                    }
                    return undefined;
                },
                enumerable: false,
                configurable: true
            });
            return TypeObject;
        }());
        var SignatureObject = /** @class */ (function () {
            function SignatureObject(checker, flags) {
                this.checker = checker;
                this.flags = flags;
            }
            SignatureObject.prototype.getDeclaration = function () {
                return this.declaration;
            };
            SignatureObject.prototype.getTypeParameters = function () {
                return this.typeParameters;
            };
            SignatureObject.prototype.getParameters = function () {
                return this.parameters;
            };
            SignatureObject.prototype.getReturnType = function () {
                return this.checker.getReturnTypeOfSignature(this);
            };
            SignatureObject.prototype.getDocumentationComment = function () {
                return this.documentationComment || (this.documentationComment = getDocumentationComment(ts.singleElementArray(this.declaration), this.checker));
            };
            SignatureObject.prototype.getJsDocTags = function () {
                if (this.jsDocTags === undefined) {
                    this.jsDocTags = this.declaration ? getJsDocTagsOfSignature(this.declaration, this.checker) : [];
                }
                return this.jsDocTags;
            };
            return SignatureObject;
        }());
        /**
         * Returns whether or not the given node has a JSDoc "inheritDoc" tag on it.
         * @param node the Node in question.
         * @returns `true` if `node` has a JSDoc "inheritDoc" tag on it, otherwise `false`.
         */
        function hasJSDocInheritDocTag(node) {
            return ts.getJSDocTags(node).some(function (tag) { return tag.tagName.text === "inheritDoc"; });
        }
        function getJsDocTagsOfSignature(declaration, checker) {
            var tags = ts.JsDoc.getJsDocTagsFromDeclarations([declaration], checker);
            if (tags.length === 0 || hasJSDocInheritDocTag(declaration)) {
                var inheritedTags = findBaseOfDeclaration(checker, declaration, function (symbol) { var _a; return ((_a = symbol.declarations) === null || _a === void 0 ? void 0 : _a.length) === 1 ? symbol.getJsDocTags() : undefined; });
                if (inheritedTags) {
                    tags = __spreadArray(__spreadArray([], inheritedTags, true), tags, true);
                }
            }
            return tags;
        }
        function getDocumentationComment(declarations, checker) {
            if (!declarations)
                return ts.emptyArray;
            var doc = ts.JsDoc.getJsDocCommentsFromDeclarations(declarations, checker);
            if (checker && (doc.length === 0 || declarations.some(hasJSDocInheritDocTag))) {
                var seenSymbols_1 = new ts.Set();
                for (var _i = 0, declarations_4 = declarations; _i < declarations_4.length; _i++) {
                    var declaration = declarations_4[_i];
                    var inheritedDocs = findBaseOfDeclaration(checker, declaration, function (symbol) {
                        if (!seenSymbols_1.has(symbol)) {
                            seenSymbols_1.add(symbol);
                            return symbol.getDocumentationComment(checker);
                        }
                    });
                    // TODO: GH#16312 Return a ReadonlyArray, avoid copying inheritedDocs
                    if (inheritedDocs)
                        doc = doc.length === 0 ? inheritedDocs.slice() : inheritedDocs.concat(ts.lineBreakPart(), doc);
                }
            }
            return doc;
        }
        function findBaseOfDeclaration(checker, declaration, cb) {
            var _a;
            var classOrInterfaceDeclaration = ((_a = declaration.parent) === null || _a === void 0 ? void 0 : _a.kind) === 169 /* Constructor */ ? declaration.parent.parent : declaration.parent;
            if (!classOrInterfaceDeclaration) {
                return;
            }
            return ts.firstDefined(ts.getAllSuperTypeNodes(classOrInterfaceDeclaration), function (superTypeNode) {
                var symbol = checker.getPropertyOfType(checker.getTypeAtLocation(superTypeNode), declaration.symbol.name);
                return symbol ? cb(symbol) : undefined;
            });
        }
        var SourceFileObject = /** @class */ (function (_super) {
            __extends(SourceFileObject, _super);
            function SourceFileObject(kind, pos, end) {
                var _this = _super.call(this, kind, pos, end) || this;
                _this.kind = 300 /* SourceFile */;
                return _this;
            }
            SourceFileObject.prototype.update = function (newText, textChangeRange) {
                return ts.updateSourceFile(this, newText, textChangeRange);
            };
            SourceFileObject.prototype.getLineAndCharacterOfPosition = function (position) {
                return ts.getLineAndCharacterOfPosition(this, position);
            };
            SourceFileObject.prototype.getLineStarts = function () {
                return ts.getLineStarts(this);
            };
            SourceFileObject.prototype.getPositionOfLineAndCharacter = function (line, character, allowEdits) {
                return ts.computePositionOfLineAndCharacter(ts.getLineStarts(this), line, character, this.text, allowEdits);
            };
            SourceFileObject.prototype.getLineEndOfPosition = function (pos) {
                var line = this.getLineAndCharacterOfPosition(pos).line;
                var lineStarts = this.getLineStarts();
                var lastCharPos;
                if (line + 1 >= lineStarts.length) {
                    lastCharPos = this.getEnd();
                }
                if (!lastCharPos) {
                    lastCharPos = lineStarts[line + 1] - 1;
                }
                var fullText = this.getFullText();
                // if the new line is "\r\n", we should return the last non-new-line-character position
                return fullText[lastCharPos] === "\n" && fullText[lastCharPos - 1] === "\r" ? lastCharPos - 1 : lastCharPos;
            };
            SourceFileObject.prototype.getNamedDeclarations = function () {
                if (!this.namedDeclarations) {
                    this.namedDeclarations = this.computeNamedDeclarations();
                }
                return this.namedDeclarations;
            };
            SourceFileObject.prototype.computeNamedDeclarations = function () {
                var result = ts.createMultiMap();
                this.forEachChild(visit);
                return result;
                function addDeclaration(declaration) {
                    var name = getDeclarationName(declaration);
                    if (name) {
                        result.add(name, declaration);
                    }
                }
                function getDeclarations(name) {
                    var declarations = result.get(name);
                    if (!declarations) {
                        result.set(name, declarations = []);
                    }
                    return declarations;
                }
                function getDeclarationName(declaration) {
                    var name = ts.getNonAssignedNameOfDeclaration(declaration);
                    return name && (ts.isComputedPropertyName(name) && ts.isPropertyAccessExpression(name.expression) ? name.expression.name.text
                        : ts.isPropertyName(name) ? ts.getNameFromPropertyName(name) : undefined);
                }
                function visit(node) {
                    switch (node.kind) {
                        case 254 /* FunctionDeclaration */:
                        case 211 /* FunctionExpression */:
                        case 167 /* MethodDeclaration */:
                        case 166 /* MethodSignature */:
                            var functionDeclaration = node;
                            var declarationName = getDeclarationName(functionDeclaration);
                            if (declarationName) {
                                var declarations = getDeclarations(declarationName);
                                var lastDeclaration = ts.lastOrUndefined(declarations);
                                // Check whether this declaration belongs to an "overload group".
                                if (lastDeclaration && functionDeclaration.parent === lastDeclaration.parent && functionDeclaration.symbol === lastDeclaration.symbol) {
                                    // Overwrite the last declaration if it was an overload
                                    // and this one is an implementation.
                                    if (functionDeclaration.body && !lastDeclaration.body) {
                                        declarations[declarations.length - 1] = functionDeclaration;
                                    }
                                }
                                else {
                                    declarations.push(functionDeclaration);
                                }
                            }
                            ts.forEachChild(node, visit);
                            break;
                        case 255 /* ClassDeclaration */:
                        case 224 /* ClassExpression */:
                        case 256 /* InterfaceDeclaration */:
                        case 257 /* TypeAliasDeclaration */:
                        case 258 /* EnumDeclaration */:
                        case 259 /* ModuleDeclaration */:
                        case 263 /* ImportEqualsDeclaration */:
                        case 273 /* ExportSpecifier */:
                        case 268 /* ImportSpecifier */:
                        case 265 /* ImportClause */:
                        case 266 /* NamespaceImport */:
                        case 170 /* GetAccessor */:
                        case 171 /* SetAccessor */:
                        case 180 /* TypeLiteral */:
                            addDeclaration(node);
                            ts.forEachChild(node, visit);
                            break;
                        case 162 /* Parameter */:
                            // Only consider parameter properties
                            if (!ts.hasSyntacticModifier(node, 16476 /* ParameterPropertyModifier */)) {
                                break;
                            }
                        // falls through
                        case 252 /* VariableDeclaration */:
                        case 201 /* BindingElement */: {
                            var decl = node;
                            if (ts.isBindingPattern(decl.name)) {
                                ts.forEachChild(decl.name, visit);
                                break;
                            }
                            if (decl.initializer) {
                                visit(decl.initializer);
                            }
                        }
                        // falls through
                        case 294 /* EnumMember */:
                        case 165 /* PropertyDeclaration */:
                        case 164 /* PropertySignature */:
                            addDeclaration(node);
                            break;
                        case 270 /* ExportDeclaration */:
                            // Handle named exports case e.g.:
                            //    export {a, b as B} from "mod";
                            var exportDeclaration = node;
                            if (exportDeclaration.exportClause) {
                                if (ts.isNamedExports(exportDeclaration.exportClause)) {
                                    ts.forEach(exportDeclaration.exportClause.elements, visit);
                                }
                                else {
                                    visit(exportDeclaration.exportClause.name);
                                }
                            }
                            break;
                        case 264 /* ImportDeclaration */:
                            var importClause = node.importClause;
                            if (importClause) {
                                // Handle default import case e.g.:
                                //    import d from "mod";
                                if (importClause.name) {
                                    addDeclaration(importClause.name);
                                }
                                // Handle named bindings in imports e.g.:
                                //    import * as NS from "mod";
                                //    import {a, b as B} from "mod";
                                if (importClause.namedBindings) {
                                    if (importClause.namedBindings.kind === 266 /* NamespaceImport */) {
                                        addDeclaration(importClause.namedBindings);
                                    }
                                    else {
                                        ts.forEach(importClause.namedBindings.elements, visit);
                                    }
                                }
                            }
                            break;
                        case 219 /* BinaryExpression */:
                            if (ts.getAssignmentDeclarationKind(node) !== 0 /* None */) {
                                addDeclaration(node);
                            }
                        // falls through
                        default:
                            ts.forEachChild(node, visit);
                    }
                }
            };
            return SourceFileObject;
        }(NodeObject));
        var SourceMapSourceObject = /** @class */ (function () {
            function SourceMapSourceObject(fileName, text, skipTrivia) {
                this.fileName = fileName;
                this.text = text;
                this.skipTrivia = skipTrivia;
            }
            SourceMapSourceObject.prototype.getLineAndCharacterOfPosition = function (pos) {
                return ts.getLineAndCharacterOfPosition(this, pos);
            };
            return SourceMapSourceObject;
        }());
        function getServicesObjectAllocator() {
            return {
                getNodeConstructor: function () { return NodeObject; },
                getTokenConstructor: function () { return TokenObject; },
                getIdentifierConstructor: function () { return IdentifierObject; },
                getPrivateIdentifierConstructor: function () { return PrivateIdentifierObject; },
                getSourceFileConstructor: function () { return SourceFileObject; },
                getSymbolConstructor: function () { return SymbolObject; },
                getTypeConstructor: function () { return TypeObject; },
                getSignatureConstructor: function () { return SignatureObject; },
                getSourceMapSourceConstructor: function () { return SourceMapSourceObject; },
            };
        }
        function toEditorSettings(optionsAsMap) {
            var allPropertiesAreCamelCased = true;
            for (var key in optionsAsMap) {
                if (ts.hasProperty(optionsAsMap, key) && !isCamelCase(key)) {
                    allPropertiesAreCamelCased = false;
                    break;
                }
            }
            if (allPropertiesAreCamelCased) {
                return optionsAsMap;
            }
            var settings = {};
            for (var key in optionsAsMap) {
                if (ts.hasProperty(optionsAsMap, key)) {
                    var newKey = isCamelCase(key) ? key : key.charAt(0).toLowerCase() + key.substr(1);
                    settings[newKey] = optionsAsMap[key];
                }
            }
            return settings;
        }
        ts.toEditorSettings = toEditorSettings;
        function isCamelCase(s) {
            return !s.length || s.charAt(0) === s.charAt(0).toLowerCase();
        }
        function displayPartsToString(displayParts) {
            if (displayParts) {
                return ts.map(displayParts, function (displayPart) { return displayPart.text; }).join("");
            }
            return "";
        }
        ts.displayPartsToString = displayPartsToString;
        function getDefaultCompilerOptions() {
            // Always default to "ScriptTarget.ES5" for the language service
            return {
                target: 1 /* ES5 */,
                jsx: 1 /* Preserve */
            };
        }
        ts.getDefaultCompilerOptions = getDefaultCompilerOptions;
        function getSupportedCodeFixes() {
            return ts.codefix.getSupportedErrorCodes();
        }
        ts.getSupportedCodeFixes = getSupportedCodeFixes;
        // Cache host information about script Should be refreshed
        // at each language service public entry point, since we don't know when
        // the set of scripts handled by the host changes.
        var HostCache = /** @class */ (function () {
            function HostCache(host, getCanonicalFileName) {
                this.host = host;
                // script id => script index
                this.currentDirectory = host.getCurrentDirectory();
                this.fileNameToEntry = new ts.Map();
                // Initialize the list with the root file names
                var rootFileNames = host.getScriptFileNames();
                for (var _i = 0, rootFileNames_1 = rootFileNames; _i < rootFileNames_1.length; _i++) {
                    var fileName = rootFileNames_1[_i];
                    this.createEntry(fileName, ts.toPath(fileName, this.currentDirectory, getCanonicalFileName));
                }
            }
            HostCache.prototype.createEntry = function (fileName, path) {
                var entry;
                var scriptSnapshot = this.host.getScriptSnapshot(fileName);
                if (scriptSnapshot) {
                    entry = {
                        hostFileName: fileName,
                        version: this.host.getScriptVersion(fileName),
                        scriptSnapshot: scriptSnapshot,
                        scriptKind: ts.getScriptKind(fileName, this.host)
                    };
                }
                else {
                    entry = fileName;
                }
                this.fileNameToEntry.set(path, entry);
                return entry;
            };
            HostCache.prototype.getEntryByPath = function (path) {
                return this.fileNameToEntry.get(path);
            };
            HostCache.prototype.getHostFileInformation = function (path) {
                var entry = this.fileNameToEntry.get(path);
                return !ts.isString(entry) ? entry : undefined;
            };
            HostCache.prototype.getOrCreateEntryByPath = function (fileName, path) {
                var info = this.getEntryByPath(path) || this.createEntry(fileName, path);
                return ts.isString(info) ? undefined : info; // TODO: GH#18217
            };
            HostCache.prototype.getRootFileNames = function () {
                var names = [];
                this.fileNameToEntry.forEach(function (entry) {
                    if (ts.isString(entry)) {
                        names.push(entry);
                    }
                    else {
                        names.push(entry.hostFileName);
                    }
                });
                return names;
            };
            HostCache.prototype.getScriptSnapshot = function (path) {
                var file = this.getHostFileInformation(path);
                return (file && file.scriptSnapshot); // TODO: GH#18217
            };
            return HostCache;
        }());
        var SyntaxTreeCache = /** @class */ (function () {
            function SyntaxTreeCache(host) {
                this.host = host;
            }
            SyntaxTreeCache.prototype.getCurrentSourceFile = function (fileName) {
                var scriptSnapshot = this.host.getScriptSnapshot(fileName);
                if (!scriptSnapshot) {
                    // The host does not know about this file.
                    throw new Error("Could not find file: '" + fileName + "'.");
                }
                var scriptKind = ts.getScriptKind(fileName, this.host);
                var version = this.host.getScriptVersion(fileName);
                var sourceFile;
                if (this.currentFileName !== fileName) {
                    // This is a new file, just parse it
                    sourceFile = createLanguageServiceSourceFile(fileName, scriptSnapshot, 99 /* Latest */, version, /*setNodeParents*/ true, scriptKind);
                }
                else if (this.currentFileVersion !== version) {
                    // This is the same file, just a newer version. Incrementally parse the file.
                    var editRange = scriptSnapshot.getChangeRange(this.currentFileScriptSnapshot);
                    sourceFile = updateLanguageServiceSourceFile(this.currentSourceFile, scriptSnapshot, version, editRange);
                }
                if (sourceFile) {
                    // All done, ensure state is up to date
                    this.currentFileVersion = version;
                    this.currentFileName = fileName;
                    this.currentFileScriptSnapshot = scriptSnapshot;
                    this.currentSourceFile = sourceFile;
                }
                return this.currentSourceFile;
            };
            return SyntaxTreeCache;
        }());
        function setSourceFileFields(sourceFile, scriptSnapshot, version) {
            sourceFile.version = version;
            sourceFile.scriptSnapshot = scriptSnapshot;
        }
        function createLanguageServiceSourceFile(fileName, scriptSnapshot, scriptTarget, version, setNodeParents, scriptKind) {
            var sourceFile = ts.createSourceFile(fileName, ts.getSnapshotText(scriptSnapshot), scriptTarget, setNodeParents, scriptKind);
            setSourceFileFields(sourceFile, scriptSnapshot, version);
            return sourceFile;
        }
        ts.createLanguageServiceSourceFile = createLanguageServiceSourceFile;
        function updateLanguageServiceSourceFile(sourceFile, scriptSnapshot, version, textChangeRange, aggressiveChecks) {
            // If we were given a text change range, and our version or open-ness changed, then
            // incrementally parse this file.
            if (textChangeRange) {
                if (version !== sourceFile.version) {
                    var newText = void 0;
                    // grab the fragment from the beginning of the original text to the beginning of the span
                    var prefix = textChangeRange.span.start !== 0
                        ? sourceFile.text.substr(0, textChangeRange.span.start)
                        : "";
                    // grab the fragment from the end of the span till the end of the original text
                    var suffix = ts.textSpanEnd(textChangeRange.span) !== sourceFile.text.length
                        ? sourceFile.text.substr(ts.textSpanEnd(textChangeRange.span))
                        : "";
                    if (textChangeRange.newLength === 0) {
                        // edit was a deletion - just combine prefix and suffix
                        newText = prefix && suffix ? prefix + suffix : prefix || suffix;
                    }
                    else {
                        // it was actual edit, fetch the fragment of new text that correspond to new span
                        var changedText = scriptSnapshot.getText(textChangeRange.span.start, textChangeRange.span.start + textChangeRange.newLength);
                        // combine prefix, changed text and suffix
                        newText = prefix && suffix
                            ? prefix + changedText + suffix
                            : prefix
                                ? (prefix + changedText)
                                : (changedText + suffix);
                    }
                    var newSourceFile = ts.updateSourceFile(sourceFile, newText, textChangeRange, aggressiveChecks);
                    setSourceFileFields(newSourceFile, scriptSnapshot, version);
                    // after incremental parsing nameTable might not be up-to-date
                    // drop it so it can be lazily recreated later
                    newSourceFile.nameTable = undefined;
                    // dispose all resources held by old script snapshot
                    if (sourceFile !== newSourceFile && sourceFile.scriptSnapshot) {
                        if (sourceFile.scriptSnapshot.dispose) {
                            sourceFile.scriptSnapshot.dispose();
                        }
                        sourceFile.scriptSnapshot = undefined;
                    }
                    return newSourceFile;
                }
            }
            // Otherwise, just create a new source file.
            return createLanguageServiceSourceFile(sourceFile.fileName, scriptSnapshot, sourceFile.languageVersion, version, /*setNodeParents*/ true, sourceFile.scriptKind);
        }
        ts.updateLanguageServiceSourceFile = updateLanguageServiceSourceFile;
        var NoopCancellationToken = {
            isCancellationRequested: ts.returnFalse,
            throwIfCancellationRequested: ts.noop,
        };
        var CancellationTokenObject = /** @class */ (function () {
            function CancellationTokenObject(cancellationToken) {
                this.cancellationToken = cancellationToken;
            }
            CancellationTokenObject.prototype.isCancellationRequested = function () {
                return this.cancellationToken.isCancellationRequested();
            };
            CancellationTokenObject.prototype.throwIfCancellationRequested = function () {
                if (this.isCancellationRequested()) {
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.instant("session" /* Session */, "cancellationThrown", { kind: "CancellationTokenObject" });
                    throw new ts.OperationCanceledException();
                }
            };
            return CancellationTokenObject;
        }());
        /* @internal */
        /** A cancellation that throttles calls to the host */
        var ThrottledCancellationToken = /** @class */ (function () {
            function ThrottledCancellationToken(hostCancellationToken, throttleWaitMilliseconds) {
                if (throttleWaitMilliseconds === void 0) {
                    throttleWaitMilliseconds = 20;
                }
                this.hostCancellationToken = hostCancellationToken;
                this.throttleWaitMilliseconds = throttleWaitMilliseconds;
                // Store when we last tried to cancel.  Checking cancellation can be expensive (as we have
                // to marshall over to the host layer).  So we only bother actually checking once enough
                // time has passed.
                this.lastCancellationCheckTime = 0;
            }
            ThrottledCancellationToken.prototype.isCancellationRequested = function () {
                var time = ts.timestamp();
                var duration = Math.abs(time - this.lastCancellationCheckTime);
                if (duration >= this.throttleWaitMilliseconds) {
                    // Check no more than once every throttle wait milliseconds
                    this.lastCancellationCheckTime = time;
                    return this.hostCancellationToken.isCancellationRequested();
                }
                return false;
            };
            ThrottledCancellationToken.prototype.throwIfCancellationRequested = function () {
                if (this.isCancellationRequested()) {
                    ts.tracing === null || ts.tracing === void 0 ? void 0 : ts.tracing.instant("session" /* Session */, "cancellationThrown", { kind: "ThrottledCancellationToken" });
                    throw new ts.OperationCanceledException();
                }
            };
            return ThrottledCancellationToken;
        }());
        ts.ThrottledCancellationToken = ThrottledCancellationToken;
        var invalidOperationsInPartialSemanticMode = [
            "getSemanticDiagnostics",
            "getSuggestionDiagnostics",
            "getCompilerOptionsDiagnostics",
            "getSemanticClassifications",
            "getEncodedSemanticClassifications",
            "getCodeFixesAtPosition",
            "getCombinedCodeFix",
            "applyCodeActionCommand",
            "organizeImports",
            "getEditsForFileRename",
            "getEmitOutput",
            "getApplicableRefactors",
            "getEditsForRefactor",
            "prepareCallHierarchy",
            "provideCallHierarchyIncomingCalls",
            "provideCallHierarchyOutgoingCalls",
            "provideInlayHints"
        ];
        var invalidOperationsInSyntacticMode = __spreadArray(__spreadArray([], invalidOperationsInPartialSemanticMode, true), [
            "getCompletionsAtPosition",
            "getCompletionEntryDetails",
            "getCompletionEntrySymbol",
            "getSignatureHelpItems",
            "getQuickInfoAtPosition",
            "getDefinitionAtPosition",
            "getDefinitionAndBoundSpan",
            "getImplementationAtPosition",
            "getTypeDefinitionAtPosition",
            "getReferencesAtPosition",
            "findReferences",
            "getOccurrencesAtPosition",
            "getDocumentHighlights",
            "getNavigateToItems",
            "getRenameInfo",
            "findRenameLocations",
            "getApplicableRefactors",
        ], false);
        function createLanguageService(host, documentRegistry, syntaxOnlyOrLanguageServiceMode) {
            var _a;
            if (documentRegistry === void 0) {
                documentRegistry = ts.createDocumentRegistry(host.useCaseSensitiveFileNames && host.useCaseSensitiveFileNames(), host.getCurrentDirectory());
            }
            var languageServiceMode;
            if (syntaxOnlyOrLanguageServiceMode === undefined) {
                languageServiceMode = ts.LanguageServiceMode.Semantic;
            }
            else if (typeof syntaxOnlyOrLanguageServiceMode === "boolean") {
                // languageServiceMode = SyntaxOnly
                languageServiceMode = syntaxOnlyOrLanguageServiceMode ? ts.LanguageServiceMode.Syntactic : ts.LanguageServiceMode.Semantic;
            }
            else {
                languageServiceMode = syntaxOnlyOrLanguageServiceMode;
            }
            var syntaxTreeCache = new SyntaxTreeCache(host);
            var program;
            var lastProjectVersion;
            var lastTypesRootVersion = 0;
            var cancellationToken = host.getCancellationToken
                ? new CancellationTokenObject(host.getCancellationToken())
                : NoopCancellationToken;
            var currentDirectory = host.getCurrentDirectory();
            // Check if the localized messages json is set, otherwise query the host for it
            if (!ts.localizedDiagnosticMessages && host.getLocalizedDiagnosticMessages) {
                ts.setLocalizedDiagnosticMessages(host.getLocalizedDiagnosticMessages());
            }
            function log(message) {
                if (host.log) {
                    host.log(message);
                }
            }
            var useCaseSensitiveFileNames = ts.hostUsesCaseSensitiveFileNames(host);
            var getCanonicalFileName = ts.createGetCanonicalFileName(useCaseSensitiveFileNames);
            var sourceMapper = ts.getSourceMapper({
                useCaseSensitiveFileNames: function () { return useCaseSensitiveFileNames; },
                getCurrentDirectory: function () { return currentDirectory; },
                getProgram: getProgram,
                fileExists: ts.maybeBind(host, host.fileExists),
                readFile: ts.maybeBind(host, host.readFile),
                getDocumentPositionMapper: ts.maybeBind(host, host.getDocumentPositionMapper),
                getSourceFileLike: ts.maybeBind(host, host.getSourceFileLike),
                log: log
            });
            function getValidSourceFile(fileName) {
                var sourceFile = program.getSourceFile(fileName);
                if (!sourceFile) {
                    var error = new Error("Could not find source file: '" + fileName + "'.");
                    // We've been having trouble debugging this, so attach sidecar data for the tsserver log.
                    // See https://github.com/microsoft/TypeScript/issues/30180.
                    error.ProgramFiles = program.getSourceFiles().map(function (f) { return f.fileName; });
                    throw error;
                }
                return sourceFile;
            }
            function synchronizeHostData() {
                var _a, _b, _c;
                ts.Debug.assert(languageServiceMode !== ts.LanguageServiceMode.Syntactic);
                // perform fast check if host supports it
                if (host.getProjectVersion) {
                    var hostProjectVersion = host.getProjectVersion();
                    if (hostProjectVersion) {
                        if (lastProjectVersion === hostProjectVersion && !((_a = host.hasChangedAutomaticTypeDirectiveNames) === null || _a === void 0 ? void 0 : _a.call(host))) {
                            return;
                        }
                        lastProjectVersion = hostProjectVersion;
                    }
                }
                var typeRootsVersion = host.getTypeRootsVersion ? host.getTypeRootsVersion() : 0;
                if (lastTypesRootVersion !== typeRootsVersion) {
                    log("TypeRoots version has changed; provide new program");
                    program = undefined; // TODO: GH#18217
                    lastTypesRootVersion = typeRootsVersion;
                }
                // Get a fresh cache of the host information
                var hostCache = new HostCache(host, getCanonicalFileName);
                var rootFileNames = hostCache.getRootFileNames();
                var newSettings = host.getCompilationSettings() || getDefaultCompilerOptions();
                var hasInvalidatedResolution = host.hasInvalidatedResolution || ts.returnFalse;
                var hasChangedAutomaticTypeDirectiveNames = ts.maybeBind(host, host.hasChangedAutomaticTypeDirectiveNames);
                var projectReferences = (_b = host.getProjectReferences) === null || _b === void 0 ? void 0 : _b.call(host);
                var parsedCommandLines;
                var parseConfigHost = {
                    useCaseSensitiveFileNames: useCaseSensitiveFileNames,
                    fileExists: fileExists,
                    readFile: readFile,
                    readDirectory: readDirectory,
                    trace: ts.maybeBind(host, host.trace),
                    getCurrentDirectory: function () { return currentDirectory; },
                    onUnRecoverableConfigFileDiagnostic: ts.noop,
                };
                // If the program is already up-to-date, we can reuse it
                if (ts.isProgramUptoDate(program, rootFileNames, newSettings, function (_path, fileName) { return host.getScriptVersion(fileName); }, fileExists, hasInvalidatedResolution, hasChangedAutomaticTypeDirectiveNames, getParsedCommandLine, projectReferences)) {
                    return;
                }
                // IMPORTANT - It is critical from this moment onward that we do not check
                // cancellation tokens.  We are about to mutate source files from a previous program
                // instance.  If we cancel midway through, we may end up in an inconsistent state where
                // the program points to old source files that have been invalidated because of
                // incremental parsing.
                // Now create a new compiler
                var compilerHost = {
                    getSourceFile: getOrCreateSourceFile,
                    getSourceFileByPath: getOrCreateSourceFileByPath,
                    getCancellationToken: function () { return cancellationToken; },
                    getCanonicalFileName: getCanonicalFileName,
                    useCaseSensitiveFileNames: function () { return useCaseSensitiveFileNames; },
                    getNewLine: function () { return ts.getNewLineCharacter(newSettings, function () { return ts.getNewLineOrDefaultFromHost(host); }); },
                    getDefaultLibFileName: function (options) { return host.getDefaultLibFileName(options); },
                    writeFile: ts.noop,
                    getCurrentDirectory: function () { return currentDirectory; },
                    fileExists: fileExists,
                    readFile: readFile,
                    getSymlinkCache: ts.maybeBind(host, host.getSymlinkCache),
                    realpath: ts.maybeBind(host, host.realpath),
                    directoryExists: function (directoryName) {
                        return ts.directoryProbablyExists(directoryName, host);
                    },
                    getDirectories: function (path) {
                        return host.getDirectories ? host.getDirectories(path) : [];
                    },
                    readDirectory: readDirectory,
                    onReleaseOldSourceFile: onReleaseOldSourceFile,
                    onReleaseParsedCommandLine: onReleaseParsedCommandLine,
                    hasInvalidatedResolution: hasInvalidatedResolution,
                    hasChangedAutomaticTypeDirectiveNames: hasChangedAutomaticTypeDirectiveNames,
                    trace: parseConfigHost.trace,
                    resolveModuleNames: ts.maybeBind(host, host.resolveModuleNames),
                    resolveTypeReferenceDirectives: ts.maybeBind(host, host.resolveTypeReferenceDirectives),
                    useSourceOfProjectReferenceRedirect: ts.maybeBind(host, host.useSourceOfProjectReferenceRedirect),
                    getParsedCommandLine: getParsedCommandLine,
                };
                (_c = host.setCompilerHost) === null || _c === void 0 ? void 0 : _c.call(host, compilerHost);
                var documentRegistryBucketKey = documentRegistry.getKeyForCompilationSettings(newSettings);
                var options = {
                    rootNames: rootFileNames,
                    options: newSettings,
                    host: compilerHost,
                    oldProgram: program,
                    projectReferences: projectReferences
                };
                program = ts.createProgram(options);
                // hostCache is captured in the closure for 'getOrCreateSourceFile' but it should not be used past this point.
                // It needs to be cleared to allow all collected snapshots to be released
                hostCache = undefined;
                parsedCommandLines = undefined;
                // We reset this cache on structure invalidation so we don't hold on to outdated files for long; however we can't use the `compilerHost` above,
                // Because it only functions until `hostCache` is cleared, while we'll potentially need the functionality to lazily read sourcemap files during
                // the course of whatever called `synchronizeHostData`
                sourceMapper.clearCache();
                // Make sure all the nodes in the program are both bound, and have their parent
                // pointers set property.
                program.getTypeChecker();
                return;
                function getParsedCommandLine(fileName) {
                    var path = ts.toPath(fileName, currentDirectory, getCanonicalFileName);
                    var existing = parsedCommandLines === null || parsedCommandLines === void 0 ? void 0 : parsedCommandLines.get(path);
                    if (existing !== undefined)
                        return existing || undefined;
                    var result = host.getParsedCommandLine ?
                        host.getParsedCommandLine(fileName) :
                        getParsedCommandLineOfConfigFileUsingSourceFile(fileName);
                    (parsedCommandLines || (parsedCommandLines = new ts.Map())).set(path, result || false);
                    return result;
                }
                function getParsedCommandLineOfConfigFileUsingSourceFile(configFileName) {
                    var result = getOrCreateSourceFile(configFileName, 100 /* JSON */);
                    if (!result)
                        return undefined;
                    result.path = ts.toPath(configFileName, currentDirectory, getCanonicalFileName);
                    result.resolvedPath = result.path;
                    result.originalFileName = result.fileName;
                    return ts.parseJsonSourceFileConfigFileContent(result, parseConfigHost, ts.getNormalizedAbsolutePath(ts.getDirectoryPath(configFileName), currentDirectory), 
                    /*optionsToExtend*/ undefined, ts.getNormalizedAbsolutePath(configFileName, currentDirectory));
                }
                function onReleaseParsedCommandLine(configFileName, oldResolvedRef, oldOptions) {
                    var _a;
                    if (host.getParsedCommandLine) {
                        (_a = host.onReleaseParsedCommandLine) === null || _a === void 0 ? void 0 : _a.call(host, configFileName, oldResolvedRef, oldOptions);
                    }
                    else if (oldResolvedRef) {
                        onReleaseOldSourceFile(oldResolvedRef.sourceFile, oldOptions);
                    }
                }
                function fileExists(fileName) {
                    var path = ts.toPath(fileName, currentDirectory, getCanonicalFileName);
                    var entry = hostCache && hostCache.getEntryByPath(path);
                    return entry ?
                        !ts.isString(entry) :
                        (!!host.fileExists && host.fileExists(fileName));
                }
                function readFile(fileName) {
                    // stub missing host functionality
                    var path = ts.toPath(fileName, currentDirectory, getCanonicalFileName);
                    var entry = hostCache && hostCache.getEntryByPath(path);
                    if (entry) {
                        return ts.isString(entry) ? undefined : ts.getSnapshotText(entry.scriptSnapshot);
                    }
                    return host.readFile && host.readFile(fileName);
                }
                function readDirectory(path, extensions, exclude, include, depth) {
                    ts.Debug.checkDefined(host.readDirectory, "'LanguageServiceHost.readDirectory' must be implemented to correctly process 'projectReferences'");
                    return host.readDirectory(path, extensions, exclude, include, depth);
                }
                // Release any files we have acquired in the old program but are
                // not part of the new program.
                function onReleaseOldSourceFile(oldSourceFile, oldOptions) {
                    var oldSettingsKey = documentRegistry.getKeyForCompilationSettings(oldOptions);
                    documentRegistry.releaseDocumentWithKey(oldSourceFile.resolvedPath, oldSettingsKey, oldSourceFile.scriptKind);
                }
                function getOrCreateSourceFile(fileName, languageVersion, onError, shouldCreateNewSourceFile) {
                    return getOrCreateSourceFileByPath(fileName, ts.toPath(fileName, currentDirectory, getCanonicalFileName), languageVersion, onError, shouldCreateNewSourceFile);
                }
                function getOrCreateSourceFileByPath(fileName, path, _languageVersion, _onError, shouldCreateNewSourceFile) {
                    ts.Debug.assert(hostCache !== undefined, "getOrCreateSourceFileByPath called after typical CompilerHost lifetime, check the callstack something with a reference to an old host.");
                    // The program is asking for this file, check first if the host can locate it.
                    // If the host can not locate the file, then it does not exist. return undefined
                    // to the program to allow reporting of errors for missing files.
                    var hostFileInformation = hostCache && hostCache.getOrCreateEntryByPath(fileName, path);
                    if (!hostFileInformation) {
                        return undefined;
                    }
                    // Check if the language version has changed since we last created a program; if they are the same,
                    // it is safe to reuse the sourceFiles; if not, then the shape of the AST can change, and the oldSourceFile
                    // can not be reused. we have to dump all syntax trees and create new ones.
                    if (!shouldCreateNewSourceFile) {
                        // Check if the old program had this file already
                        var oldSourceFile = program && program.getSourceFileByPath(path);
                        if (oldSourceFile) {
                            // We already had a source file for this file name.  Go to the registry to
                            // ensure that we get the right up to date version of it.  We need this to
                            // address the following race-condition.  Specifically, say we have the following:
                            //
                            //      LS1
                            //          \
                            //           DocumentRegistry
                            //          /
                            //      LS2
                            //
                            // Each LS has a reference to file 'foo.ts' at version 1.  LS2 then updates
                            // it's version of 'foo.ts' to version 2.  This will cause LS2 and the
                            // DocumentRegistry to have version 2 of the document.  However, LS1 will
                            // have version 1.  And *importantly* this source file will be *corrupt*.
                            // The act of creating version 2 of the file irrevocably damages the version
                            // 1 file.
                            //
                            // So, later when we call into LS1, we need to make sure that it doesn't use
                            // it's source file any more, and instead defers to DocumentRegistry to get
                            // either version 1, version 2 (or some other version) depending on what the
                            // host says should be used.
                            // We do not support the scenario where a host can modify a registered
                            // file's script kind, i.e. in one project some file is treated as ".ts"
                            // and in another as ".js"
                            if (hostFileInformation.scriptKind === oldSourceFile.scriptKind) {
                                return documentRegistry.updateDocumentWithKey(fileName, path, newSettings, documentRegistryBucketKey, hostFileInformation.scriptSnapshot, hostFileInformation.version, hostFileInformation.scriptKind);
                            }
                            else {
                                // Release old source file and fall through to aquire new file with new script kind
                                documentRegistry.releaseDocumentWithKey(oldSourceFile.resolvedPath, documentRegistry.getKeyForCompilationSettings(program.getCompilerOptions()), oldSourceFile.scriptKind);
                            }
                        }
                        // We didn't already have the file.  Fall through and acquire it from the registry.
                    }
                    // Could not find this file in the old program, create a new SourceFile for it.
                    return documentRegistry.acquireDocumentWithKey(fileName, path, newSettings, documentRegistryBucketKey, hostFileInformation.scriptSnapshot, hostFileInformation.version, hostFileInformation.scriptKind);
                }
            }
            // TODO: GH#18217 frequently asserted as defined
            function getProgram() {
                if (languageServiceMode === ts.LanguageServiceMode.Syntactic) {
                    ts.Debug.assert(program === undefined);
                    return undefined;
                }
                synchronizeHostData();
                return program;
            }
            function getAutoImportProvider() {
                var _a;
                return (_a = host.getPackageJsonAutoImportProvider) === null || _a === void 0 ? void 0 : _a.call(host);
            }
            function cleanupSemanticCache() {
                program = undefined; // TODO: GH#18217
            }
            function dispose() {
                if (program) {
                    // Use paths to ensure we are using correct key and paths as document registry could be created with different current directory than host
                    var key_1 = documentRegistry.getKeyForCompilationSettings(program.getCompilerOptions());
                    ts.forEach(program.getSourceFiles(), function (f) {
                        return documentRegistry.releaseDocumentWithKey(f.resolvedPath, key_1, f.scriptKind);
                    });
                    program = undefined; // TODO: GH#18217
                }
                host = undefined;
            }
            /// Diagnostics
            function getSyntacticDiagnostics(fileName) {
                synchronizeHostData();
                return program.getSyntacticDiagnostics(getValidSourceFile(fileName), cancellationToken).slice();
            }
            /**
             * getSemanticDiagnostics return array of Diagnostics. If '-d' is not enabled, only report semantic errors
             * If '-d' enabled, report both semantic and emitter errors
             */
            function getSemanticDiagnostics(fileName) {
                synchronizeHostData();
                var targetSourceFile = getValidSourceFile(fileName);
                // Only perform the action per file regardless of '-out' flag as LanguageServiceHost is expected to call this function per file.
                // Therefore only get diagnostics for given file.
                var semanticDiagnostics = program.getSemanticDiagnostics(targetSourceFile, cancellationToken);
                if (!ts.getEmitDeclarations(program.getCompilerOptions())) {
                    return semanticDiagnostics.slice();
                }
                // If '-d' is enabled, check for emitter error. One example of emitter error is export class implements non-export interface
                var declarationDiagnostics = program.getDeclarationDiagnostics(targetSourceFile, cancellationToken);
                return __spreadArray(__spreadArray([], semanticDiagnostics, true), declarationDiagnostics, true);
            }
            function getSuggestionDiagnostics(fileName) {
                synchronizeHostData();
                return ts.computeSuggestionDiagnostics(getValidSourceFile(fileName), program, cancellationToken);
            }
            function getCompilerOptionsDiagnostics() {
                synchronizeHostData();
                return __spreadArray(__spreadArray([], program.getOptionsDiagnostics(cancellationToken), true), program.getGlobalDiagnostics(cancellationToken), true);
            }
            function getCompletionsAtPosition(fileName, position, options) {
                if (options === void 0) {
                    options = ts.emptyOptions;
                }
                // Convert from deprecated options names to new names
                var fullPreferences = __assign(__assign({}, ts.identity(options)), { includeCompletionsForModuleExports: options.includeCompletionsForModuleExports || options.includeExternalModuleExports, includeCompletionsWithInsertText: options.includeCompletionsWithInsertText || options.includeInsertTextCompletions });
                synchronizeHostData();
                return ts.Completions.getCompletionsAtPosition(host, program, log, getValidSourceFile(fileName), position, fullPreferences, options.triggerCharacter, options.triggerKind, cancellationToken);
            }
            function getCompletionEntryDetails(fileName, position, name, formattingOptions, source, preferences, data) {
                if (preferences === void 0) {
                    preferences = ts.emptyOptions;
                }
                synchronizeHostData();
                return ts.Completions.getCompletionEntryDetails(program, log, getValidSourceFile(fileName), position, { name: name, source: source, data: data }, host, (formattingOptions && ts.formatting.getFormatContext(formattingOptions, host)), // TODO: GH#18217
                preferences, cancellationToken);
            }
            function getCompletionEntrySymbol(fileName, position, name, source, preferences) {
                if (preferences === void 0) {
                    preferences = ts.emptyOptions;
                }
                synchronizeHostData();
                return ts.Completions.getCompletionEntrySymbol(program, log, getValidSourceFile(fileName), position, { name: name, source: source }, host, preferences);
            }
            function getQuickInfoAtPosition(fileName, position) {
                synchronizeHostData();
                var sourceFile = getValidSourceFile(fileName);
                var node = ts.getTouchingPropertyName(sourceFile, position);
                if (node === sourceFile) {
                    // Avoid giving quickInfo for the sourceFile as a whole.
                    return undefined;
                }
                var typeChecker = program.getTypeChecker();
                var nodeForQuickInfo = getNodeForQuickInfo(node);
                var symbol = getSymbolAtLocationForQuickInfo(nodeForQuickInfo, typeChecker);
                if (!symbol || typeChecker.isUnknownSymbol(symbol)) {
                    var type_2 = shouldGetType(sourceFile, nodeForQuickInfo, position) ? typeChecker.getTypeAtLocation(nodeForQuickInfo) : undefined;
                    return type_2 && {
                        kind: "" /* unknown */,
                        kindModifiers: "" /* none */,
                        textSpan: ts.createTextSpanFromNode(nodeForQuickInfo, sourceFile),
                        displayParts: typeChecker.runWithCancellationToken(cancellationToken, function (typeChecker) { return ts.typeToDisplayParts(typeChecker, type_2, ts.getContainerNode(nodeForQuickInfo)); }),
                        documentation: type_2.symbol ? type_2.symbol.getDocumentationComment(typeChecker) : undefined,
                        tags: type_2.symbol ? type_2.symbol.getJsDocTags(typeChecker) : undefined
                    };
                }
                var _a = typeChecker.runWithCancellationToken(cancellationToken, function (typeChecker) {
                    return ts.SymbolDisplay.getSymbolDisplayPartsDocumentationAndSymbolKind(typeChecker, symbol, sourceFile, ts.getContainerNode(nodeForQuickInfo), nodeForQuickInfo);
                }), symbolKind = _a.symbolKind, displayParts = _a.displayParts, documentation = _a.documentation, tags = _a.tags;
                return {
                    kind: symbolKind,
                    kindModifiers: ts.SymbolDisplay.getSymbolModifiers(typeChecker, symbol),
                    textSpan: ts.createTextSpanFromNode(nodeForQuickInfo, sourceFile),
                    displayParts: displayParts,
                    documentation: documentation,
                    tags: tags,
                };
            }
            function getNodeForQuickInfo(node) {
                if (ts.isNewExpression(node.parent) && node.pos === node.parent.pos) {
                    return node.parent.expression;
                }
                if (ts.isNamedTupleMember(node.parent) && node.pos === node.parent.pos) {
                    return node.parent;
                }
                return node;
            }
            function shouldGetType(sourceFile, node, position) {
                switch (node.kind) {
                    case 79 /* Identifier */:
                        return !ts.isLabelName(node) && !ts.isTagName(node) && !ts.isConstTypeReference(node.parent);
                    case 204 /* PropertyAccessExpression */:
                    case 159 /* QualifiedName */:
                        // Don't return quickInfo if inside the comment in `a/**/.b`
                        return !ts.isInComment(sourceFile, position);
                    case 108 /* ThisKeyword */:
                    case 190 /* ThisType */:
                    case 106 /* SuperKeyword */:
                    case 195 /* NamedTupleMember */:
                        return true;
                    default:
                        return false;
                }
            }
            /// Goto definition
            function getDefinitionAtPosition(fileName, position) {
                synchronizeHostData();
                return ts.GoToDefinition.getDefinitionAtPosition(program, getValidSourceFile(fileName), position);
            }
            function getDefinitionAndBoundSpan(fileName, position) {
                synchronizeHostData();
                return ts.GoToDefinition.getDefinitionAndBoundSpan(program, getValidSourceFile(fileName), position);
            }
            function getTypeDefinitionAtPosition(fileName, position) {
                synchronizeHostData();
                return ts.GoToDefinition.getTypeDefinitionAtPosition(program.getTypeChecker(), getValidSourceFile(fileName), position);
            }
            /// Goto implementation
            function getImplementationAtPosition(fileName, position) {
                synchronizeHostData();
                return ts.FindAllReferences.getImplementationsAtPosition(program, cancellationToken, program.getSourceFiles(), getValidSourceFile(fileName), position);
            }
            /// References and Occurrences
            function getOccurrencesAtPosition(fileName, position) {
                return ts.flatMap(getDocumentHighlights(fileName, position, [fileName]), function (entry) { return entry.highlightSpans.map(function (highlightSpan) { return (__assign(__assign({ fileName: entry.fileName, textSpan: highlightSpan.textSpan, isWriteAccess: highlightSpan.kind === "writtenReference" /* writtenReference */, isDefinition: false }, highlightSpan.isInString && { isInString: true }), highlightSpan.contextSpan && { contextSpan: highlightSpan.contextSpan })); }); });
            }
            function getDocumentHighlights(fileName, position, filesToSearch) {
                var normalizedFileName = ts.normalizePath(fileName);
                ts.Debug.assert(filesToSearch.some(function (f) { return ts.normalizePath(f) === normalizedFileName; }));
                synchronizeHostData();
                var sourceFilesToSearch = ts.mapDefined(filesToSearch, function (fileName) { return program.getSourceFile(fileName); });
                var sourceFile = getValidSourceFile(fileName);
                return ts.DocumentHighlights.getDocumentHighlights(program, cancellationToken, sourceFile, position, sourceFilesToSearch);
            }
            function findRenameLocations(fileName, position, findInStrings, findInComments, providePrefixAndSuffixTextForRename) {
                synchronizeHostData();
                var sourceFile = getValidSourceFile(fileName);
                var node = ts.getAdjustedRenameLocation(ts.getTouchingPropertyName(sourceFile, position));
                if (ts.isIdentifier(node) && (ts.isJsxOpeningElement(node.parent) || ts.isJsxClosingElement(node.parent)) && ts.isIntrinsicJsxName(node.escapedText)) {
                    var _a = node.parent.parent, openingElement = _a.openingElement, closingElement = _a.closingElement;
                    return [openingElement, closingElement].map(function (node) {
                        var textSpan = ts.createTextSpanFromNode(node.tagName, sourceFile);
                        return __assign({ fileName: sourceFile.fileName, textSpan: textSpan }, ts.FindAllReferences.toContextSpan(textSpan, sourceFile, node.parent));
                    });
                }
                else {
                    return getReferencesWorker(node, position, { findInStrings: findInStrings, findInComments: findInComments, providePrefixAndSuffixTextForRename: providePrefixAndSuffixTextForRename, use: 2 /* Rename */ }, function (entry, originalNode, checker) { return ts.FindAllReferences.toRenameLocation(entry, originalNode, checker, providePrefixAndSuffixTextForRename || false); });
                }
            }
            function getReferencesAtPosition(fileName, position) {
                synchronizeHostData();
                return getReferencesWorker(ts.getTouchingPropertyName(getValidSourceFile(fileName), position), position, { use: 1 /* References */ }, ts.FindAllReferences.toReferenceEntry);
            }
            function getReferencesWorker(node, position, options, cb) {
                synchronizeHostData();
                // Exclude default library when renaming as commonly user don't want to change that file.
                var sourceFiles = options && options.use === 2 /* Rename */
                    ? program.getSourceFiles().filter(function (sourceFile) { return !program.isSourceFileDefaultLibrary(sourceFile); })
                    : program.getSourceFiles();
                return ts.FindAllReferences.findReferenceOrRenameEntries(program, cancellationToken, sourceFiles, node, position, options, cb);
            }
            function findReferences(fileName, position) {
                synchronizeHostData();
                return ts.FindAllReferences.findReferencedSymbols(program, cancellationToken, program.getSourceFiles(), getValidSourceFile(fileName), position);
            }
            function getFileReferences(fileName) {
                synchronizeHostData();
                return ts.FindAllReferences.Core.getReferencesForFileName(fileName, program, program.getSourceFiles()).map(ts.FindAllReferences.toReferenceEntry);
            }
            function getNavigateToItems(searchValue, maxResultCount, fileName, excludeDtsFiles) {
                if (excludeDtsFiles === void 0) {
                    excludeDtsFiles = false;
                }
                synchronizeHostData();
                var sourceFiles = fileName ? [getValidSourceFile(fileName)] : program.getSourceFiles();
                return ts.NavigateTo.getNavigateToItems(sourceFiles, program.getTypeChecker(), cancellationToken, searchValue, maxResultCount, excludeDtsFiles);
            }
            function getEmitOutput(fileName, emitOnlyDtsFiles, forceDtsEmit) {
                synchronizeHostData();
                var sourceFile = getValidSourceFile(fileName);
                var customTransformers = host.getCustomTransformers && host.getCustomTransformers();
                return ts.getFileEmitOutput(program, sourceFile, !!emitOnlyDtsFiles, cancellationToken, customTransformers, forceDtsEmit);
            }
            // Signature help
            /**
             * This is a semantic operation.
             */
            function getSignatureHelpItems(fileName, position, _a) {
                var _b = _a === void 0 ? ts.emptyOptions : _a, triggerReason = _b.triggerReason;
                synchronizeHostData();
                var sourceFile = getValidSourceFile(fileName);
                return ts.SignatureHelp.getSignatureHelpItems(program, sourceFile, position, triggerReason, cancellationToken);
            }
            /// Syntactic features
            function getNonBoundSourceFile(fileName) {
                return syntaxTreeCache.getCurrentSourceFile(fileName);
            }
            function getNameOrDottedNameSpan(fileName, startPos, _endPos) {
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                // Get node at the location
                var node = ts.getTouchingPropertyName(sourceFile, startPos);
                if (node === sourceFile) {
                    return undefined;
                }
                switch (node.kind) {
                    case 204 /* PropertyAccessExpression */:
                    case 159 /* QualifiedName */:
                    case 10 /* StringLiteral */:
                    case 95 /* FalseKeyword */:
                    case 110 /* TrueKeyword */:
                    case 104 /* NullKeyword */:
                    case 106 /* SuperKeyword */:
                    case 108 /* ThisKeyword */:
                    case 190 /* ThisType */:
                    case 79 /* Identifier */:
                        break;
                    // Cant create the text span
                    default:
                        return undefined;
                }
                var nodeForStartPos = node;
                while (true) {
                    if (ts.isRightSideOfPropertyAccess(nodeForStartPos) || ts.isRightSideOfQualifiedName(nodeForStartPos)) {
                        // If on the span is in right side of the the property or qualified name, return the span from the qualified name pos to end of this node
                        nodeForStartPos = nodeForStartPos.parent;
                    }
                    else if (ts.isNameOfModuleDeclaration(nodeForStartPos)) {
                        // If this is name of a module declarations, check if this is right side of dotted module name
                        // If parent of the module declaration which is parent of this node is module declaration and its body is the module declaration that this node is name of
                        // Then this name is name from dotted module
                        if (nodeForStartPos.parent.parent.kind === 259 /* ModuleDeclaration */ &&
                            nodeForStartPos.parent.parent.body === nodeForStartPos.parent) {
                            // Use parent module declarations name for start pos
                            nodeForStartPos = nodeForStartPos.parent.parent.name;
                        }
                        else {
                            // We have to use this name for start pos
                            break;
                        }
                    }
                    else {
                        // Is not a member expression so we have found the node for start pos
                        break;
                    }
                }
                return ts.createTextSpanFromBounds(nodeForStartPos.getStart(), node.getEnd());
            }
            function getBreakpointStatementAtPosition(fileName, position) {
                // doesn't use compiler - no need to synchronize with host
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                return ts.BreakpointResolver.spanInSourceFileAtLocation(sourceFile, position);
            }
            function getNavigationBarItems(fileName) {
                return ts.NavigationBar.getNavigationBarItems(syntaxTreeCache.getCurrentSourceFile(fileName), cancellationToken);
            }
            function getNavigationTree(fileName) {
                return ts.NavigationBar.getNavigationTree(syntaxTreeCache.getCurrentSourceFile(fileName), cancellationToken);
            }
            function getSemanticClassifications(fileName, span, format) {
                synchronizeHostData();
                var responseFormat = format || "original" /* Original */;
                if (responseFormat === "2020" /* TwentyTwenty */) {
                    return ts.classifier.v2020.getSemanticClassifications(program, cancellationToken, getValidSourceFile(fileName), span);
                }
                else {
                    return ts.getSemanticClassifications(program.getTypeChecker(), cancellationToken, getValidSourceFile(fileName), program.getClassifiableNames(), span);
                }
            }
            function getEncodedSemanticClassifications(fileName, span, format) {
                synchronizeHostData();
                var responseFormat = format || "original" /* Original */;
                if (responseFormat === "original" /* Original */) {
                    return ts.getEncodedSemanticClassifications(program.getTypeChecker(), cancellationToken, getValidSourceFile(fileName), program.getClassifiableNames(), span);
                }
                else {
                    return ts.classifier.v2020.getEncodedSemanticClassifications(program, cancellationToken, getValidSourceFile(fileName), span);
                }
            }
            function getSyntacticClassifications(fileName, span) {
                // doesn't use compiler - no need to synchronize with host
                return ts.getSyntacticClassifications(cancellationToken, syntaxTreeCache.getCurrentSourceFile(fileName), span);
            }
            function getEncodedSyntacticClassifications(fileName, span) {
                // doesn't use compiler - no need to synchronize with host
                return ts.getEncodedSyntacticClassifications(cancellationToken, syntaxTreeCache.getCurrentSourceFile(fileName), span);
            }
            function getOutliningSpans(fileName) {
                // doesn't use compiler - no need to synchronize with host
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                return ts.OutliningElementsCollector.collectElements(sourceFile, cancellationToken);
            }
            var braceMatching = new ts.Map(ts.getEntries((_a = {},
                _a[18 /* OpenBraceToken */] = 19 /* CloseBraceToken */,
                _a[20 /* OpenParenToken */] = 21 /* CloseParenToken */,
                _a[22 /* OpenBracketToken */] = 23 /* CloseBracketToken */,
                _a[31 /* GreaterThanToken */] = 29 /* LessThanToken */,
                _a)));
            braceMatching.forEach(function (value, key) { return braceMatching.set(value.toString(), Number(key)); });
            function getBraceMatchingAtPosition(fileName, position) {
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                var token = ts.getTouchingToken(sourceFile, position);
                var matchKind = token.getStart(sourceFile) === position ? braceMatching.get(token.kind.toString()) : undefined;
                var match = matchKind && ts.findChildOfKind(token.parent, matchKind, sourceFile);
                // We want to order the braces when we return the result.
                return match ? [ts.createTextSpanFromNode(token, sourceFile), ts.createTextSpanFromNode(match, sourceFile)].sort(function (a, b) { return a.start - b.start; }) : ts.emptyArray;
            }
            function getIndentationAtPosition(fileName, position, editorOptions) {
                var start = ts.timestamp();
                var settings = toEditorSettings(editorOptions);
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                log("getIndentationAtPosition: getCurrentSourceFile: " + (ts.timestamp() - start));
                start = ts.timestamp();
                var result = ts.formatting.SmartIndenter.getIndentation(position, sourceFile, settings);
                log("getIndentationAtPosition: computeIndentation  : " + (ts.timestamp() - start));
                return result;
            }
            function getFormattingEditsForRange(fileName, start, end, options) {
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                return ts.formatting.formatSelection(start, end, sourceFile, ts.formatting.getFormatContext(toEditorSettings(options), host));
            }
            function getFormattingEditsForDocument(fileName, options) {
                return ts.formatting.formatDocument(syntaxTreeCache.getCurrentSourceFile(fileName), ts.formatting.getFormatContext(toEditorSettings(options), host));
            }
            function getFormattingEditsAfterKeystroke(fileName, position, key, options) {
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                var formatContext = ts.formatting.getFormatContext(toEditorSettings(options), host);
                if (!ts.isInComment(sourceFile, position)) {
                    switch (key) {
                        case "{":
                            return ts.formatting.formatOnOpeningCurly(position, sourceFile, formatContext);
                        case "}":
                            return ts.formatting.formatOnClosingCurly(position, sourceFile, formatContext);
                        case ";":
                            return ts.formatting.formatOnSemicolon(position, sourceFile, formatContext);
                        case "\n":
                            return ts.formatting.formatOnEnter(position, sourceFile, formatContext);
                    }
                }
                return [];
            }
            function getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
                if (preferences === void 0) {
                    preferences = ts.emptyOptions;
                }
                synchronizeHostData();
                var sourceFile = getValidSourceFile(fileName);
                var span = ts.createTextSpanFromBounds(start, end);
                var formatContext = ts.formatting.getFormatContext(formatOptions, host);
                return ts.flatMap(ts.deduplicate(errorCodes, ts.equateValues, ts.compareValues), function (errorCode) {
                    cancellationToken.throwIfCancellationRequested();
                    return ts.codefix.getFixes({ errorCode: errorCode, sourceFile: sourceFile, span: span, program: program, host: host, cancellationToken: cancellationToken, formatContext: formatContext, preferences: preferences });
                });
            }
            function getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
                if (preferences === void 0) {
                    preferences = ts.emptyOptions;
                }
                synchronizeHostData();
                ts.Debug.assert(scope.type === "file");
                var sourceFile = getValidSourceFile(scope.fileName);
                var formatContext = ts.formatting.getFormatContext(formatOptions, host);
                return ts.codefix.getAllFixes({ fixId: fixId, sourceFile: sourceFile, program: program, host: host, cancellationToken: cancellationToken, formatContext: formatContext, preferences: preferences });
            }
            function organizeImports(args, formatOptions, preferences) {
                if (preferences === void 0) {
                    preferences = ts.emptyOptions;
                }
                synchronizeHostData();
                ts.Debug.assert(args.type === "file");
                var sourceFile = getValidSourceFile(args.fileName);
                var formatContext = ts.formatting.getFormatContext(formatOptions, host);
                return ts.OrganizeImports.organizeImports(sourceFile, formatContext, host, program, preferences, args.skipDestructiveCodeActions);
            }
            function getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
                if (preferences === void 0) {
                    preferences = ts.emptyOptions;
                }
                return ts.getEditsForFileRename(getProgram(), oldFilePath, newFilePath, host, ts.formatting.getFormatContext(formatOptions, host), preferences, sourceMapper);
            }
            function applyCodeActionCommand(fileName, actionOrFormatSettingsOrUndefined) {
                var action = typeof fileName === "string" ? actionOrFormatSettingsOrUndefined : fileName;
                return ts.isArray(action) ? Promise.all(action.map(function (a) { return applySingleCodeActionCommand(a); })) : applySingleCodeActionCommand(action);
            }
            function applySingleCodeActionCommand(action) {
                var getPath = function (path) { return ts.toPath(path, currentDirectory, getCanonicalFileName); };
                ts.Debug.assertEqual(action.type, "install package");
                return host.installPackage
                    ? host.installPackage({ fileName: getPath(action.file), packageName: action.packageName })
                    : Promise.reject("Host does not implement `installPackage`");
            }
            function getDocCommentTemplateAtPosition(fileName, position, options) {
                return ts.JsDoc.getDocCommentTemplateAtPosition(ts.getNewLineOrDefaultFromHost(host), syntaxTreeCache.getCurrentSourceFile(fileName), position, options);
            }
            function isValidBraceCompletionAtPosition(fileName, position, openingBrace) {
                // '<' is currently not supported, figuring out if we're in a Generic Type vs. a comparison is too
                // expensive to do during typing scenarios
                // i.e. whether we're dealing with:
                //      var x = new foo<| ( with class foo<T>{} )
                // or
                //      var y = 3 <|
                if (openingBrace === 60 /* lessThan */) {
                    return false;
                }
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                // Check if in a context where we don't want to perform any insertion
                if (ts.isInString(sourceFile, position)) {
                    return false;
                }
                if (ts.isInsideJsxElementOrAttribute(sourceFile, position)) {
                    return openingBrace === 123 /* openBrace */;
                }
                if (ts.isInTemplateString(sourceFile, position)) {
                    return false;
                }
                switch (openingBrace) {
                    case 39 /* singleQuote */:
                    case 34 /* doubleQuote */:
                    case 96 /* backtick */:
                        return !ts.isInComment(sourceFile, position);
                }
                return true;
            }
            function getJsxClosingTagAtPosition(fileName, position) {
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                var token = ts.findPrecedingToken(position, sourceFile);
                if (!token)
                    return undefined;
                var element = token.kind === 31 /* GreaterThanToken */ && ts.isJsxOpeningElement(token.parent) ? token.parent.parent
                    : ts.isJsxText(token) ? token.parent : undefined;
                if (element && isUnclosedTag(element)) {
                    return { newText: "</" + element.openingElement.tagName.getText(sourceFile) + ">" };
                }
            }
            function getLinesForRange(sourceFile, textRange) {
                return {
                    lineStarts: sourceFile.getLineStarts(),
                    firstLine: sourceFile.getLineAndCharacterOfPosition(textRange.pos).line,
                    lastLine: sourceFile.getLineAndCharacterOfPosition(textRange.end).line
                };
            }
            function toggleLineComment(fileName, textRange, insertComment) {
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                var textChanges = [];
                var _a = getLinesForRange(sourceFile, textRange), lineStarts = _a.lineStarts, firstLine = _a.firstLine, lastLine = _a.lastLine;
                var isCommenting = insertComment || false;
                var leftMostPosition = Number.MAX_VALUE;
                var lineTextStarts = new ts.Map();
                var firstNonWhitespaceCharacterRegex = new RegExp(/\S/);
                var isJsx = ts.isInsideJsxElement(sourceFile, lineStarts[firstLine]);
                var openComment = isJsx ? "{/*" : "//";
                // Check each line before any text changes.
                for (var i = firstLine; i <= lastLine; i++) {
                    var lineText = sourceFile.text.substring(lineStarts[i], sourceFile.getLineEndOfPosition(lineStarts[i]));
                    // Find the start of text and the left-most character. No-op on empty lines.
                    var regExec = firstNonWhitespaceCharacterRegex.exec(lineText);
                    if (regExec) {
                        leftMostPosition = Math.min(leftMostPosition, regExec.index);
                        lineTextStarts.set(i.toString(), regExec.index);
                        if (lineText.substr(regExec.index, openComment.length) !== openComment) {
                            isCommenting = insertComment === undefined || insertComment;
                        }
                    }
                }
                // Push all text changes.
                for (var i = firstLine; i <= lastLine; i++) {
                    // If the range is multiline and ends on a beginning of a line, don't comment/uncomment.
                    if (firstLine !== lastLine && lineStarts[i] === textRange.end) {
                        continue;
                    }
                    var lineTextStart = lineTextStarts.get(i.toString());
                    // If the line is not an empty line; otherwise no-op.
                    if (lineTextStart !== undefined) {
                        if (isJsx) {
                            textChanges.push.apply(textChanges, toggleMultilineComment(fileName, { pos: lineStarts[i] + leftMostPosition, end: sourceFile.getLineEndOfPosition(lineStarts[i]) }, isCommenting, isJsx));
                        }
                        else if (isCommenting) {
                            textChanges.push({
                                newText: openComment,
                                span: {
                                    length: 0,
                                    start: lineStarts[i] + leftMostPosition
                                }
                            });
                        }
                        else if (sourceFile.text.substr(lineStarts[i] + lineTextStart, openComment.length) === openComment) {
                            textChanges.push({
                                newText: "",
                                span: {
                                    length: openComment.length,
                                    start: lineStarts[i] + lineTextStart
                                }
                            });
                        }
                    }
                }
                return textChanges;
            }
            function toggleMultilineComment(fileName, textRange, insertComment, isInsideJsx) {
                var _a;
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                var textChanges = [];
                var text = sourceFile.text;
                var hasComment = false;
                var isCommenting = insertComment || false;
                var positions = [];
                var pos = textRange.pos;
                var isJsx = isInsideJsx !== undefined ? isInsideJsx : ts.isInsideJsxElement(sourceFile, pos);
                var openMultiline = isJsx ? "{/*" : "/*";
                var closeMultiline = isJsx ? "*/}" : "*/";
                var openMultilineRegex = isJsx ? "\\{\\/\\*" : "\\/\\*";
                var closeMultilineRegex = isJsx ? "\\*\\/\\}" : "\\*\\/";
                // Get all comment positions
                while (pos <= textRange.end) {
                    // Start of comment is considered inside comment.
                    var offset = text.substr(pos, openMultiline.length) === openMultiline ? openMultiline.length : 0;
                    var commentRange = ts.isInComment(sourceFile, pos + offset);
                    // If position is in a comment add it to the positions array.
                    if (commentRange) {
                        // Comment range doesn't include the brace character. Increase it to include them.
                        if (isJsx) {
                            commentRange.pos--;
                            commentRange.end++;
                        }
                        positions.push(commentRange.pos);
                        if (commentRange.kind === 3 /* MultiLineCommentTrivia */) {
                            positions.push(commentRange.end);
                        }
                        hasComment = true;
                        pos = commentRange.end + 1;
                    }
                    else { // If it's not in a comment range, then we need to comment the uncommented portions.
                        var newPos = text.substring(pos, textRange.end).search("(" + openMultilineRegex + ")|(" + closeMultilineRegex + ")");
                        isCommenting = insertComment !== undefined
                            ? insertComment
                            : isCommenting || !ts.isTextWhiteSpaceLike(text, pos, newPos === -1 ? textRange.end : pos + newPos); // If isCommenting is already true we don't need to check whitespace again.
                        pos = newPos === -1 ? textRange.end + 1 : pos + newPos + closeMultiline.length;
                    }
                }
                // If it didn't found a comment and isCommenting is false means is only empty space.
                // We want to insert comment in this scenario.
                if (isCommenting || !hasComment) {
                    if (((_a = ts.isInComment(sourceFile, textRange.pos)) === null || _a === void 0 ? void 0 : _a.kind) !== 2 /* SingleLineCommentTrivia */) {
                        ts.insertSorted(positions, textRange.pos, ts.compareValues);
                    }
                    ts.insertSorted(positions, textRange.end, ts.compareValues);
                    // Insert open comment if the first position is not a comment already.
                    var firstPos = positions[0];
                    if (text.substr(firstPos, openMultiline.length) !== openMultiline) {
                        textChanges.push({
                            newText: openMultiline,
                            span: {
                                length: 0,
                                start: firstPos
                            }
                        });
                    }
                    // Insert open and close comment to all positions between first and last. Exclusive.
                    for (var i = 1; i < positions.length - 1; i++) {
                        if (text.substr(positions[i] - closeMultiline.length, closeMultiline.length) !== closeMultiline) {
                            textChanges.push({
                                newText: closeMultiline,
                                span: {
                                    length: 0,
                                    start: positions[i]
                                }
                            });
                        }
                        if (text.substr(positions[i], openMultiline.length) !== openMultiline) {
                            textChanges.push({
                                newText: openMultiline,
                                span: {
                                    length: 0,
                                    start: positions[i]
                                }
                            });
                        }
                    }
                    // Insert open comment if the last position is not a comment already.
                    if (textChanges.length % 2 !== 0) {
                        textChanges.push({
                            newText: closeMultiline,
                            span: {
                                length: 0,
                                start: positions[positions.length - 1]
                            }
                        });
                    }
                }
                else {
                    // If is not commenting then remove all comments found.
                    for (var _i = 0, positions_1 = positions; _i < positions_1.length; _i++) {
                        var pos_2 = positions_1[_i];
                        var from = pos_2 - closeMultiline.length > 0 ? pos_2 - closeMultiline.length : 0;
                        var offset = text.substr(from, closeMultiline.length) === closeMultiline ? closeMultiline.length : 0;
                        textChanges.push({
                            newText: "",
                            span: {
                                length: openMultiline.length,
                                start: pos_2 - offset
                            }
                        });
                    }
                }
                return textChanges;
            }
            function commentSelection(fileName, textRange) {
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                var _a = getLinesForRange(sourceFile, textRange), firstLine = _a.firstLine, lastLine = _a.lastLine;
                // If there is a selection that is on the same line, add multiline.
                return firstLine === lastLine && textRange.pos !== textRange.end
                    ? toggleMultilineComment(fileName, textRange, /*insertComment*/ true)
                    : toggleLineComment(fileName, textRange, /*insertComment*/ true);
            }
            function uncommentSelection(fileName, textRange) {
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                var textChanges = [];
                var pos = textRange.pos;
                var end = textRange.end;
                // If cursor is not a selection we need to increase the end position
                // to include the start of the comment.
                if (pos === end) {
                    end += ts.isInsideJsxElement(sourceFile, pos) ? 2 : 1;
                }
                for (var i = pos; i <= end; i++) {
                    var commentRange = ts.isInComment(sourceFile, i);
                    if (commentRange) {
                        switch (commentRange.kind) {
                            case 2 /* SingleLineCommentTrivia */:
                                textChanges.push.apply(textChanges, toggleLineComment(fileName, { end: commentRange.end, pos: commentRange.pos + 1 }, /*insertComment*/ false));
                                break;
                            case 3 /* MultiLineCommentTrivia */:
                                textChanges.push.apply(textChanges, toggleMultilineComment(fileName, { end: commentRange.end, pos: commentRange.pos + 1 }, /*insertComment*/ false));
                        }
                        i = commentRange.end + 1;
                    }
                }
                return textChanges;
            }
            function isUnclosedTag(_a) {
                var openingElement = _a.openingElement, closingElement = _a.closingElement, parent = _a.parent;
                return !ts.tagNamesAreEquivalent(openingElement.tagName, closingElement.tagName) ||
                    ts.isJsxElement(parent) && ts.tagNamesAreEquivalent(openingElement.tagName, parent.openingElement.tagName) && isUnclosedTag(parent);
            }
            function getSpanOfEnclosingComment(fileName, position, onlyMultiLine) {
                var sourceFile = syntaxTreeCache.getCurrentSourceFile(fileName);
                var range = ts.formatting.getRangeOfEnclosingComment(sourceFile, position);
                return range && (!onlyMultiLine || range.kind === 3 /* MultiLineCommentTrivia */) ? ts.createTextSpanFromRange(range) : undefined;
            }
            function getTodoComments(fileName, descriptors) {
                // Note: while getting todo comments seems like a syntactic operation, we actually
                // treat it as a semantic operation here.  This is because we expect our host to call
                // this on every single file.  If we treat this syntactically, then that will cause
                // us to populate and throw away the tree in our syntax tree cache for each file.  By
                // treating this as a semantic operation, we can access any tree without throwing
                // anything away.
                synchronizeHostData();
                var sourceFile = getValidSourceFile(fileName);
                cancellationToken.throwIfCancellationRequested();
                var fileContents = sourceFile.text;
                var result = [];
                // Exclude node_modules files as we don't want to show the todos of external libraries.
                if (descriptors.length > 0 && !isNodeModulesFile(sourceFile.fileName)) {
                    var regExp = getTodoCommentsRegExp();
                    var matchArray = void 0;
                    while (matchArray = regExp.exec(fileContents)) {
                        cancellationToken.throwIfCancellationRequested();
                        // If we got a match, here is what the match array will look like.  Say the source text is:
                        //
                        //      "    // hack   1"
                        //
                        // The result array with the regexp:    will be:
                        //
                        //      ["// hack   1", "// ", "hack   1", undefined, "hack"]
                        //
                        // Here are the relevant capture groups:
                        //  0) The full match for the entire regexp.
                        //  1) The preamble to the message portion.
                        //  2) The message portion.
                        //  3...N) The descriptor that was matched - by index.  'undefined' for each
                        //         descriptor that didn't match.  an actual value if it did match.
                        //
                        //  i.e. 'undefined' in position 3 above means TODO(jason) didn't match.
                        //       "hack"      in position 4 means HACK did match.
                        var firstDescriptorCaptureIndex = 3;
                        ts.Debug.assert(matchArray.length === descriptors.length + firstDescriptorCaptureIndex);
                        var preamble = matchArray[1];
                        var matchPosition = matchArray.index + preamble.length;
                        // OK, we have found a match in the file.  This is only an acceptable match if
                        // it is contained within a comment.
                        if (!ts.isInComment(sourceFile, matchPosition)) {
                            continue;
                        }
                        var descriptor = void 0;
                        for (var i = 0; i < descriptors.length; i++) {
                            if (matchArray[i + firstDescriptorCaptureIndex]) {
                                descriptor = descriptors[i];
                            }
                        }
                        if (descriptor === undefined)
                            return ts.Debug.fail();
                        // We don't want to match something like 'TODOBY', so we make sure a non
                        // letter/digit follows the match.
                        if (isLetterOrDigit(fileContents.charCodeAt(matchPosition + descriptor.text.length))) {
                            continue;
                        }
                        var message = matchArray[2];
                        result.push({ descriptor: descriptor, message: message, position: matchPosition });
                    }
                }
                return result;
                function escapeRegExp(str) {
                    return str.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&");
                }
                function getTodoCommentsRegExp() {
                    // NOTE: `?:` means 'non-capture group'.  It allows us to have groups without having to
                    // filter them out later in the final result array.
                    // TODO comments can appear in one of the following forms:
                    //
                    //  1)      // TODO     or  /////////// TODO
                    //
                    //  2)      /* TODO     or  /********** TODO
                    //
                    //  3)      /*
                    //           *   TODO
                    //           */
                    //
                    // The following three regexps are used to match the start of the text up to the TODO
                    // comment portion.
                    var singleLineCommentStart = /(?:\/\/+\s*)/.source;
                    var multiLineCommentStart = /(?:\/\*+\s*)/.source;
                    var anyNumberOfSpacesAndAsterisksAtStartOfLine = /(?:^(?:\s|\*)*)/.source;
                    // Match any of the above three TODO comment start regexps.
                    // Note that the outermost group *is* a capture group.  We want to capture the preamble
                    // so that we can determine the starting position of the TODO comment match.
                    var preamble = "(" + anyNumberOfSpacesAndAsterisksAtStartOfLine + "|" + singleLineCommentStart + "|" + multiLineCommentStart + ")";
                    // Takes the descriptors and forms a regexp that matches them as if they were literals.
                    // For example, if the descriptors are "TODO(jason)" and "HACK", then this will be:
                    //
                    //      (?:(TODO\(jason\))|(HACK))
                    //
                    // Note that the outermost group is *not* a capture group, but the innermost groups
                    // *are* capture groups.  By capturing the inner literals we can determine after
                    // matching which descriptor we are dealing with.
                    var literals = "(?:" + ts.map(descriptors, function (d) { return "(" + escapeRegExp(d.text) + ")"; }).join("|") + ")";
                    // After matching a descriptor literal, the following regexp matches the rest of the
                    // text up to the end of the line (or */).
                    var endOfLineOrEndOfComment = /(?:$|\*\/)/.source;
                    var messageRemainder = /(?:.*?)/.source;
                    // This is the portion of the match we'll return as part of the TODO comment result. We
                    // match the literal portion up to the end of the line or end of comment.
                    var messagePortion = "(" + literals + messageRemainder + ")";
                    var regExpString = preamble + messagePortion + endOfLineOrEndOfComment;
                    // The final regexp will look like this:
                    // /((?:\/\/+\s*)|(?:\/\*+\s*)|(?:^(?:\s|\*)*))((?:(TODO\(jason\))|(HACK))(?:.*?))(?:$|\*\/)/gim
                    // The flags of the regexp are important here.
                    //  'g' is so that we are doing a global search and can find matches several times
                    //  in the input.
                    //
                    //  'i' is for case insensitivity (We do this to match C# TODO comment code).
                    //
                    //  'm' is so we can find matches in a multi-line input.
                    return new RegExp(regExpString, "gim");
                }
                function isLetterOrDigit(char) {
                    return (char >= 97 /* a */ && char <= 122 /* z */) ||
                        (char >= 65 /* A */ && char <= 90 /* Z */) ||
                        (char >= 48 /* _0 */ && char <= 57 /* _9 */);
                }
                function isNodeModulesFile(path) {
                    return ts.stringContains(path, "/node_modules/");
                }
            }
            function getRenameInfo(fileName, position, options) {
                synchronizeHostData();
                return ts.Rename.getRenameInfo(program, getValidSourceFile(fileName), position, options);
            }
            function getRefactorContext(file, positionOrRange, preferences, formatOptions, triggerReason, kind) {
                var _a = typeof positionOrRange === "number" ? [positionOrRange, undefined] : [positionOrRange.pos, positionOrRange.end], startPosition = _a[0], endPosition = _a[1];
                return {
                    file: file,
                    startPosition: startPosition,
                    endPosition: endPosition,
                    program: getProgram(),
                    host: host,
                    formatContext: ts.formatting.getFormatContext(formatOptions, host),
                    cancellationToken: cancellationToken,
                    preferences: preferences,
                    triggerReason: triggerReason,
                    kind: kind
                };
            }
            function getInlayHintsContext(file, span, preferences) {
                return {
                    file: file,
                    program: getProgram(),
                    host: host,
                    span: span,
                    preferences: preferences,
                    cancellationToken: cancellationToken,
                };
            }
            function getSmartSelectionRange(fileName, position) {
                return ts.SmartSelectionRange.getSmartSelectionRange(position, syntaxTreeCache.getCurrentSourceFile(fileName));
            }
            function getApplicableRefactors(fileName, positionOrRange, preferences, triggerReason, kind) {
                if (preferences === void 0) {
                    preferences = ts.emptyOptions;
                }
                synchronizeHostData();
                var file = getValidSourceFile(fileName);
                return ts.refactor.getApplicableRefactors(getRefactorContext(file, positionOrRange, preferences, ts.emptyOptions, triggerReason, kind));
            }
            function getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
                if (preferences === void 0) {
                    preferences = ts.emptyOptions;
                }
                synchronizeHostData();
                var file = getValidSourceFile(fileName);
                return ts.refactor.getEditsForRefactor(getRefactorContext(file, positionOrRange, preferences, formatOptions), refactorName, actionName);
            }
            function toLineColumnOffset(fileName, position) {
                // Go to Definition supports returning a zero-length span at position 0 for
                // non-existent files. We need to special-case the conversion of position 0
                // to avoid a crash trying to get the text for that file, since this function
                // otherwise assumes that 'fileName' is the name of a file that exists.
                if (position === 0) {
                    return { line: 0, character: 0 };
                }
                return sourceMapper.toLineColumnOffset(fileName, position);
            }
            function prepareCallHierarchy(fileName, position) {
                synchronizeHostData();
                var declarations = ts.CallHierarchy.resolveCallHierarchyDeclaration(program, ts.getTouchingPropertyName(getValidSourceFile(fileName), position));
                return declarations && ts.mapOneOrMany(declarations, function (declaration) { return ts.CallHierarchy.createCallHierarchyItem(program, declaration); });
            }
            function provideCallHierarchyIncomingCalls(fileName, position) {
                synchronizeHostData();
                var sourceFile = getValidSourceFile(fileName);
                var declaration = ts.firstOrOnly(ts.CallHierarchy.resolveCallHierarchyDeclaration(program, position === 0 ? sourceFile : ts.getTouchingPropertyName(sourceFile, position)));
                return declaration ? ts.CallHierarchy.getIncomingCalls(program, declaration, cancellationToken) : [];
            }
            function provideCallHierarchyOutgoingCalls(fileName, position) {
                synchronizeHostData();
                var sourceFile = getValidSourceFile(fileName);
                var declaration = ts.firstOrOnly(ts.CallHierarchy.resolveCallHierarchyDeclaration(program, position === 0 ? sourceFile : ts.getTouchingPropertyName(sourceFile, position)));
                return declaration ? ts.CallHierarchy.getOutgoingCalls(program, declaration) : [];
            }
            function provideInlayHints(fileName, span, preferences) {
                if (preferences === void 0) {
                    preferences = ts.emptyOptions;
                }
                synchronizeHostData();
                var sourceFile = getValidSourceFile(fileName);
                return ts.InlayHints.provideInlayHints(getInlayHintsContext(sourceFile, span, preferences));
            }
            var ls = {
                dispose: dispose,
                cleanupSemanticCache: cleanupSemanticCache,
                getSyntacticDiagnostics: getSyntacticDiagnostics,
                getSemanticDiagnostics: getSemanticDiagnostics,
                getSuggestionDiagnostics: getSuggestionDiagnostics,
                getCompilerOptionsDiagnostics: getCompilerOptionsDiagnostics,
                getSyntacticClassifications: getSyntacticClassifications,
                getSemanticClassifications: getSemanticClassifications,
                getEncodedSyntacticClassifications: getEncodedSyntacticClassifications,
                getEncodedSemanticClassifications: getEncodedSemanticClassifications,
                getCompletionsAtPosition: getCompletionsAtPosition,
                getCompletionEntryDetails: getCompletionEntryDetails,
                getCompletionEntrySymbol: getCompletionEntrySymbol,
                getSignatureHelpItems: getSignatureHelpItems,
                getQuickInfoAtPosition: getQuickInfoAtPosition,
                getDefinitionAtPosition: getDefinitionAtPosition,
                getDefinitionAndBoundSpan: getDefinitionAndBoundSpan,
                getImplementationAtPosition: getImplementationAtPosition,
                getTypeDefinitionAtPosition: getTypeDefinitionAtPosition,
                getReferencesAtPosition: getReferencesAtPosition,
                findReferences: findReferences,
                getFileReferences: getFileReferences,
                getOccurrencesAtPosition: getOccurrencesAtPosition,
                getDocumentHighlights: getDocumentHighlights,
                getNameOrDottedNameSpan: getNameOrDottedNameSpan,
                getBreakpointStatementAtPosition: getBreakpointStatementAtPosition,
                getNavigateToItems: getNavigateToItems,
                getRenameInfo: getRenameInfo,
                getSmartSelectionRange: getSmartSelectionRange,
                findRenameLocations: findRenameLocations,
                getNavigationBarItems: getNavigationBarItems,
                getNavigationTree: getNavigationTree,
                getOutliningSpans: getOutliningSpans,
                getTodoComments: getTodoComments,
                getBraceMatchingAtPosition: getBraceMatchingAtPosition,
                getIndentationAtPosition: getIndentationAtPosition,
                getFormattingEditsForRange: getFormattingEditsForRange,
                getFormattingEditsForDocument: getFormattingEditsForDocument,
                getFormattingEditsAfterKeystroke: getFormattingEditsAfterKeystroke,
                getDocCommentTemplateAtPosition: getDocCommentTemplateAtPosition,
                isValidBraceCompletionAtPosition: isValidBraceCompletionAtPosition,
                getJsxClosingTagAtPosition: getJsxClosingTagAtPosition,
                getSpanOfEnclosingComment: getSpanOfEnclosingComment,
                getCodeFixesAtPosition: getCodeFixesAtPosition,
                getCombinedCodeFix: getCombinedCodeFix,
                applyCodeActionCommand: applyCodeActionCommand,
                organizeImports: organizeImports,
                getEditsForFileRename: getEditsForFileRename,
                getEmitOutput: getEmitOutput,
                getNonBoundSourceFile: getNonBoundSourceFile,
                getProgram: getProgram,
                getAutoImportProvider: getAutoImportProvider,
                getApplicableRefactors: getApplicableRefactors,
                getEditsForRefactor: getEditsForRefactor,
                toLineColumnOffset: toLineColumnOffset,
                getSourceMapper: function () { return sourceMapper; },
                clearSourceMapperCache: function () { return sourceMapper.clearCache(); },
                prepareCallHierarchy: prepareCallHierarchy,
                provideCallHierarchyIncomingCalls: provideCallHierarchyIncomingCalls,
                provideCallHierarchyOutgoingCalls: provideCallHierarchyOutgoingCalls,
                toggleLineComment: toggleLineComment,
                toggleMultilineComment: toggleMultilineComment,
                commentSelection: commentSelection,
                uncommentSelection: uncommentSelection,
                provideInlayHints: provideInlayHints,
            };
            switch (languageServiceMode) {
                case ts.LanguageServiceMode.Semantic:
                    break;
                case ts.LanguageServiceMode.PartialSemantic:
                    invalidOperationsInPartialSemanticMode.forEach(function (key) {
                        return ls[key] = function () {
                            throw new Error("LanguageService Operation: " + key + " not allowed in LanguageServiceMode.PartialSemantic");
                        };
                    });
                    break;
                case ts.LanguageServiceMode.Syntactic:
                    invalidOperationsInSyntacticMode.forEach(function (key) {
                        return ls[key] = function () {
                            throw new Error("LanguageService Operation: " + key + " not allowed in LanguageServiceMode.Syntactic");
                        };
                    });
                    break;
                default:
                    ts.Debug.assertNever(languageServiceMode);
            }
            return ls;
        }
        ts.createLanguageService = createLanguageService;
        /* @internal */
        /** Names in the name table are escaped, so an identifier `__foo` will have a name table entry `___foo`. */
        function getNameTable(sourceFile) {
            if (!sourceFile.nameTable) {
                initializeNameTable(sourceFile);
            }
            return sourceFile.nameTable; // TODO: GH#18217
        }
        ts.getNameTable = getNameTable;
        function initializeNameTable(sourceFile) {
            var nameTable = sourceFile.nameTable = new ts.Map();
            sourceFile.forEachChild(function walk(node) {
                if (ts.isIdentifier(node) && !ts.isTagName(node) && node.escapedText || ts.isStringOrNumericLiteralLike(node) && literalIsName(node)) {
                    var text = ts.getEscapedTextOfIdentifierOrLiteral(node);
                    nameTable.set(text, nameTable.get(text) === undefined ? node.pos : -1);
                }
                else if (ts.isPrivateIdentifier(node)) {
                    var text = node.escapedText;
                    nameTable.set(text, nameTable.get(text) === undefined ? node.pos : -1);
                }
                ts.forEachChild(node, walk);
                if (ts.hasJSDocNodes(node)) {
                    for (var _i = 0, _a = node.jsDoc; _i < _a.length; _i++) {
                        var jsDoc = _a[_i];
                        ts.forEachChild(jsDoc, walk);
                    }
                }
            });
        }
        /**
         * We want to store any numbers/strings if they were a name that could be
         * related to a declaration.  So, if we have 'import x = require("something")'
         * then we want 'something' to be in the name table.  Similarly, if we have
         * "a['propname']" then we want to store "propname" in the name table.
         */
        function literalIsName(node) {
            return ts.isDeclarationName(node) ||
                node.parent.kind === 275 /* ExternalModuleReference */ ||
                isArgumentOfElementAccessExpression(node) ||
                ts.isLiteralComputedPropertyDeclarationName(node);
        }
        /**
         * Returns the containing object literal property declaration given a possible name node, e.g. "a" in x = { "a": 1 }
         */
        /* @internal */
        function getContainingObjectLiteralElement(node) {
            var element = getContainingObjectLiteralElementWorker(node);
            return element && (ts.isObjectLiteralExpression(element.parent) || ts.isJsxAttributes(element.parent)) ? element : undefined;
        }
        ts.getContainingObjectLiteralElement = getContainingObjectLiteralElement;
        function getContainingObjectLiteralElementWorker(node) {
            switch (node.kind) {
                case 10 /* StringLiteral */:
                case 14 /* NoSubstitutionTemplateLiteral */:
                case 8 /* NumericLiteral */:
                    if (node.parent.kind === 160 /* ComputedPropertyName */) {
                        return ts.isObjectLiteralElement(node.parent.parent) ? node.parent.parent : undefined;
                    }
                // falls through
                case 79 /* Identifier */:
                    return ts.isObjectLiteralElement(node.parent) &&
                        (node.parent.parent.kind === 203 /* ObjectLiteralExpression */ || node.parent.parent.kind === 284 /* JsxAttributes */) &&
                        node.parent.name === node ? node.parent : undefined;
            }
            return undefined;
        }
        function getSymbolAtLocationForQuickInfo(node, checker) {
            var object = getContainingObjectLiteralElement(node);
            if (object) {
                var contextualType = checker.getContextualType(object.parent);
                var properties = contextualType && getPropertySymbolsFromContextualType(object, checker, contextualType, /*unionSymbolOk*/ false);
                if (properties && properties.length === 1) {
                    return ts.first(properties);
                }
            }
            return checker.getSymbolAtLocation(node);
        }
        /** Gets all symbols for one property. Does not get symbols for every property. */
        /* @internal */
        function getPropertySymbolsFromContextualType(node, checker, contextualType, unionSymbolOk) {
            var name = ts.getNameFromPropertyName(node.name);
            if (!name)
                return ts.emptyArray;
            if (!contextualType.isUnion()) {
                var symbol = contextualType.getProperty(name);
                return symbol ? [symbol] : ts.emptyArray;
            }
            var discriminatedPropertySymbols = ts.mapDefined(contextualType.types, function (t) { return (ts.isObjectLiteralExpression(node.parent) || ts.isJsxAttributes(node.parent)) && checker.isTypeInvalidDueToUnionDiscriminant(t, node.parent) ? undefined : t.getProperty(name); });
            if (unionSymbolOk && (discriminatedPropertySymbols.length === 0 || discriminatedPropertySymbols.length === contextualType.types.length)) {
                var symbol = contextualType.getProperty(name);
                if (symbol)
                    return [symbol];
            }
            if (discriminatedPropertySymbols.length === 0) {
                // Bad discriminant -- do again without discriminating
                return ts.mapDefined(contextualType.types, function (t) { return t.getProperty(name); });
            }
            return discriminatedPropertySymbols;
        }
        ts.getPropertySymbolsFromContextualType = getPropertySymbolsFromContextualType;
        function isArgumentOfElementAccessExpression(node) {
            return node &&
                node.parent &&
                node.parent.kind === 205 /* ElementAccessExpression */ &&
                node.parent.argumentExpression === node;
        }
        /**
         * Get the path of the default library files (lib.d.ts) as distributed with the typescript
         * node package.
         * The functionality is not supported if the ts module is consumed outside of a node module.
         */
        function getDefaultLibFilePath(options) {
            // Check __dirname is defined and that we are on a node.js system.
            if (typeof __dirname !== "undefined") {
                return __dirname + ts.directorySeparator + ts.getDefaultLibFileName(options);
            }
            throw new Error("getDefaultLibFilePath is only supported when consumed as a node module. ");
        }
        ts.getDefaultLibFilePath = getDefaultLibFilePath;
        ts.setObjectAllocator(getServicesObjectAllocator());
    })(ts || (ts = {}));
    /* @internal */
    var ts;
    (function (ts) {
        var BreakpointResolver;
        (function (BreakpointResolver) {
            /**
             * Get the breakpoint span in given sourceFile
             */
            function spanInSourceFileAtLocation(sourceFile, position) {
                // Cannot set breakpoint in dts file
                if (sourceFile.isDeclarationFile) {
                    return undefined;
                }
                var tokenAtLocation = ts.getTokenAtPosition(sourceFile, position);
                var lineOfPosition = sourceFile.getLineAndCharacterOfPosition(position).line;
                if (sourceFile.getLineAndCharacterOfPosition(tokenAtLocation.getStart(sourceFile)).line > lineOfPosition) {
                    // Get previous token if the token is returned starts on new line
                    // eg: let x =10; |--- cursor is here
                    //     let y = 10;
                    // token at position will return let keyword on second line as the token but we would like to use
                    // token on same line if trailing trivia (comments or white spaces on same line) part of the last token on that line
                    var preceding = ts.findPrecedingToken(tokenAtLocation.pos, sourceFile);
                    // It's a blank line
                    if (!preceding || sourceFile.getLineAndCharacterOfPosition(preceding.getEnd()).line !== lineOfPosition) {
                        return undefined;
                    }
                    tokenAtLocation = preceding;
                }
                // Cannot set breakpoint in ambient declarations
                if (tokenAtLocation.flags & 8388608 /* Ambient */) {
                    return undefined;
                }
                // Get the span in the node based on its syntax
                return spanInNode(tokenAtLocation);
                function textSpan(startNode, endNode) {
                    var start = startNode.decorators ?
                        ts.skipTrivia(sourceFile.text, startNode.decorators.end) :
                        startNode.getStart(sourceFile);
                    return ts.createTextSpanFromBounds(start, (endNode || startNode).getEnd());
                }
                function textSpanEndingAtNextToken(startNode, previousTokenToFindNextEndToken) {
                    return textSpan(startNode, ts.findNextToken(previousTokenToFindNextEndToken, previousTokenToFindNextEndToken.parent, sourceFile));
                }
                function spanInNodeIfStartsOnSameLine(node, otherwiseOnNode) {
                    if (node && lineOfPosition === sourceFile.getLineAndCharacterOfPosition(node.getStart(sourceFile)).line) {
                        return spanInNode(node);
                    }
                    return spanInNode(otherwiseOnNode);
                }
                function spanInNodeArray(nodeArray) {
                    return ts.createTextSpanFromBounds(ts.skipTrivia(sourceFile.text, nodeArray.pos), nodeArray.end);
                }
                function spanInPreviousNode(node) {
                    return spanInNode(ts.findPrecedingToken(node.pos, sourceFile));
                }
                function spanInNextNode(node) {
                    return spanInNode(ts.findNextToken(node, node.parent, sourceFile));
                }
                function spanInNode(node) {
                    if (node) {
                        var parent = node.parent;
                        switch (node.kind) {
                            case 235 /* VariableStatement */:
                                // Span on first variable declaration
                                return spanInVariableDeclaration(node.declarationList.declarations[0]);
                            case 252 /* VariableDeclaration */:
                            case 165 /* PropertyDeclaration */:
                            case 164 /* PropertySignature */:
                                return spanInVariableDeclaration(node);
                            case 162 /* Parameter */:
                                return spanInParameterDeclaration(node);
                            case 254 /* FunctionDeclaration */:
                            case 167 /* MethodDeclaration */:
                            case 166 /* MethodSignature */:
                            case 170 /* GetAccessor */:
                            case 171 /* SetAccessor */:
                            case 169 /* Constructor */:
                            case 211 /* FunctionExpression */:
                            case 212 /* ArrowFunction */:
                                return spanInFunctionDeclaration(node);
                            case 233 /* Block */:
                                if (ts.isFunctionBlock(node)) {
                                    return spanInFunctionBlock(node);
                                }
                            // falls through
                            case 260 /* ModuleBlock */:
                                return spanInBlock(node);
                            case 290 /* CatchClause */:
                                return spanInBlock(node.block);
                            case 236 /* ExpressionStatement */:
                                // span on the expression
                                return textSpan(node.expression);
                            case 245 /* ReturnStatement */:
                                // span on return keyword and expression if present
                                return textSpan(node.getChildAt(0), node.expression);
                            case 239 /* WhileStatement */:
                                // Span on while(...)
                                return textSpanEndingAtNextToken(node, node.expression);
                            case 238 /* DoStatement */:
                                // span in statement of the do statement
                                return spanInNode(node.statement);
                            case 251 /* DebuggerStatement */:
                                // span on debugger keyword
                                return textSpan(node.getChildAt(0));
                            case 237 /* IfStatement */:
                                // set on if(..) span
                                return textSpanEndingAtNextToken(node, node.expression);
                            case 248 /* LabeledStatement */:
                                // span in statement
                                return spanInNode(node.statement);
                            case 244 /* BreakStatement */:
                            case 243 /* ContinueStatement */:
                                // On break or continue keyword and label if present
                                return textSpan(node.getChildAt(0), node.label);
                            case 240 /* ForStatement */:
                                return spanInForStatement(node);
                            case 241 /* ForInStatement */:
                                // span of for (a in ...)
                                return textSpanEndingAtNextToken(node, node.expression);
                            case 242 /* ForOfStatement */:
                                // span in initializer
                                return spanInInitializerOfForLike(node);
                            case 247 /* SwitchStatement */:
                                // span on switch(...)
                                return textSpanEndingAtNextToken(node, node.expression);
                            case 287 /* CaseClause */:
                            case 288 /* DefaultClause */:
                                // span in first statement of the clause
                                return spanInNode(node.statements[0]);
                            case 250 /* TryStatement */:
                                // span in try block
                                return spanInBlock(node.tryBlock);
                            case 249 /* ThrowStatement */:
                                // span in throw ...
                                return textSpan(node, node.expression);
                            case 269 /* ExportAssignment */:
                                // span on export = id
                                return textSpan(node, node.expression);
                            case 263 /* ImportEqualsDeclaration */:
                                // import statement without including semicolon
                                return textSpan(node, node.moduleReference);
                            case 264 /* ImportDeclaration */:
                                // import statement without including semicolon
                                return textSpan(node, node.moduleSpecifier);
                            case 270 /* ExportDeclaration */:
                                // import statement without including semicolon
                                return textSpan(node, node.moduleSpecifier);
                            case 259 /* ModuleDeclaration */:
                                // span on complete module if it is instantiated
                                if (ts.getModuleInstanceState(node) !== 1 /* Instantiated */) {
                                    return undefined;
                                }
                            // falls through
                            case 255 /* ClassDeclaration */:
                            case 258 /* EnumDeclaration */:
                            case 294 /* EnumMember */:
                            case 201 /* BindingElement */:
                                // span on complete node
                                return textSpan(node);
                            case 246 /* WithStatement */:
                                // span in statement
                                return spanInNode(node.statement);
                            case 163 /* Decorator */:
                                return spanInNodeArray(parent.decorators);
                            case 199 /* ObjectBindingPattern */:
                            case 200 /* ArrayBindingPattern */:
                                return spanInBindingPattern(node);
                            // No breakpoint in interface, type alias
                            case 256 /* InterfaceDeclaration */:
                            case 257 /* TypeAliasDeclaration */:
                                return undefined;
                            // Tokens:
                            case 26 /* SemicolonToken */:
                            case 1 /* EndOfFileToken */:
                                return spanInNodeIfStartsOnSameLine(ts.findPrecedingToken(node.pos, sourceFile));
                            case 27 /* CommaToken */:
                                return spanInPreviousNode(node);
                            case 18 /* OpenBraceToken */:
                                return spanInOpenBraceToken(node);
                            case 19 /* CloseBraceToken */:
                                return spanInCloseBraceToken(node);
                            case 23 /* CloseBracketToken */:
                                return spanInCloseBracketToken(node);
                            case 20 /* OpenParenToken */:
                                return spanInOpenParenToken(node);
                            case 21 /* CloseParenToken */:
                                return spanInCloseParenToken(node);
                            case 58 /* ColonToken */:
                                return spanInColonToken(node);
                            case 31 /* GreaterThanToken */:
                            case 29 /* LessThanToken */:
                                return spanInGreaterThanOrLessThanToken(node);
                            // Keywords:
                            case 115 /* WhileKeyword */:
                                return spanInWhileKeyword(node);
                            case 91 /* ElseKeyword */:
                            case 83 /* CatchKeyword */:
                            case 96 /* FinallyKeyword */:
                                return spanInNextNode(node);
                            case 158 /* OfKeyword */:
                                return spanInOfKeyword(node);
                            default:
                                // Destructuring pattern in destructuring assignment
                                // [a, b, c] of
                                // [a, b, c] = expression
                                if (ts.isArrayLiteralOrObjectLiteralDestructuringPattern(node)) {
                                    return spanInArrayLiteralOrObjectLiteralDestructuringPattern(node);
                                }
                                // Set breakpoint on identifier element of destructuring pattern
                                // `a` or `...c` or `d: x` from
                                // `[a, b, ...c]` or `{ a, b }` or `{ d: x }` from destructuring pattern
                                if ((node.kind === 79 /* Identifier */ ||
                                    node.kind === 223 /* SpreadElement */ ||
                                    node.kind === 291 /* PropertyAssignment */ ||
                                    node.kind === 292 /* ShorthandPropertyAssignment */) &&
                                    ts.isArrayLiteralOrObjectLiteralDestructuringPattern(parent)) {
                                    return textSpan(node);
                                }
                                if (node.kind === 219 /* BinaryExpression */) {
                                    var _a = node, left = _a.left, operatorToken = _a.operatorToken;
                                    // Set breakpoint in destructuring pattern if its destructuring assignment
                                    // [a, b, c] or {a, b, c} of
                                    // [a, b, c] = expression or
                                    // {a, b, c} = expression
                                    if (ts.isArrayLiteralOrObjectLiteralDestructuringPattern(left)) {
                                        return spanInArrayLiteralOrObjectLiteralDestructuringPattern(left);
                                    }
                                    if (operatorToken.kind === 63 /* EqualsToken */ && ts.isArrayLiteralOrObjectLiteralDestructuringPattern(node.parent)) {
                                        // Set breakpoint on assignment expression element of destructuring pattern
                                        // a = expression of
                                        // [a = expression, b, c] = someExpression or
                                        // { a = expression, b, c } = someExpression
                                        return textSpan(node);
                                    }
                                    if (operatorToken.kind === 27 /* CommaToken */) {
                                        return spanInNode(left);
                                    }
                                }
                                if (ts.isExpressionNode(node)) {
                                    switch (parent.kind) {
                                        case 238 /* DoStatement */:
                                            // Set span as if on while keyword
                                            return spanInPreviousNode(node);
                                        case 163 /* Decorator */:
                                            // Set breakpoint on the decorator emit
                                            return spanInNode(node.parent);
                                        case 240 /* ForStatement */:
                                        case 242 /* ForOfStatement */:
                                            return textSpan(node);
                                        case 219 /* BinaryExpression */:
                                            if (node.parent.operatorToken.kind === 27 /* CommaToken */) {
                                                // If this is a comma expression, the breakpoint is possible in this expression
                                                return textSpan(node);
                                            }
                                            break;
                                        case 212 /* ArrowFunction */:
                                            if (node.parent.body === node) {
                                                // If this is body of arrow function, it is allowed to have the breakpoint
                                                return textSpan(node);
                                            }
                                            break;
                                    }
                                }
                                switch (node.parent.kind) {
                                    case 291 /* PropertyAssignment */:
                                        // If this is name of property assignment, set breakpoint in the initializer
                                        if (node.parent.name === node &&
                                            !ts.isArrayLiteralOrObjectLiteralDestructuringPattern(node.parent.parent)) {
                                            return spanInNode(node.parent.initializer);
                                        }
                                        break;
                                    case 209 /* TypeAssertionExpression */:
                                        // Breakpoint in type assertion goes to its operand
                                        if (node.parent.type === node) {
                                            return spanInNextNode(node.parent.type);
                                        }
                                        break;
                                    case 252 /* VariableDeclaration */:
                                    case 162 /* Parameter */: {
                                        // initializer of variable/parameter declaration go to previous node
                                        var _b = node.parent, initializer = _b.initializer, type = _b.type;
                                        if (initializer === node || type === node || ts.isAssignmentOperator(node.kind)) {
                                            return spanInPreviousNode(node);
                                        }
                                        break;
                                    }
                                    case 219 /* BinaryExpression */: {
                                        var left = node.parent.left;
                                        if (ts.isArrayLiteralOrObjectLiteralDestructuringPattern(left) && node !== left) {
                                            // If initializer of destructuring assignment move to previous token
                                            return spanInPreviousNode(node);
                                        }
                                        break;
                                    }
                                    default:
                                        // return type of function go to previous token
                                        if (ts.isFunctionLike(node.parent) && node.parent.type === node) {
                                            return spanInPreviousNode(node);
                                        }
                                }
                                // Default go to parent to set the breakpoint
                                return spanInNode(node.parent);
                        }
                    }
                    function textSpanFromVariableDeclaration(variableDeclaration) {
                        if (ts.isVariableDeclarationList(variableDeclaration.parent) && variableDeclaration.parent.declarations[0] === variableDeclaration) {
                            // First declaration - include let keyword
                            return textSpan(ts.findPrecedingToken(variableDeclaration.pos, sourceFile, variableDeclaration.parent), variableDeclaration);
                        }
                        else {
                            // Span only on this declaration
                            return textSpan(variableDeclaration);
                        }
                    }
                    function spanInVariableDeclaration(variableDeclaration) {
                        // If declaration of for in statement, just set the span in parent
                        if (variableDeclaration.parent.parent.kind === 241 /* ForInStatement */) {
                            return spanInNode(variableDeclaration.parent.parent);
                        }
                        var parent = variableDeclaration.parent;
                        // If this is a destructuring pattern, set breakpoint in binding pattern
                        if (ts.isBindingPattern(variableDeclaration.name)) {
                            return spanInBindingPattern(variableDeclaration.name);
                        }
                        // Breakpoint is possible in variableDeclaration only if there is initialization
                        // or its declaration from 'for of'
                        if (variableDeclaration.initializer ||
                            ts.hasSyntacticModifier(variableDeclaration, 1 /* Export */) ||
                            parent.parent.kind === 242 /* ForOfStatement */) {
                            return textSpanFromVariableDeclaration(variableDeclaration);
                        }
                        if (ts.isVariableDeclarationList(variableDeclaration.parent) &&
                            variableDeclaration.parent.declarations[0] !== variableDeclaration) {
                            // If we cannot set breakpoint on this declaration, set it on previous one
                            // Because the variable declaration may be binding pattern and
                            // we would like to set breakpoint in last binding element if that's the case,
                            // use preceding token instead
                            return spanInNode(ts.findPrecedingToken(variableDeclaration.pos, sourceFile, variableDeclaration.parent));
                        }
                    }
                    function canHaveSpanInParameterDeclaration(parameter) {
                        // Breakpoint is possible on parameter only if it has initializer, is a rest parameter, or has public or private modifier
                        return !!parameter.initializer || parameter.dotDotDotToken !== undefined ||
                            ts.hasSyntacticModifier(parameter, 4 /* Public */ | 8 /* Private */);
                    }
                    function spanInParameterDeclaration(parameter) {
                        if (ts.isBindingPattern(parameter.name)) {
                            // Set breakpoint in binding pattern
                            return spanInBindingPattern(parameter.name);
                        }
                        else if (canHaveSpanInParameterDeclaration(parameter)) {
                            return textSpan(parameter);
                        }
                        else {
                            var functionDeclaration = parameter.parent;
                            var indexOfParameter = functionDeclaration.parameters.indexOf(parameter);
                            ts.Debug.assert(indexOfParameter !== -1);
                            if (indexOfParameter !== 0) {
                                // Not a first parameter, go to previous parameter
                                return spanInParameterDeclaration(functionDeclaration.parameters[indexOfParameter - 1]);
                            }
                            else {
                                // Set breakpoint in the function declaration body
                                return spanInNode(functionDeclaration.body);
                            }
                        }
                    }
                    function canFunctionHaveSpanInWholeDeclaration(functionDeclaration) {
                        return ts.hasSyntacticModifier(functionDeclaration, 1 /* Export */) ||
                            (functionDeclaration.parent.kind === 255 /* ClassDeclaration */ && functionDeclaration.kind !== 169 /* Constructor */);
                    }
                    function spanInFunctionDeclaration(functionDeclaration) {
                        // No breakpoints in the function signature
                        if (!functionDeclaration.body) {
                            return undefined;
                        }
                        if (canFunctionHaveSpanInWholeDeclaration(functionDeclaration)) {
                            // Set the span on whole function declaration
                            return textSpan(functionDeclaration);
                        }
                        // Set span in function body
                        return spanInNode(functionDeclaration.body);
                    }
                    function spanInFunctionBlock(block) {
                        var nodeForSpanInBlock = block.statements.length ? block.statements[0] : block.getLastToken();
                        if (canFunctionHaveSpanInWholeDeclaration(block.parent)) {
                            return spanInNodeIfStartsOnSameLine(block.parent, nodeForSpanInBlock);
                        }
                        return spanInNode(nodeForSpanInBlock);
                    }
                    function spanInBlock(block) {
                        switch (block.parent.kind) {
                            case 259 /* ModuleDeclaration */:
                                if (ts.getModuleInstanceState(block.parent) !== 1 /* Instantiated */) {
                                    return undefined;
                                }
                            // Set on parent if on same line otherwise on first statement
                            // falls through
                            case 239 /* WhileStatement */:
                            case 237 /* IfStatement */:
                            case 241 /* ForInStatement */:
                                return spanInNodeIfStartsOnSameLine(block.parent, block.statements[0]);
                            // Set span on previous token if it starts on same line otherwise on the first statement of the block
                            case 240 /* ForStatement */:
                            case 242 /* ForOfStatement */:
                                return spanInNodeIfStartsOnSameLine(ts.findPrecedingToken(block.pos, sourceFile, block.parent), block.statements[0]);
                        }
                        // Default action is to set on first statement
                        return spanInNode(block.statements[0]);
                    }
                    function spanInInitializerOfForLike(forLikeStatement) {
                        if (forLikeStatement.initializer.kind === 253 /* VariableDeclarationList */) {
                            // Declaration list - set breakpoint in first declaration
                            var variableDeclarationList = forLikeStatement.initializer;
                            if (variableDeclarationList.declarations.length > 0) {
                                return spanInNode(variableDeclarationList.declarations[0]);
                            }
                        }
                        else {
                            // Expression - set breakpoint in it
                            return spanInNode(forLikeStatement.initializer);
                        }
                    }
                    function spanInForStatement(forStatement) {
                        if (forStatement.initializer) {
                            return spanInInitializerOfForLike(forStatement);
                        }
                        if (forStatement.condition) {
                            return textSpan(forStatement.condition);
                        }
                        if (forStatement.incrementor) {
                            return textSpan(forStatement.incrementor);
                        }
                    }
                    function spanInBindingPattern(bindingPattern) {
                        // Set breakpoint in first binding element
                        var firstBindingElement = ts.forEach(bindingPattern.elements, function (element) { return element.kind !== 225 /* OmittedExpression */ ? element : undefined; });
                        if (firstBindingElement) {
                            return spanInNode(firstBindingElement);
                        }
                        // Empty binding pattern of binding element, set breakpoint on binding element
                        if (bindingPattern.parent.kind === 201 /* BindingElement */) {
                            return textSpan(bindingPattern.parent);
                        }
                        // Variable declaration is used as the span
                        return textSpanFromVariableDeclaration(bindingPattern.parent);
                    }
                    function spanInArrayLiteralOrObjectLiteralDestructuringPattern(node) {
                        ts.Debug.assert(node.kind !== 200 /* ArrayBindingPattern */ && node.kind !== 199 /* ObjectBindingPattern */);
                        var elements = node.kind === 202 /* ArrayLiteralExpression */ ? node.elements : node.properties;
                        var firstBindingElement = ts.forEach(elements, function (element) { return element.kind !== 225 /* OmittedExpression */ ? element : undefined; });
                        if (firstBindingElement) {
                            return spanInNode(firstBindingElement);
                        }
                        // Could be ArrayLiteral from destructuring assignment or
                        // just nested element in another destructuring assignment
                        // set breakpoint on assignment when parent is destructuring assignment
                        // Otherwise set breakpoint for this element
                        return textSpan(node.parent.kind === 219 /* BinaryExpression */ ? node.parent : node);
                    }
                    // Tokens:
                    function spanInOpenBraceToken(node) {
                        switch (node.parent.kind) {
                            case 258 /* EnumDeclaration */:
                                var enumDeclaration = node.parent;
                                return spanInNodeIfStartsOnSameLine(ts.findPrecedingToken(node.pos, sourceFile, node.parent), enumDeclaration.members.length ? enumDeclaration.members[0] : enumDeclaration.getLastToken(sourceFile));
                            case 255 /* ClassDeclaration */:
                                var classDeclaration = node.parent;
                                return spanInNodeIfStartsOnSameLine(ts.findPrecedingToken(node.pos, sourceFile, node.parent), classDeclaration.members.length ? classDeclaration.members[0] : classDeclaration.getLastToken(sourceFile));
                            case 261 /* CaseBlock */:
                                return spanInNodeIfStartsOnSameLine(node.parent.parent, node.parent.clauses[0]);
                        }
                        // Default to parent node
                        return spanInNode(node.parent);
                    }
                    function spanInCloseBraceToken(node) {
                        switch (node.parent.kind) {
                            case 260 /* ModuleBlock */:
                                // If this is not an instantiated module block, no bp span
                                if (ts.getModuleInstanceState(node.parent.parent) !== 1 /* Instantiated */) {
                                    return undefined;
                                }
                            // falls through
                            case 258 /* EnumDeclaration */:
                            case 255 /* ClassDeclaration */:
                                // Span on close brace token
                                return textSpan(node);
                            case 233 /* Block */:
                                if (ts.isFunctionBlock(node.parent)) {
                                    // Span on close brace token
                                    return textSpan(node);
                                }
                            // falls through
                            case 290 /* CatchClause */:
                                return spanInNode(ts.lastOrUndefined(node.parent.statements));
                            case 261 /* CaseBlock */:
                                // breakpoint in last statement of the last clause
                                var caseBlock = node.parent;
                                var lastClause = ts.lastOrUndefined(caseBlock.clauses);
                                if (lastClause) {
                                    return spanInNode(ts.lastOrUndefined(lastClause.statements));
                                }
                                return undefined;
                            case 199 /* ObjectBindingPattern */:
                                // Breakpoint in last binding element or binding pattern if it contains no elements
                                var bindingPattern = node.parent;
                                return spanInNode(ts.lastOrUndefined(bindingPattern.elements) || bindingPattern);
                            // Default to parent node
                            default:
                                if (ts.isArrayLiteralOrObjectLiteralDestructuringPattern(node.parent)) {
                                    // Breakpoint in last binding element or binding pattern if it contains no elements
                                    var objectLiteral = node.parent;
                                    return textSpan(ts.lastOrUndefined(objectLiteral.properties) || objectLiteral);
                                }
                                return spanInNode(node.parent);
                        }
                    }
                    function spanInCloseBracketToken(node) {
                        switch (node.parent.kind) {
                            case 200 /* ArrayBindingPattern */:
                                // Breakpoint in last binding element or binding pattern if it contains no elements
                                var bindingPattern = node.parent;
                                return textSpan(ts.lastOrUndefined(bindingPattern.elements) || bindingPattern);
                            default:
                                if (ts.isArrayLiteralOrObjectLiteralDestructuringPattern(node.parent)) {
                                    // Breakpoint in last binding element or binding pattern if it contains no elements
                                    var arrayLiteral = node.parent;
                                    return textSpan(ts.lastOrUndefined(arrayLiteral.elements) || arrayLiteral);
                                }
                                // Default to parent node
                                return spanInNode(node.parent);
                        }
                    }
                    function spanInOpenParenToken(node) {
                        if (node.parent.kind === 238 /* DoStatement */ || // Go to while keyword and do action instead
                            node.parent.kind === 206 /* CallExpression */ ||
                            node.parent.kind === 207 /* NewExpression */) {
                            return spanInPreviousNode(node);
                        }
                        if (node.parent.kind === 210 /* ParenthesizedExpression */) {
                            return spanInNextNode(node);
                        }
                        // Default to parent node
                        return spanInNode(node.parent);
                    }
                    function spanInCloseParenToken(node) {
                        // Is this close paren token of parameter list, set span in previous token
                        switch (node.parent.kind) {
                            case 211 /* FunctionExpression */:
                            case 254 /* FunctionDeclaration */:
                            case 212 /* ArrowFunction */:
                            case 167 /* MethodDeclaration */:
                            case 166 /* MethodSignature */:
                            case 170 /* GetAccessor */:
                            case 171 /* SetAccessor */:
                            case 169 /* Constructor */:
                            case 239 /* WhileStatement */:
                            case 238 /* DoStatement */:
                            case 240 /* ForStatement */:
                            case 242 /* ForOfStatement */:
                            case 206 /* CallExpression */:
                            case 207 /* NewExpression */:
                            case 210 /* ParenthesizedExpression */:
                                return spanInPreviousNode(node);
                            // Default to parent node
                            default:
                                return spanInNode(node.parent);
                        }
                    }
                    function spanInColonToken(node) {
                        // Is this : specifying return annotation of the function declaration
                        if (ts.isFunctionLike(node.parent) ||
                            node.parent.kind === 291 /* PropertyAssignment */ ||
                            node.parent.kind === 162 /* Parameter */) {
                            return spanInPreviousNode(node);
                        }
                        return spanInNode(node.parent);
                    }
                    function spanInGreaterThanOrLessThanToken(node) {
                        if (node.parent.kind === 209 /* TypeAssertionExpression */) {
                            return spanInNextNode(node);
                        }
                        return spanInNode(node.parent);
                    }
                    function spanInWhileKeyword(node) {
                        if (node.parent.kind === 238 /* DoStatement */) {
                            // Set span on while expression
                            return textSpanEndingAtNextToken(node, node.parent.expression);
                        }
                        // Default to parent node
                        return spanInNode(node.parent);
                    }
                    function spanInOfKeyword(node) {
                        if (node.parent.kind === 242 /* ForOfStatement */) {
                            // Set using next token
                            return spanInNextNode(node);
                        }
                        // Default to parent node
                        return spanInNode(node.parent);
                    }
                }
            }
            BreakpointResolver.spanInSourceFileAtLocation = spanInSourceFileAtLocation;
        })(BreakpointResolver = ts.BreakpointResolver || (ts.BreakpointResolver = {}));
    })(ts || (ts = {}));
    var ts;
    (function (ts) {
        /**
         * Transform one or more nodes using the supplied transformers.
         * @param source A single `Node` or an array of `Node` objects.
         * @param transformers An array of `TransformerFactory` callbacks used to process the transformation.
         * @param compilerOptions Optional compiler options.
         */
        function transform(source, transformers, compilerOptions) {
            var diagnostics = [];
            compilerOptions = ts.fixupCompilerOptions(compilerOptions, diagnostics); // TODO: GH#18217
            var nodes = ts.isArray(source) ? source : [source];
            var result = ts.transformNodes(/*resolver*/ undefined, /*emitHost*/ undefined, ts.factory, compilerOptions, nodes, transformers, /*allowDtsFiles*/ true);
            result.diagnostics = ts.concatenate(result.diagnostics, diagnostics);
            return result;
        }
        ts.transform = transform;
    })(ts || (ts = {}));
    //
    // Copyright (c) Microsoft Corporation.  All rights reserved.
    //
    // Licensed under the Apache License, Version 2.0 (the "License");
    // you may not use this file except in compliance with the License.
    // You may obtain a copy of the License at
    //   http://www.apache.org/licenses/LICENSE-2.0
    //
    // Unless required by applicable law or agreed to in writing, software
    // distributed under the License is distributed on an "AS IS" BASIS,
    // WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    // See the License for the specific language governing permissions and
    // limitations under the License.
    //
    /* @internal */
    var debugObjectHost = (function () { return this; })(); // eslint-disable-line prefer-const
    // We need to use 'null' to interface with the managed side.
    /* eslint-disable no-in-operator */
    /* @internal */
    var ts;
    (function (ts) {
        function logInternalError(logger, err) {
            if (logger) {
                logger.log("*INTERNAL ERROR* - Exception in typescript services: " + err.message);
            }
        }
        var ScriptSnapshotShimAdapter = /** @class */ (function () {
            function ScriptSnapshotShimAdapter(scriptSnapshotShim) {
                this.scriptSnapshotShim = scriptSnapshotShim;
            }
            ScriptSnapshotShimAdapter.prototype.getText = function (start, end) {
                return this.scriptSnapshotShim.getText(start, end);
            };
            ScriptSnapshotShimAdapter.prototype.getLength = function () {
                return this.scriptSnapshotShim.getLength();
            };
            ScriptSnapshotShimAdapter.prototype.getChangeRange = function (oldSnapshot) {
                var oldSnapshotShim = oldSnapshot;
                var encoded = this.scriptSnapshotShim.getChangeRange(oldSnapshotShim.scriptSnapshotShim);
                /* eslint-disable no-null/no-null */
                if (encoded === null) {
                    return null; // TODO: GH#18217
                }
                /* eslint-enable no-null/no-null */
                var decoded = JSON.parse(encoded); // TODO: GH#18217
                return ts.createTextChangeRange(ts.createTextSpan(decoded.span.start, decoded.span.length), decoded.newLength);
            };
            ScriptSnapshotShimAdapter.prototype.dispose = function () {
                // if scriptSnapshotShim is a COM object then property check becomes method call with no arguments
                // 'in' does not have this effect
                if ("dispose" in this.scriptSnapshotShim) {
                    this.scriptSnapshotShim.dispose(); // TODO: GH#18217 Can we just use `if (this.scriptSnapshotShim.dispose)`?
                }
            };
            return ScriptSnapshotShimAdapter;
        }());
        var LanguageServiceShimHostAdapter = /** @class */ (function () {
            function LanguageServiceShimHostAdapter(shimHost) {
                var _this = this;
                this.shimHost = shimHost;
                this.loggingEnabled = false;
                this.tracingEnabled = false;
                // if shimHost is a COM object then property check will become method call with no arguments.
                // 'in' does not have this effect.
                if ("getModuleResolutionsForFile" in this.shimHost) {
                    this.resolveModuleNames = function (moduleNames, containingFile) {
                        var resolutionsInFile = JSON.parse(_this.shimHost.getModuleResolutionsForFile(containingFile)); // TODO: GH#18217
                        return ts.map(moduleNames, function (name) {
                            var result = ts.getProperty(resolutionsInFile, name);
                            return result ? { resolvedFileName: result, extension: ts.extensionFromPath(result), isExternalLibraryImport: false } : undefined;
                        });
                    };
                }
                if ("directoryExists" in this.shimHost) {
                    this.directoryExists = function (directoryName) { return _this.shimHost.directoryExists(directoryName); };
                }
                if ("getTypeReferenceDirectiveResolutionsForFile" in this.shimHost) {
                    this.resolveTypeReferenceDirectives = function (typeDirectiveNames, containingFile) {
                        var typeDirectivesForFile = JSON.parse(_this.shimHost.getTypeReferenceDirectiveResolutionsForFile(containingFile)); // TODO: GH#18217
                        return ts.map(typeDirectiveNames, function (name) { return ts.getProperty(typeDirectivesForFile, name); });
                    };
                }
            }
            LanguageServiceShimHostAdapter.prototype.log = function (s) {
                if (this.loggingEnabled) {
                    this.shimHost.log(s);
                }
            };
            LanguageServiceShimHostAdapter.prototype.trace = function (s) {
                if (this.tracingEnabled) {
                    this.shimHost.trace(s);
                }
            };
            LanguageServiceShimHostAdapter.prototype.error = function (s) {
                this.shimHost.error(s);
            };
            LanguageServiceShimHostAdapter.prototype.getProjectVersion = function () {
                if (!this.shimHost.getProjectVersion) {
                    // shimmed host does not support getProjectVersion
                    return undefined; // TODO: GH#18217
                }
                return this.shimHost.getProjectVersion();
            };
            LanguageServiceShimHostAdapter.prototype.getTypeRootsVersion = function () {
                if (!this.shimHost.getTypeRootsVersion) {
                    return 0;
                }
                return this.shimHost.getTypeRootsVersion();
            };
            LanguageServiceShimHostAdapter.prototype.useCaseSensitiveFileNames = function () {
                return this.shimHost.useCaseSensitiveFileNames ? this.shimHost.useCaseSensitiveFileNames() : false;
            };
            LanguageServiceShimHostAdapter.prototype.getCompilationSettings = function () {
                var settingsJson = this.shimHost.getCompilationSettings();
                // eslint-disable-next-line no-null/no-null
                if (settingsJson === null || settingsJson === "") {
                    throw Error("LanguageServiceShimHostAdapter.getCompilationSettings: empty compilationSettings");
                }
                var compilerOptions = JSON.parse(settingsJson);
                // permit language service to handle all files (filtering should be performed on the host side)
                compilerOptions.allowNonTsExtensions = true;
                return compilerOptions;
            };
            LanguageServiceShimHostAdapter.prototype.getScriptFileNames = function () {
                var encoded = this.shimHost.getScriptFileNames();
                return JSON.parse(encoded);
            };
            LanguageServiceShimHostAdapter.prototype.getScriptSnapshot = function (fileName) {
                var scriptSnapshot = this.shimHost.getScriptSnapshot(fileName);
                return scriptSnapshot && new ScriptSnapshotShimAdapter(scriptSnapshot);
            };
            LanguageServiceShimHostAdapter.prototype.getScriptKind = function (fileName) {
                if ("getScriptKind" in this.shimHost) {
                    return this.shimHost.getScriptKind(fileName); // TODO: GH#18217
                }
                else {
                    return 0 /* Unknown */;
                }
            };
            LanguageServiceShimHostAdapter.prototype.getScriptVersion = function (fileName) {
                return this.shimHost.getScriptVersion(fileName);
            };
            LanguageServiceShimHostAdapter.prototype.getLocalizedDiagnosticMessages = function () {
                /* eslint-disable no-null/no-null */
                var diagnosticMessagesJson = this.shimHost.getLocalizedDiagnosticMessages();
                if (diagnosticMessagesJson === null || diagnosticMessagesJson === "") {
                    return null;
                }
                try {
                    return JSON.parse(diagnosticMessagesJson);
                }
                catch (e) {
                    this.log(e.description || "diagnosticMessages.generated.json has invalid JSON format");
                    return null;
                }
                /* eslint-enable no-null/no-null */
            };
            LanguageServiceShimHostAdapter.prototype.getCancellationToken = function () {
                var hostCancellationToken = this.shimHost.getCancellationToken();
                return new ts.ThrottledCancellationToken(hostCancellationToken);
            };
            LanguageServiceShimHostAdapter.prototype.getCurrentDirectory = function () {
                return this.shimHost.getCurrentDirectory();
            };
            LanguageServiceShimHostAdapter.prototype.getDirectories = function (path) {
                return JSON.parse(this.shimHost.getDirectories(path));
            };
            LanguageServiceShimHostAdapter.prototype.getDefaultLibFileName = function (options) {
                return this.shimHost.getDefaultLibFileName(JSON.stringify(options));
            };
            LanguageServiceShimHostAdapter.prototype.readDirectory = function (path, extensions, exclude, include, depth) {
                var pattern = ts.getFileMatcherPatterns(path, exclude, include, this.shimHost.useCaseSensitiveFileNames(), this.shimHost.getCurrentDirectory()); // TODO: GH#18217
                return JSON.parse(this.shimHost.readDirectory(path, JSON.stringify(extensions), JSON.stringify(pattern.basePaths), pattern.excludePattern, pattern.includeFilePattern, pattern.includeDirectoryPattern, depth));
            };
            LanguageServiceShimHostAdapter.prototype.readFile = function (path, encoding) {
                return this.shimHost.readFile(path, encoding);
            };
            LanguageServiceShimHostAdapter.prototype.fileExists = function (path) {
                return this.shimHost.fileExists(path);
            };
            return LanguageServiceShimHostAdapter;
        }());
        ts.LanguageServiceShimHostAdapter = LanguageServiceShimHostAdapter;
        var CoreServicesShimHostAdapter = /** @class */ (function () {
            function CoreServicesShimHostAdapter(shimHost) {
                var _this = this;
                this.shimHost = shimHost;
                this.useCaseSensitiveFileNames = this.shimHost.useCaseSensitiveFileNames ? this.shimHost.useCaseSensitiveFileNames() : false;
                if ("directoryExists" in this.shimHost) {
                    this.directoryExists = function (directoryName) { return _this.shimHost.directoryExists(directoryName); };
                }
                else {
                    this.directoryExists = undefined; // TODO: GH#18217
                }
                if ("realpath" in this.shimHost) {
                    this.realpath = function (path) { return _this.shimHost.realpath(path); }; // TODO: GH#18217
                }
                else {
                    this.realpath = undefined; // TODO: GH#18217
                }
            }
            CoreServicesShimHostAdapter.prototype.readDirectory = function (rootDir, extensions, exclude, include, depth) {
                var pattern = ts.getFileMatcherPatterns(rootDir, exclude, include, this.shimHost.useCaseSensitiveFileNames(), this.shimHost.getCurrentDirectory()); // TODO: GH#18217
                return JSON.parse(this.shimHost.readDirectory(rootDir, JSON.stringify(extensions), JSON.stringify(pattern.basePaths), pattern.excludePattern, pattern.includeFilePattern, pattern.includeDirectoryPattern, depth));
            };
            CoreServicesShimHostAdapter.prototype.fileExists = function (fileName) {
                return this.shimHost.fileExists(fileName);
            };
            CoreServicesShimHostAdapter.prototype.readFile = function (fileName) {
                return this.shimHost.readFile(fileName);
            };
            CoreServicesShimHostAdapter.prototype.getDirectories = function (path) {
                return JSON.parse(this.shimHost.getDirectories(path));
            };
            return CoreServicesShimHostAdapter;
        }());
        ts.CoreServicesShimHostAdapter = CoreServicesShimHostAdapter;
        function simpleForwardCall(logger, actionDescription, action, logPerformance) {
            var start;
            if (logPerformance) {
                logger.log(actionDescription);
                start = ts.timestamp();
            }
            var result = action();
            if (logPerformance) {
                var end = ts.timestamp();
                logger.log(actionDescription + " completed in " + (end - start) + " msec");
                if (ts.isString(result)) {
                    var str = result;
                    if (str.length > 128) {
                        str = str.substring(0, 128) + "...";
                    }
                    logger.log("  result.length=" + str.length + ", result='" + JSON.stringify(str) + "'");
                }
            }
            return result;
        }
        function forwardJSONCall(logger, actionDescription, action, logPerformance) {
            return forwardCall(logger, actionDescription, /*returnJson*/ true, action, logPerformance);
        }
        function forwardCall(logger, actionDescription, returnJson, action, logPerformance) {
            try {
                var result = simpleForwardCall(logger, actionDescription, action, logPerformance);
                return returnJson ? JSON.stringify({ result: result }) : result;
            }
            catch (err) {
                if (err instanceof ts.OperationCanceledException) {
                    return JSON.stringify({ canceled: true });
                }
                logInternalError(logger, err);
                err.description = actionDescription;
                return JSON.stringify({ error: err });
            }
        }
        var ShimBase = /** @class */ (function () {
            function ShimBase(factory) {
                this.factory = factory;
                factory.registerShim(this);
            }
            ShimBase.prototype.dispose = function (_dummy) {
                this.factory.unregisterShim(this);
            };
            return ShimBase;
        }());
        function realizeDiagnostics(diagnostics, newLine) {
            return diagnostics.map(function (d) { return realizeDiagnostic(d, newLine); });
        }
        ts.realizeDiagnostics = realizeDiagnostics;
        function realizeDiagnostic(diagnostic, newLine) {
            return {
                message: ts.flattenDiagnosticMessageText(diagnostic.messageText, newLine),
                start: diagnostic.start,
                length: diagnostic.length,
                category: ts.diagnosticCategoryName(diagnostic),
                code: diagnostic.code,
                reportsUnnecessary: diagnostic.reportsUnnecessary,
                reportsDeprecated: diagnostic.reportsDeprecated
            };
        }
        var LanguageServiceShimObject = /** @class */ (function (_super) {
            __extends(LanguageServiceShimObject, _super);
            function LanguageServiceShimObject(factory, host, languageService) {
                var _this = _super.call(this, factory) || this;
                _this.host = host;
                _this.languageService = languageService;
                _this.logPerformance = false;
                _this.logger = _this.host;
                return _this;
            }
            LanguageServiceShimObject.prototype.forwardJSONCall = function (actionDescription, action) {
                return forwardJSONCall(this.logger, actionDescription, action, this.logPerformance);
            };
            /// DISPOSE
            /**
             * Ensure (almost) deterministic release of internal Javascript resources when
             * some external native objects holds onto us (e.g. Com/Interop).
             */
            LanguageServiceShimObject.prototype.dispose = function (dummy) {
                this.logger.log("dispose()");
                this.languageService.dispose();
                this.languageService = null; // eslint-disable-line no-null/no-null
                // force a GC
                if (debugObjectHost && debugObjectHost.CollectGarbage) {
                    debugObjectHost.CollectGarbage();
                    this.logger.log("CollectGarbage()");
                }
                this.logger = null; // eslint-disable-line no-null/no-null
                _super.prototype.dispose.call(this, dummy);
            };
            /// REFRESH
            /**
             * Update the list of scripts known to the compiler
             */
            LanguageServiceShimObject.prototype.refresh = function (throwOnError) {
                this.forwardJSONCall("refresh(" + throwOnError + ")", function () { return null; } // eslint-disable-line no-null/no-null
                );
            };
            LanguageServiceShimObject.prototype.cleanupSemanticCache = function () {
                var _this = this;
                this.forwardJSONCall("cleanupSemanticCache()", function () {
                    _this.languageService.cleanupSemanticCache();
                    return null; // eslint-disable-line no-null/no-null
                });
            };
            LanguageServiceShimObject.prototype.realizeDiagnostics = function (diagnostics) {
                var newLine = ts.getNewLineOrDefaultFromHost(this.host);
                return realizeDiagnostics(diagnostics, newLine);
            };
            LanguageServiceShimObject.prototype.getSyntacticClassifications = function (fileName, start, length) {
                var _this = this;
                return this.forwardJSONCall("getSyntacticClassifications('" + fileName + "', " + start + ", " + length + ")", function () { return _this.languageService.getSyntacticClassifications(fileName, ts.createTextSpan(start, length)); });
            };
            LanguageServiceShimObject.prototype.getSemanticClassifications = function (fileName, start, length) {
                var _this = this;
                return this.forwardJSONCall("getSemanticClassifications('" + fileName + "', " + start + ", " + length + ")", function () { return _this.languageService.getSemanticClassifications(fileName, ts.createTextSpan(start, length)); });
            };
            LanguageServiceShimObject.prototype.getEncodedSyntacticClassifications = function (fileName, start, length) {
                var _this = this;
                return this.forwardJSONCall("getEncodedSyntacticClassifications('" + fileName + "', " + start + ", " + length + ")", 
                // directly serialize the spans out to a string.  This is much faster to decode
                // on the managed side versus a full JSON array.
                function () { return convertClassifications(_this.languageService.getEncodedSyntacticClassifications(fileName, ts.createTextSpan(start, length))); });
            };
            LanguageServiceShimObject.prototype.getEncodedSemanticClassifications = function (fileName, start, length) {
                var _this = this;
                return this.forwardJSONCall("getEncodedSemanticClassifications('" + fileName + "', " + start + ", " + length + ")", 
                // directly serialize the spans out to a string.  This is much faster to decode
                // on the managed side versus a full JSON array.
                function () { return convertClassifications(_this.languageService.getEncodedSemanticClassifications(fileName, ts.createTextSpan(start, length))); });
            };
            LanguageServiceShimObject.prototype.getSyntacticDiagnostics = function (fileName) {
                var _this = this;
                return this.forwardJSONCall("getSyntacticDiagnostics('" + fileName + "')", function () {
                    var diagnostics = _this.languageService.getSyntacticDiagnostics(fileName);
                    return _this.realizeDiagnostics(diagnostics);
                });
            };
            LanguageServiceShimObject.prototype.getSemanticDiagnostics = function (fileName) {
                var _this = this;
                return this.forwardJSONCall("getSemanticDiagnostics('" + fileName + "')", function () {
                    var diagnostics = _this.languageService.getSemanticDiagnostics(fileName);
                    return _this.realizeDiagnostics(diagnostics);
                });
            };
            LanguageServiceShimObject.prototype.getSuggestionDiagnostics = function (fileName) {
                var _this = this;
                return this.forwardJSONCall("getSuggestionDiagnostics('" + fileName + "')", function () { return _this.realizeDiagnostics(_this.languageService.getSuggestionDiagnostics(fileName)); });
            };
            LanguageServiceShimObject.prototype.getCompilerOptionsDiagnostics = function () {
                var _this = this;
                return this.forwardJSONCall("getCompilerOptionsDiagnostics()", function () {
                    var diagnostics = _this.languageService.getCompilerOptionsDiagnostics();
                    return _this.realizeDiagnostics(diagnostics);
                });
            };
            /// QUICKINFO
            /**
             * Computes a string representation of the type at the requested position
             * in the active file.
             */
            LanguageServiceShimObject.prototype.getQuickInfoAtPosition = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getQuickInfoAtPosition('" + fileName + "', " + position + ")", function () { return _this.languageService.getQuickInfoAtPosition(fileName, position); });
            };
            /// NAMEORDOTTEDNAMESPAN
            /**
             * Computes span information of the name or dotted name at the requested position
             * in the active file.
             */
            LanguageServiceShimObject.prototype.getNameOrDottedNameSpan = function (fileName, startPos, endPos) {
                var _this = this;
                return this.forwardJSONCall("getNameOrDottedNameSpan('" + fileName + "', " + startPos + ", " + endPos + ")", function () { return _this.languageService.getNameOrDottedNameSpan(fileName, startPos, endPos); });
            };
            /**
             * STATEMENTSPAN
             * Computes span information of statement at the requested position in the active file.
             */
            LanguageServiceShimObject.prototype.getBreakpointStatementAtPosition = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getBreakpointStatementAtPosition('" + fileName + "', " + position + ")", function () { return _this.languageService.getBreakpointStatementAtPosition(fileName, position); });
            };
            /// SIGNATUREHELP
            LanguageServiceShimObject.prototype.getSignatureHelpItems = function (fileName, position, options) {
                var _this = this;
                return this.forwardJSONCall("getSignatureHelpItems('" + fileName + "', " + position + ")", function () { return _this.languageService.getSignatureHelpItems(fileName, position, options); });
            };
            /// GOTO DEFINITION
            /**
             * Computes the definition location and file for the symbol
             * at the requested position.
             */
            LanguageServiceShimObject.prototype.getDefinitionAtPosition = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getDefinitionAtPosition('" + fileName + "', " + position + ")", function () { return _this.languageService.getDefinitionAtPosition(fileName, position); });
            };
            /**
             * Computes the definition location and file for the symbol
             * at the requested position.
             */
            LanguageServiceShimObject.prototype.getDefinitionAndBoundSpan = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getDefinitionAndBoundSpan('" + fileName + "', " + position + ")", function () { return _this.languageService.getDefinitionAndBoundSpan(fileName, position); });
            };
            /// GOTO Type
            /**
             * Computes the definition location of the type of the symbol
             * at the requested position.
             */
            LanguageServiceShimObject.prototype.getTypeDefinitionAtPosition = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getTypeDefinitionAtPosition('" + fileName + "', " + position + ")", function () { return _this.languageService.getTypeDefinitionAtPosition(fileName, position); });
            };
            /// GOTO Implementation
            /**
             * Computes the implementation location of the symbol
             * at the requested position.
             */
            LanguageServiceShimObject.prototype.getImplementationAtPosition = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getImplementationAtPosition('" + fileName + "', " + position + ")", function () { return _this.languageService.getImplementationAtPosition(fileName, position); });
            };
            LanguageServiceShimObject.prototype.getRenameInfo = function (fileName, position, options) {
                var _this = this;
                return this.forwardJSONCall("getRenameInfo('" + fileName + "', " + position + ")", function () { return _this.languageService.getRenameInfo(fileName, position, options); });
            };
            LanguageServiceShimObject.prototype.getSmartSelectionRange = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getSmartSelectionRange('" + fileName + "', " + position + ")", function () { return _this.languageService.getSmartSelectionRange(fileName, position); });
            };
            LanguageServiceShimObject.prototype.findRenameLocations = function (fileName, position, findInStrings, findInComments, providePrefixAndSuffixTextForRename) {
                var _this = this;
                return this.forwardJSONCall("findRenameLocations('" + fileName + "', " + position + ", " + findInStrings + ", " + findInComments + ", " + providePrefixAndSuffixTextForRename + ")", function () { return _this.languageService.findRenameLocations(fileName, position, findInStrings, findInComments, providePrefixAndSuffixTextForRename); });
            };
            /// GET BRACE MATCHING
            LanguageServiceShimObject.prototype.getBraceMatchingAtPosition = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getBraceMatchingAtPosition('" + fileName + "', " + position + ")", function () { return _this.languageService.getBraceMatchingAtPosition(fileName, position); });
            };
            LanguageServiceShimObject.prototype.isValidBraceCompletionAtPosition = function (fileName, position, openingBrace) {
                var _this = this;
                return this.forwardJSONCall("isValidBraceCompletionAtPosition('" + fileName + "', " + position + ", " + openingBrace + ")", function () { return _this.languageService.isValidBraceCompletionAtPosition(fileName, position, openingBrace); });
            };
            LanguageServiceShimObject.prototype.getSpanOfEnclosingComment = function (fileName, position, onlyMultiLine) {
                var _this = this;
                return this.forwardJSONCall("getSpanOfEnclosingComment('" + fileName + "', " + position + ")", function () { return _this.languageService.getSpanOfEnclosingComment(fileName, position, onlyMultiLine); });
            };
            /// GET SMART INDENT
            LanguageServiceShimObject.prototype.getIndentationAtPosition = function (fileName, position, options /*Services.EditorOptions*/) {
                var _this = this;
                return this.forwardJSONCall("getIndentationAtPosition('" + fileName + "', " + position + ")", function () {
                    var localOptions = JSON.parse(options);
                    return _this.languageService.getIndentationAtPosition(fileName, position, localOptions);
                });
            };
            /// GET REFERENCES
            LanguageServiceShimObject.prototype.getReferencesAtPosition = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getReferencesAtPosition('" + fileName + "', " + position + ")", function () { return _this.languageService.getReferencesAtPosition(fileName, position); });
            };
            LanguageServiceShimObject.prototype.findReferences = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("findReferences('" + fileName + "', " + position + ")", function () { return _this.languageService.findReferences(fileName, position); });
            };
            LanguageServiceShimObject.prototype.getFileReferences = function (fileName) {
                var _this = this;
                return this.forwardJSONCall("getFileReferences('" + fileName + ")", function () { return _this.languageService.getFileReferences(fileName); });
            };
            LanguageServiceShimObject.prototype.getOccurrencesAtPosition = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("getOccurrencesAtPosition('" + fileName + "', " + position + ")", function () { return _this.languageService.getOccurrencesAtPosition(fileName, position); });
            };
            LanguageServiceShimObject.prototype.getDocumentHighlights = function (fileName, position, filesToSearch) {
                var _this = this;
                return this.forwardJSONCall("getDocumentHighlights('" + fileName + "', " + position + ")", function () {
                    var results = _this.languageService.getDocumentHighlights(fileName, position, JSON.parse(filesToSearch));
                    // workaround for VS document highlighting issue - keep only items from the initial file
                    var normalizedName = ts.toFileNameLowerCase(ts.normalizeSlashes(fileName));
                    return ts.filter(results, function (r) { return ts.toFileNameLowerCase(ts.normalizeSlashes(r.fileName)) === normalizedName; });
                });
            };
            /// COMPLETION LISTS
            /**
             * Get a string based representation of the completions
             * to provide at the given source position and providing a member completion
             * list if requested.
             */
            LanguageServiceShimObject.prototype.getCompletionsAtPosition = function (fileName, position, preferences) {
                var _this = this;
                return this.forwardJSONCall("getCompletionsAtPosition('" + fileName + "', " + position + ", " + preferences + ")", function () { return _this.languageService.getCompletionsAtPosition(fileName, position, preferences); });
            };
            /** Get a string based representation of a completion list entry details */
            LanguageServiceShimObject.prototype.getCompletionEntryDetails = function (fileName, position, entryName, formatOptions, source, preferences, data) {
                var _this = this;
                return this.forwardJSONCall("getCompletionEntryDetails('" + fileName + "', " + position + ", '" + entryName + "')", function () {
                    var localOptions = formatOptions === undefined ? undefined : JSON.parse(formatOptions);
                    return _this.languageService.getCompletionEntryDetails(fileName, position, entryName, localOptions, source, preferences, data);
                });
            };
            LanguageServiceShimObject.prototype.getFormattingEditsForRange = function (fileName, start, end, options /*Services.FormatCodeOptions*/) {
                var _this = this;
                return this.forwardJSONCall("getFormattingEditsForRange('" + fileName + "', " + start + ", " + end + ")", function () {
                    var localOptions = JSON.parse(options);
                    return _this.languageService.getFormattingEditsForRange(fileName, start, end, localOptions);
                });
            };
            LanguageServiceShimObject.prototype.getFormattingEditsForDocument = function (fileName, options /*Services.FormatCodeOptions*/) {
                var _this = this;
                return this.forwardJSONCall("getFormattingEditsForDocument('" + fileName + "')", function () {
                    var localOptions = JSON.parse(options);
                    return _this.languageService.getFormattingEditsForDocument(fileName, localOptions);
                });
            };
            LanguageServiceShimObject.prototype.getFormattingEditsAfterKeystroke = function (fileName, position, key, options /*Services.FormatCodeOptions*/) {
                var _this = this;
                return this.forwardJSONCall("getFormattingEditsAfterKeystroke('" + fileName + "', " + position + ", '" + key + "')", function () {
                    var localOptions = JSON.parse(options);
                    return _this.languageService.getFormattingEditsAfterKeystroke(fileName, position, key, localOptions);
                });
            };
            LanguageServiceShimObject.prototype.getDocCommentTemplateAtPosition = function (fileName, position, options) {
                var _this = this;
                return this.forwardJSONCall("getDocCommentTemplateAtPosition('" + fileName + "', " + position + ")", function () { return _this.languageService.getDocCommentTemplateAtPosition(fileName, position, options); });
            };
            /// NAVIGATE TO
            /** Return a list of symbols that are interesting to navigate to */
            LanguageServiceShimObject.prototype.getNavigateToItems = function (searchValue, maxResultCount, fileName) {
                var _this = this;
                return this.forwardJSONCall("getNavigateToItems('" + searchValue + "', " + maxResultCount + ", " + fileName + ")", function () { return _this.languageService.getNavigateToItems(searchValue, maxResultCount, fileName); });
            };
            LanguageServiceShimObject.prototype.getNavigationBarItems = function (fileName) {
                var _this = this;
                return this.forwardJSONCall("getNavigationBarItems('" + fileName + "')", function () { return _this.languageService.getNavigationBarItems(fileName); });
            };
            LanguageServiceShimObject.prototype.getNavigationTree = function (fileName) {
                var _this = this;
                return this.forwardJSONCall("getNavigationTree('" + fileName + "')", function () { return _this.languageService.getNavigationTree(fileName); });
            };
            LanguageServiceShimObject.prototype.getOutliningSpans = function (fileName) {
                var _this = this;
                return this.forwardJSONCall("getOutliningSpans('" + fileName + "')", function () { return _this.languageService.getOutliningSpans(fileName); });
            };
            LanguageServiceShimObject.prototype.getTodoComments = function (fileName, descriptors) {
                var _this = this;
                return this.forwardJSONCall("getTodoComments('" + fileName + "')", function () { return _this.languageService.getTodoComments(fileName, JSON.parse(descriptors)); });
            };
            /// CALL HIERARCHY
            LanguageServiceShimObject.prototype.prepareCallHierarchy = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("prepareCallHierarchy('" + fileName + "', " + position + ")", function () { return _this.languageService.prepareCallHierarchy(fileName, position); });
            };
            LanguageServiceShimObject.prototype.provideCallHierarchyIncomingCalls = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("provideCallHierarchyIncomingCalls('" + fileName + "', " + position + ")", function () { return _this.languageService.provideCallHierarchyIncomingCalls(fileName, position); });
            };
            LanguageServiceShimObject.prototype.provideCallHierarchyOutgoingCalls = function (fileName, position) {
                var _this = this;
                return this.forwardJSONCall("provideCallHierarchyOutgoingCalls('" + fileName + "', " + position + ")", function () { return _this.languageService.provideCallHierarchyOutgoingCalls(fileName, position); });
            };
            LanguageServiceShimObject.prototype.provideInlayHints = function (fileName, span, preference) {
                var _this = this;
                return this.forwardJSONCall("provideInlayHints('" + fileName + "', '" + JSON.stringify(span) + "', " + JSON.stringify(preference) + ")", function () { return _this.languageService.provideInlayHints(fileName, span, preference); });
            };
            /// Emit
            LanguageServiceShimObject.prototype.getEmitOutput = function (fileName) {
                var _this = this;
                return this.forwardJSONCall("getEmitOutput('" + fileName + "')", function () {
                    var _a = _this.languageService.getEmitOutput(fileName), diagnostics = _a.diagnostics, rest = __rest(_a, ["diagnostics"]);
                    return __assign(__assign({}, rest), { diagnostics: _this.realizeDiagnostics(diagnostics) });
                });
            };
            LanguageServiceShimObject.prototype.getEmitOutputObject = function (fileName) {
                var _this = this;
                return forwardCall(this.logger, "getEmitOutput('" + fileName + "')", 
                /*returnJson*/ false, function () { return _this.languageService.getEmitOutput(fileName); }, this.logPerformance);
            };
            LanguageServiceShimObject.prototype.toggleLineComment = function (fileName, textRange) {
                var _this = this;
                return this.forwardJSONCall("toggleLineComment('" + fileName + "', '" + JSON.stringify(textRange) + "')", function () { return _this.languageService.toggleLineComment(fileName, textRange); });
            };
            LanguageServiceShimObject.prototype.toggleMultilineComment = function (fileName, textRange) {
                var _this = this;
                return this.forwardJSONCall("toggleMultilineComment('" + fileName + "', '" + JSON.stringify(textRange) + "')", function () { return _this.languageService.toggleMultilineComment(fileName, textRange); });
            };
            LanguageServiceShimObject.prototype.commentSelection = function (fileName, textRange) {
                var _this = this;
                return this.forwardJSONCall("commentSelection('" + fileName + "', '" + JSON.stringify(textRange) + "')", function () { return _this.languageService.commentSelection(fileName, textRange); });
            };
            LanguageServiceShimObject.prototype.uncommentSelection = function (fileName, textRange) {
                var _this = this;
                return this.forwardJSONCall("uncommentSelection('" + fileName + "', '" + JSON.stringify(textRange) + "')", function () { return _this.languageService.uncommentSelection(fileName, textRange); });
            };
            return LanguageServiceShimObject;
        }(ShimBase));
        function convertClassifications(classifications) {
            return { spans: classifications.spans.join(","), endOfLineState: classifications.endOfLineState };
        }
        var ClassifierShimObject = /** @class */ (function (_super) {
            __extends(ClassifierShimObject, _super);
            function ClassifierShimObject(factory, logger) {
                var _this = _super.call(this, factory) || this;
                _this.logger = logger;
                _this.logPerformance = false;
                _this.classifier = ts.createClassifier();
                return _this;
            }
            ClassifierShimObject.prototype.getEncodedLexicalClassifications = function (text, lexState, syntacticClassifierAbsent) {
                var _this = this;
                if (syntacticClassifierAbsent === void 0) {
                    syntacticClassifierAbsent = false;
                }
                return forwardJSONCall(this.logger, "getEncodedLexicalClassifications", function () { return convertClassifications(_this.classifier.getEncodedLexicalClassifications(text, lexState, syntacticClassifierAbsent)); }, this.logPerformance);
            };
            /// COLORIZATION
            ClassifierShimObject.prototype.getClassificationsForLine = function (text, lexState, classifyKeywordsInGenerics) {
                if (classifyKeywordsInGenerics === void 0) {
                    classifyKeywordsInGenerics = false;
                }
                var classification = this.classifier.getClassificationsForLine(text, lexState, classifyKeywordsInGenerics);
                var result = "";
                for (var _i = 0, _a = classification.entries; _i < _a.length; _i++) {
                    var item = _a[_i];
                    result += item.length + "\n";
                    result += item.classification + "\n";
                }
                result += classification.finalLexState;
                return result;
            };
            return ClassifierShimObject;
        }(ShimBase));
        var CoreServicesShimObject = /** @class */ (function (_super) {
            __extends(CoreServicesShimObject, _super);
            function CoreServicesShimObject(factory, logger, host) {
                var _this = _super.call(this, factory) || this;
                _this.logger = logger;
                _this.host = host;
                _this.logPerformance = false;
                return _this;
            }
            CoreServicesShimObject.prototype.forwardJSONCall = function (actionDescription, action) {
                return forwardJSONCall(this.logger, actionDescription, action, this.logPerformance);
            };
            CoreServicesShimObject.prototype.resolveModuleName = function (fileName, moduleName, compilerOptionsJson) {
                var _this = this;
                return this.forwardJSONCall("resolveModuleName('" + fileName + "')", function () {
                    var compilerOptions = JSON.parse(compilerOptionsJson);
                    var result = ts.resolveModuleName(moduleName, ts.normalizeSlashes(fileName), compilerOptions, _this.host);
                    var resolvedFileName = result.resolvedModule ? result.resolvedModule.resolvedFileName : undefined;
                    if (result.resolvedModule && result.resolvedModule.extension !== ".ts" /* Ts */ && result.resolvedModule.extension !== ".tsx" /* Tsx */ && result.resolvedModule.extension !== ".d.ts" /* Dts */) {
                        resolvedFileName = undefined;
                    }
                    return {
                        resolvedFileName: resolvedFileName,
                        failedLookupLocations: result.failedLookupLocations
                    };
                });
            };
            CoreServicesShimObject.prototype.resolveTypeReferenceDirective = function (fileName, typeReferenceDirective, compilerOptionsJson) {
                var _this = this;
                return this.forwardJSONCall("resolveTypeReferenceDirective(" + fileName + ")", function () {
                    var compilerOptions = JSON.parse(compilerOptionsJson);
                    var result = ts.resolveTypeReferenceDirective(typeReferenceDirective, ts.normalizeSlashes(fileName), compilerOptions, _this.host);
                    return {
                        resolvedFileName: result.resolvedTypeReferenceDirective ? result.resolvedTypeReferenceDirective.resolvedFileName : undefined,
                        primary: result.resolvedTypeReferenceDirective ? result.resolvedTypeReferenceDirective.primary : true,
                        failedLookupLocations: result.failedLookupLocations
                    };
                });
            };
            CoreServicesShimObject.prototype.getPreProcessedFileInfo = function (fileName, sourceTextSnapshot) {
                var _this = this;
                return this.forwardJSONCall("getPreProcessedFileInfo('" + fileName + "')", function () {
                    // for now treat files as JavaScript
                    var result = ts.preProcessFile(ts.getSnapshotText(sourceTextSnapshot), /* readImportFiles */ true, /* detectJavaScriptImports */ true);
                    return {
                        referencedFiles: _this.convertFileReferences(result.referencedFiles),
                        importedFiles: _this.convertFileReferences(result.importedFiles),
                        ambientExternalModules: result.ambientExternalModules,
                        isLibFile: result.isLibFile,
                        typeReferenceDirectives: _this.convertFileReferences(result.typeReferenceDirectives),
                        libReferenceDirectives: _this.convertFileReferences(result.libReferenceDirectives)
                    };
                });
            };
            CoreServicesShimObject.prototype.getAutomaticTypeDirectiveNames = function (compilerOptionsJson) {
                var _this = this;
                return this.forwardJSONCall("getAutomaticTypeDirectiveNames('" + compilerOptionsJson + "')", function () {
                    var compilerOptions = JSON.parse(compilerOptionsJson);
                    return ts.getAutomaticTypeDirectiveNames(compilerOptions, _this.host);
                });
            };
            CoreServicesShimObject.prototype.convertFileReferences = function (refs) {
                if (!refs) {
                    return undefined;
                }
                var result = [];
                for (var _i = 0, refs_1 = refs; _i < refs_1.length; _i++) {
                    var ref = refs_1[_i];
                    result.push({
                        path: ts.normalizeSlashes(ref.fileName),
                        position: ref.pos,
                        length: ref.end - ref.pos
                    });
                }
                return result;
            };
            CoreServicesShimObject.prototype.getTSConfigFileInfo = function (fileName, sourceTextSnapshot) {
                var _this = this;
                return this.forwardJSONCall("getTSConfigFileInfo('" + fileName + "')", function () {
                    var result = ts.parseJsonText(fileName, ts.getSnapshotText(sourceTextSnapshot));
                    var normalizedFileName = ts.normalizeSlashes(fileName);
                    var configFile = ts.parseJsonSourceFileConfigFileContent(result, _this.host, ts.getDirectoryPath(normalizedFileName), /*existingOptions*/ {}, normalizedFileName);
                    return {
                        options: configFile.options,
                        typeAcquisition: configFile.typeAcquisition,
                        files: configFile.fileNames,
                        raw: configFile.raw,
                        errors: realizeDiagnostics(__spreadArray(__spreadArray([], result.parseDiagnostics, true), configFile.errors, true), "\r\n")
                    };
                });
            };
            CoreServicesShimObject.prototype.getDefaultCompilationSettings = function () {
                return this.forwardJSONCall("getDefaultCompilationSettings()", function () { return ts.getDefaultCompilerOptions(); });
            };
            CoreServicesShimObject.prototype.discoverTypings = function (discoverTypingsJson) {
                var _this = this;
                var getCanonicalFileName = ts.createGetCanonicalFileName(/*useCaseSensitivefileNames:*/ false);
                return this.forwardJSONCall("discoverTypings()", function () {
                    var info = JSON.parse(discoverTypingsJson);
                    if (_this.safeList === undefined) {
                        _this.safeList = ts.JsTyping.loadSafeList(_this.host, ts.toPath(info.safeListPath, info.safeListPath, getCanonicalFileName));
                    }
                    return ts.JsTyping.discoverTypings(_this.host, function (msg) { return _this.logger.log(msg); }, info.fileNames, ts.toPath(info.projectRootPath, info.projectRootPath, getCanonicalFileName), _this.safeList, info.packageNameToTypingLocation, info.typeAcquisition, info.unresolvedImports, info.typesRegistry);
                });
            };
            return CoreServicesShimObject;
        }(ShimBase));
        var TypeScriptServicesFactory = /** @class */ (function () {
            function TypeScriptServicesFactory() {
                this._shims = [];
            }
            /*
             * Returns script API version.
             */
            TypeScriptServicesFactory.prototype.getServicesVersion = function () {
                return ts.servicesVersion;
            };
            TypeScriptServicesFactory.prototype.createLanguageServiceShim = function (host) {
                try {
                    if (this.documentRegistry === undefined) {
                        this.documentRegistry = ts.createDocumentRegistry(host.useCaseSensitiveFileNames && host.useCaseSensitiveFileNames(), host.getCurrentDirectory());
                    }
                    var hostAdapter = new LanguageServiceShimHostAdapter(host);
                    var languageService = ts.createLanguageService(hostAdapter, this.documentRegistry, /*syntaxOnly*/ false);
                    return new LanguageServiceShimObject(this, host, languageService);
                }
                catch (err) {
                    logInternalError(host, err);
                    throw err;
                }
            };
            TypeScriptServicesFactory.prototype.createClassifierShim = function (logger) {
                try {
                    return new ClassifierShimObject(this, logger);
                }
                catch (err) {
                    logInternalError(logger, err);
                    throw err;
                }
            };
            TypeScriptServicesFactory.prototype.createCoreServicesShim = function (host) {
                try {
                    var adapter = new CoreServicesShimHostAdapter(host);
                    return new CoreServicesShimObject(this, host, adapter);
                }
                catch (err) {
                    logInternalError(host, err);
                    throw err;
                }
            };
            TypeScriptServicesFactory.prototype.close = function () {
                // Forget all the registered shims
                ts.clear(this._shims);
                this.documentRegistry = undefined;
            };
            TypeScriptServicesFactory.prototype.registerShim = function (shim) {
                this._shims.push(shim);
            };
            TypeScriptServicesFactory.prototype.unregisterShim = function (shim) {
                for (var i = 0; i < this._shims.length; i++) {
                    if (this._shims[i] === shim) {
                        delete this._shims[i];
                        return;
                    }
                }
                throw new Error("Invalid operation");
            };
            return TypeScriptServicesFactory;
        }());
        ts.TypeScriptServicesFactory = TypeScriptServicesFactory;
    })(ts || (ts = {}));
    /* eslint-enable no-in-operator */
    // We polyfill `globalThis` here so re can reliably patch the global scope
    // in the contexts we want to in the same way across script and module formats
    /* eslint-enable no-var */
    ((function () {
        if (typeof globalThis === "object")
            return;
        try {
            Object.defineProperty(Object.prototype, "__magic__", {
                get: function () {
                    return this;
                },
                configurable: true
            });
            //@ts-ignore
            __magic__.globalThis = __magic__;
            // The previous line should have made `globalThis` globally
            // available, but it fails in Internet Explorer 10 and older.
            // Detect this failure and fall back.
            if (typeof globalThis === "undefined") {
                // Assume `window` exists.
                //@ts-ignore
                window.globalThis = window;
            }
            //@ts-ignore
            delete Object.prototype.__magic__;
        }
        catch (error) {
            // In IE8, Object.defineProperty only works on DOM objects.
            // If we hit this code path, assume `window` exists.
            //@ts-ignore
            window.globalThis = window;
        }
    })());
    // #endregion The polyfill ends here.
    // if `process` is undefined, we're probably not running in node - patch legacy members onto the global scope
    // @ts-ignore
    if (typeof process === "undefined" || process.browser) {
        /// TODO: this is used by VS, clean this up on both sides of the interface
        //@ts-ignore
        globalThis.TypeScript = globalThis.TypeScript || {};
        //@ts-ignore
        globalThis.TypeScript.Services = globalThis.TypeScript.Services || {};
        //@ts-ignore
        globalThis.TypeScript.Services.TypeScriptServicesFactory = ts.TypeScriptServicesFactory;
        // 'toolsVersion' gets consumed by the managed side, so it's not unused.
        // TODO: it should be moved into a namespace though.
        //@ts-ignore
        globalThis.toolsVersion = ts.versionMajorMinor;
    }
    if (typeof module !== "undefined" && module.exports) {
        module.exports = ts;
    }
    var ts;
    (function (ts) {
        // The following are deprecations for the public API. Deprecated exports are removed from the compiler itself
        // and compatible implementations are added here, along with an appropriate deprecation warning using
        // the `@deprecated` JSDoc tag as well as the `Debug.deprecate` API.
        //
        // Deprecations fall into one of three categories:
        //
        //   * "soft" - Soft deprecations are indicated with the `@deprecated` JSDoc Tag.
        //   * "warn" - Warning deprecations are indicated with the `@deprecated` JSDoc Tag and a diagnostic message (assuming a compatible host)
        //   * "error" - Error deprecations are indicated with the `@deprecated` JSDoc tag and will throw a `TypeError` when invoked.
        // DEPRECATION: Node factory top-level exports
        // DEPRECATION PLAN:
        //     - soft: 4.0
        //     - warn: 4.1
        //     - error: TBD
        // #region Node factory top-level exports
        // NOTE: These exports are deprecated in favor of using a `NodeFactory` instance and exist here purely for backwards compatibility reasons.
        var factoryDeprecation = { since: "4.0", warnAfter: "4.1", message: "Use the appropriate method on 'ts.factory' or the 'factory' supplied by your transformation context instead." };
        /** @deprecated Use `factory.createNodeArray` or the factory supplied by your transformation context instead. */
        ts.createNodeArray = ts.Debug.deprecate(ts.factory.createNodeArray, factoryDeprecation);
        /** @deprecated Use `factory.createNumericLiteral` or the factory supplied by your transformation context instead. */
        ts.createNumericLiteral = ts.Debug.deprecate(ts.factory.createNumericLiteral, factoryDeprecation);
        /** @deprecated Use `factory.createBigIntLiteral` or the factory supplied by your transformation context instead. */
        ts.createBigIntLiteral = ts.Debug.deprecate(ts.factory.createBigIntLiteral, factoryDeprecation);
        /** @deprecated Use `factory.createStringLiteral` or the factory supplied by your transformation context instead. */
        ts.createStringLiteral = ts.Debug.deprecate(ts.factory.createStringLiteral, factoryDeprecation);
        /** @deprecated Use `factory.createStringLiteralFromNode` or the factory supplied by your transformation context instead. */
        ts.createStringLiteralFromNode = ts.Debug.deprecate(ts.factory.createStringLiteralFromNode, factoryDeprecation);
        /** @deprecated Use `factory.createRegularExpressionLiteral` or the factory supplied by your transformation context instead. */
        ts.createRegularExpressionLiteral = ts.Debug.deprecate(ts.factory.createRegularExpressionLiteral, factoryDeprecation);
        /** @deprecated Use `factory.createLoopVariable` or the factory supplied by your transformation context instead. */
        ts.createLoopVariable = ts.Debug.deprecate(ts.factory.createLoopVariable, factoryDeprecation);
        /** @deprecated Use `factory.createUniqueName` or the factory supplied by your transformation context instead. */
        ts.createUniqueName = ts.Debug.deprecate(ts.factory.createUniqueName, factoryDeprecation);
        /** @deprecated Use `factory.createPrivateIdentifier` or the factory supplied by your transformation context instead. */
        ts.createPrivateIdentifier = ts.Debug.deprecate(ts.factory.createPrivateIdentifier, factoryDeprecation);
        /** @deprecated Use `factory.createSuper` or the factory supplied by your transformation context instead. */
        ts.createSuper = ts.Debug.deprecate(ts.factory.createSuper, factoryDeprecation);
        /** @deprecated Use `factory.createThis` or the factory supplied by your transformation context instead. */
        ts.createThis = ts.Debug.deprecate(ts.factory.createThis, factoryDeprecation);
        /** @deprecated Use `factory.createNull` or the factory supplied by your transformation context instead. */
        ts.createNull = ts.Debug.deprecate(ts.factory.createNull, factoryDeprecation);
        /** @deprecated Use `factory.createTrue` or the factory supplied by your transformation context instead. */
        ts.createTrue = ts.Debug.deprecate(ts.factory.createTrue, factoryDeprecation);
        /** @deprecated Use `factory.createFalse` or the factory supplied by your transformation context instead. */
        ts.createFalse = ts.Debug.deprecate(ts.factory.createFalse, factoryDeprecation);
        /** @deprecated Use `factory.createModifier` or the factory supplied by your transformation context instead. */
        ts.createModifier = ts.Debug.deprecate(ts.factory.createModifier, factoryDeprecation);
        /** @deprecated Use `factory.createModifiersFromModifierFlags` or the factory supplied by your transformation context instead. */
        ts.createModifiersFromModifierFlags = ts.Debug.deprecate(ts.factory.createModifiersFromModifierFlags, factoryDeprecation);
        /** @deprecated Use `factory.createQualifiedName` or the factory supplied by your transformation context instead. */
        ts.createQualifiedName = ts.Debug.deprecate(ts.factory.createQualifiedName, factoryDeprecation);
        /** @deprecated Use `factory.updateQualifiedName` or the factory supplied by your transformation context instead. */
        ts.updateQualifiedName = ts.Debug.deprecate(ts.factory.updateQualifiedName, factoryDeprecation);
        /** @deprecated Use `factory.createComputedPropertyName` or the factory supplied by your transformation context instead. */
        ts.createComputedPropertyName = ts.Debug.deprecate(ts.factory.createComputedPropertyName, factoryDeprecation);
        /** @deprecated Use `factory.updateComputedPropertyName` or the factory supplied by your transformation context instead. */
        ts.updateComputedPropertyName = ts.Debug.deprecate(ts.factory.updateComputedPropertyName, factoryDeprecation);
        /** @deprecated Use `factory.createTypeParameterDeclaration` or the factory supplied by your transformation context instead. */
        ts.createTypeParameterDeclaration = ts.Debug.deprecate(ts.factory.createTypeParameterDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateTypeParameterDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateTypeParameterDeclaration = ts.Debug.deprecate(ts.factory.updateTypeParameterDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createParameterDeclaration` or the factory supplied by your transformation context instead. */
        ts.createParameter = ts.Debug.deprecate(ts.factory.createParameterDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateParameterDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateParameter = ts.Debug.deprecate(ts.factory.updateParameterDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createDecorator` or the factory supplied by your transformation context instead. */
        ts.createDecorator = ts.Debug.deprecate(ts.factory.createDecorator, factoryDeprecation);
        /** @deprecated Use `factory.updateDecorator` or the factory supplied by your transformation context instead. */
        ts.updateDecorator = ts.Debug.deprecate(ts.factory.updateDecorator, factoryDeprecation);
        /** @deprecated Use `factory.createPropertyDeclaration` or the factory supplied by your transformation context instead. */
        ts.createProperty = ts.Debug.deprecate(ts.factory.createPropertyDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updatePropertyDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateProperty = ts.Debug.deprecate(ts.factory.updatePropertyDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createMethodDeclaration` or the factory supplied by your transformation context instead. */
        ts.createMethod = ts.Debug.deprecate(ts.factory.createMethodDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateMethodDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateMethod = ts.Debug.deprecate(ts.factory.updateMethodDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createConstructorDeclaration` or the factory supplied by your transformation context instead. */
        ts.createConstructor = ts.Debug.deprecate(ts.factory.createConstructorDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateConstructorDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateConstructor = ts.Debug.deprecate(ts.factory.updateConstructorDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createGetAccessorDeclaration` or the factory supplied by your transformation context instead. */
        ts.createGetAccessor = ts.Debug.deprecate(ts.factory.createGetAccessorDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateGetAccessorDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateGetAccessor = ts.Debug.deprecate(ts.factory.updateGetAccessorDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createSetAccessorDeclaration` or the factory supplied by your transformation context instead. */
        ts.createSetAccessor = ts.Debug.deprecate(ts.factory.createSetAccessorDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateSetAccessorDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateSetAccessor = ts.Debug.deprecate(ts.factory.updateSetAccessorDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createCallSignature` or the factory supplied by your transformation context instead. */
        ts.createCallSignature = ts.Debug.deprecate(ts.factory.createCallSignature, factoryDeprecation);
        /** @deprecated Use `factory.updateCallSignature` or the factory supplied by your transformation context instead. */
        ts.updateCallSignature = ts.Debug.deprecate(ts.factory.updateCallSignature, factoryDeprecation);
        /** @deprecated Use `factory.createConstructSignature` or the factory supplied by your transformation context instead. */
        ts.createConstructSignature = ts.Debug.deprecate(ts.factory.createConstructSignature, factoryDeprecation);
        /** @deprecated Use `factory.updateConstructSignature` or the factory supplied by your transformation context instead. */
        ts.updateConstructSignature = ts.Debug.deprecate(ts.factory.updateConstructSignature, factoryDeprecation);
        /** @deprecated Use `factory.updateIndexSignature` or the factory supplied by your transformation context instead. */
        ts.updateIndexSignature = ts.Debug.deprecate(ts.factory.updateIndexSignature, factoryDeprecation);
        /** @deprecated Use `factory.createKeywordTypeNode` or the factory supplied by your transformation context instead. */
        ts.createKeywordTypeNode = ts.Debug.deprecate(ts.factory.createKeywordTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createTypePredicateNode` or the factory supplied by your transformation context instead. */
        ts.createTypePredicateNodeWithModifier = ts.Debug.deprecate(ts.factory.createTypePredicateNode, factoryDeprecation);
        /** @deprecated Use `factory.updateTypePredicateNode` or the factory supplied by your transformation context instead. */
        ts.updateTypePredicateNodeWithModifier = ts.Debug.deprecate(ts.factory.updateTypePredicateNode, factoryDeprecation);
        /** @deprecated Use `factory.createTypeReferenceNode` or the factory supplied by your transformation context instead. */
        ts.createTypeReferenceNode = ts.Debug.deprecate(ts.factory.createTypeReferenceNode, factoryDeprecation);
        /** @deprecated Use `factory.updateTypeReferenceNode` or the factory supplied by your transformation context instead. */
        ts.updateTypeReferenceNode = ts.Debug.deprecate(ts.factory.updateTypeReferenceNode, factoryDeprecation);
        /** @deprecated Use `factory.createFunctionTypeNode` or the factory supplied by your transformation context instead. */
        ts.createFunctionTypeNode = ts.Debug.deprecate(ts.factory.createFunctionTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateFunctionTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateFunctionTypeNode = ts.Debug.deprecate(ts.factory.updateFunctionTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createConstructorTypeNode` or the factory supplied by your transformation context instead. */
        ts.createConstructorTypeNode = ts.Debug.deprecate(function (typeParameters, parameters, type) {
            return ts.factory.createConstructorTypeNode(/*modifiers*/ undefined, typeParameters, parameters, type);
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateConstructorTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateConstructorTypeNode = ts.Debug.deprecate(function (node, typeParameters, parameters, type) {
            return ts.factory.updateConstructorTypeNode(node, node.modifiers, typeParameters, parameters, type);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createTypeQueryNode` or the factory supplied by your transformation context instead. */
        ts.createTypeQueryNode = ts.Debug.deprecate(ts.factory.createTypeQueryNode, factoryDeprecation);
        /** @deprecated Use `factory.updateTypeQueryNode` or the factory supplied by your transformation context instead. */
        ts.updateTypeQueryNode = ts.Debug.deprecate(ts.factory.updateTypeQueryNode, factoryDeprecation);
        /** @deprecated Use `factory.createTypeLiteralNode` or the factory supplied by your transformation context instead. */
        ts.createTypeLiteralNode = ts.Debug.deprecate(ts.factory.createTypeLiteralNode, factoryDeprecation);
        /** @deprecated Use `factory.updateTypeLiteralNode` or the factory supplied by your transformation context instead. */
        ts.updateTypeLiteralNode = ts.Debug.deprecate(ts.factory.updateTypeLiteralNode, factoryDeprecation);
        /** @deprecated Use `factory.createArrayTypeNode` or the factory supplied by your transformation context instead. */
        ts.createArrayTypeNode = ts.Debug.deprecate(ts.factory.createArrayTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateArrayTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateArrayTypeNode = ts.Debug.deprecate(ts.factory.updateArrayTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createTupleTypeNode` or the factory supplied by your transformation context instead. */
        ts.createTupleTypeNode = ts.Debug.deprecate(ts.factory.createTupleTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateTupleTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateTupleTypeNode = ts.Debug.deprecate(ts.factory.updateTupleTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createOptionalTypeNode` or the factory supplied by your transformation context instead. */
        ts.createOptionalTypeNode = ts.Debug.deprecate(ts.factory.createOptionalTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateOptionalTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateOptionalTypeNode = ts.Debug.deprecate(ts.factory.updateOptionalTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createRestTypeNode` or the factory supplied by your transformation context instead. */
        ts.createRestTypeNode = ts.Debug.deprecate(ts.factory.createRestTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateRestTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateRestTypeNode = ts.Debug.deprecate(ts.factory.updateRestTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createUnionTypeNode` or the factory supplied by your transformation context instead. */
        ts.createUnionTypeNode = ts.Debug.deprecate(ts.factory.createUnionTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateUnionTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateUnionTypeNode = ts.Debug.deprecate(ts.factory.updateUnionTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createIntersectionTypeNode` or the factory supplied by your transformation context instead. */
        ts.createIntersectionTypeNode = ts.Debug.deprecate(ts.factory.createIntersectionTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateIntersectionTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateIntersectionTypeNode = ts.Debug.deprecate(ts.factory.updateIntersectionTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createConditionalTypeNode` or the factory supplied by your transformation context instead. */
        ts.createConditionalTypeNode = ts.Debug.deprecate(ts.factory.createConditionalTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateConditionalTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateConditionalTypeNode = ts.Debug.deprecate(ts.factory.updateConditionalTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createInferTypeNode` or the factory supplied by your transformation context instead. */
        ts.createInferTypeNode = ts.Debug.deprecate(ts.factory.createInferTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateInferTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateInferTypeNode = ts.Debug.deprecate(ts.factory.updateInferTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createImportTypeNode` or the factory supplied by your transformation context instead. */
        ts.createImportTypeNode = ts.Debug.deprecate(ts.factory.createImportTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateImportTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateImportTypeNode = ts.Debug.deprecate(ts.factory.updateImportTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createParenthesizedType` or the factory supplied by your transformation context instead. */
        ts.createParenthesizedType = ts.Debug.deprecate(ts.factory.createParenthesizedType, factoryDeprecation);
        /** @deprecated Use `factory.updateParenthesizedType` or the factory supplied by your transformation context instead. */
        ts.updateParenthesizedType = ts.Debug.deprecate(ts.factory.updateParenthesizedType, factoryDeprecation);
        /** @deprecated Use `factory.createThisTypeNode` or the factory supplied by your transformation context instead. */
        ts.createThisTypeNode = ts.Debug.deprecate(ts.factory.createThisTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateTypeOperatorNode` or the factory supplied by your transformation context instead. */
        ts.updateTypeOperatorNode = ts.Debug.deprecate(ts.factory.updateTypeOperatorNode, factoryDeprecation);
        /** @deprecated Use `factory.createIndexedAccessTypeNode` or the factory supplied by your transformation context instead. */
        ts.createIndexedAccessTypeNode = ts.Debug.deprecate(ts.factory.createIndexedAccessTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateIndexedAccessTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateIndexedAccessTypeNode = ts.Debug.deprecate(ts.factory.updateIndexedAccessTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createMappedTypeNode` or the factory supplied by your transformation context instead. */
        ts.createMappedTypeNode = ts.Debug.deprecate(ts.factory.createMappedTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateMappedTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateMappedTypeNode = ts.Debug.deprecate(ts.factory.updateMappedTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createLiteralTypeNode` or the factory supplied by your transformation context instead. */
        ts.createLiteralTypeNode = ts.Debug.deprecate(ts.factory.createLiteralTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.updateLiteralTypeNode` or the factory supplied by your transformation context instead. */
        ts.updateLiteralTypeNode = ts.Debug.deprecate(ts.factory.updateLiteralTypeNode, factoryDeprecation);
        /** @deprecated Use `factory.createObjectBindingPattern` or the factory supplied by your transformation context instead. */
        ts.createObjectBindingPattern = ts.Debug.deprecate(ts.factory.createObjectBindingPattern, factoryDeprecation);
        /** @deprecated Use `factory.updateObjectBindingPattern` or the factory supplied by your transformation context instead. */
        ts.updateObjectBindingPattern = ts.Debug.deprecate(ts.factory.updateObjectBindingPattern, factoryDeprecation);
        /** @deprecated Use `factory.createArrayBindingPattern` or the factory supplied by your transformation context instead. */
        ts.createArrayBindingPattern = ts.Debug.deprecate(ts.factory.createArrayBindingPattern, factoryDeprecation);
        /** @deprecated Use `factory.updateArrayBindingPattern` or the factory supplied by your transformation context instead. */
        ts.updateArrayBindingPattern = ts.Debug.deprecate(ts.factory.updateArrayBindingPattern, factoryDeprecation);
        /** @deprecated Use `factory.createBindingElement` or the factory supplied by your transformation context instead. */
        ts.createBindingElement = ts.Debug.deprecate(ts.factory.createBindingElement, factoryDeprecation);
        /** @deprecated Use `factory.updateBindingElement` or the factory supplied by your transformation context instead. */
        ts.updateBindingElement = ts.Debug.deprecate(ts.factory.updateBindingElement, factoryDeprecation);
        /** @deprecated Use `factory.createArrayLiteralExpression` or the factory supplied by your transformation context instead. */
        ts.createArrayLiteral = ts.Debug.deprecate(ts.factory.createArrayLiteralExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateArrayLiteralExpression` or the factory supplied by your transformation context instead. */
        ts.updateArrayLiteral = ts.Debug.deprecate(ts.factory.updateArrayLiteralExpression, factoryDeprecation);
        /** @deprecated Use `factory.createObjectLiteralExpression` or the factory supplied by your transformation context instead. */
        ts.createObjectLiteral = ts.Debug.deprecate(ts.factory.createObjectLiteralExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateObjectLiteralExpression` or the factory supplied by your transformation context instead. */
        ts.updateObjectLiteral = ts.Debug.deprecate(ts.factory.updateObjectLiteralExpression, factoryDeprecation);
        /** @deprecated Use `factory.createPropertyAccessExpression` or the factory supplied by your transformation context instead. */
        ts.createPropertyAccess = ts.Debug.deprecate(ts.factory.createPropertyAccessExpression, factoryDeprecation);
        /** @deprecated Use `factory.updatePropertyAccessExpression` or the factory supplied by your transformation context instead. */
        ts.updatePropertyAccess = ts.Debug.deprecate(ts.factory.updatePropertyAccessExpression, factoryDeprecation);
        /** @deprecated Use `factory.createPropertyAccessChain` or the factory supplied by your transformation context instead. */
        ts.createPropertyAccessChain = ts.Debug.deprecate(ts.factory.createPropertyAccessChain, factoryDeprecation);
        /** @deprecated Use `factory.updatePropertyAccessChain` or the factory supplied by your transformation context instead. */
        ts.updatePropertyAccessChain = ts.Debug.deprecate(ts.factory.updatePropertyAccessChain, factoryDeprecation);
        /** @deprecated Use `factory.createElementAccessExpression` or the factory supplied by your transformation context instead. */
        ts.createElementAccess = ts.Debug.deprecate(ts.factory.createElementAccessExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateElementAccessExpression` or the factory supplied by your transformation context instead. */
        ts.updateElementAccess = ts.Debug.deprecate(ts.factory.updateElementAccessExpression, factoryDeprecation);
        /** @deprecated Use `factory.createElementAccessChain` or the factory supplied by your transformation context instead. */
        ts.createElementAccessChain = ts.Debug.deprecate(ts.factory.createElementAccessChain, factoryDeprecation);
        /** @deprecated Use `factory.updateElementAccessChain` or the factory supplied by your transformation context instead. */
        ts.updateElementAccessChain = ts.Debug.deprecate(ts.factory.updateElementAccessChain, factoryDeprecation);
        /** @deprecated Use `factory.createCallExpression` or the factory supplied by your transformation context instead. */
        ts.createCall = ts.Debug.deprecate(ts.factory.createCallExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateCallExpression` or the factory supplied by your transformation context instead. */
        ts.updateCall = ts.Debug.deprecate(ts.factory.updateCallExpression, factoryDeprecation);
        /** @deprecated Use `factory.createCallChain` or the factory supplied by your transformation context instead. */
        ts.createCallChain = ts.Debug.deprecate(ts.factory.createCallChain, factoryDeprecation);
        /** @deprecated Use `factory.updateCallChain` or the factory supplied by your transformation context instead. */
        ts.updateCallChain = ts.Debug.deprecate(ts.factory.updateCallChain, factoryDeprecation);
        /** @deprecated Use `factory.createNewExpression` or the factory supplied by your transformation context instead. */
        ts.createNew = ts.Debug.deprecate(ts.factory.createNewExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateNewExpression` or the factory supplied by your transformation context instead. */
        ts.updateNew = ts.Debug.deprecate(ts.factory.updateNewExpression, factoryDeprecation);
        /** @deprecated Use `factory.createTypeAssertion` or the factory supplied by your transformation context instead. */
        ts.createTypeAssertion = ts.Debug.deprecate(ts.factory.createTypeAssertion, factoryDeprecation);
        /** @deprecated Use `factory.updateTypeAssertion` or the factory supplied by your transformation context instead. */
        ts.updateTypeAssertion = ts.Debug.deprecate(ts.factory.updateTypeAssertion, factoryDeprecation);
        /** @deprecated Use `factory.createParenthesizedExpression` or the factory supplied by your transformation context instead. */
        ts.createParen = ts.Debug.deprecate(ts.factory.createParenthesizedExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateParenthesizedExpression` or the factory supplied by your transformation context instead. */
        ts.updateParen = ts.Debug.deprecate(ts.factory.updateParenthesizedExpression, factoryDeprecation);
        /** @deprecated Use `factory.createFunctionExpression` or the factory supplied by your transformation context instead. */
        ts.createFunctionExpression = ts.Debug.deprecate(ts.factory.createFunctionExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateFunctionExpression` or the factory supplied by your transformation context instead. */
        ts.updateFunctionExpression = ts.Debug.deprecate(ts.factory.updateFunctionExpression, factoryDeprecation);
        /** @deprecated Use `factory.createDeleteExpression` or the factory supplied by your transformation context instead. */
        ts.createDelete = ts.Debug.deprecate(ts.factory.createDeleteExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateDeleteExpression` or the factory supplied by your transformation context instead. */
        ts.updateDelete = ts.Debug.deprecate(ts.factory.updateDeleteExpression, factoryDeprecation);
        /** @deprecated Use `factory.createTypeOfExpression` or the factory supplied by your transformation context instead. */
        ts.createTypeOf = ts.Debug.deprecate(ts.factory.createTypeOfExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateTypeOfExpression` or the factory supplied by your transformation context instead. */
        ts.updateTypeOf = ts.Debug.deprecate(ts.factory.updateTypeOfExpression, factoryDeprecation);
        /** @deprecated Use `factory.createVoidExpression` or the factory supplied by your transformation context instead. */
        ts.createVoid = ts.Debug.deprecate(ts.factory.createVoidExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateVoidExpression` or the factory supplied by your transformation context instead. */
        ts.updateVoid = ts.Debug.deprecate(ts.factory.updateVoidExpression, factoryDeprecation);
        /** @deprecated Use `factory.createAwaitExpression` or the factory supplied by your transformation context instead. */
        ts.createAwait = ts.Debug.deprecate(ts.factory.createAwaitExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateAwaitExpression` or the factory supplied by your transformation context instead. */
        ts.updateAwait = ts.Debug.deprecate(ts.factory.updateAwaitExpression, factoryDeprecation);
        /** @deprecated Use `factory.createPrefixExpression` or the factory supplied by your transformation context instead. */
        ts.createPrefix = ts.Debug.deprecate(ts.factory.createPrefixUnaryExpression, factoryDeprecation);
        /** @deprecated Use `factory.updatePrefixExpression` or the factory supplied by your transformation context instead. */
        ts.updatePrefix = ts.Debug.deprecate(ts.factory.updatePrefixUnaryExpression, factoryDeprecation);
        /** @deprecated Use `factory.createPostfixUnaryExpression` or the factory supplied by your transformation context instead. */
        ts.createPostfix = ts.Debug.deprecate(ts.factory.createPostfixUnaryExpression, factoryDeprecation);
        /** @deprecated Use `factory.updatePostfixUnaryExpression` or the factory supplied by your transformation context instead. */
        ts.updatePostfix = ts.Debug.deprecate(ts.factory.updatePostfixUnaryExpression, factoryDeprecation);
        /** @deprecated Use `factory.createBinaryExpression` or the factory supplied by your transformation context instead. */
        ts.createBinary = ts.Debug.deprecate(ts.factory.createBinaryExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateConditionalExpression` or the factory supplied by your transformation context instead. */
        ts.updateConditional = ts.Debug.deprecate(ts.factory.updateConditionalExpression, factoryDeprecation);
        /** @deprecated Use `factory.createTemplateExpression` or the factory supplied by your transformation context instead. */
        ts.createTemplateExpression = ts.Debug.deprecate(ts.factory.createTemplateExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateTemplateExpression` or the factory supplied by your transformation context instead. */
        ts.updateTemplateExpression = ts.Debug.deprecate(ts.factory.updateTemplateExpression, factoryDeprecation);
        /** @deprecated Use `factory.createTemplateHead` or the factory supplied by your transformation context instead. */
        ts.createTemplateHead = ts.Debug.deprecate(ts.factory.createTemplateHead, factoryDeprecation);
        /** @deprecated Use `factory.createTemplateMiddle` or the factory supplied by your transformation context instead. */
        ts.createTemplateMiddle = ts.Debug.deprecate(ts.factory.createTemplateMiddle, factoryDeprecation);
        /** @deprecated Use `factory.createTemplateTail` or the factory supplied by your transformation context instead. */
        ts.createTemplateTail = ts.Debug.deprecate(ts.factory.createTemplateTail, factoryDeprecation);
        /** @deprecated Use `factory.createNoSubstitutionTemplateLiteral` or the factory supplied by your transformation context instead. */
        ts.createNoSubstitutionTemplateLiteral = ts.Debug.deprecate(ts.factory.createNoSubstitutionTemplateLiteral, factoryDeprecation);
        /** @deprecated Use `factory.updateYieldExpression` or the factory supplied by your transformation context instead. */
        ts.updateYield = ts.Debug.deprecate(ts.factory.updateYieldExpression, factoryDeprecation);
        /** @deprecated Use `factory.createSpreadExpression` or the factory supplied by your transformation context instead. */
        ts.createSpread = ts.Debug.deprecate(ts.factory.createSpreadElement, factoryDeprecation);
        /** @deprecated Use `factory.updateSpreadExpression` or the factory supplied by your transformation context instead. */
        ts.updateSpread = ts.Debug.deprecate(ts.factory.updateSpreadElement, factoryDeprecation);
        /** @deprecated Use `factory.createOmittedExpression` or the factory supplied by your transformation context instead. */
        ts.createOmittedExpression = ts.Debug.deprecate(ts.factory.createOmittedExpression, factoryDeprecation);
        /** @deprecated Use `factory.createAsExpression` or the factory supplied by your transformation context instead. */
        ts.createAsExpression = ts.Debug.deprecate(ts.factory.createAsExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateAsExpression` or the factory supplied by your transformation context instead. */
        ts.updateAsExpression = ts.Debug.deprecate(ts.factory.updateAsExpression, factoryDeprecation);
        /** @deprecated Use `factory.createNonNullExpression` or the factory supplied by your transformation context instead. */
        ts.createNonNullExpression = ts.Debug.deprecate(ts.factory.createNonNullExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateNonNullExpression` or the factory supplied by your transformation context instead. */
        ts.updateNonNullExpression = ts.Debug.deprecate(ts.factory.updateNonNullExpression, factoryDeprecation);
        /** @deprecated Use `factory.createNonNullChain` or the factory supplied by your transformation context instead. */
        ts.createNonNullChain = ts.Debug.deprecate(ts.factory.createNonNullChain, factoryDeprecation);
        /** @deprecated Use `factory.updateNonNullChain` or the factory supplied by your transformation context instead. */
        ts.updateNonNullChain = ts.Debug.deprecate(ts.factory.updateNonNullChain, factoryDeprecation);
        /** @deprecated Use `factory.createMetaProperty` or the factory supplied by your transformation context instead. */
        ts.createMetaProperty = ts.Debug.deprecate(ts.factory.createMetaProperty, factoryDeprecation);
        /** @deprecated Use `factory.updateMetaProperty` or the factory supplied by your transformation context instead. */
        ts.updateMetaProperty = ts.Debug.deprecate(ts.factory.updateMetaProperty, factoryDeprecation);
        /** @deprecated Use `factory.createTemplateSpan` or the factory supplied by your transformation context instead. */
        ts.createTemplateSpan = ts.Debug.deprecate(ts.factory.createTemplateSpan, factoryDeprecation);
        /** @deprecated Use `factory.updateTemplateSpan` or the factory supplied by your transformation context instead. */
        ts.updateTemplateSpan = ts.Debug.deprecate(ts.factory.updateTemplateSpan, factoryDeprecation);
        /** @deprecated Use `factory.createSemicolonClassElement` or the factory supplied by your transformation context instead. */
        ts.createSemicolonClassElement = ts.Debug.deprecate(ts.factory.createSemicolonClassElement, factoryDeprecation);
        /** @deprecated Use `factory.createBlock` or the factory supplied by your transformation context instead. */
        ts.createBlock = ts.Debug.deprecate(ts.factory.createBlock, factoryDeprecation);
        /** @deprecated Use `factory.updateBlock` or the factory supplied by your transformation context instead. */
        ts.updateBlock = ts.Debug.deprecate(ts.factory.updateBlock, factoryDeprecation);
        /** @deprecated Use `factory.createVariableStatement` or the factory supplied by your transformation context instead. */
        ts.createVariableStatement = ts.Debug.deprecate(ts.factory.createVariableStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateVariableStatement` or the factory supplied by your transformation context instead. */
        ts.updateVariableStatement = ts.Debug.deprecate(ts.factory.updateVariableStatement, factoryDeprecation);
        /** @deprecated Use `factory.createEmptyStatement` or the factory supplied by your transformation context instead. */
        ts.createEmptyStatement = ts.Debug.deprecate(ts.factory.createEmptyStatement, factoryDeprecation);
        /** @deprecated Use `factory.createExpressionStatement` or the factory supplied by your transformation context instead. */
        ts.createExpressionStatement = ts.Debug.deprecate(ts.factory.createExpressionStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateExpressionStatement` or the factory supplied by your transformation context instead. */
        ts.updateExpressionStatement = ts.Debug.deprecate(ts.factory.updateExpressionStatement, factoryDeprecation);
        /** @deprecated Use `factory.createExpressionStatement` or the factory supplied by your transformation context instead. */
        ts.createStatement = ts.Debug.deprecate(ts.factory.createExpressionStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateExpressionStatement` or the factory supplied by your transformation context instead. */
        ts.updateStatement = ts.Debug.deprecate(ts.factory.updateExpressionStatement, factoryDeprecation);
        /** @deprecated Use `factory.createIfStatement` or the factory supplied by your transformation context instead. */
        ts.createIf = ts.Debug.deprecate(ts.factory.createIfStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateIfStatement` or the factory supplied by your transformation context instead. */
        ts.updateIf = ts.Debug.deprecate(ts.factory.updateIfStatement, factoryDeprecation);
        /** @deprecated Use `factory.createDoStatement` or the factory supplied by your transformation context instead. */
        ts.createDo = ts.Debug.deprecate(ts.factory.createDoStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateDoStatement` or the factory supplied by your transformation context instead. */
        ts.updateDo = ts.Debug.deprecate(ts.factory.updateDoStatement, factoryDeprecation);
        /** @deprecated Use `factory.createWhileStatement` or the factory supplied by your transformation context instead. */
        ts.createWhile = ts.Debug.deprecate(ts.factory.createWhileStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateWhileStatement` or the factory supplied by your transformation context instead. */
        ts.updateWhile = ts.Debug.deprecate(ts.factory.updateWhileStatement, factoryDeprecation);
        /** @deprecated Use `factory.createForStatement` or the factory supplied by your transformation context instead. */
        ts.createFor = ts.Debug.deprecate(ts.factory.createForStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateForStatement` or the factory supplied by your transformation context instead. */
        ts.updateFor = ts.Debug.deprecate(ts.factory.updateForStatement, factoryDeprecation);
        /** @deprecated Use `factory.createForInStatement` or the factory supplied by your transformation context instead. */
        ts.createForIn = ts.Debug.deprecate(ts.factory.createForInStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateForInStatement` or the factory supplied by your transformation context instead. */
        ts.updateForIn = ts.Debug.deprecate(ts.factory.updateForInStatement, factoryDeprecation);
        /** @deprecated Use `factory.createForOfStatement` or the factory supplied by your transformation context instead. */
        ts.createForOf = ts.Debug.deprecate(ts.factory.createForOfStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateForOfStatement` or the factory supplied by your transformation context instead. */
        ts.updateForOf = ts.Debug.deprecate(ts.factory.updateForOfStatement, factoryDeprecation);
        /** @deprecated Use `factory.createContinueStatement` or the factory supplied by your transformation context instead. */
        ts.createContinue = ts.Debug.deprecate(ts.factory.createContinueStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateContinueStatement` or the factory supplied by your transformation context instead. */
        ts.updateContinue = ts.Debug.deprecate(ts.factory.updateContinueStatement, factoryDeprecation);
        /** @deprecated Use `factory.createBreakStatement` or the factory supplied by your transformation context instead. */
        ts.createBreak = ts.Debug.deprecate(ts.factory.createBreakStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateBreakStatement` or the factory supplied by your transformation context instead. */
        ts.updateBreak = ts.Debug.deprecate(ts.factory.updateBreakStatement, factoryDeprecation);
        /** @deprecated Use `factory.createReturnStatement` or the factory supplied by your transformation context instead. */
        ts.createReturn = ts.Debug.deprecate(ts.factory.createReturnStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateReturnStatement` or the factory supplied by your transformation context instead. */
        ts.updateReturn = ts.Debug.deprecate(ts.factory.updateReturnStatement, factoryDeprecation);
        /** @deprecated Use `factory.createWithStatement` or the factory supplied by your transformation context instead. */
        ts.createWith = ts.Debug.deprecate(ts.factory.createWithStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateWithStatement` or the factory supplied by your transformation context instead. */
        ts.updateWith = ts.Debug.deprecate(ts.factory.updateWithStatement, factoryDeprecation);
        /** @deprecated Use `factory.createSwitchStatement` or the factory supplied by your transformation context instead. */
        ts.createSwitch = ts.Debug.deprecate(ts.factory.createSwitchStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateSwitchStatement` or the factory supplied by your transformation context instead. */
        ts.updateSwitch = ts.Debug.deprecate(ts.factory.updateSwitchStatement, factoryDeprecation);
        /** @deprecated Use `factory.createLabelStatement` or the factory supplied by your transformation context instead. */
        ts.createLabel = ts.Debug.deprecate(ts.factory.createLabeledStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateLabelStatement` or the factory supplied by your transformation context instead. */
        ts.updateLabel = ts.Debug.deprecate(ts.factory.updateLabeledStatement, factoryDeprecation);
        /** @deprecated Use `factory.createThrowStatement` or the factory supplied by your transformation context instead. */
        ts.createThrow = ts.Debug.deprecate(ts.factory.createThrowStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateThrowStatement` or the factory supplied by your transformation context instead. */
        ts.updateThrow = ts.Debug.deprecate(ts.factory.updateThrowStatement, factoryDeprecation);
        /** @deprecated Use `factory.createTryStatement` or the factory supplied by your transformation context instead. */
        ts.createTry = ts.Debug.deprecate(ts.factory.createTryStatement, factoryDeprecation);
        /** @deprecated Use `factory.updateTryStatement` or the factory supplied by your transformation context instead. */
        ts.updateTry = ts.Debug.deprecate(ts.factory.updateTryStatement, factoryDeprecation);
        /** @deprecated Use `factory.createDebuggerStatement` or the factory supplied by your transformation context instead. */
        ts.createDebuggerStatement = ts.Debug.deprecate(ts.factory.createDebuggerStatement, factoryDeprecation);
        /** @deprecated Use `factory.createVariableDeclarationList` or the factory supplied by your transformation context instead. */
        ts.createVariableDeclarationList = ts.Debug.deprecate(ts.factory.createVariableDeclarationList, factoryDeprecation);
        /** @deprecated Use `factory.updateVariableDeclarationList` or the factory supplied by your transformation context instead. */
        ts.updateVariableDeclarationList = ts.Debug.deprecate(ts.factory.updateVariableDeclarationList, factoryDeprecation);
        /** @deprecated Use `factory.createFunctionDeclaration` or the factory supplied by your transformation context instead. */
        ts.createFunctionDeclaration = ts.Debug.deprecate(ts.factory.createFunctionDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateFunctionDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateFunctionDeclaration = ts.Debug.deprecate(ts.factory.updateFunctionDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createClassDeclaration` or the factory supplied by your transformation context instead. */
        ts.createClassDeclaration = ts.Debug.deprecate(ts.factory.createClassDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateClassDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateClassDeclaration = ts.Debug.deprecate(ts.factory.updateClassDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createInterfaceDeclaration` or the factory supplied by your transformation context instead. */
        ts.createInterfaceDeclaration = ts.Debug.deprecate(ts.factory.createInterfaceDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateInterfaceDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateInterfaceDeclaration = ts.Debug.deprecate(ts.factory.updateInterfaceDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createTypeAliasDeclaration` or the factory supplied by your transformation context instead. */
        ts.createTypeAliasDeclaration = ts.Debug.deprecate(ts.factory.createTypeAliasDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateTypeAliasDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateTypeAliasDeclaration = ts.Debug.deprecate(ts.factory.updateTypeAliasDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createEnumDeclaration` or the factory supplied by your transformation context instead. */
        ts.createEnumDeclaration = ts.Debug.deprecate(ts.factory.createEnumDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateEnumDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateEnumDeclaration = ts.Debug.deprecate(ts.factory.updateEnumDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createModuleDeclaration` or the factory supplied by your transformation context instead. */
        ts.createModuleDeclaration = ts.Debug.deprecate(ts.factory.createModuleDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateModuleDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateModuleDeclaration = ts.Debug.deprecate(ts.factory.updateModuleDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createModuleBlock` or the factory supplied by your transformation context instead. */
        ts.createModuleBlock = ts.Debug.deprecate(ts.factory.createModuleBlock, factoryDeprecation);
        /** @deprecated Use `factory.updateModuleBlock` or the factory supplied by your transformation context instead. */
        ts.updateModuleBlock = ts.Debug.deprecate(ts.factory.updateModuleBlock, factoryDeprecation);
        /** @deprecated Use `factory.createCaseBlock` or the factory supplied by your transformation context instead. */
        ts.createCaseBlock = ts.Debug.deprecate(ts.factory.createCaseBlock, factoryDeprecation);
        /** @deprecated Use `factory.updateCaseBlock` or the factory supplied by your transformation context instead. */
        ts.updateCaseBlock = ts.Debug.deprecate(ts.factory.updateCaseBlock, factoryDeprecation);
        /** @deprecated Use `factory.createNamespaceExportDeclaration` or the factory supplied by your transformation context instead. */
        ts.createNamespaceExportDeclaration = ts.Debug.deprecate(ts.factory.createNamespaceExportDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateNamespaceExportDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateNamespaceExportDeclaration = ts.Debug.deprecate(ts.factory.updateNamespaceExportDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createImportEqualsDeclaration` or the factory supplied by your transformation context instead. */
        ts.createImportEqualsDeclaration = ts.Debug.deprecate(ts.factory.createImportEqualsDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateImportEqualsDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateImportEqualsDeclaration = ts.Debug.deprecate(ts.factory.updateImportEqualsDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createImportDeclaration` or the factory supplied by your transformation context instead. */
        ts.createImportDeclaration = ts.Debug.deprecate(ts.factory.createImportDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.updateImportDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateImportDeclaration = ts.Debug.deprecate(ts.factory.updateImportDeclaration, factoryDeprecation);
        /** @deprecated Use `factory.createNamespaceImport` or the factory supplied by your transformation context instead. */
        ts.createNamespaceImport = ts.Debug.deprecate(ts.factory.createNamespaceImport, factoryDeprecation);
        /** @deprecated Use `factory.updateNamespaceImport` or the factory supplied by your transformation context instead. */
        ts.updateNamespaceImport = ts.Debug.deprecate(ts.factory.updateNamespaceImport, factoryDeprecation);
        /** @deprecated Use `factory.createNamedImports` or the factory supplied by your transformation context instead. */
        ts.createNamedImports = ts.Debug.deprecate(ts.factory.createNamedImports, factoryDeprecation);
        /** @deprecated Use `factory.updateNamedImports` or the factory supplied by your transformation context instead. */
        ts.updateNamedImports = ts.Debug.deprecate(ts.factory.updateNamedImports, factoryDeprecation);
        /** @deprecated Use `factory.createImportSpecifier` or the factory supplied by your transformation context instead. */
        ts.createImportSpecifier = ts.Debug.deprecate(ts.factory.createImportSpecifier, factoryDeprecation);
        /** @deprecated Use `factory.updateImportSpecifier` or the factory supplied by your transformation context instead. */
        ts.updateImportSpecifier = ts.Debug.deprecate(ts.factory.updateImportSpecifier, factoryDeprecation);
        /** @deprecated Use `factory.createExportAssignment` or the factory supplied by your transformation context instead. */
        ts.createExportAssignment = ts.Debug.deprecate(ts.factory.createExportAssignment, factoryDeprecation);
        /** @deprecated Use `factory.updateExportAssignment` or the factory supplied by your transformation context instead. */
        ts.updateExportAssignment = ts.Debug.deprecate(ts.factory.updateExportAssignment, factoryDeprecation);
        /** @deprecated Use `factory.createNamedExports` or the factory supplied by your transformation context instead. */
        ts.createNamedExports = ts.Debug.deprecate(ts.factory.createNamedExports, factoryDeprecation);
        /** @deprecated Use `factory.updateNamedExports` or the factory supplied by your transformation context instead. */
        ts.updateNamedExports = ts.Debug.deprecate(ts.factory.updateNamedExports, factoryDeprecation);
        /** @deprecated Use `factory.createExportSpecifier` or the factory supplied by your transformation context instead. */
        ts.createExportSpecifier = ts.Debug.deprecate(ts.factory.createExportSpecifier, factoryDeprecation);
        /** @deprecated Use `factory.updateExportSpecifier` or the factory supplied by your transformation context instead. */
        ts.updateExportSpecifier = ts.Debug.deprecate(ts.factory.updateExportSpecifier, factoryDeprecation);
        /** @deprecated Use `factory.createExternalModuleReference` or the factory supplied by your transformation context instead. */
        ts.createExternalModuleReference = ts.Debug.deprecate(ts.factory.createExternalModuleReference, factoryDeprecation);
        /** @deprecated Use `factory.updateExternalModuleReference` or the factory supplied by your transformation context instead. */
        ts.updateExternalModuleReference = ts.Debug.deprecate(ts.factory.updateExternalModuleReference, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocTypeExpression` or the factory supplied by your transformation context instead. */
        ts.createJSDocTypeExpression = ts.Debug.deprecate(ts.factory.createJSDocTypeExpression, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocTypeTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocTypeTag = ts.Debug.deprecate(ts.factory.createJSDocTypeTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocReturnTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocReturnTag = ts.Debug.deprecate(ts.factory.createJSDocReturnTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocThisTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocThisTag = ts.Debug.deprecate(ts.factory.createJSDocThisTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocComment` or the factory supplied by your transformation context instead. */
        ts.createJSDocComment = ts.Debug.deprecate(ts.factory.createJSDocComment, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocParameterTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocParameterTag = ts.Debug.deprecate(ts.factory.createJSDocParameterTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocClassTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocClassTag = ts.Debug.deprecate(ts.factory.createJSDocClassTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocAugmentsTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocAugmentsTag = ts.Debug.deprecate(ts.factory.createJSDocAugmentsTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocEnumTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocEnumTag = ts.Debug.deprecate(ts.factory.createJSDocEnumTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocTemplateTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocTemplateTag = ts.Debug.deprecate(ts.factory.createJSDocTemplateTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocTypedefTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocTypedefTag = ts.Debug.deprecate(ts.factory.createJSDocTypedefTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocCallbackTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocCallbackTag = ts.Debug.deprecate(ts.factory.createJSDocCallbackTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocSignature` or the factory supplied by your transformation context instead. */
        ts.createJSDocSignature = ts.Debug.deprecate(ts.factory.createJSDocSignature, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocPropertyTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocPropertyTag = ts.Debug.deprecate(ts.factory.createJSDocPropertyTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocTypeLiteral` or the factory supplied by your transformation context instead. */
        ts.createJSDocTypeLiteral = ts.Debug.deprecate(ts.factory.createJSDocTypeLiteral, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocImplementsTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocImplementsTag = ts.Debug.deprecate(ts.factory.createJSDocImplementsTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocAuthorTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocAuthorTag = ts.Debug.deprecate(ts.factory.createJSDocAuthorTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocPublicTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocPublicTag = ts.Debug.deprecate(ts.factory.createJSDocPublicTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocPrivateTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocPrivateTag = ts.Debug.deprecate(ts.factory.createJSDocPrivateTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocProtectedTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocProtectedTag = ts.Debug.deprecate(ts.factory.createJSDocProtectedTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocReadonlyTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocReadonlyTag = ts.Debug.deprecate(ts.factory.createJSDocReadonlyTag, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocUnknownTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocTag = ts.Debug.deprecate(ts.factory.createJSDocUnknownTag, factoryDeprecation);
        /** @deprecated Use `factory.createJsxElement` or the factory supplied by your transformation context instead. */
        ts.createJsxElement = ts.Debug.deprecate(ts.factory.createJsxElement, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxElement` or the factory supplied by your transformation context instead. */
        ts.updateJsxElement = ts.Debug.deprecate(ts.factory.updateJsxElement, factoryDeprecation);
        /** @deprecated Use `factory.createJsxSelfClosingElement` or the factory supplied by your transformation context instead. */
        ts.createJsxSelfClosingElement = ts.Debug.deprecate(ts.factory.createJsxSelfClosingElement, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxSelfClosingElement` or the factory supplied by your transformation context instead. */
        ts.updateJsxSelfClosingElement = ts.Debug.deprecate(ts.factory.updateJsxSelfClosingElement, factoryDeprecation);
        /** @deprecated Use `factory.createJsxOpeningElement` or the factory supplied by your transformation context instead. */
        ts.createJsxOpeningElement = ts.Debug.deprecate(ts.factory.createJsxOpeningElement, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxOpeningElement` or the factory supplied by your transformation context instead. */
        ts.updateJsxOpeningElement = ts.Debug.deprecate(ts.factory.updateJsxOpeningElement, factoryDeprecation);
        /** @deprecated Use `factory.createJsxClosingElement` or the factory supplied by your transformation context instead. */
        ts.createJsxClosingElement = ts.Debug.deprecate(ts.factory.createJsxClosingElement, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxClosingElement` or the factory supplied by your transformation context instead. */
        ts.updateJsxClosingElement = ts.Debug.deprecate(ts.factory.updateJsxClosingElement, factoryDeprecation);
        /** @deprecated Use `factory.createJsxFragment` or the factory supplied by your transformation context instead. */
        ts.createJsxFragment = ts.Debug.deprecate(ts.factory.createJsxFragment, factoryDeprecation);
        /** @deprecated Use `factory.createJsxText` or the factory supplied by your transformation context instead. */
        ts.createJsxText = ts.Debug.deprecate(ts.factory.createJsxText, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxText` or the factory supplied by your transformation context instead. */
        ts.updateJsxText = ts.Debug.deprecate(ts.factory.updateJsxText, factoryDeprecation);
        /** @deprecated Use `factory.createJsxOpeningFragment` or the factory supplied by your transformation context instead. */
        ts.createJsxOpeningFragment = ts.Debug.deprecate(ts.factory.createJsxOpeningFragment, factoryDeprecation);
        /** @deprecated Use `factory.createJsxJsxClosingFragment` or the factory supplied by your transformation context instead. */
        ts.createJsxJsxClosingFragment = ts.Debug.deprecate(ts.factory.createJsxJsxClosingFragment, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxFragment` or the factory supplied by your transformation context instead. */
        ts.updateJsxFragment = ts.Debug.deprecate(ts.factory.updateJsxFragment, factoryDeprecation);
        /** @deprecated Use `factory.createJsxAttribute` or the factory supplied by your transformation context instead. */
        ts.createJsxAttribute = ts.Debug.deprecate(ts.factory.createJsxAttribute, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxAttribute` or the factory supplied by your transformation context instead. */
        ts.updateJsxAttribute = ts.Debug.deprecate(ts.factory.updateJsxAttribute, factoryDeprecation);
        /** @deprecated Use `factory.createJsxAttributes` or the factory supplied by your transformation context instead. */
        ts.createJsxAttributes = ts.Debug.deprecate(ts.factory.createJsxAttributes, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxAttributes` or the factory supplied by your transformation context instead. */
        ts.updateJsxAttributes = ts.Debug.deprecate(ts.factory.updateJsxAttributes, factoryDeprecation);
        /** @deprecated Use `factory.createJsxSpreadAttribute` or the factory supplied by your transformation context instead. */
        ts.createJsxSpreadAttribute = ts.Debug.deprecate(ts.factory.createJsxSpreadAttribute, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxSpreadAttribute` or the factory supplied by your transformation context instead. */
        ts.updateJsxSpreadAttribute = ts.Debug.deprecate(ts.factory.updateJsxSpreadAttribute, factoryDeprecation);
        /** @deprecated Use `factory.createJsxExpression` or the factory supplied by your transformation context instead. */
        ts.createJsxExpression = ts.Debug.deprecate(ts.factory.createJsxExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateJsxExpression` or the factory supplied by your transformation context instead. */
        ts.updateJsxExpression = ts.Debug.deprecate(ts.factory.updateJsxExpression, factoryDeprecation);
        /** @deprecated Use `factory.createCaseClause` or the factory supplied by your transformation context instead. */
        ts.createCaseClause = ts.Debug.deprecate(ts.factory.createCaseClause, factoryDeprecation);
        /** @deprecated Use `factory.updateCaseClause` or the factory supplied by your transformation context instead. */
        ts.updateCaseClause = ts.Debug.deprecate(ts.factory.updateCaseClause, factoryDeprecation);
        /** @deprecated Use `factory.createDefaultClause` or the factory supplied by your transformation context instead. */
        ts.createDefaultClause = ts.Debug.deprecate(ts.factory.createDefaultClause, factoryDeprecation);
        /** @deprecated Use `factory.updateDefaultClause` or the factory supplied by your transformation context instead. */
        ts.updateDefaultClause = ts.Debug.deprecate(ts.factory.updateDefaultClause, factoryDeprecation);
        /** @deprecated Use `factory.createHeritageClause` or the factory supplied by your transformation context instead. */
        ts.createHeritageClause = ts.Debug.deprecate(ts.factory.createHeritageClause, factoryDeprecation);
        /** @deprecated Use `factory.updateHeritageClause` or the factory supplied by your transformation context instead. */
        ts.updateHeritageClause = ts.Debug.deprecate(ts.factory.updateHeritageClause, factoryDeprecation);
        /** @deprecated Use `factory.createCatchClause` or the factory supplied by your transformation context instead. */
        ts.createCatchClause = ts.Debug.deprecate(ts.factory.createCatchClause, factoryDeprecation);
        /** @deprecated Use `factory.updateCatchClause` or the factory supplied by your transformation context instead. */
        ts.updateCatchClause = ts.Debug.deprecate(ts.factory.updateCatchClause, factoryDeprecation);
        /** @deprecated Use `factory.createPropertyAssignment` or the factory supplied by your transformation context instead. */
        ts.createPropertyAssignment = ts.Debug.deprecate(ts.factory.createPropertyAssignment, factoryDeprecation);
        /** @deprecated Use `factory.updatePropertyAssignment` or the factory supplied by your transformation context instead. */
        ts.updatePropertyAssignment = ts.Debug.deprecate(ts.factory.updatePropertyAssignment, factoryDeprecation);
        /** @deprecated Use `factory.createShorthandPropertyAssignment` or the factory supplied by your transformation context instead. */
        ts.createShorthandPropertyAssignment = ts.Debug.deprecate(ts.factory.createShorthandPropertyAssignment, factoryDeprecation);
        /** @deprecated Use `factory.updateShorthandPropertyAssignment` or the factory supplied by your transformation context instead. */
        ts.updateShorthandPropertyAssignment = ts.Debug.deprecate(ts.factory.updateShorthandPropertyAssignment, factoryDeprecation);
        /** @deprecated Use `factory.createSpreadAssignment` or the factory supplied by your transformation context instead. */
        ts.createSpreadAssignment = ts.Debug.deprecate(ts.factory.createSpreadAssignment, factoryDeprecation);
        /** @deprecated Use `factory.updateSpreadAssignment` or the factory supplied by your transformation context instead. */
        ts.updateSpreadAssignment = ts.Debug.deprecate(ts.factory.updateSpreadAssignment, factoryDeprecation);
        /** @deprecated Use `factory.createEnumMember` or the factory supplied by your transformation context instead. */
        ts.createEnumMember = ts.Debug.deprecate(ts.factory.createEnumMember, factoryDeprecation);
        /** @deprecated Use `factory.updateEnumMember` or the factory supplied by your transformation context instead. */
        ts.updateEnumMember = ts.Debug.deprecate(ts.factory.updateEnumMember, factoryDeprecation);
        /** @deprecated Use `factory.updateSourceFile` or the factory supplied by your transformation context instead. */
        ts.updateSourceFileNode = ts.Debug.deprecate(ts.factory.updateSourceFile, factoryDeprecation);
        /** @deprecated Use `factory.createNotEmittedStatement` or the factory supplied by your transformation context instead. */
        ts.createNotEmittedStatement = ts.Debug.deprecate(ts.factory.createNotEmittedStatement, factoryDeprecation);
        /** @deprecated Use `factory.createPartiallyEmittedExpression` or the factory supplied by your transformation context instead. */
        ts.createPartiallyEmittedExpression = ts.Debug.deprecate(ts.factory.createPartiallyEmittedExpression, factoryDeprecation);
        /** @deprecated Use `factory.updatePartiallyEmittedExpression` or the factory supplied by your transformation context instead. */
        ts.updatePartiallyEmittedExpression = ts.Debug.deprecate(ts.factory.updatePartiallyEmittedExpression, factoryDeprecation);
        /** @deprecated Use `factory.createCommaListExpression` or the factory supplied by your transformation context instead. */
        ts.createCommaList = ts.Debug.deprecate(ts.factory.createCommaListExpression, factoryDeprecation);
        /** @deprecated Use `factory.updateCommaListExpression` or the factory supplied by your transformation context instead. */
        ts.updateCommaList = ts.Debug.deprecate(ts.factory.updateCommaListExpression, factoryDeprecation);
        /** @deprecated Use `factory.createBundle` or the factory supplied by your transformation context instead. */
        ts.createBundle = ts.Debug.deprecate(ts.factory.createBundle, factoryDeprecation);
        /** @deprecated Use `factory.updateBundle` or the factory supplied by your transformation context instead. */
        ts.updateBundle = ts.Debug.deprecate(ts.factory.updateBundle, factoryDeprecation);
        /** @deprecated Use `factory.createImmediatelyInvokedFunctionExpression` or the factory supplied by your transformation context instead. */
        ts.createImmediatelyInvokedFunctionExpression = ts.Debug.deprecate(ts.factory.createImmediatelyInvokedFunctionExpression, factoryDeprecation);
        /** @deprecated Use `factory.createImmediatelyInvokedArrowFunction` or the factory supplied by your transformation context instead. */
        ts.createImmediatelyInvokedArrowFunction = ts.Debug.deprecate(ts.factory.createImmediatelyInvokedArrowFunction, factoryDeprecation);
        /** @deprecated Use `factory.createVoidZero` or the factory supplied by your transformation context instead. */
        ts.createVoidZero = ts.Debug.deprecate(ts.factory.createVoidZero, factoryDeprecation);
        /** @deprecated Use `factory.createExportDefault` or the factory supplied by your transformation context instead. */
        ts.createExportDefault = ts.Debug.deprecate(ts.factory.createExportDefault, factoryDeprecation);
        /** @deprecated Use `factory.createExternalModuleExport` or the factory supplied by your transformation context instead. */
        ts.createExternalModuleExport = ts.Debug.deprecate(ts.factory.createExternalModuleExport, factoryDeprecation);
        /** @deprecated Use `factory.createNamespaceExport` or the factory supplied by your transformation context instead. */
        ts.createNamespaceExport = ts.Debug.deprecate(ts.factory.createNamespaceExport, factoryDeprecation);
        /** @deprecated Use `factory.updateNamespaceExport` or the factory supplied by your transformation context instead. */
        ts.updateNamespaceExport = ts.Debug.deprecate(ts.factory.updateNamespaceExport, factoryDeprecation);
        /** @deprecated Use `factory.createToken` or the factory supplied by your transformation context instead. */
        ts.createToken = ts.Debug.deprecate(function createToken(kind) {
            return ts.factory.createToken(kind);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createIdentifier` or the factory supplied by your transformation context instead. */
        ts.createIdentifier = ts.Debug.deprecate(function createIdentifier(text) {
            return ts.factory.createIdentifier(text, /*typeArguments*/ undefined, /*originalKeywordKind*/ undefined);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createTempVariable` or the factory supplied by your transformation context instead. */
        ts.createTempVariable = ts.Debug.deprecate(function createTempVariable(recordTempVariable) {
            return ts.factory.createTempVariable(recordTempVariable, /*reserveInNestedScopes*/ undefined);
        }, factoryDeprecation);
        /** @deprecated Use `factory.getGeneratedNameForNode` or the factory supplied by your transformation context instead. */
        ts.getGeneratedNameForNode = ts.Debug.deprecate(function getGeneratedNameForNode(node) {
            return ts.factory.getGeneratedNameForNode(node, /*flags*/ undefined);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createUniqueName(text, GeneratedIdentifierFlags.Optimistic)` or the factory supplied by your transformation context instead. */
        ts.createOptimisticUniqueName = ts.Debug.deprecate(function createOptimisticUniqueName(text) {
            return ts.factory.createUniqueName(text, 16 /* Optimistic */);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createUniqueName(text, GeneratedIdentifierFlags.Optimistic | GeneratedIdentifierFlags.FileLevel)` or the factory supplied by your transformation context instead. */
        ts.createFileLevelUniqueName = ts.Debug.deprecate(function createFileLevelUniqueName(text) {
            return ts.factory.createUniqueName(text, 16 /* Optimistic */ | 32 /* FileLevel */);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createIndexSignature` or the factory supplied by your transformation context instead. */
        ts.createIndexSignature = ts.Debug.deprecate(function createIndexSignature(decorators, modifiers, parameters, type) {
            return ts.factory.createIndexSignature(decorators, modifiers, parameters, type);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createTypePredicateNode` or the factory supplied by your transformation context instead. */
        ts.createTypePredicateNode = ts.Debug.deprecate(function createTypePredicateNode(parameterName, type) {
            return ts.factory.createTypePredicateNode(/*assertsModifier*/ undefined, parameterName, type);
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateTypePredicateNode` or the factory supplied by your transformation context instead. */
        ts.updateTypePredicateNode = ts.Debug.deprecate(function updateTypePredicateNode(node, parameterName, type) {
            return ts.factory.updateTypePredicateNode(node, /*assertsModifier*/ undefined, parameterName, type);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createStringLiteral`, `factory.createStringLiteralFromNode`, `factory.createNumericLiteral`, `factory.createBigIntLiteral`, `factory.createTrue`, `factory.createFalse`, or the factory supplied by your transformation context instead. */
        ts.createLiteral = ts.Debug.deprecate(function createLiteral(value) {
            if (typeof value === "number") {
                return ts.factory.createNumericLiteral(value);
            }
            // eslint-disable-next-line no-in-operator
            if (typeof value === "object" && "base10Value" in value) { // PseudoBigInt
                return ts.factory.createBigIntLiteral(value);
            }
            if (typeof value === "boolean") {
                return value ? ts.factory.createTrue() : ts.factory.createFalse();
            }
            if (typeof value === "string") {
                return ts.factory.createStringLiteral(value, /*isSingleQuote*/ undefined);
            }
            return ts.factory.createStringLiteralFromNode(value);
        }, { since: "4.0", warnAfter: "4.1", message: "Use `factory.createStringLiteral`, `factory.createStringLiteralFromNode`, `factory.createNumericLiteral`, `factory.createBigIntLiteral`, `factory.createTrue`, `factory.createFalse`, or the factory supplied by your transformation context instead." });
        /** @deprecated Use `factory.createMethodSignature` or the factory supplied by your transformation context instead. */
        ts.createMethodSignature = ts.Debug.deprecate(function createMethodSignature(typeParameters, parameters, type, name, questionToken) {
            return ts.factory.createMethodSignature(/*modifiers*/ undefined, name, questionToken, typeParameters, parameters, type);
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateMethodSignature` or the factory supplied by your transformation context instead. */
        ts.updateMethodSignature = ts.Debug.deprecate(function updateMethodSignature(node, typeParameters, parameters, type, name, questionToken) {
            return ts.factory.updateMethodSignature(node, node.modifiers, name, questionToken, typeParameters, parameters, type);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createTypeOperatorNode` or the factory supplied by your transformation context instead. */
        ts.createTypeOperatorNode = ts.Debug.deprecate(function createTypeOperatorNode(operatorOrType, type) {
            var operator;
            if (type) {
                operator = operatorOrType;
            }
            else {
                type = operatorOrType;
                operator = 139 /* KeyOfKeyword */;
            }
            return ts.factory.createTypeOperatorNode(operator, type);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createTaggedTemplate` or the factory supplied by your transformation context instead. */
        ts.createTaggedTemplate = ts.Debug.deprecate(function createTaggedTemplate(tag, typeArgumentsOrTemplate, template) {
            var typeArguments;
            if (template) {
                typeArguments = typeArgumentsOrTemplate;
            }
            else {
                template = typeArgumentsOrTemplate;
            }
            return ts.factory.createTaggedTemplateExpression(tag, typeArguments, template);
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateTaggedTemplate` or the factory supplied by your transformation context instead. */
        ts.updateTaggedTemplate = ts.Debug.deprecate(function updateTaggedTemplate(node, tag, typeArgumentsOrTemplate, template) {
            var typeArguments;
            if (template) {
                typeArguments = typeArgumentsOrTemplate;
            }
            else {
                template = typeArgumentsOrTemplate;
            }
            return ts.factory.updateTaggedTemplateExpression(node, tag, typeArguments, template);
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateBinary` or the factory supplied by your transformation context instead. */
        ts.updateBinary = ts.Debug.deprecate(function updateBinary(node, left, right, operator) {
            if (operator === void 0) {
                operator = node.operatorToken;
            }
            if (typeof operator === "number") {
                operator = operator === node.operatorToken.kind ? node.operatorToken : ts.factory.createToken(operator);
            }
            return ts.factory.updateBinaryExpression(node, left, operator, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createConditional` or the factory supplied by your transformation context instead. */
        ts.createConditional = ts.Debug.deprecate(function createConditional(condition, questionTokenOrWhenTrue, whenTrueOrWhenFalse, colonToken, whenFalse) {
            return arguments.length === 5 ? ts.factory.createConditionalExpression(condition, questionTokenOrWhenTrue, whenTrueOrWhenFalse, colonToken, whenFalse) :
                arguments.length === 3 ? ts.factory.createConditionalExpression(condition, ts.factory.createToken(57 /* QuestionToken */), questionTokenOrWhenTrue, ts.factory.createToken(58 /* ColonToken */), whenTrueOrWhenFalse) :
                    ts.Debug.fail("Argument count mismatch");
        }, factoryDeprecation);
        /** @deprecated Use `factory.createYield` or the factory supplied by your transformation context instead. */
        ts.createYield = ts.Debug.deprecate(function createYield(asteriskTokenOrExpression, expression) {
            var asteriskToken;
            if (expression) {
                asteriskToken = asteriskTokenOrExpression;
            }
            else {
                expression = asteriskTokenOrExpression;
            }
            return ts.factory.createYieldExpression(asteriskToken, expression);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createClassExpression` or the factory supplied by your transformation context instead. */
        ts.createClassExpression = ts.Debug.deprecate(function createClassExpression(modifiers, name, typeParameters, heritageClauses, members) {
            return ts.factory.createClassExpression(/*decorators*/ undefined, modifiers, name, typeParameters, heritageClauses, members);
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateClassExpression` or the factory supplied by your transformation context instead. */
        ts.updateClassExpression = ts.Debug.deprecate(function updateClassExpression(node, modifiers, name, typeParameters, heritageClauses, members) {
            return ts.factory.updateClassExpression(node, /*decorators*/ undefined, modifiers, name, typeParameters, heritageClauses, members);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createPropertySignature` or the factory supplied by your transformation context instead. */
        ts.createPropertySignature = ts.Debug.deprecate(function createPropertySignature(modifiers, name, questionToken, type, initializer) {
            var node = ts.factory.createPropertySignature(modifiers, name, questionToken, type);
            node.initializer = initializer;
            return node;
        }, factoryDeprecation);
        /** @deprecated Use `factory.updatePropertySignature` or the factory supplied by your transformation context instead. */
        ts.updatePropertySignature = ts.Debug.deprecate(function updatePropertySignature(node, modifiers, name, questionToken, type, initializer) {
            var updated = ts.factory.updatePropertySignature(node, modifiers, name, questionToken, type);
            if (node.initializer !== initializer) {
                if (updated === node) {
                    updated = ts.factory.cloneNode(node);
                }
                updated.initializer = initializer;
            }
            return updated;
        }, factoryDeprecation);
        /** @deprecated Use `factory.createExpressionWithTypeArguments` or the factory supplied by your transformation context instead. */
        ts.createExpressionWithTypeArguments = ts.Debug.deprecate(function createExpressionWithTypeArguments(typeArguments, expression) {
            return ts.factory.createExpressionWithTypeArguments(expression, typeArguments);
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateExpressionWithTypeArguments` or the factory supplied by your transformation context instead. */
        ts.updateExpressionWithTypeArguments = ts.Debug.deprecate(function updateExpressionWithTypeArguments(node, typeArguments, expression) {
            return ts.factory.updateExpressionWithTypeArguments(node, expression, typeArguments);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createArrowFunction` or the factory supplied by your transformation context instead. */
        ts.createArrowFunction = ts.Debug.deprecate(function createArrowFunction(modifiers, typeParameters, parameters, type, equalsGreaterThanTokenOrBody, body) {
            return arguments.length === 6 ? ts.factory.createArrowFunction(modifiers, typeParameters, parameters, type, equalsGreaterThanTokenOrBody, body) :
                arguments.length === 5 ? ts.factory.createArrowFunction(modifiers, typeParameters, parameters, type, /*equalsGreaterThanToken*/ undefined, equalsGreaterThanTokenOrBody) :
                    ts.Debug.fail("Argument count mismatch");
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateArrowFunction` or the factory supplied by your transformation context instead. */
        ts.updateArrowFunction = ts.Debug.deprecate(function updateArrowFunction(node, modifiers, typeParameters, parameters, type, equalsGreaterThanTokenOrBody, body) {
            return arguments.length === 7 ? ts.factory.updateArrowFunction(node, modifiers, typeParameters, parameters, type, equalsGreaterThanTokenOrBody, body) :
                arguments.length === 6 ? ts.factory.updateArrowFunction(node, modifiers, typeParameters, parameters, type, node.equalsGreaterThanToken, equalsGreaterThanTokenOrBody) :
                    ts.Debug.fail("Argument count mismatch");
        }, factoryDeprecation);
        /** @deprecated Use `factory.createVariableDeclaration` or the factory supplied by your transformation context instead. */
        ts.createVariableDeclaration = ts.Debug.deprecate(function createVariableDeclaration(name, exclamationTokenOrType, typeOrInitializer, initializer) {
            return arguments.length === 4 ? ts.factory.createVariableDeclaration(name, exclamationTokenOrType, typeOrInitializer, initializer) :
                arguments.length >= 1 && arguments.length <= 3 ? ts.factory.createVariableDeclaration(name, /*exclamationToken*/ undefined, exclamationTokenOrType, typeOrInitializer) :
                    ts.Debug.fail("Argument count mismatch");
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateVariableDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateVariableDeclaration = ts.Debug.deprecate(function updateVariableDeclaration(node, name, exclamationTokenOrType, typeOrInitializer, initializer) {
            return arguments.length === 5 ? ts.factory.updateVariableDeclaration(node, name, exclamationTokenOrType, typeOrInitializer, initializer) :
                arguments.length === 4 ? ts.factory.updateVariableDeclaration(node, name, node.exclamationToken, exclamationTokenOrType, typeOrInitializer) :
                    ts.Debug.fail("Argument count mismatch");
        }, factoryDeprecation);
        /** @deprecated Use `factory.createImportClause` or the factory supplied by your transformation context instead. */
        ts.createImportClause = ts.Debug.deprecate(function createImportClause(name, namedBindings, isTypeOnly) {
            if (isTypeOnly === void 0) {
                isTypeOnly = false;
            }
            return ts.factory.createImportClause(isTypeOnly, name, namedBindings);
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateImportClause` or the factory supplied by your transformation context instead. */
        ts.updateImportClause = ts.Debug.deprecate(function updateImportClause(node, name, namedBindings, isTypeOnly) {
            return ts.factory.updateImportClause(node, isTypeOnly, name, namedBindings);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createExportDeclaration` or the factory supplied by your transformation context instead. */
        ts.createExportDeclaration = ts.Debug.deprecate(function createExportDeclaration(decorators, modifiers, exportClause, moduleSpecifier, isTypeOnly) {
            if (isTypeOnly === void 0) {
                isTypeOnly = false;
            }
            return ts.factory.createExportDeclaration(decorators, modifiers, isTypeOnly, exportClause, moduleSpecifier);
        }, factoryDeprecation);
        /** @deprecated Use `factory.updateExportDeclaration` or the factory supplied by your transformation context instead. */
        ts.updateExportDeclaration = ts.Debug.deprecate(function updateExportDeclaration(node, decorators, modifiers, exportClause, moduleSpecifier, isTypeOnly) {
            return ts.factory.updateExportDeclaration(node, decorators, modifiers, isTypeOnly, exportClause, moduleSpecifier);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createJSDocParameterTag` or the factory supplied by your transformation context instead. */
        ts.createJSDocParamTag = ts.Debug.deprecate(function createJSDocParamTag(name, isBracketed, typeExpression, comment) {
            return ts.factory.createJSDocParameterTag(/*tagName*/ undefined, name, isBracketed, typeExpression, /*isNameFirst*/ false, comment ? ts.factory.createNodeArray([ts.factory.createJSDocText(comment)]) : undefined);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createComma` or the factory supplied by your transformation context instead. */
        ts.createComma = ts.Debug.deprecate(function createComma(left, right) {
            return ts.factory.createComma(left, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createLessThan` or the factory supplied by your transformation context instead. */
        ts.createLessThan = ts.Debug.deprecate(function createLessThan(left, right) {
            return ts.factory.createLessThan(left, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createAssignment` or the factory supplied by your transformation context instead. */
        ts.createAssignment = ts.Debug.deprecate(function createAssignment(left, right) {
            return ts.factory.createAssignment(left, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createStrictEquality` or the factory supplied by your transformation context instead. */
        ts.createStrictEquality = ts.Debug.deprecate(function createStrictEquality(left, right) {
            return ts.factory.createStrictEquality(left, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createStrictInequality` or the factory supplied by your transformation context instead. */
        ts.createStrictInequality = ts.Debug.deprecate(function createStrictInequality(left, right) {
            return ts.factory.createStrictInequality(left, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createAdd` or the factory supplied by your transformation context instead. */
        ts.createAdd = ts.Debug.deprecate(function createAdd(left, right) {
            return ts.factory.createAdd(left, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createSubtract` or the factory supplied by your transformation context instead. */
        ts.createSubtract = ts.Debug.deprecate(function createSubtract(left, right) {
            return ts.factory.createSubtract(left, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createLogicalAnd` or the factory supplied by your transformation context instead. */
        ts.createLogicalAnd = ts.Debug.deprecate(function createLogicalAnd(left, right) {
            return ts.factory.createLogicalAnd(left, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createLogicalOr` or the factory supplied by your transformation context instead. */
        ts.createLogicalOr = ts.Debug.deprecate(function createLogicalOr(left, right) {
            return ts.factory.createLogicalOr(left, right);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createPostfixIncrement` or the factory supplied by your transformation context instead. */
        ts.createPostfixIncrement = ts.Debug.deprecate(function createPostfixIncrement(operand) {
            return ts.factory.createPostfixIncrement(operand);
        }, factoryDeprecation);
        /** @deprecated Use `factory.createLogicalNot` or the factory supplied by your transformation context instead. */
        ts.createLogicalNot = ts.Debug.deprecate(function createLogicalNot(operand) {
            return ts.factory.createLogicalNot(operand);
        }, factoryDeprecation);
        /** @deprecated Use an appropriate `factory` method instead. */
        ts.createNode = ts.Debug.deprecate(function createNode(kind, pos, end) {
            if (pos === void 0) {
                pos = 0;
            }
            if (end === void 0) {
                end = 0;
            }
            return ts.setTextRangePosEnd(kind === 300 /* SourceFile */ ? ts.parseBaseNodeFactory.createBaseSourceFileNode(kind) :
                kind === 79 /* Identifier */ ? ts.parseBaseNodeFactory.createBaseIdentifierNode(kind) :
                    kind === 80 /* PrivateIdentifier */ ? ts.parseBaseNodeFactory.createBasePrivateIdentifierNode(kind) :
                        !ts.isNodeKind(kind) ? ts.parseBaseNodeFactory.createBaseTokenNode(kind) :
                            ts.parseBaseNodeFactory.createBaseNode(kind), pos, end);
        }, { since: "4.0", warnAfter: "4.1", message: "Use an appropriate `factory` method instead." });
        /**
         * Creates a shallow, memberwise clone of a node ~for mutation~ with its `pos`, `end`, and `parent` set.
         *
         * NOTE: It is unsafe to change any properties of a `Node` that relate to its AST children, as those changes won't be
         * captured with respect to transformations.
         *
         * @deprecated Use an appropriate `factory.update...` method instead, use `setCommentRange` or `setSourceMapRange`, and avoid setting `parent`.
         */
        ts.getMutableClone = ts.Debug.deprecate(function getMutableClone(node) {
            var clone = ts.factory.cloneNode(node);
            ts.setTextRange(clone, node);
            ts.setParent(clone, node.parent);
            return clone;
        }, { since: "4.0", warnAfter: "4.1", message: "Use an appropriate `factory.update...` method instead, use `setCommentRange` or `setSourceMapRange`, and avoid setting `parent`." });
        // #endregion Node Factory top-level exports
        // DEPRECATION: Renamed node tests
        // DEPRECATION PLAN:
        //     - soft: 4.0
        //     - warn: 4.1
        //     - error: TBD
        // #region Renamed node Tests
        /** @deprecated Use `isTypeAssertionExpression` instead. */
        ts.isTypeAssertion = ts.Debug.deprecate(function isTypeAssertion(node) {
            return node.kind === 209 /* TypeAssertionExpression */;
        }, {
            since: "4.0",
            warnAfter: "4.1",
            message: "Use `isTypeAssertionExpression` instead."
        });
        // #endregion
        // DEPRECATION: Renamed node tests
        // DEPRECATION PLAN:
        //     - soft: 4.2
        //     - warn: 4.3
        //     - error: TBD
        // #region Renamed node Tests
        /**
         * @deprecated Use `isMemberName` instead.
         */
        ts.isIdentifierOrPrivateIdentifier = ts.Debug.deprecate(function isIdentifierOrPrivateIdentifier(node) {
            return ts.isMemberName(node);
        }, {
            since: "4.2",
            warnAfter: "4.3",
            message: "Use `isMemberName` instead."
        });
        // #endregion Renamed node Tests
    })(ts || (ts = {}));
},
/* source-map-support/source-map-support.js */ function _(require, module, exports, __esModule, __esExport) {
    var SourceMapConsumer = require(23) /* source-map */.SourceMapConsumer;
    var path = require('path');
    var fs;
    try {
        fs = require('fs');
        if (!fs.existsSync || !fs.readFileSync) {
            // fs doesn't have all methods we need
            fs = null;
        }
    }
    catch (err) {
        /* nop */
    }
    var bufferFrom = require(34) /* buffer-from */;
    /**
     * Requires a module which is protected against bundler minification.
     *
     * @param {NodeModule} mod
     * @param {string} request
     */
    function dynamicRequire(mod, request) {
        return mod.require(request);
    }
    // Only install once if called multiple times
    var errorFormatterInstalled = false;
    var uncaughtShimInstalled = false;
    // If true, the caches are reset before a stack trace formatting operation
    var emptyCacheBetweenOperations = false;
    // Supports {browser, node, auto}
    var environment = "auto";
    // Maps a file path to a string containing the file contents
    var fileContentsCache = {};
    // Maps a file path to a source map for that file
    var sourceMapCache = {};
    // Regex for detecting source maps
    var reSourceMap = /^data:application\/json[^,]+base64,/;
    // Priority list of retrieve handlers
    var retrieveFileHandlers = [];
    var retrieveMapHandlers = [];
    function isInBrowser() {
        if (environment === "browser")
            return true;
        if (environment === "node")
            return false;
        return ((typeof window !== 'undefined') && (typeof XMLHttpRequest === 'function') && !(window.require && window.module && window.process && window.process.type === "renderer"));
    }
    function hasGlobalProcessEventEmitter() {
        return ((typeof process === 'object') && (process !== null) && (typeof process.on === 'function'));
    }
    function handlerExec(list) {
        return function (arg) {
            for (var i = 0; i < list.length; i++) {
                var ret = list[i](arg);
                if (ret) {
                    return ret;
                }
            }
            return null;
        };
    }
    var retrieveFile = handlerExec(retrieveFileHandlers);
    retrieveFileHandlers.push(function (path) {
        // Trim the path to make sure there is no extra whitespace.
        path = path.trim();
        if (/^file:/.test(path)) {
            // existsSync/readFileSync can't handle file protocol, but once stripped, it works
            path = path.replace(/file:\/\/\/(\w:)?/, function (protocol, drive) {
                return drive ?
                    '' : // file:///C:/dir/file -> C:/dir/file
                    '/'; // file:///root-dir/file -> /root-dir/file
            });
        }
        if (path in fileContentsCache) {
            return fileContentsCache[path];
        }
        var contents = '';
        try {
            if (!fs) {
                // Use SJAX if we are in the browser
                var xhr = new XMLHttpRequest();
                xhr.open('GET', path, /** async */ false);
                xhr.send(null);
                if (xhr.readyState === 4 && xhr.status === 200) {
                    contents = xhr.responseText;
                }
            }
            else if (fs.existsSync(path)) {
                // Otherwise, use the filesystem
                contents = fs.readFileSync(path, 'utf8');
            }
        }
        catch (er) {
            /* ignore any errors */
        }
        return fileContentsCache[path] = contents;
    });
    // Support URLs relative to a directory, but be careful about a protocol prefix
    // in case we are in the browser (i.e. directories may start with "http://" or "file:///")
    function supportRelativeURL(file, url) {
        if (!file)
            return url;
        var dir = path.dirname(file);
        var match = /^\w+:\/\/[^\/]*/.exec(dir);
        var protocol = match ? match[0] : '';
        var startPath = dir.slice(protocol.length);
        if (protocol && /^\/\w\:/.test(startPath)) {
            // handle file:///C:/ paths
            protocol += '/';
            return protocol + path.resolve(dir.slice(protocol.length), url).replace(/\\/g, '/');
        }
        return protocol + path.resolve(dir.slice(protocol.length), url);
    }
    function retrieveSourceMapURL(source) {
        var fileData;
        if (isInBrowser()) {
            try {
                var xhr = new XMLHttpRequest();
                xhr.open('GET', source, false);
                xhr.send(null);
                fileData = xhr.readyState === 4 ? xhr.responseText : null;
                // Support providing a sourceMappingURL via the SourceMap header
                var sourceMapHeader = xhr.getResponseHeader("SourceMap") ||
                    xhr.getResponseHeader("X-SourceMap");
                if (sourceMapHeader) {
                    return sourceMapHeader;
                }
            }
            catch (e) {
            }
        }
        // Get the URL of the source map
        fileData = retrieveFile(source);
        var re = /(?:\/\/[@#][\s]*sourceMappingURL=([^\s'"]+)[\s]*$)|(?:\/\*[@#][\s]*sourceMappingURL=([^\s*'"]+)[\s]*(?:\*\/)[\s]*$)/mg;
        // Keep executing the search to find the *last* sourceMappingURL to avoid
        // picking up sourceMappingURLs from comments, strings, etc.
        var lastMatch, match;
        while (match = re.exec(fileData))
            lastMatch = match;
        if (!lastMatch)
            return null;
        return lastMatch[1];
    }
    ;
    // Can be overridden by the retrieveSourceMap option to install. Takes a
    // generated source filename; returns a {map, optional url} object, or null if
    // there is no source map.  The map field may be either a string or the parsed
    // JSON object (ie, it must be a valid argument to the SourceMapConsumer
    // constructor).
    var retrieveSourceMap = handlerExec(retrieveMapHandlers);
    retrieveMapHandlers.push(function (source) {
        var sourceMappingURL = retrieveSourceMapURL(source);
        if (!sourceMappingURL)
            return null;
        // Read the contents of the source map
        var sourceMapData;
        if (reSourceMap.test(sourceMappingURL)) {
            // Support source map URL as a data url
            var rawData = sourceMappingURL.slice(sourceMappingURL.indexOf(',') + 1);
            sourceMapData = bufferFrom(rawData, "base64").toString();
            sourceMappingURL = source;
        }
        else {
            // Support source map URLs relative to the source URL
            sourceMappingURL = supportRelativeURL(source, sourceMappingURL);
            sourceMapData = retrieveFile(sourceMappingURL);
        }
        if (!sourceMapData) {
            return null;
        }
        return {
            url: sourceMappingURL,
            map: sourceMapData
        };
    });
    function mapSourcePosition(position) {
        var sourceMap = sourceMapCache[position.source];
        if (!sourceMap) {
            // Call the (overrideable) retrieveSourceMap function to get the source map.
            var urlAndMap = retrieveSourceMap(position.source);
            if (urlAndMap) {
                sourceMap = sourceMapCache[position.source] = {
                    url: urlAndMap.url,
                    map: new SourceMapConsumer(urlAndMap.map)
                };
                // Load all sources stored inline with the source map into the file cache
                // to pretend like they are already loaded. They may not exist on disk.
                if (sourceMap.map.sourcesContent) {
                    sourceMap.map.sources.forEach(function (source, i) {
                        var contents = sourceMap.map.sourcesContent[i];
                        if (contents) {
                            var url = supportRelativeURL(sourceMap.url, source);
                            fileContentsCache[url] = contents;
                        }
                    });
                }
            }
            else {
                sourceMap = sourceMapCache[position.source] = {
                    url: null,
                    map: null
                };
            }
        }
        // Resolve the source URL relative to the URL of the source map
        if (sourceMap && sourceMap.map && typeof sourceMap.map.originalPositionFor === 'function') {
            var originalPosition = sourceMap.map.originalPositionFor(position);
            // Only return the original position if a matching line was found. If no
            // matching line is found then we return position instead, which will cause
            // the stack trace to print the path and line for the compiled file. It is
            // better to give a precise location in the compiled file than a vague
            // location in the original file.
            if (originalPosition.source !== null) {
                originalPosition.source = supportRelativeURL(sourceMap.url, originalPosition.source);
                return originalPosition;
            }
        }
        return position;
    }
    // Parses code generated by FormatEvalOrigin(), a function inside V8:
    // https://code.google.com/p/v8/source/browse/trunk/src/messages.js
    function mapEvalOrigin(origin) {
        // Most eval() calls are in this format
        var match = /^eval at ([^(]+) \((.+):(\d+):(\d+)\)$/.exec(origin);
        if (match) {
            var position = mapSourcePosition({
                source: match[2],
                line: +match[3],
                column: match[4] - 1
            });
            return 'eval at ' + match[1] + ' (' + position.source + ':' +
                position.line + ':' + (position.column + 1) + ')';
        }
        // Parse nested eval() calls using recursion
        match = /^eval at ([^(]+) \((.+)\)$/.exec(origin);
        if (match) {
            return 'eval at ' + match[1] + ' (' + mapEvalOrigin(match[2]) + ')';
        }
        // Make sure we still return useful information if we didn't find anything
        return origin;
    }
    // This is copied almost verbatim from the V8 source code at
    // https://code.google.com/p/v8/source/browse/trunk/src/messages.js. The
    // implementation of wrapCallSite() used to just forward to the actual source
    // code of CallSite.prototype.toString but unfortunately a new release of V8
    // did something to the prototype chain and broke the shim. The only fix I
    // could find was copy/paste.
    function CallSiteToString() {
        var fileName;
        var fileLocation = "";
        if (this.isNative()) {
            fileLocation = "native";
        }
        else {
            fileName = this.getScriptNameOrSourceURL();
            if (!fileName && this.isEval()) {
                fileLocation = this.getEvalOrigin();
                fileLocation += ", "; // Expecting source position to follow.
            }
            if (fileName) {
                fileLocation += fileName;
            }
            else {
                // Source code does not originate from a file and is not native, but we
                // can still get the source position inside the source string, e.g. in
                // an eval string.
                fileLocation += "<anonymous>";
            }
            var lineNumber = this.getLineNumber();
            if (lineNumber != null) {
                fileLocation += ":" + lineNumber;
                var columnNumber = this.getColumnNumber();
                if (columnNumber) {
                    fileLocation += ":" + columnNumber;
                }
            }
        }
        var line = "";
        var functionName = this.getFunctionName();
        var addSuffix = true;
        var isConstructor = this.isConstructor();
        var isMethodCall = !(this.isToplevel() || isConstructor);
        if (isMethodCall) {
            var typeName = this.getTypeName();
            // Fixes shim to be backward compatable with Node v0 to v4
            if (typeName === "[object Object]") {
                typeName = "null";
            }
            var methodName = this.getMethodName();
            if (functionName) {
                if (typeName && functionName.indexOf(typeName) != 0) {
                    line += typeName + ".";
                }
                line += functionName;
                if (methodName && functionName.indexOf("." + methodName) != functionName.length - methodName.length - 1) {
                    line += " [as " + methodName + "]";
                }
            }
            else {
                line += typeName + "." + (methodName || "<anonymous>");
            }
        }
        else if (isConstructor) {
            line += "new " + (functionName || "<anonymous>");
        }
        else if (functionName) {
            line += functionName;
        }
        else {
            line += fileLocation;
            addSuffix = false;
        }
        if (addSuffix) {
            line += " (" + fileLocation + ")";
        }
        return line;
    }
    function cloneCallSite(frame) {
        var object = {};
        Object.getOwnPropertyNames(Object.getPrototypeOf(frame)).forEach(function (name) {
            object[name] = /^(?:is|get)/.test(name) ? function () { return frame[name].call(frame); } : frame[name];
        });
        object.toString = CallSiteToString;
        return object;
    }
    function wrapCallSite(frame, state) {
        // provides interface backward compatibility
        if (state === undefined) {
            state = { nextPosition: null, curPosition: null };
        }
        if (frame.isNative()) {
            state.curPosition = null;
            return frame;
        }
        // Most call sites will return the source file from getFileName(), but code
        // passed to eval() ending in "//# sourceURL=..." will return the source file
        // from getScriptNameOrSourceURL() instead
        var source = frame.getFileName() || frame.getScriptNameOrSourceURL();
        if (source) {
            var line = frame.getLineNumber();
            var column = frame.getColumnNumber() - 1;
            // Fix position in Node where some (internal) code is prepended.
            // See https://github.com/evanw/node-source-map-support/issues/36
            // Header removed in node at ^10.16 || >=11.11.0
            // v11 is not an LTS candidate, we can just test the one version with it.
            // Test node versions for: 10.16-19, 10.20+, 12-19, 20-99, 100+, or 11.11
            var noHeader = /^v(10\.1[6-9]|10\.[2-9][0-9]|10\.[0-9]{3,}|1[2-9]\d*|[2-9]\d|\d{3,}|11\.11)/;
            var headerLength = noHeader.test(process.version) ? 0 : 62;
            if (line === 1 && column > headerLength && !isInBrowser() && !frame.isEval()) {
                column -= headerLength;
            }
            var position = mapSourcePosition({
                source: source,
                line: line,
                column: column
            });
            state.curPosition = position;
            frame = cloneCallSite(frame);
            var originalFunctionName = frame.getFunctionName;
            frame.getFunctionName = function () {
                if (state.nextPosition == null) {
                    return originalFunctionName();
                }
                return state.nextPosition.name || originalFunctionName();
            };
            frame.getFileName = function () { return position.source; };
            frame.getLineNumber = function () { return position.line; };
            frame.getColumnNumber = function () { return position.column + 1; };
            frame.getScriptNameOrSourceURL = function () { return position.source; };
            return frame;
        }
        // Code called using eval() needs special handling
        var origin = frame.isEval() && frame.getEvalOrigin();
        if (origin) {
            origin = mapEvalOrigin(origin);
            frame = cloneCallSite(frame);
            frame.getEvalOrigin = function () { return origin; };
            return frame;
        }
        // If we get here then we were unable to change the source position
        return frame;
    }
    // This function is part of the V8 stack trace API, for more info see:
    // https://v8.dev/docs/stack-trace-api
    function prepareStackTrace(error, stack) {
        if (emptyCacheBetweenOperations) {
            fileContentsCache = {};
            sourceMapCache = {};
        }
        var name = error.name || 'Error';
        var message = error.message || '';
        var errorString = name + ": " + message;
        var state = { nextPosition: null, curPosition: null };
        var processedStack = [];
        for (var i = stack.length - 1; i >= 0; i--) {
            processedStack.push('\n    at ' + wrapCallSite(stack[i], state));
            state.nextPosition = state.curPosition;
        }
        state.curPosition = state.nextPosition = null;
        return errorString + processedStack.reverse().join('');
    }
    // Generate position and snippet of original source with pointer
    function getErrorSource(error) {
        var match = /\n    at [^(]+ \((.*):(\d+):(\d+)\)/.exec(error.stack);
        if (match) {
            var source = match[1];
            var line = +match[2];
            var column = +match[3];
            // Support the inline sourceContents inside the source map
            var contents = fileContentsCache[source];
            // Support files on disk
            if (!contents && fs && fs.existsSync(source)) {
                try {
                    contents = fs.readFileSync(source, 'utf8');
                }
                catch (er) {
                    contents = '';
                }
            }
            // Format the line from the original source code like node does
            if (contents) {
                var code = contents.split(/(?:\r\n|\r|\n)/)[line - 1];
                if (code) {
                    return source + ':' + line + '\n' + code + '\n' +
                        new Array(column).join(' ') + '^';
                }
            }
        }
        return null;
    }
    function printErrorAndExit(error) {
        var source = getErrorSource(error);
        // Ensure error is printed synchronously and not truncated
        if (process.stderr._handle && process.stderr._handle.setBlocking) {
            process.stderr._handle.setBlocking(true);
        }
        if (source) {
            console.error();
            console.error(source);
        }
        console.error(error.stack);
        process.exit(1);
    }
    function shimEmitUncaughtException() {
        var origEmit = process.emit;
        process.emit = function (type) {
            if (type === 'uncaughtException') {
                var hasStack = (arguments[1] && arguments[1].stack);
                var hasListeners = (this.listeners(type).length > 0);
                if (hasStack && !hasListeners) {
                    return printErrorAndExit(arguments[1]);
                }
            }
            return origEmit.apply(this, arguments);
        };
    }
    var originalRetrieveFileHandlers = retrieveFileHandlers.slice(0);
    var originalRetrieveMapHandlers = retrieveMapHandlers.slice(0);
    exports.wrapCallSite = wrapCallSite;
    exports.getErrorSource = getErrorSource;
    exports.mapSourcePosition = mapSourcePosition;
    exports.retrieveSourceMap = retrieveSourceMap;
    exports.install = function (options) {
        options = options || {};
        if (options.environment) {
            environment = options.environment;
            if (["node", "browser", "auto"].indexOf(environment) === -1) {
                throw new Error("environment " + environment + " was unknown. Available options are {auto, browser, node}");
            }
        }
        // Allow sources to be found by methods other than reading the files
        // directly from disk.
        if (options.retrieveFile) {
            if (options.overrideRetrieveFile) {
                retrieveFileHandlers.length = 0;
            }
            retrieveFileHandlers.unshift(options.retrieveFile);
        }
        // Allow source maps to be found by methods other than reading the files
        // directly from disk.
        if (options.retrieveSourceMap) {
            if (options.overrideRetrieveSourceMap) {
                retrieveMapHandlers.length = 0;
            }
            retrieveMapHandlers.unshift(options.retrieveSourceMap);
        }
        // Support runtime transpilers that include inline source maps
        if (options.hookRequire && !isInBrowser()) {
            // Use dynamicRequire to avoid including in browser bundles
            var Module = dynamicRequire(module, 'module');
            var $compile = Module.prototype._compile;
            if (!$compile.__sourceMapSupport) {
                Module.prototype._compile = function (content, filename) {
                    fileContentsCache[filename] = content;
                    sourceMapCache[filename] = undefined;
                    return $compile.call(this, content, filename);
                };
                Module.prototype._compile.__sourceMapSupport = true;
            }
        }
        // Configure options
        if (!emptyCacheBetweenOperations) {
            emptyCacheBetweenOperations = 'emptyCacheBetweenOperations' in options ?
                options.emptyCacheBetweenOperations : false;
        }
        // Install the error reformatter
        if (!errorFormatterInstalled) {
            errorFormatterInstalled = true;
            Error.prepareStackTrace = prepareStackTrace;
        }
        if (!uncaughtShimInstalled) {
            var installHandler = 'handleUncaughtExceptions' in options ?
                options.handleUncaughtExceptions : true;
            // Do not override 'uncaughtException' with our own handler in Node.js
            // Worker threads. Workers pass the error to the main thread as an event,
            // rather than printing something to stderr and exiting.
            try {
                // We need to use `dynamicRequire` because `require` on it's own will be optimized by WebPack/Browserify.
                var worker_threads = dynamicRequire(module, 'worker_threads');
                if (worker_threads.isMainThread === false) {
                    installHandler = false;
                }
            }
            catch (e) { }
            // Provide the option to not install the uncaught exception handler. This is
            // to support other uncaught exception handlers (in test frameworks, for
            // example). If this handler is not installed and there are no other uncaught
            // exception handlers, uncaught exceptions will be caught by node's built-in
            // exception handler and the process will still be terminated. However, the
            // generated JavaScript code will be shown above the stack trace instead of
            // the original source code.
            if (installHandler && hasGlobalProcessEventEmitter()) {
                uncaughtShimInstalled = true;
                shimEmitUncaughtException();
            }
        }
    };
    exports.resetRetrieveHandlers = function () {
        retrieveFileHandlers.length = 0;
        retrieveMapHandlers.length = 0;
        retrieveFileHandlers = originalRetrieveFileHandlers.slice(0);
        retrieveMapHandlers = originalRetrieveMapHandlers.slice(0);
        retrieveSourceMap = handlerExec(retrieveMapHandlers);
        retrieveFile = handlerExec(retrieveFileHandlers);
    };
},
/* source-map/source-map.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
     * Copyright 2009-2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE.txt or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    exports.SourceMapGenerator = require(24) /* ./lib/source-map-generator */.SourceMapGenerator;
    exports.SourceMapConsumer = require(30) /* ./lib/source-map-consumer */.SourceMapConsumer;
    exports.SourceNode = require(33) /* ./lib/source-node */.SourceNode;
},
/* source-map/lib/source-map-generator.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    var base64VLQ = require(25) /* ./base64-vlq */;
    var util = require(27) /* ./util */;
    var ArraySet = require(28) /* ./array-set */.ArraySet;
    var MappingList = require(29) /* ./mapping-list */.MappingList;
    /**
     * An instance of the SourceMapGenerator represents a source map which is
     * being built incrementally. You may pass an object with the following
     * properties:
     *
     *   - file: The filename of the generated source.
     *   - sourceRoot: A root for all relative URLs in this source map.
     */
    function SourceMapGenerator(aArgs) {
        if (!aArgs) {
            aArgs = {};
        }
        this._file = util.getArg(aArgs, 'file', null);
        this._sourceRoot = util.getArg(aArgs, 'sourceRoot', null);
        this._skipValidation = util.getArg(aArgs, 'skipValidation', false);
        this._sources = new ArraySet();
        this._names = new ArraySet();
        this._mappings = new MappingList();
        this._sourcesContents = null;
    }
    SourceMapGenerator.prototype._version = 3;
    /**
     * Creates a new SourceMapGenerator based on a SourceMapConsumer
     *
     * @param aSourceMapConsumer The SourceMap.
     */
    SourceMapGenerator.fromSourceMap =
        function SourceMapGenerator_fromSourceMap(aSourceMapConsumer) {
            var sourceRoot = aSourceMapConsumer.sourceRoot;
            var generator = new SourceMapGenerator({
                file: aSourceMapConsumer.file,
                sourceRoot: sourceRoot
            });
            aSourceMapConsumer.eachMapping(function (mapping) {
                var newMapping = {
                    generated: {
                        line: mapping.generatedLine,
                        column: mapping.generatedColumn
                    }
                };
                if (mapping.source != null) {
                    newMapping.source = mapping.source;
                    if (sourceRoot != null) {
                        newMapping.source = util.relative(sourceRoot, newMapping.source);
                    }
                    newMapping.original = {
                        line: mapping.originalLine,
                        column: mapping.originalColumn
                    };
                    if (mapping.name != null) {
                        newMapping.name = mapping.name;
                    }
                }
                generator.addMapping(newMapping);
            });
            aSourceMapConsumer.sources.forEach(function (sourceFile) {
                var sourceRelative = sourceFile;
                if (sourceRoot !== null) {
                    sourceRelative = util.relative(sourceRoot, sourceFile);
                }
                if (!generator._sources.has(sourceRelative)) {
                    generator._sources.add(sourceRelative);
                }
                var content = aSourceMapConsumer.sourceContentFor(sourceFile);
                if (content != null) {
                    generator.setSourceContent(sourceFile, content);
                }
            });
            return generator;
        };
    /**
     * Add a single mapping from original source line and column to the generated
     * source's line and column for this source map being created. The mapping
     * object should have the following properties:
     *
     *   - generated: An object with the generated line and column positions.
     *   - original: An object with the original line and column positions.
     *   - source: The original source file (relative to the sourceRoot).
     *   - name: An optional original token name for this mapping.
     */
    SourceMapGenerator.prototype.addMapping =
        function SourceMapGenerator_addMapping(aArgs) {
            var generated = util.getArg(aArgs, 'generated');
            var original = util.getArg(aArgs, 'original', null);
            var source = util.getArg(aArgs, 'source', null);
            var name = util.getArg(aArgs, 'name', null);
            if (!this._skipValidation) {
                this._validateMapping(generated, original, source, name);
            }
            if (source != null) {
                source = String(source);
                if (!this._sources.has(source)) {
                    this._sources.add(source);
                }
            }
            if (name != null) {
                name = String(name);
                if (!this._names.has(name)) {
                    this._names.add(name);
                }
            }
            this._mappings.add({
                generatedLine: generated.line,
                generatedColumn: generated.column,
                originalLine: original != null && original.line,
                originalColumn: original != null && original.column,
                source: source,
                name: name
            });
        };
    /**
     * Set the source content for a source file.
     */
    SourceMapGenerator.prototype.setSourceContent =
        function SourceMapGenerator_setSourceContent(aSourceFile, aSourceContent) {
            var source = aSourceFile;
            if (this._sourceRoot != null) {
                source = util.relative(this._sourceRoot, source);
            }
            if (aSourceContent != null) {
                // Add the source content to the _sourcesContents map.
                // Create a new _sourcesContents map if the property is null.
                if (!this._sourcesContents) {
                    this._sourcesContents = Object.create(null);
                }
                this._sourcesContents[util.toSetString(source)] = aSourceContent;
            }
            else if (this._sourcesContents) {
                // Remove the source file from the _sourcesContents map.
                // If the _sourcesContents map is empty, set the property to null.
                delete this._sourcesContents[util.toSetString(source)];
                if (Object.keys(this._sourcesContents).length === 0) {
                    this._sourcesContents = null;
                }
            }
        };
    /**
     * Applies the mappings of a sub-source-map for a specific source file to the
     * source map being generated. Each mapping to the supplied source file is
     * rewritten using the supplied source map. Note: The resolution for the
     * resulting mappings is the minimium of this map and the supplied map.
     *
     * @param aSourceMapConsumer The source map to be applied.
     * @param aSourceFile Optional. The filename of the source file.
     *        If omitted, SourceMapConsumer's file property will be used.
     * @param aSourceMapPath Optional. The dirname of the path to the source map
     *        to be applied. If relative, it is relative to the SourceMapConsumer.
     *        This parameter is needed when the two source maps aren't in the same
     *        directory, and the source map to be applied contains relative source
     *        paths. If so, those relative source paths need to be rewritten
     *        relative to the SourceMapGenerator.
     */
    SourceMapGenerator.prototype.applySourceMap =
        function SourceMapGenerator_applySourceMap(aSourceMapConsumer, aSourceFile, aSourceMapPath) {
            var sourceFile = aSourceFile;
            // If aSourceFile is omitted, we will use the file property of the SourceMap
            if (aSourceFile == null) {
                if (aSourceMapConsumer.file == null) {
                    throw new Error('SourceMapGenerator.prototype.applySourceMap requires either an explicit source file, ' +
                        'or the source map\'s "file" property. Both were omitted.');
                }
                sourceFile = aSourceMapConsumer.file;
            }
            var sourceRoot = this._sourceRoot;
            // Make "sourceFile" relative if an absolute Url is passed.
            if (sourceRoot != null) {
                sourceFile = util.relative(sourceRoot, sourceFile);
            }
            // Applying the SourceMap can add and remove items from the sources and
            // the names array.
            var newSources = new ArraySet();
            var newNames = new ArraySet();
            // Find mappings for the "sourceFile"
            this._mappings.unsortedForEach(function (mapping) {
                if (mapping.source === sourceFile && mapping.originalLine != null) {
                    // Check if it can be mapped by the source map, then update the mapping.
                    var original = aSourceMapConsumer.originalPositionFor({
                        line: mapping.originalLine,
                        column: mapping.originalColumn
                    });
                    if (original.source != null) {
                        // Copy mapping
                        mapping.source = original.source;
                        if (aSourceMapPath != null) {
                            mapping.source = util.join(aSourceMapPath, mapping.source);
                        }
                        if (sourceRoot != null) {
                            mapping.source = util.relative(sourceRoot, mapping.source);
                        }
                        mapping.originalLine = original.line;
                        mapping.originalColumn = original.column;
                        if (original.name != null) {
                            mapping.name = original.name;
                        }
                    }
                }
                var source = mapping.source;
                if (source != null && !newSources.has(source)) {
                    newSources.add(source);
                }
                var name = mapping.name;
                if (name != null && !newNames.has(name)) {
                    newNames.add(name);
                }
            }, this);
            this._sources = newSources;
            this._names = newNames;
            // Copy sourcesContents of applied map.
            aSourceMapConsumer.sources.forEach(function (sourceFile) {
                var content = aSourceMapConsumer.sourceContentFor(sourceFile);
                if (content != null) {
                    if (aSourceMapPath != null) {
                        sourceFile = util.join(aSourceMapPath, sourceFile);
                    }
                    if (sourceRoot != null) {
                        sourceFile = util.relative(sourceRoot, sourceFile);
                    }
                    this.setSourceContent(sourceFile, content);
                }
            }, this);
        };
    /**
     * A mapping can have one of the three levels of data:
     *
     *   1. Just the generated position.
     *   2. The Generated position, original position, and original source.
     *   3. Generated and original position, original source, as well as a name
     *      token.
     *
     * To maintain consistency, we validate that any new mapping being added falls
     * in to one of these categories.
     */
    SourceMapGenerator.prototype._validateMapping =
        function SourceMapGenerator_validateMapping(aGenerated, aOriginal, aSource, aName) {
            // When aOriginal is truthy but has empty values for .line and .column,
            // it is most likely a programmer error. In this case we throw a very
            // specific error message to try to guide them the right way.
            // For example: https://github.com/Polymer/polymer-bundler/pull/519
            if (aOriginal && typeof aOriginal.line !== 'number' && typeof aOriginal.column !== 'number') {
                throw new Error('original.line and original.column are not numbers -- you probably meant to omit ' +
                    'the original mapping entirely and only map the generated position. If so, pass ' +
                    'null for the original mapping instead of an object with empty or null values.');
            }
            if (aGenerated && 'line' in aGenerated && 'column' in aGenerated
                && aGenerated.line > 0 && aGenerated.column >= 0
                && !aOriginal && !aSource && !aName) {
                // Case 1.
                return;
            }
            else if (aGenerated && 'line' in aGenerated && 'column' in aGenerated
                && aOriginal && 'line' in aOriginal && 'column' in aOriginal
                && aGenerated.line > 0 && aGenerated.column >= 0
                && aOriginal.line > 0 && aOriginal.column >= 0
                && aSource) {
                // Cases 2 and 3.
                return;
            }
            else {
                throw new Error('Invalid mapping: ' + JSON.stringify({
                    generated: aGenerated,
                    source: aSource,
                    original: aOriginal,
                    name: aName
                }));
            }
        };
    /**
     * Serialize the accumulated mappings in to the stream of base 64 VLQs
     * specified by the source map format.
     */
    SourceMapGenerator.prototype._serializeMappings =
        function SourceMapGenerator_serializeMappings() {
            var previousGeneratedColumn = 0;
            var previousGeneratedLine = 1;
            var previousOriginalColumn = 0;
            var previousOriginalLine = 0;
            var previousName = 0;
            var previousSource = 0;
            var result = '';
            var next;
            var mapping;
            var nameIdx;
            var sourceIdx;
            var mappings = this._mappings.toArray();
            for (var i = 0, len = mappings.length; i < len; i++) {
                mapping = mappings[i];
                next = '';
                if (mapping.generatedLine !== previousGeneratedLine) {
                    previousGeneratedColumn = 0;
                    while (mapping.generatedLine !== previousGeneratedLine) {
                        next += ';';
                        previousGeneratedLine++;
                    }
                }
                else {
                    if (i > 0) {
                        if (!util.compareByGeneratedPositionsInflated(mapping, mappings[i - 1])) {
                            continue;
                        }
                        next += ',';
                    }
                }
                next += base64VLQ.encode(mapping.generatedColumn
                    - previousGeneratedColumn);
                previousGeneratedColumn = mapping.generatedColumn;
                if (mapping.source != null) {
                    sourceIdx = this._sources.indexOf(mapping.source);
                    next += base64VLQ.encode(sourceIdx - previousSource);
                    previousSource = sourceIdx;
                    // lines are stored 0-based in SourceMap spec version 3
                    next += base64VLQ.encode(mapping.originalLine - 1
                        - previousOriginalLine);
                    previousOriginalLine = mapping.originalLine - 1;
                    next += base64VLQ.encode(mapping.originalColumn
                        - previousOriginalColumn);
                    previousOriginalColumn = mapping.originalColumn;
                    if (mapping.name != null) {
                        nameIdx = this._names.indexOf(mapping.name);
                        next += base64VLQ.encode(nameIdx - previousName);
                        previousName = nameIdx;
                    }
                }
                result += next;
            }
            return result;
        };
    SourceMapGenerator.prototype._generateSourcesContent =
        function SourceMapGenerator_generateSourcesContent(aSources, aSourceRoot) {
            return aSources.map(function (source) {
                if (!this._sourcesContents) {
                    return null;
                }
                if (aSourceRoot != null) {
                    source = util.relative(aSourceRoot, source);
                }
                var key = util.toSetString(source);
                return Object.prototype.hasOwnProperty.call(this._sourcesContents, key)
                    ? this._sourcesContents[key]
                    : null;
            }, this);
        };
    /**
     * Externalize the source map.
     */
    SourceMapGenerator.prototype.toJSON =
        function SourceMapGenerator_toJSON() {
            var map = {
                version: this._version,
                sources: this._sources.toArray(),
                names: this._names.toArray(),
                mappings: this._serializeMappings()
            };
            if (this._file != null) {
                map.file = this._file;
            }
            if (this._sourceRoot != null) {
                map.sourceRoot = this._sourceRoot;
            }
            if (this._sourcesContents) {
                map.sourcesContent = this._generateSourcesContent(map.sources, map.sourceRoot);
            }
            return map;
        };
    /**
     * Render the source map being generated to a string.
     */
    SourceMapGenerator.prototype.toString =
        function SourceMapGenerator_toString() {
            return JSON.stringify(this.toJSON());
        };
    exports.SourceMapGenerator = SourceMapGenerator;
},
/* source-map/lib/base64-vlq.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     *
     * Based on the Base 64 VLQ implementation in Closure Compiler:
     * https://code.google.com/p/closure-compiler/source/browse/trunk/src/com/google/debugging/sourcemap/Base64VLQ.java
     *
     * Copyright 2011 The Closure Compiler Authors. All rights reserved.
     * Redistribution and use in source and binary forms, with or without
     * modification, are permitted provided that the following conditions are
     * met:
     *
     *  * Redistributions of source code must retain the above copyright
     *    notice, this list of conditions and the following disclaimer.
     *  * Redistributions in binary form must reproduce the above
     *    copyright notice, this list of conditions and the following
     *    disclaimer in the documentation and/or other materials provided
     *    with the distribution.
     *  * Neither the name of Google Inc. nor the names of its
     *    contributors may be used to endorse or promote products derived
     *    from this software without specific prior written permission.
     *
     * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
     * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
     * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
     * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
     * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
     * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
     * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
     * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
     * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
     * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
     * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
     */
    var base64 = require(26) /* ./base64 */;
    // A single base 64 digit can contain 6 bits of data. For the base 64 variable
    // length quantities we use in the source map spec, the first bit is the sign,
    // the next four bits are the actual value, and the 6th bit is the
    // continuation bit. The continuation bit tells us whether there are more
    // digits in this value following this digit.
    //
    //   Continuation
    //   |    Sign
    //   |    |
    //   V    V
    //   101011
    var VLQ_BASE_SHIFT = 5;
    // binary: 100000
    var VLQ_BASE = 1 << VLQ_BASE_SHIFT;
    // binary: 011111
    var VLQ_BASE_MASK = VLQ_BASE - 1;
    // binary: 100000
    var VLQ_CONTINUATION_BIT = VLQ_BASE;
    /**
     * Converts from a two-complement value to a value where the sign bit is
     * placed in the least significant bit.  For example, as decimals:
     *   1 becomes 2 (10 binary), -1 becomes 3 (11 binary)
     *   2 becomes 4 (100 binary), -2 becomes 5 (101 binary)
     */
    function toVLQSigned(aValue) {
        return aValue < 0
            ? ((-aValue) << 1) + 1
            : (aValue << 1) + 0;
    }
    /**
     * Converts to a two-complement value from a value where the sign bit is
     * placed in the least significant bit.  For example, as decimals:
     *   2 (10 binary) becomes 1, 3 (11 binary) becomes -1
     *   4 (100 binary) becomes 2, 5 (101 binary) becomes -2
     */
    function fromVLQSigned(aValue) {
        var isNegative = (aValue & 1) === 1;
        var shifted = aValue >> 1;
        return isNegative
            ? -shifted
            : shifted;
    }
    /**
     * Returns the base 64 VLQ encoded value.
     */
    exports.encode = function base64VLQ_encode(aValue) {
        var encoded = "";
        var digit;
        var vlq = toVLQSigned(aValue);
        do {
            digit = vlq & VLQ_BASE_MASK;
            vlq >>>= VLQ_BASE_SHIFT;
            if (vlq > 0) {
                // There are still more digits in this value, so we must make sure the
                // continuation bit is marked.
                digit |= VLQ_CONTINUATION_BIT;
            }
            encoded += base64.encode(digit);
        } while (vlq > 0);
        return encoded;
    };
    /**
     * Decodes the next base 64 VLQ value from the given string and returns the
     * value and the rest of the string via the out parameter.
     */
    exports.decode = function base64VLQ_decode(aStr, aIndex, aOutParam) {
        var strLen = aStr.length;
        var result = 0;
        var shift = 0;
        var continuation, digit;
        do {
            if (aIndex >= strLen) {
                throw new Error("Expected more digits in base 64 VLQ value.");
            }
            digit = base64.decode(aStr.charCodeAt(aIndex++));
            if (digit === -1) {
                throw new Error("Invalid base64 digit: " + aStr.charAt(aIndex - 1));
            }
            continuation = !!(digit & VLQ_CONTINUATION_BIT);
            digit &= VLQ_BASE_MASK;
            result = result + (digit << shift);
            shift += VLQ_BASE_SHIFT;
        } while (continuation);
        aOutParam.value = fromVLQSigned(result);
        aOutParam.rest = aIndex;
    };
},
/* source-map/lib/base64.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    var intToCharMap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'.split('');
    /**
     * Encode an integer in the range of 0 to 63 to a single base 64 digit.
     */
    exports.encode = function (number) {
        if (0 <= number && number < intToCharMap.length) {
            return intToCharMap[number];
        }
        throw new TypeError("Must be between 0 and 63: " + number);
    };
    /**
     * Decode a single base 64 character code digit to an integer. Returns -1 on
     * failure.
     */
    exports.decode = function (charCode) {
        var bigA = 65; // 'A'
        var bigZ = 90; // 'Z'
        var littleA = 97; // 'a'
        var littleZ = 122; // 'z'
        var zero = 48; // '0'
        var nine = 57; // '9'
        var plus = 43; // '+'
        var slash = 47; // '/'
        var littleOffset = 26;
        var numberOffset = 52;
        // 0 - 25: ABCDEFGHIJKLMNOPQRSTUVWXYZ
        if (bigA <= charCode && charCode <= bigZ) {
            return (charCode - bigA);
        }
        // 26 - 51: abcdefghijklmnopqrstuvwxyz
        if (littleA <= charCode && charCode <= littleZ) {
            return (charCode - littleA + littleOffset);
        }
        // 52 - 61: 0123456789
        if (zero <= charCode && charCode <= nine) {
            return (charCode - zero + numberOffset);
        }
        // 62: +
        if (charCode == plus) {
            return 62;
        }
        // 63: /
        if (charCode == slash) {
            return 63;
        }
        // Invalid base64 digit.
        return -1;
    };
},
/* source-map/lib/util.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    /**
     * This is a helper function for getting values from parameter/options
     * objects.
     *
     * @param args The object we are extracting values from
     * @param name The name of the property we are getting.
     * @param defaultValue An optional value to return if the property is missing
     * from the object. If this is not specified and the property is missing, an
     * error will be thrown.
     */
    function getArg(aArgs, aName, aDefaultValue) {
        if (aName in aArgs) {
            return aArgs[aName];
        }
        else if (arguments.length === 3) {
            return aDefaultValue;
        }
        else {
            throw new Error('"' + aName + '" is a required argument.');
        }
    }
    exports.getArg = getArg;
    var urlRegexp = /^(?:([\w+\-.]+):)?\/\/(?:(\w+:\w+)@)?([\w.-]*)(?::(\d+))?(.*)$/;
    var dataUrlRegexp = /^data:.+\,.+$/;
    function urlParse(aUrl) {
        var match = aUrl.match(urlRegexp);
        if (!match) {
            return null;
        }
        return {
            scheme: match[1],
            auth: match[2],
            host: match[3],
            port: match[4],
            path: match[5]
        };
    }
    exports.urlParse = urlParse;
    function urlGenerate(aParsedUrl) {
        var url = '';
        if (aParsedUrl.scheme) {
            url += aParsedUrl.scheme + ':';
        }
        url += '//';
        if (aParsedUrl.auth) {
            url += aParsedUrl.auth + '@';
        }
        if (aParsedUrl.host) {
            url += aParsedUrl.host;
        }
        if (aParsedUrl.port) {
            url += ":" + aParsedUrl.port;
        }
        if (aParsedUrl.path) {
            url += aParsedUrl.path;
        }
        return url;
    }
    exports.urlGenerate = urlGenerate;
    /**
     * Normalizes a path, or the path portion of a URL:
     *
     * - Replaces consecutive slashes with one slash.
     * - Removes unnecessary '.' parts.
     * - Removes unnecessary '<dir>/..' parts.
     *
     * Based on code in the Node.js 'path' core module.
     *
     * @param aPath The path or url to normalize.
     */
    function normalize(aPath) {
        var path = aPath;
        var url = urlParse(aPath);
        if (url) {
            if (!url.path) {
                return aPath;
            }
            path = url.path;
        }
        var isAbsolute = exports.isAbsolute(path);
        var parts = path.split(/\/+/);
        for (var part, up = 0, i = parts.length - 1; i >= 0; i--) {
            part = parts[i];
            if (part === '.') {
                parts.splice(i, 1);
            }
            else if (part === '..') {
                up++;
            }
            else if (up > 0) {
                if (part === '') {
                    // The first part is blank if the path is absolute. Trying to go
                    // above the root is a no-op. Therefore we can remove all '..' parts
                    // directly after the root.
                    parts.splice(i + 1, up);
                    up = 0;
                }
                else {
                    parts.splice(i, 2);
                    up--;
                }
            }
        }
        path = parts.join('/');
        if (path === '') {
            path = isAbsolute ? '/' : '.';
        }
        if (url) {
            url.path = path;
            return urlGenerate(url);
        }
        return path;
    }
    exports.normalize = normalize;
    /**
     * Joins two paths/URLs.
     *
     * @param aRoot The root path or URL.
     * @param aPath The path or URL to be joined with the root.
     *
     * - If aPath is a URL or a data URI, aPath is returned, unless aPath is a
     *   scheme-relative URL: Then the scheme of aRoot, if any, is prepended
     *   first.
     * - Otherwise aPath is a path. If aRoot is a URL, then its path portion
     *   is updated with the result and aRoot is returned. Otherwise the result
     *   is returned.
     *   - If aPath is absolute, the result is aPath.
     *   - Otherwise the two paths are joined with a slash.
     * - Joining for example 'http://' and 'www.example.com' is also supported.
     */
    function join(aRoot, aPath) {
        if (aRoot === "") {
            aRoot = ".";
        }
        if (aPath === "") {
            aPath = ".";
        }
        var aPathUrl = urlParse(aPath);
        var aRootUrl = urlParse(aRoot);
        if (aRootUrl) {
            aRoot = aRootUrl.path || '/';
        }
        // `join(foo, '//www.example.org')`
        if (aPathUrl && !aPathUrl.scheme) {
            if (aRootUrl) {
                aPathUrl.scheme = aRootUrl.scheme;
            }
            return urlGenerate(aPathUrl);
        }
        if (aPathUrl || aPath.match(dataUrlRegexp)) {
            return aPath;
        }
        // `join('http://', 'www.example.com')`
        if (aRootUrl && !aRootUrl.host && !aRootUrl.path) {
            aRootUrl.host = aPath;
            return urlGenerate(aRootUrl);
        }
        var joined = aPath.charAt(0) === '/'
            ? aPath
            : normalize(aRoot.replace(/\/+$/, '') + '/' + aPath);
        if (aRootUrl) {
            aRootUrl.path = joined;
            return urlGenerate(aRootUrl);
        }
        return joined;
    }
    exports.join = join;
    exports.isAbsolute = function (aPath) {
        return aPath.charAt(0) === '/' || urlRegexp.test(aPath);
    };
    /**
     * Make a path relative to a URL or another path.
     *
     * @param aRoot The root path or URL.
     * @param aPath The path or URL to be made relative to aRoot.
     */
    function relative(aRoot, aPath) {
        if (aRoot === "") {
            aRoot = ".";
        }
        aRoot = aRoot.replace(/\/$/, '');
        // It is possible for the path to be above the root. In this case, simply
        // checking whether the root is a prefix of the path won't work. Instead, we
        // need to remove components from the root one by one, until either we find
        // a prefix that fits, or we run out of components to remove.
        var level = 0;
        while (aPath.indexOf(aRoot + '/') !== 0) {
            var index = aRoot.lastIndexOf("/");
            if (index < 0) {
                return aPath;
            }
            // If the only part of the root that is left is the scheme (i.e. http://,
            // file:///, etc.), one or more slashes (/), or simply nothing at all, we
            // have exhausted all components, so the path is not relative to the root.
            aRoot = aRoot.slice(0, index);
            if (aRoot.match(/^([^\/]+:\/)?\/*$/)) {
                return aPath;
            }
            ++level;
        }
        // Make sure we add a "../" for each component we removed from the root.
        return Array(level + 1).join("../") + aPath.substr(aRoot.length + 1);
    }
    exports.relative = relative;
    var supportsNullProto = (function () {
        var obj = Object.create(null);
        return !('__proto__' in obj);
    }());
    function identity(s) {
        return s;
    }
    /**
     * Because behavior goes wacky when you set `__proto__` on objects, we
     * have to prefix all the strings in our set with an arbitrary character.
     *
     * See https://github.com/mozilla/source-map/pull/31 and
     * https://github.com/mozilla/source-map/issues/30
     *
     * @param String aStr
     */
    function toSetString(aStr) {
        if (isProtoString(aStr)) {
            return '$' + aStr;
        }
        return aStr;
    }
    exports.toSetString = supportsNullProto ? identity : toSetString;
    function fromSetString(aStr) {
        if (isProtoString(aStr)) {
            return aStr.slice(1);
        }
        return aStr;
    }
    exports.fromSetString = supportsNullProto ? identity : fromSetString;
    function isProtoString(s) {
        if (!s) {
            return false;
        }
        var length = s.length;
        if (length < 9 /* "__proto__".length */) {
            return false;
        }
        if (s.charCodeAt(length - 1) !== 95 /* '_' */ ||
            s.charCodeAt(length - 2) !== 95 /* '_' */ ||
            s.charCodeAt(length - 3) !== 111 /* 'o' */ ||
            s.charCodeAt(length - 4) !== 116 /* 't' */ ||
            s.charCodeAt(length - 5) !== 111 /* 'o' */ ||
            s.charCodeAt(length - 6) !== 114 /* 'r' */ ||
            s.charCodeAt(length - 7) !== 112 /* 'p' */ ||
            s.charCodeAt(length - 8) !== 95 /* '_' */ ||
            s.charCodeAt(length - 9) !== 95 /* '_' */) {
            return false;
        }
        for (var i = length - 10; i >= 0; i--) {
            if (s.charCodeAt(i) !== 36 /* '$' */) {
                return false;
            }
        }
        return true;
    }
    /**
     * Comparator between two mappings where the original positions are compared.
     *
     * Optionally pass in `true` as `onlyCompareGenerated` to consider two
     * mappings with the same original source/line/column, but different generated
     * line and column the same. Useful when searching for a mapping with a
     * stubbed out mapping.
     */
    function compareByOriginalPositions(mappingA, mappingB, onlyCompareOriginal) {
        var cmp = strcmp(mappingA.source, mappingB.source);
        if (cmp !== 0) {
            return cmp;
        }
        cmp = mappingA.originalLine - mappingB.originalLine;
        if (cmp !== 0) {
            return cmp;
        }
        cmp = mappingA.originalColumn - mappingB.originalColumn;
        if (cmp !== 0 || onlyCompareOriginal) {
            return cmp;
        }
        cmp = mappingA.generatedColumn - mappingB.generatedColumn;
        if (cmp !== 0) {
            return cmp;
        }
        cmp = mappingA.generatedLine - mappingB.generatedLine;
        if (cmp !== 0) {
            return cmp;
        }
        return strcmp(mappingA.name, mappingB.name);
    }
    exports.compareByOriginalPositions = compareByOriginalPositions;
    /**
     * Comparator between two mappings with deflated source and name indices where
     * the generated positions are compared.
     *
     * Optionally pass in `true` as `onlyCompareGenerated` to consider two
     * mappings with the same generated line and column, but different
     * source/name/original line and column the same. Useful when searching for a
     * mapping with a stubbed out mapping.
     */
    function compareByGeneratedPositionsDeflated(mappingA, mappingB, onlyCompareGenerated) {
        var cmp = mappingA.generatedLine - mappingB.generatedLine;
        if (cmp !== 0) {
            return cmp;
        }
        cmp = mappingA.generatedColumn - mappingB.generatedColumn;
        if (cmp !== 0 || onlyCompareGenerated) {
            return cmp;
        }
        cmp = strcmp(mappingA.source, mappingB.source);
        if (cmp !== 0) {
            return cmp;
        }
        cmp = mappingA.originalLine - mappingB.originalLine;
        if (cmp !== 0) {
            return cmp;
        }
        cmp = mappingA.originalColumn - mappingB.originalColumn;
        if (cmp !== 0) {
            return cmp;
        }
        return strcmp(mappingA.name, mappingB.name);
    }
    exports.compareByGeneratedPositionsDeflated = compareByGeneratedPositionsDeflated;
    function strcmp(aStr1, aStr2) {
        if (aStr1 === aStr2) {
            return 0;
        }
        if (aStr1 === null) {
            return 1; // aStr2 !== null
        }
        if (aStr2 === null) {
            return -1; // aStr1 !== null
        }
        if (aStr1 > aStr2) {
            return 1;
        }
        return -1;
    }
    /**
     * Comparator between two mappings with inflated source and name strings where
     * the generated positions are compared.
     */
    function compareByGeneratedPositionsInflated(mappingA, mappingB) {
        var cmp = mappingA.generatedLine - mappingB.generatedLine;
        if (cmp !== 0) {
            return cmp;
        }
        cmp = mappingA.generatedColumn - mappingB.generatedColumn;
        if (cmp !== 0) {
            return cmp;
        }
        cmp = strcmp(mappingA.source, mappingB.source);
        if (cmp !== 0) {
            return cmp;
        }
        cmp = mappingA.originalLine - mappingB.originalLine;
        if (cmp !== 0) {
            return cmp;
        }
        cmp = mappingA.originalColumn - mappingB.originalColumn;
        if (cmp !== 0) {
            return cmp;
        }
        return strcmp(mappingA.name, mappingB.name);
    }
    exports.compareByGeneratedPositionsInflated = compareByGeneratedPositionsInflated;
    /**
     * Strip any JSON XSSI avoidance prefix from the string (as documented
     * in the source maps specification), and then parse the string as
     * JSON.
     */
    function parseSourceMapInput(str) {
        return JSON.parse(str.replace(/^\)]}'[^\n]*\n/, ''));
    }
    exports.parseSourceMapInput = parseSourceMapInput;
    /**
     * Compute the URL of a source given the the source root, the source's
     * URL, and the source map's URL.
     */
    function computeSourceURL(sourceRoot, sourceURL, sourceMapURL) {
        sourceURL = sourceURL || '';
        if (sourceRoot) {
            // This follows what Chrome does.
            if (sourceRoot[sourceRoot.length - 1] !== '/' && sourceURL[0] !== '/') {
                sourceRoot += '/';
            }
            // The spec says:
            //   Line 4: An optional source root, useful for relocating source
            //   files on a server or removing repeated values in the
            //   “sources” entry.  This value is prepended to the individual
            //   entries in the “source” field.
            sourceURL = sourceRoot + sourceURL;
        }
        // Historically, SourceMapConsumer did not take the sourceMapURL as
        // a parameter.  This mode is still somewhat supported, which is why
        // this code block is conditional.  However, it's preferable to pass
        // the source map URL to SourceMapConsumer, so that this function
        // can implement the source URL resolution algorithm as outlined in
        // the spec.  This block is basically the equivalent of:
        //    new URL(sourceURL, sourceMapURL).toString()
        // ... except it avoids using URL, which wasn't available in the
        // older releases of node still supported by this library.
        //
        // The spec says:
        //   If the sources are not absolute URLs after prepending of the
        //   “sourceRoot”, the sources are resolved relative to the
        //   SourceMap (like resolving script src in a html document).
        if (sourceMapURL) {
            var parsed = urlParse(sourceMapURL);
            if (!parsed) {
                throw new Error("sourceMapURL could not be parsed");
            }
            if (parsed.path) {
                // Strip the last path component, but keep the "/".
                var index = parsed.path.lastIndexOf('/');
                if (index >= 0) {
                    parsed.path = parsed.path.substring(0, index + 1);
                }
            }
            sourceURL = join(urlGenerate(parsed), sourceURL);
        }
        return normalize(sourceURL);
    }
    exports.computeSourceURL = computeSourceURL;
},
/* source-map/lib/array-set.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    var util = require(27) /* ./util */;
    var has = Object.prototype.hasOwnProperty;
    var hasNativeMap = typeof Map !== "undefined";
    /**
     * A data structure which is a combination of an array and a set. Adding a new
     * member is O(1), testing for membership is O(1), and finding the index of an
     * element is O(1). Removing elements from the set is not supported. Only
     * strings are supported for membership.
     */
    function ArraySet() {
        this._array = [];
        this._set = hasNativeMap ? new Map() : Object.create(null);
    }
    /**
     * Static method for creating ArraySet instances from an existing array.
     */
    ArraySet.fromArray = function ArraySet_fromArray(aArray, aAllowDuplicates) {
        var set = new ArraySet();
        for (var i = 0, len = aArray.length; i < len; i++) {
            set.add(aArray[i], aAllowDuplicates);
        }
        return set;
    };
    /**
     * Return how many unique items are in this ArraySet. If duplicates have been
     * added, than those do not count towards the size.
     *
     * @returns Number
     */
    ArraySet.prototype.size = function ArraySet_size() {
        return hasNativeMap ? this._set.size : Object.getOwnPropertyNames(this._set).length;
    };
    /**
     * Add the given string to this set.
     *
     * @param String aStr
     */
    ArraySet.prototype.add = function ArraySet_add(aStr, aAllowDuplicates) {
        var sStr = hasNativeMap ? aStr : util.toSetString(aStr);
        var isDuplicate = hasNativeMap ? this.has(aStr) : has.call(this._set, sStr);
        var idx = this._array.length;
        if (!isDuplicate || aAllowDuplicates) {
            this._array.push(aStr);
        }
        if (!isDuplicate) {
            if (hasNativeMap) {
                this._set.set(aStr, idx);
            }
            else {
                this._set[sStr] = idx;
            }
        }
    };
    /**
     * Is the given string a member of this set?
     *
     * @param String aStr
     */
    ArraySet.prototype.has = function ArraySet_has(aStr) {
        if (hasNativeMap) {
            return this._set.has(aStr);
        }
        else {
            var sStr = util.toSetString(aStr);
            return has.call(this._set, sStr);
        }
    };
    /**
     * What is the index of the given string in the array?
     *
     * @param String aStr
     */
    ArraySet.prototype.indexOf = function ArraySet_indexOf(aStr) {
        if (hasNativeMap) {
            var idx = this._set.get(aStr);
            if (idx >= 0) {
                return idx;
            }
        }
        else {
            var sStr = util.toSetString(aStr);
            if (has.call(this._set, sStr)) {
                return this._set[sStr];
            }
        }
        throw new Error('"' + aStr + '" is not in the set.');
    };
    /**
     * What is the element at the given index?
     *
     * @param Number aIdx
     */
    ArraySet.prototype.at = function ArraySet_at(aIdx) {
        if (aIdx >= 0 && aIdx < this._array.length) {
            return this._array[aIdx];
        }
        throw new Error('No element indexed by ' + aIdx);
    };
    /**
     * Returns the array representation of this set (which has the proper indices
     * indicated by indexOf). Note that this is a copy of the internal array used
     * for storing the members so that no one can mess with internal state.
     */
    ArraySet.prototype.toArray = function ArraySet_toArray() {
        return this._array.slice();
    };
    exports.ArraySet = ArraySet;
},
/* source-map/lib/mapping-list.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2014 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    var util = require(27) /* ./util */;
    /**
     * Determine whether mappingB is after mappingA with respect to generated
     * position.
     */
    function generatedPositionAfter(mappingA, mappingB) {
        // Optimized for most common case
        var lineA = mappingA.generatedLine;
        var lineB = mappingB.generatedLine;
        var columnA = mappingA.generatedColumn;
        var columnB = mappingB.generatedColumn;
        return lineB > lineA || lineB == lineA && columnB >= columnA ||
            util.compareByGeneratedPositionsInflated(mappingA, mappingB) <= 0;
    }
    /**
     * A data structure to provide a sorted view of accumulated mappings in a
     * performance conscious manner. It trades a neglibable overhead in general
     * case for a large speedup in case of mappings being added in order.
     */
    function MappingList() {
        this._array = [];
        this._sorted = true;
        // Serves as infimum
        this._last = { generatedLine: -1, generatedColumn: 0 };
    }
    /**
     * Iterate through internal items. This method takes the same arguments that
     * `Array.prototype.forEach` takes.
     *
     * NOTE: The order of the mappings is NOT guaranteed.
     */
    MappingList.prototype.unsortedForEach =
        function MappingList_forEach(aCallback, aThisArg) {
            this._array.forEach(aCallback, aThisArg);
        };
    /**
     * Add the given source mapping.
     *
     * @param Object aMapping
     */
    MappingList.prototype.add = function MappingList_add(aMapping) {
        if (generatedPositionAfter(this._last, aMapping)) {
            this._last = aMapping;
            this._array.push(aMapping);
        }
        else {
            this._sorted = false;
            this._array.push(aMapping);
        }
    };
    /**
     * Returns the flat, sorted array of mappings. The mappings are sorted by
     * generated position.
     *
     * WARNING: This method returns internal data without copying, for
     * performance. The return value must NOT be mutated, and should be treated as
     * an immutable borrow. If you want to take ownership, you must make your own
     * copy.
     */
    MappingList.prototype.toArray = function MappingList_toArray() {
        if (!this._sorted) {
            this._array.sort(util.compareByGeneratedPositionsInflated);
            this._sorted = true;
        }
        return this._array;
    };
    exports.MappingList = MappingList;
},
/* source-map/lib/source-map-consumer.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    var util = require(27) /* ./util */;
    var binarySearch = require(31) /* ./binary-search */;
    var ArraySet = require(28) /* ./array-set */.ArraySet;
    var base64VLQ = require(25) /* ./base64-vlq */;
    var quickSort = require(32) /* ./quick-sort */.quickSort;
    function SourceMapConsumer(aSourceMap, aSourceMapURL) {
        var sourceMap = aSourceMap;
        if (typeof aSourceMap === 'string') {
            sourceMap = util.parseSourceMapInput(aSourceMap);
        }
        return sourceMap.sections != null
            ? new IndexedSourceMapConsumer(sourceMap, aSourceMapURL)
            : new BasicSourceMapConsumer(sourceMap, aSourceMapURL);
    }
    SourceMapConsumer.fromSourceMap = function (aSourceMap, aSourceMapURL) {
        return BasicSourceMapConsumer.fromSourceMap(aSourceMap, aSourceMapURL);
    };
    /**
     * The version of the source mapping spec that we are consuming.
     */
    SourceMapConsumer.prototype._version = 3;
    // `__generatedMappings` and `__originalMappings` are arrays that hold the
    // parsed mapping coordinates from the source map's "mappings" attribute. They
    // are lazily instantiated, accessed via the `_generatedMappings` and
    // `_originalMappings` getters respectively, and we only parse the mappings
    // and create these arrays once queried for a source location. We jump through
    // these hoops because there can be many thousands of mappings, and parsing
    // them is expensive, so we only want to do it if we must.
    //
    // Each object in the arrays is of the form:
    //
    //     {
    //       generatedLine: The line number in the generated code,
    //       generatedColumn: The column number in the generated code,
    //       source: The path to the original source file that generated this
    //               chunk of code,
    //       originalLine: The line number in the original source that
    //                     corresponds to this chunk of generated code,
    //       originalColumn: The column number in the original source that
    //                       corresponds to this chunk of generated code,
    //       name: The name of the original symbol which generated this chunk of
    //             code.
    //     }
    //
    // All properties except for `generatedLine` and `generatedColumn` can be
    // `null`.
    //
    // `_generatedMappings` is ordered by the generated positions.
    //
    // `_originalMappings` is ordered by the original positions.
    SourceMapConsumer.prototype.__generatedMappings = null;
    Object.defineProperty(SourceMapConsumer.prototype, '_generatedMappings', {
        configurable: true,
        enumerable: true,
        get: function () {
            if (!this.__generatedMappings) {
                this._parseMappings(this._mappings, this.sourceRoot);
            }
            return this.__generatedMappings;
        }
    });
    SourceMapConsumer.prototype.__originalMappings = null;
    Object.defineProperty(SourceMapConsumer.prototype, '_originalMappings', {
        configurable: true,
        enumerable: true,
        get: function () {
            if (!this.__originalMappings) {
                this._parseMappings(this._mappings, this.sourceRoot);
            }
            return this.__originalMappings;
        }
    });
    SourceMapConsumer.prototype._charIsMappingSeparator =
        function SourceMapConsumer_charIsMappingSeparator(aStr, index) {
            var c = aStr.charAt(index);
            return c === ";" || c === ",";
        };
    /**
     * Parse the mappings in a string in to a data structure which we can easily
     * query (the ordered arrays in the `this.__generatedMappings` and
     * `this.__originalMappings` properties).
     */
    SourceMapConsumer.prototype._parseMappings =
        function SourceMapConsumer_parseMappings(aStr, aSourceRoot) {
            throw new Error("Subclasses must implement _parseMappings");
        };
    SourceMapConsumer.GENERATED_ORDER = 1;
    SourceMapConsumer.ORIGINAL_ORDER = 2;
    SourceMapConsumer.GREATEST_LOWER_BOUND = 1;
    SourceMapConsumer.LEAST_UPPER_BOUND = 2;
    /**
     * Iterate over each mapping between an original source/line/column and a
     * generated line/column in this source map.
     *
     * @param Function aCallback
     *        The function that is called with each mapping.
     * @param Object aContext
     *        Optional. If specified, this object will be the value of `this` every
     *        time that `aCallback` is called.
     * @param aOrder
     *        Either `SourceMapConsumer.GENERATED_ORDER` or
     *        `SourceMapConsumer.ORIGINAL_ORDER`. Specifies whether you want to
     *        iterate over the mappings sorted by the generated file's line/column
     *        order or the original's source/line/column order, respectively. Defaults to
     *        `SourceMapConsumer.GENERATED_ORDER`.
     */
    SourceMapConsumer.prototype.eachMapping =
        function SourceMapConsumer_eachMapping(aCallback, aContext, aOrder) {
            var context = aContext || null;
            var order = aOrder || SourceMapConsumer.GENERATED_ORDER;
            var mappings;
            switch (order) {
                case SourceMapConsumer.GENERATED_ORDER:
                    mappings = this._generatedMappings;
                    break;
                case SourceMapConsumer.ORIGINAL_ORDER:
                    mappings = this._originalMappings;
                    break;
                default:
                    throw new Error("Unknown order of iteration.");
            }
            var sourceRoot = this.sourceRoot;
            mappings.map(function (mapping) {
                var source = mapping.source === null ? null : this._sources.at(mapping.source);
                source = util.computeSourceURL(sourceRoot, source, this._sourceMapURL);
                return {
                    source: source,
                    generatedLine: mapping.generatedLine,
                    generatedColumn: mapping.generatedColumn,
                    originalLine: mapping.originalLine,
                    originalColumn: mapping.originalColumn,
                    name: mapping.name === null ? null : this._names.at(mapping.name)
                };
            }, this).forEach(aCallback, context);
        };
    /**
     * Returns all generated line and column information for the original source,
     * line, and column provided. If no column is provided, returns all mappings
     * corresponding to a either the line we are searching for or the next
     * closest line that has any mappings. Otherwise, returns all mappings
     * corresponding to the given line and either the column we are searching for
     * or the next closest column that has any offsets.
     *
     * The only argument is an object with the following properties:
     *
     *   - source: The filename of the original source.
     *   - line: The line number in the original source.  The line number is 1-based.
     *   - column: Optional. the column number in the original source.
     *    The column number is 0-based.
     *
     * and an array of objects is returned, each with the following properties:
     *
     *   - line: The line number in the generated source, or null.  The
     *    line number is 1-based.
     *   - column: The column number in the generated source, or null.
     *    The column number is 0-based.
     */
    SourceMapConsumer.prototype.allGeneratedPositionsFor =
        function SourceMapConsumer_allGeneratedPositionsFor(aArgs) {
            var line = util.getArg(aArgs, 'line');
            // When there is no exact match, BasicSourceMapConsumer.prototype._findMapping
            // returns the index of the closest mapping less than the needle. By
            // setting needle.originalColumn to 0, we thus find the last mapping for
            // the given line, provided such a mapping exists.
            var needle = {
                source: util.getArg(aArgs, 'source'),
                originalLine: line,
                originalColumn: util.getArg(aArgs, 'column', 0)
            };
            needle.source = this._findSourceIndex(needle.source);
            if (needle.source < 0) {
                return [];
            }
            var mappings = [];
            var index = this._findMapping(needle, this._originalMappings, "originalLine", "originalColumn", util.compareByOriginalPositions, binarySearch.LEAST_UPPER_BOUND);
            if (index >= 0) {
                var mapping = this._originalMappings[index];
                if (aArgs.column === undefined) {
                    var originalLine = mapping.originalLine;
                    // Iterate until either we run out of mappings, or we run into
                    // a mapping for a different line than the one we found. Since
                    // mappings are sorted, this is guaranteed to find all mappings for
                    // the line we found.
                    while (mapping && mapping.originalLine === originalLine) {
                        mappings.push({
                            line: util.getArg(mapping, 'generatedLine', null),
                            column: util.getArg(mapping, 'generatedColumn', null),
                            lastColumn: util.getArg(mapping, 'lastGeneratedColumn', null)
                        });
                        mapping = this._originalMappings[++index];
                    }
                }
                else {
                    var originalColumn = mapping.originalColumn;
                    // Iterate until either we run out of mappings, or we run into
                    // a mapping for a different line than the one we were searching for.
                    // Since mappings are sorted, this is guaranteed to find all mappings for
                    // the line we are searching for.
                    while (mapping &&
                        mapping.originalLine === line &&
                        mapping.originalColumn == originalColumn) {
                        mappings.push({
                            line: util.getArg(mapping, 'generatedLine', null),
                            column: util.getArg(mapping, 'generatedColumn', null),
                            lastColumn: util.getArg(mapping, 'lastGeneratedColumn', null)
                        });
                        mapping = this._originalMappings[++index];
                    }
                }
            }
            return mappings;
        };
    exports.SourceMapConsumer = SourceMapConsumer;
    /**
     * A BasicSourceMapConsumer instance represents a parsed source map which we can
     * query for information about the original file positions by giving it a file
     * position in the generated source.
     *
     * The first parameter is the raw source map (either as a JSON string, or
     * already parsed to an object). According to the spec, source maps have the
     * following attributes:
     *
     *   - version: Which version of the source map spec this map is following.
     *   - sources: An array of URLs to the original source files.
     *   - names: An array of identifiers which can be referrenced by individual mappings.
     *   - sourceRoot: Optional. The URL root from which all sources are relative.
     *   - sourcesContent: Optional. An array of contents of the original source files.
     *   - mappings: A string of base64 VLQs which contain the actual mappings.
     *   - file: Optional. The generated file this source map is associated with.
     *
     * Here is an example source map, taken from the source map spec[0]:
     *
     *     {
     *       version : 3,
     *       file: "out.js",
     *       sourceRoot : "",
     *       sources: ["foo.js", "bar.js"],
     *       names: ["src", "maps", "are", "fun"],
     *       mappings: "AA,AB;;ABCDE;"
     *     }
     *
     * The second parameter, if given, is a string whose value is the URL
     * at which the source map was found.  This URL is used to compute the
     * sources array.
     *
     * [0]: https://docs.google.com/document/d/1U1RGAehQwRypUTovF1KRlpiOFze0b-_2gc6fAH0KY0k/edit?pli=1#
     */
    function BasicSourceMapConsumer(aSourceMap, aSourceMapURL) {
        var sourceMap = aSourceMap;
        if (typeof aSourceMap === 'string') {
            sourceMap = util.parseSourceMapInput(aSourceMap);
        }
        var version = util.getArg(sourceMap, 'version');
        var sources = util.getArg(sourceMap, 'sources');
        // Sass 3.3 leaves out the 'names' array, so we deviate from the spec (which
        // requires the array) to play nice here.
        var names = util.getArg(sourceMap, 'names', []);
        var sourceRoot = util.getArg(sourceMap, 'sourceRoot', null);
        var sourcesContent = util.getArg(sourceMap, 'sourcesContent', null);
        var mappings = util.getArg(sourceMap, 'mappings');
        var file = util.getArg(sourceMap, 'file', null);
        // Once again, Sass deviates from the spec and supplies the version as a
        // string rather than a number, so we use loose equality checking here.
        if (version != this._version) {
            throw new Error('Unsupported version: ' + version);
        }
        if (sourceRoot) {
            sourceRoot = util.normalize(sourceRoot);
        }
        sources = sources
            .map(String)
            // Some source maps produce relative source paths like "./foo.js" instead of
            // "foo.js".  Normalize these first so that future comparisons will succeed.
            // See bugzil.la/1090768.
            .map(util.normalize)
            // Always ensure that absolute sources are internally stored relative to
            // the source root, if the source root is absolute. Not doing this would
            // be particularly problematic when the source root is a prefix of the
            // source (valid, but why??). See github issue #199 and bugzil.la/1188982.
            .map(function (source) {
            return sourceRoot && util.isAbsolute(sourceRoot) && util.isAbsolute(source)
                ? util.relative(sourceRoot, source)
                : source;
        });
        // Pass `true` below to allow duplicate names and sources. While source maps
        // are intended to be compressed and deduplicated, the TypeScript compiler
        // sometimes generates source maps with duplicates in them. See Github issue
        // #72 and bugzil.la/889492.
        this._names = ArraySet.fromArray(names.map(String), true);
        this._sources = ArraySet.fromArray(sources, true);
        this._absoluteSources = this._sources.toArray().map(function (s) {
            return util.computeSourceURL(sourceRoot, s, aSourceMapURL);
        });
        this.sourceRoot = sourceRoot;
        this.sourcesContent = sourcesContent;
        this._mappings = mappings;
        this._sourceMapURL = aSourceMapURL;
        this.file = file;
    }
    BasicSourceMapConsumer.prototype = Object.create(SourceMapConsumer.prototype);
    BasicSourceMapConsumer.prototype.consumer = SourceMapConsumer;
    /**
     * Utility function to find the index of a source.  Returns -1 if not
     * found.
     */
    BasicSourceMapConsumer.prototype._findSourceIndex = function (aSource) {
        var relativeSource = aSource;
        if (this.sourceRoot != null) {
            relativeSource = util.relative(this.sourceRoot, relativeSource);
        }
        if (this._sources.has(relativeSource)) {
            return this._sources.indexOf(relativeSource);
        }
        // Maybe aSource is an absolute URL as returned by |sources|.  In
        // this case we can't simply undo the transform.
        var i;
        for (i = 0; i < this._absoluteSources.length; ++i) {
            if (this._absoluteSources[i] == aSource) {
                return i;
            }
        }
        return -1;
    };
    /**
     * Create a BasicSourceMapConsumer from a SourceMapGenerator.
     *
     * @param SourceMapGenerator aSourceMap
     *        The source map that will be consumed.
     * @param String aSourceMapURL
     *        The URL at which the source map can be found (optional)
     * @returns BasicSourceMapConsumer
     */
    BasicSourceMapConsumer.fromSourceMap =
        function SourceMapConsumer_fromSourceMap(aSourceMap, aSourceMapURL) {
            var smc = Object.create(BasicSourceMapConsumer.prototype);
            var names = smc._names = ArraySet.fromArray(aSourceMap._names.toArray(), true);
            var sources = smc._sources = ArraySet.fromArray(aSourceMap._sources.toArray(), true);
            smc.sourceRoot = aSourceMap._sourceRoot;
            smc.sourcesContent = aSourceMap._generateSourcesContent(smc._sources.toArray(), smc.sourceRoot);
            smc.file = aSourceMap._file;
            smc._sourceMapURL = aSourceMapURL;
            smc._absoluteSources = smc._sources.toArray().map(function (s) {
                return util.computeSourceURL(smc.sourceRoot, s, aSourceMapURL);
            });
            // Because we are modifying the entries (by converting string sources and
            // names to indices into the sources and names ArraySets), we have to make
            // a copy of the entry or else bad things happen. Shared mutable state
            // strikes again! See github issue #191.
            var generatedMappings = aSourceMap._mappings.toArray().slice();
            var destGeneratedMappings = smc.__generatedMappings = [];
            var destOriginalMappings = smc.__originalMappings = [];
            for (var i = 0, length = generatedMappings.length; i < length; i++) {
                var srcMapping = generatedMappings[i];
                var destMapping = new Mapping;
                destMapping.generatedLine = srcMapping.generatedLine;
                destMapping.generatedColumn = srcMapping.generatedColumn;
                if (srcMapping.source) {
                    destMapping.source = sources.indexOf(srcMapping.source);
                    destMapping.originalLine = srcMapping.originalLine;
                    destMapping.originalColumn = srcMapping.originalColumn;
                    if (srcMapping.name) {
                        destMapping.name = names.indexOf(srcMapping.name);
                    }
                    destOriginalMappings.push(destMapping);
                }
                destGeneratedMappings.push(destMapping);
            }
            quickSort(smc.__originalMappings, util.compareByOriginalPositions);
            return smc;
        };
    /**
     * The version of the source mapping spec that we are consuming.
     */
    BasicSourceMapConsumer.prototype._version = 3;
    /**
     * The list of original sources.
     */
    Object.defineProperty(BasicSourceMapConsumer.prototype, 'sources', {
        get: function () {
            return this._absoluteSources.slice();
        }
    });
    /**
     * Provide the JIT with a nice shape / hidden class.
     */
    function Mapping() {
        this.generatedLine = 0;
        this.generatedColumn = 0;
        this.source = null;
        this.originalLine = null;
        this.originalColumn = null;
        this.name = null;
    }
    /**
     * Parse the mappings in a string in to a data structure which we can easily
     * query (the ordered arrays in the `this.__generatedMappings` and
     * `this.__originalMappings` properties).
     */
    BasicSourceMapConsumer.prototype._parseMappings =
        function SourceMapConsumer_parseMappings(aStr, aSourceRoot) {
            var generatedLine = 1;
            var previousGeneratedColumn = 0;
            var previousOriginalLine = 0;
            var previousOriginalColumn = 0;
            var previousSource = 0;
            var previousName = 0;
            var length = aStr.length;
            var index = 0;
            var cachedSegments = {};
            var temp = {};
            var originalMappings = [];
            var generatedMappings = [];
            var mapping, str, segment, end, value;
            while (index < length) {
                if (aStr.charAt(index) === ';') {
                    generatedLine++;
                    index++;
                    previousGeneratedColumn = 0;
                }
                else if (aStr.charAt(index) === ',') {
                    index++;
                }
                else {
                    mapping = new Mapping();
                    mapping.generatedLine = generatedLine;
                    // Because each offset is encoded relative to the previous one,
                    // many segments often have the same encoding. We can exploit this
                    // fact by caching the parsed variable length fields of each segment,
                    // allowing us to avoid a second parse if we encounter the same
                    // segment again.
                    for (end = index; end < length; end++) {
                        if (this._charIsMappingSeparator(aStr, end)) {
                            break;
                        }
                    }
                    str = aStr.slice(index, end);
                    segment = cachedSegments[str];
                    if (segment) {
                        index += str.length;
                    }
                    else {
                        segment = [];
                        while (index < end) {
                            base64VLQ.decode(aStr, index, temp);
                            value = temp.value;
                            index = temp.rest;
                            segment.push(value);
                        }
                        if (segment.length === 2) {
                            throw new Error('Found a source, but no line and column');
                        }
                        if (segment.length === 3) {
                            throw new Error('Found a source and line, but no column');
                        }
                        cachedSegments[str] = segment;
                    }
                    // Generated column.
                    mapping.generatedColumn = previousGeneratedColumn + segment[0];
                    previousGeneratedColumn = mapping.generatedColumn;
                    if (segment.length > 1) {
                        // Original source.
                        mapping.source = previousSource + segment[1];
                        previousSource += segment[1];
                        // Original line.
                        mapping.originalLine = previousOriginalLine + segment[2];
                        previousOriginalLine = mapping.originalLine;
                        // Lines are stored 0-based
                        mapping.originalLine += 1;
                        // Original column.
                        mapping.originalColumn = previousOriginalColumn + segment[3];
                        previousOriginalColumn = mapping.originalColumn;
                        if (segment.length > 4) {
                            // Original name.
                            mapping.name = previousName + segment[4];
                            previousName += segment[4];
                        }
                    }
                    generatedMappings.push(mapping);
                    if (typeof mapping.originalLine === 'number') {
                        originalMappings.push(mapping);
                    }
                }
            }
            quickSort(generatedMappings, util.compareByGeneratedPositionsDeflated);
            this.__generatedMappings = generatedMappings;
            quickSort(originalMappings, util.compareByOriginalPositions);
            this.__originalMappings = originalMappings;
        };
    /**
     * Find the mapping that best matches the hypothetical "needle" mapping that
     * we are searching for in the given "haystack" of mappings.
     */
    BasicSourceMapConsumer.prototype._findMapping =
        function SourceMapConsumer_findMapping(aNeedle, aMappings, aLineName, aColumnName, aComparator, aBias) {
            // To return the position we are searching for, we must first find the
            // mapping for the given position and then return the opposite position it
            // points to. Because the mappings are sorted, we can use binary search to
            // find the best mapping.
            if (aNeedle[aLineName] <= 0) {
                throw new TypeError('Line must be greater than or equal to 1, got '
                    + aNeedle[aLineName]);
            }
            if (aNeedle[aColumnName] < 0) {
                throw new TypeError('Column must be greater than or equal to 0, got '
                    + aNeedle[aColumnName]);
            }
            return binarySearch.search(aNeedle, aMappings, aComparator, aBias);
        };
    /**
     * Compute the last column for each generated mapping. The last column is
     * inclusive.
     */
    BasicSourceMapConsumer.prototype.computeColumnSpans =
        function SourceMapConsumer_computeColumnSpans() {
            for (var index = 0; index < this._generatedMappings.length; ++index) {
                var mapping = this._generatedMappings[index];
                // Mappings do not contain a field for the last generated columnt. We
                // can come up with an optimistic estimate, however, by assuming that
                // mappings are contiguous (i.e. given two consecutive mappings, the
                // first mapping ends where the second one starts).
                if (index + 1 < this._generatedMappings.length) {
                    var nextMapping = this._generatedMappings[index + 1];
                    if (mapping.generatedLine === nextMapping.generatedLine) {
                        mapping.lastGeneratedColumn = nextMapping.generatedColumn - 1;
                        continue;
                    }
                }
                // The last mapping for each line spans the entire line.
                mapping.lastGeneratedColumn = Infinity;
            }
        };
    /**
     * Returns the original source, line, and column information for the generated
     * source's line and column positions provided. The only argument is an object
     * with the following properties:
     *
     *   - line: The line number in the generated source.  The line number
     *     is 1-based.
     *   - column: The column number in the generated source.  The column
     *     number is 0-based.
     *   - bias: Either 'SourceMapConsumer.GREATEST_LOWER_BOUND' or
     *     'SourceMapConsumer.LEAST_UPPER_BOUND'. Specifies whether to return the
     *     closest element that is smaller than or greater than the one we are
     *     searching for, respectively, if the exact element cannot be found.
     *     Defaults to 'SourceMapConsumer.GREATEST_LOWER_BOUND'.
     *
     * and an object is returned with the following properties:
     *
     *   - source: The original source file, or null.
     *   - line: The line number in the original source, or null.  The
     *     line number is 1-based.
     *   - column: The column number in the original source, or null.  The
     *     column number is 0-based.
     *   - name: The original identifier, or null.
     */
    BasicSourceMapConsumer.prototype.originalPositionFor =
        function SourceMapConsumer_originalPositionFor(aArgs) {
            var needle = {
                generatedLine: util.getArg(aArgs, 'line'),
                generatedColumn: util.getArg(aArgs, 'column')
            };
            var index = this._findMapping(needle, this._generatedMappings, "generatedLine", "generatedColumn", util.compareByGeneratedPositionsDeflated, util.getArg(aArgs, 'bias', SourceMapConsumer.GREATEST_LOWER_BOUND));
            if (index >= 0) {
                var mapping = this._generatedMappings[index];
                if (mapping.generatedLine === needle.generatedLine) {
                    var source = util.getArg(mapping, 'source', null);
                    if (source !== null) {
                        source = this._sources.at(source);
                        source = util.computeSourceURL(this.sourceRoot, source, this._sourceMapURL);
                    }
                    var name = util.getArg(mapping, 'name', null);
                    if (name !== null) {
                        name = this._names.at(name);
                    }
                    return {
                        source: source,
                        line: util.getArg(mapping, 'originalLine', null),
                        column: util.getArg(mapping, 'originalColumn', null),
                        name: name
                    };
                }
            }
            return {
                source: null,
                line: null,
                column: null,
                name: null
            };
        };
    /**
     * Return true if we have the source content for every source in the source
     * map, false otherwise.
     */
    BasicSourceMapConsumer.prototype.hasContentsOfAllSources =
        function BasicSourceMapConsumer_hasContentsOfAllSources() {
            if (!this.sourcesContent) {
                return false;
            }
            return this.sourcesContent.length >= this._sources.size() &&
                !this.sourcesContent.some(function (sc) { return sc == null; });
        };
    /**
     * Returns the original source content. The only argument is the url of the
     * original source file. Returns null if no original source content is
     * available.
     */
    BasicSourceMapConsumer.prototype.sourceContentFor =
        function SourceMapConsumer_sourceContentFor(aSource, nullOnMissing) {
            if (!this.sourcesContent) {
                return null;
            }
            var index = this._findSourceIndex(aSource);
            if (index >= 0) {
                return this.sourcesContent[index];
            }
            var relativeSource = aSource;
            if (this.sourceRoot != null) {
                relativeSource = util.relative(this.sourceRoot, relativeSource);
            }
            var url;
            if (this.sourceRoot != null
                && (url = util.urlParse(this.sourceRoot))) {
                // XXX: file:// URIs and absolute paths lead to unexpected behavior for
                // many users. We can help them out when they expect file:// URIs to
                // behave like it would if they were running a local HTTP server. See
                // https://bugzilla.mozilla.org/show_bug.cgi?id=885597.
                var fileUriAbsPath = relativeSource.replace(/^file:\/\//, "");
                if (url.scheme == "file"
                    && this._sources.has(fileUriAbsPath)) {
                    return this.sourcesContent[this._sources.indexOf(fileUriAbsPath)];
                }
                if ((!url.path || url.path == "/")
                    && this._sources.has("/" + relativeSource)) {
                    return this.sourcesContent[this._sources.indexOf("/" + relativeSource)];
                }
            }
            // This function is used recursively from
            // IndexedSourceMapConsumer.prototype.sourceContentFor. In that case, we
            // don't want to throw if we can't find the source - we just want to
            // return null, so we provide a flag to exit gracefully.
            if (nullOnMissing) {
                return null;
            }
            else {
                throw new Error('"' + relativeSource + '" is not in the SourceMap.');
            }
        };
    /**
     * Returns the generated line and column information for the original source,
     * line, and column positions provided. The only argument is an object with
     * the following properties:
     *
     *   - source: The filename of the original source.
     *   - line: The line number in the original source.  The line number
     *     is 1-based.
     *   - column: The column number in the original source.  The column
     *     number is 0-based.
     *   - bias: Either 'SourceMapConsumer.GREATEST_LOWER_BOUND' or
     *     'SourceMapConsumer.LEAST_UPPER_BOUND'. Specifies whether to return the
     *     closest element that is smaller than or greater than the one we are
     *     searching for, respectively, if the exact element cannot be found.
     *     Defaults to 'SourceMapConsumer.GREATEST_LOWER_BOUND'.
     *
     * and an object is returned with the following properties:
     *
     *   - line: The line number in the generated source, or null.  The
     *     line number is 1-based.
     *   - column: The column number in the generated source, or null.
     *     The column number is 0-based.
     */
    BasicSourceMapConsumer.prototype.generatedPositionFor =
        function SourceMapConsumer_generatedPositionFor(aArgs) {
            var source = util.getArg(aArgs, 'source');
            source = this._findSourceIndex(source);
            if (source < 0) {
                return {
                    line: null,
                    column: null,
                    lastColumn: null
                };
            }
            var needle = {
                source: source,
                originalLine: util.getArg(aArgs, 'line'),
                originalColumn: util.getArg(aArgs, 'column')
            };
            var index = this._findMapping(needle, this._originalMappings, "originalLine", "originalColumn", util.compareByOriginalPositions, util.getArg(aArgs, 'bias', SourceMapConsumer.GREATEST_LOWER_BOUND));
            if (index >= 0) {
                var mapping = this._originalMappings[index];
                if (mapping.source === needle.source) {
                    return {
                        line: util.getArg(mapping, 'generatedLine', null),
                        column: util.getArg(mapping, 'generatedColumn', null),
                        lastColumn: util.getArg(mapping, 'lastGeneratedColumn', null)
                    };
                }
            }
            return {
                line: null,
                column: null,
                lastColumn: null
            };
        };
    exports.BasicSourceMapConsumer = BasicSourceMapConsumer;
    /**
     * An IndexedSourceMapConsumer instance represents a parsed source map which
     * we can query for information. It differs from BasicSourceMapConsumer in
     * that it takes "indexed" source maps (i.e. ones with a "sections" field) as
     * input.
     *
     * The first parameter is a raw source map (either as a JSON string, or already
     * parsed to an object). According to the spec for indexed source maps, they
     * have the following attributes:
     *
     *   - version: Which version of the source map spec this map is following.
     *   - file: Optional. The generated file this source map is associated with.
     *   - sections: A list of section definitions.
     *
     * Each value under the "sections" field has two fields:
     *   - offset: The offset into the original specified at which this section
     *       begins to apply, defined as an object with a "line" and "column"
     *       field.
     *   - map: A source map definition. This source map could also be indexed,
     *       but doesn't have to be.
     *
     * Instead of the "map" field, it's also possible to have a "url" field
     * specifying a URL to retrieve a source map from, but that's currently
     * unsupported.
     *
     * Here's an example source map, taken from the source map spec[0], but
     * modified to omit a section which uses the "url" field.
     *
     *  {
     *    version : 3,
     *    file: "app.js",
     *    sections: [{
     *      offset: {line:100, column:10},
     *      map: {
     *        version : 3,
     *        file: "section.js",
     *        sources: ["foo.js", "bar.js"],
     *        names: ["src", "maps", "are", "fun"],
     *        mappings: "AAAA,E;;ABCDE;"
     *      }
     *    }],
     *  }
     *
     * The second parameter, if given, is a string whose value is the URL
     * at which the source map was found.  This URL is used to compute the
     * sources array.
     *
     * [0]: https://docs.google.com/document/d/1U1RGAehQwRypUTovF1KRlpiOFze0b-_2gc6fAH0KY0k/edit#heading=h.535es3xeprgt
     */
    function IndexedSourceMapConsumer(aSourceMap, aSourceMapURL) {
        var sourceMap = aSourceMap;
        if (typeof aSourceMap === 'string') {
            sourceMap = util.parseSourceMapInput(aSourceMap);
        }
        var version = util.getArg(sourceMap, 'version');
        var sections = util.getArg(sourceMap, 'sections');
        if (version != this._version) {
            throw new Error('Unsupported version: ' + version);
        }
        this._sources = new ArraySet();
        this._names = new ArraySet();
        var lastOffset = {
            line: -1,
            column: 0
        };
        this._sections = sections.map(function (s) {
            if (s.url) {
                // The url field will require support for asynchronicity.
                // See https://github.com/mozilla/source-map/issues/16
                throw new Error('Support for url field in sections not implemented.');
            }
            var offset = util.getArg(s, 'offset');
            var offsetLine = util.getArg(offset, 'line');
            var offsetColumn = util.getArg(offset, 'column');
            if (offsetLine < lastOffset.line ||
                (offsetLine === lastOffset.line && offsetColumn < lastOffset.column)) {
                throw new Error('Section offsets must be ordered and non-overlapping.');
            }
            lastOffset = offset;
            return {
                generatedOffset: {
                    // The offset fields are 0-based, but we use 1-based indices when
                    // encoding/decoding from VLQ.
                    generatedLine: offsetLine + 1,
                    generatedColumn: offsetColumn + 1
                },
                consumer: new SourceMapConsumer(util.getArg(s, 'map'), aSourceMapURL)
            };
        });
    }
    IndexedSourceMapConsumer.prototype = Object.create(SourceMapConsumer.prototype);
    IndexedSourceMapConsumer.prototype.constructor = SourceMapConsumer;
    /**
     * The version of the source mapping spec that we are consuming.
     */
    IndexedSourceMapConsumer.prototype._version = 3;
    /**
     * The list of original sources.
     */
    Object.defineProperty(IndexedSourceMapConsumer.prototype, 'sources', {
        get: function () {
            var sources = [];
            for (var i = 0; i < this._sections.length; i++) {
                for (var j = 0; j < this._sections[i].consumer.sources.length; j++) {
                    sources.push(this._sections[i].consumer.sources[j]);
                }
            }
            return sources;
        }
    });
    /**
     * Returns the original source, line, and column information for the generated
     * source's line and column positions provided. The only argument is an object
     * with the following properties:
     *
     *   - line: The line number in the generated source.  The line number
     *     is 1-based.
     *   - column: The column number in the generated source.  The column
     *     number is 0-based.
     *
     * and an object is returned with the following properties:
     *
     *   - source: The original source file, or null.
     *   - line: The line number in the original source, or null.  The
     *     line number is 1-based.
     *   - column: The column number in the original source, or null.  The
     *     column number is 0-based.
     *   - name: The original identifier, or null.
     */
    IndexedSourceMapConsumer.prototype.originalPositionFor =
        function IndexedSourceMapConsumer_originalPositionFor(aArgs) {
            var needle = {
                generatedLine: util.getArg(aArgs, 'line'),
                generatedColumn: util.getArg(aArgs, 'column')
            };
            // Find the section containing the generated position we're trying to map
            // to an original position.
            var sectionIndex = binarySearch.search(needle, this._sections, function (needle, section) {
                var cmp = needle.generatedLine - section.generatedOffset.generatedLine;
                if (cmp) {
                    return cmp;
                }
                return (needle.generatedColumn -
                    section.generatedOffset.generatedColumn);
            });
            var section = this._sections[sectionIndex];
            if (!section) {
                return {
                    source: null,
                    line: null,
                    column: null,
                    name: null
                };
            }
            return section.consumer.originalPositionFor({
                line: needle.generatedLine -
                    (section.generatedOffset.generatedLine - 1),
                column: needle.generatedColumn -
                    (section.generatedOffset.generatedLine === needle.generatedLine
                        ? section.generatedOffset.generatedColumn - 1
                        : 0),
                bias: aArgs.bias
            });
        };
    /**
     * Return true if we have the source content for every source in the source
     * map, false otherwise.
     */
    IndexedSourceMapConsumer.prototype.hasContentsOfAllSources =
        function IndexedSourceMapConsumer_hasContentsOfAllSources() {
            return this._sections.every(function (s) {
                return s.consumer.hasContentsOfAllSources();
            });
        };
    /**
     * Returns the original source content. The only argument is the url of the
     * original source file. Returns null if no original source content is
     * available.
     */
    IndexedSourceMapConsumer.prototype.sourceContentFor =
        function IndexedSourceMapConsumer_sourceContentFor(aSource, nullOnMissing) {
            for (var i = 0; i < this._sections.length; i++) {
                var section = this._sections[i];
                var content = section.consumer.sourceContentFor(aSource, true);
                if (content) {
                    return content;
                }
            }
            if (nullOnMissing) {
                return null;
            }
            else {
                throw new Error('"' + aSource + '" is not in the SourceMap.');
            }
        };
    /**
     * Returns the generated line and column information for the original source,
     * line, and column positions provided. The only argument is an object with
     * the following properties:
     *
     *   - source: The filename of the original source.
     *   - line: The line number in the original source.  The line number
     *     is 1-based.
     *   - column: The column number in the original source.  The column
     *     number is 0-based.
     *
     * and an object is returned with the following properties:
     *
     *   - line: The line number in the generated source, or null.  The
     *     line number is 1-based.
     *   - column: The column number in the generated source, or null.
     *     The column number is 0-based.
     */
    IndexedSourceMapConsumer.prototype.generatedPositionFor =
        function IndexedSourceMapConsumer_generatedPositionFor(aArgs) {
            for (var i = 0; i < this._sections.length; i++) {
                var section = this._sections[i];
                // Only consider this section if the requested source is in the list of
                // sources of the consumer.
                if (section.consumer._findSourceIndex(util.getArg(aArgs, 'source')) === -1) {
                    continue;
                }
                var generatedPosition = section.consumer.generatedPositionFor(aArgs);
                if (generatedPosition) {
                    var ret = {
                        line: generatedPosition.line +
                            (section.generatedOffset.generatedLine - 1),
                        column: generatedPosition.column +
                            (section.generatedOffset.generatedLine === generatedPosition.line
                                ? section.generatedOffset.generatedColumn - 1
                                : 0)
                    };
                    return ret;
                }
            }
            return {
                line: null,
                column: null
            };
        };
    /**
     * Parse the mappings in a string in to a data structure which we can easily
     * query (the ordered arrays in the `this.__generatedMappings` and
     * `this.__originalMappings` properties).
     */
    IndexedSourceMapConsumer.prototype._parseMappings =
        function IndexedSourceMapConsumer_parseMappings(aStr, aSourceRoot) {
            this.__generatedMappings = [];
            this.__originalMappings = [];
            for (var i = 0; i < this._sections.length; i++) {
                var section = this._sections[i];
                var sectionMappings = section.consumer._generatedMappings;
                for (var j = 0; j < sectionMappings.length; j++) {
                    var mapping = sectionMappings[j];
                    var source = section.consumer._sources.at(mapping.source);
                    source = util.computeSourceURL(section.consumer.sourceRoot, source, this._sourceMapURL);
                    this._sources.add(source);
                    source = this._sources.indexOf(source);
                    var name = null;
                    if (mapping.name) {
                        name = section.consumer._names.at(mapping.name);
                        this._names.add(name);
                        name = this._names.indexOf(name);
                    }
                    // The mappings coming from the consumer for the section have
                    // generated positions relative to the start of the section, so we
                    // need to offset them to be relative to the start of the concatenated
                    // generated file.
                    var adjustedMapping = {
                        source: source,
                        generatedLine: mapping.generatedLine +
                            (section.generatedOffset.generatedLine - 1),
                        generatedColumn: mapping.generatedColumn +
                            (section.generatedOffset.generatedLine === mapping.generatedLine
                                ? section.generatedOffset.generatedColumn - 1
                                : 0),
                        originalLine: mapping.originalLine,
                        originalColumn: mapping.originalColumn,
                        name: name
                    };
                    this.__generatedMappings.push(adjustedMapping);
                    if (typeof adjustedMapping.originalLine === 'number') {
                        this.__originalMappings.push(adjustedMapping);
                    }
                }
            }
            quickSort(this.__generatedMappings, util.compareByGeneratedPositionsDeflated);
            quickSort(this.__originalMappings, util.compareByOriginalPositions);
        };
    exports.IndexedSourceMapConsumer = IndexedSourceMapConsumer;
},
/* source-map/lib/binary-search.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    exports.GREATEST_LOWER_BOUND = 1;
    exports.LEAST_UPPER_BOUND = 2;
    /**
     * Recursive implementation of binary search.
     *
     * @param aLow Indices here and lower do not contain the needle.
     * @param aHigh Indices here and higher do not contain the needle.
     * @param aNeedle The element being searched for.
     * @param aHaystack The non-empty array being searched.
     * @param aCompare Function which takes two elements and returns -1, 0, or 1.
     * @param aBias Either 'binarySearch.GREATEST_LOWER_BOUND' or
     *     'binarySearch.LEAST_UPPER_BOUND'. Specifies whether to return the
     *     closest element that is smaller than or greater than the one we are
     *     searching for, respectively, if the exact element cannot be found.
     */
    function recursiveSearch(aLow, aHigh, aNeedle, aHaystack, aCompare, aBias) {
        // This function terminates when one of the following is true:
        //
        //   1. We find the exact element we are looking for.
        //
        //   2. We did not find the exact element, but we can return the index of
        //      the next-closest element.
        //
        //   3. We did not find the exact element, and there is no next-closest
        //      element than the one we are searching for, so we return -1.
        var mid = Math.floor((aHigh - aLow) / 2) + aLow;
        var cmp = aCompare(aNeedle, aHaystack[mid], true);
        if (cmp === 0) {
            // Found the element we are looking for.
            return mid;
        }
        else if (cmp > 0) {
            // Our needle is greater than aHaystack[mid].
            if (aHigh - mid > 1) {
                // The element is in the upper half.
                return recursiveSearch(mid, aHigh, aNeedle, aHaystack, aCompare, aBias);
            }
            // The exact needle element was not found in this haystack. Determine if
            // we are in termination case (3) or (2) and return the appropriate thing.
            if (aBias == exports.LEAST_UPPER_BOUND) {
                return aHigh < aHaystack.length ? aHigh : -1;
            }
            else {
                return mid;
            }
        }
        else {
            // Our needle is less than aHaystack[mid].
            if (mid - aLow > 1) {
                // The element is in the lower half.
                return recursiveSearch(aLow, mid, aNeedle, aHaystack, aCompare, aBias);
            }
            // we are in termination case (3) or (2) and return the appropriate thing.
            if (aBias == exports.LEAST_UPPER_BOUND) {
                return mid;
            }
            else {
                return aLow < 0 ? -1 : aLow;
            }
        }
    }
    /**
     * This is an implementation of binary search which will always try and return
     * the index of the closest element if there is no exact hit. This is because
     * mappings between original and generated line/col pairs are single points,
     * and there is an implicit region between each of them, so a miss just means
     * that you aren't on the very start of a region.
     *
     * @param aNeedle The element you are looking for.
     * @param aHaystack The array that is being searched.
     * @param aCompare A function which takes the needle and an element in the
     *     array and returns -1, 0, or 1 depending on whether the needle is less
     *     than, equal to, or greater than the element, respectively.
     * @param aBias Either 'binarySearch.GREATEST_LOWER_BOUND' or
     *     'binarySearch.LEAST_UPPER_BOUND'. Specifies whether to return the
     *     closest element that is smaller than or greater than the one we are
     *     searching for, respectively, if the exact element cannot be found.
     *     Defaults to 'binarySearch.GREATEST_LOWER_BOUND'.
     */
    exports.search = function search(aNeedle, aHaystack, aCompare, aBias) {
        if (aHaystack.length === 0) {
            return -1;
        }
        var index = recursiveSearch(-1, aHaystack.length, aNeedle, aHaystack, aCompare, aBias || exports.GREATEST_LOWER_BOUND);
        if (index < 0) {
            return -1;
        }
        // We have found either the exact element, or the next-closest element than
        // the one we are searching for. However, there may be more than one such
        // element. Make sure we always return the smallest of these.
        while (index - 1 >= 0) {
            if (aCompare(aHaystack[index], aHaystack[index - 1], true) !== 0) {
                break;
            }
            --index;
        }
        return index;
    };
},
/* source-map/lib/quick-sort.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    // It turns out that some (most?) JavaScript engines don't self-host
    // `Array.prototype.sort`. This makes sense because C++ will likely remain
    // faster than JS when doing raw CPU-intensive sorting. However, when using a
    // custom comparator function, calling back and forth between the VM's C++ and
    // JIT'd JS is rather slow *and* loses JIT type information, resulting in
    // worse generated code for the comparator function than would be optimal. In
    // fact, when sorting with a comparator, these costs outweigh the benefits of
    // sorting in C++. By using our own JS-implemented Quick Sort (below), we get
    // a ~3500ms mean speed-up in `bench/bench.html`.
    /**
     * Swap the elements indexed by `x` and `y` in the array `ary`.
     *
     * @param {Array} ary
     *        The array.
     * @param {Number} x
     *        The index of the first item.
     * @param {Number} y
     *        The index of the second item.
     */
    function swap(ary, x, y) {
        var temp = ary[x];
        ary[x] = ary[y];
        ary[y] = temp;
    }
    /**
     * Returns a random integer within the range `low .. high` inclusive.
     *
     * @param {Number} low
     *        The lower bound on the range.
     * @param {Number} high
     *        The upper bound on the range.
     */
    function randomIntInRange(low, high) {
        return Math.round(low + (Math.random() * (high - low)));
    }
    /**
     * The Quick Sort algorithm.
     *
     * @param {Array} ary
     *        An array to sort.
     * @param {function} comparator
     *        Function to use to compare two items.
     * @param {Number} p
     *        Start index of the array
     * @param {Number} r
     *        End index of the array
     */
    function doQuickSort(ary, comparator, p, r) {
        // If our lower bound is less than our upper bound, we (1) partition the
        // array into two pieces and (2) recurse on each half. If it is not, this is
        // the empty array and our base case.
        if (p < r) {
            // (1) Partitioning.
            //
            // The partitioning chooses a pivot between `p` and `r` and moves all
            // elements that are less than or equal to the pivot to the before it, and
            // all the elements that are greater than it after it. The effect is that
            // once partition is done, the pivot is in the exact place it will be when
            // the array is put in sorted order, and it will not need to be moved
            // again. This runs in O(n) time.
            // Always choose a random pivot so that an input array which is reverse
            // sorted does not cause O(n^2) running time.
            var pivotIndex = randomIntInRange(p, r);
            var i = p - 1;
            swap(ary, pivotIndex, r);
            var pivot = ary[r];
            // Immediately after `j` is incremented in this loop, the following hold
            // true:
            //
            //   * Every element in `ary[p .. i]` is less than or equal to the pivot.
            //
            //   * Every element in `ary[i+1 .. j-1]` is greater than the pivot.
            for (var j = p; j < r; j++) {
                if (comparator(ary[j], pivot) <= 0) {
                    i += 1;
                    swap(ary, i, j);
                }
            }
            swap(ary, i + 1, j);
            var q = i + 1;
            // (2) Recurse on each half.
            doQuickSort(ary, comparator, p, q - 1);
            doQuickSort(ary, comparator, q + 1, r);
        }
    }
    /**
     * Sort the given array in-place with the given comparator function.
     *
     * @param {Array} ary
     *        An array to sort.
     * @param {function} comparator
     *        Function to use to compare two items.
     */
    exports.quickSort = function (ary, comparator) {
        doQuickSort(ary, comparator, 0, ary.length - 1);
    };
},
/* source-map/lib/source-node.js */ function _(require, module, exports, __esModule, __esExport) {
    /* -*- Mode: js; js-indent-level: 2; -*- */
    /*
     * Copyright 2011 Mozilla Foundation and contributors
     * Licensed under the New BSD license. See LICENSE or:
     * http://opensource.org/licenses/BSD-3-Clause
     */
    var SourceMapGenerator = require(24) /* ./source-map-generator */.SourceMapGenerator;
    var util = require(27) /* ./util */;
    // Matches a Windows-style `\r\n` newline or a `\n` newline used by all other
    // operating systems these days (capturing the result).
    var REGEX_NEWLINE = /(\r?\n)/;
    // Newline character code for charCodeAt() comparisons
    var NEWLINE_CODE = 10;
    // Private symbol for identifying `SourceNode`s when multiple versions of
    // the source-map library are loaded. This MUST NOT CHANGE across
    // versions!
    var isSourceNode = "$$$isSourceNode$$$";
    /**
     * SourceNodes provide a way to abstract over interpolating/concatenating
     * snippets of generated JavaScript source code while maintaining the line and
     * column information associated with the original source code.
     *
     * @param aLine The original line number.
     * @param aColumn The original column number.
     * @param aSource The original source's filename.
     * @param aChunks Optional. An array of strings which are snippets of
     *        generated JS, or other SourceNodes.
     * @param aName The original identifier.
     */
    function SourceNode(aLine, aColumn, aSource, aChunks, aName) {
        this.children = [];
        this.sourceContents = {};
        this.line = aLine == null ? null : aLine;
        this.column = aColumn == null ? null : aColumn;
        this.source = aSource == null ? null : aSource;
        this.name = aName == null ? null : aName;
        this[isSourceNode] = true;
        if (aChunks != null)
            this.add(aChunks);
    }
    /**
     * Creates a SourceNode from generated code and a SourceMapConsumer.
     *
     * @param aGeneratedCode The generated code
     * @param aSourceMapConsumer The SourceMap for the generated code
     * @param aRelativePath Optional. The path that relative sources in the
     *        SourceMapConsumer should be relative to.
     */
    SourceNode.fromStringWithSourceMap =
        function SourceNode_fromStringWithSourceMap(aGeneratedCode, aSourceMapConsumer, aRelativePath) {
            // The SourceNode we want to fill with the generated code
            // and the SourceMap
            var node = new SourceNode();
            // All even indices of this array are one line of the generated code,
            // while all odd indices are the newlines between two adjacent lines
            // (since `REGEX_NEWLINE` captures its match).
            // Processed fragments are accessed by calling `shiftNextLine`.
            var remainingLines = aGeneratedCode.split(REGEX_NEWLINE);
            var remainingLinesIndex = 0;
            var shiftNextLine = function () {
                var lineContents = getNextLine();
                // The last line of a file might not have a newline.
                var newLine = getNextLine() || "";
                return lineContents + newLine;
                function getNextLine() {
                    return remainingLinesIndex < remainingLines.length ?
                        remainingLines[remainingLinesIndex++] : undefined;
                }
            };
            // We need to remember the position of "remainingLines"
            var lastGeneratedLine = 1, lastGeneratedColumn = 0;
            // The generate SourceNodes we need a code range.
            // To extract it current and last mapping is used.
            // Here we store the last mapping.
            var lastMapping = null;
            aSourceMapConsumer.eachMapping(function (mapping) {
                if (lastMapping !== null) {
                    // We add the code from "lastMapping" to "mapping":
                    // First check if there is a new line in between.
                    if (lastGeneratedLine < mapping.generatedLine) {
                        // Associate first line with "lastMapping"
                        addMappingWithCode(lastMapping, shiftNextLine());
                        lastGeneratedLine++;
                        lastGeneratedColumn = 0;
                        // The remaining code is added without mapping
                    }
                    else {
                        // There is no new line in between.
                        // Associate the code between "lastGeneratedColumn" and
                        // "mapping.generatedColumn" with "lastMapping"
                        var nextLine = remainingLines[remainingLinesIndex] || '';
                        var code = nextLine.substr(0, mapping.generatedColumn -
                            lastGeneratedColumn);
                        remainingLines[remainingLinesIndex] = nextLine.substr(mapping.generatedColumn -
                            lastGeneratedColumn);
                        lastGeneratedColumn = mapping.generatedColumn;
                        addMappingWithCode(lastMapping, code);
                        // No more remaining code, continue
                        lastMapping = mapping;
                        return;
                    }
                }
                // We add the generated code until the first mapping
                // to the SourceNode without any mapping.
                // Each line is added as separate string.
                while (lastGeneratedLine < mapping.generatedLine) {
                    node.add(shiftNextLine());
                    lastGeneratedLine++;
                }
                if (lastGeneratedColumn < mapping.generatedColumn) {
                    var nextLine = remainingLines[remainingLinesIndex] || '';
                    node.add(nextLine.substr(0, mapping.generatedColumn));
                    remainingLines[remainingLinesIndex] = nextLine.substr(mapping.generatedColumn);
                    lastGeneratedColumn = mapping.generatedColumn;
                }
                lastMapping = mapping;
            }, this);
            // We have processed all mappings.
            if (remainingLinesIndex < remainingLines.length) {
                if (lastMapping) {
                    // Associate the remaining code in the current line with "lastMapping"
                    addMappingWithCode(lastMapping, shiftNextLine());
                }
                // and add the remaining lines without any mapping
                node.add(remainingLines.splice(remainingLinesIndex).join(""));
            }
            // Copy sourcesContent into SourceNode
            aSourceMapConsumer.sources.forEach(function (sourceFile) {
                var content = aSourceMapConsumer.sourceContentFor(sourceFile);
                if (content != null) {
                    if (aRelativePath != null) {
                        sourceFile = util.join(aRelativePath, sourceFile);
                    }
                    node.setSourceContent(sourceFile, content);
                }
            });
            return node;
            function addMappingWithCode(mapping, code) {
                if (mapping === null || mapping.source === undefined) {
                    node.add(code);
                }
                else {
                    var source = aRelativePath
                        ? util.join(aRelativePath, mapping.source)
                        : mapping.source;
                    node.add(new SourceNode(mapping.originalLine, mapping.originalColumn, source, code, mapping.name));
                }
            }
        };
    /**
     * Add a chunk of generated JS to this source node.
     *
     * @param aChunk A string snippet of generated JS code, another instance of
     *        SourceNode, or an array where each member is one of those things.
     */
    SourceNode.prototype.add = function SourceNode_add(aChunk) {
        if (Array.isArray(aChunk)) {
            aChunk.forEach(function (chunk) {
                this.add(chunk);
            }, this);
        }
        else if (aChunk[isSourceNode] || typeof aChunk === "string") {
            if (aChunk) {
                this.children.push(aChunk);
            }
        }
        else {
            throw new TypeError("Expected a SourceNode, string, or an array of SourceNodes and strings. Got " + aChunk);
        }
        return this;
    };
    /**
     * Add a chunk of generated JS to the beginning of this source node.
     *
     * @param aChunk A string snippet of generated JS code, another instance of
     *        SourceNode, or an array where each member is one of those things.
     */
    SourceNode.prototype.prepend = function SourceNode_prepend(aChunk) {
        if (Array.isArray(aChunk)) {
            for (var i = aChunk.length - 1; i >= 0; i--) {
                this.prepend(aChunk[i]);
            }
        }
        else if (aChunk[isSourceNode] || typeof aChunk === "string") {
            this.children.unshift(aChunk);
        }
        else {
            throw new TypeError("Expected a SourceNode, string, or an array of SourceNodes and strings. Got " + aChunk);
        }
        return this;
    };
    /**
     * Walk over the tree of JS snippets in this node and its children. The
     * walking function is called once for each snippet of JS and is passed that
     * snippet and the its original associated source's line/column location.
     *
     * @param aFn The traversal function.
     */
    SourceNode.prototype.walk = function SourceNode_walk(aFn) {
        var chunk;
        for (var i = 0, len = this.children.length; i < len; i++) {
            chunk = this.children[i];
            if (chunk[isSourceNode]) {
                chunk.walk(aFn);
            }
            else {
                if (chunk !== '') {
                    aFn(chunk, { source: this.source,
                        line: this.line,
                        column: this.column,
                        name: this.name });
                }
            }
        }
    };
    /**
     * Like `String.prototype.join` except for SourceNodes. Inserts `aStr` between
     * each of `this.children`.
     *
     * @param aSep The separator.
     */
    SourceNode.prototype.join = function SourceNode_join(aSep) {
        var newChildren;
        var i;
        var len = this.children.length;
        if (len > 0) {
            newChildren = [];
            for (i = 0; i < len - 1; i++) {
                newChildren.push(this.children[i]);
                newChildren.push(aSep);
            }
            newChildren.push(this.children[i]);
            this.children = newChildren;
        }
        return this;
    };
    /**
     * Call String.prototype.replace on the very right-most source snippet. Useful
     * for trimming whitespace from the end of a source node, etc.
     *
     * @param aPattern The pattern to replace.
     * @param aReplacement The thing to replace the pattern with.
     */
    SourceNode.prototype.replaceRight = function SourceNode_replaceRight(aPattern, aReplacement) {
        var lastChild = this.children[this.children.length - 1];
        if (lastChild[isSourceNode]) {
            lastChild.replaceRight(aPattern, aReplacement);
        }
        else if (typeof lastChild === 'string') {
            this.children[this.children.length - 1] = lastChild.replace(aPattern, aReplacement);
        }
        else {
            this.children.push(''.replace(aPattern, aReplacement));
        }
        return this;
    };
    /**
     * Set the source content for a source file. This will be added to the SourceMapGenerator
     * in the sourcesContent field.
     *
     * @param aSourceFile The filename of the source file
     * @param aSourceContent The content of the source file
     */
    SourceNode.prototype.setSourceContent =
        function SourceNode_setSourceContent(aSourceFile, aSourceContent) {
            this.sourceContents[util.toSetString(aSourceFile)] = aSourceContent;
        };
    /**
     * Walk over the tree of SourceNodes. The walking function is called for each
     * source file content and is passed the filename and source content.
     *
     * @param aFn The traversal function.
     */
    SourceNode.prototype.walkSourceContents =
        function SourceNode_walkSourceContents(aFn) {
            for (var i = 0, len = this.children.length; i < len; i++) {
                if (this.children[i][isSourceNode]) {
                    this.children[i].walkSourceContents(aFn);
                }
            }
            var sources = Object.keys(this.sourceContents);
            for (var i = 0, len = sources.length; i < len; i++) {
                aFn(util.fromSetString(sources[i]), this.sourceContents[sources[i]]);
            }
        };
    /**
     * Return the string representation of this source node. Walks over the tree
     * and concatenates all the various snippets together to one string.
     */
    SourceNode.prototype.toString = function SourceNode_toString() {
        var str = "";
        this.walk(function (chunk) {
            str += chunk;
        });
        return str;
    };
    /**
     * Returns the string representation of this source node along with a source
     * map.
     */
    SourceNode.prototype.toStringWithSourceMap = function SourceNode_toStringWithSourceMap(aArgs) {
        var generated = {
            code: "",
            line: 1,
            column: 0
        };
        var map = new SourceMapGenerator(aArgs);
        var sourceMappingActive = false;
        var lastOriginalSource = null;
        var lastOriginalLine = null;
        var lastOriginalColumn = null;
        var lastOriginalName = null;
        this.walk(function (chunk, original) {
            generated.code += chunk;
            if (original.source !== null
                && original.line !== null
                && original.column !== null) {
                if (lastOriginalSource !== original.source
                    || lastOriginalLine !== original.line
                    || lastOriginalColumn !== original.column
                    || lastOriginalName !== original.name) {
                    map.addMapping({
                        source: original.source,
                        original: {
                            line: original.line,
                            column: original.column
                        },
                        generated: {
                            line: generated.line,
                            column: generated.column
                        },
                        name: original.name
                    });
                }
                lastOriginalSource = original.source;
                lastOriginalLine = original.line;
                lastOriginalColumn = original.column;
                lastOriginalName = original.name;
                sourceMappingActive = true;
            }
            else if (sourceMappingActive) {
                map.addMapping({
                    generated: {
                        line: generated.line,
                        column: generated.column
                    }
                });
                lastOriginalSource = null;
                sourceMappingActive = false;
            }
            for (var idx = 0, length = chunk.length; idx < length; idx++) {
                if (chunk.charCodeAt(idx) === NEWLINE_CODE) {
                    generated.line++;
                    generated.column = 0;
                    // Mappings end at eol
                    if (idx + 1 === length) {
                        lastOriginalSource = null;
                        sourceMappingActive = false;
                    }
                    else if (sourceMappingActive) {
                        map.addMapping({
                            source: original.source,
                            original: {
                                line: original.line,
                                column: original.column
                            },
                            generated: {
                                line: generated.line,
                                column: generated.column
                            },
                            name: original.name
                        });
                    }
                }
                else {
                    generated.column++;
                }
            }
        });
        this.walkSourceContents(function (sourceFile, sourceContent) {
            map.setSourceContent(sourceFile, sourceContent);
        });
        return { code: generated.code, map: map };
    };
    exports.SourceNode = SourceNode;
},
/* buffer-from/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var toString = Object.prototype.toString;
    var isModern = (typeof Buffer.alloc === 'function' &&
        typeof Buffer.allocUnsafe === 'function' &&
        typeof Buffer.from === 'function');
    function isArrayBuffer(input) {
        return toString.call(input).slice(8, -1) === 'ArrayBuffer';
    }
    function fromArrayBuffer(obj, byteOffset, length) {
        byteOffset >>>= 0;
        var maxLength = obj.byteLength - byteOffset;
        if (maxLength < 0) {
            throw new RangeError("'offset' is out of bounds");
        }
        if (length === undefined) {
            length = maxLength;
        }
        else {
            length >>>= 0;
            if (length > maxLength) {
                throw new RangeError("'length' is out of bounds");
            }
        }
        return isModern
            ? Buffer.from(obj.slice(byteOffset, byteOffset + length))
            : new Buffer(new Uint8Array(obj.slice(byteOffset, byteOffset + length)));
    }
    function fromString(string, encoding) {
        if (typeof encoding !== 'string' || encoding === '') {
            encoding = 'utf8';
        }
        if (!Buffer.isEncoding(encoding)) {
            throw new TypeError('"encoding" must be a valid string encoding');
        }
        return isModern
            ? Buffer.from(string, encoding)
            : new Buffer(string, encoding);
    }
    function bufferFrom(value, encodingOrOffset, length) {
        if (typeof value === 'number') {
            throw new TypeError('"value" argument must not be a number');
        }
        if (isArrayBuffer(value)) {
            return fromArrayBuffer(value, encodingOrOffset, length);
        }
        if (typeof value === 'string') {
            return fromString(value, encodingOrOffset);
        }
        return isModern
            ? Buffer.from(value)
            : new Buffer(value);
    }
    module.exports = bufferFrom;
},
/* build.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const cp = __importStar(require("child_process"));
    const path_1 = require("path");
    const sys_1 = require(20) /* ./sys */;
    const compiler_1 = require(36) /* ./compiler */;
    const linker_1 = require(44) /* ./linker */;
    const styles_1 = require(55) /* ./styles */;
    const preludes = __importStar(require(201) /* ./prelude */);
    const tsconfig_json = __importStar(require(202) /* ./tsconfig.ext.json */);
    const chalk_1 = __importDefault(require(37) /* chalk */);
    const { cyan, magenta, red } = chalk_1.default;
    const eslint_1 = require(203) /* eslint */;
    require(732) /* @typescript-eslint/eslint-plugin */;
    require(1117) /* @typescript-eslint/parser */;
    const readline = __importStar(require("readline"));
    const toString = Object.prototype.toString;
    function isString(obj) {
        return toString.call(obj) === "[object String]";
    }
    exports.isString = isString;
    function isObject(obj) {
        const tp = typeof obj;
        return tp === "function" || tp === "object" && !!obj;
    }
    exports.isObject = isObject;
    function isPlainObject(obj) {
        return isObject(obj) && (obj.constructor == null || obj.constructor === Object);
    }
    exports.isPlainObject = isPlainObject;
    function print(str) {
        console.log(str);
    }
    function npm_install(base_dir) {
        const npm = process.platform != "win32" ? "npm" : "npm.cmd";
        const { status } = cp.spawnSync(npm, ["install"], { stdio: "inherit", cwd: base_dir });
        if (status != null && status != 0) {
            print(`${cyan("npm install")} failed with exit code ${red(`${status}`)}.`);
            process.exit(status);
        }
    }
    function is_up_to_date(base_dir, file, metadata) {
        const contents = (0, sys_1.read)((0, path_1.join)(base_dir, file));
        if (contents == null)
            return false;
        const old_hash = metadata.signatures[file];
        if (old_hash == null)
            return false;
        const new_hash = (0, sys_1.hash)(contents);
        return old_hash == new_hash;
    }
    function needs_install(base_dir, metadata) {
        if (!(0, sys_1.directory_exists)((0, path_1.join)(base_dir, "node_modules")))
            return "New development environment.";
        else if (!is_up_to_date(base_dir, "package.json", metadata))
            return "package.json has changed.";
        else if (!is_up_to_date(base_dir, "package-lock.json", metadata))
            return "package-lock.json has changed.";
        else
            return null;
    }
    function lint(config_file, paths) {
        const engine = new eslint_1.CLIEngine({
            configFile: config_file,
            extensions: [".ts", ".js"],
        });
        const report = engine.executeOnFiles(paths);
        eslint_1.CLIEngine.outputFixes(report);
        const ok = report.errorCount == 0;
        if (!ok) {
            const formatter = engine.getFormatter();
            const output = formatter(report.results);
            for (const line of output.trim().split("\n"))
                print(line);
        }
        return ok;
    }
    async function init(base_dir, _bokehjs_dir, base_setup) {
        print(`Working directory: ${cyan(base_dir)}`);
        const setup = {
            interactive: !!base_setup.interactive,
            bokehjs_version: base_setup.bokehjs_version != null ? base_setup.bokehjs_version : base_setup.bokeh_version.split("-")[0],
            bokeh_version: base_setup.bokeh_version,
        };
        const paths = {
            bokeh_ext: (0, path_1.join)(base_dir, "bokeh.ext.json"),
            package: (0, path_1.join)(base_dir, "package.json"),
            package_lock: (0, path_1.join)(base_dir, "package-lock.json"),
            tsconfig: (0, path_1.join)(base_dir, "tsconfig.json"),
            index: (0, path_1.join)(base_dir, "index.ts"),
        };
        const is_extension = (0, sys_1.file_exists)(paths.bokeh_ext);
        if (is_extension) {
            print("Already a bokeh extension. Quitting.");
            return false;
        }
        function write_json(path, json) {
            (0, sys_1.write)(path, JSON.stringify(json, undefined, 2));
            print(`Wrote ${cyan(path)}`);
        }
        const bokeh_ext_json = {};
        write_json(paths.bokeh_ext, bokeh_ext_json);
        const package_json = {
            name: (0, path_1.basename)(base_dir),
            version: "0.0.1",
            description: "",
            license: "BSD-3-Clause",
            keywords: [],
            repository: {},
            dependencies: {
                bokehjs: `^${setup.bokehjs_version}`,
            },
            devDependencies: {},
        };
        if (setup.interactive) {
            const rl = readline.createInterface({ input: process.stdin, output: process.stdout });
            async function ask(question, default_value) {
                return new Promise((resolve, _reject) => {
                    rl.question(`${question} `, (answer) => {
                        resolve(answer.length != 0 ? answer : default_value);
                    });
                });
            }
            async function ask_yn(question) {
                const ret = await ask(`${question} [y/n]`, "y");
                switch (ret) {
                    case "y":
                        return true;
                    case "n":
                        return false;
                    default: {
                        print(`${red("Invalid input")}. Assuming no.`);
                        return false;
                    }
                }
            }
            if (await ask_yn(`Create ${cyan("package.json")}? This will allow you to specify external dependencies.`)) {
                const { name } = package_json;
                package_json.name = await ask(`  What's the extension's name? [${name}]`, name);
                const { version } = package_json;
                package_json.version = await ask(`  What's the extension's version? [${version}]`, version);
                const { description } = package_json;
                package_json.description = await ask(`  What's the extension's description? [${description}]`, description);
                write_json(paths.package, package_json);
            }
            if (await ask_yn(`Create ${cyan("tsconfig.json")}? This will allow for customized configuration and improved IDE experience.`)) {
                write_json(paths.tsconfig, tsconfig_json);
            }
            rl.close();
        }
        else {
            write_json(paths.package, package_json);
            write_json(paths.tsconfig, tsconfig_json);
        }
        (0, sys_1.write)(paths.index, "");
        print(`Created empty ${cyan("index.ts")}. This is the entry point of your extension.`);
        const rel = (0, path_1.relative)(process.cwd(), base_dir);
        print(`You can build your extension with ${magenta(`bokeh build ${rel}`)}`);
        print("All done.");
        return true;
    }
    exports.init = init;
    async function build(base_dir, bokehjs_dir, base_setup) {
        print(`Working directory: ${cyan(base_dir)}`);
        const setup = {
            rebuild: !!base_setup.rebuild,
            bokeh_version: base_setup.bokeh_version,
        };
        const bokeh_ext_json_path = (0, path_1.join)(base_dir, "bokeh.ext.json");
        const bokeh_ext = (0, sys_1.read_json)(bokeh_ext_json_path);
        if (!isPlainObject(bokeh_ext)) {
            print("Not a bokeh extension. Quitting.");
            return false;
        }
        const metadata_path = (0, path_1.join)(base_dir, ".bokeh");
        const metadata = (() => {
            let obj = (0, sys_1.read_json)(metadata_path);
            if (obj == null)
                obj = {};
            if (obj.signatures == null)
                obj.signatures = {};
            return obj;
        })();
        if (metadata.bokeh_version != setup.bokeh_version) {
            print("Using different version of bokeh, rebuilding from scratch.");
            setup.rebuild = true;
        }
        const package_json_path = (0, path_1.join)(base_dir, "package.json");
        const package_lock_json_path = (0, path_1.join)(base_dir, "package-lock.json");
        const is_package = (0, sys_1.file_exists)(package_json_path);
        if (!is_package) {
            print(`${cyan(package_json_path)} doesn't exist. Not a npm package.`);
        }
        else {
            if (setup.rebuild) {
                print(`Running ${cyan("npm install")}.`);
                npm_install(base_dir);
            }
            else {
                const result = needs_install(base_dir, metadata);
                if (result != null) {
                    print(`${result} Running ${cyan("npm install")}.`);
                    npm_install(base_dir);
                }
            }
        }
        const tsconfig_path = (0, path_1.join)(base_dir, "tsconfig.json");
        const tsconfig = (() => {
            const preconfigure = {
                baseUrl: base_dir,
                paths: {
                    "@bokehjs/*": [
                        (0, path_1.join)(bokehjs_dir, "js/lib/*"),
                        (0, path_1.join)(bokehjs_dir, "js/types/*"),
                    ],
                },
            };
            if ((0, sys_1.file_exists)(tsconfig_path)) {
                print(`Using ${cyan(tsconfig_path)}`);
                return (0, compiler_1.read_tsconfig)(tsconfig_path, is_package ? undefined : preconfigure);
            }
            else
                return (0, compiler_1.parse_tsconfig)(tsconfig_json, base_dir, preconfigure);
        })();
        if ((0, compiler_1.is_failed)(tsconfig)) {
            print((0, compiler_1.report_diagnostics)(tsconfig.diagnostics).text);
            return false;
        }
        let success = true;
        const { files, options } = tsconfig;
        const dist_dir = (0, path_1.join)(base_dir, "dist");
        const lib_dir = options.outDir ?? dist_dir;
        const dts_dir = options.declarationDir ?? lib_dir;
        const styles_dir = (0, path_1.join)(base_dir, "styles");
        const css_dir = (0, path_1.join)(dist_dir, "css");
        print("Compiling styles");
        if (!await (0, styles_1.compile_styles)(styles_dir, css_dir))
            success = false;
        (0, styles_1.wrap_css_modules)(css_dir, lib_dir, dts_dir);
        const transformers = (0, compiler_1.default_transformers)(options);
        const host = (0, compiler_1.compiler_host)(new Map(), options, bokehjs_dir);
        print(`Compiling TypeScript (${magenta(`${files.length} files`)})`);
        const tsoutput = (0, compiler_1.compile_files)(files, options, transformers, host);
        if ((0, compiler_1.is_failed)(tsoutput)) {
            print((0, compiler_1.report_diagnostics)(tsoutput.diagnostics).text);
            if (options.noEmitOnError)
                return false;
        }
        const lint_config = (0, path_1.join)(base_dir, "eslint.json");
        if ((0, sys_1.file_exists)(lint_config)) {
            print("Linting sources");
            lint(lint_config, files);
        }
        const artifact = (0, path_1.basename)(base_dir);
        const bases = [lib_dir];
        if (is_package)
            bases.push((0, path_1.join)(base_dir, "node_modules"));
        const linker = new linker_1.Linker({
            entries: [(0, path_1.join)(lib_dir, "index.js")],
            bases,
            cache: (0, path_1.join)(dist_dir, `${artifact}.json`),
            excluded: (dep) => dep == "tslib" || dep.startsWith("@bokehjs/"),
            plugin: true,
            target: "ES2017",
        });
        print("Linking modules");
        if (!setup.rebuild)
            linker.load_cache();
        const { bundles, status } = await linker.link();
        if (!status)
            success = false;
        linker.store_cache();
        const outputs = [(0, path_1.join)(dist_dir, `${artifact}.js`)];
        const min_js = (js) => (0, sys_1.rename)(js, { ext: ".min.js" });
        const license = (() => {
            if (isPlainObject(bokeh_ext.license)) {
                if (isString(bokeh_ext.license.file)) {
                    const license_path = (0, path_1.join)(base_dir, bokeh_ext.license.file);
                    const text = (0, sys_1.read)(license_path);
                    if (text != null)
                        return text;
                    else
                        print(`Failed to license text from ${magenta(license_path)}`);
                }
                if (isString(bokeh_ext.license.text)) {
                    return bokeh_ext.license.text;
                }
            }
            return null;
        })();
        const license_text = license ? `${preludes.comment(license)}\n` : "";
        const prelude_base = `${license_text}${preludes.plugin_prelude()}`;
        const prelude = { main: prelude_base, plugin: prelude_base };
        const postlude_base = preludes.plugin_postlude();
        const postlude = { main: postlude_base, plugin: postlude_base };
        // HACK {{{
        for (const bundle of bundles) {
            bundle.bases.push({});
        }
        // }}}
        function bundle(minified, outputs) {
            bundles
                .map((bundle) => bundle.assemble({ prelude, postlude, minified }))
                .map((artifact, i) => artifact.write(outputs[i]));
        }
        bundle(false, outputs);
        bundle(true, outputs.map(min_js));
        (0, sys_1.write)(metadata_path, JSON.stringify({
            bokeh_version: setup.bokeh_version,
            signatures: {
                "package.json": (0, sys_1.hash_file)(package_json_path),
                "package-lock.json": (0, sys_1.hash_file)(package_lock_json_path),
                "tsconfig.json": (0, sys_1.hash_file)(tsconfig_path),
            },
        }));
        print(`Output written to ${cyan(dist_dir)}`);
        print("All done.");
        return !(0, compiler_1.is_failed)(tsoutput) && success;
    }
    exports.build = build;
},
/* compiler.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const chalk_1 = __importDefault(require(37) /* chalk */);
    const ts = __importStar(require(21) /* typescript */);
    const path_1 = require("path");
    const transforms = __importStar(require(42) /* ./transforms */);
    const error_1 = require(43) /* ./error */;
    function is_failed(obj) {
        return "diagnostics" in obj && obj.diagnostics != null;
    }
    exports.is_failed = is_failed;
    function normalize(path) {
        return path.replace(/\\/g, "/");
    }
    const diagnostics_host = {
        getCanonicalFileName: (path) => path,
        getCurrentDirectory: ts.sys.getCurrentDirectory,
        getNewLine: () => ts.sys.newLine,
    };
    function report_diagnostics(diagnostics) {
        const errors = ts.sortAndDeduplicateDiagnostics(diagnostics);
        const text = ts.formatDiagnosticsWithColorAndContext(errors, diagnostics_host);
        return { count: errors.length, text };
    }
    exports.report_diagnostics = report_diagnostics;
    function compiler_host(inputs, options, bokehjs_dir) {
        const default_host = ts.createIncrementalCompilerHost(options);
        const host = {
            ...default_host,
            fileExists(name) {
                return inputs.get(name) != null || default_host.fileExists(name);
            },
            readFile(name) {
                return inputs.get(name) != null ? inputs.get(name) : default_host.readFile(name);
            },
            getSourceFile(name, target, _onError) {
                const source = inputs.get(name);
                if (source != null) {
                    const sf = ts.createSourceFile(name, source, target);
                    const version = default_host.createHash(source);
                    return { ...sf, version }; // version is internal to the compiler
                }
                else
                    return default_host.getSourceFile(name, target, _onError);
            },
        };
        if (bokehjs_dir != null) {
            host.getDefaultLibLocation = () => {
                // bokeh/server/static or bokehjs/build
                if ((0, path_1.basename)(bokehjs_dir) == "static")
                    return (0, path_1.join)(bokehjs_dir, "lib");
                else
                    return (0, path_1.join)((0, path_1.dirname)(bokehjs_dir), "node_modules/typescript/lib");
            };
        }
        return host;
    }
    exports.compiler_host = compiler_host;
    function default_transformers(options) {
        const transformers = {
            before: [],
            after: [],
            afterDeclarations: [],
        };
        const insert_class_name = transforms.insert_class_name();
        transformers.before.push(insert_class_name);
        // TODO: remove this in 3.0
        const add_init_class = transforms.add_init_class();
        transformers.before.push(add_init_class);
        const base = options.baseUrl;
        if (base != null) {
            const relativize_modules = transforms.relativize_modules((file, module_path) => {
                if (!module_path.startsWith(".") && !module_path.startsWith("/")) {
                    const module_file = (0, path_1.join)(base, module_path);
                    if (ts.sys.fileExists(module_file) ||
                        ts.sys.fileExists(`${module_file}.ts`) ||
                        ts.sys.fileExists((0, path_1.join)(module_file, "index.ts")) ||
                        options.outDir != null && ts.sys.fileExists((0, path_1.join)(options.outDir, `${module_path}.js`))) {
                        const rel_path = normalize((0, path_1.relative)((0, path_1.dirname)(file), module_file));
                        return rel_path.startsWith(".") ? rel_path : `./${rel_path}`;
                    }
                }
                return null;
            });
            transformers.after.push(relativize_modules);
            transformers.afterDeclarations.push(relativize_modules);
        }
        return transformers;
    }
    exports.default_transformers = default_transformers;
    function compile_files(inputs, options, transformers, host) {
        const program = ts.createIncrementalProgram({ rootNames: inputs, options, host });
        const emitted = program.emit(undefined, undefined, undefined, false, transformers);
        const diagnostics = [
            ...program.getConfigFileParsingDiagnostics(),
            ...program.getSyntacticDiagnostics(),
            ...program.getOptionsDiagnostics(),
            ...program.getGlobalDiagnostics(),
            ...program.getSemanticDiagnostics(),
            ...emitted.diagnostics,
        ];
        return diagnostics.length != 0 ? { diagnostics } : {};
    }
    exports.compile_files = compile_files;
    function parse_tsconfig(tsconfig_json, base_dir, preconfigure) {
        const host = {
            useCaseSensitiveFileNames: ts.sys.useCaseSensitiveFileNames,
            readDirectory: ts.sys.readDirectory,
            fileExists: ts.sys.fileExists,
            readFile: ts.sys.readFile,
        };
        const tsconfig = ts.parseJsonConfigFileContent(tsconfig_json, host, base_dir, preconfigure);
        if (tsconfig.errors.length != 0) {
            return { diagnostics: tsconfig.errors };
        }
        return { files: tsconfig.fileNames, options: tsconfig.options };
    }
    exports.parse_tsconfig = parse_tsconfig;
    function read_tsconfig(tsconfig_path, preconfigure) {
        const tsconfig_file = ts.readConfigFile(tsconfig_path, ts.sys.readFile);
        if (tsconfig_file.error != null) {
            return { diagnostics: [tsconfig_file.error] };
        }
        return parse_tsconfig(tsconfig_file.config, (0, path_1.dirname)(tsconfig_path), preconfigure);
    }
    exports.read_tsconfig = read_tsconfig;
    function compile_project(tsconfig_path, config) {
        const tsconfig = read_tsconfig(tsconfig_path);
        if (is_failed(tsconfig))
            return { diagnostics: tsconfig.diagnostics };
        const { files, options } = tsconfig;
        const transformers = default_transformers(tsconfig.options);
        const inputs = config.inputs?.(files) ?? new Map();
        const host = compiler_host(inputs, options, config.bokehjs_dir);
        const input_files = [...inputs.keys(), ...files];
        return compile_files(input_files, options, transformers, host);
    }
    function compile_typescript(tsconfig_path, config = {}) {
        const result = compile_project(tsconfig_path, config);
        if (is_failed(result)) {
            const { count, text } = report_diagnostics(result.diagnostics);
            throw new error_1.BuildError("typescript", `There were ${chalk_1.default.red(`${count}`)} TypeScript errors:\n${text}`);
        }
    }
    exports.compile_typescript = compile_typescript;
},
/* chalk/source/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const ansiStyles = require(12) /* ansi-styles */;
    const { stdout: stdoutColor, stderr: stderrColor } = require(38) /* supports-color */;
    const { stringReplaceAll, stringEncaseCRLFWithFirstIndex } = require(40) /* ./util */;
    const { isArray } = Array;
    // `supportsColor.level` → `ansiStyles.color[name]` mapping
    const levelMapping = [
        'ansi',
        'ansi',
        'ansi256',
        'ansi16m'
    ];
    const styles = Object.create(null);
    const applyOptions = (object, options = {}) => {
        if (options.level && !(Number.isInteger(options.level) && options.level >= 0 && options.level <= 3)) {
            throw new Error('The `level` option should be an integer from 0 to 3');
        }
        // Detect level if not set manually
        const colorLevel = stdoutColor ? stdoutColor.level : 0;
        object.level = options.level === undefined ? colorLevel : options.level;
    };
    class ChalkClass {
        constructor(options) {
            // eslint-disable-next-line no-constructor-return
            return chalkFactory(options);
        }
    }
    const chalkFactory = options => {
        const chalk = {};
        applyOptions(chalk, options);
        chalk.template = (...arguments_) => chalkTag(chalk.template, ...arguments_);
        Object.setPrototypeOf(chalk, Chalk.prototype);
        Object.setPrototypeOf(chalk.template, chalk);
        chalk.template.constructor = () => {
            throw new Error('`chalk.constructor()` is deprecated. Use `new chalk.Instance()` instead.');
        };
        chalk.template.Instance = ChalkClass;
        return chalk.template;
    };
    function Chalk(options) {
        return chalkFactory(options);
    }
    for (const [styleName, style] of Object.entries(ansiStyles)) {
        styles[styleName] = {
            get() {
                const builder = createBuilder(this, createStyler(style.open, style.close, this._styler), this._isEmpty);
                Object.defineProperty(this, styleName, { value: builder });
                return builder;
            }
        };
    }
    styles.visible = {
        get() {
            const builder = createBuilder(this, this._styler, true);
            Object.defineProperty(this, 'visible', { value: builder });
            return builder;
        }
    };
    const usedModels = ['rgb', 'hex', 'keyword', 'hsl', 'hsv', 'hwb', 'ansi', 'ansi256'];
    for (const model of usedModels) {
        styles[model] = {
            get() {
                const { level } = this;
                return function (...arguments_) {
                    const styler = createStyler(ansiStyles.color[levelMapping[level]][model](...arguments_), ansiStyles.color.close, this._styler);
                    return createBuilder(this, styler, this._isEmpty);
                };
            }
        };
    }
    for (const model of usedModels) {
        const bgModel = 'bg' + model[0].toUpperCase() + model.slice(1);
        styles[bgModel] = {
            get() {
                const { level } = this;
                return function (...arguments_) {
                    const styler = createStyler(ansiStyles.bgColor[levelMapping[level]][model](...arguments_), ansiStyles.bgColor.close, this._styler);
                    return createBuilder(this, styler, this._isEmpty);
                };
            }
        };
    }
    const proto = Object.defineProperties(() => { }, {
        ...styles,
        level: {
            enumerable: true,
            get() {
                return this._generator.level;
            },
            set(level) {
                this._generator.level = level;
            }
        }
    });
    const createStyler = (open, close, parent) => {
        let openAll;
        let closeAll;
        if (parent === undefined) {
            openAll = open;
            closeAll = close;
        }
        else {
            openAll = parent.openAll + open;
            closeAll = close + parent.closeAll;
        }
        return {
            open,
            close,
            openAll,
            closeAll,
            parent
        };
    };
    const createBuilder = (self, _styler, _isEmpty) => {
        const builder = (...arguments_) => {
            if (isArray(arguments_[0]) && isArray(arguments_[0].raw)) {
                // Called as a template literal, for example: chalk.red`2 + 3 = {bold ${2+3}}`
                return applyStyle(builder, chalkTag(builder, ...arguments_));
            }
            // Single argument is hot path, implicit coercion is faster than anything
            // eslint-disable-next-line no-implicit-coercion
            return applyStyle(builder, (arguments_.length === 1) ? ('' + arguments_[0]) : arguments_.join(' '));
        };
        // We alter the prototype because we must return a function, but there is
        // no way to create a function with a different prototype
        Object.setPrototypeOf(builder, proto);
        builder._generator = self;
        builder._styler = _styler;
        builder._isEmpty = _isEmpty;
        return builder;
    };
    const applyStyle = (self, string) => {
        if (self.level <= 0 || !string) {
            return self._isEmpty ? '' : string;
        }
        let styler = self._styler;
        if (styler === undefined) {
            return string;
        }
        const { openAll, closeAll } = styler;
        if (string.indexOf('\u001B') !== -1) {
            while (styler !== undefined) {
                // Replace any instances already present with a re-opening code
                // otherwise only the part of the string until said closing code
                // will be colored, and the rest will simply be 'plain'.
                string = stringReplaceAll(string, styler.close, styler.open);
                styler = styler.parent;
            }
        }
        // We can move both next actions out of loop, because remaining actions in loop won't have
        // any/visible effect on parts we add here. Close the styling before a linebreak and reopen
        // after next line to fix a bleed issue on macOS: https://github.com/chalk/chalk/pull/92
        const lfIndex = string.indexOf('\n');
        if (lfIndex !== -1) {
            string = stringEncaseCRLFWithFirstIndex(string, closeAll, openAll, lfIndex);
        }
        return openAll + string + closeAll;
    };
    let template;
    const chalkTag = (chalk, ...strings) => {
        const [firstString] = strings;
        if (!isArray(firstString) || !isArray(firstString.raw)) {
            // If chalk() was called by itself or with a string,
            // return the string itself as a string.
            return strings.join(' ');
        }
        const arguments_ = strings.slice(1);
        const parts = [firstString.raw[0]];
        for (let i = 1; i < firstString.length; i++) {
            parts.push(String(arguments_[i - 1]).replace(/[{}\\]/g, '\\$&'), String(firstString.raw[i]));
        }
        if (template === undefined) {
            template = require(41) /* ./templates */;
        }
        return template(chalk, parts.join(''));
    };
    Object.defineProperties(Chalk.prototype, styles);
    const chalk = Chalk(); // eslint-disable-line new-cap
    chalk.supportsColor = stdoutColor;
    chalk.stderr = Chalk({ level: stderrColor ? stderrColor.level : 0 }); // eslint-disable-line new-cap
    chalk.stderr.supportsColor = stderrColor;
    module.exports = chalk;
},
/* supports-color/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const os = require('os');
    const tty = require('tty');
    const hasFlag = require(39) /* has-flag */;
    const { env } = process;
    let forceColor;
    if (hasFlag('no-color') ||
        hasFlag('no-colors') ||
        hasFlag('color=false') ||
        hasFlag('color=never')) {
        forceColor = 0;
    }
    else if (hasFlag('color') ||
        hasFlag('colors') ||
        hasFlag('color=true') ||
        hasFlag('color=always')) {
        forceColor = 1;
    }
    if ('FORCE_COLOR' in env) {
        if (env.FORCE_COLOR === 'true') {
            forceColor = 1;
        }
        else if (env.FORCE_COLOR === 'false') {
            forceColor = 0;
        }
        else {
            forceColor = env.FORCE_COLOR.length === 0 ? 1 : Math.min(parseInt(env.FORCE_COLOR, 10), 3);
        }
    }
    function translateLevel(level) {
        if (level === 0) {
            return false;
        }
        return {
            level,
            hasBasic: true,
            has256: level >= 2,
            has16m: level >= 3
        };
    }
    function supportsColor(haveStream, streamIsTTY) {
        if (forceColor === 0) {
            return 0;
        }
        if (hasFlag('color=16m') ||
            hasFlag('color=full') ||
            hasFlag('color=truecolor')) {
            return 3;
        }
        if (hasFlag('color=256')) {
            return 2;
        }
        if (haveStream && !streamIsTTY && forceColor === undefined) {
            return 0;
        }
        const min = forceColor || 0;
        if (env.TERM === 'dumb') {
            return min;
        }
        if (process.platform === 'win32') {
            // Windows 10 build 10586 is the first Windows release that supports 256 colors.
            // Windows 10 build 14931 is the first release that supports 16m/TrueColor.
            const osRelease = os.release().split('.');
            if (Number(osRelease[0]) >= 10 &&
                Number(osRelease[2]) >= 10586) {
                return Number(osRelease[2]) >= 14931 ? 3 : 2;
            }
            return 1;
        }
        if ('CI' in env) {
            if (['TRAVIS', 'CIRCLECI', 'APPVEYOR', 'GITLAB_CI', 'GITHUB_ACTIONS', 'BUILDKITE'].some(sign => sign in env) || env.CI_NAME === 'codeship') {
                return 1;
            }
            return min;
        }
        if ('TEAMCITY_VERSION' in env) {
            return /^(9\.(0*[1-9]\d*)\.|\d{2,}\.)/.test(env.TEAMCITY_VERSION) ? 1 : 0;
        }
        if (env.COLORTERM === 'truecolor') {
            return 3;
        }
        if ('TERM_PROGRAM' in env) {
            const version = parseInt((env.TERM_PROGRAM_VERSION || '').split('.')[0], 10);
            switch (env.TERM_PROGRAM) {
                case 'iTerm.app':
                    return version >= 3 ? 3 : 2;
                case 'Apple_Terminal':
                    return 2;
                // No default
            }
        }
        if (/-256(color)?$/i.test(env.TERM)) {
            return 2;
        }
        if (/^screen|^xterm|^vt100|^vt220|^rxvt|color|ansi|cygwin|linux/i.test(env.TERM)) {
            return 1;
        }
        if ('COLORTERM' in env) {
            return 1;
        }
        return min;
    }
    function getSupportLevel(stream) {
        const level = supportsColor(stream, stream && stream.isTTY);
        return translateLevel(level);
    }
    module.exports = {
        supportsColor: getSupportLevel,
        stdout: translateLevel(supportsColor(true, tty.isatty(1))),
        stderr: translateLevel(supportsColor(true, tty.isatty(2)))
    };
},
/* has-flag/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = (flag, argv = process.argv) => {
        const prefix = flag.startsWith('-') ? '' : (flag.length === 1 ? '-' : '--');
        const position = argv.indexOf(prefix + flag);
        const terminatorPosition = argv.indexOf('--');
        return position !== -1 && (terminatorPosition === -1 || position < terminatorPosition);
    };
},
/* chalk/source/util.js */ function _(require, module, exports, __esModule, __esExport) {
    const stringReplaceAll = (string, substring, replacer) => {
        let index = string.indexOf(substring);
        if (index === -1) {
            return string;
        }
        const substringLength = substring.length;
        let endIndex = 0;
        let returnValue = '';
        do {
            returnValue += string.substr(endIndex, index - endIndex) + substring + replacer;
            endIndex = index + substringLength;
            index = string.indexOf(substring, endIndex);
        } while (index !== -1);
        returnValue += string.substr(endIndex);
        return returnValue;
    };
    const stringEncaseCRLFWithFirstIndex = (string, prefix, postfix, index) => {
        let endIndex = 0;
        let returnValue = '';
        do {
            const gotCR = string[index - 1] === '\r';
            returnValue += string.substr(endIndex, (gotCR ? index - 1 : index) - endIndex) + prefix + (gotCR ? '\r\n' : '\n') + postfix;
            endIndex = index + 1;
            index = string.indexOf('\n', endIndex);
        } while (index !== -1);
        returnValue += string.substr(endIndex);
        return returnValue;
    };
    module.exports = {
        stringReplaceAll,
        stringEncaseCRLFWithFirstIndex
    };
},
/* chalk/source/templates.js */ function _(require, module, exports, __esModule, __esExport) {
    const TEMPLATE_REGEX = /(?:\\(u(?:[a-f\d]{4}|\{[a-f\d]{1,6}\})|x[a-f\d]{2}|.))|(?:\{(~)?(\w+(?:\([^)]*\))?(?:\.\w+(?:\([^)]*\))?)*)(?:[ \t]|(?=\r?\n)))|(\})|((?:.|[\r\n\f])+?)/gi;
    const STYLE_REGEX = /(?:^|\.)(\w+)(?:\(([^)]*)\))?/g;
    const STRING_REGEX = /^(['"])((?:\\.|(?!\1)[^\\])*)\1$/;
    const ESCAPE_REGEX = /\\(u(?:[a-f\d]{4}|{[a-f\d]{1,6}})|x[a-f\d]{2}|.)|([^\\])/gi;
    const ESCAPES = new Map([
        ['n', '\n'],
        ['r', '\r'],
        ['t', '\t'],
        ['b', '\b'],
        ['f', '\f'],
        ['v', '\v'],
        ['0', '\0'],
        ['\\', '\\'],
        ['e', '\u001B'],
        ['a', '\u0007']
    ]);
    function unescape(c) {
        const u = c[0] === 'u';
        const bracket = c[1] === '{';
        if ((u && !bracket && c.length === 5) || (c[0] === 'x' && c.length === 3)) {
            return String.fromCharCode(parseInt(c.slice(1), 16));
        }
        if (u && bracket) {
            return String.fromCodePoint(parseInt(c.slice(2, -1), 16));
        }
        return ESCAPES.get(c) || c;
    }
    function parseArguments(name, arguments_) {
        const results = [];
        const chunks = arguments_.trim().split(/\s*,\s*/g);
        let matches;
        for (const chunk of chunks) {
            const number = Number(chunk);
            if (!Number.isNaN(number)) {
                results.push(number);
            }
            else if ((matches = chunk.match(STRING_REGEX))) {
                results.push(matches[2].replace(ESCAPE_REGEX, (m, escape, character) => escape ? unescape(escape) : character));
            }
            else {
                throw new Error(`Invalid Chalk template style argument: ${chunk} (in style '${name}')`);
            }
        }
        return results;
    }
    function parseStyle(style) {
        STYLE_REGEX.lastIndex = 0;
        const results = [];
        let matches;
        while ((matches = STYLE_REGEX.exec(style)) !== null) {
            const name = matches[1];
            if (matches[2]) {
                const args = parseArguments(name, matches[2]);
                results.push([name].concat(args));
            }
            else {
                results.push([name]);
            }
        }
        return results;
    }
    function buildStyle(chalk, styles) {
        const enabled = {};
        for (const layer of styles) {
            for (const style of layer.styles) {
                enabled[style[0]] = layer.inverse ? null : style.slice(1);
            }
        }
        let current = chalk;
        for (const [styleName, styles] of Object.entries(enabled)) {
            if (!Array.isArray(styles)) {
                continue;
            }
            if (!(styleName in current)) {
                throw new Error(`Unknown Chalk style: ${styleName}`);
            }
            current = styles.length > 0 ? current[styleName](...styles) : current[styleName];
        }
        return current;
    }
    module.exports = (chalk, temporary) => {
        const styles = [];
        const chunks = [];
        let chunk = [];
        // eslint-disable-next-line max-params
        temporary.replace(TEMPLATE_REGEX, (m, escapeCharacter, inverse, style, close, character) => {
            if (escapeCharacter) {
                chunk.push(unescape(escapeCharacter));
            }
            else if (style) {
                const string = chunk.join('');
                chunk = [];
                chunks.push(styles.length === 0 ? string : buildStyle(chalk, styles)(string));
                styles.push({ inverse, styles: parseStyle(style) });
            }
            else if (close) {
                if (styles.length === 0) {
                    throw new Error('Found extraneous } in Chalk template literal');
                }
                chunks.push(buildStyle(chalk, styles)(chunk.join('')));
                chunk = [];
                styles.pop();
            }
            else {
                chunk.push(character);
            }
        });
        chunks.push(chunk.join(''));
        if (styles.length > 0) {
            const errMessage = `Chalk template literal is missing ${styles.length} closing bracket${styles.length === 1 ? '' : 's'} (\`}\`)`;
            throw new Error(errMessage);
        }
        return chunks.join('');
    };
},
/* transforms.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    function apply(node, ...transforms) {
        const result = ts.transform(node, transforms);
        return result.transformed[0];
    }
    exports.apply = apply;
    function is_require(node) {
        return ts.isCallExpression(node) &&
            ts.isIdentifier(node.expression) &&
            node.expression.text === "require" &&
            node.arguments.length === 1;
    }
    function relativize_modules(relativize) {
        function relativize_specifier(source, expr) {
            if (expr != null && ts.isStringLiteralLike(expr) && expr.text.length > 0) {
                const relative = relativize(source.fileName, expr.text);
                if (relative != null)
                    return ts.createLiteral(relative);
            }
            return null;
        }
        return (context) => {
            return {
                transformSourceFile(root) {
                    const { factory } = context;
                    function visit(node) {
                        if (ts.isImportDeclaration(node)) {
                            const moduleSpecifier = relativize_specifier(root, node.moduleSpecifier);
                            if (moduleSpecifier != null) {
                                const { decorators, modifiers, importClause } = node;
                                return factory.updateImportDeclaration(node, decorators, modifiers, importClause, moduleSpecifier);
                            }
                        }
                        if (ts.isExportDeclaration(node)) {
                            const moduleSpecifier = relativize_specifier(root, node.moduleSpecifier);
                            if (moduleSpecifier != null) {
                                const { decorators, modifiers, isTypeOnly, exportClause } = node;
                                return factory.updateExportDeclaration(node, decorators, modifiers, isTypeOnly, exportClause, moduleSpecifier);
                            }
                        }
                        if (is_require(node)) {
                            const moduleSpecifier = relativize_specifier(root, node.arguments[0]);
                            if (moduleSpecifier != null) {
                                const { expression, typeArguments } = node;
                                return factory.updateCallExpression(node, expression, typeArguments, [moduleSpecifier]);
                            }
                        }
                        return ts.visitEachChild(node, visit, context);
                    }
                    return ts.visitNode(root, visit);
                },
                transformBundle(_root) {
                    throw new Error("unsupported");
                },
            };
        };
    }
    exports.relativize_modules = relativize_modules;
    function is_static(node) {
        return node.modifiers != null && node.modifiers.find((modifier) => modifier.kind == ts.SyntaxKind.StaticKeyword) != null;
    }
    function add_init_class() {
        return (context) => (root) => {
            const { factory } = context;
            function visit(node) {
                node = ts.visitEachChild(node, visit, context);
                if (ts.isClassDeclaration(node) && node.name != null) {
                    const name = `init_${node.name.getText()}`;
                    if (node.members.find((member) => ts.isMethodDeclaration(member) && member.name.getText() == name && is_static(member)) != null) {
                        const init = factory.createExpressionStatement(factory.createCallExpression(factory.createPropertyAccessExpression(node.name, name), undefined, undefined));
                        return [node, init];
                    }
                }
                return node;
            }
            return ts.visitNode(root, visit);
        };
    }
    exports.add_init_class = add_init_class;
    function insert_class_name() {
        function has__name__(node) {
            return node.members.find((member) => ts.isPropertyDeclaration(member) && member.name.getText() == "__name__" && is_static(member)) != null;
        }
        return (context) => (root) => {
            const { factory } = context;
            function visit(node) {
                node = ts.visitEachChild(node, visit, context);
                if (ts.isClassDeclaration(node) && node.name != null && !has__name__(node)) {
                    const property = factory.createPropertyDeclaration(undefined, factory.createModifiersFromModifierFlags(ts.ModifierFlags.Static), "__name__", undefined, undefined, factory.createStringLiteral(node.name.text));
                    node = factory.updateClassDeclaration(node, node.decorators, node.modifiers, node.name, node.typeParameters, node.heritageClauses, [property, ...node.members]);
                }
                return node;
            }
            return ts.visitNode(root, visit);
        };
    }
    exports.insert_class_name = insert_class_name;
    function remove_use_strict() {
        return (_context) => (root) => {
            const statements = root.statements.filter((node) => {
                if (ts.isExpressionStatement(node)) {
                    const expr = node.expression;
                    if (ts.isStringLiteral(expr) && expr.text == "use strict")
                        return false;
                }
                return true;
            });
            return ts.updateSourceFileNode(root, statements);
        };
    }
    exports.remove_use_strict = remove_use_strict;
    function collect_exports(exported) {
        return (_context) => (root) => {
            for (const statement of root.statements) {
                if (ts.isExportDeclaration(statement)) {
                    if (statement.isTypeOnly)
                        continue;
                    const { exportClause, moduleSpecifier } = statement;
                    if (moduleSpecifier == null || !ts.isStringLiteral(moduleSpecifier))
                        continue;
                    const module = moduleSpecifier.text;
                    if (exportClause == null) {
                        // export * from "module"
                        exported.push({ type: "namespace", module });
                    }
                    else if (ts.isNamespaceExport(exportClause)) {
                        // export * as name from "module"
                        const name = exportClause.name.text;
                        exported.push({ type: "namespace", name, module });
                    }
                    else if (ts.isNamedExports(exportClause)) {
                        // export {name0, name1 as nameA} from "module"
                        const bindings = [];
                        for (const elem of exportClause.elements) {
                            bindings.push([elem.propertyName?.text, elem.name.text]);
                        }
                        exported.push({ type: "bindings", bindings, module });
                    }
                }
                else if (ts.isExportAssignment(statement) && !statement.isExportEquals) {
                    // export default name
                    exported.push({ type: "named", name: "default" });
                }
                else if (ts.isClassDeclaration(statement) || ts.isFunctionDeclaration(statement)) {
                    const flags = ts.getCombinedModifierFlags(statement);
                    if (flags & ts.ModifierFlags.Export) {
                        // export class X {}
                        // export function f() {}
                        if (statement.name != null) {
                            const name = statement.name.text;
                            exported.push({ type: "named", name });
                        }
                    }
                    else if (flags & ts.ModifierFlags.ExportDefault) {
                        // export default class X {}
                        // export function f() {}
                        exported.push({ type: "named", name: "default" });
                    }
                }
            }
            return root;
        };
    }
    exports.collect_exports = collect_exports;
    function isImportCall(node) {
        return ts.isCallExpression(node) && node.expression.kind == ts.SyntaxKind.ImportKeyword;
    }
    function collect_imports(imports) {
        return (context) => (root) => {
            function visit(node) {
                if (ts.isImportDeclaration(node) || ts.isExportDeclaration(node)) {
                    const name = node.moduleSpecifier;
                    if (name != null && ts.isStringLiteral(name) && name.text.length != 0)
                        imports.add(name.text);
                }
                else if (isImportCall(node)) {
                    const [name] = node.arguments;
                    if (ts.isStringLiteral(name) && name.text.length != 0)
                        imports.add(name.text);
                }
                return ts.visitEachChild(node, visit, context);
            }
            return ts.visitNode(root, visit);
        };
    }
    exports.collect_imports = collect_imports;
    function collect_deps(source) {
        function traverse(node) {
            if (is_require(node)) {
                const [arg] = node.arguments;
                if (ts.isStringLiteral(arg) && arg.text.length > 0)
                    deps.add(arg.text);
            }
            ts.forEachChild(node, traverse);
        }
        const deps = new Set();
        traverse(source);
        return [...deps];
    }
    exports.collect_deps = collect_deps;
    function rewrite_deps(resolve) {
        return (context) => (root) => {
            const { factory } = context;
            function visit(node) {
                if (is_require(node)) {
                    const [arg] = node.arguments;
                    if (ts.isStringLiteral(arg) && arg.text.length > 0) {
                        const dep = arg.text;
                        const val = resolve(dep);
                        if (val != null) {
                            const literal = typeof val == "string" ? factory.createStringLiteral(val) : factory.createNumericLiteral(val);
                            node = factory.updateCallExpression(node, node.expression, node.typeArguments, [literal]);
                            ts.addSyntheticTrailingComment(node, ts.SyntaxKind.MultiLineCommentTrivia, ` ${dep} `, false);
                        }
                        return node;
                    }
                }
                return ts.visitEachChild(node, visit, context);
            }
            return ts.visitNode(root, visit);
        };
    }
    exports.rewrite_deps = rewrite_deps;
    // XXX: this is pretty naive, but affects very litte code
    function rename_exports() {
        return (context) => (root) => {
            const { factory } = context;
            function is_exports(node) {
                return ts.isIdentifier(node) && node.text == "exports";
            }
            const has_exports = root.statements.some((stmt) => {
                return ts.isVariableStatement(stmt) && stmt.declarationList.declarations.some((decl) => is_exports(decl.name));
            });
            if (has_exports) {
                function visit(node) {
                    if (is_exports(node)) {
                        const updated = factory.createIdentifier("exports$1");
                        const original = node;
                        ts.setOriginalNode(updated, original);
                        ts.setTextRange(updated, original);
                        return updated;
                    }
                    return ts.visitEachChild(node, visit, context);
                }
                return ts.visitNode(root, visit);
            }
            else
                return root;
        };
    }
    exports.rename_exports = rename_exports;
    /**
     * Transform `var _this = _super.call(this, seq) || this;` into `var _this = new _super(seq);`.
     */
    function es5_fix_extend_builtins() {
        return (context) => (root) => {
            const { factory } = context;
            function visit(node) {
                if (ts.isFunctionDeclaration(node)) {
                    if (node.name != null && node.name.text.endsWith("NDArray") && node.body != null) {
                        const [stmt, ...rest] = node.body.statements;
                        if (ts.isVariableStatement(stmt) && stmt.declarationList.declarations.length == 1) {
                            const [decl] = stmt.declarationList.declarations;
                            if (ts.isIdentifier(decl.name) && decl.name.text == "_this" && decl.initializer != null) {
                                const init = factory.createNewExpression(factory.createIdentifier("_super"), undefined, [factory.createIdentifier("seq")]);
                                const decl_new = factory.updateVariableDeclaration(decl, decl.name, decl.exclamationToken, decl.type, init);
                                const decls_new = factory.updateVariableDeclarationList(stmt.declarationList, [decl_new]);
                                const stmt_new = factory.updateVariableStatement(stmt, stmt.modifiers, decls_new);
                                const body = factory.createBlock([stmt_new, ...rest], true);
                                const constructor = factory.updateFunctionDeclaration(node, node.decorators, node.modifiers, node.asteriskToken, node.name, node.typeParameters, node.parameters, node.type, body);
                                return constructor;
                            }
                        }
                    }
                }
                return ts.visitEachChild(node, visit, context);
            }
            return ts.visitNode(root, visit);
        };
    }
    exports.es5_fix_extend_builtins = es5_fix_extend_builtins;
    function fix_esmodule() {
        return (context) => (root) => {
            const { factory } = context;
            let found = false;
            const statements = root.statements.map((node) => {
                if (!found && ts.isExpressionStatement(node)) {
                    const expr = node.expression;
                    if (ts.isCallExpression(expr) && expr.arguments.length == 3) {
                        const [, arg] = expr.arguments;
                        if (ts.isStringLiteral(arg) && arg.text == "__esModule") {
                            found = true;
                            const es_module = factory.createIdentifier("__esModule");
                            const call = factory.createCallExpression(es_module, [], []);
                            return factory.createExpressionStatement(call);
                        }
                    }
                }
                return node;
            });
            return ts.updateSourceFileNode(root, statements);
        };
    }
    exports.fix_esmodule = fix_esmodule;
    function remove_void0() {
        return (_context) => (root) => {
            let found = false;
            const statements = root.statements.filter((node) => {
                if (!found && ts.isExpressionStatement(node)) {
                    let { expression } = node;
                    while (ts.isBinaryExpression(expression) &&
                        ts.isPropertyAccessExpression(expression.left) &&
                        ts.isIdentifier(expression.left.expression) &&
                        expression.left.expression.text == "exports") {
                        expression = expression.right;
                    }
                    if (ts.isVoidExpression(expression)) {
                        found = true;
                        return false;
                    }
                }
                return true;
            });
            return ts.updateSourceFileNode(root, statements);
        };
    }
    exports.remove_void0 = remove_void0;
    function fix_esexports() {
        return (context) => (root) => {
            const { factory } = context;
            const statements = root.statements.map((node) => {
                if (ts.isExpressionStatement(node)) {
                    const expr = node.expression;
                    if (ts.isCallExpression(expr) && ts.isPropertyAccessExpression(expr.expression) && expr.arguments.length == 3) {
                        const { expression, name } = expr.expression;
                        if (ts.isIdentifier(expression) && expression.text == "Object" &&
                            ts.isIdentifier(name) && name.text == "defineProperty") {
                            const [exports, name, config] = expr.arguments;
                            if (ts.isIdentifier(exports) && exports.text == "exports" &&
                                ts.isStringLiteral(name) &&
                                ts.isObjectLiteralExpression(config)) {
                                for (const item of config.properties) {
                                    if (ts.isPropertyAssignment(item) &&
                                        ts.isIdentifier(item.name) && item.name.text == "get" &&
                                        ts.isFunctionExpression(item.initializer)) {
                                        const { statements } = item.initializer.body;
                                        if (statements.length == 1) {
                                            const [stmt] = statements;
                                            if (ts.isReturnStatement(stmt) && stmt.expression != null) {
                                                const es_export = factory.createIdentifier("__esExport");
                                                const call = factory.createCallExpression(es_export, [], [name, stmt.expression]);
                                                return factory.createExpressionStatement(call);
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                return node;
            });
            return ts.updateSourceFileNode(root, statements);
        };
    }
    exports.fix_esexports = fix_esexports;
    function wrap_in_function(module_name) {
        return (context) => (root) => {
            const { factory } = context;
            const p = (name) => factory.createParameterDeclaration(undefined, undefined, undefined, name);
            const params = [p("require"), p("module"), p("exports"), p("__esModule"), p("__esExport")];
            const block = factory.createBlock(root.statements, true);
            const func = factory.createFunctionDeclaration(undefined, undefined, undefined, "_", undefined, params, undefined, block);
            ts.addSyntheticLeadingComment(func, ts.SyntaxKind.MultiLineCommentTrivia, ` ${module_name} `, false);
            return ts.updateSourceFileNode(root, [func]);
        };
    }
    exports.wrap_in_function = wrap_in_function;
    function parse_es(file, code, target = ts.ScriptTarget.ES2017) {
        return ts.createSourceFile(file, code != null ? code : ts.sys.readFile(file), target, true, ts.ScriptKind.JS);
    }
    exports.parse_es = parse_es;
    function print_es(source) {
        const printer = ts.createPrinter();
        return printer.printNode(ts.EmitHint.SourceFile, source, source);
    }
    exports.print_es = print_es;
},
/* error.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    class BuildError extends Error {
        constructor(component, message) {
            super(message);
            this.component = component;
        }
    }
    exports.BuildError = BuildError;
    BuildError.__name__ = "BuildError";
},
/* linker.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const path_1 = require("path");
    const module_1 = __importDefault(require("module"));
    const crypto_1 = __importDefault(require("crypto"));
    const ts = __importStar(require(21) /* typescript */);
    const terser = __importStar(require(45) /* terser */);
    const chalk_1 = __importDefault(require(37) /* chalk */);
    const combine = __importStar(require(47) /* combine-source-map */);
    const convert = __importStar(require(48) /* convert-source-map */);
    const sys_1 = require(20) /* ./sys */;
    const compiler_1 = require(36) /* ./compiler */;
    const transforms = __importStar(require(42) /* ./transforms */);
    const error_1 = require(43) /* ./error */;
    const graph_1 = require(54) /* ./graph */;
    const root_path = process.cwd();
    const cache_version = 4;
    const to_obj = (map) => {
        const obj = Object.create(null);
        for (const [key, val] of map) {
            obj[key] = val;
        }
        return obj;
    };
    const dense_assembly = {
        prefix: "[",
        suffix: "]",
        wrap: (_, source) => source,
    };
    const sparse_assembly = {
        prefix: "{",
        suffix: "}",
        wrap: (id, source) => `${id}: ${source}`,
    };
    class Bundle {
        constructor(entry, artifacts, builtins, bases = []) {
            this.entry = entry;
            this.artifacts = artifacts;
            this.builtins = builtins;
            this.bases = bases;
        }
        assemble(options) {
            let line = 0;
            let sources = "";
            const sourcemap = combine.create();
            const aliases = new Map();
            const externals = new Map();
            const newlines = (source) => {
                const result = source.match(/\n/g);
                return result != null ? result.length : 0;
            };
            const safe_id = (module) => {
                const { id } = module;
                return typeof id == "number" ? id.toString() : JSON.stringify(id);
            };
            const { entry, artifacts, bases } = this;
            const minified = options.minified ?? false;
            const plugin = bases.length != 0;
            const { prefix, suffix, wrap } = !plugin ? dense_assembly : sparse_assembly;
            const prelude = !plugin ? options.prelude.main : options.prelude.plugin;
            const postlude = !plugin ? options.postlude.main : options.postlude.plugin;
            sources += `${prelude}(${prefix}\n`;
            line += newlines(sources);
            for (const artifact of artifacts) {
                const { module } = artifact;
                if (module.canonical != null)
                    aliases.set(module.canonical, module.id);
                for (const external of module.externals)
                    externals.set(external, true);
                for (const external of module.shims)
                    externals.set(external, false);
                const start = wrap(safe_id(module), "");
                sources += start;
                line += newlines(start);
                const source_with_sourcemap = minified ? artifact.code.min_source : artifact.code.source;
                const source = combine.removeComments(source_with_sourcemap).trimRight();
                sources += source;
                const map_path = (0, path_1.join)("@@", (0, path_1.relative)(root_path, module.file));
                sourcemap.addFile({ source: source_with_sourcemap, sourceFile: map_path }, { line });
                line += newlines(source);
                const end = ",\n";
                sources += end;
                line += newlines(end);
            }
            const aliases_json = JSON.stringify(to_obj(aliases));
            const externals_json = JSON.stringify(to_obj(externals));
            sources += `${suffix}, ${safe_id(entry)}, ${aliases_json}, ${externals_json});`;
            if (postlude != null)
                sources += postlude;
            const source_map = convert.fromBase64(sourcemap.base64()).toObject();
            return new Artifact(sources, minified ? null : source_map, aliases);
        }
        *collect_exported(entry) {
            for (const item of entry.exported) {
                switch (item.type) {
                    case "bindings": {
                        for (const [, name] of item.bindings) {
                            yield name;
                        }
                        break;
                    }
                    case "named": {
                        yield item.name;
                        break;
                    }
                    case "namespace": {
                        const { name } = item;
                        if (name != null)
                            yield name;
                        else {
                            const module = entry.dependencies.get(item.module);
                            if (module != null) {
                                yield* this.collect_exported(module);
                            }
                        }
                        break;
                    }
                }
            }
        }
    }
    exports.Bundle = Bundle;
    Bundle.__name__ = "Bundle";
    class Artifact {
        constructor(source, sourcemap, exported) {
            this.source = source;
            this.sourcemap = sourcemap;
            this.exported = exported;
        }
        full_source(name) {
            if (this.sourcemap != null)
                return `${this.source}\n${convert.generateMapFileComment(name)}\n`;
            else
                return `${this.source}\n`;
        }
        get module_names() {
            return Array.from(this.exported.keys());
        }
        write(path) {
            const dir = (0, path_1.dirname)(path);
            const name = (0, path_1.basename)(path, ".js");
            (0, sys_1.write)(path, this.full_source(`${name}.js.map`));
            if (this.sourcemap != null) {
                (0, sys_1.write)((0, path_1.join)(dir, `${name}.js.map`), JSON.stringify(this.sourcemap));
            }
        }
    }
    exports.Artifact = Artifact;
    Artifact.__name__ = "Artifact";
    class Linker {
        constructor(opts) {
            this.ext = ".js";
            this.entries = opts.entries.map((path) => (0, path_1.resolve)(path));
            this.bases = (opts.bases ?? []).map((path) => (0, path_1.resolve)(path));
            this.excludes = new Set((opts.excludes ?? []).map((path) => (0, path_1.resolve)(path)));
            this.external_modules = new Set((opts.externals ?? []).filter((s) => typeof s === "string"));
            this.external_regex = (opts.externals ?? []).filter((s) => s instanceof RegExp);
            this.excluded = opts.excluded ?? (() => false);
            this.builtins = opts.builtins ?? false;
            this.exports = new Set(opts.exports ?? []);
            if (this.builtins) {
                this.external_modules.add("module");
                this.external_modules.add("constants");
                for (const lib of module_1.default.builtinModules)
                    this.external_modules.add(lib);
            }
            for (const entry of this.entries) {
                if (!(0, sys_1.file_exists)(entry))
                    throw new error_1.BuildError("linker", `entry path ${entry} doesn't exist or isn't a file`);
            }
            for (const base of this.bases) {
                if (!(0, sys_1.directory_exists)(base))
                    throw new error_1.BuildError("linker", `base path ${base} doesn't exist or isn't a directory`);
            }
            this.cache_path = opts.cache;
            this.cache = new Map();
            this.target = opts.target ?? null;
            this.es_modules = opts.es_modules ?? true;
            this.minify = opts.minify ?? true;
            this.plugin = opts.plugin ?? false;
            this.shims = new Set(opts.shims ?? []);
            this.detect_cycles = opts.detect_cycles ?? true;
        }
        is_external(dep) {
            return this.external_modules.has(dep) || this.external_regex.some((re) => re.test(dep));
        }
        is_shimmed(dep) {
            return this.shims.has(dep);
        }
        async link() {
            let status = true;
            const [entries] = this.resolve(this.entries);
            const [main, ...plugins] = entries;
            const dirnames = plugins.map((plugin) => (0, path_1.dirname)(plugin.file));
            const is_excluded = (module) => {
                return dirnames.find((e) => module.file.startsWith(e)) != null;
            };
            const main_modules = this.reachable(main, is_excluded);
            const parents = new Set(main_modules);
            const plugin_modules = [];
            for (const plugin of plugins) {
                const files = this.reachable(plugin, (module) => parents.has(module));
                plugin_modules.push(files);
            }
            const all_modules = main_modules.concat(...plugin_modules);
            if (!this.plugin)
                all_modules.forEach((module, i) => module.id = i);
            else
                all_modules.forEach((module) => module.id = module.hash.slice(0, 10));
            if (this.detect_cycles) {
                function is_internal(module) {
                    return module.canonical != null;
                }
                const nodes = all_modules.filter(is_internal);
                const graph = new Map();
                for (const node of nodes) {
                    const deps = [...node.dependencies.values()];
                    graph.set(node, deps.filter(is_internal));
                }
                const cycles = (0, graph_1.detect_cycles)(graph);
                if (cycles.length != 0) {
                    status = false;
                    for (const cycle of cycles) {
                        console.log(`${chalk_1.default.red("\u2717")} cycle detected:`);
                        for (const module of cycle) {
                            const is_last = cycle[cycle.length - 1] == module;
                            console.log(`${is_last ? "\u2514" : "\u251c"}\u2500 ${module.canonical ?? module.base_path}`);
                        }
                    }
                }
            }
            const transformers = (module) => {
                const transformers = [];
                const remove_use_strict = transforms.remove_use_strict();
                transformers.push(remove_use_strict);
                const fix_esmodule = transforms.fix_esmodule();
                transformers.push(fix_esmodule);
                const remove_void0 = transforms.remove_void0();
                transformers.push(remove_void0);
                const fix_esexports = transforms.fix_esexports();
                transformers.push(fix_esexports);
                const rewrite_deps = transforms.rewrite_deps((dep) => {
                    const module_dep = module.dependencies.get(dep);
                    return module_dep != null ? module_dep.id : undefined;
                });
                transformers.push(rewrite_deps);
                transformers.push(transforms.wrap_in_function(module.base_path));
                return transformers;
            };
            const print = (module) => {
                let ast = module.ast ?? this.parse_module(module);
                ast = transforms.apply(ast, ...transformers(module));
                const source = transforms.print_es(ast);
                return convert.removeMapFileComments(source);
            };
            const deps_changed = (module, cached) => {
                if (module.dependencies.size != cached.dependencies.size)
                    return false;
                for (const [dep, module_dep] of module.dependencies) {
                    const cached_dep = cached.dependencies.get(dep);
                    if (cached_dep == null || cached_dep.id != module_dep.id)
                        return true;
                }
                return false;
            };
            const artifacts = async (modules) => {
                const result = [];
                for (const module of modules) {
                    const cached = this.cache.get(module.file);
                    let code;
                    if (module.changed || (cached != null && deps_changed(module, cached.module))) {
                        const source = print(module);
                        const ecma = this.target == "ES2020" ? 2020 : (this.target == "ES2017" ? 2017 : 5);
                        const minified = this.minify ? await minify(module, source, ecma) : { min_source: source };
                        code = { source, ...minified };
                    }
                    else
                        code = cached.code;
                    result.push({ module, code });
                }
                return result;
            };
            const main_bundle = new Bundle(main, await artifacts(main_modules), this.builtins);
            const plugin_bundles = [];
            for (let j = 0; j < plugins.length; j++) {
                const plugin_bundle = new Bundle(plugins[j], await artifacts(plugin_modules[j]), this.builtins, [main_bundle]);
                plugin_bundles.push(plugin_bundle);
            }
            const bundles = [main_bundle, ...plugin_bundles];
            this.cache.clear();
            for (const bundle of bundles) {
                for (const artifact of bundle.artifacts) {
                    this.cache.set(artifact.module.file, artifact);
                }
            }
            return { bundles, status };
        }
        load_cache() {
            const { cache_path } = this;
            if (cache_path == null || !(0, sys_1.file_exists)(cache_path))
                return;
            this.cache.clear();
            const json = JSON.parse((0, sys_1.read)(cache_path));
            if (json.version !== cache_version) {
                console.warn("ignoring cache due to format version mismatch");
                return;
            }
            for (const { module, code } of json.artifacts) {
                const artifact = {
                    module: {
                        ...module,
                        ast: undefined,
                        changed: false,
                        dependencies: new Map(),
                        dependency_map: new Map(module.dependency_map),
                        dependency_paths: new Map(module.dependency_paths),
                        externals: new Set(module.externals),
                        shims: new Set(module.shims),
                    },
                    code,
                };
                this.cache.set(artifact.module.file, artifact);
            }
            for (const { module } of this.cache.values()) {
                for (const [dep, file] of module.dependency_paths) {
                    const file_entry = this.cache.get(file);
                    if (file_entry == null)
                        throw new Error(`${file} not in cache`);
                    module.dependencies.set(dep, file_entry.module);
                }
            }
        }
        store_cache() {
            if (this.cache_path == null)
                return;
            const artifacts = [];
            for (const artifact of this.cache.values()) {
                const module = {
                    ...artifact.module,
                    changed: undefined,
                    ast: undefined,
                    dependencies: undefined,
                };
                artifacts.push({
                    module: {
                        ...module,
                        dependency_map: [...module.dependency_map.entries()],
                        dependency_paths: [...module.dependency_paths.entries()],
                        externals: [...module.externals.values()],
                        shims: [...module.shims.values()],
                    },
                    code: artifact.code,
                });
            }
            const json = JSON.stringify({ version: cache_version, artifacts });
            (0, sys_1.write)(this.cache_path, json);
        }
        resolve_package(dir) {
            const index = (() => {
                const pkg_path = (0, path_1.join)(dir, "package.json");
                if ((0, sys_1.file_exists)(pkg_path)) {
                    const pkg = JSON.parse((0, sys_1.read)(pkg_path));
                    if (this.target != null && pkg.module != null)
                        return pkg.module;
                    if (pkg.main != null)
                        return pkg.main;
                }
                return "index.js";
            })();
            const path = (0, path_1.join)(dir, index);
            if ((0, sys_1.file_exists)(path))
                return path;
            const js_file = `${path}.js`;
            if ((0, sys_1.file_exists)(js_file))
                return js_file;
            const json_file = `${path}.json`;
            if ((0, sys_1.file_exists)(json_file))
                return json_file;
            if ((0, sys_1.directory_exists)(path)) {
                const index = (0, path_1.join)(path, "index.js");
                if ((0, sys_1.file_exists)(index))
                    return index;
            }
            return null;
        }
        resolve_relative(dep, parent) {
            const path = (0, path_1.resolve)((0, path_1.dirname)(parent.file), dep);
            if ((0, sys_1.file_exists)(path))
                return path;
            const js_file = `${path}.js`;
            const json_file = `${path}.json`;
            const has_js_file = (0, sys_1.file_exists)(js_file);
            const has_json_file = (0, sys_1.file_exists)(json_file);
            const has_file = has_js_file ?? has_json_file;
            if ((0, sys_1.directory_exists)(path)) {
                const pkg_file = this.resolve_package(path);
                if (pkg_file != null) {
                    if (!has_file)
                        return pkg_file;
                    else
                        return new error_1.BuildError("linker", `both ${has_js_file ? js_file : json_file} and ${pkg_file} exist`);
                }
            }
            if (has_js_file)
                return js_file;
            else if (has_json_file)
                return json_file;
            else
                return new error_1.BuildError("linker", `can't resolve '${dep}' from '${parent.file}'`);
        }
        resolve_absolute(dep, parent) {
            for (const base of this.bases) {
                let path = (0, path_1.join)(base, dep);
                if ((0, sys_1.file_exists)(path))
                    return path;
                const js_file = `${path}.js`;
                if ((0, sys_1.file_exists)(js_file))
                    return js_file;
                const json_file = `${path}.json`;
                if ((0, sys_1.file_exists)(json_file))
                    return json_file;
                if ((0, sys_1.directory_exists)(path)) {
                    const file = this.resolve_package(path);
                    if (file != null)
                        return file;
                }
                if (parent.file.startsWith(base)) {
                    let base_path = parent.file;
                    while (true) {
                        base_path = (0, path_1.dirname)(base_path);
                        if (base_path == base)
                            break;
                        path = (0, path_1.join)(base_path, "node_modules", dep);
                        if ((0, sys_1.directory_exists)(path)) {
                            const file = this.resolve_package(path);
                            if (file != null)
                                return file;
                        }
                    }
                }
            }
            return new error_1.BuildError("linker", `can't resolve '${dep}' from '${parent.file}'`);
        }
        resolve_file(dep, parent) {
            if (dep.startsWith("."))
                return this.resolve_relative(dep, parent);
            else
                return this.resolve_absolute(dep, parent);
        }
        parse_module({ file, source }) {
            return transforms.parse_es(file, source);
        }
        new_module(file) {
            let source = (0, sys_1.read)(file);
            if (source == null) {
                throw new error_1.BuildError("linker", `'${file} doesn't exist`);
            }
            const hash = crypto_1.default.createHash("sha256").update(source).digest("hex");
            const type = (() => {
                switch ((0, path_1.extname)(file)) {
                    case ".json": return "json";
                    case ".css": return "css";
                    case ".mjs": return "js";
                    case ".cjs": return "js";
                    case ".js": return "js";
                    default:
                        throw new error_1.BuildError("linker", `unsupported extension of ${file}`);
                }
            })();
            const export_type = this.es_modules ? "default" : "=";
            switch (type) {
                case "json":
                    source = `\
const json = ${source};
export ${export_type} json;
`;
                    break;
                case "css":
                    source = `\
const css = \`${source}\`;
export ${export_type} css;
`;
                    break;
            }
            const [base, base_path, canonical, resolution] = (() => {
                const [primary, ...secondary] = this.bases;
                function canonicalize(path) {
                    return path.replace(/\.js$/, "").replace(/\\/g, "/");
                }
                function get_package(base, path) {
                    const root = (0, path_1.join)(base, path);
                    base = (0, path_1.normalize)(base);
                    path = (0, path_1.normalize)(root);
                    while (path != base) {
                        if ((0, sys_1.directory_exists)(path)) {
                            const pkg_path = (0, path_1.join)(path, "package.json");
                            if ((0, sys_1.file_exists)(pkg_path))
                                return { dir: path, pkg: JSON.parse((0, sys_1.read)(pkg_path)) };
                        }
                        path = (0, path_1.dirname)(path);
                    }
                    throw new error_1.BuildError("linker", `can't resolve package.json for ${root}`);
                }
                const path = (0, path_1.relative)(primary, file);
                if (!path.startsWith("..")) {
                    return [primary, path, canonicalize(path), "ESM"];
                }
                for (const base of secondary) {
                    const path = (0, path_1.relative)(base, file);
                    if (!path.startsWith("..")) {
                        if (type == "js") {
                            const { dir, pkg } = get_package(base, path);
                            const reso = pkg.type == "module" || pkg.module != null ? "ESM" : "CJS";
                            const entry = pkg.module ?? pkg.name;
                            const primary = entry != null && (0, path_1.join)(dir, entry) == (0, path_1.join)(base, path);
                            const name = canonicalize(primary ? (0, path_1.basename)(dir) : path);
                            const exported = this.exports.has(name);
                            return [base, path, exported ? name : undefined, reso];
                        }
                        else {
                            return [base, path, undefined, "ESM"];
                        }
                    }
                }
                throw new error_1.BuildError("linker", `${file} is not under any of base paths`);
            })();
            const cached = this.cache.get(file);
            let ast;
            let dependency_paths;
            let externals;
            let shims;
            let exported = [];
            const changed = cached == null || cached.module.hash != hash;
            if (changed) {
                let collected = null;
                if ((this.target != null && resolution == "ESM") || type == "json") {
                    const { ES2020, ES2017, ES5 } = ts.ScriptTarget;
                    const target = this.target == "ES2020" ? ES2020 : (this.target == "ES2017" ? ES2017 : ES5);
                    const imports = new Set();
                    if (canonical != "tslib") {
                        imports.add("tslib");
                    }
                    const transform = {
                        before: [transforms.collect_imports(imports), transforms.rename_exports(), transforms.collect_exports(exported)],
                        after: [],
                    };
                    if (canonical == "core/util/ndarray" && target == ES5) {
                        transform.after.push(transforms.es5_fix_extend_builtins());
                    }
                    // XXX: .json extension will cause an internal error
                    const { output, error } = transpile(type == "json" ? `${file}.ts` : file, source, target, transform);
                    if (error)
                        throw new error_1.BuildError("linker", error);
                    else {
                        source = output;
                        collected = [...imports];
                    }
                }
                ast = this.parse_module({ file, source });
                if (collected == null)
                    collected = transforms.collect_deps(ast);
                const filtered = collected.filter((dep) => !this.is_external(dep) && !this.excluded(dep) && !this.is_shimmed(dep));
                dependency_paths = new Map();
                for (const dep of filtered) {
                    const resolved = this.resolve_file(dep, { file });
                    if (resolved instanceof Error)
                        console.log(resolved);
                    else
                        dependency_paths.set(dep, resolved);
                }
                externals = new Set(collected.filter((dep) => this.is_external(dep)));
                shims = new Set(collected.filter((dep) => this.is_shimmed(dep)));
            }
            else {
                dependency_paths = cached.module.dependency_paths;
                exported = cached.module.exported;
                externals = cached.module.externals;
                shims = cached.module.shims;
                source = cached.module.source;
            }
            return {
                file,
                base,
                base_path,
                canonical,
                resolution,
                id: NaN,
                hash,
                changed,
                source,
                ast,
                type,
                dependency_paths,
                dependency_map: new Map(),
                dependencies: new Map(),
                exported,
                externals,
                shims,
            };
        }
        resolve(files) {
            const visited = new Map();
            const pending = [...files];
            for (;;) {
                const file = pending.shift();
                if (file == null)
                    break;
                if (visited.has(file) || this.excludes.has(file))
                    continue;
                const module = this.new_module(file);
                visited.set(module.file, module);
                pending.unshift(...module.dependency_paths.values());
            }
            for (const module of visited.values()) {
                for (const [dep, file] of module.dependency_paths) {
                    module.dependencies.set(dep, visited.get(file));
                }
            }
            const entries = files.map((file) => visited.get(file));
            return [entries, [...visited.values()]];
        }
        reachable(entry, is_excluded) {
            const reached = new Set();
            const pending = [entry];
            for (;;) {
                const module = pending.shift();
                if (module == null)
                    break;
                if (reached.has(module) || this.excludes.has(module.file) || is_excluded(module))
                    continue;
                reached.add(module);
                pending.unshift(...module.dependencies.values());
            }
            return [...reached.values()];
        }
    }
    exports.Linker = Linker;
    Linker.__name__ = "Linker";
    function transpile(file, source, target, transformers) {
        const { outputText: output, diagnostics } = ts.transpileModule(source, {
            fileName: file,
            reportDiagnostics: true,
            compilerOptions: {
                target,
                module: ts.ModuleKind.CommonJS,
                esModuleInterop: true,
                importHelpers: true,
                downlevelIteration: true,
            },
            transformers,
        });
        if (diagnostics == null || diagnostics.length == 0)
            return { output };
        else {
            const { text } = (0, compiler_1.report_diagnostics)(diagnostics);
            return { output, error: text };
        }
    }
    exports.transpile = transpile;
    async function minify(module, source, ecma) {
        const name = (0, path_1.basename)(module.file);
        const min_js = (0, sys_1.rename)(name, { ext: ".min.js" });
        const min_js_map = (0, sys_1.rename)(name, { ext: ".min.js.map" });
        const minify_opts = {
            ecma,
            format: {
                comments: /^!|copyright|license|\(c\)/i,
            },
            sourceMap: {
                filename: (0, path_1.basename)(min_js),
                url: (0, path_1.basename)(min_js_map),
            },
        };
        try {
            const { code, map } = await terser.minify(source, minify_opts);
            return {
                min_source: code ?? "",
                min_map: typeof map === "string" ? map : undefined,
            };
        }
        catch (error) {
            throw new error_1.BuildError("linker", `${error}`);
        }
    }
    exports.minify = minify;
},
/* terser/dist/bundle.min.js */ function _(require, module, exports, __esModule, __esExport) {
    (function (global, factory) {
        typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require(23) /* source-map */) :
            typeof define === 'function' && define.amd ? define(['exports', 'source-map'], factory) :
                (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.Terser = {}, global.sourceMap));
    }(this, (function (exports, MOZ_SourceMap) {
        'use strict';
        function _interopDefaultLegacy(e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }
        var MOZ_SourceMap__default = /*#__PURE__*/ _interopDefaultLegacy(MOZ_SourceMap);
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        function characters(str) {
            return str.split("");
        }
        function member(name, array) {
            return array.includes(name);
        }
        class DefaultsError extends Error {
            constructor(msg, defs) {
                super();
                this.name = "DefaultsError";
                this.message = msg;
                this.defs = defs;
            }
        }
        function defaults(args, defs, croak) {
            if (args === true) {
                args = {};
            }
            else if (args != null && typeof args === "object") {
                args = { ...args };
            }
            const ret = args || {};
            if (croak)
                for (const i in ret)
                    if (HOP(ret, i) && !HOP(defs, i)) {
                        throw new DefaultsError("`" + i + "` is not a supported option", defs);
                    }
            for (const i in defs)
                if (HOP(defs, i)) {
                    if (!args || !HOP(args, i)) {
                        ret[i] = defs[i];
                    }
                    else if (i === "ecma") {
                        let ecma = args[i] | 0;
                        if (ecma > 5 && ecma < 2015)
                            ecma += 2009;
                        ret[i] = ecma;
                    }
                    else {
                        ret[i] = (args && HOP(args, i)) ? args[i] : defs[i];
                    }
                }
            return ret;
        }
        function noop() { }
        function return_false() { return false; }
        function return_true() { return true; }
        function return_this() { return this; }
        function return_null() { return null; }
        var MAP = (function () {
            function MAP(a, f, backwards) {
                var ret = [], top = [], i;
                function doit() {
                    var val = f(a[i], i);
                    var is_last = val instanceof Last;
                    if (is_last)
                        val = val.v;
                    if (val instanceof AtTop) {
                        val = val.v;
                        if (val instanceof Splice) {
                            top.push.apply(top, backwards ? val.v.slice().reverse() : val.v);
                        }
                        else {
                            top.push(val);
                        }
                    }
                    else if (val !== skip) {
                        if (val instanceof Splice) {
                            ret.push.apply(ret, backwards ? val.v.slice().reverse() : val.v);
                        }
                        else {
                            ret.push(val);
                        }
                    }
                    return is_last;
                }
                if (Array.isArray(a)) {
                    if (backwards) {
                        for (i = a.length; --i >= 0;)
                            if (doit())
                                break;
                        ret.reverse();
                        top.reverse();
                    }
                    else {
                        for (i = 0; i < a.length; ++i)
                            if (doit())
                                break;
                    }
                }
                else {
                    for (i in a)
                        if (HOP(a, i))
                            if (doit())
                                break;
                }
                return top.concat(ret);
            }
            MAP.at_top = function (val) { return new AtTop(val); };
            MAP.splice = function (val) { return new Splice(val); };
            MAP.last = function (val) { return new Last(val); };
            var skip = MAP.skip = {};
            function AtTop(val) { this.v = val; }
            function Splice(val) { this.v = val; }
            function Last(val) { this.v = val; }
            return MAP;
        })();
        function make_node(ctor, orig, props) {
            if (!props)
                props = {};
            if (orig) {
                if (!props.start)
                    props.start = orig.start;
                if (!props.end)
                    props.end = orig.end;
            }
            return new ctor(props);
        }
        function push_uniq(array, el) {
            if (!array.includes(el))
                array.push(el);
        }
        function string_template(text, props) {
            return text.replace(/{(.+?)}/g, function (str, p) {
                return props && props[p];
            });
        }
        function remove(array, el) {
            for (var i = array.length; --i >= 0;) {
                if (array[i] === el)
                    array.splice(i, 1);
            }
        }
        function mergeSort(array, cmp) {
            if (array.length < 2)
                return array.slice();
            function merge(a, b) {
                var r = [], ai = 0, bi = 0, i = 0;
                while (ai < a.length && bi < b.length) {
                    cmp(a[ai], b[bi]) <= 0
                        ? r[i++] = a[ai++]
                        : r[i++] = b[bi++];
                }
                if (ai < a.length)
                    r.push.apply(r, a.slice(ai));
                if (bi < b.length)
                    r.push.apply(r, b.slice(bi));
                return r;
            }
            function _ms(a) {
                if (a.length <= 1)
                    return a;
                var m = Math.floor(a.length / 2), left = a.slice(0, m), right = a.slice(m);
                left = _ms(left);
                right = _ms(right);
                return merge(left, right);
            }
            return _ms(array);
        }
        function makePredicate(words) {
            if (!Array.isArray(words))
                words = words.split(" ");
            return new Set(words.sort());
        }
        function map_add(map, key, value) {
            if (map.has(key)) {
                map.get(key).push(value);
            }
            else {
                map.set(key, [value]);
            }
        }
        function map_from_object(obj) {
            var map = new Map();
            for (var key in obj) {
                if (HOP(obj, key) && key.charAt(0) === "$") {
                    map.set(key.substr(1), obj[key]);
                }
            }
            return map;
        }
        function map_to_object(map) {
            var obj = Object.create(null);
            map.forEach(function (value, key) {
                obj["$" + key] = value;
            });
            return obj;
        }
        function HOP(obj, prop) {
            return Object.prototype.hasOwnProperty.call(obj, prop);
        }
        function keep_name(keep_setting, name) {
            return keep_setting === true
                || (keep_setting instanceof RegExp && keep_setting.test(name));
        }
        var lineTerminatorEscape = {
            "\0": "0",
            "\n": "n",
            "\r": "r",
            "\u2028": "u2028",
            "\u2029": "u2029",
        };
        function regexp_source_fix(source) {
            // V8 does not escape line terminators in regexp patterns in node 12
            // We'll also remove literal \0
            return source.replace(/[\0\n\r\u2028\u2029]/g, function (match, offset) {
                var escaped = source[offset - 1] == "\\"
                    && (source[offset - 2] != "\\"
                        || /(?:^|[^\\])(?:\\{2})*$/.test(source.slice(0, offset - 1)));
                return (escaped ? "" : "\\") + lineTerminatorEscape[match];
            });
        }
        const all_flags = "gimuy";
        function sort_regexp_flags(flags) {
            const existing_flags = new Set(flags.split(""));
            let out = "";
            for (const flag of all_flags) {
                if (existing_flags.has(flag)) {
                    out += flag;
                    existing_flags.delete(flag);
                }
            }
            if (existing_flags.size) {
                // Flags Terser doesn't know about
                existing_flags.forEach(flag => { out += flag; });
            }
            return out;
        }
        function has_annotation(node, annotation) {
            return node._annotations & annotation;
        }
        function set_annotation(node, annotation) {
            node._annotations |= annotation;
        }
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
            Parser based on parse-js (http://marijn.haverbeke.nl/parse-js/).
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        var LATEST_RAW = ""; // Only used for numbers and template strings
        var LATEST_TEMPLATE_END = true;
        var KEYWORDS = "break case catch class const continue debugger default delete do else export extends finally for function if in instanceof let new return switch throw try typeof var void while with";
        var KEYWORDS_ATOM = "false null true";
        var RESERVED_WORDS = "enum implements import interface package private protected public static super this " + KEYWORDS_ATOM + " " + KEYWORDS;
        var KEYWORDS_BEFORE_EXPRESSION = "return new delete throw else case yield await";
        KEYWORDS = makePredicate(KEYWORDS);
        RESERVED_WORDS = makePredicate(RESERVED_WORDS);
        KEYWORDS_BEFORE_EXPRESSION = makePredicate(KEYWORDS_BEFORE_EXPRESSION);
        KEYWORDS_ATOM = makePredicate(KEYWORDS_ATOM);
        var OPERATOR_CHARS = makePredicate(characters("+-*&%=<>!?|~^"));
        var RE_NUM_LITERAL = /[0-9a-f]/i;
        var RE_HEX_NUMBER = /^0x[0-9a-f]+$/i;
        var RE_OCT_NUMBER = /^0[0-7]+$/;
        var RE_ES6_OCT_NUMBER = /^0o[0-7]+$/i;
        var RE_BIN_NUMBER = /^0b[01]+$/i;
        var RE_DEC_NUMBER = /^\d*\.?\d*(?:e[+-]?\d*(?:\d\.?|\.?\d)\d*)?$/i;
        var RE_BIG_INT = /^(0[xob])?[0-9a-f]+n$/i;
        var OPERATORS = makePredicate([
            "in",
            "instanceof",
            "typeof",
            "new",
            "void",
            "delete",
            "++",
            "--",
            "+",
            "-",
            "!",
            "~",
            "&",
            "|",
            "^",
            "*",
            "**",
            "/",
            "%",
            ">>",
            "<<",
            ">>>",
            "<",
            ">",
            "<=",
            ">=",
            "==",
            "===",
            "!=",
            "!==",
            "?",
            "=",
            "+=",
            "-=",
            "||=",
            "&&=",
            "??=",
            "/=",
            "*=",
            "**=",
            "%=",
            ">>=",
            "<<=",
            ">>>=",
            "|=",
            "^=",
            "&=",
            "&&",
            "??",
            "||",
        ]);
        var WHITESPACE_CHARS = makePredicate(characters(" \u00a0\n\r\t\f\u000b\u200b\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a\u2028\u2029\u202f\u205f\u3000\uFEFF"));
        var NEWLINE_CHARS = makePredicate(characters("\n\r\u2028\u2029"));
        var PUNC_AFTER_EXPRESSION = makePredicate(characters(";]),:"));
        var PUNC_BEFORE_EXPRESSION = makePredicate(characters("[{(,;:"));
        var PUNC_CHARS = makePredicate(characters("[]{}(),;:"));
        /* -----[ Tokenizer ]----- */
        // surrogate safe regexps adapted from https://github.com/mathiasbynens/unicode-8.0.0/tree/89b412d8a71ecca9ed593d9e9fa073ab64acfebe/Binary_Property
        var UNICODE = {
            ID_Start: /[$A-Z_a-z\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u08A0-\u08B4\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0AF9\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D\u0C58-\u0C5A\u0C60\u0C61\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D5F-\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191E\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2118-\u211D\u2124\u2126\u2128\u212A-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u3005-\u3007\u3021-\u3029\u3031-\u3035\u3038-\u303C\u3041-\u3096\u309B-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA69D\uA6A0-\uA6EF\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AD\uA7B0-\uA7B7\uA7F7-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA8FD\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uA9E0-\uA9E4\uA9E6-\uA9EF\uA9FA-\uA9FE\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA7E-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDD40-\uDD74\uDE80-\uDE9C\uDEA0-\uDED0\uDF00-\uDF1F\uDF30-\uDF4A\uDF50-\uDF75\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF\uDFD1-\uDFD5]|\uD801[\uDC00-\uDC9D\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00\uDE10-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE4\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC03-\uDC37\uDC83-\uDCAF\uDCD0-\uDCE8\uDD03-\uDD26\uDD50-\uDD72\uDD76\uDD83-\uDDB2\uDDC1-\uDDC4\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE2B\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEDE\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3D\uDF50\uDF5D-\uDF61]|\uD805[\uDC80-\uDCAF\uDCC4\uDCC5\uDCC7\uDD80-\uDDAE\uDDD8-\uDDDB\uDE00-\uDE2F\uDE44\uDE80-\uDEAA\uDF00-\uDF19]|\uD806[\uDCA0-\uDCDF\uDCFF\uDEC0-\uDEF8]|\uD808[\uDC00-\uDF99]|\uD809[\uDC00-\uDC6E\uDC80-\uDD43]|[\uD80C\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDED0-\uDEED\uDF00-\uDF2F\uDF40-\uDF43\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50\uDF93-\uDF9F]|\uD82C[\uDC00\uDC01]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB]|\uD83A[\uDC00-\uDCC4]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1]|\uD87E[\uDC00-\uDE1D]/,
            ID_Continue: /(?:[$0-9A-Z_a-z\xAA\xB5\xB7\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0300-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u0483-\u0487\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u0591-\u05BD\u05BF\u05C1\u05C2\u05C4\u05C5\u05C7\u05D0-\u05EA\u05F0-\u05F2\u0610-\u061A\u0620-\u0669\u066E-\u06D3\u06D5-\u06DC\u06DF-\u06E8\u06EA-\u06FC\u06FF\u0710-\u074A\u074D-\u07B1\u07C0-\u07F5\u07FA\u0800-\u082D\u0840-\u085B\u08A0-\u08B4\u08E3-\u0963\u0966-\u096F\u0971-\u0983\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BC-\u09C4\u09C7\u09C8\u09CB-\u09CE\u09D7\u09DC\u09DD\u09DF-\u09E3\u09E6-\u09F1\u0A01-\u0A03\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A3C\u0A3E-\u0A42\u0A47\u0A48\u0A4B-\u0A4D\u0A51\u0A59-\u0A5C\u0A5E\u0A66-\u0A75\u0A81-\u0A83\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABC-\u0AC5\u0AC7-\u0AC9\u0ACB-\u0ACD\u0AD0\u0AE0-\u0AE3\u0AE6-\u0AEF\u0AF9\u0B01-\u0B03\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3C-\u0B44\u0B47\u0B48\u0B4B-\u0B4D\u0B56\u0B57\u0B5C\u0B5D\u0B5F-\u0B63\u0B66-\u0B6F\u0B71\u0B82\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BBE-\u0BC2\u0BC6-\u0BC8\u0BCA-\u0BCD\u0BD0\u0BD7\u0BE6-\u0BEF\u0C00-\u0C03\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D-\u0C44\u0C46-\u0C48\u0C4A-\u0C4D\u0C55\u0C56\u0C58-\u0C5A\u0C60-\u0C63\u0C66-\u0C6F\u0C81-\u0C83\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBC-\u0CC4\u0CC6-\u0CC8\u0CCA-\u0CCD\u0CD5\u0CD6\u0CDE\u0CE0-\u0CE3\u0CE6-\u0CEF\u0CF1\u0CF2\u0D01-\u0D03\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D-\u0D44\u0D46-\u0D48\u0D4A-\u0D4E\u0D57\u0D5F-\u0D63\u0D66-\u0D6F\u0D7A-\u0D7F\u0D82\u0D83\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0DCA\u0DCF-\u0DD4\u0DD6\u0DD8-\u0DDF\u0DE6-\u0DEF\u0DF2\u0DF3\u0E01-\u0E3A\u0E40-\u0E4E\u0E50-\u0E59\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB9\u0EBB-\u0EBD\u0EC0-\u0EC4\u0EC6\u0EC8-\u0ECD\u0ED0-\u0ED9\u0EDC-\u0EDF\u0F00\u0F18\u0F19\u0F20-\u0F29\u0F35\u0F37\u0F39\u0F3E-\u0F47\u0F49-\u0F6C\u0F71-\u0F84\u0F86-\u0F97\u0F99-\u0FBC\u0FC6\u1000-\u1049\u1050-\u109D\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u135D-\u135F\u1369-\u1371\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1714\u1720-\u1734\u1740-\u1753\u1760-\u176C\u176E-\u1770\u1772\u1773\u1780-\u17D3\u17D7\u17DC\u17DD\u17E0-\u17E9\u180B-\u180D\u1810-\u1819\u1820-\u1877\u1880-\u18AA\u18B0-\u18F5\u1900-\u191E\u1920-\u192B\u1930-\u193B\u1946-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u19D0-\u19DA\u1A00-\u1A1B\u1A20-\u1A5E\u1A60-\u1A7C\u1A7F-\u1A89\u1A90-\u1A99\u1AA7\u1AB0-\u1ABD\u1B00-\u1B4B\u1B50-\u1B59\u1B6B-\u1B73\u1B80-\u1BF3\u1C00-\u1C37\u1C40-\u1C49\u1C4D-\u1C7D\u1CD0-\u1CD2\u1CD4-\u1CF6\u1CF8\u1CF9\u1D00-\u1DF5\u1DFC-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u200C\u200D\u203F\u2040\u2054\u2071\u207F\u2090-\u209C\u20D0-\u20DC\u20E1\u20E5-\u20F0\u2102\u2107\u210A-\u2113\u2115\u2118-\u211D\u2124\u2126\u2128\u212A-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D7F-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2DE0-\u2DFF\u3005-\u3007\u3021-\u302F\u3031-\u3035\u3038-\u303C\u3041-\u3096\u3099-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA62B\uA640-\uA66F\uA674-\uA67D\uA67F-\uA6F1\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AD\uA7B0-\uA7B7\uA7F7-\uA827\uA840-\uA873\uA880-\uA8C4\uA8D0-\uA8D9\uA8E0-\uA8F7\uA8FB\uA8FD\uA900-\uA92D\uA930-\uA953\uA960-\uA97C\uA980-\uA9C0\uA9CF-\uA9D9\uA9E0-\uA9FE\uAA00-\uAA36\uAA40-\uAA4D\uAA50-\uAA59\uAA60-\uAA76\uAA7A-\uAAC2\uAADB-\uAADD\uAAE0-\uAAEF\uAAF2-\uAAF6\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABEA\uABEC\uABED\uABF0-\uABF9\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE00-\uFE0F\uFE20-\uFE2F\uFE33\uFE34\uFE4D-\uFE4F\uFE70-\uFE74\uFE76-\uFEFC\uFF10-\uFF19\uFF21-\uFF3A\uFF3F\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDD40-\uDD74\uDDFD\uDE80-\uDE9C\uDEA0-\uDED0\uDEE0\uDF00-\uDF1F\uDF30-\uDF4A\uDF50-\uDF7A\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF\uDFD1-\uDFD5]|\uD801[\uDC00-\uDC9D\uDCA0-\uDCA9\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00-\uDE03\uDE05\uDE06\uDE0C-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE38-\uDE3A\uDE3F\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE6\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC00-\uDC46\uDC66-\uDC6F\uDC7F-\uDCBA\uDCD0-\uDCE8\uDCF0-\uDCF9\uDD00-\uDD34\uDD36-\uDD3F\uDD50-\uDD73\uDD76\uDD80-\uDDC4\uDDCA-\uDDCC\uDDD0-\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE37\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEEA\uDEF0-\uDEF9\uDF00-\uDF03\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3C-\uDF44\uDF47\uDF48\uDF4B-\uDF4D\uDF50\uDF57\uDF5D-\uDF63\uDF66-\uDF6C\uDF70-\uDF74]|\uD805[\uDC80-\uDCC5\uDCC7\uDCD0-\uDCD9\uDD80-\uDDB5\uDDB8-\uDDC0\uDDD8-\uDDDD\uDE00-\uDE40\uDE44\uDE50-\uDE59\uDE80-\uDEB7\uDEC0-\uDEC9\uDF00-\uDF19\uDF1D-\uDF2B\uDF30-\uDF39]|\uD806[\uDCA0-\uDCE9\uDCFF\uDEC0-\uDEF8]|\uD808[\uDC00-\uDF99]|\uD809[\uDC00-\uDC6E\uDC80-\uDD43]|[\uD80C\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDE60-\uDE69\uDED0-\uDEED\uDEF0-\uDEF4\uDF00-\uDF36\uDF40-\uDF43\uDF50-\uDF59\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50-\uDF7E\uDF8F-\uDF9F]|\uD82C[\uDC00\uDC01]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99\uDC9D\uDC9E]|\uD834[\uDD65-\uDD69\uDD6D-\uDD72\uDD7B-\uDD82\uDD85-\uDD8B\uDDAA-\uDDAD\uDE42-\uDE44]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB\uDFCE-\uDFFF]|\uD836[\uDE00-\uDE36\uDE3B-\uDE6C\uDE75\uDE84\uDE9B-\uDE9F\uDEA1-\uDEAF]|\uD83A[\uDC00-\uDCC4\uDCD0-\uDCD6]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1]|\uD87E[\uDC00-\uDE1D]|\uDB40[\uDD00-\uDDEF])+/,
        };
        function get_full_char(str, pos) {
            if (is_surrogate_pair_head(str.charCodeAt(pos))) {
                if (is_surrogate_pair_tail(str.charCodeAt(pos + 1))) {
                    return str.charAt(pos) + str.charAt(pos + 1);
                }
            }
            else if (is_surrogate_pair_tail(str.charCodeAt(pos))) {
                if (is_surrogate_pair_head(str.charCodeAt(pos - 1))) {
                    return str.charAt(pos - 1) + str.charAt(pos);
                }
            }
            return str.charAt(pos);
        }
        function get_full_char_code(str, pos) {
            // https://en.wikipedia.org/wiki/Universal_Character_Set_characters#Surrogates
            if (is_surrogate_pair_head(str.charCodeAt(pos))) {
                return 0x10000 + (str.charCodeAt(pos) - 0xd800 << 10) + str.charCodeAt(pos + 1) - 0xdc00;
            }
            return str.charCodeAt(pos);
        }
        function get_full_char_length(str) {
            var surrogates = 0;
            for (var i = 0; i < str.length; i++) {
                if (is_surrogate_pair_head(str.charCodeAt(i)) && is_surrogate_pair_tail(str.charCodeAt(i + 1))) {
                    surrogates++;
                    i++;
                }
            }
            return str.length - surrogates;
        }
        function from_char_code(code) {
            // Based on https://github.com/mathiasbynens/String.fromCodePoint/blob/master/fromcodepoint.js
            if (code > 0xFFFF) {
                code -= 0x10000;
                return (String.fromCharCode((code >> 10) + 0xD800) +
                    String.fromCharCode((code % 0x400) + 0xDC00));
            }
            return String.fromCharCode(code);
        }
        function is_surrogate_pair_head(code) {
            return code >= 0xd800 && code <= 0xdbff;
        }
        function is_surrogate_pair_tail(code) {
            return code >= 0xdc00 && code <= 0xdfff;
        }
        function is_digit(code) {
            return code >= 48 && code <= 57;
        }
        function is_identifier_start(ch) {
            return UNICODE.ID_Start.test(ch);
        }
        function is_identifier_char(ch) {
            return UNICODE.ID_Continue.test(ch);
        }
        const BASIC_IDENT = /^[a-z_$][a-z0-9_$]*$/i;
        function is_basic_identifier_string(str) {
            return BASIC_IDENT.test(str);
        }
        function is_identifier_string(str, allow_surrogates) {
            if (BASIC_IDENT.test(str)) {
                return true;
            }
            if (!allow_surrogates && /[\ud800-\udfff]/.test(str)) {
                return false;
            }
            var match = UNICODE.ID_Start.exec(str);
            if (!match || match.index !== 0) {
                return false;
            }
            str = str.slice(match[0].length);
            if (!str) {
                return true;
            }
            match = UNICODE.ID_Continue.exec(str);
            return !!match && match[0].length === str.length;
        }
        function parse_js_number(num, allow_e = true) {
            if (!allow_e && num.includes("e")) {
                return NaN;
            }
            if (RE_HEX_NUMBER.test(num)) {
                return parseInt(num.substr(2), 16);
            }
            else if (RE_OCT_NUMBER.test(num)) {
                return parseInt(num.substr(1), 8);
            }
            else if (RE_ES6_OCT_NUMBER.test(num)) {
                return parseInt(num.substr(2), 8);
            }
            else if (RE_BIN_NUMBER.test(num)) {
                return parseInt(num.substr(2), 2);
            }
            else if (RE_DEC_NUMBER.test(num)) {
                return parseFloat(num);
            }
            else {
                var val = parseFloat(num);
                if (val == num)
                    return val;
            }
        }
        class JS_Parse_Error extends Error {
            constructor(message, filename, line, col, pos) {
                super();
                this.name = "SyntaxError";
                this.message = message;
                this.filename = filename;
                this.line = line;
                this.col = col;
                this.pos = pos;
            }
        }
        function js_error(message, filename, line, col, pos) {
            throw new JS_Parse_Error(message, filename, line, col, pos);
        }
        function is_token(token, type, val) {
            return token.type == type && (val == null || token.value == val);
        }
        var EX_EOF = {};
        function tokenizer($TEXT, filename, html5_comments, shebang) {
            var S = {
                text: $TEXT,
                filename: filename,
                pos: 0,
                tokpos: 0,
                line: 1,
                tokline: 0,
                col: 0,
                tokcol: 0,
                newline_before: false,
                regex_allowed: false,
                brace_counter: 0,
                template_braces: [],
                comments_before: [],
                directives: {},
                directive_stack: []
            };
            function peek() { return get_full_char(S.text, S.pos); }
            // Used because parsing ?. involves a lookahead for a digit
            function is_option_chain_op() {
                const must_be_dot = S.text.charCodeAt(S.pos + 1) === 46;
                if (!must_be_dot)
                    return false;
                const cannot_be_digit = S.text.charCodeAt(S.pos + 2);
                return cannot_be_digit < 48 || cannot_be_digit > 57;
            }
            function next(signal_eof, in_string) {
                var ch = get_full_char(S.text, S.pos++);
                if (signal_eof && !ch)
                    throw EX_EOF;
                if (NEWLINE_CHARS.has(ch)) {
                    S.newline_before = S.newline_before || !in_string;
                    ++S.line;
                    S.col = 0;
                    if (ch == "\r" && peek() == "\n") {
                        // treat a \r\n sequence as a single \n
                        ++S.pos;
                        ch = "\n";
                    }
                }
                else {
                    if (ch.length > 1) {
                        ++S.pos;
                        ++S.col;
                    }
                    ++S.col;
                }
                return ch;
            }
            function forward(i) {
                while (i--)
                    next();
            }
            function looking_at(str) {
                return S.text.substr(S.pos, str.length) == str;
            }
            function find_eol() {
                var text = S.text;
                for (var i = S.pos, n = S.text.length; i < n; ++i) {
                    var ch = text[i];
                    if (NEWLINE_CHARS.has(ch))
                        return i;
                }
                return -1;
            }
            function find(what, signal_eof) {
                var pos = S.text.indexOf(what, S.pos);
                if (signal_eof && pos == -1)
                    throw EX_EOF;
                return pos;
            }
            function start_token() {
                S.tokline = S.line;
                S.tokcol = S.col;
                S.tokpos = S.pos;
            }
            var prev_was_dot = false;
            var previous_token = null;
            function token(type, value, is_comment) {
                S.regex_allowed = ((type == "operator" && !UNARY_POSTFIX.has(value)) ||
                    (type == "keyword" && KEYWORDS_BEFORE_EXPRESSION.has(value)) ||
                    (type == "punc" && PUNC_BEFORE_EXPRESSION.has(value))) ||
                    (type == "arrow");
                if (type == "punc" && (value == "." || value == "?.")) {
                    prev_was_dot = true;
                }
                else if (!is_comment) {
                    prev_was_dot = false;
                }
                const line = S.tokline;
                const col = S.tokcol;
                const pos = S.tokpos;
                const nlb = S.newline_before;
                const file = filename;
                let comments_before = [];
                let comments_after = [];
                if (!is_comment) {
                    comments_before = S.comments_before;
                    comments_after = S.comments_before = [];
                }
                S.newline_before = false;
                const tok = new AST_Token(type, value, line, col, pos, nlb, comments_before, comments_after, file);
                if (!is_comment)
                    previous_token = tok;
                return tok;
            }
            function skip_whitespace() {
                while (WHITESPACE_CHARS.has(peek()))
                    next();
            }
            function read_while(pred) {
                var ret = "", ch, i = 0;
                while ((ch = peek()) && pred(ch, i++))
                    ret += next();
                return ret;
            }
            function parse_error(err) {
                js_error(err, filename, S.tokline, S.tokcol, S.tokpos);
            }
            function read_num(prefix) {
                var has_e = false, after_e = false, has_x = false, has_dot = prefix == ".", is_big_int = false, numeric_separator = false;
                var num = read_while(function (ch, i) {
                    if (is_big_int)
                        return false;
                    var code = ch.charCodeAt(0);
                    switch (code) {
                        case 95: // _
                            return (numeric_separator = true);
                        case 98:
                        case 66: // bB
                            return (has_x = true); // Can occur in hex sequence, don't return false yet
                        case 111:
                        case 79: // oO
                        case 120:
                        case 88: // xX
                            return has_x ? false : (has_x = true);
                        case 101:
                        case 69: // eE
                            return has_x ? true : has_e ? false : (has_e = after_e = true);
                        case 45: // -
                            return after_e || (i == 0 && !prefix);
                        case 43: // +
                            return after_e;
                        case (after_e = false, 46): // .
                            return (!has_dot && !has_x && !has_e) ? (has_dot = true) : false;
                    }
                    if (ch === "n") {
                        is_big_int = true;
                        return true;
                    }
                    return RE_NUM_LITERAL.test(ch);
                });
                if (prefix)
                    num = prefix + num;
                LATEST_RAW = num;
                if (RE_OCT_NUMBER.test(num) && next_token.has_directive("use strict")) {
                    parse_error("Legacy octal literals are not allowed in strict mode");
                }
                if (numeric_separator) {
                    if (num.endsWith("_")) {
                        parse_error("Numeric separators are not allowed at the end of numeric literals");
                    }
                    else if (num.includes("__")) {
                        parse_error("Only one underscore is allowed as numeric separator");
                    }
                    num = num.replace(/_/g, "");
                }
                if (num.endsWith("n")) {
                    const without_n = num.slice(0, -1);
                    const allow_e = RE_HEX_NUMBER.test(without_n);
                    const valid = parse_js_number(without_n, allow_e);
                    if (!has_dot && RE_BIG_INT.test(num) && !isNaN(valid))
                        return token("big_int", without_n);
                    parse_error("Invalid or unexpected token");
                }
                var valid = parse_js_number(num);
                if (!isNaN(valid)) {
                    return token("num", valid);
                }
                else {
                    parse_error("Invalid syntax: " + num);
                }
            }
            function is_octal(ch) {
                return ch >= "0" && ch <= "7";
            }
            function read_escaped_char(in_string, strict_hex, template_string) {
                var ch = next(true, in_string);
                switch (ch.charCodeAt(0)) {
                    case 110: return "\n";
                    case 114: return "\r";
                    case 116: return "\t";
                    case 98: return "\b";
                    case 118: return "\u000b"; // \v
                    case 102: return "\f";
                    case 120: return String.fromCharCode(hex_bytes(2, strict_hex)); // \x
                    case 117: // \u
                        if (peek() == "{") {
                            next(true);
                            if (peek() === "}")
                                parse_error("Expecting hex-character between {}");
                            while (peek() == "0")
                                next(true); // No significance
                            var result, length = find("}", true) - S.pos;
                            // Avoid 32 bit integer overflow (1 << 32 === 1)
                            // We know first character isn't 0 and thus out of range anyway
                            if (length > 6 || (result = hex_bytes(length, strict_hex)) > 0x10FFFF) {
                                parse_error("Unicode reference out of bounds");
                            }
                            next(true);
                            return from_char_code(result);
                        }
                        return String.fromCharCode(hex_bytes(4, strict_hex));
                    case 10: return ""; // newline
                    case 13: // \r
                        if (peek() == "\n") { // DOS newline
                            next(true, in_string);
                            return "";
                        }
                }
                if (is_octal(ch)) {
                    if (template_string && strict_hex) {
                        const represents_null_character = ch === "0" && !is_octal(peek());
                        if (!represents_null_character) {
                            parse_error("Octal escape sequences are not allowed in template strings");
                        }
                    }
                    return read_octal_escape_sequence(ch, strict_hex);
                }
                return ch;
            }
            function read_octal_escape_sequence(ch, strict_octal) {
                // Read
                var p = peek();
                if (p >= "0" && p <= "7") {
                    ch += next(true);
                    if (ch[0] <= "3" && (p = peek()) >= "0" && p <= "7")
                        ch += next(true);
                }
                // Parse
                if (ch === "0")
                    return "\0";
                if (ch.length > 0 && next_token.has_directive("use strict") && strict_octal)
                    parse_error("Legacy octal escape sequences are not allowed in strict mode");
                return String.fromCharCode(parseInt(ch, 8));
            }
            function hex_bytes(n, strict_hex) {
                var num = 0;
                for (; n > 0; --n) {
                    if (!strict_hex && isNaN(parseInt(peek(), 16))) {
                        return parseInt(num, 16) || "";
                    }
                    var digit = next(true);
                    if (isNaN(parseInt(digit, 16)))
                        parse_error("Invalid hex-character pattern in string");
                    num += digit;
                }
                return parseInt(num, 16);
            }
            var read_string = with_eof_error("Unterminated string constant", function () {
                const start_pos = S.pos;
                var quote = next(), ret = [];
                for (;;) {
                    var ch = next(true, true);
                    if (ch == "\\")
                        ch = read_escaped_char(true, true);
                    else if (ch == "\r" || ch == "\n")
                        parse_error("Unterminated string constant");
                    else if (ch == quote)
                        break;
                    ret.push(ch);
                }
                var tok = token("string", ret.join(""));
                LATEST_RAW = S.text.slice(start_pos, S.pos);
                tok.quote = quote;
                return tok;
            });
            var read_template_characters = with_eof_error("Unterminated template", function (begin) {
                if (begin) {
                    S.template_braces.push(S.brace_counter);
                }
                var content = "", raw = "", ch, tok;
                next(true, true);
                while ((ch = next(true, true)) != "`") {
                    if (ch == "\r") {
                        if (peek() == "\n")
                            ++S.pos;
                        ch = "\n";
                    }
                    else if (ch == "$" && peek() == "{") {
                        next(true, true);
                        S.brace_counter++;
                        tok = token(begin ? "template_head" : "template_substitution", content);
                        LATEST_RAW = raw;
                        LATEST_TEMPLATE_END = false;
                        return tok;
                    }
                    raw += ch;
                    if (ch == "\\") {
                        var tmp = S.pos;
                        var prev_is_tag = previous_token && (previous_token.type === "name" || previous_token.type === "punc" && (previous_token.value === ")" || previous_token.value === "]"));
                        ch = read_escaped_char(true, !prev_is_tag, true);
                        raw += S.text.substr(tmp, S.pos - tmp);
                    }
                    content += ch;
                }
                S.template_braces.pop();
                tok = token(begin ? "template_head" : "template_substitution", content);
                LATEST_RAW = raw;
                LATEST_TEMPLATE_END = true;
                return tok;
            });
            function skip_line_comment(type) {
                var regex_allowed = S.regex_allowed;
                var i = find_eol(), ret;
                if (i == -1) {
                    ret = S.text.substr(S.pos);
                    S.pos = S.text.length;
                }
                else {
                    ret = S.text.substring(S.pos, i);
                    S.pos = i;
                }
                S.col = S.tokcol + (S.pos - S.tokpos);
                S.comments_before.push(token(type, ret, true));
                S.regex_allowed = regex_allowed;
                return next_token;
            }
            var skip_multiline_comment = with_eof_error("Unterminated multiline comment", function () {
                var regex_allowed = S.regex_allowed;
                var i = find("*/", true);
                var text = S.text.substring(S.pos, i).replace(/\r\n|\r|\u2028|\u2029/g, "\n");
                // update stream position
                forward(get_full_char_length(text) /* text length doesn't count \r\n as 2 char while S.pos - i does */ + 2);
                S.comments_before.push(token("comment2", text, true));
                S.newline_before = S.newline_before || text.includes("\n");
                S.regex_allowed = regex_allowed;
                return next_token;
            });
            var read_name = with_eof_error("Unterminated identifier name", function () {
                var name = [], ch, escaped = false;
                var read_escaped_identifier_char = function () {
                    escaped = true;
                    next();
                    if (peek() !== "u") {
                        parse_error("Expecting UnicodeEscapeSequence -- uXXXX or u{XXXX}");
                    }
                    return read_escaped_char(false, true);
                };
                // Read first character (ID_Start)
                if ((ch = peek()) === "\\") {
                    ch = read_escaped_identifier_char();
                    if (!is_identifier_start(ch)) {
                        parse_error("First identifier char is an invalid identifier char");
                    }
                }
                else if (is_identifier_start(ch)) {
                    next();
                }
                else {
                    return "";
                }
                name.push(ch);
                // Read ID_Continue
                while ((ch = peek()) != null) {
                    if ((ch = peek()) === "\\") {
                        ch = read_escaped_identifier_char();
                        if (!is_identifier_char(ch)) {
                            parse_error("Invalid escaped identifier char");
                        }
                    }
                    else {
                        if (!is_identifier_char(ch)) {
                            break;
                        }
                        next();
                    }
                    name.push(ch);
                }
                const name_str = name.join("");
                if (RESERVED_WORDS.has(name_str) && escaped) {
                    parse_error("Escaped characters are not allowed in keywords");
                }
                return name_str;
            });
            var read_regexp = with_eof_error("Unterminated regular expression", function (source) {
                var prev_backslash = false, ch, in_class = false;
                while ((ch = next(true)))
                    if (NEWLINE_CHARS.has(ch)) {
                        parse_error("Unexpected line terminator");
                    }
                    else if (prev_backslash) {
                        source += "\\" + ch;
                        prev_backslash = false;
                    }
                    else if (ch == "[") {
                        in_class = true;
                        source += ch;
                    }
                    else if (ch == "]" && in_class) {
                        in_class = false;
                        source += ch;
                    }
                    else if (ch == "/" && !in_class) {
                        break;
                    }
                    else if (ch == "\\") {
                        prev_backslash = true;
                    }
                    else {
                        source += ch;
                    }
                const flags = read_name();
                return token("regexp", "/" + source + "/" + flags);
            });
            function read_operator(prefix) {
                function grow(op) {
                    if (!peek())
                        return op;
                    var bigger = op + peek();
                    if (OPERATORS.has(bigger)) {
                        next();
                        return grow(bigger);
                    }
                    else {
                        return op;
                    }
                }
                return token("operator", grow(prefix || next()));
            }
            function handle_slash() {
                next();
                switch (peek()) {
                    case "/":
                        next();
                        return skip_line_comment("comment1");
                    case "*":
                        next();
                        return skip_multiline_comment();
                }
                return S.regex_allowed ? read_regexp("") : read_operator("/");
            }
            function handle_eq_sign() {
                next();
                if (peek() === ">") {
                    next();
                    return token("arrow", "=>");
                }
                else {
                    return read_operator("=");
                }
            }
            function handle_dot() {
                next();
                if (is_digit(peek().charCodeAt(0))) {
                    return read_num(".");
                }
                if (peek() === ".") {
                    next(); // Consume second dot
                    next(); // Consume third dot
                    return token("expand", "...");
                }
                return token("punc", ".");
            }
            function read_word() {
                var word = read_name();
                if (prev_was_dot)
                    return token("name", word);
                return KEYWORDS_ATOM.has(word) ? token("atom", word)
                    : !KEYWORDS.has(word) ? token("name", word)
                        : OPERATORS.has(word) ? token("operator", word)
                            : token("keyword", word);
            }
            function read_private_word() {
                next();
                return token("privatename", read_name());
            }
            function with_eof_error(eof_error, cont) {
                return function (x) {
                    try {
                        return cont(x);
                    }
                    catch (ex) {
                        if (ex === EX_EOF)
                            parse_error(eof_error);
                        else
                            throw ex;
                    }
                };
            }
            function next_token(force_regexp) {
                if (force_regexp != null)
                    return read_regexp(force_regexp);
                if (shebang && S.pos == 0 && looking_at("#!")) {
                    start_token();
                    forward(2);
                    skip_line_comment("comment5");
                }
                for (;;) {
                    skip_whitespace();
                    start_token();
                    if (html5_comments) {
                        if (looking_at("<!--")) {
                            forward(4);
                            skip_line_comment("comment3");
                            continue;
                        }
                        if (looking_at("-->") && S.newline_before) {
                            forward(3);
                            skip_line_comment("comment4");
                            continue;
                        }
                    }
                    var ch = peek();
                    if (!ch)
                        return token("eof");
                    var code = ch.charCodeAt(0);
                    switch (code) {
                        case 34:
                        case 39: return read_string();
                        case 46: return handle_dot();
                        case 47: {
                            var tok = handle_slash();
                            if (tok === next_token)
                                continue;
                            return tok;
                        }
                        case 61: return handle_eq_sign();
                        case 63: {
                            if (!is_option_chain_op())
                                break; // Handled below
                            next(); // ?
                            next(); // .
                            return token("punc", "?.");
                        }
                        case 96: return read_template_characters(true);
                        case 123:
                            S.brace_counter++;
                            break;
                        case 125:
                            S.brace_counter--;
                            if (S.template_braces.length > 0
                                && S.template_braces[S.template_braces.length - 1] === S.brace_counter)
                                return read_template_characters(false);
                            break;
                    }
                    if (is_digit(code))
                        return read_num();
                    if (PUNC_CHARS.has(ch))
                        return token("punc", next());
                    if (OPERATOR_CHARS.has(ch))
                        return read_operator();
                    if (code == 92 || is_identifier_start(ch))
                        return read_word();
                    if (code == 35)
                        return read_private_word();
                    break;
                }
                parse_error("Unexpected character '" + ch + "'");
            }
            next_token.next = next;
            next_token.peek = peek;
            next_token.context = function (nc) {
                if (nc)
                    S = nc;
                return S;
            };
            next_token.add_directive = function (directive) {
                S.directive_stack[S.directive_stack.length - 1].push(directive);
                if (S.directives[directive] === undefined) {
                    S.directives[directive] = 1;
                }
                else {
                    S.directives[directive]++;
                }
            };
            next_token.push_directives_stack = function () {
                S.directive_stack.push([]);
            };
            next_token.pop_directives_stack = function () {
                var directives = S.directive_stack[S.directive_stack.length - 1];
                for (var i = 0; i < directives.length; i++) {
                    S.directives[directives[i]]--;
                }
                S.directive_stack.pop();
            };
            next_token.has_directive = function (directive) {
                return S.directives[directive] > 0;
            };
            return next_token;
        }
        /* -----[ Parser (constants) ]----- */
        var UNARY_PREFIX = makePredicate([
            "typeof",
            "void",
            "delete",
            "--",
            "++",
            "!",
            "~",
            "-",
            "+"
        ]);
        var UNARY_POSTFIX = makePredicate(["--", "++"]);
        var ASSIGNMENT = makePredicate(["=", "+=", "-=", "??=", "&&=", "||=", "/=", "*=", "**=", "%=", ">>=", "<<=", ">>>=", "|=", "^=", "&="]);
        var LOGICAL_ASSIGNMENT = makePredicate(["??=", "&&=", "||="]);
        var PRECEDENCE = (function (a, ret) {
            for (var i = 0; i < a.length; ++i) {
                var b = a[i];
                for (var j = 0; j < b.length; ++j) {
                    ret[b[j]] = i + 1;
                }
            }
            return ret;
        })([
            ["||"],
            ["??"],
            ["&&"],
            ["|"],
            ["^"],
            ["&"],
            ["==", "===", "!=", "!=="],
            ["<", ">", "<=", ">=", "in", "instanceof"],
            [">>", "<<", ">>>"],
            ["+", "-"],
            ["*", "/", "%"],
            ["**"]
        ], {});
        var ATOMIC_START_TOKEN = makePredicate(["atom", "num", "big_int", "string", "regexp", "name"]);
        /* -----[ Parser ]----- */
        function parse($TEXT, options) {
            // maps start tokens to count of comments found outside of their parens
            // Example: /* I count */ ( /* I don't */ foo() )
            // Useful because comments_before property of call with parens outside
            // contains both comments inside and outside these parens. Used to find the
            // right #__PURE__ comments for an expression
            const outer_comments_before_counts = new WeakMap();
            options = defaults(options, {
                bare_returns: false,
                ecma: null,
                expression: false,
                filename: null,
                html5_comments: true,
                module: false,
                shebang: true,
                strict: false,
                toplevel: null,
            }, true);
            var S = {
                input: (typeof $TEXT == "string"
                    ? tokenizer($TEXT, options.filename, options.html5_comments, options.shebang)
                    : $TEXT),
                token: null,
                prev: null,
                peeked: null,
                in_function: 0,
                in_async: -1,
                in_generator: -1,
                in_directives: true,
                in_loop: 0,
                labels: []
            };
            S.token = next();
            function is(type, value) {
                return is_token(S.token, type, value);
            }
            function peek() { return S.peeked || (S.peeked = S.input()); }
            function next() {
                S.prev = S.token;
                if (!S.peeked)
                    peek();
                S.token = S.peeked;
                S.peeked = null;
                S.in_directives = S.in_directives && (S.token.type == "string" || is("punc", ";"));
                return S.token;
            }
            function prev() {
                return S.prev;
            }
            function croak(msg, line, col, pos) {
                var ctx = S.input.context();
                js_error(msg, ctx.filename, line != null ? line : ctx.tokline, col != null ? col : ctx.tokcol, pos != null ? pos : ctx.tokpos);
            }
            function token_error(token, msg) {
                croak(msg, token.line, token.col);
            }
            function unexpected(token) {
                if (token == null)
                    token = S.token;
                token_error(token, "Unexpected token: " + token.type + " (" + token.value + ")");
            }
            function expect_token(type, val) {
                if (is(type, val)) {
                    return next();
                }
                token_error(S.token, "Unexpected token " + S.token.type + " «" + S.token.value + "»" + ", expected " + type + " «" + val + "»");
            }
            function expect(punc) { return expect_token("punc", punc); }
            function has_newline_before(token) {
                return token.nlb || !token.comments_before.every((comment) => !comment.nlb);
            }
            function can_insert_semicolon() {
                return !options.strict
                    && (is("eof") || is("punc", "}") || has_newline_before(S.token));
            }
            function is_in_generator() {
                return S.in_generator === S.in_function;
            }
            function is_in_async() {
                return S.in_async === S.in_function;
            }
            function can_await() {
                return (S.in_async === S.in_function
                    || S.in_function === 0 && S.input.has_directive("use strict"));
            }
            function semicolon(optional) {
                if (is("punc", ";"))
                    next();
                else if (!optional && !can_insert_semicolon())
                    unexpected();
            }
            function parenthesised() {
                expect("(");
                var exp = expression(true);
                expect(")");
                return exp;
            }
            function embed_tokens(parser) {
                return function _embed_tokens_wrapper(...args) {
                    const start = S.token;
                    const expr = parser(...args);
                    expr.start = start;
                    expr.end = prev();
                    return expr;
                };
            }
            function handle_regexp() {
                if (is("operator", "/") || is("operator", "/=")) {
                    S.peeked = null;
                    S.token = S.input(S.token.value.substr(1)); // force regexp
                }
            }
            var statement = embed_tokens(function statement(is_export_default, is_for_body, is_if_body) {
                handle_regexp();
                switch (S.token.type) {
                    case "string":
                        if (S.in_directives) {
                            var token = peek();
                            if (!LATEST_RAW.includes("\\")
                                && (is_token(token, "punc", ";")
                                    || is_token(token, "punc", "}")
                                    || has_newline_before(token)
                                    || is_token(token, "eof"))) {
                                S.input.add_directive(S.token.value);
                            }
                            else {
                                S.in_directives = false;
                            }
                        }
                        var dir = S.in_directives, stat = simple_statement();
                        return dir && stat.body instanceof AST_String ? new AST_Directive(stat.body) : stat;
                    case "template_head":
                    case "num":
                    case "big_int":
                    case "regexp":
                    case "operator":
                    case "atom":
                        return simple_statement();
                    case "name":
                        if (S.token.value == "async" && is_token(peek(), "keyword", "function")) {
                            next();
                            next();
                            if (is_for_body) {
                                croak("functions are not allowed as the body of a loop");
                            }
                            return function_(AST_Defun, false, true, is_export_default);
                        }
                        if (S.token.value == "import" && !is_token(peek(), "punc", "(") && !is_token(peek(), "punc", ".")) {
                            next();
                            var node = import_();
                            semicolon();
                            return node;
                        }
                        return is_token(peek(), "punc", ":")
                            ? labeled_statement()
                            : simple_statement();
                    case "punc":
                        switch (S.token.value) {
                            case "{":
                                return new AST_BlockStatement({
                                    start: S.token,
                                    body: block_(),
                                    end: prev()
                                });
                            case "[":
                            case "(":
                                return simple_statement();
                            case ";":
                                S.in_directives = false;
                                next();
                                return new AST_EmptyStatement();
                            default:
                                unexpected();
                        }
                    case "keyword":
                        switch (S.token.value) {
                            case "break":
                                next();
                                return break_cont(AST_Break);
                            case "continue":
                                next();
                                return break_cont(AST_Continue);
                            case "debugger":
                                next();
                                semicolon();
                                return new AST_Debugger();
                            case "do":
                                next();
                                var body = in_loop(statement);
                                expect_token("keyword", "while");
                                var condition = parenthesised();
                                semicolon(true);
                                return new AST_Do({
                                    body: body,
                                    condition: condition
                                });
                            case "while":
                                next();
                                return new AST_While({
                                    condition: parenthesised(),
                                    body: in_loop(function () { return statement(false, true); })
                                });
                            case "for":
                                next();
                                return for_();
                            case "class":
                                next();
                                if (is_for_body) {
                                    croak("classes are not allowed as the body of a loop");
                                }
                                if (is_if_body) {
                                    croak("classes are not allowed as the body of an if");
                                }
                                return class_(AST_DefClass, is_export_default);
                            case "function":
                                next();
                                if (is_for_body) {
                                    croak("functions are not allowed as the body of a loop");
                                }
                                return function_(AST_Defun, false, false, is_export_default);
                            case "if":
                                next();
                                return if_();
                            case "return":
                                if (S.in_function == 0 && !options.bare_returns)
                                    croak("'return' outside of function");
                                next();
                                var value = null;
                                if (is("punc", ";")) {
                                    next();
                                }
                                else if (!can_insert_semicolon()) {
                                    value = expression(true);
                                    semicolon();
                                }
                                return new AST_Return({
                                    value: value
                                });
                            case "switch":
                                next();
                                return new AST_Switch({
                                    expression: parenthesised(),
                                    body: in_loop(switch_body_)
                                });
                            case "throw":
                                next();
                                if (has_newline_before(S.token))
                                    croak("Illegal newline after 'throw'");
                                var value = expression(true);
                                semicolon();
                                return new AST_Throw({
                                    value: value
                                });
                            case "try":
                                next();
                                return try_();
                            case "var":
                                next();
                                var node = var_();
                                semicolon();
                                return node;
                            case "let":
                                next();
                                var node = let_();
                                semicolon();
                                return node;
                            case "const":
                                next();
                                var node = const_();
                                semicolon();
                                return node;
                            case "with":
                                if (S.input.has_directive("use strict")) {
                                    croak("Strict mode may not include a with statement");
                                }
                                next();
                                return new AST_With({
                                    expression: parenthesised(),
                                    body: statement()
                                });
                            case "export":
                                if (!is_token(peek(), "punc", "(")) {
                                    next();
                                    var node = export_();
                                    if (is("punc", ";"))
                                        semicolon();
                                    return node;
                                }
                        }
                }
                unexpected();
            });
            function labeled_statement() {
                var label = as_symbol(AST_Label);
                if (label.name === "await" && is_in_async()) {
                    token_error(S.prev, "await cannot be used as label inside async function");
                }
                if (S.labels.some((l) => l.name === label.name)) {
                    // ECMA-262, 12.12: An ECMAScript program is considered
                    // syntactically incorrect if it contains a
                    // LabelledStatement that is enclosed by a
                    // LabelledStatement with the same Identifier as label.
                    croak("Label " + label.name + " defined twice");
                }
                expect(":");
                S.labels.push(label);
                var stat = statement();
                S.labels.pop();
                if (!(stat instanceof AST_IterationStatement)) {
                    // check for `continue` that refers to this label.
                    // those should be reported as syntax errors.
                    // https://github.com/mishoo/UglifyJS2/issues/287
                    label.references.forEach(function (ref) {
                        if (ref instanceof AST_Continue) {
                            ref = ref.label.start;
                            croak("Continue label `" + label.name + "` refers to non-IterationStatement.", ref.line, ref.col, ref.pos);
                        }
                    });
                }
                return new AST_LabeledStatement({ body: stat, label: label });
            }
            function simple_statement(tmp) {
                return new AST_SimpleStatement({ body: (tmp = expression(true), semicolon(), tmp) });
            }
            function break_cont(type) {
                var label = null, ldef;
                if (!can_insert_semicolon()) {
                    label = as_symbol(AST_LabelRef, true);
                }
                if (label != null) {
                    ldef = S.labels.find((l) => l.name === label.name);
                    if (!ldef)
                        croak("Undefined label " + label.name);
                    label.thedef = ldef;
                }
                else if (S.in_loop == 0)
                    croak(type.TYPE + " not inside a loop or switch");
                semicolon();
                var stat = new type({ label: label });
                if (ldef)
                    ldef.references.push(stat);
                return stat;
            }
            function for_() {
                var for_await_error = "`for await` invalid in this context";
                var await_tok = S.token;
                if (await_tok.type == "name" && await_tok.value == "await") {
                    if (!can_await()) {
                        token_error(await_tok, for_await_error);
                    }
                    next();
                }
                else {
                    await_tok = false;
                }
                expect("(");
                var init = null;
                if (!is("punc", ";")) {
                    init =
                        is("keyword", "var") ? (next(), var_(true)) :
                            is("keyword", "let") ? (next(), let_(true)) :
                                is("keyword", "const") ? (next(), const_(true)) :
                                    expression(true, true);
                    var is_in = is("operator", "in");
                    var is_of = is("name", "of");
                    if (await_tok && !is_of) {
                        token_error(await_tok, for_await_error);
                    }
                    if (is_in || is_of) {
                        if (init instanceof AST_Definitions) {
                            if (init.definitions.length > 1)
                                token_error(init.start, "Only one variable declaration allowed in for..in loop");
                        }
                        else if (!(is_assignable(init) || (init = to_destructuring(init)) instanceof AST_Destructuring)) {
                            token_error(init.start, "Invalid left-hand side in for..in loop");
                        }
                        next();
                        if (is_in) {
                            return for_in(init);
                        }
                        else {
                            return for_of(init, !!await_tok);
                        }
                    }
                }
                else if (await_tok) {
                    token_error(await_tok, for_await_error);
                }
                return regular_for(init);
            }
            function regular_for(init) {
                expect(";");
                var test = is("punc", ";") ? null : expression(true);
                expect(";");
                var step = is("punc", ")") ? null : expression(true);
                expect(")");
                return new AST_For({
                    init: init,
                    condition: test,
                    step: step,
                    body: in_loop(function () { return statement(false, true); })
                });
            }
            function for_of(init, is_await) {
                var lhs = init instanceof AST_Definitions ? init.definitions[0].name : null;
                var obj = expression(true);
                expect(")");
                return new AST_ForOf({
                    await: is_await,
                    init: init,
                    name: lhs,
                    object: obj,
                    body: in_loop(function () { return statement(false, true); })
                });
            }
            function for_in(init) {
                var obj = expression(true);
                expect(")");
                return new AST_ForIn({
                    init: init,
                    object: obj,
                    body: in_loop(function () { return statement(false, true); })
                });
            }
            var arrow_function = function (start, argnames, is_async) {
                if (has_newline_before(S.token)) {
                    croak("Unexpected newline before arrow (=>)");
                }
                expect_token("arrow", "=>");
                var body = _function_body(is("punc", "{"), false, is_async);
                var end = body instanceof Array && body.length ? body[body.length - 1].end :
                    body instanceof Array ? start :
                        body.end;
                return new AST_Arrow({
                    start: start,
                    end: end,
                    async: is_async,
                    argnames: argnames,
                    body: body
                });
            };
            var function_ = function (ctor, is_generator_property, is_async, is_export_default) {
                var in_statement = ctor === AST_Defun;
                var is_generator = is("operator", "*");
                if (is_generator) {
                    next();
                }
                var name = is("name") ? as_symbol(in_statement ? AST_SymbolDefun : AST_SymbolLambda) : null;
                if (in_statement && !name) {
                    if (is_export_default) {
                        ctor = AST_Function;
                    }
                    else {
                        unexpected();
                    }
                }
                if (name && ctor !== AST_Accessor && !(name instanceof AST_SymbolDeclaration))
                    unexpected(prev());
                var args = [];
                var body = _function_body(true, is_generator || is_generator_property, is_async, name, args);
                return new ctor({
                    start: args.start,
                    end: body.end,
                    is_generator: is_generator,
                    async: is_async,
                    name: name,
                    argnames: args,
                    body: body
                });
            };
            function track_used_binding_identifiers(is_parameter, strict) {
                var parameters = new Set();
                var duplicate = false;
                var default_assignment = false;
                var spread = false;
                var strict_mode = !!strict;
                var tracker = {
                    add_parameter: function (token) {
                        if (parameters.has(token.value)) {
                            if (duplicate === false) {
                                duplicate = token;
                            }
                            tracker.check_strict();
                        }
                        else {
                            parameters.add(token.value);
                            if (is_parameter) {
                                switch (token.value) {
                                    case "arguments":
                                    case "eval":
                                    case "yield":
                                        if (strict_mode) {
                                            token_error(token, "Unexpected " + token.value + " identifier as parameter inside strict mode");
                                        }
                                        break;
                                    default:
                                        if (RESERVED_WORDS.has(token.value)) {
                                            unexpected();
                                        }
                                }
                            }
                        }
                    },
                    mark_default_assignment: function (token) {
                        if (default_assignment === false) {
                            default_assignment = token;
                        }
                    },
                    mark_spread: function (token) {
                        if (spread === false) {
                            spread = token;
                        }
                    },
                    mark_strict_mode: function () {
                        strict_mode = true;
                    },
                    is_strict: function () {
                        return default_assignment !== false || spread !== false || strict_mode;
                    },
                    check_strict: function () {
                        if (tracker.is_strict() && duplicate !== false) {
                            token_error(duplicate, "Parameter " + duplicate.value + " was used already");
                        }
                    }
                };
                return tracker;
            }
            function parameters(params) {
                var used_parameters = track_used_binding_identifiers(true, S.input.has_directive("use strict"));
                expect("(");
                while (!is("punc", ")")) {
                    var param = parameter(used_parameters);
                    params.push(param);
                    if (!is("punc", ")")) {
                        expect(",");
                    }
                    if (param instanceof AST_Expansion) {
                        break;
                    }
                }
                next();
            }
            function parameter(used_parameters, symbol_type) {
                var param;
                var expand = false;
                if (used_parameters === undefined) {
                    used_parameters = track_used_binding_identifiers(true, S.input.has_directive("use strict"));
                }
                if (is("expand", "...")) {
                    expand = S.token;
                    used_parameters.mark_spread(S.token);
                    next();
                }
                param = binding_element(used_parameters, symbol_type);
                if (is("operator", "=") && expand === false) {
                    used_parameters.mark_default_assignment(S.token);
                    next();
                    param = new AST_DefaultAssign({
                        start: param.start,
                        left: param,
                        operator: "=",
                        right: expression(false),
                        end: S.token
                    });
                }
                if (expand !== false) {
                    if (!is("punc", ")")) {
                        unexpected();
                    }
                    param = new AST_Expansion({
                        start: expand,
                        expression: param,
                        end: expand
                    });
                }
                used_parameters.check_strict();
                return param;
            }
            function binding_element(used_parameters, symbol_type) {
                var elements = [];
                var first = true;
                var is_expand = false;
                var expand_token;
                var first_token = S.token;
                if (used_parameters === undefined) {
                    used_parameters = track_used_binding_identifiers(false, S.input.has_directive("use strict"));
                }
                symbol_type = symbol_type === undefined ? AST_SymbolFunarg : symbol_type;
                if (is("punc", "[")) {
                    next();
                    while (!is("punc", "]")) {
                        if (first) {
                            first = false;
                        }
                        else {
                            expect(",");
                        }
                        if (is("expand", "...")) {
                            is_expand = true;
                            expand_token = S.token;
                            used_parameters.mark_spread(S.token);
                            next();
                        }
                        if (is("punc")) {
                            switch (S.token.value) {
                                case ",":
                                    elements.push(new AST_Hole({
                                        start: S.token,
                                        end: S.token
                                    }));
                                    continue;
                                case "]": // Trailing comma after last element
                                    break;
                                case "[":
                                case "{":
                                    elements.push(binding_element(used_parameters, symbol_type));
                                    break;
                                default:
                                    unexpected();
                            }
                        }
                        else if (is("name")) {
                            used_parameters.add_parameter(S.token);
                            elements.push(as_symbol(symbol_type));
                        }
                        else {
                            croak("Invalid function parameter");
                        }
                        if (is("operator", "=") && is_expand === false) {
                            used_parameters.mark_default_assignment(S.token);
                            next();
                            elements[elements.length - 1] = new AST_DefaultAssign({
                                start: elements[elements.length - 1].start,
                                left: elements[elements.length - 1],
                                operator: "=",
                                right: expression(false),
                                end: S.token
                            });
                        }
                        if (is_expand) {
                            if (!is("punc", "]")) {
                                croak("Rest element must be last element");
                            }
                            elements[elements.length - 1] = new AST_Expansion({
                                start: expand_token,
                                expression: elements[elements.length - 1],
                                end: expand_token
                            });
                        }
                    }
                    expect("]");
                    used_parameters.check_strict();
                    return new AST_Destructuring({
                        start: first_token,
                        names: elements,
                        is_array: true,
                        end: prev()
                    });
                }
                else if (is("punc", "{")) {
                    next();
                    while (!is("punc", "}")) {
                        if (first) {
                            first = false;
                        }
                        else {
                            expect(",");
                        }
                        if (is("expand", "...")) {
                            is_expand = true;
                            expand_token = S.token;
                            used_parameters.mark_spread(S.token);
                            next();
                        }
                        if (is("name") && (is_token(peek(), "punc") || is_token(peek(), "operator")) && [",", "}", "="].includes(peek().value)) {
                            used_parameters.add_parameter(S.token);
                            var start = prev();
                            var value = as_symbol(symbol_type);
                            if (is_expand) {
                                elements.push(new AST_Expansion({
                                    start: expand_token,
                                    expression: value,
                                    end: value.end,
                                }));
                            }
                            else {
                                elements.push(new AST_ObjectKeyVal({
                                    start: start,
                                    key: value.name,
                                    value: value,
                                    end: value.end,
                                }));
                            }
                        }
                        else if (is("punc", "}")) {
                            continue; // Allow trailing hole
                        }
                        else {
                            var property_token = S.token;
                            var property = as_property_name();
                            if (property === null) {
                                unexpected(prev());
                            }
                            else if (prev().type === "name" && !is("punc", ":")) {
                                elements.push(new AST_ObjectKeyVal({
                                    start: prev(),
                                    key: property,
                                    value: new symbol_type({
                                        start: prev(),
                                        name: property,
                                        end: prev()
                                    }),
                                    end: prev()
                                }));
                            }
                            else {
                                expect(":");
                                elements.push(new AST_ObjectKeyVal({
                                    start: property_token,
                                    quote: property_token.quote,
                                    key: property,
                                    value: binding_element(used_parameters, symbol_type),
                                    end: prev()
                                }));
                            }
                        }
                        if (is_expand) {
                            if (!is("punc", "}")) {
                                croak("Rest element must be last element");
                            }
                        }
                        else if (is("operator", "=")) {
                            used_parameters.mark_default_assignment(S.token);
                            next();
                            elements[elements.length - 1].value = new AST_DefaultAssign({
                                start: elements[elements.length - 1].value.start,
                                left: elements[elements.length - 1].value,
                                operator: "=",
                                right: expression(false),
                                end: S.token
                            });
                        }
                    }
                    expect("}");
                    used_parameters.check_strict();
                    return new AST_Destructuring({
                        start: first_token,
                        names: elements,
                        is_array: false,
                        end: prev()
                    });
                }
                else if (is("name")) {
                    used_parameters.add_parameter(S.token);
                    return as_symbol(symbol_type);
                }
                else {
                    croak("Invalid function parameter");
                }
            }
            function params_or_seq_(allow_arrows, maybe_sequence) {
                var spread_token;
                var invalid_sequence;
                var trailing_comma;
                var a = [];
                expect("(");
                while (!is("punc", ")")) {
                    if (spread_token)
                        unexpected(spread_token);
                    if (is("expand", "...")) {
                        spread_token = S.token;
                        if (maybe_sequence)
                            invalid_sequence = S.token;
                        next();
                        a.push(new AST_Expansion({
                            start: prev(),
                            expression: expression(),
                            end: S.token,
                        }));
                    }
                    else {
                        a.push(expression());
                    }
                    if (!is("punc", ")")) {
                        expect(",");
                        if (is("punc", ")")) {
                            trailing_comma = prev();
                            if (maybe_sequence)
                                invalid_sequence = trailing_comma;
                        }
                    }
                }
                expect(")");
                if (allow_arrows && is("arrow", "=>")) {
                    if (spread_token && trailing_comma)
                        unexpected(trailing_comma);
                }
                else if (invalid_sequence) {
                    unexpected(invalid_sequence);
                }
                return a;
            }
            function _function_body(block, generator, is_async, name, args) {
                var loop = S.in_loop;
                var labels = S.labels;
                var current_generator = S.in_generator;
                var current_async = S.in_async;
                ++S.in_function;
                if (generator)
                    S.in_generator = S.in_function;
                if (is_async)
                    S.in_async = S.in_function;
                if (args)
                    parameters(args);
                if (block)
                    S.in_directives = true;
                S.in_loop = 0;
                S.labels = [];
                if (block) {
                    S.input.push_directives_stack();
                    var a = block_();
                    if (name)
                        _verify_symbol(name);
                    if (args)
                        args.forEach(_verify_symbol);
                    S.input.pop_directives_stack();
                }
                else {
                    var a = [new AST_Return({
                            start: S.token,
                            value: expression(false),
                            end: S.token
                        })];
                }
                --S.in_function;
                S.in_loop = loop;
                S.labels = labels;
                S.in_generator = current_generator;
                S.in_async = current_async;
                return a;
            }
            function _await_expression() {
                // Previous token must be "await" and not be interpreted as an identifier
                if (!can_await()) {
                    croak("Unexpected await expression outside async function", S.prev.line, S.prev.col, S.prev.pos);
                }
                // the await expression is parsed as a unary expression in Babel
                return new AST_Await({
                    start: prev(),
                    end: S.token,
                    expression: maybe_unary(true),
                });
            }
            function _yield_expression() {
                // Previous token must be keyword yield and not be interpret as an identifier
                if (!is_in_generator()) {
                    croak("Unexpected yield expression outside generator function", S.prev.line, S.prev.col, S.prev.pos);
                }
                var start = S.token;
                var star = false;
                var has_expression = true;
                // Attempt to get expression or star (and then the mandatory expression)
                // behind yield on the same line.
                //
                // If nothing follows on the same line of the yieldExpression,
                // it should default to the value `undefined` for yield to return.
                // In that case, the `undefined` stored as `null` in ast.
                //
                // Note 1: It isn't allowed for yield* to close without an expression
                // Note 2: If there is a nlb between yield and star, it is interpret as
                //         yield <explicit undefined> <inserted automatic semicolon> *
                if (can_insert_semicolon() ||
                    (is("punc") && PUNC_AFTER_EXPRESSION.has(S.token.value))) {
                    has_expression = false;
                }
                else if (is("operator", "*")) {
                    star = true;
                    next();
                }
                return new AST_Yield({
                    start: start,
                    is_star: star,
                    expression: has_expression ? expression() : null,
                    end: prev()
                });
            }
            function if_() {
                var cond = parenthesised(), body = statement(false, false, true), belse = null;
                if (is("keyword", "else")) {
                    next();
                    belse = statement(false, false, true);
                }
                return new AST_If({
                    condition: cond,
                    body: body,
                    alternative: belse
                });
            }
            function block_() {
                expect("{");
                var a = [];
                while (!is("punc", "}")) {
                    if (is("eof"))
                        unexpected();
                    a.push(statement());
                }
                next();
                return a;
            }
            function switch_body_() {
                expect("{");
                var a = [], cur = null, branch = null, tmp;
                while (!is("punc", "}")) {
                    if (is("eof"))
                        unexpected();
                    if (is("keyword", "case")) {
                        if (branch)
                            branch.end = prev();
                        cur = [];
                        branch = new AST_Case({
                            start: (tmp = S.token, next(), tmp),
                            expression: expression(true),
                            body: cur
                        });
                        a.push(branch);
                        expect(":");
                    }
                    else if (is("keyword", "default")) {
                        if (branch)
                            branch.end = prev();
                        cur = [];
                        branch = new AST_Default({
                            start: (tmp = S.token, next(), expect(":"), tmp),
                            body: cur
                        });
                        a.push(branch);
                    }
                    else {
                        if (!cur)
                            unexpected();
                        cur.push(statement());
                    }
                }
                if (branch)
                    branch.end = prev();
                next();
                return a;
            }
            function try_() {
                var body = block_(), bcatch = null, bfinally = null;
                if (is("keyword", "catch")) {
                    var start = S.token;
                    next();
                    if (is("punc", "{")) {
                        var name = null;
                    }
                    else {
                        expect("(");
                        var name = parameter(undefined, AST_SymbolCatch);
                        expect(")");
                    }
                    bcatch = new AST_Catch({
                        start: start,
                        argname: name,
                        body: block_(),
                        end: prev()
                    });
                }
                if (is("keyword", "finally")) {
                    var start = S.token;
                    next();
                    bfinally = new AST_Finally({
                        start: start,
                        body: block_(),
                        end: prev()
                    });
                }
                if (!bcatch && !bfinally)
                    croak("Missing catch/finally blocks");
                return new AST_Try({
                    body: body,
                    bcatch: bcatch,
                    bfinally: bfinally
                });
            }
            function vardefs(no_in, kind) {
                var a = [];
                var def;
                for (;;) {
                    var sym_type = kind === "var" ? AST_SymbolVar :
                        kind === "const" ? AST_SymbolConst :
                            kind === "let" ? AST_SymbolLet : null;
                    if (is("punc", "{") || is("punc", "[")) {
                        def = new AST_VarDef({
                            start: S.token,
                            name: binding_element(undefined, sym_type),
                            value: is("operator", "=") ? (expect_token("operator", "="), expression(false, no_in)) : null,
                            end: prev()
                        });
                    }
                    else {
                        def = new AST_VarDef({
                            start: S.token,
                            name: as_symbol(sym_type),
                            value: is("operator", "=")
                                ? (next(), expression(false, no_in))
                                : !no_in && kind === "const"
                                    ? croak("Missing initializer in const declaration") : null,
                            end: prev()
                        });
                        if (def.name.name == "import")
                            croak("Unexpected token: import");
                    }
                    a.push(def);
                    if (!is("punc", ","))
                        break;
                    next();
                }
                return a;
            }
            var var_ = function (no_in) {
                return new AST_Var({
                    start: prev(),
                    definitions: vardefs(no_in, "var"),
                    end: prev()
                });
            };
            var let_ = function (no_in) {
                return new AST_Let({
                    start: prev(),
                    definitions: vardefs(no_in, "let"),
                    end: prev()
                });
            };
            var const_ = function (no_in) {
                return new AST_Const({
                    start: prev(),
                    definitions: vardefs(no_in, "const"),
                    end: prev()
                });
            };
            var new_ = function (allow_calls) {
                var start = S.token;
                expect_token("operator", "new");
                if (is("punc", ".")) {
                    next();
                    expect_token("name", "target");
                    return subscripts(new AST_NewTarget({
                        start: start,
                        end: prev()
                    }), allow_calls);
                }
                var newexp = expr_atom(false), args;
                if (is("punc", "(")) {
                    next();
                    args = expr_list(")", true);
                }
                else {
                    args = [];
                }
                var call = new AST_New({
                    start: start,
                    expression: newexp,
                    args: args,
                    end: prev()
                });
                annotate(call);
                return subscripts(call, allow_calls);
            };
            function as_atom_node() {
                var tok = S.token, ret;
                switch (tok.type) {
                    case "name":
                        ret = _make_symbol(AST_SymbolRef);
                        break;
                    case "num":
                        ret = new AST_Number({
                            start: tok,
                            end: tok,
                            value: tok.value,
                            raw: LATEST_RAW
                        });
                        break;
                    case "big_int":
                        ret = new AST_BigInt({ start: tok, end: tok, value: tok.value });
                        break;
                    case "string":
                        ret = new AST_String({
                            start: tok,
                            end: tok,
                            value: tok.value,
                            quote: tok.quote
                        });
                        break;
                    case "regexp":
                        const [_, source, flags] = tok.value.match(/^\/(.*)\/(\w*)$/);
                        ret = new AST_RegExp({ start: tok, end: tok, value: { source, flags } });
                        break;
                    case "atom":
                        switch (tok.value) {
                            case "false":
                                ret = new AST_False({ start: tok, end: tok });
                                break;
                            case "true":
                                ret = new AST_True({ start: tok, end: tok });
                                break;
                            case "null":
                                ret = new AST_Null({ start: tok, end: tok });
                                break;
                        }
                        break;
                }
                next();
                return ret;
            }
            function to_fun_args(ex, default_seen_above) {
                var insert_default = function (ex, default_value) {
                    if (default_value) {
                        return new AST_DefaultAssign({
                            start: ex.start,
                            left: ex,
                            operator: "=",
                            right: default_value,
                            end: default_value.end
                        });
                    }
                    return ex;
                };
                if (ex instanceof AST_Object) {
                    return insert_default(new AST_Destructuring({
                        start: ex.start,
                        end: ex.end,
                        is_array: false,
                        names: ex.properties.map(prop => to_fun_args(prop))
                    }), default_seen_above);
                }
                else if (ex instanceof AST_ObjectKeyVal) {
                    ex.value = to_fun_args(ex.value);
                    return insert_default(ex, default_seen_above);
                }
                else if (ex instanceof AST_Hole) {
                    return ex;
                }
                else if (ex instanceof AST_Destructuring) {
                    ex.names = ex.names.map(name => to_fun_args(name));
                    return insert_default(ex, default_seen_above);
                }
                else if (ex instanceof AST_SymbolRef) {
                    return insert_default(new AST_SymbolFunarg({
                        name: ex.name,
                        start: ex.start,
                        end: ex.end
                    }), default_seen_above);
                }
                else if (ex instanceof AST_Expansion) {
                    ex.expression = to_fun_args(ex.expression);
                    return insert_default(ex, default_seen_above);
                }
                else if (ex instanceof AST_Array) {
                    return insert_default(new AST_Destructuring({
                        start: ex.start,
                        end: ex.end,
                        is_array: true,
                        names: ex.elements.map(elm => to_fun_args(elm))
                    }), default_seen_above);
                }
                else if (ex instanceof AST_Assign) {
                    return insert_default(to_fun_args(ex.left, ex.right), default_seen_above);
                }
                else if (ex instanceof AST_DefaultAssign) {
                    ex.left = to_fun_args(ex.left);
                    return ex;
                }
                else {
                    croak("Invalid function parameter", ex.start.line, ex.start.col);
                }
            }
            var expr_atom = function (allow_calls, allow_arrows) {
                if (is("operator", "new")) {
                    return new_(allow_calls);
                }
                if (is("operator", "import")) {
                    return import_meta();
                }
                var start = S.token;
                var peeked;
                var async = is("name", "async")
                    && (peeked = peek()).value != "["
                    && peeked.type != "arrow"
                    && as_atom_node();
                if (is("punc")) {
                    switch (S.token.value) {
                        case "(":
                            if (async && !allow_calls)
                                break;
                            var exprs = params_or_seq_(allow_arrows, !async);
                            if (allow_arrows && is("arrow", "=>")) {
                                return arrow_function(start, exprs.map(e => to_fun_args(e)), !!async);
                            }
                            var ex = async ? new AST_Call({
                                expression: async,
                                args: exprs
                            }) : exprs.length == 1 ? exprs[0] : new AST_Sequence({
                                expressions: exprs
                            });
                            if (ex.start) {
                                const outer_comments_before = start.comments_before.length;
                                outer_comments_before_counts.set(start, outer_comments_before);
                                ex.start.comments_before.unshift(...start.comments_before);
                                start.comments_before = ex.start.comments_before;
                                if (outer_comments_before == 0 && start.comments_before.length > 0) {
                                    var comment = start.comments_before[0];
                                    if (!comment.nlb) {
                                        comment.nlb = start.nlb;
                                        start.nlb = false;
                                    }
                                }
                                start.comments_after = ex.start.comments_after;
                            }
                            ex.start = start;
                            var end = prev();
                            if (ex.end) {
                                end.comments_before = ex.end.comments_before;
                                ex.end.comments_after.push(...end.comments_after);
                                end.comments_after = ex.end.comments_after;
                            }
                            ex.end = end;
                            if (ex instanceof AST_Call)
                                annotate(ex);
                            return subscripts(ex, allow_calls);
                        case "[":
                            return subscripts(array_(), allow_calls);
                        case "{":
                            return subscripts(object_or_destructuring_(), allow_calls);
                    }
                    if (!async)
                        unexpected();
                }
                if (allow_arrows && is("name") && is_token(peek(), "arrow")) {
                    var param = new AST_SymbolFunarg({
                        name: S.token.value,
                        start: start,
                        end: start,
                    });
                    next();
                    return arrow_function(start, [param], !!async);
                }
                if (is("keyword", "function")) {
                    next();
                    var func = function_(AST_Function, false, !!async);
                    func.start = start;
                    func.end = prev();
                    return subscripts(func, allow_calls);
                }
                if (async)
                    return subscripts(async, allow_calls);
                if (is("keyword", "class")) {
                    next();
                    var cls = class_(AST_ClassExpression);
                    cls.start = start;
                    cls.end = prev();
                    return subscripts(cls, allow_calls);
                }
                if (is("template_head")) {
                    return subscripts(template_string(), allow_calls);
                }
                if (ATOMIC_START_TOKEN.has(S.token.type)) {
                    return subscripts(as_atom_node(), allow_calls);
                }
                unexpected();
            };
            function template_string() {
                var segments = [], start = S.token;
                segments.push(new AST_TemplateSegment({
                    start: S.token,
                    raw: LATEST_RAW,
                    value: S.token.value,
                    end: S.token
                }));
                while (!LATEST_TEMPLATE_END) {
                    next();
                    handle_regexp();
                    segments.push(expression(true));
                    segments.push(new AST_TemplateSegment({
                        start: S.token,
                        raw: LATEST_RAW,
                        value: S.token.value,
                        end: S.token
                    }));
                }
                next();
                return new AST_TemplateString({
                    start: start,
                    segments: segments,
                    end: S.token
                });
            }
            function expr_list(closing, allow_trailing_comma, allow_empty) {
                var first = true, a = [];
                while (!is("punc", closing)) {
                    if (first)
                        first = false;
                    else
                        expect(",");
                    if (allow_trailing_comma && is("punc", closing))
                        break;
                    if (is("punc", ",") && allow_empty) {
                        a.push(new AST_Hole({ start: S.token, end: S.token }));
                    }
                    else if (is("expand", "...")) {
                        next();
                        a.push(new AST_Expansion({ start: prev(), expression: expression(), end: S.token }));
                    }
                    else {
                        a.push(expression(false));
                    }
                }
                next();
                return a;
            }
            var array_ = embed_tokens(function () {
                expect("[");
                return new AST_Array({
                    elements: expr_list("]", !options.strict, true)
                });
            });
            var create_accessor = embed_tokens((is_generator, is_async) => {
                return function_(AST_Accessor, is_generator, is_async);
            });
            var object_or_destructuring_ = embed_tokens(function object_or_destructuring_() {
                var start = S.token, first = true, a = [];
                expect("{");
                while (!is("punc", "}")) {
                    if (first)
                        first = false;
                    else
                        expect(",");
                    if (!options.strict && is("punc", "}"))
                        // allow trailing comma
                        break;
                    start = S.token;
                    if (start.type == "expand") {
                        next();
                        a.push(new AST_Expansion({
                            start: start,
                            expression: expression(false),
                            end: prev(),
                        }));
                        continue;
                    }
                    var name = as_property_name();
                    var value;
                    // Check property and fetch value
                    if (!is("punc", ":")) {
                        var concise = concise_method_or_getset(name, start);
                        if (concise) {
                            a.push(concise);
                            continue;
                        }
                        value = new AST_SymbolRef({
                            start: prev(),
                            name: name,
                            end: prev()
                        });
                    }
                    else if (name === null) {
                        unexpected(prev());
                    }
                    else {
                        next(); // `:` - see first condition
                        value = expression(false);
                    }
                    // Check for default value and alter value accordingly if necessary
                    if (is("operator", "=")) {
                        next();
                        value = new AST_Assign({
                            start: start,
                            left: value,
                            operator: "=",
                            right: expression(false),
                            logical: false,
                            end: prev()
                        });
                    }
                    // Create property
                    a.push(new AST_ObjectKeyVal({
                        start: start,
                        quote: start.quote,
                        key: name instanceof AST_Node ? name : "" + name,
                        value: value,
                        end: prev()
                    }));
                }
                next();
                return new AST_Object({ properties: a });
            });
            function class_(KindOfClass, is_export_default) {
                var start, method, class_name, extends_, a = [];
                S.input.push_directives_stack(); // Push directive stack, but not scope stack
                S.input.add_directive("use strict");
                if (S.token.type == "name" && S.token.value != "extends") {
                    class_name = as_symbol(KindOfClass === AST_DefClass ? AST_SymbolDefClass : AST_SymbolClass);
                }
                if (KindOfClass === AST_DefClass && !class_name) {
                    if (is_export_default) {
                        KindOfClass = AST_ClassExpression;
                    }
                    else {
                        unexpected();
                    }
                }
                if (S.token.value == "extends") {
                    next();
                    extends_ = expression(true);
                }
                expect("{");
                while (is("punc", ";")) {
                    next();
                } // Leading semicolons are okay in class bodies.
                while (!is("punc", "}")) {
                    start = S.token;
                    method = concise_method_or_getset(as_property_name(), start, true);
                    if (!method) {
                        unexpected();
                    }
                    a.push(method);
                    while (is("punc", ";")) {
                        next();
                    }
                }
                S.input.pop_directives_stack();
                next();
                return new KindOfClass({
                    start: start,
                    name: class_name,
                    extends: extends_,
                    properties: a,
                    end: prev(),
                });
            }
            function concise_method_or_getset(name, start, is_class) {
                const get_symbol_ast = (name, SymbolClass = AST_SymbolMethod) => {
                    if (typeof name === "string" || typeof name === "number") {
                        return new SymbolClass({
                            start,
                            name: "" + name,
                            end: prev()
                        });
                    }
                    else if (name === null) {
                        unexpected();
                    }
                    return name;
                };
                const is_not_method_start = () => !is("punc", "(") && !is("punc", ",") && !is("punc", "}") && !is("operator", "=");
                var is_async = false;
                var is_static = false;
                var is_generator = false;
                var is_private = false;
                var accessor_type = null;
                if (is_class && name === "static" && is_not_method_start()) {
                    is_static = true;
                    name = as_property_name();
                }
                if (name === "async" && is_not_method_start()) {
                    is_async = true;
                    name = as_property_name();
                }
                if (prev().type === "operator" && prev().value === "*") {
                    is_generator = true;
                    name = as_property_name();
                }
                if ((name === "get" || name === "set") && is_not_method_start()) {
                    accessor_type = name;
                    name = as_property_name();
                }
                if (prev().type === "privatename") {
                    is_private = true;
                }
                const property_token = prev();
                if (accessor_type != null) {
                    if (!is_private) {
                        const AccessorClass = accessor_type === "get"
                            ? AST_ObjectGetter
                            : AST_ObjectSetter;
                        name = get_symbol_ast(name);
                        return new AccessorClass({
                            start,
                            static: is_static,
                            key: name,
                            quote: name instanceof AST_SymbolMethod ? property_token.quote : undefined,
                            value: create_accessor(),
                            end: prev()
                        });
                    }
                    else {
                        const AccessorClass = accessor_type === "get"
                            ? AST_PrivateGetter
                            : AST_PrivateSetter;
                        return new AccessorClass({
                            start,
                            static: is_static,
                            key: get_symbol_ast(name),
                            value: create_accessor(),
                            end: prev(),
                        });
                    }
                }
                if (is("punc", "(")) {
                    name = get_symbol_ast(name);
                    const AST_MethodVariant = is_private
                        ? AST_PrivateMethod
                        : AST_ConciseMethod;
                    var node = new AST_MethodVariant({
                        start: start,
                        static: is_static,
                        is_generator: is_generator,
                        async: is_async,
                        key: name,
                        quote: name instanceof AST_SymbolMethod ?
                            property_token.quote : undefined,
                        value: create_accessor(is_generator, is_async),
                        end: prev()
                    });
                    return node;
                }
                if (is_class) {
                    const key = get_symbol_ast(name, AST_SymbolClassProperty);
                    const quote = key instanceof AST_SymbolClassProperty
                        ? property_token.quote
                        : undefined;
                    const AST_ClassPropertyVariant = is_private
                        ? AST_ClassPrivateProperty
                        : AST_ClassProperty;
                    if (is("operator", "=")) {
                        next();
                        return new AST_ClassPropertyVariant({
                            start,
                            static: is_static,
                            quote,
                            key,
                            value: expression(false),
                            end: prev()
                        });
                    }
                    else if (is("name")
                        || is("privatename")
                        || is("operator", "*")
                        || is("punc", ";")
                        || is("punc", "}")) {
                        return new AST_ClassPropertyVariant({
                            start,
                            static: is_static,
                            quote,
                            key,
                            end: prev()
                        });
                    }
                }
            }
            function import_() {
                var start = prev();
                var imported_name;
                var imported_names;
                if (is("name")) {
                    imported_name = as_symbol(AST_SymbolImport);
                }
                if (is("punc", ",")) {
                    next();
                }
                imported_names = map_names(true);
                if (imported_names || imported_name) {
                    expect_token("name", "from");
                }
                var mod_str = S.token;
                if (mod_str.type !== "string") {
                    unexpected();
                }
                next();
                return new AST_Import({
                    start: start,
                    imported_name: imported_name,
                    imported_names: imported_names,
                    module_name: new AST_String({
                        start: mod_str,
                        value: mod_str.value,
                        quote: mod_str.quote,
                        end: mod_str,
                    }),
                    end: S.token,
                });
            }
            function import_meta() {
                var start = S.token;
                expect_token("operator", "import");
                expect_token("punc", ".");
                expect_token("name", "meta");
                return subscripts(new AST_ImportMeta({
                    start: start,
                    end: prev()
                }), false);
            }
            function map_name(is_import) {
                function make_symbol(type) {
                    return new type({
                        name: as_property_name(),
                        start: prev(),
                        end: prev()
                    });
                }
                var foreign_type = is_import ? AST_SymbolImportForeign : AST_SymbolExportForeign;
                var type = is_import ? AST_SymbolImport : AST_SymbolExport;
                var start = S.token;
                var foreign_name;
                var name;
                if (is_import) {
                    foreign_name = make_symbol(foreign_type);
                }
                else {
                    name = make_symbol(type);
                }
                if (is("name", "as")) {
                    next(); // The "as" word
                    if (is_import) {
                        name = make_symbol(type);
                    }
                    else {
                        foreign_name = make_symbol(foreign_type);
                    }
                }
                else if (is_import) {
                    name = new type(foreign_name);
                }
                else {
                    foreign_name = new foreign_type(name);
                }
                return new AST_NameMapping({
                    start: start,
                    foreign_name: foreign_name,
                    name: name,
                    end: prev(),
                });
            }
            function map_nameAsterisk(is_import, name) {
                var foreign_type = is_import ? AST_SymbolImportForeign : AST_SymbolExportForeign;
                var type = is_import ? AST_SymbolImport : AST_SymbolExport;
                var start = S.token;
                var foreign_name;
                var end = prev();
                name = name || new type({
                    name: "*",
                    start: start,
                    end: end,
                });
                foreign_name = new foreign_type({
                    name: "*",
                    start: start,
                    end: end,
                });
                return new AST_NameMapping({
                    start: start,
                    foreign_name: foreign_name,
                    name: name,
                    end: end,
                });
            }
            function map_names(is_import) {
                var names;
                if (is("punc", "{")) {
                    next();
                    names = [];
                    while (!is("punc", "}")) {
                        names.push(map_name(is_import));
                        if (is("punc", ",")) {
                            next();
                        }
                    }
                    next();
                }
                else if (is("operator", "*")) {
                    var name;
                    next();
                    if (is_import && is("name", "as")) {
                        next(); // The "as" word
                        name = as_symbol(is_import ? AST_SymbolImport : AST_SymbolExportForeign);
                    }
                    names = [map_nameAsterisk(is_import, name)];
                }
                return names;
            }
            function export_() {
                var start = S.token;
                var is_default;
                var exported_names;
                if (is("keyword", "default")) {
                    is_default = true;
                    next();
                }
                else if (exported_names = map_names(false)) {
                    if (is("name", "from")) {
                        next();
                        var mod_str = S.token;
                        if (mod_str.type !== "string") {
                            unexpected();
                        }
                        next();
                        return new AST_Export({
                            start: start,
                            is_default: is_default,
                            exported_names: exported_names,
                            module_name: new AST_String({
                                start: mod_str,
                                value: mod_str.value,
                                quote: mod_str.quote,
                                end: mod_str,
                            }),
                            end: prev(),
                        });
                    }
                    else {
                        return new AST_Export({
                            start: start,
                            is_default: is_default,
                            exported_names: exported_names,
                            end: prev(),
                        });
                    }
                }
                var node;
                var exported_value;
                var exported_definition;
                if (is("punc", "{")
                    || is_default
                        && (is("keyword", "class") || is("keyword", "function"))
                        && is_token(peek(), "punc")) {
                    exported_value = expression(false);
                    semicolon();
                }
                else if ((node = statement(is_default)) instanceof AST_Definitions && is_default) {
                    unexpected(node.start);
                }
                else if (node instanceof AST_Definitions
                    || node instanceof AST_Defun
                    || node instanceof AST_DefClass) {
                    exported_definition = node;
                }
                else if (node instanceof AST_ClassExpression
                    || node instanceof AST_Function) {
                    exported_value = node;
                }
                else if (node instanceof AST_SimpleStatement) {
                    exported_value = node.body;
                }
                else {
                    unexpected(node.start);
                }
                return new AST_Export({
                    start: start,
                    is_default: is_default,
                    exported_value: exported_value,
                    exported_definition: exported_definition,
                    end: prev(),
                });
            }
            function as_property_name() {
                var tmp = S.token;
                switch (tmp.type) {
                    case "punc":
                        if (tmp.value === "[") {
                            next();
                            var ex = expression(false);
                            expect("]");
                            return ex;
                        }
                        else
                            unexpected(tmp);
                    case "operator":
                        if (tmp.value === "*") {
                            next();
                            return null;
                        }
                        if (!["delete", "in", "instanceof", "new", "typeof", "void"].includes(tmp.value)) {
                            unexpected(tmp);
                        }
                    /* falls through */
                    case "name":
                    case "privatename":
                    case "string":
                    case "num":
                    case "big_int":
                    case "keyword":
                    case "atom":
                        next();
                        return tmp.value;
                    default:
                        unexpected(tmp);
                }
            }
            function as_name() {
                var tmp = S.token;
                if (tmp.type != "name" && tmp.type != "privatename")
                    unexpected();
                next();
                return tmp.value;
            }
            function _make_symbol(type) {
                var name = S.token.value;
                return new (name == "this" ? AST_This :
                    name == "super" ? AST_Super :
                        type)({
                    name: String(name),
                    start: S.token,
                    end: S.token
                });
            }
            function _verify_symbol(sym) {
                var name = sym.name;
                if (is_in_generator() && name == "yield") {
                    token_error(sym.start, "Yield cannot be used as identifier inside generators");
                }
                if (S.input.has_directive("use strict")) {
                    if (name == "yield") {
                        token_error(sym.start, "Unexpected yield identifier inside strict mode");
                    }
                    if (sym instanceof AST_SymbolDeclaration && (name == "arguments" || name == "eval")) {
                        token_error(sym.start, "Unexpected " + name + " in strict mode");
                    }
                }
            }
            function as_symbol(type, noerror) {
                if (!is("name")) {
                    if (!noerror)
                        croak("Name expected");
                    return null;
                }
                var sym = _make_symbol(type);
                _verify_symbol(sym);
                next();
                return sym;
            }
            // Annotate AST_Call, AST_Lambda or AST_New with the special comments
            function annotate(node) {
                var start = node.start;
                var comments = start.comments_before;
                const comments_outside_parens = outer_comments_before_counts.get(start);
                var i = comments_outside_parens != null ? comments_outside_parens : comments.length;
                while (--i >= 0) {
                    var comment = comments[i];
                    if (/[@#]__/.test(comment.value)) {
                        if (/[@#]__PURE__/.test(comment.value)) {
                            set_annotation(node, _PURE);
                            break;
                        }
                        if (/[@#]__INLINE__/.test(comment.value)) {
                            set_annotation(node, _INLINE);
                            break;
                        }
                        if (/[@#]__NOINLINE__/.test(comment.value)) {
                            set_annotation(node, _NOINLINE);
                            break;
                        }
                    }
                }
            }
            var subscripts = function (expr, allow_calls, is_chain) {
                var start = expr.start;
                if (is("punc", ".")) {
                    next();
                    const AST_DotVariant = is("privatename") ? AST_DotHash : AST_Dot;
                    return subscripts(new AST_DotVariant({
                        start: start,
                        expression: expr,
                        optional: false,
                        property: as_name(),
                        end: prev()
                    }), allow_calls, is_chain);
                }
                if (is("punc", "[")) {
                    next();
                    var prop = expression(true);
                    expect("]");
                    return subscripts(new AST_Sub({
                        start: start,
                        expression: expr,
                        optional: false,
                        property: prop,
                        end: prev()
                    }), allow_calls, is_chain);
                }
                if (allow_calls && is("punc", "(")) {
                    next();
                    var call = new AST_Call({
                        start: start,
                        expression: expr,
                        optional: false,
                        args: call_args(),
                        end: prev()
                    });
                    annotate(call);
                    return subscripts(call, true, is_chain);
                }
                if (is("punc", "?.")) {
                    next();
                    let chain_contents;
                    if (allow_calls && is("punc", "(")) {
                        next();
                        const call = new AST_Call({
                            start,
                            optional: true,
                            expression: expr,
                            args: call_args(),
                            end: prev()
                        });
                        annotate(call);
                        chain_contents = subscripts(call, true, true);
                    }
                    else if (is("name") || is("privatename")) {
                        const AST_DotVariant = is("privatename") ? AST_DotHash : AST_Dot;
                        chain_contents = subscripts(new AST_DotVariant({
                            start,
                            expression: expr,
                            optional: true,
                            property: as_name(),
                            end: prev()
                        }), allow_calls, true);
                    }
                    else if (is("punc", "[")) {
                        next();
                        const property = expression(true);
                        expect("]");
                        chain_contents = subscripts(new AST_Sub({
                            start,
                            expression: expr,
                            optional: true,
                            property,
                            end: prev()
                        }), allow_calls, true);
                    }
                    if (!chain_contents)
                        unexpected();
                    if (chain_contents instanceof AST_Chain)
                        return chain_contents;
                    return new AST_Chain({
                        start,
                        expression: chain_contents,
                        end: prev()
                    });
                }
                if (is("template_head")) {
                    if (is_chain) {
                        // a?.b`c` is a syntax error
                        unexpected();
                    }
                    return subscripts(new AST_PrefixedTemplateString({
                        start: start,
                        prefix: expr,
                        template_string: template_string(),
                        end: prev()
                    }), allow_calls);
                }
                return expr;
            };
            function call_args() {
                var args = [];
                while (!is("punc", ")")) {
                    if (is("expand", "...")) {
                        next();
                        args.push(new AST_Expansion({
                            start: prev(),
                            expression: expression(false),
                            end: prev()
                        }));
                    }
                    else {
                        args.push(expression(false));
                    }
                    if (!is("punc", ")")) {
                        expect(",");
                    }
                }
                next();
                return args;
            }
            var maybe_unary = function (allow_calls, allow_arrows) {
                var start = S.token;
                if (start.type == "name" && start.value == "await" && can_await()) {
                    next();
                    return _await_expression();
                }
                if (is("operator") && UNARY_PREFIX.has(start.value)) {
                    next();
                    handle_regexp();
                    var ex = make_unary(AST_UnaryPrefix, start, maybe_unary(allow_calls));
                    ex.start = start;
                    ex.end = prev();
                    return ex;
                }
                var val = expr_atom(allow_calls, allow_arrows);
                while (is("operator") && UNARY_POSTFIX.has(S.token.value) && !has_newline_before(S.token)) {
                    if (val instanceof AST_Arrow)
                        unexpected();
                    val = make_unary(AST_UnaryPostfix, S.token, val);
                    val.start = start;
                    val.end = S.token;
                    next();
                }
                return val;
            };
            function make_unary(ctor, token, expr) {
                var op = token.value;
                switch (op) {
                    case "++":
                    case "--":
                        if (!is_assignable(expr))
                            croak("Invalid use of " + op + " operator", token.line, token.col, token.pos);
                        break;
                    case "delete":
                        if (expr instanceof AST_SymbolRef && S.input.has_directive("use strict"))
                            croak("Calling delete on expression not allowed in strict mode", expr.start.line, expr.start.col, expr.start.pos);
                        break;
                }
                return new ctor({ operator: op, expression: expr });
            }
            var expr_op = function (left, min_prec, no_in) {
                var op = is("operator") ? S.token.value : null;
                if (op == "in" && no_in)
                    op = null;
                if (op == "**" && left instanceof AST_UnaryPrefix
                    /* unary token in front not allowed - parenthesis required */
                    && !is_token(left.start, "punc", "(")
                    && left.operator !== "--" && left.operator !== "++")
                    unexpected(left.start);
                var prec = op != null ? PRECEDENCE[op] : null;
                if (prec != null && (prec > min_prec || (op === "**" && min_prec === prec))) {
                    next();
                    var right = expr_op(maybe_unary(true), prec, no_in);
                    return expr_op(new AST_Binary({
                        start: left.start,
                        left: left,
                        operator: op,
                        right: right,
                        end: right.end
                    }), min_prec, no_in);
                }
                return left;
            };
            function expr_ops(no_in) {
                return expr_op(maybe_unary(true, true), 0, no_in);
            }
            var maybe_conditional = function (no_in) {
                var start = S.token;
                var expr = expr_ops(no_in);
                if (is("operator", "?")) {
                    next();
                    var yes = expression(false);
                    expect(":");
                    return new AST_Conditional({
                        start: start,
                        condition: expr,
                        consequent: yes,
                        alternative: expression(false, no_in),
                        end: prev()
                    });
                }
                return expr;
            };
            function is_assignable(expr) {
                return expr instanceof AST_PropAccess || expr instanceof AST_SymbolRef;
            }
            function to_destructuring(node) {
                if (node instanceof AST_Object) {
                    node = new AST_Destructuring({
                        start: node.start,
                        names: node.properties.map(to_destructuring),
                        is_array: false,
                        end: node.end
                    });
                }
                else if (node instanceof AST_Array) {
                    var names = [];
                    for (var i = 0; i < node.elements.length; i++) {
                        // Only allow expansion as last element
                        if (node.elements[i] instanceof AST_Expansion) {
                            if (i + 1 !== node.elements.length) {
                                token_error(node.elements[i].start, "Spread must the be last element in destructuring array");
                            }
                            node.elements[i].expression = to_destructuring(node.elements[i].expression);
                        }
                        names.push(to_destructuring(node.elements[i]));
                    }
                    node = new AST_Destructuring({
                        start: node.start,
                        names: names,
                        is_array: true,
                        end: node.end
                    });
                }
                else if (node instanceof AST_ObjectProperty) {
                    node.value = to_destructuring(node.value);
                }
                else if (node instanceof AST_Assign) {
                    node = new AST_DefaultAssign({
                        start: node.start,
                        left: node.left,
                        operator: "=",
                        right: node.right,
                        end: node.end
                    });
                }
                return node;
            }
            // In ES6, AssignmentExpression can also be an ArrowFunction
            var maybe_assign = function (no_in) {
                handle_regexp();
                var start = S.token;
                if (start.type == "name" && start.value == "yield") {
                    if (is_in_generator()) {
                        next();
                        return _yield_expression();
                    }
                    else if (S.input.has_directive("use strict")) {
                        token_error(S.token, "Unexpected yield identifier inside strict mode");
                    }
                }
                var left = maybe_conditional(no_in);
                var val = S.token.value;
                if (is("operator") && ASSIGNMENT.has(val)) {
                    if (is_assignable(left) || (left = to_destructuring(left)) instanceof AST_Destructuring) {
                        next();
                        return new AST_Assign({
                            start: start,
                            left: left,
                            operator: val,
                            right: maybe_assign(no_in),
                            logical: LOGICAL_ASSIGNMENT.has(val),
                            end: prev()
                        });
                    }
                    croak("Invalid assignment");
                }
                return left;
            };
            var expression = function (commas, no_in) {
                var start = S.token;
                var exprs = [];
                while (true) {
                    exprs.push(maybe_assign(no_in));
                    if (!commas || !is("punc", ","))
                        break;
                    next();
                    commas = true;
                }
                return exprs.length == 1 ? exprs[0] : new AST_Sequence({
                    start: start,
                    expressions: exprs,
                    end: peek()
                });
            };
            function in_loop(cont) {
                ++S.in_loop;
                var ret = cont();
                --S.in_loop;
                return ret;
            }
            if (options.expression) {
                return expression(true);
            }
            return (function parse_toplevel() {
                var start = S.token;
                var body = [];
                S.input.push_directives_stack();
                if (options.module)
                    S.input.add_directive("use strict");
                while (!is("eof")) {
                    body.push(statement());
                }
                S.input.pop_directives_stack();
                var end = prev();
                var toplevel = options.toplevel;
                if (toplevel) {
                    toplevel.body = toplevel.body.concat(body);
                    toplevel.end = end;
                }
                else {
                    toplevel = new AST_Toplevel({ start: start, body: body, end: end });
                }
                return toplevel;
            })();
        }
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        function DEFNODE(type, props, methods, base = AST_Node) {
            if (!props)
                props = [];
            else
                props = props.split(/\s+/);
            var self_props = props;
            if (base && base.PROPS)
                props = props.concat(base.PROPS);
            var code = "return function AST_" + type + "(props){ if (props) { ";
            for (var i = props.length; --i >= 0;) {
                code += "this." + props[i] + " = props." + props[i] + ";";
            }
            const proto = base && Object.create(base.prototype);
            if (proto && proto.initialize || (methods && methods.initialize))
                code += "this.initialize();";
            code += "}";
            code += "this.flags = 0;";
            code += "}";
            var ctor = new Function(code)();
            if (proto) {
                ctor.prototype = proto;
                ctor.BASE = base;
            }
            if (base)
                base.SUBCLASSES.push(ctor);
            ctor.prototype.CTOR = ctor;
            ctor.prototype.constructor = ctor;
            ctor.PROPS = props || null;
            ctor.SELF_PROPS = self_props;
            ctor.SUBCLASSES = [];
            if (type) {
                ctor.prototype.TYPE = ctor.TYPE = type;
            }
            if (methods)
                for (i in methods)
                    if (HOP(methods, i)) {
                        if (i[0] === "$") {
                            ctor[i.substr(1)] = methods[i];
                        }
                        else {
                            ctor.prototype[i] = methods[i];
                        }
                    }
            ctor.DEFMETHOD = function (name, method) {
                this.prototype[name] = method;
            };
            return ctor;
        }
        const has_tok_flag = (tok, flag) => Boolean(tok.flags & flag);
        const set_tok_flag = (tok, flag, truth) => {
            if (truth) {
                tok.flags |= flag;
            }
            else {
                tok.flags &= ~flag;
            }
        };
        const TOK_FLAG_NLB = 0b0001;
        const TOK_FLAG_QUOTE_SINGLE = 0b0010;
        const TOK_FLAG_QUOTE_EXISTS = 0b0100;
        class AST_Token {
            constructor(type, value, line, col, pos, nlb, comments_before, comments_after, file) {
                this.flags = (nlb ? 1 : 0);
                this.type = type;
                this.value = value;
                this.line = line;
                this.col = col;
                this.pos = pos;
                this.comments_before = comments_before;
                this.comments_after = comments_after;
                this.file = file;
                Object.seal(this);
            }
            get nlb() {
                return has_tok_flag(this, TOK_FLAG_NLB);
            }
            set nlb(new_nlb) {
                set_tok_flag(this, TOK_FLAG_NLB, new_nlb);
            }
            get quote() {
                return !has_tok_flag(this, TOK_FLAG_QUOTE_EXISTS)
                    ? ""
                    : (has_tok_flag(this, TOK_FLAG_QUOTE_SINGLE) ? "'" : '"');
            }
            set quote(quote_type) {
                set_tok_flag(this, TOK_FLAG_QUOTE_SINGLE, quote_type === "'");
                set_tok_flag(this, TOK_FLAG_QUOTE_EXISTS, !!quote_type);
            }
        }
        var AST_Node = DEFNODE("Node", "start end", {
            _clone: function (deep) {
                if (deep) {
                    var self = this.clone();
                    return self.transform(new TreeTransformer(function (node) {
                        if (node !== self) {
                            return node.clone(true);
                        }
                    }));
                }
                return new this.CTOR(this);
            },
            clone: function (deep) {
                return this._clone(deep);
            },
            $documentation: "Base class of all AST nodes",
            $propdoc: {
                start: "[AST_Token] The first token of this node",
                end: "[AST_Token] The last token of this node"
            },
            _walk: function (visitor) {
                return visitor._visit(this);
            },
            walk: function (visitor) {
                return this._walk(visitor); // not sure the indirection will be any help
            },
            _children_backwards: () => { }
        }, null);
        /* -----[ statements ]----- */
        var AST_Statement = DEFNODE("Statement", null, {
            $documentation: "Base class of all statements",
        });
        var AST_Debugger = DEFNODE("Debugger", null, {
            $documentation: "Represents a debugger statement",
        }, AST_Statement);
        var AST_Directive = DEFNODE("Directive", "value quote", {
            $documentation: "Represents a directive, like \"use strict\";",
            $propdoc: {
                value: "[string] The value of this directive as a plain string (it's not an AST_String!)",
                quote: "[string] the original quote character"
            },
        }, AST_Statement);
        var AST_SimpleStatement = DEFNODE("SimpleStatement", "body", {
            $documentation: "A statement consisting of an expression, i.e. a = 1 + 2",
            $propdoc: {
                body: "[AST_Node] an expression node (should not be instanceof AST_Statement)"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.body._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.body);
            }
        }, AST_Statement);
        function walk_body(node, visitor) {
            const body = node.body;
            for (var i = 0, len = body.length; i < len; i++) {
                body[i]._walk(visitor);
            }
        }
        function clone_block_scope(deep) {
            var clone = this._clone(deep);
            if (this.block_scope) {
                clone.block_scope = this.block_scope.clone();
            }
            return clone;
        }
        var AST_Block = DEFNODE("Block", "body block_scope", {
            $documentation: "A body of statements (usually braced)",
            $propdoc: {
                body: "[AST_Statement*] an array of statements",
                block_scope: "[AST_Scope] the block scope"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    walk_body(this, visitor);
                });
            },
            _children_backwards(push) {
                let i = this.body.length;
                while (i--)
                    push(this.body[i]);
            },
            clone: clone_block_scope
        }, AST_Statement);
        var AST_BlockStatement = DEFNODE("BlockStatement", null, {
            $documentation: "A block statement",
        }, AST_Block);
        var AST_EmptyStatement = DEFNODE("EmptyStatement", null, {
            $documentation: "The empty statement (empty block or simply a semicolon)"
        }, AST_Statement);
        var AST_StatementWithBody = DEFNODE("StatementWithBody", "body", {
            $documentation: "Base class for all statements that contain one nested body: `For`, `ForIn`, `Do`, `While`, `With`",
            $propdoc: {
                body: "[AST_Statement] the body; this should always be present, even if it's an AST_EmptyStatement"
            }
        }, AST_Statement);
        var AST_LabeledStatement = DEFNODE("LabeledStatement", "label", {
            $documentation: "Statement with a label",
            $propdoc: {
                label: "[AST_Label] a label definition"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.label._walk(visitor);
                    this.body._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.body);
                push(this.label);
            },
            clone: function (deep) {
                var node = this._clone(deep);
                if (deep) {
                    var label = node.label;
                    var def = this.label;
                    node.walk(new TreeWalker(function (node) {
                        if (node instanceof AST_LoopControl
                            && node.label && node.label.thedef === def) {
                            node.label.thedef = label;
                            label.references.push(node);
                        }
                    }));
                }
                return node;
            }
        }, AST_StatementWithBody);
        var AST_IterationStatement = DEFNODE("IterationStatement", "block_scope", {
            $documentation: "Internal class.  All loops inherit from it.",
            $propdoc: {
                block_scope: "[AST_Scope] the block scope for this iteration statement."
            },
            clone: clone_block_scope
        }, AST_StatementWithBody);
        var AST_DWLoop = DEFNODE("DWLoop", "condition", {
            $documentation: "Base class for do/while statements",
            $propdoc: {
                condition: "[AST_Node] the loop condition.  Should not be instanceof AST_Statement"
            }
        }, AST_IterationStatement);
        var AST_Do = DEFNODE("Do", null, {
            $documentation: "A `do` statement",
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.body._walk(visitor);
                    this.condition._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.condition);
                push(this.body);
            }
        }, AST_DWLoop);
        var AST_While = DEFNODE("While", null, {
            $documentation: "A `while` statement",
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.condition._walk(visitor);
                    this.body._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.body);
                push(this.condition);
            },
        }, AST_DWLoop);
        var AST_For = DEFNODE("For", "init condition step", {
            $documentation: "A `for` statement",
            $propdoc: {
                init: "[AST_Node?] the `for` initialization code, or null if empty",
                condition: "[AST_Node?] the `for` termination clause, or null if empty",
                step: "[AST_Node?] the `for` update clause, or null if empty"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    if (this.init)
                        this.init._walk(visitor);
                    if (this.condition)
                        this.condition._walk(visitor);
                    if (this.step)
                        this.step._walk(visitor);
                    this.body._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.body);
                if (this.step)
                    push(this.step);
                if (this.condition)
                    push(this.condition);
                if (this.init)
                    push(this.init);
            },
        }, AST_IterationStatement);
        var AST_ForIn = DEFNODE("ForIn", "init object", {
            $documentation: "A `for ... in` statement",
            $propdoc: {
                init: "[AST_Node] the `for/in` initialization code",
                object: "[AST_Node] the object that we're looping through"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.init._walk(visitor);
                    this.object._walk(visitor);
                    this.body._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.body);
                if (this.object)
                    push(this.object);
                if (this.init)
                    push(this.init);
            },
        }, AST_IterationStatement);
        var AST_ForOf = DEFNODE("ForOf", "await", {
            $documentation: "A `for ... of` statement",
        }, AST_ForIn);
        var AST_With = DEFNODE("With", "expression", {
            $documentation: "A `with` statement",
            $propdoc: {
                expression: "[AST_Node] the `with` expression"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression._walk(visitor);
                    this.body._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.body);
                push(this.expression);
            },
        }, AST_StatementWithBody);
        /* -----[ scope and functions ]----- */
        var AST_Scope = DEFNODE("Scope", "variables functions uses_with uses_eval parent_scope enclosed cname", {
            $documentation: "Base class for all statements introducing a lexical scope",
            $propdoc: {
                variables: "[Map/S] a map of name -> SymbolDef for all variables/functions defined in this scope",
                uses_with: "[boolean/S] tells whether this scope uses the `with` statement",
                uses_eval: "[boolean/S] tells whether this scope contains a direct call to the global `eval`",
                parent_scope: "[AST_Scope?/S] link to the parent scope",
                enclosed: "[SymbolDef*/S] a list of all symbol definitions that are accessed from this scope or any subscopes",
                cname: "[integer/S] current index for mangling variables (used internally by the mangler)",
            },
            get_defun_scope: function () {
                var self = this;
                while (self.is_block_scope()) {
                    self = self.parent_scope;
                }
                return self;
            },
            clone: function (deep, toplevel) {
                var node = this._clone(deep);
                if (deep && this.variables && toplevel && !this._block_scope) {
                    node.figure_out_scope({}, {
                        toplevel: toplevel,
                        parent_scope: this.parent_scope
                    });
                }
                else {
                    if (this.variables)
                        node.variables = new Map(this.variables);
                    if (this.enclosed)
                        node.enclosed = this.enclosed.slice();
                    if (this._block_scope)
                        node._block_scope = this._block_scope;
                }
                return node;
            },
            pinned: function () {
                return this.uses_eval || this.uses_with;
            }
        }, AST_Block);
        var AST_Toplevel = DEFNODE("Toplevel", "globals", {
            $documentation: "The toplevel scope",
            $propdoc: {
                globals: "[Map/S] a map of name -> SymbolDef for all undeclared names",
            },
            wrap_commonjs: function (name) {
                var body = this.body;
                var wrapped_tl = "(function(exports){'$ORIG';})(typeof " + name + "=='undefined'?(" + name + "={}):" + name + ");";
                wrapped_tl = parse(wrapped_tl);
                wrapped_tl = wrapped_tl.transform(new TreeTransformer(function (node) {
                    if (node instanceof AST_Directive && node.value == "$ORIG") {
                        return MAP.splice(body);
                    }
                }));
                return wrapped_tl;
            },
            wrap_enclose: function (args_values) {
                if (typeof args_values != "string")
                    args_values = "";
                var index = args_values.indexOf(":");
                if (index < 0)
                    index = args_values.length;
                var body = this.body;
                return parse([
                    "(function(",
                    args_values.slice(0, index),
                    '){"$ORIG"})(',
                    args_values.slice(index + 1),
                    ")"
                ].join("")).transform(new TreeTransformer(function (node) {
                    if (node instanceof AST_Directive && node.value == "$ORIG") {
                        return MAP.splice(body);
                    }
                }));
            }
        }, AST_Scope);
        var AST_Expansion = DEFNODE("Expansion", "expression", {
            $documentation: "An expandible argument, such as ...rest, a splat, such as [1,2,...all], or an expansion in a variable declaration, such as var [first, ...rest] = list",
            $propdoc: {
                expression: "[AST_Node] the thing to be expanded"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression.walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.expression);
            },
        });
        var AST_Lambda = DEFNODE("Lambda", "name argnames uses_arguments is_generator async", {
            $documentation: "Base class for functions",
            $propdoc: {
                name: "[AST_SymbolDeclaration?] the name of this function",
                argnames: "[AST_SymbolFunarg|AST_Destructuring|AST_Expansion|AST_DefaultAssign*] array of function arguments, destructurings, or expanding arguments",
                uses_arguments: "[boolean/S] tells whether this function accesses the arguments array",
                is_generator: "[boolean] is this a generator method",
                async: "[boolean] is this method async",
            },
            args_as_names: function () {
                var out = [];
                for (var i = 0; i < this.argnames.length; i++) {
                    if (this.argnames[i] instanceof AST_Destructuring) {
                        out.push(...this.argnames[i].all_symbols());
                    }
                    else {
                        out.push(this.argnames[i]);
                    }
                }
                return out;
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    if (this.name)
                        this.name._walk(visitor);
                    var argnames = this.argnames;
                    for (var i = 0, len = argnames.length; i < len; i++) {
                        argnames[i]._walk(visitor);
                    }
                    walk_body(this, visitor);
                });
            },
            _children_backwards(push) {
                let i = this.body.length;
                while (i--)
                    push(this.body[i]);
                i = this.argnames.length;
                while (i--)
                    push(this.argnames[i]);
                if (this.name)
                    push(this.name);
            },
            is_braceless() {
                return this.body[0] instanceof AST_Return && this.body[0].value;
            },
            // Default args and expansion don't count, so .argnames.length doesn't cut it
            length_property() {
                let length = 0;
                for (const arg of this.argnames) {
                    if (arg instanceof AST_SymbolFunarg || arg instanceof AST_Destructuring) {
                        length++;
                    }
                }
                return length;
            }
        }, AST_Scope);
        var AST_Accessor = DEFNODE("Accessor", null, {
            $documentation: "A setter/getter function.  The `name` property is always null."
        }, AST_Lambda);
        var AST_Function = DEFNODE("Function", null, {
            $documentation: "A function expression"
        }, AST_Lambda);
        var AST_Arrow = DEFNODE("Arrow", null, {
            $documentation: "An ES6 Arrow function ((a) => b)"
        }, AST_Lambda);
        var AST_Defun = DEFNODE("Defun", null, {
            $documentation: "A function definition"
        }, AST_Lambda);
        /* -----[ DESTRUCTURING ]----- */
        var AST_Destructuring = DEFNODE("Destructuring", "names is_array", {
            $documentation: "A destructuring of several names. Used in destructuring assignment and with destructuring function argument names",
            $propdoc: {
                "names": "[AST_Node*] Array of properties or elements",
                "is_array": "[Boolean] Whether the destructuring represents an object or array"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.names.forEach(function (name) {
                        name._walk(visitor);
                    });
                });
            },
            _children_backwards(push) {
                let i = this.names.length;
                while (i--)
                    push(this.names[i]);
            },
            all_symbols: function () {
                var out = [];
                this.walk(new TreeWalker(function (node) {
                    if (node instanceof AST_Symbol) {
                        out.push(node);
                    }
                }));
                return out;
            }
        });
        var AST_PrefixedTemplateString = DEFNODE("PrefixedTemplateString", "template_string prefix", {
            $documentation: "A templatestring with a prefix, such as String.raw`foobarbaz`",
            $propdoc: {
                template_string: "[AST_TemplateString] The template string",
                prefix: "[AST_Node] The prefix, which will get called."
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.prefix._walk(visitor);
                    this.template_string._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.template_string);
                push(this.prefix);
            },
        });
        var AST_TemplateString = DEFNODE("TemplateString", "segments", {
            $documentation: "A template string literal",
            $propdoc: {
                segments: "[AST_Node*] One or more segments, starting with AST_TemplateSegment. AST_Node may follow AST_TemplateSegment, but each AST_Node must be followed by AST_TemplateSegment."
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.segments.forEach(function (seg) {
                        seg._walk(visitor);
                    });
                });
            },
            _children_backwards(push) {
                let i = this.segments.length;
                while (i--)
                    push(this.segments[i]);
            }
        });
        var AST_TemplateSegment = DEFNODE("TemplateSegment", "value raw", {
            $documentation: "A segment of a template string literal",
            $propdoc: {
                value: "Content of the segment",
                raw: "Raw source of the segment",
            }
        });
        /* -----[ JUMPS ]----- */
        var AST_Jump = DEFNODE("Jump", null, {
            $documentation: "Base class for “jumps” (for now that's `return`, `throw`, `break` and `continue`)"
        }, AST_Statement);
        var AST_Exit = DEFNODE("Exit", "value", {
            $documentation: "Base class for “exits” (`return` and `throw`)",
            $propdoc: {
                value: "[AST_Node?] the value returned or thrown by this statement; could be null for AST_Return"
            },
            _walk: function (visitor) {
                return visitor._visit(this, this.value && function () {
                    this.value._walk(visitor);
                });
            },
            _children_backwards(push) {
                if (this.value)
                    push(this.value);
            },
        }, AST_Jump);
        var AST_Return = DEFNODE("Return", null, {
            $documentation: "A `return` statement"
        }, AST_Exit);
        var AST_Throw = DEFNODE("Throw", null, {
            $documentation: "A `throw` statement"
        }, AST_Exit);
        var AST_LoopControl = DEFNODE("LoopControl", "label", {
            $documentation: "Base class for loop control statements (`break` and `continue`)",
            $propdoc: {
                label: "[AST_LabelRef?] the label, or null if none",
            },
            _walk: function (visitor) {
                return visitor._visit(this, this.label && function () {
                    this.label._walk(visitor);
                });
            },
            _children_backwards(push) {
                if (this.label)
                    push(this.label);
            },
        }, AST_Jump);
        var AST_Break = DEFNODE("Break", null, {
            $documentation: "A `break` statement"
        }, AST_LoopControl);
        var AST_Continue = DEFNODE("Continue", null, {
            $documentation: "A `continue` statement"
        }, AST_LoopControl);
        var AST_Await = DEFNODE("Await", "expression", {
            $documentation: "An `await` statement",
            $propdoc: {
                expression: "[AST_Node] the mandatory expression being awaited",
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.expression);
            },
        });
        var AST_Yield = DEFNODE("Yield", "expression is_star", {
            $documentation: "A `yield` statement",
            $propdoc: {
                expression: "[AST_Node?] the value returned or thrown by this statement; could be null (representing undefined) but only when is_star is set to false",
                is_star: "[Boolean] Whether this is a yield or yield* statement"
            },
            _walk: function (visitor) {
                return visitor._visit(this, this.expression && function () {
                    this.expression._walk(visitor);
                });
            },
            _children_backwards(push) {
                if (this.expression)
                    push(this.expression);
            }
        });
        /* -----[ IF ]----- */
        var AST_If = DEFNODE("If", "condition alternative", {
            $documentation: "A `if` statement",
            $propdoc: {
                condition: "[AST_Node] the `if` condition",
                alternative: "[AST_Statement?] the `else` part, or null if not present"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.condition._walk(visitor);
                    this.body._walk(visitor);
                    if (this.alternative)
                        this.alternative._walk(visitor);
                });
            },
            _children_backwards(push) {
                if (this.alternative) {
                    push(this.alternative);
                }
                push(this.body);
                push(this.condition);
            }
        }, AST_StatementWithBody);
        /* -----[ SWITCH ]----- */
        var AST_Switch = DEFNODE("Switch", "expression", {
            $documentation: "A `switch` statement",
            $propdoc: {
                expression: "[AST_Node] the `switch` “discriminant”"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression._walk(visitor);
                    walk_body(this, visitor);
                });
            },
            _children_backwards(push) {
                let i = this.body.length;
                while (i--)
                    push(this.body[i]);
                push(this.expression);
            }
        }, AST_Block);
        var AST_SwitchBranch = DEFNODE("SwitchBranch", null, {
            $documentation: "Base class for `switch` branches",
        }, AST_Block);
        var AST_Default = DEFNODE("Default", null, {
            $documentation: "A `default` switch branch",
        }, AST_SwitchBranch);
        var AST_Case = DEFNODE("Case", "expression", {
            $documentation: "A `case` switch branch",
            $propdoc: {
                expression: "[AST_Node] the `case` expression"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression._walk(visitor);
                    walk_body(this, visitor);
                });
            },
            _children_backwards(push) {
                let i = this.body.length;
                while (i--)
                    push(this.body[i]);
                push(this.expression);
            },
        }, AST_SwitchBranch);
        /* -----[ EXCEPTIONS ]----- */
        var AST_Try = DEFNODE("Try", "bcatch bfinally", {
            $documentation: "A `try` statement",
            $propdoc: {
                bcatch: "[AST_Catch?] the catch block, or null if not present",
                bfinally: "[AST_Finally?] the finally block, or null if not present"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    walk_body(this, visitor);
                    if (this.bcatch)
                        this.bcatch._walk(visitor);
                    if (this.bfinally)
                        this.bfinally._walk(visitor);
                });
            },
            _children_backwards(push) {
                if (this.bfinally)
                    push(this.bfinally);
                if (this.bcatch)
                    push(this.bcatch);
                let i = this.body.length;
                while (i--)
                    push(this.body[i]);
            },
        }, AST_Block);
        var AST_Catch = DEFNODE("Catch", "argname", {
            $documentation: "A `catch` node; only makes sense as part of a `try` statement",
            $propdoc: {
                argname: "[AST_SymbolCatch|AST_Destructuring|AST_Expansion|AST_DefaultAssign] symbol for the exception"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    if (this.argname)
                        this.argname._walk(visitor);
                    walk_body(this, visitor);
                });
            },
            _children_backwards(push) {
                let i = this.body.length;
                while (i--)
                    push(this.body[i]);
                if (this.argname)
                    push(this.argname);
            },
        }, AST_Block);
        var AST_Finally = DEFNODE("Finally", null, {
            $documentation: "A `finally` node; only makes sense as part of a `try` statement"
        }, AST_Block);
        /* -----[ VAR/CONST ]----- */
        var AST_Definitions = DEFNODE("Definitions", "definitions", {
            $documentation: "Base class for `var` or `const` nodes (variable declarations/initializations)",
            $propdoc: {
                definitions: "[AST_VarDef*] array of variable definitions"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    var definitions = this.definitions;
                    for (var i = 0, len = definitions.length; i < len; i++) {
                        definitions[i]._walk(visitor);
                    }
                });
            },
            _children_backwards(push) {
                let i = this.definitions.length;
                while (i--)
                    push(this.definitions[i]);
            },
        }, AST_Statement);
        var AST_Var = DEFNODE("Var", null, {
            $documentation: "A `var` statement"
        }, AST_Definitions);
        var AST_Let = DEFNODE("Let", null, {
            $documentation: "A `let` statement"
        }, AST_Definitions);
        var AST_Const = DEFNODE("Const", null, {
            $documentation: "A `const` statement"
        }, AST_Definitions);
        var AST_VarDef = DEFNODE("VarDef", "name value", {
            $documentation: "A variable declaration; only appears in a AST_Definitions node",
            $propdoc: {
                name: "[AST_Destructuring|AST_SymbolConst|AST_SymbolLet|AST_SymbolVar] name of the variable",
                value: "[AST_Node?] initializer, or null of there's no initializer"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.name._walk(visitor);
                    if (this.value)
                        this.value._walk(visitor);
                });
            },
            _children_backwards(push) {
                if (this.value)
                    push(this.value);
                push(this.name);
            },
        });
        var AST_NameMapping = DEFNODE("NameMapping", "foreign_name name", {
            $documentation: "The part of the export/import statement that declare names from a module.",
            $propdoc: {
                foreign_name: "[AST_SymbolExportForeign|AST_SymbolImportForeign] The name being exported/imported (as specified in the module)",
                name: "[AST_SymbolExport|AST_SymbolImport] The name as it is visible to this module."
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.foreign_name._walk(visitor);
                    this.name._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.name);
                push(this.foreign_name);
            },
        });
        var AST_Import = DEFNODE("Import", "imported_name imported_names module_name", {
            $documentation: "An `import` statement",
            $propdoc: {
                imported_name: "[AST_SymbolImport] The name of the variable holding the module's default export.",
                imported_names: "[AST_NameMapping*] The names of non-default imported variables",
                module_name: "[AST_String] String literal describing where this module came from",
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    if (this.imported_name) {
                        this.imported_name._walk(visitor);
                    }
                    if (this.imported_names) {
                        this.imported_names.forEach(function (name_import) {
                            name_import._walk(visitor);
                        });
                    }
                    this.module_name._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.module_name);
                if (this.imported_names) {
                    let i = this.imported_names.length;
                    while (i--)
                        push(this.imported_names[i]);
                }
                if (this.imported_name)
                    push(this.imported_name);
            },
        });
        var AST_ImportMeta = DEFNODE("ImportMeta", null, {
            $documentation: "A reference to import.meta",
        });
        var AST_Export = DEFNODE("Export", "exported_definition exported_value is_default exported_names module_name", {
            $documentation: "An `export` statement",
            $propdoc: {
                exported_definition: "[AST_Defun|AST_Definitions|AST_DefClass?] An exported definition",
                exported_value: "[AST_Node?] An exported value",
                exported_names: "[AST_NameMapping*?] List of exported names",
                module_name: "[AST_String?] Name of the file to load exports from",
                is_default: "[Boolean] Whether this is the default exported value of this module"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    if (this.exported_definition) {
                        this.exported_definition._walk(visitor);
                    }
                    if (this.exported_value) {
                        this.exported_value._walk(visitor);
                    }
                    if (this.exported_names) {
                        this.exported_names.forEach(function (name_export) {
                            name_export._walk(visitor);
                        });
                    }
                    if (this.module_name) {
                        this.module_name._walk(visitor);
                    }
                });
            },
            _children_backwards(push) {
                if (this.module_name)
                    push(this.module_name);
                if (this.exported_names) {
                    let i = this.exported_names.length;
                    while (i--)
                        push(this.exported_names[i]);
                }
                if (this.exported_value)
                    push(this.exported_value);
                if (this.exported_definition)
                    push(this.exported_definition);
            }
        }, AST_Statement);
        /* -----[ OTHER ]----- */
        var AST_Call = DEFNODE("Call", "expression args optional _annotations", {
            $documentation: "A function call expression",
            $propdoc: {
                expression: "[AST_Node] expression to invoke as function",
                args: "[AST_Node*] array of arguments",
                optional: "[boolean] whether this is an optional call (IE ?.() )",
                _annotations: "[number] bitfield containing information about the call"
            },
            initialize() {
                if (this._annotations == null)
                    this._annotations = 0;
            },
            _walk(visitor) {
                return visitor._visit(this, function () {
                    var args = this.args;
                    for (var i = 0, len = args.length; i < len; i++) {
                        args[i]._walk(visitor);
                    }
                    this.expression._walk(visitor); // TODO why do we need to crawl this last?
                });
            },
            _children_backwards(push) {
                let i = this.args.length;
                while (i--)
                    push(this.args[i]);
                push(this.expression);
            },
        });
        var AST_New = DEFNODE("New", null, {
            $documentation: "An object instantiation.  Derives from a function call since it has exactly the same properties"
        }, AST_Call);
        var AST_Sequence = DEFNODE("Sequence", "expressions", {
            $documentation: "A sequence expression (comma-separated expressions)",
            $propdoc: {
                expressions: "[AST_Node*] array of expressions (at least two)"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expressions.forEach(function (node) {
                        node._walk(visitor);
                    });
                });
            },
            _children_backwards(push) {
                let i = this.expressions.length;
                while (i--)
                    push(this.expressions[i]);
            },
        });
        var AST_PropAccess = DEFNODE("PropAccess", "expression property optional", {
            $documentation: "Base class for property access expressions, i.e. `a.foo` or `a[\"foo\"]`",
            $propdoc: {
                expression: "[AST_Node] the “container” expression",
                property: "[AST_Node|string] the property to access.  For AST_Dot & AST_DotHash this is always a plain string, while for AST_Sub it's an arbitrary AST_Node",
                optional: "[boolean] whether this is an optional property access (IE ?.)"
            }
        });
        var AST_Dot = DEFNODE("Dot", "quote", {
            $documentation: "A dotted property access expression",
            $propdoc: {
                quote: "[string] the original quote character when transformed from AST_Sub",
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.expression);
            },
        }, AST_PropAccess);
        var AST_DotHash = DEFNODE("DotHash", "", {
            $documentation: "A dotted property access to a private property",
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.expression);
            },
        }, AST_PropAccess);
        var AST_Sub = DEFNODE("Sub", null, {
            $documentation: "Index-style property access, i.e. `a[\"foo\"]`",
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression._walk(visitor);
                    this.property._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.property);
                push(this.expression);
            },
        }, AST_PropAccess);
        var AST_Chain = DEFNODE("Chain", "expression", {
            $documentation: "A chain expression like a?.b?.(c)?.[d]",
            $propdoc: {
                expression: "[AST_Call|AST_Dot|AST_DotHash|AST_Sub] chain element."
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.expression);
            },
        });
        var AST_Unary = DEFNODE("Unary", "operator expression", {
            $documentation: "Base class for unary expressions",
            $propdoc: {
                operator: "[string] the operator",
                expression: "[AST_Node] expression that this unary operator applies to"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.expression._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.expression);
            },
        });
        var AST_UnaryPrefix = DEFNODE("UnaryPrefix", null, {
            $documentation: "Unary prefix expression, i.e. `typeof i` or `++i`"
        }, AST_Unary);
        var AST_UnaryPostfix = DEFNODE("UnaryPostfix", null, {
            $documentation: "Unary postfix expression, i.e. `i++`"
        }, AST_Unary);
        var AST_Binary = DEFNODE("Binary", "operator left right", {
            $documentation: "Binary expression, i.e. `a + b`",
            $propdoc: {
                left: "[AST_Node] left-hand side expression",
                operator: "[string] the operator",
                right: "[AST_Node] right-hand side expression"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.left._walk(visitor);
                    this.right._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.right);
                push(this.left);
            },
        });
        var AST_Conditional = DEFNODE("Conditional", "condition consequent alternative", {
            $documentation: "Conditional expression using the ternary operator, i.e. `a ? b : c`",
            $propdoc: {
                condition: "[AST_Node]",
                consequent: "[AST_Node]",
                alternative: "[AST_Node]"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    this.condition._walk(visitor);
                    this.consequent._walk(visitor);
                    this.alternative._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.alternative);
                push(this.consequent);
                push(this.condition);
            },
        });
        var AST_Assign = DEFNODE("Assign", "logical", {
            $documentation: "An assignment expression — `a = b + 5`",
            $propdoc: {
                logical: "Whether it's a logical assignment"
            }
        }, AST_Binary);
        var AST_DefaultAssign = DEFNODE("DefaultAssign", null, {
            $documentation: "A default assignment expression like in `(a = 3) => a`"
        }, AST_Binary);
        /* -----[ LITERALS ]----- */
        var AST_Array = DEFNODE("Array", "elements", {
            $documentation: "An array literal",
            $propdoc: {
                elements: "[AST_Node*] array of elements"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    var elements = this.elements;
                    for (var i = 0, len = elements.length; i < len; i++) {
                        elements[i]._walk(visitor);
                    }
                });
            },
            _children_backwards(push) {
                let i = this.elements.length;
                while (i--)
                    push(this.elements[i]);
            },
        });
        var AST_Object = DEFNODE("Object", "properties", {
            $documentation: "An object literal",
            $propdoc: {
                properties: "[AST_ObjectProperty*] array of properties"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    var properties = this.properties;
                    for (var i = 0, len = properties.length; i < len; i++) {
                        properties[i]._walk(visitor);
                    }
                });
            },
            _children_backwards(push) {
                let i = this.properties.length;
                while (i--)
                    push(this.properties[i]);
            },
        });
        var AST_ObjectProperty = DEFNODE("ObjectProperty", "key value", {
            $documentation: "Base class for literal object properties",
            $propdoc: {
                key: "[string|AST_Node] property name. For ObjectKeyVal this is a string. For getters, setters and computed property this is an AST_Node.",
                value: "[AST_Node] property value.  For getters and setters this is an AST_Accessor."
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    if (this.key instanceof AST_Node)
                        this.key._walk(visitor);
                    this.value._walk(visitor);
                });
            },
            _children_backwards(push) {
                push(this.value);
                if (this.key instanceof AST_Node)
                    push(this.key);
            }
        });
        var AST_ObjectKeyVal = DEFNODE("ObjectKeyVal", "quote", {
            $documentation: "A key: value object property",
            $propdoc: {
                quote: "[string] the original quote character"
            },
            computed_key() {
                return this.key instanceof AST_Node;
            }
        }, AST_ObjectProperty);
        var AST_PrivateSetter = DEFNODE("PrivateSetter", "static", {
            $propdoc: {
                static: "[boolean] whether this is a static private setter"
            },
            $documentation: "A private setter property",
            computed_key() {
                return false;
            }
        }, AST_ObjectProperty);
        var AST_PrivateGetter = DEFNODE("PrivateGetter", "static", {
            $propdoc: {
                static: "[boolean] whether this is a static private getter"
            },
            $documentation: "A private getter property",
            computed_key() {
                return false;
            }
        }, AST_ObjectProperty);
        var AST_ObjectSetter = DEFNODE("ObjectSetter", "quote static", {
            $propdoc: {
                quote: "[string|undefined] the original quote character, if any",
                static: "[boolean] whether this is a static setter (classes only)"
            },
            $documentation: "An object setter property",
            computed_key() {
                return !(this.key instanceof AST_SymbolMethod);
            }
        }, AST_ObjectProperty);
        var AST_ObjectGetter = DEFNODE("ObjectGetter", "quote static", {
            $propdoc: {
                quote: "[string|undefined] the original quote character, if any",
                static: "[boolean] whether this is a static getter (classes only)"
            },
            $documentation: "An object getter property",
            computed_key() {
                return !(this.key instanceof AST_SymbolMethod);
            }
        }, AST_ObjectProperty);
        var AST_ConciseMethod = DEFNODE("ConciseMethod", "quote static is_generator async", {
            $propdoc: {
                quote: "[string|undefined] the original quote character, if any",
                static: "[boolean] is this method static (classes only)",
                is_generator: "[boolean] is this a generator method",
                async: "[boolean] is this method async",
            },
            $documentation: "An ES6 concise method inside an object or class",
            computed_key() {
                return !(this.key instanceof AST_SymbolMethod);
            }
        }, AST_ObjectProperty);
        var AST_PrivateMethod = DEFNODE("PrivateMethod", "", {
            $documentation: "A private class method inside a class",
        }, AST_ConciseMethod);
        var AST_Class = DEFNODE("Class", "name extends properties", {
            $propdoc: {
                name: "[AST_SymbolClass|AST_SymbolDefClass?] optional class name.",
                extends: "[AST_Node]? optional parent class",
                properties: "[AST_ObjectProperty*] array of properties"
            },
            $documentation: "An ES6 class",
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    if (this.name) {
                        this.name._walk(visitor);
                    }
                    if (this.extends) {
                        this.extends._walk(visitor);
                    }
                    this.properties.forEach((prop) => prop._walk(visitor));
                });
            },
            _children_backwards(push) {
                let i = this.properties.length;
                while (i--)
                    push(this.properties[i]);
                if (this.extends)
                    push(this.extends);
                if (this.name)
                    push(this.name);
            },
        }, AST_Scope /* TODO a class might have a scope but it's not a scope */);
        var AST_ClassProperty = DEFNODE("ClassProperty", "static quote", {
            $documentation: "A class property",
            $propdoc: {
                static: "[boolean] whether this is a static key",
                quote: "[string] which quote is being used"
            },
            _walk: function (visitor) {
                return visitor._visit(this, function () {
                    if (this.key instanceof AST_Node)
                        this.key._walk(visitor);
                    if (this.value instanceof AST_Node)
                        this.value._walk(visitor);
                });
            },
            _children_backwards(push) {
                if (this.value instanceof AST_Node)
                    push(this.value);
                if (this.key instanceof AST_Node)
                    push(this.key);
            },
            computed_key() {
                return !(this.key instanceof AST_SymbolClassProperty);
            }
        }, AST_ObjectProperty);
        var AST_ClassPrivateProperty = DEFNODE("ClassProperty", "", {
            $documentation: "A class property for a private property",
        }, AST_ClassProperty);
        var AST_DefClass = DEFNODE("DefClass", null, {
            $documentation: "A class definition",
        }, AST_Class);
        var AST_ClassExpression = DEFNODE("ClassExpression", null, {
            $documentation: "A class expression."
        }, AST_Class);
        var AST_Symbol = DEFNODE("Symbol", "scope name thedef", {
            $propdoc: {
                name: "[string] name of this symbol",
                scope: "[AST_Scope/S] the current scope (not necessarily the definition scope)",
                thedef: "[SymbolDef/S] the definition of this symbol"
            },
            $documentation: "Base class for all symbols"
        });
        var AST_NewTarget = DEFNODE("NewTarget", null, {
            $documentation: "A reference to new.target"
        });
        var AST_SymbolDeclaration = DEFNODE("SymbolDeclaration", "init", {
            $documentation: "A declaration symbol (symbol in var/const, function name or argument, symbol in catch)",
        }, AST_Symbol);
        var AST_SymbolVar = DEFNODE("SymbolVar", null, {
            $documentation: "Symbol defining a variable",
        }, AST_SymbolDeclaration);
        var AST_SymbolBlockDeclaration = DEFNODE("SymbolBlockDeclaration", null, {
            $documentation: "Base class for block-scoped declaration symbols"
        }, AST_SymbolDeclaration);
        var AST_SymbolConst = DEFNODE("SymbolConst", null, {
            $documentation: "A constant declaration"
        }, AST_SymbolBlockDeclaration);
        var AST_SymbolLet = DEFNODE("SymbolLet", null, {
            $documentation: "A block-scoped `let` declaration"
        }, AST_SymbolBlockDeclaration);
        var AST_SymbolFunarg = DEFNODE("SymbolFunarg", null, {
            $documentation: "Symbol naming a function argument",
        }, AST_SymbolVar);
        var AST_SymbolDefun = DEFNODE("SymbolDefun", null, {
            $documentation: "Symbol defining a function",
        }, AST_SymbolDeclaration);
        var AST_SymbolMethod = DEFNODE("SymbolMethod", null, {
            $documentation: "Symbol in an object defining a method",
        }, AST_Symbol);
        var AST_SymbolClassProperty = DEFNODE("SymbolClassProperty", null, {
            $documentation: "Symbol for a class property",
        }, AST_Symbol);
        var AST_SymbolLambda = DEFNODE("SymbolLambda", null, {
            $documentation: "Symbol naming a function expression",
        }, AST_SymbolDeclaration);
        var AST_SymbolDefClass = DEFNODE("SymbolDefClass", null, {
            $documentation: "Symbol naming a class's name in a class declaration. Lexically scoped to its containing scope, and accessible within the class."
        }, AST_SymbolBlockDeclaration);
        var AST_SymbolClass = DEFNODE("SymbolClass", null, {
            $documentation: "Symbol naming a class's name. Lexically scoped to the class."
        }, AST_SymbolDeclaration);
        var AST_SymbolCatch = DEFNODE("SymbolCatch", null, {
            $documentation: "Symbol naming the exception in catch",
        }, AST_SymbolBlockDeclaration);
        var AST_SymbolImport = DEFNODE("SymbolImport", null, {
            $documentation: "Symbol referring to an imported name",
        }, AST_SymbolBlockDeclaration);
        var AST_SymbolImportForeign = DEFNODE("SymbolImportForeign", null, {
            $documentation: "A symbol imported from a module, but it is defined in the other module, and its real name is irrelevant for this module's purposes",
        }, AST_Symbol);
        var AST_Label = DEFNODE("Label", "references", {
            $documentation: "Symbol naming a label (declaration)",
            $propdoc: {
                references: "[AST_LoopControl*] a list of nodes referring to this label"
            },
            initialize: function () {
                this.references = [];
                this.thedef = this;
            }
        }, AST_Symbol);
        var AST_SymbolRef = DEFNODE("SymbolRef", null, {
            $documentation: "Reference to some symbol (not definition/declaration)",
        }, AST_Symbol);
        var AST_SymbolExport = DEFNODE("SymbolExport", null, {
            $documentation: "Symbol referring to a name to export",
        }, AST_SymbolRef);
        var AST_SymbolExportForeign = DEFNODE("SymbolExportForeign", null, {
            $documentation: "A symbol exported from this module, but it is used in the other module, and its real name is irrelevant for this module's purposes",
        }, AST_Symbol);
        var AST_LabelRef = DEFNODE("LabelRef", null, {
            $documentation: "Reference to a label symbol",
        }, AST_Symbol);
        var AST_This = DEFNODE("This", null, {
            $documentation: "The `this` symbol",
        }, AST_Symbol);
        var AST_Super = DEFNODE("Super", null, {
            $documentation: "The `super` symbol",
        }, AST_This);
        var AST_Constant = DEFNODE("Constant", null, {
            $documentation: "Base class for all constants",
            getValue: function () {
                return this.value;
            }
        });
        var AST_String = DEFNODE("String", "value quote", {
            $documentation: "A string literal",
            $propdoc: {
                value: "[string] the contents of this string",
                quote: "[string] the original quote character"
            }
        }, AST_Constant);
        var AST_Number = DEFNODE("Number", "value raw", {
            $documentation: "A number literal",
            $propdoc: {
                value: "[number] the numeric value",
                raw: "[string] numeric value as string"
            }
        }, AST_Constant);
        var AST_BigInt = DEFNODE("BigInt", "value", {
            $documentation: "A big int literal",
            $propdoc: {
                value: "[string] big int value"
            }
        }, AST_Constant);
        var AST_RegExp = DEFNODE("RegExp", "value", {
            $documentation: "A regexp literal",
            $propdoc: {
                value: "[RegExp] the actual regexp",
            }
        }, AST_Constant);
        var AST_Atom = DEFNODE("Atom", null, {
            $documentation: "Base class for atoms",
        }, AST_Constant);
        var AST_Null = DEFNODE("Null", null, {
            $documentation: "The `null` atom",
            value: null
        }, AST_Atom);
        var AST_NaN = DEFNODE("NaN", null, {
            $documentation: "The impossible value",
            value: 0 / 0
        }, AST_Atom);
        var AST_Undefined = DEFNODE("Undefined", null, {
            $documentation: "The `undefined` value",
            value: (function () { }())
        }, AST_Atom);
        var AST_Hole = DEFNODE("Hole", null, {
            $documentation: "A hole in an array",
            value: (function () { }())
        }, AST_Atom);
        var AST_Infinity = DEFNODE("Infinity", null, {
            $documentation: "The `Infinity` value",
            value: 1 / 0
        }, AST_Atom);
        var AST_Boolean = DEFNODE("Boolean", null, {
            $documentation: "Base class for booleans",
        }, AST_Atom);
        var AST_False = DEFNODE("False", null, {
            $documentation: "The `false` atom",
            value: false
        }, AST_Boolean);
        var AST_True = DEFNODE("True", null, {
            $documentation: "The `true` atom",
            value: true
        }, AST_Boolean);
        /* -----[ Walk function ]---- */
        /**
         * Walk nodes in depth-first search fashion.
         * Callback can return `walk_abort` symbol to stop iteration.
         * It can also return `true` to stop iteration just for child nodes.
         * Iteration can be stopped and continued by passing the `to_visit` argument,
         * which is given to the callback in the second argument.
         **/
        function walk(node, cb, to_visit = [node]) {
            const push = to_visit.push.bind(to_visit);
            while (to_visit.length) {
                const node = to_visit.pop();
                const ret = cb(node, to_visit);
                if (ret) {
                    if (ret === walk_abort)
                        return true;
                    continue;
                }
                node._children_backwards(push);
            }
            return false;
        }
        function walk_parent(node, cb, initial_stack) {
            const to_visit = [node];
            const push = to_visit.push.bind(to_visit);
            const stack = initial_stack ? initial_stack.slice() : [];
            const parent_pop_indices = [];
            let current;
            const info = {
                parent: (n = 0) => {
                    if (n === -1) {
                        return current;
                    }
                    // [ p1 p0 ] [ 1 0 ]
                    if (initial_stack && n >= stack.length) {
                        n -= stack.length;
                        return initial_stack[initial_stack.length - (n + 1)];
                    }
                    return stack[stack.length - (1 + n)];
                },
            };
            while (to_visit.length) {
                current = to_visit.pop();
                while (parent_pop_indices.length &&
                    to_visit.length == parent_pop_indices[parent_pop_indices.length - 1]) {
                    stack.pop();
                    parent_pop_indices.pop();
                }
                const ret = cb(current, info);
                if (ret) {
                    if (ret === walk_abort)
                        return true;
                    continue;
                }
                const visit_length = to_visit.length;
                current._children_backwards(push);
                // Push only if we're going to traverse the children
                if (to_visit.length > visit_length) {
                    stack.push(current);
                    parent_pop_indices.push(visit_length - 1);
                }
            }
            return false;
        }
        const walk_abort = Symbol("abort walk");
        /* -----[ TreeWalker ]----- */
        class TreeWalker {
            constructor(callback) {
                this.visit = callback;
                this.stack = [];
                this.directives = Object.create(null);
            }
            _visit(node, descend) {
                this.push(node);
                var ret = this.visit(node, descend ? function () {
                    descend.call(node);
                } : noop);
                if (!ret && descend) {
                    descend.call(node);
                }
                this.pop();
                return ret;
            }
            parent(n) {
                return this.stack[this.stack.length - 2 - (n || 0)];
            }
            push(node) {
                if (node instanceof AST_Lambda) {
                    this.directives = Object.create(this.directives);
                }
                else if (node instanceof AST_Directive && !this.directives[node.value]) {
                    this.directives[node.value] = node;
                }
                else if (node instanceof AST_Class) {
                    this.directives = Object.create(this.directives);
                    if (!this.directives["use strict"]) {
                        this.directives["use strict"] = node;
                    }
                }
                this.stack.push(node);
            }
            pop() {
                var node = this.stack.pop();
                if (node instanceof AST_Lambda || node instanceof AST_Class) {
                    this.directives = Object.getPrototypeOf(this.directives);
                }
            }
            self() {
                return this.stack[this.stack.length - 1];
            }
            find_parent(type) {
                var stack = this.stack;
                for (var i = stack.length; --i >= 0;) {
                    var x = stack[i];
                    if (x instanceof type)
                        return x;
                }
            }
            has_directive(type) {
                var dir = this.directives[type];
                if (dir)
                    return dir;
                var node = this.stack[this.stack.length - 1];
                if (node instanceof AST_Scope && node.body) {
                    for (var i = 0; i < node.body.length; ++i) {
                        var st = node.body[i];
                        if (!(st instanceof AST_Directive))
                            break;
                        if (st.value == type)
                            return st;
                    }
                }
            }
            loopcontrol_target(node) {
                var stack = this.stack;
                if (node.label)
                    for (var i = stack.length; --i >= 0;) {
                        var x = stack[i];
                        if (x instanceof AST_LabeledStatement && x.label.name == node.label.name)
                            return x.body;
                    }
                else
                    for (var i = stack.length; --i >= 0;) {
                        var x = stack[i];
                        if (x instanceof AST_IterationStatement
                            || node instanceof AST_Break && x instanceof AST_Switch)
                            return x;
                    }
            }
        }
        // Tree transformer helpers.
        class TreeTransformer extends TreeWalker {
            constructor(before, after) {
                super();
                this.before = before;
                this.after = after;
            }
        }
        const _PURE = 0b00000001;
        const _INLINE = 0b00000010;
        const _NOINLINE = 0b00000100;
        var ast = /*#__PURE__*/ Object.freeze({
            __proto__: null,
            AST_Accessor: AST_Accessor,
            AST_Array: AST_Array,
            AST_Arrow: AST_Arrow,
            AST_Assign: AST_Assign,
            AST_Atom: AST_Atom,
            AST_Await: AST_Await,
            AST_BigInt: AST_BigInt,
            AST_Binary: AST_Binary,
            AST_Block: AST_Block,
            AST_BlockStatement: AST_BlockStatement,
            AST_Boolean: AST_Boolean,
            AST_Break: AST_Break,
            AST_Call: AST_Call,
            AST_Case: AST_Case,
            AST_Catch: AST_Catch,
            AST_Chain: AST_Chain,
            AST_Class: AST_Class,
            AST_ClassExpression: AST_ClassExpression,
            AST_ClassPrivateProperty: AST_ClassPrivateProperty,
            AST_ClassProperty: AST_ClassProperty,
            AST_ConciseMethod: AST_ConciseMethod,
            AST_Conditional: AST_Conditional,
            AST_Const: AST_Const,
            AST_Constant: AST_Constant,
            AST_Continue: AST_Continue,
            AST_Debugger: AST_Debugger,
            AST_Default: AST_Default,
            AST_DefaultAssign: AST_DefaultAssign,
            AST_DefClass: AST_DefClass,
            AST_Definitions: AST_Definitions,
            AST_Defun: AST_Defun,
            AST_Destructuring: AST_Destructuring,
            AST_Directive: AST_Directive,
            AST_Do: AST_Do,
            AST_Dot: AST_Dot,
            AST_DotHash: AST_DotHash,
            AST_DWLoop: AST_DWLoop,
            AST_EmptyStatement: AST_EmptyStatement,
            AST_Exit: AST_Exit,
            AST_Expansion: AST_Expansion,
            AST_Export: AST_Export,
            AST_False: AST_False,
            AST_Finally: AST_Finally,
            AST_For: AST_For,
            AST_ForIn: AST_ForIn,
            AST_ForOf: AST_ForOf,
            AST_Function: AST_Function,
            AST_Hole: AST_Hole,
            AST_If: AST_If,
            AST_Import: AST_Import,
            AST_ImportMeta: AST_ImportMeta,
            AST_Infinity: AST_Infinity,
            AST_IterationStatement: AST_IterationStatement,
            AST_Jump: AST_Jump,
            AST_Label: AST_Label,
            AST_LabeledStatement: AST_LabeledStatement,
            AST_LabelRef: AST_LabelRef,
            AST_Lambda: AST_Lambda,
            AST_Let: AST_Let,
            AST_LoopControl: AST_LoopControl,
            AST_NameMapping: AST_NameMapping,
            AST_NaN: AST_NaN,
            AST_New: AST_New,
            AST_NewTarget: AST_NewTarget,
            AST_Node: AST_Node,
            AST_Null: AST_Null,
            AST_Number: AST_Number,
            AST_Object: AST_Object,
            AST_ObjectGetter: AST_ObjectGetter,
            AST_ObjectKeyVal: AST_ObjectKeyVal,
            AST_ObjectProperty: AST_ObjectProperty,
            AST_ObjectSetter: AST_ObjectSetter,
            AST_PrefixedTemplateString: AST_PrefixedTemplateString,
            AST_PrivateGetter: AST_PrivateGetter,
            AST_PrivateMethod: AST_PrivateMethod,
            AST_PrivateSetter: AST_PrivateSetter,
            AST_PropAccess: AST_PropAccess,
            AST_RegExp: AST_RegExp,
            AST_Return: AST_Return,
            AST_Scope: AST_Scope,
            AST_Sequence: AST_Sequence,
            AST_SimpleStatement: AST_SimpleStatement,
            AST_Statement: AST_Statement,
            AST_StatementWithBody: AST_StatementWithBody,
            AST_String: AST_String,
            AST_Sub: AST_Sub,
            AST_Super: AST_Super,
            AST_Switch: AST_Switch,
            AST_SwitchBranch: AST_SwitchBranch,
            AST_Symbol: AST_Symbol,
            AST_SymbolBlockDeclaration: AST_SymbolBlockDeclaration,
            AST_SymbolCatch: AST_SymbolCatch,
            AST_SymbolClass: AST_SymbolClass,
            AST_SymbolClassProperty: AST_SymbolClassProperty,
            AST_SymbolConst: AST_SymbolConst,
            AST_SymbolDeclaration: AST_SymbolDeclaration,
            AST_SymbolDefClass: AST_SymbolDefClass,
            AST_SymbolDefun: AST_SymbolDefun,
            AST_SymbolExport: AST_SymbolExport,
            AST_SymbolExportForeign: AST_SymbolExportForeign,
            AST_SymbolFunarg: AST_SymbolFunarg,
            AST_SymbolImport: AST_SymbolImport,
            AST_SymbolImportForeign: AST_SymbolImportForeign,
            AST_SymbolLambda: AST_SymbolLambda,
            AST_SymbolLet: AST_SymbolLet,
            AST_SymbolMethod: AST_SymbolMethod,
            AST_SymbolRef: AST_SymbolRef,
            AST_SymbolVar: AST_SymbolVar,
            AST_TemplateSegment: AST_TemplateSegment,
            AST_TemplateString: AST_TemplateString,
            AST_This: AST_This,
            AST_Throw: AST_Throw,
            AST_Token: AST_Token,
            AST_Toplevel: AST_Toplevel,
            AST_True: AST_True,
            AST_Try: AST_Try,
            AST_Unary: AST_Unary,
            AST_UnaryPostfix: AST_UnaryPostfix,
            AST_UnaryPrefix: AST_UnaryPrefix,
            AST_Undefined: AST_Undefined,
            AST_Var: AST_Var,
            AST_VarDef: AST_VarDef,
            AST_While: AST_While,
            AST_With: AST_With,
            AST_Yield: AST_Yield,
            TreeTransformer: TreeTransformer,
            TreeWalker: TreeWalker,
            walk: walk,
            walk_abort: walk_abort,
            walk_body: walk_body,
            walk_parent: walk_parent,
            _INLINE: _INLINE,
            _NOINLINE: _NOINLINE,
            _PURE: _PURE
        });
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        function def_transform(node, descend) {
            node.DEFMETHOD("transform", function (tw, in_list) {
                let transformed = undefined;
                tw.push(this);
                if (tw.before)
                    transformed = tw.before(this, descend, in_list);
                if (transformed === undefined) {
                    transformed = this;
                    descend(transformed, tw);
                    if (tw.after) {
                        const after_ret = tw.after(transformed, in_list);
                        if (after_ret !== undefined)
                            transformed = after_ret;
                    }
                }
                tw.pop();
                return transformed;
            });
        }
        function do_list(list, tw) {
            return MAP(list, function (node) {
                return node.transform(tw, true);
            });
        }
        def_transform(AST_Node, noop);
        def_transform(AST_LabeledStatement, function (self, tw) {
            self.label = self.label.transform(tw);
            self.body = self.body.transform(tw);
        });
        def_transform(AST_SimpleStatement, function (self, tw) {
            self.body = self.body.transform(tw);
        });
        def_transform(AST_Block, function (self, tw) {
            self.body = do_list(self.body, tw);
        });
        def_transform(AST_Do, function (self, tw) {
            self.body = self.body.transform(tw);
            self.condition = self.condition.transform(tw);
        });
        def_transform(AST_While, function (self, tw) {
            self.condition = self.condition.transform(tw);
            self.body = self.body.transform(tw);
        });
        def_transform(AST_For, function (self, tw) {
            if (self.init)
                self.init = self.init.transform(tw);
            if (self.condition)
                self.condition = self.condition.transform(tw);
            if (self.step)
                self.step = self.step.transform(tw);
            self.body = self.body.transform(tw);
        });
        def_transform(AST_ForIn, function (self, tw) {
            self.init = self.init.transform(tw);
            self.object = self.object.transform(tw);
            self.body = self.body.transform(tw);
        });
        def_transform(AST_With, function (self, tw) {
            self.expression = self.expression.transform(tw);
            self.body = self.body.transform(tw);
        });
        def_transform(AST_Exit, function (self, tw) {
            if (self.value)
                self.value = self.value.transform(tw);
        });
        def_transform(AST_LoopControl, function (self, tw) {
            if (self.label)
                self.label = self.label.transform(tw);
        });
        def_transform(AST_If, function (self, tw) {
            self.condition = self.condition.transform(tw);
            self.body = self.body.transform(tw);
            if (self.alternative)
                self.alternative = self.alternative.transform(tw);
        });
        def_transform(AST_Switch, function (self, tw) {
            self.expression = self.expression.transform(tw);
            self.body = do_list(self.body, tw);
        });
        def_transform(AST_Case, function (self, tw) {
            self.expression = self.expression.transform(tw);
            self.body = do_list(self.body, tw);
        });
        def_transform(AST_Try, function (self, tw) {
            self.body = do_list(self.body, tw);
            if (self.bcatch)
                self.bcatch = self.bcatch.transform(tw);
            if (self.bfinally)
                self.bfinally = self.bfinally.transform(tw);
        });
        def_transform(AST_Catch, function (self, tw) {
            if (self.argname)
                self.argname = self.argname.transform(tw);
            self.body = do_list(self.body, tw);
        });
        def_transform(AST_Definitions, function (self, tw) {
            self.definitions = do_list(self.definitions, tw);
        });
        def_transform(AST_VarDef, function (self, tw) {
            self.name = self.name.transform(tw);
            if (self.value)
                self.value = self.value.transform(tw);
        });
        def_transform(AST_Destructuring, function (self, tw) {
            self.names = do_list(self.names, tw);
        });
        def_transform(AST_Lambda, function (self, tw) {
            if (self.name)
                self.name = self.name.transform(tw);
            self.argnames = do_list(self.argnames, tw);
            if (self.body instanceof AST_Node) {
                self.body = self.body.transform(tw);
            }
            else {
                self.body = do_list(self.body, tw);
            }
        });
        def_transform(AST_Call, function (self, tw) {
            self.expression = self.expression.transform(tw);
            self.args = do_list(self.args, tw);
        });
        def_transform(AST_Sequence, function (self, tw) {
            const result = do_list(self.expressions, tw);
            self.expressions = result.length
                ? result
                : [new AST_Number({ value: 0 })];
        });
        def_transform(AST_Dot, function (self, tw) {
            self.expression = self.expression.transform(tw);
        });
        def_transform(AST_Sub, function (self, tw) {
            self.expression = self.expression.transform(tw);
            self.property = self.property.transform(tw);
        });
        def_transform(AST_Chain, function (self, tw) {
            self.expression = self.expression.transform(tw);
        });
        def_transform(AST_Yield, function (self, tw) {
            if (self.expression)
                self.expression = self.expression.transform(tw);
        });
        def_transform(AST_Await, function (self, tw) {
            self.expression = self.expression.transform(tw);
        });
        def_transform(AST_Unary, function (self, tw) {
            self.expression = self.expression.transform(tw);
        });
        def_transform(AST_Binary, function (self, tw) {
            self.left = self.left.transform(tw);
            self.right = self.right.transform(tw);
        });
        def_transform(AST_Conditional, function (self, tw) {
            self.condition = self.condition.transform(tw);
            self.consequent = self.consequent.transform(tw);
            self.alternative = self.alternative.transform(tw);
        });
        def_transform(AST_Array, function (self, tw) {
            self.elements = do_list(self.elements, tw);
        });
        def_transform(AST_Object, function (self, tw) {
            self.properties = do_list(self.properties, tw);
        });
        def_transform(AST_ObjectProperty, function (self, tw) {
            if (self.key instanceof AST_Node) {
                self.key = self.key.transform(tw);
            }
            if (self.value)
                self.value = self.value.transform(tw);
        });
        def_transform(AST_Class, function (self, tw) {
            if (self.name)
                self.name = self.name.transform(tw);
            if (self.extends)
                self.extends = self.extends.transform(tw);
            self.properties = do_list(self.properties, tw);
        });
        def_transform(AST_Expansion, function (self, tw) {
            self.expression = self.expression.transform(tw);
        });
        def_transform(AST_NameMapping, function (self, tw) {
            self.foreign_name = self.foreign_name.transform(tw);
            self.name = self.name.transform(tw);
        });
        def_transform(AST_Import, function (self, tw) {
            if (self.imported_name)
                self.imported_name = self.imported_name.transform(tw);
            if (self.imported_names)
                do_list(self.imported_names, tw);
            self.module_name = self.module_name.transform(tw);
        });
        def_transform(AST_Export, function (self, tw) {
            if (self.exported_definition)
                self.exported_definition = self.exported_definition.transform(tw);
            if (self.exported_value)
                self.exported_value = self.exported_value.transform(tw);
            if (self.exported_names)
                do_list(self.exported_names, tw);
            if (self.module_name)
                self.module_name = self.module_name.transform(tw);
        });
        def_transform(AST_TemplateString, function (self, tw) {
            self.segments = do_list(self.segments, tw);
        });
        def_transform(AST_PrefixedTemplateString, function (self, tw) {
            self.prefix = self.prefix.transform(tw);
            self.template_string = self.template_string.transform(tw);
        });
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        (function () {
            var normalize_directives = function (body) {
                var in_directive = true;
                for (var i = 0; i < body.length; i++) {
                    if (in_directive && body[i] instanceof AST_Statement && body[i].body instanceof AST_String) {
                        body[i] = new AST_Directive({
                            start: body[i].start,
                            end: body[i].end,
                            value: body[i].body.value
                        });
                    }
                    else if (in_directive && !(body[i] instanceof AST_Statement && body[i].body instanceof AST_String)) {
                        in_directive = false;
                    }
                }
                return body;
            };
            var MOZ_TO_ME = {
                Program: function (M) {
                    return new AST_Toplevel({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        body: normalize_directives(M.body.map(from_moz))
                    });
                },
                ArrayPattern: function (M) {
                    return new AST_Destructuring({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        names: M.elements.map(function (elm) {
                            if (elm === null) {
                                return new AST_Hole();
                            }
                            return from_moz(elm);
                        }),
                        is_array: true
                    });
                },
                ObjectPattern: function (M) {
                    return new AST_Destructuring({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        names: M.properties.map(from_moz),
                        is_array: false
                    });
                },
                AssignmentPattern: function (M) {
                    return new AST_DefaultAssign({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        left: from_moz(M.left),
                        operator: "=",
                        right: from_moz(M.right)
                    });
                },
                SpreadElement: function (M) {
                    return new AST_Expansion({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        expression: from_moz(M.argument)
                    });
                },
                RestElement: function (M) {
                    return new AST_Expansion({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        expression: from_moz(M.argument)
                    });
                },
                TemplateElement: function (M) {
                    return new AST_TemplateSegment({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        value: M.value.cooked,
                        raw: M.value.raw
                    });
                },
                TemplateLiteral: function (M) {
                    var segments = [];
                    for (var i = 0; i < M.quasis.length; i++) {
                        segments.push(from_moz(M.quasis[i]));
                        if (M.expressions[i]) {
                            segments.push(from_moz(M.expressions[i]));
                        }
                    }
                    return new AST_TemplateString({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        segments: segments
                    });
                },
                TaggedTemplateExpression: function (M) {
                    return new AST_PrefixedTemplateString({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        template_string: from_moz(M.quasi),
                        prefix: from_moz(M.tag)
                    });
                },
                FunctionDeclaration: function (M) {
                    return new AST_Defun({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        name: from_moz(M.id),
                        argnames: M.params.map(from_moz),
                        is_generator: M.generator,
                        async: M.async,
                        body: normalize_directives(from_moz(M.body).body)
                    });
                },
                FunctionExpression: function (M) {
                    return new AST_Function({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        name: from_moz(M.id),
                        argnames: M.params.map(from_moz),
                        is_generator: M.generator,
                        async: M.async,
                        body: normalize_directives(from_moz(M.body).body)
                    });
                },
                ArrowFunctionExpression: function (M) {
                    const body = M.body.type === "BlockStatement"
                        ? from_moz(M.body).body
                        : [make_node(AST_Return, {}, { value: from_moz(M.body) })];
                    return new AST_Arrow({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        argnames: M.params.map(from_moz),
                        body,
                        async: M.async,
                    });
                },
                ExpressionStatement: function (M) {
                    return new AST_SimpleStatement({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        body: from_moz(M.expression)
                    });
                },
                TryStatement: function (M) {
                    var handlers = M.handlers || [M.handler];
                    if (handlers.length > 1 || M.guardedHandlers && M.guardedHandlers.length) {
                        throw new Error("Multiple catch clauses are not supported.");
                    }
                    return new AST_Try({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        body: from_moz(M.block).body,
                        bcatch: from_moz(handlers[0]),
                        bfinally: M.finalizer ? new AST_Finally(from_moz(M.finalizer)) : null
                    });
                },
                Property: function (M) {
                    var key = M.key;
                    var args = {
                        start: my_start_token(key || M.value),
                        end: my_end_token(M.value),
                        key: key.type == "Identifier" ? key.name : key.value,
                        value: from_moz(M.value)
                    };
                    if (M.computed) {
                        args.key = from_moz(M.key);
                    }
                    if (M.method) {
                        args.is_generator = M.value.generator;
                        args.async = M.value.async;
                        if (!M.computed) {
                            args.key = new AST_SymbolMethod({ name: args.key });
                        }
                        else {
                            args.key = from_moz(M.key);
                        }
                        return new AST_ConciseMethod(args);
                    }
                    if (M.kind == "init") {
                        if (key.type != "Identifier" && key.type != "Literal") {
                            args.key = from_moz(key);
                        }
                        return new AST_ObjectKeyVal(args);
                    }
                    if (typeof args.key === "string" || typeof args.key === "number") {
                        args.key = new AST_SymbolMethod({
                            name: args.key
                        });
                    }
                    args.value = new AST_Accessor(args.value);
                    if (M.kind == "get")
                        return new AST_ObjectGetter(args);
                    if (M.kind == "set")
                        return new AST_ObjectSetter(args);
                    if (M.kind == "method") {
                        args.async = M.value.async;
                        args.is_generator = M.value.generator;
                        args.quote = M.computed ? "\"" : null;
                        return new AST_ConciseMethod(args);
                    }
                },
                MethodDefinition: function (M) {
                    var args = {
                        start: my_start_token(M),
                        end: my_end_token(M),
                        key: M.computed ? from_moz(M.key) : new AST_SymbolMethod({ name: M.key.name || M.key.value }),
                        value: from_moz(M.value),
                        static: M.static,
                    };
                    if (M.kind == "get") {
                        return new AST_ObjectGetter(args);
                    }
                    if (M.kind == "set") {
                        return new AST_ObjectSetter(args);
                    }
                    args.is_generator = M.value.generator;
                    args.async = M.value.async;
                    return new AST_ConciseMethod(args);
                },
                FieldDefinition: function (M) {
                    let key;
                    if (M.computed) {
                        key = from_moz(M.key);
                    }
                    else {
                        if (M.key.type !== "Identifier")
                            throw new Error("Non-Identifier key in FieldDefinition");
                        key = from_moz(M.key);
                    }
                    return new AST_ClassProperty({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        key,
                        value: from_moz(M.value),
                        static: M.static,
                    });
                },
                PropertyDefinition: function (M) {
                    let key;
                    if (M.computed) {
                        key = from_moz(M.key);
                    }
                    else {
                        if (M.key.type !== "Identifier")
                            throw new Error("Non-Identifier key in PropertyDefinition");
                        key = from_moz(M.key);
                    }
                    return new AST_ClassProperty({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        key,
                        value: from_moz(M.value),
                        static: M.static,
                    });
                },
                ArrayExpression: function (M) {
                    return new AST_Array({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        elements: M.elements.map(function (elem) {
                            return elem === null ? new AST_Hole() : from_moz(elem);
                        })
                    });
                },
                ObjectExpression: function (M) {
                    return new AST_Object({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        properties: M.properties.map(function (prop) {
                            if (prop.type === "SpreadElement") {
                                return from_moz(prop);
                            }
                            prop.type = "Property";
                            return from_moz(prop);
                        })
                    });
                },
                SequenceExpression: function (M) {
                    return new AST_Sequence({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        expressions: M.expressions.map(from_moz)
                    });
                },
                MemberExpression: function (M) {
                    return new (M.computed ? AST_Sub : AST_Dot)({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        property: M.computed ? from_moz(M.property) : M.property.name,
                        expression: from_moz(M.object),
                        optional: M.optional || false
                    });
                },
                ChainExpression: function (M) {
                    return new AST_Chain({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        expression: from_moz(M.expression)
                    });
                },
                SwitchCase: function (M) {
                    return new (M.test ? AST_Case : AST_Default)({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        expression: from_moz(M.test),
                        body: M.consequent.map(from_moz)
                    });
                },
                VariableDeclaration: function (M) {
                    return new (M.kind === "const" ? AST_Const :
                        M.kind === "let" ? AST_Let : AST_Var)({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        definitions: M.declarations.map(from_moz)
                    });
                },
                ImportDeclaration: function (M) {
                    var imported_name = null;
                    var imported_names = null;
                    M.specifiers.forEach(function (specifier) {
                        if (specifier.type === "ImportSpecifier") {
                            if (!imported_names) {
                                imported_names = [];
                            }
                            imported_names.push(new AST_NameMapping({
                                start: my_start_token(specifier),
                                end: my_end_token(specifier),
                                foreign_name: from_moz(specifier.imported),
                                name: from_moz(specifier.local)
                            }));
                        }
                        else if (specifier.type === "ImportDefaultSpecifier") {
                            imported_name = from_moz(specifier.local);
                        }
                        else if (specifier.type === "ImportNamespaceSpecifier") {
                            if (!imported_names) {
                                imported_names = [];
                            }
                            imported_names.push(new AST_NameMapping({
                                start: my_start_token(specifier),
                                end: my_end_token(specifier),
                                foreign_name: new AST_SymbolImportForeign({ name: "*" }),
                                name: from_moz(specifier.local)
                            }));
                        }
                    });
                    return new AST_Import({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        imported_name: imported_name,
                        imported_names: imported_names,
                        module_name: from_moz(M.source)
                    });
                },
                ExportAllDeclaration: function (M) {
                    return new AST_Export({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        exported_names: [
                            new AST_NameMapping({
                                name: new AST_SymbolExportForeign({ name: "*" }),
                                foreign_name: new AST_SymbolExportForeign({ name: "*" })
                            })
                        ],
                        module_name: from_moz(M.source)
                    });
                },
                ExportNamedDeclaration: function (M) {
                    return new AST_Export({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        exported_definition: from_moz(M.declaration),
                        exported_names: M.specifiers && M.specifiers.length ? M.specifiers.map(function (specifier) {
                            return new AST_NameMapping({
                                foreign_name: from_moz(specifier.exported),
                                name: from_moz(specifier.local)
                            });
                        }) : null,
                        module_name: from_moz(M.source)
                    });
                },
                ExportDefaultDeclaration: function (M) {
                    return new AST_Export({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        exported_value: from_moz(M.declaration),
                        is_default: true
                    });
                },
                Literal: function (M) {
                    var val = M.value, args = {
                        start: my_start_token(M),
                        end: my_end_token(M)
                    };
                    var rx = M.regex;
                    if (rx && rx.pattern) {
                        // RegExpLiteral as per ESTree AST spec
                        args.value = {
                            source: rx.pattern,
                            flags: rx.flags
                        };
                        return new AST_RegExp(args);
                    }
                    else if (rx) {
                        // support legacy RegExp
                        const rx_source = M.raw || val;
                        const match = rx_source.match(/^\/(.*)\/(\w*)$/);
                        if (!match)
                            throw new Error("Invalid regex source " + rx_source);
                        const [_, source, flags] = match;
                        args.value = { source, flags };
                        return new AST_RegExp(args);
                    }
                    if (val === null)
                        return new AST_Null(args);
                    switch (typeof val) {
                        case "string":
                            args.value = val;
                            return new AST_String(args);
                        case "number":
                            args.value = val;
                            args.raw = M.raw || val.toString();
                            return new AST_Number(args);
                        case "boolean":
                            return new (val ? AST_True : AST_False)(args);
                    }
                },
                MetaProperty: function (M) {
                    if (M.meta.name === "new" && M.property.name === "target") {
                        return new AST_NewTarget({
                            start: my_start_token(M),
                            end: my_end_token(M)
                        });
                    }
                    else if (M.meta.name === "import" && M.property.name === "meta") {
                        return new AST_ImportMeta({
                            start: my_start_token(M),
                            end: my_end_token(M)
                        });
                    }
                },
                Identifier: function (M) {
                    var p = FROM_MOZ_STACK[FROM_MOZ_STACK.length - 2];
                    return new (p.type == "LabeledStatement" ? AST_Label
                        : p.type == "VariableDeclarator" && p.id === M ? (p.kind == "const" ? AST_SymbolConst : p.kind == "let" ? AST_SymbolLet : AST_SymbolVar)
                            : /Import.*Specifier/.test(p.type) ? (p.local === M ? AST_SymbolImport : AST_SymbolImportForeign)
                                : p.type == "ExportSpecifier" ? (p.local === M ? AST_SymbolExport : AST_SymbolExportForeign)
                                    : p.type == "FunctionExpression" ? (p.id === M ? AST_SymbolLambda : AST_SymbolFunarg)
                                        : p.type == "FunctionDeclaration" ? (p.id === M ? AST_SymbolDefun : AST_SymbolFunarg)
                                            : p.type == "ArrowFunctionExpression" ? (p.params.includes(M)) ? AST_SymbolFunarg : AST_SymbolRef
                                                : p.type == "ClassExpression" ? (p.id === M ? AST_SymbolClass : AST_SymbolRef)
                                                    : p.type == "Property" ? (p.key === M && p.computed || p.value === M ? AST_SymbolRef : AST_SymbolMethod)
                                                        : p.type == "PropertyDefinition" || p.type === "FieldDefinition" ? (p.key === M && p.computed || p.value === M ? AST_SymbolRef : AST_SymbolClassProperty)
                                                            : p.type == "ClassDeclaration" ? (p.id === M ? AST_SymbolDefClass : AST_SymbolRef)
                                                                : p.type == "MethodDefinition" ? (p.computed ? AST_SymbolRef : AST_SymbolMethod)
                                                                    : p.type == "CatchClause" ? AST_SymbolCatch
                                                                        : p.type == "BreakStatement" || p.type == "ContinueStatement" ? AST_LabelRef
                                                                            : AST_SymbolRef)({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        name: M.name
                    });
                },
                BigIntLiteral(M) {
                    return new AST_BigInt({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        value: M.value
                    });
                }
            };
            MOZ_TO_ME.UpdateExpression =
                MOZ_TO_ME.UnaryExpression = function To_Moz_Unary(M) {
                    var prefix = "prefix" in M ? M.prefix
                        : M.type == "UnaryExpression" ? true : false;
                    return new (prefix ? AST_UnaryPrefix : AST_UnaryPostfix)({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        operator: M.operator,
                        expression: from_moz(M.argument)
                    });
                };
            MOZ_TO_ME.ClassDeclaration =
                MOZ_TO_ME.ClassExpression = function From_Moz_Class(M) {
                    return new (M.type === "ClassDeclaration" ? AST_DefClass : AST_ClassExpression)({
                        start: my_start_token(M),
                        end: my_end_token(M),
                        name: from_moz(M.id),
                        extends: from_moz(M.superClass),
                        properties: M.body.body.map(from_moz)
                    });
                };
            map("EmptyStatement", AST_EmptyStatement);
            map("BlockStatement", AST_BlockStatement, "body@body");
            map("IfStatement", AST_If, "test>condition, consequent>body, alternate>alternative");
            map("LabeledStatement", AST_LabeledStatement, "label>label, body>body");
            map("BreakStatement", AST_Break, "label>label");
            map("ContinueStatement", AST_Continue, "label>label");
            map("WithStatement", AST_With, "object>expression, body>body");
            map("SwitchStatement", AST_Switch, "discriminant>expression, cases@body");
            map("ReturnStatement", AST_Return, "argument>value");
            map("ThrowStatement", AST_Throw, "argument>value");
            map("WhileStatement", AST_While, "test>condition, body>body");
            map("DoWhileStatement", AST_Do, "test>condition, body>body");
            map("ForStatement", AST_For, "init>init, test>condition, update>step, body>body");
            map("ForInStatement", AST_ForIn, "left>init, right>object, body>body");
            map("ForOfStatement", AST_ForOf, "left>init, right>object, body>body, await=await");
            map("AwaitExpression", AST_Await, "argument>expression");
            map("YieldExpression", AST_Yield, "argument>expression, delegate=is_star");
            map("DebuggerStatement", AST_Debugger);
            map("VariableDeclarator", AST_VarDef, "id>name, init>value");
            map("CatchClause", AST_Catch, "param>argname, body%body");
            map("ThisExpression", AST_This);
            map("Super", AST_Super);
            map("BinaryExpression", AST_Binary, "operator=operator, left>left, right>right");
            map("LogicalExpression", AST_Binary, "operator=operator, left>left, right>right");
            map("AssignmentExpression", AST_Assign, "operator=operator, left>left, right>right");
            map("ConditionalExpression", AST_Conditional, "test>condition, consequent>consequent, alternate>alternative");
            map("NewExpression", AST_New, "callee>expression, arguments@args");
            map("CallExpression", AST_Call, "callee>expression, optional=optional, arguments@args");
            def_to_moz(AST_Toplevel, function To_Moz_Program(M) {
                return to_moz_scope("Program", M);
            });
            def_to_moz(AST_Expansion, function To_Moz_Spread(M) {
                return {
                    type: to_moz_in_destructuring() ? "RestElement" : "SpreadElement",
                    argument: to_moz(M.expression)
                };
            });
            def_to_moz(AST_PrefixedTemplateString, function To_Moz_TaggedTemplateExpression(M) {
                return {
                    type: "TaggedTemplateExpression",
                    tag: to_moz(M.prefix),
                    quasi: to_moz(M.template_string)
                };
            });
            def_to_moz(AST_TemplateString, function To_Moz_TemplateLiteral(M) {
                var quasis = [];
                var expressions = [];
                for (var i = 0; i < M.segments.length; i++) {
                    if (i % 2 !== 0) {
                        expressions.push(to_moz(M.segments[i]));
                    }
                    else {
                        quasis.push({
                            type: "TemplateElement",
                            value: {
                                raw: M.segments[i].raw,
                                cooked: M.segments[i].value
                            },
                            tail: i === M.segments.length - 1
                        });
                    }
                }
                return {
                    type: "TemplateLiteral",
                    quasis: quasis,
                    expressions: expressions
                };
            });
            def_to_moz(AST_Defun, function To_Moz_FunctionDeclaration(M) {
                return {
                    type: "FunctionDeclaration",
                    id: to_moz(M.name),
                    params: M.argnames.map(to_moz),
                    generator: M.is_generator,
                    async: M.async,
                    body: to_moz_scope("BlockStatement", M)
                };
            });
            def_to_moz(AST_Function, function To_Moz_FunctionExpression(M, parent) {
                var is_generator = parent.is_generator !== undefined ?
                    parent.is_generator : M.is_generator;
                return {
                    type: "FunctionExpression",
                    id: to_moz(M.name),
                    params: M.argnames.map(to_moz),
                    generator: is_generator,
                    async: M.async,
                    body: to_moz_scope("BlockStatement", M)
                };
            });
            def_to_moz(AST_Arrow, function To_Moz_ArrowFunctionExpression(M) {
                var body = {
                    type: "BlockStatement",
                    body: M.body.map(to_moz)
                };
                return {
                    type: "ArrowFunctionExpression",
                    params: M.argnames.map(to_moz),
                    async: M.async,
                    body: body
                };
            });
            def_to_moz(AST_Destructuring, function To_Moz_ObjectPattern(M) {
                if (M.is_array) {
                    return {
                        type: "ArrayPattern",
                        elements: M.names.map(to_moz)
                    };
                }
                return {
                    type: "ObjectPattern",
                    properties: M.names.map(to_moz)
                };
            });
            def_to_moz(AST_Directive, function To_Moz_Directive(M) {
                return {
                    type: "ExpressionStatement",
                    expression: {
                        type: "Literal",
                        value: M.value,
                        raw: M.print_to_string()
                    },
                    directive: M.value
                };
            });
            def_to_moz(AST_SimpleStatement, function To_Moz_ExpressionStatement(M) {
                return {
                    type: "ExpressionStatement",
                    expression: to_moz(M.body)
                };
            });
            def_to_moz(AST_SwitchBranch, function To_Moz_SwitchCase(M) {
                return {
                    type: "SwitchCase",
                    test: to_moz(M.expression),
                    consequent: M.body.map(to_moz)
                };
            });
            def_to_moz(AST_Try, function To_Moz_TryStatement(M) {
                return {
                    type: "TryStatement",
                    block: to_moz_block(M),
                    handler: to_moz(M.bcatch),
                    guardedHandlers: [],
                    finalizer: to_moz(M.bfinally)
                };
            });
            def_to_moz(AST_Catch, function To_Moz_CatchClause(M) {
                return {
                    type: "CatchClause",
                    param: to_moz(M.argname),
                    guard: null,
                    body: to_moz_block(M)
                };
            });
            def_to_moz(AST_Definitions, function To_Moz_VariableDeclaration(M) {
                return {
                    type: "VariableDeclaration",
                    kind: M instanceof AST_Const ? "const" :
                        M instanceof AST_Let ? "let" : "var",
                    declarations: M.definitions.map(to_moz)
                };
            });
            def_to_moz(AST_Export, function To_Moz_ExportDeclaration(M) {
                if (M.exported_names) {
                    if (M.exported_names[0].name.name === "*") {
                        return {
                            type: "ExportAllDeclaration",
                            source: to_moz(M.module_name)
                        };
                    }
                    return {
                        type: "ExportNamedDeclaration",
                        specifiers: M.exported_names.map(function (name_mapping) {
                            return {
                                type: "ExportSpecifier",
                                exported: to_moz(name_mapping.foreign_name),
                                local: to_moz(name_mapping.name)
                            };
                        }),
                        declaration: to_moz(M.exported_definition),
                        source: to_moz(M.module_name)
                    };
                }
                return {
                    type: M.is_default ? "ExportDefaultDeclaration" : "ExportNamedDeclaration",
                    declaration: to_moz(M.exported_value || M.exported_definition)
                };
            });
            def_to_moz(AST_Import, function To_Moz_ImportDeclaration(M) {
                var specifiers = [];
                if (M.imported_name) {
                    specifiers.push({
                        type: "ImportDefaultSpecifier",
                        local: to_moz(M.imported_name)
                    });
                }
                if (M.imported_names && M.imported_names[0].foreign_name.name === "*") {
                    specifiers.push({
                        type: "ImportNamespaceSpecifier",
                        local: to_moz(M.imported_names[0].name)
                    });
                }
                else if (M.imported_names) {
                    M.imported_names.forEach(function (name_mapping) {
                        specifiers.push({
                            type: "ImportSpecifier",
                            local: to_moz(name_mapping.name),
                            imported: to_moz(name_mapping.foreign_name)
                        });
                    });
                }
                return {
                    type: "ImportDeclaration",
                    specifiers: specifiers,
                    source: to_moz(M.module_name)
                };
            });
            def_to_moz(AST_ImportMeta, function To_Moz_MetaProperty() {
                return {
                    type: "MetaProperty",
                    meta: {
                        type: "Identifier",
                        name: "import"
                    },
                    property: {
                        type: "Identifier",
                        name: "meta"
                    }
                };
            });
            def_to_moz(AST_Sequence, function To_Moz_SequenceExpression(M) {
                return {
                    type: "SequenceExpression",
                    expressions: M.expressions.map(to_moz)
                };
            });
            def_to_moz(AST_DotHash, function To_Moz_PrivateMemberExpression(M) {
                return {
                    type: "MemberExpression",
                    object: to_moz(M.expression),
                    computed: false,
                    property: {
                        type: "PrivateIdentifier",
                        name: M.property
                    },
                    optional: M.optional
                };
            });
            def_to_moz(AST_PropAccess, function To_Moz_MemberExpression(M) {
                var isComputed = M instanceof AST_Sub;
                return {
                    type: "MemberExpression",
                    object: to_moz(M.expression),
                    computed: isComputed,
                    property: isComputed ? to_moz(M.property) : { type: "Identifier", name: M.property },
                    optional: M.optional
                };
            });
            def_to_moz(AST_Chain, function To_Moz_ChainExpression(M) {
                return {
                    type: "ChainExpression",
                    expression: to_moz(M.expression)
                };
            });
            def_to_moz(AST_Unary, function To_Moz_Unary(M) {
                return {
                    type: M.operator == "++" || M.operator == "--" ? "UpdateExpression" : "UnaryExpression",
                    operator: M.operator,
                    prefix: M instanceof AST_UnaryPrefix,
                    argument: to_moz(M.expression)
                };
            });
            def_to_moz(AST_Binary, function To_Moz_BinaryExpression(M) {
                if (M.operator == "=" && to_moz_in_destructuring()) {
                    return {
                        type: "AssignmentPattern",
                        left: to_moz(M.left),
                        right: to_moz(M.right)
                    };
                }
                const type = M.operator == "&&" || M.operator == "||" || M.operator === "??"
                    ? "LogicalExpression"
                    : "BinaryExpression";
                return {
                    type,
                    left: to_moz(M.left),
                    operator: M.operator,
                    right: to_moz(M.right)
                };
            });
            def_to_moz(AST_Array, function To_Moz_ArrayExpression(M) {
                return {
                    type: "ArrayExpression",
                    elements: M.elements.map(to_moz)
                };
            });
            def_to_moz(AST_Object, function To_Moz_ObjectExpression(M) {
                return {
                    type: "ObjectExpression",
                    properties: M.properties.map(to_moz)
                };
            });
            def_to_moz(AST_ObjectProperty, function To_Moz_Property(M, parent) {
                var key = M.key instanceof AST_Node ? to_moz(M.key) : {
                    type: "Identifier",
                    value: M.key
                };
                if (typeof M.key === "number") {
                    key = {
                        type: "Literal",
                        value: Number(M.key)
                    };
                }
                if (typeof M.key === "string") {
                    key = {
                        type: "Identifier",
                        name: M.key
                    };
                }
                var kind;
                var string_or_num = typeof M.key === "string" || typeof M.key === "number";
                var computed = string_or_num ? false : !(M.key instanceof AST_Symbol) || M.key instanceof AST_SymbolRef;
                if (M instanceof AST_ObjectKeyVal) {
                    kind = "init";
                    computed = !string_or_num;
                }
                else if (M instanceof AST_ObjectGetter) {
                    kind = "get";
                }
                else if (M instanceof AST_ObjectSetter) {
                    kind = "set";
                }
                if (M instanceof AST_PrivateGetter || M instanceof AST_PrivateSetter) {
                    const kind = M instanceof AST_PrivateGetter ? "get" : "set";
                    return {
                        type: "MethodDefinition",
                        computed: false,
                        kind: kind,
                        static: M.static,
                        key: {
                            type: "PrivateIdentifier",
                            name: M.key.name
                        },
                        value: to_moz(M.value)
                    };
                }
                if (M instanceof AST_ClassPrivateProperty) {
                    return {
                        type: "PropertyDefinition",
                        key: {
                            type: "PrivateIdentifier",
                            name: M.key.name
                        },
                        value: to_moz(M.value),
                        computed: false,
                        static: M.static
                    };
                }
                if (M instanceof AST_ClassProperty) {
                    return {
                        type: "PropertyDefinition",
                        key,
                        value: to_moz(M.value),
                        computed,
                        static: M.static
                    };
                }
                if (parent instanceof AST_Class) {
                    return {
                        type: "MethodDefinition",
                        computed: computed,
                        kind: kind,
                        static: M.static,
                        key: to_moz(M.key),
                        value: to_moz(M.value)
                    };
                }
                return {
                    type: "Property",
                    computed: computed,
                    kind: kind,
                    key: key,
                    value: to_moz(M.value)
                };
            });
            def_to_moz(AST_ConciseMethod, function To_Moz_MethodDefinition(M, parent) {
                if (parent instanceof AST_Object) {
                    return {
                        type: "Property",
                        computed: !(M.key instanceof AST_Symbol) || M.key instanceof AST_SymbolRef,
                        kind: "init",
                        method: true,
                        shorthand: false,
                        key: to_moz(M.key),
                        value: to_moz(M.value)
                    };
                }
                const key = M instanceof AST_PrivateMethod
                    ? {
                        type: "PrivateIdentifier",
                        name: M.key.name
                    }
                    : to_moz(M.key);
                return {
                    type: "MethodDefinition",
                    kind: M.key === "constructor" ? "constructor" : "method",
                    key,
                    value: to_moz(M.value),
                    computed: !(M.key instanceof AST_Symbol) || M.key instanceof AST_SymbolRef,
                    static: M.static,
                };
            });
            def_to_moz(AST_Class, function To_Moz_Class(M) {
                var type = M instanceof AST_ClassExpression ? "ClassExpression" : "ClassDeclaration";
                return {
                    type: type,
                    superClass: to_moz(M.extends),
                    id: M.name ? to_moz(M.name) : null,
                    body: {
                        type: "ClassBody",
                        body: M.properties.map(to_moz)
                    }
                };
            });
            def_to_moz(AST_NewTarget, function To_Moz_MetaProperty() {
                return {
                    type: "MetaProperty",
                    meta: {
                        type: "Identifier",
                        name: "new"
                    },
                    property: {
                        type: "Identifier",
                        name: "target"
                    }
                };
            });
            def_to_moz(AST_Symbol, function To_Moz_Identifier(M, parent) {
                if (M instanceof AST_SymbolMethod && parent.quote) {
                    return {
                        type: "Literal",
                        value: M.name
                    };
                }
                var def = M.definition();
                return {
                    type: "Identifier",
                    name: def ? def.mangled_name || def.name : M.name
                };
            });
            def_to_moz(AST_RegExp, function To_Moz_RegExpLiteral(M) {
                const pattern = M.value.source;
                const flags = M.value.flags;
                return {
                    type: "Literal",
                    value: null,
                    raw: M.print_to_string(),
                    regex: { pattern, flags }
                };
            });
            def_to_moz(AST_Constant, function To_Moz_Literal(M) {
                var value = M.value;
                return {
                    type: "Literal",
                    value: value,
                    raw: M.raw || M.print_to_string()
                };
            });
            def_to_moz(AST_Atom, function To_Moz_Atom(M) {
                return {
                    type: "Identifier",
                    name: String(M.value)
                };
            });
            def_to_moz(AST_BigInt, M => ({
                type: "BigIntLiteral",
                value: M.value
            }));
            AST_Boolean.DEFMETHOD("to_mozilla_ast", AST_Constant.prototype.to_mozilla_ast);
            AST_Null.DEFMETHOD("to_mozilla_ast", AST_Constant.prototype.to_mozilla_ast);
            AST_Hole.DEFMETHOD("to_mozilla_ast", function To_Moz_ArrayHole() { return null; });
            AST_Block.DEFMETHOD("to_mozilla_ast", AST_BlockStatement.prototype.to_mozilla_ast);
            AST_Lambda.DEFMETHOD("to_mozilla_ast", AST_Function.prototype.to_mozilla_ast);
            /* -----[ tools ]----- */
            function my_start_token(moznode) {
                var loc = moznode.loc, start = loc && loc.start;
                var range = moznode.range;
                return new AST_Token("", "", start && start.line || 0, start && start.column || 0, range ? range[0] : moznode.start, false, [], [], loc && loc.source);
            }
            function my_end_token(moznode) {
                var loc = moznode.loc, end = loc && loc.end;
                var range = moznode.range;
                return new AST_Token("", "", end && end.line || 0, end && end.column || 0, range ? range[0] : moznode.end, false, [], [], loc && loc.source);
            }
            function map(moztype, mytype, propmap) {
                var moz_to_me = "function From_Moz_" + moztype + "(M){\n";
                moz_to_me += "return new U2." + mytype.name + "({\n" +
                    "start: my_start_token(M),\n" +
                    "end: my_end_token(M)";
                var me_to_moz = "function To_Moz_" + moztype + "(M){\n";
                me_to_moz += "return {\n" +
                    "type: " + JSON.stringify(moztype);
                if (propmap)
                    propmap.split(/\s*,\s*/).forEach(function (prop) {
                        var m = /([a-z0-9$_]+)([=@>%])([a-z0-9$_]+)/i.exec(prop);
                        if (!m)
                            throw new Error("Can't understand property map: " + prop);
                        var moz = m[1], how = m[2], my = m[3];
                        moz_to_me += ",\n" + my + ": ";
                        me_to_moz += ",\n" + moz + ": ";
                        switch (how) {
                            case "@":
                                moz_to_me += "M." + moz + ".map(from_moz)";
                                me_to_moz += "M." + my + ".map(to_moz)";
                                break;
                            case ">":
                                moz_to_me += "from_moz(M." + moz + ")";
                                me_to_moz += "to_moz(M." + my + ")";
                                break;
                            case "=":
                                moz_to_me += "M." + moz;
                                me_to_moz += "M." + my;
                                break;
                            case "%":
                                moz_to_me += "from_moz(M." + moz + ").body";
                                me_to_moz += "to_moz_block(M)";
                                break;
                            default:
                                throw new Error("Can't understand operator in propmap: " + prop);
                        }
                    });
                moz_to_me += "\n})\n}";
                me_to_moz += "\n}\n}";
                moz_to_me = new Function("U2", "my_start_token", "my_end_token", "from_moz", "return(" + moz_to_me + ")")(ast, my_start_token, my_end_token, from_moz);
                me_to_moz = new Function("to_moz", "to_moz_block", "to_moz_scope", "return(" + me_to_moz + ")")(to_moz, to_moz_block, to_moz_scope);
                MOZ_TO_ME[moztype] = moz_to_me;
                def_to_moz(mytype, me_to_moz);
            }
            var FROM_MOZ_STACK = null;
            function from_moz(node) {
                FROM_MOZ_STACK.push(node);
                var ret = node != null ? MOZ_TO_ME[node.type](node) : null;
                FROM_MOZ_STACK.pop();
                return ret;
            }
            AST_Node.from_mozilla_ast = function (node) {
                var save_stack = FROM_MOZ_STACK;
                FROM_MOZ_STACK = [];
                var ast = from_moz(node);
                FROM_MOZ_STACK = save_stack;
                return ast;
            };
            function set_moz_loc(mynode, moznode) {
                var start = mynode.start;
                var end = mynode.end;
                if (!(start && end)) {
                    return moznode;
                }
                if (start.pos != null && end.endpos != null) {
                    moznode.range = [start.pos, end.endpos];
                }
                if (start.line) {
                    moznode.loc = {
                        start: { line: start.line, column: start.col },
                        end: end.endline ? { line: end.endline, column: end.endcol } : null
                    };
                    if (start.file) {
                        moznode.loc.source = start.file;
                    }
                }
                return moznode;
            }
            function def_to_moz(mytype, handler) {
                mytype.DEFMETHOD("to_mozilla_ast", function (parent) {
                    return set_moz_loc(this, handler(this, parent));
                });
            }
            var TO_MOZ_STACK = null;
            function to_moz(node) {
                if (TO_MOZ_STACK === null) {
                    TO_MOZ_STACK = [];
                }
                TO_MOZ_STACK.push(node);
                var ast = node != null ? node.to_mozilla_ast(TO_MOZ_STACK[TO_MOZ_STACK.length - 2]) : null;
                TO_MOZ_STACK.pop();
                if (TO_MOZ_STACK.length === 0) {
                    TO_MOZ_STACK = null;
                }
                return ast;
            }
            function to_moz_in_destructuring() {
                var i = TO_MOZ_STACK.length;
                while (i--) {
                    if (TO_MOZ_STACK[i] instanceof AST_Destructuring) {
                        return true;
                    }
                }
                return false;
            }
            function to_moz_block(node) {
                return {
                    type: "BlockStatement",
                    body: node.body.map(to_moz)
                };
            }
            function to_moz_scope(type, node) {
                var body = node.body.map(to_moz);
                if (node.body[0] instanceof AST_SimpleStatement && node.body[0].body instanceof AST_String) {
                    body.unshift(to_moz(new AST_EmptyStatement(node.body[0])));
                }
                return {
                    type: type,
                    body: body
                };
            }
        })();
        // return true if the node at the top of the stack (that means the
        // innermost node in the current output) is lexically the first in
        // a statement.
        function first_in_statement(stack) {
            let node = stack.parent(-1);
            for (let i = 0, p; p = stack.parent(i); i++) {
                if (p instanceof AST_Statement && p.body === node)
                    return true;
                if ((p instanceof AST_Sequence && p.expressions[0] === node) ||
                    (p.TYPE === "Call" && p.expression === node) ||
                    (p instanceof AST_PrefixedTemplateString && p.prefix === node) ||
                    (p instanceof AST_Dot && p.expression === node) ||
                    (p instanceof AST_Sub && p.expression === node) ||
                    (p instanceof AST_Conditional && p.condition === node) ||
                    (p instanceof AST_Binary && p.left === node) ||
                    (p instanceof AST_UnaryPostfix && p.expression === node)) {
                    node = p;
                }
                else {
                    return false;
                }
            }
        }
        // Returns whether the leftmost item in the expression is an object
        function left_is_object(node) {
            if (node instanceof AST_Object)
                return true;
            if (node instanceof AST_Sequence)
                return left_is_object(node.expressions[0]);
            if (node.TYPE === "Call")
                return left_is_object(node.expression);
            if (node instanceof AST_PrefixedTemplateString)
                return left_is_object(node.prefix);
            if (node instanceof AST_Dot || node instanceof AST_Sub)
                return left_is_object(node.expression);
            if (node instanceof AST_Conditional)
                return left_is_object(node.condition);
            if (node instanceof AST_Binary)
                return left_is_object(node.left);
            if (node instanceof AST_UnaryPostfix)
                return left_is_object(node.expression);
            return false;
        }
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        const EXPECT_DIRECTIVE = /^$|[;{][\s\n]*$/;
        const CODE_LINE_BREAK = 10;
        const CODE_SPACE = 32;
        const r_annotation = /[@#]__(PURE|INLINE|NOINLINE)__/g;
        function is_some_comments(comment) {
            // multiline comment
            return ((comment.type === "comment2" || comment.type === "comment1")
                && /@preserve|@lic|@cc_on|^\**!/i.test(comment.value));
        }
        function OutputStream(options) {
            var readonly = !options;
            options = defaults(options, {
                ascii_only: false,
                beautify: false,
                braces: false,
                comments: "some",
                ecma: 5,
                ie8: false,
                indent_level: 4,
                indent_start: 0,
                inline_script: true,
                keep_numbers: false,
                keep_quoted_props: false,
                max_line_len: false,
                preamble: null,
                preserve_annotations: false,
                quote_keys: false,
                quote_style: 0,
                safari10: false,
                semicolons: true,
                shebang: true,
                shorthand: undefined,
                source_map: null,
                webkit: false,
                width: 80,
                wrap_iife: false,
                wrap_func_args: true,
            }, true);
            if (options.shorthand === undefined)
                options.shorthand = options.ecma > 5;
            // Convert comment option to RegExp if neccessary and set up comments filter
            var comment_filter = return_false; // Default case, throw all comments away
            if (options.comments) {
                let comments = options.comments;
                if (typeof options.comments === "string" && /^\/.*\/[a-zA-Z]*$/.test(options.comments)) {
                    var regex_pos = options.comments.lastIndexOf("/");
                    comments = new RegExp(options.comments.substr(1, regex_pos - 1), options.comments.substr(regex_pos + 1));
                }
                if (comments instanceof RegExp) {
                    comment_filter = function (comment) {
                        return comment.type != "comment5" && comments.test(comment.value);
                    };
                }
                else if (typeof comments === "function") {
                    comment_filter = function (comment) {
                        return comment.type != "comment5" && comments(this, comment);
                    };
                }
                else if (comments === "some") {
                    comment_filter = is_some_comments;
                }
                else { // NOTE includes "all" option
                    comment_filter = return_true;
                }
            }
            var indentation = 0;
            var current_col = 0;
            var current_line = 1;
            var current_pos = 0;
            var OUTPUT = "";
            let printed_comments = new Set();
            var to_utf8 = options.ascii_only ? function (str, identifier) {
                if (options.ecma >= 2015 && !options.safari10) {
                    str = str.replace(/[\ud800-\udbff][\udc00-\udfff]/g, function (ch) {
                        var code = get_full_char_code(ch, 0).toString(16);
                        return "\\u{" + code + "}";
                    });
                }
                return str.replace(/[\u0000-\u001f\u007f-\uffff]/g, function (ch) {
                    var code = ch.charCodeAt(0).toString(16);
                    if (code.length <= 2 && !identifier) {
                        while (code.length < 2)
                            code = "0" + code;
                        return "\\x" + code;
                    }
                    else {
                        while (code.length < 4)
                            code = "0" + code;
                        return "\\u" + code;
                    }
                });
            } : function (str) {
                return str.replace(/[\ud800-\udbff][\udc00-\udfff]|([\ud800-\udbff]|[\udc00-\udfff])/g, function (match, lone) {
                    if (lone) {
                        return "\\u" + lone.charCodeAt(0).toString(16);
                    }
                    return match;
                });
            };
            function make_string(str, quote) {
                var dq = 0, sq = 0;
                str = str.replace(/[\\\b\f\n\r\v\t\x22\x27\u2028\u2029\0\ufeff]/g, function (s, i) {
                    switch (s) {
                        case '"':
                            ++dq;
                            return '"';
                        case "'":
                            ++sq;
                            return "'";
                        case "\\": return "\\\\";
                        case "\n": return "\\n";
                        case "\r": return "\\r";
                        case "\t": return "\\t";
                        case "\b": return "\\b";
                        case "\f": return "\\f";
                        case "\x0B": return options.ie8 ? "\\x0B" : "\\v";
                        case "\u2028": return "\\u2028";
                        case "\u2029": return "\\u2029";
                        case "\ufeff": return "\\ufeff";
                        case "\0":
                            return /[0-9]/.test(get_full_char(str, i + 1)) ? "\\x00" : "\\0";
                    }
                    return s;
                });
                function quote_single() {
                    return "'" + str.replace(/\x27/g, "\\'") + "'";
                }
                function quote_double() {
                    return '"' + str.replace(/\x22/g, '\\"') + '"';
                }
                function quote_template() {
                    return "`" + str.replace(/`/g, "\\`") + "`";
                }
                str = to_utf8(str);
                if (quote === "`")
                    return quote_template();
                switch (options.quote_style) {
                    case 1:
                        return quote_single();
                    case 2:
                        return quote_double();
                    case 3:
                        return quote == "'" ? quote_single() : quote_double();
                    default:
                        return dq > sq ? quote_single() : quote_double();
                }
            }
            function encode_string(str, quote) {
                var ret = make_string(str, quote);
                if (options.inline_script) {
                    ret = ret.replace(/<\x2f(script)([>\/\t\n\f\r ])/gi, "<\\/$1$2");
                    ret = ret.replace(/\x3c!--/g, "\\x3c!--");
                    ret = ret.replace(/--\x3e/g, "--\\x3e");
                }
                return ret;
            }
            function make_name(name) {
                name = name.toString();
                name = to_utf8(name, true);
                return name;
            }
            function make_indent(back) {
                return " ".repeat(options.indent_start + indentation - back * options.indent_level);
            }
            /* -----[ beautification/minification ]----- */
            var has_parens = false;
            var might_need_space = false;
            var might_need_semicolon = false;
            var might_add_newline = 0;
            var need_newline_indented = false;
            var need_space = false;
            var newline_insert = -1;
            var last = "";
            var mapping_token, mapping_name, mappings = options.source_map && [];
            var do_add_mapping = mappings ? function () {
                mappings.forEach(function (mapping) {
                    try {
                        let name = !mapping.name && mapping.token.type == "name" ? mapping.token.value : mapping.name;
                        if (name instanceof AST_Symbol) {
                            name = name.name;
                        }
                        options.source_map.add(mapping.token.file, mapping.line, mapping.col, mapping.token.line, mapping.token.col, is_basic_identifier_string(name) ? name : undefined);
                    }
                    catch (ex) {
                        // Ignore bad mapping
                    }
                });
                mappings = [];
            } : noop;
            var ensure_line_len = options.max_line_len ? function () {
                if (current_col > options.max_line_len) {
                    if (might_add_newline) {
                        var left = OUTPUT.slice(0, might_add_newline);
                        var right = OUTPUT.slice(might_add_newline);
                        if (mappings) {
                            var delta = right.length - current_col;
                            mappings.forEach(function (mapping) {
                                mapping.line++;
                                mapping.col += delta;
                            });
                        }
                        OUTPUT = left + "\n" + right;
                        current_line++;
                        current_pos++;
                        current_col = right.length;
                    }
                }
                if (might_add_newline) {
                    might_add_newline = 0;
                    do_add_mapping();
                }
            } : noop;
            var requireSemicolonChars = makePredicate("( [ + * / - , . `");
            function print(str) {
                str = String(str);
                var ch = get_full_char(str, 0);
                if (need_newline_indented && ch) {
                    need_newline_indented = false;
                    if (ch !== "\n") {
                        print("\n");
                        indent();
                    }
                }
                if (need_space && ch) {
                    need_space = false;
                    if (!/[\s;})]/.test(ch)) {
                        space();
                    }
                }
                newline_insert = -1;
                var prev = last.charAt(last.length - 1);
                if (might_need_semicolon) {
                    might_need_semicolon = false;
                    if (prev === ":" && ch === "}" || (!ch || !";}".includes(ch)) && prev !== ";") {
                        if (options.semicolons || requireSemicolonChars.has(ch)) {
                            OUTPUT += ";";
                            current_col++;
                            current_pos++;
                        }
                        else {
                            ensure_line_len();
                            if (current_col > 0) {
                                OUTPUT += "\n";
                                current_pos++;
                                current_line++;
                                current_col = 0;
                            }
                            if (/^\s+$/.test(str)) {
                                // reset the semicolon flag, since we didn't print one
                                // now and might still have to later
                                might_need_semicolon = true;
                            }
                        }
                        if (!options.beautify)
                            might_need_space = false;
                    }
                }
                if (might_need_space) {
                    if ((is_identifier_char(prev)
                        && (is_identifier_char(ch) || ch == "\\"))
                        || (ch == "/" && ch == prev)
                        || ((ch == "+" || ch == "-") && ch == last)) {
                        OUTPUT += " ";
                        current_col++;
                        current_pos++;
                    }
                    might_need_space = false;
                }
                if (mapping_token) {
                    mappings.push({
                        token: mapping_token,
                        name: mapping_name,
                        line: current_line,
                        col: current_col
                    });
                    mapping_token = false;
                    if (!might_add_newline)
                        do_add_mapping();
                }
                OUTPUT += str;
                has_parens = str[str.length - 1] == "(";
                current_pos += str.length;
                var a = str.split(/\r?\n/), n = a.length - 1;
                current_line += n;
                current_col += a[0].length;
                if (n > 0) {
                    ensure_line_len();
                    current_col = a[n].length;
                }
                last = str;
            }
            var star = function () {
                print("*");
            };
            var space = options.beautify ? function () {
                print(" ");
            } : function () {
                might_need_space = true;
            };
            var indent = options.beautify ? function (half) {
                if (options.beautify) {
                    print(make_indent(half ? 0.5 : 0));
                }
            } : noop;
            var with_indent = options.beautify ? function (col, cont) {
                if (col === true)
                    col = next_indent();
                var save_indentation = indentation;
                indentation = col;
                var ret = cont();
                indentation = save_indentation;
                return ret;
            } : function (col, cont) { return cont(); };
            var newline = options.beautify ? function () {
                if (newline_insert < 0)
                    return print("\n");
                if (OUTPUT[newline_insert] != "\n") {
                    OUTPUT = OUTPUT.slice(0, newline_insert) + "\n" + OUTPUT.slice(newline_insert);
                    current_pos++;
                    current_line++;
                }
                newline_insert++;
            } : options.max_line_len ? function () {
                ensure_line_len();
                might_add_newline = OUTPUT.length;
            } : noop;
            var semicolon = options.beautify ? function () {
                print(";");
            } : function () {
                might_need_semicolon = true;
            };
            function force_semicolon() {
                might_need_semicolon = false;
                print(";");
            }
            function next_indent() {
                return indentation + options.indent_level;
            }
            function with_block(cont) {
                var ret;
                print("{");
                newline();
                with_indent(next_indent(), function () {
                    ret = cont();
                });
                indent();
                print("}");
                return ret;
            }
            function with_parens(cont) {
                print("(");
                //XXX: still nice to have that for argument lists
                //var ret = with_indent(current_col, cont);
                var ret = cont();
                print(")");
                return ret;
            }
            function with_square(cont) {
                print("[");
                //var ret = with_indent(current_col, cont);
                var ret = cont();
                print("]");
                return ret;
            }
            function comma() {
                print(",");
                space();
            }
            function colon() {
                print(":");
                space();
            }
            var add_mapping = mappings ? function (token, name) {
                mapping_token = token;
                mapping_name = name;
            } : noop;
            function get() {
                if (might_add_newline) {
                    ensure_line_len();
                }
                return OUTPUT;
            }
            function has_nlb() {
                let n = OUTPUT.length - 1;
                while (n >= 0) {
                    const code = OUTPUT.charCodeAt(n);
                    if (code === CODE_LINE_BREAK) {
                        return true;
                    }
                    if (code !== CODE_SPACE) {
                        return false;
                    }
                    n--;
                }
                return true;
            }
            function filter_comment(comment) {
                if (!options.preserve_annotations) {
                    comment = comment.replace(r_annotation, " ");
                }
                if (/^\s*$/.test(comment)) {
                    return "";
                }
                return comment.replace(/(<\s*\/\s*)(script)/i, "<\\/$2");
            }
            function prepend_comments(node) {
                var self = this;
                var start = node.start;
                if (!start)
                    return;
                var printed_comments = self.printed_comments;
                // There cannot be a newline between return and its value.
                const return_with_value = node instanceof AST_Exit && node.value;
                if (start.comments_before
                    && printed_comments.has(start.comments_before)) {
                    if (return_with_value) {
                        start.comments_before = [];
                    }
                    else {
                        return;
                    }
                }
                var comments = start.comments_before;
                if (!comments) {
                    comments = start.comments_before = [];
                }
                printed_comments.add(comments);
                if (return_with_value) {
                    var tw = new TreeWalker(function (node) {
                        var parent = tw.parent();
                        if (parent instanceof AST_Exit
                            || parent instanceof AST_Binary && parent.left === node
                            || parent.TYPE == "Call" && parent.expression === node
                            || parent instanceof AST_Conditional && parent.condition === node
                            || parent instanceof AST_Dot && parent.expression === node
                            || parent instanceof AST_Sequence && parent.expressions[0] === node
                            || parent instanceof AST_Sub && parent.expression === node
                            || parent instanceof AST_UnaryPostfix) {
                            if (!node.start)
                                return;
                            var text = node.start.comments_before;
                            if (text && !printed_comments.has(text)) {
                                printed_comments.add(text);
                                comments = comments.concat(text);
                            }
                        }
                        else {
                            return true;
                        }
                    });
                    tw.push(node);
                    node.value.walk(tw);
                }
                if (current_pos == 0) {
                    if (comments.length > 0 && options.shebang && comments[0].type === "comment5"
                        && !printed_comments.has(comments[0])) {
                        print("#!" + comments.shift().value + "\n");
                        indent();
                    }
                    var preamble = options.preamble;
                    if (preamble) {
                        print(preamble.replace(/\r\n?|[\n\u2028\u2029]|\s*$/g, "\n"));
                    }
                }
                comments = comments.filter(comment_filter, node).filter(c => !printed_comments.has(c));
                if (comments.length == 0)
                    return;
                var last_nlb = has_nlb();
                comments.forEach(function (c, i) {
                    printed_comments.add(c);
                    if (!last_nlb) {
                        if (c.nlb) {
                            print("\n");
                            indent();
                            last_nlb = true;
                        }
                        else if (i > 0) {
                            space();
                        }
                    }
                    if (/comment[134]/.test(c.type)) {
                        var value = filter_comment(c.value);
                        if (value) {
                            print("//" + value + "\n");
                            indent();
                        }
                        last_nlb = true;
                    }
                    else if (c.type == "comment2") {
                        var value = filter_comment(c.value);
                        if (value) {
                            print("/*" + value + "*/");
                        }
                        last_nlb = false;
                    }
                });
                if (!last_nlb) {
                    if (start.nlb) {
                        print("\n");
                        indent();
                    }
                    else {
                        space();
                    }
                }
            }
            function append_comments(node, tail) {
                var self = this;
                var token = node.end;
                if (!token)
                    return;
                var printed_comments = self.printed_comments;
                var comments = token[tail ? "comments_before" : "comments_after"];
                if (!comments || printed_comments.has(comments))
                    return;
                if (!(node instanceof AST_Statement || comments.every((c) => !/comment[134]/.test(c.type))))
                    return;
                printed_comments.add(comments);
                var insert = OUTPUT.length;
                comments.filter(comment_filter, node).forEach(function (c, i) {
                    if (printed_comments.has(c))
                        return;
                    printed_comments.add(c);
                    need_space = false;
                    if (need_newline_indented) {
                        print("\n");
                        indent();
                        need_newline_indented = false;
                    }
                    else if (c.nlb && (i > 0 || !has_nlb())) {
                        print("\n");
                        indent();
                    }
                    else if (i > 0 || !tail) {
                        space();
                    }
                    if (/comment[134]/.test(c.type)) {
                        const value = filter_comment(c.value);
                        if (value) {
                            print("//" + value);
                        }
                        need_newline_indented = true;
                    }
                    else if (c.type == "comment2") {
                        const value = filter_comment(c.value);
                        if (value) {
                            print("/*" + value + "*/");
                        }
                        need_space = true;
                    }
                });
                if (OUTPUT.length > insert)
                    newline_insert = insert;
            }
            var stack = [];
            return {
                get: get,
                toString: get,
                indent: indent,
                in_directive: false,
                use_asm: null,
                active_scope: null,
                indentation: function () { return indentation; },
                current_width: function () { return current_col - indentation; },
                should_break: function () { return options.width && this.current_width() >= options.width; },
                has_parens: function () { return has_parens; },
                newline: newline,
                print: print,
                star: star,
                space: space,
                comma: comma,
                colon: colon,
                last: function () { return last; },
                semicolon: semicolon,
                force_semicolon: force_semicolon,
                to_utf8: to_utf8,
                print_name: function (name) { print(make_name(name)); },
                print_string: function (str, quote, escape_directive) {
                    var encoded = encode_string(str, quote);
                    if (escape_directive === true && !encoded.includes("\\")) {
                        // Insert semicolons to break directive prologue
                        if (!EXPECT_DIRECTIVE.test(OUTPUT)) {
                            force_semicolon();
                        }
                        force_semicolon();
                    }
                    print(encoded);
                },
                print_template_string_chars: function (str) {
                    var encoded = encode_string(str, "`").replace(/\${/g, "\\${");
                    return print(encoded.substr(1, encoded.length - 2));
                },
                encode_string: encode_string,
                next_indent: next_indent,
                with_indent: with_indent,
                with_block: with_block,
                with_parens: with_parens,
                with_square: with_square,
                add_mapping: add_mapping,
                option: function (opt) { return options[opt]; },
                printed_comments: printed_comments,
                prepend_comments: readonly ? noop : prepend_comments,
                append_comments: readonly || comment_filter === return_false ? noop : append_comments,
                line: function () { return current_line; },
                col: function () { return current_col; },
                pos: function () { return current_pos; },
                push_node: function (node) { stack.push(node); },
                pop_node: function () { return stack.pop(); },
                parent: function (n) {
                    return stack[stack.length - 2 - (n || 0)];
                }
            };
        }
        /* -----[ code generators ]----- */
        (function () {
            /* -----[ utils ]----- */
            function DEFPRINT(nodetype, generator) {
                nodetype.DEFMETHOD("_codegen", generator);
            }
            AST_Node.DEFMETHOD("print", function (output, force_parens) {
                var self = this, generator = self._codegen;
                if (self instanceof AST_Scope) {
                    output.active_scope = self;
                }
                else if (!output.use_asm && self instanceof AST_Directive && self.value == "use asm") {
                    output.use_asm = output.active_scope;
                }
                function doit() {
                    output.prepend_comments(self);
                    self.add_source_map(output);
                    generator(self, output);
                    output.append_comments(self);
                }
                output.push_node(self);
                if (force_parens || self.needs_parens(output)) {
                    output.with_parens(doit);
                }
                else {
                    doit();
                }
                output.pop_node();
                if (self === output.use_asm) {
                    output.use_asm = null;
                }
            });
            AST_Node.DEFMETHOD("_print", AST_Node.prototype.print);
            AST_Node.DEFMETHOD("print_to_string", function (options) {
                var output = OutputStream(options);
                this.print(output);
                return output.get();
            });
            /* -----[ PARENTHESES ]----- */
            function PARENS(nodetype, func) {
                if (Array.isArray(nodetype)) {
                    nodetype.forEach(function (nodetype) {
                        PARENS(nodetype, func);
                    });
                }
                else {
                    nodetype.DEFMETHOD("needs_parens", func);
                }
            }
            PARENS(AST_Node, return_false);
            // a function expression needs parens around it when it's provably
            // the first token to appear in a statement.
            PARENS(AST_Function, function (output) {
                if (!output.has_parens() && first_in_statement(output)) {
                    return true;
                }
                if (output.option("webkit")) {
                    var p = output.parent();
                    if (p instanceof AST_PropAccess && p.expression === this) {
                        return true;
                    }
                }
                if (output.option("wrap_iife")) {
                    var p = output.parent();
                    if (p instanceof AST_Call && p.expression === this) {
                        return true;
                    }
                }
                if (output.option("wrap_func_args")) {
                    var p = output.parent();
                    if (p instanceof AST_Call && p.args.includes(this)) {
                        return true;
                    }
                }
                return false;
            });
            PARENS(AST_Arrow, function (output) {
                var p = output.parent();
                if (output.option("wrap_func_args")
                    && p instanceof AST_Call
                    && p.args.includes(this)) {
                    return true;
                }
                return p instanceof AST_PropAccess && p.expression === this;
            });
            // same goes for an object literal (as in AST_Function), because
            // otherwise {...} would be interpreted as a block of code.
            PARENS(AST_Object, function (output) {
                return !output.has_parens() && first_in_statement(output);
            });
            PARENS(AST_ClassExpression, first_in_statement);
            PARENS(AST_Unary, function (output) {
                var p = output.parent();
                return p instanceof AST_PropAccess && p.expression === this
                    || p instanceof AST_Call && p.expression === this
                    || p instanceof AST_Binary
                        && p.operator === "**"
                        && this instanceof AST_UnaryPrefix
                        && p.left === this
                        && this.operator !== "++"
                        && this.operator !== "--";
            });
            PARENS(AST_Await, function (output) {
                var p = output.parent();
                return p instanceof AST_PropAccess && p.expression === this
                    || p instanceof AST_Call && p.expression === this
                    || p instanceof AST_Binary && p.operator === "**" && p.left === this
                    || output.option("safari10") && p instanceof AST_UnaryPrefix;
            });
            PARENS(AST_Sequence, function (output) {
                var p = output.parent();
                return p instanceof AST_Call // (foo, bar)() or foo(1, (2, 3), 4)
                    || p instanceof AST_Unary // !(foo, bar, baz)
                    || p instanceof AST_Binary // 1 + (2, 3) + 4 ==> 8
                    || p instanceof AST_VarDef // var a = (1, 2), b = a + a; ==> b == 4
                    || p instanceof AST_PropAccess // (1, {foo:2}).foo or (1, {foo:2})["foo"] ==> 2
                    || p instanceof AST_Array // [ 1, (2, 3), 4 ] ==> [ 1, 3, 4 ]
                    || p instanceof AST_ObjectProperty // { foo: (1, 2) }.foo ==> 2
                    || p instanceof AST_Conditional /* (false, true) ? (a = 10, b = 20) : (c = 30)
                                                                   * ==> 20 (side effect, set a := 10 and b := 20) */
                    || p instanceof AST_Arrow // x => (x, x)
                    || p instanceof AST_DefaultAssign // x => (x = (0, function(){}))
                    || p instanceof AST_Expansion // [...(a, b)]
                    || p instanceof AST_ForOf && this === p.object // for (e of (foo, bar)) {}
                    || p instanceof AST_Yield // yield (foo, bar)
                    || p instanceof AST_Export // export default (foo, bar)
                ;
            });
            PARENS(AST_Binary, function (output) {
                var p = output.parent();
                // (foo && bar)()
                if (p instanceof AST_Call && p.expression === this)
                    return true;
                // typeof (foo && bar)
                if (p instanceof AST_Unary)
                    return true;
                // (foo && bar)["prop"], (foo && bar).prop
                if (p instanceof AST_PropAccess && p.expression === this)
                    return true;
                // this deals with precedence: 3 * (2 + 1)
                if (p instanceof AST_Binary) {
                    const po = p.operator;
                    const so = this.operator;
                    if (so === "??" && (po === "||" || po === "&&")) {
                        return true;
                    }
                    if (po === "??" && (so === "||" || so === "&&")) {
                        return true;
                    }
                    const pp = PRECEDENCE[po];
                    const sp = PRECEDENCE[so];
                    if (pp > sp
                        || (pp == sp
                            && (this === p.right || po == "**"))) {
                        return true;
                    }
                }
            });
            PARENS(AST_Yield, function (output) {
                var p = output.parent();
                // (yield 1) + (yield 2)
                // a = yield 3
                if (p instanceof AST_Binary && p.operator !== "=")
                    return true;
                // (yield 1)()
                // new (yield 1)()
                if (p instanceof AST_Call && p.expression === this)
                    return true;
                // (yield 1) ? yield 2 : yield 3
                if (p instanceof AST_Conditional && p.condition === this)
                    return true;
                // -(yield 4)
                if (p instanceof AST_Unary)
                    return true;
                // (yield x).foo
                // (yield x)['foo']
                if (p instanceof AST_PropAccess && p.expression === this)
                    return true;
            });
            PARENS(AST_PropAccess, function (output) {
                var p = output.parent();
                if (p instanceof AST_New && p.expression === this) {
                    // i.e. new (foo.bar().baz)
                    //
                    // if there's one call into this subtree, then we need
                    // parens around it too, otherwise the call will be
                    // interpreted as passing the arguments to the upper New
                    // expression.
                    return walk(this, node => {
                        if (node instanceof AST_Scope)
                            return true;
                        if (node instanceof AST_Call) {
                            return walk_abort; // makes walk() return true.
                        }
                    });
                }
            });
            PARENS(AST_Call, function (output) {
                var p = output.parent(), p1;
                if (p instanceof AST_New && p.expression === this
                    || p instanceof AST_Export && p.is_default && this.expression instanceof AST_Function)
                    return true;
                // workaround for Safari bug.
                // https://bugs.webkit.org/show_bug.cgi?id=123506
                return this.expression instanceof AST_Function
                    && p instanceof AST_PropAccess
                    && p.expression === this
                    && (p1 = output.parent(1)) instanceof AST_Assign
                    && p1.left === p;
            });
            PARENS(AST_New, function (output) {
                var p = output.parent();
                if (this.args.length === 0
                    && (p instanceof AST_PropAccess // (new Date).getTime(), (new Date)["getTime"]()
                        || p instanceof AST_Call && p.expression === this)) // (new foo)(bar)
                    return true;
            });
            PARENS(AST_Number, function (output) {
                var p = output.parent();
                if (p instanceof AST_PropAccess && p.expression === this) {
                    var value = this.getValue();
                    if (value < 0 || /^0/.test(make_num(value))) {
                        return true;
                    }
                }
            });
            PARENS(AST_BigInt, function (output) {
                var p = output.parent();
                if (p instanceof AST_PropAccess && p.expression === this) {
                    var value = this.getValue();
                    if (value.startsWith("-")) {
                        return true;
                    }
                }
            });
            PARENS([AST_Assign, AST_Conditional], function (output) {
                var p = output.parent();
                // !(a = false) → true
                if (p instanceof AST_Unary)
                    return true;
                // 1 + (a = 2) + 3 → 6, side effect setting a = 2
                if (p instanceof AST_Binary && !(p instanceof AST_Assign))
                    return true;
                // (a = func)() —or— new (a = Object)()
                if (p instanceof AST_Call && p.expression === this)
                    return true;
                // (a = foo) ? bar : baz
                if (p instanceof AST_Conditional && p.condition === this)
                    return true;
                // (a = foo)["prop"] —or— (a = foo).prop
                if (p instanceof AST_PropAccess && p.expression === this)
                    return true;
                // ({a, b} = {a: 1, b: 2}), a destructuring assignment
                if (this instanceof AST_Assign && this.left instanceof AST_Destructuring && this.left.is_array === false)
                    return true;
            });
            /* -----[ PRINTERS ]----- */
            DEFPRINT(AST_Directive, function (self, output) {
                output.print_string(self.value, self.quote);
                output.semicolon();
            });
            DEFPRINT(AST_Expansion, function (self, output) {
                output.print("...");
                self.expression.print(output);
            });
            DEFPRINT(AST_Destructuring, function (self, output) {
                output.print(self.is_array ? "[" : "{");
                var len = self.names.length;
                self.names.forEach(function (name, i) {
                    if (i > 0)
                        output.comma();
                    name.print(output);
                    // If the final element is a hole, we need to make sure it
                    // doesn't look like a trailing comma, by inserting an actual
                    // trailing comma.
                    if (i == len - 1 && name instanceof AST_Hole)
                        output.comma();
                });
                output.print(self.is_array ? "]" : "}");
            });
            DEFPRINT(AST_Debugger, function (self, output) {
                output.print("debugger");
                output.semicolon();
            });
            /* -----[ statements ]----- */
            function display_body(body, is_toplevel, output, allow_directives) {
                var last = body.length - 1;
                output.in_directive = allow_directives;
                body.forEach(function (stmt, i) {
                    if (output.in_directive === true && !(stmt instanceof AST_Directive ||
                        stmt instanceof AST_EmptyStatement ||
                        (stmt instanceof AST_SimpleStatement && stmt.body instanceof AST_String))) {
                        output.in_directive = false;
                    }
                    if (!(stmt instanceof AST_EmptyStatement)) {
                        output.indent();
                        stmt.print(output);
                        if (!(i == last && is_toplevel)) {
                            output.newline();
                            if (is_toplevel)
                                output.newline();
                        }
                    }
                    if (output.in_directive === true &&
                        stmt instanceof AST_SimpleStatement &&
                        stmt.body instanceof AST_String) {
                        output.in_directive = false;
                    }
                });
                output.in_directive = false;
            }
            AST_StatementWithBody.DEFMETHOD("_do_print_body", function (output) {
                force_statement(this.body, output);
            });
            DEFPRINT(AST_Statement, function (self, output) {
                self.body.print(output);
                output.semicolon();
            });
            DEFPRINT(AST_Toplevel, function (self, output) {
                display_body(self.body, true, output, true);
                output.print("");
            });
            DEFPRINT(AST_LabeledStatement, function (self, output) {
                self.label.print(output);
                output.colon();
                self.body.print(output);
            });
            DEFPRINT(AST_SimpleStatement, function (self, output) {
                self.body.print(output);
                output.semicolon();
            });
            function print_braced_empty(self, output) {
                output.print("{");
                output.with_indent(output.next_indent(), function () {
                    output.append_comments(self, true);
                });
                output.print("}");
            }
            function print_braced(self, output, allow_directives) {
                if (self.body.length > 0) {
                    output.with_block(function () {
                        display_body(self.body, false, output, allow_directives);
                    });
                }
                else
                    print_braced_empty(self, output);
            }
            DEFPRINT(AST_BlockStatement, function (self, output) {
                print_braced(self, output);
            });
            DEFPRINT(AST_EmptyStatement, function (self, output) {
                output.semicolon();
            });
            DEFPRINT(AST_Do, function (self, output) {
                output.print("do");
                output.space();
                make_block(self.body, output);
                output.space();
                output.print("while");
                output.space();
                output.with_parens(function () {
                    self.condition.print(output);
                });
                output.semicolon();
            });
            DEFPRINT(AST_While, function (self, output) {
                output.print("while");
                output.space();
                output.with_parens(function () {
                    self.condition.print(output);
                });
                output.space();
                self._do_print_body(output);
            });
            DEFPRINT(AST_For, function (self, output) {
                output.print("for");
                output.space();
                output.with_parens(function () {
                    if (self.init) {
                        if (self.init instanceof AST_Definitions) {
                            self.init.print(output);
                        }
                        else {
                            parenthesize_for_noin(self.init, output, true);
                        }
                        output.print(";");
                        output.space();
                    }
                    else {
                        output.print(";");
                    }
                    if (self.condition) {
                        self.condition.print(output);
                        output.print(";");
                        output.space();
                    }
                    else {
                        output.print(";");
                    }
                    if (self.step) {
                        self.step.print(output);
                    }
                });
                output.space();
                self._do_print_body(output);
            });
            DEFPRINT(AST_ForIn, function (self, output) {
                output.print("for");
                if (self.await) {
                    output.space();
                    output.print("await");
                }
                output.space();
                output.with_parens(function () {
                    self.init.print(output);
                    output.space();
                    output.print(self instanceof AST_ForOf ? "of" : "in");
                    output.space();
                    self.object.print(output);
                });
                output.space();
                self._do_print_body(output);
            });
            DEFPRINT(AST_With, function (self, output) {
                output.print("with");
                output.space();
                output.with_parens(function () {
                    self.expression.print(output);
                });
                output.space();
                self._do_print_body(output);
            });
            /* -----[ functions ]----- */
            AST_Lambda.DEFMETHOD("_do_print", function (output, nokeyword) {
                var self = this;
                if (!nokeyword) {
                    if (self.async) {
                        output.print("async");
                        output.space();
                    }
                    output.print("function");
                    if (self.is_generator) {
                        output.star();
                    }
                    if (self.name) {
                        output.space();
                    }
                }
                if (self.name instanceof AST_Symbol) {
                    self.name.print(output);
                }
                else if (nokeyword && self.name instanceof AST_Node) {
                    output.with_square(function () {
                        self.name.print(output); // Computed method name
                    });
                }
                output.with_parens(function () {
                    self.argnames.forEach(function (arg, i) {
                        if (i)
                            output.comma();
                        arg.print(output);
                    });
                });
                output.space();
                print_braced(self, output, true);
            });
            DEFPRINT(AST_Lambda, function (self, output) {
                self._do_print(output);
            });
            DEFPRINT(AST_PrefixedTemplateString, function (self, output) {
                var tag = self.prefix;
                var parenthesize_tag = tag instanceof AST_Lambda
                    || tag instanceof AST_Binary
                    || tag instanceof AST_Conditional
                    || tag instanceof AST_Sequence
                    || tag instanceof AST_Unary
                    || tag instanceof AST_Dot && tag.expression instanceof AST_Object;
                if (parenthesize_tag)
                    output.print("(");
                self.prefix.print(output);
                if (parenthesize_tag)
                    output.print(")");
                self.template_string.print(output);
            });
            DEFPRINT(AST_TemplateString, function (self, output) {
                var is_tagged = output.parent() instanceof AST_PrefixedTemplateString;
                output.print("`");
                for (var i = 0; i < self.segments.length; i++) {
                    if (!(self.segments[i] instanceof AST_TemplateSegment)) {
                        output.print("${");
                        self.segments[i].print(output);
                        output.print("}");
                    }
                    else if (is_tagged) {
                        output.print(self.segments[i].raw);
                    }
                    else {
                        output.print_template_string_chars(self.segments[i].value);
                    }
                }
                output.print("`");
            });
            DEFPRINT(AST_TemplateSegment, function (self, output) {
                output.print_template_string_chars(self.value);
            });
            AST_Arrow.DEFMETHOD("_do_print", function (output) {
                var self = this;
                var parent = output.parent();
                var needs_parens = (parent instanceof AST_Binary && !(parent instanceof AST_Assign)) ||
                    parent instanceof AST_Unary ||
                    (parent instanceof AST_Call && self === parent.expression);
                if (needs_parens) {
                    output.print("(");
                }
                if (self.async) {
                    output.print("async");
                    output.space();
                }
                if (self.argnames.length === 1 && self.argnames[0] instanceof AST_Symbol) {
                    self.argnames[0].print(output);
                }
                else {
                    output.with_parens(function () {
                        self.argnames.forEach(function (arg, i) {
                            if (i)
                                output.comma();
                            arg.print(output);
                        });
                    });
                }
                output.space();
                output.print("=>");
                output.space();
                const first_statement = self.body[0];
                if (self.body.length === 1
                    && first_statement instanceof AST_Return) {
                    const returned = first_statement.value;
                    if (!returned) {
                        output.print("{}");
                    }
                    else if (left_is_object(returned)) {
                        output.print("(");
                        returned.print(output);
                        output.print(")");
                    }
                    else {
                        returned.print(output);
                    }
                }
                else {
                    print_braced(self, output);
                }
                if (needs_parens) {
                    output.print(")");
                }
            });
            /* -----[ exits ]----- */
            AST_Exit.DEFMETHOD("_do_print", function (output, kind) {
                output.print(kind);
                if (this.value) {
                    output.space();
                    const comments = this.value.start.comments_before;
                    if (comments && comments.length && !output.printed_comments.has(comments)) {
                        output.print("(");
                        this.value.print(output);
                        output.print(")");
                    }
                    else {
                        this.value.print(output);
                    }
                }
                output.semicolon();
            });
            DEFPRINT(AST_Return, function (self, output) {
                self._do_print(output, "return");
            });
            DEFPRINT(AST_Throw, function (self, output) {
                self._do_print(output, "throw");
            });
            /* -----[ yield ]----- */
            DEFPRINT(AST_Yield, function (self, output) {
                var star = self.is_star ? "*" : "";
                output.print("yield" + star);
                if (self.expression) {
                    output.space();
                    self.expression.print(output);
                }
            });
            DEFPRINT(AST_Await, function (self, output) {
                output.print("await");
                output.space();
                var e = self.expression;
                var parens = !(e instanceof AST_Call
                    || e instanceof AST_SymbolRef
                    || e instanceof AST_PropAccess
                    || e instanceof AST_Unary
                    || e instanceof AST_Constant
                    || e instanceof AST_Await
                    || e instanceof AST_Object);
                if (parens)
                    output.print("(");
                self.expression.print(output);
                if (parens)
                    output.print(")");
            });
            /* -----[ loop control ]----- */
            AST_LoopControl.DEFMETHOD("_do_print", function (output, kind) {
                output.print(kind);
                if (this.label) {
                    output.space();
                    this.label.print(output);
                }
                output.semicolon();
            });
            DEFPRINT(AST_Break, function (self, output) {
                self._do_print(output, "break");
            });
            DEFPRINT(AST_Continue, function (self, output) {
                self._do_print(output, "continue");
            });
            /* -----[ if ]----- */
            function make_then(self, output) {
                var b = self.body;
                if (output.option("braces")
                    || output.option("ie8") && b instanceof AST_Do)
                    return make_block(b, output);
                // The squeezer replaces "block"-s that contain only a single
                // statement with the statement itself; technically, the AST
                // is correct, but this can create problems when we output an
                // IF having an ELSE clause where the THEN clause ends in an
                // IF *without* an ELSE block (then the outer ELSE would refer
                // to the inner IF).  This function checks for this case and
                // adds the block braces if needed.
                if (!b)
                    return output.force_semicolon();
                while (true) {
                    if (b instanceof AST_If) {
                        if (!b.alternative) {
                            make_block(self.body, output);
                            return;
                        }
                        b = b.alternative;
                    }
                    else if (b instanceof AST_StatementWithBody) {
                        b = b.body;
                    }
                    else
                        break;
                }
                force_statement(self.body, output);
            }
            DEFPRINT(AST_If, function (self, output) {
                output.print("if");
                output.space();
                output.with_parens(function () {
                    self.condition.print(output);
                });
                output.space();
                if (self.alternative) {
                    make_then(self, output);
                    output.space();
                    output.print("else");
                    output.space();
                    if (self.alternative instanceof AST_If)
                        self.alternative.print(output);
                    else
                        force_statement(self.alternative, output);
                }
                else {
                    self._do_print_body(output);
                }
            });
            /* -----[ switch ]----- */
            DEFPRINT(AST_Switch, function (self, output) {
                output.print("switch");
                output.space();
                output.with_parens(function () {
                    self.expression.print(output);
                });
                output.space();
                var last = self.body.length - 1;
                if (last < 0)
                    print_braced_empty(self, output);
                else
                    output.with_block(function () {
                        self.body.forEach(function (branch, i) {
                            output.indent(true);
                            branch.print(output);
                            if (i < last && branch.body.length > 0)
                                output.newline();
                        });
                    });
            });
            AST_SwitchBranch.DEFMETHOD("_do_print_body", function (output) {
                output.newline();
                this.body.forEach(function (stmt) {
                    output.indent();
                    stmt.print(output);
                    output.newline();
                });
            });
            DEFPRINT(AST_Default, function (self, output) {
                output.print("default:");
                self._do_print_body(output);
            });
            DEFPRINT(AST_Case, function (self, output) {
                output.print("case");
                output.space();
                self.expression.print(output);
                output.print(":");
                self._do_print_body(output);
            });
            /* -----[ exceptions ]----- */
            DEFPRINT(AST_Try, function (self, output) {
                output.print("try");
                output.space();
                print_braced(self, output);
                if (self.bcatch) {
                    output.space();
                    self.bcatch.print(output);
                }
                if (self.bfinally) {
                    output.space();
                    self.bfinally.print(output);
                }
            });
            DEFPRINT(AST_Catch, function (self, output) {
                output.print("catch");
                if (self.argname) {
                    output.space();
                    output.with_parens(function () {
                        self.argname.print(output);
                    });
                }
                output.space();
                print_braced(self, output);
            });
            DEFPRINT(AST_Finally, function (self, output) {
                output.print("finally");
                output.space();
                print_braced(self, output);
            });
            /* -----[ var/const ]----- */
            AST_Definitions.DEFMETHOD("_do_print", function (output, kind) {
                output.print(kind);
                output.space();
                this.definitions.forEach(function (def, i) {
                    if (i)
                        output.comma();
                    def.print(output);
                });
                var p = output.parent();
                var in_for = p instanceof AST_For || p instanceof AST_ForIn;
                var output_semicolon = !in_for || p && p.init !== this;
                if (output_semicolon)
                    output.semicolon();
            });
            DEFPRINT(AST_Let, function (self, output) {
                self._do_print(output, "let");
            });
            DEFPRINT(AST_Var, function (self, output) {
                self._do_print(output, "var");
            });
            DEFPRINT(AST_Const, function (self, output) {
                self._do_print(output, "const");
            });
            DEFPRINT(AST_Import, function (self, output) {
                output.print("import");
                output.space();
                if (self.imported_name) {
                    self.imported_name.print(output);
                }
                if (self.imported_name && self.imported_names) {
                    output.print(",");
                    output.space();
                }
                if (self.imported_names) {
                    if (self.imported_names.length === 1 && self.imported_names[0].foreign_name.name === "*") {
                        self.imported_names[0].print(output);
                    }
                    else {
                        output.print("{");
                        self.imported_names.forEach(function (name_import, i) {
                            output.space();
                            name_import.print(output);
                            if (i < self.imported_names.length - 1) {
                                output.print(",");
                            }
                        });
                        output.space();
                        output.print("}");
                    }
                }
                if (self.imported_name || self.imported_names) {
                    output.space();
                    output.print("from");
                    output.space();
                }
                self.module_name.print(output);
                output.semicolon();
            });
            DEFPRINT(AST_ImportMeta, function (self, output) {
                output.print("import.meta");
            });
            DEFPRINT(AST_NameMapping, function (self, output) {
                var is_import = output.parent() instanceof AST_Import;
                var definition = self.name.definition();
                var names_are_different = (definition && definition.mangled_name || self.name.name) !==
                    self.foreign_name.name;
                if (names_are_different) {
                    if (is_import) {
                        output.print(self.foreign_name.name);
                    }
                    else {
                        self.name.print(output);
                    }
                    output.space();
                    output.print("as");
                    output.space();
                    if (is_import) {
                        self.name.print(output);
                    }
                    else {
                        output.print(self.foreign_name.name);
                    }
                }
                else {
                    self.name.print(output);
                }
            });
            DEFPRINT(AST_Export, function (self, output) {
                output.print("export");
                output.space();
                if (self.is_default) {
                    output.print("default");
                    output.space();
                }
                if (self.exported_names) {
                    if (self.exported_names.length === 1 && self.exported_names[0].name.name === "*") {
                        self.exported_names[0].print(output);
                    }
                    else {
                        output.print("{");
                        self.exported_names.forEach(function (name_export, i) {
                            output.space();
                            name_export.print(output);
                            if (i < self.exported_names.length - 1) {
                                output.print(",");
                            }
                        });
                        output.space();
                        output.print("}");
                    }
                }
                else if (self.exported_value) {
                    self.exported_value.print(output);
                }
                else if (self.exported_definition) {
                    self.exported_definition.print(output);
                    if (self.exported_definition instanceof AST_Definitions)
                        return;
                }
                if (self.module_name) {
                    output.space();
                    output.print("from");
                    output.space();
                    self.module_name.print(output);
                }
                if (self.exported_value
                    && !(self.exported_value instanceof AST_Defun ||
                        self.exported_value instanceof AST_Function ||
                        self.exported_value instanceof AST_Class)
                    || self.module_name
                    || self.exported_names) {
                    output.semicolon();
                }
            });
            function parenthesize_for_noin(node, output, noin) {
                var parens = false;
                // need to take some precautions here:
                //    https://github.com/mishoo/UglifyJS2/issues/60
                if (noin) {
                    parens = walk(node, node => {
                        if (node instanceof AST_Scope)
                            return true;
                        if (node instanceof AST_Binary && node.operator == "in") {
                            return walk_abort; // makes walk() return true
                        }
                    });
                }
                node.print(output, parens);
            }
            DEFPRINT(AST_VarDef, function (self, output) {
                self.name.print(output);
                if (self.value) {
                    output.space();
                    output.print("=");
                    output.space();
                    var p = output.parent(1);
                    var noin = p instanceof AST_For || p instanceof AST_ForIn;
                    parenthesize_for_noin(self.value, output, noin);
                }
            });
            /* -----[ other expressions ]----- */
            DEFPRINT(AST_Call, function (self, output) {
                self.expression.print(output);
                if (self instanceof AST_New && self.args.length === 0)
                    return;
                if (self.expression instanceof AST_Call || self.expression instanceof AST_Lambda) {
                    output.add_mapping(self.start);
                }
                if (self.optional)
                    output.print("?.");
                output.with_parens(function () {
                    self.args.forEach(function (expr, i) {
                        if (i)
                            output.comma();
                        expr.print(output);
                    });
                });
            });
            DEFPRINT(AST_New, function (self, output) {
                output.print("new");
                output.space();
                AST_Call.prototype._codegen(self, output);
            });
            AST_Sequence.DEFMETHOD("_do_print", function (output) {
                this.expressions.forEach(function (node, index) {
                    if (index > 0) {
                        output.comma();
                        if (output.should_break()) {
                            output.newline();
                            output.indent();
                        }
                    }
                    node.print(output);
                });
            });
            DEFPRINT(AST_Sequence, function (self, output) {
                self._do_print(output);
                // var p = output.parent();
                // if (p instanceof AST_Statement) {
                //     output.with_indent(output.next_indent(), function(){
                //         self._do_print(output);
                //     });
                // } else {
                //     self._do_print(output);
                // }
            });
            DEFPRINT(AST_Dot, function (self, output) {
                var expr = self.expression;
                expr.print(output);
                var prop = self.property;
                var print_computed = RESERVED_WORDS.has(prop)
                    ? output.option("ie8")
                    : !is_identifier_string(prop, output.option("ecma") >= 2015 || output.option("safari10"));
                if (self.optional)
                    output.print("?.");
                if (print_computed) {
                    output.print("[");
                    output.add_mapping(self.end);
                    output.print_string(prop);
                    output.print("]");
                }
                else {
                    if (expr instanceof AST_Number && expr.getValue() >= 0) {
                        if (!/[xa-f.)]/i.test(output.last())) {
                            output.print(".");
                        }
                    }
                    if (!self.optional)
                        output.print(".");
                    // the name after dot would be mapped about here.
                    output.add_mapping(self.end);
                    output.print_name(prop);
                }
            });
            DEFPRINT(AST_DotHash, function (self, output) {
                var expr = self.expression;
                expr.print(output);
                var prop = self.property;
                if (self.optional)
                    output.print("?");
                output.print(".#");
                output.print_name(prop);
            });
            DEFPRINT(AST_Sub, function (self, output) {
                self.expression.print(output);
                if (self.optional)
                    output.print("?.");
                output.print("[");
                self.property.print(output);
                output.print("]");
            });
            DEFPRINT(AST_Chain, function (self, output) {
                self.expression.print(output);
            });
            DEFPRINT(AST_UnaryPrefix, function (self, output) {
                var op = self.operator;
                output.print(op);
                if (/^[a-z]/i.test(op)
                    || (/[+-]$/.test(op)
                        && self.expression instanceof AST_UnaryPrefix
                        && /^[+-]/.test(self.expression.operator))) {
                    output.space();
                }
                self.expression.print(output);
            });
            DEFPRINT(AST_UnaryPostfix, function (self, output) {
                self.expression.print(output);
                output.print(self.operator);
            });
            DEFPRINT(AST_Binary, function (self, output) {
                var op = self.operator;
                self.left.print(output);
                if (op[0] == ">" /* ">>" ">>>" ">" ">=" */
                    && self.left instanceof AST_UnaryPostfix
                    && self.left.operator == "--") {
                    // space is mandatory to avoid outputting -->
                    output.print(" ");
                }
                else {
                    // the space is optional depending on "beautify"
                    output.space();
                }
                output.print(op);
                if ((op == "<" || op == "<<")
                    && self.right instanceof AST_UnaryPrefix
                    && self.right.operator == "!"
                    && self.right.expression instanceof AST_UnaryPrefix
                    && self.right.expression.operator == "--") {
                    // space is mandatory to avoid outputting <!--
                    output.print(" ");
                }
                else {
                    // the space is optional depending on "beautify"
                    output.space();
                }
                self.right.print(output);
            });
            DEFPRINT(AST_Conditional, function (self, output) {
                self.condition.print(output);
                output.space();
                output.print("?");
                output.space();
                self.consequent.print(output);
                output.space();
                output.colon();
                self.alternative.print(output);
            });
            /* -----[ literals ]----- */
            DEFPRINT(AST_Array, function (self, output) {
                output.with_square(function () {
                    var a = self.elements, len = a.length;
                    if (len > 0)
                        output.space();
                    a.forEach(function (exp, i) {
                        if (i)
                            output.comma();
                        exp.print(output);
                        // If the final element is a hole, we need to make sure it
                        // doesn't look like a trailing comma, by inserting an actual
                        // trailing comma.
                        if (i === len - 1 && exp instanceof AST_Hole)
                            output.comma();
                    });
                    if (len > 0)
                        output.space();
                });
            });
            DEFPRINT(AST_Object, function (self, output) {
                if (self.properties.length > 0)
                    output.with_block(function () {
                        self.properties.forEach(function (prop, i) {
                            if (i) {
                                output.print(",");
                                output.newline();
                            }
                            output.indent();
                            prop.print(output);
                        });
                        output.newline();
                    });
                else
                    print_braced_empty(self, output);
            });
            DEFPRINT(AST_Class, function (self, output) {
                output.print("class");
                output.space();
                if (self.name) {
                    self.name.print(output);
                    output.space();
                }
                if (self.extends) {
                    var parens = (!(self.extends instanceof AST_SymbolRef)
                        && !(self.extends instanceof AST_PropAccess)
                        && !(self.extends instanceof AST_ClassExpression)
                        && !(self.extends instanceof AST_Function));
                    output.print("extends");
                    if (parens) {
                        output.print("(");
                    }
                    else {
                        output.space();
                    }
                    self.extends.print(output);
                    if (parens) {
                        output.print(")");
                    }
                    else {
                        output.space();
                    }
                }
                if (self.properties.length > 0)
                    output.with_block(function () {
                        self.properties.forEach(function (prop, i) {
                            if (i) {
                                output.newline();
                            }
                            output.indent();
                            prop.print(output);
                        });
                        output.newline();
                    });
                else
                    output.print("{}");
            });
            DEFPRINT(AST_NewTarget, function (self, output) {
                output.print("new.target");
            });
            function print_property_name(key, quote, output) {
                if (output.option("quote_keys")) {
                    return output.print_string(key);
                }
                if ("" + +key == key && key >= 0) {
                    if (output.option("keep_numbers")) {
                        return output.print(key);
                    }
                    return output.print(make_num(key));
                }
                var print_string = RESERVED_WORDS.has(key)
                    ? output.option("ie8")
                    : (output.option("ecma") < 2015 || output.option("safari10")
                        ? !is_basic_identifier_string(key)
                        : !is_identifier_string(key, true));
                if (print_string || (quote && output.option("keep_quoted_props"))) {
                    return output.print_string(key, quote);
                }
                return output.print_name(key);
            }
            DEFPRINT(AST_ObjectKeyVal, function (self, output) {
                function get_name(self) {
                    var def = self.definition();
                    return def ? def.mangled_name || def.name : self.name;
                }
                var allowShortHand = output.option("shorthand");
                if (allowShortHand &&
                    self.value instanceof AST_Symbol &&
                    is_identifier_string(self.key, output.option("ecma") >= 2015 || output.option("safari10")) &&
                    get_name(self.value) === self.key &&
                    !RESERVED_WORDS.has(self.key)) {
                    print_property_name(self.key, self.quote, output);
                }
                else if (allowShortHand &&
                    self.value instanceof AST_DefaultAssign &&
                    self.value.left instanceof AST_Symbol &&
                    is_identifier_string(self.key, output.option("ecma") >= 2015 || output.option("safari10")) &&
                    get_name(self.value.left) === self.key) {
                    print_property_name(self.key, self.quote, output);
                    output.space();
                    output.print("=");
                    output.space();
                    self.value.right.print(output);
                }
                else {
                    if (!(self.key instanceof AST_Node)) {
                        print_property_name(self.key, self.quote, output);
                    }
                    else {
                        output.with_square(function () {
                            self.key.print(output);
                        });
                    }
                    output.colon();
                    self.value.print(output);
                }
            });
            DEFPRINT(AST_ClassPrivateProperty, (self, output) => {
                if (self.static) {
                    output.print("static");
                    output.space();
                }
                output.print("#");
                print_property_name(self.key.name, self.quote, output);
                if (self.value) {
                    output.print("=");
                    self.value.print(output);
                }
                output.semicolon();
            });
            DEFPRINT(AST_ClassProperty, (self, output) => {
                if (self.static) {
                    output.print("static");
                    output.space();
                }
                if (self.key instanceof AST_SymbolClassProperty) {
                    print_property_name(self.key.name, self.quote, output);
                }
                else {
                    output.print("[");
                    self.key.print(output);
                    output.print("]");
                }
                if (self.value) {
                    output.print("=");
                    self.value.print(output);
                }
                output.semicolon();
            });
            AST_ObjectProperty.DEFMETHOD("_print_getter_setter", function (type, is_private, output) {
                var self = this;
                if (self.static) {
                    output.print("static");
                    output.space();
                }
                if (type) {
                    output.print(type);
                    output.space();
                }
                if (self.key instanceof AST_SymbolMethod) {
                    if (is_private)
                        output.print("#");
                    print_property_name(self.key.name, self.quote, output);
                }
                else {
                    output.with_square(function () {
                        self.key.print(output);
                    });
                }
                self.value._do_print(output, true);
            });
            DEFPRINT(AST_ObjectSetter, function (self, output) {
                self._print_getter_setter("set", false, output);
            });
            DEFPRINT(AST_ObjectGetter, function (self, output) {
                self._print_getter_setter("get", false, output);
            });
            DEFPRINT(AST_PrivateSetter, function (self, output) {
                self._print_getter_setter("set", true, output);
            });
            DEFPRINT(AST_PrivateGetter, function (self, output) {
                self._print_getter_setter("get", true, output);
            });
            DEFPRINT(AST_PrivateMethod, function (self, output) {
                var type;
                if (self.is_generator && self.async) {
                    type = "async*";
                }
                else if (self.is_generator) {
                    type = "*";
                }
                else if (self.async) {
                    type = "async";
                }
                self._print_getter_setter(type, true, output);
            });
            DEFPRINT(AST_ConciseMethod, function (self, output) {
                var type;
                if (self.is_generator && self.async) {
                    type = "async*";
                }
                else if (self.is_generator) {
                    type = "*";
                }
                else if (self.async) {
                    type = "async";
                }
                self._print_getter_setter(type, false, output);
            });
            AST_Symbol.DEFMETHOD("_do_print", function (output) {
                var def = this.definition();
                output.print_name(def ? def.mangled_name || def.name : this.name);
            });
            DEFPRINT(AST_Symbol, function (self, output) {
                self._do_print(output);
            });
            DEFPRINT(AST_Hole, noop);
            DEFPRINT(AST_This, function (self, output) {
                output.print("this");
            });
            DEFPRINT(AST_Super, function (self, output) {
                output.print("super");
            });
            DEFPRINT(AST_Constant, function (self, output) {
                output.print(self.getValue());
            });
            DEFPRINT(AST_String, function (self, output) {
                output.print_string(self.getValue(), self.quote, output.in_directive);
            });
            DEFPRINT(AST_Number, function (self, output) {
                if ((output.option("keep_numbers") || output.use_asm) && self.raw) {
                    output.print(self.raw);
                }
                else {
                    output.print(make_num(self.getValue()));
                }
            });
            DEFPRINT(AST_BigInt, function (self, output) {
                output.print(self.getValue() + "n");
            });
            const r_slash_script = /(<\s*\/\s*script)/i;
            const slash_script_replace = (_, $1) => $1.replace("/", "\\/");
            DEFPRINT(AST_RegExp, function (self, output) {
                let { source, flags } = self.getValue();
                source = regexp_source_fix(source);
                flags = flags ? sort_regexp_flags(flags) : "";
                source = source.replace(r_slash_script, slash_script_replace);
                output.print(output.to_utf8(`/${source}/${flags}`));
                const parent = output.parent();
                if (parent instanceof AST_Binary
                    && /^\w/.test(parent.operator)
                    && parent.left === self) {
                    output.print(" ");
                }
            });
            function force_statement(stat, output) {
                if (output.option("braces")) {
                    make_block(stat, output);
                }
                else {
                    if (!stat || stat instanceof AST_EmptyStatement)
                        output.force_semicolon();
                    else
                        stat.print(output);
                }
            }
            function best_of(a) {
                var best = a[0], len = best.length;
                for (var i = 1; i < a.length; ++i) {
                    if (a[i].length < len) {
                        best = a[i];
                        len = best.length;
                    }
                }
                return best;
            }
            function make_num(num) {
                var str = num.toString(10).replace(/^0\./, ".").replace("e+", "e");
                var candidates = [str];
                if (Math.floor(num) === num) {
                    if (num < 0) {
                        candidates.push("-0x" + (-num).toString(16).toLowerCase());
                    }
                    else {
                        candidates.push("0x" + num.toString(16).toLowerCase());
                    }
                }
                var match, len, digits;
                if (match = /^\.0+/.exec(str)) {
                    len = match[0].length;
                    digits = str.slice(len);
                    candidates.push(digits + "e-" + (digits.length + len - 1));
                }
                else if (match = /0+$/.exec(str)) {
                    len = match[0].length;
                    candidates.push(str.slice(0, -len) + "e" + len);
                }
                else if (match = /^(\d)\.(\d+)e(-?\d+)$/.exec(str)) {
                    candidates.push(match[1] + match[2] + "e" + (match[3] - match[2].length));
                }
                return best_of(candidates);
            }
            function make_block(stmt, output) {
                if (!stmt || stmt instanceof AST_EmptyStatement)
                    output.print("{}");
                else if (stmt instanceof AST_BlockStatement)
                    stmt.print(output);
                else
                    output.with_block(function () {
                        output.indent();
                        stmt.print(output);
                        output.newline();
                    });
            }
            /* -----[ source map generators ]----- */
            function DEFMAP(nodetype, generator) {
                nodetype.forEach(function (nodetype) {
                    nodetype.DEFMETHOD("add_source_map", generator);
                });
            }
            DEFMAP([
                // We could easily add info for ALL nodes, but it seems to me that
                // would be quite wasteful, hence this noop in the base class.
                AST_Node,
                // since the label symbol will mark it
                AST_LabeledStatement,
                AST_Toplevel,
            ], noop);
            // XXX: I'm not exactly sure if we need it for all of these nodes,
            // or if we should add even more.
            DEFMAP([
                AST_Array,
                AST_BlockStatement,
                AST_Catch,
                AST_Class,
                AST_Constant,
                AST_Debugger,
                AST_Definitions,
                AST_Directive,
                AST_Finally,
                AST_Jump,
                AST_Lambda,
                AST_New,
                AST_Object,
                AST_StatementWithBody,
                AST_Symbol,
                AST_Switch,
                AST_SwitchBranch,
                AST_TemplateString,
                AST_TemplateSegment,
                AST_Try,
            ], function (output) {
                output.add_mapping(this.start);
            });
            DEFMAP([
                AST_ObjectGetter,
                AST_ObjectSetter,
            ], function (output) {
                output.add_mapping(this.start, this.key.name);
            });
            DEFMAP([AST_ObjectProperty], function (output) {
                output.add_mapping(this.start, this.key);
            });
        })();
        const shallow_cmp = (node1, node2) => {
            return (node1 === null && node2 === null
                || node1.TYPE === node2.TYPE && node1.shallow_cmp(node2));
        };
        const equivalent_to = (tree1, tree2) => {
            if (!shallow_cmp(tree1, tree2))
                return false;
            const walk_1_state = [tree1];
            const walk_2_state = [tree2];
            const walk_1_push = walk_1_state.push.bind(walk_1_state);
            const walk_2_push = walk_2_state.push.bind(walk_2_state);
            while (walk_1_state.length && walk_2_state.length) {
                const node_1 = walk_1_state.pop();
                const node_2 = walk_2_state.pop();
                if (!shallow_cmp(node_1, node_2))
                    return false;
                node_1._children_backwards(walk_1_push);
                node_2._children_backwards(walk_2_push);
                if (walk_1_state.length !== walk_2_state.length) {
                    // Different number of children
                    return false;
                }
            }
            return walk_1_state.length == 0 && walk_2_state.length == 0;
        };
        // Creates a shallow compare function
        const mkshallow = (props) => {
            const comparisons = Object
                .keys(props)
                .map(key => {
                if (props[key] === "eq") {
                    return `this.${key} === other.${key}`;
                }
                else if (props[key] === "exist") {
                    return `(this.${key} == null ? other.${key} == null : this.${key} === other.${key})`;
                }
                else {
                    throw new Error(`mkshallow: Unexpected instruction: ${props[key]}`);
                }
            })
                .join(" && ");
            return new Function("other", "return " + comparisons);
        };
        const pass_through = () => true;
        AST_Node.prototype.shallow_cmp = function () {
            throw new Error("did not find a shallow_cmp function for " + this.constructor.name);
        };
        AST_Debugger.prototype.shallow_cmp = pass_through;
        AST_Directive.prototype.shallow_cmp = mkshallow({ value: "eq" });
        AST_SimpleStatement.prototype.shallow_cmp = pass_through;
        AST_Block.prototype.shallow_cmp = pass_through;
        AST_EmptyStatement.prototype.shallow_cmp = pass_through;
        AST_LabeledStatement.prototype.shallow_cmp = mkshallow({ "label.name": "eq" });
        AST_Do.prototype.shallow_cmp = pass_through;
        AST_While.prototype.shallow_cmp = pass_through;
        AST_For.prototype.shallow_cmp = mkshallow({
            init: "exist",
            condition: "exist",
            step: "exist"
        });
        AST_ForIn.prototype.shallow_cmp = pass_through;
        AST_ForOf.prototype.shallow_cmp = pass_through;
        AST_With.prototype.shallow_cmp = pass_through;
        AST_Toplevel.prototype.shallow_cmp = pass_through;
        AST_Expansion.prototype.shallow_cmp = pass_through;
        AST_Lambda.prototype.shallow_cmp = mkshallow({
            is_generator: "eq",
            async: "eq"
        });
        AST_Destructuring.prototype.shallow_cmp = mkshallow({
            is_array: "eq"
        });
        AST_PrefixedTemplateString.prototype.shallow_cmp = pass_through;
        AST_TemplateString.prototype.shallow_cmp = pass_through;
        AST_TemplateSegment.prototype.shallow_cmp = mkshallow({
            "value": "eq"
        });
        AST_Jump.prototype.shallow_cmp = pass_through;
        AST_LoopControl.prototype.shallow_cmp = pass_through;
        AST_Await.prototype.shallow_cmp = pass_through;
        AST_Yield.prototype.shallow_cmp = mkshallow({
            is_star: "eq"
        });
        AST_If.prototype.shallow_cmp = mkshallow({
            alternative: "exist"
        });
        AST_Switch.prototype.shallow_cmp = pass_through;
        AST_SwitchBranch.prototype.shallow_cmp = pass_through;
        AST_Try.prototype.shallow_cmp = mkshallow({
            bcatch: "exist",
            bfinally: "exist"
        });
        AST_Catch.prototype.shallow_cmp = mkshallow({
            argname: "exist"
        });
        AST_Finally.prototype.shallow_cmp = pass_through;
        AST_Definitions.prototype.shallow_cmp = pass_through;
        AST_VarDef.prototype.shallow_cmp = mkshallow({
            value: "exist"
        });
        AST_NameMapping.prototype.shallow_cmp = pass_through;
        AST_Import.prototype.shallow_cmp = mkshallow({
            imported_name: "exist",
            imported_names: "exist"
        });
        AST_ImportMeta.prototype.shallow_cmp = pass_through;
        AST_Export.prototype.shallow_cmp = mkshallow({
            exported_definition: "exist",
            exported_value: "exist",
            exported_names: "exist",
            module_name: "eq",
            is_default: "eq",
        });
        AST_Call.prototype.shallow_cmp = pass_through;
        AST_Sequence.prototype.shallow_cmp = pass_through;
        AST_PropAccess.prototype.shallow_cmp = pass_through;
        AST_Chain.prototype.shallow_cmp = pass_through;
        AST_Dot.prototype.shallow_cmp = mkshallow({
            property: "eq"
        });
        AST_DotHash.prototype.shallow_cmp = mkshallow({
            property: "eq"
        });
        AST_Unary.prototype.shallow_cmp = mkshallow({
            operator: "eq"
        });
        AST_Binary.prototype.shallow_cmp = mkshallow({
            operator: "eq"
        });
        AST_Conditional.prototype.shallow_cmp = pass_through;
        AST_Array.prototype.shallow_cmp = pass_through;
        AST_Object.prototype.shallow_cmp = pass_through;
        AST_ObjectProperty.prototype.shallow_cmp = pass_through;
        AST_ObjectKeyVal.prototype.shallow_cmp = mkshallow({
            key: "eq"
        });
        AST_ObjectSetter.prototype.shallow_cmp = mkshallow({
            static: "eq"
        });
        AST_ObjectGetter.prototype.shallow_cmp = mkshallow({
            static: "eq"
        });
        AST_ConciseMethod.prototype.shallow_cmp = mkshallow({
            static: "eq",
            is_generator: "eq",
            async: "eq",
        });
        AST_Class.prototype.shallow_cmp = mkshallow({
            name: "exist",
            extends: "exist",
        });
        AST_ClassProperty.prototype.shallow_cmp = mkshallow({
            static: "eq"
        });
        AST_Symbol.prototype.shallow_cmp = mkshallow({
            name: "eq"
        });
        AST_NewTarget.prototype.shallow_cmp = pass_through;
        AST_This.prototype.shallow_cmp = pass_through;
        AST_Super.prototype.shallow_cmp = pass_through;
        AST_String.prototype.shallow_cmp = mkshallow({
            value: "eq"
        });
        AST_Number.prototype.shallow_cmp = mkshallow({
            value: "eq"
        });
        AST_BigInt.prototype.shallow_cmp = mkshallow({
            value: "eq"
        });
        AST_RegExp.prototype.shallow_cmp = function (other) {
            return (this.value.flags === other.value.flags
                && this.value.source === other.value.source);
        };
        AST_Atom.prototype.shallow_cmp = pass_through;
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        const MASK_EXPORT_DONT_MANGLE = 1 << 0;
        const MASK_EXPORT_WANT_MANGLE = 1 << 1;
        let function_defs = null;
        let unmangleable_names = null;
        class SymbolDef {
            constructor(scope, orig, init) {
                this.name = orig.name;
                this.orig = [orig];
                this.init = init;
                this.eliminated = 0;
                this.assignments = 0;
                this.scope = scope;
                this.replaced = 0;
                this.global = false;
                this.export = 0;
                this.mangled_name = null;
                this.undeclared = false;
                this.id = SymbolDef.next_id++;
                this.chained = false;
                this.direct_access = false;
                this.escaped = 0;
                this.recursive_refs = 0;
                this.references = [];
                this.should_replace = undefined;
                this.single_use = false;
                this.fixed = false;
                Object.seal(this);
            }
            fixed_value() {
                if (!this.fixed || this.fixed instanceof AST_Node)
                    return this.fixed;
                return this.fixed();
            }
            unmangleable(options) {
                if (!options)
                    options = {};
                if (function_defs &&
                    function_defs.has(this.id) &&
                    keep_name(options.keep_fnames, this.orig[0].name))
                    return true;
                return this.global && !options.toplevel
                    || (this.export & MASK_EXPORT_DONT_MANGLE)
                    || this.undeclared
                    || !options.eval && this.scope.pinned()
                    || (this.orig[0] instanceof AST_SymbolLambda
                        || this.orig[0] instanceof AST_SymbolDefun) && keep_name(options.keep_fnames, this.orig[0].name)
                    || this.orig[0] instanceof AST_SymbolMethod
                    || (this.orig[0] instanceof AST_SymbolClass
                        || this.orig[0] instanceof AST_SymbolDefClass) && keep_name(options.keep_classnames, this.orig[0].name);
            }
            mangle(options) {
                const cache = options.cache && options.cache.props;
                if (this.global && cache && cache.has(this.name)) {
                    this.mangled_name = cache.get(this.name);
                }
                else if (!this.mangled_name && !this.unmangleable(options)) {
                    var s = this.scope;
                    var sym = this.orig[0];
                    if (options.ie8 && sym instanceof AST_SymbolLambda)
                        s = s.parent_scope;
                    const redefinition = redefined_catch_def(this);
                    this.mangled_name = redefinition
                        ? redefinition.mangled_name || redefinition.name
                        : s.next_mangled(options, this);
                    if (this.global && cache) {
                        cache.set(this.name, this.mangled_name);
                    }
                }
            }
        }
        SymbolDef.next_id = 1;
        function redefined_catch_def(def) {
            if (def.orig[0] instanceof AST_SymbolCatch
                && def.scope.is_block_scope()) {
                return def.scope.get_defun_scope().variables.get(def.name);
            }
        }
        AST_Scope.DEFMETHOD("figure_out_scope", function (options, { parent_scope = null, toplevel = this } = {}) {
            options = defaults(options, {
                cache: null,
                ie8: false,
                safari10: false,
            });
            if (!(toplevel instanceof AST_Toplevel)) {
                throw new Error("Invalid toplevel scope");
            }
            // pass 1: setup scope chaining and handle definitions
            var scope = this.parent_scope = parent_scope;
            var labels = new Map();
            var defun = null;
            var in_destructuring = null;
            var for_scopes = [];
            var tw = new TreeWalker((node, descend) => {
                if (node.is_block_scope()) {
                    const save_scope = scope;
                    node.block_scope = scope = new AST_Scope(node);
                    scope._block_scope = true;
                    // AST_Try in the AST sadly *is* (not has) a body itself,
                    // and its catch and finally branches are children of the AST_Try itself
                    const parent_scope = node instanceof AST_Catch
                        ? save_scope.parent_scope
                        : save_scope;
                    scope.init_scope_vars(parent_scope);
                    scope.uses_with = save_scope.uses_with;
                    scope.uses_eval = save_scope.uses_eval;
                    if (options.safari10) {
                        if (node instanceof AST_For || node instanceof AST_ForIn) {
                            for_scopes.push(scope);
                        }
                    }
                    if (node instanceof AST_Switch) {
                        // XXX: HACK! Ensure the switch expression gets the correct scope (the parent scope) and the body gets the contained scope
                        // AST_Switch has a scope within the body, but it itself "is a block scope"
                        // This means the switched expression has to belong to the outer scope
                        // while the body inside belongs to the switch itself.
                        // This is pretty nasty and warrants an AST change similar to AST_Try (read above)
                        const the_block_scope = scope;
                        scope = save_scope;
                        node.expression.walk(tw);
                        scope = the_block_scope;
                        for (let i = 0; i < node.body.length; i++) {
                            node.body[i].walk(tw);
                        }
                    }
                    else {
                        descend();
                    }
                    scope = save_scope;
                    return true;
                }
                if (node instanceof AST_Destructuring) {
                    const save_destructuring = in_destructuring;
                    in_destructuring = node;
                    descend();
                    in_destructuring = save_destructuring;
                    return true;
                }
                if (node instanceof AST_Scope) {
                    node.init_scope_vars(scope);
                    var save_scope = scope;
                    var save_defun = defun;
                    var save_labels = labels;
                    defun = scope = node;
                    labels = new Map();
                    descend();
                    scope = save_scope;
                    defun = save_defun;
                    labels = save_labels;
                    return true; // don't descend again in TreeWalker
                }
                if (node instanceof AST_LabeledStatement) {
                    var l = node.label;
                    if (labels.has(l.name)) {
                        throw new Error(string_template("Label {name} defined twice", l));
                    }
                    labels.set(l.name, l);
                    descend();
                    labels.delete(l.name);
                    return true; // no descend again
                }
                if (node instanceof AST_With) {
                    for (var s = scope; s; s = s.parent_scope)
                        s.uses_with = true;
                    return;
                }
                if (node instanceof AST_Symbol) {
                    node.scope = scope;
                }
                if (node instanceof AST_Label) {
                    node.thedef = node;
                    node.references = [];
                }
                if (node instanceof AST_SymbolLambda) {
                    defun.def_function(node, node.name == "arguments" ? undefined : defun);
                }
                else if (node instanceof AST_SymbolDefun) {
                    // Careful here, the scope where this should be defined is
                    // the parent scope.  The reason is that we enter a new
                    // scope when we encounter the AST_Defun node (which is
                    // instanceof AST_Scope) but we get to the symbol a bit
                    // later.
                    const closest_scope = defun.parent_scope;
                    // In strict mode, function definitions are block-scoped
                    node.scope = tw.directives["use strict"]
                        ? closest_scope
                        : closest_scope.get_defun_scope();
                    mark_export(node.scope.def_function(node, defun), 1);
                }
                else if (node instanceof AST_SymbolClass) {
                    mark_export(defun.def_variable(node, defun), 1);
                }
                else if (node instanceof AST_SymbolImport) {
                    scope.def_variable(node);
                }
                else if (node instanceof AST_SymbolDefClass) {
                    // This deals with the name of the class being available
                    // inside the class.
                    mark_export((node.scope = defun.parent_scope).def_function(node, defun), 1);
                }
                else if (node instanceof AST_SymbolVar
                    || node instanceof AST_SymbolLet
                    || node instanceof AST_SymbolConst
                    || node instanceof AST_SymbolCatch) {
                    var def;
                    if (node instanceof AST_SymbolBlockDeclaration) {
                        def = scope.def_variable(node, null);
                    }
                    else {
                        def = defun.def_variable(node, node.TYPE == "SymbolVar" ? null : undefined);
                    }
                    if (!def.orig.every((sym) => {
                        if (sym === node)
                            return true;
                        if (node instanceof AST_SymbolBlockDeclaration) {
                            return sym instanceof AST_SymbolLambda;
                        }
                        return !(sym instanceof AST_SymbolLet || sym instanceof AST_SymbolConst);
                    })) {
                        js_error(`"${node.name}" is redeclared`, node.start.file, node.start.line, node.start.col, node.start.pos);
                    }
                    if (!(node instanceof AST_SymbolFunarg))
                        mark_export(def, 2);
                    if (defun !== scope) {
                        node.mark_enclosed();
                        var def = scope.find_variable(node);
                        if (node.thedef !== def) {
                            node.thedef = def;
                            node.reference();
                        }
                    }
                }
                else if (node instanceof AST_LabelRef) {
                    var sym = labels.get(node.name);
                    if (!sym)
                        throw new Error(string_template("Undefined label {name} [{line},{col}]", {
                            name: node.name,
                            line: node.start.line,
                            col: node.start.col
                        }));
                    node.thedef = sym;
                }
                if (!(scope instanceof AST_Toplevel) && (node instanceof AST_Export || node instanceof AST_Import)) {
                    js_error(`"${node.TYPE}" statement may only appear at the top level`, node.start.file, node.start.line, node.start.col, node.start.pos);
                }
            });
            this.walk(tw);
            function mark_export(def, level) {
                if (in_destructuring) {
                    var i = 0;
                    do {
                        level++;
                    } while (tw.parent(i++) !== in_destructuring);
                }
                var node = tw.parent(level);
                if (def.export = node instanceof AST_Export ? MASK_EXPORT_DONT_MANGLE : 0) {
                    var exported = node.exported_definition;
                    if ((exported instanceof AST_Defun || exported instanceof AST_DefClass) && node.is_default) {
                        def.export = MASK_EXPORT_WANT_MANGLE;
                    }
                }
            }
            // pass 2: find back references and eval
            const is_toplevel = this instanceof AST_Toplevel;
            if (is_toplevel) {
                this.globals = new Map();
            }
            var tw = new TreeWalker(node => {
                if (node instanceof AST_LoopControl && node.label) {
                    node.label.thedef.references.push(node);
                    return true;
                }
                if (node instanceof AST_SymbolRef) {
                    var name = node.name;
                    if (name == "eval" && tw.parent() instanceof AST_Call) {
                        for (var s = node.scope; s && !s.uses_eval; s = s.parent_scope) {
                            s.uses_eval = true;
                        }
                    }
                    var sym;
                    if (tw.parent() instanceof AST_NameMapping && tw.parent(1).module_name
                        || !(sym = node.scope.find_variable(name))) {
                        sym = toplevel.def_global(node);
                        if (node instanceof AST_SymbolExport)
                            sym.export = MASK_EXPORT_DONT_MANGLE;
                    }
                    else if (sym.scope instanceof AST_Lambda && name == "arguments") {
                        sym.scope.uses_arguments = true;
                    }
                    node.thedef = sym;
                    node.reference();
                    if (node.scope.is_block_scope()
                        && !(sym.orig[0] instanceof AST_SymbolBlockDeclaration)) {
                        node.scope = node.scope.get_defun_scope();
                    }
                    return true;
                }
                // ensure mangling works if catch reuses a scope variable
                var def;
                if (node instanceof AST_SymbolCatch && (def = redefined_catch_def(node.definition()))) {
                    var s = node.scope;
                    while (s) {
                        push_uniq(s.enclosed, def);
                        if (s === def.scope)
                            break;
                        s = s.parent_scope;
                    }
                }
            });
            this.walk(tw);
            // pass 3: work around IE8 and Safari catch scope bugs
            if (options.ie8 || options.safari10) {
                walk(this, node => {
                    if (node instanceof AST_SymbolCatch) {
                        var name = node.name;
                        var refs = node.thedef.references;
                        var scope = node.scope.get_defun_scope();
                        var def = scope.find_variable(name)
                            || toplevel.globals.get(name)
                            || scope.def_variable(node);
                        refs.forEach(function (ref) {
                            ref.thedef = def;
                            ref.reference();
                        });
                        node.thedef = def;
                        node.reference();
                        return true;
                    }
                });
            }
            // pass 4: add symbol definitions to loop scopes
            // Safari/Webkit bug workaround - loop init let variable shadowing argument.
            // https://github.com/mishoo/UglifyJS2/issues/1753
            // https://bugs.webkit.org/show_bug.cgi?id=171041
            if (options.safari10) {
                for (const scope of for_scopes) {
                    scope.parent_scope.variables.forEach(function (def) {
                        push_uniq(scope.enclosed, def);
                    });
                }
            }
        });
        AST_Toplevel.DEFMETHOD("def_global", function (node) {
            var globals = this.globals, name = node.name;
            if (globals.has(name)) {
                return globals.get(name);
            }
            else {
                var g = new SymbolDef(this, node);
                g.undeclared = true;
                g.global = true;
                globals.set(name, g);
                return g;
            }
        });
        AST_Scope.DEFMETHOD("init_scope_vars", function (parent_scope) {
            this.variables = new Map(); // map name to AST_SymbolVar (variables defined in this scope; includes functions)
            this.uses_with = false; // will be set to true if this or some nested scope uses the `with` statement
            this.uses_eval = false; // will be set to true if this or nested scope uses the global `eval`
            this.parent_scope = parent_scope; // the parent scope
            this.enclosed = []; // a list of variables from this or outer scope(s) that are referenced from this or inner scopes
            this.cname = -1; // the current index for mangling functions/variables
        });
        AST_Scope.DEFMETHOD("conflicting_def", function (name) {
            return (this.enclosed.find(def => def.name === name)
                || this.variables.has(name)
                || (this.parent_scope && this.parent_scope.conflicting_def(name)));
        });
        AST_Scope.DEFMETHOD("conflicting_def_shallow", function (name) {
            return (this.enclosed.find(def => def.name === name)
                || this.variables.has(name));
        });
        AST_Scope.DEFMETHOD("add_child_scope", function (scope) {
            // `scope` is going to be moved into `this` right now.
            // Update the required scopes' information
            if (scope.parent_scope === this)
                return;
            scope.parent_scope = this;
            // TODO uses_with, uses_eval, etc
            const scope_ancestry = (() => {
                const ancestry = [];
                let cur = this;
                do {
                    ancestry.push(cur);
                } while ((cur = cur.parent_scope));
                ancestry.reverse();
                return ancestry;
            })();
            const new_scope_enclosed_set = new Set(scope.enclosed);
            const to_enclose = [];
            for (const scope_topdown of scope_ancestry) {
                to_enclose.forEach(e => push_uniq(scope_topdown.enclosed, e));
                for (const def of scope_topdown.variables.values()) {
                    if (new_scope_enclosed_set.has(def)) {
                        push_uniq(to_enclose, def);
                        push_uniq(scope_topdown.enclosed, def);
                    }
                }
            }
        });
        function find_scopes_visible_from(scopes) {
            const found_scopes = new Set();
            for (const scope of new Set(scopes)) {
                (function bubble_up(scope) {
                    if (scope == null || found_scopes.has(scope))
                        return;
                    found_scopes.add(scope);
                    bubble_up(scope.parent_scope);
                })(scope);
            }
            return [...found_scopes];
        }
        // Creates a symbol during compression
        AST_Scope.DEFMETHOD("create_symbol", function (SymClass, { source, tentative_name, scope, conflict_scopes = [scope], init = null } = {}) {
            let symbol_name;
            conflict_scopes = find_scopes_visible_from(conflict_scopes);
            if (tentative_name) {
                // Implement hygiene (no new names are conflicting with existing names)
                tentative_name =
                    symbol_name =
                        tentative_name.replace(/(?:^[^a-z_$]|[^a-z0-9_$])/ig, "_");
                let i = 0;
                while (conflict_scopes.find(s => s.conflicting_def_shallow(symbol_name))) {
                    symbol_name = tentative_name + "$" + i++;
                }
            }
            if (!symbol_name) {
                throw new Error("No symbol name could be generated in create_symbol()");
            }
            const symbol = make_node(SymClass, source, {
                name: symbol_name,
                scope
            });
            this.def_variable(symbol, init || null);
            symbol.mark_enclosed();
            return symbol;
        });
        AST_Node.DEFMETHOD("is_block_scope", return_false);
        AST_Class.DEFMETHOD("is_block_scope", return_false);
        AST_Lambda.DEFMETHOD("is_block_scope", return_false);
        AST_Toplevel.DEFMETHOD("is_block_scope", return_false);
        AST_SwitchBranch.DEFMETHOD("is_block_scope", return_false);
        AST_Block.DEFMETHOD("is_block_scope", return_true);
        AST_Scope.DEFMETHOD("is_block_scope", function () {
            return this._block_scope || false;
        });
        AST_IterationStatement.DEFMETHOD("is_block_scope", return_true);
        AST_Lambda.DEFMETHOD("init_scope_vars", function () {
            AST_Scope.prototype.init_scope_vars.apply(this, arguments);
            this.uses_arguments = false;
            this.def_variable(new AST_SymbolFunarg({
                name: "arguments",
                start: this.start,
                end: this.end
            }));
        });
        AST_Arrow.DEFMETHOD("init_scope_vars", function () {
            AST_Scope.prototype.init_scope_vars.apply(this, arguments);
            this.uses_arguments = false;
        });
        AST_Symbol.DEFMETHOD("mark_enclosed", function () {
            var def = this.definition();
            var s = this.scope;
            while (s) {
                push_uniq(s.enclosed, def);
                if (s === def.scope)
                    break;
                s = s.parent_scope;
            }
        });
        AST_Symbol.DEFMETHOD("reference", function () {
            this.definition().references.push(this);
            this.mark_enclosed();
        });
        AST_Scope.DEFMETHOD("find_variable", function (name) {
            if (name instanceof AST_Symbol)
                name = name.name;
            return this.variables.get(name)
                || (this.parent_scope && this.parent_scope.find_variable(name));
        });
        AST_Scope.DEFMETHOD("def_function", function (symbol, init) {
            var def = this.def_variable(symbol, init);
            if (!def.init || def.init instanceof AST_Defun)
                def.init = init;
            return def;
        });
        AST_Scope.DEFMETHOD("def_variable", function (symbol, init) {
            var def = this.variables.get(symbol.name);
            if (def) {
                def.orig.push(symbol);
                if (def.init && (def.scope !== symbol.scope || def.init instanceof AST_Function)) {
                    def.init = init;
                }
            }
            else {
                def = new SymbolDef(this, symbol, init);
                this.variables.set(symbol.name, def);
                def.global = !this.parent_scope;
            }
            return symbol.thedef = def;
        });
        function next_mangled(scope, options) {
            var ext = scope.enclosed;
            out: while (true) {
                var m = base54(++scope.cname);
                if (RESERVED_WORDS.has(m))
                    continue; // skip over "do"
                // https://github.com/mishoo/UglifyJS2/issues/242 -- do not
                // shadow a name reserved from mangling.
                if (options.reserved.has(m))
                    continue;
                // Functions with short names might collide with base54 output
                // and therefore cause collisions when keep_fnames is true.
                if (unmangleable_names && unmangleable_names.has(m))
                    continue out;
                // we must ensure that the mangled name does not shadow a name
                // from some parent scope that is referenced in this or in
                // inner scopes.
                for (let i = ext.length; --i >= 0;) {
                    const def = ext[i];
                    const name = def.mangled_name || (def.unmangleable(options) && def.name);
                    if (m == name)
                        continue out;
                }
                return m;
            }
        }
        AST_Scope.DEFMETHOD("next_mangled", function (options) {
            return next_mangled(this, options);
        });
        AST_Toplevel.DEFMETHOD("next_mangled", function (options) {
            let name;
            const mangled_names = this.mangled_names;
            do {
                name = next_mangled(this, options);
            } while (mangled_names.has(name));
            return name;
        });
        AST_Function.DEFMETHOD("next_mangled", function (options, def) {
            // #179, #326
            // in Safari strict mode, something like (function x(x){...}) is a syntax error;
            // a function expression's argument cannot shadow the function expression's name
            var tricky_def = def.orig[0] instanceof AST_SymbolFunarg && this.name && this.name.definition();
            // the function's mangled_name is null when keep_fnames is true
            var tricky_name = tricky_def ? tricky_def.mangled_name || tricky_def.name : null;
            while (true) {
                var name = next_mangled(this, options);
                if (!tricky_name || tricky_name != name)
                    return name;
            }
        });
        AST_Symbol.DEFMETHOD("unmangleable", function (options) {
            var def = this.definition();
            return !def || def.unmangleable(options);
        });
        // labels are always mangleable
        AST_Label.DEFMETHOD("unmangleable", return_false);
        AST_Symbol.DEFMETHOD("unreferenced", function () {
            return !this.definition().references.length && !this.scope.pinned();
        });
        AST_Symbol.DEFMETHOD("definition", function () {
            return this.thedef;
        });
        AST_Symbol.DEFMETHOD("global", function () {
            return this.thedef.global;
        });
        AST_Toplevel.DEFMETHOD("_default_mangler_options", function (options) {
            options = defaults(options, {
                eval: false,
                ie8: false,
                keep_classnames: false,
                keep_fnames: false,
                module: false,
                reserved: [],
                toplevel: false,
            });
            if (options.module)
                options.toplevel = true;
            if (!Array.isArray(options.reserved)
                && !(options.reserved instanceof Set)) {
                options.reserved = [];
            }
            options.reserved = new Set(options.reserved);
            // Never mangle arguments
            options.reserved.add("arguments");
            return options;
        });
        AST_Toplevel.DEFMETHOD("mangle_names", function (options) {
            options = this._default_mangler_options(options);
            // We only need to mangle declaration nodes.  Special logic wired
            // into the code generator will display the mangled name if it's
            // present (and for AST_SymbolRef-s it'll use the mangled name of
            // the AST_SymbolDeclaration that it points to).
            var lname = -1;
            var to_mangle = [];
            if (options.keep_fnames) {
                function_defs = new Set();
            }
            const mangled_names = this.mangled_names = new Set();
            if (options.cache) {
                this.globals.forEach(collect);
                if (options.cache.props) {
                    options.cache.props.forEach(function (mangled_name) {
                        mangled_names.add(mangled_name);
                    });
                }
            }
            var tw = new TreeWalker(function (node, descend) {
                if (node instanceof AST_LabeledStatement) {
                    // lname is incremented when we get to the AST_Label
                    var save_nesting = lname;
                    descend();
                    lname = save_nesting;
                    return true; // don't descend again in TreeWalker
                }
                if (node instanceof AST_Scope) {
                    node.variables.forEach(collect);
                    return;
                }
                if (node.is_block_scope()) {
                    node.block_scope.variables.forEach(collect);
                    return;
                }
                if (function_defs
                    && node instanceof AST_VarDef
                    && node.value instanceof AST_Lambda
                    && !node.value.name
                    && keep_name(options.keep_fnames, node.name.name)) {
                    function_defs.add(node.name.definition().id);
                    return;
                }
                if (node instanceof AST_Label) {
                    let name;
                    do {
                        name = base54(++lname);
                    } while (RESERVED_WORDS.has(name));
                    node.mangled_name = name;
                    return true;
                }
                if (!(options.ie8 || options.safari10) && node instanceof AST_SymbolCatch) {
                    to_mangle.push(node.definition());
                    return;
                }
            });
            this.walk(tw);
            if (options.keep_fnames || options.keep_classnames) {
                unmangleable_names = new Set();
                // Collect a set of short names which are unmangleable,
                // for use in avoiding collisions in next_mangled.
                to_mangle.forEach(def => {
                    if (def.name.length < 6 && def.unmangleable(options)) {
                        unmangleable_names.add(def.name);
                    }
                });
            }
            to_mangle.forEach(def => { def.mangle(options); });
            function_defs = null;
            unmangleable_names = null;
            function collect(symbol) {
                const should_mangle = !options.reserved.has(symbol.name)
                    && !(symbol.export & MASK_EXPORT_DONT_MANGLE);
                if (should_mangle) {
                    to_mangle.push(symbol);
                }
            }
        });
        AST_Toplevel.DEFMETHOD("find_colliding_names", function (options) {
            const cache = options.cache && options.cache.props;
            const avoid = new Set();
            options.reserved.forEach(to_avoid);
            this.globals.forEach(add_def);
            this.walk(new TreeWalker(function (node) {
                if (node instanceof AST_Scope)
                    node.variables.forEach(add_def);
                if (node instanceof AST_SymbolCatch)
                    add_def(node.definition());
            }));
            return avoid;
            function to_avoid(name) {
                avoid.add(name);
            }
            function add_def(def) {
                var name = def.name;
                if (def.global && cache && cache.has(name))
                    name = cache.get(name);
                else if (!def.unmangleable(options))
                    return;
                to_avoid(name);
            }
        });
        AST_Toplevel.DEFMETHOD("expand_names", function (options) {
            base54.reset();
            base54.sort();
            options = this._default_mangler_options(options);
            var avoid = this.find_colliding_names(options);
            var cname = 0;
            this.globals.forEach(rename);
            this.walk(new TreeWalker(function (node) {
                if (node instanceof AST_Scope)
                    node.variables.forEach(rename);
                if (node instanceof AST_SymbolCatch)
                    rename(node.definition());
            }));
            function next_name() {
                var name;
                do {
                    name = base54(cname++);
                } while (avoid.has(name) || RESERVED_WORDS.has(name));
                return name;
            }
            function rename(def) {
                if (def.global && options.cache)
                    return;
                if (def.unmangleable(options))
                    return;
                if (options.reserved.has(def.name))
                    return;
                const redefinition = redefined_catch_def(def);
                const name = def.name = redefinition ? redefinition.name : next_name();
                def.orig.forEach(function (sym) {
                    sym.name = name;
                });
                def.references.forEach(function (sym) {
                    sym.name = name;
                });
            }
        });
        AST_Node.DEFMETHOD("tail_node", return_this);
        AST_Sequence.DEFMETHOD("tail_node", function () {
            return this.expressions[this.expressions.length - 1];
        });
        AST_Toplevel.DEFMETHOD("compute_char_frequency", function (options) {
            options = this._default_mangler_options(options);
            try {
                AST_Node.prototype.print = function (stream, force_parens) {
                    this._print(stream, force_parens);
                    if (this instanceof AST_Symbol && !this.unmangleable(options)) {
                        base54.consider(this.name, -1);
                    }
                    else if (options.properties) {
                        if (this instanceof AST_DotHash) {
                            base54.consider("#" + this.property, -1);
                        }
                        else if (this instanceof AST_Dot) {
                            base54.consider(this.property, -1);
                        }
                        else if (this instanceof AST_Sub) {
                            skip_string(this.property);
                        }
                    }
                };
                base54.consider(this.print_to_string(), 1);
            }
            finally {
                AST_Node.prototype.print = AST_Node.prototype._print;
            }
            base54.sort();
            function skip_string(node) {
                if (node instanceof AST_String) {
                    base54.consider(node.value, -1);
                }
                else if (node instanceof AST_Conditional) {
                    skip_string(node.consequent);
                    skip_string(node.alternative);
                }
                else if (node instanceof AST_Sequence) {
                    skip_string(node.tail_node());
                }
            }
        });
        const base54 = (() => {
            const leading = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ$_".split("");
            const digits = "0123456789".split("");
            let chars;
            let frequency;
            function reset() {
                frequency = new Map();
                leading.forEach(function (ch) {
                    frequency.set(ch, 0);
                });
                digits.forEach(function (ch) {
                    frequency.set(ch, 0);
                });
            }
            base54.consider = function (str, delta) {
                for (var i = str.length; --i >= 0;) {
                    frequency.set(str[i], frequency.get(str[i]) + delta);
                }
            };
            function compare(a, b) {
                return frequency.get(b) - frequency.get(a);
            }
            base54.sort = function () {
                chars = mergeSort(leading, compare).concat(mergeSort(digits, compare));
            };
            base54.reset = reset;
            reset();
            function base54(num) {
                var ret = "", base = 54;
                num++;
                do {
                    num--;
                    ret += chars[num % base];
                    num = Math.floor(num / base);
                    base = 64;
                } while (num > 0);
                return ret;
            }
            return base54;
        })();
        let mangle_options = undefined;
        AST_Node.prototype.size = function (compressor, stack) {
            mangle_options = compressor && compressor.mangle_options;
            let size = 0;
            walk_parent(this, (node, info) => {
                size += node._size(info);
                // Braceless arrow functions have fake "return" statements
                if (node instanceof AST_Arrow && node.is_braceless()) {
                    size += node.body[0].value._size(info);
                    return true;
                }
            }, stack || (compressor && compressor.stack));
            // just to save a bit of memory
            mangle_options = undefined;
            return size;
        };
        AST_Node.prototype._size = () => 0;
        AST_Debugger.prototype._size = () => 8;
        AST_Directive.prototype._size = function () {
            // TODO string encoding stuff
            return 2 + this.value.length;
        };
        const list_overhead = (array) => array.length && array.length - 1;
        AST_Block.prototype._size = function () {
            return 2 + list_overhead(this.body);
        };
        AST_Toplevel.prototype._size = function () {
            return list_overhead(this.body);
        };
        AST_EmptyStatement.prototype._size = () => 1;
        AST_LabeledStatement.prototype._size = () => 2; // x:
        AST_Do.prototype._size = () => 9;
        AST_While.prototype._size = () => 7;
        AST_For.prototype._size = () => 8;
        AST_ForIn.prototype._size = () => 8;
        // AST_ForOf inherits ^
        AST_With.prototype._size = () => 6;
        AST_Expansion.prototype._size = () => 3;
        const lambda_modifiers = func => (func.is_generator ? 1 : 0) + (func.async ? 6 : 0);
        AST_Accessor.prototype._size = function () {
            return lambda_modifiers(this) + 4 + list_overhead(this.argnames) + list_overhead(this.body);
        };
        AST_Function.prototype._size = function (info) {
            const first = !!first_in_statement(info);
            return (first * 2) + lambda_modifiers(this) + 12 + list_overhead(this.argnames) + list_overhead(this.body);
        };
        AST_Defun.prototype._size = function () {
            return lambda_modifiers(this) + 13 + list_overhead(this.argnames) + list_overhead(this.body);
        };
        AST_Arrow.prototype._size = function () {
            let args_and_arrow = 2 + list_overhead(this.argnames);
            if (!(this.argnames.length === 1
                && this.argnames[0] instanceof AST_Symbol)) {
                args_and_arrow += 2;
            }
            const body_overhead = this.is_braceless() ? 0 : list_overhead(this.body) + 2;
            return lambda_modifiers(this) + args_and_arrow + body_overhead;
        };
        AST_Destructuring.prototype._size = () => 2;
        AST_TemplateString.prototype._size = function () {
            return 2 + (Math.floor(this.segments.length / 2) * 3); /* "${}" */
        };
        AST_TemplateSegment.prototype._size = function () {
            return this.value.length;
        };
        AST_Return.prototype._size = function () {
            return this.value ? 7 : 6;
        };
        AST_Throw.prototype._size = () => 6;
        AST_Break.prototype._size = function () {
            return this.label ? 6 : 5;
        };
        AST_Continue.prototype._size = function () {
            return this.label ? 9 : 8;
        };
        AST_If.prototype._size = () => 4;
        AST_Switch.prototype._size = function () {
            return 8 + list_overhead(this.body);
        };
        AST_Case.prototype._size = function () {
            return 5 + list_overhead(this.body);
        };
        AST_Default.prototype._size = function () {
            return 8 + list_overhead(this.body);
        };
        AST_Try.prototype._size = function () {
            return 3 + list_overhead(this.body);
        };
        AST_Catch.prototype._size = function () {
            let size = 7 + list_overhead(this.body);
            if (this.argname) {
                size += 2;
            }
            return size;
        };
        AST_Finally.prototype._size = function () {
            return 7 + list_overhead(this.body);
        };
        /*#__INLINE__*/
        const def_size = (size, def) => size + list_overhead(def.definitions);
        AST_Var.prototype._size = function () {
            return def_size(4, this);
        };
        AST_Let.prototype._size = function () {
            return def_size(4, this);
        };
        AST_Const.prototype._size = function () {
            return def_size(6, this);
        };
        AST_VarDef.prototype._size = function () {
            return this.value ? 1 : 0;
        };
        AST_NameMapping.prototype._size = function () {
            // foreign name isn't mangled
            return this.name ? 4 : 0;
        };
        AST_Import.prototype._size = function () {
            // import
            let size = 6;
            if (this.imported_name)
                size += 1;
            // from
            if (this.imported_name || this.imported_names)
                size += 5;
            // braces, and the commas
            if (this.imported_names) {
                size += 2 + list_overhead(this.imported_names);
            }
            return size;
        };
        AST_ImportMeta.prototype._size = () => 11;
        AST_Export.prototype._size = function () {
            let size = 7 + (this.is_default ? 8 : 0);
            if (this.exported_value) {
                size += this.exported_value._size();
            }
            if (this.exported_names) {
                // Braces and commas
                size += 2 + list_overhead(this.exported_names);
            }
            if (this.module_name) {
                // "from "
                size += 5;
            }
            return size;
        };
        AST_Call.prototype._size = function () {
            if (this.optional) {
                return 4 + list_overhead(this.args);
            }
            return 2 + list_overhead(this.args);
        };
        AST_New.prototype._size = function () {
            return 6 + list_overhead(this.args);
        };
        AST_Sequence.prototype._size = function () {
            return list_overhead(this.expressions);
        };
        AST_Dot.prototype._size = function () {
            if (this.optional) {
                return this.property.length + 2;
            }
            return this.property.length + 1;
        };
        AST_DotHash.prototype._size = function () {
            if (this.optional) {
                return this.property.length + 3;
            }
            return this.property.length + 2;
        };
        AST_Sub.prototype._size = function () {
            return this.optional ? 4 : 2;
        };
        AST_Unary.prototype._size = function () {
            if (this.operator === "typeof")
                return 7;
            if (this.operator === "void")
                return 5;
            return this.operator.length;
        };
        AST_Binary.prototype._size = function (info) {
            if (this.operator === "in")
                return 4;
            let size = this.operator.length;
            if ((this.operator === "+" || this.operator === "-")
                && this.right instanceof AST_Unary && this.right.operator === this.operator) {
                // 1+ +a > needs space between the +
                size += 1;
            }
            if (this.needs_parens(info)) {
                size += 2;
            }
            return size;
        };
        AST_Conditional.prototype._size = () => 3;
        AST_Array.prototype._size = function () {
            return 2 + list_overhead(this.elements);
        };
        AST_Object.prototype._size = function (info) {
            let base = 2;
            if (first_in_statement(info)) {
                base += 2; // parens
            }
            return base + list_overhead(this.properties);
        };
        /*#__INLINE__*/
        const key_size = key => typeof key === "string" ? key.length : 0;
        AST_ObjectKeyVal.prototype._size = function () {
            return key_size(this.key) + 1;
        };
        /*#__INLINE__*/
        const static_size = is_static => is_static ? 7 : 0;
        AST_ObjectGetter.prototype._size = function () {
            return 5 + static_size(this.static) + key_size(this.key);
        };
        AST_ObjectSetter.prototype._size = function () {
            return 5 + static_size(this.static) + key_size(this.key);
        };
        AST_ConciseMethod.prototype._size = function () {
            return static_size(this.static) + key_size(this.key) + lambda_modifiers(this);
        };
        AST_PrivateMethod.prototype._size = function () {
            return AST_ConciseMethod.prototype._size.call(this) + 1;
        };
        AST_PrivateGetter.prototype._size = AST_PrivateSetter.prototype._size = function () {
            return AST_ConciseMethod.prototype._size.call(this) + 4;
        };
        AST_Class.prototype._size = function () {
            return ((this.name ? 8 : 7)
                + (this.extends ? 8 : 0));
        };
        AST_ClassProperty.prototype._size = function () {
            return (static_size(this.static)
                + (typeof this.key === "string" ? this.key.length + 2 : 0)
                + (this.value ? 1 : 0));
        };
        AST_ClassPrivateProperty.prototype._size = function () {
            return AST_ClassProperty.prototype._size.call(this) + 1;
        };
        AST_Symbol.prototype._size = function () {
            return !mangle_options || this.definition().unmangleable(mangle_options)
                ? this.name.length
                : 1;
        };
        // TODO take propmangle into account
        AST_SymbolClassProperty.prototype._size = function () {
            return this.name.length;
        };
        AST_SymbolRef.prototype._size = AST_SymbolDeclaration.prototype._size = function () {
            const { name, thedef } = this;
            if (thedef && thedef.global)
                return name.length;
            if (name === "arguments")
                return 9;
            return AST_Symbol.prototype._size.call(this);
        };
        AST_NewTarget.prototype._size = () => 10;
        AST_SymbolImportForeign.prototype._size = function () {
            return this.name.length;
        };
        AST_SymbolExportForeign.prototype._size = function () {
            return this.name.length;
        };
        AST_This.prototype._size = () => 4;
        AST_Super.prototype._size = () => 5;
        AST_String.prototype._size = function () {
            return this.value.length + 2;
        };
        AST_Number.prototype._size = function () {
            const { value } = this;
            if (value === 0)
                return 1;
            if (value > 0 && Math.floor(value) === value) {
                return Math.floor(Math.log10(value) + 1);
            }
            return value.toString().length;
        };
        AST_BigInt.prototype._size = function () {
            return this.value.length;
        };
        AST_RegExp.prototype._size = function () {
            return this.value.toString().length;
        };
        AST_Null.prototype._size = () => 4;
        AST_NaN.prototype._size = () => 3;
        AST_Undefined.prototype._size = () => 6; // "void 0"
        AST_Hole.prototype._size = () => 0; // comma is taken into account
        AST_Infinity.prototype._size = () => 8;
        AST_True.prototype._size = () => 4;
        AST_False.prototype._size = () => 5;
        AST_Await.prototype._size = () => 6;
        AST_Yield.prototype._size = () => 6;
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        const UNUSED = 0b00000001;
        const TRUTHY = 0b00000010;
        const FALSY = 0b00000100;
        const UNDEFINED = 0b00001000;
        const INLINED = 0b00010000;
        // Nodes to which values are ever written. Used when keep_assign is part of the unused option string.
        const WRITE_ONLY = 0b00100000;
        // information specific to a single compression pass
        const SQUEEZED = 0b0000000100000000;
        const OPTIMIZED = 0b0000001000000000;
        const TOP = 0b0000010000000000;
        const CLEAR_BETWEEN_PASSES = SQUEEZED | OPTIMIZED | TOP;
        const has_flag = (node, flag) => node.flags & flag;
        const set_flag = (node, flag) => { node.flags |= flag; };
        const clear_flag = (node, flag) => { node.flags &= ~flag; };
        class Compressor extends TreeWalker {
            constructor(options, { false_by_default = false, mangle_options = false }) {
                super();
                if (options.defaults !== undefined && !options.defaults)
                    false_by_default = true;
                this.options = defaults(options, {
                    arguments: false,
                    arrows: !false_by_default,
                    booleans: !false_by_default,
                    booleans_as_integers: false,
                    collapse_vars: !false_by_default,
                    comparisons: !false_by_default,
                    computed_props: !false_by_default,
                    conditionals: !false_by_default,
                    dead_code: !false_by_default,
                    defaults: true,
                    directives: !false_by_default,
                    drop_console: false,
                    drop_debugger: !false_by_default,
                    ecma: 5,
                    evaluate: !false_by_default,
                    expression: false,
                    global_defs: false,
                    hoist_funs: false,
                    hoist_props: !false_by_default,
                    hoist_vars: false,
                    ie8: false,
                    if_return: !false_by_default,
                    inline: !false_by_default,
                    join_vars: !false_by_default,
                    keep_classnames: false,
                    keep_fargs: true,
                    keep_fnames: false,
                    keep_infinity: false,
                    loops: !false_by_default,
                    module: false,
                    negate_iife: !false_by_default,
                    passes: 1,
                    properties: !false_by_default,
                    pure_getters: !false_by_default && "strict",
                    pure_funcs: null,
                    reduce_funcs: !false_by_default,
                    reduce_vars: !false_by_default,
                    sequences: !false_by_default,
                    side_effects: !false_by_default,
                    switches: !false_by_default,
                    top_retain: null,
                    toplevel: !!(options && options["top_retain"]),
                    typeofs: !false_by_default,
                    unsafe: false,
                    unsafe_arrows: false,
                    unsafe_comps: false,
                    unsafe_Function: false,
                    unsafe_math: false,
                    unsafe_symbols: false,
                    unsafe_methods: false,
                    unsafe_proto: false,
                    unsafe_regexp: false,
                    unsafe_undefined: false,
                    unused: !false_by_default,
                    warnings: false // legacy
                }, true);
                var global_defs = this.options["global_defs"];
                if (typeof global_defs == "object")
                    for (var key in global_defs) {
                        if (key[0] === "@" && HOP(global_defs, key)) {
                            global_defs[key.slice(1)] = parse(global_defs[key], {
                                expression: true
                            });
                        }
                    }
                if (this.options["inline"] === true)
                    this.options["inline"] = 3;
                var pure_funcs = this.options["pure_funcs"];
                if (typeof pure_funcs == "function") {
                    this.pure_funcs = pure_funcs;
                }
                else {
                    this.pure_funcs = pure_funcs ? function (node) {
                        return !pure_funcs.includes(node.expression.print_to_string());
                    } : return_true;
                }
                var top_retain = this.options["top_retain"];
                if (top_retain instanceof RegExp) {
                    this.top_retain = function (def) {
                        return top_retain.test(def.name);
                    };
                }
                else if (typeof top_retain == "function") {
                    this.top_retain = top_retain;
                }
                else if (top_retain) {
                    if (typeof top_retain == "string") {
                        top_retain = top_retain.split(/,/);
                    }
                    this.top_retain = function (def) {
                        return top_retain.includes(def.name);
                    };
                }
                if (this.options["module"]) {
                    this.directives["use strict"] = true;
                    this.options["toplevel"] = true;
                }
                var toplevel = this.options["toplevel"];
                this.toplevel = typeof toplevel == "string" ? {
                    funcs: /funcs/.test(toplevel),
                    vars: /vars/.test(toplevel)
                } : {
                    funcs: toplevel,
                    vars: toplevel
                };
                var sequences = this.options["sequences"];
                this.sequences_limit = sequences == 1 ? 800 : sequences | 0;
                this.evaluated_regexps = new Map();
                this._toplevel = undefined;
                this.mangle_options = mangle_options;
            }
            option(key) {
                return this.options[key];
            }
            exposed(def) {
                if (def.export)
                    return true;
                if (def.global)
                    for (var i = 0, len = def.orig.length; i < len; i++)
                        if (!this.toplevel[def.orig[i] instanceof AST_SymbolDefun ? "funcs" : "vars"])
                            return true;
                return false;
            }
            in_boolean_context() {
                if (!this.option("booleans"))
                    return false;
                var self = this.self();
                for (var i = 0, p; p = this.parent(i); i++) {
                    if (p instanceof AST_SimpleStatement
                        || p instanceof AST_Conditional && p.condition === self
                        || p instanceof AST_DWLoop && p.condition === self
                        || p instanceof AST_For && p.condition === self
                        || p instanceof AST_If && p.condition === self
                        || p instanceof AST_UnaryPrefix && p.operator == "!" && p.expression === self) {
                        return true;
                    }
                    if (p instanceof AST_Binary
                        && (p.operator == "&&"
                            || p.operator == "||"
                            || p.operator == "??")
                        || p instanceof AST_Conditional
                        || p.tail_node() === self) {
                        self = p;
                    }
                    else {
                        return false;
                    }
                }
            }
            get_toplevel() {
                return this._toplevel;
            }
            compress(toplevel) {
                toplevel = toplevel.resolve_defines(this);
                this._toplevel = toplevel;
                if (this.option("expression")) {
                    this._toplevel.process_expression(true);
                }
                var passes = +this.options.passes || 1;
                var min_count = 1 / 0;
                var stopping = false;
                var mangle = { ie8: this.option("ie8") };
                for (var pass = 0; pass < passes; pass++) {
                    this._toplevel.figure_out_scope(mangle);
                    if (pass === 0 && this.option("drop_console")) {
                        // must be run before reduce_vars and compress pass
                        this._toplevel = this._toplevel.drop_console();
                    }
                    if (pass > 0 || this.option("reduce_vars")) {
                        this._toplevel.reset_opt_flags(this);
                    }
                    this._toplevel = this._toplevel.transform(this);
                    if (passes > 1) {
                        let count = 0;
                        walk(this._toplevel, () => { count++; });
                        if (count < min_count) {
                            min_count = count;
                            stopping = false;
                        }
                        else if (stopping) {
                            break;
                        }
                        else {
                            stopping = true;
                        }
                    }
                }
                if (this.option("expression")) {
                    this._toplevel.process_expression(false);
                }
                toplevel = this._toplevel;
                this._toplevel = undefined;
                return toplevel;
            }
            before(node, descend) {
                if (has_flag(node, SQUEEZED))
                    return node;
                var was_scope = false;
                if (node instanceof AST_Scope) {
                    node = node.hoist_properties(this);
                    node = node.hoist_declarations(this);
                    was_scope = true;
                }
                // Before https://github.com/mishoo/UglifyJS2/pull/1602 AST_Node.optimize()
                // would call AST_Node.transform() if a different instance of AST_Node is
                // produced after def_optimize().
                // This corrupts TreeWalker.stack, which cause AST look-ups to malfunction.
                // Migrate and defer all children's AST_Node.transform() to below, which
                // will now happen after this parent AST_Node has been properly substituted
                // thus gives a consistent AST snapshot.
                descend(node, this);
                // Existing code relies on how AST_Node.optimize() worked, and omitting the
                // following replacement call would result in degraded efficiency of both
                // output and performance.
                descend(node, this);
                var opt = node.optimize(this);
                if (was_scope && opt instanceof AST_Scope) {
                    opt.drop_unused(this);
                    descend(opt, this);
                }
                if (opt === node)
                    set_flag(opt, SQUEEZED);
                return opt;
            }
        }
        function def_optimize(node, optimizer) {
            node.DEFMETHOD("optimize", function (compressor) {
                var self = this;
                if (has_flag(self, OPTIMIZED))
                    return self;
                if (compressor.has_directive("use asm"))
                    return self;
                var opt = optimizer(self, compressor);
                set_flag(opt, OPTIMIZED);
                return opt;
            });
        }
        def_optimize(AST_Node, function (self) {
            return self;
        });
        AST_Toplevel.DEFMETHOD("drop_console", function () {
            return this.transform(new TreeTransformer(function (self) {
                if (self.TYPE == "Call") {
                    var exp = self.expression;
                    if (exp instanceof AST_PropAccess) {
                        var name = exp.expression;
                        while (name.expression) {
                            name = name.expression;
                        }
                        if (is_undeclared_ref(name) && name.name == "console") {
                            return make_node(AST_Undefined, self);
                        }
                    }
                }
            }));
        });
        AST_Node.DEFMETHOD("equivalent_to", function (node) {
            return equivalent_to(this, node);
        });
        AST_Scope.DEFMETHOD("process_expression", function (insert, compressor) {
            var self = this;
            var tt = new TreeTransformer(function (node) {
                if (insert && node instanceof AST_SimpleStatement) {
                    return make_node(AST_Return, node, {
                        value: node.body
                    });
                }
                if (!insert && node instanceof AST_Return) {
                    if (compressor) {
                        var value = node.value && node.value.drop_side_effect_free(compressor, true);
                        return value ? make_node(AST_SimpleStatement, node, {
                            body: value
                        }) : make_node(AST_EmptyStatement, node);
                    }
                    return make_node(AST_SimpleStatement, node, {
                        body: node.value || make_node(AST_UnaryPrefix, node, {
                            operator: "void",
                            expression: make_node(AST_Number, node, {
                                value: 0
                            })
                        })
                    });
                }
                if (node instanceof AST_Class || node instanceof AST_Lambda && node !== self) {
                    return node;
                }
                if (node instanceof AST_Block) {
                    var index = node.body.length - 1;
                    if (index >= 0) {
                        node.body[index] = node.body[index].transform(tt);
                    }
                }
                else if (node instanceof AST_If) {
                    node.body = node.body.transform(tt);
                    if (node.alternative) {
                        node.alternative = node.alternative.transform(tt);
                    }
                }
                else if (node instanceof AST_With) {
                    node.body = node.body.transform(tt);
                }
                return node;
            });
            self.transform(tt);
        });
        function read_property(obj, key) {
            key = get_value(key);
            if (key instanceof AST_Node)
                return;
            var value;
            if (obj instanceof AST_Array) {
                var elements = obj.elements;
                if (key == "length")
                    return make_node_from_constant(elements.length, obj);
                if (typeof key == "number" && key in elements)
                    value = elements[key];
            }
            else if (obj instanceof AST_Object) {
                key = "" + key;
                var props = obj.properties;
                for (var i = props.length; --i >= 0;) {
                    var prop = props[i];
                    if (!(prop instanceof AST_ObjectKeyVal))
                        return;
                    if (!value && props[i].key === key)
                        value = props[i].value;
                }
            }
            return value instanceof AST_SymbolRef && value.fixed_value() || value;
        }
        function is_modified(compressor, tw, node, value, level, immutable) {
            var parent = tw.parent(level);
            var lhs = is_lhs(node, parent);
            if (lhs)
                return lhs;
            if (!immutable
                && parent instanceof AST_Call
                && parent.expression === node
                && !(value instanceof AST_Arrow)
                && !(value instanceof AST_Class)
                && !parent.is_callee_pure(compressor)
                && (!(value instanceof AST_Function)
                    || !(parent instanceof AST_New) && value.contains_this())) {
                return true;
            }
            if (parent instanceof AST_Array) {
                return is_modified(compressor, tw, parent, parent, level + 1);
            }
            if (parent instanceof AST_ObjectKeyVal && node === parent.value) {
                var obj = tw.parent(level + 1);
                return is_modified(compressor, tw, obj, obj, level + 2);
            }
            if (parent instanceof AST_PropAccess && parent.expression === node) {
                var prop = read_property(value, parent.property);
                return !immutable && is_modified(compressor, tw, parent, prop, level + 1);
            }
        }
        (function (def_reduce_vars) {
            def_reduce_vars(AST_Node, noop);
            function reset_def(compressor, def) {
                def.assignments = 0;
                def.chained = false;
                def.direct_access = false;
                def.escaped = 0;
                def.recursive_refs = 0;
                def.references = [];
                def.single_use = undefined;
                if (def.scope.pinned()) {
                    def.fixed = false;
                }
                else if (def.orig[0] instanceof AST_SymbolConst || !compressor.exposed(def)) {
                    def.fixed = def.init;
                }
                else {
                    def.fixed = false;
                }
            }
            function reset_variables(tw, compressor, node) {
                node.variables.forEach(function (def) {
                    reset_def(compressor, def);
                    if (def.fixed === null) {
                        tw.defs_to_safe_ids.set(def.id, tw.safe_ids);
                        mark(tw, def, true);
                    }
                    else if (def.fixed) {
                        tw.loop_ids.set(def.id, tw.in_loop);
                        mark(tw, def, true);
                    }
                });
            }
            function reset_block_variables(compressor, node) {
                if (node.block_scope)
                    node.block_scope.variables.forEach((def) => {
                        reset_def(compressor, def);
                    });
            }
            function push(tw) {
                tw.safe_ids = Object.create(tw.safe_ids);
            }
            function pop(tw) {
                tw.safe_ids = Object.getPrototypeOf(tw.safe_ids);
            }
            function mark(tw, def, safe) {
                tw.safe_ids[def.id] = safe;
            }
            function safe_to_read(tw, def) {
                if (def.single_use == "m")
                    return false;
                if (tw.safe_ids[def.id]) {
                    if (def.fixed == null) {
                        var orig = def.orig[0];
                        if (orig instanceof AST_SymbolFunarg || orig.name == "arguments")
                            return false;
                        def.fixed = make_node(AST_Undefined, orig);
                    }
                    return true;
                }
                return def.fixed instanceof AST_Defun;
            }
            function safe_to_assign(tw, def, scope, value) {
                if (def.fixed === undefined)
                    return true;
                let def_safe_ids;
                if (def.fixed === null
                    && (def_safe_ids = tw.defs_to_safe_ids.get(def.id))) {
                    def_safe_ids[def.id] = false;
                    tw.defs_to_safe_ids.delete(def.id);
                    return true;
                }
                if (!HOP(tw.safe_ids, def.id))
                    return false;
                if (!safe_to_read(tw, def))
                    return false;
                if (def.fixed === false)
                    return false;
                if (def.fixed != null && (!value || def.references.length > def.assignments))
                    return false;
                if (def.fixed instanceof AST_Defun) {
                    return value instanceof AST_Node && def.fixed.parent_scope === scope;
                }
                return def.orig.every((sym) => {
                    return !(sym instanceof AST_SymbolConst
                        || sym instanceof AST_SymbolDefun
                        || sym instanceof AST_SymbolLambda);
                });
            }
            function ref_once(tw, compressor, def) {
                return compressor.option("unused")
                    && !def.scope.pinned()
                    && def.references.length - def.recursive_refs == 1
                    && tw.loop_ids.get(def.id) === tw.in_loop;
            }
            function is_immutable(value) {
                if (!value)
                    return false;
                return value.is_constant()
                    || value instanceof AST_Lambda
                    || value instanceof AST_This;
            }
            // A definition "escapes" when its value can leave the point of use.
            // Example: `a = b || c`
            // In this example, "b" and "c" are escaping, because they're going into "a"
            //
            // def.escaped is != 0 when it escapes.
            //
            // When greater than 1, it means that N chained properties will be read off
            // of that def before an escape occurs. This is useful for evaluating
            // property accesses, where you need to know when to stop.
            function mark_escaped(tw, d, scope, node, value, level = 0, depth = 1) {
                var parent = tw.parent(level);
                if (value) {
                    if (value.is_constant())
                        return;
                    if (value instanceof AST_ClassExpression)
                        return;
                }
                if (parent instanceof AST_Assign && (parent.operator === "=" || parent.logical) && node === parent.right
                    || parent instanceof AST_Call && (node !== parent.expression || parent instanceof AST_New)
                    || parent instanceof AST_Exit && node === parent.value && node.scope !== d.scope
                    || parent instanceof AST_VarDef && node === parent.value
                    || parent instanceof AST_Yield && node === parent.value && node.scope !== d.scope) {
                    if (depth > 1 && !(value && value.is_constant_expression(scope)))
                        depth = 1;
                    if (!d.escaped || d.escaped > depth)
                        d.escaped = depth;
                    return;
                }
                else if (parent instanceof AST_Array
                    || parent instanceof AST_Await
                    || parent instanceof AST_Binary && lazy_op.has(parent.operator)
                    || parent instanceof AST_Conditional && node !== parent.condition
                    || parent instanceof AST_Expansion
                    || parent instanceof AST_Sequence && node === parent.tail_node()) {
                    mark_escaped(tw, d, scope, parent, parent, level + 1, depth);
                }
                else if (parent instanceof AST_ObjectKeyVal && node === parent.value) {
                    var obj = tw.parent(level + 1);
                    mark_escaped(tw, d, scope, obj, obj, level + 2, depth);
                }
                else if (parent instanceof AST_PropAccess && node === parent.expression) {
                    value = read_property(value, parent.property);
                    mark_escaped(tw, d, scope, parent, value, level + 1, depth + 1);
                    if (value)
                        return;
                }
                if (level > 0)
                    return;
                if (parent instanceof AST_Sequence && node !== parent.tail_node())
                    return;
                if (parent instanceof AST_SimpleStatement)
                    return;
                d.direct_access = true;
            }
            const suppress = node => walk(node, node => {
                if (!(node instanceof AST_Symbol))
                    return;
                var d = node.definition();
                if (!d)
                    return;
                if (node instanceof AST_SymbolRef)
                    d.references.push(node);
                d.fixed = false;
            });
            def_reduce_vars(AST_Accessor, function (tw, descend, compressor) {
                push(tw);
                reset_variables(tw, compressor, this);
                descend();
                pop(tw);
                return true;
            });
            def_reduce_vars(AST_Assign, function (tw, descend, compressor) {
                var node = this;
                if (node.left instanceof AST_Destructuring) {
                    suppress(node.left);
                    return;
                }
                const finish_walk = () => {
                    if (node.logical) {
                        node.left.walk(tw);
                        push(tw);
                        node.right.walk(tw);
                        pop(tw);
                        return true;
                    }
                };
                var sym = node.left;
                if (!(sym instanceof AST_SymbolRef))
                    return finish_walk();
                var def = sym.definition();
                var safe = safe_to_assign(tw, def, sym.scope, node.right);
                def.assignments++;
                if (!safe)
                    return finish_walk();
                var fixed = def.fixed;
                if (!fixed && node.operator != "=" && !node.logical)
                    return finish_walk();
                var eq = node.operator == "=";
                var value = eq ? node.right : node;
                if (is_modified(compressor, tw, node, value, 0))
                    return finish_walk();
                def.references.push(sym);
                if (!node.logical) {
                    if (!eq)
                        def.chained = true;
                    def.fixed = eq ? function () {
                        return node.right;
                    } : function () {
                        return make_node(AST_Binary, node, {
                            operator: node.operator.slice(0, -1),
                            left: fixed instanceof AST_Node ? fixed : fixed(),
                            right: node.right
                        });
                    };
                }
                if (node.logical) {
                    mark(tw, def, false);
                    push(tw);
                    node.right.walk(tw);
                    pop(tw);
                    return true;
                }
                mark(tw, def, false);
                node.right.walk(tw);
                mark(tw, def, true);
                mark_escaped(tw, def, sym.scope, node, value, 0, 1);
                return true;
            });
            def_reduce_vars(AST_Binary, function (tw) {
                if (!lazy_op.has(this.operator))
                    return;
                this.left.walk(tw);
                push(tw);
                this.right.walk(tw);
                pop(tw);
                return true;
            });
            def_reduce_vars(AST_Block, function (tw, descend, compressor) {
                reset_block_variables(compressor, this);
            });
            def_reduce_vars(AST_Case, function (tw) {
                push(tw);
                this.expression.walk(tw);
                pop(tw);
                push(tw);
                walk_body(this, tw);
                pop(tw);
                return true;
            });
            def_reduce_vars(AST_Class, function (tw, descend) {
                clear_flag(this, INLINED);
                push(tw);
                descend();
                pop(tw);
                return true;
            });
            def_reduce_vars(AST_Conditional, function (tw) {
                this.condition.walk(tw);
                push(tw);
                this.consequent.walk(tw);
                pop(tw);
                push(tw);
                this.alternative.walk(tw);
                pop(tw);
                return true;
            });
            def_reduce_vars(AST_Chain, function (tw, descend) {
                // Chains' conditions apply left-to-right, cumulatively.
                // If we walk normally we don't go in that order because we would pop before pushing again
                // Solution: AST_PropAccess and AST_Call push when they are optional, and never pop.
                // Then we pop everything when they are done being walked.
                const safe_ids = tw.safe_ids;
                descend();
                // Unroll back to start
                tw.safe_ids = safe_ids;
                return true;
            });
            def_reduce_vars(AST_Call, function (tw) {
                this.expression.walk(tw);
                if (this.optional) {
                    // Never pop -- it's popped at AST_Chain above
                    push(tw);
                }
                for (const arg of this.args)
                    arg.walk(tw);
                return true;
            });
            def_reduce_vars(AST_PropAccess, function (tw) {
                if (!this.optional)
                    return;
                this.expression.walk(tw);
                // Never pop -- it's popped at AST_Chain above
                push(tw);
                if (this.property instanceof AST_Node)
                    this.property.walk(tw);
                return true;
            });
            def_reduce_vars(AST_Default, function (tw, descend) {
                push(tw);
                descend();
                pop(tw);
                return true;
            });
            function mark_lambda(tw, descend, compressor) {
                clear_flag(this, INLINED);
                push(tw);
                reset_variables(tw, compressor, this);
                if (this.uses_arguments) {
                    descend();
                    pop(tw);
                    return;
                }
                var iife;
                if (!this.name
                    && (iife = tw.parent()) instanceof AST_Call
                    && iife.expression === this
                    && !iife.args.some(arg => arg instanceof AST_Expansion)
                    && this.argnames.every(arg_name => arg_name instanceof AST_Symbol)) {
                    // Virtually turn IIFE parameters into variable definitions:
                    //   (function(a,b) {...})(c,d) => (function() {var a=c,b=d; ...})()
                    // So existing transformation rules can work on them.
                    this.argnames.forEach((arg, i) => {
                        if (!arg.definition)
                            return;
                        var d = arg.definition();
                        // Avoid setting fixed when there's more than one origin for a variable value
                        if (d.orig.length > 1)
                            return;
                        if (d.fixed === undefined && (!this.uses_arguments || tw.has_directive("use strict"))) {
                            d.fixed = function () {
                                return iife.args[i] || make_node(AST_Undefined, iife);
                            };
                            tw.loop_ids.set(d.id, tw.in_loop);
                            mark(tw, d, true);
                        }
                        else {
                            d.fixed = false;
                        }
                    });
                }
                descend();
                pop(tw);
                return true;
            }
            def_reduce_vars(AST_Lambda, mark_lambda);
            def_reduce_vars(AST_Do, function (tw, descend, compressor) {
                reset_block_variables(compressor, this);
                const saved_loop = tw.in_loop;
                tw.in_loop = this;
                push(tw);
                this.body.walk(tw);
                if (has_break_or_continue(this)) {
                    pop(tw);
                    push(tw);
                }
                this.condition.walk(tw);
                pop(tw);
                tw.in_loop = saved_loop;
                return true;
            });
            def_reduce_vars(AST_For, function (tw, descend, compressor) {
                reset_block_variables(compressor, this);
                if (this.init)
                    this.init.walk(tw);
                const saved_loop = tw.in_loop;
                tw.in_loop = this;
                push(tw);
                if (this.condition)
                    this.condition.walk(tw);
                this.body.walk(tw);
                if (this.step) {
                    if (has_break_or_continue(this)) {
                        pop(tw);
                        push(tw);
                    }
                    this.step.walk(tw);
                }
                pop(tw);
                tw.in_loop = saved_loop;
                return true;
            });
            def_reduce_vars(AST_ForIn, function (tw, descend, compressor) {
                reset_block_variables(compressor, this);
                suppress(this.init);
                this.object.walk(tw);
                const saved_loop = tw.in_loop;
                tw.in_loop = this;
                push(tw);
                this.body.walk(tw);
                pop(tw);
                tw.in_loop = saved_loop;
                return true;
            });
            def_reduce_vars(AST_If, function (tw) {
                this.condition.walk(tw);
                push(tw);
                this.body.walk(tw);
                pop(tw);
                if (this.alternative) {
                    push(tw);
                    this.alternative.walk(tw);
                    pop(tw);
                }
                return true;
            });
            def_reduce_vars(AST_LabeledStatement, function (tw) {
                push(tw);
                this.body.walk(tw);
                pop(tw);
                return true;
            });
            def_reduce_vars(AST_SymbolCatch, function () {
                this.definition().fixed = false;
            });
            def_reduce_vars(AST_SymbolRef, function (tw, descend, compressor) {
                var d = this.definition();
                d.references.push(this);
                if (d.references.length == 1
                    && !d.fixed
                    && d.orig[0] instanceof AST_SymbolDefun) {
                    tw.loop_ids.set(d.id, tw.in_loop);
                }
                var fixed_value;
                if (d.fixed === undefined || !safe_to_read(tw, d)) {
                    d.fixed = false;
                }
                else if (d.fixed) {
                    fixed_value = this.fixed_value();
                    if (fixed_value instanceof AST_Lambda
                        && recursive_ref(tw, d)) {
                        d.recursive_refs++;
                    }
                    else if (fixed_value
                        && !compressor.exposed(d)
                        && ref_once(tw, compressor, d)) {
                        d.single_use =
                            fixed_value instanceof AST_Lambda && !fixed_value.pinned()
                                || fixed_value instanceof AST_Class
                                || d.scope === this.scope && fixed_value.is_constant_expression();
                    }
                    else {
                        d.single_use = false;
                    }
                    if (is_modified(compressor, tw, this, fixed_value, 0, is_immutable(fixed_value))) {
                        if (d.single_use) {
                            d.single_use = "m";
                        }
                        else {
                            d.fixed = false;
                        }
                    }
                }
                mark_escaped(tw, d, this.scope, this, fixed_value, 0, 1);
            });
            def_reduce_vars(AST_Toplevel, function (tw, descend, compressor) {
                this.globals.forEach(function (def) {
                    reset_def(compressor, def);
                });
                reset_variables(tw, compressor, this);
            });
            def_reduce_vars(AST_Try, function (tw, descend, compressor) {
                reset_block_variables(compressor, this);
                push(tw);
                walk_body(this, tw);
                pop(tw);
                if (this.bcatch) {
                    push(tw);
                    this.bcatch.walk(tw);
                    pop(tw);
                }
                if (this.bfinally)
                    this.bfinally.walk(tw);
                return true;
            });
            def_reduce_vars(AST_Unary, function (tw) {
                var node = this;
                if (node.operator !== "++" && node.operator !== "--")
                    return;
                var exp = node.expression;
                if (!(exp instanceof AST_SymbolRef))
                    return;
                var def = exp.definition();
                var safe = safe_to_assign(tw, def, exp.scope, true);
                def.assignments++;
                if (!safe)
                    return;
                var fixed = def.fixed;
                if (!fixed)
                    return;
                def.references.push(exp);
                def.chained = true;
                def.fixed = function () {
                    return make_node(AST_Binary, node, {
                        operator: node.operator.slice(0, -1),
                        left: make_node(AST_UnaryPrefix, node, {
                            operator: "+",
                            expression: fixed instanceof AST_Node ? fixed : fixed()
                        }),
                        right: make_node(AST_Number, node, {
                            value: 1
                        })
                    });
                };
                mark(tw, def, true);
                return true;
            });
            def_reduce_vars(AST_VarDef, function (tw, descend) {
                var node = this;
                if (node.name instanceof AST_Destructuring) {
                    suppress(node.name);
                    return;
                }
                var d = node.name.definition();
                if (node.value) {
                    if (safe_to_assign(tw, d, node.name.scope, node.value)) {
                        d.fixed = function () {
                            return node.value;
                        };
                        tw.loop_ids.set(d.id, tw.in_loop);
                        mark(tw, d, false);
                        descend();
                        mark(tw, d, true);
                        return true;
                    }
                    else {
                        d.fixed = false;
                    }
                }
            });
            def_reduce_vars(AST_While, function (tw, descend, compressor) {
                reset_block_variables(compressor, this);
                const saved_loop = tw.in_loop;
                tw.in_loop = this;
                push(tw);
                descend();
                pop(tw);
                tw.in_loop = saved_loop;
                return true;
            });
        })(function (node, func) {
            node.DEFMETHOD("reduce_vars", func);
        });
        AST_Toplevel.DEFMETHOD("reset_opt_flags", function (compressor) {
            const self = this;
            const reduce_vars = compressor.option("reduce_vars");
            const preparation = new TreeWalker(function (node, descend) {
                clear_flag(node, CLEAR_BETWEEN_PASSES);
                if (reduce_vars) {
                    if (compressor.top_retain
                        && node instanceof AST_Defun // Only functions are retained
                        && preparation.parent() === self) {
                        set_flag(node, TOP);
                    }
                    return node.reduce_vars(preparation, descend, compressor);
                }
            });
            // Stack of look-up tables to keep track of whether a `SymbolDef` has been
            // properly assigned before use:
            // - `push()` & `pop()` when visiting conditional branches
            preparation.safe_ids = Object.create(null);
            preparation.in_loop = null;
            preparation.loop_ids = new Map();
            preparation.defs_to_safe_ids = new Map();
            self.walk(preparation);
        });
        AST_Symbol.DEFMETHOD("fixed_value", function () {
            var fixed = this.thedef.fixed;
            if (!fixed || fixed instanceof AST_Node)
                return fixed;
            return fixed();
        });
        AST_SymbolRef.DEFMETHOD("is_immutable", function () {
            var orig = this.definition().orig;
            return orig.length == 1 && orig[0] instanceof AST_SymbolLambda;
        });
        function is_func_expr(node) {
            return node instanceof AST_Arrow || node instanceof AST_Function;
        }
        function is_lhs_read_only(lhs) {
            if (lhs instanceof AST_This)
                return true;
            if (lhs instanceof AST_SymbolRef)
                return lhs.definition().orig[0] instanceof AST_SymbolLambda;
            if (lhs instanceof AST_PropAccess) {
                lhs = lhs.expression;
                if (lhs instanceof AST_SymbolRef) {
                    if (lhs.is_immutable())
                        return false;
                    lhs = lhs.fixed_value();
                }
                if (!lhs)
                    return true;
                if (lhs instanceof AST_RegExp)
                    return false;
                if (lhs instanceof AST_Constant)
                    return true;
                return is_lhs_read_only(lhs);
            }
            return false;
        }
        function is_ref_of(ref, type) {
            if (!(ref instanceof AST_SymbolRef))
                return false;
            var orig = ref.definition().orig;
            for (var i = orig.length; --i >= 0;) {
                if (orig[i] instanceof type)
                    return true;
            }
        }
        function find_scope(tw) {
            for (let i = 0;; i++) {
                const p = tw.parent(i);
                if (p instanceof AST_Toplevel)
                    return p;
                if (p instanceof AST_Lambda)
                    return p;
                if (p.block_scope)
                    return p.block_scope;
            }
        }
        function find_variable(compressor, name) {
            var scope, i = 0;
            while (scope = compressor.parent(i++)) {
                if (scope instanceof AST_Scope)
                    break;
                if (scope instanceof AST_Catch && scope.argname) {
                    scope = scope.argname.definition().scope;
                    break;
                }
            }
            return scope.find_variable(name);
        }
        function make_sequence(orig, expressions) {
            if (expressions.length == 1)
                return expressions[0];
            if (expressions.length == 0)
                throw new Error("trying to create a sequence with length zero!");
            return make_node(AST_Sequence, orig, {
                expressions: expressions.reduce(merge_sequence, [])
            });
        }
        function make_node_from_constant(val, orig) {
            switch (typeof val) {
                case "string":
                    return make_node(AST_String, orig, {
                        value: val
                    });
                case "number":
                    if (isNaN(val))
                        return make_node(AST_NaN, orig);
                    if (isFinite(val)) {
                        return 1 / val < 0 ? make_node(AST_UnaryPrefix, orig, {
                            operator: "-",
                            expression: make_node(AST_Number, orig, { value: -val })
                        }) : make_node(AST_Number, orig, { value: val });
                    }
                    return val < 0 ? make_node(AST_UnaryPrefix, orig, {
                        operator: "-",
                        expression: make_node(AST_Infinity, orig)
                    }) : make_node(AST_Infinity, orig);
                case "boolean":
                    return make_node(val ? AST_True : AST_False, orig);
                case "undefined":
                    return make_node(AST_Undefined, orig);
                default:
                    if (val === null) {
                        return make_node(AST_Null, orig, { value: null });
                    }
                    if (val instanceof RegExp) {
                        return make_node(AST_RegExp, orig, {
                            value: {
                                source: regexp_source_fix(val.source),
                                flags: val.flags
                            }
                        });
                    }
                    throw new Error(string_template("Can't handle constant of type: {type}", {
                        type: typeof val
                    }));
            }
        }
        // we shouldn't compress (1,func)(something) to
        // func(something) because that changes the meaning of
        // the func (becomes lexical instead of global).
        function maintain_this_binding(parent, orig, val) {
            if (parent instanceof AST_UnaryPrefix && parent.operator == "delete"
                || parent instanceof AST_Call && parent.expression === orig
                    && (val instanceof AST_PropAccess || val instanceof AST_SymbolRef && val.name == "eval")) {
                return make_sequence(orig, [make_node(AST_Number, orig, { value: 0 }), val]);
            }
            return val;
        }
        function merge_sequence(array, node) {
            if (node instanceof AST_Sequence) {
                array.push(...node.expressions);
            }
            else {
                array.push(node);
            }
            return array;
        }
        function as_statement_array(thing) {
            if (thing === null)
                return [];
            if (thing instanceof AST_BlockStatement)
                return thing.body;
            if (thing instanceof AST_EmptyStatement)
                return [];
            if (thing instanceof AST_Statement)
                return [thing];
            throw new Error("Can't convert thing to statement array");
        }
        function is_empty(thing) {
            if (thing === null)
                return true;
            if (thing instanceof AST_EmptyStatement)
                return true;
            if (thing instanceof AST_BlockStatement)
                return thing.body.length == 0;
            return false;
        }
        function can_be_evicted_from_block(node) {
            return !(node instanceof AST_DefClass ||
                node instanceof AST_Defun ||
                node instanceof AST_Let ||
                node instanceof AST_Const ||
                node instanceof AST_Export ||
                node instanceof AST_Import);
        }
        function loop_body(x) {
            if (x instanceof AST_IterationStatement) {
                return x.body instanceof AST_BlockStatement ? x.body : x;
            }
            return x;
        }
        function is_iife_call(node) {
            // Used to determine whether the node can benefit from negation.
            // Not the case with arrow functions (you need an extra set of parens).
            if (node.TYPE != "Call")
                return false;
            return node.expression instanceof AST_Function || is_iife_call(node.expression);
        }
        function is_undeclared_ref(node) {
            return node instanceof AST_SymbolRef && node.definition().undeclared;
        }
        var global_names = makePredicate("Array Boolean clearInterval clearTimeout console Date decodeURI decodeURIComponent encodeURI encodeURIComponent Error escape eval EvalError Function isFinite isNaN JSON Math Number parseFloat parseInt RangeError ReferenceError RegExp Object setInterval setTimeout String SyntaxError TypeError unescape URIError");
        AST_SymbolRef.DEFMETHOD("is_declared", function (compressor) {
            return !this.definition().undeclared
                || compressor.option("unsafe") && global_names.has(this.name);
        });
        var identifier_atom = makePredicate("Infinity NaN undefined");
        function is_identifier_atom(node) {
            return node instanceof AST_Infinity
                || node instanceof AST_NaN
                || node instanceof AST_Undefined;
        }
        // Tighten a bunch of statements together. Used whenever there is a block.
        function tighten_body(statements, compressor) {
            var in_loop, in_try;
            var scope = compressor.find_parent(AST_Scope).get_defun_scope();
            find_loop_scope_try();
            var CHANGED, max_iter = 10;
            do {
                CHANGED = false;
                eliminate_spurious_blocks(statements);
                if (compressor.option("dead_code")) {
                    eliminate_dead_code(statements, compressor);
                }
                if (compressor.option("if_return")) {
                    handle_if_return(statements, compressor);
                }
                if (compressor.sequences_limit > 0) {
                    sequencesize(statements, compressor);
                    sequencesize_2(statements, compressor);
                }
                if (compressor.option("join_vars")) {
                    join_consecutive_vars(statements);
                }
                if (compressor.option("collapse_vars")) {
                    collapse(statements, compressor);
                }
            } while (CHANGED && max_iter-- > 0);
            function find_loop_scope_try() {
                var node = compressor.self(), level = 0;
                do {
                    if (node instanceof AST_Catch || node instanceof AST_Finally) {
                        level++;
                    }
                    else if (node instanceof AST_IterationStatement) {
                        in_loop = true;
                    }
                    else if (node instanceof AST_Scope) {
                        scope = node;
                        break;
                    }
                    else if (node instanceof AST_Try) {
                        in_try = true;
                    }
                } while (node = compressor.parent(level++));
            }
            // Search from right to left for assignment-like expressions:
            // - `var a = x;`
            // - `a = x;`
            // - `++a`
            // For each candidate, scan from left to right for first usage, then try
            // to fold assignment into the site for compression.
            // Will not attempt to collapse assignments into or past code blocks
            // which are not sequentially executed, e.g. loops and conditionals.
            function collapse(statements, compressor) {
                if (scope.pinned())
                    return statements;
                var args;
                var candidates = [];
                var stat_index = statements.length;
                var scanner = new TreeTransformer(function (node) {
                    if (abort)
                        return node;
                    // Skip nodes before `candidate` as quickly as possible
                    if (!hit) {
                        if (node !== hit_stack[hit_index])
                            return node;
                        hit_index++;
                        if (hit_index < hit_stack.length)
                            return handle_custom_scan_order(node);
                        hit = true;
                        stop_after = find_stop(node, 0);
                        if (stop_after === node)
                            abort = true;
                        return node;
                    }
                    // Stop immediately if these node types are encountered
                    var parent = scanner.parent();
                    if (node instanceof AST_Assign
                        && (node.logical || node.operator != "=" && lhs.equivalent_to(node.left))
                        || node instanceof AST_Await
                        || node instanceof AST_Call && lhs instanceof AST_PropAccess && lhs.equivalent_to(node.expression)
                        || node instanceof AST_Debugger
                        || node instanceof AST_Destructuring
                        || node instanceof AST_Expansion
                            && node.expression instanceof AST_Symbol
                            && (node.expression instanceof AST_This
                                || node.expression.definition().references.length > 1)
                        || node instanceof AST_IterationStatement && !(node instanceof AST_For)
                        || node instanceof AST_LoopControl
                        || node instanceof AST_Try
                        || node instanceof AST_With
                        || node instanceof AST_Yield
                        || node instanceof AST_Export
                        || node instanceof AST_Class
                        || parent instanceof AST_For && node !== parent.init
                        || !replace_all
                            && (node instanceof AST_SymbolRef
                                && !node.is_declared(compressor)
                                && !pure_prop_access_globals.has(node))
                        || node instanceof AST_SymbolRef
                            && parent instanceof AST_Call
                            && has_annotation(parent, _NOINLINE)) {
                        abort = true;
                        return node;
                    }
                    // Stop only if candidate is found within conditional branches
                    if (!stop_if_hit && (!lhs_local || !replace_all)
                        && (parent instanceof AST_Binary && lazy_op.has(parent.operator) && parent.left !== node
                            || parent instanceof AST_Conditional && parent.condition !== node
                            || parent instanceof AST_If && parent.condition !== node)) {
                        stop_if_hit = parent;
                    }
                    // Replace variable with assignment when found
                    if (can_replace
                        && !(node instanceof AST_SymbolDeclaration)
                        && lhs.equivalent_to(node)) {
                        if (stop_if_hit) {
                            abort = true;
                            return node;
                        }
                        if (is_lhs(node, parent)) {
                            if (value_def)
                                replaced++;
                            return node;
                        }
                        else {
                            replaced++;
                            if (value_def && candidate instanceof AST_VarDef)
                                return node;
                        }
                        CHANGED = abort = true;
                        if (candidate instanceof AST_UnaryPostfix) {
                            return make_node(AST_UnaryPrefix, candidate, candidate);
                        }
                        if (candidate instanceof AST_VarDef) {
                            var def = candidate.name.definition();
                            var value = candidate.value;
                            if (def.references.length - def.replaced == 1 && !compressor.exposed(def)) {
                                def.replaced++;
                                if (funarg && is_identifier_atom(value)) {
                                    return value.transform(compressor);
                                }
                                else {
                                    return maintain_this_binding(parent, node, value);
                                }
                            }
                            return make_node(AST_Assign, candidate, {
                                operator: "=",
                                logical: false,
                                left: make_node(AST_SymbolRef, candidate.name, candidate.name),
                                right: value
                            });
                        }
                        clear_flag(candidate, WRITE_ONLY);
                        return candidate;
                    }
                    // These node types have child nodes that execute sequentially,
                    // but are otherwise not safe to scan into or beyond them.
                    var sym;
                    if (node instanceof AST_Call
                        || node instanceof AST_Exit
                            && (side_effects || lhs instanceof AST_PropAccess || may_modify(lhs))
                        || node instanceof AST_PropAccess
                            && (side_effects || node.expression.may_throw_on_access(compressor))
                        || node instanceof AST_SymbolRef
                            && (lvalues.get(node.name) || side_effects && may_modify(node))
                        || node instanceof AST_VarDef && node.value
                            && (lvalues.has(node.name.name) || side_effects && may_modify(node.name))
                        || (sym = is_lhs(node.left, node))
                            && (sym instanceof AST_PropAccess || lvalues.has(sym.name))
                        || may_throw
                            && (in_try ? node.has_side_effects(compressor) : side_effects_external(node))) {
                        stop_after = node;
                        if (node instanceof AST_Scope)
                            abort = true;
                    }
                    return handle_custom_scan_order(node);
                }, function (node) {
                    if (abort)
                        return;
                    if (stop_after === node)
                        abort = true;
                    if (stop_if_hit === node)
                        stop_if_hit = null;
                });
                var multi_replacer = new TreeTransformer(function (node) {
                    if (abort)
                        return node;
                    // Skip nodes before `candidate` as quickly as possible
                    if (!hit) {
                        if (node !== hit_stack[hit_index])
                            return node;
                        hit_index++;
                        if (hit_index < hit_stack.length)
                            return;
                        hit = true;
                        return node;
                    }
                    // Replace variable when found
                    if (node instanceof AST_SymbolRef
                        && node.name == def.name) {
                        if (!--replaced)
                            abort = true;
                        if (is_lhs(node, multi_replacer.parent()))
                            return node;
                        def.replaced++;
                        value_def.replaced--;
                        return candidate.value;
                    }
                    // Skip (non-executed) functions and (leading) default case in switch statements
                    if (node instanceof AST_Default || node instanceof AST_Scope)
                        return node;
                });
                while (--stat_index >= 0) {
                    // Treat parameters as collapsible in IIFE, i.e.
                    //   function(a, b){ ... }(x());
                    // would be translated into equivalent assignments:
                    //   var a = x(), b = undefined;
                    if (stat_index == 0 && compressor.option("unused"))
                        extract_args();
                    // Find collapsible assignments
                    var hit_stack = [];
                    extract_candidates(statements[stat_index]);
                    while (candidates.length > 0) {
                        hit_stack = candidates.pop();
                        var hit_index = 0;
                        var candidate = hit_stack[hit_stack.length - 1];
                        var value_def = null;
                        var stop_after = null;
                        var stop_if_hit = null;
                        var lhs = get_lhs(candidate);
                        if (!lhs || is_lhs_read_only(lhs) || lhs.has_side_effects(compressor))
                            continue;
                        // Locate symbols which may execute code outside of scanning range
                        var lvalues = get_lvalues(candidate);
                        var lhs_local = is_lhs_local(lhs);
                        if (lhs instanceof AST_SymbolRef)
                            lvalues.set(lhs.name, false);
                        var side_effects = value_has_side_effects(candidate);
                        var replace_all = replace_all_symbols();
                        var may_throw = candidate.may_throw(compressor);
                        var funarg = candidate.name instanceof AST_SymbolFunarg;
                        var hit = funarg;
                        var abort = false, replaced = 0, can_replace = !args || !hit;
                        if (!can_replace) {
                            for (var j = compressor.self().argnames.lastIndexOf(candidate.name) + 1; !abort && j < args.length; j++) {
                                args[j].transform(scanner);
                            }
                            can_replace = true;
                        }
                        for (var i = stat_index; !abort && i < statements.length; i++) {
                            statements[i].transform(scanner);
                        }
                        if (value_def) {
                            var def = candidate.name.definition();
                            if (abort && def.references.length - def.replaced > replaced)
                                replaced = false;
                            else {
                                abort = false;
                                hit_index = 0;
                                hit = funarg;
                                for (var i = stat_index; !abort && i < statements.length; i++) {
                                    statements[i].transform(multi_replacer);
                                }
                                value_def.single_use = false;
                            }
                        }
                        if (replaced && !remove_candidate(candidate))
                            statements.splice(stat_index, 1);
                    }
                }
                function handle_custom_scan_order(node) {
                    // Skip (non-executed) functions
                    if (node instanceof AST_Scope)
                        return node;
                    // Scan case expressions first in a switch statement
                    if (node instanceof AST_Switch) {
                        node.expression = node.expression.transform(scanner);
                        for (var i = 0, len = node.body.length; !abort && i < len; i++) {
                            var branch = node.body[i];
                            if (branch instanceof AST_Case) {
                                if (!hit) {
                                    if (branch !== hit_stack[hit_index])
                                        continue;
                                    hit_index++;
                                }
                                branch.expression = branch.expression.transform(scanner);
                                if (!replace_all)
                                    break;
                            }
                        }
                        abort = true;
                        return node;
                    }
                }
                function redefined_within_scope(def, scope) {
                    if (def.global)
                        return false;
                    let cur_scope = def.scope;
                    while (cur_scope && cur_scope !== scope) {
                        if (cur_scope.variables.has(def.name))
                            return true;
                        cur_scope = cur_scope.parent_scope;
                    }
                    return false;
                }
                function has_overlapping_symbol(fn, arg, fn_strict) {
                    var found = false, scan_this = !(fn instanceof AST_Arrow);
                    arg.walk(new TreeWalker(function (node, descend) {
                        if (found)
                            return true;
                        if (node instanceof AST_SymbolRef && (fn.variables.has(node.name) || redefined_within_scope(node.definition(), fn))) {
                            var s = node.definition().scope;
                            if (s !== scope)
                                while (s = s.parent_scope) {
                                    if (s === scope)
                                        return true;
                                }
                            return found = true;
                        }
                        if ((fn_strict || scan_this) && node instanceof AST_This) {
                            return found = true;
                        }
                        if (node instanceof AST_Scope && !(node instanceof AST_Arrow)) {
                            var prev = scan_this;
                            scan_this = false;
                            descend();
                            scan_this = prev;
                            return true;
                        }
                    }));
                    return found;
                }
                function extract_args() {
                    var iife, fn = compressor.self();
                    if (is_func_expr(fn)
                        && !fn.name
                        && !fn.uses_arguments
                        && !fn.pinned()
                        && (iife = compressor.parent()) instanceof AST_Call
                        && iife.expression === fn
                        && iife.args.every((arg) => !(arg instanceof AST_Expansion))) {
                        var fn_strict = compressor.has_directive("use strict");
                        if (fn_strict && !member(fn_strict, fn.body))
                            fn_strict = false;
                        var len = fn.argnames.length;
                        args = iife.args.slice(len);
                        var names = new Set();
                        for (var i = len; --i >= 0;) {
                            var sym = fn.argnames[i];
                            var arg = iife.args[i];
                            // The following two line fix is a duplicate of the fix at
                            // https://github.com/terser/terser/commit/011d3eb08cefe6922c7d1bdfa113fc4aeaca1b75
                            // This might mean that these two pieces of code (one here in collapse_vars and another in reduce_vars
                            // Might be doing the exact same thing.
                            const def = sym.definition && sym.definition();
                            const is_reassigned = def && def.orig.length > 1;
                            if (is_reassigned)
                                continue;
                            args.unshift(make_node(AST_VarDef, sym, {
                                name: sym,
                                value: arg
                            }));
                            if (names.has(sym.name))
                                continue;
                            names.add(sym.name);
                            if (sym instanceof AST_Expansion) {
                                var elements = iife.args.slice(i);
                                if (elements.every((arg) => !has_overlapping_symbol(fn, arg, fn_strict))) {
                                    candidates.unshift([make_node(AST_VarDef, sym, {
                                            name: sym.expression,
                                            value: make_node(AST_Array, iife, {
                                                elements: elements
                                            })
                                        })]);
                                }
                            }
                            else {
                                if (!arg) {
                                    arg = make_node(AST_Undefined, sym).transform(compressor);
                                }
                                else if (arg instanceof AST_Lambda && arg.pinned()
                                    || has_overlapping_symbol(fn, arg, fn_strict)) {
                                    arg = null;
                                }
                                if (arg)
                                    candidates.unshift([make_node(AST_VarDef, sym, {
                                            name: sym,
                                            value: arg
                                        })]);
                            }
                        }
                    }
                }
                function extract_candidates(expr) {
                    hit_stack.push(expr);
                    if (expr instanceof AST_Assign) {
                        if (!expr.left.has_side_effects(compressor)
                            && !(expr.right instanceof AST_Chain)) {
                            candidates.push(hit_stack.slice());
                        }
                        extract_candidates(expr.right);
                    }
                    else if (expr instanceof AST_Binary) {
                        extract_candidates(expr.left);
                        extract_candidates(expr.right);
                    }
                    else if (expr instanceof AST_Call && !has_annotation(expr, _NOINLINE)) {
                        extract_candidates(expr.expression);
                        expr.args.forEach(extract_candidates);
                    }
                    else if (expr instanceof AST_Case) {
                        extract_candidates(expr.expression);
                    }
                    else if (expr instanceof AST_Conditional) {
                        extract_candidates(expr.condition);
                        extract_candidates(expr.consequent);
                        extract_candidates(expr.alternative);
                    }
                    else if (expr instanceof AST_Definitions) {
                        var len = expr.definitions.length;
                        // limit number of trailing variable definitions for consideration
                        var i = len - 200;
                        if (i < 0)
                            i = 0;
                        for (; i < len; i++) {
                            extract_candidates(expr.definitions[i]);
                        }
                    }
                    else if (expr instanceof AST_DWLoop) {
                        extract_candidates(expr.condition);
                        if (!(expr.body instanceof AST_Block)) {
                            extract_candidates(expr.body);
                        }
                    }
                    else if (expr instanceof AST_Exit) {
                        if (expr.value)
                            extract_candidates(expr.value);
                    }
                    else if (expr instanceof AST_For) {
                        if (expr.init)
                            extract_candidates(expr.init);
                        if (expr.condition)
                            extract_candidates(expr.condition);
                        if (expr.step)
                            extract_candidates(expr.step);
                        if (!(expr.body instanceof AST_Block)) {
                            extract_candidates(expr.body);
                        }
                    }
                    else if (expr instanceof AST_ForIn) {
                        extract_candidates(expr.object);
                        if (!(expr.body instanceof AST_Block)) {
                            extract_candidates(expr.body);
                        }
                    }
                    else if (expr instanceof AST_If) {
                        extract_candidates(expr.condition);
                        if (!(expr.body instanceof AST_Block)) {
                            extract_candidates(expr.body);
                        }
                        if (expr.alternative && !(expr.alternative instanceof AST_Block)) {
                            extract_candidates(expr.alternative);
                        }
                    }
                    else if (expr instanceof AST_Sequence) {
                        expr.expressions.forEach(extract_candidates);
                    }
                    else if (expr instanceof AST_SimpleStatement) {
                        extract_candidates(expr.body);
                    }
                    else if (expr instanceof AST_Switch) {
                        extract_candidates(expr.expression);
                        expr.body.forEach(extract_candidates);
                    }
                    else if (expr instanceof AST_Unary) {
                        if (expr.operator == "++" || expr.operator == "--") {
                            candidates.push(hit_stack.slice());
                        }
                    }
                    else if (expr instanceof AST_VarDef) {
                        if (expr.value && !(expr.value instanceof AST_Chain)) {
                            candidates.push(hit_stack.slice());
                            extract_candidates(expr.value);
                        }
                    }
                    hit_stack.pop();
                }
                function find_stop(node, level, write_only) {
                    var parent = scanner.parent(level);
                    if (parent instanceof AST_Assign) {
                        if (write_only
                            && !parent.logical
                            && !(parent.left instanceof AST_PropAccess
                                || lvalues.has(parent.left.name))) {
                            return find_stop(parent, level + 1, write_only);
                        }
                        return node;
                    }
                    if (parent instanceof AST_Binary) {
                        if (write_only && (!lazy_op.has(parent.operator) || parent.left === node)) {
                            return find_stop(parent, level + 1, write_only);
                        }
                        return node;
                    }
                    if (parent instanceof AST_Call)
                        return node;
                    if (parent instanceof AST_Case)
                        return node;
                    if (parent instanceof AST_Conditional) {
                        if (write_only && parent.condition === node) {
                            return find_stop(parent, level + 1, write_only);
                        }
                        return node;
                    }
                    if (parent instanceof AST_Definitions) {
                        return find_stop(parent, level + 1, true);
                    }
                    if (parent instanceof AST_Exit) {
                        return write_only ? find_stop(parent, level + 1, write_only) : node;
                    }
                    if (parent instanceof AST_If) {
                        if (write_only && parent.condition === node) {
                            return find_stop(parent, level + 1, write_only);
                        }
                        return node;
                    }
                    if (parent instanceof AST_IterationStatement)
                        return node;
                    if (parent instanceof AST_Sequence) {
                        return find_stop(parent, level + 1, parent.tail_node() !== node);
                    }
                    if (parent instanceof AST_SimpleStatement) {
                        return find_stop(parent, level + 1, true);
                    }
                    if (parent instanceof AST_Switch)
                        return node;
                    if (parent instanceof AST_VarDef)
                        return node;
                    return null;
                }
                function mangleable_var(var_def) {
                    var value = var_def.value;
                    if (!(value instanceof AST_SymbolRef))
                        return;
                    if (value.name == "arguments")
                        return;
                    var def = value.definition();
                    if (def.undeclared)
                        return;
                    return value_def = def;
                }
                function get_lhs(expr) {
                    if (expr instanceof AST_Assign && expr.logical) {
                        return false;
                    }
                    else if (expr instanceof AST_VarDef && expr.name instanceof AST_SymbolDeclaration) {
                        var def = expr.name.definition();
                        if (!member(expr.name, def.orig))
                            return;
                        var referenced = def.references.length - def.replaced;
                        if (!referenced)
                            return;
                        var declared = def.orig.length - def.eliminated;
                        if (declared > 1 && !(expr.name instanceof AST_SymbolFunarg)
                            || (referenced > 1 ? mangleable_var(expr) : !compressor.exposed(def))) {
                            return make_node(AST_SymbolRef, expr.name, expr.name);
                        }
                    }
                    else {
                        const lhs = expr instanceof AST_Assign
                            ? expr.left
                            : expr.expression;
                        return !is_ref_of(lhs, AST_SymbolConst)
                            && !is_ref_of(lhs, AST_SymbolLet) && lhs;
                    }
                }
                function get_rvalue(expr) {
                    if (expr instanceof AST_Assign) {
                        return expr.right;
                    }
                    else {
                        return expr.value;
                    }
                }
                function get_lvalues(expr) {
                    var lvalues = new Map();
                    if (expr instanceof AST_Unary)
                        return lvalues;
                    var tw = new TreeWalker(function (node) {
                        var sym = node;
                        while (sym instanceof AST_PropAccess)
                            sym = sym.expression;
                        if (sym instanceof AST_SymbolRef || sym instanceof AST_This) {
                            lvalues.set(sym.name, lvalues.get(sym.name) || is_modified(compressor, tw, node, node, 0));
                        }
                    });
                    get_rvalue(expr).walk(tw);
                    return lvalues;
                }
                function remove_candidate(expr) {
                    if (expr.name instanceof AST_SymbolFunarg) {
                        var iife = compressor.parent(), argnames = compressor.self().argnames;
                        var index = argnames.indexOf(expr.name);
                        if (index < 0) {
                            iife.args.length = Math.min(iife.args.length, argnames.length - 1);
                        }
                        else {
                            var args = iife.args;
                            if (args[index])
                                args[index] = make_node(AST_Number, args[index], {
                                    value: 0
                                });
                        }
                        return true;
                    }
                    var found = false;
                    return statements[stat_index].transform(new TreeTransformer(function (node, descend, in_list) {
                        if (found)
                            return node;
                        if (node === expr || node.body === expr) {
                            found = true;
                            if (node instanceof AST_VarDef) {
                                node.value = node.name instanceof AST_SymbolConst
                                    ? make_node(AST_Undefined, node.value) // `const` always needs value.
                                    : null;
                                return node;
                            }
                            return in_list ? MAP.skip : null;
                        }
                    }, function (node) {
                        if (node instanceof AST_Sequence)
                            switch (node.expressions.length) {
                                case 0: return null;
                                case 1: return node.expressions[0];
                            }
                    }));
                }
                function is_lhs_local(lhs) {
                    while (lhs instanceof AST_PropAccess)
                        lhs = lhs.expression;
                    return lhs instanceof AST_SymbolRef
                        && lhs.definition().scope === scope
                        && !(in_loop
                            && (lvalues.has(lhs.name)
                                || candidate instanceof AST_Unary
                                || (candidate instanceof AST_Assign
                                    && !candidate.logical
                                    && candidate.operator != "=")));
                }
                function value_has_side_effects(expr) {
                    if (expr instanceof AST_Unary)
                        return unary_side_effects.has(expr.operator);
                    return get_rvalue(expr).has_side_effects(compressor);
                }
                function replace_all_symbols() {
                    if (side_effects)
                        return false;
                    if (value_def)
                        return true;
                    if (lhs instanceof AST_SymbolRef) {
                        var def = lhs.definition();
                        if (def.references.length - def.replaced == (candidate instanceof AST_VarDef ? 1 : 2)) {
                            return true;
                        }
                    }
                    return false;
                }
                function may_modify(sym) {
                    if (!sym.definition)
                        return true; // AST_Destructuring
                    var def = sym.definition();
                    if (def.orig.length == 1 && def.orig[0] instanceof AST_SymbolDefun)
                        return false;
                    if (def.scope.get_defun_scope() !== scope)
                        return true;
                    return !def.references.every((ref) => {
                        var s = ref.scope.get_defun_scope();
                        // "block" scope within AST_Catch
                        if (s.TYPE == "Scope")
                            s = s.parent_scope;
                        return s === scope;
                    });
                }
                function side_effects_external(node, lhs) {
                    if (node instanceof AST_Assign)
                        return side_effects_external(node.left, true);
                    if (node instanceof AST_Unary)
                        return side_effects_external(node.expression, true);
                    if (node instanceof AST_VarDef)
                        return node.value && side_effects_external(node.value);
                    if (lhs) {
                        if (node instanceof AST_Dot)
                            return side_effects_external(node.expression, true);
                        if (node instanceof AST_Sub)
                            return side_effects_external(node.expression, true);
                        if (node instanceof AST_SymbolRef)
                            return node.definition().scope !== scope;
                    }
                    return false;
                }
            }
            function eliminate_spurious_blocks(statements) {
                var seen_dirs = [];
                for (var i = 0; i < statements.length;) {
                    var stat = statements[i];
                    if (stat instanceof AST_BlockStatement && stat.body.every(can_be_evicted_from_block)) {
                        CHANGED = true;
                        eliminate_spurious_blocks(stat.body);
                        statements.splice(i, 1, ...stat.body);
                        i += stat.body.length;
                    }
                    else if (stat instanceof AST_EmptyStatement) {
                        CHANGED = true;
                        statements.splice(i, 1);
                    }
                    else if (stat instanceof AST_Directive) {
                        if (seen_dirs.indexOf(stat.value) < 0) {
                            i++;
                            seen_dirs.push(stat.value);
                        }
                        else {
                            CHANGED = true;
                            statements.splice(i, 1);
                        }
                    }
                    else
                        i++;
                }
            }
            function handle_if_return(statements, compressor) {
                var self = compressor.self();
                var multiple_if_returns = has_multiple_if_returns(statements);
                var in_lambda = self instanceof AST_Lambda;
                for (var i = statements.length; --i >= 0;) {
                    var stat = statements[i];
                    var j = next_index(i);
                    var next = statements[j];
                    if (in_lambda && !next && stat instanceof AST_Return) {
                        if (!stat.value) {
                            CHANGED = true;
                            statements.splice(i, 1);
                            continue;
                        }
                        if (stat.value instanceof AST_UnaryPrefix && stat.value.operator == "void") {
                            CHANGED = true;
                            statements[i] = make_node(AST_SimpleStatement, stat, {
                                body: stat.value.expression
                            });
                            continue;
                        }
                    }
                    if (stat instanceof AST_If) {
                        var ab = aborts(stat.body);
                        if (can_merge_flow(ab)) {
                            if (ab.label) {
                                remove(ab.label.thedef.references, ab);
                            }
                            CHANGED = true;
                            stat = stat.clone();
                            stat.condition = stat.condition.negate(compressor);
                            var body = as_statement_array_with_return(stat.body, ab);
                            stat.body = make_node(AST_BlockStatement, stat, {
                                body: as_statement_array(stat.alternative).concat(extract_functions())
                            });
                            stat.alternative = make_node(AST_BlockStatement, stat, {
                                body: body
                            });
                            statements[i] = stat.transform(compressor);
                            continue;
                        }
                        var ab = aborts(stat.alternative);
                        if (can_merge_flow(ab)) {
                            if (ab.label) {
                                remove(ab.label.thedef.references, ab);
                            }
                            CHANGED = true;
                            stat = stat.clone();
                            stat.body = make_node(AST_BlockStatement, stat.body, {
                                body: as_statement_array(stat.body).concat(extract_functions())
                            });
                            var body = as_statement_array_with_return(stat.alternative, ab);
                            stat.alternative = make_node(AST_BlockStatement, stat.alternative, {
                                body: body
                            });
                            statements[i] = stat.transform(compressor);
                            continue;
                        }
                    }
                    if (stat instanceof AST_If && stat.body instanceof AST_Return) {
                        var value = stat.body.value;
                        //---
                        // pretty silly case, but:
                        // if (foo()) return; return; ==> foo(); return;
                        if (!value && !stat.alternative
                            && (in_lambda && !next || next instanceof AST_Return && !next.value)) {
                            CHANGED = true;
                            statements[i] = make_node(AST_SimpleStatement, stat.condition, {
                                body: stat.condition
                            });
                            continue;
                        }
                        //---
                        // if (foo()) return x; return y; ==> return foo() ? x : y;
                        if (value && !stat.alternative && next instanceof AST_Return && next.value) {
                            CHANGED = true;
                            stat = stat.clone();
                            stat.alternative = next;
                            statements[i] = stat.transform(compressor);
                            statements.splice(j, 1);
                            continue;
                        }
                        //---
                        // if (foo()) return x; [ return ; ] ==> return foo() ? x : undefined;
                        if (value && !stat.alternative
                            && (!next && in_lambda && multiple_if_returns
                                || next instanceof AST_Return)) {
                            CHANGED = true;
                            stat = stat.clone();
                            stat.alternative = next || make_node(AST_Return, stat, {
                                value: null
                            });
                            statements[i] = stat.transform(compressor);
                            if (next)
                                statements.splice(j, 1);
                            continue;
                        }
                        //---
                        // if (a) return b; if (c) return d; e; ==> return a ? b : c ? d : void e;
                        //
                        // if sequences is not enabled, this can lead to an endless loop (issue #866).
                        // however, with sequences on this helps producing slightly better output for
                        // the example code.
                        var prev = statements[prev_index(i)];
                        if (compressor.option("sequences") && in_lambda && !stat.alternative
                            && prev instanceof AST_If && prev.body instanceof AST_Return
                            && next_index(j) == statements.length && next instanceof AST_SimpleStatement) {
                            CHANGED = true;
                            stat = stat.clone();
                            stat.alternative = make_node(AST_BlockStatement, next, {
                                body: [
                                    next,
                                    make_node(AST_Return, next, {
                                        value: null
                                    })
                                ]
                            });
                            statements[i] = stat.transform(compressor);
                            statements.splice(j, 1);
                            continue;
                        }
                    }
                }
                function has_multiple_if_returns(statements) {
                    var n = 0;
                    for (var i = statements.length; --i >= 0;) {
                        var stat = statements[i];
                        if (stat instanceof AST_If && stat.body instanceof AST_Return) {
                            if (++n > 1)
                                return true;
                        }
                    }
                    return false;
                }
                function is_return_void(value) {
                    return !value || value instanceof AST_UnaryPrefix && value.operator == "void";
                }
                function can_merge_flow(ab) {
                    if (!ab)
                        return false;
                    for (var j = i + 1, len = statements.length; j < len; j++) {
                        var stat = statements[j];
                        if (stat instanceof AST_Const || stat instanceof AST_Let)
                            return false;
                    }
                    var lct = ab instanceof AST_LoopControl ? compressor.loopcontrol_target(ab) : null;
                    return ab instanceof AST_Return && in_lambda && is_return_void(ab.value)
                        || ab instanceof AST_Continue && self === loop_body(lct)
                        || ab instanceof AST_Break && lct instanceof AST_BlockStatement && self === lct;
                }
                function extract_functions() {
                    var tail = statements.slice(i + 1);
                    statements.length = i + 1;
                    return tail.filter(function (stat) {
                        if (stat instanceof AST_Defun) {
                            statements.push(stat);
                            return false;
                        }
                        return true;
                    });
                }
                function as_statement_array_with_return(node, ab) {
                    var body = as_statement_array(node).slice(0, -1);
                    if (ab.value) {
                        body.push(make_node(AST_SimpleStatement, ab.value, {
                            body: ab.value.expression
                        }));
                    }
                    return body;
                }
                function next_index(i) {
                    for (var j = i + 1, len = statements.length; j < len; j++) {
                        var stat = statements[j];
                        if (!(stat instanceof AST_Var && declarations_only(stat))) {
                            break;
                        }
                    }
                    return j;
                }
                function prev_index(i) {
                    for (var j = i; --j >= 0;) {
                        var stat = statements[j];
                        if (!(stat instanceof AST_Var && declarations_only(stat))) {
                            break;
                        }
                    }
                    return j;
                }
            }
            function eliminate_dead_code(statements, compressor) {
                var has_quit;
                var self = compressor.self();
                for (var i = 0, n = 0, len = statements.length; i < len; i++) {
                    var stat = statements[i];
                    if (stat instanceof AST_LoopControl) {
                        var lct = compressor.loopcontrol_target(stat);
                        if (stat instanceof AST_Break
                            && !(lct instanceof AST_IterationStatement)
                            && loop_body(lct) === self
                            || stat instanceof AST_Continue
                                && loop_body(lct) === self) {
                            if (stat.label) {
                                remove(stat.label.thedef.references, stat);
                            }
                        }
                        else {
                            statements[n++] = stat;
                        }
                    }
                    else {
                        statements[n++] = stat;
                    }
                    if (aborts(stat)) {
                        has_quit = statements.slice(i + 1);
                        break;
                    }
                }
                statements.length = n;
                CHANGED = n != len;
                if (has_quit)
                    has_quit.forEach(function (stat) {
                        trim_unreachable_code(compressor, stat, statements);
                    });
            }
            function declarations_only(node) {
                return node.definitions.every((var_def) => !var_def.value);
            }
            function sequencesize(statements, compressor) {
                if (statements.length < 2)
                    return;
                var seq = [], n = 0;
                function push_seq() {
                    if (!seq.length)
                        return;
                    var body = make_sequence(seq[0], seq);
                    statements[n++] = make_node(AST_SimpleStatement, body, { body: body });
                    seq = [];
                }
                for (var i = 0, len = statements.length; i < len; i++) {
                    var stat = statements[i];
                    if (stat instanceof AST_SimpleStatement) {
                        if (seq.length >= compressor.sequences_limit)
                            push_seq();
                        var body = stat.body;
                        if (seq.length > 0)
                            body = body.drop_side_effect_free(compressor);
                        if (body)
                            merge_sequence(seq, body);
                    }
                    else if (stat instanceof AST_Definitions && declarations_only(stat)
                        || stat instanceof AST_Defun) {
                        statements[n++] = stat;
                    }
                    else {
                        push_seq();
                        statements[n++] = stat;
                    }
                }
                push_seq();
                statements.length = n;
                if (n != len)
                    CHANGED = true;
            }
            function to_simple_statement(block, decls) {
                if (!(block instanceof AST_BlockStatement))
                    return block;
                var stat = null;
                for (var i = 0, len = block.body.length; i < len; i++) {
                    var line = block.body[i];
                    if (line instanceof AST_Var && declarations_only(line)) {
                        decls.push(line);
                    }
                    else if (stat) {
                        return false;
                    }
                    else {
                        stat = line;
                    }
                }
                return stat;
            }
            function sequencesize_2(statements, compressor) {
                function cons_seq(right) {
                    n--;
                    CHANGED = true;
                    var left = prev.body;
                    return make_sequence(left, [left, right]).transform(compressor);
                }
                var n = 0, prev;
                for (var i = 0; i < statements.length; i++) {
                    var stat = statements[i];
                    if (prev) {
                        if (stat instanceof AST_Exit) {
                            stat.value = cons_seq(stat.value || make_node(AST_Undefined, stat).transform(compressor));
                        }
                        else if (stat instanceof AST_For) {
                            if (!(stat.init instanceof AST_Definitions)) {
                                const abort = walk(prev.body, node => {
                                    if (node instanceof AST_Scope)
                                        return true;
                                    if (node instanceof AST_Binary
                                        && node.operator === "in") {
                                        return walk_abort;
                                    }
                                });
                                if (!abort) {
                                    if (stat.init)
                                        stat.init = cons_seq(stat.init);
                                    else {
                                        stat.init = prev.body;
                                        n--;
                                        CHANGED = true;
                                    }
                                }
                            }
                        }
                        else if (stat instanceof AST_ForIn) {
                            if (!(stat.init instanceof AST_Const) && !(stat.init instanceof AST_Let)) {
                                stat.object = cons_seq(stat.object);
                            }
                        }
                        else if (stat instanceof AST_If) {
                            stat.condition = cons_seq(stat.condition);
                        }
                        else if (stat instanceof AST_Switch) {
                            stat.expression = cons_seq(stat.expression);
                        }
                        else if (stat instanceof AST_With) {
                            stat.expression = cons_seq(stat.expression);
                        }
                    }
                    if (compressor.option("conditionals") && stat instanceof AST_If) {
                        var decls = [];
                        var body = to_simple_statement(stat.body, decls);
                        var alt = to_simple_statement(stat.alternative, decls);
                        if (body !== false && alt !== false && decls.length > 0) {
                            var len = decls.length;
                            decls.push(make_node(AST_If, stat, {
                                condition: stat.condition,
                                body: body || make_node(AST_EmptyStatement, stat.body),
                                alternative: alt
                            }));
                            decls.unshift(n, 1);
                            [].splice.apply(statements, decls);
                            i += len;
                            n += len + 1;
                            prev = null;
                            CHANGED = true;
                            continue;
                        }
                    }
                    statements[n++] = stat;
                    prev = stat instanceof AST_SimpleStatement ? stat : null;
                }
                statements.length = n;
            }
            function join_object_assignments(defn, body) {
                if (!(defn instanceof AST_Definitions))
                    return;
                var def = defn.definitions[defn.definitions.length - 1];
                if (!(def.value instanceof AST_Object))
                    return;
                var exprs;
                if (body instanceof AST_Assign && !body.logical) {
                    exprs = [body];
                }
                else if (body instanceof AST_Sequence) {
                    exprs = body.expressions.slice();
                }
                if (!exprs)
                    return;
                var trimmed = false;
                do {
                    var node = exprs[0];
                    if (!(node instanceof AST_Assign))
                        break;
                    if (node.operator != "=")
                        break;
                    if (!(node.left instanceof AST_PropAccess))
                        break;
                    var sym = node.left.expression;
                    if (!(sym instanceof AST_SymbolRef))
                        break;
                    if (def.name.name != sym.name)
                        break;
                    if (!node.right.is_constant_expression(scope))
                        break;
                    var prop = node.left.property;
                    if (prop instanceof AST_Node) {
                        prop = prop.evaluate(compressor);
                    }
                    if (prop instanceof AST_Node)
                        break;
                    prop = "" + prop;
                    var diff = compressor.option("ecma") < 2015
                        && compressor.has_directive("use strict") ? function (node) {
                        return node.key != prop && (node.key && node.key.name != prop);
                    } : function (node) {
                        return node.key && node.key.name != prop;
                    };
                    if (!def.value.properties.every(diff))
                        break;
                    var p = def.value.properties.filter(function (p) { return p.key === prop; })[0];
                    if (!p) {
                        def.value.properties.push(make_node(AST_ObjectKeyVal, node, {
                            key: prop,
                            value: node.right
                        }));
                    }
                    else {
                        p.value = new AST_Sequence({
                            start: p.start,
                            expressions: [p.value.clone(), node.right.clone()],
                            end: p.end
                        });
                    }
                    exprs.shift();
                    trimmed = true;
                } while (exprs.length);
                return trimmed && exprs;
            }
            function join_consecutive_vars(statements) {
                var defs;
                for (var i = 0, j = -1, len = statements.length; i < len; i++) {
                    var stat = statements[i];
                    var prev = statements[j];
                    if (stat instanceof AST_Definitions) {
                        if (prev && prev.TYPE == stat.TYPE) {
                            prev.definitions = prev.definitions.concat(stat.definitions);
                            CHANGED = true;
                        }
                        else if (defs && defs.TYPE == stat.TYPE && declarations_only(stat)) {
                            defs.definitions = defs.definitions.concat(stat.definitions);
                            CHANGED = true;
                        }
                        else {
                            statements[++j] = stat;
                            defs = stat;
                        }
                    }
                    else if (stat instanceof AST_Exit) {
                        stat.value = extract_object_assignments(stat.value);
                    }
                    else if (stat instanceof AST_For) {
                        var exprs = join_object_assignments(prev, stat.init);
                        if (exprs) {
                            CHANGED = true;
                            stat.init = exprs.length ? make_sequence(stat.init, exprs) : null;
                            statements[++j] = stat;
                        }
                        else if (prev instanceof AST_Var && (!stat.init || stat.init.TYPE == prev.TYPE)) {
                            if (stat.init) {
                                prev.definitions = prev.definitions.concat(stat.init.definitions);
                            }
                            stat.init = prev;
                            statements[j] = stat;
                            CHANGED = true;
                        }
                        else if (defs && stat.init && defs.TYPE == stat.init.TYPE && declarations_only(stat.init)) {
                            defs.definitions = defs.definitions.concat(stat.init.definitions);
                            stat.init = null;
                            statements[++j] = stat;
                            CHANGED = true;
                        }
                        else {
                            statements[++j] = stat;
                        }
                    }
                    else if (stat instanceof AST_ForIn) {
                        stat.object = extract_object_assignments(stat.object);
                    }
                    else if (stat instanceof AST_If) {
                        stat.condition = extract_object_assignments(stat.condition);
                    }
                    else if (stat instanceof AST_SimpleStatement) {
                        var exprs = join_object_assignments(prev, stat.body);
                        if (exprs) {
                            CHANGED = true;
                            if (!exprs.length)
                                continue;
                            stat.body = make_sequence(stat.body, exprs);
                        }
                        statements[++j] = stat;
                    }
                    else if (stat instanceof AST_Switch) {
                        stat.expression = extract_object_assignments(stat.expression);
                    }
                    else if (stat instanceof AST_With) {
                        stat.expression = extract_object_assignments(stat.expression);
                    }
                    else {
                        statements[++j] = stat;
                    }
                }
                statements.length = j + 1;
                function extract_object_assignments(value) {
                    statements[++j] = stat;
                    var exprs = join_object_assignments(prev, value);
                    if (exprs) {
                        CHANGED = true;
                        if (exprs.length) {
                            return make_sequence(value, exprs);
                        }
                        else if (value instanceof AST_Sequence) {
                            return value.tail_node().left;
                        }
                        else {
                            return value.left;
                        }
                    }
                    return value;
                }
            }
        }
        function trim_unreachable_code(compressor, stat, target) {
            walk(stat, node => {
                if (node instanceof AST_Var) {
                    node.remove_initializers();
                    target.push(node);
                    return true;
                }
                if (node instanceof AST_Defun
                    && (node === stat || !compressor.has_directive("use strict"))) {
                    target.push(node === stat ? node : make_node(AST_Var, node, {
                        definitions: [
                            make_node(AST_VarDef, node, {
                                name: make_node(AST_SymbolVar, node.name, node.name),
                                value: null
                            })
                        ]
                    }));
                    return true;
                }
                if (node instanceof AST_Export || node instanceof AST_Import) {
                    target.push(node);
                    return true;
                }
                if (node instanceof AST_Scope) {
                    return true;
                }
            });
        }
        function get_value(key) {
            if (key instanceof AST_Constant) {
                return key.getValue();
            }
            if (key instanceof AST_UnaryPrefix
                && key.operator == "void"
                && key.expression instanceof AST_Constant) {
                return;
            }
            return key;
        }
        function is_undefined(node, compressor) {
            return (has_flag(node, UNDEFINED)
                || node instanceof AST_Undefined
                || node instanceof AST_UnaryPrefix
                    && node.operator == "void"
                    && !node.expression.has_side_effects(compressor));
        }
        // may_throw_on_access()
        // returns true if this node may be null, undefined or contain `AST_Accessor`
        (function (def_may_throw_on_access) {
            AST_Node.DEFMETHOD("may_throw_on_access", function (compressor) {
                return !compressor.option("pure_getters")
                    || this._dot_throw(compressor);
            });
            function is_strict(compressor) {
                return /strict/.test(compressor.option("pure_getters"));
            }
            def_may_throw_on_access(AST_Node, is_strict);
            def_may_throw_on_access(AST_Null, return_true);
            def_may_throw_on_access(AST_Undefined, return_true);
            def_may_throw_on_access(AST_Constant, return_false);
            def_may_throw_on_access(AST_Array, return_false);
            def_may_throw_on_access(AST_Object, function (compressor) {
                if (!is_strict(compressor))
                    return false;
                for (var i = this.properties.length; --i >= 0;)
                    if (this.properties[i]._dot_throw(compressor))
                        return true;
                return false;
            });
            // Do not be as strict with classes as we are with objects.
            // Hopefully the community is not going to abuse static getters and setters.
            // https://github.com/terser/terser/issues/724#issuecomment-643655656
            def_may_throw_on_access(AST_Class, return_false);
            def_may_throw_on_access(AST_ObjectProperty, return_false);
            def_may_throw_on_access(AST_ObjectGetter, return_true);
            def_may_throw_on_access(AST_Expansion, function (compressor) {
                return this.expression._dot_throw(compressor);
            });
            def_may_throw_on_access(AST_Function, return_false);
            def_may_throw_on_access(AST_Arrow, return_false);
            def_may_throw_on_access(AST_UnaryPostfix, return_false);
            def_may_throw_on_access(AST_UnaryPrefix, function () {
                return this.operator == "void";
            });
            def_may_throw_on_access(AST_Binary, function (compressor) {
                return (this.operator == "&&" || this.operator == "||" || this.operator == "??")
                    && (this.left._dot_throw(compressor) || this.right._dot_throw(compressor));
            });
            def_may_throw_on_access(AST_Assign, function (compressor) {
                if (this.logical)
                    return true;
                return this.operator == "="
                    && this.right._dot_throw(compressor);
            });
            def_may_throw_on_access(AST_Conditional, function (compressor) {
                return this.consequent._dot_throw(compressor)
                    || this.alternative._dot_throw(compressor);
            });
            def_may_throw_on_access(AST_Dot, function (compressor) {
                if (!is_strict(compressor))
                    return false;
                if (this.property == "prototype") {
                    return !(this.expression instanceof AST_Function
                        || this.expression instanceof AST_Class);
                }
                return true;
            });
            def_may_throw_on_access(AST_Chain, function (compressor) {
                return this.expression._dot_throw(compressor);
            });
            def_may_throw_on_access(AST_Sequence, function (compressor) {
                return this.tail_node()._dot_throw(compressor);
            });
            def_may_throw_on_access(AST_SymbolRef, function (compressor) {
                if (this.name === "arguments")
                    return false;
                if (has_flag(this, UNDEFINED))
                    return true;
                if (!is_strict(compressor))
                    return false;
                if (is_undeclared_ref(this) && this.is_declared(compressor))
                    return false;
                if (this.is_immutable())
                    return false;
                var fixed = this.fixed_value();
                return !fixed || fixed._dot_throw(compressor);
            });
        })(function (node, func) {
            node.DEFMETHOD("_dot_throw", func);
        });
        /* -----[ boolean/negation helpers ]----- */
        // methods to determine whether an expression has a boolean result type
        (function (def_is_boolean) {
            const unary_bool = makePredicate("! delete");
            const binary_bool = makePredicate("in instanceof == != === !== < <= >= >");
            def_is_boolean(AST_Node, return_false);
            def_is_boolean(AST_UnaryPrefix, function () {
                return unary_bool.has(this.operator);
            });
            def_is_boolean(AST_Binary, function () {
                return binary_bool.has(this.operator)
                    || lazy_op.has(this.operator)
                        && this.left.is_boolean()
                        && this.right.is_boolean();
            });
            def_is_boolean(AST_Conditional, function () {
                return this.consequent.is_boolean() && this.alternative.is_boolean();
            });
            def_is_boolean(AST_Assign, function () {
                return this.operator == "=" && this.right.is_boolean();
            });
            def_is_boolean(AST_Sequence, function () {
                return this.tail_node().is_boolean();
            });
            def_is_boolean(AST_True, return_true);
            def_is_boolean(AST_False, return_true);
        })(function (node, func) {
            node.DEFMETHOD("is_boolean", func);
        });
        // methods to determine if an expression has a numeric result type
        (function (def_is_number) {
            def_is_number(AST_Node, return_false);
            def_is_number(AST_Number, return_true);
            var unary = makePredicate("+ - ~ ++ --");
            def_is_number(AST_Unary, function () {
                return unary.has(this.operator);
            });
            var binary = makePredicate("- * / % & | ^ << >> >>>");
            def_is_number(AST_Binary, function (compressor) {
                return binary.has(this.operator) || this.operator == "+"
                    && this.left.is_number(compressor)
                    && this.right.is_number(compressor);
            });
            def_is_number(AST_Assign, function (compressor) {
                return binary.has(this.operator.slice(0, -1))
                    || this.operator == "=" && this.right.is_number(compressor);
            });
            def_is_number(AST_Sequence, function (compressor) {
                return this.tail_node().is_number(compressor);
            });
            def_is_number(AST_Conditional, function (compressor) {
                return this.consequent.is_number(compressor) && this.alternative.is_number(compressor);
            });
        })(function (node, func) {
            node.DEFMETHOD("is_number", func);
        });
        // methods to determine if an expression has a string result type
        (function (def_is_string) {
            def_is_string(AST_Node, return_false);
            def_is_string(AST_String, return_true);
            def_is_string(AST_TemplateString, return_true);
            def_is_string(AST_UnaryPrefix, function () {
                return this.operator == "typeof";
            });
            def_is_string(AST_Binary, function (compressor) {
                return this.operator == "+" &&
                    (this.left.is_string(compressor) || this.right.is_string(compressor));
            });
            def_is_string(AST_Assign, function (compressor) {
                return (this.operator == "=" || this.operator == "+=") && this.right.is_string(compressor);
            });
            def_is_string(AST_Sequence, function (compressor) {
                return this.tail_node().is_string(compressor);
            });
            def_is_string(AST_Conditional, function (compressor) {
                return this.consequent.is_string(compressor) && this.alternative.is_string(compressor);
            });
        })(function (node, func) {
            node.DEFMETHOD("is_string", func);
        });
        var lazy_op = makePredicate("&& || ??");
        var unary_side_effects = makePredicate("delete ++ --");
        function is_lhs(node, parent) {
            if (parent instanceof AST_Unary && unary_side_effects.has(parent.operator))
                return parent.expression;
            if (parent instanceof AST_Assign && parent.left === node)
                return node;
        }
        (function (def_find_defs) {
            function to_node(value, orig) {
                if (value instanceof AST_Node)
                    return make_node(value.CTOR, orig, value);
                if (Array.isArray(value))
                    return make_node(AST_Array, orig, {
                        elements: value.map(function (value) {
                            return to_node(value, orig);
                        })
                    });
                if (value && typeof value == "object") {
                    var props = [];
                    for (var key in value)
                        if (HOP(value, key)) {
                            props.push(make_node(AST_ObjectKeyVal, orig, {
                                key: key,
                                value: to_node(value[key], orig)
                            }));
                        }
                    return make_node(AST_Object, orig, {
                        properties: props
                    });
                }
                return make_node_from_constant(value, orig);
            }
            AST_Toplevel.DEFMETHOD("resolve_defines", function (compressor) {
                if (!compressor.option("global_defs"))
                    return this;
                this.figure_out_scope({ ie8: compressor.option("ie8") });
                return this.transform(new TreeTransformer(function (node) {
                    var def = node._find_defs(compressor, "");
                    if (!def)
                        return;
                    var level = 0, child = node, parent;
                    while (parent = this.parent(level++)) {
                        if (!(parent instanceof AST_PropAccess))
                            break;
                        if (parent.expression !== child)
                            break;
                        child = parent;
                    }
                    if (is_lhs(child, parent)) {
                        return;
                    }
                    return def;
                }));
            });
            def_find_defs(AST_Node, noop);
            def_find_defs(AST_Chain, function (compressor, suffix) {
                return this.expression._find_defs(compressor, suffix);
            });
            def_find_defs(AST_Dot, function (compressor, suffix) {
                return this.expression._find_defs(compressor, "." + this.property + suffix);
            });
            def_find_defs(AST_SymbolDeclaration, function () {
                if (!this.global())
                    return;
            });
            def_find_defs(AST_SymbolRef, function (compressor, suffix) {
                if (!this.global())
                    return;
                var defines = compressor.option("global_defs");
                var name = this.name + suffix;
                if (HOP(defines, name))
                    return to_node(defines[name], this);
            });
        })(function (node, func) {
            node.DEFMETHOD("_find_defs", func);
        });
        function best_of_expression(ast1, ast2) {
            return ast1.size() > ast2.size() ? ast2 : ast1;
        }
        function best_of_statement(ast1, ast2) {
            return best_of_expression(make_node(AST_SimpleStatement, ast1, {
                body: ast1
            }), make_node(AST_SimpleStatement, ast2, {
                body: ast2
            })).body;
        }
        function best_of(compressor, ast1, ast2) {
            return (first_in_statement(compressor) ? best_of_statement : best_of_expression)(ast1, ast2);
        }
        function convert_to_predicate(obj) {
            const out = new Map();
            for (var key of Object.keys(obj)) {
                out.set(key, makePredicate(obj[key]));
            }
            return out;
        }
        var object_fns = [
            "constructor",
            "toString",
            "valueOf",
        ];
        var native_fns = convert_to_predicate({
            Array: [
                "indexOf",
                "join",
                "lastIndexOf",
                "slice",
            ].concat(object_fns),
            Boolean: object_fns,
            Function: object_fns,
            Number: [
                "toExponential",
                "toFixed",
                "toPrecision",
            ].concat(object_fns),
            Object: object_fns,
            RegExp: [
                "test",
            ].concat(object_fns),
            String: [
                "charAt",
                "charCodeAt",
                "concat",
                "indexOf",
                "italics",
                "lastIndexOf",
                "match",
                "replace",
                "search",
                "slice",
                "split",
                "substr",
                "substring",
                "toLowerCase",
                "toUpperCase",
                "trim",
            ].concat(object_fns),
        });
        var static_fns = convert_to_predicate({
            Array: [
                "isArray",
            ],
            Math: [
                "abs",
                "acos",
                "asin",
                "atan",
                "ceil",
                "cos",
                "exp",
                "floor",
                "log",
                "round",
                "sin",
                "sqrt",
                "tan",
                "atan2",
                "pow",
                "max",
                "min",
            ],
            Number: [
                "isFinite",
                "isNaN",
            ],
            Object: [
                "create",
                "getOwnPropertyDescriptor",
                "getOwnPropertyNames",
                "getPrototypeOf",
                "isExtensible",
                "isFrozen",
                "isSealed",
                "keys",
            ],
            String: [
                "fromCharCode",
            ],
        });
        // methods to evaluate a constant expression
        (function (def_eval) {
            // If the node has been successfully reduced to a constant,
            // then its value is returned; otherwise the element itself
            // is returned.
            // They can be distinguished as constant value is never a
            // descendant of AST_Node.
            AST_Node.DEFMETHOD("evaluate", function (compressor) {
                if (!compressor.option("evaluate"))
                    return this;
                var val = this._eval(compressor, 1);
                if (!val || val instanceof RegExp)
                    return val;
                if (typeof val == "function" || typeof val == "object")
                    return this;
                return val;
            });
            var unaryPrefix = makePredicate("! ~ - + void");
            AST_Node.DEFMETHOD("is_constant", function () {
                // Accomodate when compress option evaluate=false
                // as well as the common constant expressions !0 and -1
                if (this instanceof AST_Constant) {
                    return !(this instanceof AST_RegExp);
                }
                else {
                    return this instanceof AST_UnaryPrefix
                        && this.expression instanceof AST_Constant
                        && unaryPrefix.has(this.operator);
                }
            });
            def_eval(AST_Statement, function () {
                throw new Error(string_template("Cannot evaluate a statement [{file}:{line},{col}]", this.start));
            });
            def_eval(AST_Lambda, return_this);
            def_eval(AST_Class, return_this);
            def_eval(AST_Node, return_this);
            def_eval(AST_Constant, function () {
                return this.getValue();
            });
            def_eval(AST_BigInt, return_this);
            def_eval(AST_RegExp, function (compressor) {
                let evaluated = compressor.evaluated_regexps.get(this);
                if (evaluated === undefined) {
                    try {
                        evaluated = (0, eval)(this.print_to_string());
                    }
                    catch (e) {
                        evaluated = null;
                    }
                    compressor.evaluated_regexps.set(this, evaluated);
                }
                return evaluated || this;
            });
            def_eval(AST_TemplateString, function () {
                if (this.segments.length !== 1)
                    return this;
                return this.segments[0].value;
            });
            def_eval(AST_Function, function (compressor) {
                if (compressor.option("unsafe")) {
                    var fn = function () { };
                    fn.node = this;
                    fn.toString = () => this.print_to_string();
                    return fn;
                }
                return this;
            });
            def_eval(AST_Array, function (compressor, depth) {
                if (compressor.option("unsafe")) {
                    var elements = [];
                    for (var i = 0, len = this.elements.length; i < len; i++) {
                        var element = this.elements[i];
                        var value = element._eval(compressor, depth);
                        if (element === value)
                            return this;
                        elements.push(value);
                    }
                    return elements;
                }
                return this;
            });
            def_eval(AST_Object, function (compressor, depth) {
                if (compressor.option("unsafe")) {
                    var val = {};
                    for (var i = 0, len = this.properties.length; i < len; i++) {
                        var prop = this.properties[i];
                        if (prop instanceof AST_Expansion)
                            return this;
                        var key = prop.key;
                        if (key instanceof AST_Symbol) {
                            key = key.name;
                        }
                        else if (key instanceof AST_Node) {
                            key = key._eval(compressor, depth);
                            if (key === prop.key)
                                return this;
                        }
                        if (typeof Object.prototype[key] === "function") {
                            return this;
                        }
                        if (prop.value instanceof AST_Function)
                            continue;
                        val[key] = prop.value._eval(compressor, depth);
                        if (val[key] === prop.value)
                            return this;
                    }
                    return val;
                }
                return this;
            });
            var non_converting_unary = makePredicate("! typeof void");
            def_eval(AST_UnaryPrefix, function (compressor, depth) {
                var e = this.expression;
                // Function would be evaluated to an array and so typeof would
                // incorrectly return 'object'. Hence making is a special case.
                if (compressor.option("typeofs")
                    && this.operator == "typeof"
                    && (e instanceof AST_Lambda
                        || e instanceof AST_SymbolRef
                            && e.fixed_value() instanceof AST_Lambda)) {
                    return typeof function () { };
                }
                if (!non_converting_unary.has(this.operator))
                    depth++;
                e = e._eval(compressor, depth);
                if (e === this.expression)
                    return this;
                switch (this.operator) {
                    case "!": return !e;
                    case "typeof":
                        // typeof <RegExp> returns "object" or "function" on different platforms
                        // so cannot evaluate reliably
                        if (e instanceof RegExp)
                            return this;
                        return typeof e;
                    case "void": return void e;
                    case "~": return ~e;
                    case "-": return -e;
                    case "+": return +e;
                }
                return this;
            });
            var non_converting_binary = makePredicate("&& || ?? === !==");
            const identity_comparison = makePredicate("== != === !==");
            const has_identity = value => typeof value === "object"
                || typeof value === "function"
                || typeof value === "symbol";
            def_eval(AST_Binary, function (compressor, depth) {
                if (!non_converting_binary.has(this.operator))
                    depth++;
                var left = this.left._eval(compressor, depth);
                if (left === this.left)
                    return this;
                var right = this.right._eval(compressor, depth);
                if (right === this.right)
                    return this;
                var result;
                if (left != null
                    && right != null
                    && identity_comparison.has(this.operator)
                    && has_identity(left)
                    && has_identity(right)
                    && typeof left === typeof right) {
                    // Do not compare by reference
                    return this;
                }
                switch (this.operator) {
                    case "&&":
                        result = left && right;
                        break;
                    case "||":
                        result = left || right;
                        break;
                    case "??":
                        result = left != null ? left : right;
                        break;
                    case "|":
                        result = left | right;
                        break;
                    case "&":
                        result = left & right;
                        break;
                    case "^":
                        result = left ^ right;
                        break;
                    case "+":
                        result = left + right;
                        break;
                    case "*":
                        result = left * right;
                        break;
                    case "**":
                        result = Math.pow(left, right);
                        break;
                    case "/":
                        result = left / right;
                        break;
                    case "%":
                        result = left % right;
                        break;
                    case "-":
                        result = left - right;
                        break;
                    case "<<":
                        result = left << right;
                        break;
                    case ">>":
                        result = left >> right;
                        break;
                    case ">>>":
                        result = left >>> right;
                        break;
                    case "==":
                        result = left == right;
                        break;
                    case "===":
                        result = left === right;
                        break;
                    case "!=":
                        result = left != right;
                        break;
                    case "!==":
                        result = left !== right;
                        break;
                    case "<":
                        result = left < right;
                        break;
                    case "<=":
                        result = left <= right;
                        break;
                    case ">":
                        result = left > right;
                        break;
                    case ">=":
                        result = left >= right;
                        break;
                    default:
                        return this;
                }
                if (isNaN(result) && compressor.find_parent(AST_With)) {
                    // leave original expression as is
                    return this;
                }
                return result;
            });
            def_eval(AST_Conditional, function (compressor, depth) {
                var condition = this.condition._eval(compressor, depth);
                if (condition === this.condition)
                    return this;
                var node = condition ? this.consequent : this.alternative;
                var value = node._eval(compressor, depth);
                return value === node ? this : value;
            });
            // Set of AST_SymbolRef which are currently being evaluated.
            // Avoids infinite recursion of ._eval()
            const reentrant_ref_eval = new Set();
            def_eval(AST_SymbolRef, function (compressor, depth) {
                if (reentrant_ref_eval.has(this))
                    return this;
                var fixed = this.fixed_value();
                if (!fixed)
                    return this;
                reentrant_ref_eval.add(this);
                const value = fixed._eval(compressor, depth);
                reentrant_ref_eval.delete(this);
                if (value === fixed)
                    return this;
                if (value && typeof value == "object") {
                    var escaped = this.definition().escaped;
                    if (escaped && depth > escaped)
                        return this;
                }
                return value;
            });
            var global_objs = { Array, Math, Number, Object, String };
            var static_values = convert_to_predicate({
                Math: [
                    "E",
                    "LN10",
                    "LN2",
                    "LOG2E",
                    "LOG10E",
                    "PI",
                    "SQRT1_2",
                    "SQRT2",
                ],
                Number: [
                    "MAX_VALUE",
                    "MIN_VALUE",
                    "NaN",
                    "NEGATIVE_INFINITY",
                    "POSITIVE_INFINITY",
                ],
            });
            const regexp_flags = new Set([
                "dotAll",
                "global",
                "ignoreCase",
                "multiline",
                "sticky",
                "unicode",
            ]);
            def_eval(AST_PropAccess, function (compressor, depth) {
                if (this.optional) {
                    const obj = this.expression._eval(compressor, depth);
                    if (obj == null)
                        return undefined;
                }
                if (compressor.option("unsafe")) {
                    var key = this.property;
                    if (key instanceof AST_Node) {
                        key = key._eval(compressor, depth);
                        if (key === this.property)
                            return this;
                    }
                    var exp = this.expression;
                    var val;
                    if (is_undeclared_ref(exp)) {
                        var aa;
                        var first_arg = exp.name === "hasOwnProperty"
                            && key === "call"
                            && (aa = compressor.parent() && compressor.parent().args)
                            && (aa && aa[0]
                                && aa[0].evaluate(compressor));
                        first_arg = first_arg instanceof AST_Dot ? first_arg.expression : first_arg;
                        if (first_arg == null || first_arg.thedef && first_arg.thedef.undeclared) {
                            return this.clone();
                        }
                        var static_value = static_values.get(exp.name);
                        if (!static_value || !static_value.has(key))
                            return this;
                        val = global_objs[exp.name];
                    }
                    else {
                        val = exp._eval(compressor, depth + 1);
                        if (val instanceof RegExp) {
                            if (key == "source") {
                                return regexp_source_fix(val.source);
                            }
                            else if (key == "flags" || regexp_flags.has(key)) {
                                return val[key];
                            }
                        }
                        if (!val || val === exp || !HOP(val, key))
                            return this;
                        if (typeof val == "function")
                            switch (key) {
                                case "name":
                                    return val.node.name ? val.node.name.name : "";
                                case "length":
                                    return val.node.length_property();
                                default:
                                    return this;
                            }
                    }
                    return val[key];
                }
                return this;
            });
            def_eval(AST_Chain, function (compressor, depth) {
                const evaluated = this.expression._eval(compressor, depth);
                return evaluated === this.expression ? this : evaluated;
            });
            def_eval(AST_Call, function (compressor, depth) {
                var exp = this.expression;
                if (this.optional) {
                    const callee = this.expression._eval(compressor, depth);
                    if (callee == null)
                        return undefined;
                }
                if (compressor.option("unsafe") && exp instanceof AST_PropAccess) {
                    var key = exp.property;
                    if (key instanceof AST_Node) {
                        key = key._eval(compressor, depth);
                        if (key === exp.property)
                            return this;
                    }
                    var val;
                    var e = exp.expression;
                    if (is_undeclared_ref(e)) {
                        var first_arg = e.name === "hasOwnProperty" &&
                            key === "call" &&
                            (this.args[0] && this.args[0].evaluate(compressor));
                        first_arg = first_arg instanceof AST_Dot ? first_arg.expression : first_arg;
                        if ((first_arg == null || first_arg.thedef && first_arg.thedef.undeclared)) {
                            return this.clone();
                        }
                        var static_fn = static_fns.get(e.name);
                        if (!static_fn || !static_fn.has(key))
                            return this;
                        val = global_objs[e.name];
                    }
                    else {
                        val = e._eval(compressor, depth + 1);
                        if (val === e || !val)
                            return this;
                        var native_fn = native_fns.get(val.constructor.name);
                        if (!native_fn || !native_fn.has(key))
                            return this;
                    }
                    var args = [];
                    for (var i = 0, len = this.args.length; i < len; i++) {
                        var arg = this.args[i];
                        var value = arg._eval(compressor, depth);
                        if (arg === value)
                            return this;
                        if (arg instanceof AST_Lambda)
                            return this;
                        args.push(value);
                    }
                    try {
                        return val[key].apply(val, args);
                    }
                    catch (ex) {
                        // We don't really care
                    }
                }
                return this;
            });
            def_eval(AST_New, return_this);
        })(function (node, func) {
            node.DEFMETHOD("_eval", func);
        });
        // method to negate an expression
        (function (def_negate) {
            function basic_negation(exp) {
                return make_node(AST_UnaryPrefix, exp, {
                    operator: "!",
                    expression: exp
                });
            }
            function best(orig, alt, first_in_statement) {
                var negated = basic_negation(orig);
                if (first_in_statement) {
                    var stat = make_node(AST_SimpleStatement, alt, {
                        body: alt
                    });
                    return best_of_expression(negated, stat) === stat ? alt : negated;
                }
                return best_of_expression(negated, alt);
            }
            def_negate(AST_Node, function () {
                return basic_negation(this);
            });
            def_negate(AST_Statement, function () {
                throw new Error("Cannot negate a statement");
            });
            def_negate(AST_Function, function () {
                return basic_negation(this);
            });
            def_negate(AST_Arrow, function () {
                return basic_negation(this);
            });
            def_negate(AST_UnaryPrefix, function () {
                if (this.operator == "!")
                    return this.expression;
                return basic_negation(this);
            });
            def_negate(AST_Sequence, function (compressor) {
                var expressions = this.expressions.slice();
                expressions.push(expressions.pop().negate(compressor));
                return make_sequence(this, expressions);
            });
            def_negate(AST_Conditional, function (compressor, first_in_statement) {
                var self = this.clone();
                self.consequent = self.consequent.negate(compressor);
                self.alternative = self.alternative.negate(compressor);
                return best(this, self, first_in_statement);
            });
            def_negate(AST_Binary, function (compressor, first_in_statement) {
                var self = this.clone(), op = this.operator;
                if (compressor.option("unsafe_comps")) {
                    switch (op) {
                        case "<=":
                            self.operator = ">";
                            return self;
                        case "<":
                            self.operator = ">=";
                            return self;
                        case ">=":
                            self.operator = "<";
                            return self;
                        case ">":
                            self.operator = "<=";
                            return self;
                    }
                }
                switch (op) {
                    case "==":
                        self.operator = "!=";
                        return self;
                    case "!=":
                        self.operator = "==";
                        return self;
                    case "===":
                        self.operator = "!==";
                        return self;
                    case "!==":
                        self.operator = "===";
                        return self;
                    case "&&":
                        self.operator = "||";
                        self.left = self.left.negate(compressor, first_in_statement);
                        self.right = self.right.negate(compressor);
                        return best(this, self, first_in_statement);
                    case "||":
                        self.operator = "&&";
                        self.left = self.left.negate(compressor, first_in_statement);
                        self.right = self.right.negate(compressor);
                        return best(this, self, first_in_statement);
                    case "??":
                        self.right = self.right.negate(compressor);
                        return best(this, self, first_in_statement);
                }
                return basic_negation(this);
            });
        })(function (node, func) {
            node.DEFMETHOD("negate", function (compressor, first_in_statement) {
                return func.call(this, compressor, first_in_statement);
            });
        });
        var global_pure_fns = makePredicate("Boolean decodeURI decodeURIComponent Date encodeURI encodeURIComponent Error escape EvalError isFinite isNaN Number Object parseFloat parseInt RangeError ReferenceError String SyntaxError TypeError unescape URIError");
        AST_Call.DEFMETHOD("is_callee_pure", function (compressor) {
            if (compressor.option("unsafe")) {
                var expr = this.expression;
                var first_arg = (this.args && this.args[0] && this.args[0].evaluate(compressor));
                if (expr.expression && expr.expression.name === "hasOwnProperty" &&
                    (first_arg == null || first_arg.thedef && first_arg.thedef.undeclared)) {
                    return false;
                }
                if (is_undeclared_ref(expr) && global_pure_fns.has(expr.name))
                    return true;
                let static_fn;
                if (expr instanceof AST_Dot
                    && is_undeclared_ref(expr.expression)
                    && (static_fn = static_fns.get(expr.expression.name))
                    && static_fn.has(expr.property)) {
                    return true;
                }
            }
            return !!has_annotation(this, _PURE) || !compressor.pure_funcs(this);
        });
        AST_Node.DEFMETHOD("is_call_pure", return_false);
        AST_Dot.DEFMETHOD("is_call_pure", function (compressor) {
            if (!compressor.option("unsafe"))
                return;
            const expr = this.expression;
            let map;
            if (expr instanceof AST_Array) {
                map = native_fns.get("Array");
            }
            else if (expr.is_boolean()) {
                map = native_fns.get("Boolean");
            }
            else if (expr.is_number(compressor)) {
                map = native_fns.get("Number");
            }
            else if (expr instanceof AST_RegExp) {
                map = native_fns.get("RegExp");
            }
            else if (expr.is_string(compressor)) {
                map = native_fns.get("String");
            }
            else if (!this.may_throw_on_access(compressor)) {
                map = native_fns.get("Object");
            }
            return map && map.has(this.property);
        });
        const pure_prop_access_globals = new Set([
            "Number",
            "String",
            "Array",
            "Object",
            "Function",
            "Promise",
        ]);
        // determine if expression has side effects
        (function (def_has_side_effects) {
            def_has_side_effects(AST_Node, return_true);
            def_has_side_effects(AST_EmptyStatement, return_false);
            def_has_side_effects(AST_Constant, return_false);
            def_has_side_effects(AST_This, return_false);
            function any(list, compressor) {
                for (var i = list.length; --i >= 0;)
                    if (list[i].has_side_effects(compressor))
                        return true;
                return false;
            }
            def_has_side_effects(AST_Block, function (compressor) {
                return any(this.body, compressor);
            });
            def_has_side_effects(AST_Call, function (compressor) {
                if (!this.is_callee_pure(compressor)
                    && (!this.expression.is_call_pure(compressor)
                        || this.expression.has_side_effects(compressor))) {
                    return true;
                }
                return any(this.args, compressor);
            });
            def_has_side_effects(AST_Switch, function (compressor) {
                return this.expression.has_side_effects(compressor)
                    || any(this.body, compressor);
            });
            def_has_side_effects(AST_Case, function (compressor) {
                return this.expression.has_side_effects(compressor)
                    || any(this.body, compressor);
            });
            def_has_side_effects(AST_Try, function (compressor) {
                return any(this.body, compressor)
                    || this.bcatch && this.bcatch.has_side_effects(compressor)
                    || this.bfinally && this.bfinally.has_side_effects(compressor);
            });
            def_has_side_effects(AST_If, function (compressor) {
                return this.condition.has_side_effects(compressor)
                    || this.body && this.body.has_side_effects(compressor)
                    || this.alternative && this.alternative.has_side_effects(compressor);
            });
            def_has_side_effects(AST_LabeledStatement, function (compressor) {
                return this.body.has_side_effects(compressor);
            });
            def_has_side_effects(AST_SimpleStatement, function (compressor) {
                return this.body.has_side_effects(compressor);
            });
            def_has_side_effects(AST_Lambda, return_false);
            def_has_side_effects(AST_Class, function (compressor) {
                if (this.extends && this.extends.has_side_effects(compressor)) {
                    return true;
                }
                return any(this.properties, compressor);
            });
            def_has_side_effects(AST_Binary, function (compressor) {
                return this.left.has_side_effects(compressor)
                    || this.right.has_side_effects(compressor);
            });
            def_has_side_effects(AST_Assign, return_true);
            def_has_side_effects(AST_Conditional, function (compressor) {
                return this.condition.has_side_effects(compressor)
                    || this.consequent.has_side_effects(compressor)
                    || this.alternative.has_side_effects(compressor);
            });
            def_has_side_effects(AST_Unary, function (compressor) {
                return unary_side_effects.has(this.operator)
                    || this.expression.has_side_effects(compressor);
            });
            def_has_side_effects(AST_SymbolRef, function (compressor) {
                return !this.is_declared(compressor) && !pure_prop_access_globals.has(this.name);
            });
            def_has_side_effects(AST_SymbolClassProperty, return_false);
            def_has_side_effects(AST_SymbolDeclaration, return_false);
            def_has_side_effects(AST_Object, function (compressor) {
                return any(this.properties, compressor);
            });
            def_has_side_effects(AST_ObjectProperty, function (compressor) {
                return (this.computed_key() && this.key.has_side_effects(compressor)
                    || this.value && this.value.has_side_effects(compressor));
            });
            def_has_side_effects(AST_ClassProperty, function (compressor) {
                return (this.computed_key() && this.key.has_side_effects(compressor)
                    || this.static && this.value && this.value.has_side_effects(compressor));
            });
            def_has_side_effects(AST_ConciseMethod, function (compressor) {
                return this.computed_key() && this.key.has_side_effects(compressor);
            });
            def_has_side_effects(AST_ObjectGetter, function (compressor) {
                return this.computed_key() && this.key.has_side_effects(compressor);
            });
            def_has_side_effects(AST_ObjectSetter, function (compressor) {
                return this.computed_key() && this.key.has_side_effects(compressor);
            });
            def_has_side_effects(AST_Array, function (compressor) {
                return any(this.elements, compressor);
            });
            def_has_side_effects(AST_Dot, function (compressor) {
                return !this.optional && this.expression.may_throw_on_access(compressor)
                    || this.expression.has_side_effects(compressor);
            });
            def_has_side_effects(AST_Sub, function (compressor) {
                if (this.optional && is_nullish(this.expression, compressor)) {
                    return false;
                }
                return !this.optional && this.expression.may_throw_on_access(compressor)
                    || this.expression.has_side_effects(compressor)
                    || this.property.has_side_effects(compressor);
            });
            def_has_side_effects(AST_Chain, function (compressor) {
                return this.expression.has_side_effects(compressor);
            });
            def_has_side_effects(AST_Sequence, function (compressor) {
                return any(this.expressions, compressor);
            });
            def_has_side_effects(AST_Definitions, function (compressor) {
                return any(this.definitions, compressor);
            });
            def_has_side_effects(AST_VarDef, function () {
                return this.value;
            });
            def_has_side_effects(AST_TemplateSegment, return_false);
            def_has_side_effects(AST_TemplateString, function (compressor) {
                return any(this.segments, compressor);
            });
        })(function (node, func) {
            node.DEFMETHOD("has_side_effects", func);
        });
        // determine if expression may throw
        (function (def_may_throw) {
            def_may_throw(AST_Node, return_true);
            def_may_throw(AST_Constant, return_false);
            def_may_throw(AST_EmptyStatement, return_false);
            def_may_throw(AST_Lambda, return_false);
            def_may_throw(AST_SymbolDeclaration, return_false);
            def_may_throw(AST_This, return_false);
            function any(list, compressor) {
                for (var i = list.length; --i >= 0;)
                    if (list[i].may_throw(compressor))
                        return true;
                return false;
            }
            def_may_throw(AST_Class, function (compressor) {
                if (this.extends && this.extends.may_throw(compressor))
                    return true;
                return any(this.properties, compressor);
            });
            def_may_throw(AST_Array, function (compressor) {
                return any(this.elements, compressor);
            });
            def_may_throw(AST_Assign, function (compressor) {
                if (this.right.may_throw(compressor))
                    return true;
                if (!compressor.has_directive("use strict")
                    && this.operator == "="
                    && this.left instanceof AST_SymbolRef) {
                    return false;
                }
                return this.left.may_throw(compressor);
            });
            def_may_throw(AST_Binary, function (compressor) {
                return this.left.may_throw(compressor)
                    || this.right.may_throw(compressor);
            });
            def_may_throw(AST_Block, function (compressor) {
                return any(this.body, compressor);
            });
            def_may_throw(AST_Call, function (compressor) {
                if (this.optional && is_nullish(this.expression, compressor))
                    return false;
                if (any(this.args, compressor))
                    return true;
                if (this.is_callee_pure(compressor))
                    return false;
                if (this.expression.may_throw(compressor))
                    return true;
                return !(this.expression instanceof AST_Lambda)
                    || any(this.expression.body, compressor);
            });
            def_may_throw(AST_Case, function (compressor) {
                return this.expression.may_throw(compressor)
                    || any(this.body, compressor);
            });
            def_may_throw(AST_Conditional, function (compressor) {
                return this.condition.may_throw(compressor)
                    || this.consequent.may_throw(compressor)
                    || this.alternative.may_throw(compressor);
            });
            def_may_throw(AST_Definitions, function (compressor) {
                return any(this.definitions, compressor);
            });
            def_may_throw(AST_If, function (compressor) {
                return this.condition.may_throw(compressor)
                    || this.body && this.body.may_throw(compressor)
                    || this.alternative && this.alternative.may_throw(compressor);
            });
            def_may_throw(AST_LabeledStatement, function (compressor) {
                return this.body.may_throw(compressor);
            });
            def_may_throw(AST_Object, function (compressor) {
                return any(this.properties, compressor);
            });
            def_may_throw(AST_ObjectProperty, function (compressor) {
                // TODO key may throw too
                return this.value ? this.value.may_throw(compressor) : false;
            });
            def_may_throw(AST_ClassProperty, function (compressor) {
                return (this.computed_key() && this.key.may_throw(compressor)
                    || this.static && this.value && this.value.may_throw(compressor));
            });
            def_may_throw(AST_ConciseMethod, function (compressor) {
                return this.computed_key() && this.key.may_throw(compressor);
            });
            def_may_throw(AST_ObjectGetter, function (compressor) {
                return this.computed_key() && this.key.may_throw(compressor);
            });
            def_may_throw(AST_ObjectSetter, function (compressor) {
                return this.computed_key() && this.key.may_throw(compressor);
            });
            def_may_throw(AST_Return, function (compressor) {
                return this.value && this.value.may_throw(compressor);
            });
            def_may_throw(AST_Sequence, function (compressor) {
                return any(this.expressions, compressor);
            });
            def_may_throw(AST_SimpleStatement, function (compressor) {
                return this.body.may_throw(compressor);
            });
            def_may_throw(AST_Dot, function (compressor) {
                return !this.optional && this.expression.may_throw_on_access(compressor)
                    || this.expression.may_throw(compressor);
            });
            def_may_throw(AST_Sub, function (compressor) {
                if (this.optional && is_nullish(this.expression, compressor))
                    return false;
                return !this.optional && this.expression.may_throw_on_access(compressor)
                    || this.expression.may_throw(compressor)
                    || this.property.may_throw(compressor);
            });
            def_may_throw(AST_Chain, function (compressor) {
                return this.expression.may_throw(compressor);
            });
            def_may_throw(AST_Switch, function (compressor) {
                return this.expression.may_throw(compressor)
                    || any(this.body, compressor);
            });
            def_may_throw(AST_SymbolRef, function (compressor) {
                return !this.is_declared(compressor) && !pure_prop_access_globals.has(this.name);
            });
            def_may_throw(AST_SymbolClassProperty, return_false);
            def_may_throw(AST_Try, function (compressor) {
                return this.bcatch ? this.bcatch.may_throw(compressor) : any(this.body, compressor)
                    || this.bfinally && this.bfinally.may_throw(compressor);
            });
            def_may_throw(AST_Unary, function (compressor) {
                if (this.operator == "typeof" && this.expression instanceof AST_SymbolRef)
                    return false;
                return this.expression.may_throw(compressor);
            });
            def_may_throw(AST_VarDef, function (compressor) {
                if (!this.value)
                    return false;
                return this.value.may_throw(compressor);
            });
        })(function (node, func) {
            node.DEFMETHOD("may_throw", func);
        });
        // determine if expression is constant
        (function (def_is_constant_expression) {
            function all_refs_local(scope) {
                let result = true;
                walk(this, node => {
                    if (node instanceof AST_SymbolRef) {
                        if (has_flag(this, INLINED)) {
                            result = false;
                            return walk_abort;
                        }
                        var def = node.definition();
                        if (member(def, this.enclosed)
                            && !this.variables.has(def.name)) {
                            if (scope) {
                                var scope_def = scope.find_variable(node);
                                if (def.undeclared ? !scope_def : scope_def === def) {
                                    result = "f";
                                    return true;
                                }
                            }
                            result = false;
                            return walk_abort;
                        }
                        return true;
                    }
                    if (node instanceof AST_This && this instanceof AST_Arrow) {
                        // TODO check arguments too!
                        result = false;
                        return walk_abort;
                    }
                });
                return result;
            }
            def_is_constant_expression(AST_Node, return_false);
            def_is_constant_expression(AST_Constant, return_true);
            def_is_constant_expression(AST_Class, function (scope) {
                if (this.extends && !this.extends.is_constant_expression(scope)) {
                    return false;
                }
                for (const prop of this.properties) {
                    if (prop.computed_key() && !prop.key.is_constant_expression(scope)) {
                        return false;
                    }
                    if (prop.static && prop.value && !prop.value.is_constant_expression(scope)) {
                        return false;
                    }
                }
                return all_refs_local.call(this, scope);
            });
            def_is_constant_expression(AST_Lambda, all_refs_local);
            def_is_constant_expression(AST_Unary, function () {
                return this.expression.is_constant_expression();
            });
            def_is_constant_expression(AST_Binary, function () {
                return this.left.is_constant_expression()
                    && this.right.is_constant_expression();
            });
            def_is_constant_expression(AST_Array, function () {
                return this.elements.every((l) => l.is_constant_expression());
            });
            def_is_constant_expression(AST_Object, function () {
                return this.properties.every((l) => l.is_constant_expression());
            });
            def_is_constant_expression(AST_ObjectProperty, function () {
                return !!(!(this.key instanceof AST_Node) && this.value && this.value.is_constant_expression());
            });
        })(function (node, func) {
            node.DEFMETHOD("is_constant_expression", func);
        });
        // tell me if a statement aborts
        function aborts(thing) {
            return thing && thing.aborts();
        }
        (function (def_aborts) {
            def_aborts(AST_Statement, return_null);
            def_aborts(AST_Jump, return_this);
            function block_aborts() {
                for (var i = 0; i < this.body.length; i++) {
                    if (aborts(this.body[i])) {
                        return this.body[i];
                    }
                }
                return null;
            }
            def_aborts(AST_Import, function () { return null; });
            def_aborts(AST_BlockStatement, block_aborts);
            def_aborts(AST_SwitchBranch, block_aborts);
            def_aborts(AST_If, function () {
                return this.alternative && aborts(this.body) && aborts(this.alternative) && this;
            });
        })(function (node, func) {
            node.DEFMETHOD("aborts", func);
        });
        /* -----[ optimizers ]----- */
        var directives = new Set(["use asm", "use strict"]);
        def_optimize(AST_Directive, function (self, compressor) {
            if (compressor.option("directives")
                && (!directives.has(self.value) || compressor.has_directive(self.value) !== self)) {
                return make_node(AST_EmptyStatement, self);
            }
            return self;
        });
        def_optimize(AST_Debugger, function (self, compressor) {
            if (compressor.option("drop_debugger"))
                return make_node(AST_EmptyStatement, self);
            return self;
        });
        def_optimize(AST_LabeledStatement, function (self, compressor) {
            if (self.body instanceof AST_Break
                && compressor.loopcontrol_target(self.body) === self.body) {
                return make_node(AST_EmptyStatement, self);
            }
            return self.label.references.length == 0 ? self.body : self;
        });
        def_optimize(AST_Block, function (self, compressor) {
            tighten_body(self.body, compressor);
            return self;
        });
        function can_be_extracted_from_if_block(node) {
            return !(node instanceof AST_Const
                || node instanceof AST_Let
                || node instanceof AST_Class);
        }
        def_optimize(AST_BlockStatement, function (self, compressor) {
            tighten_body(self.body, compressor);
            switch (self.body.length) {
                case 1:
                    if (!compressor.has_directive("use strict")
                        && compressor.parent() instanceof AST_If
                        && can_be_extracted_from_if_block(self.body[0])
                        || can_be_evicted_from_block(self.body[0])) {
                        return self.body[0];
                    }
                    break;
                case 0: return make_node(AST_EmptyStatement, self);
            }
            return self;
        });
        function opt_AST_Lambda(self, compressor) {
            tighten_body(self.body, compressor);
            if (compressor.option("side_effects")
                && self.body.length == 1
                && self.body[0] === compressor.has_directive("use strict")) {
                self.body.length = 0;
            }
            return self;
        }
        def_optimize(AST_Lambda, opt_AST_Lambda);
        const r_keep_assign = /keep_assign/;
        AST_Scope.DEFMETHOD("drop_unused", function (compressor) {
            if (!compressor.option("unused"))
                return;
            if (compressor.has_directive("use asm"))
                return;
            var self = this;
            if (self.pinned())
                return;
            var drop_funcs = !(self instanceof AST_Toplevel) || compressor.toplevel.funcs;
            var drop_vars = !(self instanceof AST_Toplevel) || compressor.toplevel.vars;
            const assign_as_unused = r_keep_assign.test(compressor.option("unused")) ? return_false : function (node) {
                if (node instanceof AST_Assign
                    && !node.logical
                    && (has_flag(node, WRITE_ONLY) || node.operator == "=")) {
                    return node.left;
                }
                if (node instanceof AST_Unary && has_flag(node, WRITE_ONLY)) {
                    return node.expression;
                }
            };
            var in_use_ids = new Map();
            var fixed_ids = new Map();
            if (self instanceof AST_Toplevel && compressor.top_retain) {
                self.variables.forEach(function (def) {
                    if (compressor.top_retain(def) && !in_use_ids.has(def.id)) {
                        in_use_ids.set(def.id, def);
                    }
                });
            }
            var var_defs_by_id = new Map();
            var initializations = new Map();
            // pass 1: find out which symbols are directly used in
            // this scope (not in nested scopes).
            var scope = this;
            var tw = new TreeWalker(function (node, descend) {
                if (node instanceof AST_Lambda && node.uses_arguments && !tw.has_directive("use strict")) {
                    node.argnames.forEach(function (argname) {
                        if (!(argname instanceof AST_SymbolDeclaration))
                            return;
                        var def = argname.definition();
                        if (!in_use_ids.has(def.id)) {
                            in_use_ids.set(def.id, def);
                        }
                    });
                }
                if (node === self)
                    return;
                if (node instanceof AST_Defun || node instanceof AST_DefClass) {
                    var node_def = node.name.definition();
                    const in_export = tw.parent() instanceof AST_Export;
                    if (in_export || !drop_funcs && scope === self) {
                        if (node_def.global && !in_use_ids.has(node_def.id)) {
                            in_use_ids.set(node_def.id, node_def);
                        }
                    }
                    if (node instanceof AST_DefClass) {
                        if (node.extends
                            && (node.extends.has_side_effects(compressor)
                                || node.extends.may_throw(compressor))) {
                            node.extends.walk(tw);
                        }
                        for (const prop of node.properties) {
                            if (prop.has_side_effects(compressor) ||
                                prop.may_throw(compressor)) {
                                prop.walk(tw);
                            }
                        }
                    }
                    map_add(initializations, node_def.id, node);
                    return true; // don't go in nested scopes
                }
                if (node instanceof AST_SymbolFunarg && scope === self) {
                    map_add(var_defs_by_id, node.definition().id, node);
                }
                if (node instanceof AST_Definitions && scope === self) {
                    const in_export = tw.parent() instanceof AST_Export;
                    node.definitions.forEach(function (def) {
                        if (def.name instanceof AST_SymbolVar) {
                            map_add(var_defs_by_id, def.name.definition().id, def);
                        }
                        if (in_export || !drop_vars) {
                            walk(def.name, node => {
                                if (node instanceof AST_SymbolDeclaration) {
                                    const def = node.definition();
                                    if ((in_export || def.global)
                                        && !in_use_ids.has(def.id)) {
                                        in_use_ids.set(def.id, def);
                                    }
                                }
                            });
                        }
                        if (def.value) {
                            if (def.name instanceof AST_Destructuring) {
                                def.walk(tw);
                            }
                            else {
                                var node_def = def.name.definition();
                                map_add(initializations, node_def.id, def.value);
                                if (!node_def.chained && def.name.fixed_value() === def.value) {
                                    fixed_ids.set(node_def.id, def);
                                }
                            }
                            if (def.value.has_side_effects(compressor)) {
                                def.value.walk(tw);
                            }
                        }
                    });
                    return true;
                }
                return scan_ref_scoped(node, descend);
            });
            self.walk(tw);
            // pass 2: for every used symbol we need to walk its
            // initialization code to figure out if it uses other
            // symbols (that may not be in_use).
            tw = new TreeWalker(scan_ref_scoped);
            in_use_ids.forEach(function (def) {
                var init = initializations.get(def.id);
                if (init)
                    init.forEach(function (init) {
                        init.walk(tw);
                    });
            });
            // pass 3: we should drop declarations not in_use
            var tt = new TreeTransformer(function before(node, descend, in_list) {
                var parent = tt.parent();
                if (drop_vars) {
                    const sym = assign_as_unused(node);
                    if (sym instanceof AST_SymbolRef) {
                        var def = sym.definition();
                        var in_use = in_use_ids.has(def.id);
                        if (node instanceof AST_Assign) {
                            if (!in_use || fixed_ids.has(def.id) && fixed_ids.get(def.id) !== node) {
                                return maintain_this_binding(parent, node, node.right.transform(tt));
                            }
                        }
                        else if (!in_use)
                            return in_list ? MAP.skip : make_node(AST_Number, node, {
                                value: 0
                            });
                    }
                }
                if (scope !== self)
                    return;
                var def;
                if (node.name
                    && (node instanceof AST_ClassExpression
                        && !keep_name(compressor.option("keep_classnames"), (def = node.name.definition()).name)
                        || node instanceof AST_Function
                            && !keep_name(compressor.option("keep_fnames"), (def = node.name.definition()).name))) {
                    // any declarations with same name will overshadow
                    // name of this anonymous function and can therefore
                    // never be used anywhere
                    if (!in_use_ids.has(def.id) || def.orig.length > 1)
                        node.name = null;
                }
                if (node instanceof AST_Lambda && !(node instanceof AST_Accessor)) {
                    var trim = !compressor.option("keep_fargs");
                    for (var a = node.argnames, i = a.length; --i >= 0;) {
                        var sym = a[i];
                        if (sym instanceof AST_Expansion) {
                            sym = sym.expression;
                        }
                        if (sym instanceof AST_DefaultAssign) {
                            sym = sym.left;
                        }
                        // Do not drop destructuring arguments.
                        // They constitute a type assertion, so dropping
                        // them would stop that TypeError which would happen
                        // if someone called it with an incorrectly formatted
                        // parameter.
                        if (!(sym instanceof AST_Destructuring) && !in_use_ids.has(sym.definition().id)) {
                            set_flag(sym, UNUSED);
                            if (trim) {
                                a.pop();
                            }
                        }
                        else {
                            trim = false;
                        }
                    }
                }
                if ((node instanceof AST_Defun || node instanceof AST_DefClass) && node !== self) {
                    const def = node.name.definition();
                    let keep = def.global && !drop_funcs || in_use_ids.has(def.id);
                    if (!keep) {
                        def.eliminated++;
                        if (node instanceof AST_DefClass) {
                            // Classes might have extends with side effects
                            const side_effects = node.drop_side_effect_free(compressor);
                            if (side_effects) {
                                return make_node(AST_SimpleStatement, node, {
                                    body: side_effects
                                });
                            }
                        }
                        return in_list ? MAP.skip : make_node(AST_EmptyStatement, node);
                    }
                }
                if (node instanceof AST_Definitions && !(parent instanceof AST_ForIn && parent.init === node)) {
                    var drop_block = !(parent instanceof AST_Toplevel) && !(node instanceof AST_Var);
                    // place uninitialized names at the start
                    var body = [], head = [], tail = [];
                    // for unused names whose initialization has
                    // side effects, we can cascade the init. code
                    // into the next one, or next statement.
                    var side_effects = [];
                    node.definitions.forEach(function (def) {
                        if (def.value)
                            def.value = def.value.transform(tt);
                        var is_destructure = def.name instanceof AST_Destructuring;
                        var sym = is_destructure
                            ? new SymbolDef(null, { name: "<destructure>" }) /* fake SymbolDef */
                            : def.name.definition();
                        if (drop_block && sym.global)
                            return tail.push(def);
                        if (!(drop_vars || drop_block)
                            || is_destructure
                                && (def.name.names.length
                                    || def.name.is_array
                                    || compressor.option("pure_getters") != true)
                            || in_use_ids.has(sym.id)) {
                            if (def.value && fixed_ids.has(sym.id) && fixed_ids.get(sym.id) !== def) {
                                def.value = def.value.drop_side_effect_free(compressor);
                            }
                            if (def.name instanceof AST_SymbolVar) {
                                var var_defs = var_defs_by_id.get(sym.id);
                                if (var_defs.length > 1 && (!def.value || sym.orig.indexOf(def.name) > sym.eliminated)) {
                                    if (def.value) {
                                        var ref = make_node(AST_SymbolRef, def.name, def.name);
                                        sym.references.push(ref);
                                        var assign = make_node(AST_Assign, def, {
                                            operator: "=",
                                            logical: false,
                                            left: ref,
                                            right: def.value
                                        });
                                        if (fixed_ids.get(sym.id) === def) {
                                            fixed_ids.set(sym.id, assign);
                                        }
                                        side_effects.push(assign.transform(tt));
                                    }
                                    remove(var_defs, def);
                                    sym.eliminated++;
                                    return;
                                }
                            }
                            if (def.value) {
                                if (side_effects.length > 0) {
                                    if (tail.length > 0) {
                                        side_effects.push(def.value);
                                        def.value = make_sequence(def.value, side_effects);
                                    }
                                    else {
                                        body.push(make_node(AST_SimpleStatement, node, {
                                            body: make_sequence(node, side_effects)
                                        }));
                                    }
                                    side_effects = [];
                                }
                                tail.push(def);
                            }
                            else {
                                head.push(def);
                            }
                        }
                        else if (sym.orig[0] instanceof AST_SymbolCatch) {
                            var value = def.value && def.value.drop_side_effect_free(compressor);
                            if (value)
                                side_effects.push(value);
                            def.value = null;
                            head.push(def);
                        }
                        else {
                            var value = def.value && def.value.drop_side_effect_free(compressor);
                            if (value) {
                                side_effects.push(value);
                            }
                            sym.eliminated++;
                        }
                    });
                    if (head.length > 0 || tail.length > 0) {
                        node.definitions = head.concat(tail);
                        body.push(node);
                    }
                    if (side_effects.length > 0) {
                        body.push(make_node(AST_SimpleStatement, node, {
                            body: make_sequence(node, side_effects)
                        }));
                    }
                    switch (body.length) {
                        case 0:
                            return in_list ? MAP.skip : make_node(AST_EmptyStatement, node);
                        case 1:
                            return body[0];
                        default:
                            return in_list ? MAP.splice(body) : make_node(AST_BlockStatement, node, {
                                body: body
                            });
                    }
                }
                // certain combination of unused name + side effect leads to:
                //    https://github.com/mishoo/UglifyJS2/issues/44
                //    https://github.com/mishoo/UglifyJS2/issues/1830
                //    https://github.com/mishoo/UglifyJS2/issues/1838
                // that's an invalid AST.
                // We fix it at this stage by moving the `var` outside the `for`.
                if (node instanceof AST_For) {
                    descend(node, this);
                    var block;
                    if (node.init instanceof AST_BlockStatement) {
                        block = node.init;
                        node.init = block.body.pop();
                        block.body.push(node);
                    }
                    if (node.init instanceof AST_SimpleStatement) {
                        node.init = node.init.body;
                    }
                    else if (is_empty(node.init)) {
                        node.init = null;
                    }
                    return !block ? node : in_list ? MAP.splice(block.body) : block;
                }
                if (node instanceof AST_LabeledStatement
                    && node.body instanceof AST_For) {
                    descend(node, this);
                    if (node.body instanceof AST_BlockStatement) {
                        var block = node.body;
                        node.body = block.body.pop();
                        block.body.push(node);
                        return in_list ? MAP.splice(block.body) : block;
                    }
                    return node;
                }
                if (node instanceof AST_BlockStatement) {
                    descend(node, this);
                    if (in_list && node.body.every(can_be_evicted_from_block)) {
                        return MAP.splice(node.body);
                    }
                    return node;
                }
                if (node instanceof AST_Scope) {
                    const save_scope = scope;
                    scope = node;
                    descend(node, this);
                    scope = save_scope;
                    return node;
                }
            });
            self.transform(tt);
            function scan_ref_scoped(node, descend) {
                var node_def;
                const sym = assign_as_unused(node);
                if (sym instanceof AST_SymbolRef
                    && !is_ref_of(node.left, AST_SymbolBlockDeclaration)
                    && self.variables.get(sym.name) === (node_def = sym.definition())) {
                    if (node instanceof AST_Assign) {
                        node.right.walk(tw);
                        if (!node_def.chained && node.left.fixed_value() === node.right) {
                            fixed_ids.set(node_def.id, node);
                        }
                    }
                    return true;
                }
                if (node instanceof AST_SymbolRef) {
                    node_def = node.definition();
                    if (!in_use_ids.has(node_def.id)) {
                        in_use_ids.set(node_def.id, node_def);
                        if (node_def.orig[0] instanceof AST_SymbolCatch) {
                            const redef = node_def.scope.is_block_scope()
                                && node_def.scope.get_defun_scope().variables.get(node_def.name);
                            if (redef)
                                in_use_ids.set(redef.id, redef);
                        }
                    }
                    return true;
                }
                if (node instanceof AST_Scope) {
                    var save_scope = scope;
                    scope = node;
                    descend();
                    scope = save_scope;
                    return true;
                }
            }
        });
        AST_Scope.DEFMETHOD("hoist_declarations", function (compressor) {
            var self = this;
            if (compressor.has_directive("use asm"))
                return self;
            // Hoisting makes no sense in an arrow func
            if (!Array.isArray(self.body))
                return self;
            var hoist_funs = compressor.option("hoist_funs");
            var hoist_vars = compressor.option("hoist_vars");
            if (hoist_funs || hoist_vars) {
                var dirs = [];
                var hoisted = [];
                var vars = new Map(), vars_found = 0, var_decl = 0;
                // let's count var_decl first, we seem to waste a lot of
                // space if we hoist `var` when there's only one.
                walk(self, node => {
                    if (node instanceof AST_Scope && node !== self)
                        return true;
                    if (node instanceof AST_Var) {
                        ++var_decl;
                        return true;
                    }
                });
                hoist_vars = hoist_vars && var_decl > 1;
                var tt = new TreeTransformer(function before(node) {
                    if (node !== self) {
                        if (node instanceof AST_Directive) {
                            dirs.push(node);
                            return make_node(AST_EmptyStatement, node);
                        }
                        if (hoist_funs && node instanceof AST_Defun
                            && !(tt.parent() instanceof AST_Export)
                            && tt.parent() === self) {
                            hoisted.push(node);
                            return make_node(AST_EmptyStatement, node);
                        }
                        if (hoist_vars
                            && node instanceof AST_Var
                            && !node.definitions.some(def => def.name instanceof AST_Destructuring)) {
                            node.definitions.forEach(function (def) {
                                vars.set(def.name.name, def);
                                ++vars_found;
                            });
                            var seq = node.to_assignments(compressor);
                            var p = tt.parent();
                            if (p instanceof AST_ForIn && p.init === node) {
                                if (seq == null) {
                                    var def = node.definitions[0].name;
                                    return make_node(AST_SymbolRef, def, def);
                                }
                                return seq;
                            }
                            if (p instanceof AST_For && p.init === node) {
                                return seq;
                            }
                            if (!seq)
                                return make_node(AST_EmptyStatement, node);
                            return make_node(AST_SimpleStatement, node, {
                                body: seq
                            });
                        }
                        if (node instanceof AST_Scope)
                            return node; // to avoid descending in nested scopes
                    }
                });
                self = self.transform(tt);
                if (vars_found > 0) {
                    // collect only vars which don't show up in self's arguments list
                    var defs = [];
                    const is_lambda = self instanceof AST_Lambda;
                    const args_as_names = is_lambda ? self.args_as_names() : null;
                    vars.forEach((def, name) => {
                        if (is_lambda && args_as_names.some((x) => x.name === def.name.name)) {
                            vars.delete(name);
                        }
                        else {
                            def = def.clone();
                            def.value = null;
                            defs.push(def);
                            vars.set(name, def);
                        }
                    });
                    if (defs.length > 0) {
                        // try to merge in assignments
                        for (var i = 0; i < self.body.length;) {
                            if (self.body[i] instanceof AST_SimpleStatement) {
                                var expr = self.body[i].body, sym, assign;
                                if (expr instanceof AST_Assign
                                    && expr.operator == "="
                                    && (sym = expr.left) instanceof AST_Symbol
                                    && vars.has(sym.name)) {
                                    var def = vars.get(sym.name);
                                    if (def.value)
                                        break;
                                    def.value = expr.right;
                                    remove(defs, def);
                                    defs.push(def);
                                    self.body.splice(i, 1);
                                    continue;
                                }
                                if (expr instanceof AST_Sequence
                                    && (assign = expr.expressions[0]) instanceof AST_Assign
                                    && assign.operator == "="
                                    && (sym = assign.left) instanceof AST_Symbol
                                    && vars.has(sym.name)) {
                                    var def = vars.get(sym.name);
                                    if (def.value)
                                        break;
                                    def.value = assign.right;
                                    remove(defs, def);
                                    defs.push(def);
                                    self.body[i].body = make_sequence(expr, expr.expressions.slice(1));
                                    continue;
                                }
                            }
                            if (self.body[i] instanceof AST_EmptyStatement) {
                                self.body.splice(i, 1);
                                continue;
                            }
                            if (self.body[i] instanceof AST_BlockStatement) {
                                self.body.splice(i, 1, ...self.body[i].body);
                                continue;
                            }
                            break;
                        }
                        defs = make_node(AST_Var, self, {
                            definitions: defs
                        });
                        hoisted.push(defs);
                    }
                }
                self.body = dirs.concat(hoisted, self.body);
            }
            return self;
        });
        AST_Scope.DEFMETHOD("hoist_properties", function (compressor) {
            var self = this;
            if (!compressor.option("hoist_props") || compressor.has_directive("use asm"))
                return self;
            var top_retain = self instanceof AST_Toplevel && compressor.top_retain || return_false;
            var defs_by_id = new Map();
            var hoister = new TreeTransformer(function (node, descend) {
                if (node instanceof AST_Definitions
                    && hoister.parent() instanceof AST_Export)
                    return node;
                if (node instanceof AST_VarDef) {
                    const sym = node.name;
                    let def;
                    let value;
                    if (sym.scope === self
                        && (def = sym.definition()).escaped != 1
                        && !def.assignments
                        && !def.direct_access
                        && !def.single_use
                        && !compressor.exposed(def)
                        && !top_retain(def)
                        && (value = sym.fixed_value()) === node.value
                        && value instanceof AST_Object
                        && !value.properties.some(prop => prop instanceof AST_Expansion || prop.computed_key())) {
                        descend(node, this);
                        const defs = new Map();
                        const assignments = [];
                        value.properties.forEach(({ key, value }) => {
                            const scope = find_scope(hoister);
                            const symbol = self.create_symbol(sym.CTOR, {
                                source: sym,
                                scope,
                                conflict_scopes: new Set([
                                    scope,
                                    ...sym.definition().references.map(ref => ref.scope)
                                ]),
                                tentative_name: sym.name + "_" + key
                            });
                            defs.set(String(key), symbol.definition());
                            assignments.push(make_node(AST_VarDef, node, {
                                name: symbol,
                                value
                            }));
                        });
                        defs_by_id.set(def.id, defs);
                        return MAP.splice(assignments);
                    }
                }
                else if (node instanceof AST_PropAccess
                    && node.expression instanceof AST_SymbolRef) {
                    const defs = defs_by_id.get(node.expression.definition().id);
                    if (defs) {
                        const def = defs.get(String(get_value(node.property)));
                        const sym = make_node(AST_SymbolRef, node, {
                            name: def.name,
                            scope: node.expression.scope,
                            thedef: def
                        });
                        sym.reference({});
                        return sym;
                    }
                }
            });
            return self.transform(hoister);
        });
        // drop_side_effect_free()
        // remove side-effect-free parts which only affects return value
        (function (def_drop_side_effect_free) {
            // Drop side-effect-free elements from an array of expressions.
            // Returns an array of expressions with side-effects or null
            // if all elements were dropped. Note: original array may be
            // returned if nothing changed.
            function trim(nodes, compressor, first_in_statement) {
                var len = nodes.length;
                if (!len)
                    return null;
                var ret = [], changed = false;
                for (var i = 0; i < len; i++) {
                    var node = nodes[i].drop_side_effect_free(compressor, first_in_statement);
                    changed |= node !== nodes[i];
                    if (node) {
                        ret.push(node);
                        first_in_statement = false;
                    }
                }
                return changed ? ret.length ? ret : null : nodes;
            }
            def_drop_side_effect_free(AST_Node, return_this);
            def_drop_side_effect_free(AST_Constant, return_null);
            def_drop_side_effect_free(AST_This, return_null);
            def_drop_side_effect_free(AST_Call, function (compressor, first_in_statement) {
                if (this.optional && is_nullish(this.expression, compressor)) {
                    return make_node(AST_Undefined, this);
                }
                if (!this.is_callee_pure(compressor)) {
                    if (this.expression.is_call_pure(compressor)) {
                        var exprs = this.args.slice();
                        exprs.unshift(this.expression.expression);
                        exprs = trim(exprs, compressor, first_in_statement);
                        return exprs && make_sequence(this, exprs);
                    }
                    if (is_func_expr(this.expression)
                        && (!this.expression.name || !this.expression.name.definition().references.length)) {
                        var node = this.clone();
                        node.expression.process_expression(false, compressor);
                        return node;
                    }
                    return this;
                }
                var args = trim(this.args, compressor, first_in_statement);
                return args && make_sequence(this, args);
            });
            def_drop_side_effect_free(AST_Accessor, return_null);
            def_drop_side_effect_free(AST_Function, return_null);
            def_drop_side_effect_free(AST_Arrow, return_null);
            def_drop_side_effect_free(AST_Class, function (compressor) {
                const with_effects = [];
                const trimmed_extends = this.extends && this.extends.drop_side_effect_free(compressor);
                if (trimmed_extends)
                    with_effects.push(trimmed_extends);
                for (const prop of this.properties) {
                    const trimmed_prop = prop.drop_side_effect_free(compressor);
                    if (trimmed_prop)
                        with_effects.push(trimmed_prop);
                }
                if (!with_effects.length)
                    return null;
                return make_sequence(this, with_effects);
            });
            def_drop_side_effect_free(AST_Binary, function (compressor, first_in_statement) {
                var right = this.right.drop_side_effect_free(compressor);
                if (!right)
                    return this.left.drop_side_effect_free(compressor, first_in_statement);
                if (lazy_op.has(this.operator)) {
                    if (right === this.right)
                        return this;
                    var node = this.clone();
                    node.right = right;
                    return node;
                }
                else {
                    var left = this.left.drop_side_effect_free(compressor, first_in_statement);
                    if (!left)
                        return this.right.drop_side_effect_free(compressor, first_in_statement);
                    return make_sequence(this, [left, right]);
                }
            });
            def_drop_side_effect_free(AST_Assign, function (compressor) {
                if (this.logical)
                    return this;
                var left = this.left;
                if (left.has_side_effects(compressor)
                    || compressor.has_directive("use strict")
                        && left instanceof AST_PropAccess
                        && left.expression.is_constant()) {
                    return this;
                }
                set_flag(this, WRITE_ONLY);
                while (left instanceof AST_PropAccess) {
                    left = left.expression;
                }
                if (left.is_constant_expression(compressor.find_parent(AST_Scope))) {
                    return this.right.drop_side_effect_free(compressor);
                }
                return this;
            });
            def_drop_side_effect_free(AST_Conditional, function (compressor) {
                var consequent = this.consequent.drop_side_effect_free(compressor);
                var alternative = this.alternative.drop_side_effect_free(compressor);
                if (consequent === this.consequent && alternative === this.alternative)
                    return this;
                if (!consequent)
                    return alternative ? make_node(AST_Binary, this, {
                        operator: "||",
                        left: this.condition,
                        right: alternative
                    }) : this.condition.drop_side_effect_free(compressor);
                if (!alternative)
                    return make_node(AST_Binary, this, {
                        operator: "&&",
                        left: this.condition,
                        right: consequent
                    });
                var node = this.clone();
                node.consequent = consequent;
                node.alternative = alternative;
                return node;
            });
            def_drop_side_effect_free(AST_Unary, function (compressor, first_in_statement) {
                if (unary_side_effects.has(this.operator)) {
                    if (!this.expression.has_side_effects(compressor)) {
                        set_flag(this, WRITE_ONLY);
                    }
                    else {
                        clear_flag(this, WRITE_ONLY);
                    }
                    return this;
                }
                if (this.operator == "typeof" && this.expression instanceof AST_SymbolRef)
                    return null;
                var expression = this.expression.drop_side_effect_free(compressor, first_in_statement);
                if (first_in_statement && expression && is_iife_call(expression)) {
                    if (expression === this.expression && this.operator == "!")
                        return this;
                    return expression.negate(compressor, first_in_statement);
                }
                return expression;
            });
            def_drop_side_effect_free(AST_SymbolRef, function (compressor) {
                const safe_access = this.is_declared(compressor)
                    || pure_prop_access_globals.has(this.name);
                return safe_access ? null : this;
            });
            def_drop_side_effect_free(AST_Object, function (compressor, first_in_statement) {
                var values = trim(this.properties, compressor, first_in_statement);
                return values && make_sequence(this, values);
            });
            def_drop_side_effect_free(AST_ObjectProperty, function (compressor, first_in_statement) {
                const computed_key = this instanceof AST_ObjectKeyVal && this.key instanceof AST_Node;
                const key = computed_key && this.key.drop_side_effect_free(compressor, first_in_statement);
                const value = this.value && this.value.drop_side_effect_free(compressor, first_in_statement);
                if (key && value) {
                    return make_sequence(this, [key, value]);
                }
                return key || value;
            });
            def_drop_side_effect_free(AST_ClassProperty, function (compressor) {
                const key = this.computed_key() && this.key.drop_side_effect_free(compressor);
                const value = this.static && this.value
                    && this.value.drop_side_effect_free(compressor);
                if (key && value)
                    return make_sequence(this, [key, value]);
                return key || value || null;
            });
            def_drop_side_effect_free(AST_ConciseMethod, function () {
                return this.computed_key() ? this.key : null;
            });
            def_drop_side_effect_free(AST_ObjectGetter, function () {
                return this.computed_key() ? this.key : null;
            });
            def_drop_side_effect_free(AST_ObjectSetter, function () {
                return this.computed_key() ? this.key : null;
            });
            def_drop_side_effect_free(AST_Array, function (compressor, first_in_statement) {
                var values = trim(this.elements, compressor, first_in_statement);
                return values && make_sequence(this, values);
            });
            def_drop_side_effect_free(AST_Dot, function (compressor, first_in_statement) {
                if (this.optional) {
                    return is_nullish(this.expression, compressor) ? make_node(AST_Undefined, this) : this;
                }
                if (this.expression.may_throw_on_access(compressor))
                    return this;
                return this.expression.drop_side_effect_free(compressor, first_in_statement);
            });
            def_drop_side_effect_free(AST_Sub, function (compressor, first_in_statement) {
                if (this.optional) {
                    return is_nullish(this.expression, compressor) ? make_node(AST_Undefined, this) : this;
                }
                if (this.expression.may_throw_on_access(compressor))
                    return this;
                var expression = this.expression.drop_side_effect_free(compressor, first_in_statement);
                if (!expression)
                    return this.property.drop_side_effect_free(compressor, first_in_statement);
                var property = this.property.drop_side_effect_free(compressor);
                if (!property)
                    return expression;
                return make_sequence(this, [expression, property]);
            });
            def_drop_side_effect_free(AST_Chain, function (compressor, first_in_statement) {
                return this.expression.drop_side_effect_free(compressor, first_in_statement);
            });
            def_drop_side_effect_free(AST_Sequence, function (compressor) {
                var last = this.tail_node();
                var expr = last.drop_side_effect_free(compressor);
                if (expr === last)
                    return this;
                var expressions = this.expressions.slice(0, -1);
                if (expr)
                    expressions.push(expr);
                if (!expressions.length) {
                    return make_node(AST_Number, this, { value: 0 });
                }
                return make_sequence(this, expressions);
            });
            def_drop_side_effect_free(AST_Expansion, function (compressor, first_in_statement) {
                return this.expression.drop_side_effect_free(compressor, first_in_statement);
            });
            def_drop_side_effect_free(AST_TemplateSegment, return_null);
            def_drop_side_effect_free(AST_TemplateString, function (compressor) {
                var values = trim(this.segments, compressor, first_in_statement);
                return values && make_sequence(this, values);
            });
        })(function (node, func) {
            node.DEFMETHOD("drop_side_effect_free", func);
        });
        def_optimize(AST_SimpleStatement, function (self, compressor) {
            if (compressor.option("side_effects")) {
                var body = self.body;
                var node = body.drop_side_effect_free(compressor, true);
                if (!node) {
                    return make_node(AST_EmptyStatement, self);
                }
                if (node !== body) {
                    return make_node(AST_SimpleStatement, self, { body: node });
                }
            }
            return self;
        });
        def_optimize(AST_While, function (self, compressor) {
            return compressor.option("loops") ? make_node(AST_For, self, self).optimize(compressor) : self;
        });
        function has_break_or_continue(loop, parent) {
            var found = false;
            var tw = new TreeWalker(function (node) {
                if (found || node instanceof AST_Scope)
                    return true;
                if (node instanceof AST_LoopControl && tw.loopcontrol_target(node) === loop) {
                    return found = true;
                }
            });
            if (parent instanceof AST_LabeledStatement)
                tw.push(parent);
            tw.push(loop);
            loop.body.walk(tw);
            return found;
        }
        def_optimize(AST_Do, function (self, compressor) {
            if (!compressor.option("loops"))
                return self;
            var cond = self.condition.tail_node().evaluate(compressor);
            if (!(cond instanceof AST_Node)) {
                if (cond)
                    return make_node(AST_For, self, {
                        body: make_node(AST_BlockStatement, self.body, {
                            body: [
                                self.body,
                                make_node(AST_SimpleStatement, self.condition, {
                                    body: self.condition
                                })
                            ]
                        })
                    }).optimize(compressor);
                if (!has_break_or_continue(self, compressor.parent())) {
                    return make_node(AST_BlockStatement, self.body, {
                        body: [
                            self.body,
                            make_node(AST_SimpleStatement, self.condition, {
                                body: self.condition
                            })
                        ]
                    }).optimize(compressor);
                }
            }
            return self;
        });
        function if_break_in_loop(self, compressor) {
            var first = self.body instanceof AST_BlockStatement ? self.body.body[0] : self.body;
            if (compressor.option("dead_code") && is_break(first)) {
                var body = [];
                if (self.init instanceof AST_Statement) {
                    body.push(self.init);
                }
                else if (self.init) {
                    body.push(make_node(AST_SimpleStatement, self.init, {
                        body: self.init
                    }));
                }
                if (self.condition) {
                    body.push(make_node(AST_SimpleStatement, self.condition, {
                        body: self.condition
                    }));
                }
                trim_unreachable_code(compressor, self.body, body);
                return make_node(AST_BlockStatement, self, {
                    body: body
                });
            }
            if (first instanceof AST_If) {
                if (is_break(first.body)) {
                    if (self.condition) {
                        self.condition = make_node(AST_Binary, self.condition, {
                            left: self.condition,
                            operator: "&&",
                            right: first.condition.negate(compressor),
                        });
                    }
                    else {
                        self.condition = first.condition.negate(compressor);
                    }
                    drop_it(first.alternative);
                }
                else if (is_break(first.alternative)) {
                    if (self.condition) {
                        self.condition = make_node(AST_Binary, self.condition, {
                            left: self.condition,
                            operator: "&&",
                            right: first.condition,
                        });
                    }
                    else {
                        self.condition = first.condition;
                    }
                    drop_it(first.body);
                }
            }
            return self;
            function is_break(node) {
                return node instanceof AST_Break
                    && compressor.loopcontrol_target(node) === compressor.self();
            }
            function drop_it(rest) {
                rest = as_statement_array(rest);
                if (self.body instanceof AST_BlockStatement) {
                    self.body = self.body.clone();
                    self.body.body = rest.concat(self.body.body.slice(1));
                    self.body = self.body.transform(compressor);
                }
                else {
                    self.body = make_node(AST_BlockStatement, self.body, {
                        body: rest
                    }).transform(compressor);
                }
                self = if_break_in_loop(self, compressor);
            }
        }
        def_optimize(AST_For, function (self, compressor) {
            if (!compressor.option("loops"))
                return self;
            if (compressor.option("side_effects") && self.init) {
                self.init = self.init.drop_side_effect_free(compressor);
            }
            if (self.condition) {
                var cond = self.condition.evaluate(compressor);
                if (!(cond instanceof AST_Node)) {
                    if (cond)
                        self.condition = null;
                    else if (!compressor.option("dead_code")) {
                        var orig = self.condition;
                        self.condition = make_node_from_constant(cond, self.condition);
                        self.condition = best_of_expression(self.condition.transform(compressor), orig);
                    }
                }
                if (compressor.option("dead_code")) {
                    if (cond instanceof AST_Node)
                        cond = self.condition.tail_node().evaluate(compressor);
                    if (!cond) {
                        var body = [];
                        trim_unreachable_code(compressor, self.body, body);
                        if (self.init instanceof AST_Statement) {
                            body.push(self.init);
                        }
                        else if (self.init) {
                            body.push(make_node(AST_SimpleStatement, self.init, {
                                body: self.init
                            }));
                        }
                        body.push(make_node(AST_SimpleStatement, self.condition, {
                            body: self.condition
                        }));
                        return make_node(AST_BlockStatement, self, { body: body }).optimize(compressor);
                    }
                }
            }
            return if_break_in_loop(self, compressor);
        });
        def_optimize(AST_If, function (self, compressor) {
            if (is_empty(self.alternative))
                self.alternative = null;
            if (!compressor.option("conditionals"))
                return self;
            // if condition can be statically determined, drop
            // one of the blocks.  note, statically determined implies
            // “has no side effects”; also it doesn't work for cases like
            // `x && true`, though it probably should.
            var cond = self.condition.evaluate(compressor);
            if (!compressor.option("dead_code") && !(cond instanceof AST_Node)) {
                var orig = self.condition;
                self.condition = make_node_from_constant(cond, orig);
                self.condition = best_of_expression(self.condition.transform(compressor), orig);
            }
            if (compressor.option("dead_code")) {
                if (cond instanceof AST_Node)
                    cond = self.condition.tail_node().evaluate(compressor);
                if (!cond) {
                    var body = [];
                    trim_unreachable_code(compressor, self.body, body);
                    body.push(make_node(AST_SimpleStatement, self.condition, {
                        body: self.condition
                    }));
                    if (self.alternative)
                        body.push(self.alternative);
                    return make_node(AST_BlockStatement, self, { body: body }).optimize(compressor);
                }
                else if (!(cond instanceof AST_Node)) {
                    var body = [];
                    body.push(make_node(AST_SimpleStatement, self.condition, {
                        body: self.condition
                    }));
                    body.push(self.body);
                    if (self.alternative) {
                        trim_unreachable_code(compressor, self.alternative, body);
                    }
                    return make_node(AST_BlockStatement, self, { body: body }).optimize(compressor);
                }
            }
            var negated = self.condition.negate(compressor);
            var self_condition_length = self.condition.size();
            var negated_length = negated.size();
            var negated_is_best = negated_length < self_condition_length;
            if (self.alternative && negated_is_best) {
                negated_is_best = false; // because we already do the switch here.
                // no need to swap values of self_condition_length and negated_length
                // here because they are only used in an equality comparison later on.
                self.condition = negated;
                var tmp = self.body;
                self.body = self.alternative || make_node(AST_EmptyStatement, self);
                self.alternative = tmp;
            }
            if (is_empty(self.body) && is_empty(self.alternative)) {
                return make_node(AST_SimpleStatement, self.condition, {
                    body: self.condition.clone()
                }).optimize(compressor);
            }
            if (self.body instanceof AST_SimpleStatement
                && self.alternative instanceof AST_SimpleStatement) {
                return make_node(AST_SimpleStatement, self, {
                    body: make_node(AST_Conditional, self, {
                        condition: self.condition,
                        consequent: self.body.body,
                        alternative: self.alternative.body
                    })
                }).optimize(compressor);
            }
            if (is_empty(self.alternative) && self.body instanceof AST_SimpleStatement) {
                if (self_condition_length === negated_length && !negated_is_best
                    && self.condition instanceof AST_Binary && self.condition.operator == "||") {
                    // although the code length of self.condition and negated are the same,
                    // negated does not require additional surrounding parentheses.
                    // see https://github.com/mishoo/UglifyJS2/issues/979
                    negated_is_best = true;
                }
                if (negated_is_best)
                    return make_node(AST_SimpleStatement, self, {
                        body: make_node(AST_Binary, self, {
                            operator: "||",
                            left: negated,
                            right: self.body.body
                        })
                    }).optimize(compressor);
                return make_node(AST_SimpleStatement, self, {
                    body: make_node(AST_Binary, self, {
                        operator: "&&",
                        left: self.condition,
                        right: self.body.body
                    })
                }).optimize(compressor);
            }
            if (self.body instanceof AST_EmptyStatement
                && self.alternative instanceof AST_SimpleStatement) {
                return make_node(AST_SimpleStatement, self, {
                    body: make_node(AST_Binary, self, {
                        operator: "||",
                        left: self.condition,
                        right: self.alternative.body
                    })
                }).optimize(compressor);
            }
            if (self.body instanceof AST_Exit
                && self.alternative instanceof AST_Exit
                && self.body.TYPE == self.alternative.TYPE) {
                return make_node(self.body.CTOR, self, {
                    value: make_node(AST_Conditional, self, {
                        condition: self.condition,
                        consequent: self.body.value || make_node(AST_Undefined, self.body),
                        alternative: self.alternative.value || make_node(AST_Undefined, self.alternative)
                    }).transform(compressor)
                }).optimize(compressor);
            }
            if (self.body instanceof AST_If
                && !self.body.alternative
                && !self.alternative) {
                self = make_node(AST_If, self, {
                    condition: make_node(AST_Binary, self.condition, {
                        operator: "&&",
                        left: self.condition,
                        right: self.body.condition
                    }),
                    body: self.body.body,
                    alternative: null
                });
            }
            if (aborts(self.body)) {
                if (self.alternative) {
                    var alt = self.alternative;
                    self.alternative = null;
                    return make_node(AST_BlockStatement, self, {
                        body: [self, alt]
                    }).optimize(compressor);
                }
            }
            if (aborts(self.alternative)) {
                var body = self.body;
                self.body = self.alternative;
                self.condition = negated_is_best ? negated : self.condition.negate(compressor);
                self.alternative = null;
                return make_node(AST_BlockStatement, self, {
                    body: [self, body]
                }).optimize(compressor);
            }
            return self;
        });
        def_optimize(AST_Switch, function (self, compressor) {
            if (!compressor.option("switches"))
                return self;
            var branch;
            var value = self.expression.evaluate(compressor);
            if (!(value instanceof AST_Node)) {
                var orig = self.expression;
                self.expression = make_node_from_constant(value, orig);
                self.expression = best_of_expression(self.expression.transform(compressor), orig);
            }
            if (!compressor.option("dead_code"))
                return self;
            if (value instanceof AST_Node) {
                value = self.expression.tail_node().evaluate(compressor);
            }
            var decl = [];
            var body = [];
            var default_branch;
            var exact_match;
            for (var i = 0, len = self.body.length; i < len && !exact_match; i++) {
                branch = self.body[i];
                if (branch instanceof AST_Default) {
                    if (!default_branch) {
                        default_branch = branch;
                    }
                    else {
                        eliminate_branch(branch, body[body.length - 1]);
                    }
                }
                else if (!(value instanceof AST_Node)) {
                    var exp = branch.expression.evaluate(compressor);
                    if (!(exp instanceof AST_Node) && exp !== value) {
                        eliminate_branch(branch, body[body.length - 1]);
                        continue;
                    }
                    if (exp instanceof AST_Node)
                        exp = branch.expression.tail_node().evaluate(compressor);
                    if (exp === value) {
                        exact_match = branch;
                        if (default_branch) {
                            var default_index = body.indexOf(default_branch);
                            body.splice(default_index, 1);
                            eliminate_branch(default_branch, body[default_index - 1]);
                            default_branch = null;
                        }
                    }
                }
                if (aborts(branch)) {
                    var prev = body[body.length - 1];
                    if (aborts(prev) && prev.body.length == branch.body.length
                        && make_node(AST_BlockStatement, prev, prev).equivalent_to(make_node(AST_BlockStatement, branch, branch))) {
                        prev.body = [];
                    }
                }
                body.push(branch);
            }
            while (i < len)
                eliminate_branch(self.body[i++], body[body.length - 1]);
            if (body.length > 0) {
                body[0].body = decl.concat(body[0].body);
            }
            self.body = body;
            while (branch = body[body.length - 1]) {
                var stat = branch.body[branch.body.length - 1];
                if (stat instanceof AST_Break && compressor.loopcontrol_target(stat) === self)
                    branch.body.pop();
                if (branch.body.length || branch instanceof AST_Case
                    && (default_branch || branch.expression.has_side_effects(compressor)))
                    break;
                if (body.pop() === default_branch)
                    default_branch = null;
            }
            if (body.length == 0) {
                return make_node(AST_BlockStatement, self, {
                    body: decl.concat(make_node(AST_SimpleStatement, self.expression, {
                        body: self.expression
                    }))
                }).optimize(compressor);
            }
            if (body.length == 1 && (body[0] === exact_match || body[0] === default_branch)) {
                var has_break = false;
                var tw = new TreeWalker(function (node) {
                    if (has_break
                        || node instanceof AST_Lambda
                        || node instanceof AST_SimpleStatement)
                        return true;
                    if (node instanceof AST_Break && tw.loopcontrol_target(node) === self)
                        has_break = true;
                });
                self.walk(tw);
                if (!has_break) {
                    var statements = body[0].body.slice();
                    var exp = body[0].expression;
                    if (exp)
                        statements.unshift(make_node(AST_SimpleStatement, exp, {
                            body: exp
                        }));
                    statements.unshift(make_node(AST_SimpleStatement, self.expression, {
                        body: self.expression
                    }));
                    return make_node(AST_BlockStatement, self, {
                        body: statements
                    }).optimize(compressor);
                }
            }
            return self;
            function eliminate_branch(branch, prev) {
                if (prev && !aborts(prev)) {
                    prev.body = prev.body.concat(branch.body);
                }
                else {
                    trim_unreachable_code(compressor, branch, decl);
                }
            }
        });
        def_optimize(AST_Try, function (self, compressor) {
            tighten_body(self.body, compressor);
            if (self.bcatch && self.bfinally && self.bfinally.body.every(is_empty))
                self.bfinally = null;
            if (compressor.option("dead_code") && self.body.every(is_empty)) {
                var body = [];
                if (self.bcatch) {
                    trim_unreachable_code(compressor, self.bcatch, body);
                }
                if (self.bfinally)
                    body.push(...self.bfinally.body);
                return make_node(AST_BlockStatement, self, {
                    body: body
                }).optimize(compressor);
            }
            return self;
        });
        AST_Definitions.DEFMETHOD("remove_initializers", function () {
            var decls = [];
            this.definitions.forEach(function (def) {
                if (def.name instanceof AST_SymbolDeclaration) {
                    def.value = null;
                    decls.push(def);
                }
                else {
                    walk(def.name, node => {
                        if (node instanceof AST_SymbolDeclaration) {
                            decls.push(make_node(AST_VarDef, def, {
                                name: node,
                                value: null
                            }));
                        }
                    });
                }
            });
            this.definitions = decls;
        });
        AST_Definitions.DEFMETHOD("to_assignments", function (compressor) {
            var reduce_vars = compressor.option("reduce_vars");
            var assignments = [];
            for (const def of this.definitions) {
                if (def.value) {
                    var name = make_node(AST_SymbolRef, def.name, def.name);
                    assignments.push(make_node(AST_Assign, def, {
                        operator: "=",
                        logical: false,
                        left: name,
                        right: def.value
                    }));
                    if (reduce_vars)
                        name.definition().fixed = false;
                }
                else if (def.value) {
                    // Because it's a destructuring, do not turn into an assignment.
                    var varDef = make_node(AST_VarDef, def, {
                        name: def.name,
                        value: def.value
                    });
                    var var_ = make_node(AST_Var, def, {
                        definitions: [varDef]
                    });
                    assignments.push(var_);
                }
                const thedef = def.name.definition();
                thedef.eliminated++;
                thedef.replaced--;
            }
            if (assignments.length == 0)
                return null;
            return make_sequence(this, assignments);
        });
        def_optimize(AST_Definitions, function (self) {
            if (self.definitions.length == 0)
                return make_node(AST_EmptyStatement, self);
            return self;
        });
        def_optimize(AST_VarDef, function (self, compressor) {
            if (self.name instanceof AST_SymbolLet
                && self.value != null
                && is_undefined(self.value, compressor)) {
                self.value = null;
            }
            return self;
        });
        def_optimize(AST_Import, function (self) {
            return self;
        });
        // TODO this only works with AST_Defun, shouldn't it work for other ways of defining functions?
        function retain_top_func(fn, compressor) {
            return compressor.top_retain
                && fn instanceof AST_Defun
                && has_flag(fn, TOP)
                && fn.name
                && compressor.top_retain(fn.name);
        }
        def_optimize(AST_Call, function (self, compressor) {
            var exp = self.expression;
            var fn = exp;
            inline_array_like_spread(self.args);
            var simple_args = self.args.every((arg) => !(arg instanceof AST_Expansion));
            if (compressor.option("reduce_vars")
                && fn instanceof AST_SymbolRef
                && !has_annotation(self, _NOINLINE)) {
                const fixed = fn.fixed_value();
                if (!retain_top_func(fixed, compressor)) {
                    fn = fixed;
                }
            }
            if (self.optional && is_nullish(fn, compressor)) {
                return make_node(AST_Undefined, self);
            }
            var is_func = fn instanceof AST_Lambda;
            if (is_func && fn.pinned())
                return self;
            if (compressor.option("unused")
                && simple_args
                && is_func
                && !fn.uses_arguments) {
                var pos = 0, last = 0;
                for (var i = 0, len = self.args.length; i < len; i++) {
                    if (fn.argnames[i] instanceof AST_Expansion) {
                        if (has_flag(fn.argnames[i].expression, UNUSED))
                            while (i < len) {
                                var node = self.args[i++].drop_side_effect_free(compressor);
                                if (node) {
                                    self.args[pos++] = node;
                                }
                            }
                        else
                            while (i < len) {
                                self.args[pos++] = self.args[i++];
                            }
                        last = pos;
                        break;
                    }
                    var trim = i >= fn.argnames.length;
                    if (trim || has_flag(fn.argnames[i], UNUSED)) {
                        var node = self.args[i].drop_side_effect_free(compressor);
                        if (node) {
                            self.args[pos++] = node;
                        }
                        else if (!trim) {
                            self.args[pos++] = make_node(AST_Number, self.args[i], {
                                value: 0
                            });
                            continue;
                        }
                    }
                    else {
                        self.args[pos++] = self.args[i];
                    }
                    last = pos;
                }
                self.args.length = last;
            }
            if (compressor.option("unsafe")) {
                if (is_undeclared_ref(exp))
                    switch (exp.name) {
                        case "Array":
                            if (self.args.length != 1) {
                                return make_node(AST_Array, self, {
                                    elements: self.args
                                }).optimize(compressor);
                            }
                            else if (self.args[0] instanceof AST_Number && self.args[0].value <= 11) {
                                const elements = [];
                                for (let i = 0; i < self.args[0].value; i++)
                                    elements.push(new AST_Hole);
                                return new AST_Array({ elements });
                            }
                            break;
                        case "Object":
                            if (self.args.length == 0) {
                                return make_node(AST_Object, self, {
                                    properties: []
                                });
                            }
                            break;
                        case "String":
                            if (self.args.length == 0)
                                return make_node(AST_String, self, {
                                    value: ""
                                });
                            if (self.args.length <= 1)
                                return make_node(AST_Binary, self, {
                                    left: self.args[0],
                                    operator: "+",
                                    right: make_node(AST_String, self, { value: "" })
                                }).optimize(compressor);
                            break;
                        case "Number":
                            if (self.args.length == 0)
                                return make_node(AST_Number, self, {
                                    value: 0
                                });
                            if (self.args.length == 1 && compressor.option("unsafe_math")) {
                                return make_node(AST_UnaryPrefix, self, {
                                    expression: self.args[0],
                                    operator: "+"
                                }).optimize(compressor);
                            }
                            break;
                        case "Symbol":
                            if (self.args.length == 1 && self.args[0] instanceof AST_String && compressor.option("unsafe_symbols"))
                                self.args.length = 0;
                            break;
                        case "Boolean":
                            if (self.args.length == 0)
                                return make_node(AST_False, self);
                            if (self.args.length == 1)
                                return make_node(AST_UnaryPrefix, self, {
                                    expression: make_node(AST_UnaryPrefix, self, {
                                        expression: self.args[0],
                                        operator: "!"
                                    }),
                                    operator: "!"
                                }).optimize(compressor);
                            break;
                        case "RegExp":
                            var params = [];
                            if (self.args.length >= 1
                                && self.args.length <= 2
                                && self.args.every((arg) => {
                                    var value = arg.evaluate(compressor);
                                    params.push(value);
                                    return arg !== value;
                                })) {
                                let [source, flags] = params;
                                source = regexp_source_fix(new RegExp(source).source);
                                const rx = make_node(AST_RegExp, self, {
                                    value: { source, flags }
                                });
                                if (rx._eval(compressor) !== rx) {
                                    return rx;
                                }
                            }
                            break;
                    }
                else if (exp instanceof AST_Dot)
                    switch (exp.property) {
                        case "toString":
                            if (self.args.length == 0 && !exp.expression.may_throw_on_access(compressor)) {
                                return make_node(AST_Binary, self, {
                                    left: make_node(AST_String, self, { value: "" }),
                                    operator: "+",
                                    right: exp.expression
                                }).optimize(compressor);
                            }
                            break;
                        case "join":
                            if (exp.expression instanceof AST_Array)
                                EXIT: {
                                    var separator;
                                    if (self.args.length > 0) {
                                        separator = self.args[0].evaluate(compressor);
                                        if (separator === self.args[0])
                                            break EXIT; // not a constant
                                    }
                                    var elements = [];
                                    var consts = [];
                                    for (var i = 0, len = exp.expression.elements.length; i < len; i++) {
                                        var el = exp.expression.elements[i];
                                        if (el instanceof AST_Expansion)
                                            break EXIT;
                                        var value = el.evaluate(compressor);
                                        if (value !== el) {
                                            consts.push(value);
                                        }
                                        else {
                                            if (consts.length > 0) {
                                                elements.push(make_node(AST_String, self, {
                                                    value: consts.join(separator)
                                                }));
                                                consts.length = 0;
                                            }
                                            elements.push(el);
                                        }
                                    }
                                    if (consts.length > 0) {
                                        elements.push(make_node(AST_String, self, {
                                            value: consts.join(separator)
                                        }));
                                    }
                                    if (elements.length == 0)
                                        return make_node(AST_String, self, { value: "" });
                                    if (elements.length == 1) {
                                        if (elements[0].is_string(compressor)) {
                                            return elements[0];
                                        }
                                        return make_node(AST_Binary, elements[0], {
                                            operator: "+",
                                            left: make_node(AST_String, self, { value: "" }),
                                            right: elements[0]
                                        });
                                    }
                                    if (separator == "") {
                                        var first;
                                        if (elements[0].is_string(compressor)
                                            || elements[1].is_string(compressor)) {
                                            first = elements.shift();
                                        }
                                        else {
                                            first = make_node(AST_String, self, { value: "" });
                                        }
                                        return elements.reduce(function (prev, el) {
                                            return make_node(AST_Binary, el, {
                                                operator: "+",
                                                left: prev,
                                                right: el
                                            });
                                        }, first).optimize(compressor);
                                    }
                                    // need this awkward cloning to not affect original element
                                    // best_of will decide which one to get through.
                                    var node = self.clone();
                                    node.expression = node.expression.clone();
                                    node.expression.expression = node.expression.expression.clone();
                                    node.expression.expression.elements = elements;
                                    return best_of(compressor, self, node);
                                }
                            break;
                        case "charAt":
                            if (exp.expression.is_string(compressor)) {
                                var arg = self.args[0];
                                var index = arg ? arg.evaluate(compressor) : 0;
                                if (index !== arg) {
                                    return make_node(AST_Sub, exp, {
                                        expression: exp.expression,
                                        property: make_node_from_constant(index | 0, arg || exp)
                                    }).optimize(compressor);
                                }
                            }
                            break;
                        case "apply":
                            if (self.args.length == 2 && self.args[1] instanceof AST_Array) {
                                var args = self.args[1].elements.slice();
                                args.unshift(self.args[0]);
                                return make_node(AST_Call, self, {
                                    expression: make_node(AST_Dot, exp, {
                                        expression: exp.expression,
                                        optional: false,
                                        property: "call"
                                    }),
                                    args: args
                                }).optimize(compressor);
                            }
                            break;
                        case "call":
                            var func = exp.expression;
                            if (func instanceof AST_SymbolRef) {
                                func = func.fixed_value();
                            }
                            if (func instanceof AST_Lambda && !func.contains_this()) {
                                return (self.args.length ? make_sequence(this, [
                                    self.args[0],
                                    make_node(AST_Call, self, {
                                        expression: exp.expression,
                                        args: self.args.slice(1)
                                    })
                                ]) : make_node(AST_Call, self, {
                                    expression: exp.expression,
                                    args: []
                                })).optimize(compressor);
                            }
                            break;
                    }
            }
            if (compressor.option("unsafe_Function")
                && is_undeclared_ref(exp)
                && exp.name == "Function") {
                // new Function() => function(){}
                if (self.args.length == 0)
                    return make_node(AST_Function, self, {
                        argnames: [],
                        body: []
                    }).optimize(compressor);
                if (self.args.every((x) => x instanceof AST_String)) {
                    // quite a corner-case, but we can handle it:
                    //   https://github.com/mishoo/UglifyJS2/issues/203
                    // if the code argument is a constant, then we can minify it.
                    try {
                        var code = "n(function(" + self.args.slice(0, -1).map(function (arg) {
                            return arg.value;
                        }).join(",") + "){" + self.args[self.args.length - 1].value + "})";
                        var ast = parse(code);
                        var mangle = { ie8: compressor.option("ie8") };
                        ast.figure_out_scope(mangle);
                        var comp = new Compressor(compressor.options, {
                            mangle_options: compressor.mangle_options
                        });
                        ast = ast.transform(comp);
                        ast.figure_out_scope(mangle);
                        base54.reset();
                        ast.compute_char_frequency(mangle);
                        ast.mangle_names(mangle);
                        var fun;
                        walk(ast, node => {
                            if (is_func_expr(node)) {
                                fun = node;
                                return walk_abort;
                            }
                        });
                        var code = OutputStream();
                        AST_BlockStatement.prototype._codegen.call(fun, fun, code);
                        self.args = [
                            make_node(AST_String, self, {
                                value: fun.argnames.map(function (arg) {
                                    return arg.print_to_string();
                                }).join(",")
                            }),
                            make_node(AST_String, self.args[self.args.length - 1], {
                                value: code.get().replace(/^{|}$/g, "")
                            })
                        ];
                        return self;
                    }
                    catch (ex) {
                        if (!(ex instanceof JS_Parse_Error)) {
                            throw ex;
                        }
                        // Otherwise, it crashes at runtime. Or maybe it's nonstandard syntax.
                    }
                }
            }
            var stat = is_func && fn.body[0];
            var is_regular_func = is_func && !fn.is_generator && !fn.async;
            var can_inline = is_regular_func && compressor.option("inline") && !self.is_callee_pure(compressor);
            if (can_inline && stat instanceof AST_Return) {
                let returned = stat.value;
                if (!returned || returned.is_constant_expression()) {
                    if (returned) {
                        returned = returned.clone(true);
                    }
                    else {
                        returned = make_node(AST_Undefined, self);
                    }
                    const args = self.args.concat(returned);
                    return make_sequence(self, args).optimize(compressor);
                }
                // optimize identity function
                if (fn.argnames.length === 1
                    && (fn.argnames[0] instanceof AST_SymbolFunarg)
                    && self.args.length < 2
                    && returned instanceof AST_SymbolRef
                    && returned.name === fn.argnames[0].name) {
                    const replacement = (self.args[0] || make_node(AST_Undefined)).optimize(compressor);
                    let parent;
                    if (replacement instanceof AST_PropAccess
                        && (parent = compressor.parent()) instanceof AST_Call
                        && parent.expression === self) {
                        // identity function was being used to remove `this`, like in
                        //
                        // id(bag.no_this)(...)
                        //
                        // Replace with a larger but more effish (0, bag.no_this) wrapper.
                        return make_sequence(self, [
                            make_node(AST_Number, self, { value: 0 }),
                            replacement
                        ]);
                    }
                    // replace call with first argument or undefined if none passed
                    return replacement;
                }
            }
            if (can_inline) {
                var scope, in_loop, level = -1;
                let def;
                let returned_value;
                let nearest_scope;
                if (simple_args
                    && !fn.uses_arguments
                    && !(compressor.parent() instanceof AST_Class)
                    && !(fn.name && fn instanceof AST_Function)
                    && (returned_value = can_flatten_body(stat))
                    && (exp === fn
                        || has_annotation(self, _INLINE)
                        || compressor.option("unused")
                            && (def = exp.definition()).references.length == 1
                            && !recursive_ref(compressor, def)
                            && fn.is_constant_expression(exp.scope))
                    && !has_annotation(self, _PURE | _NOINLINE)
                    && !fn.contains_this()
                    && can_inject_symbols()
                    && (nearest_scope = find_scope(compressor))
                    && !scope_encloses_variables_in_this_scope(nearest_scope, fn)
                    && !(function in_default_assign() {
                        // Due to the fact function parameters have their own scope
                        // which can't use `var something` in the function body within,
                        // we simply don't inline into DefaultAssign.
                        let i = 0;
                        let p;
                        while ((p = compressor.parent(i++))) {
                            if (p instanceof AST_DefaultAssign)
                                return true;
                            if (p instanceof AST_Block)
                                break;
                        }
                        return false;
                    })()
                    && !(scope instanceof AST_Class)) {
                    set_flag(fn, SQUEEZED);
                    nearest_scope.add_child_scope(fn);
                    return make_sequence(self, flatten_fn(returned_value)).optimize(compressor);
                }
            }
            if (can_inline && has_annotation(self, _INLINE)) {
                set_flag(fn, SQUEEZED);
                fn = make_node(fn.CTOR === AST_Defun ? AST_Function : fn.CTOR, fn, fn);
                fn.figure_out_scope({}, {
                    parent_scope: find_scope(compressor),
                    toplevel: compressor.get_toplevel()
                });
                return make_node(AST_Call, self, {
                    expression: fn,
                    args: self.args,
                }).optimize(compressor);
            }
            const can_drop_this_call = is_regular_func && compressor.option("side_effects") && fn.body.every(is_empty);
            if (can_drop_this_call) {
                var args = self.args.concat(make_node(AST_Undefined, self));
                return make_sequence(self, args).optimize(compressor);
            }
            if (compressor.option("negate_iife")
                && compressor.parent() instanceof AST_SimpleStatement
                && is_iife_call(self)) {
                return self.negate(compressor, true);
            }
            var ev = self.evaluate(compressor);
            if (ev !== self) {
                ev = make_node_from_constant(ev, self).optimize(compressor);
                return best_of(compressor, ev, self);
            }
            return self;
            function return_value(stat) {
                if (!stat)
                    return make_node(AST_Undefined, self);
                if (stat instanceof AST_Return) {
                    if (!stat.value)
                        return make_node(AST_Undefined, self);
                    return stat.value.clone(true);
                }
                if (stat instanceof AST_SimpleStatement) {
                    return make_node(AST_UnaryPrefix, stat, {
                        operator: "void",
                        expression: stat.body.clone(true)
                    });
                }
            }
            function can_flatten_body(stat) {
                var body = fn.body;
                var len = body.length;
                if (compressor.option("inline") < 3) {
                    return len == 1 && return_value(stat);
                }
                stat = null;
                for (var i = 0; i < len; i++) {
                    var line = body[i];
                    if (line instanceof AST_Var) {
                        if (stat && !line.definitions.every((var_def) => !var_def.value)) {
                            return false;
                        }
                    }
                    else if (stat) {
                        return false;
                    }
                    else if (!(line instanceof AST_EmptyStatement)) {
                        stat = line;
                    }
                }
                return return_value(stat);
            }
            function can_inject_args(block_scoped, safe_to_inject) {
                for (var i = 0, len = fn.argnames.length; i < len; i++) {
                    var arg = fn.argnames[i];
                    if (arg instanceof AST_DefaultAssign) {
                        if (has_flag(arg.left, UNUSED))
                            continue;
                        return false;
                    }
                    if (arg instanceof AST_Destructuring)
                        return false;
                    if (arg instanceof AST_Expansion) {
                        if (has_flag(arg.expression, UNUSED))
                            continue;
                        return false;
                    }
                    if (has_flag(arg, UNUSED))
                        continue;
                    if (!safe_to_inject
                        || block_scoped.has(arg.name)
                        || identifier_atom.has(arg.name)
                        || scope.conflicting_def(arg.name)) {
                        return false;
                    }
                    if (in_loop)
                        in_loop.push(arg.definition());
                }
                return true;
            }
            function can_inject_vars(block_scoped, safe_to_inject) {
                var len = fn.body.length;
                for (var i = 0; i < len; i++) {
                    var stat = fn.body[i];
                    if (!(stat instanceof AST_Var))
                        continue;
                    if (!safe_to_inject)
                        return false;
                    for (var j = stat.definitions.length; --j >= 0;) {
                        var name = stat.definitions[j].name;
                        if (name instanceof AST_Destructuring
                            || block_scoped.has(name.name)
                            || identifier_atom.has(name.name)
                            || scope.conflicting_def(name.name)) {
                            return false;
                        }
                        if (in_loop)
                            in_loop.push(name.definition());
                    }
                }
                return true;
            }
            function can_inject_symbols() {
                var block_scoped = new Set();
                do {
                    scope = compressor.parent(++level);
                    if (scope.is_block_scope() && scope.block_scope) {
                        // TODO this is sometimes undefined during compression.
                        // But it should always have a value!
                        scope.block_scope.variables.forEach(function (variable) {
                            block_scoped.add(variable.name);
                        });
                    }
                    if (scope instanceof AST_Catch) {
                        // TODO can we delete? AST_Catch is a block scope.
                        if (scope.argname) {
                            block_scoped.add(scope.argname.name);
                        }
                    }
                    else if (scope instanceof AST_IterationStatement) {
                        in_loop = [];
                    }
                    else if (scope instanceof AST_SymbolRef) {
                        if (scope.fixed_value() instanceof AST_Scope)
                            return false;
                    }
                } while (!(scope instanceof AST_Scope));
                var safe_to_inject = !(scope instanceof AST_Toplevel) || compressor.toplevel.vars;
                var inline = compressor.option("inline");
                if (!can_inject_vars(block_scoped, inline >= 3 && safe_to_inject))
                    return false;
                if (!can_inject_args(block_scoped, inline >= 2 && safe_to_inject))
                    return false;
                return !in_loop || in_loop.length == 0 || !is_reachable(fn, in_loop);
            }
            function append_var(decls, expressions, name, value) {
                var def = name.definition();
                // Name already exists, only when a function argument had the same name
                const already_appended = scope.variables.has(name.name);
                if (!already_appended) {
                    scope.variables.set(name.name, def);
                    scope.enclosed.push(def);
                    decls.push(make_node(AST_VarDef, name, {
                        name: name,
                        value: null
                    }));
                }
                var sym = make_node(AST_SymbolRef, name, name);
                def.references.push(sym);
                if (value)
                    expressions.push(make_node(AST_Assign, self, {
                        operator: "=",
                        logical: false,
                        left: sym,
                        right: value.clone()
                    }));
            }
            function flatten_args(decls, expressions) {
                var len = fn.argnames.length;
                for (var i = self.args.length; --i >= len;) {
                    expressions.push(self.args[i]);
                }
                for (i = len; --i >= 0;) {
                    var name = fn.argnames[i];
                    var value = self.args[i];
                    if (has_flag(name, UNUSED) || !name.name || scope.conflicting_def(name.name)) {
                        if (value)
                            expressions.push(value);
                    }
                    else {
                        var symbol = make_node(AST_SymbolVar, name, name);
                        name.definition().orig.push(symbol);
                        if (!value && in_loop)
                            value = make_node(AST_Undefined, self);
                        append_var(decls, expressions, symbol, value);
                    }
                }
                decls.reverse();
                expressions.reverse();
            }
            function flatten_vars(decls, expressions) {
                var pos = expressions.length;
                for (var i = 0, lines = fn.body.length; i < lines; i++) {
                    var stat = fn.body[i];
                    if (!(stat instanceof AST_Var))
                        continue;
                    for (var j = 0, defs = stat.definitions.length; j < defs; j++) {
                        var var_def = stat.definitions[j];
                        var name = var_def.name;
                        append_var(decls, expressions, name, var_def.value);
                        if (in_loop && fn.argnames.every((argname) => argname.name != name.name)) {
                            var def = fn.variables.get(name.name);
                            var sym = make_node(AST_SymbolRef, name, name);
                            def.references.push(sym);
                            expressions.splice(pos++, 0, make_node(AST_Assign, var_def, {
                                operator: "=",
                                logical: false,
                                left: sym,
                                right: make_node(AST_Undefined, name)
                            }));
                        }
                    }
                }
            }
            function flatten_fn(returned_value) {
                var decls = [];
                var expressions = [];
                flatten_args(decls, expressions);
                flatten_vars(decls, expressions);
                expressions.push(returned_value);
                if (decls.length) {
                    const i = scope.body.indexOf(compressor.parent(level - 1)) + 1;
                    scope.body.splice(i, 0, make_node(AST_Var, fn, {
                        definitions: decls
                    }));
                }
                return expressions.map(exp => exp.clone(true));
            }
        });
        def_optimize(AST_New, function (self, compressor) {
            if (compressor.option("unsafe") &&
                is_undeclared_ref(self.expression) &&
                ["Object", "RegExp", "Function", "Error", "Array"].includes(self.expression.name))
                return make_node(AST_Call, self, self).transform(compressor);
            return self;
        });
        def_optimize(AST_Sequence, function (self, compressor) {
            if (!compressor.option("side_effects"))
                return self;
            var expressions = [];
            filter_for_side_effects();
            var end = expressions.length - 1;
            trim_right_for_undefined();
            if (end == 0) {
                self = maintain_this_binding(compressor.parent(), compressor.self(), expressions[0]);
                if (!(self instanceof AST_Sequence))
                    self = self.optimize(compressor);
                return self;
            }
            self.expressions = expressions;
            return self;
            function filter_for_side_effects() {
                var first = first_in_statement(compressor);
                var last = self.expressions.length - 1;
                self.expressions.forEach(function (expr, index) {
                    if (index < last)
                        expr = expr.drop_side_effect_free(compressor, first);
                    if (expr) {
                        merge_sequence(expressions, expr);
                        first = false;
                    }
                });
            }
            function trim_right_for_undefined() {
                while (end > 0 && is_undefined(expressions[end], compressor))
                    end--;
                if (end < expressions.length - 1) {
                    expressions[end] = make_node(AST_UnaryPrefix, self, {
                        operator: "void",
                        expression: expressions[end]
                    });
                    expressions.length = end + 1;
                }
            }
        });
        AST_Unary.DEFMETHOD("lift_sequences", function (compressor) {
            if (compressor.option("sequences")) {
                if (this.expression instanceof AST_Sequence) {
                    var x = this.expression.expressions.slice();
                    var e = this.clone();
                    e.expression = x.pop();
                    x.push(e);
                    return make_sequence(this, x).optimize(compressor);
                }
            }
            return this;
        });
        def_optimize(AST_UnaryPostfix, function (self, compressor) {
            return self.lift_sequences(compressor);
        });
        def_optimize(AST_UnaryPrefix, function (self, compressor) {
            var e = self.expression;
            if (self.operator == "delete"
                && !(e instanceof AST_SymbolRef
                    || e instanceof AST_PropAccess
                    || is_identifier_atom(e))) {
                if (e instanceof AST_Sequence) {
                    const exprs = e.expressions.slice();
                    exprs.push(make_node(AST_True, self));
                    return make_sequence(self, exprs).optimize(compressor);
                }
                return make_sequence(self, [e, make_node(AST_True, self)]).optimize(compressor);
            }
            var seq = self.lift_sequences(compressor);
            if (seq !== self) {
                return seq;
            }
            if (compressor.option("side_effects") && self.operator == "void") {
                e = e.drop_side_effect_free(compressor);
                if (e) {
                    self.expression = e;
                    return self;
                }
                else {
                    return make_node(AST_Undefined, self).optimize(compressor);
                }
            }
            if (compressor.in_boolean_context()) {
                switch (self.operator) {
                    case "!":
                        if (e instanceof AST_UnaryPrefix && e.operator == "!") {
                            // !!foo ==> foo, if we're in boolean context
                            return e.expression;
                        }
                        if (e instanceof AST_Binary) {
                            self = best_of(compressor, self, e.negate(compressor, first_in_statement(compressor)));
                        }
                        break;
                    case "typeof":
                        // typeof always returns a non-empty string, thus it's
                        // always true in booleans
                        // And we don't need to check if it's undeclared, because in typeof, that's OK
                        return (e instanceof AST_SymbolRef ? make_node(AST_True, self) : make_sequence(self, [
                            e,
                            make_node(AST_True, self)
                        ])).optimize(compressor);
                }
            }
            if (self.operator == "-" && e instanceof AST_Infinity) {
                e = e.transform(compressor);
            }
            if (e instanceof AST_Binary
                && (self.operator == "+" || self.operator == "-")
                && (e.operator == "*" || e.operator == "/" || e.operator == "%")) {
                return make_node(AST_Binary, self, {
                    operator: e.operator,
                    left: make_node(AST_UnaryPrefix, e.left, {
                        operator: self.operator,
                        expression: e.left
                    }),
                    right: e.right
                });
            }
            // avoids infinite recursion of numerals
            if (self.operator != "-"
                || !(e instanceof AST_Number || e instanceof AST_Infinity || e instanceof AST_BigInt)) {
                var ev = self.evaluate(compressor);
                if (ev !== self) {
                    ev = make_node_from_constant(ev, self).optimize(compressor);
                    return best_of(compressor, ev, self);
                }
            }
            return self;
        });
        AST_Binary.DEFMETHOD("lift_sequences", function (compressor) {
            if (compressor.option("sequences")) {
                if (this.left instanceof AST_Sequence) {
                    var x = this.left.expressions.slice();
                    var e = this.clone();
                    e.left = x.pop();
                    x.push(e);
                    return make_sequence(this, x).optimize(compressor);
                }
                if (this.right instanceof AST_Sequence && !this.left.has_side_effects(compressor)) {
                    var assign = this.operator == "=" && this.left instanceof AST_SymbolRef;
                    var x = this.right.expressions;
                    var last = x.length - 1;
                    for (var i = 0; i < last; i++) {
                        if (!assign && x[i].has_side_effects(compressor))
                            break;
                    }
                    if (i == last) {
                        x = x.slice();
                        var e = this.clone();
                        e.right = x.pop();
                        x.push(e);
                        return make_sequence(this, x).optimize(compressor);
                    }
                    else if (i > 0) {
                        var e = this.clone();
                        e.right = make_sequence(this.right, x.slice(i));
                        x = x.slice(0, i);
                        x.push(e);
                        return make_sequence(this, x).optimize(compressor);
                    }
                }
            }
            return this;
        });
        var commutativeOperators = makePredicate("== === != !== * & | ^");
        function is_object(node) {
            return node instanceof AST_Array
                || node instanceof AST_Lambda
                || node instanceof AST_Object
                || node instanceof AST_Class;
        }
        def_optimize(AST_Binary, function (self, compressor) {
            function reversible() {
                return self.left.is_constant()
                    || self.right.is_constant()
                    || !self.left.has_side_effects(compressor)
                        && !self.right.has_side_effects(compressor);
            }
            function reverse(op) {
                if (reversible()) {
                    if (op)
                        self.operator = op;
                    var tmp = self.left;
                    self.left = self.right;
                    self.right = tmp;
                }
            }
            if (commutativeOperators.has(self.operator)) {
                if (self.right.is_constant()
                    && !self.left.is_constant()) {
                    // if right is a constant, whatever side effects the
                    // left side might have could not influence the
                    // result.  hence, force switch.
                    if (!(self.left instanceof AST_Binary
                        && PRECEDENCE[self.left.operator] >= PRECEDENCE[self.operator])) {
                        reverse();
                    }
                }
            }
            self = self.lift_sequences(compressor);
            if (compressor.option("comparisons"))
                switch (self.operator) {
                    case "===":
                    case "!==":
                        var is_strict_comparison = true;
                        if ((self.left.is_string(compressor) && self.right.is_string(compressor)) ||
                            (self.left.is_number(compressor) && self.right.is_number(compressor)) ||
                            (self.left.is_boolean() && self.right.is_boolean()) ||
                            self.left.equivalent_to(self.right)) {
                            self.operator = self.operator.substr(0, 2);
                        }
                    // XXX: intentionally falling down to the next case
                    case "==":
                    case "!=":
                        // void 0 == x => null == x
                        if (!is_strict_comparison && is_undefined(self.left, compressor)) {
                            self.left = make_node(AST_Null, self.left);
                        }
                        else if (compressor.option("typeofs")
                            // "undefined" == typeof x => undefined === x
                            && self.left instanceof AST_String
                            && self.left.value == "undefined"
                            && self.right instanceof AST_UnaryPrefix
                            && self.right.operator == "typeof") {
                            var expr = self.right.expression;
                            if (expr instanceof AST_SymbolRef ? expr.is_declared(compressor)
                                : !(expr instanceof AST_PropAccess && compressor.option("ie8"))) {
                                self.right = expr;
                                self.left = make_node(AST_Undefined, self.left).optimize(compressor);
                                if (self.operator.length == 2)
                                    self.operator += "=";
                            }
                        }
                        else if (self.left instanceof AST_SymbolRef
                            // obj !== obj => false
                            && self.right instanceof AST_SymbolRef
                            && self.left.definition() === self.right.definition()
                            && is_object(self.left.fixed_value())) {
                            return make_node(self.operator[0] == "=" ? AST_True : AST_False, self);
                        }
                        break;
                    case "&&":
                    case "||":
                        var lhs = self.left;
                        if (lhs.operator == self.operator) {
                            lhs = lhs.right;
                        }
                        if (lhs instanceof AST_Binary
                            && lhs.operator == (self.operator == "&&" ? "!==" : "===")
                            && self.right instanceof AST_Binary
                            && lhs.operator == self.right.operator
                            && (is_undefined(lhs.left, compressor) && self.right.left instanceof AST_Null
                                || lhs.left instanceof AST_Null && is_undefined(self.right.left, compressor))
                            && !lhs.right.has_side_effects(compressor)
                            && lhs.right.equivalent_to(self.right.right)) {
                            var combined = make_node(AST_Binary, self, {
                                operator: lhs.operator.slice(0, -1),
                                left: make_node(AST_Null, self),
                                right: lhs.right
                            });
                            if (lhs !== self.left) {
                                combined = make_node(AST_Binary, self, {
                                    operator: self.operator,
                                    left: self.left.left,
                                    right: combined
                                });
                            }
                            return combined;
                        }
                        break;
                }
            if (self.operator == "+" && compressor.in_boolean_context()) {
                var ll = self.left.evaluate(compressor);
                var rr = self.right.evaluate(compressor);
                if (ll && typeof ll == "string") {
                    return make_sequence(self, [
                        self.right,
                        make_node(AST_True, self)
                    ]).optimize(compressor);
                }
                if (rr && typeof rr == "string") {
                    return make_sequence(self, [
                        self.left,
                        make_node(AST_True, self)
                    ]).optimize(compressor);
                }
            }
            if (compressor.option("comparisons") && self.is_boolean()) {
                if (!(compressor.parent() instanceof AST_Binary)
                    || compressor.parent() instanceof AST_Assign) {
                    var negated = make_node(AST_UnaryPrefix, self, {
                        operator: "!",
                        expression: self.negate(compressor, first_in_statement(compressor))
                    });
                    self = best_of(compressor, self, negated);
                }
                if (compressor.option("unsafe_comps")) {
                    switch (self.operator) {
                        case "<":
                            reverse(">");
                            break;
                        case "<=":
                            reverse(">=");
                            break;
                    }
                }
            }
            if (self.operator == "+") {
                if (self.right instanceof AST_String
                    && self.right.getValue() == ""
                    && self.left.is_string(compressor)) {
                    return self.left;
                }
                if (self.left instanceof AST_String
                    && self.left.getValue() == ""
                    && self.right.is_string(compressor)) {
                    return self.right;
                }
                if (self.left instanceof AST_Binary
                    && self.left.operator == "+"
                    && self.left.left instanceof AST_String
                    && self.left.left.getValue() == ""
                    && self.right.is_string(compressor)) {
                    self.left = self.left.right;
                    return self;
                }
            }
            if (compressor.option("evaluate")) {
                switch (self.operator) {
                    case "&&":
                        var ll = has_flag(self.left, TRUTHY)
                            ? true
                            : has_flag(self.left, FALSY)
                                ? false
                                : self.left.evaluate(compressor);
                        if (!ll) {
                            return maintain_this_binding(compressor.parent(), compressor.self(), self.left).optimize(compressor);
                        }
                        else if (!(ll instanceof AST_Node)) {
                            return make_sequence(self, [self.left, self.right]).optimize(compressor);
                        }
                        var rr = self.right.evaluate(compressor);
                        if (!rr) {
                            if (compressor.in_boolean_context()) {
                                return make_sequence(self, [
                                    self.left,
                                    make_node(AST_False, self)
                                ]).optimize(compressor);
                            }
                            else {
                                set_flag(self, FALSY);
                            }
                        }
                        else if (!(rr instanceof AST_Node)) {
                            var parent = compressor.parent();
                            if (parent.operator == "&&" && parent.left === compressor.self() || compressor.in_boolean_context()) {
                                return self.left.optimize(compressor);
                            }
                        }
                        // x || false && y ---> x ? y : false
                        if (self.left.operator == "||") {
                            var lr = self.left.right.evaluate(compressor);
                            if (!lr)
                                return make_node(AST_Conditional, self, {
                                    condition: self.left.left,
                                    consequent: self.right,
                                    alternative: self.left.right
                                }).optimize(compressor);
                        }
                        break;
                    case "||":
                        var ll = has_flag(self.left, TRUTHY)
                            ? true
                            : has_flag(self.left, FALSY)
                                ? false
                                : self.left.evaluate(compressor);
                        if (!ll) {
                            return make_sequence(self, [self.left, self.right]).optimize(compressor);
                        }
                        else if (!(ll instanceof AST_Node)) {
                            return maintain_this_binding(compressor.parent(), compressor.self(), self.left).optimize(compressor);
                        }
                        var rr = self.right.evaluate(compressor);
                        if (!rr) {
                            var parent = compressor.parent();
                            if (parent.operator == "||" && parent.left === compressor.self() || compressor.in_boolean_context()) {
                                return self.left.optimize(compressor);
                            }
                        }
                        else if (!(rr instanceof AST_Node)) {
                            if (compressor.in_boolean_context()) {
                                return make_sequence(self, [
                                    self.left,
                                    make_node(AST_True, self)
                                ]).optimize(compressor);
                            }
                            else {
                                set_flag(self, TRUTHY);
                            }
                        }
                        if (self.left.operator == "&&") {
                            var lr = self.left.right.evaluate(compressor);
                            if (lr && !(lr instanceof AST_Node))
                                return make_node(AST_Conditional, self, {
                                    condition: self.left.left,
                                    consequent: self.left.right,
                                    alternative: self.right
                                }).optimize(compressor);
                        }
                        break;
                    case "??":
                        if (is_nullish(self.left, compressor)) {
                            return self.right;
                        }
                        var ll = self.left.evaluate(compressor);
                        if (!(ll instanceof AST_Node)) {
                            // if we know the value for sure we can simply compute right away.
                            return ll == null ? self.right : self.left;
                        }
                        if (compressor.in_boolean_context()) {
                            const rr = self.right.evaluate(compressor);
                            if (!(rr instanceof AST_Node) && !rr) {
                                return self.left;
                            }
                        }
                }
                var associative = true;
                switch (self.operator) {
                    case "+":
                        // (x + "foo") + "bar" => x + "foobar"
                        if (self.right instanceof AST_Constant
                            && self.left instanceof AST_Binary
                            && self.left.operator == "+"
                            && self.left.is_string(compressor)) {
                            var binary = make_node(AST_Binary, self, {
                                operator: "+",
                                left: self.left.right,
                                right: self.right,
                            });
                            var r = binary.optimize(compressor);
                            if (binary !== r) {
                                self = make_node(AST_Binary, self, {
                                    operator: "+",
                                    left: self.left.left,
                                    right: r
                                });
                            }
                        }
                        // (x + "foo") + ("bar" + y) => (x + "foobar") + y
                        if (self.left instanceof AST_Binary
                            && self.left.operator == "+"
                            && self.left.is_string(compressor)
                            && self.right instanceof AST_Binary
                            && self.right.operator == "+"
                            && self.right.is_string(compressor)) {
                            var binary = make_node(AST_Binary, self, {
                                operator: "+",
                                left: self.left.right,
                                right: self.right.left,
                            });
                            var m = binary.optimize(compressor);
                            if (binary !== m) {
                                self = make_node(AST_Binary, self, {
                                    operator: "+",
                                    left: make_node(AST_Binary, self.left, {
                                        operator: "+",
                                        left: self.left.left,
                                        right: m
                                    }),
                                    right: self.right.right
                                });
                            }
                        }
                        // a + -b => a - b
                        if (self.right instanceof AST_UnaryPrefix
                            && self.right.operator == "-"
                            && self.left.is_number(compressor)) {
                            self = make_node(AST_Binary, self, {
                                operator: "-",
                                left: self.left,
                                right: self.right.expression
                            });
                            break;
                        }
                        // -a + b => b - a
                        if (self.left instanceof AST_UnaryPrefix
                            && self.left.operator == "-"
                            && reversible()
                            && self.right.is_number(compressor)) {
                            self = make_node(AST_Binary, self, {
                                operator: "-",
                                left: self.right,
                                right: self.left.expression
                            });
                            break;
                        }
                        // `foo${bar}baz` + 1 => `foo${bar}baz1`
                        if (self.left instanceof AST_TemplateString) {
                            var l = self.left;
                            var r = self.right.evaluate(compressor);
                            if (r != self.right) {
                                l.segments[l.segments.length - 1].value += String(r);
                                return l;
                            }
                        }
                        // 1 + `foo${bar}baz` => `1foo${bar}baz`
                        if (self.right instanceof AST_TemplateString) {
                            var r = self.right;
                            var l = self.left.evaluate(compressor);
                            if (l != self.left) {
                                r.segments[0].value = String(l) + r.segments[0].value;
                                return r;
                            }
                        }
                        // `1${bar}2` + `foo${bar}baz` => `1${bar}2foo${bar}baz`
                        if (self.left instanceof AST_TemplateString
                            && self.right instanceof AST_TemplateString) {
                            var l = self.left;
                            var segments = l.segments;
                            var r = self.right;
                            segments[segments.length - 1].value += r.segments[0].value;
                            for (var i = 1; i < r.segments.length; i++) {
                                segments.push(r.segments[i]);
                            }
                            return l;
                        }
                    case "*":
                        associative = compressor.option("unsafe_math");
                    case "&":
                    case "|":
                    case "^":
                        // a + +b => +b + a
                        if (self.left.is_number(compressor)
                            && self.right.is_number(compressor)
                            && reversible()
                            && !(self.left instanceof AST_Binary
                                && self.left.operator != self.operator
                                && PRECEDENCE[self.left.operator] >= PRECEDENCE[self.operator])) {
                            var reversed = make_node(AST_Binary, self, {
                                operator: self.operator,
                                left: self.right,
                                right: self.left
                            });
                            if (self.right instanceof AST_Constant
                                && !(self.left instanceof AST_Constant)) {
                                self = best_of(compressor, reversed, self);
                            }
                            else {
                                self = best_of(compressor, self, reversed);
                            }
                        }
                        if (associative && self.is_number(compressor)) {
                            // a + (b + c) => (a + b) + c
                            if (self.right instanceof AST_Binary
                                && self.right.operator == self.operator) {
                                self = make_node(AST_Binary, self, {
                                    operator: self.operator,
                                    left: make_node(AST_Binary, self.left, {
                                        operator: self.operator,
                                        left: self.left,
                                        right: self.right.left,
                                        start: self.left.start,
                                        end: self.right.left.end
                                    }),
                                    right: self.right.right
                                });
                            }
                            // (n + 2) + 3 => 5 + n
                            // (2 * n) * 3 => 6 + n
                            if (self.right instanceof AST_Constant
                                && self.left instanceof AST_Binary
                                && self.left.operator == self.operator) {
                                if (self.left.left instanceof AST_Constant) {
                                    self = make_node(AST_Binary, self, {
                                        operator: self.operator,
                                        left: make_node(AST_Binary, self.left, {
                                            operator: self.operator,
                                            left: self.left.left,
                                            right: self.right,
                                            start: self.left.left.start,
                                            end: self.right.end
                                        }),
                                        right: self.left.right
                                    });
                                }
                                else if (self.left.right instanceof AST_Constant) {
                                    self = make_node(AST_Binary, self, {
                                        operator: self.operator,
                                        left: make_node(AST_Binary, self.left, {
                                            operator: self.operator,
                                            left: self.left.right,
                                            right: self.right,
                                            start: self.left.right.start,
                                            end: self.right.end
                                        }),
                                        right: self.left.left
                                    });
                                }
                            }
                            // (a | 1) | (2 | d) => (3 | a) | b
                            if (self.left instanceof AST_Binary
                                && self.left.operator == self.operator
                                && self.left.right instanceof AST_Constant
                                && self.right instanceof AST_Binary
                                && self.right.operator == self.operator
                                && self.right.left instanceof AST_Constant) {
                                self = make_node(AST_Binary, self, {
                                    operator: self.operator,
                                    left: make_node(AST_Binary, self.left, {
                                        operator: self.operator,
                                        left: make_node(AST_Binary, self.left.left, {
                                            operator: self.operator,
                                            left: self.left.right,
                                            right: self.right.left,
                                            start: self.left.right.start,
                                            end: self.right.left.end
                                        }),
                                        right: self.left.left
                                    }),
                                    right: self.right.right
                                });
                            }
                        }
                }
            }
            // x && (y && z)  ==>  x && y && z
            // x || (y || z)  ==>  x || y || z
            // x + ("y" + z)  ==>  x + "y" + z
            // "x" + (y + "z")==>  "x" + y + "z"
            if (self.right instanceof AST_Binary
                && self.right.operator == self.operator
                && (lazy_op.has(self.operator)
                    || (self.operator == "+"
                        && (self.right.left.is_string(compressor)
                            || (self.left.is_string(compressor)
                                && self.right.right.is_string(compressor)))))) {
                self.left = make_node(AST_Binary, self.left, {
                    operator: self.operator,
                    left: self.left.transform(compressor),
                    right: self.right.left.transform(compressor)
                });
                self.right = self.right.right.transform(compressor);
                return self.transform(compressor);
            }
            var ev = self.evaluate(compressor);
            if (ev !== self) {
                ev = make_node_from_constant(ev, self).optimize(compressor);
                return best_of(compressor, ev, self);
            }
            return self;
        });
        def_optimize(AST_SymbolExport, function (self) {
            return self;
        });
        function recursive_ref(compressor, def) {
            var node;
            for (var i = 0; node = compressor.parent(i); i++) {
                if (node instanceof AST_Lambda
                    || node instanceof AST_Class) {
                    var name = node.name;
                    if (name && name.definition() === def)
                        break;
                }
            }
            return node;
        }
        function within_array_or_object_literal(compressor) {
            var node, level = 0;
            while (node = compressor.parent(level++)) {
                if (node instanceof AST_Statement)
                    return false;
                if (node instanceof AST_Array
                    || node instanceof AST_ObjectKeyVal
                    || node instanceof AST_Object) {
                    return true;
                }
            }
            return false;
        }
        def_optimize(AST_SymbolRef, function (self, compressor) {
            if (!compressor.option("ie8")
                && is_undeclared_ref(self)
                && !compressor.find_parent(AST_With)) {
                switch (self.name) {
                    case "undefined":
                        return make_node(AST_Undefined, self).optimize(compressor);
                    case "NaN":
                        return make_node(AST_NaN, self).optimize(compressor);
                    case "Infinity":
                        return make_node(AST_Infinity, self).optimize(compressor);
                }
            }
            const parent = compressor.parent();
            if (compressor.option("reduce_vars") && is_lhs(self, parent) !== self) {
                const def = self.definition();
                const nearest_scope = find_scope(compressor);
                if (compressor.top_retain && def.global && compressor.top_retain(def)) {
                    def.fixed = false;
                    def.single_use = false;
                    return self;
                }
                let fixed = self.fixed_value();
                let single_use = def.single_use
                    && !(parent instanceof AST_Call
                        && (parent.is_callee_pure(compressor))
                        || has_annotation(parent, _NOINLINE))
                    && !(parent instanceof AST_Export
                        && fixed instanceof AST_Lambda
                        && fixed.name);
                if (single_use && fixed instanceof AST_Node) {
                    single_use =
                        !fixed.has_side_effects(compressor)
                            && !fixed.may_throw(compressor);
                }
                if (single_use && (fixed instanceof AST_Lambda || fixed instanceof AST_Class)) {
                    if (retain_top_func(fixed, compressor)) {
                        single_use = false;
                    }
                    else if (def.scope !== self.scope
                        && (def.escaped == 1
                            || has_flag(fixed, INLINED)
                            || within_array_or_object_literal(compressor)
                            || !compressor.option("reduce_funcs"))) {
                        single_use = false;
                    }
                    else if (recursive_ref(compressor, def)) {
                        single_use = false;
                    }
                    else if (def.scope !== self.scope || def.orig[0] instanceof AST_SymbolFunarg) {
                        single_use = fixed.is_constant_expression(self.scope);
                        if (single_use == "f") {
                            var scope = self.scope;
                            do {
                                if (scope instanceof AST_Defun || is_func_expr(scope)) {
                                    set_flag(scope, INLINED);
                                }
                            } while (scope = scope.parent_scope);
                        }
                    }
                }
                if (single_use && fixed instanceof AST_Lambda) {
                    single_use =
                        def.scope === self.scope
                            && !scope_encloses_variables_in_this_scope(nearest_scope, fixed)
                            || parent instanceof AST_Call
                                && parent.expression === self
                                && !scope_encloses_variables_in_this_scope(nearest_scope, fixed)
                                && !(fixed.name && fixed.name.definition().recursive_refs > 0);
                }
                if (single_use && fixed) {
                    if (fixed instanceof AST_DefClass) {
                        set_flag(fixed, SQUEEZED);
                        fixed = make_node(AST_ClassExpression, fixed, fixed);
                    }
                    if (fixed instanceof AST_Defun) {
                        set_flag(fixed, SQUEEZED);
                        fixed = make_node(AST_Function, fixed, fixed);
                    }
                    if (def.recursive_refs > 0 && fixed.name instanceof AST_SymbolDefun) {
                        const defun_def = fixed.name.definition();
                        let lambda_def = fixed.variables.get(fixed.name.name);
                        let name = lambda_def && lambda_def.orig[0];
                        if (!(name instanceof AST_SymbolLambda)) {
                            name = make_node(AST_SymbolLambda, fixed.name, fixed.name);
                            name.scope = fixed;
                            fixed.name = name;
                            lambda_def = fixed.def_function(name);
                        }
                        walk(fixed, node => {
                            if (node instanceof AST_SymbolRef && node.definition() === defun_def) {
                                node.thedef = lambda_def;
                                lambda_def.references.push(node);
                            }
                        });
                    }
                    if ((fixed instanceof AST_Lambda || fixed instanceof AST_Class)
                        && fixed.parent_scope !== nearest_scope) {
                        fixed = fixed.clone(true, compressor.get_toplevel());
                        nearest_scope.add_child_scope(fixed);
                    }
                    return fixed.optimize(compressor);
                }
                // multiple uses
                if (fixed) {
                    let replace;
                    if (fixed instanceof AST_This) {
                        if (!(def.orig[0] instanceof AST_SymbolFunarg)
                            && def.references.every((ref) => def.scope === ref.scope)) {
                            replace = fixed;
                        }
                    }
                    else {
                        var ev = fixed.evaluate(compressor);
                        if (ev !== fixed
                            && (compressor.option("unsafe_regexp") || !(ev instanceof RegExp))) {
                            replace = make_node_from_constant(ev, fixed);
                        }
                    }
                    if (replace) {
                        const name_length = self.size(compressor);
                        const replace_size = replace.size(compressor);
                        let overhead = 0;
                        if (compressor.option("unused") && !compressor.exposed(def)) {
                            overhead =
                                (name_length + 2 + replace_size) /
                                    (def.references.length - def.assignments);
                        }
                        if (replace_size <= name_length + overhead) {
                            return replace;
                        }
                    }
                }
            }
            return self;
        });
        function scope_encloses_variables_in_this_scope(scope, pulled_scope) {
            for (const enclosed of pulled_scope.enclosed) {
                if (pulled_scope.variables.has(enclosed.name)) {
                    continue;
                }
                const looked_up = scope.find_variable(enclosed.name);
                if (looked_up) {
                    if (looked_up === enclosed)
                        continue;
                    return true;
                }
            }
            return false;
        }
        function is_atomic(lhs, self) {
            return lhs instanceof AST_SymbolRef || lhs.TYPE === self.TYPE;
        }
        def_optimize(AST_Undefined, function (self, compressor) {
            if (compressor.option("unsafe_undefined")) {
                var undef = find_variable(compressor, "undefined");
                if (undef) {
                    var ref = make_node(AST_SymbolRef, self, {
                        name: "undefined",
                        scope: undef.scope,
                        thedef: undef
                    });
                    set_flag(ref, UNDEFINED);
                    return ref;
                }
            }
            var lhs = is_lhs(compressor.self(), compressor.parent());
            if (lhs && is_atomic(lhs, self))
                return self;
            return make_node(AST_UnaryPrefix, self, {
                operator: "void",
                expression: make_node(AST_Number, self, {
                    value: 0
                })
            });
        });
        def_optimize(AST_Infinity, function (self, compressor) {
            var lhs = is_lhs(compressor.self(), compressor.parent());
            if (lhs && is_atomic(lhs, self))
                return self;
            if (compressor.option("keep_infinity")
                && !(lhs && !is_atomic(lhs, self))
                && !find_variable(compressor, "Infinity")) {
                return self;
            }
            return make_node(AST_Binary, self, {
                operator: "/",
                left: make_node(AST_Number, self, {
                    value: 1
                }),
                right: make_node(AST_Number, self, {
                    value: 0
                })
            });
        });
        def_optimize(AST_NaN, function (self, compressor) {
            var lhs = is_lhs(compressor.self(), compressor.parent());
            if (lhs && !is_atomic(lhs, self)
                || find_variable(compressor, "NaN")) {
                return make_node(AST_Binary, self, {
                    operator: "/",
                    left: make_node(AST_Number, self, {
                        value: 0
                    }),
                    right: make_node(AST_Number, self, {
                        value: 0
                    })
                });
            }
            return self;
        });
        function is_reachable(self, defs) {
            const find_ref = node => {
                if (node instanceof AST_SymbolRef && member(node.definition(), defs)) {
                    return walk_abort;
                }
            };
            return walk_parent(self, (node, info) => {
                if (node instanceof AST_Scope && node !== self) {
                    var parent = info.parent();
                    if (parent instanceof AST_Call && parent.expression === node)
                        return;
                    if (walk(node, find_ref))
                        return walk_abort;
                    return true;
                }
            });
        }
        const ASSIGN_OPS = makePredicate("+ - / * % >> << >>> | ^ &");
        const ASSIGN_OPS_COMMUTATIVE = makePredicate("* | ^ &");
        def_optimize(AST_Assign, function (self, compressor) {
            if (self.logical) {
                return self.lift_sequences(compressor);
            }
            var def;
            if (compressor.option("dead_code")
                && self.left instanceof AST_SymbolRef
                && (def = self.left.definition()).scope === compressor.find_parent(AST_Lambda)) {
                var level = 0, node, parent = self;
                do {
                    node = parent;
                    parent = compressor.parent(level++);
                    if (parent instanceof AST_Exit) {
                        if (in_try(level, parent))
                            break;
                        if (is_reachable(def.scope, [def]))
                            break;
                        if (self.operator == "=")
                            return self.right;
                        def.fixed = false;
                        return make_node(AST_Binary, self, {
                            operator: self.operator.slice(0, -1),
                            left: self.left,
                            right: self.right
                        }).optimize(compressor);
                    }
                } while (parent instanceof AST_Binary && parent.right === node
                    || parent instanceof AST_Sequence && parent.tail_node() === node);
            }
            self = self.lift_sequences(compressor);
            if (self.operator == "=" && self.left instanceof AST_SymbolRef && self.right instanceof AST_Binary) {
                // x = expr1 OP expr2
                if (self.right.left instanceof AST_SymbolRef
                    && self.right.left.name == self.left.name
                    && ASSIGN_OPS.has(self.right.operator)) {
                    // x = x - 2  --->  x -= 2
                    self.operator = self.right.operator + "=";
                    self.right = self.right.right;
                }
                else if (self.right.right instanceof AST_SymbolRef
                    && self.right.right.name == self.left.name
                    && ASSIGN_OPS_COMMUTATIVE.has(self.right.operator)
                    && !self.right.left.has_side_effects(compressor)) {
                    // x = 2 & x  --->  x &= 2
                    self.operator = self.right.operator + "=";
                    self.right = self.right.left;
                }
            }
            return self;
            function in_try(level, node) {
                var right = self.right;
                self.right = make_node(AST_Null, right);
                var may_throw = node.may_throw(compressor);
                self.right = right;
                var scope = self.left.definition().scope;
                var parent;
                while ((parent = compressor.parent(level++)) !== scope) {
                    if (parent instanceof AST_Try) {
                        if (parent.bfinally)
                            return true;
                        if (may_throw && parent.bcatch)
                            return true;
                    }
                }
            }
        });
        def_optimize(AST_DefaultAssign, function (self, compressor) {
            if (!compressor.option("evaluate")) {
                return self;
            }
            var evaluateRight = self.right.evaluate(compressor);
            // `[x = undefined] = foo` ---> `[x] = foo`
            if (evaluateRight === undefined) {
                self = self.left;
            }
            else if (evaluateRight !== self.right) {
                evaluateRight = make_node_from_constant(evaluateRight, self.right);
                self.right = best_of_expression(evaluateRight, self.right);
            }
            return self;
        });
        function is_nullish(node, compressor) {
            let fixed;
            return (node instanceof AST_Null
                || is_undefined(node, compressor)
                || (node instanceof AST_SymbolRef
                    && (fixed = node.definition().fixed) instanceof AST_Node
                    && is_nullish(fixed, compressor))
                // Recurse into those optional chains!
                || node instanceof AST_PropAccess && node.optional && is_nullish(node.expression, compressor)
                || node instanceof AST_Call && node.optional && is_nullish(node.expression, compressor)
                || node instanceof AST_Chain && is_nullish(node.expression, compressor));
        }
        function is_nullish_check(check, check_subject, compressor) {
            if (check_subject.may_throw(compressor))
                return false;
            let nullish_side;
            // foo == null
            if (check instanceof AST_Binary
                && check.operator === "=="
                // which side is nullish?
                && ((nullish_side = is_nullish(check.left, compressor) && check.left)
                    || (nullish_side = is_nullish(check.right, compressor) && check.right))
                // is the other side the same as the check_subject
                && (nullish_side === check.left
                    ? check.right
                    : check.left).equivalent_to(check_subject)) {
                return true;
            }
            // foo === null || foo === undefined
            if (check instanceof AST_Binary && check.operator === "||") {
                let null_cmp;
                let undefined_cmp;
                const find_comparison = cmp => {
                    if (!(cmp instanceof AST_Binary
                        && (cmp.operator === "===" || cmp.operator === "=="))) {
                        return false;
                    }
                    let found = 0;
                    let defined_side;
                    if (cmp.left instanceof AST_Null) {
                        found++;
                        null_cmp = cmp;
                        defined_side = cmp.right;
                    }
                    if (cmp.right instanceof AST_Null) {
                        found++;
                        null_cmp = cmp;
                        defined_side = cmp.left;
                    }
                    if (is_undefined(cmp.left, compressor)) {
                        found++;
                        undefined_cmp = cmp;
                        defined_side = cmp.right;
                    }
                    if (is_undefined(cmp.right, compressor)) {
                        found++;
                        undefined_cmp = cmp;
                        defined_side = cmp.left;
                    }
                    if (found !== 1) {
                        return false;
                    }
                    if (!defined_side.equivalent_to(check_subject)) {
                        return false;
                    }
                    return true;
                };
                if (!find_comparison(check.left))
                    return false;
                if (!find_comparison(check.right))
                    return false;
                if (null_cmp && undefined_cmp && null_cmp !== undefined_cmp) {
                    return true;
                }
            }
            return false;
        }
        def_optimize(AST_Conditional, function (self, compressor) {
            if (!compressor.option("conditionals"))
                return self;
            // This looks like lift_sequences(), should probably be under "sequences"
            if (self.condition instanceof AST_Sequence) {
                var expressions = self.condition.expressions.slice();
                self.condition = expressions.pop();
                expressions.push(self);
                return make_sequence(self, expressions);
            }
            var cond = self.condition.evaluate(compressor);
            if (cond !== self.condition) {
                if (cond) {
                    return maintain_this_binding(compressor.parent(), compressor.self(), self.consequent);
                }
                else {
                    return maintain_this_binding(compressor.parent(), compressor.self(), self.alternative);
                }
            }
            var negated = cond.negate(compressor, first_in_statement(compressor));
            if (best_of(compressor, cond, negated) === negated) {
                self = make_node(AST_Conditional, self, {
                    condition: negated,
                    consequent: self.alternative,
                    alternative: self.consequent
                });
            }
            var condition = self.condition;
            var consequent = self.consequent;
            var alternative = self.alternative;
            // x?x:y --> x||y
            if (condition instanceof AST_SymbolRef
                && consequent instanceof AST_SymbolRef
                && condition.definition() === consequent.definition()) {
                return make_node(AST_Binary, self, {
                    operator: "||",
                    left: condition,
                    right: alternative
                });
            }
            // if (foo) exp = something; else exp = something_else;
            //                   |
            //                   v
            // exp = foo ? something : something_else;
            if (consequent instanceof AST_Assign
                && alternative instanceof AST_Assign
                && consequent.operator === alternative.operator
                && consequent.logical === alternative.logical
                && consequent.left.equivalent_to(alternative.left)
                && (!self.condition.has_side_effects(compressor)
                    || consequent.operator == "="
                        && !consequent.left.has_side_effects(compressor))) {
                return make_node(AST_Assign, self, {
                    operator: consequent.operator,
                    left: consequent.left,
                    logical: consequent.logical,
                    right: make_node(AST_Conditional, self, {
                        condition: self.condition,
                        consequent: consequent.right,
                        alternative: alternative.right
                    })
                });
            }
            // x ? y(a) : y(b) --> y(x ? a : b)
            var arg_index;
            if (consequent instanceof AST_Call
                && alternative.TYPE === consequent.TYPE
                && consequent.args.length > 0
                && consequent.args.length == alternative.args.length
                && consequent.expression.equivalent_to(alternative.expression)
                && !self.condition.has_side_effects(compressor)
                && !consequent.expression.has_side_effects(compressor)
                && typeof (arg_index = single_arg_diff()) == "number") {
                var node = consequent.clone();
                node.args[arg_index] = make_node(AST_Conditional, self, {
                    condition: self.condition,
                    consequent: consequent.args[arg_index],
                    alternative: alternative.args[arg_index]
                });
                return node;
            }
            // a ? b : c ? b : d --> (a || c) ? b : d
            if (alternative instanceof AST_Conditional
                && consequent.equivalent_to(alternative.consequent)) {
                return make_node(AST_Conditional, self, {
                    condition: make_node(AST_Binary, self, {
                        operator: "||",
                        left: condition,
                        right: alternative.condition
                    }),
                    consequent: consequent,
                    alternative: alternative.alternative
                }).optimize(compressor);
            }
            // a == null ? b : a -> a ?? b
            if (compressor.option("ecma") >= 2020 &&
                is_nullish_check(condition, alternative, compressor)) {
                return make_node(AST_Binary, self, {
                    operator: "??",
                    left: alternative,
                    right: consequent
                }).optimize(compressor);
            }
            // a ? b : (c, b) --> (a || c), b
            if (alternative instanceof AST_Sequence
                && consequent.equivalent_to(alternative.expressions[alternative.expressions.length - 1])) {
                return make_sequence(self, [
                    make_node(AST_Binary, self, {
                        operator: "||",
                        left: condition,
                        right: make_sequence(self, alternative.expressions.slice(0, -1))
                    }),
                    consequent
                ]).optimize(compressor);
            }
            // a ? b : (c && b) --> (a || c) && b
            if (alternative instanceof AST_Binary
                && alternative.operator == "&&"
                && consequent.equivalent_to(alternative.right)) {
                return make_node(AST_Binary, self, {
                    operator: "&&",
                    left: make_node(AST_Binary, self, {
                        operator: "||",
                        left: condition,
                        right: alternative.left
                    }),
                    right: consequent
                }).optimize(compressor);
            }
            // x?y?z:a:a --> x&&y?z:a
            if (consequent instanceof AST_Conditional
                && consequent.alternative.equivalent_to(alternative)) {
                return make_node(AST_Conditional, self, {
                    condition: make_node(AST_Binary, self, {
                        left: self.condition,
                        operator: "&&",
                        right: consequent.condition
                    }),
                    consequent: consequent.consequent,
                    alternative: alternative
                });
            }
            // x ? y : y --> x, y
            if (consequent.equivalent_to(alternative)) {
                return make_sequence(self, [
                    self.condition,
                    consequent
                ]).optimize(compressor);
            }
            // x ? y || z : z --> x && y || z
            if (consequent instanceof AST_Binary
                && consequent.operator == "||"
                && consequent.right.equivalent_to(alternative)) {
                return make_node(AST_Binary, self, {
                    operator: "||",
                    left: make_node(AST_Binary, self, {
                        operator: "&&",
                        left: self.condition,
                        right: consequent.left
                    }),
                    right: alternative
                }).optimize(compressor);
            }
            var in_bool = compressor.in_boolean_context();
            if (is_true(self.consequent)) {
                if (is_false(self.alternative)) {
                    // c ? true : false ---> !!c
                    return booleanize(self.condition);
                }
                // c ? true : x ---> !!c || x
                return make_node(AST_Binary, self, {
                    operator: "||",
                    left: booleanize(self.condition),
                    right: self.alternative
                });
            }
            if (is_false(self.consequent)) {
                if (is_true(self.alternative)) {
                    // c ? false : true ---> !c
                    return booleanize(self.condition.negate(compressor));
                }
                // c ? false : x ---> !c && x
                return make_node(AST_Binary, self, {
                    operator: "&&",
                    left: booleanize(self.condition.negate(compressor)),
                    right: self.alternative
                });
            }
            if (is_true(self.alternative)) {
                // c ? x : true ---> !c || x
                return make_node(AST_Binary, self, {
                    operator: "||",
                    left: booleanize(self.condition.negate(compressor)),
                    right: self.consequent
                });
            }
            if (is_false(self.alternative)) {
                // c ? x : false ---> !!c && x
                return make_node(AST_Binary, self, {
                    operator: "&&",
                    left: booleanize(self.condition),
                    right: self.consequent
                });
            }
            return self;
            function booleanize(node) {
                if (node.is_boolean())
                    return node;
                // !!expression
                return make_node(AST_UnaryPrefix, node, {
                    operator: "!",
                    expression: node.negate(compressor)
                });
            }
            // AST_True or !0
            function is_true(node) {
                return node instanceof AST_True
                    || in_bool
                        && node instanceof AST_Constant
                        && node.getValue()
                    || (node instanceof AST_UnaryPrefix
                        && node.operator == "!"
                        && node.expression instanceof AST_Constant
                        && !node.expression.getValue());
            }
            // AST_False or !1
            function is_false(node) {
                return node instanceof AST_False
                    || in_bool
                        && node instanceof AST_Constant
                        && !node.getValue()
                    || (node instanceof AST_UnaryPrefix
                        && node.operator == "!"
                        && node.expression instanceof AST_Constant
                        && node.expression.getValue());
            }
            function single_arg_diff() {
                var a = consequent.args;
                var b = alternative.args;
                for (var i = 0, len = a.length; i < len; i++) {
                    if (a[i] instanceof AST_Expansion)
                        return;
                    if (!a[i].equivalent_to(b[i])) {
                        if (b[i] instanceof AST_Expansion)
                            return;
                        for (var j = i + 1; j < len; j++) {
                            if (a[j] instanceof AST_Expansion)
                                return;
                            if (!a[j].equivalent_to(b[j]))
                                return;
                        }
                        return i;
                    }
                }
            }
        });
        def_optimize(AST_Boolean, function (self, compressor) {
            if (compressor.in_boolean_context())
                return make_node(AST_Number, self, {
                    value: +self.value
                });
            var p = compressor.parent();
            if (compressor.option("booleans_as_integers")) {
                if (p instanceof AST_Binary && (p.operator == "===" || p.operator == "!==")) {
                    p.operator = p.operator.replace(/=$/, "");
                }
                return make_node(AST_Number, self, {
                    value: +self.value
                });
            }
            if (compressor.option("booleans")) {
                if (p instanceof AST_Binary && (p.operator == "=="
                    || p.operator == "!=")) {
                    return make_node(AST_Number, self, {
                        value: +self.value
                    });
                }
                return make_node(AST_UnaryPrefix, self, {
                    operator: "!",
                    expression: make_node(AST_Number, self, {
                        value: 1 - self.value
                    })
                });
            }
            return self;
        });
        function safe_to_flatten(value, compressor) {
            if (value instanceof AST_SymbolRef) {
                value = value.fixed_value();
            }
            if (!value)
                return false;
            if (!(value instanceof AST_Lambda || value instanceof AST_Class))
                return true;
            if (!(value instanceof AST_Lambda && value.contains_this()))
                return true;
            return compressor.parent() instanceof AST_New;
        }
        AST_PropAccess.DEFMETHOD("flatten_object", function (key, compressor) {
            if (!compressor.option("properties"))
                return;
            if (key === "__proto__")
                return;
            var arrows = compressor.option("unsafe_arrows") && compressor.option("ecma") >= 2015;
            var expr = this.expression;
            if (expr instanceof AST_Object) {
                var props = expr.properties;
                for (var i = props.length; --i >= 0;) {
                    var prop = props[i];
                    if ("" + (prop instanceof AST_ConciseMethod ? prop.key.name : prop.key) == key) {
                        const all_props_flattenable = props.every((p) => (p instanceof AST_ObjectKeyVal
                            || arrows && p instanceof AST_ConciseMethod && !p.is_generator)
                            && !p.computed_key());
                        if (!all_props_flattenable)
                            return;
                        if (!safe_to_flatten(prop.value, compressor))
                            return;
                        return make_node(AST_Sub, this, {
                            expression: make_node(AST_Array, expr, {
                                elements: props.map(function (prop) {
                                    var v = prop.value;
                                    if (v instanceof AST_Accessor) {
                                        v = make_node(AST_Function, v, v);
                                    }
                                    var k = prop.key;
                                    if (k instanceof AST_Node && !(k instanceof AST_SymbolMethod)) {
                                        return make_sequence(prop, [k, v]);
                                    }
                                    return v;
                                })
                            }),
                            property: make_node(AST_Number, this, {
                                value: i
                            })
                        });
                    }
                }
            }
        });
        def_optimize(AST_Sub, function (self, compressor) {
            var expr = self.expression;
            var prop = self.property;
            if (compressor.option("properties")) {
                var key = prop.evaluate(compressor);
                if (key !== prop) {
                    if (typeof key == "string") {
                        if (key == "undefined") {
                            key = undefined;
                        }
                        else {
                            var value = parseFloat(key);
                            if (value.toString() == key) {
                                key = value;
                            }
                        }
                    }
                    prop = self.property = best_of_expression(prop, make_node_from_constant(key, prop).transform(compressor));
                    var property = "" + key;
                    if (is_basic_identifier_string(property)
                        && property.length <= prop.size() + 1) {
                        return make_node(AST_Dot, self, {
                            expression: expr,
                            optional: self.optional,
                            property: property,
                            quote: prop.quote,
                        }).optimize(compressor);
                    }
                }
            }
            var fn;
            OPT_ARGUMENTS: if (compressor.option("arguments")
                && expr instanceof AST_SymbolRef
                && expr.name == "arguments"
                && expr.definition().orig.length == 1
                && (fn = expr.scope) instanceof AST_Lambda
                && fn.uses_arguments
                && !(fn instanceof AST_Arrow)
                && prop instanceof AST_Number) {
                var index = prop.getValue();
                var params = new Set();
                var argnames = fn.argnames;
                for (var n = 0; n < argnames.length; n++) {
                    if (!(argnames[n] instanceof AST_SymbolFunarg)) {
                        break OPT_ARGUMENTS; // destructuring parameter - bail
                    }
                    var param = argnames[n].name;
                    if (params.has(param)) {
                        break OPT_ARGUMENTS; // duplicate parameter - bail
                    }
                    params.add(param);
                }
                var argname = fn.argnames[index];
                if (argname && compressor.has_directive("use strict")) {
                    var def = argname.definition();
                    if (!compressor.option("reduce_vars") || def.assignments || def.orig.length > 1) {
                        argname = null;
                    }
                }
                else if (!argname && !compressor.option("keep_fargs") && index < fn.argnames.length + 5) {
                    while (index >= fn.argnames.length) {
                        argname = fn.create_symbol(AST_SymbolFunarg, {
                            source: fn,
                            scope: fn,
                            tentative_name: "argument_" + fn.argnames.length,
                        });
                        fn.argnames.push(argname);
                    }
                }
                if (argname) {
                    var sym = make_node(AST_SymbolRef, self, argname);
                    sym.reference({});
                    clear_flag(argname, UNUSED);
                    return sym;
                }
            }
            if (is_lhs(self, compressor.parent()))
                return self;
            if (key !== prop) {
                var sub = self.flatten_object(property, compressor);
                if (sub) {
                    expr = self.expression = sub.expression;
                    prop = self.property = sub.property;
                }
            }
            if (compressor.option("properties") && compressor.option("side_effects")
                && prop instanceof AST_Number && expr instanceof AST_Array) {
                var index = prop.getValue();
                var elements = expr.elements;
                var retValue = elements[index];
                FLATTEN: if (safe_to_flatten(retValue, compressor)) {
                    var flatten = true;
                    var values = [];
                    for (var i = elements.length; --i > index;) {
                        var value = elements[i].drop_side_effect_free(compressor);
                        if (value) {
                            values.unshift(value);
                            if (flatten && value.has_side_effects(compressor))
                                flatten = false;
                        }
                    }
                    if (retValue instanceof AST_Expansion)
                        break FLATTEN;
                    retValue = retValue instanceof AST_Hole ? make_node(AST_Undefined, retValue) : retValue;
                    if (!flatten)
                        values.unshift(retValue);
                    while (--i >= 0) {
                        var value = elements[i];
                        if (value instanceof AST_Expansion)
                            break FLATTEN;
                        value = value.drop_side_effect_free(compressor);
                        if (value)
                            values.unshift(value);
                        else
                            index--;
                    }
                    if (flatten) {
                        values.push(retValue);
                        return make_sequence(self, values).optimize(compressor);
                    }
                    else
                        return make_node(AST_Sub, self, {
                            expression: make_node(AST_Array, expr, {
                                elements: values
                            }),
                            property: make_node(AST_Number, prop, {
                                value: index
                            })
                        });
                }
            }
            var ev = self.evaluate(compressor);
            if (ev !== self) {
                ev = make_node_from_constant(ev, self).optimize(compressor);
                return best_of(compressor, ev, self);
            }
            if (self.optional && is_nullish(self.expression, compressor)) {
                return make_node(AST_Undefined, self);
            }
            return self;
        });
        def_optimize(AST_Chain, function (self, compressor) {
            self.expression = self.expression.optimize(compressor);
            return self;
        });
        AST_Lambda.DEFMETHOD("contains_this", function () {
            return walk(this, node => {
                if (node instanceof AST_This)
                    return walk_abort;
                if (node !== this
                    && node instanceof AST_Scope
                    && !(node instanceof AST_Arrow)) {
                    return true;
                }
            });
        });
        def_optimize(AST_Dot, function (self, compressor) {
            const parent = compressor.parent();
            if (is_lhs(self, parent))
                return self;
            if (compressor.option("unsafe_proto")
                && self.expression instanceof AST_Dot
                && self.expression.property == "prototype") {
                var exp = self.expression.expression;
                if (is_undeclared_ref(exp))
                    switch (exp.name) {
                        case "Array":
                            self.expression = make_node(AST_Array, self.expression, {
                                elements: []
                            });
                            break;
                        case "Function":
                            self.expression = make_node(AST_Function, self.expression, {
                                argnames: [],
                                body: []
                            });
                            break;
                        case "Number":
                            self.expression = make_node(AST_Number, self.expression, {
                                value: 0
                            });
                            break;
                        case "Object":
                            self.expression = make_node(AST_Object, self.expression, {
                                properties: []
                            });
                            break;
                        case "RegExp":
                            self.expression = make_node(AST_RegExp, self.expression, {
                                value: { source: "t", flags: "" }
                            });
                            break;
                        case "String":
                            self.expression = make_node(AST_String, self.expression, {
                                value: ""
                            });
                            break;
                    }
            }
            if (!(parent instanceof AST_Call) || !has_annotation(parent, _NOINLINE)) {
                const sub = self.flatten_object(self.property, compressor);
                if (sub)
                    return sub.optimize(compressor);
            }
            let ev = self.evaluate(compressor);
            if (ev !== self) {
                ev = make_node_from_constant(ev, self).optimize(compressor);
                return best_of(compressor, ev, self);
            }
            if (self.optional && is_nullish(self.expression, compressor)) {
                return make_node(AST_Undefined, self);
            }
            return self;
        });
        function literals_in_boolean_context(self, compressor) {
            if (compressor.in_boolean_context()) {
                return best_of(compressor, self, make_sequence(self, [
                    self,
                    make_node(AST_True, self)
                ]).optimize(compressor));
            }
            return self;
        }
        function inline_array_like_spread(elements) {
            for (var i = 0; i < elements.length; i++) {
                var el = elements[i];
                if (el instanceof AST_Expansion) {
                    var expr = el.expression;
                    if (expr instanceof AST_Array
                        && !expr.elements.some(elm => elm instanceof AST_Hole)) {
                        elements.splice(i, 1, ...expr.elements);
                        // Step back one, as the element at i is now new.
                        i--;
                    }
                    // In array-like spread, spreading a non-iterable value is TypeError.
                    // We therefore can’t optimize anything else, unlike with object spread.
                }
            }
        }
        def_optimize(AST_Array, function (self, compressor) {
            var optimized = literals_in_boolean_context(self, compressor);
            if (optimized !== self) {
                return optimized;
            }
            inline_array_like_spread(self.elements);
            return self;
        });
        function inline_object_prop_spread(props, compressor) {
            for (var i = 0; i < props.length; i++) {
                var prop = props[i];
                if (prop instanceof AST_Expansion) {
                    const expr = prop.expression;
                    if (expr instanceof AST_Object
                        && expr.properties.every(prop => prop instanceof AST_ObjectKeyVal)) {
                        props.splice(i, 1, ...expr.properties);
                        // Step back one, as the property at i is now new.
                        i--;
                    }
                    else if (expr instanceof AST_Constant
                        && !(expr instanceof AST_String)) {
                        // Unlike array-like spread, in object spread, spreading a
                        // non-iterable value silently does nothing; it is thus safe
                        // to remove. AST_String is the only iterable AST_Constant.
                        props.splice(i, 1);
                    }
                    else if (is_nullish(expr, compressor)) {
                        // Likewise, null and undefined can be silently removed.
                        props.splice(i, 1);
                    }
                }
            }
        }
        def_optimize(AST_Object, function (self, compressor) {
            var optimized = literals_in_boolean_context(self, compressor);
            if (optimized !== self) {
                return optimized;
            }
            inline_object_prop_spread(self.properties, compressor);
            return self;
        });
        def_optimize(AST_RegExp, literals_in_boolean_context);
        def_optimize(AST_Return, function (self, compressor) {
            if (self.value && is_undefined(self.value, compressor)) {
                self.value = null;
            }
            return self;
        });
        def_optimize(AST_Arrow, opt_AST_Lambda);
        def_optimize(AST_Function, function (self, compressor) {
            self = opt_AST_Lambda(self, compressor);
            if (compressor.option("unsafe_arrows")
                && compressor.option("ecma") >= 2015
                && !self.name
                && !self.is_generator
                && !self.uses_arguments
                && !self.pinned()) {
                const uses_this = walk(self, node => {
                    if (node instanceof AST_This)
                        return walk_abort;
                });
                if (!uses_this)
                    return make_node(AST_Arrow, self, self).optimize(compressor);
            }
            return self;
        });
        def_optimize(AST_Class, function (self) {
            // HACK to avoid compress failure.
            // AST_Class is not really an AST_Scope/AST_Block as it lacks a body.
            return self;
        });
        def_optimize(AST_Yield, function (self, compressor) {
            if (self.expression && !self.is_star && is_undefined(self.expression, compressor)) {
                self.expression = null;
            }
            return self;
        });
        def_optimize(AST_TemplateString, function (self, compressor) {
            if (!compressor.option("evaluate")
                || compressor.parent() instanceof AST_PrefixedTemplateString) {
                return self;
            }
            var segments = [];
            for (var i = 0; i < self.segments.length; i++) {
                var segment = self.segments[i];
                if (segment instanceof AST_Node) {
                    var result = segment.evaluate(compressor);
                    // Evaluate to constant value
                    // Constant value shorter than ${segment}
                    if (result !== segment && (result + "").length <= segment.size() + "${}".length) {
                        // There should always be a previous and next segment if segment is a node
                        segments[segments.length - 1].value = segments[segments.length - 1].value + result + self.segments[++i].value;
                        continue;
                    }
                    // `before ${`innerBefore ${any} innerAfter`} after` => `before innerBefore ${any} innerAfter after`
                    // TODO:
                    // `before ${'test' + foo} after` => `before innerBefore ${any} innerAfter after`
                    // `before ${foo + 'test} after` => `before innerBefore ${any} innerAfter after`
                    if (segment instanceof AST_TemplateString) {
                        var inners = segment.segments;
                        segments[segments.length - 1].value += inners[0].value;
                        for (var j = 1; j < inners.length; j++) {
                            segment = inners[j];
                            segments.push(segment);
                        }
                        continue;
                    }
                }
                segments.push(segment);
            }
            self.segments = segments;
            // `foo` => "foo"
            if (segments.length == 1) {
                return make_node(AST_String, self, segments[0]);
            }
            if (segments.length === 3
                && segments[1] instanceof AST_Node
                && (segments[1].is_string(compressor)
                    || segments[1].is_number(compressor)
                    || is_nullish(segments[1], compressor)
                    || compressor.option("unsafe"))) {
                // `foo${bar}` => "foo" + bar
                if (segments[2].value === "") {
                    return make_node(AST_Binary, self, {
                        operator: "+",
                        left: make_node(AST_String, self, {
                            value: segments[0].value,
                        }),
                        right: segments[1],
                    });
                }
                // `${bar}baz` => bar + "baz"
                if (segments[0].value === "") {
                    return make_node(AST_Binary, self, {
                        operator: "+",
                        left: segments[1],
                        right: make_node(AST_String, self, {
                            value: segments[2].value,
                        }),
                    });
                }
            }
            return self;
        });
        def_optimize(AST_PrefixedTemplateString, function (self) {
            return self;
        });
        // ["p"]:1 ---> p:1
        // [42]:1 ---> 42:1
        function lift_key(self, compressor) {
            if (!compressor.option("computed_props"))
                return self;
            // save a comparison in the typical case
            if (!(self.key instanceof AST_Constant))
                return self;
            // allow certain acceptable props as not all AST_Constants are true constants
            if (self.key instanceof AST_String || self.key instanceof AST_Number) {
                if (self.key.value === "__proto__")
                    return self;
                if (self.key.value == "constructor"
                    && compressor.parent() instanceof AST_Class)
                    return self;
                if (self instanceof AST_ObjectKeyVal) {
                    self.key = self.key.value;
                }
                else if (self instanceof AST_ClassProperty) {
                    self.key = make_node(AST_SymbolClassProperty, self.key, {
                        name: self.key.value
                    });
                }
                else {
                    self.key = make_node(AST_SymbolMethod, self.key, {
                        name: self.key.value
                    });
                }
            }
            return self;
        }
        def_optimize(AST_ObjectProperty, lift_key);
        def_optimize(AST_ConciseMethod, function (self, compressor) {
            lift_key(self, compressor);
            // p(){return x;} ---> p:()=>x
            if (compressor.option("arrows")
                && compressor.parent() instanceof AST_Object
                && !self.is_generator
                && !self.value.uses_arguments
                && !self.value.pinned()
                && self.value.body.length == 1
                && self.value.body[0] instanceof AST_Return
                && self.value.body[0].value
                && !self.value.contains_this()) {
                var arrow = make_node(AST_Arrow, self.value, self.value);
                arrow.async = self.async;
                arrow.is_generator = self.is_generator;
                return make_node(AST_ObjectKeyVal, self, {
                    key: self.key instanceof AST_SymbolMethod ? self.key.name : self.key,
                    value: arrow,
                    quote: self.quote,
                });
            }
            return self;
        });
        def_optimize(AST_ObjectKeyVal, function (self, compressor) {
            lift_key(self, compressor);
            // p:function(){} ---> p(){}
            // p:function*(){} ---> *p(){}
            // p:async function(){} ---> async p(){}
            // p:()=>{} ---> p(){}
            // p:async()=>{} ---> async p(){}
            var unsafe_methods = compressor.option("unsafe_methods");
            if (unsafe_methods
                && compressor.option("ecma") >= 2015
                && (!(unsafe_methods instanceof RegExp) || unsafe_methods.test(self.key + ""))) {
                var key = self.key;
                var value = self.value;
                var is_arrow_with_block = value instanceof AST_Arrow
                    && Array.isArray(value.body)
                    && !value.contains_this();
                if ((is_arrow_with_block || value instanceof AST_Function) && !value.name) {
                    return make_node(AST_ConciseMethod, self, {
                        async: value.async,
                        is_generator: value.is_generator,
                        key: key instanceof AST_Node ? key : make_node(AST_SymbolMethod, self, {
                            name: key,
                        }),
                        value: make_node(AST_Accessor, value, value),
                        quote: self.quote,
                    });
                }
            }
            return self;
        });
        def_optimize(AST_Destructuring, function (self, compressor) {
            if (compressor.option("pure_getters") == true
                && compressor.option("unused")
                && !self.is_array
                && Array.isArray(self.names)
                && !is_destructuring_export_decl(compressor)
                && !(self.names[self.names.length - 1] instanceof AST_Expansion)) {
                var keep = [];
                for (var i = 0; i < self.names.length; i++) {
                    var elem = self.names[i];
                    if (!(elem instanceof AST_ObjectKeyVal
                        && typeof elem.key == "string"
                        && elem.value instanceof AST_SymbolDeclaration
                        && !should_retain(compressor, elem.value.definition()))) {
                        keep.push(elem);
                    }
                }
                if (keep.length != self.names.length) {
                    self.names = keep;
                }
            }
            return self;
            function is_destructuring_export_decl(compressor) {
                var ancestors = [/^VarDef$/, /^(Const|Let|Var)$/, /^Export$/];
                for (var a = 0, p = 0, len = ancestors.length; a < len; p++) {
                    var parent = compressor.parent(p);
                    if (!parent)
                        return false;
                    if (a === 0 && parent.TYPE == "Destructuring")
                        continue;
                    if (!ancestors[a].test(parent.TYPE)) {
                        return false;
                    }
                    a++;
                }
                return true;
            }
            function should_retain(compressor, def) {
                if (def.references.length)
                    return true;
                if (!def.global)
                    return false;
                if (compressor.toplevel.vars) {
                    if (compressor.top_retain) {
                        return compressor.top_retain(def);
                    }
                    return false;
                }
                return true;
            }
        });
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        // a small wrapper around fitzgen's source-map library
        async function SourceMap(options) {
            options = defaults(options, {
                file: null,
                root: null,
                orig: null,
                orig_line_diff: 0,
                dest_line_diff: 0,
            });
            var orig_map;
            var generator = new MOZ_SourceMap__default['default'].SourceMapGenerator({
                file: options.file,
                sourceRoot: options.root
            });
            if (options.orig) {
                orig_map = await new MOZ_SourceMap__default['default'].SourceMapConsumer(options.orig);
                orig_map.sources.forEach(function (source) {
                    var sourceContent = orig_map.sourceContentFor(source, true);
                    if (sourceContent) {
                        generator.setSourceContent(source, sourceContent);
                    }
                });
            }
            function add(source, gen_line, gen_col, orig_line, orig_col, name) {
                if (orig_map) {
                    var info = orig_map.originalPositionFor({
                        line: orig_line,
                        column: orig_col
                    });
                    if (info.source === null) {
                        return;
                    }
                    source = info.source;
                    orig_line = info.line;
                    orig_col = info.column;
                    name = info.name || name;
                }
                generator.addMapping({
                    generated: { line: gen_line + options.dest_line_diff, column: gen_col },
                    original: { line: orig_line + options.orig_line_diff, column: orig_col },
                    source: source,
                    name: name
                });
            }
            return {
                add: add,
                get: function () { return generator; },
                toString: function () { return generator.toString(); },
                destroy: function () {
                    if (orig_map && orig_map.destroy) {
                        orig_map.destroy();
                    }
                }
            };
        }
        var domprops = [
            "$&",
            "$'",
            "$*",
            "$+",
            "$1",
            "$2",
            "$3",
            "$4",
            "$5",
            "$6",
            "$7",
            "$8",
            "$9",
            "$_",
            "$`",
            "$input",
            "-moz-animation",
            "-moz-animation-delay",
            "-moz-animation-direction",
            "-moz-animation-duration",
            "-moz-animation-fill-mode",
            "-moz-animation-iteration-count",
            "-moz-animation-name",
            "-moz-animation-play-state",
            "-moz-animation-timing-function",
            "-moz-appearance",
            "-moz-backface-visibility",
            "-moz-border-end",
            "-moz-border-end-color",
            "-moz-border-end-style",
            "-moz-border-end-width",
            "-moz-border-image",
            "-moz-border-start",
            "-moz-border-start-color",
            "-moz-border-start-style",
            "-moz-border-start-width",
            "-moz-box-align",
            "-moz-box-direction",
            "-moz-box-flex",
            "-moz-box-ordinal-group",
            "-moz-box-orient",
            "-moz-box-pack",
            "-moz-box-sizing",
            "-moz-float-edge",
            "-moz-font-feature-settings",
            "-moz-font-language-override",
            "-moz-force-broken-image-icon",
            "-moz-hyphens",
            "-moz-image-region",
            "-moz-margin-end",
            "-moz-margin-start",
            "-moz-orient",
            "-moz-osx-font-smoothing",
            "-moz-outline-radius",
            "-moz-outline-radius-bottomleft",
            "-moz-outline-radius-bottomright",
            "-moz-outline-radius-topleft",
            "-moz-outline-radius-topright",
            "-moz-padding-end",
            "-moz-padding-start",
            "-moz-perspective",
            "-moz-perspective-origin",
            "-moz-tab-size",
            "-moz-text-size-adjust",
            "-moz-transform",
            "-moz-transform-origin",
            "-moz-transform-style",
            "-moz-transition",
            "-moz-transition-delay",
            "-moz-transition-duration",
            "-moz-transition-property",
            "-moz-transition-timing-function",
            "-moz-user-focus",
            "-moz-user-input",
            "-moz-user-modify",
            "-moz-user-select",
            "-moz-window-dragging",
            "-webkit-align-content",
            "-webkit-align-items",
            "-webkit-align-self",
            "-webkit-animation",
            "-webkit-animation-delay",
            "-webkit-animation-direction",
            "-webkit-animation-duration",
            "-webkit-animation-fill-mode",
            "-webkit-animation-iteration-count",
            "-webkit-animation-name",
            "-webkit-animation-play-state",
            "-webkit-animation-timing-function",
            "-webkit-appearance",
            "-webkit-backface-visibility",
            "-webkit-background-clip",
            "-webkit-background-origin",
            "-webkit-background-size",
            "-webkit-border-bottom-left-radius",
            "-webkit-border-bottom-right-radius",
            "-webkit-border-image",
            "-webkit-border-radius",
            "-webkit-border-top-left-radius",
            "-webkit-border-top-right-radius",
            "-webkit-box-align",
            "-webkit-box-direction",
            "-webkit-box-flex",
            "-webkit-box-ordinal-group",
            "-webkit-box-orient",
            "-webkit-box-pack",
            "-webkit-box-shadow",
            "-webkit-box-sizing",
            "-webkit-filter",
            "-webkit-flex",
            "-webkit-flex-basis",
            "-webkit-flex-direction",
            "-webkit-flex-flow",
            "-webkit-flex-grow",
            "-webkit-flex-shrink",
            "-webkit-flex-wrap",
            "-webkit-justify-content",
            "-webkit-line-clamp",
            "-webkit-mask",
            "-webkit-mask-clip",
            "-webkit-mask-composite",
            "-webkit-mask-image",
            "-webkit-mask-origin",
            "-webkit-mask-position",
            "-webkit-mask-position-x",
            "-webkit-mask-position-y",
            "-webkit-mask-repeat",
            "-webkit-mask-size",
            "-webkit-order",
            "-webkit-perspective",
            "-webkit-perspective-origin",
            "-webkit-text-fill-color",
            "-webkit-text-size-adjust",
            "-webkit-text-stroke",
            "-webkit-text-stroke-color",
            "-webkit-text-stroke-width",
            "-webkit-transform",
            "-webkit-transform-origin",
            "-webkit-transform-style",
            "-webkit-transition",
            "-webkit-transition-delay",
            "-webkit-transition-duration",
            "-webkit-transition-property",
            "-webkit-transition-timing-function",
            "-webkit-user-select",
            "0",
            "1",
            "10",
            "11",
            "12",
            "13",
            "14",
            "15",
            "16",
            "17",
            "18",
            "19",
            "2",
            "20",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "@@iterator",
            "ABORT_ERR",
            "ACTIVE",
            "ACTIVE_ATTRIBUTES",
            "ACTIVE_TEXTURE",
            "ACTIVE_UNIFORMS",
            "ACTIVE_UNIFORM_BLOCKS",
            "ADDITION",
            "ALIASED_LINE_WIDTH_RANGE",
            "ALIASED_POINT_SIZE_RANGE",
            "ALLOW_KEYBOARD_INPUT",
            "ALLPASS",
            "ALPHA",
            "ALPHA_BITS",
            "ALREADY_SIGNALED",
            "ALT_MASK",
            "ALWAYS",
            "ANY_SAMPLES_PASSED",
            "ANY_SAMPLES_PASSED_CONSERVATIVE",
            "ANY_TYPE",
            "ANY_UNORDERED_NODE_TYPE",
            "ARRAY_BUFFER",
            "ARRAY_BUFFER_BINDING",
            "ATTACHED_SHADERS",
            "ATTRIBUTE_NODE",
            "AT_TARGET",
            "AbortController",
            "AbortSignal",
            "AbsoluteOrientationSensor",
            "AbstractRange",
            "Accelerometer",
            "AddSearchProvider",
            "AggregateError",
            "AnalyserNode",
            "Animation",
            "AnimationEffect",
            "AnimationEvent",
            "AnimationPlaybackEvent",
            "AnimationTimeline",
            "AnonXMLHttpRequest",
            "Any",
            "ApplicationCache",
            "ApplicationCacheErrorEvent",
            "Array",
            "ArrayBuffer",
            "ArrayType",
            "Atomics",
            "Attr",
            "Audio",
            "AudioBuffer",
            "AudioBufferSourceNode",
            "AudioContext",
            "AudioDestinationNode",
            "AudioListener",
            "AudioNode",
            "AudioParam",
            "AudioParamMap",
            "AudioProcessingEvent",
            "AudioScheduledSourceNode",
            "AudioStreamTrack",
            "AudioWorklet",
            "AudioWorkletNode",
            "AuthenticatorAssertionResponse",
            "AuthenticatorAttestationResponse",
            "AuthenticatorResponse",
            "AutocompleteErrorEvent",
            "BACK",
            "BAD_BOUNDARYPOINTS_ERR",
            "BAD_REQUEST",
            "BANDPASS",
            "BLEND",
            "BLEND_COLOR",
            "BLEND_DST_ALPHA",
            "BLEND_DST_RGB",
            "BLEND_EQUATION",
            "BLEND_EQUATION_ALPHA",
            "BLEND_EQUATION_RGB",
            "BLEND_SRC_ALPHA",
            "BLEND_SRC_RGB",
            "BLUE_BITS",
            "BLUR",
            "BOOL",
            "BOOLEAN_TYPE",
            "BOOL_VEC2",
            "BOOL_VEC3",
            "BOOL_VEC4",
            "BOTH",
            "BROWSER_DEFAULT_WEBGL",
            "BUBBLING_PHASE",
            "BUFFER_SIZE",
            "BUFFER_USAGE",
            "BYTE",
            "BYTES_PER_ELEMENT",
            "BackgroundFetchManager",
            "BackgroundFetchRecord",
            "BackgroundFetchRegistration",
            "BarProp",
            "BarcodeDetector",
            "BaseAudioContext",
            "BaseHref",
            "BatteryManager",
            "BeforeInstallPromptEvent",
            "BeforeLoadEvent",
            "BeforeUnloadEvent",
            "BigInt",
            "BigInt64Array",
            "BigUint64Array",
            "BiquadFilterNode",
            "Blob",
            "BlobEvent",
            "Bluetooth",
            "BluetoothCharacteristicProperties",
            "BluetoothDevice",
            "BluetoothRemoteGATTCharacteristic",
            "BluetoothRemoteGATTDescriptor",
            "BluetoothRemoteGATTServer",
            "BluetoothRemoteGATTService",
            "BluetoothUUID",
            "Boolean",
            "BroadcastChannel",
            "ByteLengthQueuingStrategy",
            "CAPTURING_PHASE",
            "CCW",
            "CDATASection",
            "CDATA_SECTION_NODE",
            "CHANGE",
            "CHARSET_RULE",
            "CHECKING",
            "CLAMP_TO_EDGE",
            "CLICK",
            "CLOSED",
            "CLOSING",
            "COLOR",
            "COLOR_ATTACHMENT0",
            "COLOR_ATTACHMENT1",
            "COLOR_ATTACHMENT10",
            "COLOR_ATTACHMENT11",
            "COLOR_ATTACHMENT12",
            "COLOR_ATTACHMENT13",
            "COLOR_ATTACHMENT14",
            "COLOR_ATTACHMENT15",
            "COLOR_ATTACHMENT2",
            "COLOR_ATTACHMENT3",
            "COLOR_ATTACHMENT4",
            "COLOR_ATTACHMENT5",
            "COLOR_ATTACHMENT6",
            "COLOR_ATTACHMENT7",
            "COLOR_ATTACHMENT8",
            "COLOR_ATTACHMENT9",
            "COLOR_BUFFER_BIT",
            "COLOR_CLEAR_VALUE",
            "COLOR_WRITEMASK",
            "COMMENT_NODE",
            "COMPARE_REF_TO_TEXTURE",
            "COMPILE_STATUS",
            "COMPRESSED_RGBA_S3TC_DXT1_EXT",
            "COMPRESSED_RGBA_S3TC_DXT3_EXT",
            "COMPRESSED_RGBA_S3TC_DXT5_EXT",
            "COMPRESSED_RGB_S3TC_DXT1_EXT",
            "COMPRESSED_TEXTURE_FORMATS",
            "CONDITION_SATISFIED",
            "CONFIGURATION_UNSUPPORTED",
            "CONNECTING",
            "CONSTANT_ALPHA",
            "CONSTANT_COLOR",
            "CONSTRAINT_ERR",
            "CONTEXT_LOST_WEBGL",
            "CONTROL_MASK",
            "COPY_READ_BUFFER",
            "COPY_READ_BUFFER_BINDING",
            "COPY_WRITE_BUFFER",
            "COPY_WRITE_BUFFER_BINDING",
            "COUNTER_STYLE_RULE",
            "CSS",
            "CSS2Properties",
            "CSSAnimation",
            "CSSCharsetRule",
            "CSSConditionRule",
            "CSSCounterStyleRule",
            "CSSFontFaceRule",
            "CSSFontFeatureValuesRule",
            "CSSGroupingRule",
            "CSSImageValue",
            "CSSImportRule",
            "CSSKeyframeRule",
            "CSSKeyframesRule",
            "CSSKeywordValue",
            "CSSMathInvert",
            "CSSMathMax",
            "CSSMathMin",
            "CSSMathNegate",
            "CSSMathProduct",
            "CSSMathSum",
            "CSSMathValue",
            "CSSMatrixComponent",
            "CSSMediaRule",
            "CSSMozDocumentRule",
            "CSSNameSpaceRule",
            "CSSNamespaceRule",
            "CSSNumericArray",
            "CSSNumericValue",
            "CSSPageRule",
            "CSSPerspective",
            "CSSPositionValue",
            "CSSPrimitiveValue",
            "CSSRotate",
            "CSSRule",
            "CSSRuleList",
            "CSSScale",
            "CSSSkew",
            "CSSSkewX",
            "CSSSkewY",
            "CSSStyleDeclaration",
            "CSSStyleRule",
            "CSSStyleSheet",
            "CSSStyleValue",
            "CSSSupportsRule",
            "CSSTransformComponent",
            "CSSTransformValue",
            "CSSTransition",
            "CSSTranslate",
            "CSSUnitValue",
            "CSSUnknownRule",
            "CSSUnparsedValue",
            "CSSValue",
            "CSSValueList",
            "CSSVariableReferenceValue",
            "CSSVariablesDeclaration",
            "CSSVariablesRule",
            "CSSViewportRule",
            "CSS_ATTR",
            "CSS_CM",
            "CSS_COUNTER",
            "CSS_CUSTOM",
            "CSS_DEG",
            "CSS_DIMENSION",
            "CSS_EMS",
            "CSS_EXS",
            "CSS_FILTER_BLUR",
            "CSS_FILTER_BRIGHTNESS",
            "CSS_FILTER_CONTRAST",
            "CSS_FILTER_CUSTOM",
            "CSS_FILTER_DROP_SHADOW",
            "CSS_FILTER_GRAYSCALE",
            "CSS_FILTER_HUE_ROTATE",
            "CSS_FILTER_INVERT",
            "CSS_FILTER_OPACITY",
            "CSS_FILTER_REFERENCE",
            "CSS_FILTER_SATURATE",
            "CSS_FILTER_SEPIA",
            "CSS_GRAD",
            "CSS_HZ",
            "CSS_IDENT",
            "CSS_IN",
            "CSS_INHERIT",
            "CSS_KHZ",
            "CSS_MATRIX",
            "CSS_MATRIX3D",
            "CSS_MM",
            "CSS_MS",
            "CSS_NUMBER",
            "CSS_PC",
            "CSS_PERCENTAGE",
            "CSS_PERSPECTIVE",
            "CSS_PRIMITIVE_VALUE",
            "CSS_PT",
            "CSS_PX",
            "CSS_RAD",
            "CSS_RECT",
            "CSS_RGBCOLOR",
            "CSS_ROTATE",
            "CSS_ROTATE3D",
            "CSS_ROTATEX",
            "CSS_ROTATEY",
            "CSS_ROTATEZ",
            "CSS_S",
            "CSS_SCALE",
            "CSS_SCALE3D",
            "CSS_SCALEX",
            "CSS_SCALEY",
            "CSS_SCALEZ",
            "CSS_SKEW",
            "CSS_SKEWX",
            "CSS_SKEWY",
            "CSS_STRING",
            "CSS_TRANSLATE",
            "CSS_TRANSLATE3D",
            "CSS_TRANSLATEX",
            "CSS_TRANSLATEY",
            "CSS_TRANSLATEZ",
            "CSS_UNKNOWN",
            "CSS_URI",
            "CSS_VALUE_LIST",
            "CSS_VH",
            "CSS_VMAX",
            "CSS_VMIN",
            "CSS_VW",
            "CULL_FACE",
            "CULL_FACE_MODE",
            "CURRENT_PROGRAM",
            "CURRENT_QUERY",
            "CURRENT_VERTEX_ATTRIB",
            "CUSTOM",
            "CW",
            "Cache",
            "CacheStorage",
            "CanvasCaptureMediaStream",
            "CanvasCaptureMediaStreamTrack",
            "CanvasGradient",
            "CanvasPattern",
            "CanvasRenderingContext2D",
            "CaretPosition",
            "ChannelMergerNode",
            "ChannelSplitterNode",
            "CharacterData",
            "ClientRect",
            "ClientRectList",
            "Clipboard",
            "ClipboardEvent",
            "ClipboardItem",
            "CloseEvent",
            "Collator",
            "CommandEvent",
            "Comment",
            "CompileError",
            "CompositionEvent",
            "CompressionStream",
            "Console",
            "ConstantSourceNode",
            "Controllers",
            "ConvolverNode",
            "CountQueuingStrategy",
            "Counter",
            "Credential",
            "CredentialsContainer",
            "Crypto",
            "CryptoKey",
            "CustomElementRegistry",
            "CustomEvent",
            "DATABASE_ERR",
            "DATA_CLONE_ERR",
            "DATA_ERR",
            "DBLCLICK",
            "DECR",
            "DECR_WRAP",
            "DELETE_STATUS",
            "DEPTH",
            "DEPTH24_STENCIL8",
            "DEPTH32F_STENCIL8",
            "DEPTH_ATTACHMENT",
            "DEPTH_BITS",
            "DEPTH_BUFFER_BIT",
            "DEPTH_CLEAR_VALUE",
            "DEPTH_COMPONENT",
            "DEPTH_COMPONENT16",
            "DEPTH_COMPONENT24",
            "DEPTH_COMPONENT32F",
            "DEPTH_FUNC",
            "DEPTH_RANGE",
            "DEPTH_STENCIL",
            "DEPTH_STENCIL_ATTACHMENT",
            "DEPTH_TEST",
            "DEPTH_WRITEMASK",
            "DEVICE_INELIGIBLE",
            "DIRECTION_DOWN",
            "DIRECTION_LEFT",
            "DIRECTION_RIGHT",
            "DIRECTION_UP",
            "DISABLED",
            "DISPATCH_REQUEST_ERR",
            "DITHER",
            "DOCUMENT_FRAGMENT_NODE",
            "DOCUMENT_NODE",
            "DOCUMENT_POSITION_CONTAINED_BY",
            "DOCUMENT_POSITION_CONTAINS",
            "DOCUMENT_POSITION_DISCONNECTED",
            "DOCUMENT_POSITION_FOLLOWING",
            "DOCUMENT_POSITION_IMPLEMENTATION_SPECIFIC",
            "DOCUMENT_POSITION_PRECEDING",
            "DOCUMENT_TYPE_NODE",
            "DOMCursor",
            "DOMError",
            "DOMException",
            "DOMImplementation",
            "DOMImplementationLS",
            "DOMMatrix",
            "DOMMatrixReadOnly",
            "DOMParser",
            "DOMPoint",
            "DOMPointReadOnly",
            "DOMQuad",
            "DOMRect",
            "DOMRectList",
            "DOMRectReadOnly",
            "DOMRequest",
            "DOMSTRING_SIZE_ERR",
            "DOMSettableTokenList",
            "DOMStringList",
            "DOMStringMap",
            "DOMTokenList",
            "DOMTransactionEvent",
            "DOM_DELTA_LINE",
            "DOM_DELTA_PAGE",
            "DOM_DELTA_PIXEL",
            "DOM_INPUT_METHOD_DROP",
            "DOM_INPUT_METHOD_HANDWRITING",
            "DOM_INPUT_METHOD_IME",
            "DOM_INPUT_METHOD_KEYBOARD",
            "DOM_INPUT_METHOD_MULTIMODAL",
            "DOM_INPUT_METHOD_OPTION",
            "DOM_INPUT_METHOD_PASTE",
            "DOM_INPUT_METHOD_SCRIPT",
            "DOM_INPUT_METHOD_UNKNOWN",
            "DOM_INPUT_METHOD_VOICE",
            "DOM_KEY_LOCATION_JOYSTICK",
            "DOM_KEY_LOCATION_LEFT",
            "DOM_KEY_LOCATION_MOBILE",
            "DOM_KEY_LOCATION_NUMPAD",
            "DOM_KEY_LOCATION_RIGHT",
            "DOM_KEY_LOCATION_STANDARD",
            "DOM_VK_0",
            "DOM_VK_1",
            "DOM_VK_2",
            "DOM_VK_3",
            "DOM_VK_4",
            "DOM_VK_5",
            "DOM_VK_6",
            "DOM_VK_7",
            "DOM_VK_8",
            "DOM_VK_9",
            "DOM_VK_A",
            "DOM_VK_ACCEPT",
            "DOM_VK_ADD",
            "DOM_VK_ALT",
            "DOM_VK_ALTGR",
            "DOM_VK_AMPERSAND",
            "DOM_VK_ASTERISK",
            "DOM_VK_AT",
            "DOM_VK_ATTN",
            "DOM_VK_B",
            "DOM_VK_BACKSPACE",
            "DOM_VK_BACK_QUOTE",
            "DOM_VK_BACK_SLASH",
            "DOM_VK_BACK_SPACE",
            "DOM_VK_C",
            "DOM_VK_CANCEL",
            "DOM_VK_CAPS_LOCK",
            "DOM_VK_CIRCUMFLEX",
            "DOM_VK_CLEAR",
            "DOM_VK_CLOSE_BRACKET",
            "DOM_VK_CLOSE_CURLY_BRACKET",
            "DOM_VK_CLOSE_PAREN",
            "DOM_VK_COLON",
            "DOM_VK_COMMA",
            "DOM_VK_CONTEXT_MENU",
            "DOM_VK_CONTROL",
            "DOM_VK_CONVERT",
            "DOM_VK_CRSEL",
            "DOM_VK_CTRL",
            "DOM_VK_D",
            "DOM_VK_DECIMAL",
            "DOM_VK_DELETE",
            "DOM_VK_DIVIDE",
            "DOM_VK_DOLLAR",
            "DOM_VK_DOUBLE_QUOTE",
            "DOM_VK_DOWN",
            "DOM_VK_E",
            "DOM_VK_EISU",
            "DOM_VK_END",
            "DOM_VK_ENTER",
            "DOM_VK_EQUALS",
            "DOM_VK_EREOF",
            "DOM_VK_ESCAPE",
            "DOM_VK_EXCLAMATION",
            "DOM_VK_EXECUTE",
            "DOM_VK_EXSEL",
            "DOM_VK_F",
            "DOM_VK_F1",
            "DOM_VK_F10",
            "DOM_VK_F11",
            "DOM_VK_F12",
            "DOM_VK_F13",
            "DOM_VK_F14",
            "DOM_VK_F15",
            "DOM_VK_F16",
            "DOM_VK_F17",
            "DOM_VK_F18",
            "DOM_VK_F19",
            "DOM_VK_F2",
            "DOM_VK_F20",
            "DOM_VK_F21",
            "DOM_VK_F22",
            "DOM_VK_F23",
            "DOM_VK_F24",
            "DOM_VK_F25",
            "DOM_VK_F26",
            "DOM_VK_F27",
            "DOM_VK_F28",
            "DOM_VK_F29",
            "DOM_VK_F3",
            "DOM_VK_F30",
            "DOM_VK_F31",
            "DOM_VK_F32",
            "DOM_VK_F33",
            "DOM_VK_F34",
            "DOM_VK_F35",
            "DOM_VK_F36",
            "DOM_VK_F4",
            "DOM_VK_F5",
            "DOM_VK_F6",
            "DOM_VK_F7",
            "DOM_VK_F8",
            "DOM_VK_F9",
            "DOM_VK_FINAL",
            "DOM_VK_FRONT",
            "DOM_VK_G",
            "DOM_VK_GREATER_THAN",
            "DOM_VK_H",
            "DOM_VK_HANGUL",
            "DOM_VK_HANJA",
            "DOM_VK_HASH",
            "DOM_VK_HELP",
            "DOM_VK_HK_TOGGLE",
            "DOM_VK_HOME",
            "DOM_VK_HYPHEN_MINUS",
            "DOM_VK_I",
            "DOM_VK_INSERT",
            "DOM_VK_J",
            "DOM_VK_JUNJA",
            "DOM_VK_K",
            "DOM_VK_KANA",
            "DOM_VK_KANJI",
            "DOM_VK_L",
            "DOM_VK_LEFT",
            "DOM_VK_LEFT_TAB",
            "DOM_VK_LESS_THAN",
            "DOM_VK_M",
            "DOM_VK_META",
            "DOM_VK_MODECHANGE",
            "DOM_VK_MULTIPLY",
            "DOM_VK_N",
            "DOM_VK_NONCONVERT",
            "DOM_VK_NUMPAD0",
            "DOM_VK_NUMPAD1",
            "DOM_VK_NUMPAD2",
            "DOM_VK_NUMPAD3",
            "DOM_VK_NUMPAD4",
            "DOM_VK_NUMPAD5",
            "DOM_VK_NUMPAD6",
            "DOM_VK_NUMPAD7",
            "DOM_VK_NUMPAD8",
            "DOM_VK_NUMPAD9",
            "DOM_VK_NUM_LOCK",
            "DOM_VK_O",
            "DOM_VK_OEM_1",
            "DOM_VK_OEM_102",
            "DOM_VK_OEM_2",
            "DOM_VK_OEM_3",
            "DOM_VK_OEM_4",
            "DOM_VK_OEM_5",
            "DOM_VK_OEM_6",
            "DOM_VK_OEM_7",
            "DOM_VK_OEM_8",
            "DOM_VK_OEM_COMMA",
            "DOM_VK_OEM_MINUS",
            "DOM_VK_OEM_PERIOD",
            "DOM_VK_OEM_PLUS",
            "DOM_VK_OPEN_BRACKET",
            "DOM_VK_OPEN_CURLY_BRACKET",
            "DOM_VK_OPEN_PAREN",
            "DOM_VK_P",
            "DOM_VK_PA1",
            "DOM_VK_PAGEDOWN",
            "DOM_VK_PAGEUP",
            "DOM_VK_PAGE_DOWN",
            "DOM_VK_PAGE_UP",
            "DOM_VK_PAUSE",
            "DOM_VK_PERCENT",
            "DOM_VK_PERIOD",
            "DOM_VK_PIPE",
            "DOM_VK_PLAY",
            "DOM_VK_PLUS",
            "DOM_VK_PRINT",
            "DOM_VK_PRINTSCREEN",
            "DOM_VK_PROCESSKEY",
            "DOM_VK_PROPERITES",
            "DOM_VK_Q",
            "DOM_VK_QUESTION_MARK",
            "DOM_VK_QUOTE",
            "DOM_VK_R",
            "DOM_VK_REDO",
            "DOM_VK_RETURN",
            "DOM_VK_RIGHT",
            "DOM_VK_S",
            "DOM_VK_SCROLL_LOCK",
            "DOM_VK_SELECT",
            "DOM_VK_SEMICOLON",
            "DOM_VK_SEPARATOR",
            "DOM_VK_SHIFT",
            "DOM_VK_SLASH",
            "DOM_VK_SLEEP",
            "DOM_VK_SPACE",
            "DOM_VK_SUBTRACT",
            "DOM_VK_T",
            "DOM_VK_TAB",
            "DOM_VK_TILDE",
            "DOM_VK_U",
            "DOM_VK_UNDERSCORE",
            "DOM_VK_UNDO",
            "DOM_VK_UNICODE",
            "DOM_VK_UP",
            "DOM_VK_V",
            "DOM_VK_VOLUME_DOWN",
            "DOM_VK_VOLUME_MUTE",
            "DOM_VK_VOLUME_UP",
            "DOM_VK_W",
            "DOM_VK_WIN",
            "DOM_VK_WINDOW",
            "DOM_VK_WIN_ICO_00",
            "DOM_VK_WIN_ICO_CLEAR",
            "DOM_VK_WIN_ICO_HELP",
            "DOM_VK_WIN_OEM_ATTN",
            "DOM_VK_WIN_OEM_AUTO",
            "DOM_VK_WIN_OEM_BACKTAB",
            "DOM_VK_WIN_OEM_CLEAR",
            "DOM_VK_WIN_OEM_COPY",
            "DOM_VK_WIN_OEM_CUSEL",
            "DOM_VK_WIN_OEM_ENLW",
            "DOM_VK_WIN_OEM_FINISH",
            "DOM_VK_WIN_OEM_FJ_JISHO",
            "DOM_VK_WIN_OEM_FJ_LOYA",
            "DOM_VK_WIN_OEM_FJ_MASSHOU",
            "DOM_VK_WIN_OEM_FJ_ROYA",
            "DOM_VK_WIN_OEM_FJ_TOUROKU",
            "DOM_VK_WIN_OEM_JUMP",
            "DOM_VK_WIN_OEM_PA1",
            "DOM_VK_WIN_OEM_PA2",
            "DOM_VK_WIN_OEM_PA3",
            "DOM_VK_WIN_OEM_RESET",
            "DOM_VK_WIN_OEM_WSCTRL",
            "DOM_VK_X",
            "DOM_VK_XF86XK_ADD_FAVORITE",
            "DOM_VK_XF86XK_APPLICATION_LEFT",
            "DOM_VK_XF86XK_APPLICATION_RIGHT",
            "DOM_VK_XF86XK_AUDIO_CYCLE_TRACK",
            "DOM_VK_XF86XK_AUDIO_FORWARD",
            "DOM_VK_XF86XK_AUDIO_LOWER_VOLUME",
            "DOM_VK_XF86XK_AUDIO_MEDIA",
            "DOM_VK_XF86XK_AUDIO_MUTE",
            "DOM_VK_XF86XK_AUDIO_NEXT",
            "DOM_VK_XF86XK_AUDIO_PAUSE",
            "DOM_VK_XF86XK_AUDIO_PLAY",
            "DOM_VK_XF86XK_AUDIO_PREV",
            "DOM_VK_XF86XK_AUDIO_RAISE_VOLUME",
            "DOM_VK_XF86XK_AUDIO_RANDOM_PLAY",
            "DOM_VK_XF86XK_AUDIO_RECORD",
            "DOM_VK_XF86XK_AUDIO_REPEAT",
            "DOM_VK_XF86XK_AUDIO_REWIND",
            "DOM_VK_XF86XK_AUDIO_STOP",
            "DOM_VK_XF86XK_AWAY",
            "DOM_VK_XF86XK_BACK",
            "DOM_VK_XF86XK_BACK_FORWARD",
            "DOM_VK_XF86XK_BATTERY",
            "DOM_VK_XF86XK_BLUE",
            "DOM_VK_XF86XK_BLUETOOTH",
            "DOM_VK_XF86XK_BOOK",
            "DOM_VK_XF86XK_BRIGHTNESS_ADJUST",
            "DOM_VK_XF86XK_CALCULATOR",
            "DOM_VK_XF86XK_CALENDAR",
            "DOM_VK_XF86XK_CD",
            "DOM_VK_XF86XK_CLOSE",
            "DOM_VK_XF86XK_COMMUNITY",
            "DOM_VK_XF86XK_CONTRAST_ADJUST",
            "DOM_VK_XF86XK_COPY",
            "DOM_VK_XF86XK_CUT",
            "DOM_VK_XF86XK_CYCLE_ANGLE",
            "DOM_VK_XF86XK_DISPLAY",
            "DOM_VK_XF86XK_DOCUMENTS",
            "DOM_VK_XF86XK_DOS",
            "DOM_VK_XF86XK_EJECT",
            "DOM_VK_XF86XK_EXCEL",
            "DOM_VK_XF86XK_EXPLORER",
            "DOM_VK_XF86XK_FAVORITES",
            "DOM_VK_XF86XK_FINANCE",
            "DOM_VK_XF86XK_FORWARD",
            "DOM_VK_XF86XK_FRAME_BACK",
            "DOM_VK_XF86XK_FRAME_FORWARD",
            "DOM_VK_XF86XK_GAME",
            "DOM_VK_XF86XK_GO",
            "DOM_VK_XF86XK_GREEN",
            "DOM_VK_XF86XK_HIBERNATE",
            "DOM_VK_XF86XK_HISTORY",
            "DOM_VK_XF86XK_HOME_PAGE",
            "DOM_VK_XF86XK_HOT_LINKS",
            "DOM_VK_XF86XK_I_TOUCH",
            "DOM_VK_XF86XK_KBD_BRIGHTNESS_DOWN",
            "DOM_VK_XF86XK_KBD_BRIGHTNESS_UP",
            "DOM_VK_XF86XK_KBD_LIGHT_ON_OFF",
            "DOM_VK_XF86XK_LAUNCH0",
            "DOM_VK_XF86XK_LAUNCH1",
            "DOM_VK_XF86XK_LAUNCH2",
            "DOM_VK_XF86XK_LAUNCH3",
            "DOM_VK_XF86XK_LAUNCH4",
            "DOM_VK_XF86XK_LAUNCH5",
            "DOM_VK_XF86XK_LAUNCH6",
            "DOM_VK_XF86XK_LAUNCH7",
            "DOM_VK_XF86XK_LAUNCH8",
            "DOM_VK_XF86XK_LAUNCH9",
            "DOM_VK_XF86XK_LAUNCH_A",
            "DOM_VK_XF86XK_LAUNCH_B",
            "DOM_VK_XF86XK_LAUNCH_C",
            "DOM_VK_XF86XK_LAUNCH_D",
            "DOM_VK_XF86XK_LAUNCH_E",
            "DOM_VK_XF86XK_LAUNCH_F",
            "DOM_VK_XF86XK_LIGHT_BULB",
            "DOM_VK_XF86XK_LOG_OFF",
            "DOM_VK_XF86XK_MAIL",
            "DOM_VK_XF86XK_MAIL_FORWARD",
            "DOM_VK_XF86XK_MARKET",
            "DOM_VK_XF86XK_MEETING",
            "DOM_VK_XF86XK_MEMO",
            "DOM_VK_XF86XK_MENU_KB",
            "DOM_VK_XF86XK_MENU_PB",
            "DOM_VK_XF86XK_MESSENGER",
            "DOM_VK_XF86XK_MON_BRIGHTNESS_DOWN",
            "DOM_VK_XF86XK_MON_BRIGHTNESS_UP",
            "DOM_VK_XF86XK_MUSIC",
            "DOM_VK_XF86XK_MY_COMPUTER",
            "DOM_VK_XF86XK_MY_SITES",
            "DOM_VK_XF86XK_NEW",
            "DOM_VK_XF86XK_NEWS",
            "DOM_VK_XF86XK_OFFICE_HOME",
            "DOM_VK_XF86XK_OPEN",
            "DOM_VK_XF86XK_OPEN_URL",
            "DOM_VK_XF86XK_OPTION",
            "DOM_VK_XF86XK_PASTE",
            "DOM_VK_XF86XK_PHONE",
            "DOM_VK_XF86XK_PICTURES",
            "DOM_VK_XF86XK_POWER_DOWN",
            "DOM_VK_XF86XK_POWER_OFF",
            "DOM_VK_XF86XK_RED",
            "DOM_VK_XF86XK_REFRESH",
            "DOM_VK_XF86XK_RELOAD",
            "DOM_VK_XF86XK_REPLY",
            "DOM_VK_XF86XK_ROCKER_DOWN",
            "DOM_VK_XF86XK_ROCKER_ENTER",
            "DOM_VK_XF86XK_ROCKER_UP",
            "DOM_VK_XF86XK_ROTATE_WINDOWS",
            "DOM_VK_XF86XK_ROTATION_KB",
            "DOM_VK_XF86XK_ROTATION_PB",
            "DOM_VK_XF86XK_SAVE",
            "DOM_VK_XF86XK_SCREEN_SAVER",
            "DOM_VK_XF86XK_SCROLL_CLICK",
            "DOM_VK_XF86XK_SCROLL_DOWN",
            "DOM_VK_XF86XK_SCROLL_UP",
            "DOM_VK_XF86XK_SEARCH",
            "DOM_VK_XF86XK_SEND",
            "DOM_VK_XF86XK_SHOP",
            "DOM_VK_XF86XK_SPELL",
            "DOM_VK_XF86XK_SPLIT_SCREEN",
            "DOM_VK_XF86XK_STANDBY",
            "DOM_VK_XF86XK_START",
            "DOM_VK_XF86XK_STOP",
            "DOM_VK_XF86XK_SUBTITLE",
            "DOM_VK_XF86XK_SUPPORT",
            "DOM_VK_XF86XK_SUSPEND",
            "DOM_VK_XF86XK_TASK_PANE",
            "DOM_VK_XF86XK_TERMINAL",
            "DOM_VK_XF86XK_TIME",
            "DOM_VK_XF86XK_TOOLS",
            "DOM_VK_XF86XK_TOP_MENU",
            "DOM_VK_XF86XK_TO_DO_LIST",
            "DOM_VK_XF86XK_TRAVEL",
            "DOM_VK_XF86XK_USER1KB",
            "DOM_VK_XF86XK_USER2KB",
            "DOM_VK_XF86XK_USER_PB",
            "DOM_VK_XF86XK_UWB",
            "DOM_VK_XF86XK_VENDOR_HOME",
            "DOM_VK_XF86XK_VIDEO",
            "DOM_VK_XF86XK_VIEW",
            "DOM_VK_XF86XK_WAKE_UP",
            "DOM_VK_XF86XK_WEB_CAM",
            "DOM_VK_XF86XK_WHEEL_BUTTON",
            "DOM_VK_XF86XK_WLAN",
            "DOM_VK_XF86XK_WORD",
            "DOM_VK_XF86XK_WWW",
            "DOM_VK_XF86XK_XFER",
            "DOM_VK_XF86XK_YELLOW",
            "DOM_VK_XF86XK_ZOOM_IN",
            "DOM_VK_XF86XK_ZOOM_OUT",
            "DOM_VK_Y",
            "DOM_VK_Z",
            "DOM_VK_ZOOM",
            "DONE",
            "DONT_CARE",
            "DOWNLOADING",
            "DRAGDROP",
            "DRAW_BUFFER0",
            "DRAW_BUFFER1",
            "DRAW_BUFFER10",
            "DRAW_BUFFER11",
            "DRAW_BUFFER12",
            "DRAW_BUFFER13",
            "DRAW_BUFFER14",
            "DRAW_BUFFER15",
            "DRAW_BUFFER2",
            "DRAW_BUFFER3",
            "DRAW_BUFFER4",
            "DRAW_BUFFER5",
            "DRAW_BUFFER6",
            "DRAW_BUFFER7",
            "DRAW_BUFFER8",
            "DRAW_BUFFER9",
            "DRAW_FRAMEBUFFER",
            "DRAW_FRAMEBUFFER_BINDING",
            "DST_ALPHA",
            "DST_COLOR",
            "DYNAMIC_COPY",
            "DYNAMIC_DRAW",
            "DYNAMIC_READ",
            "DataChannel",
            "DataTransfer",
            "DataTransferItem",
            "DataTransferItemList",
            "DataView",
            "Date",
            "DateTimeFormat",
            "DecompressionStream",
            "DelayNode",
            "DeprecationReportBody",
            "DesktopNotification",
            "DesktopNotificationCenter",
            "DeviceLightEvent",
            "DeviceMotionEvent",
            "DeviceMotionEventAcceleration",
            "DeviceMotionEventRotationRate",
            "DeviceOrientationEvent",
            "DeviceProximityEvent",
            "DeviceStorage",
            "DeviceStorageChangeEvent",
            "Directory",
            "DisplayNames",
            "Document",
            "DocumentFragment",
            "DocumentTimeline",
            "DocumentType",
            "DragEvent",
            "DynamicsCompressorNode",
            "E",
            "ELEMENT_ARRAY_BUFFER",
            "ELEMENT_ARRAY_BUFFER_BINDING",
            "ELEMENT_NODE",
            "EMPTY",
            "ENCODING_ERR",
            "ENDED",
            "END_TO_END",
            "END_TO_START",
            "ENTITY_NODE",
            "ENTITY_REFERENCE_NODE",
            "EPSILON",
            "EQUAL",
            "EQUALPOWER",
            "ERROR",
            "EXPONENTIAL_DISTANCE",
            "Element",
            "ElementInternals",
            "ElementQuery",
            "EnterPictureInPictureEvent",
            "Entity",
            "EntityReference",
            "Error",
            "ErrorEvent",
            "EvalError",
            "Event",
            "EventException",
            "EventSource",
            "EventTarget",
            "External",
            "FASTEST",
            "FIDOSDK",
            "FILTER_ACCEPT",
            "FILTER_INTERRUPT",
            "FILTER_REJECT",
            "FILTER_SKIP",
            "FINISHED_STATE",
            "FIRST_ORDERED_NODE_TYPE",
            "FLOAT",
            "FLOAT_32_UNSIGNED_INT_24_8_REV",
            "FLOAT_MAT2",
            "FLOAT_MAT2x3",
            "FLOAT_MAT2x4",
            "FLOAT_MAT3",
            "FLOAT_MAT3x2",
            "FLOAT_MAT3x4",
            "FLOAT_MAT4",
            "FLOAT_MAT4x2",
            "FLOAT_MAT4x3",
            "FLOAT_VEC2",
            "FLOAT_VEC3",
            "FLOAT_VEC4",
            "FOCUS",
            "FONT_FACE_RULE",
            "FONT_FEATURE_VALUES_RULE",
            "FRAGMENT_SHADER",
            "FRAGMENT_SHADER_DERIVATIVE_HINT",
            "FRAGMENT_SHADER_DERIVATIVE_HINT_OES",
            "FRAMEBUFFER",
            "FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE",
            "FRAMEBUFFER_ATTACHMENT_BLUE_SIZE",
            "FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING",
            "FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE",
            "FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE",
            "FRAMEBUFFER_ATTACHMENT_GREEN_SIZE",
            "FRAMEBUFFER_ATTACHMENT_OBJECT_NAME",
            "FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE",
            "FRAMEBUFFER_ATTACHMENT_RED_SIZE",
            "FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE",
            "FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE",
            "FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER",
            "FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL",
            "FRAMEBUFFER_BINDING",
            "FRAMEBUFFER_COMPLETE",
            "FRAMEBUFFER_DEFAULT",
            "FRAMEBUFFER_INCOMPLETE_ATTACHMENT",
            "FRAMEBUFFER_INCOMPLETE_DIMENSIONS",
            "FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT",
            "FRAMEBUFFER_INCOMPLETE_MULTISAMPLE",
            "FRAMEBUFFER_UNSUPPORTED",
            "FRONT",
            "FRONT_AND_BACK",
            "FRONT_FACE",
            "FUNC_ADD",
            "FUNC_REVERSE_SUBTRACT",
            "FUNC_SUBTRACT",
            "FeaturePolicy",
            "FeaturePolicyViolationReportBody",
            "FederatedCredential",
            "Feed",
            "FeedEntry",
            "File",
            "FileError",
            "FileList",
            "FileReader",
            "FileSystem",
            "FileSystemDirectoryEntry",
            "FileSystemDirectoryReader",
            "FileSystemEntry",
            "FileSystemFileEntry",
            "FinalizationRegistry",
            "FindInPage",
            "Float32Array",
            "Float64Array",
            "FocusEvent",
            "FontFace",
            "FontFaceSet",
            "FontFaceSetLoadEvent",
            "FormData",
            "FormDataEvent",
            "FragmentDirective",
            "Function",
            "GENERATE_MIPMAP_HINT",
            "GEQUAL",
            "GREATER",
            "GREEN_BITS",
            "GainNode",
            "Gamepad",
            "GamepadAxisMoveEvent",
            "GamepadButton",
            "GamepadButtonEvent",
            "GamepadEvent",
            "GamepadHapticActuator",
            "GamepadPose",
            "Geolocation",
            "GeolocationCoordinates",
            "GeolocationPosition",
            "GeolocationPositionError",
            "GestureEvent",
            "Global",
            "Gyroscope",
            "HALF_FLOAT",
            "HAVE_CURRENT_DATA",
            "HAVE_ENOUGH_DATA",
            "HAVE_FUTURE_DATA",
            "HAVE_METADATA",
            "HAVE_NOTHING",
            "HEADERS_RECEIVED",
            "HIDDEN",
            "HIERARCHY_REQUEST_ERR",
            "HIGHPASS",
            "HIGHSHELF",
            "HIGH_FLOAT",
            "HIGH_INT",
            "HORIZONTAL",
            "HORIZONTAL_AXIS",
            "HRTF",
            "HTMLAllCollection",
            "HTMLAnchorElement",
            "HTMLAppletElement",
            "HTMLAreaElement",
            "HTMLAudioElement",
            "HTMLBRElement",
            "HTMLBaseElement",
            "HTMLBaseFontElement",
            "HTMLBlockquoteElement",
            "HTMLBodyElement",
            "HTMLButtonElement",
            "HTMLCanvasElement",
            "HTMLCollection",
            "HTMLCommandElement",
            "HTMLContentElement",
            "HTMLDListElement",
            "HTMLDataElement",
            "HTMLDataListElement",
            "HTMLDetailsElement",
            "HTMLDialogElement",
            "HTMLDirectoryElement",
            "HTMLDivElement",
            "HTMLDocument",
            "HTMLElement",
            "HTMLEmbedElement",
            "HTMLFieldSetElement",
            "HTMLFontElement",
            "HTMLFormControlsCollection",
            "HTMLFormElement",
            "HTMLFrameElement",
            "HTMLFrameSetElement",
            "HTMLHRElement",
            "HTMLHeadElement",
            "HTMLHeadingElement",
            "HTMLHtmlElement",
            "HTMLIFrameElement",
            "HTMLImageElement",
            "HTMLInputElement",
            "HTMLIsIndexElement",
            "HTMLKeygenElement",
            "HTMLLIElement",
            "HTMLLabelElement",
            "HTMLLegendElement",
            "HTMLLinkElement",
            "HTMLMapElement",
            "HTMLMarqueeElement",
            "HTMLMediaElement",
            "HTMLMenuElement",
            "HTMLMenuItemElement",
            "HTMLMetaElement",
            "HTMLMeterElement",
            "HTMLModElement",
            "HTMLOListElement",
            "HTMLObjectElement",
            "HTMLOptGroupElement",
            "HTMLOptionElement",
            "HTMLOptionsCollection",
            "HTMLOutputElement",
            "HTMLParagraphElement",
            "HTMLParamElement",
            "HTMLPictureElement",
            "HTMLPreElement",
            "HTMLProgressElement",
            "HTMLPropertiesCollection",
            "HTMLQuoteElement",
            "HTMLScriptElement",
            "HTMLSelectElement",
            "HTMLShadowElement",
            "HTMLSlotElement",
            "HTMLSourceElement",
            "HTMLSpanElement",
            "HTMLStyleElement",
            "HTMLTableCaptionElement",
            "HTMLTableCellElement",
            "HTMLTableColElement",
            "HTMLTableElement",
            "HTMLTableRowElement",
            "HTMLTableSectionElement",
            "HTMLTemplateElement",
            "HTMLTextAreaElement",
            "HTMLTimeElement",
            "HTMLTitleElement",
            "HTMLTrackElement",
            "HTMLUListElement",
            "HTMLUnknownElement",
            "HTMLVideoElement",
            "HashChangeEvent",
            "Headers",
            "History",
            "Hz",
            "ICE_CHECKING",
            "ICE_CLOSED",
            "ICE_COMPLETED",
            "ICE_CONNECTED",
            "ICE_FAILED",
            "ICE_GATHERING",
            "ICE_WAITING",
            "IDBCursor",
            "IDBCursorWithValue",
            "IDBDatabase",
            "IDBDatabaseException",
            "IDBFactory",
            "IDBFileHandle",
            "IDBFileRequest",
            "IDBIndex",
            "IDBKeyRange",
            "IDBMutableFile",
            "IDBObjectStore",
            "IDBOpenDBRequest",
            "IDBRequest",
            "IDBTransaction",
            "IDBVersionChangeEvent",
            "IDLE",
            "IIRFilterNode",
            "IMPLEMENTATION_COLOR_READ_FORMAT",
            "IMPLEMENTATION_COLOR_READ_TYPE",
            "IMPORT_RULE",
            "INCR",
            "INCR_WRAP",
            "INDEX_SIZE_ERR",
            "INT",
            "INTERLEAVED_ATTRIBS",
            "INT_2_10_10_10_REV",
            "INT_SAMPLER_2D",
            "INT_SAMPLER_2D_ARRAY",
            "INT_SAMPLER_3D",
            "INT_SAMPLER_CUBE",
            "INT_VEC2",
            "INT_VEC3",
            "INT_VEC4",
            "INUSE_ATTRIBUTE_ERR",
            "INVALID_ACCESS_ERR",
            "INVALID_CHARACTER_ERR",
            "INVALID_ENUM",
            "INVALID_EXPRESSION_ERR",
            "INVALID_FRAMEBUFFER_OPERATION",
            "INVALID_INDEX",
            "INVALID_MODIFICATION_ERR",
            "INVALID_NODE_TYPE_ERR",
            "INVALID_OPERATION",
            "INVALID_STATE_ERR",
            "INVALID_VALUE",
            "INVERSE_DISTANCE",
            "INVERT",
            "IceCandidate",
            "IdleDeadline",
            "Image",
            "ImageBitmap",
            "ImageBitmapRenderingContext",
            "ImageCapture",
            "ImageData",
            "Infinity",
            "InputDeviceCapabilities",
            "InputDeviceInfo",
            "InputEvent",
            "InputMethodContext",
            "InstallTrigger",
            "InstallTriggerImpl",
            "Instance",
            "Int16Array",
            "Int32Array",
            "Int8Array",
            "Intent",
            "InternalError",
            "IntersectionObserver",
            "IntersectionObserverEntry",
            "Intl",
            "IsSearchProviderInstalled",
            "Iterator",
            "JSON",
            "KEEP",
            "KEYDOWN",
            "KEYFRAMES_RULE",
            "KEYFRAME_RULE",
            "KEYPRESS",
            "KEYUP",
            "KeyEvent",
            "Keyboard",
            "KeyboardEvent",
            "KeyboardLayoutMap",
            "KeyframeEffect",
            "LENGTHADJUST_SPACING",
            "LENGTHADJUST_SPACINGANDGLYPHS",
            "LENGTHADJUST_UNKNOWN",
            "LEQUAL",
            "LESS",
            "LINEAR",
            "LINEAR_DISTANCE",
            "LINEAR_MIPMAP_LINEAR",
            "LINEAR_MIPMAP_NEAREST",
            "LINES",
            "LINE_LOOP",
            "LINE_STRIP",
            "LINE_WIDTH",
            "LINK_STATUS",
            "LIVE",
            "LN10",
            "LN2",
            "LOADED",
            "LOADING",
            "LOG10E",
            "LOG2E",
            "LOWPASS",
            "LOWSHELF",
            "LOW_FLOAT",
            "LOW_INT",
            "LSException",
            "LSParserFilter",
            "LUMINANCE",
            "LUMINANCE_ALPHA",
            "LargestContentfulPaint",
            "LayoutShift",
            "LayoutShiftAttribution",
            "LinearAccelerationSensor",
            "LinkError",
            "ListFormat",
            "LocalMediaStream",
            "Locale",
            "Location",
            "Lock",
            "LockManager",
            "MAX",
            "MAX_3D_TEXTURE_SIZE",
            "MAX_ARRAY_TEXTURE_LAYERS",
            "MAX_CLIENT_WAIT_TIMEOUT_WEBGL",
            "MAX_COLOR_ATTACHMENTS",
            "MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS",
            "MAX_COMBINED_TEXTURE_IMAGE_UNITS",
            "MAX_COMBINED_UNIFORM_BLOCKS",
            "MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS",
            "MAX_CUBE_MAP_TEXTURE_SIZE",
            "MAX_DRAW_BUFFERS",
            "MAX_ELEMENTS_INDICES",
            "MAX_ELEMENTS_VERTICES",
            "MAX_ELEMENT_INDEX",
            "MAX_FRAGMENT_INPUT_COMPONENTS",
            "MAX_FRAGMENT_UNIFORM_BLOCKS",
            "MAX_FRAGMENT_UNIFORM_COMPONENTS",
            "MAX_FRAGMENT_UNIFORM_VECTORS",
            "MAX_PROGRAM_TEXEL_OFFSET",
            "MAX_RENDERBUFFER_SIZE",
            "MAX_SAFE_INTEGER",
            "MAX_SAMPLES",
            "MAX_SERVER_WAIT_TIMEOUT",
            "MAX_TEXTURE_IMAGE_UNITS",
            "MAX_TEXTURE_LOD_BIAS",
            "MAX_TEXTURE_MAX_ANISOTROPY_EXT",
            "MAX_TEXTURE_SIZE",
            "MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS",
            "MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS",
            "MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS",
            "MAX_UNIFORM_BLOCK_SIZE",
            "MAX_UNIFORM_BUFFER_BINDINGS",
            "MAX_VALUE",
            "MAX_VARYING_COMPONENTS",
            "MAX_VARYING_VECTORS",
            "MAX_VERTEX_ATTRIBS",
            "MAX_VERTEX_OUTPUT_COMPONENTS",
            "MAX_VERTEX_TEXTURE_IMAGE_UNITS",
            "MAX_VERTEX_UNIFORM_BLOCKS",
            "MAX_VERTEX_UNIFORM_COMPONENTS",
            "MAX_VERTEX_UNIFORM_VECTORS",
            "MAX_VIEWPORT_DIMS",
            "MEDIA_ERR_ABORTED",
            "MEDIA_ERR_DECODE",
            "MEDIA_ERR_ENCRYPTED",
            "MEDIA_ERR_NETWORK",
            "MEDIA_ERR_SRC_NOT_SUPPORTED",
            "MEDIA_KEYERR_CLIENT",
            "MEDIA_KEYERR_DOMAIN",
            "MEDIA_KEYERR_HARDWARECHANGE",
            "MEDIA_KEYERR_OUTPUT",
            "MEDIA_KEYERR_SERVICE",
            "MEDIA_KEYERR_UNKNOWN",
            "MEDIA_RULE",
            "MEDIUM_FLOAT",
            "MEDIUM_INT",
            "META_MASK",
            "MIDIAccess",
            "MIDIConnectionEvent",
            "MIDIInput",
            "MIDIInputMap",
            "MIDIMessageEvent",
            "MIDIOutput",
            "MIDIOutputMap",
            "MIDIPort",
            "MIN",
            "MIN_PROGRAM_TEXEL_OFFSET",
            "MIN_SAFE_INTEGER",
            "MIN_VALUE",
            "MIRRORED_REPEAT",
            "MODE_ASYNCHRONOUS",
            "MODE_SYNCHRONOUS",
            "MODIFICATION",
            "MOUSEDOWN",
            "MOUSEDRAG",
            "MOUSEMOVE",
            "MOUSEOUT",
            "MOUSEOVER",
            "MOUSEUP",
            "MOZ_KEYFRAMES_RULE",
            "MOZ_KEYFRAME_RULE",
            "MOZ_SOURCE_CURSOR",
            "MOZ_SOURCE_ERASER",
            "MOZ_SOURCE_KEYBOARD",
            "MOZ_SOURCE_MOUSE",
            "MOZ_SOURCE_PEN",
            "MOZ_SOURCE_TOUCH",
            "MOZ_SOURCE_UNKNOWN",
            "MSGESTURE_FLAG_BEGIN",
            "MSGESTURE_FLAG_CANCEL",
            "MSGESTURE_FLAG_END",
            "MSGESTURE_FLAG_INERTIA",
            "MSGESTURE_FLAG_NONE",
            "MSPOINTER_TYPE_MOUSE",
            "MSPOINTER_TYPE_PEN",
            "MSPOINTER_TYPE_TOUCH",
            "MS_ASYNC_CALLBACK_STATUS_ASSIGN_DELEGATE",
            "MS_ASYNC_CALLBACK_STATUS_CANCEL",
            "MS_ASYNC_CALLBACK_STATUS_CHOOSEANY",
            "MS_ASYNC_CALLBACK_STATUS_ERROR",
            "MS_ASYNC_CALLBACK_STATUS_JOIN",
            "MS_ASYNC_OP_STATUS_CANCELED",
            "MS_ASYNC_OP_STATUS_ERROR",
            "MS_ASYNC_OP_STATUS_SUCCESS",
            "MS_MANIPULATION_STATE_ACTIVE",
            "MS_MANIPULATION_STATE_CANCELLED",
            "MS_MANIPULATION_STATE_COMMITTED",
            "MS_MANIPULATION_STATE_DRAGGING",
            "MS_MANIPULATION_STATE_INERTIA",
            "MS_MANIPULATION_STATE_PRESELECT",
            "MS_MANIPULATION_STATE_SELECTING",
            "MS_MANIPULATION_STATE_STOPPED",
            "MS_MEDIA_ERR_ENCRYPTED",
            "MS_MEDIA_KEYERR_CLIENT",
            "MS_MEDIA_KEYERR_DOMAIN",
            "MS_MEDIA_KEYERR_HARDWARECHANGE",
            "MS_MEDIA_KEYERR_OUTPUT",
            "MS_MEDIA_KEYERR_SERVICE",
            "MS_MEDIA_KEYERR_UNKNOWN",
            "Map",
            "Math",
            "MathMLElement",
            "MediaCapabilities",
            "MediaCapabilitiesInfo",
            "MediaController",
            "MediaDeviceInfo",
            "MediaDevices",
            "MediaElementAudioSourceNode",
            "MediaEncryptedEvent",
            "MediaError",
            "MediaKeyError",
            "MediaKeyEvent",
            "MediaKeyMessageEvent",
            "MediaKeyNeededEvent",
            "MediaKeySession",
            "MediaKeyStatusMap",
            "MediaKeySystemAccess",
            "MediaKeys",
            "MediaList",
            "MediaMetadata",
            "MediaQueryList",
            "MediaQueryListEvent",
            "MediaRecorder",
            "MediaRecorderErrorEvent",
            "MediaSession",
            "MediaSettingsRange",
            "MediaSource",
            "MediaStream",
            "MediaStreamAudioDestinationNode",
            "MediaStreamAudioSourceNode",
            "MediaStreamEvent",
            "MediaStreamTrack",
            "MediaStreamTrackAudioSourceNode",
            "MediaStreamTrackEvent",
            "Memory",
            "MessageChannel",
            "MessageEvent",
            "MessagePort",
            "Methods",
            "MimeType",
            "MimeTypeArray",
            "Module",
            "MouseEvent",
            "MouseScrollEvent",
            "MozAnimation",
            "MozAnimationDelay",
            "MozAnimationDirection",
            "MozAnimationDuration",
            "MozAnimationFillMode",
            "MozAnimationIterationCount",
            "MozAnimationName",
            "MozAnimationPlayState",
            "MozAnimationTimingFunction",
            "MozAppearance",
            "MozBackfaceVisibility",
            "MozBinding",
            "MozBorderBottomColors",
            "MozBorderEnd",
            "MozBorderEndColor",
            "MozBorderEndStyle",
            "MozBorderEndWidth",
            "MozBorderImage",
            "MozBorderLeftColors",
            "MozBorderRightColors",
            "MozBorderStart",
            "MozBorderStartColor",
            "MozBorderStartStyle",
            "MozBorderStartWidth",
            "MozBorderTopColors",
            "MozBoxAlign",
            "MozBoxDirection",
            "MozBoxFlex",
            "MozBoxOrdinalGroup",
            "MozBoxOrient",
            "MozBoxPack",
            "MozBoxSizing",
            "MozCSSKeyframeRule",
            "MozCSSKeyframesRule",
            "MozColumnCount",
            "MozColumnFill",
            "MozColumnGap",
            "MozColumnRule",
            "MozColumnRuleColor",
            "MozColumnRuleStyle",
            "MozColumnRuleWidth",
            "MozColumnWidth",
            "MozColumns",
            "MozContactChangeEvent",
            "MozFloatEdge",
            "MozFontFeatureSettings",
            "MozFontLanguageOverride",
            "MozForceBrokenImageIcon",
            "MozHyphens",
            "MozImageRegion",
            "MozMarginEnd",
            "MozMarginStart",
            "MozMmsEvent",
            "MozMmsMessage",
            "MozMobileMessageThread",
            "MozOSXFontSmoothing",
            "MozOrient",
            "MozOsxFontSmoothing",
            "MozOutlineRadius",
            "MozOutlineRadiusBottomleft",
            "MozOutlineRadiusBottomright",
            "MozOutlineRadiusTopleft",
            "MozOutlineRadiusTopright",
            "MozPaddingEnd",
            "MozPaddingStart",
            "MozPerspective",
            "MozPerspectiveOrigin",
            "MozPowerManager",
            "MozSettingsEvent",
            "MozSmsEvent",
            "MozSmsMessage",
            "MozStackSizing",
            "MozTabSize",
            "MozTextAlignLast",
            "MozTextDecorationColor",
            "MozTextDecorationLine",
            "MozTextDecorationStyle",
            "MozTextSizeAdjust",
            "MozTransform",
            "MozTransformOrigin",
            "MozTransformStyle",
            "MozTransition",
            "MozTransitionDelay",
            "MozTransitionDuration",
            "MozTransitionProperty",
            "MozTransitionTimingFunction",
            "MozUserFocus",
            "MozUserInput",
            "MozUserModify",
            "MozUserSelect",
            "MozWindowDragging",
            "MozWindowShadow",
            "MutationEvent",
            "MutationObserver",
            "MutationRecord",
            "NAMESPACE_ERR",
            "NAMESPACE_RULE",
            "NEAREST",
            "NEAREST_MIPMAP_LINEAR",
            "NEAREST_MIPMAP_NEAREST",
            "NEGATIVE_INFINITY",
            "NETWORK_EMPTY",
            "NETWORK_ERR",
            "NETWORK_IDLE",
            "NETWORK_LOADED",
            "NETWORK_LOADING",
            "NETWORK_NO_SOURCE",
            "NEVER",
            "NEW",
            "NEXT",
            "NEXT_NO_DUPLICATE",
            "NICEST",
            "NODE_AFTER",
            "NODE_BEFORE",
            "NODE_BEFORE_AND_AFTER",
            "NODE_INSIDE",
            "NONE",
            "NON_TRANSIENT_ERR",
            "NOTATION_NODE",
            "NOTCH",
            "NOTEQUAL",
            "NOT_ALLOWED_ERR",
            "NOT_FOUND_ERR",
            "NOT_READABLE_ERR",
            "NOT_SUPPORTED_ERR",
            "NO_DATA_ALLOWED_ERR",
            "NO_ERR",
            "NO_ERROR",
            "NO_MODIFICATION_ALLOWED_ERR",
            "NUMBER_TYPE",
            "NUM_COMPRESSED_TEXTURE_FORMATS",
            "NaN",
            "NamedNodeMap",
            "NavigationPreloadManager",
            "Navigator",
            "NearbyLinks",
            "NetworkInformation",
            "Node",
            "NodeFilter",
            "NodeIterator",
            "NodeList",
            "Notation",
            "Notification",
            "NotifyPaintEvent",
            "Number",
            "NumberFormat",
            "OBJECT_TYPE",
            "OBSOLETE",
            "OK",
            "ONE",
            "ONE_MINUS_CONSTANT_ALPHA",
            "ONE_MINUS_CONSTANT_COLOR",
            "ONE_MINUS_DST_ALPHA",
            "ONE_MINUS_DST_COLOR",
            "ONE_MINUS_SRC_ALPHA",
            "ONE_MINUS_SRC_COLOR",
            "OPEN",
            "OPENED",
            "OPENING",
            "ORDERED_NODE_ITERATOR_TYPE",
            "ORDERED_NODE_SNAPSHOT_TYPE",
            "OTHER_ERROR",
            "OUT_OF_MEMORY",
            "Object",
            "OfflineAudioCompletionEvent",
            "OfflineAudioContext",
            "OfflineResourceList",
            "OffscreenCanvas",
            "OffscreenCanvasRenderingContext2D",
            "Option",
            "OrientationSensor",
            "OscillatorNode",
            "OverconstrainedError",
            "OverflowEvent",
            "PACK_ALIGNMENT",
            "PACK_ROW_LENGTH",
            "PACK_SKIP_PIXELS",
            "PACK_SKIP_ROWS",
            "PAGE_RULE",
            "PARSE_ERR",
            "PATHSEG_ARC_ABS",
            "PATHSEG_ARC_REL",
            "PATHSEG_CLOSEPATH",
            "PATHSEG_CURVETO_CUBIC_ABS",
            "PATHSEG_CURVETO_CUBIC_REL",
            "PATHSEG_CURVETO_CUBIC_SMOOTH_ABS",
            "PATHSEG_CURVETO_CUBIC_SMOOTH_REL",
            "PATHSEG_CURVETO_QUADRATIC_ABS",
            "PATHSEG_CURVETO_QUADRATIC_REL",
            "PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS",
            "PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL",
            "PATHSEG_LINETO_ABS",
            "PATHSEG_LINETO_HORIZONTAL_ABS",
            "PATHSEG_LINETO_HORIZONTAL_REL",
            "PATHSEG_LINETO_REL",
            "PATHSEG_LINETO_VERTICAL_ABS",
            "PATHSEG_LINETO_VERTICAL_REL",
            "PATHSEG_MOVETO_ABS",
            "PATHSEG_MOVETO_REL",
            "PATHSEG_UNKNOWN",
            "PATH_EXISTS_ERR",
            "PEAKING",
            "PERMISSION_DENIED",
            "PERSISTENT",
            "PI",
            "PIXEL_PACK_BUFFER",
            "PIXEL_PACK_BUFFER_BINDING",
            "PIXEL_UNPACK_BUFFER",
            "PIXEL_UNPACK_BUFFER_BINDING",
            "PLAYING_STATE",
            "POINTS",
            "POLYGON_OFFSET_FACTOR",
            "POLYGON_OFFSET_FILL",
            "POLYGON_OFFSET_UNITS",
            "POSITION_UNAVAILABLE",
            "POSITIVE_INFINITY",
            "PREV",
            "PREV_NO_DUPLICATE",
            "PROCESSING_INSTRUCTION_NODE",
            "PageChangeEvent",
            "PageTransitionEvent",
            "PaintRequest",
            "PaintRequestList",
            "PannerNode",
            "PasswordCredential",
            "Path2D",
            "PaymentAddress",
            "PaymentInstruments",
            "PaymentManager",
            "PaymentMethodChangeEvent",
            "PaymentRequest",
            "PaymentRequestUpdateEvent",
            "PaymentResponse",
            "Performance",
            "PerformanceElementTiming",
            "PerformanceEntry",
            "PerformanceEventTiming",
            "PerformanceLongTaskTiming",
            "PerformanceMark",
            "PerformanceMeasure",
            "PerformanceNavigation",
            "PerformanceNavigationTiming",
            "PerformanceObserver",
            "PerformanceObserverEntryList",
            "PerformancePaintTiming",
            "PerformanceResourceTiming",
            "PerformanceServerTiming",
            "PerformanceTiming",
            "PeriodicSyncManager",
            "PeriodicWave",
            "PermissionStatus",
            "Permissions",
            "PhotoCapabilities",
            "PictureInPictureWindow",
            "Plugin",
            "PluginArray",
            "PluralRules",
            "PointerEvent",
            "PopStateEvent",
            "PopupBlockedEvent",
            "Presentation",
            "PresentationAvailability",
            "PresentationConnection",
            "PresentationConnectionAvailableEvent",
            "PresentationConnectionCloseEvent",
            "PresentationConnectionList",
            "PresentationReceiver",
            "PresentationRequest",
            "ProcessingInstruction",
            "ProgressEvent",
            "Promise",
            "PromiseRejectionEvent",
            "PropertyNodeList",
            "Proxy",
            "PublicKeyCredential",
            "PushManager",
            "PushSubscription",
            "PushSubscriptionOptions",
            "Q",
            "QUERY_RESULT",
            "QUERY_RESULT_AVAILABLE",
            "QUOTA_ERR",
            "QUOTA_EXCEEDED_ERR",
            "QueryInterface",
            "R11F_G11F_B10F",
            "R16F",
            "R16I",
            "R16UI",
            "R32F",
            "R32I",
            "R32UI",
            "R8",
            "R8I",
            "R8UI",
            "R8_SNORM",
            "RASTERIZER_DISCARD",
            "READ_BUFFER",
            "READ_FRAMEBUFFER",
            "READ_FRAMEBUFFER_BINDING",
            "READ_ONLY",
            "READ_ONLY_ERR",
            "READ_WRITE",
            "RED",
            "RED_BITS",
            "RED_INTEGER",
            "REMOVAL",
            "RENDERBUFFER",
            "RENDERBUFFER_ALPHA_SIZE",
            "RENDERBUFFER_BINDING",
            "RENDERBUFFER_BLUE_SIZE",
            "RENDERBUFFER_DEPTH_SIZE",
            "RENDERBUFFER_GREEN_SIZE",
            "RENDERBUFFER_HEIGHT",
            "RENDERBUFFER_INTERNAL_FORMAT",
            "RENDERBUFFER_RED_SIZE",
            "RENDERBUFFER_SAMPLES",
            "RENDERBUFFER_STENCIL_SIZE",
            "RENDERBUFFER_WIDTH",
            "RENDERER",
            "RENDERING_INTENT_ABSOLUTE_COLORIMETRIC",
            "RENDERING_INTENT_AUTO",
            "RENDERING_INTENT_PERCEPTUAL",
            "RENDERING_INTENT_RELATIVE_COLORIMETRIC",
            "RENDERING_INTENT_SATURATION",
            "RENDERING_INTENT_UNKNOWN",
            "REPEAT",
            "REPLACE",
            "RG",
            "RG16F",
            "RG16I",
            "RG16UI",
            "RG32F",
            "RG32I",
            "RG32UI",
            "RG8",
            "RG8I",
            "RG8UI",
            "RG8_SNORM",
            "RGB",
            "RGB10_A2",
            "RGB10_A2UI",
            "RGB16F",
            "RGB16I",
            "RGB16UI",
            "RGB32F",
            "RGB32I",
            "RGB32UI",
            "RGB565",
            "RGB5_A1",
            "RGB8",
            "RGB8I",
            "RGB8UI",
            "RGB8_SNORM",
            "RGB9_E5",
            "RGBA",
            "RGBA16F",
            "RGBA16I",
            "RGBA16UI",
            "RGBA32F",
            "RGBA32I",
            "RGBA32UI",
            "RGBA4",
            "RGBA8",
            "RGBA8I",
            "RGBA8UI",
            "RGBA8_SNORM",
            "RGBA_INTEGER",
            "RGBColor",
            "RGB_INTEGER",
            "RG_INTEGER",
            "ROTATION_CLOCKWISE",
            "ROTATION_COUNTERCLOCKWISE",
            "RTCCertificate",
            "RTCDTMFSender",
            "RTCDTMFToneChangeEvent",
            "RTCDataChannel",
            "RTCDataChannelEvent",
            "RTCDtlsTransport",
            "RTCError",
            "RTCErrorEvent",
            "RTCIceCandidate",
            "RTCIceTransport",
            "RTCPeerConnection",
            "RTCPeerConnectionIceErrorEvent",
            "RTCPeerConnectionIceEvent",
            "RTCRtpReceiver",
            "RTCRtpSender",
            "RTCRtpTransceiver",
            "RTCSctpTransport",
            "RTCSessionDescription",
            "RTCStatsReport",
            "RTCTrackEvent",
            "RadioNodeList",
            "Range",
            "RangeError",
            "RangeException",
            "ReadableStream",
            "ReadableStreamDefaultReader",
            "RecordErrorEvent",
            "Rect",
            "ReferenceError",
            "Reflect",
            "RegExp",
            "RelativeOrientationSensor",
            "RelativeTimeFormat",
            "RemotePlayback",
            "Report",
            "ReportBody",
            "ReportingObserver",
            "Request",
            "ResizeObserver",
            "ResizeObserverEntry",
            "ResizeObserverSize",
            "Response",
            "RuntimeError",
            "SAMPLER_2D",
            "SAMPLER_2D_ARRAY",
            "SAMPLER_2D_ARRAY_SHADOW",
            "SAMPLER_2D_SHADOW",
            "SAMPLER_3D",
            "SAMPLER_BINDING",
            "SAMPLER_CUBE",
            "SAMPLER_CUBE_SHADOW",
            "SAMPLES",
            "SAMPLE_ALPHA_TO_COVERAGE",
            "SAMPLE_BUFFERS",
            "SAMPLE_COVERAGE",
            "SAMPLE_COVERAGE_INVERT",
            "SAMPLE_COVERAGE_VALUE",
            "SAWTOOTH",
            "SCHEDULED_STATE",
            "SCISSOR_BOX",
            "SCISSOR_TEST",
            "SCROLL_PAGE_DOWN",
            "SCROLL_PAGE_UP",
            "SDP_ANSWER",
            "SDP_OFFER",
            "SDP_PRANSWER",
            "SECURITY_ERR",
            "SELECT",
            "SEPARATE_ATTRIBS",
            "SERIALIZE_ERR",
            "SEVERITY_ERROR",
            "SEVERITY_FATAL_ERROR",
            "SEVERITY_WARNING",
            "SHADER_COMPILER",
            "SHADER_TYPE",
            "SHADING_LANGUAGE_VERSION",
            "SHIFT_MASK",
            "SHORT",
            "SHOWING",
            "SHOW_ALL",
            "SHOW_ATTRIBUTE",
            "SHOW_CDATA_SECTION",
            "SHOW_COMMENT",
            "SHOW_DOCUMENT",
            "SHOW_DOCUMENT_FRAGMENT",
            "SHOW_DOCUMENT_TYPE",
            "SHOW_ELEMENT",
            "SHOW_ENTITY",
            "SHOW_ENTITY_REFERENCE",
            "SHOW_NOTATION",
            "SHOW_PROCESSING_INSTRUCTION",
            "SHOW_TEXT",
            "SIGNALED",
            "SIGNED_NORMALIZED",
            "SINE",
            "SOUNDFIELD",
            "SQLException",
            "SQRT1_2",
            "SQRT2",
            "SQUARE",
            "SRC_ALPHA",
            "SRC_ALPHA_SATURATE",
            "SRC_COLOR",
            "SRGB",
            "SRGB8",
            "SRGB8_ALPHA8",
            "START_TO_END",
            "START_TO_START",
            "STATIC_COPY",
            "STATIC_DRAW",
            "STATIC_READ",
            "STENCIL",
            "STENCIL_ATTACHMENT",
            "STENCIL_BACK_FAIL",
            "STENCIL_BACK_FUNC",
            "STENCIL_BACK_PASS_DEPTH_FAIL",
            "STENCIL_BACK_PASS_DEPTH_PASS",
            "STENCIL_BACK_REF",
            "STENCIL_BACK_VALUE_MASK",
            "STENCIL_BACK_WRITEMASK",
            "STENCIL_BITS",
            "STENCIL_BUFFER_BIT",
            "STENCIL_CLEAR_VALUE",
            "STENCIL_FAIL",
            "STENCIL_FUNC",
            "STENCIL_INDEX",
            "STENCIL_INDEX8",
            "STENCIL_PASS_DEPTH_FAIL",
            "STENCIL_PASS_DEPTH_PASS",
            "STENCIL_REF",
            "STENCIL_TEST",
            "STENCIL_VALUE_MASK",
            "STENCIL_WRITEMASK",
            "STREAM_COPY",
            "STREAM_DRAW",
            "STREAM_READ",
            "STRING_TYPE",
            "STYLE_RULE",
            "SUBPIXEL_BITS",
            "SUPPORTS_RULE",
            "SVGAElement",
            "SVGAltGlyphDefElement",
            "SVGAltGlyphElement",
            "SVGAltGlyphItemElement",
            "SVGAngle",
            "SVGAnimateColorElement",
            "SVGAnimateElement",
            "SVGAnimateMotionElement",
            "SVGAnimateTransformElement",
            "SVGAnimatedAngle",
            "SVGAnimatedBoolean",
            "SVGAnimatedEnumeration",
            "SVGAnimatedInteger",
            "SVGAnimatedLength",
            "SVGAnimatedLengthList",
            "SVGAnimatedNumber",
            "SVGAnimatedNumberList",
            "SVGAnimatedPreserveAspectRatio",
            "SVGAnimatedRect",
            "SVGAnimatedString",
            "SVGAnimatedTransformList",
            "SVGAnimationElement",
            "SVGCircleElement",
            "SVGClipPathElement",
            "SVGColor",
            "SVGComponentTransferFunctionElement",
            "SVGCursorElement",
            "SVGDefsElement",
            "SVGDescElement",
            "SVGDiscardElement",
            "SVGDocument",
            "SVGElement",
            "SVGElementInstance",
            "SVGElementInstanceList",
            "SVGEllipseElement",
            "SVGException",
            "SVGFEBlendElement",
            "SVGFEColorMatrixElement",
            "SVGFEComponentTransferElement",
            "SVGFECompositeElement",
            "SVGFEConvolveMatrixElement",
            "SVGFEDiffuseLightingElement",
            "SVGFEDisplacementMapElement",
            "SVGFEDistantLightElement",
            "SVGFEDropShadowElement",
            "SVGFEFloodElement",
            "SVGFEFuncAElement",
            "SVGFEFuncBElement",
            "SVGFEFuncGElement",
            "SVGFEFuncRElement",
            "SVGFEGaussianBlurElement",
            "SVGFEImageElement",
            "SVGFEMergeElement",
            "SVGFEMergeNodeElement",
            "SVGFEMorphologyElement",
            "SVGFEOffsetElement",
            "SVGFEPointLightElement",
            "SVGFESpecularLightingElement",
            "SVGFESpotLightElement",
            "SVGFETileElement",
            "SVGFETurbulenceElement",
            "SVGFilterElement",
            "SVGFontElement",
            "SVGFontFaceElement",
            "SVGFontFaceFormatElement",
            "SVGFontFaceNameElement",
            "SVGFontFaceSrcElement",
            "SVGFontFaceUriElement",
            "SVGForeignObjectElement",
            "SVGGElement",
            "SVGGeometryElement",
            "SVGGlyphElement",
            "SVGGlyphRefElement",
            "SVGGradientElement",
            "SVGGraphicsElement",
            "SVGHKernElement",
            "SVGImageElement",
            "SVGLength",
            "SVGLengthList",
            "SVGLineElement",
            "SVGLinearGradientElement",
            "SVGMPathElement",
            "SVGMarkerElement",
            "SVGMaskElement",
            "SVGMatrix",
            "SVGMetadataElement",
            "SVGMissingGlyphElement",
            "SVGNumber",
            "SVGNumberList",
            "SVGPaint",
            "SVGPathElement",
            "SVGPathSeg",
            "SVGPathSegArcAbs",
            "SVGPathSegArcRel",
            "SVGPathSegClosePath",
            "SVGPathSegCurvetoCubicAbs",
            "SVGPathSegCurvetoCubicRel",
            "SVGPathSegCurvetoCubicSmoothAbs",
            "SVGPathSegCurvetoCubicSmoothRel",
            "SVGPathSegCurvetoQuadraticAbs",
            "SVGPathSegCurvetoQuadraticRel",
            "SVGPathSegCurvetoQuadraticSmoothAbs",
            "SVGPathSegCurvetoQuadraticSmoothRel",
            "SVGPathSegLinetoAbs",
            "SVGPathSegLinetoHorizontalAbs",
            "SVGPathSegLinetoHorizontalRel",
            "SVGPathSegLinetoRel",
            "SVGPathSegLinetoVerticalAbs",
            "SVGPathSegLinetoVerticalRel",
            "SVGPathSegList",
            "SVGPathSegMovetoAbs",
            "SVGPathSegMovetoRel",
            "SVGPatternElement",
            "SVGPoint",
            "SVGPointList",
            "SVGPolygonElement",
            "SVGPolylineElement",
            "SVGPreserveAspectRatio",
            "SVGRadialGradientElement",
            "SVGRect",
            "SVGRectElement",
            "SVGRenderingIntent",
            "SVGSVGElement",
            "SVGScriptElement",
            "SVGSetElement",
            "SVGStopElement",
            "SVGStringList",
            "SVGStyleElement",
            "SVGSwitchElement",
            "SVGSymbolElement",
            "SVGTRefElement",
            "SVGTSpanElement",
            "SVGTextContentElement",
            "SVGTextElement",
            "SVGTextPathElement",
            "SVGTextPositioningElement",
            "SVGTitleElement",
            "SVGTransform",
            "SVGTransformList",
            "SVGUnitTypes",
            "SVGUseElement",
            "SVGVKernElement",
            "SVGViewElement",
            "SVGViewSpec",
            "SVGZoomAndPan",
            "SVGZoomEvent",
            "SVG_ANGLETYPE_DEG",
            "SVG_ANGLETYPE_GRAD",
            "SVG_ANGLETYPE_RAD",
            "SVG_ANGLETYPE_UNKNOWN",
            "SVG_ANGLETYPE_UNSPECIFIED",
            "SVG_CHANNEL_A",
            "SVG_CHANNEL_B",
            "SVG_CHANNEL_G",
            "SVG_CHANNEL_R",
            "SVG_CHANNEL_UNKNOWN",
            "SVG_COLORTYPE_CURRENTCOLOR",
            "SVG_COLORTYPE_RGBCOLOR",
            "SVG_COLORTYPE_RGBCOLOR_ICCCOLOR",
            "SVG_COLORTYPE_UNKNOWN",
            "SVG_EDGEMODE_DUPLICATE",
            "SVG_EDGEMODE_NONE",
            "SVG_EDGEMODE_UNKNOWN",
            "SVG_EDGEMODE_WRAP",
            "SVG_FEBLEND_MODE_COLOR",
            "SVG_FEBLEND_MODE_COLOR_BURN",
            "SVG_FEBLEND_MODE_COLOR_DODGE",
            "SVG_FEBLEND_MODE_DARKEN",
            "SVG_FEBLEND_MODE_DIFFERENCE",
            "SVG_FEBLEND_MODE_EXCLUSION",
            "SVG_FEBLEND_MODE_HARD_LIGHT",
            "SVG_FEBLEND_MODE_HUE",
            "SVG_FEBLEND_MODE_LIGHTEN",
            "SVG_FEBLEND_MODE_LUMINOSITY",
            "SVG_FEBLEND_MODE_MULTIPLY",
            "SVG_FEBLEND_MODE_NORMAL",
            "SVG_FEBLEND_MODE_OVERLAY",
            "SVG_FEBLEND_MODE_SATURATION",
            "SVG_FEBLEND_MODE_SCREEN",
            "SVG_FEBLEND_MODE_SOFT_LIGHT",
            "SVG_FEBLEND_MODE_UNKNOWN",
            "SVG_FECOLORMATRIX_TYPE_HUEROTATE",
            "SVG_FECOLORMATRIX_TYPE_LUMINANCETOALPHA",
            "SVG_FECOLORMATRIX_TYPE_MATRIX",
            "SVG_FECOLORMATRIX_TYPE_SATURATE",
            "SVG_FECOLORMATRIX_TYPE_UNKNOWN",
            "SVG_FECOMPONENTTRANSFER_TYPE_DISCRETE",
            "SVG_FECOMPONENTTRANSFER_TYPE_GAMMA",
            "SVG_FECOMPONENTTRANSFER_TYPE_IDENTITY",
            "SVG_FECOMPONENTTRANSFER_TYPE_LINEAR",
            "SVG_FECOMPONENTTRANSFER_TYPE_TABLE",
            "SVG_FECOMPONENTTRANSFER_TYPE_UNKNOWN",
            "SVG_FECOMPOSITE_OPERATOR_ARITHMETIC",
            "SVG_FECOMPOSITE_OPERATOR_ATOP",
            "SVG_FECOMPOSITE_OPERATOR_IN",
            "SVG_FECOMPOSITE_OPERATOR_OUT",
            "SVG_FECOMPOSITE_OPERATOR_OVER",
            "SVG_FECOMPOSITE_OPERATOR_UNKNOWN",
            "SVG_FECOMPOSITE_OPERATOR_XOR",
            "SVG_INVALID_VALUE_ERR",
            "SVG_LENGTHTYPE_CM",
            "SVG_LENGTHTYPE_EMS",
            "SVG_LENGTHTYPE_EXS",
            "SVG_LENGTHTYPE_IN",
            "SVG_LENGTHTYPE_MM",
            "SVG_LENGTHTYPE_NUMBER",
            "SVG_LENGTHTYPE_PC",
            "SVG_LENGTHTYPE_PERCENTAGE",
            "SVG_LENGTHTYPE_PT",
            "SVG_LENGTHTYPE_PX",
            "SVG_LENGTHTYPE_UNKNOWN",
            "SVG_MARKERUNITS_STROKEWIDTH",
            "SVG_MARKERUNITS_UNKNOWN",
            "SVG_MARKERUNITS_USERSPACEONUSE",
            "SVG_MARKER_ORIENT_ANGLE",
            "SVG_MARKER_ORIENT_AUTO",
            "SVG_MARKER_ORIENT_UNKNOWN",
            "SVG_MASKTYPE_ALPHA",
            "SVG_MASKTYPE_LUMINANCE",
            "SVG_MATRIX_NOT_INVERTABLE",
            "SVG_MEETORSLICE_MEET",
            "SVG_MEETORSLICE_SLICE",
            "SVG_MEETORSLICE_UNKNOWN",
            "SVG_MORPHOLOGY_OPERATOR_DILATE",
            "SVG_MORPHOLOGY_OPERATOR_ERODE",
            "SVG_MORPHOLOGY_OPERATOR_UNKNOWN",
            "SVG_PAINTTYPE_CURRENTCOLOR",
            "SVG_PAINTTYPE_NONE",
            "SVG_PAINTTYPE_RGBCOLOR",
            "SVG_PAINTTYPE_RGBCOLOR_ICCCOLOR",
            "SVG_PAINTTYPE_UNKNOWN",
            "SVG_PAINTTYPE_URI",
            "SVG_PAINTTYPE_URI_CURRENTCOLOR",
            "SVG_PAINTTYPE_URI_NONE",
            "SVG_PAINTTYPE_URI_RGBCOLOR",
            "SVG_PAINTTYPE_URI_RGBCOLOR_ICCCOLOR",
            "SVG_PRESERVEASPECTRATIO_NONE",
            "SVG_PRESERVEASPECTRATIO_UNKNOWN",
            "SVG_PRESERVEASPECTRATIO_XMAXYMAX",
            "SVG_PRESERVEASPECTRATIO_XMAXYMID",
            "SVG_PRESERVEASPECTRATIO_XMAXYMIN",
            "SVG_PRESERVEASPECTRATIO_XMIDYMAX",
            "SVG_PRESERVEASPECTRATIO_XMIDYMID",
            "SVG_PRESERVEASPECTRATIO_XMIDYMIN",
            "SVG_PRESERVEASPECTRATIO_XMINYMAX",
            "SVG_PRESERVEASPECTRATIO_XMINYMID",
            "SVG_PRESERVEASPECTRATIO_XMINYMIN",
            "SVG_SPREADMETHOD_PAD",
            "SVG_SPREADMETHOD_REFLECT",
            "SVG_SPREADMETHOD_REPEAT",
            "SVG_SPREADMETHOD_UNKNOWN",
            "SVG_STITCHTYPE_NOSTITCH",
            "SVG_STITCHTYPE_STITCH",
            "SVG_STITCHTYPE_UNKNOWN",
            "SVG_TRANSFORM_MATRIX",
            "SVG_TRANSFORM_ROTATE",
            "SVG_TRANSFORM_SCALE",
            "SVG_TRANSFORM_SKEWX",
            "SVG_TRANSFORM_SKEWY",
            "SVG_TRANSFORM_TRANSLATE",
            "SVG_TRANSFORM_UNKNOWN",
            "SVG_TURBULENCE_TYPE_FRACTALNOISE",
            "SVG_TURBULENCE_TYPE_TURBULENCE",
            "SVG_TURBULENCE_TYPE_UNKNOWN",
            "SVG_UNIT_TYPE_OBJECTBOUNDINGBOX",
            "SVG_UNIT_TYPE_UNKNOWN",
            "SVG_UNIT_TYPE_USERSPACEONUSE",
            "SVG_WRONG_TYPE_ERR",
            "SVG_ZOOMANDPAN_DISABLE",
            "SVG_ZOOMANDPAN_MAGNIFY",
            "SVG_ZOOMANDPAN_UNKNOWN",
            "SYNC_CONDITION",
            "SYNC_FENCE",
            "SYNC_FLAGS",
            "SYNC_FLUSH_COMMANDS_BIT",
            "SYNC_GPU_COMMANDS_COMPLETE",
            "SYNC_STATUS",
            "SYNTAX_ERR",
            "SavedPages",
            "Screen",
            "ScreenOrientation",
            "Script",
            "ScriptProcessorNode",
            "ScrollAreaEvent",
            "SecurityPolicyViolationEvent",
            "Selection",
            "Sensor",
            "SensorErrorEvent",
            "ServiceWorker",
            "ServiceWorkerContainer",
            "ServiceWorkerRegistration",
            "SessionDescription",
            "Set",
            "ShadowRoot",
            "SharedArrayBuffer",
            "SharedWorker",
            "SimpleGestureEvent",
            "SourceBuffer",
            "SourceBufferList",
            "SpeechSynthesis",
            "SpeechSynthesisErrorEvent",
            "SpeechSynthesisEvent",
            "SpeechSynthesisUtterance",
            "SpeechSynthesisVoice",
            "StaticRange",
            "StereoPannerNode",
            "StopIteration",
            "Storage",
            "StorageEvent",
            "StorageManager",
            "String",
            "StructType",
            "StylePropertyMap",
            "StylePropertyMapReadOnly",
            "StyleSheet",
            "StyleSheetList",
            "SubmitEvent",
            "SubtleCrypto",
            "Symbol",
            "SyncManager",
            "SyntaxError",
            "TEMPORARY",
            "TEXTPATH_METHODTYPE_ALIGN",
            "TEXTPATH_METHODTYPE_STRETCH",
            "TEXTPATH_METHODTYPE_UNKNOWN",
            "TEXTPATH_SPACINGTYPE_AUTO",
            "TEXTPATH_SPACINGTYPE_EXACT",
            "TEXTPATH_SPACINGTYPE_UNKNOWN",
            "TEXTURE",
            "TEXTURE0",
            "TEXTURE1",
            "TEXTURE10",
            "TEXTURE11",
            "TEXTURE12",
            "TEXTURE13",
            "TEXTURE14",
            "TEXTURE15",
            "TEXTURE16",
            "TEXTURE17",
            "TEXTURE18",
            "TEXTURE19",
            "TEXTURE2",
            "TEXTURE20",
            "TEXTURE21",
            "TEXTURE22",
            "TEXTURE23",
            "TEXTURE24",
            "TEXTURE25",
            "TEXTURE26",
            "TEXTURE27",
            "TEXTURE28",
            "TEXTURE29",
            "TEXTURE3",
            "TEXTURE30",
            "TEXTURE31",
            "TEXTURE4",
            "TEXTURE5",
            "TEXTURE6",
            "TEXTURE7",
            "TEXTURE8",
            "TEXTURE9",
            "TEXTURE_2D",
            "TEXTURE_2D_ARRAY",
            "TEXTURE_3D",
            "TEXTURE_BASE_LEVEL",
            "TEXTURE_BINDING_2D",
            "TEXTURE_BINDING_2D_ARRAY",
            "TEXTURE_BINDING_3D",
            "TEXTURE_BINDING_CUBE_MAP",
            "TEXTURE_COMPARE_FUNC",
            "TEXTURE_COMPARE_MODE",
            "TEXTURE_CUBE_MAP",
            "TEXTURE_CUBE_MAP_NEGATIVE_X",
            "TEXTURE_CUBE_MAP_NEGATIVE_Y",
            "TEXTURE_CUBE_MAP_NEGATIVE_Z",
            "TEXTURE_CUBE_MAP_POSITIVE_X",
            "TEXTURE_CUBE_MAP_POSITIVE_Y",
            "TEXTURE_CUBE_MAP_POSITIVE_Z",
            "TEXTURE_IMMUTABLE_FORMAT",
            "TEXTURE_IMMUTABLE_LEVELS",
            "TEXTURE_MAG_FILTER",
            "TEXTURE_MAX_ANISOTROPY_EXT",
            "TEXTURE_MAX_LEVEL",
            "TEXTURE_MAX_LOD",
            "TEXTURE_MIN_FILTER",
            "TEXTURE_MIN_LOD",
            "TEXTURE_WRAP_R",
            "TEXTURE_WRAP_S",
            "TEXTURE_WRAP_T",
            "TEXT_NODE",
            "TIMEOUT",
            "TIMEOUT_ERR",
            "TIMEOUT_EXPIRED",
            "TIMEOUT_IGNORED",
            "TOO_LARGE_ERR",
            "TRANSACTION_INACTIVE_ERR",
            "TRANSFORM_FEEDBACK",
            "TRANSFORM_FEEDBACK_ACTIVE",
            "TRANSFORM_FEEDBACK_BINDING",
            "TRANSFORM_FEEDBACK_BUFFER",
            "TRANSFORM_FEEDBACK_BUFFER_BINDING",
            "TRANSFORM_FEEDBACK_BUFFER_MODE",
            "TRANSFORM_FEEDBACK_BUFFER_SIZE",
            "TRANSFORM_FEEDBACK_BUFFER_START",
            "TRANSFORM_FEEDBACK_PAUSED",
            "TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN",
            "TRANSFORM_FEEDBACK_VARYINGS",
            "TRIANGLE",
            "TRIANGLES",
            "TRIANGLE_FAN",
            "TRIANGLE_STRIP",
            "TYPE_BACK_FORWARD",
            "TYPE_ERR",
            "TYPE_MISMATCH_ERR",
            "TYPE_NAVIGATE",
            "TYPE_RELOAD",
            "TYPE_RESERVED",
            "Table",
            "TaskAttributionTiming",
            "Text",
            "TextDecoder",
            "TextDecoderStream",
            "TextEncoder",
            "TextEncoderStream",
            "TextEvent",
            "TextMetrics",
            "TextTrack",
            "TextTrackCue",
            "TextTrackCueList",
            "TextTrackList",
            "TimeEvent",
            "TimeRanges",
            "Touch",
            "TouchEvent",
            "TouchList",
            "TrackEvent",
            "TransformStream",
            "TransitionEvent",
            "TreeWalker",
            "TrustedHTML",
            "TrustedScript",
            "TrustedScriptURL",
            "TrustedTypePolicy",
            "TrustedTypePolicyFactory",
            "TypeError",
            "TypedObject",
            "U2F",
            "UIEvent",
            "UNCACHED",
            "UNIFORM_ARRAY_STRIDE",
            "UNIFORM_BLOCK_ACTIVE_UNIFORMS",
            "UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES",
            "UNIFORM_BLOCK_BINDING",
            "UNIFORM_BLOCK_DATA_SIZE",
            "UNIFORM_BLOCK_INDEX",
            "UNIFORM_BLOCK_REFERENCED_BY_FRAGMENT_SHADER",
            "UNIFORM_BLOCK_REFERENCED_BY_VERTEX_SHADER",
            "UNIFORM_BUFFER",
            "UNIFORM_BUFFER_BINDING",
            "UNIFORM_BUFFER_OFFSET_ALIGNMENT",
            "UNIFORM_BUFFER_SIZE",
            "UNIFORM_BUFFER_START",
            "UNIFORM_IS_ROW_MAJOR",
            "UNIFORM_MATRIX_STRIDE",
            "UNIFORM_OFFSET",
            "UNIFORM_SIZE",
            "UNIFORM_TYPE",
            "UNKNOWN_ERR",
            "UNKNOWN_RULE",
            "UNMASKED_RENDERER_WEBGL",
            "UNMASKED_VENDOR_WEBGL",
            "UNORDERED_NODE_ITERATOR_TYPE",
            "UNORDERED_NODE_SNAPSHOT_TYPE",
            "UNPACK_ALIGNMENT",
            "UNPACK_COLORSPACE_CONVERSION_WEBGL",
            "UNPACK_FLIP_Y_WEBGL",
            "UNPACK_IMAGE_HEIGHT",
            "UNPACK_PREMULTIPLY_ALPHA_WEBGL",
            "UNPACK_ROW_LENGTH",
            "UNPACK_SKIP_IMAGES",
            "UNPACK_SKIP_PIXELS",
            "UNPACK_SKIP_ROWS",
            "UNSCHEDULED_STATE",
            "UNSENT",
            "UNSIGNALED",
            "UNSIGNED_BYTE",
            "UNSIGNED_INT",
            "UNSIGNED_INT_10F_11F_11F_REV",
            "UNSIGNED_INT_24_8",
            "UNSIGNED_INT_2_10_10_10_REV",
            "UNSIGNED_INT_5_9_9_9_REV",
            "UNSIGNED_INT_SAMPLER_2D",
            "UNSIGNED_INT_SAMPLER_2D_ARRAY",
            "UNSIGNED_INT_SAMPLER_3D",
            "UNSIGNED_INT_SAMPLER_CUBE",
            "UNSIGNED_INT_VEC2",
            "UNSIGNED_INT_VEC3",
            "UNSIGNED_INT_VEC4",
            "UNSIGNED_NORMALIZED",
            "UNSIGNED_SHORT",
            "UNSIGNED_SHORT_4_4_4_4",
            "UNSIGNED_SHORT_5_5_5_1",
            "UNSIGNED_SHORT_5_6_5",
            "UNSPECIFIED_EVENT_TYPE_ERR",
            "UPDATEREADY",
            "URIError",
            "URL",
            "URLSearchParams",
            "URLUnencoded",
            "URL_MISMATCH_ERR",
            "USB",
            "USBAlternateInterface",
            "USBConfiguration",
            "USBConnectionEvent",
            "USBDevice",
            "USBEndpoint",
            "USBInTransferResult",
            "USBInterface",
            "USBIsochronousInTransferPacket",
            "USBIsochronousInTransferResult",
            "USBIsochronousOutTransferPacket",
            "USBIsochronousOutTransferResult",
            "USBOutTransferResult",
            "UTC",
            "Uint16Array",
            "Uint32Array",
            "Uint8Array",
            "Uint8ClampedArray",
            "UserActivation",
            "UserMessageHandler",
            "UserMessageHandlersNamespace",
            "UserProximityEvent",
            "VALIDATE_STATUS",
            "VALIDATION_ERR",
            "VARIABLES_RULE",
            "VENDOR",
            "VERSION",
            "VERSION_CHANGE",
            "VERSION_ERR",
            "VERTEX_ARRAY_BINDING",
            "VERTEX_ATTRIB_ARRAY_BUFFER_BINDING",
            "VERTEX_ATTRIB_ARRAY_DIVISOR",
            "VERTEX_ATTRIB_ARRAY_DIVISOR_ANGLE",
            "VERTEX_ATTRIB_ARRAY_ENABLED",
            "VERTEX_ATTRIB_ARRAY_INTEGER",
            "VERTEX_ATTRIB_ARRAY_NORMALIZED",
            "VERTEX_ATTRIB_ARRAY_POINTER",
            "VERTEX_ATTRIB_ARRAY_SIZE",
            "VERTEX_ATTRIB_ARRAY_STRIDE",
            "VERTEX_ATTRIB_ARRAY_TYPE",
            "VERTEX_SHADER",
            "VERTICAL",
            "VERTICAL_AXIS",
            "VER_ERR",
            "VIEWPORT",
            "VIEWPORT_RULE",
            "VRDisplay",
            "VRDisplayCapabilities",
            "VRDisplayEvent",
            "VREyeParameters",
            "VRFieldOfView",
            "VRFrameData",
            "VRPose",
            "VRStageParameters",
            "VTTCue",
            "VTTRegion",
            "ValidityState",
            "VideoPlaybackQuality",
            "VideoStreamTrack",
            "VisualViewport",
            "WAIT_FAILED",
            "WEBKIT_FILTER_RULE",
            "WEBKIT_KEYFRAMES_RULE",
            "WEBKIT_KEYFRAME_RULE",
            "WEBKIT_REGION_RULE",
            "WRONG_DOCUMENT_ERR",
            "WakeLock",
            "WakeLockSentinel",
            "WasmAnyRef",
            "WaveShaperNode",
            "WeakMap",
            "WeakRef",
            "WeakSet",
            "WebAssembly",
            "WebGL2RenderingContext",
            "WebGLActiveInfo",
            "WebGLBuffer",
            "WebGLContextEvent",
            "WebGLFramebuffer",
            "WebGLProgram",
            "WebGLQuery",
            "WebGLRenderbuffer",
            "WebGLRenderingContext",
            "WebGLSampler",
            "WebGLShader",
            "WebGLShaderPrecisionFormat",
            "WebGLSync",
            "WebGLTexture",
            "WebGLTransformFeedback",
            "WebGLUniformLocation",
            "WebGLVertexArray",
            "WebGLVertexArrayObject",
            "WebKitAnimationEvent",
            "WebKitBlobBuilder",
            "WebKitCSSFilterRule",
            "WebKitCSSFilterValue",
            "WebKitCSSKeyframeRule",
            "WebKitCSSKeyframesRule",
            "WebKitCSSMatrix",
            "WebKitCSSRegionRule",
            "WebKitCSSTransformValue",
            "WebKitDataCue",
            "WebKitGamepad",
            "WebKitMediaKeyError",
            "WebKitMediaKeyMessageEvent",
            "WebKitMediaKeySession",
            "WebKitMediaKeys",
            "WebKitMediaSource",
            "WebKitMutationObserver",
            "WebKitNamespace",
            "WebKitPlaybackTargetAvailabilityEvent",
            "WebKitPoint",
            "WebKitShadowRoot",
            "WebKitSourceBuffer",
            "WebKitSourceBufferList",
            "WebKitTransitionEvent",
            "WebSocket",
            "WebkitAlignContent",
            "WebkitAlignItems",
            "WebkitAlignSelf",
            "WebkitAnimation",
            "WebkitAnimationDelay",
            "WebkitAnimationDirection",
            "WebkitAnimationDuration",
            "WebkitAnimationFillMode",
            "WebkitAnimationIterationCount",
            "WebkitAnimationName",
            "WebkitAnimationPlayState",
            "WebkitAnimationTimingFunction",
            "WebkitAppearance",
            "WebkitBackfaceVisibility",
            "WebkitBackgroundClip",
            "WebkitBackgroundOrigin",
            "WebkitBackgroundSize",
            "WebkitBorderBottomLeftRadius",
            "WebkitBorderBottomRightRadius",
            "WebkitBorderImage",
            "WebkitBorderRadius",
            "WebkitBorderTopLeftRadius",
            "WebkitBorderTopRightRadius",
            "WebkitBoxAlign",
            "WebkitBoxDirection",
            "WebkitBoxFlex",
            "WebkitBoxOrdinalGroup",
            "WebkitBoxOrient",
            "WebkitBoxPack",
            "WebkitBoxShadow",
            "WebkitBoxSizing",
            "WebkitFilter",
            "WebkitFlex",
            "WebkitFlexBasis",
            "WebkitFlexDirection",
            "WebkitFlexFlow",
            "WebkitFlexGrow",
            "WebkitFlexShrink",
            "WebkitFlexWrap",
            "WebkitJustifyContent",
            "WebkitLineClamp",
            "WebkitMask",
            "WebkitMaskClip",
            "WebkitMaskComposite",
            "WebkitMaskImage",
            "WebkitMaskOrigin",
            "WebkitMaskPosition",
            "WebkitMaskPositionX",
            "WebkitMaskPositionY",
            "WebkitMaskRepeat",
            "WebkitMaskSize",
            "WebkitOrder",
            "WebkitPerspective",
            "WebkitPerspectiveOrigin",
            "WebkitTextFillColor",
            "WebkitTextSizeAdjust",
            "WebkitTextStroke",
            "WebkitTextStrokeColor",
            "WebkitTextStrokeWidth",
            "WebkitTransform",
            "WebkitTransformOrigin",
            "WebkitTransformStyle",
            "WebkitTransition",
            "WebkitTransitionDelay",
            "WebkitTransitionDuration",
            "WebkitTransitionProperty",
            "WebkitTransitionTimingFunction",
            "WebkitUserSelect",
            "WheelEvent",
            "Window",
            "Worker",
            "Worklet",
            "WritableStream",
            "WritableStreamDefaultWriter",
            "XMLDocument",
            "XMLHttpRequest",
            "XMLHttpRequestEventTarget",
            "XMLHttpRequestException",
            "XMLHttpRequestProgressEvent",
            "XMLHttpRequestUpload",
            "XMLSerializer",
            "XMLStylesheetProcessingInstruction",
            "XPathEvaluator",
            "XPathException",
            "XPathExpression",
            "XPathNSResolver",
            "XPathResult",
            "XRBoundedReferenceSpace",
            "XRDOMOverlayState",
            "XRFrame",
            "XRHitTestResult",
            "XRHitTestSource",
            "XRInputSource",
            "XRInputSourceArray",
            "XRInputSourceEvent",
            "XRInputSourcesChangeEvent",
            "XRLayer",
            "XRPose",
            "XRRay",
            "XRReferenceSpace",
            "XRReferenceSpaceEvent",
            "XRRenderState",
            "XRRigidTransform",
            "XRSession",
            "XRSessionEvent",
            "XRSpace",
            "XRSystem",
            "XRTransientInputHitTestResult",
            "XRTransientInputHitTestSource",
            "XRView",
            "XRViewerPose",
            "XRViewport",
            "XRWebGLLayer",
            "XSLTProcessor",
            "ZERO",
            "_XD0M_",
            "_YD0M_",
            "__defineGetter__",
            "__defineSetter__",
            "__lookupGetter__",
            "__lookupSetter__",
            "__opera",
            "__proto__",
            "_browserjsran",
            "a",
            "aLink",
            "abbr",
            "abort",
            "aborted",
            "abs",
            "absolute",
            "acceleration",
            "accelerationIncludingGravity",
            "accelerator",
            "accept",
            "acceptCharset",
            "acceptNode",
            "accessKey",
            "accessKeyLabel",
            "accuracy",
            "acos",
            "acosh",
            "action",
            "actionURL",
            "actions",
            "activated",
            "active",
            "activeCues",
            "activeElement",
            "activeSourceBuffers",
            "activeSourceCount",
            "activeTexture",
            "activeVRDisplays",
            "actualBoundingBoxAscent",
            "actualBoundingBoxDescent",
            "actualBoundingBoxLeft",
            "actualBoundingBoxRight",
            "add",
            "addAll",
            "addBehavior",
            "addCandidate",
            "addColorStop",
            "addCue",
            "addElement",
            "addEventListener",
            "addFilter",
            "addFromString",
            "addFromUri",
            "addIceCandidate",
            "addImport",
            "addListener",
            "addModule",
            "addNamed",
            "addPageRule",
            "addPath",
            "addPointer",
            "addRange",
            "addRegion",
            "addRule",
            "addSearchEngine",
            "addSourceBuffer",
            "addStream",
            "addTextTrack",
            "addTrack",
            "addTransceiver",
            "addWakeLockListener",
            "added",
            "addedNodes",
            "additionalName",
            "additiveSymbols",
            "addons",
            "address",
            "addressLine",
            "adoptNode",
            "adoptedStyleSheets",
            "adr",
            "advance",
            "after",
            "album",
            "alert",
            "algorithm",
            "align",
            "align-content",
            "align-items",
            "align-self",
            "alignContent",
            "alignItems",
            "alignSelf",
            "alignmentBaseline",
            "alinkColor",
            "all",
            "allSettled",
            "allow",
            "allowFullscreen",
            "allowPaymentRequest",
            "allowedDirections",
            "allowedFeatures",
            "allowedToPlay",
            "allowsFeature",
            "alpha",
            "alt",
            "altGraphKey",
            "altHtml",
            "altKey",
            "altLeft",
            "alternate",
            "alternateSetting",
            "alternates",
            "altitude",
            "altitudeAccuracy",
            "amplitude",
            "ancestorOrigins",
            "anchor",
            "anchorNode",
            "anchorOffset",
            "anchors",
            "and",
            "angle",
            "angularAcceleration",
            "angularVelocity",
            "animVal",
            "animate",
            "animatedInstanceRoot",
            "animatedNormalizedPathSegList",
            "animatedPathSegList",
            "animatedPoints",
            "animation",
            "animation-delay",
            "animation-direction",
            "animation-duration",
            "animation-fill-mode",
            "animation-iteration-count",
            "animation-name",
            "animation-play-state",
            "animation-timing-function",
            "animationDelay",
            "animationDirection",
            "animationDuration",
            "animationFillMode",
            "animationIterationCount",
            "animationName",
            "animationPlayState",
            "animationStartTime",
            "animationTimingFunction",
            "animationsPaused",
            "anniversary",
            "antialias",
            "anticipatedRemoval",
            "any",
            "app",
            "appCodeName",
            "appMinorVersion",
            "appName",
            "appNotifications",
            "appVersion",
            "appearance",
            "append",
            "appendBuffer",
            "appendChild",
            "appendData",
            "appendItem",
            "appendMedium",
            "appendNamed",
            "appendRule",
            "appendStream",
            "appendWindowEnd",
            "appendWindowStart",
            "applets",
            "applicationCache",
            "applicationServerKey",
            "apply",
            "applyConstraints",
            "applyElement",
            "arc",
            "arcTo",
            "archive",
            "areas",
            "arguments",
            "ariaAtomic",
            "ariaAutoComplete",
            "ariaBusy",
            "ariaChecked",
            "ariaColCount",
            "ariaColIndex",
            "ariaColSpan",
            "ariaCurrent",
            "ariaDescription",
            "ariaDisabled",
            "ariaExpanded",
            "ariaHasPopup",
            "ariaHidden",
            "ariaKeyShortcuts",
            "ariaLabel",
            "ariaLevel",
            "ariaLive",
            "ariaModal",
            "ariaMultiLine",
            "ariaMultiSelectable",
            "ariaOrientation",
            "ariaPlaceholder",
            "ariaPosInSet",
            "ariaPressed",
            "ariaReadOnly",
            "ariaRelevant",
            "ariaRequired",
            "ariaRoleDescription",
            "ariaRowCount",
            "ariaRowIndex",
            "ariaRowSpan",
            "ariaSelected",
            "ariaSetSize",
            "ariaSort",
            "ariaValueMax",
            "ariaValueMin",
            "ariaValueNow",
            "ariaValueText",
            "arrayBuffer",
            "artist",
            "artwork",
            "as",
            "asIntN",
            "asUintN",
            "asin",
            "asinh",
            "assert",
            "assign",
            "assignedElements",
            "assignedNodes",
            "assignedSlot",
            "async",
            "asyncIterator",
            "atEnd",
            "atan",
            "atan2",
            "atanh",
            "atob",
            "attachEvent",
            "attachInternals",
            "attachShader",
            "attachShadow",
            "attachments",
            "attack",
            "attestationObject",
            "attrChange",
            "attrName",
            "attributeFilter",
            "attributeName",
            "attributeNamespace",
            "attributeOldValue",
            "attributeStyleMap",
            "attributes",
            "attribution",
            "audioBitsPerSecond",
            "audioTracks",
            "audioWorklet",
            "authenticatedSignedWrites",
            "authenticatorData",
            "autoIncrement",
            "autobuffer",
            "autocapitalize",
            "autocomplete",
            "autocorrect",
            "autofocus",
            "automationRate",
            "autoplay",
            "availHeight",
            "availLeft",
            "availTop",
            "availWidth",
            "availability",
            "available",
            "aversion",
            "ax",
            "axes",
            "axis",
            "ay",
            "azimuth",
            "b",
            "back",
            "backface-visibility",
            "backfaceVisibility",
            "background",
            "background-attachment",
            "background-blend-mode",
            "background-clip",
            "background-color",
            "background-image",
            "background-origin",
            "background-position",
            "background-position-x",
            "background-position-y",
            "background-repeat",
            "background-size",
            "backgroundAttachment",
            "backgroundBlendMode",
            "backgroundClip",
            "backgroundColor",
            "backgroundFetch",
            "backgroundImage",
            "backgroundOrigin",
            "backgroundPosition",
            "backgroundPositionX",
            "backgroundPositionY",
            "backgroundRepeat",
            "backgroundSize",
            "badInput",
            "badge",
            "balance",
            "baseFrequencyX",
            "baseFrequencyY",
            "baseLatency",
            "baseLayer",
            "baseNode",
            "baseOffset",
            "baseURI",
            "baseVal",
            "baselineShift",
            "battery",
            "bday",
            "before",
            "beginElement",
            "beginElementAt",
            "beginPath",
            "beginQuery",
            "beginTransformFeedback",
            "behavior",
            "behaviorCookie",
            "behaviorPart",
            "behaviorUrns",
            "beta",
            "bezierCurveTo",
            "bgColor",
            "bgProperties",
            "bias",
            "big",
            "bigint64",
            "biguint64",
            "binaryType",
            "bind",
            "bindAttribLocation",
            "bindBuffer",
            "bindBufferBase",
            "bindBufferRange",
            "bindFramebuffer",
            "bindRenderbuffer",
            "bindSampler",
            "bindTexture",
            "bindTransformFeedback",
            "bindVertexArray",
            "blendColor",
            "blendEquation",
            "blendEquationSeparate",
            "blendFunc",
            "blendFuncSeparate",
            "blink",
            "blitFramebuffer",
            "blob",
            "block-size",
            "blockDirection",
            "blockSize",
            "blockedURI",
            "blue",
            "bluetooth",
            "blur",
            "body",
            "bodyUsed",
            "bold",
            "bookmarks",
            "booleanValue",
            "border",
            "border-block",
            "border-block-color",
            "border-block-end",
            "border-block-end-color",
            "border-block-end-style",
            "border-block-end-width",
            "border-block-start",
            "border-block-start-color",
            "border-block-start-style",
            "border-block-start-width",
            "border-block-style",
            "border-block-width",
            "border-bottom",
            "border-bottom-color",
            "border-bottom-left-radius",
            "border-bottom-right-radius",
            "border-bottom-style",
            "border-bottom-width",
            "border-collapse",
            "border-color",
            "border-end-end-radius",
            "border-end-start-radius",
            "border-image",
            "border-image-outset",
            "border-image-repeat",
            "border-image-slice",
            "border-image-source",
            "border-image-width",
            "border-inline",
            "border-inline-color",
            "border-inline-end",
            "border-inline-end-color",
            "border-inline-end-style",
            "border-inline-end-width",
            "border-inline-start",
            "border-inline-start-color",
            "border-inline-start-style",
            "border-inline-start-width",
            "border-inline-style",
            "border-inline-width",
            "border-left",
            "border-left-color",
            "border-left-style",
            "border-left-width",
            "border-radius",
            "border-right",
            "border-right-color",
            "border-right-style",
            "border-right-width",
            "border-spacing",
            "border-start-end-radius",
            "border-start-start-radius",
            "border-style",
            "border-top",
            "border-top-color",
            "border-top-left-radius",
            "border-top-right-radius",
            "border-top-style",
            "border-top-width",
            "border-width",
            "borderBlock",
            "borderBlockColor",
            "borderBlockEnd",
            "borderBlockEndColor",
            "borderBlockEndStyle",
            "borderBlockEndWidth",
            "borderBlockStart",
            "borderBlockStartColor",
            "borderBlockStartStyle",
            "borderBlockStartWidth",
            "borderBlockStyle",
            "borderBlockWidth",
            "borderBottom",
            "borderBottomColor",
            "borderBottomLeftRadius",
            "borderBottomRightRadius",
            "borderBottomStyle",
            "borderBottomWidth",
            "borderBoxSize",
            "borderCollapse",
            "borderColor",
            "borderColorDark",
            "borderColorLight",
            "borderEndEndRadius",
            "borderEndStartRadius",
            "borderImage",
            "borderImageOutset",
            "borderImageRepeat",
            "borderImageSlice",
            "borderImageSource",
            "borderImageWidth",
            "borderInline",
            "borderInlineColor",
            "borderInlineEnd",
            "borderInlineEndColor",
            "borderInlineEndStyle",
            "borderInlineEndWidth",
            "borderInlineStart",
            "borderInlineStartColor",
            "borderInlineStartStyle",
            "borderInlineStartWidth",
            "borderInlineStyle",
            "borderInlineWidth",
            "borderLeft",
            "borderLeftColor",
            "borderLeftStyle",
            "borderLeftWidth",
            "borderRadius",
            "borderRight",
            "borderRightColor",
            "borderRightStyle",
            "borderRightWidth",
            "borderSpacing",
            "borderStartEndRadius",
            "borderStartStartRadius",
            "borderStyle",
            "borderTop",
            "borderTopColor",
            "borderTopLeftRadius",
            "borderTopRightRadius",
            "borderTopStyle",
            "borderTopWidth",
            "borderWidth",
            "bottom",
            "bottomMargin",
            "bound",
            "boundElements",
            "boundingClientRect",
            "boundingHeight",
            "boundingLeft",
            "boundingTop",
            "boundingWidth",
            "bounds",
            "boundsGeometry",
            "box-decoration-break",
            "box-shadow",
            "box-sizing",
            "boxDecorationBreak",
            "boxShadow",
            "boxSizing",
            "break-after",
            "break-before",
            "break-inside",
            "breakAfter",
            "breakBefore",
            "breakInside",
            "broadcast",
            "browserLanguage",
            "btoa",
            "bubbles",
            "buffer",
            "bufferData",
            "bufferDepth",
            "bufferSize",
            "bufferSubData",
            "buffered",
            "bufferedAmount",
            "bufferedAmountLowThreshold",
            "buildID",
            "buildNumber",
            "button",
            "buttonID",
            "buttons",
            "byteLength",
            "byteOffset",
            "bytesWritten",
            "c",
            "cache",
            "caches",
            "call",
            "caller",
            "canBeFormatted",
            "canBeMounted",
            "canBeShared",
            "canHaveChildren",
            "canHaveHTML",
            "canInsertDTMF",
            "canMakePayment",
            "canPlayType",
            "canPresent",
            "canTrickleIceCandidates",
            "cancel",
            "cancelAndHoldAtTime",
            "cancelAnimationFrame",
            "cancelBubble",
            "cancelIdleCallback",
            "cancelScheduledValues",
            "cancelVideoFrameCallback",
            "cancelWatchAvailability",
            "cancelable",
            "candidate",
            "canonicalUUID",
            "canvas",
            "capabilities",
            "caption",
            "caption-side",
            "captionSide",
            "capture",
            "captureEvents",
            "captureStackTrace",
            "captureStream",
            "caret-color",
            "caretBidiLevel",
            "caretColor",
            "caretPositionFromPoint",
            "caretRangeFromPoint",
            "cast",
            "catch",
            "category",
            "cbrt",
            "cd",
            "ceil",
            "cellIndex",
            "cellPadding",
            "cellSpacing",
            "cells",
            "ch",
            "chOff",
            "chain",
            "challenge",
            "changeType",
            "changedTouches",
            "channel",
            "channelCount",
            "channelCountMode",
            "channelInterpretation",
            "char",
            "charAt",
            "charCode",
            "charCodeAt",
            "charIndex",
            "charLength",
            "characterData",
            "characterDataOldValue",
            "characterSet",
            "characteristic",
            "charging",
            "chargingTime",
            "charset",
            "check",
            "checkEnclosure",
            "checkFramebufferStatus",
            "checkIntersection",
            "checkValidity",
            "checked",
            "childElementCount",
            "childList",
            "childNodes",
            "children",
            "chrome",
            "ciphertext",
            "cite",
            "city",
            "claimInterface",
            "claimed",
            "classList",
            "className",
            "classid",
            "clear",
            "clearAppBadge",
            "clearAttributes",
            "clearBufferfi",
            "clearBufferfv",
            "clearBufferiv",
            "clearBufferuiv",
            "clearColor",
            "clearData",
            "clearDepth",
            "clearHalt",
            "clearImmediate",
            "clearInterval",
            "clearLiveSeekableRange",
            "clearMarks",
            "clearMaxGCPauseAccumulator",
            "clearMeasures",
            "clearParameters",
            "clearRect",
            "clearResourceTimings",
            "clearShadow",
            "clearStencil",
            "clearTimeout",
            "clearWatch",
            "click",
            "clickCount",
            "clientDataJSON",
            "clientHeight",
            "clientInformation",
            "clientLeft",
            "clientRect",
            "clientRects",
            "clientTop",
            "clientWaitSync",
            "clientWidth",
            "clientX",
            "clientY",
            "clip",
            "clip-path",
            "clip-rule",
            "clipBottom",
            "clipLeft",
            "clipPath",
            "clipPathUnits",
            "clipRight",
            "clipRule",
            "clipTop",
            "clipboard",
            "clipboardData",
            "clone",
            "cloneContents",
            "cloneNode",
            "cloneRange",
            "close",
            "closePath",
            "closed",
            "closest",
            "clz",
            "clz32",
            "cm",
            "cmp",
            "code",
            "codeBase",
            "codePointAt",
            "codeType",
            "colSpan",
            "collapse",
            "collapseToEnd",
            "collapseToStart",
            "collapsed",
            "collect",
            "colno",
            "color",
            "color-adjust",
            "color-interpolation",
            "color-interpolation-filters",
            "colorAdjust",
            "colorDepth",
            "colorInterpolation",
            "colorInterpolationFilters",
            "colorMask",
            "colorType",
            "cols",
            "column-count",
            "column-fill",
            "column-gap",
            "column-rule",
            "column-rule-color",
            "column-rule-style",
            "column-rule-width",
            "column-span",
            "column-width",
            "columnCount",
            "columnFill",
            "columnGap",
            "columnNumber",
            "columnRule",
            "columnRuleColor",
            "columnRuleStyle",
            "columnRuleWidth",
            "columnSpan",
            "columnWidth",
            "columns",
            "command",
            "commit",
            "commitPreferences",
            "commitStyles",
            "commonAncestorContainer",
            "compact",
            "compareBoundaryPoints",
            "compareDocumentPosition",
            "compareEndPoints",
            "compareExchange",
            "compareNode",
            "comparePoint",
            "compatMode",
            "compatible",
            "compile",
            "compileShader",
            "compileStreaming",
            "complete",
            "component",
            "componentFromPoint",
            "composed",
            "composedPath",
            "composite",
            "compositionEndOffset",
            "compositionStartOffset",
            "compressedTexImage2D",
            "compressedTexImage3D",
            "compressedTexSubImage2D",
            "compressedTexSubImage3D",
            "computedStyleMap",
            "concat",
            "conditionText",
            "coneInnerAngle",
            "coneOuterAngle",
            "coneOuterGain",
            "configuration",
            "configurationName",
            "configurationValue",
            "configurations",
            "confirm",
            "confirmComposition",
            "confirmSiteSpecificTrackingException",
            "confirmWebWideTrackingException",
            "connect",
            "connectEnd",
            "connectShark",
            "connectStart",
            "connected",
            "connection",
            "connectionList",
            "connectionSpeed",
            "connectionState",
            "connections",
            "console",
            "consolidate",
            "constraint",
            "constrictionActive",
            "construct",
            "constructor",
            "contactID",
            "contain",
            "containerId",
            "containerName",
            "containerSrc",
            "containerType",
            "contains",
            "containsNode",
            "content",
            "contentBoxSize",
            "contentDocument",
            "contentEditable",
            "contentHint",
            "contentOverflow",
            "contentRect",
            "contentScriptType",
            "contentStyleType",
            "contentType",
            "contentWindow",
            "context",
            "contextMenu",
            "contextmenu",
            "continue",
            "continuePrimaryKey",
            "continuous",
            "control",
            "controlTransferIn",
            "controlTransferOut",
            "controller",
            "controls",
            "controlsList",
            "convertPointFromNode",
            "convertQuadFromNode",
            "convertRectFromNode",
            "convertToBlob",
            "convertToSpecifiedUnits",
            "cookie",
            "cookieEnabled",
            "coords",
            "copyBufferSubData",
            "copyFromChannel",
            "copyTexImage2D",
            "copyTexSubImage2D",
            "copyTexSubImage3D",
            "copyToChannel",
            "copyWithin",
            "correspondingElement",
            "correspondingUseElement",
            "corruptedVideoFrames",
            "cos",
            "cosh",
            "count",
            "countReset",
            "counter-increment",
            "counter-reset",
            "counter-set",
            "counterIncrement",
            "counterReset",
            "counterSet",
            "country",
            "cpuClass",
            "cpuSleepAllowed",
            "create",
            "createAnalyser",
            "createAnswer",
            "createAttribute",
            "createAttributeNS",
            "createBiquadFilter",
            "createBuffer",
            "createBufferSource",
            "createCDATASection",
            "createCSSStyleSheet",
            "createCaption",
            "createChannelMerger",
            "createChannelSplitter",
            "createComment",
            "createConstantSource",
            "createContextualFragment",
            "createControlRange",
            "createConvolver",
            "createDTMFSender",
            "createDataChannel",
            "createDelay",
            "createDelayNode",
            "createDocument",
            "createDocumentFragment",
            "createDocumentType",
            "createDynamicsCompressor",
            "createElement",
            "createElementNS",
            "createEntityReference",
            "createEvent",
            "createEventObject",
            "createExpression",
            "createFramebuffer",
            "createFunction",
            "createGain",
            "createGainNode",
            "createHTML",
            "createHTMLDocument",
            "createIIRFilter",
            "createImageBitmap",
            "createImageData",
            "createIndex",
            "createJavaScriptNode",
            "createLinearGradient",
            "createMediaElementSource",
            "createMediaKeys",
            "createMediaStreamDestination",
            "createMediaStreamSource",
            "createMediaStreamTrackSource",
            "createMutableFile",
            "createNSResolver",
            "createNodeIterator",
            "createNotification",
            "createObjectStore",
            "createObjectURL",
            "createOffer",
            "createOscillator",
            "createPanner",
            "createPattern",
            "createPeriodicWave",
            "createPolicy",
            "createPopup",
            "createProcessingInstruction",
            "createProgram",
            "createQuery",
            "createRadialGradient",
            "createRange",
            "createRangeCollection",
            "createReader",
            "createRenderbuffer",
            "createSVGAngle",
            "createSVGLength",
            "createSVGMatrix",
            "createSVGNumber",
            "createSVGPathSegArcAbs",
            "createSVGPathSegArcRel",
            "createSVGPathSegClosePath",
            "createSVGPathSegCurvetoCubicAbs",
            "createSVGPathSegCurvetoCubicRel",
            "createSVGPathSegCurvetoCubicSmoothAbs",
            "createSVGPathSegCurvetoCubicSmoothRel",
            "createSVGPathSegCurvetoQuadraticAbs",
            "createSVGPathSegCurvetoQuadraticRel",
            "createSVGPathSegCurvetoQuadraticSmoothAbs",
            "createSVGPathSegCurvetoQuadraticSmoothRel",
            "createSVGPathSegLinetoAbs",
            "createSVGPathSegLinetoHorizontalAbs",
            "createSVGPathSegLinetoHorizontalRel",
            "createSVGPathSegLinetoRel",
            "createSVGPathSegLinetoVerticalAbs",
            "createSVGPathSegLinetoVerticalRel",
            "createSVGPathSegMovetoAbs",
            "createSVGPathSegMovetoRel",
            "createSVGPoint",
            "createSVGRect",
            "createSVGTransform",
            "createSVGTransformFromMatrix",
            "createSampler",
            "createScript",
            "createScriptProcessor",
            "createScriptURL",
            "createSession",
            "createShader",
            "createShadowRoot",
            "createStereoPanner",
            "createStyleSheet",
            "createTBody",
            "createTFoot",
            "createTHead",
            "createTextNode",
            "createTextRange",
            "createTexture",
            "createTouch",
            "createTouchList",
            "createTransformFeedback",
            "createTreeWalker",
            "createVertexArray",
            "createWaveShaper",
            "creationTime",
            "credentials",
            "crossOrigin",
            "crossOriginIsolated",
            "crypto",
            "csi",
            "csp",
            "cssFloat",
            "cssRules",
            "cssText",
            "cssValueType",
            "ctrlKey",
            "ctrlLeft",
            "cues",
            "cullFace",
            "currentDirection",
            "currentLocalDescription",
            "currentNode",
            "currentPage",
            "currentRect",
            "currentRemoteDescription",
            "currentScale",
            "currentScript",
            "currentSrc",
            "currentState",
            "currentStyle",
            "currentTarget",
            "currentTime",
            "currentTranslate",
            "currentView",
            "cursor",
            "curve",
            "customElements",
            "customError",
            "cx",
            "cy",
            "d",
            "data",
            "dataFld",
            "dataFormatAs",
            "dataLoss",
            "dataLossMessage",
            "dataPageSize",
            "dataSrc",
            "dataTransfer",
            "database",
            "databases",
            "dataset",
            "dateTime",
            "db",
            "debug",
            "debuggerEnabled",
            "declare",
            "decode",
            "decodeAudioData",
            "decodeURI",
            "decodeURIComponent",
            "decodedBodySize",
            "decoding",
            "decodingInfo",
            "decrypt",
            "default",
            "defaultCharset",
            "defaultChecked",
            "defaultMuted",
            "defaultPlaybackRate",
            "defaultPolicy",
            "defaultPrevented",
            "defaultRequest",
            "defaultSelected",
            "defaultStatus",
            "defaultURL",
            "defaultValue",
            "defaultView",
            "defaultstatus",
            "defer",
            "define",
            "defineMagicFunction",
            "defineMagicVariable",
            "defineProperties",
            "defineProperty",
            "deg",
            "delay",
            "delayTime",
            "delegatesFocus",
            "delete",
            "deleteBuffer",
            "deleteCaption",
            "deleteCell",
            "deleteContents",
            "deleteData",
            "deleteDatabase",
            "deleteFramebuffer",
            "deleteFromDocument",
            "deleteIndex",
            "deleteMedium",
            "deleteObjectStore",
            "deleteProgram",
            "deleteProperty",
            "deleteQuery",
            "deleteRenderbuffer",
            "deleteRow",
            "deleteRule",
            "deleteSampler",
            "deleteShader",
            "deleteSync",
            "deleteTFoot",
            "deleteTHead",
            "deleteTexture",
            "deleteTransformFeedback",
            "deleteVertexArray",
            "deliverChangeRecords",
            "delivery",
            "deliveryInfo",
            "deliveryStatus",
            "deliveryTimestamp",
            "delta",
            "deltaMode",
            "deltaX",
            "deltaY",
            "deltaZ",
            "dependentLocality",
            "depthFar",
            "depthFunc",
            "depthMask",
            "depthNear",
            "depthRange",
            "deref",
            "deriveBits",
            "deriveKey",
            "description",
            "deselectAll",
            "designMode",
            "desiredSize",
            "destination",
            "destinationURL",
            "detach",
            "detachEvent",
            "detachShader",
            "detail",
            "details",
            "detect",
            "detune",
            "device",
            "deviceClass",
            "deviceId",
            "deviceMemory",
            "devicePixelContentBoxSize",
            "devicePixelRatio",
            "deviceProtocol",
            "deviceSubclass",
            "deviceVersionMajor",
            "deviceVersionMinor",
            "deviceVersionSubminor",
            "deviceXDPI",
            "deviceYDPI",
            "didTimeout",
            "diffuseConstant",
            "digest",
            "dimensions",
            "dir",
            "dirName",
            "direction",
            "dirxml",
            "disable",
            "disablePictureInPicture",
            "disableRemotePlayback",
            "disableVertexAttribArray",
            "disabled",
            "dischargingTime",
            "disconnect",
            "disconnectShark",
            "dispatchEvent",
            "display",
            "displayId",
            "displayName",
            "disposition",
            "distanceModel",
            "div",
            "divisor",
            "djsapi",
            "djsproxy",
            "doImport",
            "doNotTrack",
            "doScroll",
            "doctype",
            "document",
            "documentElement",
            "documentMode",
            "documentURI",
            "dolphin",
            "dolphinGameCenter",
            "dolphininfo",
            "dolphinmeta",
            "domComplete",
            "domContentLoadedEventEnd",
            "domContentLoadedEventStart",
            "domInteractive",
            "domLoading",
            "domOverlayState",
            "domain",
            "domainLookupEnd",
            "domainLookupStart",
            "dominant-baseline",
            "dominantBaseline",
            "done",
            "dopplerFactor",
            "dotAll",
            "downDegrees",
            "downlink",
            "download",
            "downloadTotal",
            "downloaded",
            "dpcm",
            "dpi",
            "dppx",
            "dragDrop",
            "draggable",
            "drawArrays",
            "drawArraysInstanced",
            "drawArraysInstancedANGLE",
            "drawBuffers",
            "drawCustomFocusRing",
            "drawElements",
            "drawElementsInstanced",
            "drawElementsInstancedANGLE",
            "drawFocusIfNeeded",
            "drawImage",
            "drawImageFromRect",
            "drawRangeElements",
            "drawSystemFocusRing",
            "drawingBufferHeight",
            "drawingBufferWidth",
            "dropEffect",
            "droppedVideoFrames",
            "dropzone",
            "dtmf",
            "dump",
            "dumpProfile",
            "duplicate",
            "durability",
            "duration",
            "dvname",
            "dvnum",
            "dx",
            "dy",
            "dynsrc",
            "e",
            "edgeMode",
            "effect",
            "effectAllowed",
            "effectiveDirective",
            "effectiveType",
            "elapsedTime",
            "element",
            "elementFromPoint",
            "elementTiming",
            "elements",
            "elementsFromPoint",
            "elevation",
            "ellipse",
            "em",
            "email",
            "embeds",
            "emma",
            "empty",
            "empty-cells",
            "emptyCells",
            "emptyHTML",
            "emptyScript",
            "emulatedPosition",
            "enable",
            "enableBackground",
            "enableDelegations",
            "enableStyleSheetsForSet",
            "enableVertexAttribArray",
            "enabled",
            "enabledPlugin",
            "encode",
            "encodeInto",
            "encodeURI",
            "encodeURIComponent",
            "encodedBodySize",
            "encoding",
            "encodingInfo",
            "encrypt",
            "enctype",
            "end",
            "endContainer",
            "endElement",
            "endElementAt",
            "endOfStream",
            "endOffset",
            "endQuery",
            "endTime",
            "endTransformFeedback",
            "ended",
            "endpoint",
            "endpointNumber",
            "endpoints",
            "endsWith",
            "enterKeyHint",
            "entities",
            "entries",
            "entryType",
            "enumerate",
            "enumerateDevices",
            "enumerateEditable",
            "environmentBlendMode",
            "equals",
            "error",
            "errorCode",
            "errorDetail",
            "errorText",
            "escape",
            "estimate",
            "eval",
            "evaluate",
            "event",
            "eventPhase",
            "every",
            "ex",
            "exception",
            "exchange",
            "exec",
            "execCommand",
            "execCommandShowHelp",
            "execScript",
            "exitFullscreen",
            "exitPictureInPicture",
            "exitPointerLock",
            "exitPresent",
            "exp",
            "expand",
            "expandEntityReferences",
            "expando",
            "expansion",
            "expiration",
            "expirationTime",
            "expires",
            "expiryDate",
            "explicitOriginalTarget",
            "expm1",
            "exponent",
            "exponentialRampToValueAtTime",
            "exportKey",
            "exports",
            "extend",
            "extensions",
            "extentNode",
            "extentOffset",
            "external",
            "externalResourcesRequired",
            "extractContents",
            "extractable",
            "eye",
            "f",
            "face",
            "factoryReset",
            "failureReason",
            "fallback",
            "family",
            "familyName",
            "farthestViewportElement",
            "fastSeek",
            "fatal",
            "featureId",
            "featurePolicy",
            "featureSettings",
            "features",
            "fenceSync",
            "fetch",
            "fetchStart",
            "fftSize",
            "fgColor",
            "fieldOfView",
            "file",
            "fileCreatedDate",
            "fileHandle",
            "fileModifiedDate",
            "fileName",
            "fileSize",
            "fileUpdatedDate",
            "filename",
            "files",
            "filesystem",
            "fill",
            "fill-opacity",
            "fill-rule",
            "fillLightMode",
            "fillOpacity",
            "fillRect",
            "fillRule",
            "fillStyle",
            "fillText",
            "filter",
            "filterResX",
            "filterResY",
            "filterUnits",
            "filters",
            "finally",
            "find",
            "findIndex",
            "findRule",
            "findText",
            "finish",
            "finished",
            "fireEvent",
            "firesTouchEvents",
            "firstChild",
            "firstElementChild",
            "firstPage",
            "fixed",
            "flags",
            "flat",
            "flatMap",
            "flex",
            "flex-basis",
            "flex-direction",
            "flex-flow",
            "flex-grow",
            "flex-shrink",
            "flex-wrap",
            "flexBasis",
            "flexDirection",
            "flexFlow",
            "flexGrow",
            "flexShrink",
            "flexWrap",
            "flipX",
            "flipY",
            "float",
            "float32",
            "float64",
            "flood-color",
            "flood-opacity",
            "floodColor",
            "floodOpacity",
            "floor",
            "flush",
            "focus",
            "focusNode",
            "focusOffset",
            "font",
            "font-family",
            "font-feature-settings",
            "font-kerning",
            "font-language-override",
            "font-optical-sizing",
            "font-size",
            "font-size-adjust",
            "font-stretch",
            "font-style",
            "font-synthesis",
            "font-variant",
            "font-variant-alternates",
            "font-variant-caps",
            "font-variant-east-asian",
            "font-variant-ligatures",
            "font-variant-numeric",
            "font-variant-position",
            "font-variation-settings",
            "font-weight",
            "fontFamily",
            "fontFeatureSettings",
            "fontKerning",
            "fontLanguageOverride",
            "fontOpticalSizing",
            "fontSize",
            "fontSizeAdjust",
            "fontSmoothingEnabled",
            "fontStretch",
            "fontStyle",
            "fontSynthesis",
            "fontVariant",
            "fontVariantAlternates",
            "fontVariantCaps",
            "fontVariantEastAsian",
            "fontVariantLigatures",
            "fontVariantNumeric",
            "fontVariantPosition",
            "fontVariationSettings",
            "fontWeight",
            "fontcolor",
            "fontfaces",
            "fonts",
            "fontsize",
            "for",
            "forEach",
            "force",
            "forceRedraw",
            "form",
            "formAction",
            "formData",
            "formEnctype",
            "formMethod",
            "formNoValidate",
            "formTarget",
            "format",
            "formatToParts",
            "forms",
            "forward",
            "forwardX",
            "forwardY",
            "forwardZ",
            "foundation",
            "fr",
            "fragmentDirective",
            "frame",
            "frameBorder",
            "frameElement",
            "frameSpacing",
            "framebuffer",
            "framebufferHeight",
            "framebufferRenderbuffer",
            "framebufferTexture2D",
            "framebufferTextureLayer",
            "framebufferWidth",
            "frames",
            "freeSpace",
            "freeze",
            "frequency",
            "frequencyBinCount",
            "from",
            "fromCharCode",
            "fromCodePoint",
            "fromElement",
            "fromEntries",
            "fromFloat32Array",
            "fromFloat64Array",
            "fromMatrix",
            "fromPoint",
            "fromQuad",
            "fromRect",
            "frontFace",
            "fround",
            "fullPath",
            "fullScreen",
            "fullscreen",
            "fullscreenElement",
            "fullscreenEnabled",
            "fx",
            "fy",
            "gain",
            "gamepad",
            "gamma",
            "gap",
            "gatheringState",
            "gatt",
            "genderIdentity",
            "generateCertificate",
            "generateKey",
            "generateMipmap",
            "generateRequest",
            "geolocation",
            "gestureObject",
            "get",
            "getActiveAttrib",
            "getActiveUniform",
            "getActiveUniformBlockName",
            "getActiveUniformBlockParameter",
            "getActiveUniforms",
            "getAdjacentText",
            "getAll",
            "getAllKeys",
            "getAllResponseHeaders",
            "getAllowlistForFeature",
            "getAnimations",
            "getAsFile",
            "getAsString",
            "getAttachedShaders",
            "getAttribLocation",
            "getAttribute",
            "getAttributeNS",
            "getAttributeNames",
            "getAttributeNode",
            "getAttributeNodeNS",
            "getAttributeType",
            "getAudioTracks",
            "getAvailability",
            "getBBox",
            "getBattery",
            "getBigInt64",
            "getBigUint64",
            "getBlob",
            "getBookmark",
            "getBoundingClientRect",
            "getBounds",
            "getBoxQuads",
            "getBufferParameter",
            "getBufferSubData",
            "getByteFrequencyData",
            "getByteTimeDomainData",
            "getCSSCanvasContext",
            "getCTM",
            "getCandidateWindowClientRect",
            "getCanonicalLocales",
            "getCapabilities",
            "getChannelData",
            "getCharNumAtPosition",
            "getCharacteristic",
            "getCharacteristics",
            "getClientExtensionResults",
            "getClientRect",
            "getClientRects",
            "getCoalescedEvents",
            "getCompositionAlternatives",
            "getComputedStyle",
            "getComputedTextLength",
            "getComputedTiming",
            "getConfiguration",
            "getConstraints",
            "getContext",
            "getContextAttributes",
            "getContributingSources",
            "getCounterValue",
            "getCueAsHTML",
            "getCueById",
            "getCurrentPosition",
            "getCurrentTime",
            "getData",
            "getDatabaseNames",
            "getDate",
            "getDay",
            "getDefaultComputedStyle",
            "getDescriptor",
            "getDescriptors",
            "getDestinationInsertionPoints",
            "getDevices",
            "getDirectory",
            "getDisplayMedia",
            "getDistributedNodes",
            "getEditable",
            "getElementById",
            "getElementsByClassName",
            "getElementsByName",
            "getElementsByTagName",
            "getElementsByTagNameNS",
            "getEnclosureList",
            "getEndPositionOfChar",
            "getEntries",
            "getEntriesByName",
            "getEntriesByType",
            "getError",
            "getExtension",
            "getExtentOfChar",
            "getEyeParameters",
            "getFeature",
            "getFile",
            "getFiles",
            "getFilesAndDirectories",
            "getFingerprints",
            "getFloat32",
            "getFloat64",
            "getFloatFrequencyData",
            "getFloatTimeDomainData",
            "getFloatValue",
            "getFragDataLocation",
            "getFrameData",
            "getFramebufferAttachmentParameter",
            "getFrequencyResponse",
            "getFullYear",
            "getGamepads",
            "getHitTestResults",
            "getHitTestResultsForTransientInput",
            "getHours",
            "getIdentityAssertion",
            "getIds",
            "getImageData",
            "getIndexedParameter",
            "getInstalledRelatedApps",
            "getInt16",
            "getInt32",
            "getInt8",
            "getInternalformatParameter",
            "getIntersectionList",
            "getItem",
            "getItems",
            "getKey",
            "getKeyframes",
            "getLayers",
            "getLayoutMap",
            "getLineDash",
            "getLocalCandidates",
            "getLocalParameters",
            "getLocalStreams",
            "getMarks",
            "getMatchedCSSRules",
            "getMaxGCPauseSinceClear",
            "getMeasures",
            "getMetadata",
            "getMilliseconds",
            "getMinutes",
            "getModifierState",
            "getMonth",
            "getNamedItem",
            "getNamedItemNS",
            "getNativeFramebufferScaleFactor",
            "getNotifications",
            "getNotifier",
            "getNumberOfChars",
            "getOffsetReferenceSpace",
            "getOutputTimestamp",
            "getOverrideHistoryNavigationMode",
            "getOverrideStyle",
            "getOwnPropertyDescriptor",
            "getOwnPropertyDescriptors",
            "getOwnPropertyNames",
            "getOwnPropertySymbols",
            "getParameter",
            "getParameters",
            "getParent",
            "getPathSegAtLength",
            "getPhotoCapabilities",
            "getPhotoSettings",
            "getPointAtLength",
            "getPose",
            "getPredictedEvents",
            "getPreference",
            "getPreferenceDefault",
            "getPresentationAttribute",
            "getPreventDefault",
            "getPrimaryService",
            "getPrimaryServices",
            "getProgramInfoLog",
            "getProgramParameter",
            "getPropertyCSSValue",
            "getPropertyPriority",
            "getPropertyShorthand",
            "getPropertyType",
            "getPropertyValue",
            "getPrototypeOf",
            "getQuery",
            "getQueryParameter",
            "getRGBColorValue",
            "getRandomValues",
            "getRangeAt",
            "getReader",
            "getReceivers",
            "getRectValue",
            "getRegistration",
            "getRegistrations",
            "getRemoteCandidates",
            "getRemoteCertificates",
            "getRemoteParameters",
            "getRemoteStreams",
            "getRenderbufferParameter",
            "getResponseHeader",
            "getRoot",
            "getRootNode",
            "getRotationOfChar",
            "getSVGDocument",
            "getSamplerParameter",
            "getScreenCTM",
            "getSeconds",
            "getSelectedCandidatePair",
            "getSelection",
            "getSenders",
            "getService",
            "getSettings",
            "getShaderInfoLog",
            "getShaderParameter",
            "getShaderPrecisionFormat",
            "getShaderSource",
            "getSimpleDuration",
            "getSiteIcons",
            "getSources",
            "getSpeculativeParserUrls",
            "getStartPositionOfChar",
            "getStartTime",
            "getState",
            "getStats",
            "getStatusForPolicy",
            "getStorageUpdates",
            "getStreamById",
            "getStringValue",
            "getSubStringLength",
            "getSubscription",
            "getSupportedConstraints",
            "getSupportedExtensions",
            "getSupportedFormats",
            "getSyncParameter",
            "getSynchronizationSources",
            "getTags",
            "getTargetRanges",
            "getTexParameter",
            "getTime",
            "getTimezoneOffset",
            "getTiming",
            "getTotalLength",
            "getTrackById",
            "getTracks",
            "getTransceivers",
            "getTransform",
            "getTransformFeedbackVarying",
            "getTransformToElement",
            "getTransports",
            "getType",
            "getTypeMapping",
            "getUTCDate",
            "getUTCDay",
            "getUTCFullYear",
            "getUTCHours",
            "getUTCMilliseconds",
            "getUTCMinutes",
            "getUTCMonth",
            "getUTCSeconds",
            "getUint16",
            "getUint32",
            "getUint8",
            "getUniform",
            "getUniformBlockIndex",
            "getUniformIndices",
            "getUniformLocation",
            "getUserMedia",
            "getVRDisplays",
            "getValues",
            "getVarDate",
            "getVariableValue",
            "getVertexAttrib",
            "getVertexAttribOffset",
            "getVideoPlaybackQuality",
            "getVideoTracks",
            "getViewerPose",
            "getViewport",
            "getVoices",
            "getWakeLockState",
            "getWriter",
            "getYear",
            "givenName",
            "global",
            "globalAlpha",
            "globalCompositeOperation",
            "globalThis",
            "glyphOrientationHorizontal",
            "glyphOrientationVertical",
            "glyphRef",
            "go",
            "grabFrame",
            "grad",
            "gradientTransform",
            "gradientUnits",
            "grammars",
            "green",
            "grid",
            "grid-area",
            "grid-auto-columns",
            "grid-auto-flow",
            "grid-auto-rows",
            "grid-column",
            "grid-column-end",
            "grid-column-gap",
            "grid-column-start",
            "grid-gap",
            "grid-row",
            "grid-row-end",
            "grid-row-gap",
            "grid-row-start",
            "grid-template",
            "grid-template-areas",
            "grid-template-columns",
            "grid-template-rows",
            "gridArea",
            "gridAutoColumns",
            "gridAutoFlow",
            "gridAutoRows",
            "gridColumn",
            "gridColumnEnd",
            "gridColumnGap",
            "gridColumnStart",
            "gridGap",
            "gridRow",
            "gridRowEnd",
            "gridRowGap",
            "gridRowStart",
            "gridTemplate",
            "gridTemplateAreas",
            "gridTemplateColumns",
            "gridTemplateRows",
            "gripSpace",
            "group",
            "groupCollapsed",
            "groupEnd",
            "groupId",
            "hadRecentInput",
            "hand",
            "handedness",
            "hapticActuators",
            "hardwareConcurrency",
            "has",
            "hasAttribute",
            "hasAttributeNS",
            "hasAttributes",
            "hasBeenActive",
            "hasChildNodes",
            "hasComposition",
            "hasEnrolledInstrument",
            "hasExtension",
            "hasExternalDisplay",
            "hasFeature",
            "hasFocus",
            "hasInstance",
            "hasLayout",
            "hasOrientation",
            "hasOwnProperty",
            "hasPointerCapture",
            "hasPosition",
            "hasReading",
            "hasStorageAccess",
            "hash",
            "head",
            "headers",
            "heading",
            "height",
            "hidden",
            "hide",
            "hideFocus",
            "high",
            "highWaterMark",
            "hint",
            "history",
            "honorificPrefix",
            "honorificSuffix",
            "horizontalOverflow",
            "host",
            "hostCandidate",
            "hostname",
            "href",
            "hrefTranslate",
            "hreflang",
            "hspace",
            "html5TagCheckInerface",
            "htmlFor",
            "htmlText",
            "httpEquiv",
            "httpRequestStatusCode",
            "hwTimestamp",
            "hyphens",
            "hypot",
            "iccId",
            "iceConnectionState",
            "iceGatheringState",
            "iceTransport",
            "icon",
            "iconURL",
            "id",
            "identifier",
            "identity",
            "idpLoginUrl",
            "ignoreBOM",
            "ignoreCase",
            "ignoreDepthValues",
            "image-orientation",
            "image-rendering",
            "imageHeight",
            "imageOrientation",
            "imageRendering",
            "imageSizes",
            "imageSmoothingEnabled",
            "imageSmoothingQuality",
            "imageSrcset",
            "imageWidth",
            "images",
            "ime-mode",
            "imeMode",
            "implementation",
            "importKey",
            "importNode",
            "importStylesheet",
            "imports",
            "impp",
            "imul",
            "in",
            "in1",
            "in2",
            "inBandMetadataTrackDispatchType",
            "inRange",
            "includes",
            "incremental",
            "indeterminate",
            "index",
            "indexNames",
            "indexOf",
            "indexedDB",
            "indicate",
            "inertiaDestinationX",
            "inertiaDestinationY",
            "info",
            "init",
            "initAnimationEvent",
            "initBeforeLoadEvent",
            "initClipboardEvent",
            "initCloseEvent",
            "initCommandEvent",
            "initCompositionEvent",
            "initCustomEvent",
            "initData",
            "initDataType",
            "initDeviceMotionEvent",
            "initDeviceOrientationEvent",
            "initDragEvent",
            "initErrorEvent",
            "initEvent",
            "initFocusEvent",
            "initGestureEvent",
            "initHashChangeEvent",
            "initKeyEvent",
            "initKeyboardEvent",
            "initMSManipulationEvent",
            "initMessageEvent",
            "initMouseEvent",
            "initMouseScrollEvent",
            "initMouseWheelEvent",
            "initMutationEvent",
            "initNSMouseEvent",
            "initOverflowEvent",
            "initPageEvent",
            "initPageTransitionEvent",
            "initPointerEvent",
            "initPopStateEvent",
            "initProgressEvent",
            "initScrollAreaEvent",
            "initSimpleGestureEvent",
            "initStorageEvent",
            "initTextEvent",
            "initTimeEvent",
            "initTouchEvent",
            "initTransitionEvent",
            "initUIEvent",
            "initWebKitAnimationEvent",
            "initWebKitTransitionEvent",
            "initWebKitWheelEvent",
            "initWheelEvent",
            "initialTime",
            "initialize",
            "initiatorType",
            "inline-size",
            "inlineSize",
            "inlineVerticalFieldOfView",
            "inner",
            "innerHTML",
            "innerHeight",
            "innerText",
            "innerWidth",
            "input",
            "inputBuffer",
            "inputEncoding",
            "inputMethod",
            "inputMode",
            "inputSource",
            "inputSources",
            "inputType",
            "inputs",
            "insertAdjacentElement",
            "insertAdjacentHTML",
            "insertAdjacentText",
            "insertBefore",
            "insertCell",
            "insertDTMF",
            "insertData",
            "insertItemBefore",
            "insertNode",
            "insertRow",
            "insertRule",
            "inset",
            "inset-block",
            "inset-block-end",
            "inset-block-start",
            "inset-inline",
            "inset-inline-end",
            "inset-inline-start",
            "insetBlock",
            "insetBlockEnd",
            "insetBlockStart",
            "insetInline",
            "insetInlineEnd",
            "insetInlineStart",
            "installing",
            "instanceRoot",
            "instantiate",
            "instantiateStreaming",
            "instruments",
            "int16",
            "int32",
            "int8",
            "integrity",
            "interactionMode",
            "intercept",
            "interfaceClass",
            "interfaceName",
            "interfaceNumber",
            "interfaceProtocol",
            "interfaceSubclass",
            "interfaces",
            "interimResults",
            "internalSubset",
            "interpretation",
            "intersectionRatio",
            "intersectionRect",
            "intersectsNode",
            "interval",
            "invalidIteratorState",
            "invalidateFramebuffer",
            "invalidateSubFramebuffer",
            "inverse",
            "invertSelf",
            "is",
            "is2D",
            "isActive",
            "isAlternate",
            "isArray",
            "isBingCurrentSearchDefault",
            "isBuffer",
            "isCandidateWindowVisible",
            "isChar",
            "isCollapsed",
            "isComposing",
            "isConcatSpreadable",
            "isConnected",
            "isContentEditable",
            "isContentHandlerRegistered",
            "isContextLost",
            "isDefaultNamespace",
            "isDirectory",
            "isDisabled",
            "isEnabled",
            "isEqual",
            "isEqualNode",
            "isExtensible",
            "isExternalCTAP2SecurityKeySupported",
            "isFile",
            "isFinite",
            "isFramebuffer",
            "isFrozen",
            "isGenerator",
            "isHTML",
            "isHistoryNavigation",
            "isId",
            "isIdentity",
            "isInjected",
            "isInteger",
            "isIntersecting",
            "isLockFree",
            "isMap",
            "isMultiLine",
            "isNaN",
            "isOpen",
            "isPointInFill",
            "isPointInPath",
            "isPointInRange",
            "isPointInStroke",
            "isPrefAlternate",
            "isPresenting",
            "isPrimary",
            "isProgram",
            "isPropertyImplicit",
            "isProtocolHandlerRegistered",
            "isPrototypeOf",
            "isQuery",
            "isRenderbuffer",
            "isSafeInteger",
            "isSameNode",
            "isSampler",
            "isScript",
            "isScriptURL",
            "isSealed",
            "isSecureContext",
            "isSessionSupported",
            "isShader",
            "isSupported",
            "isSync",
            "isTextEdit",
            "isTexture",
            "isTransformFeedback",
            "isTrusted",
            "isTypeSupported",
            "isUserVerifyingPlatformAuthenticatorAvailable",
            "isVertexArray",
            "isView",
            "isVisible",
            "isochronousTransferIn",
            "isochronousTransferOut",
            "isolation",
            "italics",
            "item",
            "itemId",
            "itemProp",
            "itemRef",
            "itemScope",
            "itemType",
            "itemValue",
            "items",
            "iterateNext",
            "iterationComposite",
            "iterator",
            "javaEnabled",
            "jobTitle",
            "join",
            "json",
            "justify-content",
            "justify-items",
            "justify-self",
            "justifyContent",
            "justifyItems",
            "justifySelf",
            "k1",
            "k2",
            "k3",
            "k4",
            "kHz",
            "keepalive",
            "kernelMatrix",
            "kernelUnitLengthX",
            "kernelUnitLengthY",
            "kerning",
            "key",
            "keyCode",
            "keyFor",
            "keyIdentifier",
            "keyLightEnabled",
            "keyLocation",
            "keyPath",
            "keyStatuses",
            "keySystem",
            "keyText",
            "keyUsage",
            "keyboard",
            "keys",
            "keytype",
            "kind",
            "knee",
            "label",
            "labels",
            "lang",
            "language",
            "languages",
            "largeArcFlag",
            "lastChild",
            "lastElementChild",
            "lastEventId",
            "lastIndex",
            "lastIndexOf",
            "lastInputTime",
            "lastMatch",
            "lastMessageSubject",
            "lastMessageType",
            "lastModified",
            "lastModifiedDate",
            "lastPage",
            "lastParen",
            "lastState",
            "lastStyleSheetSet",
            "latitude",
            "layerX",
            "layerY",
            "layoutFlow",
            "layoutGrid",
            "layoutGridChar",
            "layoutGridLine",
            "layoutGridMode",
            "layoutGridType",
            "lbound",
            "left",
            "leftContext",
            "leftDegrees",
            "leftMargin",
            "leftProjectionMatrix",
            "leftViewMatrix",
            "length",
            "lengthAdjust",
            "lengthComputable",
            "letter-spacing",
            "letterSpacing",
            "level",
            "lighting-color",
            "lightingColor",
            "limitingConeAngle",
            "line",
            "line-break",
            "line-height",
            "lineAlign",
            "lineBreak",
            "lineCap",
            "lineDashOffset",
            "lineHeight",
            "lineJoin",
            "lineNumber",
            "lineTo",
            "lineWidth",
            "linearAcceleration",
            "linearRampToValueAtTime",
            "linearVelocity",
            "lineno",
            "lines",
            "link",
            "linkColor",
            "linkProgram",
            "links",
            "list",
            "list-style",
            "list-style-image",
            "list-style-position",
            "list-style-type",
            "listStyle",
            "listStyleImage",
            "listStylePosition",
            "listStyleType",
            "listener",
            "load",
            "loadEventEnd",
            "loadEventStart",
            "loadTime",
            "loadTimes",
            "loaded",
            "loading",
            "localDescription",
            "localName",
            "localService",
            "localStorage",
            "locale",
            "localeCompare",
            "location",
            "locationbar",
            "lock",
            "locked",
            "lockedFile",
            "locks",
            "log",
            "log10",
            "log1p",
            "log2",
            "logicalXDPI",
            "logicalYDPI",
            "longDesc",
            "longitude",
            "lookupNamespaceURI",
            "lookupPrefix",
            "loop",
            "loopEnd",
            "loopStart",
            "looping",
            "low",
            "lower",
            "lowerBound",
            "lowerOpen",
            "lowsrc",
            "m11",
            "m12",
            "m13",
            "m14",
            "m21",
            "m22",
            "m23",
            "m24",
            "m31",
            "m32",
            "m33",
            "m34",
            "m41",
            "m42",
            "m43",
            "m44",
            "makeXRCompatible",
            "manifest",
            "manufacturer",
            "manufacturerName",
            "map",
            "mapping",
            "margin",
            "margin-block",
            "margin-block-end",
            "margin-block-start",
            "margin-bottom",
            "margin-inline",
            "margin-inline-end",
            "margin-inline-start",
            "margin-left",
            "margin-right",
            "margin-top",
            "marginBlock",
            "marginBlockEnd",
            "marginBlockStart",
            "marginBottom",
            "marginHeight",
            "marginInline",
            "marginInlineEnd",
            "marginInlineStart",
            "marginLeft",
            "marginRight",
            "marginTop",
            "marginWidth",
            "mark",
            "marker",
            "marker-end",
            "marker-mid",
            "marker-offset",
            "marker-start",
            "markerEnd",
            "markerHeight",
            "markerMid",
            "markerOffset",
            "markerStart",
            "markerUnits",
            "markerWidth",
            "marks",
            "mask",
            "mask-clip",
            "mask-composite",
            "mask-image",
            "mask-mode",
            "mask-origin",
            "mask-position",
            "mask-position-x",
            "mask-position-y",
            "mask-repeat",
            "mask-size",
            "mask-type",
            "maskClip",
            "maskComposite",
            "maskContentUnits",
            "maskImage",
            "maskMode",
            "maskOrigin",
            "maskPosition",
            "maskPositionX",
            "maskPositionY",
            "maskRepeat",
            "maskSize",
            "maskType",
            "maskUnits",
            "match",
            "matchAll",
            "matchMedia",
            "matchMedium",
            "matches",
            "matrix",
            "matrixTransform",
            "max",
            "max-block-size",
            "max-height",
            "max-inline-size",
            "max-width",
            "maxActions",
            "maxAlternatives",
            "maxBlockSize",
            "maxChannelCount",
            "maxChannels",
            "maxConnectionsPerServer",
            "maxDecibels",
            "maxDistance",
            "maxHeight",
            "maxInlineSize",
            "maxLayers",
            "maxLength",
            "maxMessageSize",
            "maxPacketLifeTime",
            "maxRetransmits",
            "maxTouchPoints",
            "maxValue",
            "maxWidth",
            "measure",
            "measureText",
            "media",
            "mediaCapabilities",
            "mediaDevices",
            "mediaElement",
            "mediaGroup",
            "mediaKeys",
            "mediaSession",
            "mediaStream",
            "mediaText",
            "meetOrSlice",
            "memory",
            "menubar",
            "mergeAttributes",
            "message",
            "messageClass",
            "messageHandlers",
            "messageType",
            "metaKey",
            "metadata",
            "method",
            "methodDetails",
            "methodName",
            "mid",
            "mimeType",
            "mimeTypes",
            "min",
            "min-block-size",
            "min-height",
            "min-inline-size",
            "min-width",
            "minBlockSize",
            "minDecibels",
            "minHeight",
            "minInlineSize",
            "minLength",
            "minValue",
            "minWidth",
            "miterLimit",
            "mix-blend-mode",
            "mixBlendMode",
            "mm",
            "mode",
            "modify",
            "mount",
            "move",
            "moveBy",
            "moveEnd",
            "moveFirst",
            "moveFocusDown",
            "moveFocusLeft",
            "moveFocusRight",
            "moveFocusUp",
            "moveNext",
            "moveRow",
            "moveStart",
            "moveTo",
            "moveToBookmark",
            "moveToElementText",
            "moveToPoint",
            "movementX",
            "movementY",
            "mozAdd",
            "mozAnimationStartTime",
            "mozAnon",
            "mozApps",
            "mozAudioCaptured",
            "mozAudioChannelType",
            "mozAutoplayEnabled",
            "mozCancelAnimationFrame",
            "mozCancelFullScreen",
            "mozCancelRequestAnimationFrame",
            "mozCaptureStream",
            "mozCaptureStreamUntilEnded",
            "mozClearDataAt",
            "mozContact",
            "mozContacts",
            "mozCreateFileHandle",
            "mozCurrentTransform",
            "mozCurrentTransformInverse",
            "mozCursor",
            "mozDash",
            "mozDashOffset",
            "mozDecodedFrames",
            "mozExitPointerLock",
            "mozFillRule",
            "mozFragmentEnd",
            "mozFrameDelay",
            "mozFullScreen",
            "mozFullScreenElement",
            "mozFullScreenEnabled",
            "mozGetAll",
            "mozGetAllKeys",
            "mozGetAsFile",
            "mozGetDataAt",
            "mozGetMetadata",
            "mozGetUserMedia",
            "mozHasAudio",
            "mozHasItem",
            "mozHidden",
            "mozImageSmoothingEnabled",
            "mozIndexedDB",
            "mozInnerScreenX",
            "mozInnerScreenY",
            "mozInputSource",
            "mozIsTextField",
            "mozItem",
            "mozItemCount",
            "mozItems",
            "mozLength",
            "mozLockOrientation",
            "mozMatchesSelector",
            "mozMovementX",
            "mozMovementY",
            "mozOpaque",
            "mozOrientation",
            "mozPaintCount",
            "mozPaintedFrames",
            "mozParsedFrames",
            "mozPay",
            "mozPointerLockElement",
            "mozPresentedFrames",
            "mozPreservesPitch",
            "mozPressure",
            "mozPrintCallback",
            "mozRTCIceCandidate",
            "mozRTCPeerConnection",
            "mozRTCSessionDescription",
            "mozRemove",
            "mozRequestAnimationFrame",
            "mozRequestFullScreen",
            "mozRequestPointerLock",
            "mozSetDataAt",
            "mozSetImageElement",
            "mozSourceNode",
            "mozSrcObject",
            "mozSystem",
            "mozTCPSocket",
            "mozTextStyle",
            "mozTypesAt",
            "mozUnlockOrientation",
            "mozUserCancelled",
            "mozVisibilityState",
            "ms",
            "msAnimation",
            "msAnimationDelay",
            "msAnimationDirection",
            "msAnimationDuration",
            "msAnimationFillMode",
            "msAnimationIterationCount",
            "msAnimationName",
            "msAnimationPlayState",
            "msAnimationStartTime",
            "msAnimationTimingFunction",
            "msBackfaceVisibility",
            "msBlockProgression",
            "msCSSOMElementFloatMetrics",
            "msCaching",
            "msCachingEnabled",
            "msCancelRequestAnimationFrame",
            "msCapsLockWarningOff",
            "msClearImmediate",
            "msClose",
            "msContentZoomChaining",
            "msContentZoomFactor",
            "msContentZoomLimit",
            "msContentZoomLimitMax",
            "msContentZoomLimitMin",
            "msContentZoomSnap",
            "msContentZoomSnapPoints",
            "msContentZoomSnapType",
            "msContentZooming",
            "msConvertURL",
            "msCrypto",
            "msDoNotTrack",
            "msElementsFromPoint",
            "msElementsFromRect",
            "msExitFullscreen",
            "msExtendedCode",
            "msFillRule",
            "msFirstPaint",
            "msFlex",
            "msFlexAlign",
            "msFlexDirection",
            "msFlexFlow",
            "msFlexItemAlign",
            "msFlexLinePack",
            "msFlexNegative",
            "msFlexOrder",
            "msFlexPack",
            "msFlexPositive",
            "msFlexPreferredSize",
            "msFlexWrap",
            "msFlowFrom",
            "msFlowInto",
            "msFontFeatureSettings",
            "msFullscreenElement",
            "msFullscreenEnabled",
            "msGetInputContext",
            "msGetRegionContent",
            "msGetUntransformedBounds",
            "msGraphicsTrustStatus",
            "msGridColumn",
            "msGridColumnAlign",
            "msGridColumnSpan",
            "msGridColumns",
            "msGridRow",
            "msGridRowAlign",
            "msGridRowSpan",
            "msGridRows",
            "msHidden",
            "msHighContrastAdjust",
            "msHyphenateLimitChars",
            "msHyphenateLimitLines",
            "msHyphenateLimitZone",
            "msHyphens",
            "msImageSmoothingEnabled",
            "msImeAlign",
            "msIndexedDB",
            "msInterpolationMode",
            "msIsStaticHTML",
            "msKeySystem",
            "msKeys",
            "msLaunchUri",
            "msLockOrientation",
            "msManipulationViewsEnabled",
            "msMatchMedia",
            "msMatchesSelector",
            "msMaxTouchPoints",
            "msOrientation",
            "msOverflowStyle",
            "msPerspective",
            "msPerspectiveOrigin",
            "msPlayToDisabled",
            "msPlayToPreferredSourceUri",
            "msPlayToPrimary",
            "msPointerEnabled",
            "msRegionOverflow",
            "msReleasePointerCapture",
            "msRequestAnimationFrame",
            "msRequestFullscreen",
            "msSaveBlob",
            "msSaveOrOpenBlob",
            "msScrollChaining",
            "msScrollLimit",
            "msScrollLimitXMax",
            "msScrollLimitXMin",
            "msScrollLimitYMax",
            "msScrollLimitYMin",
            "msScrollRails",
            "msScrollSnapPointsX",
            "msScrollSnapPointsY",
            "msScrollSnapType",
            "msScrollSnapX",
            "msScrollSnapY",
            "msScrollTranslation",
            "msSetImmediate",
            "msSetMediaKeys",
            "msSetPointerCapture",
            "msTextCombineHorizontal",
            "msTextSizeAdjust",
            "msToBlob",
            "msTouchAction",
            "msTouchSelect",
            "msTraceAsyncCallbackCompleted",
            "msTraceAsyncCallbackStarting",
            "msTraceAsyncOperationCompleted",
            "msTraceAsyncOperationStarting",
            "msTransform",
            "msTransformOrigin",
            "msTransformStyle",
            "msTransition",
            "msTransitionDelay",
            "msTransitionDuration",
            "msTransitionProperty",
            "msTransitionTimingFunction",
            "msUnlockOrientation",
            "msUpdateAsyncCallbackRelation",
            "msUserSelect",
            "msVisibilityState",
            "msWrapFlow",
            "msWrapMargin",
            "msWrapThrough",
            "msWriteProfilerMark",
            "msZoom",
            "msZoomTo",
            "mt",
            "mul",
            "multiEntry",
            "multiSelectionObj",
            "multiline",
            "multiple",
            "multiply",
            "multiplySelf",
            "mutableFile",
            "muted",
            "n",
            "name",
            "nameProp",
            "namedItem",
            "namedRecordset",
            "names",
            "namespaceURI",
            "namespaces",
            "naturalHeight",
            "naturalWidth",
            "navigate",
            "navigation",
            "navigationMode",
            "navigationPreload",
            "navigationStart",
            "navigator",
            "near",
            "nearestViewportElement",
            "negative",
            "negotiated",
            "netscape",
            "networkState",
            "newScale",
            "newTranslate",
            "newURL",
            "newValue",
            "newValueSpecifiedUnits",
            "newVersion",
            "newhome",
            "next",
            "nextElementSibling",
            "nextHopProtocol",
            "nextNode",
            "nextPage",
            "nextSibling",
            "nickname",
            "noHref",
            "noModule",
            "noResize",
            "noShade",
            "noValidate",
            "noWrap",
            "node",
            "nodeName",
            "nodeType",
            "nodeValue",
            "nonce",
            "normalize",
            "normalizedPathSegList",
            "notationName",
            "notations",
            "note",
            "noteGrainOn",
            "noteOff",
            "noteOn",
            "notify",
            "now",
            "numOctaves",
            "number",
            "numberOfChannels",
            "numberOfInputs",
            "numberOfItems",
            "numberOfOutputs",
            "numberValue",
            "oMatchesSelector",
            "object",
            "object-fit",
            "object-position",
            "objectFit",
            "objectPosition",
            "objectStore",
            "objectStoreNames",
            "objectType",
            "observe",
            "of",
            "offscreenBuffering",
            "offset",
            "offset-anchor",
            "offset-distance",
            "offset-path",
            "offset-rotate",
            "offsetAnchor",
            "offsetDistance",
            "offsetHeight",
            "offsetLeft",
            "offsetNode",
            "offsetParent",
            "offsetPath",
            "offsetRotate",
            "offsetTop",
            "offsetWidth",
            "offsetX",
            "offsetY",
            "ok",
            "oldURL",
            "oldValue",
            "oldVersion",
            "olderShadowRoot",
            "onLine",
            "onabort",
            "onabsolutedeviceorientation",
            "onactivate",
            "onactive",
            "onaddsourcebuffer",
            "onaddstream",
            "onaddtrack",
            "onafterprint",
            "onafterscriptexecute",
            "onafterupdate",
            "onanimationcancel",
            "onanimationend",
            "onanimationiteration",
            "onanimationstart",
            "onappinstalled",
            "onaudioend",
            "onaudioprocess",
            "onaudiostart",
            "onautocomplete",
            "onautocompleteerror",
            "onauxclick",
            "onbeforeactivate",
            "onbeforecopy",
            "onbeforecut",
            "onbeforedeactivate",
            "onbeforeeditfocus",
            "onbeforeinstallprompt",
            "onbeforepaste",
            "onbeforeprint",
            "onbeforescriptexecute",
            "onbeforeunload",
            "onbeforeupdate",
            "onbeforexrselect",
            "onbegin",
            "onblocked",
            "onblur",
            "onbounce",
            "onboundary",
            "onbufferedamountlow",
            "oncached",
            "oncancel",
            "oncandidatewindowhide",
            "oncandidatewindowshow",
            "oncandidatewindowupdate",
            "oncanplay",
            "oncanplaythrough",
            "once",
            "oncellchange",
            "onchange",
            "oncharacteristicvaluechanged",
            "onchargingchange",
            "onchargingtimechange",
            "onchecking",
            "onclick",
            "onclose",
            "onclosing",
            "oncompassneedscalibration",
            "oncomplete",
            "onconnect",
            "onconnecting",
            "onconnectionavailable",
            "onconnectionstatechange",
            "oncontextmenu",
            "oncontrollerchange",
            "oncontrolselect",
            "oncopy",
            "oncuechange",
            "oncut",
            "ondataavailable",
            "ondatachannel",
            "ondatasetchanged",
            "ondatasetcomplete",
            "ondblclick",
            "ondeactivate",
            "ondevicechange",
            "ondevicelight",
            "ondevicemotion",
            "ondeviceorientation",
            "ondeviceorientationabsolute",
            "ondeviceproximity",
            "ondischargingtimechange",
            "ondisconnect",
            "ondisplay",
            "ondownloading",
            "ondrag",
            "ondragend",
            "ondragenter",
            "ondragexit",
            "ondragleave",
            "ondragover",
            "ondragstart",
            "ondrop",
            "ondurationchange",
            "onemptied",
            "onencrypted",
            "onend",
            "onended",
            "onenter",
            "onenterpictureinpicture",
            "onerror",
            "onerrorupdate",
            "onexit",
            "onfilterchange",
            "onfinish",
            "onfocus",
            "onfocusin",
            "onfocusout",
            "onformdata",
            "onfreeze",
            "onfullscreenchange",
            "onfullscreenerror",
            "ongatheringstatechange",
            "ongattserverdisconnected",
            "ongesturechange",
            "ongestureend",
            "ongesturestart",
            "ongotpointercapture",
            "onhashchange",
            "onhelp",
            "onicecandidate",
            "onicecandidateerror",
            "oniceconnectionstatechange",
            "onicegatheringstatechange",
            "oninactive",
            "oninput",
            "oninputsourceschange",
            "oninvalid",
            "onkeydown",
            "onkeypress",
            "onkeystatuseschange",
            "onkeyup",
            "onlanguagechange",
            "onlayoutcomplete",
            "onleavepictureinpicture",
            "onlevelchange",
            "onload",
            "onloadeddata",
            "onloadedmetadata",
            "onloadend",
            "onloading",
            "onloadingdone",
            "onloadingerror",
            "onloadstart",
            "onlosecapture",
            "onlostpointercapture",
            "only",
            "onmark",
            "onmessage",
            "onmessageerror",
            "onmidimessage",
            "onmousedown",
            "onmouseenter",
            "onmouseleave",
            "onmousemove",
            "onmouseout",
            "onmouseover",
            "onmouseup",
            "onmousewheel",
            "onmove",
            "onmoveend",
            "onmovestart",
            "onmozfullscreenchange",
            "onmozfullscreenerror",
            "onmozorientationchange",
            "onmozpointerlockchange",
            "onmozpointerlockerror",
            "onmscontentzoom",
            "onmsfullscreenchange",
            "onmsfullscreenerror",
            "onmsgesturechange",
            "onmsgesturedoubletap",
            "onmsgestureend",
            "onmsgesturehold",
            "onmsgesturestart",
            "onmsgesturetap",
            "onmsgotpointercapture",
            "onmsinertiastart",
            "onmslostpointercapture",
            "onmsmanipulationstatechanged",
            "onmsneedkey",
            "onmsorientationchange",
            "onmspointercancel",
            "onmspointerdown",
            "onmspointerenter",
            "onmspointerhover",
            "onmspointerleave",
            "onmspointermove",
            "onmspointerout",
            "onmspointerover",
            "onmspointerup",
            "onmssitemodejumplistitemremoved",
            "onmsthumbnailclick",
            "onmute",
            "onnegotiationneeded",
            "onnomatch",
            "onnoupdate",
            "onobsolete",
            "onoffline",
            "ononline",
            "onopen",
            "onorientationchange",
            "onpagechange",
            "onpagehide",
            "onpageshow",
            "onpaste",
            "onpause",
            "onpayerdetailchange",
            "onpaymentmethodchange",
            "onplay",
            "onplaying",
            "onpluginstreamstart",
            "onpointercancel",
            "onpointerdown",
            "onpointerenter",
            "onpointerleave",
            "onpointerlockchange",
            "onpointerlockerror",
            "onpointermove",
            "onpointerout",
            "onpointerover",
            "onpointerrawupdate",
            "onpointerup",
            "onpopstate",
            "onprocessorerror",
            "onprogress",
            "onpropertychange",
            "onratechange",
            "onreading",
            "onreadystatechange",
            "onrejectionhandled",
            "onrelease",
            "onremove",
            "onremovesourcebuffer",
            "onremovestream",
            "onremovetrack",
            "onrepeat",
            "onreset",
            "onresize",
            "onresizeend",
            "onresizestart",
            "onresourcetimingbufferfull",
            "onresult",
            "onresume",
            "onrowenter",
            "onrowexit",
            "onrowsdelete",
            "onrowsinserted",
            "onscroll",
            "onsearch",
            "onsecuritypolicyviolation",
            "onseeked",
            "onseeking",
            "onselect",
            "onselectedcandidatepairchange",
            "onselectend",
            "onselectionchange",
            "onselectstart",
            "onshippingaddresschange",
            "onshippingoptionchange",
            "onshow",
            "onsignalingstatechange",
            "onsoundend",
            "onsoundstart",
            "onsourceclose",
            "onsourceclosed",
            "onsourceended",
            "onsourceopen",
            "onspeechend",
            "onspeechstart",
            "onsqueeze",
            "onsqueezeend",
            "onsqueezestart",
            "onstalled",
            "onstart",
            "onstatechange",
            "onstop",
            "onstorage",
            "onstoragecommit",
            "onsubmit",
            "onsuccess",
            "onsuspend",
            "onterminate",
            "ontextinput",
            "ontimeout",
            "ontimeupdate",
            "ontoggle",
            "ontonechange",
            "ontouchcancel",
            "ontouchend",
            "ontouchmove",
            "ontouchstart",
            "ontrack",
            "ontransitioncancel",
            "ontransitionend",
            "ontransitionrun",
            "ontransitionstart",
            "onunhandledrejection",
            "onunload",
            "onunmute",
            "onupdate",
            "onupdateend",
            "onupdatefound",
            "onupdateready",
            "onupdatestart",
            "onupgradeneeded",
            "onuserproximity",
            "onversionchange",
            "onvisibilitychange",
            "onvoiceschanged",
            "onvolumechange",
            "onvrdisplayactivate",
            "onvrdisplayconnect",
            "onvrdisplaydeactivate",
            "onvrdisplaydisconnect",
            "onvrdisplaypresentchange",
            "onwaiting",
            "onwaitingforkey",
            "onwarning",
            "onwebkitanimationend",
            "onwebkitanimationiteration",
            "onwebkitanimationstart",
            "onwebkitcurrentplaybacktargetiswirelesschanged",
            "onwebkitfullscreenchange",
            "onwebkitfullscreenerror",
            "onwebkitkeyadded",
            "onwebkitkeyerror",
            "onwebkitkeymessage",
            "onwebkitneedkey",
            "onwebkitorientationchange",
            "onwebkitplaybacktargetavailabilitychanged",
            "onwebkitpointerlockchange",
            "onwebkitpointerlockerror",
            "onwebkitresourcetimingbufferfull",
            "onwebkittransitionend",
            "onwheel",
            "onzoom",
            "opacity",
            "open",
            "openCursor",
            "openDatabase",
            "openKeyCursor",
            "opened",
            "opener",
            "opera",
            "operationType",
            "operator",
            "opr",
            "optimum",
            "options",
            "or",
            "order",
            "orderX",
            "orderY",
            "ordered",
            "org",
            "organization",
            "orient",
            "orientAngle",
            "orientType",
            "orientation",
            "orientationX",
            "orientationY",
            "orientationZ",
            "origin",
            "originalPolicy",
            "originalTarget",
            "orphans",
            "oscpu",
            "outerHTML",
            "outerHeight",
            "outerText",
            "outerWidth",
            "outline",
            "outline-color",
            "outline-offset",
            "outline-style",
            "outline-width",
            "outlineColor",
            "outlineOffset",
            "outlineStyle",
            "outlineWidth",
            "outputBuffer",
            "outputLatency",
            "outputs",
            "overflow",
            "overflow-anchor",
            "overflow-block",
            "overflow-inline",
            "overflow-wrap",
            "overflow-x",
            "overflow-y",
            "overflowAnchor",
            "overflowBlock",
            "overflowInline",
            "overflowWrap",
            "overflowX",
            "overflowY",
            "overrideMimeType",
            "oversample",
            "overscroll-behavior",
            "overscroll-behavior-block",
            "overscroll-behavior-inline",
            "overscroll-behavior-x",
            "overscroll-behavior-y",
            "overscrollBehavior",
            "overscrollBehaviorBlock",
            "overscrollBehaviorInline",
            "overscrollBehaviorX",
            "overscrollBehaviorY",
            "ownKeys",
            "ownerDocument",
            "ownerElement",
            "ownerNode",
            "ownerRule",
            "ownerSVGElement",
            "owningElement",
            "p1",
            "p2",
            "p3",
            "p4",
            "packetSize",
            "packets",
            "pad",
            "padEnd",
            "padStart",
            "padding",
            "padding-block",
            "padding-block-end",
            "padding-block-start",
            "padding-bottom",
            "padding-inline",
            "padding-inline-end",
            "padding-inline-start",
            "padding-left",
            "padding-right",
            "padding-top",
            "paddingBlock",
            "paddingBlockEnd",
            "paddingBlockStart",
            "paddingBottom",
            "paddingInline",
            "paddingInlineEnd",
            "paddingInlineStart",
            "paddingLeft",
            "paddingRight",
            "paddingTop",
            "page",
            "page-break-after",
            "page-break-before",
            "page-break-inside",
            "pageBreakAfter",
            "pageBreakBefore",
            "pageBreakInside",
            "pageCount",
            "pageLeft",
            "pageTop",
            "pageX",
            "pageXOffset",
            "pageY",
            "pageYOffset",
            "pages",
            "paint-order",
            "paintOrder",
            "paintRequests",
            "paintType",
            "paintWorklet",
            "palette",
            "pan",
            "panningModel",
            "parameters",
            "parent",
            "parentElement",
            "parentNode",
            "parentRule",
            "parentStyleSheet",
            "parentTextEdit",
            "parentWindow",
            "parse",
            "parseAll",
            "parseFloat",
            "parseFromString",
            "parseInt",
            "part",
            "participants",
            "passive",
            "password",
            "pasteHTML",
            "path",
            "pathLength",
            "pathSegList",
            "pathSegType",
            "pathSegTypeAsLetter",
            "pathname",
            "pattern",
            "patternContentUnits",
            "patternMismatch",
            "patternTransform",
            "patternUnits",
            "pause",
            "pauseAnimations",
            "pauseOnExit",
            "pauseProfilers",
            "pauseTransformFeedback",
            "paused",
            "payerEmail",
            "payerName",
            "payerPhone",
            "paymentManager",
            "pc",
            "peerIdentity",
            "pending",
            "pendingLocalDescription",
            "pendingRemoteDescription",
            "percent",
            "performance",
            "periodicSync",
            "permission",
            "permissionState",
            "permissions",
            "persist",
            "persisted",
            "personalbar",
            "perspective",
            "perspective-origin",
            "perspectiveOrigin",
            "phone",
            "phoneticFamilyName",
            "phoneticGivenName",
            "photo",
            "pictureInPictureElement",
            "pictureInPictureEnabled",
            "pictureInPictureWindow",
            "ping",
            "pipeThrough",
            "pipeTo",
            "pitch",
            "pixelBottom",
            "pixelDepth",
            "pixelHeight",
            "pixelLeft",
            "pixelRight",
            "pixelStorei",
            "pixelTop",
            "pixelUnitToMillimeterX",
            "pixelUnitToMillimeterY",
            "pixelWidth",
            "place-content",
            "place-items",
            "place-self",
            "placeContent",
            "placeItems",
            "placeSelf",
            "placeholder",
            "platform",
            "platforms",
            "play",
            "playEffect",
            "playState",
            "playbackRate",
            "playbackState",
            "playbackTime",
            "played",
            "playoutDelayHint",
            "playsInline",
            "plugins",
            "pluginspage",
            "pname",
            "pointer-events",
            "pointerBeforeReferenceNode",
            "pointerEnabled",
            "pointerEvents",
            "pointerId",
            "pointerLockElement",
            "pointerType",
            "points",
            "pointsAtX",
            "pointsAtY",
            "pointsAtZ",
            "polygonOffset",
            "pop",
            "populateMatrix",
            "popupWindowFeatures",
            "popupWindowName",
            "popupWindowURI",
            "port",
            "port1",
            "port2",
            "ports",
            "posBottom",
            "posHeight",
            "posLeft",
            "posRight",
            "posTop",
            "posWidth",
            "pose",
            "position",
            "positionAlign",
            "positionX",
            "positionY",
            "positionZ",
            "postError",
            "postMessage",
            "postalCode",
            "poster",
            "pow",
            "powerEfficient",
            "powerOff",
            "preMultiplySelf",
            "precision",
            "preferredStyleSheetSet",
            "preferredStylesheetSet",
            "prefix",
            "preload",
            "prepend",
            "presentation",
            "preserveAlpha",
            "preserveAspectRatio",
            "preserveAspectRatioString",
            "pressed",
            "pressure",
            "prevValue",
            "preventDefault",
            "preventExtensions",
            "preventSilentAccess",
            "previousElementSibling",
            "previousNode",
            "previousPage",
            "previousRect",
            "previousScale",
            "previousSibling",
            "previousTranslate",
            "primaryKey",
            "primitiveType",
            "primitiveUnits",
            "principals",
            "print",
            "priority",
            "privateKey",
            "probablySupportsContext",
            "process",
            "processIceMessage",
            "processingEnd",
            "processingStart",
            "product",
            "productId",
            "productName",
            "productSub",
            "profile",
            "profileEnd",
            "profiles",
            "projectionMatrix",
            "promise",
            "prompt",
            "properties",
            "propertyIsEnumerable",
            "propertyName",
            "protocol",
            "protocolLong",
            "prototype",
            "provider",
            "pseudoClass",
            "pseudoElement",
            "pt",
            "publicId",
            "publicKey",
            "published",
            "pulse",
            "push",
            "pushManager",
            "pushNotification",
            "pushState",
            "put",
            "putImageData",
            "px",
            "quadraticCurveTo",
            "qualifier",
            "quaternion",
            "query",
            "queryCommandEnabled",
            "queryCommandIndeterm",
            "queryCommandState",
            "queryCommandSupported",
            "queryCommandText",
            "queryCommandValue",
            "querySelector",
            "querySelectorAll",
            "queueMicrotask",
            "quote",
            "quotes",
            "r",
            "r1",
            "r2",
            "race",
            "rad",
            "radiogroup",
            "radiusX",
            "radiusY",
            "random",
            "range",
            "rangeCount",
            "rangeMax",
            "rangeMin",
            "rangeOffset",
            "rangeOverflow",
            "rangeParent",
            "rangeUnderflow",
            "rate",
            "ratio",
            "raw",
            "rawId",
            "read",
            "readAsArrayBuffer",
            "readAsBinaryString",
            "readAsBlob",
            "readAsDataURL",
            "readAsText",
            "readBuffer",
            "readEntries",
            "readOnly",
            "readPixels",
            "readReportRequested",
            "readText",
            "readValue",
            "readable",
            "ready",
            "readyState",
            "reason",
            "reboot",
            "receivedAlert",
            "receiver",
            "receivers",
            "recipient",
            "reconnect",
            "recordNumber",
            "recordsAvailable",
            "recordset",
            "rect",
            "red",
            "redEyeReduction",
            "redirect",
            "redirectCount",
            "redirectEnd",
            "redirectStart",
            "redirected",
            "reduce",
            "reduceRight",
            "reduction",
            "refDistance",
            "refX",
            "refY",
            "referenceNode",
            "referenceSpace",
            "referrer",
            "referrerPolicy",
            "refresh",
            "region",
            "regionAnchorX",
            "regionAnchorY",
            "regionId",
            "regions",
            "register",
            "registerContentHandler",
            "registerElement",
            "registerProperty",
            "registerProtocolHandler",
            "reject",
            "rel",
            "relList",
            "relatedAddress",
            "relatedNode",
            "relatedPort",
            "relatedTarget",
            "release",
            "releaseCapture",
            "releaseEvents",
            "releaseInterface",
            "releaseLock",
            "releasePointerCapture",
            "releaseShaderCompiler",
            "reliable",
            "reliableWrite",
            "reload",
            "rem",
            "remainingSpace",
            "remote",
            "remoteDescription",
            "remove",
            "removeAllRanges",
            "removeAttribute",
            "removeAttributeNS",
            "removeAttributeNode",
            "removeBehavior",
            "removeChild",
            "removeCue",
            "removeEventListener",
            "removeFilter",
            "removeImport",
            "removeItem",
            "removeListener",
            "removeNamedItem",
            "removeNamedItemNS",
            "removeNode",
            "removeParameter",
            "removeProperty",
            "removeRange",
            "removeRegion",
            "removeRule",
            "removeSiteSpecificTrackingException",
            "removeSourceBuffer",
            "removeStream",
            "removeTrack",
            "removeVariable",
            "removeWakeLockListener",
            "removeWebWideTrackingException",
            "removed",
            "removedNodes",
            "renderHeight",
            "renderState",
            "renderTime",
            "renderWidth",
            "renderbufferStorage",
            "renderbufferStorageMultisample",
            "renderedBuffer",
            "renderingMode",
            "renotify",
            "repeat",
            "replace",
            "replaceAdjacentText",
            "replaceAll",
            "replaceChild",
            "replaceChildren",
            "replaceData",
            "replaceId",
            "replaceItem",
            "replaceNode",
            "replaceState",
            "replaceSync",
            "replaceTrack",
            "replaceWholeText",
            "replaceWith",
            "reportValidity",
            "request",
            "requestAnimationFrame",
            "requestAutocomplete",
            "requestData",
            "requestDevice",
            "requestFrame",
            "requestFullscreen",
            "requestHitTestSource",
            "requestHitTestSourceForTransientInput",
            "requestId",
            "requestIdleCallback",
            "requestMIDIAccess",
            "requestMediaKeySystemAccess",
            "requestPermission",
            "requestPictureInPicture",
            "requestPointerLock",
            "requestPresent",
            "requestReferenceSpace",
            "requestSession",
            "requestStart",
            "requestStorageAccess",
            "requestSubmit",
            "requestVideoFrameCallback",
            "requestingWindow",
            "requireInteraction",
            "required",
            "requiredExtensions",
            "requiredFeatures",
            "reset",
            "resetPose",
            "resetTransform",
            "resize",
            "resizeBy",
            "resizeTo",
            "resolve",
            "response",
            "responseBody",
            "responseEnd",
            "responseReady",
            "responseStart",
            "responseText",
            "responseType",
            "responseURL",
            "responseXML",
            "restartIce",
            "restore",
            "result",
            "resultIndex",
            "resultType",
            "results",
            "resume",
            "resumeProfilers",
            "resumeTransformFeedback",
            "retry",
            "returnValue",
            "rev",
            "reverse",
            "reversed",
            "revocable",
            "revokeObjectURL",
            "rgbColor",
            "right",
            "rightContext",
            "rightDegrees",
            "rightMargin",
            "rightProjectionMatrix",
            "rightViewMatrix",
            "role",
            "rolloffFactor",
            "root",
            "rootBounds",
            "rootElement",
            "rootMargin",
            "rotate",
            "rotateAxisAngle",
            "rotateAxisAngleSelf",
            "rotateFromVector",
            "rotateFromVectorSelf",
            "rotateSelf",
            "rotation",
            "rotationAngle",
            "rotationRate",
            "round",
            "row-gap",
            "rowGap",
            "rowIndex",
            "rowSpan",
            "rows",
            "rtcpTransport",
            "rtt",
            "ruby-align",
            "ruby-position",
            "rubyAlign",
            "rubyOverhang",
            "rubyPosition",
            "rules",
            "runtime",
            "runtimeStyle",
            "rx",
            "ry",
            "s",
            "safari",
            "sample",
            "sampleCoverage",
            "sampleRate",
            "samplerParameterf",
            "samplerParameteri",
            "sandbox",
            "save",
            "saveData",
            "scale",
            "scale3d",
            "scale3dSelf",
            "scaleNonUniform",
            "scaleNonUniformSelf",
            "scaleSelf",
            "scheme",
            "scissor",
            "scope",
            "scopeName",
            "scoped",
            "screen",
            "screenBrightness",
            "screenEnabled",
            "screenLeft",
            "screenPixelToMillimeterX",
            "screenPixelToMillimeterY",
            "screenTop",
            "screenX",
            "screenY",
            "scriptURL",
            "scripts",
            "scroll",
            "scroll-behavior",
            "scroll-margin",
            "scroll-margin-block",
            "scroll-margin-block-end",
            "scroll-margin-block-start",
            "scroll-margin-bottom",
            "scroll-margin-inline",
            "scroll-margin-inline-end",
            "scroll-margin-inline-start",
            "scroll-margin-left",
            "scroll-margin-right",
            "scroll-margin-top",
            "scroll-padding",
            "scroll-padding-block",
            "scroll-padding-block-end",
            "scroll-padding-block-start",
            "scroll-padding-bottom",
            "scroll-padding-inline",
            "scroll-padding-inline-end",
            "scroll-padding-inline-start",
            "scroll-padding-left",
            "scroll-padding-right",
            "scroll-padding-top",
            "scroll-snap-align",
            "scroll-snap-type",
            "scrollAmount",
            "scrollBehavior",
            "scrollBy",
            "scrollByLines",
            "scrollByPages",
            "scrollDelay",
            "scrollHeight",
            "scrollIntoView",
            "scrollIntoViewIfNeeded",
            "scrollLeft",
            "scrollLeftMax",
            "scrollMargin",
            "scrollMarginBlock",
            "scrollMarginBlockEnd",
            "scrollMarginBlockStart",
            "scrollMarginBottom",
            "scrollMarginInline",
            "scrollMarginInlineEnd",
            "scrollMarginInlineStart",
            "scrollMarginLeft",
            "scrollMarginRight",
            "scrollMarginTop",
            "scrollMaxX",
            "scrollMaxY",
            "scrollPadding",
            "scrollPaddingBlock",
            "scrollPaddingBlockEnd",
            "scrollPaddingBlockStart",
            "scrollPaddingBottom",
            "scrollPaddingInline",
            "scrollPaddingInlineEnd",
            "scrollPaddingInlineStart",
            "scrollPaddingLeft",
            "scrollPaddingRight",
            "scrollPaddingTop",
            "scrollRestoration",
            "scrollSnapAlign",
            "scrollSnapType",
            "scrollTo",
            "scrollTop",
            "scrollTopMax",
            "scrollWidth",
            "scrollX",
            "scrollY",
            "scrollbar-color",
            "scrollbar-width",
            "scrollbar3dLightColor",
            "scrollbarArrowColor",
            "scrollbarBaseColor",
            "scrollbarColor",
            "scrollbarDarkShadowColor",
            "scrollbarFaceColor",
            "scrollbarHighlightColor",
            "scrollbarShadowColor",
            "scrollbarTrackColor",
            "scrollbarWidth",
            "scrollbars",
            "scrolling",
            "scrollingElement",
            "sctp",
            "sctpCauseCode",
            "sdp",
            "sdpLineNumber",
            "sdpMLineIndex",
            "sdpMid",
            "seal",
            "search",
            "searchBox",
            "searchBoxJavaBridge_",
            "searchParams",
            "sectionRowIndex",
            "secureConnectionStart",
            "security",
            "seed",
            "seekToNextFrame",
            "seekable",
            "seeking",
            "select",
            "selectAllChildren",
            "selectAlternateInterface",
            "selectConfiguration",
            "selectNode",
            "selectNodeContents",
            "selectNodes",
            "selectSingleNode",
            "selectSubString",
            "selected",
            "selectedIndex",
            "selectedOptions",
            "selectedStyleSheetSet",
            "selectedStylesheetSet",
            "selection",
            "selectionDirection",
            "selectionEnd",
            "selectionStart",
            "selector",
            "selectorText",
            "self",
            "send",
            "sendAsBinary",
            "sendBeacon",
            "sender",
            "sentAlert",
            "sentTimestamp",
            "separator",
            "serialNumber",
            "serializeToString",
            "serverTiming",
            "service",
            "serviceWorker",
            "session",
            "sessionId",
            "sessionStorage",
            "set",
            "setActionHandler",
            "setActive",
            "setAlpha",
            "setAppBadge",
            "setAttribute",
            "setAttributeNS",
            "setAttributeNode",
            "setAttributeNodeNS",
            "setBaseAndExtent",
            "setBigInt64",
            "setBigUint64",
            "setBingCurrentSearchDefault",
            "setCapture",
            "setCodecPreferences",
            "setColor",
            "setCompositeOperation",
            "setConfiguration",
            "setCurrentTime",
            "setCustomValidity",
            "setData",
            "setDate",
            "setDragImage",
            "setEnd",
            "setEndAfter",
            "setEndBefore",
            "setEndPoint",
            "setFillColor",
            "setFilterRes",
            "setFloat32",
            "setFloat64",
            "setFloatValue",
            "setFormValue",
            "setFullYear",
            "setHeaderValue",
            "setHours",
            "setIdentityProvider",
            "setImmediate",
            "setInt16",
            "setInt32",
            "setInt8",
            "setInterval",
            "setItem",
            "setKeyframes",
            "setLineCap",
            "setLineDash",
            "setLineJoin",
            "setLineWidth",
            "setLiveSeekableRange",
            "setLocalDescription",
            "setMatrix",
            "setMatrixValue",
            "setMediaKeys",
            "setMilliseconds",
            "setMinutes",
            "setMiterLimit",
            "setMonth",
            "setNamedItem",
            "setNamedItemNS",
            "setNonUserCodeExceptions",
            "setOrientToAngle",
            "setOrientToAuto",
            "setOrientation",
            "setOverrideHistoryNavigationMode",
            "setPaint",
            "setParameter",
            "setParameters",
            "setPeriodicWave",
            "setPointerCapture",
            "setPosition",
            "setPositionState",
            "setPreference",
            "setProperty",
            "setPrototypeOf",
            "setRGBColor",
            "setRGBColorICCColor",
            "setRadius",
            "setRangeText",
            "setRemoteDescription",
            "setRequestHeader",
            "setResizable",
            "setResourceTimingBufferSize",
            "setRotate",
            "setScale",
            "setSeconds",
            "setSelectionRange",
            "setServerCertificate",
            "setShadow",
            "setSinkId",
            "setSkewX",
            "setSkewY",
            "setStart",
            "setStartAfter",
            "setStartBefore",
            "setStdDeviation",
            "setStreams",
            "setStringValue",
            "setStrokeColor",
            "setSuggestResult",
            "setTargetAtTime",
            "setTargetValueAtTime",
            "setTime",
            "setTimeout",
            "setTransform",
            "setTranslate",
            "setUTCDate",
            "setUTCFullYear",
            "setUTCHours",
            "setUTCMilliseconds",
            "setUTCMinutes",
            "setUTCMonth",
            "setUTCSeconds",
            "setUint16",
            "setUint32",
            "setUint8",
            "setUri",
            "setValidity",
            "setValueAtTime",
            "setValueCurveAtTime",
            "setVariable",
            "setVelocity",
            "setVersion",
            "setYear",
            "settingName",
            "settingValue",
            "sex",
            "shaderSource",
            "shadowBlur",
            "shadowColor",
            "shadowOffsetX",
            "shadowOffsetY",
            "shadowRoot",
            "shape",
            "shape-image-threshold",
            "shape-margin",
            "shape-outside",
            "shape-rendering",
            "shapeImageThreshold",
            "shapeMargin",
            "shapeOutside",
            "shapeRendering",
            "sheet",
            "shift",
            "shiftKey",
            "shiftLeft",
            "shippingAddress",
            "shippingOption",
            "shippingType",
            "show",
            "showHelp",
            "showModal",
            "showModalDialog",
            "showModelessDialog",
            "showNotification",
            "sidebar",
            "sign",
            "signal",
            "signalingState",
            "signature",
            "silent",
            "sin",
            "singleNodeValue",
            "sinh",
            "sinkId",
            "sittingToStandingTransform",
            "size",
            "sizeToContent",
            "sizeX",
            "sizeZ",
            "sizes",
            "skewX",
            "skewXSelf",
            "skewY",
            "skewYSelf",
            "slice",
            "slope",
            "slot",
            "small",
            "smil",
            "smooth",
            "smoothingTimeConstant",
            "snapToLines",
            "snapshotItem",
            "snapshotLength",
            "some",
            "sort",
            "sortingCode",
            "source",
            "sourceBuffer",
            "sourceBuffers",
            "sourceCapabilities",
            "sourceFile",
            "sourceIndex",
            "sources",
            "spacing",
            "span",
            "speak",
            "speakAs",
            "speaking",
            "species",
            "specified",
            "specularConstant",
            "specularExponent",
            "speechSynthesis",
            "speed",
            "speedOfSound",
            "spellcheck",
            "splice",
            "split",
            "splitText",
            "spreadMethod",
            "sqrt",
            "src",
            "srcElement",
            "srcFilter",
            "srcObject",
            "srcUrn",
            "srcdoc",
            "srclang",
            "srcset",
            "stack",
            "stackTraceLimit",
            "stacktrace",
            "stageParameters",
            "standalone",
            "standby",
            "start",
            "startContainer",
            "startIce",
            "startMessages",
            "startNotifications",
            "startOffset",
            "startProfiling",
            "startRendering",
            "startShark",
            "startTime",
            "startsWith",
            "state",
            "status",
            "statusCode",
            "statusMessage",
            "statusText",
            "statusbar",
            "stdDeviationX",
            "stdDeviationY",
            "stencilFunc",
            "stencilFuncSeparate",
            "stencilMask",
            "stencilMaskSeparate",
            "stencilOp",
            "stencilOpSeparate",
            "step",
            "stepDown",
            "stepMismatch",
            "stepUp",
            "sticky",
            "stitchTiles",
            "stop",
            "stop-color",
            "stop-opacity",
            "stopColor",
            "stopImmediatePropagation",
            "stopNotifications",
            "stopOpacity",
            "stopProfiling",
            "stopPropagation",
            "stopShark",
            "stopped",
            "storage",
            "storageArea",
            "storageName",
            "storageStatus",
            "store",
            "storeSiteSpecificTrackingException",
            "storeWebWideTrackingException",
            "stpVersion",
            "stream",
            "streams",
            "stretch",
            "strike",
            "string",
            "stringValue",
            "stringify",
            "stroke",
            "stroke-dasharray",
            "stroke-dashoffset",
            "stroke-linecap",
            "stroke-linejoin",
            "stroke-miterlimit",
            "stroke-opacity",
            "stroke-width",
            "strokeDasharray",
            "strokeDashoffset",
            "strokeLinecap",
            "strokeLinejoin",
            "strokeMiterlimit",
            "strokeOpacity",
            "strokeRect",
            "strokeStyle",
            "strokeText",
            "strokeWidth",
            "style",
            "styleFloat",
            "styleMap",
            "styleMedia",
            "styleSheet",
            "styleSheetSets",
            "styleSheets",
            "sub",
            "subarray",
            "subject",
            "submit",
            "submitFrame",
            "submitter",
            "subscribe",
            "substr",
            "substring",
            "substringData",
            "subtle",
            "subtree",
            "suffix",
            "suffixes",
            "summary",
            "sup",
            "supported",
            "supportedContentEncodings",
            "supportedEntryTypes",
            "supports",
            "supportsSession",
            "surfaceScale",
            "surroundContents",
            "suspend",
            "suspendRedraw",
            "swapCache",
            "swapNode",
            "sweepFlag",
            "symbols",
            "sync",
            "sysexEnabled",
            "system",
            "systemCode",
            "systemId",
            "systemLanguage",
            "systemXDPI",
            "systemYDPI",
            "tBodies",
            "tFoot",
            "tHead",
            "tabIndex",
            "table",
            "table-layout",
            "tableLayout",
            "tableValues",
            "tag",
            "tagName",
            "tagUrn",
            "tags",
            "taintEnabled",
            "takePhoto",
            "takeRecords",
            "tan",
            "tangentialPressure",
            "tanh",
            "target",
            "targetElement",
            "targetRayMode",
            "targetRaySpace",
            "targetTouches",
            "targetX",
            "targetY",
            "tcpType",
            "tee",
            "tel",
            "terminate",
            "test",
            "texImage2D",
            "texImage3D",
            "texParameterf",
            "texParameteri",
            "texStorage2D",
            "texStorage3D",
            "texSubImage2D",
            "texSubImage3D",
            "text",
            "text-align",
            "text-align-last",
            "text-anchor",
            "text-combine-upright",
            "text-decoration",
            "text-decoration-color",
            "text-decoration-line",
            "text-decoration-skip-ink",
            "text-decoration-style",
            "text-decoration-thickness",
            "text-emphasis",
            "text-emphasis-color",
            "text-emphasis-position",
            "text-emphasis-style",
            "text-indent",
            "text-justify",
            "text-orientation",
            "text-overflow",
            "text-rendering",
            "text-shadow",
            "text-transform",
            "text-underline-offset",
            "text-underline-position",
            "textAlign",
            "textAlignLast",
            "textAnchor",
            "textAutospace",
            "textBaseline",
            "textCombineUpright",
            "textContent",
            "textDecoration",
            "textDecorationBlink",
            "textDecorationColor",
            "textDecorationLine",
            "textDecorationLineThrough",
            "textDecorationNone",
            "textDecorationOverline",
            "textDecorationSkipInk",
            "textDecorationStyle",
            "textDecorationThickness",
            "textDecorationUnderline",
            "textEmphasis",
            "textEmphasisColor",
            "textEmphasisPosition",
            "textEmphasisStyle",
            "textIndent",
            "textJustify",
            "textJustifyTrim",
            "textKashida",
            "textKashidaSpace",
            "textLength",
            "textOrientation",
            "textOverflow",
            "textRendering",
            "textShadow",
            "textTracks",
            "textTransform",
            "textUnderlineOffset",
            "textUnderlinePosition",
            "then",
            "threadId",
            "threshold",
            "thresholds",
            "tiltX",
            "tiltY",
            "time",
            "timeEnd",
            "timeLog",
            "timeOrigin",
            "timeRemaining",
            "timeStamp",
            "timecode",
            "timeline",
            "timelineTime",
            "timeout",
            "timestamp",
            "timestampOffset",
            "timing",
            "title",
            "to",
            "toArray",
            "toBlob",
            "toDataURL",
            "toDateString",
            "toElement",
            "toExponential",
            "toFixed",
            "toFloat32Array",
            "toFloat64Array",
            "toGMTString",
            "toISOString",
            "toJSON",
            "toLocaleDateString",
            "toLocaleFormat",
            "toLocaleLowerCase",
            "toLocaleString",
            "toLocaleTimeString",
            "toLocaleUpperCase",
            "toLowerCase",
            "toMatrix",
            "toMethod",
            "toPrecision",
            "toPrimitive",
            "toSdp",
            "toSource",
            "toStaticHTML",
            "toString",
            "toStringTag",
            "toSum",
            "toTimeString",
            "toUTCString",
            "toUpperCase",
            "toggle",
            "toggleAttribute",
            "toggleLongPressEnabled",
            "tone",
            "toneBuffer",
            "tooLong",
            "tooShort",
            "toolbar",
            "top",
            "topMargin",
            "total",
            "totalFrameDelay",
            "totalVideoFrames",
            "touch-action",
            "touchAction",
            "touched",
            "touches",
            "trace",
            "track",
            "trackVisibility",
            "transaction",
            "transactions",
            "transceiver",
            "transferControlToOffscreen",
            "transferFromImageBitmap",
            "transferImageBitmap",
            "transferIn",
            "transferOut",
            "transferSize",
            "transferToImageBitmap",
            "transform",
            "transform-box",
            "transform-origin",
            "transform-style",
            "transformBox",
            "transformFeedbackVaryings",
            "transformOrigin",
            "transformPoint",
            "transformString",
            "transformStyle",
            "transformToDocument",
            "transformToFragment",
            "transition",
            "transition-delay",
            "transition-duration",
            "transition-property",
            "transition-timing-function",
            "transitionDelay",
            "transitionDuration",
            "transitionProperty",
            "transitionTimingFunction",
            "translate",
            "translateSelf",
            "translationX",
            "translationY",
            "transport",
            "trim",
            "trimEnd",
            "trimLeft",
            "trimRight",
            "trimStart",
            "trueSpeed",
            "trunc",
            "truncate",
            "trustedTypes",
            "turn",
            "twist",
            "type",
            "typeDetail",
            "typeMismatch",
            "typeMustMatch",
            "types",
            "u2f",
            "ubound",
            "uint16",
            "uint32",
            "uint8",
            "uint8Clamped",
            "undefined",
            "unescape",
            "uneval",
            "unicode",
            "unicode-bidi",
            "unicodeBidi",
            "unicodeRange",
            "uniform1f",
            "uniform1fv",
            "uniform1i",
            "uniform1iv",
            "uniform1ui",
            "uniform1uiv",
            "uniform2f",
            "uniform2fv",
            "uniform2i",
            "uniform2iv",
            "uniform2ui",
            "uniform2uiv",
            "uniform3f",
            "uniform3fv",
            "uniform3i",
            "uniform3iv",
            "uniform3ui",
            "uniform3uiv",
            "uniform4f",
            "uniform4fv",
            "uniform4i",
            "uniform4iv",
            "uniform4ui",
            "uniform4uiv",
            "uniformBlockBinding",
            "uniformMatrix2fv",
            "uniformMatrix2x3fv",
            "uniformMatrix2x4fv",
            "uniformMatrix3fv",
            "uniformMatrix3x2fv",
            "uniformMatrix3x4fv",
            "uniformMatrix4fv",
            "uniformMatrix4x2fv",
            "uniformMatrix4x3fv",
            "unique",
            "uniqueID",
            "uniqueNumber",
            "unit",
            "unitType",
            "units",
            "unloadEventEnd",
            "unloadEventStart",
            "unlock",
            "unmount",
            "unobserve",
            "unpause",
            "unpauseAnimations",
            "unreadCount",
            "unregister",
            "unregisterContentHandler",
            "unregisterProtocolHandler",
            "unscopables",
            "unselectable",
            "unshift",
            "unsubscribe",
            "unsuspendRedraw",
            "unsuspendRedrawAll",
            "unwatch",
            "unwrapKey",
            "upDegrees",
            "upX",
            "upY",
            "upZ",
            "update",
            "updateCommands",
            "updateIce",
            "updateInterval",
            "updatePlaybackRate",
            "updateRenderState",
            "updateSettings",
            "updateTiming",
            "updateViaCache",
            "updateWith",
            "updated",
            "updating",
            "upgrade",
            "upload",
            "uploadTotal",
            "uploaded",
            "upper",
            "upperBound",
            "upperOpen",
            "uri",
            "url",
            "urn",
            "urns",
            "usages",
            "usb",
            "usbVersionMajor",
            "usbVersionMinor",
            "usbVersionSubminor",
            "useCurrentView",
            "useMap",
            "useProgram",
            "usedSpace",
            "user-select",
            "userActivation",
            "userAgent",
            "userChoice",
            "userHandle",
            "userHint",
            "userLanguage",
            "userSelect",
            "userVisibleOnly",
            "username",
            "usernameFragment",
            "utterance",
            "uuid",
            "v8BreakIterator",
            "vAlign",
            "vLink",
            "valid",
            "validate",
            "validateProgram",
            "validationMessage",
            "validity",
            "value",
            "valueAsDate",
            "valueAsNumber",
            "valueAsString",
            "valueInSpecifiedUnits",
            "valueMissing",
            "valueOf",
            "valueText",
            "valueType",
            "values",
            "variable",
            "variant",
            "variationSettings",
            "vector-effect",
            "vectorEffect",
            "velocityAngular",
            "velocityExpansion",
            "velocityX",
            "velocityY",
            "vendor",
            "vendorId",
            "vendorSub",
            "verify",
            "version",
            "vertexAttrib1f",
            "vertexAttrib1fv",
            "vertexAttrib2f",
            "vertexAttrib2fv",
            "vertexAttrib3f",
            "vertexAttrib3fv",
            "vertexAttrib4f",
            "vertexAttrib4fv",
            "vertexAttribDivisor",
            "vertexAttribDivisorANGLE",
            "vertexAttribI4i",
            "vertexAttribI4iv",
            "vertexAttribI4ui",
            "vertexAttribI4uiv",
            "vertexAttribIPointer",
            "vertexAttribPointer",
            "vertical",
            "vertical-align",
            "verticalAlign",
            "verticalOverflow",
            "vh",
            "vibrate",
            "vibrationActuator",
            "videoBitsPerSecond",
            "videoHeight",
            "videoTracks",
            "videoWidth",
            "view",
            "viewBox",
            "viewBoxString",
            "viewTarget",
            "viewTargetString",
            "viewport",
            "viewportAnchorX",
            "viewportAnchorY",
            "viewportElement",
            "views",
            "violatedDirective",
            "visibility",
            "visibilityState",
            "visible",
            "visualViewport",
            "vlinkColor",
            "vmax",
            "vmin",
            "voice",
            "voiceURI",
            "volume",
            "vrml",
            "vspace",
            "vw",
            "w",
            "wait",
            "waitSync",
            "waiting",
            "wake",
            "wakeLock",
            "wand",
            "warn",
            "wasClean",
            "wasDiscarded",
            "watch",
            "watchAvailability",
            "watchPosition",
            "webdriver",
            "webkitAddKey",
            "webkitAlignContent",
            "webkitAlignItems",
            "webkitAlignSelf",
            "webkitAnimation",
            "webkitAnimationDelay",
            "webkitAnimationDirection",
            "webkitAnimationDuration",
            "webkitAnimationFillMode",
            "webkitAnimationIterationCount",
            "webkitAnimationName",
            "webkitAnimationPlayState",
            "webkitAnimationTimingFunction",
            "webkitAppearance",
            "webkitAudioContext",
            "webkitAudioDecodedByteCount",
            "webkitAudioPannerNode",
            "webkitBackfaceVisibility",
            "webkitBackground",
            "webkitBackgroundAttachment",
            "webkitBackgroundClip",
            "webkitBackgroundColor",
            "webkitBackgroundImage",
            "webkitBackgroundOrigin",
            "webkitBackgroundPosition",
            "webkitBackgroundPositionX",
            "webkitBackgroundPositionY",
            "webkitBackgroundRepeat",
            "webkitBackgroundSize",
            "webkitBackingStorePixelRatio",
            "webkitBorderBottomLeftRadius",
            "webkitBorderBottomRightRadius",
            "webkitBorderImage",
            "webkitBorderImageOutset",
            "webkitBorderImageRepeat",
            "webkitBorderImageSlice",
            "webkitBorderImageSource",
            "webkitBorderImageWidth",
            "webkitBorderRadius",
            "webkitBorderTopLeftRadius",
            "webkitBorderTopRightRadius",
            "webkitBoxAlign",
            "webkitBoxDirection",
            "webkitBoxFlex",
            "webkitBoxOrdinalGroup",
            "webkitBoxOrient",
            "webkitBoxPack",
            "webkitBoxShadow",
            "webkitBoxSizing",
            "webkitCancelAnimationFrame",
            "webkitCancelFullScreen",
            "webkitCancelKeyRequest",
            "webkitCancelRequestAnimationFrame",
            "webkitClearResourceTimings",
            "webkitClosedCaptionsVisible",
            "webkitConvertPointFromNodeToPage",
            "webkitConvertPointFromPageToNode",
            "webkitCreateShadowRoot",
            "webkitCurrentFullScreenElement",
            "webkitCurrentPlaybackTargetIsWireless",
            "webkitDecodedFrameCount",
            "webkitDirectionInvertedFromDevice",
            "webkitDisplayingFullscreen",
            "webkitDroppedFrameCount",
            "webkitEnterFullScreen",
            "webkitEnterFullscreen",
            "webkitEntries",
            "webkitExitFullScreen",
            "webkitExitFullscreen",
            "webkitExitPointerLock",
            "webkitFilter",
            "webkitFlex",
            "webkitFlexBasis",
            "webkitFlexDirection",
            "webkitFlexFlow",
            "webkitFlexGrow",
            "webkitFlexShrink",
            "webkitFlexWrap",
            "webkitFullScreenKeyboardInputAllowed",
            "webkitFullscreenElement",
            "webkitFullscreenEnabled",
            "webkitGenerateKeyRequest",
            "webkitGetAsEntry",
            "webkitGetDatabaseNames",
            "webkitGetEntries",
            "webkitGetEntriesByName",
            "webkitGetEntriesByType",
            "webkitGetFlowByName",
            "webkitGetGamepads",
            "webkitGetImageDataHD",
            "webkitGetNamedFlows",
            "webkitGetRegionFlowRanges",
            "webkitGetUserMedia",
            "webkitHasClosedCaptions",
            "webkitHidden",
            "webkitIDBCursor",
            "webkitIDBDatabase",
            "webkitIDBDatabaseError",
            "webkitIDBDatabaseException",
            "webkitIDBFactory",
            "webkitIDBIndex",
            "webkitIDBKeyRange",
            "webkitIDBObjectStore",
            "webkitIDBRequest",
            "webkitIDBTransaction",
            "webkitImageSmoothingEnabled",
            "webkitIndexedDB",
            "webkitInitMessageEvent",
            "webkitIsFullScreen",
            "webkitJustifyContent",
            "webkitKeys",
            "webkitLineClamp",
            "webkitLineDashOffset",
            "webkitLockOrientation",
            "webkitMask",
            "webkitMaskClip",
            "webkitMaskComposite",
            "webkitMaskImage",
            "webkitMaskOrigin",
            "webkitMaskPosition",
            "webkitMaskPositionX",
            "webkitMaskPositionY",
            "webkitMaskRepeat",
            "webkitMaskSize",
            "webkitMatchesSelector",
            "webkitMediaStream",
            "webkitNotifications",
            "webkitOfflineAudioContext",
            "webkitOrder",
            "webkitOrientation",
            "webkitPeerConnection00",
            "webkitPersistentStorage",
            "webkitPerspective",
            "webkitPerspectiveOrigin",
            "webkitPointerLockElement",
            "webkitPostMessage",
            "webkitPreservesPitch",
            "webkitPutImageDataHD",
            "webkitRTCPeerConnection",
            "webkitRegionOverset",
            "webkitRelativePath",
            "webkitRequestAnimationFrame",
            "webkitRequestFileSystem",
            "webkitRequestFullScreen",
            "webkitRequestFullscreen",
            "webkitRequestPointerLock",
            "webkitResolveLocalFileSystemURL",
            "webkitSetMediaKeys",
            "webkitSetResourceTimingBufferSize",
            "webkitShadowRoot",
            "webkitShowPlaybackTargetPicker",
            "webkitSlice",
            "webkitSpeechGrammar",
            "webkitSpeechGrammarList",
            "webkitSpeechRecognition",
            "webkitSpeechRecognitionError",
            "webkitSpeechRecognitionEvent",
            "webkitStorageInfo",
            "webkitSupportsFullscreen",
            "webkitTemporaryStorage",
            "webkitTextFillColor",
            "webkitTextSizeAdjust",
            "webkitTextStroke",
            "webkitTextStrokeColor",
            "webkitTextStrokeWidth",
            "webkitTransform",
            "webkitTransformOrigin",
            "webkitTransformStyle",
            "webkitTransition",
            "webkitTransitionDelay",
            "webkitTransitionDuration",
            "webkitTransitionProperty",
            "webkitTransitionTimingFunction",
            "webkitURL",
            "webkitUnlockOrientation",
            "webkitUserSelect",
            "webkitVideoDecodedByteCount",
            "webkitVisibilityState",
            "webkitWirelessVideoPlaybackDisabled",
            "webkitdirectory",
            "webkitdropzone",
            "webstore",
            "weight",
            "whatToShow",
            "wheelDelta",
            "wheelDeltaX",
            "wheelDeltaY",
            "whenDefined",
            "which",
            "white-space",
            "whiteSpace",
            "wholeText",
            "widows",
            "width",
            "will-change",
            "willChange",
            "willValidate",
            "window",
            "withCredentials",
            "word-break",
            "word-spacing",
            "word-wrap",
            "wordBreak",
            "wordSpacing",
            "wordWrap",
            "workerStart",
            "wrap",
            "wrapKey",
            "writable",
            "writableAuxiliaries",
            "write",
            "writeText",
            "writeValue",
            "writeWithoutResponse",
            "writeln",
            "writing-mode",
            "writingMode",
            "x",
            "x1",
            "x2",
            "xChannelSelector",
            "xmlEncoding",
            "xmlStandalone",
            "xmlVersion",
            "xmlbase",
            "xmllang",
            "xmlspace",
            "xor",
            "xr",
            "y",
            "y1",
            "y2",
            "yChannelSelector",
            "yandex",
            "z",
            "z-index",
            "zIndex",
            "zoom",
            "zoomAndPan",
            "zoomRectScreen",
        ];
        /***********************************************************************
        
          A JavaScript tokenizer / parser / beautifier / compressor.
          https://github.com/mishoo/UglifyJS2
        
          -------------------------------- (C) ---------------------------------
        
                                   Author: Mihai Bazon
                                 <mihai.bazon@gmail.com>
                               http://mihai.bazon.net/blog
        
          Distributed under the BSD license:
        
            Copyright 2012 (c) Mihai Bazon <mihai.bazon@gmail.com>
        
            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions
            are met:
        
                * Redistributions of source code must retain the above
                  copyright notice, this list of conditions and the following
                  disclaimer.
        
                * Redistributions in binary form must reproduce the above
                  copyright notice, this list of conditions and the following
                  disclaimer in the documentation and/or other materials
                  provided with the distribution.
        
            THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
            EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
            IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
            PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
            LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
            OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
            PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
            PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
            THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
            TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
            THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
            SUCH DAMAGE.
        
         ***********************************************************************/
        function find_builtins(reserved) {
            domprops.forEach(add);
            // Compatibility fix for some standard defined globals not defined on every js environment
            var new_globals = ["Symbol", "Map", "Promise", "Proxy", "Reflect", "Set", "WeakMap", "WeakSet"];
            var objects = {};
            var global_ref = typeof global === "object" ? global : self;
            new_globals.forEach(function (new_global) {
                objects[new_global] = global_ref[new_global] || new Function();
            });
            [
                "null",
                "true",
                "false",
                "NaN",
                "Infinity",
                "-Infinity",
                "undefined",
            ].forEach(add);
            [Object, Array, Function, Number,
                String, Boolean, Error, Math,
                Date, RegExp, objects.Symbol, ArrayBuffer,
                DataView, decodeURI, decodeURIComponent,
                encodeURI, encodeURIComponent, eval, EvalError,
                Float32Array, Float64Array, Int8Array, Int16Array,
                Int32Array, isFinite, isNaN, JSON, objects.Map, parseFloat,
                parseInt, objects.Promise, objects.Proxy, RangeError, ReferenceError,
                objects.Reflect, objects.Set, SyntaxError, TypeError, Uint8Array,
                Uint8ClampedArray, Uint16Array, Uint32Array, URIError,
                objects.WeakMap, objects.WeakSet
            ].forEach(function (ctor) {
                Object.getOwnPropertyNames(ctor).map(add);
                if (ctor.prototype) {
                    Object.getOwnPropertyNames(ctor.prototype).map(add);
                }
            });
            function add(name) {
                reserved.add(name);
            }
        }
        function reserve_quoted_keys(ast, reserved) {
            function add(name) {
                push_uniq(reserved, name);
            }
            ast.walk(new TreeWalker(function (node) {
                if (node instanceof AST_ObjectKeyVal && node.quote) {
                    add(node.key);
                }
                else if (node instanceof AST_ObjectProperty && node.quote) {
                    add(node.key.name);
                }
                else if (node instanceof AST_Sub) {
                    addStrings(node.property, add);
                }
            }));
        }
        function addStrings(node, add) {
            node.walk(new TreeWalker(function (node) {
                if (node instanceof AST_Sequence) {
                    addStrings(node.tail_node(), add);
                }
                else if (node instanceof AST_String) {
                    add(node.value);
                }
                else if (node instanceof AST_Conditional) {
                    addStrings(node.consequent, add);
                    addStrings(node.alternative, add);
                }
                return true;
            }));
        }
        function mangle_properties(ast, options) {
            options = defaults(options, {
                builtins: false,
                cache: null,
                debug: false,
                keep_quoted: false,
                only_cache: false,
                regex: null,
                reserved: null,
                undeclared: false,
            }, true);
            var reserved_option = options.reserved;
            if (!Array.isArray(reserved_option))
                reserved_option = [reserved_option];
            var reserved = new Set(reserved_option);
            if (!options.builtins)
                find_builtins(reserved);
            var cname = -1;
            var cprivate = -1;
            var cache;
            var private_cache = new Map();
            if (options.cache) {
                cache = options.cache.props;
                cache.forEach(function (mangled_name) {
                    reserved.add(mangled_name);
                });
            }
            else {
                cache = new Map();
            }
            var regex = options.regex && new RegExp(options.regex);
            // note debug is either false (disabled), or a string of the debug suffix to use (enabled).
            // note debug may be enabled as an empty string, which is falsey. Also treat passing 'true'
            // the same as passing an empty string.
            var debug = options.debug !== false;
            var debug_name_suffix;
            if (debug) {
                debug_name_suffix = (options.debug === true ? "" : options.debug);
            }
            var names_to_mangle = new Set();
            var unmangleable = new Set();
            var private_properties = new Set();
            var keep_quoted_strict = options.keep_quoted === "strict";
            // step 1: find candidates to mangle
            ast.walk(new TreeWalker(function (node) {
                if (node instanceof AST_ClassPrivateProperty
                    || node instanceof AST_PrivateMethod) {
                    private_properties.add(node.key.name);
                }
                else if (node instanceof AST_DotHash) {
                    private_properties.add(node.property);
                }
                else if (node instanceof AST_ObjectKeyVal) {
                    if (typeof node.key == "string" &&
                        (!keep_quoted_strict || !node.quote)) {
                        add(node.key);
                    }
                }
                else if (node instanceof AST_ObjectProperty) {
                    // setter or getter, since KeyVal is handled above
                    if (!keep_quoted_strict || !node.key.end.quote) {
                        add(node.key.name);
                    }
                }
                else if (node instanceof AST_Dot) {
                    var declared = !!options.undeclared;
                    if (!declared) {
                        var root = node;
                        while (root.expression) {
                            root = root.expression;
                        }
                        declared = !(root.thedef && root.thedef.undeclared);
                    }
                    if (declared &&
                        (!keep_quoted_strict || !node.quote)) {
                        add(node.property);
                    }
                }
                else if (node instanceof AST_Sub) {
                    if (!keep_quoted_strict) {
                        addStrings(node.property, add);
                    }
                }
                else if (node instanceof AST_Call
                    && node.expression.print_to_string() == "Object.defineProperty") {
                    addStrings(node.args[1], add);
                }
                else if (node instanceof AST_Binary && node.operator === "in") {
                    addStrings(node.left, add);
                }
            }));
            // step 2: transform the tree, renaming properties
            return ast.transform(new TreeTransformer(function (node) {
                if (node instanceof AST_ClassPrivateProperty
                    || node instanceof AST_PrivateMethod) {
                    node.key.name = mangle_private(node.key.name);
                }
                else if (node instanceof AST_DotHash) {
                    node.property = mangle_private(node.property);
                }
                else if (node instanceof AST_ObjectKeyVal) {
                    if (typeof node.key == "string" &&
                        (!keep_quoted_strict || !node.quote)) {
                        node.key = mangle(node.key);
                    }
                }
                else if (node instanceof AST_ObjectProperty) {
                    // setter, getter, method or class field
                    if (!keep_quoted_strict || !node.key.end.quote) {
                        node.key.name = mangle(node.key.name);
                    }
                }
                else if (node instanceof AST_Dot) {
                    if (!keep_quoted_strict || !node.quote) {
                        node.property = mangle(node.property);
                    }
                }
                else if (!options.keep_quoted && node instanceof AST_Sub) {
                    node.property = mangleStrings(node.property);
                }
                else if (node instanceof AST_Call
                    && node.expression.print_to_string() == "Object.defineProperty") {
                    node.args[1] = mangleStrings(node.args[1]);
                }
                else if (node instanceof AST_Binary && node.operator === "in") {
                    node.left = mangleStrings(node.left);
                }
            }));
            // only function declarations after this line
            function can_mangle(name) {
                if (unmangleable.has(name))
                    return false;
                if (reserved.has(name))
                    return false;
                if (options.only_cache) {
                    return cache.has(name);
                }
                if (/^-?[0-9]+(\.[0-9]+)?(e[+-][0-9]+)?$/.test(name))
                    return false;
                return true;
            }
            function should_mangle(name) {
                if (regex && !regex.test(name))
                    return false;
                if (reserved.has(name))
                    return false;
                return cache.has(name)
                    || names_to_mangle.has(name);
            }
            function add(name) {
                if (can_mangle(name))
                    names_to_mangle.add(name);
                if (!should_mangle(name)) {
                    unmangleable.add(name);
                }
            }
            function mangle(name) {
                if (!should_mangle(name)) {
                    return name;
                }
                var mangled = cache.get(name);
                if (!mangled) {
                    if (debug) {
                        // debug mode: use a prefix and suffix to preserve readability, e.g. o.foo -> o._$foo$NNN_.
                        var debug_mangled = "_$" + name + "$" + debug_name_suffix + "_";
                        if (can_mangle(debug_mangled)) {
                            mangled = debug_mangled;
                        }
                    }
                    // either debug mode is off, or it is on and we could not use the mangled name
                    if (!mangled) {
                        do {
                            mangled = base54(++cname);
                        } while (!can_mangle(mangled));
                    }
                    cache.set(name, mangled);
                }
                return mangled;
            }
            function mangle_private(name) {
                let mangled = private_cache.get(name);
                if (!mangled) {
                    mangled = base54(++cprivate);
                    private_cache.set(name, mangled);
                }
                return mangled;
            }
            function mangleStrings(node) {
                return node.transform(new TreeTransformer(function (node) {
                    if (node instanceof AST_Sequence) {
                        var last = node.expressions.length - 1;
                        node.expressions[last] = mangleStrings(node.expressions[last]);
                    }
                    else if (node instanceof AST_String) {
                        node.value = mangle(node.value);
                    }
                    else if (node instanceof AST_Conditional) {
                        node.consequent = mangleStrings(node.consequent);
                        node.alternative = mangleStrings(node.alternative);
                    }
                    return node;
                }));
            }
        }
        var to_ascii = typeof atob == "undefined" ? function (b64) {
            return Buffer.from(b64, "base64").toString();
        } : atob;
        var to_base64 = typeof btoa == "undefined" ? function (str) {
            return Buffer.from(str).toString("base64");
        } : btoa;
        function read_source_map(code) {
            var match = /(?:^|[^.])\/\/# sourceMappingURL=data:application\/json(;[\w=-]*)?;base64,([+/0-9A-Za-z]*=*)\s*$/.exec(code);
            if (!match) {
                console.warn("inline source map not found");
                return null;
            }
            return to_ascii(match[2]);
        }
        function set_shorthand(name, options, keys) {
            if (options[name]) {
                keys.forEach(function (key) {
                    if (options[key]) {
                        if (typeof options[key] != "object")
                            options[key] = {};
                        if (!(name in options[key]))
                            options[key][name] = options[name];
                    }
                });
            }
        }
        function init_cache(cache) {
            if (!cache)
                return;
            if (!("props" in cache)) {
                cache.props = new Map();
            }
            else if (!(cache.props instanceof Map)) {
                cache.props = map_from_object(cache.props);
            }
        }
        function cache_to_json(cache) {
            return {
                props: map_to_object(cache.props)
            };
        }
        async function minify(files, options) {
            options = defaults(options, {
                compress: {},
                ecma: undefined,
                enclose: false,
                ie8: false,
                keep_classnames: undefined,
                keep_fnames: false,
                mangle: {},
                module: false,
                nameCache: null,
                output: null,
                format: null,
                parse: {},
                rename: undefined,
                safari10: false,
                sourceMap: false,
                spidermonkey: false,
                timings: false,
                toplevel: false,
                warnings: false,
                wrap: false,
            }, true);
            var timings = options.timings && {
                start: Date.now()
            };
            if (options.keep_classnames === undefined) {
                options.keep_classnames = options.keep_fnames;
            }
            if (options.rename === undefined) {
                options.rename = options.compress && options.mangle;
            }
            if (options.output && options.format) {
                throw new Error("Please only specify either output or format option, preferrably format.");
            }
            options.format = options.format || options.output || {};
            set_shorthand("ecma", options, ["parse", "compress", "format"]);
            set_shorthand("ie8", options, ["compress", "mangle", "format"]);
            set_shorthand("keep_classnames", options, ["compress", "mangle"]);
            set_shorthand("keep_fnames", options, ["compress", "mangle"]);
            set_shorthand("module", options, ["parse", "compress", "mangle"]);
            set_shorthand("safari10", options, ["mangle", "format"]);
            set_shorthand("toplevel", options, ["compress", "mangle"]);
            set_shorthand("warnings", options, ["compress"]); // legacy
            var quoted_props;
            if (options.mangle) {
                options.mangle = defaults(options.mangle, {
                    cache: options.nameCache && (options.nameCache.vars || {}),
                    eval: false,
                    ie8: false,
                    keep_classnames: false,
                    keep_fnames: false,
                    module: false,
                    properties: false,
                    reserved: [],
                    safari10: false,
                    toplevel: false,
                }, true);
                if (options.mangle.properties) {
                    if (typeof options.mangle.properties != "object") {
                        options.mangle.properties = {};
                    }
                    if (options.mangle.properties.keep_quoted) {
                        quoted_props = options.mangle.properties.reserved;
                        if (!Array.isArray(quoted_props))
                            quoted_props = [];
                        options.mangle.properties.reserved = quoted_props;
                    }
                    if (options.nameCache && !("cache" in options.mangle.properties)) {
                        options.mangle.properties.cache = options.nameCache.props || {};
                    }
                }
                init_cache(options.mangle.cache);
                init_cache(options.mangle.properties.cache);
            }
            if (options.sourceMap) {
                options.sourceMap = defaults(options.sourceMap, {
                    asObject: false,
                    content: null,
                    filename: null,
                    includeSources: false,
                    root: null,
                    url: null,
                }, true);
            }
            if (timings)
                timings.parse = Date.now();
            var toplevel;
            if (files instanceof AST_Toplevel) {
                toplevel = files;
            }
            else {
                if (typeof files == "string" || (options.parse.spidermonkey && !Array.isArray(files))) {
                    files = [files];
                }
                options.parse = options.parse || {};
                options.parse.toplevel = null;
                if (options.parse.spidermonkey) {
                    options.parse.toplevel = AST_Node.from_mozilla_ast(Object.keys(files).reduce(function (toplevel, name) {
                        if (!toplevel)
                            return files[name];
                        toplevel.body = toplevel.body.concat(files[name].body);
                        return toplevel;
                    }, null));
                }
                else {
                    delete options.parse.spidermonkey;
                    for (var name in files)
                        if (HOP(files, name)) {
                            options.parse.filename = name;
                            options.parse.toplevel = parse(files[name], options.parse);
                            if (options.sourceMap && options.sourceMap.content == "inline") {
                                if (Object.keys(files).length > 1)
                                    throw new Error("inline source map only works with singular input");
                                options.sourceMap.content = read_source_map(files[name]);
                            }
                        }
                }
                toplevel = options.parse.toplevel;
            }
            if (quoted_props && options.mangle.properties.keep_quoted !== "strict") {
                reserve_quoted_keys(toplevel, quoted_props);
            }
            if (options.wrap) {
                toplevel = toplevel.wrap_commonjs(options.wrap);
            }
            if (options.enclose) {
                toplevel = toplevel.wrap_enclose(options.enclose);
            }
            if (timings)
                timings.rename = Date.now();
            if (timings)
                timings.compress = Date.now();
            if (options.compress) {
                toplevel = new Compressor(options.compress, {
                    mangle_options: options.mangle
                }).compress(toplevel);
            }
            if (timings)
                timings.scope = Date.now();
            if (options.mangle)
                toplevel.figure_out_scope(options.mangle);
            if (timings)
                timings.mangle = Date.now();
            if (options.mangle) {
                base54.reset();
                toplevel.compute_char_frequency(options.mangle);
                toplevel.mangle_names(options.mangle);
            }
            if (timings)
                timings.properties = Date.now();
            if (options.mangle && options.mangle.properties) {
                toplevel = mangle_properties(toplevel, options.mangle.properties);
            }
            if (timings)
                timings.format = Date.now();
            var result = {};
            if (options.format.ast) {
                result.ast = toplevel;
            }
            if (options.format.spidermonkey) {
                result.ast = toplevel.to_mozilla_ast();
            }
            if (!HOP(options.format, "code") || options.format.code) {
                if (options.sourceMap) {
                    options.format.source_map = await SourceMap({
                        file: options.sourceMap.filename,
                        orig: options.sourceMap.content,
                        root: options.sourceMap.root
                    });
                    if (options.sourceMap.includeSources) {
                        if (files instanceof AST_Toplevel) {
                            throw new Error("original source content unavailable");
                        }
                        else
                            for (var name in files)
                                if (HOP(files, name)) {
                                    options.format.source_map.get().setSourceContent(name, files[name]);
                                }
                    }
                }
                delete options.format.ast;
                delete options.format.code;
                delete options.format.spidermonkey;
                var stream = OutputStream(options.format);
                toplevel.print(stream);
                result.code = stream.get();
                if (options.sourceMap) {
                    if (options.sourceMap.asObject) {
                        result.map = options.format.source_map.get().toJSON();
                    }
                    else {
                        result.map = options.format.source_map.toString();
                    }
                    if (options.sourceMap.url == "inline") {
                        var sourceMap = typeof result.map === "object" ? JSON.stringify(result.map) : result.map;
                        result.code += "\n//# sourceMappingURL=data:application/json;charset=utf-8;base64," + to_base64(sourceMap);
                    }
                    else if (options.sourceMap.url) {
                        result.code += "\n//# sourceMappingURL=" + options.sourceMap.url;
                    }
                }
            }
            if (options.nameCache && options.mangle) {
                if (options.mangle.cache)
                    options.nameCache.vars = cache_to_json(options.mangle.cache);
                if (options.mangle.properties && options.mangle.properties.cache) {
                    options.nameCache.props = cache_to_json(options.mangle.properties.cache);
                }
            }
            if (options.format && options.format.source_map) {
                options.format.source_map.destroy();
            }
            if (timings) {
                timings.end = Date.now();
                result.timings = {
                    parse: 1e-3 * (timings.rename - timings.parse),
                    rename: 1e-3 * (timings.compress - timings.rename),
                    compress: 1e-3 * (timings.scope - timings.compress),
                    scope: 1e-3 * (timings.mangle - timings.scope),
                    mangle: 1e-3 * (timings.properties - timings.mangle),
                    properties: 1e-3 * (timings.format - timings.properties),
                    format: 1e-3 * (timings.end - timings.format),
                    total: 1e-3 * (timings.end - timings.start)
                };
            }
            return result;
        }
        async function run_cli({ program, packageJson, fs, path }) {
            const skip_keys = new Set(["cname", "parent_scope", "scope", "uses_eval", "uses_with"]);
            var files = {};
            var options = {
                compress: false,
                mangle: false
            };
            const default_options = await _default_options();
            program.version(packageJson.name + " " + packageJson.version);
            program.parseArgv = program.parse;
            program.parse = undefined;
            if (process.argv.includes("ast"))
                program.helpInformation = describe_ast;
            else if (process.argv.includes("options"))
                program.helpInformation = function () {
                    var text = [];
                    for (var option in default_options) {
                        text.push("--" + (option === "sourceMap" ? "source-map" : option) + " options:");
                        text.push(format_object(default_options[option]));
                        text.push("");
                    }
                    return text.join("\n");
                };
            program.option("-p, --parse <options>", "Specify parser options.", parse_js());
            program.option("-c, --compress [options]", "Enable compressor/specify compressor options.", parse_js());
            program.option("-m, --mangle [options]", "Mangle names/specify mangler options.", parse_js());
            program.option("--mangle-props [options]", "Mangle properties/specify mangler options.", parse_js());
            program.option("-f, --format [options]", "Format options.", parse_js());
            program.option("-b, --beautify [options]", "Alias for --format.", parse_js());
            program.option("-o, --output <file>", "Output file (default STDOUT).");
            program.option("--comments [filter]", "Preserve copyright comments in the output.");
            program.option("--config-file <file>", "Read minify() options from JSON file.");
            program.option("-d, --define <expr>[=value]", "Global definitions.", parse_js("define"));
            program.option("--ecma <version>", "Specify ECMAScript release: 5, 2015, 2016 or 2017...");
            program.option("-e, --enclose [arg[,...][:value[,...]]]", "Embed output in a big function with configurable arguments and values.");
            program.option("--ie8", "Support non-standard Internet Explorer 8.");
            program.option("--keep-classnames", "Do not mangle/drop class names.");
            program.option("--keep-fnames", "Do not mangle/drop function names. Useful for code relying on Function.prototype.name.");
            program.option("--module", "Input is an ES6 module");
            program.option("--name-cache <file>", "File to hold mangled name mappings.");
            program.option("--rename", "Force symbol expansion.");
            program.option("--no-rename", "Disable symbol expansion.");
            program.option("--safari10", "Support non-standard Safari 10.");
            program.option("--source-map [options]", "Enable source map/specify source map options.", parse_js());
            program.option("--timings", "Display operations run time on STDERR.");
            program.option("--toplevel", "Compress and/or mangle variables in toplevel scope.");
            program.option("--wrap <name>", "Embed everything as a function with “exports” corresponding to “name” globally.");
            program.arguments("[files...]").parseArgv(process.argv);
            if (program.configFile) {
                options = JSON.parse(read_file(program.configFile));
            }
            if (!program.output && program.sourceMap && program.sourceMap.url != "inline") {
                fatal("ERROR: cannot write source map to STDOUT");
            }
            [
                "compress",
                "enclose",
                "ie8",
                "mangle",
                "module",
                "safari10",
                "sourceMap",
                "toplevel",
                "wrap"
            ].forEach(function (name) {
                if (name in program) {
                    options[name] = program[name];
                }
            });
            if ("ecma" in program) {
                if (program.ecma != (program.ecma | 0))
                    fatal("ERROR: ecma must be an integer");
                const ecma = program.ecma | 0;
                if (ecma > 5 && ecma < 2015)
                    options.ecma = ecma + 2009;
                else
                    options.ecma = ecma;
            }
            if (program.format || program.beautify) {
                const chosenOption = program.format || program.beautify;
                options.format = typeof chosenOption === "object" ? chosenOption : {};
            }
            if (program.comments) {
                if (typeof options.format != "object")
                    options.format = {};
                options.format.comments = typeof program.comments == "string" ? (program.comments == "false" ? false : program.comments) : "some";
            }
            if (program.define) {
                if (typeof options.compress != "object")
                    options.compress = {};
                if (typeof options.compress.global_defs != "object")
                    options.compress.global_defs = {};
                for (var expr in program.define) {
                    options.compress.global_defs[expr] = program.define[expr];
                }
            }
            if (program.keepClassnames) {
                options.keep_classnames = true;
            }
            if (program.keepFnames) {
                options.keep_fnames = true;
            }
            if (program.mangleProps) {
                if (program.mangleProps.domprops) {
                    delete program.mangleProps.domprops;
                }
                else {
                    if (typeof program.mangleProps != "object")
                        program.mangleProps = {};
                    if (!Array.isArray(program.mangleProps.reserved))
                        program.mangleProps.reserved = [];
                }
                if (typeof options.mangle != "object")
                    options.mangle = {};
                options.mangle.properties = program.mangleProps;
            }
            if (program.nameCache) {
                options.nameCache = JSON.parse(read_file(program.nameCache, "{}"));
            }
            if (program.output == "ast") {
                options.format = {
                    ast: true,
                    code: false
                };
            }
            if (program.parse) {
                if (!program.parse.acorn && !program.parse.spidermonkey) {
                    options.parse = program.parse;
                }
                else if (program.sourceMap && program.sourceMap.content == "inline") {
                    fatal("ERROR: inline source map only works with built-in parser");
                }
            }
            if (~program.rawArgs.indexOf("--rename")) {
                options.rename = true;
            }
            else if (!program.rename) {
                options.rename = false;
            }
            let convert_path = name => name;
            if (typeof program.sourceMap == "object" && "base" in program.sourceMap) {
                convert_path = function () {
                    var base = program.sourceMap.base;
                    delete options.sourceMap.base;
                    return function (name) {
                        return path.relative(base, name);
                    };
                }();
            }
            let filesList;
            if (options.files && options.files.length) {
                filesList = options.files;
                delete options.files;
            }
            else if (program.args.length) {
                filesList = program.args;
            }
            if (filesList) {
                simple_glob(filesList).forEach(function (name) {
                    files[convert_path(name)] = read_file(name);
                });
            }
            else {
                await new Promise((resolve) => {
                    var chunks = [];
                    process.stdin.setEncoding("utf8");
                    process.stdin.on("data", function (chunk) {
                        chunks.push(chunk);
                    }).on("end", function () {
                        files = [chunks.join("")];
                        resolve();
                    });
                    process.stdin.resume();
                });
            }
            await run_cli();
            function convert_ast(fn) {
                return AST_Node.from_mozilla_ast(Object.keys(files).reduce(fn, null));
            }
            async function run_cli() {
                var content = program.sourceMap && program.sourceMap.content;
                if (content && content !== "inline") {
                    options.sourceMap.content = read_file(content, content);
                }
                if (program.timings)
                    options.timings = true;
                try {
                    if (program.parse) {
                        if (program.parse.acorn) {
                            files = convert_ast(function (toplevel, name) {
                                return require(46) /* acorn */.parse(files[name], {
                                    ecmaVersion: 2018,
                                    locations: true,
                                    program: toplevel,
                                    sourceFile: name,
                                    sourceType: options.module || program.parse.module ? "module" : "script"
                                });
                            });
                        }
                        else if (program.parse.spidermonkey) {
                            files = convert_ast(function (toplevel, name) {
                                var obj = JSON.parse(files[name]);
                                if (!toplevel)
                                    return obj;
                                toplevel.body = toplevel.body.concat(obj.body);
                                return toplevel;
                            });
                        }
                    }
                }
                catch (ex) {
                    fatal(ex);
                }
                let result;
                try {
                    result = await minify(files, options);
                }
                catch (ex) {
                    if (ex.name == "SyntaxError") {
                        print_error("Parse error at " + ex.filename + ":" + ex.line + "," + ex.col);
                        var col = ex.col;
                        var lines = files[ex.filename].split(/\r?\n/);
                        var line = lines[ex.line - 1];
                        if (!line && !col) {
                            line = lines[ex.line - 2];
                            col = line.length;
                        }
                        if (line) {
                            var limit = 70;
                            if (col > limit) {
                                line = line.slice(col - limit);
                                col = limit;
                            }
                            print_error(line.slice(0, 80));
                            print_error(line.slice(0, col).replace(/\S/g, " ") + "^");
                        }
                    }
                    if (ex.defs) {
                        print_error("Supported options:");
                        print_error(format_object(ex.defs));
                    }
                    fatal(ex);
                    return;
                }
                if (program.output == "ast") {
                    if (!options.compress && !options.mangle) {
                        result.ast.figure_out_scope({});
                    }
                    console.log(JSON.stringify(result.ast, function (key, value) {
                        if (value)
                            switch (key) {
                                case "thedef":
                                    return symdef(value);
                                case "enclosed":
                                    return value.length ? value.map(symdef) : undefined;
                                case "variables":
                                case "globals":
                                    return value.size ? collect_from_map(value, symdef) : undefined;
                            }
                        if (skip_keys.has(key))
                            return;
                        if (value instanceof AST_Token)
                            return;
                        if (value instanceof Map)
                            return;
                        if (value instanceof AST_Node) {
                            var result = {
                                _class: "AST_" + value.TYPE
                            };
                            if (value.block_scope) {
                                result.variables = value.block_scope.variables;
                                result.enclosed = value.block_scope.enclosed;
                            }
                            value.CTOR.PROPS.forEach(function (prop) {
                                result[prop] = value[prop];
                            });
                            return result;
                        }
                        return value;
                    }, 2));
                }
                else if (program.output == "spidermonkey") {
                    try {
                        const minified = await minify(result.code, {
                            compress: false,
                            mangle: false,
                            format: {
                                ast: true,
                                code: false
                            }
                        });
                        console.log(JSON.stringify(minified.ast.to_mozilla_ast(), null, 2));
                    }
                    catch (ex) {
                        fatal(ex);
                        return;
                    }
                }
                else if (program.output) {
                    fs.writeFileSync(program.output, result.code);
                    if (options.sourceMap && options.sourceMap.url !== "inline" && result.map) {
                        fs.writeFileSync(program.output + ".map", result.map);
                    }
                }
                else {
                    console.log(result.code);
                }
                if (program.nameCache) {
                    fs.writeFileSync(program.nameCache, JSON.stringify(options.nameCache));
                }
                if (result.timings)
                    for (var phase in result.timings) {
                        print_error("- " + phase + ": " + result.timings[phase].toFixed(3) + "s");
                    }
            }
            function fatal(message) {
                if (message instanceof Error)
                    message = message.stack.replace(/^\S*?Error:/, "ERROR:");
                print_error(message);
                process.exit(1);
            }
            // A file glob function that only supports "*" and "?" wildcards in the basename.
            // Example: "foo/bar/*baz??.*.js"
            // Argument `glob` may be a string or an array of strings.
            // Returns an array of strings. Garbage in, garbage out.
            function simple_glob(glob) {
                if (Array.isArray(glob)) {
                    return [].concat.apply([], glob.map(simple_glob));
                }
                if (glob && glob.match(/[*?]/)) {
                    var dir = path.dirname(glob);
                    try {
                        var entries = fs.readdirSync(dir);
                    }
                    catch (ex) { }
                    if (entries) {
                        var pattern = "^" + path.basename(glob)
                            .replace(/[.+^$[\]\\(){}]/g, "\\$&")
                            .replace(/\*/g, "[^/\\\\]*")
                            .replace(/\?/g, "[^/\\\\]") + "$";
                        var mod = process.platform === "win32" ? "i" : "";
                        var rx = new RegExp(pattern, mod);
                        var results = entries.filter(function (name) {
                            return rx.test(name);
                        }).map(function (name) {
                            return path.join(dir, name);
                        });
                        if (results.length)
                            return results;
                    }
                }
                return [glob];
            }
            function read_file(path, default_value) {
                try {
                    return fs.readFileSync(path, "utf8");
                }
                catch (ex) {
                    if ((ex.code == "ENOENT" || ex.code == "ENAMETOOLONG") && default_value != null)
                        return default_value;
                    fatal(ex);
                }
            }
            function parse_js(flag) {
                return function (value, options) {
                    options = options || {};
                    try {
                        walk(parse(value, { expression: true }), node => {
                            if (node instanceof AST_Assign) {
                                var name = node.left.print_to_string();
                                var value = node.right;
                                if (flag) {
                                    options[name] = value;
                                }
                                else if (value instanceof AST_Array) {
                                    options[name] = value.elements.map(to_string);
                                }
                                else if (value instanceof AST_RegExp) {
                                    value = value.value;
                                    options[name] = new RegExp(value.source, value.flags);
                                }
                                else {
                                    options[name] = to_string(value);
                                }
                                return true;
                            }
                            if (node instanceof AST_Symbol || node instanceof AST_PropAccess) {
                                var name = node.print_to_string();
                                options[name] = true;
                                return true;
                            }
                            if (!(node instanceof AST_Sequence))
                                throw node;
                            function to_string(value) {
                                return value instanceof AST_Constant ? value.getValue() : value.print_to_string({
                                    quote_keys: true
                                });
                            }
                        });
                    }
                    catch (ex) {
                        if (flag) {
                            fatal("Error parsing arguments for '" + flag + "': " + value);
                        }
                        else {
                            options[value] = null;
                        }
                    }
                    return options;
                };
            }
            function symdef(def) {
                var ret = (1e6 + def.id) + " " + def.name;
                if (def.mangled_name)
                    ret += " " + def.mangled_name;
                return ret;
            }
            function collect_from_map(map, callback) {
                var result = [];
                map.forEach(function (def) {
                    result.push(callback(def));
                });
                return result;
            }
            function format_object(obj) {
                var lines = [];
                var padding = "";
                Object.keys(obj).map(function (name) {
                    if (padding.length < name.length)
                        padding = Array(name.length + 1).join(" ");
                    return [name, JSON.stringify(obj[name])];
                }).forEach(function (tokens) {
                    lines.push("  " + tokens[0] + padding.slice(tokens[0].length - 2) + tokens[1]);
                });
                return lines.join("\n");
            }
            function print_error(msg) {
                process.stderr.write(msg);
                process.stderr.write("\n");
            }
            function describe_ast() {
                var out = OutputStream({ beautify: true });
                function doitem(ctor) {
                    out.print("AST_" + ctor.TYPE);
                    const props = ctor.SELF_PROPS.filter(prop => !/^\$/.test(prop));
                    if (props.length > 0) {
                        out.space();
                        out.with_parens(function () {
                            props.forEach(function (prop, i) {
                                if (i)
                                    out.space();
                                out.print(prop);
                            });
                        });
                    }
                    if (ctor.documentation) {
                        out.space();
                        out.print_string(ctor.documentation);
                    }
                    if (ctor.SUBCLASSES.length > 0) {
                        out.space();
                        out.with_block(function () {
                            ctor.SUBCLASSES.forEach(function (ctor) {
                                out.indent();
                                doitem(ctor);
                                out.newline();
                            });
                        });
                    }
                }
                doitem(AST_Node);
                return out + "\n";
            }
        }
        async function _default_options() {
            const defs = {};
            Object.keys(infer_options({ 0: 0 })).forEach((component) => {
                const options = infer_options({
                    [component]: { 0: 0 }
                });
                if (options)
                    defs[component] = options;
            });
            return defs;
        }
        async function infer_options(options) {
            try {
                await minify("", options);
            }
            catch (error) {
                return error.defs;
            }
        }
        exports._default_options = _default_options;
        exports._run_cli = run_cli;
        exports.minify = minify;
    })));
},
/* acorn/dist/acorn.js */ function _(require, module, exports, __esModule, __esExport) {
    (function (global, factory) {
        typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
            typeof define === 'function' && define.amd ? define(['exports'], factory) :
                (global = global || self, factory(global.acorn = {}));
    }(this, (function (exports) {
        'use strict';
        // Reserved word lists for various dialects of the language
        var reservedWords = {
            3: "abstract boolean byte char class double enum export extends final float goto implements import int interface long native package private protected public short static super synchronized throws transient volatile",
            5: "class enum extends super const export import",
            6: "enum",
            strict: "implements interface let package private protected public static yield",
            strictBind: "eval arguments"
        };
        // And the keywords
        var ecma5AndLessKeywords = "break case catch continue debugger default do else finally for function if return switch throw try var while with null true false instanceof typeof void delete new in this";
        var keywords = {
            5: ecma5AndLessKeywords,
            "5module": ecma5AndLessKeywords + " export import",
            6: ecma5AndLessKeywords + " const class extends export import super"
        };
        var keywordRelationalOperator = /^in(stanceof)?$/;
        // ## Character categories
        // Big ugly regular expressions that match characters in the
        // whitespace, identifier, and identifier-start categories. These
        // are only applied when a character is found to actually have a
        // code point above 128.
        // Generated by `bin/generate-identifier-regex.js`.
        var nonASCIIidentifierStartChars = "\xaa\xb5\xba\xc0-\xd6\xd8-\xf6\xf8-\u02c1\u02c6-\u02d1\u02e0-\u02e4\u02ec\u02ee\u0370-\u0374\u0376\u0377\u037a-\u037d\u037f\u0386\u0388-\u038a\u038c\u038e-\u03a1\u03a3-\u03f5\u03f7-\u0481\u048a-\u052f\u0531-\u0556\u0559\u0560-\u0588\u05d0-\u05ea\u05ef-\u05f2\u0620-\u064a\u066e\u066f\u0671-\u06d3\u06d5\u06e5\u06e6\u06ee\u06ef\u06fa-\u06fc\u06ff\u0710\u0712-\u072f\u074d-\u07a5\u07b1\u07ca-\u07ea\u07f4\u07f5\u07fa\u0800-\u0815\u081a\u0824\u0828\u0840-\u0858\u0860-\u086a\u08a0-\u08b4\u08b6-\u08c7\u0904-\u0939\u093d\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098c\u098f\u0990\u0993-\u09a8\u09aa-\u09b0\u09b2\u09b6-\u09b9\u09bd\u09ce\u09dc\u09dd\u09df-\u09e1\u09f0\u09f1\u09fc\u0a05-\u0a0a\u0a0f\u0a10\u0a13-\u0a28\u0a2a-\u0a30\u0a32\u0a33\u0a35\u0a36\u0a38\u0a39\u0a59-\u0a5c\u0a5e\u0a72-\u0a74\u0a85-\u0a8d\u0a8f-\u0a91\u0a93-\u0aa8\u0aaa-\u0ab0\u0ab2\u0ab3\u0ab5-\u0ab9\u0abd\u0ad0\u0ae0\u0ae1\u0af9\u0b05-\u0b0c\u0b0f\u0b10\u0b13-\u0b28\u0b2a-\u0b30\u0b32\u0b33\u0b35-\u0b39\u0b3d\u0b5c\u0b5d\u0b5f-\u0b61\u0b71\u0b83\u0b85-\u0b8a\u0b8e-\u0b90\u0b92-\u0b95\u0b99\u0b9a\u0b9c\u0b9e\u0b9f\u0ba3\u0ba4\u0ba8-\u0baa\u0bae-\u0bb9\u0bd0\u0c05-\u0c0c\u0c0e-\u0c10\u0c12-\u0c28\u0c2a-\u0c39\u0c3d\u0c58-\u0c5a\u0c60\u0c61\u0c80\u0c85-\u0c8c\u0c8e-\u0c90\u0c92-\u0ca8\u0caa-\u0cb3\u0cb5-\u0cb9\u0cbd\u0cde\u0ce0\u0ce1\u0cf1\u0cf2\u0d04-\u0d0c\u0d0e-\u0d10\u0d12-\u0d3a\u0d3d\u0d4e\u0d54-\u0d56\u0d5f-\u0d61\u0d7a-\u0d7f\u0d85-\u0d96\u0d9a-\u0db1\u0db3-\u0dbb\u0dbd\u0dc0-\u0dc6\u0e01-\u0e30\u0e32\u0e33\u0e40-\u0e46\u0e81\u0e82\u0e84\u0e86-\u0e8a\u0e8c-\u0ea3\u0ea5\u0ea7-\u0eb0\u0eb2\u0eb3\u0ebd\u0ec0-\u0ec4\u0ec6\u0edc-\u0edf\u0f00\u0f40-\u0f47\u0f49-\u0f6c\u0f88-\u0f8c\u1000-\u102a\u103f\u1050-\u1055\u105a-\u105d\u1061\u1065\u1066\u106e-\u1070\u1075-\u1081\u108e\u10a0-\u10c5\u10c7\u10cd\u10d0-\u10fa\u10fc-\u1248\u124a-\u124d\u1250-\u1256\u1258\u125a-\u125d\u1260-\u1288\u128a-\u128d\u1290-\u12b0\u12b2-\u12b5\u12b8-\u12be\u12c0\u12c2-\u12c5\u12c8-\u12d6\u12d8-\u1310\u1312-\u1315\u1318-\u135a\u1380-\u138f\u13a0-\u13f5\u13f8-\u13fd\u1401-\u166c\u166f-\u167f\u1681-\u169a\u16a0-\u16ea\u16ee-\u16f8\u1700-\u170c\u170e-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176c\u176e-\u1770\u1780-\u17b3\u17d7\u17dc\u1820-\u1878\u1880-\u18a8\u18aa\u18b0-\u18f5\u1900-\u191e\u1950-\u196d\u1970-\u1974\u1980-\u19ab\u19b0-\u19c9\u1a00-\u1a16\u1a20-\u1a54\u1aa7\u1b05-\u1b33\u1b45-\u1b4b\u1b83-\u1ba0\u1bae\u1baf\u1bba-\u1be5\u1c00-\u1c23\u1c4d-\u1c4f\u1c5a-\u1c7d\u1c80-\u1c88\u1c90-\u1cba\u1cbd-\u1cbf\u1ce9-\u1cec\u1cee-\u1cf3\u1cf5\u1cf6\u1cfa\u1d00-\u1dbf\u1e00-\u1f15\u1f18-\u1f1d\u1f20-\u1f45\u1f48-\u1f4d\u1f50-\u1f57\u1f59\u1f5b\u1f5d\u1f5f-\u1f7d\u1f80-\u1fb4\u1fb6-\u1fbc\u1fbe\u1fc2-\u1fc4\u1fc6-\u1fcc\u1fd0-\u1fd3\u1fd6-\u1fdb\u1fe0-\u1fec\u1ff2-\u1ff4\u1ff6-\u1ffc\u2071\u207f\u2090-\u209c\u2102\u2107\u210a-\u2113\u2115\u2118-\u211d\u2124\u2126\u2128\u212a-\u2139\u213c-\u213f\u2145-\u2149\u214e\u2160-\u2188\u2c00-\u2c2e\u2c30-\u2c5e\u2c60-\u2ce4\u2ceb-\u2cee\u2cf2\u2cf3\u2d00-\u2d25\u2d27\u2d2d\u2d30-\u2d67\u2d6f\u2d80-\u2d96\u2da0-\u2da6\u2da8-\u2dae\u2db0-\u2db6\u2db8-\u2dbe\u2dc0-\u2dc6\u2dc8-\u2dce\u2dd0-\u2dd6\u2dd8-\u2dde\u3005-\u3007\u3021-\u3029\u3031-\u3035\u3038-\u303c\u3041-\u3096\u309b-\u309f\u30a1-\u30fa\u30fc-\u30ff\u3105-\u312f\u3131-\u318e\u31a0-\u31bf\u31f0-\u31ff\u3400-\u4dbf\u4e00-\u9ffc\ua000-\ua48c\ua4d0-\ua4fd\ua500-\ua60c\ua610-\ua61f\ua62a\ua62b\ua640-\ua66e\ua67f-\ua69d\ua6a0-\ua6ef\ua717-\ua71f\ua722-\ua788\ua78b-\ua7bf\ua7c2-\ua7ca\ua7f5-\ua801\ua803-\ua805\ua807-\ua80a\ua80c-\ua822\ua840-\ua873\ua882-\ua8b3\ua8f2-\ua8f7\ua8fb\ua8fd\ua8fe\ua90a-\ua925\ua930-\ua946\ua960-\ua97c\ua984-\ua9b2\ua9cf\ua9e0-\ua9e4\ua9e6-\ua9ef\ua9fa-\ua9fe\uaa00-\uaa28\uaa40-\uaa42\uaa44-\uaa4b\uaa60-\uaa76\uaa7a\uaa7e-\uaaaf\uaab1\uaab5\uaab6\uaab9-\uaabd\uaac0\uaac2\uaadb-\uaadd\uaae0-\uaaea\uaaf2-\uaaf4\uab01-\uab06\uab09-\uab0e\uab11-\uab16\uab20-\uab26\uab28-\uab2e\uab30-\uab5a\uab5c-\uab69\uab70-\uabe2\uac00-\ud7a3\ud7b0-\ud7c6\ud7cb-\ud7fb\uf900-\ufa6d\ufa70-\ufad9\ufb00-\ufb06\ufb13-\ufb17\ufb1d\ufb1f-\ufb28\ufb2a-\ufb36\ufb38-\ufb3c\ufb3e\ufb40\ufb41\ufb43\ufb44\ufb46-\ufbb1\ufbd3-\ufd3d\ufd50-\ufd8f\ufd92-\ufdc7\ufdf0-\ufdfb\ufe70-\ufe74\ufe76-\ufefc\uff21-\uff3a\uff41-\uff5a\uff66-\uffbe\uffc2-\uffc7\uffca-\uffcf\uffd2-\uffd7\uffda-\uffdc";
        var nonASCIIidentifierChars = "\u200c\u200d\xb7\u0300-\u036f\u0387\u0483-\u0487\u0591-\u05bd\u05bf\u05c1\u05c2\u05c4\u05c5\u05c7\u0610-\u061a\u064b-\u0669\u0670\u06d6-\u06dc\u06df-\u06e4\u06e7\u06e8\u06ea-\u06ed\u06f0-\u06f9\u0711\u0730-\u074a\u07a6-\u07b0\u07c0-\u07c9\u07eb-\u07f3\u07fd\u0816-\u0819\u081b-\u0823\u0825-\u0827\u0829-\u082d\u0859-\u085b\u08d3-\u08e1\u08e3-\u0903\u093a-\u093c\u093e-\u094f\u0951-\u0957\u0962\u0963\u0966-\u096f\u0981-\u0983\u09bc\u09be-\u09c4\u09c7\u09c8\u09cb-\u09cd\u09d7\u09e2\u09e3\u09e6-\u09ef\u09fe\u0a01-\u0a03\u0a3c\u0a3e-\u0a42\u0a47\u0a48\u0a4b-\u0a4d\u0a51\u0a66-\u0a71\u0a75\u0a81-\u0a83\u0abc\u0abe-\u0ac5\u0ac7-\u0ac9\u0acb-\u0acd\u0ae2\u0ae3\u0ae6-\u0aef\u0afa-\u0aff\u0b01-\u0b03\u0b3c\u0b3e-\u0b44\u0b47\u0b48\u0b4b-\u0b4d\u0b55-\u0b57\u0b62\u0b63\u0b66-\u0b6f\u0b82\u0bbe-\u0bc2\u0bc6-\u0bc8\u0bca-\u0bcd\u0bd7\u0be6-\u0bef\u0c00-\u0c04\u0c3e-\u0c44\u0c46-\u0c48\u0c4a-\u0c4d\u0c55\u0c56\u0c62\u0c63\u0c66-\u0c6f\u0c81-\u0c83\u0cbc\u0cbe-\u0cc4\u0cc6-\u0cc8\u0cca-\u0ccd\u0cd5\u0cd6\u0ce2\u0ce3\u0ce6-\u0cef\u0d00-\u0d03\u0d3b\u0d3c\u0d3e-\u0d44\u0d46-\u0d48\u0d4a-\u0d4d\u0d57\u0d62\u0d63\u0d66-\u0d6f\u0d81-\u0d83\u0dca\u0dcf-\u0dd4\u0dd6\u0dd8-\u0ddf\u0de6-\u0def\u0df2\u0df3\u0e31\u0e34-\u0e3a\u0e47-\u0e4e\u0e50-\u0e59\u0eb1\u0eb4-\u0ebc\u0ec8-\u0ecd\u0ed0-\u0ed9\u0f18\u0f19\u0f20-\u0f29\u0f35\u0f37\u0f39\u0f3e\u0f3f\u0f71-\u0f84\u0f86\u0f87\u0f8d-\u0f97\u0f99-\u0fbc\u0fc6\u102b-\u103e\u1040-\u1049\u1056-\u1059\u105e-\u1060\u1062-\u1064\u1067-\u106d\u1071-\u1074\u1082-\u108d\u108f-\u109d\u135d-\u135f\u1369-\u1371\u1712-\u1714\u1732-\u1734\u1752\u1753\u1772\u1773\u17b4-\u17d3\u17dd\u17e0-\u17e9\u180b-\u180d\u1810-\u1819\u18a9\u1920-\u192b\u1930-\u193b\u1946-\u194f\u19d0-\u19da\u1a17-\u1a1b\u1a55-\u1a5e\u1a60-\u1a7c\u1a7f-\u1a89\u1a90-\u1a99\u1ab0-\u1abd\u1abf\u1ac0\u1b00-\u1b04\u1b34-\u1b44\u1b50-\u1b59\u1b6b-\u1b73\u1b80-\u1b82\u1ba1-\u1bad\u1bb0-\u1bb9\u1be6-\u1bf3\u1c24-\u1c37\u1c40-\u1c49\u1c50-\u1c59\u1cd0-\u1cd2\u1cd4-\u1ce8\u1ced\u1cf4\u1cf7-\u1cf9\u1dc0-\u1df9\u1dfb-\u1dff\u203f\u2040\u2054\u20d0-\u20dc\u20e1\u20e5-\u20f0\u2cef-\u2cf1\u2d7f\u2de0-\u2dff\u302a-\u302f\u3099\u309a\ua620-\ua629\ua66f\ua674-\ua67d\ua69e\ua69f\ua6f0\ua6f1\ua802\ua806\ua80b\ua823-\ua827\ua82c\ua880\ua881\ua8b4-\ua8c5\ua8d0-\ua8d9\ua8e0-\ua8f1\ua8ff-\ua909\ua926-\ua92d\ua947-\ua953\ua980-\ua983\ua9b3-\ua9c0\ua9d0-\ua9d9\ua9e5\ua9f0-\ua9f9\uaa29-\uaa36\uaa43\uaa4c\uaa4d\uaa50-\uaa59\uaa7b-\uaa7d\uaab0\uaab2-\uaab4\uaab7\uaab8\uaabe\uaabf\uaac1\uaaeb-\uaaef\uaaf5\uaaf6\uabe3-\uabea\uabec\uabed\uabf0-\uabf9\ufb1e\ufe00-\ufe0f\ufe20-\ufe2f\ufe33\ufe34\ufe4d-\ufe4f\uff10-\uff19\uff3f";
        var nonASCIIidentifierStart = new RegExp("[" + nonASCIIidentifierStartChars + "]");
        var nonASCIIidentifier = new RegExp("[" + nonASCIIidentifierStartChars + nonASCIIidentifierChars + "]");
        nonASCIIidentifierStartChars = nonASCIIidentifierChars = null;
        // These are a run-length and offset encoded representation of the
        // >0xffff code points that are a valid part of identifiers. The
        // offset starts at 0x10000, and each pair of numbers represents an
        // offset to the next range, and then a size of the range. They were
        // generated by bin/generate-identifier-regex.js
        // eslint-disable-next-line comma-spacing
        var astralIdentifierStartCodes = [0, 11, 2, 25, 2, 18, 2, 1, 2, 14, 3, 13, 35, 122, 70, 52, 268, 28, 4, 48, 48, 31, 14, 29, 6, 37, 11, 29, 3, 35, 5, 7, 2, 4, 43, 157, 19, 35, 5, 35, 5, 39, 9, 51, 157, 310, 10, 21, 11, 7, 153, 5, 3, 0, 2, 43, 2, 1, 4, 0, 3, 22, 11, 22, 10, 30, 66, 18, 2, 1, 11, 21, 11, 25, 71, 55, 7, 1, 65, 0, 16, 3, 2, 2, 2, 28, 43, 28, 4, 28, 36, 7, 2, 27, 28, 53, 11, 21, 11, 18, 14, 17, 111, 72, 56, 50, 14, 50, 14, 35, 349, 41, 7, 1, 79, 28, 11, 0, 9, 21, 107, 20, 28, 22, 13, 52, 76, 44, 33, 24, 27, 35, 30, 0, 3, 0, 9, 34, 4, 0, 13, 47, 15, 3, 22, 0, 2, 0, 36, 17, 2, 24, 85, 6, 2, 0, 2, 3, 2, 14, 2, 9, 8, 46, 39, 7, 3, 1, 3, 21, 2, 6, 2, 1, 2, 4, 4, 0, 19, 0, 13, 4, 159, 52, 19, 3, 21, 2, 31, 47, 21, 1, 2, 0, 185, 46, 42, 3, 37, 47, 21, 0, 60, 42, 14, 0, 72, 26, 230, 43, 117, 63, 32, 7, 3, 0, 3, 7, 2, 1, 2, 23, 16, 0, 2, 0, 95, 7, 3, 38, 17, 0, 2, 0, 29, 0, 11, 39, 8, 0, 22, 0, 12, 45, 20, 0, 35, 56, 264, 8, 2, 36, 18, 0, 50, 29, 113, 6, 2, 1, 2, 37, 22, 0, 26, 5, 2, 1, 2, 31, 15, 0, 328, 18, 190, 0, 80, 921, 103, 110, 18, 195, 2749, 1070, 4050, 582, 8634, 568, 8, 30, 114, 29, 19, 47, 17, 3, 32, 20, 6, 18, 689, 63, 129, 74, 6, 0, 67, 12, 65, 1, 2, 0, 29, 6135, 9, 1237, 43, 8, 8952, 286, 50, 2, 18, 3, 9, 395, 2309, 106, 6, 12, 4, 8, 8, 9, 5991, 84, 2, 70, 2, 1, 3, 0, 3, 1, 3, 3, 2, 11, 2, 0, 2, 6, 2, 64, 2, 3, 3, 7, 2, 6, 2, 27, 2, 3, 2, 4, 2, 0, 4, 6, 2, 339, 3, 24, 2, 24, 2, 30, 2, 24, 2, 30, 2, 24, 2, 30, 2, 24, 2, 30, 2, 24, 2, 7, 2357, 44, 11, 6, 17, 0, 370, 43, 1301, 196, 60, 67, 8, 0, 1205, 3, 2, 26, 2, 1, 2, 0, 3, 0, 2, 9, 2, 3, 2, 0, 2, 0, 7, 0, 5, 0, 2, 0, 2, 0, 2, 2, 2, 1, 2, 0, 3, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 1, 2, 0, 3, 3, 2, 6, 2, 3, 2, 3, 2, 0, 2, 9, 2, 16, 6, 2, 2, 4, 2, 16, 4421, 42717, 35, 4148, 12, 221, 3, 5761, 15, 7472, 3104, 541, 1507, 4938];
        // eslint-disable-next-line comma-spacing
        var astralIdentifierCodes = [509, 0, 227, 0, 150, 4, 294, 9, 1368, 2, 2, 1, 6, 3, 41, 2, 5, 0, 166, 1, 574, 3, 9, 9, 370, 1, 154, 10, 176, 2, 54, 14, 32, 9, 16, 3, 46, 10, 54, 9, 7, 2, 37, 13, 2, 9, 6, 1, 45, 0, 13, 2, 49, 13, 9, 3, 2, 11, 83, 11, 7, 0, 161, 11, 6, 9, 7, 3, 56, 1, 2, 6, 3, 1, 3, 2, 10, 0, 11, 1, 3, 6, 4, 4, 193, 17, 10, 9, 5, 0, 82, 19, 13, 9, 214, 6, 3, 8, 28, 1, 83, 16, 16, 9, 82, 12, 9, 9, 84, 14, 5, 9, 243, 14, 166, 9, 71, 5, 2, 1, 3, 3, 2, 0, 2, 1, 13, 9, 120, 6, 3, 6, 4, 0, 29, 9, 41, 6, 2, 3, 9, 0, 10, 10, 47, 15, 406, 7, 2, 7, 17, 9, 57, 21, 2, 13, 123, 5, 4, 0, 2, 1, 2, 6, 2, 0, 9, 9, 49, 4, 2, 1, 2, 4, 9, 9, 330, 3, 19306, 9, 135, 4, 60, 6, 26, 9, 1014, 0, 2, 54, 8, 3, 82, 0, 12, 1, 19628, 1, 5319, 4, 4, 5, 9, 7, 3, 6, 31, 3, 149, 2, 1418, 49, 513, 54, 5, 49, 9, 0, 15, 0, 23, 4, 2, 14, 1361, 6, 2, 16, 3, 6, 2, 1, 2, 4, 262, 6, 10, 9, 419, 13, 1495, 6, 110, 6, 6, 9, 4759, 9, 787719, 239];
        // This has a complexity linear to the value of the code. The
        // assumption is that looking up astral identifier characters is
        // rare.
        function isInAstralSet(code, set) {
            var pos = 0x10000;
            for (var i = 0; i < set.length; i += 2) {
                pos += set[i];
                if (pos > code) {
                    return false;
                }
                pos += set[i + 1];
                if (pos >= code) {
                    return true;
                }
            }
        }
        // Test whether a given character code starts an identifier.
        function isIdentifierStart(code, astral) {
            if (code < 65) {
                return code === 36;
            }
            if (code < 91) {
                return true;
            }
            if (code < 97) {
                return code === 95;
            }
            if (code < 123) {
                return true;
            }
            if (code <= 0xffff) {
                return code >= 0xaa && nonASCIIidentifierStart.test(String.fromCharCode(code));
            }
            if (astral === false) {
                return false;
            }
            return isInAstralSet(code, astralIdentifierStartCodes);
        }
        // Test whether a given character is part of an identifier.
        function isIdentifierChar(code, astral) {
            if (code < 48) {
                return code === 36;
            }
            if (code < 58) {
                return true;
            }
            if (code < 65) {
                return false;
            }
            if (code < 91) {
                return true;
            }
            if (code < 97) {
                return code === 95;
            }
            if (code < 123) {
                return true;
            }
            if (code <= 0xffff) {
                return code >= 0xaa && nonASCIIidentifier.test(String.fromCharCode(code));
            }
            if (astral === false) {
                return false;
            }
            return isInAstralSet(code, astralIdentifierStartCodes) || isInAstralSet(code, astralIdentifierCodes);
        }
        // ## Token types
        // The assignment of fine-grained, information-carrying type objects
        // allows the tokenizer to store the information it has about a
        // token in a way that is very cheap for the parser to look up.
        // All token type variables start with an underscore, to make them
        // easy to recognize.
        // The `beforeExpr` property is used to disambiguate between regular
        // expressions and divisions. It is set on all token types that can
        // be followed by an expression (thus, a slash after them would be a
        // regular expression).
        //
        // The `startsExpr` property is used to check if the token ends a
        // `yield` expression. It is set on all token types that either can
        // directly start an expression (like a quotation mark) or can
        // continue an expression (like the body of a string).
        //
        // `isLoop` marks a keyword as starting a loop, which is important
        // to know when parsing a label, in order to allow or disallow
        // continue jumps to that label.
        var TokenType = function TokenType(label, conf) {
            if (conf === void 0)
                conf = {};
            this.label = label;
            this.keyword = conf.keyword;
            this.beforeExpr = !!conf.beforeExpr;
            this.startsExpr = !!conf.startsExpr;
            this.isLoop = !!conf.isLoop;
            this.isAssign = !!conf.isAssign;
            this.prefix = !!conf.prefix;
            this.postfix = !!conf.postfix;
            this.binop = conf.binop || null;
            this.updateContext = null;
        };
        function binop(name, prec) {
            return new TokenType(name, { beforeExpr: true, binop: prec });
        }
        var beforeExpr = { beforeExpr: true }, startsExpr = { startsExpr: true };
        // Map keyword names to token types.
        var keywords$1 = {};
        // Succinct definitions of keyword token types
        function kw(name, options) {
            if (options === void 0)
                options = {};
            options.keyword = name;
            return keywords$1[name] = new TokenType(name, options);
        }
        var types = {
            num: new TokenType("num", startsExpr),
            regexp: new TokenType("regexp", startsExpr),
            string: new TokenType("string", startsExpr),
            name: new TokenType("name", startsExpr),
            privateId: new TokenType("privateId", startsExpr),
            eof: new TokenType("eof"),
            // Punctuation token types.
            bracketL: new TokenType("[", { beforeExpr: true, startsExpr: true }),
            bracketR: new TokenType("]"),
            braceL: new TokenType("{", { beforeExpr: true, startsExpr: true }),
            braceR: new TokenType("}"),
            parenL: new TokenType("(", { beforeExpr: true, startsExpr: true }),
            parenR: new TokenType(")"),
            comma: new TokenType(",", beforeExpr),
            semi: new TokenType(";", beforeExpr),
            colon: new TokenType(":", beforeExpr),
            dot: new TokenType("."),
            question: new TokenType("?", beforeExpr),
            questionDot: new TokenType("?."),
            arrow: new TokenType("=>", beforeExpr),
            template: new TokenType("template"),
            invalidTemplate: new TokenType("invalidTemplate"),
            ellipsis: new TokenType("...", beforeExpr),
            backQuote: new TokenType("`", startsExpr),
            dollarBraceL: new TokenType("${", { beforeExpr: true, startsExpr: true }),
            // Operators. These carry several kinds of properties to help the
            // parser use them properly (the presence of these properties is
            // what categorizes them as operators).
            //
            // `binop`, when present, specifies that this operator is a binary
            // operator, and will refer to its precedence.
            //
            // `prefix` and `postfix` mark the operator as a prefix or postfix
            // unary operator.
            //
            // `isAssign` marks all of `=`, `+=`, `-=` etcetera, which act as
            // binary operators with a very low precedence, that should result
            // in AssignmentExpression nodes.
            eq: new TokenType("=", { beforeExpr: true, isAssign: true }),
            assign: new TokenType("_=", { beforeExpr: true, isAssign: true }),
            incDec: new TokenType("++/--", { prefix: true, postfix: true, startsExpr: true }),
            prefix: new TokenType("!/~", { beforeExpr: true, prefix: true, startsExpr: true }),
            logicalOR: binop("||", 1),
            logicalAND: binop("&&", 2),
            bitwiseOR: binop("|", 3),
            bitwiseXOR: binop("^", 4),
            bitwiseAND: binop("&", 5),
            equality: binop("==/!=/===/!==", 6),
            relational: binop("</>/<=/>=", 7),
            bitShift: binop("<</>>/>>>", 8),
            plusMin: new TokenType("+/-", { beforeExpr: true, binop: 9, prefix: true, startsExpr: true }),
            modulo: binop("%", 10),
            star: binop("*", 10),
            slash: binop("/", 10),
            starstar: new TokenType("**", { beforeExpr: true }),
            coalesce: binop("??", 1),
            // Keyword token types.
            _break: kw("break"),
            _case: kw("case", beforeExpr),
            _catch: kw("catch"),
            _continue: kw("continue"),
            _debugger: kw("debugger"),
            _default: kw("default", beforeExpr),
            _do: kw("do", { isLoop: true, beforeExpr: true }),
            _else: kw("else", beforeExpr),
            _finally: kw("finally"),
            _for: kw("for", { isLoop: true }),
            _function: kw("function", startsExpr),
            _if: kw("if"),
            _return: kw("return", beforeExpr),
            _switch: kw("switch"),
            _throw: kw("throw", beforeExpr),
            _try: kw("try"),
            _var: kw("var"),
            _const: kw("const"),
            _while: kw("while", { isLoop: true }),
            _with: kw("with"),
            _new: kw("new", { beforeExpr: true, startsExpr: true }),
            _this: kw("this", startsExpr),
            _super: kw("super", startsExpr),
            _class: kw("class", startsExpr),
            _extends: kw("extends", beforeExpr),
            _export: kw("export"),
            _import: kw("import", startsExpr),
            _null: kw("null", startsExpr),
            _true: kw("true", startsExpr),
            _false: kw("false", startsExpr),
            _in: kw("in", { beforeExpr: true, binop: 7 }),
            _instanceof: kw("instanceof", { beforeExpr: true, binop: 7 }),
            _typeof: kw("typeof", { beforeExpr: true, prefix: true, startsExpr: true }),
            _void: kw("void", { beforeExpr: true, prefix: true, startsExpr: true }),
            _delete: kw("delete", { beforeExpr: true, prefix: true, startsExpr: true })
        };
        // Matches a whole line break (where CRLF is considered a single
        // line break). Used to count lines.
        var lineBreak = /\r\n?|\n|\u2028|\u2029/;
        var lineBreakG = new RegExp(lineBreak.source, "g");
        function isNewLine(code, ecma2019String) {
            return code === 10 || code === 13 || (!ecma2019String && (code === 0x2028 || code === 0x2029));
        }
        var nonASCIIwhitespace = /[\u1680\u2000-\u200a\u202f\u205f\u3000\ufeff]/;
        var skipWhiteSpace = /(?:\s|\/\/.*|\/\*[^]*?\*\/)*/g;
        var ref = Object.prototype;
        var hasOwnProperty = ref.hasOwnProperty;
        var toString = ref.toString;
        // Checks if an object has a property.
        function has(obj, propName) {
            return hasOwnProperty.call(obj, propName);
        }
        var isArray = Array.isArray || (function (obj) {
            return (toString.call(obj) === "[object Array]");
        });
        function wordsRegexp(words) {
            return new RegExp("^(?:" + words.replace(/ /g, "|") + ")$");
        }
        // These are used when `options.locations` is on, for the
        // `startLoc` and `endLoc` properties.
        var Position = function Position(line, col) {
            this.line = line;
            this.column = col;
        };
        Position.prototype.offset = function offset(n) {
            return new Position(this.line, this.column + n);
        };
        var SourceLocation = function SourceLocation(p, start, end) {
            this.start = start;
            this.end = end;
            if (p.sourceFile !== null) {
                this.source = p.sourceFile;
            }
        };
        // The `getLineInfo` function is mostly useful when the
        // `locations` option is off (for performance reasons) and you
        // want to find the line/column position for a given character
        // offset. `input` should be the code string that the offset refers
        // into.
        function getLineInfo(input, offset) {
            for (var line = 1, cur = 0;;) {
                lineBreakG.lastIndex = cur;
                var match = lineBreakG.exec(input);
                if (match && match.index < offset) {
                    ++line;
                    cur = match.index + match[0].length;
                }
                else {
                    return new Position(line, offset - cur);
                }
            }
        }
        // A second argument must be given to configure the parser process.
        // These options are recognized (only `ecmaVersion` is required):
        var defaultOptions = {
            // `ecmaVersion` indicates the ECMAScript version to parse. Must be
            // either 3, 5, 6 (or 2015), 7 (2016), 8 (2017), 9 (2018), 10
            // (2019), 11 (2020), 12 (2021), 13 (2022), or `"latest"` (the
            // latest version the library supports). This influences support
            // for strict mode, the set of reserved words, and support for
            // new syntax features.
            ecmaVersion: null,
            // `sourceType` indicates the mode the code should be parsed in.
            // Can be either `"script"` or `"module"`. This influences global
            // strict mode and parsing of `import` and `export` declarations.
            sourceType: "script",
            // `onInsertedSemicolon` can be a callback that will be called
            // when a semicolon is automatically inserted. It will be passed
            // the position of the comma as an offset, and if `locations` is
            // enabled, it is given the location as a `{line, column}` object
            // as second argument.
            onInsertedSemicolon: null,
            // `onTrailingComma` is similar to `onInsertedSemicolon`, but for
            // trailing commas.
            onTrailingComma: null,
            // By default, reserved words are only enforced if ecmaVersion >= 5.
            // Set `allowReserved` to a boolean value to explicitly turn this on
            // an off. When this option has the value "never", reserved words
            // and keywords can also not be used as property names.
            allowReserved: null,
            // When enabled, a return at the top level is not considered an
            // error.
            allowReturnOutsideFunction: false,
            // When enabled, import/export statements are not constrained to
            // appearing at the top of the program, and an import.meta expression
            // in a script isn't considered an error.
            allowImportExportEverywhere: false,
            // By default, await identifiers are allowed to appear at the top-level scope only if ecmaVersion >= 2022.
            // When enabled, await identifiers are allowed to appear at the top-level scope,
            // but they are still not allowed in non-async functions.
            allowAwaitOutsideFunction: null,
            // When enabled, super identifiers are not constrained to
            // appearing in methods and do not raise an error when they appear elsewhere.
            allowSuperOutsideMethod: null,
            // When enabled, hashbang directive in the beginning of file
            // is allowed and treated as a line comment.
            allowHashBang: false,
            // When `locations` is on, `loc` properties holding objects with
            // `start` and `end` properties in `{line, column}` form (with
            // line being 1-based and column 0-based) will be attached to the
            // nodes.
            locations: false,
            // A function can be passed as `onToken` option, which will
            // cause Acorn to call that function with object in the same
            // format as tokens returned from `tokenizer().getToken()`. Note
            // that you are not allowed to call the parser from the
            // callback—that will corrupt its internal state.
            onToken: null,
            // A function can be passed as `onComment` option, which will
            // cause Acorn to call that function with `(block, text, start,
            // end)` parameters whenever a comment is skipped. `block` is a
            // boolean indicating whether this is a block (`/* */`) comment,
            // `text` is the content of the comment, and `start` and `end` are
            // character offsets that denote the start and end of the comment.
            // When the `locations` option is on, two more parameters are
            // passed, the full `{line, column}` locations of the start and
            // end of the comments. Note that you are not allowed to call the
            // parser from the callback—that will corrupt its internal state.
            onComment: null,
            // Nodes have their start and end characters offsets recorded in
            // `start` and `end` properties (directly on the node, rather than
            // the `loc` object, which holds line/column data. To also add a
            // [semi-standardized][range] `range` property holding a `[start,
            // end]` array with the same numbers, set the `ranges` option to
            // `true`.
            //
            // [range]: https://bugzilla.mozilla.org/show_bug.cgi?id=745678
            ranges: false,
            // It is possible to parse multiple files into a single AST by
            // passing the tree produced by parsing the first file as
            // `program` option in subsequent parses. This will add the
            // toplevel forms of the parsed file to the `Program` (top) node
            // of an existing parse tree.
            program: null,
            // When `locations` is on, you can pass this to record the source
            // file in every node's `loc` object.
            sourceFile: null,
            // This value, if given, is stored in every node, whether
            // `locations` is on or off.
            directSourceFile: null,
            // When enabled, parenthesized expressions are represented by
            // (non-standard) ParenthesizedExpression nodes
            preserveParens: false
        };
        // Interpret and default an options object
        var warnedAboutEcmaVersion = false;
        function getOptions(opts) {
            var options = {};
            for (var opt in defaultOptions) {
                options[opt] = opts && has(opts, opt) ? opts[opt] : defaultOptions[opt];
            }
            if (options.ecmaVersion === "latest") {
                options.ecmaVersion = 1e8;
            }
            else if (options.ecmaVersion == null) {
                if (!warnedAboutEcmaVersion && typeof console === "object" && console.warn) {
                    warnedAboutEcmaVersion = true;
                    console.warn("Since Acorn 8.0.0, options.ecmaVersion is required.\nDefaulting to 2020, but this will stop working in the future.");
                }
                options.ecmaVersion = 11;
            }
            else if (options.ecmaVersion >= 2015) {
                options.ecmaVersion -= 2009;
            }
            if (options.allowReserved == null) {
                options.allowReserved = options.ecmaVersion < 5;
            }
            if (isArray(options.onToken)) {
                var tokens = options.onToken;
                options.onToken = function (token) { return tokens.push(token); };
            }
            if (isArray(options.onComment)) {
                options.onComment = pushComment(options, options.onComment);
            }
            return options;
        }
        function pushComment(options, array) {
            return function (block, text, start, end, startLoc, endLoc) {
                var comment = {
                    type: block ? "Block" : "Line",
                    value: text,
                    start: start,
                    end: end
                };
                if (options.locations) {
                    comment.loc = new SourceLocation(this, startLoc, endLoc);
                }
                if (options.ranges) {
                    comment.range = [start, end];
                }
                array.push(comment);
            };
        }
        // Each scope gets a bitset that may contain these flags
        var SCOPE_TOP = 1, SCOPE_FUNCTION = 2, SCOPE_VAR = SCOPE_TOP | SCOPE_FUNCTION, SCOPE_ASYNC = 4, SCOPE_GENERATOR = 8, SCOPE_ARROW = 16, SCOPE_SIMPLE_CATCH = 32, SCOPE_SUPER = 64, SCOPE_DIRECT_SUPER = 128;
        function functionFlags(async, generator) {
            return SCOPE_FUNCTION | (async ? SCOPE_ASYNC : 0) | (generator ? SCOPE_GENERATOR : 0);
        }
        // Used in checkLVal* and declareName to determine the type of a binding
        var BIND_NONE = 0, // Not a binding
        BIND_VAR = 1, // Var-style binding
        BIND_LEXICAL = 2, // Let- or const-style binding
        BIND_FUNCTION = 3, // Function declaration
        BIND_SIMPLE_CATCH = 4, // Simple (identifier pattern) catch binding
        BIND_OUTSIDE = 5; // Special case for function names as bound inside the function
        var Parser = function Parser(options, input, startPos) {
            this.options = options = getOptions(options);
            this.sourceFile = options.sourceFile;
            this.keywords = wordsRegexp(keywords[options.ecmaVersion >= 6 ? 6 : options.sourceType === "module" ? "5module" : 5]);
            var reserved = "";
            if (options.allowReserved !== true) {
                reserved = reservedWords[options.ecmaVersion >= 6 ? 6 : options.ecmaVersion === 5 ? 5 : 3];
                if (options.sourceType === "module") {
                    reserved += " await";
                }
            }
            this.reservedWords = wordsRegexp(reserved);
            var reservedStrict = (reserved ? reserved + " " : "") + reservedWords.strict;
            this.reservedWordsStrict = wordsRegexp(reservedStrict);
            this.reservedWordsStrictBind = wordsRegexp(reservedStrict + " " + reservedWords.strictBind);
            this.input = String(input);
            // Used to signal to callers of `readWord1` whether the word
            // contained any escape sequences. This is needed because words with
            // escape sequences must not be interpreted as keywords.
            this.containsEsc = false;
            // Set up token state
            // The current position of the tokenizer in the input.
            if (startPos) {
                this.pos = startPos;
                this.lineStart = this.input.lastIndexOf("\n", startPos - 1) + 1;
                this.curLine = this.input.slice(0, this.lineStart).split(lineBreak).length;
            }
            else {
                this.pos = this.lineStart = 0;
                this.curLine = 1;
            }
            // Properties of the current token:
            // Its type
            this.type = types.eof;
            // For tokens that include more information than their type, the value
            this.value = null;
            // Its start and end offset
            this.start = this.end = this.pos;
            // And, if locations are used, the {line, column} object
            // corresponding to those offsets
            this.startLoc = this.endLoc = this.curPosition();
            // Position information for the previous token
            this.lastTokEndLoc = this.lastTokStartLoc = null;
            this.lastTokStart = this.lastTokEnd = this.pos;
            // The context stack is used to superficially track syntactic
            // context to predict whether a regular expression is allowed in a
            // given position.
            this.context = this.initialContext();
            this.exprAllowed = true;
            // Figure out if it's a module code.
            this.inModule = options.sourceType === "module";
            this.strict = this.inModule || this.strictDirective(this.pos);
            // Used to signify the start of a potential arrow function
            this.potentialArrowAt = -1;
            this.potentialArrowInForAwait = false;
            // Positions to delayed-check that yield/await does not exist in default parameters.
            this.yieldPos = this.awaitPos = this.awaitIdentPos = 0;
            // Labels in scope.
            this.labels = [];
            // Thus-far undefined exports.
            this.undefinedExports = Object.create(null);
            // If enabled, skip leading hashbang line.
            if (this.pos === 0 && options.allowHashBang && this.input.slice(0, 2) === "#!") {
                this.skipLineComment(2);
            }
            // Scope tracking for duplicate variable names (see scope.js)
            this.scopeStack = [];
            this.enterScope(SCOPE_TOP);
            // For RegExp validation
            this.regexpState = null;
            // The stack of private names.
            // Each element has two properties: 'declared' and 'used'.
            // When it exited from the outermost class definition, all used private names must be declared.
            this.privateNameStack = [];
        };
        var prototypeAccessors = { inFunction: { configurable: true }, inGenerator: { configurable: true }, inAsync: { configurable: true }, canAwait: { configurable: true }, allowSuper: { configurable: true }, allowDirectSuper: { configurable: true }, treatFunctionsAsVar: { configurable: true }, inNonArrowFunction: { configurable: true } };
        Parser.prototype.parse = function parse() {
            var node = this.options.program || this.startNode();
            this.nextToken();
            return this.parseTopLevel(node);
        };
        prototypeAccessors.inFunction.get = function () { return (this.currentVarScope().flags & SCOPE_FUNCTION) > 0; };
        prototypeAccessors.inGenerator.get = function () { return (this.currentVarScope().flags & SCOPE_GENERATOR) > 0 && !this.currentVarScope().inClassFieldInit; };
        prototypeAccessors.inAsync.get = function () { return (this.currentVarScope().flags & SCOPE_ASYNC) > 0 && !this.currentVarScope().inClassFieldInit; };
        prototypeAccessors.canAwait.get = function () {
            for (var i = this.scopeStack.length - 1; i >= 0; i--) {
                var scope = this.scopeStack[i];
                if (scope.inClassFieldInit) {
                    return false;
                }
                if (scope.flags & SCOPE_FUNCTION) {
                    return (scope.flags & SCOPE_ASYNC) > 0;
                }
            }
            return (this.inModule && this.options.ecmaVersion >= 13) || this.options.allowAwaitOutsideFunction;
        };
        prototypeAccessors.allowSuper.get = function () {
            var ref = this.currentThisScope();
            var flags = ref.flags;
            var inClassFieldInit = ref.inClassFieldInit;
            return (flags & SCOPE_SUPER) > 0 || inClassFieldInit || this.options.allowSuperOutsideMethod;
        };
        prototypeAccessors.allowDirectSuper.get = function () { return (this.currentThisScope().flags & SCOPE_DIRECT_SUPER) > 0; };
        prototypeAccessors.treatFunctionsAsVar.get = function () { return this.treatFunctionsAsVarInScope(this.currentScope()); };
        prototypeAccessors.inNonArrowFunction.get = function () {
            var ref = this.currentThisScope();
            var flags = ref.flags;
            var inClassFieldInit = ref.inClassFieldInit;
            return (flags & SCOPE_FUNCTION) > 0 || inClassFieldInit;
        };
        Parser.extend = function extend() {
            var plugins = [], len = arguments.length;
            while (len--)
                plugins[len] = arguments[len];
            var cls = this;
            for (var i = 0; i < plugins.length; i++) {
                cls = plugins[i](cls);
            }
            return cls;
        };
        Parser.parse = function parse(input, options) {
            return new this(options, input).parse();
        };
        Parser.parseExpressionAt = function parseExpressionAt(input, pos, options) {
            var parser = new this(options, input, pos);
            parser.nextToken();
            return parser.parseExpression();
        };
        Parser.tokenizer = function tokenizer(input, options) {
            return new this(options, input);
        };
        Object.defineProperties(Parser.prototype, prototypeAccessors);
        var pp = Parser.prototype;
        // ## Parser utilities
        var literal = /^(?:'((?:\\.|[^'\\])*?)'|"((?:\\.|[^"\\])*?)")/;
        pp.strictDirective = function (start) {
            for (;;) {
                // Try to find string literal.
                skipWhiteSpace.lastIndex = start;
                start += skipWhiteSpace.exec(this.input)[0].length;
                var match = literal.exec(this.input.slice(start));
                if (!match) {
                    return false;
                }
                if ((match[1] || match[2]) === "use strict") {
                    skipWhiteSpace.lastIndex = start + match[0].length;
                    var spaceAfter = skipWhiteSpace.exec(this.input), end = spaceAfter.index + spaceAfter[0].length;
                    var next = this.input.charAt(end);
                    return next === ";" || next === "}" ||
                        (lineBreak.test(spaceAfter[0]) &&
                            !(/[(`.[+\-/*%<>=,?^&]/.test(next) || next === "!" && this.input.charAt(end + 1) === "="));
                }
                start += match[0].length;
                // Skip semicolon, if any.
                skipWhiteSpace.lastIndex = start;
                start += skipWhiteSpace.exec(this.input)[0].length;
                if (this.input[start] === ";") {
                    start++;
                }
            }
        };
        // Predicate that tests whether the next token is of the given
        // type, and if yes, consumes it as a side effect.
        pp.eat = function (type) {
            if (this.type === type) {
                this.next();
                return true;
            }
            else {
                return false;
            }
        };
        // Tests whether parsed token is a contextual keyword.
        pp.isContextual = function (name) {
            return this.type === types.name && this.value === name && !this.containsEsc;
        };
        // Consumes contextual keyword if possible.
        pp.eatContextual = function (name) {
            if (!this.isContextual(name)) {
                return false;
            }
            this.next();
            return true;
        };
        // Asserts that following token is given contextual keyword.
        pp.expectContextual = function (name) {
            if (!this.eatContextual(name)) {
                this.unexpected();
            }
        };
        // Test whether a semicolon can be inserted at the current position.
        pp.canInsertSemicolon = function () {
            return this.type === types.eof ||
                this.type === types.braceR ||
                lineBreak.test(this.input.slice(this.lastTokEnd, this.start));
        };
        pp.insertSemicolon = function () {
            if (this.canInsertSemicolon()) {
                if (this.options.onInsertedSemicolon) {
                    this.options.onInsertedSemicolon(this.lastTokEnd, this.lastTokEndLoc);
                }
                return true;
            }
        };
        // Consume a semicolon, or, failing that, see if we are allowed to
        // pretend that there is a semicolon at this position.
        pp.semicolon = function () {
            if (!this.eat(types.semi) && !this.insertSemicolon()) {
                this.unexpected();
            }
        };
        pp.afterTrailingComma = function (tokType, notNext) {
            if (this.type === tokType) {
                if (this.options.onTrailingComma) {
                    this.options.onTrailingComma(this.lastTokStart, this.lastTokStartLoc);
                }
                if (!notNext) {
                    this.next();
                }
                return true;
            }
        };
        // Expect a token of a given type. If found, consume it, otherwise,
        // raise an unexpected token error.
        pp.expect = function (type) {
            this.eat(type) || this.unexpected();
        };
        // Raise an unexpected token error.
        pp.unexpected = function (pos) {
            this.raise(pos != null ? pos : this.start, "Unexpected token");
        };
        function DestructuringErrors() {
            this.shorthandAssign =
                this.trailingComma =
                    this.parenthesizedAssign =
                        this.parenthesizedBind =
                            this.doubleProto =
                                -1;
        }
        pp.checkPatternErrors = function (refDestructuringErrors, isAssign) {
            if (!refDestructuringErrors) {
                return;
            }
            if (refDestructuringErrors.trailingComma > -1) {
                this.raiseRecoverable(refDestructuringErrors.trailingComma, "Comma is not permitted after the rest element");
            }
            var parens = isAssign ? refDestructuringErrors.parenthesizedAssign : refDestructuringErrors.parenthesizedBind;
            if (parens > -1) {
                this.raiseRecoverable(parens, "Parenthesized pattern");
            }
        };
        pp.checkExpressionErrors = function (refDestructuringErrors, andThrow) {
            if (!refDestructuringErrors) {
                return false;
            }
            var shorthandAssign = refDestructuringErrors.shorthandAssign;
            var doubleProto = refDestructuringErrors.doubleProto;
            if (!andThrow) {
                return shorthandAssign >= 0 || doubleProto >= 0;
            }
            if (shorthandAssign >= 0) {
                this.raise(shorthandAssign, "Shorthand property assignments are valid only in destructuring patterns");
            }
            if (doubleProto >= 0) {
                this.raiseRecoverable(doubleProto, "Redefinition of __proto__ property");
            }
        };
        pp.checkYieldAwaitInDefaultParams = function () {
            if (this.yieldPos && (!this.awaitPos || this.yieldPos < this.awaitPos)) {
                this.raise(this.yieldPos, "Yield expression cannot be a default value");
            }
            if (this.awaitPos) {
                this.raise(this.awaitPos, "Await expression cannot be a default value");
            }
        };
        pp.isSimpleAssignTarget = function (expr) {
            if (expr.type === "ParenthesizedExpression") {
                return this.isSimpleAssignTarget(expr.expression);
            }
            return expr.type === "Identifier" || expr.type === "MemberExpression";
        };
        var pp$1 = Parser.prototype;
        // ### Statement parsing
        // Parse a program. Initializes the parser, reads any number of
        // statements, and wraps them in a Program node.  Optionally takes a
        // `program` argument.  If present, the statements will be appended
        // to its body instead of creating a new node.
        pp$1.parseTopLevel = function (node) {
            var exports = Object.create(null);
            if (!node.body) {
                node.body = [];
            }
            while (this.type !== types.eof) {
                var stmt = this.parseStatement(null, true, exports);
                node.body.push(stmt);
            }
            if (this.inModule) {
                for (var i = 0, list = Object.keys(this.undefinedExports); i < list.length; i += 1) {
                    var name = list[i];
                    this.raiseRecoverable(this.undefinedExports[name].start, ("Export '" + name + "' is not defined"));
                }
            }
            this.adaptDirectivePrologue(node.body);
            this.next();
            node.sourceType = this.options.sourceType;
            return this.finishNode(node, "Program");
        };
        var loopLabel = { kind: "loop" }, switchLabel = { kind: "switch" };
        pp$1.isLet = function (context) {
            if (this.options.ecmaVersion < 6 || !this.isContextual("let")) {
                return false;
            }
            skipWhiteSpace.lastIndex = this.pos;
            var skip = skipWhiteSpace.exec(this.input);
            var next = this.pos + skip[0].length, nextCh = this.input.charCodeAt(next);
            // For ambiguous cases, determine if a LexicalDeclaration (or only a
            // Statement) is allowed here. If context is not empty then only a Statement
            // is allowed. However, `let [` is an explicit negative lookahead for
            // ExpressionStatement, so special-case it first.
            if (nextCh === 91 || nextCh === 92 || nextCh > 0xd7ff && nextCh < 0xdc00) {
                return true;
            } // '[', '/', astral
            if (context) {
                return false;
            }
            if (nextCh === 123) {
                return true;
            } // '{'
            if (isIdentifierStart(nextCh, true)) {
                var pos = next + 1;
                while (isIdentifierChar(nextCh = this.input.charCodeAt(pos), true)) {
                    ++pos;
                }
                if (nextCh === 92 || nextCh > 0xd7ff && nextCh < 0xdc00) {
                    return true;
                }
                var ident = this.input.slice(next, pos);
                if (!keywordRelationalOperator.test(ident)) {
                    return true;
                }
            }
            return false;
        };
        // check 'async [no LineTerminator here] function'
        // - 'async /*foo*/ function' is OK.
        // - 'async /*\n*/ function' is invalid.
        pp$1.isAsyncFunction = function () {
            if (this.options.ecmaVersion < 8 || !this.isContextual("async")) {
                return false;
            }
            skipWhiteSpace.lastIndex = this.pos;
            var skip = skipWhiteSpace.exec(this.input);
            var next = this.pos + skip[0].length, after;
            return !lineBreak.test(this.input.slice(this.pos, next)) &&
                this.input.slice(next, next + 8) === "function" &&
                (next + 8 === this.input.length ||
                    !(isIdentifierChar(after = this.input.charCodeAt(next + 8)) || after > 0xd7ff && after < 0xdc00));
        };
        // Parse a single statement.
        //
        // If expecting a statement and finding a slash operator, parse a
        // regular expression literal. This is to handle cases like
        // `if (foo) /blah/.exec(foo)`, where looking at the previous token
        // does not help.
        pp$1.parseStatement = function (context, topLevel, exports) {
            var starttype = this.type, node = this.startNode(), kind;
            if (this.isLet(context)) {
                starttype = types._var;
                kind = "let";
            }
            // Most types of statements are recognized by the keyword they
            // start with. Many are trivial to parse, some require a bit of
            // complexity.
            switch (starttype) {
                case types._break:
                case types._continue: return this.parseBreakContinueStatement(node, starttype.keyword);
                case types._debugger: return this.parseDebuggerStatement(node);
                case types._do: return this.parseDoStatement(node);
                case types._for: return this.parseForStatement(node);
                case types._function:
                    // Function as sole body of either an if statement or a labeled statement
                    // works, but not when it is part of a labeled statement that is the sole
                    // body of an if statement.
                    if ((context && (this.strict || context !== "if" && context !== "label")) && this.options.ecmaVersion >= 6) {
                        this.unexpected();
                    }
                    return this.parseFunctionStatement(node, false, !context);
                case types._class:
                    if (context) {
                        this.unexpected();
                    }
                    return this.parseClass(node, true);
                case types._if: return this.parseIfStatement(node);
                case types._return: return this.parseReturnStatement(node);
                case types._switch: return this.parseSwitchStatement(node);
                case types._throw: return this.parseThrowStatement(node);
                case types._try: return this.parseTryStatement(node);
                case types._const:
                case types._var:
                    kind = kind || this.value;
                    if (context && kind !== "var") {
                        this.unexpected();
                    }
                    return this.parseVarStatement(node, kind);
                case types._while: return this.parseWhileStatement(node);
                case types._with: return this.parseWithStatement(node);
                case types.braceL: return this.parseBlock(true, node);
                case types.semi: return this.parseEmptyStatement(node);
                case types._export:
                case types._import:
                    if (this.options.ecmaVersion > 10 && starttype === types._import) {
                        skipWhiteSpace.lastIndex = this.pos;
                        var skip = skipWhiteSpace.exec(this.input);
                        var next = this.pos + skip[0].length, nextCh = this.input.charCodeAt(next);
                        if (nextCh === 40 || nextCh === 46) // '(' or '.'
                         {
                            return this.parseExpressionStatement(node, this.parseExpression());
                        }
                    }
                    if (!this.options.allowImportExportEverywhere) {
                        if (!topLevel) {
                            this.raise(this.start, "'import' and 'export' may only appear at the top level");
                        }
                        if (!this.inModule) {
                            this.raise(this.start, "'import' and 'export' may appear only with 'sourceType: module'");
                        }
                    }
                    return starttype === types._import ? this.parseImport(node) : this.parseExport(node, exports);
                // If the statement does not start with a statement keyword or a
                // brace, it's an ExpressionStatement or LabeledStatement. We
                // simply start parsing an expression, and afterwards, if the
                // next token is a colon and the expression was a simple
                // Identifier node, we switch to interpreting it as a label.
                default:
                    if (this.isAsyncFunction()) {
                        if (context) {
                            this.unexpected();
                        }
                        this.next();
                        return this.parseFunctionStatement(node, true, !context);
                    }
                    var maybeName = this.value, expr = this.parseExpression();
                    if (starttype === types.name && expr.type === "Identifier" && this.eat(types.colon)) {
                        return this.parseLabeledStatement(node, maybeName, expr, context);
                    }
                    else {
                        return this.parseExpressionStatement(node, expr);
                    }
            }
        };
        pp$1.parseBreakContinueStatement = function (node, keyword) {
            var isBreak = keyword === "break";
            this.next();
            if (this.eat(types.semi) || this.insertSemicolon()) {
                node.label = null;
            }
            else if (this.type !== types.name) {
                this.unexpected();
            }
            else {
                node.label = this.parseIdent();
                this.semicolon();
            }
            // Verify that there is an actual destination to break or
            // continue to.
            var i = 0;
            for (; i < this.labels.length; ++i) {
                var lab = this.labels[i];
                if (node.label == null || lab.name === node.label.name) {
                    if (lab.kind != null && (isBreak || lab.kind === "loop")) {
                        break;
                    }
                    if (node.label && isBreak) {
                        break;
                    }
                }
            }
            if (i === this.labels.length) {
                this.raise(node.start, "Unsyntactic " + keyword);
            }
            return this.finishNode(node, isBreak ? "BreakStatement" : "ContinueStatement");
        };
        pp$1.parseDebuggerStatement = function (node) {
            this.next();
            this.semicolon();
            return this.finishNode(node, "DebuggerStatement");
        };
        pp$1.parseDoStatement = function (node) {
            this.next();
            this.labels.push(loopLabel);
            node.body = this.parseStatement("do");
            this.labels.pop();
            this.expect(types._while);
            node.test = this.parseParenExpression();
            if (this.options.ecmaVersion >= 6) {
                this.eat(types.semi);
            }
            else {
                this.semicolon();
            }
            return this.finishNode(node, "DoWhileStatement");
        };
        // Disambiguating between a `for` and a `for`/`in` or `for`/`of`
        // loop is non-trivial. Basically, we have to parse the init `var`
        // statement or expression, disallowing the `in` operator (see
        // the second parameter to `parseExpression`), and then check
        // whether the next token is `in` or `of`. When there is no init
        // part (semicolon immediately after the opening parenthesis), it
        // is a regular `for` loop.
        pp$1.parseForStatement = function (node) {
            this.next();
            var awaitAt = (this.options.ecmaVersion >= 9 && this.canAwait && this.eatContextual("await")) ? this.lastTokStart : -1;
            this.labels.push(loopLabel);
            this.enterScope(0);
            this.expect(types.parenL);
            if (this.type === types.semi) {
                if (awaitAt > -1) {
                    this.unexpected(awaitAt);
                }
                return this.parseFor(node, null);
            }
            var isLet = this.isLet();
            if (this.type === types._var || this.type === types._const || isLet) {
                var init$1 = this.startNode(), kind = isLet ? "let" : this.value;
                this.next();
                this.parseVar(init$1, true, kind);
                this.finishNode(init$1, "VariableDeclaration");
                if ((this.type === types._in || (this.options.ecmaVersion >= 6 && this.isContextual("of"))) && init$1.declarations.length === 1) {
                    if (this.options.ecmaVersion >= 9) {
                        if (this.type === types._in) {
                            if (awaitAt > -1) {
                                this.unexpected(awaitAt);
                            }
                        }
                        else {
                            node.await = awaitAt > -1;
                        }
                    }
                    return this.parseForIn(node, init$1);
                }
                if (awaitAt > -1) {
                    this.unexpected(awaitAt);
                }
                return this.parseFor(node, init$1);
            }
            var refDestructuringErrors = new DestructuringErrors;
            var init = this.parseExpression(awaitAt > -1 ? "await" : true, refDestructuringErrors);
            if (this.type === types._in || (this.options.ecmaVersion >= 6 && this.isContextual("of"))) {
                if (this.options.ecmaVersion >= 9) {
                    if (this.type === types._in) {
                        if (awaitAt > -1) {
                            this.unexpected(awaitAt);
                        }
                    }
                    else {
                        node.await = awaitAt > -1;
                    }
                }
                this.toAssignable(init, false, refDestructuringErrors);
                this.checkLValPattern(init);
                return this.parseForIn(node, init);
            }
            else {
                this.checkExpressionErrors(refDestructuringErrors, true);
            }
            if (awaitAt > -1) {
                this.unexpected(awaitAt);
            }
            return this.parseFor(node, init);
        };
        pp$1.parseFunctionStatement = function (node, isAsync, declarationPosition) {
            this.next();
            return this.parseFunction(node, FUNC_STATEMENT | (declarationPosition ? 0 : FUNC_HANGING_STATEMENT), false, isAsync);
        };
        pp$1.parseIfStatement = function (node) {
            this.next();
            node.test = this.parseParenExpression();
            // allow function declarations in branches, but only in non-strict mode
            node.consequent = this.parseStatement("if");
            node.alternate = this.eat(types._else) ? this.parseStatement("if") : null;
            return this.finishNode(node, "IfStatement");
        };
        pp$1.parseReturnStatement = function (node) {
            if (!this.inFunction && !this.options.allowReturnOutsideFunction) {
                this.raise(this.start, "'return' outside of function");
            }
            this.next();
            // In `return` (and `break`/`continue`), the keywords with
            // optional arguments, we eagerly look for a semicolon or the
            // possibility to insert one.
            if (this.eat(types.semi) || this.insertSemicolon()) {
                node.argument = null;
            }
            else {
                node.argument = this.parseExpression();
                this.semicolon();
            }
            return this.finishNode(node, "ReturnStatement");
        };
        pp$1.parseSwitchStatement = function (node) {
            this.next();
            node.discriminant = this.parseParenExpression();
            node.cases = [];
            this.expect(types.braceL);
            this.labels.push(switchLabel);
            this.enterScope(0);
            // Statements under must be grouped (by label) in SwitchCase
            // nodes. `cur` is used to keep the node that we are currently
            // adding statements to.
            var cur;
            for (var sawDefault = false; this.type !== types.braceR;) {
                if (this.type === types._case || this.type === types._default) {
                    var isCase = this.type === types._case;
                    if (cur) {
                        this.finishNode(cur, "SwitchCase");
                    }
                    node.cases.push(cur = this.startNode());
                    cur.consequent = [];
                    this.next();
                    if (isCase) {
                        cur.test = this.parseExpression();
                    }
                    else {
                        if (sawDefault) {
                            this.raiseRecoverable(this.lastTokStart, "Multiple default clauses");
                        }
                        sawDefault = true;
                        cur.test = null;
                    }
                    this.expect(types.colon);
                }
                else {
                    if (!cur) {
                        this.unexpected();
                    }
                    cur.consequent.push(this.parseStatement(null));
                }
            }
            this.exitScope();
            if (cur) {
                this.finishNode(cur, "SwitchCase");
            }
            this.next(); // Closing brace
            this.labels.pop();
            return this.finishNode(node, "SwitchStatement");
        };
        pp$1.parseThrowStatement = function (node) {
            this.next();
            if (lineBreak.test(this.input.slice(this.lastTokEnd, this.start))) {
                this.raise(this.lastTokEnd, "Illegal newline after throw");
            }
            node.argument = this.parseExpression();
            this.semicolon();
            return this.finishNode(node, "ThrowStatement");
        };
        // Reused empty array added for node fields that are always empty.
        var empty = [];
        pp$1.parseTryStatement = function (node) {
            this.next();
            node.block = this.parseBlock();
            node.handler = null;
            if (this.type === types._catch) {
                var clause = this.startNode();
                this.next();
                if (this.eat(types.parenL)) {
                    clause.param = this.parseBindingAtom();
                    var simple = clause.param.type === "Identifier";
                    this.enterScope(simple ? SCOPE_SIMPLE_CATCH : 0);
                    this.checkLValPattern(clause.param, simple ? BIND_SIMPLE_CATCH : BIND_LEXICAL);
                    this.expect(types.parenR);
                }
                else {
                    if (this.options.ecmaVersion < 10) {
                        this.unexpected();
                    }
                    clause.param = null;
                    this.enterScope(0);
                }
                clause.body = this.parseBlock(false);
                this.exitScope();
                node.handler = this.finishNode(clause, "CatchClause");
            }
            node.finalizer = this.eat(types._finally) ? this.parseBlock() : null;
            if (!node.handler && !node.finalizer) {
                this.raise(node.start, "Missing catch or finally clause");
            }
            return this.finishNode(node, "TryStatement");
        };
        pp$1.parseVarStatement = function (node, kind) {
            this.next();
            this.parseVar(node, false, kind);
            this.semicolon();
            return this.finishNode(node, "VariableDeclaration");
        };
        pp$1.parseWhileStatement = function (node) {
            this.next();
            node.test = this.parseParenExpression();
            this.labels.push(loopLabel);
            node.body = this.parseStatement("while");
            this.labels.pop();
            return this.finishNode(node, "WhileStatement");
        };
        pp$1.parseWithStatement = function (node) {
            if (this.strict) {
                this.raise(this.start, "'with' in strict mode");
            }
            this.next();
            node.object = this.parseParenExpression();
            node.body = this.parseStatement("with");
            return this.finishNode(node, "WithStatement");
        };
        pp$1.parseEmptyStatement = function (node) {
            this.next();
            return this.finishNode(node, "EmptyStatement");
        };
        pp$1.parseLabeledStatement = function (node, maybeName, expr, context) {
            for (var i$1 = 0, list = this.labels; i$1 < list.length; i$1 += 1) {
                var label = list[i$1];
                if (label.name === maybeName) {
                    this.raise(expr.start, "Label '" + maybeName + "' is already declared");
                }
            }
            var kind = this.type.isLoop ? "loop" : this.type === types._switch ? "switch" : null;
            for (var i = this.labels.length - 1; i >= 0; i--) {
                var label$1 = this.labels[i];
                if (label$1.statementStart === node.start) {
                    // Update information about previous labels on this node
                    label$1.statementStart = this.start;
                    label$1.kind = kind;
                }
                else {
                    break;
                }
            }
            this.labels.push({ name: maybeName, kind: kind, statementStart: this.start });
            node.body = this.parseStatement(context ? context.indexOf("label") === -1 ? context + "label" : context : "label");
            this.labels.pop();
            node.label = expr;
            return this.finishNode(node, "LabeledStatement");
        };
        pp$1.parseExpressionStatement = function (node, expr) {
            node.expression = expr;
            this.semicolon();
            return this.finishNode(node, "ExpressionStatement");
        };
        // Parse a semicolon-enclosed block of statements, handling `"use
        // strict"` declarations when `allowStrict` is true (used for
        // function bodies).
        pp$1.parseBlock = function (createNewLexicalScope, node, exitStrict) {
            if (createNewLexicalScope === void 0)
                createNewLexicalScope = true;
            if (node === void 0)
                node = this.startNode();
            node.body = [];
            this.expect(types.braceL);
            if (createNewLexicalScope) {
                this.enterScope(0);
            }
            while (this.type !== types.braceR) {
                var stmt = this.parseStatement(null);
                node.body.push(stmt);
            }
            if (exitStrict) {
                this.strict = false;
            }
            this.next();
            if (createNewLexicalScope) {
                this.exitScope();
            }
            return this.finishNode(node, "BlockStatement");
        };
        // Parse a regular `for` loop. The disambiguation code in
        // `parseStatement` will already have parsed the init statement or
        // expression.
        pp$1.parseFor = function (node, init) {
            node.init = init;
            this.expect(types.semi);
            node.test = this.type === types.semi ? null : this.parseExpression();
            this.expect(types.semi);
            node.update = this.type === types.parenR ? null : this.parseExpression();
            this.expect(types.parenR);
            node.body = this.parseStatement("for");
            this.exitScope();
            this.labels.pop();
            return this.finishNode(node, "ForStatement");
        };
        // Parse a `for`/`in` and `for`/`of` loop, which are almost
        // same from parser's perspective.
        pp$1.parseForIn = function (node, init) {
            var isForIn = this.type === types._in;
            this.next();
            if (init.type === "VariableDeclaration" &&
                init.declarations[0].init != null &&
                (!isForIn ||
                    this.options.ecmaVersion < 8 ||
                    this.strict ||
                    init.kind !== "var" ||
                    init.declarations[0].id.type !== "Identifier")) {
                this.raise(init.start, ((isForIn ? "for-in" : "for-of") + " loop variable declaration may not have an initializer"));
            }
            node.left = init;
            node.right = isForIn ? this.parseExpression() : this.parseMaybeAssign();
            this.expect(types.parenR);
            node.body = this.parseStatement("for");
            this.exitScope();
            this.labels.pop();
            return this.finishNode(node, isForIn ? "ForInStatement" : "ForOfStatement");
        };
        // Parse a list of variable declarations.
        pp$1.parseVar = function (node, isFor, kind) {
            node.declarations = [];
            node.kind = kind;
            for (;;) {
                var decl = this.startNode();
                this.parseVarId(decl, kind);
                if (this.eat(types.eq)) {
                    decl.init = this.parseMaybeAssign(isFor);
                }
                else if (kind === "const" && !(this.type === types._in || (this.options.ecmaVersion >= 6 && this.isContextual("of")))) {
                    this.unexpected();
                }
                else if (decl.id.type !== "Identifier" && !(isFor && (this.type === types._in || this.isContextual("of")))) {
                    this.raise(this.lastTokEnd, "Complex binding patterns require an initialization value");
                }
                else {
                    decl.init = null;
                }
                node.declarations.push(this.finishNode(decl, "VariableDeclarator"));
                if (!this.eat(types.comma)) {
                    break;
                }
            }
            return node;
        };
        pp$1.parseVarId = function (decl, kind) {
            decl.id = this.parseBindingAtom();
            this.checkLValPattern(decl.id, kind === "var" ? BIND_VAR : BIND_LEXICAL, false);
        };
        var FUNC_STATEMENT = 1, FUNC_HANGING_STATEMENT = 2, FUNC_NULLABLE_ID = 4;
        // Parse a function declaration or literal (depending on the
        // `statement & FUNC_STATEMENT`).
        // Remove `allowExpressionBody` for 7.0.0, as it is only called with false
        pp$1.parseFunction = function (node, statement, allowExpressionBody, isAsync) {
            this.initFunction(node);
            if (this.options.ecmaVersion >= 9 || this.options.ecmaVersion >= 6 && !isAsync) {
                if (this.type === types.star && (statement & FUNC_HANGING_STATEMENT)) {
                    this.unexpected();
                }
                node.generator = this.eat(types.star);
            }
            if (this.options.ecmaVersion >= 8) {
                node.async = !!isAsync;
            }
            if (statement & FUNC_STATEMENT) {
                node.id = (statement & FUNC_NULLABLE_ID) && this.type !== types.name ? null : this.parseIdent();
                if (node.id && !(statement & FUNC_HANGING_STATEMENT)) 
                // If it is a regular function declaration in sloppy mode, then it is
                // subject to Annex B semantics (BIND_FUNCTION). Otherwise, the binding
                // mode depends on properties of the current scope (see
                // treatFunctionsAsVar).
                {
                    this.checkLValSimple(node.id, (this.strict || node.generator || node.async) ? this.treatFunctionsAsVar ? BIND_VAR : BIND_LEXICAL : BIND_FUNCTION);
                }
            }
            var oldYieldPos = this.yieldPos, oldAwaitPos = this.awaitPos, oldAwaitIdentPos = this.awaitIdentPos;
            this.yieldPos = 0;
            this.awaitPos = 0;
            this.awaitIdentPos = 0;
            this.enterScope(functionFlags(node.async, node.generator));
            if (!(statement & FUNC_STATEMENT)) {
                node.id = this.type === types.name ? this.parseIdent() : null;
            }
            this.parseFunctionParams(node);
            this.parseFunctionBody(node, allowExpressionBody, false);
            this.yieldPos = oldYieldPos;
            this.awaitPos = oldAwaitPos;
            this.awaitIdentPos = oldAwaitIdentPos;
            return this.finishNode(node, (statement & FUNC_STATEMENT) ? "FunctionDeclaration" : "FunctionExpression");
        };
        pp$1.parseFunctionParams = function (node) {
            this.expect(types.parenL);
            node.params = this.parseBindingList(types.parenR, false, this.options.ecmaVersion >= 8);
            this.checkYieldAwaitInDefaultParams();
        };
        // Parse a class declaration or literal (depending on the
        // `isStatement` parameter).
        pp$1.parseClass = function (node, isStatement) {
            this.next();
            // ecma-262 14.6 Class Definitions
            // A class definition is always strict mode code.
            var oldStrict = this.strict;
            this.strict = true;
            this.parseClassId(node, isStatement);
            this.parseClassSuper(node);
            var privateNameMap = this.enterClassBody();
            var classBody = this.startNode();
            var hadConstructor = false;
            classBody.body = [];
            this.expect(types.braceL);
            while (this.type !== types.braceR) {
                var element = this.parseClassElement(node.superClass !== null);
                if (element) {
                    classBody.body.push(element);
                    if (element.type === "MethodDefinition" && element.kind === "constructor") {
                        if (hadConstructor) {
                            this.raise(element.start, "Duplicate constructor in the same class");
                        }
                        hadConstructor = true;
                    }
                    else if (element.key.type === "PrivateIdentifier" && isPrivateNameConflicted(privateNameMap, element)) {
                        this.raiseRecoverable(element.key.start, ("Identifier '#" + (element.key.name) + "' has already been declared"));
                    }
                }
            }
            this.strict = oldStrict;
            this.next();
            node.body = this.finishNode(classBody, "ClassBody");
            this.exitClassBody();
            return this.finishNode(node, isStatement ? "ClassDeclaration" : "ClassExpression");
        };
        pp$1.parseClassElement = function (constructorAllowsSuper) {
            if (this.eat(types.semi)) {
                return null;
            }
            var ecmaVersion = this.options.ecmaVersion;
            var node = this.startNode();
            var keyName = "";
            var isGenerator = false;
            var isAsync = false;
            var kind = "method";
            // Parse modifiers
            node.static = false;
            if (this.eatContextual("static")) {
                if (this.isClassElementNameStart() || this.type === types.star) {
                    node.static = true;
                }
                else {
                    keyName = "static";
                }
            }
            if (!keyName && ecmaVersion >= 8 && this.eatContextual("async")) {
                if ((this.isClassElementNameStart() || this.type === types.star) && !this.canInsertSemicolon()) {
                    isAsync = true;
                }
                else {
                    keyName = "async";
                }
            }
            if (!keyName && (ecmaVersion >= 9 || !isAsync) && this.eat(types.star)) {
                isGenerator = true;
            }
            if (!keyName && !isAsync && !isGenerator) {
                var lastValue = this.value;
                if (this.eatContextual("get") || this.eatContextual("set")) {
                    if (this.isClassElementNameStart()) {
                        kind = lastValue;
                    }
                    else {
                        keyName = lastValue;
                    }
                }
            }
            // Parse element name
            if (keyName) {
                // 'async', 'get', 'set', or 'static' were not a keyword contextually.
                // The last token is any of those. Make it the element name.
                node.computed = false;
                node.key = this.startNodeAt(this.lastTokStart, this.lastTokStartLoc);
                node.key.name = keyName;
                this.finishNode(node.key, "Identifier");
            }
            else {
                this.parseClassElementName(node);
            }
            // Parse element value
            if (ecmaVersion < 13 || this.type === types.parenL || kind !== "method" || isGenerator || isAsync) {
                var isConstructor = !node.static && checkKeyName(node, "constructor");
                var allowsDirectSuper = isConstructor && constructorAllowsSuper;
                // Couldn't move this check into the 'parseClassMethod' method for backward compatibility.
                if (isConstructor && kind !== "method") {
                    this.raise(node.key.start, "Constructor can't have get/set modifier");
                }
                node.kind = isConstructor ? "constructor" : kind;
                this.parseClassMethod(node, isGenerator, isAsync, allowsDirectSuper);
            }
            else {
                this.parseClassField(node);
            }
            return node;
        };
        pp$1.isClassElementNameStart = function () {
            return (this.type === types.name ||
                this.type === types.privateId ||
                this.type === types.num ||
                this.type === types.string ||
                this.type === types.bracketL ||
                this.type.keyword);
        };
        pp$1.parseClassElementName = function (element) {
            if (this.type === types.privateId) {
                if (this.value === "constructor") {
                    this.raise(this.start, "Classes can't have an element named '#constructor'");
                }
                element.computed = false;
                element.key = this.parsePrivateIdent();
            }
            else {
                this.parsePropertyName(element);
            }
        };
        pp$1.parseClassMethod = function (method, isGenerator, isAsync, allowsDirectSuper) {
            // Check key and flags
            var key = method.key;
            if (method.kind === "constructor") {
                if (isGenerator) {
                    this.raise(key.start, "Constructor can't be a generator");
                }
                if (isAsync) {
                    this.raise(key.start, "Constructor can't be an async method");
                }
            }
            else if (method.static && checkKeyName(method, "prototype")) {
                this.raise(key.start, "Classes may not have a static property named prototype");
            }
            // Parse value
            var value = method.value = this.parseMethod(isGenerator, isAsync, allowsDirectSuper);
            // Check value
            if (method.kind === "get" && value.params.length !== 0) {
                this.raiseRecoverable(value.start, "getter should have no params");
            }
            if (method.kind === "set" && value.params.length !== 1) {
                this.raiseRecoverable(value.start, "setter should have exactly one param");
            }
            if (method.kind === "set" && value.params[0].type === "RestElement") {
                this.raiseRecoverable(value.params[0].start, "Setter cannot use rest params");
            }
            return this.finishNode(method, "MethodDefinition");
        };
        pp$1.parseClassField = function (field) {
            if (checkKeyName(field, "constructor")) {
                this.raise(field.key.start, "Classes can't have a field named 'constructor'");
            }
            else if (field.static && checkKeyName(field, "prototype")) {
                this.raise(field.key.start, "Classes can't have a static field named 'prototype'");
            }
            if (this.eat(types.eq)) {
                // To raise SyntaxError if 'arguments' exists in the initializer.
                var scope = this.currentThisScope();
                var inClassFieldInit = scope.inClassFieldInit;
                scope.inClassFieldInit = true;
                field.value = this.parseMaybeAssign();
                scope.inClassFieldInit = inClassFieldInit;
            }
            else {
                field.value = null;
            }
            this.semicolon();
            return this.finishNode(field, "PropertyDefinition");
        };
        pp$1.parseClassId = function (node, isStatement) {
            if (this.type === types.name) {
                node.id = this.parseIdent();
                if (isStatement) {
                    this.checkLValSimple(node.id, BIND_LEXICAL, false);
                }
            }
            else {
                if (isStatement === true) {
                    this.unexpected();
                }
                node.id = null;
            }
        };
        pp$1.parseClassSuper = function (node) {
            node.superClass = this.eat(types._extends) ? this.parseExprSubscripts() : null;
        };
        pp$1.enterClassBody = function () {
            var element = { declared: Object.create(null), used: [] };
            this.privateNameStack.push(element);
            return element.declared;
        };
        pp$1.exitClassBody = function () {
            var ref = this.privateNameStack.pop();
            var declared = ref.declared;
            var used = ref.used;
            var len = this.privateNameStack.length;
            var parent = len === 0 ? null : this.privateNameStack[len - 1];
            for (var i = 0; i < used.length; ++i) {
                var id = used[i];
                if (!has(declared, id.name)) {
                    if (parent) {
                        parent.used.push(id);
                    }
                    else {
                        this.raiseRecoverable(id.start, ("Private field '#" + (id.name) + "' must be declared in an enclosing class"));
                    }
                }
            }
        };
        function isPrivateNameConflicted(privateNameMap, element) {
            var name = element.key.name;
            var curr = privateNameMap[name];
            var next = "true";
            if (element.type === "MethodDefinition" && (element.kind === "get" || element.kind === "set")) {
                next = (element.static ? "s" : "i") + element.kind;
            }
            // `class { get #a(){}; static set #a(_){} }` is also conflict.
            if (curr === "iget" && next === "iset" ||
                curr === "iset" && next === "iget" ||
                curr === "sget" && next === "sset" ||
                curr === "sset" && next === "sget") {
                privateNameMap[name] = "true";
                return false;
            }
            else if (!curr) {
                privateNameMap[name] = next;
                return false;
            }
            else {
                return true;
            }
        }
        function checkKeyName(node, name) {
            var computed = node.computed;
            var key = node.key;
            return !computed && (key.type === "Identifier" && key.name === name ||
                key.type === "Literal" && key.value === name);
        }
        // Parses module export declaration.
        pp$1.parseExport = function (node, exports) {
            this.next();
            // export * from '...'
            if (this.eat(types.star)) {
                if (this.options.ecmaVersion >= 11) {
                    if (this.eatContextual("as")) {
                        node.exported = this.parseIdent(true);
                        this.checkExport(exports, node.exported.name, this.lastTokStart);
                    }
                    else {
                        node.exported = null;
                    }
                }
                this.expectContextual("from");
                if (this.type !== types.string) {
                    this.unexpected();
                }
                node.source = this.parseExprAtom();
                this.semicolon();
                return this.finishNode(node, "ExportAllDeclaration");
            }
            if (this.eat(types._default)) { // export default ...
                this.checkExport(exports, "default", this.lastTokStart);
                var isAsync;
                if (this.type === types._function || (isAsync = this.isAsyncFunction())) {
                    var fNode = this.startNode();
                    this.next();
                    if (isAsync) {
                        this.next();
                    }
                    node.declaration = this.parseFunction(fNode, FUNC_STATEMENT | FUNC_NULLABLE_ID, false, isAsync);
                }
                else if (this.type === types._class) {
                    var cNode = this.startNode();
                    node.declaration = this.parseClass(cNode, "nullableID");
                }
                else {
                    node.declaration = this.parseMaybeAssign();
                    this.semicolon();
                }
                return this.finishNode(node, "ExportDefaultDeclaration");
            }
            // export var|const|let|function|class ...
            if (this.shouldParseExportStatement()) {
                node.declaration = this.parseStatement(null);
                if (node.declaration.type === "VariableDeclaration") {
                    this.checkVariableExport(exports, node.declaration.declarations);
                }
                else {
                    this.checkExport(exports, node.declaration.id.name, node.declaration.id.start);
                }
                node.specifiers = [];
                node.source = null;
            }
            else { // export { x, y as z } [from '...']
                node.declaration = null;
                node.specifiers = this.parseExportSpecifiers(exports);
                if (this.eatContextual("from")) {
                    if (this.type !== types.string) {
                        this.unexpected();
                    }
                    node.source = this.parseExprAtom();
                }
                else {
                    for (var i = 0, list = node.specifiers; i < list.length; i += 1) {
                        // check for keywords used as local names
                        var spec = list[i];
                        this.checkUnreserved(spec.local);
                        // check if export is defined
                        this.checkLocalExport(spec.local);
                    }
                    node.source = null;
                }
                this.semicolon();
            }
            return this.finishNode(node, "ExportNamedDeclaration");
        };
        pp$1.checkExport = function (exports, name, pos) {
            if (!exports) {
                return;
            }
            if (has(exports, name)) {
                this.raiseRecoverable(pos, "Duplicate export '" + name + "'");
            }
            exports[name] = true;
        };
        pp$1.checkPatternExport = function (exports, pat) {
            var type = pat.type;
            if (type === "Identifier") {
                this.checkExport(exports, pat.name, pat.start);
            }
            else if (type === "ObjectPattern") {
                for (var i = 0, list = pat.properties; i < list.length; i += 1) {
                    var prop = list[i];
                    this.checkPatternExport(exports, prop);
                }
            }
            else if (type === "ArrayPattern") {
                for (var i$1 = 0, list$1 = pat.elements; i$1 < list$1.length; i$1 += 1) {
                    var elt = list$1[i$1];
                    if (elt) {
                        this.checkPatternExport(exports, elt);
                    }
                }
            }
            else if (type === "Property") {
                this.checkPatternExport(exports, pat.value);
            }
            else if (type === "AssignmentPattern") {
                this.checkPatternExport(exports, pat.left);
            }
            else if (type === "RestElement") {
                this.checkPatternExport(exports, pat.argument);
            }
            else if (type === "ParenthesizedExpression") {
                this.checkPatternExport(exports, pat.expression);
            }
        };
        pp$1.checkVariableExport = function (exports, decls) {
            if (!exports) {
                return;
            }
            for (var i = 0, list = decls; i < list.length; i += 1) {
                var decl = list[i];
                this.checkPatternExport(exports, decl.id);
            }
        };
        pp$1.shouldParseExportStatement = function () {
            return this.type.keyword === "var" ||
                this.type.keyword === "const" ||
                this.type.keyword === "class" ||
                this.type.keyword === "function" ||
                this.isLet() ||
                this.isAsyncFunction();
        };
        // Parses a comma-separated list of module exports.
        pp$1.parseExportSpecifiers = function (exports) {
            var nodes = [], first = true;
            // export { x, y as z } [from '...']
            this.expect(types.braceL);
            while (!this.eat(types.braceR)) {
                if (!first) {
                    this.expect(types.comma);
                    if (this.afterTrailingComma(types.braceR)) {
                        break;
                    }
                }
                else {
                    first = false;
                }
                var node = this.startNode();
                node.local = this.parseIdent(true);
                node.exported = this.eatContextual("as") ? this.parseIdent(true) : node.local;
                this.checkExport(exports, node.exported.name, node.exported.start);
                nodes.push(this.finishNode(node, "ExportSpecifier"));
            }
            return nodes;
        };
        // Parses import declaration.
        pp$1.parseImport = function (node) {
            this.next();
            // import '...'
            if (this.type === types.string) {
                node.specifiers = empty;
                node.source = this.parseExprAtom();
            }
            else {
                node.specifiers = this.parseImportSpecifiers();
                this.expectContextual("from");
                node.source = this.type === types.string ? this.parseExprAtom() : this.unexpected();
            }
            this.semicolon();
            return this.finishNode(node, "ImportDeclaration");
        };
        // Parses a comma-separated list of module imports.
        pp$1.parseImportSpecifiers = function () {
            var nodes = [], first = true;
            if (this.type === types.name) {
                // import defaultObj, { x, y as z } from '...'
                var node = this.startNode();
                node.local = this.parseIdent();
                this.checkLValSimple(node.local, BIND_LEXICAL);
                nodes.push(this.finishNode(node, "ImportDefaultSpecifier"));
                if (!this.eat(types.comma)) {
                    return nodes;
                }
            }
            if (this.type === types.star) {
                var node$1 = this.startNode();
                this.next();
                this.expectContextual("as");
                node$1.local = this.parseIdent();
                this.checkLValSimple(node$1.local, BIND_LEXICAL);
                nodes.push(this.finishNode(node$1, "ImportNamespaceSpecifier"));
                return nodes;
            }
            this.expect(types.braceL);
            while (!this.eat(types.braceR)) {
                if (!first) {
                    this.expect(types.comma);
                    if (this.afterTrailingComma(types.braceR)) {
                        break;
                    }
                }
                else {
                    first = false;
                }
                var node$2 = this.startNode();
                node$2.imported = this.parseIdent(true);
                if (this.eatContextual("as")) {
                    node$2.local = this.parseIdent();
                }
                else {
                    this.checkUnreserved(node$2.imported);
                    node$2.local = node$2.imported;
                }
                this.checkLValSimple(node$2.local, BIND_LEXICAL);
                nodes.push(this.finishNode(node$2, "ImportSpecifier"));
            }
            return nodes;
        };
        // Set `ExpressionStatement#directive` property for directive prologues.
        pp$1.adaptDirectivePrologue = function (statements) {
            for (var i = 0; i < statements.length && this.isDirectiveCandidate(statements[i]); ++i) {
                statements[i].directive = statements[i].expression.raw.slice(1, -1);
            }
        };
        pp$1.isDirectiveCandidate = function (statement) {
            return (statement.type === "ExpressionStatement" &&
                statement.expression.type === "Literal" &&
                typeof statement.expression.value === "string" &&
                // Reject parenthesized strings.
                (this.input[statement.start] === "\"" || this.input[statement.start] === "'"));
        };
        var pp$2 = Parser.prototype;
        // Convert existing expression atom to assignable pattern
        // if possible.
        pp$2.toAssignable = function (node, isBinding, refDestructuringErrors) {
            if (this.options.ecmaVersion >= 6 && node) {
                switch (node.type) {
                    case "Identifier":
                        if (this.inAsync && node.name === "await") {
                            this.raise(node.start, "Cannot use 'await' as identifier inside an async function");
                        }
                        break;
                    case "ObjectPattern":
                    case "ArrayPattern":
                    case "AssignmentPattern":
                    case "RestElement":
                        break;
                    case "ObjectExpression":
                        node.type = "ObjectPattern";
                        if (refDestructuringErrors) {
                            this.checkPatternErrors(refDestructuringErrors, true);
                        }
                        for (var i = 0, list = node.properties; i < list.length; i += 1) {
                            var prop = list[i];
                            this.toAssignable(prop, isBinding);
                            // Early error:
                            //   AssignmentRestProperty[Yield, Await] :
                            //     `...` DestructuringAssignmentTarget[Yield, Await]
                            //
                            //   It is a Syntax Error if |DestructuringAssignmentTarget| is an |ArrayLiteral| or an |ObjectLiteral|.
                            if (prop.type === "RestElement" &&
                                (prop.argument.type === "ArrayPattern" || prop.argument.type === "ObjectPattern")) {
                                this.raise(prop.argument.start, "Unexpected token");
                            }
                        }
                        break;
                    case "Property":
                        // AssignmentProperty has type === "Property"
                        if (node.kind !== "init") {
                            this.raise(node.key.start, "Object pattern can't contain getter or setter");
                        }
                        this.toAssignable(node.value, isBinding);
                        break;
                    case "ArrayExpression":
                        node.type = "ArrayPattern";
                        if (refDestructuringErrors) {
                            this.checkPatternErrors(refDestructuringErrors, true);
                        }
                        this.toAssignableList(node.elements, isBinding);
                        break;
                    case "SpreadElement":
                        node.type = "RestElement";
                        this.toAssignable(node.argument, isBinding);
                        if (node.argument.type === "AssignmentPattern") {
                            this.raise(node.argument.start, "Rest elements cannot have a default value");
                        }
                        break;
                    case "AssignmentExpression":
                        if (node.operator !== "=") {
                            this.raise(node.left.end, "Only '=' operator can be used for specifying default value.");
                        }
                        node.type = "AssignmentPattern";
                        delete node.operator;
                        this.toAssignable(node.left, isBinding);
                        break;
                    case "ParenthesizedExpression":
                        this.toAssignable(node.expression, isBinding, refDestructuringErrors);
                        break;
                    case "ChainExpression":
                        this.raiseRecoverable(node.start, "Optional chaining cannot appear in left-hand side");
                        break;
                    case "MemberExpression":
                        if (!isBinding) {
                            break;
                        }
                    default:
                        this.raise(node.start, "Assigning to rvalue");
                }
            }
            else if (refDestructuringErrors) {
                this.checkPatternErrors(refDestructuringErrors, true);
            }
            return node;
        };
        // Convert list of expression atoms to binding list.
        pp$2.toAssignableList = function (exprList, isBinding) {
            var end = exprList.length;
            for (var i = 0; i < end; i++) {
                var elt = exprList[i];
                if (elt) {
                    this.toAssignable(elt, isBinding);
                }
            }
            if (end) {
                var last = exprList[end - 1];
                if (this.options.ecmaVersion === 6 && isBinding && last && last.type === "RestElement" && last.argument.type !== "Identifier") {
                    this.unexpected(last.argument.start);
                }
            }
            return exprList;
        };
        // Parses spread element.
        pp$2.parseSpread = function (refDestructuringErrors) {
            var node = this.startNode();
            this.next();
            node.argument = this.parseMaybeAssign(false, refDestructuringErrors);
            return this.finishNode(node, "SpreadElement");
        };
        pp$2.parseRestBinding = function () {
            var node = this.startNode();
            this.next();
            // RestElement inside of a function parameter must be an identifier
            if (this.options.ecmaVersion === 6 && this.type !== types.name) {
                this.unexpected();
            }
            node.argument = this.parseBindingAtom();
            return this.finishNode(node, "RestElement");
        };
        // Parses lvalue (assignable) atom.
        pp$2.parseBindingAtom = function () {
            if (this.options.ecmaVersion >= 6) {
                switch (this.type) {
                    case types.bracketL:
                        var node = this.startNode();
                        this.next();
                        node.elements = this.parseBindingList(types.bracketR, true, true);
                        return this.finishNode(node, "ArrayPattern");
                    case types.braceL:
                        return this.parseObj(true);
                }
            }
            return this.parseIdent();
        };
        pp$2.parseBindingList = function (close, allowEmpty, allowTrailingComma) {
            var elts = [], first = true;
            while (!this.eat(close)) {
                if (first) {
                    first = false;
                }
                else {
                    this.expect(types.comma);
                }
                if (allowEmpty && this.type === types.comma) {
                    elts.push(null);
                }
                else if (allowTrailingComma && this.afterTrailingComma(close)) {
                    break;
                }
                else if (this.type === types.ellipsis) {
                    var rest = this.parseRestBinding();
                    this.parseBindingListItem(rest);
                    elts.push(rest);
                    if (this.type === types.comma) {
                        this.raise(this.start, "Comma is not permitted after the rest element");
                    }
                    this.expect(close);
                    break;
                }
                else {
                    var elem = this.parseMaybeDefault(this.start, this.startLoc);
                    this.parseBindingListItem(elem);
                    elts.push(elem);
                }
            }
            return elts;
        };
        pp$2.parseBindingListItem = function (param) {
            return param;
        };
        // Parses assignment pattern around given atom if possible.
        pp$2.parseMaybeDefault = function (startPos, startLoc, left) {
            left = left || this.parseBindingAtom();
            if (this.options.ecmaVersion < 6 || !this.eat(types.eq)) {
                return left;
            }
            var node = this.startNodeAt(startPos, startLoc);
            node.left = left;
            node.right = this.parseMaybeAssign();
            return this.finishNode(node, "AssignmentPattern");
        };
        // The following three functions all verify that a node is an lvalue —
        // something that can be bound, or assigned to. In order to do so, they perform
        // a variety of checks:
        //
        // - Check that none of the bound/assigned-to identifiers are reserved words.
        // - Record name declarations for bindings in the appropriate scope.
        // - Check duplicate argument names, if checkClashes is set.
        //
        // If a complex binding pattern is encountered (e.g., object and array
        // destructuring), the entire pattern is recursively checked.
        //
        // There are three versions of checkLVal*() appropriate for different
        // circumstances:
        //
        // - checkLValSimple() shall be used if the syntactic construct supports
        //   nothing other than identifiers and member expressions. Parenthesized
        //   expressions are also correctly handled. This is generally appropriate for
        //   constructs for which the spec says
        //
        //   > It is a Syntax Error if AssignmentTargetType of [the production] is not
        //   > simple.
        //
        //   It is also appropriate for checking if an identifier is valid and not
        //   defined elsewhere, like import declarations or function/class identifiers.
        //
        //   Examples where this is used include:
        //     a += …;
        //     import a from '…';
        //   where a is the node to be checked.
        //
        // - checkLValPattern() shall be used if the syntactic construct supports
        //   anything checkLValSimple() supports, as well as object and array
        //   destructuring patterns. This is generally appropriate for constructs for
        //   which the spec says
        //
        //   > It is a Syntax Error if [the production] is neither an ObjectLiteral nor
        //   > an ArrayLiteral and AssignmentTargetType of [the production] is not
        //   > simple.
        //
        //   Examples where this is used include:
        //     (a = …);
        //     const a = …;
        //     try { … } catch (a) { … }
        //   where a is the node to be checked.
        //
        // - checkLValInnerPattern() shall be used if the syntactic construct supports
        //   anything checkLValPattern() supports, as well as default assignment
        //   patterns, rest elements, and other constructs that may appear within an
        //   object or array destructuring pattern.
        //
        //   As a special case, function parameters also use checkLValInnerPattern(),
        //   as they also support defaults and rest constructs.
        //
        // These functions deliberately support both assignment and binding constructs,
        // as the logic for both is exceedingly similar. If the node is the target of
        // an assignment, then bindingType should be set to BIND_NONE. Otherwise, it
        // should be set to the appropriate BIND_* constant, like BIND_VAR or
        // BIND_LEXICAL.
        //
        // If the function is called with a non-BIND_NONE bindingType, then
        // additionally a checkClashes object may be specified to allow checking for
        // duplicate argument names. checkClashes is ignored if the provided construct
        // is an assignment (i.e., bindingType is BIND_NONE).
        pp$2.checkLValSimple = function (expr, bindingType, checkClashes) {
            if (bindingType === void 0)
                bindingType = BIND_NONE;
            var isBind = bindingType !== BIND_NONE;
            switch (expr.type) {
                case "Identifier":
                    if (this.strict && this.reservedWordsStrictBind.test(expr.name)) {
                        this.raiseRecoverable(expr.start, (isBind ? "Binding " : "Assigning to ") + expr.name + " in strict mode");
                    }
                    if (isBind) {
                        if (bindingType === BIND_LEXICAL && expr.name === "let") {
                            this.raiseRecoverable(expr.start, "let is disallowed as a lexically bound name");
                        }
                        if (checkClashes) {
                            if (has(checkClashes, expr.name)) {
                                this.raiseRecoverable(expr.start, "Argument name clash");
                            }
                            checkClashes[expr.name] = true;
                        }
                        if (bindingType !== BIND_OUTSIDE) {
                            this.declareName(expr.name, bindingType, expr.start);
                        }
                    }
                    break;
                case "ChainExpression":
                    this.raiseRecoverable(expr.start, "Optional chaining cannot appear in left-hand side");
                    break;
                case "MemberExpression":
                    if (isBind) {
                        this.raiseRecoverable(expr.start, "Binding member expression");
                    }
                    break;
                case "ParenthesizedExpression":
                    if (isBind) {
                        this.raiseRecoverable(expr.start, "Binding parenthesized expression");
                    }
                    return this.checkLValSimple(expr.expression, bindingType, checkClashes);
                default:
                    this.raise(expr.start, (isBind ? "Binding" : "Assigning to") + " rvalue");
            }
        };
        pp$2.checkLValPattern = function (expr, bindingType, checkClashes) {
            if (bindingType === void 0)
                bindingType = BIND_NONE;
            switch (expr.type) {
                case "ObjectPattern":
                    for (var i = 0, list = expr.properties; i < list.length; i += 1) {
                        var prop = list[i];
                        this.checkLValInnerPattern(prop, bindingType, checkClashes);
                    }
                    break;
                case "ArrayPattern":
                    for (var i$1 = 0, list$1 = expr.elements; i$1 < list$1.length; i$1 += 1) {
                        var elem = list$1[i$1];
                        if (elem) {
                            this.checkLValInnerPattern(elem, bindingType, checkClashes);
                        }
                    }
                    break;
                default:
                    this.checkLValSimple(expr, bindingType, checkClashes);
            }
        };
        pp$2.checkLValInnerPattern = function (expr, bindingType, checkClashes) {
            if (bindingType === void 0)
                bindingType = BIND_NONE;
            switch (expr.type) {
                case "Property":
                    // AssignmentProperty has type === "Property"
                    this.checkLValInnerPattern(expr.value, bindingType, checkClashes);
                    break;
                case "AssignmentPattern":
                    this.checkLValPattern(expr.left, bindingType, checkClashes);
                    break;
                case "RestElement":
                    this.checkLValPattern(expr.argument, bindingType, checkClashes);
                    break;
                default:
                    this.checkLValPattern(expr, bindingType, checkClashes);
            }
        };
        // A recursive descent parser operates by defining functions for all
        var pp$3 = Parser.prototype;
        // Check if property name clashes with already added.
        // Object/class getters and setters are not allowed to clash —
        // either with each other or with an init property — and in
        // strict mode, init properties are also not allowed to be repeated.
        pp$3.checkPropClash = function (prop, propHash, refDestructuringErrors) {
            if (this.options.ecmaVersion >= 9 && prop.type === "SpreadElement") {
                return;
            }
            if (this.options.ecmaVersion >= 6 && (prop.computed || prop.method || prop.shorthand)) {
                return;
            }
            var key = prop.key;
            var name;
            switch (key.type) {
                case "Identifier":
                    name = key.name;
                    break;
                case "Literal":
                    name = String(key.value);
                    break;
                default: return;
            }
            var kind = prop.kind;
            if (this.options.ecmaVersion >= 6) {
                if (name === "__proto__" && kind === "init") {
                    if (propHash.proto) {
                        if (refDestructuringErrors) {
                            if (refDestructuringErrors.doubleProto < 0) {
                                refDestructuringErrors.doubleProto = key.start;
                            }
                            // Backwards-compat kludge. Can be removed in version 6.0
                        }
                        else {
                            this.raiseRecoverable(key.start, "Redefinition of __proto__ property");
                        }
                    }
                    propHash.proto = true;
                }
                return;
            }
            name = "$" + name;
            var other = propHash[name];
            if (other) {
                var redefinition;
                if (kind === "init") {
                    redefinition = this.strict && other.init || other.get || other.set;
                }
                else {
                    redefinition = other.init || other[kind];
                }
                if (redefinition) {
                    this.raiseRecoverable(key.start, "Redefinition of property");
                }
            }
            else {
                other = propHash[name] = {
                    init: false,
                    get: false,
                    set: false
                };
            }
            other[kind] = true;
        };
        // ### Expression parsing
        // These nest, from the most general expression type at the top to
        // 'atomic', nondivisible expression types at the bottom. Most of
        // the functions will simply let the function(s) below them parse,
        // and, *if* the syntactic construct they handle is present, wrap
        // the AST node that the inner parser gave them in another node.
        // Parse a full expression. The optional arguments are used to
        // forbid the `in` operator (in for loops initalization expressions)
        // and provide reference for storing '=' operator inside shorthand
        // property assignment in contexts where both object expression
        // and object pattern might appear (so it's possible to raise
        // delayed syntax error at correct position).
        pp$3.parseExpression = function (forInit, refDestructuringErrors) {
            var startPos = this.start, startLoc = this.startLoc;
            var expr = this.parseMaybeAssign(forInit, refDestructuringErrors);
            if (this.type === types.comma) {
                var node = this.startNodeAt(startPos, startLoc);
                node.expressions = [expr];
                while (this.eat(types.comma)) {
                    node.expressions.push(this.parseMaybeAssign(forInit, refDestructuringErrors));
                }
                return this.finishNode(node, "SequenceExpression");
            }
            return expr;
        };
        // Parse an assignment expression. This includes applications of
        // operators like `+=`.
        pp$3.parseMaybeAssign = function (forInit, refDestructuringErrors, afterLeftParse) {
            if (this.isContextual("yield")) {
                if (this.inGenerator) {
                    return this.parseYield(forInit);
                }
                // The tokenizer will assume an expression is allowed after
                // `yield`, but this isn't that kind of yield
                else {
                    this.exprAllowed = false;
                }
            }
            var ownDestructuringErrors = false, oldParenAssign = -1, oldTrailingComma = -1;
            if (refDestructuringErrors) {
                oldParenAssign = refDestructuringErrors.parenthesizedAssign;
                oldTrailingComma = refDestructuringErrors.trailingComma;
                refDestructuringErrors.parenthesizedAssign = refDestructuringErrors.trailingComma = -1;
            }
            else {
                refDestructuringErrors = new DestructuringErrors;
                ownDestructuringErrors = true;
            }
            var startPos = this.start, startLoc = this.startLoc;
            if (this.type === types.parenL || this.type === types.name) {
                this.potentialArrowAt = this.start;
                this.potentialArrowInForAwait = forInit === "await";
            }
            var left = this.parseMaybeConditional(forInit, refDestructuringErrors);
            if (afterLeftParse) {
                left = afterLeftParse.call(this, left, startPos, startLoc);
            }
            if (this.type.isAssign) {
                var node = this.startNodeAt(startPos, startLoc);
                node.operator = this.value;
                if (this.type === types.eq) {
                    left = this.toAssignable(left, false, refDestructuringErrors);
                }
                if (!ownDestructuringErrors) {
                    refDestructuringErrors.parenthesizedAssign = refDestructuringErrors.trailingComma = refDestructuringErrors.doubleProto = -1;
                }
                if (refDestructuringErrors.shorthandAssign >= left.start) {
                    refDestructuringErrors.shorthandAssign = -1;
                } // reset because shorthand default was used correctly
                if (this.type === types.eq) {
                    this.checkLValPattern(left);
                }
                else {
                    this.checkLValSimple(left);
                }
                node.left = left;
                this.next();
                node.right = this.parseMaybeAssign(forInit);
                return this.finishNode(node, "AssignmentExpression");
            }
            else {
                if (ownDestructuringErrors) {
                    this.checkExpressionErrors(refDestructuringErrors, true);
                }
            }
            if (oldParenAssign > -1) {
                refDestructuringErrors.parenthesizedAssign = oldParenAssign;
            }
            if (oldTrailingComma > -1) {
                refDestructuringErrors.trailingComma = oldTrailingComma;
            }
            return left;
        };
        // Parse a ternary conditional (`?:`) operator.
        pp$3.parseMaybeConditional = function (forInit, refDestructuringErrors) {
            var startPos = this.start, startLoc = this.startLoc;
            var expr = this.parseExprOps(forInit, refDestructuringErrors);
            if (this.checkExpressionErrors(refDestructuringErrors)) {
                return expr;
            }
            if (this.eat(types.question)) {
                var node = this.startNodeAt(startPos, startLoc);
                node.test = expr;
                node.consequent = this.parseMaybeAssign();
                this.expect(types.colon);
                node.alternate = this.parseMaybeAssign(forInit);
                return this.finishNode(node, "ConditionalExpression");
            }
            return expr;
        };
        // Start the precedence parser.
        pp$3.parseExprOps = function (forInit, refDestructuringErrors) {
            var startPos = this.start, startLoc = this.startLoc;
            var expr = this.parseMaybeUnary(refDestructuringErrors, false);
            if (this.checkExpressionErrors(refDestructuringErrors)) {
                return expr;
            }
            return expr.start === startPos && expr.type === "ArrowFunctionExpression" ? expr : this.parseExprOp(expr, startPos, startLoc, -1, forInit);
        };
        // Parse binary operators with the operator precedence parsing
        // algorithm. `left` is the left-hand side of the operator.
        // `minPrec` provides context that allows the function to stop and
        // defer further parser to one of its callers when it encounters an
        // operator that has a lower precedence than the set it is parsing.
        pp$3.parseExprOp = function (left, leftStartPos, leftStartLoc, minPrec, forInit) {
            var prec = this.type.binop;
            if (prec != null && (!forInit || this.type !== types._in)) {
                if (prec > minPrec) {
                    var logical = this.type === types.logicalOR || this.type === types.logicalAND;
                    var coalesce = this.type === types.coalesce;
                    if (coalesce) {
                        // Handle the precedence of `tt.coalesce` as equal to the range of logical expressions.
                        // In other words, `node.right` shouldn't contain logical expressions in order to check the mixed error.
                        prec = types.logicalAND.binop;
                    }
                    var op = this.value;
                    this.next();
                    var startPos = this.start, startLoc = this.startLoc;
                    var right = this.parseExprOp(this.parseMaybeUnary(null, false), startPos, startLoc, prec, forInit);
                    var node = this.buildBinary(leftStartPos, leftStartLoc, left, right, op, logical || coalesce);
                    if ((logical && this.type === types.coalesce) || (coalesce && (this.type === types.logicalOR || this.type === types.logicalAND))) {
                        this.raiseRecoverable(this.start, "Logical expressions and coalesce expressions cannot be mixed. Wrap either by parentheses");
                    }
                    return this.parseExprOp(node, leftStartPos, leftStartLoc, minPrec, forInit);
                }
            }
            return left;
        };
        pp$3.buildBinary = function (startPos, startLoc, left, right, op, logical) {
            var node = this.startNodeAt(startPos, startLoc);
            node.left = left;
            node.operator = op;
            node.right = right;
            return this.finishNode(node, logical ? "LogicalExpression" : "BinaryExpression");
        };
        // Parse unary operators, both prefix and postfix.
        pp$3.parseMaybeUnary = function (refDestructuringErrors, sawUnary, incDec) {
            var startPos = this.start, startLoc = this.startLoc, expr;
            if (this.isContextual("await") && this.canAwait) {
                expr = this.parseAwait();
                sawUnary = true;
            }
            else if (this.type.prefix) {
                var node = this.startNode(), update = this.type === types.incDec;
                node.operator = this.value;
                node.prefix = true;
                this.next();
                node.argument = this.parseMaybeUnary(null, true, update);
                this.checkExpressionErrors(refDestructuringErrors, true);
                if (update) {
                    this.checkLValSimple(node.argument);
                }
                else if (this.strict && node.operator === "delete" &&
                    node.argument.type === "Identifier") {
                    this.raiseRecoverable(node.start, "Deleting local variable in strict mode");
                }
                else if (node.operator === "delete" && isPrivateFieldAccess(node.argument)) {
                    this.raiseRecoverable(node.start, "Private fields can not be deleted");
                }
                else {
                    sawUnary = true;
                }
                expr = this.finishNode(node, update ? "UpdateExpression" : "UnaryExpression");
            }
            else {
                expr = this.parseExprSubscripts(refDestructuringErrors);
                if (this.checkExpressionErrors(refDestructuringErrors)) {
                    return expr;
                }
                while (this.type.postfix && !this.canInsertSemicolon()) {
                    var node$1 = this.startNodeAt(startPos, startLoc);
                    node$1.operator = this.value;
                    node$1.prefix = false;
                    node$1.argument = expr;
                    this.checkLValSimple(expr);
                    this.next();
                    expr = this.finishNode(node$1, "UpdateExpression");
                }
            }
            if (!incDec && this.eat(types.starstar)) {
                if (sawUnary) {
                    this.unexpected(this.lastTokStart);
                }
                else {
                    return this.buildBinary(startPos, startLoc, expr, this.parseMaybeUnary(null, false), "**", false);
                }
            }
            else {
                return expr;
            }
        };
        function isPrivateFieldAccess(node) {
            return (node.type === "MemberExpression" && node.property.type === "PrivateIdentifier" ||
                node.type === "ChainExpression" && isPrivateFieldAccess(node.expression));
        }
        // Parse call, dot, and `[]`-subscript expressions.
        pp$3.parseExprSubscripts = function (refDestructuringErrors) {
            var startPos = this.start, startLoc = this.startLoc;
            var expr = this.parseExprAtom(refDestructuringErrors);
            if (expr.type === "ArrowFunctionExpression" && this.input.slice(this.lastTokStart, this.lastTokEnd) !== ")") {
                return expr;
            }
            var result = this.parseSubscripts(expr, startPos, startLoc);
            if (refDestructuringErrors && result.type === "MemberExpression") {
                if (refDestructuringErrors.parenthesizedAssign >= result.start) {
                    refDestructuringErrors.parenthesizedAssign = -1;
                }
                if (refDestructuringErrors.parenthesizedBind >= result.start) {
                    refDestructuringErrors.parenthesizedBind = -1;
                }
                if (refDestructuringErrors.trailingComma >= result.start) {
                    refDestructuringErrors.trailingComma = -1;
                }
            }
            return result;
        };
        pp$3.parseSubscripts = function (base, startPos, startLoc, noCalls) {
            var maybeAsyncArrow = this.options.ecmaVersion >= 8 && base.type === "Identifier" && base.name === "async" &&
                this.lastTokEnd === base.end && !this.canInsertSemicolon() && base.end - base.start === 5 &&
                this.potentialArrowAt === base.start;
            var optionalChained = false;
            while (true) {
                var element = this.parseSubscript(base, startPos, startLoc, noCalls, maybeAsyncArrow, optionalChained);
                if (element.optional) {
                    optionalChained = true;
                }
                if (element === base || element.type === "ArrowFunctionExpression") {
                    if (optionalChained) {
                        var chainNode = this.startNodeAt(startPos, startLoc);
                        chainNode.expression = element;
                        element = this.finishNode(chainNode, "ChainExpression");
                    }
                    return element;
                }
                base = element;
            }
        };
        pp$3.parseSubscript = function (base, startPos, startLoc, noCalls, maybeAsyncArrow, optionalChained) {
            var optionalSupported = this.options.ecmaVersion >= 11;
            var optional = optionalSupported && this.eat(types.questionDot);
            if (noCalls && optional) {
                this.raise(this.lastTokStart, "Optional chaining cannot appear in the callee of new expressions");
            }
            var computed = this.eat(types.bracketL);
            if (computed || (optional && this.type !== types.parenL && this.type !== types.backQuote) || this.eat(types.dot)) {
                var node = this.startNodeAt(startPos, startLoc);
                node.object = base;
                if (computed) {
                    node.property = this.parseExpression();
                    this.expect(types.bracketR);
                }
                else if (this.type === types.privateId && base.type !== "Super") {
                    node.property = this.parsePrivateIdent();
                }
                else {
                    node.property = this.parseIdent(this.options.allowReserved !== "never");
                }
                node.computed = !!computed;
                if (optionalSupported) {
                    node.optional = optional;
                }
                base = this.finishNode(node, "MemberExpression");
            }
            else if (!noCalls && this.eat(types.parenL)) {
                var refDestructuringErrors = new DestructuringErrors, oldYieldPos = this.yieldPos, oldAwaitPos = this.awaitPos, oldAwaitIdentPos = this.awaitIdentPos;
                this.yieldPos = 0;
                this.awaitPos = 0;
                this.awaitIdentPos = 0;
                var exprList = this.parseExprList(types.parenR, this.options.ecmaVersion >= 8, false, refDestructuringErrors);
                if (maybeAsyncArrow && !optional && !this.canInsertSemicolon() && this.eat(types.arrow)) {
                    this.checkPatternErrors(refDestructuringErrors, false);
                    this.checkYieldAwaitInDefaultParams();
                    if (this.awaitIdentPos > 0) {
                        this.raise(this.awaitIdentPos, "Cannot use 'await' as identifier inside an async function");
                    }
                    this.yieldPos = oldYieldPos;
                    this.awaitPos = oldAwaitPos;
                    this.awaitIdentPos = oldAwaitIdentPos;
                    return this.parseArrowExpression(this.startNodeAt(startPos, startLoc), exprList, true);
                }
                this.checkExpressionErrors(refDestructuringErrors, true);
                this.yieldPos = oldYieldPos || this.yieldPos;
                this.awaitPos = oldAwaitPos || this.awaitPos;
                this.awaitIdentPos = oldAwaitIdentPos || this.awaitIdentPos;
                var node$1 = this.startNodeAt(startPos, startLoc);
                node$1.callee = base;
                node$1.arguments = exprList;
                if (optionalSupported) {
                    node$1.optional = optional;
                }
                base = this.finishNode(node$1, "CallExpression");
            }
            else if (this.type === types.backQuote) {
                if (optional || optionalChained) {
                    this.raise(this.start, "Optional chaining cannot appear in the tag of tagged template expressions");
                }
                var node$2 = this.startNodeAt(startPos, startLoc);
                node$2.tag = base;
                node$2.quasi = this.parseTemplate({ isTagged: true });
                base = this.finishNode(node$2, "TaggedTemplateExpression");
            }
            return base;
        };
        // Parse an atomic expression — either a single token that is an
        // expression, an expression started by a keyword like `function` or
        // `new`, or an expression wrapped in punctuation like `()`, `[]`,
        // or `{}`.
        pp$3.parseExprAtom = function (refDestructuringErrors) {
            // If a division operator appears in an expression position, the
            // tokenizer got confused, and we force it to read a regexp instead.
            if (this.type === types.slash) {
                this.readRegexp();
            }
            var node, canBeArrow = this.potentialArrowAt === this.start;
            switch (this.type) {
                case types._super:
                    if (!this.allowSuper) {
                        this.raise(this.start, "'super' keyword outside a method");
                    }
                    node = this.startNode();
                    this.next();
                    if (this.type === types.parenL && !this.allowDirectSuper) {
                        this.raise(node.start, "super() call outside constructor of a subclass");
                    }
                    // The `super` keyword can appear at below:
                    // SuperProperty:
                    //     super [ Expression ]
                    //     super . IdentifierName
                    // SuperCall:
                    //     super ( Arguments )
                    if (this.type !== types.dot && this.type !== types.bracketL && this.type !== types.parenL) {
                        this.unexpected();
                    }
                    return this.finishNode(node, "Super");
                case types._this:
                    node = this.startNode();
                    this.next();
                    return this.finishNode(node, "ThisExpression");
                case types.name:
                    var startPos = this.start, startLoc = this.startLoc, containsEsc = this.containsEsc;
                    var id = this.parseIdent(false);
                    if (this.options.ecmaVersion >= 8 && !containsEsc && id.name === "async" && !this.canInsertSemicolon() && this.eat(types._function)) {
                        return this.parseFunction(this.startNodeAt(startPos, startLoc), 0, false, true);
                    }
                    if (canBeArrow && !this.canInsertSemicolon()) {
                        if (this.eat(types.arrow)) {
                            return this.parseArrowExpression(this.startNodeAt(startPos, startLoc), [id], false);
                        }
                        if (this.options.ecmaVersion >= 8 && id.name === "async" && this.type === types.name && !containsEsc &&
                            (!this.potentialArrowInForAwait || this.value !== "of" || this.containsEsc)) {
                            id = this.parseIdent(false);
                            if (this.canInsertSemicolon() || !this.eat(types.arrow)) {
                                this.unexpected();
                            }
                            return this.parseArrowExpression(this.startNodeAt(startPos, startLoc), [id], true);
                        }
                    }
                    return id;
                case types.regexp:
                    var value = this.value;
                    node = this.parseLiteral(value.value);
                    node.regex = { pattern: value.pattern, flags: value.flags };
                    return node;
                case types.num:
                case types.string:
                    return this.parseLiteral(this.value);
                case types._null:
                case types._true:
                case types._false:
                    node = this.startNode();
                    node.value = this.type === types._null ? null : this.type === types._true;
                    node.raw = this.type.keyword;
                    this.next();
                    return this.finishNode(node, "Literal");
                case types.parenL:
                    var start = this.start, expr = this.parseParenAndDistinguishExpression(canBeArrow);
                    if (refDestructuringErrors) {
                        if (refDestructuringErrors.parenthesizedAssign < 0 && !this.isSimpleAssignTarget(expr)) {
                            refDestructuringErrors.parenthesizedAssign = start;
                        }
                        if (refDestructuringErrors.parenthesizedBind < 0) {
                            refDestructuringErrors.parenthesizedBind = start;
                        }
                    }
                    return expr;
                case types.bracketL:
                    node = this.startNode();
                    this.next();
                    node.elements = this.parseExprList(types.bracketR, true, true, refDestructuringErrors);
                    return this.finishNode(node, "ArrayExpression");
                case types.braceL:
                    return this.parseObj(false, refDestructuringErrors);
                case types._function:
                    node = this.startNode();
                    this.next();
                    return this.parseFunction(node, 0);
                case types._class:
                    return this.parseClass(this.startNode(), false);
                case types._new:
                    return this.parseNew();
                case types.backQuote:
                    return this.parseTemplate();
                case types._import:
                    if (this.options.ecmaVersion >= 11) {
                        return this.parseExprImport();
                    }
                    else {
                        return this.unexpected();
                    }
                default:
                    this.unexpected();
            }
        };
        pp$3.parseExprImport = function () {
            var node = this.startNode();
            // Consume `import` as an identifier for `import.meta`.
            // Because `this.parseIdent(true)` doesn't check escape sequences, it needs the check of `this.containsEsc`.
            if (this.containsEsc) {
                this.raiseRecoverable(this.start, "Escape sequence in keyword import");
            }
            var meta = this.parseIdent(true);
            switch (this.type) {
                case types.parenL:
                    return this.parseDynamicImport(node);
                case types.dot:
                    node.meta = meta;
                    return this.parseImportMeta(node);
                default:
                    this.unexpected();
            }
        };
        pp$3.parseDynamicImport = function (node) {
            this.next(); // skip `(`
            // Parse node.source.
            node.source = this.parseMaybeAssign();
            // Verify ending.
            if (!this.eat(types.parenR)) {
                var errorPos = this.start;
                if (this.eat(types.comma) && this.eat(types.parenR)) {
                    this.raiseRecoverable(errorPos, "Trailing comma is not allowed in import()");
                }
                else {
                    this.unexpected(errorPos);
                }
            }
            return this.finishNode(node, "ImportExpression");
        };
        pp$3.parseImportMeta = function (node) {
            this.next(); // skip `.`
            var containsEsc = this.containsEsc;
            node.property = this.parseIdent(true);
            if (node.property.name !== "meta") {
                this.raiseRecoverable(node.property.start, "The only valid meta property for import is 'import.meta'");
            }
            if (containsEsc) {
                this.raiseRecoverable(node.start, "'import.meta' must not contain escaped characters");
            }
            if (this.options.sourceType !== "module" && !this.options.allowImportExportEverywhere) {
                this.raiseRecoverable(node.start, "Cannot use 'import.meta' outside a module");
            }
            return this.finishNode(node, "MetaProperty");
        };
        pp$3.parseLiteral = function (value) {
            var node = this.startNode();
            node.value = value;
            node.raw = this.input.slice(this.start, this.end);
            if (node.raw.charCodeAt(node.raw.length - 1) === 110) {
                node.bigint = node.raw.slice(0, -1).replace(/_/g, "");
            }
            this.next();
            return this.finishNode(node, "Literal");
        };
        pp$3.parseParenExpression = function () {
            this.expect(types.parenL);
            var val = this.parseExpression();
            this.expect(types.parenR);
            return val;
        };
        pp$3.parseParenAndDistinguishExpression = function (canBeArrow) {
            var startPos = this.start, startLoc = this.startLoc, val, allowTrailingComma = this.options.ecmaVersion >= 8;
            if (this.options.ecmaVersion >= 6) {
                this.next();
                var innerStartPos = this.start, innerStartLoc = this.startLoc;
                var exprList = [], first = true, lastIsComma = false;
                var refDestructuringErrors = new DestructuringErrors, oldYieldPos = this.yieldPos, oldAwaitPos = this.awaitPos, spreadStart;
                this.yieldPos = 0;
                this.awaitPos = 0;
                // Do not save awaitIdentPos to allow checking awaits nested in parameters
                while (this.type !== types.parenR) {
                    first ? first = false : this.expect(types.comma);
                    if (allowTrailingComma && this.afterTrailingComma(types.parenR, true)) {
                        lastIsComma = true;
                        break;
                    }
                    else if (this.type === types.ellipsis) {
                        spreadStart = this.start;
                        exprList.push(this.parseParenItem(this.parseRestBinding()));
                        if (this.type === types.comma) {
                            this.raise(this.start, "Comma is not permitted after the rest element");
                        }
                        break;
                    }
                    else {
                        exprList.push(this.parseMaybeAssign(false, refDestructuringErrors, this.parseParenItem));
                    }
                }
                var innerEndPos = this.start, innerEndLoc = this.startLoc;
                this.expect(types.parenR);
                if (canBeArrow && !this.canInsertSemicolon() && this.eat(types.arrow)) {
                    this.checkPatternErrors(refDestructuringErrors, false);
                    this.checkYieldAwaitInDefaultParams();
                    this.yieldPos = oldYieldPos;
                    this.awaitPos = oldAwaitPos;
                    return this.parseParenArrowList(startPos, startLoc, exprList);
                }
                if (!exprList.length || lastIsComma) {
                    this.unexpected(this.lastTokStart);
                }
                if (spreadStart) {
                    this.unexpected(spreadStart);
                }
                this.checkExpressionErrors(refDestructuringErrors, true);
                this.yieldPos = oldYieldPos || this.yieldPos;
                this.awaitPos = oldAwaitPos || this.awaitPos;
                if (exprList.length > 1) {
                    val = this.startNodeAt(innerStartPos, innerStartLoc);
                    val.expressions = exprList;
                    this.finishNodeAt(val, "SequenceExpression", innerEndPos, innerEndLoc);
                }
                else {
                    val = exprList[0];
                }
            }
            else {
                val = this.parseParenExpression();
            }
            if (this.options.preserveParens) {
                var par = this.startNodeAt(startPos, startLoc);
                par.expression = val;
                return this.finishNode(par, "ParenthesizedExpression");
            }
            else {
                return val;
            }
        };
        pp$3.parseParenItem = function (item) {
            return item;
        };
        pp$3.parseParenArrowList = function (startPos, startLoc, exprList) {
            return this.parseArrowExpression(this.startNodeAt(startPos, startLoc), exprList);
        };
        // New's precedence is slightly tricky. It must allow its argument to
        // be a `[]` or dot subscript expression, but not a call — at least,
        // not without wrapping it in parentheses. Thus, it uses the noCalls
        // argument to parseSubscripts to prevent it from consuming the
        // argument list.
        var empty$1 = [];
        pp$3.parseNew = function () {
            if (this.containsEsc) {
                this.raiseRecoverable(this.start, "Escape sequence in keyword new");
            }
            var node = this.startNode();
            var meta = this.parseIdent(true);
            if (this.options.ecmaVersion >= 6 && this.eat(types.dot)) {
                node.meta = meta;
                var containsEsc = this.containsEsc;
                node.property = this.parseIdent(true);
                if (node.property.name !== "target") {
                    this.raiseRecoverable(node.property.start, "The only valid meta property for new is 'new.target'");
                }
                if (containsEsc) {
                    this.raiseRecoverable(node.start, "'new.target' must not contain escaped characters");
                }
                if (!this.inNonArrowFunction) {
                    this.raiseRecoverable(node.start, "'new.target' can only be used in functions");
                }
                return this.finishNode(node, "MetaProperty");
            }
            var startPos = this.start, startLoc = this.startLoc, isImport = this.type === types._import;
            node.callee = this.parseSubscripts(this.parseExprAtom(), startPos, startLoc, true);
            if (isImport && node.callee.type === "ImportExpression") {
                this.raise(startPos, "Cannot use new with import()");
            }
            if (this.eat(types.parenL)) {
                node.arguments = this.parseExprList(types.parenR, this.options.ecmaVersion >= 8, false);
            }
            else {
                node.arguments = empty$1;
            }
            return this.finishNode(node, "NewExpression");
        };
        // Parse template expression.
        pp$3.parseTemplateElement = function (ref) {
            var isTagged = ref.isTagged;
            var elem = this.startNode();
            if (this.type === types.invalidTemplate) {
                if (!isTagged) {
                    this.raiseRecoverable(this.start, "Bad escape sequence in untagged template literal");
                }
                elem.value = {
                    raw: this.value,
                    cooked: null
                };
            }
            else {
                elem.value = {
                    raw: this.input.slice(this.start, this.end).replace(/\r\n?/g, "\n"),
                    cooked: this.value
                };
            }
            this.next();
            elem.tail = this.type === types.backQuote;
            return this.finishNode(elem, "TemplateElement");
        };
        pp$3.parseTemplate = function (ref) {
            if (ref === void 0)
                ref = {};
            var isTagged = ref.isTagged;
            if (isTagged === void 0)
                isTagged = false;
            var node = this.startNode();
            this.next();
            node.expressions = [];
            var curElt = this.parseTemplateElement({ isTagged: isTagged });
            node.quasis = [curElt];
            while (!curElt.tail) {
                if (this.type === types.eof) {
                    this.raise(this.pos, "Unterminated template literal");
                }
                this.expect(types.dollarBraceL);
                node.expressions.push(this.parseExpression());
                this.expect(types.braceR);
                node.quasis.push(curElt = this.parseTemplateElement({ isTagged: isTagged }));
            }
            this.next();
            return this.finishNode(node, "TemplateLiteral");
        };
        pp$3.isAsyncProp = function (prop) {
            return !prop.computed && prop.key.type === "Identifier" && prop.key.name === "async" &&
                (this.type === types.name || this.type === types.num || this.type === types.string || this.type === types.bracketL || this.type.keyword || (this.options.ecmaVersion >= 9 && this.type === types.star)) &&
                !lineBreak.test(this.input.slice(this.lastTokEnd, this.start));
        };
        // Parse an object literal or binding pattern.
        pp$3.parseObj = function (isPattern, refDestructuringErrors) {
            var node = this.startNode(), first = true, propHash = {};
            node.properties = [];
            this.next();
            while (!this.eat(types.braceR)) {
                if (!first) {
                    this.expect(types.comma);
                    if (this.options.ecmaVersion >= 5 && this.afterTrailingComma(types.braceR)) {
                        break;
                    }
                }
                else {
                    first = false;
                }
                var prop = this.parseProperty(isPattern, refDestructuringErrors);
                if (!isPattern) {
                    this.checkPropClash(prop, propHash, refDestructuringErrors);
                }
                node.properties.push(prop);
            }
            return this.finishNode(node, isPattern ? "ObjectPattern" : "ObjectExpression");
        };
        pp$3.parseProperty = function (isPattern, refDestructuringErrors) {
            var prop = this.startNode(), isGenerator, isAsync, startPos, startLoc;
            if (this.options.ecmaVersion >= 9 && this.eat(types.ellipsis)) {
                if (isPattern) {
                    prop.argument = this.parseIdent(false);
                    if (this.type === types.comma) {
                        this.raise(this.start, "Comma is not permitted after the rest element");
                    }
                    return this.finishNode(prop, "RestElement");
                }
                // To disallow parenthesized identifier via `this.toAssignable()`.
                if (this.type === types.parenL && refDestructuringErrors) {
                    if (refDestructuringErrors.parenthesizedAssign < 0) {
                        refDestructuringErrors.parenthesizedAssign = this.start;
                    }
                    if (refDestructuringErrors.parenthesizedBind < 0) {
                        refDestructuringErrors.parenthesizedBind = this.start;
                    }
                }
                // Parse argument.
                prop.argument = this.parseMaybeAssign(false, refDestructuringErrors);
                // To disallow trailing comma via `this.toAssignable()`.
                if (this.type === types.comma && refDestructuringErrors && refDestructuringErrors.trailingComma < 0) {
                    refDestructuringErrors.trailingComma = this.start;
                }
                // Finish
                return this.finishNode(prop, "SpreadElement");
            }
            if (this.options.ecmaVersion >= 6) {
                prop.method = false;
                prop.shorthand = false;
                if (isPattern || refDestructuringErrors) {
                    startPos = this.start;
                    startLoc = this.startLoc;
                }
                if (!isPattern) {
                    isGenerator = this.eat(types.star);
                }
            }
            var containsEsc = this.containsEsc;
            this.parsePropertyName(prop);
            if (!isPattern && !containsEsc && this.options.ecmaVersion >= 8 && !isGenerator && this.isAsyncProp(prop)) {
                isAsync = true;
                isGenerator = this.options.ecmaVersion >= 9 && this.eat(types.star);
                this.parsePropertyName(prop, refDestructuringErrors);
            }
            else {
                isAsync = false;
            }
            this.parsePropertyValue(prop, isPattern, isGenerator, isAsync, startPos, startLoc, refDestructuringErrors, containsEsc);
            return this.finishNode(prop, "Property");
        };
        pp$3.parsePropertyValue = function (prop, isPattern, isGenerator, isAsync, startPos, startLoc, refDestructuringErrors, containsEsc) {
            if ((isGenerator || isAsync) && this.type === types.colon) {
                this.unexpected();
            }
            if (this.eat(types.colon)) {
                prop.value = isPattern ? this.parseMaybeDefault(this.start, this.startLoc) : this.parseMaybeAssign(false, refDestructuringErrors);
                prop.kind = "init";
            }
            else if (this.options.ecmaVersion >= 6 && this.type === types.parenL) {
                if (isPattern) {
                    this.unexpected();
                }
                prop.kind = "init";
                prop.method = true;
                prop.value = this.parseMethod(isGenerator, isAsync);
            }
            else if (!isPattern && !containsEsc &&
                this.options.ecmaVersion >= 5 && !prop.computed && prop.key.type === "Identifier" &&
                (prop.key.name === "get" || prop.key.name === "set") &&
                (this.type !== types.comma && this.type !== types.braceR && this.type !== types.eq)) {
                if (isGenerator || isAsync) {
                    this.unexpected();
                }
                prop.kind = prop.key.name;
                this.parsePropertyName(prop);
                prop.value = this.parseMethod(false);
                var paramCount = prop.kind === "get" ? 0 : 1;
                if (prop.value.params.length !== paramCount) {
                    var start = prop.value.start;
                    if (prop.kind === "get") {
                        this.raiseRecoverable(start, "getter should have no params");
                    }
                    else {
                        this.raiseRecoverable(start, "setter should have exactly one param");
                    }
                }
                else {
                    if (prop.kind === "set" && prop.value.params[0].type === "RestElement") {
                        this.raiseRecoverable(prop.value.params[0].start, "Setter cannot use rest params");
                    }
                }
            }
            else if (this.options.ecmaVersion >= 6 && !prop.computed && prop.key.type === "Identifier") {
                if (isGenerator || isAsync) {
                    this.unexpected();
                }
                this.checkUnreserved(prop.key);
                if (prop.key.name === "await" && !this.awaitIdentPos) {
                    this.awaitIdentPos = startPos;
                }
                prop.kind = "init";
                if (isPattern) {
                    prop.value = this.parseMaybeDefault(startPos, startLoc, this.copyNode(prop.key));
                }
                else if (this.type === types.eq && refDestructuringErrors) {
                    if (refDestructuringErrors.shorthandAssign < 0) {
                        refDestructuringErrors.shorthandAssign = this.start;
                    }
                    prop.value = this.parseMaybeDefault(startPos, startLoc, this.copyNode(prop.key));
                }
                else {
                    prop.value = this.copyNode(prop.key);
                }
                prop.shorthand = true;
            }
            else {
                this.unexpected();
            }
        };
        pp$3.parsePropertyName = function (prop) {
            if (this.options.ecmaVersion >= 6) {
                if (this.eat(types.bracketL)) {
                    prop.computed = true;
                    prop.key = this.parseMaybeAssign();
                    this.expect(types.bracketR);
                    return prop.key;
                }
                else {
                    prop.computed = false;
                }
            }
            return prop.key = this.type === types.num || this.type === types.string ? this.parseExprAtom() : this.parseIdent(this.options.allowReserved !== "never");
        };
        // Initialize empty function node.
        pp$3.initFunction = function (node) {
            node.id = null;
            if (this.options.ecmaVersion >= 6) {
                node.generator = node.expression = false;
            }
            if (this.options.ecmaVersion >= 8) {
                node.async = false;
            }
        };
        // Parse object or class method.
        pp$3.parseMethod = function (isGenerator, isAsync, allowDirectSuper) {
            var node = this.startNode(), oldYieldPos = this.yieldPos, oldAwaitPos = this.awaitPos, oldAwaitIdentPos = this.awaitIdentPos;
            this.initFunction(node);
            if (this.options.ecmaVersion >= 6) {
                node.generator = isGenerator;
            }
            if (this.options.ecmaVersion >= 8) {
                node.async = !!isAsync;
            }
            this.yieldPos = 0;
            this.awaitPos = 0;
            this.awaitIdentPos = 0;
            this.enterScope(functionFlags(isAsync, node.generator) | SCOPE_SUPER | (allowDirectSuper ? SCOPE_DIRECT_SUPER : 0));
            this.expect(types.parenL);
            node.params = this.parseBindingList(types.parenR, false, this.options.ecmaVersion >= 8);
            this.checkYieldAwaitInDefaultParams();
            this.parseFunctionBody(node, false, true);
            this.yieldPos = oldYieldPos;
            this.awaitPos = oldAwaitPos;
            this.awaitIdentPos = oldAwaitIdentPos;
            return this.finishNode(node, "FunctionExpression");
        };
        // Parse arrow function expression with given parameters.
        pp$3.parseArrowExpression = function (node, params, isAsync) {
            var oldYieldPos = this.yieldPos, oldAwaitPos = this.awaitPos, oldAwaitIdentPos = this.awaitIdentPos;
            this.enterScope(functionFlags(isAsync, false) | SCOPE_ARROW);
            this.initFunction(node);
            if (this.options.ecmaVersion >= 8) {
                node.async = !!isAsync;
            }
            this.yieldPos = 0;
            this.awaitPos = 0;
            this.awaitIdentPos = 0;
            node.params = this.toAssignableList(params, true);
            this.parseFunctionBody(node, true, false);
            this.yieldPos = oldYieldPos;
            this.awaitPos = oldAwaitPos;
            this.awaitIdentPos = oldAwaitIdentPos;
            return this.finishNode(node, "ArrowFunctionExpression");
        };
        // Parse function body and check parameters.
        pp$3.parseFunctionBody = function (node, isArrowFunction, isMethod) {
            var isExpression = isArrowFunction && this.type !== types.braceL;
            var oldStrict = this.strict, useStrict = false;
            if (isExpression) {
                node.body = this.parseMaybeAssign();
                node.expression = true;
                this.checkParams(node, false);
            }
            else {
                var nonSimple = this.options.ecmaVersion >= 7 && !this.isSimpleParamList(node.params);
                if (!oldStrict || nonSimple) {
                    useStrict = this.strictDirective(this.end);
                    // If this is a strict mode function, verify that argument names
                    // are not repeated, and it does not try to bind the words `eval`
                    // or `arguments`.
                    if (useStrict && nonSimple) {
                        this.raiseRecoverable(node.start, "Illegal 'use strict' directive in function with non-simple parameter list");
                    }
                }
                // Start a new scope with regard to labels and the `inFunction`
                // flag (restore them to their old value afterwards).
                var oldLabels = this.labels;
                this.labels = [];
                if (useStrict) {
                    this.strict = true;
                }
                // Add the params to varDeclaredNames to ensure that an error is thrown
                // if a let/const declaration in the function clashes with one of the params.
                this.checkParams(node, !oldStrict && !useStrict && !isArrowFunction && !isMethod && this.isSimpleParamList(node.params));
                // Ensure the function name isn't a forbidden identifier in strict mode, e.g. 'eval'
                if (this.strict && node.id) {
                    this.checkLValSimple(node.id, BIND_OUTSIDE);
                }
                node.body = this.parseBlock(false, undefined, useStrict && !oldStrict);
                node.expression = false;
                this.adaptDirectivePrologue(node.body.body);
                this.labels = oldLabels;
            }
            this.exitScope();
        };
        pp$3.isSimpleParamList = function (params) {
            for (var i = 0, list = params; i < list.length; i += 1) {
                var param = list[i];
                if (param.type !== "Identifier") {
                    return false;
                }
            }
            return true;
        };
        // Checks function params for various disallowed patterns such as using "eval"
        // or "arguments" and duplicate parameters.
        pp$3.checkParams = function (node, allowDuplicates) {
            var nameHash = Object.create(null);
            for (var i = 0, list = node.params; i < list.length; i += 1) {
                var param = list[i];
                this.checkLValInnerPattern(param, BIND_VAR, allowDuplicates ? null : nameHash);
            }
        };
        // Parses a comma-separated list of expressions, and returns them as
        // an array. `close` is the token type that ends the list, and
        // `allowEmpty` can be turned on to allow subsequent commas with
        // nothing in between them to be parsed as `null` (which is needed
        // for array literals).
        pp$3.parseExprList = function (close, allowTrailingComma, allowEmpty, refDestructuringErrors) {
            var elts = [], first = true;
            while (!this.eat(close)) {
                if (!first) {
                    this.expect(types.comma);
                    if (allowTrailingComma && this.afterTrailingComma(close)) {
                        break;
                    }
                }
                else {
                    first = false;
                }
                var elt = (void 0);
                if (allowEmpty && this.type === types.comma) {
                    elt = null;
                }
                else if (this.type === types.ellipsis) {
                    elt = this.parseSpread(refDestructuringErrors);
                    if (refDestructuringErrors && this.type === types.comma && refDestructuringErrors.trailingComma < 0) {
                        refDestructuringErrors.trailingComma = this.start;
                    }
                }
                else {
                    elt = this.parseMaybeAssign(false, refDestructuringErrors);
                }
                elts.push(elt);
            }
            return elts;
        };
        pp$3.checkUnreserved = function (ref) {
            var start = ref.start;
            var end = ref.end;
            var name = ref.name;
            if (this.inGenerator && name === "yield") {
                this.raiseRecoverable(start, "Cannot use 'yield' as identifier inside a generator");
            }
            if (this.inAsync && name === "await") {
                this.raiseRecoverable(start, "Cannot use 'await' as identifier inside an async function");
            }
            if (this.currentThisScope().inClassFieldInit && name === "arguments") {
                this.raiseRecoverable(start, "Cannot use 'arguments' in class field initializer");
            }
            if (this.keywords.test(name)) {
                this.raise(start, ("Unexpected keyword '" + name + "'"));
            }
            if (this.options.ecmaVersion < 6 &&
                this.input.slice(start, end).indexOf("\\") !== -1) {
                return;
            }
            var re = this.strict ? this.reservedWordsStrict : this.reservedWords;
            if (re.test(name)) {
                if (!this.inAsync && name === "await") {
                    this.raiseRecoverable(start, "Cannot use keyword 'await' outside an async function");
                }
                this.raiseRecoverable(start, ("The keyword '" + name + "' is reserved"));
            }
        };
        // Parse the next token as an identifier. If `liberal` is true (used
        // when parsing properties), it will also convert keywords into
        // identifiers.
        pp$3.parseIdent = function (liberal, isBinding) {
            var node = this.startNode();
            if (this.type === types.name) {
                node.name = this.value;
            }
            else if (this.type.keyword) {
                node.name = this.type.keyword;
                // To fix https://github.com/acornjs/acorn/issues/575
                // `class` and `function` keywords push new context into this.context.
                // But there is no chance to pop the context if the keyword is consumed as an identifier such as a property name.
                // If the previous token is a dot, this does not apply because the context-managing code already ignored the keyword
                if ((node.name === "class" || node.name === "function") &&
                    (this.lastTokEnd !== this.lastTokStart + 1 || this.input.charCodeAt(this.lastTokStart) !== 46)) {
                    this.context.pop();
                }
            }
            else {
                this.unexpected();
            }
            this.next(!!liberal);
            this.finishNode(node, "Identifier");
            if (!liberal) {
                this.checkUnreserved(node);
                if (node.name === "await" && !this.awaitIdentPos) {
                    this.awaitIdentPos = node.start;
                }
            }
            return node;
        };
        pp$3.parsePrivateIdent = function () {
            var node = this.startNode();
            if (this.type === types.privateId) {
                node.name = this.value;
            }
            else {
                this.unexpected();
            }
            this.next();
            this.finishNode(node, "PrivateIdentifier");
            // For validating existence
            if (this.privateNameStack.length === 0) {
                this.raise(node.start, ("Private field '#" + (node.name) + "' must be declared in an enclosing class"));
            }
            else {
                this.privateNameStack[this.privateNameStack.length - 1].used.push(node);
            }
            return node;
        };
        // Parses yield expression inside generator.
        pp$3.parseYield = function (forInit) {
            if (!this.yieldPos) {
                this.yieldPos = this.start;
            }
            var node = this.startNode();
            this.next();
            if (this.type === types.semi || this.canInsertSemicolon() || (this.type !== types.star && !this.type.startsExpr)) {
                node.delegate = false;
                node.argument = null;
            }
            else {
                node.delegate = this.eat(types.star);
                node.argument = this.parseMaybeAssign(forInit);
            }
            return this.finishNode(node, "YieldExpression");
        };
        pp$3.parseAwait = function () {
            if (!this.awaitPos) {
                this.awaitPos = this.start;
            }
            var node = this.startNode();
            this.next();
            node.argument = this.parseMaybeUnary(null, true);
            return this.finishNode(node, "AwaitExpression");
        };
        var pp$4 = Parser.prototype;
        // This function is used to raise exceptions on parse errors. It
        // takes an offset integer (into the current `input`) to indicate
        // the location of the error, attaches the position to the end
        // of the error message, and then raises a `SyntaxError` with that
        // message.
        pp$4.raise = function (pos, message) {
            var loc = getLineInfo(this.input, pos);
            message += " (" + loc.line + ":" + loc.column + ")";
            var err = new SyntaxError(message);
            err.pos = pos;
            err.loc = loc;
            err.raisedAt = this.pos;
            throw err;
        };
        pp$4.raiseRecoverable = pp$4.raise;
        pp$4.curPosition = function () {
            if (this.options.locations) {
                return new Position(this.curLine, this.pos - this.lineStart);
            }
        };
        var pp$5 = Parser.prototype;
        var Scope = function Scope(flags) {
            this.flags = flags;
            // A list of var-declared names in the current lexical scope
            this.var = [];
            // A list of lexically-declared names in the current lexical scope
            this.lexical = [];
            // A list of lexically-declared FunctionDeclaration names in the current lexical scope
            this.functions = [];
            // A switch to disallow the identifier reference 'arguments'
            this.inClassFieldInit = false;
        };
        // The functions in this module keep track of declared variables in the current scope in order to detect duplicate variable names.
        pp$5.enterScope = function (flags) {
            this.scopeStack.push(new Scope(flags));
        };
        pp$5.exitScope = function () {
            this.scopeStack.pop();
        };
        // The spec says:
        // > At the top level of a function, or script, function declarations are
        // > treated like var declarations rather than like lexical declarations.
        pp$5.treatFunctionsAsVarInScope = function (scope) {
            return (scope.flags & SCOPE_FUNCTION) || !this.inModule && (scope.flags & SCOPE_TOP);
        };
        pp$5.declareName = function (name, bindingType, pos) {
            var redeclared = false;
            if (bindingType === BIND_LEXICAL) {
                var scope = this.currentScope();
                redeclared = scope.lexical.indexOf(name) > -1 || scope.functions.indexOf(name) > -1 || scope.var.indexOf(name) > -1;
                scope.lexical.push(name);
                if (this.inModule && (scope.flags & SCOPE_TOP)) {
                    delete this.undefinedExports[name];
                }
            }
            else if (bindingType === BIND_SIMPLE_CATCH) {
                var scope$1 = this.currentScope();
                scope$1.lexical.push(name);
            }
            else if (bindingType === BIND_FUNCTION) {
                var scope$2 = this.currentScope();
                if (this.treatFunctionsAsVar) {
                    redeclared = scope$2.lexical.indexOf(name) > -1;
                }
                else {
                    redeclared = scope$2.lexical.indexOf(name) > -1 || scope$2.var.indexOf(name) > -1;
                }
                scope$2.functions.push(name);
            }
            else {
                for (var i = this.scopeStack.length - 1; i >= 0; --i) {
                    var scope$3 = this.scopeStack[i];
                    if (scope$3.lexical.indexOf(name) > -1 && !((scope$3.flags & SCOPE_SIMPLE_CATCH) && scope$3.lexical[0] === name) ||
                        !this.treatFunctionsAsVarInScope(scope$3) && scope$3.functions.indexOf(name) > -1) {
                        redeclared = true;
                        break;
                    }
                    scope$3.var.push(name);
                    if (this.inModule && (scope$3.flags & SCOPE_TOP)) {
                        delete this.undefinedExports[name];
                    }
                    if (scope$3.flags & SCOPE_VAR) {
                        break;
                    }
                }
            }
            if (redeclared) {
                this.raiseRecoverable(pos, ("Identifier '" + name + "' has already been declared"));
            }
        };
        pp$5.checkLocalExport = function (id) {
            // scope.functions must be empty as Module code is always strict.
            if (this.scopeStack[0].lexical.indexOf(id.name) === -1 &&
                this.scopeStack[0].var.indexOf(id.name) === -1) {
                this.undefinedExports[id.name] = id;
            }
        };
        pp$5.currentScope = function () {
            return this.scopeStack[this.scopeStack.length - 1];
        };
        pp$5.currentVarScope = function () {
            for (var i = this.scopeStack.length - 1;; i--) {
                var scope = this.scopeStack[i];
                if (scope.flags & SCOPE_VAR) {
                    return scope;
                }
            }
        };
        // Could be useful for `this`, `new.target`, `super()`, `super.property`, and `super[property]`.
        pp$5.currentThisScope = function () {
            for (var i = this.scopeStack.length - 1;; i--) {
                var scope = this.scopeStack[i];
                if (scope.flags & SCOPE_VAR && !(scope.flags & SCOPE_ARROW)) {
                    return scope;
                }
            }
        };
        var Node = function Node(parser, pos, loc) {
            this.type = "";
            this.start = pos;
            this.end = 0;
            if (parser.options.locations) {
                this.loc = new SourceLocation(parser, loc);
            }
            if (parser.options.directSourceFile) {
                this.sourceFile = parser.options.directSourceFile;
            }
            if (parser.options.ranges) {
                this.range = [pos, 0];
            }
        };
        // Start an AST node, attaching a start offset.
        var pp$6 = Parser.prototype;
        pp$6.startNode = function () {
            return new Node(this, this.start, this.startLoc);
        };
        pp$6.startNodeAt = function (pos, loc) {
            return new Node(this, pos, loc);
        };
        // Finish an AST node, adding `type` and `end` properties.
        function finishNodeAt(node, type, pos, loc) {
            node.type = type;
            node.end = pos;
            if (this.options.locations) {
                node.loc.end = loc;
            }
            if (this.options.ranges) {
                node.range[1] = pos;
            }
            return node;
        }
        pp$6.finishNode = function (node, type) {
            return finishNodeAt.call(this, node, type, this.lastTokEnd, this.lastTokEndLoc);
        };
        // Finish node at given position
        pp$6.finishNodeAt = function (node, type, pos, loc) {
            return finishNodeAt.call(this, node, type, pos, loc);
        };
        pp$6.copyNode = function (node) {
            var newNode = new Node(this, node.start, this.startLoc);
            for (var prop in node) {
                newNode[prop] = node[prop];
            }
            return newNode;
        };
        // The algorithm used to determine whether a regexp can appear at a
        var TokContext = function TokContext(token, isExpr, preserveSpace, override, generator) {
            this.token = token;
            this.isExpr = !!isExpr;
            this.preserveSpace = !!preserveSpace;
            this.override = override;
            this.generator = !!generator;
        };
        var types$1 = {
            b_stat: new TokContext("{", false),
            b_expr: new TokContext("{", true),
            b_tmpl: new TokContext("${", false),
            p_stat: new TokContext("(", false),
            p_expr: new TokContext("(", true),
            q_tmpl: new TokContext("`", true, true, function (p) { return p.tryReadTemplateToken(); }),
            f_stat: new TokContext("function", false),
            f_expr: new TokContext("function", true),
            f_expr_gen: new TokContext("function", true, false, null, true),
            f_gen: new TokContext("function", false, false, null, true)
        };
        var pp$7 = Parser.prototype;
        pp$7.initialContext = function () {
            return [types$1.b_stat];
        };
        pp$7.braceIsBlock = function (prevType) {
            var parent = this.curContext();
            if (parent === types$1.f_expr || parent === types$1.f_stat) {
                return true;
            }
            if (prevType === types.colon && (parent === types$1.b_stat || parent === types$1.b_expr)) {
                return !parent.isExpr;
            }
            // The check for `tt.name && exprAllowed` detects whether we are
            // after a `yield` or `of` construct. See the `updateContext` for
            // `tt.name`.
            if (prevType === types._return || prevType === types.name && this.exprAllowed) {
                return lineBreak.test(this.input.slice(this.lastTokEnd, this.start));
            }
            if (prevType === types._else || prevType === types.semi || prevType === types.eof || prevType === types.parenR || prevType === types.arrow) {
                return true;
            }
            if (prevType === types.braceL) {
                return parent === types$1.b_stat;
            }
            if (prevType === types._var || prevType === types._const || prevType === types.name) {
                return false;
            }
            return !this.exprAllowed;
        };
        pp$7.inGeneratorContext = function () {
            for (var i = this.context.length - 1; i >= 1; i--) {
                var context = this.context[i];
                if (context.token === "function") {
                    return context.generator;
                }
            }
            return false;
        };
        pp$7.updateContext = function (prevType) {
            var update, type = this.type;
            if (type.keyword && prevType === types.dot) {
                this.exprAllowed = false;
            }
            else if (update = type.updateContext) {
                update.call(this, prevType);
            }
            else {
                this.exprAllowed = type.beforeExpr;
            }
        };
        // Token-specific context update code
        types.parenR.updateContext = types.braceR.updateContext = function () {
            if (this.context.length === 1) {
                this.exprAllowed = true;
                return;
            }
            var out = this.context.pop();
            if (out === types$1.b_stat && this.curContext().token === "function") {
                out = this.context.pop();
            }
            this.exprAllowed = !out.isExpr;
        };
        types.braceL.updateContext = function (prevType) {
            this.context.push(this.braceIsBlock(prevType) ? types$1.b_stat : types$1.b_expr);
            this.exprAllowed = true;
        };
        types.dollarBraceL.updateContext = function () {
            this.context.push(types$1.b_tmpl);
            this.exprAllowed = true;
        };
        types.parenL.updateContext = function (prevType) {
            var statementParens = prevType === types._if || prevType === types._for || prevType === types._with || prevType === types._while;
            this.context.push(statementParens ? types$1.p_stat : types$1.p_expr);
            this.exprAllowed = true;
        };
        types.incDec.updateContext = function () {
            // tokExprAllowed stays unchanged
        };
        types._function.updateContext = types._class.updateContext = function (prevType) {
            if (prevType.beforeExpr && prevType !== types._else &&
                !(prevType === types.semi && this.curContext() !== types$1.p_stat) &&
                !(prevType === types._return && lineBreak.test(this.input.slice(this.lastTokEnd, this.start))) &&
                !((prevType === types.colon || prevType === types.braceL) && this.curContext() === types$1.b_stat)) {
                this.context.push(types$1.f_expr);
            }
            else {
                this.context.push(types$1.f_stat);
            }
            this.exprAllowed = false;
        };
        types.backQuote.updateContext = function () {
            if (this.curContext() === types$1.q_tmpl) {
                this.context.pop();
            }
            else {
                this.context.push(types$1.q_tmpl);
            }
            this.exprAllowed = false;
        };
        types.star.updateContext = function (prevType) {
            if (prevType === types._function) {
                var index = this.context.length - 1;
                if (this.context[index] === types$1.f_expr) {
                    this.context[index] = types$1.f_expr_gen;
                }
                else {
                    this.context[index] = types$1.f_gen;
                }
            }
            this.exprAllowed = true;
        };
        types.name.updateContext = function (prevType) {
            var allowed = false;
            if (this.options.ecmaVersion >= 6 && prevType !== types.dot) {
                if (this.value === "of" && !this.exprAllowed ||
                    this.value === "yield" && this.inGeneratorContext()) {
                    allowed = true;
                }
            }
            this.exprAllowed = allowed;
        };
        // This file contains Unicode properties extracted from the ECMAScript
        // specification. The lists are extracted like so:
        // $$('#table-binary-unicode-properties > figure > table > tbody > tr > td:nth-child(1) code').map(el => el.innerText)
        // #table-binary-unicode-properties
        var ecma9BinaryProperties = "ASCII ASCII_Hex_Digit AHex Alphabetic Alpha Any Assigned Bidi_Control Bidi_C Bidi_Mirrored Bidi_M Case_Ignorable CI Cased Changes_When_Casefolded CWCF Changes_When_Casemapped CWCM Changes_When_Lowercased CWL Changes_When_NFKC_Casefolded CWKCF Changes_When_Titlecased CWT Changes_When_Uppercased CWU Dash Default_Ignorable_Code_Point DI Deprecated Dep Diacritic Dia Emoji Emoji_Component Emoji_Modifier Emoji_Modifier_Base Emoji_Presentation Extender Ext Grapheme_Base Gr_Base Grapheme_Extend Gr_Ext Hex_Digit Hex IDS_Binary_Operator IDSB IDS_Trinary_Operator IDST ID_Continue IDC ID_Start IDS Ideographic Ideo Join_Control Join_C Logical_Order_Exception LOE Lowercase Lower Math Noncharacter_Code_Point NChar Pattern_Syntax Pat_Syn Pattern_White_Space Pat_WS Quotation_Mark QMark Radical Regional_Indicator RI Sentence_Terminal STerm Soft_Dotted SD Terminal_Punctuation Term Unified_Ideograph UIdeo Uppercase Upper Variation_Selector VS White_Space space XID_Continue XIDC XID_Start XIDS";
        var ecma10BinaryProperties = ecma9BinaryProperties + " Extended_Pictographic";
        var ecma11BinaryProperties = ecma10BinaryProperties;
        var ecma12BinaryProperties = ecma11BinaryProperties + " EBase EComp EMod EPres ExtPict";
        var unicodeBinaryProperties = {
            9: ecma9BinaryProperties,
            10: ecma10BinaryProperties,
            11: ecma11BinaryProperties,
            12: ecma12BinaryProperties
        };
        // #table-unicode-general-category-values
        var unicodeGeneralCategoryValues = "Cased_Letter LC Close_Punctuation Pe Connector_Punctuation Pc Control Cc cntrl Currency_Symbol Sc Dash_Punctuation Pd Decimal_Number Nd digit Enclosing_Mark Me Final_Punctuation Pf Format Cf Initial_Punctuation Pi Letter L Letter_Number Nl Line_Separator Zl Lowercase_Letter Ll Mark M Combining_Mark Math_Symbol Sm Modifier_Letter Lm Modifier_Symbol Sk Nonspacing_Mark Mn Number N Open_Punctuation Ps Other C Other_Letter Lo Other_Number No Other_Punctuation Po Other_Symbol So Paragraph_Separator Zp Private_Use Co Punctuation P punct Separator Z Space_Separator Zs Spacing_Mark Mc Surrogate Cs Symbol S Titlecase_Letter Lt Unassigned Cn Uppercase_Letter Lu";
        // #table-unicode-script-values
        var ecma9ScriptValues = "Adlam Adlm Ahom Ahom Anatolian_Hieroglyphs Hluw Arabic Arab Armenian Armn Avestan Avst Balinese Bali Bamum Bamu Bassa_Vah Bass Batak Batk Bengali Beng Bhaiksuki Bhks Bopomofo Bopo Brahmi Brah Braille Brai Buginese Bugi Buhid Buhd Canadian_Aboriginal Cans Carian Cari Caucasian_Albanian Aghb Chakma Cakm Cham Cham Cherokee Cher Common Zyyy Coptic Copt Qaac Cuneiform Xsux Cypriot Cprt Cyrillic Cyrl Deseret Dsrt Devanagari Deva Duployan Dupl Egyptian_Hieroglyphs Egyp Elbasan Elba Ethiopic Ethi Georgian Geor Glagolitic Glag Gothic Goth Grantha Gran Greek Grek Gujarati Gujr Gurmukhi Guru Han Hani Hangul Hang Hanunoo Hano Hatran Hatr Hebrew Hebr Hiragana Hira Imperial_Aramaic Armi Inherited Zinh Qaai Inscriptional_Pahlavi Phli Inscriptional_Parthian Prti Javanese Java Kaithi Kthi Kannada Knda Katakana Kana Kayah_Li Kali Kharoshthi Khar Khmer Khmr Khojki Khoj Khudawadi Sind Lao Laoo Latin Latn Lepcha Lepc Limbu Limb Linear_A Lina Linear_B Linb Lisu Lisu Lycian Lyci Lydian Lydi Mahajani Mahj Malayalam Mlym Mandaic Mand Manichaean Mani Marchen Marc Masaram_Gondi Gonm Meetei_Mayek Mtei Mende_Kikakui Mend Meroitic_Cursive Merc Meroitic_Hieroglyphs Mero Miao Plrd Modi Modi Mongolian Mong Mro Mroo Multani Mult Myanmar Mymr Nabataean Nbat New_Tai_Lue Talu Newa Newa Nko Nkoo Nushu Nshu Ogham Ogam Ol_Chiki Olck Old_Hungarian Hung Old_Italic Ital Old_North_Arabian Narb Old_Permic Perm Old_Persian Xpeo Old_South_Arabian Sarb Old_Turkic Orkh Oriya Orya Osage Osge Osmanya Osma Pahawh_Hmong Hmng Palmyrene Palm Pau_Cin_Hau Pauc Phags_Pa Phag Phoenician Phnx Psalter_Pahlavi Phlp Rejang Rjng Runic Runr Samaritan Samr Saurashtra Saur Sharada Shrd Shavian Shaw Siddham Sidd SignWriting Sgnw Sinhala Sinh Sora_Sompeng Sora Soyombo Soyo Sundanese Sund Syloti_Nagri Sylo Syriac Syrc Tagalog Tglg Tagbanwa Tagb Tai_Le Tale Tai_Tham Lana Tai_Viet Tavt Takri Takr Tamil Taml Tangut Tang Telugu Telu Thaana Thaa Thai Thai Tibetan Tibt Tifinagh Tfng Tirhuta Tirh Ugaritic Ugar Vai Vaii Warang_Citi Wara Yi Yiii Zanabazar_Square Zanb";
        var ecma10ScriptValues = ecma9ScriptValues + " Dogra Dogr Gunjala_Gondi Gong Hanifi_Rohingya Rohg Makasar Maka Medefaidrin Medf Old_Sogdian Sogo Sogdian Sogd";
        var ecma11ScriptValues = ecma10ScriptValues + " Elymaic Elym Nandinagari Nand Nyiakeng_Puachue_Hmong Hmnp Wancho Wcho";
        var ecma12ScriptValues = ecma11ScriptValues + " Chorasmian Chrs Diak Dives_Akuru Khitan_Small_Script Kits Yezi Yezidi";
        var unicodeScriptValues = {
            9: ecma9ScriptValues,
            10: ecma10ScriptValues,
            11: ecma11ScriptValues,
            12: ecma12ScriptValues
        };
        var data = {};
        function buildUnicodeData(ecmaVersion) {
            var d = data[ecmaVersion] = {
                binary: wordsRegexp(unicodeBinaryProperties[ecmaVersion] + " " + unicodeGeneralCategoryValues),
                nonBinary: {
                    General_Category: wordsRegexp(unicodeGeneralCategoryValues),
                    Script: wordsRegexp(unicodeScriptValues[ecmaVersion])
                }
            };
            d.nonBinary.Script_Extensions = d.nonBinary.Script;
            d.nonBinary.gc = d.nonBinary.General_Category;
            d.nonBinary.sc = d.nonBinary.Script;
            d.nonBinary.scx = d.nonBinary.Script_Extensions;
        }
        buildUnicodeData(9);
        buildUnicodeData(10);
        buildUnicodeData(11);
        buildUnicodeData(12);
        var pp$8 = Parser.prototype;
        var RegExpValidationState = function RegExpValidationState(parser) {
            this.parser = parser;
            this.validFlags = "gim" + (parser.options.ecmaVersion >= 6 ? "uy" : "") + (parser.options.ecmaVersion >= 9 ? "s" : "") + (parser.options.ecmaVersion >= 13 ? "d" : "");
            this.unicodeProperties = data[parser.options.ecmaVersion >= 12 ? 12 : parser.options.ecmaVersion];
            this.source = "";
            this.flags = "";
            this.start = 0;
            this.switchU = false;
            this.switchN = false;
            this.pos = 0;
            this.lastIntValue = 0;
            this.lastStringValue = "";
            this.lastAssertionIsQuantifiable = false;
            this.numCapturingParens = 0;
            this.maxBackReference = 0;
            this.groupNames = [];
            this.backReferenceNames = [];
        };
        RegExpValidationState.prototype.reset = function reset(start, pattern, flags) {
            var unicode = flags.indexOf("u") !== -1;
            this.start = start | 0;
            this.source = pattern + "";
            this.flags = flags;
            this.switchU = unicode && this.parser.options.ecmaVersion >= 6;
            this.switchN = unicode && this.parser.options.ecmaVersion >= 9;
        };
        RegExpValidationState.prototype.raise = function raise(message) {
            this.parser.raiseRecoverable(this.start, ("Invalid regular expression: /" + (this.source) + "/: " + message));
        };
        // If u flag is given, this returns the code point at the index (it combines a surrogate pair).
        // Otherwise, this returns the code unit of the index (can be a part of a surrogate pair).
        RegExpValidationState.prototype.at = function at(i, forceU) {
            if (forceU === void 0)
                forceU = false;
            var s = this.source;
            var l = s.length;
            if (i >= l) {
                return -1;
            }
            var c = s.charCodeAt(i);
            if (!(forceU || this.switchU) || c <= 0xD7FF || c >= 0xE000 || i + 1 >= l) {
                return c;
            }
            var next = s.charCodeAt(i + 1);
            return next >= 0xDC00 && next <= 0xDFFF ? (c << 10) + next - 0x35FDC00 : c;
        };
        RegExpValidationState.prototype.nextIndex = function nextIndex(i, forceU) {
            if (forceU === void 0)
                forceU = false;
            var s = this.source;
            var l = s.length;
            if (i >= l) {
                return l;
            }
            var c = s.charCodeAt(i), next;
            if (!(forceU || this.switchU) || c <= 0xD7FF || c >= 0xE000 || i + 1 >= l ||
                (next = s.charCodeAt(i + 1)) < 0xDC00 || next > 0xDFFF) {
                return i + 1;
            }
            return i + 2;
        };
        RegExpValidationState.prototype.current = function current(forceU) {
            if (forceU === void 0)
                forceU = false;
            return this.at(this.pos, forceU);
        };
        RegExpValidationState.prototype.lookahead = function lookahead(forceU) {
            if (forceU === void 0)
                forceU = false;
            return this.at(this.nextIndex(this.pos, forceU), forceU);
        };
        RegExpValidationState.prototype.advance = function advance(forceU) {
            if (forceU === void 0)
                forceU = false;
            this.pos = this.nextIndex(this.pos, forceU);
        };
        RegExpValidationState.prototype.eat = function eat(ch, forceU) {
            if (forceU === void 0)
                forceU = false;
            if (this.current(forceU) === ch) {
                this.advance(forceU);
                return true;
            }
            return false;
        };
        function codePointToString(ch) {
            if (ch <= 0xFFFF) {
                return String.fromCharCode(ch);
            }
            ch -= 0x10000;
            return String.fromCharCode((ch >> 10) + 0xD800, (ch & 0x03FF) + 0xDC00);
        }
        /**
         * Validate the flags part of a given RegExpLiteral.
         *
         * @param {RegExpValidationState} state The state to validate RegExp.
         * @returns {void}
         */
        pp$8.validateRegExpFlags = function (state) {
            var validFlags = state.validFlags;
            var flags = state.flags;
            for (var i = 0; i < flags.length; i++) {
                var flag = flags.charAt(i);
                if (validFlags.indexOf(flag) === -1) {
                    this.raise(state.start, "Invalid regular expression flag");
                }
                if (flags.indexOf(flag, i + 1) > -1) {
                    this.raise(state.start, "Duplicate regular expression flag");
                }
            }
        };
        /**
         * Validate the pattern part of a given RegExpLiteral.
         *
         * @param {RegExpValidationState} state The state to validate RegExp.
         * @returns {void}
         */
        pp$8.validateRegExpPattern = function (state) {
            this.regexp_pattern(state);
            // The goal symbol for the parse is |Pattern[~U, ~N]|. If the result of
            // parsing contains a |GroupName|, reparse with the goal symbol
            // |Pattern[~U, +N]| and use this result instead. Throw a *SyntaxError*
            // exception if _P_ did not conform to the grammar, if any elements of _P_
            // were not matched by the parse, or if any Early Error conditions exist.
            if (!state.switchN && this.options.ecmaVersion >= 9 && state.groupNames.length > 0) {
                state.switchN = true;
                this.regexp_pattern(state);
            }
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-Pattern
        pp$8.regexp_pattern = function (state) {
            state.pos = 0;
            state.lastIntValue = 0;
            state.lastStringValue = "";
            state.lastAssertionIsQuantifiable = false;
            state.numCapturingParens = 0;
            state.maxBackReference = 0;
            state.groupNames.length = 0;
            state.backReferenceNames.length = 0;
            this.regexp_disjunction(state);
            if (state.pos !== state.source.length) {
                // Make the same messages as V8.
                if (state.eat(0x29 /* ) */)) {
                    state.raise("Unmatched ')'");
                }
                if (state.eat(0x5D /* ] */) || state.eat(0x7D /* } */)) {
                    state.raise("Lone quantifier brackets");
                }
            }
            if (state.maxBackReference > state.numCapturingParens) {
                state.raise("Invalid escape");
            }
            for (var i = 0, list = state.backReferenceNames; i < list.length; i += 1) {
                var name = list[i];
                if (state.groupNames.indexOf(name) === -1) {
                    state.raise("Invalid named capture referenced");
                }
            }
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-Disjunction
        pp$8.regexp_disjunction = function (state) {
            this.regexp_alternative(state);
            while (state.eat(0x7C /* | */)) {
                this.regexp_alternative(state);
            }
            // Make the same message as V8.
            if (this.regexp_eatQuantifier(state, true)) {
                state.raise("Nothing to repeat");
            }
            if (state.eat(0x7B /* { */)) {
                state.raise("Lone quantifier brackets");
            }
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-Alternative
        pp$8.regexp_alternative = function (state) {
            while (state.pos < state.source.length && this.regexp_eatTerm(state)) { }
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-Term
        pp$8.regexp_eatTerm = function (state) {
            if (this.regexp_eatAssertion(state)) {
                // Handle `QuantifiableAssertion Quantifier` alternative.
                // `state.lastAssertionIsQuantifiable` is true if the last eaten Assertion
                // is a QuantifiableAssertion.
                if (state.lastAssertionIsQuantifiable && this.regexp_eatQuantifier(state)) {
                    // Make the same message as V8.
                    if (state.switchU) {
                        state.raise("Invalid quantifier");
                    }
                }
                return true;
            }
            if (state.switchU ? this.regexp_eatAtom(state) : this.regexp_eatExtendedAtom(state)) {
                this.regexp_eatQuantifier(state);
                return true;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-Assertion
        pp$8.regexp_eatAssertion = function (state) {
            var start = state.pos;
            state.lastAssertionIsQuantifiable = false;
            // ^, $
            if (state.eat(0x5E /* ^ */) || state.eat(0x24 /* $ */)) {
                return true;
            }
            // \b \B
            if (state.eat(0x5C /* \ */)) {
                if (state.eat(0x42 /* B */) || state.eat(0x62 /* b */)) {
                    return true;
                }
                state.pos = start;
            }
            // Lookahead / Lookbehind
            if (state.eat(0x28 /* ( */) && state.eat(0x3F /* ? */)) {
                var lookbehind = false;
                if (this.options.ecmaVersion >= 9) {
                    lookbehind = state.eat(0x3C /* < */);
                }
                if (state.eat(0x3D /* = */) || state.eat(0x21 /* ! */)) {
                    this.regexp_disjunction(state);
                    if (!state.eat(0x29 /* ) */)) {
                        state.raise("Unterminated group");
                    }
                    state.lastAssertionIsQuantifiable = !lookbehind;
                    return true;
                }
            }
            state.pos = start;
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-Quantifier
        pp$8.regexp_eatQuantifier = function (state, noError) {
            if (noError === void 0)
                noError = false;
            if (this.regexp_eatQuantifierPrefix(state, noError)) {
                state.eat(0x3F /* ? */);
                return true;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-QuantifierPrefix
        pp$8.regexp_eatQuantifierPrefix = function (state, noError) {
            return (state.eat(0x2A /* * */) ||
                state.eat(0x2B /* + */) ||
                state.eat(0x3F /* ? */) ||
                this.regexp_eatBracedQuantifier(state, noError));
        };
        pp$8.regexp_eatBracedQuantifier = function (state, noError) {
            var start = state.pos;
            if (state.eat(0x7B /* { */)) {
                var min = 0, max = -1;
                if (this.regexp_eatDecimalDigits(state)) {
                    min = state.lastIntValue;
                    if (state.eat(0x2C /* , */) && this.regexp_eatDecimalDigits(state)) {
                        max = state.lastIntValue;
                    }
                    if (state.eat(0x7D /* } */)) {
                        // SyntaxError in https://www.ecma-international.org/ecma-262/8.0/#sec-term
                        if (max !== -1 && max < min && !noError) {
                            state.raise("numbers out of order in {} quantifier");
                        }
                        return true;
                    }
                }
                if (state.switchU && !noError) {
                    state.raise("Incomplete quantifier");
                }
                state.pos = start;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-Atom
        pp$8.regexp_eatAtom = function (state) {
            return (this.regexp_eatPatternCharacters(state) ||
                state.eat(0x2E /* . */) ||
                this.regexp_eatReverseSolidusAtomEscape(state) ||
                this.regexp_eatCharacterClass(state) ||
                this.regexp_eatUncapturingGroup(state) ||
                this.regexp_eatCapturingGroup(state));
        };
        pp$8.regexp_eatReverseSolidusAtomEscape = function (state) {
            var start = state.pos;
            if (state.eat(0x5C /* \ */)) {
                if (this.regexp_eatAtomEscape(state)) {
                    return true;
                }
                state.pos = start;
            }
            return false;
        };
        pp$8.regexp_eatUncapturingGroup = function (state) {
            var start = state.pos;
            if (state.eat(0x28 /* ( */)) {
                if (state.eat(0x3F /* ? */) && state.eat(0x3A /* : */)) {
                    this.regexp_disjunction(state);
                    if (state.eat(0x29 /* ) */)) {
                        return true;
                    }
                    state.raise("Unterminated group");
                }
                state.pos = start;
            }
            return false;
        };
        pp$8.regexp_eatCapturingGroup = function (state) {
            if (state.eat(0x28 /* ( */)) {
                if (this.options.ecmaVersion >= 9) {
                    this.regexp_groupSpecifier(state);
                }
                else if (state.current() === 0x3F /* ? */) {
                    state.raise("Invalid group");
                }
                this.regexp_disjunction(state);
                if (state.eat(0x29 /* ) */)) {
                    state.numCapturingParens += 1;
                    return true;
                }
                state.raise("Unterminated group");
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-ExtendedAtom
        pp$8.regexp_eatExtendedAtom = function (state) {
            return (state.eat(0x2E /* . */) ||
                this.regexp_eatReverseSolidusAtomEscape(state) ||
                this.regexp_eatCharacterClass(state) ||
                this.regexp_eatUncapturingGroup(state) ||
                this.regexp_eatCapturingGroup(state) ||
                this.regexp_eatInvalidBracedQuantifier(state) ||
                this.regexp_eatExtendedPatternCharacter(state));
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-InvalidBracedQuantifier
        pp$8.regexp_eatInvalidBracedQuantifier = function (state) {
            if (this.regexp_eatBracedQuantifier(state, true)) {
                state.raise("Nothing to repeat");
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-SyntaxCharacter
        pp$8.regexp_eatSyntaxCharacter = function (state) {
            var ch = state.current();
            if (isSyntaxCharacter(ch)) {
                state.lastIntValue = ch;
                state.advance();
                return true;
            }
            return false;
        };
        function isSyntaxCharacter(ch) {
            return (ch === 0x24 /* $ */ ||
                ch >= 0x28 /* ( */ && ch <= 0x2B /* + */ ||
                ch === 0x2E /* . */ ||
                ch === 0x3F /* ? */ ||
                ch >= 0x5B /* [ */ && ch <= 0x5E /* ^ */ ||
                ch >= 0x7B /* { */ && ch <= 0x7D /* } */);
        }
        // https://www.ecma-international.org/ecma-262/8.0/#prod-PatternCharacter
        // But eat eager.
        pp$8.regexp_eatPatternCharacters = function (state) {
            var start = state.pos;
            var ch = 0;
            while ((ch = state.current()) !== -1 && !isSyntaxCharacter(ch)) {
                state.advance();
            }
            return state.pos !== start;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-ExtendedPatternCharacter
        pp$8.regexp_eatExtendedPatternCharacter = function (state) {
            var ch = state.current();
            if (ch !== -1 &&
                ch !== 0x24 /* $ */ &&
                !(ch >= 0x28 /* ( */ && ch <= 0x2B /* + */) &&
                ch !== 0x2E /* . */ &&
                ch !== 0x3F /* ? */ &&
                ch !== 0x5B /* [ */ &&
                ch !== 0x5E /* ^ */ &&
                ch !== 0x7C /* | */) {
                state.advance();
                return true;
            }
            return false;
        };
        // GroupSpecifier ::
        //   [empty]
        //   `?` GroupName
        pp$8.regexp_groupSpecifier = function (state) {
            if (state.eat(0x3F /* ? */)) {
                if (this.regexp_eatGroupName(state)) {
                    if (state.groupNames.indexOf(state.lastStringValue) !== -1) {
                        state.raise("Duplicate capture group name");
                    }
                    state.groupNames.push(state.lastStringValue);
                    return;
                }
                state.raise("Invalid group");
            }
        };
        // GroupName ::
        //   `<` RegExpIdentifierName `>`
        // Note: this updates `state.lastStringValue` property with the eaten name.
        pp$8.regexp_eatGroupName = function (state) {
            state.lastStringValue = "";
            if (state.eat(0x3C /* < */)) {
                if (this.regexp_eatRegExpIdentifierName(state) && state.eat(0x3E /* > */)) {
                    return true;
                }
                state.raise("Invalid capture group name");
            }
            return false;
        };
        // RegExpIdentifierName ::
        //   RegExpIdentifierStart
        //   RegExpIdentifierName RegExpIdentifierPart
        // Note: this updates `state.lastStringValue` property with the eaten name.
        pp$8.regexp_eatRegExpIdentifierName = function (state) {
            state.lastStringValue = "";
            if (this.regexp_eatRegExpIdentifierStart(state)) {
                state.lastStringValue += codePointToString(state.lastIntValue);
                while (this.regexp_eatRegExpIdentifierPart(state)) {
                    state.lastStringValue += codePointToString(state.lastIntValue);
                }
                return true;
            }
            return false;
        };
        // RegExpIdentifierStart ::
        //   UnicodeIDStart
        //   `$`
        //   `_`
        //   `\` RegExpUnicodeEscapeSequence[+U]
        pp$8.regexp_eatRegExpIdentifierStart = function (state) {
            var start = state.pos;
            var forceU = this.options.ecmaVersion >= 11;
            var ch = state.current(forceU);
            state.advance(forceU);
            if (ch === 0x5C /* \ */ && this.regexp_eatRegExpUnicodeEscapeSequence(state, forceU)) {
                ch = state.lastIntValue;
            }
            if (isRegExpIdentifierStart(ch)) {
                state.lastIntValue = ch;
                return true;
            }
            state.pos = start;
            return false;
        };
        function isRegExpIdentifierStart(ch) {
            return isIdentifierStart(ch, true) || ch === 0x24 /* $ */ || ch === 0x5F; /* _ */
        }
        // RegExpIdentifierPart ::
        //   UnicodeIDContinue
        //   `$`
        //   `_`
        //   `\` RegExpUnicodeEscapeSequence[+U]
        //   <ZWNJ>
        //   <ZWJ>
        pp$8.regexp_eatRegExpIdentifierPart = function (state) {
            var start = state.pos;
            var forceU = this.options.ecmaVersion >= 11;
            var ch = state.current(forceU);
            state.advance(forceU);
            if (ch === 0x5C /* \ */ && this.regexp_eatRegExpUnicodeEscapeSequence(state, forceU)) {
                ch = state.lastIntValue;
            }
            if (isRegExpIdentifierPart(ch)) {
                state.lastIntValue = ch;
                return true;
            }
            state.pos = start;
            return false;
        };
        function isRegExpIdentifierPart(ch) {
            return isIdentifierChar(ch, true) || ch === 0x24 /* $ */ || ch === 0x5F /* _ */ || ch === 0x200C /* <ZWNJ> */ || ch === 0x200D; /* <ZWJ> */
        }
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-AtomEscape
        pp$8.regexp_eatAtomEscape = function (state) {
            if (this.regexp_eatBackReference(state) ||
                this.regexp_eatCharacterClassEscape(state) ||
                this.regexp_eatCharacterEscape(state) ||
                (state.switchN && this.regexp_eatKGroupName(state))) {
                return true;
            }
            if (state.switchU) {
                // Make the same message as V8.
                if (state.current() === 0x63 /* c */) {
                    state.raise("Invalid unicode escape");
                }
                state.raise("Invalid escape");
            }
            return false;
        };
        pp$8.regexp_eatBackReference = function (state) {
            var start = state.pos;
            if (this.regexp_eatDecimalEscape(state)) {
                var n = state.lastIntValue;
                if (state.switchU) {
                    // For SyntaxError in https://www.ecma-international.org/ecma-262/8.0/#sec-atomescape
                    if (n > state.maxBackReference) {
                        state.maxBackReference = n;
                    }
                    return true;
                }
                if (n <= state.numCapturingParens) {
                    return true;
                }
                state.pos = start;
            }
            return false;
        };
        pp$8.regexp_eatKGroupName = function (state) {
            if (state.eat(0x6B /* k */)) {
                if (this.regexp_eatGroupName(state)) {
                    state.backReferenceNames.push(state.lastStringValue);
                    return true;
                }
                state.raise("Invalid named reference");
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-CharacterEscape
        pp$8.regexp_eatCharacterEscape = function (state) {
            return (this.regexp_eatControlEscape(state) ||
                this.regexp_eatCControlLetter(state) ||
                this.regexp_eatZero(state) ||
                this.regexp_eatHexEscapeSequence(state) ||
                this.regexp_eatRegExpUnicodeEscapeSequence(state, false) ||
                (!state.switchU && this.regexp_eatLegacyOctalEscapeSequence(state)) ||
                this.regexp_eatIdentityEscape(state));
        };
        pp$8.regexp_eatCControlLetter = function (state) {
            var start = state.pos;
            if (state.eat(0x63 /* c */)) {
                if (this.regexp_eatControlLetter(state)) {
                    return true;
                }
                state.pos = start;
            }
            return false;
        };
        pp$8.regexp_eatZero = function (state) {
            if (state.current() === 0x30 /* 0 */ && !isDecimalDigit(state.lookahead())) {
                state.lastIntValue = 0;
                state.advance();
                return true;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-ControlEscape
        pp$8.regexp_eatControlEscape = function (state) {
            var ch = state.current();
            if (ch === 0x74 /* t */) {
                state.lastIntValue = 0x09; /* \t */
                state.advance();
                return true;
            }
            if (ch === 0x6E /* n */) {
                state.lastIntValue = 0x0A; /* \n */
                state.advance();
                return true;
            }
            if (ch === 0x76 /* v */) {
                state.lastIntValue = 0x0B; /* \v */
                state.advance();
                return true;
            }
            if (ch === 0x66 /* f */) {
                state.lastIntValue = 0x0C; /* \f */
                state.advance();
                return true;
            }
            if (ch === 0x72 /* r */) {
                state.lastIntValue = 0x0D; /* \r */
                state.advance();
                return true;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-ControlLetter
        pp$8.regexp_eatControlLetter = function (state) {
            var ch = state.current();
            if (isControlLetter(ch)) {
                state.lastIntValue = ch % 0x20;
                state.advance();
                return true;
            }
            return false;
        };
        function isControlLetter(ch) {
            return ((ch >= 0x41 /* A */ && ch <= 0x5A /* Z */) ||
                (ch >= 0x61 /* a */ && ch <= 0x7A /* z */));
        }
        // https://www.ecma-international.org/ecma-262/8.0/#prod-RegExpUnicodeEscapeSequence
        pp$8.regexp_eatRegExpUnicodeEscapeSequence = function (state, forceU) {
            if (forceU === void 0)
                forceU = false;
            var start = state.pos;
            var switchU = forceU || state.switchU;
            if (state.eat(0x75 /* u */)) {
                if (this.regexp_eatFixedHexDigits(state, 4)) {
                    var lead = state.lastIntValue;
                    if (switchU && lead >= 0xD800 && lead <= 0xDBFF) {
                        var leadSurrogateEnd = state.pos;
                        if (state.eat(0x5C /* \ */) && state.eat(0x75 /* u */) && this.regexp_eatFixedHexDigits(state, 4)) {
                            var trail = state.lastIntValue;
                            if (trail >= 0xDC00 && trail <= 0xDFFF) {
                                state.lastIntValue = (lead - 0xD800) * 0x400 + (trail - 0xDC00) + 0x10000;
                                return true;
                            }
                        }
                        state.pos = leadSurrogateEnd;
                        state.lastIntValue = lead;
                    }
                    return true;
                }
                if (switchU &&
                    state.eat(0x7B /* { */) &&
                    this.regexp_eatHexDigits(state) &&
                    state.eat(0x7D /* } */) &&
                    isValidUnicode(state.lastIntValue)) {
                    return true;
                }
                if (switchU) {
                    state.raise("Invalid unicode escape");
                }
                state.pos = start;
            }
            return false;
        };
        function isValidUnicode(ch) {
            return ch >= 0 && ch <= 0x10FFFF;
        }
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-IdentityEscape
        pp$8.regexp_eatIdentityEscape = function (state) {
            if (state.switchU) {
                if (this.regexp_eatSyntaxCharacter(state)) {
                    return true;
                }
                if (state.eat(0x2F /* / */)) {
                    state.lastIntValue = 0x2F; /* / */
                    return true;
                }
                return false;
            }
            var ch = state.current();
            if (ch !== 0x63 /* c */ && (!state.switchN || ch !== 0x6B /* k */)) {
                state.lastIntValue = ch;
                state.advance();
                return true;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-DecimalEscape
        pp$8.regexp_eatDecimalEscape = function (state) {
            state.lastIntValue = 0;
            var ch = state.current();
            if (ch >= 0x31 /* 1 */ && ch <= 0x39 /* 9 */) {
                do {
                    state.lastIntValue = 10 * state.lastIntValue + (ch - 0x30 /* 0 */);
                    state.advance();
                } while ((ch = state.current()) >= 0x30 /* 0 */ && ch <= 0x39 /* 9 */);
                return true;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-CharacterClassEscape
        pp$8.regexp_eatCharacterClassEscape = function (state) {
            var ch = state.current();
            if (isCharacterClassEscape(ch)) {
                state.lastIntValue = -1;
                state.advance();
                return true;
            }
            if (state.switchU &&
                this.options.ecmaVersion >= 9 &&
                (ch === 0x50 /* P */ || ch === 0x70 /* p */)) {
                state.lastIntValue = -1;
                state.advance();
                if (state.eat(0x7B /* { */) &&
                    this.regexp_eatUnicodePropertyValueExpression(state) &&
                    state.eat(0x7D /* } */)) {
                    return true;
                }
                state.raise("Invalid property name");
            }
            return false;
        };
        function isCharacterClassEscape(ch) {
            return (ch === 0x64 /* d */ ||
                ch === 0x44 /* D */ ||
                ch === 0x73 /* s */ ||
                ch === 0x53 /* S */ ||
                ch === 0x77 /* w */ ||
                ch === 0x57 /* W */);
        }
        // UnicodePropertyValueExpression ::
        //   UnicodePropertyName `=` UnicodePropertyValue
        //   LoneUnicodePropertyNameOrValue
        pp$8.regexp_eatUnicodePropertyValueExpression = function (state) {
            var start = state.pos;
            // UnicodePropertyName `=` UnicodePropertyValue
            if (this.regexp_eatUnicodePropertyName(state) && state.eat(0x3D /* = */)) {
                var name = state.lastStringValue;
                if (this.regexp_eatUnicodePropertyValue(state)) {
                    var value = state.lastStringValue;
                    this.regexp_validateUnicodePropertyNameAndValue(state, name, value);
                    return true;
                }
            }
            state.pos = start;
            // LoneUnicodePropertyNameOrValue
            if (this.regexp_eatLoneUnicodePropertyNameOrValue(state)) {
                var nameOrValue = state.lastStringValue;
                this.regexp_validateUnicodePropertyNameOrValue(state, nameOrValue);
                return true;
            }
            return false;
        };
        pp$8.regexp_validateUnicodePropertyNameAndValue = function (state, name, value) {
            if (!has(state.unicodeProperties.nonBinary, name)) {
                state.raise("Invalid property name");
            }
            if (!state.unicodeProperties.nonBinary[name].test(value)) {
                state.raise("Invalid property value");
            }
        };
        pp$8.regexp_validateUnicodePropertyNameOrValue = function (state, nameOrValue) {
            if (!state.unicodeProperties.binary.test(nameOrValue)) {
                state.raise("Invalid property name");
            }
        };
        // UnicodePropertyName ::
        //   UnicodePropertyNameCharacters
        pp$8.regexp_eatUnicodePropertyName = function (state) {
            var ch = 0;
            state.lastStringValue = "";
            while (isUnicodePropertyNameCharacter(ch = state.current())) {
                state.lastStringValue += codePointToString(ch);
                state.advance();
            }
            return state.lastStringValue !== "";
        };
        function isUnicodePropertyNameCharacter(ch) {
            return isControlLetter(ch) || ch === 0x5F; /* _ */
        }
        // UnicodePropertyValue ::
        //   UnicodePropertyValueCharacters
        pp$8.regexp_eatUnicodePropertyValue = function (state) {
            var ch = 0;
            state.lastStringValue = "";
            while (isUnicodePropertyValueCharacter(ch = state.current())) {
                state.lastStringValue += codePointToString(ch);
                state.advance();
            }
            return state.lastStringValue !== "";
        };
        function isUnicodePropertyValueCharacter(ch) {
            return isUnicodePropertyNameCharacter(ch) || isDecimalDigit(ch);
        }
        // LoneUnicodePropertyNameOrValue ::
        //   UnicodePropertyValueCharacters
        pp$8.regexp_eatLoneUnicodePropertyNameOrValue = function (state) {
            return this.regexp_eatUnicodePropertyValue(state);
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-CharacterClass
        pp$8.regexp_eatCharacterClass = function (state) {
            if (state.eat(0x5B /* [ */)) {
                state.eat(0x5E /* ^ */);
                this.regexp_classRanges(state);
                if (state.eat(0x5D /* ] */)) {
                    return true;
                }
                // Unreachable since it threw "unterminated regular expression" error before.
                state.raise("Unterminated character class");
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-ClassRanges
        // https://www.ecma-international.org/ecma-262/8.0/#prod-NonemptyClassRanges
        // https://www.ecma-international.org/ecma-262/8.0/#prod-NonemptyClassRangesNoDash
        pp$8.regexp_classRanges = function (state) {
            while (this.regexp_eatClassAtom(state)) {
                var left = state.lastIntValue;
                if (state.eat(0x2D /* - */) && this.regexp_eatClassAtom(state)) {
                    var right = state.lastIntValue;
                    if (state.switchU && (left === -1 || right === -1)) {
                        state.raise("Invalid character class");
                    }
                    if (left !== -1 && right !== -1 && left > right) {
                        state.raise("Range out of order in character class");
                    }
                }
            }
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-ClassAtom
        // https://www.ecma-international.org/ecma-262/8.0/#prod-ClassAtomNoDash
        pp$8.regexp_eatClassAtom = function (state) {
            var start = state.pos;
            if (state.eat(0x5C /* \ */)) {
                if (this.regexp_eatClassEscape(state)) {
                    return true;
                }
                if (state.switchU) {
                    // Make the same message as V8.
                    var ch$1 = state.current();
                    if (ch$1 === 0x63 /* c */ || isOctalDigit(ch$1)) {
                        state.raise("Invalid class escape");
                    }
                    state.raise("Invalid escape");
                }
                state.pos = start;
            }
            var ch = state.current();
            if (ch !== 0x5D /* ] */) {
                state.lastIntValue = ch;
                state.advance();
                return true;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-ClassEscape
        pp$8.regexp_eatClassEscape = function (state) {
            var start = state.pos;
            if (state.eat(0x62 /* b */)) {
                state.lastIntValue = 0x08; /* <BS> */
                return true;
            }
            if (state.switchU && state.eat(0x2D /* - */)) {
                state.lastIntValue = 0x2D; /* - */
                return true;
            }
            if (!state.switchU && state.eat(0x63 /* c */)) {
                if (this.regexp_eatClassControlLetter(state)) {
                    return true;
                }
                state.pos = start;
            }
            return (this.regexp_eatCharacterClassEscape(state) ||
                this.regexp_eatCharacterEscape(state));
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-ClassControlLetter
        pp$8.regexp_eatClassControlLetter = function (state) {
            var ch = state.current();
            if (isDecimalDigit(ch) || ch === 0x5F /* _ */) {
                state.lastIntValue = ch % 0x20;
                state.advance();
                return true;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-HexEscapeSequence
        pp$8.regexp_eatHexEscapeSequence = function (state) {
            var start = state.pos;
            if (state.eat(0x78 /* x */)) {
                if (this.regexp_eatFixedHexDigits(state, 2)) {
                    return true;
                }
                if (state.switchU) {
                    state.raise("Invalid escape");
                }
                state.pos = start;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-DecimalDigits
        pp$8.regexp_eatDecimalDigits = function (state) {
            var start = state.pos;
            var ch = 0;
            state.lastIntValue = 0;
            while (isDecimalDigit(ch = state.current())) {
                state.lastIntValue = 10 * state.lastIntValue + (ch - 0x30 /* 0 */);
                state.advance();
            }
            return state.pos !== start;
        };
        function isDecimalDigit(ch) {
            return ch >= 0x30 /* 0 */ && ch <= 0x39; /* 9 */
        }
        // https://www.ecma-international.org/ecma-262/8.0/#prod-HexDigits
        pp$8.regexp_eatHexDigits = function (state) {
            var start = state.pos;
            var ch = 0;
            state.lastIntValue = 0;
            while (isHexDigit(ch = state.current())) {
                state.lastIntValue = 16 * state.lastIntValue + hexToInt(ch);
                state.advance();
            }
            return state.pos !== start;
        };
        function isHexDigit(ch) {
            return ((ch >= 0x30 /* 0 */ && ch <= 0x39 /* 9 */) ||
                (ch >= 0x41 /* A */ && ch <= 0x46 /* F */) ||
                (ch >= 0x61 /* a */ && ch <= 0x66 /* f */));
        }
        function hexToInt(ch) {
            if (ch >= 0x41 /* A */ && ch <= 0x46 /* F */) {
                return 10 + (ch - 0x41 /* A */);
            }
            if (ch >= 0x61 /* a */ && ch <= 0x66 /* f */) {
                return 10 + (ch - 0x61 /* a */);
            }
            return ch - 0x30; /* 0 */
        }
        // https://www.ecma-international.org/ecma-262/8.0/#prod-annexB-LegacyOctalEscapeSequence
        // Allows only 0-377(octal) i.e. 0-255(decimal).
        pp$8.regexp_eatLegacyOctalEscapeSequence = function (state) {
            if (this.regexp_eatOctalDigit(state)) {
                var n1 = state.lastIntValue;
                if (this.regexp_eatOctalDigit(state)) {
                    var n2 = state.lastIntValue;
                    if (n1 <= 3 && this.regexp_eatOctalDigit(state)) {
                        state.lastIntValue = n1 * 64 + n2 * 8 + state.lastIntValue;
                    }
                    else {
                        state.lastIntValue = n1 * 8 + n2;
                    }
                }
                else {
                    state.lastIntValue = n1;
                }
                return true;
            }
            return false;
        };
        // https://www.ecma-international.org/ecma-262/8.0/#prod-OctalDigit
        pp$8.regexp_eatOctalDigit = function (state) {
            var ch = state.current();
            if (isOctalDigit(ch)) {
                state.lastIntValue = ch - 0x30; /* 0 */
                state.advance();
                return true;
            }
            state.lastIntValue = 0;
            return false;
        };
        function isOctalDigit(ch) {
            return ch >= 0x30 /* 0 */ && ch <= 0x37; /* 7 */
        }
        // https://www.ecma-international.org/ecma-262/8.0/#prod-Hex4Digits
        // https://www.ecma-international.org/ecma-262/8.0/#prod-HexDigit
        // And HexDigit HexDigit in https://www.ecma-international.org/ecma-262/8.0/#prod-HexEscapeSequence
        pp$8.regexp_eatFixedHexDigits = function (state, length) {
            var start = state.pos;
            state.lastIntValue = 0;
            for (var i = 0; i < length; ++i) {
                var ch = state.current();
                if (!isHexDigit(ch)) {
                    state.pos = start;
                    return false;
                }
                state.lastIntValue = 16 * state.lastIntValue + hexToInt(ch);
                state.advance();
            }
            return true;
        };
        // Object type used to represent tokens. Note that normally, tokens
        // simply exist as properties on the parser object. This is only
        // used for the onToken callback and the external tokenizer.
        var Token = function Token(p) {
            this.type = p.type;
            this.value = p.value;
            this.start = p.start;
            this.end = p.end;
            if (p.options.locations) {
                this.loc = new SourceLocation(p, p.startLoc, p.endLoc);
            }
            if (p.options.ranges) {
                this.range = [p.start, p.end];
            }
        };
        // ## Tokenizer
        var pp$9 = Parser.prototype;
        // Move to the next token
        pp$9.next = function (ignoreEscapeSequenceInKeyword) {
            if (!ignoreEscapeSequenceInKeyword && this.type.keyword && this.containsEsc) {
                this.raiseRecoverable(this.start, "Escape sequence in keyword " + this.type.keyword);
            }
            if (this.options.onToken) {
                this.options.onToken(new Token(this));
            }
            this.lastTokEnd = this.end;
            this.lastTokStart = this.start;
            this.lastTokEndLoc = this.endLoc;
            this.lastTokStartLoc = this.startLoc;
            this.nextToken();
        };
        pp$9.getToken = function () {
            this.next();
            return new Token(this);
        };
        // If we're in an ES6 environment, make parsers iterable
        if (typeof Symbol !== "undefined") {
            pp$9[Symbol.iterator] = function () {
                var this$1 = this;
                return {
                    next: function () {
                        var token = this$1.getToken();
                        return {
                            done: token.type === types.eof,
                            value: token
                        };
                    }
                };
            };
        }
        // Toggle strict mode. Re-reads the next number or string to please
        // pedantic tests (`"use strict"; 010;` should fail).
        pp$9.curContext = function () {
            return this.context[this.context.length - 1];
        };
        // Read a single token, updating the parser object's token-related
        // properties.
        pp$9.nextToken = function () {
            var curContext = this.curContext();
            if (!curContext || !curContext.preserveSpace) {
                this.skipSpace();
            }
            this.start = this.pos;
            if (this.options.locations) {
                this.startLoc = this.curPosition();
            }
            if (this.pos >= this.input.length) {
                return this.finishToken(types.eof);
            }
            if (curContext.override) {
                return curContext.override(this);
            }
            else {
                this.readToken(this.fullCharCodeAtPos());
            }
        };
        pp$9.readToken = function (code) {
            // Identifier or keyword. '\uXXXX' sequences are allowed in
            // identifiers, so '\' also dispatches to that.
            if (isIdentifierStart(code, this.options.ecmaVersion >= 6) || code === 92 /* '\' */) {
                return this.readWord();
            }
            return this.getTokenFromCode(code);
        };
        pp$9.fullCharCodeAtPos = function () {
            var code = this.input.charCodeAt(this.pos);
            if (code <= 0xd7ff || code >= 0xdc00) {
                return code;
            }
            var next = this.input.charCodeAt(this.pos + 1);
            return next <= 0xdbff || next >= 0xe000 ? code : (code << 10) + next - 0x35fdc00;
        };
        pp$9.skipBlockComment = function () {
            var startLoc = this.options.onComment && this.curPosition();
            var start = this.pos, end = this.input.indexOf("*/", this.pos += 2);
            if (end === -1) {
                this.raise(this.pos - 2, "Unterminated comment");
            }
            this.pos = end + 2;
            if (this.options.locations) {
                lineBreakG.lastIndex = start;
                var match;
                while ((match = lineBreakG.exec(this.input)) && match.index < this.pos) {
                    ++this.curLine;
                    this.lineStart = match.index + match[0].length;
                }
            }
            if (this.options.onComment) {
                this.options.onComment(true, this.input.slice(start + 2, end), start, this.pos, startLoc, this.curPosition());
            }
        };
        pp$9.skipLineComment = function (startSkip) {
            var start = this.pos;
            var startLoc = this.options.onComment && this.curPosition();
            var ch = this.input.charCodeAt(this.pos += startSkip);
            while (this.pos < this.input.length && !isNewLine(ch)) {
                ch = this.input.charCodeAt(++this.pos);
            }
            if (this.options.onComment) {
                this.options.onComment(false, this.input.slice(start + startSkip, this.pos), start, this.pos, startLoc, this.curPosition());
            }
        };
        // Called at the start of the parse and after every token. Skips
        // whitespace and comments, and.
        pp$9.skipSpace = function () {
            loop: while (this.pos < this.input.length) {
                var ch = this.input.charCodeAt(this.pos);
                switch (ch) {
                    case 32:
                    case 160: // ' '
                        ++this.pos;
                        break;
                    case 13:
                        if (this.input.charCodeAt(this.pos + 1) === 10) {
                            ++this.pos;
                        }
                    case 10:
                    case 8232:
                    case 8233:
                        ++this.pos;
                        if (this.options.locations) {
                            ++this.curLine;
                            this.lineStart = this.pos;
                        }
                        break;
                    case 47: // '/'
                        switch (this.input.charCodeAt(this.pos + 1)) {
                            case 42: // '*'
                                this.skipBlockComment();
                                break;
                            case 47:
                                this.skipLineComment(2);
                                break;
                            default:
                                break loop;
                        }
                        break;
                    default:
                        if (ch > 8 && ch < 14 || ch >= 5760 && nonASCIIwhitespace.test(String.fromCharCode(ch))) {
                            ++this.pos;
                        }
                        else {
                            break loop;
                        }
                }
            }
        };
        // Called at the end of every token. Sets `end`, `val`, and
        // maintains `context` and `exprAllowed`, and skips the space after
        // the token, so that the next one's `start` will point at the
        // right position.
        pp$9.finishToken = function (type, val) {
            this.end = this.pos;
            if (this.options.locations) {
                this.endLoc = this.curPosition();
            }
            var prevType = this.type;
            this.type = type;
            this.value = val;
            this.updateContext(prevType);
        };
        // ### Token reading
        // This is the function that is called to fetch the next token. It
        // is somewhat obscure, because it works in character codes rather
        // than characters, and because operator parsing has been inlined
        // into it.
        //
        // All in the name of speed.
        //
        pp$9.readToken_dot = function () {
            var next = this.input.charCodeAt(this.pos + 1);
            if (next >= 48 && next <= 57) {
                return this.readNumber(true);
            }
            var next2 = this.input.charCodeAt(this.pos + 2);
            if (this.options.ecmaVersion >= 6 && next === 46 && next2 === 46) { // 46 = dot '.'
                this.pos += 3;
                return this.finishToken(types.ellipsis);
            }
            else {
                ++this.pos;
                return this.finishToken(types.dot);
            }
        };
        pp$9.readToken_slash = function () {
            var next = this.input.charCodeAt(this.pos + 1);
            if (this.exprAllowed) {
                ++this.pos;
                return this.readRegexp();
            }
            if (next === 61) {
                return this.finishOp(types.assign, 2);
            }
            return this.finishOp(types.slash, 1);
        };
        pp$9.readToken_mult_modulo_exp = function (code) {
            var next = this.input.charCodeAt(this.pos + 1);
            var size = 1;
            var tokentype = code === 42 ? types.star : types.modulo;
            // exponentiation operator ** and **=
            if (this.options.ecmaVersion >= 7 && code === 42 && next === 42) {
                ++size;
                tokentype = types.starstar;
                next = this.input.charCodeAt(this.pos + 2);
            }
            if (next === 61) {
                return this.finishOp(types.assign, size + 1);
            }
            return this.finishOp(tokentype, size);
        };
        pp$9.readToken_pipe_amp = function (code) {
            var next = this.input.charCodeAt(this.pos + 1);
            if (next === code) {
                if (this.options.ecmaVersion >= 12) {
                    var next2 = this.input.charCodeAt(this.pos + 2);
                    if (next2 === 61) {
                        return this.finishOp(types.assign, 3);
                    }
                }
                return this.finishOp(code === 124 ? types.logicalOR : types.logicalAND, 2);
            }
            if (next === 61) {
                return this.finishOp(types.assign, 2);
            }
            return this.finishOp(code === 124 ? types.bitwiseOR : types.bitwiseAND, 1);
        };
        pp$9.readToken_caret = function () {
            var next = this.input.charCodeAt(this.pos + 1);
            if (next === 61) {
                return this.finishOp(types.assign, 2);
            }
            return this.finishOp(types.bitwiseXOR, 1);
        };
        pp$9.readToken_plus_min = function (code) {
            var next = this.input.charCodeAt(this.pos + 1);
            if (next === code) {
                if (next === 45 && !this.inModule && this.input.charCodeAt(this.pos + 2) === 62 &&
                    (this.lastTokEnd === 0 || lineBreak.test(this.input.slice(this.lastTokEnd, this.pos)))) {
                    // A `-->` line comment
                    this.skipLineComment(3);
                    this.skipSpace();
                    return this.nextToken();
                }
                return this.finishOp(types.incDec, 2);
            }
            if (next === 61) {
                return this.finishOp(types.assign, 2);
            }
            return this.finishOp(types.plusMin, 1);
        };
        pp$9.readToken_lt_gt = function (code) {
            var next = this.input.charCodeAt(this.pos + 1);
            var size = 1;
            if (next === code) {
                size = code === 62 && this.input.charCodeAt(this.pos + 2) === 62 ? 3 : 2;
                if (this.input.charCodeAt(this.pos + size) === 61) {
                    return this.finishOp(types.assign, size + 1);
                }
                return this.finishOp(types.bitShift, size);
            }
            if (next === 33 && code === 60 && !this.inModule && this.input.charCodeAt(this.pos + 2) === 45 &&
                this.input.charCodeAt(this.pos + 3) === 45) {
                // `<!--`, an XML-style comment that should be interpreted as a line comment
                this.skipLineComment(4);
                this.skipSpace();
                return this.nextToken();
            }
            if (next === 61) {
                size = 2;
            }
            return this.finishOp(types.relational, size);
        };
        pp$9.readToken_eq_excl = function (code) {
            var next = this.input.charCodeAt(this.pos + 1);
            if (next === 61) {
                return this.finishOp(types.equality, this.input.charCodeAt(this.pos + 2) === 61 ? 3 : 2);
            }
            if (code === 61 && next === 62 && this.options.ecmaVersion >= 6) { // '=>'
                this.pos += 2;
                return this.finishToken(types.arrow);
            }
            return this.finishOp(code === 61 ? types.eq : types.prefix, 1);
        };
        pp$9.readToken_question = function () {
            var ecmaVersion = this.options.ecmaVersion;
            if (ecmaVersion >= 11) {
                var next = this.input.charCodeAt(this.pos + 1);
                if (next === 46) {
                    var next2 = this.input.charCodeAt(this.pos + 2);
                    if (next2 < 48 || next2 > 57) {
                        return this.finishOp(types.questionDot, 2);
                    }
                }
                if (next === 63) {
                    if (ecmaVersion >= 12) {
                        var next2$1 = this.input.charCodeAt(this.pos + 2);
                        if (next2$1 === 61) {
                            return this.finishOp(types.assign, 3);
                        }
                    }
                    return this.finishOp(types.coalesce, 2);
                }
            }
            return this.finishOp(types.question, 1);
        };
        pp$9.readToken_numberSign = function () {
            var ecmaVersion = this.options.ecmaVersion;
            var code = 35; // '#'
            if (ecmaVersion >= 13) {
                ++this.pos;
                code = this.fullCharCodeAtPos();
                if (isIdentifierStart(code, true) || code === 92 /* '\' */) {
                    return this.finishToken(types.privateId, this.readWord1());
                }
            }
            this.raise(this.pos, "Unexpected character '" + codePointToString$1(code) + "'");
        };
        pp$9.getTokenFromCode = function (code) {
            switch (code) {
                // The interpretation of a dot depends on whether it is followed
                // by a digit or another two dots.
                case 46: // '.'
                    return this.readToken_dot();
                // Punctuation tokens.
                case 40:
                    ++this.pos;
                    return this.finishToken(types.parenL);
                case 41:
                    ++this.pos;
                    return this.finishToken(types.parenR);
                case 59:
                    ++this.pos;
                    return this.finishToken(types.semi);
                case 44:
                    ++this.pos;
                    return this.finishToken(types.comma);
                case 91:
                    ++this.pos;
                    return this.finishToken(types.bracketL);
                case 93:
                    ++this.pos;
                    return this.finishToken(types.bracketR);
                case 123:
                    ++this.pos;
                    return this.finishToken(types.braceL);
                case 125:
                    ++this.pos;
                    return this.finishToken(types.braceR);
                case 58:
                    ++this.pos;
                    return this.finishToken(types.colon);
                case 96: // '`'
                    if (this.options.ecmaVersion < 6) {
                        break;
                    }
                    ++this.pos;
                    return this.finishToken(types.backQuote);
                case 48: // '0'
                    var next = this.input.charCodeAt(this.pos + 1);
                    if (next === 120 || next === 88) {
                        return this.readRadixNumber(16);
                    } // '0x', '0X' - hex number
                    if (this.options.ecmaVersion >= 6) {
                        if (next === 111 || next === 79) {
                            return this.readRadixNumber(8);
                        } // '0o', '0O' - octal number
                        if (next === 98 || next === 66) {
                            return this.readRadixNumber(2);
                        } // '0b', '0B' - binary number
                    }
                // Anything else beginning with a digit is an integer, octal
                // number, or float.
                case 49:
                case 50:
                case 51:
                case 52:
                case 53:
                case 54:
                case 55:
                case 56:
                case 57: // 1-9
                    return this.readNumber(false);
                // Quotes produce strings.
                case 34:
                case 39: // '"', "'"
                    return this.readString(code);
                // Operators are parsed inline in tiny state machines. '=' (61) is
                // often referred to. `finishOp` simply skips the amount of
                // characters it is given as second argument, and returns a token
                // of the type given by its first argument.
                case 47: // '/'
                    return this.readToken_slash();
                case 37:
                case 42: // '%*'
                    return this.readToken_mult_modulo_exp(code);
                case 124:
                case 38: // '|&'
                    return this.readToken_pipe_amp(code);
                case 94: // '^'
                    return this.readToken_caret();
                case 43:
                case 45: // '+-'
                    return this.readToken_plus_min(code);
                case 60:
                case 62: // '<>'
                    return this.readToken_lt_gt(code);
                case 61:
                case 33: // '=!'
                    return this.readToken_eq_excl(code);
                case 63: // '?'
                    return this.readToken_question();
                case 126: // '~'
                    return this.finishOp(types.prefix, 1);
                case 35: // '#'
                    return this.readToken_numberSign();
            }
            this.raise(this.pos, "Unexpected character '" + codePointToString$1(code) + "'");
        };
        pp$9.finishOp = function (type, size) {
            var str = this.input.slice(this.pos, this.pos + size);
            this.pos += size;
            return this.finishToken(type, str);
        };
        pp$9.readRegexp = function () {
            var escaped, inClass, start = this.pos;
            for (;;) {
                if (this.pos >= this.input.length) {
                    this.raise(start, "Unterminated regular expression");
                }
                var ch = this.input.charAt(this.pos);
                if (lineBreak.test(ch)) {
                    this.raise(start, "Unterminated regular expression");
                }
                if (!escaped) {
                    if (ch === "[") {
                        inClass = true;
                    }
                    else if (ch === "]" && inClass) {
                        inClass = false;
                    }
                    else if (ch === "/" && !inClass) {
                        break;
                    }
                    escaped = ch === "\\";
                }
                else {
                    escaped = false;
                }
                ++this.pos;
            }
            var pattern = this.input.slice(start, this.pos);
            ++this.pos;
            var flagsStart = this.pos;
            var flags = this.readWord1();
            if (this.containsEsc) {
                this.unexpected(flagsStart);
            }
            // Validate pattern
            var state = this.regexpState || (this.regexpState = new RegExpValidationState(this));
            state.reset(start, pattern, flags);
            this.validateRegExpFlags(state);
            this.validateRegExpPattern(state);
            // Create Literal#value property value.
            var value = null;
            try {
                value = new RegExp(pattern, flags);
            }
            catch (e) {
                // ESTree requires null if it failed to instantiate RegExp object.
                // https://github.com/estree/estree/blob/a27003adf4fd7bfad44de9cef372a2eacd527b1c/es5.md#regexpliteral
            }
            return this.finishToken(types.regexp, { pattern: pattern, flags: flags, value: value });
        };
        // Read an integer in the given radix. Return null if zero digits
        // were read, the integer value otherwise. When `len` is given, this
        // will return `null` unless the integer has exactly `len` digits.
        pp$9.readInt = function (radix, len, maybeLegacyOctalNumericLiteral) {
            // `len` is used for character escape sequences. In that case, disallow separators.
            var allowSeparators = this.options.ecmaVersion >= 12 && len === undefined;
            // `maybeLegacyOctalNumericLiteral` is true if it doesn't have prefix (0x,0o,0b)
            // and isn't fraction part nor exponent part. In that case, if the first digit
            // is zero then disallow separators.
            var isLegacyOctalNumericLiteral = maybeLegacyOctalNumericLiteral && this.input.charCodeAt(this.pos) === 48;
            var start = this.pos, total = 0, lastCode = 0;
            for (var i = 0, e = len == null ? Infinity : len; i < e; ++i, ++this.pos) {
                var code = this.input.charCodeAt(this.pos), val = (void 0);
                if (allowSeparators && code === 95) {
                    if (isLegacyOctalNumericLiteral) {
                        this.raiseRecoverable(this.pos, "Numeric separator is not allowed in legacy octal numeric literals");
                    }
                    if (lastCode === 95) {
                        this.raiseRecoverable(this.pos, "Numeric separator must be exactly one underscore");
                    }
                    if (i === 0) {
                        this.raiseRecoverable(this.pos, "Numeric separator is not allowed at the first of digits");
                    }
                    lastCode = code;
                    continue;
                }
                if (code >= 97) {
                    val = code - 97 + 10;
                } // a
                else if (code >= 65) {
                    val = code - 65 + 10;
                } // A
                else if (code >= 48 && code <= 57) {
                    val = code - 48;
                } // 0-9
                else {
                    val = Infinity;
                }
                if (val >= radix) {
                    break;
                }
                lastCode = code;
                total = total * radix + val;
            }
            if (allowSeparators && lastCode === 95) {
                this.raiseRecoverable(this.pos - 1, "Numeric separator is not allowed at the last of digits");
            }
            if (this.pos === start || len != null && this.pos - start !== len) {
                return null;
            }
            return total;
        };
        function stringToNumber(str, isLegacyOctalNumericLiteral) {
            if (isLegacyOctalNumericLiteral) {
                return parseInt(str, 8);
            }
            // `parseFloat(value)` stops parsing at the first numeric separator then returns a wrong value.
            return parseFloat(str.replace(/_/g, ""));
        }
        function stringToBigInt(str) {
            if (typeof BigInt !== "function") {
                return null;
            }
            // `BigInt(value)` throws syntax error if the string contains numeric separators.
            return BigInt(str.replace(/_/g, ""));
        }
        pp$9.readRadixNumber = function (radix) {
            var start = this.pos;
            this.pos += 2; // 0x
            var val = this.readInt(radix);
            if (val == null) {
                this.raise(this.start + 2, "Expected number in radix " + radix);
            }
            if (this.options.ecmaVersion >= 11 && this.input.charCodeAt(this.pos) === 110) {
                val = stringToBigInt(this.input.slice(start, this.pos));
                ++this.pos;
            }
            else if (isIdentifierStart(this.fullCharCodeAtPos())) {
                this.raise(this.pos, "Identifier directly after number");
            }
            return this.finishToken(types.num, val);
        };
        // Read an integer, octal integer, or floating-point number.
        pp$9.readNumber = function (startsWithDot) {
            var start = this.pos;
            if (!startsWithDot && this.readInt(10, undefined, true) === null) {
                this.raise(start, "Invalid number");
            }
            var octal = this.pos - start >= 2 && this.input.charCodeAt(start) === 48;
            if (octal && this.strict) {
                this.raise(start, "Invalid number");
            }
            var next = this.input.charCodeAt(this.pos);
            if (!octal && !startsWithDot && this.options.ecmaVersion >= 11 && next === 110) {
                var val$1 = stringToBigInt(this.input.slice(start, this.pos));
                ++this.pos;
                if (isIdentifierStart(this.fullCharCodeAtPos())) {
                    this.raise(this.pos, "Identifier directly after number");
                }
                return this.finishToken(types.num, val$1);
            }
            if (octal && /[89]/.test(this.input.slice(start, this.pos))) {
                octal = false;
            }
            if (next === 46 && !octal) { // '.'
                ++this.pos;
                this.readInt(10);
                next = this.input.charCodeAt(this.pos);
            }
            if ((next === 69 || next === 101) && !octal) { // 'eE'
                next = this.input.charCodeAt(++this.pos);
                if (next === 43 || next === 45) {
                    ++this.pos;
                } // '+-'
                if (this.readInt(10) === null) {
                    this.raise(start, "Invalid number");
                }
            }
            if (isIdentifierStart(this.fullCharCodeAtPos())) {
                this.raise(this.pos, "Identifier directly after number");
            }
            var val = stringToNumber(this.input.slice(start, this.pos), octal);
            return this.finishToken(types.num, val);
        };
        // Read a string value, interpreting backslash-escapes.
        pp$9.readCodePoint = function () {
            var ch = this.input.charCodeAt(this.pos), code;
            if (ch === 123) { // '{'
                if (this.options.ecmaVersion < 6) {
                    this.unexpected();
                }
                var codePos = ++this.pos;
                code = this.readHexChar(this.input.indexOf("}", this.pos) - this.pos);
                ++this.pos;
                if (code > 0x10FFFF) {
                    this.invalidStringToken(codePos, "Code point out of bounds");
                }
            }
            else {
                code = this.readHexChar(4);
            }
            return code;
        };
        function codePointToString$1(code) {
            // UTF-16 Decoding
            if (code <= 0xFFFF) {
                return String.fromCharCode(code);
            }
            code -= 0x10000;
            return String.fromCharCode((code >> 10) + 0xD800, (code & 1023) + 0xDC00);
        }
        pp$9.readString = function (quote) {
            var out = "", chunkStart = ++this.pos;
            for (;;) {
                if (this.pos >= this.input.length) {
                    this.raise(this.start, "Unterminated string constant");
                }
                var ch = this.input.charCodeAt(this.pos);
                if (ch === quote) {
                    break;
                }
                if (ch === 92) { // '\'
                    out += this.input.slice(chunkStart, this.pos);
                    out += this.readEscapedChar(false);
                    chunkStart = this.pos;
                }
                else {
                    if (isNewLine(ch, this.options.ecmaVersion >= 10)) {
                        this.raise(this.start, "Unterminated string constant");
                    }
                    ++this.pos;
                }
            }
            out += this.input.slice(chunkStart, this.pos++);
            return this.finishToken(types.string, out);
        };
        // Reads template string tokens.
        var INVALID_TEMPLATE_ESCAPE_ERROR = {};
        pp$9.tryReadTemplateToken = function () {
            this.inTemplateElement = true;
            try {
                this.readTmplToken();
            }
            catch (err) {
                if (err === INVALID_TEMPLATE_ESCAPE_ERROR) {
                    this.readInvalidTemplateToken();
                }
                else {
                    throw err;
                }
            }
            this.inTemplateElement = false;
        };
        pp$9.invalidStringToken = function (position, message) {
            if (this.inTemplateElement && this.options.ecmaVersion >= 9) {
                throw INVALID_TEMPLATE_ESCAPE_ERROR;
            }
            else {
                this.raise(position, message);
            }
        };
        pp$9.readTmplToken = function () {
            var out = "", chunkStart = this.pos;
            for (;;) {
                if (this.pos >= this.input.length) {
                    this.raise(this.start, "Unterminated template");
                }
                var ch = this.input.charCodeAt(this.pos);
                if (ch === 96 || ch === 36 && this.input.charCodeAt(this.pos + 1) === 123) { // '`', '${'
                    if (this.pos === this.start && (this.type === types.template || this.type === types.invalidTemplate)) {
                        if (ch === 36) {
                            this.pos += 2;
                            return this.finishToken(types.dollarBraceL);
                        }
                        else {
                            ++this.pos;
                            return this.finishToken(types.backQuote);
                        }
                    }
                    out += this.input.slice(chunkStart, this.pos);
                    return this.finishToken(types.template, out);
                }
                if (ch === 92) { // '\'
                    out += this.input.slice(chunkStart, this.pos);
                    out += this.readEscapedChar(true);
                    chunkStart = this.pos;
                }
                else if (isNewLine(ch)) {
                    out += this.input.slice(chunkStart, this.pos);
                    ++this.pos;
                    switch (ch) {
                        case 13:
                            if (this.input.charCodeAt(this.pos) === 10) {
                                ++this.pos;
                            }
                        case 10:
                            out += "\n";
                            break;
                        default:
                            out += String.fromCharCode(ch);
                            break;
                    }
                    if (this.options.locations) {
                        ++this.curLine;
                        this.lineStart = this.pos;
                    }
                    chunkStart = this.pos;
                }
                else {
                    ++this.pos;
                }
            }
        };
        // Reads a template token to search for the end, without validating any escape sequences
        pp$9.readInvalidTemplateToken = function () {
            for (; this.pos < this.input.length; this.pos++) {
                switch (this.input[this.pos]) {
                    case "\\":
                        ++this.pos;
                        break;
                    case "$":
                        if (this.input[this.pos + 1] !== "{") {
                            break;
                        }
                    // falls through
                    case "`":
                        return this.finishToken(types.invalidTemplate, this.input.slice(this.start, this.pos));
                    // no default
                }
            }
            this.raise(this.start, "Unterminated template");
        };
        // Used to read escaped characters
        pp$9.readEscapedChar = function (inTemplate) {
            var ch = this.input.charCodeAt(++this.pos);
            ++this.pos;
            switch (ch) {
                case 110: return "\n"; // 'n' -> '\n'
                case 114: return "\r"; // 'r' -> '\r'
                case 120: return String.fromCharCode(this.readHexChar(2)); // 'x'
                case 117: return codePointToString$1(this.readCodePoint()); // 'u'
                case 116: return "\t"; // 't' -> '\t'
                case 98: return "\b"; // 'b' -> '\b'
                case 118: return "\u000b"; // 'v' -> '\u000b'
                case 102: return "\f"; // 'f' -> '\f'
                case 13: if (this.input.charCodeAt(this.pos) === 10) {
                    ++this.pos;
                } // '\r\n'
                case 10: // ' \n'
                    if (this.options.locations) {
                        this.lineStart = this.pos;
                        ++this.curLine;
                    }
                    return "";
                case 56:
                case 57:
                    if (this.strict) {
                        this.invalidStringToken(this.pos - 1, "Invalid escape sequence");
                    }
                    if (inTemplate) {
                        var codePos = this.pos - 1;
                        this.invalidStringToken(codePos, "Invalid escape sequence in template string");
                        return null;
                    }
                default:
                    if (ch >= 48 && ch <= 55) {
                        var octalStr = this.input.substr(this.pos - 1, 3).match(/^[0-7]+/)[0];
                        var octal = parseInt(octalStr, 8);
                        if (octal > 255) {
                            octalStr = octalStr.slice(0, -1);
                            octal = parseInt(octalStr, 8);
                        }
                        this.pos += octalStr.length - 1;
                        ch = this.input.charCodeAt(this.pos);
                        if ((octalStr !== "0" || ch === 56 || ch === 57) && (this.strict || inTemplate)) {
                            this.invalidStringToken(this.pos - 1 - octalStr.length, inTemplate
                                ? "Octal literal in template string"
                                : "Octal literal in strict mode");
                        }
                        return String.fromCharCode(octal);
                    }
                    if (isNewLine(ch)) {
                        // Unicode new line characters after \ get removed from output in both
                        // template literals and strings
                        return "";
                    }
                    return String.fromCharCode(ch);
            }
        };
        // Used to read character escape sequences ('\x', '\u', '\U').
        pp$9.readHexChar = function (len) {
            var codePos = this.pos;
            var n = this.readInt(16, len);
            if (n === null) {
                this.invalidStringToken(codePos, "Bad character escape sequence");
            }
            return n;
        };
        // Read an identifier, and return it as a string. Sets `this.containsEsc`
        // to whether the word contained a '\u' escape.
        //
        // Incrementally adds only escaped chars, adding other chunks as-is
        // as a micro-optimization.
        pp$9.readWord1 = function () {
            this.containsEsc = false;
            var word = "", first = true, chunkStart = this.pos;
            var astral = this.options.ecmaVersion >= 6;
            while (this.pos < this.input.length) {
                var ch = this.fullCharCodeAtPos();
                if (isIdentifierChar(ch, astral)) {
                    this.pos += ch <= 0xffff ? 1 : 2;
                }
                else if (ch === 92) { // "\"
                    this.containsEsc = true;
                    word += this.input.slice(chunkStart, this.pos);
                    var escStart = this.pos;
                    if (this.input.charCodeAt(++this.pos) !== 117) // "u"
                     {
                        this.invalidStringToken(this.pos, "Expecting Unicode escape sequence \\uXXXX");
                    }
                    ++this.pos;
                    var esc = this.readCodePoint();
                    if (!(first ? isIdentifierStart : isIdentifierChar)(esc, astral)) {
                        this.invalidStringToken(escStart, "Invalid Unicode escape");
                    }
                    word += codePointToString$1(esc);
                    chunkStart = this.pos;
                }
                else {
                    break;
                }
                first = false;
            }
            return word + this.input.slice(chunkStart, this.pos);
        };
        // Read an identifier or keyword token. Will check for reserved
        // words when necessary.
        pp$9.readWord = function () {
            var word = this.readWord1();
            var type = types.name;
            if (this.keywords.test(word)) {
                type = keywords$1[word];
            }
            return this.finishToken(type, word);
        };
        // Acorn is a tiny, fast JavaScript parser written in JavaScript.
        var version = "8.4.1";
        Parser.acorn = {
            Parser: Parser,
            version: version,
            defaultOptions: defaultOptions,
            Position: Position,
            SourceLocation: SourceLocation,
            getLineInfo: getLineInfo,
            Node: Node,
            TokenType: TokenType,
            tokTypes: types,
            keywordTypes: keywords$1,
            TokContext: TokContext,
            tokContexts: types$1,
            isIdentifierChar: isIdentifierChar,
            isIdentifierStart: isIdentifierStart,
            Token: Token,
            isNewLine: isNewLine,
            lineBreak: lineBreak,
            lineBreakG: lineBreakG,
            nonASCIIwhitespace: nonASCIIwhitespace
        };
        // The main exported interface (under `self.acorn` when in the
        // browser) is a `parse` function that takes a code string and
        // returns an abstract syntax tree as specified by [Mozilla parser
        // API][api].
        //
        // [api]: https://developer.mozilla.org/en-US/docs/SpiderMonkey/Parser_API
        function parse(input, options) {
            return Parser.parse(input, options);
        }
        // This function tries to parse a single expression at a given
        // offset in a string. Useful for parsing mixed-language formats
        // that embed JavaScript expressions.
        function parseExpressionAt(input, pos, options) {
            return Parser.parseExpressionAt(input, pos, options);
        }
        // Acorn is organized as a tokenizer and a recursive-descent parser.
        // The `tokenizer` export provides an interface to the tokenizer.
        function tokenizer(input, options) {
            return Parser.tokenizer(input, options);
        }
        exports.Node = Node;
        exports.Parser = Parser;
        exports.Position = Position;
        exports.SourceLocation = SourceLocation;
        exports.TokContext = TokContext;
        exports.Token = Token;
        exports.TokenType = TokenType;
        exports.defaultOptions = defaultOptions;
        exports.getLineInfo = getLineInfo;
        exports.isIdentifierChar = isIdentifierChar;
        exports.isIdentifierStart = isIdentifierStart;
        exports.isNewLine = isNewLine;
        exports.keywordTypes = keywords$1;
        exports.lineBreak = lineBreak;
        exports.lineBreakG = lineBreakG;
        exports.nonASCIIwhitespace = nonASCIIwhitespace;
        exports.parse = parse;
        exports.parseExpressionAt = parseExpressionAt;
        exports.tokContexts = types$1;
        exports.tokTypes = types;
        exports.tokenizer = tokenizer;
        exports.version = version;
        Object.defineProperty(exports, '__esModule', { value: true });
    })));
},
/* combine-source-map/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var path = require('path');
    var convert = require(48) /* convert-source-map */;
    var memoize = require(50) /* lodash.memoize */;
    var createGenerator = require(51) /* inline-source-map */;
    var pathIsAbsolute = require(52) /* ./lib/path-is-absolute */;
    var mappingsFromMap = require(53) /* ./lib/mappings-from-map */;
    var protocolRx = /^[a-z]+:\/\//;
    /**
     * Rebases a relative path in 'sourceFile' to be relative
     * to the path where 'sourceFile' is located.
     *
     * This is necessary before adding relative paths to the
     * new combined map to ensure all paths are relative to their
     * original source.
     *
     * The 'sourceRoot' from the original source map is joined
     * as well to ensure the complete path.
     *
     * Resulting paths that are absolute are passed along directly.
     *
     * @param sourceFile {String} path to the original source file that references a map
     * @param relativeRoot {String} sourceRoot in sourceFile's map to combine with relativePath
     * @param relativePath {String} source path from sourceFile's map
     */
    var rebaseRelativePath = memoize(function (sourceFile, relativeRoot, relativePath) {
        if (!relativePath) {
            return relativePath;
        }
        // join relative path to root (e.g. 'src/' + 'file.js')
        var relativeRootedPath = relativeRoot ? path.join(relativeRoot, relativePath) : relativePath;
        relativeRootedPath = relativeRootedPath.replace(/\\/g, '/');
        sourceFile = sourceFile.replace(/\\/g, '/');
        if (sourceFile === relativeRootedPath || // same path,
            pathIsAbsolute(relativeRootedPath) || // absolute path, nor
            protocolRx.test(relativeRootedPath)) { // absolute protocol need rebasing
            return relativeRootedPath;
        }
        // make relative to source file
        return path.join(path.dirname(sourceFile), relativeRootedPath).replace(/\\/g, '/');
    }, function (a, b, c) {
        return a + '::' + b + '::' + c;
    });
    function resolveMap(source) {
        var gen = convert.fromSource(source);
        return gen ? gen.toObject() : null;
    }
    function hasInlinedSource(existingMap) {
        return existingMap.sourcesContent && !!existingMap.sourcesContent[0];
    }
    function Combiner(file, sourceRoot) {
        // since we include the original code in the map sourceRoot actually not needed
        this.generator = createGenerator({ file: file || 'generated.js', sourceRoot: sourceRoot });
    }
    Combiner.prototype._addGeneratedMap = function (sourceFile, source, offset) {
        this.generator.addGeneratedMappings(sourceFile, source, offset);
        this.generator.addSourceContent(sourceFile, source);
        return this;
    };
    Combiner.prototype._addExistingMap = function (sourceFile, source, existingMap, offset) {
        var mappings = mappingsFromMap(existingMap);
        // add all of the sources from the map
        for (var i = 0, len = existingMap.sources.length; i < len; i++) {
            if (!existingMap.sourcesContent)
                continue;
            this.generator.addSourceContent(rebaseRelativePath(sourceFile, existingMap.sourceRoot, existingMap.sources[i]), existingMap.sourcesContent[i]);
        }
        // add the mappings, preserving the original mapping 'source'
        mappings.forEach(function (mapping) {
            // Add the mappings one at a time because 'inline-source-map' doesn't handle
            // mapping source filenames. The mapping.source already takes sourceRoot into account
            // per the SMConsumer.eachMapping function, so pass null for the root here.
            this.generator.addMappings(rebaseRelativePath(sourceFile, null, mapping.source), [mapping], offset);
        }, this);
        return this;
    };
    /**
     * Adds map to underlying source map.
     * If source contains a source map comment that has the source of the original file inlined it will offset these
     * mappings and include them.
     * If no source map comment is found or it has no source inlined, mappings for the file will be generated and included
     *
     * @name addMap
     * @function
     * @param opts {Object} { sourceFile: {String}, source: {String} }
     * @param offset {Object} { line: {Number}, column: {Number} }
     */
    Combiner.prototype.addFile = function (opts, offset) {
        offset = offset || {};
        if (!offset.hasOwnProperty('line'))
            offset.line = 0;
        if (!offset.hasOwnProperty('column'))
            offset.column = 0;
        var existingMap = resolveMap(opts.source);
        return existingMap && hasInlinedSource(existingMap)
            ? this._addExistingMap(opts.sourceFile, opts.source, existingMap, offset)
            : this._addGeneratedMap(opts.sourceFile, opts.source, offset);
    };
    /**
    * @name base64
    * @function
    * @return {String} base64 encoded combined source map
    */
    Combiner.prototype.base64 = function () {
        return this.generator.base64Encode();
    };
    /**
     * @name comment
     * @function
     * @return {String} base64 encoded sourceMappingUrl comment of the combined source map
     */
    Combiner.prototype.comment = function () {
        return this.generator.inlineMappingUrl();
    };
    /**
     * @name create
     * @function
     * @param file {String} optional name of the generated file
     * @param sourceRoot {String} optional sourceRoot of the map to be generated
     * @return {Object} Combiner instance to which source maps can be added and later combined
     */
    exports.create = function (file, sourceRoot) { return new Combiner(file, sourceRoot); };
    /**
     * @name removeComments
     * @function
     * @param src
     * @return {String} src with all sourceMappingUrl comments removed
     */
    exports.removeComments = function (src) {
        if (!src.replace)
            return src;
        return src.replace(convert.commentRegex, '').replace(convert.mapFileCommentRegex, '');
    };
},
/* convert-source-map/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var fs = require('fs');
    var path = require('path');
    var SafeBuffer = require(49) /* safe-buffer */;
    __esExport('commentRegex', /^\s*\/(?:\/|\*)[@#]\s+sourceMappingURL=data:(?:application|text)\/json;(?:charset[:=]\S+?;)?base64,(?:.*)$/mg);
    __esExport('mapFileCommentRegex', /(?:\/\/[@#][ \t]+sourceMappingURL=([^\s'"`]+?)[ \t]*$)|(?:\/\*[@#][ \t]+sourceMappingURL=([^\*]+?)[ \t]*(?:\*\/){1}[ \t]*$)/mg);
    function decodeBase64(base64) {
        return (SafeBuffer.Buffer.from(base64, 'base64') || "").toString();
    }
    function stripComment(sm) {
        return sm.split(',').pop();
    }
    function readFromFileMap(sm, dir) {
        // NOTE: this will only work on the server since it attempts to read the map file
        var r = exports.mapFileCommentRegex.exec(sm);
        // for some odd reason //# .. captures in 1 and /* .. */ in 2
        var filename = r[1] || r[2];
        var filepath = path.resolve(dir, filename);
        try {
            return fs.readFileSync(filepath, 'utf8');
        }
        catch (e) {
            throw new Error('An error occurred while trying to read the map file at ' + filepath + '\n' + e);
        }
    }
    function Converter(sm, opts) {
        opts = opts || {};
        if (opts.isFileComment)
            sm = readFromFileMap(sm, opts.commentFileDir);
        if (opts.hasComment)
            sm = stripComment(sm);
        if (opts.isEncoded)
            sm = decodeBase64(sm);
        if (opts.isJSON || opts.isEncoded)
            sm = JSON.parse(sm);
        this.sourcemap = sm;
    }
    Converter.prototype.toJSON = function (space) {
        return JSON.stringify(this.sourcemap, null, space);
    };
    Converter.prototype.toBase64 = function () {
        var json = this.toJSON();
        return (SafeBuffer.Buffer.from(json, 'utf8') || "").toString('base64');
    };
    Converter.prototype.toComment = function (options) {
        var base64 = this.toBase64();
        var data = 'sourceMappingURL=data:application/json;charset=utf-8;base64,' + base64;
        return options && options.multiline ? '/*# ' + data + ' */' : '//# ' + data;
    };
    // returns copy instead of original
    Converter.prototype.toObject = function () {
        return JSON.parse(this.toJSON());
    };
    Converter.prototype.addProperty = function (key, value) {
        if (this.sourcemap.hasOwnProperty(key))
            throw new Error('property "' + key + '" already exists on the sourcemap, use set property instead');
        return this.setProperty(key, value);
    };
    Converter.prototype.setProperty = function (key, value) {
        this.sourcemap[key] = value;
        return this;
    };
    Converter.prototype.getProperty = function (key) {
        return this.sourcemap[key];
    };
    exports.fromObject = function (obj) {
        return new Converter(obj);
    };
    exports.fromJSON = function (json) {
        return new Converter(json, { isJSON: true });
    };
    exports.fromBase64 = function (base64) {
        return new Converter(base64, { isEncoded: true });
    };
    exports.fromComment = function (comment) {
        comment = comment
            .replace(/^\/\*/g, '//')
            .replace(/\*\/$/g, '');
        return new Converter(comment, { isEncoded: true, hasComment: true });
    };
    exports.fromMapFileComment = function (comment, dir) {
        return new Converter(comment, { commentFileDir: dir, isFileComment: true, isJSON: true });
    };
    // Finds last sourcemap comment in file or returns null if none was found
    exports.fromSource = function (content) {
        var m = content.match(exports.commentRegex);
        return m ? exports.fromComment(m.pop()) : null;
    };
    // Finds last sourcemap comment in file or returns null if none was found
    exports.fromMapFileSource = function (content, dir) {
        var m = content.match(exports.mapFileCommentRegex);
        return m ? exports.fromMapFileComment(m.pop(), dir) : null;
    };
    exports.removeComments = function (src) {
        return src.replace(exports.commentRegex, '');
    };
    exports.removeMapFileComments = function (src) {
        return src.replace(exports.mapFileCommentRegex, '');
    };
    exports.generateMapFileComment = function (file, options) {
        var data = 'sourceMappingURL=' + file;
        return options && options.multiline ? '/*# ' + data + ' */' : '//# ' + data;
    };
},
/* safe-buffer/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint-disable node/no-deprecated-api */
    var buffer = require('buffer');
    var Buffer = buffer.Buffer;
    // alternative to using Object.keys for old browsers
    function copyProps(src, dst) {
        for (var key in src) {
            dst[key] = src[key];
        }
    }
    if (Buffer.from && Buffer.alloc && Buffer.allocUnsafe && Buffer.allocUnsafeSlow) {
        module.exports = buffer;
    }
    else {
        // Copy properties from require('buffer')
        copyProps(buffer, exports);
        exports.Buffer = SafeBuffer;
    }
    function SafeBuffer(arg, encodingOrOffset, length) {
        return Buffer(arg, encodingOrOffset, length);
    }
    // Copy static methods from Buffer
    copyProps(Buffer, SafeBuffer);
    SafeBuffer.from = function (arg, encodingOrOffset, length) {
        if (typeof arg === 'number') {
            throw new TypeError('Argument must not be a number');
        }
        return Buffer(arg, encodingOrOffset, length);
    };
    SafeBuffer.alloc = function (size, fill, encoding) {
        if (typeof size !== 'number') {
            throw new TypeError('Argument must be a number');
        }
        var buf = Buffer(size);
        if (fill !== undefined) {
            if (typeof encoding === 'string') {
                buf.fill(fill, encoding);
            }
            else {
                buf.fill(fill);
            }
        }
        else {
            buf.fill(0);
        }
        return buf;
    };
    SafeBuffer.allocUnsafe = function (size) {
        if (typeof size !== 'number') {
            throw new TypeError('Argument must be a number');
        }
        return Buffer(size);
    };
    SafeBuffer.allocUnsafeSlow = function (size) {
        if (typeof size !== 'number') {
            throw new TypeError('Argument must be a number');
        }
        return buffer.SlowBuffer(size);
    };
},
/* lodash.memoize/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * lodash 3.0.4 (Custom Build) <https://lodash.com/>
     * Build: `lodash modern modularize exports="npm" -o ./`
     * Copyright 2012-2015 The Dojo Foundation <http://dojofoundation.org/>
     * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
     * Copyright 2009-2015 Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
     * Available under MIT license <https://lodash.com/license>
     */
    /** Used as the `TypeError` message for "Functions" methods. */
    var FUNC_ERROR_TEXT = 'Expected a function';
    /** Used for native method references. */
    var objectProto = Object.prototype;
    /** Used to check objects for own properties. */
    var hasOwnProperty = objectProto.hasOwnProperty;
    /**
     * Creates a cache object to store key/value pairs.
     *
     * @private
     * @static
     * @name Cache
     * @memberOf _.memoize
     */
    function MapCache() {
        this.__data__ = {};
    }
    /**
     * Removes `key` and its value from the cache.
     *
     * @private
     * @name delete
     * @memberOf _.memoize.Cache
     * @param {string} key The key of the value to remove.
     * @returns {boolean} Returns `true` if the entry was removed successfully, else `false`.
     */
    function mapDelete(key) {
        return this.has(key) && delete this.__data__[key];
    }
    /**
     * Gets the cached value for `key`.
     *
     * @private
     * @name get
     * @memberOf _.memoize.Cache
     * @param {string} key The key of the value to get.
     * @returns {*} Returns the cached value.
     */
    function mapGet(key) {
        return key == '__proto__' ? undefined : this.__data__[key];
    }
    /**
     * Checks if a cached value for `key` exists.
     *
     * @private
     * @name has
     * @memberOf _.memoize.Cache
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function mapHas(key) {
        return key != '__proto__' && hasOwnProperty.call(this.__data__, key);
    }
    /**
     * Sets `value` to `key` of the cache.
     *
     * @private
     * @name set
     * @memberOf _.memoize.Cache
     * @param {string} key The key of the value to cache.
     * @param {*} value The value to cache.
     * @returns {Object} Returns the cache object.
     */
    function mapSet(key, value) {
        if (key != '__proto__') {
            this.__data__[key] = value;
        }
        return this;
    }
    /**
     * Creates a function that memoizes the result of `func`. If `resolver` is
     * provided it determines the cache key for storing the result based on the
     * arguments provided to the memoized function. By default, the first argument
     * provided to the memoized function is coerced to a string and used as the
     * cache key. The `func` is invoked with the `this` binding of the memoized
     * function.
     *
     * **Note:** The cache is exposed as the `cache` property on the memoized
     * function. Its creation may be customized by replacing the `_.memoize.Cache`
     * constructor with one whose instances implement the [`Map`](https://people.mozilla.org/~jorendorff/es6-draft.html#sec-properties-of-the-map-prototype-object)
     * method interface of `get`, `has`, and `set`.
     *
     * @static
     * @memberOf _
     * @category Function
     * @param {Function} func The function to have its output memoized.
     * @param {Function} [resolver] The function to resolve the cache key.
     * @returns {Function} Returns the new memoizing function.
     * @example
     *
     * var upperCase = _.memoize(function(string) {
     *   return string.toUpperCase();
     * });
     *
     * upperCase('fred');
     * // => 'FRED'
     *
     * // modifying the result cache
     * upperCase.cache.set('fred', 'BARNEY');
     * upperCase('fred');
     * // => 'BARNEY'
     *
     * // replacing `_.memoize.Cache`
     * var object = { 'user': 'fred' };
     * var other = { 'user': 'barney' };
     * var identity = _.memoize(_.identity);
     *
     * identity(object);
     * // => { 'user': 'fred' }
     * identity(other);
     * // => { 'user': 'fred' }
     *
     * _.memoize.Cache = WeakMap;
     * var identity = _.memoize(_.identity);
     *
     * identity(object);
     * // => { 'user': 'fred' }
     * identity(other);
     * // => { 'user': 'barney' }
     */
    function memoize(func, resolver) {
        if (typeof func != 'function' || (resolver && typeof resolver != 'function')) {
            throw new TypeError(FUNC_ERROR_TEXT);
        }
        var memoized = function () {
            var args = arguments, key = resolver ? resolver.apply(this, args) : args[0], cache = memoized.cache;
            if (cache.has(key)) {
                return cache.get(key);
            }
            var result = func.apply(this, args);
            memoized.cache = cache.set(key, result);
            return result;
        };
        memoized.cache = new memoize.Cache;
        return memoized;
    }
    // Add functions to the `Map` cache.
    MapCache.prototype['delete'] = mapDelete;
    MapCache.prototype.get = mapGet;
    MapCache.prototype.has = mapHas;
    MapCache.prototype.set = mapSet;
    // Assign cache to `_.memoize`.
    memoize.Cache = MapCache;
    module.exports = memoize;
},
/* inline-source-map/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var SourceMapGenerator = require(23) /* source-map */.SourceMapGenerator;
    function offsetMapping(mapping, offset) {
        return { line: offset.line + mapping.line, column: offset.column + mapping.column };
    }
    function newlinesIn(src) {
        if (!src)
            return 0;
        var newlines = src.match(/\n/g);
        return newlines ? newlines.length : 0;
    }
    function Generator(opts) {
        opts = opts || {};
        this.generator = new SourceMapGenerator({ file: opts.file || '', sourceRoot: opts.sourceRoot || '' });
        this.sourcesContent = undefined;
        this.opts = opts;
    }
    /**
     * Adds the given mappings to the generator and offsets them if offset is given
     *
     * @name addMappings
     * @function
     * @param sourceFile {String} name of the source file
     * @param mappings {Array{{Object}} each object has the form { original: { line: _, column: _ }, generated: { line: _, column: _ } }
     * @param offset {Object} offset to apply to each mapping. Has the form { line: _, column: _ }
     * @return {Object} the generator to allow chaining
     */
    Generator.prototype.addMappings = function (sourceFile, mappings, offset) {
        var generator = this.generator;
        offset = offset || {};
        offset.line = offset.hasOwnProperty('line') ? offset.line : 0;
        offset.column = offset.hasOwnProperty('column') ? offset.column : 0;
        mappings.forEach(function (m) {
            // only set source if we have original position to handle edgecase (see inline-source-map tests)
            generator.addMapping({
                source: m.original ? sourceFile : undefined,
                original: m.original,
                generated: offsetMapping(m.generated, offset)
            });
        });
        return this;
    };
    /**
     * Generates mappings for the given source, assuming that no translation from original to generated is necessary.
     *
     * @name addGeneratedMappings
     * @function
     * @param sourceFile {String} name of the source file
     * @param source {String} source of the file
     * @param offset {Object} offset to apply to each mapping. Has the form { line: _, column: _ }
     * @return {Object} the generator to allow chaining
     */
    Generator.prototype.addGeneratedMappings = function (sourceFile, source, offset) {
        var mappings = [], linesToGenerate = newlinesIn(source) + 1;
        for (var line = 1; line <= linesToGenerate; line++) {
            var location = { line: line, column: 0 };
            mappings.push({ original: location, generated: location });
        }
        return this.addMappings(sourceFile, mappings, offset);
    };
    /**
     * Adds source content for the given source file.
     *
     * @name addSourceContent
     * @function
     * @param sourceFile {String} The source file for which a mapping is included
     * @param sourcesContent {String} The content of the source file
     * @return {Object} The generator to allow chaining
     */
    Generator.prototype.addSourceContent = function (sourceFile, sourcesContent) {
        this.sourcesContent = this.sourcesContent || {};
        this.sourcesContent[sourceFile] = sourcesContent;
        return this;
    };
    /**
     * @name base64Encode
     * @function
     * @return {String} bas64 encoded representation of the added mappings
     */
    Generator.prototype.base64Encode = function () {
        var map = this.toString();
        return new Buffer(map).toString('base64');
    };
    /**
     * @name inlineMappingUrl
     * @function
     * @return {String} comment with base64 encoded representation of the added mappings. Can be inlined at the end of the generated file.
     */
    Generator.prototype.inlineMappingUrl = function () {
        var charset = this.opts.charset || 'utf-8';
        return '//# sourceMappingURL=data:application/json;charset=' + charset + ';base64,' + this.base64Encode();
    };
    Generator.prototype.toJSON = function () {
        var map = this.generator.toJSON();
        if (!this.sourcesContent)
            return map;
        var toSourcesContent = (function (s) {
            if (typeof this.sourcesContent[s] === 'string') {
                return this.sourcesContent[s];
            }
            else {
                return null;
            }
        }).bind(this);
        map.sourcesContent = map.sources.map(toSourcesContent);
        return map;
    };
    Generator.prototype.toString = function () {
        return JSON.stringify(this);
    };
    Generator.prototype._mappings = function () {
        return this.generator._mappings._array;
    };
    Generator.prototype.gen = function () {
        return this.generator;
    };
    module.exports = function (opts) { return new Generator(opts); };
    module.exports.Generator = Generator;
},
/* combine-source-map/lib/path-is-absolute.js */ function _(require, module, exports, __esModule, __esExport) {
    function posix(path) {
        return path.charAt(0) === '/';
    }
    ;
    function win32(path) {
        // https://github.com/joyent/node/blob/b3fcc245fb25539909ef1d5eaa01dbf92e168633/lib/path.js#L56
        var splitDeviceRe = /^([a-zA-Z]:|[\\\/]{2}[^\\\/]+[\\\/]+[^\\\/]+)?([\\\/])?([\s\S]*?)$/;
        var result = splitDeviceRe.exec(path);
        var device = result[1] || '';
        var isUnc = !!device && device.charAt(1) !== ':';
        // UNC paths are always absolute
        return !!result[2] || isUnc;
    }
    ;
    module.exports = process.platform === 'win32' ? win32 : posix;
    module.exports.posix = posix;
    module.exports.win32 = win32;
},
/* combine-source-map/lib/mappings-from-map.js */ function _(require, module, exports, __esModule, __esExport) {
    var SMConsumer = require(23) /* source-map */.SourceMapConsumer;
    /**
     * @name mappingsFromMap
     * @function
     * @param map {Object} the JSON.parse()'ed map
     * @return {Array} array of mappings
     */
    module.exports = function (map) {
        var consumer = new SMConsumer(map);
        var mappings = [];
        consumer.eachMapping(function (mapping) {
            // only set source if we have original position to handle edgecase (see inline-source-map tests)
            mappings.push({
                original: mapping.originalColumn != null ? {
                    column: mapping.originalColumn,
                    line: mapping.originalLine
                } : undefined,
                generated: {
                    column: mapping.generatedColumn,
                    line: mapping.generatedLine
                },
                source: mapping.originalColumn != null ? mapping.source : undefined,
                name: mapping.name
            });
        });
        return mappings;
    };
},
/* graph.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function detect_cycles(graph) {
        const cycles = [];
        const state = new Map();
        for (const node of graph.keys()) {
            state.set(node, { visited: false, explored: false });
        }
        function detect_cycle(nodes) {
            const node = nodes[0];
            const entry = state.get(node);
            if (entry.visited)
                return false;
            if (entry.explored) {
                cycles.push(nodes);
                return true;
            }
            entry.explored = true;
            const neighbors = graph.get(node);
            for (const neighbor of neighbors) {
                const { visited } = state.get(neighbor);
                if (!visited) {
                    const cycle = detect_cycle([neighbor, ...nodes]);
                    if (cycle)
                        break;
                }
            }
            entry.explored = false;
            entry.visited = true;
            return false;
        }
        for (const node of graph.keys()) {
            detect_cycle([node]);
        }
        return cycles;
    }
    exports.detect_cycles = detect_cycles;
},
/* styles.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const path_1 = require("path");
    const less_1 = __importDefault(require(56) /* less */);
    const chalk_1 = __importDefault(require(37) /* chalk */);
    const sys_1 = require(20) /* ./sys */;
    async function compile_styles(styles_dir, css_dir) {
        let success = true;
        for (const src of (0, sys_1.scan)(styles_dir, [".less", ".css"])) {
            if ((0, path_1.basename)(src).startsWith("_"))
                continue;
            try {
                const style = (0, sys_1.read)(src);
                const { css } = await less_1.default.render(style, { filename: src });
                const dst = (0, sys_1.rename)(src, { base: styles_dir, dir: css_dir, ext: ".css" });
                (0, sys_1.write)(dst, css);
            }
            catch (error) {
                success = false;
                console.log(`${chalk_1.default.red("\u2717")} failed to compile ${chalk_1.default.magenta(src)}:`);
                console.log(`${error}`);
            }
        }
        return success;
    }
    exports.compile_styles = compile_styles;
    function wrap_css_modules(css_dir, js_dir, dts_dir) {
        for (const css_path of (0, sys_1.scan)(css_dir, [".css"])) {
            const js = `\
const css = \`\n${(0, sys_1.read)(css_path)}\`;
export default css;
`;
            const dts = `\
declare const css: string;
export default css;
`;
            const sub_path = (0, path_1.relative)(css_dir, css_path);
            (0, sys_1.write)(`${(0, path_1.join)(js_dir, "styles", sub_path)}.js`, js);
            (0, sys_1.write)(`${(0, path_1.join)(dts_dir, "styles", sub_path)}.d.ts`, dts);
        }
    }
    exports.wrap_css_modules = wrap_css_modules;
},
/* less/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = require(57) /* ./lib/less-node */.default;
},
/* less/lib/less-node/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var environment_1 = tslib_1.__importDefault(require(59) /* ./environment */);
    var file_manager_1 = tslib_1.__importDefault(require(62) /* ./file-manager */);
    var url_file_manager_1 = tslib_1.__importDefault(require(69) /* ./url-file-manager */);
    var less_1 = tslib_1.__importDefault(require(107) /* ../less */);
    var less = less_1.default(environment_1.default, [new file_manager_1.default(), new url_file_manager_1.default()]);
    var lessc_helper_1 = tslib_1.__importDefault(require(194) /* ./lessc-helper */);
    // allow people to create less with their own environment
    less.createFromEnvironment = less_1.default;
    less.lesscHelper = lessc_helper_1.default;
    less.PluginLoader = require(195) /* ./plugin-loader */.default;
    less.fs = require(63) /* ./fs */.default;
    less.FileManager = file_manager_1.default;
    less.UrlFileManager = url_file_manager_1.default;
    // Set up options
    less.options = require(196) /* ../less/default-options */.default();
    // provide image-size functionality
    require(197) /* ./image-size */.default(less.environment);
    exports.default = less;
},
/* tslib/tslib.js */ function _(require, module, exports, __esModule, __esExport) {
    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation.
    
    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.
    
    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */
    /* global global, define, System, Reflect, Promise */
    var __extends;
    var __assign;
    var __rest;
    var __decorate;
    var __param;
    var __metadata;
    var __awaiter;
    var __generator;
    var __exportStar;
    var __values;
    var __read;
    var __spread;
    var __spreadArrays;
    var __spreadArray;
    var __await;
    var __asyncGenerator;
    var __asyncDelegator;
    var __asyncValues;
    var __makeTemplateObject;
    var __importStar;
    var __importDefault;
    var __classPrivateFieldGet;
    var __classPrivateFieldSet;
    var __createBinding;
    (function (factory) {
        var root = typeof global === "object" ? global : typeof self === "object" ? self : typeof this === "object" ? this : {};
        if (typeof define === "function" && define.amd) {
            define("tslib", ["exports"], function (exports) { factory(createExporter(root, createExporter(exports))); });
        }
        else if (typeof module === "object" && typeof module.exports === "object") {
            factory(createExporter(root, createExporter(module.exports)));
        }
        else {
            factory(createExporter(root));
        }
        function createExporter(exports, previous) {
            if (exports !== root) {
                if (typeof Object.create === "function") {
                    Object.defineProperty(exports, "__esModule", { value: true });
                }
                else {
                    exports.__esModule = true;
                }
            }
            return function (id, v) { return exports[id] = previous ? previous(id, v) : v; };
        }
    })(function (exporter) {
        var extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b)
                if (Object.prototype.hasOwnProperty.call(b, p))
                    d[p] = b[p]; };
        __extends = function (d, b) {
            if (typeof b !== "function" && b !== null)
                throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
            extendStatics(d, b);
            function __() { this.constructor = d; }
            d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
        };
        __assign = Object.assign || function (t) {
            for (var s, i = 1, n = arguments.length; i < n; i++) {
                s = arguments[i];
                for (var p in s)
                    if (Object.prototype.hasOwnProperty.call(s, p))
                        t[p] = s[p];
            }
            return t;
        };
        __rest = function (s, e) {
            var t = {};
            for (var p in s)
                if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
                    t[p] = s[p];
            if (s != null && typeof Object.getOwnPropertySymbols === "function")
                for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
                    if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                        t[p[i]] = s[p[i]];
                }
            return t;
        };
        __decorate = function (decorators, target, key, desc) {
            var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
            if (typeof Reflect === "object" && typeof Reflect.decorate === "function")
                r = Reflect.decorate(decorators, target, key, desc);
            else
                for (var i = decorators.length - 1; i >= 0; i--)
                    if (d = decorators[i])
                        r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
            return c > 3 && r && Object.defineProperty(target, key, r), r;
        };
        __param = function (paramIndex, decorator) {
            return function (target, key) { decorator(target, key, paramIndex); };
        };
        __metadata = function (metadataKey, metadataValue) {
            if (typeof Reflect === "object" && typeof Reflect.metadata === "function")
                return Reflect.metadata(metadataKey, metadataValue);
        };
        __awaiter = function (thisArg, _arguments, P, generator) {
            function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
            return new (P || (P = Promise))(function (resolve, reject) {
                function fulfilled(value) { try {
                    step(generator.next(value));
                }
                catch (e) {
                    reject(e);
                } }
                function rejected(value) { try {
                    step(generator["throw"](value));
                }
                catch (e) {
                    reject(e);
                } }
                function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
                step((generator = generator.apply(thisArg, _arguments || [])).next());
            });
        };
        __generator = function (thisArg, body) {
            var _ = { label: 0, sent: function () { if (t[0] & 1)
                    throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
            return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function () { return this; }), g;
            function verb(n) { return function (v) { return step([n, v]); }; }
            function step(op) {
                if (f)
                    throw new TypeError("Generator is already executing.");
                while (_)
                    try {
                        if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done)
                            return t;
                        if (y = 0, t)
                            op = [op[0] & 2, t.value];
                        switch (op[0]) {
                            case 0:
                            case 1:
                                t = op;
                                break;
                            case 4:
                                _.label++;
                                return { value: op[1], done: false };
                            case 5:
                                _.label++;
                                y = op[1];
                                op = [0];
                                continue;
                            case 7:
                                op = _.ops.pop();
                                _.trys.pop();
                                continue;
                            default:
                                if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) {
                                    _ = 0;
                                    continue;
                                }
                                if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) {
                                    _.label = op[1];
                                    break;
                                }
                                if (op[0] === 6 && _.label < t[1]) {
                                    _.label = t[1];
                                    t = op;
                                    break;
                                }
                                if (t && _.label < t[2]) {
                                    _.label = t[2];
                                    _.ops.push(op);
                                    break;
                                }
                                if (t[2])
                                    _.ops.pop();
                                _.trys.pop();
                                continue;
                        }
                        op = body.call(thisArg, _);
                    }
                    catch (e) {
                        op = [6, e];
                        y = 0;
                    }
                    finally {
                        f = t = 0;
                    }
                if (op[0] & 5)
                    throw op[1];
                return { value: op[0] ? op[1] : void 0, done: true };
            }
        };
        __exportStar = function (m, o) {
            for (var p in m)
                if (p !== "default" && !Object.prototype.hasOwnProperty.call(o, p))
                    __createBinding(o, m, p);
        };
        __createBinding = Object.create ? (function (o, m, k, k2) {
            if (k2 === undefined)
                k2 = k;
            Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
        }) : (function (o, m, k, k2) {
            if (k2 === undefined)
                k2 = k;
            o[k2] = m[k];
        });
        __values = function (o) {
            var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
            if (m)
                return m.call(o);
            if (o && typeof o.length === "number")
                return {
                    next: function () {
                        if (o && i >= o.length)
                            o = void 0;
                        return { value: o && o[i++], done: !o };
                    }
                };
            throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
        };
        __read = function (o, n) {
            var m = typeof Symbol === "function" && o[Symbol.iterator];
            if (!m)
                return o;
            var i = m.call(o), r, ar = [], e;
            try {
                while ((n === void 0 || n-- > 0) && !(r = i.next()).done)
                    ar.push(r.value);
            }
            catch (error) {
                e = { error: error };
            }
            finally {
                try {
                    if (r && !r.done && (m = i["return"]))
                        m.call(i);
                }
                finally {
                    if (e)
                        throw e.error;
                }
            }
            return ar;
        };
        /** @deprecated */
        __spread = function () {
            for (var ar = [], i = 0; i < arguments.length; i++)
                ar = ar.concat(__read(arguments[i]));
            return ar;
        };
        /** @deprecated */
        __spreadArrays = function () {
            for (var s = 0, i = 0, il = arguments.length; i < il; i++)
                s += arguments[i].length;
            for (var r = Array(s), k = 0, i = 0; i < il; i++)
                for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
                    r[k] = a[j];
            return r;
        };
        __spreadArray = function (to, from, pack) {
            if (pack || arguments.length === 2)
                for (var i = 0, l = from.length, ar; i < l; i++) {
                    if (ar || !(i in from)) {
                        if (!ar)
                            ar = Array.prototype.slice.call(from, 0, i);
                        ar[i] = from[i];
                    }
                }
            return to.concat(ar || from);
        };
        __await = function (v) {
            return this instanceof __await ? (this.v = v, this) : new __await(v);
        };
        __asyncGenerator = function (thisArg, _arguments, generator) {
            if (!Symbol.asyncIterator)
                throw new TypeError("Symbol.asyncIterator is not defined.");
            var g = generator.apply(thisArg, _arguments || []), i, q = [];
            return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
            function verb(n) { if (g[n])
                i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
            function resume(n, v) { try {
                step(g[n](v));
            }
            catch (e) {
                settle(q[0][3], e);
            } }
            function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
            function fulfill(value) { resume("next", value); }
            function reject(value) { resume("throw", value); }
            function settle(f, v) { if (f(v), q.shift(), q.length)
                resume(q[0][0], q[0][1]); }
        };
        __asyncDelegator = function (o) {
            var i, p;
            return i = {}, verb("next"), verb("throw", function (e) { throw e; }), verb("return"), i[Symbol.iterator] = function () { return this; }, i;
            function verb(n, f) { i[n] = o[n] ? function (v) { return (p = !p) ? { value: __await(o[n](v)), done: n === "return" } : f ? f(v) : v; } : f; }
        };
        __asyncValues = function (o) {
            if (!Symbol.asyncIterator)
                throw new TypeError("Symbol.asyncIterator is not defined.");
            var m = o[Symbol.asyncIterator], i;
            return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
            function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
            function settle(resolve, reject, d, v) { Promise.resolve(v).then(function (v) { resolve({ value: v, done: d }); }, reject); }
        };
        __makeTemplateObject = function (cooked, raw) {
            if (Object.defineProperty) {
                Object.defineProperty(cooked, "raw", { value: raw });
            }
            else {
                cooked.raw = raw;
            }
            return cooked;
        };
        var __setModuleDefault = Object.create ? (function (o, v) {
            Object.defineProperty(o, "default", { enumerable: true, value: v });
        }) : function (o, v) {
            o["default"] = v;
        };
        __importStar = function (mod) {
            if (mod && mod.__esModule)
                return mod;
            var result = {};
            if (mod != null)
                for (var k in mod)
                    if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                        __createBinding(result, mod, k);
            __setModuleDefault(result, mod);
            return result;
        };
        __importDefault = function (mod) {
            return (mod && mod.__esModule) ? mod : { "default": mod };
        };
        __classPrivateFieldGet = function (receiver, state, kind, f) {
            if (kind === "a" && !f)
                throw new TypeError("Private accessor was defined without a getter");
            if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
                throw new TypeError("Cannot read private member from an object whose class did not declare it");
            return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
        };
        __classPrivateFieldSet = function (receiver, state, value, kind, f) {
            if (kind === "m")
                throw new TypeError("Private method is not writable");
            if (kind === "a" && !f)
                throw new TypeError("Private accessor was defined without a setter");
            if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
                throw new TypeError("Cannot write private member to an object whose class did not declare it");
            return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
        };
        exporter("__extends", __extends);
        exporter("__assign", __assign);
        exporter("__rest", __rest);
        exporter("__decorate", __decorate);
        exporter("__param", __param);
        exporter("__metadata", __metadata);
        exporter("__awaiter", __awaiter);
        exporter("__generator", __generator);
        exporter("__exportStar", __exportStar);
        exporter("__createBinding", __createBinding);
        exporter("__values", __values);
        exporter("__read", __read);
        exporter("__spread", __spread);
        exporter("__spreadArrays", __spreadArrays);
        exporter("__spreadArray", __spreadArray);
        exporter("__await", __await);
        exporter("__asyncGenerator", __asyncGenerator);
        exporter("__asyncDelegator", __asyncDelegator);
        exporter("__asyncValues", __asyncValues);
        exporter("__makeTemplateObject", __makeTemplateObject);
        exporter("__importStar", __importStar);
        exporter("__importDefault", __importDefault);
        exporter("__classPrivateFieldGet", __classPrivateFieldGet);
        exporter("__classPrivateFieldSet", __classPrivateFieldSet);
    });
},
/* less/lib/less-node/environment.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.default = {
        encodeBase64: function encodeBase64(str) {
            // Avoid Buffer constructor on newer versions of Node.js.
            var buffer = (Buffer.from ? Buffer.from(str) : (new Buffer(str)));
            return buffer.toString('base64');
        },
        mimeLookup: function (filename) {
            return require(60) /* mime */.lookup(filename);
        },
        charsetLookup: function (mime) {
            return require(60) /* mime */.charsets.lookup(mime);
        },
        getSourceMapGenerator: function getSourceMapGenerator() {
            return require(23) /* source-map */.SourceMapGenerator;
        }
    };
},
/* mime/mime.js */ function _(require, module, exports, __esModule, __esExport) {
    var path = require('path');
    var fs = require('fs');
    function Mime() {
        // Map of extension -> mime type
        this.types = Object.create(null);
        // Map of mime type -> extension
        this.extensions = Object.create(null);
    }
    /**
     * Define mimetype -> extension mappings.  Each key is a mime-type that maps
     * to an array of extensions associated with the type.  The first extension is
     * used as the default extension for the type.
     *
     * e.g. mime.define({'audio/ogg', ['oga', 'ogg', 'spx']});
     *
     * @param map (Object) type definitions
     */
    Mime.prototype.define = function (map) {
        for (var type in map) {
            var exts = map[type];
            for (var i = 0; i < exts.length; i++) {
                if (process.env.DEBUG_MIME && this.types[exts[i]]) {
                    console.warn((this._loading || "define()").replace(/.*\//, ''), 'changes "' + exts[i] + '" extension type from ' +
                        this.types[exts[i]] + ' to ' + type);
                }
                this.types[exts[i]] = type;
            }
            // Default extension is the first one we encounter
            if (!this.extensions[type]) {
                this.extensions[type] = exts[0];
            }
        }
    };
    /**
     * Load an Apache2-style ".types" file
     *
     * This may be called multiple times (it's expected).  Where files declare
     * overlapping types/extensions, the last file wins.
     *
     * @param file (String) path of file to load.
     */
    Mime.prototype.load = function (file) {
        this._loading = file;
        // Read file and split into lines
        var map = {}, content = fs.readFileSync(file, 'ascii'), lines = content.split(/[\r\n]+/);
        lines.forEach(function (line) {
            // Clean up whitespace/comments, and split into fields
            var fields = line.replace(/\s*#.*|^\s*|\s*$/g, '').split(/\s+/);
            map[fields.shift()] = fields;
        });
        this.define(map);
        this._loading = null;
    };
    /**
     * Lookup a mime type based on extension
     */
    Mime.prototype.lookup = function (path, fallback) {
        var ext = path.replace(/^.*[\.\/\\]/, '').toLowerCase();
        return this.types[ext] || fallback || this.default_type;
    };
    /**
     * Return file extension associated with a mime type
     */
    Mime.prototype.extension = function (mimeType) {
        var type = mimeType.match(/^\s*([^;\s]*)(?:;|\s|$)/)[1].toLowerCase();
        return this.extensions[type];
    };
    // Default instance
    var mime = new Mime();
    // Define built-in types
    mime.define(require(61) /* ./types.json */);
    // Default type
    mime.default_type = mime.lookup('bin');
    //
    // Additional API specific to the default instance
    //
    mime.Mime = Mime;
    /**
     * Lookup a charset based on mime type.
     */
    mime.charsets = {
        lookup: function (mimeType, fallback) {
            // Assume text types are utf8
            return (/^text\/|^application\/(javascript|json)/).test(mimeType) ? 'UTF-8' : fallback;
        }
    };
    module.exports = mime;
},
/* mime/types.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = { "application/andrew-inset": ["ez"], "application/applixware": ["aw"], "application/atom+xml": ["atom"], "application/atomcat+xml": ["atomcat"], "application/atomsvc+xml": ["atomsvc"], "application/bdoc": ["bdoc"], "application/ccxml+xml": ["ccxml"], "application/cdmi-capability": ["cdmia"], "application/cdmi-container": ["cdmic"], "application/cdmi-domain": ["cdmid"], "application/cdmi-object": ["cdmio"], "application/cdmi-queue": ["cdmiq"], "application/cu-seeme": ["cu"], "application/dash+xml": ["mpd"], "application/davmount+xml": ["davmount"], "application/docbook+xml": ["dbk"], "application/dssc+der": ["dssc"], "application/dssc+xml": ["xdssc"], "application/ecmascript": ["ecma"], "application/emma+xml": ["emma"], "application/epub+zip": ["epub"], "application/exi": ["exi"], "application/font-tdpfr": ["pfr"], "application/font-woff": [], "application/font-woff2": [], "application/geo+json": ["geojson"], "application/gml+xml": ["gml"], "application/gpx+xml": ["gpx"], "application/gxf": ["gxf"], "application/gzip": ["gz"], "application/hyperstudio": ["stk"], "application/inkml+xml": ["ink", "inkml"], "application/ipfix": ["ipfix"], "application/java-archive": ["jar", "war", "ear"], "application/java-serialized-object": ["ser"], "application/java-vm": ["class"], "application/javascript": ["js", "mjs"], "application/json": ["json", "map"], "application/json5": ["json5"], "application/jsonml+json": ["jsonml"], "application/ld+json": ["jsonld"], "application/lost+xml": ["lostxml"], "application/mac-binhex40": ["hqx"], "application/mac-compactpro": ["cpt"], "application/mads+xml": ["mads"], "application/manifest+json": ["webmanifest"], "application/marc": ["mrc"], "application/marcxml+xml": ["mrcx"], "application/mathematica": ["ma", "nb", "mb"], "application/mathml+xml": ["mathml"], "application/mbox": ["mbox"], "application/mediaservercontrol+xml": ["mscml"], "application/metalink+xml": ["metalink"], "application/metalink4+xml": ["meta4"], "application/mets+xml": ["mets"], "application/mods+xml": ["mods"], "application/mp21": ["m21", "mp21"], "application/mp4": ["mp4s", "m4p"], "application/msword": ["doc", "dot"], "application/mxf": ["mxf"], "application/octet-stream": ["bin", "dms", "lrf", "mar", "so", "dist", "distz", "pkg", "bpk", "dump", "elc", "deploy", "exe", "dll", "deb", "dmg", "iso", "img", "msi", "msp", "msm", "buffer"], "application/oda": ["oda"], "application/oebps-package+xml": ["opf"], "application/ogg": ["ogx"], "application/omdoc+xml": ["omdoc"], "application/onenote": ["onetoc", "onetoc2", "onetmp", "onepkg"], "application/oxps": ["oxps"], "application/patch-ops-error+xml": ["xer"], "application/pdf": ["pdf"], "application/pgp-encrypted": ["pgp"], "application/pgp-signature": ["asc", "sig"], "application/pics-rules": ["prf"], "application/pkcs10": ["p10"], "application/pkcs7-mime": ["p7m", "p7c"], "application/pkcs7-signature": ["p7s"], "application/pkcs8": ["p8"], "application/pkix-attr-cert": ["ac"], "application/pkix-cert": ["cer"], "application/pkix-crl": ["crl"], "application/pkix-pkipath": ["pkipath"], "application/pkixcmp": ["pki"], "application/pls+xml": ["pls"], "application/postscript": ["ai", "eps", "ps"], "application/prs.cww": ["cww"], "application/pskc+xml": ["pskcxml"], "application/raml+yaml": ["raml"], "application/rdf+xml": ["rdf"], "application/reginfo+xml": ["rif"], "application/relax-ng-compact-syntax": ["rnc"], "application/resource-lists+xml": ["rl"], "application/resource-lists-diff+xml": ["rld"], "application/rls-services+xml": ["rs"], "application/rpki-ghostbusters": ["gbr"], "application/rpki-manifest": ["mft"], "application/rpki-roa": ["roa"], "application/rsd+xml": ["rsd"], "application/rss+xml": ["rss"], "application/rtf": ["rtf"], "application/sbml+xml": ["sbml"], "application/scvp-cv-request": ["scq"], "application/scvp-cv-response": ["scs"], "application/scvp-vp-request": ["spq"], "application/scvp-vp-response": ["spp"], "application/sdp": ["sdp"], "application/set-payment-initiation": ["setpay"], "application/set-registration-initiation": ["setreg"], "application/shf+xml": ["shf"], "application/smil+xml": ["smi", "smil"], "application/sparql-query": ["rq"], "application/sparql-results+xml": ["srx"], "application/srgs": ["gram"], "application/srgs+xml": ["grxml"], "application/sru+xml": ["sru"], "application/ssdl+xml": ["ssdl"], "application/ssml+xml": ["ssml"], "application/tei+xml": ["tei", "teicorpus"], "application/thraud+xml": ["tfi"], "application/timestamped-data": ["tsd"], "application/vnd.3gpp.pic-bw-large": ["plb"], "application/vnd.3gpp.pic-bw-small": ["psb"], "application/vnd.3gpp.pic-bw-var": ["pvb"], "application/vnd.3gpp2.tcap": ["tcap"], "application/vnd.3m.post-it-notes": ["pwn"], "application/vnd.accpac.simply.aso": ["aso"], "application/vnd.accpac.simply.imp": ["imp"], "application/vnd.acucobol": ["acu"], "application/vnd.acucorp": ["atc", "acutc"], "application/vnd.adobe.air-application-installer-package+zip": ["air"], "application/vnd.adobe.formscentral.fcdt": ["fcdt"], "application/vnd.adobe.fxp": ["fxp", "fxpl"], "application/vnd.adobe.xdp+xml": ["xdp"], "application/vnd.adobe.xfdf": ["xfdf"], "application/vnd.ahead.space": ["ahead"], "application/vnd.airzip.filesecure.azf": ["azf"], "application/vnd.airzip.filesecure.azs": ["azs"], "application/vnd.amazon.ebook": ["azw"], "application/vnd.americandynamics.acc": ["acc"], "application/vnd.amiga.ami": ["ami"], "application/vnd.android.package-archive": ["apk"], "application/vnd.anser-web-certificate-issue-initiation": ["cii"], "application/vnd.anser-web-funds-transfer-initiation": ["fti"], "application/vnd.antix.game-component": ["atx"], "application/vnd.apple.installer+xml": ["mpkg"], "application/vnd.apple.mpegurl": ["m3u8"], "application/vnd.apple.pkpass": ["pkpass"], "application/vnd.aristanetworks.swi": ["swi"], "application/vnd.astraea-software.iota": ["iota"], "application/vnd.audiograph": ["aep"], "application/vnd.blueice.multipass": ["mpm"], "application/vnd.bmi": ["bmi"], "application/vnd.businessobjects": ["rep"], "application/vnd.chemdraw+xml": ["cdxml"], "application/vnd.chipnuts.karaoke-mmd": ["mmd"], "application/vnd.cinderella": ["cdy"], "application/vnd.claymore": ["cla"], "application/vnd.cloanto.rp9": ["rp9"], "application/vnd.clonk.c4group": ["c4g", "c4d", "c4f", "c4p", "c4u"], "application/vnd.cluetrust.cartomobile-config": ["c11amc"], "application/vnd.cluetrust.cartomobile-config-pkg": ["c11amz"], "application/vnd.commonspace": ["csp"], "application/vnd.contact.cmsg": ["cdbcmsg"], "application/vnd.cosmocaller": ["cmc"], "application/vnd.crick.clicker": ["clkx"], "application/vnd.crick.clicker.keyboard": ["clkk"], "application/vnd.crick.clicker.palette": ["clkp"], "application/vnd.crick.clicker.template": ["clkt"], "application/vnd.crick.clicker.wordbank": ["clkw"], "application/vnd.criticaltools.wbs+xml": ["wbs"], "application/vnd.ctc-posml": ["pml"], "application/vnd.cups-ppd": ["ppd"], "application/vnd.curl.car": ["car"], "application/vnd.curl.pcurl": ["pcurl"], "application/vnd.dart": ["dart"], "application/vnd.data-vision.rdz": ["rdz"], "application/vnd.dece.data": ["uvf", "uvvf", "uvd", "uvvd"], "application/vnd.dece.ttml+xml": ["uvt", "uvvt"], "application/vnd.dece.unspecified": ["uvx", "uvvx"], "application/vnd.dece.zip": ["uvz", "uvvz"], "application/vnd.denovo.fcselayout-link": ["fe_launch"], "application/vnd.dna": ["dna"], "application/vnd.dolby.mlp": ["mlp"], "application/vnd.dpgraph": ["dpg"], "application/vnd.dreamfactory": ["dfac"], "application/vnd.ds-keypoint": ["kpxx"], "application/vnd.dvb.ait": ["ait"], "application/vnd.dvb.service": ["svc"], "application/vnd.dynageo": ["geo"], "application/vnd.ecowin.chart": ["mag"], "application/vnd.enliven": ["nml"], "application/vnd.epson.esf": ["esf"], "application/vnd.epson.msf": ["msf"], "application/vnd.epson.quickanime": ["qam"], "application/vnd.epson.salt": ["slt"], "application/vnd.epson.ssf": ["ssf"], "application/vnd.eszigno3+xml": ["es3", "et3"], "application/vnd.ezpix-album": ["ez2"], "application/vnd.ezpix-package": ["ez3"], "application/vnd.fdf": ["fdf"], "application/vnd.fdsn.mseed": ["mseed"], "application/vnd.fdsn.seed": ["seed", "dataless"], "application/vnd.flographit": ["gph"], "application/vnd.fluxtime.clip": ["ftc"], "application/vnd.framemaker": ["fm", "frame", "maker", "book"], "application/vnd.frogans.fnc": ["fnc"], "application/vnd.frogans.ltf": ["ltf"], "application/vnd.fsc.weblaunch": ["fsc"], "application/vnd.fujitsu.oasys": ["oas"], "application/vnd.fujitsu.oasys2": ["oa2"], "application/vnd.fujitsu.oasys3": ["oa3"], "application/vnd.fujitsu.oasysgp": ["fg5"], "application/vnd.fujitsu.oasysprs": ["bh2"], "application/vnd.fujixerox.ddd": ["ddd"], "application/vnd.fujixerox.docuworks": ["xdw"], "application/vnd.fujixerox.docuworks.binder": ["xbd"], "application/vnd.fuzzysheet": ["fzs"], "application/vnd.genomatix.tuxedo": ["txd"], "application/vnd.geogebra.file": ["ggb"], "application/vnd.geogebra.tool": ["ggt"], "application/vnd.geometry-explorer": ["gex", "gre"], "application/vnd.geonext": ["gxt"], "application/vnd.geoplan": ["g2w"], "application/vnd.geospace": ["g3w"], "application/vnd.gmx": ["gmx"], "application/vnd.google-apps.document": ["gdoc"], "application/vnd.google-apps.presentation": ["gslides"], "application/vnd.google-apps.spreadsheet": ["gsheet"], "application/vnd.google-earth.kml+xml": ["kml"], "application/vnd.google-earth.kmz": ["kmz"], "application/vnd.grafeq": ["gqf", "gqs"], "application/vnd.groove-account": ["gac"], "application/vnd.groove-help": ["ghf"], "application/vnd.groove-identity-message": ["gim"], "application/vnd.groove-injector": ["grv"], "application/vnd.groove-tool-message": ["gtm"], "application/vnd.groove-tool-template": ["tpl"], "application/vnd.groove-vcard": ["vcg"], "application/vnd.hal+xml": ["hal"], "application/vnd.handheld-entertainment+xml": ["zmm"], "application/vnd.hbci": ["hbci"], "application/vnd.hhe.lesson-player": ["les"], "application/vnd.hp-hpgl": ["hpgl"], "application/vnd.hp-hpid": ["hpid"], "application/vnd.hp-hps": ["hps"], "application/vnd.hp-jlyt": ["jlt"], "application/vnd.hp-pcl": ["pcl"], "application/vnd.hp-pclxl": ["pclxl"], "application/vnd.hydrostatix.sof-data": ["sfd-hdstx"], "application/vnd.ibm.minipay": ["mpy"], "application/vnd.ibm.modcap": ["afp", "listafp", "list3820"], "application/vnd.ibm.rights-management": ["irm"], "application/vnd.ibm.secure-container": ["sc"], "application/vnd.iccprofile": ["icc", "icm"], "application/vnd.igloader": ["igl"], "application/vnd.immervision-ivp": ["ivp"], "application/vnd.immervision-ivu": ["ivu"], "application/vnd.insors.igm": ["igm"], "application/vnd.intercon.formnet": ["xpw", "xpx"], "application/vnd.intergeo": ["i2g"], "application/vnd.intu.qbo": ["qbo"], "application/vnd.intu.qfx": ["qfx"], "application/vnd.ipunplugged.rcprofile": ["rcprofile"], "application/vnd.irepository.package+xml": ["irp"], "application/vnd.is-xpr": ["xpr"], "application/vnd.isac.fcs": ["fcs"], "application/vnd.jam": ["jam"], "application/vnd.jcp.javame.midlet-rms": ["rms"], "application/vnd.jisp": ["jisp"], "application/vnd.joost.joda-archive": ["joda"], "application/vnd.kahootz": ["ktz", "ktr"], "application/vnd.kde.karbon": ["karbon"], "application/vnd.kde.kchart": ["chrt"], "application/vnd.kde.kformula": ["kfo"], "application/vnd.kde.kivio": ["flw"], "application/vnd.kde.kontour": ["kon"], "application/vnd.kde.kpresenter": ["kpr", "kpt"], "application/vnd.kde.kspread": ["ksp"], "application/vnd.kde.kword": ["kwd", "kwt"], "application/vnd.kenameaapp": ["htke"], "application/vnd.kidspiration": ["kia"], "application/vnd.kinar": ["kne", "knp"], "application/vnd.koan": ["skp", "skd", "skt", "skm"], "application/vnd.kodak-descriptor": ["sse"], "application/vnd.las.las+xml": ["lasxml"], "application/vnd.llamagraphics.life-balance.desktop": ["lbd"], "application/vnd.llamagraphics.life-balance.exchange+xml": ["lbe"], "application/vnd.lotus-1-2-3": ["123"], "application/vnd.lotus-approach": ["apr"], "application/vnd.lotus-freelance": ["pre"], "application/vnd.lotus-notes": ["nsf"], "application/vnd.lotus-organizer": ["org"], "application/vnd.lotus-screencam": ["scm"], "application/vnd.lotus-wordpro": ["lwp"], "application/vnd.macports.portpkg": ["portpkg"], "application/vnd.mcd": ["mcd"], "application/vnd.medcalcdata": ["mc1"], "application/vnd.mediastation.cdkey": ["cdkey"], "application/vnd.mfer": ["mwf"], "application/vnd.mfmp": ["mfm"], "application/vnd.micrografx.flo": ["flo"], "application/vnd.micrografx.igx": ["igx"], "application/vnd.mif": ["mif"], "application/vnd.mobius.daf": ["daf"], "application/vnd.mobius.dis": ["dis"], "application/vnd.mobius.mbk": ["mbk"], "application/vnd.mobius.mqy": ["mqy"], "application/vnd.mobius.msl": ["msl"], "application/vnd.mobius.plc": ["plc"], "application/vnd.mobius.txf": ["txf"], "application/vnd.mophun.application": ["mpn"], "application/vnd.mophun.certificate": ["mpc"], "application/vnd.mozilla.xul+xml": ["xul"], "application/vnd.ms-artgalry": ["cil"], "application/vnd.ms-cab-compressed": ["cab"], "application/vnd.ms-excel": ["xls", "xlm", "xla", "xlc", "xlt", "xlw"], "application/vnd.ms-excel.addin.macroenabled.12": ["xlam"], "application/vnd.ms-excel.sheet.binary.macroenabled.12": ["xlsb"], "application/vnd.ms-excel.sheet.macroenabled.12": ["xlsm"], "application/vnd.ms-excel.template.macroenabled.12": ["xltm"], "application/vnd.ms-fontobject": ["eot"], "application/vnd.ms-htmlhelp": ["chm"], "application/vnd.ms-ims": ["ims"], "application/vnd.ms-lrm": ["lrm"], "application/vnd.ms-officetheme": ["thmx"], "application/vnd.ms-outlook": ["msg"], "application/vnd.ms-pki.seccat": ["cat"], "application/vnd.ms-pki.stl": ["stl"], "application/vnd.ms-powerpoint": ["ppt", "pps", "pot"], "application/vnd.ms-powerpoint.addin.macroenabled.12": ["ppam"], "application/vnd.ms-powerpoint.presentation.macroenabled.12": ["pptm"], "application/vnd.ms-powerpoint.slide.macroenabled.12": ["sldm"], "application/vnd.ms-powerpoint.slideshow.macroenabled.12": ["ppsm"], "application/vnd.ms-powerpoint.template.macroenabled.12": ["potm"], "application/vnd.ms-project": ["mpp", "mpt"], "application/vnd.ms-word.document.macroenabled.12": ["docm"], "application/vnd.ms-word.template.macroenabled.12": ["dotm"], "application/vnd.ms-works": ["wps", "wks", "wcm", "wdb"], "application/vnd.ms-wpl": ["wpl"], "application/vnd.ms-xpsdocument": ["xps"], "application/vnd.mseq": ["mseq"], "application/vnd.musician": ["mus"], "application/vnd.muvee.style": ["msty"], "application/vnd.mynfc": ["taglet"], "application/vnd.neurolanguage.nlu": ["nlu"], "application/vnd.nitf": ["ntf", "nitf"], "application/vnd.noblenet-directory": ["nnd"], "application/vnd.noblenet-sealer": ["nns"], "application/vnd.noblenet-web": ["nnw"], "application/vnd.nokia.n-gage.data": ["ngdat"], "application/vnd.nokia.n-gage.symbian.install": ["n-gage"], "application/vnd.nokia.radio-preset": ["rpst"], "application/vnd.nokia.radio-presets": ["rpss"], "application/vnd.novadigm.edm": ["edm"], "application/vnd.novadigm.edx": ["edx"], "application/vnd.novadigm.ext": ["ext"], "application/vnd.oasis.opendocument.chart": ["odc"], "application/vnd.oasis.opendocument.chart-template": ["otc"], "application/vnd.oasis.opendocument.database": ["odb"], "application/vnd.oasis.opendocument.formula": ["odf"], "application/vnd.oasis.opendocument.formula-template": ["odft"], "application/vnd.oasis.opendocument.graphics": ["odg"], "application/vnd.oasis.opendocument.graphics-template": ["otg"], "application/vnd.oasis.opendocument.image": ["odi"], "application/vnd.oasis.opendocument.image-template": ["oti"], "application/vnd.oasis.opendocument.presentation": ["odp"], "application/vnd.oasis.opendocument.presentation-template": ["otp"], "application/vnd.oasis.opendocument.spreadsheet": ["ods"], "application/vnd.oasis.opendocument.spreadsheet-template": ["ots"], "application/vnd.oasis.opendocument.text": ["odt"], "application/vnd.oasis.opendocument.text-master": ["odm"], "application/vnd.oasis.opendocument.text-template": ["ott"], "application/vnd.oasis.opendocument.text-web": ["oth"], "application/vnd.olpc-sugar": ["xo"], "application/vnd.oma.dd2+xml": ["dd2"], "application/vnd.openofficeorg.extension": ["oxt"], "application/vnd.openxmlformats-officedocument.presentationml.presentation": ["pptx"], "application/vnd.openxmlformats-officedocument.presentationml.slide": ["sldx"], "application/vnd.openxmlformats-officedocument.presentationml.slideshow": ["ppsx"], "application/vnd.openxmlformats-officedocument.presentationml.template": ["potx"], "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": ["xlsx"], "application/vnd.openxmlformats-officedocument.spreadsheetml.template": ["xltx"], "application/vnd.openxmlformats-officedocument.wordprocessingml.document": ["docx"], "application/vnd.openxmlformats-officedocument.wordprocessingml.template": ["dotx"], "application/vnd.osgeo.mapguide.package": ["mgp"], "application/vnd.osgi.dp": ["dp"], "application/vnd.osgi.subsystem": ["esa"], "application/vnd.palm": ["pdb", "pqa", "oprc"], "application/vnd.pawaafile": ["paw"], "application/vnd.pg.format": ["str"], "application/vnd.pg.osasli": ["ei6"], "application/vnd.picsel": ["efif"], "application/vnd.pmi.widget": ["wg"], "application/vnd.pocketlearn": ["plf"], "application/vnd.powerbuilder6": ["pbd"], "application/vnd.previewsystems.box": ["box"], "application/vnd.proteus.magazine": ["mgz"], "application/vnd.publishare-delta-tree": ["qps"], "application/vnd.pvi.ptid1": ["ptid"], "application/vnd.quark.quarkxpress": ["qxd", "qxt", "qwd", "qwt", "qxl", "qxb"], "application/vnd.realvnc.bed": ["bed"], "application/vnd.recordare.musicxml": ["mxl"], "application/vnd.recordare.musicxml+xml": ["musicxml"], "application/vnd.rig.cryptonote": ["cryptonote"], "application/vnd.rim.cod": ["cod"], "application/vnd.rn-realmedia": ["rm"], "application/vnd.rn-realmedia-vbr": ["rmvb"], "application/vnd.route66.link66+xml": ["link66"], "application/vnd.sailingtracker.track": ["st"], "application/vnd.seemail": ["see"], "application/vnd.sema": ["sema"], "application/vnd.semd": ["semd"], "application/vnd.semf": ["semf"], "application/vnd.shana.informed.formdata": ["ifm"], "application/vnd.shana.informed.formtemplate": ["itp"], "application/vnd.shana.informed.interchange": ["iif"], "application/vnd.shana.informed.package": ["ipk"], "application/vnd.simtech-mindmapper": ["twd", "twds"], "application/vnd.smaf": ["mmf"], "application/vnd.smart.teacher": ["teacher"], "application/vnd.solent.sdkm+xml": ["sdkm", "sdkd"], "application/vnd.spotfire.dxp": ["dxp"], "application/vnd.spotfire.sfs": ["sfs"], "application/vnd.stardivision.calc": ["sdc"], "application/vnd.stardivision.draw": ["sda"], "application/vnd.stardivision.impress": ["sdd"], "application/vnd.stardivision.math": ["smf"], "application/vnd.stardivision.writer": ["sdw", "vor"], "application/vnd.stardivision.writer-global": ["sgl"], "application/vnd.stepmania.package": ["smzip"], "application/vnd.stepmania.stepchart": ["sm"], "application/vnd.sun.wadl+xml": ["wadl"], "application/vnd.sun.xml.calc": ["sxc"], "application/vnd.sun.xml.calc.template": ["stc"], "application/vnd.sun.xml.draw": ["sxd"], "application/vnd.sun.xml.draw.template": ["std"], "application/vnd.sun.xml.impress": ["sxi"], "application/vnd.sun.xml.impress.template": ["sti"], "application/vnd.sun.xml.math": ["sxm"], "application/vnd.sun.xml.writer": ["sxw"], "application/vnd.sun.xml.writer.global": ["sxg"], "application/vnd.sun.xml.writer.template": ["stw"], "application/vnd.sus-calendar": ["sus", "susp"], "application/vnd.svd": ["svd"], "application/vnd.symbian.install": ["sis", "sisx"], "application/vnd.syncml+xml": ["xsm"], "application/vnd.syncml.dm+wbxml": ["bdm"], "application/vnd.syncml.dm+xml": ["xdm"], "application/vnd.tao.intent-module-archive": ["tao"], "application/vnd.tcpdump.pcap": ["pcap", "cap", "dmp"], "application/vnd.tmobile-livetv": ["tmo"], "application/vnd.trid.tpt": ["tpt"], "application/vnd.triscape.mxs": ["mxs"], "application/vnd.trueapp": ["tra"], "application/vnd.ufdl": ["ufd", "ufdl"], "application/vnd.uiq.theme": ["utz"], "application/vnd.umajin": ["umj"], "application/vnd.unity": ["unityweb"], "application/vnd.uoml+xml": ["uoml"], "application/vnd.vcx": ["vcx"], "application/vnd.visio": ["vsd", "vst", "vss", "vsw"], "application/vnd.visionary": ["vis"], "application/vnd.vsf": ["vsf"], "application/vnd.wap.wbxml": ["wbxml"], "application/vnd.wap.wmlc": ["wmlc"], "application/vnd.wap.wmlscriptc": ["wmlsc"], "application/vnd.webturbo": ["wtb"], "application/vnd.wolfram.player": ["nbp"], "application/vnd.wordperfect": ["wpd"], "application/vnd.wqd": ["wqd"], "application/vnd.wt.stf": ["stf"], "application/vnd.xara": ["xar"], "application/vnd.xfdl": ["xfdl"], "application/vnd.yamaha.hv-dic": ["hvd"], "application/vnd.yamaha.hv-script": ["hvs"], "application/vnd.yamaha.hv-voice": ["hvp"], "application/vnd.yamaha.openscoreformat": ["osf"], "application/vnd.yamaha.openscoreformat.osfpvg+xml": ["osfpvg"], "application/vnd.yamaha.smaf-audio": ["saf"], "application/vnd.yamaha.smaf-phrase": ["spf"], "application/vnd.yellowriver-custom-menu": ["cmp"], "application/vnd.zul": ["zir", "zirz"], "application/vnd.zzazz.deck+xml": ["zaz"], "application/voicexml+xml": ["vxml"], "application/wasm": ["wasm"], "application/widget": ["wgt"], "application/winhlp": ["hlp"], "application/wsdl+xml": ["wsdl"], "application/wspolicy+xml": ["wspolicy"], "application/x-7z-compressed": ["7z"], "application/x-abiword": ["abw"], "application/x-ace-compressed": ["ace"], "application/x-apple-diskimage": [], "application/x-arj": ["arj"], "application/x-authorware-bin": ["aab", "x32", "u32", "vox"], "application/x-authorware-map": ["aam"], "application/x-authorware-seg": ["aas"], "application/x-bcpio": ["bcpio"], "application/x-bdoc": [], "application/x-bittorrent": ["torrent"], "application/x-blorb": ["blb", "blorb"], "application/x-bzip": ["bz"], "application/x-bzip2": ["bz2", "boz"], "application/x-cbr": ["cbr", "cba", "cbt", "cbz", "cb7"], "application/x-cdlink": ["vcd"], "application/x-cfs-compressed": ["cfs"], "application/x-chat": ["chat"], "application/x-chess-pgn": ["pgn"], "application/x-chrome-extension": ["crx"], "application/x-cocoa": ["cco"], "application/x-conference": ["nsc"], "application/x-cpio": ["cpio"], "application/x-csh": ["csh"], "application/x-debian-package": ["udeb"], "application/x-dgc-compressed": ["dgc"], "application/x-director": ["dir", "dcr", "dxr", "cst", "cct", "cxt", "w3d", "fgd", "swa"], "application/x-doom": ["wad"], "application/x-dtbncx+xml": ["ncx"], "application/x-dtbook+xml": ["dtb"], "application/x-dtbresource+xml": ["res"], "application/x-dvi": ["dvi"], "application/x-envoy": ["evy"], "application/x-eva": ["eva"], "application/x-font-bdf": ["bdf"], "application/x-font-ghostscript": ["gsf"], "application/x-font-linux-psf": ["psf"], "application/x-font-pcf": ["pcf"], "application/x-font-snf": ["snf"], "application/x-font-type1": ["pfa", "pfb", "pfm", "afm"], "application/x-freearc": ["arc"], "application/x-futuresplash": ["spl"], "application/x-gca-compressed": ["gca"], "application/x-glulx": ["ulx"], "application/x-gnumeric": ["gnumeric"], "application/x-gramps-xml": ["gramps"], "application/x-gtar": ["gtar"], "application/x-hdf": ["hdf"], "application/x-httpd-php": ["php"], "application/x-install-instructions": ["install"], "application/x-iso9660-image": [], "application/x-java-archive-diff": ["jardiff"], "application/x-java-jnlp-file": ["jnlp"], "application/x-latex": ["latex"], "application/x-lua-bytecode": ["luac"], "application/x-lzh-compressed": ["lzh", "lha"], "application/x-makeself": ["run"], "application/x-mie": ["mie"], "application/x-mobipocket-ebook": ["prc", "mobi"], "application/x-ms-application": ["application"], "application/x-ms-shortcut": ["lnk"], "application/x-ms-wmd": ["wmd"], "application/x-ms-wmz": ["wmz"], "application/x-ms-xbap": ["xbap"], "application/x-msaccess": ["mdb"], "application/x-msbinder": ["obd"], "application/x-mscardfile": ["crd"], "application/x-msclip": ["clp"], "application/x-msdos-program": [], "application/x-msdownload": ["com", "bat"], "application/x-msmediaview": ["mvb", "m13", "m14"], "application/x-msmetafile": ["wmf", "emf", "emz"], "application/x-msmoney": ["mny"], "application/x-mspublisher": ["pub"], "application/x-msschedule": ["scd"], "application/x-msterminal": ["trm"], "application/x-mswrite": ["wri"], "application/x-netcdf": ["nc", "cdf"], "application/x-ns-proxy-autoconfig": ["pac"], "application/x-nzb": ["nzb"], "application/x-perl": ["pl", "pm"], "application/x-pilot": [], "application/x-pkcs12": ["p12", "pfx"], "application/x-pkcs7-certificates": ["p7b", "spc"], "application/x-pkcs7-certreqresp": ["p7r"], "application/x-rar-compressed": ["rar"], "application/x-redhat-package-manager": ["rpm"], "application/x-research-info-systems": ["ris"], "application/x-sea": ["sea"], "application/x-sh": ["sh"], "application/x-shar": ["shar"], "application/x-shockwave-flash": ["swf"], "application/x-silverlight-app": ["xap"], "application/x-sql": ["sql"], "application/x-stuffit": ["sit"], "application/x-stuffitx": ["sitx"], "application/x-subrip": ["srt"], "application/x-sv4cpio": ["sv4cpio"], "application/x-sv4crc": ["sv4crc"], "application/x-t3vm-image": ["t3"], "application/x-tads": ["gam"], "application/x-tar": ["tar"], "application/x-tcl": ["tcl", "tk"], "application/x-tex": ["tex"], "application/x-tex-tfm": ["tfm"], "application/x-texinfo": ["texinfo", "texi"], "application/x-tgif": ["obj"], "application/x-ustar": ["ustar"], "application/x-virtualbox-hdd": ["hdd"], "application/x-virtualbox-ova": ["ova"], "application/x-virtualbox-ovf": ["ovf"], "application/x-virtualbox-vbox": ["vbox"], "application/x-virtualbox-vbox-extpack": ["vbox-extpack"], "application/x-virtualbox-vdi": ["vdi"], "application/x-virtualbox-vhd": ["vhd"], "application/x-virtualbox-vmdk": ["vmdk"], "application/x-wais-source": ["src"], "application/x-web-app-manifest+json": ["webapp"], "application/x-x509-ca-cert": ["der", "crt", "pem"], "application/x-xfig": ["fig"], "application/x-xliff+xml": ["xlf"], "application/x-xpinstall": ["xpi"], "application/x-xz": ["xz"], "application/x-zmachine": ["z1", "z2", "z3", "z4", "z5", "z6", "z7", "z8"], "application/xaml+xml": ["xaml"], "application/xcap-diff+xml": ["xdf"], "application/xenc+xml": ["xenc"], "application/xhtml+xml": ["xhtml", "xht"], "application/xml": ["xml", "xsl", "xsd", "rng"], "application/xml-dtd": ["dtd"], "application/xop+xml": ["xop"], "application/xproc+xml": ["xpl"], "application/xslt+xml": ["xslt"], "application/xspf+xml": ["xspf"], "application/xv+xml": ["mxml", "xhvml", "xvml", "xvm"], "application/yang": ["yang"], "application/yin+xml": ["yin"], "application/zip": ["zip"], "audio/3gpp": [], "audio/adpcm": ["adp"], "audio/basic": ["au", "snd"], "audio/midi": ["mid", "midi", "kar", "rmi"], "audio/mp3": [], "audio/mp4": ["m4a", "mp4a"], "audio/mpeg": ["mpga", "mp2", "mp2a", "mp3", "m2a", "m3a"], "audio/ogg": ["oga", "ogg", "spx"], "audio/s3m": ["s3m"], "audio/silk": ["sil"], "audio/vnd.dece.audio": ["uva", "uvva"], "audio/vnd.digital-winds": ["eol"], "audio/vnd.dra": ["dra"], "audio/vnd.dts": ["dts"], "audio/vnd.dts.hd": ["dtshd"], "audio/vnd.lucent.voice": ["lvp"], "audio/vnd.ms-playready.media.pya": ["pya"], "audio/vnd.nuera.ecelp4800": ["ecelp4800"], "audio/vnd.nuera.ecelp7470": ["ecelp7470"], "audio/vnd.nuera.ecelp9600": ["ecelp9600"], "audio/vnd.rip": ["rip"], "audio/wav": ["wav"], "audio/wave": [], "audio/webm": ["weba"], "audio/x-aac": ["aac"], "audio/x-aiff": ["aif", "aiff", "aifc"], "audio/x-caf": ["caf"], "audio/x-flac": ["flac"], "audio/x-m4a": [], "audio/x-matroska": ["mka"], "audio/x-mpegurl": ["m3u"], "audio/x-ms-wax": ["wax"], "audio/x-ms-wma": ["wma"], "audio/x-pn-realaudio": ["ram", "ra"], "audio/x-pn-realaudio-plugin": ["rmp"], "audio/x-realaudio": [], "audio/x-wav": [], "audio/xm": ["xm"], "chemical/x-cdx": ["cdx"], "chemical/x-cif": ["cif"], "chemical/x-cmdf": ["cmdf"], "chemical/x-cml": ["cml"], "chemical/x-csml": ["csml"], "chemical/x-xyz": ["xyz"], "font/collection": ["ttc"], "font/otf": ["otf"], "font/ttf": ["ttf"], "font/woff": ["woff"], "font/woff2": ["woff2"], "image/apng": ["apng"], "image/bmp": ["bmp"], "image/cgm": ["cgm"], "image/g3fax": ["g3"], "image/gif": ["gif"], "image/ief": ["ief"], "image/jp2": ["jp2", "jpg2"], "image/jpeg": ["jpeg", "jpg", "jpe"], "image/jpm": ["jpm"], "image/jpx": ["jpx", "jpf"], "image/ktx": ["ktx"], "image/png": ["png"], "image/prs.btif": ["btif"], "image/sgi": ["sgi"], "image/svg+xml": ["svg", "svgz"], "image/tiff": ["tiff", "tif"], "image/vnd.adobe.photoshop": ["psd"], "image/vnd.dece.graphic": ["uvi", "uvvi", "uvg", "uvvg"], "image/vnd.djvu": ["djvu", "djv"], "image/vnd.dvb.subtitle": [], "image/vnd.dwg": ["dwg"], "image/vnd.dxf": ["dxf"], "image/vnd.fastbidsheet": ["fbs"], "image/vnd.fpx": ["fpx"], "image/vnd.fst": ["fst"], "image/vnd.fujixerox.edmics-mmr": ["mmr"], "image/vnd.fujixerox.edmics-rlc": ["rlc"], "image/vnd.ms-modi": ["mdi"], "image/vnd.ms-photo": ["wdp"], "image/vnd.net-fpx": ["npx"], "image/vnd.wap.wbmp": ["wbmp"], "image/vnd.xiff": ["xif"], "image/webp": ["webp"], "image/x-3ds": ["3ds"], "image/x-cmu-raster": ["ras"], "image/x-cmx": ["cmx"], "image/x-freehand": ["fh", "fhc", "fh4", "fh5", "fh7"], "image/x-icon": ["ico"], "image/x-jng": ["jng"], "image/x-mrsid-image": ["sid"], "image/x-ms-bmp": [], "image/x-pcx": ["pcx"], "image/x-pict": ["pic", "pct"], "image/x-portable-anymap": ["pnm"], "image/x-portable-bitmap": ["pbm"], "image/x-portable-graymap": ["pgm"], "image/x-portable-pixmap": ["ppm"], "image/x-rgb": ["rgb"], "image/x-tga": ["tga"], "image/x-xbitmap": ["xbm"], "image/x-xpixmap": ["xpm"], "image/x-xwindowdump": ["xwd"], "message/rfc822": ["eml", "mime"], "model/gltf+json": ["gltf"], "model/gltf-binary": ["glb"], "model/iges": ["igs", "iges"], "model/mesh": ["msh", "mesh", "silo"], "model/vnd.collada+xml": ["dae"], "model/vnd.dwf": ["dwf"], "model/vnd.gdl": ["gdl"], "model/vnd.gtw": ["gtw"], "model/vnd.mts": ["mts"], "model/vnd.vtu": ["vtu"], "model/vrml": ["wrl", "vrml"], "model/x3d+binary": ["x3db", "x3dbz"], "model/x3d+vrml": ["x3dv", "x3dvz"], "model/x3d+xml": ["x3d", "x3dz"], "text/cache-manifest": ["appcache", "manifest"], "text/calendar": ["ics", "ifb"], "text/coffeescript": ["coffee", "litcoffee"], "text/css": ["css"], "text/csv": ["csv"], "text/hjson": ["hjson"], "text/html": ["html", "htm", "shtml"], "text/jade": ["jade"], "text/jsx": ["jsx"], "text/less": ["less"], "text/markdown": ["markdown", "md"], "text/mathml": ["mml"], "text/n3": ["n3"], "text/plain": ["txt", "text", "conf", "def", "list", "log", "in", "ini"], "text/prs.lines.tag": ["dsc"], "text/richtext": ["rtx"], "text/rtf": [], "text/sgml": ["sgml", "sgm"], "text/slim": ["slim", "slm"], "text/stylus": ["stylus", "styl"], "text/tab-separated-values": ["tsv"], "text/troff": ["t", "tr", "roff", "man", "me", "ms"], "text/turtle": ["ttl"], "text/uri-list": ["uri", "uris", "urls"], "text/vcard": ["vcard"], "text/vnd.curl": ["curl"], "text/vnd.curl.dcurl": ["dcurl"], "text/vnd.curl.mcurl": ["mcurl"], "text/vnd.curl.scurl": ["scurl"], "text/vnd.dvb.subtitle": ["sub"], "text/vnd.fly": ["fly"], "text/vnd.fmi.flexstor": ["flx"], "text/vnd.graphviz": ["gv"], "text/vnd.in3d.3dml": ["3dml"], "text/vnd.in3d.spot": ["spot"], "text/vnd.sun.j2me.app-descriptor": ["jad"], "text/vnd.wap.wml": ["wml"], "text/vnd.wap.wmlscript": ["wmls"], "text/vtt": ["vtt"], "text/x-asm": ["s", "asm"], "text/x-c": ["c", "cc", "cxx", "cpp", "h", "hh", "dic"], "text/x-component": ["htc"], "text/x-fortran": ["f", "for", "f77", "f90"], "text/x-handlebars-template": ["hbs"], "text/x-java-source": ["java"], "text/x-lua": ["lua"], "text/x-markdown": ["mkd"], "text/x-nfo": ["nfo"], "text/x-opml": ["opml"], "text/x-org": [], "text/x-pascal": ["p", "pas"], "text/x-processing": ["pde"], "text/x-sass": ["sass"], "text/x-scss": ["scss"], "text/x-setext": ["etx"], "text/x-sfv": ["sfv"], "text/x-suse-ymp": ["ymp"], "text/x-uuencode": ["uu"], "text/x-vcalendar": ["vcs"], "text/x-vcard": ["vcf"], "text/xml": [], "text/yaml": ["yaml", "yml"], "video/3gpp": ["3gp", "3gpp"], "video/3gpp2": ["3g2"], "video/h261": ["h261"], "video/h263": ["h263"], "video/h264": ["h264"], "video/jpeg": ["jpgv"], "video/jpm": ["jpgm"], "video/mj2": ["mj2", "mjp2"], "video/mp2t": ["ts"], "video/mp4": ["mp4", "mp4v", "mpg4"], "video/mpeg": ["mpeg", "mpg", "mpe", "m1v", "m2v"], "video/ogg": ["ogv"], "video/quicktime": ["qt", "mov"], "video/vnd.dece.hd": ["uvh", "uvvh"], "video/vnd.dece.mobile": ["uvm", "uvvm"], "video/vnd.dece.pd": ["uvp", "uvvp"], "video/vnd.dece.sd": ["uvs", "uvvs"], "video/vnd.dece.video": ["uvv", "uvvv"], "video/vnd.dvb.file": ["dvb"], "video/vnd.fvt": ["fvt"], "video/vnd.mpegurl": ["mxu", "m4u"], "video/vnd.ms-playready.media.pyv": ["pyv"], "video/vnd.uvvu.mp4": ["uvu", "uvvu"], "video/vnd.vivo": ["viv"], "video/webm": ["webm"], "video/x-f4v": ["f4v"], "video/x-fli": ["fli"], "video/x-flv": ["flv"], "video/x-m4v": ["m4v"], "video/x-matroska": ["mkv", "mk3d", "mks"], "video/x-mng": ["mng"], "video/x-ms-asf": ["asf", "asx"], "video/x-ms-vob": ["vob"], "video/x-ms-wm": ["wm"], "video/x-ms-wmv": ["wmv"], "video/x-ms-wmx": ["wmx"], "video/x-ms-wvx": ["wvx"], "video/x-msvideo": ["avi"], "video/x-sgi-movie": ["movie"], "video/x-smv": ["smv"], "x-conference/x-cooltalk": ["ice"] };
    module.exports = json;
},
/* less/lib/less-node/file-manager.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var path_1 = tslib_1.__importDefault(require("path"));
    var fs_1 = tslib_1.__importDefault(require(63) /* ./fs */);
    var abstract_file_manager_js_1 = tslib_1.__importDefault(require(68) /* ../less/environment/abstract-file-manager.js */);
    var FileManager = function () { };
    FileManager.prototype = Object.assign(new abstract_file_manager_js_1.default(), {
        supports: function () {
            return true;
        },
        supportsSync: function () {
            return true;
        },
        loadFile: function (filename, currentDirectory, options, environment, callback) {
            var fullFilename;
            var isAbsoluteFilename = this.isPathAbsolute(filename);
            var filenamesTried = [];
            var self = this;
            var prefix = filename.slice(0, 1);
            var explicit = prefix === '.' || prefix === '/';
            var result = null;
            var isNodeModule = false;
            var npmPrefix = 'npm://';
            options = options || {};
            var paths = isAbsoluteFilename ? [''] : [currentDirectory];
            if (options.paths) {
                paths.push.apply(paths, options.paths);
            }
            if (!isAbsoluteFilename && paths.indexOf('.') === -1) {
                paths.push('.');
            }
            var prefixes = options.prefixes || [''];
            var fileParts = this.extractUrlParts(filename);
            if (options.syncImport) {
                getFileData(returnData, returnData);
                if (callback) {
                    callback(result.error, result);
                }
                else {
                    return result;
                }
            }
            else {
                // promise is guaranteed to be asyncronous
                // which helps as it allows the file handle
                // to be closed before it continues with the next file
                return new Promise(getFileData);
            }
            function returnData(data) {
                if (!data.filename) {
                    result = { error: data };
                }
                else {
                    result = data;
                }
            }
            function getFileData(fulfill, reject) {
                (function tryPathIndex(i) {
                    if (i < paths.length) {
                        (function tryPrefix(j) {
                            if (j < prefixes.length) {
                                isNodeModule = false;
                                fullFilename = fileParts.rawPath + prefixes[j] + fileParts.filename;
                                if (paths[i]) {
                                    fullFilename = path_1.default.join(paths[i], fullFilename);
                                }
                                if (!explicit && paths[i] === '.') {
                                    try {
                                        fullFilename = require.resolve(fullFilename);
                                        isNodeModule = true;
                                    }
                                    catch (e) {
                                        filenamesTried.push(npmPrefix + fullFilename);
                                        tryWithExtension();
                                    }
                                }
                                else {
                                    tryWithExtension();
                                }
                                function tryWithExtension() {
                                    var extFilename = options.ext ? self.tryAppendExtension(fullFilename, options.ext) : fullFilename;
                                    if (extFilename !== fullFilename && !explicit && paths[i] === '.') {
                                        try {
                                            fullFilename = require.resolve(extFilename);
                                            isNodeModule = true;
                                        }
                                        catch (e) {
                                            filenamesTried.push(npmPrefix + extFilename);
                                            fullFilename = extFilename;
                                        }
                                    }
                                    else {
                                        fullFilename = extFilename;
                                    }
                                }
                                var readFileArgs = [fullFilename];
                                if (!options.rawBuffer) {
                                    readFileArgs.push('utf-8');
                                }
                                if (options.syncImport) {
                                    try {
                                        var data = fs_1.default.readFileSync.apply(this, readFileArgs);
                                        fulfill({ contents: data, filename: fullFilename });
                                    }
                                    catch (e) {
                                        filenamesTried.push(isNodeModule ? npmPrefix + fullFilename : fullFilename);
                                        return tryPrefix(j + 1);
                                    }
                                }
                                else {
                                    readFileArgs.push(function (e, data) {
                                        if (e) {
                                            filenamesTried.push(isNodeModule ? npmPrefix + fullFilename : fullFilename);
                                            return tryPrefix(j + 1);
                                        }
                                        fulfill({ contents: data, filename: fullFilename });
                                    });
                                    fs_1.default.readFile.apply(this, readFileArgs);
                                }
                            }
                            else {
                                tryPathIndex(i + 1);
                            }
                        })(0);
                    }
                    else {
                        reject({ type: 'File', message: "'" + filename + "' wasn't found. Tried - " + filenamesTried.join(',') });
                    }
                }(0));
            }
        },
        loadFileSync: function (filename, currentDirectory, options, environment) {
            options.syncImport = true;
            return this.loadFile(filename, currentDirectory, options, environment);
        }
    });
    exports.default = FileManager;
},
/* less/lib/less-node/fs.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var fs;
    try {
        fs = require(64) /* graceful-fs */;
    }
    catch (e) {
        fs = require('fs');
    }
    exports.default = fs;
},
/* graceful-fs/graceful-fs.js */ function _(require, module, exports, __esModule, __esExport) {
    var fs = require('fs');
    var polyfills = require(65) /* ./polyfills.js */;
    var legacy = require(66) /* ./legacy-streams.js */;
    var clone = require(67) /* ./clone.js */;
    var util = require('util');
    /* istanbul ignore next - node 0.x polyfill */
    var gracefulQueue;
    var previousSymbol;
    /* istanbul ignore else - node 0.x polyfill */
    if (typeof Symbol === 'function' && typeof Symbol.for === 'function') {
        gracefulQueue = Symbol.for('graceful-fs.queue');
        // This is used in testing by future versions
        previousSymbol = Symbol.for('graceful-fs.previous');
    }
    else {
        gracefulQueue = '___graceful-fs.queue';
        previousSymbol = '___graceful-fs.previous';
    }
    function noop() { }
    function publishQueue(context, queue) {
        Object.defineProperty(context, gracefulQueue, {
            get: function () {
                return queue;
            }
        });
    }
    var debug = noop;
    if (util.debuglog)
        debug = util.debuglog('gfs4');
    else if (/\bgfs4\b/i.test(process.env.NODE_DEBUG || ''))
        debug = function () {
            var m = util.format.apply(util, arguments);
            m = 'GFS4: ' + m.split(/\n/).join('\nGFS4: ');
            console.error(m);
        };
    // Once time initialization
    if (!fs[gracefulQueue]) {
        // This queue can be shared by multiple loaded instances
        var queue = global[gracefulQueue] || [];
        publishQueue(fs, queue);
        // Patch fs.close/closeSync to shared queue version, because we need
        // to retry() whenever a close happens *anywhere* in the program.
        // This is essential when multiple graceful-fs instances are
        // in play at the same time.
        fs.close = (function (fs$close) {
            function close(fd, cb) {
                return fs$close.call(fs, fd, function (err) {
                    // This function uses the graceful-fs shared queue
                    if (!err) {
                        retry();
                    }
                    if (typeof cb === 'function')
                        cb.apply(this, arguments);
                });
            }
            Object.defineProperty(close, previousSymbol, {
                value: fs$close
            });
            return close;
        })(fs.close);
        fs.closeSync = (function (fs$closeSync) {
            function closeSync(fd) {
                // This function uses the graceful-fs shared queue
                fs$closeSync.apply(fs, arguments);
                retry();
            }
            Object.defineProperty(closeSync, previousSymbol, {
                value: fs$closeSync
            });
            return closeSync;
        })(fs.closeSync);
        if (/\bgfs4\b/i.test(process.env.NODE_DEBUG || '')) {
            process.on('exit', function () {
                debug(fs[gracefulQueue]);
                require('assert').equal(fs[gracefulQueue].length, 0);
            });
        }
    }
    if (!global[gracefulQueue]) {
        publishQueue(global, fs[gracefulQueue]);
    }
    module.exports = patch(clone(fs));
    if (process.env.TEST_GRACEFUL_FS_GLOBAL_PATCH && !fs.__patched) {
        module.exports = patch(fs);
        fs.__patched = true;
    }
    function patch(fs) {
        // Everything that references the open() function needs to be in here
        polyfills(fs);
        fs.gracefulify = patch;
        fs.createReadStream = createReadStream;
        fs.createWriteStream = createWriteStream;
        var fs$readFile = fs.readFile;
        fs.readFile = readFile;
        function readFile(path, options, cb) {
            if (typeof options === 'function')
                cb = options, options = null;
            return go$readFile(path, options, cb);
            function go$readFile(path, options, cb) {
                return fs$readFile(path, options, function (err) {
                    if (err && (err.code === 'EMFILE' || err.code === 'ENFILE'))
                        enqueue([go$readFile, [path, options, cb]]);
                    else {
                        if (typeof cb === 'function')
                            cb.apply(this, arguments);
                        retry();
                    }
                });
            }
        }
        var fs$writeFile = fs.writeFile;
        fs.writeFile = writeFile;
        function writeFile(path, data, options, cb) {
            if (typeof options === 'function')
                cb = options, options = null;
            return go$writeFile(path, data, options, cb);
            function go$writeFile(path, data, options, cb) {
                return fs$writeFile(path, data, options, function (err) {
                    if (err && (err.code === 'EMFILE' || err.code === 'ENFILE'))
                        enqueue([go$writeFile, [path, data, options, cb]]);
                    else {
                        if (typeof cb === 'function')
                            cb.apply(this, arguments);
                        retry();
                    }
                });
            }
        }
        var fs$appendFile = fs.appendFile;
        if (fs$appendFile)
            fs.appendFile = appendFile;
        function appendFile(path, data, options, cb) {
            if (typeof options === 'function')
                cb = options, options = null;
            return go$appendFile(path, data, options, cb);
            function go$appendFile(path, data, options, cb) {
                return fs$appendFile(path, data, options, function (err) {
                    if (err && (err.code === 'EMFILE' || err.code === 'ENFILE'))
                        enqueue([go$appendFile, [path, data, options, cb]]);
                    else {
                        if (typeof cb === 'function')
                            cb.apply(this, arguments);
                        retry();
                    }
                });
            }
        }
        var fs$copyFile = fs.copyFile;
        if (fs$copyFile)
            fs.copyFile = copyFile;
        function copyFile(src, dest, flags, cb) {
            if (typeof flags === 'function') {
                cb = flags;
                flags = 0;
            }
            return fs$copyFile(src, dest, flags, function (err) {
                if (err && (err.code === 'EMFILE' || err.code === 'ENFILE'))
                    enqueue([fs$copyFile, [src, dest, flags, cb]]);
                else {
                    if (typeof cb === 'function')
                        cb.apply(this, arguments);
                    retry();
                }
            });
        }
        var fs$readdir = fs.readdir;
        fs.readdir = readdir;
        function readdir(path, options, cb) {
            var args = [path];
            if (typeof options !== 'function') {
                args.push(options);
            }
            else {
                cb = options;
            }
            args.push(go$readdir$cb);
            return go$readdir(args);
            function go$readdir$cb(err, files) {
                if (files && files.sort)
                    files.sort();
                if (err && (err.code === 'EMFILE' || err.code === 'ENFILE'))
                    enqueue([go$readdir, [args]]);
                else {
                    if (typeof cb === 'function')
                        cb.apply(this, arguments);
                    retry();
                }
            }
        }
        function go$readdir(args) {
            return fs$readdir.apply(fs, args);
        }
        if (process.version.substr(0, 4) === 'v0.8') {
            var legStreams = legacy(fs);
            ReadStream = legStreams.ReadStream;
            WriteStream = legStreams.WriteStream;
        }
        var fs$ReadStream = fs.ReadStream;
        if (fs$ReadStream) {
            ReadStream.prototype = Object.create(fs$ReadStream.prototype);
            ReadStream.prototype.open = ReadStream$open;
        }
        var fs$WriteStream = fs.WriteStream;
        if (fs$WriteStream) {
            WriteStream.prototype = Object.create(fs$WriteStream.prototype);
            WriteStream.prototype.open = WriteStream$open;
        }
        Object.defineProperty(fs, 'ReadStream', {
            get: function () {
                return ReadStream;
            },
            set: function (val) {
                ReadStream = val;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(fs, 'WriteStream', {
            get: function () {
                return WriteStream;
            },
            set: function (val) {
                WriteStream = val;
            },
            enumerable: true,
            configurable: true
        });
        // legacy names
        var FileReadStream = ReadStream;
        Object.defineProperty(fs, 'FileReadStream', {
            get: function () {
                return FileReadStream;
            },
            set: function (val) {
                FileReadStream = val;
            },
            enumerable: true,
            configurable: true
        });
        var FileWriteStream = WriteStream;
        Object.defineProperty(fs, 'FileWriteStream', {
            get: function () {
                return FileWriteStream;
            },
            set: function (val) {
                FileWriteStream = val;
            },
            enumerable: true,
            configurable: true
        });
        function ReadStream(path, options) {
            if (this instanceof ReadStream)
                return fs$ReadStream.apply(this, arguments), this;
            else
                return ReadStream.apply(Object.create(ReadStream.prototype), arguments);
        }
        function ReadStream$open() {
            var that = this;
            open(that.path, that.flags, that.mode, function (err, fd) {
                if (err) {
                    if (that.autoClose)
                        that.destroy();
                    that.emit('error', err);
                }
                else {
                    that.fd = fd;
                    that.emit('open', fd);
                    that.read();
                }
            });
        }
        function WriteStream(path, options) {
            if (this instanceof WriteStream)
                return fs$WriteStream.apply(this, arguments), this;
            else
                return WriteStream.apply(Object.create(WriteStream.prototype), arguments);
        }
        function WriteStream$open() {
            var that = this;
            open(that.path, that.flags, that.mode, function (err, fd) {
                if (err) {
                    that.destroy();
                    that.emit('error', err);
                }
                else {
                    that.fd = fd;
                    that.emit('open', fd);
                }
            });
        }
        function createReadStream(path, options) {
            return new fs.ReadStream(path, options);
        }
        function createWriteStream(path, options) {
            return new fs.WriteStream(path, options);
        }
        var fs$open = fs.open;
        fs.open = open;
        function open(path, flags, mode, cb) {
            if (typeof mode === 'function')
                cb = mode, mode = null;
            return go$open(path, flags, mode, cb);
            function go$open(path, flags, mode, cb) {
                return fs$open(path, flags, mode, function (err, fd) {
                    if (err && (err.code === 'EMFILE' || err.code === 'ENFILE'))
                        enqueue([go$open, [path, flags, mode, cb]]);
                    else {
                        if (typeof cb === 'function')
                            cb.apply(this, arguments);
                        retry();
                    }
                });
            }
        }
        return fs;
    }
    function enqueue(elem) {
        debug('ENQUEUE', elem[0].name, elem[1]);
        fs[gracefulQueue].push(elem);
    }
    function retry() {
        var elem = fs[gracefulQueue].shift();
        if (elem) {
            debug('RETRY', elem[0].name, elem[1]);
            elem[0].apply(null, elem[1]);
        }
    }
},
/* graceful-fs/polyfills.js */ function _(require, module, exports, __esModule, __esExport) {
    var constants = require('constants');
    var origCwd = process.cwd;
    var cwd = null;
    var platform = process.env.GRACEFUL_FS_PLATFORM || process.platform;
    process.cwd = function () {
        if (!cwd)
            cwd = origCwd.call(process);
        return cwd;
    };
    try {
        process.cwd();
    }
    catch (er) { }
    // This check is needed until node.js 12 is required
    if (typeof process.chdir === 'function') {
        var chdir = process.chdir;
        process.chdir = function (d) {
            cwd = null;
            chdir.call(process, d);
        };
        if (Object.setPrototypeOf)
            Object.setPrototypeOf(process.chdir, chdir);
    }
    module.exports = patch;
    function patch(fs) {
        // (re-)implement some things that are known busted or missing.
        // lchmod, broken prior to 0.6.2
        // back-port the fix here.
        if (constants.hasOwnProperty('O_SYMLINK') &&
            process.version.match(/^v0\.6\.[0-2]|^v0\.5\./)) {
            patchLchmod(fs);
        }
        // lutimes implementation, or no-op
        if (!fs.lutimes) {
            patchLutimes(fs);
        }
        // https://github.com/isaacs/node-graceful-fs/issues/4
        // Chown should not fail on einval or eperm if non-root.
        // It should not fail on enosys ever, as this just indicates
        // that a fs doesn't support the intended operation.
        fs.chown = chownFix(fs.chown);
        fs.fchown = chownFix(fs.fchown);
        fs.lchown = chownFix(fs.lchown);
        fs.chmod = chmodFix(fs.chmod);
        fs.fchmod = chmodFix(fs.fchmod);
        fs.lchmod = chmodFix(fs.lchmod);
        fs.chownSync = chownFixSync(fs.chownSync);
        fs.fchownSync = chownFixSync(fs.fchownSync);
        fs.lchownSync = chownFixSync(fs.lchownSync);
        fs.chmodSync = chmodFixSync(fs.chmodSync);
        fs.fchmodSync = chmodFixSync(fs.fchmodSync);
        fs.lchmodSync = chmodFixSync(fs.lchmodSync);
        fs.stat = statFix(fs.stat);
        fs.fstat = statFix(fs.fstat);
        fs.lstat = statFix(fs.lstat);
        fs.statSync = statFixSync(fs.statSync);
        fs.fstatSync = statFixSync(fs.fstatSync);
        fs.lstatSync = statFixSync(fs.lstatSync);
        // if lchmod/lchown do not exist, then make them no-ops
        if (!fs.lchmod) {
            fs.lchmod = function (path, mode, cb) {
                if (cb)
                    process.nextTick(cb);
            };
            fs.lchmodSync = function () { };
        }
        if (!fs.lchown) {
            fs.lchown = function (path, uid, gid, cb) {
                if (cb)
                    process.nextTick(cb);
            };
            fs.lchownSync = function () { };
        }
        // on Windows, A/V software can lock the directory, causing this
        // to fail with an EACCES or EPERM if the directory contains newly
        // created files.  Try again on failure, for up to 60 seconds.
        // Set the timeout this long because some Windows Anti-Virus, such as Parity
        // bit9, may lock files for up to a minute, causing npm package install
        // failures. Also, take care to yield the scheduler. Windows scheduling gives
        // CPU to a busy looping process, which can cause the program causing the lock
        // contention to be starved of CPU by node, so the contention doesn't resolve.
        if (platform === "win32") {
            fs.rename = (function (fs$rename) {
                return function (from, to, cb) {
                    var start = Date.now();
                    var backoff = 0;
                    fs$rename(from, to, function CB(er) {
                        if (er
                            && (er.code === "EACCES" || er.code === "EPERM")
                            && Date.now() - start < 60000) {
                            setTimeout(function () {
                                fs.stat(to, function (stater, st) {
                                    if (stater && stater.code === "ENOENT")
                                        fs$rename(from, to, CB);
                                    else
                                        cb(er);
                                });
                            }, backoff);
                            if (backoff < 100)
                                backoff += 10;
                            return;
                        }
                        if (cb)
                            cb(er);
                    });
                };
            })(fs.rename);
        }
        // if read() returns EAGAIN, then just try it again.
        fs.read = (function (fs$read) {
            function read(fd, buffer, offset, length, position, callback_) {
                var callback;
                if (callback_ && typeof callback_ === 'function') {
                    var eagCounter = 0;
                    callback = function (er, _, __) {
                        if (er && er.code === 'EAGAIN' && eagCounter < 10) {
                            eagCounter++;
                            return fs$read.call(fs, fd, buffer, offset, length, position, callback);
                        }
                        callback_.apply(this, arguments);
                    };
                }
                return fs$read.call(fs, fd, buffer, offset, length, position, callback);
            }
            // This ensures `util.promisify` works as it does for native `fs.read`.
            if (Object.setPrototypeOf)
                Object.setPrototypeOf(read, fs$read);
            return read;
        })(fs.read);
        fs.readSync = (function (fs$readSync) {
            return function (fd, buffer, offset, length, position) {
                var eagCounter = 0;
                while (true) {
                    try {
                        return fs$readSync.call(fs, fd, buffer, offset, length, position);
                    }
                    catch (er) {
                        if (er.code === 'EAGAIN' && eagCounter < 10) {
                            eagCounter++;
                            continue;
                        }
                        throw er;
                    }
                }
            };
        })(fs.readSync);
        function patchLchmod(fs) {
            fs.lchmod = function (path, mode, callback) {
                fs.open(path, constants.O_WRONLY | constants.O_SYMLINK, mode, function (err, fd) {
                    if (err) {
                        if (callback)
                            callback(err);
                        return;
                    }
                    // prefer to return the chmod error, if one occurs,
                    // but still try to close, and report closing errors if they occur.
                    fs.fchmod(fd, mode, function (err) {
                        fs.close(fd, function (err2) {
                            if (callback)
                                callback(err || err2);
                        });
                    });
                });
            };
            fs.lchmodSync = function (path, mode) {
                var fd = fs.openSync(path, constants.O_WRONLY | constants.O_SYMLINK, mode);
                // prefer to return the chmod error, if one occurs,
                // but still try to close, and report closing errors if they occur.
                var threw = true;
                var ret;
                try {
                    ret = fs.fchmodSync(fd, mode);
                    threw = false;
                }
                finally {
                    if (threw) {
                        try {
                            fs.closeSync(fd);
                        }
                        catch (er) { }
                    }
                    else {
                        fs.closeSync(fd);
                    }
                }
                return ret;
            };
        }
        function patchLutimes(fs) {
            if (constants.hasOwnProperty("O_SYMLINK")) {
                fs.lutimes = function (path, at, mt, cb) {
                    fs.open(path, constants.O_SYMLINK, function (er, fd) {
                        if (er) {
                            if (cb)
                                cb(er);
                            return;
                        }
                        fs.futimes(fd, at, mt, function (er) {
                            fs.close(fd, function (er2) {
                                if (cb)
                                    cb(er || er2);
                            });
                        });
                    });
                };
                fs.lutimesSync = function (path, at, mt) {
                    var fd = fs.openSync(path, constants.O_SYMLINK);
                    var ret;
                    var threw = true;
                    try {
                        ret = fs.futimesSync(fd, at, mt);
                        threw = false;
                    }
                    finally {
                        if (threw) {
                            try {
                                fs.closeSync(fd);
                            }
                            catch (er) { }
                        }
                        else {
                            fs.closeSync(fd);
                        }
                    }
                    return ret;
                };
            }
            else {
                fs.lutimes = function (_a, _b, _c, cb) { if (cb)
                    process.nextTick(cb); };
                fs.lutimesSync = function () { };
            }
        }
        function chmodFix(orig) {
            if (!orig)
                return orig;
            return function (target, mode, cb) {
                return orig.call(fs, target, mode, function (er) {
                    if (chownErOk(er))
                        er = null;
                    if (cb)
                        cb.apply(this, arguments);
                });
            };
        }
        function chmodFixSync(orig) {
            if (!orig)
                return orig;
            return function (target, mode) {
                try {
                    return orig.call(fs, target, mode);
                }
                catch (er) {
                    if (!chownErOk(er))
                        throw er;
                }
            };
        }
        function chownFix(orig) {
            if (!orig)
                return orig;
            return function (target, uid, gid, cb) {
                return orig.call(fs, target, uid, gid, function (er) {
                    if (chownErOk(er))
                        er = null;
                    if (cb)
                        cb.apply(this, arguments);
                });
            };
        }
        function chownFixSync(orig) {
            if (!orig)
                return orig;
            return function (target, uid, gid) {
                try {
                    return orig.call(fs, target, uid, gid);
                }
                catch (er) {
                    if (!chownErOk(er))
                        throw er;
                }
            };
        }
        function statFix(orig) {
            if (!orig)
                return orig;
            // Older versions of Node erroneously returned signed integers for
            // uid + gid.
            return function (target, options, cb) {
                if (typeof options === 'function') {
                    cb = options;
                    options = null;
                }
                function callback(er, stats) {
                    if (stats) {
                        if (stats.uid < 0)
                            stats.uid += 0x100000000;
                        if (stats.gid < 0)
                            stats.gid += 0x100000000;
                    }
                    if (cb)
                        cb.apply(this, arguments);
                }
                return options ? orig.call(fs, target, options, callback)
                    : orig.call(fs, target, callback);
            };
        }
        function statFixSync(orig) {
            if (!orig)
                return orig;
            // Older versions of Node erroneously returned signed integers for
            // uid + gid.
            return function (target, options) {
                var stats = options ? orig.call(fs, target, options)
                    : orig.call(fs, target);
                if (stats.uid < 0)
                    stats.uid += 0x100000000;
                if (stats.gid < 0)
                    stats.gid += 0x100000000;
                return stats;
            };
        }
        // ENOSYS means that the fs doesn't support the op. Just ignore
        // that, because it doesn't matter.
        //
        // if there's no getuid, or if getuid() is something other
        // than 0, and the error is EINVAL or EPERM, then just ignore
        // it.
        //
        // This specific case is a silent failure in cp, install, tar,
        // and most other unix tools that manage permissions.
        //
        // When running as root, or if other types of errors are
        // encountered, then it's strict.
        function chownErOk(er) {
            if (!er)
                return true;
            if (er.code === "ENOSYS")
                return true;
            var nonroot = !process.getuid || process.getuid() !== 0;
            if (nonroot) {
                if (er.code === "EINVAL" || er.code === "EPERM")
                    return true;
            }
            return false;
        }
    }
},
/* graceful-fs/legacy-streams.js */ function _(require, module, exports, __esModule, __esExport) {
    var Stream = require('stream').Stream;
    module.exports = legacy;
    function legacy(fs) {
        return {
            ReadStream: ReadStream,
            WriteStream: WriteStream
        };
        function ReadStream(path, options) {
            if (!(this instanceof ReadStream))
                return new ReadStream(path, options);
            Stream.call(this);
            var self = this;
            this.path = path;
            this.fd = null;
            this.readable = true;
            this.paused = false;
            this.flags = 'r';
            this.mode = 438; /*=0666*/
            this.bufferSize = 64 * 1024;
            options = options || {};
            // Mixin options into this
            var keys = Object.keys(options);
            for (var index = 0, length = keys.length; index < length; index++) {
                var key = keys[index];
                this[key] = options[key];
            }
            if (this.encoding)
                this.setEncoding(this.encoding);
            if (this.start !== undefined) {
                if ('number' !== typeof this.start) {
                    throw TypeError('start must be a Number');
                }
                if (this.end === undefined) {
                    this.end = Infinity;
                }
                else if ('number' !== typeof this.end) {
                    throw TypeError('end must be a Number');
                }
                if (this.start > this.end) {
                    throw new Error('start must be <= end');
                }
                this.pos = this.start;
            }
            if (this.fd !== null) {
                process.nextTick(function () {
                    self._read();
                });
                return;
            }
            fs.open(this.path, this.flags, this.mode, function (err, fd) {
                if (err) {
                    self.emit('error', err);
                    self.readable = false;
                    return;
                }
                self.fd = fd;
                self.emit('open', fd);
                self._read();
            });
        }
        function WriteStream(path, options) {
            if (!(this instanceof WriteStream))
                return new WriteStream(path, options);
            Stream.call(this);
            this.path = path;
            this.fd = null;
            this.writable = true;
            this.flags = 'w';
            this.encoding = 'binary';
            this.mode = 438; /*=0666*/
            this.bytesWritten = 0;
            options = options || {};
            // Mixin options into this
            var keys = Object.keys(options);
            for (var index = 0, length = keys.length; index < length; index++) {
                var key = keys[index];
                this[key] = options[key];
            }
            if (this.start !== undefined) {
                if ('number' !== typeof this.start) {
                    throw TypeError('start must be a Number');
                }
                if (this.start < 0) {
                    throw new Error('start must be >= zero');
                }
                this.pos = this.start;
            }
            this.busy = false;
            this._queue = [];
            if (this.fd === null) {
                this._open = fs.open;
                this._queue.push([this._open, this.path, this.flags, this.mode, undefined]);
                this.flush();
            }
        }
    }
},
/* graceful-fs/clone.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = clone;
    var getPrototypeOf = Object.getPrototypeOf || function (obj) {
        return obj.__proto__;
    };
    function clone(obj) {
        if (obj === null || typeof obj !== 'object')
            return obj;
        if (obj instanceof Object)
            var copy = { __proto__: getPrototypeOf(obj) };
        else
            var copy = Object.create(null);
        Object.getOwnPropertyNames(obj).forEach(function (key) {
            Object.defineProperty(copy, key, Object.getOwnPropertyDescriptor(obj, key));
        });
        return copy;
    }
},
/* less/lib/less/environment/abstract-file-manager.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var AbstractFileManager = /** @class */ (function () {
        function AbstractFileManager() {
        }
        AbstractFileManager.prototype.getPath = function (filename) {
            var j = filename.lastIndexOf('?');
            if (j > 0) {
                filename = filename.slice(0, j);
            }
            j = filename.lastIndexOf('/');
            if (j < 0) {
                j = filename.lastIndexOf('\\');
            }
            if (j < 0) {
                return '';
            }
            return filename.slice(0, j + 1);
        };
        AbstractFileManager.prototype.tryAppendExtension = function (path, ext) {
            return /(\.[a-z]*$)|([\?;].*)$/.test(path) ? path : path + ext;
        };
        AbstractFileManager.prototype.tryAppendLessExtension = function (path) {
            return this.tryAppendExtension(path, '.less');
        };
        AbstractFileManager.prototype.supportsSync = function () {
            return false;
        };
        AbstractFileManager.prototype.alwaysMakePathsAbsolute = function () {
            return false;
        };
        AbstractFileManager.prototype.isPathAbsolute = function (filename) {
            return (/^(?:[a-z-]+:|\/|\\|#)/i).test(filename);
        };
        // TODO: pull out / replace?
        AbstractFileManager.prototype.join = function (basePath, laterPath) {
            if (!basePath) {
                return laterPath;
            }
            return basePath + laterPath;
        };
        AbstractFileManager.prototype.pathDiff = function (url, baseUrl) {
            // diff between two paths to create a relative path
            var urlParts = this.extractUrlParts(url);
            var baseUrlParts = this.extractUrlParts(baseUrl);
            var i;
            var max;
            var urlDirectories;
            var baseUrlDirectories;
            var diff = '';
            if (urlParts.hostPart !== baseUrlParts.hostPart) {
                return '';
            }
            max = Math.max(baseUrlParts.directories.length, urlParts.directories.length);
            for (i = 0; i < max; i++) {
                if (baseUrlParts.directories[i] !== urlParts.directories[i]) {
                    break;
                }
            }
            baseUrlDirectories = baseUrlParts.directories.slice(i);
            urlDirectories = urlParts.directories.slice(i);
            for (i = 0; i < baseUrlDirectories.length - 1; i++) {
                diff += '../';
            }
            for (i = 0; i < urlDirectories.length - 1; i++) {
                diff += urlDirectories[i] + "/";
            }
            return diff;
        };
        // helper function, not part of API
        AbstractFileManager.prototype.extractUrlParts = function (url, baseUrl) {
            // urlParts[1] = protocol://hostname/ OR /
            // urlParts[2] = / if path relative to host base
            // urlParts[3] = directories
            // urlParts[4] = filename
            // urlParts[5] = parameters
            var urlPartsRegex = /^((?:[a-z-]+:)?\/{2}(?:[^\/\?#]*\/)|([\/\\]))?((?:[^\/\\\?#]*[\/\\])*)([^\/\\\?#]*)([#\?].*)?$/i;
            var urlParts = url.match(urlPartsRegex);
            var returner = {};
            var rawDirectories = [];
            var directories = [];
            var i;
            var baseUrlParts;
            if (!urlParts) {
                throw new Error("Could not parse sheet href - '" + url + "'");
            }
            // Stylesheets in IE don't always return the full path
            if (baseUrl && (!urlParts[1] || urlParts[2])) {
                baseUrlParts = baseUrl.match(urlPartsRegex);
                if (!baseUrlParts) {
                    throw new Error("Could not parse page url - '" + baseUrl + "'");
                }
                urlParts[1] = urlParts[1] || baseUrlParts[1] || '';
                if (!urlParts[2]) {
                    urlParts[3] = baseUrlParts[3] + urlParts[3];
                }
            }
            if (urlParts[3]) {
                rawDirectories = urlParts[3].replace(/\\/g, '/').split('/');
                // collapse '..' and skip '.'
                for (i = 0; i < rawDirectories.length; i++) {
                    if (rawDirectories[i] === '..') {
                        directories.pop();
                    }
                    else if (rawDirectories[i] !== '.') {
                        directories.push(rawDirectories[i]);
                    }
                }
            }
            returner.hostPart = urlParts[1];
            returner.directories = directories;
            returner.rawPath = (urlParts[1] || '') + rawDirectories.join('/');
            returner.path = (urlParts[1] || '') + directories.join('/');
            returner.filename = urlParts[4];
            returner.fileUrl = returner.path + (urlParts[4] || '');
            returner.url = returner.fileUrl + (urlParts[5] || '');
            return returner;
        };
        return AbstractFileManager;
    }());
    exports.default = AbstractFileManager;
},
/* less/lib/less-node/url-file-manager.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var isUrlRe = /^(?:https?:)?\/\//i;
    var url_1 = tslib_1.__importDefault(require("url"));
    var request;
    var abstract_file_manager_js_1 = tslib_1.__importDefault(require(68) /* ../less/environment/abstract-file-manager.js */);
    var logger_1 = tslib_1.__importDefault(require(70) /* ../less/logger */);
    var UrlFileManager = function () { };
    UrlFileManager.prototype = Object.assign(new abstract_file_manager_js_1.default(), {
        supports: function (filename, currentDirectory, options, environment) {
            return isUrlRe.test(filename) || isUrlRe.test(currentDirectory);
        },
        loadFile: function (filename, currentDirectory, options, environment) {
            return new Promise(function (fulfill, reject) {
                if (request === undefined) {
                    try {
                        request = require(71) /* needle */;
                    }
                    catch (e) {
                        request = null;
                    }
                }
                if (!request) {
                    reject({ type: 'File', message: 'optional dependency \'native-request\' required to import over http(s)\n' });
                    return;
                }
                var urlStr = isUrlRe.test(filename) ? filename : url_1.default.resolve(currentDirectory, filename);
                /** native-request currently has a bug */
                var hackUrlStr = urlStr.indexOf('?') === -1 ? urlStr + '?' : urlStr;
                request.get(hackUrlStr, { follow_max: 5 }, function (err, resp, body) {
                    if (err || resp && resp.statusCode >= 400) {
                        var message = resp && resp.statusCode === 404
                            ? "resource '" + urlStr + "' was not found\n"
                            : "resource '" + urlStr + "' gave this Error:\n  " + (err || resp.statusMessage || resp.statusCode) + "\n";
                        reject({ type: 'File', message: message });
                        return;
                    }
                    if (resp.statusCode >= 300) {
                        reject({ type: 'File', message: "resource '" + urlStr + "' caused too many redirects" });
                        return;
                    }
                    body = body.toString('utf8');
                    if (!body) {
                        logger_1.default.warn("Warning: Empty body (HTTP " + resp.statusCode + ") returned by \"" + urlStr + "\"");
                    }
                    fulfill({ contents: body || '', filename: urlStr });
                });
            });
        }
    });
    exports.default = UrlFileManager;
},
/* less/lib/less/logger.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.default = {
        error: function (msg) {
            this._fireEvent('error', msg);
        },
        warn: function (msg) {
            this._fireEvent('warn', msg);
        },
        info: function (msg) {
            this._fireEvent('info', msg);
        },
        debug: function (msg) {
            this._fireEvent('debug', msg);
        },
        addListener: function (listener) {
            this._listeners.push(listener);
        },
        removeListener: function (listener) {
            for (var i = 0; i < this._listeners.length; i++) {
                if (this._listeners[i] === listener) {
                    this._listeners.splice(i, 1);
                    return;
                }
            }
        },
        _fireEvent: function (type, msg) {
            for (var i = 0; i < this._listeners.length; i++) {
                var logFunction = this._listeners[i][type];
                if (logFunction) {
                    logFunction(msg);
                }
            }
        },
        _listeners: []
    };
},
/* needle/lib/needle.js */ function _(require, module, exports, __esModule, __esExport) {
    //////////////////////////////////////////
    // Needle -- HTTP Client for Node.js
    // Written by Tomás Pollak <tomas@forkhq.com>
    // (c) 2012-2020 - Fork Ltd.
    // MIT Licensed
    //////////////////////////////////////////
    var fs = require('fs'), http = require('http'), https = require('https'), url = require('url'), stream = require('stream'), debug = require(72) /* debug */('needle'), stringify = require(77) /* ./querystring */.build, multipart = require(78) /* ./multipart */, auth = require(79) /* ./auth */, cookies = require(80) /* ./cookies */, parsers = require(81) /* ./parsers */, decoder = require(83) /* ./decoder */;
    //////////////////////////////////////////
    // variabilia
    var version = require(106) /* ../package.json */.version;
    var user_agent = 'Needle/' + version;
    user_agent += ' (Node.js ' + process.version + '; ' + process.platform + ' ' + process.arch + ')';
    var tls_options = 'agent pfx key passphrase cert ca ciphers rejectUnauthorized secureProtocol checkServerIdentity family';
    // older versions of node (< 0.11.4) prevent the runtime from exiting
    // because of connections in keep-alive state. so if this is the case
    // we'll default new requests to set a Connection: close header.
    var close_by_default = !http.Agent || http.Agent.defaultMaxSockets != Infinity;
    // see if we have Object.assign. otherwise fall back to util._extend
    var extend = Object.assign ? Object.assign : require('util')._extend;
    // these are the status codes that Needle interprets as redirects.
    var redirect_codes = [301, 302, 303, 307, 308];
    //////////////////////////////////////////
    // decompressors for gzip/deflate/br bodies
    function bind_opts(fn, options) {
        return fn.bind(null, options);
    }
    var decompressors = {};
    try {
        var zlib = require('zlib');
        // Enable Z_SYNC_FLUSH to avoid Z_BUF_ERROR errors (Node PR #2595)
        var zlib_options = {
            flush: zlib.Z_SYNC_FLUSH,
            finishFlush: zlib.Z_SYNC_FLUSH
        };
        var br_options = {
            flush: zlib.BROTLI_OPERATION_FLUSH,
            finishFlush: zlib.BROTLI_OPERATION_FLUSH
        };
        decompressors['x-deflate'] = bind_opts(zlib.Inflate, zlib_options);
        decompressors['deflate'] = bind_opts(zlib.Inflate, zlib_options);
        decompressors['x-gzip'] = bind_opts(zlib.Gunzip, zlib_options);
        decompressors['gzip'] = bind_opts(zlib.Gunzip, zlib_options);
        if (typeof zlib.BrotliDecompress === 'function') {
            decompressors['br'] = bind_opts(zlib.BrotliDecompress, br_options);
        }
    }
    catch (e) { /* zlib not available */ }
    //////////////////////////////////////////
    // options and aliases
    var defaults = {
        // data
        boundary: '--------------------NODENEEDLEHTTPCLIENT',
        encoding: 'utf8',
        parse_response: 'all',
        proxy: null,
        // headers
        headers: {},
        accept: '*/*',
        user_agent: user_agent,
        // numbers
        open_timeout: 10000,
        response_timeout: 0,
        read_timeout: 0,
        follow_max: 0,
        stream_length: -1,
        // booleans
        compressed: false,
        decode_response: true,
        parse_cookies: true,
        follow_set_cookies: false,
        follow_set_referer: false,
        follow_keep_method: false,
        follow_if_same_host: false,
        follow_if_same_protocol: false,
        follow_if_same_location: false
    };
    var aliased = {
        options: {
            decode: 'decode_response',
            parse: 'parse_response',
            timeout: 'open_timeout',
            follow: 'follow_max'
        },
        inverted: {}
    };
    // only once, invert aliased keys so we can get passed options.
    Object.keys(aliased.options).map(function (k) {
        var value = aliased.options[k];
        aliased.inverted[value] = k;
    });
    //////////////////////////////////////////
    // helpers
    function keys_by_type(type) {
        return Object.keys(defaults).map(function (el) {
            if (defaults[el] !== null && defaults[el].constructor == type)
                return el;
        }).filter(function (el) { return el; });
    }
    function parse_content_type(header) {
        if (!header || header === '')
            return {};
        var found, charset = 'utf8', arr = header.split(';');
        if (arr.length > 1 && (found = arr[1].match(/charset=(.+)/)))
            charset = found[1];
        return { type: arr[0], charset: charset };
    }
    function is_stream(obj) {
        return typeof obj.pipe === 'function';
    }
    function get_stream_length(stream, given_length, cb) {
        if (given_length > 0)
            return cb(given_length);
        if (stream.end !== void 0 && stream.end !== Infinity && stream.start !== void 0)
            return cb((stream.end + 1) - (stream.start || 0));
        fs.stat(stream.path, function (err, stat) {
            cb(stat ? stat.size - (stream.start || 0) : null);
        });
    }
    //////////////////////////////////////////
    // the main act
    function Needle(method, uri, data, options, callback) {
        // if (!(this instanceof Needle)) {
        //   return new Needle(method, uri, data, options, callback);
        // }
        if (typeof uri !== 'string')
            throw new TypeError('URL must be a string, not ' + uri);
        this.method = method;
        this.uri = uri;
        this.data = data;
        if (typeof options == 'function') {
            this.callback = options;
            this.options = {};
        }
        else {
            this.callback = callback;
            this.options = options;
        }
    }
    Needle.prototype.setup = function (uri, options) {
        function get_option(key, fallback) {
            // if original is in options, return that value
            if (typeof options[key] != 'undefined')
                return options[key];
            // otherwise, return value from alias or fallback/undefined
            return typeof options[aliased.inverted[key]] != 'undefined'
                ? options[aliased.inverted[key]] : fallback;
        }
        function check_value(expected, key) {
            var value = get_option(key), type = typeof value;
            if (type != 'undefined' && type != expected)
                throw new TypeError(type + ' received for ' + key + ', but expected a ' + expected);
            return (type == expected) ? value : defaults[key];
        }
        //////////////////////////////////////////////////
        // the basics
        var config = {
            http_opts: {
                localAddress: get_option('localAddress', undefined)
            },
            headers: {},
            output: options.output,
            proxy: get_option('proxy', defaults.proxy),
            parser: get_option('parse_response', defaults.parse_response),
            encoding: options.encoding || (options.multipart ? 'binary' : defaults.encoding)
        };
        keys_by_type(Boolean).forEach(function (key) {
            config[key] = check_value('boolean', key);
        });
        keys_by_type(Number).forEach(function (key) {
            config[key] = check_value('number', key);
        });
        // populate http_opts with given TLS options
        tls_options.split(' ').forEach(function (key) {
            if (typeof options[key] != 'undefined') {
                config.http_opts[key] = options[key];
                if (typeof options.agent == 'undefined')
                    config.http_opts.agent = false; // otherwise tls options are skipped
            }
        });
        //////////////////////////////////////////////////
        // headers, cookies
        for (var key in defaults.headers)
            config.headers[key] = defaults.headers[key];
        config.headers['accept'] = options.accept || defaults.accept;
        config.headers['user-agent'] = options.user_agent || defaults.user_agent;
        if (options.content_type)
            config.headers['content-type'] = options.content_type;
        // set connection header if opts.connection was passed, or if node < 0.11.4 (close)
        if (options.connection || close_by_default)
            config.headers['connection'] = options.connection || 'close';
        if ((options.compressed || defaults.compressed) && typeof zlib != 'undefined')
            config.headers['accept-encoding'] = decompressors['br'] ? 'gzip, deflate, br' : 'gzip, deflate';
        if (options.cookies)
            config.headers['cookie'] = cookies.write(options.cookies);
        //////////////////////////////////////////////////
        // basic/digest auth
        if (uri.match(/[^\/]@/)) { // url contains user:pass@host, so parse it.
            var parts = (url.parse(uri).auth || '').split(':');
            options.username = parts[0];
            options.password = parts[1];
        }
        if (options.username) {
            if (options.auth && (options.auth == 'auto' || options.auth == 'digest')) {
                config.credentials = [options.username, options.password];
            }
            else {
                config.headers['authorization'] = auth.basic(options.username, options.password);
            }
        }
        // if proxy is present, set auth header from either url or proxy_user option.
        if (config.proxy) {
            if (config.proxy.indexOf('http') === -1)
                config.proxy = 'http://' + config.proxy;
            if (config.proxy.indexOf('@') !== -1) {
                var proxy = (url.parse(config.proxy).auth || '').split(':');
                options.proxy_user = proxy[0];
                options.proxy_pass = proxy[1];
            }
            if (options.proxy_user)
                config.headers['proxy-authorization'] = auth.basic(options.proxy_user, options.proxy_pass);
        }
        // now that all our headers are set, overwrite them if instructed.
        for (var h in options.headers)
            config.headers[h.toLowerCase()] = options.headers[h];
        config.uri_modifier = get_option('uri_modifier', null);
        return config;
    };
    Needle.prototype.start = function () {
        var out = new stream.PassThrough({ objectMode: false }), uri = this.uri, data = this.data, method = this.method, callback = (typeof this.options == 'function') ? this.options : this.callback, options = this.options || {};
        // if no 'http' is found on URL, prepend it.
        if (uri.indexOf('http') === -1)
            uri = uri.replace(/^(\/\/)?/, 'http://');
        var self = this, body, waiting = false, config = this.setup(uri, options);
        // unless options.json was set to false, assume boss also wants JSON if content-type matches.
        var json = options.json || (options.json !== false && config.headers['content-type'] == 'application/json');
        if (data) {
            if (options.multipart) { // boss says we do multipart. so we do it.
                var boundary = options.boundary || defaults.boundary;
                waiting = true;
                multipart.build(data, boundary, function (err, parts) {
                    if (err)
                        throw (err);
                    config.headers['content-type'] = 'multipart/form-data; boundary=' + boundary;
                    next(parts);
                });
            }
            else if (is_stream(data)) {
                if (method.toUpperCase() == 'GET')
                    throw new Error('Refusing to pipe() a stream via GET. Did you mean .post?');
                if (config.stream_length > 0 || (config.stream_length === 0 && data.path)) {
                    // ok, let's get the stream's length and set it as the content-length header.
                    // this prevents some servers from cutting us off before all the data is sent.
                    waiting = true;
                    get_stream_length(data, config.stream_length, function (length) {
                        data.length = length;
                        next(data);
                    });
                }
                else {
                    // if the boss doesn't want us to get the stream's length, or if it doesn't
                    // have a file descriptor for that purpose, then just head on.
                    body = data;
                }
            }
            else if (Buffer.isBuffer(data)) {
                body = data; // use the raw buffer as request body.
            }
            else if (method.toUpperCase() == 'GET' && !json) {
                // append the data to the URI as a querystring.
                uri = uri.replace(/\?.*|$/, '?' + stringify(data));
            }
            else { // string or object data, no multipart.
                // if string, leave it as it is, otherwise, stringify.
                body = (typeof (data) === 'string') ? data
                    : json ? JSON.stringify(data) : stringify(data);
                // ensure we have a buffer so bytecount is correct.
                body = Buffer.from(body, config.encoding);
            }
        }
        function next(body) {
            if (body) {
                if (body.length)
                    config.headers['content-length'] = body.length;
                // if no content-type was passed, determine if json or not.
                if (!config.headers['content-type']) {
                    config.headers['content-type'] = json
                        ? 'application/json; charset=utf-8'
                        : 'application/x-www-form-urlencoded'; // no charset says W3 spec.
                }
            }
            // unless a specific accept header was set, assume json: true wants JSON back.
            if (options.json && (!options.accept && !(options.headers || {}).accept))
                config.headers['accept'] = 'application/json';
            self.send_request(1, method, uri, config, body, out, callback);
        }
        if (!waiting)
            next(body);
        return out;
    };
    Needle.prototype.get_request_opts = function (method, uri, config) {
        var opts = config.http_opts, proxy = config.proxy, remote = proxy ? url.parse(proxy) : url.parse(uri);
        opts.protocol = remote.protocol;
        opts.host = remote.hostname;
        opts.port = remote.port || (remote.protocol == 'https:' ? 443 : 80);
        opts.path = proxy ? uri : remote.pathname + (remote.search || '');
        opts.method = method;
        opts.headers = config.headers;
        if (!opts.headers['host']) {
            // if using proxy, make sure the host header shows the final destination
            var target = proxy ? url.parse(uri) : remote;
            opts.headers['host'] = target.hostname;
            // and if a non standard port was passed, append it to the port header
            if (target.port && [80, 443].indexOf(target.port) === -1) {
                opts.headers['host'] += ':' + target.port;
            }
        }
        return opts;
    };
    Needle.prototype.should_follow = function (location, config, original) {
        if (!location)
            return false;
        // returns true if location contains matching property (host or protocol)
        function matches(property) {
            var property = original[property];
            return location.indexOf(property) !== -1;
        }
        // first, check whether the requested location is actually different from the original
        if (!config.follow_if_same_location && location === original)
            return false;
        if (config.follow_if_same_host && !matches('host'))
            return false; // host does not match, so not following
        if (config.follow_if_same_protocol && !matches('protocol'))
            return false; // procotol does not match, so not following
        return true;
    };
    Needle.prototype.send_request = function (count, method, uri, config, post_data, out, callback) {
        if (typeof config.uri_modifier === 'function') {
            var modified_uri = config.uri_modifier(uri);
            debug('Modifying request URI', uri + ' => ' + modified_uri);
            uri = modified_uri;
        }
        var timer, returned = 0, self = this, request_opts = this.get_request_opts(method, uri, config), protocol = request_opts.protocol == 'https:' ? https : http;
        function done(err, resp) {
            if (returned++ > 0)
                return debug('Already finished, stopping here.');
            if (timer)
                clearTimeout(timer);
            request.removeListener('error', had_error);
            if (callback)
                return callback(err, resp, resp ? resp.body : undefined);
            // NOTE: this event used to be called 'end', but the behaviour was confusing
            // when errors ocurred, because the stream would still emit an 'end' event.
            out.emit('done', err);
        }
        function had_error(err) {
            debug('Request error', err);
            out.emit('err', err);
            done(err || new Error('Unknown error when making request.'));
        }
        function set_timeout(type, milisecs) {
            if (timer)
                clearTimeout(timer);
            if (milisecs <= 0)
                return;
            timer = setTimeout(function () {
                out.emit('timeout', type);
                request.abort();
                // also invoke done() to terminate job on read_timeout
                if (type == 'read')
                    done(new Error(type + ' timeout'));
            }, milisecs);
        }
        // handle errors on the underlying socket, that may be closed while writing
        // for an example case, see test/long_string_spec.js. we make sure this
        // scenario ocurred by verifying the socket's writable & destroyed states.
        function on_socket_end() {
            if (returned && !this.writable && this.destroyed === false) {
                this.destroy();
                had_error(new Error('Remote end closed socket abruptly.'));
            }
        }
        debug('Making request #' + count, request_opts);
        var request = protocol.request(request_opts, function (resp) {
            var headers = resp.headers;
            debug('Got response', resp.statusCode, headers);
            out.emit('response', resp);
            set_timeout('read', config.read_timeout);
            // if we got cookies, parse them unless we were instructed not to. make sure to include any
            // cookies that might have been set on previous redirects.
            if (config.parse_cookies && (headers['set-cookie'] || config.previous_resp_cookies)) {
                resp.cookies = extend(config.previous_resp_cookies || {}, cookies.read(headers['set-cookie']));
                debug('Got cookies', resp.cookies);
            }
            // if redirect code is found, determine if we should follow it according to the given options.
            if (redirect_codes.indexOf(resp.statusCode) !== -1 && self.should_follow(headers.location, config, uri)) {
                // clear timer before following redirects to prevent unexpected setTimeout consequence
                clearTimeout(timer);
                if (count <= config.follow_max) {
                    out.emit('redirect', headers.location);
                    // unless 'follow_keep_method' is true, rewrite the request to GET before continuing.
                    if (!config.follow_keep_method) {
                        method = 'GET';
                        post_data = null;
                        delete config.headers['content-length']; // in case the original was a multipart POST request.
                    }
                    // if follow_set_cookies is true, insert cookies in the next request's headers.
                    // we set both the original request cookies plus any response cookies we might have received.
                    if (config.follow_set_cookies) {
                        var request_cookies = cookies.read(config.headers['cookie']);
                        config.previous_resp_cookies = resp.cookies;
                        if (Object.keys(request_cookies).length || Object.keys(resp.cookies || {}).length) {
                            config.headers['cookie'] = cookies.write(extend(request_cookies, resp.cookies));
                        }
                    }
                    else if (config.headers['cookie']) {
                        debug('Clearing original request cookie', config.headers['cookie']);
                        delete config.headers['cookie'];
                    }
                    if (config.follow_set_referer)
                        config.headers['referer'] = encodeURI(uri); // the original, not the destination URL.
                    config.headers['host'] = null; // clear previous Host header to avoid conflicts.
                    debug('Redirecting to ' + url.resolve(uri, headers.location));
                    return self.send_request(++count, method, url.resolve(uri, headers.location), config, post_data, out, callback);
                }
                else if (config.follow_max > 0) {
                    return done(new Error('Max redirects reached. Possible loop in: ' + headers.location));
                }
            }
            // if auth is requested and credentials were not passed, resend request, provided we have user/pass.
            if (resp.statusCode == 401 && headers['www-authenticate'] && config.credentials) {
                if (!config.headers['authorization']) { // only if authentication hasn't been sent
                    var auth_header = auth.header(headers['www-authenticate'], config.credentials, request_opts);
                    if (auth_header) {
                        config.headers['authorization'] = auth_header;
                        return self.send_request(count, method, uri, config, post_data, out, callback);
                    }
                }
            }
            // ok, so we got a valid (non-redirect & authorized) response. let's notify the stream guys.
            out.emit('header', resp.statusCode, headers);
            out.emit('headers', headers);
            var pipeline = [], mime = parse_content_type(headers['content-type']), text_response = mime.type && mime.type.indexOf('text/') != -1;
            // To start, if our body is compressed and we're able to inflate it, do it.
            if (headers['content-encoding'] && decompressors[headers['content-encoding']]) {
                var decompressor = decompressors[headers['content-encoding']]();
                // make sure we catch errors triggered by the decompressor.
                decompressor.on('error', had_error);
                pipeline.push(decompressor);
            }
            // If parse is enabled and we have a parser for it, then go for it.
            if (config.parser && parsers[mime.type]) {
                // If a specific parser was requested, make sure we don't parse other types.
                var parser_name = config.parser.toString().toLowerCase();
                if (['xml', 'json'].indexOf(parser_name) == -1 || parsers[mime.type].name == parser_name) {
                    // OK, so either we're parsing all content types or the one requested matches.
                    out.parser = parsers[mime.type].name;
                    pipeline.push(parsers[mime.type].fn());
                    // Set objectMode on out stream to improve performance.
                    out._writableState.objectMode = true;
                    out._readableState.objectMode = true;
                }
                // If we're not parsing, and unless decoding was disabled, we'll try
                // decoding non UTF-8 bodies to UTF-8, using the iconv-lite library.
            }
            else if (text_response && config.decode_response
                && mime.charset) {
                pipeline.push(decoder(mime.charset));
            }
            // And `out` is the stream we finally push the decoded/parsed output to.
            pipeline.push(out);
            // Now, release the kraken!
            var tmp = resp;
            while (pipeline.length) {
                tmp = tmp.pipe(pipeline.shift());
            }
            // If the user has requested and output file, pipe the output stream to it.
            // In stream mode, we will still get the response stream to play with.
            if (config.output && resp.statusCode == 200) {
                // for some reason, simply piping resp to the writable stream doesn't
                // work all the time (stream gets cut in the middle with no warning).
                // so we'll manually need to do the readable/write(chunk) trick.
                var file = fs.createWriteStream(config.output);
                file.on('error', had_error);
                out.on('end', function () {
                    if (file.writable)
                        file.end();
                });
                file.on('close', function () {
                    delete out.file;
                });
                out.on('readable', function () {
                    var chunk;
                    while ((chunk = this.read()) !== null) {
                        if (file.writable)
                            file.write(chunk);
                        // if callback was requested, also push it to resp.body
                        if (resp.body)
                            resp.body.push(chunk);
                    }
                });
                out.file = file;
            }
            // Only aggregate the full body if a callback was requested.
            if (callback) {
                resp.raw = [];
                resp.body = [];
                resp.bytes = 0;
                // Gather and count the amount of (raw) bytes using a PassThrough stream.
                var clean_pipe = new stream.PassThrough();
                resp.pipe(clean_pipe);
                clean_pipe.on('readable', function () {
                    var chunk;
                    while ((chunk = this.read()) != null) {
                        resp.bytes += chunk.length;
                        resp.raw.push(chunk);
                    }
                });
                // Listen on the 'readable' event to aggregate the chunks, but only if
                // file output wasn't requested. Otherwise we'd have two stream readers.
                if (!config.output || resp.statusCode != 200) {
                    out.on('readable', function () {
                        var chunk;
                        while ((chunk = this.read()) !== null) {
                            // We're either pushing buffers or objects, never strings.
                            if (typeof chunk == 'string')
                                chunk = Buffer.from(chunk);
                            // Push all chunks to resp.body. We'll bind them in resp.end().
                            resp.body.push(chunk);
                        }
                    });
                }
            }
            // And set the .body property once all data is in.
            out.on('end', function () {
                if (resp.body) { // callback mode
                    // we want to be able to access to the raw data later, so keep a reference.
                    resp.raw = Buffer.concat(resp.raw);
                    // if parse was successful, we should have an array with one object
                    if (resp.body[0] !== undefined && !Buffer.isBuffer(resp.body[0])) {
                        // that's our body right there.
                        resp.body = resp.body[0];
                        // set the parser property on our response. we may want to check.
                        if (out.parser)
                            resp.parser = out.parser;
                    }
                    else { // we got one or several buffers. string or binary.
                        resp.body = Buffer.concat(resp.body);
                        // if we're here and parsed is true, it means we tried to but it didn't work.
                        // so given that we got a text response, let's stringify it.
                        if (text_response || out.parser) {
                            resp.body = resp.body.toString();
                        }
                    }
                }
                // if an output file is being written to, make sure the callback
                // is triggered after all data has been written to it.
                if (out.file) {
                    out.file.on('close', function () {
                        done(null, resp, resp.body);
                    });
                }
                else { // elvis has left the building.
                    done(null, resp, resp.body);
                }
            });
        }); // end request call
        // unless open_timeout was disabled, set a timeout to abort the request.
        set_timeout('open', config.open_timeout);
        // handle errors on the request object. things might get bumpy.
        request.on('error', had_error);
        // make sure timer is cleared if request is aborted (issue #257)
        request.once('abort', function () {
            if (timer)
                clearTimeout(timer);
        });
        // handle socket 'end' event to ensure we don't get delayed EPIPE errors.
        request.once('socket', function (socket) {
            if (socket.connecting) {
                socket.once('connect', function () {
                    set_timeout('response', config.response_timeout);
                });
            }
            else {
                set_timeout('response', config.response_timeout);
            }
            // console.log(socket);
            if (!socket.on_socket_end) {
                socket.on_socket_end = on_socket_end;
                socket.once('end', function () { process.nextTick(on_socket_end.bind(socket)); });
            }
        });
        if (post_data) {
            if (is_stream(post_data)) {
                post_data.pipe(request);
            }
            else {
                request.write(post_data, config.encoding);
                request.end();
            }
        }
        else {
            request.end();
        }
        out.request = request;
        return out;
    };
    //////////////////////////////////////////
    // exports
    if (typeof Promise !== 'undefined') {
        module.exports = function () {
            var verb, args = [].slice.call(arguments);
            if (args[0].match(/\.|\//)) // first argument looks like a URL
                verb = (args.length > 2) ? 'post' : 'get';
            else
                verb = args.shift();
            if (verb.match(/get|head/) && args.length == 2)
                args.splice(1, 0, null); // assume no data if head/get with two args (url, options)
            return new Promise(function (resolve, reject) {
                module.exports.request(verb, args[0], args[1], args[2], function (err, resp) {
                    return err ? reject(err) : resolve(resp);
                });
            });
        };
    }
    module.exports.version = version;
    module.exports.defaults = function (obj) {
        for (var key in obj) {
            var target_key = aliased.options[key] || key;
            if (defaults.hasOwnProperty(target_key) && typeof obj[key] != 'undefined') {
                if (target_key != 'parse_response' && target_key != 'proxy') {
                    // ensure type matches the original, except for proxy/parse_response that can be null/bool or string
                    var valid_type = defaults[target_key].constructor.name;
                    if (obj[key].constructor.name != valid_type)
                        throw new TypeError('Invalid type for ' + key + ', should be ' + valid_type);
                }
                defaults[target_key] = obj[key];
            }
            else {
                throw new Error('Invalid property for defaults:' + target_key);
            }
        }
        return defaults;
    };
    'head get'.split(' ').forEach(function (method) {
        module.exports[method] = function (uri, options, callback) {
            return new Needle(method, uri, null, options, callback).start();
        };
    });
    'post put patch delete'.split(' ').forEach(function (method) {
        module.exports[method] = function (uri, data, options, callback) {
            return new Needle(method, uri, data, options, callback).start();
        };
    });
    module.exports.request = function (method, uri, data, opts, callback) {
        return new Needle(method, uri, data, opts, callback).start();
    };
},
/* debug/src/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Detect Electron renderer / nwjs process, which is node, but we should
     * treat as a browser.
     */
    if (typeof process === 'undefined' || process.type === 'renderer' || process.browser === true || process.__nwjs) {
        module.exports = require(73) /* ./browser.js */;
    }
    else {
        module.exports = require(76) /* ./node.js */;
    }
},
/* debug/src/browser.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint-env browser */
    /**
     * This is the web browser implementation of `debug()`.
     */
    exports.formatArgs = formatArgs;
    exports.save = save;
    exports.load = load;
    exports.useColors = useColors;
    exports.storage = localstorage();
    exports.destroy = (() => {
        let warned = false;
        return () => {
            if (!warned) {
                warned = true;
                console.warn('Instance method `debug.destroy()` is deprecated and no longer does anything. It will be removed in the next major version of `debug`.');
            }
        };
    })();
    /**
     * Colors.
     */
    exports.colors = [
        '#0000CC',
        '#0000FF',
        '#0033CC',
        '#0033FF',
        '#0066CC',
        '#0066FF',
        '#0099CC',
        '#0099FF',
        '#00CC00',
        '#00CC33',
        '#00CC66',
        '#00CC99',
        '#00CCCC',
        '#00CCFF',
        '#3300CC',
        '#3300FF',
        '#3333CC',
        '#3333FF',
        '#3366CC',
        '#3366FF',
        '#3399CC',
        '#3399FF',
        '#33CC00',
        '#33CC33',
        '#33CC66',
        '#33CC99',
        '#33CCCC',
        '#33CCFF',
        '#6600CC',
        '#6600FF',
        '#6633CC',
        '#6633FF',
        '#66CC00',
        '#66CC33',
        '#9900CC',
        '#9900FF',
        '#9933CC',
        '#9933FF',
        '#99CC00',
        '#99CC33',
        '#CC0000',
        '#CC0033',
        '#CC0066',
        '#CC0099',
        '#CC00CC',
        '#CC00FF',
        '#CC3300',
        '#CC3333',
        '#CC3366',
        '#CC3399',
        '#CC33CC',
        '#CC33FF',
        '#CC6600',
        '#CC6633',
        '#CC9900',
        '#CC9933',
        '#CCCC00',
        '#CCCC33',
        '#FF0000',
        '#FF0033',
        '#FF0066',
        '#FF0099',
        '#FF00CC',
        '#FF00FF',
        '#FF3300',
        '#FF3333',
        '#FF3366',
        '#FF3399',
        '#FF33CC',
        '#FF33FF',
        '#FF6600',
        '#FF6633',
        '#FF9900',
        '#FF9933',
        '#FFCC00',
        '#FFCC33'
    ];
    /**
     * Currently only WebKit-based Web Inspectors, Firefox >= v31,
     * and the Firebug extension (any Firefox version) are known
     * to support "%c" CSS customizations.
     *
     * TODO: add a `localStorage` variable to explicitly enable/disable colors
     */
    // eslint-disable-next-line complexity
    function useColors() {
        // NB: In an Electron preload script, document will be defined but not fully
        // initialized. Since we know we're in Chrome, we'll just detect this case
        // explicitly
        if (typeof window !== 'undefined' && window.process && (window.process.type === 'renderer' || window.process.__nwjs)) {
            return true;
        }
        // Internet Explorer and Edge do not support colors.
        if (typeof navigator !== 'undefined' && navigator.userAgent && navigator.userAgent.toLowerCase().match(/(edge|trident)\/(\d+)/)) {
            return false;
        }
        // Is webkit? http://stackoverflow.com/a/16459606/376773
        // document is undefined in react-native: https://github.com/facebook/react-native/pull/1632
        return (typeof document !== 'undefined' && document.documentElement && document.documentElement.style && document.documentElement.style.WebkitAppearance) ||
            // Is firebug? http://stackoverflow.com/a/398120/376773
            (typeof window !== 'undefined' && window.console && (window.console.firebug || (window.console.exception && window.console.table))) ||
            // Is firefox >= v31?
            // https://developer.mozilla.org/en-US/docs/Tools/Web_Console#Styling_messages
            (typeof navigator !== 'undefined' && navigator.userAgent && navigator.userAgent.toLowerCase().match(/firefox\/(\d+)/) && parseInt(RegExp.$1, 10) >= 31) ||
            // Double check webkit in userAgent just in case we are in a worker
            (typeof navigator !== 'undefined' && navigator.userAgent && navigator.userAgent.toLowerCase().match(/applewebkit\/(\d+)/));
    }
    /**
     * Colorize log arguments if enabled.
     *
     * @api public
     */
    function formatArgs(args) {
        args[0] = (this.useColors ? '%c' : '') +
            this.namespace +
            (this.useColors ? ' %c' : ' ') +
            args[0] +
            (this.useColors ? '%c ' : ' ') +
            '+' + module.exports.humanize(this.diff);
        if (!this.useColors) {
            return;
        }
        const c = 'color: ' + this.color;
        args.splice(1, 0, c, 'color: inherit');
        // The final "%c" is somewhat tricky, because there could be other
        // arguments passed either before or after the %c, so we need to
        // figure out the correct index to insert the CSS into
        let index = 0;
        let lastC = 0;
        args[0].replace(/%[a-zA-Z%]/g, match => {
            if (match === '%%') {
                return;
            }
            index++;
            if (match === '%c') {
                // We only are interested in the *last* %c
                // (the user may have provided their own)
                lastC = index;
            }
        });
        args.splice(lastC, 0, c);
    }
    /**
     * Invokes `console.debug()` when available.
     * No-op when `console.debug` is not a "function".
     * If `console.debug` is not available, falls back
     * to `console.log`.
     *
     * @api public
     */
    exports.log = console.debug || console.log || (() => { });
    /**
     * Save `namespaces`.
     *
     * @param {String} namespaces
     * @api private
     */
    function save(namespaces) {
        try {
            if (namespaces) {
                exports.storage.setItem('debug', namespaces);
            }
            else {
                exports.storage.removeItem('debug');
            }
        }
        catch (error) {
            // Swallow
            // XXX (@Qix-) should we be logging these?
        }
    }
    /**
     * Load `namespaces`.
     *
     * @return {String} returns the previously persisted debug modes
     * @api private
     */
    function load() {
        let r;
        try {
            r = exports.storage.getItem('debug');
        }
        catch (error) {
            // Swallow
            // XXX (@Qix-) should we be logging these?
        }
        // If debug isn't set in LS, and we're in Electron, try to load $DEBUG
        if (!r && typeof process !== 'undefined' && 'env' in process) {
            r = process.env.DEBUG;
        }
        return r;
    }
    /**
     * Localstorage attempts to return the localstorage.
     *
     * This is necessary because safari throws
     * when a user disables cookies/localstorage
     * and you attempt to access it.
     *
     * @return {LocalStorage}
     * @api private
     */
    function localstorage() {
        try {
            // TVMLKit (Apple TV JS Runtime) does not have a window object, just localStorage in the global context
            // The Browser also has localStorage in the global context.
            return localStorage;
        }
        catch (error) {
            // Swallow
            // XXX (@Qix-) should we be logging these?
        }
    }
    module.exports = require(74) /* ./common */(exports);
    const { formatters } = module.exports;
    /**
     * Map %j to `JSON.stringify()`, since no Web Inspectors do that by default.
     */
    formatters.j = function (v) {
        try {
            return JSON.stringify(v);
        }
        catch (error) {
            return '[UnexpectedJSONParseError]: ' + error.message;
        }
    };
},
/* debug/src/common.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * This is the common logic for both the Node.js and web browser
     * implementations of `debug()`.
     */
    function setup(env) {
        createDebug.debug = createDebug;
        createDebug.default = createDebug;
        createDebug.coerce = coerce;
        createDebug.disable = disable;
        createDebug.enable = enable;
        createDebug.enabled = enabled;
        createDebug.humanize = require(75) /* ms */;
        createDebug.destroy = destroy;
        Object.keys(env).forEach(key => {
            createDebug[key] = env[key];
        });
        /**
        * The currently active debug mode names, and names to skip.
        */
        createDebug.names = [];
        createDebug.skips = [];
        /**
        * Map of special "%n" handling functions, for the debug "format" argument.
        *
        * Valid key names are a single, lower or upper-case letter, i.e. "n" and "N".
        */
        createDebug.formatters = {};
        /**
        * Selects a color for a debug namespace
        * @param {String} namespace The namespace string for the for the debug instance to be colored
        * @return {Number|String} An ANSI color code for the given namespace
        * @api private
        */
        function selectColor(namespace) {
            let hash = 0;
            for (let i = 0; i < namespace.length; i++) {
                hash = ((hash << 5) - hash) + namespace.charCodeAt(i);
                hash |= 0; // Convert to 32bit integer
            }
            return createDebug.colors[Math.abs(hash) % createDebug.colors.length];
        }
        createDebug.selectColor = selectColor;
        /**
        * Create a debugger with the given `namespace`.
        *
        * @param {String} namespace
        * @return {Function}
        * @api public
        */
        function createDebug(namespace) {
            let prevTime;
            let enableOverride = null;
            function debug(...args) {
                // Disabled?
                if (!debug.enabled) {
                    return;
                }
                const self = debug;
                // Set `diff` timestamp
                const curr = Number(new Date());
                const ms = curr - (prevTime || curr);
                self.diff = ms;
                self.prev = prevTime;
                self.curr = curr;
                prevTime = curr;
                args[0] = createDebug.coerce(args[0]);
                if (typeof args[0] !== 'string') {
                    // Anything else let's inspect with %O
                    args.unshift('%O');
                }
                // Apply any `formatters` transformations
                let index = 0;
                args[0] = args[0].replace(/%([a-zA-Z%])/g, (match, format) => {
                    // If we encounter an escaped % then don't increase the array index
                    if (match === '%%') {
                        return '%';
                    }
                    index++;
                    const formatter = createDebug.formatters[format];
                    if (typeof formatter === 'function') {
                        const val = args[index];
                        match = formatter.call(self, val);
                        // Now we need to remove `args[index]` since it's inlined in the `format`
                        args.splice(index, 1);
                        index--;
                    }
                    return match;
                });
                // Apply env-specific formatting (colors, etc.)
                createDebug.formatArgs.call(self, args);
                const logFn = self.log || createDebug.log;
                logFn.apply(self, args);
            }
            debug.namespace = namespace;
            debug.useColors = createDebug.useColors();
            debug.color = createDebug.selectColor(namespace);
            debug.extend = extend;
            debug.destroy = createDebug.destroy; // XXX Temporary. Will be removed in the next major release.
            Object.defineProperty(debug, 'enabled', {
                enumerable: true,
                configurable: false,
                get: () => enableOverride === null ? createDebug.enabled(namespace) : enableOverride,
                set: v => {
                    enableOverride = v;
                }
            });
            // Env-specific initialization logic for debug instances
            if (typeof createDebug.init === 'function') {
                createDebug.init(debug);
            }
            return debug;
        }
        function extend(namespace, delimiter) {
            const newDebug = createDebug(this.namespace + (typeof delimiter === 'undefined' ? ':' : delimiter) + namespace);
            newDebug.log = this.log;
            return newDebug;
        }
        /**
        * Enables a debug mode by namespaces. This can include modes
        * separated by a colon and wildcards.
        *
        * @param {String} namespaces
        * @api public
        */
        function enable(namespaces) {
            createDebug.save(namespaces);
            createDebug.names = [];
            createDebug.skips = [];
            let i;
            const split = (typeof namespaces === 'string' ? namespaces : '').split(/[\s,]+/);
            const len = split.length;
            for (i = 0; i < len; i++) {
                if (!split[i]) {
                    // ignore empty strings
                    continue;
                }
                namespaces = split[i].replace(/\*/g, '.*?');
                if (namespaces[0] === '-') {
                    createDebug.skips.push(new RegExp('^' + namespaces.substr(1) + '$'));
                }
                else {
                    createDebug.names.push(new RegExp('^' + namespaces + '$'));
                }
            }
        }
        /**
        * Disable debug output.
        *
        * @return {String} namespaces
        * @api public
        */
        function disable() {
            const namespaces = [
                ...createDebug.names.map(toNamespace),
                ...createDebug.skips.map(toNamespace).map(namespace => '-' + namespace)
            ].join(',');
            createDebug.enable('');
            return namespaces;
        }
        /**
        * Returns true if the given mode name is enabled, false otherwise.
        *
        * @param {String} name
        * @return {Boolean}
        * @api public
        */
        function enabled(name) {
            if (name[name.length - 1] === '*') {
                return true;
            }
            let i;
            let len;
            for (i = 0, len = createDebug.skips.length; i < len; i++) {
                if (createDebug.skips[i].test(name)) {
                    return false;
                }
            }
            for (i = 0, len = createDebug.names.length; i < len; i++) {
                if (createDebug.names[i].test(name)) {
                    return true;
                }
            }
            return false;
        }
        /**
        * Convert regexp to namespace
        *
        * @param {RegExp} regxep
        * @return {String} namespace
        * @api private
        */
        function toNamespace(regexp) {
            return regexp.toString()
                .substring(2, regexp.toString().length - 2)
                .replace(/\.\*\?$/, '*');
        }
        /**
        * Coerce `val`.
        *
        * @param {Mixed} val
        * @return {Mixed}
        * @api private
        */
        function coerce(val) {
            if (val instanceof Error) {
                return val.stack || val.message;
            }
            return val;
        }
        /**
        * XXX DO NOT USE. This is a temporary stub function.
        * XXX It WILL be removed in the next major release.
        */
        function destroy() {
            console.warn('Instance method `debug.destroy()` is deprecated and no longer does anything. It will be removed in the next major version of `debug`.');
        }
        createDebug.enable(createDebug.load());
        return createDebug;
    }
    module.exports = setup;
},
/* ms/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Helpers.
     */
    var s = 1000;
    var m = s * 60;
    var h = m * 60;
    var d = h * 24;
    var w = d * 7;
    var y = d * 365.25;
    /**
     * Parse or format the given `val`.
     *
     * Options:
     *
     *  - `long` verbose formatting [false]
     *
     * @param {String|Number} val
     * @param {Object} [options]
     * @throws {Error} throw an error if val is not a non-empty string or a number
     * @return {String|Number}
     * @api public
     */
    module.exports = function (val, options) {
        options = options || {};
        var type = typeof val;
        if (type === 'string' && val.length > 0) {
            return parse(val);
        }
        else if (type === 'number' && isFinite(val)) {
            return options.long ? fmtLong(val) : fmtShort(val);
        }
        throw new Error('val is not a non-empty string or a valid number. val=' +
            JSON.stringify(val));
    };
    /**
     * Parse the given `str` and return milliseconds.
     *
     * @param {String} str
     * @return {Number}
     * @api private
     */
    function parse(str) {
        str = String(str);
        if (str.length > 100) {
            return;
        }
        var match = /^(-?(?:\d+)?\.?\d+) *(milliseconds?|msecs?|ms|seconds?|secs?|s|minutes?|mins?|m|hours?|hrs?|h|days?|d|weeks?|w|years?|yrs?|y)?$/i.exec(str);
        if (!match) {
            return;
        }
        var n = parseFloat(match[1]);
        var type = (match[2] || 'ms').toLowerCase();
        switch (type) {
            case 'years':
            case 'year':
            case 'yrs':
            case 'yr':
            case 'y':
                return n * y;
            case 'weeks':
            case 'week':
            case 'w':
                return n * w;
            case 'days':
            case 'day':
            case 'd':
                return n * d;
            case 'hours':
            case 'hour':
            case 'hrs':
            case 'hr':
            case 'h':
                return n * h;
            case 'minutes':
            case 'minute':
            case 'mins':
            case 'min':
            case 'm':
                return n * m;
            case 'seconds':
            case 'second':
            case 'secs':
            case 'sec':
            case 's':
                return n * s;
            case 'milliseconds':
            case 'millisecond':
            case 'msecs':
            case 'msec':
            case 'ms':
                return n;
            default:
                return undefined;
        }
    }
    /**
     * Short format for `ms`.
     *
     * @param {Number} ms
     * @return {String}
     * @api private
     */
    function fmtShort(ms) {
        var msAbs = Math.abs(ms);
        if (msAbs >= d) {
            return Math.round(ms / d) + 'd';
        }
        if (msAbs >= h) {
            return Math.round(ms / h) + 'h';
        }
        if (msAbs >= m) {
            return Math.round(ms / m) + 'm';
        }
        if (msAbs >= s) {
            return Math.round(ms / s) + 's';
        }
        return ms + 'ms';
    }
    /**
     * Long format for `ms`.
     *
     * @param {Number} ms
     * @return {String}
     * @api private
     */
    function fmtLong(ms) {
        var msAbs = Math.abs(ms);
        if (msAbs >= d) {
            return plural(ms, msAbs, d, 'day');
        }
        if (msAbs >= h) {
            return plural(ms, msAbs, h, 'hour');
        }
        if (msAbs >= m) {
            return plural(ms, msAbs, m, 'minute');
        }
        if (msAbs >= s) {
            return plural(ms, msAbs, s, 'second');
        }
        return ms + ' ms';
    }
    /**
     * Pluralization helper.
     */
    function plural(ms, msAbs, n, name) {
        var isPlural = msAbs >= n * 1.5;
        return Math.round(ms / n) + ' ' + name + (isPlural ? 's' : '');
    }
},
/* debug/src/node.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Module dependencies.
     */
    const tty = require('tty');
    const util = require('util');
    /**
     * This is the Node.js implementation of `debug()`.
     */
    exports.init = init;
    exports.log = log;
    exports.formatArgs = formatArgs;
    exports.save = save;
    exports.load = load;
    exports.useColors = useColors;
    exports.destroy = util.deprecate(() => { }, 'Instance method `debug.destroy()` is deprecated and no longer does anything. It will be removed in the next major version of `debug`.');
    /**
     * Colors.
     */
    exports.colors = [6, 2, 3, 4, 5, 1];
    try {
        // Optional dependency (as in, doesn't need to be installed, NOT like optionalDependencies in package.json)
        // eslint-disable-next-line import/no-extraneous-dependencies
        const supportsColor = require(38) /* supports-color */;
        if (supportsColor && (supportsColor.stderr || supportsColor).level >= 2) {
            exports.colors = [
                20,
                21,
                26,
                27,
                32,
                33,
                38,
                39,
                40,
                41,
                42,
                43,
                44,
                45,
                56,
                57,
                62,
                63,
                68,
                69,
                74,
                75,
                76,
                77,
                78,
                79,
                80,
                81,
                92,
                93,
                98,
                99,
                112,
                113,
                128,
                129,
                134,
                135,
                148,
                149,
                160,
                161,
                162,
                163,
                164,
                165,
                166,
                167,
                168,
                169,
                170,
                171,
                172,
                173,
                178,
                179,
                184,
                185,
                196,
                197,
                198,
                199,
                200,
                201,
                202,
                203,
                204,
                205,
                206,
                207,
                208,
                209,
                214,
                215,
                220,
                221
            ];
        }
    }
    catch (error) {
        // Swallow - we only care if `supports-color` is available; it doesn't have to be.
    }
    /**
     * Build up the default `inspectOpts` object from the environment variables.
     *
     *   $ DEBUG_COLORS=no DEBUG_DEPTH=10 DEBUG_SHOW_HIDDEN=enabled node script.js
     */
    exports.inspectOpts = Object.keys(process.env).filter(key => {
        return /^debug_/i.test(key);
    }).reduce((obj, key) => {
        // Camel-case
        const prop = key
            .substring(6)
            .toLowerCase()
            .replace(/_([a-z])/g, (_, k) => {
            return k.toUpperCase();
        });
        // Coerce string value into JS value
        let val = process.env[key];
        if (/^(yes|on|true|enabled)$/i.test(val)) {
            val = true;
        }
        else if (/^(no|off|false|disabled)$/i.test(val)) {
            val = false;
        }
        else if (val === 'null') {
            val = null;
        }
        else {
            val = Number(val);
        }
        obj[prop] = val;
        return obj;
    }, {});
    /**
     * Is stdout a TTY? Colored output is enabled when `true`.
     */
    function useColors() {
        return 'colors' in exports.inspectOpts ?
            Boolean(exports.inspectOpts.colors) :
            tty.isatty(process.stderr.fd);
    }
    /**
     * Adds ANSI color escape codes if enabled.
     *
     * @api public
     */
    function formatArgs(args) {
        const { namespace: name, useColors } = this;
        if (useColors) {
            const c = this.color;
            const colorCode = '\u001B[3' + (c < 8 ? c : '8;5;' + c);
            const prefix = `  ${colorCode};1m${name} \u001B[0m`;
            args[0] = prefix + args[0].split('\n').join('\n' + prefix);
            args.push(colorCode + 'm+' + module.exports.humanize(this.diff) + '\u001B[0m');
        }
        else {
            args[0] = getDate() + name + ' ' + args[0];
        }
    }
    function getDate() {
        if (exports.inspectOpts.hideDate) {
            return '';
        }
        return new Date().toISOString() + ' ';
    }
    /**
     * Invokes `util.format()` with the specified arguments and writes to stderr.
     */
    function log(...args) {
        return process.stderr.write(util.format(...args) + '\n');
    }
    /**
     * Save `namespaces`.
     *
     * @param {String} namespaces
     * @api private
     */
    function save(namespaces) {
        if (namespaces) {
            process.env.DEBUG = namespaces;
        }
        else {
            // If you set a process.env field to null or undefined, it gets cast to the
            // string 'null' or 'undefined'. Just delete instead.
            delete process.env.DEBUG;
        }
    }
    /**
     * Load `namespaces`.
     *
     * @return {String} returns the previously persisted debug modes
     * @api private
     */
    function load() {
        return process.env.DEBUG;
    }
    /**
     * Init logic for `debug` instances.
     *
     * Create a new `inspectOpts` object in case `useColors` is set
     * differently for a particular `debug` instance.
     */
    function init(debug) {
        debug.inspectOpts = {};
        const keys = Object.keys(exports.inspectOpts);
        for (let i = 0; i < keys.length; i++) {
            debug.inspectOpts[keys[i]] = exports.inspectOpts[keys[i]];
        }
    }
    module.exports = require(74) /* ./common */(exports);
    const { formatters } = module.exports;
    /**
     * Map %o to `util.inspect()`, all on a single line.
     */
    formatters.o = function (v) {
        this.inspectOpts.colors = this.useColors;
        return util.inspect(v, this.inspectOpts)
            .split('\n')
            .map(str => str.trim())
            .join(' ');
    };
    /**
     * Map %O to `util.inspect()`, allowing multiple lines if needed.
     */
    formatters.O = function (v) {
        this.inspectOpts.colors = this.useColors;
        return util.inspect(v, this.inspectOpts);
    };
},
/* needle/lib/querystring.js */ function _(require, module, exports, __esModule, __esExport) {
    // based on the qs module, but handles null objects as expected
    // fixes by Tomas Pollak.
    var toString = Object.prototype.toString;
    function stringify(obj, prefix) {
        if (prefix && (obj === null || typeof obj == 'undefined')) {
            return prefix + '=';
        }
        else if (toString.call(obj) == '[object Array]') {
            return stringifyArray(obj, prefix);
        }
        else if (toString.call(obj) == '[object Object]') {
            return stringifyObject(obj, prefix);
        }
        else if (toString.call(obj) == '[object Date]') {
            return obj.toISOString();
        }
        else if (prefix) { // string inside array or hash
            return prefix + '=' + encodeURIComponent(String(obj));
        }
        else if (String(obj).indexOf('=') !== -1) { // string with equal sign
            return String(obj);
        }
        else {
            throw new TypeError('Cannot build a querystring out of: ' + obj);
        }
    }
    ;
    function stringifyArray(arr, prefix) {
        var ret = [];
        for (var i = 0, len = arr.length; i < len; i++) {
            if (prefix)
                ret.push(stringify(arr[i], prefix + '[]'));
            else
                ret.push(stringify(arr[i]));
        }
        return ret.join('&');
    }
    function stringifyObject(obj, prefix) {
        var ret = [];
        Object.keys(obj).forEach(function (key) {
            ret.push(stringify(obj[key], prefix
                ? prefix + '[' + encodeURIComponent(key) + ']'
                : encodeURIComponent(key)));
        });
        return ret.join('&');
    }
    exports.build = stringify;
},
/* needle/lib/multipart.js */ function _(require, module, exports, __esModule, __esExport) {
    var readFile = require('fs').readFile, basename = require('path').basename;
    exports.build = function (data, boundary, callback) {
        if (typeof data != 'object' || typeof data.pipe == 'function')
            return callback(new Error('Multipart builder expects data as key/val object.'));
        var body = '', object = flatten(data), count = Object.keys(object).length;
        if (count === 0)
            return callback(new Error('Empty multipart body. Invalid data.'));
        function done(err, section) {
            if (err)
                return callback(err);
            if (section)
                body += section;
            --count || callback(null, body + '--' + boundary + '--');
        }
        ;
        for (var key in object) {
            var value = object[key];
            if (value === null || typeof value == 'undefined') {
                done();
            }
            else if (Buffer.isBuffer(value)) {
                var part = { buffer: value, content_type: 'application/octet-stream' };
                generate_part(key, part, boundary, done);
            }
            else {
                var part = (value.buffer || value.file || value.content_type) ? value : { value: value };
                generate_part(key, part, boundary, done);
            }
        }
    };
    function generate_part(name, part, boundary, callback) {
        var return_part = '--' + boundary + '\r\n';
        return_part += 'Content-Disposition: form-data; name="' + name + '"';
        function append(data, filename) {
            if (data) {
                var binary = part.content_type.indexOf('text') == -1;
                return_part += '; filename="' + encodeURIComponent(filename) + '"\r\n';
                if (binary)
                    return_part += 'Content-Transfer-Encoding: binary\r\n';
                return_part += 'Content-Type: ' + part.content_type + '\r\n\r\n';
                return_part += binary ? data.toString('binary') : data.toString('utf8');
            }
            callback(null, return_part + '\r\n');
        }
        ;
        if ((part.file || part.buffer) && part.content_type) {
            var filename = part.filename ? part.filename : part.file ? basename(part.file) : name;
            if (part.buffer)
                return append(part.buffer, filename);
            readFile(part.file, function (err, data) {
                if (err)
                    return callback(err);
                append(data, filename);
            });
        }
        else {
            if (typeof part.value == 'object')
                return callback(new Error('Object received for ' + name + ', expected string.'));
            if (part.content_type) {
                return_part += '\r\n';
                return_part += 'Content-Type: ' + part.content_type;
            }
            return_part += '\r\n\r\n';
            return_part += Buffer.from(String(part.value), 'utf8').toString('binary');
            append();
        }
    }
    // flattens nested objects for multipart body
    function flatten(object, into, prefix) {
        into = into || {};
        for (var key in object) {
            var prefix_key = prefix ? prefix + '[' + key + ']' : key;
            var prop = object[key];
            if (prop && typeof prop === 'object' && !(prop.buffer || prop.file || prop.content_type))
                flatten(prop, into, prefix_key);
            else
                into[prefix_key] = prop;
        }
        return into;
    }
},
/* needle/lib/auth.js */ function _(require, module, exports, __esModule, __esExport) {
    var createHash = require('crypto').createHash;
    function get_header(header, credentials, opts) {
        var type = header.split(' ')[0], user = credentials[0], pass = credentials[1];
        if (type == 'Digest') {
            return digest.generate(header, user, pass, opts.method, opts.path);
        }
        else if (type == 'Basic') {
            return basic(user, pass);
        }
    }
    ////////////////////
    // basic
    function md5(string) {
        return createHash('md5').update(string).digest('hex');
    }
    function basic(user, pass) {
        var str = typeof pass == 'undefined' ? user : [user, pass].join(':');
        return 'Basic ' + Buffer.from(str).toString('base64');
    }
    ////////////////////
    // digest
    // logic inspired from https://github.com/simme/node-http-digest-client
    var digest = {};
    digest.parse_header = function (header) {
        var challenge = {}, matches = header.match(/([a-z0-9_-]+)="?([a-z0-9=\/\.@\s-\+)()]+)"?/gi);
        for (var i = 0, l = matches.length; i < l; i++) {
            var parts = matches[i].split('='), key = parts.shift(), val = parts.join('=').replace(/^"/, '').replace(/"$/, '');
            challenge[key] = val;
        }
        return challenge;
    };
    digest.update_nc = function (nc) {
        var max = 99999999;
        nc++;
        if (nc > max)
            nc = 1;
        var padding = new Array(8).join('0') + '';
        nc = nc + '';
        return padding.substr(0, 8 - nc.length) + nc;
    };
    digest.generate = function (header, user, pass, method, path) {
        var nc = 1, cnonce = null, challenge = digest.parse_header(header);
        var ha1 = md5(user + ':' + challenge.realm + ':' + pass), ha2 = md5(method.toUpperCase() + ':' + path), resp = [ha1, challenge.nonce];
        if (typeof challenge.qop === 'string') {
            cnonce = md5(Math.random().toString(36)).substr(0, 8);
            nc = digest.update_nc(nc);
            resp = resp.concat(nc, cnonce);
            resp = resp.concat(challenge.qop, ha2);
        }
        else {
            resp = resp.concat(ha2);
        }
        var params = {
            uri: path,
            realm: challenge.realm,
            nonce: challenge.nonce,
            username: user,
            response: md5(resp.join(':'))
        };
        if (challenge.qop) {
            params.qop = challenge.qop;
        }
        if (challenge.opaque) {
            params.opaque = challenge.opaque;
        }
        if (cnonce) {
            params.nc = nc;
            params.cnonce = cnonce;
        }
        header = [];
        for (var k in params)
            header.push(k + '="' + params[k] + '"');
        return 'Digest ' + header.join(', ');
    };
    module.exports = {
        header: get_header,
        basic: basic,
        digest: digest.generate
    };
},
/* needle/lib/cookies.js */ function _(require, module, exports, __esModule, __esExport) {
    //  Simple cookie handling implementation based on the standard RFC 6265.
    //
    //  This module just has two functionalities:
    //    - Parse a set-cookie-header as a key value object
    //    - Write a cookie-string from a key value object
    //
    //  All cookie attributes are ignored.
    var unescape = require('querystring').unescape;
    var COOKIE_PAIR = /^([^=\s]+)\s*=\s*("?)\s*(.*)\s*\2\s*$/;
    var EXCLUDED_CHARS = /[\x00-\x1F\x7F\x3B\x3B\s\"\,\\"%]/g;
    var TRAILING_SEMICOLON = /\x3B+$/;
    var SEP_SEMICOLON = /\s*\x3B\s*/;
    // i know these should be 'const', but I'd like to keep
    // supporting earlier node.js versions as long as I can. :)
    var KEY_INDEX = 1; // index of key from COOKIE_PAIR match
    var VALUE_INDEX = 3; // index of value from COOKIE_PAIR match
    // Returns a copy str trimmed and without trainling semicolon.
    function cleanCookieString(str) {
        return str.trim().replace(/\x3B+$/, '');
    }
    function getFirstPair(str) {
        var index = str.indexOf('\x3B');
        return index === -1 ? str : str.substr(0, index);
    }
    // Returns a encoded copy of str based on RFC6265 S4.1.1.
    function encodeCookieComponent(str) {
        return str.toString().replace(EXCLUDED_CHARS, encodeURIComponent);
    }
    // Parses a set-cookie-string based on the standard defined in RFC6265 S4.1.1.
    function parseSetCookieString(str) {
        str = cleanCookieString(str);
        str = getFirstPair(str);
        var res = COOKIE_PAIR.exec(str);
        if (!res || !res[VALUE_INDEX])
            return null;
        return {
            name: unescape(res[KEY_INDEX]),
            value: unescape(res[VALUE_INDEX])
        };
    }
    // Parses a set-cookie-header and returns a key/value object.
    // Each key represents the name of a cookie.
    function parseSetCookieHeader(header) {
        if (!header)
            return {};
        header = Array.isArray(header) ? header : [header];
        return header.reduce(function (res, str) {
            var cookie = parseSetCookieString(str);
            if (cookie)
                res[cookie.name] = cookie.value;
            return res;
        }, {});
    }
    // Writes a set-cookie-string based on the standard definded in RFC6265 S4.1.1.
    function writeCookieString(obj) {
        return Object.keys(obj).reduce(function (str, name) {
            var encodedName = encodeCookieComponent(name);
            var encodedValue = encodeCookieComponent(obj[name]);
            str += (str ? '; ' : '') + encodedName + '=' + encodedValue;
            return str;
        }, '');
    }
    // returns a key/val object from an array of cookie strings
    exports.read = parseSetCookieHeader;
    // writes a cookie string header
    exports.write = writeCookieString;
},
/* needle/lib/parsers.js */ function _(require, module, exports, __esModule, __esExport) {
    //////////////////////////////////////////
    // Defines mappings between content-type
    // and the appropriate parsers.
    //////////////////////////////////////////
    var Transform = require('stream').Transform;
    var sax = require(82) /* sax */;
    function parseXML(str, cb) {
        var obj, current, parser = sax.parser(true, { trim: true, lowercase: true });
        parser.onerror = parser.onend = done;
        function done(err) {
            parser.onerror = parser.onend = function () { };
            cb(err, obj);
        }
        function newElement(name, attributes) {
            return {
                name: name || '',
                value: '',
                attributes: attributes || {},
                children: []
            };
        }
        parser.oncdata = parser.ontext = function (t) {
            if (current)
                current.value += t;
        };
        parser.onopentag = function (node) {
            var element = newElement(node.name, node.attributes);
            if (current) {
                element.parent = current;
                current.children.push(element);
            }
            else { // root object
                obj = element;
            }
            current = element;
        };
        parser.onclosetag = function () {
            if (typeof current.parent !== 'undefined') {
                var just_closed = current;
                current = current.parent;
                delete just_closed.parent;
            }
        };
        parser.write(str).close();
    }
    function parserFactory(name, fn) {
        function parser() {
            var chunks = [], stream = new Transform({ objectMode: true });
            // Buffer all our data
            stream._transform = function (chunk, encoding, done) {
                chunks.push(chunk);
                done();
            };
            // And call the parser when all is there.
            stream._flush = function (done) {
                var self = this, data = Buffer.concat(chunks);
                try {
                    fn(data, function (err, result) {
                        if (err)
                            throw err;
                        self.push(result);
                    });
                }
                catch (err) {
                    self.push(data); // just pass the original data
                }
                finally {
                    done();
                }
            };
            return stream;
        }
        return { fn: parser, name: name };
    }
    var parsers = {};
    function buildParser(name, types, fn) {
        var parser = parserFactory(name, fn);
        types.forEach(function (type) {
            parsers[type] = parser;
        });
    }
    buildParser('json', [
        'application/json',
        'text/javascript'
    ], function (buffer, cb) {
        var err, data;
        try {
            data = JSON.parse(buffer);
        }
        catch (e) {
            err = e;
        }
        cb(err, data);
    });
    buildParser('xml', [
        'text/xml',
        'application/xml',
        'application/rdf+xml',
        'application/rss+xml',
        'application/atom+xml'
    ], function (buffer, cb) {
        parseXML(buffer.toString(), function (err, obj) {
            cb(err, obj);
        });
    });
    module.exports = parsers;
    module.exports.use = buildParser;
},
/* sax/lib/sax.js */ function _(require, module, exports, __esModule, __esExport) {
    ;
    (function (sax) {
        sax.parser = function (strict, opt) { return new SAXParser(strict, opt); };
        sax.SAXParser = SAXParser;
        sax.SAXStream = SAXStream;
        sax.createStream = createStream;
        // When we pass the MAX_BUFFER_LENGTH position, start checking for buffer overruns.
        // When we check, schedule the next check for MAX_BUFFER_LENGTH - (max(buffer lengths)),
        // since that's the earliest that a buffer overrun could occur.  This way, checks are
        // as rare as required, but as often as necessary to ensure never crossing this bound.
        // Furthermore, buffers are only tested at most once per write(), so passing a very
        // large string into write() might have undesirable effects, but this is manageable by
        // the caller, so it is assumed to be safe.  Thus, a call to write() may, in the extreme
        // edge case, result in creating at most one complete copy of the string passed in.
        // Set to Infinity to have unlimited buffers.
        sax.MAX_BUFFER_LENGTH = 64 * 1024;
        var buffers = [
            'comment', 'sgmlDecl', 'textNode', 'tagName', 'doctype',
            'procInstName', 'procInstBody', 'entity', 'attribName',
            'attribValue', 'cdata', 'script'
        ];
        sax.EVENTS = [
            'text',
            'processinginstruction',
            'sgmldeclaration',
            'doctype',
            'comment',
            'opentagstart',
            'attribute',
            'opentag',
            'closetag',
            'opencdata',
            'cdata',
            'closecdata',
            'error',
            'end',
            'ready',
            'script',
            'opennamespace',
            'closenamespace'
        ];
        function SAXParser(strict, opt) {
            if (!(this instanceof SAXParser)) {
                return new SAXParser(strict, opt);
            }
            var parser = this;
            clearBuffers(parser);
            parser.q = parser.c = '';
            parser.bufferCheckPosition = sax.MAX_BUFFER_LENGTH;
            parser.opt = opt || {};
            parser.opt.lowercase = parser.opt.lowercase || parser.opt.lowercasetags;
            parser.looseCase = parser.opt.lowercase ? 'toLowerCase' : 'toUpperCase';
            parser.tags = [];
            parser.closed = parser.closedRoot = parser.sawRoot = false;
            parser.tag = parser.error = null;
            parser.strict = !!strict;
            parser.noscript = !!(strict || parser.opt.noscript);
            parser.state = S.BEGIN;
            parser.strictEntities = parser.opt.strictEntities;
            parser.ENTITIES = parser.strictEntities ? Object.create(sax.XML_ENTITIES) : Object.create(sax.ENTITIES);
            parser.attribList = [];
            // namespaces form a prototype chain.
            // it always points at the current tag,
            // which protos to its parent tag.
            if (parser.opt.xmlns) {
                parser.ns = Object.create(rootNS);
            }
            // mostly just for error reporting
            parser.trackPosition = parser.opt.position !== false;
            if (parser.trackPosition) {
                parser.position = parser.line = parser.column = 0;
            }
            emit(parser, 'onready');
        }
        if (!Object.create) {
            Object.create = function (o) {
                function F() { }
                F.prototype = o;
                var newf = new F();
                return newf;
            };
        }
        if (!Object.keys) {
            Object.keys = function (o) {
                var a = [];
                for (var i in o)
                    if (o.hasOwnProperty(i))
                        a.push(i);
                return a;
            };
        }
        function checkBufferLength(parser) {
            var maxAllowed = Math.max(sax.MAX_BUFFER_LENGTH, 10);
            var maxActual = 0;
            for (var i = 0, l = buffers.length; i < l; i++) {
                var len = parser[buffers[i]].length;
                if (len > maxAllowed) {
                    // Text/cdata nodes can get big, and since they're buffered,
                    // we can get here under normal conditions.
                    // Avoid issues by emitting the text node now,
                    // so at least it won't get any bigger.
                    switch (buffers[i]) {
                        case 'textNode':
                            closeText(parser);
                            break;
                        case 'cdata':
                            emitNode(parser, 'oncdata', parser.cdata);
                            parser.cdata = '';
                            break;
                        case 'script':
                            emitNode(parser, 'onscript', parser.script);
                            parser.script = '';
                            break;
                        default:
                            error(parser, 'Max buffer length exceeded: ' + buffers[i]);
                    }
                }
                maxActual = Math.max(maxActual, len);
            }
            // schedule the next check for the earliest possible buffer overrun.
            var m = sax.MAX_BUFFER_LENGTH - maxActual;
            parser.bufferCheckPosition = m + parser.position;
        }
        function clearBuffers(parser) {
            for (var i = 0, l = buffers.length; i < l; i++) {
                parser[buffers[i]] = '';
            }
        }
        function flushBuffers(parser) {
            closeText(parser);
            if (parser.cdata !== '') {
                emitNode(parser, 'oncdata', parser.cdata);
                parser.cdata = '';
            }
            if (parser.script !== '') {
                emitNode(parser, 'onscript', parser.script);
                parser.script = '';
            }
        }
        SAXParser.prototype = {
            end: function () { end(this); },
            write: write,
            resume: function () { this.error = null; return this; },
            close: function () { return this.write(null); },
            flush: function () { flushBuffers(this); }
        };
        var Stream;
        try {
            Stream = require('stream').Stream;
        }
        catch (ex) {
            Stream = function () { };
        }
        var streamWraps = sax.EVENTS.filter(function (ev) {
            return ev !== 'error' && ev !== 'end';
        });
        function createStream(strict, opt) {
            return new SAXStream(strict, opt);
        }
        function SAXStream(strict, opt) {
            if (!(this instanceof SAXStream)) {
                return new SAXStream(strict, opt);
            }
            Stream.apply(this);
            this._parser = new SAXParser(strict, opt);
            this.writable = true;
            this.readable = true;
            var me = this;
            this._parser.onend = function () {
                me.emit('end');
            };
            this._parser.onerror = function (er) {
                me.emit('error', er);
                // if didn't throw, then means error was handled.
                // go ahead and clear error, so we can write again.
                me._parser.error = null;
            };
            this._decoder = null;
            streamWraps.forEach(function (ev) {
                Object.defineProperty(me, 'on' + ev, {
                    get: function () {
                        return me._parser['on' + ev];
                    },
                    set: function (h) {
                        if (!h) {
                            me.removeAllListeners(ev);
                            me._parser['on' + ev] = h;
                            return h;
                        }
                        me.on(ev, h);
                    },
                    enumerable: true,
                    configurable: false
                });
            });
        }
        SAXStream.prototype = Object.create(Stream.prototype, {
            constructor: {
                value: SAXStream
            }
        });
        SAXStream.prototype.write = function (data) {
            if (typeof Buffer === 'function' &&
                typeof Buffer.isBuffer === 'function' &&
                Buffer.isBuffer(data)) {
                if (!this._decoder) {
                    var SD = require('string_decoder').StringDecoder;
                    this._decoder = new SD('utf8');
                }
                data = this._decoder.write(data);
            }
            this._parser.write(data.toString());
            this.emit('data', data);
            return true;
        };
        SAXStream.prototype.end = function (chunk) {
            if (chunk && chunk.length) {
                this.write(chunk);
            }
            this._parser.end();
            return true;
        };
        SAXStream.prototype.on = function (ev, handler) {
            var me = this;
            if (!me._parser['on' + ev] && streamWraps.indexOf(ev) !== -1) {
                me._parser['on' + ev] = function () {
                    var args = arguments.length === 1 ? [arguments[0]] : Array.apply(null, arguments);
                    args.splice(0, 0, ev);
                    me.emit.apply(me, args);
                };
            }
            return Stream.prototype.on.call(me, ev, handler);
        };
        // this really needs to be replaced with character classes.
        // XML allows all manner of ridiculous numbers and digits.
        var CDATA = '[CDATA[';
        var DOCTYPE = 'DOCTYPE';
        var XML_NAMESPACE = 'http://www.w3.org/XML/1998/namespace';
        var XMLNS_NAMESPACE = 'http://www.w3.org/2000/xmlns/';
        var rootNS = { xml: XML_NAMESPACE, xmlns: XMLNS_NAMESPACE };
        // http://www.w3.org/TR/REC-xml/#NT-NameStartChar
        // This implementation works on strings, a single character at a time
        // as such, it cannot ever support astral-plane characters (10000-EFFFF)
        // without a significant breaking change to either this  parser, or the
        // JavaScript language.  Implementation of an emoji-capable xml parser
        // is left as an exercise for the reader.
        var nameStart = /[:_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD]/;
        var nameBody = /[:_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\u00B7\u0300-\u036F\u203F-\u2040.\d-]/;
        var entityStart = /[#:_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD]/;
        var entityBody = /[#:_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\u00B7\u0300-\u036F\u203F-\u2040.\d-]/;
        function isWhitespace(c) {
            return c === ' ' || c === '\n' || c === '\r' || c === '\t';
        }
        function isQuote(c) {
            return c === '"' || c === '\'';
        }
        function isAttribEnd(c) {
            return c === '>' || isWhitespace(c);
        }
        function isMatch(regex, c) {
            return regex.test(c);
        }
        function notMatch(regex, c) {
            return !isMatch(regex, c);
        }
        var S = 0;
        sax.STATE = {
            BEGIN: S++,
            BEGIN_WHITESPACE: S++,
            TEXT: S++,
            TEXT_ENTITY: S++,
            OPEN_WAKA: S++,
            SGML_DECL: S++,
            SGML_DECL_QUOTED: S++,
            DOCTYPE: S++,
            DOCTYPE_QUOTED: S++,
            DOCTYPE_DTD: S++,
            DOCTYPE_DTD_QUOTED: S++,
            COMMENT_STARTING: S++,
            COMMENT: S++,
            COMMENT_ENDING: S++,
            COMMENT_ENDED: S++,
            CDATA: S++,
            CDATA_ENDING: S++,
            CDATA_ENDING_2: S++,
            PROC_INST: S++,
            PROC_INST_BODY: S++,
            PROC_INST_ENDING: S++,
            OPEN_TAG: S++,
            OPEN_TAG_SLASH: S++,
            ATTRIB: S++,
            ATTRIB_NAME: S++,
            ATTRIB_NAME_SAW_WHITE: S++,
            ATTRIB_VALUE: S++,
            ATTRIB_VALUE_QUOTED: S++,
            ATTRIB_VALUE_CLOSED: S++,
            ATTRIB_VALUE_UNQUOTED: S++,
            ATTRIB_VALUE_ENTITY_Q: S++,
            ATTRIB_VALUE_ENTITY_U: S++,
            CLOSE_TAG: S++,
            CLOSE_TAG_SAW_WHITE: S++,
            SCRIPT: S++,
            SCRIPT_ENDING: S++ // <script> ... <
        };
        sax.XML_ENTITIES = {
            'amp': '&',
            'gt': '>',
            'lt': '<',
            'quot': '"',
            'apos': "'"
        };
        sax.ENTITIES = {
            'amp': '&',
            'gt': '>',
            'lt': '<',
            'quot': '"',
            'apos': "'",
            'AElig': 198,
            'Aacute': 193,
            'Acirc': 194,
            'Agrave': 192,
            'Aring': 197,
            'Atilde': 195,
            'Auml': 196,
            'Ccedil': 199,
            'ETH': 208,
            'Eacute': 201,
            'Ecirc': 202,
            'Egrave': 200,
            'Euml': 203,
            'Iacute': 205,
            'Icirc': 206,
            'Igrave': 204,
            'Iuml': 207,
            'Ntilde': 209,
            'Oacute': 211,
            'Ocirc': 212,
            'Ograve': 210,
            'Oslash': 216,
            'Otilde': 213,
            'Ouml': 214,
            'THORN': 222,
            'Uacute': 218,
            'Ucirc': 219,
            'Ugrave': 217,
            'Uuml': 220,
            'Yacute': 221,
            'aacute': 225,
            'acirc': 226,
            'aelig': 230,
            'agrave': 224,
            'aring': 229,
            'atilde': 227,
            'auml': 228,
            'ccedil': 231,
            'eacute': 233,
            'ecirc': 234,
            'egrave': 232,
            'eth': 240,
            'euml': 235,
            'iacute': 237,
            'icirc': 238,
            'igrave': 236,
            'iuml': 239,
            'ntilde': 241,
            'oacute': 243,
            'ocirc': 244,
            'ograve': 242,
            'oslash': 248,
            'otilde': 245,
            'ouml': 246,
            'szlig': 223,
            'thorn': 254,
            'uacute': 250,
            'ucirc': 251,
            'ugrave': 249,
            'uuml': 252,
            'yacute': 253,
            'yuml': 255,
            'copy': 169,
            'reg': 174,
            'nbsp': 160,
            'iexcl': 161,
            'cent': 162,
            'pound': 163,
            'curren': 164,
            'yen': 165,
            'brvbar': 166,
            'sect': 167,
            'uml': 168,
            'ordf': 170,
            'laquo': 171,
            'not': 172,
            'shy': 173,
            'macr': 175,
            'deg': 176,
            'plusmn': 177,
            'sup1': 185,
            'sup2': 178,
            'sup3': 179,
            'acute': 180,
            'micro': 181,
            'para': 182,
            'middot': 183,
            'cedil': 184,
            'ordm': 186,
            'raquo': 187,
            'frac14': 188,
            'frac12': 189,
            'frac34': 190,
            'iquest': 191,
            'times': 215,
            'divide': 247,
            'OElig': 338,
            'oelig': 339,
            'Scaron': 352,
            'scaron': 353,
            'Yuml': 376,
            'fnof': 402,
            'circ': 710,
            'tilde': 732,
            'Alpha': 913,
            'Beta': 914,
            'Gamma': 915,
            'Delta': 916,
            'Epsilon': 917,
            'Zeta': 918,
            'Eta': 919,
            'Theta': 920,
            'Iota': 921,
            'Kappa': 922,
            'Lambda': 923,
            'Mu': 924,
            'Nu': 925,
            'Xi': 926,
            'Omicron': 927,
            'Pi': 928,
            'Rho': 929,
            'Sigma': 931,
            'Tau': 932,
            'Upsilon': 933,
            'Phi': 934,
            'Chi': 935,
            'Psi': 936,
            'Omega': 937,
            'alpha': 945,
            'beta': 946,
            'gamma': 947,
            'delta': 948,
            'epsilon': 949,
            'zeta': 950,
            'eta': 951,
            'theta': 952,
            'iota': 953,
            'kappa': 954,
            'lambda': 955,
            'mu': 956,
            'nu': 957,
            'xi': 958,
            'omicron': 959,
            'pi': 960,
            'rho': 961,
            'sigmaf': 962,
            'sigma': 963,
            'tau': 964,
            'upsilon': 965,
            'phi': 966,
            'chi': 967,
            'psi': 968,
            'omega': 969,
            'thetasym': 977,
            'upsih': 978,
            'piv': 982,
            'ensp': 8194,
            'emsp': 8195,
            'thinsp': 8201,
            'zwnj': 8204,
            'zwj': 8205,
            'lrm': 8206,
            'rlm': 8207,
            'ndash': 8211,
            'mdash': 8212,
            'lsquo': 8216,
            'rsquo': 8217,
            'sbquo': 8218,
            'ldquo': 8220,
            'rdquo': 8221,
            'bdquo': 8222,
            'dagger': 8224,
            'Dagger': 8225,
            'bull': 8226,
            'hellip': 8230,
            'permil': 8240,
            'prime': 8242,
            'Prime': 8243,
            'lsaquo': 8249,
            'rsaquo': 8250,
            'oline': 8254,
            'frasl': 8260,
            'euro': 8364,
            'image': 8465,
            'weierp': 8472,
            'real': 8476,
            'trade': 8482,
            'alefsym': 8501,
            'larr': 8592,
            'uarr': 8593,
            'rarr': 8594,
            'darr': 8595,
            'harr': 8596,
            'crarr': 8629,
            'lArr': 8656,
            'uArr': 8657,
            'rArr': 8658,
            'dArr': 8659,
            'hArr': 8660,
            'forall': 8704,
            'part': 8706,
            'exist': 8707,
            'empty': 8709,
            'nabla': 8711,
            'isin': 8712,
            'notin': 8713,
            'ni': 8715,
            'prod': 8719,
            'sum': 8721,
            'minus': 8722,
            'lowast': 8727,
            'radic': 8730,
            'prop': 8733,
            'infin': 8734,
            'ang': 8736,
            'and': 8743,
            'or': 8744,
            'cap': 8745,
            'cup': 8746,
            'int': 8747,
            'there4': 8756,
            'sim': 8764,
            'cong': 8773,
            'asymp': 8776,
            'ne': 8800,
            'equiv': 8801,
            'le': 8804,
            'ge': 8805,
            'sub': 8834,
            'sup': 8835,
            'nsub': 8836,
            'sube': 8838,
            'supe': 8839,
            'oplus': 8853,
            'otimes': 8855,
            'perp': 8869,
            'sdot': 8901,
            'lceil': 8968,
            'rceil': 8969,
            'lfloor': 8970,
            'rfloor': 8971,
            'lang': 9001,
            'rang': 9002,
            'loz': 9674,
            'spades': 9824,
            'clubs': 9827,
            'hearts': 9829,
            'diams': 9830
        };
        Object.keys(sax.ENTITIES).forEach(function (key) {
            var e = sax.ENTITIES[key];
            var s = typeof e === 'number' ? String.fromCharCode(e) : e;
            sax.ENTITIES[key] = s;
        });
        for (var s in sax.STATE) {
            sax.STATE[sax.STATE[s]] = s;
        }
        // shorthand
        S = sax.STATE;
        function emit(parser, event, data) {
            parser[event] && parser[event](data);
        }
        function emitNode(parser, nodeType, data) {
            if (parser.textNode)
                closeText(parser);
            emit(parser, nodeType, data);
        }
        function closeText(parser) {
            parser.textNode = textopts(parser.opt, parser.textNode);
            if (parser.textNode)
                emit(parser, 'ontext', parser.textNode);
            parser.textNode = '';
        }
        function textopts(opt, text) {
            if (opt.trim)
                text = text.trim();
            if (opt.normalize)
                text = text.replace(/\s+/g, ' ');
            return text;
        }
        function error(parser, er) {
            closeText(parser);
            if (parser.trackPosition) {
                er += '\nLine: ' + parser.line +
                    '\nColumn: ' + parser.column +
                    '\nChar: ' + parser.c;
            }
            er = new Error(er);
            parser.error = er;
            emit(parser, 'onerror', er);
            return parser;
        }
        function end(parser) {
            if (parser.sawRoot && !parser.closedRoot)
                strictFail(parser, 'Unclosed root tag');
            if ((parser.state !== S.BEGIN) &&
                (parser.state !== S.BEGIN_WHITESPACE) &&
                (parser.state !== S.TEXT)) {
                error(parser, 'Unexpected end');
            }
            closeText(parser);
            parser.c = '';
            parser.closed = true;
            emit(parser, 'onend');
            SAXParser.call(parser, parser.strict, parser.opt);
            return parser;
        }
        function strictFail(parser, message) {
            if (typeof parser !== 'object' || !(parser instanceof SAXParser)) {
                throw new Error('bad call to strictFail');
            }
            if (parser.strict) {
                error(parser, message);
            }
        }
        function newTag(parser) {
            if (!parser.strict)
                parser.tagName = parser.tagName[parser.looseCase]();
            var parent = parser.tags[parser.tags.length - 1] || parser;
            var tag = parser.tag = { name: parser.tagName, attributes: {} };
            // will be overridden if tag contails an xmlns="foo" or xmlns:foo="bar"
            if (parser.opt.xmlns) {
                tag.ns = parent.ns;
            }
            parser.attribList.length = 0;
            emitNode(parser, 'onopentagstart', tag);
        }
        function qname(name, attribute) {
            var i = name.indexOf(':');
            var qualName = i < 0 ? ['', name] : name.split(':');
            var prefix = qualName[0];
            var local = qualName[1];
            // <x "xmlns"="http://foo">
            if (attribute && name === 'xmlns') {
                prefix = 'xmlns';
                local = '';
            }
            return { prefix: prefix, local: local };
        }
        function attrib(parser) {
            if (!parser.strict) {
                parser.attribName = parser.attribName[parser.looseCase]();
            }
            if (parser.attribList.indexOf(parser.attribName) !== -1 ||
                parser.tag.attributes.hasOwnProperty(parser.attribName)) {
                parser.attribName = parser.attribValue = '';
                return;
            }
            if (parser.opt.xmlns) {
                var qn = qname(parser.attribName, true);
                var prefix = qn.prefix;
                var local = qn.local;
                if (prefix === 'xmlns') {
                    // namespace binding attribute. push the binding into scope
                    if (local === 'xml' && parser.attribValue !== XML_NAMESPACE) {
                        strictFail(parser, 'xml: prefix must be bound to ' + XML_NAMESPACE + '\n' +
                            'Actual: ' + parser.attribValue);
                    }
                    else if (local === 'xmlns' && parser.attribValue !== XMLNS_NAMESPACE) {
                        strictFail(parser, 'xmlns: prefix must be bound to ' + XMLNS_NAMESPACE + '\n' +
                            'Actual: ' + parser.attribValue);
                    }
                    else {
                        var tag = parser.tag;
                        var parent = parser.tags[parser.tags.length - 1] || parser;
                        if (tag.ns === parent.ns) {
                            tag.ns = Object.create(parent.ns);
                        }
                        tag.ns[local] = parser.attribValue;
                    }
                }
                // defer onattribute events until all attributes have been seen
                // so any new bindings can take effect. preserve attribute order
                // so deferred events can be emitted in document order
                parser.attribList.push([parser.attribName, parser.attribValue]);
            }
            else {
                // in non-xmlns mode, we can emit the event right away
                parser.tag.attributes[parser.attribName] = parser.attribValue;
                emitNode(parser, 'onattribute', {
                    name: parser.attribName,
                    value: parser.attribValue
                });
            }
            parser.attribName = parser.attribValue = '';
        }
        function openTag(parser, selfClosing) {
            if (parser.opt.xmlns) {
                // emit namespace binding events
                var tag = parser.tag;
                // add namespace info to tag
                var qn = qname(parser.tagName);
                tag.prefix = qn.prefix;
                tag.local = qn.local;
                tag.uri = tag.ns[qn.prefix] || '';
                if (tag.prefix && !tag.uri) {
                    strictFail(parser, 'Unbound namespace prefix: ' +
                        JSON.stringify(parser.tagName));
                    tag.uri = qn.prefix;
                }
                var parent = parser.tags[parser.tags.length - 1] || parser;
                if (tag.ns && parent.ns !== tag.ns) {
                    Object.keys(tag.ns).forEach(function (p) {
                        emitNode(parser, 'onopennamespace', {
                            prefix: p,
                            uri: tag.ns[p]
                        });
                    });
                }
                // handle deferred onattribute events
                // Note: do not apply default ns to attributes:
                //   http://www.w3.org/TR/REC-xml-names/#defaulting
                for (var i = 0, l = parser.attribList.length; i < l; i++) {
                    var nv = parser.attribList[i];
                    var name = nv[0];
                    var value = nv[1];
                    var qualName = qname(name, true);
                    var prefix = qualName.prefix;
                    var local = qualName.local;
                    var uri = prefix === '' ? '' : (tag.ns[prefix] || '');
                    var a = {
                        name: name,
                        value: value,
                        prefix: prefix,
                        local: local,
                        uri: uri
                    };
                    // if there's any attributes with an undefined namespace,
                    // then fail on them now.
                    if (prefix && prefix !== 'xmlns' && !uri) {
                        strictFail(parser, 'Unbound namespace prefix: ' +
                            JSON.stringify(prefix));
                        a.uri = prefix;
                    }
                    parser.tag.attributes[name] = a;
                    emitNode(parser, 'onattribute', a);
                }
                parser.attribList.length = 0;
            }
            parser.tag.isSelfClosing = !!selfClosing;
            // process the tag
            parser.sawRoot = true;
            parser.tags.push(parser.tag);
            emitNode(parser, 'onopentag', parser.tag);
            if (!selfClosing) {
                // special case for <script> in non-strict mode.
                if (!parser.noscript && parser.tagName.toLowerCase() === 'script') {
                    parser.state = S.SCRIPT;
                }
                else {
                    parser.state = S.TEXT;
                }
                parser.tag = null;
                parser.tagName = '';
            }
            parser.attribName = parser.attribValue = '';
            parser.attribList.length = 0;
        }
        function closeTag(parser) {
            if (!parser.tagName) {
                strictFail(parser, 'Weird empty close tag.');
                parser.textNode += '</>';
                parser.state = S.TEXT;
                return;
            }
            if (parser.script) {
                if (parser.tagName !== 'script') {
                    parser.script += '</' + parser.tagName + '>';
                    parser.tagName = '';
                    parser.state = S.SCRIPT;
                    return;
                }
                emitNode(parser, 'onscript', parser.script);
                parser.script = '';
            }
            // first make sure that the closing tag actually exists.
            // <a><b></c></b></a> will close everything, otherwise.
            var t = parser.tags.length;
            var tagName = parser.tagName;
            if (!parser.strict) {
                tagName = tagName[parser.looseCase]();
            }
            var closeTo = tagName;
            while (t--) {
                var close = parser.tags[t];
                if (close.name !== closeTo) {
                    // fail the first time in strict mode
                    strictFail(parser, 'Unexpected close tag');
                }
                else {
                    break;
                }
            }
            // didn't find it.  we already failed for strict, so just abort.
            if (t < 0) {
                strictFail(parser, 'Unmatched closing tag: ' + parser.tagName);
                parser.textNode += '</' + parser.tagName + '>';
                parser.state = S.TEXT;
                return;
            }
            parser.tagName = tagName;
            var s = parser.tags.length;
            while (s-- > t) {
                var tag = parser.tag = parser.tags.pop();
                parser.tagName = parser.tag.name;
                emitNode(parser, 'onclosetag', parser.tagName);
                var x = {};
                for (var i in tag.ns) {
                    x[i] = tag.ns[i];
                }
                var parent = parser.tags[parser.tags.length - 1] || parser;
                if (parser.opt.xmlns && tag.ns !== parent.ns) {
                    // remove namespace bindings introduced by tag
                    Object.keys(tag.ns).forEach(function (p) {
                        var n = tag.ns[p];
                        emitNode(parser, 'onclosenamespace', { prefix: p, uri: n });
                    });
                }
            }
            if (t === 0)
                parser.closedRoot = true;
            parser.tagName = parser.attribValue = parser.attribName = '';
            parser.attribList.length = 0;
            parser.state = S.TEXT;
        }
        function parseEntity(parser) {
            var entity = parser.entity;
            var entityLC = entity.toLowerCase();
            var num;
            var numStr = '';
            if (parser.ENTITIES[entity]) {
                return parser.ENTITIES[entity];
            }
            if (parser.ENTITIES[entityLC]) {
                return parser.ENTITIES[entityLC];
            }
            entity = entityLC;
            if (entity.charAt(0) === '#') {
                if (entity.charAt(1) === 'x') {
                    entity = entity.slice(2);
                    num = parseInt(entity, 16);
                    numStr = num.toString(16);
                }
                else {
                    entity = entity.slice(1);
                    num = parseInt(entity, 10);
                    numStr = num.toString(10);
                }
            }
            entity = entity.replace(/^0+/, '');
            if (isNaN(num) || numStr.toLowerCase() !== entity) {
                strictFail(parser, 'Invalid character entity');
                return '&' + parser.entity + ';';
            }
            return String.fromCodePoint(num);
        }
        function beginWhiteSpace(parser, c) {
            if (c === '<') {
                parser.state = S.OPEN_WAKA;
                parser.startTagPosition = parser.position;
            }
            else if (!isWhitespace(c)) {
                // have to process this as a text node.
                // weird, but happens.
                strictFail(parser, 'Non-whitespace before first tag.');
                parser.textNode = c;
                parser.state = S.TEXT;
            }
        }
        function charAt(chunk, i) {
            var result = '';
            if (i < chunk.length) {
                result = chunk.charAt(i);
            }
            return result;
        }
        function write(chunk) {
            var parser = this;
            if (this.error) {
                throw this.error;
            }
            if (parser.closed) {
                return error(parser, 'Cannot write after close. Assign an onready handler.');
            }
            if (chunk === null) {
                return end(parser);
            }
            if (typeof chunk === 'object') {
                chunk = chunk.toString();
            }
            var i = 0;
            var c = '';
            while (true) {
                c = charAt(chunk, i++);
                parser.c = c;
                if (!c) {
                    break;
                }
                if (parser.trackPosition) {
                    parser.position++;
                    if (c === '\n') {
                        parser.line++;
                        parser.column = 0;
                    }
                    else {
                        parser.column++;
                    }
                }
                switch (parser.state) {
                    case S.BEGIN:
                        parser.state = S.BEGIN_WHITESPACE;
                        if (c === '\uFEFF') {
                            continue;
                        }
                        beginWhiteSpace(parser, c);
                        continue;
                    case S.BEGIN_WHITESPACE:
                        beginWhiteSpace(parser, c);
                        continue;
                    case S.TEXT:
                        if (parser.sawRoot && !parser.closedRoot) {
                            var starti = i - 1;
                            while (c && c !== '<' && c !== '&') {
                                c = charAt(chunk, i++);
                                if (c && parser.trackPosition) {
                                    parser.position++;
                                    if (c === '\n') {
                                        parser.line++;
                                        parser.column = 0;
                                    }
                                    else {
                                        parser.column++;
                                    }
                                }
                            }
                            parser.textNode += chunk.substring(starti, i - 1);
                        }
                        if (c === '<' && !(parser.sawRoot && parser.closedRoot && !parser.strict)) {
                            parser.state = S.OPEN_WAKA;
                            parser.startTagPosition = parser.position;
                        }
                        else {
                            if (!isWhitespace(c) && (!parser.sawRoot || parser.closedRoot)) {
                                strictFail(parser, 'Text data outside of root node.');
                            }
                            if (c === '&') {
                                parser.state = S.TEXT_ENTITY;
                            }
                            else {
                                parser.textNode += c;
                            }
                        }
                        continue;
                    case S.SCRIPT:
                        // only non-strict
                        if (c === '<') {
                            parser.state = S.SCRIPT_ENDING;
                        }
                        else {
                            parser.script += c;
                        }
                        continue;
                    case S.SCRIPT_ENDING:
                        if (c === '/') {
                            parser.state = S.CLOSE_TAG;
                        }
                        else {
                            parser.script += '<' + c;
                            parser.state = S.SCRIPT;
                        }
                        continue;
                    case S.OPEN_WAKA:
                        // either a /, ?, !, or text is coming next.
                        if (c === '!') {
                            parser.state = S.SGML_DECL;
                            parser.sgmlDecl = '';
                        }
                        else if (isWhitespace(c)) {
                            // wait for it...
                        }
                        else if (isMatch(nameStart, c)) {
                            parser.state = S.OPEN_TAG;
                            parser.tagName = c;
                        }
                        else if (c === '/') {
                            parser.state = S.CLOSE_TAG;
                            parser.tagName = '';
                        }
                        else if (c === '?') {
                            parser.state = S.PROC_INST;
                            parser.procInstName = parser.procInstBody = '';
                        }
                        else {
                            strictFail(parser, 'Unencoded <');
                            // if there was some whitespace, then add that in.
                            if (parser.startTagPosition + 1 < parser.position) {
                                var pad = parser.position - parser.startTagPosition;
                                c = new Array(pad).join(' ') + c;
                            }
                            parser.textNode += '<' + c;
                            parser.state = S.TEXT;
                        }
                        continue;
                    case S.SGML_DECL:
                        if ((parser.sgmlDecl + c).toUpperCase() === CDATA) {
                            emitNode(parser, 'onopencdata');
                            parser.state = S.CDATA;
                            parser.sgmlDecl = '';
                            parser.cdata = '';
                        }
                        else if (parser.sgmlDecl + c === '--') {
                            parser.state = S.COMMENT;
                            parser.comment = '';
                            parser.sgmlDecl = '';
                        }
                        else if ((parser.sgmlDecl + c).toUpperCase() === DOCTYPE) {
                            parser.state = S.DOCTYPE;
                            if (parser.doctype || parser.sawRoot) {
                                strictFail(parser, 'Inappropriately located doctype declaration');
                            }
                            parser.doctype = '';
                            parser.sgmlDecl = '';
                        }
                        else if (c === '>') {
                            emitNode(parser, 'onsgmldeclaration', parser.sgmlDecl);
                            parser.sgmlDecl = '';
                            parser.state = S.TEXT;
                        }
                        else if (isQuote(c)) {
                            parser.state = S.SGML_DECL_QUOTED;
                            parser.sgmlDecl += c;
                        }
                        else {
                            parser.sgmlDecl += c;
                        }
                        continue;
                    case S.SGML_DECL_QUOTED:
                        if (c === parser.q) {
                            parser.state = S.SGML_DECL;
                            parser.q = '';
                        }
                        parser.sgmlDecl += c;
                        continue;
                    case S.DOCTYPE:
                        if (c === '>') {
                            parser.state = S.TEXT;
                            emitNode(parser, 'ondoctype', parser.doctype);
                            parser.doctype = true; // just remember that we saw it.
                        }
                        else {
                            parser.doctype += c;
                            if (c === '[') {
                                parser.state = S.DOCTYPE_DTD;
                            }
                            else if (isQuote(c)) {
                                parser.state = S.DOCTYPE_QUOTED;
                                parser.q = c;
                            }
                        }
                        continue;
                    case S.DOCTYPE_QUOTED:
                        parser.doctype += c;
                        if (c === parser.q) {
                            parser.q = '';
                            parser.state = S.DOCTYPE;
                        }
                        continue;
                    case S.DOCTYPE_DTD:
                        parser.doctype += c;
                        if (c === ']') {
                            parser.state = S.DOCTYPE;
                        }
                        else if (isQuote(c)) {
                            parser.state = S.DOCTYPE_DTD_QUOTED;
                            parser.q = c;
                        }
                        continue;
                    case S.DOCTYPE_DTD_QUOTED:
                        parser.doctype += c;
                        if (c === parser.q) {
                            parser.state = S.DOCTYPE_DTD;
                            parser.q = '';
                        }
                        continue;
                    case S.COMMENT:
                        if (c === '-') {
                            parser.state = S.COMMENT_ENDING;
                        }
                        else {
                            parser.comment += c;
                        }
                        continue;
                    case S.COMMENT_ENDING:
                        if (c === '-') {
                            parser.state = S.COMMENT_ENDED;
                            parser.comment = textopts(parser.opt, parser.comment);
                            if (parser.comment) {
                                emitNode(parser, 'oncomment', parser.comment);
                            }
                            parser.comment = '';
                        }
                        else {
                            parser.comment += '-' + c;
                            parser.state = S.COMMENT;
                        }
                        continue;
                    case S.COMMENT_ENDED:
                        if (c !== '>') {
                            strictFail(parser, 'Malformed comment');
                            // allow <!-- blah -- bloo --> in non-strict mode,
                            // which is a comment of " blah -- bloo "
                            parser.comment += '--' + c;
                            parser.state = S.COMMENT;
                        }
                        else {
                            parser.state = S.TEXT;
                        }
                        continue;
                    case S.CDATA:
                        if (c === ']') {
                            parser.state = S.CDATA_ENDING;
                        }
                        else {
                            parser.cdata += c;
                        }
                        continue;
                    case S.CDATA_ENDING:
                        if (c === ']') {
                            parser.state = S.CDATA_ENDING_2;
                        }
                        else {
                            parser.cdata += ']' + c;
                            parser.state = S.CDATA;
                        }
                        continue;
                    case S.CDATA_ENDING_2:
                        if (c === '>') {
                            if (parser.cdata) {
                                emitNode(parser, 'oncdata', parser.cdata);
                            }
                            emitNode(parser, 'onclosecdata');
                            parser.cdata = '';
                            parser.state = S.TEXT;
                        }
                        else if (c === ']') {
                            parser.cdata += ']';
                        }
                        else {
                            parser.cdata += ']]' + c;
                            parser.state = S.CDATA;
                        }
                        continue;
                    case S.PROC_INST:
                        if (c === '?') {
                            parser.state = S.PROC_INST_ENDING;
                        }
                        else if (isWhitespace(c)) {
                            parser.state = S.PROC_INST_BODY;
                        }
                        else {
                            parser.procInstName += c;
                        }
                        continue;
                    case S.PROC_INST_BODY:
                        if (!parser.procInstBody && isWhitespace(c)) {
                            continue;
                        }
                        else if (c === '?') {
                            parser.state = S.PROC_INST_ENDING;
                        }
                        else {
                            parser.procInstBody += c;
                        }
                        continue;
                    case S.PROC_INST_ENDING:
                        if (c === '>') {
                            emitNode(parser, 'onprocessinginstruction', {
                                name: parser.procInstName,
                                body: parser.procInstBody
                            });
                            parser.procInstName = parser.procInstBody = '';
                            parser.state = S.TEXT;
                        }
                        else {
                            parser.procInstBody += '?' + c;
                            parser.state = S.PROC_INST_BODY;
                        }
                        continue;
                    case S.OPEN_TAG:
                        if (isMatch(nameBody, c)) {
                            parser.tagName += c;
                        }
                        else {
                            newTag(parser);
                            if (c === '>') {
                                openTag(parser);
                            }
                            else if (c === '/') {
                                parser.state = S.OPEN_TAG_SLASH;
                            }
                            else {
                                if (!isWhitespace(c)) {
                                    strictFail(parser, 'Invalid character in tag name');
                                }
                                parser.state = S.ATTRIB;
                            }
                        }
                        continue;
                    case S.OPEN_TAG_SLASH:
                        if (c === '>') {
                            openTag(parser, true);
                            closeTag(parser);
                        }
                        else {
                            strictFail(parser, 'Forward-slash in opening tag not followed by >');
                            parser.state = S.ATTRIB;
                        }
                        continue;
                    case S.ATTRIB:
                        // haven't read the attribute name yet.
                        if (isWhitespace(c)) {
                            continue;
                        }
                        else if (c === '>') {
                            openTag(parser);
                        }
                        else if (c === '/') {
                            parser.state = S.OPEN_TAG_SLASH;
                        }
                        else if (isMatch(nameStart, c)) {
                            parser.attribName = c;
                            parser.attribValue = '';
                            parser.state = S.ATTRIB_NAME;
                        }
                        else {
                            strictFail(parser, 'Invalid attribute name');
                        }
                        continue;
                    case S.ATTRIB_NAME:
                        if (c === '=') {
                            parser.state = S.ATTRIB_VALUE;
                        }
                        else if (c === '>') {
                            strictFail(parser, 'Attribute without value');
                            parser.attribValue = parser.attribName;
                            attrib(parser);
                            openTag(parser);
                        }
                        else if (isWhitespace(c)) {
                            parser.state = S.ATTRIB_NAME_SAW_WHITE;
                        }
                        else if (isMatch(nameBody, c)) {
                            parser.attribName += c;
                        }
                        else {
                            strictFail(parser, 'Invalid attribute name');
                        }
                        continue;
                    case S.ATTRIB_NAME_SAW_WHITE:
                        if (c === '=') {
                            parser.state = S.ATTRIB_VALUE;
                        }
                        else if (isWhitespace(c)) {
                            continue;
                        }
                        else {
                            strictFail(parser, 'Attribute without value');
                            parser.tag.attributes[parser.attribName] = '';
                            parser.attribValue = '';
                            emitNode(parser, 'onattribute', {
                                name: parser.attribName,
                                value: ''
                            });
                            parser.attribName = '';
                            if (c === '>') {
                                openTag(parser);
                            }
                            else if (isMatch(nameStart, c)) {
                                parser.attribName = c;
                                parser.state = S.ATTRIB_NAME;
                            }
                            else {
                                strictFail(parser, 'Invalid attribute name');
                                parser.state = S.ATTRIB;
                            }
                        }
                        continue;
                    case S.ATTRIB_VALUE:
                        if (isWhitespace(c)) {
                            continue;
                        }
                        else if (isQuote(c)) {
                            parser.q = c;
                            parser.state = S.ATTRIB_VALUE_QUOTED;
                        }
                        else {
                            strictFail(parser, 'Unquoted attribute value');
                            parser.state = S.ATTRIB_VALUE_UNQUOTED;
                            parser.attribValue = c;
                        }
                        continue;
                    case S.ATTRIB_VALUE_QUOTED:
                        if (c !== parser.q) {
                            if (c === '&') {
                                parser.state = S.ATTRIB_VALUE_ENTITY_Q;
                            }
                            else {
                                parser.attribValue += c;
                            }
                            continue;
                        }
                        attrib(parser);
                        parser.q = '';
                        parser.state = S.ATTRIB_VALUE_CLOSED;
                        continue;
                    case S.ATTRIB_VALUE_CLOSED:
                        if (isWhitespace(c)) {
                            parser.state = S.ATTRIB;
                        }
                        else if (c === '>') {
                            openTag(parser);
                        }
                        else if (c === '/') {
                            parser.state = S.OPEN_TAG_SLASH;
                        }
                        else if (isMatch(nameStart, c)) {
                            strictFail(parser, 'No whitespace between attributes');
                            parser.attribName = c;
                            parser.attribValue = '';
                            parser.state = S.ATTRIB_NAME;
                        }
                        else {
                            strictFail(parser, 'Invalid attribute name');
                        }
                        continue;
                    case S.ATTRIB_VALUE_UNQUOTED:
                        if (!isAttribEnd(c)) {
                            if (c === '&') {
                                parser.state = S.ATTRIB_VALUE_ENTITY_U;
                            }
                            else {
                                parser.attribValue += c;
                            }
                            continue;
                        }
                        attrib(parser);
                        if (c === '>') {
                            openTag(parser);
                        }
                        else {
                            parser.state = S.ATTRIB;
                        }
                        continue;
                    case S.CLOSE_TAG:
                        if (!parser.tagName) {
                            if (isWhitespace(c)) {
                                continue;
                            }
                            else if (notMatch(nameStart, c)) {
                                if (parser.script) {
                                    parser.script += '</' + c;
                                    parser.state = S.SCRIPT;
                                }
                                else {
                                    strictFail(parser, 'Invalid tagname in closing tag.');
                                }
                            }
                            else {
                                parser.tagName = c;
                            }
                        }
                        else if (c === '>') {
                            closeTag(parser);
                        }
                        else if (isMatch(nameBody, c)) {
                            parser.tagName += c;
                        }
                        else if (parser.script) {
                            parser.script += '</' + parser.tagName;
                            parser.tagName = '';
                            parser.state = S.SCRIPT;
                        }
                        else {
                            if (!isWhitespace(c)) {
                                strictFail(parser, 'Invalid tagname in closing tag');
                            }
                            parser.state = S.CLOSE_TAG_SAW_WHITE;
                        }
                        continue;
                    case S.CLOSE_TAG_SAW_WHITE:
                        if (isWhitespace(c)) {
                            continue;
                        }
                        if (c === '>') {
                            closeTag(parser);
                        }
                        else {
                            strictFail(parser, 'Invalid characters in closing tag');
                        }
                        continue;
                    case S.TEXT_ENTITY:
                    case S.ATTRIB_VALUE_ENTITY_Q:
                    case S.ATTRIB_VALUE_ENTITY_U:
                        var returnState;
                        var buffer;
                        switch (parser.state) {
                            case S.TEXT_ENTITY:
                                returnState = S.TEXT;
                                buffer = 'textNode';
                                break;
                            case S.ATTRIB_VALUE_ENTITY_Q:
                                returnState = S.ATTRIB_VALUE_QUOTED;
                                buffer = 'attribValue';
                                break;
                            case S.ATTRIB_VALUE_ENTITY_U:
                                returnState = S.ATTRIB_VALUE_UNQUOTED;
                                buffer = 'attribValue';
                                break;
                        }
                        if (c === ';') {
                            parser[buffer] += parseEntity(parser);
                            parser.entity = '';
                            parser.state = returnState;
                        }
                        else if (isMatch(parser.entity.length ? entityBody : entityStart, c)) {
                            parser.entity += c;
                        }
                        else {
                            strictFail(parser, 'Invalid character in entity name');
                            parser[buffer] += '&' + parser.entity + c;
                            parser.entity = '';
                            parser.state = returnState;
                        }
                        continue;
                    default:
                        throw new Error(parser, 'Unknown state: ' + parser.state);
                }
            } // while
            if (parser.position >= parser.bufferCheckPosition) {
                checkBufferLength(parser);
            }
            return parser;
        }
        /*! http://mths.be/fromcodepoint v0.1.0 by @mathias */
        /* istanbul ignore next */
        if (!String.fromCodePoint) {
            (function () {
                var stringFromCharCode = String.fromCharCode;
                var floor = Math.floor;
                var fromCodePoint = function () {
                    var MAX_SIZE = 0x4000;
                    var codeUnits = [];
                    var highSurrogate;
                    var lowSurrogate;
                    var index = -1;
                    var length = arguments.length;
                    if (!length) {
                        return '';
                    }
                    var result = '';
                    while (++index < length) {
                        var codePoint = Number(arguments[index]);
                        if (!isFinite(codePoint) || // `NaN`, `+Infinity`, or `-Infinity`
                            codePoint < 0 || // not a valid Unicode code point
                            codePoint > 0x10FFFF || // not a valid Unicode code point
                            floor(codePoint) !== codePoint // not an integer
                        ) {
                            throw RangeError('Invalid code point: ' + codePoint);
                        }
                        if (codePoint <= 0xFFFF) { // BMP code point
                            codeUnits.push(codePoint);
                        }
                        else { // Astral code point; split in surrogate halves
                            // http://mathiasbynens.be/notes/javascript-encoding#surrogate-formulae
                            codePoint -= 0x10000;
                            highSurrogate = (codePoint >> 10) + 0xD800;
                            lowSurrogate = (codePoint % 0x400) + 0xDC00;
                            codeUnits.push(highSurrogate, lowSurrogate);
                        }
                        if (index + 1 === length || codeUnits.length > MAX_SIZE) {
                            result += stringFromCharCode.apply(null, codeUnits);
                            codeUnits.length = 0;
                        }
                    }
                    return result;
                };
                /* istanbul ignore next */
                if (Object.defineProperty) {
                    Object.defineProperty(String, 'fromCodePoint', {
                        value: fromCodePoint,
                        configurable: true,
                        writable: true
                    });
                }
                else {
                    String.fromCodePoint = fromCodePoint;
                }
            }());
        }
    })(typeof exports === 'undefined' ? this.sax = {} : exports);
},
/* needle/lib/decoder.js */ function _(require, module, exports, __esModule, __esExport) {
    var iconv, inherits = require('util').inherits, stream = require('stream');
    var regex = /(?:charset|encoding)\s*=\s*['"]? *([\w\-]+)/i;
    inherits(StreamDecoder, stream.Transform);
    function StreamDecoder(charset) {
        if (!(this instanceof StreamDecoder))
            return new StreamDecoder(charset);
        stream.Transform.call(this, charset);
        this.charset = charset;
        this.parsed_chunk = false;
    }
    StreamDecoder.prototype._transform = function (chunk, encoding, done) {
        var res, found;
        // try get charset from chunk, just once
        if (this.charset == 'utf8' && !this.parsed_chunk) {
            this.parsed_chunk = true;
            var matches = regex.exec(chunk.toString());
            if (matches) {
                found = matches[1].toLowerCase();
                this.charset = found == 'utf-8' ? 'utf8' : found;
            }
        }
        try {
            res = iconv.decode(chunk, this.charset);
        }
        catch (e) { // something went wrong, just return original chunk
            res = chunk;
        }
        this.push(res);
        done();
    };
    module.exports = function (charset) {
        try {
            if (!iconv)
                iconv = require(84) /* iconv-lite */;
        }
        catch (e) {
            /* iconv not found */
        }
        if (iconv)
            return new StreamDecoder(charset);
        else
            return new stream.PassThrough;
    };
},
/* iconv-lite/lib/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // Some environments don't have global Buffer (e.g. React Native).
    // Solution would be installing npm modules "buffer" and "stream" explicitly.
    var Buffer = require(85) /* safer-buffer */.Buffer;
    var bomHandling = require(86) /* ./bom-handling */, iconv = module.exports;
    // All codecs and aliases are kept here, keyed by encoding name/alias.
    // They are lazy loaded in `iconv.getCodec` from `encodings/index.js`.
    iconv.encodings = null;
    // Characters emitted in case of error.
    iconv.defaultCharUnicode = '�';
    iconv.defaultCharSingleByte = '?';
    // Public API.
    iconv.encode = function encode(str, encoding, options) {
        str = "" + (str || ""); // Ensure string.
        var encoder = iconv.getEncoder(encoding, options);
        var res = encoder.write(str);
        var trail = encoder.end();
        return (trail && trail.length > 0) ? Buffer.concat([res, trail]) : res;
    };
    iconv.decode = function decode(buf, encoding, options) {
        if (typeof buf === 'string') {
            if (!iconv.skipDecodeWarning) {
                console.error('Iconv-lite warning: decode()-ing strings is deprecated. Refer to https://github.com/ashtuchkin/iconv-lite/wiki/Use-Buffers-when-decoding');
                iconv.skipDecodeWarning = true;
            }
            buf = Buffer.from("" + (buf || ""), "binary"); // Ensure buffer.
        }
        var decoder = iconv.getDecoder(encoding, options);
        var res = decoder.write(buf);
        var trail = decoder.end();
        return trail ? (res + trail) : res;
    };
    iconv.encodingExists = function encodingExists(enc) {
        try {
            iconv.getCodec(enc);
            return true;
        }
        catch (e) {
            return false;
        }
    };
    // Legacy aliases to convert functions
    iconv.toEncoding = iconv.encode;
    iconv.fromEncoding = iconv.decode;
    // Search for a codec in iconv.encodings. Cache codec data in iconv._codecDataCache.
    iconv._codecDataCache = {};
    iconv.getCodec = function getCodec(encoding) {
        if (!iconv.encodings)
            iconv.encodings = require(87) /* ../encodings */; // Lazy load all encoding definitions.
        // Canonicalize encoding name: strip all non-alphanumeric chars and appended year.
        var enc = iconv._canonicalizeEncoding(encoding);
        // Traverse iconv.encodings to find actual codec.
        var codecOptions = {};
        while (true) {
            var codec = iconv._codecDataCache[enc];
            if (codec)
                return codec;
            var codecDef = iconv.encodings[enc];
            switch (typeof codecDef) {
                case "string": // Direct alias to other encoding.
                    enc = codecDef;
                    break;
                case "object": // Alias with options. Can be layered.
                    for (var key in codecDef)
                        codecOptions[key] = codecDef[key];
                    if (!codecOptions.encodingName)
                        codecOptions.encodingName = enc;
                    enc = codecDef.type;
                    break;
                case "function": // Codec itself.
                    if (!codecOptions.encodingName)
                        codecOptions.encodingName = enc;
                    // The codec function must load all tables and return object with .encoder and .decoder methods.
                    // It'll be called only once (for each different options object).
                    codec = new codecDef(codecOptions, iconv);
                    iconv._codecDataCache[codecOptions.encodingName] = codec; // Save it to be reused later.
                    return codec;
                default:
                    throw new Error("Encoding not recognized: '" + encoding + "' (searched as: '" + enc + "')");
            }
        }
    };
    iconv._canonicalizeEncoding = function (encoding) {
        // Canonicalize encoding name: strip all non-alphanumeric chars and appended year.
        return ('' + encoding).toLowerCase().replace(/:\d{4}$|[^0-9a-z]/g, "");
    };
    iconv.getEncoder = function getEncoder(encoding, options) {
        var codec = iconv.getCodec(encoding), encoder = new codec.encoder(options, codec);
        if (codec.bomAware && options && options.addBOM)
            encoder = new bomHandling.PrependBOM(encoder, options);
        return encoder;
    };
    iconv.getDecoder = function getDecoder(encoding, options) {
        var codec = iconv.getCodec(encoding), decoder = new codec.decoder(options, codec);
        if (codec.bomAware && !(options && options.stripBOM === false))
            decoder = new bomHandling.StripBOM(decoder, options);
        return decoder;
    };
    // Load extensions in Node. All of them are omitted in Browserify build via 'browser' field in package.json.
    var nodeVer = typeof process !== 'undefined' && process.versions && process.versions.node;
    if (nodeVer) {
        // Load streaming support in Node v0.10+
        var nodeVerArr = nodeVer.split(".").map(Number);
        if (nodeVerArr[0] > 0 || nodeVerArr[1] >= 10) {
            require(104) /* ./streams */(iconv);
        }
        // Load Node primitive extensions.
        require(105) /* ./extend-node */(iconv);
    }
    if ("Ā" != "\u0100") {
        console.error("iconv-lite warning: javascript files use encoding different from utf-8. See https://github.com/ashtuchkin/iconv-lite/wiki/Javascript-source-file-encodings for more info.");
    }
},
/* safer-buffer/safer.js */ function _(require, module, exports, __esModule, __esExport) {
    var buffer = require('buffer');
    var Buffer = buffer.Buffer;
    var safer = {};
    var key;
    for (key in buffer) {
        if (!buffer.hasOwnProperty(key))
            continue;
        if (key === 'SlowBuffer' || key === 'Buffer')
            continue;
        safer[key] = buffer[key];
    }
    var Safer = safer.Buffer = {};
    for (key in Buffer) {
        if (!Buffer.hasOwnProperty(key))
            continue;
        if (key === 'allocUnsafe' || key === 'allocUnsafeSlow')
            continue;
        Safer[key] = Buffer[key];
    }
    safer.Buffer.prototype = Buffer.prototype;
    if (!Safer.from || Safer.from === Uint8Array.from) {
        Safer.from = function (value, encodingOrOffset, length) {
            if (typeof value === 'number') {
                throw new TypeError('The "value" argument must not be of type number. Received type ' + typeof value);
            }
            if (value && typeof value.length === 'undefined') {
                throw new TypeError('The first argument must be one of type string, Buffer, ArrayBuffer, Array, or Array-like Object. Received type ' + typeof value);
            }
            return Buffer(value, encodingOrOffset, length);
        };
    }
    if (!Safer.alloc) {
        Safer.alloc = function (size, fill, encoding) {
            if (typeof size !== 'number') {
                throw new TypeError('The "size" argument must be of type number. Received type ' + typeof size);
            }
            if (size < 0 || size >= 2 * (1 << 30)) {
                throw new RangeError('The value "' + size + '" is invalid for option "size"');
            }
            var buf = Buffer(size);
            if (!fill || fill.length === 0) {
                buf.fill(0);
            }
            else if (typeof encoding === 'string') {
                buf.fill(fill, encoding);
            }
            else {
                buf.fill(fill);
            }
            return buf;
        };
    }
    if (!safer.kStringMaxLength) {
        try {
            safer.kStringMaxLength = process.binding('buffer').kStringMaxLength;
        }
        catch (e) {
            // we can't determine kStringMaxLength in environments where process.binding
            // is unsupported, so let's not set it
        }
    }
    if (!safer.constants) {
        safer.constants = {
            MAX_LENGTH: safer.kMaxLength
        };
        if (safer.kStringMaxLength) {
            safer.constants.MAX_STRING_LENGTH = safer.kStringMaxLength;
        }
    }
    module.exports = safer;
},
/* iconv-lite/lib/bom-handling.js */ function _(require, module, exports, __esModule, __esExport) {
    var BOMChar = '\uFEFF';
    exports.PrependBOM = PrependBOMWrapper;
    function PrependBOMWrapper(encoder, options) {
        this.encoder = encoder;
        this.addBOM = true;
    }
    PrependBOMWrapper.prototype.write = function (str) {
        if (this.addBOM) {
            str = BOMChar + str;
            this.addBOM = false;
        }
        return this.encoder.write(str);
    };
    PrependBOMWrapper.prototype.end = function () {
        return this.encoder.end();
    };
    //------------------------------------------------------------------------------
    exports.StripBOM = StripBOMWrapper;
    function StripBOMWrapper(decoder, options) {
        this.decoder = decoder;
        this.pass = false;
        this.options = options || {};
    }
    StripBOMWrapper.prototype.write = function (buf) {
        var res = this.decoder.write(buf);
        if (this.pass || !res)
            return res;
        if (res[0] === BOMChar) {
            res = res.slice(1);
            if (typeof this.options.stripBOM === 'function')
                this.options.stripBOM();
        }
        this.pass = true;
        return res;
    };
    StripBOMWrapper.prototype.end = function () {
        return this.decoder.end();
    };
},
/* iconv-lite/encodings/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // Update this array if you add/rename/remove files in this directory.
    // We support Browserify by skipping automatic module discovery and requiring modules directly.
    var modules = [
        require(88) /* ./internal */,
        require(89) /* ./utf16 */,
        require(90) /* ./utf7 */,
        require(91) /* ./sbcs-codec */,
        require(92) /* ./sbcs-data */,
        require(93) /* ./sbcs-data-generated */,
        require(94) /* ./dbcs-codec */,
        require(95) /* ./dbcs-data */,
    ];
    // Put all encoding/alias/codec definitions to single object and export it. 
    for (var i = 0; i < modules.length; i++) {
        var module = modules[i];
        for (var enc in module)
            if (Object.prototype.hasOwnProperty.call(module, enc))
                exports[enc] = module[enc];
    }
},
/* iconv-lite/encodings/internal.js */ function _(require, module, exports, __esModule, __esExport) {
    var Buffer = require(85) /* safer-buffer */.Buffer;
    // Export Node.js internal encodings.
    module.exports = {
        // Encodings
        utf8: { type: "_internal", bomAware: true },
        cesu8: { type: "_internal", bomAware: true },
        unicode11utf8: "utf8",
        ucs2: { type: "_internal", bomAware: true },
        utf16le: "ucs2",
        binary: { type: "_internal" },
        base64: { type: "_internal" },
        hex: { type: "_internal" },
        // Codec.
        _internal: InternalCodec,
    };
    //------------------------------------------------------------------------------
    function InternalCodec(codecOptions, iconv) {
        this.enc = codecOptions.encodingName;
        this.bomAware = codecOptions.bomAware;
        if (this.enc === "base64")
            this.encoder = InternalEncoderBase64;
        else if (this.enc === "cesu8") {
            this.enc = "utf8"; // Use utf8 for decoding.
            this.encoder = InternalEncoderCesu8;
            // Add decoder for versions of Node not supporting CESU-8
            if (Buffer.from('eda0bdedb2a9', 'hex').toString() !== '💩') {
                this.decoder = InternalDecoderCesu8;
                this.defaultCharUnicode = iconv.defaultCharUnicode;
            }
        }
    }
    InternalCodec.prototype.encoder = InternalEncoder;
    InternalCodec.prototype.decoder = InternalDecoder;
    //------------------------------------------------------------------------------
    // We use node.js internal decoder. Its signature is the same as ours.
    var StringDecoder = require('string_decoder').StringDecoder;
    if (!StringDecoder.prototype.end) // Node v0.8 doesn't have this method.
        StringDecoder.prototype.end = function () { };
    function InternalDecoder(options, codec) {
        StringDecoder.call(this, codec.enc);
    }
    InternalDecoder.prototype = StringDecoder.prototype;
    //------------------------------------------------------------------------------
    // Encoder is mostly trivial
    function InternalEncoder(options, codec) {
        this.enc = codec.enc;
    }
    InternalEncoder.prototype.write = function (str) {
        return Buffer.from(str, this.enc);
    };
    InternalEncoder.prototype.end = function () {
    };
    //------------------------------------------------------------------------------
    // Except base64 encoder, which must keep its state.
    function InternalEncoderBase64(options, codec) {
        this.prevStr = '';
    }
    InternalEncoderBase64.prototype.write = function (str) {
        str = this.prevStr + str;
        var completeQuads = str.length - (str.length % 4);
        this.prevStr = str.slice(completeQuads);
        str = str.slice(0, completeQuads);
        return Buffer.from(str, "base64");
    };
    InternalEncoderBase64.prototype.end = function () {
        return Buffer.from(this.prevStr, "base64");
    };
    //------------------------------------------------------------------------------
    // CESU-8 encoder is also special.
    function InternalEncoderCesu8(options, codec) {
    }
    InternalEncoderCesu8.prototype.write = function (str) {
        var buf = Buffer.alloc(str.length * 3), bufIdx = 0;
        for (var i = 0; i < str.length; i++) {
            var charCode = str.charCodeAt(i);
            // Naive implementation, but it works because CESU-8 is especially easy
            // to convert from UTF-16 (which all JS strings are encoded in).
            if (charCode < 0x80)
                buf[bufIdx++] = charCode;
            else if (charCode < 0x800) {
                buf[bufIdx++] = 0xC0 + (charCode >>> 6);
                buf[bufIdx++] = 0x80 + (charCode & 0x3f);
            }
            else { // charCode will always be < 0x10000 in javascript.
                buf[bufIdx++] = 0xE0 + (charCode >>> 12);
                buf[bufIdx++] = 0x80 + ((charCode >>> 6) & 0x3f);
                buf[bufIdx++] = 0x80 + (charCode & 0x3f);
            }
        }
        return buf.slice(0, bufIdx);
    };
    InternalEncoderCesu8.prototype.end = function () {
    };
    //------------------------------------------------------------------------------
    // CESU-8 decoder is not implemented in Node v4.0+
    function InternalDecoderCesu8(options, codec) {
        this.acc = 0;
        this.contBytes = 0;
        this.accBytes = 0;
        this.defaultCharUnicode = codec.defaultCharUnicode;
    }
    InternalDecoderCesu8.prototype.write = function (buf) {
        var acc = this.acc, contBytes = this.contBytes, accBytes = this.accBytes, res = '';
        for (var i = 0; i < buf.length; i++) {
            var curByte = buf[i];
            if ((curByte & 0xC0) !== 0x80) { // Leading byte
                if (contBytes > 0) { // Previous code is invalid
                    res += this.defaultCharUnicode;
                    contBytes = 0;
                }
                if (curByte < 0x80) { // Single-byte code
                    res += String.fromCharCode(curByte);
                }
                else if (curByte < 0xE0) { // Two-byte code
                    acc = curByte & 0x1F;
                    contBytes = 1;
                    accBytes = 1;
                }
                else if (curByte < 0xF0) { // Three-byte code
                    acc = curByte & 0x0F;
                    contBytes = 2;
                    accBytes = 1;
                }
                else { // Four or more are not supported for CESU-8.
                    res += this.defaultCharUnicode;
                }
            }
            else { // Continuation byte
                if (contBytes > 0) { // We're waiting for it.
                    acc = (acc << 6) | (curByte & 0x3f);
                    contBytes--;
                    accBytes++;
                    if (contBytes === 0) {
                        // Check for overlong encoding, but support Modified UTF-8 (encoding NULL as C0 80)
                        if (accBytes === 2 && acc < 0x80 && acc > 0)
                            res += this.defaultCharUnicode;
                        else if (accBytes === 3 && acc < 0x800)
                            res += this.defaultCharUnicode;
                        else
                            // Actually add character.
                            res += String.fromCharCode(acc);
                    }
                }
                else { // Unexpected continuation byte
                    res += this.defaultCharUnicode;
                }
            }
        }
        this.acc = acc;
        this.contBytes = contBytes;
        this.accBytes = accBytes;
        return res;
    };
    InternalDecoderCesu8.prototype.end = function () {
        var res = 0;
        if (this.contBytes > 0)
            res += this.defaultCharUnicode;
        return res;
    };
},
/* iconv-lite/encodings/utf16.js */ function _(require, module, exports, __esModule, __esExport) {
    var Buffer = require(85) /* safer-buffer */.Buffer;
    // Note: UTF16-LE (or UCS2) codec is Node.js native. See encodings/internal.js
    // == UTF16-BE codec. ==========================================================
    exports.utf16be = Utf16BECodec;
    function Utf16BECodec() {
    }
    Utf16BECodec.prototype.encoder = Utf16BEEncoder;
    Utf16BECodec.prototype.decoder = Utf16BEDecoder;
    Utf16BECodec.prototype.bomAware = true;
    // -- Encoding
    function Utf16BEEncoder() {
    }
    Utf16BEEncoder.prototype.write = function (str) {
        var buf = Buffer.from(str, 'ucs2');
        for (var i = 0; i < buf.length; i += 2) {
            var tmp = buf[i];
            buf[i] = buf[i + 1];
            buf[i + 1] = tmp;
        }
        return buf;
    };
    Utf16BEEncoder.prototype.end = function () {
    };
    // -- Decoding
    function Utf16BEDecoder() {
        this.overflowByte = -1;
    }
    Utf16BEDecoder.prototype.write = function (buf) {
        if (buf.length == 0)
            return '';
        var buf2 = Buffer.alloc(buf.length + 1), i = 0, j = 0;
        if (this.overflowByte !== -1) {
            buf2[0] = buf[0];
            buf2[1] = this.overflowByte;
            i = 1;
            j = 2;
        }
        for (; i < buf.length - 1; i += 2, j += 2) {
            buf2[j] = buf[i + 1];
            buf2[j + 1] = buf[i];
        }
        this.overflowByte = (i == buf.length - 1) ? buf[buf.length - 1] : -1;
        return buf2.slice(0, j).toString('ucs2');
    };
    Utf16BEDecoder.prototype.end = function () {
    };
    // == UTF-16 codec =============================================================
    // Decoder chooses automatically from UTF-16LE and UTF-16BE using BOM and space-based heuristic.
    // Defaults to UTF-16LE, as it's prevalent and default in Node.
    // http://en.wikipedia.org/wiki/UTF-16 and http://encoding.spec.whatwg.org/#utf-16le
    // Decoder default can be changed: iconv.decode(buf, 'utf16', {defaultEncoding: 'utf-16be'});
    // Encoder uses UTF-16LE and prepends BOM (which can be overridden with addBOM: false).
    exports.utf16 = Utf16Codec;
    function Utf16Codec(codecOptions, iconv) {
        this.iconv = iconv;
    }
    Utf16Codec.prototype.encoder = Utf16Encoder;
    Utf16Codec.prototype.decoder = Utf16Decoder;
    // -- Encoding (pass-through)
    function Utf16Encoder(options, codec) {
        options = options || {};
        if (options.addBOM === undefined)
            options.addBOM = true;
        this.encoder = codec.iconv.getEncoder('utf-16le', options);
    }
    Utf16Encoder.prototype.write = function (str) {
        return this.encoder.write(str);
    };
    Utf16Encoder.prototype.end = function () {
        return this.encoder.end();
    };
    // -- Decoding
    function Utf16Decoder(options, codec) {
        this.decoder = null;
        this.initialBytes = [];
        this.initialBytesLen = 0;
        this.options = options || {};
        this.iconv = codec.iconv;
    }
    Utf16Decoder.prototype.write = function (buf) {
        if (!this.decoder) {
            // Codec is not chosen yet. Accumulate initial bytes.
            this.initialBytes.push(buf);
            this.initialBytesLen += buf.length;
            if (this.initialBytesLen < 16) // We need more bytes to use space heuristic (see below)
                return '';
            // We have enough bytes -> detect endianness.
            var buf = Buffer.concat(this.initialBytes), encoding = detectEncoding(buf, this.options.defaultEncoding);
            this.decoder = this.iconv.getDecoder(encoding, this.options);
            this.initialBytes.length = this.initialBytesLen = 0;
        }
        return this.decoder.write(buf);
    };
    Utf16Decoder.prototype.end = function () {
        if (!this.decoder) {
            var buf = Buffer.concat(this.initialBytes), encoding = detectEncoding(buf, this.options.defaultEncoding);
            this.decoder = this.iconv.getDecoder(encoding, this.options);
            var res = this.decoder.write(buf), trail = this.decoder.end();
            return trail ? (res + trail) : res;
        }
        return this.decoder.end();
    };
    function detectEncoding(buf, defaultEncoding) {
        var enc = defaultEncoding || 'utf-16le';
        if (buf.length >= 2) {
            // Check BOM.
            if (buf[0] == 0xFE && buf[1] == 0xFF) // UTF-16BE BOM
                enc = 'utf-16be';
            else if (buf[0] == 0xFF && buf[1] == 0xFE) // UTF-16LE BOM
                enc = 'utf-16le';
            else {
                // No BOM found. Try to deduce encoding from initial content.
                // Most of the time, the content has ASCII chars (U+00**), but the opposite (U+**00) is uncommon.
                // So, we count ASCII as if it was LE or BE, and decide from that.
                var asciiCharsLE = 0, asciiCharsBE = 0, // Counts of chars in both positions
                _len = Math.min(buf.length - (buf.length % 2), 64); // Len is always even.
                for (var i = 0; i < _len; i += 2) {
                    if (buf[i] === 0 && buf[i + 1] !== 0)
                        asciiCharsBE++;
                    if (buf[i] !== 0 && buf[i + 1] === 0)
                        asciiCharsLE++;
                }
                if (asciiCharsBE > asciiCharsLE)
                    enc = 'utf-16be';
                else if (asciiCharsBE < asciiCharsLE)
                    enc = 'utf-16le';
            }
        }
        return enc;
    }
},
/* iconv-lite/encodings/utf7.js */ function _(require, module, exports, __esModule, __esExport) {
    var Buffer = require(85) /* safer-buffer */.Buffer;
    // UTF-7 codec, according to https://tools.ietf.org/html/rfc2152
    // See also below a UTF-7-IMAP codec, according to http://tools.ietf.org/html/rfc3501#section-5.1.3
    exports.utf7 = Utf7Codec;
    exports.unicode11utf7 = 'utf7'; // Alias UNICODE-1-1-UTF-7
    function Utf7Codec(codecOptions, iconv) {
        this.iconv = iconv;
    }
    ;
    Utf7Codec.prototype.encoder = Utf7Encoder;
    Utf7Codec.prototype.decoder = Utf7Decoder;
    Utf7Codec.prototype.bomAware = true;
    // -- Encoding
    var nonDirectChars = /[^A-Za-z0-9'\(\),-\.\/:\? \n\r\t]+/g;
    function Utf7Encoder(options, codec) {
        this.iconv = codec.iconv;
    }
    Utf7Encoder.prototype.write = function (str) {
        // Naive implementation.
        // Non-direct chars are encoded as "+<base64>-"; single "+" char is encoded as "+-".
        return Buffer.from(str.replace(nonDirectChars, function (chunk) {
            return "+" + (chunk === '+' ? '' :
                this.iconv.encode(chunk, 'utf16-be').toString('base64').replace(/=+$/, ''))
                + "-";
        }.bind(this)));
    };
    Utf7Encoder.prototype.end = function () {
    };
    // -- Decoding
    function Utf7Decoder(options, codec) {
        this.iconv = codec.iconv;
        this.inBase64 = false;
        this.base64Accum = '';
    }
    var base64Regex = /[A-Za-z0-9\/+]/;
    var base64Chars = [];
    for (var i = 0; i < 256; i++)
        base64Chars[i] = base64Regex.test(String.fromCharCode(i));
    var plusChar = '+'.charCodeAt(0), minusChar = '-'.charCodeAt(0), andChar = '&'.charCodeAt(0);
    Utf7Decoder.prototype.write = function (buf) {
        var res = "", lastI = 0, inBase64 = this.inBase64, base64Accum = this.base64Accum;
        // The decoder is more involved as we must handle chunks in stream.
        for (var i = 0; i < buf.length; i++) {
            if (!inBase64) { // We're in direct mode.
                // Write direct chars until '+'
                if (buf[i] == plusChar) {
                    res += this.iconv.decode(buf.slice(lastI, i), "ascii"); // Write direct chars.
                    lastI = i + 1;
                    inBase64 = true;
                }
            }
            else { // We decode base64.
                if (!base64Chars[buf[i]]) { // Base64 ended.
                    if (i == lastI && buf[i] == minusChar) { // "+-" -> "+"
                        res += "+";
                    }
                    else {
                        var b64str = base64Accum + buf.slice(lastI, i).toString();
                        res += this.iconv.decode(Buffer.from(b64str, 'base64'), "utf16-be");
                    }
                    if (buf[i] != minusChar) // Minus is absorbed after base64.
                        i--;
                    lastI = i + 1;
                    inBase64 = false;
                    base64Accum = '';
                }
            }
        }
        if (!inBase64) {
            res += this.iconv.decode(buf.slice(lastI), "ascii"); // Write direct chars.
        }
        else {
            var b64str = base64Accum + buf.slice(lastI).toString();
            var canBeDecoded = b64str.length - (b64str.length % 8); // Minimal chunk: 2 quads -> 2x3 bytes -> 3 chars.
            base64Accum = b64str.slice(canBeDecoded); // The rest will be decoded in future.
            b64str = b64str.slice(0, canBeDecoded);
            res += this.iconv.decode(Buffer.from(b64str, 'base64'), "utf16-be");
        }
        this.inBase64 = inBase64;
        this.base64Accum = base64Accum;
        return res;
    };
    Utf7Decoder.prototype.end = function () {
        var res = "";
        if (this.inBase64 && this.base64Accum.length > 0)
            res = this.iconv.decode(Buffer.from(this.base64Accum, 'base64'), "utf16-be");
        this.inBase64 = false;
        this.base64Accum = '';
        return res;
    };
    // UTF-7-IMAP codec.
    // RFC3501 Sec. 5.1.3 Modified UTF-7 (http://tools.ietf.org/html/rfc3501#section-5.1.3)
    // Differences:
    //  * Base64 part is started by "&" instead of "+"
    //  * Direct characters are 0x20-0x7E, except "&" (0x26)
    //  * In Base64, "," is used instead of "/"
    //  * Base64 must not be used to represent direct characters.
    //  * No implicit shift back from Base64 (should always end with '-')
    //  * String must end in non-shifted position.
    //  * "-&" while in base64 is not allowed.
    exports.utf7imap = Utf7IMAPCodec;
    function Utf7IMAPCodec(codecOptions, iconv) {
        this.iconv = iconv;
    }
    ;
    Utf7IMAPCodec.prototype.encoder = Utf7IMAPEncoder;
    Utf7IMAPCodec.prototype.decoder = Utf7IMAPDecoder;
    Utf7IMAPCodec.prototype.bomAware = true;
    // -- Encoding
    function Utf7IMAPEncoder(options, codec) {
        this.iconv = codec.iconv;
        this.inBase64 = false;
        this.base64Accum = Buffer.alloc(6);
        this.base64AccumIdx = 0;
    }
    Utf7IMAPEncoder.prototype.write = function (str) {
        var inBase64 = this.inBase64, base64Accum = this.base64Accum, base64AccumIdx = this.base64AccumIdx, buf = Buffer.alloc(str.length * 5 + 10), bufIdx = 0;
        for (var i = 0; i < str.length; i++) {
            var uChar = str.charCodeAt(i);
            if (0x20 <= uChar && uChar <= 0x7E) { // Direct character or '&'.
                if (inBase64) {
                    if (base64AccumIdx > 0) {
                        bufIdx += buf.write(base64Accum.slice(0, base64AccumIdx).toString('base64').replace(/\//g, ',').replace(/=+$/, ''), bufIdx);
                        base64AccumIdx = 0;
                    }
                    buf[bufIdx++] = minusChar; // Write '-', then go to direct mode.
                    inBase64 = false;
                }
                if (!inBase64) {
                    buf[bufIdx++] = uChar; // Write direct character
                    if (uChar === andChar) // Ampersand -> '&-'
                        buf[bufIdx++] = minusChar;
                }
            }
            else { // Non-direct character
                if (!inBase64) {
                    buf[bufIdx++] = andChar; // Write '&', then go to base64 mode.
                    inBase64 = true;
                }
                if (inBase64) {
                    base64Accum[base64AccumIdx++] = uChar >> 8;
                    base64Accum[base64AccumIdx++] = uChar & 0xFF;
                    if (base64AccumIdx == base64Accum.length) {
                        bufIdx += buf.write(base64Accum.toString('base64').replace(/\//g, ','), bufIdx);
                        base64AccumIdx = 0;
                    }
                }
            }
        }
        this.inBase64 = inBase64;
        this.base64AccumIdx = base64AccumIdx;
        return buf.slice(0, bufIdx);
    };
    Utf7IMAPEncoder.prototype.end = function () {
        var buf = Buffer.alloc(10), bufIdx = 0;
        if (this.inBase64) {
            if (this.base64AccumIdx > 0) {
                bufIdx += buf.write(this.base64Accum.slice(0, this.base64AccumIdx).toString('base64').replace(/\//g, ',').replace(/=+$/, ''), bufIdx);
                this.base64AccumIdx = 0;
            }
            buf[bufIdx++] = minusChar; // Write '-', then go to direct mode.
            this.inBase64 = false;
        }
        return buf.slice(0, bufIdx);
    };
    // -- Decoding
    function Utf7IMAPDecoder(options, codec) {
        this.iconv = codec.iconv;
        this.inBase64 = false;
        this.base64Accum = '';
    }
    var base64IMAPChars = base64Chars.slice();
    base64IMAPChars[','.charCodeAt(0)] = true;
    Utf7IMAPDecoder.prototype.write = function (buf) {
        var res = "", lastI = 0, inBase64 = this.inBase64, base64Accum = this.base64Accum;
        // The decoder is more involved as we must handle chunks in stream.
        // It is forgiving, closer to standard UTF-7 (for example, '-' is optional at the end).
        for (var i = 0; i < buf.length; i++) {
            if (!inBase64) { // We're in direct mode.
                // Write direct chars until '&'
                if (buf[i] == andChar) {
                    res += this.iconv.decode(buf.slice(lastI, i), "ascii"); // Write direct chars.
                    lastI = i + 1;
                    inBase64 = true;
                }
            }
            else { // We decode base64.
                if (!base64IMAPChars[buf[i]]) { // Base64 ended.
                    if (i == lastI && buf[i] == minusChar) { // "&-" -> "&"
                        res += "&";
                    }
                    else {
                        var b64str = base64Accum + buf.slice(lastI, i).toString().replace(/,/g, '/');
                        res += this.iconv.decode(Buffer.from(b64str, 'base64'), "utf16-be");
                    }
                    if (buf[i] != minusChar) // Minus may be absorbed after base64.
                        i--;
                    lastI = i + 1;
                    inBase64 = false;
                    base64Accum = '';
                }
            }
        }
        if (!inBase64) {
            res += this.iconv.decode(buf.slice(lastI), "ascii"); // Write direct chars.
        }
        else {
            var b64str = base64Accum + buf.slice(lastI).toString().replace(/,/g, '/');
            var canBeDecoded = b64str.length - (b64str.length % 8); // Minimal chunk: 2 quads -> 2x3 bytes -> 3 chars.
            base64Accum = b64str.slice(canBeDecoded); // The rest will be decoded in future.
            b64str = b64str.slice(0, canBeDecoded);
            res += this.iconv.decode(Buffer.from(b64str, 'base64'), "utf16-be");
        }
        this.inBase64 = inBase64;
        this.base64Accum = base64Accum;
        return res;
    };
    Utf7IMAPDecoder.prototype.end = function () {
        var res = "";
        if (this.inBase64 && this.base64Accum.length > 0)
            res = this.iconv.decode(Buffer.from(this.base64Accum, 'base64'), "utf16-be");
        this.inBase64 = false;
        this.base64Accum = '';
        return res;
    };
},
/* iconv-lite/encodings/sbcs-codec.js */ function _(require, module, exports, __esModule, __esExport) {
    var Buffer = require(85) /* safer-buffer */.Buffer;
    // Single-byte codec. Needs a 'chars' string parameter that contains 256 or 128 chars that
    // correspond to encoded bytes (if 128 - then lower half is ASCII). 
    exports._sbcs = SBCSCodec;
    function SBCSCodec(codecOptions, iconv) {
        if (!codecOptions)
            throw new Error("SBCS codec is called without the data.");
        // Prepare char buffer for decoding.
        if (!codecOptions.chars || (codecOptions.chars.length !== 128 && codecOptions.chars.length !== 256))
            throw new Error("Encoding '" + codecOptions.type + "' has incorrect 'chars' (must be of len 128 or 256)");
        if (codecOptions.chars.length === 128) {
            var asciiString = "";
            for (var i = 0; i < 128; i++)
                asciiString += String.fromCharCode(i);
            codecOptions.chars = asciiString + codecOptions.chars;
        }
        this.decodeBuf = Buffer.from(codecOptions.chars, 'ucs2');
        // Encoding buffer.
        var encodeBuf = Buffer.alloc(65536, iconv.defaultCharSingleByte.charCodeAt(0));
        for (var i = 0; i < codecOptions.chars.length; i++)
            encodeBuf[codecOptions.chars.charCodeAt(i)] = i;
        this.encodeBuf = encodeBuf;
    }
    SBCSCodec.prototype.encoder = SBCSEncoder;
    SBCSCodec.prototype.decoder = SBCSDecoder;
    function SBCSEncoder(options, codec) {
        this.encodeBuf = codec.encodeBuf;
    }
    SBCSEncoder.prototype.write = function (str) {
        var buf = Buffer.alloc(str.length);
        for (var i = 0; i < str.length; i++)
            buf[i] = this.encodeBuf[str.charCodeAt(i)];
        return buf;
    };
    SBCSEncoder.prototype.end = function () {
    };
    function SBCSDecoder(options, codec) {
        this.decodeBuf = codec.decodeBuf;
    }
    SBCSDecoder.prototype.write = function (buf) {
        // Strings are immutable in JS -> we use ucs2 buffer to speed up computations.
        var decodeBuf = this.decodeBuf;
        var newBuf = Buffer.alloc(buf.length * 2);
        var idx1 = 0, idx2 = 0;
        for (var i = 0; i < buf.length; i++) {
            idx1 = buf[i] * 2;
            idx2 = i * 2;
            newBuf[idx2] = decodeBuf[idx1];
            newBuf[idx2 + 1] = decodeBuf[idx1 + 1];
        }
        return newBuf.toString('ucs2');
    };
    SBCSDecoder.prototype.end = function () {
    };
},
/* iconv-lite/encodings/sbcs-data.js */ function _(require, module, exports, __esModule, __esExport) {
    // Manually added data to be used by sbcs codec in addition to generated one.
    module.exports = {
        // Not supported by iconv, not sure why.
        "10029": "maccenteuro",
        "maccenteuro": {
            "type": "_sbcs",
            "chars": "ÄĀāÉĄÖÜáąČäčĆćéŹźĎíďĒēĖóėôöõúĚěü†°Ę£§•¶ß®©™ę¨≠ģĮįĪ≤≥īĶ∂∑łĻļĽľĹĺŅņŃ¬√ńŇ∆«»… ňŐÕőŌ–—“”‘’÷◊ōŔŕŘ‹›řŖŗŠ‚„šŚśÁŤťÍŽžŪÓÔūŮÚůŰűŲųÝýķŻŁżĢˇ"
        },
        "808": "cp808",
        "ibm808": "cp808",
        "cp808": {
            "type": "_sbcs",
            "chars": "АБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмноп░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀рстуфхцчшщъыьэюяЁёЄєЇїЎў°∙·√№€■ "
        },
        "mik": {
            "type": "_sbcs",
            "chars": "АБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмнопрстуфхцчшщъыьэюя└┴┬├─┼╣║╚╔╩╦╠═╬┐░▒▓│┤№§╗╝┘┌█▄▌▐▀αßΓπΣσµτΦΘΩδ∞φε∩≡±≥≤⌠⌡÷≈°∙·√ⁿ²■ "
        },
        // Aliases of generated encodings.
        "ascii8bit": "ascii",
        "usascii": "ascii",
        "ansix34": "ascii",
        "ansix341968": "ascii",
        "ansix341986": "ascii",
        "csascii": "ascii",
        "cp367": "ascii",
        "ibm367": "ascii",
        "isoir6": "ascii",
        "iso646us": "ascii",
        "iso646irv": "ascii",
        "us": "ascii",
        "latin1": "iso88591",
        "latin2": "iso88592",
        "latin3": "iso88593",
        "latin4": "iso88594",
        "latin5": "iso88599",
        "latin6": "iso885910",
        "latin7": "iso885913",
        "latin8": "iso885914",
        "latin9": "iso885915",
        "latin10": "iso885916",
        "csisolatin1": "iso88591",
        "csisolatin2": "iso88592",
        "csisolatin3": "iso88593",
        "csisolatin4": "iso88594",
        "csisolatincyrillic": "iso88595",
        "csisolatinarabic": "iso88596",
        "csisolatingreek": "iso88597",
        "csisolatinhebrew": "iso88598",
        "csisolatin5": "iso88599",
        "csisolatin6": "iso885910",
        "l1": "iso88591",
        "l2": "iso88592",
        "l3": "iso88593",
        "l4": "iso88594",
        "l5": "iso88599",
        "l6": "iso885910",
        "l7": "iso885913",
        "l8": "iso885914",
        "l9": "iso885915",
        "l10": "iso885916",
        "isoir14": "iso646jp",
        "isoir57": "iso646cn",
        "isoir100": "iso88591",
        "isoir101": "iso88592",
        "isoir109": "iso88593",
        "isoir110": "iso88594",
        "isoir144": "iso88595",
        "isoir127": "iso88596",
        "isoir126": "iso88597",
        "isoir138": "iso88598",
        "isoir148": "iso88599",
        "isoir157": "iso885910",
        "isoir166": "tis620",
        "isoir179": "iso885913",
        "isoir199": "iso885914",
        "isoir203": "iso885915",
        "isoir226": "iso885916",
        "cp819": "iso88591",
        "ibm819": "iso88591",
        "cyrillic": "iso88595",
        "arabic": "iso88596",
        "arabic8": "iso88596",
        "ecma114": "iso88596",
        "asmo708": "iso88596",
        "greek": "iso88597",
        "greek8": "iso88597",
        "ecma118": "iso88597",
        "elot928": "iso88597",
        "hebrew": "iso88598",
        "hebrew8": "iso88598",
        "turkish": "iso88599",
        "turkish8": "iso88599",
        "thai": "iso885911",
        "thai8": "iso885911",
        "celtic": "iso885914",
        "celtic8": "iso885914",
        "isoceltic": "iso885914",
        "tis6200": "tis620",
        "tis62025291": "tis620",
        "tis62025330": "tis620",
        "10000": "macroman",
        "10006": "macgreek",
        "10007": "maccyrillic",
        "10079": "maciceland",
        "10081": "macturkish",
        "cspc8codepage437": "cp437",
        "cspc775baltic": "cp775",
        "cspc850multilingual": "cp850",
        "cspcp852": "cp852",
        "cspc862latinhebrew": "cp862",
        "cpgr": "cp869",
        "msee": "cp1250",
        "mscyrl": "cp1251",
        "msansi": "cp1252",
        "msgreek": "cp1253",
        "msturk": "cp1254",
        "mshebr": "cp1255",
        "msarab": "cp1256",
        "winbaltrim": "cp1257",
        "cp20866": "koi8r",
        "20866": "koi8r",
        "ibm878": "koi8r",
        "cskoi8r": "koi8r",
        "cp21866": "koi8u",
        "21866": "koi8u",
        "ibm1168": "koi8u",
        "strk10482002": "rk1048",
        "tcvn5712": "tcvn",
        "tcvn57121": "tcvn",
        "gb198880": "iso646cn",
        "cn": "iso646cn",
        "csiso14jisc6220ro": "iso646jp",
        "jisc62201969ro": "iso646jp",
        "jp": "iso646jp",
        "cshproman8": "hproman8",
        "r8": "hproman8",
        "roman8": "hproman8",
        "xroman8": "hproman8",
        "ibm1051": "hproman8",
        "mac": "macintosh",
        "csmacintosh": "macintosh",
    };
},
/* iconv-lite/encodings/sbcs-data-generated.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated data for sbcs codec. Don't edit manually. Regenerate using generation/gen-sbcs.js script.
    module.exports = {
        "437": "cp437",
        "737": "cp737",
        "775": "cp775",
        "850": "cp850",
        "852": "cp852",
        "855": "cp855",
        "856": "cp856",
        "857": "cp857",
        "858": "cp858",
        "860": "cp860",
        "861": "cp861",
        "862": "cp862",
        "863": "cp863",
        "864": "cp864",
        "865": "cp865",
        "866": "cp866",
        "869": "cp869",
        "874": "windows874",
        "922": "cp922",
        "1046": "cp1046",
        "1124": "cp1124",
        "1125": "cp1125",
        "1129": "cp1129",
        "1133": "cp1133",
        "1161": "cp1161",
        "1162": "cp1162",
        "1163": "cp1163",
        "1250": "windows1250",
        "1251": "windows1251",
        "1252": "windows1252",
        "1253": "windows1253",
        "1254": "windows1254",
        "1255": "windows1255",
        "1256": "windows1256",
        "1257": "windows1257",
        "1258": "windows1258",
        "28591": "iso88591",
        "28592": "iso88592",
        "28593": "iso88593",
        "28594": "iso88594",
        "28595": "iso88595",
        "28596": "iso88596",
        "28597": "iso88597",
        "28598": "iso88598",
        "28599": "iso88599",
        "28600": "iso885910",
        "28601": "iso885911",
        "28603": "iso885913",
        "28604": "iso885914",
        "28605": "iso885915",
        "28606": "iso885916",
        "windows874": {
            "type": "_sbcs",
            "chars": "€����…�����������‘’“”•–—�������� กขฃคฅฆงจฉชซฌญฎฏฐฑฒณดตถทธนบปผฝพฟภมยรฤลฦวศษสหฬอฮฯะัาำิีึืฺุู����฿เแโใไๅๆ็่้๊๋์ํ๎๏๐๑๒๓๔๕๖๗๘๙๚๛����"
        },
        "win874": "windows874",
        "cp874": "windows874",
        "windows1250": {
            "type": "_sbcs",
            "chars": "€�‚�„…†‡�‰Š‹ŚŤŽŹ�‘’“”•–—�™š›śťžź ˇ˘Ł¤Ą¦§¨©Ş«¬­®Ż°±˛ł´µ¶·¸ąş»Ľ˝ľżŔÁÂĂÄĹĆÇČÉĘËĚÍÎĎĐŃŇÓÔŐÖ×ŘŮÚŰÜÝŢßŕáâăäĺćçčéęëěíîďđńňóôőö÷řůúűüýţ˙"
        },
        "win1250": "windows1250",
        "cp1250": "windows1250",
        "windows1251": {
            "type": "_sbcs",
            "chars": "ЂЃ‚ѓ„…†‡€‰Љ‹ЊЌЋЏђ‘’“”•–—�™љ›њќћџ ЎўЈ¤Ґ¦§Ё©Є«¬­®Ї°±Ііґµ¶·ё№є»јЅѕїАБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмнопрстуфхцчшщъыьэюя"
        },
        "win1251": "windows1251",
        "cp1251": "windows1251",
        "windows1252": {
            "type": "_sbcs",
            "chars": "€�‚ƒ„…†‡ˆ‰Š‹Œ�Ž��‘’“”•–—˜™š›œ�žŸ ¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ×ØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõö÷øùúûüýþÿ"
        },
        "win1252": "windows1252",
        "cp1252": "windows1252",
        "windows1253": {
            "type": "_sbcs",
            "chars": "€�‚ƒ„…†‡�‰�‹�����‘’“”•–—�™�›���� ΅Ά£¤¥¦§¨©�«¬­®―°±²³΄µ¶·ΈΉΊ»Ό½ΎΏΐΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡ�ΣΤΥΦΧΨΩΪΫάέήίΰαβγδεζηθικλμνξοπρςστυφχψωϊϋόύώ�"
        },
        "win1253": "windows1253",
        "cp1253": "windows1253",
        "windows1254": {
            "type": "_sbcs",
            "chars": "€�‚ƒ„…†‡ˆ‰Š‹Œ����‘’“”•–—˜™š›œ��Ÿ ¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏĞÑÒÓÔÕÖ×ØÙÚÛÜİŞßàáâãäåæçèéêëìíîïğñòóôõö÷øùúûüışÿ"
        },
        "win1254": "windows1254",
        "cp1254": "windows1254",
        "windows1255": {
            "type": "_sbcs",
            "chars": "€�‚ƒ„…†‡ˆ‰�‹�����‘’“”•–—˜™�›���� ¡¢£₪¥¦§¨©×«¬­®¯°±²³´µ¶·¸¹÷»¼½¾¿ְֱֲֳִֵֶַָֹֺֻּֽ־ֿ׀ׁׂ׃װױײ׳״�������אבגדהוזחטיךכלםמןנסעףפץצקרשת��‎‏�"
        },
        "win1255": "windows1255",
        "cp1255": "windows1255",
        "windows1256": {
            "type": "_sbcs",
            "chars": "€پ‚ƒ„…†‡ˆ‰ٹ‹Œچژڈگ‘’“”•–—ک™ڑ›œ‌‍ں ،¢£¤¥¦§¨©ھ«¬­®¯°±²³´µ¶·¸¹؛»¼½¾؟ہءآأؤإئابةتثجحخدذرزسشصض×طظعغـفقكàلâمنهوçèéêëىيîïًٌٍَôُِ÷ّùْûü‎‏ے"
        },
        "win1256": "windows1256",
        "cp1256": "windows1256",
        "windows1257": {
            "type": "_sbcs",
            "chars": "€�‚�„…†‡�‰�‹�¨ˇ¸�‘’“”•–—�™�›�¯˛� �¢£¤�¦§Ø©Ŗ«¬­®Æ°±²³´µ¶·ø¹ŗ»¼½¾æĄĮĀĆÄÅĘĒČÉŹĖĢĶĪĻŠŃŅÓŌÕÖ×ŲŁŚŪÜŻŽßąįāćäåęēčéźėģķīļšńņóōõö÷ųłśūüżž˙"
        },
        "win1257": "windows1257",
        "cp1257": "windows1257",
        "windows1258": {
            "type": "_sbcs",
            "chars": "€�‚ƒ„…†‡ˆ‰�‹Œ����‘’“”•–—˜™�›œ��Ÿ ¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂĂÄÅÆÇÈÉÊË̀ÍÎÏĐÑ̉ÓÔƠÖ×ØÙÚÛÜỮßàáâăäåæçèéêë́íîïđṇ̃óôơö÷øùúûüư₫ÿ"
        },
        "win1258": "windows1258",
        "cp1258": "windows1258",
        "iso88591": {
            "type": "_sbcs",
            "chars": " ¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ×ØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõö÷øùúûüýþÿ"
        },
        "cp28591": "iso88591",
        "iso88592": {
            "type": "_sbcs",
            "chars": " Ą˘Ł¤ĽŚ§¨ŠŞŤŹ­ŽŻ°ą˛ł´ľśˇ¸šşťź˝žżŔÁÂĂÄĹĆÇČÉĘËĚÍÎĎĐŃŇÓÔŐÖ×ŘŮÚŰÜÝŢßŕáâăäĺćçčéęëěíîďđńňóôőö÷řůúűüýţ˙"
        },
        "cp28592": "iso88592",
        "iso88593": {
            "type": "_sbcs",
            "chars": " Ħ˘£¤�Ĥ§¨İŞĞĴ­�Ż°ħ²³´µĥ·¸ışğĵ½�żÀÁÂ�ÄĊĈÇÈÉÊËÌÍÎÏ�ÑÒÓÔĠÖ×ĜÙÚÛÜŬŜßàáâ�äċĉçèéêëìíîï�ñòóôġö÷ĝùúûüŭŝ˙"
        },
        "cp28593": "iso88593",
        "iso88594": {
            "type": "_sbcs",
            "chars": " ĄĸŖ¤ĨĻ§¨ŠĒĢŦ­Ž¯°ą˛ŗ´ĩļˇ¸šēģŧŊžŋĀÁÂÃÄÅÆĮČÉĘËĖÍÎĪĐŅŌĶÔÕÖ×ØŲÚÛÜŨŪßāáâãäåæįčéęëėíîīđņōķôõö÷øųúûüũū˙"
        },
        "cp28594": "iso88594",
        "iso88595": {
            "type": "_sbcs",
            "chars": " ЁЂЃЄЅІЇЈЉЊЋЌ­ЎЏАБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмнопрстуфхцчшщъыьэюя№ёђѓєѕіїјљњћќ§ўџ"
        },
        "cp28595": "iso88595",
        "iso88596": {
            "type": "_sbcs",
            "chars": " ���¤�������،­�������������؛���؟�ءآأؤإئابةتثجحخدذرزسشصضطظعغ�����ـفقكلمنهوىيًٌٍَُِّْ�������������"
        },
        "cp28596": "iso88596",
        "iso88597": {
            "type": "_sbcs",
            "chars": " ‘’£€₯¦§¨©ͺ«¬­�―°±²³΄΅Ά·ΈΉΊ»Ό½ΎΏΐΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡ�ΣΤΥΦΧΨΩΪΫάέήίΰαβγδεζηθικλμνξοπρςστυφχψωϊϋόύώ�"
        },
        "cp28597": "iso88597",
        "iso88598": {
            "type": "_sbcs",
            "chars": " �¢£¤¥¦§¨©×«¬­®¯°±²³´µ¶·¸¹÷»¼½¾��������������������������������‗אבגדהוזחטיךכלםמןנסעףפץצקרשת��‎‏�"
        },
        "cp28598": "iso88598",
        "iso88599": {
            "type": "_sbcs",
            "chars": " ¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏĞÑÒÓÔÕÖ×ØÙÚÛÜİŞßàáâãäåæçèéêëìíîïğñòóôõö÷øùúûüışÿ"
        },
        "cp28599": "iso88599",
        "iso885910": {
            "type": "_sbcs",
            "chars": " ĄĒĢĪĨĶ§ĻĐŠŦŽ­ŪŊ°ąēģīĩķ·ļđšŧž―ūŋĀÁÂÃÄÅÆĮČÉĘËĖÍÎÏÐŅŌÓÔÕÖŨØŲÚÛÜÝÞßāáâãäåæįčéęëėíîïðņōóôõöũøųúûüýþĸ"
        },
        "cp28600": "iso885910",
        "iso885911": {
            "type": "_sbcs",
            "chars": " กขฃคฅฆงจฉชซฌญฎฏฐฑฒณดตถทธนบปผฝพฟภมยรฤลฦวศษสหฬอฮฯะัาำิีึืฺุู����฿เแโใไๅๆ็่้๊๋์ํ๎๏๐๑๒๓๔๕๖๗๘๙๚๛����"
        },
        "cp28601": "iso885911",
        "iso885913": {
            "type": "_sbcs",
            "chars": " ”¢£¤„¦§Ø©Ŗ«¬­®Æ°±²³“µ¶·ø¹ŗ»¼½¾æĄĮĀĆÄÅĘĒČÉŹĖĢĶĪĻŠŃŅÓŌÕÖ×ŲŁŚŪÜŻŽßąįāćäåęēčéźėģķīļšńņóōõö÷ųłśūüżž’"
        },
        "cp28603": "iso885913",
        "iso885914": {
            "type": "_sbcs",
            "chars": " Ḃḃ£ĊċḊ§Ẁ©ẂḋỲ­®ŸḞḟĠġṀṁ¶ṖẁṗẃṠỳẄẅṡÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏŴÑÒÓÔÕÖṪØÙÚÛÜÝŶßàáâãäåæçèéêëìíîïŵñòóôõöṫøùúûüýŷÿ"
        },
        "cp28604": "iso885914",
        "iso885915": {
            "type": "_sbcs",
            "chars": " ¡¢£€¥Š§š©ª«¬­®¯°±²³Žµ¶·ž¹º»ŒœŸ¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ×ØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõö÷øùúûüýþÿ"
        },
        "cp28605": "iso885915",
        "iso885916": {
            "type": "_sbcs",
            "chars": " ĄąŁ€„Š§š©Ș«Ź­źŻ°±ČłŽ”¶·žčș»ŒœŸżÀÁÂĂÄĆÆÇÈÉÊËÌÍÎÏĐŃÒÓÔŐÖŚŰÙÚÛÜĘȚßàáâăäćæçèéêëìíîïđńòóôőöśűùúûüęțÿ"
        },
        "cp28606": "iso885916",
        "cp437": {
            "type": "_sbcs",
            "chars": "ÇüéâäàåçêëèïîìÄÅÉæÆôöòûùÿÖÜ¢£¥₧ƒáíóúñÑªº¿⌐¬½¼¡«»░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀αßΓπΣσµτΦΘΩδ∞φε∩≡±≥≤⌠⌡÷≈°∙·√ⁿ²■ "
        },
        "ibm437": "cp437",
        "csibm437": "cp437",
        "cp737": {
            "type": "_sbcs",
            "chars": "ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩαβγδεζηθικλμνξοπρσςτυφχψ░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀ωάέήϊίόύϋώΆΈΉΊΌΎΏ±≥≤ΪΫ÷≈°∙·√ⁿ²■ "
        },
        "ibm737": "cp737",
        "csibm737": "cp737",
        "cp775": {
            "type": "_sbcs",
            "chars": "ĆüéāäģåćłēŖŗīŹÄÅÉæÆōöĢ¢ŚśÖÜø£Ø×¤ĀĪóŻżź”¦©®¬½¼Ł«»░▒▓│┤ĄČĘĖ╣║╗╝ĮŠ┐└┴┬├─┼ŲŪ╚╔╩╦╠═╬Žąčęėįšųūž┘┌█▄▌▐▀ÓßŌŃõÕµńĶķĻļņĒŅ’­±“¾¶§÷„°∙·¹³²■ "
        },
        "ibm775": "cp775",
        "csibm775": "cp775",
        "cp850": {
            "type": "_sbcs",
            "chars": "ÇüéâäàåçêëèïîìÄÅÉæÆôöòûùÿÖÜø£Ø×ƒáíóúñÑªº¿®¬½¼¡«»░▒▓│┤ÁÂÀ©╣║╗╝¢¥┐└┴┬├─┼ãÃ╚╔╩╦╠═╬¤ðÐÊËÈıÍÎÏ┘┌█▄¦Ì▀ÓßÔÒõÕµþÞÚÛÙýÝ¯´­±‗¾¶§÷¸°¨·¹³²■ "
        },
        "ibm850": "cp850",
        "csibm850": "cp850",
        "cp852": {
            "type": "_sbcs",
            "chars": "ÇüéâäůćçłëŐőîŹÄĆÉĹĺôöĽľŚśÖÜŤťŁ×čáíóúĄąŽžĘę¬źČş«»░▒▓│┤ÁÂĚŞ╣║╗╝Żż┐└┴┬├─┼Ăă╚╔╩╦╠═╬¤đĐĎËďŇÍÎě┘┌█▄ŢŮ▀ÓßÔŃńňŠšŔÚŕŰýÝţ´­˝˛ˇ˘§÷¸°¨˙űŘř■ "
        },
        "ibm852": "cp852",
        "csibm852": "cp852",
        "cp855": {
            "type": "_sbcs",
            "chars": "ђЂѓЃёЁєЄѕЅіІїЇјЈљЉњЊћЋќЌўЎџЏюЮъЪаАбБцЦдДеЕфФгГ«»░▒▓│┤хХиИ╣║╗╝йЙ┐└┴┬├─┼кК╚╔╩╦╠═╬¤лЛмМнНоОп┘┌█▄Пя▀ЯрРсСтТуУжЖвВьЬ№­ыЫзЗшШэЭщЩчЧ§■ "
        },
        "ibm855": "cp855",
        "csibm855": "cp855",
        "cp856": {
            "type": "_sbcs",
            "chars": "אבגדהוזחטיךכלםמןנסעףפץצקרשת�£�×����������®¬½¼�«»░▒▓│┤���©╣║╗╝¢¥┐└┴┬├─┼��╚╔╩╦╠═╬¤���������┘┌█▄¦�▀������µ�������¯´­±‗¾¶§÷¸°¨·¹³²■ "
        },
        "ibm856": "cp856",
        "csibm856": "cp856",
        "cp857": {
            "type": "_sbcs",
            "chars": "ÇüéâäàåçêëèïîıÄÅÉæÆôöòûùİÖÜø£ØŞşáíóúñÑĞğ¿®¬½¼¡«»░▒▓│┤ÁÂÀ©╣║╗╝¢¥┐└┴┬├─┼ãÃ╚╔╩╦╠═╬¤ºªÊËÈ�ÍÎÏ┘┌█▄¦Ì▀ÓßÔÒõÕµ�×ÚÛÙìÿ¯´­±�¾¶§÷¸°¨·¹³²■ "
        },
        "ibm857": "cp857",
        "csibm857": "cp857",
        "cp858": {
            "type": "_sbcs",
            "chars": "ÇüéâäàåçêëèïîìÄÅÉæÆôöòûùÿÖÜø£Ø×ƒáíóúñÑªº¿®¬½¼¡«»░▒▓│┤ÁÂÀ©╣║╗╝¢¥┐└┴┬├─┼ãÃ╚╔╩╦╠═╬¤ðÐÊËÈ€ÍÎÏ┘┌█▄¦Ì▀ÓßÔÒõÕµþÞÚÛÙýÝ¯´­±‗¾¶§÷¸°¨·¹³²■ "
        },
        "ibm858": "cp858",
        "csibm858": "cp858",
        "cp860": {
            "type": "_sbcs",
            "chars": "ÇüéâãàÁçêÊèÍÔìÃÂÉÀÈôõòÚùÌÕÜ¢£Ù₧ÓáíóúñÑªº¿Ò¬½¼¡«»░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀αßΓπΣσµτΦΘΩδ∞φε∩≡±≥≤⌠⌡÷≈°∙·√ⁿ²■ "
        },
        "ibm860": "cp860",
        "csibm860": "cp860",
        "cp861": {
            "type": "_sbcs",
            "chars": "ÇüéâäàåçêëèÐðÞÄÅÉæÆôöþûÝýÖÜø£Ø₧ƒáíóúÁÍÓÚ¿⌐¬½¼¡«»░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀αßΓπΣσµτΦΘΩδ∞φε∩≡±≥≤⌠⌡÷≈°∙·√ⁿ²■ "
        },
        "ibm861": "cp861",
        "csibm861": "cp861",
        "cp862": {
            "type": "_sbcs",
            "chars": "אבגדהוזחטיךכלםמןנסעףפץצקרשת¢£¥₧ƒáíóúñÑªº¿⌐¬½¼¡«»░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀αßΓπΣσµτΦΘΩδ∞φε∩≡±≥≤⌠⌡÷≈°∙·√ⁿ²■ "
        },
        "ibm862": "cp862",
        "csibm862": "cp862",
        "cp863": {
            "type": "_sbcs",
            "chars": "ÇüéâÂà¶çêëèïî‗À§ÉÈÊôËÏûù¤ÔÜ¢£ÙÛƒ¦´óú¨¸³¯Î⌐¬½¼¾«»░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀αßΓπΣσµτΦΘΩδ∞φε∩≡±≥≤⌠⌡÷≈°∙·√ⁿ²■ "
        },
        "ibm863": "cp863",
        "csibm863": "cp863",
        "cp864": {
            "type": "_sbcs",
            "chars": "\u0000\u0001\u0002\u0003\u0004\u0005\u0006\u0007\b\t\n\u000b\f\r\u000e\u000f\u0010\u0011\u0012\u0013\u0014\u0015\u0016\u0017\u0018\u0019\u001a\u001b\u001c\u001d\u001e\u001f !\"#$٪&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~°·∙√▒─│┼┤┬├┴┐┌└┘β∞φ±½¼≈«»ﻷﻸ��ﻻﻼ� ­ﺂ£¤ﺄ��ﺎﺏﺕﺙ،ﺝﺡﺥ٠١٢٣٤٥٦٧٨٩ﻑ؛ﺱﺵﺹ؟¢ﺀﺁﺃﺅﻊﺋﺍﺑﺓﺗﺛﺟﺣﺧﺩﺫﺭﺯﺳﺷﺻﺿﻁﻅﻋﻏ¦¬÷×ﻉـﻓﻗﻛﻟﻣﻧﻫﻭﻯﻳﺽﻌﻎﻍﻡﹽّﻥﻩﻬﻰﻲﻐﻕﻵﻶﻝﻙﻱ■�"
        },
        "ibm864": "cp864",
        "csibm864": "cp864",
        "cp865": {
            "type": "_sbcs",
            "chars": "ÇüéâäàåçêëèïîìÄÅÉæÆôöòûùÿÖÜø£Ø₧ƒáíóúñÑªº¿⌐¬½¼¡«¤░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀αßΓπΣσµτΦΘΩδ∞φε∩≡±≥≤⌠⌡÷≈°∙·√ⁿ²■ "
        },
        "ibm865": "cp865",
        "csibm865": "cp865",
        "cp866": {
            "type": "_sbcs",
            "chars": "АБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмноп░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀рстуфхцчшщъыьэюяЁёЄєЇїЎў°∙·√№¤■ "
        },
        "ibm866": "cp866",
        "csibm866": "cp866",
        "cp869": {
            "type": "_sbcs",
            "chars": "������Ά�·¬¦‘’Έ―ΉΊΪΌ��ΎΫ©Ώ²³ά£έήίϊΐόύΑΒΓΔΕΖΗ½ΘΙ«»░▒▓│┤ΚΛΜΝ╣║╗╝ΞΟ┐└┴┬├─┼ΠΡ╚╔╩╦╠═╬ΣΤΥΦΧΨΩαβγ┘┌█▄δε▀ζηθικλμνξοπρσςτ΄­±υφχ§ψ΅°¨ωϋΰώ■ "
        },
        "ibm869": "cp869",
        "csibm869": "cp869",
        "cp922": {
            "type": "_sbcs",
            "chars": " ¡¢£¤¥¦§¨©ª«¬­®‾°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏŠÑÒÓÔÕÖ×ØÙÚÛÜÝŽßàáâãäåæçèéêëìíîïšñòóôõö÷øùúûüýžÿ"
        },
        "ibm922": "cp922",
        "csibm922": "cp922",
        "cp1046": {
            "type": "_sbcs",
            "chars": "ﺈ×÷ﹱ■│─┐┌└┘ﹹﹻﹽﹿﹷﺊﻰﻳﻲﻎﻏﻐﻶﻸﻺﻼ ¤ﺋﺑﺗﺛﺟﺣ،­ﺧﺳ٠١٢٣٤٥٦٧٨٩ﺷ؛ﺻﺿﻊ؟ﻋءآأؤإئابةتثجحخدذرزسشصضطﻇعغﻌﺂﺄﺎﻓـفقكلمنهوىيًٌٍَُِّْﻗﻛﻟﻵﻷﻹﻻﻣﻧﻬﻩ�"
        },
        "ibm1046": "cp1046",
        "csibm1046": "cp1046",
        "cp1124": {
            "type": "_sbcs",
            "chars": " ЁЂҐЄЅІЇЈЉЊЋЌ­ЎЏАБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмнопрстуфхцчшщъыьэюя№ёђґєѕіїјљњћќ§ўџ"
        },
        "ibm1124": "cp1124",
        "csibm1124": "cp1124",
        "cp1125": {
            "type": "_sbcs",
            "chars": "АБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмноп░▒▓│┤╡╢╖╕╣║╗╝╜╛┐└┴┬├─┼╞╟╚╔╩╦╠═╬╧╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀рстуфхцчшщъыьэюяЁёҐґЄєІіЇї·√№¤■ "
        },
        "ibm1125": "cp1125",
        "csibm1125": "cp1125",
        "cp1129": {
            "type": "_sbcs",
            "chars": " ¡¢£¤¥¦§œ©ª«¬­®¯°±²³Ÿµ¶·Œ¹º»¼½¾¿ÀÁÂĂÄÅÆÇÈÉÊË̀ÍÎÏĐÑ̉ÓÔƠÖ×ØÙÚÛÜỮßàáâăäåæçèéêë́íîïđṇ̃óôơö÷øùúûüư₫ÿ"
        },
        "ibm1129": "cp1129",
        "csibm1129": "cp1129",
        "cp1133": {
            "type": "_sbcs",
            "chars": " ກຂຄງຈສຊຍດຕຖທນບປຜຝພຟມຢຣລວຫອຮ���ຯະາຳິີຶືຸູຼັົຽ���ເແໂໃໄ່້໊໋໌ໍໆ�ໜໝ₭����������������໐໑໒໓໔໕໖໗໘໙��¢¬¦�"
        },
        "ibm1133": "cp1133",
        "csibm1133": "cp1133",
        "cp1161": {
            "type": "_sbcs",
            "chars": "��������������������������������่กขฃคฅฆงจฉชซฌญฎฏฐฑฒณดตถทธนบปผฝพฟภมยรฤลฦวศษสหฬอฮฯะัาำิีึืฺุู้๊๋€฿เแโใไๅๆ็่้๊๋์ํ๎๏๐๑๒๓๔๕๖๗๘๙๚๛¢¬¦ "
        },
        "ibm1161": "cp1161",
        "csibm1161": "cp1161",
        "cp1162": {
            "type": "_sbcs",
            "chars": "€…‘’“”•–— กขฃคฅฆงจฉชซฌญฎฏฐฑฒณดตถทธนบปผฝพฟภมยรฤลฦวศษสหฬอฮฯะัาำิีึืฺุู����฿เแโใไๅๆ็่้๊๋์ํ๎๏๐๑๒๓๔๕๖๗๘๙๚๛����"
        },
        "ibm1162": "cp1162",
        "csibm1162": "cp1162",
        "cp1163": {
            "type": "_sbcs",
            "chars": " ¡¢£€¥¦§œ©ª«¬­®¯°±²³Ÿµ¶·Œ¹º»¼½¾¿ÀÁÂĂÄÅÆÇÈÉÊË̀ÍÎÏĐÑ̉ÓÔƠÖ×ØÙÚÛÜỮßàáâăäåæçèéêë́íîïđṇ̃óôơö÷øùúûüư₫ÿ"
        },
        "ibm1163": "cp1163",
        "csibm1163": "cp1163",
        "maccroatian": {
            "type": "_sbcs",
            "chars": "ÄÅÇÉÑÖÜáàâäãåçéèêëíìîïñóòôöõúùûü†°¢£§•¶ß®Š™´¨≠ŽØ∞±≤≥∆µ∂∑∏š∫ªºΩžø¿¡¬√ƒ≈Ć«Č… ÀÃÕŒœĐ—“”‘’÷◊�©⁄¤‹›Æ»–·‚„‰ÂćÁčÈÍÎÏÌÓÔđÒÚÛÙıˆ˜¯πË˚¸Êæˇ"
        },
        "maccyrillic": {
            "type": "_sbcs",
            "chars": "АБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯ†°¢£§•¶І®©™Ђђ≠Ѓѓ∞±≤≥іµ∂ЈЄєЇїЉљЊњјЅ¬√ƒ≈∆«»… ЋћЌќѕ–—“”‘’÷„ЎўЏџ№Ёёяабвгдежзийклмнопрстуфхцчшщъыьэю¤"
        },
        "macgreek": {
            "type": "_sbcs",
            "chars": "Ä¹²É³ÖÜ΅àâä΄¨çéèêë£™îï•½‰ôö¦­ùûü†ΓΔΘΛΞΠß®©ΣΪ§≠°·Α±≤≥¥ΒΕΖΗΙΚΜΦΫΨΩάΝ¬ΟΡ≈Τ«»… ΥΧΆΈœ–―“”‘’÷ΉΊΌΎέήίόΏύαβψδεφγηιξκλμνοπώρστθωςχυζϊϋΐΰ�"
        },
        "maciceland": {
            "type": "_sbcs",
            "chars": "ÄÅÇÉÑÖÜáàâäãåçéèêëíìîïñóòôöõúùûüÝ°¢£§•¶ß®©™´¨≠ÆØ∞±≤≥¥µ∂∑∏π∫ªºΩæø¿¡¬√ƒ≈∆«»… ÀÃÕŒœ–—“”‘’÷◊ÿŸ⁄¤ÐðÞþý·‚„‰ÂÊÁËÈÍÎÏÌÓÔ�ÒÚÛÙıˆ˜¯˘˙˚¸˝˛ˇ"
        },
        "macroman": {
            "type": "_sbcs",
            "chars": "ÄÅÇÉÑÖÜáàâäãåçéèêëíìîïñóòôöõúùûü†°¢£§•¶ß®©™´¨≠ÆØ∞±≤≥¥µ∂∑∏π∫ªºΩæø¿¡¬√ƒ≈∆«»… ÀÃÕŒœ–—“”‘’÷◊ÿŸ⁄¤‹›ﬁﬂ‡·‚„‰ÂÊÁËÈÍÎÏÌÓÔ�ÒÚÛÙıˆ˜¯˘˙˚¸˝˛ˇ"
        },
        "macromania": {
            "type": "_sbcs",
            "chars": "ÄÅÇÉÑÖÜáàâäãåçéèêëíìîïñóòôöõúùûü†°¢£§•¶ß®©™´¨≠ĂŞ∞±≤≥¥µ∂∑∏π∫ªºΩăş¿¡¬√ƒ≈∆«»… ÀÃÕŒœ–—“”‘’÷◊ÿŸ⁄¤‹›Ţţ‡·‚„‰ÂÊÁËÈÍÎÏÌÓÔ�ÒÚÛÙıˆ˜¯˘˙˚¸˝˛ˇ"
        },
        "macthai": {
            "type": "_sbcs",
            "chars": "«»…“”�•‘’� กขฃคฅฆงจฉชซฌญฎฏฐฑฒณดตถทธนบปผฝพฟภมยรฤลฦวศษสหฬอฮฯะัาำิีึืฺุู﻿​–—฿เแโใไๅๆ็่้๊๋์ํ™๏๐๑๒๓๔๕๖๗๘๙®©����"
        },
        "macturkish": {
            "type": "_sbcs",
            "chars": "ÄÅÇÉÑÖÜáàâäãåçéèêëíìîïñóòôöõúùûü†°¢£§•¶ß®©™´¨≠ÆØ∞±≤≥¥µ∂∑∏π∫ªºΩæø¿¡¬√ƒ≈∆«»… ÀÃÕŒœ–—“”‘’÷◊ÿŸĞğİıŞş‡·‚„‰ÂÊÁËÈÍÎÏÌÓÔ�ÒÚÛÙ�ˆ˜¯˘˙˚¸˝˛ˇ"
        },
        "macukraine": {
            "type": "_sbcs",
            "chars": "АБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯ†°Ґ£§•¶І®©™Ђђ≠Ѓѓ∞±≤≥іµґЈЄєЇїЉљЊњјЅ¬√ƒ≈∆«»… ЋћЌќѕ–—“”‘’÷„ЎўЏџ№Ёёяабвгдежзийклмнопрстуфхцчшщъыьэю¤"
        },
        "koi8r": {
            "type": "_sbcs",
            "chars": "─│┌┐└┘├┤┬┴┼▀▄█▌▐░▒▓⌠■∙√≈≤≥ ⌡°²·÷═║╒ё╓╔╕╖╗╘╙╚╛╜╝╞╟╠╡Ё╢╣╤╥╦╧╨╩╪╫╬©юабцдефгхийклмнопярстужвьызшэщчъЮАБЦДЕФГХИЙКЛМНОПЯРСТУЖВЬЫЗШЭЩЧЪ"
        },
        "koi8u": {
            "type": "_sbcs",
            "chars": "─│┌┐└┘├┤┬┴┼▀▄█▌▐░▒▓⌠■∙√≈≤≥ ⌡°²·÷═║╒ёє╔ії╗╘╙╚╛ґ╝╞╟╠╡ЁЄ╣ІЇ╦╧╨╩╪Ґ╬©юабцдефгхийклмнопярстужвьызшэщчъЮАБЦДЕФГХИЙКЛМНОПЯРСТУЖВЬЫЗШЭЩЧЪ"
        },
        "koi8ru": {
            "type": "_sbcs",
            "chars": "─│┌┐└┘├┤┬┴┼▀▄█▌▐░▒▓⌠■∙√≈≤≥ ⌡°²·÷═║╒ёє╔ії╗╘╙╚╛ґў╞╟╠╡ЁЄ╣ІЇ╦╧╨╩╪ҐЎ©юабцдефгхийклмнопярстужвьызшэщчъЮАБЦДЕФГХИЙКЛМНОПЯРСТУЖВЬЫЗШЭЩЧЪ"
        },
        "koi8t": {
            "type": "_sbcs",
            "chars": "қғ‚Ғ„…†‡�‰ҳ‹ҲҷҶ�Қ‘’“”•–—�™�›�����ӯӮё¤ӣ¦§���«¬­®�°±²Ё�Ӣ¶·�№�»���©юабцдефгхийклмнопярстужвьызшэщчъЮАБЦДЕФГХИЙКЛМНОПЯРСТУЖВЬЫЗШЭЩЧЪ"
        },
        "armscii8": {
            "type": "_sbcs",
            "chars": " �և։)(»«—.՝,-֊…՜՛՞ԱաԲբԳգԴդԵեԶզԷէԸըԹթԺժԻիԼլԽխԾծԿկՀհՁձՂղՃճՄմՅյՆնՇշՈոՉչՊպՋջՌռՍսՎվՏտՐրՑցՒւՓփՔքՕօՖֆ՚�"
        },
        "rk1048": {
            "type": "_sbcs",
            "chars": "ЂЃ‚ѓ„…†‡€‰Љ‹ЊҚҺЏђ‘’“”•–—�™љ›њқһџ ҰұӘ¤Ө¦§Ё©Ғ«¬­®Ү°±Ііөµ¶·ё№ғ»әҢңүАБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмнопрстуфхцчшщъыьэюя"
        },
        "tcvn": {
            "type": "_sbcs",
            "chars": "\u0000ÚỤ\u0003ỪỬỮ\u0007\b\t\n\u000b\f\r\u000e\u000f\u0010ỨỰỲỶỸÝỴ\u0018\u0019\u001a\u001b\u001c\u001d\u001e\u001f !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~ÀẢÃÁẠẶẬÈẺẼÉẸỆÌỈĨÍỊÒỎÕÓỌỘỜỞỠỚỢÙỦŨ ĂÂÊÔƠƯĐăâêôơưđẶ̀̀̉̃́àảãáạẲằẳẵắẴẮẦẨẪẤỀặầẩẫấậèỂẻẽéẹềểễếệìỉỄẾỒĩíịòỔỏõóọồổỗốộờởỡớợùỖủũúụừửữứựỳỷỹýỵỐ"
        },
        "georgianacademy": {
            "type": "_sbcs",
            "chars": "‚ƒ„…†‡ˆ‰Š‹Œ‘’“”•–—˜™š›œŸ ¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿აბგდევზთიკლმნოპჟრსტუფქღყშჩცძწჭხჯჰჱჲჳჴჵჶçèéêëìíîïðñòóôõö÷øùúûüýþÿ"
        },
        "georgianps": {
            "type": "_sbcs",
            "chars": "‚ƒ„…†‡ˆ‰Š‹Œ‘’“”•–—˜™š›œŸ ¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿აბგდევზჱთიკლმნჲოპჟრსტჳუფქღყშჩცძწჭხჴჯჰჵæçèéêëìíîïðñòóôõö÷øùúûüýþÿ"
        },
        "pt154": {
            "type": "_sbcs",
            "chars": "ҖҒӮғ„…ҶҮҲүҠӢҢҚҺҸҗ‘’“”•–—ҳҷҡӣңқһҹ ЎўЈӨҘҰ§Ё©Ә«¬ӯ®Ҝ°ұІіҙө¶·ё№ә»јҪҫҝАБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмнопрстуфхцчшщъыьэюя"
        },
        "viscii": {
            "type": "_sbcs",
            "chars": "\u0000\u0001Ẳ\u0003\u0004ẴẪ\u0007\b\t\n\u000b\f\r\u000e\u000f\u0010\u0011\u0012\u0013Ỷ\u0015\u0016\u0017\u0018Ỹ\u001a\u001b\u001c\u001dỴ\u001f !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~ẠẮẰẶẤẦẨẬẼẸẾỀỂỄỆỐỒỔỖỘỢỚỜỞỊỎỌỈỦŨỤỲÕắằặấầẩậẽẹếềểễệốồổỗỠƠộờởịỰỨỪỬơớƯÀÁÂÃẢĂẳẵÈÉÊẺÌÍĨỳĐứÒÓÔạỷừửÙÚỹỵÝỡưàáâãảăữẫèéêẻìíĩỉđựòóôõỏọụùúũủýợỮ"
        },
        "iso646cn": {
            "type": "_sbcs",
            "chars": "\u0000\u0001\u0002\u0003\u0004\u0005\u0006\u0007\b\t\n\u000b\f\r\u000e\u000f\u0010\u0011\u0012\u0013\u0014\u0015\u0016\u0017\u0018\u0019\u001a\u001b\u001c\u001d\u001e\u001f !\"#¥%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}‾��������������������������������������������������������������������������������������������������������������������������������"
        },
        "iso646jp": {
            "type": "_sbcs",
            "chars": "\u0000\u0001\u0002\u0003\u0004\u0005\u0006\u0007\b\t\n\u000b\f\r\u000e\u000f\u0010\u0011\u0012\u0013\u0014\u0015\u0016\u0017\u0018\u0019\u001a\u001b\u001c\u001d\u001e\u001f !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[¥]^_`abcdefghijklmnopqrstuvwxyz{|}‾��������������������������������������������������������������������������������������������������������������������������������"
        },
        "hproman8": {
            "type": "_sbcs",
            "chars": " ÀÂÈÊËÎÏ´ˋˆ¨˜ÙÛ₤¯Ýý°ÇçÑñ¡¿¤£¥§ƒ¢âêôûáéóúàèòùäëöüÅîØÆåíøæÄìÖÜÉïßÔÁÃãÐðÍÌÓÒÕõŠšÚŸÿÞþ·µ¶¾—¼½ªº«■»±�"
        },
        "macintosh": {
            "type": "_sbcs",
            "chars": "ÄÅÇÉÑÖÜáàâäãåçéèêëíìîïñóòôöõúùûü†°¢£§•¶ß®©™´¨≠ÆØ∞±≤≥¥µ∂∑∏π∫ªºΩæø¿¡¬√ƒ≈∆«»… ÀÃÕŒœ–—“”‘’÷◊ÿŸ⁄¤‹›ﬁﬂ‡·‚„‰ÂÊÁËÈÍÎÏÌÓÔ�ÒÚÛÙıˆ˜¯˘˙˚¸˝˛ˇ"
        },
        "ascii": {
            "type": "_sbcs",
            "chars": "��������������������������������������������������������������������������������������������������������������������������������"
        },
        "tis620": {
            "type": "_sbcs",
            "chars": "���������������������������������กขฃคฅฆงจฉชซฌญฎฏฐฑฒณดตถทธนบปผฝพฟภมยรฤลฦวศษสหฬอฮฯะัาำิีึืฺุู����฿เแโใไๅๆ็่้๊๋์ํ๎๏๐๑๒๓๔๕๖๗๘๙๚๛����"
        }
    };
},
/* iconv-lite/encodings/dbcs-codec.js */ function _(require, module, exports, __esModule, __esExport) {
    var Buffer = require(85) /* safer-buffer */.Buffer;
    // Multibyte codec. In this scheme, a character is represented by 1 or more bytes.
    // Our codec supports UTF-16 surrogates, extensions for GB18030 and unicode sequences.
    // To save memory and loading time, we read table files only when requested.
    exports._dbcs = DBCSCodec;
    var UNASSIGNED = -1, GB18030_CODE = -2, SEQ_START = -10, NODE_START = -1000, UNASSIGNED_NODE = new Array(0x100), DEF_CHAR = -1;
    for (var i = 0; i < 0x100; i++)
        UNASSIGNED_NODE[i] = UNASSIGNED;
    // Class DBCSCodec reads and initializes mapping tables.
    function DBCSCodec(codecOptions, iconv) {
        this.encodingName = codecOptions.encodingName;
        if (!codecOptions)
            throw new Error("DBCS codec is called without the data.");
        if (!codecOptions.table)
            throw new Error("Encoding '" + this.encodingName + "' has no data.");
        // Load tables.
        var mappingTable = codecOptions.table();
        // Decode tables: MBCS -> Unicode.
        // decodeTables is a trie, encoded as an array of arrays of integers. Internal arrays are trie nodes and all have len = 256.
        // Trie root is decodeTables[0].
        // Values: >=  0 -> unicode character code. can be > 0xFFFF
        //         == UNASSIGNED -> unknown/unassigned sequence.
        //         == GB18030_CODE -> this is the end of a GB18030 4-byte sequence.
        //         <= NODE_START -> index of the next node in our trie to process next byte.
        //         <= SEQ_START  -> index of the start of a character code sequence, in decodeTableSeq.
        this.decodeTables = [];
        this.decodeTables[0] = UNASSIGNED_NODE.slice(0); // Create root node.
        // Sometimes a MBCS char corresponds to a sequence of unicode chars. We store them as arrays of integers here. 
        this.decodeTableSeq = [];
        // Actual mapping tables consist of chunks. Use them to fill up decode tables.
        for (var i = 0; i < mappingTable.length; i++)
            this._addDecodeChunk(mappingTable[i]);
        this.defaultCharUnicode = iconv.defaultCharUnicode;
        // Encode tables: Unicode -> DBCS.
        // `encodeTable` is array mapping from unicode char to encoded char. All its values are integers for performance.
        // Because it can be sparse, it is represented as array of buckets by 256 chars each. Bucket can be null.
        // Values: >=  0 -> it is a normal char. Write the value (if <=256 then 1 byte, if <=65536 then 2 bytes, etc.).
        //         == UNASSIGNED -> no conversion found. Output a default char.
        //         <= SEQ_START  -> it's an index in encodeTableSeq, see below. The character starts a sequence.
        this.encodeTable = [];
        // `encodeTableSeq` is used when a sequence of unicode characters is encoded as a single code. We use a tree of
        // objects where keys correspond to characters in sequence and leafs are the encoded dbcs values. A special DEF_CHAR key
        // means end of sequence (needed when one sequence is a strict subsequence of another).
        // Objects are kept separately from encodeTable to increase performance.
        this.encodeTableSeq = [];
        // Some chars can be decoded, but need not be encoded.
        var skipEncodeChars = {};
        if (codecOptions.encodeSkipVals)
            for (var i = 0; i < codecOptions.encodeSkipVals.length; i++) {
                var val = codecOptions.encodeSkipVals[i];
                if (typeof val === 'number')
                    skipEncodeChars[val] = true;
                else
                    for (var j = val.from; j <= val.to; j++)
                        skipEncodeChars[j] = true;
            }
        // Use decode trie to recursively fill out encode tables.
        this._fillEncodeTable(0, 0, skipEncodeChars);
        // Add more encoding pairs when needed.
        if (codecOptions.encodeAdd) {
            for (var uChar in codecOptions.encodeAdd)
                if (Object.prototype.hasOwnProperty.call(codecOptions.encodeAdd, uChar))
                    this._setEncodeChar(uChar.charCodeAt(0), codecOptions.encodeAdd[uChar]);
        }
        this.defCharSB = this.encodeTable[0][iconv.defaultCharSingleByte.charCodeAt(0)];
        if (this.defCharSB === UNASSIGNED)
            this.defCharSB = this.encodeTable[0]['?'];
        if (this.defCharSB === UNASSIGNED)
            this.defCharSB = "?".charCodeAt(0);
        // Load & create GB18030 tables when needed.
        if (typeof codecOptions.gb18030 === 'function') {
            this.gb18030 = codecOptions.gb18030(); // Load GB18030 ranges.
            // Add GB18030 decode tables.
            var thirdByteNodeIdx = this.decodeTables.length;
            var thirdByteNode = this.decodeTables[thirdByteNodeIdx] = UNASSIGNED_NODE.slice(0);
            var fourthByteNodeIdx = this.decodeTables.length;
            var fourthByteNode = this.decodeTables[fourthByteNodeIdx] = UNASSIGNED_NODE.slice(0);
            for (var i = 0x81; i <= 0xFE; i++) {
                var secondByteNodeIdx = NODE_START - this.decodeTables[0][i];
                var secondByteNode = this.decodeTables[secondByteNodeIdx];
                for (var j = 0x30; j <= 0x39; j++)
                    secondByteNode[j] = NODE_START - thirdByteNodeIdx;
            }
            for (var i = 0x81; i <= 0xFE; i++)
                thirdByteNode[i] = NODE_START - fourthByteNodeIdx;
            for (var i = 0x30; i <= 0x39; i++)
                fourthByteNode[i] = GB18030_CODE;
        }
    }
    DBCSCodec.prototype.encoder = DBCSEncoder;
    DBCSCodec.prototype.decoder = DBCSDecoder;
    // Decoder helpers
    DBCSCodec.prototype._getDecodeTrieNode = function (addr) {
        var bytes = [];
        for (; addr > 0; addr >>= 8)
            bytes.push(addr & 0xFF);
        if (bytes.length == 0)
            bytes.push(0);
        var node = this.decodeTables[0];
        for (var i = bytes.length - 1; i > 0; i--) { // Traverse nodes deeper into the trie.
            var val = node[bytes[i]];
            if (val == UNASSIGNED) { // Create new node.
                node[bytes[i]] = NODE_START - this.decodeTables.length;
                this.decodeTables.push(node = UNASSIGNED_NODE.slice(0));
            }
            else if (val <= NODE_START) { // Existing node.
                node = this.decodeTables[NODE_START - val];
            }
            else
                throw new Error("Overwrite byte in " + this.encodingName + ", addr: " + addr.toString(16));
        }
        return node;
    };
    DBCSCodec.prototype._addDecodeChunk = function (chunk) {
        // First element of chunk is the hex mbcs code where we start.
        var curAddr = parseInt(chunk[0], 16);
        // Choose the decoding node where we'll write our chars.
        var writeTable = this._getDecodeTrieNode(curAddr);
        curAddr = curAddr & 0xFF;
        // Write all other elements of the chunk to the table.
        for (var k = 1; k < chunk.length; k++) {
            var part = chunk[k];
            if (typeof part === "string") { // String, write as-is.
                for (var l = 0; l < part.length;) {
                    var code = part.charCodeAt(l++);
                    if (0xD800 <= code && code < 0xDC00) { // Decode surrogate
                        var codeTrail = part.charCodeAt(l++);
                        if (0xDC00 <= codeTrail && codeTrail < 0xE000)
                            writeTable[curAddr++] = 0x10000 + (code - 0xD800) * 0x400 + (codeTrail - 0xDC00);
                        else
                            throw new Error("Incorrect surrogate pair in " + this.encodingName + " at chunk " + chunk[0]);
                    }
                    else if (0x0FF0 < code && code <= 0x0FFF) { // Character sequence (our own encoding used)
                        var len = 0xFFF - code + 2;
                        var seq = [];
                        for (var m = 0; m < len; m++)
                            seq.push(part.charCodeAt(l++)); // Simple variation: don't support surrogates or subsequences in seq.
                        writeTable[curAddr++] = SEQ_START - this.decodeTableSeq.length;
                        this.decodeTableSeq.push(seq);
                    }
                    else
                        writeTable[curAddr++] = code; // Basic char
                }
            }
            else if (typeof part === "number") { // Integer, meaning increasing sequence starting with prev character.
                var charCode = writeTable[curAddr - 1] + 1;
                for (var l = 0; l < part; l++)
                    writeTable[curAddr++] = charCode++;
            }
            else
                throw new Error("Incorrect type '" + typeof part + "' given in " + this.encodingName + " at chunk " + chunk[0]);
        }
        if (curAddr > 0xFF)
            throw new Error("Incorrect chunk in " + this.encodingName + " at addr " + chunk[0] + ": too long" + curAddr);
    };
    // Encoder helpers
    DBCSCodec.prototype._getEncodeBucket = function (uCode) {
        var high = uCode >> 8; // This could be > 0xFF because of astral characters.
        if (this.encodeTable[high] === undefined)
            this.encodeTable[high] = UNASSIGNED_NODE.slice(0); // Create bucket on demand.
        return this.encodeTable[high];
    };
    DBCSCodec.prototype._setEncodeChar = function (uCode, dbcsCode) {
        var bucket = this._getEncodeBucket(uCode);
        var low = uCode & 0xFF;
        if (bucket[low] <= SEQ_START)
            this.encodeTableSeq[SEQ_START - bucket[low]][DEF_CHAR] = dbcsCode; // There's already a sequence, set a single-char subsequence of it.
        else if (bucket[low] == UNASSIGNED)
            bucket[low] = dbcsCode;
    };
    DBCSCodec.prototype._setEncodeSequence = function (seq, dbcsCode) {
        // Get the root of character tree according to first character of the sequence.
        var uCode = seq[0];
        var bucket = this._getEncodeBucket(uCode);
        var low = uCode & 0xFF;
        var node;
        if (bucket[low] <= SEQ_START) {
            // There's already a sequence with  - use it.
            node = this.encodeTableSeq[SEQ_START - bucket[low]];
        }
        else {
            // There was no sequence object - allocate a new one.
            node = {};
            if (bucket[low] !== UNASSIGNED)
                node[DEF_CHAR] = bucket[low]; // If a char was set before - make it a single-char subsequence.
            bucket[low] = SEQ_START - this.encodeTableSeq.length;
            this.encodeTableSeq.push(node);
        }
        // Traverse the character tree, allocating new nodes as needed.
        for (var j = 1; j < seq.length - 1; j++) {
            var oldVal = node[uCode];
            if (typeof oldVal === 'object')
                node = oldVal;
            else {
                node = node[uCode] = {};
                if (oldVal !== undefined)
                    node[DEF_CHAR] = oldVal;
            }
        }
        // Set the leaf to given dbcsCode.
        uCode = seq[seq.length - 1];
        node[uCode] = dbcsCode;
    };
    DBCSCodec.prototype._fillEncodeTable = function (nodeIdx, prefix, skipEncodeChars) {
        var node = this.decodeTables[nodeIdx];
        for (var i = 0; i < 0x100; i++) {
            var uCode = node[i];
            var mbCode = prefix + i;
            if (skipEncodeChars[mbCode])
                continue;
            if (uCode >= 0)
                this._setEncodeChar(uCode, mbCode);
            else if (uCode <= NODE_START)
                this._fillEncodeTable(NODE_START - uCode, mbCode << 8, skipEncodeChars);
            else if (uCode <= SEQ_START)
                this._setEncodeSequence(this.decodeTableSeq[SEQ_START - uCode], mbCode);
        }
    };
    // == Encoder ==================================================================
    function DBCSEncoder(options, codec) {
        // Encoder state
        this.leadSurrogate = -1;
        this.seqObj = undefined;
        // Static data
        this.encodeTable = codec.encodeTable;
        this.encodeTableSeq = codec.encodeTableSeq;
        this.defaultCharSingleByte = codec.defCharSB;
        this.gb18030 = codec.gb18030;
    }
    DBCSEncoder.prototype.write = function (str) {
        var newBuf = Buffer.alloc(str.length * (this.gb18030 ? 4 : 3)), leadSurrogate = this.leadSurrogate, seqObj = this.seqObj, nextChar = -1, i = 0, j = 0;
        while (true) {
            // 0. Get next character.
            if (nextChar === -1) {
                if (i == str.length)
                    break;
                var uCode = str.charCodeAt(i++);
            }
            else {
                var uCode = nextChar;
                nextChar = -1;
            }
            // 1. Handle surrogates.
            if (0xD800 <= uCode && uCode < 0xE000) { // Char is one of surrogates.
                if (uCode < 0xDC00) { // We've got lead surrogate.
                    if (leadSurrogate === -1) {
                        leadSurrogate = uCode;
                        continue;
                    }
                    else {
                        leadSurrogate = uCode;
                        // Double lead surrogate found.
                        uCode = UNASSIGNED;
                    }
                }
                else { // We've got trail surrogate.
                    if (leadSurrogate !== -1) {
                        uCode = 0x10000 + (leadSurrogate - 0xD800) * 0x400 + (uCode - 0xDC00);
                        leadSurrogate = -1;
                    }
                    else {
                        // Incomplete surrogate pair - only trail surrogate found.
                        uCode = UNASSIGNED;
                    }
                }
            }
            else if (leadSurrogate !== -1) {
                // Incomplete surrogate pair - only lead surrogate found.
                nextChar = uCode;
                uCode = UNASSIGNED; // Write an error, then current char.
                leadSurrogate = -1;
            }
            // 2. Convert uCode character.
            var dbcsCode = UNASSIGNED;
            if (seqObj !== undefined && uCode != UNASSIGNED) { // We are in the middle of the sequence
                var resCode = seqObj[uCode];
                if (typeof resCode === 'object') { // Sequence continues.
                    seqObj = resCode;
                    continue;
                }
                else if (typeof resCode == 'number') { // Sequence finished. Write it.
                    dbcsCode = resCode;
                }
                else if (resCode == undefined) { // Current character is not part of the sequence.
                    // Try default character for this sequence
                    resCode = seqObj[DEF_CHAR];
                    if (resCode !== undefined) {
                        dbcsCode = resCode; // Found. Write it.
                        nextChar = uCode; // Current character will be written too in the next iteration.
                    }
                    else {
                        // TODO: What if we have no default? (resCode == undefined)
                        // Then, we should write first char of the sequence as-is and try the rest recursively.
                        // Didn't do it for now because no encoding has this situation yet.
                        // Currently, just skip the sequence and write current char.
                    }
                }
                seqObj = undefined;
            }
            else if (uCode >= 0) { // Regular character
                var subtable = this.encodeTable[uCode >> 8];
                if (subtable !== undefined)
                    dbcsCode = subtable[uCode & 0xFF];
                if (dbcsCode <= SEQ_START) { // Sequence start
                    seqObj = this.encodeTableSeq[SEQ_START - dbcsCode];
                    continue;
                }
                if (dbcsCode == UNASSIGNED && this.gb18030) {
                    // Use GB18030 algorithm to find character(s) to write.
                    var idx = findIdx(this.gb18030.uChars, uCode);
                    if (idx != -1) {
                        var dbcsCode = this.gb18030.gbChars[idx] + (uCode - this.gb18030.uChars[idx]);
                        newBuf[j++] = 0x81 + Math.floor(dbcsCode / 12600);
                        dbcsCode = dbcsCode % 12600;
                        newBuf[j++] = 0x30 + Math.floor(dbcsCode / 1260);
                        dbcsCode = dbcsCode % 1260;
                        newBuf[j++] = 0x81 + Math.floor(dbcsCode / 10);
                        dbcsCode = dbcsCode % 10;
                        newBuf[j++] = 0x30 + dbcsCode;
                        continue;
                    }
                }
            }
            // 3. Write dbcsCode character.
            if (dbcsCode === UNASSIGNED)
                dbcsCode = this.defaultCharSingleByte;
            if (dbcsCode < 0x100) {
                newBuf[j++] = dbcsCode;
            }
            else if (dbcsCode < 0x10000) {
                newBuf[j++] = dbcsCode >> 8; // high byte
                newBuf[j++] = dbcsCode & 0xFF; // low byte
            }
            else {
                newBuf[j++] = dbcsCode >> 16;
                newBuf[j++] = (dbcsCode >> 8) & 0xFF;
                newBuf[j++] = dbcsCode & 0xFF;
            }
        }
        this.seqObj = seqObj;
        this.leadSurrogate = leadSurrogate;
        return newBuf.slice(0, j);
    };
    DBCSEncoder.prototype.end = function () {
        if (this.leadSurrogate === -1 && this.seqObj === undefined)
            return; // All clean. Most often case.
        var newBuf = Buffer.alloc(10), j = 0;
        if (this.seqObj) { // We're in the sequence.
            var dbcsCode = this.seqObj[DEF_CHAR];
            if (dbcsCode !== undefined) { // Write beginning of the sequence.
                if (dbcsCode < 0x100) {
                    newBuf[j++] = dbcsCode;
                }
                else {
                    newBuf[j++] = dbcsCode >> 8; // high byte
                    newBuf[j++] = dbcsCode & 0xFF; // low byte
                }
            }
            else {
                // See todo above.
            }
            this.seqObj = undefined;
        }
        if (this.leadSurrogate !== -1) {
            // Incomplete surrogate pair - only lead surrogate found.
            newBuf[j++] = this.defaultCharSingleByte;
            this.leadSurrogate = -1;
        }
        return newBuf.slice(0, j);
    };
    // Export for testing
    DBCSEncoder.prototype.findIdx = findIdx;
    // == Decoder ==================================================================
    function DBCSDecoder(options, codec) {
        // Decoder state
        this.nodeIdx = 0;
        this.prevBuf = Buffer.alloc(0);
        // Static data
        this.decodeTables = codec.decodeTables;
        this.decodeTableSeq = codec.decodeTableSeq;
        this.defaultCharUnicode = codec.defaultCharUnicode;
        this.gb18030 = codec.gb18030;
    }
    DBCSDecoder.prototype.write = function (buf) {
        var newBuf = Buffer.alloc(buf.length * 2), nodeIdx = this.nodeIdx, prevBuf = this.prevBuf, prevBufOffset = this.prevBuf.length, seqStart = -this.prevBuf.length, // idx of the start of current parsed sequence.
        uCode;
        if (prevBufOffset > 0) // Make prev buf overlap a little to make it easier to slice later.
            prevBuf = Buffer.concat([prevBuf, buf.slice(0, 10)]);
        for (var i = 0, j = 0; i < buf.length; i++) {
            var curByte = (i >= 0) ? buf[i] : prevBuf[i + prevBufOffset];
            // Lookup in current trie node.
            var uCode = this.decodeTables[nodeIdx][curByte];
            if (uCode >= 0) {
                // Normal character, just use it.
            }
            else if (uCode === UNASSIGNED) { // Unknown char.
                // TODO: Callback with seq.
                //var curSeq = (seqStart >= 0) ? buf.slice(seqStart, i+1) : prevBuf.slice(seqStart + prevBufOffset, i+1 + prevBufOffset);
                i = seqStart; // Try to parse again, after skipping first byte of the sequence ('i' will be incremented by 'for' cycle).
                uCode = this.defaultCharUnicode.charCodeAt(0);
            }
            else if (uCode === GB18030_CODE) {
                var curSeq = (seqStart >= 0) ? buf.slice(seqStart, i + 1) : prevBuf.slice(seqStart + prevBufOffset, i + 1 + prevBufOffset);
                var ptr = (curSeq[0] - 0x81) * 12600 + (curSeq[1] - 0x30) * 1260 + (curSeq[2] - 0x81) * 10 + (curSeq[3] - 0x30);
                var idx = findIdx(this.gb18030.gbChars, ptr);
                uCode = this.gb18030.uChars[idx] + ptr - this.gb18030.gbChars[idx];
            }
            else if (uCode <= NODE_START) { // Go to next trie node.
                nodeIdx = NODE_START - uCode;
                continue;
            }
            else if (uCode <= SEQ_START) { // Output a sequence of chars.
                var seq = this.decodeTableSeq[SEQ_START - uCode];
                for (var k = 0; k < seq.length - 1; k++) {
                    uCode = seq[k];
                    newBuf[j++] = uCode & 0xFF;
                    newBuf[j++] = uCode >> 8;
                }
                uCode = seq[seq.length - 1];
            }
            else
                throw new Error("iconv-lite internal error: invalid decoding table value " + uCode + " at " + nodeIdx + "/" + curByte);
            // Write the character to buffer, handling higher planes using surrogate pair.
            if (uCode > 0xFFFF) {
                uCode -= 0x10000;
                var uCodeLead = 0xD800 + Math.floor(uCode / 0x400);
                newBuf[j++] = uCodeLead & 0xFF;
                newBuf[j++] = uCodeLead >> 8;
                uCode = 0xDC00 + uCode % 0x400;
            }
            newBuf[j++] = uCode & 0xFF;
            newBuf[j++] = uCode >> 8;
            // Reset trie node.
            nodeIdx = 0;
            seqStart = i + 1;
        }
        this.nodeIdx = nodeIdx;
        this.prevBuf = (seqStart >= 0) ? buf.slice(seqStart) : prevBuf.slice(seqStart + prevBufOffset);
        return newBuf.slice(0, j).toString('ucs2');
    };
    DBCSDecoder.prototype.end = function () {
        var ret = '';
        // Try to parse all remaining chars.
        while (this.prevBuf.length > 0) {
            // Skip 1 character in the buffer.
            ret += this.defaultCharUnicode;
            var buf = this.prevBuf.slice(1);
            // Parse remaining as usual.
            this.prevBuf = Buffer.alloc(0);
            this.nodeIdx = 0;
            if (buf.length > 0)
                ret += this.write(buf);
        }
        this.nodeIdx = 0;
        return ret;
    };
    // Binary search for GB18030. Returns largest i such that table[i] <= val.
    function findIdx(table, val) {
        if (table[0] > val)
            return -1;
        var l = 0, r = table.length;
        while (l < r - 1) { // always table[l] <= val < table[r]
            var mid = l + Math.floor((r - l + 1) / 2);
            if (table[mid] <= val)
                l = mid;
            else
                r = mid;
        }
        return l;
    }
},
/* iconv-lite/encodings/dbcs-data.js */ function _(require, module, exports, __esModule, __esExport) {
    // Description of supported double byte encodings and aliases.
    // Tables are not require()-d until they are needed to speed up library load.
    // require()-s are direct to support Browserify.
    module.exports = {
        // == Japanese/ShiftJIS ====================================================
        // All japanese encodings are based on JIS X set of standards:
        // JIS X 0201 - Single-byte encoding of ASCII + ¥ + Kana chars at 0xA1-0xDF.
        // JIS X 0208 - Main set of 6879 characters, placed in 94x94 plane, to be encoded by 2 bytes. 
        //              Has several variations in 1978, 1983, 1990 and 1997.
        // JIS X 0212 - Supplementary plane of 6067 chars in 94x94 plane. 1990. Effectively dead.
        // JIS X 0213 - Extension and modern replacement of 0208 and 0212. Total chars: 11233.
        //              2 planes, first is superset of 0208, second - revised 0212.
        //              Introduced in 2000, revised 2004. Some characters are in Unicode Plane 2 (0x2xxxx)
        // Byte encodings are:
        //  * Shift_JIS: Compatible with 0201, uses not defined chars in top half as lead bytes for double-byte
        //               encoding of 0208. Lead byte ranges: 0x81-0x9F, 0xE0-0xEF; Trail byte ranges: 0x40-0x7E, 0x80-0x9E, 0x9F-0xFC.
        //               Windows CP932 is a superset of Shift_JIS. Some companies added more chars, notably KDDI.
        //  * EUC-JP:    Up to 3 bytes per character. Used mostly on *nixes.
        //               0x00-0x7F       - lower part of 0201
        //               0x8E, 0xA1-0xDF - upper part of 0201
        //               (0xA1-0xFE)x2   - 0208 plane (94x94).
        //               0x8F, (0xA1-0xFE)x2 - 0212 plane (94x94).
        //  * JIS X 208: 7-bit, direct encoding of 0208. Byte ranges: 0x21-0x7E (94 values). Uncommon.
        //               Used as-is in ISO2022 family.
        //  * ISO2022-JP: Stateful encoding, with escape sequences to switch between ASCII, 
        //                0201-1976 Roman, 0208-1978, 0208-1983.
        //  * ISO2022-JP-1: Adds esc seq for 0212-1990.
        //  * ISO2022-JP-2: Adds esc seq for GB2313-1980, KSX1001-1992, ISO8859-1, ISO8859-7.
        //  * ISO2022-JP-3: Adds esc seq for 0201-1976 Kana set, 0213-2000 Planes 1, 2.
        //  * ISO2022-JP-2004: Adds 0213-2004 Plane 1.
        //
        // After JIS X 0213 appeared, Shift_JIS-2004, EUC-JISX0213 and ISO2022-JP-2004 followed, with just changing the planes.
        //
        // Overall, it seems that it's a mess :( http://www8.plala.or.jp/tkubota1/unicode-symbols-map2.html
        'shiftjis': {
            type: '_dbcs',
            table: function () { return require(96) /* ./tables/shiftjis.json */; },
            encodeAdd: { '\u00a5': 0x5C, '\u203E': 0x7E },
            encodeSkipVals: [{ from: 0xED40, to: 0xF940 }],
        },
        'csshiftjis': 'shiftjis',
        'mskanji': 'shiftjis',
        'sjis': 'shiftjis',
        'windows31j': 'shiftjis',
        'ms31j': 'shiftjis',
        'xsjis': 'shiftjis',
        'windows932': 'shiftjis',
        'ms932': 'shiftjis',
        '932': 'shiftjis',
        'cp932': 'shiftjis',
        'eucjp': {
            type: '_dbcs',
            table: function () { return require(97) /* ./tables/eucjp.json */; },
            encodeAdd: { '\u00a5': 0x5C, '\u203E': 0x7E },
        },
        // TODO: KDDI extension to Shift_JIS
        // TODO: IBM CCSID 942 = CP932, but F0-F9 custom chars and other char changes.
        // TODO: IBM CCSID 943 = Shift_JIS = CP932 with original Shift_JIS lower 128 chars.
        // == Chinese/GBK ==========================================================
        // http://en.wikipedia.org/wiki/GBK
        // We mostly implement W3C recommendation: https://www.w3.org/TR/encoding/#gbk-encoder
        // Oldest GB2312 (1981, ~7600 chars) is a subset of CP936
        'gb2312': 'cp936',
        'gb231280': 'cp936',
        'gb23121980': 'cp936',
        'csgb2312': 'cp936',
        'csiso58gb231280': 'cp936',
        'euccn': 'cp936',
        // Microsoft's CP936 is a subset and approximation of GBK.
        'windows936': 'cp936',
        'ms936': 'cp936',
        '936': 'cp936',
        'cp936': {
            type: '_dbcs',
            table: function () { return require(98) /* ./tables/cp936.json */; },
        },
        // GBK (~22000 chars) is an extension of CP936 that added user-mapped chars and some other.
        'gbk': {
            type: '_dbcs',
            table: function () { return require(98) /* ./tables/cp936.json */.concat(require(99) /* ./tables/gbk-added.json */); },
        },
        'xgbk': 'gbk',
        'isoir58': 'gbk',
        // GB18030 is an algorithmic extension of GBK.
        // Main source: https://www.w3.org/TR/encoding/#gbk-encoder
        // http://icu-project.org/docs/papers/gb18030.html
        // http://source.icu-project.org/repos/icu/data/trunk/charset/data/xml/gb-18030-2000.xml
        // http://www.khngai.com/chinese/charmap/tblgbk.php?page=0
        'gb18030': {
            type: '_dbcs',
            table: function () { return require(98) /* ./tables/cp936.json */.concat(require(99) /* ./tables/gbk-added.json */); },
            gb18030: function () { return require(100) /* ./tables/gb18030-ranges.json */; },
            encodeSkipVals: [0x80],
            encodeAdd: { '€': 0xA2E3 },
        },
        'chinese': 'gb18030',
        // == Korean ===============================================================
        // EUC-KR, KS_C_5601 and KS X 1001 are exactly the same.
        'windows949': 'cp949',
        'ms949': 'cp949',
        '949': 'cp949',
        'cp949': {
            type: '_dbcs',
            table: function () { return require(101) /* ./tables/cp949.json */; },
        },
        'cseuckr': 'cp949',
        'csksc56011987': 'cp949',
        'euckr': 'cp949',
        'isoir149': 'cp949',
        'korean': 'cp949',
        'ksc56011987': 'cp949',
        'ksc56011989': 'cp949',
        'ksc5601': 'cp949',
        // == Big5/Taiwan/Hong Kong ================================================
        // There are lots of tables for Big5 and cp950. Please see the following links for history:
        // http://moztw.org/docs/big5/  http://www.haible.de/bruno/charsets/conversion-tables/Big5.html
        // Variations, in roughly number of defined chars:
        //  * Windows CP 950: Microsoft variant of Big5. Canonical: http://www.unicode.org/Public/MAPPINGS/VENDORS/MICSFT/WINDOWS/CP950.TXT
        //  * Windows CP 951: Microsoft variant of Big5-HKSCS-2001. Seems to be never public. http://me.abelcheung.org/articles/research/what-is-cp951/
        //  * Big5-2003 (Taiwan standard) almost superset of cp950.
        //  * Unicode-at-on (UAO) / Mozilla 1.8. Falling out of use on the Web. Not supported by other browsers.
        //  * Big5-HKSCS (-2001, -2004, -2008). Hong Kong standard. 
        //    many unicode code points moved from PUA to Supplementary plane (U+2XXXX) over the years.
        //    Plus, it has 4 combining sequences.
        //    Seems that Mozilla refused to support it for 10 yrs. https://bugzilla.mozilla.org/show_bug.cgi?id=162431 https://bugzilla.mozilla.org/show_bug.cgi?id=310299
        //    because big5-hkscs is the only encoding to include astral characters in non-algorithmic way.
        //    Implementations are not consistent within browsers; sometimes labeled as just big5.
        //    MS Internet Explorer switches from big5 to big5-hkscs when a patch applied.
        //    Great discussion & recap of what's going on https://bugzilla.mozilla.org/show_bug.cgi?id=912470#c31
        //    In the encoder, it might make sense to support encoding old PUA mappings to Big5 bytes seq-s.
        //    Official spec: http://www.ogcio.gov.hk/en/business/tech_promotion/ccli/terms/doc/2003cmp_2008.txt
        //                   http://www.ogcio.gov.hk/tc/business/tech_promotion/ccli/terms/doc/hkscs-2008-big5-iso.txt
        // 
        // Current understanding of how to deal with Big5(-HKSCS) is in the Encoding Standard, http://encoding.spec.whatwg.org/#big5-encoder
        // Unicode mapping (http://www.unicode.org/Public/MAPPINGS/OBSOLETE/EASTASIA/OTHER/BIG5.TXT) is said to be wrong.
        'windows950': 'cp950',
        'ms950': 'cp950',
        '950': 'cp950',
        'cp950': {
            type: '_dbcs',
            table: function () { return require(102) /* ./tables/cp950.json */; },
        },
        // Big5 has many variations and is an extension of cp950. We use Encoding Standard's as a consensus.
        'big5': 'big5hkscs',
        'big5hkscs': {
            type: '_dbcs',
            table: function () { return require(102) /* ./tables/cp950.json */.concat(require(103) /* ./tables/big5-added.json */); },
            encodeSkipVals: [0xa2cc],
        },
        'cnbig5': 'big5hkscs',
        'csbig5': 'big5hkscs',
        'xxbig5': 'big5hkscs',
    };
},
/* iconv-lite/encodings/tables/shiftjis.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = [
        ["0", "\u0000", 128],
        ["a1", "｡", 62],
        ["8140", "　、。，．・：；？！゛゜´｀¨＾￣＿ヽヾゝゞ〃仝々〆〇ー―‐／＼～∥｜…‥‘’“”（）〔〕［］｛｝〈", 9, "＋－±×"],
        ["8180", "÷＝≠＜＞≦≧∞∴♂♀°′″℃￥＄￠￡％＃＆＊＠§☆★○●◎◇◆□■△▲▽▼※〒→←↑↓〓"],
        ["81b8", "∈∋⊆⊇⊂⊃∪∩"],
        ["81c8", "∧∨￢⇒⇔∀∃"],
        ["81da", "∠⊥⌒∂∇≡≒≪≫√∽∝∵∫∬"],
        ["81f0", "Å‰♯♭♪†‡¶"],
        ["81fc", "◯"],
        ["824f", "０", 9],
        ["8260", "Ａ", 25],
        ["8281", "ａ", 25],
        ["829f", "ぁ", 82],
        ["8340", "ァ", 62],
        ["8380", "ム", 22],
        ["839f", "Α", 16, "Σ", 6],
        ["83bf", "α", 16, "σ", 6],
        ["8440", "А", 5, "ЁЖ", 25],
        ["8470", "а", 5, "ёж", 7],
        ["8480", "о", 17],
        ["849f", "─│┌┐┘└├┬┤┴┼━┃┏┓┛┗┣┳┫┻╋┠┯┨┷┿┝┰┥┸╂"],
        ["8740", "①", 19, "Ⅰ", 9],
        ["875f", "㍉㌔㌢㍍㌘㌧㌃㌶㍑㍗㌍㌦㌣㌫㍊㌻㎜㎝㎞㎎㎏㏄㎡"],
        ["877e", "㍻"],
        ["8780", "〝〟№㏍℡㊤", 4, "㈱㈲㈹㍾㍽㍼≒≡∫∮∑√⊥∠∟⊿∵∩∪"],
        ["889f", "亜唖娃阿哀愛挨姶逢葵茜穐悪握渥旭葦芦鯵梓圧斡扱宛姐虻飴絢綾鮎或粟袷安庵按暗案闇鞍杏以伊位依偉囲夷委威尉惟意慰易椅為畏異移維緯胃萎衣謂違遺医井亥域育郁磯一壱溢逸稲茨芋鰯允印咽員因姻引飲淫胤蔭"],
        ["8940", "院陰隠韻吋右宇烏羽迂雨卯鵜窺丑碓臼渦嘘唄欝蔚鰻姥厩浦瓜閏噂云運雲荏餌叡営嬰影映曳栄永泳洩瑛盈穎頴英衛詠鋭液疫益駅悦謁越閲榎厭円"],
        ["8980", "園堰奄宴延怨掩援沿演炎焔煙燕猿縁艶苑薗遠鉛鴛塩於汚甥凹央奥往応押旺横欧殴王翁襖鴬鴎黄岡沖荻億屋憶臆桶牡乙俺卸恩温穏音下化仮何伽価佳加可嘉夏嫁家寡科暇果架歌河火珂禍禾稼箇花苛茄荷華菓蝦課嘩貨迦過霞蚊俄峨我牙画臥芽蛾賀雅餓駕介会解回塊壊廻快怪悔恢懐戒拐改"],
        ["8a40", "魁晦械海灰界皆絵芥蟹開階貝凱劾外咳害崖慨概涯碍蓋街該鎧骸浬馨蛙垣柿蛎鈎劃嚇各廓拡撹格核殻獲確穫覚角赫較郭閣隔革学岳楽額顎掛笠樫"],
        ["8a80", "橿梶鰍潟割喝恰括活渇滑葛褐轄且鰹叶椛樺鞄株兜竃蒲釜鎌噛鴨栢茅萱粥刈苅瓦乾侃冠寒刊勘勧巻喚堪姦完官寛干幹患感慣憾換敢柑桓棺款歓汗漢澗潅環甘監看竿管簡緩缶翰肝艦莞観諌貫還鑑間閑関陥韓館舘丸含岸巌玩癌眼岩翫贋雁頑顔願企伎危喜器基奇嬉寄岐希幾忌揮机旗既期棋棄"],
        ["8b40", "機帰毅気汽畿祈季稀紀徽規記貴起軌輝飢騎鬼亀偽儀妓宜戯技擬欺犠疑祇義蟻誼議掬菊鞠吉吃喫桔橘詰砧杵黍却客脚虐逆丘久仇休及吸宮弓急救"],
        ["8b80", "朽求汲泣灸球究窮笈級糾給旧牛去居巨拒拠挙渠虚許距鋸漁禦魚亨享京供侠僑兇競共凶協匡卿叫喬境峡強彊怯恐恭挟教橋況狂狭矯胸脅興蕎郷鏡響饗驚仰凝尭暁業局曲極玉桐粁僅勤均巾錦斤欣欽琴禁禽筋緊芹菌衿襟謹近金吟銀九倶句区狗玖矩苦躯駆駈駒具愚虞喰空偶寓遇隅串櫛釧屑屈"],
        ["8c40", "掘窟沓靴轡窪熊隈粂栗繰桑鍬勲君薫訓群軍郡卦袈祁係傾刑兄啓圭珪型契形径恵慶慧憩掲携敬景桂渓畦稽系経継繋罫茎荊蛍計詣警軽頚鶏芸迎鯨"],
        ["8c80", "劇戟撃激隙桁傑欠決潔穴結血訣月件倹倦健兼券剣喧圏堅嫌建憲懸拳捲検権牽犬献研硯絹県肩見謙賢軒遣鍵険顕験鹸元原厳幻弦減源玄現絃舷言諺限乎個古呼固姑孤己庫弧戸故枯湖狐糊袴股胡菰虎誇跨鈷雇顧鼓五互伍午呉吾娯後御悟梧檎瑚碁語誤護醐乞鯉交佼侯候倖光公功効勾厚口向"],
        ["8d40", "后喉坑垢好孔孝宏工巧巷幸広庚康弘恒慌抗拘控攻昂晃更杭校梗構江洪浩港溝甲皇硬稿糠紅紘絞綱耕考肯肱腔膏航荒行衡講貢購郊酵鉱砿鋼閤降"],
        ["8d80", "項香高鴻剛劫号合壕拷濠豪轟麹克刻告国穀酷鵠黒獄漉腰甑忽惚骨狛込此頃今困坤墾婚恨懇昏昆根梱混痕紺艮魂些佐叉唆嵯左差査沙瑳砂詐鎖裟坐座挫債催再最哉塞妻宰彩才採栽歳済災采犀砕砦祭斎細菜裁載際剤在材罪財冴坂阪堺榊肴咲崎埼碕鷺作削咋搾昨朔柵窄策索錯桜鮭笹匙冊刷"],
        ["8e40", "察拶撮擦札殺薩雑皐鯖捌錆鮫皿晒三傘参山惨撒散桟燦珊産算纂蚕讃賛酸餐斬暫残仕仔伺使刺司史嗣四士始姉姿子屍市師志思指支孜斯施旨枝止"],
        ["8e80", "死氏獅祉私糸紙紫肢脂至視詞詩試誌諮資賜雌飼歯事似侍児字寺慈持時次滋治爾璽痔磁示而耳自蒔辞汐鹿式識鴫竺軸宍雫七叱執失嫉室悉湿漆疾質実蔀篠偲柴芝屡蕊縞舎写射捨赦斜煮社紗者謝車遮蛇邪借勺尺杓灼爵酌釈錫若寂弱惹主取守手朱殊狩珠種腫趣酒首儒受呪寿授樹綬需囚収周"],
        ["8f40", "宗就州修愁拾洲秀秋終繍習臭舟蒐衆襲讐蹴輯週酋酬集醜什住充十従戎柔汁渋獣縦重銃叔夙宿淑祝縮粛塾熟出術述俊峻春瞬竣舜駿准循旬楯殉淳"],
        ["8f80", "準潤盾純巡遵醇順処初所暑曙渚庶緒署書薯藷諸助叙女序徐恕鋤除傷償勝匠升召哨商唱嘗奨妾娼宵将小少尚庄床廠彰承抄招掌捷昇昌昭晶松梢樟樵沼消渉湘焼焦照症省硝礁祥称章笑粧紹肖菖蒋蕉衝裳訟証詔詳象賞醤鉦鍾鐘障鞘上丈丞乗冗剰城場壌嬢常情擾条杖浄状畳穣蒸譲醸錠嘱埴飾"],
        ["9040", "拭植殖燭織職色触食蝕辱尻伸信侵唇娠寝審心慎振新晋森榛浸深申疹真神秦紳臣芯薪親診身辛進針震人仁刃塵壬尋甚尽腎訊迅陣靭笥諏須酢図厨"],
        ["9080", "逗吹垂帥推水炊睡粋翠衰遂酔錐錘随瑞髄崇嵩数枢趨雛据杉椙菅頗雀裾澄摺寸世瀬畝是凄制勢姓征性成政整星晴棲栖正清牲生盛精聖声製西誠誓請逝醒青静斉税脆隻席惜戚斥昔析石積籍績脊責赤跡蹟碩切拙接摂折設窃節説雪絶舌蝉仙先千占宣専尖川戦扇撰栓栴泉浅洗染潜煎煽旋穿箭線"],
        ["9140", "繊羨腺舛船薦詮賎践選遷銭銑閃鮮前善漸然全禅繕膳糎噌塑岨措曾曽楚狙疏疎礎祖租粗素組蘇訴阻遡鼠僧創双叢倉喪壮奏爽宋層匝惣想捜掃挿掻"],
        ["9180", "操早曹巣槍槽漕燥争痩相窓糟総綜聡草荘葬蒼藻装走送遭鎗霜騒像増憎臓蔵贈造促側則即息捉束測足速俗属賊族続卒袖其揃存孫尊損村遜他多太汰詑唾堕妥惰打柁舵楕陀駄騨体堆対耐岱帯待怠態戴替泰滞胎腿苔袋貸退逮隊黛鯛代台大第醍題鷹滝瀧卓啄宅托択拓沢濯琢託鐸濁諾茸凧蛸只"],
        ["9240", "叩但達辰奪脱巽竪辿棚谷狸鱈樽誰丹単嘆坦担探旦歎淡湛炭短端箪綻耽胆蛋誕鍛団壇弾断暖檀段男談値知地弛恥智池痴稚置致蜘遅馳築畜竹筑蓄"],
        ["9280", "逐秩窒茶嫡着中仲宙忠抽昼柱注虫衷註酎鋳駐樗瀦猪苧著貯丁兆凋喋寵帖帳庁弔張彫徴懲挑暢朝潮牒町眺聴脹腸蝶調諜超跳銚長頂鳥勅捗直朕沈珍賃鎮陳津墜椎槌追鎚痛通塚栂掴槻佃漬柘辻蔦綴鍔椿潰坪壷嬬紬爪吊釣鶴亭低停偵剃貞呈堤定帝底庭廷弟悌抵挺提梯汀碇禎程締艇訂諦蹄逓"],
        ["9340", "邸鄭釘鼎泥摘擢敵滴的笛適鏑溺哲徹撤轍迭鉄典填天展店添纏甜貼転顛点伝殿澱田電兎吐堵塗妬屠徒斗杜渡登菟賭途都鍍砥砺努度土奴怒倒党冬"],
        ["9380", "凍刀唐塔塘套宕島嶋悼投搭東桃梼棟盗淘湯涛灯燈当痘祷等答筒糖統到董蕩藤討謄豆踏逃透鐙陶頭騰闘働動同堂導憧撞洞瞳童胴萄道銅峠鴇匿得徳涜特督禿篤毒独読栃橡凸突椴届鳶苫寅酉瀞噸屯惇敦沌豚遁頓呑曇鈍奈那内乍凪薙謎灘捺鍋楢馴縄畷南楠軟難汝二尼弐迩匂賑肉虹廿日乳入"],
        ["9440", "如尿韮任妊忍認濡禰祢寧葱猫熱年念捻撚燃粘乃廼之埜嚢悩濃納能脳膿農覗蚤巴把播覇杷波派琶破婆罵芭馬俳廃拝排敗杯盃牌背肺輩配倍培媒梅"],
        ["9480", "楳煤狽買売賠陪這蝿秤矧萩伯剥博拍柏泊白箔粕舶薄迫曝漠爆縛莫駁麦函箱硲箸肇筈櫨幡肌畑畠八鉢溌発醗髪伐罰抜筏閥鳩噺塙蛤隼伴判半反叛帆搬斑板氾汎版犯班畔繁般藩販範釆煩頒飯挽晩番盤磐蕃蛮匪卑否妃庇彼悲扉批披斐比泌疲皮碑秘緋罷肥被誹費避非飛樋簸備尾微枇毘琵眉美"],
        ["9540", "鼻柊稗匹疋髭彦膝菱肘弼必畢筆逼桧姫媛紐百謬俵彪標氷漂瓢票表評豹廟描病秒苗錨鋲蒜蛭鰭品彬斌浜瀕貧賓頻敏瓶不付埠夫婦富冨布府怖扶敷"],
        ["9580", "斧普浮父符腐膚芙譜負賦赴阜附侮撫武舞葡蕪部封楓風葺蕗伏副復幅服福腹複覆淵弗払沸仏物鮒分吻噴墳憤扮焚奮粉糞紛雰文聞丙併兵塀幣平弊柄並蔽閉陛米頁僻壁癖碧別瞥蔑箆偏変片篇編辺返遍便勉娩弁鞭保舗鋪圃捕歩甫補輔穂募墓慕戊暮母簿菩倣俸包呆報奉宝峰峯崩庖抱捧放方朋"],
        ["9640", "法泡烹砲縫胞芳萌蓬蜂褒訪豊邦鋒飽鳳鵬乏亡傍剖坊妨帽忘忙房暴望某棒冒紡肪膨謀貌貿鉾防吠頬北僕卜墨撲朴牧睦穆釦勃没殆堀幌奔本翻凡盆"],
        ["9680", "摩磨魔麻埋妹昧枚毎哩槙幕膜枕鮪柾鱒桝亦俣又抹末沫迄侭繭麿万慢満漫蔓味未魅巳箕岬密蜜湊蓑稔脈妙粍民眠務夢無牟矛霧鵡椋婿娘冥名命明盟迷銘鳴姪牝滅免棉綿緬面麺摸模茂妄孟毛猛盲網耗蒙儲木黙目杢勿餅尤戻籾貰問悶紋門匁也冶夜爺耶野弥矢厄役約薬訳躍靖柳薮鑓愉愈油癒"],
        ["9740", "諭輸唯佑優勇友宥幽悠憂揖有柚湧涌猶猷由祐裕誘遊邑郵雄融夕予余与誉輿預傭幼妖容庸揚揺擁曜楊様洋溶熔用窯羊耀葉蓉要謡踊遥陽養慾抑欲"],
        ["9780", "沃浴翌翼淀羅螺裸来莱頼雷洛絡落酪乱卵嵐欄濫藍蘭覧利吏履李梨理璃痢裏裡里離陸律率立葎掠略劉流溜琉留硫粒隆竜龍侶慮旅虜了亮僚両凌寮料梁涼猟療瞭稜糧良諒遼量陵領力緑倫厘林淋燐琳臨輪隣鱗麟瑠塁涙累類令伶例冷励嶺怜玲礼苓鈴隷零霊麗齢暦歴列劣烈裂廉恋憐漣煉簾練聯"],
        ["9840", "蓮連錬呂魯櫓炉賂路露労婁廊弄朗楼榔浪漏牢狼篭老聾蝋郎六麓禄肋録論倭和話歪賄脇惑枠鷲亙亘鰐詫藁蕨椀湾碗腕"],
        ["989f", "弌丐丕个丱丶丼丿乂乖乘亂亅豫亊舒弍于亞亟亠亢亰亳亶从仍仄仆仂仗仞仭仟价伉佚估佛佝佗佇佶侈侏侘佻佩佰侑佯來侖儘俔俟俎俘俛俑俚俐俤俥倚倨倔倪倥倅伜俶倡倩倬俾俯們倆偃假會偕偐偈做偖偬偸傀傚傅傴傲"],
        ["9940", "僉僊傳僂僖僞僥僭僣僮價僵儉儁儂儖儕儔儚儡儺儷儼儻儿兀兒兌兔兢竸兩兪兮冀冂囘册冉冏冑冓冕冖冤冦冢冩冪冫决冱冲冰况冽凅凉凛几處凩凭"],
        ["9980", "凰凵凾刄刋刔刎刧刪刮刳刹剏剄剋剌剞剔剪剴剩剳剿剽劍劔劒剱劈劑辨辧劬劭劼劵勁勍勗勞勣勦飭勠勳勵勸勹匆匈甸匍匐匏匕匚匣匯匱匳匸區卆卅丗卉卍凖卞卩卮夘卻卷厂厖厠厦厥厮厰厶參簒雙叟曼燮叮叨叭叺吁吽呀听吭吼吮吶吩吝呎咏呵咎呟呱呷呰咒呻咀呶咄咐咆哇咢咸咥咬哄哈咨"],
        ["9a40", "咫哂咤咾咼哘哥哦唏唔哽哮哭哺哢唹啀啣啌售啜啅啖啗唸唳啝喙喀咯喊喟啻啾喘喞單啼喃喩喇喨嗚嗅嗟嗄嗜嗤嗔嘔嗷嘖嗾嗽嘛嗹噎噐營嘴嘶嘲嘸"],
        ["9a80", "噫噤嘯噬噪嚆嚀嚊嚠嚔嚏嚥嚮嚶嚴囂嚼囁囃囀囈囎囑囓囗囮囹圀囿圄圉圈國圍圓團圖嗇圜圦圷圸坎圻址坏坩埀垈坡坿垉垓垠垳垤垪垰埃埆埔埒埓堊埖埣堋堙堝塲堡塢塋塰毀塒堽塹墅墹墟墫墺壞墻墸墮壅壓壑壗壙壘壥壜壤壟壯壺壹壻壼壽夂夊夐夛梦夥夬夭夲夸夾竒奕奐奎奚奘奢奠奧奬奩"],
        ["9b40", "奸妁妝佞侫妣妲姆姨姜妍姙姚娥娟娑娜娉娚婀婬婉娵娶婢婪媚媼媾嫋嫂媽嫣嫗嫦嫩嫖嫺嫻嬌嬋嬖嬲嫐嬪嬶嬾孃孅孀孑孕孚孛孥孩孰孳孵學斈孺宀"],
        ["9b80", "它宦宸寃寇寉寔寐寤實寢寞寥寫寰寶寳尅將專對尓尠尢尨尸尹屁屆屎屓屐屏孱屬屮乢屶屹岌岑岔妛岫岻岶岼岷峅岾峇峙峩峽峺峭嶌峪崋崕崗嵜崟崛崑崔崢崚崙崘嵌嵒嵎嵋嵬嵳嵶嶇嶄嶂嶢嶝嶬嶮嶽嶐嶷嶼巉巍巓巒巖巛巫已巵帋帚帙帑帛帶帷幄幃幀幎幗幔幟幢幤幇幵并幺麼广庠廁廂廈廐廏"],
        ["9c40", "廖廣廝廚廛廢廡廨廩廬廱廳廰廴廸廾弃弉彝彜弋弑弖弩弭弸彁彈彌彎弯彑彖彗彙彡彭彳彷徃徂彿徊很徑徇從徙徘徠徨徭徼忖忻忤忸忱忝悳忿怡恠"],
        ["9c80", "怙怐怩怎怱怛怕怫怦怏怺恚恁恪恷恟恊恆恍恣恃恤恂恬恫恙悁悍惧悃悚悄悛悖悗悒悧悋惡悸惠惓悴忰悽惆悵惘慍愕愆惶惷愀惴惺愃愡惻惱愍愎慇愾愨愧慊愿愼愬愴愽慂慄慳慷慘慙慚慫慴慯慥慱慟慝慓慵憙憖憇憬憔憚憊憑憫憮懌懊應懷懈懃懆憺懋罹懍懦懣懶懺懴懿懽懼懾戀戈戉戍戌戔戛"],
        ["9d40", "戞戡截戮戰戲戳扁扎扞扣扛扠扨扼抂抉找抒抓抖拔抃抔拗拑抻拏拿拆擔拈拜拌拊拂拇抛拉挌拮拱挧挂挈拯拵捐挾捍搜捏掖掎掀掫捶掣掏掉掟掵捫"],
        ["9d80", "捩掾揩揀揆揣揉插揶揄搖搴搆搓搦搶攝搗搨搏摧摯摶摎攪撕撓撥撩撈撼據擒擅擇撻擘擂擱擧舉擠擡抬擣擯攬擶擴擲擺攀擽攘攜攅攤攣攫攴攵攷收攸畋效敖敕敍敘敞敝敲數斂斃變斛斟斫斷旃旆旁旄旌旒旛旙无旡旱杲昊昃旻杳昵昶昴昜晏晄晉晁晞晝晤晧晨晟晢晰暃暈暎暉暄暘暝曁暹曉暾暼"],
        ["9e40", "曄暸曖曚曠昿曦曩曰曵曷朏朖朞朦朧霸朮朿朶杁朸朷杆杞杠杙杣杤枉杰枩杼杪枌枋枦枡枅枷柯枴柬枳柩枸柤柞柝柢柮枹柎柆柧檜栞框栩桀桍栲桎"],
        ["9e80", "梳栫桙档桷桿梟梏梭梔條梛梃檮梹桴梵梠梺椏梍桾椁棊椈棘椢椦棡椌棍棔棧棕椶椒椄棗棣椥棹棠棯椨椪椚椣椡棆楹楷楜楸楫楔楾楮椹楴椽楙椰楡楞楝榁楪榲榮槐榿槁槓榾槎寨槊槝榻槃榧樮榑榠榜榕榴槞槨樂樛槿權槹槲槧樅榱樞槭樔槫樊樒櫁樣樓橄樌橲樶橸橇橢橙橦橈樸樢檐檍檠檄檢檣"],
        ["9f40", "檗蘗檻櫃櫂檸檳檬櫞櫑櫟檪櫚櫪櫻欅蘖櫺欒欖鬱欟欸欷盜欹飮歇歃歉歐歙歔歛歟歡歸歹歿殀殄殃殍殘殕殞殤殪殫殯殲殱殳殷殼毆毋毓毟毬毫毳毯"],
        ["9f80", "麾氈氓气氛氤氣汞汕汢汪沂沍沚沁沛汾汨汳沒沐泄泱泓沽泗泅泝沮沱沾沺泛泯泙泪洟衍洶洫洽洸洙洵洳洒洌浣涓浤浚浹浙涎涕濤涅淹渕渊涵淇淦涸淆淬淞淌淨淒淅淺淙淤淕淪淮渭湮渮渙湲湟渾渣湫渫湶湍渟湃渺湎渤滿渝游溂溪溘滉溷滓溽溯滄溲滔滕溏溥滂溟潁漑灌滬滸滾漿滲漱滯漲滌"],
        ["e040", "漾漓滷澆潺潸澁澀潯潛濳潭澂潼潘澎澑濂潦澳澣澡澤澹濆澪濟濕濬濔濘濱濮濛瀉瀋濺瀑瀁瀏濾瀛瀚潴瀝瀘瀟瀰瀾瀲灑灣炙炒炯烱炬炸炳炮烟烋烝"],
        ["e080", "烙焉烽焜焙煥煕熈煦煢煌煖煬熏燻熄熕熨熬燗熹熾燒燉燔燎燠燬燧燵燼燹燿爍爐爛爨爭爬爰爲爻爼爿牀牆牋牘牴牾犂犁犇犒犖犢犧犹犲狃狆狄狎狒狢狠狡狹狷倏猗猊猜猖猝猴猯猩猥猾獎獏默獗獪獨獰獸獵獻獺珈玳珎玻珀珥珮珞璢琅瑯琥珸琲琺瑕琿瑟瑙瑁瑜瑩瑰瑣瑪瑶瑾璋璞璧瓊瓏瓔珱"],
        ["e140", "瓠瓣瓧瓩瓮瓲瓰瓱瓸瓷甄甃甅甌甎甍甕甓甞甦甬甼畄畍畊畉畛畆畚畩畤畧畫畭畸當疆疇畴疊疉疂疔疚疝疥疣痂疳痃疵疽疸疼疱痍痊痒痙痣痞痾痿"],
        ["e180", "痼瘁痰痺痲痳瘋瘍瘉瘟瘧瘠瘡瘢瘤瘴瘰瘻癇癈癆癜癘癡癢癨癩癪癧癬癰癲癶癸發皀皃皈皋皎皖皓皙皚皰皴皸皹皺盂盍盖盒盞盡盥盧盪蘯盻眈眇眄眩眤眞眥眦眛眷眸睇睚睨睫睛睥睿睾睹瞎瞋瞑瞠瞞瞰瞶瞹瞿瞼瞽瞻矇矍矗矚矜矣矮矼砌砒礦砠礪硅碎硴碆硼碚碌碣碵碪碯磑磆磋磔碾碼磅磊磬"],
        ["e240", "磧磚磽磴礇礒礑礙礬礫祀祠祗祟祚祕祓祺祿禊禝禧齋禪禮禳禹禺秉秕秧秬秡秣稈稍稘稙稠稟禀稱稻稾稷穃穗穉穡穢穩龝穰穹穽窈窗窕窘窖窩竈窰"],
        ["e280", "窶竅竄窿邃竇竊竍竏竕竓站竚竝竡竢竦竭竰笂笏笊笆笳笘笙笞笵笨笶筐筺笄筍笋筌筅筵筥筴筧筰筱筬筮箝箘箟箍箜箚箋箒箏筝箙篋篁篌篏箴篆篝篩簑簔篦篥籠簀簇簓篳篷簗簍篶簣簧簪簟簷簫簽籌籃籔籏籀籐籘籟籤籖籥籬籵粃粐粤粭粢粫粡粨粳粲粱粮粹粽糀糅糂糘糒糜糢鬻糯糲糴糶糺紆"],
        ["e340", "紂紜紕紊絅絋紮紲紿紵絆絳絖絎絲絨絮絏絣經綉絛綏絽綛綺綮綣綵緇綽綫總綢綯緜綸綟綰緘緝緤緞緻緲緡縅縊縣縡縒縱縟縉縋縢繆繦縻縵縹繃縷"],
        ["e380", "縲縺繧繝繖繞繙繚繹繪繩繼繻纃緕繽辮繿纈纉續纒纐纓纔纖纎纛纜缸缺罅罌罍罎罐网罕罔罘罟罠罨罩罧罸羂羆羃羈羇羌羔羞羝羚羣羯羲羹羮羶羸譱翅翆翊翕翔翡翦翩翳翹飜耆耄耋耒耘耙耜耡耨耿耻聊聆聒聘聚聟聢聨聳聲聰聶聹聽聿肄肆肅肛肓肚肭冐肬胛胥胙胝胄胚胖脉胯胱脛脩脣脯腋"],
        ["e440", "隋腆脾腓腑胼腱腮腥腦腴膃膈膊膀膂膠膕膤膣腟膓膩膰膵膾膸膽臀臂膺臉臍臑臙臘臈臚臟臠臧臺臻臾舁舂舅與舊舍舐舖舩舫舸舳艀艙艘艝艚艟艤"],
        ["e480", "艢艨艪艫舮艱艷艸艾芍芒芫芟芻芬苡苣苟苒苴苳苺莓范苻苹苞茆苜茉苙茵茴茖茲茱荀茹荐荅茯茫茗茘莅莚莪莟莢莖茣莎莇莊荼莵荳荵莠莉莨菴萓菫菎菽萃菘萋菁菷萇菠菲萍萢萠莽萸蔆菻葭萪萼蕚蒄葷葫蒭葮蒂葩葆萬葯葹萵蓊葢蒹蒿蒟蓙蓍蒻蓚蓐蓁蓆蓖蒡蔡蓿蓴蔗蔘蔬蔟蔕蔔蓼蕀蕣蕘蕈"],
        ["e540", "蕁蘂蕋蕕薀薤薈薑薊薨蕭薔薛藪薇薜蕷蕾薐藉薺藏薹藐藕藝藥藜藹蘊蘓蘋藾藺蘆蘢蘚蘰蘿虍乕虔號虧虱蚓蚣蚩蚪蚋蚌蚶蚯蛄蛆蚰蛉蠣蚫蛔蛞蛩蛬"],
        ["e580", "蛟蛛蛯蜒蜆蜈蜀蜃蛻蜑蜉蜍蛹蜊蜴蜿蜷蜻蜥蜩蜚蝠蝟蝸蝌蝎蝴蝗蝨蝮蝙蝓蝣蝪蠅螢螟螂螯蟋螽蟀蟐雖螫蟄螳蟇蟆螻蟯蟲蟠蠏蠍蟾蟶蟷蠎蟒蠑蠖蠕蠢蠡蠱蠶蠹蠧蠻衄衂衒衙衞衢衫袁衾袞衵衽袵衲袂袗袒袮袙袢袍袤袰袿袱裃裄裔裘裙裝裹褂裼裴裨裲褄褌褊褓襃褞褥褪褫襁襄褻褶褸襌褝襠襞"],
        ["e640", "襦襤襭襪襯襴襷襾覃覈覊覓覘覡覩覦覬覯覲覺覽覿觀觚觜觝觧觴觸訃訖訐訌訛訝訥訶詁詛詒詆詈詼詭詬詢誅誂誄誨誡誑誥誦誚誣諄諍諂諚諫諳諧"],
        ["e680", "諤諱謔諠諢諷諞諛謌謇謚諡謖謐謗謠謳鞫謦謫謾謨譁譌譏譎證譖譛譚譫譟譬譯譴譽讀讌讎讒讓讖讙讚谺豁谿豈豌豎豐豕豢豬豸豺貂貉貅貊貍貎貔豼貘戝貭貪貽貲貳貮貶賈賁賤賣賚賽賺賻贄贅贊贇贏贍贐齎贓賍贔贖赧赭赱赳趁趙跂趾趺跏跚跖跌跛跋跪跫跟跣跼踈踉跿踝踞踐踟蹂踵踰踴蹊"],
        ["e740", "蹇蹉蹌蹐蹈蹙蹤蹠踪蹣蹕蹶蹲蹼躁躇躅躄躋躊躓躑躔躙躪躡躬躰軆躱躾軅軈軋軛軣軼軻軫軾輊輅輕輒輙輓輜輟輛輌輦輳輻輹轅轂輾轌轉轆轎轗轜"],
        ["e780", "轢轣轤辜辟辣辭辯辷迚迥迢迪迯邇迴逅迹迺逑逕逡逍逞逖逋逧逶逵逹迸遏遐遑遒逎遉逾遖遘遞遨遯遶隨遲邂遽邁邀邊邉邏邨邯邱邵郢郤扈郛鄂鄒鄙鄲鄰酊酖酘酣酥酩酳酲醋醉醂醢醫醯醪醵醴醺釀釁釉釋釐釖釟釡釛釼釵釶鈞釿鈔鈬鈕鈑鉞鉗鉅鉉鉤鉈銕鈿鉋鉐銜銖銓銛鉚鋏銹銷鋩錏鋺鍄錮"],
        ["e840", "錙錢錚錣錺錵錻鍜鍠鍼鍮鍖鎰鎬鎭鎔鎹鏖鏗鏨鏥鏘鏃鏝鏐鏈鏤鐚鐔鐓鐃鐇鐐鐶鐫鐵鐡鐺鑁鑒鑄鑛鑠鑢鑞鑪鈩鑰鑵鑷鑽鑚鑼鑾钁鑿閂閇閊閔閖閘閙"],
        ["e880", "閠閨閧閭閼閻閹閾闊濶闃闍闌闕闔闖關闡闥闢阡阨阮阯陂陌陏陋陷陜陞陝陟陦陲陬隍隘隕隗險隧隱隲隰隴隶隸隹雎雋雉雍襍雜霍雕雹霄霆霈霓霎霑霏霖霙霤霪霰霹霽霾靄靆靈靂靉靜靠靤靦靨勒靫靱靹鞅靼鞁靺鞆鞋鞏鞐鞜鞨鞦鞣鞳鞴韃韆韈韋韜韭齏韲竟韶韵頏頌頸頤頡頷頽顆顏顋顫顯顰"],
        ["e940", "顱顴顳颪颯颱颶飄飃飆飩飫餃餉餒餔餘餡餝餞餤餠餬餮餽餾饂饉饅饐饋饑饒饌饕馗馘馥馭馮馼駟駛駝駘駑駭駮駱駲駻駸騁騏騅駢騙騫騷驅驂驀驃"],
        ["e980", "騾驕驍驛驗驟驢驥驤驩驫驪骭骰骼髀髏髑髓體髞髟髢髣髦髯髫髮髴髱髷髻鬆鬘鬚鬟鬢鬣鬥鬧鬨鬩鬪鬮鬯鬲魄魃魏魍魎魑魘魴鮓鮃鮑鮖鮗鮟鮠鮨鮴鯀鯊鮹鯆鯏鯑鯒鯣鯢鯤鯔鯡鰺鯲鯱鯰鰕鰔鰉鰓鰌鰆鰈鰒鰊鰄鰮鰛鰥鰤鰡鰰鱇鰲鱆鰾鱚鱠鱧鱶鱸鳧鳬鳰鴉鴈鳫鴃鴆鴪鴦鶯鴣鴟鵄鴕鴒鵁鴿鴾鵆鵈"],
        ["ea40", "鵝鵞鵤鵑鵐鵙鵲鶉鶇鶫鵯鵺鶚鶤鶩鶲鷄鷁鶻鶸鶺鷆鷏鷂鷙鷓鷸鷦鷭鷯鷽鸚鸛鸞鹵鹹鹽麁麈麋麌麒麕麑麝麥麩麸麪麭靡黌黎黏黐黔黜點黝黠黥黨黯"],
        ["ea80", "黴黶黷黹黻黼黽鼇鼈皷鼕鼡鼬鼾齊齒齔齣齟齠齡齦齧齬齪齷齲齶龕龜龠堯槇遙瑤凜熙"],
        ["ed40", "纊褜鍈銈蓜俉炻昱棈鋹曻彅丨仡仼伀伃伹佖侒侊侚侔俍偀倢俿倞偆偰偂傔僴僘兊兤冝冾凬刕劜劦勀勛匀匇匤卲厓厲叝﨎咜咊咩哿喆坙坥垬埈埇﨏"],
        ["ed80", "塚增墲夋奓奛奝奣妤妺孖寀甯寘寬尞岦岺峵崧嵓﨑嵂嵭嶸嶹巐弡弴彧德忞恝悅悊惞惕愠惲愑愷愰憘戓抦揵摠撝擎敎昀昕昻昉昮昞昤晥晗晙晴晳暙暠暲暿曺朎朗杦枻桒柀栁桄棏﨓楨﨔榘槢樰橫橆橳橾櫢櫤毖氿汜沆汯泚洄涇浯涖涬淏淸淲淼渹湜渧渼溿澈澵濵瀅瀇瀨炅炫焏焄煜煆煇凞燁燾犱"],
        ["ee40", "犾猤猪獷玽珉珖珣珒琇珵琦琪琩琮瑢璉璟甁畯皂皜皞皛皦益睆劯砡硎硤硺礰礼神祥禔福禛竑竧靖竫箞精絈絜綷綠緖繒罇羡羽茁荢荿菇菶葈蒴蕓蕙"],
        ["ee80", "蕫﨟薰蘒﨡蠇裵訒訷詹誧誾諟諸諶譓譿賰賴贒赶﨣軏﨤逸遧郞都鄕鄧釚釗釞釭釮釤釥鈆鈐鈊鈺鉀鈼鉎鉙鉑鈹鉧銧鉷鉸鋧鋗鋙鋐﨧鋕鋠鋓錥錡鋻﨨錞鋿錝錂鍰鍗鎤鏆鏞鏸鐱鑅鑈閒隆﨩隝隯霳霻靃靍靏靑靕顗顥飯飼餧館馞驎髙髜魵魲鮏鮱鮻鰀鵰鵫鶴鸙黑"],
        ["eeef", "ⅰ", 9, "￢￤＇＂"],
        ["f040", "", 62],
        ["f080", "", 124],
        ["f140", "", 62],
        ["f180", "", 124],
        ["f240", "", 62],
        ["f280", "", 124],
        ["f340", "", 62],
        ["f380", "", 124],
        ["f440", "", 62],
        ["f480", "", 124],
        ["f540", "", 62],
        ["f580", "", 124],
        ["f640", "", 62],
        ["f680", "", 124],
        ["f740", "", 62],
        ["f780", "", 124],
        ["f840", "", 62],
        ["f880", "", 124],
        ["f940", ""],
        ["fa40", "ⅰ", 9, "Ⅰ", 9, "￢￤＇＂㈱№℡∵纊褜鍈銈蓜俉炻昱棈鋹曻彅丨仡仼伀伃伹佖侒侊侚侔俍偀倢俿倞偆偰偂傔僴僘兊"],
        ["fa80", "兤冝冾凬刕劜劦勀勛匀匇匤卲厓厲叝﨎咜咊咩哿喆坙坥垬埈埇﨏塚增墲夋奓奛奝奣妤妺孖寀甯寘寬尞岦岺峵崧嵓﨑嵂嵭嶸嶹巐弡弴彧德忞恝悅悊惞惕愠惲愑愷愰憘戓抦揵摠撝擎敎昀昕昻昉昮昞昤晥晗晙晴晳暙暠暲暿曺朎朗杦枻桒柀栁桄棏﨓楨﨔榘槢樰橫橆橳橾櫢櫤毖氿汜沆汯泚洄涇浯"],
        ["fb40", "涖涬淏淸淲淼渹湜渧渼溿澈澵濵瀅瀇瀨炅炫焏焄煜煆煇凞燁燾犱犾猤猪獷玽珉珖珣珒琇珵琦琪琩琮瑢璉璟甁畯皂皜皞皛皦益睆劯砡硎硤硺礰礼神"],
        ["fb80", "祥禔福禛竑竧靖竫箞精絈絜綷綠緖繒罇羡羽茁荢荿菇菶葈蒴蕓蕙蕫﨟薰蘒﨡蠇裵訒訷詹誧誾諟諸諶譓譿賰賴贒赶﨣軏﨤逸遧郞都鄕鄧釚釗釞釭釮釤釥鈆鈐鈊鈺鉀鈼鉎鉙鉑鈹鉧銧鉷鉸鋧鋗鋙鋐﨧鋕鋠鋓錥錡鋻﨨錞鋿錝錂鍰鍗鎤鏆鏞鏸鐱鑅鑈閒隆﨩隝隯霳霻靃靍靏靑靕顗顥飯飼餧館馞驎髙"],
        ["fc40", "髜魵魲鮏鮱鮻鰀鵰鵫鶴鸙黑"]
    ];
    module.exports = json;
},
/* iconv-lite/encodings/tables/eucjp.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = [
        ["0", "\u0000", 127],
        ["8ea1", "｡", 62],
        ["a1a1", "　、。，．・：；？！゛゜´｀¨＾￣＿ヽヾゝゞ〃仝々〆〇ー―‐／＼～∥｜…‥‘’“”（）〔〕［］｛｝〈", 9, "＋－±×÷＝≠＜＞≦≧∞∴♂♀°′″℃￥＄￠￡％＃＆＊＠§☆★○●◎◇"],
        ["a2a1", "◆□■△▲▽▼※〒→←↑↓〓"],
        ["a2ba", "∈∋⊆⊇⊂⊃∪∩"],
        ["a2ca", "∧∨￢⇒⇔∀∃"],
        ["a2dc", "∠⊥⌒∂∇≡≒≪≫√∽∝∵∫∬"],
        ["a2f2", "Å‰♯♭♪†‡¶"],
        ["a2fe", "◯"],
        ["a3b0", "０", 9],
        ["a3c1", "Ａ", 25],
        ["a3e1", "ａ", 25],
        ["a4a1", "ぁ", 82],
        ["a5a1", "ァ", 85],
        ["a6a1", "Α", 16, "Σ", 6],
        ["a6c1", "α", 16, "σ", 6],
        ["a7a1", "А", 5, "ЁЖ", 25],
        ["a7d1", "а", 5, "ёж", 25],
        ["a8a1", "─│┌┐┘└├┬┤┴┼━┃┏┓┛┗┣┳┫┻╋┠┯┨┷┿┝┰┥┸╂"],
        ["ada1", "①", 19, "Ⅰ", 9],
        ["adc0", "㍉㌔㌢㍍㌘㌧㌃㌶㍑㍗㌍㌦㌣㌫㍊㌻㎜㎝㎞㎎㎏㏄㎡"],
        ["addf", "㍻〝〟№㏍℡㊤", 4, "㈱㈲㈹㍾㍽㍼≒≡∫∮∑√⊥∠∟⊿∵∩∪"],
        ["b0a1", "亜唖娃阿哀愛挨姶逢葵茜穐悪握渥旭葦芦鯵梓圧斡扱宛姐虻飴絢綾鮎或粟袷安庵按暗案闇鞍杏以伊位依偉囲夷委威尉惟意慰易椅為畏異移維緯胃萎衣謂違遺医井亥域育郁磯一壱溢逸稲茨芋鰯允印咽員因姻引飲淫胤蔭"],
        ["b1a1", "院陰隠韻吋右宇烏羽迂雨卯鵜窺丑碓臼渦嘘唄欝蔚鰻姥厩浦瓜閏噂云運雲荏餌叡営嬰影映曳栄永泳洩瑛盈穎頴英衛詠鋭液疫益駅悦謁越閲榎厭円園堰奄宴延怨掩援沿演炎焔煙燕猿縁艶苑薗遠鉛鴛塩於汚甥凹央奥往応"],
        ["b2a1", "押旺横欧殴王翁襖鴬鴎黄岡沖荻億屋憶臆桶牡乙俺卸恩温穏音下化仮何伽価佳加可嘉夏嫁家寡科暇果架歌河火珂禍禾稼箇花苛茄荷華菓蝦課嘩貨迦過霞蚊俄峨我牙画臥芽蛾賀雅餓駕介会解回塊壊廻快怪悔恢懐戒拐改"],
        ["b3a1", "魁晦械海灰界皆絵芥蟹開階貝凱劾外咳害崖慨概涯碍蓋街該鎧骸浬馨蛙垣柿蛎鈎劃嚇各廓拡撹格核殻獲確穫覚角赫較郭閣隔革学岳楽額顎掛笠樫橿梶鰍潟割喝恰括活渇滑葛褐轄且鰹叶椛樺鞄株兜竃蒲釜鎌噛鴨栢茅萱"],
        ["b4a1", "粥刈苅瓦乾侃冠寒刊勘勧巻喚堪姦完官寛干幹患感慣憾換敢柑桓棺款歓汗漢澗潅環甘監看竿管簡緩缶翰肝艦莞観諌貫還鑑間閑関陥韓館舘丸含岸巌玩癌眼岩翫贋雁頑顔願企伎危喜器基奇嬉寄岐希幾忌揮机旗既期棋棄"],
        ["b5a1", "機帰毅気汽畿祈季稀紀徽規記貴起軌輝飢騎鬼亀偽儀妓宜戯技擬欺犠疑祇義蟻誼議掬菊鞠吉吃喫桔橘詰砧杵黍却客脚虐逆丘久仇休及吸宮弓急救朽求汲泣灸球究窮笈級糾給旧牛去居巨拒拠挙渠虚許距鋸漁禦魚亨享京"],
        ["b6a1", "供侠僑兇競共凶協匡卿叫喬境峡強彊怯恐恭挟教橋況狂狭矯胸脅興蕎郷鏡響饗驚仰凝尭暁業局曲極玉桐粁僅勤均巾錦斤欣欽琴禁禽筋緊芹菌衿襟謹近金吟銀九倶句区狗玖矩苦躯駆駈駒具愚虞喰空偶寓遇隅串櫛釧屑屈"],
        ["b7a1", "掘窟沓靴轡窪熊隈粂栗繰桑鍬勲君薫訓群軍郡卦袈祁係傾刑兄啓圭珪型契形径恵慶慧憩掲携敬景桂渓畦稽系経継繋罫茎荊蛍計詣警軽頚鶏芸迎鯨劇戟撃激隙桁傑欠決潔穴結血訣月件倹倦健兼券剣喧圏堅嫌建憲懸拳捲"],
        ["b8a1", "検権牽犬献研硯絹県肩見謙賢軒遣鍵険顕験鹸元原厳幻弦減源玄現絃舷言諺限乎個古呼固姑孤己庫弧戸故枯湖狐糊袴股胡菰虎誇跨鈷雇顧鼓五互伍午呉吾娯後御悟梧檎瑚碁語誤護醐乞鯉交佼侯候倖光公功効勾厚口向"],
        ["b9a1", "后喉坑垢好孔孝宏工巧巷幸広庚康弘恒慌抗拘控攻昂晃更杭校梗構江洪浩港溝甲皇硬稿糠紅紘絞綱耕考肯肱腔膏航荒行衡講貢購郊酵鉱砿鋼閤降項香高鴻剛劫号合壕拷濠豪轟麹克刻告国穀酷鵠黒獄漉腰甑忽惚骨狛込"],
        ["baa1", "此頃今困坤墾婚恨懇昏昆根梱混痕紺艮魂些佐叉唆嵯左差査沙瑳砂詐鎖裟坐座挫債催再最哉塞妻宰彩才採栽歳済災采犀砕砦祭斎細菜裁載際剤在材罪財冴坂阪堺榊肴咲崎埼碕鷺作削咋搾昨朔柵窄策索錯桜鮭笹匙冊刷"],
        ["bba1", "察拶撮擦札殺薩雑皐鯖捌錆鮫皿晒三傘参山惨撒散桟燦珊産算纂蚕讃賛酸餐斬暫残仕仔伺使刺司史嗣四士始姉姿子屍市師志思指支孜斯施旨枝止死氏獅祉私糸紙紫肢脂至視詞詩試誌諮資賜雌飼歯事似侍児字寺慈持時"],
        ["bca1", "次滋治爾璽痔磁示而耳自蒔辞汐鹿式識鴫竺軸宍雫七叱執失嫉室悉湿漆疾質実蔀篠偲柴芝屡蕊縞舎写射捨赦斜煮社紗者謝車遮蛇邪借勺尺杓灼爵酌釈錫若寂弱惹主取守手朱殊狩珠種腫趣酒首儒受呪寿授樹綬需囚収周"],
        ["bda1", "宗就州修愁拾洲秀秋終繍習臭舟蒐衆襲讐蹴輯週酋酬集醜什住充十従戎柔汁渋獣縦重銃叔夙宿淑祝縮粛塾熟出術述俊峻春瞬竣舜駿准循旬楯殉淳準潤盾純巡遵醇順処初所暑曙渚庶緒署書薯藷諸助叙女序徐恕鋤除傷償"],
        ["bea1", "勝匠升召哨商唱嘗奨妾娼宵将小少尚庄床廠彰承抄招掌捷昇昌昭晶松梢樟樵沼消渉湘焼焦照症省硝礁祥称章笑粧紹肖菖蒋蕉衝裳訟証詔詳象賞醤鉦鍾鐘障鞘上丈丞乗冗剰城場壌嬢常情擾条杖浄状畳穣蒸譲醸錠嘱埴飾"],
        ["bfa1", "拭植殖燭織職色触食蝕辱尻伸信侵唇娠寝審心慎振新晋森榛浸深申疹真神秦紳臣芯薪親診身辛進針震人仁刃塵壬尋甚尽腎訊迅陣靭笥諏須酢図厨逗吹垂帥推水炊睡粋翠衰遂酔錐錘随瑞髄崇嵩数枢趨雛据杉椙菅頗雀裾"],
        ["c0a1", "澄摺寸世瀬畝是凄制勢姓征性成政整星晴棲栖正清牲生盛精聖声製西誠誓請逝醒青静斉税脆隻席惜戚斥昔析石積籍績脊責赤跡蹟碩切拙接摂折設窃節説雪絶舌蝉仙先千占宣専尖川戦扇撰栓栴泉浅洗染潜煎煽旋穿箭線"],
        ["c1a1", "繊羨腺舛船薦詮賎践選遷銭銑閃鮮前善漸然全禅繕膳糎噌塑岨措曾曽楚狙疏疎礎祖租粗素組蘇訴阻遡鼠僧創双叢倉喪壮奏爽宋層匝惣想捜掃挿掻操早曹巣槍槽漕燥争痩相窓糟総綜聡草荘葬蒼藻装走送遭鎗霜騒像増憎"],
        ["c2a1", "臓蔵贈造促側則即息捉束測足速俗属賊族続卒袖其揃存孫尊損村遜他多太汰詑唾堕妥惰打柁舵楕陀駄騨体堆対耐岱帯待怠態戴替泰滞胎腿苔袋貸退逮隊黛鯛代台大第醍題鷹滝瀧卓啄宅托択拓沢濯琢託鐸濁諾茸凧蛸只"],
        ["c3a1", "叩但達辰奪脱巽竪辿棚谷狸鱈樽誰丹単嘆坦担探旦歎淡湛炭短端箪綻耽胆蛋誕鍛団壇弾断暖檀段男談値知地弛恥智池痴稚置致蜘遅馳築畜竹筑蓄逐秩窒茶嫡着中仲宙忠抽昼柱注虫衷註酎鋳駐樗瀦猪苧著貯丁兆凋喋寵"],
        ["c4a1", "帖帳庁弔張彫徴懲挑暢朝潮牒町眺聴脹腸蝶調諜超跳銚長頂鳥勅捗直朕沈珍賃鎮陳津墜椎槌追鎚痛通塚栂掴槻佃漬柘辻蔦綴鍔椿潰坪壷嬬紬爪吊釣鶴亭低停偵剃貞呈堤定帝底庭廷弟悌抵挺提梯汀碇禎程締艇訂諦蹄逓"],
        ["c5a1", "邸鄭釘鼎泥摘擢敵滴的笛適鏑溺哲徹撤轍迭鉄典填天展店添纏甜貼転顛点伝殿澱田電兎吐堵塗妬屠徒斗杜渡登菟賭途都鍍砥砺努度土奴怒倒党冬凍刀唐塔塘套宕島嶋悼投搭東桃梼棟盗淘湯涛灯燈当痘祷等答筒糖統到"],
        ["c6a1", "董蕩藤討謄豆踏逃透鐙陶頭騰闘働動同堂導憧撞洞瞳童胴萄道銅峠鴇匿得徳涜特督禿篤毒独読栃橡凸突椴届鳶苫寅酉瀞噸屯惇敦沌豚遁頓呑曇鈍奈那内乍凪薙謎灘捺鍋楢馴縄畷南楠軟難汝二尼弐迩匂賑肉虹廿日乳入"],
        ["c7a1", "如尿韮任妊忍認濡禰祢寧葱猫熱年念捻撚燃粘乃廼之埜嚢悩濃納能脳膿農覗蚤巴把播覇杷波派琶破婆罵芭馬俳廃拝排敗杯盃牌背肺輩配倍培媒梅楳煤狽買売賠陪這蝿秤矧萩伯剥博拍柏泊白箔粕舶薄迫曝漠爆縛莫駁麦"],
        ["c8a1", "函箱硲箸肇筈櫨幡肌畑畠八鉢溌発醗髪伐罰抜筏閥鳩噺塙蛤隼伴判半反叛帆搬斑板氾汎版犯班畔繁般藩販範釆煩頒飯挽晩番盤磐蕃蛮匪卑否妃庇彼悲扉批披斐比泌疲皮碑秘緋罷肥被誹費避非飛樋簸備尾微枇毘琵眉美"],
        ["c9a1", "鼻柊稗匹疋髭彦膝菱肘弼必畢筆逼桧姫媛紐百謬俵彪標氷漂瓢票表評豹廟描病秒苗錨鋲蒜蛭鰭品彬斌浜瀕貧賓頻敏瓶不付埠夫婦富冨布府怖扶敷斧普浮父符腐膚芙譜負賦赴阜附侮撫武舞葡蕪部封楓風葺蕗伏副復幅服"],
        ["caa1", "福腹複覆淵弗払沸仏物鮒分吻噴墳憤扮焚奮粉糞紛雰文聞丙併兵塀幣平弊柄並蔽閉陛米頁僻壁癖碧別瞥蔑箆偏変片篇編辺返遍便勉娩弁鞭保舗鋪圃捕歩甫補輔穂募墓慕戊暮母簿菩倣俸包呆報奉宝峰峯崩庖抱捧放方朋"],
        ["cba1", "法泡烹砲縫胞芳萌蓬蜂褒訪豊邦鋒飽鳳鵬乏亡傍剖坊妨帽忘忙房暴望某棒冒紡肪膨謀貌貿鉾防吠頬北僕卜墨撲朴牧睦穆釦勃没殆堀幌奔本翻凡盆摩磨魔麻埋妹昧枚毎哩槙幕膜枕鮪柾鱒桝亦俣又抹末沫迄侭繭麿万慢満"],
        ["cca1", "漫蔓味未魅巳箕岬密蜜湊蓑稔脈妙粍民眠務夢無牟矛霧鵡椋婿娘冥名命明盟迷銘鳴姪牝滅免棉綿緬面麺摸模茂妄孟毛猛盲網耗蒙儲木黙目杢勿餅尤戻籾貰問悶紋門匁也冶夜爺耶野弥矢厄役約薬訳躍靖柳薮鑓愉愈油癒"],
        ["cda1", "諭輸唯佑優勇友宥幽悠憂揖有柚湧涌猶猷由祐裕誘遊邑郵雄融夕予余与誉輿預傭幼妖容庸揚揺擁曜楊様洋溶熔用窯羊耀葉蓉要謡踊遥陽養慾抑欲沃浴翌翼淀羅螺裸来莱頼雷洛絡落酪乱卵嵐欄濫藍蘭覧利吏履李梨理璃"],
        ["cea1", "痢裏裡里離陸律率立葎掠略劉流溜琉留硫粒隆竜龍侶慮旅虜了亮僚両凌寮料梁涼猟療瞭稜糧良諒遼量陵領力緑倫厘林淋燐琳臨輪隣鱗麟瑠塁涙累類令伶例冷励嶺怜玲礼苓鈴隷零霊麗齢暦歴列劣烈裂廉恋憐漣煉簾練聯"],
        ["cfa1", "蓮連錬呂魯櫓炉賂路露労婁廊弄朗楼榔浪漏牢狼篭老聾蝋郎六麓禄肋録論倭和話歪賄脇惑枠鷲亙亘鰐詫藁蕨椀湾碗腕"],
        ["d0a1", "弌丐丕个丱丶丼丿乂乖乘亂亅豫亊舒弍于亞亟亠亢亰亳亶从仍仄仆仂仗仞仭仟价伉佚估佛佝佗佇佶侈侏侘佻佩佰侑佯來侖儘俔俟俎俘俛俑俚俐俤俥倚倨倔倪倥倅伜俶倡倩倬俾俯們倆偃假會偕偐偈做偖偬偸傀傚傅傴傲"],
        ["d1a1", "僉僊傳僂僖僞僥僭僣僮價僵儉儁儂儖儕儔儚儡儺儷儼儻儿兀兒兌兔兢竸兩兪兮冀冂囘册冉冏冑冓冕冖冤冦冢冩冪冫决冱冲冰况冽凅凉凛几處凩凭凰凵凾刄刋刔刎刧刪刮刳刹剏剄剋剌剞剔剪剴剩剳剿剽劍劔劒剱劈劑辨"],
        ["d2a1", "辧劬劭劼劵勁勍勗勞勣勦飭勠勳勵勸勹匆匈甸匍匐匏匕匚匣匯匱匳匸區卆卅丗卉卍凖卞卩卮夘卻卷厂厖厠厦厥厮厰厶參簒雙叟曼燮叮叨叭叺吁吽呀听吭吼吮吶吩吝呎咏呵咎呟呱呷呰咒呻咀呶咄咐咆哇咢咸咥咬哄哈咨"],
        ["d3a1", "咫哂咤咾咼哘哥哦唏唔哽哮哭哺哢唹啀啣啌售啜啅啖啗唸唳啝喙喀咯喊喟啻啾喘喞單啼喃喩喇喨嗚嗅嗟嗄嗜嗤嗔嘔嗷嘖嗾嗽嘛嗹噎噐營嘴嘶嘲嘸噫噤嘯噬噪嚆嚀嚊嚠嚔嚏嚥嚮嚶嚴囂嚼囁囃囀囈囎囑囓囗囮囹圀囿圄圉"],
        ["d4a1", "圈國圍圓團圖嗇圜圦圷圸坎圻址坏坩埀垈坡坿垉垓垠垳垤垪垰埃埆埔埒埓堊埖埣堋堙堝塲堡塢塋塰毀塒堽塹墅墹墟墫墺壞墻墸墮壅壓壑壗壙壘壥壜壤壟壯壺壹壻壼壽夂夊夐夛梦夥夬夭夲夸夾竒奕奐奎奚奘奢奠奧奬奩"],
        ["d5a1", "奸妁妝佞侫妣妲姆姨姜妍姙姚娥娟娑娜娉娚婀婬婉娵娶婢婪媚媼媾嫋嫂媽嫣嫗嫦嫩嫖嫺嫻嬌嬋嬖嬲嫐嬪嬶嬾孃孅孀孑孕孚孛孥孩孰孳孵學斈孺宀它宦宸寃寇寉寔寐寤實寢寞寥寫寰寶寳尅將專對尓尠尢尨尸尹屁屆屎屓"],
        ["d6a1", "屐屏孱屬屮乢屶屹岌岑岔妛岫岻岶岼岷峅岾峇峙峩峽峺峭嶌峪崋崕崗嵜崟崛崑崔崢崚崙崘嵌嵒嵎嵋嵬嵳嵶嶇嶄嶂嶢嶝嶬嶮嶽嶐嶷嶼巉巍巓巒巖巛巫已巵帋帚帙帑帛帶帷幄幃幀幎幗幔幟幢幤幇幵并幺麼广庠廁廂廈廐廏"],
        ["d7a1", "廖廣廝廚廛廢廡廨廩廬廱廳廰廴廸廾弃弉彝彜弋弑弖弩弭弸彁彈彌彎弯彑彖彗彙彡彭彳彷徃徂彿徊很徑徇從徙徘徠徨徭徼忖忻忤忸忱忝悳忿怡恠怙怐怩怎怱怛怕怫怦怏怺恚恁恪恷恟恊恆恍恣恃恤恂恬恫恙悁悍惧悃悚"],
        ["d8a1", "悄悛悖悗悒悧悋惡悸惠惓悴忰悽惆悵惘慍愕愆惶惷愀惴惺愃愡惻惱愍愎慇愾愨愧慊愿愼愬愴愽慂慄慳慷慘慙慚慫慴慯慥慱慟慝慓慵憙憖憇憬憔憚憊憑憫憮懌懊應懷懈懃懆憺懋罹懍懦懣懶懺懴懿懽懼懾戀戈戉戍戌戔戛"],
        ["d9a1", "戞戡截戮戰戲戳扁扎扞扣扛扠扨扼抂抉找抒抓抖拔抃抔拗拑抻拏拿拆擔拈拜拌拊拂拇抛拉挌拮拱挧挂挈拯拵捐挾捍搜捏掖掎掀掫捶掣掏掉掟掵捫捩掾揩揀揆揣揉插揶揄搖搴搆搓搦搶攝搗搨搏摧摯摶摎攪撕撓撥撩撈撼"],
        ["daa1", "據擒擅擇撻擘擂擱擧舉擠擡抬擣擯攬擶擴擲擺攀擽攘攜攅攤攣攫攴攵攷收攸畋效敖敕敍敘敞敝敲數斂斃變斛斟斫斷旃旆旁旄旌旒旛旙无旡旱杲昊昃旻杳昵昶昴昜晏晄晉晁晞晝晤晧晨晟晢晰暃暈暎暉暄暘暝曁暹曉暾暼"],
        ["dba1", "曄暸曖曚曠昿曦曩曰曵曷朏朖朞朦朧霸朮朿朶杁朸朷杆杞杠杙杣杤枉杰枩杼杪枌枋枦枡枅枷柯枴柬枳柩枸柤柞柝柢柮枹柎柆柧檜栞框栩桀桍栲桎梳栫桙档桷桿梟梏梭梔條梛梃檮梹桴梵梠梺椏梍桾椁棊椈棘椢椦棡椌棍"],
        ["dca1", "棔棧棕椶椒椄棗棣椥棹棠棯椨椪椚椣椡棆楹楷楜楸楫楔楾楮椹楴椽楙椰楡楞楝榁楪榲榮槐榿槁槓榾槎寨槊槝榻槃榧樮榑榠榜榕榴槞槨樂樛槿權槹槲槧樅榱樞槭樔槫樊樒櫁樣樓橄樌橲樶橸橇橢橙橦橈樸樢檐檍檠檄檢檣"],
        ["dda1", "檗蘗檻櫃櫂檸檳檬櫞櫑櫟檪櫚櫪櫻欅蘖櫺欒欖鬱欟欸欷盜欹飮歇歃歉歐歙歔歛歟歡歸歹歿殀殄殃殍殘殕殞殤殪殫殯殲殱殳殷殼毆毋毓毟毬毫毳毯麾氈氓气氛氤氣汞汕汢汪沂沍沚沁沛汾汨汳沒沐泄泱泓沽泗泅泝沮沱沾"],
        ["dea1", "沺泛泯泙泪洟衍洶洫洽洸洙洵洳洒洌浣涓浤浚浹浙涎涕濤涅淹渕渊涵淇淦涸淆淬淞淌淨淒淅淺淙淤淕淪淮渭湮渮渙湲湟渾渣湫渫湶湍渟湃渺湎渤滿渝游溂溪溘滉溷滓溽溯滄溲滔滕溏溥滂溟潁漑灌滬滸滾漿滲漱滯漲滌"],
        ["dfa1", "漾漓滷澆潺潸澁澀潯潛濳潭澂潼潘澎澑濂潦澳澣澡澤澹濆澪濟濕濬濔濘濱濮濛瀉瀋濺瀑瀁瀏濾瀛瀚潴瀝瀘瀟瀰瀾瀲灑灣炙炒炯烱炬炸炳炮烟烋烝烙焉烽焜焙煥煕熈煦煢煌煖煬熏燻熄熕熨熬燗熹熾燒燉燔燎燠燬燧燵燼"],
        ["e0a1", "燹燿爍爐爛爨爭爬爰爲爻爼爿牀牆牋牘牴牾犂犁犇犒犖犢犧犹犲狃狆狄狎狒狢狠狡狹狷倏猗猊猜猖猝猴猯猩猥猾獎獏默獗獪獨獰獸獵獻獺珈玳珎玻珀珥珮珞璢琅瑯琥珸琲琺瑕琿瑟瑙瑁瑜瑩瑰瑣瑪瑶瑾璋璞璧瓊瓏瓔珱"],
        ["e1a1", "瓠瓣瓧瓩瓮瓲瓰瓱瓸瓷甄甃甅甌甎甍甕甓甞甦甬甼畄畍畊畉畛畆畚畩畤畧畫畭畸當疆疇畴疊疉疂疔疚疝疥疣痂疳痃疵疽疸疼疱痍痊痒痙痣痞痾痿痼瘁痰痺痲痳瘋瘍瘉瘟瘧瘠瘡瘢瘤瘴瘰瘻癇癈癆癜癘癡癢癨癩癪癧癬癰"],
        ["e2a1", "癲癶癸發皀皃皈皋皎皖皓皙皚皰皴皸皹皺盂盍盖盒盞盡盥盧盪蘯盻眈眇眄眩眤眞眥眦眛眷眸睇睚睨睫睛睥睿睾睹瞎瞋瞑瞠瞞瞰瞶瞹瞿瞼瞽瞻矇矍矗矚矜矣矮矼砌砒礦砠礪硅碎硴碆硼碚碌碣碵碪碯磑磆磋磔碾碼磅磊磬"],
        ["e3a1", "磧磚磽磴礇礒礑礙礬礫祀祠祗祟祚祕祓祺祿禊禝禧齋禪禮禳禹禺秉秕秧秬秡秣稈稍稘稙稠稟禀稱稻稾稷穃穗穉穡穢穩龝穰穹穽窈窗窕窘窖窩竈窰窶竅竄窿邃竇竊竍竏竕竓站竚竝竡竢竦竭竰笂笏笊笆笳笘笙笞笵笨笶筐"],
        ["e4a1", "筺笄筍笋筌筅筵筥筴筧筰筱筬筮箝箘箟箍箜箚箋箒箏筝箙篋篁篌篏箴篆篝篩簑簔篦篥籠簀簇簓篳篷簗簍篶簣簧簪簟簷簫簽籌籃籔籏籀籐籘籟籤籖籥籬籵粃粐粤粭粢粫粡粨粳粲粱粮粹粽糀糅糂糘糒糜糢鬻糯糲糴糶糺紆"],
        ["e5a1", "紂紜紕紊絅絋紮紲紿紵絆絳絖絎絲絨絮絏絣經綉絛綏絽綛綺綮綣綵緇綽綫總綢綯緜綸綟綰緘緝緤緞緻緲緡縅縊縣縡縒縱縟縉縋縢繆繦縻縵縹繃縷縲縺繧繝繖繞繙繚繹繪繩繼繻纃緕繽辮繿纈纉續纒纐纓纔纖纎纛纜缸缺"],
        ["e6a1", "罅罌罍罎罐网罕罔罘罟罠罨罩罧罸羂羆羃羈羇羌羔羞羝羚羣羯羲羹羮羶羸譱翅翆翊翕翔翡翦翩翳翹飜耆耄耋耒耘耙耜耡耨耿耻聊聆聒聘聚聟聢聨聳聲聰聶聹聽聿肄肆肅肛肓肚肭冐肬胛胥胙胝胄胚胖脉胯胱脛脩脣脯腋"],
        ["e7a1", "隋腆脾腓腑胼腱腮腥腦腴膃膈膊膀膂膠膕膤膣腟膓膩膰膵膾膸膽臀臂膺臉臍臑臙臘臈臚臟臠臧臺臻臾舁舂舅與舊舍舐舖舩舫舸舳艀艙艘艝艚艟艤艢艨艪艫舮艱艷艸艾芍芒芫芟芻芬苡苣苟苒苴苳苺莓范苻苹苞茆苜茉苙"],
        ["e8a1", "茵茴茖茲茱荀茹荐荅茯茫茗茘莅莚莪莟莢莖茣莎莇莊荼莵荳荵莠莉莨菴萓菫菎菽萃菘萋菁菷萇菠菲萍萢萠莽萸蔆菻葭萪萼蕚蒄葷葫蒭葮蒂葩葆萬葯葹萵蓊葢蒹蒿蒟蓙蓍蒻蓚蓐蓁蓆蓖蒡蔡蓿蓴蔗蔘蔬蔟蔕蔔蓼蕀蕣蕘蕈"],
        ["e9a1", "蕁蘂蕋蕕薀薤薈薑薊薨蕭薔薛藪薇薜蕷蕾薐藉薺藏薹藐藕藝藥藜藹蘊蘓蘋藾藺蘆蘢蘚蘰蘿虍乕虔號虧虱蚓蚣蚩蚪蚋蚌蚶蚯蛄蛆蚰蛉蠣蚫蛔蛞蛩蛬蛟蛛蛯蜒蜆蜈蜀蜃蛻蜑蜉蜍蛹蜊蜴蜿蜷蜻蜥蜩蜚蝠蝟蝸蝌蝎蝴蝗蝨蝮蝙"],
        ["eaa1", "蝓蝣蝪蠅螢螟螂螯蟋螽蟀蟐雖螫蟄螳蟇蟆螻蟯蟲蟠蠏蠍蟾蟶蟷蠎蟒蠑蠖蠕蠢蠡蠱蠶蠹蠧蠻衄衂衒衙衞衢衫袁衾袞衵衽袵衲袂袗袒袮袙袢袍袤袰袿袱裃裄裔裘裙裝裹褂裼裴裨裲褄褌褊褓襃褞褥褪褫襁襄褻褶褸襌褝襠襞"],
        ["eba1", "襦襤襭襪襯襴襷襾覃覈覊覓覘覡覩覦覬覯覲覺覽覿觀觚觜觝觧觴觸訃訖訐訌訛訝訥訶詁詛詒詆詈詼詭詬詢誅誂誄誨誡誑誥誦誚誣諄諍諂諚諫諳諧諤諱謔諠諢諷諞諛謌謇謚諡謖謐謗謠謳鞫謦謫謾謨譁譌譏譎證譖譛譚譫"],
        ["eca1", "譟譬譯譴譽讀讌讎讒讓讖讙讚谺豁谿豈豌豎豐豕豢豬豸豺貂貉貅貊貍貎貔豼貘戝貭貪貽貲貳貮貶賈賁賤賣賚賽賺賻贄贅贊贇贏贍贐齎贓賍贔贖赧赭赱赳趁趙跂趾趺跏跚跖跌跛跋跪跫跟跣跼踈踉跿踝踞踐踟蹂踵踰踴蹊"],
        ["eda1", "蹇蹉蹌蹐蹈蹙蹤蹠踪蹣蹕蹶蹲蹼躁躇躅躄躋躊躓躑躔躙躪躡躬躰軆躱躾軅軈軋軛軣軼軻軫軾輊輅輕輒輙輓輜輟輛輌輦輳輻輹轅轂輾轌轉轆轎轗轜轢轣轤辜辟辣辭辯辷迚迥迢迪迯邇迴逅迹迺逑逕逡逍逞逖逋逧逶逵逹迸"],
        ["eea1", "遏遐遑遒逎遉逾遖遘遞遨遯遶隨遲邂遽邁邀邊邉邏邨邯邱邵郢郤扈郛鄂鄒鄙鄲鄰酊酖酘酣酥酩酳酲醋醉醂醢醫醯醪醵醴醺釀釁釉釋釐釖釟釡釛釼釵釶鈞釿鈔鈬鈕鈑鉞鉗鉅鉉鉤鉈銕鈿鉋鉐銜銖銓銛鉚鋏銹銷鋩錏鋺鍄錮"],
        ["efa1", "錙錢錚錣錺錵錻鍜鍠鍼鍮鍖鎰鎬鎭鎔鎹鏖鏗鏨鏥鏘鏃鏝鏐鏈鏤鐚鐔鐓鐃鐇鐐鐶鐫鐵鐡鐺鑁鑒鑄鑛鑠鑢鑞鑪鈩鑰鑵鑷鑽鑚鑼鑾钁鑿閂閇閊閔閖閘閙閠閨閧閭閼閻閹閾闊濶闃闍闌闕闔闖關闡闥闢阡阨阮阯陂陌陏陋陷陜陞"],
        ["f0a1", "陝陟陦陲陬隍隘隕隗險隧隱隲隰隴隶隸隹雎雋雉雍襍雜霍雕雹霄霆霈霓霎霑霏霖霙霤霪霰霹霽霾靄靆靈靂靉靜靠靤靦靨勒靫靱靹鞅靼鞁靺鞆鞋鞏鞐鞜鞨鞦鞣鞳鞴韃韆韈韋韜韭齏韲竟韶韵頏頌頸頤頡頷頽顆顏顋顫顯顰"],
        ["f1a1", "顱顴顳颪颯颱颶飄飃飆飩飫餃餉餒餔餘餡餝餞餤餠餬餮餽餾饂饉饅饐饋饑饒饌饕馗馘馥馭馮馼駟駛駝駘駑駭駮駱駲駻駸騁騏騅駢騙騫騷驅驂驀驃騾驕驍驛驗驟驢驥驤驩驫驪骭骰骼髀髏髑髓體髞髟髢髣髦髯髫髮髴髱髷"],
        ["f2a1", "髻鬆鬘鬚鬟鬢鬣鬥鬧鬨鬩鬪鬮鬯鬲魄魃魏魍魎魑魘魴鮓鮃鮑鮖鮗鮟鮠鮨鮴鯀鯊鮹鯆鯏鯑鯒鯣鯢鯤鯔鯡鰺鯲鯱鯰鰕鰔鰉鰓鰌鰆鰈鰒鰊鰄鰮鰛鰥鰤鰡鰰鱇鰲鱆鰾鱚鱠鱧鱶鱸鳧鳬鳰鴉鴈鳫鴃鴆鴪鴦鶯鴣鴟鵄鴕鴒鵁鴿鴾鵆鵈"],
        ["f3a1", "鵝鵞鵤鵑鵐鵙鵲鶉鶇鶫鵯鵺鶚鶤鶩鶲鷄鷁鶻鶸鶺鷆鷏鷂鷙鷓鷸鷦鷭鷯鷽鸚鸛鸞鹵鹹鹽麁麈麋麌麒麕麑麝麥麩麸麪麭靡黌黎黏黐黔黜點黝黠黥黨黯黴黶黷黹黻黼黽鼇鼈皷鼕鼡鼬鼾齊齒齔齣齟齠齡齦齧齬齪齷齲齶龕龜龠"],
        ["f4a1", "堯槇遙瑤凜熙"],
        ["f9a1", "纊褜鍈銈蓜俉炻昱棈鋹曻彅丨仡仼伀伃伹佖侒侊侚侔俍偀倢俿倞偆偰偂傔僴僘兊兤冝冾凬刕劜劦勀勛匀匇匤卲厓厲叝﨎咜咊咩哿喆坙坥垬埈埇﨏塚增墲夋奓奛奝奣妤妺孖寀甯寘寬尞岦岺峵崧嵓﨑嵂嵭嶸嶹巐弡弴彧德"],
        ["faa1", "忞恝悅悊惞惕愠惲愑愷愰憘戓抦揵摠撝擎敎昀昕昻昉昮昞昤晥晗晙晴晳暙暠暲暿曺朎朗杦枻桒柀栁桄棏﨓楨﨔榘槢樰橫橆橳橾櫢櫤毖氿汜沆汯泚洄涇浯涖涬淏淸淲淼渹湜渧渼溿澈澵濵瀅瀇瀨炅炫焏焄煜煆煇凞燁燾犱"],
        ["fba1", "犾猤猪獷玽珉珖珣珒琇珵琦琪琩琮瑢璉璟甁畯皂皜皞皛皦益睆劯砡硎硤硺礰礼神祥禔福禛竑竧靖竫箞精絈絜綷綠緖繒罇羡羽茁荢荿菇菶葈蒴蕓蕙蕫﨟薰蘒﨡蠇裵訒訷詹誧誾諟諸諶譓譿賰賴贒赶﨣軏﨤逸遧郞都鄕鄧釚"],
        ["fca1", "釗釞釭釮釤釥鈆鈐鈊鈺鉀鈼鉎鉙鉑鈹鉧銧鉷鉸鋧鋗鋙鋐﨧鋕鋠鋓錥錡鋻﨨錞鋿錝錂鍰鍗鎤鏆鏞鏸鐱鑅鑈閒隆﨩隝隯霳霻靃靍靏靑靕顗顥飯飼餧館馞驎髙髜魵魲鮏鮱鮻鰀鵰鵫鶴鸙黑"],
        ["fcf1", "ⅰ", 9, "￢￤＇＂"],
        ["8fa2af", "˘ˇ¸˙˝¯˛˚～΄΅"],
        ["8fa2c2", "¡¦¿"],
        ["8fa2eb", "ºª©®™¤№"],
        ["8fa6e1", "ΆΈΉΊΪ"],
        ["8fa6e7", "Ό"],
        ["8fa6e9", "ΎΫ"],
        ["8fa6ec", "Ώ"],
        ["8fa6f1", "άέήίϊΐόςύϋΰώ"],
        ["8fa7c2", "Ђ", 10, "ЎЏ"],
        ["8fa7f2", "ђ", 10, "ўџ"],
        ["8fa9a1", "ÆĐ"],
        ["8fa9a4", "Ħ"],
        ["8fa9a6", "Ĳ"],
        ["8fa9a8", "ŁĿ"],
        ["8fa9ab", "ŊØŒ"],
        ["8fa9af", "ŦÞ"],
        ["8fa9c1", "æđðħıĳĸłŀŉŋøœßŧþ"],
        ["8faaa1", "ÁÀÄÂĂǍĀĄÅÃĆĈČÇĊĎÉÈËÊĚĖĒĘ"],
        ["8faaba", "ĜĞĢĠĤÍÌÏÎǏİĪĮĨĴĶĹĽĻŃŇŅÑÓÒÖÔǑŐŌÕŔŘŖŚŜŠŞŤŢÚÙÜÛŬǓŰŪŲŮŨǗǛǙǕŴÝŸŶŹŽŻ"],
        ["8faba1", "áàäâăǎāąåãćĉčçċďéèëêěėēęǵĝğ"],
        ["8fabbd", "ġĥíìïîǐ"],
        ["8fabc5", "īįĩĵķĺľļńňņñóòöôǒőōõŕřŗśŝšşťţúùüûŭǔűūųůũǘǜǚǖŵýÿŷźžż"],
        ["8fb0a1", "丂丄丅丌丒丟丣两丨丫丮丯丰丵乀乁乄乇乑乚乜乣乨乩乴乵乹乿亍亖亗亝亯亹仃仐仚仛仠仡仢仨仯仱仳仵份仾仿伀伂伃伈伋伌伒伕伖众伙伮伱你伳伵伷伹伻伾佀佂佈佉佋佌佒佔佖佘佟佣佪佬佮佱佷佸佹佺佽佾侁侂侄"],
        ["8fb1a1", "侅侉侊侌侎侐侒侓侔侗侙侚侞侟侲侷侹侻侼侽侾俀俁俅俆俈俉俋俌俍俏俒俜俠俢俰俲俼俽俿倀倁倄倇倊倌倎倐倓倗倘倛倜倝倞倢倧倮倰倲倳倵偀偁偂偅偆偊偌偎偑偒偓偗偙偟偠偢偣偦偧偪偭偰偱倻傁傃傄傆傊傎傏傐"],
        ["8fb2a1", "傒傓傔傖傛傜傞", 4, "傪傯傰傹傺傽僀僃僄僇僌僎僐僓僔僘僜僝僟僢僤僦僨僩僯僱僶僺僾儃儆儇儈儋儌儍儎僲儐儗儙儛儜儝儞儣儧儨儬儭儯儱儳儴儵儸儹兂兊兏兓兕兗兘兟兤兦兾冃冄冋冎冘冝冡冣冭冸冺冼冾冿凂"],
        ["8fb3a1", "凈减凑凒凓凕凘凞凢凥凮凲凳凴凷刁刂刅划刓刕刖刘刢刨刱刲刵刼剅剉剕剗剘剚剜剟剠剡剦剮剷剸剹劀劂劅劊劌劓劕劖劗劘劚劜劤劥劦劧劯劰劶劷劸劺劻劽勀勄勆勈勌勏勑勔勖勛勜勡勥勨勩勪勬勰勱勴勶勷匀匃匊匋"],
        ["8fb4a1", "匌匑匓匘匛匜匞匟匥匧匨匩匫匬匭匰匲匵匼匽匾卂卌卋卙卛卡卣卥卬卭卲卹卾厃厇厈厎厓厔厙厝厡厤厪厫厯厲厴厵厷厸厺厽叀叅叏叒叓叕叚叝叞叠另叧叵吂吓吚吡吧吨吪启吱吴吵呃呄呇呍呏呞呢呤呦呧呩呫呭呮呴呿"],
        ["8fb5a1", "咁咃咅咈咉咍咑咕咖咜咟咡咦咧咩咪咭咮咱咷咹咺咻咿哆哊响哎哠哪哬哯哶哼哾哿唀唁唅唈唉唌唍唎唕唪唫唲唵唶唻唼唽啁啇啉啊啍啐啑啘啚啛啞啠啡啤啦啿喁喂喆喈喎喏喑喒喓喔喗喣喤喭喲喿嗁嗃嗆嗉嗋嗌嗎嗑嗒"],
        ["8fb6a1", "嗓嗗嗘嗛嗞嗢嗩嗶嗿嘅嘈嘊嘍", 5, "嘙嘬嘰嘳嘵嘷嘹嘻嘼嘽嘿噀噁噃噄噆噉噋噍噏噔噞噠噡噢噣噦噩噭噯噱噲噵嚄嚅嚈嚋嚌嚕嚙嚚嚝嚞嚟嚦嚧嚨嚩嚫嚬嚭嚱嚳嚷嚾囅囉囊囋囏囐囌囍囙囜囝囟囡囤", 4, "囱囫园"],
        ["8fb7a1", "囶囷圁圂圇圊圌圑圕圚圛圝圠圢圣圤圥圩圪圬圮圯圳圴圽圾圿坅坆坌坍坒坢坥坧坨坫坭", 4, "坳坴坵坷坹坺坻坼坾垁垃垌垔垗垙垚垜垝垞垟垡垕垧垨垩垬垸垽埇埈埌埏埕埝埞埤埦埧埩埭埰埵埶埸埽埾埿堃堄堈堉埡"],
        ["8fb8a1", "堌堍堛堞堟堠堦堧堭堲堹堿塉塌塍塏塐塕塟塡塤塧塨塸塼塿墀墁墇墈墉墊墌墍墏墐墔墖墝墠墡墢墦墩墱墲壄墼壂壈壍壎壐壒壔壖壚壝壡壢壩壳夅夆夋夌夒夓夔虁夝夡夣夤夨夯夰夳夵夶夿奃奆奒奓奙奛奝奞奟奡奣奫奭"],
        ["8fb9a1", "奯奲奵奶她奻奼妋妌妎妒妕妗妟妤妧妭妮妯妰妳妷妺妼姁姃姄姈姊姍姒姝姞姟姣姤姧姮姯姱姲姴姷娀娄娌娍娎娒娓娞娣娤娧娨娪娭娰婄婅婇婈婌婐婕婞婣婥婧婭婷婺婻婾媋媐媓媖媙媜媞媟媠媢媧媬媱媲媳媵媸媺媻媿"],
        ["8fbaa1", "嫄嫆嫈嫏嫚嫜嫠嫥嫪嫮嫵嫶嫽嬀嬁嬈嬗嬴嬙嬛嬝嬡嬥嬭嬸孁孋孌孒孖孞孨孮孯孼孽孾孿宁宄宆宊宎宐宑宓宔宖宨宩宬宭宯宱宲宷宺宼寀寁寍寏寖", 4, "寠寯寱寴寽尌尗尞尟尣尦尩尫尬尮尰尲尵尶屙屚屜屢屣屧屨屩"],
        ["8fbba1", "屭屰屴屵屺屻屼屽岇岈岊岏岒岝岟岠岢岣岦岪岲岴岵岺峉峋峒峝峗峮峱峲峴崁崆崍崒崫崣崤崦崧崱崴崹崽崿嵂嵃嵆嵈嵕嵑嵙嵊嵟嵠嵡嵢嵤嵪嵭嵰嵹嵺嵾嵿嶁嶃嶈嶊嶒嶓嶔嶕嶙嶛嶟嶠嶧嶫嶰嶴嶸嶹巃巇巋巐巎巘巙巠巤"],
        ["8fbca1", "巩巸巹帀帇帍帒帔帕帘帟帠帮帨帲帵帾幋幐幉幑幖幘幛幜幞幨幪", 4, "幰庀庋庎庢庤庥庨庪庬庱庳庽庾庿廆廌廋廎廑廒廔廕廜廞廥廫异弆弇弈弎弙弜弝弡弢弣弤弨弫弬弮弰弴弶弻弽弿彀彄彅彇彍彐彔彘彛彠彣彤彧"],
        ["8fbda1", "彯彲彴彵彸彺彽彾徉徍徏徖徜徝徢徧徫徤徬徯徰徱徸忄忇忈忉忋忐", 4, "忞忡忢忨忩忪忬忭忮忯忲忳忶忺忼怇怊怍怓怔怗怘怚怟怤怭怳怵恀恇恈恉恌恑恔恖恗恝恡恧恱恾恿悂悆悈悊悎悑悓悕悘悝悞悢悤悥您悰悱悷"],
        ["8fbea1", "悻悾惂惄惈惉惊惋惎惏惔惕惙惛惝惞惢惥惲惵惸惼惽愂愇愊愌愐", 4, "愖愗愙愜愞愢愪愫愰愱愵愶愷愹慁慅慆慉慞慠慬慲慸慻慼慿憀憁憃憄憋憍憒憓憗憘憜憝憟憠憥憨憪憭憸憹憼懀懁懂懎懏懕懜懝懞懟懡懢懧懩懥"],
        ["8fbfa1", "懬懭懯戁戃戄戇戓戕戜戠戢戣戧戩戫戹戽扂扃扄扆扌扐扑扒扔扖扚扜扤扭扯扳扺扽抍抎抏抐抦抨抳抶抷抺抾抿拄拎拕拖拚拪拲拴拼拽挃挄挊挋挍挐挓挖挘挩挪挭挵挶挹挼捁捂捃捄捆捊捋捎捒捓捔捘捛捥捦捬捭捱捴捵"],
        ["8fc0a1", "捸捼捽捿掂掄掇掊掐掔掕掙掚掞掤掦掭掮掯掽揁揅揈揎揑揓揔揕揜揠揥揪揬揲揳揵揸揹搉搊搐搒搔搘搞搠搢搤搥搩搪搯搰搵搽搿摋摏摑摒摓摔摚摛摜摝摟摠摡摣摭摳摴摻摽撅撇撏撐撑撘撙撛撝撟撡撣撦撨撬撳撽撾撿"],
        ["8fc1a1", "擄擉擊擋擌擎擐擑擕擗擤擥擩擪擭擰擵擷擻擿攁攄攈攉攊攏攓攔攖攙攛攞攟攢攦攩攮攱攺攼攽敃敇敉敐敒敔敟敠敧敫敺敽斁斅斊斒斕斘斝斠斣斦斮斲斳斴斿旂旈旉旎旐旔旖旘旟旰旲旴旵旹旾旿昀昄昈昉昍昑昒昕昖昝"],
        ["8fc2a1", "昞昡昢昣昤昦昩昪昫昬昮昰昱昳昹昷晀晅晆晊晌晑晎晗晘晙晛晜晠晡曻晪晫晬晾晳晵晿晷晸晹晻暀晼暋暌暍暐暒暙暚暛暜暟暠暤暭暱暲暵暻暿曀曂曃曈曌曎曏曔曛曟曨曫曬曮曺朅朇朎朓朙朜朠朢朳朾杅杇杈杌杔杕杝"],
        ["8fc3a1", "杦杬杮杴杶杻极构枎枏枑枓枖枘枙枛枰枱枲枵枻枼枽柹柀柂柃柅柈柉柒柗柙柜柡柦柰柲柶柷桒栔栙栝栟栨栧栬栭栯栰栱栳栻栿桄桅桊桌桕桗桘桛桫桮", 4, "桵桹桺桻桼梂梄梆梈梖梘梚梜梡梣梥梩梪梮梲梻棅棈棌棏"],
        ["8fc4a1", "棐棑棓棖棙棜棝棥棨棪棫棬棭棰棱棵棶棻棼棽椆椉椊椐椑椓椖椗椱椳椵椸椻楂楅楉楎楗楛楣楤楥楦楨楩楬楰楱楲楺楻楿榀榍榒榖榘榡榥榦榨榫榭榯榷榸榺榼槅槈槑槖槗槢槥槮槯槱槳槵槾樀樁樃樏樑樕樚樝樠樤樨樰樲"],
        ["8fc5a1", "樴樷樻樾樿橅橆橉橊橎橐橑橒橕橖橛橤橧橪橱橳橾檁檃檆檇檉檋檑檛檝檞檟檥檫檯檰檱檴檽檾檿櫆櫉櫈櫌櫐櫔櫕櫖櫜櫝櫤櫧櫬櫰櫱櫲櫼櫽欂欃欆欇欉欏欐欑欗欛欞欤欨欫欬欯欵欶欻欿歆歊歍歒歖歘歝歠歧歫歮歰歵歽"],
        ["8fc6a1", "歾殂殅殗殛殟殠殢殣殨殩殬殭殮殰殸殹殽殾毃毄毉毌毖毚毡毣毦毧毮毱毷毹毿氂氄氅氉氍氎氐氒氙氟氦氧氨氬氮氳氵氶氺氻氿汊汋汍汏汒汔汙汛汜汫汭汯汴汶汸汹汻沅沆沇沉沔沕沗沘沜沟沰沲沴泂泆泍泏泐泑泒泔泖"],
        ["8fc7a1", "泚泜泠泧泩泫泬泮泲泴洄洇洊洎洏洑洓洚洦洧洨汧洮洯洱洹洼洿浗浞浟浡浥浧浯浰浼涂涇涑涒涔涖涗涘涪涬涴涷涹涽涿淄淈淊淎淏淖淛淝淟淠淢淥淩淯淰淴淶淼渀渄渞渢渧渲渶渹渻渼湄湅湈湉湋湏湑湒湓湔湗湜湝湞"],
        ["8fc8a1", "湢湣湨湳湻湽溍溓溙溠溧溭溮溱溳溻溿滀滁滃滇滈滊滍滎滏滫滭滮滹滻滽漄漈漊漌漍漖漘漚漛漦漩漪漯漰漳漶漻漼漭潏潑潒潓潗潙潚潝潞潡潢潨潬潽潾澃澇澈澋澌澍澐澒澓澔澖澚澟澠澥澦澧澨澮澯澰澵澶澼濅濇濈濊"],
        ["8fc9a1", "濚濞濨濩濰濵濹濼濽瀀瀅瀆瀇瀍瀗瀠瀣瀯瀴瀷瀹瀼灃灄灈灉灊灋灔灕灝灞灎灤灥灬灮灵灶灾炁炅炆炔", 4, "炛炤炫炰炱炴炷烊烑烓烔烕烖烘烜烤烺焃", 4, "焋焌焏焞焠焫焭焯焰焱焸煁煅煆煇煊煋煐煒煗煚煜煞煠"],
        ["8fcaa1", "煨煹熀熅熇熌熒熚熛熠熢熯熰熲熳熺熿燀燁燄燋燌燓燖燙燚燜燸燾爀爇爈爉爓爗爚爝爟爤爫爯爴爸爹牁牂牃牅牎牏牐牓牕牖牚牜牞牠牣牨牫牮牯牱牷牸牻牼牿犄犉犍犎犓犛犨犭犮犱犴犾狁狇狉狌狕狖狘狟狥狳狴狺狻"],
        ["8fcba1", "狾猂猄猅猇猋猍猒猓猘猙猞猢猤猧猨猬猱猲猵猺猻猽獃獍獐獒獖獘獝獞獟獠獦獧獩獫獬獮獯獱獷獹獼玀玁玃玅玆玎玐玓玕玗玘玜玞玟玠玢玥玦玪玫玭玵玷玹玼玽玿珅珆珉珋珌珏珒珓珖珙珝珡珣珦珧珩珴珵珷珹珺珻珽"],
        ["8fcca1", "珿琀琁琄琇琊琑琚琛琤琦琨", 9, "琹瑀瑃瑄瑆瑇瑋瑍瑑瑒瑗瑝瑢瑦瑧瑨瑫瑭瑮瑱瑲璀璁璅璆璇璉璏璐璑璒璘璙璚璜璟璠璡璣璦璨璩璪璫璮璯璱璲璵璹璻璿瓈瓉瓌瓐瓓瓘瓚瓛瓞瓟瓤瓨瓪瓫瓯瓴瓺瓻瓼瓿甆"],
        ["8fcda1", "甒甖甗甠甡甤甧甩甪甯甶甹甽甾甿畀畃畇畈畎畐畒畗畞畟畡畯畱畹", 5, "疁疅疐疒疓疕疙疜疢疤疴疺疿痀痁痄痆痌痎痏痗痜痟痠痡痤痧痬痮痯痱痹瘀瘂瘃瘄瘇瘈瘊瘌瘏瘒瘓瘕瘖瘙瘛瘜瘝瘞瘣瘥瘦瘩瘭瘲瘳瘵瘸瘹"],
        ["8fcea1", "瘺瘼癊癀癁癃癄癅癉癋癕癙癟癤癥癭癮癯癱癴皁皅皌皍皕皛皜皝皟皠皢", 6, "皪皭皽盁盅盉盋盌盎盔盙盠盦盨盬盰盱盶盹盼眀眆眊眎眒眔眕眗眙眚眜眢眨眭眮眯眴眵眶眹眽眾睂睅睆睊睍睎睏睒睖睗睜睞睟睠睢"],
        ["8fcfa1", "睤睧睪睬睰睲睳睴睺睽瞀瞄瞌瞍瞔瞕瞖瞚瞟瞢瞧瞪瞮瞯瞱瞵瞾矃矉矑矒矕矙矞矟矠矤矦矪矬矰矱矴矸矻砅砆砉砍砎砑砝砡砢砣砭砮砰砵砷硃硄硇硈硌硎硒硜硞硠硡硣硤硨硪确硺硾碊碏碔碘碡碝碞碟碤碨碬碭碰碱碲碳"],
        ["8fd0a1", "碻碽碿磇磈磉磌磎磒磓磕磖磤磛磟磠磡磦磪磲磳礀磶磷磺磻磿礆礌礐礚礜礞礟礠礥礧礩礭礱礴礵礻礽礿祄祅祆祊祋祏祑祔祘祛祜祧祩祫祲祹祻祼祾禋禌禑禓禔禕禖禘禛禜禡禨禩禫禯禱禴禸离秂秄秇秈秊秏秔秖秚秝秞"],
        ["8fd1a1", "秠秢秥秪秫秭秱秸秼稂稃稇稉稊稌稑稕稛稞稡稧稫稭稯稰稴稵稸稹稺穄穅穇穈穌穕穖穙穜穝穟穠穥穧穪穭穵穸穾窀窂窅窆窊窋窐窑窔窞窠窣窬窳窵窹窻窼竆竉竌竎竑竛竨竩竫竬竱竴竻竽竾笇笔笟笣笧笩笪笫笭笮笯笰"],
        ["8fd2a1", "笱笴笽笿筀筁筇筎筕筠筤筦筩筪筭筯筲筳筷箄箉箎箐箑箖箛箞箠箥箬箯箰箲箵箶箺箻箼箽篂篅篈篊篔篖篗篙篚篛篨篪篲篴篵篸篹篺篼篾簁簂簃簄簆簉簋簌簎簏簙簛簠簥簦簨簬簱簳簴簶簹簺籆籊籕籑籒籓籙", 5],
        ["8fd3a1", "籡籣籧籩籭籮籰籲籹籼籽粆粇粏粔粞粠粦粰粶粷粺粻粼粿糄糇糈糉糍糏糓糔糕糗糙糚糝糦糩糫糵紃紇紈紉紏紑紒紓紖紝紞紣紦紪紭紱紼紽紾絀絁絇絈絍絑絓絗絙絚絜絝絥絧絪絰絸絺絻絿綁綂綃綅綆綈綋綌綍綑綖綗綝"],
        ["8fd4a1", "綞綦綧綪綳綶綷綹緂", 4, "緌緍緎緗緙縀緢緥緦緪緫緭緱緵緶緹緺縈縐縑縕縗縜縝縠縧縨縬縭縯縳縶縿繄繅繇繎繐繒繘繟繡繢繥繫繮繯繳繸繾纁纆纇纊纍纑纕纘纚纝纞缼缻缽缾缿罃罄罇罏罒罓罛罜罝罡罣罤罥罦罭"],
        ["8fd5a1", "罱罽罾罿羀羋羍羏羐羑羖羗羜羡羢羦羪羭羴羼羿翀翃翈翎翏翛翟翣翥翨翬翮翯翲翺翽翾翿耇耈耊耍耎耏耑耓耔耖耝耞耟耠耤耦耬耮耰耴耵耷耹耺耼耾聀聄聠聤聦聭聱聵肁肈肎肜肞肦肧肫肸肹胈胍胏胒胔胕胗胘胠胭胮"],
        ["8fd6a1", "胰胲胳胶胹胺胾脃脋脖脗脘脜脞脠脤脧脬脰脵脺脼腅腇腊腌腒腗腠腡腧腨腩腭腯腷膁膐膄膅膆膋膎膖膘膛膞膢膮膲膴膻臋臃臅臊臎臏臕臗臛臝臞臡臤臫臬臰臱臲臵臶臸臹臽臿舀舃舏舓舔舙舚舝舡舢舨舲舴舺艃艄艅艆"],
        ["8fd7a1", "艋艎艏艑艖艜艠艣艧艭艴艻艽艿芀芁芃芄芇芉芊芎芑芔芖芘芚芛芠芡芣芤芧芨芩芪芮芰芲芴芷芺芼芾芿苆苐苕苚苠苢苤苨苪苭苯苶苷苽苾茀茁茇茈茊茋荔茛茝茞茟茡茢茬茭茮茰茳茷茺茼茽荂荃荄荇荍荎荑荕荖荗荰荸"],
        ["8fd8a1", "荽荿莀莂莄莆莍莒莔莕莘莙莛莜莝莦莧莩莬莾莿菀菇菉菏菐菑菔菝荓菨菪菶菸菹菼萁萆萊萏萑萕萙莭萯萹葅葇葈葊葍葏葑葒葖葘葙葚葜葠葤葥葧葪葰葳葴葶葸葼葽蒁蒅蒒蒓蒕蒞蒦蒨蒩蒪蒯蒱蒴蒺蒽蒾蓀蓂蓇蓈蓌蓏蓓"],
        ["8fd9a1", "蓜蓧蓪蓯蓰蓱蓲蓷蔲蓺蓻蓽蔂蔃蔇蔌蔎蔐蔜蔞蔢蔣蔤蔥蔧蔪蔫蔯蔳蔴蔶蔿蕆蕏", 4, "蕖蕙蕜", 6, "蕤蕫蕯蕹蕺蕻蕽蕿薁薅薆薉薋薌薏薓薘薝薟薠薢薥薧薴薶薷薸薼薽薾薿藂藇藊藋藎薭藘藚藟藠藦藨藭藳藶藼"],
        ["8fdaa1", "藿蘀蘄蘅蘍蘎蘐蘑蘒蘘蘙蘛蘞蘡蘧蘩蘶蘸蘺蘼蘽虀虂虆虒虓虖虗虘虙虝虠", 4, "虩虬虯虵虶虷虺蚍蚑蚖蚘蚚蚜蚡蚦蚧蚨蚭蚱蚳蚴蚵蚷蚸蚹蚿蛀蛁蛃蛅蛑蛒蛕蛗蛚蛜蛠蛣蛥蛧蚈蛺蛼蛽蜄蜅蜇蜋蜎蜏蜐蜓蜔蜙蜞蜟蜡蜣"],
        ["8fdba1", "蜨蜮蜯蜱蜲蜹蜺蜼蜽蜾蝀蝃蝅蝍蝘蝝蝡蝤蝥蝯蝱蝲蝻螃", 6, "螋螌螐螓螕螗螘螙螞螠螣螧螬螭螮螱螵螾螿蟁蟈蟉蟊蟎蟕蟖蟙蟚蟜蟟蟢蟣蟤蟪蟫蟭蟱蟳蟸蟺蟿蠁蠃蠆蠉蠊蠋蠐蠙蠒蠓蠔蠘蠚蠛蠜蠞蠟蠨蠭蠮蠰蠲蠵"],
        ["8fdca1", "蠺蠼衁衃衅衈衉衊衋衎衑衕衖衘衚衜衟衠衤衩衱衹衻袀袘袚袛袜袟袠袨袪袺袽袾裀裊", 4, "裑裒裓裛裞裧裯裰裱裵裷褁褆褍褎褏褕褖褘褙褚褜褠褦褧褨褰褱褲褵褹褺褾襀襂襅襆襉襏襒襗襚襛襜襡襢襣襫襮襰襳襵襺"],
        ["8fdda1", "襻襼襽覉覍覐覔覕覛覜覟覠覥覰覴覵覶覷覼觔", 4, "觥觩觫觭觱觳觶觹觽觿訄訅訇訏訑訒訔訕訞訠訢訤訦訫訬訯訵訷訽訾詀詃詅詇詉詍詎詓詖詗詘詜詝詡詥詧詵詶詷詹詺詻詾詿誀誃誆誋誏誐誒誖誗誙誟誧誩誮誯誳"],
        ["8fdea1", "誶誷誻誾諃諆諈諉諊諑諓諔諕諗諝諟諬諰諴諵諶諼諿謅謆謋謑謜謞謟謊謭謰謷謼譂", 4, "譈譒譓譔譙譍譞譣譭譶譸譹譼譾讁讄讅讋讍讏讔讕讜讞讟谸谹谽谾豅豇豉豋豏豑豓豔豗豘豛豝豙豣豤豦豨豩豭豳豵豶豻豾貆"],
        ["8fdfa1", "貇貋貐貒貓貙貛貜貤貹貺賅賆賉賋賏賖賕賙賝賡賨賬賯賰賲賵賷賸賾賿贁贃贉贒贗贛赥赩赬赮赿趂趄趈趍趐趑趕趞趟趠趦趫趬趯趲趵趷趹趻跀跅跆跇跈跊跎跑跔跕跗跙跤跥跧跬跰趼跱跲跴跽踁踄踅踆踋踑踔踖踠踡踢"],
        ["8fe0a1", "踣踦踧踱踳踶踷踸踹踽蹀蹁蹋蹍蹎蹏蹔蹛蹜蹝蹞蹡蹢蹩蹬蹭蹯蹰蹱蹹蹺蹻躂躃躉躐躒躕躚躛躝躞躢躧躩躭躮躳躵躺躻軀軁軃軄軇軏軑軔軜軨軮軰軱軷軹軺軭輀輂輇輈輏輐輖輗輘輞輠輡輣輥輧輨輬輭輮輴輵輶輷輺轀轁"],
        ["8fe1a1", "轃轇轏轑", 4, "轘轝轞轥辝辠辡辤辥辦辵辶辸达迀迁迆迊迋迍运迒迓迕迠迣迤迨迮迱迵迶迻迾适逄逈逌逘逛逨逩逯逪逬逭逳逴逷逿遃遄遌遛遝遢遦遧遬遰遴遹邅邈邋邌邎邐邕邗邘邙邛邠邡邢邥邰邲邳邴邶邽郌邾郃"],
        ["8fe2a1", "郄郅郇郈郕郗郘郙郜郝郟郥郒郶郫郯郰郴郾郿鄀鄄鄅鄆鄈鄍鄐鄔鄖鄗鄘鄚鄜鄞鄠鄥鄢鄣鄧鄩鄮鄯鄱鄴鄶鄷鄹鄺鄼鄽酃酇酈酏酓酗酙酚酛酡酤酧酭酴酹酺酻醁醃醅醆醊醎醑醓醔醕醘醞醡醦醨醬醭醮醰醱醲醳醶醻醼醽醿"],
        ["8fe3a1", "釂釃釅釓釔釗釙釚釞釤釥釩釪釬", 5, "釷釹釻釽鈀鈁鈄鈅鈆鈇鈉鈊鈌鈐鈒鈓鈖鈘鈜鈝鈣鈤鈥鈦鈨鈮鈯鈰鈳鈵鈶鈸鈹鈺鈼鈾鉀鉂鉃鉆鉇鉊鉍鉎鉏鉑鉘鉙鉜鉝鉠鉡鉥鉧鉨鉩鉮鉯鉰鉵", 4, "鉻鉼鉽鉿銈銉銊銍銎銒銗"],
        ["8fe4a1", "銙銟銠銤銥銧銨銫銯銲銶銸銺銻銼銽銿", 4, "鋅鋆鋇鋈鋋鋌鋍鋎鋐鋓鋕鋗鋘鋙鋜鋝鋟鋠鋡鋣鋥鋧鋨鋬鋮鋰鋹鋻鋿錀錂錈錍錑錔錕錜錝錞錟錡錤錥錧錩錪錳錴錶錷鍇鍈鍉鍐鍑鍒鍕鍗鍘鍚鍞鍤鍥鍧鍩鍪鍭鍯鍰鍱鍳鍴鍶"],
        ["8fe5a1", "鍺鍽鍿鎀鎁鎂鎈鎊鎋鎍鎏鎒鎕鎘鎛鎞鎡鎣鎤鎦鎨鎫鎴鎵鎶鎺鎩鏁鏄鏅鏆鏇鏉", 4, "鏓鏙鏜鏞鏟鏢鏦鏧鏹鏷鏸鏺鏻鏽鐁鐂鐄鐈鐉鐍鐎鐏鐕鐖鐗鐟鐮鐯鐱鐲鐳鐴鐻鐿鐽鑃鑅鑈鑊鑌鑕鑙鑜鑟鑡鑣鑨鑫鑭鑮鑯鑱鑲钄钃镸镹"],
        ["8fe6a1", "镾閄閈閌閍閎閝閞閟閡閦閩閫閬閴閶閺閽閿闆闈闉闋闐闑闒闓闙闚闝闞闟闠闤闦阝阞阢阤阥阦阬阱阳阷阸阹阺阼阽陁陒陔陖陗陘陡陮陴陻陼陾陿隁隂隃隄隉隑隖隚隝隟隤隥隦隩隮隯隳隺雊雒嶲雘雚雝雞雟雩雯雱雺霂"],
        ["8fe7a1", "霃霅霉霚霛霝霡霢霣霨霱霳靁靃靊靎靏靕靗靘靚靛靣靧靪靮靳靶靷靸靻靽靿鞀鞉鞕鞖鞗鞙鞚鞞鞟鞢鞬鞮鞱鞲鞵鞶鞸鞹鞺鞼鞾鞿韁韄韅韇韉韊韌韍韎韐韑韔韗韘韙韝韞韠韛韡韤韯韱韴韷韸韺頇頊頙頍頎頔頖頜頞頠頣頦"],
        ["8fe8a1", "頫頮頯頰頲頳頵頥頾顄顇顊顑顒顓顖顗顙顚顢顣顥顦顪顬颫颭颮颰颴颷颸颺颻颿飂飅飈飌飡飣飥飦飧飪飳飶餂餇餈餑餕餖餗餚餛餜餟餢餦餧餫餱", 4, "餹餺餻餼饀饁饆饇饈饍饎饔饘饙饛饜饞饟饠馛馝馟馦馰馱馲馵"],
        ["8fe9a1", "馹馺馽馿駃駉駓駔駙駚駜駞駧駪駫駬駰駴駵駹駽駾騂騃騄騋騌騐騑騖騞騠騢騣騤騧騭騮騳騵騶騸驇驁驄驊驋驌驎驑驔驖驝骪骬骮骯骲骴骵骶骹骻骾骿髁髃髆髈髎髐髒髕髖髗髛髜髠髤髥髧髩髬髲髳髵髹髺髽髿", 4],
        ["8feaa1", "鬄鬅鬈鬉鬋鬌鬍鬎鬐鬒鬖鬙鬛鬜鬠鬦鬫鬭鬳鬴鬵鬷鬹鬺鬽魈魋魌魕魖魗魛魞魡魣魥魦魨魪", 4, "魳魵魷魸魹魿鮀鮄鮅鮆鮇鮉鮊鮋鮍鮏鮐鮔鮚鮝鮞鮦鮧鮩鮬鮰鮱鮲鮷鮸鮻鮼鮾鮿鯁鯇鯈鯎鯐鯗鯘鯝鯟鯥鯧鯪鯫鯯鯳鯷鯸"],
        ["8feba1", "鯹鯺鯽鯿鰀鰂鰋鰏鰑鰖鰘鰙鰚鰜鰞鰢鰣鰦", 4, "鰱鰵鰶鰷鰽鱁鱃鱄鱅鱉鱊鱎鱏鱐鱓鱔鱖鱘鱛鱝鱞鱟鱣鱩鱪鱜鱫鱨鱮鱰鱲鱵鱷鱻鳦鳲鳷鳹鴋鴂鴑鴗鴘鴜鴝鴞鴯鴰鴲鴳鴴鴺鴼鵅鴽鵂鵃鵇鵊鵓鵔鵟鵣鵢鵥鵩鵪鵫鵰鵶鵷鵻"],
        ["8feca1", "鵼鵾鶃鶄鶆鶊鶍鶎鶒鶓鶕鶖鶗鶘鶡鶪鶬鶮鶱鶵鶹鶼鶿鷃鷇鷉鷊鷔鷕鷖鷗鷚鷞鷟鷠鷥鷧鷩鷫鷮鷰鷳鷴鷾鸊鸂鸇鸎鸐鸑鸒鸕鸖鸙鸜鸝鹺鹻鹼麀麂麃麄麅麇麎麏麖麘麛麞麤麨麬麮麯麰麳麴麵黆黈黋黕黟黤黧黬黭黮黰黱黲黵"],
        ["8feda1", "黸黿鼂鼃鼉鼏鼐鼑鼒鼔鼖鼗鼙鼚鼛鼟鼢鼦鼪鼫鼯鼱鼲鼴鼷鼹鼺鼼鼽鼿齁齃", 4, "齓齕齖齗齘齚齝齞齨齩齭", 4, "齳齵齺齽龏龐龑龒龔龖龗龞龡龢龣龥"]
    ];
    module.exports = json;
},
/* iconv-lite/encodings/tables/cp936.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = [
        ["0", "\u0000", 127, "€"],
        ["8140", "丂丄丅丆丏丒丗丟丠両丣並丩丮丯丱丳丵丷丼乀乁乂乄乆乊乑乕乗乚乛乢乣乤乥乧乨乪", 5, "乲乴", 9, "乿", 6, "亇亊"],
        ["8180", "亐亖亗亙亜亝亞亣亪亯亰亱亴亶亷亸亹亼亽亾仈仌仏仐仒仚仛仜仠仢仦仧仩仭仮仯仱仴仸仹仺仼仾伀伂", 6, "伋伌伒", 4, "伜伝伡伣伨伩伬伭伮伱伳伵伷伹伻伾", 4, "佄佅佇", 5, "佒佔佖佡佢佦佨佪佫佭佮佱佲併佷佸佹佺佽侀侁侂侅來侇侊侌侎侐侒侓侕侖侘侙侚侜侞侟価侢"],
        ["8240", "侤侫侭侰", 4, "侶", 8, "俀俁係俆俇俈俉俋俌俍俒", 4, "俙俛俠俢俤俥俧俫俬俰俲俴俵俶俷俹俻俼俽俿", 11],
        ["8280", "個倎倐們倓倕倖倗倛倝倞倠倢倣値倧倫倯", 10, "倻倽倿偀偁偂偄偅偆偉偊偋偍偐", 4, "偖偗偘偙偛偝", 7, "偦", 5, "偭", 8, "偸偹偺偼偽傁傂傃傄傆傇傉傊傋傌傎", 20, "傤傦傪傫傭", 4, "傳", 6, "傼"],
        ["8340", "傽", 17, "僐", 5, "僗僘僙僛", 10, "僨僩僪僫僯僰僱僲僴僶", 4, "僼", 9, "儈"],
        ["8380", "儉儊儌", 5, "儓", 13, "儢", 28, "兂兇兊兌兎兏児兒兓兗兘兙兛兝", 4, "兣兤兦內兩兪兯兲兺兾兿冃冄円冇冊冋冎冏冐冑冓冔冘冚冝冞冟冡冣冦", 4, "冭冮冴冸冹冺冾冿凁凂凃凅凈凊凍凎凐凒", 5],
        ["8440", "凘凙凚凜凞凟凢凣凥", 5, "凬凮凱凲凴凷凾刄刅刉刋刌刏刐刓刔刕刜刞刟刡刢刣別刦刧刪刬刯刱刲刴刵刼刾剄", 5, "剋剎剏剒剓剕剗剘"],
        ["8480", "剙剚剛剝剟剠剢剣剤剦剨剫剬剭剮剰剱剳", 9, "剾劀劃", 4, "劉", 6, "劑劒劔", 6, "劜劤劥劦劧劮劯劰労", 9, "勀勁勂勄勅勆勈勊勌勍勎勏勑勓勔動勗務", 5, "勠勡勢勣勥", 10, "勱", 7, "勻勼勽匁匂匃匄匇匉匊匋匌匎"],
        ["8540", "匑匒匓匔匘匛匜匞匟匢匤匥匧匨匩匫匬匭匯", 9, "匼匽區卂卄卆卋卌卍卐協単卙卛卝卥卨卪卬卭卲卶卹卻卼卽卾厀厁厃厇厈厊厎厏"],
        ["8580", "厐", 4, "厖厗厙厛厜厞厠厡厤厧厪厫厬厭厯", 6, "厷厸厹厺厼厽厾叀參", 4, "収叏叐叒叓叕叚叜叝叞叡叢叧叴叺叾叿吀吂吅吇吋吔吘吙吚吜吢吤吥吪吰吳吶吷吺吽吿呁呂呄呅呇呉呌呍呎呏呑呚呝", 4, "呣呥呧呩", 7, "呴呹呺呾呿咁咃咅咇咈咉咊咍咑咓咗咘咜咞咟咠咡"],
        ["8640", "咢咥咮咰咲咵咶咷咹咺咼咾哃哅哊哋哖哘哛哠", 4, "哫哬哯哰哱哴", 5, "哻哾唀唂唃唄唅唈唊", 4, "唒唓唕", 5, "唜唝唞唟唡唥唦"],
        ["8680", "唨唩唫唭唲唴唵唶唸唹唺唻唽啀啂啅啇啈啋", 4, "啑啒啓啔啗", 4, "啝啞啟啠啢啣啨啩啫啯", 5, "啹啺啽啿喅喆喌喍喎喐喒喓喕喖喗喚喛喞喠", 6, "喨", 8, "喲喴営喸喺喼喿", 4, "嗆嗇嗈嗊嗋嗎嗏嗐嗕嗗", 4, "嗞嗠嗢嗧嗩嗭嗮嗰嗱嗴嗶嗸", 4, "嗿嘂嘃嘄嘅"],
        ["8740", "嘆嘇嘊嘋嘍嘐", 7, "嘙嘚嘜嘝嘠嘡嘢嘥嘦嘨嘩嘪嘫嘮嘯嘰嘳嘵嘷嘸嘺嘼嘽嘾噀", 11, "噏", 4, "噕噖噚噛噝", 4],
        ["8780", "噣噥噦噧噭噮噯噰噲噳噴噵噷噸噹噺噽", 7, "嚇", 6, "嚐嚑嚒嚔", 14, "嚤", 10, "嚰", 6, "嚸嚹嚺嚻嚽", 12, "囋", 8, "囕囖囘囙囜団囥", 5, "囬囮囯囲図囶囷囸囻囼圀圁圂圅圇國", 6],
        ["8840", "園", 9, "圝圞圠圡圢圤圥圦圧圫圱圲圴", 4, "圼圽圿坁坃坄坅坆坈坉坋坒", 4, "坘坙坢坣坥坧坬坮坰坱坲坴坵坸坹坺坽坾坿垀"],
        ["8880", "垁垇垈垉垊垍", 4, "垔", 6, "垜垝垞垟垥垨垪垬垯垰垱垳垵垶垷垹", 8, "埄", 6, "埌埍埐埑埓埖埗埛埜埞埡埢埣埥", 7, "埮埰埱埲埳埵埶執埻埼埾埿堁堃堄堅堈堉堊堌堎堏堐堒堓堔堖堗堘堚堛堜堝堟堢堣堥", 4, "堫", 4, "報堲堳場堶", 7],
        ["8940", "堾", 5, "塅", 6, "塎塏塐塒塓塕塖塗塙", 4, "塟", 5, "塦", 4, "塭", 16, "塿墂墄墆墇墈墊墋墌"],
        ["8980", "墍", 4, "墔", 4, "墛墜墝墠", 7, "墪", 17, "墽墾墿壀壂壃壄壆", 10, "壒壓壔壖", 13, "壥", 5, "壭壯壱売壴壵壷壸壺", 7, "夃夅夆夈", 4, "夎夐夑夒夓夗夘夛夝夞夠夡夢夣夦夨夬夰夲夳夵夶夻"],
        ["8a40", "夽夾夿奀奃奅奆奊奌奍奐奒奓奙奛", 4, "奡奣奤奦", 12, "奵奷奺奻奼奾奿妀妅妉妋妌妎妏妐妑妔妕妘妚妛妜妝妟妠妡妢妦"],
        ["8a80", "妧妬妭妰妱妳", 5, "妺妼妽妿", 6, "姇姈姉姌姍姎姏姕姖姙姛姞", 4, "姤姦姧姩姪姫姭", 11, "姺姼姽姾娀娂娊娋娍娎娏娐娒娔娕娖娗娙娚娛娝娞娡娢娤娦娧娨娪", 6, "娳娵娷", 4, "娽娾娿婁", 4, "婇婈婋", 9, "婖婗婘婙婛", 5],
        ["8b40", "婡婣婤婥婦婨婩婫", 8, "婸婹婻婼婽婾媀", 17, "媓", 6, "媜", 13, "媫媬"],
        ["8b80", "媭", 4, "媴媶媷媹", 4, "媿嫀嫃", 5, "嫊嫋嫍", 4, "嫓嫕嫗嫙嫚嫛嫝嫞嫟嫢嫤嫥嫧嫨嫪嫬", 4, "嫲", 22, "嬊", 11, "嬘", 25, "嬳嬵嬶嬸", 7, "孁", 6],
        ["8c40", "孈", 7, "孒孖孞孠孡孧孨孫孭孮孯孲孴孶孷學孹孻孼孾孿宂宆宊宍宎宐宑宒宔宖実宧宨宩宬宭宮宯宱宲宷宺宻宼寀寁寃寈寉寊寋寍寎寏"],
        ["8c80", "寑寔", 8, "寠寢寣實寧審", 4, "寯寱", 6, "寽対尀専尃尅將專尋尌對導尐尒尓尗尙尛尞尟尠尡尣尦尨尩尪尫尭尮尯尰尲尳尵尶尷屃屄屆屇屌屍屒屓屔屖屗屘屚屛屜屝屟屢層屧", 6, "屰屲", 6, "屻屼屽屾岀岃", 4, "岉岊岋岎岏岒岓岕岝", 4, "岤", 4],
        ["8d40", "岪岮岯岰岲岴岶岹岺岻岼岾峀峂峃峅", 5, "峌", 5, "峓", 5, "峚", 6, "峢峣峧峩峫峬峮峯峱", 9, "峼", 4],
        ["8d80", "崁崄崅崈", 5, "崏", 4, "崕崗崘崙崚崜崝崟", 4, "崥崨崪崫崬崯", 4, "崵", 7, "崿", 7, "嵈嵉嵍", 10, "嵙嵚嵜嵞", 10, "嵪嵭嵮嵰嵱嵲嵳嵵", 12, "嶃", 21, "嶚嶛嶜嶞嶟嶠"],
        ["8e40", "嶡", 21, "嶸", 12, "巆", 6, "巎", 12, "巜巟巠巣巤巪巬巭"],
        ["8e80", "巰巵巶巸", 4, "巿帀帄帇帉帊帋帍帎帒帓帗帞", 7, "帨", 4, "帯帰帲", 4, "帹帺帾帿幀幁幃幆", 5, "幍", 6, "幖", 4, "幜幝幟幠幣", 14, "幵幷幹幾庁庂広庅庈庉庌庍庎庒庘庛庝庡庢庣庤庨", 4, "庮", 4, "庴庺庻庼庽庿", 6],
        ["8f40", "廆廇廈廋", 5, "廔廕廗廘廙廚廜", 11, "廩廫", 8, "廵廸廹廻廼廽弅弆弇弉弌弍弎弐弒弔弖弙弚弜弝弞弡弢弣弤"],
        ["8f80", "弨弫弬弮弰弲", 6, "弻弽弾弿彁", 14, "彑彔彙彚彛彜彞彟彠彣彥彧彨彫彮彯彲彴彵彶彸彺彽彾彿徃徆徍徎徏徑従徔徖徚徛徝從徟徠徢", 5, "復徫徬徯", 5, "徶徸徹徺徻徾", 4, "忇忈忊忋忎忓忔忕忚忛応忞忟忢忣忥忦忨忩忬忯忰忲忳忴忶忷忹忺忼怇"],
        ["9040", "怈怉怋怌怐怑怓怗怘怚怞怟怢怣怤怬怭怮怰", 4, "怶", 4, "怽怾恀恄", 6, "恌恎恏恑恓恔恖恗恘恛恜恞恟恠恡恥恦恮恱恲恴恵恷恾悀"],
        ["9080", "悁悂悅悆悇悈悊悋悎悏悐悑悓悕悗悘悙悜悞悡悢悤悥悧悩悪悮悰悳悵悶悷悹悺悽", 7, "惇惈惉惌", 4, "惒惓惔惖惗惙惛惞惡", 4, "惪惱惲惵惷惸惻", 4, "愂愃愄愅愇愊愋愌愐", 4, "愖愗愘愙愛愜愝愞愡愢愥愨愩愪愬", 18, "慀", 6],
        ["9140", "慇慉態慍慏慐慒慓慔慖", 6, "慞慟慠慡慣慤慥慦慩", 6, "慱慲慳慴慶慸", 18, "憌憍憏", 4, "憕"],
        ["9180", "憖", 6, "憞", 8, "憪憫憭", 9, "憸", 5, "憿懀懁懃", 4, "應懌", 4, "懓懕", 16, "懧", 13, "懶", 8, "戀", 5, "戇戉戓戔戙戜戝戞戠戣戦戧戨戩戫戭戯戰戱戲戵戶戸", 4, "扂扄扅扆扊"],
        ["9240", "扏扐払扖扗扙扚扜", 6, "扤扥扨扱扲扴扵扷扸扺扻扽抁抂抃抅抆抇抈抋", 5, "抔抙抜抝択抣抦抧抩抪抭抮抯抰抲抳抴抶抷抸抺抾拀拁"],
        ["9280", "拃拋拏拑拕拝拞拠拡拤拪拫拰拲拵拸拹拺拻挀挃挄挅挆挊挋挌挍挏挐挒挓挔挕挗挘挙挜挦挧挩挬挭挮挰挱挳", 5, "挻挼挾挿捀捁捄捇捈捊捑捒捓捔捖", 7, "捠捤捥捦捨捪捫捬捯捰捲捳捴捵捸捹捼捽捾捿掁掃掄掅掆掋掍掑掓掔掕掗掙", 6, "採掤掦掫掯掱掲掵掶掹掻掽掿揀"],
        ["9340", "揁揂揃揅揇揈揊揋揌揑揓揔揕揗", 6, "揟揢揤", 4, "揫揬揮揯揰揱揳揵揷揹揺揻揼揾搃搄搆", 4, "損搎搑搒搕", 5, "搝搟搢搣搤"],
        ["9380", "搥搧搨搩搫搮", 5, "搵", 4, "搻搼搾摀摂摃摉摋", 6, "摓摕摖摗摙", 4, "摟", 7, "摨摪摫摬摮", 9, "摻", 6, "撃撆撈", 8, "撓撔撗撘撚撛撜撝撟", 4, "撥撦撧撨撪撫撯撱撲撳撴撶撹撻撽撾撿擁擃擄擆", 6, "擏擑擓擔擕擖擙據"],
        ["9440", "擛擜擝擟擠擡擣擥擧", 24, "攁", 7, "攊", 7, "攓", 4, "攙", 8],
        ["9480", "攢攣攤攦", 4, "攬攭攰攱攲攳攷攺攼攽敀", 4, "敆敇敊敋敍敎敐敒敓敔敗敘敚敜敟敠敡敤敥敧敨敩敪敭敮敯敱敳敵敶數", 14, "斈斉斊斍斎斏斒斔斕斖斘斚斝斞斠斢斣斦斨斪斬斮斱", 7, "斺斻斾斿旀旂旇旈旉旊旍旐旑旓旔旕旘", 7, "旡旣旤旪旫"],
        ["9540", "旲旳旴旵旸旹旻", 4, "昁昄昅昇昈昉昋昍昐昑昒昖昗昘昚昛昜昞昡昢昣昤昦昩昪昫昬昮昰昲昳昷", 4, "昽昿晀時晄", 6, "晍晎晐晑晘"],
        ["9580", "晙晛晜晝晞晠晢晣晥晧晩", 4, "晱晲晳晵晸晹晻晼晽晿暀暁暃暅暆暈暉暊暋暍暎暏暐暒暓暔暕暘", 4, "暞", 8, "暩", 4, "暯", 4, "暵暶暷暸暺暻暼暽暿", 25, "曚曞", 7, "曧曨曪", 5, "曱曵曶書曺曻曽朁朂會"],
        ["9640", "朄朅朆朇朌朎朏朑朒朓朖朘朙朚朜朞朠", 5, "朧朩朮朰朲朳朶朷朸朹朻朼朾朿杁杄杅杇杊杋杍杒杔杕杗", 4, "杝杢杣杤杦杧杫杬杮東杴杶"],
        ["9680", "杸杹杺杻杽枀枂枃枅枆枈枊枌枍枎枏枑枒枓枔枖枙枛枟枠枡枤枦枩枬枮枱枲枴枹", 7, "柂柅", 9, "柕柖柗柛柟柡柣柤柦柧柨柪柫柭柮柲柵", 7, "柾栁栂栃栄栆栍栐栒栔栕栘", 4, "栞栟栠栢", 6, "栫", 6, "栴栵栶栺栻栿桇桋桍桏桒桖", 5],
        ["9740", "桜桝桞桟桪桬", 7, "桵桸", 8, "梂梄梇", 7, "梐梑梒梔梕梖梘", 9, "梣梤梥梩梪梫梬梮梱梲梴梶梷梸"],
        ["9780", "梹", 6, "棁棃", 5, "棊棌棎棏棐棑棓棔棖棗棙棛", 4, "棡棢棤", 9, "棯棲棳棴棶棷棸棻棽棾棿椀椂椃椄椆", 4, "椌椏椑椓", 11, "椡椢椣椥", 7, "椮椯椱椲椳椵椶椷椸椺椻椼椾楀楁楃", 16, "楕楖楘楙楛楜楟"],
        ["9840", "楡楢楤楥楧楨楩楪楬業楯楰楲", 4, "楺楻楽楾楿榁榃榅榊榋榌榎", 5, "榖榗榙榚榝", 9, "榩榪榬榮榯榰榲榳榵榶榸榹榺榼榽"],
        ["9880", "榾榿槀槂", 7, "構槍槏槑槒槓槕", 5, "槜槝槞槡", 11, "槮槯槰槱槳", 9, "槾樀", 9, "樋", 11, "標", 5, "樠樢", 5, "権樫樬樭樮樰樲樳樴樶", 6, "樿", 4, "橅橆橈", 7, "橑", 6, "橚"],
        ["9940", "橜", 4, "橢橣橤橦", 10, "橲", 6, "橺橻橽橾橿檁檂檃檅", 8, "檏檒", 4, "檘", 7, "檡", 5],
        ["9980", "檧檨檪檭", 114, "欥欦欨", 6],
        ["9a40", "欯欰欱欳欴欵欶欸欻欼欽欿歀歁歂歄歅歈歊歋歍", 11, "歚", 7, "歨歩歫", 13, "歺歽歾歿殀殅殈"],
        ["9a80", "殌殎殏殐殑殔殕殗殘殙殜", 4, "殢", 7, "殫", 7, "殶殸", 6, "毀毃毄毆", 4, "毌毎毐毑毘毚毜", 4, "毢", 7, "毬毭毮毰毱毲毴毶毷毸毺毻毼毾", 6, "氈", 4, "氎氒気氜氝氞氠氣氥氫氬氭氱氳氶氷氹氺氻氼氾氿汃汄汅汈汋", 4, "汑汒汓汖汘"],
        ["9b40", "汙汚汢汣汥汦汧汫", 4, "汱汳汵汷汸決汻汼汿沀沄沇沊沋沍沎沑沒沕沖沗沘沚沜沝沞沠沢沨沬沯沰沴沵沶沷沺泀況泂泃泆泇泈泋泍泎泏泑泒泘"],
        ["9b80", "泙泚泜泝泟泤泦泧泩泬泭泲泴泹泿洀洂洃洅洆洈洉洊洍洏洐洑洓洔洕洖洘洜洝洟", 5, "洦洨洩洬洭洯洰洴洶洷洸洺洿浀浂浄浉浌浐浕浖浗浘浛浝浟浡浢浤浥浧浨浫浬浭浰浱浲浳浵浶浹浺浻浽", 4, "涃涄涆涇涊涋涍涏涐涒涖", 4, "涜涢涥涬涭涰涱涳涴涶涷涹", 5, "淁淂淃淈淉淊"],
        ["9c40", "淍淎淏淐淒淓淔淕淗淚淛淜淟淢淣淥淧淨淩淪淭淯淰淲淴淵淶淸淺淽", 7, "渆渇済渉渋渏渒渓渕渘渙減渜渞渟渢渦渧渨渪測渮渰渱渳渵"],
        ["9c80", "渶渷渹渻", 7, "湅", 7, "湏湐湑湒湕湗湙湚湜湝湞湠", 10, "湬湭湯", 14, "満溁溂溄溇溈溊", 4, "溑", 6, "溙溚溛溝溞溠溡溣溤溦溨溩溫溬溭溮溰溳溵溸溹溼溾溿滀滃滄滅滆滈滉滊滌滍滎滐滒滖滘滙滛滜滝滣滧滪", 5],
        ["9d40", "滰滱滲滳滵滶滷滸滺", 7, "漃漄漅漇漈漊", 4, "漐漑漒漖", 9, "漡漢漣漥漦漧漨漬漮漰漲漴漵漷", 6, "漿潀潁潂"],
        ["9d80", "潃潄潅潈潉潊潌潎", 9, "潙潚潛潝潟潠潡潣潤潥潧", 5, "潯潰潱潳潵潶潷潹潻潽", 6, "澅澆澇澊澋澏", 12, "澝澞澟澠澢", 4, "澨", 10, "澴澵澷澸澺", 5, "濁濃", 5, "濊", 6, "濓", 10, "濟濢濣濤濥"],
        ["9e40", "濦", 7, "濰", 32, "瀒", 7, "瀜", 6, "瀤", 6],
        ["9e80", "瀫", 9, "瀶瀷瀸瀺", 17, "灍灎灐", 13, "灟", 11, "灮灱灲灳灴灷灹灺灻災炁炂炃炄炆炇炈炋炌炍炏炐炑炓炗炘炚炛炞", 12, "炰炲炴炵炶為炾炿烄烅烆烇烉烋", 12, "烚"],
        ["9f40", "烜烝烞烠烡烢烣烥烪烮烰", 6, "烸烺烻烼烾", 10, "焋", 4, "焑焒焔焗焛", 10, "焧", 7, "焲焳焴"],
        ["9f80", "焵焷", 13, "煆煇煈煉煋煍煏", 12, "煝煟", 4, "煥煩", 4, "煯煰煱煴煵煶煷煹煻煼煾", 5, "熅", 4, "熋熌熍熎熐熑熒熓熕熖熗熚", 4, "熡", 6, "熩熪熫熭", 5, "熴熶熷熸熺", 8, "燄", 9, "燏", 4],
        ["a040", "燖", 9, "燡燢燣燤燦燨", 5, "燯", 9, "燺", 11, "爇", 19],
        ["a080", "爛爜爞", 9, "爩爫爭爮爯爲爳爴爺爼爾牀", 6, "牉牊牋牎牏牐牑牓牔牕牗牘牚牜牞牠牣牤牥牨牪牫牬牭牰牱牳牴牶牷牸牻牼牽犂犃犅", 4, "犌犎犐犑犓", 11, "犠", 11, "犮犱犲犳犵犺", 6, "狅狆狇狉狊狋狌狏狑狓狔狕狖狘狚狛"],
        ["a1a1", "　、。·ˉˇ¨〃々—～‖…‘’“”〔〕〈", 7, "〖〗【】±×÷∶∧∨∑∏∪∩∈∷√⊥∥∠⌒⊙∫∮≡≌≈∽∝≠≮≯≤≥∞∵∴♂♀°′″℃＄¤￠￡‰§№☆★○●◎◇◆□■△▲※→←↑↓〓"],
        ["a2a1", "ⅰ", 9],
        ["a2b1", "⒈", 19, "⑴", 19, "①", 9],
        ["a2e5", "㈠", 9],
        ["a2f1", "Ⅰ", 11],
        ["a3a1", "！＂＃￥％", 88, "￣"],
        ["a4a1", "ぁ", 82],
        ["a5a1", "ァ", 85],
        ["a6a1", "Α", 16, "Σ", 6],
        ["a6c1", "α", 16, "σ", 6],
        ["a6e0", "︵︶︹︺︿﹀︽︾﹁﹂﹃﹄"],
        ["a6ee", "︻︼︷︸︱"],
        ["a6f4", "︳︴"],
        ["a7a1", "А", 5, "ЁЖ", 25],
        ["a7d1", "а", 5, "ёж", 25],
        ["a840", "ˊˋ˙–―‥‵℅℉↖↗↘↙∕∟∣≒≦≧⊿═", 35, "▁", 6],
        ["a880", "█", 7, "▓▔▕▼▽◢◣◤◥☉⊕〒〝〞"],
        ["a8a1", "āáǎàēéěèīíǐìōóǒòūúǔùǖǘǚǜüêɑ"],
        ["a8bd", "ńň"],
        ["a8c0", "ɡ"],
        ["a8c5", "ㄅ", 36],
        ["a940", "〡", 8, "㊣㎎㎏㎜㎝㎞㎡㏄㏎㏑㏒㏕︰￢￤"],
        ["a959", "℡㈱"],
        ["a95c", "‐"],
        ["a960", "ー゛゜ヽヾ〆ゝゞ﹉", 9, "﹔﹕﹖﹗﹙", 8],
        ["a980", "﹢", 4, "﹨﹩﹪﹫"],
        ["a996", "〇"],
        ["a9a4", "─", 75],
        ["aa40", "狜狝狟狢", 5, "狪狫狵狶狹狽狾狿猀猂猄", 5, "猋猌猍猏猐猑猒猔猘猙猚猟猠猣猤猦猧猨猭猯猰猲猳猵猶猺猻猼猽獀", 8],
        ["aa80", "獉獊獋獌獎獏獑獓獔獕獖獘", 7, "獡", 10, "獮獰獱"],
        ["ab40", "獲", 11, "獿", 4, "玅玆玈玊玌玍玏玐玒玓玔玕玗玘玙玚玜玝玞玠玡玣", 5, "玪玬玭玱玴玵玶玸玹玼玽玾玿珁珃", 4],
        ["ab80", "珋珌珎珒", 6, "珚珛珜珝珟珡珢珣珤珦珨珪珫珬珮珯珰珱珳", 4],
        ["ac40", "珸", 10, "琄琇琈琋琌琍琎琑", 8, "琜", 5, "琣琤琧琩琫琭琯琱琲琷", 4, "琽琾琿瑀瑂", 11],
        ["ac80", "瑎", 6, "瑖瑘瑝瑠", 12, "瑮瑯瑱", 4, "瑸瑹瑺"],
        ["ad40", "瑻瑼瑽瑿璂璄璅璆璈璉璊璌璍璏璑", 10, "璝璟", 7, "璪", 15, "璻", 12],
        ["ad80", "瓈", 9, "瓓", 8, "瓝瓟瓡瓥瓧", 6, "瓰瓱瓲"],
        ["ae40", "瓳瓵瓸", 6, "甀甁甂甃甅", 7, "甎甐甒甔甕甖甗甛甝甞甠", 4, "甦甧甪甮甴甶甹甼甽甿畁畂畃畄畆畇畉畊畍畐畑畒畓畕畖畗畘"],
        ["ae80", "畝", 7, "畧畨畩畫", 6, "畳畵當畷畺", 4, "疀疁疂疄疅疇"],
        ["af40", "疈疉疊疌疍疎疐疓疕疘疛疜疞疢疦", 4, "疭疶疷疺疻疿痀痁痆痋痌痎痏痐痑痓痗痙痚痜痝痟痠痡痥痩痬痭痮痯痲痳痵痶痷痸痺痻痽痾瘂瘄瘆瘇"],
        ["af80", "瘈瘉瘋瘍瘎瘏瘑瘒瘓瘔瘖瘚瘜瘝瘞瘡瘣瘧瘨瘬瘮瘯瘱瘲瘶瘷瘹瘺瘻瘽癁療癄"],
        ["b040", "癅", 6, "癎", 5, "癕癗", 4, "癝癟癠癡癢癤", 6, "癬癭癮癰", 7, "癹発發癿皀皁皃皅皉皊皌皍皏皐皒皔皕皗皘皚皛"],
        ["b080", "皜", 7, "皥", 8, "皯皰皳皵", 9, "盀盁盃啊阿埃挨哎唉哀皑癌蔼矮艾碍爱隘鞍氨安俺按暗岸胺案肮昂盎凹敖熬翱袄傲奥懊澳芭捌扒叭吧笆八疤巴拔跋靶把耙坝霸罢爸白柏百摆佰败拜稗斑班搬扳般颁板版扮拌伴瓣半办绊邦帮梆榜膀绑棒磅蚌镑傍谤苞胞包褒剥"],
        ["b140", "盄盇盉盋盌盓盕盙盚盜盝盞盠", 4, "盦", 7, "盰盳盵盶盷盺盻盽盿眀眂眃眅眆眊県眎", 10, "眛眜眝眞眡眣眤眥眧眪眫"],
        ["b180", "眬眮眰", 4, "眹眻眽眾眿睂睄睅睆睈", 7, "睒", 7, "睜薄雹保堡饱宝抱报暴豹鲍爆杯碑悲卑北辈背贝钡倍狈备惫焙被奔苯本笨崩绷甭泵蹦迸逼鼻比鄙笔彼碧蓖蔽毕毙毖币庇痹闭敝弊必辟壁臂避陛鞭边编贬扁便变卞辨辩辫遍标彪膘表鳖憋别瘪彬斌濒滨宾摈兵冰柄丙秉饼炳"],
        ["b240", "睝睞睟睠睤睧睩睪睭", 11, "睺睻睼瞁瞂瞃瞆", 5, "瞏瞐瞓", 11, "瞡瞣瞤瞦瞨瞫瞭瞮瞯瞱瞲瞴瞶", 4],
        ["b280", "瞼瞾矀", 12, "矎", 8, "矘矙矚矝", 4, "矤病并玻菠播拨钵波博勃搏铂箔伯帛舶脖膊渤泊驳捕卜哺补埠不布步簿部怖擦猜裁材才财睬踩采彩菜蔡餐参蚕残惭惨灿苍舱仓沧藏操糙槽曹草厕策侧册测层蹭插叉茬茶查碴搽察岔差诧拆柴豺搀掺蝉馋谗缠铲产阐颤昌猖"],
        ["b340", "矦矨矪矯矰矱矲矴矵矷矹矺矻矼砃", 5, "砊砋砎砏砐砓砕砙砛砞砠砡砢砤砨砪砫砮砯砱砲砳砵砶砽砿硁硂硃硄硆硈硉硊硋硍硏硑硓硔硘硙硚"],
        ["b380", "硛硜硞", 11, "硯", 7, "硸硹硺硻硽", 6, "场尝常长偿肠厂敞畅唱倡超抄钞朝嘲潮巢吵炒车扯撤掣彻澈郴臣辰尘晨忱沉陈趁衬撑称城橙成呈乘程惩澄诚承逞骋秤吃痴持匙池迟弛驰耻齿侈尺赤翅斥炽充冲虫崇宠抽酬畴踌稠愁筹仇绸瞅丑臭初出橱厨躇锄雏滁除楚"],
        ["b440", "碄碅碆碈碊碋碏碐碒碔碕碖碙碝碞碠碢碤碦碨", 7, "碵碶碷碸確碻碼碽碿磀磂磃磄磆磇磈磌磍磎磏磑磒磓磖磗磘磚", 9],
        ["b480", "磤磥磦磧磩磪磫磭", 4, "磳磵磶磸磹磻", 5, "礂礃礄礆", 6, "础储矗搐触处揣川穿椽传船喘串疮窗幢床闯创吹炊捶锤垂春椿醇唇淳纯蠢戳绰疵茨磁雌辞慈瓷词此刺赐次聪葱囱匆从丛凑粗醋簇促蹿篡窜摧崔催脆瘁粹淬翠村存寸磋撮搓措挫错搭达答瘩打大呆歹傣戴带殆代贷袋待逮"],
        ["b540", "礍", 5, "礔", 9, "礟", 4, "礥", 14, "礵", 4, "礽礿祂祃祄祅祇祊", 8, "祔祕祘祙祡祣"],
        ["b580", "祤祦祩祪祫祬祮祰", 6, "祹祻", 4, "禂禃禆禇禈禉禋禌禍禎禐禑禒怠耽担丹单郸掸胆旦氮但惮淡诞弹蛋当挡党荡档刀捣蹈倒岛祷导到稻悼道盗德得的蹬灯登等瞪凳邓堤低滴迪敌笛狄涤翟嫡抵底地蒂第帝弟递缔颠掂滇碘点典靛垫电佃甸店惦奠淀殿碉叼雕凋刁掉吊钓调跌爹碟蝶迭谍叠"],
        ["b640", "禓", 6, "禛", 11, "禨", 10, "禴", 4, "禼禿秂秄秅秇秈秊秌秎秏秐秓秔秖秗秙", 5, "秠秡秢秥秨秪"],
        ["b680", "秬秮秱", 6, "秹秺秼秾秿稁稄稅稇稈稉稊稌稏", 4, "稕稖稘稙稛稜丁盯叮钉顶鼎锭定订丢东冬董懂动栋侗恫冻洞兜抖斗陡豆逗痘都督毒犊独读堵睹赌杜镀肚度渡妒端短锻段断缎堆兑队对墩吨蹲敦顿囤钝盾遁掇哆多夺垛躲朵跺舵剁惰堕蛾峨鹅俄额讹娥恶厄扼遏鄂饿恩而儿耳尔饵洱二"],
        ["b740", "稝稟稡稢稤", 14, "稴稵稶稸稺稾穀", 5, "穇", 9, "穒", 4, "穘", 16],
        ["b780", "穩", 6, "穱穲穳穵穻穼穽穾窂窅窇窉窊窋窌窎窏窐窓窔窙窚窛窞窡窢贰发罚筏伐乏阀法珐藩帆番翻樊矾钒繁凡烦反返范贩犯饭泛坊芳方肪房防妨仿访纺放菲非啡飞肥匪诽吠肺废沸费芬酚吩氛分纷坟焚汾粉奋份忿愤粪丰封枫蜂峰锋风疯烽逢冯缝讽奉凤佛否夫敷肤孵扶拂辐幅氟符伏俘服"],
        ["b840", "窣窤窧窩窪窫窮", 4, "窴", 10, "竀", 10, "竌", 9, "竗竘竚竛竜竝竡竢竤竧", 5, "竮竰竱竲竳"],
        ["b880", "竴", 4, "竻竼竾笀笁笂笅笇笉笌笍笎笐笒笓笖笗笘笚笜笝笟笡笢笣笧笩笭浮涪福袱弗甫抚辅俯釜斧脯腑府腐赴副覆赋复傅付阜父腹负富讣附妇缚咐噶嘎该改概钙盖溉干甘杆柑竿肝赶感秆敢赣冈刚钢缸肛纲岗港杠篙皋高膏羔糕搞镐稿告哥歌搁戈鸽胳疙割革葛格蛤阁隔铬个各给根跟耕更庚羹"],
        ["b940", "笯笰笲笴笵笶笷笹笻笽笿", 5, "筆筈筊筍筎筓筕筗筙筜筞筟筡筣", 10, "筯筰筳筴筶筸筺筼筽筿箁箂箃箄箆", 6, "箎箏"],
        ["b980", "箑箒箓箖箘箙箚箛箞箟箠箣箤箥箮箯箰箲箳箵箶箷箹", 7, "篂篃範埂耿梗工攻功恭龚供躬公宫弓巩汞拱贡共钩勾沟苟狗垢构购够辜菇咕箍估沽孤姑鼓古蛊骨谷股故顾固雇刮瓜剐寡挂褂乖拐怪棺关官冠观管馆罐惯灌贯光广逛瑰规圭硅归龟闺轨鬼诡癸桂柜跪贵刽辊滚棍锅郭国果裹过哈"],
        ["ba40", "篅篈築篊篋篍篎篏篐篒篔", 4, "篛篜篞篟篠篢篣篤篧篨篩篫篬篭篯篰篲", 4, "篸篹篺篻篽篿", 7, "簈簉簊簍簎簐", 5, "簗簘簙"],
        ["ba80", "簚", 4, "簠", 5, "簨簩簫", 12, "簹", 5, "籂骸孩海氦亥害骇酣憨邯韩含涵寒函喊罕翰撼捍旱憾悍焊汗汉夯杭航壕嚎豪毫郝好耗号浩呵喝荷菏核禾和何合盒貉阂河涸赫褐鹤贺嘿黑痕很狠恨哼亨横衡恒轰哄烘虹鸿洪宏弘红喉侯猴吼厚候后呼乎忽瑚壶葫胡蝴狐糊湖"],
        ["bb40", "籃", 9, "籎", 36, "籵", 5, "籾", 9],
        ["bb80", "粈粊", 6, "粓粔粖粙粚粛粠粡粣粦粧粨粩粫粬粭粯粰粴", 4, "粺粻弧虎唬护互沪户花哗华猾滑画划化话槐徊怀淮坏欢环桓还缓换患唤痪豢焕涣宦幻荒慌黄磺蝗簧皇凰惶煌晃幌恍谎灰挥辉徽恢蛔回毁悔慧卉惠晦贿秽会烩汇讳诲绘荤昏婚魂浑混豁活伙火获或惑霍货祸击圾基机畸稽积箕"],
        ["bc40", "粿糀糂糃糄糆糉糋糎", 6, "糘糚糛糝糞糡", 6, "糩", 5, "糰", 7, "糹糺糼", 13, "紋", 5],
        ["bc80", "紑", 14, "紡紣紤紥紦紨紩紪紬紭紮細", 6, "肌饥迹激讥鸡姬绩缉吉极棘辑籍集及急疾汲即嫉级挤几脊己蓟技冀季伎祭剂悸济寄寂计记既忌际妓继纪嘉枷夹佳家加荚颊贾甲钾假稼价架驾嫁歼监坚尖笺间煎兼肩艰奸缄茧检柬碱硷拣捡简俭剪减荐槛鉴践贱见键箭件"],
        ["bd40", "紷", 54, "絯", 7],
        ["bd80", "絸", 32, "健舰剑饯渐溅涧建僵姜将浆江疆蒋桨奖讲匠酱降蕉椒礁焦胶交郊浇骄娇嚼搅铰矫侥脚狡角饺缴绞剿教酵轿较叫窖揭接皆秸街阶截劫节桔杰捷睫竭洁结解姐戒藉芥界借介疥诫届巾筋斤金今津襟紧锦仅谨进靳晋禁近烬浸"],
        ["be40", "継", 12, "綧", 6, "綯", 42],
        ["be80", "線", 32, "尽劲荆兢茎睛晶鲸京惊精粳经井警景颈静境敬镜径痉靖竟竞净炯窘揪究纠玖韭久灸九酒厩救旧臼舅咎就疚鞠拘狙疽居驹菊局咀矩举沮聚拒据巨具距踞锯俱句惧炬剧捐鹃娟倦眷卷绢撅攫抉掘倔爵觉决诀绝均菌钧军君峻"],
        ["bf40", "緻", 62],
        ["bf80", "縺縼", 4, "繂", 4, "繈", 21, "俊竣浚郡骏喀咖卡咯开揩楷凯慨刊堪勘坎砍看康慷糠扛抗亢炕考拷烤靠坷苛柯棵磕颗科壳咳可渴克刻客课肯啃垦恳坑吭空恐孔控抠口扣寇枯哭窟苦酷库裤夸垮挎跨胯块筷侩快宽款匡筐狂框矿眶旷况亏盔岿窥葵奎魁傀"],
        ["c040", "繞", 35, "纃", 23, "纜纝纞"],
        ["c080", "纮纴纻纼绖绤绬绹缊缐缞缷缹缻", 6, "罃罆", 9, "罒罓馈愧溃坤昆捆困括扩廓阔垃拉喇蜡腊辣啦莱来赖蓝婪栏拦篮阑兰澜谰揽览懒缆烂滥琅榔狼廊郎朗浪捞劳牢老佬姥酪烙涝勒乐雷镭蕾磊累儡垒擂肋类泪棱楞冷厘梨犁黎篱狸离漓理李里鲤礼莉荔吏栗丽厉励砾历利傈例俐"],
        ["c140", "罖罙罛罜罝罞罠罣", 4, "罫罬罭罯罰罳罵罶罷罸罺罻罼罽罿羀羂", 7, "羋羍羏", 4, "羕", 4, "羛羜羠羢羣羥羦羨", 6, "羱"],
        ["c180", "羳", 4, "羺羻羾翀翂翃翄翆翇翈翉翋翍翏", 4, "翖翗翙", 5, "翢翣痢立粒沥隶力璃哩俩联莲连镰廉怜涟帘敛脸链恋炼练粮凉梁粱良两辆量晾亮谅撩聊僚疗燎寥辽潦了撂镣廖料列裂烈劣猎琳林磷霖临邻鳞淋凛赁吝拎玲菱零龄铃伶羚凌灵陵岭领另令溜琉榴硫馏留刘瘤流柳六龙聋咙笼窿"],
        ["c240", "翤翧翨翪翫翬翭翯翲翴", 6, "翽翾翿耂耇耈耉耊耎耏耑耓耚耛耝耞耟耡耣耤耫", 5, "耲耴耹耺耼耾聀聁聄聅聇聈聉聎聏聐聑聓聕聖聗"],
        ["c280", "聙聛", 13, "聫", 5, "聲", 11, "隆垄拢陇楼娄搂篓漏陋芦卢颅庐炉掳卤虏鲁麓碌露路赂鹿潞禄录陆戮驴吕铝侣旅履屡缕虑氯律率滤绿峦挛孪滦卵乱掠略抡轮伦仑沦纶论萝螺罗逻锣箩骡裸落洛骆络妈麻玛码蚂马骂嘛吗埋买麦卖迈脉瞒馒蛮满蔓曼慢漫"],
        ["c340", "聾肁肂肅肈肊肍", 5, "肔肕肗肙肞肣肦肧肨肬肰肳肵肶肸肹肻胅胇", 4, "胏", 6, "胘胟胠胢胣胦胮胵胷胹胻胾胿脀脁脃脄脅脇脈脋"],
        ["c380", "脌脕脗脙脛脜脝脟", 12, "脭脮脰脳脴脵脷脹", 4, "脿谩芒茫盲氓忙莽猫茅锚毛矛铆卯茂冒帽貌贸么玫枚梅酶霉煤没眉媒镁每美昧寐妹媚门闷们萌蒙檬盟锰猛梦孟眯醚靡糜迷谜弥米秘觅泌蜜密幂棉眠绵冕免勉娩缅面苗描瞄藐秒渺庙妙蔑灭民抿皿敏悯闽明螟鸣铭名命谬摸"],
        ["c440", "腀", 5, "腇腉腍腎腏腒腖腗腘腛", 4, "腡腢腣腤腦腨腪腫腬腯腲腳腵腶腷腸膁膃", 4, "膉膋膌膍膎膐膒", 5, "膙膚膞", 4, "膤膥"],
        ["c480", "膧膩膫", 7, "膴", 5, "膼膽膾膿臄臅臇臈臉臋臍", 6, "摹蘑模膜磨摩魔抹末莫墨默沫漠寞陌谋牟某拇牡亩姆母墓暮幕募慕木目睦牧穆拿哪呐钠那娜纳氖乃奶耐奈南男难囊挠脑恼闹淖呢馁内嫩能妮霓倪泥尼拟你匿腻逆溺蔫拈年碾撵捻念娘酿鸟尿捏聂孽啮镊镍涅您柠狞凝宁"],
        ["c540", "臔", 14, "臤臥臦臨臩臫臮", 4, "臵", 5, "臽臿舃與", 4, "舎舏舑舓舕", 5, "舝舠舤舥舦舧舩舮舲舺舼舽舿"],
        ["c580", "艀艁艂艃艅艆艈艊艌艍艎艐", 7, "艙艛艜艝艞艠", 7, "艩拧泞牛扭钮纽脓浓农弄奴努怒女暖虐疟挪懦糯诺哦欧鸥殴藕呕偶沤啪趴爬帕怕琶拍排牌徘湃派攀潘盘磐盼畔判叛乓庞旁耪胖抛咆刨炮袍跑泡呸胚培裴赔陪配佩沛喷盆砰抨烹澎彭蓬棚硼篷膨朋鹏捧碰坯砒霹批披劈琵毗"],
        ["c640", "艪艫艬艭艱艵艶艷艸艻艼芀芁芃芅芆芇芉芌芐芓芔芕芖芚芛芞芠芢芣芧芲芵芶芺芻芼芿苀苂苃苅苆苉苐苖苙苚苝苢苧苨苩苪苬苭苮苰苲苳苵苶苸"],
        ["c680", "苺苼", 4, "茊茋茍茐茒茓茖茘茙茝", 9, "茩茪茮茰茲茷茻茽啤脾疲皮匹痞僻屁譬篇偏片骗飘漂瓢票撇瞥拼频贫品聘乒坪苹萍平凭瓶评屏坡泼颇婆破魄迫粕剖扑铺仆莆葡菩蒲埔朴圃普浦谱曝瀑期欺栖戚妻七凄漆柒沏其棋奇歧畦崎脐齐旗祈祁骑起岂乞企启契砌器气迄弃汽泣讫掐"],
        ["c740", "茾茿荁荂荄荅荈荊", 4, "荓荕", 4, "荝荢荰", 6, "荹荺荾", 6, "莇莈莊莋莌莍莏莐莑莔莕莖莗莙莚莝莟莡", 6, "莬莭莮"],
        ["c780", "莯莵莻莾莿菂菃菄菆菈菉菋菍菎菐菑菒菓菕菗菙菚菛菞菢菣菤菦菧菨菫菬菭恰洽牵扦钎铅千迁签仟谦乾黔钱钳前潜遣浅谴堑嵌欠歉枪呛腔羌墙蔷强抢橇锹敲悄桥瞧乔侨巧鞘撬翘峭俏窍切茄且怯窃钦侵亲秦琴勤芹擒禽寝沁青轻氢倾卿清擎晴氰情顷请庆琼穷秋丘邱球求囚酋泅趋区蛆曲躯屈驱渠"],
        ["c840", "菮華菳", 4, "菺菻菼菾菿萀萂萅萇萈萉萊萐萒", 5, "萙萚萛萞", 5, "萩", 7, "萲", 5, "萹萺萻萾", 7, "葇葈葉"],
        ["c880", "葊", 6, "葒", 4, "葘葝葞葟葠葢葤", 4, "葪葮葯葰葲葴葷葹葻葼取娶龋趣去圈颧权醛泉全痊拳犬券劝缺炔瘸却鹊榷确雀裙群然燃冉染瓤壤攘嚷让饶扰绕惹热壬仁人忍韧任认刃妊纫扔仍日戎茸蓉荣融熔溶容绒冗揉柔肉茹蠕儒孺如辱乳汝入褥软阮蕊瑞锐闰润若弱撒洒萨腮鳃塞赛三叁"],
        ["c940", "葽", 4, "蒃蒄蒅蒆蒊蒍蒏", 7, "蒘蒚蒛蒝蒞蒟蒠蒢", 12, "蒰蒱蒳蒵蒶蒷蒻蒼蒾蓀蓂蓃蓅蓆蓇蓈蓋蓌蓎蓏蓒蓔蓕蓗"],
        ["c980", "蓘", 4, "蓞蓡蓢蓤蓧", 4, "蓭蓮蓯蓱", 10, "蓽蓾蔀蔁蔂伞散桑嗓丧搔骚扫嫂瑟色涩森僧莎砂杀刹沙纱傻啥煞筛晒珊苫杉山删煽衫闪陕擅赡膳善汕扇缮墒伤商赏晌上尚裳梢捎稍烧芍勺韶少哨邵绍奢赊蛇舌舍赦摄射慑涉社设砷申呻伸身深娠绅神沈审婶甚肾慎渗声生甥牲升绳"],
        ["ca40", "蔃", 8, "蔍蔎蔏蔐蔒蔔蔕蔖蔘蔙蔛蔜蔝蔞蔠蔢", 8, "蔭", 9, "蔾", 4, "蕄蕅蕆蕇蕋", 10],
        ["ca80", "蕗蕘蕚蕛蕜蕝蕟", 4, "蕥蕦蕧蕩", 8, "蕳蕵蕶蕷蕸蕼蕽蕿薀薁省盛剩胜圣师失狮施湿诗尸虱十石拾时什食蚀实识史矢使屎驶始式示士世柿事拭誓逝势是嗜噬适仕侍释饰氏市恃室视试收手首守寿授售受瘦兽蔬枢梳殊抒输叔舒淑疏书赎孰熟薯暑曙署蜀黍鼠属术述树束戍竖墅庶数漱"],
        ["cb40", "薂薃薆薈", 6, "薐", 10, "薝", 6, "薥薦薧薩薫薬薭薱", 5, "薸薺", 6, "藂", 6, "藊", 4, "藑藒"],
        ["cb80", "藔藖", 5, "藝", 6, "藥藦藧藨藪", 14, "恕刷耍摔衰甩帅栓拴霜双爽谁水睡税吮瞬顺舜说硕朔烁斯撕嘶思私司丝死肆寺嗣四伺似饲巳松耸怂颂送宋讼诵搜艘擞嗽苏酥俗素速粟僳塑溯宿诉肃酸蒜算虽隋随绥髓碎岁穗遂隧祟孙损笋蓑梭唆缩琐索锁所塌他它她塔"],
        ["cc40", "藹藺藼藽藾蘀", 4, "蘆", 10, "蘒蘓蘔蘕蘗", 15, "蘨蘪", 13, "蘹蘺蘻蘽蘾蘿虀"],
        ["cc80", "虁", 11, "虒虓處", 4, "虛虜虝號虠虡虣", 7, "獭挞蹋踏胎苔抬台泰酞太态汰坍摊贪瘫滩坛檀痰潭谭谈坦毯袒碳探叹炭汤塘搪堂棠膛唐糖倘躺淌趟烫掏涛滔绦萄桃逃淘陶讨套特藤腾疼誊梯剔踢锑提题蹄啼体替嚏惕涕剃屉天添填田甜恬舔腆挑条迢眺跳贴铁帖厅听烃"],
        ["cd40", "虭虯虰虲", 6, "蚃", 6, "蚎", 4, "蚔蚖", 5, "蚞", 4, "蚥蚦蚫蚭蚮蚲蚳蚷蚸蚹蚻", 4, "蛁蛂蛃蛅蛈蛌蛍蛒蛓蛕蛖蛗蛚蛜"],
        ["cd80", "蛝蛠蛡蛢蛣蛥蛦蛧蛨蛪蛫蛬蛯蛵蛶蛷蛺蛻蛼蛽蛿蜁蜄蜅蜆蜋蜌蜎蜏蜐蜑蜔蜖汀廷停亭庭挺艇通桐酮瞳同铜彤童桶捅筒统痛偷投头透凸秃突图徒途涂屠土吐兔湍团推颓腿蜕褪退吞屯臀拖托脱鸵陀驮驼椭妥拓唾挖哇蛙洼娃瓦袜歪外豌弯湾玩顽丸烷完碗挽晚皖惋宛婉万腕汪王亡枉网往旺望忘妄威"],
        ["ce40", "蜙蜛蜝蜟蜠蜤蜦蜧蜨蜪蜫蜬蜭蜯蜰蜲蜳蜵蜶蜸蜹蜺蜼蜽蝀", 6, "蝊蝋蝍蝏蝐蝑蝒蝔蝕蝖蝘蝚", 5, "蝡蝢蝦", 7, "蝯蝱蝲蝳蝵"],
        ["ce80", "蝷蝸蝹蝺蝿螀螁螄螆螇螉螊螌螎", 4, "螔螕螖螘", 6, "螠", 4, "巍微危韦违桅围唯惟为潍维苇萎委伟伪尾纬未蔚味畏胃喂魏位渭谓尉慰卫瘟温蚊文闻纹吻稳紊问嗡翁瓮挝蜗涡窝我斡卧握沃巫呜钨乌污诬屋无芜梧吾吴毋武五捂午舞伍侮坞戊雾晤物勿务悟误昔熙析西硒矽晰嘻吸锡牺"],
        ["cf40", "螥螦螧螩螪螮螰螱螲螴螶螷螸螹螻螼螾螿蟁", 4, "蟇蟈蟉蟌", 4, "蟔", 6, "蟜蟝蟞蟟蟡蟢蟣蟤蟦蟧蟨蟩蟫蟬蟭蟯", 9],
        ["cf80", "蟺蟻蟼蟽蟿蠀蠁蠂蠄", 5, "蠋", 7, "蠔蠗蠘蠙蠚蠜", 4, "蠣稀息希悉膝夕惜熄烯溪汐犀檄袭席习媳喜铣洗系隙戏细瞎虾匣霞辖暇峡侠狭下厦夏吓掀锨先仙鲜纤咸贤衔舷闲涎弦嫌显险现献县腺馅羡宪陷限线相厢镶香箱襄湘乡翔祥详想响享项巷橡像向象萧硝霄削哮嚣销消宵淆晓"],
        ["d040", "蠤", 13, "蠳", 5, "蠺蠻蠽蠾蠿衁衂衃衆", 5, "衎", 5, "衕衖衘衚", 6, "衦衧衪衭衯衱衳衴衵衶衸衹衺"],
        ["d080", "衻衼袀袃袆袇袉袊袌袎袏袐袑袓袔袕袗", 4, "袝", 4, "袣袥", 5, "小孝校肖啸笑效楔些歇蝎鞋协挟携邪斜胁谐写械卸蟹懈泄泻谢屑薪芯锌欣辛新忻心信衅星腥猩惺兴刑型形邢行醒幸杏性姓兄凶胸匈汹雄熊休修羞朽嗅锈秀袖绣墟戌需虚嘘须徐许蓄酗叙旭序畜恤絮婿绪续轩喧宣悬旋玄"],
        ["d140", "袬袮袯袰袲", 4, "袸袹袺袻袽袾袿裀裃裄裇裈裊裋裌裍裏裐裑裓裖裗裚", 4, "裠裡裦裧裩", 6, "裲裵裶裷裺裻製裿褀褁褃", 5],
        ["d180", "褉褋", 4, "褑褔", 4, "褜", 4, "褢褣褤褦褧褨褩褬褭褮褯褱褲褳褵褷选癣眩绚靴薛学穴雪血勋熏循旬询寻驯巡殉汛训讯逊迅压押鸦鸭呀丫芽牙蚜崖衙涯雅哑亚讶焉咽阉烟淹盐严研蜒岩延言颜阎炎沿奄掩眼衍演艳堰燕厌砚雁唁彦焰宴谚验殃央鸯秧杨扬佯疡羊洋阳氧仰痒养样漾邀腰妖瑶"],
        ["d240", "褸", 8, "襂襃襅", 24, "襠", 5, "襧", 19, "襼"],
        ["d280", "襽襾覀覂覄覅覇", 26, "摇尧遥窑谣姚咬舀药要耀椰噎耶爷野冶也页掖业叶曳腋夜液一壹医揖铱依伊衣颐夷遗移仪胰疑沂宜姨彝椅蚁倚已乙矣以艺抑易邑屹亿役臆逸肄疫亦裔意毅忆义益溢诣议谊译异翼翌绎茵荫因殷音阴姻吟银淫寅饮尹引隐"],
        ["d340", "覢", 30, "觃觍觓觔觕觗觘觙觛觝觟觠觡觢觤觧觨觩觪觬觭觮觰觱觲觴", 6],
        ["d380", "觻", 4, "訁", 5, "計", 21, "印英樱婴鹰应缨莹萤营荧蝇迎赢盈影颖硬映哟拥佣臃痈庸雍踊蛹咏泳涌永恿勇用幽优悠忧尤由邮铀犹油游酉有友右佑釉诱又幼迂淤于盂榆虞愚舆余俞逾鱼愉渝渔隅予娱雨与屿禹宇语羽玉域芋郁吁遇喻峪御愈欲狱育誉"],
        ["d440", "訞", 31, "訿", 8, "詉", 21],
        ["d480", "詟", 25, "詺", 6, "浴寓裕预豫驭鸳渊冤元垣袁原援辕园员圆猿源缘远苑愿怨院曰约越跃钥岳粤月悦阅耘云郧匀陨允运蕴酝晕韵孕匝砸杂栽哉灾宰载再在咱攒暂赞赃脏葬遭糟凿藻枣早澡蚤躁噪造皂灶燥责择则泽贼怎增憎曾赠扎喳渣札轧"],
        ["d540", "誁", 7, "誋", 7, "誔", 46],
        ["d580", "諃", 32, "铡闸眨栅榨咋乍炸诈摘斋宅窄债寨瞻毡詹粘沾盏斩辗崭展蘸栈占战站湛绽樟章彰漳张掌涨杖丈帐账仗胀瘴障招昭找沼赵照罩兆肇召遮折哲蛰辙者锗蔗这浙珍斟真甄砧臻贞针侦枕疹诊震振镇阵蒸挣睁征狰争怔整拯正政"],
        ["d640", "諤", 34, "謈", 27],
        ["d680", "謤謥謧", 30, "帧症郑证芝枝支吱蜘知肢脂汁之织职直植殖执值侄址指止趾只旨纸志挚掷至致置帜峙制智秩稚质炙痔滞治窒中盅忠钟衷终种肿重仲众舟周州洲诌粥轴肘帚咒皱宙昼骤珠株蛛朱猪诸诛逐竹烛煮拄瞩嘱主著柱助蛀贮铸筑"],
        ["d740", "譆", 31, "譧", 4, "譭", 25],
        ["d780", "讇", 24, "讬讱讻诇诐诪谉谞住注祝驻抓爪拽专砖转撰赚篆桩庄装妆撞壮状椎锥追赘坠缀谆准捉拙卓桌琢茁酌啄着灼浊兹咨资姿滋淄孜紫仔籽滓子自渍字鬃棕踪宗综总纵邹走奏揍租足卒族祖诅阻组钻纂嘴醉最罪尊遵昨左佐柞做作坐座"],
        ["d840", "谸", 8, "豂豃豄豅豈豊豋豍", 7, "豖豗豘豙豛", 5, "豣", 6, "豬", 6, "豴豵豶豷豻", 6, "貃貄貆貇"],
        ["d880", "貈貋貍", 6, "貕貖貗貙", 20, "亍丌兀丐廿卅丕亘丞鬲孬噩丨禺丿匕乇夭爻卮氐囟胤馗毓睾鼗丶亟鼐乜乩亓芈孛啬嘏仄厍厝厣厥厮靥赝匚叵匦匮匾赜卦卣刂刈刎刭刳刿剀剌剞剡剜蒯剽劂劁劐劓冂罔亻仃仉仂仨仡仫仞伛仳伢佤仵伥伧伉伫佞佧攸佚佝"],
        ["d940", "貮", 62],
        ["d980", "賭", 32, "佟佗伲伽佶佴侑侉侃侏佾佻侪佼侬侔俦俨俪俅俚俣俜俑俟俸倩偌俳倬倏倮倭俾倜倌倥倨偾偃偕偈偎偬偻傥傧傩傺僖儆僭僬僦僮儇儋仝氽佘佥俎龠汆籴兮巽黉馘冁夔勹匍訇匐凫夙兕亠兖亳衮袤亵脔裒禀嬴蠃羸冫冱冽冼"],
        ["da40", "贎", 14, "贠赑赒赗赟赥赨赩赪赬赮赯赱赲赸", 8, "趂趃趆趇趈趉趌", 4, "趒趓趕", 9, "趠趡"],
        ["da80", "趢趤", 12, "趲趶趷趹趻趽跀跁跂跅跇跈跉跊跍跐跒跓跔凇冖冢冥讠讦讧讪讴讵讷诂诃诋诏诎诒诓诔诖诘诙诜诟诠诤诨诩诮诰诳诶诹诼诿谀谂谄谇谌谏谑谒谔谕谖谙谛谘谝谟谠谡谥谧谪谫谮谯谲谳谵谶卩卺阝阢阡阱阪阽阼陂陉陔陟陧陬陲陴隈隍隗隰邗邛邝邙邬邡邴邳邶邺"],
        ["db40", "跕跘跙跜跠跡跢跥跦跧跩跭跮跰跱跲跴跶跼跾", 6, "踆踇踈踋踍踎踐踑踒踓踕", 7, "踠踡踤", 4, "踫踭踰踲踳踴踶踷踸踻踼踾"],
        ["db80", "踿蹃蹅蹆蹌", 4, "蹓", 5, "蹚", 11, "蹧蹨蹪蹫蹮蹱邸邰郏郅邾郐郄郇郓郦郢郜郗郛郫郯郾鄄鄢鄞鄣鄱鄯鄹酃酆刍奂劢劬劭劾哿勐勖勰叟燮矍廴凵凼鬯厶弁畚巯坌垩垡塾墼壅壑圩圬圪圳圹圮圯坜圻坂坩垅坫垆坼坻坨坭坶坳垭垤垌垲埏垧垴垓垠埕埘埚埙埒垸埴埯埸埤埝"],
        ["dc40", "蹳蹵蹷", 4, "蹽蹾躀躂躃躄躆躈", 6, "躑躒躓躕", 6, "躝躟", 11, "躭躮躰躱躳", 6, "躻", 7],
        ["dc80", "軃", 10, "軏", 21, "堋堍埽埭堀堞堙塄堠塥塬墁墉墚墀馨鼙懿艹艽艿芏芊芨芄芎芑芗芙芫芸芾芰苈苊苣芘芷芮苋苌苁芩芴芡芪芟苄苎芤苡茉苷苤茏茇苜苴苒苘茌苻苓茑茚茆茔茕苠苕茜荑荛荜茈莒茼茴茱莛荞茯荏荇荃荟荀茗荠茭茺茳荦荥"],
        ["dd40", "軥", 62],
        ["dd80", "輤", 32, "荨茛荩荬荪荭荮莰荸莳莴莠莪莓莜莅荼莶莩荽莸荻莘莞莨莺莼菁萁菥菘堇萘萋菝菽菖萜萸萑萆菔菟萏萃菸菹菪菅菀萦菰菡葜葑葚葙葳蒇蒈葺蒉葸萼葆葩葶蒌蒎萱葭蓁蓍蓐蓦蒽蓓蓊蒿蒺蓠蒡蒹蒴蒗蓥蓣蔌甍蔸蓰蔹蔟蔺"],
        ["de40", "轅", 32, "轪辀辌辒辝辠辡辢辤辥辦辧辪辬辭辮辯農辳辴辵辷辸辺辻込辿迀迃迆"],
        ["de80", "迉", 4, "迏迒迖迗迚迠迡迣迧迬迯迱迲迴迵迶迺迻迼迾迿逇逈逌逎逓逕逘蕖蔻蓿蓼蕙蕈蕨蕤蕞蕺瞢蕃蕲蕻薤薨薇薏蕹薮薜薅薹薷薰藓藁藜藿蘧蘅蘩蘖蘼廾弈夼奁耷奕奚奘匏尢尥尬尴扌扪抟抻拊拚拗拮挢拶挹捋捃掭揶捱捺掎掴捭掬掊捩掮掼揲揸揠揿揄揞揎摒揆掾摅摁搋搛搠搌搦搡摞撄摭撖"],
        ["df40", "這逜連逤逥逧", 5, "逰", 4, "逷逹逺逽逿遀遃遅遆遈", 4, "過達違遖遙遚遜", 5, "遤遦遧適遪遫遬遯", 4, "遶", 6, "遾邁"],
        ["df80", "還邅邆邇邉邊邌", 4, "邒邔邖邘邚邜邞邟邠邤邥邧邨邩邫邭邲邷邼邽邿郀摺撷撸撙撺擀擐擗擤擢攉攥攮弋忒甙弑卟叱叽叩叨叻吒吖吆呋呒呓呔呖呃吡呗呙吣吲咂咔呷呱呤咚咛咄呶呦咝哐咭哂咴哒咧咦哓哔呲咣哕咻咿哌哙哚哜咩咪咤哝哏哞唛哧唠哽唔哳唢唣唏唑唧唪啧喏喵啉啭啁啕唿啐唼"],
        ["e040", "郂郃郆郈郉郋郌郍郒郔郕郖郘郙郚郞郟郠郣郤郥郩郪郬郮郰郱郲郳郵郶郷郹郺郻郼郿鄀鄁鄃鄅", 19, "鄚鄛鄜"],
        ["e080", "鄝鄟鄠鄡鄤", 10, "鄰鄲", 6, "鄺", 8, "酄唷啖啵啶啷唳唰啜喋嗒喃喱喹喈喁喟啾嗖喑啻嗟喽喾喔喙嗪嗷嗉嘟嗑嗫嗬嗔嗦嗝嗄嗯嗥嗲嗳嗌嗍嗨嗵嗤辔嘞嘈嘌嘁嘤嘣嗾嘀嘧嘭噘嘹噗嘬噍噢噙噜噌噔嚆噤噱噫噻噼嚅嚓嚯囔囗囝囡囵囫囹囿圄圊圉圜帏帙帔帑帱帻帼"],
        ["e140", "酅酇酈酑酓酔酕酖酘酙酛酜酟酠酦酧酨酫酭酳酺酻酼醀", 4, "醆醈醊醎醏醓", 6, "醜", 5, "醤", 5, "醫醬醰醱醲醳醶醷醸醹醻"],
        ["e180", "醼", 10, "釈釋釐釒", 9, "針", 8, "帷幄幔幛幞幡岌屺岍岐岖岈岘岙岑岚岜岵岢岽岬岫岱岣峁岷峄峒峤峋峥崂崃崧崦崮崤崞崆崛嵘崾崴崽嵬嵛嵯嵝嵫嵋嵊嵩嵴嶂嶙嶝豳嶷巅彳彷徂徇徉後徕徙徜徨徭徵徼衢彡犭犰犴犷犸狃狁狎狍狒狨狯狩狲狴狷猁狳猃狺"],
        ["e240", "釦", 62],
        ["e280", "鈥", 32, "狻猗猓猡猊猞猝猕猢猹猥猬猸猱獐獍獗獠獬獯獾舛夥飧夤夂饣饧", 5, "饴饷饽馀馄馇馊馍馐馑馓馔馕庀庑庋庖庥庠庹庵庾庳赓廒廑廛廨廪膺忄忉忖忏怃忮怄忡忤忾怅怆忪忭忸怙怵怦怛怏怍怩怫怊怿怡恸恹恻恺恂"],
        ["e340", "鉆", 45, "鉵", 16],
        ["e380", "銆", 7, "銏", 24, "恪恽悖悚悭悝悃悒悌悛惬悻悱惝惘惆惚悴愠愦愕愣惴愀愎愫慊慵憬憔憧憷懔懵忝隳闩闫闱闳闵闶闼闾阃阄阆阈阊阋阌阍阏阒阕阖阗阙阚丬爿戕氵汔汜汊沣沅沐沔沌汨汩汴汶沆沩泐泔沭泷泸泱泗沲泠泖泺泫泮沱泓泯泾"],
        ["e440", "銨", 5, "銯", 24, "鋉", 31],
        ["e480", "鋩", 32, "洹洧洌浃浈洇洄洙洎洫浍洮洵洚浏浒浔洳涑浯涞涠浞涓涔浜浠浼浣渚淇淅淞渎涿淠渑淦淝淙渖涫渌涮渫湮湎湫溲湟溆湓湔渲渥湄滟溱溘滠漭滢溥溧溽溻溷滗溴滏溏滂溟潢潆潇漤漕滹漯漶潋潴漪漉漩澉澍澌潸潲潼潺濑"],
        ["e540", "錊", 51, "錿", 10],
        ["e580", "鍊", 31, "鍫濉澧澹澶濂濡濮濞濠濯瀚瀣瀛瀹瀵灏灞宀宄宕宓宥宸甯骞搴寤寮褰寰蹇謇辶迓迕迥迮迤迩迦迳迨逅逄逋逦逑逍逖逡逵逶逭逯遄遑遒遐遨遘遢遛暹遴遽邂邈邃邋彐彗彖彘尻咫屐屙孱屣屦羼弪弩弭艴弼鬻屮妁妃妍妩妪妣"],
        ["e640", "鍬", 34, "鎐", 27],
        ["e680", "鎬", 29, "鏋鏌鏍妗姊妫妞妤姒妲妯姗妾娅娆姝娈姣姘姹娌娉娲娴娑娣娓婀婧婊婕娼婢婵胬媪媛婷婺媾嫫媲嫒嫔媸嫠嫣嫱嫖嫦嫘嫜嬉嬗嬖嬲嬷孀尕尜孚孥孳孑孓孢驵驷驸驺驿驽骀骁骅骈骊骐骒骓骖骘骛骜骝骟骠骢骣骥骧纟纡纣纥纨纩"],
        ["e740", "鏎", 7, "鏗", 54],
        ["e780", "鐎", 32, "纭纰纾绀绁绂绉绋绌绐绔绗绛绠绡绨绫绮绯绱绲缍绶绺绻绾缁缂缃缇缈缋缌缏缑缒缗缙缜缛缟缡", 6, "缪缫缬缭缯", 4, "缵幺畿巛甾邕玎玑玮玢玟珏珂珑玷玳珀珉珈珥珙顼琊珩珧珞玺珲琏琪瑛琦琥琨琰琮琬"],
        ["e840", "鐯", 14, "鐿", 43, "鑬鑭鑮鑯"],
        ["e880", "鑰", 20, "钑钖钘铇铏铓铔铚铦铻锜锠琛琚瑁瑜瑗瑕瑙瑷瑭瑾璜璎璀璁璇璋璞璨璩璐璧瓒璺韪韫韬杌杓杞杈杩枥枇杪杳枘枧杵枨枞枭枋杷杼柰栉柘栊柩枰栌柙枵柚枳柝栀柃枸柢栎柁柽栲栳桠桡桎桢桄桤梃栝桕桦桁桧桀栾桊桉栩梵梏桴桷梓桫棂楮棼椟椠棹"],
        ["e940", "锧锳锽镃镈镋镕镚镠镮镴镵長", 7, "門", 42],
        ["e980", "閫", 32, "椤棰椋椁楗棣椐楱椹楠楂楝榄楫榀榘楸椴槌榇榈槎榉楦楣楹榛榧榻榫榭槔榱槁槊槟榕槠榍槿樯槭樗樘橥槲橄樾檠橐橛樵檎橹樽樨橘橼檑檐檩檗檫猷獒殁殂殇殄殒殓殍殚殛殡殪轫轭轱轲轳轵轶轸轷轹轺轼轾辁辂辄辇辋"],
        ["ea40", "闌", 27, "闬闿阇阓阘阛阞阠阣", 6, "阫阬阭阯阰阷阸阹阺阾陁陃陊陎陏陑陒陓陖陗"],
        ["ea80", "陘陙陚陜陝陞陠陣陥陦陫陭", 4, "陳陸", 12, "隇隉隊辍辎辏辘辚軎戋戗戛戟戢戡戥戤戬臧瓯瓴瓿甏甑甓攴旮旯旰昊昙杲昃昕昀炅曷昝昴昱昶昵耆晟晔晁晏晖晡晗晷暄暌暧暝暾曛曜曦曩贲贳贶贻贽赀赅赆赈赉赇赍赕赙觇觊觋觌觎觏觐觑牮犟牝牦牯牾牿犄犋犍犏犒挈挲掰"],
        ["eb40", "隌階隑隒隓隕隖隚際隝", 9, "隨", 7, "隱隲隴隵隷隸隺隻隿雂雃雈雊雋雐雑雓雔雖", 9, "雡", 6, "雫"],
        ["eb80", "雬雭雮雰雱雲雴雵雸雺電雼雽雿霂霃霅霊霋霌霐霑霒霔霕霗", 4, "霝霟霠搿擘耄毪毳毽毵毹氅氇氆氍氕氘氙氚氡氩氤氪氲攵敕敫牍牒牖爰虢刖肟肜肓肼朊肽肱肫肭肴肷胧胨胩胪胛胂胄胙胍胗朐胝胫胱胴胭脍脎胲胼朕脒豚脶脞脬脘脲腈腌腓腴腙腚腱腠腩腼腽腭腧塍媵膈膂膑滕膣膪臌朦臊膻"],
        ["ec40", "霡", 8, "霫霬霮霯霱霳", 4, "霺霻霼霽霿", 18, "靔靕靗靘靚靜靝靟靣靤靦靧靨靪", 7],
        ["ec80", "靲靵靷", 4, "靽", 7, "鞆", 4, "鞌鞎鞏鞐鞓鞕鞖鞗鞙", 4, "臁膦欤欷欹歃歆歙飑飒飓飕飙飚殳彀毂觳斐齑斓於旆旄旃旌旎旒旖炀炜炖炝炻烀炷炫炱烨烊焐焓焖焯焱煳煜煨煅煲煊煸煺熘熳熵熨熠燠燔燧燹爝爨灬焘煦熹戾戽扃扈扉礻祀祆祉祛祜祓祚祢祗祠祯祧祺禅禊禚禧禳忑忐"],
        ["ed40", "鞞鞟鞡鞢鞤", 6, "鞬鞮鞰鞱鞳鞵", 46],
        ["ed80", "韤韥韨韮", 4, "韴韷", 23, "怼恝恚恧恁恙恣悫愆愍慝憩憝懋懑戆肀聿沓泶淼矶矸砀砉砗砘砑斫砭砜砝砹砺砻砟砼砥砬砣砩硎硭硖硗砦硐硇硌硪碛碓碚碇碜碡碣碲碹碥磔磙磉磬磲礅磴礓礤礞礴龛黹黻黼盱眄眍盹眇眈眚眢眙眭眦眵眸睐睑睇睃睚睨"],
        ["ee40", "頏", 62],
        ["ee80", "顎", 32, "睢睥睿瞍睽瞀瞌瞑瞟瞠瞰瞵瞽町畀畎畋畈畛畲畹疃罘罡罟詈罨罴罱罹羁罾盍盥蠲钅钆钇钋钊钌钍钏钐钔钗钕钚钛钜钣钤钫钪钭钬钯钰钲钴钶", 4, "钼钽钿铄铈", 6, "铐铑铒铕铖铗铙铘铛铞铟铠铢铤铥铧铨铪"],
        ["ef40", "顯", 5, "颋颎颒颕颙颣風", 37, "飏飐飔飖飗飛飜飝飠", 4],
        ["ef80", "飥飦飩", 30, "铩铫铮铯铳铴铵铷铹铼铽铿锃锂锆锇锉锊锍锎锏锒", 4, "锘锛锝锞锟锢锪锫锩锬锱锲锴锶锷锸锼锾锿镂锵镄镅镆镉镌镎镏镒镓镔镖镗镘镙镛镞镟镝镡镢镤", 8, "镯镱镲镳锺矧矬雉秕秭秣秫稆嵇稃稂稞稔"],
        ["f040", "餈", 4, "餎餏餑", 28, "餯", 26],
        ["f080", "饊", 9, "饖", 12, "饤饦饳饸饹饻饾馂馃馉稹稷穑黏馥穰皈皎皓皙皤瓞瓠甬鸠鸢鸨", 4, "鸲鸱鸶鸸鸷鸹鸺鸾鹁鹂鹄鹆鹇鹈鹉鹋鹌鹎鹑鹕鹗鹚鹛鹜鹞鹣鹦", 6, "鹱鹭鹳疒疔疖疠疝疬疣疳疴疸痄疱疰痃痂痖痍痣痨痦痤痫痧瘃痱痼痿瘐瘀瘅瘌瘗瘊瘥瘘瘕瘙"],
        ["f140", "馌馎馚", 10, "馦馧馩", 47],
        ["f180", "駙", 32, "瘛瘼瘢瘠癀瘭瘰瘿瘵癃瘾瘳癍癞癔癜癖癫癯翊竦穸穹窀窆窈窕窦窠窬窨窭窳衤衩衲衽衿袂袢裆袷袼裉裢裎裣裥裱褚裼裨裾裰褡褙褓褛褊褴褫褶襁襦襻疋胥皲皴矜耒耔耖耜耠耢耥耦耧耩耨耱耋耵聃聆聍聒聩聱覃顸颀颃"],
        ["f240", "駺", 62],
        ["f280", "騹", 32, "颉颌颍颏颔颚颛颞颟颡颢颥颦虍虔虬虮虿虺虼虻蚨蚍蚋蚬蚝蚧蚣蚪蚓蚩蚶蛄蚵蛎蚰蚺蚱蚯蛉蛏蚴蛩蛱蛲蛭蛳蛐蜓蛞蛴蛟蛘蛑蜃蜇蛸蜈蜊蜍蜉蜣蜻蜞蜥蜮蜚蜾蝈蜴蜱蜩蜷蜿螂蜢蝽蝾蝻蝠蝰蝌蝮螋蝓蝣蝼蝤蝙蝥螓螯螨蟒"],
        ["f340", "驚", 17, "驲骃骉骍骎骔骕骙骦骩", 6, "骲骳骴骵骹骻骽骾骿髃髄髆", 4, "髍髎髏髐髒體髕髖髗髙髚髛髜"],
        ["f380", "髝髞髠髢髣髤髥髧髨髩髪髬髮髰", 8, "髺髼", 6, "鬄鬅鬆蟆螈螅螭螗螃螫蟥螬螵螳蟋蟓螽蟑蟀蟊蟛蟪蟠蟮蠖蠓蟾蠊蠛蠡蠹蠼缶罂罄罅舐竺竽笈笃笄笕笊笫笏筇笸笪笙笮笱笠笥笤笳笾笞筘筚筅筵筌筝筠筮筻筢筲筱箐箦箧箸箬箝箨箅箪箜箢箫箴篑篁篌篝篚篥篦篪簌篾篼簏簖簋"],
        ["f440", "鬇鬉", 5, "鬐鬑鬒鬔", 10, "鬠鬡鬢鬤", 10, "鬰鬱鬳", 7, "鬽鬾鬿魀魆魊魋魌魎魐魒魓魕", 5],
        ["f480", "魛", 32, "簟簪簦簸籁籀臾舁舂舄臬衄舡舢舣舭舯舨舫舸舻舳舴舾艄艉艋艏艚艟艨衾袅袈裘裟襞羝羟羧羯羰羲籼敉粑粝粜粞粢粲粼粽糁糇糌糍糈糅糗糨艮暨羿翎翕翥翡翦翩翮翳糸絷綦綮繇纛麸麴赳趄趔趑趱赧赭豇豉酊酐酎酏酤"],
        ["f540", "魼", 62],
        ["f580", "鮻", 32, "酢酡酰酩酯酽酾酲酴酹醌醅醐醍醑醢醣醪醭醮醯醵醴醺豕鹾趸跫踅蹙蹩趵趿趼趺跄跖跗跚跞跎跏跛跆跬跷跸跣跹跻跤踉跽踔踝踟踬踮踣踯踺蹀踹踵踽踱蹉蹁蹂蹑蹒蹊蹰蹶蹼蹯蹴躅躏躔躐躜躞豸貂貊貅貘貔斛觖觞觚觜"],
        ["f640", "鯜", 62],
        ["f680", "鰛", 32, "觥觫觯訾謦靓雩雳雯霆霁霈霏霎霪霭霰霾龀龃龅", 5, "龌黾鼋鼍隹隼隽雎雒瞿雠銎銮鋈錾鍪鏊鎏鐾鑫鱿鲂鲅鲆鲇鲈稣鲋鲎鲐鲑鲒鲔鲕鲚鲛鲞", 5, "鲥", 4, "鲫鲭鲮鲰", 7, "鲺鲻鲼鲽鳄鳅鳆鳇鳊鳋"],
        ["f740", "鰼", 62],
        ["f780", "鱻鱽鱾鲀鲃鲄鲉鲊鲌鲏鲓鲖鲗鲘鲙鲝鲪鲬鲯鲹鲾", 4, "鳈鳉鳑鳒鳚鳛鳠鳡鳌", 4, "鳓鳔鳕鳗鳘鳙鳜鳝鳟鳢靼鞅鞑鞒鞔鞯鞫鞣鞲鞴骱骰骷鹘骶骺骼髁髀髅髂髋髌髑魅魃魇魉魈魍魑飨餍餮饕饔髟髡髦髯髫髻髭髹鬈鬏鬓鬟鬣麽麾縻麂麇麈麋麒鏖麝麟黛黜黝黠黟黢黩黧黥黪黯鼢鼬鼯鼹鼷鼽鼾齄"],
        ["f840", "鳣", 62],
        ["f880", "鴢", 32],
        ["f940", "鵃", 62],
        ["f980", "鶂", 32],
        ["fa40", "鶣", 62],
        ["fa80", "鷢", 32],
        ["fb40", "鸃", 27, "鸤鸧鸮鸰鸴鸻鸼鹀鹍鹐鹒鹓鹔鹖鹙鹝鹟鹠鹡鹢鹥鹮鹯鹲鹴", 9, "麀"],
        ["fb80", "麁麃麄麅麆麉麊麌", 5, "麔", 8, "麞麠", 5, "麧麨麩麪"],
        ["fc40", "麫", 8, "麵麶麷麹麺麼麿", 4, "黅黆黇黈黊黋黌黐黒黓黕黖黗黙黚點黡黣黤黦黨黫黬黭黮黰", 8, "黺黽黿", 6],
        ["fc80", "鼆", 4, "鼌鼏鼑鼒鼔鼕鼖鼘鼚", 5, "鼡鼣", 8, "鼭鼮鼰鼱"],
        ["fd40", "鼲", 4, "鼸鼺鼼鼿", 4, "齅", 10, "齒", 38],
        ["fd80", "齹", 5, "龁龂龍", 11, "龜龝龞龡", 4, "郎凉秊裏隣"],
        ["fe40", "兀嗀﨎﨏﨑﨓﨔礼﨟蘒﨡﨣﨤﨧﨨﨩"]
    ];
    module.exports = json;
},
/* iconv-lite/encodings/tables/gbk-added.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = [
        ["a140", "", 62],
        ["a180", "", 32],
        ["a240", "", 62],
        ["a280", "", 32],
        ["a2ab", "", 5],
        ["a2e3", "€"],
        ["a2ef", ""],
        ["a2fd", ""],
        ["a340", "", 62],
        ["a380", "", 31, "　"],
        ["a440", "", 62],
        ["a480", "", 32],
        ["a4f4", "", 10],
        ["a540", "", 62],
        ["a580", "", 32],
        ["a5f7", "", 7],
        ["a640", "", 62],
        ["a680", "", 32],
        ["a6b9", "", 7],
        ["a6d9", "", 6],
        ["a6ec", ""],
        ["a6f3", ""],
        ["a6f6", "", 8],
        ["a740", "", 62],
        ["a780", "", 32],
        ["a7c2", "", 14],
        ["a7f2", "", 12],
        ["a896", "", 10],
        ["a8bc", ""],
        ["a8bf", "ǹ"],
        ["a8c1", ""],
        ["a8ea", "", 20],
        ["a958", ""],
        ["a95b", ""],
        ["a95d", ""],
        ["a989", "〾⿰", 11],
        ["a997", "", 12],
        ["a9f0", "", 14],
        ["aaa1", "", 93],
        ["aba1", "", 93],
        ["aca1", "", 93],
        ["ada1", "", 93],
        ["aea1", "", 93],
        ["afa1", "", 93],
        ["d7fa", "", 4],
        ["f8a1", "", 93],
        ["f9a1", "", 93],
        ["faa1", "", 93],
        ["fba1", "", 93],
        ["fca1", "", 93],
        ["fda1", "", 93],
        ["fe50", "⺁⺄㑳㑇⺈⺋㖞㘚㘎⺌⺗㥮㤘㧏㧟㩳㧐㭎㱮㳠⺧⺪䁖䅟⺮䌷⺳⺶⺷䎱䎬⺻䏝䓖䙡䙌"],
        ["fe80", "䜣䜩䝼䞍⻊䥇䥺䥽䦂䦃䦅䦆䦟䦛䦷䦶䲣䲟䲠䲡䱷䲢䴓", 6, "䶮", 93]
    ];
    module.exports = json;
},
/* iconv-lite/encodings/tables/gb18030-ranges.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = { "uChars": [128, 165, 169, 178, 184, 216, 226, 235, 238, 244, 248, 251, 253, 258, 276, 284, 300, 325, 329, 334, 364, 463, 465, 467, 469, 471, 473, 475, 477, 506, 594, 610, 712, 716, 730, 930, 938, 962, 970, 1026, 1104, 1106, 8209, 8215, 8218, 8222, 8231, 8241, 8244, 8246, 8252, 8365, 8452, 8454, 8458, 8471, 8482, 8556, 8570, 8596, 8602, 8713, 8720, 8722, 8726, 8731, 8737, 8740, 8742, 8748, 8751, 8760, 8766, 8777, 8781, 8787, 8802, 8808, 8816, 8854, 8858, 8870, 8896, 8979, 9322, 9372, 9548, 9588, 9616, 9622, 9634, 9652, 9662, 9672, 9676, 9680, 9702, 9735, 9738, 9793, 9795, 11906, 11909, 11913, 11917, 11928, 11944, 11947, 11951, 11956, 11960, 11964, 11979, 12284, 12292, 12312, 12319, 12330, 12351, 12436, 12447, 12535, 12543, 12586, 12842, 12850, 12964, 13200, 13215, 13218, 13253, 13263, 13267, 13270, 13384, 13428, 13727, 13839, 13851, 14617, 14703, 14801, 14816, 14964, 15183, 15471, 15585, 16471, 16736, 17208, 17325, 17330, 17374, 17623, 17997, 18018, 18212, 18218, 18301, 18318, 18760, 18811, 18814, 18820, 18823, 18844, 18848, 18872, 19576, 19620, 19738, 19887, 40870, 59244, 59336, 59367, 59413, 59417, 59423, 59431, 59437, 59443, 59452, 59460, 59478, 59493, 63789, 63866, 63894, 63976, 63986, 64016, 64018, 64021, 64025, 64034, 64037, 64042, 65074, 65093, 65107, 65112, 65127, 65132, 65375, 65510, 65536], "gbChars": [0, 36, 38, 45, 50, 81, 89, 95, 96, 100, 103, 104, 105, 109, 126, 133, 148, 172, 175, 179, 208, 306, 307, 308, 309, 310, 311, 312, 313, 341, 428, 443, 544, 545, 558, 741, 742, 749, 750, 805, 819, 820, 7922, 7924, 7925, 7927, 7934, 7943, 7944, 7945, 7950, 8062, 8148, 8149, 8152, 8164, 8174, 8236, 8240, 8262, 8264, 8374, 8380, 8381, 8384, 8388, 8390, 8392, 8393, 8394, 8396, 8401, 8406, 8416, 8419, 8424, 8437, 8439, 8445, 8482, 8485, 8496, 8521, 8603, 8936, 8946, 9046, 9050, 9063, 9066, 9076, 9092, 9100, 9108, 9111, 9113, 9131, 9162, 9164, 9218, 9219, 11329, 11331, 11334, 11336, 11346, 11361, 11363, 11366, 11370, 11372, 11375, 11389, 11682, 11686, 11687, 11692, 11694, 11714, 11716, 11723, 11725, 11730, 11736, 11982, 11989, 12102, 12336, 12348, 12350, 12384, 12393, 12395, 12397, 12510, 12553, 12851, 12962, 12973, 13738, 13823, 13919, 13933, 14080, 14298, 14585, 14698, 15583, 15847, 16318, 16434, 16438, 16481, 16729, 17102, 17122, 17315, 17320, 17402, 17418, 17859, 17909, 17911, 17915, 17916, 17936, 17939, 17961, 18664, 18703, 18814, 18962, 19043, 33469, 33470, 33471, 33484, 33485, 33490, 33497, 33501, 33505, 33513, 33520, 33536, 33550, 37845, 37921, 37948, 38029, 38038, 38064, 38065, 38066, 38069, 38075, 38076, 38078, 39108, 39109, 39113, 39114, 39115, 39116, 39265, 39394, 189000] };
    module.exports = json;
},
/* iconv-lite/encodings/tables/cp949.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = [
        ["0", "\u0000", 127],
        ["8141", "갂갃갅갆갋", 4, "갘갞갟갡갢갣갥", 6, "갮갲갳갴"],
        ["8161", "갵갶갷갺갻갽갾갿걁", 9, "걌걎", 5, "걕"],
        ["8181", "걖걗걙걚걛걝", 18, "걲걳걵걶걹걻", 4, "겂겇겈겍겎겏겑겒겓겕", 6, "겞겢", 5, "겫겭겮겱", 6, "겺겾겿곀곂곃곅곆곇곉곊곋곍", 7, "곖곘", 7, "곢곣곥곦곩곫곭곮곲곴곷", 4, "곾곿괁괂괃괅괇", 4, "괎괐괒괓"],
        ["8241", "괔괕괖괗괙괚괛괝괞괟괡", 7, "괪괫괮", 5],
        ["8261", "괶괷괹괺괻괽", 6, "굆굈굊", 5, "굑굒굓굕굖굗"],
        ["8281", "굙", 7, "굢굤", 7, "굮굯굱굲굷굸굹굺굾궀궃", 4, "궊궋궍궎궏궑", 10, "궞", 5, "궥", 17, "궸", 7, "귂귃귅귆귇귉", 6, "귒귔", 7, "귝귞귟귡귢귣귥", 18],
        ["8341", "귺귻귽귾긂", 5, "긊긌긎", 5, "긕", 7],
        ["8361", "긝", 18, "긲긳긵긶긹긻긼"],
        ["8381", "긽긾긿깂깄깇깈깉깋깏깑깒깓깕깗", 4, "깞깢깣깤깦깧깪깫깭깮깯깱", 6, "깺깾", 5, "꺆", 5, "꺍", 46, "꺿껁껂껃껅", 6, "껎껒", 5, "껚껛껝", 8],
        ["8441", "껦껧껩껪껬껮", 5, "껵껶껷껹껺껻껽", 8],
        ["8461", "꼆꼉꼊꼋꼌꼎꼏꼑", 18],
        ["8481", "꼤", 7, "꼮꼯꼱꼳꼵", 6, "꼾꽀꽄꽅꽆꽇꽊", 5, "꽑", 10, "꽞", 5, "꽦", 18, "꽺", 5, "꾁꾂꾃꾅꾆꾇꾉", 6, "꾒꾓꾔꾖", 5, "꾝", 26, "꾺꾻꾽꾾"],
        ["8541", "꾿꿁", 5, "꿊꿌꿏", 4, "꿕", 6, "꿝", 4],
        ["8561", "꿢", 5, "꿪", 5, "꿲꿳꿵꿶꿷꿹", 6, "뀂뀃"],
        ["8581", "뀅", 6, "뀍뀎뀏뀑뀒뀓뀕", 6, "뀞", 9, "뀩", 26, "끆끇끉끋끍끏끐끑끒끖끘끚끛끜끞", 29, "끾끿낁낂낃낅", 6, "낎낐낒", 5, "낛낝낞낣낤"],
        ["8641", "낥낦낧낪낰낲낶낷낹낺낻낽", 6, "냆냊", 5, "냒"],
        ["8661", "냓냕냖냗냙", 6, "냡냢냣냤냦", 10],
        ["8681", "냱", 22, "넊넍넎넏넑넔넕넖넗넚넞", 4, "넦넧넩넪넫넭", 6, "넶넺", 5, "녂녃녅녆녇녉", 6, "녒녓녖녗녙녚녛녝녞녟녡", 22, "녺녻녽녾녿놁놃", 4, "놊놌놎놏놐놑놕놖놗놙놚놛놝"],
        ["8741", "놞", 9, "놩", 15],
        ["8761", "놹", 18, "뇍뇎뇏뇑뇒뇓뇕"],
        ["8781", "뇖", 5, "뇞뇠", 7, "뇪뇫뇭뇮뇯뇱", 7, "뇺뇼뇾", 5, "눆눇눉눊눍", 6, "눖눘눚", 5, "눡", 18, "눵", 6, "눽", 26, "뉙뉚뉛뉝뉞뉟뉡", 6, "뉪", 4],
        ["8841", "뉯", 4, "뉶", 5, "뉽", 6, "늆늇늈늊", 4],
        ["8861", "늏늒늓늕늖늗늛", 4, "늢늤늧늨늩늫늭늮늯늱늲늳늵늶늷"],
        ["8881", "늸", 15, "닊닋닍닎닏닑닓", 4, "닚닜닞닟닠닡닣닧닩닪닰닱닲닶닼닽닾댂댃댅댆댇댉", 6, "댒댖", 5, "댝", 54, "덗덙덚덝덠덡덢덣"],
        ["8941", "덦덨덪덬덭덯덲덳덵덶덷덹", 6, "뎂뎆", 5, "뎍"],
        ["8961", "뎎뎏뎑뎒뎓뎕", 10, "뎢", 5, "뎩뎪뎫뎭"],
        ["8981", "뎮", 21, "돆돇돉돊돍돏돑돒돓돖돘돚돜돞돟돡돢돣돥돦돧돩", 18, "돽", 18, "됑", 6, "됙됚됛됝됞됟됡", 6, "됪됬", 7, "됵", 15],
        ["8a41", "둅", 10, "둒둓둕둖둗둙", 6, "둢둤둦"],
        ["8a61", "둧", 4, "둭", 18, "뒁뒂"],
        ["8a81", "뒃", 4, "뒉", 19, "뒞", 5, "뒥뒦뒧뒩뒪뒫뒭", 7, "뒶뒸뒺", 5, "듁듂듃듅듆듇듉", 6, "듑듒듓듔듖", 5, "듞듟듡듢듥듧", 4, "듮듰듲", 5, "듹", 26, "딖딗딙딚딝"],
        ["8b41", "딞", 5, "딦딫", 4, "딲딳딵딶딷딹", 6, "땂땆"],
        ["8b61", "땇땈땉땊땎땏땑땒땓땕", 6, "땞땢", 8],
        ["8b81", "땫", 52, "떢떣떥떦떧떩떬떭떮떯떲떶", 4, "떾떿뗁뗂뗃뗅", 6, "뗎뗒", 5, "뗙", 18, "뗭", 18],
        ["8c41", "똀", 15, "똒똓똕똖똗똙", 4],
        ["8c61", "똞", 6, "똦", 5, "똭", 6, "똵", 5],
        ["8c81", "똻", 12, "뙉", 26, "뙥뙦뙧뙩", 50, "뚞뚟뚡뚢뚣뚥", 5, "뚭뚮뚯뚰뚲", 16],
        ["8d41", "뛃", 16, "뛕", 8],
        ["8d61", "뛞", 17, "뛱뛲뛳뛵뛶뛷뛹뛺"],
        ["8d81", "뛻", 4, "뜂뜃뜄뜆", 33, "뜪뜫뜭뜮뜱", 6, "뜺뜼", 7, "띅띆띇띉띊띋띍", 6, "띖", 9, "띡띢띣띥띦띧띩", 6, "띲띴띶", 5, "띾띿랁랂랃랅", 6, "랎랓랔랕랚랛랝랞"],
        ["8e41", "랟랡", 6, "랪랮", 5, "랶랷랹", 8],
        ["8e61", "럂", 4, "럈럊", 19],
        ["8e81", "럞", 13, "럮럯럱럲럳럵", 6, "럾렂", 4, "렊렋렍렎렏렑", 6, "렚렜렞", 5, "렦렧렩렪렫렭", 6, "렶렺", 5, "롁롂롃롅", 11, "롒롔", 7, "롞롟롡롢롣롥", 6, "롮롰롲", 5, "롹롺롻롽", 7],
        ["8f41", "뢅", 7, "뢎", 17],
        ["8f61", "뢠", 7, "뢩", 6, "뢱뢲뢳뢵뢶뢷뢹", 4],
        ["8f81", "뢾뢿룂룄룆", 5, "룍룎룏룑룒룓룕", 7, "룞룠룢", 5, "룪룫룭룮룯룱", 6, "룺룼룾", 5, "뤅", 18, "뤙", 6, "뤡", 26, "뤾뤿륁륂륃륅", 6, "륍륎륐륒", 5],
        ["9041", "륚륛륝륞륟륡", 6, "륪륬륮", 5, "륶륷륹륺륻륽"],
        ["9061", "륾", 5, "릆릈릋릌릏", 15],
        ["9081", "릟", 12, "릮릯릱릲릳릵", 6, "릾맀맂", 5, "맊맋맍맓", 4, "맚맜맟맠맢맦맧맩맪맫맭", 6, "맶맻", 4, "먂", 5, "먉", 11, "먖", 33, "먺먻먽먾먿멁멃멄멅멆"],
        ["9141", "멇멊멌멏멐멑멒멖멗멙멚멛멝", 6, "멦멪", 5],
        ["9161", "멲멳멵멶멷멹", 9, "몆몈몉몊몋몍", 5],
        ["9181", "몓", 20, "몪몭몮몯몱몳", 4, "몺몼몾", 5, "뫅뫆뫇뫉", 14, "뫚", 33, "뫽뫾뫿묁묂묃묅", 7, "묎묐묒", 5, "묙묚묛묝묞묟묡", 6],
        ["9241", "묨묪묬", 7, "묷묹묺묿", 4, "뭆뭈뭊뭋뭌뭎뭑뭒"],
        ["9261", "뭓뭕뭖뭗뭙", 7, "뭢뭤", 7, "뭭", 4],
        ["9281", "뭲", 21, "뮉뮊뮋뮍뮎뮏뮑", 18, "뮥뮦뮧뮩뮪뮫뮭", 6, "뮵뮶뮸", 7, "믁믂믃믅믆믇믉", 6, "믑믒믔", 35, "믺믻믽믾밁"],
        ["9341", "밃", 4, "밊밎밐밒밓밙밚밠밡밢밣밦밨밪밫밬밮밯밲밳밵"],
        ["9361", "밶밷밹", 6, "뱂뱆뱇뱈뱊뱋뱎뱏뱑", 8],
        ["9381", "뱚뱛뱜뱞", 37, "벆벇벉벊벍벏", 4, "벖벘벛", 4, "벢벣벥벦벩", 6, "벲벶", 5, "벾벿볁볂볃볅", 7, "볎볒볓볔볖볗볙볚볛볝", 22, "볷볹볺볻볽"],
        ["9441", "볾", 5, "봆봈봊", 5, "봑봒봓봕", 8],
        ["9461", "봞", 5, "봥", 6, "봭", 12],
        ["9481", "봺", 5, "뵁", 6, "뵊뵋뵍뵎뵏뵑", 6, "뵚", 9, "뵥뵦뵧뵩", 22, "붂붃붅붆붋", 4, "붒붔붖붗붘붛붝", 6, "붥", 10, "붱", 6, "붹", 24],
        ["9541", "뷒뷓뷖뷗뷙뷚뷛뷝", 11, "뷪", 5, "뷱"],
        ["9561", "뷲뷳뷵뷶뷷뷹", 6, "븁븂븄븆", 5, "븎븏븑븒븓"],
        ["9581", "븕", 6, "븞븠", 35, "빆빇빉빊빋빍빏", 4, "빖빘빜빝빞빟빢빣빥빦빧빩빫", 4, "빲빶", 4, "빾빿뺁뺂뺃뺅", 6, "뺎뺒", 5, "뺚", 13, "뺩", 14],
        ["9641", "뺸", 23, "뻒뻓"],
        ["9661", "뻕뻖뻙", 6, "뻡뻢뻦", 5, "뻭", 8],
        ["9681", "뻶", 10, "뼂", 5, "뼊", 13, "뼚뼞", 33, "뽂뽃뽅뽆뽇뽉", 6, "뽒뽓뽔뽖", 44],
        ["9741", "뾃", 16, "뾕", 8],
        ["9761", "뾞", 17, "뾱", 7],
        ["9781", "뾹", 11, "뿆", 5, "뿎뿏뿑뿒뿓뿕", 6, "뿝뿞뿠뿢", 89, "쀽쀾쀿"],
        ["9841", "쁀", 16, "쁒", 5, "쁙쁚쁛"],
        ["9861", "쁝쁞쁟쁡", 6, "쁪", 15],
        ["9881", "쁺", 21, "삒삓삕삖삗삙", 6, "삢삤삦", 5, "삮삱삲삷", 4, "삾샂샃샄샆샇샊샋샍샎샏샑", 6, "샚샞", 5, "샦샧샩샪샫샭", 6, "샶샸샺", 5, "섁섂섃섅섆섇섉", 6, "섑섒섓섔섖", 5, "섡섢섥섨섩섪섫섮"],
        ["9941", "섲섳섴섵섷섺섻섽섾섿셁", 6, "셊셎", 5, "셖셗"],
        ["9961", "셙셚셛셝", 6, "셦셪", 5, "셱셲셳셵셶셷셹셺셻"],
        ["9981", "셼", 8, "솆", 5, "솏솑솒솓솕솗", 4, "솞솠솢솣솤솦솧솪솫솭솮솯솱", 11, "솾", 5, "쇅쇆쇇쇉쇊쇋쇍", 6, "쇕쇖쇙", 6, "쇡쇢쇣쇥쇦쇧쇩", 6, "쇲쇴", 7, "쇾쇿숁숂숃숅", 6, "숎숐숒", 5, "숚숛숝숞숡숢숣"],
        ["9a41", "숤숥숦숧숪숬숮숰숳숵", 16],
        ["9a61", "쉆쉇쉉", 6, "쉒쉓쉕쉖쉗쉙", 6, "쉡쉢쉣쉤쉦"],
        ["9a81", "쉧", 4, "쉮쉯쉱쉲쉳쉵", 6, "쉾슀슂", 5, "슊", 5, "슑", 6, "슙슚슜슞", 5, "슦슧슩슪슫슮", 5, "슶슸슺", 33, "싞싟싡싢싥", 5, "싮싰싲싳싴싵싷싺싽싾싿쌁", 6, "쌊쌋쌎쌏"],
        ["9b41", "쌐쌑쌒쌖쌗쌙쌚쌛쌝", 6, "쌦쌧쌪", 8],
        ["9b61", "쌳", 17, "썆", 7],
        ["9b81", "썎", 25, "썪썫썭썮썯썱썳", 4, "썺썻썾", 5, "쎅쎆쎇쎉쎊쎋쎍", 50, "쏁", 22, "쏚"],
        ["9c41", "쏛쏝쏞쏡쏣", 4, "쏪쏫쏬쏮", 5, "쏶쏷쏹", 5],
        ["9c61", "쏿", 8, "쐉", 6, "쐑", 9],
        ["9c81", "쐛", 8, "쐥", 6, "쐭쐮쐯쐱쐲쐳쐵", 6, "쐾", 9, "쑉", 26, "쑦쑧쑩쑪쑫쑭", 6, "쑶쑷쑸쑺", 5, "쒁", 18, "쒕", 6, "쒝", 12],
        ["9d41", "쒪", 13, "쒹쒺쒻쒽", 8],
        ["9d61", "쓆", 25],
        ["9d81", "쓠", 8, "쓪", 5, "쓲쓳쓵쓶쓷쓹쓻쓼쓽쓾씂", 9, "씍씎씏씑씒씓씕", 6, "씝", 10, "씪씫씭씮씯씱", 6, "씺씼씾", 5, "앆앇앋앏앐앑앒앖앚앛앜앟앢앣앥앦앧앩", 6, "앲앶", 5, "앾앿얁얂얃얅얆얈얉얊얋얎얐얒얓얔"],
        ["9e41", "얖얙얚얛얝얞얟얡", 7, "얪", 9, "얶"],
        ["9e61", "얷얺얿", 4, "엋엍엏엒엓엕엖엗엙", 6, "엢엤엦엧"],
        ["9e81", "엨엩엪엫엯엱엲엳엵엸엹엺엻옂옃옄옉옊옋옍옎옏옑", 6, "옚옝", 6, "옦옧옩옪옫옯옱옲옶옸옺옼옽옾옿왂왃왅왆왇왉", 6, "왒왖", 5, "왞왟왡", 10, "왭왮왰왲", 5, "왺왻왽왾왿욁", 6, "욊욌욎", 5, "욖욗욙욚욛욝", 6, "욦"],
        ["9f41", "욨욪", 5, "욲욳욵욶욷욻", 4, "웂웄웆", 5, "웎"],
        ["9f61", "웏웑웒웓웕", 6, "웞웟웢", 5, "웪웫웭웮웯웱웲"],
        ["9f81", "웳", 4, "웺웻웼웾", 5, "윆윇윉윊윋윍", 6, "윖윘윚", 5, "윢윣윥윦윧윩", 6, "윲윴윶윸윹윺윻윾윿읁읂읃읅", 4, "읋읎읐읙읚읛읝읞읟읡", 6, "읩읪읬", 7, "읶읷읹읺읻읿잀잁잂잆잋잌잍잏잒잓잕잙잛", 4, "잢잧", 4, "잮잯잱잲잳잵잶잷"],
        ["a041", "잸잹잺잻잾쟂", 5, "쟊쟋쟍쟏쟑", 6, "쟙쟚쟛쟜"],
        ["a061", "쟞", 5, "쟥쟦쟧쟩쟪쟫쟭", 13],
        ["a081", "쟻", 4, "젂젃젅젆젇젉젋", 4, "젒젔젗", 4, "젞젟젡젢젣젥", 6, "젮젰젲", 5, "젹젺젻젽젾젿졁", 6, "졊졋졎", 5, "졕", 26, "졲졳졵졶졷졹졻", 4, "좂좄좈좉좊좎", 5, "좕", 7, "좞좠좢좣좤"],
        ["a141", "좥좦좧좩", 18, "좾좿죀죁"],
        ["a161", "죂죃죅죆죇죉죊죋죍", 6, "죖죘죚", 5, "죢죣죥"],
        ["a181", "죦", 14, "죶", 5, "죾죿줁줂줃줇", 4, "줎　、。·‥…¨〃­―∥＼∼‘’“”〔〕〈", 9, "±×÷≠≤≥∞∴°′″℃Å￠￡￥♂♀∠⊥⌒∂∇≡≒§※☆★○●◎◇◆□■△▲▽▼→←↑↓↔〓≪≫√∽∝∵∫∬∈∋⊆⊇⊂⊃∪∩∧∨￢"],
        ["a241", "줐줒", 5, "줙", 18],
        ["a261", "줭", 6, "줵", 18],
        ["a281", "쥈", 7, "쥒쥓쥕쥖쥗쥙", 6, "쥢쥤", 7, "쥭쥮쥯⇒⇔∀∃´～ˇ˘˝˚˙¸˛¡¿ː∮∑∏¤℉‰◁◀▷▶♤♠♡♥♧♣⊙◈▣◐◑▒▤▥▨▧▦▩♨☏☎☜☞¶†‡↕↗↙↖↘♭♩♪♬㉿㈜№㏇™㏂㏘℡€®"],
        ["a341", "쥱쥲쥳쥵", 6, "쥽", 10, "즊즋즍즎즏"],
        ["a361", "즑", 6, "즚즜즞", 16],
        ["a381", "즯", 16, "짂짃짅짆짉짋", 4, "짒짔짗짘짛！", 58, "￦］", 32, "￣"],
        ["a441", "짞짟짡짣짥짦짨짩짪짫짮짲", 5, "짺짻짽짾짿쨁쨂쨃쨄"],
        ["a461", "쨅쨆쨇쨊쨎", 5, "쨕쨖쨗쨙", 12],
        ["a481", "쨦쨧쨨쨪", 28, "ㄱ", 93],
        ["a541", "쩇", 4, "쩎쩏쩑쩒쩓쩕", 6, "쩞쩢", 5, "쩩쩪"],
        ["a561", "쩫", 17, "쩾", 5, "쪅쪆"],
        ["a581", "쪇", 16, "쪙", 14, "ⅰ", 9],
        ["a5b0", "Ⅰ", 9],
        ["a5c1", "Α", 16, "Σ", 6],
        ["a5e1", "α", 16, "σ", 6],
        ["a641", "쪨", 19, "쪾쪿쫁쫂쫃쫅"],
        ["a661", "쫆", 5, "쫎쫐쫒쫔쫕쫖쫗쫚", 5, "쫡", 6],
        ["a681", "쫨쫩쫪쫫쫭", 6, "쫵", 18, "쬉쬊─│┌┐┘└├┬┤┴┼━┃┏┓┛┗┣┳┫┻╋┠┯┨┷┿┝┰┥┸╂┒┑┚┙┖┕┎┍┞┟┡┢┦┧┩┪┭┮┱┲┵┶┹┺┽┾╀╁╃", 7],
        ["a741", "쬋", 4, "쬑쬒쬓쬕쬖쬗쬙", 6, "쬢", 7],
        ["a761", "쬪", 22, "쭂쭃쭄"],
        ["a781", "쭅쭆쭇쭊쭋쭍쭎쭏쭑", 6, "쭚쭛쭜쭞", 5, "쭥", 7, "㎕㎖㎗ℓ㎘㏄㎣㎤㎥㎦㎙", 9, "㏊㎍㎎㎏㏏㎈㎉㏈㎧㎨㎰", 9, "㎀", 4, "㎺", 5, "㎐", 4, "Ω㏀㏁㎊㎋㎌㏖㏅㎭㎮㎯㏛㎩㎪㎫㎬㏝㏐㏓㏃㏉㏜㏆"],
        ["a841", "쭭", 10, "쭺", 14],
        ["a861", "쮉", 18, "쮝", 6],
        ["a881", "쮤", 19, "쮹", 11, "ÆÐªĦ"],
        ["a8a6", "Ĳ"],
        ["a8a8", "ĿŁØŒºÞŦŊ"],
        ["a8b1", "㉠", 27, "ⓐ", 25, "①", 14, "½⅓⅔¼¾⅛⅜⅝⅞"],
        ["a941", "쯅", 14, "쯕", 10],
        ["a961", "쯠쯡쯢쯣쯥쯦쯨쯪", 18],
        ["a981", "쯽", 14, "찎찏찑찒찓찕", 6, "찞찟찠찣찤æđðħıĳĸŀłøœßþŧŋŉ㈀", 27, "⒜", 25, "⑴", 14, "¹²³⁴ⁿ₁₂₃₄"],
        ["aa41", "찥찦찪찫찭찯찱", 6, "찺찿", 4, "챆챇챉챊챋챍챎"],
        ["aa61", "챏", 4, "챖챚", 5, "챡챢챣챥챧챩", 6, "챱챲"],
        ["aa81", "챳챴챶", 29, "ぁ", 82],
        ["ab41", "첔첕첖첗첚첛첝첞첟첡", 6, "첪첮", 5, "첶첷첹"],
        ["ab61", "첺첻첽", 6, "쳆쳈쳊", 5, "쳑쳒쳓쳕", 5],
        ["ab81", "쳛", 8, "쳥", 6, "쳭쳮쳯쳱", 12, "ァ", 85],
        ["ac41", "쳾쳿촀촂", 5, "촊촋촍촎촏촑", 6, "촚촜촞촟촠"],
        ["ac61", "촡촢촣촥촦촧촩촪촫촭", 11, "촺", 4],
        ["ac81", "촿", 28, "쵝쵞쵟А", 5, "ЁЖ", 25],
        ["acd1", "а", 5, "ёж", 25],
        ["ad41", "쵡쵢쵣쵥", 6, "쵮쵰쵲", 5, "쵹", 7],
        ["ad61", "춁", 6, "춉", 10, "춖춗춙춚춛춝춞춟"],
        ["ad81", "춠춡춢춣춦춨춪", 5, "춱", 18, "췅"],
        ["ae41", "췆", 5, "췍췎췏췑", 16],
        ["ae61", "췢", 5, "췩췪췫췭췮췯췱", 6, "췺췼췾", 4],
        ["ae81", "츃츅츆츇츉츊츋츍", 6, "츕츖츗츘츚", 5, "츢츣츥츦츧츩츪츫"],
        ["af41", "츬츭츮츯츲츴츶", 19],
        ["af61", "칊", 13, "칚칛칝칞칢", 5, "칪칬"],
        ["af81", "칮", 5, "칶칷칹칺칻칽", 6, "캆캈캊", 5, "캒캓캕캖캗캙"],
        ["b041", "캚", 5, "캢캦", 5, "캮", 12],
        ["b061", "캻", 5, "컂", 19],
        ["b081", "컖", 13, "컦컧컩컪컭", 6, "컶컺", 5, "가각간갇갈갉갊감", 7, "같", 4, "갠갤갬갭갯갰갱갸갹갼걀걋걍걔걘걜거걱건걷걸걺검겁것겄겅겆겉겊겋게겐겔겜겝겟겠겡겨격겪견겯결겸겹겻겼경곁계곈곌곕곗고곡곤곧골곪곬곯곰곱곳공곶과곽관괄괆"],
        ["b141", "켂켃켅켆켇켉", 6, "켒켔켖", 5, "켝켞켟켡켢켣"],
        ["b161", "켥", 6, "켮켲", 5, "켹", 11],
        ["b181", "콅", 14, "콖콗콙콚콛콝", 6, "콦콨콪콫콬괌괍괏광괘괜괠괩괬괭괴괵괸괼굄굅굇굉교굔굘굡굣구국군굳굴굵굶굻굼굽굿궁궂궈궉권궐궜궝궤궷귀귁귄귈귐귑귓규균귤그극근귿글긁금급긋긍긔기긱긴긷길긺김깁깃깅깆깊까깍깎깐깔깖깜깝깟깠깡깥깨깩깬깰깸"],
        ["b241", "콭콮콯콲콳콵콶콷콹", 6, "쾁쾂쾃쾄쾆", 5, "쾍"],
        ["b261", "쾎", 18, "쾢", 5, "쾩"],
        ["b281", "쾪", 5, "쾱", 18, "쿅", 6, "깹깻깼깽꺄꺅꺌꺼꺽꺾껀껄껌껍껏껐껑께껙껜껨껫껭껴껸껼꼇꼈꼍꼐꼬꼭꼰꼲꼴꼼꼽꼿꽁꽂꽃꽈꽉꽐꽜꽝꽤꽥꽹꾀꾄꾈꾐꾑꾕꾜꾸꾹꾼꿀꿇꿈꿉꿋꿍꿎꿔꿜꿨꿩꿰꿱꿴꿸뀀뀁뀄뀌뀐뀔뀜뀝뀨끄끅끈끊끌끎끓끔끕끗끙"],
        ["b341", "쿌", 19, "쿢쿣쿥쿦쿧쿩"],
        ["b361", "쿪", 5, "쿲쿴쿶", 5, "쿽쿾쿿퀁퀂퀃퀅", 5],
        ["b381", "퀋", 5, "퀒", 5, "퀙", 19, "끝끼끽낀낄낌낍낏낑나낙낚난낟날낡낢남납낫", 4, "낱낳내낵낸낼냄냅냇냈냉냐냑냔냘냠냥너넉넋넌널넒넓넘넙넛넜넝넣네넥넨넬넴넵넷넸넹녀녁년녈념녑녔녕녘녜녠노녹논놀놂놈놉놋농높놓놔놘놜놨뇌뇐뇔뇜뇝"],
        ["b441", "퀮", 5, "퀶퀷퀹퀺퀻퀽", 6, "큆큈큊", 5],
        ["b461", "큑큒큓큕큖큗큙", 6, "큡", 10, "큮큯"],
        ["b481", "큱큲큳큵", 6, "큾큿킀킂", 18, "뇟뇨뇩뇬뇰뇹뇻뇽누눅눈눋눌눔눕눗눙눠눴눼뉘뉜뉠뉨뉩뉴뉵뉼늄늅늉느늑는늘늙늚늠늡늣능늦늪늬늰늴니닉닌닐닒님닙닛닝닢다닥닦단닫", 4, "닳담답닷", 4, "닿대댁댄댈댐댑댓댔댕댜더덕덖던덛덜덞덟덤덥"],
        ["b541", "킕", 14, "킦킧킩킪킫킭", 5],
        ["b561", "킳킶킸킺", 5, "탂탃탅탆탇탊", 5, "탒탖", 4],
        ["b581", "탛탞탟탡탢탣탥", 6, "탮탲", 5, "탹", 11, "덧덩덫덮데덱덴델뎀뎁뎃뎄뎅뎌뎐뎔뎠뎡뎨뎬도독돈돋돌돎돐돔돕돗동돛돝돠돤돨돼됐되된될됨됩됫됴두둑둔둘둠둡둣둥둬뒀뒈뒝뒤뒨뒬뒵뒷뒹듀듄듈듐듕드득든듣들듦듬듭듯등듸디딕딘딛딜딤딥딧딨딩딪따딱딴딸"],
        ["b641", "턅", 7, "턎", 17],
        ["b661", "턠", 15, "턲턳턵턶턷턹턻턼턽턾"],
        ["b681", "턿텂텆", 5, "텎텏텑텒텓텕", 6, "텞텠텢", 5, "텩텪텫텭땀땁땃땄땅땋때땍땐땔땜땝땟땠땡떠떡떤떨떪떫떰떱떳떴떵떻떼떽뗀뗄뗌뗍뗏뗐뗑뗘뗬또똑똔똘똥똬똴뙈뙤뙨뚜뚝뚠뚤뚫뚬뚱뛔뛰뛴뛸뜀뜁뜅뜨뜩뜬뜯뜰뜸뜹뜻띄띈띌띔띕띠띤띨띰띱띳띵라락란랄람랍랏랐랑랒랖랗"],
        ["b741", "텮", 13, "텽", 6, "톅톆톇톉톊"],
        ["b761", "톋", 20, "톢톣톥톦톧"],
        ["b781", "톩", 6, "톲톴톶톷톸톹톻톽톾톿퇁", 14, "래랙랜랠램랩랫랬랭랴략랸럇량러럭런럴럼럽럿렀렁렇레렉렌렐렘렙렛렝려력련렬렴렵렷렸령례롄롑롓로록론롤롬롭롯롱롸롼뢍뢨뢰뢴뢸룀룁룃룅료룐룔룝룟룡루룩룬룰룸룹룻룽뤄뤘뤠뤼뤽륀륄륌륏륑류륙륜률륨륩"],
        ["b841", "퇐", 7, "퇙", 17],
        ["b861", "퇫", 8, "퇵퇶퇷퇹", 13],
        ["b881", "툈툊", 5, "툑", 24, "륫륭르륵른를름릅릇릉릊릍릎리릭린릴림립릿링마막만많", 4, "맘맙맛망맞맡맣매맥맨맬맴맵맷맸맹맺먀먁먈먕머먹먼멀멂멈멉멋멍멎멓메멕멘멜멤멥멧멨멩며멱면멸몃몄명몇몌모목몫몬몰몲몸몹못몽뫄뫈뫘뫙뫼"],
        ["b941", "툪툫툮툯툱툲툳툵", 6, "툾퉀퉂", 5, "퉉퉊퉋퉌"],
        ["b961", "퉍", 14, "퉝", 6, "퉥퉦퉧퉨"],
        ["b981", "퉩", 22, "튂튃튅튆튇튉튊튋튌묀묄묍묏묑묘묜묠묩묫무묵묶문묻물묽묾뭄뭅뭇뭉뭍뭏뭐뭔뭘뭡뭣뭬뮈뮌뮐뮤뮨뮬뮴뮷므믄믈믐믓미믹민믿밀밂밈밉밋밌밍및밑바", 4, "받", 4, "밤밥밧방밭배백밴밸뱀뱁뱃뱄뱅뱉뱌뱍뱐뱝버벅번벋벌벎범법벗"],
        ["ba41", "튍튎튏튒튓튔튖", 5, "튝튞튟튡튢튣튥", 6, "튭"],
        ["ba61", "튮튯튰튲", 5, "튺튻튽튾틁틃", 4, "틊틌", 5],
        ["ba81", "틒틓틕틖틗틙틚틛틝", 6, "틦", 9, "틲틳틵틶틷틹틺벙벚베벡벤벧벨벰벱벳벴벵벼벽변별볍볏볐병볕볘볜보복볶본볼봄봅봇봉봐봔봤봬뵀뵈뵉뵌뵐뵘뵙뵤뵨부북분붇불붉붊붐붑붓붕붙붚붜붤붰붸뷔뷕뷘뷜뷩뷰뷴뷸븀븃븅브븍븐블븜븝븟비빅빈빌빎빔빕빗빙빚빛빠빡빤"],
        ["bb41", "틻", 4, "팂팄팆", 5, "팏팑팒팓팕팗", 4, "팞팢팣"],
        ["bb61", "팤팦팧팪팫팭팮팯팱", 6, "팺팾", 5, "퍆퍇퍈퍉"],
        ["bb81", "퍊", 31, "빨빪빰빱빳빴빵빻빼빽뺀뺄뺌뺍뺏뺐뺑뺘뺙뺨뻐뻑뻔뻗뻘뻠뻣뻤뻥뻬뼁뼈뼉뼘뼙뼛뼜뼝뽀뽁뽄뽈뽐뽑뽕뾔뾰뿅뿌뿍뿐뿔뿜뿟뿡쀼쁑쁘쁜쁠쁨쁩삐삑삔삘삠삡삣삥사삭삯산삳살삵삶삼삽삿샀상샅새색샌샐샘샙샛샜생샤"],
        ["bc41", "퍪", 17, "퍾퍿펁펂펃펅펆펇"],
        ["bc61", "펈펉펊펋펎펒", 5, "펚펛펝펞펟펡", 6, "펪펬펮"],
        ["bc81", "펯", 4, "펵펶펷펹펺펻펽", 6, "폆폇폊", 5, "폑", 5, "샥샨샬샴샵샷샹섀섄섈섐섕서", 4, "섣설섦섧섬섭섯섰성섶세섹센셀셈셉셋셌셍셔셕션셜셤셥셧셨셩셰셴셸솅소속솎손솔솖솜솝솟송솥솨솩솬솰솽쇄쇈쇌쇔쇗쇘쇠쇤쇨쇰쇱쇳쇼쇽숀숄숌숍숏숑수숙순숟술숨숩숫숭"],
        ["bd41", "폗폙", 7, "폢폤", 7, "폮폯폱폲폳폵폶폷"],
        ["bd61", "폸폹폺폻폾퐀퐂", 5, "퐉", 13],
        ["bd81", "퐗", 5, "퐞", 25, "숯숱숲숴쉈쉐쉑쉔쉘쉠쉥쉬쉭쉰쉴쉼쉽쉿슁슈슉슐슘슛슝스슥슨슬슭슴습슷승시식신싣실싫심십싯싱싶싸싹싻싼쌀쌈쌉쌌쌍쌓쌔쌕쌘쌜쌤쌥쌨쌩썅써썩썬썰썲썸썹썼썽쎄쎈쎌쏀쏘쏙쏜쏟쏠쏢쏨쏩쏭쏴쏵쏸쐈쐐쐤쐬쐰"],
        ["be41", "퐸", 7, "푁푂푃푅", 14],
        ["be61", "푔", 7, "푝푞푟푡푢푣푥", 7, "푮푰푱푲"],
        ["be81", "푳", 4, "푺푻푽푾풁풃", 4, "풊풌풎", 5, "풕", 8, "쐴쐼쐽쑈쑤쑥쑨쑬쑴쑵쑹쒀쒔쒜쒸쒼쓩쓰쓱쓴쓸쓺쓿씀씁씌씐씔씜씨씩씬씰씸씹씻씽아악안앉않알앍앎앓암압앗았앙앝앞애액앤앨앰앱앳앴앵야약얀얄얇얌얍얏양얕얗얘얜얠얩어억언얹얻얼얽얾엄", 6, "엌엎"],
        ["bf41", "풞", 10, "풪", 14],
        ["bf61", "풹", 18, "퓍퓎퓏퓑퓒퓓퓕"],
        ["bf81", "퓖", 5, "퓝퓞퓠", 7, "퓩퓪퓫퓭퓮퓯퓱", 6, "퓹퓺퓼에엑엔엘엠엡엣엥여역엮연열엶엷염", 5, "옅옆옇예옌옐옘옙옛옜오옥온올옭옮옰옳옴옵옷옹옻와왁완왈왐왑왓왔왕왜왝왠왬왯왱외왹왼욀욈욉욋욍요욕욘욜욤욥욧용우욱운울욹욺움웁웃웅워웍원월웜웝웠웡웨"],
        ["c041", "퓾", 5, "픅픆픇픉픊픋픍", 6, "픖픘", 5],
        ["c061", "픞", 25],
        ["c081", "픸픹픺픻픾픿핁핂핃핅", 6, "핎핐핒", 5, "핚핛핝핞핟핡핢핣웩웬웰웸웹웽위윅윈윌윔윕윗윙유육윤율윰윱윳융윷으윽은을읊음읍읏응", 7, "읜읠읨읫이익인일읽읾잃임입잇있잉잊잎자작잔잖잗잘잚잠잡잣잤장잦재잭잰잴잼잽잿쟀쟁쟈쟉쟌쟎쟐쟘쟝쟤쟨쟬저적전절젊"],
        ["c141", "핤핦핧핪핬핮", 5, "핶핷핹핺핻핽", 6, "햆햊햋"],
        ["c161", "햌햍햎햏햑", 19, "햦햧"],
        ["c181", "햨", 31, "점접젓정젖제젝젠젤젬젭젯젱져젼졀졈졉졌졍졔조족존졸졺좀좁좃종좆좇좋좌좍좔좝좟좡좨좼좽죄죈죌죔죕죗죙죠죡죤죵주죽준줄줅줆줌줍줏중줘줬줴쥐쥑쥔쥘쥠쥡쥣쥬쥰쥴쥼즈즉즌즐즘즙즛증지직진짇질짊짐집짓"],
        ["c241", "헊헋헍헎헏헑헓", 4, "헚헜헞", 5, "헦헧헩헪헫헭헮"],
        ["c261", "헯", 4, "헶헸헺", 5, "혂혃혅혆혇혉", 6, "혒"],
        ["c281", "혖", 5, "혝혞혟혡혢혣혥", 7, "혮", 9, "혺혻징짖짙짚짜짝짠짢짤짧짬짭짯짰짱째짹짼쨀쨈쨉쨋쨌쨍쨔쨘쨩쩌쩍쩐쩔쩜쩝쩟쩠쩡쩨쩽쪄쪘쪼쪽쫀쫄쫌쫍쫏쫑쫓쫘쫙쫠쫬쫴쬈쬐쬔쬘쬠쬡쭁쭈쭉쭌쭐쭘쭙쭝쭤쭸쭹쮜쮸쯔쯤쯧쯩찌찍찐찔찜찝찡찢찧차착찬찮찰참찹찻"],
        ["c341", "혽혾혿홁홂홃홄홆홇홊홌홎홏홐홒홓홖홗홙홚홛홝", 4],
        ["c361", "홢", 4, "홨홪", 5, "홲홳홵", 11],
        ["c381", "횁횂횄횆", 5, "횎횏횑횒횓횕", 7, "횞횠횢", 5, "횩횪찼창찾채책챈챌챔챕챗챘챙챠챤챦챨챰챵처척천철첨첩첫첬청체첵첸첼쳄쳅쳇쳉쳐쳔쳤쳬쳰촁초촉촌촐촘촙촛총촤촨촬촹최쵠쵤쵬쵭쵯쵱쵸춈추축춘출춤춥춧충춰췄췌췐취췬췰췸췹췻췽츄츈츌츔츙츠측츤츨츰츱츳층"],
        ["c441", "횫횭횮횯횱", 7, "횺횼", 7, "훆훇훉훊훋"],
        ["c461", "훍훎훏훐훒훓훕훖훘훚", 5, "훡훢훣훥훦훧훩", 4],
        ["c481", "훮훯훱훲훳훴훶", 5, "훾훿휁휂휃휅", 11, "휒휓휔치칙친칟칠칡침칩칫칭카칵칸칼캄캅캇캉캐캑캔캘캠캡캣캤캥캬캭컁커컥컨컫컬컴컵컷컸컹케켁켄켈켐켑켓켕켜켠켤켬켭켯켰켱켸코콕콘콜콤콥콧콩콰콱콴콸쾀쾅쾌쾡쾨쾰쿄쿠쿡쿤쿨쿰쿱쿳쿵쿼퀀퀄퀑퀘퀭퀴퀵퀸퀼"],
        ["c541", "휕휖휗휚휛휝휞휟휡", 6, "휪휬휮", 5, "휶휷휹"],
        ["c561", "휺휻휽", 6, "흅흆흈흊", 5, "흒흓흕흚", 4],
        ["c581", "흟흢흤흦흧흨흪흫흭흮흯흱흲흳흵", 6, "흾흿힀힂", 5, "힊힋큄큅큇큉큐큔큘큠크큭큰클큼큽킁키킥킨킬킴킵킷킹타탁탄탈탉탐탑탓탔탕태택탠탤탬탭탯탰탱탸턍터턱턴털턺텀텁텃텄텅테텍텐텔템텝텟텡텨텬텼톄톈토톡톤톨톰톱톳통톺톼퇀퇘퇴퇸툇툉툐투툭툰툴툼툽툿퉁퉈퉜"],
        ["c641", "힍힎힏힑", 6, "힚힜힞", 5],
        ["c6a1", "퉤튀튁튄튈튐튑튕튜튠튤튬튱트특튼튿틀틂틈틉틋틔틘틜틤틥티틱틴틸팀팁팃팅파팍팎판팔팖팜팝팟팠팡팥패팩팬팰팸팹팻팼팽퍄퍅퍼퍽펀펄펌펍펏펐펑페펙펜펠펨펩펫펭펴편펼폄폅폈평폐폘폡폣포폭폰폴폼폽폿퐁"],
        ["c7a1", "퐈퐝푀푄표푠푤푭푯푸푹푼푿풀풂품풉풋풍풔풩퓌퓐퓔퓜퓟퓨퓬퓰퓸퓻퓽프픈플픔픕픗피픽핀필핌핍핏핑하학한할핥함합핫항해핵핸핼햄햅햇했행햐향허헉헌헐헒험헙헛헝헤헥헨헬헴헵헷헹혀혁현혈혐협혓혔형혜혠"],
        ["c8a1", "혤혭호혹혼홀홅홈홉홋홍홑화확환활홧황홰홱홴횃횅회획횐횔횝횟횡효횬횰횹횻후훅훈훌훑훔훗훙훠훤훨훰훵훼훽휀휄휑휘휙휜휠휨휩휫휭휴휵휸휼흄흇흉흐흑흔흖흗흘흙흠흡흣흥흩희흰흴흼흽힁히힉힌힐힘힙힛힝"],
        ["caa1", "伽佳假價加可呵哥嘉嫁家暇架枷柯歌珂痂稼苛茄街袈訶賈跏軻迦駕刻却各恪慤殼珏脚覺角閣侃刊墾奸姦干幹懇揀杆柬桿澗癎看磵稈竿簡肝艮艱諫間乫喝曷渴碣竭葛褐蝎鞨勘坎堪嵌感憾戡敢柑橄減甘疳監瞰紺邯鑑鑒龕"],
        ["cba1", "匣岬甲胛鉀閘剛堈姜岡崗康强彊慷江畺疆糠絳綱羌腔舡薑襁講鋼降鱇介价個凱塏愷愾慨改槪漑疥皆盖箇芥蓋豈鎧開喀客坑更粳羹醵倨去居巨拒据據擧渠炬祛距踞車遽鉅鋸乾件健巾建愆楗腱虔蹇鍵騫乞傑杰桀儉劍劒檢"],
        ["cca1", "瞼鈐黔劫怯迲偈憩揭擊格檄激膈覡隔堅牽犬甄絹繭肩見譴遣鵑抉決潔結缺訣兼慊箝謙鉗鎌京俓倞傾儆勁勍卿坰境庚徑慶憬擎敬景暻更梗涇炅烱璟璥瓊痙硬磬竟競絅經耕耿脛莖警輕逕鏡頃頸驚鯨係啓堺契季屆悸戒桂械"],
        ["cda1", "棨溪界癸磎稽系繫繼計誡谿階鷄古叩告呱固姑孤尻庫拷攷故敲暠枯槁沽痼皐睾稿羔考股膏苦苽菰藁蠱袴誥賈辜錮雇顧高鼓哭斛曲梏穀谷鵠困坤崑昆梱棍滾琨袞鯤汨滑骨供公共功孔工恐恭拱控攻珙空蚣貢鞏串寡戈果瓜"],
        ["cea1", "科菓誇課跨過鍋顆廓槨藿郭串冠官寬慣棺款灌琯瓘管罐菅觀貫關館刮恝括适侊光匡壙廣曠洸炚狂珖筐胱鑛卦掛罫乖傀塊壞怪愧拐槐魁宏紘肱轟交僑咬喬嬌嶠巧攪敎校橋狡皎矯絞翹膠蕎蛟較轎郊餃驕鮫丘久九仇俱具勾"],
        ["cfa1", "區口句咎嘔坵垢寇嶇廐懼拘救枸柩構歐毆毬求溝灸狗玖球瞿矩究絿耉臼舅舊苟衢謳購軀逑邱鉤銶駒驅鳩鷗龜國局菊鞠鞫麴君窘群裙軍郡堀屈掘窟宮弓穹窮芎躬倦券勸卷圈拳捲權淃眷厥獗蕨蹶闕机櫃潰詭軌饋句晷歸貴"],
        ["d0a1", "鬼龜叫圭奎揆槻珪硅窺竅糾葵規赳逵閨勻均畇筠菌鈞龜橘克剋劇戟棘極隙僅劤勤懃斤根槿瑾筋芹菫覲謹近饉契今妗擒昑檎琴禁禽芩衾衿襟金錦伋及急扱汲級給亘兢矜肯企伎其冀嗜器圻基埼夔奇妓寄岐崎己幾忌技旗旣"],
        ["d1a1", "朞期杞棋棄機欺氣汽沂淇玘琦琪璂璣畸畿碁磯祁祇祈祺箕紀綺羈耆耭肌記譏豈起錡錤飢饑騎騏驥麒緊佶吉拮桔金喫儺喇奈娜懦懶拏拿癩", 5, "那樂", 4, "諾酪駱亂卵暖欄煖爛蘭難鸞捏捺南嵐枏楠湳濫男藍襤拉"],
        ["d2a1", "納臘蠟衲囊娘廊", 4, "乃來內奈柰耐冷女年撚秊念恬拈捻寧寗努勞奴弩怒擄櫓爐瑙盧", 5, "駑魯", 10, "濃籠聾膿農惱牢磊腦賂雷尿壘", 7, "嫩訥杻紐勒", 5, "能菱陵尼泥匿溺多茶"],
        ["d3a1", "丹亶但單團壇彖斷旦檀段湍短端簞緞蛋袒鄲鍛撻澾獺疸達啖坍憺擔曇淡湛潭澹痰聃膽蕁覃談譚錟沓畓答踏遝唐堂塘幢戇撞棠當糖螳黨代垈坮大對岱帶待戴擡玳臺袋貸隊黛宅德悳倒刀到圖堵塗導屠島嶋度徒悼挑掉搗桃"],
        ["d4a1", "棹櫂淘渡滔濤燾盜睹禱稻萄覩賭跳蹈逃途道都鍍陶韜毒瀆牘犢獨督禿篤纛讀墩惇敦旽暾沌焞燉豚頓乭突仝冬凍動同憧東桐棟洞潼疼瞳童胴董銅兜斗杜枓痘竇荳讀豆逗頭屯臀芚遁遯鈍得嶝橙燈登等藤謄鄧騰喇懶拏癩羅"],
        ["d5a1", "蘿螺裸邏樂洛烙珞絡落諾酪駱丹亂卵欄欒瀾爛蘭鸞剌辣嵐擥攬欖濫籃纜藍襤覽拉臘蠟廊朗浪狼琅瑯螂郞來崍徠萊冷掠略亮倆兩凉梁樑粮粱糧良諒輛量侶儷勵呂廬慮戾旅櫚濾礪藜蠣閭驢驪麗黎力曆歷瀝礫轢靂憐戀攣漣"],
        ["d6a1", "煉璉練聯蓮輦連鍊冽列劣洌烈裂廉斂殮濂簾獵令伶囹寧岺嶺怜玲笭羚翎聆逞鈴零靈領齡例澧禮醴隷勞怒撈擄櫓潞瀘爐盧老蘆虜路輅露魯鷺鹵碌祿綠菉錄鹿麓論壟弄朧瀧瓏籠聾儡瀨牢磊賂賚賴雷了僚寮廖料燎療瞭聊蓼"],
        ["d7a1", "遼鬧龍壘婁屢樓淚漏瘻累縷蔞褸鏤陋劉旒柳榴流溜瀏琉瑠留瘤硫謬類六戮陸侖倫崙淪綸輪律慄栗率隆勒肋凜凌楞稜綾菱陵俚利厘吏唎履悧李梨浬犁狸理璃異痢籬罹羸莉裏裡里釐離鯉吝潾燐璘藺躪隣鱗麟林淋琳臨霖砬"],
        ["d8a1", "立笠粒摩瑪痲碼磨馬魔麻寞幕漠膜莫邈万卍娩巒彎慢挽晩曼滿漫灣瞞萬蔓蠻輓饅鰻唜抹末沫茉襪靺亡妄忘忙望網罔芒茫莽輞邙埋妹媒寐昧枚梅每煤罵買賣邁魅脈貊陌驀麥孟氓猛盲盟萌冪覓免冕勉棉沔眄眠綿緬面麵滅"],
        ["d9a1", "蔑冥名命明暝椧溟皿瞑茗蓂螟酩銘鳴袂侮冒募姆帽慕摸摹暮某模母毛牟牡瑁眸矛耗芼茅謀謨貌木沐牧目睦穆鶩歿沒夢朦蒙卯墓妙廟描昴杳渺猫竗苗錨務巫憮懋戊拇撫无楙武毋無珷畝繆舞茂蕪誣貿霧鵡墨默們刎吻問文"],
        ["daa1", "汶紊紋聞蚊門雯勿沕物味媚尾嵋彌微未梶楣渼湄眉米美薇謎迷靡黴岷悶愍憫敏旻旼民泯玟珉緡閔密蜜謐剝博拍搏撲朴樸泊珀璞箔粕縛膊舶薄迫雹駁伴半反叛拌搬攀斑槃泮潘班畔瘢盤盼磐磻礬絆般蟠返頒飯勃拔撥渤潑"],
        ["dba1", "發跋醱鉢髮魃倣傍坊妨尨幇彷房放方旁昉枋榜滂磅紡肪膀舫芳蒡蚌訪謗邦防龐倍俳北培徘拜排杯湃焙盃背胚裴裵褙賠輩配陪伯佰帛柏栢白百魄幡樊煩燔番磻繁蕃藩飜伐筏罰閥凡帆梵氾汎泛犯範范法琺僻劈壁擘檗璧癖"],
        ["dca1", "碧蘗闢霹便卞弁變辨辯邊別瞥鱉鼈丙倂兵屛幷昞昺柄棅炳甁病秉竝輧餠騈保堡報寶普步洑湺潽珤甫菩補褓譜輔伏僕匐卜宓復服福腹茯蔔複覆輹輻馥鰒本乶俸奉封峯峰捧棒烽熢琫縫蓬蜂逢鋒鳳不付俯傅剖副否咐埠夫婦"],
        ["dda1", "孚孵富府復扶敷斧浮溥父符簿缶腐腑膚艀芙莩訃負賦賻赴趺部釜阜附駙鳧北分吩噴墳奔奮忿憤扮昐汾焚盆粉糞紛芬賁雰不佛弗彿拂崩朋棚硼繃鵬丕備匕匪卑妃婢庇悲憊扉批斐枇榧比毖毗毘沸泌琵痺砒碑秕秘粃緋翡肥"],
        ["dea1", "脾臂菲蜚裨誹譬費鄙非飛鼻嚬嬪彬斌檳殯浜濱瀕牝玭貧賓頻憑氷聘騁乍事些仕伺似使俟僿史司唆嗣四士奢娑寫寺射巳師徙思捨斜斯柶査梭死沙泗渣瀉獅砂社祀祠私篩紗絲肆舍莎蓑蛇裟詐詞謝賜赦辭邪飼駟麝削數朔索"],
        ["dfa1", "傘刪山散汕珊産疝算蒜酸霰乷撒殺煞薩三參杉森渗芟蔘衫揷澁鈒颯上傷像償商喪嘗孀尙峠常床庠廂想桑橡湘爽牀狀相祥箱翔裳觴詳象賞霜塞璽賽嗇塞穡索色牲生甥省笙墅壻嶼序庶徐恕抒捿敍暑曙書栖棲犀瑞筮絮緖署"],
        ["e0a1", "胥舒薯西誓逝鋤黍鼠夕奭席惜昔晳析汐淅潟石碩蓆釋錫仙僊先善嬋宣扇敾旋渲煽琁瑄璇璿癬禪線繕羨腺膳船蘚蟬詵跣選銑鐥饍鮮卨屑楔泄洩渫舌薛褻設說雪齧剡暹殲纖蟾贍閃陝攝涉燮葉城姓宬性惺成星晟猩珹盛省筬"],
        ["e1a1", "聖聲腥誠醒世勢歲洗稅笹細說貰召嘯塑宵小少巢所掃搔昭梳沼消溯瀟炤燒甦疏疎瘙笑篠簫素紹蔬蕭蘇訴逍遡邵銷韶騷俗屬束涑粟續謖贖速孫巽損蓀遜飡率宋悚松淞訟誦送頌刷殺灑碎鎖衰釗修受嗽囚垂壽嫂守岫峀帥愁"],
        ["e2a1", "戍手授搜收數樹殊水洙漱燧狩獸琇璲瘦睡秀穗竪粹綏綬繡羞脩茱蒐蓚藪袖誰讐輸遂邃酬銖銹隋隧隨雖需須首髓鬚叔塾夙孰宿淑潚熟琡璹肅菽巡徇循恂旬栒楯橓殉洵淳珣盾瞬筍純脣舜荀蓴蕣詢諄醇錞順馴戌術述鉥崇崧"],
        ["e3a1", "嵩瑟膝蝨濕拾習褶襲丞乘僧勝升承昇繩蠅陞侍匙嘶始媤尸屎屍市弑恃施是時枾柴猜矢示翅蒔蓍視試詩諡豕豺埴寔式息拭植殖湜熄篒蝕識軾食飾伸侁信呻娠宸愼新晨燼申神紳腎臣莘薪藎蜃訊身辛辰迅失室實悉審尋心沁"],
        ["e4a1", "沈深瀋甚芯諶什十拾雙氏亞俄兒啞娥峨我牙芽莪蛾衙訝阿雅餓鴉鵝堊岳嶽幄惡愕握樂渥鄂鍔顎鰐齷安岸按晏案眼雁鞍顔鮟斡謁軋閼唵岩巖庵暗癌菴闇壓押狎鴨仰央怏昻殃秧鴦厓哀埃崖愛曖涯碍艾隘靄厄扼掖液縊腋額"],
        ["e5a1", "櫻罌鶯鸚也倻冶夜惹揶椰爺耶若野弱掠略約若葯蒻藥躍亮佯兩凉壤孃恙揚攘敭暘梁楊樣洋瀁煬痒瘍禳穰糧羊良襄諒讓釀陽量養圄御於漁瘀禦語馭魚齬億憶抑檍臆偃堰彦焉言諺孼蘖俺儼嚴奄掩淹嶪業円予余勵呂女如廬"],
        ["e6a1", "旅歟汝濾璵礖礪與艅茹輿轝閭餘驪麗黎亦力域役易曆歷疫繹譯轢逆驛嚥堧姸娟宴年延憐戀捐挻撚椽沇沿涎涓淵演漣烟然煙煉燃燕璉硏硯秊筵緣練縯聯衍軟輦蓮連鉛鍊鳶列劣咽悅涅烈熱裂說閱厭廉念捻染殮炎焰琰艶苒"],
        ["e7a1", "簾閻髥鹽曄獵燁葉令囹塋寧嶺嶸影怜映暎楹榮永泳渶潁濚瀛瀯煐營獰玲瑛瑩瓔盈穎纓羚聆英詠迎鈴鍈零霙靈領乂倪例刈叡曳汭濊猊睿穢芮藝蘂禮裔詣譽豫醴銳隸霓預五伍俉傲午吾吳嗚塢墺奧娛寤悟惡懊敖旿晤梧汚澳"],
        ["e8a1", "烏熬獒筽蜈誤鰲鼇屋沃獄玉鈺溫瑥瘟穩縕蘊兀壅擁瓮甕癰翁邕雍饔渦瓦窩窪臥蛙蝸訛婉完宛梡椀浣玩琓琬碗緩翫脘腕莞豌阮頑曰往旺枉汪王倭娃歪矮外嵬巍猥畏了僚僥凹堯夭妖姚寥寮尿嶢拗搖撓擾料曜樂橈燎燿瑤療"],
        ["e9a1", "窈窯繇繞耀腰蓼蟯要謠遙遼邀饒慾欲浴縟褥辱俑傭冗勇埇墉容庸慂榕涌湧溶熔瑢用甬聳茸蓉踊鎔鏞龍于佑偶優又友右宇寓尤愚憂旴牛玗瑀盂祐禑禹紆羽芋藕虞迂遇郵釪隅雨雩勖彧旭昱栯煜稶郁頊云暈橒殞澐熉耘芸蕓"],
        ["eaa1", "運隕雲韻蔚鬱亐熊雄元原員圓園垣媛嫄寃怨愿援沅洹湲源爰猿瑗苑袁轅遠阮院願鴛月越鉞位偉僞危圍委威尉慰暐渭爲瑋緯胃萎葦蔿蝟衛褘謂違韋魏乳侑儒兪劉唯喩孺宥幼幽庾悠惟愈愉揄攸有杻柔柚柳楡楢油洧流游溜"],
        ["eba1", "濡猶猷琉瑜由留癒硫紐維臾萸裕誘諛諭踰蹂遊逾遺酉釉鍮類六堉戮毓肉育陸倫允奫尹崙淪潤玧胤贇輪鈗閏律慄栗率聿戎瀜絨融隆垠恩慇殷誾銀隱乙吟淫蔭陰音飮揖泣邑凝應膺鷹依倚儀宜意懿擬椅毅疑矣義艤薏蟻衣誼"],
        ["eca1", "議醫二以伊利吏夷姨履已弛彛怡易李梨泥爾珥理異痍痢移罹而耳肄苡荑裏裡貽貳邇里離飴餌匿溺瀷益翊翌翼謚人仁刃印吝咽因姻寅引忍湮燐璘絪茵藺蚓認隣靭靷鱗麟一佚佾壹日溢逸鎰馹任壬妊姙恁林淋稔臨荏賃入卄"],
        ["eda1", "立笠粒仍剩孕芿仔刺咨姉姿子字孜恣慈滋炙煮玆瓷疵磁紫者自茨蔗藉諮資雌作勺嚼斫昨灼炸爵綽芍酌雀鵲孱棧殘潺盞岑暫潛箴簪蠶雜丈仗匠場墻壯奬將帳庄張掌暲杖樟檣欌漿牆狀獐璋章粧腸臟臧莊葬蔣薔藏裝贓醬長"],
        ["eea1", "障再哉在宰才材栽梓渽滓災縡裁財載齋齎爭箏諍錚佇低儲咀姐底抵杵楮樗沮渚狙猪疽箸紵苧菹著藷詛貯躇這邸雎齟勣吊嫡寂摘敵滴狄炙的積笛籍績翟荻謫賊赤跡蹟迪迹適鏑佃佺傳全典前剪塡塼奠專展廛悛戰栓殿氈澱"],
        ["efa1", "煎琠田甸畑癲筌箋箭篆纏詮輾轉鈿銓錢鐫電顚顫餞切截折浙癤竊節絶占岾店漸点粘霑鮎點接摺蝶丁井亭停偵呈姃定幀庭廷征情挺政整旌晶晸柾楨檉正汀淀淨渟湞瀞炡玎珽町睛碇禎程穽精綎艇訂諪貞鄭酊釘鉦鋌錠霆靖"],
        ["f0a1", "靜頂鼎制劑啼堤帝弟悌提梯濟祭第臍薺製諸蹄醍除際霽題齊俎兆凋助嘲弔彫措操早晁曺曹朝條棗槽漕潮照燥爪璪眺祖祚租稠窕粗糟組繰肇藻蚤詔調趙躁造遭釣阻雕鳥族簇足鏃存尊卒拙猝倧宗從悰慫棕淙琮種終綜縱腫"],
        ["f1a1", "踪踵鍾鐘佐坐左座挫罪主住侏做姝胄呪周嗾奏宙州廚晝朱柱株注洲湊澍炷珠疇籌紂紬綢舟蛛註誅走躊輳週酎酒鑄駐竹粥俊儁准埈寯峻晙樽浚準濬焌畯竣蠢逡遵雋駿茁中仲衆重卽櫛楫汁葺增憎曾拯烝甑症繒蒸證贈之只"],
        ["f2a1", "咫地址志持指摯支旨智枝枳止池沚漬知砥祉祗紙肢脂至芝芷蜘誌識贄趾遲直稙稷織職唇嗔塵振搢晉晋桭榛殄津溱珍瑨璡畛疹盡眞瞋秦縉縝臻蔯袗診賑軫辰進鎭陣陳震侄叱姪嫉帙桎瓆疾秩窒膣蛭質跌迭斟朕什執潗緝輯"],
        ["f3a1", "鏶集徵懲澄且侘借叉嗟嵯差次此磋箚茶蹉車遮捉搾着窄錯鑿齪撰澯燦璨瓚竄簒纂粲纘讚贊鑽餐饌刹察擦札紮僭參塹慘慙懺斬站讒讖倉倡創唱娼廠彰愴敞昌昶暢槍滄漲猖瘡窓脹艙菖蒼債埰寀寨彩採砦綵菜蔡采釵冊柵策"],
        ["f4a1", "責凄妻悽處倜刺剔尺慽戚拓擲斥滌瘠脊蹠陟隻仟千喘天川擅泉淺玔穿舛薦賤踐遷釧闡阡韆凸哲喆徹撤澈綴輟轍鐵僉尖沾添甛瞻簽籤詹諂堞妾帖捷牒疊睫諜貼輒廳晴淸聽菁請靑鯖切剃替涕滯締諦逮遞體初剿哨憔抄招梢"],
        ["f5a1", "椒楚樵炒焦硝礁礎秒稍肖艸苕草蕉貂超酢醋醮促囑燭矗蜀觸寸忖村邨叢塚寵悤憁摠總聰蔥銃撮催崔最墜抽推椎楸樞湫皺秋芻萩諏趨追鄒酋醜錐錘鎚雛騶鰍丑畜祝竺筑築縮蓄蹙蹴軸逐春椿瑃出朮黜充忠沖蟲衝衷悴膵萃"],
        ["f6a1", "贅取吹嘴娶就炊翠聚脆臭趣醉驟鷲側仄厠惻測層侈値嗤峙幟恥梔治淄熾痔痴癡稚穉緇緻置致蚩輜雉馳齒則勅飭親七柒漆侵寢枕沈浸琛砧針鍼蟄秤稱快他咤唾墮妥惰打拖朶楕舵陀馱駝倬卓啄坼度托拓擢晫柝濁濯琢琸託"],
        ["f7a1", "鐸呑嘆坦彈憚歎灘炭綻誕奪脫探眈耽貪塔搭榻宕帑湯糖蕩兌台太怠態殆汰泰笞胎苔跆邰颱宅擇澤撑攄兎吐土討慟桶洞痛筒統通堆槌腿褪退頹偸套妬投透鬪慝特闖坡婆巴把播擺杷波派爬琶破罷芭跛頗判坂板版瓣販辦鈑"],
        ["f8a1", "阪八叭捌佩唄悖敗沛浿牌狽稗覇貝彭澎烹膨愎便偏扁片篇編翩遍鞭騙貶坪平枰萍評吠嬖幣廢弊斃肺蔽閉陛佈包匍匏咆哺圃布怖抛抱捕暴泡浦疱砲胞脯苞葡蒲袍褒逋鋪飽鮑幅暴曝瀑爆輻俵剽彪慓杓標漂瓢票表豹飇飄驃"],
        ["f9a1", "品稟楓諷豊風馮彼披疲皮被避陂匹弼必泌珌畢疋筆苾馝乏逼下何厦夏廈昰河瑕荷蝦賀遐霞鰕壑學虐謔鶴寒恨悍旱汗漢澣瀚罕翰閑閒限韓割轄函含咸啣喊檻涵緘艦銜陷鹹合哈盒蛤閤闔陜亢伉姮嫦巷恒抗杭桁沆港缸肛航"],
        ["faa1", "行降項亥偕咳垓奚孩害懈楷海瀣蟹解該諧邂駭骸劾核倖幸杏荇行享向嚮珦鄕響餉饗香噓墟虛許憲櫶獻軒歇險驗奕爀赫革俔峴弦懸晛泫炫玄玹現眩睍絃絢縣舷衒見賢鉉顯孑穴血頁嫌俠協夾峽挾浹狹脅脇莢鋏頰亨兄刑型"],
        ["fba1", "形泂滎瀅灐炯熒珩瑩荊螢衡逈邢鎣馨兮彗惠慧暳蕙蹊醯鞋乎互呼壕壺好岵弧戶扈昊晧毫浩淏湖滸澔濠濩灝狐琥瑚瓠皓祜糊縞胡芦葫蒿虎號蝴護豪鎬頀顥惑或酷婚昏混渾琿魂忽惚笏哄弘汞泓洪烘紅虹訌鴻化和嬅樺火畵"],
        ["fca1", "禍禾花華話譁貨靴廓擴攫確碻穫丸喚奐宦幻患換歡晥桓渙煥環紈還驩鰥活滑猾豁闊凰幌徨恍惶愰慌晃晄榥況湟滉潢煌璜皇篁簧荒蝗遑隍黃匯回廻徊恢悔懷晦會檜淮澮灰獪繪膾茴蛔誨賄劃獲宖橫鐄哮嚆孝效斅曉梟涍淆"],
        ["fda1", "爻肴酵驍侯候厚后吼喉嗅帿後朽煦珝逅勛勳塤壎焄熏燻薰訓暈薨喧暄煊萱卉喙毁彙徽揮暉煇諱輝麾休携烋畦虧恤譎鷸兇凶匈洶胸黑昕欣炘痕吃屹紇訖欠欽歆吸恰洽翕興僖凞喜噫囍姬嬉希憙憘戱晞曦熙熹熺犧禧稀羲詰"]
    ];
    module.exports = json;
},
/* iconv-lite/encodings/tables/cp950.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = [
        ["0", "\u0000", 127],
        ["a140", "　，、。．‧；：？！︰…‥﹐﹑﹒·﹔﹕﹖﹗｜–︱—︳╴︴﹏（）︵︶｛｝︷︸〔〕︹︺【】︻︼《》︽︾〈〉︿﹀「」﹁﹂『』﹃﹄﹙﹚"],
        ["a1a1", "﹛﹜﹝﹞‘’“”〝〞‵′＃＆＊※§〃○●△▲◎☆★◇◆□■▽▼㊣℅¯￣＿ˍ﹉﹊﹍﹎﹋﹌﹟﹠﹡＋－×÷±√＜＞＝≦≧≠∞≒≡﹢", 4, "～∩∪⊥∠∟⊿㏒㏑∫∮∵∴♀♂⊕⊙↑↓←→↖↗↙↘∥∣／"],
        ["a240", "＼∕﹨＄￥〒￠￡％＠℃℉﹩﹪﹫㏕㎜㎝㎞㏎㎡㎎㎏㏄°兙兛兞兝兡兣嗧瓩糎▁", 7, "▏▎▍▌▋▊▉┼┴┬┤├▔─│▕┌┐└┘╭"],
        ["a2a1", "╮╰╯═╞╪╡◢◣◥◤╱╲╳０", 9, "Ⅰ", 9, "〡", 8, "十卄卅Ａ", 25, "ａ", 21],
        ["a340", "ｗｘｙｚΑ", 16, "Σ", 6, "α", 16, "σ", 6, "ㄅ", 10],
        ["a3a1", "ㄐ", 25, "˙ˉˊˇˋ"],
        ["a3e1", "€"],
        ["a440", "一乙丁七乃九了二人儿入八几刀刁力匕十卜又三下丈上丫丸凡久么也乞于亡兀刃勺千叉口土士夕大女子孑孓寸小尢尸山川工己已巳巾干廾弋弓才"],
        ["a4a1", "丑丐不中丰丹之尹予云井互五亢仁什仃仆仇仍今介仄元允內六兮公冗凶分切刈勻勾勿化匹午升卅卞厄友及反壬天夫太夭孔少尤尺屯巴幻廿弔引心戈戶手扎支文斗斤方日曰月木欠止歹毋比毛氏水火爪父爻片牙牛犬王丙"],
        ["a540", "世丕且丘主乍乏乎以付仔仕他仗代令仙仞充兄冉冊冬凹出凸刊加功包匆北匝仟半卉卡占卯卮去可古右召叮叩叨叼司叵叫另只史叱台句叭叻四囚外"],
        ["a5a1", "央失奴奶孕它尼巨巧左市布平幼弁弘弗必戊打扔扒扑斥旦朮本未末札正母民氐永汁汀氾犯玄玉瓜瓦甘生用甩田由甲申疋白皮皿目矛矢石示禾穴立丞丟乒乓乩亙交亦亥仿伉伙伊伕伍伐休伏仲件任仰仳份企伋光兇兆先全"],
        ["a640", "共再冰列刑划刎刖劣匈匡匠印危吉吏同吊吐吁吋各向名合吃后吆吒因回囝圳地在圭圬圯圩夙多夷夸妄奸妃好她如妁字存宇守宅安寺尖屹州帆并年"],
        ["a6a1", "式弛忙忖戎戌戍成扣扛托收早旨旬旭曲曳有朽朴朱朵次此死氖汝汗汙江池汐汕污汛汍汎灰牟牝百竹米糸缶羊羽老考而耒耳聿肉肋肌臣自至臼舌舛舟艮色艾虫血行衣西阡串亨位住佇佗佞伴佛何估佐佑伽伺伸佃佔似但佣"],
        ["a740", "作你伯低伶余佝佈佚兌克免兵冶冷別判利刪刨劫助努劬匣即卵吝吭吞吾否呎吧呆呃吳呈呂君吩告吹吻吸吮吵吶吠吼呀吱含吟听囪困囤囫坊坑址坍"],
        ["a7a1", "均坎圾坐坏圻壯夾妝妒妨妞妣妙妖妍妤妓妊妥孝孜孚孛完宋宏尬局屁尿尾岐岑岔岌巫希序庇床廷弄弟彤形彷役忘忌志忍忱快忸忪戒我抄抗抖技扶抉扭把扼找批扳抒扯折扮投抓抑抆改攻攸旱更束李杏材村杜杖杞杉杆杠"],
        ["a840", "杓杗步每求汞沙沁沈沉沅沛汪決沐汰沌汨沖沒汽沃汲汾汴沆汶沍沔沘沂灶灼災灸牢牡牠狄狂玖甬甫男甸皂盯矣私秀禿究系罕肖肓肝肘肛肚育良芒"],
        ["a8a1", "芋芍見角言谷豆豕貝赤走足身車辛辰迂迆迅迄巡邑邢邪邦那酉釆里防阮阱阪阬並乖乳事些亞享京佯依侍佳使佬供例來侃佰併侈佩佻侖佾侏侑佺兔兒兕兩具其典冽函刻券刷刺到刮制剁劾劻卒協卓卑卦卷卸卹取叔受味呵"],
        ["a940", "咖呸咕咀呻呷咄咒咆呼咐呱呶和咚呢周咋命咎固垃坷坪坩坡坦坤坼夜奉奇奈奄奔妾妻委妹妮姑姆姐姍始姓姊妯妳姒姅孟孤季宗定官宜宙宛尚屈居"],
        ["a9a1", "屆岷岡岸岩岫岱岳帘帚帖帕帛帑幸庚店府底庖延弦弧弩往征彿彼忝忠忽念忿怏怔怯怵怖怪怕怡性怩怫怛或戕房戾所承拉拌拄抿拂抹拒招披拓拔拋拈抨抽押拐拙拇拍抵拚抱拘拖拗拆抬拎放斧於旺昔易昌昆昂明昀昏昕昊"],
        ["aa40", "昇服朋杭枋枕東果杳杷枇枝林杯杰板枉松析杵枚枓杼杪杲欣武歧歿氓氛泣注泳沱泌泥河沽沾沼波沫法泓沸泄油況沮泗泅泱沿治泡泛泊沬泯泜泖泠"],
        ["aaa1", "炕炎炒炊炙爬爭爸版牧物狀狎狙狗狐玩玨玟玫玥甽疝疙疚的盂盲直知矽社祀祁秉秈空穹竺糾罔羌羋者肺肥肢肱股肫肩肴肪肯臥臾舍芳芝芙芭芽芟芹花芬芥芯芸芣芰芾芷虎虱初表軋迎返近邵邸邱邶采金長門阜陀阿阻附"],
        ["ab40", "陂隹雨青非亟亭亮信侵侯便俠俑俏保促侶俘俟俊俗侮俐俄係俚俎俞侷兗冒冑冠剎剃削前剌剋則勇勉勃勁匍南卻厚叛咬哀咨哎哉咸咦咳哇哂咽咪品"],
        ["aba1", "哄哈咯咫咱咻咩咧咿囿垂型垠垣垢城垮垓奕契奏奎奐姜姘姿姣姨娃姥姪姚姦威姻孩宣宦室客宥封屎屏屍屋峙峒巷帝帥帟幽庠度建弈弭彥很待徊律徇後徉怒思怠急怎怨恍恰恨恢恆恃恬恫恪恤扁拜挖按拼拭持拮拽指拱拷"],
        ["ac40", "拯括拾拴挑挂政故斫施既春昭映昧是星昨昱昤曷柿染柱柔某柬架枯柵柩柯柄柑枴柚查枸柏柞柳枰柙柢柝柒歪殃殆段毒毗氟泉洋洲洪流津洌洱洞洗"],
        ["aca1", "活洽派洶洛泵洹洧洸洩洮洵洎洫炫為炳炬炯炭炸炮炤爰牲牯牴狩狠狡玷珊玻玲珍珀玳甚甭畏界畎畋疫疤疥疢疣癸皆皇皈盈盆盃盅省盹相眉看盾盼眇矜砂研砌砍祆祉祈祇禹禺科秒秋穿突竿竽籽紂紅紀紉紇約紆缸美羿耄"],
        ["ad40", "耐耍耑耶胖胥胚胃胄背胡胛胎胞胤胝致舢苧范茅苣苛苦茄若茂茉苒苗英茁苜苔苑苞苓苟苯茆虐虹虻虺衍衫要觔計訂訃貞負赴赳趴軍軌述迦迢迪迥"],
        ["ada1", "迭迫迤迨郊郎郁郃酋酊重閂限陋陌降面革韋韭音頁風飛食首香乘亳倌倍倣俯倦倥俸倩倖倆值借倚倒們俺倀倔倨俱倡個候倘俳修倭倪俾倫倉兼冤冥冢凍凌准凋剖剜剔剛剝匪卿原厝叟哨唐唁唷哼哥哲唆哺唔哩哭員唉哮哪"],
        ["ae40", "哦唧唇哽唏圃圄埂埔埋埃堉夏套奘奚娑娘娜娟娛娓姬娠娣娩娥娌娉孫屘宰害家宴宮宵容宸射屑展屐峭峽峻峪峨峰島崁峴差席師庫庭座弱徒徑徐恙"],
        ["aea1", "恣恥恐恕恭恩息悄悟悚悍悔悌悅悖扇拳挈拿捎挾振捕捂捆捏捉挺捐挽挪挫挨捍捌效敉料旁旅時晉晏晃晒晌晅晁書朔朕朗校核案框桓根桂桔栩梳栗桌桑栽柴桐桀格桃株桅栓栘桁殊殉殷氣氧氨氦氤泰浪涕消涇浦浸海浙涓"],
        ["af40", "浬涉浮浚浴浩涌涊浹涅浥涔烊烘烤烙烈烏爹特狼狹狽狸狷玆班琉珮珠珪珞畔畝畜畚留疾病症疲疳疽疼疹痂疸皋皰益盍盎眩真眠眨矩砰砧砸砝破砷"],
        ["afa1", "砥砭砠砟砲祕祐祠祟祖神祝祗祚秤秣秧租秦秩秘窄窈站笆笑粉紡紗紋紊素索純紐紕級紜納紙紛缺罟羔翅翁耆耘耕耙耗耽耿胱脂胰脅胭胴脆胸胳脈能脊胼胯臭臬舀舐航舫舨般芻茫荒荔荊茸荐草茵茴荏茲茹茶茗荀茱茨荃"],
        ["b040", "虔蚊蚪蚓蚤蚩蚌蚣蚜衰衷袁袂衽衹記訐討訌訕訊託訓訖訏訑豈豺豹財貢起躬軒軔軏辱送逆迷退迺迴逃追逅迸邕郡郝郢酒配酌釘針釗釜釙閃院陣陡"],
        ["b0a1", "陛陝除陘陞隻飢馬骨高鬥鬲鬼乾偺偽停假偃偌做偉健偶偎偕偵側偷偏倏偯偭兜冕凰剪副勒務勘動匐匏匙匿區匾參曼商啪啦啄啞啡啃啊唱啖問啕唯啤唸售啜唬啣唳啁啗圈國圉域堅堊堆埠埤基堂堵執培夠奢娶婁婉婦婪婀"],
        ["b140", "娼婢婚婆婊孰寇寅寄寂宿密尉專將屠屜屝崇崆崎崛崖崢崑崩崔崙崤崧崗巢常帶帳帷康庸庶庵庾張強彗彬彩彫得徙從徘御徠徜恿患悉悠您惋悴惦悽"],
        ["b1a1", "情悻悵惜悼惘惕惆惟悸惚惇戚戛扈掠控捲掖探接捷捧掘措捱掩掉掃掛捫推掄授掙採掬排掏掀捻捩捨捺敝敖救教敗啟敏敘敕敔斜斛斬族旋旌旎晝晚晤晨晦晞曹勗望梁梯梢梓梵桿桶梱梧梗械梃棄梭梆梅梔條梨梟梡梂欲殺"],
        ["b240", "毫毬氫涎涼淳淙液淡淌淤添淺清淇淋涯淑涮淞淹涸混淵淅淒渚涵淚淫淘淪深淮淨淆淄涪淬涿淦烹焉焊烽烯爽牽犁猜猛猖猓猙率琅琊球理現琍瓠瓶"],
        ["b2a1", "瓷甜產略畦畢異疏痔痕疵痊痍皎盔盒盛眷眾眼眶眸眺硫硃硎祥票祭移窒窕笠笨笛第符笙笞笮粒粗粕絆絃統紮紹紼絀細紳組累終紲紱缽羞羚翌翎習耜聊聆脯脖脣脫脩脰脤舂舵舷舶船莎莞莘荸莢莖莽莫莒莊莓莉莠荷荻荼"],
        ["b340", "莆莧處彪蛇蛀蚶蛄蚵蛆蛋蚱蚯蛉術袞袈被袒袖袍袋覓規訪訝訣訥許設訟訛訢豉豚販責貫貨貪貧赧赦趾趺軛軟這逍通逗連速逝逐逕逞造透逢逖逛途"],
        ["b3a1", "部郭都酗野釵釦釣釧釭釩閉陪陵陳陸陰陴陶陷陬雀雪雩章竟頂頃魚鳥鹵鹿麥麻傢傍傅備傑傀傖傘傚最凱割剴創剩勞勝勛博厥啻喀喧啼喊喝喘喂喜喪喔喇喋喃喳單喟唾喲喚喻喬喱啾喉喫喙圍堯堪場堤堰報堡堝堠壹壺奠"],
        ["b440", "婷媚婿媒媛媧孳孱寒富寓寐尊尋就嵌嵐崴嵇巽幅帽幀幃幾廊廁廂廄弼彭復循徨惑惡悲悶惠愜愣惺愕惰惻惴慨惱愎惶愉愀愒戟扉掣掌描揀揩揉揆揍"],
        ["b4a1", "插揣提握揖揭揮捶援揪換摒揚揹敞敦敢散斑斐斯普晰晴晶景暑智晾晷曾替期朝棺棕棠棘棗椅棟棵森棧棹棒棲棣棋棍植椒椎棉棚楮棻款欺欽殘殖殼毯氮氯氬港游湔渡渲湧湊渠渥渣減湛湘渤湖湮渭渦湯渴湍渺測湃渝渾滋"],
        ["b540", "溉渙湎湣湄湲湩湟焙焚焦焰無然煮焜牌犄犀猶猥猴猩琺琪琳琢琥琵琶琴琯琛琦琨甥甦畫番痢痛痣痙痘痞痠登發皖皓皴盜睏短硝硬硯稍稈程稅稀窘"],
        ["b5a1", "窗窖童竣等策筆筐筒答筍筋筏筑粟粥絞結絨絕紫絮絲絡給絢絰絳善翔翕耋聒肅腕腔腋腑腎脹腆脾腌腓腴舒舜菩萃菸萍菠菅萋菁華菱菴著萊菰萌菌菽菲菊萸萎萄菜萇菔菟虛蛟蛙蛭蛔蛛蛤蛐蛞街裁裂袱覃視註詠評詞証詁"],
        ["b640", "詔詛詐詆訴診訶詖象貂貯貼貳貽賁費賀貴買貶貿貸越超趁跎距跋跚跑跌跛跆軻軸軼辜逮逵週逸進逶鄂郵鄉郾酣酥量鈔鈕鈣鈉鈞鈍鈐鈇鈑閔閏開閑"],
        ["b6a1", "間閒閎隊階隋陽隅隆隍陲隄雁雅雄集雇雯雲韌項順須飧飪飯飩飲飭馮馭黃黍黑亂傭債傲傳僅傾催傷傻傯僇剿剷剽募勦勤勢勣匯嗟嗨嗓嗦嗎嗜嗇嗑嗣嗤嗯嗚嗡嗅嗆嗥嗉園圓塞塑塘塗塚塔填塌塭塊塢塒塋奧嫁嫉嫌媾媽媼"],
        ["b740", "媳嫂媲嵩嵯幌幹廉廈弒彙徬微愚意慈感想愛惹愁愈慎慌慄慍愾愴愧愍愆愷戡戢搓搾搞搪搭搽搬搏搜搔損搶搖搗搆敬斟新暗暉暇暈暖暄暘暍會榔業"],
        ["b7a1", "楚楷楠楔極椰概楊楨楫楞楓楹榆楝楣楛歇歲毀殿毓毽溢溯滓溶滂源溝滇滅溥溘溼溺溫滑準溜滄滔溪溧溴煎煙煩煤煉照煜煬煦煌煥煞煆煨煖爺牒猷獅猿猾瑯瑚瑕瑟瑞瑁琿瑙瑛瑜當畸瘀痰瘁痲痱痺痿痴痳盞盟睛睫睦睞督"],
        ["b840", "睹睪睬睜睥睨睢矮碎碰碗碘碌碉硼碑碓硿祺祿禁萬禽稜稚稠稔稟稞窟窠筷節筠筮筧粱粳粵經絹綑綁綏絛置罩罪署義羨群聖聘肆肄腱腰腸腥腮腳腫"],
        ["b8a1", "腹腺腦舅艇蒂葷落萱葵葦葫葉葬葛萼萵葡董葩葭葆虞虜號蛹蜓蜈蜇蜀蛾蛻蜂蜃蜆蜊衙裟裔裙補裘裝裡裊裕裒覜解詫該詳試詩詰誇詼詣誠話誅詭詢詮詬詹詻訾詨豢貊貉賊資賈賄貲賃賂賅跡跟跨路跳跺跪跤跦躲較載軾輊"],
        ["b940", "辟農運遊道遂達逼違遐遇遏過遍遑逾遁鄒鄗酬酪酩釉鈷鉗鈸鈽鉀鈾鉛鉋鉤鉑鈴鉉鉍鉅鈹鈿鉚閘隘隔隕雍雋雉雊雷電雹零靖靴靶預頑頓頊頒頌飼飴"],
        ["b9a1", "飽飾馳馱馴髡鳩麂鼎鼓鼠僧僮僥僖僭僚僕像僑僱僎僩兢凳劃劂匱厭嗾嘀嘛嘗嗽嘔嘆嘉嘍嘎嗷嘖嘟嘈嘐嗶團圖塵塾境墓墊塹墅塽壽夥夢夤奪奩嫡嫦嫩嫗嫖嫘嫣孵寞寧寡寥實寨寢寤察對屢嶄嶇幛幣幕幗幔廓廖弊彆彰徹慇"],
        ["ba40", "愿態慷慢慣慟慚慘慵截撇摘摔撤摸摟摺摑摧搴摭摻敲斡旗旖暢暨暝榜榨榕槁榮槓構榛榷榻榫榴槐槍榭槌榦槃榣歉歌氳漳演滾漓滴漩漾漠漬漏漂漢"],
        ["baa1", "滿滯漆漱漸漲漣漕漫漯澈漪滬漁滲滌滷熔熙煽熊熄熒爾犒犖獄獐瑤瑣瑪瑰瑭甄疑瘧瘍瘋瘉瘓盡監瞄睽睿睡磁碟碧碳碩碣禎福禍種稱窪窩竭端管箕箋筵算箝箔箏箸箇箄粹粽精綻綰綜綽綾綠緊綴網綱綺綢綿綵綸維緒緇綬"],
        ["bb40", "罰翠翡翟聞聚肇腐膀膏膈膊腿膂臧臺與舔舞艋蓉蒿蓆蓄蒙蒞蒲蒜蓋蒸蓀蓓蒐蒼蓑蓊蜿蜜蜻蜢蜥蜴蜘蝕蜷蜩裳褂裴裹裸製裨褚裯誦誌語誣認誡誓誤"],
        ["bba1", "說誥誨誘誑誚誧豪貍貌賓賑賒赫趙趕跼輔輒輕輓辣遠遘遜遣遙遞遢遝遛鄙鄘鄞酵酸酷酴鉸銀銅銘銖鉻銓銜銨鉼銑閡閨閩閣閥閤隙障際雌雒需靼鞅韶頗領颯颱餃餅餌餉駁骯骰髦魁魂鳴鳶鳳麼鼻齊億儀僻僵價儂儈儉儅凜"],
        ["bc40", "劇劈劉劍劊勰厲嘮嘻嘹嘲嘿嘴嘩噓噎噗噴嘶嘯嘰墀墟增墳墜墮墩墦奭嬉嫻嬋嫵嬌嬈寮寬審寫層履嶝嶔幢幟幡廢廚廟廝廣廠彈影德徵慶慧慮慝慕憂"],
        ["bca1", "慼慰慫慾憧憐憫憎憬憚憤憔憮戮摩摯摹撞撲撈撐撰撥撓撕撩撒撮播撫撚撬撙撢撳敵敷數暮暫暴暱樣樟槨樁樞標槽模樓樊槳樂樅槭樑歐歎殤毅毆漿潼澄潑潦潔澆潭潛潸潮澎潺潰潤澗潘滕潯潠潟熟熬熱熨牖犛獎獗瑩璋璃"],
        ["bd40", "瑾璀畿瘠瘩瘟瘤瘦瘡瘢皚皺盤瞎瞇瞌瞑瞋磋磅確磊碾磕碼磐稿稼穀稽稷稻窯窮箭箱範箴篆篇篁箠篌糊締練緯緻緘緬緝編緣線緞緩綞緙緲緹罵罷羯"],
        ["bda1", "翩耦膛膜膝膠膚膘蔗蔽蔚蓮蔬蔭蔓蔑蔣蔡蔔蓬蔥蓿蔆螂蝴蝶蝠蝦蝸蝨蝙蝗蝌蝓衛衝褐複褒褓褕褊誼諒談諄誕請諸課諉諂調誰論諍誶誹諛豌豎豬賠賞賦賤賬賭賢賣賜質賡赭趟趣踫踐踝踢踏踩踟踡踞躺輝輛輟輩輦輪輜輞"],
        ["be40", "輥適遮遨遭遷鄰鄭鄧鄱醇醉醋醃鋅銻銷鋪銬鋤鋁銳銼鋒鋇鋰銲閭閱霄霆震霉靠鞍鞋鞏頡頫頜颳養餓餒餘駝駐駟駛駑駕駒駙骷髮髯鬧魅魄魷魯鴆鴉"],
        ["bea1", "鴃麩麾黎墨齒儒儘儔儐儕冀冪凝劑劓勳噙噫噹噩噤噸噪器噥噱噯噬噢噶壁墾壇壅奮嬝嬴學寰導彊憲憑憩憊懍憶憾懊懈戰擅擁擋撻撼據擄擇擂操撿擒擔撾整曆曉暹曄曇暸樽樸樺橙橫橘樹橄橢橡橋橇樵機橈歙歷氅濂澱澡"],
        ["bf40", "濃澤濁澧澳激澹澶澦澠澴熾燉燐燒燈燕熹燎燙燜燃燄獨璜璣璘璟璞瓢甌甍瘴瘸瘺盧盥瞠瞞瞟瞥磨磚磬磧禦積穎穆穌穋窺篙簑築篤篛篡篩篦糕糖縊"],
        ["bfa1", "縑縈縛縣縞縝縉縐罹羲翰翱翮耨膳膩膨臻興艘艙蕊蕙蕈蕨蕩蕃蕉蕭蕪蕞螃螟螞螢融衡褪褲褥褫褡親覦諦諺諫諱謀諜諧諮諾謁謂諷諭諳諶諼豫豭貓賴蹄踱踴蹂踹踵輻輯輸輳辨辦遵遴選遲遼遺鄴醒錠錶鋸錳錯錢鋼錫錄錚"],
        ["c040", "錐錦錡錕錮錙閻隧隨險雕霎霑霖霍霓霏靛靜靦鞘頰頸頻頷頭頹頤餐館餞餛餡餚駭駢駱骸骼髻髭鬨鮑鴕鴣鴦鴨鴒鴛默黔龍龜優償儡儲勵嚎嚀嚐嚅嚇"],
        ["c0a1", "嚏壕壓壑壎嬰嬪嬤孺尷屨嶼嶺嶽嶸幫彌徽應懂懇懦懋戲戴擎擊擘擠擰擦擬擱擢擭斂斃曙曖檀檔檄檢檜櫛檣橾檗檐檠歜殮毚氈濘濱濟濠濛濤濫濯澀濬濡濩濕濮濰燧營燮燦燥燭燬燴燠爵牆獰獲璩環璦璨癆療癌盪瞳瞪瞰瞬"],
        ["c140", "瞧瞭矯磷磺磴磯礁禧禪穗窿簇簍篾篷簌篠糠糜糞糢糟糙糝縮績繆縷縲繃縫總縱繅繁縴縹繈縵縿縯罄翳翼聱聲聰聯聳臆臃膺臂臀膿膽臉膾臨舉艱薪"],
        ["c1a1", "薄蕾薜薑薔薯薛薇薨薊虧蟀蟑螳蟒蟆螫螻螺蟈蟋褻褶襄褸褽覬謎謗謙講謊謠謝謄謐豁谿豳賺賽購賸賻趨蹉蹋蹈蹊轄輾轂轅輿避遽還邁邂邀鄹醣醞醜鍍鎂錨鍵鍊鍥鍋錘鍾鍬鍛鍰鍚鍔闊闋闌闈闆隱隸雖霜霞鞠韓顆颶餵騁"],
        ["c240", "駿鮮鮫鮪鮭鴻鴿麋黏點黜黝黛鼾齋叢嚕嚮壙壘嬸彝懣戳擴擲擾攆擺擻擷斷曜朦檳檬櫃檻檸櫂檮檯歟歸殯瀉瀋濾瀆濺瀑瀏燻燼燾燸獷獵璧璿甕癖癘"],
        ["c2a1", "癒瞽瞿瞻瞼礎禮穡穢穠竄竅簫簧簪簞簣簡糧織繕繞繚繡繒繙罈翹翻職聶臍臏舊藏薩藍藐藉薰薺薹薦蟯蟬蟲蟠覆覲觴謨謹謬謫豐贅蹙蹣蹦蹤蹟蹕軀轉轍邇邃邈醫醬釐鎔鎊鎖鎢鎳鎮鎬鎰鎘鎚鎗闔闖闐闕離雜雙雛雞霤鞣鞦"],
        ["c340", "鞭韹額顏題顎顓颺餾餿餽餮馥騎髁鬃鬆魏魎魍鯊鯉鯽鯈鯀鵑鵝鵠黠鼕鼬儳嚥壞壟壢寵龐廬懲懷懶懵攀攏曠曝櫥櫝櫚櫓瀛瀟瀨瀚瀝瀕瀘爆爍牘犢獸"],
        ["c3a1", "獺璽瓊瓣疇疆癟癡矇礙禱穫穩簾簿簸簽簷籀繫繭繹繩繪羅繳羶羹羸臘藩藝藪藕藤藥藷蟻蠅蠍蟹蟾襠襟襖襞譁譜識證譚譎譏譆譙贈贊蹼蹲躇蹶蹬蹺蹴轔轎辭邊邋醱醮鏡鏑鏟鏃鏈鏜鏝鏖鏢鏍鏘鏤鏗鏨關隴難霪霧靡韜韻類"],
        ["c440", "願顛颼饅饉騖騙鬍鯨鯧鯖鯛鶉鵡鵲鵪鵬麒麗麓麴勸嚨嚷嚶嚴嚼壤孀孃孽寶巉懸懺攘攔攙曦朧櫬瀾瀰瀲爐獻瓏癢癥礦礪礬礫竇競籌籃籍糯糰辮繽繼"],
        ["c4a1", "纂罌耀臚艦藻藹蘑藺蘆蘋蘇蘊蠔蠕襤覺觸議譬警譯譟譫贏贍躉躁躅躂醴釋鐘鐃鏽闡霰飄饒饑馨騫騰騷騵鰓鰍鹹麵黨鼯齟齣齡儷儸囁囀囂夔屬巍懼懾攝攜斕曩櫻欄櫺殲灌爛犧瓖瓔癩矓籐纏續羼蘗蘭蘚蠣蠢蠡蠟襪襬覽譴"],
        ["c540", "護譽贓躊躍躋轟辯醺鐮鐳鐵鐺鐸鐲鐫闢霸霹露響顧顥饗驅驃驀騾髏魔魑鰭鰥鶯鶴鷂鶸麝黯鼙齜齦齧儼儻囈囊囉孿巔巒彎懿攤權歡灑灘玀瓤疊癮癬"],
        ["c5a1", "禳籠籟聾聽臟襲襯觼讀贖贗躑躓轡酈鑄鑑鑒霽霾韃韁顫饕驕驍髒鬚鱉鰱鰾鰻鷓鷗鼴齬齪龔囌巖戀攣攫攪曬欐瓚竊籤籣籥纓纖纔臢蘸蘿蠱變邐邏鑣鑠鑤靨顯饜驚驛驗髓體髑鱔鱗鱖鷥麟黴囑壩攬灞癱癲矗罐羈蠶蠹衢讓讒"],
        ["c640", "讖艷贛釀鑪靂靈靄韆顰驟鬢魘鱟鷹鷺鹼鹽鼇齷齲廳欖灣籬籮蠻觀躡釁鑲鑰顱饞髖鬣黌灤矚讚鑷韉驢驥纜讜躪釅鑽鑾鑼鱷鱸黷豔鑿鸚爨驪鬱鸛鸞籲"],
        ["c940", "乂乜凵匚厂万丌乇亍囗兀屮彳丏冇与丮亓仂仉仈冘勼卬厹圠夃夬尐巿旡殳毌气爿丱丼仨仜仩仡仝仚刌匜卌圢圣夗夯宁宄尒尻屴屳帄庀庂忉戉扐氕"],
        ["c9a1", "氶汃氿氻犮犰玊禸肊阞伎优伬仵伔仱伀价伈伝伂伅伢伓伄仴伒冱刓刉刐劦匢匟卍厊吇囡囟圮圪圴夼妀奼妅奻奾奷奿孖尕尥屼屺屻屾巟幵庄异弚彴忕忔忏扜扞扤扡扦扢扙扠扚扥旯旮朾朹朸朻机朿朼朳氘汆汒汜汏汊汔汋"],
        ["ca40", "汌灱牞犴犵玎甪癿穵网艸艼芀艽艿虍襾邙邗邘邛邔阢阤阠阣佖伻佢佉体佤伾佧佒佟佁佘伭伳伿佡冏冹刜刞刡劭劮匉卣卲厎厏吰吷吪呔呅吙吜吥吘"],
        ["caa1", "吽呏呁吨吤呇囮囧囥坁坅坌坉坋坒夆奀妦妘妠妗妎妢妐妏妧妡宎宒尨尪岍岏岈岋岉岒岊岆岓岕巠帊帎庋庉庌庈庍弅弝彸彶忒忑忐忭忨忮忳忡忤忣忺忯忷忻怀忴戺抃抌抎抏抔抇扱扻扺扰抁抈扷扽扲扴攷旰旴旳旲旵杅杇"],
        ["cb40", "杙杕杌杈杝杍杚杋毐氙氚汸汧汫沄沋沏汱汯汩沚汭沇沕沜汦汳汥汻沎灴灺牣犿犽狃狆狁犺狅玕玗玓玔玒町甹疔疕皁礽耴肕肙肐肒肜芐芏芅芎芑芓"],
        ["cba1", "芊芃芄豸迉辿邟邡邥邞邧邠阰阨阯阭丳侘佼侅佽侀侇佶佴侉侄佷佌侗佪侚佹侁佸侐侜侔侞侒侂侕佫佮冞冼冾刵刲刳剆刱劼匊匋匼厒厔咇呿咁咑咂咈呫呺呾呥呬呴呦咍呯呡呠咘呣呧呤囷囹坯坲坭坫坱坰坶垀坵坻坳坴坢"],
        ["cc40", "坨坽夌奅妵妺姏姎妲姌姁妶妼姃姖妱妽姀姈妴姇孢孥宓宕屄屇岮岤岠岵岯岨岬岟岣岭岢岪岧岝岥岶岰岦帗帔帙弨弢弣弤彔徂彾彽忞忥怭怦怙怲怋"],
        ["cca1", "怴怊怗怳怚怞怬怢怍怐怮怓怑怌怉怜戔戽抭抴拑抾抪抶拊抮抳抯抻抩抰抸攽斨斻昉旼昄昒昈旻昃昋昍昅旽昑昐曶朊枅杬枎枒杶杻枘枆构杴枍枌杺枟枑枙枃杽极杸杹枔欥殀歾毞氝沓泬泫泮泙沶泔沭泧沷泐泂沺泃泆泭泲"],
        ["cd40", "泒泝沴沊沝沀泞泀洰泍泇沰泹泏泩泑炔炘炅炓炆炄炑炖炂炚炃牪狖狋狘狉狜狒狔狚狌狑玤玡玭玦玢玠玬玝瓝瓨甿畀甾疌疘皯盳盱盰盵矸矼矹矻矺"],
        ["cda1", "矷祂礿秅穸穻竻籵糽耵肏肮肣肸肵肭舠芠苀芫芚芘芛芵芧芮芼芞芺芴芨芡芩苂芤苃芶芢虰虯虭虮豖迒迋迓迍迖迕迗邲邴邯邳邰阹阽阼阺陃俍俅俓侲俉俋俁俔俜俙侻侳俛俇俖侺俀侹俬剄剉勀勂匽卼厗厖厙厘咺咡咭咥哏"],
        ["ce40", "哃茍咷咮哖咶哅哆咠呰咼咢咾呲哞咰垵垞垟垤垌垗垝垛垔垘垏垙垥垚垕壴复奓姡姞姮娀姱姝姺姽姼姶姤姲姷姛姩姳姵姠姾姴姭宨屌峐峘峌峗峋峛"],
        ["cea1", "峞峚峉峇峊峖峓峔峏峈峆峎峟峸巹帡帢帣帠帤庰庤庢庛庣庥弇弮彖徆怷怹恔恲恞恅恓恇恉恛恌恀恂恟怤恄恘恦恮扂扃拏挍挋拵挎挃拫拹挏挌拸拶挀挓挔拺挕拻拰敁敃斪斿昶昡昲昵昜昦昢昳昫昺昝昴昹昮朏朐柁柲柈枺"],
        ["cf40", "柜枻柸柘柀枷柅柫柤柟枵柍枳柷柶柮柣柂枹柎柧柰枲柼柆柭柌枮柦柛柺柉柊柃柪柋欨殂殄殶毖毘毠氠氡洨洴洭洟洼洿洒洊泚洳洄洙洺洚洑洀洝浂"],
        ["cfa1", "洁洘洷洃洏浀洇洠洬洈洢洉洐炷炟炾炱炰炡炴炵炩牁牉牊牬牰牳牮狊狤狨狫狟狪狦狣玅珌珂珈珅玹玶玵玴珫玿珇玾珃珆玸珋瓬瓮甮畇畈疧疪癹盄眈眃眄眅眊盷盻盺矧矨砆砑砒砅砐砏砎砉砃砓祊祌祋祅祄秕种秏秖秎窀"],
        ["d040", "穾竑笀笁籺籸籹籿粀粁紃紈紁罘羑羍羾耇耎耏耔耷胘胇胠胑胈胂胐胅胣胙胜胊胕胉胏胗胦胍臿舡芔苙苾苹茇苨茀苕茺苫苖苴苬苡苲苵茌苻苶苰苪"],
        ["d0a1", "苤苠苺苳苭虷虴虼虳衁衎衧衪衩觓訄訇赲迣迡迮迠郱邽邿郕郅邾郇郋郈釔釓陔陏陑陓陊陎倞倅倇倓倢倰倛俵俴倳倷倬俶俷倗倜倠倧倵倯倱倎党冔冓凊凄凅凈凎剡剚剒剞剟剕剢勍匎厞唦哢唗唒哧哳哤唚哿唄唈哫唑唅哱"],
        ["d140", "唊哻哷哸哠唎唃唋圁圂埌堲埕埒垺埆垽垼垸垶垿埇埐垹埁夎奊娙娖娭娮娕娏娗娊娞娳孬宧宭宬尃屖屔峬峿峮峱峷崀峹帩帨庨庮庪庬弳弰彧恝恚恧"],
        ["d1a1", "恁悢悈悀悒悁悝悃悕悛悗悇悜悎戙扆拲挐捖挬捄捅挶捃揤挹捋捊挼挩捁挴捘捔捙挭捇挳捚捑挸捗捀捈敊敆旆旃旄旂晊晟晇晑朒朓栟栚桉栲栳栻桋桏栖栱栜栵栫栭栯桎桄栴栝栒栔栦栨栮桍栺栥栠欬欯欭欱欴歭肂殈毦毤"],
        ["d240", "毨毣毢毧氥浺浣浤浶洍浡涒浘浢浭浯涑涍淯浿涆浞浧浠涗浰浼浟涂涘洯浨涋浾涀涄洖涃浻浽浵涐烜烓烑烝烋缹烢烗烒烞烠烔烍烅烆烇烚烎烡牂牸"],
        ["d2a1", "牷牶猀狺狴狾狶狳狻猁珓珙珥珖玼珧珣珩珜珒珛珔珝珚珗珘珨瓞瓟瓴瓵甡畛畟疰痁疻痄痀疿疶疺皊盉眝眛眐眓眒眣眑眕眙眚眢眧砣砬砢砵砯砨砮砫砡砩砳砪砱祔祛祏祜祓祒祑秫秬秠秮秭秪秜秞秝窆窉窅窋窌窊窇竘笐"],
        ["d340", "笄笓笅笏笈笊笎笉笒粄粑粊粌粈粍粅紞紝紑紎紘紖紓紟紒紏紌罜罡罞罠罝罛羖羒翃翂翀耖耾耹胺胲胹胵脁胻脀舁舯舥茳茭荄茙荑茥荖茿荁茦茜茢"],
        ["d3a1", "荂荎茛茪茈茼荍茖茤茠茷茯茩荇荅荌荓茞茬荋茧荈虓虒蚢蚨蚖蚍蚑蚞蚇蚗蚆蚋蚚蚅蚥蚙蚡蚧蚕蚘蚎蚝蚐蚔衃衄衭衵衶衲袀衱衿衯袃衾衴衼訒豇豗豻貤貣赶赸趵趷趶軑軓迾迵适迿迻逄迼迶郖郠郙郚郣郟郥郘郛郗郜郤酐"],
        ["d440", "酎酏釕釢釚陜陟隼飣髟鬯乿偰偪偡偞偠偓偋偝偲偈偍偁偛偊偢倕偅偟偩偫偣偤偆偀偮偳偗偑凐剫剭剬剮勖勓匭厜啵啶唼啍啐唴唪啑啢唶唵唰啒啅"],
        ["d4a1", "唌唲啥啎唹啈唭唻啀啋圊圇埻堔埢埶埜埴堀埭埽堈埸堋埳埏堇埮埣埲埥埬埡堎埼堐埧堁堌埱埩埰堍堄奜婠婘婕婧婞娸娵婭婐婟婥婬婓婤婗婃婝婒婄婛婈媎娾婍娹婌婰婩婇婑婖婂婜孲孮寁寀屙崞崋崝崚崠崌崨崍崦崥崏"],
        ["d540", "崰崒崣崟崮帾帴庱庴庹庲庳弶弸徛徖徟悊悐悆悾悰悺惓惔惏惤惙惝惈悱惛悷惊悿惃惍惀挲捥掊掂捽掽掞掭掝掗掫掎捯掇掐据掯捵掜捭掮捼掤挻掟"],
        ["d5a1", "捸掅掁掑掍捰敓旍晥晡晛晙晜晢朘桹梇梐梜桭桮梮梫楖桯梣梬梩桵桴梲梏桷梒桼桫桲梪梀桱桾梛梖梋梠梉梤桸桻梑梌梊桽欶欳欷欸殑殏殍殎殌氪淀涫涴涳湴涬淩淢涷淶淔渀淈淠淟淖涾淥淜淝淛淴淊涽淭淰涺淕淂淏淉"],
        ["d640", "淐淲淓淽淗淍淣涻烺焍烷焗烴焌烰焄烳焐烼烿焆焓焀烸烶焋焂焎牾牻牼牿猝猗猇猑猘猊猈狿猏猞玈珶珸珵琄琁珽琇琀珺珼珿琌琋珴琈畤畣痎痒痏"],
        ["d6a1", "痋痌痑痐皏皉盓眹眯眭眱眲眴眳眽眥眻眵硈硒硉硍硊硌砦硅硐祤祧祩祪祣祫祡离秺秸秶秷窏窔窐笵筇笴笥笰笢笤笳笘笪笝笱笫笭笯笲笸笚笣粔粘粖粣紵紽紸紶紺絅紬紩絁絇紾紿絊紻紨罣羕羜羝羛翊翋翍翐翑翇翏翉耟"],
        ["d740", "耞耛聇聃聈脘脥脙脛脭脟脬脞脡脕脧脝脢舑舸舳舺舴舲艴莐莣莨莍荺荳莤荴莏莁莕莙荵莔莩荽莃莌莝莛莪莋荾莥莯莈莗莰荿莦莇莮荶莚虙虖蚿蚷"],
        ["d7a1", "蛂蛁蛅蚺蚰蛈蚹蚳蚸蛌蚴蚻蚼蛃蚽蚾衒袉袕袨袢袪袚袑袡袟袘袧袙袛袗袤袬袌袓袎覂觖觙觕訰訧訬訞谹谻豜豝豽貥赽赻赹趼跂趹趿跁軘軞軝軜軗軠軡逤逋逑逜逌逡郯郪郰郴郲郳郔郫郬郩酖酘酚酓酕釬釴釱釳釸釤釹釪"],
        ["d840", "釫釷釨釮镺閆閈陼陭陫陱陯隿靪頄飥馗傛傕傔傞傋傣傃傌傎傝偨傜傒傂傇兟凔匒匑厤厧喑喨喥喭啷噅喢喓喈喏喵喁喣喒喤啽喌喦啿喕喡喎圌堩堷"],
        ["d8a1", "堙堞堧堣堨埵塈堥堜堛堳堿堶堮堹堸堭堬堻奡媯媔媟婺媢媞婸媦婼媥媬媕媮娷媄媊媗媃媋媩婻婽媌媜媏媓媝寪寍寋寔寑寊寎尌尰崷嵃嵫嵁嵋崿崵嵑嵎嵕崳崺嵒崽崱嵙嵂崹嵉崸崼崲崶嵀嵅幄幁彘徦徥徫惉悹惌惢惎惄愔"],
        ["d940", "惲愊愖愅惵愓惸惼惾惁愃愘愝愐惿愄愋扊掔掱掰揎揥揨揯揃撝揳揊揠揶揕揲揵摡揟掾揝揜揄揘揓揂揇揌揋揈揰揗揙攲敧敪敤敜敨敥斌斝斞斮旐旒"],
        ["d9a1", "晼晬晻暀晱晹晪晲朁椌棓椄棜椪棬棪棱椏棖棷棫棤棶椓椐棳棡椇棌椈楰梴椑棯棆椔棸棐棽棼棨椋椊椗棎棈棝棞棦棴棑椆棔棩椕椥棇欹欻欿欼殔殗殙殕殽毰毲毳氰淼湆湇渟湉溈渼渽湅湢渫渿湁湝湳渜渳湋湀湑渻渃渮湞"],
        ["da40", "湨湜湡渱渨湠湱湫渹渢渰湓湥渧湸湤湷湕湹湒湦渵渶湚焠焞焯烻焮焱焣焥焢焲焟焨焺焛牋牚犈犉犆犅犋猒猋猰猢猱猳猧猲猭猦猣猵猌琮琬琰琫琖"],
        ["daa1", "琚琡琭琱琤琣琝琩琠琲瓻甯畯畬痧痚痡痦痝痟痤痗皕皒盚睆睇睄睍睅睊睎睋睌矞矬硠硤硥硜硭硱硪确硰硩硨硞硢祴祳祲祰稂稊稃稌稄窙竦竤筊笻筄筈筌筎筀筘筅粢粞粨粡絘絯絣絓絖絧絪絏絭絜絫絒絔絩絑絟絎缾缿罥"],
        ["db40", "罦羢羠羡翗聑聏聐胾胔腃腊腒腏腇脽腍脺臦臮臷臸臹舄舼舽舿艵茻菏菹萣菀菨萒菧菤菼菶萐菆菈菫菣莿萁菝菥菘菿菡菋菎菖菵菉萉萏菞萑萆菂菳"],
        ["dba1", "菕菺菇菑菪萓菃菬菮菄菻菗菢萛菛菾蛘蛢蛦蛓蛣蛚蛪蛝蛫蛜蛬蛩蛗蛨蛑衈衖衕袺裗袹袸裀袾袶袼袷袽袲褁裉覕覘覗觝觚觛詎詍訹詙詀詗詘詄詅詒詈詑詊詌詏豟貁貀貺貾貰貹貵趄趀趉跘跓跍跇跖跜跏跕跙跈跗跅軯軷軺"],
        ["dc40", "軹軦軮軥軵軧軨軶軫軱軬軴軩逭逴逯鄆鄬鄄郿郼鄈郹郻鄁鄀鄇鄅鄃酡酤酟酢酠鈁鈊鈥鈃鈚鈦鈏鈌鈀鈒釿釽鈆鈄鈧鈂鈜鈤鈙鈗鈅鈖镻閍閌閐隇陾隈"],
        ["dca1", "隉隃隀雂雈雃雱雰靬靰靮頇颩飫鳦黹亃亄亶傽傿僆傮僄僊傴僈僂傰僁傺傱僋僉傶傸凗剺剸剻剼嗃嗛嗌嗐嗋嗊嗝嗀嗔嗄嗩喿嗒喍嗏嗕嗢嗖嗈嗲嗍嗙嗂圔塓塨塤塏塍塉塯塕塎塝塙塥塛堽塣塱壼嫇嫄嫋媺媸媱媵媰媿嫈媻嫆"],
        ["dd40", "媷嫀嫊媴媶嫍媹媐寖寘寙尟尳嵱嵣嵊嵥嵲嵬嵞嵨嵧嵢巰幏幎幊幍幋廅廌廆廋廇彀徯徭惷慉慊愫慅愶愲愮慆愯慏愩慀戠酨戣戥戤揅揱揫搐搒搉搠搤"],
        ["dda1", "搳摃搟搕搘搹搷搢搣搌搦搰搨摁搵搯搊搚摀搥搧搋揧搛搮搡搎敯斒旓暆暌暕暐暋暊暙暔晸朠楦楟椸楎楢楱椿楅楪椹楂楗楙楺楈楉椵楬椳椽楥棰楸椴楩楀楯楄楶楘楁楴楌椻楋椷楜楏楑椲楒椯楻椼歆歅歃歂歈歁殛嗀毻毼"],
        ["de40", "毹毷毸溛滖滈溏滀溟溓溔溠溱溹滆滒溽滁溞滉溷溰滍溦滏溲溾滃滜滘溙溒溎溍溤溡溿溳滐滊溗溮溣煇煔煒煣煠煁煝煢煲煸煪煡煂煘煃煋煰煟煐煓"],
        ["dea1", "煄煍煚牏犍犌犑犐犎猼獂猻猺獀獊獉瑄瑊瑋瑒瑑瑗瑀瑏瑐瑎瑂瑆瑍瑔瓡瓿瓾瓽甝畹畷榃痯瘏瘃痷痾痼痹痸瘐痻痶痭痵痽皙皵盝睕睟睠睒睖睚睩睧睔睙睭矠碇碚碔碏碄碕碅碆碡碃硹碙碀碖硻祼禂祽祹稑稘稙稒稗稕稢稓"],
        ["df40", "稛稐窣窢窞竫筦筤筭筴筩筲筥筳筱筰筡筸筶筣粲粴粯綈綆綀綍絿綅絺綎絻綃絼綌綔綄絽綒罭罫罧罨罬羦羥羧翛翜耡腤腠腷腜腩腛腢腲朡腞腶腧腯"],
        ["dfa1", "腄腡舝艉艄艀艂艅蓱萿葖葶葹蒏蒍葥葑葀蒆葧萰葍葽葚葙葴葳葝蔇葞萷萺萴葺葃葸萲葅萩菙葋萯葂萭葟葰萹葎葌葒葯蓅蒎萻葇萶萳葨葾葄萫葠葔葮葐蜋蜄蛷蜌蛺蛖蛵蝍蛸蜎蜉蜁蛶蜍蜅裖裋裍裎裞裛裚裌裐覅覛觟觥觤"],
        ["e040", "觡觠觢觜触詶誆詿詡訿詷誂誄詵誃誁詴詺谼豋豊豥豤豦貆貄貅賌赨赩趑趌趎趏趍趓趔趐趒跰跠跬跱跮跐跩跣跢跧跲跫跴輆軿輁輀輅輇輈輂輋遒逿"],
        ["e0a1", "遄遉逽鄐鄍鄏鄑鄖鄔鄋鄎酮酯鉈鉒鈰鈺鉦鈳鉥鉞銃鈮鉊鉆鉭鉬鉏鉠鉧鉯鈶鉡鉰鈱鉔鉣鉐鉲鉎鉓鉌鉖鈲閟閜閞閛隒隓隑隗雎雺雽雸雵靳靷靸靲頏頍頎颬飶飹馯馲馰馵骭骫魛鳪鳭鳧麀黽僦僔僗僨僳僛僪僝僤僓僬僰僯僣僠"],
        ["e140", "凘劀劁勩勫匰厬嘧嘕嘌嘒嗼嘏嘜嘁嘓嘂嗺嘝嘄嗿嗹墉塼墐墘墆墁塿塴墋塺墇墑墎塶墂墈塻墔墏壾奫嫜嫮嫥嫕嫪嫚嫭嫫嫳嫢嫠嫛嫬嫞嫝嫙嫨嫟孷寠"],
        ["e1a1", "寣屣嶂嶀嵽嶆嵺嶁嵷嶊嶉嶈嵾嵼嶍嵹嵿幘幙幓廘廑廗廎廜廕廙廒廔彄彃彯徶愬愨慁慞慱慳慒慓慲慬憀慴慔慺慛慥愻慪慡慖戩戧戫搫摍摛摝摴摶摲摳摽摵摦撦摎撂摞摜摋摓摠摐摿搿摬摫摙摥摷敳斠暡暠暟朅朄朢榱榶槉"],
        ["e240", "榠槎榖榰榬榼榑榙榎榧榍榩榾榯榿槄榽榤槔榹槊榚槏榳榓榪榡榞槙榗榐槂榵榥槆歊歍歋殞殟殠毃毄毾滎滵滱漃漥滸漷滻漮漉潎漙漚漧漘漻漒滭漊"],
        ["e2a1", "漶潳滹滮漭潀漰漼漵滫漇漎潃漅滽滶漹漜滼漺漟漍漞漈漡熇熐熉熀熅熂熏煻熆熁熗牄牓犗犕犓獃獍獑獌瑢瑳瑱瑵瑲瑧瑮甀甂甃畽疐瘖瘈瘌瘕瘑瘊瘔皸瞁睼瞅瞂睮瞀睯睾瞃碲碪碴碭碨硾碫碞碥碠碬碢碤禘禊禋禖禕禔禓"],
        ["e340", "禗禈禒禐稫穊稰稯稨稦窨窫窬竮箈箜箊箑箐箖箍箌箛箎箅箘劄箙箤箂粻粿粼粺綧綷緂綣綪緁緀緅綝緎緄緆緋緌綯綹綖綼綟綦綮綩綡緉罳翢翣翥翞"],
        ["e3a1", "耤聝聜膉膆膃膇膍膌膋舕蒗蒤蒡蒟蒺蓎蓂蒬蒮蒫蒹蒴蓁蓍蒪蒚蒱蓐蒝蒧蒻蒢蒔蓇蓌蒛蒩蒯蒨蓖蒘蒶蓏蒠蓗蓔蓒蓛蒰蒑虡蜳蜣蜨蝫蝀蜮蜞蜡蜙蜛蝃蜬蝁蜾蝆蜠蜲蜪蜭蜼蜒蜺蜱蜵蝂蜦蜧蜸蜤蜚蜰蜑裷裧裱裲裺裾裮裼裶裻"],
        ["e440", "裰裬裫覝覡覟覞觩觫觨誫誙誋誒誏誖谽豨豩賕賏賗趖踉踂跿踍跽踊踃踇踆踅跾踀踄輐輑輎輍鄣鄜鄠鄢鄟鄝鄚鄤鄡鄛酺酲酹酳銥銤鉶銛鉺銠銔銪銍"],
        ["e4a1", "銦銚銫鉹銗鉿銣鋮銎銂銕銢鉽銈銡銊銆銌銙銧鉾銇銩銝銋鈭隞隡雿靘靽靺靾鞃鞀鞂靻鞄鞁靿韎韍頖颭颮餂餀餇馝馜駃馹馻馺駂馽駇骱髣髧鬾鬿魠魡魟鳱鳲鳵麧僿儃儰僸儆儇僶僾儋儌僽儊劋劌勱勯噈噂噌嘵噁噊噉噆噘"],
        ["e540", "噚噀嘳嘽嘬嘾嘸嘪嘺圚墫墝墱墠墣墯墬墥墡壿嫿嫴嫽嫷嫶嬃嫸嬂嫹嬁嬇嬅嬏屧嶙嶗嶟嶒嶢嶓嶕嶠嶜嶡嶚嶞幩幝幠幜緳廛廞廡彉徲憋憃慹憱憰憢憉"],
        ["e5a1", "憛憓憯憭憟憒憪憡憍慦憳戭摮摰撖撠撅撗撜撏撋撊撌撣撟摨撱撘敶敺敹敻斲斳暵暰暩暲暷暪暯樀樆樗槥槸樕槱槤樠槿槬槢樛樝槾樧槲槮樔槷槧橀樈槦槻樍槼槫樉樄樘樥樏槶樦樇槴樖歑殥殣殢殦氁氀毿氂潁漦潾澇濆澒"],
        ["e640", "澍澉澌潢潏澅潚澖潶潬澂潕潲潒潐潗澔澓潝漀潡潫潽潧澐潓澋潩潿澕潣潷潪潻熲熯熛熰熠熚熩熵熝熥熞熤熡熪熜熧熳犘犚獘獒獞獟獠獝獛獡獚獙"],
        ["e6a1", "獢璇璉璊璆璁瑽璅璈瑼瑹甈甇畾瘥瘞瘙瘝瘜瘣瘚瘨瘛皜皝皞皛瞍瞏瞉瞈磍碻磏磌磑磎磔磈磃磄磉禚禡禠禜禢禛歶稹窲窴窳箷篋箾箬篎箯箹篊箵糅糈糌糋緷緛緪緧緗緡縃緺緦緶緱緰緮緟罶羬羰羭翭翫翪翬翦翨聤聧膣膟"],
        ["e740", "膞膕膢膙膗舖艏艓艒艐艎艑蔤蔻蔏蔀蔩蔎蔉蔍蔟蔊蔧蔜蓻蔫蓺蔈蔌蓴蔪蓲蔕蓷蓫蓳蓼蔒蓪蓩蔖蓾蔨蔝蔮蔂蓽蔞蓶蔱蔦蓧蓨蓰蓯蓹蔘蔠蔰蔋蔙蔯虢"],
        ["e7a1", "蝖蝣蝤蝷蟡蝳蝘蝔蝛蝒蝡蝚蝑蝞蝭蝪蝐蝎蝟蝝蝯蝬蝺蝮蝜蝥蝏蝻蝵蝢蝧蝩衚褅褌褔褋褗褘褙褆褖褑褎褉覢覤覣觭觰觬諏諆誸諓諑諔諕誻諗誾諀諅諘諃誺誽諙谾豍貏賥賟賙賨賚賝賧趠趜趡趛踠踣踥踤踮踕踛踖踑踙踦踧"],
        ["e840", "踔踒踘踓踜踗踚輬輤輘輚輠輣輖輗遳遰遯遧遫鄯鄫鄩鄪鄲鄦鄮醅醆醊醁醂醄醀鋐鋃鋄鋀鋙銶鋏鋱鋟鋘鋩鋗鋝鋌鋯鋂鋨鋊鋈鋎鋦鋍鋕鋉鋠鋞鋧鋑鋓"],
        ["e8a1", "銵鋡鋆銴镼閬閫閮閰隤隢雓霅霈霂靚鞊鞎鞈韐韏頞頝頦頩頨頠頛頧颲餈飺餑餔餖餗餕駜駍駏駓駔駎駉駖駘駋駗駌骳髬髫髳髲髱魆魃魧魴魱魦魶魵魰魨魤魬鳼鳺鳽鳿鳷鴇鴀鳹鳻鴈鴅鴄麃黓鼏鼐儜儓儗儚儑凞匴叡噰噠噮"],
        ["e940", "噳噦噣噭噲噞噷圜圛壈墽壉墿墺壂墼壆嬗嬙嬛嬡嬔嬓嬐嬖嬨嬚嬠嬞寯嶬嶱嶩嶧嶵嶰嶮嶪嶨嶲嶭嶯嶴幧幨幦幯廩廧廦廨廥彋徼憝憨憖懅憴懆懁懌憺"],
        ["e9a1", "憿憸憌擗擖擐擏擉撽撉擃擛擳擙攳敿敼斢曈暾曀曊曋曏暽暻暺曌朣樴橦橉橧樲橨樾橝橭橶橛橑樨橚樻樿橁橪橤橐橏橔橯橩橠樼橞橖橕橍橎橆歕歔歖殧殪殫毈毇氄氃氆澭濋澣濇澼濎濈潞濄澽澞濊澨瀄澥澮澺澬澪濏澿澸"],
        ["ea40", "澢濉澫濍澯澲澰燅燂熿熸燖燀燁燋燔燊燇燏熽燘熼燆燚燛犝犞獩獦獧獬獥獫獪瑿璚璠璔璒璕璡甋疀瘯瘭瘱瘽瘳瘼瘵瘲瘰皻盦瞚瞝瞡瞜瞛瞢瞣瞕瞙"],
        ["eaa1", "瞗磝磩磥磪磞磣磛磡磢磭磟磠禤穄穈穇窶窸窵窱窷篞篣篧篝篕篥篚篨篹篔篪篢篜篫篘篟糒糔糗糐糑縒縡縗縌縟縠縓縎縜縕縚縢縋縏縖縍縔縥縤罃罻罼罺羱翯耪耩聬膱膦膮膹膵膫膰膬膴膲膷膧臲艕艖艗蕖蕅蕫蕍蕓蕡蕘"],
        ["eb40", "蕀蕆蕤蕁蕢蕄蕑蕇蕣蔾蕛蕱蕎蕮蕵蕕蕧蕠薌蕦蕝蕔蕥蕬虣虥虤螛螏螗螓螒螈螁螖螘蝹螇螣螅螐螑螝螄螔螜螚螉褞褦褰褭褮褧褱褢褩褣褯褬褟觱諠"],
        ["eba1", "諢諲諴諵諝謔諤諟諰諈諞諡諨諿諯諻貑貒貐賵賮賱賰賳赬赮趥趧踳踾踸蹀蹅踶踼踽蹁踰踿躽輶輮輵輲輹輷輴遶遹遻邆郺鄳鄵鄶醓醐醑醍醏錧錞錈錟錆錏鍺錸錼錛錣錒錁鍆錭錎錍鋋錝鋺錥錓鋹鋷錴錂錤鋿錩錹錵錪錔錌"],
        ["ec40", "錋鋾錉錀鋻錖閼闍閾閹閺閶閿閵閽隩雔霋霒霐鞙鞗鞔韰韸頵頯頲餤餟餧餩馞駮駬駥駤駰駣駪駩駧骹骿骴骻髶髺髹髷鬳鮀鮅鮇魼魾魻鮂鮓鮒鮐魺鮕"],
        ["eca1", "魽鮈鴥鴗鴠鴞鴔鴩鴝鴘鴢鴐鴙鴟麈麆麇麮麭黕黖黺鼒鼽儦儥儢儤儠儩勴嚓嚌嚍嚆嚄嚃噾嚂噿嚁壖壔壏壒嬭嬥嬲嬣嬬嬧嬦嬯嬮孻寱寲嶷幬幪徾徻懃憵憼懧懠懥懤懨懞擯擩擣擫擤擨斁斀斶旚曒檍檖檁檥檉檟檛檡檞檇檓檎"],
        ["ed40", "檕檃檨檤檑橿檦檚檅檌檒歛殭氉濌澩濴濔濣濜濭濧濦濞濲濝濢濨燡燱燨燲燤燰燢獳獮獯璗璲璫璐璪璭璱璥璯甐甑甒甏疄癃癈癉癇皤盩瞵瞫瞲瞷瞶"],
        ["eda1", "瞴瞱瞨矰磳磽礂磻磼磲礅磹磾礄禫禨穜穛穖穘穔穚窾竀竁簅簏篲簀篿篻簎篴簋篳簂簉簃簁篸篽簆篰篱簐簊糨縭縼繂縳顈縸縪繉繀繇縩繌縰縻縶繄縺罅罿罾罽翴翲耬膻臄臌臊臅臇膼臩艛艚艜薃薀薏薧薕薠薋薣蕻薤薚薞"],
        ["ee40", "蕷蕼薉薡蕺蕸蕗薎薖薆薍薙薝薁薢薂薈薅蕹蕶薘薐薟虨螾螪螭蟅螰螬螹螵螼螮蟉蟃蟂蟌螷螯蟄蟊螴螶螿螸螽蟞螲褵褳褼褾襁襒褷襂覭覯覮觲觳謞"],
        ["eea1", "謘謖謑謅謋謢謏謒謕謇謍謈謆謜謓謚豏豰豲豱豯貕貔賹赯蹎蹍蹓蹐蹌蹇轃轀邅遾鄸醚醢醛醙醟醡醝醠鎡鎃鎯鍤鍖鍇鍼鍘鍜鍶鍉鍐鍑鍠鍭鎏鍌鍪鍹鍗鍕鍒鍏鍱鍷鍻鍡鍞鍣鍧鎀鍎鍙闇闀闉闃闅閷隮隰隬霠霟霘霝霙鞚鞡鞜"],
        ["ef40", "鞞鞝韕韔韱顁顄顊顉顅顃餥餫餬餪餳餲餯餭餱餰馘馣馡騂駺駴駷駹駸駶駻駽駾駼騃骾髾髽鬁髼魈鮚鮨鮞鮛鮦鮡鮥鮤鮆鮢鮠鮯鴳鵁鵧鴶鴮鴯鴱鴸鴰"],
        ["efa1", "鵅鵂鵃鴾鴷鵀鴽翵鴭麊麉麍麰黈黚黻黿鼤鼣鼢齔龠儱儭儮嚘嚜嚗嚚嚝嚙奰嬼屩屪巀幭幮懘懟懭懮懱懪懰懫懖懩擿攄擽擸攁攃擼斔旛曚曛曘櫅檹檽櫡櫆檺檶檷櫇檴檭歞毉氋瀇瀌瀍瀁瀅瀔瀎濿瀀濻瀦濼濷瀊爁燿燹爃燽獶"],
        ["f040", "璸瓀璵瓁璾璶璻瓂甔甓癜癤癙癐癓癗癚皦皽盬矂瞺磿礌礓礔礉礐礒礑禭禬穟簜簩簙簠簟簭簝簦簨簢簥簰繜繐繖繣繘繢繟繑繠繗繓羵羳翷翸聵臑臒"],
        ["f0a1", "臐艟艞薴藆藀藃藂薳薵薽藇藄薿藋藎藈藅薱薶藒蘤薸薷薾虩蟧蟦蟢蟛蟫蟪蟥蟟蟳蟤蟔蟜蟓蟭蟘蟣螤蟗蟙蠁蟴蟨蟝襓襋襏襌襆襐襑襉謪謧謣謳謰謵譇謯謼謾謱謥謷謦謶謮謤謻謽謺豂豵貙貘貗賾贄贂贀蹜蹢蹠蹗蹖蹞蹥蹧"],
        ["f140", "蹛蹚蹡蹝蹩蹔轆轇轈轋鄨鄺鄻鄾醨醥醧醯醪鎵鎌鎒鎷鎛鎝鎉鎧鎎鎪鎞鎦鎕鎈鎙鎟鎍鎱鎑鎲鎤鎨鎴鎣鎥闒闓闑隳雗雚巂雟雘雝霣霢霥鞬鞮鞨鞫鞤鞪"],
        ["f1a1", "鞢鞥韗韙韖韘韺顐顑顒颸饁餼餺騏騋騉騍騄騑騊騅騇騆髀髜鬈鬄鬅鬩鬵魊魌魋鯇鯆鯃鮿鯁鮵鮸鯓鮶鯄鮹鮽鵜鵓鵏鵊鵛鵋鵙鵖鵌鵗鵒鵔鵟鵘鵚麎麌黟鼁鼀鼖鼥鼫鼪鼩鼨齌齕儴儵劖勷厴嚫嚭嚦嚧嚪嚬壚壝壛夒嬽嬾嬿巃幰"],
        ["f240", "徿懻攇攐攍攉攌攎斄旞旝曞櫧櫠櫌櫑櫙櫋櫟櫜櫐櫫櫏櫍櫞歠殰氌瀙瀧瀠瀖瀫瀡瀢瀣瀩瀗瀤瀜瀪爌爊爇爂爅犥犦犤犣犡瓋瓅璷瓃甖癠矉矊矄矱礝礛"],
        ["f2a1", "礡礜礗礞禰穧穨簳簼簹簬簻糬糪繶繵繸繰繷繯繺繲繴繨罋罊羃羆羷翽翾聸臗臕艤艡艣藫藱藭藙藡藨藚藗藬藲藸藘藟藣藜藑藰藦藯藞藢蠀蟺蠃蟶蟷蠉蠌蠋蠆蟼蠈蟿蠊蠂襢襚襛襗襡襜襘襝襙覈覷覶觶譐譈譊譀譓譖譔譋譕"],
        ["f340", "譑譂譒譗豃豷豶貚贆贇贉趬趪趭趫蹭蹸蹳蹪蹯蹻軂轒轑轏轐轓辴酀鄿醰醭鏞鏇鏏鏂鏚鏐鏹鏬鏌鏙鎩鏦鏊鏔鏮鏣鏕鏄鏎鏀鏒鏧镽闚闛雡霩霫霬霨霦"],
        ["f3a1", "鞳鞷鞶韝韞韟顜顙顝顗颿颽颻颾饈饇饃馦馧騚騕騥騝騤騛騢騠騧騣騞騜騔髂鬋鬊鬎鬌鬷鯪鯫鯠鯞鯤鯦鯢鯰鯔鯗鯬鯜鯙鯥鯕鯡鯚鵷鶁鶊鶄鶈鵱鶀鵸鶆鶋鶌鵽鵫鵴鵵鵰鵩鶅鵳鵻鶂鵯鵹鵿鶇鵨麔麑黀黼鼭齀齁齍齖齗齘匷嚲"],
        ["f440", "嚵嚳壣孅巆巇廮廯忀忁懹攗攖攕攓旟曨曣曤櫳櫰櫪櫨櫹櫱櫮櫯瀼瀵瀯瀷瀴瀱灂瀸瀿瀺瀹灀瀻瀳灁爓爔犨獽獼璺皫皪皾盭矌矎矏矍矲礥礣礧礨礤礩"],
        ["f4a1", "禲穮穬穭竷籉籈籊籇籅糮繻繾纁纀羺翿聹臛臙舋艨艩蘢藿蘁藾蘛蘀藶蘄蘉蘅蘌藽蠙蠐蠑蠗蠓蠖襣襦覹觷譠譪譝譨譣譥譧譭趮躆躈躄轙轖轗轕轘轚邍酃酁醷醵醲醳鐋鐓鏻鐠鐏鐔鏾鐕鐐鐨鐙鐍鏵鐀鏷鐇鐎鐖鐒鏺鐉鏸鐊鏿"],
        ["f540", "鏼鐌鏶鐑鐆闞闠闟霮霯鞹鞻韽韾顠顢顣顟飁飂饐饎饙饌饋饓騲騴騱騬騪騶騩騮騸騭髇髊髆鬐鬒鬑鰋鰈鯷鰅鰒鯸鱀鰇鰎鰆鰗鰔鰉鶟鶙鶤鶝鶒鶘鶐鶛"],
        ["f5a1", "鶠鶔鶜鶪鶗鶡鶚鶢鶨鶞鶣鶿鶩鶖鶦鶧麙麛麚黥黤黧黦鼰鼮齛齠齞齝齙龑儺儹劘劗囃嚽嚾孈孇巋巏廱懽攛欂櫼欃櫸欀灃灄灊灈灉灅灆爝爚爙獾甗癪矐礭礱礯籔籓糲纊纇纈纋纆纍罍羻耰臝蘘蘪蘦蘟蘣蘜蘙蘧蘮蘡蘠蘩蘞蘥"],
        ["f640", "蠩蠝蠛蠠蠤蠜蠫衊襭襩襮襫觺譹譸譅譺譻贐贔趯躎躌轞轛轝酆酄酅醹鐿鐻鐶鐩鐽鐼鐰鐹鐪鐷鐬鑀鐱闥闤闣霵霺鞿韡顤飉飆飀饘饖騹騽驆驄驂驁騺"],
        ["f6a1", "騿髍鬕鬗鬘鬖鬺魒鰫鰝鰜鰬鰣鰨鰩鰤鰡鶷鶶鶼鷁鷇鷊鷏鶾鷅鷃鶻鶵鷎鶹鶺鶬鷈鶱鶭鷌鶳鷍鶲鹺麜黫黮黭鼛鼘鼚鼱齎齥齤龒亹囆囅囋奱孋孌巕巑廲攡攠攦攢欋欈欉氍灕灖灗灒爞爟犩獿瓘瓕瓙瓗癭皭礵禴穰穱籗籜籙籛籚"],
        ["f740", "糴糱纑罏羇臞艫蘴蘵蘳蘬蘲蘶蠬蠨蠦蠪蠥襱覿覾觻譾讄讂讆讅譿贕躕躔躚躒躐躖躗轠轢酇鑌鑐鑊鑋鑏鑇鑅鑈鑉鑆霿韣顪顩飋饔饛驎驓驔驌驏驈驊"],
        ["f7a1", "驉驒驐髐鬙鬫鬻魖魕鱆鱈鰿鱄鰹鰳鱁鰼鰷鰴鰲鰽鰶鷛鷒鷞鷚鷋鷐鷜鷑鷟鷩鷙鷘鷖鷵鷕鷝麶黰鼵鼳鼲齂齫龕龢儽劙壨壧奲孍巘蠯彏戁戃戄攩攥斖曫欑欒欏毊灛灚爢玂玁玃癰矔籧籦纕艬蘺虀蘹蘼蘱蘻蘾蠰蠲蠮蠳襶襴襳觾"],
        ["f840", "讌讎讋讈豅贙躘轤轣醼鑢鑕鑝鑗鑞韄韅頀驖驙鬞鬟鬠鱒鱘鱐鱊鱍鱋鱕鱙鱌鱎鷻鷷鷯鷣鷫鷸鷤鷶鷡鷮鷦鷲鷰鷢鷬鷴鷳鷨鷭黂黐黲黳鼆鼜鼸鼷鼶齃齏"],
        ["f8a1", "齱齰齮齯囓囍孎屭攭曭曮欓灟灡灝灠爣瓛瓥矕礸禷禶籪纗羉艭虃蠸蠷蠵衋讔讕躞躟躠躝醾醽釂鑫鑨鑩雥靆靃靇韇韥驞髕魙鱣鱧鱦鱢鱞鱠鸂鷾鸇鸃鸆鸅鸀鸁鸉鷿鷽鸄麠鼞齆齴齵齶囔攮斸欘欙欗欚灢爦犪矘矙礹籩籫糶纚"],
        ["f940", "纘纛纙臠臡虆虇虈襹襺襼襻觿讘讙躥躤躣鑮鑭鑯鑱鑳靉顲饟鱨鱮鱭鸋鸍鸐鸏鸒鸑麡黵鼉齇齸齻齺齹圞灦籯蠼趲躦釃鑴鑸鑶鑵驠鱴鱳鱱鱵鸔鸓黶鼊"],
        ["f9a1", "龤灨灥糷虪蠾蠽蠿讞貜躩軉靋顳顴飌饡馫驤驦驧鬤鸕鸗齈戇欞爧虌躨钂钀钁驩驨鬮鸙爩虋讟钃鱹麷癵驫鱺鸝灩灪麤齾齉龘碁銹裏墻恒粧嫺╔╦╗╠╬╣╚╩╝╒╤╕╞╪╡╘╧╛╓╥╖╟╫╢╙╨╜║═╭╮╰╯▓"]
    ];
    module.exports = json;
},
/* iconv-lite/encodings/tables/big5-added.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = [
        ["8740", "䏰䰲䘃䖦䕸𧉧䵷䖳𧲱䳢𧳅㮕䜶䝄䱇䱀𤊿𣘗𧍒𦺋𧃒䱗𪍑䝏䗚䲅𧱬䴇䪤䚡𦬣爥𥩔𡩣𣸆𣽡晍囻"],
        ["8767", "綕夝𨮹㷴霴𧯯寛𡵞媤㘥𩺰嫑宷峼杮薓𩥅瑡璝㡵𡵓𣚞𦀡㻬"],
        ["87a1", "𥣞㫵竼龗𤅡𨤍𣇪𠪊𣉞䌊蒄龖鐯䤰蘓墖靊鈘秐稲晠権袝瑌篅枂稬剏遆㓦珄𥶹瓆鿇垳䤯呌䄱𣚎堘穲𧭥讏䚮𦺈䆁𥶙箮𢒼鿈𢓁𢓉𢓌鿉蔄𣖻䂴鿊䓡𪷿拁灮鿋"],
        ["8840", "㇀", 4, "𠄌㇅𠃑𠃍㇆㇇𠃋𡿨㇈𠃊㇉㇊㇋㇌𠄎㇍㇎ĀÁǍÀĒÉĚÈŌÓǑÒ࿿Ê̄Ế࿿Ê̌ỀÊāáǎàɑēéěèīíǐìōóǒòūúǔùǖǘǚ"],
        ["88a1", "ǜü࿿ê̄ế࿿ê̌ềêɡ⏚⏛"],
        ["8940", "𪎩𡅅"],
        ["8943", "攊"],
        ["8946", "丽滝鵎釟"],
        ["894c", "𧜵撑会伨侨兖兴农凤务动医华发变团声处备夲头学实実岚庆总斉柾栄桥济炼电纤纬纺织经统缆缷艺苏药视设询车轧轮"],
        ["89a1", "琑糼緍楆竉刧"],
        ["89ab", "醌碸酞肼"],
        ["89b0", "贋胶𠧧"],
        ["89b5", "肟黇䳍鷉鸌䰾𩷶𧀎鸊𪄳㗁"],
        ["89c1", "溚舾甙"],
        ["89c5", "䤑马骏龙禇𨑬𡷊𠗐𢫦两亁亀亇亿仫伷㑌侽㹈倃傈㑽㒓㒥円夅凛凼刅争剹劐匧㗇厩㕑厰㕓参吣㕭㕲㚁咓咣咴咹哐哯唘唣唨㖘唿㖥㖿嗗㗅"],
        ["8a40", "𧶄唥"],
        ["8a43", "𠱂𠴕𥄫喐𢳆㧬𠍁蹆𤶸𩓥䁓𨂾睺𢰸㨴䟕𨅝𦧲𤷪擝𠵼𠾴𠳕𡃴撍蹾𠺖𠰋𠽤𢲩𨉖𤓓"],
        ["8a64", "𠵆𩩍𨃩䟴𤺧𢳂骲㩧𩗴㿭㔆𥋇𩟔𧣈𢵄鵮頕"],
        ["8a76", "䏙𦂥撴哣𢵌𢯊𡁷㧻𡁯"],
        ["8aa1", "𦛚𦜖𧦠擪𥁒𠱃蹨𢆡𨭌𠜱"],
        ["8aac", "䠋𠆩㿺塳𢶍"],
        ["8ab2", "𤗈𠓼𦂗𠽌𠶖啹䂻䎺"],
        ["8abb", "䪴𢩦𡂝膪飵𠶜捹㧾𢝵跀嚡摼㹃"],
        ["8ac9", "𪘁𠸉𢫏𢳉"],
        ["8ace", "𡃈𣧂㦒㨆𨊛㕸𥹉𢃇噒𠼱𢲲𩜠㒼氽𤸻"],
        ["8adf", "𧕴𢺋𢈈𪙛𨳍𠹺𠰴𦠜羓𡃏𢠃𢤹㗻𥇣𠺌𠾍𠺪㾓𠼰𠵇𡅏𠹌"],
        ["8af6", "𠺫𠮩𠵈𡃀𡄽㿹𢚖搲𠾭"],
        ["8b40", "𣏴𧘹𢯎𠵾𠵿𢱑𢱕㨘𠺘𡃇𠼮𪘲𦭐𨳒𨶙𨳊閪哌苄喹"],
        ["8b55", "𩻃鰦骶𧝞𢷮煀腭胬尜𦕲脴㞗卟𨂽醶𠻺𠸏𠹷𠻻㗝𤷫㘉𠳖嚯𢞵𡃉𠸐𠹸𡁸𡅈𨈇𡑕𠹹𤹐𢶤婔𡀝𡀞𡃵𡃶垜𠸑"],
        ["8ba1", "𧚔𨋍𠾵𠹻𥅾㜃𠾶𡆀𥋘𪊽𤧚𡠺𤅷𨉼墙剨㘚𥜽箲孨䠀䬬鼧䧧鰟鮍𥭴𣄽嗻㗲嚉丨夂𡯁屮靑𠂆乛亻㔾尣彑忄㣺扌攵歺氵氺灬爫丬犭𤣩罒礻糹罓𦉪㓁"],
        ["8bde", "𦍋耂肀𦘒𦥑卝衤见𧢲讠贝钅镸长门𨸏韦页风飞饣𩠐鱼鸟黄歯龜丷𠂇阝户钢"],
        ["8c40", "倻淾𩱳龦㷉袏𤅎灷峵䬠𥇍㕙𥴰愢𨨲辧釶熑朙玺𣊁𪄇㲋𡦀䬐磤琂冮𨜏䀉橣𪊺䈣蘏𠩯稪𩥇𨫪靕灍匤𢁾鏴盙𨧣龧矝亣俰傼丯众龨吴綋墒壐𡶶庒庙忂𢜒斋"],
        ["8ca1", "𣏹椙橃𣱣泿"],
        ["8ca7", "爀𤔅玌㻛𤨓嬕璹讃𥲤𥚕窓篬糃繬苸薗龩袐龪躹龫迏蕟駠鈡龬𨶹𡐿䁱䊢娚"],
        ["8cc9", "顨杫䉶圽"],
        ["8cce", "藖𤥻芿𧄍䲁𦵴嵻𦬕𦾾龭龮宖龯曧繛湗秊㶈䓃𣉖𢞖䎚䔶"],
        ["8ce6", "峕𣬚諹屸㴒𣕑嵸龲煗䕘𤃬𡸣䱷㥸㑊𠆤𦱁諌侴𠈹妿腬顖𩣺弻"],
        ["8d40", "𠮟"],
        ["8d42", "𢇁𨥭䄂䚻𩁹㼇龳𪆵䃸㟖䛷𦱆䅼𨚲𧏿䕭㣔𥒚䕡䔛䶉䱻䵶䗪㿈𤬏㙡䓞䒽䇭崾嵈嵖㷼㠏嶤嶹㠠㠸幂庽弥徃㤈㤔㤿㥍惗愽峥㦉憷憹懏㦸戬抐拥挘㧸嚱"],
        ["8da1", "㨃揢揻搇摚㩋擀崕嘡龟㪗斆㪽旿晓㫲暒㬢朖㭂枤栀㭘桊梄㭲㭱㭻椉楃牜楤榟榅㮼槖㯝橥橴橱檂㯬檙㯲檫檵櫔櫶殁毁毪汵沪㳋洂洆洦涁㳯涤涱渕渘温溆𨧀溻滢滚齿滨滩漤漴㵆𣽁澁澾㵪㵵熷岙㶊瀬㶑灐灔灯灿炉𠌥䏁㗱𠻘"],
        ["8e40", "𣻗垾𦻓焾𥟠㙎榢𨯩孴穉𥣡𩓙穥穽𥦬窻窰竂竃燑𦒍䇊竚竝竪䇯咲𥰁笋筕笩𥌎𥳾箢筯莜𥮴𦱿篐萡箒箸𥴠㶭𥱥蒒篺簆簵𥳁籄粃𤢂粦晽𤕸糉糇糦籴糳糵糎"],
        ["8ea1", "繧䔝𦹄絝𦻖璍綉綫焵綳緒𤁗𦀩緤㴓緵𡟹緥𨍭縝𦄡𦅚繮纒䌫鑬縧罀罁罇礶𦋐駡羗𦍑羣𡙡𠁨䕜𣝦䔃𨌺翺𦒉者耈耝耨耯𪂇𦳃耻耼聡𢜔䦉𦘦𣷣𦛨朥肧𨩈脇脚墰𢛶汿𦒘𤾸擧𡒊舘𡡞橓𤩥𤪕䑺舩𠬍𦩒𣵾俹𡓽蓢荢𦬊𤦧𣔰𡝳𣷸芪椛芳䇛"],
        ["8f40", "蕋苐茚𠸖𡞴㛁𣅽𣕚艻苢茘𣺋𦶣𦬅𦮗𣗎㶿茝嗬莅䔋𦶥莬菁菓㑾𦻔橗蕚㒖𦹂𢻯葘𥯤葱㷓䓤檧葊𣲵祘蒨𦮖𦹷𦹃蓞萏莑䒠蒓蓤𥲑䉀𥳀䕃蔴嫲𦺙䔧蕳䔖枿蘖"],
        ["8fa1", "𨘥𨘻藁𧂈蘂𡖂𧃍䕫䕪蘨㙈𡢢号𧎚虾蝱𪃸蟮𢰧螱蟚蠏噡虬桖䘏衅衆𧗠𣶹𧗤衞袜䙛袴袵揁装睷𧜏覇覊覦覩覧覼𨨥觧𧤤𧪽誜瞓釾誐𧩙竩𧬺𣾏䜓𧬸煼謌謟𥐰𥕥謿譌譍誩𤩺讐讛誯𡛟䘕衏貛𧵔𧶏貫㜥𧵓賖𧶘𧶽贒贃𡤐賛灜贑𤳉㻐起"],
        ["9040", "趩𨀂𡀔𤦊㭼𨆼𧄌竧躭躶軃鋔輙輭𨍥𨐒辥錃𪊟𠩐辳䤪𨧞𨔽𣶻廸𣉢迹𪀔𨚼𨔁𢌥㦀𦻗逷𨔼𧪾遡𨕬𨘋邨𨜓郄𨛦邮都酧㫰醩釄粬𨤳𡺉鈎沟鉁鉢𥖹銹𨫆𣲛𨬌𥗛"],
        ["90a1", "𠴱錬鍫𨫡𨯫炏嫃𨫢𨫥䥥鉄𨯬𨰹𨯿鍳鑛躼閅閦鐦閠濶䊹𢙺𨛘𡉼𣸮䧟氜陻隖䅬隣𦻕懚隶磵𨫠隽双䦡𦲸𠉴𦐐𩂯𩃥𤫑𡤕𣌊霱虂霶䨏䔽䖅𤫩灵孁霛靜𩇕靗孊𩇫靟鐥僐𣂷𣂼鞉鞟鞱鞾韀韒韠𥑬韮琜𩐳響韵𩐝𧥺䫑頴頳顋顦㬎𧅵㵑𠘰𤅜"],
        ["9140", "𥜆飊颷飈飇䫿𦴧𡛓喰飡飦飬鍸餹𤨩䭲𩡗𩤅駵騌騻騐驘𥜥㛄𩂱𩯕髠髢𩬅髴䰎鬔鬭𨘀倴鬴𦦨㣃𣁽魐魀𩴾婅𡡣鮎𤉋鰂鯿鰌𩹨鷔𩾷𪆒𪆫𪃡𪄣𪇟鵾鶃𪄴鸎梈"],
        ["91a1", "鷄𢅛𪆓𪈠𡤻𪈳鴹𪂹𪊴麐麕麞麢䴴麪麯𤍤黁㭠㧥㴝伲㞾𨰫鼂鼈䮖鐤𦶢鼗鼖鼹嚟嚊齅馸𩂋韲葿齢齩竜龎爖䮾𤥵𤦻煷𤧸𤍈𤩑玞𨯚𡣺禟𨥾𨸶鍩鏳𨩄鋬鎁鏋𨥬𤒹爗㻫睲穃烐𤑳𤏸煾𡟯炣𡢾𣖙㻇𡢅𥐯𡟸㜢𡛻𡠹㛡𡝴𡣑𥽋㜣𡛀坛𤨥𡏾𡊨"],
        ["9240", "𡏆𡒶蔃𣚦蔃葕𤦔𧅥𣸱𥕜𣻻𧁒䓴𣛮𩦝𦼦柹㜳㰕㷧塬𡤢栐䁗𣜿𤃡𤂋𤄏𦰡哋嚞𦚱嚒𠿟𠮨𠸍鏆𨬓鎜仸儫㠙𤐶亼𠑥𠍿佋侊𥙑婨𠆫𠏋㦙𠌊𠐔㐵伩𠋀𨺳𠉵諚𠈌亘"],
        ["92a1", "働儍侢伃𤨎𣺊佂倮偬傁俌俥偘僼兙兛兝兞湶𣖕𣸹𣺿浲𡢄𣺉冨凃𠗠䓝𠒣𠒒𠒑赺𨪜𠜎剙劤𠡳勡鍮䙺熌𤎌𠰠𤦬𡃤槑𠸝瑹㻞璙琔瑖玘䮎𤪼𤂍叐㖄爏𤃉喴𠍅响𠯆圝鉝雴鍦埝垍坿㘾壋媙𨩆𡛺𡝯𡜐娬妸銏婾嫏娒𥥆𡧳𡡡𤊕㛵洅瑃娡𥺃"],
        ["9340", "媁𨯗𠐓鏠璌𡌃焅䥲鐈𨧻鎽㞠尞岞幞幈𡦖𡥼𣫮廍孏𡤃𡤄㜁𡢠㛝𡛾㛓脪𨩇𡶺𣑲𨦨弌弎𡤧𡞫婫𡜻孄蘔𧗽衠恾𢡠𢘫忛㺸𢖯𢖾𩂈𦽳懀𠀾𠁆𢘛憙憘恵𢲛𢴇𤛔𩅍"],
        ["93a1", "摱𤙥𢭪㨩𢬢𣑐𩣪𢹸挷𪑛撶挱揑𤧣𢵧护𢲡搻敫楲㯴𣂎𣊭𤦉𣊫唍𣋠𡣙𩐿曎𣊉𣆳㫠䆐𥖄𨬢𥖏𡛼𥕛𥐥磮𣄃𡠪𣈴㑤𣈏𣆂𤋉暎𦴤晫䮓昰𧡰𡷫晣𣋒𣋡昞𥡲㣑𣠺𣞼㮙𣞢𣏾瓐㮖枏𤘪梶栞㯄檾㡣𣟕𤒇樳橒櫉欅𡤒攑梘橌㯗橺歗𣿀𣲚鎠鋲𨯪𨫋"],
        ["9440", "銉𨀞𨧜鑧涥漋𤧬浧𣽿㶏渄𤀼娽渊塇洤硂焻𤌚𤉶烱牐犇犔𤞏𤜥兹𤪤𠗫瑺𣻸𣙟𤩊𤤗𥿡㼆㺱𤫟𨰣𣼵悧㻳瓌琼鎇琷䒟𦷪䕑疃㽣𤳙𤴆㽘畕癳𪗆㬙瑨𨫌𤦫𤦎㫻"],
        ["94a1", "㷍𤩎㻿𤧅𤣳釺圲鍂𨫣𡡤僟𥈡𥇧睸𣈲眎眏睻𤚗𣞁㩞𤣰琸璛㺿𤪺𤫇䃈𤪖𦆮錇𥖁砞碍碈磒珐祙𧝁𥛣䄎禛蒖禥樭𣻺稺秴䅮𡛦䄲鈵秱𠵌𤦌𠊙𣶺𡝮㖗啫㕰㚪𠇔𠰍竢婙𢛵𥪯𥪜娍𠉛磰娪𥯆竾䇹籝籭䈑𥮳𥺼𥺦糍𤧹𡞰粎籼粮檲緜縇緓罎𦉡"],
        ["9540", "𦅜𧭈綗𥺂䉪𦭵𠤖柖𠁎𣗏埄𦐒𦏸𤥢翝笧𠠬𥫩𥵃笌𥸎駦虅驣樜𣐿㧢𤧷𦖭騟𦖠蒀𧄧𦳑䓪脷䐂胆脉腂𦞴飃𦩂艢艥𦩑葓𦶧蘐𧈛媆䅿𡡀嬫𡢡嫤𡣘蚠蜨𣶏蠭𧐢娂"],
        ["95a1", "衮佅袇袿裦襥襍𥚃襔𧞅𧞄𨯵𨯙𨮜𨧹㺭蒣䛵䛏㟲訽訜𩑈彍鈫𤊄旔焩烄𡡅鵭貟賩𧷜妚矃姰䍮㛔踪躧𤰉輰轊䋴汘澻𢌡䢛潹溋𡟚鯩㚵𤤯邻邗啱䤆醻鐄𨩋䁢𨫼鐧𨰝𨰻蓥訫閙閧閗閖𨴴瑅㻂𤣿𤩂𤏪㻧𣈥随𨻧𨹦𨹥㻌𤧭𤩸𣿮琒瑫㻼靁𩂰"],
        ["9640", "桇䨝𩂓𥟟靝鍨𨦉𨰦𨬯𦎾銺嬑譩䤼珹𤈛鞛靱餸𠼦巁𨯅𤪲頟𩓚鋶𩗗釥䓀𨭐𤩧𨭤飜𨩅㼀鈪䤥萔餻饍𧬆㷽馛䭯馪驜𨭥𥣈檏騡嫾騯𩣱䮐𩥈馼䮽䮗鍽塲𡌂堢𤦸"],
        ["96a1", "𡓨硄𢜟𣶸棅㵽鑘㤧慐𢞁𢥫愇鱏鱓鱻鰵鰐魿鯏𩸭鮟𪇵𪃾鴡䲮𤄄鸘䲰鴌𪆴𪃭𪃳𩤯鶥蒽𦸒𦿟𦮂藼䔳𦶤𦺄𦷰萠藮𦸀𣟗𦁤秢𣖜𣙀䤭𤧞㵢鏛銾鍈𠊿碹鉷鑍俤㑀遤𥕝砽硔碶硋𡝗𣇉𤥁㚚佲濚濙瀞瀞吔𤆵垻壳垊鴖埗焴㒯𤆬燫𦱀𤾗嬨𡞵𨩉"],
        ["9740", "愌嫎娋䊼𤒈㜬䭻𨧼鎻鎸𡣖𠼝葲𦳀𡐓𤋺𢰦𤏁妔𣶷𦝁綨𦅛𦂤𤦹𤦋𨧺鋥珢㻩璴𨭣𡢟㻡𤪳櫘珳珻㻖𤨾𤪔𡟙𤩦𠎧𡐤𤧥瑈𤤖炥𤥶銄珦鍟𠓾錱𨫎𨨖鎆𨯧𥗕䤵𨪂煫"],
        ["97a1", "𤥃𠳿嚤𠘚𠯫𠲸唂秄𡟺緾𡛂𤩐𡡒䔮鐁㜊𨫀𤦭妰𡢿𡢃𧒄媡㛢𣵛㚰鉟婹𨪁𡡢鍴㳍𠪴䪖㦊僴㵩㵌𡎜煵䋻𨈘渏𩃤䓫浗𧹏灧沯㳖𣿭𣸭渂漌㵯𠏵畑㚼㓈䚀㻚䡱姄鉮䤾轁𨰜𦯀堒埈㛖𡑒烾𤍢𤩱𢿣𡊰𢎽梹楧𡎘𣓥𧯴𣛟𨪃𣟖𣏺𤲟樚𣚭𦲷萾䓟䓎"],
        ["9840", "𦴦𦵑𦲂𦿞漗𧄉茽𡜺菭𦲀𧁓𡟛妉媂𡞳婡婱𡤅𤇼㜭姯𡜼㛇熎鎐暚𤊥婮娫𤊓樫𣻹𧜶𤑛𤋊焝𤉙𨧡侰𦴨峂𤓎𧹍𤎽樌𤉖𡌄炦焳𤏩㶥泟勇𤩏繥姫崯㷳彜𤩝𡟟綤萦"],
        ["98a1", "咅𣫺𣌀𠈔坾𠣕𠘙㿥𡾞𪊶瀃𩅛嵰玏糓𨩙𩐠俈翧狍猐𧫴猸猹𥛶獁獈㺩𧬘遬燵𤣲珡臶㻊県㻑沢国琙琞琟㻢㻰㻴㻺瓓㼎㽓畂畭畲疍㽼痈痜㿀癍㿗癴㿜発𤽜熈嘣覀塩䀝睃䀹条䁅㗛瞘䁪䁯属瞾矋売砘点砜䂨砹硇硑硦葈𥔵礳栃礲䄃"],
        ["9940", "䄉禑禙辻稆込䅧窑䆲窼艹䇄竏竛䇏両筢筬筻簒簛䉠䉺类粜䊌粸䊔糭输烀𠳏総緔緐緽羮羴犟䎗耠耥笹耮耱联㷌垴炠肷胩䏭脌猪脎脒畠脔䐁㬹腖腙腚"],
        ["99a1", "䐓堺腼膄䐥膓䐭膥埯臁臤艔䒏芦艶苊苘苿䒰荗险榊萅烵葤惣蒈䔄蒾蓡蓸蔐蔸蕒䔻蕯蕰藠䕷虲蚒蚲蛯际螋䘆䘗袮裿褤襇覑𧥧訩訸誔誴豑賔賲贜䞘塟跃䟭仮踺嗘坔蹱嗵躰䠷軎転軤軭軲辷迁迊迌逳駄䢭飠鈓䤞鈨鉘鉫銱銮銿"],
        ["9a40", "鋣鋫鋳鋴鋽鍃鎄鎭䥅䥑麿鐗匁鐝鐭鐾䥪鑔鑹锭関䦧间阳䧥枠䨤靀䨵鞲韂噔䫤惨颹䬙飱塄餎餙冴餜餷饂饝饢䭰駅䮝騼鬏窃魩鮁鯝鯱鯴䱭鰠㝯𡯂鵉鰺"],
        ["9aa1", "黾噐鶓鶽鷀鷼银辶鹻麬麱麽黆铜黢黱黸竈齄𠂔𠊷𠎠椚铃妬𠓗塀铁㞹𠗕𠘕𠙶𡚺块煳𠫂𠫍𠮿呪吆𠯋咞𠯻𠰻𠱓𠱥𠱼惧𠲍噺𠲵𠳝𠳭𠵯𠶲𠷈楕鰯螥𠸄𠸎𠻗𠾐𠼭𠹳尠𠾼帋𡁜𡁏𡁶朞𡁻𡂈𡂖㙇𡂿𡃓𡄯𡄻卤蒭𡋣𡍵𡌶讁𡕷𡘙𡟃𡟇乸炻𡠭𡥪"],
        ["9b40", "𡨭𡩅𡰪𡱰𡲬𡻈拃𡻕𡼕熘桕𢁅槩㛈𢉼𢏗𢏺𢜪𢡱𢥏苽𢥧𢦓𢫕覥𢫨辠𢬎鞸𢬿顇骽𢱌"],
        ["9b62", "𢲈𢲷𥯨𢴈𢴒𢶷𢶕𢹂𢽴𢿌𣀳𣁦𣌟𣏞徱晈暿𧩹𣕧𣗳爁𤦺矗𣘚𣜖纇𠍆墵朎"],
        ["9ba1", "椘𣪧𧙗𥿢𣸑𣺹𧗾𢂚䣐䪸𤄙𨪚𤋮𤌍𤀻𤌴𤎖𤩅𠗊凒𠘑妟𡺨㮾𣳿𤐄𤓖垈𤙴㦛𤜯𨗨𩧉㝢𢇃譞𨭎駖𤠒𤣻𤨕爉𤫀𠱸奥𤺥𤾆𠝹軚𥀬劏圿煱𥊙𥐙𣽊𤪧喼𥑆𥑮𦭒釔㑳𥔿𧘲𥕞䜘𥕢𥕦𥟇𤤿𥡝偦㓻𣏌惞𥤃䝼𨥈𥪮𥮉𥰆𡶐垡煑澶𦄂𧰒遖𦆲𤾚譢𦐂𦑊"],
        ["9c40", "嵛𦯷輶𦒄𡤜諪𤧶𦒈𣿯𦔒䯀𦖿𦚵𢜛鑥𥟡憕娧晉侻嚹𤔡𦛼乪𤤴陖涏𦲽㘘襷𦞙𦡮𦐑𦡞營𦣇筂𩃀𠨑𦤦鄄𦤹穅鷰𦧺騦𦨭㙟𦑩𠀡禃𦨴𦭛崬𣔙菏𦮝䛐𦲤画补𦶮墶"],
        ["9ca1", "㜜𢖍𧁋𧇍㱔𧊀𧊅銁𢅺𧊋錰𧋦𤧐氹钟𧑐𠻸蠧裵𢤦𨑳𡞱溸𤨪𡠠㦤㚹尐秣䔿暶𩲭𩢤襃𧟌𧡘囖䃟𡘊㦡𣜯𨃨𡏅熭荦𧧝𩆨婧䲷𧂯𨦫𧧽𧨊𧬋𧵦𤅺筃祾𨀉澵𪋟樃𨌘厢𦸇鎿栶靝𨅯𨀣𦦵𡏭𣈯𨁈嶅𨰰𨂃圕頣𨥉嶫𤦈斾槕叒𤪥𣾁㰑朶𨂐𨃴𨄮𡾡𨅏"],
        ["9d40", "𨆉𨆯𨈚𨌆𨌯𨎊㗊𨑨𨚪䣺揦𨥖砈鉕𨦸䏲𨧧䏟𨧨𨭆𨯔姸𨰉輋𨿅𩃬筑𩄐𩄼㷷𩅞𤫊运犏嚋𩓧𩗩𩖰𩖸𩜲𩣑𩥉𩥪𩧃𩨨𩬎𩵚𩶛纟𩻸𩼣䲤镇𪊓熢𪋿䶑递𪗋䶜𠲜达嗁"],
        ["9da1", "辺𢒰边𤪓䔉繿潖檱仪㓤𨬬𧢝㜺躀𡟵𨀤𨭬𨮙𧨾𦚯㷫𧙕𣲷𥘵𥥖亚𥺁𦉘嚿𠹭踎孭𣺈𤲞揞拐𡟶𡡻攰嘭𥱊吚𥌑㷆𩶘䱽嘢嘞罉𥻘奵𣵀蝰东𠿪𠵉𣚺脗鵞贘瘻鱅癎瞹鍅吲腈苷嘥脲萘肽嗪祢噃吖𠺝㗎嘅嗱曱𨋢㘭甴嗰喺咗啲𠱁𠲖廐𥅈𠹶𢱢"],
        ["9e40", "𠺢麫絚嗞𡁵抝靭咔賍燶酶揼掹揾啩𢭃鱲𢺳冚㓟𠶧冧呍唞唓癦踭𦢊疱肶蠄螆裇膶萜𡃁䓬猄𤜆宐茋𦢓噻𢛴𧴯𤆣𧵳𦻐𧊶酰𡇙鈈𣳼𪚩𠺬𠻹牦𡲢䝎𤿂𧿹𠿫䃺"],
        ["9ea1", "鱝攟𢶠䣳𤟠𩵼𠿬𠸊恢𧖣𠿭"],
        ["9ead", "𦁈𡆇熣纎鵐业丄㕷嬍沲卧㚬㧜卽㚥𤘘墚𤭮舭呋垪𥪕𠥹"],
        ["9ec5", "㩒𢑥獴𩺬䴉鯭𣳾𩼰䱛𤾩𩖞𩿞葜𣶶𧊲𦞳𣜠挮紥𣻷𣸬㨪逈勌㹴㙺䗩𠒎癀嫰𠺶硺𧼮墧䂿噼鮋嵴癔𪐴麅䳡痹㟻愙𣃚𤏲"],
        ["9ef5", "噝𡊩垧𤥣𩸆刴𧂮㖭汊鵼"],
        ["9f40", "籖鬹埞𡝬屓擓𩓐𦌵𧅤蚭𠴨𦴢𤫢𠵱"],
        ["9f4f", "凾𡼏嶎霃𡷑麁遌笟鬂峑箣扨挵髿篏鬪籾鬮籂粆鰕篼鬉鼗鰛𤤾齚啳寃俽麘俲剠㸆勑坧偖妷帒韈鶫轜呩鞴饀鞺匬愰"],
        ["9fa1", "椬叚鰊鴂䰻陁榀傦畆𡝭駚剳"],
        ["9fae", "酙隁酜"],
        ["9fb2", "酑𨺗捿𦴣櫊嘑醎畺抅𠏼獏籰𥰡𣳽"],
        ["9fc1", "𤤙盖鮝个𠳔莾衂"],
        ["9fc9", "届槀僭坺刟巵从氱𠇲伹咜哚劚趂㗾弌㗳"],
        ["9fdb", "歒酼龥鮗頮颴骺麨麄煺笔"],
        ["9fe7", "毺蠘罸"],
        ["9feb", "嘠𪙊蹷齓"],
        ["9ff0", "跔蹏鸜踁抂𨍽踨蹵竓𤩷稾磘泪詧瘇"],
        ["a040", "𨩚鼦泎蟖痃𪊲硓咢贌狢獱謭猂瓱賫𤪻蘯徺袠䒷"],
        ["a055", "𡠻𦸅"],
        ["a058", "詾𢔛"],
        ["a05b", "惽癧髗鵄鍮鮏蟵"],
        ["a063", "蠏賷猬霡鮰㗖犲䰇籑饊𦅙慙䰄麖慽"],
        ["a073", "坟慯抦戹拎㩜懢厪𣏵捤栂㗒"],
        ["a0a1", "嵗𨯂迚𨸹"],
        ["a0a6", "僙𡵆礆匲阸𠼻䁥"],
        ["a0ae", "矾"],
        ["a0b0", "糂𥼚糚稭聦聣絍甅瓲覔舚朌聢𧒆聛瓰脃眤覉𦟌畓𦻑螩蟎臈螌詉貭譃眫瓸蓚㘵榲趦"],
        ["a0d4", "覩瑨涹蟁𤀑瓧㷛煶悤憜㳑煢恷"],
        ["a0e2", "罱𨬭牐惩䭾删㰘𣳇𥻗𧙖𥔱𡥄𡋾𩤃𦷜𧂭峁𦆭𨨏𣙷𠃮𦡆𤼎䕢嬟𦍌齐麦𦉫"],
        ["a3c0", "␀", 31, "␡"],
        ["c6a1", "①", 9, "⑴", 9, "ⅰ", 9, "丶丿亅亠冂冖冫勹匸卩厶夊宀巛⼳广廴彐彡攴无疒癶辵隶¨ˆヽヾゝゞ〃仝々〆〇ー［］✽ぁ", 23],
        ["c740", "す", 58, "ァアィイ"],
        ["c7a1", "ゥ", 81, "А", 5, "ЁЖ", 4],
        ["c840", "Л", 26, "ёж", 25, "⇧↸↹㇏𠃌乚𠂊刂䒑"],
        ["c8a1", "龰冈龱𧘇"],
        ["c8cd", "￢￤＇＂㈱№℡゛゜⺀⺄⺆⺇⺈⺊⺌⺍⺕⺜⺝⺥⺧⺪⺬⺮⺶⺼⺾⻆⻊⻌⻍⻏⻖⻗⻞⻣"],
        ["c8f5", "ʃɐɛɔɵœøŋʊɪ"],
        ["f9fe", "￭"],
        ["fa40", "𠕇鋛𠗟𣿅蕌䊵珯况㙉𤥂𨧤鍄𡧛苮𣳈砼杄拟𤤳𨦪𠊠𦮳𡌅侫𢓭倈𦴩𧪄𣘀𤪱𢔓倩𠍾徤𠎀𠍇滛𠐟偽儁㑺儎顬㝃萖𤦤𠒇兠𣎴兪𠯿𢃼𠋥𢔰𠖎𣈳𡦃宂蝽𠖳𣲙冲冸"],
        ["faa1", "鴴凉减凑㳜凓𤪦决凢卂凭菍椾𣜭彻刋刦刼劵剗劔効勅簕蕂勠蘍𦬓包𨫞啉滙𣾀𠥔𣿬匳卄𠯢泋𡜦栛珕恊㺪㣌𡛨燝䒢卭却𨚫卾卿𡖖𡘓矦厓𨪛厠厫厮玧𥝲㽙玜叁叅汉义埾叙㪫𠮏叠𣿫𢶣叶𠱷吓灹唫晗浛呭𦭓𠵴啝咏咤䞦𡜍𠻝㶴𠵍"],
        ["fb40", "𨦼𢚘啇䳭启琗喆喩嘅𡣗𤀺䕒𤐵暳𡂴嘷曍𣊊暤暭噍噏磱囱鞇叾圀囯园𨭦㘣𡉏坆𤆥汮炋坂㚱𦱾埦𡐖堃𡑔𤍣堦𤯵塜墪㕡壠壜𡈼壻寿坃𪅐𤉸鏓㖡够梦㛃湙"],
        ["fba1", "𡘾娤啓𡚒蔅姉𠵎𦲁𦴪𡟜姙𡟻𡞲𦶦浱𡠨𡛕姹𦹅媫婣㛦𤦩婷㜈媖瑥嫓𦾡𢕔㶅𡤑㜲𡚸広勐孶斈孼𧨎䀄䡝𠈄寕慠𡨴𥧌𠖥寳宝䴐尅𡭄尓珎尔𡲥𦬨屉䣝岅峩峯嶋𡷹𡸷崐崘嵆𡺤岺巗苼㠭𤤁𢁉𢅳芇㠶㯂帮檊幵幺𤒼𠳓厦亷廐厨𡝱帉廴𨒂"],
        ["fc40", "廹廻㢠廼栾鐛弍𠇁弢㫞䢮𡌺强𦢈𢏐彘𢑱彣鞽𦹮彲鍀𨨶徧嶶㵟𥉐𡽪𧃸𢙨釖𠊞𨨩怱暅𡡷㥣㷇㘹垐𢞴祱㹀悞悤悳𤦂𤦏𧩓璤僡媠慤萤慂慈𦻒憁凴𠙖憇宪𣾷"],
        ["fca1", "𢡟懓𨮝𩥝懐㤲𢦀𢣁怣慜攞掋𠄘担𡝰拕𢸍捬𤧟㨗搸揸𡎎𡟼撐澊𢸶頔𤂌𥜝擡擥鑻㩦携㩗敍漖𤨨𤨣斅敭敟𣁾斵𤥀䬷旑䃘𡠩无旣忟𣐀昘𣇷𣇸晄𣆤𣆥晋𠹵晧𥇦晳晴𡸽𣈱𨗴𣇈𥌓矅𢣷馤朂𤎜𤨡㬫槺𣟂杞杧杢𤇍𩃭柗䓩栢湐鈼栁𣏦𦶠桝"],
        ["fd40", "𣑯槡樋𨫟楳棃𣗍椁椀㴲㨁𣘼㮀枬楡𨩊䋼椶榘㮡𠏉荣傐槹𣙙𢄪橅𣜃檝㯳枱櫈𩆜㰍欝𠤣惞欵歴𢟍溵𣫛𠎵𡥘㝀吡𣭚毡𣻼毜氷𢒋𤣱𦭑汚舦汹𣶼䓅𣶽𤆤𤤌𤤀"],
        ["fda1", "𣳉㛥㳫𠴲鮃𣇹𢒑羏样𦴥𦶡𦷫涖浜湼漄𤥿𤂅𦹲蔳𦽴凇沜渝萮𨬡港𣸯瑓𣾂秌湏媑𣁋濸㜍澝𣸰滺𡒗𤀽䕕鏰潄潜㵎潴𩅰㴻澟𤅄濓𤂑𤅕𤀹𣿰𣾴𤄿凟𤅖𤅗𤅀𦇝灋灾炧炁烌烕烖烟䄄㷨熴熖𤉷焫煅媈煊煮岜𤍥煏鍢𤋁焬𤑚𤨧𤨢熺𨯨炽爎"],
        ["fe40", "鑂爕夑鑃爤鍁𥘅爮牀𤥴梽牕牗㹕𣁄栍漽犂猪猫𤠣𨠫䣭𨠄猨献珏玪𠰺𦨮珉瑉𤇢𡛧𤨤昣㛅𤦷𤦍𤧻珷琕椃𤨦琹𠗃㻗瑜𢢭瑠𨺲瑇珤瑶莹瑬㜰瑴鏱樬璂䥓𤪌"],
        ["fea1", "𤅟𤩹𨮏孆𨰃𡢞瓈𡦈甎瓩甞𨻙𡩋寗𨺬鎅畍畊畧畮𤾂㼄𤴓疎瑝疞疴瘂瘬癑癏癯癶𦏵皐臯㟸𦤑𦤎皡皥皷盌𦾟葢𥂝𥅽𡸜眞眦着撯𥈠睘𣊬瞯𨥤𨥨𡛁矴砉𡍶𤨒棊碯磇磓隥礮𥗠磗礴碱𧘌辸袄𨬫𦂃𢘜禆褀椂禀𥡗禝𧬹礼禩渪𧄦㺨秆𩄍秔"]
    ];
    module.exports = json;
},
/* iconv-lite/lib/streams.js */ function _(require, module, exports, __esModule, __esExport) {
    var Buffer = require("buffer").Buffer, Transform = require("stream").Transform;
    // == Exports ==================================================================
    module.exports = function (iconv) {
        // Additional Public API.
        iconv.encodeStream = function encodeStream(encoding, options) {
            return new IconvLiteEncoderStream(iconv.getEncoder(encoding, options), options);
        };
        iconv.decodeStream = function decodeStream(encoding, options) {
            return new IconvLiteDecoderStream(iconv.getDecoder(encoding, options), options);
        };
        iconv.supportsStreams = true;
        // Not published yet.
        iconv.IconvLiteEncoderStream = IconvLiteEncoderStream;
        iconv.IconvLiteDecoderStream = IconvLiteDecoderStream;
        iconv._collect = IconvLiteDecoderStream.prototype.collect;
    };
    // == Encoder stream =======================================================
    function IconvLiteEncoderStream(conv, options) {
        this.conv = conv;
        options = options || {};
        options.decodeStrings = false; // We accept only strings, so we don't need to decode them.
        Transform.call(this, options);
    }
    IconvLiteEncoderStream.prototype = Object.create(Transform.prototype, {
        constructor: { value: IconvLiteEncoderStream }
    });
    IconvLiteEncoderStream.prototype._transform = function (chunk, encoding, done) {
        if (typeof chunk != 'string')
            return done(new Error("Iconv encoding stream needs strings as its input."));
        try {
            var res = this.conv.write(chunk);
            if (res && res.length)
                this.push(res);
            done();
        }
        catch (e) {
            done(e);
        }
    };
    IconvLiteEncoderStream.prototype._flush = function (done) {
        try {
            var res = this.conv.end();
            if (res && res.length)
                this.push(res);
            done();
        }
        catch (e) {
            done(e);
        }
    };
    IconvLiteEncoderStream.prototype.collect = function (cb) {
        var chunks = [];
        this.on('error', cb);
        this.on('data', function (chunk) { chunks.push(chunk); });
        this.on('end', function () {
            cb(null, Buffer.concat(chunks));
        });
        return this;
    };
    // == Decoder stream =======================================================
    function IconvLiteDecoderStream(conv, options) {
        this.conv = conv;
        options = options || {};
        options.encoding = this.encoding = 'utf8'; // We output strings.
        Transform.call(this, options);
    }
    IconvLiteDecoderStream.prototype = Object.create(Transform.prototype, {
        constructor: { value: IconvLiteDecoderStream }
    });
    IconvLiteDecoderStream.prototype._transform = function (chunk, encoding, done) {
        if (!Buffer.isBuffer(chunk))
            return done(new Error("Iconv decoding stream needs buffers as its input."));
        try {
            var res = this.conv.write(chunk);
            if (res && res.length)
                this.push(res, this.encoding);
            done();
        }
        catch (e) {
            done(e);
        }
    };
    IconvLiteDecoderStream.prototype._flush = function (done) {
        try {
            var res = this.conv.end();
            if (res && res.length)
                this.push(res, this.encoding);
            done();
        }
        catch (e) {
            done(e);
        }
    };
    IconvLiteDecoderStream.prototype.collect = function (cb) {
        var res = '';
        this.on('error', cb);
        this.on('data', function (chunk) { res += chunk; });
        this.on('end', function () {
            cb(null, res);
        });
        return this;
    };
},
/* iconv-lite/lib/extend-node.js */ function _(require, module, exports, __esModule, __esExport) {
    var Buffer = require("buffer").Buffer;
    // Note: not polyfilled with safer-buffer on a purpose, as overrides Buffer
    // == Extend Node primitives to use iconv-lite =================================
    module.exports = function (iconv) {
        var original = undefined; // Place to keep original methods.
        // Node authors rewrote Buffer internals to make it compatible with
        // Uint8Array and we cannot patch key functions since then.
        // Note: this does use older Buffer API on a purpose
        iconv.supportsNodeEncodingsExtension = !(Buffer.from || new Buffer(0) instanceof Uint8Array);
        iconv.extendNodeEncodings = function extendNodeEncodings() {
            if (original)
                return;
            original = {};
            if (!iconv.supportsNodeEncodingsExtension) {
                console.error("ACTION NEEDED: require('iconv-lite').extendNodeEncodings() is not supported in your version of Node");
                console.error("See more info at https://github.com/ashtuchkin/iconv-lite/wiki/Node-v4-compatibility");
                return;
            }
            var nodeNativeEncodings = {
                'hex': true, 'utf8': true, 'utf-8': true, 'ascii': true, 'binary': true,
                'base64': true, 'ucs2': true, 'ucs-2': true, 'utf16le': true, 'utf-16le': true,
            };
            Buffer.isNativeEncoding = function (enc) {
                return enc && nodeNativeEncodings[enc.toLowerCase()];
            };
            // -- SlowBuffer -----------------------------------------------------------
            var SlowBuffer = require('buffer').SlowBuffer;
            original.SlowBufferToString = SlowBuffer.prototype.toString;
            SlowBuffer.prototype.toString = function (encoding, start, end) {
                encoding = String(encoding || 'utf8').toLowerCase();
                // Use native conversion when possible
                if (Buffer.isNativeEncoding(encoding))
                    return original.SlowBufferToString.call(this, encoding, start, end);
                // Otherwise, use our decoding method.
                if (typeof start == 'undefined')
                    start = 0;
                if (typeof end == 'undefined')
                    end = this.length;
                return iconv.decode(this.slice(start, end), encoding);
            };
            original.SlowBufferWrite = SlowBuffer.prototype.write;
            SlowBuffer.prototype.write = function (string, offset, length, encoding) {
                // Support both (string, offset, length, encoding)
                // and the legacy (string, encoding, offset, length)
                if (isFinite(offset)) {
                    if (!isFinite(length)) {
                        encoding = length;
                        length = undefined;
                    }
                }
                else { // legacy
                    var swap = encoding;
                    encoding = offset;
                    offset = length;
                    length = swap;
                }
                offset = +offset || 0;
                var remaining = this.length - offset;
                if (!length) {
                    length = remaining;
                }
                else {
                    length = +length;
                    if (length > remaining) {
                        length = remaining;
                    }
                }
                encoding = String(encoding || 'utf8').toLowerCase();
                // Use native conversion when possible
                if (Buffer.isNativeEncoding(encoding))
                    return original.SlowBufferWrite.call(this, string, offset, length, encoding);
                if (string.length > 0 && (length < 0 || offset < 0))
                    throw new RangeError('attempt to write beyond buffer bounds');
                // Otherwise, use our encoding method.
                var buf = iconv.encode(string, encoding);
                if (buf.length < length)
                    length = buf.length;
                buf.copy(this, offset, 0, length);
                return length;
            };
            // -- Buffer ---------------------------------------------------------------
            original.BufferIsEncoding = Buffer.isEncoding;
            Buffer.isEncoding = function (encoding) {
                return Buffer.isNativeEncoding(encoding) || iconv.encodingExists(encoding);
            };
            original.BufferByteLength = Buffer.byteLength;
            Buffer.byteLength = SlowBuffer.byteLength = function (str, encoding) {
                encoding = String(encoding || 'utf8').toLowerCase();
                // Use native conversion when possible
                if (Buffer.isNativeEncoding(encoding))
                    return original.BufferByteLength.call(this, str, encoding);
                // Slow, I know, but we don't have a better way yet.
                return iconv.encode(str, encoding).length;
            };
            original.BufferToString = Buffer.prototype.toString;
            Buffer.prototype.toString = function (encoding, start, end) {
                encoding = String(encoding || 'utf8').toLowerCase();
                // Use native conversion when possible
                if (Buffer.isNativeEncoding(encoding))
                    return original.BufferToString.call(this, encoding, start, end);
                // Otherwise, use our decoding method.
                if (typeof start == 'undefined')
                    start = 0;
                if (typeof end == 'undefined')
                    end = this.length;
                return iconv.decode(this.slice(start, end), encoding);
            };
            original.BufferWrite = Buffer.prototype.write;
            Buffer.prototype.write = function (string, offset, length, encoding) {
                var _offset = offset, _length = length, _encoding = encoding;
                // Support both (string, offset, length, encoding)
                // and the legacy (string, encoding, offset, length)
                if (isFinite(offset)) {
                    if (!isFinite(length)) {
                        encoding = length;
                        length = undefined;
                    }
                }
                else { // legacy
                    var swap = encoding;
                    encoding = offset;
                    offset = length;
                    length = swap;
                }
                encoding = String(encoding || 'utf8').toLowerCase();
                // Use native conversion when possible
                if (Buffer.isNativeEncoding(encoding))
                    return original.BufferWrite.call(this, string, _offset, _length, _encoding);
                offset = +offset || 0;
                var remaining = this.length - offset;
                if (!length) {
                    length = remaining;
                }
                else {
                    length = +length;
                    if (length > remaining) {
                        length = remaining;
                    }
                }
                if (string.length > 0 && (length < 0 || offset < 0))
                    throw new RangeError('attempt to write beyond buffer bounds');
                // Otherwise, use our encoding method.
                var buf = iconv.encode(string, encoding);
                if (buf.length < length)
                    length = buf.length;
                buf.copy(this, offset, 0, length);
                return length;
                // TODO: Set _charsWritten.
            };
            // -- Readable -------------------------------------------------------------
            if (iconv.supportsStreams) {
                var Readable = require('stream').Readable;
                original.ReadableSetEncoding = Readable.prototype.setEncoding;
                Readable.prototype.setEncoding = function setEncoding(enc, options) {
                    // Use our own decoder, it has the same interface.
                    // We cannot use original function as it doesn't handle BOM-s.
                    this._readableState.decoder = iconv.getDecoder(enc, options);
                    this._readableState.encoding = enc;
                };
                Readable.prototype.collect = iconv._collect;
            }
        };
        // Remove iconv-lite Node primitive extensions.
        iconv.undoExtendNodeEncodings = function undoExtendNodeEncodings() {
            if (!iconv.supportsNodeEncodingsExtension)
                return;
            if (!original)
                throw new Error("require('iconv-lite').undoExtendNodeEncodings(): Nothing to undo; extendNodeEncodings() is not called.");
            delete Buffer.isNativeEncoding;
            var SlowBuffer = require('buffer').SlowBuffer;
            SlowBuffer.prototype.toString = original.SlowBufferToString;
            SlowBuffer.prototype.write = original.SlowBufferWrite;
            Buffer.isEncoding = original.BufferIsEncoding;
            Buffer.byteLength = original.BufferByteLength;
            Buffer.prototype.toString = original.BufferToString;
            Buffer.prototype.write = original.BufferWrite;
            if (iconv.supportsStreams) {
                var Readable = require('stream').Readable;
                Readable.prototype.setEncoding = original.ReadableSetEncoding;
                delete Readable.prototype.collect;
            }
            original = undefined;
        };
    };
},
/* needle/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "needle",
        "version": "2.6.0",
        "description": "The leanest and most handsome HTTP client in the Nodelands.",
        "keywords": [
            "http",
            "https",
            "simple",
            "request",
            "client",
            "multipart",
            "upload",
            "proxy",
            "deflate",
            "timeout",
            "charset",
            "iconv",
            "cookie",
            "redirect"
        ],
        "tags": [
            "http",
            "https",
            "simple",
            "request",
            "client",
            "multipart",
            "upload",
            "proxy",
            "deflate",
            "timeout",
            "charset",
            "iconv",
            "cookie",
            "redirect"
        ],
        "author": "Tomás Pollak <tomas@forkhq.com>",
        "repository": {
            "type": "git",
            "url": "https://github.com/tomas/needle.git"
        },
        "dependencies": {
            "debug": "^3.2.6",
            "iconv-lite": "^0.4.4",
            "sax": "^1.2.4"
        },
        "devDependencies": {
            "JSONStream": "^1.3.5",
            "jschardet": "^1.6.0",
            "mocha": "^5.2.0",
            "q": "^1.5.1",
            "should": "^13.2.3",
            "sinon": "^2.3.0",
            "xml2js": "^0.4.19"
        },
        "scripts": {
            "test": "mocha test"
        },
        "directories": {
            "lib": "./lib"
        },
        "main": "./lib/needle",
        "bin": {
            "needle": "./bin/needle"
        },
        "license": "MIT",
        "engines": {
            "node": ">= 4.4.x"
        }
    };
    module.exports = json;
},
/* less/lib/less/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var environment_1 = tslib_1.__importDefault(require(108) /* ./environment/environment */);
    var data_1 = tslib_1.__importDefault(require(109) /* ./data */);
    var tree_1 = tslib_1.__importDefault(require(112) /* ./tree */);
    var abstract_file_manager_1 = tslib_1.__importDefault(require(68) /* ./environment/abstract-file-manager */);
    var abstract_plugin_loader_1 = tslib_1.__importDefault(require(160) /* ./environment/abstract-plugin-loader */);
    var visitors_1 = tslib_1.__importDefault(require(161) /* ./visitors */);
    var parser_1 = tslib_1.__importDefault(require(169) /* ./parser/parser */);
    var functions_1 = tslib_1.__importDefault(require(172) /* ./functions */);
    var contexts_1 = tslib_1.__importDefault(require(132) /* ./contexts */);
    var less_error_1 = tslib_1.__importDefault(require(120) /* ./less-error */);
    var transform_tree_1 = tslib_1.__importDefault(require(184) /* ./transform-tree */);
    var utils = tslib_1.__importStar(require(121) /* ./utils */);
    var plugin_manager_1 = tslib_1.__importDefault(require(185) /* ./plugin-manager */);
    var logger_1 = tslib_1.__importDefault(require(70) /* ./logger */);
    var source_map_output_1 = tslib_1.__importDefault(require(186) /* ./source-map-output */);
    var source_map_builder_1 = tslib_1.__importDefault(require(187) /* ./source-map-builder */);
    var parse_tree_1 = tslib_1.__importDefault(require(188) /* ./parse-tree */);
    var import_manager_1 = tslib_1.__importDefault(require(189) /* ./import-manager */);
    var parse_1 = tslib_1.__importDefault(require(190) /* ./parse */);
    var render_1 = tslib_1.__importDefault(require(191) /* ./render */);
    var package_json_1 = require(192) /* ../../package.json */;
    var parse_node_version_1 = tslib_1.__importDefault(require(193) /* parse-node-version */);
    function default_1(environment, fileManagers) {
        var sourceMapOutput, sourceMapBuilder, parseTree, importManager;
        environment = new environment_1.default(environment, fileManagers);
        sourceMapOutput = source_map_output_1.default(environment);
        sourceMapBuilder = source_map_builder_1.default(sourceMapOutput, environment);
        parseTree = parse_tree_1.default(sourceMapBuilder);
        importManager = import_manager_1.default(environment);
        var render = render_1.default(environment, parseTree, importManager);
        var parse = parse_1.default(environment, parseTree, importManager);
        var v = parse_node_version_1.default("v" + package_json_1.version);
        var initial = {
            version: [v.major, v.minor, v.patch],
            data: data_1.default,
            tree: tree_1.default,
            Environment: environment_1.default,
            AbstractFileManager: abstract_file_manager_1.default,
            AbstractPluginLoader: abstract_plugin_loader_1.default,
            environment: environment,
            visitors: visitors_1.default,
            Parser: parser_1.default,
            functions: functions_1.default(environment),
            contexts: contexts_1.default,
            SourceMapOutput: sourceMapOutput,
            SourceMapBuilder: sourceMapBuilder,
            ParseTree: parseTree,
            ImportManager: importManager,
            render: render,
            parse: parse,
            LessError: less_error_1.default,
            transformTree: transform_tree_1.default,
            utils: utils,
            PluginManager: plugin_manager_1.default,
            logger: logger_1.default
        };
        // Create a public API
        var ctor = function (t) {
            return function () {
                var obj = Object.create(t.prototype);
                t.apply(obj, Array.prototype.slice.call(arguments, 0));
                return obj;
            };
        };
        var t;
        var api = Object.create(initial);
        for (var n in initial.tree) {
            /* eslint guard-for-in: 0 */
            t = initial.tree[n];
            if (typeof t === 'function') {
                api[n.toLowerCase()] = ctor(t);
            }
            else {
                api[n] = Object.create(null);
                for (var o in t) {
                    /* eslint guard-for-in: 0 */
                    api[n][o.toLowerCase()] = ctor(t[o]);
                }
            }
        }
        /**
         * Some of the functions assume a `this` context of the API object,
         * which causes it to fail when wrapped for ES6 imports.
         *
         * An assumed `this` should be removed in the future.
         */
        initial.parse = initial.parse.bind(api);
        initial.render = initial.render.bind(api);
        return api;
    }
    exports.default = default_1;
    ;
},
/* less/lib/less/environment/environment.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var logger_1 = tslib_1.__importDefault(require(70) /* ../logger */);
    var Environment = /** @class */ (function () {
        function Environment(externalEnvironment, fileManagers) {
            this.fileManagers = fileManagers || [];
            externalEnvironment = externalEnvironment || {};
            var optionalFunctions = ['encodeBase64', 'mimeLookup', 'charsetLookup', 'getSourceMapGenerator'];
            var requiredFunctions = [];
            var functions = requiredFunctions.concat(optionalFunctions);
            for (var i = 0; i < functions.length; i++) {
                var propName = functions[i];
                var environmentFunc = externalEnvironment[propName];
                if (environmentFunc) {
                    this[propName] = environmentFunc.bind(externalEnvironment);
                }
                else if (i < requiredFunctions.length) {
                    this.warn("missing required function in environment - " + propName);
                }
            }
        }
        Environment.prototype.getFileManager = function (filename, currentDirectory, options, environment, isSync) {
            if (!filename) {
                logger_1.default.warn('getFileManager called with no filename.. Please report this issue. continuing.');
            }
            if (currentDirectory == null) {
                logger_1.default.warn('getFileManager called with null directory.. Please report this issue. continuing.');
            }
            var fileManagers = this.fileManagers;
            if (options.pluginManager) {
                fileManagers = [].concat(fileManagers).concat(options.pluginManager.getFileManagers());
            }
            for (var i = fileManagers.length - 1; i >= 0; i--) {
                var fileManager = fileManagers[i];
                if (fileManager[isSync ? 'supportsSync' : 'supports'](filename, currentDirectory, options, environment)) {
                    return fileManager;
                }
            }
            return null;
        };
        Environment.prototype.addFileManager = function (fileManager) {
            this.fileManagers.push(fileManager);
        };
        Environment.prototype.clearFileManagers = function () {
            this.fileManagers = [];
        };
        return Environment;
    }());
    exports.default = Environment;
},
/* less/lib/less/data/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var colors_1 = tslib_1.__importDefault(require(110) /* ./colors */);
    var unit_conversions_1 = tslib_1.__importDefault(require(111) /* ./unit-conversions */);
    exports.default = { colors: colors_1.default, unitConversions: unit_conversions_1.default };
},
/* less/lib/less/data/colors.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.default = {
        'aliceblue': '#f0f8ff',
        'antiquewhite': '#faebd7',
        'aqua': '#00ffff',
        'aquamarine': '#7fffd4',
        'azure': '#f0ffff',
        'beige': '#f5f5dc',
        'bisque': '#ffe4c4',
        'black': '#000000',
        'blanchedalmond': '#ffebcd',
        'blue': '#0000ff',
        'blueviolet': '#8a2be2',
        'brown': '#a52a2a',
        'burlywood': '#deb887',
        'cadetblue': '#5f9ea0',
        'chartreuse': '#7fff00',
        'chocolate': '#d2691e',
        'coral': '#ff7f50',
        'cornflowerblue': '#6495ed',
        'cornsilk': '#fff8dc',
        'crimson': '#dc143c',
        'cyan': '#00ffff',
        'darkblue': '#00008b',
        'darkcyan': '#008b8b',
        'darkgoldenrod': '#b8860b',
        'darkgray': '#a9a9a9',
        'darkgrey': '#a9a9a9',
        'darkgreen': '#006400',
        'darkkhaki': '#bdb76b',
        'darkmagenta': '#8b008b',
        'darkolivegreen': '#556b2f',
        'darkorange': '#ff8c00',
        'darkorchid': '#9932cc',
        'darkred': '#8b0000',
        'darksalmon': '#e9967a',
        'darkseagreen': '#8fbc8f',
        'darkslateblue': '#483d8b',
        'darkslategray': '#2f4f4f',
        'darkslategrey': '#2f4f4f',
        'darkturquoise': '#00ced1',
        'darkviolet': '#9400d3',
        'deeppink': '#ff1493',
        'deepskyblue': '#00bfff',
        'dimgray': '#696969',
        'dimgrey': '#696969',
        'dodgerblue': '#1e90ff',
        'firebrick': '#b22222',
        'floralwhite': '#fffaf0',
        'forestgreen': '#228b22',
        'fuchsia': '#ff00ff',
        'gainsboro': '#dcdcdc',
        'ghostwhite': '#f8f8ff',
        'gold': '#ffd700',
        'goldenrod': '#daa520',
        'gray': '#808080',
        'grey': '#808080',
        'green': '#008000',
        'greenyellow': '#adff2f',
        'honeydew': '#f0fff0',
        'hotpink': '#ff69b4',
        'indianred': '#cd5c5c',
        'indigo': '#4b0082',
        'ivory': '#fffff0',
        'khaki': '#f0e68c',
        'lavender': '#e6e6fa',
        'lavenderblush': '#fff0f5',
        'lawngreen': '#7cfc00',
        'lemonchiffon': '#fffacd',
        'lightblue': '#add8e6',
        'lightcoral': '#f08080',
        'lightcyan': '#e0ffff',
        'lightgoldenrodyellow': '#fafad2',
        'lightgray': '#d3d3d3',
        'lightgrey': '#d3d3d3',
        'lightgreen': '#90ee90',
        'lightpink': '#ffb6c1',
        'lightsalmon': '#ffa07a',
        'lightseagreen': '#20b2aa',
        'lightskyblue': '#87cefa',
        'lightslategray': '#778899',
        'lightslategrey': '#778899',
        'lightsteelblue': '#b0c4de',
        'lightyellow': '#ffffe0',
        'lime': '#00ff00',
        'limegreen': '#32cd32',
        'linen': '#faf0e6',
        'magenta': '#ff00ff',
        'maroon': '#800000',
        'mediumaquamarine': '#66cdaa',
        'mediumblue': '#0000cd',
        'mediumorchid': '#ba55d3',
        'mediumpurple': '#9370d8',
        'mediumseagreen': '#3cb371',
        'mediumslateblue': '#7b68ee',
        'mediumspringgreen': '#00fa9a',
        'mediumturquoise': '#48d1cc',
        'mediumvioletred': '#c71585',
        'midnightblue': '#191970',
        'mintcream': '#f5fffa',
        'mistyrose': '#ffe4e1',
        'moccasin': '#ffe4b5',
        'navajowhite': '#ffdead',
        'navy': '#000080',
        'oldlace': '#fdf5e6',
        'olive': '#808000',
        'olivedrab': '#6b8e23',
        'orange': '#ffa500',
        'orangered': '#ff4500',
        'orchid': '#da70d6',
        'palegoldenrod': '#eee8aa',
        'palegreen': '#98fb98',
        'paleturquoise': '#afeeee',
        'palevioletred': '#d87093',
        'papayawhip': '#ffefd5',
        'peachpuff': '#ffdab9',
        'peru': '#cd853f',
        'pink': '#ffc0cb',
        'plum': '#dda0dd',
        'powderblue': '#b0e0e6',
        'purple': '#800080',
        'rebeccapurple': '#663399',
        'red': '#ff0000',
        'rosybrown': '#bc8f8f',
        'royalblue': '#4169e1',
        'saddlebrown': '#8b4513',
        'salmon': '#fa8072',
        'sandybrown': '#f4a460',
        'seagreen': '#2e8b57',
        'seashell': '#fff5ee',
        'sienna': '#a0522d',
        'silver': '#c0c0c0',
        'skyblue': '#87ceeb',
        'slateblue': '#6a5acd',
        'slategray': '#708090',
        'slategrey': '#708090',
        'snow': '#fffafa',
        'springgreen': '#00ff7f',
        'steelblue': '#4682b4',
        'tan': '#d2b48c',
        'teal': '#008080',
        'thistle': '#d8bfd8',
        'tomato': '#ff6347',
        'turquoise': '#40e0d0',
        'violet': '#ee82ee',
        'wheat': '#f5deb3',
        'white': '#ffffff',
        'whitesmoke': '#f5f5f5',
        'yellow': '#ffff00',
        'yellowgreen': '#9acd32'
    };
},
/* less/lib/less/data/unit-conversions.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.default = {
        length: {
            'm': 1,
            'cm': 0.01,
            'mm': 0.001,
            'in': 0.0254,
            'px': 0.0254 / 96,
            'pt': 0.0254 / 72,
            'pc': 0.0254 / 72 * 12
        },
        duration: {
            's': 1,
            'ms': 0.001
        },
        angle: {
            'rad': 1 / (2 * Math.PI),
            'deg': 1 / 360,
            'grad': 1 / 400,
            'turn': 1
        }
    };
},
/* less/lib/less/tree/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var color_1 = tslib_1.__importDefault(require(114) /* ./color */);
    var atrule_1 = tslib_1.__importDefault(require(115) /* ./atrule */);
    var detached_ruleset_1 = tslib_1.__importDefault(require(135) /* ./detached-ruleset */);
    var operation_1 = tslib_1.__importDefault(require(136) /* ./operation */);
    var dimension_1 = tslib_1.__importDefault(require(137) /* ./dimension */);
    var unit_1 = tslib_1.__importDefault(require(138) /* ./unit */);
    var keyword_1 = tslib_1.__importDefault(require(128) /* ./keyword */);
    var variable_1 = tslib_1.__importDefault(require(139) /* ./variable */);
    var property_1 = tslib_1.__importDefault(require(143) /* ./property */);
    var ruleset_1 = tslib_1.__importDefault(require(125) /* ./ruleset */);
    var element_1 = tslib_1.__importDefault(require(117) /* ./element */);
    var attribute_1 = tslib_1.__importDefault(require(144) /* ./attribute */);
    var combinator_1 = tslib_1.__importDefault(require(119) /* ./combinator */);
    var selector_1 = tslib_1.__importDefault(require(116) /* ./selector */);
    var quoted_1 = tslib_1.__importDefault(require(145) /* ./quoted */);
    var expression_1 = tslib_1.__importDefault(require(142) /* ./expression */);
    var declaration_1 = tslib_1.__importDefault(require(126) /* ./declaration */);
    var call_1 = tslib_1.__importDefault(require(140) /* ./call */);
    var url_1 = tslib_1.__importDefault(require(146) /* ./url */);
    var import_1 = tslib_1.__importDefault(require(147) /* ./import */);
    var comment_1 = tslib_1.__importDefault(require(130) /* ./comment */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ./anonymous */);
    var value_1 = tslib_1.__importDefault(require(127) /* ./value */);
    var javascript_1 = tslib_1.__importDefault(require(149) /* ./javascript */);
    var assignment_1 = tslib_1.__importDefault(require(151) /* ./assignment */);
    var condition_1 = tslib_1.__importDefault(require(152) /* ./condition */);
    var paren_1 = tslib_1.__importDefault(require(118) /* ./paren */);
    var media_1 = tslib_1.__importDefault(require(148) /* ./media */);
    var unicode_descriptor_1 = tslib_1.__importDefault(require(153) /* ./unicode-descriptor */);
    var negative_1 = tslib_1.__importDefault(require(154) /* ./negative */);
    var extend_1 = tslib_1.__importDefault(require(155) /* ./extend */);
    var variable_call_1 = tslib_1.__importDefault(require(156) /* ./variable-call */);
    var namespace_value_1 = tslib_1.__importDefault(require(157) /* ./namespace-value */);
    // mixins
    var mixin_call_1 = tslib_1.__importDefault(require(158) /* ./mixin-call */);
    var mixin_definition_1 = tslib_1.__importDefault(require(159) /* ./mixin-definition */);
    exports.default = {
        Node: node_1.default, Color: color_1.default, AtRule: atrule_1.default, DetachedRuleset: detached_ruleset_1.default, Operation: operation_1.default,
        Dimension: dimension_1.default, Unit: unit_1.default, Keyword: keyword_1.default, Variable: variable_1.default, Property: property_1.default,
        Ruleset: ruleset_1.default, Element: element_1.default, Attribute: attribute_1.default, Combinator: combinator_1.default, Selector: selector_1.default,
        Quoted: quoted_1.default, Expression: expression_1.default, Declaration: declaration_1.default, Call: call_1.default, URL: url_1.default, Import: import_1.default,
        Comment: comment_1.default, Anonymous: anonymous_1.default, Value: value_1.default, JavaScript: javascript_1.default, Assignment: assignment_1.default,
        Condition: condition_1.default, Paren: paren_1.default, Media: media_1.default, UnicodeDescriptor: unicode_descriptor_1.default, Negative: negative_1.default,
        Extend: extend_1.default, VariableCall: variable_call_1.default, NamespaceValue: namespace_value_1.default,
        mixin: {
            Call: mixin_call_1.default,
            Definition: mixin_definition_1.default
        }
    };
},
/* less/lib/less/tree/node.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    /**
     * The reason why Node is a class and other nodes simply do not extend
     * from Node (since we're transpiling) is due to this issue:
     *
     * https://github.com/less/less.js/issues/3434
     */
    var Node = /** @class */ (function () {
        function Node() {
            this.parent = null;
            this.visibilityBlocks = undefined;
            this.nodeVisible = undefined;
            this.rootNode = null;
            this.parsed = null;
            var self = this;
            Object.defineProperty(this, 'currentFileInfo', {
                get: function () { return self.fileInfo(); }
            });
            Object.defineProperty(this, 'index', {
                get: function () { return self.getIndex(); }
            });
        }
        Node.prototype.setParent = function (nodes, parent) {
            function set(node) {
                if (node && node instanceof Node) {
                    node.parent = parent;
                }
            }
            if (Array.isArray(nodes)) {
                nodes.forEach(set);
            }
            else {
                set(nodes);
            }
        };
        Node.prototype.getIndex = function () {
            return this._index || (this.parent && this.parent.getIndex()) || 0;
        };
        Node.prototype.fileInfo = function () {
            return this._fileInfo || (this.parent && this.parent.fileInfo()) || {};
        };
        Node.prototype.isRulesetLike = function () { return false; };
        Node.prototype.toCSS = function (context) {
            var strs = [];
            this.genCSS(context, {
                add: function (chunk, fileInfo, index) {
                    strs.push(chunk);
                },
                isEmpty: function () {
                    return strs.length === 0;
                }
            });
            return strs.join('');
        };
        Node.prototype.genCSS = function (context, output) {
            output.add(this.value);
        };
        Node.prototype.accept = function (visitor) {
            this.value = visitor.visit(this.value);
        };
        Node.prototype.eval = function () { return this; };
        Node.prototype._operate = function (context, op, a, b) {
            switch (op) {
                case '+': return a + b;
                case '-': return a - b;
                case '*': return a * b;
                case '/': return a / b;
            }
        };
        Node.prototype.fround = function (context, value) {
            var precision = context && context.numPrecision;
            // add "epsilon" to ensure numbers like 1.000000005 (represented as 1.000000004999...) are properly rounded:
            return (precision) ? Number((value + 2e-16).toFixed(precision)) : value;
        };
        Node.compare = function (a, b) {
            /* returns:
             -1: a < b
             0: a = b
             1: a > b
             and *any* other value for a != b (e.g. undefined, NaN, -2 etc.) */
            if ((a.compare) &&
                // for "symmetric results" force toCSS-based comparison
                // of Quoted or Anonymous if either value is one of those
                !(b.type === 'Quoted' || b.type === 'Anonymous')) {
                return a.compare(b);
            }
            else if (b.compare) {
                return -b.compare(a);
            }
            else if (a.type !== b.type) {
                return undefined;
            }
            a = a.value;
            b = b.value;
            if (!Array.isArray(a)) {
                return a === b ? 0 : undefined;
            }
            if (a.length !== b.length) {
                return undefined;
            }
            for (var i = 0; i < a.length; i++) {
                if (Node.compare(a[i], b[i]) !== 0) {
                    return undefined;
                }
            }
            return 0;
        };
        Node.numericCompare = function (a, b) {
            return a < b ? -1
                : a === b ? 0
                    : a > b ? 1 : undefined;
        };
        // Returns true if this node represents root of ast imported by reference
        Node.prototype.blocksVisibility = function () {
            if (this.visibilityBlocks == null) {
                this.visibilityBlocks = 0;
            }
            return this.visibilityBlocks !== 0;
        };
        Node.prototype.addVisibilityBlock = function () {
            if (this.visibilityBlocks == null) {
                this.visibilityBlocks = 0;
            }
            this.visibilityBlocks = this.visibilityBlocks + 1;
        };
        Node.prototype.removeVisibilityBlock = function () {
            if (this.visibilityBlocks == null) {
                this.visibilityBlocks = 0;
            }
            this.visibilityBlocks = this.visibilityBlocks - 1;
        };
        // Turns on node visibility - if called node will be shown in output regardless
        // of whether it comes from import by reference or not
        Node.prototype.ensureVisibility = function () {
            this.nodeVisible = true;
        };
        // Turns off node visibility - if called node will NOT be shown in output regardless
        // of whether it comes from import by reference or not
        Node.prototype.ensureInvisibility = function () {
            this.nodeVisible = false;
        };
        // return values:
        // false - the node must not be visible
        // true - the node must be visible
        // undefined or null - the node has the same visibility as its parent
        Node.prototype.isVisible = function () {
            return this.nodeVisible;
        };
        Node.prototype.visibilityInfo = function () {
            return {
                visibilityBlocks: this.visibilityBlocks,
                nodeVisible: this.nodeVisible
            };
        };
        Node.prototype.copyVisibilityInfo = function (info) {
            if (!info) {
                return;
            }
            this.visibilityBlocks = info.visibilityBlocks;
            this.nodeVisible = info.nodeVisible;
        };
        return Node;
    }());
    exports.default = Node;
},
/* less/lib/less/tree/color.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var colors_1 = tslib_1.__importDefault(require(110) /* ../data/colors */);
    //
    // RGB Colors - #ff0014, #eee
    //
    var Color = function (rgb, a, originalForm) {
        var self = this;
        //
        // The end goal here, is to parse the arguments
        // into an integer triplet, such as `128, 255, 0`
        //
        // This facilitates operations and conversions.
        //
        if (Array.isArray(rgb)) {
            this.rgb = rgb;
        }
        else if (rgb.length >= 6) {
            this.rgb = [];
            rgb.match(/.{2}/g).map(function (c, i) {
                if (i < 3) {
                    self.rgb.push(parseInt(c, 16));
                }
                else {
                    self.alpha = (parseInt(c, 16)) / 255;
                }
            });
        }
        else {
            this.rgb = [];
            rgb.split('').map(function (c, i) {
                if (i < 3) {
                    self.rgb.push(parseInt(c + c, 16));
                }
                else {
                    self.alpha = (parseInt(c + c, 16)) / 255;
                }
            });
        }
        this.alpha = this.alpha || (typeof a === 'number' ? a : 1);
        if (typeof originalForm !== 'undefined') {
            this.value = originalForm;
        }
    };
    Color.prototype = Object.assign(new node_1.default(), {
        type: 'Color',
        luma: function () {
            var r = this.rgb[0] / 255, g = this.rgb[1] / 255, b = this.rgb[2] / 255;
            r = (r <= 0.03928) ? r / 12.92 : Math.pow(((r + 0.055) / 1.055), 2.4);
            g = (g <= 0.03928) ? g / 12.92 : Math.pow(((g + 0.055) / 1.055), 2.4);
            b = (b <= 0.03928) ? b / 12.92 : Math.pow(((b + 0.055) / 1.055), 2.4);
            return 0.2126 * r + 0.7152 * g + 0.0722 * b;
        },
        genCSS: function (context, output) {
            output.add(this.toCSS(context));
        },
        toCSS: function (context, doNotCompress) {
            var compress = context && context.compress && !doNotCompress;
            var color;
            var alpha;
            var colorFunction;
            var args = [];
            // `value` is set if this color was originally
            // converted from a named color string so we need
            // to respect this and try to output named color too.
            alpha = this.fround(context, this.alpha);
            if (this.value) {
                if (this.value.indexOf('rgb') === 0) {
                    if (alpha < 1) {
                        colorFunction = 'rgba';
                    }
                }
                else if (this.value.indexOf('hsl') === 0) {
                    if (alpha < 1) {
                        colorFunction = 'hsla';
                    }
                    else {
                        colorFunction = 'hsl';
                    }
                }
                else {
                    return this.value;
                }
            }
            else {
                if (alpha < 1) {
                    colorFunction = 'rgba';
                }
            }
            switch (colorFunction) {
                case 'rgba':
                    args = this.rgb.map(function (c) {
                        return clamp(Math.round(c), 255);
                    }).concat(clamp(alpha, 1));
                    break;
                case 'hsla':
                    args.push(clamp(alpha, 1));
                case 'hsl':
                    color = this.toHSL();
                    args = [
                        this.fround(context, color.h),
                        this.fround(context, color.s * 100) + "%",
                        this.fround(context, color.l * 100) + "%"
                    ].concat(args);
            }
            if (colorFunction) {
                // Values are capped between `0` and `255`, rounded and zero-padded.
                return colorFunction + "(" + args.join("," + (compress ? '' : ' ')) + ")";
            }
            color = this.toRGB();
            if (compress) {
                var splitcolor = color.split('');
                // Convert color to short format
                if (splitcolor[1] === splitcolor[2] && splitcolor[3] === splitcolor[4] && splitcolor[5] === splitcolor[6]) {
                    color = "#" + splitcolor[1] + splitcolor[3] + splitcolor[5];
                }
            }
            return color;
        },
        //
        // Operations have to be done per-channel, if not,
        // channels will spill onto each other. Once we have
        // our result, in the form of an integer triplet,
        // we create a new Color node to hold the result.
        //
        operate: function (context, op, other) {
            var rgb = new Array(3);
            var alpha = this.alpha * (1 - other.alpha) + other.alpha;
            for (var c = 0; c < 3; c++) {
                rgb[c] = this._operate(context, op, this.rgb[c], other.rgb[c]);
            }
            return new Color(rgb, alpha);
        },
        toRGB: function () {
            return toHex(this.rgb);
        },
        toHSL: function () {
            var r = this.rgb[0] / 255, g = this.rgb[1] / 255, b = this.rgb[2] / 255, a = this.alpha;
            var max = Math.max(r, g, b), min = Math.min(r, g, b);
            var h;
            var s;
            var l = (max + min) / 2;
            var d = max - min;
            if (max === min) {
                h = s = 0;
            }
            else {
                s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
                switch (max) {
                    case r:
                        h = (g - b) / d + (g < b ? 6 : 0);
                        break;
                    case g:
                        h = (b - r) / d + 2;
                        break;
                    case b:
                        h = (r - g) / d + 4;
                        break;
                }
                h /= 6;
            }
            return { h: h * 360, s: s, l: l, a: a };
        },
        // Adapted from http://mjijackson.com/2008/02/rgb-to-hsl-and-rgb-to-hsv-color-model-conversion-algorithms-in-javascript
        toHSV: function () {
            var r = this.rgb[0] / 255, g = this.rgb[1] / 255, b = this.rgb[2] / 255, a = this.alpha;
            var max = Math.max(r, g, b), min = Math.min(r, g, b);
            var h;
            var s;
            var v = max;
            var d = max - min;
            if (max === 0) {
                s = 0;
            }
            else {
                s = d / max;
            }
            if (max === min) {
                h = 0;
            }
            else {
                switch (max) {
                    case r:
                        h = (g - b) / d + (g < b ? 6 : 0);
                        break;
                    case g:
                        h = (b - r) / d + 2;
                        break;
                    case b:
                        h = (r - g) / d + 4;
                        break;
                }
                h /= 6;
            }
            return { h: h * 360, s: s, v: v, a: a };
        },
        toARGB: function () {
            return toHex([this.alpha * 255].concat(this.rgb));
        },
        compare: function (x) {
            return (x.rgb &&
                x.rgb[0] === this.rgb[0] &&
                x.rgb[1] === this.rgb[1] &&
                x.rgb[2] === this.rgb[2] &&
                x.alpha === this.alpha) ? 0 : undefined;
        }
    });
    Color.fromKeyword = function (keyword) {
        var c;
        var key = keyword.toLowerCase();
        if (colors_1.default.hasOwnProperty(key)) {
            c = new Color(colors_1.default[key].slice(1));
        }
        else if (key === 'transparent') {
            c = new Color([0, 0, 0], 0);
        }
        if (c) {
            c.value = keyword;
            return c;
        }
    };
    function clamp(v, max) {
        return Math.min(Math.max(v, 0), max);
    }
    function toHex(v) {
        return "#" + v.map(function (c) {
            c = clamp(Math.round(c), 255);
            return (c < 16 ? '0' : '') + c.toString(16);
        }).join('');
    }
    exports.default = Color;
},
/* less/lib/less/tree/atrule.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var selector_1 = tslib_1.__importDefault(require(116) /* ./selector */);
    var ruleset_1 = tslib_1.__importDefault(require(125) /* ./ruleset */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ./anonymous */);
    var AtRule = function (name, value, rules, index, currentFileInfo, debugInfo, isRooted, visibilityInfo) {
        var i;
        this.name = name;
        this.value = (value instanceof node_1.default) ? value : (value ? new anonymous_1.default(value) : value);
        if (rules) {
            if (Array.isArray(rules)) {
                this.rules = rules;
            }
            else {
                this.rules = [rules];
                this.rules[0].selectors = (new selector_1.default([], null, null, index, currentFileInfo)).createEmptySelectors();
            }
            for (i = 0; i < this.rules.length; i++) {
                this.rules[i].allowImports = true;
            }
            this.setParent(this.rules, this);
        }
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.debugInfo = debugInfo;
        this.isRooted = isRooted || false;
        this.copyVisibilityInfo(visibilityInfo);
        this.allowRoot = true;
    };
    AtRule.prototype = Object.assign(new node_1.default(), {
        type: 'AtRule',
        accept: function (visitor) {
            var value = this.value, rules = this.rules;
            if (rules) {
                this.rules = visitor.visitArray(rules);
            }
            if (value) {
                this.value = visitor.visit(value);
            }
        },
        isRulesetLike: function () {
            return this.rules || !this.isCharset();
        },
        isCharset: function () {
            return '@charset' === this.name;
        },
        genCSS: function (context, output) {
            var value = this.value, rules = this.rules;
            output.add(this.name, this.fileInfo(), this.getIndex());
            if (value) {
                output.add(' ');
                value.genCSS(context, output);
            }
            if (rules) {
                this.outputRuleset(context, output, rules);
            }
            else {
                output.add(';');
            }
        },
        eval: function (context) {
            var mediaPathBackup, mediaBlocksBackup, value = this.value, rules = this.rules;
            // media stored inside other atrule should not bubble over it
            // backpup media bubbling information
            mediaPathBackup = context.mediaPath;
            mediaBlocksBackup = context.mediaBlocks;
            // deleted media bubbling information
            context.mediaPath = [];
            context.mediaBlocks = [];
            if (value) {
                value = value.eval(context);
            }
            if (rules) {
                // assuming that there is only one rule at this point - that is how parser constructs the rule
                rules = [rules[0].eval(context)];
                rules[0].root = true;
            }
            // restore media bubbling information
            context.mediaPath = mediaPathBackup;
            context.mediaBlocks = mediaBlocksBackup;
            return new AtRule(this.name, value, rules, this.getIndex(), this.fileInfo(), this.debugInfo, this.isRooted, this.visibilityInfo());
        },
        variable: function (name) {
            if (this.rules) {
                // assuming that there is only one rule at this point - that is how parser constructs the rule
                return ruleset_1.default.prototype.variable.call(this.rules[0], name);
            }
        },
        find: function () {
            if (this.rules) {
                // assuming that there is only one rule at this point - that is how parser constructs the rule
                return ruleset_1.default.prototype.find.apply(this.rules[0], arguments);
            }
        },
        rulesets: function () {
            if (this.rules) {
                // assuming that there is only one rule at this point - that is how parser constructs the rule
                return ruleset_1.default.prototype.rulesets.apply(this.rules[0]);
            }
        },
        outputRuleset: function (context, output, rules) {
            var ruleCnt = rules.length;
            var i;
            context.tabLevel = (context.tabLevel | 0) + 1;
            // Compressed
            if (context.compress) {
                output.add('{');
                for (i = 0; i < ruleCnt; i++) {
                    rules[i].genCSS(context, output);
                }
                output.add('}');
                context.tabLevel--;
                return;
            }
            // Non-compressed
            var tabSetStr = "\n" + Array(context.tabLevel).join('  '), tabRuleStr = tabSetStr + "  ";
            if (!ruleCnt) {
                output.add(" {" + tabSetStr + "}");
            }
            else {
                output.add(" {" + tabRuleStr);
                rules[0].genCSS(context, output);
                for (i = 1; i < ruleCnt; i++) {
                    output.add(tabRuleStr);
                    rules[i].genCSS(context, output);
                }
                output.add(tabSetStr + "}");
            }
            context.tabLevel--;
        }
    });
    exports.default = AtRule;
},
/* less/lib/less/tree/selector.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var element_1 = tslib_1.__importDefault(require(117) /* ./element */);
    var less_error_1 = tslib_1.__importDefault(require(120) /* ../less-error */);
    var Selector = function (elements, extendList, condition, index, currentFileInfo, visibilityInfo) {
        this.extendList = extendList;
        this.condition = condition;
        this.evaldCondition = !condition;
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.elements = this.getElements(elements);
        this.mixinElements_ = undefined;
        this.copyVisibilityInfo(visibilityInfo);
        this.setParent(this.elements, this);
    };
    Selector.prototype = Object.assign(new node_1.default(), {
        type: 'Selector',
        accept: function (visitor) {
            if (this.elements) {
                this.elements = visitor.visitArray(this.elements);
            }
            if (this.extendList) {
                this.extendList = visitor.visitArray(this.extendList);
            }
            if (this.condition) {
                this.condition = visitor.visit(this.condition);
            }
        },
        createDerived: function (elements, extendList, evaldCondition) {
            elements = this.getElements(elements);
            var newSelector = new Selector(elements, extendList || this.extendList, null, this.getIndex(), this.fileInfo(), this.visibilityInfo());
            newSelector.evaldCondition = (evaldCondition != null) ? evaldCondition : this.evaldCondition;
            newSelector.mediaEmpty = this.mediaEmpty;
            return newSelector;
        },
        getElements: function (els) {
            if (!els) {
                return [new element_1.default('', '&', false, this._index, this._fileInfo)];
            }
            if (typeof els === 'string') {
                this.parse.parseNode(els, ['selector'], this._index, this._fileInfo, function (err, result) {
                    if (err) {
                        throw new less_error_1.default({
                            index: err.index,
                            message: err.message
                        }, this.parse.imports, this._fileInfo.filename);
                    }
                    els = result[0].elements;
                });
            }
            return els;
        },
        createEmptySelectors: function () {
            var el = new element_1.default('', '&', false, this._index, this._fileInfo), sels = [new Selector([el], null, null, this._index, this._fileInfo)];
            sels[0].mediaEmpty = true;
            return sels;
        },
        match: function (other) {
            var elements = this.elements;
            var len = elements.length;
            var olen;
            var i;
            other = other.mixinElements();
            olen = other.length;
            if (olen === 0 || len < olen) {
                return 0;
            }
            else {
                for (i = 0; i < olen; i++) {
                    if (elements[i].value !== other[i]) {
                        return 0;
                    }
                }
            }
            return olen; // return number of matched elements
        },
        mixinElements: function () {
            if (this.mixinElements_) {
                return this.mixinElements_;
            }
            var elements = this.elements.map(function (v) {
                return v.combinator.value + (v.value.value || v.value);
            }).join('').match(/[,&#\*\.\w-]([\w-]|(\\.))*/g);
            if (elements) {
                if (elements[0] === '&') {
                    elements.shift();
                }
            }
            else {
                elements = [];
            }
            return (this.mixinElements_ = elements);
        },
        isJustParentSelector: function () {
            return !this.mediaEmpty &&
                this.elements.length === 1 &&
                this.elements[0].value === '&' &&
                (this.elements[0].combinator.value === ' ' || this.elements[0].combinator.value === '');
        },
        eval: function (context) {
            var evaldCondition = this.condition && this.condition.eval(context);
            var elements = this.elements;
            var extendList = this.extendList;
            elements = elements && elements.map(function (e) { return e.eval(context); });
            extendList = extendList && extendList.map(function (extend) { return extend.eval(context); });
            return this.createDerived(elements, extendList, evaldCondition);
        },
        genCSS: function (context, output) {
            var i, element;
            if ((!context || !context.firstSelector) && this.elements[0].combinator.value === '') {
                output.add(' ', this.fileInfo(), this.getIndex());
            }
            for (i = 0; i < this.elements.length; i++) {
                element = this.elements[i];
                element.genCSS(context, output);
            }
        },
        getIsOutput: function () {
            return this.evaldCondition;
        }
    });
    exports.default = Selector;
},
/* less/lib/less/tree/element.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var paren_1 = tslib_1.__importDefault(require(118) /* ./paren */);
    var combinator_1 = tslib_1.__importDefault(require(119) /* ./combinator */);
    var Element = function (combinator, value, isVariable, index, currentFileInfo, visibilityInfo) {
        this.combinator = combinator instanceof combinator_1.default ?
            combinator : new combinator_1.default(combinator);
        if (typeof value === 'string') {
            this.value = value.trim();
        }
        else if (value) {
            this.value = value;
        }
        else {
            this.value = '';
        }
        this.isVariable = isVariable;
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.copyVisibilityInfo(visibilityInfo);
        this.setParent(this.combinator, this);
    };
    Element.prototype = Object.assign(new node_1.default(), {
        type: 'Element',
        accept: function (visitor) {
            var value = this.value;
            this.combinator = visitor.visit(this.combinator);
            if (typeof value === 'object') {
                this.value = visitor.visit(value);
            }
        },
        eval: function (context) {
            return new Element(this.combinator, this.value.eval ? this.value.eval(context) : this.value, this.isVariable, this.getIndex(), this.fileInfo(), this.visibilityInfo());
        },
        clone: function () {
            return new Element(this.combinator, this.value, this.isVariable, this.getIndex(), this.fileInfo(), this.visibilityInfo());
        },
        genCSS: function (context, output) {
            output.add(this.toCSS(context), this.fileInfo(), this.getIndex());
        },
        toCSS: function (context) {
            context = context || {};
            var value = this.value;
            var firstSelector = context.firstSelector;
            if (value instanceof paren_1.default) {
                // selector in parens should not be affected by outer selector
                // flags (breaks only interpolated selectors - see #1973)
                context.firstSelector = true;
            }
            value = value.toCSS ? value.toCSS(context) : value;
            context.firstSelector = firstSelector;
            if (value === '' && this.combinator.value.charAt(0) === '&') {
                return '';
            }
            else {
                return this.combinator.toCSS(context) + value;
            }
        }
    });
    exports.default = Element;
},
/* less/lib/less/tree/paren.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var Paren = function (node) {
        this.value = node;
    };
    Paren.prototype = Object.assign(new node_1.default(), {
        type: 'Paren',
        genCSS: function (context, output) {
            output.add('(');
            this.value.genCSS(context, output);
            output.add(')');
        },
        eval: function (context) {
            return new Paren(this.value.eval(context));
        }
    });
    exports.default = Paren;
},
/* less/lib/less/tree/combinator.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var _noSpaceCombinators = {
        '': true,
        ' ': true,
        '|': true
    };
    var Combinator = function (value) {
        if (value === ' ') {
            this.value = ' ';
            this.emptyOrWhitespace = true;
        }
        else {
            this.value = value ? value.trim() : '';
            this.emptyOrWhitespace = this.value === '';
        }
    };
    Combinator.prototype = Object.assign(new node_1.default(), {
        type: 'Combinator',
        genCSS: function (context, output) {
            var spaceOrEmpty = (context.compress || _noSpaceCombinators[this.value]) ? '' : ' ';
            output.add(spaceOrEmpty + this.value + spaceOrEmpty);
        }
    });
    exports.default = Combinator;
},
/* less/lib/less/less-error.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var utils = tslib_1.__importStar(require(121) /* ./utils */);
    var anonymousFunc = /(<anonymous>|Function):(\d+):(\d+)/;
    /**
     * This is a centralized class of any error that could be thrown internally (mostly by the parser).
     * Besides standard .message it keeps some additional data like a path to the file where the error
     * occurred along with line and column numbers.
     *
     * @class
     * @extends Error
     * @type {module.LessError}
     *
     * @prop {string} type
     * @prop {string} filename
     * @prop {number} index
     * @prop {number} line
     * @prop {number} column
     * @prop {number} callLine
     * @prop {number} callExtract
     * @prop {string[]} extract
     *
     * @param {Object} e              - An error object to wrap around or just a descriptive object
     * @param {Object} fileContentMap - An object with file contents in 'contents' property (like importManager) @todo - move to fileManager?
     * @param {string} [currentFilename]
     */
    var LessError = function (e, fileContentMap, currentFilename) {
        Error.call(this);
        var filename = e.filename || currentFilename;
        this.message = e.message;
        this.stack = e.stack;
        if (fileContentMap && filename) {
            var input = fileContentMap.contents[filename];
            var loc = utils.getLocation(e.index, input);
            var line = loc.line;
            var col = loc.column;
            var callLine = e.call && utils.getLocation(e.call, input).line;
            var lines = input ? input.split('\n') : '';
            this.type = e.type || 'Syntax';
            this.filename = filename;
            this.index = e.index;
            this.line = typeof line === 'number' ? line + 1 : null;
            this.column = col;
            if (!this.line && this.stack) {
                var found = this.stack.match(anonymousFunc);
                /**
                 * We have to figure out how this environment stringifies anonymous functions
                 * so we can correctly map plugin errors.
                 *
                 * Note, in Node 8, the output of anonymous funcs varied based on parameters
                 * being present or not, so we inject dummy params.
                 */
                var func = new Function('a', 'throw new Error()');
                var lineAdjust = 0;
                try {
                    func();
                }
                catch (e) {
                    var match = e.stack.match(anonymousFunc);
                    var line = parseInt(match[2]);
                    lineAdjust = 1 - line;
                }
                if (found) {
                    if (found[2]) {
                        this.line = parseInt(found[2]) + lineAdjust;
                    }
                    if (found[3]) {
                        this.column = parseInt(found[3]);
                    }
                }
            }
            this.callLine = callLine + 1;
            this.callExtract = lines[callLine];
            this.extract = [
                lines[this.line - 2],
                lines[this.line - 1],
                lines[this.line]
            ];
        }
    };
    if (typeof Object.create === 'undefined') {
        var F = function () { };
        F.prototype = Error.prototype;
        LessError.prototype = new F();
    }
    else {
        LessError.prototype = Object.create(Error.prototype);
    }
    LessError.prototype.constructor = LessError;
    /**
     * An overridden version of the default Object.prototype.toString
     * which uses additional information to create a helpful message.
     *
     * @param {Object} options
     * @returns {string}
     */
    LessError.prototype.toString = function (options) {
        options = options || {};
        var message = '';
        var extract = this.extract || [];
        var error = [];
        var stylize = function (str) { return str; };
        if (options.stylize) {
            var type = typeof options.stylize;
            if (type !== 'function') {
                throw Error("options.stylize should be a function, got a " + type + "!");
            }
            stylize = options.stylize;
        }
        if (this.line !== null) {
            if (typeof extract[0] === 'string') {
                error.push(stylize(this.line - 1 + " " + extract[0], 'grey'));
            }
            if (typeof extract[1] === 'string') {
                var errorTxt = this.line + " ";
                if (extract[1]) {
                    errorTxt += extract[1].slice(0, this.column) +
                        stylize(stylize(stylize(extract[1].substr(this.column, 1), 'bold') +
                            extract[1].slice(this.column + 1), 'red'), 'inverse');
                }
                error.push(errorTxt);
            }
            if (typeof extract[2] === 'string') {
                error.push(stylize(this.line + 1 + " " + extract[2], 'grey'));
            }
            error = error.join('\n') + stylize('', 'reset') + "\n";
        }
        message += stylize(this.type + "Error: " + this.message, 'red');
        if (this.filename) {
            message += stylize(' in ', 'red') + this.filename;
        }
        if (this.line) {
            message += stylize(" on line " + this.line + ", column " + (this.column + 1) + ":", 'grey');
        }
        message += "\n" + error;
        if (this.callLine) {
            message += stylize('from ', 'red') + (this.filename || '') + "/n";
            message += stylize(this.callLine, 'grey') + " " + this.callExtract + "/n";
        }
        return message;
    };
    exports.default = LessError;
},
/* less/lib/less/utils.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    /* jshint proto: true */
    var Constants = tslib_1.__importStar(require(122) /* ./constants */);
    var copy_anything_1 = require(123) /* copy-anything */;
    function getLocation(index, inputStream) {
        var n = index + 1;
        var line = null;
        var column = -1;
        while (--n >= 0 && inputStream.charAt(n) !== '\n') {
            column++;
        }
        if (typeof index === 'number') {
            line = (inputStream.slice(0, index).match(/\n/g) || '').length;
        }
        return {
            line: line,
            column: column
        };
    }
    exports.getLocation = getLocation;
    function copyArray(arr) {
        var i;
        var length = arr.length;
        var copy = new Array(length);
        for (i = 0; i < length; i++) {
            copy[i] = arr[i];
        }
        return copy;
    }
    exports.copyArray = copyArray;
    function clone(obj) {
        var cloned = {};
        for (var prop in obj) {
            if (obj.hasOwnProperty(prop)) {
                cloned[prop] = obj[prop];
            }
        }
        return cloned;
    }
    exports.clone = clone;
    function defaults(obj1, obj2) {
        var newObj = obj2 || {};
        if (!obj2._defaults) {
            newObj = {};
            var defaults_1 = copy_anything_1.copy(obj1);
            newObj._defaults = defaults_1;
            var cloned = obj2 ? copy_anything_1.copy(obj2) : {};
            Object.assign(newObj, defaults_1, cloned);
        }
        return newObj;
    }
    exports.defaults = defaults;
    function copyOptions(obj1, obj2) {
        if (obj2 && obj2._defaults) {
            return obj2;
        }
        var opts = defaults(obj1, obj2);
        if (opts.strictMath) {
            opts.math = Constants.Math.PARENS;
        }
        // Back compat with changed relativeUrls option
        if (opts.relativeUrls) {
            opts.rewriteUrls = Constants.RewriteUrls.ALL;
        }
        if (typeof opts.math === 'string') {
            switch (opts.math.toLowerCase()) {
                case 'always':
                    opts.math = Constants.Math.ALWAYS;
                    break;
                case 'parens-division':
                    opts.math = Constants.Math.PARENS_DIVISION;
                    break;
                case 'strict':
                case 'parens':
                    opts.math = Constants.Math.PARENS;
                    break;
                default:
                    opts.math = Constants.Math.PARENS;
            }
        }
        if (typeof opts.rewriteUrls === 'string') {
            switch (opts.rewriteUrls.toLowerCase()) {
                case 'off':
                    opts.rewriteUrls = Constants.RewriteUrls.OFF;
                    break;
                case 'local':
                    opts.rewriteUrls = Constants.RewriteUrls.LOCAL;
                    break;
                case 'all':
                    opts.rewriteUrls = Constants.RewriteUrls.ALL;
                    break;
            }
        }
        return opts;
    }
    exports.copyOptions = copyOptions;
    function merge(obj1, obj2) {
        for (var prop in obj2) {
            if (obj2.hasOwnProperty(prop)) {
                obj1[prop] = obj2[prop];
            }
        }
        return obj1;
    }
    exports.merge = merge;
    function flattenArray(arr, result) {
        if (result === void 0) {
            result = [];
        }
        for (var i = 0, length_1 = arr.length; i < length_1; i++) {
            var value = arr[i];
            if (Array.isArray(value)) {
                flattenArray(value, result);
            }
            else {
                if (value !== undefined) {
                    result.push(value);
                }
            }
        }
        return result;
    }
    exports.flattenArray = flattenArray;
},
/* less/lib/less/constants.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.Math = {
        ALWAYS: 0,
        PARENS_DIVISION: 1,
        PARENS: 2
        // removed - STRICT_LEGACY: 3
    };
    exports.RewriteUrls = {
        OFF: 0,
        LOCAL: 1,
        ALL: 2
    };
},
/* copy-anything/dist/index.cjs.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var isWhat = require(124) /* is-what */;
    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation.
    
    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.
    
    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */
    function __spreadArrays() {
        for (var s = 0, i = 0, il = arguments.length; i < il; i++)
            s += arguments[i].length;
        for (var r = Array(s), k = 0, i = 0; i < il; i++)
            for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
                r[k] = a[j];
        return r;
    }
    function assignProp(carry, key, newVal, originalObject, includeNonenumerable) {
        var propType = {}.propertyIsEnumerable.call(originalObject, key)
            ? 'enumerable'
            : 'nonenumerable';
        if (propType === 'enumerable')
            carry[key] = newVal;
        if (includeNonenumerable && propType === 'nonenumerable') {
            Object.defineProperty(carry, key, {
                value: newVal,
                enumerable: false,
                writable: true,
                configurable: true,
            });
        }
    }
    /**
     * Copy (clone) an object and all its props recursively to get rid of any prop referenced of the original object. Arrays are also cloned, however objects inside arrays are still linked.
     *
     * @export
     * @template T
     * @param {T} target Target can be anything
     * @param {Options} [options = {}] Options can be `props` or `nonenumerable`
     * @returns {T} the target with replaced values
     * @export
     */
    function copy(target, options) {
        if (options === void 0) {
            options = {};
        }
        if (isWhat.isArray(target))
            return target.map(function (item) { return copy(item, options); });
        if (!isWhat.isPlainObject(target))
            return target;
        var props = Object.getOwnPropertyNames(target);
        var symbols = Object.getOwnPropertySymbols(target);
        return __spreadArrays(props, symbols).reduce(function (carry, key) {
            if (isWhat.isArray(options.props) && !options.props.includes(key)) {
                return carry;
            }
            var val = target[key];
            var newVal = copy(val, options);
            assignProp(carry, key, newVal, target, options.nonenumerable);
            return carry;
        }, {});
    }
    exports.copy = copy;
},
/* is-what/dist/index.cjs.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    /**
     * Returns the object type of the given payload
     *
     * @param {*} payload
     * @returns {string}
     */
    function getType(payload) {
        return Object.prototype.toString.call(payload).slice(8, -1);
    }
    /**
     * Returns whether the payload is undefined
     *
     * @param {*} payload
     * @returns {payload is undefined}
     */
    function isUndefined(payload) {
        return getType(payload) === 'Undefined';
    }
    /**
     * Returns whether the payload is null
     *
     * @param {*} payload
     * @returns {payload is null}
     */
    function isNull(payload) {
        return getType(payload) === 'Null';
    }
    /**
     * Returns whether the payload is a plain JavaScript object (excluding special classes or objects with other prototypes)
     *
     * @param {*} payload
     * @returns {payload is PlainObject}
     */
    function isPlainObject(payload) {
        if (getType(payload) !== 'Object')
            return false;
        return payload.constructor === Object && Object.getPrototypeOf(payload) === Object.prototype;
    }
    /**
     * Returns whether the payload is a plain JavaScript object (excluding special classes or objects with other prototypes)
     *
     * @param {*} payload
     * @returns {payload is PlainObject}
     */
    function isObject(payload) {
        return isPlainObject(payload);
    }
    /**
     * Returns whether the payload is a an empty object (excluding special classes or objects with other prototypes)
     *
     * @param {*} payload
     * @returns {payload is { [K in any]: never }}
     */
    function isEmptyObject(payload) {
        return isPlainObject(payload) && Object.keys(payload).length === 0;
    }
    /**
     * Returns whether the payload is a an empty object (excluding special classes or objects with other prototypes)
     *
     * @param {*} payload
     * @returns {payload is PlainObject}
     */
    function isFullObject(payload) {
        return isPlainObject(payload) && Object.keys(payload).length > 0;
    }
    /**
     * Returns whether the payload is an any kind of object (including special classes or objects with different prototypes)
     *
     * @param {*} payload
     * @returns {payload is PlainObject}
     */
    function isAnyObject(payload) {
        return getType(payload) === 'Object';
    }
    /**
     * Returns whether the payload is an object like a type passed in < >
     *
     * Usage: isObjectLike<{id: any}>(payload) // will make sure it's an object and has an `id` prop.
     *
     * @template T this must be passed in < >
     * @param {*} payload
     * @returns {payload is T}
     */
    function isObjectLike(payload) {
        return isAnyObject(payload);
    }
    /**
     * Returns whether the payload is a function (regular or async)
     *
     * @param {*} payload
     * @returns {payload is AnyFunction}
     */
    function isFunction(payload) {
        return typeof payload === 'function';
    }
    /**
     * Returns whether the payload is an array
     *
     * @param {any} payload
     * @returns {payload is any[]}
     */
    function isArray(payload) {
        return getType(payload) === 'Array';
    }
    /**
     * Returns whether the payload is a an array with at least 1 item
     *
     * @param {*} payload
     * @returns {payload is any[]}
     */
    function isFullArray(payload) {
        return isArray(payload) && payload.length > 0;
    }
    /**
     * Returns whether the payload is a an empty array
     *
     * @param {*} payload
     * @returns {payload is []}
     */
    function isEmptyArray(payload) {
        return isArray(payload) && payload.length === 0;
    }
    /**
     * Returns whether the payload is a string
     *
     * @param {*} payload
     * @returns {payload is string}
     */
    function isString(payload) {
        return getType(payload) === 'String';
    }
    /**
     * Returns whether the payload is a string, BUT returns false for ''
     *
     * @param {*} payload
     * @returns {payload is string}
     */
    function isFullString(payload) {
        return isString(payload) && payload !== '';
    }
    /**
     * Returns whether the payload is ''
     *
     * @param {*} payload
     * @returns {payload is string}
     */
    function isEmptyString(payload) {
        return payload === '';
    }
    /**
     * Returns whether the payload is a number (but not NaN)
     *
     * This will return `false` for `NaN`!!
     *
     * @param {*} payload
     * @returns {payload is number}
     */
    function isNumber(payload) {
        return getType(payload) === 'Number' && !isNaN(payload);
    }
    /**
     * Returns whether the payload is a boolean
     *
     * @param {*} payload
     * @returns {payload is boolean}
     */
    function isBoolean(payload) {
        return getType(payload) === 'Boolean';
    }
    /**
     * Returns whether the payload is a regular expression (RegExp)
     *
     * @param {*} payload
     * @returns {payload is RegExp}
     */
    function isRegExp(payload) {
        return getType(payload) === 'RegExp';
    }
    /**
     * Returns whether the payload is a Map
     *
     * @param {*} payload
     * @returns {payload is Map<any, any>}
     */
    function isMap(payload) {
        return getType(payload) === 'Map';
    }
    /**
     * Returns whether the payload is a WeakMap
     *
     * @param {*} payload
     * @returns {payload is WeakMap<any, any>}
     */
    function isWeakMap(payload) {
        return getType(payload) === 'WeakMap';
    }
    /**
     * Returns whether the payload is a Set
     *
     * @param {*} payload
     * @returns {payload is Set<any>}
     */
    function isSet(payload) {
        return getType(payload) === 'Set';
    }
    /**
     * Returns whether the payload is a WeakSet
     *
     * @param {*} payload
     * @returns {payload is WeakSet<any>}
     */
    function isWeakSet(payload) {
        return getType(payload) === 'WeakSet';
    }
    /**
     * Returns whether the payload is a Symbol
     *
     * @param {*} payload
     * @returns {payload is symbol}
     */
    function isSymbol(payload) {
        return getType(payload) === 'Symbol';
    }
    /**
     * Returns whether the payload is a Date, and that the date is valid
     *
     * @param {*} payload
     * @returns {payload is Date}
     */
    function isDate(payload) {
        return getType(payload) === 'Date' && !isNaN(payload);
    }
    /**
     * Returns whether the payload is a Blob
     *
     * @param {*} payload
     * @returns {payload is Blob}
     */
    function isBlob(payload) {
        return getType(payload) === 'Blob';
    }
    /**
     * Returns whether the payload is a File
     *
     * @param {*} payload
     * @returns {payload is File}
     */
    function isFile(payload) {
        return getType(payload) === 'File';
    }
    /**
     * Returns whether the payload is a Promise
     *
     * @param {*} payload
     * @returns {payload is Promise<any>}
     */
    function isPromise(payload) {
        return getType(payload) === 'Promise';
    }
    /**
     * Returns whether the payload is an Error
     *
     * @param {*} payload
     * @returns {payload is Error}
     */
    function isError(payload) {
        return getType(payload) === 'Error';
    }
    /**
     * Returns whether the payload is literally the value `NaN` (it's `NaN` and also a `number`)
     *
     * @param {*} payload
     * @returns {payload is typeof NaN}
     */
    function isNaNValue(payload) {
        return getType(payload) === 'Number' && isNaN(payload);
    }
    /**
     * Returns whether the payload is a primitive type (eg. Boolean | Null | Undefined | Number | String | Symbol)
     *
     * @param {*} payload
     * @returns {(payload is boolean | null | undefined | number | string | symbol)}
     */
    function isPrimitive(payload) {
        return (isBoolean(payload) ||
            isNull(payload) ||
            isUndefined(payload) ||
            isNumber(payload) ||
            isString(payload) ||
            isSymbol(payload));
    }
    /**
     * Returns true whether the payload is null or undefined
     *
     * @param {*} payload
     * @returns {(payload is null | undefined)}
     */
    var isNullOrUndefined = isOneOf(isNull, isUndefined);
    function isOneOf(a, b, c, d, e) {
        return function (value) {
            return a(value) || b(value) || (!!c && c(value)) || (!!d && d(value)) || (!!e && e(value));
        };
    }
    /**
     * Does a generic check to check that the given payload is of a given type.
     * In cases like Number, it will return true for NaN as NaN is a Number (thanks javascript!);
     * It will, however, differentiate between object and null
     *
     * @template T
     * @param {*} payload
     * @param {T} type
     * @throws {TypeError} Will throw type error if type is an invalid type
     * @returns {payload is T}
     */
    function isType(payload, type) {
        if (!(type instanceof Function)) {
            throw new TypeError('Type must be a function');
        }
        if (!Object.prototype.hasOwnProperty.call(type, 'prototype')) {
            throw new TypeError('Type is not a class');
        }
        // Classes usually have names (as functions usually have names)
        var name = type.name;
        return getType(payload) === name || Boolean(payload && payload.constructor === type);
    }
    exports.getType = getType;
    exports.isAnyObject = isAnyObject;
    exports.isArray = isArray;
    exports.isBlob = isBlob;
    exports.isBoolean = isBoolean;
    exports.isDate = isDate;
    exports.isEmptyArray = isEmptyArray;
    exports.isEmptyObject = isEmptyObject;
    exports.isEmptyString = isEmptyString;
    exports.isError = isError;
    exports.isFile = isFile;
    exports.isFullArray = isFullArray;
    exports.isFullObject = isFullObject;
    exports.isFullString = isFullString;
    exports.isFunction = isFunction;
    exports.isMap = isMap;
    exports.isNaNValue = isNaNValue;
    exports.isNull = isNull;
    exports.isNullOrUndefined = isNullOrUndefined;
    exports.isNumber = isNumber;
    exports.isObject = isObject;
    exports.isObjectLike = isObjectLike;
    exports.isOneOf = isOneOf;
    exports.isPlainObject = isPlainObject;
    exports.isPrimitive = isPrimitive;
    exports.isPromise = isPromise;
    exports.isRegExp = isRegExp;
    exports.isSet = isSet;
    exports.isString = isString;
    exports.isSymbol = isSymbol;
    exports.isType = isType;
    exports.isUndefined = isUndefined;
    exports.isWeakMap = isWeakMap;
    exports.isWeakSet = isWeakSet;
},
/* less/lib/less/tree/ruleset.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var declaration_1 = tslib_1.__importDefault(require(126) /* ./declaration */);
    var keyword_1 = tslib_1.__importDefault(require(128) /* ./keyword */);
    var comment_1 = tslib_1.__importDefault(require(130) /* ./comment */);
    var paren_1 = tslib_1.__importDefault(require(118) /* ./paren */);
    var selector_1 = tslib_1.__importDefault(require(116) /* ./selector */);
    var element_1 = tslib_1.__importDefault(require(117) /* ./element */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ./anonymous */);
    var contexts_1 = tslib_1.__importDefault(require(132) /* ../contexts */);
    var function_registry_1 = tslib_1.__importDefault(require(133) /* ../functions/function-registry */);
    var default_1 = tslib_1.__importDefault(require(134) /* ../functions/default */);
    var debug_info_1 = tslib_1.__importDefault(require(131) /* ./debug-info */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    var Ruleset = function (selectors, rules, strictImports, visibilityInfo) {
        this.selectors = selectors;
        this.rules = rules;
        this._lookups = {};
        this._variables = null;
        this._properties = null;
        this.strictImports = strictImports;
        this.copyVisibilityInfo(visibilityInfo);
        this.allowRoot = true;
        this.setParent(this.selectors, this);
        this.setParent(this.rules, this);
    };
    Ruleset.prototype = Object.assign(new node_1.default(), {
        type: 'Ruleset',
        isRuleset: true,
        isRulesetLike: function () { return true; },
        accept: function (visitor) {
            if (this.paths) {
                this.paths = visitor.visitArray(this.paths, true);
            }
            else if (this.selectors) {
                this.selectors = visitor.visitArray(this.selectors);
            }
            if (this.rules && this.rules.length) {
                this.rules = visitor.visitArray(this.rules);
            }
        },
        eval: function (context) {
            var that = this;
            var selectors;
            var selCnt;
            var selector;
            var i;
            var hasVariable;
            var hasOnePassingSelector = false;
            if (this.selectors && (selCnt = this.selectors.length)) {
                selectors = new Array(selCnt);
                default_1.default.error({
                    type: 'Syntax',
                    message: 'it is currently only allowed in parametric mixin guards,'
                });
                for (i = 0; i < selCnt; i++) {
                    selector = this.selectors[i].eval(context);
                    for (var j = 0; j < selector.elements.length; j++) {
                        if (selector.elements[j].isVariable) {
                            hasVariable = true;
                            break;
                        }
                    }
                    selectors[i] = selector;
                    if (selector.evaldCondition) {
                        hasOnePassingSelector = true;
                    }
                }
                if (hasVariable) {
                    var toParseSelectors = new Array(selCnt);
                    for (i = 0; i < selCnt; i++) {
                        selector = selectors[i];
                        toParseSelectors[i] = selector.toCSS(context);
                    }
                    this.parse.parseNode(toParseSelectors.join(','), ["selectors"], selectors[0].getIndex(), selectors[0].fileInfo(), function (err, result) {
                        if (result) {
                            selectors = utils.flattenArray(result);
                        }
                    });
                }
                default_1.default.reset();
            }
            else {
                hasOnePassingSelector = true;
            }
            var rules = this.rules ? utils.copyArray(this.rules) : null;
            var ruleset = new Ruleset(selectors, rules, this.strictImports, this.visibilityInfo());
            var rule;
            var subRule;
            ruleset.originalRuleset = this;
            ruleset.root = this.root;
            ruleset.firstRoot = this.firstRoot;
            ruleset.allowImports = this.allowImports;
            if (this.debugInfo) {
                ruleset.debugInfo = this.debugInfo;
            }
            if (!hasOnePassingSelector) {
                rules.length = 0;
            }
            // inherit a function registry from the frames stack when possible;
            // otherwise from the global registry
            ruleset.functionRegistry = (function (frames) {
                var i = 0;
                var n = frames.length;
                var found;
                for (; i !== n; ++i) {
                    found = frames[i].functionRegistry;
                    if (found) {
                        return found;
                    }
                }
                return function_registry_1.default;
            }(context.frames)).inherit();
            // push the current ruleset to the frames stack
            var ctxFrames = context.frames;
            ctxFrames.unshift(ruleset);
            // currrent selectors
            var ctxSelectors = context.selectors;
            if (!ctxSelectors) {
                context.selectors = ctxSelectors = [];
            }
            ctxSelectors.unshift(this.selectors);
            // Evaluate imports
            if (ruleset.root || ruleset.allowImports || !ruleset.strictImports) {
                ruleset.evalImports(context);
            }
            // Store the frames around mixin definitions,
            // so they can be evaluated like closures when the time comes.
            var rsRules = ruleset.rules;
            for (i = 0; (rule = rsRules[i]); i++) {
                if (rule.evalFirst) {
                    rsRules[i] = rule.eval(context);
                }
            }
            var mediaBlockCount = (context.mediaBlocks && context.mediaBlocks.length) || 0;
            // Evaluate mixin calls.
            for (i = 0; (rule = rsRules[i]); i++) {
                if (rule.type === 'MixinCall') {
                    /* jshint loopfunc:true */
                    rules = rule.eval(context).filter(function (r) {
                        if ((r instanceof declaration_1.default) && r.variable) {
                            // do not pollute the scope if the variable is
                            // already there. consider returning false here
                            // but we need a way to "return" variable from mixins
                            return !(ruleset.variable(r.name));
                        }
                        return true;
                    });
                    rsRules.splice.apply(rsRules, [i, 1].concat(rules));
                    i += rules.length - 1;
                    ruleset.resetCache();
                }
                else if (rule.type === 'VariableCall') {
                    /* jshint loopfunc:true */
                    rules = rule.eval(context).rules.filter(function (r) {
                        if ((r instanceof declaration_1.default) && r.variable) {
                            // do not pollute the scope at all
                            return false;
                        }
                        return true;
                    });
                    rsRules.splice.apply(rsRules, [i, 1].concat(rules));
                    i += rules.length - 1;
                    ruleset.resetCache();
                }
            }
            // Evaluate everything else
            for (i = 0; (rule = rsRules[i]); i++) {
                if (!rule.evalFirst) {
                    rsRules[i] = rule = rule.eval ? rule.eval(context) : rule;
                }
            }
            // Evaluate everything else
            for (i = 0; (rule = rsRules[i]); i++) {
                // for rulesets, check if it is a css guard and can be removed
                if (rule instanceof Ruleset && rule.selectors && rule.selectors.length === 1) {
                    // check if it can be folded in (e.g. & where)
                    if (rule.selectors[0] && rule.selectors[0].isJustParentSelector()) {
                        rsRules.splice(i--, 1);
                        for (var j = 0; (subRule = rule.rules[j]); j++) {
                            if (subRule instanceof node_1.default) {
                                subRule.copyVisibilityInfo(rule.visibilityInfo());
                                if (!(subRule instanceof declaration_1.default) || !subRule.variable) {
                                    rsRules.splice(++i, 0, subRule);
                                }
                            }
                        }
                    }
                }
            }
            // Pop the stack
            ctxFrames.shift();
            ctxSelectors.shift();
            if (context.mediaBlocks) {
                for (i = mediaBlockCount; i < context.mediaBlocks.length; i++) {
                    context.mediaBlocks[i].bubbleSelectors(selectors);
                }
            }
            return ruleset;
        },
        evalImports: function (context) {
            var rules = this.rules;
            var i;
            var importRules;
            if (!rules) {
                return;
            }
            for (i = 0; i < rules.length; i++) {
                if (rules[i].type === 'Import') {
                    importRules = rules[i].eval(context);
                    if (importRules && (importRules.length || importRules.length === 0)) {
                        rules.splice.apply(rules, [i, 1].concat(importRules));
                        i += importRules.length - 1;
                    }
                    else {
                        rules.splice(i, 1, importRules);
                    }
                    this.resetCache();
                }
            }
        },
        makeImportant: function () {
            var result = new Ruleset(this.selectors, this.rules.map(function (r) {
                if (r.makeImportant) {
                    return r.makeImportant();
                }
                else {
                    return r;
                }
            }), this.strictImports, this.visibilityInfo());
            return result;
        },
        matchArgs: function (args) {
            return !args || args.length === 0;
        },
        // lets you call a css selector with a guard
        matchCondition: function (args, context) {
            var lastSelector = this.selectors[this.selectors.length - 1];
            if (!lastSelector.evaldCondition) {
                return false;
            }
            if (lastSelector.condition &&
                !lastSelector.condition.eval(new contexts_1.default.Eval(context, context.frames))) {
                return false;
            }
            return true;
        },
        resetCache: function () {
            this._rulesets = null;
            this._variables = null;
            this._properties = null;
            this._lookups = {};
        },
        variables: function () {
            if (!this._variables) {
                this._variables = !this.rules ? {} : this.rules.reduce(function (hash, r) {
                    if (r instanceof declaration_1.default && r.variable === true) {
                        hash[r.name] = r;
                    }
                    // when evaluating variables in an import statement, imports have not been eval'd
                    // so we need to go inside import statements.
                    // guard against root being a string (in the case of inlined less)
                    if (r.type === 'Import' && r.root && r.root.variables) {
                        var vars = r.root.variables();
                        for (var name_1 in vars) {
                            if (vars.hasOwnProperty(name_1)) {
                                hash[name_1] = r.root.variable(name_1);
                            }
                        }
                    }
                    return hash;
                }, {});
            }
            return this._variables;
        },
        properties: function () {
            if (!this._properties) {
                this._properties = !this.rules ? {} : this.rules.reduce(function (hash, r) {
                    if (r instanceof declaration_1.default && r.variable !== true) {
                        var name_2 = (r.name.length === 1) && (r.name[0] instanceof keyword_1.default) ?
                            r.name[0].value : r.name;
                        // Properties don't overwrite as they can merge
                        if (!hash["$" + name_2]) {
                            hash["$" + name_2] = [r];
                        }
                        else {
                            hash["$" + name_2].push(r);
                        }
                    }
                    return hash;
                }, {});
            }
            return this._properties;
        },
        variable: function (name) {
            var decl = this.variables()[name];
            if (decl) {
                return this.parseValue(decl);
            }
        },
        property: function (name) {
            var decl = this.properties()[name];
            if (decl) {
                return this.parseValue(decl);
            }
        },
        lastDeclaration: function () {
            for (var i = this.rules.length; i > 0; i--) {
                var decl = this.rules[i - 1];
                if (decl instanceof declaration_1.default) {
                    return this.parseValue(decl);
                }
            }
        },
        parseValue: function (toParse) {
            var self = this;
            function transformDeclaration(decl) {
                if (decl.value instanceof anonymous_1.default && !decl.parsed) {
                    if (typeof decl.value.value === 'string') {
                        this.parse.parseNode(decl.value.value, ['value', 'important'], decl.value.getIndex(), decl.fileInfo(), function (err, result) {
                            if (err) {
                                decl.parsed = true;
                            }
                            if (result) {
                                decl.value = result[0];
                                decl.important = result[1] || '';
                                decl.parsed = true;
                            }
                        });
                    }
                    else {
                        decl.parsed = true;
                    }
                    return decl;
                }
                else {
                    return decl;
                }
            }
            if (!Array.isArray(toParse)) {
                return transformDeclaration.call(self, toParse);
            }
            else {
                var nodes_1 = [];
                toParse.forEach(function (n) {
                    nodes_1.push(transformDeclaration.call(self, n));
                });
                return nodes_1;
            }
        },
        rulesets: function () {
            if (!this.rules) {
                return [];
            }
            var filtRules = [];
            var rules = this.rules;
            var i;
            var rule;
            for (i = 0; (rule = rules[i]); i++) {
                if (rule.isRuleset) {
                    filtRules.push(rule);
                }
            }
            return filtRules;
        },
        prependRule: function (rule) {
            var rules = this.rules;
            if (rules) {
                rules.unshift(rule);
            }
            else {
                this.rules = [rule];
            }
            this.setParent(rule, this);
        },
        find: function (selector, self, filter) {
            self = self || this;
            var rules = [];
            var match;
            var foundMixins;
            var key = selector.toCSS();
            if (key in this._lookups) {
                return this._lookups[key];
            }
            this.rulesets().forEach(function (rule) {
                if (rule !== self) {
                    for (var j = 0; j < rule.selectors.length; j++) {
                        match = selector.match(rule.selectors[j]);
                        if (match) {
                            if (selector.elements.length > match) {
                                if (!filter || filter(rule)) {
                                    foundMixins = rule.find(new selector_1.default(selector.elements.slice(match)), self, filter);
                                    for (var i = 0; i < foundMixins.length; ++i) {
                                        foundMixins[i].path.push(rule);
                                    }
                                    Array.prototype.push.apply(rules, foundMixins);
                                }
                            }
                            else {
                                rules.push({ rule: rule, path: [] });
                            }
                            break;
                        }
                    }
                }
            });
            this._lookups[key] = rules;
            return rules;
        },
        genCSS: function (context, output) {
            var i;
            var j;
            var charsetRuleNodes = [];
            var ruleNodes = [];
            var // Line number debugging
            debugInfo;
            var rule;
            var path;
            context.tabLevel = (context.tabLevel || 0);
            if (!this.root) {
                context.tabLevel++;
            }
            var tabRuleStr = context.compress ? '' : Array(context.tabLevel + 1).join('  ');
            var tabSetStr = context.compress ? '' : Array(context.tabLevel).join('  ');
            var sep;
            var charsetNodeIndex = 0;
            var importNodeIndex = 0;
            for (i = 0; (rule = this.rules[i]); i++) {
                if (rule instanceof comment_1.default) {
                    if (importNodeIndex === i) {
                        importNodeIndex++;
                    }
                    ruleNodes.push(rule);
                }
                else if (rule.isCharset && rule.isCharset()) {
                    ruleNodes.splice(charsetNodeIndex, 0, rule);
                    charsetNodeIndex++;
                    importNodeIndex++;
                }
                else if (rule.type === 'Import') {
                    ruleNodes.splice(importNodeIndex, 0, rule);
                    importNodeIndex++;
                }
                else {
                    ruleNodes.push(rule);
                }
            }
            ruleNodes = charsetRuleNodes.concat(ruleNodes);
            // If this is the root node, we don't render
            // a selector, or {}.
            if (!this.root) {
                debugInfo = debug_info_1.default(context, this, tabSetStr);
                if (debugInfo) {
                    output.add(debugInfo);
                    output.add(tabSetStr);
                }
                var paths = this.paths;
                var pathCnt = paths.length;
                var pathSubCnt = void 0;
                sep = context.compress ? ',' : (",\n" + tabSetStr);
                for (i = 0; i < pathCnt; i++) {
                    path = paths[i];
                    if (!(pathSubCnt = path.length)) {
                        continue;
                    }
                    if (i > 0) {
                        output.add(sep);
                    }
                    context.firstSelector = true;
                    path[0].genCSS(context, output);
                    context.firstSelector = false;
                    for (j = 1; j < pathSubCnt; j++) {
                        path[j].genCSS(context, output);
                    }
                }
                output.add((context.compress ? '{' : ' {\n') + tabRuleStr);
            }
            // Compile rules and rulesets
            for (i = 0; (rule = ruleNodes[i]); i++) {
                if (i + 1 === ruleNodes.length) {
                    context.lastRule = true;
                }
                var currentLastRule = context.lastRule;
                if (rule.isRulesetLike(rule)) {
                    context.lastRule = false;
                }
                if (rule.genCSS) {
                    rule.genCSS(context, output);
                }
                else if (rule.value) {
                    output.add(rule.value.toString());
                }
                context.lastRule = currentLastRule;
                if (!context.lastRule && rule.isVisible()) {
                    output.add(context.compress ? '' : ("\n" + tabRuleStr));
                }
                else {
                    context.lastRule = false;
                }
            }
            if (!this.root) {
                output.add((context.compress ? '}' : "\n" + tabSetStr + "}"));
                context.tabLevel--;
            }
            if (!output.isEmpty() && !context.compress && this.firstRoot) {
                output.add('\n');
            }
        },
        joinSelectors: function (paths, context, selectors) {
            for (var s = 0; s < selectors.length; s++) {
                this.joinSelector(paths, context, selectors[s]);
            }
        },
        joinSelector: function (paths, context, selector) {
            function createParenthesis(elementsToPak, originalElement) {
                var replacementParen, j;
                if (elementsToPak.length === 0) {
                    replacementParen = new paren_1.default(elementsToPak[0]);
                }
                else {
                    var insideParent = new Array(elementsToPak.length);
                    for (j = 0; j < elementsToPak.length; j++) {
                        insideParent[j] = new element_1.default(null, elementsToPak[j], originalElement.isVariable, originalElement._index, originalElement._fileInfo);
                    }
                    replacementParen = new paren_1.default(new selector_1.default(insideParent));
                }
                return replacementParen;
            }
            function createSelector(containedElement, originalElement) {
                var element, selector;
                element = new element_1.default(null, containedElement, originalElement.isVariable, originalElement._index, originalElement._fileInfo);
                selector = new selector_1.default([element]);
                return selector;
            }
            // joins selector path from `beginningPath` with selector path in `addPath`
            // `replacedElement` contains element that is being replaced by `addPath`
            // returns concatenated path
            function addReplacementIntoPath(beginningPath, addPath, replacedElement, originalSelector) {
                var newSelectorPath, lastSelector, newJoinedSelector;
                // our new selector path
                newSelectorPath = [];
                // construct the joined selector - if & is the first thing this will be empty,
                // if not newJoinedSelector will be the last set of elements in the selector
                if (beginningPath.length > 0) {
                    newSelectorPath = utils.copyArray(beginningPath);
                    lastSelector = newSelectorPath.pop();
                    newJoinedSelector = originalSelector.createDerived(utils.copyArray(lastSelector.elements));
                }
                else {
                    newJoinedSelector = originalSelector.createDerived([]);
                }
                if (addPath.length > 0) {
                    // /deep/ is a CSS4 selector - (removed, so should deprecate)
                    // that is valid without anything in front of it
                    // so if the & does not have a combinator that is "" or " " then
                    // and there is a combinator on the parent, then grab that.
                    // this also allows + a { & .b { .a & { ... though not sure why you would want to do that
                    var combinator = replacedElement.combinator;
                    var parentEl = addPath[0].elements[0];
                    if (combinator.emptyOrWhitespace && !parentEl.combinator.emptyOrWhitespace) {
                        combinator = parentEl.combinator;
                    }
                    // join the elements so far with the first part of the parent
                    newJoinedSelector.elements.push(new element_1.default(combinator, parentEl.value, replacedElement.isVariable, replacedElement._index, replacedElement._fileInfo));
                    newJoinedSelector.elements = newJoinedSelector.elements.concat(addPath[0].elements.slice(1));
                }
                // now add the joined selector - but only if it is not empty
                if (newJoinedSelector.elements.length !== 0) {
                    newSelectorPath.push(newJoinedSelector);
                }
                // put together the parent selectors after the join (e.g. the rest of the parent)
                if (addPath.length > 1) {
                    var restOfPath = addPath.slice(1);
                    restOfPath = restOfPath.map(function (selector) {
                        return selector.createDerived(selector.elements, []);
                    });
                    newSelectorPath = newSelectorPath.concat(restOfPath);
                }
                return newSelectorPath;
            }
            // joins selector path from `beginningPath` with every selector path in `addPaths` array
            // `replacedElement` contains element that is being replaced by `addPath`
            // returns array with all concatenated paths
            function addAllReplacementsIntoPath(beginningPath, addPaths, replacedElement, originalSelector, result) {
                var j;
                for (j = 0; j < beginningPath.length; j++) {
                    var newSelectorPath = addReplacementIntoPath(beginningPath[j], addPaths, replacedElement, originalSelector);
                    result.push(newSelectorPath);
                }
                return result;
            }
            function mergeElementsOnToSelectors(elements, selectors) {
                var i, sel;
                if (elements.length === 0) {
                    return;
                }
                if (selectors.length === 0) {
                    selectors.push([new selector_1.default(elements)]);
                    return;
                }
                for (i = 0; (sel = selectors[i]); i++) {
                    // if the previous thing in sel is a parent this needs to join on to it
                    if (sel.length > 0) {
                        sel[sel.length - 1] = sel[sel.length - 1].createDerived(sel[sel.length - 1].elements.concat(elements));
                    }
                    else {
                        sel.push(new selector_1.default(elements));
                    }
                }
            }
            // replace all parent selectors inside `inSelector` by content of `context` array
            // resulting selectors are returned inside `paths` array
            // returns true if `inSelector` contained at least one parent selector
            function replaceParentSelector(paths, context, inSelector) {
                // The paths are [[Selector]]
                // The first list is a list of comma separated selectors
                // The inner list is a list of inheritance separated selectors
                // e.g.
                // .a, .b {
                //   .c {
                //   }
                // }
                // == [[.a] [.c]] [[.b] [.c]]
                //
                var i, j, k, currentElements, newSelectors, selectorsMultiplied, sel, el, hadParentSelector = false, length, lastSelector;
                function findNestedSelector(element) {
                    var maybeSelector;
                    if (!(element.value instanceof paren_1.default)) {
                        return null;
                    }
                    maybeSelector = element.value.value;
                    if (!(maybeSelector instanceof selector_1.default)) {
                        return null;
                    }
                    return maybeSelector;
                }
                // the elements from the current selector so far
                currentElements = [];
                // the current list of new selectors to add to the path.
                // We will build it up. We initiate it with one empty selector as we "multiply" the new selectors
                // by the parents
                newSelectors = [
                    []
                ];
                for (i = 0; (el = inSelector.elements[i]); i++) {
                    // non parent reference elements just get added
                    if (el.value !== '&') {
                        var nestedSelector = findNestedSelector(el);
                        if (nestedSelector != null) {
                            // merge the current list of non parent selector elements
                            // on to the current list of selectors to add
                            mergeElementsOnToSelectors(currentElements, newSelectors);
                            var nestedPaths = [];
                            var replaced = void 0;
                            var replacedNewSelectors = [];
                            replaced = replaceParentSelector(nestedPaths, context, nestedSelector);
                            hadParentSelector = hadParentSelector || replaced;
                            // the nestedPaths array should have only one member - replaceParentSelector does not multiply selectors
                            for (k = 0; k < nestedPaths.length; k++) {
                                var replacementSelector = createSelector(createParenthesis(nestedPaths[k], el), el);
                                addAllReplacementsIntoPath(newSelectors, [replacementSelector], el, inSelector, replacedNewSelectors);
                            }
                            newSelectors = replacedNewSelectors;
                            currentElements = [];
                        }
                        else {
                            currentElements.push(el);
                        }
                    }
                    else {
                        hadParentSelector = true;
                        // the new list of selectors to add
                        selectorsMultiplied = [];
                        // merge the current list of non parent selector elements
                        // on to the current list of selectors to add
                        mergeElementsOnToSelectors(currentElements, newSelectors);
                        // loop through our current selectors
                        for (j = 0; j < newSelectors.length; j++) {
                            sel = newSelectors[j];
                            // if we don't have any parent paths, the & might be in a mixin so that it can be used
                            // whether there are parents or not
                            if (context.length === 0) {
                                // the combinator used on el should now be applied to the next element instead so that
                                // it is not lost
                                if (sel.length > 0) {
                                    sel[0].elements.push(new element_1.default(el.combinator, '', el.isVariable, el._index, el._fileInfo));
                                }
                                selectorsMultiplied.push(sel);
                            }
                            else {
                                // and the parent selectors
                                for (k = 0; k < context.length; k++) {
                                    // We need to put the current selectors
                                    // then join the last selector's elements on to the parents selectors
                                    var newSelectorPath = addReplacementIntoPath(sel, context[k], el, inSelector);
                                    // add that to our new set of selectors
                                    selectorsMultiplied.push(newSelectorPath);
                                }
                            }
                        }
                        // our new selectors has been multiplied, so reset the state
                        newSelectors = selectorsMultiplied;
                        currentElements = [];
                    }
                }
                // if we have any elements left over (e.g. .a& .b == .b)
                // add them on to all the current selectors
                mergeElementsOnToSelectors(currentElements, newSelectors);
                for (i = 0; i < newSelectors.length; i++) {
                    length = newSelectors[i].length;
                    if (length > 0) {
                        paths.push(newSelectors[i]);
                        lastSelector = newSelectors[i][length - 1];
                        newSelectors[i][length - 1] = lastSelector.createDerived(lastSelector.elements, inSelector.extendList);
                    }
                }
                return hadParentSelector;
            }
            function deriveSelector(visibilityInfo, deriveFrom) {
                var newSelector = deriveFrom.createDerived(deriveFrom.elements, deriveFrom.extendList, deriveFrom.evaldCondition);
                newSelector.copyVisibilityInfo(visibilityInfo);
                return newSelector;
            }
            // joinSelector code follows
            var i, newPaths, hadParentSelector;
            newPaths = [];
            hadParentSelector = replaceParentSelector(newPaths, context, selector);
            if (!hadParentSelector) {
                if (context.length > 0) {
                    newPaths = [];
                    for (i = 0; i < context.length; i++) {
                        var concatenated = context[i].map(deriveSelector.bind(this, selector.visibilityInfo()));
                        concatenated.push(selector);
                        newPaths.push(concatenated);
                    }
                }
                else {
                    newPaths = [[selector]];
                }
            }
            for (i = 0; i < newPaths.length; i++) {
                paths.push(newPaths[i]);
            }
        }
    });
    exports.default = Ruleset;
},
/* less/lib/less/tree/declaration.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var value_1 = tslib_1.__importDefault(require(127) /* ./value */);
    var keyword_1 = tslib_1.__importDefault(require(128) /* ./keyword */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ./anonymous */);
    var Constants = tslib_1.__importStar(require(122) /* ../constants */);
    var MATH = Constants.Math;
    function evalName(context, name) {
        var value = '';
        var i;
        var n = name.length;
        var output = { add: function (s) { value += s; } };
        for (i = 0; i < n; i++) {
            name[i].eval(context).genCSS(context, output);
        }
        return value;
    }
    var Declaration = function (name, value, important, merge, index, currentFileInfo, inline, variable) {
        this.name = name;
        this.value = (value instanceof node_1.default) ? value : new value_1.default([value ? new anonymous_1.default(value) : null]);
        this.important = important ? " " + important.trim() : '';
        this.merge = merge;
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.inline = inline || false;
        this.variable = (variable !== undefined) ? variable
            : (name.charAt && (name.charAt(0) === '@'));
        this.allowRoot = true;
        this.setParent(this.value, this);
    };
    Declaration.prototype = Object.assign(new node_1.default(), {
        type: 'Declaration',
        genCSS: function (context, output) {
            output.add(this.name + (context.compress ? ':' : ': '), this.fileInfo(), this.getIndex());
            try {
                this.value.genCSS(context, output);
            }
            catch (e) {
                e.index = this._index;
                e.filename = this._fileInfo.filename;
                throw e;
            }
            output.add(this.important + ((this.inline || (context.lastRule && context.compress)) ? '' : ';'), this._fileInfo, this._index);
        },
        eval: function (context) {
            var mathBypass = false, prevMath, name = this.name, evaldValue, variable = this.variable;
            if (typeof name !== 'string') {
                // expand 'primitive' name directly to get
                // things faster (~10% for benchmark.less):
                name = (name.length === 1) && (name[0] instanceof keyword_1.default) ?
                    name[0].value : evalName(context, name);
                variable = false; // never treat expanded interpolation as new variable name
            }
            // @todo remove when parens-division is default
            if (name === 'font' && context.math === MATH.ALWAYS) {
                mathBypass = true;
                prevMath = context.math;
                context.math = MATH.PARENS_DIVISION;
            }
            try {
                context.importantScope.push({});
                evaldValue = this.value.eval(context);
                if (!this.variable && evaldValue.type === 'DetachedRuleset') {
                    throw { message: 'Rulesets cannot be evaluated on a property.',
                        index: this.getIndex(), filename: this.fileInfo().filename };
                }
                var important = this.important;
                var importantResult = context.importantScope.pop();
                if (!important && importantResult.important) {
                    important = importantResult.important;
                }
                return new Declaration(name, evaldValue, important, this.merge, this.getIndex(), this.fileInfo(), this.inline, variable);
            }
            catch (e) {
                if (typeof e.index !== 'number') {
                    e.index = this.getIndex();
                    e.filename = this.fileInfo().filename;
                }
                throw e;
            }
            finally {
                if (mathBypass) {
                    context.math = prevMath;
                }
            }
        },
        makeImportant: function () {
            return new Declaration(this.name, this.value, '!important', this.merge, this.getIndex(), this.fileInfo(), this.inline);
        }
    });
    exports.default = Declaration;
},
/* less/lib/less/tree/value.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var Value = function (value) {
        if (!value) {
            throw new Error('Value requires an array argument');
        }
        if (!Array.isArray(value)) {
            this.value = [value];
        }
        else {
            this.value = value;
        }
    };
    Value.prototype = Object.assign(new node_1.default(), {
        type: 'Value',
        accept: function (visitor) {
            if (this.value) {
                this.value = visitor.visitArray(this.value);
            }
        },
        eval: function (context) {
            if (this.value.length === 1) {
                return this.value[0].eval(context);
            }
            else {
                return new Value(this.value.map(function (v) {
                    return v.eval(context);
                }));
            }
        },
        genCSS: function (context, output) {
            var i;
            for (i = 0; i < this.value.length; i++) {
                this.value[i].genCSS(context, output);
                if (i + 1 < this.value.length) {
                    output.add((context && context.compress) ? ',' : ', ');
                }
            }
        }
    });
    exports.default = Value;
},
/* less/lib/less/tree/keyword.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var Keyword = function (value) {
        this.value = value;
    };
    Keyword.prototype = Object.assign(new node_1.default(), {
        type: 'Keyword',
        genCSS: function (context, output) {
            if (this.value === '%') {
                throw { type: 'Syntax', message: 'Invalid % without number' };
            }
            output.add(this.value);
        }
    });
    Keyword.True = new Keyword('true');
    Keyword.False = new Keyword('false');
    exports.default = Keyword;
},
/* less/lib/less/tree/anonymous.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var Anonymous = function (value, index, currentFileInfo, mapLines, rulesetLike, visibilityInfo) {
        this.value = value;
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.mapLines = mapLines;
        this.rulesetLike = (typeof rulesetLike === 'undefined') ? false : rulesetLike;
        this.allowRoot = true;
        this.copyVisibilityInfo(visibilityInfo);
    };
    Anonymous.prototype = Object.assign(new node_1.default(), {
        type: 'Anonymous',
        eval: function () {
            return new Anonymous(this.value, this._index, this._fileInfo, this.mapLines, this.rulesetLike, this.visibilityInfo());
        },
        compare: function (other) {
            return other.toCSS && this.toCSS() === other.toCSS() ? 0 : undefined;
        },
        isRulesetLike: function () {
            return this.rulesetLike;
        },
        genCSS: function (context, output) {
            this.nodeVisible = Boolean(this.value);
            if (this.nodeVisible) {
                output.add(this.value, this._fileInfo, this._index, this.mapLines);
            }
        }
    });
    exports.default = Anonymous;
},
/* less/lib/less/tree/comment.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var debug_info_1 = tslib_1.__importDefault(require(131) /* ./debug-info */);
    var Comment = function (value, isLineComment, index, currentFileInfo) {
        this.value = value;
        this.isLineComment = isLineComment;
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.allowRoot = true;
    };
    Comment.prototype = Object.assign(new node_1.default(), {
        type: 'Comment',
        genCSS: function (context, output) {
            if (this.debugInfo) {
                output.add(debug_info_1.default(context, this), this.fileInfo(), this.getIndex());
            }
            output.add(this.value);
        },
        isSilent: function (context) {
            var isCompressed = context.compress && this.value[2] !== '!';
            return this.isLineComment || isCompressed;
        }
    });
    exports.default = Comment;
},
/* less/lib/less/tree/debug-info.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var debugInfo = /** @class */ (function () {
        function debugInfo(context, ctx, lineSeparator) {
            var result = '';
            if (context.dumpLineNumbers && !context.compress) {
                switch (context.dumpLineNumbers) {
                    case 'comments':
                        result = debugInfo.asComment(ctx);
                        break;
                    case 'mediaquery':
                        result = debugInfo.asMediaQuery(ctx);
                        break;
                    case 'all':
                        result = debugInfo.asComment(ctx) + (lineSeparator || '') + debugInfo.asMediaQuery(ctx);
                        break;
                }
            }
            return result;
        }
        debugInfo.asComment = function (ctx) {
            return "/* line " + ctx.debugInfo.lineNumber + ", " + ctx.debugInfo.fileName + " */\n";
        };
        debugInfo.asMediaQuery = function (ctx) {
            var filenameWithProtocol = ctx.debugInfo.fileName;
            if (!/^[a-z]+:\/\//i.test(filenameWithProtocol)) {
                filenameWithProtocol = "file://" + filenameWithProtocol;
            }
            return "@media -sass-debug-info{filename{font-family:" + filenameWithProtocol.replace(/([.:\/\\])/g, function (a) {
                if (a == '\\') {
                    a = '\/';
                }
                return "\\" + a;
            }) + "}line{font-family:\\00003" + ctx.debugInfo.lineNumber + "}}\n";
        };
        return debugInfo;
    }());
    exports.default = debugInfo;
},
/* less/lib/less/contexts.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var contexts = {};
    exports.default = contexts;
    var Constants = tslib_1.__importStar(require(122) /* ./constants */);
    var copyFromOriginal = function copyFromOriginal(original, destination, propertiesToCopy) {
        if (!original) {
            return;
        }
        for (var i = 0; i < propertiesToCopy.length; i++) {
            if (original.hasOwnProperty(propertiesToCopy[i])) {
                destination[propertiesToCopy[i]] = original[propertiesToCopy[i]];
            }
        }
    };
    /*
     parse is used whilst parsing
     */
    var parseCopyProperties = [
        // options
        'paths',
        'rewriteUrls',
        'rootpath',
        'strictImports',
        'insecure',
        'dumpLineNumbers',
        'compress',
        'syncImport',
        'chunkInput',
        'mime',
        'useFileCache',
        // context
        'processImports',
        // Used by the import manager to stop multiple import visitors being created.
        'pluginManager' // Used as the plugin manager for the session
    ];
    contexts.Parse = function (options) {
        copyFromOriginal(options, this, parseCopyProperties);
        if (typeof this.paths === 'string') {
            this.paths = [this.paths];
        }
    };
    var evalCopyProperties = [
        'paths',
        'compress',
        'math',
        'strictUnits',
        'sourceMap',
        'importMultiple',
        'urlArgs',
        'javascriptEnabled',
        'pluginManager',
        'importantScope',
        'rewriteUrls' // option - whether to adjust URL's to be relative
    ];
    contexts.Eval = function (options, frames) {
        copyFromOriginal(options, this, evalCopyProperties);
        if (typeof this.paths === 'string') {
            this.paths = [this.paths];
        }
        this.frames = frames || [];
        this.importantScope = this.importantScope || [];
    };
    contexts.Eval.prototype.enterCalc = function () {
        if (!this.calcStack) {
            this.calcStack = [];
        }
        this.calcStack.push(true);
        this.inCalc = true;
    };
    contexts.Eval.prototype.exitCalc = function () {
        this.calcStack.pop();
        if (!this.calcStack.length) {
            this.inCalc = false;
        }
    };
    contexts.Eval.prototype.inParenthesis = function () {
        if (!this.parensStack) {
            this.parensStack = [];
        }
        this.parensStack.push(true);
    };
    contexts.Eval.prototype.outOfParenthesis = function () {
        this.parensStack.pop();
    };
    contexts.Eval.prototype.inCalc = false;
    contexts.Eval.prototype.mathOn = true;
    contexts.Eval.prototype.isMathOn = function (op) {
        if (!this.mathOn) {
            return false;
        }
        if (op === '/' && this.math !== Constants.Math.ALWAYS && (!this.parensStack || !this.parensStack.length)) {
            return false;
        }
        if (this.math > Constants.Math.PARENS_DIVISION) {
            return this.parensStack && this.parensStack.length;
        }
        return true;
    };
    contexts.Eval.prototype.pathRequiresRewrite = function (path) {
        var isRelative = this.rewriteUrls === Constants.RewriteUrls.LOCAL ? isPathLocalRelative : isPathRelative;
        return isRelative(path);
    };
    contexts.Eval.prototype.rewritePath = function (path, rootpath) {
        var newPath;
        rootpath = rootpath || '';
        newPath = this.normalizePath(rootpath + path);
        // If a path was explicit relative and the rootpath was not an absolute path
        // we must ensure that the new path is also explicit relative.
        if (isPathLocalRelative(path) &&
            isPathRelative(rootpath) &&
            isPathLocalRelative(newPath) === false) {
            newPath = "./" + newPath;
        }
        return newPath;
    };
    contexts.Eval.prototype.normalizePath = function (path) {
        var segments = path.split('/').reverse();
        var segment;
        path = [];
        while (segments.length !== 0) {
            segment = segments.pop();
            switch (segment) {
                case '.':
                    break;
                case '..':
                    if ((path.length === 0) || (path[path.length - 1] === '..')) {
                        path.push(segment);
                    }
                    else {
                        path.pop();
                    }
                    break;
                default:
                    path.push(segment);
                    break;
            }
        }
        return path.join('/');
    };
    function isPathRelative(path) {
        return !/^(?:[a-z-]+:|\/|#)/i.test(path);
    }
    function isPathLocalRelative(path) {
        return path.charAt(0) === '.';
    }
},
/* less/lib/less/functions/function-registry.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function makeRegistry(base) {
        return {
            _data: {},
            add: function (name, func) {
                // precautionary case conversion, as later querying of
                // the registry by function-caller uses lower case as well.
                name = name.toLowerCase();
                if (this._data.hasOwnProperty(name)) {
                    // TODO warn
                }
                this._data[name] = func;
            },
            addMultiple: function (functions) {
                var _this = this;
                Object.keys(functions).forEach(function (name) {
                    _this.add(name, functions[name]);
                });
            },
            get: function (name) {
                return this._data[name] || (base && base.get(name));
            },
            getLocalFunctions: function () {
                return this._data;
            },
            inherit: function () {
                return makeRegistry(this);
            },
            create: function (base) {
                return makeRegistry(base);
            }
        };
    }
    exports.default = makeRegistry(null);
},
/* less/lib/less/functions/default.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var keyword_1 = tslib_1.__importDefault(require(128) /* ../tree/keyword */);
    var defaultFunc = {
        eval: function () {
            var v = this.value_;
            var e = this.error_;
            if (e) {
                throw e;
            }
            if (v != null) {
                return v ? keyword_1.default.True : keyword_1.default.False;
            }
        },
        value: function (v) {
            this.value_ = v;
        },
        error: function (e) {
            this.error_ = e;
        },
        reset: function () {
            this.value_ = this.error_ = null;
        }
    };
    exports.default = defaultFunc;
},
/* less/lib/less/tree/detached-ruleset.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var contexts_1 = tslib_1.__importDefault(require(132) /* ../contexts */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    var DetachedRuleset = function (ruleset, frames) {
        this.ruleset = ruleset;
        this.frames = frames;
        this.setParent(this.ruleset, this);
    };
    DetachedRuleset.prototype = Object.assign(new node_1.default(), {
        type: 'DetachedRuleset',
        evalFirst: true,
        accept: function (visitor) {
            this.ruleset = visitor.visit(this.ruleset);
        },
        eval: function (context) {
            var frames = this.frames || utils.copyArray(context.frames);
            return new DetachedRuleset(this.ruleset, frames);
        },
        callEval: function (context) {
            return this.ruleset.eval(this.frames ? new contexts_1.default.Eval(context, this.frames.concat(context.frames)) : context);
        }
    });
    exports.default = DetachedRuleset;
},
/* less/lib/less/tree/operation.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var color_1 = tslib_1.__importDefault(require(114) /* ./color */);
    var dimension_1 = tslib_1.__importDefault(require(137) /* ./dimension */);
    var Constants = tslib_1.__importStar(require(122) /* ../constants */);
    var MATH = Constants.Math;
    var Operation = function (op, operands, isSpaced) {
        this.op = op.trim();
        this.operands = operands;
        this.isSpaced = isSpaced;
    };
    Operation.prototype = Object.assign(new node_1.default(), {
        type: 'Operation',
        accept: function (visitor) {
            this.operands = visitor.visitArray(this.operands);
        },
        eval: function (context) {
            var a = this.operands[0].eval(context), b = this.operands[1].eval(context), op;
            if (context.isMathOn(this.op)) {
                op = this.op === './' ? '/' : this.op;
                if (a instanceof dimension_1.default && b instanceof color_1.default) {
                    a = a.toColor();
                }
                if (b instanceof dimension_1.default && a instanceof color_1.default) {
                    b = b.toColor();
                }
                if (!a.operate || !b.operate) {
                    if ((a instanceof Operation || b instanceof Operation)
                        && a.op === '/' && context.math === MATH.PARENS_DIVISION) {
                        return new Operation(this.op, [a, b], this.isSpaced);
                    }
                    throw { type: 'Operation',
                        message: 'Operation on an invalid type' };
                }
                return a.operate(context, op, b);
            }
            else {
                return new Operation(this.op, [a, b], this.isSpaced);
            }
        },
        genCSS: function (context, output) {
            this.operands[0].genCSS(context, output);
            if (this.isSpaced) {
                output.add(' ');
            }
            output.add(this.op);
            if (this.isSpaced) {
                output.add(' ');
            }
            this.operands[1].genCSS(context, output);
        }
    });
    exports.default = Operation;
},
/* less/lib/less/tree/dimension.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var unit_conversions_1 = tslib_1.__importDefault(require(111) /* ../data/unit-conversions */);
    var unit_1 = tslib_1.__importDefault(require(138) /* ./unit */);
    var color_1 = tslib_1.__importDefault(require(114) /* ./color */);
    //
    // A number with a unit
    //
    var Dimension = function (value, unit) {
        this.value = parseFloat(value);
        if (isNaN(this.value)) {
            throw new Error('Dimension is not a number.');
        }
        this.unit = (unit && unit instanceof unit_1.default) ? unit :
            new unit_1.default(unit ? [unit] : undefined);
        this.setParent(this.unit, this);
    };
    Dimension.prototype = Object.assign(new node_1.default(), {
        type: 'Dimension',
        accept: function (visitor) {
            this.unit = visitor.visit(this.unit);
        },
        eval: function (context) {
            return this;
        },
        toColor: function () {
            return new color_1.default([this.value, this.value, this.value]);
        },
        genCSS: function (context, output) {
            if ((context && context.strictUnits) && !this.unit.isSingular()) {
                throw new Error("Multiple units in dimension. Correct the units or use the unit function. Bad unit: " + this.unit.toString());
            }
            var value = this.fround(context, this.value);
            var strValue = String(value);
            if (value !== 0 && value < 0.000001 && value > -0.000001) {
                // would be output 1e-6 etc.
                strValue = value.toFixed(20).replace(/0+$/, '');
            }
            if (context && context.compress) {
                // Zero values doesn't need a unit
                if (value === 0 && this.unit.isLength()) {
                    output.add(strValue);
                    return;
                }
                // Float values doesn't need a leading zero
                if (value > 0 && value < 1) {
                    strValue = (strValue).substr(1);
                }
            }
            output.add(strValue);
            this.unit.genCSS(context, output);
        },
        // In an operation between two Dimensions,
        // we default to the first Dimension's unit,
        // so `1px + 2` will yield `3px`.
        operate: function (context, op, other) {
            /* jshint noempty:false */
            var value = this._operate(context, op, this.value, other.value);
            var unit = this.unit.clone();
            if (op === '+' || op === '-') {
                if (unit.numerator.length === 0 && unit.denominator.length === 0) {
                    unit = other.unit.clone();
                    if (this.unit.backupUnit) {
                        unit.backupUnit = this.unit.backupUnit;
                    }
                }
                else if (other.unit.numerator.length === 0 && unit.denominator.length === 0) {
                    // do nothing
                }
                else {
                    other = other.convertTo(this.unit.usedUnits());
                    if (context.strictUnits && other.unit.toString() !== unit.toString()) {
                        throw new Error("Incompatible units. Change the units or use the unit function. "
                            + ("Bad units: '" + unit.toString() + "' and '" + other.unit.toString() + "'."));
                    }
                    value = this._operate(context, op, this.value, other.value);
                }
            }
            else if (op === '*') {
                unit.numerator = unit.numerator.concat(other.unit.numerator).sort();
                unit.denominator = unit.denominator.concat(other.unit.denominator).sort();
                unit.cancel();
            }
            else if (op === '/') {
                unit.numerator = unit.numerator.concat(other.unit.denominator).sort();
                unit.denominator = unit.denominator.concat(other.unit.numerator).sort();
                unit.cancel();
            }
            return new Dimension(value, unit);
        },
        compare: function (other) {
            var a, b;
            if (!(other instanceof Dimension)) {
                return undefined;
            }
            if (this.unit.isEmpty() || other.unit.isEmpty()) {
                a = this;
                b = other;
            }
            else {
                a = this.unify();
                b = other.unify();
                if (a.unit.compare(b.unit) !== 0) {
                    return undefined;
                }
            }
            return node_1.default.numericCompare(a.value, b.value);
        },
        unify: function () {
            return this.convertTo({ length: 'px', duration: 's', angle: 'rad' });
        },
        convertTo: function (conversions) {
            var value = this.value;
            var unit = this.unit.clone();
            var i;
            var groupName;
            var group;
            var targetUnit;
            var derivedConversions = {};
            var applyUnit;
            if (typeof conversions === 'string') {
                for (i in unit_conversions_1.default) {
                    if (unit_conversions_1.default[i].hasOwnProperty(conversions)) {
                        derivedConversions = {};
                        derivedConversions[i] = conversions;
                    }
                }
                conversions = derivedConversions;
            }
            applyUnit = function (atomicUnit, denominator) {
                /* jshint loopfunc:true */
                if (group.hasOwnProperty(atomicUnit)) {
                    if (denominator) {
                        value = value / (group[atomicUnit] / group[targetUnit]);
                    }
                    else {
                        value = value * (group[atomicUnit] / group[targetUnit]);
                    }
                    return targetUnit;
                }
                return atomicUnit;
            };
            for (groupName in conversions) {
                if (conversions.hasOwnProperty(groupName)) {
                    targetUnit = conversions[groupName];
                    group = unit_conversions_1.default[groupName];
                    unit.map(applyUnit);
                }
            }
            unit.cancel();
            return new Dimension(value, unit);
        }
    });
    exports.default = Dimension;
},
/* less/lib/less/tree/unit.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var unit_conversions_1 = tslib_1.__importDefault(require(111) /* ../data/unit-conversions */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    var Unit = function (numerator, denominator, backupUnit) {
        this.numerator = numerator ? utils.copyArray(numerator).sort() : [];
        this.denominator = denominator ? utils.copyArray(denominator).sort() : [];
        if (backupUnit) {
            this.backupUnit = backupUnit;
        }
        else if (numerator && numerator.length) {
            this.backupUnit = numerator[0];
        }
    };
    Unit.prototype = Object.assign(new node_1.default(), {
        type: 'Unit',
        clone: function () {
            return new Unit(utils.copyArray(this.numerator), utils.copyArray(this.denominator), this.backupUnit);
        },
        genCSS: function (context, output) {
            // Dimension checks the unit is singular and throws an error if in strict math mode.
            var strictUnits = context && context.strictUnits;
            if (this.numerator.length === 1) {
                output.add(this.numerator[0]); // the ideal situation
            }
            else if (!strictUnits && this.backupUnit) {
                output.add(this.backupUnit);
            }
            else if (!strictUnits && this.denominator.length) {
                output.add(this.denominator[0]);
            }
        },
        toString: function () {
            var i, returnStr = this.numerator.join('*');
            for (i = 0; i < this.denominator.length; i++) {
                returnStr += "/" + this.denominator[i];
            }
            return returnStr;
        },
        compare: function (other) {
            return this.is(other.toString()) ? 0 : undefined;
        },
        is: function (unitString) {
            return this.toString().toUpperCase() === unitString.toUpperCase();
        },
        isLength: function () {
            return RegExp('^(px|em|ex|ch|rem|in|cm|mm|pc|pt|ex|vw|vh|vmin|vmax)$', 'gi').test(this.toCSS());
        },
        isEmpty: function () {
            return this.numerator.length === 0 && this.denominator.length === 0;
        },
        isSingular: function () {
            return this.numerator.length <= 1 && this.denominator.length === 0;
        },
        map: function (callback) {
            var i;
            for (i = 0; i < this.numerator.length; i++) {
                this.numerator[i] = callback(this.numerator[i], false);
            }
            for (i = 0; i < this.denominator.length; i++) {
                this.denominator[i] = callback(this.denominator[i], true);
            }
        },
        usedUnits: function () {
            var group;
            var result = {};
            var mapUnit;
            var groupName;
            mapUnit = function (atomicUnit) {
                /* jshint loopfunc:true */
                if (group.hasOwnProperty(atomicUnit) && !result[groupName]) {
                    result[groupName] = atomicUnit;
                }
                return atomicUnit;
            };
            for (groupName in unit_conversions_1.default) {
                if (unit_conversions_1.default.hasOwnProperty(groupName)) {
                    group = unit_conversions_1.default[groupName];
                    this.map(mapUnit);
                }
            }
            return result;
        },
        cancel: function () {
            var counter = {};
            var atomicUnit;
            var i;
            for (i = 0; i < this.numerator.length; i++) {
                atomicUnit = this.numerator[i];
                counter[atomicUnit] = (counter[atomicUnit] || 0) + 1;
            }
            for (i = 0; i < this.denominator.length; i++) {
                atomicUnit = this.denominator[i];
                counter[atomicUnit] = (counter[atomicUnit] || 0) - 1;
            }
            this.numerator = [];
            this.denominator = [];
            for (atomicUnit in counter) {
                if (counter.hasOwnProperty(atomicUnit)) {
                    var count = counter[atomicUnit];
                    if (count > 0) {
                        for (i = 0; i < count; i++) {
                            this.numerator.push(atomicUnit);
                        }
                    }
                    else if (count < 0) {
                        for (i = 0; i < -count; i++) {
                            this.denominator.push(atomicUnit);
                        }
                    }
                }
            }
            this.numerator.sort();
            this.denominator.sort();
        }
    });
    exports.default = Unit;
},
/* less/lib/less/tree/variable.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var call_1 = tslib_1.__importDefault(require(140) /* ./call */);
    var Variable = function (name, index, currentFileInfo) {
        this.name = name;
        this._index = index;
        this._fileInfo = currentFileInfo;
    };
    Variable.prototype = Object.assign(new node_1.default(), {
        type: 'Variable',
        eval: function (context) {
            var variable, name = this.name;
            if (name.indexOf('@@') === 0) {
                name = "@" + new Variable(name.slice(1), this.getIndex(), this.fileInfo()).eval(context).value;
            }
            if (this.evaluating) {
                throw { type: 'Name', message: "Recursive variable definition for " + name, filename: this.fileInfo().filename,
                    index: this.getIndex() };
            }
            this.evaluating = true;
            variable = this.find(context.frames, function (frame) {
                var v = frame.variable(name);
                if (v) {
                    if (v.important) {
                        var importantScope = context.importantScope[context.importantScope.length - 1];
                        importantScope.important = v.important;
                    }
                    // If in calc, wrap vars in a function call to cascade evaluate args first
                    if (context.inCalc) {
                        return (new call_1.default('_SELF', [v.value])).eval(context);
                    }
                    else {
                        return v.value.eval(context);
                    }
                }
            });
            if (variable) {
                this.evaluating = false;
                return variable;
            }
            else {
                throw { type: 'Name', message: "variable " + name + " is undefined", filename: this.fileInfo().filename,
                    index: this.getIndex() };
            }
        },
        find: function (obj, fun) {
            for (var i = 0, r = void 0; i < obj.length; i++) {
                r = fun.call(obj, obj[i]);
                if (r) {
                    return r;
                }
            }
            return null;
        }
    });
    exports.default = Variable;
},
/* less/lib/less/tree/call.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ./anonymous */);
    var function_caller_1 = tslib_1.__importDefault(require(141) /* ../functions/function-caller */);
    //
    // A function call node.
    //
    var Call = function (name, args, index, currentFileInfo) {
        this.name = name;
        this.args = args;
        this.calc = name === 'calc';
        this._index = index;
        this._fileInfo = currentFileInfo;
    };
    Call.prototype = Object.assign(new node_1.default(), {
        type: 'Call',
        accept: function (visitor) {
            if (this.args) {
                this.args = visitor.visitArray(this.args);
            }
        },
        //
        // When evaluating a function call,
        // we either find the function in the functionRegistry,
        // in which case we call it, passing the  evaluated arguments,
        // if this returns null or we cannot find the function, we
        // simply print it out as it appeared originally [2].
        //
        // The reason why we evaluate the arguments, is in the case where
        // we try to pass a variable to a function, like: `saturate(@color)`.
        // The function should receive the value, not the variable.
        //
        eval: function (context) {
            var _this = this;
            /**
             * Turn off math for calc(), and switch back on for evaluating nested functions
             */
            var currentMathContext = context.mathOn;
            context.mathOn = !this.calc;
            if (this.calc || context.inCalc) {
                context.enterCalc();
            }
            var exitCalc = function () {
                if (_this.calc || context.inCalc) {
                    context.exitCalc();
                }
                context.mathOn = currentMathContext;
            };
            var result;
            var funcCaller = new function_caller_1.default(this.name, context, this.getIndex(), this.fileInfo());
            if (funcCaller.isValid()) {
                try {
                    result = funcCaller.call(this.args);
                    exitCalc();
                }
                catch (e) {
                    if (e.hasOwnProperty('line') && e.hasOwnProperty('column')) {
                        throw e;
                    }
                    throw {
                        type: e.type || 'Runtime',
                        message: "Error evaluating function `" + this.name + "`" + (e.message ? ": " + e.message : ''),
                        index: this.getIndex(),
                        filename: this.fileInfo().filename,
                        line: e.lineNumber,
                        column: e.columnNumber
                    };
                }
            }
            if (result !== null && result !== undefined) {
                // Results that that are not nodes are cast as Anonymous nodes
                // Falsy values or booleans are returned as empty nodes
                if (!(result instanceof node_1.default)) {
                    if (!result || result === true) {
                        result = new anonymous_1.default(null);
                    }
                    else {
                        result = new anonymous_1.default(result.toString());
                    }
                }
                result._index = this._index;
                result._fileInfo = this._fileInfo;
                return result;
            }
            var args = this.args.map(function (a) { return a.eval(context); });
            exitCalc();
            return new Call(this.name, args, this.getIndex(), this.fileInfo());
        },
        genCSS: function (context, output) {
            output.add(this.name + "(", this.fileInfo(), this.getIndex());
            for (var i = 0; i < this.args.length; i++) {
                this.args[i].genCSS(context, output);
                if (i + 1 < this.args.length) {
                    output.add(', ');
                }
            }
            output.add(')');
        }
    });
    exports.default = Call;
},
/* less/lib/less/functions/function-caller.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var expression_1 = tslib_1.__importDefault(require(142) /* ../tree/expression */);
    var functionCaller = /** @class */ (function () {
        function functionCaller(name, context, index, currentFileInfo) {
            this.name = name.toLowerCase();
            this.index = index;
            this.context = context;
            this.currentFileInfo = currentFileInfo;
            this.func = context.frames[0].functionRegistry.get(this.name);
        }
        functionCaller.prototype.isValid = function () {
            return Boolean(this.func);
        };
        functionCaller.prototype.call = function (args) {
            var _this = this;
            if (!(Array.isArray(args))) {
                args = [args];
            }
            var evalArgs = this.func.evalArgs;
            if (evalArgs !== false) {
                args = args.map(function (a) { return a.eval(_this.context); });
            }
            var commentFilter = function (item) { return !(item.type === 'Comment'); };
            // This code is terrible and should be replaced as per this issue...
            // https://github.com/less/less.js/issues/2477
            args = args
                .filter(commentFilter)
                .map(function (item) {
                if (item.type === 'Expression') {
                    var subNodes = item.value.filter(commentFilter);
                    if (subNodes.length === 1) {
                        return subNodes[0];
                    }
                    else {
                        return new expression_1.default(subNodes);
                    }
                }
                return item;
            });
            if (evalArgs === false) {
                return this.func.apply(this, tslib_1.__spreadArrays([this.context], args));
            }
            return this.func.apply(this, args);
        };
        return functionCaller;
    }());
    exports.default = functionCaller;
},
/* less/lib/less/tree/expression.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var paren_1 = tslib_1.__importDefault(require(118) /* ./paren */);
    var comment_1 = tslib_1.__importDefault(require(130) /* ./comment */);
    var dimension_1 = tslib_1.__importDefault(require(137) /* ./dimension */);
    var Constants = tslib_1.__importStar(require(122) /* ../constants */);
    var MATH = Constants.Math;
    var Expression = function (value, noSpacing) {
        this.value = value;
        this.noSpacing = noSpacing;
        if (!value) {
            throw new Error('Expression requires an array parameter');
        }
    };
    Expression.prototype = Object.assign(new node_1.default(), {
        type: 'Expression',
        accept: function (visitor) {
            this.value = visitor.visitArray(this.value);
        },
        eval: function (context) {
            var returnValue;
            var mathOn = context.isMathOn();
            var inParenthesis = this.parens;
            var doubleParen = false;
            if (inParenthesis) {
                context.inParenthesis();
            }
            if (this.value.length > 1) {
                returnValue = new Expression(this.value.map(function (e) {
                    if (!e.eval) {
                        return e;
                    }
                    return e.eval(context);
                }), this.noSpacing);
            }
            else if (this.value.length === 1) {
                if (this.value[0].parens && !this.value[0].parensInOp && !context.inCalc) {
                    doubleParen = true;
                }
                returnValue = this.value[0].eval(context);
            }
            else {
                returnValue = this;
            }
            if (inParenthesis) {
                context.outOfParenthesis();
            }
            if (this.parens && this.parensInOp && !mathOn && !doubleParen
                && (!(returnValue instanceof dimension_1.default))) {
                returnValue = new paren_1.default(returnValue);
            }
            return returnValue;
        },
        genCSS: function (context, output) {
            for (var i = 0; i < this.value.length; i++) {
                this.value[i].genCSS(context, output);
                if (!this.noSpacing && i + 1 < this.value.length) {
                    output.add(' ');
                }
            }
        },
        throwAwayComments: function () {
            this.value = this.value.filter(function (v) {
                return !(v instanceof comment_1.default);
            });
        }
    });
    exports.default = Expression;
},
/* less/lib/less/tree/property.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var declaration_1 = tslib_1.__importDefault(require(126) /* ./declaration */);
    var Property = function (name, index, currentFileInfo) {
        this.name = name;
        this._index = index;
        this._fileInfo = currentFileInfo;
    };
    Property.prototype = Object.assign(new node_1.default(), {
        type: 'Property',
        eval: function (context) {
            var property;
            var name = this.name;
            // TODO: shorten this reference
            var mergeRules = context.pluginManager.less.visitors.ToCSSVisitor.prototype._mergeRules;
            if (this.evaluating) {
                throw { type: 'Name', message: "Recursive property reference for " + name, filename: this.fileInfo().filename,
                    index: this.getIndex() };
            }
            this.evaluating = true;
            property = this.find(context.frames, function (frame) {
                var v;
                var vArr = frame.property(name);
                if (vArr) {
                    for (var i = 0; i < vArr.length; i++) {
                        v = vArr[i];
                        vArr[i] = new declaration_1.default(v.name, v.value, v.important, v.merge, v.index, v.currentFileInfo, v.inline, v.variable);
                    }
                    mergeRules(vArr);
                    v = vArr[vArr.length - 1];
                    if (v.important) {
                        var importantScope = context.importantScope[context.importantScope.length - 1];
                        importantScope.important = v.important;
                    }
                    v = v.value.eval(context);
                    return v;
                }
            });
            if (property) {
                this.evaluating = false;
                return property;
            }
            else {
                throw { type: 'Name', message: "Property '" + name + "' is undefined", filename: this.currentFileInfo.filename,
                    index: this.index };
            }
        },
        find: function (obj, fun) {
            for (var i = 0, r = void 0; i < obj.length; i++) {
                r = fun.call(obj, obj[i]);
                if (r) {
                    return r;
                }
            }
            return null;
        }
    });
    exports.default = Property;
},
/* less/lib/less/tree/attribute.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var Attribute = function (key, op, value) {
        this.key = key;
        this.op = op;
        this.value = value;
    };
    Attribute.prototype = Object.assign(new node_1.default(), {
        type: 'Attribute',
        eval: function (context) {
            return new Attribute(this.key.eval ? this.key.eval(context) : this.key, this.op, (this.value && this.value.eval) ? this.value.eval(context) : this.value);
        },
        genCSS: function (context, output) {
            output.add(this.toCSS(context));
        },
        toCSS: function (context) {
            var value = this.key.toCSS ? this.key.toCSS(context) : this.key;
            if (this.op) {
                value += this.op;
                value += (this.value.toCSS ? this.value.toCSS(context) : this.value);
            }
            return "[" + value + "]";
        }
    });
    exports.default = Attribute;
},
/* less/lib/less/tree/quoted.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var variable_1 = tslib_1.__importDefault(require(139) /* ./variable */);
    var property_1 = tslib_1.__importDefault(require(143) /* ./property */);
    var Quoted = function (str, content, escaped, index, currentFileInfo) {
        this.escaped = (escaped == null) ? true : escaped;
        this.value = content || '';
        this.quote = str.charAt(0);
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.variableRegex = /@\{([\w-]+)\}/g;
        this.propRegex = /\$\{([\w-]+)\}/g;
        this.allowRoot = escaped;
    };
    Quoted.prototype = Object.assign(new node_1.default(), {
        type: 'Quoted',
        genCSS: function (context, output) {
            if (!this.escaped) {
                output.add(this.quote, this.fileInfo(), this.getIndex());
            }
            output.add(this.value);
            if (!this.escaped) {
                output.add(this.quote);
            }
        },
        containsVariables: function () {
            return this.value.match(this.variableRegex);
        },
        eval: function (context) {
            var that = this;
            var value = this.value;
            var variableReplacement = function (_, name) {
                var v = new variable_1.default("@" + name, that.getIndex(), that.fileInfo()).eval(context, true);
                return (v instanceof Quoted) ? v.value : v.toCSS();
            };
            var propertyReplacement = function (_, name) {
                var v = new property_1.default("$" + name, that.getIndex(), that.fileInfo()).eval(context, true);
                return (v instanceof Quoted) ? v.value : v.toCSS();
            };
            function iterativeReplace(value, regexp, replacementFnc) {
                var evaluatedValue = value;
                do {
                    value = evaluatedValue.toString();
                    evaluatedValue = value.replace(regexp, replacementFnc);
                } while (value !== evaluatedValue);
                return evaluatedValue;
            }
            value = iterativeReplace(value, this.variableRegex, variableReplacement);
            value = iterativeReplace(value, this.propRegex, propertyReplacement);
            return new Quoted(this.quote + value + this.quote, value, this.escaped, this.getIndex(), this.fileInfo());
        },
        compare: function (other) {
            // when comparing quoted strings allow the quote to differ
            if (other.type === 'Quoted' && !this.escaped && !other.escaped) {
                return node_1.default.numericCompare(this.value, other.value);
            }
            else {
                return other.toCSS && this.toCSS() === other.toCSS() ? 0 : undefined;
            }
        }
    });
    exports.default = Quoted;
},
/* less/lib/less/tree/url.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    function escapePath(path) {
        return path.replace(/[\(\)'"\s]/g, function (match) { return "\\" + match; });
    }
    var URL = function (val, index, currentFileInfo, isEvald) {
        this.value = val;
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.isEvald = isEvald;
    };
    URL.prototype = Object.assign(new node_1.default(), {
        type: 'Url',
        accept: function (visitor) {
            this.value = visitor.visit(this.value);
        },
        genCSS: function (context, output) {
            output.add('url(');
            this.value.genCSS(context, output);
            output.add(')');
        },
        eval: function (context) {
            var val = this.value.eval(context);
            var rootpath;
            if (!this.isEvald) {
                // Add the rootpath if the URL requires a rewrite
                rootpath = this.fileInfo() && this.fileInfo().rootpath;
                if (typeof rootpath === 'string' &&
                    typeof val.value === 'string' &&
                    context.pathRequiresRewrite(val.value)) {
                    if (!val.quote) {
                        rootpath = escapePath(rootpath);
                    }
                    val.value = context.rewritePath(val.value, rootpath);
                }
                else {
                    val.value = context.normalizePath(val.value);
                }
                // Add url args if enabled
                if (context.urlArgs) {
                    if (!val.value.match(/^\s*data:/)) {
                        var delimiter = val.value.indexOf('?') === -1 ? '?' : '&';
                        var urlArgs = delimiter + context.urlArgs;
                        if (val.value.indexOf('#') !== -1) {
                            val.value = val.value.replace('#', urlArgs + "#");
                        }
                        else {
                            val.value += urlArgs;
                        }
                    }
                }
            }
            return new URL(val, this.getIndex(), this.fileInfo(), true);
        }
    });
    exports.default = URL;
},
/* less/lib/less/tree/import.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var media_1 = tslib_1.__importDefault(require(148) /* ./media */);
    var url_1 = tslib_1.__importDefault(require(146) /* ./url */);
    var quoted_1 = tslib_1.__importDefault(require(145) /* ./quoted */);
    var ruleset_1 = tslib_1.__importDefault(require(125) /* ./ruleset */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ./anonymous */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    var less_error_1 = tslib_1.__importDefault(require(120) /* ../less-error */);
    //
    // CSS @import node
    //
    // The general strategy here is that we don't want to wait
    // for the parsing to be completed, before we start importing
    // the file. That's because in the context of a browser,
    // most of the time will be spent waiting for the server to respond.
    //
    // On creation, we push the import path to our import queue, though
    // `import,push`, we also pass it a callback, which it'll call once
    // the file has been fetched, and parsed.
    //
    var Import = function (path, features, options, index, currentFileInfo, visibilityInfo) {
        this.options = options;
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.path = path;
        this.features = features;
        this.allowRoot = true;
        if (this.options.less !== undefined || this.options.inline) {
            this.css = !this.options.less || this.options.inline;
        }
        else {
            var pathValue = this.getPath();
            if (pathValue && /[#\.\&\?]css([\?;].*)?$/.test(pathValue)) {
                this.css = true;
            }
        }
        this.copyVisibilityInfo(visibilityInfo);
        this.setParent(this.features, this);
        this.setParent(this.path, this);
    };
    Import.prototype = Object.assign(new node_1.default(), {
        type: 'Import',
        accept: function (visitor) {
            if (this.features) {
                this.features = visitor.visit(this.features);
            }
            this.path = visitor.visit(this.path);
            if (!this.options.isPlugin && !this.options.inline && this.root) {
                this.root = visitor.visit(this.root);
            }
        },
        genCSS: function (context, output) {
            if (this.css && this.path._fileInfo.reference === undefined) {
                output.add('@import ', this._fileInfo, this._index);
                this.path.genCSS(context, output);
                if (this.features) {
                    output.add(' ');
                    this.features.genCSS(context, output);
                }
                output.add(';');
            }
        },
        getPath: function () {
            return (this.path instanceof url_1.default) ?
                this.path.value.value : this.path.value;
        },
        isVariableImport: function () {
            var path = this.path;
            if (path instanceof url_1.default) {
                path = path.value;
            }
            if (path instanceof quoted_1.default) {
                return path.containsVariables();
            }
            return true;
        },
        evalForImport: function (context) {
            var path = this.path;
            if (path instanceof url_1.default) {
                path = path.value;
            }
            return new Import(path.eval(context), this.features, this.options, this._index, this._fileInfo, this.visibilityInfo());
        },
        evalPath: function (context) {
            var path = this.path.eval(context);
            var fileInfo = this._fileInfo;
            if (!(path instanceof url_1.default)) {
                // Add the rootpath if the URL requires a rewrite
                var pathValue = path.value;
                if (fileInfo &&
                    pathValue &&
                    context.pathRequiresRewrite(pathValue)) {
                    path.value = context.rewritePath(pathValue, fileInfo.rootpath);
                }
                else {
                    path.value = context.normalizePath(path.value);
                }
            }
            return path;
        },
        eval: function (context) {
            var result = this.doEval(context);
            if (this.options.reference || this.blocksVisibility()) {
                if (result.length || result.length === 0) {
                    result.forEach(function (node) {
                        node.addVisibilityBlock();
                    });
                }
                else {
                    result.addVisibilityBlock();
                }
            }
            return result;
        },
        doEval: function (context) {
            var ruleset;
            var registry;
            var features = this.features && this.features.eval(context);
            if (this.options.isPlugin) {
                if (this.root && this.root.eval) {
                    try {
                        this.root.eval(context);
                    }
                    catch (e) {
                        e.message = 'Plugin error during evaluation';
                        throw new less_error_1.default(e, this.root.imports, this.root.filename);
                    }
                }
                registry = context.frames[0] && context.frames[0].functionRegistry;
                if (registry && this.root && this.root.functions) {
                    registry.addMultiple(this.root.functions);
                }
                return [];
            }
            if (this.skip) {
                if (typeof this.skip === 'function') {
                    this.skip = this.skip();
                }
                if (this.skip) {
                    return [];
                }
            }
            if (this.options.inline) {
                var contents = new anonymous_1.default(this.root, 0, {
                    filename: this.importedFilename,
                    reference: this.path._fileInfo && this.path._fileInfo.reference
                }, true, true);
                return this.features ? new media_1.default([contents], this.features.value) : [contents];
            }
            else if (this.css) {
                var newImport = new Import(this.evalPath(context), features, this.options, this._index);
                if (!newImport.css && this.error) {
                    throw this.error;
                }
                return newImport;
            }
            else if (this.root) {
                ruleset = new ruleset_1.default(null, utils.copyArray(this.root.rules));
                ruleset.evalImports(context);
                return this.features ? new media_1.default(ruleset.rules, this.features.value) : ruleset.rules;
            }
            else {
                return [];
            }
        }
    });
    exports.default = Import;
},
/* less/lib/less/tree/media.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var ruleset_1 = tslib_1.__importDefault(require(125) /* ./ruleset */);
    var value_1 = tslib_1.__importDefault(require(127) /* ./value */);
    var selector_1 = tslib_1.__importDefault(require(116) /* ./selector */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ./anonymous */);
    var expression_1 = tslib_1.__importDefault(require(142) /* ./expression */);
    var atrule_1 = tslib_1.__importDefault(require(115) /* ./atrule */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    var Media = function (value, features, index, currentFileInfo, visibilityInfo) {
        this._index = index;
        this._fileInfo = currentFileInfo;
        var selectors = (new selector_1.default([], null, null, this._index, this._fileInfo)).createEmptySelectors();
        this.features = new value_1.default(features);
        this.rules = [new ruleset_1.default(selectors, value)];
        this.rules[0].allowImports = true;
        this.copyVisibilityInfo(visibilityInfo);
        this.allowRoot = true;
        this.setParent(selectors, this);
        this.setParent(this.features, this);
        this.setParent(this.rules, this);
    };
    Media.prototype = Object.assign(new atrule_1.default(), {
        type: 'Media',
        isRulesetLike: function () {
            return true;
        },
        accept: function (visitor) {
            if (this.features) {
                this.features = visitor.visit(this.features);
            }
            if (this.rules) {
                this.rules = visitor.visitArray(this.rules);
            }
        },
        genCSS: function (context, output) {
            output.add('@media ', this._fileInfo, this._index);
            this.features.genCSS(context, output);
            this.outputRuleset(context, output, this.rules);
        },
        eval: function (context) {
            if (!context.mediaBlocks) {
                context.mediaBlocks = [];
                context.mediaPath = [];
            }
            var media = new Media(null, [], this._index, this._fileInfo, this.visibilityInfo());
            if (this.debugInfo) {
                this.rules[0].debugInfo = this.debugInfo;
                media.debugInfo = this.debugInfo;
            }
            media.features = this.features.eval(context);
            context.mediaPath.push(media);
            context.mediaBlocks.push(media);
            this.rules[0].functionRegistry = context.frames[0].functionRegistry.inherit();
            context.frames.unshift(this.rules[0]);
            media.rules = [this.rules[0].eval(context)];
            context.frames.shift();
            context.mediaPath.pop();
            return context.mediaPath.length === 0 ? media.evalTop(context) :
                media.evalNested(context);
        },
        evalTop: function (context) {
            var result = this;
            // Render all dependent Media blocks.
            if (context.mediaBlocks.length > 1) {
                var selectors = (new selector_1.default([], null, null, this.getIndex(), this.fileInfo())).createEmptySelectors();
                result = new ruleset_1.default(selectors, context.mediaBlocks);
                result.multiMedia = true;
                result.copyVisibilityInfo(this.visibilityInfo());
                this.setParent(result, this);
            }
            delete context.mediaBlocks;
            delete context.mediaPath;
            return result;
        },
        evalNested: function (context) {
            var i;
            var value;
            var path = context.mediaPath.concat([this]);
            // Extract the media-query conditions separated with `,` (OR).
            for (i = 0; i < path.length; i++) {
                value = path[i].features instanceof value_1.default ?
                    path[i].features.value : path[i].features;
                path[i] = Array.isArray(value) ? value : [value];
            }
            // Trace all permutations to generate the resulting media-query.
            //
            // (a, b and c) with nested (d, e) ->
            //    a and d
            //    a and e
            //    b and c and d
            //    b and c and e
            this.features = new value_1.default(this.permute(path).map(function (path) {
                path = path.map(function (fragment) { return fragment.toCSS ? fragment : new anonymous_1.default(fragment); });
                for (i = path.length - 1; i > 0; i--) {
                    path.splice(i, 0, new anonymous_1.default('and'));
                }
                return new expression_1.default(path);
            }));
            this.setParent(this.features, this);
            // Fake a tree-node that doesn't output anything.
            return new ruleset_1.default([], []);
        },
        permute: function (arr) {
            if (arr.length === 0) {
                return [];
            }
            else if (arr.length === 1) {
                return arr[0];
            }
            else {
                var result = [];
                var rest = this.permute(arr.slice(1));
                for (var i = 0; i < rest.length; i++) {
                    for (var j = 0; j < arr[0].length; j++) {
                        result.push([arr[0][j]].concat(rest[i]));
                    }
                }
                return result;
            }
        },
        bubbleSelectors: function (selectors) {
            if (!selectors) {
                return;
            }
            this.rules = [new ruleset_1.default(utils.copyArray(selectors), [this.rules[0]])];
            this.setParent(this.rules, this);
        }
    });
    exports.default = Media;
},
/* less/lib/less/tree/javascript.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var js_eval_node_1 = tslib_1.__importDefault(require(150) /* ./js-eval-node */);
    var dimension_1 = tslib_1.__importDefault(require(137) /* ./dimension */);
    var quoted_1 = tslib_1.__importDefault(require(145) /* ./quoted */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ./anonymous */);
    var JavaScript = function (string, escaped, index, currentFileInfo) {
        this.escaped = escaped;
        this.expression = string;
        this._index = index;
        this._fileInfo = currentFileInfo;
    };
    JavaScript.prototype = Object.assign(new js_eval_node_1.default(), {
        type: 'JavaScript',
        eval: function (context) {
            var result = this.evaluateJavaScript(this.expression, context);
            var type = typeof result;
            if (type === 'number' && !isNaN(result)) {
                return new dimension_1.default(result);
            }
            else if (type === 'string') {
                return new quoted_1.default("\"" + result + "\"", result, this.escaped, this._index);
            }
            else if (Array.isArray(result)) {
                return new anonymous_1.default(result.join(', '));
            }
            else {
                return new anonymous_1.default(result);
            }
        }
    });
    exports.default = JavaScript;
},
/* less/lib/less/tree/js-eval-node.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var variable_1 = tslib_1.__importDefault(require(139) /* ./variable */);
    var JsEvalNode = function () { };
    JsEvalNode.prototype = Object.assign(new node_1.default(), {
        evaluateJavaScript: function (expression, context) {
            var result;
            var that = this;
            var evalContext = {};
            if (!context.javascriptEnabled) {
                throw { message: 'Inline JavaScript is not enabled. Is it set in your options?',
                    filename: this.fileInfo().filename,
                    index: this.getIndex() };
            }
            expression = expression.replace(/@\{([\w-]+)\}/g, function (_, name) {
                return that.jsify(new variable_1.default("@" + name, that.getIndex(), that.fileInfo()).eval(context));
            });
            try {
                expression = new Function("return (" + expression + ")");
            }
            catch (e) {
                throw { message: "JavaScript evaluation error: " + e.message + " from `" + expression + "`", filename: this.fileInfo().filename,
                    index: this.getIndex() };
            }
            var variables = context.frames[0].variables();
            for (var k in variables) {
                if (variables.hasOwnProperty(k)) {
                    /* jshint loopfunc:true */
                    evalContext[k.slice(1)] = {
                        value: variables[k].value,
                        toJS: function () {
                            return this.value.eval(context).toCSS();
                        }
                    };
                }
            }
            try {
                result = expression.call(evalContext);
            }
            catch (e) {
                throw { message: "JavaScript evaluation error: '" + e.name + ": " + e.message.replace(/["]/g, '\'') + "'", filename: this.fileInfo().filename,
                    index: this.getIndex() };
            }
            return result;
        },
        jsify: function (obj) {
            if (Array.isArray(obj.value) && (obj.value.length > 1)) {
                return "[" + obj.value.map(function (v) { return v.toCSS(); }).join(', ') + "]";
            }
            else {
                return obj.toCSS();
            }
        }
    });
    exports.default = JsEvalNode;
},
/* less/lib/less/tree/assignment.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var Assignment = function (key, val) {
        this.key = key;
        this.value = val;
    };
    Assignment.prototype = Object.assign(new node_1.default(), {
        type: 'Assignment',
        accept: function (visitor) {
            this.value = visitor.visit(this.value);
        },
        eval: function (context) {
            if (this.value.eval) {
                return new Assignment(this.key, this.value.eval(context));
            }
            return this;
        },
        genCSS: function (context, output) {
            output.add(this.key + "=");
            if (this.value.genCSS) {
                this.value.genCSS(context, output);
            }
            else {
                output.add(this.value);
            }
        }
    });
    exports.default = Assignment;
},
/* less/lib/less/tree/condition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var Condition = function (op, l, r, i, negate) {
        this.op = op.trim();
        this.lvalue = l;
        this.rvalue = r;
        this._index = i;
        this.negate = negate;
    };
    Condition.prototype = Object.assign(new node_1.default(), {
        type: 'Condition',
        accept: function (visitor) {
            this.lvalue = visitor.visit(this.lvalue);
            this.rvalue = visitor.visit(this.rvalue);
        },
        eval: function (context) {
            var result = (function (op, a, b) {
                switch (op) {
                    case 'and': return a && b;
                    case 'or': return a || b;
                    default:
                        switch (node_1.default.compare(a, b)) {
                            case -1:
                                return op === '<' || op === '=<' || op === '<=';
                            case 0:
                                return op === '=' || op === '>=' || op === '=<' || op === '<=';
                            case 1:
                                return op === '>' || op === '>=';
                            default:
                                return false;
                        }
                }
            })(this.op, this.lvalue.eval(context), this.rvalue.eval(context));
            return this.negate ? !result : result;
        }
    });
    exports.default = Condition;
},
/* less/lib/less/tree/unicode-descriptor.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var UnicodeDescriptor = function (value) {
        this.value = value;
    };
    UnicodeDescriptor.prototype = Object.assign(new node_1.default(), {
        type: 'UnicodeDescriptor'
    });
    exports.default = UnicodeDescriptor;
},
/* less/lib/less/tree/negative.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var operation_1 = tslib_1.__importDefault(require(136) /* ./operation */);
    var dimension_1 = tslib_1.__importDefault(require(137) /* ./dimension */);
    var Negative = function (node) {
        this.value = node;
    };
    Negative.prototype = Object.assign(new node_1.default(), {
        type: 'Negative',
        genCSS: function (context, output) {
            output.add('-');
            this.value.genCSS(context, output);
        },
        eval: function (context) {
            if (context.isMathOn()) {
                return (new operation_1.default('*', [new dimension_1.default(-1), this.value])).eval(context);
            }
            return new Negative(this.value.eval(context));
        }
    });
    exports.default = Negative;
},
/* less/lib/less/tree/extend.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var selector_1 = tslib_1.__importDefault(require(116) /* ./selector */);
    var Extend = function (selector, option, index, currentFileInfo, visibilityInfo) {
        this.selector = selector;
        this.option = option;
        this.object_id = Extend.next_id++;
        this.parent_ids = [this.object_id];
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.copyVisibilityInfo(visibilityInfo);
        this.allowRoot = true;
        switch (option) {
            case 'all':
                this.allowBefore = true;
                this.allowAfter = true;
                break;
            default:
                this.allowBefore = false;
                this.allowAfter = false;
                break;
        }
        this.setParent(this.selector, this);
    };
    Extend.prototype = Object.assign(new node_1.default(), {
        type: 'Extend',
        accept: function (visitor) {
            this.selector = visitor.visit(this.selector);
        },
        eval: function (context) {
            return new Extend(this.selector.eval(context), this.option, this.getIndex(), this.fileInfo(), this.visibilityInfo());
        },
        clone: function (context) {
            return new Extend(this.selector, this.option, this.getIndex(), this.fileInfo(), this.visibilityInfo());
        },
        // it concatenates (joins) all selectors in selector array
        findSelfSelectors: function (selectors) {
            var selfElements = [], i, selectorElements;
            for (i = 0; i < selectors.length; i++) {
                selectorElements = selectors[i].elements;
                // duplicate the logic in genCSS function inside the selector node.
                // future TODO - move both logics into the selector joiner visitor
                if (i > 0 && selectorElements.length && selectorElements[0].combinator.value === '') {
                    selectorElements[0].combinator.value = ' ';
                }
                selfElements = selfElements.concat(selectors[i].elements);
            }
            this.selfSelectors = [new selector_1.default(selfElements)];
            this.selfSelectors[0].copyVisibilityInfo(this.visibilityInfo());
        }
    });
    Extend.next_id = 0;
    exports.default = Extend;
},
/* less/lib/less/tree/variable-call.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var variable_1 = tslib_1.__importDefault(require(139) /* ./variable */);
    var ruleset_1 = tslib_1.__importDefault(require(125) /* ./ruleset */);
    var detached_ruleset_1 = tslib_1.__importDefault(require(135) /* ./detached-ruleset */);
    var less_error_1 = tslib_1.__importDefault(require(120) /* ../less-error */);
    var VariableCall = function (variable, index, currentFileInfo) {
        this.variable = variable;
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.allowRoot = true;
    };
    VariableCall.prototype = Object.assign(new node_1.default(), {
        type: 'VariableCall',
        eval: function (context) {
            var rules;
            var detachedRuleset = new variable_1.default(this.variable, this.getIndex(), this.fileInfo()).eval(context);
            var error = new less_error_1.default({ message: "Could not evaluate variable call " + this.variable });
            if (!detachedRuleset.ruleset) {
                if (detachedRuleset.rules) {
                    rules = detachedRuleset;
                }
                else if (Array.isArray(detachedRuleset)) {
                    rules = new ruleset_1.default('', detachedRuleset);
                }
                else if (Array.isArray(detachedRuleset.value)) {
                    rules = new ruleset_1.default('', detachedRuleset.value);
                }
                else {
                    throw error;
                }
                detachedRuleset = new detached_ruleset_1.default(rules);
            }
            if (detachedRuleset.ruleset) {
                return detachedRuleset.callEval(context);
            }
            throw error;
        }
    });
    exports.default = VariableCall;
},
/* less/lib/less/tree/namespace-value.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var variable_1 = tslib_1.__importDefault(require(139) /* ./variable */);
    var ruleset_1 = tslib_1.__importDefault(require(125) /* ./ruleset */);
    var selector_1 = tslib_1.__importDefault(require(116) /* ./selector */);
    var NamespaceValue = function (ruleCall, lookups, index, fileInfo) {
        this.value = ruleCall;
        this.lookups = lookups;
        this._index = index;
        this._fileInfo = fileInfo;
    };
    NamespaceValue.prototype = Object.assign(new node_1.default(), {
        type: 'NamespaceValue',
        eval: function (context) {
            var i, j, name, rules = this.value.eval(context);
            for (i = 0; i < this.lookups.length; i++) {
                name = this.lookups[i];
                /**
                 * Eval'd DRs return rulesets.
                 * Eval'd mixins return rules, so let's make a ruleset if we need it.
                 * We need to do this because of late parsing of values
                 */
                if (Array.isArray(rules)) {
                    rules = new ruleset_1.default([new selector_1.default()], rules);
                }
                if (name === '') {
                    rules = rules.lastDeclaration();
                }
                else if (name.charAt(0) === '@') {
                    if (name.charAt(1) === '@') {
                        name = "@" + new variable_1.default(name.substr(1)).eval(context).value;
                    }
                    if (rules.variables) {
                        rules = rules.variable(name);
                    }
                    if (!rules) {
                        throw { type: 'Name', message: "variable " + name + " not found", filename: this.fileInfo().filename,
                            index: this.getIndex() };
                    }
                }
                else {
                    if (name.substring(0, 2) === '$@') {
                        name = "$" + new variable_1.default(name.substr(1)).eval(context).value;
                    }
                    else {
                        name = name.charAt(0) === '$' ? name : "$" + name;
                    }
                    if (rules.properties) {
                        rules = rules.property(name);
                    }
                    if (!rules) {
                        throw { type: 'Name', message: "property \"" + name.substr(1) + "\" not found", filename: this.fileInfo().filename,
                            index: this.getIndex() };
                    }
                    // Properties are an array of values, since a ruleset can have multiple props.
                    // We pick the last one (the "cascaded" value)
                    rules = rules[rules.length - 1];
                }
                if (rules.value) {
                    rules = rules.eval(context).value;
                }
                if (rules.ruleset) {
                    rules = rules.ruleset.eval(context);
                }
            }
            return rules;
        }
    });
    exports.default = NamespaceValue;
},
/* less/lib/less/tree/mixin-call.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var node_1 = tslib_1.__importDefault(require(113) /* ./node */);
    var selector_1 = tslib_1.__importDefault(require(116) /* ./selector */);
    var mixin_definition_1 = tslib_1.__importDefault(require(159) /* ./mixin-definition */);
    var default_1 = tslib_1.__importDefault(require(134) /* ../functions/default */);
    var MixinCall = function (elements, args, index, currentFileInfo, important) {
        this.selector = new selector_1.default(elements);
        this.arguments = args || [];
        this._index = index;
        this._fileInfo = currentFileInfo;
        this.important = important;
        this.allowRoot = true;
        this.setParent(this.selector, this);
    };
    MixinCall.prototype = Object.assign(new node_1.default(), {
        type: 'MixinCall',
        accept: function (visitor) {
            if (this.selector) {
                this.selector = visitor.visit(this.selector);
            }
            if (this.arguments.length) {
                this.arguments = visitor.visitArray(this.arguments);
            }
        },
        eval: function (context) {
            var mixins;
            var mixin;
            var mixinPath;
            var args = [];
            var arg;
            var argValue;
            var rules = [];
            var match = false;
            var i;
            var m;
            var f;
            var isRecursive;
            var isOneFound;
            var candidates = [];
            var candidate;
            var conditionResult = [];
            var defaultResult;
            var defFalseEitherCase = -1;
            var defNone = 0;
            var defTrue = 1;
            var defFalse = 2;
            var count;
            var originalRuleset;
            var noArgumentsFilter;
            this.selector = this.selector.eval(context);
            function calcDefGroup(mixin, mixinPath) {
                var f, p, namespace;
                for (f = 0; f < 2; f++) {
                    conditionResult[f] = true;
                    default_1.default.value(f);
                    for (p = 0; p < mixinPath.length && conditionResult[f]; p++) {
                        namespace = mixinPath[p];
                        if (namespace.matchCondition) {
                            conditionResult[f] = conditionResult[f] && namespace.matchCondition(null, context);
                        }
                    }
                    if (mixin.matchCondition) {
                        conditionResult[f] = conditionResult[f] && mixin.matchCondition(args, context);
                    }
                }
                if (conditionResult[0] || conditionResult[1]) {
                    if (conditionResult[0] != conditionResult[1]) {
                        return conditionResult[1] ?
                            defTrue : defFalse;
                    }
                    return defNone;
                }
                return defFalseEitherCase;
            }
            for (i = 0; i < this.arguments.length; i++) {
                arg = this.arguments[i];
                argValue = arg.value.eval(context);
                if (arg.expand && Array.isArray(argValue.value)) {
                    argValue = argValue.value;
                    for (m = 0; m < argValue.length; m++) {
                        args.push({ value: argValue[m] });
                    }
                }
                else {
                    args.push({ name: arg.name, value: argValue });
                }
            }
            noArgumentsFilter = function (rule) { return rule.matchArgs(null, context); };
            for (i = 0; i < context.frames.length; i++) {
                if ((mixins = context.frames[i].find(this.selector, null, noArgumentsFilter)).length > 0) {
                    isOneFound = true;
                    // To make `default()` function independent of definition order we have two "subpasses" here.
                    // At first we evaluate each guard *twice* (with `default() == true` and `default() == false`),
                    // and build candidate list with corresponding flags. Then, when we know all possible matches,
                    // we make a final decision.
                    for (m = 0; m < mixins.length; m++) {
                        mixin = mixins[m].rule;
                        mixinPath = mixins[m].path;
                        isRecursive = false;
                        for (f = 0; f < context.frames.length; f++) {
                            if ((!(mixin instanceof mixin_definition_1.default)) && mixin === (context.frames[f].originalRuleset || context.frames[f])) {
                                isRecursive = true;
                                break;
                            }
                        }
                        if (isRecursive) {
                            continue;
                        }
                        if (mixin.matchArgs(args, context)) {
                            candidate = { mixin: mixin, group: calcDefGroup(mixin, mixinPath) };
                            if (candidate.group !== defFalseEitherCase) {
                                candidates.push(candidate);
                            }
                            match = true;
                        }
                    }
                    default_1.default.reset();
                    count = [0, 0, 0];
                    for (m = 0; m < candidates.length; m++) {
                        count[candidates[m].group]++;
                    }
                    if (count[defNone] > 0) {
                        defaultResult = defFalse;
                    }
                    else {
                        defaultResult = defTrue;
                        if ((count[defTrue] + count[defFalse]) > 1) {
                            throw { type: 'Runtime', message: "Ambiguous use of `default()` found when matching for `" + this.format(args) + "`", index: this.getIndex(), filename: this.fileInfo().filename };
                        }
                    }
                    for (m = 0; m < candidates.length; m++) {
                        candidate = candidates[m].group;
                        if ((candidate === defNone) || (candidate === defaultResult)) {
                            try {
                                mixin = candidates[m].mixin;
                                if (!(mixin instanceof mixin_definition_1.default)) {
                                    originalRuleset = mixin.originalRuleset || mixin;
                                    mixin = new mixin_definition_1.default('', [], mixin.rules, null, false, null, originalRuleset.visibilityInfo());
                                    mixin.originalRuleset = originalRuleset;
                                }
                                var newRules = mixin.evalCall(context, args, this.important).rules;
                                this._setVisibilityToReplacement(newRules);
                                Array.prototype.push.apply(rules, newRules);
                            }
                            catch (e) {
                                throw { message: e.message, index: this.getIndex(), filename: this.fileInfo().filename, stack: e.stack };
                            }
                        }
                    }
                    if (match) {
                        return rules;
                    }
                }
            }
            if (isOneFound) {
                throw { type: 'Runtime', message: "No matching definition was found for `" + this.format(args) + "`", index: this.getIndex(), filename: this.fileInfo().filename };
            }
            else {
                throw { type: 'Name', message: this.selector.toCSS().trim() + " is undefined", index: this.getIndex(), filename: this.fileInfo().filename };
            }
        },
        _setVisibilityToReplacement: function (replacement) {
            var i, rule;
            if (this.blocksVisibility()) {
                for (i = 0; i < replacement.length; i++) {
                    rule = replacement[i];
                    rule.addVisibilityBlock();
                }
            }
        },
        format: function (args) {
            return this.selector.toCSS().trim() + "(" + (args ? args.map(function (a) {
                var argValue = '';
                if (a.name) {
                    argValue += a.name + ":";
                }
                if (a.value.toCSS) {
                    argValue += a.value.toCSS();
                }
                else {
                    argValue += '???';
                }
                return argValue;
            }).join(', ') : '') + ")";
        }
    });
    exports.default = MixinCall;
},
/* less/lib/less/tree/mixin-definition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var selector_1 = tslib_1.__importDefault(require(116) /* ./selector */);
    var element_1 = tslib_1.__importDefault(require(117) /* ./element */);
    var ruleset_1 = tslib_1.__importDefault(require(125) /* ./ruleset */);
    var declaration_1 = tslib_1.__importDefault(require(126) /* ./declaration */);
    var detached_ruleset_1 = tslib_1.__importDefault(require(135) /* ./detached-ruleset */);
    var expression_1 = tslib_1.__importDefault(require(142) /* ./expression */);
    var contexts_1 = tslib_1.__importDefault(require(132) /* ../contexts */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    var Definition = function (name, params, rules, condition, variadic, frames, visibilityInfo) {
        this.name = name || 'anonymous mixin';
        this.selectors = [new selector_1.default([new element_1.default(null, name, false, this._index, this._fileInfo)])];
        this.params = params;
        this.condition = condition;
        this.variadic = variadic;
        this.arity = params.length;
        this.rules = rules;
        this._lookups = {};
        var optionalParameters = [];
        this.required = params.reduce(function (count, p) {
            if (!p.name || (p.name && !p.value)) {
                return count + 1;
            }
            else {
                optionalParameters.push(p.name);
                return count;
            }
        }, 0);
        this.optionalParameters = optionalParameters;
        this.frames = frames;
        this.copyVisibilityInfo(visibilityInfo);
        this.allowRoot = true;
    };
    Definition.prototype = Object.assign(new ruleset_1.default(), {
        type: 'MixinDefinition',
        evalFirst: true,
        accept: function (visitor) {
            if (this.params && this.params.length) {
                this.params = visitor.visitArray(this.params);
            }
            this.rules = visitor.visitArray(this.rules);
            if (this.condition) {
                this.condition = visitor.visit(this.condition);
            }
        },
        evalParams: function (context, mixinEnv, args, evaldArguments) {
            /* jshint boss:true */
            var frame = new ruleset_1.default(null, null);
            var varargs;
            var arg;
            var params = utils.copyArray(this.params);
            var i;
            var j;
            var val;
            var name;
            var isNamedFound;
            var argIndex;
            var argsLength = 0;
            if (mixinEnv.frames && mixinEnv.frames[0] && mixinEnv.frames[0].functionRegistry) {
                frame.functionRegistry = mixinEnv.frames[0].functionRegistry.inherit();
            }
            mixinEnv = new contexts_1.default.Eval(mixinEnv, [frame].concat(mixinEnv.frames));
            if (args) {
                args = utils.copyArray(args);
                argsLength = args.length;
                for (i = 0; i < argsLength; i++) {
                    arg = args[i];
                    if (name = (arg && arg.name)) {
                        isNamedFound = false;
                        for (j = 0; j < params.length; j++) {
                            if (!evaldArguments[j] && name === params[j].name) {
                                evaldArguments[j] = arg.value.eval(context);
                                frame.prependRule(new declaration_1.default(name, arg.value.eval(context)));
                                isNamedFound = true;
                                break;
                            }
                        }
                        if (isNamedFound) {
                            args.splice(i, 1);
                            i--;
                            continue;
                        }
                        else {
                            throw { type: 'Runtime', message: "Named argument for " + this.name + " " + args[i].name + " not found" };
                        }
                    }
                }
            }
            argIndex = 0;
            for (i = 0; i < params.length; i++) {
                if (evaldArguments[i]) {
                    continue;
                }
                arg = args && args[argIndex];
                if (name = params[i].name) {
                    if (params[i].variadic) {
                        varargs = [];
                        for (j = argIndex; j < argsLength; j++) {
                            varargs.push(args[j].value.eval(context));
                        }
                        frame.prependRule(new declaration_1.default(name, new expression_1.default(varargs).eval(context)));
                    }
                    else {
                        val = arg && arg.value;
                        if (val) {
                            // This was a mixin call, pass in a detached ruleset of it's eval'd rules
                            if (Array.isArray(val)) {
                                val = new detached_ruleset_1.default(new ruleset_1.default('', val));
                            }
                            else {
                                val = val.eval(context);
                            }
                        }
                        else if (params[i].value) {
                            val = params[i].value.eval(mixinEnv);
                            frame.resetCache();
                        }
                        else {
                            throw { type: 'Runtime', message: "wrong number of arguments for " + this.name + " (" + argsLength + " for " + this.arity + ")" };
                        }
                        frame.prependRule(new declaration_1.default(name, val));
                        evaldArguments[i] = val;
                    }
                }
                if (params[i].variadic && args) {
                    for (j = argIndex; j < argsLength; j++) {
                        evaldArguments[j] = args[j].value.eval(context);
                    }
                }
                argIndex++;
            }
            return frame;
        },
        makeImportant: function () {
            var rules = !this.rules ? this.rules : this.rules.map(function (r) {
                if (r.makeImportant) {
                    return r.makeImportant(true);
                }
                else {
                    return r;
                }
            });
            var result = new Definition(this.name, this.params, rules, this.condition, this.variadic, this.frames);
            return result;
        },
        eval: function (context) {
            return new Definition(this.name, this.params, this.rules, this.condition, this.variadic, this.frames || utils.copyArray(context.frames));
        },
        evalCall: function (context, args, important) {
            var _arguments = [];
            var mixinFrames = this.frames ? this.frames.concat(context.frames) : context.frames;
            var frame = this.evalParams(context, new contexts_1.default.Eval(context, mixinFrames), args, _arguments);
            var rules;
            var ruleset;
            frame.prependRule(new declaration_1.default('@arguments', new expression_1.default(_arguments).eval(context)));
            rules = utils.copyArray(this.rules);
            ruleset = new ruleset_1.default(null, rules);
            ruleset.originalRuleset = this;
            ruleset = ruleset.eval(new contexts_1.default.Eval(context, [this, frame].concat(mixinFrames)));
            if (important) {
                ruleset = ruleset.makeImportant();
            }
            return ruleset;
        },
        matchCondition: function (args, context) {
            if (this.condition && !this.condition.eval(new contexts_1.default.Eval(context, [this.evalParams(context, /* the parameter variables */ new contexts_1.default.Eval(context, this.frames ? this.frames.concat(context.frames) : context.frames), args, [])]
                .concat(this.frames || []) // the parent namespace/mixin frames
                .concat(context.frames)))) { // the current environment frames
                return false;
            }
            return true;
        },
        matchArgs: function (args, context) {
            var allArgsCnt = (args && args.length) || 0;
            var len;
            var optionalParameters = this.optionalParameters;
            var requiredArgsCnt = !args ? 0 : args.reduce(function (count, p) {
                if (optionalParameters.indexOf(p.name) < 0) {
                    return count + 1;
                }
                else {
                    return count;
                }
            }, 0);
            if (!this.variadic) {
                if (requiredArgsCnt < this.required) {
                    return false;
                }
                if (allArgsCnt > this.params.length) {
                    return false;
                }
            }
            else {
                if (requiredArgsCnt < (this.required - 1)) {
                    return false;
                }
            }
            // check patterns
            len = Math.min(requiredArgsCnt, this.arity);
            for (var i = 0; i < len; i++) {
                if (!this.params[i].name && !this.params[i].variadic) {
                    if (args[i].value.eval(context).toCSS() != this.params[i].value.eval(context).toCSS()) {
                        return false;
                    }
                }
            }
            return true;
        }
    });
    exports.default = Definition;
},
/* less/lib/less/environment/abstract-plugin-loader.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var function_registry_1 = tslib_1.__importDefault(require(133) /* ../functions/function-registry */);
    var less_error_1 = tslib_1.__importDefault(require(120) /* ../less-error */);
    var AbstractPluginLoader = /** @class */ (function () {
        function AbstractPluginLoader() {
            // Implemented by Node.js plugin loader
            this.require = function () {
                return null;
            };
        }
        AbstractPluginLoader.prototype.evalPlugin = function (contents, context, imports, pluginOptions, fileInfo) {
            var loader, registry, pluginObj, localModule, pluginManager, filename, result;
            pluginManager = context.pluginManager;
            if (fileInfo) {
                if (typeof fileInfo === 'string') {
                    filename = fileInfo;
                }
                else {
                    filename = fileInfo.filename;
                }
            }
            var shortname = (new this.less.FileManager()).extractUrlParts(filename).filename;
            if (filename) {
                pluginObj = pluginManager.get(filename);
                if (pluginObj) {
                    result = this.trySetOptions(pluginObj, filename, shortname, pluginOptions);
                    if (result) {
                        return result;
                    }
                    try {
                        if (pluginObj.use) {
                            pluginObj.use.call(this.context, pluginObj);
                        }
                    }
                    catch (e) {
                        e.message = e.message || 'Error during @plugin call';
                        return new less_error_1.default(e, imports, filename);
                    }
                    return pluginObj;
                }
            }
            localModule = {
                exports: {},
                pluginManager: pluginManager,
                fileInfo: fileInfo
            };
            registry = function_registry_1.default.create();
            var registerPlugin = function (obj) {
                pluginObj = obj;
            };
            try {
                loader = new Function('module', 'require', 'registerPlugin', 'functions', 'tree', 'less', 'fileInfo', contents);
                loader(localModule, this.require(filename), registerPlugin, registry, this.less.tree, this.less, fileInfo);
            }
            catch (e) {
                return new less_error_1.default(e, imports, filename);
            }
            if (!pluginObj) {
                pluginObj = localModule.exports;
            }
            pluginObj = this.validatePlugin(pluginObj, filename, shortname);
            if (pluginObj instanceof less_error_1.default) {
                return pluginObj;
            }
            if (pluginObj) {
                pluginObj.imports = imports;
                pluginObj.filename = filename;
                // For < 3.x (or unspecified minVersion) - setOptions() before install()
                if (!pluginObj.minVersion || this.compareVersion('3.0.0', pluginObj.minVersion) < 0) {
                    result = this.trySetOptions(pluginObj, filename, shortname, pluginOptions);
                    if (result) {
                        return result;
                    }
                }
                // Run on first load
                pluginManager.addPlugin(pluginObj, fileInfo.filename, registry);
                pluginObj.functions = registry.getLocalFunctions();
                // Need to call setOptions again because the pluginObj might have functions
                result = this.trySetOptions(pluginObj, filename, shortname, pluginOptions);
                if (result) {
                    return result;
                }
                // Run every @plugin call
                try {
                    if (pluginObj.use) {
                        pluginObj.use.call(this.context, pluginObj);
                    }
                }
                catch (e) {
                    e.message = e.message || 'Error during @plugin call';
                    return new less_error_1.default(e, imports, filename);
                }
            }
            else {
                return new less_error_1.default({ message: 'Not a valid plugin' }, imports, filename);
            }
            return pluginObj;
        };
        AbstractPluginLoader.prototype.trySetOptions = function (plugin, filename, name, options) {
            if (options && !plugin.setOptions) {
                return new less_error_1.default({
                    message: "Options have been provided but the plugin " + name + " does not support any options."
                });
            }
            try {
                plugin.setOptions && plugin.setOptions(options);
            }
            catch (e) {
                return new less_error_1.default(e);
            }
        };
        AbstractPluginLoader.prototype.validatePlugin = function (plugin, filename, name) {
            if (plugin) {
                // support plugins being a function
                // so that the plugin can be more usable programmatically
                if (typeof plugin === 'function') {
                    plugin = new plugin();
                }
                if (plugin.minVersion) {
                    if (this.compareVersion(plugin.minVersion, this.less.version) < 0) {
                        return new less_error_1.default({
                            message: "Plugin " + name + " requires version " + this.versionToString(plugin.minVersion)
                        });
                    }
                }
                return plugin;
            }
            return null;
        };
        AbstractPluginLoader.prototype.compareVersion = function (aVersion, bVersion) {
            if (typeof aVersion === 'string') {
                aVersion = aVersion.match(/^(\d+)\.?(\d+)?\.?(\d+)?/);
                aVersion.shift();
            }
            for (var i = 0; i < aVersion.length; i++) {
                if (aVersion[i] !== bVersion[i]) {
                    return parseInt(aVersion[i]) > parseInt(bVersion[i]) ? -1 : 1;
                }
            }
            return 0;
        };
        AbstractPluginLoader.prototype.versionToString = function (version) {
            var versionString = '';
            for (var i = 0; i < version.length; i++) {
                versionString += (versionString ? '.' : '') + version[i];
            }
            return versionString;
        };
        AbstractPluginLoader.prototype.printUsage = function (plugins) {
            for (var i = 0; i < plugins.length; i++) {
                var plugin = plugins[i];
                if (plugin.printUsage) {
                    plugin.printUsage();
                }
            }
        };
        return AbstractPluginLoader;
    }());
    exports.default = AbstractPluginLoader;
},
/* less/lib/less/visitors/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var visitor_1 = tslib_1.__importDefault(require(162) /* ./visitor */);
    var import_visitor_1 = tslib_1.__importDefault(require(163) /* ./import-visitor */);
    var set_tree_visibility_visitor_1 = tslib_1.__importDefault(require(165) /* ./set-tree-visibility-visitor */);
    var extend_visitor_1 = tslib_1.__importDefault(require(166) /* ./extend-visitor */);
    var join_selector_visitor_1 = tslib_1.__importDefault(require(167) /* ./join-selector-visitor */);
    var to_css_visitor_1 = tslib_1.__importDefault(require(168) /* ./to-css-visitor */);
    exports.default = {
        Visitor: visitor_1.default,
        ImportVisitor: import_visitor_1.default,
        MarkVisibleSelectorsVisitor: set_tree_visibility_visitor_1.default,
        ExtendVisitor: extend_visitor_1.default,
        JoinSelectorVisitor: join_selector_visitor_1.default,
        ToCSSVisitor: to_css_visitor_1.default
    };
},
/* less/lib/less/visitors/visitor.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var tree_1 = tslib_1.__importDefault(require(112) /* ../tree */);
    var _visitArgs = { visitDeeper: true };
    var _hasIndexed = false;
    function _noop(node) {
        return node;
    }
    function indexNodeTypes(parent, ticker) {
        // add .typeIndex to tree node types for lookup table
        var key, child;
        for (key in parent) {
            /* eslint guard-for-in: 0 */
            child = parent[key];
            switch (typeof child) {
                case 'function':
                    // ignore bound functions directly on tree which do not have a prototype
                    // or aren't nodes
                    if (child.prototype && child.prototype.type) {
                        child.prototype.typeIndex = ticker++;
                    }
                    break;
                case 'object':
                    ticker = indexNodeTypes(child, ticker);
                    break;
            }
        }
        return ticker;
    }
    var Visitor = /** @class */ (function () {
        function Visitor(implementation) {
            this._implementation = implementation;
            this._visitInCache = {};
            this._visitOutCache = {};
            if (!_hasIndexed) {
                indexNodeTypes(tree_1.default, 1);
                _hasIndexed = true;
            }
        }
        Visitor.prototype.visit = function (node) {
            if (!node) {
                return node;
            }
            var nodeTypeIndex = node.typeIndex;
            if (!nodeTypeIndex) {
                // MixinCall args aren't a node type?
                if (node.value && node.value.typeIndex) {
                    this.visit(node.value);
                }
                return node;
            }
            var impl = this._implementation;
            var func = this._visitInCache[nodeTypeIndex];
            var funcOut = this._visitOutCache[nodeTypeIndex];
            var visitArgs = _visitArgs;
            var fnName;
            visitArgs.visitDeeper = true;
            if (!func) {
                fnName = "visit" + node.type;
                func = impl[fnName] || _noop;
                funcOut = impl[fnName + "Out"] || _noop;
                this._visitInCache[nodeTypeIndex] = func;
                this._visitOutCache[nodeTypeIndex] = funcOut;
            }
            if (func !== _noop) {
                var newNode = func.call(impl, node, visitArgs);
                if (node && impl.isReplacing) {
                    node = newNode;
                }
            }
            if (visitArgs.visitDeeper && node) {
                if (node.length) {
                    for (var i = 0, cnt = node.length; i < cnt; i++) {
                        if (node[i].accept) {
                            node[i].accept(this);
                        }
                    }
                }
                else if (node.accept) {
                    node.accept(this);
                }
            }
            if (funcOut != _noop) {
                funcOut.call(impl, node);
            }
            return node;
        };
        Visitor.prototype.visitArray = function (nodes, nonReplacing) {
            if (!nodes) {
                return nodes;
            }
            var cnt = nodes.length;
            var i;
            // Non-replacing
            if (nonReplacing || !this._implementation.isReplacing) {
                for (i = 0; i < cnt; i++) {
                    this.visit(nodes[i]);
                }
                return nodes;
            }
            // Replacing
            var out = [];
            for (i = 0; i < cnt; i++) {
                var evald = this.visit(nodes[i]);
                if (evald === undefined) {
                    continue;
                }
                if (!evald.splice) {
                    out.push(evald);
                }
                else if (evald.length) {
                    this.flatten(evald, out);
                }
            }
            return out;
        };
        Visitor.prototype.flatten = function (arr, out) {
            if (!out) {
                out = [];
            }
            var cnt, i, item, nestedCnt, j, nestedItem;
            for (i = 0, cnt = arr.length; i < cnt; i++) {
                item = arr[i];
                if (item === undefined) {
                    continue;
                }
                if (!item.splice) {
                    out.push(item);
                    continue;
                }
                for (j = 0, nestedCnt = item.length; j < nestedCnt; j++) {
                    nestedItem = item[j];
                    if (nestedItem === undefined) {
                        continue;
                    }
                    if (!nestedItem.splice) {
                        out.push(nestedItem);
                    }
                    else if (nestedItem.length) {
                        this.flatten(nestedItem, out);
                    }
                }
            }
            return out;
        };
        return Visitor;
    }());
    exports.default = Visitor;
},
/* less/lib/less/visitors/import-visitor.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var contexts_1 = tslib_1.__importDefault(require(132) /* ../contexts */);
    var visitor_1 = tslib_1.__importDefault(require(162) /* ./visitor */);
    var import_sequencer_1 = tslib_1.__importDefault(require(164) /* ./import-sequencer */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    var ImportVisitor = function (importer, finish) {
        this._visitor = new visitor_1.default(this);
        this._importer = importer;
        this._finish = finish;
        this.context = new contexts_1.default.Eval();
        this.importCount = 0;
        this.onceFileDetectionMap = {};
        this.recursionDetector = {};
        this._sequencer = new import_sequencer_1.default(this._onSequencerEmpty.bind(this));
    };
    ImportVisitor.prototype = {
        isReplacing: false,
        run: function (root) {
            try {
                // process the contents
                this._visitor.visit(root);
            }
            catch (e) {
                this.error = e;
            }
            this.isFinished = true;
            this._sequencer.tryRun();
        },
        _onSequencerEmpty: function () {
            if (!this.isFinished) {
                return;
            }
            this._finish(this.error);
        },
        visitImport: function (importNode, visitArgs) {
            var inlineCSS = importNode.options.inline;
            if (!importNode.css || inlineCSS) {
                var context = new contexts_1.default.Eval(this.context, utils.copyArray(this.context.frames));
                var importParent = context.frames[0];
                this.importCount++;
                if (importNode.isVariableImport()) {
                    this._sequencer.addVariableImport(this.processImportNode.bind(this, importNode, context, importParent));
                }
                else {
                    this.processImportNode(importNode, context, importParent);
                }
            }
            visitArgs.visitDeeper = false;
        },
        processImportNode: function (importNode, context, importParent) {
            var evaldImportNode;
            var inlineCSS = importNode.options.inline;
            try {
                evaldImportNode = importNode.evalForImport(context);
            }
            catch (e) {
                if (!e.filename) {
                    e.index = importNode.getIndex();
                    e.filename = importNode.fileInfo().filename;
                }
                // attempt to eval properly and treat as css
                importNode.css = true;
                // if that fails, this error will be thrown
                importNode.error = e;
            }
            if (evaldImportNode && (!evaldImportNode.css || inlineCSS)) {
                if (evaldImportNode.options.multiple) {
                    context.importMultiple = true;
                }
                // try appending if we haven't determined if it is css or not
                var tryAppendLessExtension = evaldImportNode.css === undefined;
                for (var i = 0; i < importParent.rules.length; i++) {
                    if (importParent.rules[i] === importNode) {
                        importParent.rules[i] = evaldImportNode;
                        break;
                    }
                }
                var onImported = this.onImported.bind(this, evaldImportNode, context), sequencedOnImported = this._sequencer.addImport(onImported);
                this._importer.push(evaldImportNode.getPath(), tryAppendLessExtension, evaldImportNode.fileInfo(), evaldImportNode.options, sequencedOnImported);
            }
            else {
                this.importCount--;
                if (this.isFinished) {
                    this._sequencer.tryRun();
                }
            }
        },
        onImported: function (importNode, context, e, root, importedAtRoot, fullPath) {
            if (e) {
                if (!e.filename) {
                    e.index = importNode.getIndex();
                    e.filename = importNode.fileInfo().filename;
                }
                this.error = e;
            }
            var importVisitor = this, inlineCSS = importNode.options.inline, isPlugin = importNode.options.isPlugin, isOptional = importNode.options.optional, duplicateImport = importedAtRoot || fullPath in importVisitor.recursionDetector;
            if (!context.importMultiple) {
                if (duplicateImport) {
                    importNode.skip = true;
                }
                else {
                    importNode.skip = function () {
                        if (fullPath in importVisitor.onceFileDetectionMap) {
                            return true;
                        }
                        importVisitor.onceFileDetectionMap[fullPath] = true;
                        return false;
                    };
                }
            }
            if (!fullPath && isOptional) {
                importNode.skip = true;
            }
            if (root) {
                importNode.root = root;
                importNode.importedFilename = fullPath;
                if (!inlineCSS && !isPlugin && (context.importMultiple || !duplicateImport)) {
                    importVisitor.recursionDetector[fullPath] = true;
                    var oldContext = this.context;
                    this.context = context;
                    try {
                        this._visitor.visit(root);
                    }
                    catch (e) {
                        this.error = e;
                    }
                    this.context = oldContext;
                }
            }
            importVisitor.importCount--;
            if (importVisitor.isFinished) {
                importVisitor._sequencer.tryRun();
            }
        },
        visitDeclaration: function (declNode, visitArgs) {
            if (declNode.value.type === 'DetachedRuleset') {
                this.context.frames.unshift(declNode);
            }
            else {
                visitArgs.visitDeeper = false;
            }
        },
        visitDeclarationOut: function (declNode) {
            if (declNode.value.type === 'DetachedRuleset') {
                this.context.frames.shift();
            }
        },
        visitAtRule: function (atRuleNode, visitArgs) {
            this.context.frames.unshift(atRuleNode);
        },
        visitAtRuleOut: function (atRuleNode) {
            this.context.frames.shift();
        },
        visitMixinDefinition: function (mixinDefinitionNode, visitArgs) {
            this.context.frames.unshift(mixinDefinitionNode);
        },
        visitMixinDefinitionOut: function (mixinDefinitionNode) {
            this.context.frames.shift();
        },
        visitRuleset: function (rulesetNode, visitArgs) {
            this.context.frames.unshift(rulesetNode);
        },
        visitRulesetOut: function (rulesetNode) {
            this.context.frames.shift();
        },
        visitMedia: function (mediaNode, visitArgs) {
            this.context.frames.unshift(mediaNode.rules[0]);
        },
        visitMediaOut: function (mediaNode) {
            this.context.frames.shift();
        }
    };
    exports.default = ImportVisitor;
},
/* less/lib/less/visitors/import-sequencer.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var ImportSequencer = /** @class */ (function () {
        function ImportSequencer(onSequencerEmpty) {
            this.imports = [];
            this.variableImports = [];
            this._onSequencerEmpty = onSequencerEmpty;
            this._currentDepth = 0;
        }
        ImportSequencer.prototype.addImport = function (callback) {
            var importSequencer = this, importItem = {
                callback: callback,
                args: null,
                isReady: false
            };
            this.imports.push(importItem);
            return function () {
                importItem.args = Array.prototype.slice.call(arguments, 0);
                importItem.isReady = true;
                importSequencer.tryRun();
            };
        };
        ImportSequencer.prototype.addVariableImport = function (callback) {
            this.variableImports.push(callback);
        };
        ImportSequencer.prototype.tryRun = function () {
            this._currentDepth++;
            try {
                while (true) {
                    while (this.imports.length > 0) {
                        var importItem = this.imports[0];
                        if (!importItem.isReady) {
                            return;
                        }
                        this.imports = this.imports.slice(1);
                        importItem.callback.apply(null, importItem.args);
                    }
                    if (this.variableImports.length === 0) {
                        break;
                    }
                    var variableImport = this.variableImports[0];
                    this.variableImports = this.variableImports.slice(1);
                    variableImport();
                }
            }
            finally {
                this._currentDepth--;
            }
            if (this._currentDepth === 0 && this._onSequencerEmpty) {
                this._onSequencerEmpty();
            }
        };
        return ImportSequencer;
    }());
    exports.default = ImportSequencer;
},
/* less/lib/less/visitors/set-tree-visibility-visitor.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var SetTreeVisibilityVisitor = /** @class */ (function () {
        function SetTreeVisibilityVisitor(visible) {
            this.visible = visible;
        }
        SetTreeVisibilityVisitor.prototype.run = function (root) {
            this.visit(root);
        };
        SetTreeVisibilityVisitor.prototype.visitArray = function (nodes) {
            if (!nodes) {
                return nodes;
            }
            var cnt = nodes.length;
            var i;
            for (i = 0; i < cnt; i++) {
                this.visit(nodes[i]);
            }
            return nodes;
        };
        SetTreeVisibilityVisitor.prototype.visit = function (node) {
            if (!node) {
                return node;
            }
            if (node.constructor === Array) {
                return this.visitArray(node);
            }
            if (!node.blocksVisibility || node.blocksVisibility()) {
                return node;
            }
            if (this.visible) {
                node.ensureVisibility();
            }
            else {
                node.ensureInvisibility();
            }
            node.accept(this);
            return node;
        };
        return SetTreeVisibilityVisitor;
    }());
    exports.default = SetTreeVisibilityVisitor;
},
/* less/lib/less/visitors/extend-visitor.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var tree_1 = tslib_1.__importDefault(require(112) /* ../tree */);
    var visitor_1 = tslib_1.__importDefault(require(162) /* ./visitor */);
    var logger_1 = tslib_1.__importDefault(require(70) /* ../logger */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    /* jshint loopfunc:true */
    var ExtendFinderVisitor = /** @class */ (function () {
        function ExtendFinderVisitor() {
            this._visitor = new visitor_1.default(this);
            this.contexts = [];
            this.allExtendsStack = [[]];
        }
        ExtendFinderVisitor.prototype.run = function (root) {
            root = this._visitor.visit(root);
            root.allExtends = this.allExtendsStack[0];
            return root;
        };
        ExtendFinderVisitor.prototype.visitDeclaration = function (declNode, visitArgs) {
            visitArgs.visitDeeper = false;
        };
        ExtendFinderVisitor.prototype.visitMixinDefinition = function (mixinDefinitionNode, visitArgs) {
            visitArgs.visitDeeper = false;
        };
        ExtendFinderVisitor.prototype.visitRuleset = function (rulesetNode, visitArgs) {
            if (rulesetNode.root) {
                return;
            }
            var i;
            var j;
            var extend;
            var allSelectorsExtendList = [];
            var extendList;
            // get &:extend(.a); rules which apply to all selectors in this ruleset
            var rules = rulesetNode.rules, ruleCnt = rules ? rules.length : 0;
            for (i = 0; i < ruleCnt; i++) {
                if (rulesetNode.rules[i] instanceof tree_1.default.Extend) {
                    allSelectorsExtendList.push(rules[i]);
                    rulesetNode.extendOnEveryPath = true;
                }
            }
            // now find every selector and apply the extends that apply to all extends
            // and the ones which apply to an individual extend
            var paths = rulesetNode.paths;
            for (i = 0; i < paths.length; i++) {
                var selectorPath = paths[i], selector = selectorPath[selectorPath.length - 1], selExtendList = selector.extendList;
                extendList = selExtendList ? utils.copyArray(selExtendList).concat(allSelectorsExtendList)
                    : allSelectorsExtendList;
                if (extendList) {
                    extendList = extendList.map(function (allSelectorsExtend) {
                        return allSelectorsExtend.clone();
                    });
                }
                for (j = 0; j < extendList.length; j++) {
                    this.foundExtends = true;
                    extend = extendList[j];
                    extend.findSelfSelectors(selectorPath);
                    extend.ruleset = rulesetNode;
                    if (j === 0) {
                        extend.firstExtendOnThisSelectorPath = true;
                    }
                    this.allExtendsStack[this.allExtendsStack.length - 1].push(extend);
                }
            }
            this.contexts.push(rulesetNode.selectors);
        };
        ExtendFinderVisitor.prototype.visitRulesetOut = function (rulesetNode) {
            if (!rulesetNode.root) {
                this.contexts.length = this.contexts.length - 1;
            }
        };
        ExtendFinderVisitor.prototype.visitMedia = function (mediaNode, visitArgs) {
            mediaNode.allExtends = [];
            this.allExtendsStack.push(mediaNode.allExtends);
        };
        ExtendFinderVisitor.prototype.visitMediaOut = function (mediaNode) {
            this.allExtendsStack.length = this.allExtendsStack.length - 1;
        };
        ExtendFinderVisitor.prototype.visitAtRule = function (atRuleNode, visitArgs) {
            atRuleNode.allExtends = [];
            this.allExtendsStack.push(atRuleNode.allExtends);
        };
        ExtendFinderVisitor.prototype.visitAtRuleOut = function (atRuleNode) {
            this.allExtendsStack.length = this.allExtendsStack.length - 1;
        };
        return ExtendFinderVisitor;
    }());
    var ProcessExtendsVisitor = /** @class */ (function () {
        function ProcessExtendsVisitor() {
            this._visitor = new visitor_1.default(this);
        }
        ProcessExtendsVisitor.prototype.run = function (root) {
            var extendFinder = new ExtendFinderVisitor();
            this.extendIndices = {};
            extendFinder.run(root);
            if (!extendFinder.foundExtends) {
                return root;
            }
            root.allExtends = root.allExtends.concat(this.doExtendChaining(root.allExtends, root.allExtends));
            this.allExtendsStack = [root.allExtends];
            var newRoot = this._visitor.visit(root);
            this.checkExtendsForNonMatched(root.allExtends);
            return newRoot;
        };
        ProcessExtendsVisitor.prototype.checkExtendsForNonMatched = function (extendList) {
            var indices = this.extendIndices;
            extendList.filter(function (extend) {
                return !extend.hasFoundMatches && extend.parent_ids.length == 1;
            }).forEach(function (extend) {
                var selector = '_unknown_';
                try {
                    selector = extend.selector.toCSS({});
                }
                catch (_) { }
                if (!indices[extend.index + " " + selector]) {
                    indices[extend.index + " " + selector] = true;
                    logger_1.default.warn("extend '" + selector + "' has no matches");
                }
            });
        };
        ProcessExtendsVisitor.prototype.doExtendChaining = function (extendsList, extendsListTarget, iterationCount) {
            //
            // chaining is different from normal extension.. if we extend an extend then we are not just copying, altering
            // and pasting the selector we would do normally, but we are also adding an extend with the same target selector
            // this means this new extend can then go and alter other extends
            //
            // this method deals with all the chaining work - without it, extend is flat and doesn't work on other extend selectors
            // this is also the most expensive.. and a match on one selector can cause an extension of a selector we had already
            // processed if we look at each selector at a time, as is done in visitRuleset
            var extendIndex;
            var targetExtendIndex;
            var matches;
            var extendsToAdd = [];
            var newSelector;
            var extendVisitor = this;
            var selectorPath;
            var extend;
            var targetExtend;
            var newExtend;
            iterationCount = iterationCount || 0;
            // loop through comparing every extend with every target extend.
            // a target extend is the one on the ruleset we are looking at copy/edit/pasting in place
            // e.g.  .a:extend(.b) {}  and .b:extend(.c) {} then the first extend extends the second one
            // and the second is the target.
            // the separation into two lists allows us to process a subset of chains with a bigger set, as is the
            // case when processing media queries
            for (extendIndex = 0; extendIndex < extendsList.length; extendIndex++) {
                for (targetExtendIndex = 0; targetExtendIndex < extendsListTarget.length; targetExtendIndex++) {
                    extend = extendsList[extendIndex];
                    targetExtend = extendsListTarget[targetExtendIndex];
                    // look for circular references
                    if (extend.parent_ids.indexOf(targetExtend.object_id) >= 0) {
                        continue;
                    }
                    // find a match in the target extends self selector (the bit before :extend)
                    selectorPath = [targetExtend.selfSelectors[0]];
                    matches = extendVisitor.findMatch(extend, selectorPath);
                    if (matches.length) {
                        extend.hasFoundMatches = true;
                        // we found a match, so for each self selector..
                        extend.selfSelectors.forEach(function (selfSelector) {
                            var info = targetExtend.visibilityInfo();
                            // process the extend as usual
                            newSelector = extendVisitor.extendSelector(matches, selectorPath, selfSelector, extend.isVisible());
                            // but now we create a new extend from it
                            newExtend = new (tree_1.default.Extend)(targetExtend.selector, targetExtend.option, 0, targetExtend.fileInfo(), info);
                            newExtend.selfSelectors = newSelector;
                            // add the extend onto the list of extends for that selector
                            newSelector[newSelector.length - 1].extendList = [newExtend];
                            // record that we need to add it.
                            extendsToAdd.push(newExtend);
                            newExtend.ruleset = targetExtend.ruleset;
                            // remember its parents for circular references
                            newExtend.parent_ids = newExtend.parent_ids.concat(targetExtend.parent_ids, extend.parent_ids);
                            // only process the selector once.. if we have :extend(.a,.b) then multiple
                            // extends will look at the same selector path, so when extending
                            // we know that any others will be duplicates in terms of what is added to the css
                            if (targetExtend.firstExtendOnThisSelectorPath) {
                                newExtend.firstExtendOnThisSelectorPath = true;
                                targetExtend.ruleset.paths.push(newSelector);
                            }
                        });
                    }
                }
            }
            if (extendsToAdd.length) {
                // try to detect circular references to stop a stack overflow.
                // may no longer be needed.
                this.extendChainCount++;
                if (iterationCount > 100) {
                    var selectorOne = '{unable to calculate}';
                    var selectorTwo = '{unable to calculate}';
                    try {
                        selectorOne = extendsToAdd[0].selfSelectors[0].toCSS();
                        selectorTwo = extendsToAdd[0].selector.toCSS();
                    }
                    catch (e) { }
                    throw { message: "extend circular reference detected. One of the circular extends is currently:" + selectorOne + ":extend(" + selectorTwo + ")" };
                }
                // now process the new extends on the existing rules so that we can handle a extending b extending c extending
                // d extending e...
                return extendsToAdd.concat(extendVisitor.doExtendChaining(extendsToAdd, extendsListTarget, iterationCount + 1));
            }
            else {
                return extendsToAdd;
            }
        };
        ProcessExtendsVisitor.prototype.visitDeclaration = function (ruleNode, visitArgs) {
            visitArgs.visitDeeper = false;
        };
        ProcessExtendsVisitor.prototype.visitMixinDefinition = function (mixinDefinitionNode, visitArgs) {
            visitArgs.visitDeeper = false;
        };
        ProcessExtendsVisitor.prototype.visitSelector = function (selectorNode, visitArgs) {
            visitArgs.visitDeeper = false;
        };
        ProcessExtendsVisitor.prototype.visitRuleset = function (rulesetNode, visitArgs) {
            if (rulesetNode.root) {
                return;
            }
            var matches;
            var pathIndex;
            var extendIndex;
            var allExtends = this.allExtendsStack[this.allExtendsStack.length - 1];
            var selectorsToAdd = [];
            var extendVisitor = this;
            var selectorPath;
            // look at each selector path in the ruleset, find any extend matches and then copy, find and replace
            for (extendIndex = 0; extendIndex < allExtends.length; extendIndex++) {
                for (pathIndex = 0; pathIndex < rulesetNode.paths.length; pathIndex++) {
                    selectorPath = rulesetNode.paths[pathIndex];
                    // extending extends happens initially, before the main pass
                    if (rulesetNode.extendOnEveryPath) {
                        continue;
                    }
                    var extendList = selectorPath[selectorPath.length - 1].extendList;
                    if (extendList && extendList.length) {
                        continue;
                    }
                    matches = this.findMatch(allExtends[extendIndex], selectorPath);
                    if (matches.length) {
                        allExtends[extendIndex].hasFoundMatches = true;
                        allExtends[extendIndex].selfSelectors.forEach(function (selfSelector) {
                            var extendedSelectors;
                            extendedSelectors = extendVisitor.extendSelector(matches, selectorPath, selfSelector, allExtends[extendIndex].isVisible());
                            selectorsToAdd.push(extendedSelectors);
                        });
                    }
                }
            }
            rulesetNode.paths = rulesetNode.paths.concat(selectorsToAdd);
        };
        ProcessExtendsVisitor.prototype.findMatch = function (extend, haystackSelectorPath) {
            //
            // look through the haystack selector path to try and find the needle - extend.selector
            // returns an array of selector matches that can then be replaced
            //
            var haystackSelectorIndex;
            var hackstackSelector;
            var hackstackElementIndex;
            var haystackElement;
            var targetCombinator;
            var i;
            var extendVisitor = this;
            var needleElements = extend.selector.elements;
            var potentialMatches = [];
            var potentialMatch;
            var matches = [];
            // loop through the haystack elements
            for (haystackSelectorIndex = 0; haystackSelectorIndex < haystackSelectorPath.length; haystackSelectorIndex++) {
                hackstackSelector = haystackSelectorPath[haystackSelectorIndex];
                for (hackstackElementIndex = 0; hackstackElementIndex < hackstackSelector.elements.length; hackstackElementIndex++) {
                    haystackElement = hackstackSelector.elements[hackstackElementIndex];
                    // if we allow elements before our match we can add a potential match every time. otherwise only at the first element.
                    if (extend.allowBefore || (haystackSelectorIndex === 0 && hackstackElementIndex === 0)) {
                        potentialMatches.push({ pathIndex: haystackSelectorIndex, index: hackstackElementIndex, matched: 0,
                            initialCombinator: haystackElement.combinator });
                    }
                    for (i = 0; i < potentialMatches.length; i++) {
                        potentialMatch = potentialMatches[i];
                        // selectors add " " onto the first element. When we use & it joins the selectors together, but if we don't
                        // then each selector in haystackSelectorPath has a space before it added in the toCSS phase. so we need to
                        // work out what the resulting combinator will be
                        targetCombinator = haystackElement.combinator.value;
                        if (targetCombinator === '' && hackstackElementIndex === 0) {
                            targetCombinator = ' ';
                        }
                        // if we don't match, null our match to indicate failure
                        if (!extendVisitor.isElementValuesEqual(needleElements[potentialMatch.matched].value, haystackElement.value) ||
                            (potentialMatch.matched > 0 && needleElements[potentialMatch.matched].combinator.value !== targetCombinator)) {
                            potentialMatch = null;
                        }
                        else {
                            potentialMatch.matched++;
                        }
                        // if we are still valid and have finished, test whether we have elements after and whether these are allowed
                        if (potentialMatch) {
                            potentialMatch.finished = potentialMatch.matched === needleElements.length;
                            if (potentialMatch.finished &&
                                (!extend.allowAfter &&
                                    (hackstackElementIndex + 1 < hackstackSelector.elements.length || haystackSelectorIndex + 1 < haystackSelectorPath.length))) {
                                potentialMatch = null;
                            }
                        }
                        // if null we remove, if not, we are still valid, so either push as a valid match or continue
                        if (potentialMatch) {
                            if (potentialMatch.finished) {
                                potentialMatch.length = needleElements.length;
                                potentialMatch.endPathIndex = haystackSelectorIndex;
                                potentialMatch.endPathElementIndex = hackstackElementIndex + 1; // index after end of match
                                potentialMatches.length = 0; // we don't allow matches to overlap, so start matching again
                                matches.push(potentialMatch);
                            }
                        }
                        else {
                            potentialMatches.splice(i, 1);
                            i--;
                        }
                    }
                }
            }
            return matches;
        };
        ProcessExtendsVisitor.prototype.isElementValuesEqual = function (elementValue1, elementValue2) {
            if (typeof elementValue1 === 'string' || typeof elementValue2 === 'string') {
                return elementValue1 === elementValue2;
            }
            if (elementValue1 instanceof tree_1.default.Attribute) {
                if (elementValue1.op !== elementValue2.op || elementValue1.key !== elementValue2.key) {
                    return false;
                }
                if (!elementValue1.value || !elementValue2.value) {
                    if (elementValue1.value || elementValue2.value) {
                        return false;
                    }
                    return true;
                }
                elementValue1 = elementValue1.value.value || elementValue1.value;
                elementValue2 = elementValue2.value.value || elementValue2.value;
                return elementValue1 === elementValue2;
            }
            elementValue1 = elementValue1.value;
            elementValue2 = elementValue2.value;
            if (elementValue1 instanceof tree_1.default.Selector) {
                if (!(elementValue2 instanceof tree_1.default.Selector) || elementValue1.elements.length !== elementValue2.elements.length) {
                    return false;
                }
                for (var i = 0; i < elementValue1.elements.length; i++) {
                    if (elementValue1.elements[i].combinator.value !== elementValue2.elements[i].combinator.value) {
                        if (i !== 0 || (elementValue1.elements[i].combinator.value || ' ') !== (elementValue2.elements[i].combinator.value || ' ')) {
                            return false;
                        }
                    }
                    if (!this.isElementValuesEqual(elementValue1.elements[i].value, elementValue2.elements[i].value)) {
                        return false;
                    }
                }
                return true;
            }
            return false;
        };
        ProcessExtendsVisitor.prototype.extendSelector = function (matches, selectorPath, replacementSelector, isVisible) {
            // for a set of matches, replace each match with the replacement selector
            var currentSelectorPathIndex = 0, currentSelectorPathElementIndex = 0, path = [], matchIndex, selector, firstElement, match, newElements;
            for (matchIndex = 0; matchIndex < matches.length; matchIndex++) {
                match = matches[matchIndex];
                selector = selectorPath[match.pathIndex];
                firstElement = new tree_1.default.Element(match.initialCombinator, replacementSelector.elements[0].value, replacementSelector.elements[0].isVariable, replacementSelector.elements[0].getIndex(), replacementSelector.elements[0].fileInfo());
                if (match.pathIndex > currentSelectorPathIndex && currentSelectorPathElementIndex > 0) {
                    path[path.length - 1].elements = path[path.length - 1]
                        .elements.concat(selectorPath[currentSelectorPathIndex].elements.slice(currentSelectorPathElementIndex));
                    currentSelectorPathElementIndex = 0;
                    currentSelectorPathIndex++;
                }
                newElements = selector.elements
                    .slice(currentSelectorPathElementIndex, match.index)
                    .concat([firstElement])
                    .concat(replacementSelector.elements.slice(1));
                if (currentSelectorPathIndex === match.pathIndex && matchIndex > 0) {
                    path[path.length - 1].elements =
                        path[path.length - 1].elements.concat(newElements);
                }
                else {
                    path = path.concat(selectorPath.slice(currentSelectorPathIndex, match.pathIndex));
                    path.push(new tree_1.default.Selector(newElements));
                }
                currentSelectorPathIndex = match.endPathIndex;
                currentSelectorPathElementIndex = match.endPathElementIndex;
                if (currentSelectorPathElementIndex >= selectorPath[currentSelectorPathIndex].elements.length) {
                    currentSelectorPathElementIndex = 0;
                    currentSelectorPathIndex++;
                }
            }
            if (currentSelectorPathIndex < selectorPath.length && currentSelectorPathElementIndex > 0) {
                path[path.length - 1].elements = path[path.length - 1]
                    .elements.concat(selectorPath[currentSelectorPathIndex].elements.slice(currentSelectorPathElementIndex));
                currentSelectorPathIndex++;
            }
            path = path.concat(selectorPath.slice(currentSelectorPathIndex, selectorPath.length));
            path = path.map(function (currentValue) {
                // we can re-use elements here, because the visibility property matters only for selectors
                var derived = currentValue.createDerived(currentValue.elements);
                if (isVisible) {
                    derived.ensureVisibility();
                }
                else {
                    derived.ensureInvisibility();
                }
                return derived;
            });
            return path;
        };
        ProcessExtendsVisitor.prototype.visitMedia = function (mediaNode, visitArgs) {
            var newAllExtends = mediaNode.allExtends.concat(this.allExtendsStack[this.allExtendsStack.length - 1]);
            newAllExtends = newAllExtends.concat(this.doExtendChaining(newAllExtends, mediaNode.allExtends));
            this.allExtendsStack.push(newAllExtends);
        };
        ProcessExtendsVisitor.prototype.visitMediaOut = function (mediaNode) {
            var lastIndex = this.allExtendsStack.length - 1;
            this.allExtendsStack.length = lastIndex;
        };
        ProcessExtendsVisitor.prototype.visitAtRule = function (atRuleNode, visitArgs) {
            var newAllExtends = atRuleNode.allExtends.concat(this.allExtendsStack[this.allExtendsStack.length - 1]);
            newAllExtends = newAllExtends.concat(this.doExtendChaining(newAllExtends, atRuleNode.allExtends));
            this.allExtendsStack.push(newAllExtends);
        };
        ProcessExtendsVisitor.prototype.visitAtRuleOut = function (atRuleNode) {
            var lastIndex = this.allExtendsStack.length - 1;
            this.allExtendsStack.length = lastIndex;
        };
        return ProcessExtendsVisitor;
    }());
    exports.default = ProcessExtendsVisitor;
},
/* less/lib/less/visitors/join-selector-visitor.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var visitor_1 = tslib_1.__importDefault(require(162) /* ./visitor */);
    var JoinSelectorVisitor = /** @class */ (function () {
        function JoinSelectorVisitor() {
            this.contexts = [[]];
            this._visitor = new visitor_1.default(this);
        }
        JoinSelectorVisitor.prototype.run = function (root) {
            return this._visitor.visit(root);
        };
        JoinSelectorVisitor.prototype.visitDeclaration = function (declNode, visitArgs) {
            visitArgs.visitDeeper = false;
        };
        JoinSelectorVisitor.prototype.visitMixinDefinition = function (mixinDefinitionNode, visitArgs) {
            visitArgs.visitDeeper = false;
        };
        JoinSelectorVisitor.prototype.visitRuleset = function (rulesetNode, visitArgs) {
            var context = this.contexts[this.contexts.length - 1];
            var paths = [];
            var selectors;
            this.contexts.push(paths);
            if (!rulesetNode.root) {
                selectors = rulesetNode.selectors;
                if (selectors) {
                    selectors = selectors.filter(function (selector) { return selector.getIsOutput(); });
                    rulesetNode.selectors = selectors.length ? selectors : (selectors = null);
                    if (selectors) {
                        rulesetNode.joinSelectors(paths, context, selectors);
                    }
                }
                if (!selectors) {
                    rulesetNode.rules = null;
                }
                rulesetNode.paths = paths;
            }
        };
        JoinSelectorVisitor.prototype.visitRulesetOut = function (rulesetNode) {
            this.contexts.length = this.contexts.length - 1;
        };
        JoinSelectorVisitor.prototype.visitMedia = function (mediaNode, visitArgs) {
            var context = this.contexts[this.contexts.length - 1];
            mediaNode.rules[0].root = (context.length === 0 || context[0].multiMedia);
        };
        JoinSelectorVisitor.prototype.visitAtRule = function (atRuleNode, visitArgs) {
            var context = this.contexts[this.contexts.length - 1];
            if (atRuleNode.rules && atRuleNode.rules.length) {
                atRuleNode.rules[0].root = (atRuleNode.isRooted || context.length === 0 || null);
            }
        };
        return JoinSelectorVisitor;
    }());
    exports.default = JoinSelectorVisitor;
},
/* less/lib/less/visitors/to-css-visitor.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var tree_1 = tslib_1.__importDefault(require(112) /* ../tree */);
    var visitor_1 = tslib_1.__importDefault(require(162) /* ./visitor */);
    var CSSVisitorUtils = /** @class */ (function () {
        function CSSVisitorUtils(context) {
            this._visitor = new visitor_1.default(this);
            this._context = context;
        }
        CSSVisitorUtils.prototype.containsSilentNonBlockedChild = function (bodyRules) {
            var rule;
            if (!bodyRules) {
                return false;
            }
            for (var r = 0; r < bodyRules.length; r++) {
                rule = bodyRules[r];
                if (rule.isSilent && rule.isSilent(this._context) && !rule.blocksVisibility()) {
                    // the atrule contains something that was referenced (likely by extend)
                    // therefore it needs to be shown in output too
                    return true;
                }
            }
            return false;
        };
        CSSVisitorUtils.prototype.keepOnlyVisibleChilds = function (owner) {
            if (owner && owner.rules) {
                owner.rules = owner.rules.filter(function (thing) { return thing.isVisible(); });
            }
        };
        CSSVisitorUtils.prototype.isEmpty = function (owner) {
            return (owner && owner.rules)
                ? (owner.rules.length === 0) : true;
        };
        CSSVisitorUtils.prototype.hasVisibleSelector = function (rulesetNode) {
            return (rulesetNode && rulesetNode.paths)
                ? (rulesetNode.paths.length > 0) : false;
        };
        CSSVisitorUtils.prototype.resolveVisibility = function (node, originalRules) {
            if (!node.blocksVisibility()) {
                if (this.isEmpty(node) && !this.containsSilentNonBlockedChild(originalRules)) {
                    return;
                }
                return node;
            }
            var compiledRulesBody = node.rules[0];
            this.keepOnlyVisibleChilds(compiledRulesBody);
            if (this.isEmpty(compiledRulesBody)) {
                return;
            }
            node.ensureVisibility();
            node.removeVisibilityBlock();
            return node;
        };
        CSSVisitorUtils.prototype.isVisibleRuleset = function (rulesetNode) {
            if (rulesetNode.firstRoot) {
                return true;
            }
            if (this.isEmpty(rulesetNode)) {
                return false;
            }
            if (!rulesetNode.root && !this.hasVisibleSelector(rulesetNode)) {
                return false;
            }
            return true;
        };
        return CSSVisitorUtils;
    }());
    var ToCSSVisitor = function (context) {
        this._visitor = new visitor_1.default(this);
        this._context = context;
        this.utils = new CSSVisitorUtils(context);
    };
    ToCSSVisitor.prototype = {
        isReplacing: true,
        run: function (root) {
            return this._visitor.visit(root);
        },
        visitDeclaration: function (declNode, visitArgs) {
            if (declNode.blocksVisibility() || declNode.variable) {
                return;
            }
            return declNode;
        },
        visitMixinDefinition: function (mixinNode, visitArgs) {
            // mixin definitions do not get eval'd - this means they keep state
            // so we have to clear that state here so it isn't used if toCSS is called twice
            mixinNode.frames = [];
        },
        visitExtend: function (extendNode, visitArgs) {
        },
        visitComment: function (commentNode, visitArgs) {
            if (commentNode.blocksVisibility() || commentNode.isSilent(this._context)) {
                return;
            }
            return commentNode;
        },
        visitMedia: function (mediaNode, visitArgs) {
            var originalRules = mediaNode.rules[0].rules;
            mediaNode.accept(this._visitor);
            visitArgs.visitDeeper = false;
            return this.utils.resolveVisibility(mediaNode, originalRules);
        },
        visitImport: function (importNode, visitArgs) {
            if (importNode.blocksVisibility()) {
                return;
            }
            return importNode;
        },
        visitAtRule: function (atRuleNode, visitArgs) {
            if (atRuleNode.rules && atRuleNode.rules.length) {
                return this.visitAtRuleWithBody(atRuleNode, visitArgs);
            }
            else {
                return this.visitAtRuleWithoutBody(atRuleNode, visitArgs);
            }
        },
        visitAnonymous: function (anonymousNode, visitArgs) {
            if (!anonymousNode.blocksVisibility()) {
                anonymousNode.accept(this._visitor);
                return anonymousNode;
            }
        },
        visitAtRuleWithBody: function (atRuleNode, visitArgs) {
            // if there is only one nested ruleset and that one has no path, then it is
            // just fake ruleset
            function hasFakeRuleset(atRuleNode) {
                var bodyRules = atRuleNode.rules;
                return bodyRules.length === 1 && (!bodyRules[0].paths || bodyRules[0].paths.length === 0);
            }
            function getBodyRules(atRuleNode) {
                var nodeRules = atRuleNode.rules;
                if (hasFakeRuleset(atRuleNode)) {
                    return nodeRules[0].rules;
                }
                return nodeRules;
            }
            // it is still true that it is only one ruleset in array
            // this is last such moment
            // process childs
            var originalRules = getBodyRules(atRuleNode);
            atRuleNode.accept(this._visitor);
            visitArgs.visitDeeper = false;
            if (!this.utils.isEmpty(atRuleNode)) {
                this._mergeRules(atRuleNode.rules[0].rules);
            }
            return this.utils.resolveVisibility(atRuleNode, originalRules);
        },
        visitAtRuleWithoutBody: function (atRuleNode, visitArgs) {
            if (atRuleNode.blocksVisibility()) {
                return;
            }
            if (atRuleNode.name === '@charset') {
                // Only output the debug info together with subsequent @charset definitions
                // a comment (or @media statement) before the actual @charset atrule would
                // be considered illegal css as it has to be on the first line
                if (this.charset) {
                    if (atRuleNode.debugInfo) {
                        var comment = new tree_1.default.Comment("/* " + atRuleNode.toCSS(this._context).replace(/\n/g, '') + " */\n");
                        comment.debugInfo = atRuleNode.debugInfo;
                        return this._visitor.visit(comment);
                    }
                    return;
                }
                this.charset = true;
            }
            return atRuleNode;
        },
        checkValidNodes: function (rules, isRoot) {
            if (!rules) {
                return;
            }
            for (var i = 0; i < rules.length; i++) {
                var ruleNode = rules[i];
                if (isRoot && ruleNode instanceof tree_1.default.Declaration && !ruleNode.variable) {
                    throw { message: 'Properties must be inside selector blocks. They cannot be in the root',
                        index: ruleNode.getIndex(), filename: ruleNode.fileInfo() && ruleNode.fileInfo().filename };
                }
                if (ruleNode instanceof tree_1.default.Call) {
                    throw { message: "Function '" + ruleNode.name + "' did not return a root node", index: ruleNode.getIndex(), filename: ruleNode.fileInfo() && ruleNode.fileInfo().filename };
                }
                if (ruleNode.type && !ruleNode.allowRoot) {
                    throw { message: ruleNode.type + " node returned by a function is not valid here", index: ruleNode.getIndex(), filename: ruleNode.fileInfo() && ruleNode.fileInfo().filename };
                }
            }
        },
        visitRuleset: function (rulesetNode, visitArgs) {
            // at this point rulesets are nested into each other
            var rule;
            var rulesets = [];
            this.checkValidNodes(rulesetNode.rules, rulesetNode.firstRoot);
            if (!rulesetNode.root) {
                // remove invisible paths
                this._compileRulesetPaths(rulesetNode);
                // remove rulesets from this ruleset body and compile them separately
                var nodeRules = rulesetNode.rules;
                var nodeRuleCnt = nodeRules ? nodeRules.length : 0;
                for (var i = 0; i < nodeRuleCnt;) {
                    rule = nodeRules[i];
                    if (rule && rule.rules) {
                        // visit because we are moving them out from being a child
                        rulesets.push(this._visitor.visit(rule));
                        nodeRules.splice(i, 1);
                        nodeRuleCnt--;
                        continue;
                    }
                    i++;
                }
                // accept the visitor to remove rules and refactor itself
                // then we can decide nogw whether we want it or not
                // compile body
                if (nodeRuleCnt > 0) {
                    rulesetNode.accept(this._visitor);
                }
                else {
                    rulesetNode.rules = null;
                }
                visitArgs.visitDeeper = false;
            }
            else { // if (! rulesetNode.root) {
                rulesetNode.accept(this._visitor);
                visitArgs.visitDeeper = false;
            }
            if (rulesetNode.rules) {
                this._mergeRules(rulesetNode.rules);
                this._removeDuplicateRules(rulesetNode.rules);
            }
            // now decide whether we keep the ruleset
            if (this.utils.isVisibleRuleset(rulesetNode)) {
                rulesetNode.ensureVisibility();
                rulesets.splice(0, 0, rulesetNode);
            }
            if (rulesets.length === 1) {
                return rulesets[0];
            }
            return rulesets;
        },
        _compileRulesetPaths: function (rulesetNode) {
            if (rulesetNode.paths) {
                rulesetNode.paths = rulesetNode.paths
                    .filter(function (p) {
                    var i;
                    if (p[0].elements[0].combinator.value === ' ') {
                        p[0].elements[0].combinator = new (tree_1.default.Combinator)('');
                    }
                    for (i = 0; i < p.length; i++) {
                        if (p[i].isVisible() && p[i].getIsOutput()) {
                            return true;
                        }
                    }
                    return false;
                });
            }
        },
        _removeDuplicateRules: function (rules) {
            if (!rules) {
                return;
            }
            // remove duplicates
            var ruleCache = {};
            var ruleList;
            var rule;
            var i;
            for (i = rules.length - 1; i >= 0; i--) {
                rule = rules[i];
                if (rule instanceof tree_1.default.Declaration) {
                    if (!ruleCache[rule.name]) {
                        ruleCache[rule.name] = rule;
                    }
                    else {
                        ruleList = ruleCache[rule.name];
                        if (ruleList instanceof tree_1.default.Declaration) {
                            ruleList = ruleCache[rule.name] = [ruleCache[rule.name].toCSS(this._context)];
                        }
                        var ruleCSS = rule.toCSS(this._context);
                        if (ruleList.indexOf(ruleCSS) !== -1) {
                            rules.splice(i, 1);
                        }
                        else {
                            ruleList.push(ruleCSS);
                        }
                    }
                }
            }
        },
        _mergeRules: function (rules) {
            if (!rules) {
                return;
            }
            var groups = {};
            var groupsArr = [];
            for (var i = 0; i < rules.length; i++) {
                var rule = rules[i];
                if (rule.merge) {
                    var key = rule.name;
                    groups[key] ? rules.splice(i--, 1) :
                        groupsArr.push(groups[key] = []);
                    groups[key].push(rule);
                }
            }
            groupsArr.forEach(function (group) {
                if (group.length > 0) {
                    var result_1 = group[0];
                    var space_1 = [];
                    var comma_1 = [new tree_1.default.Expression(space_1)];
                    group.forEach(function (rule) {
                        if ((rule.merge === '+') && (space_1.length > 0)) {
                            comma_1.push(new tree_1.default.Expression(space_1 = []));
                        }
                        space_1.push(rule.value);
                        result_1.important = result_1.important || rule.important;
                    });
                    result_1.value = new tree_1.default.Value(comma_1);
                }
            });
        }
    };
    exports.default = ToCSSVisitor;
},
/* less/lib/less/parser/parser.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var less_error_1 = tslib_1.__importDefault(require(120) /* ../less-error */);
    var tree_1 = tslib_1.__importDefault(require(112) /* ../tree */);
    var visitors_1 = tslib_1.__importDefault(require(161) /* ../visitors */);
    var parser_input_1 = tslib_1.__importDefault(require(170) /* ./parser-input */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    var function_registry_1 = tslib_1.__importDefault(require(133) /* ../functions/function-registry */);
    //
    // less.js - parser
    //
    //    A relatively straight-forward predictive parser.
    //    There is no tokenization/lexing stage, the input is parsed
    //    in one sweep.
    //
    //    To make the parser fast enough to run in the browser, several
    //    optimization had to be made:
    //
    //    - Matching and slicing on a huge input is often cause of slowdowns.
    //      The solution is to chunkify the input into smaller strings.
    //      The chunks are stored in the `chunks` var,
    //      `j` holds the current chunk index, and `currentPos` holds
    //      the index of the current chunk in relation to `input`.
    //      This gives us an almost 4x speed-up.
    //
    //    - In many cases, we don't need to match individual tokens;
    //      for example, if a value doesn't hold any variables, operations
    //      or dynamic references, the parser can effectively 'skip' it,
    //      treating it as a literal.
    //      An example would be '1px solid #000' - which evaluates to itself,
    //      we don't need to know what the individual components are.
    //      The drawback, of course is that you don't get the benefits of
    //      syntax-checking on the CSS. This gives us a 50% speed-up in the parser,
    //      and a smaller speed-up in the code-gen.
    //
    //
    //    Token matching is done with the `$` function, which either takes
    //    a terminal string or regexp, or a non-terminal function to call.
    //    It also takes care of moving all the indices forwards.
    //
    var Parser = function Parser(context, imports, fileInfo) {
        var parsers;
        var parserInput = parser_input_1.default();
        function error(msg, type) {
            throw new less_error_1.default({
                index: parserInput.i,
                filename: fileInfo.filename,
                type: type || 'Syntax',
                message: msg
            }, imports);
        }
        function expect(arg, msg) {
            // some older browsers return typeof 'function' for RegExp
            var result = (arg instanceof Function) ? arg.call(parsers) : parserInput.$re(arg);
            if (result) {
                return result;
            }
            error(msg || (typeof arg === 'string'
                ? "expected '" + arg + "' got '" + parserInput.currentChar() + "'"
                : 'unexpected token'));
        }
        // Specialization of expect()
        function expectChar(arg, msg) {
            if (parserInput.$char(arg)) {
                return arg;
            }
            error(msg || "expected '" + arg + "' got '" + parserInput.currentChar() + "'");
        }
        function getDebugInfo(index) {
            var filename = fileInfo.filename;
            return {
                lineNumber: utils.getLocation(index, parserInput.getInput()).line + 1,
                fileName: filename
            };
        }
        /**
         *  Used after initial parsing to create nodes on the fly
         *
         *  @param {String} str          - string to parse
         *  @param {Array}  parseList    - array of parsers to run input through e.g. ["value", "important"]
         *  @param {Number} currentIndex - start number to begin indexing
         *  @param {Object} fileInfo     - fileInfo to attach to created nodes
         */
        function parseNode(str, parseList, currentIndex, fileInfo, callback) {
            var result;
            var returnNodes = [];
            var parser = parserInput;
            try {
                parser.start(str, false, function fail(msg, index) {
                    callback({
                        message: msg,
                        index: index + currentIndex
                    });
                });
                for (var x = 0, p = void 0, i = void 0; (p = parseList[x]); x++) {
                    i = parser.i;
                    result = parsers[p]();
                    if (result) {
                        try {
                            result._index = i + currentIndex;
                            result._fileInfo = fileInfo;
                        }
                        catch (e) { }
                        returnNodes.push(result);
                    }
                    else {
                        returnNodes.push(null);
                    }
                }
                var endInfo = parser.end();
                if (endInfo.isFinished) {
                    callback(null, returnNodes);
                }
                else {
                    callback(true, null);
                }
            }
            catch (e) {
                throw new less_error_1.default({
                    index: e.index + currentIndex,
                    message: e.message
                }, imports, fileInfo.filename);
            }
        }
        //
        // The Parser
        //
        return {
            parserInput: parserInput,
            imports: imports,
            fileInfo: fileInfo,
            parseNode: parseNode,
            //
            // Parse an input string into an abstract syntax tree,
            // @param str A string containing 'less' markup
            // @param callback call `callback` when done.
            // @param [additionalData] An optional map which can contains vars - a map (key, value) of variables to apply
            //
            parse: function (str, callback, additionalData) {
                var root;
                var error = null;
                var globalVars;
                var modifyVars;
                var ignored;
                var preText = '';
                globalVars = (additionalData && additionalData.globalVars) ? Parser.serializeVars(additionalData.globalVars) + "\n" : '';
                modifyVars = (additionalData && additionalData.modifyVars) ? "\n" + Parser.serializeVars(additionalData.modifyVars) : '';
                if (context.pluginManager) {
                    var preProcessors = context.pluginManager.getPreProcessors();
                    for (var i = 0; i < preProcessors.length; i++) {
                        str = preProcessors[i].process(str, { context: context, imports: imports, fileInfo: fileInfo });
                    }
                }
                if (globalVars || (additionalData && additionalData.banner)) {
                    preText = ((additionalData && additionalData.banner) ? additionalData.banner : '') + globalVars;
                    ignored = imports.contentsIgnoredChars;
                    ignored[fileInfo.filename] = ignored[fileInfo.filename] || 0;
                    ignored[fileInfo.filename] += preText.length;
                }
                str = str.replace(/\r\n?/g, '\n');
                // Remove potential UTF Byte Order Mark
                str = preText + str.replace(/^\uFEFF/, '') + modifyVars;
                imports.contents[fileInfo.filename] = str;
                // Start with the primary rule.
                // The whole syntax tree is held under a Ruleset node,
                // with the `root` property set to true, so no `{}` are
                // output. The callback is called when the input is parsed.
                try {
                    parserInput.start(str, context.chunkInput, function fail(msg, index) {
                        throw new less_error_1.default({
                            index: index,
                            type: 'Parse',
                            message: msg,
                            filename: fileInfo.filename
                        }, imports);
                    });
                    tree_1.default.Node.prototype.parse = this;
                    root = new tree_1.default.Ruleset(null, this.parsers.primary());
                    tree_1.default.Node.prototype.rootNode = root;
                    root.root = true;
                    root.firstRoot = true;
                    root.functionRegistry = function_registry_1.default.inherit();
                }
                catch (e) {
                    return callback(new less_error_1.default(e, imports, fileInfo.filename));
                }
                // If `i` is smaller than the `input.length - 1`,
                // it means the parser wasn't able to parse the whole
                // string, so we've got a parsing error.
                //
                // We try to extract a \n delimited string,
                // showing the line where the parse error occurred.
                // We split it up into two parts (the part which parsed,
                // and the part which didn't), so we can color them differently.
                var endInfo = parserInput.end();
                if (!endInfo.isFinished) {
                    var message = endInfo.furthestPossibleErrorMessage;
                    if (!message) {
                        message = 'Unrecognised input';
                        if (endInfo.furthestChar === '}') {
                            message += '. Possibly missing opening \'{\'';
                        }
                        else if (endInfo.furthestChar === ')') {
                            message += '. Possibly missing opening \'(\'';
                        }
                        else if (endInfo.furthestReachedEnd) {
                            message += '. Possibly missing something';
                        }
                    }
                    error = new less_error_1.default({
                        type: 'Parse',
                        message: message,
                        index: endInfo.furthest,
                        filename: fileInfo.filename
                    }, imports);
                }
                var finish = function (e) {
                    e = error || e || imports.error;
                    if (e) {
                        if (!(e instanceof less_error_1.default)) {
                            e = new less_error_1.default(e, imports, fileInfo.filename);
                        }
                        return callback(e);
                    }
                    else {
                        return callback(null, root);
                    }
                };
                if (context.processImports !== false) {
                    new visitors_1.default.ImportVisitor(imports, finish)
                        .run(root);
                }
                else {
                    return finish();
                }
            },
            //
            // Here in, the parsing rules/functions
            //
            // The basic structure of the syntax tree generated is as follows:
            //
            //   Ruleset ->  Declaration -> Value -> Expression -> Entity
            //
            // Here's some Less code:
            //
            //    .class {
            //      color: #fff;
            //      border: 1px solid #000;
            //      width: @w + 4px;
            //      > .child {...}
            //    }
            //
            // And here's what the parse tree might look like:
            //
            //     Ruleset (Selector '.class', [
            //         Declaration ("color",  Value ([Expression [Color #fff]]))
            //         Declaration ("border", Value ([Expression [Dimension 1px][Keyword "solid"][Color #000]]))
            //         Declaration ("width",  Value ([Expression [Operation " + " [Variable "@w"][Dimension 4px]]]))
            //         Ruleset (Selector [Element '>', '.child'], [...])
            //     ])
            //
            //  In general, most rules will try to parse a token with the `$re()` function, and if the return
            //  value is truly, will return a new node, of the relevant type. Sometimes, we need to check
            //  first, before parsing, that's when we use `peek()`.
            //
            parsers: parsers = {
                //
                // The `primary` rule is the *entry* and *exit* point of the parser.
                // The rules here can appear at any level of the parse tree.
                //
                // The recursive nature of the grammar is an interplay between the `block`
                // rule, which represents `{ ... }`, the `ruleset` rule, and this `primary` rule,
                // as represented by this simplified grammar:
                //
                //     primary  →  (ruleset | declaration)+
                //     ruleset  →  selector+ block
                //     block    →  '{' primary '}'
                //
                // Only at one point is the primary rule not called from the
                // block rule: at the root level.
                //
                primary: function () {
                    var mixin = this.mixin;
                    var root = [];
                    var node;
                    while (true) {
                        while (true) {
                            node = this.comment();
                            if (!node) {
                                break;
                            }
                            root.push(node);
                        }
                        // always process comments before deciding if finished
                        if (parserInput.finished) {
                            break;
                        }
                        if (parserInput.peek('}')) {
                            break;
                        }
                        node = this.extendRule();
                        if (node) {
                            root = root.concat(node);
                            continue;
                        }
                        node = mixin.definition() || this.declaration() || mixin.call(false, false) ||
                            this.ruleset() || this.variableCall() || this.entities.call() || this.atrule();
                        if (node) {
                            root.push(node);
                        }
                        else {
                            var foundSemiColon = false;
                            while (parserInput.$char(';')) {
                                foundSemiColon = true;
                            }
                            if (!foundSemiColon) {
                                break;
                            }
                        }
                    }
                    return root;
                },
                // comments are collected by the main parsing mechanism and then assigned to nodes
                // where the current structure allows it
                comment: function () {
                    if (parserInput.commentStore.length) {
                        var comment = parserInput.commentStore.shift();
                        return new (tree_1.default.Comment)(comment.text, comment.isLineComment, comment.index, fileInfo);
                    }
                },
                //
                // Entities are tokens which can be found inside an Expression
                //
                entities: {
                    mixinLookup: function () {
                        return parsers.mixin.call(true, true);
                    },
                    //
                    // A string, which supports escaping " and '
                    //
                    //     "milky way" 'he\'s the one!'
                    //
                    quoted: function (forceEscaped) {
                        var str;
                        var index = parserInput.i;
                        var isEscaped = false;
                        parserInput.save();
                        if (parserInput.$char('~')) {
                            isEscaped = true;
                        }
                        else if (forceEscaped) {
                            parserInput.restore();
                            return;
                        }
                        str = parserInput.$quoted();
                        if (!str) {
                            parserInput.restore();
                            return;
                        }
                        parserInput.forget();
                        return new (tree_1.default.Quoted)(str.charAt(0), str.substr(1, str.length - 2), isEscaped, index, fileInfo);
                    },
                    //
                    // A catch-all word, such as:
                    //
                    //     black border-collapse
                    //
                    keyword: function () {
                        var k = parserInput.$char('%') || parserInput.$re(/^\[?(?:[\w-]|\\(?:[A-Fa-f0-9]{1,6} ?|[^A-Fa-f0-9]))+\]?/);
                        if (k) {
                            return tree_1.default.Color.fromKeyword(k) || new (tree_1.default.Keyword)(k);
                        }
                    },
                    //
                    // A function call
                    //
                    //     rgb(255, 0, 255)
                    //
                    // The arguments are parsed with the `entities.arguments` parser.
                    //
                    call: function () {
                        var name;
                        var args;
                        var func;
                        var index = parserInput.i;
                        // http://jsperf.com/case-insensitive-regex-vs-strtolower-then-regex/18
                        if (parserInput.peek(/^url\(/i)) {
                            return;
                        }
                        parserInput.save();
                        name = parserInput.$re(/^([\w-]+|%|~|progid:[\w\.]+)\(/);
                        if (!name) {
                            parserInput.forget();
                            return;
                        }
                        name = name[1];
                        func = this.customFuncCall(name);
                        if (func) {
                            args = func.parse();
                            if (args && func.stop) {
                                parserInput.forget();
                                return args;
                            }
                        }
                        args = this.arguments(args);
                        if (!parserInput.$char(')')) {
                            parserInput.restore('Could not parse call arguments or missing \')\'');
                            return;
                        }
                        parserInput.forget();
                        return new (tree_1.default.Call)(name, args, index, fileInfo);
                    },
                    //
                    // Parsing rules for functions with non-standard args, e.g.:
                    //
                    //     boolean(not(2 > 1))
                    //
                    //     This is a quick prototype, to be modified/improved when
                    //     more custom-parsed funcs come (e.g. `selector(...)`)
                    //
                    customFuncCall: function (name) {
                        /* Ideally the table is to be moved out of here for faster perf.,
                           but it's quite tricky since it relies on all these `parsers`
                           and `expect` available only here */
                        return {
                            alpha: f(parsers.ieAlpha, true),
                            boolean: f(condition),
                            'if': f(condition)
                        }[name.toLowerCase()];
                        function f(parse, stop) {
                            return {
                                parse: parse,
                                stop: stop // when true - stop after parse() and return its result, 
                                // otherwise continue for plain args
                            };
                        }
                        function condition() {
                            return [expect(parsers.condition, 'expected condition')];
                        }
                    },
                    arguments: function (prevArgs) {
                        var argsComma = prevArgs || [];
                        var argsSemiColon = [];
                        var isSemiColonSeparated;
                        var value;
                        parserInput.save();
                        while (true) {
                            if (prevArgs) {
                                prevArgs = false;
                            }
                            else {
                                value = parsers.detachedRuleset() || this.assignment() || parsers.expression();
                                if (!value) {
                                    break;
                                }
                                if (value.value && value.value.length == 1) {
                                    value = value.value[0];
                                }
                                argsComma.push(value);
                            }
                            if (parserInput.$char(',')) {
                                continue;
                            }
                            if (parserInput.$char(';') || isSemiColonSeparated) {
                                isSemiColonSeparated = true;
                                value = (argsComma.length < 1) ? argsComma[0]
                                    : new tree_1.default.Value(argsComma);
                                argsSemiColon.push(value);
                                argsComma = [];
                            }
                        }
                        parserInput.forget();
                        return isSemiColonSeparated ? argsSemiColon : argsComma;
                    },
                    literal: function () {
                        return this.dimension() ||
                            this.color() ||
                            this.quoted() ||
                            this.unicodeDescriptor();
                    },
                    // Assignments are argument entities for calls.
                    // They are present in ie filter properties as shown below.
                    //
                    //     filter: progid:DXImageTransform.Microsoft.Alpha( *opacity=50* )
                    //
                    assignment: function () {
                        var key;
                        var value;
                        parserInput.save();
                        key = parserInput.$re(/^\w+(?=\s?=)/i);
                        if (!key) {
                            parserInput.restore();
                            return;
                        }
                        if (!parserInput.$char('=')) {
                            parserInput.restore();
                            return;
                        }
                        value = parsers.entity();
                        if (value) {
                            parserInput.forget();
                            return new (tree_1.default.Assignment)(key, value);
                        }
                        else {
                            parserInput.restore();
                        }
                    },
                    //
                    // Parse url() tokens
                    //
                    // We use a specific rule for urls, because they don't really behave like
                    // standard function calls. The difference is that the argument doesn't have
                    // to be enclosed within a string, so it can't be parsed as an Expression.
                    //
                    url: function () {
                        var value;
                        var index = parserInput.i;
                        parserInput.autoCommentAbsorb = false;
                        if (!parserInput.$str('url(')) {
                            parserInput.autoCommentAbsorb = true;
                            return;
                        }
                        value = this.quoted() || this.variable() || this.property() ||
                            parserInput.$re(/^(?:(?:\\[\(\)'"])|[^\(\)'"])+/) || '';
                        parserInput.autoCommentAbsorb = true;
                        expectChar(')');
                        return new (tree_1.default.URL)((value.value != null ||
                            value instanceof tree_1.default.Variable ||
                            value instanceof tree_1.default.Property) ?
                            value : new (tree_1.default.Anonymous)(value, index), index, fileInfo);
                    },
                    //
                    // A Variable entity, such as `@fink`, in
                    //
                    //     width: @fink + 2px
                    //
                    // We use a different parser for variable definitions,
                    // see `parsers.variable`.
                    //
                    variable: function () {
                        var ch;
                        var name;
                        var index = parserInput.i;
                        parserInput.save();
                        if (parserInput.currentChar() === '@' && (name = parserInput.$re(/^@@?[\w-]+/))) {
                            ch = parserInput.currentChar();
                            if (ch === '(' || ch === '[' && !parserInput.prevChar().match(/^\s/)) {
                                // this may be a VariableCall lookup
                                var result = parsers.variableCall(name);
                                if (result) {
                                    parserInput.forget();
                                    return result;
                                }
                            }
                            parserInput.forget();
                            return new (tree_1.default.Variable)(name, index, fileInfo);
                        }
                        parserInput.restore();
                    },
                    // A variable entity using the protective {} e.g. @{var}
                    variableCurly: function () {
                        var curly;
                        var index = parserInput.i;
                        if (parserInput.currentChar() === '@' && (curly = parserInput.$re(/^@\{([\w-]+)\}/))) {
                            return new (tree_1.default.Variable)("@" + curly[1], index, fileInfo);
                        }
                    },
                    //
                    // A Property accessor, such as `$color`, in
                    //
                    //     background-color: $color
                    //
                    property: function () {
                        var name;
                        var index = parserInput.i;
                        if (parserInput.currentChar() === '$' && (name = parserInput.$re(/^\$[\w-]+/))) {
                            return new (tree_1.default.Property)(name, index, fileInfo);
                        }
                    },
                    // A property entity useing the protective {} e.g. ${prop}
                    propertyCurly: function () {
                        var curly;
                        var index = parserInput.i;
                        if (parserInput.currentChar() === '$' && (curly = parserInput.$re(/^\$\{([\w-]+)\}/))) {
                            return new (tree_1.default.Property)("$" + curly[1], index, fileInfo);
                        }
                    },
                    //
                    // A Hexadecimal color
                    //
                    //     #4F3C2F
                    //
                    // `rgb` and `hsl` colors are parsed through the `entities.call` parser.
                    //
                    color: function () {
                        var rgb;
                        parserInput.save();
                        if (parserInput.currentChar() === '#' && (rgb = parserInput.$re(/^#([A-Fa-f0-9]{8}|[A-Fa-f0-9]{6}|[A-Fa-f0-9]{3,4})([\w.#\[])?/))) {
                            if (!rgb[2]) {
                                parserInput.forget();
                                return new (tree_1.default.Color)(rgb[1], undefined, rgb[0]);
                            }
                        }
                        parserInput.restore();
                    },
                    colorKeyword: function () {
                        parserInput.save();
                        var autoCommentAbsorb = parserInput.autoCommentAbsorb;
                        parserInput.autoCommentAbsorb = false;
                        var k = parserInput.$re(/^[_A-Za-z-][_A-Za-z0-9-]+/);
                        parserInput.autoCommentAbsorb = autoCommentAbsorb;
                        if (!k) {
                            parserInput.forget();
                            return;
                        }
                        parserInput.restore();
                        var color = tree_1.default.Color.fromKeyword(k);
                        if (color) {
                            parserInput.$str(k);
                            return color;
                        }
                    },
                    //
                    // A Dimension, that is, a number and a unit
                    //
                    //     0.5em 95%
                    //
                    dimension: function () {
                        if (parserInput.peekNotNumeric()) {
                            return;
                        }
                        var value = parserInput.$re(/^([+-]?\d*\.?\d+)(%|[a-z_]+)?/i);
                        if (value) {
                            return new (tree_1.default.Dimension)(value[1], value[2]);
                        }
                    },
                    //
                    // A unicode descriptor, as is used in unicode-range
                    //
                    // U+0??  or U+00A1-00A9
                    //
                    unicodeDescriptor: function () {
                        var ud;
                        ud = parserInput.$re(/^U\+[0-9a-fA-F?]+(\-[0-9a-fA-F?]+)?/);
                        if (ud) {
                            return new (tree_1.default.UnicodeDescriptor)(ud[0]);
                        }
                    },
                    //
                    // JavaScript code to be evaluated
                    //
                    //     `window.location.href`
                    //
                    javascript: function () {
                        var js;
                        var index = parserInput.i;
                        parserInput.save();
                        var escape = parserInput.$char('~');
                        var jsQuote = parserInput.$char('`');
                        if (!jsQuote) {
                            parserInput.restore();
                            return;
                        }
                        js = parserInput.$re(/^[^`]*`/);
                        if (js) {
                            parserInput.forget();
                            return new (tree_1.default.JavaScript)(js.substr(0, js.length - 1), Boolean(escape), index, fileInfo);
                        }
                        parserInput.restore('invalid javascript definition');
                    }
                },
                //
                // The variable part of a variable definition. Used in the `rule` parser
                //
                //     @fink:
                //
                variable: function () {
                    var name;
                    if (parserInput.currentChar() === '@' && (name = parserInput.$re(/^(@[\w-]+)\s*:/))) {
                        return name[1];
                    }
                },
                //
                // Call a variable value to retrieve a detached ruleset
                // or a value from a detached ruleset's rules.
                //
                //     @fink();
                //     @fink;
                //     color: @fink[@color];
                //
                variableCall: function (parsedName) {
                    var lookups;
                    var i = parserInput.i;
                    var inValue = !!parsedName;
                    var name = parsedName;
                    parserInput.save();
                    if (name || (parserInput.currentChar() === '@'
                        && (name = parserInput.$re(/^(@[\w-]+)(\(\s*\))?/)))) {
                        lookups = this.mixin.ruleLookups();
                        if (!lookups && ((inValue && parserInput.$str('()') !== '()') || (name[2] !== '()'))) {
                            parserInput.restore('Missing \'[...]\' lookup in variable call');
                            return;
                        }
                        if (!inValue) {
                            name = name[1];
                        }
                        var call = new tree_1.default.VariableCall(name, i, fileInfo);
                        if (!inValue && parsers.end()) {
                            parserInput.forget();
                            return call;
                        }
                        else {
                            parserInput.forget();
                            return new tree_1.default.NamespaceValue(call, lookups, i, fileInfo);
                        }
                    }
                    parserInput.restore();
                },
                //
                // extend syntax - used to extend selectors
                //
                extend: function (isRule) {
                    var elements;
                    var e;
                    var index = parserInput.i;
                    var option;
                    var extendList;
                    var extend;
                    if (!parserInput.$str(isRule ? '&:extend(' : ':extend(')) {
                        return;
                    }
                    do {
                        option = null;
                        elements = null;
                        while (!(option = parserInput.$re(/^(all)(?=\s*(\)|,))/))) {
                            e = this.element();
                            if (!e) {
                                break;
                            }
                            if (elements) {
                                elements.push(e);
                            }
                            else {
                                elements = [e];
                            }
                        }
                        option = option && option[1];
                        if (!elements) {
                            error('Missing target selector for :extend().');
                        }
                        extend = new (tree_1.default.Extend)(new (tree_1.default.Selector)(elements), option, index, fileInfo);
                        if (extendList) {
                            extendList.push(extend);
                        }
                        else {
                            extendList = [extend];
                        }
                    } while (parserInput.$char(','));
                    expect(/^\)/);
                    if (isRule) {
                        expect(/^;/);
                    }
                    return extendList;
                },
                //
                // extendRule - used in a rule to extend all the parent selectors
                //
                extendRule: function () {
                    return this.extend(true);
                },
                //
                // Mixins
                //
                mixin: {
                    //
                    // A Mixin call, with an optional argument list
                    //
                    //     #mixins > .square(#fff);
                    //     #mixins.square(#fff);
                    //     .rounded(4px, black);
                    //     .button;
                    //
                    // We can lookup / return a value using the lookup syntax:
                    //
                    //     color: #mixin.square(#fff)[@color];
                    //
                    // The `while` loop is there because mixins can be
                    // namespaced, but we only support the child and descendant
                    // selector for now.
                    //
                    call: function (inValue, getLookup) {
                        var s = parserInput.currentChar();
                        var important = false;
                        var lookups;
                        var index = parserInput.i;
                        var elements;
                        var args;
                        var hasParens;
                        if (s !== '.' && s !== '#') {
                            return;
                        }
                        parserInput.save(); // stop us absorbing part of an invalid selector
                        elements = this.elements();
                        if (elements) {
                            if (parserInput.$char('(')) {
                                args = this.args(true).args;
                                expectChar(')');
                                hasParens = true;
                            }
                            if (getLookup !== false) {
                                lookups = this.ruleLookups();
                            }
                            if (getLookup === true && !lookups) {
                                parserInput.restore();
                                return;
                            }
                            if (inValue && !lookups && !hasParens) {
                                // This isn't a valid in-value mixin call
                                parserInput.restore();
                                return;
                            }
                            if (!inValue && parsers.important()) {
                                important = true;
                            }
                            if (inValue || parsers.end()) {
                                parserInput.forget();
                                var mixin = new (tree_1.default.mixin.Call)(elements, args, index, fileInfo, !lookups && important);
                                if (lookups) {
                                    return new tree_1.default.NamespaceValue(mixin, lookups);
                                }
                                else {
                                    return mixin;
                                }
                            }
                        }
                        parserInput.restore();
                    },
                    /**
                     * Matching elements for mixins
                     * (Start with . or # and can have > )
                     */
                    elements: function () {
                        var elements;
                        var e;
                        var c;
                        var elem;
                        var elemIndex;
                        var re = /^[#.](?:[\w-]|\\(?:[A-Fa-f0-9]{1,6} ?|[^A-Fa-f0-9]))+/;
                        while (true) {
                            elemIndex = parserInput.i;
                            e = parserInput.$re(re);
                            if (!e) {
                                break;
                            }
                            elem = new (tree_1.default.Element)(c, e, false, elemIndex, fileInfo);
                            if (elements) {
                                elements.push(elem);
                            }
                            else {
                                elements = [elem];
                            }
                            c = parserInput.$char('>');
                        }
                        return elements;
                    },
                    args: function (isCall) {
                        var entities = parsers.entities;
                        var returner = { args: null, variadic: false };
                        var expressions = [];
                        var argsSemiColon = [];
                        var argsComma = [];
                        var isSemiColonSeparated;
                        var expressionContainsNamed;
                        var name;
                        var nameLoop;
                        var value;
                        var arg;
                        var expand;
                        var hasSep = true;
                        parserInput.save();
                        while (true) {
                            if (isCall) {
                                arg = parsers.detachedRuleset() || parsers.expression();
                            }
                            else {
                                parserInput.commentStore.length = 0;
                                if (parserInput.$str('...')) {
                                    returner.variadic = true;
                                    if (parserInput.$char(';') && !isSemiColonSeparated) {
                                        isSemiColonSeparated = true;
                                    }
                                    (isSemiColonSeparated ? argsSemiColon : argsComma)
                                        .push({ variadic: true });
                                    break;
                                }
                                arg = entities.variable() || entities.property() || entities.literal() || entities.keyword() || this.call(true);
                            }
                            if (!arg || !hasSep) {
                                break;
                            }
                            nameLoop = null;
                            if (arg.throwAwayComments) {
                                arg.throwAwayComments();
                            }
                            value = arg;
                            var val = null;
                            if (isCall) {
                                // Variable
                                if (arg.value && arg.value.length == 1) {
                                    val = arg.value[0];
                                }
                            }
                            else {
                                val = arg;
                            }
                            if (val && (val instanceof tree_1.default.Variable || val instanceof tree_1.default.Property)) {
                                if (parserInput.$char(':')) {
                                    if (expressions.length > 0) {
                                        if (isSemiColonSeparated) {
                                            error('Cannot mix ; and , as delimiter types');
                                        }
                                        expressionContainsNamed = true;
                                    }
                                    value = parsers.detachedRuleset() || parsers.expression();
                                    if (!value) {
                                        if (isCall) {
                                            error('could not understand value for named argument');
                                        }
                                        else {
                                            parserInput.restore();
                                            returner.args = [];
                                            return returner;
                                        }
                                    }
                                    nameLoop = (name = val.name);
                                }
                                else if (parserInput.$str('...')) {
                                    if (!isCall) {
                                        returner.variadic = true;
                                        if (parserInput.$char(';') && !isSemiColonSeparated) {
                                            isSemiColonSeparated = true;
                                        }
                                        (isSemiColonSeparated ? argsSemiColon : argsComma)
                                            .push({ name: arg.name, variadic: true });
                                        break;
                                    }
                                    else {
                                        expand = true;
                                    }
                                }
                                else if (!isCall) {
                                    name = nameLoop = val.name;
                                    value = null;
                                }
                            }
                            if (value) {
                                expressions.push(value);
                            }
                            argsComma.push({ name: nameLoop, value: value, expand: expand });
                            if (parserInput.$char(',')) {
                                hasSep = true;
                                continue;
                            }
                            hasSep = parserInput.$char(';') === ';';
                            if (hasSep || isSemiColonSeparated) {
                                if (expressionContainsNamed) {
                                    error('Cannot mix ; and , as delimiter types');
                                }
                                isSemiColonSeparated = true;
                                if (expressions.length > 1) {
                                    value = new (tree_1.default.Value)(expressions);
                                }
                                argsSemiColon.push({ name: name, value: value, expand: expand });
                                name = null;
                                expressions = [];
                                expressionContainsNamed = false;
                            }
                        }
                        parserInput.forget();
                        returner.args = isSemiColonSeparated ? argsSemiColon : argsComma;
                        return returner;
                    },
                    //
                    // A Mixin definition, with a list of parameters
                    //
                    //     .rounded (@radius: 2px, @color) {
                    //        ...
                    //     }
                    //
                    // Until we have a finer grained state-machine, we have to
                    // do a look-ahead, to make sure we don't have a mixin call.
                    // See the `rule` function for more information.
                    //
                    // We start by matching `.rounded (`, and then proceed on to
                    // the argument list, which has optional default values.
                    // We store the parameters in `params`, with a `value` key,
                    // if there is a value, such as in the case of `@radius`.
                    //
                    // Once we've got our params list, and a closing `)`, we parse
                    // the `{...}` block.
                    //
                    definition: function () {
                        var name;
                        var params = [];
                        var match;
                        var ruleset;
                        var cond;
                        var variadic = false;
                        if ((parserInput.currentChar() !== '.' && parserInput.currentChar() !== '#') ||
                            parserInput.peek(/^[^{]*\}/)) {
                            return;
                        }
                        parserInput.save();
                        match = parserInput.$re(/^([#.](?:[\w-]|\\(?:[A-Fa-f0-9]{1,6} ?|[^A-Fa-f0-9]))+)\s*\(/);
                        if (match) {
                            name = match[1];
                            var argInfo = this.args(false);
                            params = argInfo.args;
                            variadic = argInfo.variadic;
                            // .mixincall("@{a}");
                            // looks a bit like a mixin definition..
                            // also
                            // .mixincall(@a: {rule: set;});
                            // so we have to be nice and restore
                            if (!parserInput.$char(')')) {
                                parserInput.restore('Missing closing \')\'');
                                return;
                            }
                            parserInput.commentStore.length = 0;
                            if (parserInput.$str('when')) { // Guard
                                cond = expect(parsers.conditions, 'expected condition');
                            }
                            ruleset = parsers.block();
                            if (ruleset) {
                                parserInput.forget();
                                return new (tree_1.default.mixin.Definition)(name, params, ruleset, cond, variadic);
                            }
                            else {
                                parserInput.restore();
                            }
                        }
                        else {
                            parserInput.restore();
                        }
                    },
                    ruleLookups: function () {
                        var rule;
                        var args;
                        var lookups = [];
                        if (parserInput.currentChar() !== '[') {
                            return;
                        }
                        while (true) {
                            parserInput.save();
                            args = null;
                            rule = this.lookupValue();
                            if (!rule && rule !== '') {
                                parserInput.restore();
                                break;
                            }
                            lookups.push(rule);
                            parserInput.forget();
                        }
                        if (lookups.length > 0) {
                            return lookups;
                        }
                    },
                    lookupValue: function () {
                        parserInput.save();
                        if (!parserInput.$char('[')) {
                            parserInput.restore();
                            return;
                        }
                        var name = parserInput.$re(/^(?:[@$]{0,2})[_a-zA-Z0-9-]*/);
                        if (!parserInput.$char(']')) {
                            parserInput.restore();
                            return;
                        }
                        if (name || name === '') {
                            parserInput.forget();
                            return name;
                        }
                        parserInput.restore();
                    }
                },
                //
                // Entities are the smallest recognized token,
                // and can be found inside a rule's value.
                //
                entity: function () {
                    var entities = this.entities;
                    return this.comment() || entities.literal() || entities.variable() || entities.url() ||
                        entities.property() || entities.call() || entities.keyword() || this.mixin.call(true) ||
                        entities.javascript();
                },
                //
                // A Declaration terminator. Note that we use `peek()` to check for '}',
                // because the `block` rule will be expecting it, but we still need to make sure
                // it's there, if ';' was omitted.
                //
                end: function () {
                    return parserInput.$char(';') || parserInput.peek('}');
                },
                //
                // IE's alpha function
                //
                //     alpha(opacity=88)
                //
                ieAlpha: function () {
                    var value;
                    // http://jsperf.com/case-insensitive-regex-vs-strtolower-then-regex/18
                    if (!parserInput.$re(/^opacity=/i)) {
                        return;
                    }
                    value = parserInput.$re(/^\d+/);
                    if (!value) {
                        value = expect(parsers.entities.variable, 'Could not parse alpha');
                        value = "@{" + value.name.slice(1) + "}";
                    }
                    expectChar(')');
                    return new tree_1.default.Quoted('', "alpha(opacity=" + value + ")");
                },
                //
                // A Selector Element
                //
                //     div
                //     + h1
                //     #socks
                //     input[type="text"]
                //
                // Elements are the building blocks for Selectors,
                // they are made out of a `Combinator` (see combinator rule),
                // and an element name, such as a tag a class, or `*`.
                //
                element: function () {
                    var e;
                    var c;
                    var v;
                    var index = parserInput.i;
                    c = this.combinator();
                    e = parserInput.$re(/^(?:\d+\.\d+|\d+)%/) ||
                        parserInput.$re(/^(?:[.#]?|:*)(?:[\w-]|[^\x00-\x9f]|\\(?:[A-Fa-f0-9]{1,6} ?|[^A-Fa-f0-9]))+/) ||
                        parserInput.$char('*') || parserInput.$char('&') || this.attribute() ||
                        parserInput.$re(/^\([^&()@]+\)/) || parserInput.$re(/^[\.#:](?=@)/) ||
                        this.entities.variableCurly();
                    if (!e) {
                        parserInput.save();
                        if (parserInput.$char('(')) {
                            if ((v = this.selector(false)) && parserInput.$char(')')) {
                                e = new (tree_1.default.Paren)(v);
                                parserInput.forget();
                            }
                            else {
                                parserInput.restore('Missing closing \')\'');
                            }
                        }
                        else {
                            parserInput.forget();
                        }
                    }
                    if (e) {
                        return new (tree_1.default.Element)(c, e, e instanceof tree_1.default.Variable, index, fileInfo);
                    }
                },
                //
                // Combinators combine elements together, in a Selector.
                //
                // Because our parser isn't white-space sensitive, special care
                // has to be taken, when parsing the descendant combinator, ` `,
                // as it's an empty space. We have to check the previous character
                // in the input, to see if it's a ` ` character. More info on how
                // we deal with this in *combinator.js*.
                //
                combinator: function () {
                    var c = parserInput.currentChar();
                    if (c === '/') {
                        parserInput.save();
                        var slashedCombinator = parserInput.$re(/^\/[a-z]+\//i);
                        if (slashedCombinator) {
                            parserInput.forget();
                            return new (tree_1.default.Combinator)(slashedCombinator);
                        }
                        parserInput.restore();
                    }
                    if (c === '>' || c === '+' || c === '~' || c === '|' || c === '^') {
                        parserInput.i++;
                        if (c === '^' && parserInput.currentChar() === '^') {
                            c = '^^';
                            parserInput.i++;
                        }
                        while (parserInput.isWhitespace()) {
                            parserInput.i++;
                        }
                        return new (tree_1.default.Combinator)(c);
                    }
                    else if (parserInput.isWhitespace(-1)) {
                        return new (tree_1.default.Combinator)(' ');
                    }
                    else {
                        return new (tree_1.default.Combinator)(null);
                    }
                },
                //
                // A CSS Selector
                // with less extensions e.g. the ability to extend and guard
                //
                //     .class > div + h1
                //     li a:hover
                //
                // Selectors are made out of one or more Elements, see above.
                //
                selector: function (isLess) {
                    var index = parserInput.i;
                    var elements;
                    var extendList;
                    var c;
                    var e;
                    var allExtends;
                    var when;
                    var condition;
                    isLess = isLess !== false;
                    while ((isLess && (extendList = this.extend())) || (isLess && (when = parserInput.$str('when'))) || (e = this.element())) {
                        if (when) {
                            condition = expect(this.conditions, 'expected condition');
                        }
                        else if (condition) {
                            error('CSS guard can only be used at the end of selector');
                        }
                        else if (extendList) {
                            if (allExtends) {
                                allExtends = allExtends.concat(extendList);
                            }
                            else {
                                allExtends = extendList;
                            }
                        }
                        else {
                            if (allExtends) {
                                error('Extend can only be used at the end of selector');
                            }
                            c = parserInput.currentChar();
                            if (elements) {
                                elements.push(e);
                            }
                            else {
                                elements = [e];
                            }
                            e = null;
                        }
                        if (c === '{' || c === '}' || c === ';' || c === ',' || c === ')') {
                            break;
                        }
                    }
                    if (elements) {
                        return new (tree_1.default.Selector)(elements, allExtends, condition, index, fileInfo);
                    }
                    if (allExtends) {
                        error('Extend must be used to extend a selector, it cannot be used on its own');
                    }
                },
                selectors: function () {
                    var s;
                    var selectors;
                    while (true) {
                        s = this.selector();
                        if (!s) {
                            break;
                        }
                        if (selectors) {
                            selectors.push(s);
                        }
                        else {
                            selectors = [s];
                        }
                        parserInput.commentStore.length = 0;
                        if (s.condition && selectors.length > 1) {
                            error("Guards are only currently allowed on a single selector.");
                        }
                        if (!parserInput.$char(',')) {
                            break;
                        }
                        if (s.condition) {
                            error("Guards are only currently allowed on a single selector.");
                        }
                        parserInput.commentStore.length = 0;
                    }
                    return selectors;
                },
                attribute: function () {
                    if (!parserInput.$char('[')) {
                        return;
                    }
                    var entities = this.entities;
                    var key;
                    var val;
                    var op;
                    if (!(key = entities.variableCurly())) {
                        key = expect(/^(?:[_A-Za-z0-9-\*]*\|)?(?:[_A-Za-z0-9-]|\\.)+/);
                    }
                    op = parserInput.$re(/^[|~*$^]?=/);
                    if (op) {
                        val = entities.quoted() || parserInput.$re(/^[0-9]+%/) || parserInput.$re(/^[\w-]+/) || entities.variableCurly();
                    }
                    expectChar(']');
                    return new (tree_1.default.Attribute)(key, op, val);
                },
                //
                // The `block` rule is used by `ruleset` and `mixin.definition`.
                // It's a wrapper around the `primary` rule, with added `{}`.
                //
                block: function () {
                    var content;
                    if (parserInput.$char('{') && (content = this.primary()) && parserInput.$char('}')) {
                        return content;
                    }
                },
                blockRuleset: function () {
                    var block = this.block();
                    if (block) {
                        block = new tree_1.default.Ruleset(null, block);
                    }
                    return block;
                },
                detachedRuleset: function () {
                    var argInfo;
                    var params;
                    var variadic;
                    parserInput.save();
                    if (parserInput.$re(/^[.#]\(/)) {
                        /**
                         * DR args currently only implemented for each() function, and not
                         * yet settable as `@dr: #(@arg) {}`
                         * This should be done when DRs are merged with mixins.
                         * See: https://github.com/less/less-meta/issues/16
                         */
                        argInfo = this.mixin.args(false);
                        params = argInfo.args;
                        variadic = argInfo.variadic;
                        if (!parserInput.$char(')')) {
                            parserInput.restore();
                            return;
                        }
                    }
                    var blockRuleset = this.blockRuleset();
                    if (blockRuleset) {
                        parserInput.forget();
                        if (params) {
                            return new tree_1.default.mixin.Definition(null, params, blockRuleset, null, variadic);
                        }
                        return new tree_1.default.DetachedRuleset(blockRuleset);
                    }
                    parserInput.restore();
                },
                //
                // div, .class, body > p {...}
                //
                ruleset: function () {
                    var selectors;
                    var rules;
                    var debugInfo;
                    parserInput.save();
                    if (context.dumpLineNumbers) {
                        debugInfo = getDebugInfo(parserInput.i);
                    }
                    selectors = this.selectors();
                    if (selectors && (rules = this.block())) {
                        parserInput.forget();
                        var ruleset = new (tree_1.default.Ruleset)(selectors, rules, context.strictImports);
                        if (context.dumpLineNumbers) {
                            ruleset.debugInfo = debugInfo;
                        }
                        return ruleset;
                    }
                    else {
                        parserInput.restore();
                    }
                },
                declaration: function () {
                    var name;
                    var value;
                    var index = parserInput.i;
                    var hasDR;
                    var c = parserInput.currentChar();
                    var important;
                    var merge;
                    var isVariable;
                    if (c === '.' || c === '#' || c === '&' || c === ':') {
                        return;
                    }
                    parserInput.save();
                    name = this.variable() || this.ruleProperty();
                    if (name) {
                        isVariable = typeof name === 'string';
                        if (isVariable) {
                            value = this.detachedRuleset();
                            if (value) {
                                hasDR = true;
                            }
                        }
                        parserInput.commentStore.length = 0;
                        if (!value) {
                            // a name returned by this.ruleProperty() is always an array of the form:
                            // [string-1, ..., string-n, ""] or [string-1, ..., string-n, "+"]
                            // where each item is a tree.Keyword or tree.Variable
                            merge = !isVariable && name.length > 1 && name.pop().value;
                            // Custom property values get permissive parsing
                            if (name[0].value && name[0].value.slice(0, 2) === '--') {
                                value = this.permissiveValue();
                            }
                            // Try to store values as anonymous
                            // If we need the value later we'll re-parse it in ruleset.parseValue
                            else {
                                value = this.anonymousValue();
                            }
                            if (value) {
                                parserInput.forget();
                                // anonymous values absorb the end ';' which is required for them to work
                                return new (tree_1.default.Declaration)(name, value, false, merge, index, fileInfo);
                            }
                            if (!value) {
                                value = this.value();
                            }
                            if (value) {
                                important = this.important();
                            }
                            else if (isVariable) {
                                // As a last resort, try permissiveValue
                                value = this.permissiveValue();
                            }
                        }
                        if (value && (this.end() || hasDR)) {
                            parserInput.forget();
                            return new (tree_1.default.Declaration)(name, value, important, merge, index, fileInfo);
                        }
                        else {
                            parserInput.restore();
                        }
                    }
                    else {
                        parserInput.restore();
                    }
                },
                anonymousValue: function () {
                    var index = parserInput.i;
                    var match = parserInput.$re(/^([^.#@\$+\/'"*`(;{}-]*);/);
                    if (match) {
                        return new (tree_1.default.Anonymous)(match[1], index);
                    }
                },
                /**
                 * Used for custom properties, at-rules, and variables (as fallback)
                 * Parses almost anything inside of {} [] () "" blocks
                 * until it reaches outer-most tokens.
                 *
                 * First, it will try to parse comments and entities to reach
                 * the end. This is mostly like the Expression parser except no
                 * math is allowed.
                 */
                permissiveValue: function (untilTokens) {
                    var i;
                    var e;
                    var done;
                    var value;
                    var tok = untilTokens || ';';
                    var index = parserInput.i;
                    var result = [];
                    function testCurrentChar() {
                        var char = parserInput.currentChar();
                        if (typeof tok === 'string') {
                            return char === tok;
                        }
                        else {
                            return tok.test(char);
                        }
                    }
                    if (testCurrentChar()) {
                        return;
                    }
                    value = [];
                    do {
                        e = this.comment();
                        if (e) {
                            value.push(e);
                            continue;
                        }
                        e = this.entity();
                        if (e) {
                            value.push(e);
                        }
                    } while (e);
                    done = testCurrentChar();
                    if (value.length > 0) {
                        value = new (tree_1.default.Expression)(value);
                        if (done) {
                            return value;
                        }
                        else {
                            result.push(value);
                        }
                        // Preserve space before $parseUntil as it will not
                        if (parserInput.prevChar() === ' ') {
                            result.push(new tree_1.default.Anonymous(' ', index));
                        }
                    }
                    parserInput.save();
                    value = parserInput.$parseUntil(tok);
                    if (value) {
                        if (typeof value === 'string') {
                            error("Expected '" + value + "'", 'Parse');
                        }
                        if (value.length === 1 && value[0] === ' ') {
                            parserInput.forget();
                            return new tree_1.default.Anonymous('', index);
                        }
                        var item = void 0;
                        for (i = 0; i < value.length; i++) {
                            item = value[i];
                            if (Array.isArray(item)) {
                                // Treat actual quotes as normal quoted values
                                result.push(new tree_1.default.Quoted(item[0], item[1], true, index, fileInfo));
                            }
                            else {
                                if (i === value.length - 1) {
                                    item = item.trim();
                                }
                                // Treat like quoted values, but replace vars like unquoted expressions
                                var quote = new tree_1.default.Quoted('\'', item, true, index, fileInfo);
                                quote.variableRegex = /@([\w-]+)/g;
                                quote.propRegex = /\$([\w-]+)/g;
                                result.push(quote);
                            }
                        }
                        parserInput.forget();
                        return new tree_1.default.Expression(result, true);
                    }
                    parserInput.restore();
                },
                //
                // An @import atrule
                //
                //     @import "lib";
                //
                // Depending on our environment, importing is done differently:
                // In the browser, it's an XHR request, in Node, it would be a
                // file-system operation. The function used for importing is
                // stored in `import`, which we pass to the Import constructor.
                //
                'import': function () {
                    var path;
                    var features;
                    var index = parserInput.i;
                    var dir = parserInput.$re(/^@import?\s+/);
                    if (dir) {
                        var options = (dir ? this.importOptions() : null) || {};
                        if ((path = this.entities.quoted() || this.entities.url())) {
                            features = this.mediaFeatures();
                            if (!parserInput.$char(';')) {
                                parserInput.i = index;
                                error('missing semi-colon or unrecognised media features on import');
                            }
                            features = features && new (tree_1.default.Value)(features);
                            return new (tree_1.default.Import)(path, features, options, index, fileInfo);
                        }
                        else {
                            parserInput.i = index;
                            error('malformed import statement');
                        }
                    }
                },
                importOptions: function () {
                    var o;
                    var options = {};
                    var optionName;
                    var value;
                    // list of options, surrounded by parens
                    if (!parserInput.$char('(')) {
                        return null;
                    }
                    do {
                        o = this.importOption();
                        if (o) {
                            optionName = o;
                            value = true;
                            switch (optionName) {
                                case 'css':
                                    optionName = 'less';
                                    value = false;
                                    break;
                                case 'once':
                                    optionName = 'multiple';
                                    value = false;
                                    break;
                            }
                            options[optionName] = value;
                            if (!parserInput.$char(',')) {
                                break;
                            }
                        }
                    } while (o);
                    expectChar(')');
                    return options;
                },
                importOption: function () {
                    var opt = parserInput.$re(/^(less|css|multiple|once|inline|reference|optional)/);
                    if (opt) {
                        return opt[1];
                    }
                },
                mediaFeature: function () {
                    var entities = this.entities;
                    var nodes = [];
                    var e;
                    var p;
                    parserInput.save();
                    do {
                        e = entities.keyword() || entities.variable() || entities.mixinLookup();
                        if (e) {
                            nodes.push(e);
                        }
                        else if (parserInput.$char('(')) {
                            p = this.property();
                            e = this.value();
                            if (parserInput.$char(')')) {
                                if (p && e) {
                                    nodes.push(new (tree_1.default.Paren)(new (tree_1.default.Declaration)(p, e, null, null, parserInput.i, fileInfo, true)));
                                }
                                else if (e) {
                                    nodes.push(new (tree_1.default.Paren)(e));
                                }
                                else {
                                    error('badly formed media feature definition');
                                }
                            }
                            else {
                                error('Missing closing \')\'', 'Parse');
                            }
                        }
                    } while (e);
                    parserInput.forget();
                    if (nodes.length > 0) {
                        return new (tree_1.default.Expression)(nodes);
                    }
                },
                mediaFeatures: function () {
                    var entities = this.entities;
                    var features = [];
                    var e;
                    do {
                        e = this.mediaFeature();
                        if (e) {
                            features.push(e);
                            if (!parserInput.$char(',')) {
                                break;
                            }
                        }
                        else {
                            e = entities.variable() || entities.mixinLookup();
                            if (e) {
                                features.push(e);
                                if (!parserInput.$char(',')) {
                                    break;
                                }
                            }
                        }
                    } while (e);
                    return features.length > 0 ? features : null;
                },
                media: function () {
                    var features;
                    var rules;
                    var media;
                    var debugInfo;
                    var index = parserInput.i;
                    if (context.dumpLineNumbers) {
                        debugInfo = getDebugInfo(index);
                    }
                    parserInput.save();
                    if (parserInput.$str('@media')) {
                        features = this.mediaFeatures();
                        rules = this.block();
                        if (!rules) {
                            error('media definitions require block statements after any features');
                        }
                        parserInput.forget();
                        media = new (tree_1.default.Media)(rules, features, index, fileInfo);
                        if (context.dumpLineNumbers) {
                            media.debugInfo = debugInfo;
                        }
                        return media;
                    }
                    parserInput.restore();
                },
                //
                // A @plugin directive, used to import plugins dynamically.
                //
                //     @plugin (args) "lib";
                //
                plugin: function () {
                    var path;
                    var args;
                    var options;
                    var index = parserInput.i;
                    var dir = parserInput.$re(/^@plugin?\s+/);
                    if (dir) {
                        args = this.pluginArgs();
                        if (args) {
                            options = {
                                pluginArgs: args,
                                isPlugin: true
                            };
                        }
                        else {
                            options = { isPlugin: true };
                        }
                        if ((path = this.entities.quoted() || this.entities.url())) {
                            if (!parserInput.$char(';')) {
                                parserInput.i = index;
                                error('missing semi-colon on @plugin');
                            }
                            return new (tree_1.default.Import)(path, null, options, index, fileInfo);
                        }
                        else {
                            parserInput.i = index;
                            error('malformed @plugin statement');
                        }
                    }
                },
                pluginArgs: function () {
                    // list of options, surrounded by parens
                    parserInput.save();
                    if (!parserInput.$char('(')) {
                        parserInput.restore();
                        return null;
                    }
                    var args = parserInput.$re(/^\s*([^\);]+)\)\s*/);
                    if (args[1]) {
                        parserInput.forget();
                        return args[1].trim();
                    }
                    else {
                        parserInput.restore();
                        return null;
                    }
                },
                //
                // A CSS AtRule
                //
                //     @charset "utf-8";
                //
                atrule: function () {
                    var index = parserInput.i;
                    var name;
                    var value;
                    var rules;
                    var nonVendorSpecificName;
                    var hasIdentifier;
                    var hasExpression;
                    var hasUnknown;
                    var hasBlock = true;
                    var isRooted = true;
                    if (parserInput.currentChar() !== '@') {
                        return;
                    }
                    value = this['import']() || this.plugin() || this.media();
                    if (value) {
                        return value;
                    }
                    parserInput.save();
                    name = parserInput.$re(/^@[a-z-]+/);
                    if (!name) {
                        return;
                    }
                    nonVendorSpecificName = name;
                    if (name.charAt(1) == '-' && name.indexOf('-', 2) > 0) {
                        nonVendorSpecificName = "@" + name.slice(name.indexOf('-', 2) + 1);
                    }
                    switch (nonVendorSpecificName) {
                        case '@charset':
                            hasIdentifier = true;
                            hasBlock = false;
                            break;
                        case '@namespace':
                            hasExpression = true;
                            hasBlock = false;
                            break;
                        case '@keyframes':
                        case '@counter-style':
                            hasIdentifier = true;
                            break;
                        case '@document':
                        case '@supports':
                            hasUnknown = true;
                            isRooted = false;
                            break;
                        default:
                            hasUnknown = true;
                            break;
                    }
                    parserInput.commentStore.length = 0;
                    if (hasIdentifier) {
                        value = this.entity();
                        if (!value) {
                            error("expected " + name + " identifier");
                        }
                    }
                    else if (hasExpression) {
                        value = this.expression();
                        if (!value) {
                            error("expected " + name + " expression");
                        }
                    }
                    else if (hasUnknown) {
                        value = this.permissiveValue(/^[{;]/);
                        hasBlock = (parserInput.currentChar() === '{');
                        if (!value) {
                            if (!hasBlock && parserInput.currentChar() !== ';') {
                                error(name + " rule is missing block or ending semi-colon");
                            }
                        }
                        else if (!value.value) {
                            value = null;
                        }
                    }
                    if (hasBlock) {
                        rules = this.blockRuleset();
                    }
                    if (rules || (!hasBlock && value && parserInput.$char(';'))) {
                        parserInput.forget();
                        return new (tree_1.default.AtRule)(name, value, rules, index, fileInfo, context.dumpLineNumbers ? getDebugInfo(index) : null, isRooted);
                    }
                    parserInput.restore('at-rule options not recognised');
                },
                //
                // A Value is a comma-delimited list of Expressions
                //
                //     font-family: Baskerville, Georgia, serif;
                //
                // In a Rule, a Value represents everything after the `:`,
                // and before the `;`.
                //
                value: function () {
                    var e;
                    var expressions = [];
                    var index = parserInput.i;
                    do {
                        e = this.expression();
                        if (e) {
                            expressions.push(e);
                            if (!parserInput.$char(',')) {
                                break;
                            }
                        }
                    } while (e);
                    if (expressions.length > 0) {
                        return new (tree_1.default.Value)(expressions, index);
                    }
                },
                important: function () {
                    if (parserInput.currentChar() === '!') {
                        return parserInput.$re(/^! *important/);
                    }
                },
                sub: function () {
                    var a;
                    var e;
                    parserInput.save();
                    if (parserInput.$char('(')) {
                        a = this.addition();
                        if (a && parserInput.$char(')')) {
                            parserInput.forget();
                            e = new (tree_1.default.Expression)([a]);
                            e.parens = true;
                            return e;
                        }
                        parserInput.restore('Expected \')\'');
                        return;
                    }
                    parserInput.restore();
                },
                multiplication: function () {
                    var m;
                    var a;
                    var op;
                    var operation;
                    var isSpaced;
                    m = this.operand();
                    if (m) {
                        isSpaced = parserInput.isWhitespace(-1);
                        while (true) {
                            if (parserInput.peek(/^\/[*\/]/)) {
                                break;
                            }
                            parserInput.save();
                            op = parserInput.$char('/') || parserInput.$char('*') || parserInput.$str('./');
                            if (!op) {
                                parserInput.forget();
                                break;
                            }
                            a = this.operand();
                            if (!a) {
                                parserInput.restore();
                                break;
                            }
                            parserInput.forget();
                            m.parensInOp = true;
                            a.parensInOp = true;
                            operation = new (tree_1.default.Operation)(op, [operation || m, a], isSpaced);
                            isSpaced = parserInput.isWhitespace(-1);
                        }
                        return operation || m;
                    }
                },
                addition: function () {
                    var m;
                    var a;
                    var op;
                    var operation;
                    var isSpaced;
                    m = this.multiplication();
                    if (m) {
                        isSpaced = parserInput.isWhitespace(-1);
                        while (true) {
                            op = parserInput.$re(/^[-+]\s+/) || (!isSpaced && (parserInput.$char('+') || parserInput.$char('-')));
                            if (!op) {
                                break;
                            }
                            a = this.multiplication();
                            if (!a) {
                                break;
                            }
                            m.parensInOp = true;
                            a.parensInOp = true;
                            operation = new (tree_1.default.Operation)(op, [operation || m, a], isSpaced);
                            isSpaced = parserInput.isWhitespace(-1);
                        }
                        return operation || m;
                    }
                },
                conditions: function () {
                    var a;
                    var b;
                    var index = parserInput.i;
                    var condition;
                    a = this.condition(true);
                    if (a) {
                        while (true) {
                            if (!parserInput.peek(/^,\s*(not\s*)?\(/) || !parserInput.$char(',')) {
                                break;
                            }
                            b = this.condition(true);
                            if (!b) {
                                break;
                            }
                            condition = new (tree_1.default.Condition)('or', condition || a, b, index);
                        }
                        return condition || a;
                    }
                },
                condition: function (needsParens) {
                    var result;
                    var logical;
                    var next;
                    function or() {
                        return parserInput.$str('or');
                    }
                    result = this.conditionAnd(needsParens);
                    if (!result) {
                        return;
                    }
                    logical = or();
                    if (logical) {
                        next = this.condition(needsParens);
                        if (next) {
                            result = new (tree_1.default.Condition)(logical, result, next);
                        }
                        else {
                            return;
                        }
                    }
                    return result;
                },
                conditionAnd: function (needsParens) {
                    var result;
                    var logical;
                    var next;
                    var self = this;
                    function insideCondition() {
                        var cond = self.negatedCondition(needsParens) || self.parenthesisCondition(needsParens);
                        if (!cond && !needsParens) {
                            return self.atomicCondition(needsParens);
                        }
                        return cond;
                    }
                    function and() {
                        return parserInput.$str('and');
                    }
                    result = insideCondition();
                    if (!result) {
                        return;
                    }
                    logical = and();
                    if (logical) {
                        next = this.conditionAnd(needsParens);
                        if (next) {
                            result = new (tree_1.default.Condition)(logical, result, next);
                        }
                        else {
                            return;
                        }
                    }
                    return result;
                },
                negatedCondition: function (needsParens) {
                    if (parserInput.$str('not')) {
                        var result = this.parenthesisCondition(needsParens);
                        if (result) {
                            result.negate = !result.negate;
                        }
                        return result;
                    }
                },
                parenthesisCondition: function (needsParens) {
                    function tryConditionFollowedByParenthesis(me) {
                        var body;
                        parserInput.save();
                        body = me.condition(needsParens);
                        if (!body) {
                            parserInput.restore();
                            return;
                        }
                        if (!parserInput.$char(')')) {
                            parserInput.restore();
                            return;
                        }
                        parserInput.forget();
                        return body;
                    }
                    var body;
                    parserInput.save();
                    if (!parserInput.$str('(')) {
                        parserInput.restore();
                        return;
                    }
                    body = tryConditionFollowedByParenthesis(this);
                    if (body) {
                        parserInput.forget();
                        return body;
                    }
                    body = this.atomicCondition(needsParens);
                    if (!body) {
                        parserInput.restore();
                        return;
                    }
                    if (!parserInput.$char(')')) {
                        parserInput.restore("expected ')' got '" + parserInput.currentChar() + "'");
                        return;
                    }
                    parserInput.forget();
                    return body;
                },
                atomicCondition: function (needsParens) {
                    var entities = this.entities;
                    var index = parserInput.i;
                    var a;
                    var b;
                    var c;
                    var op;
                    function cond() {
                        return this.addition() || entities.keyword() || entities.quoted() || entities.mixinLookup();
                    }
                    cond = cond.bind(this);
                    a = cond();
                    if (a) {
                        if (parserInput.$char('>')) {
                            if (parserInput.$char('=')) {
                                op = '>=';
                            }
                            else {
                                op = '>';
                            }
                        }
                        else if (parserInput.$char('<')) {
                            if (parserInput.$char('=')) {
                                op = '<=';
                            }
                            else {
                                op = '<';
                            }
                        }
                        else if (parserInput.$char('=')) {
                            if (parserInput.$char('>')) {
                                op = '=>';
                            }
                            else if (parserInput.$char('<')) {
                                op = '=<';
                            }
                            else {
                                op = '=';
                            }
                        }
                        if (op) {
                            b = cond();
                            if (b) {
                                c = new (tree_1.default.Condition)(op, a, b, index, false);
                            }
                            else {
                                error('expected expression');
                            }
                        }
                        else {
                            c = new (tree_1.default.Condition)('=', a, new (tree_1.default.Keyword)('true'), index, false);
                        }
                        return c;
                    }
                },
                //
                // An operand is anything that can be part of an operation,
                // such as a Color, or a Variable
                //
                operand: function () {
                    var entities = this.entities;
                    var negate;
                    if (parserInput.peek(/^-[@\$\(]/)) {
                        negate = parserInput.$char('-');
                    }
                    var o = this.sub() || entities.dimension() ||
                        entities.color() || entities.variable() ||
                        entities.property() || entities.call() ||
                        entities.quoted(true) || entities.colorKeyword() ||
                        entities.mixinLookup();
                    if (negate) {
                        o.parensInOp = true;
                        o = new (tree_1.default.Negative)(o);
                    }
                    return o;
                },
                //
                // Expressions either represent mathematical operations,
                // or white-space delimited Entities.
                //
                //     1px solid black
                //     @var * 2
                //
                expression: function () {
                    var entities = [];
                    var e;
                    var delim;
                    var index = parserInput.i;
                    do {
                        e = this.comment();
                        if (e) {
                            entities.push(e);
                            continue;
                        }
                        e = this.addition() || this.entity();
                        if (e instanceof tree_1.default.Comment) {
                            e = null;
                        }
                        if (e) {
                            entities.push(e);
                            // operations do not allow keyword "/" dimension (e.g. small/20px) so we support that here
                            if (!parserInput.peek(/^\/[\/*]/)) {
                                delim = parserInput.$char('/');
                                if (delim) {
                                    entities.push(new (tree_1.default.Anonymous)(delim, index));
                                }
                            }
                        }
                    } while (e);
                    if (entities.length > 0) {
                        return new (tree_1.default.Expression)(entities);
                    }
                },
                property: function () {
                    var name = parserInput.$re(/^(\*?-?[_a-zA-Z0-9-]+)\s*:/);
                    if (name) {
                        return name[1];
                    }
                },
                ruleProperty: function () {
                    var name = [];
                    var index = [];
                    var s;
                    var k;
                    parserInput.save();
                    var simpleProperty = parserInput.$re(/^([_a-zA-Z0-9-]+)\s*:/);
                    if (simpleProperty) {
                        name = [new (tree_1.default.Keyword)(simpleProperty[1])];
                        parserInput.forget();
                        return name;
                    }
                    function match(re) {
                        var i = parserInput.i;
                        var chunk = parserInput.$re(re);
                        if (chunk) {
                            index.push(i);
                            return name.push(chunk[1]);
                        }
                    }
                    match(/^(\*?)/);
                    while (true) {
                        if (!match(/^((?:[\w-]+)|(?:[@\$]\{[\w-]+\}))/)) {
                            break;
                        }
                    }
                    if ((name.length > 1) && match(/^((?:\+_|\+)?)\s*:/)) {
                        parserInput.forget();
                        // at last, we have the complete match now. move forward,
                        // convert name particles to tree objects and return:
                        if (name[0] === '') {
                            name.shift();
                            index.shift();
                        }
                        for (k = 0; k < name.length; k++) {
                            s = name[k];
                            name[k] = (s.charAt(0) !== '@' && s.charAt(0) !== '$') ?
                                new (tree_1.default.Keyword)(s) :
                                (s.charAt(0) === '@' ?
                                    new (tree_1.default.Variable)("@" + s.slice(2, -1), index[k], fileInfo) :
                                    new (tree_1.default.Property)("$" + s.slice(2, -1), index[k], fileInfo));
                        }
                        return name;
                    }
                    parserInput.restore();
                }
            }
        };
    };
    Parser.serializeVars = function (vars) {
        var s = '';
        for (var name_1 in vars) {
            if (Object.hasOwnProperty.call(vars, name_1)) {
                var value = vars[name_1];
                s += ((name_1[0] === '@') ? '' : '@') + name_1 + ": " + value + ((String(value).slice(-1) === ';') ? '' : ';');
            }
        }
        return s;
    };
    exports.default = Parser;
},
/* less/lib/less/parser/parser-input.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var chunker_1 = tslib_1.__importDefault(require(171) /* ./chunker */);
    exports.default = (function () {
        var // Less input string
        input;
        var // current chunk
        j;
        var // holds state for backtracking
        saveStack = [];
        var // furthest index the parser has gone to
        furthest;
        var // if this is furthest we got to, this is the probably cause
        furthestPossibleErrorMessage;
        var // chunkified input
        chunks;
        var // current chunk
        current;
        var // index of current chunk, in `input`
        currentPos;
        var parserInput = {};
        var CHARCODE_SPACE = 32;
        var CHARCODE_TAB = 9;
        var CHARCODE_LF = 10;
        var CHARCODE_CR = 13;
        var CHARCODE_PLUS = 43;
        var CHARCODE_COMMA = 44;
        var CHARCODE_FORWARD_SLASH = 47;
        var CHARCODE_9 = 57;
        function skipWhitespace(length) {
            var oldi = parserInput.i;
            var oldj = j;
            var curr = parserInput.i - currentPos;
            var endIndex = parserInput.i + current.length - curr;
            var mem = (parserInput.i += length);
            var inp = input;
            var c;
            var nextChar;
            var comment;
            for (; parserInput.i < endIndex; parserInput.i++) {
                c = inp.charCodeAt(parserInput.i);
                if (parserInput.autoCommentAbsorb && c === CHARCODE_FORWARD_SLASH) {
                    nextChar = inp.charAt(parserInput.i + 1);
                    if (nextChar === '/') {
                        comment = { index: parserInput.i, isLineComment: true };
                        var nextNewLine = inp.indexOf('\n', parserInput.i + 2);
                        if (nextNewLine < 0) {
                            nextNewLine = endIndex;
                        }
                        parserInput.i = nextNewLine;
                        comment.text = inp.substr(comment.index, parserInput.i - comment.index);
                        parserInput.commentStore.push(comment);
                        continue;
                    }
                    else if (nextChar === '*') {
                        var nextStarSlash = inp.indexOf('*/', parserInput.i + 2);
                        if (nextStarSlash >= 0) {
                            comment = {
                                index: parserInput.i,
                                text: inp.substr(parserInput.i, nextStarSlash + 2 - parserInput.i),
                                isLineComment: false
                            };
                            parserInput.i += comment.text.length - 1;
                            parserInput.commentStore.push(comment);
                            continue;
                        }
                    }
                    break;
                }
                if ((c !== CHARCODE_SPACE) && (c !== CHARCODE_LF) && (c !== CHARCODE_TAB) && (c !== CHARCODE_CR)) {
                    break;
                }
            }
            current = current.slice(length + parserInput.i - mem + curr);
            currentPos = parserInput.i;
            if (!current.length) {
                if (j < chunks.length - 1) {
                    current = chunks[++j];
                    skipWhitespace(0); // skip space at the beginning of a chunk
                    return true; // things changed
                }
                parserInput.finished = true;
            }
            return oldi !== parserInput.i || oldj !== j;
        }
        parserInput.save = function () {
            currentPos = parserInput.i;
            saveStack.push({ current: current, i: parserInput.i, j: j });
        };
        parserInput.restore = function (possibleErrorMessage) {
            if (parserInput.i > furthest || (parserInput.i === furthest && possibleErrorMessage && !furthestPossibleErrorMessage)) {
                furthest = parserInput.i;
                furthestPossibleErrorMessage = possibleErrorMessage;
            }
            var state = saveStack.pop();
            current = state.current;
            currentPos = parserInput.i = state.i;
            j = state.j;
        };
        parserInput.forget = function () {
            saveStack.pop();
        };
        parserInput.isWhitespace = function (offset) {
            var pos = parserInput.i + (offset || 0);
            var code = input.charCodeAt(pos);
            return (code === CHARCODE_SPACE || code === CHARCODE_CR || code === CHARCODE_TAB || code === CHARCODE_LF);
        };
        // Specialization of $(tok)
        parserInput.$re = function (tok) {
            if (parserInput.i > currentPos) {
                current = current.slice(parserInput.i - currentPos);
                currentPos = parserInput.i;
            }
            var m = tok.exec(current);
            if (!m) {
                return null;
            }
            skipWhitespace(m[0].length);
            if (typeof m === 'string') {
                return m;
            }
            return m.length === 1 ? m[0] : m;
        };
        parserInput.$char = function (tok) {
            if (input.charAt(parserInput.i) !== tok) {
                return null;
            }
            skipWhitespace(1);
            return tok;
        };
        parserInput.$str = function (tok) {
            var tokLength = tok.length;
            // https://jsperf.com/string-startswith/21
            for (var i = 0; i < tokLength; i++) {
                if (input.charAt(parserInput.i + i) !== tok.charAt(i)) {
                    return null;
                }
            }
            skipWhitespace(tokLength);
            return tok;
        };
        parserInput.$quoted = function (loc) {
            var pos = loc || parserInput.i;
            var startChar = input.charAt(pos);
            if (startChar !== '\'' && startChar !== '"') {
                return;
            }
            var length = input.length;
            var currentPosition = pos;
            for (var i = 1; i + currentPosition < length; i++) {
                var nextChar = input.charAt(i + currentPosition);
                switch (nextChar) {
                    case '\\':
                        i++;
                        continue;
                    case '\r':
                    case '\n':
                        break;
                    case startChar:
                        var str = input.substr(currentPosition, i + 1);
                        if (!loc && loc !== 0) {
                            skipWhitespace(i + 1);
                            return str;
                        }
                        return [startChar, str];
                    default:
                }
            }
            return null;
        };
        /**
         * Permissive parsing. Ignores everything except matching {} [] () and quotes
         * until matching token (outside of blocks)
         */
        parserInput.$parseUntil = function (tok) {
            var quote = '';
            var returnVal = null;
            var inComment = false;
            var blockDepth = 0;
            var blockStack = [];
            var parseGroups = [];
            var length = input.length;
            var startPos = parserInput.i;
            var lastPos = parserInput.i;
            var i = parserInput.i;
            var loop = true;
            var testChar;
            if (typeof tok === 'string') {
                testChar = function (char) { return char === tok; };
            }
            else {
                testChar = function (char) { return tok.test(char); };
            }
            do {
                var prevChar = void 0;
                var nextChar = input.charAt(i);
                if (blockDepth === 0 && testChar(nextChar)) {
                    returnVal = input.substr(lastPos, i - lastPos);
                    if (returnVal) {
                        parseGroups.push(returnVal);
                    }
                    else {
                        parseGroups.push(' ');
                    }
                    returnVal = parseGroups;
                    skipWhitespace(i - startPos);
                    loop = false;
                }
                else {
                    if (inComment) {
                        if (nextChar === '*' &&
                            input.charAt(i + 1) === '/') {
                            i++;
                            blockDepth--;
                            inComment = false;
                        }
                        i++;
                        continue;
                    }
                    switch (nextChar) {
                        case '\\':
                            i++;
                            nextChar = input.charAt(i);
                            parseGroups.push(input.substr(lastPos, i - lastPos + 1));
                            lastPos = i + 1;
                            break;
                        case '/':
                            if (input.charAt(i + 1) === '*') {
                                i++;
                                inComment = true;
                                blockDepth++;
                            }
                            break;
                        case '\'':
                        case '"':
                            quote = parserInput.$quoted(i);
                            if (quote) {
                                parseGroups.push(input.substr(lastPos, i - lastPos), quote);
                                i += quote[1].length - 1;
                                lastPos = i + 1;
                            }
                            else {
                                skipWhitespace(i - startPos);
                                returnVal = nextChar;
                                loop = false;
                            }
                            break;
                        case '{':
                            blockStack.push('}');
                            blockDepth++;
                            break;
                        case '(':
                            blockStack.push(')');
                            blockDepth++;
                            break;
                        case '[':
                            blockStack.push(']');
                            blockDepth++;
                            break;
                        case '}':
                        case ')':
                        case ']':
                            var expected = blockStack.pop();
                            if (nextChar === expected) {
                                blockDepth--;
                            }
                            else {
                                // move the parser to the error and return expected
                                skipWhitespace(i - startPos);
                                returnVal = expected;
                                loop = false;
                            }
                    }
                    i++;
                    if (i > length) {
                        loop = false;
                    }
                }
                prevChar = nextChar;
            } while (loop);
            return returnVal ? returnVal : null;
        };
        parserInput.autoCommentAbsorb = true;
        parserInput.commentStore = [];
        parserInput.finished = false;
        // Same as $(), but don't change the state of the parser,
        // just return the match.
        parserInput.peek = function (tok) {
            if (typeof tok === 'string') {
                // https://jsperf.com/string-startswith/21
                for (var i = 0; i < tok.length; i++) {
                    if (input.charAt(parserInput.i + i) !== tok.charAt(i)) {
                        return false;
                    }
                }
                return true;
            }
            else {
                return tok.test(current);
            }
        };
        // Specialization of peek()
        // TODO remove or change some currentChar calls to peekChar
        parserInput.peekChar = function (tok) { return input.charAt(parserInput.i) === tok; };
        parserInput.currentChar = function () { return input.charAt(parserInput.i); };
        parserInput.prevChar = function () { return input.charAt(parserInput.i - 1); };
        parserInput.getInput = function () { return input; };
        parserInput.peekNotNumeric = function () {
            var c = input.charCodeAt(parserInput.i);
            // Is the first char of the dimension 0-9, '.', '+' or '-'
            return (c > CHARCODE_9 || c < CHARCODE_PLUS) || c === CHARCODE_FORWARD_SLASH || c === CHARCODE_COMMA;
        };
        parserInput.start = function (str, chunkInput, failFunction) {
            input = str;
            parserInput.i = j = currentPos = furthest = 0;
            // chunking apparently makes things quicker (but my tests indicate
            // it might actually make things slower in node at least)
            // and it is a non-perfect parse - it can't recognise
            // unquoted urls, meaning it can't distinguish comments
            // meaning comments with quotes or {}() in them get 'counted'
            // and then lead to parse errors.
            // In addition if the chunking chunks in the wrong place we might
            // not be able to parse a parser statement in one go
            // this is officially deprecated but can be switched on via an option
            // in the case it causes too much performance issues.
            if (chunkInput) {
                chunks = chunker_1.default(str, failFunction);
            }
            else {
                chunks = [str];
            }
            current = chunks[0];
            skipWhitespace(0);
        };
        parserInput.end = function () {
            var message;
            var isFinished = parserInput.i >= input.length;
            if (parserInput.i < furthest) {
                message = furthestPossibleErrorMessage;
                parserInput.i = furthest;
            }
            return {
                isFinished: isFinished,
                furthest: parserInput.i,
                furthestPossibleErrorMessage: message,
                furthestReachedEnd: parserInput.i >= input.length - 1,
                furthestChar: input[parserInput.i]
            };
        };
        return parserInput;
    });
},
/* less/lib/less/parser/chunker.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    // Split the input into chunks.
    function default_1(input, fail) {
        var len = input.length;
        var level = 0;
        var parenLevel = 0;
        var lastOpening;
        var lastOpeningParen;
        var lastMultiComment;
        var lastMultiCommentEndBrace;
        var chunks = [];
        var emitFrom = 0;
        var chunkerCurrentIndex;
        var currentChunkStartIndex;
        var cc;
        var cc2;
        var matched;
        function emitChunk(force) {
            var len = chunkerCurrentIndex - emitFrom;
            if (((len < 512) && !force) || !len) {
                return;
            }
            chunks.push(input.slice(emitFrom, chunkerCurrentIndex + 1));
            emitFrom = chunkerCurrentIndex + 1;
        }
        for (chunkerCurrentIndex = 0; chunkerCurrentIndex < len; chunkerCurrentIndex++) {
            cc = input.charCodeAt(chunkerCurrentIndex);
            if (((cc >= 97) && (cc <= 122)) || (cc < 34)) {
                // a-z or whitespace
                continue;
            }
            switch (cc) {
                case 40: // (
                    parenLevel++;
                    lastOpeningParen = chunkerCurrentIndex;
                    continue;
                case 41: // )
                    if (--parenLevel < 0) {
                        return fail('missing opening `(`', chunkerCurrentIndex);
                    }
                    continue;
                case 59: // ;
                    if (!parenLevel) {
                        emitChunk();
                    }
                    continue;
                case 123: // {
                    level++;
                    lastOpening = chunkerCurrentIndex;
                    continue;
                case 125: // }
                    if (--level < 0) {
                        return fail('missing opening `{`', chunkerCurrentIndex);
                    }
                    if (!level && !parenLevel) {
                        emitChunk();
                    }
                    continue;
                case 92: // \
                    if (chunkerCurrentIndex < len - 1) {
                        chunkerCurrentIndex++;
                        continue;
                    }
                    return fail('unescaped `\\`', chunkerCurrentIndex);
                case 34:
                case 39:
                case 96: // ", ' and `
                    matched = 0;
                    currentChunkStartIndex = chunkerCurrentIndex;
                    for (chunkerCurrentIndex = chunkerCurrentIndex + 1; chunkerCurrentIndex < len; chunkerCurrentIndex++) {
                        cc2 = input.charCodeAt(chunkerCurrentIndex);
                        if (cc2 > 96) {
                            continue;
                        }
                        if (cc2 == cc) {
                            matched = 1;
                            break;
                        }
                        if (cc2 == 92) { // \
                            if (chunkerCurrentIndex == len - 1) {
                                return fail('unescaped `\\`', chunkerCurrentIndex);
                            }
                            chunkerCurrentIndex++;
                        }
                    }
                    if (matched) {
                        continue;
                    }
                    return fail("unmatched `" + String.fromCharCode(cc) + "`", currentChunkStartIndex);
                case 47: // /, check for comment
                    if (parenLevel || (chunkerCurrentIndex == len - 1)) {
                        continue;
                    }
                    cc2 = input.charCodeAt(chunkerCurrentIndex + 1);
                    if (cc2 == 47) {
                        // //, find lnfeed
                        for (chunkerCurrentIndex = chunkerCurrentIndex + 2; chunkerCurrentIndex < len; chunkerCurrentIndex++) {
                            cc2 = input.charCodeAt(chunkerCurrentIndex);
                            if ((cc2 <= 13) && ((cc2 == 10) || (cc2 == 13))) {
                                break;
                            }
                        }
                    }
                    else if (cc2 == 42) {
                        // /*, find */
                        lastMultiComment = currentChunkStartIndex = chunkerCurrentIndex;
                        for (chunkerCurrentIndex = chunkerCurrentIndex + 2; chunkerCurrentIndex < len - 1; chunkerCurrentIndex++) {
                            cc2 = input.charCodeAt(chunkerCurrentIndex);
                            if (cc2 == 125) {
                                lastMultiCommentEndBrace = chunkerCurrentIndex;
                            }
                            if (cc2 != 42) {
                                continue;
                            }
                            if (input.charCodeAt(chunkerCurrentIndex + 1) == 47) {
                                break;
                            }
                        }
                        if (chunkerCurrentIndex == len - 1) {
                            return fail('missing closing `*/`', currentChunkStartIndex);
                        }
                        chunkerCurrentIndex++;
                    }
                    continue;
                case 42: // *, check for unmatched */
                    if ((chunkerCurrentIndex < len - 1) && (input.charCodeAt(chunkerCurrentIndex + 1) == 47)) {
                        return fail('unmatched `/*`', chunkerCurrentIndex);
                    }
                    continue;
            }
        }
        if (level !== 0) {
            if ((lastMultiComment > lastOpening) && (lastMultiCommentEndBrace > lastMultiComment)) {
                return fail('missing closing `}` or `*/`', lastOpening);
            }
            else {
                return fail('missing closing `}`', lastOpening);
            }
        }
        else if (parenLevel !== 0) {
            return fail('missing closing `)`', lastOpeningParen);
        }
        emitChunk(true);
        return chunks;
    }
    exports.default = default_1;
    ;
},
/* less/lib/less/functions/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var function_registry_1 = tslib_1.__importDefault(require(133) /* ./function-registry */);
    var function_caller_1 = tslib_1.__importDefault(require(141) /* ./function-caller */);
    var boolean_1 = tslib_1.__importDefault(require(173) /* ./boolean */);
    var default_1 = tslib_1.__importDefault(require(134) /* ./default */);
    var color_1 = tslib_1.__importDefault(require(174) /* ./color */);
    var color_blending_1 = tslib_1.__importDefault(require(175) /* ./color-blending */);
    var data_uri_1 = tslib_1.__importDefault(require(176) /* ./data-uri */);
    var list_1 = tslib_1.__importDefault(require(177) /* ./list */);
    var math_1 = tslib_1.__importDefault(require(178) /* ./math */);
    var number_1 = tslib_1.__importDefault(require(180) /* ./number */);
    var string_1 = tslib_1.__importDefault(require(181) /* ./string */);
    var svg_1 = tslib_1.__importDefault(require(182) /* ./svg */);
    var types_1 = tslib_1.__importDefault(require(183) /* ./types */);
    exports.default = (function (environment) {
        var functions = { functionRegistry: function_registry_1.default, functionCaller: function_caller_1.default };
        // register functions
        function_registry_1.default.addMultiple(boolean_1.default);
        function_registry_1.default.add('default', default_1.default.eval.bind(default_1.default));
        function_registry_1.default.addMultiple(color_1.default);
        function_registry_1.default.addMultiple(color_blending_1.default);
        function_registry_1.default.addMultiple(data_uri_1.default(environment));
        function_registry_1.default.addMultiple(list_1.default);
        function_registry_1.default.addMultiple(math_1.default);
        function_registry_1.default.addMultiple(number_1.default);
        function_registry_1.default.addMultiple(string_1.default);
        function_registry_1.default.addMultiple(svg_1.default(environment));
        function_registry_1.default.addMultiple(types_1.default);
        return functions;
    });
},
/* less/lib/less/functions/boolean.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ../tree/anonymous */);
    var keyword_1 = tslib_1.__importDefault(require(128) /* ../tree/keyword */);
    function boolean(condition) {
        return condition ? keyword_1.default.True : keyword_1.default.False;
    }
    /**
     * Functions with evalArgs set to false are sent context
     * as the first argument.
     */
    function If(context, condition, trueValue, falseValue) {
        return condition.eval(context) ? trueValue.eval(context)
            : (falseValue ? falseValue.eval(context) : new anonymous_1.default);
    }
    If.evalArgs = false;
    function isdefined(context, variable) {
        try {
            variable.eval(context);
            return keyword_1.default.True;
        }
        catch (e) {
            return keyword_1.default.False;
        }
    }
    isdefined.evalArgs = false;
    exports.default = { isdefined: isdefined, boolean: boolean, 'if': If };
},
/* less/lib/less/functions/color.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var dimension_1 = tslib_1.__importDefault(require(137) /* ../tree/dimension */);
    var color_1 = tslib_1.__importDefault(require(114) /* ../tree/color */);
    var quoted_1 = tslib_1.__importDefault(require(145) /* ../tree/quoted */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ../tree/anonymous */);
    var expression_1 = tslib_1.__importDefault(require(142) /* ../tree/expression */);
    var operation_1 = tslib_1.__importDefault(require(136) /* ../tree/operation */);
    var colorFunctions;
    function clamp(val) {
        return Math.min(1, Math.max(0, val));
    }
    function hsla(origColor, hsl) {
        var color = colorFunctions.hsla(hsl.h, hsl.s, hsl.l, hsl.a);
        if (color) {
            if (origColor.value &&
                /^(rgb|hsl)/.test(origColor.value)) {
                color.value = origColor.value;
            }
            else {
                color.value = 'rgb';
            }
            return color;
        }
    }
    function toHSL(color) {
        if (color.toHSL) {
            return color.toHSL();
        }
        else {
            throw new Error('Argument cannot be evaluated to a color');
        }
    }
    function toHSV(color) {
        if (color.toHSV) {
            return color.toHSV();
        }
        else {
            throw new Error('Argument cannot be evaluated to a color');
        }
    }
    function number(n) {
        if (n instanceof dimension_1.default) {
            return parseFloat(n.unit.is('%') ? n.value / 100 : n.value);
        }
        else if (typeof n === 'number') {
            return n;
        }
        else {
            throw {
                type: 'Argument',
                message: 'color functions take numbers as parameters'
            };
        }
    }
    function scaled(n, size) {
        if (n instanceof dimension_1.default && n.unit.is('%')) {
            return parseFloat(n.value * size / 100);
        }
        else {
            return number(n);
        }
    }
    colorFunctions = {
        rgb: function (r, g, b) {
            var a = 1;
            /**
             * Comma-less syntax
             *   e.g. rgb(0 128 255 / 50%)
             */
            if (r instanceof expression_1.default) {
                var val = r.value;
                r = val[0];
                g = val[1];
                b = val[2];
                /**
                 * @todo - should this be normalized in
                 *   function caller? Or parsed differently?
                 */
                if (b instanceof operation_1.default) {
                    var op = b;
                    b = op.operands[0];
                    a = op.operands[1];
                }
            }
            var color = colorFunctions.rgba(r, g, b, a);
            if (color) {
                color.value = 'rgb';
                return color;
            }
        },
        rgba: function (r, g, b, a) {
            try {
                if (r instanceof color_1.default) {
                    if (g) {
                        a = number(g);
                    }
                    else {
                        a = r.alpha;
                    }
                    return new color_1.default(r.rgb, a, 'rgba');
                }
                var rgb = [r, g, b].map(function (c) { return scaled(c, 255); });
                a = number(a);
                return new color_1.default(rgb, a, 'rgba');
            }
            catch (e) { }
        },
        hsl: function (h, s, l) {
            var a = 1;
            if (h instanceof expression_1.default) {
                var val = h.value;
                h = val[0];
                s = val[1];
                l = val[2];
                if (l instanceof operation_1.default) {
                    var op = l;
                    l = op.operands[0];
                    a = op.operands[1];
                }
            }
            var color = colorFunctions.hsla(h, s, l, a);
            if (color) {
                color.value = 'hsl';
                return color;
            }
        },
        hsla: function (h, s, l, a) {
            try {
                if (h instanceof color_1.default) {
                    if (s) {
                        a = number(s);
                    }
                    else {
                        a = h.alpha;
                    }
                    return new color_1.default(h.rgb, a, 'hsla');
                }
                var m1_1;
                var m2_1;
                function hue(h) {
                    h = h < 0 ? h + 1 : (h > 1 ? h - 1 : h);
                    if (h * 6 < 1) {
                        return m1_1 + (m2_1 - m1_1) * h * 6;
                    }
                    else if (h * 2 < 1) {
                        return m2_1;
                    }
                    else if (h * 3 < 2) {
                        return m1_1 + (m2_1 - m1_1) * (2 / 3 - h) * 6;
                    }
                    else {
                        return m1_1;
                    }
                }
                h = (number(h) % 360) / 360;
                s = clamp(number(s));
                l = clamp(number(l));
                a = clamp(number(a));
                m2_1 = l <= 0.5 ? l * (s + 1) : l + s - l * s;
                m1_1 = l * 2 - m2_1;
                var rgb = [
                    hue(h + 1 / 3) * 255,
                    hue(h) * 255,
                    hue(h - 1 / 3) * 255
                ];
                a = number(a);
                return new color_1.default(rgb, a, 'hsla');
            }
            catch (e) { }
        },
        hsv: function (h, s, v) {
            return colorFunctions.hsva(h, s, v, 1.0);
        },
        hsva: function (h, s, v, a) {
            h = ((number(h) % 360) / 360) * 360;
            s = number(s);
            v = number(v);
            a = number(a);
            var i;
            var f;
            i = Math.floor((h / 60) % 6);
            f = (h / 60) - i;
            var vs = [v,
                v * (1 - s),
                v * (1 - f * s),
                v * (1 - (1 - f) * s)];
            var perm = [[0, 3, 1],
                [2, 0, 1],
                [1, 0, 3],
                [1, 2, 0],
                [3, 1, 0],
                [0, 1, 2]];
            return colorFunctions.rgba(vs[perm[i][0]] * 255, vs[perm[i][1]] * 255, vs[perm[i][2]] * 255, a);
        },
        hue: function (color) {
            return new dimension_1.default(toHSL(color).h);
        },
        saturation: function (color) {
            return new dimension_1.default(toHSL(color).s * 100, '%');
        },
        lightness: function (color) {
            return new dimension_1.default(toHSL(color).l * 100, '%');
        },
        hsvhue: function (color) {
            return new dimension_1.default(toHSV(color).h);
        },
        hsvsaturation: function (color) {
            return new dimension_1.default(toHSV(color).s * 100, '%');
        },
        hsvvalue: function (color) {
            return new dimension_1.default(toHSV(color).v * 100, '%');
        },
        red: function (color) {
            return new dimension_1.default(color.rgb[0]);
        },
        green: function (color) {
            return new dimension_1.default(color.rgb[1]);
        },
        blue: function (color) {
            return new dimension_1.default(color.rgb[2]);
        },
        alpha: function (color) {
            return new dimension_1.default(toHSL(color).a);
        },
        luma: function (color) {
            return new dimension_1.default(color.luma() * color.alpha * 100, '%');
        },
        luminance: function (color) {
            var luminance = (0.2126 * color.rgb[0] / 255) +
                (0.7152 * color.rgb[1] / 255) +
                (0.0722 * color.rgb[2] / 255);
            return new dimension_1.default(luminance * color.alpha * 100, '%');
        },
        saturate: function (color, amount, method) {
            // filter: saturate(3.2);
            // should be kept as is, so check for color
            if (!color.rgb) {
                return null;
            }
            var hsl = toHSL(color);
            if (typeof method !== 'undefined' && method.value === 'relative') {
                hsl.s += hsl.s * amount.value / 100;
            }
            else {
                hsl.s += amount.value / 100;
            }
            hsl.s = clamp(hsl.s);
            return hsla(color, hsl);
        },
        desaturate: function (color, amount, method) {
            var hsl = toHSL(color);
            if (typeof method !== 'undefined' && method.value === 'relative') {
                hsl.s -= hsl.s * amount.value / 100;
            }
            else {
                hsl.s -= amount.value / 100;
            }
            hsl.s = clamp(hsl.s);
            return hsla(color, hsl);
        },
        lighten: function (color, amount, method) {
            var hsl = toHSL(color);
            if (typeof method !== 'undefined' && method.value === 'relative') {
                hsl.l += hsl.l * amount.value / 100;
            }
            else {
                hsl.l += amount.value / 100;
            }
            hsl.l = clamp(hsl.l);
            return hsla(color, hsl);
        },
        darken: function (color, amount, method) {
            var hsl = toHSL(color);
            if (typeof method !== 'undefined' && method.value === 'relative') {
                hsl.l -= hsl.l * amount.value / 100;
            }
            else {
                hsl.l -= amount.value / 100;
            }
            hsl.l = clamp(hsl.l);
            return hsla(color, hsl);
        },
        fadein: function (color, amount, method) {
            var hsl = toHSL(color);
            if (typeof method !== 'undefined' && method.value === 'relative') {
                hsl.a += hsl.a * amount.value / 100;
            }
            else {
                hsl.a += amount.value / 100;
            }
            hsl.a = clamp(hsl.a);
            return hsla(color, hsl);
        },
        fadeout: function (color, amount, method) {
            var hsl = toHSL(color);
            if (typeof method !== 'undefined' && method.value === 'relative') {
                hsl.a -= hsl.a * amount.value / 100;
            }
            else {
                hsl.a -= amount.value / 100;
            }
            hsl.a = clamp(hsl.a);
            return hsla(color, hsl);
        },
        fade: function (color, amount) {
            var hsl = toHSL(color);
            hsl.a = amount.value / 100;
            hsl.a = clamp(hsl.a);
            return hsla(color, hsl);
        },
        spin: function (color, amount) {
            var hsl = toHSL(color);
            var hue = (hsl.h + amount.value) % 360;
            hsl.h = hue < 0 ? 360 + hue : hue;
            return hsla(color, hsl);
        },
        //
        // Copyright (c) 2006-2009 Hampton Catlin, Natalie Weizenbaum, and Chris Eppstein
        // http://sass-lang.com
        //
        mix: function (color1, color2, weight) {
            if (!weight) {
                weight = new dimension_1.default(50);
            }
            var p = weight.value / 100.0;
            var w = p * 2 - 1;
            var a = toHSL(color1).a - toHSL(color2).a;
            var w1 = (((w * a == -1) ? w : (w + a) / (1 + w * a)) + 1) / 2.0;
            var w2 = 1 - w1;
            var rgb = [color1.rgb[0] * w1 + color2.rgb[0] * w2,
                color1.rgb[1] * w1 + color2.rgb[1] * w2,
                color1.rgb[2] * w1 + color2.rgb[2] * w2];
            var alpha = color1.alpha * p + color2.alpha * (1 - p);
            return new color_1.default(rgb, alpha);
        },
        greyscale: function (color) {
            return colorFunctions.desaturate(color, new dimension_1.default(100));
        },
        contrast: function (color, dark, light, threshold) {
            // filter: contrast(3.2);
            // should be kept as is, so check for color
            if (!color.rgb) {
                return null;
            }
            if (typeof light === 'undefined') {
                light = colorFunctions.rgba(255, 255, 255, 1.0);
            }
            if (typeof dark === 'undefined') {
                dark = colorFunctions.rgba(0, 0, 0, 1.0);
            }
            // Figure out which is actually light and dark:
            if (dark.luma() > light.luma()) {
                var t = light;
                light = dark;
                dark = t;
            }
            if (typeof threshold === 'undefined') {
                threshold = 0.43;
            }
            else {
                threshold = number(threshold);
            }
            if (color.luma() < threshold) {
                return light;
            }
            else {
                return dark;
            }
        },
        // Changes made in 2.7.0 - Reverted in 3.0.0
        // contrast: function (color, color1, color2, threshold) {
        //     // Return which of `color1` and `color2` has the greatest contrast with `color`
        //     // according to the standard WCAG contrast ratio calculation.
        //     // http://www.w3.org/TR/WCAG20/#contrast-ratiodef
        //     // The threshold param is no longer used, in line with SASS.
        //     // filter: contrast(3.2);
        //     // should be kept as is, so check for color
        //     if (!color.rgb) {
        //         return null;
        //     }
        //     if (typeof color1 === 'undefined') {
        //         color1 = colorFunctions.rgba(0, 0, 0, 1.0);
        //     }
        //     if (typeof color2 === 'undefined') {
        //         color2 = colorFunctions.rgba(255, 255, 255, 1.0);
        //     }
        //     var contrast1, contrast2;
        //     var luma = color.luma();
        //     var luma1 = color1.luma();
        //     var luma2 = color2.luma();
        //     // Calculate contrast ratios for each color
        //     if (luma > luma1) {
        //         contrast1 = (luma + 0.05) / (luma1 + 0.05);
        //     } else {
        //         contrast1 = (luma1 + 0.05) / (luma + 0.05);
        //     }
        //     if (luma > luma2) {
        //         contrast2 = (luma + 0.05) / (luma2 + 0.05);
        //     } else {
        //         contrast2 = (luma2 + 0.05) / (luma + 0.05);
        //     }
        //     if (contrast1 > contrast2) {
        //         return color1;
        //     } else {
        //         return color2;
        //     }
        // },
        argb: function (color) {
            return new anonymous_1.default(color.toARGB());
        },
        color: function (c) {
            if ((c instanceof quoted_1.default) &&
                (/^#([A-Fa-f0-9]{8}|[A-Fa-f0-9]{6}|[A-Fa-f0-9]{3,4})$/i.test(c.value))) {
                var val = c.value.slice(1);
                return new color_1.default(val, undefined, "#" + val);
            }
            if ((c instanceof color_1.default) || (c = color_1.default.fromKeyword(c.value))) {
                c.value = undefined;
                return c;
            }
            throw {
                type: 'Argument',
                message: 'argument must be a color keyword or 3|4|6|8 digit hex e.g. #FFF'
            };
        },
        tint: function (color, amount) {
            return colorFunctions.mix(colorFunctions.rgb(255, 255, 255), color, amount);
        },
        shade: function (color, amount) {
            return colorFunctions.mix(colorFunctions.rgb(0, 0, 0), color, amount);
        }
    };
    exports.default = colorFunctions;
},
/* less/lib/less/functions/color-blending.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var color_1 = tslib_1.__importDefault(require(114) /* ../tree/color */);
    // Color Blending
    // ref: http://www.w3.org/TR/compositing-1
    function colorBlend(mode, color1, color2) {
        var ab = color1.alpha; // result
        var // backdrop
        cb;
        var as = color2.alpha;
        var // source
        cs;
        var ar;
        var cr;
        var r = [];
        ar = as + ab * (1 - as);
        for (var i = 0; i < 3; i++) {
            cb = color1.rgb[i] / 255;
            cs = color2.rgb[i] / 255;
            cr = mode(cb, cs);
            if (ar) {
                cr = (as * cs + ab * (cb -
                    as * (cb + cs - cr))) / ar;
            }
            r[i] = cr * 255;
        }
        return new color_1.default(r, ar);
    }
    var colorBlendModeFunctions = {
        multiply: function (cb, cs) {
            return cb * cs;
        },
        screen: function (cb, cs) {
            return cb + cs - cb * cs;
        },
        overlay: function (cb, cs) {
            cb *= 2;
            return (cb <= 1) ?
                colorBlendModeFunctions.multiply(cb, cs) :
                colorBlendModeFunctions.screen(cb - 1, cs);
        },
        softlight: function (cb, cs) {
            var d = 1;
            var e = cb;
            if (cs > 0.5) {
                e = 1;
                d = (cb > 0.25) ? Math.sqrt(cb)
                    : ((16 * cb - 12) * cb + 4) * cb;
            }
            return cb - (1 - 2 * cs) * e * (d - cb);
        },
        hardlight: function (cb, cs) {
            return colorBlendModeFunctions.overlay(cs, cb);
        },
        difference: function (cb, cs) {
            return Math.abs(cb - cs);
        },
        exclusion: function (cb, cs) {
            return cb + cs - 2 * cb * cs;
        },
        // non-w3c functions:
        average: function (cb, cs) {
            return (cb + cs) / 2;
        },
        negation: function (cb, cs) {
            return 1 - Math.abs(cb + cs - 1);
        }
    };
    for (var f in colorBlendModeFunctions) {
        if (colorBlendModeFunctions.hasOwnProperty(f)) {
            colorBlend[f] = colorBlend.bind(null, colorBlendModeFunctions[f]);
        }
    }
    exports.default = colorBlend;
},
/* less/lib/less/functions/data-uri.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var quoted_1 = tslib_1.__importDefault(require(145) /* ../tree/quoted */);
    var url_1 = tslib_1.__importDefault(require(146) /* ../tree/url */);
    var utils = tslib_1.__importStar(require(121) /* ../utils */);
    var logger_1 = tslib_1.__importDefault(require(70) /* ../logger */);
    exports.default = (function (environment) {
        var fallback = function (functionThis, node) { return new url_1.default(node, functionThis.index, functionThis.currentFileInfo).eval(functionThis.context); };
        return { 'data-uri': function (mimetypeNode, filePathNode) {
                if (!filePathNode) {
                    filePathNode = mimetypeNode;
                    mimetypeNode = null;
                }
                var mimetype = mimetypeNode && mimetypeNode.value;
                var filePath = filePathNode.value;
                var currentFileInfo = this.currentFileInfo;
                var currentDirectory = currentFileInfo.rewriteUrls ?
                    currentFileInfo.currentDirectory : currentFileInfo.entryPath;
                var fragmentStart = filePath.indexOf('#');
                var fragment = '';
                if (fragmentStart !== -1) {
                    fragment = filePath.slice(fragmentStart);
                    filePath = filePath.slice(0, fragmentStart);
                }
                var context = utils.clone(this.context);
                context.rawBuffer = true;
                var fileManager = environment.getFileManager(filePath, currentDirectory, context, environment, true);
                if (!fileManager) {
                    return fallback(this, filePathNode);
                }
                var useBase64 = false;
                // detect the mimetype if not given
                if (!mimetypeNode) {
                    mimetype = environment.mimeLookup(filePath);
                    if (mimetype === 'image/svg+xml') {
                        useBase64 = false;
                    }
                    else {
                        // use base 64 unless it's an ASCII or UTF-8 format
                        var charset = environment.charsetLookup(mimetype);
                        useBase64 = ['US-ASCII', 'UTF-8'].indexOf(charset) < 0;
                    }
                    if (useBase64) {
                        mimetype += ';base64';
                    }
                }
                else {
                    useBase64 = /;base64$/.test(mimetype);
                }
                var fileSync = fileManager.loadFileSync(filePath, currentDirectory, context, environment);
                if (!fileSync.contents) {
                    logger_1.default.warn("Skipped data-uri embedding of " + filePath + " because file not found");
                    return fallback(this, filePathNode || mimetypeNode);
                }
                var buf = fileSync.contents;
                if (useBase64 && !environment.encodeBase64) {
                    return fallback(this, filePathNode);
                }
                buf = useBase64 ? environment.encodeBase64(buf) : encodeURIComponent(buf);
                var uri = "data:" + mimetype + "," + buf + fragment;
                return new url_1.default(new quoted_1.default("\"" + uri + "\"", uri, false, this.index, this.currentFileInfo), this.index, this.currentFileInfo);
            } };
    });
},
/* less/lib/less/functions/list.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var comment_1 = tslib_1.__importDefault(require(130) /* ../tree/comment */);
    var node_1 = tslib_1.__importDefault(require(113) /* ../tree/node */);
    var dimension_1 = tslib_1.__importDefault(require(137) /* ../tree/dimension */);
    var declaration_1 = tslib_1.__importDefault(require(126) /* ../tree/declaration */);
    var expression_1 = tslib_1.__importDefault(require(142) /* ../tree/expression */);
    var ruleset_1 = tslib_1.__importDefault(require(125) /* ../tree/ruleset */);
    var selector_1 = tslib_1.__importDefault(require(116) /* ../tree/selector */);
    var element_1 = tslib_1.__importDefault(require(117) /* ../tree/element */);
    var quoted_1 = tslib_1.__importDefault(require(145) /* ../tree/quoted */);
    var value_1 = tslib_1.__importDefault(require(127) /* ../tree/value */);
    var getItemsFromNode = function (node) {
        // handle non-array values as an array of length 1
        // return 'undefined' if index is invalid
        var items = Array.isArray(node.value) ?
            node.value : Array(node);
        return items;
    };
    exports.default = {
        _SELF: function (n) {
            return n;
        },
        '~': function () {
            var expr = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                expr[_i] = arguments[_i];
            }
            if (expr.length === 1) {
                return expr[0];
            }
            return new value_1.default(expr);
        },
        extract: function (values, index) {
            // (1-based index)
            index = index.value - 1;
            return getItemsFromNode(values)[index];
        },
        length: function (values) {
            return new dimension_1.default(getItemsFromNode(values).length);
        },
        /**
         * Creates a Less list of incremental values.
         * Modeled after Lodash's range function, also exists natively in PHP
         *
         * @param {Dimension} [start=1]
         * @param {Dimension} end  - e.g. 10 or 10px - unit is added to output
         * @param {Dimension} [step=1]
         */
        range: function (start, end, step) {
            var from;
            var to;
            var stepValue = 1;
            var list = [];
            if (end) {
                to = end;
                from = start.value;
                if (step) {
                    stepValue = step.value;
                }
            }
            else {
                from = 1;
                to = start;
            }
            for (var i = from; i <= to.value; i += stepValue) {
                list.push(new dimension_1.default(i, to.unit));
            }
            return new expression_1.default(list);
        },
        each: function (list, rs) {
            var _this = this;
            var rules = [];
            var newRules;
            var iterator;
            var tryEval = function (val) {
                if (val instanceof node_1.default) {
                    return val.eval(_this.context);
                }
                return val;
            };
            if (list.value && !(list instanceof quoted_1.default)) {
                if (Array.isArray(list.value)) {
                    iterator = list.value.map(tryEval);
                }
                else {
                    iterator = [tryEval(list.value)];
                }
            }
            else if (list.ruleset) {
                iterator = tryEval(list.ruleset).rules;
            }
            else if (list.rules) {
                iterator = list.rules.map(tryEval);
            }
            else if (Array.isArray(list)) {
                iterator = list.map(tryEval);
            }
            else {
                iterator = [tryEval(list)];
            }
            var valueName = '@value';
            var keyName = '@key';
            var indexName = '@index';
            if (rs.params) {
                valueName = rs.params[0] && rs.params[0].name;
                keyName = rs.params[1] && rs.params[1].name;
                indexName = rs.params[2] && rs.params[2].name;
                rs = rs.rules;
            }
            else {
                rs = rs.ruleset;
            }
            for (var i = 0; i < iterator.length; i++) {
                var key = void 0;
                var value = void 0;
                var item = iterator[i];
                if (item instanceof declaration_1.default) {
                    key = typeof item.name === 'string' ? item.name : item.name[0].value;
                    value = item.value;
                }
                else {
                    key = new dimension_1.default(i + 1);
                    value = item;
                }
                if (item instanceof comment_1.default) {
                    continue;
                }
                newRules = rs.rules.slice(0);
                if (valueName) {
                    newRules.push(new declaration_1.default(valueName, value, false, false, this.index, this.currentFileInfo));
                }
                if (indexName) {
                    newRules.push(new declaration_1.default(indexName, new dimension_1.default(i + 1), false, false, this.index, this.currentFileInfo));
                }
                if (keyName) {
                    newRules.push(new declaration_1.default(keyName, key, false, false, this.index, this.currentFileInfo));
                }
                rules.push(new ruleset_1.default([new (selector_1.default)([new element_1.default("", '&')])], newRules, rs.strictImports, rs.visibilityInfo()));
            }
            return new ruleset_1.default([new (selector_1.default)([new element_1.default("", '&')])], rules, rs.strictImports, rs.visibilityInfo()).eval(this.context);
        }
    };
},
/* less/lib/less/functions/math.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var math_helper_js_1 = tslib_1.__importDefault(require(179) /* ./math-helper.js */);
    var mathFunctions = {
        // name,  unit
        ceil: null,
        floor: null,
        sqrt: null,
        abs: null,
        tan: '',
        sin: '',
        cos: '',
        atan: 'rad',
        asin: 'rad',
        acos: 'rad'
    };
    for (var f in mathFunctions) {
        if (mathFunctions.hasOwnProperty(f)) {
            mathFunctions[f] = math_helper_js_1.default.bind(null, Math[f], mathFunctions[f]);
        }
    }
    mathFunctions.round = function (n, f) {
        var fraction = typeof f === 'undefined' ? 0 : f.value;
        return math_helper_js_1.default(function (num) { return num.toFixed(fraction); }, null, n);
    };
    exports.default = mathFunctions;
},
/* less/lib/less/functions/math-helper.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var dimension_1 = tslib_1.__importDefault(require(137) /* ../tree/dimension */);
    var MathHelper = function (fn, unit, n) {
        if (!(n instanceof dimension_1.default)) {
            throw { type: 'Argument', message: 'argument must be a number' };
        }
        if (unit == null) {
            unit = n.unit;
        }
        else {
            n = n.unify();
        }
        return new dimension_1.default(fn(parseFloat(n.value)), unit);
    };
    exports.default = MathHelper;
},
/* less/lib/less/functions/number.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var dimension_1 = tslib_1.__importDefault(require(137) /* ../tree/dimension */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ../tree/anonymous */);
    var math_helper_js_1 = tslib_1.__importDefault(require(179) /* ./math-helper.js */);
    var minMax = function (isMin, args) {
        args = Array.prototype.slice.call(args);
        switch (args.length) {
            case 0: throw { type: 'Argument', message: 'one or more arguments required' };
        }
        var i; // key is the unit.toString() for unified Dimension values,
        var j;
        var current;
        var currentUnified;
        var referenceUnified;
        var unit;
        var unitStatic;
        var unitClone;
        var // elems only contains original argument values.
        order = [];
        var values = {};
        // value is the index into the order array.
        for (i = 0; i < args.length; i++) {
            current = args[i];
            if (!(current instanceof dimension_1.default)) {
                if (Array.isArray(args[i].value)) {
                    Array.prototype.push.apply(args, Array.prototype.slice.call(args[i].value));
                }
                continue;
            }
            currentUnified = current.unit.toString() === '' && unitClone !== undefined ? new dimension_1.default(current.value, unitClone).unify() : current.unify();
            unit = currentUnified.unit.toString() === '' && unitStatic !== undefined ? unitStatic : currentUnified.unit.toString();
            unitStatic = unit !== '' && unitStatic === undefined || unit !== '' && order[0].unify().unit.toString() === '' ? unit : unitStatic;
            unitClone = unit !== '' && unitClone === undefined ? current.unit.toString() : unitClone;
            j = values[''] !== undefined && unit !== '' && unit === unitStatic ? values[''] : values[unit];
            if (j === undefined) {
                if (unitStatic !== undefined && unit !== unitStatic) {
                    throw { type: 'Argument', message: 'incompatible types' };
                }
                values[unit] = order.length;
                order.push(current);
                continue;
            }
            referenceUnified = order[j].unit.toString() === '' && unitClone !== undefined ? new dimension_1.default(order[j].value, unitClone).unify() : order[j].unify();
            if (isMin && currentUnified.value < referenceUnified.value ||
                !isMin && currentUnified.value > referenceUnified.value) {
                order[j] = current;
            }
        }
        if (order.length == 1) {
            return order[0];
        }
        args = order.map(function (a) { return a.toCSS(this.context); }).join(this.context.compress ? ',' : ', ');
        return new anonymous_1.default((isMin ? 'min' : 'max') + "(" + args + ")");
    };
    exports.default = {
        min: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            try {
                return minMax(true, args);
            }
            catch (e) { }
        },
        max: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            try {
                return minMax(false, args);
            }
            catch (e) { }
        },
        convert: function (val, unit) {
            return val.convertTo(unit.value);
        },
        pi: function () {
            return new dimension_1.default(Math.PI);
        },
        mod: function (a, b) {
            return new dimension_1.default(a.value % b.value, a.unit);
        },
        pow: function (x, y) {
            if (typeof x === 'number' && typeof y === 'number') {
                x = new dimension_1.default(x);
                y = new dimension_1.default(y);
            }
            else if (!(x instanceof dimension_1.default) || !(y instanceof dimension_1.default)) {
                throw { type: 'Argument', message: 'arguments must be numbers' };
            }
            return new dimension_1.default(Math.pow(x.value, y.value), x.unit);
        },
        percentage: function (n) {
            var result = math_helper_js_1.default(function (num) { return num * 100; }, '%', n);
            return result;
        }
    };
},
/* less/lib/less/functions/string.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var quoted_1 = tslib_1.__importDefault(require(145) /* ../tree/quoted */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ../tree/anonymous */);
    var javascript_1 = tslib_1.__importDefault(require(149) /* ../tree/javascript */);
    exports.default = {
        e: function (str) {
            return new quoted_1.default('"', str instanceof javascript_1.default ? str.evaluated : str.value, true);
        },
        escape: function (str) {
            return new anonymous_1.default(encodeURI(str.value).replace(/=/g, '%3D').replace(/:/g, '%3A').replace(/#/g, '%23').replace(/;/g, '%3B')
                .replace(/\(/g, '%28').replace(/\)/g, '%29'));
        },
        replace: function (string, pattern, replacement, flags) {
            var result = string.value;
            replacement = (replacement.type === 'Quoted') ?
                replacement.value : replacement.toCSS();
            result = result.replace(new RegExp(pattern.value, flags ? flags.value : ''), replacement);
            return new quoted_1.default(string.quote || '', result, string.escaped);
        },
        '%': function (string /* arg, arg, ... */) {
            var args = Array.prototype.slice.call(arguments, 1);
            var result = string.value;
            var _loop_1 = function (i) {
                /* jshint loopfunc:true */
                result = result.replace(/%[sda]/i, function (token) {
                    var value = ((args[i].type === 'Quoted') &&
                        token.match(/s/i)) ? args[i].value : args[i].toCSS();
                    return token.match(/[A-Z]$/) ? encodeURIComponent(value) : value;
                });
            };
            for (var i = 0; i < args.length; i++) {
                _loop_1(i);
            }
            result = result.replace(/%%/g, '%');
            return new quoted_1.default(string.quote || '', result, string.escaped);
        }
    };
},
/* less/lib/less/functions/svg.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var dimension_1 = tslib_1.__importDefault(require(137) /* ../tree/dimension */);
    var color_1 = tslib_1.__importDefault(require(114) /* ../tree/color */);
    var expression_1 = tslib_1.__importDefault(require(142) /* ../tree/expression */);
    var quoted_1 = tslib_1.__importDefault(require(145) /* ../tree/quoted */);
    var url_1 = tslib_1.__importDefault(require(146) /* ../tree/url */);
    exports.default = (function (environment) {
        return { 'svg-gradient': function (direction) {
                var stops;
                var gradientDirectionSvg;
                var gradientType = 'linear';
                var rectangleDimension = 'x="0" y="0" width="1" height="1"';
                var renderEnv = { compress: false };
                var returner;
                var directionValue = direction.toCSS(renderEnv);
                var i;
                var color;
                var position;
                var positionValue;
                var alpha;
                function throwArgumentDescriptor() {
                    throw { type: 'Argument',
                        message: 'svg-gradient expects direction, start_color [start_position], [color position,]...,' +
                            ' end_color [end_position] or direction, color list' };
                }
                if (arguments.length == 2) {
                    if (arguments[1].value.length < 2) {
                        throwArgumentDescriptor();
                    }
                    stops = arguments[1].value;
                }
                else if (arguments.length < 3) {
                    throwArgumentDescriptor();
                }
                else {
                    stops = Array.prototype.slice.call(arguments, 1);
                }
                switch (directionValue) {
                    case 'to bottom':
                        gradientDirectionSvg = 'x1="0%" y1="0%" x2="0%" y2="100%"';
                        break;
                    case 'to right':
                        gradientDirectionSvg = 'x1="0%" y1="0%" x2="100%" y2="0%"';
                        break;
                    case 'to bottom right':
                        gradientDirectionSvg = 'x1="0%" y1="0%" x2="100%" y2="100%"';
                        break;
                    case 'to top right':
                        gradientDirectionSvg = 'x1="0%" y1="100%" x2="100%" y2="0%"';
                        break;
                    case 'ellipse':
                    case 'ellipse at center':
                        gradientType = 'radial';
                        gradientDirectionSvg = 'cx="50%" cy="50%" r="75%"';
                        rectangleDimension = 'x="-50" y="-50" width="101" height="101"';
                        break;
                    default:
                        throw { type: 'Argument', message: 'svg-gradient direction must be \'to bottom\', \'to right\',' +
                                ' \'to bottom right\', \'to top right\' or \'ellipse at center\'' };
                }
                returner = "<svg xmlns=\"http://www.w3.org/2000/svg\" viewBox=\"0 0 1 1\"><" + gradientType + "Gradient id=\"g\" " + gradientDirectionSvg + ">";
                for (i = 0; i < stops.length; i += 1) {
                    if (stops[i] instanceof expression_1.default) {
                        color = stops[i].value[0];
                        position = stops[i].value[1];
                    }
                    else {
                        color = stops[i];
                        position = undefined;
                    }
                    if (!(color instanceof color_1.default) || (!((i === 0 || i + 1 === stops.length) && position === undefined) && !(position instanceof dimension_1.default))) {
                        throwArgumentDescriptor();
                    }
                    positionValue = position ? position.toCSS(renderEnv) : i === 0 ? '0%' : '100%';
                    alpha = color.alpha;
                    returner += "<stop offset=\"" + positionValue + "\" stop-color=\"" + color.toRGB() + "\"" + (alpha < 1 ? " stop-opacity=\"" + alpha + "\"" : '') + "/>";
                }
                returner += "</" + gradientType + "Gradient><rect " + rectangleDimension + " fill=\"url(#g)\" /></svg>";
                returner = encodeURIComponent(returner);
                returner = "data:image/svg+xml," + returner;
                return new url_1.default(new quoted_1.default("'" + returner + "'", returner, false, this.index, this.currentFileInfo), this.index, this.currentFileInfo);
            } };
    });
},
/* less/lib/less/functions/types.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var keyword_1 = tslib_1.__importDefault(require(128) /* ../tree/keyword */);
    var detached_ruleset_1 = tslib_1.__importDefault(require(135) /* ../tree/detached-ruleset */);
    var dimension_1 = tslib_1.__importDefault(require(137) /* ../tree/dimension */);
    var color_1 = tslib_1.__importDefault(require(114) /* ../tree/color */);
    var quoted_1 = tslib_1.__importDefault(require(145) /* ../tree/quoted */);
    var anonymous_1 = tslib_1.__importDefault(require(129) /* ../tree/anonymous */);
    var url_1 = tslib_1.__importDefault(require(146) /* ../tree/url */);
    var operation_1 = tslib_1.__importDefault(require(136) /* ../tree/operation */);
    var isa = function (n, Type) { return (n instanceof Type) ? keyword_1.default.True : keyword_1.default.False; };
    var isunit = function (n, unit) {
        if (unit === undefined) {
            throw { type: 'Argument', message: 'missing the required second argument to isunit.' };
        }
        unit = typeof unit.value === 'string' ? unit.value : unit;
        if (typeof unit !== 'string') {
            throw { type: 'Argument', message: 'Second argument to isunit should be a unit or a string.' };
        }
        return (n instanceof dimension_1.default) && n.unit.is(unit) ? keyword_1.default.True : keyword_1.default.False;
    };
    exports.default = {
        isruleset: function (n) {
            return isa(n, detached_ruleset_1.default);
        },
        iscolor: function (n) {
            return isa(n, color_1.default);
        },
        isnumber: function (n) {
            return isa(n, dimension_1.default);
        },
        isstring: function (n) {
            return isa(n, quoted_1.default);
        },
        iskeyword: function (n) {
            return isa(n, keyword_1.default);
        },
        isurl: function (n) {
            return isa(n, url_1.default);
        },
        ispixel: function (n) {
            return isunit(n, 'px');
        },
        ispercentage: function (n) {
            return isunit(n, '%');
        },
        isem: function (n) {
            return isunit(n, 'em');
        },
        isunit: isunit,
        unit: function (val, unit) {
            if (!(val instanceof dimension_1.default)) {
                throw { type: 'Argument', message: "the first argument to unit must be a number" + (val instanceof operation_1.default ? '. Have you forgotten parenthesis?' : '') };
            }
            if (unit) {
                if (unit instanceof keyword_1.default) {
                    unit = unit.value;
                }
                else {
                    unit = unit.toCSS();
                }
            }
            else {
                unit = '';
            }
            return new dimension_1.default(val.value, unit);
        },
        'get-unit': function (n) {
            return new anonymous_1.default(n.unit);
        }
    };
},
/* less/lib/less/transform-tree.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var contexts_1 = tslib_1.__importDefault(require(132) /* ./contexts */);
    var visitors_1 = tslib_1.__importDefault(require(161) /* ./visitors */);
    var tree_1 = tslib_1.__importDefault(require(112) /* ./tree */);
    function default_1(root, options) {
        options = options || {};
        var evaldRoot;
        var variables = options.variables;
        var evalEnv = new contexts_1.default.Eval(options);
        //
        // Allows setting variables with a hash, so:
        //
        //   `{ color: new tree.Color('#f01') }` will become:
        //
        //   new tree.Declaration('@color',
        //     new tree.Value([
        //       new tree.Expression([
        //         new tree.Color('#f01')
        //       ])
        //     ])
        //   )
        //
        if (typeof variables === 'object' && !Array.isArray(variables)) {
            variables = Object.keys(variables).map(function (k) {
                var value = variables[k];
                if (!(value instanceof tree_1.default.Value)) {
                    if (!(value instanceof tree_1.default.Expression)) {
                        value = new tree_1.default.Expression([value]);
                    }
                    value = new tree_1.default.Value([value]);
                }
                return new tree_1.default.Declaration("@" + k, value, false, null, 0);
            });
            evalEnv.frames = [new tree_1.default.Ruleset(null, variables)];
        }
        var visitors = [
            new visitors_1.default.JoinSelectorVisitor(),
            new visitors_1.default.MarkVisibleSelectorsVisitor(true),
            new visitors_1.default.ExtendVisitor(),
            new visitors_1.default.ToCSSVisitor({ compress: Boolean(options.compress) })
        ];
        var preEvalVisitors = [];
        var v;
        var visitorIterator;
        /**
         * first() / get() allows visitors to be added while visiting
         *
         * @todo Add scoping for visitors just like functions for @plugin; right now they're global
         */
        if (options.pluginManager) {
            visitorIterator = options.pluginManager.visitor();
            for (var i = 0; i < 2; i++) {
                visitorIterator.first();
                while ((v = visitorIterator.get())) {
                    if (v.isPreEvalVisitor) {
                        if (i === 0 || preEvalVisitors.indexOf(v) === -1) {
                            preEvalVisitors.push(v);
                            v.run(root);
                        }
                    }
                    else {
                        if (i === 0 || visitors.indexOf(v) === -1) {
                            if (v.isPreVisitor) {
                                visitors.unshift(v);
                            }
                            else {
                                visitors.push(v);
                            }
                        }
                    }
                }
            }
        }
        evaldRoot = root.eval(evalEnv);
        for (var i = 0; i < visitors.length; i++) {
            visitors[i].run(evaldRoot);
        }
        // Run any remaining visitors added after eval pass
        if (options.pluginManager) {
            visitorIterator.first();
            while ((v = visitorIterator.get())) {
                if (visitors.indexOf(v) === -1 && preEvalVisitors.indexOf(v) === -1) {
                    v.run(evaldRoot);
                }
            }
        }
        return evaldRoot;
    }
    exports.default = default_1;
    ;
},
/* less/lib/less/plugin-manager.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    /**
     * Plugin Manager
     */
    var PluginManager = /** @class */ (function () {
        function PluginManager(less) {
            this.less = less;
            this.visitors = [];
            this.preProcessors = [];
            this.postProcessors = [];
            this.installedPlugins = [];
            this.fileManagers = [];
            this.iterator = -1;
            this.pluginCache = {};
            this.Loader = new less.PluginLoader(less);
        }
        /**
         * Adds all the plugins in the array
         * @param {Array} plugins
         */
        PluginManager.prototype.addPlugins = function (plugins) {
            if (plugins) {
                for (var i = 0; i < plugins.length; i++) {
                    this.addPlugin(plugins[i]);
                }
            }
        };
        /**
         *
         * @param plugin
         * @param {String} filename
         */
        PluginManager.prototype.addPlugin = function (plugin, filename, functionRegistry) {
            this.installedPlugins.push(plugin);
            if (filename) {
                this.pluginCache[filename] = plugin;
            }
            if (plugin.install) {
                plugin.install(this.less, this, functionRegistry || this.less.functions.functionRegistry);
            }
        };
        /**
         *
         * @param filename
         */
        PluginManager.prototype.get = function (filename) {
            return this.pluginCache[filename];
        };
        /**
         * Adds a visitor. The visitor object has options on itself to determine
         * when it should run.
         * @param visitor
         */
        PluginManager.prototype.addVisitor = function (visitor) {
            this.visitors.push(visitor);
        };
        /**
         * Adds a pre processor object
         * @param {object} preProcessor
         * @param {number} priority - guidelines 1 = before import, 1000 = import, 2000 = after import
         */
        PluginManager.prototype.addPreProcessor = function (preProcessor, priority) {
            var indexToInsertAt;
            for (indexToInsertAt = 0; indexToInsertAt < this.preProcessors.length; indexToInsertAt++) {
                if (this.preProcessors[indexToInsertAt].priority >= priority) {
                    break;
                }
            }
            this.preProcessors.splice(indexToInsertAt, 0, { preProcessor: preProcessor, priority: priority });
        };
        /**
         * Adds a post processor object
         * @param {object} postProcessor
         * @param {number} priority - guidelines 1 = before compression, 1000 = compression, 2000 = after compression
         */
        PluginManager.prototype.addPostProcessor = function (postProcessor, priority) {
            var indexToInsertAt;
            for (indexToInsertAt = 0; indexToInsertAt < this.postProcessors.length; indexToInsertAt++) {
                if (this.postProcessors[indexToInsertAt].priority >= priority) {
                    break;
                }
            }
            this.postProcessors.splice(indexToInsertAt, 0, { postProcessor: postProcessor, priority: priority });
        };
        /**
         *
         * @param manager
         */
        PluginManager.prototype.addFileManager = function (manager) {
            this.fileManagers.push(manager);
        };
        /**
         *
         * @returns {Array}
         * @private
         */
        PluginManager.prototype.getPreProcessors = function () {
            var preProcessors = [];
            for (var i = 0; i < this.preProcessors.length; i++) {
                preProcessors.push(this.preProcessors[i].preProcessor);
            }
            return preProcessors;
        };
        /**
         *
         * @returns {Array}
         * @private
         */
        PluginManager.prototype.getPostProcessors = function () {
            var postProcessors = [];
            for (var i = 0; i < this.postProcessors.length; i++) {
                postProcessors.push(this.postProcessors[i].postProcessor);
            }
            return postProcessors;
        };
        /**
         *
         * @returns {Array}
         * @private
         */
        PluginManager.prototype.getVisitors = function () {
            return this.visitors;
        };
        PluginManager.prototype.visitor = function () {
            var self = this;
            return {
                first: function () {
                    self.iterator = -1;
                    return self.visitors[self.iterator];
                },
                get: function () {
                    self.iterator += 1;
                    return self.visitors[self.iterator];
                }
            };
        };
        /**
         *
         * @returns {Array}
         * @private
         */
        PluginManager.prototype.getFileManagers = function () {
            return this.fileManagers;
        };
        return PluginManager;
    }());
    var pm;
    var PluginManagerFactory = function (less, newFactory) {
        if (newFactory || !pm) {
            pm = new PluginManager(less);
        }
        return pm;
    };
    //
    exports.default = PluginManagerFactory;
},
/* less/lib/less/source-map-output.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function default_1(environment) {
        var SourceMapOutput = /** @class */ (function () {
            function SourceMapOutput(options) {
                this._css = [];
                this._rootNode = options.rootNode;
                this._contentsMap = options.contentsMap;
                this._contentsIgnoredCharsMap = options.contentsIgnoredCharsMap;
                if (options.sourceMapFilename) {
                    this._sourceMapFilename = options.sourceMapFilename.replace(/\\/g, '/');
                }
                this._outputFilename = options.outputFilename;
                this.sourceMapURL = options.sourceMapURL;
                if (options.sourceMapBasepath) {
                    this._sourceMapBasepath = options.sourceMapBasepath.replace(/\\/g, '/');
                }
                if (options.sourceMapRootpath) {
                    this._sourceMapRootpath = options.sourceMapRootpath.replace(/\\/g, '/');
                    if (this._sourceMapRootpath.charAt(this._sourceMapRootpath.length - 1) !== '/') {
                        this._sourceMapRootpath += '/';
                    }
                }
                else {
                    this._sourceMapRootpath = '';
                }
                this._outputSourceFiles = options.outputSourceFiles;
                this._sourceMapGeneratorConstructor = environment.getSourceMapGenerator();
                this._lineNumber = 0;
                this._column = 0;
            }
            SourceMapOutput.prototype.removeBasepath = function (path) {
                if (this._sourceMapBasepath && path.indexOf(this._sourceMapBasepath) === 0) {
                    path = path.substring(this._sourceMapBasepath.length);
                    if (path.charAt(0) === '\\' || path.charAt(0) === '/') {
                        path = path.substring(1);
                    }
                }
                return path;
            };
            SourceMapOutput.prototype.normalizeFilename = function (filename) {
                filename = filename.replace(/\\/g, '/');
                filename = this.removeBasepath(filename);
                return (this._sourceMapRootpath || '') + filename;
            };
            SourceMapOutput.prototype.add = function (chunk, fileInfo, index, mapLines) {
                // ignore adding empty strings
                if (!chunk) {
                    return;
                }
                var lines, sourceLines, columns, sourceColumns, i;
                if (fileInfo && fileInfo.filename) {
                    var inputSource = this._contentsMap[fileInfo.filename];
                    // remove vars/banner added to the top of the file
                    if (this._contentsIgnoredCharsMap[fileInfo.filename]) {
                        // adjust the index
                        index -= this._contentsIgnoredCharsMap[fileInfo.filename];
                        if (index < 0) {
                            index = 0;
                        }
                        // adjust the source
                        inputSource = inputSource.slice(this._contentsIgnoredCharsMap[fileInfo.filename]);
                    }
                    /**
                     * ignore empty content, or failsafe
                     * if contents map is incorrect
                     */
                    if (inputSource === undefined) {
                        this._css.push(chunk);
                        return;
                    }
                    inputSource = inputSource.substring(0, index);
                    sourceLines = inputSource.split('\n');
                    sourceColumns = sourceLines[sourceLines.length - 1];
                }
                lines = chunk.split('\n');
                columns = lines[lines.length - 1];
                if (fileInfo && fileInfo.filename) {
                    if (!mapLines) {
                        this._sourceMapGenerator.addMapping({ generated: { line: this._lineNumber + 1, column: this._column },
                            original: { line: sourceLines.length, column: sourceColumns.length },
                            source: this.normalizeFilename(fileInfo.filename) });
                    }
                    else {
                        for (i = 0; i < lines.length; i++) {
                            this._sourceMapGenerator.addMapping({ generated: { line: this._lineNumber + i + 1, column: i === 0 ? this._column : 0 },
                                original: { line: sourceLines.length + i, column: i === 0 ? sourceColumns.length : 0 },
                                source: this.normalizeFilename(fileInfo.filename) });
                        }
                    }
                }
                if (lines.length === 1) {
                    this._column += columns.length;
                }
                else {
                    this._lineNumber += lines.length - 1;
                    this._column = columns.length;
                }
                this._css.push(chunk);
            };
            SourceMapOutput.prototype.isEmpty = function () {
                return this._css.length === 0;
            };
            SourceMapOutput.prototype.toCSS = function (context) {
                this._sourceMapGenerator = new this._sourceMapGeneratorConstructor({ file: this._outputFilename, sourceRoot: null });
                if (this._outputSourceFiles) {
                    for (var filename in this._contentsMap) {
                        if (this._contentsMap.hasOwnProperty(filename)) {
                            var source = this._contentsMap[filename];
                            if (this._contentsIgnoredCharsMap[filename]) {
                                source = source.slice(this._contentsIgnoredCharsMap[filename]);
                            }
                            this._sourceMapGenerator.setSourceContent(this.normalizeFilename(filename), source);
                        }
                    }
                }
                this._rootNode.genCSS(context, this);
                if (this._css.length > 0) {
                    var sourceMapURL = void 0;
                    var sourceMapContent = JSON.stringify(this._sourceMapGenerator.toJSON());
                    if (this.sourceMapURL) {
                        sourceMapURL = this.sourceMapURL;
                    }
                    else if (this._sourceMapFilename) {
                        sourceMapURL = this._sourceMapFilename;
                    }
                    this.sourceMapURL = sourceMapURL;
                    this.sourceMap = sourceMapContent;
                }
                return this._css.join('');
            };
            return SourceMapOutput;
        }());
        return SourceMapOutput;
    }
    exports.default = default_1;
    ;
},
/* less/lib/less/source-map-builder.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function default_1(SourceMapOutput, environment) {
        var SourceMapBuilder = /** @class */ (function () {
            function SourceMapBuilder(options) {
                this.options = options;
            }
            SourceMapBuilder.prototype.toCSS = function (rootNode, options, imports) {
                var sourceMapOutput = new SourceMapOutput({
                    contentsIgnoredCharsMap: imports.contentsIgnoredChars,
                    rootNode: rootNode,
                    contentsMap: imports.contents,
                    sourceMapFilename: this.options.sourceMapFilename,
                    sourceMapURL: this.options.sourceMapURL,
                    outputFilename: this.options.sourceMapOutputFilename,
                    sourceMapBasepath: this.options.sourceMapBasepath,
                    sourceMapRootpath: this.options.sourceMapRootpath,
                    outputSourceFiles: this.options.outputSourceFiles,
                    sourceMapGenerator: this.options.sourceMapGenerator,
                    sourceMapFileInline: this.options.sourceMapFileInline,
                    disableSourcemapAnnotation: this.options.disableSourcemapAnnotation
                });
                var css = sourceMapOutput.toCSS(options);
                this.sourceMap = sourceMapOutput.sourceMap;
                this.sourceMapURL = sourceMapOutput.sourceMapURL;
                if (this.options.sourceMapInputFilename) {
                    this.sourceMapInputFilename = sourceMapOutput.normalizeFilename(this.options.sourceMapInputFilename);
                }
                if (this.options.sourceMapBasepath !== undefined && this.sourceMapURL !== undefined) {
                    this.sourceMapURL = sourceMapOutput.removeBasepath(this.sourceMapURL);
                }
                return css + this.getCSSAppendage();
            };
            SourceMapBuilder.prototype.getCSSAppendage = function () {
                var sourceMapURL = this.sourceMapURL;
                if (this.options.sourceMapFileInline) {
                    if (this.sourceMap === undefined) {
                        return '';
                    }
                    sourceMapURL = "data:application/json;base64," + environment.encodeBase64(this.sourceMap);
                }
                if (this.options.disableSourcemapAnnotation) {
                    return '';
                }
                if (sourceMapURL) {
                    return "/*# sourceMappingURL=" + sourceMapURL + " */";
                }
                return '';
            };
            SourceMapBuilder.prototype.getExternalSourceMap = function () {
                return this.sourceMap;
            };
            SourceMapBuilder.prototype.setExternalSourceMap = function (sourceMap) {
                this.sourceMap = sourceMap;
            };
            SourceMapBuilder.prototype.isInline = function () {
                return this.options.sourceMapFileInline;
            };
            SourceMapBuilder.prototype.getSourceMapURL = function () {
                return this.sourceMapURL;
            };
            SourceMapBuilder.prototype.getOutputFilename = function () {
                return this.options.sourceMapOutputFilename;
            };
            SourceMapBuilder.prototype.getInputFilename = function () {
                return this.sourceMapInputFilename;
            };
            return SourceMapBuilder;
        }());
        return SourceMapBuilder;
    }
    exports.default = default_1;
    ;
},
/* less/lib/less/parse-tree.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var less_error_1 = tslib_1.__importDefault(require(120) /* ./less-error */);
    var transform_tree_1 = tslib_1.__importDefault(require(184) /* ./transform-tree */);
    var logger_1 = tslib_1.__importDefault(require(70) /* ./logger */);
    function default_1(SourceMapBuilder) {
        var ParseTree = /** @class */ (function () {
            function ParseTree(root, imports) {
                this.root = root;
                this.imports = imports;
            }
            ParseTree.prototype.toCSS = function (options) {
                var evaldRoot;
                var result = {};
                var sourceMapBuilder;
                try {
                    evaldRoot = transform_tree_1.default(this.root, options);
                }
                catch (e) {
                    throw new less_error_1.default(e, this.imports);
                }
                try {
                    var compress = Boolean(options.compress);
                    if (compress) {
                        logger_1.default.warn('The compress option has been deprecated. ' +
                            'We recommend you use a dedicated css minifier, for instance see less-plugin-clean-css.');
                    }
                    var toCSSOptions = {
                        compress: compress,
                        dumpLineNumbers: options.dumpLineNumbers,
                        strictUnits: Boolean(options.strictUnits),
                        numPrecision: 8
                    };
                    if (options.sourceMap) {
                        sourceMapBuilder = new SourceMapBuilder(options.sourceMap);
                        result.css = sourceMapBuilder.toCSS(evaldRoot, toCSSOptions, this.imports);
                    }
                    else {
                        result.css = evaldRoot.toCSS(toCSSOptions);
                    }
                }
                catch (e) {
                    throw new less_error_1.default(e, this.imports);
                }
                if (options.pluginManager) {
                    var postProcessors = options.pluginManager.getPostProcessors();
                    for (var i = 0; i < postProcessors.length; i++) {
                        result.css = postProcessors[i].process(result.css, { sourceMap: sourceMapBuilder, options: options, imports: this.imports });
                    }
                }
                if (options.sourceMap) {
                    result.map = sourceMapBuilder.getExternalSourceMap();
                }
                result.imports = [];
                for (var file in this.imports.files) {
                    if (this.imports.files.hasOwnProperty(file) && file !== this.imports.rootFilename) {
                        result.imports.push(file);
                    }
                }
                return result;
            };
            return ParseTree;
        }());
        return ParseTree;
    }
    exports.default = default_1;
    ;
},
/* less/lib/less/import-manager.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var contexts_1 = tslib_1.__importDefault(require(132) /* ./contexts */);
    var parser_1 = tslib_1.__importDefault(require(169) /* ./parser/parser */);
    var less_error_1 = tslib_1.__importDefault(require(120) /* ./less-error */);
    var utils = tslib_1.__importStar(require(121) /* ./utils */);
    var logger_1 = tslib_1.__importDefault(require(70) /* ./logger */);
    function default_1(environment) {
        // FileInfo = {
        //  'rewriteUrls' - option - whether to adjust URL's to be relative
        //  'filename' - full resolved filename of current file
        //  'rootpath' - path to append to normal URLs for this node
        //  'currentDirectory' - path to the current file, absolute
        //  'rootFilename' - filename of the base file
        //  'entryPath' - absolute path to the entry file
        //  'reference' - whether the file should not be output and only output parts that are referenced
        var ImportManager = /** @class */ (function () {
            function ImportManager(less, context, rootFileInfo) {
                this.less = less;
                this.rootFilename = rootFileInfo.filename;
                this.paths = context.paths || []; // Search paths, when importing
                this.contents = {}; // map - filename to contents of all the files
                this.contentsIgnoredChars = {}; // map - filename to lines at the beginning of each file to ignore
                this.mime = context.mime;
                this.error = null;
                this.context = context;
                // Deprecated? Unused outside of here, could be useful.
                this.queue = []; // Files which haven't been imported yet
                this.files = {}; // Holds the imported parse trees.
            }
            /**
             * Add an import to be imported
             * @param path - the raw path
             * @param tryAppendExtension - whether to try appending a file extension (.less or .js if the path has no extension)
             * @param currentFileInfo - the current file info (used for instance to work out relative paths)
             * @param importOptions - import options
             * @param callback - callback for when it is imported
             */
            ImportManager.prototype.push = function (path, tryAppendExtension, currentFileInfo, importOptions, callback) {
                var importManager = this, pluginLoader = this.context.pluginManager.Loader;
                this.queue.push(path);
                var fileParsedFunc = function (e, root, fullPath) {
                    importManager.queue.splice(importManager.queue.indexOf(path), 1); // Remove the path from the queue
                    var importedEqualsRoot = fullPath === importManager.rootFilename;
                    if (importOptions.optional && e) {
                        callback(null, { rules: [] }, false, null);
                        logger_1.default.info("The file " + fullPath + " was skipped because it was not found and the import was marked optional.");
                    }
                    else {
                        // Inline imports aren't cached here.
                        // If we start to cache them, please make sure they won't conflict with non-inline imports of the
                        // same name as they used to do before this comment and the condition below have been added.
                        if (!importManager.files[fullPath] && !importOptions.inline) {
                            importManager.files[fullPath] = { root: root, options: importOptions };
                        }
                        if (e && !importManager.error) {
                            importManager.error = e;
                        }
                        callback(e, root, importedEqualsRoot, fullPath);
                    }
                };
                var newFileInfo = {
                    rewriteUrls: this.context.rewriteUrls,
                    entryPath: currentFileInfo.entryPath,
                    rootpath: currentFileInfo.rootpath,
                    rootFilename: currentFileInfo.rootFilename
                };
                var fileManager = environment.getFileManager(path, currentFileInfo.currentDirectory, this.context, environment);
                if (!fileManager) {
                    fileParsedFunc({ message: "Could not find a file-manager for " + path });
                    return;
                }
                var loadFileCallback = function (loadedFile) {
                    var plugin;
                    var resolvedFilename = loadedFile.filename;
                    var contents = loadedFile.contents.replace(/^\uFEFF/, '');
                    // Pass on an updated rootpath if path of imported file is relative and file
                    // is in a (sub|sup) directory
                    //
                    // Examples:
                    // - If path of imported file is 'module/nav/nav.less' and rootpath is 'less/',
                    //   then rootpath should become 'less/module/nav/'
                    // - If path of imported file is '../mixins.less' and rootpath is 'less/',
                    //   then rootpath should become 'less/../'
                    newFileInfo.currentDirectory = fileManager.getPath(resolvedFilename);
                    if (newFileInfo.rewriteUrls) {
                        newFileInfo.rootpath = fileManager.join((importManager.context.rootpath || ''), fileManager.pathDiff(newFileInfo.currentDirectory, newFileInfo.entryPath));
                        if (!fileManager.isPathAbsolute(newFileInfo.rootpath) && fileManager.alwaysMakePathsAbsolute()) {
                            newFileInfo.rootpath = fileManager.join(newFileInfo.entryPath, newFileInfo.rootpath);
                        }
                    }
                    newFileInfo.filename = resolvedFilename;
                    var newEnv = new contexts_1.default.Parse(importManager.context);
                    newEnv.processImports = false;
                    importManager.contents[resolvedFilename] = contents;
                    if (currentFileInfo.reference || importOptions.reference) {
                        newFileInfo.reference = true;
                    }
                    if (importOptions.isPlugin) {
                        plugin = pluginLoader.evalPlugin(contents, newEnv, importManager, importOptions.pluginArgs, newFileInfo);
                        if (plugin instanceof less_error_1.default) {
                            fileParsedFunc(plugin, null, resolvedFilename);
                        }
                        else {
                            fileParsedFunc(null, plugin, resolvedFilename);
                        }
                    }
                    else if (importOptions.inline) {
                        fileParsedFunc(null, contents, resolvedFilename);
                    }
                    else {
                        // import (multiple) parse trees apparently get altered and can't be cached.
                        // TODO: investigate why this is
                        if (importManager.files[resolvedFilename]
                            && !importManager.files[resolvedFilename].options.multiple
                            && !importOptions.multiple) {
                            fileParsedFunc(null, importManager.files[resolvedFilename].root, resolvedFilename);
                        }
                        else {
                            new parser_1.default(newEnv, importManager, newFileInfo).parse(contents, function (e, root) {
                                fileParsedFunc(e, root, resolvedFilename);
                            });
                        }
                    }
                };
                var loadedFile;
                var promise;
                var context = utils.clone(this.context);
                if (tryAppendExtension) {
                    context.ext = importOptions.isPlugin ? '.js' : '.less';
                }
                if (importOptions.isPlugin) {
                    context.mime = 'application/javascript';
                    if (context.syncImport) {
                        loadedFile = pluginLoader.loadPluginSync(path, currentFileInfo.currentDirectory, context, environment, fileManager);
                    }
                    else {
                        promise = pluginLoader.loadPlugin(path, currentFileInfo.currentDirectory, context, environment, fileManager);
                    }
                }
                else {
                    if (context.syncImport) {
                        loadedFile = fileManager.loadFileSync(path, currentFileInfo.currentDirectory, context, environment);
                    }
                    else {
                        promise = fileManager.loadFile(path, currentFileInfo.currentDirectory, context, environment, function (err, loadedFile) {
                            if (err) {
                                fileParsedFunc(err);
                            }
                            else {
                                loadFileCallback(loadedFile);
                            }
                        });
                    }
                }
                if (loadedFile) {
                    if (!loadedFile.filename) {
                        fileParsedFunc(loadedFile);
                    }
                    else {
                        loadFileCallback(loadedFile);
                    }
                }
                else if (promise) {
                    promise.then(loadFileCallback, fileParsedFunc);
                }
            };
            return ImportManager;
        }());
        return ImportManager;
    }
    exports.default = default_1;
    ;
},
/* less/lib/less/parse.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var contexts_1 = tslib_1.__importDefault(require(132) /* ./contexts */);
    var parser_1 = tslib_1.__importDefault(require(169) /* ./parser/parser */);
    var plugin_manager_1 = tslib_1.__importDefault(require(185) /* ./plugin-manager */);
    var less_error_1 = tslib_1.__importDefault(require(120) /* ./less-error */);
    var utils = tslib_1.__importStar(require(121) /* ./utils */);
    function default_1(environment, ParseTree, ImportManager) {
        var parse = function (input, options, callback) {
            if (typeof options === 'function') {
                callback = options;
                options = utils.copyOptions(this.options, {});
            }
            else {
                options = utils.copyOptions(this.options, options || {});
            }
            if (!callback) {
                var self_1 = this;
                return new Promise(function (resolve, reject) {
                    parse.call(self_1, input, options, function (err, output) {
                        if (err) {
                            reject(err);
                        }
                        else {
                            resolve(output);
                        }
                    });
                });
            }
            else {
                var context_1;
                var rootFileInfo = void 0;
                var pluginManager_1 = new plugin_manager_1.default(this, !options.reUsePluginManager);
                options.pluginManager = pluginManager_1;
                context_1 = new contexts_1.default.Parse(options);
                if (options.rootFileInfo) {
                    rootFileInfo = options.rootFileInfo;
                }
                else {
                    var filename = options.filename || 'input';
                    var entryPath = filename.replace(/[^\/\\]*$/, '');
                    rootFileInfo = {
                        filename: filename,
                        rewriteUrls: context_1.rewriteUrls,
                        rootpath: context_1.rootpath || '',
                        currentDirectory: entryPath,
                        entryPath: entryPath,
                        rootFilename: filename
                    };
                    // add in a missing trailing slash
                    if (rootFileInfo.rootpath && rootFileInfo.rootpath.slice(-1) !== '/') {
                        rootFileInfo.rootpath += '/';
                    }
                }
                var imports_1 = new ImportManager(this, context_1, rootFileInfo);
                this.importManager = imports_1;
                // TODO: allow the plugins to be just a list of paths or names
                // Do an async plugin queue like lessc
                if (options.plugins) {
                    options.plugins.forEach(function (plugin) {
                        var evalResult, contents;
                        if (plugin.fileContent) {
                            contents = plugin.fileContent.replace(/^\uFEFF/, '');
                            evalResult = pluginManager_1.Loader.evalPlugin(contents, context_1, imports_1, plugin.options, plugin.filename);
                            if (evalResult instanceof less_error_1.default) {
                                return callback(evalResult);
                            }
                        }
                        else {
                            pluginManager_1.addPlugin(plugin);
                        }
                    });
                }
                new parser_1.default(context_1, imports_1, rootFileInfo)
                    .parse(input, function (e, root) {
                    if (e) {
                        return callback(e);
                    }
                    callback(null, root, imports_1, options);
                }, options);
            }
        };
        return parse;
    }
    exports.default = default_1;
    ;
},
/* less/lib/less/render.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var utils = tslib_1.__importStar(require(121) /* ./utils */);
    function default_1(environment, ParseTree, ImportManager) {
        var render = function (input, options, callback) {
            if (typeof options === 'function') {
                callback = options;
                options = utils.copyOptions(this.options, {});
            }
            else {
                options = utils.copyOptions(this.options, options || {});
            }
            if (!callback) {
                var self_1 = this;
                return new Promise(function (resolve, reject) {
                    render.call(self_1, input, options, function (err, output) {
                        if (err) {
                            reject(err);
                        }
                        else {
                            resolve(output);
                        }
                    });
                });
            }
            else {
                this.parse(input, options, function (err, root, imports, options) {
                    if (err) {
                        return callback(err);
                    }
                    var result;
                    try {
                        var parseTree = new ParseTree(root, imports);
                        result = parseTree.toCSS(options);
                    }
                    catch (err) {
                        return callback(err);
                    }
                    callback(null, result);
                });
            }
        };
        return render;
    }
    exports.default = default_1;
    ;
},
/* less/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "less",
        "version": "4.1.1",
        "description": "Leaner CSS",
        "homepage": "http://lesscss.org",
        "author": {
            "name": "Alexis Sellier",
            "email": "self@cloudhead.net"
        },
        "contributors": [
            "The Core Less Team"
        ],
        "bugs": {
            "url": "https://github.com/less/less.js/issues"
        },
        "repository": {
            "type": "git",
            "url": "https://github.com/less/less.js.git"
        },
        "master": {
            "url": "https://github.com/less/less.js/blob/master/",
            "raw": "https://raw.githubusercontent.com/less/less.js/master/"
        },
        "license": "Apache-2.0",
        "bin": {
            "lessc": "./bin/lessc"
        },
        "main": "index",
        "module": "./lib/less-node/index",
        "directories": {
            "test": "./test"
        },
        "browser": "./dist/less.js",
        "engines": {
            "node": ">=6"
        },
        "scripts": {
            "test": "grunt test",
            "grunt": "grunt",
            "build": "npm-run-all clean compile",
            "clean": "shx rm -rf ./lib tsconfig.tsbuildinfo",
            "compile": "tsc -p tsconfig.json",
            "copy:root": "shx cp -rf ./dist ../../",
            "dev": "tsc -p tsconfig.json -w",
            "prepublishOnly": "grunt dist"
        },
        "optionalDependencies": {
            "errno": "^0.1.1",
            "graceful-fs": "^4.1.2",
            "image-size": "~0.5.0",
            "make-dir": "^2.1.0",
            "mime": "^1.4.1",
            "needle": "^2.5.2",
            "source-map": "~0.6.0"
        },
        "devDependencies": {
            "@less/test-data": "^4.1.0",
            "@less/test-import-module": "^4.0.0",
            "@rollup/plugin-commonjs": "^17.0.0",
            "@rollup/plugin-json": "^4.1.0",
            "@rollup/plugin-node-resolve": "^11.0.0",
            "@typescript-eslint/eslint-plugin": "^3.3.0",
            "@typescript-eslint/parser": "^3.3.0",
            "benny": "^3.6.12",
            "bootstrap-less-port": "0.3.0",
            "chai": "^4.2.0",
            "diff": "^3.2.0",
            "eslint": "^6.8.0",
            "fs-extra": "^8.1.0",
            "git-rev": "^0.2.1",
            "globby": "^10.0.1",
            "grunt": "^1.0.4",
            "grunt-cli": "^1.3.2",
            "grunt-contrib-clean": "^1.0.0",
            "grunt-contrib-connect": "^1.0.2",
            "grunt-eslint": "^21.1.0",
            "grunt-saucelabs": "^9.0.1",
            "grunt-shell": "^1.3.0",
            "html-template-tag": "^3.2.0",
            "jit-grunt": "^0.10.0",
            "less-plugin-autoprefix": "^1.5.1",
            "less-plugin-clean-css": "^1.5.1",
            "minimist": "^1.2.0",
            "mocha": "^6.2.1",
            "mocha-headless-chrome": "^2.0.3",
            "mocha-teamcity-reporter": "^3.0.0",
            "nock": "^11.8.2",
            "npm-run-all": "^4.1.5",
            "performance-now": "^0.2.0",
            "phin": "^2.2.3",
            "promise": "^7.1.1",
            "read-glob": "^3.0.0",
            "resolve": "^1.17.0",
            "rollup": "^2.34.1",
            "rollup-plugin-terser": "^5.1.1",
            "rollup-plugin-typescript2": "^0.29.0",
            "semver": "^6.3.0",
            "shx": "^0.3.2",
            "time-grunt": "^1.3.0",
            "ts-node": "^8.4.1",
            "typescript": "^4.1.3",
            "uikit": "2.27.4"
        },
        "keywords": [
            "compile less",
            "css nesting",
            "css variable",
            "css",
            "gradients css",
            "gradients css3",
            "less compiler",
            "less css",
            "less mixins",
            "less",
            "less.js",
            "lesscss",
            "mixins",
            "nested css",
            "parser",
            "preprocessor",
            "bootstrap css",
            "bootstrap less",
            "style",
            "styles",
            "stylesheet",
            "variables in css",
            "css less"
        ],
        "rawcurrent": "https://raw.github.com/less/less.js/v",
        "sourcearchive": "https://github.com/less/less.js/archive/v",
        "dependencies": {
            "copy-anything": "^2.0.1",
            "parse-node-version": "^1.0.1",
            "tslib": "^1.10.0"
        }
    };
    module.exports = json;
},
/* parse-node-version/index.js */ function _(require, module, exports, __esModule, __esExport) {
    function parseNodeVersion(version) {
        var match = version.match(/^v(\d{1,2})\.(\d{1,2})\.(\d{1,2})(?:-([0-9A-Za-z-.]+))?(?:\+([0-9A-Za-z-.]+))?$/); // eslint-disable-line max-len
        if (!match) {
            throw new Error('Unable to parse: ' + version);
        }
        var res = {
            major: parseInt(match[1], 10),
            minor: parseInt(match[2], 10),
            patch: parseInt(match[3], 10),
            pre: match[4] || '',
            build: match[5] || '',
        };
        return res;
    }
    module.exports = parseNodeVersion;
},
/* less/lib/less-node/lessc-helper.js */ function _(require, module, exports, __esModule, __esExport) {
    // lessc_helper.js
    //
    //      helper functions for lessc
    var lessc_helper = {
        // Stylize a string
        stylize: function (str, style) {
            var styles = {
                'reset': [0, 0],
                'bold': [1, 22],
                'inverse': [7, 27],
                'underline': [4, 24],
                'yellow': [33, 39],
                'green': [32, 39],
                'red': [31, 39],
                'grey': [90, 39]
            };
            return "\u001B[" + styles[style][0] + "m" + str + "\u001B[" + styles[style][1] + "m";
        },
        // Print command line options
        printUsage: function () {
            console.log('usage: lessc [option option=parameter ...] <source> [destination]');
            console.log('');
            console.log('If source is set to `-\' (dash or hyphen-minus), input is read from stdin.');
            console.log('');
            console.log('options:');
            console.log('  -h, --help                   Prints help (this message) and exit.');
            console.log('  --include-path=PATHS         Sets include paths. Separated by `:\'. `;\' also supported on windows.');
            console.log('  -M, --depends                Outputs a makefile import dependency list to stdout.');
            console.log('  --no-color                   Disables colorized output.');
            console.log('  --ie-compat                  Enables IE8 compatibility checks.');
            console.log('  --js                         Enables inline JavaScript in less files');
            console.log('  -l, --lint                   Syntax check only (lint).');
            console.log('  -s, --silent                 Suppresses output of error messages.');
            console.log('  --strict-imports             Forces evaluation of imports.');
            console.log('  --insecure                   Allows imports from insecure https hosts.');
            console.log('  -v, --version                Prints version number and exit.');
            console.log('  --verbose                    Be verbose.');
            console.log('  --source-map[=FILENAME]      Outputs a v3 sourcemap to the filename (or output filename.map).');
            console.log('  --source-map-rootpath=X      Adds this path onto the sourcemap filename and less file paths.');
            console.log('  --source-map-basepath=X      Sets sourcemap base path, defaults to current working directory.');
            console.log('  --source-map-include-source  Puts the less files into the map instead of referencing them.');
            console.log('  --source-map-inline          Puts the map (and any less files) as a base64 data uri into the output css file.');
            console.log('  --source-map-url=URL         Sets a custom URL to map file, for sourceMappingURL comment');
            console.log('                               in generated CSS file.');
            console.log('  --source-map-no-annotation   Excludes the sourceMappingURL comment from the output css file.');
            console.log('  -rp, --rootpath=URL          Sets rootpath for url rewriting in relative imports and urls');
            console.log('                               Works with or without the relative-urls option.');
            console.log('  -ru=, --rewrite-urls=        Rewrites URLs to make them relative to the base less file.');
            console.log('    all|local|off              \'all\' rewrites all URLs, \'local\' just those starting with a \'.\'');
            console.log('');
            console.log('  -m=, --math=');
            console.log('     always                    Less will eagerly perform math operations always.');
            console.log('     parens-division           Math performed except for division (/) operator');
            console.log('     parens | strict           Math only performed inside parentheses');
            console.log('     strict-legacy             Parens required in very strict terms (legacy --strict-math)');
            console.log('');
            console.log('  -su=on|off                   Allows mixed units, e.g. 1px+1em or 1px*1px which have units');
            console.log('  --strict-units=on|off        that cannot be represented.');
            console.log('  --global-var=\'VAR=VALUE\'     Defines a variable that can be referenced by the file.');
            console.log('  --modify-var=\'VAR=VALUE\'     Modifies a variable already declared in the file.');
            console.log('  --url-args=\'QUERYSTRING\'     Adds params into url tokens (e.g. 42, cb=42 or \'a=1&b=2\')');
            console.log('  --plugin=PLUGIN=OPTIONS      Loads a plugin. You can also omit the --plugin= if the plugin begins');
            console.log('                               less-plugin. E.g. the clean css plugin is called less-plugin-clean-css');
            console.log('                               once installed (npm install less-plugin-clean-css), use either with');
            console.log('                               --plugin=less-plugin-clean-css or just --clean-css');
            console.log('                               specify options afterwards e.g. --plugin=less-plugin-clean-css="advanced"');
            console.log('                               or --clean-css="advanced"');
            console.log('');
            console.log('-------------------------- Deprecated ----------------');
            console.log('  -sm=on|off               Legacy parens-only math. Use --math');
            console.log('  --strict-math=on|off     ');
            console.log('');
            console.log('  --line-numbers=TYPE      Outputs filename and line numbers.');
            console.log('                           TYPE can be either \'comments\', which will output');
            console.log('                           the debug info within comments, \'mediaquery\'');
            console.log('                           that will output the information within a fake');
            console.log('                           media query which is compatible with the SASS');
            console.log('                           format, and \'all\' which will do both.');
            console.log('  -x, --compress           Compresses output by removing some whitespaces.');
            console.log('                           We recommend you use a dedicated minifer like less-plugin-clean-css');
            console.log('');
            console.log('Report bugs to: http://github.com/less/less.js/issues');
            console.log('Home page: <http://lesscss.org/>');
        }
    };
    // Exports helper functions
    for (var h in lessc_helper) {
        if (lessc_helper.hasOwnProperty(h)) {
            exports[h] = lessc_helper[h];
        }
    }
},
/* less/lib/less-node/plugin-loader.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var path_1 = tslib_1.__importDefault(require("path"));
    var abstract_plugin_loader_js_1 = tslib_1.__importDefault(require(160) /* ../less/environment/abstract-plugin-loader.js */);
    /**
     * Node Plugin Loader
     */
    var PluginLoader = function (less) {
        this.less = less;
        this.require = function (prefix) {
            prefix = path_1.default.dirname(prefix);
            return function (id) {
                var str = id.substr(0, 2);
                if (str === '..' || str === './') {
                    return require(path_1.default.join(prefix, id));
                }
                else {
                    return require(id);
                }
            };
        };
    };
    PluginLoader.prototype = Object.assign(new abstract_plugin_loader_js_1.default(), {
        loadPlugin: function (filename, basePath, context, environment, fileManager) {
            var prefix = filename.slice(0, 1);
            var explicit = prefix === '.' || prefix === '/' || filename.slice(-3).toLowerCase() === '.js';
            if (!explicit) {
                context.prefixes = ['less-plugin-', ''];
            }
            if (context.syncImport) {
                return fileManager.loadFileSync(filename, basePath, context, environment);
            }
            return new Promise(function (fulfill, reject) {
                fileManager.loadFile(filename, basePath, context, environment).then(function (data) {
                    try {
                        fulfill(data);
                    }
                    catch (e) {
                        console.log(e);
                        reject(e);
                    }
                }).catch(function (err) {
                    reject(err);
                });
            });
        },
        loadPluginSync: function (filename, basePath, context, environment, fileManager) {
            context.syncImport = true;
            return this.loadPlugin(filename, basePath, context, environment, fileManager);
        }
    });
    exports.default = PluginLoader;
},
/* less/lib/less/default-options.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    // Export a new default each time
    function default_1() {
        return {
            /* Inline Javascript - @plugin still allowed */
            javascriptEnabled: false,
            /* Outputs a makefile import dependency list to stdout. */
            depends: false,
            /* (DEPRECATED) Compress using less built-in compression.
             * This does an okay job but does not utilise all the tricks of
             * dedicated css compression. */
            compress: false,
            /* Runs the less parser and just reports errors without any output. */
            lint: false,
            /* Sets available include paths.
             * If the file in an @import rule does not exist at that exact location,
             * less will look for it at the location(s) passed to this option.
             * You might use this for instance to specify a path to a library which
             * you want to be referenced simply and relatively in the less files. */
            paths: [],
            /* color output in the terminal */
            color: true,
            /* The strictImports controls whether the compiler will allow an @import inside of either
             * @media blocks or (a later addition) other selector blocks.
             * See: https://github.com/less/less.js/issues/656 */
            strictImports: false,
            /* Allow Imports from Insecure HTTPS Hosts */
            insecure: false,
            /* Allows you to add a path to every generated import and url in your css.
             * This does not affect less import statements that are processed, just ones
             * that are left in the output css. */
            rootpath: '',
            /* By default URLs are kept as-is, so if you import a file in a sub-directory
             * that references an image, exactly the same URL will be output in the css.
             * This option allows you to re-write URL's in imported files so that the
             * URL is always relative to the base imported file */
            rewriteUrls: false,
            /* How to process math
             *   0 always           - eagerly try to solve all operations
             *   1 parens-division  - require parens for division "/"
             *   2 parens | strict  - require parens for all operations
             *   3 strict-legacy    - legacy strict behavior (super-strict)
             */
            math: 1,
            /* Without this option, less attempts to guess at the output unit when it does maths. */
            strictUnits: false,
            /* Effectively the declaration is put at the top of your base Less file,
             * meaning it can be used but it also can be overridden if this variable
             * is defined in the file. */
            globalVars: null,
            /* As opposed to the global variable option, this puts the declaration at the
             * end of your base file, meaning it will override anything defined in your Less file. */
            modifyVars: null,
            /* This option allows you to specify a argument to go on to every URL.  */
            urlArgs: ''
        };
    }
    exports.default = default_1;
    ;
},
/* less/lib/less-node/image-size.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var tslib_1 = require(58) /* tslib */;
    var dimension_1 = tslib_1.__importDefault(require(137) /* ../less/tree/dimension */);
    var expression_1 = tslib_1.__importDefault(require(142) /* ../less/tree/expression */);
    var function_registry_1 = tslib_1.__importDefault(require(133) /* ./../less/functions/function-registry */);
    exports.default = (function (environment) {
        function imageSize(functionContext, filePathNode) {
            var filePath = filePathNode.value;
            var currentFileInfo = functionContext.currentFileInfo;
            var currentDirectory = currentFileInfo.rewriteUrls ?
                currentFileInfo.currentDirectory : currentFileInfo.entryPath;
            var fragmentStart = filePath.indexOf('#');
            var fragment = '';
            if (fragmentStart !== -1) {
                fragment = filePath.slice(fragmentStart);
                filePath = filePath.slice(0, fragmentStart);
            }
            var fileManager = environment.getFileManager(filePath, currentDirectory, functionContext.context, environment, true);
            if (!fileManager) {
                throw {
                    type: 'File',
                    message: "Can not set up FileManager for " + filePathNode
                };
            }
            var fileSync = fileManager.loadFileSync(filePath, currentDirectory, functionContext.context, environment);
            if (fileSync.error) {
                throw fileSync.error;
            }
            var sizeOf = require(198) /* image-size */;
            return sizeOf(fileSync.filename);
        }
        var imageFunctions = {
            'image-size': function (filePathNode) {
                var size = imageSize(this, filePathNode);
                return new expression_1.default([
                    new dimension_1.default(size.width, 'px'),
                    new dimension_1.default(size.height, 'px')
                ]);
            },
            'image-width': function (filePathNode) {
                var size = imageSize(this, filePathNode);
                return new dimension_1.default(size.width, 'px');
            },
            'image-height': function (filePathNode) {
                var size = imageSize(this, filePathNode);
                return new dimension_1.default(size.height, 'px');
            }
        };
        function_registry_1.default.addMultiple(imageFunctions);
    });
},
/* image-size/lib/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var fs = require('fs');
    var path = require('path');
    var detector = require(199) /* ./detector */;
    var handlers = {};
    var types = require(200) /* ./types */;
    // load all available handlers
    types.forEach(function (type) {
        handlers[type] = require('./types/' + type);
    });
    // Maximum buffer size, with a default of 128 kilobytes.
    // TO-DO: make this adaptive based on the initial signature of the image
    var MaxBufferSize = 128 * 1024;
    function lookup(buffer, filepath) {
        // detect the file type.. don't rely on the extension
        var type = detector(buffer, filepath);
        // find an appropriate handler for this file type
        if (type in handlers) {
            var size = handlers[type].calculate(buffer, filepath);
            if (size !== false) {
                size.type = type;
                return size;
            }
        }
        // throw up, if we don't understand the file
        throw new TypeError('unsupported file type: ' + type + ' (file: ' + filepath + ')');
    }
    function asyncFileToBuffer(filepath, callback) {
        // open the file in read only mode
        fs.open(filepath, 'r', function (err, descriptor) {
            if (err) {
                return callback(err);
            }
            var size = fs.fstatSync(descriptor).size;
            if (size <= 0) {
                return callback(new Error("File size is not greater than 0 —— " + filepath));
            }
            var bufferSize = Math.min(size, MaxBufferSize);
            var buffer = new Buffer(bufferSize);
            // read first buffer block from the file, asynchronously
            fs.read(descriptor, buffer, 0, bufferSize, 0, function (err) {
                if (err) {
                    return callback(err);
                }
                // close the file, we are done
                fs.close(descriptor, function (err) {
                    callback(err, buffer);
                });
            });
        });
    }
    function syncFileToBuffer(filepath) {
        // read from the file, synchronously
        var descriptor = fs.openSync(filepath, 'r');
        var size = fs.fstatSync(descriptor).size;
        var bufferSize = Math.min(size, MaxBufferSize);
        var buffer = new Buffer(bufferSize);
        fs.readSync(descriptor, buffer, 0, bufferSize, 0);
        fs.closeSync(descriptor);
        return buffer;
    }
    /**
     * @params input - buffer or relative/absolute path of the image file
     * @params callback - optional function for async detection
     */
    module.exports = function (input, callback) {
        // Handle buffer input
        if (Buffer.isBuffer(input)) {
            return lookup(input);
        }
        // input should be a string at this point
        if (typeof input !== 'string') {
            throw new TypeError('invalid invocation');
        }
        // resolve the file path
        var filepath = path.resolve(input);
        if (typeof callback === 'function') {
            asyncFileToBuffer(filepath, function (err, buffer) {
                if (err) {
                    return callback(err);
                }
                // return the dimensions
                var dimensions;
                try {
                    dimensions = lookup(buffer, filepath);
                }
                catch (e) {
                    err = e;
                }
                callback(err, dimensions);
            });
        }
        else {
            var buffer = syncFileToBuffer(filepath);
            return lookup(buffer, filepath);
        }
    };
    module.exports.types = types;
},
/* image-size/lib/detector.js */ function _(require, module, exports, __esModule, __esExport) {
    var typeMap = {};
    var types = require(200) /* ./types */;
    // load all available handlers
    types.forEach(function (type) {
        typeMap[type] = require('./types/' + type).detect;
    });
    module.exports = function (buffer, filepath) {
        var type, result;
        for (type in typeMap) {
            result = typeMap[type](buffer, filepath);
            if (result) {
                return type;
            }
        }
    };
},
/* image-size/lib/types.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = [
        'bmp',
        'gif',
        'jpg',
        'png',
        'psd',
        'svg',
        'tiff',
        'webp',
        'dds'
    ];
},
/* prelude.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    // TODO import license from "../../../LICENSE.txt"
    const license = `\
Copyright (c) 2012 - 2021, Anaconda, Inc., and Bokeh Contributors
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of Anaconda nor the names of any contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.
`;
    function comment(text) {
        return `/*!${"\n"}${text.trim().split("\n").map((line) => ` * ${line}`).join("\n")}${"\n"} */`;
    }
    exports.comment = comment;
    function str(obj) {
        return JSON.stringify(obj);
    }
    const loader = `\
(function(modules, entry, aliases, externals) {
  if (aliases === undefined) aliases = {};
  if (externals === undefined) externals = {};

  const cache = {};

  const normalize = function(name) {
    if (typeof name === "number")
      return name;

    if (name === "bokehjs")
      return entry;

    if (!externals[name]) {
      const prefix = "@bokehjs/"
      if (name.slice(0, prefix.length) === prefix)
        name = name.slice(prefix.length)
    }

    const alias = aliases[name]
    if (alias != null)
      return alias;

    const trailing = name.length > 0 && name[name.lenght-1] === "/";
    const index = aliases[name + (trailing ? "" : "/") + "index"];
    if (index != null)
      return index;

    return name;
  }

  const require = function(name) {
    let mod = cache[name];
    if (!mod) {
      const id = normalize(name);

      mod = cache[id];
      if (!mod) {
        if (!modules[id]) {
          if (externals[id] === false || (externals[id] == true && parent_require)) {
            try {
              mod = {exports: externals[id] ? parent_require(id) : {}};
              cache[id] = cache[name] = mod;
              return mod.exports;
            } catch (e) {}
          }

          const err = new Error("Cannot find module '" + name + "'");
          err.code = 'MODULE_NOT_FOUND';
          throw err;
        }

        mod = {exports: {}};
        cache[id] = cache[name] = mod;

        function __esModule() {
          Object.defineProperty(mod.exports, "__esModule", {value: true});
        }

        function __esExport(name, value) {
          Object.defineProperty(mod.exports, name, {
            enumerable: true, get: function () { return value; }
          });
        }

        modules[id].call(mod.exports, require, mod, mod.exports, __esModule, __esExport);
      } else {
        cache[name] = mod;
      }
    }

    return mod.exports;
  }
  require.resolve = function(name) {
    return ""
  }

  const main = require(entry);
  main.require = require;

  if (typeof Proxy !== "undefined") {
    // allow Bokeh.loader["@bokehjs/module/name"] syntax
    main.loader = new Proxy({}, {
      get: function(_obj, module) {
        return require(module);
      }
    });
  }

  main.register_plugin = function(plugin_modules, plugin_entry, plugin_aliases, plugin_externals) {
    if (plugin_aliases === undefined) plugin_aliases = {};
    if (plugin_externals === undefined) plugin_externals = {};

    for (let name in plugin_modules) {
      modules[name] = plugin_modules[name];
    }

    for (let name in plugin_aliases) {
      aliases[name] = plugin_aliases[name];
    }

    for (let name in plugin_externals) {
      externals[name] = plugin_externals[name];
    }

    const plugin = require(plugin_entry);

    for (let name in plugin) {
      main[name] = plugin[name];
    }

    return plugin;
  }

  return main;
})
`;
    function prelude_esm() {
        return `\
${comment(license)}
const main = ${loader}\
`;
    }
    exports.prelude_esm = prelude_esm;
    function postlude_esm() {
        return "export default main;";
    }
    exports.postlude_esm = postlude_esm;
    function plugin_prelude_esm() {
        return `\
${comment(license)}
import main from "./bokeh.esm.js";

const plugin = (function(modules, entry, aliases, externals) {
  main.register_plugin(modules, entry, aliases);
})\
`;
    }
    exports.plugin_prelude_esm = plugin_prelude_esm;
    function plugin_postlude_esm() {
        return "export default plugin;";
    }
    exports.plugin_postlude_esm = plugin_postlude_esm;
    function prelude() {
        return `\
${comment(license)}
(function(root, factory) {
  const bokeh = factory();
  bokeh.__bokeh__ = true;
  if (typeof root.Bokeh === "undefined" || typeof root.Bokeh.__bokeh__ === "undefined") {
    root.Bokeh = bokeh;
  }
  const Bokeh = root.Bokeh;
  Bokeh[bokeh.version] = bokeh;
})(this, function() {
  let define;
  const parent_require = typeof require === "function" && require
  return ${loader}\
`;
    }
    exports.prelude = prelude;
    function postlude() {
        return "});";
    }
    exports.postlude = postlude;
    function plugin_postlude() {
        return "});";
    }
    exports.plugin_postlude = plugin_postlude;
    function plugin_prelude(options) {
        return `\
${comment(license)}
(function(root, factory) {
  factory(root["Bokeh"], ${str(options?.version)});
})(this, function(Bokeh, version) {
  let define;
  return (function(modules, entry, aliases, externals) {
    const bokeh = typeof Bokeh !== "undefined" && (version != null ? Bokeh[version] : Bokeh);
    if (bokeh != null) {
      return bokeh.register_plugin(modules, entry, aliases);
    } else {
      throw new Error("Cannot find Bokeh " + version + ". You have to load it prior to loading plugins.");
    }
  })
`;
    }
    exports.plugin_prelude = plugin_prelude;
    function default_prelude(options) {
        return `\
(function(root, factory) {
  ${options?.global != null ? `root[${str(options.global)}] = factory()` : "Object.assign(root, factory())"};
})(this, function() {
  const parent_require = typeof require === "function" && require
  return ${loader}
`;
    }
    exports.default_prelude = default_prelude;
},
/* tsconfig.ext.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "compilerOptions": {
            "noImplicitAny": true,
            "noImplicitThis": true,
            "noImplicitReturns": true,
            "noUnusedLocals": true,
            "noUnusedParameters": true,
            "noImplicitOverride": false,
            "strictNullChecks": true,
            "strictBindCallApply": true,
            "strictFunctionTypes": true,
            "strictPropertyInitialization": false,
            "useUnknownInCatchVariables": true,
            "alwaysStrict": true,
            "noErrorTruncation": true,
            "noEmitOnError": false,
            "allowJs": true,
            "declaration": true,
            "sourceMap": true,
            "declarationMap": true,
            "importHelpers": false,
            "experimentalDecorators": true,
            "module": "ES2020",
            "moduleResolution": "node",
            "resolveJsonModule": true,
            "skipLibCheck": true,
            "esModuleInterop": true,
            "target": "ES2017",
            "lib": ["es2017", "dom", "dom.iterable"],
            "baseUrl": ".",
            "outDir": "./dist/lib"
        },
        "include": ["./**/*.ts"]
    };
    module.exports = json;
},
/* eslint/lib/api.js */ function _(require, module, exports, __esModule, __esExport) {
    const { CLIEngine } = require(204) /* ./cli-engine */;
    const { ESLint } = require(724) /* ./eslint */;
    const { Linter } = require(330) /* ./linter */;
    const { RuleTester } = require(726) /* ./rule-tester */;
    const { SourceCode } = require(348) /* ./source-code */;
    module.exports = {
        Linter,
        CLIEngine,
        ESLint,
        RuleTester,
        SourceCode
    };
    // DOTO: remove deprecated API.
    let deprecatedLinterInstance = null;
    Object.defineProperty(module.exports, "linter", {
        enumerable: false,
        get() {
            if (!deprecatedLinterInstance) {
                deprecatedLinterInstance = new Linter();
            }
            return deprecatedLinterInstance;
        }
    });
},
/* eslint/lib/cli-engine/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const { CLIEngine } = require(205) /* ./cli-engine */;
    module.exports = {
        CLIEngine
    };
},
/* eslint/lib/cli-engine/cli-engine.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
     * The CLI object should *not* call process.exit() directly. It should only return
     * exit codes. This allows other programs to use the CLI object and still control
     * when the program exits.
     */
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const fs = require("fs");
    const path = require("path");
    const defaultOptions = require(206) /* ../../conf/default-cli-options */;
    const pkg = require(207) /* ../../package.json */;
    const { Legacy: { ConfigOps, naming, CascadingConfigArrayFactory, IgnorePattern, getUsedExtractedConfigs, ModuleResolver } } = require(208) /* @eslint/eslintrc */;
    const { FileEnumerator } = require(325) /* ./file-enumerator */;
    const { Linter } = require(330) /* ../linter */;
    const builtInRules = require(391) /* ../rules */;
    const loadRules = require(702) /* ./load-rules */;
    const hash = require(703) /* ./hash */;
    const LintResultCache = require(705) /* ./lint-result-cache */;
    const debug = require(72) /* debug */("eslint:cli-engine");
    const validFixTypes = new Set(["problem", "suggestion", "layout"]);
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    // For VSCode IntelliSense
    /** @typedef {import("../shared/types").ConfigData} ConfigData */
    /** @typedef {import("../shared/types").DeprecatedRuleInfo} DeprecatedRuleInfo */
    /** @typedef {import("../shared/types").LintMessage} LintMessage */
    /** @typedef {import("../shared/types").ParserOptions} ParserOptions */
    /** @typedef {import("../shared/types").Plugin} Plugin */
    /** @typedef {import("../shared/types").RuleConf} RuleConf */
    /** @typedef {import("../shared/types").Rule} Rule */
    /** @typedef {ReturnType<CascadingConfigArrayFactory["getConfigArrayForFile"]>} ConfigArray */
    /** @typedef {ReturnType<ConfigArray["extractConfig"]>} ExtractedConfig */
    /**
     * The options to configure a CLI engine with.
     * @typedef {Object} CLIEngineOptions
     * @property {boolean} [allowInlineConfig] Enable or disable inline configuration comments.
     * @property {ConfigData} [baseConfig] Base config object, extended by all configs used with this CLIEngine instance
     * @property {boolean} [cache] Enable result caching.
     * @property {string} [cacheLocation] The cache file to use instead of .eslintcache.
     * @property {string} [configFile] The configuration file to use.
     * @property {string} [cwd] The value to use for the current working directory.
     * @property {string[]} [envs] An array of environments to load.
     * @property {string[]|null} [extensions] An array of file extensions to check.
     * @property {boolean|Function} [fix] Execute in autofix mode. If a function, should return a boolean.
     * @property {string[]} [fixTypes] Array of rule types to apply fixes for.
     * @property {string[]} [globals] An array of global variables to declare.
     * @property {boolean} [ignore] False disables use of .eslintignore.
     * @property {string} [ignorePath] The ignore file to use instead of .eslintignore.
     * @property {string|string[]} [ignorePattern] One or more glob patterns to ignore.
     * @property {boolean} [useEslintrc] False disables looking for .eslintrc
     * @property {string} [parser] The name of the parser to use.
     * @property {ParserOptions} [parserOptions] An object of parserOption settings to use.
     * @property {string[]} [plugins] An array of plugins to load.
     * @property {Record<string,RuleConf>} [rules] An object of rules to use.
     * @property {string[]} [rulePaths] An array of directories to load custom rules from.
     * @property {boolean} [reportUnusedDisableDirectives] `true` adds reports for unused eslint-disable directives
     * @property {boolean} [globInputPaths] Set to false to skip glob resolution of input file paths to lint (default: true). If false, each input file paths is assumed to be a non-glob path to an existing file.
     * @property {string} [resolvePluginsRelativeTo] The folder where plugins should be resolved from, defaulting to the CWD
     */
    /**
     * A linting result.
     * @typedef {Object} LintResult
     * @property {string} filePath The path to the file that was linted.
     * @property {LintMessage[]} messages All of the messages for the result.
     * @property {number} errorCount Number of errors for the result.
     * @property {number} warningCount Number of warnings for the result.
     * @property {number} fixableErrorCount Number of fixable errors for the result.
     * @property {number} fixableWarningCount Number of fixable warnings for the result.
     * @property {string} [source] The source code of the file that was linted.
     * @property {string} [output] The source code of the file that was linted, with as many fixes applied as possible.
     */
    /**
     * Linting results.
     * @typedef {Object} LintReport
     * @property {LintResult[]} results All of the result.
     * @property {number} errorCount Number of errors for the result.
     * @property {number} warningCount Number of warnings for the result.
     * @property {number} fixableErrorCount Number of fixable errors for the result.
     * @property {number} fixableWarningCount Number of fixable warnings for the result.
     * @property {DeprecatedRuleInfo[]} usedDeprecatedRules The list of used deprecated rules.
     */
    /**
     * Private data for CLIEngine.
     * @typedef {Object} CLIEngineInternalSlots
     * @property {Map<string, Plugin>} additionalPluginPool The map for additional plugins.
     * @property {string} cacheFilePath The path to the cache of lint results.
     * @property {CascadingConfigArrayFactory} configArrayFactory The factory of configs.
     * @property {(filePath: string) => boolean} defaultIgnores The default predicate function to check if a file ignored or not.
     * @property {FileEnumerator} fileEnumerator The file enumerator.
     * @property {ConfigArray[]} lastConfigArrays The list of config arrays that the last `executeOnFiles` or `executeOnText` used.
     * @property {LintResultCache|null} lintResultCache The cache of lint results.
     * @property {Linter} linter The linter instance which has loaded rules.
     * @property {CLIEngineOptions} options The normalized options of this instance.
     */
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /** @type {WeakMap<CLIEngine, CLIEngineInternalSlots>} */
    const internalSlotsMap = new WeakMap();
    /**
     * Determines if each fix type in an array is supported by ESLint and throws
     * an error if not.
     * @param {string[]} fixTypes An array of fix types to check.
     * @returns {void}
     * @throws {Error} If an invalid fix type is found.
     */
    function validateFixTypes(fixTypes) {
        for (const fixType of fixTypes) {
            if (!validFixTypes.has(fixType)) {
                throw new Error(`Invalid fix type "${fixType}" found.`);
            }
        }
    }
    /**
     * It will calculate the error and warning count for collection of messages per file
     * @param {LintMessage[]} messages Collection of messages
     * @returns {Object} Contains the stats
     * @private
     */
    function calculateStatsPerFile(messages) {
        return messages.reduce((stat, message) => {
            if (message.fatal || message.severity === 2) {
                stat.errorCount++;
                if (message.fatal) {
                    stat.fatalErrorCount++;
                }
                if (message.fix) {
                    stat.fixableErrorCount++;
                }
            }
            else {
                stat.warningCount++;
                if (message.fix) {
                    stat.fixableWarningCount++;
                }
            }
            return stat;
        }, {
            errorCount: 0,
            fatalErrorCount: 0,
            warningCount: 0,
            fixableErrorCount: 0,
            fixableWarningCount: 0
        });
    }
    /**
     * It will calculate the error and warning count for collection of results from all files
     * @param {LintResult[]} results Collection of messages from all the files
     * @returns {Object} Contains the stats
     * @private
     */
    function calculateStatsPerRun(results) {
        return results.reduce((stat, result) => {
            stat.errorCount += result.errorCount;
            stat.fatalErrorCount += result.fatalErrorCount;
            stat.warningCount += result.warningCount;
            stat.fixableErrorCount += result.fixableErrorCount;
            stat.fixableWarningCount += result.fixableWarningCount;
            return stat;
        }, {
            errorCount: 0,
            fatalErrorCount: 0,
            warningCount: 0,
            fixableErrorCount: 0,
            fixableWarningCount: 0
        });
    }
    /**
     * Processes an source code using ESLint.
     * @param {Object} config The config object.
     * @param {string} config.text The source code to verify.
     * @param {string} config.cwd The path to the current working directory.
     * @param {string|undefined} config.filePath The path to the file of `text`. If this is undefined, it uses `<text>`.
     * @param {ConfigArray} config.config The config.
     * @param {boolean} config.fix If `true` then it does fix.
     * @param {boolean} config.allowInlineConfig If `true` then it uses directive comments.
     * @param {boolean} config.reportUnusedDisableDirectives If `true` then it reports unused `eslint-disable` comments.
     * @param {FileEnumerator} config.fileEnumerator The file enumerator to check if a path is a target or not.
     * @param {Linter} config.linter The linter instance to verify.
     * @returns {LintResult} The result of linting.
     * @private
     */
    function verifyText({ text, cwd, filePath: providedFilePath, config, fix, allowInlineConfig, reportUnusedDisableDirectives, fileEnumerator, linter }) {
        const filePath = providedFilePath || "<text>";
        debug(`Lint ${filePath}`);
        /*
         * Verify.
         * `config.extractConfig(filePath)` requires an absolute path, but `linter`
         * doesn't know CWD, so it gives `linter` an absolute path always.
         */
        const filePathToVerify = filePath === "<text>" ? path.join(cwd, filePath) : filePath;
        const { fixed, messages, output } = linter.verifyAndFix(text, config, {
            allowInlineConfig,
            filename: filePathToVerify,
            fix,
            reportUnusedDisableDirectives,
            /**
             * Check if the linter should adopt a given code block or not.
             * @param {string} blockFilename The virtual filename of a code block.
             * @returns {boolean} `true` if the linter should adopt the code block.
             */
            filterCodeBlock(blockFilename) {
                return fileEnumerator.isTargetPath(blockFilename);
            }
        });
        // Tweak and return.
        const result = {
            filePath,
            messages,
            ...calculateStatsPerFile(messages)
        };
        if (fixed) {
            result.output = output;
        }
        if (result.errorCount + result.warningCount > 0 &&
            typeof result.output === "undefined") {
            result.source = text;
        }
        return result;
    }
    /**
     * Returns result with warning by ignore settings
     * @param {string} filePath File path of checked code
     * @param {string} baseDir  Absolute path of base directory
     * @returns {LintResult} Result with single warning
     * @private
     */
    function createIgnoreResult(filePath, baseDir) {
        let message;
        const isHidden = filePath.split(path.sep)
            .find(segment => /^\./u.test(segment));
        const isInNodeModules = baseDir && path.relative(baseDir, filePath).startsWith("node_modules");
        if (isHidden) {
            message = "File ignored by default.  Use a negated ignore pattern (like \"--ignore-pattern '!<relative/path/to/filename>'\") to override.";
        }
        else if (isInNodeModules) {
            message = "File ignored by default. Use \"--ignore-pattern '!node_modules/*'\" to override.";
        }
        else {
            message = "File ignored because of a matching ignore pattern. Use \"--no-ignore\" to override.";
        }
        return {
            filePath: path.resolve(filePath),
            messages: [
                {
                    fatal: false,
                    severity: 1,
                    message
                }
            ],
            errorCount: 0,
            warningCount: 1,
            fixableErrorCount: 0,
            fixableWarningCount: 0
        };
    }
    /**
     * Get a rule.
     * @param {string} ruleId The rule ID to get.
     * @param {ConfigArray[]} configArrays The config arrays that have plugin rules.
     * @returns {Rule|null} The rule or null.
     */
    function getRule(ruleId, configArrays) {
        for (const configArray of configArrays) {
            const rule = configArray.pluginRules.get(ruleId);
            if (rule) {
                return rule;
            }
        }
        return builtInRules.get(ruleId) || null;
    }
    /**
     * Collect used deprecated rules.
     * @param {ConfigArray[]} usedConfigArrays The config arrays which were used.
     * @returns {IterableIterator<DeprecatedRuleInfo>} Used deprecated rules.
     */
    function* iterateRuleDeprecationWarnings(usedConfigArrays) {
        const processedRuleIds = new Set();
        // Flatten used configs.
        /** @type {ExtractedConfig[]} */
        const configs = [].concat(...usedConfigArrays.map(getUsedExtractedConfigs));
        // Traverse rule configs.
        for (const config of configs) {
            for (const [ruleId, ruleConfig] of Object.entries(config.rules)) {
                // Skip if it was processed.
                if (processedRuleIds.has(ruleId)) {
                    continue;
                }
                processedRuleIds.add(ruleId);
                // Skip if it's not used.
                if (!ConfigOps.getRuleSeverity(ruleConfig)) {
                    continue;
                }
                const rule = getRule(ruleId, usedConfigArrays);
                // Skip if it's not deprecated.
                if (!(rule && rule.meta && rule.meta.deprecated)) {
                    continue;
                }
                // This rule was used and deprecated.
                yield {
                    ruleId,
                    replacedBy: rule.meta.replacedBy || []
                };
            }
        }
    }
    /**
     * Checks if the given message is an error message.
     * @param {LintMessage} message The message to check.
     * @returns {boolean} Whether or not the message is an error message.
     * @private
     */
    function isErrorMessage(message) {
        return message.severity === 2;
    }
    /**
     * return the cacheFile to be used by eslint, based on whether the provided parameter is
     * a directory or looks like a directory (ends in `path.sep`), in which case the file
     * name will be the `cacheFile/.cache_hashOfCWD`
     *
     * if cacheFile points to a file or looks like a file then in will just use that file
     * @param {string} cacheFile The name of file to be used to store the cache
     * @param {string} cwd Current working directory
     * @returns {string} the resolved path to the cache file
     */
    function getCacheFile(cacheFile, cwd) {
        /*
         * make sure the path separators are normalized for the environment/os
         * keeping the trailing path separator if present
         */
        const normalizedCacheFile = path.normalize(cacheFile);
        const resolvedCacheFile = path.resolve(cwd, normalizedCacheFile);
        const looksLikeADirectory = normalizedCacheFile.slice(-1) === path.sep;
        /**
         * return the name for the cache file in case the provided parameter is a directory
         * @returns {string} the resolved path to the cacheFile
         */
        function getCacheFileForDirectory() {
            return path.join(resolvedCacheFile, `.cache_${hash(cwd)}`);
        }
        let fileStats;
        try {
            fileStats = fs.lstatSync(resolvedCacheFile);
        }
        catch {
            fileStats = null;
        }
        /*
         * in case the file exists we need to verify if the provided path
         * is a directory or a file. If it is a directory we want to create a file
         * inside that directory
         */
        if (fileStats) {
            /*
             * is a directory or is a file, but the original file the user provided
             * looks like a directory but `path.resolve` removed the `last path.sep`
             * so we need to still treat this like a directory
             */
            if (fileStats.isDirectory() || looksLikeADirectory) {
                return getCacheFileForDirectory();
            }
            // is file so just use that file
            return resolvedCacheFile;
        }
        /*
         * here we known the file or directory doesn't exist,
         * so we will try to infer if its a directory if it looks like a directory
         * for the current operating system.
         */
        // if the last character passed is a path separator we assume is a directory
        if (looksLikeADirectory) {
            return getCacheFileForDirectory();
        }
        return resolvedCacheFile;
    }
    /**
     * Convert a string array to a boolean map.
     * @param {string[]|null} keys The keys to assign true.
     * @param {boolean} defaultValue The default value for each property.
     * @param {string} displayName The property name which is used in error message.
     * @returns {Record<string,boolean>} The boolean map.
     */
    function toBooleanMap(keys, defaultValue, displayName) {
        if (keys && !Array.isArray(keys)) {
            throw new Error(`${displayName} must be an array.`);
        }
        if (keys && keys.length > 0) {
            return keys.reduce((map, def) => {
                const [key, value] = def.split(":");
                if (key !== "__proto__") {
                    map[key] = value === void 0
                        ? defaultValue
                        : value === "true";
                }
                return map;
            }, {});
        }
        return void 0;
    }
    /**
     * Create a config data from CLI options.
     * @param {CLIEngineOptions} options The options
     * @returns {ConfigData|null} The created config data.
     */
    function createConfigDataFromOptions(options) {
        const { ignorePattern, parser, parserOptions, plugins, rules } = options;
        const env = toBooleanMap(options.envs, true, "envs");
        const globals = toBooleanMap(options.globals, false, "globals");
        if (env === void 0 &&
            globals === void 0 &&
            (ignorePattern === void 0 || ignorePattern.length === 0) &&
            parser === void 0 &&
            parserOptions === void 0 &&
            plugins === void 0 &&
            rules === void 0) {
            return null;
        }
        return {
            env,
            globals,
            ignorePatterns: ignorePattern,
            parser,
            parserOptions,
            plugins,
            rules
        };
    }
    /**
     * Checks whether a directory exists at the given location
     * @param {string} resolvedPath A path from the CWD
     * @returns {boolean} `true` if a directory exists
     */
    function directoryExists(resolvedPath) {
        try {
            return fs.statSync(resolvedPath).isDirectory();
        }
        catch (error) {
            if (error && (error.code === "ENOENT" || error.code === "ENOTDIR")) {
                return false;
            }
            throw error;
        }
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    class CLIEngine {
        /**
         * Creates a new instance of the core CLI engine.
         * @param {CLIEngineOptions} providedOptions The options for this instance.
         */
        constructor(providedOptions) {
            const options = Object.assign(Object.create(null), defaultOptions, { cwd: process.cwd() }, providedOptions);
            if (options.fix === void 0) {
                options.fix = false;
            }
            const additionalPluginPool = new Map();
            const cacheFilePath = getCacheFile(options.cacheLocation || options.cacheFile, options.cwd);
            const configArrayFactory = new CascadingConfigArrayFactory({
                additionalPluginPool,
                baseConfig: options.baseConfig || null,
                cliConfig: createConfigDataFromOptions(options),
                cwd: options.cwd,
                ignorePath: options.ignorePath,
                resolvePluginsRelativeTo: options.resolvePluginsRelativeTo,
                rulePaths: options.rulePaths,
                specificConfigPath: options.configFile,
                useEslintrc: options.useEslintrc,
                builtInRules,
                loadRules,
                eslintRecommendedPath: path.resolve(__dirname, "../../conf/eslint-recommended.js"),
                eslintAllPath: path.resolve(__dirname, "../../conf/eslint-all.js")
            });
            const fileEnumerator = new FileEnumerator({
                configArrayFactory,
                cwd: options.cwd,
                extensions: options.extensions,
                globInputPaths: options.globInputPaths,
                errorOnUnmatchedPattern: options.errorOnUnmatchedPattern,
                ignore: options.ignore
            });
            const lintResultCache = options.cache ? new LintResultCache(cacheFilePath, options.cacheStrategy) : null;
            const linter = new Linter({ cwd: options.cwd });
            /** @type {ConfigArray[]} */
            const lastConfigArrays = [configArrayFactory.getConfigArrayForFile()];
            // Store private data.
            internalSlotsMap.set(this, {
                additionalPluginPool,
                cacheFilePath,
                configArrayFactory,
                defaultIgnores: IgnorePattern.createDefaultIgnore(options.cwd),
                fileEnumerator,
                lastConfigArrays,
                lintResultCache,
                linter,
                options
            });
            // setup special filter for fixes
            if (options.fix && options.fixTypes && options.fixTypes.length > 0) {
                debug(`Using fix types ${options.fixTypes}`);
                // throw an error if any invalid fix types are found
                validateFixTypes(options.fixTypes);
                // convert to Set for faster lookup
                const fixTypes = new Set(options.fixTypes);
                // save original value of options.fix in case it's a function
                const originalFix = (typeof options.fix === "function")
                    ? options.fix : () => true;
                options.fix = message => {
                    const rule = message.ruleId && getRule(message.ruleId, lastConfigArrays);
                    const matches = rule && rule.meta && fixTypes.has(rule.meta.type);
                    return matches && originalFix(message);
                };
            }
        }
        getRules() {
            const { lastConfigArrays } = internalSlotsMap.get(this);
            return new Map(function* () {
                yield* builtInRules;
                for (const configArray of lastConfigArrays) {
                    yield* configArray.pluginRules;
                }
            }());
        }
        /**
         * Returns results that only contains errors.
         * @param {LintResult[]} results The results to filter.
         * @returns {LintResult[]} The filtered results.
         */
        static getErrorResults(results) {
            const filtered = [];
            results.forEach(result => {
                const filteredMessages = result.messages.filter(isErrorMessage);
                if (filteredMessages.length > 0) {
                    filtered.push({
                        ...result,
                        messages: filteredMessages,
                        errorCount: filteredMessages.length,
                        warningCount: 0,
                        fixableErrorCount: result.fixableErrorCount,
                        fixableWarningCount: 0
                    });
                }
            });
            return filtered;
        }
        /**
         * Outputs fixes from the given results to files.
         * @param {LintReport} report The report object created by CLIEngine.
         * @returns {void}
         */
        static outputFixes(report) {
            report.results.filter(result => Object.prototype.hasOwnProperty.call(result, "output")).forEach(result => {
                fs.writeFileSync(result.filePath, result.output);
            });
        }
        /**
         * Add a plugin by passing its configuration
         * @param {string} name Name of the plugin.
         * @param {Plugin} pluginObject Plugin configuration object.
         * @returns {void}
         */
        addPlugin(name, pluginObject) {
            const { additionalPluginPool, configArrayFactory, lastConfigArrays } = internalSlotsMap.get(this);
            additionalPluginPool.set(name, pluginObject);
            configArrayFactory.clearCache();
            lastConfigArrays.length = 1;
            lastConfigArrays[0] = configArrayFactory.getConfigArrayForFile();
        }
        /**
         * Resolves the patterns passed into executeOnFiles() into glob-based patterns
         * for easier handling.
         * @param {string[]} patterns The file patterns passed on the command line.
         * @returns {string[]} The equivalent glob patterns.
         */
        resolveFileGlobPatterns(patterns) {
            const { options } = internalSlotsMap.get(this);
            if (options.globInputPaths === false) {
                return patterns.filter(Boolean);
            }
            const extensions = (options.extensions || [".js"]).map(ext => ext.replace(/^\./u, ""));
            const dirSuffix = `/**/*.{${extensions.join(",")}}`;
            return patterns.filter(Boolean).map(pathname => {
                const resolvedPath = path.resolve(options.cwd, pathname);
                const newPath = directoryExists(resolvedPath)
                    ? pathname.replace(/[/\\]$/u, "") + dirSuffix
                    : pathname;
                return path.normalize(newPath).replace(/\\/gu, "/");
            });
        }
        /**
         * Executes the current configuration on an array of file and directory names.
         * @param {string[]} patterns An array of file and directory names.
         * @returns {LintReport} The results for all files that were linted.
         */
        executeOnFiles(patterns) {
            const { cacheFilePath, fileEnumerator, lastConfigArrays, lintResultCache, linter, options: { allowInlineConfig, cache, cwd, fix, reportUnusedDisableDirectives } } = internalSlotsMap.get(this);
            const results = [];
            const startTime = Date.now();
            // Clear the last used config arrays.
            lastConfigArrays.length = 0;
            // Delete cache file; should this do here?
            if (!cache) {
                try {
                    fs.unlinkSync(cacheFilePath);
                }
                catch (error) {
                    const errorCode = error && error.code;
                    // Ignore errors when no such file exists or file system is read only (and cache file does not exist)
                    if (errorCode !== "ENOENT" && !(errorCode === "EROFS" && !fs.existsSync(cacheFilePath))) {
                        throw error;
                    }
                }
            }
            // Iterate source code files.
            for (const { config, filePath, ignored } of fileEnumerator.iterateFiles(patterns)) {
                if (ignored) {
                    results.push(createIgnoreResult(filePath, cwd));
                    continue;
                }
                /*
                 * Store used configs for:
                 * - this method uses to collect used deprecated rules.
                 * - `getRules()` method uses to collect all loaded rules.
                 * - `--fix-type` option uses to get the loaded rule's meta data.
                 */
                if (!lastConfigArrays.includes(config)) {
                    lastConfigArrays.push(config);
                }
                // Skip if there is cached result.
                if (lintResultCache) {
                    const cachedResult = lintResultCache.getCachedLintResults(filePath, config);
                    if (cachedResult) {
                        const hadMessages = cachedResult.messages &&
                            cachedResult.messages.length > 0;
                        if (hadMessages && fix) {
                            debug(`Reprocessing cached file to allow autofix: ${filePath}`);
                        }
                        else {
                            debug(`Skipping file since it hasn't changed: ${filePath}`);
                            results.push(cachedResult);
                            continue;
                        }
                    }
                }
                // Do lint.
                const result = verifyText({
                    text: fs.readFileSync(filePath, "utf8"),
                    filePath,
                    config,
                    cwd,
                    fix,
                    allowInlineConfig,
                    reportUnusedDisableDirectives,
                    fileEnumerator,
                    linter
                });
                results.push(result);
                /*
                 * Store the lint result in the LintResultCache.
                 * NOTE: The LintResultCache will remove the file source and any
                 * other properties that are difficult to serialize, and will
                 * hydrate those properties back in on future lint runs.
                 */
                if (lintResultCache) {
                    lintResultCache.setCachedLintResults(filePath, config, result);
                }
            }
            // Persist the cache to disk.
            if (lintResultCache) {
                lintResultCache.reconcile();
            }
            debug(`Linting complete in: ${Date.now() - startTime}ms`);
            let usedDeprecatedRules;
            return {
                results,
                ...calculateStatsPerRun(results),
                // Initialize it lazily because CLI and `ESLint` API don't use it.
                get usedDeprecatedRules() {
                    if (!usedDeprecatedRules) {
                        usedDeprecatedRules = Array.from(iterateRuleDeprecationWarnings(lastConfigArrays));
                    }
                    return usedDeprecatedRules;
                }
            };
        }
        /**
         * Executes the current configuration on text.
         * @param {string} text A string of JavaScript code to lint.
         * @param {string} [filename] An optional string representing the texts filename.
         * @param {boolean} [warnIgnored] Always warn when a file is ignored
         * @returns {LintReport} The results for the linting.
         */
        executeOnText(text, filename, warnIgnored) {
            const { configArrayFactory, fileEnumerator, lastConfigArrays, linter, options: { allowInlineConfig, cwd, fix, reportUnusedDisableDirectives } } = internalSlotsMap.get(this);
            const results = [];
            const startTime = Date.now();
            const resolvedFilename = filename && path.resolve(cwd, filename);
            // Clear the last used config arrays.
            lastConfigArrays.length = 0;
            if (resolvedFilename && this.isPathIgnored(resolvedFilename)) {
                if (warnIgnored) {
                    results.push(createIgnoreResult(resolvedFilename, cwd));
                }
            }
            else {
                const config = configArrayFactory.getConfigArrayForFile(resolvedFilename || "__placeholder__.js");
                /*
                 * Store used configs for:
                 * - this method uses to collect used deprecated rules.
                 * - `getRules()` method uses to collect all loaded rules.
                 * - `--fix-type` option uses to get the loaded rule's meta data.
                 */
                lastConfigArrays.push(config);
                // Do lint.
                results.push(verifyText({
                    text,
                    filePath: resolvedFilename,
                    config,
                    cwd,
                    fix,
                    allowInlineConfig,
                    reportUnusedDisableDirectives,
                    fileEnumerator,
                    linter
                }));
            }
            debug(`Linting complete in: ${Date.now() - startTime}ms`);
            let usedDeprecatedRules;
            return {
                results,
                ...calculateStatsPerRun(results),
                // Initialize it lazily because CLI and `ESLint` API don't use it.
                get usedDeprecatedRules() {
                    if (!usedDeprecatedRules) {
                        usedDeprecatedRules = Array.from(iterateRuleDeprecationWarnings(lastConfigArrays));
                    }
                    return usedDeprecatedRules;
                }
            };
        }
        /**
         * Returns a configuration object for the given file based on the CLI options.
         * This is the same logic used by the ESLint CLI executable to determine
         * configuration for each file it processes.
         * @param {string} filePath The path of the file to retrieve a config object for.
         * @returns {ConfigData} A configuration object for the file.
         */
        getConfigForFile(filePath) {
            const { configArrayFactory, options } = internalSlotsMap.get(this);
            const absolutePath = path.resolve(options.cwd, filePath);
            if (directoryExists(absolutePath)) {
                throw Object.assign(new Error("'filePath' should not be a directory path."), { messageTemplate: "print-config-with-directory-path" });
            }
            return configArrayFactory
                .getConfigArrayForFile(absolutePath)
                .extractConfig(absolutePath)
                .toCompatibleObjectAsConfigFileContent();
        }
        /**
         * Checks if a given path is ignored by ESLint.
         * @param {string} filePath The path of the file to check.
         * @returns {boolean} Whether or not the given path is ignored.
         */
        isPathIgnored(filePath) {
            const { configArrayFactory, defaultIgnores, options: { cwd, ignore } } = internalSlotsMap.get(this);
            const absolutePath = path.resolve(cwd, filePath);
            if (ignore) {
                const config = configArrayFactory
                    .getConfigArrayForFile(absolutePath)
                    .extractConfig(absolutePath);
                const ignores = config.ignores || defaultIgnores;
                return ignores(absolutePath);
            }
            return defaultIgnores(absolutePath);
        }
        /**
         * Returns the formatter representing the given format or null if the `format` is not a string.
         * @param {string} [format] The name of the format to load or the path to a
         *      custom formatter.
         * @returns {(Function|null)} The formatter function or null if the `format` is not a string.
         */
        getFormatter(format) {
            // default is stylish
            const resolvedFormatName = format || "stylish";
            // only strings are valid formatters
            if (typeof resolvedFormatName === "string") {
                // replace \ with / for Windows compatibility
                const normalizedFormatName = resolvedFormatName.replace(/\\/gu, "/");
                const slots = internalSlotsMap.get(this);
                const cwd = slots ? slots.options.cwd : process.cwd();
                const namespace = naming.getNamespaceFromTerm(normalizedFormatName);
                let formatterPath;
                // if there's a slash, then it's a file (TODO: this check seems dubious for scoped npm packages)
                if (!namespace && normalizedFormatName.indexOf("/") > -1) {
                    formatterPath = path.resolve(cwd, normalizedFormatName);
                }
                else {
                    try {
                        const npmFormat = naming.normalizePackageName(normalizedFormatName, "eslint-formatter");
                        formatterPath = ModuleResolver.resolve(npmFormat, path.join(cwd, "__placeholder__.js"));
                    }
                    catch {
                        formatterPath = path.resolve(__dirname, "formatters", normalizedFormatName);
                    }
                }
                try {
                    return require(formatterPath);
                }
                catch (ex) {
                    ex.message = `There was a problem loading formatter: ${formatterPath}\nError: ${ex.message}`;
                    throw ex;
                }
            }
            else {
                return null;
            }
        }
    }
    CLIEngine.version = pkg.version;
    CLIEngine.getFormatter = CLIEngine.prototype.getFormatter;
    module.exports = {
        CLIEngine,
        /**
         * Get the internal slots of a given CLIEngine instance for tests.
         * @param {CLIEngine} instance The CLIEngine instance to get.
         * @returns {CLIEngineInternalSlots} The internal slots.
         */
        getCLIEngineInternalSlots(instance) {
            return internalSlotsMap.get(instance);
        }
    };
},
/* eslint/conf/default-cli-options.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        configFile: null,
        baseConfig: false,
        rulePaths: [],
        useEslintrc: true,
        envs: [],
        globals: [],
        extensions: null,
        ignore: true,
        ignorePath: void 0,
        cache: false,
        /*
         * in order to honor the cacheFile option if specified
         * this option should not have a default value otherwise
         * it will always be used
         */
        cacheLocation: "",
        cacheFile: ".eslintcache",
        cacheStrategy: "metadata",
        fix: false,
        allowInlineConfig: true,
        reportUnusedDisableDirectives: void 0,
        globInputPaths: true
    };
},
/* eslint/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "eslint",
        "version": "7.32.0",
        "author": "Nicholas C. Zakas <nicholas+npm@nczconsulting.com>",
        "description": "An AST-based pattern checker for JavaScript.",
        "bin": {
            "eslint": "./bin/eslint.js"
        },
        "main": "./lib/api.js",
        "scripts": {
            "test": "node Makefile.js test",
            "test:cli": "mocha",
            "lint": "node Makefile.js lint",
            "fix": "node Makefile.js lint -- fix",
            "fuzz": "node Makefile.js fuzz",
            "generate-release": "node Makefile.js generateRelease",
            "generate-alpharelease": "node Makefile.js generatePrerelease -- alpha",
            "generate-betarelease": "node Makefile.js generatePrerelease -- beta",
            "generate-rcrelease": "node Makefile.js generatePrerelease -- rc",
            "publish-release": "node Makefile.js publishRelease",
            "docs": "node Makefile.js docs",
            "gensite": "node Makefile.js gensite",
            "webpack": "node Makefile.js webpack",
            "perf": "node Makefile.js perf"
        },
        "gitHooks": {
            "pre-commit": "lint-staged"
        },
        "lint-staged": {
            "*.js": "eslint --fix",
            "*.md": "markdownlint"
        },
        "files": [
            "LICENSE",
            "README.md",
            "bin",
            "conf",
            "lib",
            "messages"
        ],
        "repository": "eslint/eslint",
        "funding": "https://opencollective.com/eslint",
        "homepage": "https://eslint.org",
        "bugs": "https://github.com/eslint/eslint/issues/",
        "dependencies": {
            "@babel/code-frame": "7.12.11",
            "@eslint/eslintrc": "^0.4.3",
            "@humanwhocodes/config-array": "^0.5.0",
            "ajv": "^6.10.0",
            "chalk": "^4.0.0",
            "cross-spawn": "^7.0.2",
            "debug": "^4.0.1",
            "doctrine": "^3.0.0",
            "enquirer": "^2.3.5",
            "escape-string-regexp": "^4.0.0",
            "eslint-scope": "^5.1.1",
            "eslint-utils": "^2.1.0",
            "eslint-visitor-keys": "^2.0.0",
            "espree": "^7.3.1",
            "esquery": "^1.4.0",
            "esutils": "^2.0.2",
            "fast-deep-equal": "^3.1.3",
            "file-entry-cache": "^6.0.1",
            "functional-red-black-tree": "^1.0.1",
            "glob-parent": "^5.1.2",
            "globals": "^13.6.0",
            "ignore": "^4.0.6",
            "import-fresh": "^3.0.0",
            "imurmurhash": "^0.1.4",
            "is-glob": "^4.0.0",
            "js-yaml": "^3.13.1",
            "json-stable-stringify-without-jsonify": "^1.0.1",
            "levn": "^0.4.1",
            "lodash.merge": "^4.6.2",
            "minimatch": "^3.0.4",
            "natural-compare": "^1.4.0",
            "optionator": "^0.9.1",
            "progress": "^2.0.0",
            "regexpp": "^3.1.0",
            "semver": "^7.2.1",
            "strip-ansi": "^6.0.0",
            "strip-json-comments": "^3.1.0",
            "table": "^6.0.9",
            "text-table": "^0.2.0",
            "v8-compile-cache": "^2.0.3"
        },
        "devDependencies": {
            "@babel/core": "^7.4.3",
            "@babel/preset-env": "^7.4.3",
            "babel-loader": "^8.0.5",
            "chai": "^4.0.1",
            "cheerio": "^0.22.0",
            "common-tags": "^1.8.0",
            "core-js": "^3.1.3",
            "dateformat": "^3.0.3",
            "ejs": "^3.0.2",
            "eslint": "file:.",
            "eslint-config-eslint": "file:packages/eslint-config-eslint",
            "eslint-plugin-eslint-plugin": "^3.5.3",
            "eslint-plugin-internal-rules": "file:tools/internal-rules",
            "eslint-plugin-jsdoc": "^25.4.3",
            "eslint-plugin-node": "^11.1.0",
            "eslint-release": "^2.0.0",
            "eslump": "^3.0.0",
            "esprima": "^4.0.1",
            "fs-teardown": "0.1.1",
            "glob": "^7.1.6",
            "jsdoc": "^3.5.5",
            "karma": "^6.1.1",
            "karma-chrome-launcher": "^3.1.0",
            "karma-mocha": "^2.0.1",
            "karma-mocha-reporter": "^2.2.5",
            "karma-webpack": "^5.0.0",
            "lint-staged": "^10.1.2",
            "load-perf": "^0.2.0",
            "markdownlint": "^0.19.0",
            "markdownlint-cli": "^0.22.0",
            "memfs": "^3.0.1",
            "mocha": "^8.3.2",
            "mocha-junit-reporter": "^2.0.0",
            "node-polyfill-webpack-plugin": "^1.0.3",
            "npm-license": "^0.3.3",
            "nyc": "^15.0.1",
            "proxyquire": "^2.0.1",
            "puppeteer": "^7.1.0",
            "recast": "^0.19.0",
            "regenerator-runtime": "^0.13.2",
            "shelljs": "^0.8.2",
            "sinon": "^9.0.1",
            "temp": "^0.9.0",
            "webpack": "^5.23.0",
            "webpack-cli": "^4.5.0",
            "yorkie": "^2.0.0"
        },
        "keywords": [
            "ast",
            "lint",
            "javascript",
            "ecmascript",
            "espree"
        ],
        "license": "MIT",
        "engines": {
            "node": "^10.12.0 || >=12.0.0"
        }
    };
    module.exports = json;
},
/* @eslint/eslintrc/lib/index.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { ConfigArrayFactory, createContext: createConfigArrayFactoryContext } = require(209) /* ./config-array-factory */;
    const { CascadingConfigArrayFactory } = require(323) /* ./cascading-config-array-factory */;
    const ModuleResolver = require(271) /* ./shared/relative-module-resolver */;
    const { ConfigArray, getUsedExtractedConfigs } = require(272) /* ./config-array */;
    const { ConfigDependency } = require(277) /* ./config-array/config-dependency */;
    const { ExtractedConfig } = require(274) /* ./config-array/extracted-config */;
    const { IgnorePattern } = require(275) /* ./config-array/ignore-pattern */;
    const { OverrideTester } = require(278) /* ./config-array/override-tester */;
    const ConfigOps = require(220) /* ./shared/config-ops */;
    const ConfigValidator = require(215) /* ./shared/config-validator */;
    const naming = require(270) /* ./shared/naming */;
    const { FlatCompat } = require(324) /* ./flat-compat */;
    //-----------------------------------------------------------------------------
    // Exports
    //-----------------------------------------------------------------------------
    module.exports = {
        Legacy: {
            ConfigArray,
            createConfigArrayFactoryContext,
            CascadingConfigArrayFactory,
            ConfigArrayFactory,
            ConfigDependency,
            ExtractedConfig,
            IgnorePattern,
            OverrideTester,
            getUsedExtractedConfigs,
            // shared
            ConfigOps,
            ConfigValidator,
            ModuleResolver,
            naming
        },
        FlatCompat
    };
},
/* @eslint/eslintrc/lib/config-array-factory.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const fs = require("fs");
    const path = require("path");
    const importFresh = require(210) /* import-fresh */;
    const stripComments = require(214) /* strip-json-comments */;
    const ConfigValidator = require(215) /* ./shared/config-validator */;
    const naming = require(270) /* ./shared/naming */;
    const ModuleResolver = require(271) /* ./shared/relative-module-resolver */;
    const { ConfigArray, ConfigDependency, IgnorePattern, OverrideTester } = require(272) /* ./config-array */;
    const debug = require(72) /* debug */("eslintrc:config-array-factory");
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const configFilenames = [
        ".eslintrc.js",
        ".eslintrc.cjs",
        ".eslintrc.yaml",
        ".eslintrc.yml",
        ".eslintrc.json",
        ".eslintrc",
        "package.json"
    ];
    // Define types for VSCode IntelliSense.
    /** @typedef {import("./shared/types").ConfigData} ConfigData */
    /** @typedef {import("./shared/types").OverrideConfigData} OverrideConfigData */
    /** @typedef {import("./shared/types").Parser} Parser */
    /** @typedef {import("./shared/types").Plugin} Plugin */
    /** @typedef {import("./shared/types").Rule} Rule */
    /** @typedef {import("./config-array/config-dependency").DependentParser} DependentParser */
    /** @typedef {import("./config-array/config-dependency").DependentPlugin} DependentPlugin */
    /** @typedef {ConfigArray[0]} ConfigArrayElement */
    /**
     * @typedef {Object} ConfigArrayFactoryOptions
     * @property {Map<string,Plugin>} [additionalPluginPool] The map for additional plugins.
     * @property {string} [cwd] The path to the current working directory.
     * @property {string} [resolvePluginsRelativeTo] A path to the directory that plugins should be resolved from. Defaults to `cwd`.
     * @property {Map<string,Rule>} builtInRules The rules that are built in to ESLint.
     * @property {Object} [resolver=ModuleResolver] The module resolver object.
     * @property {string} eslintAllPath The path to the definitions for eslint:all.
     * @property {string} eslintRecommendedPath The path to the definitions for eslint:recommended.
     */
    /**
     * @typedef {Object} ConfigArrayFactoryInternalSlots
     * @property {Map<string,Plugin>} additionalPluginPool The map for additional plugins.
     * @property {string} cwd The path to the current working directory.
     * @property {string | undefined} resolvePluginsRelativeTo An absolute path the the directory that plugins should be resolved from.
     * @property {Map<string,Rule>} builtInRules The rules that are built in to ESLint.
     * @property {Object} [resolver=ModuleResolver] The module resolver object.
     * @property {string} eslintAllPath The path to the definitions for eslint:all.
     * @property {string} eslintRecommendedPath The path to the definitions for eslint:recommended.
     */
    /**
     * @typedef {Object} ConfigArrayFactoryLoadingContext
     * @property {string} filePath The path to the current configuration.
     * @property {string} matchBasePath The base path to resolve relative paths in `overrides[].files`, `overrides[].excludedFiles`, and `ignorePatterns`.
     * @property {string} name The name of the current configuration.
     * @property {string} pluginBasePath The base path to resolve plugins.
     * @property {"config" | "ignore" | "implicit-processor"} type The type of the current configuration. This is `"config"` in normal. This is `"ignore"` if it came from `.eslintignore`. This is `"implicit-processor"` if it came from legacy file-extension processors.
     */
    /**
     * @typedef {Object} ConfigArrayFactoryLoadingContext
     * @property {string} filePath The path to the current configuration.
     * @property {string} matchBasePath The base path to resolve relative paths in `overrides[].files`, `overrides[].excludedFiles`, and `ignorePatterns`.
     * @property {string} name The name of the current configuration.
     * @property {"config" | "ignore" | "implicit-processor"} type The type of the current configuration. This is `"config"` in normal. This is `"ignore"` if it came from `.eslintignore`. This is `"implicit-processor"` if it came from legacy file-extension processors.
     */
    /** @type {WeakMap<ConfigArrayFactory, ConfigArrayFactoryInternalSlots>} */
    const internalSlotsMap = new WeakMap();
    /**
     * Check if a given string is a file path.
     * @param {string} nameOrPath A module name or file path.
     * @returns {boolean} `true` if the `nameOrPath` is a file path.
     */
    function isFilePath(nameOrPath) {
        return (/^\.{1,2}[/\\]/u.test(nameOrPath) ||
            path.isAbsolute(nameOrPath));
    }
    /**
     * Convenience wrapper for synchronously reading file contents.
     * @param {string} filePath The filename to read.
     * @returns {string} The file contents, with the BOM removed.
     * @private
     */
    function readFile(filePath) {
        return fs.readFileSync(filePath, "utf8").replace(/^\ufeff/u, "");
    }
    /**
     * Loads a YAML configuration from a file.
     * @param {string} filePath The filename to load.
     * @returns {ConfigData} The configuration object from the file.
     * @throws {Error} If the file cannot be read.
     * @private
     */
    function loadYAMLConfigFile(filePath) {
        debug(`Loading YAML config file: ${filePath}`);
        // lazy load YAML to improve performance when not used
        const yaml = require(283) /* js-yaml */;
        try {
            // empty YAML file can be null, so always use
            return yaml.safeLoad(readFile(filePath)) || {};
        }
        catch (e) {
            debug(`Error reading YAML file: ${filePath}`);
            e.message = `Cannot read config file: ${filePath}\nError: ${e.message}`;
            throw e;
        }
    }
    /**
     * Loads a JSON configuration from a file.
     * @param {string} filePath The filename to load.
     * @returns {ConfigData} The configuration object from the file.
     * @throws {Error} If the file cannot be read.
     * @private
     */
    function loadJSONConfigFile(filePath) {
        debug(`Loading JSON config file: ${filePath}`);
        try {
            return JSON.parse(stripComments(readFile(filePath)));
        }
        catch (e) {
            debug(`Error reading JSON file: ${filePath}`);
            e.message = `Cannot read config file: ${filePath}\nError: ${e.message}`;
            e.messageTemplate = "failed-to-read-json";
            e.messageData = {
                path: filePath,
                message: e.message
            };
            throw e;
        }
    }
    /**
     * Loads a legacy (.eslintrc) configuration from a file.
     * @param {string} filePath The filename to load.
     * @returns {ConfigData} The configuration object from the file.
     * @throws {Error} If the file cannot be read.
     * @private
     */
    function loadLegacyConfigFile(filePath) {
        debug(`Loading legacy config file: ${filePath}`);
        // lazy load YAML to improve performance when not used
        const yaml = require(283) /* js-yaml */;
        try {
            return yaml.safeLoad(stripComments(readFile(filePath))) || /* istanbul ignore next */ {};
        }
        catch (e) {
            debug("Error reading YAML file: %s\n%o", filePath, e);
            e.message = `Cannot read config file: ${filePath}\nError: ${e.message}`;
            throw e;
        }
    }
    /**
     * Loads a JavaScript configuration from a file.
     * @param {string} filePath The filename to load.
     * @returns {ConfigData} The configuration object from the file.
     * @throws {Error} If the file cannot be read.
     * @private
     */
    function loadJSConfigFile(filePath) {
        debug(`Loading JS config file: ${filePath}`);
        try {
            return importFresh(filePath);
        }
        catch (e) {
            debug(`Error reading JavaScript file: ${filePath}`);
            e.message = `Cannot read config file: ${filePath}\nError: ${e.message}`;
            throw e;
        }
    }
    /**
     * Loads a configuration from a package.json file.
     * @param {string} filePath The filename to load.
     * @returns {ConfigData} The configuration object from the file.
     * @throws {Error} If the file cannot be read.
     * @private
     */
    function loadPackageJSONConfigFile(filePath) {
        debug(`Loading package.json config file: ${filePath}`);
        try {
            const packageData = loadJSONConfigFile(filePath);
            if (!Object.hasOwnProperty.call(packageData, "eslintConfig")) {
                throw Object.assign(new Error("package.json file doesn't have 'eslintConfig' field."), { code: "ESLINT_CONFIG_FIELD_NOT_FOUND" });
            }
            return packageData.eslintConfig;
        }
        catch (e) {
            debug(`Error reading package.json file: ${filePath}`);
            e.message = `Cannot read config file: ${filePath}\nError: ${e.message}`;
            throw e;
        }
    }
    /**
     * Loads a `.eslintignore` from a file.
     * @param {string} filePath The filename to load.
     * @returns {string[]} The ignore patterns from the file.
     * @private
     */
    function loadESLintIgnoreFile(filePath) {
        debug(`Loading .eslintignore file: ${filePath}`);
        try {
            return readFile(filePath)
                .split(/\r?\n/gu)
                .filter(line => line.trim() !== "" && !line.startsWith("#"));
        }
        catch (e) {
            debug(`Error reading .eslintignore file: ${filePath}`);
            e.message = `Cannot read .eslintignore file: ${filePath}\nError: ${e.message}`;
            throw e;
        }
    }
    /**
     * Creates an error to notify about a missing config to extend from.
     * @param {string} configName The name of the missing config.
     * @param {string} importerName The name of the config that imported the missing config
     * @param {string} messageTemplate The text template to source error strings from.
     * @returns {Error} The error object to throw
     * @private
     */
    function configInvalidError(configName, importerName, messageTemplate) {
        return Object.assign(new Error(`Failed to load config "${configName}" to extend from.`), {
            messageTemplate,
            messageData: { configName, importerName }
        });
    }
    /**
     * Loads a configuration file regardless of the source. Inspects the file path
     * to determine the correctly way to load the config file.
     * @param {string} filePath The path to the configuration.
     * @returns {ConfigData|null} The configuration information.
     * @private
     */
    function loadConfigFile(filePath) {
        switch (path.extname(filePath)) {
            case ".js":
            case ".cjs":
                return loadJSConfigFile(filePath);
            case ".json":
                if (path.basename(filePath) === "package.json") {
                    return loadPackageJSONConfigFile(filePath);
                }
                return loadJSONConfigFile(filePath);
            case ".yaml":
            case ".yml":
                return loadYAMLConfigFile(filePath);
            default:
                return loadLegacyConfigFile(filePath);
        }
    }
    /**
     * Write debug log.
     * @param {string} request The requested module name.
     * @param {string} relativeTo The file path to resolve the request relative to.
     * @param {string} filePath The resolved file path.
     * @returns {void}
     */
    function writeDebugLogForLoading(request, relativeTo, filePath) {
        /* istanbul ignore next */
        if (debug.enabled) {
            let nameAndVersion = null;
            try {
                const packageJsonPath = ModuleResolver.resolve(`${request}/package.json`, relativeTo);
                const { version = "unknown" } = require(packageJsonPath);
                nameAndVersion = `${request}@${version}`;
            }
            catch (error) {
                debug("package.json was not found:", error.message);
                nameAndVersion = request;
            }
            debug("Loaded: %s (%s)", nameAndVersion, filePath);
        }
    }
    /**
     * Create a new context with default values.
     * @param {ConfigArrayFactoryInternalSlots} slots The internal slots.
     * @param {"config" | "ignore" | "implicit-processor" | undefined} providedType The type of the current configuration. Default is `"config"`.
     * @param {string | undefined} providedName The name of the current configuration. Default is the relative path from `cwd` to `filePath`.
     * @param {string | undefined} providedFilePath The path to the current configuration. Default is empty string.
     * @param {string | undefined} providedMatchBasePath The type of the current configuration. Default is the directory of `filePath` or `cwd`.
     * @returns {ConfigArrayFactoryLoadingContext} The created context.
     */
    function createContext({ cwd, resolvePluginsRelativeTo }, providedType, providedName, providedFilePath, providedMatchBasePath) {
        const filePath = providedFilePath
            ? path.resolve(cwd, providedFilePath)
            : "";
        const matchBasePath = (providedMatchBasePath && path.resolve(cwd, providedMatchBasePath)) ||
            (filePath && path.dirname(filePath)) ||
            cwd;
        const name = providedName ||
            (filePath && path.relative(cwd, filePath)) ||
            "";
        const pluginBasePath = resolvePluginsRelativeTo ||
            (filePath && path.dirname(filePath)) ||
            cwd;
        const type = providedType || "config";
        return { filePath, matchBasePath, name, pluginBasePath, type };
    }
    /**
     * Normalize a given plugin.
     * - Ensure the object to have four properties: configs, environments, processors, and rules.
     * - Ensure the object to not have other properties.
     * @param {Plugin} plugin The plugin to normalize.
     * @returns {Plugin} The normalized plugin.
     */
    function normalizePlugin(plugin) {
        return {
            configs: plugin.configs || {},
            environments: plugin.environments || {},
            processors: plugin.processors || {},
            rules: plugin.rules || {}
        };
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * The factory of `ConfigArray` objects.
     */
    class ConfigArrayFactory {
        /**
         * Initialize this instance.
         * @param {ConfigArrayFactoryOptions} [options] The map for additional plugins.
         */
        constructor({ additionalPluginPool = new Map(), cwd = process.cwd(), resolvePluginsRelativeTo, builtInRules, resolver = ModuleResolver, eslintAllPath, eslintRecommendedPath } = {}) {
            internalSlotsMap.set(this, {
                additionalPluginPool,
                cwd,
                resolvePluginsRelativeTo: resolvePluginsRelativeTo &&
                    path.resolve(cwd, resolvePluginsRelativeTo),
                builtInRules,
                resolver,
                eslintAllPath,
                eslintRecommendedPath
            });
        }
        /**
         * Create `ConfigArray` instance from a config data.
         * @param {ConfigData|null} configData The config data to create.
         * @param {Object} [options] The options.
         * @param {string} [options.basePath] The base path to resolve relative paths in `overrides[].files`, `overrides[].excludedFiles`, and `ignorePatterns`.
         * @param {string} [options.filePath] The path to this config data.
         * @param {string} [options.name] The config name.
         * @returns {ConfigArray} Loaded config.
         */
        create(configData, { basePath, filePath, name } = {}) {
            if (!configData) {
                return new ConfigArray();
            }
            const slots = internalSlotsMap.get(this);
            const ctx = createContext(slots, "config", name, filePath, basePath);
            const elements = this._normalizeConfigData(configData, ctx);
            return new ConfigArray(...elements);
        }
        /**
         * Load a config file.
         * @param {string} filePath The path to a config file.
         * @param {Object} [options] The options.
         * @param {string} [options.basePath] The base path to resolve relative paths in `overrides[].files`, `overrides[].excludedFiles`, and `ignorePatterns`.
         * @param {string} [options.name] The config name.
         * @returns {ConfigArray} Loaded config.
         */
        loadFile(filePath, { basePath, name } = {}) {
            const slots = internalSlotsMap.get(this);
            const ctx = createContext(slots, "config", name, filePath, basePath);
            return new ConfigArray(...this._loadConfigData(ctx));
        }
        /**
         * Load the config file on a given directory if exists.
         * @param {string} directoryPath The path to a directory.
         * @param {Object} [options] The options.
         * @param {string} [options.basePath] The base path to resolve relative paths in `overrides[].files`, `overrides[].excludedFiles`, and `ignorePatterns`.
         * @param {string} [options.name] The config name.
         * @returns {ConfigArray} Loaded config. An empty `ConfigArray` if any config doesn't exist.
         */
        loadInDirectory(directoryPath, { basePath, name } = {}) {
            const slots = internalSlotsMap.get(this);
            for (const filename of configFilenames) {
                const ctx = createContext(slots, "config", name, path.join(directoryPath, filename), basePath);
                if (fs.existsSync(ctx.filePath) && fs.statSync(ctx.filePath).isFile()) {
                    let configData;
                    try {
                        configData = loadConfigFile(ctx.filePath);
                    }
                    catch (error) {
                        if (!error || error.code !== "ESLINT_CONFIG_FIELD_NOT_FOUND") {
                            throw error;
                        }
                    }
                    if (configData) {
                        debug(`Config file found: ${ctx.filePath}`);
                        return new ConfigArray(...this._normalizeConfigData(configData, ctx));
                    }
                }
            }
            debug(`Config file not found on ${directoryPath}`);
            return new ConfigArray();
        }
        /**
         * Check if a config file on a given directory exists or not.
         * @param {string} directoryPath The path to a directory.
         * @returns {string | null} The path to the found config file. If not found then null.
         */
        static getPathToConfigFileInDirectory(directoryPath) {
            for (const filename of configFilenames) {
                const filePath = path.join(directoryPath, filename);
                if (fs.existsSync(filePath)) {
                    if (filename === "package.json") {
                        try {
                            loadPackageJSONConfigFile(filePath);
                            return filePath;
                        }
                        catch { /* ignore */ }
                    }
                    else {
                        return filePath;
                    }
                }
            }
            return null;
        }
        /**
         * Load `.eslintignore` file.
         * @param {string} filePath The path to a `.eslintignore` file to load.
         * @returns {ConfigArray} Loaded config. An empty `ConfigArray` if any config doesn't exist.
         */
        loadESLintIgnore(filePath) {
            const slots = internalSlotsMap.get(this);
            const ctx = createContext(slots, "ignore", void 0, filePath, slots.cwd);
            const ignorePatterns = loadESLintIgnoreFile(ctx.filePath);
            return new ConfigArray(...this._normalizeESLintIgnoreData(ignorePatterns, ctx));
        }
        /**
         * Load `.eslintignore` file in the current working directory.
         * @returns {ConfigArray} Loaded config. An empty `ConfigArray` if any config doesn't exist.
         */
        loadDefaultESLintIgnore() {
            const slots = internalSlotsMap.get(this);
            const eslintIgnorePath = path.resolve(slots.cwd, ".eslintignore");
            const packageJsonPath = path.resolve(slots.cwd, "package.json");
            if (fs.existsSync(eslintIgnorePath)) {
                return this.loadESLintIgnore(eslintIgnorePath);
            }
            if (fs.existsSync(packageJsonPath)) {
                const data = loadJSONConfigFile(packageJsonPath);
                if (Object.hasOwnProperty.call(data, "eslintIgnore")) {
                    if (!Array.isArray(data.eslintIgnore)) {
                        throw new Error("Package.json eslintIgnore property requires an array of paths");
                    }
                    const ctx = createContext(slots, "ignore", "eslintIgnore in package.json", packageJsonPath, slots.cwd);
                    return new ConfigArray(...this._normalizeESLintIgnoreData(data.eslintIgnore, ctx));
                }
            }
            return new ConfigArray();
        }
        /**
         * Load a given config file.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} Loaded config.
         * @private
         */
        _loadConfigData(ctx) {
            return this._normalizeConfigData(loadConfigFile(ctx.filePath), ctx);
        }
        /**
         * Normalize a given `.eslintignore` data to config array elements.
         * @param {string[]} ignorePatterns The patterns to ignore files.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} The normalized config.
         * @private
         */
        *_normalizeESLintIgnoreData(ignorePatterns, ctx) {
            const elements = this._normalizeObjectConfigData({ ignorePatterns }, ctx);
            // Set `ignorePattern.loose` flag for backward compatibility.
            for (const element of elements) {
                if (element.ignorePattern) {
                    element.ignorePattern.loose = true;
                }
                yield element;
            }
        }
        /**
         * Normalize a given config to an array.
         * @param {ConfigData} configData The config data to normalize.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} The normalized config.
         * @private
         */
        _normalizeConfigData(configData, ctx) {
            const validator = new ConfigValidator();
            validator.validateConfigSchema(configData, ctx.name || ctx.filePath);
            return this._normalizeObjectConfigData(configData, ctx);
        }
        /**
         * Normalize a given config to an array.
         * @param {ConfigData|OverrideConfigData} configData The config data to normalize.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} The normalized config.
         * @private
         */
        *_normalizeObjectConfigData(configData, ctx) {
            const { files, excludedFiles, ...configBody } = configData;
            const criteria = OverrideTester.create(files, excludedFiles, ctx.matchBasePath);
            const elements = this._normalizeObjectConfigDataBody(configBody, ctx);
            // Apply the criteria to every element.
            for (const element of elements) {
                /*
                 * Merge the criteria.
                 * This is for the `overrides` entries that came from the
                 * configurations of `overrides[].extends`.
                 */
                element.criteria = OverrideTester.and(criteria, element.criteria);
                /*
                 * Remove `root` property to ignore `root` settings which came from
                 * `extends` in `overrides`.
                 */
                if (element.criteria) {
                    element.root = void 0;
                }
                yield element;
            }
        }
        /**
         * Normalize a given config to an array.
         * @param {ConfigData} configData The config data to normalize.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} The normalized config.
         * @private
         */
        *_normalizeObjectConfigDataBody({ env, extends: extend, globals, ignorePatterns, noInlineConfig, parser: parserName, parserOptions, plugins: pluginList, processor, reportUnusedDisableDirectives, root, rules, settings, overrides: overrideList = [] }, ctx) {
            const extendList = Array.isArray(extend) ? extend : [extend];
            const ignorePattern = ignorePatterns && new IgnorePattern(Array.isArray(ignorePatterns) ? ignorePatterns : [ignorePatterns], ctx.matchBasePath);
            // Flatten `extends`.
            for (const extendName of extendList.filter(Boolean)) {
                yield* this._loadExtends(extendName, ctx);
            }
            // Load parser & plugins.
            const parser = parserName && this._loadParser(parserName, ctx);
            const plugins = pluginList && this._loadPlugins(pluginList, ctx);
            // Yield pseudo config data for file extension processors.
            if (plugins) {
                yield* this._takeFileExtensionProcessors(plugins, ctx);
            }
            // Yield the config data except `extends` and `overrides`.
            yield {
                // Debug information.
                type: ctx.type,
                name: ctx.name,
                filePath: ctx.filePath,
                // Config data.
                criteria: null,
                env,
                globals,
                ignorePattern,
                noInlineConfig,
                parser,
                parserOptions,
                plugins,
                processor,
                reportUnusedDisableDirectives,
                root,
                rules,
                settings
            };
            // Flatten `overries`.
            for (let i = 0; i < overrideList.length; ++i) {
                yield* this._normalizeObjectConfigData(overrideList[i], { ...ctx, name: `${ctx.name}#overrides[${i}]` });
            }
        }
        /**
         * Load configs of an element in `extends`.
         * @param {string} extendName The name of a base config.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} The normalized config.
         * @private
         */
        _loadExtends(extendName, ctx) {
            debug("Loading {extends:%j} relative to %s", extendName, ctx.filePath);
            try {
                if (extendName.startsWith("eslint:")) {
                    return this._loadExtendedBuiltInConfig(extendName, ctx);
                }
                if (extendName.startsWith("plugin:")) {
                    return this._loadExtendedPluginConfig(extendName, ctx);
                }
                return this._loadExtendedShareableConfig(extendName, ctx);
            }
            catch (error) {
                error.message += `\nReferenced from: ${ctx.filePath || ctx.name}`;
                throw error;
            }
        }
        /**
         * Load configs of an element in `extends`.
         * @param {string} extendName The name of a base config.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} The normalized config.
         * @private
         */
        _loadExtendedBuiltInConfig(extendName, ctx) {
            const { eslintAllPath, eslintRecommendedPath } = internalSlotsMap.get(this);
            if (extendName === "eslint:recommended") {
                return this._loadConfigData({
                    ...ctx,
                    filePath: eslintRecommendedPath,
                    name: `${ctx.name} » ${extendName}`
                });
            }
            if (extendName === "eslint:all") {
                return this._loadConfigData({
                    ...ctx,
                    filePath: eslintAllPath,
                    name: `${ctx.name} » ${extendName}`
                });
            }
            throw configInvalidError(extendName, ctx.name, "extend-config-missing");
        }
        /**
         * Load configs of an element in `extends`.
         * @param {string} extendName The name of a base config.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} The normalized config.
         * @private
         */
        _loadExtendedPluginConfig(extendName, ctx) {
            const slashIndex = extendName.lastIndexOf("/");
            if (slashIndex === -1) {
                throw configInvalidError(extendName, ctx.filePath, "plugin-invalid");
            }
            const pluginName = extendName.slice("plugin:".length, slashIndex);
            const configName = extendName.slice(slashIndex + 1);
            if (isFilePath(pluginName)) {
                throw new Error("'extends' cannot use a file path for plugins.");
            }
            const plugin = this._loadPlugin(pluginName, ctx);
            const configData = plugin.definition &&
                plugin.definition.configs[configName];
            if (configData) {
                return this._normalizeConfigData(configData, {
                    ...ctx,
                    filePath: plugin.filePath || ctx.filePath,
                    name: `${ctx.name} » plugin:${plugin.id}/${configName}`
                });
            }
            throw plugin.error || configInvalidError(extendName, ctx.filePath, "extend-config-missing");
        }
        /**
         * Load configs of an element in `extends`.
         * @param {string} extendName The name of a base config.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} The normalized config.
         * @private
         */
        _loadExtendedShareableConfig(extendName, ctx) {
            const { cwd, resolver } = internalSlotsMap.get(this);
            const relativeTo = ctx.filePath || path.join(cwd, "__placeholder__.js");
            let request;
            if (isFilePath(extendName)) {
                request = extendName;
            }
            else if (extendName.startsWith(".")) {
                request = `./${extendName}`; // For backward compatibility. A ton of tests depended on this behavior.
            }
            else {
                request = naming.normalizePackageName(extendName, "eslint-config");
            }
            let filePath;
            try {
                filePath = resolver.resolve(request, relativeTo);
            }
            catch (error) {
                /* istanbul ignore else */
                if (error && error.code === "MODULE_NOT_FOUND") {
                    throw configInvalidError(extendName, ctx.filePath, "extend-config-missing");
                }
                throw error;
            }
            writeDebugLogForLoading(request, relativeTo, filePath);
            return this._loadConfigData({
                ...ctx,
                filePath,
                name: `${ctx.name} » ${request}`
            });
        }
        /**
         * Load given plugins.
         * @param {string[]} names The plugin names to load.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {Record<string,DependentPlugin>} The loaded parser.
         * @private
         */
        _loadPlugins(names, ctx) {
            return names.reduce((map, name) => {
                if (isFilePath(name)) {
                    throw new Error("Plugins array cannot includes file paths.");
                }
                const plugin = this._loadPlugin(name, ctx);
                map[plugin.id] = plugin;
                return map;
            }, {});
        }
        /**
         * Load a given parser.
         * @param {string} nameOrPath The package name or the path to a parser file.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {DependentParser} The loaded parser.
         */
        _loadParser(nameOrPath, ctx) {
            debug("Loading parser %j from %s", nameOrPath, ctx.filePath);
            const { cwd } = internalSlotsMap.get(this);
            const relativeTo = ctx.filePath || path.join(cwd, "__placeholder__.js");
            try {
                const filePath = ModuleResolver.resolve(nameOrPath, relativeTo);
                writeDebugLogForLoading(nameOrPath, relativeTo, filePath);
                return new ConfigDependency({
                    definition: require(filePath),
                    filePath,
                    id: nameOrPath,
                    importerName: ctx.name,
                    importerPath: ctx.filePath
                });
            }
            catch (error) {
                // If the parser name is "espree", load the espree of ESLint.
                if (nameOrPath === "espree") {
                    debug("Fallback espree.");
                    return new ConfigDependency({
                        definition: require(313) /* espree */,
                        filePath: require.resolve("espree"),
                        id: nameOrPath,
                        importerName: ctx.name,
                        importerPath: ctx.filePath
                    });
                }
                debug("Failed to load parser '%s' declared in '%s'.", nameOrPath, ctx.name);
                error.message = `Failed to load parser '${nameOrPath}' declared in '${ctx.name}': ${error.message}`;
                return new ConfigDependency({
                    error,
                    id: nameOrPath,
                    importerName: ctx.name,
                    importerPath: ctx.filePath
                });
            }
        }
        /**
         * Load a given plugin.
         * @param {string} name The plugin name to load.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {DependentPlugin} The loaded plugin.
         * @private
         */
        _loadPlugin(name, ctx) {
            debug("Loading plugin %j from %s", name, ctx.filePath);
            const { additionalPluginPool } = internalSlotsMap.get(this);
            const request = naming.normalizePackageName(name, "eslint-plugin");
            const id = naming.getShorthandName(request, "eslint-plugin");
            const relativeTo = path.join(ctx.pluginBasePath, "__placeholder__.js");
            if (name.match(/\s+/u)) {
                const error = Object.assign(new Error(`Whitespace found in plugin name '${name}'`), {
                    messageTemplate: "whitespace-found",
                    messageData: { pluginName: request }
                });
                return new ConfigDependency({
                    error,
                    id,
                    importerName: ctx.name,
                    importerPath: ctx.filePath
                });
            }
            // Check for additional pool.
            const plugin = additionalPluginPool.get(request) ||
                additionalPluginPool.get(id);
            if (plugin) {
                return new ConfigDependency({
                    definition: normalizePlugin(plugin),
                    filePath: "",
                    id,
                    importerName: ctx.name,
                    importerPath: ctx.filePath
                });
            }
            let filePath;
            let error;
            try {
                filePath = ModuleResolver.resolve(request, relativeTo);
            }
            catch (resolveError) {
                error = resolveError;
                /* istanbul ignore else */
                if (error && error.code === "MODULE_NOT_FOUND") {
                    error.messageTemplate = "plugin-missing";
                    error.messageData = {
                        pluginName: request,
                        resolvePluginsRelativeTo: ctx.pluginBasePath,
                        importerName: ctx.name
                    };
                }
            }
            if (filePath) {
                try {
                    writeDebugLogForLoading(request, relativeTo, filePath);
                    const startTime = Date.now();
                    const pluginDefinition = require(filePath);
                    debug(`Plugin ${filePath} loaded in: ${Date.now() - startTime}ms`);
                    return new ConfigDependency({
                        definition: normalizePlugin(pluginDefinition),
                        filePath,
                        id,
                        importerName: ctx.name,
                        importerPath: ctx.filePath
                    });
                }
                catch (loadError) {
                    error = loadError;
                }
            }
            debug("Failed to load plugin '%s' declared in '%s'.", name, ctx.name);
            error.message = `Failed to load plugin '${name}' declared in '${ctx.name}': ${error.message}`;
            return new ConfigDependency({
                error,
                id,
                importerName: ctx.name,
                importerPath: ctx.filePath
            });
        }
        /**
         * Take file expression processors as config array elements.
         * @param {Record<string,DependentPlugin>} plugins The plugin definitions.
         * @param {ConfigArrayFactoryLoadingContext} ctx The loading context.
         * @returns {IterableIterator<ConfigArrayElement>} The config array elements of file expression processors.
         * @private
         */
        *_takeFileExtensionProcessors(plugins, ctx) {
            for (const pluginId of Object.keys(plugins)) {
                const processors = plugins[pluginId] &&
                    plugins[pluginId].definition &&
                    plugins[pluginId].definition.processors;
                if (!processors) {
                    continue;
                }
                for (const processorId of Object.keys(processors)) {
                    if (processorId.startsWith(".")) {
                        yield* this._normalizeObjectConfigData({
                            files: [`*${processorId}`],
                            processor: `${pluginId}/${processorId}`
                        }, {
                            ...ctx,
                            type: "implicit-processor",
                            name: `${ctx.name}#processors["${pluginId}/${processorId}"]`
                        });
                    }
                }
            }
        }
    }
    module.exports = { ConfigArrayFactory, createContext };
},
/* import-fresh/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const path = require('path');
    const resolveFrom = require(211) /* resolve-from */;
    const parentModule = require(212) /* parent-module */;
    module.exports = moduleId => {
        if (typeof moduleId !== 'string') {
            throw new TypeError('Expected a string');
        }
        const parentPath = parentModule(__filename);
        const cwd = parentPath ? path.dirname(parentPath) : __dirname;
        const filePath = resolveFrom(cwd, moduleId);
        const oldModule = require.cache[filePath];
        // Delete itself from module parent
        if (oldModule && oldModule.parent) {
            let i = oldModule.parent.children.length;
            while (i--) {
                if (oldModule.parent.children[i].id === filePath) {
                    oldModule.parent.children.splice(i, 1);
                }
            }
        }
        delete require.cache[filePath]; // Delete module from cache
        const parent = require.cache[parentPath]; // If `filePath` and `parentPath` are the same, cache will already be deleted so we won't get a memory leak in next step
        return parent === undefined ? require(filePath) : parent.require(filePath); // In case cache doesn't have parent, fall back to normal require
    };
},
/* resolve-from/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const path = require('path');
    const Module = require('module');
    const fs = require('fs');
    const resolveFrom = (fromDir, moduleId, silent) => {
        if (typeof fromDir !== 'string') {
            throw new TypeError(`Expected \`fromDir\` to be of type \`string\`, got \`${typeof fromDir}\``);
        }
        if (typeof moduleId !== 'string') {
            throw new TypeError(`Expected \`moduleId\` to be of type \`string\`, got \`${typeof moduleId}\``);
        }
        try {
            fromDir = fs.realpathSync(fromDir);
        }
        catch (err) {
            if (err.code === 'ENOENT') {
                fromDir = path.resolve(fromDir);
            }
            else if (silent) {
                return null;
            }
            else {
                throw err;
            }
        }
        const fromFile = path.join(fromDir, 'noop.js');
        const resolveFileName = () => Module._resolveFilename(moduleId, {
            id: fromFile,
            filename: fromFile,
            paths: Module._nodeModulePaths(fromDir)
        });
        if (silent) {
            try {
                return resolveFileName();
            }
            catch (err) {
                return null;
            }
        }
        return resolveFileName();
    };
    module.exports = (fromDir, moduleId) => resolveFrom(fromDir, moduleId);
    module.exports.silent = (fromDir, moduleId) => resolveFrom(fromDir, moduleId, true);
},
/* parent-module/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const callsites = require(213) /* callsites */;
    module.exports = filepath => {
        const stacks = callsites();
        if (!filepath) {
            return stacks[2].getFileName();
        }
        let seenVal = false;
        // Skip the first stack as it's this function
        stacks.shift();
        for (const stack of stacks) {
            const parentFilepath = stack.getFileName();
            if (typeof parentFilepath !== 'string') {
                continue;
            }
            if (parentFilepath === filepath) {
                seenVal = true;
                continue;
            }
            // Skip native modules
            if (parentFilepath === 'module.js') {
                continue;
            }
            if (seenVal && parentFilepath !== filepath) {
                return parentFilepath;
            }
        }
    };
},
/* callsites/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const callsites = () => {
        const _prepareStackTrace = Error.prepareStackTrace;
        Error.prepareStackTrace = (_, stack) => stack;
        const stack = new Error().stack.slice(1);
        Error.prepareStackTrace = _prepareStackTrace;
        return stack;
    };
    module.exports = callsites;
    // TODO: Remove this for the next major release
    module.exports.default = callsites;
},
/* strip-json-comments/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const singleComment = Symbol('singleComment');
    const multiComment = Symbol('multiComment');
    const stripWithoutWhitespace = () => '';
    const stripWithWhitespace = (string, start, end) => string.slice(start, end).replace(/\S/g, ' ');
    const isEscaped = (jsonString, quotePosition) => {
        let index = quotePosition - 1;
        let backslashCount = 0;
        while (jsonString[index] === '\\') {
            index -= 1;
            backslashCount += 1;
        }
        return Boolean(backslashCount % 2);
    };
    module.exports = (jsonString, options = {}) => {
        if (typeof jsonString !== 'string') {
            throw new TypeError(`Expected argument \`jsonString\` to be a \`string\`, got \`${typeof jsonString}\``);
        }
        const strip = options.whitespace === false ? stripWithoutWhitespace : stripWithWhitespace;
        let insideString = false;
        let insideComment = false;
        let offset = 0;
        let result = '';
        for (let i = 0; i < jsonString.length; i++) {
            const currentCharacter = jsonString[i];
            const nextCharacter = jsonString[i + 1];
            if (!insideComment && currentCharacter === '"') {
                const escaped = isEscaped(jsonString, i);
                if (!escaped) {
                    insideString = !insideString;
                }
            }
            if (insideString) {
                continue;
            }
            if (!insideComment && currentCharacter + nextCharacter === '//') {
                result += jsonString.slice(offset, i);
                offset = i;
                insideComment = singleComment;
                i++;
            }
            else if (insideComment === singleComment && currentCharacter + nextCharacter === '\r\n') {
                i++;
                insideComment = false;
                result += strip(jsonString, offset, i);
                offset = i;
                continue;
            }
            else if (insideComment === singleComment && currentCharacter === '\n') {
                insideComment = false;
                result += strip(jsonString, offset, i);
                offset = i;
            }
            else if (!insideComment && currentCharacter + nextCharacter === '/*') {
                result += jsonString.slice(offset, i);
                offset = i;
                insideComment = multiComment;
                i++;
                continue;
            }
            else if (insideComment === multiComment && currentCharacter + nextCharacter === '*/') {
                i++;
                insideComment = false;
                result += strip(jsonString, offset, i + 1);
                offset = i + 1;
                continue;
            }
        }
        return result + (insideComment ? strip(jsonString.slice(offset)) : jsonString.slice(offset));
    };
},
/* @eslint/eslintrc/lib/shared/config-validator.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint class-methods-use-this: "off" */
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const util = require("util"), configSchema = require(216) /* ../../conf/config-schema */, BuiltInEnvironments = require(217) /* ../../conf/environments */, ConfigOps = require(220) /* ./config-ops */, { emitDeprecationWarning } = require(221) /* ./deprecation-warnings */;
    const ajv = require(222) /* ./ajv */();
    const ruleValidators = new WeakMap();
    const noop = Function.prototype;
    //------------------------------------------------------------------------------
    // Private
    //------------------------------------------------------------------------------
    let validateSchema;
    const severityMap = {
        error: 2,
        warn: 1,
        off: 0
    };
    const validated = new WeakSet();
    //-----------------------------------------------------------------------------
    // Exports
    //-----------------------------------------------------------------------------
    module.exports = class ConfigValidator {
        constructor({ builtInRules = new Map() } = {}) {
            this.builtInRules = builtInRules;
        }
        /**
         * Gets a complete options schema for a rule.
         * @param {{create: Function, schema: (Array|null)}} rule A new-style rule object
         * @returns {Object} JSON Schema for the rule's options.
         */
        getRuleOptionsSchema(rule) {
            if (!rule) {
                return null;
            }
            const schema = rule.schema || rule.meta && rule.meta.schema;
            // Given a tuple of schemas, insert warning level at the beginning
            if (Array.isArray(schema)) {
                if (schema.length) {
                    return {
                        type: "array",
                        items: schema,
                        minItems: 0,
                        maxItems: schema.length
                    };
                }
                return {
                    type: "array",
                    minItems: 0,
                    maxItems: 0
                };
            }
            // Given a full schema, leave it alone
            return schema || null;
        }
        /**
         * Validates a rule's severity and returns the severity value. Throws an error if the severity is invalid.
         * @param {options} options The given options for the rule.
         * @returns {number|string} The rule's severity value
         */
        validateRuleSeverity(options) {
            const severity = Array.isArray(options) ? options[0] : options;
            const normSeverity = typeof severity === "string" ? severityMap[severity.toLowerCase()] : severity;
            if (normSeverity === 0 || normSeverity === 1 || normSeverity === 2) {
                return normSeverity;
            }
            throw new Error(`\tSeverity should be one of the following: 0 = off, 1 = warn, 2 = error (you passed '${util.inspect(severity).replace(/'/gu, "\"").replace(/\n/gu, "")}').\n`);
        }
        /**
         * Validates the non-severity options passed to a rule, based on its schema.
         * @param {{create: Function}} rule The rule to validate
         * @param {Array} localOptions The options for the rule, excluding severity
         * @returns {void}
         */
        validateRuleSchema(rule, localOptions) {
            if (!ruleValidators.has(rule)) {
                const schema = this.getRuleOptionsSchema(rule);
                if (schema) {
                    ruleValidators.set(rule, ajv.compile(schema));
                }
            }
            const validateRule = ruleValidators.get(rule);
            if (validateRule) {
                validateRule(localOptions);
                if (validateRule.errors) {
                    throw new Error(validateRule.errors.map(error => `\tValue ${JSON.stringify(error.data)} ${error.message}.\n`).join(""));
                }
            }
        }
        /**
         * Validates a rule's options against its schema.
         * @param {{create: Function}|null} rule The rule that the config is being validated for
         * @param {string} ruleId The rule's unique name.
         * @param {Array|number} options The given options for the rule.
         * @param {string|null} source The name of the configuration source to report in any errors. If null or undefined,
         * no source is prepended to the message.
         * @returns {void}
         */
        validateRuleOptions(rule, ruleId, options, source = null) {
            try {
                const severity = this.validateRuleSeverity(options);
                if (severity !== 0) {
                    this.validateRuleSchema(rule, Array.isArray(options) ? options.slice(1) : []);
                }
            }
            catch (err) {
                const enhancedMessage = `Configuration for rule "${ruleId}" is invalid:\n${err.message}`;
                if (typeof source === "string") {
                    throw new Error(`${source}:\n\t${enhancedMessage}`);
                }
                else {
                    throw new Error(enhancedMessage);
                }
            }
        }
        /**
         * Validates an environment object
         * @param {Object} environment The environment config object to validate.
         * @param {string} source The name of the configuration source to report in any errors.
         * @param {function(envId:string): Object} [getAdditionalEnv] A map from strings to loaded environments.
         * @returns {void}
         */
        validateEnvironment(environment, source, getAdditionalEnv = noop) {
            // not having an environment is ok
            if (!environment) {
                return;
            }
            Object.keys(environment).forEach(id => {
                const env = getAdditionalEnv(id) || BuiltInEnvironments.get(id) || null;
                if (!env) {
                    const message = `${source}:\n\tEnvironment key "${id}" is unknown\n`;
                    throw new Error(message);
                }
            });
        }
        /**
         * Validates a rules config object
         * @param {Object} rulesConfig The rules config object to validate.
         * @param {string} source The name of the configuration source to report in any errors.
         * @param {function(ruleId:string): Object} getAdditionalRule A map from strings to loaded rules
         * @returns {void}
         */
        validateRules(rulesConfig, source, getAdditionalRule = noop) {
            if (!rulesConfig) {
                return;
            }
            Object.keys(rulesConfig).forEach(id => {
                const rule = getAdditionalRule(id) || this.builtInRules.get(id) || null;
                this.validateRuleOptions(rule, id, rulesConfig[id], source);
            });
        }
        /**
         * Validates a `globals` section of a config file
         * @param {Object} globalsConfig The `globals` section
         * @param {string|null} source The name of the configuration source to report in the event of an error.
         * @returns {void}
         */
        validateGlobals(globalsConfig, source = null) {
            if (!globalsConfig) {
                return;
            }
            Object.entries(globalsConfig)
                .forEach(([configuredGlobal, configuredValue]) => {
                try {
                    ConfigOps.normalizeConfigGlobal(configuredValue);
                }
                catch (err) {
                    throw new Error(`ESLint configuration of global '${configuredGlobal}' in ${source} is invalid:\n${err.message}`);
                }
            });
        }
        /**
         * Validate `processor` configuration.
         * @param {string|undefined} processorName The processor name.
         * @param {string} source The name of config file.
         * @param {function(id:string): Processor} getProcessor The getter of defined processors.
         * @returns {void}
         */
        validateProcessor(processorName, source, getProcessor) {
            if (processorName && !getProcessor(processorName)) {
                throw new Error(`ESLint configuration of processor in '${source}' is invalid: '${processorName}' was not found.`);
            }
        }
        /**
         * Formats an array of schema validation errors.
         * @param {Array} errors An array of error messages to format.
         * @returns {string} Formatted error message
         */
        formatErrors(errors) {
            return errors.map(error => {
                if (error.keyword === "additionalProperties") {
                    const formattedPropertyPath = error.dataPath.length ? `${error.dataPath.slice(1)}.${error.params.additionalProperty}` : error.params.additionalProperty;
                    return `Unexpected top-level property "${formattedPropertyPath}"`;
                }
                if (error.keyword === "type") {
                    const formattedField = error.dataPath.slice(1);
                    const formattedExpectedType = Array.isArray(error.schema) ? error.schema.join("/") : error.schema;
                    const formattedValue = JSON.stringify(error.data);
                    return `Property "${formattedField}" is the wrong type (expected ${formattedExpectedType} but got \`${formattedValue}\`)`;
                }
                const field = error.dataPath[0] === "." ? error.dataPath.slice(1) : error.dataPath;
                return `"${field}" ${error.message}. Value: ${JSON.stringify(error.data)}`;
            }).map(message => `\t- ${message}.\n`).join("");
        }
        /**
         * Validates the top level properties of the config object.
         * @param {Object} config The config object to validate.
         * @param {string} source The name of the configuration source to report in any errors.
         * @returns {void}
         */
        validateConfigSchema(config, source = null) {
            validateSchema = validateSchema || ajv.compile(configSchema);
            if (!validateSchema(config)) {
                throw new Error(`ESLint configuration in ${source} is invalid:\n${this.formatErrors(validateSchema.errors)}`);
            }
            if (Object.hasOwnProperty.call(config, "ecmaFeatures")) {
                emitDeprecationWarning(source, "ESLINT_LEGACY_ECMAFEATURES");
            }
        }
        /**
         * Validates an entire config object.
         * @param {Object} config The config object to validate.
         * @param {string} source The name of the configuration source to report in any errors.
         * @param {function(ruleId:string): Object} [getAdditionalRule] A map from strings to loaded rules.
         * @param {function(envId:string): Object} [getAdditionalEnv] A map from strings to loaded envs.
         * @returns {void}
         */
        validate(config, source, getAdditionalRule, getAdditionalEnv) {
            this.validateConfigSchema(config, source);
            this.validateRules(config.rules, source, getAdditionalRule);
            this.validateEnvironment(config.env, source, getAdditionalEnv);
            this.validateGlobals(config.globals, source);
            for (const override of config.overrides || []) {
                this.validateRules(override.rules, source, getAdditionalRule);
                this.validateEnvironment(override.env, source, getAdditionalEnv);
                this.validateGlobals(config.globals, source);
            }
        }
        /**
         * Validate config array object.
         * @param {ConfigArray} configArray The config array to validate.
         * @returns {void}
         */
        validateConfigArray(configArray) {
            const getPluginEnv = Map.prototype.get.bind(configArray.pluginEnvironments);
            const getPluginProcessor = Map.prototype.get.bind(configArray.pluginProcessors);
            const getPluginRule = Map.prototype.get.bind(configArray.pluginRules);
            // Validate.
            for (const element of configArray) {
                if (validated.has(element)) {
                    continue;
                }
                validated.add(element);
                this.validateEnvironment(element.env, element.name, getPluginEnv);
                this.validateGlobals(element.globals, element.name);
                this.validateProcessor(element.processor, element.name, getPluginProcessor);
                this.validateRules(element.rules, element.name, getPluginRule);
            }
        }
    };
},
/* @eslint/eslintrc/conf/config-schema.js */ function _(require, module, exports, __esModule, __esExport) {
    const baseConfigProperties = {
        $schema: { type: "string" },
        env: { type: "object" },
        extends: { $ref: "#/definitions/stringOrStrings" },
        globals: { type: "object" },
        overrides: {
            type: "array",
            items: { $ref: "#/definitions/overrideConfig" },
            additionalItems: false
        },
        parser: { type: ["string", "null"] },
        parserOptions: { type: "object" },
        plugins: { type: "array" },
        processor: { type: "string" },
        rules: { type: "object" },
        settings: { type: "object" },
        noInlineConfig: { type: "boolean" },
        reportUnusedDisableDirectives: { type: "boolean" },
        ecmaFeatures: { type: "object" } // deprecated; logs a warning when used
    };
    const configSchema = {
        definitions: {
            stringOrStrings: {
                oneOf: [
                    { type: "string" },
                    {
                        type: "array",
                        items: { type: "string" },
                        additionalItems: false
                    }
                ]
            },
            stringOrStringsRequired: {
                oneOf: [
                    { type: "string" },
                    {
                        type: "array",
                        items: { type: "string" },
                        additionalItems: false,
                        minItems: 1
                    }
                ]
            },
            // Config at top-level.
            objectConfig: {
                type: "object",
                properties: {
                    root: { type: "boolean" },
                    ignorePatterns: { $ref: "#/definitions/stringOrStrings" },
                    ...baseConfigProperties
                },
                additionalProperties: false
            },
            // Config in `overrides`.
            overrideConfig: {
                type: "object",
                properties: {
                    excludedFiles: { $ref: "#/definitions/stringOrStrings" },
                    files: { $ref: "#/definitions/stringOrStringsRequired" },
                    ...baseConfigProperties
                },
                required: ["files"],
                additionalProperties: false
            }
        },
        $ref: "#/definitions/objectConfig"
    };
    module.exports = configSchema;
},
/* @eslint/eslintrc/conf/environments.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const globals = require(218) /* globals */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Get the object that has difference.
     * @param {Record<string,boolean>} current The newer object.
     * @param {Record<string,boolean>} prev The older object.
     * @returns {Record<string,boolean>} The difference object.
     */
    function getDiff(current, prev) {
        const retv = {};
        for (const [key, value] of Object.entries(current)) {
            if (!Object.hasOwnProperty.call(prev, key)) {
                retv[key] = value;
            }
        }
        return retv;
    }
    const newGlobals2015 = getDiff(globals.es2015, globals.es5); // 19 variables such as Promise, Map, ...
    const newGlobals2017 = {
        Atomics: false,
        SharedArrayBuffer: false
    };
    const newGlobals2020 = {
        BigInt: false,
        BigInt64Array: false,
        BigUint64Array: false,
        globalThis: false
    };
    const newGlobals2021 = {
        AggregateError: false,
        FinalizationRegistry: false,
        WeakRef: false
    };
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /** @type {Map<string, import("../lib/shared/types").Environment>} */
    module.exports = new Map(Object.entries({
        // Language
        builtin: {
            globals: globals.es5
        },
        es6: {
            globals: newGlobals2015,
            parserOptions: {
                ecmaVersion: 6
            }
        },
        es2015: {
            globals: newGlobals2015,
            parserOptions: {
                ecmaVersion: 6
            }
        },
        es2017: {
            globals: { ...newGlobals2015, ...newGlobals2017 },
            parserOptions: {
                ecmaVersion: 8
            }
        },
        es2020: {
            globals: { ...newGlobals2015, ...newGlobals2017, ...newGlobals2020 },
            parserOptions: {
                ecmaVersion: 11
            }
        },
        es2021: {
            globals: { ...newGlobals2015, ...newGlobals2017, ...newGlobals2020, ...newGlobals2021 },
            parserOptions: {
                ecmaVersion: 12
            }
        },
        // Platforms
        browser: {
            globals: globals.browser
        },
        node: {
            globals: globals.node,
            parserOptions: {
                ecmaFeatures: {
                    globalReturn: true
                }
            }
        },
        "shared-node-browser": {
            globals: globals["shared-node-browser"]
        },
        worker: {
            globals: globals.worker
        },
        serviceworker: {
            globals: globals.serviceworker
        },
        // Frameworks
        commonjs: {
            globals: globals.commonjs,
            parserOptions: {
                ecmaFeatures: {
                    globalReturn: true
                }
            }
        },
        amd: {
            globals: globals.amd
        },
        mocha: {
            globals: globals.mocha
        },
        jasmine: {
            globals: globals.jasmine
        },
        jest: {
            globals: globals.jest
        },
        phantomjs: {
            globals: globals.phantomjs
        },
        jquery: {
            globals: globals.jquery
        },
        qunit: {
            globals: globals.qunit
        },
        prototypejs: {
            globals: globals.prototypejs
        },
        shelljs: {
            globals: globals.shelljs
        },
        meteor: {
            globals: globals.meteor
        },
        mongo: {
            globals: globals.mongo
        },
        protractor: {
            globals: globals.protractor
        },
        applescript: {
            globals: globals.applescript
        },
        nashorn: {
            globals: globals.nashorn
        },
        atomtest: {
            globals: globals.atomtest
        },
        embertest: {
            globals: globals.embertest
        },
        webextensions: {
            globals: globals.webextensions
        },
        greasemonkey: {
            globals: globals.greasemonkey
        }
    }));
},
/* globals/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = require(219) /* ./globals.json */;
},
/* globals/globals.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "builtin": {
            "AggregateError": false,
            "Array": false,
            "ArrayBuffer": false,
            "Atomics": false,
            "BigInt": false,
            "BigInt64Array": false,
            "BigUint64Array": false,
            "Boolean": false,
            "constructor": false,
            "DataView": false,
            "Date": false,
            "decodeURI": false,
            "decodeURIComponent": false,
            "encodeURI": false,
            "encodeURIComponent": false,
            "Error": false,
            "escape": false,
            "eval": false,
            "EvalError": false,
            "FinalizationRegistry": false,
            "Float32Array": false,
            "Float64Array": false,
            "Function": false,
            "globalThis": false,
            "hasOwnProperty": false,
            "Infinity": false,
            "Int16Array": false,
            "Int32Array": false,
            "Int8Array": false,
            "isFinite": false,
            "isNaN": false,
            "isPrototypeOf": false,
            "JSON": false,
            "Map": false,
            "Math": false,
            "NaN": false,
            "Number": false,
            "Object": false,
            "parseFloat": false,
            "parseInt": false,
            "Promise": false,
            "propertyIsEnumerable": false,
            "Proxy": false,
            "RangeError": false,
            "ReferenceError": false,
            "Reflect": false,
            "RegExp": false,
            "Set": false,
            "SharedArrayBuffer": false,
            "String": false,
            "Symbol": false,
            "SyntaxError": false,
            "toLocaleString": false,
            "toString": false,
            "TypeError": false,
            "Uint16Array": false,
            "Uint32Array": false,
            "Uint8Array": false,
            "Uint8ClampedArray": false,
            "undefined": false,
            "unescape": false,
            "URIError": false,
            "valueOf": false,
            "WeakMap": false,
            "WeakRef": false,
            "WeakSet": false
        },
        "es5": {
            "Array": false,
            "Boolean": false,
            "constructor": false,
            "Date": false,
            "decodeURI": false,
            "decodeURIComponent": false,
            "encodeURI": false,
            "encodeURIComponent": false,
            "Error": false,
            "escape": false,
            "eval": false,
            "EvalError": false,
            "Function": false,
            "hasOwnProperty": false,
            "Infinity": false,
            "isFinite": false,
            "isNaN": false,
            "isPrototypeOf": false,
            "JSON": false,
            "Math": false,
            "NaN": false,
            "Number": false,
            "Object": false,
            "parseFloat": false,
            "parseInt": false,
            "propertyIsEnumerable": false,
            "RangeError": false,
            "ReferenceError": false,
            "RegExp": false,
            "String": false,
            "SyntaxError": false,
            "toLocaleString": false,
            "toString": false,
            "TypeError": false,
            "undefined": false,
            "unescape": false,
            "URIError": false,
            "valueOf": false
        },
        "es2015": {
            "Array": false,
            "ArrayBuffer": false,
            "Boolean": false,
            "constructor": false,
            "DataView": false,
            "Date": false,
            "decodeURI": false,
            "decodeURIComponent": false,
            "encodeURI": false,
            "encodeURIComponent": false,
            "Error": false,
            "escape": false,
            "eval": false,
            "EvalError": false,
            "Float32Array": false,
            "Float64Array": false,
            "Function": false,
            "hasOwnProperty": false,
            "Infinity": false,
            "Int16Array": false,
            "Int32Array": false,
            "Int8Array": false,
            "isFinite": false,
            "isNaN": false,
            "isPrototypeOf": false,
            "JSON": false,
            "Map": false,
            "Math": false,
            "NaN": false,
            "Number": false,
            "Object": false,
            "parseFloat": false,
            "parseInt": false,
            "Promise": false,
            "propertyIsEnumerable": false,
            "Proxy": false,
            "RangeError": false,
            "ReferenceError": false,
            "Reflect": false,
            "RegExp": false,
            "Set": false,
            "String": false,
            "Symbol": false,
            "SyntaxError": false,
            "toLocaleString": false,
            "toString": false,
            "TypeError": false,
            "Uint16Array": false,
            "Uint32Array": false,
            "Uint8Array": false,
            "Uint8ClampedArray": false,
            "undefined": false,
            "unescape": false,
            "URIError": false,
            "valueOf": false,
            "WeakMap": false,
            "WeakSet": false
        },
        "es2017": {
            "Array": false,
            "ArrayBuffer": false,
            "Atomics": false,
            "Boolean": false,
            "constructor": false,
            "DataView": false,
            "Date": false,
            "decodeURI": false,
            "decodeURIComponent": false,
            "encodeURI": false,
            "encodeURIComponent": false,
            "Error": false,
            "escape": false,
            "eval": false,
            "EvalError": false,
            "Float32Array": false,
            "Float64Array": false,
            "Function": false,
            "hasOwnProperty": false,
            "Infinity": false,
            "Int16Array": false,
            "Int32Array": false,
            "Int8Array": false,
            "isFinite": false,
            "isNaN": false,
            "isPrototypeOf": false,
            "JSON": false,
            "Map": false,
            "Math": false,
            "NaN": false,
            "Number": false,
            "Object": false,
            "parseFloat": false,
            "parseInt": false,
            "Promise": false,
            "propertyIsEnumerable": false,
            "Proxy": false,
            "RangeError": false,
            "ReferenceError": false,
            "Reflect": false,
            "RegExp": false,
            "Set": false,
            "SharedArrayBuffer": false,
            "String": false,
            "Symbol": false,
            "SyntaxError": false,
            "toLocaleString": false,
            "toString": false,
            "TypeError": false,
            "Uint16Array": false,
            "Uint32Array": false,
            "Uint8Array": false,
            "Uint8ClampedArray": false,
            "undefined": false,
            "unescape": false,
            "URIError": false,
            "valueOf": false,
            "WeakMap": false,
            "WeakSet": false
        },
        "es2020": {
            "Array": false,
            "ArrayBuffer": false,
            "Atomics": false,
            "BigInt": false,
            "BigInt64Array": false,
            "BigUint64Array": false,
            "Boolean": false,
            "constructor": false,
            "DataView": false,
            "Date": false,
            "decodeURI": false,
            "decodeURIComponent": false,
            "encodeURI": false,
            "encodeURIComponent": false,
            "Error": false,
            "escape": false,
            "eval": false,
            "EvalError": false,
            "Float32Array": false,
            "Float64Array": false,
            "Function": false,
            "globalThis": false,
            "hasOwnProperty": false,
            "Infinity": false,
            "Int16Array": false,
            "Int32Array": false,
            "Int8Array": false,
            "isFinite": false,
            "isNaN": false,
            "isPrototypeOf": false,
            "JSON": false,
            "Map": false,
            "Math": false,
            "NaN": false,
            "Number": false,
            "Object": false,
            "parseFloat": false,
            "parseInt": false,
            "Promise": false,
            "propertyIsEnumerable": false,
            "Proxy": false,
            "RangeError": false,
            "ReferenceError": false,
            "Reflect": false,
            "RegExp": false,
            "Set": false,
            "SharedArrayBuffer": false,
            "String": false,
            "Symbol": false,
            "SyntaxError": false,
            "toLocaleString": false,
            "toString": false,
            "TypeError": false,
            "Uint16Array": false,
            "Uint32Array": false,
            "Uint8Array": false,
            "Uint8ClampedArray": false,
            "undefined": false,
            "unescape": false,
            "URIError": false,
            "valueOf": false,
            "WeakMap": false,
            "WeakSet": false
        },
        "es2021": {
            "AggregateError": false,
            "Array": false,
            "ArrayBuffer": false,
            "Atomics": false,
            "BigInt": false,
            "BigInt64Array": false,
            "BigUint64Array": false,
            "Boolean": false,
            "constructor": false,
            "DataView": false,
            "Date": false,
            "decodeURI": false,
            "decodeURIComponent": false,
            "encodeURI": false,
            "encodeURIComponent": false,
            "Error": false,
            "escape": false,
            "eval": false,
            "EvalError": false,
            "FinalizationRegistry": false,
            "Float32Array": false,
            "Float64Array": false,
            "Function": false,
            "globalThis": false,
            "hasOwnProperty": false,
            "Infinity": false,
            "Int16Array": false,
            "Int32Array": false,
            "Int8Array": false,
            "isFinite": false,
            "isNaN": false,
            "isPrototypeOf": false,
            "JSON": false,
            "Map": false,
            "Math": false,
            "NaN": false,
            "Number": false,
            "Object": false,
            "parseFloat": false,
            "parseInt": false,
            "Promise": false,
            "propertyIsEnumerable": false,
            "Proxy": false,
            "RangeError": false,
            "ReferenceError": false,
            "Reflect": false,
            "RegExp": false,
            "Set": false,
            "SharedArrayBuffer": false,
            "String": false,
            "Symbol": false,
            "SyntaxError": false,
            "toLocaleString": false,
            "toString": false,
            "TypeError": false,
            "Uint16Array": false,
            "Uint32Array": false,
            "Uint8Array": false,
            "Uint8ClampedArray": false,
            "undefined": false,
            "unescape": false,
            "URIError": false,
            "valueOf": false,
            "WeakMap": false,
            "WeakRef": false,
            "WeakSet": false
        },
        "browser": {
            "AbortController": false,
            "AbortSignal": false,
            "addEventListener": false,
            "alert": false,
            "AnalyserNode": false,
            "Animation": false,
            "AnimationEffectReadOnly": false,
            "AnimationEffectTiming": false,
            "AnimationEffectTimingReadOnly": false,
            "AnimationEvent": false,
            "AnimationPlaybackEvent": false,
            "AnimationTimeline": false,
            "applicationCache": false,
            "ApplicationCache": false,
            "ApplicationCacheErrorEvent": false,
            "atob": false,
            "Attr": false,
            "Audio": false,
            "AudioBuffer": false,
            "AudioBufferSourceNode": false,
            "AudioContext": false,
            "AudioDestinationNode": false,
            "AudioListener": false,
            "AudioNode": false,
            "AudioParam": false,
            "AudioProcessingEvent": false,
            "AudioScheduledSourceNode": false,
            "AudioWorkletGlobalScope ": false,
            "AudioWorkletNode": false,
            "AudioWorkletProcessor": false,
            "BarProp": false,
            "BaseAudioContext": false,
            "BatteryManager": false,
            "BeforeUnloadEvent": false,
            "BiquadFilterNode": false,
            "Blob": false,
            "BlobEvent": false,
            "blur": false,
            "BroadcastChannel": false,
            "btoa": false,
            "BudgetService": false,
            "ByteLengthQueuingStrategy": false,
            "Cache": false,
            "caches": false,
            "CacheStorage": false,
            "cancelAnimationFrame": false,
            "cancelIdleCallback": false,
            "CanvasCaptureMediaStreamTrack": false,
            "CanvasGradient": false,
            "CanvasPattern": false,
            "CanvasRenderingContext2D": false,
            "ChannelMergerNode": false,
            "ChannelSplitterNode": false,
            "CharacterData": false,
            "clearInterval": false,
            "clearTimeout": false,
            "clientInformation": false,
            "ClipboardEvent": false,
            "close": false,
            "closed": false,
            "CloseEvent": false,
            "Comment": false,
            "CompositionEvent": false,
            "confirm": false,
            "console": false,
            "ConstantSourceNode": false,
            "ConvolverNode": false,
            "CountQueuingStrategy": false,
            "createImageBitmap": false,
            "Credential": false,
            "CredentialsContainer": false,
            "crypto": false,
            "Crypto": false,
            "CryptoKey": false,
            "CSS": false,
            "CSSConditionRule": false,
            "CSSFontFaceRule": false,
            "CSSGroupingRule": false,
            "CSSImportRule": false,
            "CSSKeyframeRule": false,
            "CSSKeyframesRule": false,
            "CSSMatrixComponent": false,
            "CSSMediaRule": false,
            "CSSNamespaceRule": false,
            "CSSPageRule": false,
            "CSSPerspective": false,
            "CSSRotate": false,
            "CSSRule": false,
            "CSSRuleList": false,
            "CSSScale": false,
            "CSSSkew": false,
            "CSSSkewX": false,
            "CSSSkewY": false,
            "CSSStyleDeclaration": false,
            "CSSStyleRule": false,
            "CSSStyleSheet": false,
            "CSSSupportsRule": false,
            "CSSTransformValue": false,
            "CSSTranslate": false,
            "CustomElementRegistry": false,
            "customElements": false,
            "CustomEvent": false,
            "DataTransfer": false,
            "DataTransferItem": false,
            "DataTransferItemList": false,
            "defaultstatus": false,
            "defaultStatus": false,
            "DelayNode": false,
            "DeviceMotionEvent": false,
            "DeviceOrientationEvent": false,
            "devicePixelRatio": false,
            "dispatchEvent": false,
            "document": false,
            "Document": false,
            "DocumentFragment": false,
            "DocumentType": false,
            "DOMError": false,
            "DOMException": false,
            "DOMImplementation": false,
            "DOMMatrix": false,
            "DOMMatrixReadOnly": false,
            "DOMParser": false,
            "DOMPoint": false,
            "DOMPointReadOnly": false,
            "DOMQuad": false,
            "DOMRect": false,
            "DOMRectList": false,
            "DOMRectReadOnly": false,
            "DOMStringList": false,
            "DOMStringMap": false,
            "DOMTokenList": false,
            "DragEvent": false,
            "DynamicsCompressorNode": false,
            "Element": false,
            "ErrorEvent": false,
            "event": false,
            "Event": false,
            "EventSource": false,
            "EventTarget": false,
            "external": false,
            "fetch": false,
            "File": false,
            "FileList": false,
            "FileReader": false,
            "find": false,
            "focus": false,
            "FocusEvent": false,
            "FontFace": false,
            "FontFaceSetLoadEvent": false,
            "FormData": false,
            "frameElement": false,
            "frames": false,
            "GainNode": false,
            "Gamepad": false,
            "GamepadButton": false,
            "GamepadEvent": false,
            "getComputedStyle": false,
            "getSelection": false,
            "HashChangeEvent": false,
            "Headers": false,
            "history": false,
            "History": false,
            "HTMLAllCollection": false,
            "HTMLAnchorElement": false,
            "HTMLAreaElement": false,
            "HTMLAudioElement": false,
            "HTMLBaseElement": false,
            "HTMLBodyElement": false,
            "HTMLBRElement": false,
            "HTMLButtonElement": false,
            "HTMLCanvasElement": false,
            "HTMLCollection": false,
            "HTMLContentElement": false,
            "HTMLDataElement": false,
            "HTMLDataListElement": false,
            "HTMLDetailsElement": false,
            "HTMLDialogElement": false,
            "HTMLDirectoryElement": false,
            "HTMLDivElement": false,
            "HTMLDListElement": false,
            "HTMLDocument": false,
            "HTMLElement": false,
            "HTMLEmbedElement": false,
            "HTMLFieldSetElement": false,
            "HTMLFontElement": false,
            "HTMLFormControlsCollection": false,
            "HTMLFormElement": false,
            "HTMLFrameElement": false,
            "HTMLFrameSetElement": false,
            "HTMLHeadElement": false,
            "HTMLHeadingElement": false,
            "HTMLHRElement": false,
            "HTMLHtmlElement": false,
            "HTMLIFrameElement": false,
            "HTMLImageElement": false,
            "HTMLInputElement": false,
            "HTMLLabelElement": false,
            "HTMLLegendElement": false,
            "HTMLLIElement": false,
            "HTMLLinkElement": false,
            "HTMLMapElement": false,
            "HTMLMarqueeElement": false,
            "HTMLMediaElement": false,
            "HTMLMenuElement": false,
            "HTMLMetaElement": false,
            "HTMLMeterElement": false,
            "HTMLModElement": false,
            "HTMLObjectElement": false,
            "HTMLOListElement": false,
            "HTMLOptGroupElement": false,
            "HTMLOptionElement": false,
            "HTMLOptionsCollection": false,
            "HTMLOutputElement": false,
            "HTMLParagraphElement": false,
            "HTMLParamElement": false,
            "HTMLPictureElement": false,
            "HTMLPreElement": false,
            "HTMLProgressElement": false,
            "HTMLQuoteElement": false,
            "HTMLScriptElement": false,
            "HTMLSelectElement": false,
            "HTMLShadowElement": false,
            "HTMLSlotElement": false,
            "HTMLSourceElement": false,
            "HTMLSpanElement": false,
            "HTMLStyleElement": false,
            "HTMLTableCaptionElement": false,
            "HTMLTableCellElement": false,
            "HTMLTableColElement": false,
            "HTMLTableElement": false,
            "HTMLTableRowElement": false,
            "HTMLTableSectionElement": false,
            "HTMLTemplateElement": false,
            "HTMLTextAreaElement": false,
            "HTMLTimeElement": false,
            "HTMLTitleElement": false,
            "HTMLTrackElement": false,
            "HTMLUListElement": false,
            "HTMLUnknownElement": false,
            "HTMLVideoElement": false,
            "IDBCursor": false,
            "IDBCursorWithValue": false,
            "IDBDatabase": false,
            "IDBFactory": false,
            "IDBIndex": false,
            "IDBKeyRange": false,
            "IDBObjectStore": false,
            "IDBOpenDBRequest": false,
            "IDBRequest": false,
            "IDBTransaction": false,
            "IDBVersionChangeEvent": false,
            "IdleDeadline": false,
            "IIRFilterNode": false,
            "Image": false,
            "ImageBitmap": false,
            "ImageBitmapRenderingContext": false,
            "ImageCapture": false,
            "ImageData": false,
            "indexedDB": false,
            "innerHeight": false,
            "innerWidth": false,
            "InputEvent": false,
            "IntersectionObserver": false,
            "IntersectionObserverEntry": false,
            "Intl": false,
            "isSecureContext": false,
            "KeyboardEvent": false,
            "KeyframeEffect": false,
            "KeyframeEffectReadOnly": false,
            "length": false,
            "localStorage": false,
            "location": true,
            "Location": false,
            "locationbar": false,
            "matchMedia": false,
            "MediaDeviceInfo": false,
            "MediaDevices": false,
            "MediaElementAudioSourceNode": false,
            "MediaEncryptedEvent": false,
            "MediaError": false,
            "MediaKeyMessageEvent": false,
            "MediaKeySession": false,
            "MediaKeyStatusMap": false,
            "MediaKeySystemAccess": false,
            "MediaList": false,
            "MediaQueryList": false,
            "MediaQueryListEvent": false,
            "MediaRecorder": false,
            "MediaSettingsRange": false,
            "MediaSource": false,
            "MediaStream": false,
            "MediaStreamAudioDestinationNode": false,
            "MediaStreamAudioSourceNode": false,
            "MediaStreamEvent": false,
            "MediaStreamTrack": false,
            "MediaStreamTrackEvent": false,
            "menubar": false,
            "MessageChannel": false,
            "MessageEvent": false,
            "MessagePort": false,
            "MIDIAccess": false,
            "MIDIConnectionEvent": false,
            "MIDIInput": false,
            "MIDIInputMap": false,
            "MIDIMessageEvent": false,
            "MIDIOutput": false,
            "MIDIOutputMap": false,
            "MIDIPort": false,
            "MimeType": false,
            "MimeTypeArray": false,
            "MouseEvent": false,
            "moveBy": false,
            "moveTo": false,
            "MutationEvent": false,
            "MutationObserver": false,
            "MutationRecord": false,
            "name": false,
            "NamedNodeMap": false,
            "NavigationPreloadManager": false,
            "navigator": false,
            "Navigator": false,
            "NetworkInformation": false,
            "Node": false,
            "NodeFilter": false,
            "NodeIterator": false,
            "NodeList": false,
            "Notification": false,
            "OfflineAudioCompletionEvent": false,
            "OfflineAudioContext": false,
            "offscreenBuffering": false,
            "OffscreenCanvas": true,
            "OffscreenCanvasRenderingContext2D": false,
            "onabort": true,
            "onafterprint": true,
            "onanimationend": true,
            "onanimationiteration": true,
            "onanimationstart": true,
            "onappinstalled": true,
            "onauxclick": true,
            "onbeforeinstallprompt": true,
            "onbeforeprint": true,
            "onbeforeunload": true,
            "onblur": true,
            "oncancel": true,
            "oncanplay": true,
            "oncanplaythrough": true,
            "onchange": true,
            "onclick": true,
            "onclose": true,
            "oncontextmenu": true,
            "oncuechange": true,
            "ondblclick": true,
            "ondevicemotion": true,
            "ondeviceorientation": true,
            "ondeviceorientationabsolute": true,
            "ondrag": true,
            "ondragend": true,
            "ondragenter": true,
            "ondragleave": true,
            "ondragover": true,
            "ondragstart": true,
            "ondrop": true,
            "ondurationchange": true,
            "onemptied": true,
            "onended": true,
            "onerror": true,
            "onfocus": true,
            "ongotpointercapture": true,
            "onhashchange": true,
            "oninput": true,
            "oninvalid": true,
            "onkeydown": true,
            "onkeypress": true,
            "onkeyup": true,
            "onlanguagechange": true,
            "onload": true,
            "onloadeddata": true,
            "onloadedmetadata": true,
            "onloadstart": true,
            "onlostpointercapture": true,
            "onmessage": true,
            "onmessageerror": true,
            "onmousedown": true,
            "onmouseenter": true,
            "onmouseleave": true,
            "onmousemove": true,
            "onmouseout": true,
            "onmouseover": true,
            "onmouseup": true,
            "onmousewheel": true,
            "onoffline": true,
            "ononline": true,
            "onpagehide": true,
            "onpageshow": true,
            "onpause": true,
            "onplay": true,
            "onplaying": true,
            "onpointercancel": true,
            "onpointerdown": true,
            "onpointerenter": true,
            "onpointerleave": true,
            "onpointermove": true,
            "onpointerout": true,
            "onpointerover": true,
            "onpointerup": true,
            "onpopstate": true,
            "onprogress": true,
            "onratechange": true,
            "onrejectionhandled": true,
            "onreset": true,
            "onresize": true,
            "onscroll": true,
            "onsearch": true,
            "onseeked": true,
            "onseeking": true,
            "onselect": true,
            "onstalled": true,
            "onstorage": true,
            "onsubmit": true,
            "onsuspend": true,
            "ontimeupdate": true,
            "ontoggle": true,
            "ontransitionend": true,
            "onunhandledrejection": true,
            "onunload": true,
            "onvolumechange": true,
            "onwaiting": true,
            "onwheel": true,
            "open": false,
            "openDatabase": false,
            "opener": false,
            "Option": false,
            "origin": false,
            "OscillatorNode": false,
            "outerHeight": false,
            "outerWidth": false,
            "OverconstrainedError": false,
            "PageTransitionEvent": false,
            "pageXOffset": false,
            "pageYOffset": false,
            "PannerNode": false,
            "parent": false,
            "Path2D": false,
            "PaymentAddress": false,
            "PaymentRequest": false,
            "PaymentRequestUpdateEvent": false,
            "PaymentResponse": false,
            "performance": false,
            "Performance": false,
            "PerformanceEntry": false,
            "PerformanceLongTaskTiming": false,
            "PerformanceMark": false,
            "PerformanceMeasure": false,
            "PerformanceNavigation": false,
            "PerformanceNavigationTiming": false,
            "PerformanceObserver": false,
            "PerformanceObserverEntryList": false,
            "PerformancePaintTiming": false,
            "PerformanceResourceTiming": false,
            "PerformanceTiming": false,
            "PeriodicWave": false,
            "Permissions": false,
            "PermissionStatus": false,
            "personalbar": false,
            "PhotoCapabilities": false,
            "Plugin": false,
            "PluginArray": false,
            "PointerEvent": false,
            "PopStateEvent": false,
            "postMessage": false,
            "Presentation": false,
            "PresentationAvailability": false,
            "PresentationConnection": false,
            "PresentationConnectionAvailableEvent": false,
            "PresentationConnectionCloseEvent": false,
            "PresentationConnectionList": false,
            "PresentationReceiver": false,
            "PresentationRequest": false,
            "print": false,
            "ProcessingInstruction": false,
            "ProgressEvent": false,
            "PromiseRejectionEvent": false,
            "prompt": false,
            "PushManager": false,
            "PushSubscription": false,
            "PushSubscriptionOptions": false,
            "queueMicrotask": false,
            "RadioNodeList": false,
            "Range": false,
            "ReadableStream": false,
            "registerProcessor": false,
            "RemotePlayback": false,
            "removeEventListener": false,
            "Request": false,
            "requestAnimationFrame": false,
            "requestIdleCallback": false,
            "resizeBy": false,
            "ResizeObserver": false,
            "ResizeObserverEntry": false,
            "resizeTo": false,
            "Response": false,
            "RTCCertificate": false,
            "RTCDataChannel": false,
            "RTCDataChannelEvent": false,
            "RTCDtlsTransport": false,
            "RTCIceCandidate": false,
            "RTCIceGatherer": false,
            "RTCIceTransport": false,
            "RTCPeerConnection": false,
            "RTCPeerConnectionIceEvent": false,
            "RTCRtpContributingSource": false,
            "RTCRtpReceiver": false,
            "RTCRtpSender": false,
            "RTCSctpTransport": false,
            "RTCSessionDescription": false,
            "RTCStatsReport": false,
            "RTCTrackEvent": false,
            "screen": false,
            "Screen": false,
            "screenLeft": false,
            "ScreenOrientation": false,
            "screenTop": false,
            "screenX": false,
            "screenY": false,
            "ScriptProcessorNode": false,
            "scroll": false,
            "scrollbars": false,
            "scrollBy": false,
            "scrollTo": false,
            "scrollX": false,
            "scrollY": false,
            "SecurityPolicyViolationEvent": false,
            "Selection": false,
            "self": false,
            "ServiceWorker": false,
            "ServiceWorkerContainer": false,
            "ServiceWorkerRegistration": false,
            "sessionStorage": false,
            "setInterval": false,
            "setTimeout": false,
            "ShadowRoot": false,
            "SharedWorker": false,
            "SourceBuffer": false,
            "SourceBufferList": false,
            "speechSynthesis": false,
            "SpeechSynthesisEvent": false,
            "SpeechSynthesisUtterance": false,
            "StaticRange": false,
            "status": false,
            "statusbar": false,
            "StereoPannerNode": false,
            "stop": false,
            "Storage": false,
            "StorageEvent": false,
            "StorageManager": false,
            "styleMedia": false,
            "StyleSheet": false,
            "StyleSheetList": false,
            "SubtleCrypto": false,
            "SVGAElement": false,
            "SVGAngle": false,
            "SVGAnimatedAngle": false,
            "SVGAnimatedBoolean": false,
            "SVGAnimatedEnumeration": false,
            "SVGAnimatedInteger": false,
            "SVGAnimatedLength": false,
            "SVGAnimatedLengthList": false,
            "SVGAnimatedNumber": false,
            "SVGAnimatedNumberList": false,
            "SVGAnimatedPreserveAspectRatio": false,
            "SVGAnimatedRect": false,
            "SVGAnimatedString": false,
            "SVGAnimatedTransformList": false,
            "SVGAnimateElement": false,
            "SVGAnimateMotionElement": false,
            "SVGAnimateTransformElement": false,
            "SVGAnimationElement": false,
            "SVGCircleElement": false,
            "SVGClipPathElement": false,
            "SVGComponentTransferFunctionElement": false,
            "SVGDefsElement": false,
            "SVGDescElement": false,
            "SVGDiscardElement": false,
            "SVGElement": false,
            "SVGEllipseElement": false,
            "SVGFEBlendElement": false,
            "SVGFEColorMatrixElement": false,
            "SVGFEComponentTransferElement": false,
            "SVGFECompositeElement": false,
            "SVGFEConvolveMatrixElement": false,
            "SVGFEDiffuseLightingElement": false,
            "SVGFEDisplacementMapElement": false,
            "SVGFEDistantLightElement": false,
            "SVGFEDropShadowElement": false,
            "SVGFEFloodElement": false,
            "SVGFEFuncAElement": false,
            "SVGFEFuncBElement": false,
            "SVGFEFuncGElement": false,
            "SVGFEFuncRElement": false,
            "SVGFEGaussianBlurElement": false,
            "SVGFEImageElement": false,
            "SVGFEMergeElement": false,
            "SVGFEMergeNodeElement": false,
            "SVGFEMorphologyElement": false,
            "SVGFEOffsetElement": false,
            "SVGFEPointLightElement": false,
            "SVGFESpecularLightingElement": false,
            "SVGFESpotLightElement": false,
            "SVGFETileElement": false,
            "SVGFETurbulenceElement": false,
            "SVGFilterElement": false,
            "SVGForeignObjectElement": false,
            "SVGGElement": false,
            "SVGGeometryElement": false,
            "SVGGradientElement": false,
            "SVGGraphicsElement": false,
            "SVGImageElement": false,
            "SVGLength": false,
            "SVGLengthList": false,
            "SVGLinearGradientElement": false,
            "SVGLineElement": false,
            "SVGMarkerElement": false,
            "SVGMaskElement": false,
            "SVGMatrix": false,
            "SVGMetadataElement": false,
            "SVGMPathElement": false,
            "SVGNumber": false,
            "SVGNumberList": false,
            "SVGPathElement": false,
            "SVGPatternElement": false,
            "SVGPoint": false,
            "SVGPointList": false,
            "SVGPolygonElement": false,
            "SVGPolylineElement": false,
            "SVGPreserveAspectRatio": false,
            "SVGRadialGradientElement": false,
            "SVGRect": false,
            "SVGRectElement": false,
            "SVGScriptElement": false,
            "SVGSetElement": false,
            "SVGStopElement": false,
            "SVGStringList": false,
            "SVGStyleElement": false,
            "SVGSVGElement": false,
            "SVGSwitchElement": false,
            "SVGSymbolElement": false,
            "SVGTextContentElement": false,
            "SVGTextElement": false,
            "SVGTextPathElement": false,
            "SVGTextPositioningElement": false,
            "SVGTitleElement": false,
            "SVGTransform": false,
            "SVGTransformList": false,
            "SVGTSpanElement": false,
            "SVGUnitTypes": false,
            "SVGUseElement": false,
            "SVGViewElement": false,
            "TaskAttributionTiming": false,
            "Text": false,
            "TextDecoder": false,
            "TextEncoder": false,
            "TextEvent": false,
            "TextMetrics": false,
            "TextTrack": false,
            "TextTrackCue": false,
            "TextTrackCueList": false,
            "TextTrackList": false,
            "TimeRanges": false,
            "toolbar": false,
            "top": false,
            "Touch": false,
            "TouchEvent": false,
            "TouchList": false,
            "TrackEvent": false,
            "TransitionEvent": false,
            "TreeWalker": false,
            "UIEvent": false,
            "URL": false,
            "URLSearchParams": false,
            "ValidityState": false,
            "visualViewport": false,
            "VisualViewport": false,
            "VTTCue": false,
            "WaveShaperNode": false,
            "WebAssembly": false,
            "WebGL2RenderingContext": false,
            "WebGLActiveInfo": false,
            "WebGLBuffer": false,
            "WebGLContextEvent": false,
            "WebGLFramebuffer": false,
            "WebGLProgram": false,
            "WebGLQuery": false,
            "WebGLRenderbuffer": false,
            "WebGLRenderingContext": false,
            "WebGLSampler": false,
            "WebGLShader": false,
            "WebGLShaderPrecisionFormat": false,
            "WebGLSync": false,
            "WebGLTexture": false,
            "WebGLTransformFeedback": false,
            "WebGLUniformLocation": false,
            "WebGLVertexArrayObject": false,
            "WebSocket": false,
            "WheelEvent": false,
            "window": false,
            "Window": false,
            "Worker": false,
            "WritableStream": false,
            "XMLDocument": false,
            "XMLHttpRequest": false,
            "XMLHttpRequestEventTarget": false,
            "XMLHttpRequestUpload": false,
            "XMLSerializer": false,
            "XPathEvaluator": false,
            "XPathExpression": false,
            "XPathResult": false,
            "XSLTProcessor": false
        },
        "worker": {
            "addEventListener": false,
            "applicationCache": false,
            "atob": false,
            "Blob": false,
            "BroadcastChannel": false,
            "btoa": false,
            "Cache": false,
            "caches": false,
            "clearInterval": false,
            "clearTimeout": false,
            "close": true,
            "console": false,
            "fetch": false,
            "FileReaderSync": false,
            "FormData": false,
            "Headers": false,
            "IDBCursor": false,
            "IDBCursorWithValue": false,
            "IDBDatabase": false,
            "IDBFactory": false,
            "IDBIndex": false,
            "IDBKeyRange": false,
            "IDBObjectStore": false,
            "IDBOpenDBRequest": false,
            "IDBRequest": false,
            "IDBTransaction": false,
            "IDBVersionChangeEvent": false,
            "ImageData": false,
            "importScripts": true,
            "indexedDB": false,
            "location": false,
            "MessageChannel": false,
            "MessagePort": false,
            "name": false,
            "navigator": false,
            "Notification": false,
            "onclose": true,
            "onconnect": true,
            "onerror": true,
            "onlanguagechange": true,
            "onmessage": true,
            "onoffline": true,
            "ononline": true,
            "onrejectionhandled": true,
            "onunhandledrejection": true,
            "performance": false,
            "Performance": false,
            "PerformanceEntry": false,
            "PerformanceMark": false,
            "PerformanceMeasure": false,
            "PerformanceNavigation": false,
            "PerformanceResourceTiming": false,
            "PerformanceTiming": false,
            "postMessage": true,
            "Promise": false,
            "queueMicrotask": false,
            "removeEventListener": false,
            "Request": false,
            "Response": false,
            "self": true,
            "ServiceWorkerRegistration": false,
            "setInterval": false,
            "setTimeout": false,
            "TextDecoder": false,
            "TextEncoder": false,
            "URL": false,
            "URLSearchParams": false,
            "WebSocket": false,
            "Worker": false,
            "WorkerGlobalScope": false,
            "XMLHttpRequest": false
        },
        "node": {
            "__dirname": false,
            "__filename": false,
            "AbortController": false,
            "Buffer": false,
            "clearImmediate": false,
            "clearInterval": false,
            "clearTimeout": false,
            "console": false,
            "exports": true,
            "global": false,
            "Intl": false,
            "module": false,
            "process": false,
            "queueMicrotask": false,
            "require": false,
            "setImmediate": false,
            "setInterval": false,
            "setTimeout": false,
            "TextDecoder": false,
            "TextEncoder": false,
            "URL": false,
            "URLSearchParams": false
        },
        "nodeBuiltin": {
            "AbortController": false,
            "Buffer": false,
            "clearImmediate": false,
            "clearInterval": false,
            "clearTimeout": false,
            "console": false,
            "global": false,
            "Intl": false,
            "process": false,
            "queueMicrotask": false,
            "setImmediate": false,
            "setInterval": false,
            "setTimeout": false,
            "TextDecoder": false,
            "TextEncoder": false,
            "URL": false,
            "URLSearchParams": false
        },
        "commonjs": {
            "exports": true,
            "global": false,
            "module": false,
            "require": false
        },
        "amd": {
            "define": false,
            "require": false
        },
        "mocha": {
            "after": false,
            "afterEach": false,
            "before": false,
            "beforeEach": false,
            "context": false,
            "describe": false,
            "it": false,
            "mocha": false,
            "run": false,
            "setup": false,
            "specify": false,
            "suite": false,
            "suiteSetup": false,
            "suiteTeardown": false,
            "teardown": false,
            "test": false,
            "xcontext": false,
            "xdescribe": false,
            "xit": false,
            "xspecify": false
        },
        "jasmine": {
            "afterAll": false,
            "afterEach": false,
            "beforeAll": false,
            "beforeEach": false,
            "describe": false,
            "expect": false,
            "expectAsync": false,
            "fail": false,
            "fdescribe": false,
            "fit": false,
            "it": false,
            "jasmine": false,
            "pending": false,
            "runs": false,
            "spyOn": false,
            "spyOnAllFunctions": false,
            "spyOnProperty": false,
            "waits": false,
            "waitsFor": false,
            "xdescribe": false,
            "xit": false
        },
        "jest": {
            "afterAll": false,
            "afterEach": false,
            "beforeAll": false,
            "beforeEach": false,
            "describe": false,
            "expect": false,
            "fdescribe": false,
            "fit": false,
            "it": false,
            "jest": false,
            "pit": false,
            "require": false,
            "test": false,
            "xdescribe": false,
            "xit": false,
            "xtest": false
        },
        "qunit": {
            "asyncTest": false,
            "deepEqual": false,
            "equal": false,
            "expect": false,
            "module": false,
            "notDeepEqual": false,
            "notEqual": false,
            "notOk": false,
            "notPropEqual": false,
            "notStrictEqual": false,
            "ok": false,
            "propEqual": false,
            "QUnit": false,
            "raises": false,
            "start": false,
            "stop": false,
            "strictEqual": false,
            "test": false,
            "throws": false
        },
        "phantomjs": {
            "console": true,
            "exports": true,
            "phantom": true,
            "require": true,
            "WebPage": true
        },
        "couch": {
            "emit": false,
            "exports": false,
            "getRow": false,
            "log": false,
            "module": false,
            "provides": false,
            "require": false,
            "respond": false,
            "send": false,
            "start": false,
            "sum": false
        },
        "rhino": {
            "defineClass": false,
            "deserialize": false,
            "gc": false,
            "help": false,
            "importClass": false,
            "importPackage": false,
            "java": false,
            "load": false,
            "loadClass": false,
            "Packages": false,
            "print": false,
            "quit": false,
            "readFile": false,
            "readUrl": false,
            "runCommand": false,
            "seal": false,
            "serialize": false,
            "spawn": false,
            "sync": false,
            "toint32": false,
            "version": false
        },
        "nashorn": {
            "__DIR__": false,
            "__FILE__": false,
            "__LINE__": false,
            "com": false,
            "edu": false,
            "exit": false,
            "java": false,
            "Java": false,
            "javafx": false,
            "JavaImporter": false,
            "javax": false,
            "JSAdapter": false,
            "load": false,
            "loadWithNewGlobal": false,
            "org": false,
            "Packages": false,
            "print": false,
            "quit": false
        },
        "wsh": {
            "ActiveXObject": false,
            "CollectGarbage": false,
            "Debug": false,
            "Enumerator": false,
            "GetObject": false,
            "RuntimeObject": false,
            "ScriptEngine": false,
            "ScriptEngineBuildVersion": false,
            "ScriptEngineMajorVersion": false,
            "ScriptEngineMinorVersion": false,
            "VBArray": false,
            "WScript": false,
            "WSH": false
        },
        "jquery": {
            "$": false,
            "jQuery": false
        },
        "yui": {
            "YAHOO": false,
            "YAHOO_config": false,
            "YUI": false,
            "YUI_config": false
        },
        "shelljs": {
            "cat": false,
            "cd": false,
            "chmod": false,
            "config": false,
            "cp": false,
            "dirs": false,
            "echo": false,
            "env": false,
            "error": false,
            "exec": false,
            "exit": false,
            "find": false,
            "grep": false,
            "ln": false,
            "ls": false,
            "mkdir": false,
            "mv": false,
            "popd": false,
            "pushd": false,
            "pwd": false,
            "rm": false,
            "sed": false,
            "set": false,
            "target": false,
            "tempdir": false,
            "test": false,
            "touch": false,
            "which": false
        },
        "prototypejs": {
            "$": false,
            "$$": false,
            "$A": false,
            "$break": false,
            "$continue": false,
            "$F": false,
            "$H": false,
            "$R": false,
            "$w": false,
            "Abstract": false,
            "Ajax": false,
            "Autocompleter": false,
            "Builder": false,
            "Class": false,
            "Control": false,
            "Draggable": false,
            "Draggables": false,
            "Droppables": false,
            "Effect": false,
            "Element": false,
            "Enumerable": false,
            "Event": false,
            "Field": false,
            "Form": false,
            "Hash": false,
            "Insertion": false,
            "ObjectRange": false,
            "PeriodicalExecuter": false,
            "Position": false,
            "Prototype": false,
            "Scriptaculous": false,
            "Selector": false,
            "Sortable": false,
            "SortableObserver": false,
            "Sound": false,
            "Template": false,
            "Toggle": false,
            "Try": false
        },
        "meteor": {
            "$": false,
            "Accounts": false,
            "AccountsClient": false,
            "AccountsCommon": false,
            "AccountsServer": false,
            "App": false,
            "Assets": false,
            "Blaze": false,
            "check": false,
            "Cordova": false,
            "DDP": false,
            "DDPRateLimiter": false,
            "DDPServer": false,
            "Deps": false,
            "EJSON": false,
            "Email": false,
            "HTTP": false,
            "Log": false,
            "Match": false,
            "Meteor": false,
            "Mongo": false,
            "MongoInternals": false,
            "Npm": false,
            "Package": false,
            "Plugin": false,
            "process": false,
            "Random": false,
            "ReactiveDict": false,
            "ReactiveVar": false,
            "Router": false,
            "ServiceConfiguration": false,
            "Session": false,
            "share": false,
            "Spacebars": false,
            "Template": false,
            "Tinytest": false,
            "Tracker": false,
            "UI": false,
            "Utils": false,
            "WebApp": false,
            "WebAppInternals": false
        },
        "mongo": {
            "_isWindows": false,
            "_rand": false,
            "BulkWriteResult": false,
            "cat": false,
            "cd": false,
            "connect": false,
            "db": false,
            "getHostName": false,
            "getMemInfo": false,
            "hostname": false,
            "ISODate": false,
            "listFiles": false,
            "load": false,
            "ls": false,
            "md5sumFile": false,
            "mkdir": false,
            "Mongo": false,
            "NumberInt": false,
            "NumberLong": false,
            "ObjectId": false,
            "PlanCache": false,
            "print": false,
            "printjson": false,
            "pwd": false,
            "quit": false,
            "removeFile": false,
            "rs": false,
            "sh": false,
            "UUID": false,
            "version": false,
            "WriteResult": false
        },
        "applescript": {
            "$": false,
            "Application": false,
            "Automation": false,
            "console": false,
            "delay": false,
            "Library": false,
            "ObjC": false,
            "ObjectSpecifier": false,
            "Path": false,
            "Progress": false,
            "Ref": false
        },
        "serviceworker": {
            "addEventListener": false,
            "applicationCache": false,
            "atob": false,
            "Blob": false,
            "BroadcastChannel": false,
            "btoa": false,
            "Cache": false,
            "caches": false,
            "CacheStorage": false,
            "clearInterval": false,
            "clearTimeout": false,
            "Client": false,
            "clients": false,
            "Clients": false,
            "close": true,
            "console": false,
            "ExtendableEvent": false,
            "ExtendableMessageEvent": false,
            "fetch": false,
            "FetchEvent": false,
            "FileReaderSync": false,
            "FormData": false,
            "Headers": false,
            "IDBCursor": false,
            "IDBCursorWithValue": false,
            "IDBDatabase": false,
            "IDBFactory": false,
            "IDBIndex": false,
            "IDBKeyRange": false,
            "IDBObjectStore": false,
            "IDBOpenDBRequest": false,
            "IDBRequest": false,
            "IDBTransaction": false,
            "IDBVersionChangeEvent": false,
            "ImageData": false,
            "importScripts": false,
            "indexedDB": false,
            "location": false,
            "MessageChannel": false,
            "MessagePort": false,
            "name": false,
            "navigator": false,
            "Notification": false,
            "onclose": true,
            "onconnect": true,
            "onerror": true,
            "onfetch": true,
            "oninstall": true,
            "onlanguagechange": true,
            "onmessage": true,
            "onmessageerror": true,
            "onnotificationclick": true,
            "onnotificationclose": true,
            "onoffline": true,
            "ononline": true,
            "onpush": true,
            "onpushsubscriptionchange": true,
            "onrejectionhandled": true,
            "onsync": true,
            "onunhandledrejection": true,
            "performance": false,
            "Performance": false,
            "PerformanceEntry": false,
            "PerformanceMark": false,
            "PerformanceMeasure": false,
            "PerformanceNavigation": false,
            "PerformanceResourceTiming": false,
            "PerformanceTiming": false,
            "postMessage": true,
            "Promise": false,
            "queueMicrotask": false,
            "registration": false,
            "removeEventListener": false,
            "Request": false,
            "Response": false,
            "self": false,
            "ServiceWorker": false,
            "ServiceWorkerContainer": false,
            "ServiceWorkerGlobalScope": false,
            "ServiceWorkerMessageEvent": false,
            "ServiceWorkerRegistration": false,
            "setInterval": false,
            "setTimeout": false,
            "skipWaiting": false,
            "TextDecoder": false,
            "TextEncoder": false,
            "URL": false,
            "URLSearchParams": false,
            "WebSocket": false,
            "WindowClient": false,
            "Worker": false,
            "WorkerGlobalScope": false,
            "XMLHttpRequest": false
        },
        "atomtest": {
            "advanceClock": false,
            "atom": false,
            "fakeClearInterval": false,
            "fakeClearTimeout": false,
            "fakeSetInterval": false,
            "fakeSetTimeout": false,
            "resetTimeouts": false,
            "waitsForPromise": false
        },
        "embertest": {
            "andThen": false,
            "click": false,
            "currentPath": false,
            "currentRouteName": false,
            "currentURL": false,
            "fillIn": false,
            "find": false,
            "findAll": false,
            "findWithAssert": false,
            "keyEvent": false,
            "pauseTest": false,
            "resumeTest": false,
            "triggerEvent": false,
            "visit": false,
            "wait": false
        },
        "protractor": {
            "$": false,
            "$$": false,
            "browser": false,
            "by": false,
            "By": false,
            "DartObject": false,
            "element": false,
            "protractor": false
        },
        "shared-node-browser": {
            "clearInterval": false,
            "clearTimeout": false,
            "console": false,
            "setInterval": false,
            "setTimeout": false,
            "URL": false,
            "URLSearchParams": false
        },
        "webextensions": {
            "browser": false,
            "chrome": false,
            "opr": false
        },
        "greasemonkey": {
            "cloneInto": false,
            "createObjectIn": false,
            "exportFunction": false,
            "GM": false,
            "GM_addStyle": false,
            "GM_addValueChangeListener": false,
            "GM_deleteValue": false,
            "GM_download": false,
            "GM_getResourceText": false,
            "GM_getResourceURL": false,
            "GM_getTab": false,
            "GM_getTabs": false,
            "GM_getValue": false,
            "GM_info": false,
            "GM_listValues": false,
            "GM_log": false,
            "GM_notification": false,
            "GM_openInTab": false,
            "GM_registerMenuCommand": false,
            "GM_removeValueChangeListener": false,
            "GM_saveTab": false,
            "GM_setClipboard": false,
            "GM_setValue": false,
            "GM_unregisterMenuCommand": false,
            "GM_xmlhttpRequest": false,
            "unsafeWindow": false
        },
        "devtools": {
            "$": false,
            "$_": false,
            "$$": false,
            "$0": false,
            "$1": false,
            "$2": false,
            "$3": false,
            "$4": false,
            "$x": false,
            "chrome": false,
            "clear": false,
            "copy": false,
            "debug": false,
            "dir": false,
            "dirxml": false,
            "getEventListeners": false,
            "inspect": false,
            "keys": false,
            "monitor": false,
            "monitorEvents": false,
            "profile": false,
            "profileEnd": false,
            "queryObjects": false,
            "table": false,
            "undebug": false,
            "unmonitor": false,
            "unmonitorEvents": false,
            "values": false
        }
    };
    module.exports = json;
},
/* @eslint/eslintrc/lib/shared/config-ops.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Private
    //------------------------------------------------------------------------------
    const RULE_SEVERITY_STRINGS = ["off", "warn", "error"], RULE_SEVERITY = RULE_SEVERITY_STRINGS.reduce((map, value, index) => {
        map[value] = index;
        return map;
    }, {}), VALID_SEVERITIES = [0, 1, 2, "off", "warn", "error"];
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = {
        /**
         * Normalizes the severity value of a rule's configuration to a number
         * @param {(number|string|[number, ...*]|[string, ...*])} ruleConfig A rule's configuration value, generally
         * received from the user. A valid config value is either 0, 1, 2, the string "off" (treated the same as 0),
         * the string "warn" (treated the same as 1), the string "error" (treated the same as 2), or an array
         * whose first element is one of the above values. Strings are matched case-insensitively.
         * @returns {(0|1|2)} The numeric severity value if the config value was valid, otherwise 0.
         */
        getRuleSeverity(ruleConfig) {
            const severityValue = Array.isArray(ruleConfig) ? ruleConfig[0] : ruleConfig;
            if (severityValue === 0 || severityValue === 1 || severityValue === 2) {
                return severityValue;
            }
            if (typeof severityValue === "string") {
                return RULE_SEVERITY[severityValue.toLowerCase()] || 0;
            }
            return 0;
        },
        /**
         * Converts old-style severity settings (0, 1, 2) into new-style
         * severity settings (off, warn, error) for all rules. Assumption is that severity
         * values have already been validated as correct.
         * @param {Object} config The config object to normalize.
         * @returns {void}
         */
        normalizeToStrings(config) {
            if (config.rules) {
                Object.keys(config.rules).forEach(ruleId => {
                    const ruleConfig = config.rules[ruleId];
                    if (typeof ruleConfig === "number") {
                        config.rules[ruleId] = RULE_SEVERITY_STRINGS[ruleConfig] || RULE_SEVERITY_STRINGS[0];
                    }
                    else if (Array.isArray(ruleConfig) && typeof ruleConfig[0] === "number") {
                        ruleConfig[0] = RULE_SEVERITY_STRINGS[ruleConfig[0]] || RULE_SEVERITY_STRINGS[0];
                    }
                });
            }
        },
        /**
         * Determines if the severity for the given rule configuration represents an error.
         * @param {int|string|Array} ruleConfig The configuration for an individual rule.
         * @returns {boolean} True if the rule represents an error, false if not.
         */
        isErrorSeverity(ruleConfig) {
            return module.exports.getRuleSeverity(ruleConfig) === 2;
        },
        /**
         * Checks whether a given config has valid severity or not.
         * @param {number|string|Array} ruleConfig The configuration for an individual rule.
         * @returns {boolean} `true` if the configuration has valid severity.
         */
        isValidSeverity(ruleConfig) {
            let severity = Array.isArray(ruleConfig) ? ruleConfig[0] : ruleConfig;
            if (typeof severity === "string") {
                severity = severity.toLowerCase();
            }
            return VALID_SEVERITIES.indexOf(severity) !== -1;
        },
        /**
         * Checks whether every rule of a given config has valid severity or not.
         * @param {Object} config The configuration for rules.
         * @returns {boolean} `true` if the configuration has valid severity.
         */
        isEverySeverityValid(config) {
            return Object.keys(config).every(ruleId => this.isValidSeverity(config[ruleId]));
        },
        /**
         * Normalizes a value for a global in a config
         * @param {(boolean|string|null)} configuredValue The value given for a global in configuration or in
         * a global directive comment
         * @returns {("readable"|"writeable"|"off")} The value normalized as a string
         * @throws Error if global value is invalid
         */
        normalizeConfigGlobal(configuredValue) {
            switch (configuredValue) {
                case "off":
                    return "off";
                case true:
                case "true":
                case "writeable":
                case "writable":
                    return "writable";
                case null:
                case false:
                case "false":
                case "readable":
                case "readonly":
                    return "readonly";
                default:
                    throw new Error(`'${configuredValue}' is not a valid configuration for a global (use 'readonly', 'writable', or 'off')`);
            }
        }
    };
},
/* @eslint/eslintrc/lib/shared/deprecation-warnings.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const path = require("path");
    //------------------------------------------------------------------------------
    // Private
    //------------------------------------------------------------------------------
    // Defitions for deprecation warnings.
    const deprecationWarningMessages = {
        ESLINT_LEGACY_ECMAFEATURES: "The 'ecmaFeatures' config file property is deprecated and has no effect.",
        ESLINT_PERSONAL_CONFIG_LOAD: "'~/.eslintrc.*' config files have been deprecated. " +
            "Please use a config file per project or the '--config' option.",
        ESLINT_PERSONAL_CONFIG_SUPPRESS: "'~/.eslintrc.*' config files have been deprecated. " +
            "Please remove it or add 'root:true' to the config files in your " +
            "projects in order to avoid loading '~/.eslintrc.*' accidentally."
    };
    const sourceFileErrorCache = new Set();
    /**
     * Emits a deprecation warning containing a given filepath. A new deprecation warning is emitted
     * for each unique file path, but repeated invocations with the same file path have no effect.
     * No warnings are emitted if the `--no-deprecation` or `--no-warnings` Node runtime flags are active.
     * @param {string} source The name of the configuration source to report the warning for.
     * @param {string} errorCode The warning message to show.
     * @returns {void}
     */
    function emitDeprecationWarning(source, errorCode) {
        const cacheKey = JSON.stringify({ source, errorCode });
        if (sourceFileErrorCache.has(cacheKey)) {
            return;
        }
        sourceFileErrorCache.add(cacheKey);
        const rel = path.relative(process.cwd(), source);
        const message = deprecationWarningMessages[errorCode];
        process.emitWarning(`${message} (found in "${rel}")`, "DeprecationWarning", errorCode);
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = {
        emitDeprecationWarning
    };
},
/* @eslint/eslintrc/lib/shared/ajv.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const Ajv = require(223) /* ajv */, metaSchema = require(269) /* ajv/lib/refs/json-schema-draft-04.json */;
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = (additionalOptions = {}) => {
        const ajv = new Ajv({
            meta: false,
            useDefaults: true,
            validateSchema: false,
            missingRefs: "ignore",
            verbose: true,
            schemaId: "auto",
            ...additionalOptions
        });
        ajv.addMetaSchema(metaSchema);
        // eslint-disable-next-line no-underscore-dangle
        ajv._opts.defaultMeta = metaSchema.id;
        return ajv;
    };
},
/* ajv/lib/ajv.js */ function _(require, module, exports, __esModule, __esExport) {
    var compileSchema = require(224) /* ./compile */, resolve = require(225) /* ./compile/resolve */, Cache = require(235) /* ./cache */, SchemaObject = require(230) /* ./compile/schema_obj */, stableStringify = require(233) /* fast-json-stable-stringify */, formats = require(236) /* ./compile/formats */, rules = require(237) /* ./compile/rules */, $dataMetaSchema = require(262) /* ./data */, util = require(228) /* ./compile/util */;
    module.exports = Ajv;
    Ajv.prototype.validate = validate;
    Ajv.prototype.compile = compile;
    Ajv.prototype.addSchema = addSchema;
    Ajv.prototype.addMetaSchema = addMetaSchema;
    Ajv.prototype.validateSchema = validateSchema;
    Ajv.prototype.getSchema = getSchema;
    Ajv.prototype.removeSchema = removeSchema;
    Ajv.prototype.addFormat = addFormat;
    Ajv.prototype.errorsText = errorsText;
    Ajv.prototype._addSchema = _addSchema;
    Ajv.prototype._compile = _compile;
    Ajv.prototype.compileAsync = require(263) /* ./compile/async */;
    var customKeyword = require(264) /* ./keyword */;
    Ajv.prototype.addKeyword = customKeyword.add;
    Ajv.prototype.getKeyword = customKeyword.get;
    Ajv.prototype.removeKeyword = customKeyword.remove;
    Ajv.prototype.validateKeyword = customKeyword.validate;
    var errorClasses = require(232) /* ./compile/error_classes */;
    Ajv.ValidationError = errorClasses.Validation;
    Ajv.MissingRefError = errorClasses.MissingRef;
    Ajv.$dataMetaSchema = $dataMetaSchema;
    var META_SCHEMA_ID = 'http://json-schema.org/draft-07/schema';
    var META_IGNORE_OPTIONS = ['removeAdditional', 'useDefaults', 'coerceTypes', 'strictDefaults'];
    var META_SUPPORT_DATA = ['/properties'];
    /**
     * Creates validator instance.
     * Usage: `Ajv(opts)`
     * @param {Object} opts optional options
     * @return {Object} ajv instance
     */
    function Ajv(opts) {
        if (!(this instanceof Ajv))
            return new Ajv(opts);
        opts = this._opts = util.copy(opts) || {};
        setLogger(this);
        this._schemas = {};
        this._refs = {};
        this._fragments = {};
        this._formats = formats(opts.format);
        this._cache = opts.cache || new Cache;
        this._loadingSchemas = {};
        this._compilations = [];
        this.RULES = rules();
        this._getId = chooseGetId(opts);
        opts.loopRequired = opts.loopRequired || Infinity;
        if (opts.errorDataPath == 'property')
            opts._errorDataPathProperty = true;
        if (opts.serialize === undefined)
            opts.serialize = stableStringify;
        this._metaOpts = getMetaSchemaOptions(this);
        if (opts.formats)
            addInitialFormats(this);
        if (opts.keywords)
            addInitialKeywords(this);
        addDefaultMetaSchema(this);
        if (typeof opts.meta == 'object')
            this.addMetaSchema(opts.meta);
        if (opts.nullable)
            this.addKeyword('nullable', { metaSchema: { type: 'boolean' } });
        addInitialSchemas(this);
    }
    /**
     * Validate data using schema
     * Schema will be compiled and cached (using serialized JSON as key. [fast-json-stable-stringify](https://github.com/epoberezkin/fast-json-stable-stringify) is used to serialize.
     * @this   Ajv
     * @param  {String|Object} schemaKeyRef key, ref or schema object
     * @param  {Any} data to be validated
     * @return {Boolean} validation result. Errors from the last validation will be available in `ajv.errors` (and also in compiled schema: `schema.errors`).
     */
    function validate(schemaKeyRef, data) {
        var v;
        if (typeof schemaKeyRef == 'string') {
            v = this.getSchema(schemaKeyRef);
            if (!v)
                throw new Error('no schema with key or ref "' + schemaKeyRef + '"');
        }
        else {
            var schemaObj = this._addSchema(schemaKeyRef);
            v = schemaObj.validate || this._compile(schemaObj);
        }
        var valid = v(data);
        if (v.$async !== true)
            this.errors = v.errors;
        return valid;
    }
    /**
     * Create validating function for passed schema.
     * @this   Ajv
     * @param  {Object} schema schema object
     * @param  {Boolean} _meta true if schema is a meta-schema. Used internally to compile meta schemas of custom keywords.
     * @return {Function} validating function
     */
    function compile(schema, _meta) {
        var schemaObj = this._addSchema(schema, undefined, _meta);
        return schemaObj.validate || this._compile(schemaObj);
    }
    /**
     * Adds schema to the instance.
     * @this   Ajv
     * @param {Object|Array} schema schema or array of schemas. If array is passed, `key` and other parameters will be ignored.
     * @param {String} key Optional schema key. Can be passed to `validate` method instead of schema object or id/ref. One schema per instance can have empty `id` and `key`.
     * @param {Boolean} _skipValidation true to skip schema validation. Used internally, option validateSchema should be used instead.
     * @param {Boolean} _meta true if schema is a meta-schema. Used internally, addMetaSchema should be used instead.
     * @return {Ajv} this for method chaining
     */
    function addSchema(schema, key, _skipValidation, _meta) {
        if (Array.isArray(schema)) {
            for (var i = 0; i < schema.length; i++)
                this.addSchema(schema[i], undefined, _skipValidation, _meta);
            return this;
        }
        var id = this._getId(schema);
        if (id !== undefined && typeof id != 'string')
            throw new Error('schema id must be string');
        key = resolve.normalizeId(key || id);
        checkUnique(this, key);
        this._schemas[key] = this._addSchema(schema, _skipValidation, _meta, true);
        return this;
    }
    /**
     * Add schema that will be used to validate other schemas
     * options in META_IGNORE_OPTIONS are alway set to false
     * @this   Ajv
     * @param {Object} schema schema object
     * @param {String} key optional schema key
     * @param {Boolean} skipValidation true to skip schema validation, can be used to override validateSchema option for meta-schema
     * @return {Ajv} this for method chaining
     */
    function addMetaSchema(schema, key, skipValidation) {
        this.addSchema(schema, key, skipValidation, true);
        return this;
    }
    /**
     * Validate schema
     * @this   Ajv
     * @param {Object} schema schema to validate
     * @param {Boolean} throwOrLogError pass true to throw (or log) an error if invalid
     * @return {Boolean} true if schema is valid
     */
    function validateSchema(schema, throwOrLogError) {
        var $schema = schema.$schema;
        if ($schema !== undefined && typeof $schema != 'string')
            throw new Error('$schema must be a string');
        $schema = $schema || this._opts.defaultMeta || defaultMeta(this);
        if (!$schema) {
            this.logger.warn('meta-schema not available');
            this.errors = null;
            return true;
        }
        var valid = this.validate($schema, schema);
        if (!valid && throwOrLogError) {
            var message = 'schema is invalid: ' + this.errorsText();
            if (this._opts.validateSchema == 'log')
                this.logger.error(message);
            else
                throw new Error(message);
        }
        return valid;
    }
    function defaultMeta(self) {
        var meta = self._opts.meta;
        self._opts.defaultMeta = typeof meta == 'object'
            ? self._getId(meta) || meta
            : self.getSchema(META_SCHEMA_ID)
                ? META_SCHEMA_ID
                : undefined;
        return self._opts.defaultMeta;
    }
    /**
     * Get compiled schema from the instance by `key` or `ref`.
     * @this   Ajv
     * @param  {String} keyRef `key` that was passed to `addSchema` or full schema reference (`schema.id` or resolved id).
     * @return {Function} schema validating function (with property `schema`).
     */
    function getSchema(keyRef) {
        var schemaObj = _getSchemaObj(this, keyRef);
        switch (typeof schemaObj) {
            case 'object': return schemaObj.validate || this._compile(schemaObj);
            case 'string': return this.getSchema(schemaObj);
            case 'undefined': return _getSchemaFragment(this, keyRef);
        }
    }
    function _getSchemaFragment(self, ref) {
        var res = resolve.schema.call(self, { schema: {} }, ref);
        if (res) {
            var schema = res.schema, root = res.root, baseId = res.baseId;
            var v = compileSchema.call(self, schema, root, undefined, baseId);
            self._fragments[ref] = new SchemaObject({
                ref: ref,
                fragment: true,
                schema: schema,
                root: root,
                baseId: baseId,
                validate: v
            });
            return v;
        }
    }
    function _getSchemaObj(self, keyRef) {
        keyRef = resolve.normalizeId(keyRef);
        return self._schemas[keyRef] || self._refs[keyRef] || self._fragments[keyRef];
    }
    /**
     * Remove cached schema(s).
     * If no parameter is passed all schemas but meta-schemas are removed.
     * If RegExp is passed all schemas with key/id matching pattern but meta-schemas are removed.
     * Even if schema is referenced by other schemas it still can be removed as other schemas have local references.
     * @this   Ajv
     * @param  {String|Object|RegExp} schemaKeyRef key, ref, pattern to match key/ref or schema object
     * @return {Ajv} this for method chaining
     */
    function removeSchema(schemaKeyRef) {
        if (schemaKeyRef instanceof RegExp) {
            _removeAllSchemas(this, this._schemas, schemaKeyRef);
            _removeAllSchemas(this, this._refs, schemaKeyRef);
            return this;
        }
        switch (typeof schemaKeyRef) {
            case 'undefined':
                _removeAllSchemas(this, this._schemas);
                _removeAllSchemas(this, this._refs);
                this._cache.clear();
                return this;
            case 'string':
                var schemaObj = _getSchemaObj(this, schemaKeyRef);
                if (schemaObj)
                    this._cache.del(schemaObj.cacheKey);
                delete this._schemas[schemaKeyRef];
                delete this._refs[schemaKeyRef];
                return this;
            case 'object':
                var serialize = this._opts.serialize;
                var cacheKey = serialize ? serialize(schemaKeyRef) : schemaKeyRef;
                this._cache.del(cacheKey);
                var id = this._getId(schemaKeyRef);
                if (id) {
                    id = resolve.normalizeId(id);
                    delete this._schemas[id];
                    delete this._refs[id];
                }
        }
        return this;
    }
    function _removeAllSchemas(self, schemas, regex) {
        for (var keyRef in schemas) {
            var schemaObj = schemas[keyRef];
            if (!schemaObj.meta && (!regex || regex.test(keyRef))) {
                self._cache.del(schemaObj.cacheKey);
                delete schemas[keyRef];
            }
        }
    }
    /* @this   Ajv */
    function _addSchema(schema, skipValidation, meta, shouldAddSchema) {
        if (typeof schema != 'object' && typeof schema != 'boolean')
            throw new Error('schema should be object or boolean');
        var serialize = this._opts.serialize;
        var cacheKey = serialize ? serialize(schema) : schema;
        var cached = this._cache.get(cacheKey);
        if (cached)
            return cached;
        shouldAddSchema = shouldAddSchema || this._opts.addUsedSchema !== false;
        var id = resolve.normalizeId(this._getId(schema));
        if (id && shouldAddSchema)
            checkUnique(this, id);
        var willValidate = this._opts.validateSchema !== false && !skipValidation;
        var recursiveMeta;
        if (willValidate && !(recursiveMeta = id && id == resolve.normalizeId(schema.$schema)))
            this.validateSchema(schema, true);
        var localRefs = resolve.ids.call(this, schema);
        var schemaObj = new SchemaObject({
            id: id,
            schema: schema,
            localRefs: localRefs,
            cacheKey: cacheKey,
            meta: meta
        });
        if (id[0] != '#' && shouldAddSchema)
            this._refs[id] = schemaObj;
        this._cache.put(cacheKey, schemaObj);
        if (willValidate && recursiveMeta)
            this.validateSchema(schema, true);
        return schemaObj;
    }
    /* @this   Ajv */
    function _compile(schemaObj, root) {
        if (schemaObj.compiling) {
            schemaObj.validate = callValidate;
            callValidate.schema = schemaObj.schema;
            callValidate.errors = null;
            callValidate.root = root ? root : callValidate;
            if (schemaObj.schema.$async === true)
                callValidate.$async = true;
            return callValidate;
        }
        schemaObj.compiling = true;
        var currentOpts;
        if (schemaObj.meta) {
            currentOpts = this._opts;
            this._opts = this._metaOpts;
        }
        var v;
        try {
            v = compileSchema.call(this, schemaObj.schema, root, schemaObj.localRefs);
        }
        catch (e) {
            delete schemaObj.validate;
            throw e;
        }
        finally {
            schemaObj.compiling = false;
            if (schemaObj.meta)
                this._opts = currentOpts;
        }
        schemaObj.validate = v;
        schemaObj.refs = v.refs;
        schemaObj.refVal = v.refVal;
        schemaObj.root = v.root;
        return v;
        /* @this   {*} - custom context, see passContext option */
        function callValidate() {
            /* jshint validthis: true */
            var _validate = schemaObj.validate;
            var result = _validate.apply(this, arguments);
            callValidate.errors = _validate.errors;
            return result;
        }
    }
    function chooseGetId(opts) {
        switch (opts.schemaId) {
            case 'auto': return _get$IdOrId;
            case 'id': return _getId;
            default: return _get$Id;
        }
    }
    /* @this   Ajv */
    function _getId(schema) {
        if (schema.$id)
            this.logger.warn('schema $id ignored', schema.$id);
        return schema.id;
    }
    /* @this   Ajv */
    function _get$Id(schema) {
        if (schema.id)
            this.logger.warn('schema id ignored', schema.id);
        return schema.$id;
    }
    function _get$IdOrId(schema) {
        if (schema.$id && schema.id && schema.$id != schema.id)
            throw new Error('schema $id is different from id');
        return schema.$id || schema.id;
    }
    /**
     * Convert array of error message objects to string
     * @this   Ajv
     * @param  {Array<Object>} errors optional array of validation errors, if not passed errors from the instance are used.
     * @param  {Object} options optional options with properties `separator` and `dataVar`.
     * @return {String} human readable string with all errors descriptions
     */
    function errorsText(errors, options) {
        errors = errors || this.errors;
        if (!errors)
            return 'No errors';
        options = options || {};
        var separator = options.separator === undefined ? ', ' : options.separator;
        var dataVar = options.dataVar === undefined ? 'data' : options.dataVar;
        var text = '';
        for (var i = 0; i < errors.length; i++) {
            var e = errors[i];
            if (e)
                text += dataVar + e.dataPath + ' ' + e.message + separator;
        }
        return text.slice(0, -separator.length);
    }
    /**
     * Add custom format
     * @this   Ajv
     * @param {String} name format name
     * @param {String|RegExp|Function} format string is converted to RegExp; function should return boolean (true when valid)
     * @return {Ajv} this for method chaining
     */
    function addFormat(name, format) {
        if (typeof format == 'string')
            format = new RegExp(format);
        this._formats[name] = format;
        return this;
    }
    function addDefaultMetaSchema(self) {
        var $dataSchema;
        if (self._opts.$data) {
            $dataSchema = require(268) /* ./refs/data.json */;
            self.addMetaSchema($dataSchema, $dataSchema.$id, true);
        }
        if (self._opts.meta === false)
            return;
        var metaSchema = require(267) /* ./refs/json-schema-draft-07.json */;
        if (self._opts.$data)
            metaSchema = $dataMetaSchema(metaSchema, META_SUPPORT_DATA);
        self.addMetaSchema(metaSchema, META_SCHEMA_ID, true);
        self._refs['http://json-schema.org/schema'] = META_SCHEMA_ID;
    }
    function addInitialSchemas(self) {
        var optsSchemas = self._opts.schemas;
        if (!optsSchemas)
            return;
        if (Array.isArray(optsSchemas))
            self.addSchema(optsSchemas);
        else
            for (var key in optsSchemas)
                self.addSchema(optsSchemas[key], key);
    }
    function addInitialFormats(self) {
        for (var name in self._opts.formats) {
            var format = self._opts.formats[name];
            self.addFormat(name, format);
        }
    }
    function addInitialKeywords(self) {
        for (var name in self._opts.keywords) {
            var keyword = self._opts.keywords[name];
            self.addKeyword(name, keyword);
        }
    }
    function checkUnique(self, id) {
        if (self._schemas[id] || self._refs[id])
            throw new Error('schema with key or id "' + id + '" already exists');
    }
    function getMetaSchemaOptions(self) {
        var metaOpts = util.copy(self._opts);
        for (var i = 0; i < META_IGNORE_OPTIONS.length; i++)
            delete metaOpts[META_IGNORE_OPTIONS[i]];
        return metaOpts;
    }
    function setLogger(self) {
        var logger = self._opts.logger;
        if (logger === false) {
            self.logger = { log: noop, warn: noop, error: noop };
        }
        else {
            if (logger === undefined)
                logger = console;
            if (!(typeof logger == 'object' && logger.log && logger.warn && logger.error))
                throw new Error('logger must implement log, warn and error methods');
            self.logger = logger;
        }
    }
    function noop() { }
},
/* ajv/lib/compile/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var resolve = require(225) /* ./resolve */, util = require(228) /* ./util */, errorClasses = require(232) /* ./error_classes */, stableStringify = require(233) /* fast-json-stable-stringify */;
    var validateGenerator = require(234) /* ../dotjs/validate */;
    /**
     * Functions below are used inside compiled validations function
     */
    var ucs2length = util.ucs2length;
    var equal = require(227) /* fast-deep-equal */;
    // this error is thrown by async schemas to return validation errors via exception
    var ValidationError = errorClasses.Validation;
    module.exports = compile;
    /**
     * Compiles schema to validation function
     * @this   Ajv
     * @param  {Object} schema schema object
     * @param  {Object} root object with information about the root schema for this schema
     * @param  {Object} localRefs the hash of local references inside the schema (created by resolve.id), used for inline resolution
     * @param  {String} baseId base ID for IDs in the schema
     * @return {Function} validation function
     */
    function compile(schema, root, localRefs, baseId) {
        /* jshint validthis: true, evil: true */
        /* eslint no-shadow: 0 */
        var self = this, opts = this._opts, refVal = [undefined], refs = {}, patterns = [], patternsHash = {}, defaults = [], defaultsHash = {}, customRules = [];
        root = root || { schema: schema, refVal: refVal, refs: refs };
        var c = checkCompiling.call(this, schema, root, baseId);
        var compilation = this._compilations[c.index];
        if (c.compiling)
            return (compilation.callValidate = callValidate);
        var formats = this._formats;
        var RULES = this.RULES;
        try {
            var v = localCompile(schema, root, localRefs, baseId);
            compilation.validate = v;
            var cv = compilation.callValidate;
            if (cv) {
                cv.schema = v.schema;
                cv.errors = null;
                cv.refs = v.refs;
                cv.refVal = v.refVal;
                cv.root = v.root;
                cv.$async = v.$async;
                if (opts.sourceCode)
                    cv.source = v.source;
            }
            return v;
        }
        finally {
            endCompiling.call(this, schema, root, baseId);
        }
        /* @this   {*} - custom context, see passContext option */
        function callValidate() {
            /* jshint validthis: true */
            var validate = compilation.validate;
            var result = validate.apply(this, arguments);
            callValidate.errors = validate.errors;
            return result;
        }
        function localCompile(_schema, _root, localRefs, baseId) {
            var isRoot = !_root || (_root && _root.schema == _schema);
            if (_root.schema != root.schema)
                return compile.call(self, _schema, _root, localRefs, baseId);
            var $async = _schema.$async === true;
            var sourceCode = validateGenerator({
                isTop: true,
                schema: _schema,
                isRoot: isRoot,
                baseId: baseId,
                root: _root,
                schemaPath: '',
                errSchemaPath: '#',
                errorPath: '""',
                MissingRefError: errorClasses.MissingRef,
                RULES: RULES,
                validate: validateGenerator,
                util: util,
                resolve: resolve,
                resolveRef: resolveRef,
                usePattern: usePattern,
                useDefault: useDefault,
                useCustomRule: useCustomRule,
                opts: opts,
                formats: formats,
                logger: self.logger,
                self: self
            });
            sourceCode = vars(refVal, refValCode) + vars(patterns, patternCode)
                + vars(defaults, defaultCode) + vars(customRules, customRuleCode)
                + sourceCode;
            if (opts.processCode)
                sourceCode = opts.processCode(sourceCode, _schema);
            // console.log('\n\n\n *** \n', JSON.stringify(sourceCode));
            var validate;
            try {
                var makeValidate = new Function('self', 'RULES', 'formats', 'root', 'refVal', 'defaults', 'customRules', 'equal', 'ucs2length', 'ValidationError', sourceCode);
                validate = makeValidate(self, RULES, formats, root, refVal, defaults, customRules, equal, ucs2length, ValidationError);
                refVal[0] = validate;
            }
            catch (e) {
                self.logger.error('Error compiling schema, function code:', sourceCode);
                throw e;
            }
            validate.schema = _schema;
            validate.errors = null;
            validate.refs = refs;
            validate.refVal = refVal;
            validate.root = isRoot ? validate : _root;
            if ($async)
                validate.$async = true;
            if (opts.sourceCode === true) {
                validate.source = {
                    code: sourceCode,
                    patterns: patterns,
                    defaults: defaults
                };
            }
            return validate;
        }
        function resolveRef(baseId, ref, isRoot) {
            ref = resolve.url(baseId, ref);
            var refIndex = refs[ref];
            var _refVal, refCode;
            if (refIndex !== undefined) {
                _refVal = refVal[refIndex];
                refCode = 'refVal[' + refIndex + ']';
                return resolvedRef(_refVal, refCode);
            }
            if (!isRoot && root.refs) {
                var rootRefId = root.refs[ref];
                if (rootRefId !== undefined) {
                    _refVal = root.refVal[rootRefId];
                    refCode = addLocalRef(ref, _refVal);
                    return resolvedRef(_refVal, refCode);
                }
            }
            refCode = addLocalRef(ref);
            var v = resolve.call(self, localCompile, root, ref);
            if (v === undefined) {
                var localSchema = localRefs && localRefs[ref];
                if (localSchema) {
                    v = resolve.inlineRef(localSchema, opts.inlineRefs)
                        ? localSchema
                        : compile.call(self, localSchema, root, localRefs, baseId);
                }
            }
            if (v === undefined) {
                removeLocalRef(ref);
            }
            else {
                replaceLocalRef(ref, v);
                return resolvedRef(v, refCode);
            }
        }
        function addLocalRef(ref, v) {
            var refId = refVal.length;
            refVal[refId] = v;
            refs[ref] = refId;
            return 'refVal' + refId;
        }
        function removeLocalRef(ref) {
            delete refs[ref];
        }
        function replaceLocalRef(ref, v) {
            var refId = refs[ref];
            refVal[refId] = v;
        }
        function resolvedRef(refVal, code) {
            return typeof refVal == 'object' || typeof refVal == 'boolean'
                ? { code: code, schema: refVal, inline: true }
                : { code: code, $async: refVal && !!refVal.$async };
        }
        function usePattern(regexStr) {
            var index = patternsHash[regexStr];
            if (index === undefined) {
                index = patternsHash[regexStr] = patterns.length;
                patterns[index] = regexStr;
            }
            return 'pattern' + index;
        }
        function useDefault(value) {
            switch (typeof value) {
                case 'boolean':
                case 'number':
                    return '' + value;
                case 'string':
                    return util.toQuotedString(value);
                case 'object':
                    if (value === null)
                        return 'null';
                    var valueStr = stableStringify(value);
                    var index = defaultsHash[valueStr];
                    if (index === undefined) {
                        index = defaultsHash[valueStr] = defaults.length;
                        defaults[index] = value;
                    }
                    return 'default' + index;
            }
        }
        function useCustomRule(rule, schema, parentSchema, it) {
            if (self._opts.validateSchema !== false) {
                var deps = rule.definition.dependencies;
                if (deps && !deps.every(function (keyword) {
                    return Object.prototype.hasOwnProperty.call(parentSchema, keyword);
                }))
                    throw new Error('parent schema must have all required keywords: ' + deps.join(','));
                var validateSchema = rule.definition.validateSchema;
                if (validateSchema) {
                    var valid = validateSchema(schema);
                    if (!valid) {
                        var message = 'keyword schema is invalid: ' + self.errorsText(validateSchema.errors);
                        if (self._opts.validateSchema == 'log')
                            self.logger.error(message);
                        else
                            throw new Error(message);
                    }
                }
            }
            var compile = rule.definition.compile, inline = rule.definition.inline, macro = rule.definition.macro;
            var validate;
            if (compile) {
                validate = compile.call(self, schema, parentSchema, it);
            }
            else if (macro) {
                validate = macro.call(self, schema, parentSchema, it);
                if (opts.validateSchema !== false)
                    self.validateSchema(validate, true);
            }
            else if (inline) {
                validate = inline.call(self, it, rule.keyword, schema, parentSchema);
            }
            else {
                validate = rule.definition.validate;
                if (!validate)
                    return;
            }
            if (validate === undefined)
                throw new Error('custom keyword "' + rule.keyword + '"failed to compile');
            var index = customRules.length;
            customRules[index] = validate;
            return {
                code: 'customRule' + index,
                validate: validate
            };
        }
    }
    /**
     * Checks if the schema is currently compiled
     * @this   Ajv
     * @param  {Object} schema schema to compile
     * @param  {Object} root root object
     * @param  {String} baseId base schema ID
     * @return {Object} object with properties "index" (compilation index) and "compiling" (boolean)
     */
    function checkCompiling(schema, root, baseId) {
        /* jshint validthis: true */
        var index = compIndex.call(this, schema, root, baseId);
        if (index >= 0)
            return { index: index, compiling: true };
        index = this._compilations.length;
        this._compilations[index] = {
            schema: schema,
            root: root,
            baseId: baseId
        };
        return { index: index, compiling: false };
    }
    /**
     * Removes the schema from the currently compiled list
     * @this   Ajv
     * @param  {Object} schema schema to compile
     * @param  {Object} root root object
     * @param  {String} baseId base schema ID
     */
    function endCompiling(schema, root, baseId) {
        /* jshint validthis: true */
        var i = compIndex.call(this, schema, root, baseId);
        if (i >= 0)
            this._compilations.splice(i, 1);
    }
    /**
     * Index of schema compilation in the currently compiled list
     * @this   Ajv
     * @param  {Object} schema schema to compile
     * @param  {Object} root root object
     * @param  {String} baseId base schema ID
     * @return {Integer} compilation index
     */
    function compIndex(schema, root, baseId) {
        /* jshint validthis: true */
        for (var i = 0; i < this._compilations.length; i++) {
            var c = this._compilations[i];
            if (c.schema == schema && c.root == root && c.baseId == baseId)
                return i;
        }
        return -1;
    }
    function patternCode(i, patterns) {
        return 'var pattern' + i + ' = new RegExp(' + util.toQuotedString(patterns[i]) + ');';
    }
    function defaultCode(i) {
        return 'var default' + i + ' = defaults[' + i + '];';
    }
    function refValCode(i, refVal) {
        return refVal[i] === undefined ? '' : 'var refVal' + i + ' = refVal[' + i + '];';
    }
    function customRuleCode(i) {
        return 'var customRule' + i + ' = customRules[' + i + '];';
    }
    function vars(arr, statement) {
        if (!arr.length)
            return '';
        var code = '';
        for (var i = 0; i < arr.length; i++)
            code += statement(i, arr);
        return code;
    }
},
/* ajv/lib/compile/resolve.js */ function _(require, module, exports, __esModule, __esExport) {
    var URI = require(226) /* uri-js */, equal = require(227) /* fast-deep-equal */, util = require(228) /* ./util */, SchemaObject = require(230) /* ./schema_obj */, traverse = require(231) /* json-schema-traverse */;
    module.exports = resolve;
    resolve.normalizeId = normalizeId;
    resolve.fullPath = getFullPath;
    resolve.url = resolveUrl;
    resolve.ids = resolveIds;
    resolve.inlineRef = inlineRef;
    resolve.schema = resolveSchema;
    /**
     * [resolve and compile the references ($ref)]
     * @this   Ajv
     * @param  {Function} compile reference to schema compilation funciton (localCompile)
     * @param  {Object} root object with information about the root schema for the current schema
     * @param  {String} ref reference to resolve
     * @return {Object|Function} schema object (if the schema can be inlined) or validation function
     */
    function resolve(compile, root, ref) {
        /* jshint validthis: true */
        var refVal = this._refs[ref];
        if (typeof refVal == 'string') {
            if (this._refs[refVal])
                refVal = this._refs[refVal];
            else
                return resolve.call(this, compile, root, refVal);
        }
        refVal = refVal || this._schemas[ref];
        if (refVal instanceof SchemaObject) {
            return inlineRef(refVal.schema, this._opts.inlineRefs)
                ? refVal.schema
                : refVal.validate || this._compile(refVal);
        }
        var res = resolveSchema.call(this, root, ref);
        var schema, v, baseId;
        if (res) {
            schema = res.schema;
            root = res.root;
            baseId = res.baseId;
        }
        if (schema instanceof SchemaObject) {
            v = schema.validate || compile.call(this, schema.schema, root, undefined, baseId);
        }
        else if (schema !== undefined) {
            v = inlineRef(schema, this._opts.inlineRefs)
                ? schema
                : compile.call(this, schema, root, undefined, baseId);
        }
        return v;
    }
    /**
     * Resolve schema, its root and baseId
     * @this Ajv
     * @param  {Object} root root object with properties schema, refVal, refs
     * @param  {String} ref  reference to resolve
     * @return {Object} object with properties schema, root, baseId
     */
    function resolveSchema(root, ref) {
        /* jshint validthis: true */
        var p = URI.parse(ref), refPath = _getFullPath(p), baseId = getFullPath(this._getId(root.schema));
        if (Object.keys(root.schema).length === 0 || refPath !== baseId) {
            var id = normalizeId(refPath);
            var refVal = this._refs[id];
            if (typeof refVal == 'string') {
                return resolveRecursive.call(this, root, refVal, p);
            }
            else if (refVal instanceof SchemaObject) {
                if (!refVal.validate)
                    this._compile(refVal);
                root = refVal;
            }
            else {
                refVal = this._schemas[id];
                if (refVal instanceof SchemaObject) {
                    if (!refVal.validate)
                        this._compile(refVal);
                    if (id == normalizeId(ref))
                        return { schema: refVal, root: root, baseId: baseId };
                    root = refVal;
                }
                else {
                    return;
                }
            }
            if (!root.schema)
                return;
            baseId = getFullPath(this._getId(root.schema));
        }
        return getJsonPointer.call(this, p, baseId, root.schema, root);
    }
    /* @this Ajv */
    function resolveRecursive(root, ref, parsedRef) {
        /* jshint validthis: true */
        var res = resolveSchema.call(this, root, ref);
        if (res) {
            var schema = res.schema;
            var baseId = res.baseId;
            root = res.root;
            var id = this._getId(schema);
            if (id)
                baseId = resolveUrl(baseId, id);
            return getJsonPointer.call(this, parsedRef, baseId, schema, root);
        }
    }
    var PREVENT_SCOPE_CHANGE = util.toHash(['properties', 'patternProperties', 'enum', 'dependencies', 'definitions']);
    /* @this Ajv */
    function getJsonPointer(parsedRef, baseId, schema, root) {
        /* jshint validthis: true */
        parsedRef.fragment = parsedRef.fragment || '';
        if (parsedRef.fragment.slice(0, 1) != '/')
            return;
        var parts = parsedRef.fragment.split('/');
        for (var i = 1; i < parts.length; i++) {
            var part = parts[i];
            if (part) {
                part = util.unescapeFragment(part);
                schema = schema[part];
                if (schema === undefined)
                    break;
                var id;
                if (!PREVENT_SCOPE_CHANGE[part]) {
                    id = this._getId(schema);
                    if (id)
                        baseId = resolveUrl(baseId, id);
                    if (schema.$ref) {
                        var $ref = resolveUrl(baseId, schema.$ref);
                        var res = resolveSchema.call(this, root, $ref);
                        if (res) {
                            schema = res.schema;
                            root = res.root;
                            baseId = res.baseId;
                        }
                    }
                }
            }
        }
        if (schema !== undefined && schema !== root.schema)
            return { schema: schema, root: root, baseId: baseId };
    }
    var SIMPLE_INLINED = util.toHash([
        'type', 'format', 'pattern',
        'maxLength', 'minLength',
        'maxProperties', 'minProperties',
        'maxItems', 'minItems',
        'maximum', 'minimum',
        'uniqueItems', 'multipleOf',
        'required', 'enum'
    ]);
    function inlineRef(schema, limit) {
        if (limit === false)
            return false;
        if (limit === undefined || limit === true)
            return checkNoRef(schema);
        else if (limit)
            return countKeys(schema) <= limit;
    }
    function checkNoRef(schema) {
        var item;
        if (Array.isArray(schema)) {
            for (var i = 0; i < schema.length; i++) {
                item = schema[i];
                if (typeof item == 'object' && !checkNoRef(item))
                    return false;
            }
        }
        else {
            for (var key in schema) {
                if (key == '$ref')
                    return false;
                item = schema[key];
                if (typeof item == 'object' && !checkNoRef(item))
                    return false;
            }
        }
        return true;
    }
    function countKeys(schema) {
        var count = 0, item;
        if (Array.isArray(schema)) {
            for (var i = 0; i < schema.length; i++) {
                item = schema[i];
                if (typeof item == 'object')
                    count += countKeys(item);
                if (count == Infinity)
                    return Infinity;
            }
        }
        else {
            for (var key in schema) {
                if (key == '$ref')
                    return Infinity;
                if (SIMPLE_INLINED[key]) {
                    count++;
                }
                else {
                    item = schema[key];
                    if (typeof item == 'object')
                        count += countKeys(item) + 1;
                    if (count == Infinity)
                        return Infinity;
                }
            }
        }
        return count;
    }
    function getFullPath(id, normalize) {
        if (normalize !== false)
            id = normalizeId(id);
        var p = URI.parse(id);
        return _getFullPath(p);
    }
    function _getFullPath(p) {
        return URI.serialize(p).split('#')[0] + '#';
    }
    var TRAILING_SLASH_HASH = /#\/?$/;
    function normalizeId(id) {
        return id ? id.replace(TRAILING_SLASH_HASH, '') : '';
    }
    function resolveUrl(baseId, id) {
        id = normalizeId(id);
        return URI.resolve(baseId, id);
    }
    /* @this Ajv */
    function resolveIds(schema) {
        var schemaId = normalizeId(this._getId(schema));
        var baseIds = { '': schemaId };
        var fullPaths = { '': getFullPath(schemaId, false) };
        var localRefs = {};
        var self = this;
        traverse(schema, { allKeys: true }, function (sch, jsonPtr, rootSchema, parentJsonPtr, parentKeyword, parentSchema, keyIndex) {
            if (jsonPtr === '')
                return;
            var id = self._getId(sch);
            var baseId = baseIds[parentJsonPtr];
            var fullPath = fullPaths[parentJsonPtr] + '/' + parentKeyword;
            if (keyIndex !== undefined)
                fullPath += '/' + (typeof keyIndex == 'number' ? keyIndex : util.escapeFragment(keyIndex));
            if (typeof id == 'string') {
                id = baseId = normalizeId(baseId ? URI.resolve(baseId, id) : id);
                var refVal = self._refs[id];
                if (typeof refVal == 'string')
                    refVal = self._refs[refVal];
                if (refVal && refVal.schema) {
                    if (!equal(sch, refVal.schema))
                        throw new Error('id "' + id + '" resolves to more than one schema');
                }
                else if (id != normalizeId(fullPath)) {
                    if (id[0] == '#') {
                        if (localRefs[id] && !equal(sch, localRefs[id]))
                            throw new Error('id "' + id + '" resolves to more than one schema');
                        localRefs[id] = sch;
                    }
                    else {
                        self._refs[id] = fullPath;
                    }
                }
            }
            baseIds[jsonPtr] = baseId;
            fullPaths[jsonPtr] = fullPath;
        });
        return localRefs;
    }
},
/* uri-js/dist/es5/uri.all.js */ function _(require, module, exports, __esModule, __esExport) {
    /** @license URI.js v4.4.1 (c) 2011 Gary Court. License: http://github.com/garycourt/uri-js */
    (function (global, factory) {
        typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
            typeof define === 'function' && define.amd ? define(['exports'], factory) :
                (factory((global.URI = global.URI || {})));
    }(this, (function (exports) {
        'use strict';
        function merge() {
            for (var _len = arguments.length, sets = Array(_len), _key = 0; _key < _len; _key++) {
                sets[_key] = arguments[_key];
            }
            if (sets.length > 1) {
                sets[0] = sets[0].slice(0, -1);
                var xl = sets.length - 1;
                for (var x = 1; x < xl; ++x) {
                    sets[x] = sets[x].slice(1, -1);
                }
                sets[xl] = sets[xl].slice(1);
                return sets.join('');
            }
            else {
                return sets[0];
            }
        }
        function subexp(str) {
            return "(?:" + str + ")";
        }
        function typeOf(o) {
            return o === undefined ? "undefined" : o === null ? "null" : Object.prototype.toString.call(o).split(" ").pop().split("]").shift().toLowerCase();
        }
        function toUpperCase(str) {
            return str.toUpperCase();
        }
        function toArray(obj) {
            return obj !== undefined && obj !== null ? obj instanceof Array ? obj : typeof obj.length !== "number" || obj.split || obj.setInterval || obj.call ? [obj] : Array.prototype.slice.call(obj) : [];
        }
        function assign(target, source) {
            var obj = target;
            if (source) {
                for (var key in source) {
                    obj[key] = source[key];
                }
            }
            return obj;
        }
        function buildExps(isIRI) {
            var ALPHA$$ = "[A-Za-z]", CR$ = "[\\x0D]", DIGIT$$ = "[0-9]", DQUOTE$$ = "[\\x22]", HEXDIG$$ = merge(DIGIT$$, "[A-Fa-f]"), 
            //case-insensitive
            LF$$ = "[\\x0A]", SP$$ = "[\\x20]", PCT_ENCODED$ = subexp(subexp("%[EFef]" + HEXDIG$$ + "%" + HEXDIG$$ + HEXDIG$$ + "%" + HEXDIG$$ + HEXDIG$$) + "|" + subexp("%[89A-Fa-f]" + HEXDIG$$ + "%" + HEXDIG$$ + HEXDIG$$) + "|" + subexp("%" + HEXDIG$$ + HEXDIG$$)), 
            //expanded
            GEN_DELIMS$$ = "[\\:\\/\\?\\#\\[\\]\\@]", SUB_DELIMS$$ = "[\\!\\$\\&\\'\\(\\)\\*\\+\\,\\;\\=]", RESERVED$$ = merge(GEN_DELIMS$$, SUB_DELIMS$$), UCSCHAR$$ = isIRI ? "[\\xA0-\\u200D\\u2010-\\u2029\\u202F-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF]" : "[]", 
            //subset, excludes bidi control characters
            IPRIVATE$$ = isIRI ? "[\\uE000-\\uF8FF]" : "[]", 
            //subset
            UNRESERVED$$ = merge(ALPHA$$, DIGIT$$, "[\\-\\.\\_\\~]", UCSCHAR$$), SCHEME$ = subexp(ALPHA$$ + merge(ALPHA$$, DIGIT$$, "[\\+\\-\\.]") + "*"), USERINFO$ = subexp(subexp(PCT_ENCODED$ + "|" + merge(UNRESERVED$$, SUB_DELIMS$$, "[\\:]")) + "*"), DEC_OCTET$ = subexp(subexp("25[0-5]") + "|" + subexp("2[0-4]" + DIGIT$$) + "|" + subexp("1" + DIGIT$$ + DIGIT$$) + "|" + subexp("[1-9]" + DIGIT$$) + "|" + DIGIT$$), DEC_OCTET_RELAXED$ = subexp(subexp("25[0-5]") + "|" + subexp("2[0-4]" + DIGIT$$) + "|" + subexp("1" + DIGIT$$ + DIGIT$$) + "|" + subexp("0?[1-9]" + DIGIT$$) + "|0?0?" + DIGIT$$), 
            //relaxed parsing rules
            IPV4ADDRESS$ = subexp(DEC_OCTET_RELAXED$ + "\\." + DEC_OCTET_RELAXED$ + "\\." + DEC_OCTET_RELAXED$ + "\\." + DEC_OCTET_RELAXED$), H16$ = subexp(HEXDIG$$ + "{1,4}"), LS32$ = subexp(subexp(H16$ + "\\:" + H16$) + "|" + IPV4ADDRESS$), IPV6ADDRESS1$ = subexp(subexp(H16$ + "\\:") + "{6}" + LS32$), 
            //                           6( h16 ":" ) ls32
            IPV6ADDRESS2$ = subexp("\\:\\:" + subexp(H16$ + "\\:") + "{5}" + LS32$), 
            //                      "::" 5( h16 ":" ) ls32
            IPV6ADDRESS3$ = subexp(subexp(H16$) + "?\\:\\:" + subexp(H16$ + "\\:") + "{4}" + LS32$), 
            //[               h16 ] "::" 4( h16 ":" ) ls32
            IPV6ADDRESS4$ = subexp(subexp(subexp(H16$ + "\\:") + "{0,1}" + H16$) + "?\\:\\:" + subexp(H16$ + "\\:") + "{3}" + LS32$), 
            //[ *1( h16 ":" ) h16 ] "::" 3( h16 ":" ) ls32
            IPV6ADDRESS5$ = subexp(subexp(subexp(H16$ + "\\:") + "{0,2}" + H16$) + "?\\:\\:" + subexp(H16$ + "\\:") + "{2}" + LS32$), 
            //[ *2( h16 ":" ) h16 ] "::" 2( h16 ":" ) ls32
            IPV6ADDRESS6$ = subexp(subexp(subexp(H16$ + "\\:") + "{0,3}" + H16$) + "?\\:\\:" + H16$ + "\\:" + LS32$), 
            //[ *3( h16 ":" ) h16 ] "::"    h16 ":"   ls32
            IPV6ADDRESS7$ = subexp(subexp(subexp(H16$ + "\\:") + "{0,4}" + H16$) + "?\\:\\:" + LS32$), 
            //[ *4( h16 ":" ) h16 ] "::"              ls32
            IPV6ADDRESS8$ = subexp(subexp(subexp(H16$ + "\\:") + "{0,5}" + H16$) + "?\\:\\:" + H16$), 
            //[ *5( h16 ":" ) h16 ] "::"              h16
            IPV6ADDRESS9$ = subexp(subexp(subexp(H16$ + "\\:") + "{0,6}" + H16$) + "?\\:\\:"), 
            //[ *6( h16 ":" ) h16 ] "::"
            IPV6ADDRESS$ = subexp([IPV6ADDRESS1$, IPV6ADDRESS2$, IPV6ADDRESS3$, IPV6ADDRESS4$, IPV6ADDRESS5$, IPV6ADDRESS6$, IPV6ADDRESS7$, IPV6ADDRESS8$, IPV6ADDRESS9$].join("|")), ZONEID$ = subexp(subexp(UNRESERVED$$ + "|" + PCT_ENCODED$) + "+"), 
            //RFC 6874
            IPV6ADDRZ$ = subexp(IPV6ADDRESS$ + "\\%25" + ZONEID$), 
            //RFC 6874
            IPV6ADDRZ_RELAXED$ = subexp(IPV6ADDRESS$ + subexp("\\%25|\\%(?!" + HEXDIG$$ + "{2})") + ZONEID$), 
            //RFC 6874, with relaxed parsing rules
            IPVFUTURE$ = subexp("[vV]" + HEXDIG$$ + "+\\." + merge(UNRESERVED$$, SUB_DELIMS$$, "[\\:]") + "+"), IP_LITERAL$ = subexp("\\[" + subexp(IPV6ADDRZ_RELAXED$ + "|" + IPV6ADDRESS$ + "|" + IPVFUTURE$) + "\\]"), 
            //RFC 6874
            REG_NAME$ = subexp(subexp(PCT_ENCODED$ + "|" + merge(UNRESERVED$$, SUB_DELIMS$$)) + "*"), HOST$ = subexp(IP_LITERAL$ + "|" + IPV4ADDRESS$ + "(?!" + REG_NAME$ + ")" + "|" + REG_NAME$), PORT$ = subexp(DIGIT$$ + "*"), AUTHORITY$ = subexp(subexp(USERINFO$ + "@") + "?" + HOST$ + subexp("\\:" + PORT$) + "?"), PCHAR$ = subexp(PCT_ENCODED$ + "|" + merge(UNRESERVED$$, SUB_DELIMS$$, "[\\:\\@]")), SEGMENT$ = subexp(PCHAR$ + "*"), SEGMENT_NZ$ = subexp(PCHAR$ + "+"), SEGMENT_NZ_NC$ = subexp(subexp(PCT_ENCODED$ + "|" + merge(UNRESERVED$$, SUB_DELIMS$$, "[\\@]")) + "+"), PATH_ABEMPTY$ = subexp(subexp("\\/" + SEGMENT$) + "*"), PATH_ABSOLUTE$ = subexp("\\/" + subexp(SEGMENT_NZ$ + PATH_ABEMPTY$) + "?"), 
            //simplified
            PATH_NOSCHEME$ = subexp(SEGMENT_NZ_NC$ + PATH_ABEMPTY$), 
            //simplified
            PATH_ROOTLESS$ = subexp(SEGMENT_NZ$ + PATH_ABEMPTY$), 
            //simplified
            PATH_EMPTY$ = "(?!" + PCHAR$ + ")", PATH$ = subexp(PATH_ABEMPTY$ + "|" + PATH_ABSOLUTE$ + "|" + PATH_NOSCHEME$ + "|" + PATH_ROOTLESS$ + "|" + PATH_EMPTY$), QUERY$ = subexp(subexp(PCHAR$ + "|" + merge("[\\/\\?]", IPRIVATE$$)) + "*"), FRAGMENT$ = subexp(subexp(PCHAR$ + "|[\\/\\?]") + "*"), HIER_PART$ = subexp(subexp("\\/\\/" + AUTHORITY$ + PATH_ABEMPTY$) + "|" + PATH_ABSOLUTE$ + "|" + PATH_ROOTLESS$ + "|" + PATH_EMPTY$), URI$ = subexp(SCHEME$ + "\\:" + HIER_PART$ + subexp("\\?" + QUERY$) + "?" + subexp("\\#" + FRAGMENT$) + "?"), RELATIVE_PART$ = subexp(subexp("\\/\\/" + AUTHORITY$ + PATH_ABEMPTY$) + "|" + PATH_ABSOLUTE$ + "|" + PATH_NOSCHEME$ + "|" + PATH_EMPTY$), RELATIVE$ = subexp(RELATIVE_PART$ + subexp("\\?" + QUERY$) + "?" + subexp("\\#" + FRAGMENT$) + "?"), URI_REFERENCE$ = subexp(URI$ + "|" + RELATIVE$), ABSOLUTE_URI$ = subexp(SCHEME$ + "\\:" + HIER_PART$ + subexp("\\?" + QUERY$) + "?"), GENERIC_REF$ = "^(" + SCHEME$ + ")\\:" + subexp(subexp("\\/\\/(" + subexp("(" + USERINFO$ + ")@") + "?(" + HOST$ + ")" + subexp("\\:(" + PORT$ + ")") + "?)") + "?(" + PATH_ABEMPTY$ + "|" + PATH_ABSOLUTE$ + "|" + PATH_ROOTLESS$ + "|" + PATH_EMPTY$ + ")") + subexp("\\?(" + QUERY$ + ")") + "?" + subexp("\\#(" + FRAGMENT$ + ")") + "?$", RELATIVE_REF$ = "^(){0}" + subexp(subexp("\\/\\/(" + subexp("(" + USERINFO$ + ")@") + "?(" + HOST$ + ")" + subexp("\\:(" + PORT$ + ")") + "?)") + "?(" + PATH_ABEMPTY$ + "|" + PATH_ABSOLUTE$ + "|" + PATH_NOSCHEME$ + "|" + PATH_EMPTY$ + ")") + subexp("\\?(" + QUERY$ + ")") + "?" + subexp("\\#(" + FRAGMENT$ + ")") + "?$", ABSOLUTE_REF$ = "^(" + SCHEME$ + ")\\:" + subexp(subexp("\\/\\/(" + subexp("(" + USERINFO$ + ")@") + "?(" + HOST$ + ")" + subexp("\\:(" + PORT$ + ")") + "?)") + "?(" + PATH_ABEMPTY$ + "|" + PATH_ABSOLUTE$ + "|" + PATH_ROOTLESS$ + "|" + PATH_EMPTY$ + ")") + subexp("\\?(" + QUERY$ + ")") + "?$", SAMEDOC_REF$ = "^" + subexp("\\#(" + FRAGMENT$ + ")") + "?$", AUTHORITY_REF$ = "^" + subexp("(" + USERINFO$ + ")@") + "?(" + HOST$ + ")" + subexp("\\:(" + PORT$ + ")") + "?$";
            return {
                NOT_SCHEME: new RegExp(merge("[^]", ALPHA$$, DIGIT$$, "[\\+\\-\\.]"), "g"),
                NOT_USERINFO: new RegExp(merge("[^\\%\\:]", UNRESERVED$$, SUB_DELIMS$$), "g"),
                NOT_HOST: new RegExp(merge("[^\\%\\[\\]\\:]", UNRESERVED$$, SUB_DELIMS$$), "g"),
                NOT_PATH: new RegExp(merge("[^\\%\\/\\:\\@]", UNRESERVED$$, SUB_DELIMS$$), "g"),
                NOT_PATH_NOSCHEME: new RegExp(merge("[^\\%\\/\\@]", UNRESERVED$$, SUB_DELIMS$$), "g"),
                NOT_QUERY: new RegExp(merge("[^\\%]", UNRESERVED$$, SUB_DELIMS$$, "[\\:\\@\\/\\?]", IPRIVATE$$), "g"),
                NOT_FRAGMENT: new RegExp(merge("[^\\%]", UNRESERVED$$, SUB_DELIMS$$, "[\\:\\@\\/\\?]"), "g"),
                ESCAPE: new RegExp(merge("[^]", UNRESERVED$$, SUB_DELIMS$$), "g"),
                UNRESERVED: new RegExp(UNRESERVED$$, "g"),
                OTHER_CHARS: new RegExp(merge("[^\\%]", UNRESERVED$$, RESERVED$$), "g"),
                PCT_ENCODED: new RegExp(PCT_ENCODED$, "g"),
                IPV4ADDRESS: new RegExp("^(" + IPV4ADDRESS$ + ")$"),
                IPV6ADDRESS: new RegExp("^\\[?(" + IPV6ADDRESS$ + ")" + subexp(subexp("\\%25|\\%(?!" + HEXDIG$$ + "{2})") + "(" + ZONEID$ + ")") + "?\\]?$") //RFC 6874, with relaxed parsing rules
            };
        }
        var URI_PROTOCOL = buildExps(false);
        var IRI_PROTOCOL = buildExps(true);
        var slicedToArray = function () {
            function sliceIterator(arr, i) {
                var _arr = [];
                var _n = true;
                var _d = false;
                var _e = undefined;
                try {
                    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
                        _arr.push(_s.value);
                        if (i && _arr.length === i)
                            break;
                    }
                }
                catch (err) {
                    _d = true;
                    _e = err;
                }
                finally {
                    try {
                        if (!_n && _i["return"])
                            _i["return"]();
                    }
                    finally {
                        if (_d)
                            throw _e;
                    }
                }
                return _arr;
            }
            return function (arr, i) {
                if (Array.isArray(arr)) {
                    return arr;
                }
                else if (Symbol.iterator in Object(arr)) {
                    return sliceIterator(arr, i);
                }
                else {
                    throw new TypeError("Invalid attempt to destructure non-iterable instance");
                }
            };
        }();
        var toConsumableArray = function (arr) {
            if (Array.isArray(arr)) {
                for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++)
                    arr2[i] = arr[i];
                return arr2;
            }
            else {
                return Array.from(arr);
            }
        };
        /** Highest positive signed 32-bit float value */
        var maxInt = 2147483647; // aka. 0x7FFFFFFF or 2^31-1
        /** Bootstring parameters */
        var base = 36;
        var tMin = 1;
        var tMax = 26;
        var skew = 38;
        var damp = 700;
        var initialBias = 72;
        var initialN = 128; // 0x80
        var delimiter = '-'; // '\x2D'
        /** Regular expressions */
        var regexPunycode = /^xn--/;
        var regexNonASCII = /[^\0-\x7E]/; // non-ASCII chars
        var regexSeparators = /[\x2E\u3002\uFF0E\uFF61]/g; // RFC 3490 separators
        /** Error messages */
        var errors = {
            'overflow': 'Overflow: input needs wider integers to process',
            'not-basic': 'Illegal input >= 0x80 (not a basic code point)',
            'invalid-input': 'Invalid input'
        };
        /** Convenience shortcuts */
        var baseMinusTMin = base - tMin;
        var floor = Math.floor;
        var stringFromCharCode = String.fromCharCode;
        /*--------------------------------------------------------------------------*/
        /**
         * A generic error utility function.
         * @private
         * @param {String} type The error type.
         * @returns {Error} Throws a `RangeError` with the applicable error message.
         */
        function error$1(type) {
            throw new RangeError(errors[type]);
        }
        /**
         * A generic `Array#map` utility function.
         * @private
         * @param {Array} array The array to iterate over.
         * @param {Function} callback The function that gets called for every array
         * item.
         * @returns {Array} A new array of values returned by the callback function.
         */
        function map(array, fn) {
            var result = [];
            var length = array.length;
            while (length--) {
                result[length] = fn(array[length]);
            }
            return result;
        }
        /**
         * A simple `Array#map`-like wrapper to work with domain name strings or email
         * addresses.
         * @private
         * @param {String} domain The domain name or email address.
         * @param {Function} callback The function that gets called for every
         * character.
         * @returns {Array} A new string of characters returned by the callback
         * function.
         */
        function mapDomain(string, fn) {
            var parts = string.split('@');
            var result = '';
            if (parts.length > 1) {
                // In email addresses, only the domain name should be punycoded. Leave
                // the local part (i.e. everything up to `@`) intact.
                result = parts[0] + '@';
                string = parts[1];
            }
            // Avoid `split(regex)` for IE8 compatibility. See #17.
            string = string.replace(regexSeparators, '\x2E');
            var labels = string.split('.');
            var encoded = map(labels, fn).join('.');
            return result + encoded;
        }
        /**
         * Creates an array containing the numeric code points of each Unicode
         * character in the string. While JavaScript uses UCS-2 internally,
         * this function will convert a pair of surrogate halves (each of which
         * UCS-2 exposes as separate characters) into a single code point,
         * matching UTF-16.
         * @see `punycode.ucs2.encode`
         * @see <https://mathiasbynens.be/notes/javascript-encoding>
         * @memberOf punycode.ucs2
         * @name decode
         * @param {String} string The Unicode input string (UCS-2).
         * @returns {Array} The new array of code points.
         */
        function ucs2decode(string) {
            var output = [];
            var counter = 0;
            var length = string.length;
            while (counter < length) {
                var value = string.charCodeAt(counter++);
                if (value >= 0xD800 && value <= 0xDBFF && counter < length) {
                    // It's a high surrogate, and there is a next character.
                    var extra = string.charCodeAt(counter++);
                    if ((extra & 0xFC00) == 0xDC00) {
                        // Low surrogate.
                        output.push(((value & 0x3FF) << 10) + (extra & 0x3FF) + 0x10000);
                    }
                    else {
                        // It's an unmatched surrogate; only append this code unit, in case the
                        // next code unit is the high surrogate of a surrogate pair.
                        output.push(value);
                        counter--;
                    }
                }
                else {
                    output.push(value);
                }
            }
            return output;
        }
        /**
         * Creates a string based on an array of numeric code points.
         * @see `punycode.ucs2.decode`
         * @memberOf punycode.ucs2
         * @name encode
         * @param {Array} codePoints The array of numeric code points.
         * @returns {String} The new Unicode string (UCS-2).
         */
        var ucs2encode = function ucs2encode(array) {
            return String.fromCodePoint.apply(String, toConsumableArray(array));
        };
        /**
         * Converts a basic code point into a digit/integer.
         * @see `digitToBasic()`
         * @private
         * @param {Number} codePoint The basic numeric code point value.
         * @returns {Number} The numeric value of a basic code point (for use in
         * representing integers) in the range `0` to `base - 1`, or `base` if
         * the code point does not represent a value.
         */
        var basicToDigit = function basicToDigit(codePoint) {
            if (codePoint - 0x30 < 0x0A) {
                return codePoint - 0x16;
            }
            if (codePoint - 0x41 < 0x1A) {
                return codePoint - 0x41;
            }
            if (codePoint - 0x61 < 0x1A) {
                return codePoint - 0x61;
            }
            return base;
        };
        /**
         * Converts a digit/integer into a basic code point.
         * @see `basicToDigit()`
         * @private
         * @param {Number} digit The numeric value of a basic code point.
         * @returns {Number} The basic code point whose value (when used for
         * representing integers) is `digit`, which needs to be in the range
         * `0` to `base - 1`. If `flag` is non-zero, the uppercase form is
         * used; else, the lowercase form is used. The behavior is undefined
         * if `flag` is non-zero and `digit` has no uppercase form.
         */
        var digitToBasic = function digitToBasic(digit, flag) {
            //  0..25 map to ASCII a..z or A..Z
            // 26..35 map to ASCII 0..9
            return digit + 22 + 75 * (digit < 26) - ((flag != 0) << 5);
        };
        /**
         * Bias adaptation function as per section 3.4 of RFC 3492.
         * https://tools.ietf.org/html/rfc3492#section-3.4
         * @private
         */
        var adapt = function adapt(delta, numPoints, firstTime) {
            var k = 0;
            delta = firstTime ? floor(delta / damp) : delta >> 1;
            delta += floor(delta / numPoints);
            for (; /* no initialization */ delta > baseMinusTMin * tMax >> 1; k += base) {
                delta = floor(delta / baseMinusTMin);
            }
            return floor(k + (baseMinusTMin + 1) * delta / (delta + skew));
        };
        /**
         * Converts a Punycode string of ASCII-only symbols to a string of Unicode
         * symbols.
         * @memberOf punycode
         * @param {String} input The Punycode string of ASCII-only symbols.
         * @returns {String} The resulting string of Unicode symbols.
         */
        var decode = function decode(input) {
            // Don't use UCS-2.
            var output = [];
            var inputLength = input.length;
            var i = 0;
            var n = initialN;
            var bias = initialBias;
            // Handle the basic code points: let `basic` be the number of input code
            // points before the last delimiter, or `0` if there is none, then copy
            // the first basic code points to the output.
            var basic = input.lastIndexOf(delimiter);
            if (basic < 0) {
                basic = 0;
            }
            for (var j = 0; j < basic; ++j) {
                // if it's not a basic code point
                if (input.charCodeAt(j) >= 0x80) {
                    error$1('not-basic');
                }
                output.push(input.charCodeAt(j));
            }
            // Main decoding loop: start just after the last delimiter if any basic code
            // points were copied; start at the beginning otherwise.
            for (var index = basic > 0 ? basic + 1 : 0; index < inputLength;) /* no final expression */ {
                // `index` is the index of the next character to be consumed.
                // Decode a generalized variable-length integer into `delta`,
                // which gets added to `i`. The overflow checking is easier
                // if we increase `i` as we go, then subtract off its starting
                // value at the end to obtain `delta`.
                var oldi = i;
                for (var w = 1, k = base;; /* no condition */ k += base) {
                    if (index >= inputLength) {
                        error$1('invalid-input');
                    }
                    var digit = basicToDigit(input.charCodeAt(index++));
                    if (digit >= base || digit > floor((maxInt - i) / w)) {
                        error$1('overflow');
                    }
                    i += digit * w;
                    var t = k <= bias ? tMin : k >= bias + tMax ? tMax : k - bias;
                    if (digit < t) {
                        break;
                    }
                    var baseMinusT = base - t;
                    if (w > floor(maxInt / baseMinusT)) {
                        error$1('overflow');
                    }
                    w *= baseMinusT;
                }
                var out = output.length + 1;
                bias = adapt(i - oldi, out, oldi == 0);
                // `i` was supposed to wrap around from `out` to `0`,
                // incrementing `n` each time, so we'll fix that now:
                if (floor(i / out) > maxInt - n) {
                    error$1('overflow');
                }
                n += floor(i / out);
                i %= out;
                // Insert `n` at position `i` of the output.
                output.splice(i++, 0, n);
            }
            return String.fromCodePoint.apply(String, output);
        };
        /**
         * Converts a string of Unicode symbols (e.g. a domain name label) to a
         * Punycode string of ASCII-only symbols.
         * @memberOf punycode
         * @param {String} input The string of Unicode symbols.
         * @returns {String} The resulting Punycode string of ASCII-only symbols.
         */
        var encode = function encode(input) {
            var output = [];
            // Convert the input in UCS-2 to an array of Unicode code points.
            input = ucs2decode(input);
            // Cache the length.
            var inputLength = input.length;
            // Initialize the state.
            var n = initialN;
            var delta = 0;
            var bias = initialBias;
            // Handle the basic code points.
            var _iteratorNormalCompletion = true;
            var _didIteratorError = false;
            var _iteratorError = undefined;
            try {
                for (var _iterator = input[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                    var _currentValue2 = _step.value;
                    if (_currentValue2 < 0x80) {
                        output.push(stringFromCharCode(_currentValue2));
                    }
                }
            }
            catch (err) {
                _didIteratorError = true;
                _iteratorError = err;
            }
            finally {
                try {
                    if (!_iteratorNormalCompletion && _iterator.return) {
                        _iterator.return();
                    }
                }
                finally {
                    if (_didIteratorError) {
                        throw _iteratorError;
                    }
                }
            }
            var basicLength = output.length;
            var handledCPCount = basicLength;
            // `handledCPCount` is the number of code points that have been handled;
            // `basicLength` is the number of basic code points.
            // Finish the basic string with a delimiter unless it's empty.
            if (basicLength) {
                output.push(delimiter);
            }
            // Main encoding loop:
            while (handledCPCount < inputLength) {
                // All non-basic code points < n have been handled already. Find the next
                // larger one:
                var m = maxInt;
                var _iteratorNormalCompletion2 = true;
                var _didIteratorError2 = false;
                var _iteratorError2 = undefined;
                try {
                    for (var _iterator2 = input[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                        var currentValue = _step2.value;
                        if (currentValue >= n && currentValue < m) {
                            m = currentValue;
                        }
                    }
                    // Increase `delta` enough to advance the decoder's <n,i> state to <m,0>,
                    // but guard against overflow.
                }
                catch (err) {
                    _didIteratorError2 = true;
                    _iteratorError2 = err;
                }
                finally {
                    try {
                        if (!_iteratorNormalCompletion2 && _iterator2.return) {
                            _iterator2.return();
                        }
                    }
                    finally {
                        if (_didIteratorError2) {
                            throw _iteratorError2;
                        }
                    }
                }
                var handledCPCountPlusOne = handledCPCount + 1;
                if (m - n > floor((maxInt - delta) / handledCPCountPlusOne)) {
                    error$1('overflow');
                }
                delta += (m - n) * handledCPCountPlusOne;
                n = m;
                var _iteratorNormalCompletion3 = true;
                var _didIteratorError3 = false;
                var _iteratorError3 = undefined;
                try {
                    for (var _iterator3 = input[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                        var _currentValue = _step3.value;
                        if (_currentValue < n && ++delta > maxInt) {
                            error$1('overflow');
                        }
                        if (_currentValue == n) {
                            // Represent delta as a generalized variable-length integer.
                            var q = delta;
                            for (var k = base;; /* no condition */ k += base) {
                                var t = k <= bias ? tMin : k >= bias + tMax ? tMax : k - bias;
                                if (q < t) {
                                    break;
                                }
                                var qMinusT = q - t;
                                var baseMinusT = base - t;
                                output.push(stringFromCharCode(digitToBasic(t + qMinusT % baseMinusT, 0)));
                                q = floor(qMinusT / baseMinusT);
                            }
                            output.push(stringFromCharCode(digitToBasic(q, 0)));
                            bias = adapt(delta, handledCPCountPlusOne, handledCPCount == basicLength);
                            delta = 0;
                            ++handledCPCount;
                        }
                    }
                }
                catch (err) {
                    _didIteratorError3 = true;
                    _iteratorError3 = err;
                }
                finally {
                    try {
                        if (!_iteratorNormalCompletion3 && _iterator3.return) {
                            _iterator3.return();
                        }
                    }
                    finally {
                        if (_didIteratorError3) {
                            throw _iteratorError3;
                        }
                    }
                }
                ++delta;
                ++n;
            }
            return output.join('');
        };
        /**
         * Converts a Punycode string representing a domain name or an email address
         * to Unicode. Only the Punycoded parts of the input will be converted, i.e.
         * it doesn't matter if you call it on a string that has already been
         * converted to Unicode.
         * @memberOf punycode
         * @param {String} input The Punycoded domain name or email address to
         * convert to Unicode.
         * @returns {String} The Unicode representation of the given Punycode
         * string.
         */
        var toUnicode = function toUnicode(input) {
            return mapDomain(input, function (string) {
                return regexPunycode.test(string) ? decode(string.slice(4).toLowerCase()) : string;
            });
        };
        /**
         * Converts a Unicode string representing a domain name or an email address to
         * Punycode. Only the non-ASCII parts of the domain name will be converted,
         * i.e. it doesn't matter if you call it with a domain that's already in
         * ASCII.
         * @memberOf punycode
         * @param {String} input The domain name or email address to convert, as a
         * Unicode string.
         * @returns {String} The Punycode representation of the given domain name or
         * email address.
         */
        var toASCII = function toASCII(input) {
            return mapDomain(input, function (string) {
                return regexNonASCII.test(string) ? 'xn--' + encode(string) : string;
            });
        };
        /*--------------------------------------------------------------------------*/
        /** Define the public API */
        var punycode = {
            /**
          * A string representing the current Punycode.js version number.
          * @memberOf punycode
          * @type String
          */
            'version': '2.1.0',
            /**
          * An object of methods to convert from JavaScript's internal character
          * representation (UCS-2) to Unicode code points, and back.
          * @see <https://mathiasbynens.be/notes/javascript-encoding>
          * @memberOf punycode
          * @type Object
          */
            'ucs2': {
                'decode': ucs2decode,
                'encode': ucs2encode
            },
            'decode': decode,
            'encode': encode,
            'toASCII': toASCII,
            'toUnicode': toUnicode
        };
        /**
         * URI.js
         *
         * @fileoverview An RFC 3986 compliant, scheme extendable URI parsing/validating/resolving library for JavaScript.
         * @author <a href="mailto:gary.court@gmail.com">Gary Court</a>
         * @see http://github.com/garycourt/uri-js
         */
        /**
         * Copyright 2011 Gary Court. All rights reserved.
         *
         * Redistribution and use in source and binary forms, with or without modification, are
         * permitted provided that the following conditions are met:
         *
         *    1. Redistributions of source code must retain the above copyright notice, this list of
         *       conditions and the following disclaimer.
         *
         *    2. Redistributions in binary form must reproduce the above copyright notice, this list
         *       of conditions and the following disclaimer in the documentation and/or other materials
         *       provided with the distribution.
         *
         * THIS SOFTWARE IS PROVIDED BY GARY COURT ``AS IS'' AND ANY EXPRESS OR IMPLIED
         * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
         * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GARY COURT OR
         * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
         * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
         * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
         * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
         * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
         * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
         *
         * The views and conclusions contained in the software and documentation are those of the
         * authors and should not be interpreted as representing official policies, either expressed
         * or implied, of Gary Court.
         */
        var SCHEMES = {};
        function pctEncChar(chr) {
            var c = chr.charCodeAt(0);
            var e = void 0;
            if (c < 16)
                e = "%0" + c.toString(16).toUpperCase();
            else if (c < 128)
                e = "%" + c.toString(16).toUpperCase();
            else if (c < 2048)
                e = "%" + (c >> 6 | 192).toString(16).toUpperCase() + "%" + (c & 63 | 128).toString(16).toUpperCase();
            else
                e = "%" + (c >> 12 | 224).toString(16).toUpperCase() + "%" + (c >> 6 & 63 | 128).toString(16).toUpperCase() + "%" + (c & 63 | 128).toString(16).toUpperCase();
            return e;
        }
        function pctDecChars(str) {
            var newStr = "";
            var i = 0;
            var il = str.length;
            while (i < il) {
                var c = parseInt(str.substr(i + 1, 2), 16);
                if (c < 128) {
                    newStr += String.fromCharCode(c);
                    i += 3;
                }
                else if (c >= 194 && c < 224) {
                    if (il - i >= 6) {
                        var c2 = parseInt(str.substr(i + 4, 2), 16);
                        newStr += String.fromCharCode((c & 31) << 6 | c2 & 63);
                    }
                    else {
                        newStr += str.substr(i, 6);
                    }
                    i += 6;
                }
                else if (c >= 224) {
                    if (il - i >= 9) {
                        var _c = parseInt(str.substr(i + 4, 2), 16);
                        var c3 = parseInt(str.substr(i + 7, 2), 16);
                        newStr += String.fromCharCode((c & 15) << 12 | (_c & 63) << 6 | c3 & 63);
                    }
                    else {
                        newStr += str.substr(i, 9);
                    }
                    i += 9;
                }
                else {
                    newStr += str.substr(i, 3);
                    i += 3;
                }
            }
            return newStr;
        }
        function _normalizeComponentEncoding(components, protocol) {
            function decodeUnreserved(str) {
                var decStr = pctDecChars(str);
                return !decStr.match(protocol.UNRESERVED) ? str : decStr;
            }
            if (components.scheme)
                components.scheme = String(components.scheme).replace(protocol.PCT_ENCODED, decodeUnreserved).toLowerCase().replace(protocol.NOT_SCHEME, "");
            if (components.userinfo !== undefined)
                components.userinfo = String(components.userinfo).replace(protocol.PCT_ENCODED, decodeUnreserved).replace(protocol.NOT_USERINFO, pctEncChar).replace(protocol.PCT_ENCODED, toUpperCase);
            if (components.host !== undefined)
                components.host = String(components.host).replace(protocol.PCT_ENCODED, decodeUnreserved).toLowerCase().replace(protocol.NOT_HOST, pctEncChar).replace(protocol.PCT_ENCODED, toUpperCase);
            if (components.path !== undefined)
                components.path = String(components.path).replace(protocol.PCT_ENCODED, decodeUnreserved).replace(components.scheme ? protocol.NOT_PATH : protocol.NOT_PATH_NOSCHEME, pctEncChar).replace(protocol.PCT_ENCODED, toUpperCase);
            if (components.query !== undefined)
                components.query = String(components.query).replace(protocol.PCT_ENCODED, decodeUnreserved).replace(protocol.NOT_QUERY, pctEncChar).replace(protocol.PCT_ENCODED, toUpperCase);
            if (components.fragment !== undefined)
                components.fragment = String(components.fragment).replace(protocol.PCT_ENCODED, decodeUnreserved).replace(protocol.NOT_FRAGMENT, pctEncChar).replace(protocol.PCT_ENCODED, toUpperCase);
            return components;
        }
        function _stripLeadingZeros(str) {
            return str.replace(/^0*(.*)/, "$1") || "0";
        }
        function _normalizeIPv4(host, protocol) {
            var matches = host.match(protocol.IPV4ADDRESS) || [];
            var _matches = slicedToArray(matches, 2), address = _matches[1];
            if (address) {
                return address.split(".").map(_stripLeadingZeros).join(".");
            }
            else {
                return host;
            }
        }
        function _normalizeIPv6(host, protocol) {
            var matches = host.match(protocol.IPV6ADDRESS) || [];
            var _matches2 = slicedToArray(matches, 3), address = _matches2[1], zone = _matches2[2];
            if (address) {
                var _address$toLowerCase$ = address.toLowerCase().split('::').reverse(), _address$toLowerCase$2 = slicedToArray(_address$toLowerCase$, 2), last = _address$toLowerCase$2[0], first = _address$toLowerCase$2[1];
                var firstFields = first ? first.split(":").map(_stripLeadingZeros) : [];
                var lastFields = last.split(":").map(_stripLeadingZeros);
                var isLastFieldIPv4Address = protocol.IPV4ADDRESS.test(lastFields[lastFields.length - 1]);
                var fieldCount = isLastFieldIPv4Address ? 7 : 8;
                var lastFieldsStart = lastFields.length - fieldCount;
                var fields = Array(fieldCount);
                for (var x = 0; x < fieldCount; ++x) {
                    fields[x] = firstFields[x] || lastFields[lastFieldsStart + x] || '';
                }
                if (isLastFieldIPv4Address) {
                    fields[fieldCount - 1] = _normalizeIPv4(fields[fieldCount - 1], protocol);
                }
                var allZeroFields = fields.reduce(function (acc, field, index) {
                    if (!field || field === "0") {
                        var lastLongest = acc[acc.length - 1];
                        if (lastLongest && lastLongest.index + lastLongest.length === index) {
                            lastLongest.length++;
                        }
                        else {
                            acc.push({ index: index, length: 1 });
                        }
                    }
                    return acc;
                }, []);
                var longestZeroFields = allZeroFields.sort(function (a, b) {
                    return b.length - a.length;
                })[0];
                var newHost = void 0;
                if (longestZeroFields && longestZeroFields.length > 1) {
                    var newFirst = fields.slice(0, longestZeroFields.index);
                    var newLast = fields.slice(longestZeroFields.index + longestZeroFields.length);
                    newHost = newFirst.join(":") + "::" + newLast.join(":");
                }
                else {
                    newHost = fields.join(":");
                }
                if (zone) {
                    newHost += "%" + zone;
                }
                return newHost;
            }
            else {
                return host;
            }
        }
        var URI_PARSE = /^(?:([^:\/?#]+):)?(?:\/\/((?:([^\/?#@]*)@)?(\[[^\/?#\]]+\]|[^\/?#:]*)(?:\:(\d*))?))?([^?#]*)(?:\?([^#]*))?(?:#((?:.|\n|\r)*))?/i;
        var NO_MATCH_IS_UNDEFINED = "".match(/(){0}/)[1] === undefined;
        function parse(uriString) {
            var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
            var components = {};
            var protocol = options.iri !== false ? IRI_PROTOCOL : URI_PROTOCOL;
            if (options.reference === "suffix")
                uriString = (options.scheme ? options.scheme + ":" : "") + "//" + uriString;
            var matches = uriString.match(URI_PARSE);
            if (matches) {
                if (NO_MATCH_IS_UNDEFINED) {
                    //store each component
                    components.scheme = matches[1];
                    components.userinfo = matches[3];
                    components.host = matches[4];
                    components.port = parseInt(matches[5], 10);
                    components.path = matches[6] || "";
                    components.query = matches[7];
                    components.fragment = matches[8];
                    //fix port number
                    if (isNaN(components.port)) {
                        components.port = matches[5];
                    }
                }
                else {
                    //IE FIX for improper RegExp matching
                    //store each component
                    components.scheme = matches[1] || undefined;
                    components.userinfo = uriString.indexOf("@") !== -1 ? matches[3] : undefined;
                    components.host = uriString.indexOf("//") !== -1 ? matches[4] : undefined;
                    components.port = parseInt(matches[5], 10);
                    components.path = matches[6] || "";
                    components.query = uriString.indexOf("?") !== -1 ? matches[7] : undefined;
                    components.fragment = uriString.indexOf("#") !== -1 ? matches[8] : undefined;
                    //fix port number
                    if (isNaN(components.port)) {
                        components.port = uriString.match(/\/\/(?:.|\n)*\:(?:\/|\?|\#|$)/) ? matches[4] : undefined;
                    }
                }
                if (components.host) {
                    //normalize IP hosts
                    components.host = _normalizeIPv6(_normalizeIPv4(components.host, protocol), protocol);
                }
                //determine reference type
                if (components.scheme === undefined && components.userinfo === undefined && components.host === undefined && components.port === undefined && !components.path && components.query === undefined) {
                    components.reference = "same-document";
                }
                else if (components.scheme === undefined) {
                    components.reference = "relative";
                }
                else if (components.fragment === undefined) {
                    components.reference = "absolute";
                }
                else {
                    components.reference = "uri";
                }
                //check for reference errors
                if (options.reference && options.reference !== "suffix" && options.reference !== components.reference) {
                    components.error = components.error || "URI is not a " + options.reference + " reference.";
                }
                //find scheme handler
                var schemeHandler = SCHEMES[(options.scheme || components.scheme || "").toLowerCase()];
                //check if scheme can't handle IRIs
                if (!options.unicodeSupport && (!schemeHandler || !schemeHandler.unicodeSupport)) {
                    //if host component is a domain name
                    if (components.host && (options.domainHost || schemeHandler && schemeHandler.domainHost)) {
                        //convert Unicode IDN -> ASCII IDN
                        try {
                            components.host = punycode.toASCII(components.host.replace(protocol.PCT_ENCODED, pctDecChars).toLowerCase());
                        }
                        catch (e) {
                            components.error = components.error || "Host's domain name can not be converted to ASCII via punycode: " + e;
                        }
                    }
                    //convert IRI -> URI
                    _normalizeComponentEncoding(components, URI_PROTOCOL);
                }
                else {
                    //normalize encodings
                    _normalizeComponentEncoding(components, protocol);
                }
                //perform scheme specific parsing
                if (schemeHandler && schemeHandler.parse) {
                    schemeHandler.parse(components, options);
                }
            }
            else {
                components.error = components.error || "URI can not be parsed.";
            }
            return components;
        }
        function _recomposeAuthority(components, options) {
            var protocol = options.iri !== false ? IRI_PROTOCOL : URI_PROTOCOL;
            var uriTokens = [];
            if (components.userinfo !== undefined) {
                uriTokens.push(components.userinfo);
                uriTokens.push("@");
            }
            if (components.host !== undefined) {
                //normalize IP hosts, add brackets and escape zone separator for IPv6
                uriTokens.push(_normalizeIPv6(_normalizeIPv4(String(components.host), protocol), protocol).replace(protocol.IPV6ADDRESS, function (_, $1, $2) {
                    return "[" + $1 + ($2 ? "%25" + $2 : "") + "]";
                }));
            }
            if (typeof components.port === "number" || typeof components.port === "string") {
                uriTokens.push(":");
                uriTokens.push(String(components.port));
            }
            return uriTokens.length ? uriTokens.join("") : undefined;
        }
        var RDS1 = /^\.\.?\//;
        var RDS2 = /^\/\.(\/|$)/;
        var RDS3 = /^\/\.\.(\/|$)/;
        var RDS5 = /^\/?(?:.|\n)*?(?=\/|$)/;
        function removeDotSegments(input) {
            var output = [];
            while (input.length) {
                if (input.match(RDS1)) {
                    input = input.replace(RDS1, "");
                }
                else if (input.match(RDS2)) {
                    input = input.replace(RDS2, "/");
                }
                else if (input.match(RDS3)) {
                    input = input.replace(RDS3, "/");
                    output.pop();
                }
                else if (input === "." || input === "..") {
                    input = "";
                }
                else {
                    var im = input.match(RDS5);
                    if (im) {
                        var s = im[0];
                        input = input.slice(s.length);
                        output.push(s);
                    }
                    else {
                        throw new Error("Unexpected dot segment condition");
                    }
                }
            }
            return output.join("");
        }
        function serialize(components) {
            var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
            var protocol = options.iri ? IRI_PROTOCOL : URI_PROTOCOL;
            var uriTokens = [];
            //find scheme handler
            var schemeHandler = SCHEMES[(options.scheme || components.scheme || "").toLowerCase()];
            //perform scheme specific serialization
            if (schemeHandler && schemeHandler.serialize)
                schemeHandler.serialize(components, options);
            if (components.host) {
                //if host component is an IPv6 address
                if (protocol.IPV6ADDRESS.test(components.host)) { }
                //TODO: normalize IPv6 address as per RFC 5952
                //if host component is a domain name
                else if (options.domainHost || schemeHandler && schemeHandler.domainHost) {
                    //convert IDN via punycode
                    try {
                        components.host = !options.iri ? punycode.toASCII(components.host.replace(protocol.PCT_ENCODED, pctDecChars).toLowerCase()) : punycode.toUnicode(components.host);
                    }
                    catch (e) {
                        components.error = components.error || "Host's domain name can not be converted to " + (!options.iri ? "ASCII" : "Unicode") + " via punycode: " + e;
                    }
                }
            }
            //normalize encoding
            _normalizeComponentEncoding(components, protocol);
            if (options.reference !== "suffix" && components.scheme) {
                uriTokens.push(components.scheme);
                uriTokens.push(":");
            }
            var authority = _recomposeAuthority(components, options);
            if (authority !== undefined) {
                if (options.reference !== "suffix") {
                    uriTokens.push("//");
                }
                uriTokens.push(authority);
                if (components.path && components.path.charAt(0) !== "/") {
                    uriTokens.push("/");
                }
            }
            if (components.path !== undefined) {
                var s = components.path;
                if (!options.absolutePath && (!schemeHandler || !schemeHandler.absolutePath)) {
                    s = removeDotSegments(s);
                }
                if (authority === undefined) {
                    s = s.replace(/^\/\//, "/%2F"); //don't allow the path to start with "//"
                }
                uriTokens.push(s);
            }
            if (components.query !== undefined) {
                uriTokens.push("?");
                uriTokens.push(components.query);
            }
            if (components.fragment !== undefined) {
                uriTokens.push("#");
                uriTokens.push(components.fragment);
            }
            return uriTokens.join(""); //merge tokens into a string
        }
        function resolveComponents(base, relative) {
            var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
            var skipNormalization = arguments[3];
            var target = {};
            if (!skipNormalization) {
                base = parse(serialize(base, options), options); //normalize base components
                relative = parse(serialize(relative, options), options); //normalize relative components
            }
            options = options || {};
            if (!options.tolerant && relative.scheme) {
                target.scheme = relative.scheme;
                //target.authority = relative.authority;
                target.userinfo = relative.userinfo;
                target.host = relative.host;
                target.port = relative.port;
                target.path = removeDotSegments(relative.path || "");
                target.query = relative.query;
            }
            else {
                if (relative.userinfo !== undefined || relative.host !== undefined || relative.port !== undefined) {
                    //target.authority = relative.authority;
                    target.userinfo = relative.userinfo;
                    target.host = relative.host;
                    target.port = relative.port;
                    target.path = removeDotSegments(relative.path || "");
                    target.query = relative.query;
                }
                else {
                    if (!relative.path) {
                        target.path = base.path;
                        if (relative.query !== undefined) {
                            target.query = relative.query;
                        }
                        else {
                            target.query = base.query;
                        }
                    }
                    else {
                        if (relative.path.charAt(0) === "/") {
                            target.path = removeDotSegments(relative.path);
                        }
                        else {
                            if ((base.userinfo !== undefined || base.host !== undefined || base.port !== undefined) && !base.path) {
                                target.path = "/" + relative.path;
                            }
                            else if (!base.path) {
                                target.path = relative.path;
                            }
                            else {
                                target.path = base.path.slice(0, base.path.lastIndexOf("/") + 1) + relative.path;
                            }
                            target.path = removeDotSegments(target.path);
                        }
                        target.query = relative.query;
                    }
                    //target.authority = base.authority;
                    target.userinfo = base.userinfo;
                    target.host = base.host;
                    target.port = base.port;
                }
                target.scheme = base.scheme;
            }
            target.fragment = relative.fragment;
            return target;
        }
        function resolve(baseURI, relativeURI, options) {
            var schemelessOptions = assign({ scheme: 'null' }, options);
            return serialize(resolveComponents(parse(baseURI, schemelessOptions), parse(relativeURI, schemelessOptions), schemelessOptions, true), schemelessOptions);
        }
        function normalize(uri, options) {
            if (typeof uri === "string") {
                uri = serialize(parse(uri, options), options);
            }
            else if (typeOf(uri) === "object") {
                uri = parse(serialize(uri, options), options);
            }
            return uri;
        }
        function equal(uriA, uriB, options) {
            if (typeof uriA === "string") {
                uriA = serialize(parse(uriA, options), options);
            }
            else if (typeOf(uriA) === "object") {
                uriA = serialize(uriA, options);
            }
            if (typeof uriB === "string") {
                uriB = serialize(parse(uriB, options), options);
            }
            else if (typeOf(uriB) === "object") {
                uriB = serialize(uriB, options);
            }
            return uriA === uriB;
        }
        function escapeComponent(str, options) {
            return str && str.toString().replace(!options || !options.iri ? URI_PROTOCOL.ESCAPE : IRI_PROTOCOL.ESCAPE, pctEncChar);
        }
        function unescapeComponent(str, options) {
            return str && str.toString().replace(!options || !options.iri ? URI_PROTOCOL.PCT_ENCODED : IRI_PROTOCOL.PCT_ENCODED, pctDecChars);
        }
        var handler = {
            scheme: "http",
            domainHost: true,
            parse: function parse(components, options) {
                //report missing host
                if (!components.host) {
                    components.error = components.error || "HTTP URIs must have a host.";
                }
                return components;
            },
            serialize: function serialize(components, options) {
                var secure = String(components.scheme).toLowerCase() === "https";
                //normalize the default port
                if (components.port === (secure ? 443 : 80) || components.port === "") {
                    components.port = undefined;
                }
                //normalize the empty path
                if (!components.path) {
                    components.path = "/";
                }
                //NOTE: We do not parse query strings for HTTP URIs
                //as WWW Form Url Encoded query strings are part of the HTML4+ spec,
                //and not the HTTP spec.
                return components;
            }
        };
        var handler$1 = {
            scheme: "https",
            domainHost: handler.domainHost,
            parse: handler.parse,
            serialize: handler.serialize
        };
        function isSecure(wsComponents) {
            return typeof wsComponents.secure === 'boolean' ? wsComponents.secure : String(wsComponents.scheme).toLowerCase() === "wss";
        }
        //RFC 6455
        var handler$2 = {
            scheme: "ws",
            domainHost: true,
            parse: function parse(components, options) {
                var wsComponents = components;
                //indicate if the secure flag is set
                wsComponents.secure = isSecure(wsComponents);
                //construct resouce name
                wsComponents.resourceName = (wsComponents.path || '/') + (wsComponents.query ? '?' + wsComponents.query : '');
                wsComponents.path = undefined;
                wsComponents.query = undefined;
                return wsComponents;
            },
            serialize: function serialize(wsComponents, options) {
                //normalize the default port
                if (wsComponents.port === (isSecure(wsComponents) ? 443 : 80) || wsComponents.port === "") {
                    wsComponents.port = undefined;
                }
                //ensure scheme matches secure flag
                if (typeof wsComponents.secure === 'boolean') {
                    wsComponents.scheme = wsComponents.secure ? 'wss' : 'ws';
                    wsComponents.secure = undefined;
                }
                //reconstruct path from resource name
                if (wsComponents.resourceName) {
                    var _wsComponents$resourc = wsComponents.resourceName.split('?'), _wsComponents$resourc2 = slicedToArray(_wsComponents$resourc, 2), path = _wsComponents$resourc2[0], query = _wsComponents$resourc2[1];
                    wsComponents.path = path && path !== '/' ? path : undefined;
                    wsComponents.query = query;
                    wsComponents.resourceName = undefined;
                }
                //forbid fragment component
                wsComponents.fragment = undefined;
                return wsComponents;
            }
        };
        var handler$3 = {
            scheme: "wss",
            domainHost: handler$2.domainHost,
            parse: handler$2.parse,
            serialize: handler$2.serialize
        };
        var O = {};
        var isIRI = true;
        //RFC 3986
        var UNRESERVED$$ = "[A-Za-z0-9\\-\\.\\_\\~" + (isIRI ? "\\xA0-\\u200D\\u2010-\\u2029\\u202F-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFEF" : "") + "]";
        var HEXDIG$$ = "[0-9A-Fa-f]"; //case-insensitive
        var PCT_ENCODED$ = subexp(subexp("%[EFef]" + HEXDIG$$ + "%" + HEXDIG$$ + HEXDIG$$ + "%" + HEXDIG$$ + HEXDIG$$) + "|" + subexp("%[89A-Fa-f]" + HEXDIG$$ + "%" + HEXDIG$$ + HEXDIG$$) + "|" + subexp("%" + HEXDIG$$ + HEXDIG$$)); //expanded
        //RFC 5322, except these symbols as per RFC 6068: @ : / ? # [ ] & ; =
        //const ATEXT$$ = "[A-Za-z0-9\\!\\#\\$\\%\\&\\'\\*\\+\\-\\/\\=\\?\\^\\_\\`\\{\\|\\}\\~]";
        //const WSP$$ = "[\\x20\\x09]";
        //const OBS_QTEXT$$ = "[\\x01-\\x08\\x0B\\x0C\\x0E-\\x1F\\x7F]";  //(%d1-8 / %d11-12 / %d14-31 / %d127)
        //const QTEXT$$ = merge("[\\x21\\x23-\\x5B\\x5D-\\x7E]", OBS_QTEXT$$);  //%d33 / %d35-91 / %d93-126 / obs-qtext
        //const VCHAR$$ = "[\\x21-\\x7E]";
        //const WSP$$ = "[\\x20\\x09]";
        //const OBS_QP$ = subexp("\\\\" + merge("[\\x00\\x0D\\x0A]", OBS_QTEXT$$));  //%d0 / CR / LF / obs-qtext
        //const FWS$ = subexp(subexp(WSP$$ + "*" + "\\x0D\\x0A") + "?" + WSP$$ + "+");
        //const QUOTED_PAIR$ = subexp(subexp("\\\\" + subexp(VCHAR$$ + "|" + WSP$$)) + "|" + OBS_QP$);
        //const QUOTED_STRING$ = subexp('\\"' + subexp(FWS$ + "?" + QCONTENT$) + "*" + FWS$ + "?" + '\\"');
        var ATEXT$$ = "[A-Za-z0-9\\!\\$\\%\\'\\*\\+\\-\\^\\_\\`\\{\\|\\}\\~]";
        var QTEXT$$ = "[\\!\\$\\%\\'\\(\\)\\*\\+\\,\\-\\.0-9\\<\\>A-Z\\x5E-\\x7E]";
        var VCHAR$$ = merge(QTEXT$$, "[\\\"\\\\]");
        var SOME_DELIMS$$ = "[\\!\\$\\'\\(\\)\\*\\+\\,\\;\\:\\@]";
        var UNRESERVED = new RegExp(UNRESERVED$$, "g");
        var PCT_ENCODED = new RegExp(PCT_ENCODED$, "g");
        var NOT_LOCAL_PART = new RegExp(merge("[^]", ATEXT$$, "[\\.]", '[\\"]', VCHAR$$), "g");
        var NOT_HFNAME = new RegExp(merge("[^]", UNRESERVED$$, SOME_DELIMS$$), "g");
        var NOT_HFVALUE = NOT_HFNAME;
        function decodeUnreserved(str) {
            var decStr = pctDecChars(str);
            return !decStr.match(UNRESERVED) ? str : decStr;
        }
        var handler$4 = {
            scheme: "mailto",
            parse: function parse$$1(components, options) {
                var mailtoComponents = components;
                var to = mailtoComponents.to = mailtoComponents.path ? mailtoComponents.path.split(",") : [];
                mailtoComponents.path = undefined;
                if (mailtoComponents.query) {
                    var unknownHeaders = false;
                    var headers = {};
                    var hfields = mailtoComponents.query.split("&");
                    for (var x = 0, xl = hfields.length; x < xl; ++x) {
                        var hfield = hfields[x].split("=");
                        switch (hfield[0]) {
                            case "to":
                                var toAddrs = hfield[1].split(",");
                                for (var _x = 0, _xl = toAddrs.length; _x < _xl; ++_x) {
                                    to.push(toAddrs[_x]);
                                }
                                break;
                            case "subject":
                                mailtoComponents.subject = unescapeComponent(hfield[1], options);
                                break;
                            case "body":
                                mailtoComponents.body = unescapeComponent(hfield[1], options);
                                break;
                            default:
                                unknownHeaders = true;
                                headers[unescapeComponent(hfield[0], options)] = unescapeComponent(hfield[1], options);
                                break;
                        }
                    }
                    if (unknownHeaders)
                        mailtoComponents.headers = headers;
                }
                mailtoComponents.query = undefined;
                for (var _x2 = 0, _xl2 = to.length; _x2 < _xl2; ++_x2) {
                    var addr = to[_x2].split("@");
                    addr[0] = unescapeComponent(addr[0]);
                    if (!options.unicodeSupport) {
                        //convert Unicode IDN -> ASCII IDN
                        try {
                            addr[1] = punycode.toASCII(unescapeComponent(addr[1], options).toLowerCase());
                        }
                        catch (e) {
                            mailtoComponents.error = mailtoComponents.error || "Email address's domain name can not be converted to ASCII via punycode: " + e;
                        }
                    }
                    else {
                        addr[1] = unescapeComponent(addr[1], options).toLowerCase();
                    }
                    to[_x2] = addr.join("@");
                }
                return mailtoComponents;
            },
            serialize: function serialize$$1(mailtoComponents, options) {
                var components = mailtoComponents;
                var to = toArray(mailtoComponents.to);
                if (to) {
                    for (var x = 0, xl = to.length; x < xl; ++x) {
                        var toAddr = String(to[x]);
                        var atIdx = toAddr.lastIndexOf("@");
                        var localPart = toAddr.slice(0, atIdx).replace(PCT_ENCODED, decodeUnreserved).replace(PCT_ENCODED, toUpperCase).replace(NOT_LOCAL_PART, pctEncChar);
                        var domain = toAddr.slice(atIdx + 1);
                        //convert IDN via punycode
                        try {
                            domain = !options.iri ? punycode.toASCII(unescapeComponent(domain, options).toLowerCase()) : punycode.toUnicode(domain);
                        }
                        catch (e) {
                            components.error = components.error || "Email address's domain name can not be converted to " + (!options.iri ? "ASCII" : "Unicode") + " via punycode: " + e;
                        }
                        to[x] = localPart + "@" + domain;
                    }
                    components.path = to.join(",");
                }
                var headers = mailtoComponents.headers = mailtoComponents.headers || {};
                if (mailtoComponents.subject)
                    headers["subject"] = mailtoComponents.subject;
                if (mailtoComponents.body)
                    headers["body"] = mailtoComponents.body;
                var fields = [];
                for (var name in headers) {
                    if (headers[name] !== O[name]) {
                        fields.push(name.replace(PCT_ENCODED, decodeUnreserved).replace(PCT_ENCODED, toUpperCase).replace(NOT_HFNAME, pctEncChar) + "=" + headers[name].replace(PCT_ENCODED, decodeUnreserved).replace(PCT_ENCODED, toUpperCase).replace(NOT_HFVALUE, pctEncChar));
                    }
                }
                if (fields.length) {
                    components.query = fields.join("&");
                }
                return components;
            }
        };
        var URN_PARSE = /^([^\:]+)\:(.*)/;
        //RFC 2141
        var handler$5 = {
            scheme: "urn",
            parse: function parse$$1(components, options) {
                var matches = components.path && components.path.match(URN_PARSE);
                var urnComponents = components;
                if (matches) {
                    var scheme = options.scheme || urnComponents.scheme || "urn";
                    var nid = matches[1].toLowerCase();
                    var nss = matches[2];
                    var urnScheme = scheme + ":" + (options.nid || nid);
                    var schemeHandler = SCHEMES[urnScheme];
                    urnComponents.nid = nid;
                    urnComponents.nss = nss;
                    urnComponents.path = undefined;
                    if (schemeHandler) {
                        urnComponents = schemeHandler.parse(urnComponents, options);
                    }
                }
                else {
                    urnComponents.error = urnComponents.error || "URN can not be parsed.";
                }
                return urnComponents;
            },
            serialize: function serialize$$1(urnComponents, options) {
                var scheme = options.scheme || urnComponents.scheme || "urn";
                var nid = urnComponents.nid;
                var urnScheme = scheme + ":" + (options.nid || nid);
                var schemeHandler = SCHEMES[urnScheme];
                if (schemeHandler) {
                    urnComponents = schemeHandler.serialize(urnComponents, options);
                }
                var uriComponents = urnComponents;
                var nss = urnComponents.nss;
                uriComponents.path = (nid || options.nid) + ":" + nss;
                return uriComponents;
            }
        };
        var UUID = /^[0-9A-Fa-f]{8}(?:\-[0-9A-Fa-f]{4}){3}\-[0-9A-Fa-f]{12}$/;
        //RFC 4122
        var handler$6 = {
            scheme: "urn:uuid",
            parse: function parse(urnComponents, options) {
                var uuidComponents = urnComponents;
                uuidComponents.uuid = uuidComponents.nss;
                uuidComponents.nss = undefined;
                if (!options.tolerant && (!uuidComponents.uuid || !uuidComponents.uuid.match(UUID))) {
                    uuidComponents.error = uuidComponents.error || "UUID is not valid.";
                }
                return uuidComponents;
            },
            serialize: function serialize(uuidComponents, options) {
                var urnComponents = uuidComponents;
                //normalize UUID
                urnComponents.nss = (uuidComponents.uuid || "").toLowerCase();
                return urnComponents;
            }
        };
        SCHEMES[handler.scheme] = handler;
        SCHEMES[handler$1.scheme] = handler$1;
        SCHEMES[handler$2.scheme] = handler$2;
        SCHEMES[handler$3.scheme] = handler$3;
        SCHEMES[handler$4.scheme] = handler$4;
        SCHEMES[handler$5.scheme] = handler$5;
        SCHEMES[handler$6.scheme] = handler$6;
        exports.SCHEMES = SCHEMES;
        exports.pctEncChar = pctEncChar;
        exports.pctDecChars = pctDecChars;
        exports.parse = parse;
        exports.removeDotSegments = removeDotSegments;
        exports.serialize = serialize;
        exports.resolveComponents = resolveComponents;
        exports.resolve = resolve;
        exports.normalize = normalize;
        exports.equal = equal;
        exports.escapeComponent = escapeComponent;
        exports.unescapeComponent = unescapeComponent;
        Object.defineProperty(exports, '__esModule', { value: true });
    })));
},
/* fast-deep-equal/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // do not edit .js files directly - edit src/index.jst
    module.exports = function equal(a, b) {
        if (a === b)
            return true;
        if (a && b && typeof a == 'object' && typeof b == 'object') {
            if (a.constructor !== b.constructor)
                return false;
            var length, i, keys;
            if (Array.isArray(a)) {
                length = a.length;
                if (length != b.length)
                    return false;
                for (i = length; i-- !== 0;)
                    if (!equal(a[i], b[i]))
                        return false;
                return true;
            }
            if (a.constructor === RegExp)
                return a.source === b.source && a.flags === b.flags;
            if (a.valueOf !== Object.prototype.valueOf)
                return a.valueOf() === b.valueOf();
            if (a.toString !== Object.prototype.toString)
                return a.toString() === b.toString();
            keys = Object.keys(a);
            length = keys.length;
            if (length !== Object.keys(b).length)
                return false;
            for (i = length; i-- !== 0;)
                if (!Object.prototype.hasOwnProperty.call(b, keys[i]))
                    return false;
            for (i = length; i-- !== 0;) {
                var key = keys[i];
                if (!equal(a[key], b[key]))
                    return false;
            }
            return true;
        }
        // true if both NaN, false otherwise
        return a !== a && b !== b;
    };
},
/* ajv/lib/compile/util.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        copy: copy,
        checkDataType: checkDataType,
        checkDataTypes: checkDataTypes,
        coerceToTypes: coerceToTypes,
        toHash: toHash,
        getProperty: getProperty,
        escapeQuotes: escapeQuotes,
        equal: require(227) /* fast-deep-equal */,
        ucs2length: require(229) /* ./ucs2length */,
        varOccurences: varOccurences,
        varReplace: varReplace,
        schemaHasRules: schemaHasRules,
        schemaHasRulesExcept: schemaHasRulesExcept,
        schemaUnknownRules: schemaUnknownRules,
        toQuotedString: toQuotedString,
        getPathExpr: getPathExpr,
        getPath: getPath,
        getData: getData,
        unescapeFragment: unescapeFragment,
        unescapeJsonPointer: unescapeJsonPointer,
        escapeFragment: escapeFragment,
        escapeJsonPointer: escapeJsonPointer
    };
    function copy(o, to) {
        to = to || {};
        for (var key in o)
            to[key] = o[key];
        return to;
    }
    function checkDataType(dataType, data, strictNumbers, negate) {
        var EQUAL = negate ? ' !== ' : ' === ', AND = negate ? ' || ' : ' && ', OK = negate ? '!' : '', NOT = negate ? '' : '!';
        switch (dataType) {
            case 'null': return data + EQUAL + 'null';
            case 'array': return OK + 'Array.isArray(' + data + ')';
            case 'object': return '(' + OK + data + AND +
                'typeof ' + data + EQUAL + '"object"' + AND +
                NOT + 'Array.isArray(' + data + '))';
            case 'integer': return '(typeof ' + data + EQUAL + '"number"' + AND +
                NOT + '(' + data + ' % 1)' +
                AND + data + EQUAL + data +
                (strictNumbers ? (AND + OK + 'isFinite(' + data + ')') : '') + ')';
            case 'number': return '(typeof ' + data + EQUAL + '"' + dataType + '"' +
                (strictNumbers ? (AND + OK + 'isFinite(' + data + ')') : '') + ')';
            default: return 'typeof ' + data + EQUAL + '"' + dataType + '"';
        }
    }
    function checkDataTypes(dataTypes, data, strictNumbers) {
        switch (dataTypes.length) {
            case 1: return checkDataType(dataTypes[0], data, strictNumbers, true);
            default:
                var code = '';
                var types = toHash(dataTypes);
                if (types.array && types.object) {
                    code = types.null ? '(' : '(!' + data + ' || ';
                    code += 'typeof ' + data + ' !== "object")';
                    delete types.null;
                    delete types.array;
                    delete types.object;
                }
                if (types.number)
                    delete types.integer;
                for (var t in types)
                    code += (code ? ' && ' : '') + checkDataType(t, data, strictNumbers, true);
                return code;
        }
    }
    var COERCE_TO_TYPES = toHash(['string', 'number', 'integer', 'boolean', 'null']);
    function coerceToTypes(optionCoerceTypes, dataTypes) {
        if (Array.isArray(dataTypes)) {
            var types = [];
            for (var i = 0; i < dataTypes.length; i++) {
                var t = dataTypes[i];
                if (COERCE_TO_TYPES[t])
                    types[types.length] = t;
                else if (optionCoerceTypes === 'array' && t === 'array')
                    types[types.length] = t;
            }
            if (types.length)
                return types;
        }
        else if (COERCE_TO_TYPES[dataTypes]) {
            return [dataTypes];
        }
        else if (optionCoerceTypes === 'array' && dataTypes === 'array') {
            return ['array'];
        }
    }
    function toHash(arr) {
        var hash = {};
        for (var i = 0; i < arr.length; i++)
            hash[arr[i]] = true;
        return hash;
    }
    var IDENTIFIER = /^[a-z$_][a-z$_0-9]*$/i;
    var SINGLE_QUOTE = /'|\\/g;
    function getProperty(key) {
        return typeof key == 'number'
            ? '[' + key + ']'
            : IDENTIFIER.test(key)
                ? '.' + key
                : "['" + escapeQuotes(key) + "']";
    }
    function escapeQuotes(str) {
        return str.replace(SINGLE_QUOTE, '\\$&')
            .replace(/\n/g, '\\n')
            .replace(/\r/g, '\\r')
            .replace(/\f/g, '\\f')
            .replace(/\t/g, '\\t');
    }
    function varOccurences(str, dataVar) {
        dataVar += '[^0-9]';
        var matches = str.match(new RegExp(dataVar, 'g'));
        return matches ? matches.length : 0;
    }
    function varReplace(str, dataVar, expr) {
        dataVar += '([^0-9])';
        expr = expr.replace(/\$/g, '$$$$');
        return str.replace(new RegExp(dataVar, 'g'), expr + '$1');
    }
    function schemaHasRules(schema, rules) {
        if (typeof schema == 'boolean')
            return !schema;
        for (var key in schema)
            if (rules[key])
                return true;
    }
    function schemaHasRulesExcept(schema, rules, exceptKeyword) {
        if (typeof schema == 'boolean')
            return !schema && exceptKeyword != 'not';
        for (var key in schema)
            if (key != exceptKeyword && rules[key])
                return true;
    }
    function schemaUnknownRules(schema, rules) {
        if (typeof schema == 'boolean')
            return;
        for (var key in schema)
            if (!rules[key])
                return key;
    }
    function toQuotedString(str) {
        return '\'' + escapeQuotes(str) + '\'';
    }
    function getPathExpr(currentPath, expr, jsonPointers, isNumber) {
        var path = jsonPointers // false by default
            ? '\'/\' + ' + expr + (isNumber ? '' : '.replace(/~/g, \'~0\').replace(/\\//g, \'~1\')')
            : (isNumber ? '\'[\' + ' + expr + ' + \']\'' : '\'[\\\'\' + ' + expr + ' + \'\\\']\'');
        return joinPaths(currentPath, path);
    }
    function getPath(currentPath, prop, jsonPointers) {
        var path = jsonPointers // false by default
            ? toQuotedString('/' + escapeJsonPointer(prop))
            : toQuotedString(getProperty(prop));
        return joinPaths(currentPath, path);
    }
    var JSON_POINTER = /^\/(?:[^~]|~0|~1)*$/;
    var RELATIVE_JSON_POINTER = /^([0-9]+)(#|\/(?:[^~]|~0|~1)*)?$/;
    function getData($data, lvl, paths) {
        var up, jsonPointer, data, matches;
        if ($data === '')
            return 'rootData';
        if ($data[0] == '/') {
            if (!JSON_POINTER.test($data))
                throw new Error('Invalid JSON-pointer: ' + $data);
            jsonPointer = $data;
            data = 'rootData';
        }
        else {
            matches = $data.match(RELATIVE_JSON_POINTER);
            if (!matches)
                throw new Error('Invalid JSON-pointer: ' + $data);
            up = +matches[1];
            jsonPointer = matches[2];
            if (jsonPointer == '#') {
                if (up >= lvl)
                    throw new Error('Cannot access property/index ' + up + ' levels up, current level is ' + lvl);
                return paths[lvl - up];
            }
            if (up > lvl)
                throw new Error('Cannot access data ' + up + ' levels up, current level is ' + lvl);
            data = 'data' + ((lvl - up) || '');
            if (!jsonPointer)
                return data;
        }
        var expr = data;
        var segments = jsonPointer.split('/');
        for (var i = 0; i < segments.length; i++) {
            var segment = segments[i];
            if (segment) {
                data += getProperty(unescapeJsonPointer(segment));
                expr += ' && ' + data;
            }
        }
        return expr;
    }
    function joinPaths(a, b) {
        if (a == '""')
            return b;
        return (a + ' + ' + b).replace(/([^\\])' \+ '/g, '$1');
    }
    function unescapeFragment(str) {
        return unescapeJsonPointer(decodeURIComponent(str));
    }
    function escapeFragment(str) {
        return encodeURIComponent(escapeJsonPointer(str));
    }
    function escapeJsonPointer(str) {
        return str.replace(/~/g, '~0').replace(/\//g, '~1');
    }
    function unescapeJsonPointer(str) {
        return str.replace(/~1/g, '/').replace(/~0/g, '~');
    }
},
/* ajv/lib/compile/ucs2length.js */ function _(require, module, exports, __esModule, __esExport) {
    // https://mathiasbynens.be/notes/javascript-encoding
    // https://github.com/bestiejs/punycode.js - punycode.ucs2.decode
    module.exports = function ucs2length(str) {
        var length = 0, len = str.length, pos = 0, value;
        while (pos < len) {
            length++;
            value = str.charCodeAt(pos++);
            if (value >= 0xD800 && value <= 0xDBFF && pos < len) {
                // high surrogate, and there is a next character
                value = str.charCodeAt(pos);
                if ((value & 0xFC00) == 0xDC00)
                    pos++; // low surrogate
            }
        }
        return length;
    };
},
/* ajv/lib/compile/schema_obj.js */ function _(require, module, exports, __esModule, __esExport) {
    var util = require(228) /* ./util */;
    module.exports = SchemaObject;
    function SchemaObject(obj) {
        util.copy(obj, this);
    }
},
/* json-schema-traverse/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var traverse = module.exports = function (schema, opts, cb) {
        // Legacy support for v0.3.1 and earlier.
        if (typeof opts == 'function') {
            cb = opts;
            opts = {};
        }
        cb = opts.cb || cb;
        var pre = (typeof cb == 'function') ? cb : cb.pre || function () { };
        var post = cb.post || function () { };
        _traverse(opts, pre, post, schema, '', schema);
    };
    traverse.keywords = {
        additionalItems: true,
        items: true,
        contains: true,
        additionalProperties: true,
        propertyNames: true,
        not: true
    };
    traverse.arrayKeywords = {
        items: true,
        allOf: true,
        anyOf: true,
        oneOf: true
    };
    traverse.propsKeywords = {
        definitions: true,
        properties: true,
        patternProperties: true,
        dependencies: true
    };
    traverse.skipKeywords = {
        default: true,
        enum: true,
        const: true,
        required: true,
        maximum: true,
        minimum: true,
        exclusiveMaximum: true,
        exclusiveMinimum: true,
        multipleOf: true,
        maxLength: true,
        minLength: true,
        pattern: true,
        format: true,
        maxItems: true,
        minItems: true,
        uniqueItems: true,
        maxProperties: true,
        minProperties: true
    };
    function _traverse(opts, pre, post, schema, jsonPtr, rootSchema, parentJsonPtr, parentKeyword, parentSchema, keyIndex) {
        if (schema && typeof schema == 'object' && !Array.isArray(schema)) {
            pre(schema, jsonPtr, rootSchema, parentJsonPtr, parentKeyword, parentSchema, keyIndex);
            for (var key in schema) {
                var sch = schema[key];
                if (Array.isArray(sch)) {
                    if (key in traverse.arrayKeywords) {
                        for (var i = 0; i < sch.length; i++)
                            _traverse(opts, pre, post, sch[i], jsonPtr + '/' + key + '/' + i, rootSchema, jsonPtr, key, schema, i);
                    }
                }
                else if (key in traverse.propsKeywords) {
                    if (sch && typeof sch == 'object') {
                        for (var prop in sch)
                            _traverse(opts, pre, post, sch[prop], jsonPtr + '/' + key + '/' + escapeJsonPtr(prop), rootSchema, jsonPtr, key, schema, prop);
                    }
                }
                else if (key in traverse.keywords || (opts.allKeys && !(key in traverse.skipKeywords))) {
                    _traverse(opts, pre, post, sch, jsonPtr + '/' + key, rootSchema, jsonPtr, key, schema);
                }
            }
            post(schema, jsonPtr, rootSchema, parentJsonPtr, parentKeyword, parentSchema, keyIndex);
        }
    }
    function escapeJsonPtr(str) {
        return str.replace(/~/g, '~0').replace(/\//g, '~1');
    }
},
/* ajv/lib/compile/error_classes.js */ function _(require, module, exports, __esModule, __esExport) {
    var resolve = require(225) /* ./resolve */;
    module.exports = {
        Validation: errorSubclass(ValidationError),
        MissingRef: errorSubclass(MissingRefError)
    };
    function ValidationError(errors) {
        this.message = 'validation failed';
        this.errors = errors;
        this.ajv = this.validation = true;
    }
    MissingRefError.message = function (baseId, ref) {
        return 'can\'t resolve reference ' + ref + ' from id ' + baseId;
    };
    function MissingRefError(baseId, ref, message) {
        this.message = message || MissingRefError.message(baseId, ref);
        this.missingRef = resolve.url(baseId, ref);
        this.missingSchema = resolve.normalizeId(resolve.fullPath(this.missingRef));
    }
    function errorSubclass(Subclass) {
        Subclass.prototype = Object.create(Error.prototype);
        Subclass.prototype.constructor = Subclass;
        return Subclass;
    }
},
/* fast-json-stable-stringify/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function (data, opts) {
        if (!opts)
            opts = {};
        if (typeof opts === 'function')
            opts = { cmp: opts };
        var cycles = (typeof opts.cycles === 'boolean') ? opts.cycles : false;
        var cmp = opts.cmp && (function (f) {
            return function (node) {
                return function (a, b) {
                    var aobj = { key: a, value: node[a] };
                    var bobj = { key: b, value: node[b] };
                    return f(aobj, bobj);
                };
            };
        })(opts.cmp);
        var seen = [];
        return (function stringify(node) {
            if (node && node.toJSON && typeof node.toJSON === 'function') {
                node = node.toJSON();
            }
            if (node === undefined)
                return;
            if (typeof node == 'number')
                return isFinite(node) ? '' + node : 'null';
            if (typeof node !== 'object')
                return JSON.stringify(node);
            var i, out;
            if (Array.isArray(node)) {
                out = '[';
                for (i = 0; i < node.length; i++) {
                    if (i)
                        out += ',';
                    out += stringify(node[i]) || 'null';
                }
                return out + ']';
            }
            if (node === null)
                return 'null';
            if (seen.indexOf(node) !== -1) {
                if (cycles)
                    return JSON.stringify('__cycle__');
                throw new TypeError('Converting circular structure to JSON');
            }
            var seenIndex = seen.push(node) - 1;
            var keys = Object.keys(node).sort(cmp && cmp(node));
            out = '';
            for (i = 0; i < keys.length; i++) {
                var key = keys[i];
                var value = stringify(node[key]);
                if (!value)
                    continue;
                if (out)
                    out += ',';
                out += JSON.stringify(key) + ':' + value;
            }
            seen.splice(seenIndex, 1);
            return '{' + out + '}';
        })(data);
    };
},
/* ajv/lib/dotjs/validate.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_validate(it, $keyword, $ruleType) {
        var out = '';
        var $async = it.schema.$async === true, $refKeywords = it.util.schemaHasRulesExcept(it.schema, it.RULES.all, '$ref'), $id = it.self._getId(it.schema);
        if (it.opts.strictKeywords) {
            var $unknownKwd = it.util.schemaUnknownRules(it.schema, it.RULES.keywords);
            if ($unknownKwd) {
                var $keywordsMsg = 'unknown keyword: ' + $unknownKwd;
                if (it.opts.strictKeywords === 'log')
                    it.logger.warn($keywordsMsg);
                else
                    throw new Error($keywordsMsg);
            }
        }
        if (it.isTop) {
            out += ' var validate = ';
            if ($async) {
                it.async = true;
                out += 'async ';
            }
            out += 'function(data, dataPath, parentData, parentDataProperty, rootData) { \'use strict\'; ';
            if ($id && (it.opts.sourceCode || it.opts.processCode)) {
                out += ' ' + ('/\*# sourceURL=' + $id + ' */') + ' ';
            }
        }
        if (typeof it.schema == 'boolean' || !($refKeywords || it.schema.$ref)) {
            var $keyword = 'false schema';
            var $lvl = it.level;
            var $dataLvl = it.dataLevel;
            var $schema = it.schema[$keyword];
            var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
            var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
            var $breakOnError = !it.opts.allErrors;
            var $errorKeyword;
            var $data = 'data' + ($dataLvl || '');
            var $valid = 'valid' + $lvl;
            if (it.schema === false) {
                if (it.isTop) {
                    $breakOnError = true;
                }
                else {
                    out += ' var ' + ($valid) + ' = false; ';
                }
                var $$outStack = $$outStack || [];
                $$outStack.push(out);
                out = ''; /* istanbul ignore else */
                if (it.createErrors !== false) {
                    out += ' { keyword: \'' + ($errorKeyword || 'false schema') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: {} ';
                    if (it.opts.messages !== false) {
                        out += ' , message: \'boolean schema is false\' ';
                    }
                    if (it.opts.verbose) {
                        out += ' , schema: false , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                    }
                    out += ' } ';
                }
                else {
                    out += ' {} ';
                }
                var __err = out;
                out = $$outStack.pop();
                if (!it.compositeRule && $breakOnError) {
                    /* istanbul ignore if */
                    if (it.async) {
                        out += ' throw new ValidationError([' + (__err) + ']); ';
                    }
                    else {
                        out += ' validate.errors = [' + (__err) + ']; return false; ';
                    }
                }
                else {
                    out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                }
            }
            else {
                if (it.isTop) {
                    if ($async) {
                        out += ' return data; ';
                    }
                    else {
                        out += ' validate.errors = null; return true; ';
                    }
                }
                else {
                    out += ' var ' + ($valid) + ' = true; ';
                }
            }
            if (it.isTop) {
                out += ' }; return validate; ';
            }
            return out;
        }
        if (it.isTop) {
            var $top = it.isTop, $lvl = it.level = 0, $dataLvl = it.dataLevel = 0, $data = 'data';
            it.rootId = it.resolve.fullPath(it.self._getId(it.root.schema));
            it.baseId = it.baseId || it.rootId;
            delete it.isTop;
            it.dataPathArr = [""];
            if (it.schema.default !== undefined && it.opts.useDefaults && it.opts.strictDefaults) {
                var $defaultMsg = 'default is ignored in the schema root';
                if (it.opts.strictDefaults === 'log')
                    it.logger.warn($defaultMsg);
                else
                    throw new Error($defaultMsg);
            }
            out += ' var vErrors = null; ';
            out += ' var errors = 0;     ';
            out += ' if (rootData === undefined) rootData = data; ';
        }
        else {
            var $lvl = it.level, $dataLvl = it.dataLevel, $data = 'data' + ($dataLvl || '');
            if ($id)
                it.baseId = it.resolve.url(it.baseId, $id);
            if ($async && !it.async)
                throw new Error('async schema in sync schema');
            out += ' var errs_' + ($lvl) + ' = errors;';
        }
        var $valid = 'valid' + $lvl, $breakOnError = !it.opts.allErrors, $closingBraces1 = '', $closingBraces2 = '';
        var $errorKeyword;
        var $typeSchema = it.schema.type, $typeIsArray = Array.isArray($typeSchema);
        if ($typeSchema && it.opts.nullable && it.schema.nullable === true) {
            if ($typeIsArray) {
                if ($typeSchema.indexOf('null') == -1)
                    $typeSchema = $typeSchema.concat('null');
            }
            else if ($typeSchema != 'null') {
                $typeSchema = [$typeSchema, 'null'];
                $typeIsArray = true;
            }
        }
        if ($typeIsArray && $typeSchema.length == 1) {
            $typeSchema = $typeSchema[0];
            $typeIsArray = false;
        }
        if (it.schema.$ref && $refKeywords) {
            if (it.opts.extendRefs == 'fail') {
                throw new Error('$ref: validation keywords used in schema at path "' + it.errSchemaPath + '" (see option extendRefs)');
            }
            else if (it.opts.extendRefs !== true) {
                $refKeywords = false;
                it.logger.warn('$ref: keywords ignored in schema at path "' + it.errSchemaPath + '"');
            }
        }
        if (it.schema.$comment && it.opts.$comment) {
            out += ' ' + (it.RULES.all.$comment.code(it, '$comment'));
        }
        if ($typeSchema) {
            if (it.opts.coerceTypes) {
                var $coerceToTypes = it.util.coerceToTypes(it.opts.coerceTypes, $typeSchema);
            }
            var $rulesGroup = it.RULES.types[$typeSchema];
            if ($coerceToTypes || $typeIsArray || $rulesGroup === true || ($rulesGroup && !$shouldUseGroup($rulesGroup))) {
                var $schemaPath = it.schemaPath + '.type', $errSchemaPath = it.errSchemaPath + '/type';
                var $schemaPath = it.schemaPath + '.type', $errSchemaPath = it.errSchemaPath + '/type', $method = $typeIsArray ? 'checkDataTypes' : 'checkDataType';
                out += ' if (' + (it.util[$method]($typeSchema, $data, it.opts.strictNumbers, true)) + ') { ';
                if ($coerceToTypes) {
                    var $dataType = 'dataType' + $lvl, $coerced = 'coerced' + $lvl;
                    out += ' var ' + ($dataType) + ' = typeof ' + ($data) + '; var ' + ($coerced) + ' = undefined; ';
                    if (it.opts.coerceTypes == 'array') {
                        out += ' if (' + ($dataType) + ' == \'object\' && Array.isArray(' + ($data) + ') && ' + ($data) + '.length == 1) { ' + ($data) + ' = ' + ($data) + '[0]; ' + ($dataType) + ' = typeof ' + ($data) + '; if (' + (it.util.checkDataType(it.schema.type, $data, it.opts.strictNumbers)) + ') ' + ($coerced) + ' = ' + ($data) + '; } ';
                    }
                    out += ' if (' + ($coerced) + ' !== undefined) ; ';
                    var arr1 = $coerceToTypes;
                    if (arr1) {
                        var $type, $i = -1, l1 = arr1.length - 1;
                        while ($i < l1) {
                            $type = arr1[$i += 1];
                            if ($type == 'string') {
                                out += ' else if (' + ($dataType) + ' == \'number\' || ' + ($dataType) + ' == \'boolean\') ' + ($coerced) + ' = \'\' + ' + ($data) + '; else if (' + ($data) + ' === null) ' + ($coerced) + ' = \'\'; ';
                            }
                            else if ($type == 'number' || $type == 'integer') {
                                out += ' else if (' + ($dataType) + ' == \'boolean\' || ' + ($data) + ' === null || (' + ($dataType) + ' == \'string\' && ' + ($data) + ' && ' + ($data) + ' == +' + ($data) + ' ';
                                if ($type == 'integer') {
                                    out += ' && !(' + ($data) + ' % 1)';
                                }
                                out += ')) ' + ($coerced) + ' = +' + ($data) + '; ';
                            }
                            else if ($type == 'boolean') {
                                out += ' else if (' + ($data) + ' === \'false\' || ' + ($data) + ' === 0 || ' + ($data) + ' === null) ' + ($coerced) + ' = false; else if (' + ($data) + ' === \'true\' || ' + ($data) + ' === 1) ' + ($coerced) + ' = true; ';
                            }
                            else if ($type == 'null') {
                                out += ' else if (' + ($data) + ' === \'\' || ' + ($data) + ' === 0 || ' + ($data) + ' === false) ' + ($coerced) + ' = null; ';
                            }
                            else if (it.opts.coerceTypes == 'array' && $type == 'array') {
                                out += ' else if (' + ($dataType) + ' == \'string\' || ' + ($dataType) + ' == \'number\' || ' + ($dataType) + ' == \'boolean\' || ' + ($data) + ' == null) ' + ($coerced) + ' = [' + ($data) + ']; ';
                            }
                        }
                    }
                    out += ' else {   ';
                    var $$outStack = $$outStack || [];
                    $$outStack.push(out);
                    out = ''; /* istanbul ignore else */
                    if (it.createErrors !== false) {
                        out += ' { keyword: \'' + ($errorKeyword || 'type') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { type: \'';
                        if ($typeIsArray) {
                            out += '' + ($typeSchema.join(","));
                        }
                        else {
                            out += '' + ($typeSchema);
                        }
                        out += '\' } ';
                        if (it.opts.messages !== false) {
                            out += ' , message: \'should be ';
                            if ($typeIsArray) {
                                out += '' + ($typeSchema.join(","));
                            }
                            else {
                                out += '' + ($typeSchema);
                            }
                            out += '\' ';
                        }
                        if (it.opts.verbose) {
                            out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                        }
                        out += ' } ';
                    }
                    else {
                        out += ' {} ';
                    }
                    var __err = out;
                    out = $$outStack.pop();
                    if (!it.compositeRule && $breakOnError) {
                        /* istanbul ignore if */
                        if (it.async) {
                            out += ' throw new ValidationError([' + (__err) + ']); ';
                        }
                        else {
                            out += ' validate.errors = [' + (__err) + ']; return false; ';
                        }
                    }
                    else {
                        out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                    }
                    out += ' } if (' + ($coerced) + ' !== undefined) {  ';
                    var $parentData = $dataLvl ? 'data' + (($dataLvl - 1) || '') : 'parentData', $parentDataProperty = $dataLvl ? it.dataPathArr[$dataLvl] : 'parentDataProperty';
                    out += ' ' + ($data) + ' = ' + ($coerced) + '; ';
                    if (!$dataLvl) {
                        out += 'if (' + ($parentData) + ' !== undefined)';
                    }
                    out += ' ' + ($parentData) + '[' + ($parentDataProperty) + '] = ' + ($coerced) + '; } ';
                }
                else {
                    var $$outStack = $$outStack || [];
                    $$outStack.push(out);
                    out = ''; /* istanbul ignore else */
                    if (it.createErrors !== false) {
                        out += ' { keyword: \'' + ($errorKeyword || 'type') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { type: \'';
                        if ($typeIsArray) {
                            out += '' + ($typeSchema.join(","));
                        }
                        else {
                            out += '' + ($typeSchema);
                        }
                        out += '\' } ';
                        if (it.opts.messages !== false) {
                            out += ' , message: \'should be ';
                            if ($typeIsArray) {
                                out += '' + ($typeSchema.join(","));
                            }
                            else {
                                out += '' + ($typeSchema);
                            }
                            out += '\' ';
                        }
                        if (it.opts.verbose) {
                            out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                        }
                        out += ' } ';
                    }
                    else {
                        out += ' {} ';
                    }
                    var __err = out;
                    out = $$outStack.pop();
                    if (!it.compositeRule && $breakOnError) {
                        /* istanbul ignore if */
                        if (it.async) {
                            out += ' throw new ValidationError([' + (__err) + ']); ';
                        }
                        else {
                            out += ' validate.errors = [' + (__err) + ']; return false; ';
                        }
                    }
                    else {
                        out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                    }
                }
                out += ' } ';
            }
        }
        if (it.schema.$ref && !$refKeywords) {
            out += ' ' + (it.RULES.all.$ref.code(it, '$ref')) + ' ';
            if ($breakOnError) {
                out += ' } if (errors === ';
                if ($top) {
                    out += '0';
                }
                else {
                    out += 'errs_' + ($lvl);
                }
                out += ') { ';
                $closingBraces2 += '}';
            }
        }
        else {
            var arr2 = it.RULES;
            if (arr2) {
                var $rulesGroup, i2 = -1, l2 = arr2.length - 1;
                while (i2 < l2) {
                    $rulesGroup = arr2[i2 += 1];
                    if ($shouldUseGroup($rulesGroup)) {
                        if ($rulesGroup.type) {
                            out += ' if (' + (it.util.checkDataType($rulesGroup.type, $data, it.opts.strictNumbers)) + ') { ';
                        }
                        if (it.opts.useDefaults) {
                            if ($rulesGroup.type == 'object' && it.schema.properties) {
                                var $schema = it.schema.properties, $schemaKeys = Object.keys($schema);
                                var arr3 = $schemaKeys;
                                if (arr3) {
                                    var $propertyKey, i3 = -1, l3 = arr3.length - 1;
                                    while (i3 < l3) {
                                        $propertyKey = arr3[i3 += 1];
                                        var $sch = $schema[$propertyKey];
                                        if ($sch.default !== undefined) {
                                            var $passData = $data + it.util.getProperty($propertyKey);
                                            if (it.compositeRule) {
                                                if (it.opts.strictDefaults) {
                                                    var $defaultMsg = 'default is ignored for: ' + $passData;
                                                    if (it.opts.strictDefaults === 'log')
                                                        it.logger.warn($defaultMsg);
                                                    else
                                                        throw new Error($defaultMsg);
                                                }
                                            }
                                            else {
                                                out += ' if (' + ($passData) + ' === undefined ';
                                                if (it.opts.useDefaults == 'empty') {
                                                    out += ' || ' + ($passData) + ' === null || ' + ($passData) + ' === \'\' ';
                                                }
                                                out += ' ) ' + ($passData) + ' = ';
                                                if (it.opts.useDefaults == 'shared') {
                                                    out += ' ' + (it.useDefault($sch.default)) + ' ';
                                                }
                                                else {
                                                    out += ' ' + (JSON.stringify($sch.default)) + ' ';
                                                }
                                                out += '; ';
                                            }
                                        }
                                    }
                                }
                            }
                            else if ($rulesGroup.type == 'array' && Array.isArray(it.schema.items)) {
                                var arr4 = it.schema.items;
                                if (arr4) {
                                    var $sch, $i = -1, l4 = arr4.length - 1;
                                    while ($i < l4) {
                                        $sch = arr4[$i += 1];
                                        if ($sch.default !== undefined) {
                                            var $passData = $data + '[' + $i + ']';
                                            if (it.compositeRule) {
                                                if (it.opts.strictDefaults) {
                                                    var $defaultMsg = 'default is ignored for: ' + $passData;
                                                    if (it.opts.strictDefaults === 'log')
                                                        it.logger.warn($defaultMsg);
                                                    else
                                                        throw new Error($defaultMsg);
                                                }
                                            }
                                            else {
                                                out += ' if (' + ($passData) + ' === undefined ';
                                                if (it.opts.useDefaults == 'empty') {
                                                    out += ' || ' + ($passData) + ' === null || ' + ($passData) + ' === \'\' ';
                                                }
                                                out += ' ) ' + ($passData) + ' = ';
                                                if (it.opts.useDefaults == 'shared') {
                                                    out += ' ' + (it.useDefault($sch.default)) + ' ';
                                                }
                                                else {
                                                    out += ' ' + (JSON.stringify($sch.default)) + ' ';
                                                }
                                                out += '; ';
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        var arr5 = $rulesGroup.rules;
                        if (arr5) {
                            var $rule, i5 = -1, l5 = arr5.length - 1;
                            while (i5 < l5) {
                                $rule = arr5[i5 += 1];
                                if ($shouldUseRule($rule)) {
                                    var $code = $rule.code(it, $rule.keyword, $rulesGroup.type);
                                    if ($code) {
                                        out += ' ' + ($code) + ' ';
                                        if ($breakOnError) {
                                            $closingBraces1 += '}';
                                        }
                                    }
                                }
                            }
                        }
                        if ($breakOnError) {
                            out += ' ' + ($closingBraces1) + ' ';
                            $closingBraces1 = '';
                        }
                        if ($rulesGroup.type) {
                            out += ' } ';
                            if ($typeSchema && $typeSchema === $rulesGroup.type && !$coerceToTypes) {
                                out += ' else { ';
                                var $schemaPath = it.schemaPath + '.type', $errSchemaPath = it.errSchemaPath + '/type';
                                var $$outStack = $$outStack || [];
                                $$outStack.push(out);
                                out = ''; /* istanbul ignore else */
                                if (it.createErrors !== false) {
                                    out += ' { keyword: \'' + ($errorKeyword || 'type') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { type: \'';
                                    if ($typeIsArray) {
                                        out += '' + ($typeSchema.join(","));
                                    }
                                    else {
                                        out += '' + ($typeSchema);
                                    }
                                    out += '\' } ';
                                    if (it.opts.messages !== false) {
                                        out += ' , message: \'should be ';
                                        if ($typeIsArray) {
                                            out += '' + ($typeSchema.join(","));
                                        }
                                        else {
                                            out += '' + ($typeSchema);
                                        }
                                        out += '\' ';
                                    }
                                    if (it.opts.verbose) {
                                        out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                                    }
                                    out += ' } ';
                                }
                                else {
                                    out += ' {} ';
                                }
                                var __err = out;
                                out = $$outStack.pop();
                                if (!it.compositeRule && $breakOnError) {
                                    /* istanbul ignore if */
                                    if (it.async) {
                                        out += ' throw new ValidationError([' + (__err) + ']); ';
                                    }
                                    else {
                                        out += ' validate.errors = [' + (__err) + ']; return false; ';
                                    }
                                }
                                else {
                                    out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                                }
                                out += ' } ';
                            }
                        }
                        if ($breakOnError) {
                            out += ' if (errors === ';
                            if ($top) {
                                out += '0';
                            }
                            else {
                                out += 'errs_' + ($lvl);
                            }
                            out += ') { ';
                            $closingBraces2 += '}';
                        }
                    }
                }
            }
        }
        if ($breakOnError) {
            out += ' ' + ($closingBraces2) + ' ';
        }
        if ($top) {
            if ($async) {
                out += ' if (errors === 0) return data;           ';
                out += ' else throw new ValidationError(vErrors); ';
            }
            else {
                out += ' validate.errors = vErrors; ';
                out += ' return errors === 0;       ';
            }
            out += ' }; return validate;';
        }
        else {
            out += ' var ' + ($valid) + ' = errors === errs_' + ($lvl) + ';';
        }
        function $shouldUseGroup($rulesGroup) {
            var rules = $rulesGroup.rules;
            for (var i = 0; i < rules.length; i++)
                if ($shouldUseRule(rules[i]))
                    return true;
        }
        function $shouldUseRule($rule) {
            return it.schema[$rule.keyword] !== undefined || ($rule.implements && $ruleImplementsSomeKeyword($rule));
        }
        function $ruleImplementsSomeKeyword($rule) {
            var impl = $rule.implements;
            for (var i = 0; i < impl.length; i++)
                if (it.schema[impl[i]] !== undefined)
                    return true;
        }
        return out;
    };
},
/* ajv/lib/cache.js */ function _(require, module, exports, __esModule, __esExport) {
    var Cache = module.exports = function Cache() {
        this._cache = {};
    };
    Cache.prototype.put = function Cache_put(key, value) {
        this._cache[key] = value;
    };
    Cache.prototype.get = function Cache_get(key) {
        return this._cache[key];
    };
    Cache.prototype.del = function Cache_del(key) {
        delete this._cache[key];
    };
    Cache.prototype.clear = function Cache_clear() {
        this._cache = {};
    };
},
/* ajv/lib/compile/formats.js */ function _(require, module, exports, __esModule, __esExport) {
    var util = require(228) /* ./util */;
    var DATE = /^(\d\d\d\d)-(\d\d)-(\d\d)$/;
    var DAYS = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31];
    var TIME = /^(\d\d):(\d\d):(\d\d)(\.\d+)?(z|[+-]\d\d(?::?\d\d)?)?$/i;
    var HOSTNAME = /^(?=.{1,253}\.?$)[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?(?:\.[a-z0-9](?:[-0-9a-z]{0,61}[0-9a-z])?)*\.?$/i;
    var URI = /^(?:[a-z][a-z0-9+\-.]*:)(?:\/?\/(?:(?:[a-z0-9\-._~!$&'()*+,;=:]|%[0-9a-f]{2})*@)?(?:\[(?:(?:(?:(?:[0-9a-f]{1,4}:){6}|::(?:[0-9a-f]{1,4}:){5}|(?:[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){4}|(?:(?:[0-9a-f]{1,4}:){0,1}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){3}|(?:(?:[0-9a-f]{1,4}:){0,2}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){2}|(?:(?:[0-9a-f]{1,4}:){0,3}[0-9a-f]{1,4})?::[0-9a-f]{1,4}:|(?:(?:[0-9a-f]{1,4}:){0,4}[0-9a-f]{1,4})?::)(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?))|(?:(?:[0-9a-f]{1,4}:){0,5}[0-9a-f]{1,4})?::[0-9a-f]{1,4}|(?:(?:[0-9a-f]{1,4}:){0,6}[0-9a-f]{1,4})?::)|[Vv][0-9a-f]+\.[a-z0-9\-._~!$&'()*+,;=:]+)\]|(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?)|(?:[a-z0-9\-._~!$&'()*+,;=]|%[0-9a-f]{2})*)(?::\d*)?(?:\/(?:[a-z0-9\-._~!$&'()*+,;=:@]|%[0-9a-f]{2})*)*|\/(?:(?:[a-z0-9\-._~!$&'()*+,;=:@]|%[0-9a-f]{2})+(?:\/(?:[a-z0-9\-._~!$&'()*+,;=:@]|%[0-9a-f]{2})*)*)?|(?:[a-z0-9\-._~!$&'()*+,;=:@]|%[0-9a-f]{2})+(?:\/(?:[a-z0-9\-._~!$&'()*+,;=:@]|%[0-9a-f]{2})*)*)(?:\?(?:[a-z0-9\-._~!$&'()*+,;=:@/?]|%[0-9a-f]{2})*)?(?:#(?:[a-z0-9\-._~!$&'()*+,;=:@/?]|%[0-9a-f]{2})*)?$/i;
    var URIREF = /^(?:[a-z][a-z0-9+\-.]*:)?(?:\/?\/(?:(?:[a-z0-9\-._~!$&'()*+,;=:]|%[0-9a-f]{2})*@)?(?:\[(?:(?:(?:(?:[0-9a-f]{1,4}:){6}|::(?:[0-9a-f]{1,4}:){5}|(?:[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){4}|(?:(?:[0-9a-f]{1,4}:){0,1}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){3}|(?:(?:[0-9a-f]{1,4}:){0,2}[0-9a-f]{1,4})?::(?:[0-9a-f]{1,4}:){2}|(?:(?:[0-9a-f]{1,4}:){0,3}[0-9a-f]{1,4})?::[0-9a-f]{1,4}:|(?:(?:[0-9a-f]{1,4}:){0,4}[0-9a-f]{1,4})?::)(?:[0-9a-f]{1,4}:[0-9a-f]{1,4}|(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?))|(?:(?:[0-9a-f]{1,4}:){0,5}[0-9a-f]{1,4})?::[0-9a-f]{1,4}|(?:(?:[0-9a-f]{1,4}:){0,6}[0-9a-f]{1,4})?::)|[Vv][0-9a-f]+\.[a-z0-9\-._~!$&'()*+,;=:]+)\]|(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?)|(?:[a-z0-9\-._~!$&'"()*+,;=]|%[0-9a-f]{2})*)(?::\d*)?(?:\/(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})*)*|\/(?:(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})+(?:\/(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})*)*)?|(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})+(?:\/(?:[a-z0-9\-._~!$&'"()*+,;=:@]|%[0-9a-f]{2})*)*)?(?:\?(?:[a-z0-9\-._~!$&'"()*+,;=:@/?]|%[0-9a-f]{2})*)?(?:#(?:[a-z0-9\-._~!$&'"()*+,;=:@/?]|%[0-9a-f]{2})*)?$/i;
    // uri-template: https://tools.ietf.org/html/rfc6570
    var URITEMPLATE = /^(?:(?:[^\x00-\x20"'<>%\\^`{|}]|%[0-9a-f]{2})|\{[+#./;?&=,!@|]?(?:[a-z0-9_]|%[0-9a-f]{2})+(?::[1-9][0-9]{0,3}|\*)?(?:,(?:[a-z0-9_]|%[0-9a-f]{2})+(?::[1-9][0-9]{0,3}|\*)?)*\})*$/i;
    // For the source: https://gist.github.com/dperini/729294
    // For test cases: https://mathiasbynens.be/demo/url-regex
    // @todo Delete current URL in favour of the commented out URL rule when this issue is fixed https://github.com/eslint/eslint/issues/7983.
    // var URL = /^(?:(?:https?|ftp):\/\/)(?:\S+(?::\S*)?@)?(?:(?!10(?:\.\d{1,3}){3})(?!127(?:\.\d{1,3}){3})(?!169\.254(?:\.\d{1,3}){2})(?!192\.168(?:\.\d{1,3}){2})(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|(?:(?:[a-z\u{00a1}-\u{ffff}0-9]+-)*[a-z\u{00a1}-\u{ffff}0-9]+)(?:\.(?:[a-z\u{00a1}-\u{ffff}0-9]+-)*[a-z\u{00a1}-\u{ffff}0-9]+)*(?:\.(?:[a-z\u{00a1}-\u{ffff}]{2,})))(?::\d{2,5})?(?:\/[^\s]*)?$/iu;
    var URL = /^(?:(?:http[s\u017F]?|ftp):\/\/)(?:(?:[\0-\x08\x0E-\x1F!-\x9F\xA1-\u167F\u1681-\u1FFF\u200B-\u2027\u202A-\u202E\u2030-\u205E\u2060-\u2FFF\u3001-\uD7FF\uE000-\uFEFE\uFF00-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])+(?::(?:[\0-\x08\x0E-\x1F!-\x9F\xA1-\u167F\u1681-\u1FFF\u200B-\u2027\u202A-\u202E\u2030-\u205E\u2060-\u2FFF\u3001-\uD7FF\uE000-\uFEFE\uFF00-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])*)?@)?(?:(?!10(?:\.[0-9]{1,3}){3})(?!127(?:\.[0-9]{1,3}){3})(?!169\.254(?:\.[0-9]{1,3}){2})(?!192\.168(?:\.[0-9]{1,3}){2})(?!172\.(?:1[6-9]|2[0-9]|3[01])(?:\.[0-9]{1,3}){2})(?:[1-9][0-9]?|1[0-9][0-9]|2[01][0-9]|22[0-3])(?:\.(?:1?[0-9]{1,2}|2[0-4][0-9]|25[0-5])){2}(?:\.(?:[1-9][0-9]?|1[0-9][0-9]|2[0-4][0-9]|25[0-4]))|(?:(?:(?:[0-9a-z\xA1-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])+-)*(?:[0-9a-z\xA1-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])+)(?:\.(?:(?:[0-9a-z\xA1-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])+-)*(?:[0-9a-z\xA1-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])+)*(?:\.(?:(?:[a-z\xA1-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF]){2,})))(?::[0-9]{2,5})?(?:\/(?:[\0-\x08\x0E-\x1F!-\x9F\xA1-\u167F\u1681-\u1FFF\u200B-\u2027\u202A-\u202E\u2030-\u205E\u2060-\u2FFF\u3001-\uD7FF\uE000-\uFEFE\uFF00-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])*)?$/i;
    var UUID = /^(?:urn:uuid:)?[0-9a-f]{8}-(?:[0-9a-f]{4}-){3}[0-9a-f]{12}$/i;
    var JSON_POINTER = /^(?:\/(?:[^~/]|~0|~1)*)*$/;
    var JSON_POINTER_URI_FRAGMENT = /^#(?:\/(?:[a-z0-9_\-.!$&'()*+,;:=@]|%[0-9a-f]{2}|~0|~1)*)*$/i;
    var RELATIVE_JSON_POINTER = /^(?:0|[1-9][0-9]*)(?:#|(?:\/(?:[^~/]|~0|~1)*)*)$/;
    module.exports = formats;
    function formats(mode) {
        mode = mode == 'full' ? 'full' : 'fast';
        return util.copy(formats[mode]);
    }
    formats.fast = {
        // date: http://tools.ietf.org/html/rfc3339#section-5.6
        date: /^\d\d\d\d-[0-1]\d-[0-3]\d$/,
        // date-time: http://tools.ietf.org/html/rfc3339#section-5.6
        time: /^(?:[0-2]\d:[0-5]\d:[0-5]\d|23:59:60)(?:\.\d+)?(?:z|[+-]\d\d(?::?\d\d)?)?$/i,
        'date-time': /^\d\d\d\d-[0-1]\d-[0-3]\d[t\s](?:[0-2]\d:[0-5]\d:[0-5]\d|23:59:60)(?:\.\d+)?(?:z|[+-]\d\d(?::?\d\d)?)$/i,
        // uri: https://github.com/mafintosh/is-my-json-valid/blob/master/formats.js
        uri: /^(?:[a-z][a-z0-9+\-.]*:)(?:\/?\/)?[^\s]*$/i,
        'uri-reference': /^(?:(?:[a-z][a-z0-9+\-.]*:)?\/?\/)?(?:[^\\\s#][^\s#]*)?(?:#[^\\\s]*)?$/i,
        'uri-template': URITEMPLATE,
        url: URL,
        // email (sources from jsen validator):
        // http://stackoverflow.com/questions/201323/using-a-regular-expression-to-validate-an-email-address#answer-8829363
        // http://www.w3.org/TR/html5/forms.html#valid-e-mail-address (search for 'willful violation')
        email: /^[a-z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?(?:\.[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?)*$/i,
        hostname: HOSTNAME,
        // optimized https://www.safaribooksonline.com/library/view/regular-expressions-cookbook/9780596802837/ch07s16.html
        ipv4: /^(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?)$/,
        // optimized http://stackoverflow.com/questions/53497/regular-expression-that-matches-valid-ipv6-addresses
        ipv6: /^\s*(?:(?:(?:[0-9a-f]{1,4}:){7}(?:[0-9a-f]{1,4}|:))|(?:(?:[0-9a-f]{1,4}:){6}(?::[0-9a-f]{1,4}|(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(?:(?:[0-9a-f]{1,4}:){5}(?:(?:(?::[0-9a-f]{1,4}){1,2})|:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(?:(?:[0-9a-f]{1,4}:){4}(?:(?:(?::[0-9a-f]{1,4}){1,3})|(?:(?::[0-9a-f]{1,4})?:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(?:(?:[0-9a-f]{1,4}:){3}(?:(?:(?::[0-9a-f]{1,4}){1,4})|(?:(?::[0-9a-f]{1,4}){0,2}:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(?:(?:[0-9a-f]{1,4}:){2}(?:(?:(?::[0-9a-f]{1,4}){1,5})|(?:(?::[0-9a-f]{1,4}){0,3}:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(?:(?:[0-9a-f]{1,4}:){1}(?:(?:(?::[0-9a-f]{1,4}){1,6})|(?:(?::[0-9a-f]{1,4}){0,4}:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(?::(?:(?:(?::[0-9a-f]{1,4}){1,7})|(?:(?::[0-9a-f]{1,4}){0,5}:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:)))(?:%.+)?\s*$/i,
        regex: regex,
        // uuid: http://tools.ietf.org/html/rfc4122
        uuid: UUID,
        // JSON-pointer: https://tools.ietf.org/html/rfc6901
        // uri fragment: https://tools.ietf.org/html/rfc3986#appendix-A
        'json-pointer': JSON_POINTER,
        'json-pointer-uri-fragment': JSON_POINTER_URI_FRAGMENT,
        // relative JSON-pointer: http://tools.ietf.org/html/draft-luff-relative-json-pointer-00
        'relative-json-pointer': RELATIVE_JSON_POINTER
    };
    formats.full = {
        date: date,
        time: time,
        'date-time': date_time,
        uri: uri,
        'uri-reference': URIREF,
        'uri-template': URITEMPLATE,
        url: URL,
        email: /^[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?$/i,
        hostname: HOSTNAME,
        ipv4: /^(?:(?:25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(?:25[0-5]|2[0-4]\d|[01]?\d\d?)$/,
        ipv6: /^\s*(?:(?:(?:[0-9a-f]{1,4}:){7}(?:[0-9a-f]{1,4}|:))|(?:(?:[0-9a-f]{1,4}:){6}(?::[0-9a-f]{1,4}|(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(?:(?:[0-9a-f]{1,4}:){5}(?:(?:(?::[0-9a-f]{1,4}){1,2})|:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(?:(?:[0-9a-f]{1,4}:){4}(?:(?:(?::[0-9a-f]{1,4}){1,3})|(?:(?::[0-9a-f]{1,4})?:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(?:(?:[0-9a-f]{1,4}:){3}(?:(?:(?::[0-9a-f]{1,4}){1,4})|(?:(?::[0-9a-f]{1,4}){0,2}:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(?:(?:[0-9a-f]{1,4}:){2}(?:(?:(?::[0-9a-f]{1,4}){1,5})|(?:(?::[0-9a-f]{1,4}){0,3}:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(?:(?:[0-9a-f]{1,4}:){1}(?:(?:(?::[0-9a-f]{1,4}){1,6})|(?:(?::[0-9a-f]{1,4}){0,4}:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(?::(?:(?:(?::[0-9a-f]{1,4}){1,7})|(?:(?::[0-9a-f]{1,4}){0,5}:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:)))(?:%.+)?\s*$/i,
        regex: regex,
        uuid: UUID,
        'json-pointer': JSON_POINTER,
        'json-pointer-uri-fragment': JSON_POINTER_URI_FRAGMENT,
        'relative-json-pointer': RELATIVE_JSON_POINTER
    };
    function isLeapYear(year) {
        // https://tools.ietf.org/html/rfc3339#appendix-C
        return year % 4 === 0 && (year % 100 !== 0 || year % 400 === 0);
    }
    function date(str) {
        // full-date from http://tools.ietf.org/html/rfc3339#section-5.6
        var matches = str.match(DATE);
        if (!matches)
            return false;
        var year = +matches[1];
        var month = +matches[2];
        var day = +matches[3];
        return month >= 1 && month <= 12 && day >= 1 &&
            day <= (month == 2 && isLeapYear(year) ? 29 : DAYS[month]);
    }
    function time(str, full) {
        var matches = str.match(TIME);
        if (!matches)
            return false;
        var hour = matches[1];
        var minute = matches[2];
        var second = matches[3];
        var timeZone = matches[5];
        return ((hour <= 23 && minute <= 59 && second <= 59) ||
            (hour == 23 && minute == 59 && second == 60)) &&
            (!full || timeZone);
    }
    var DATE_TIME_SEPARATOR = /t|\s/i;
    function date_time(str) {
        // http://tools.ietf.org/html/rfc3339#section-5.6
        var dateTime = str.split(DATE_TIME_SEPARATOR);
        return dateTime.length == 2 && date(dateTime[0]) && time(dateTime[1], true);
    }
    var NOT_URI_FRAGMENT = /\/|:/;
    function uri(str) {
        // http://jmrware.com/articles/2009/uri_regexp/URI_regex.html + optional protocol + required "."
        return NOT_URI_FRAGMENT.test(str) && URI.test(str);
    }
    var Z_ANCHOR = /[^\\]\\Z/;
    function regex(str) {
        if (Z_ANCHOR.test(str))
            return false;
        try {
            new RegExp(str);
            return true;
        }
        catch (e) {
            return false;
        }
    }
},
/* ajv/lib/compile/rules.js */ function _(require, module, exports, __esModule, __esExport) {
    var ruleModules = require(238) /* ../dotjs */, toHash = require(228) /* ./util */.toHash;
    module.exports = function rules() {
        var RULES = [
            { type: 'number',
                rules: [{ 'maximum': ['exclusiveMaximum'] },
                    { 'minimum': ['exclusiveMinimum'] }, 'multipleOf', 'format'] },
            { type: 'string',
                rules: ['maxLength', 'minLength', 'pattern', 'format'] },
            { type: 'array',
                rules: ['maxItems', 'minItems', 'items', 'contains', 'uniqueItems'] },
            { type: 'object',
                rules: ['maxProperties', 'minProperties', 'required', 'dependencies', 'propertyNames',
                    { 'properties': ['additionalProperties', 'patternProperties'] }] },
            { rules: ['$ref', 'const', 'enum', 'not', 'anyOf', 'oneOf', 'allOf', 'if'] }
        ];
        var ALL = ['type', '$comment'];
        var KEYWORDS = [
            '$schema', '$id', 'id', '$data', '$async', 'title',
            'description', 'default', 'definitions',
            'examples', 'readOnly', 'writeOnly',
            'contentMediaType', 'contentEncoding',
            'additionalItems', 'then', 'else'
        ];
        var TYPES = ['number', 'integer', 'string', 'array', 'object', 'boolean', 'null'];
        RULES.all = toHash(ALL);
        RULES.types = toHash(TYPES);
        RULES.forEach(function (group) {
            group.rules = group.rules.map(function (keyword) {
                var implKeywords;
                if (typeof keyword == 'object') {
                    var key = Object.keys(keyword)[0];
                    implKeywords = keyword[key];
                    keyword = key;
                    implKeywords.forEach(function (k) {
                        ALL.push(k);
                        RULES.all[k] = true;
                    });
                }
                ALL.push(keyword);
                var rule = RULES.all[keyword] = {
                    keyword: keyword,
                    code: ruleModules[keyword],
                    implements: implKeywords
                };
                return rule;
            });
            RULES.all.$comment = {
                keyword: '$comment',
                code: ruleModules.$comment
            };
            if (group.type)
                RULES.types[group.type] = group;
        });
        RULES.keywords = toHash(ALL.concat(KEYWORDS));
        RULES.custom = {};
        return RULES;
    };
},
/* ajv/lib/dotjs/index.js */ function _(require, module, exports, __esModule, __esExport) {
    //all requires must be explicit because browserify won't work with dynamic requires
    module.exports = {
        '$ref': require(239) /* ./ref */,
        allOf: require(240) /* ./allOf */,
        anyOf: require(241) /* ./anyOf */,
        '$comment': require(242) /* ./comment */,
        const: require(243) /* ./const */,
        contains: require(244) /* ./contains */,
        dependencies: require(245) /* ./dependencies */,
        'enum': require(246) /* ./enum */,
        format: require(247) /* ./format */,
        'if': require(248) /* ./if */,
        items: require(249) /* ./items */,
        maximum: require(250) /* ./_limit */,
        minimum: require(250) /* ./_limit */,
        maxItems: require(251) /* ./_limitItems */,
        minItems: require(251) /* ./_limitItems */,
        maxLength: require(252) /* ./_limitLength */,
        minLength: require(252) /* ./_limitLength */,
        maxProperties: require(253) /* ./_limitProperties */,
        minProperties: require(253) /* ./_limitProperties */,
        multipleOf: require(254) /* ./multipleOf */,
        not: require(255) /* ./not */,
        oneOf: require(256) /* ./oneOf */,
        pattern: require(257) /* ./pattern */,
        properties: require(258) /* ./properties */,
        propertyNames: require(259) /* ./propertyNames */,
        required: require(260) /* ./required */,
        uniqueItems: require(261) /* ./uniqueItems */,
        validate: require(234) /* ./validate */
    };
},
/* ajv/lib/dotjs/ref.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_ref(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $async, $refCode;
        if ($schema == '#' || $schema == '#/') {
            if (it.isRoot) {
                $async = it.async;
                $refCode = 'validate';
            }
            else {
                $async = it.root.schema.$async === true;
                $refCode = 'root.refVal[0]';
            }
        }
        else {
            var $refVal = it.resolveRef(it.baseId, $schema, it.isRoot);
            if ($refVal === undefined) {
                var $message = it.MissingRefError.message(it.baseId, $schema);
                if (it.opts.missingRefs == 'fail') {
                    it.logger.error($message);
                    var $$outStack = $$outStack || [];
                    $$outStack.push(out);
                    out = ''; /* istanbul ignore else */
                    if (it.createErrors !== false) {
                        out += ' { keyword: \'' + ('$ref') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { ref: \'' + (it.util.escapeQuotes($schema)) + '\' } ';
                        if (it.opts.messages !== false) {
                            out += ' , message: \'can\\\'t resolve reference ' + (it.util.escapeQuotes($schema)) + '\' ';
                        }
                        if (it.opts.verbose) {
                            out += ' , schema: ' + (it.util.toQuotedString($schema)) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                        }
                        out += ' } ';
                    }
                    else {
                        out += ' {} ';
                    }
                    var __err = out;
                    out = $$outStack.pop();
                    if (!it.compositeRule && $breakOnError) {
                        /* istanbul ignore if */
                        if (it.async) {
                            out += ' throw new ValidationError([' + (__err) + ']); ';
                        }
                        else {
                            out += ' validate.errors = [' + (__err) + ']; return false; ';
                        }
                    }
                    else {
                        out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                    }
                    if ($breakOnError) {
                        out += ' if (false) { ';
                    }
                }
                else if (it.opts.missingRefs == 'ignore') {
                    it.logger.warn($message);
                    if ($breakOnError) {
                        out += ' if (true) { ';
                    }
                }
                else {
                    throw new it.MissingRefError(it.baseId, $schema, $message);
                }
            }
            else if ($refVal.inline) {
                var $it = it.util.copy(it);
                $it.level++;
                var $nextValid = 'valid' + $it.level;
                $it.schema = $refVal.schema;
                $it.schemaPath = '';
                $it.errSchemaPath = $schema;
                var $code = it.validate($it).replace(/validate\.schema/g, $refVal.code);
                out += ' ' + ($code) + ' ';
                if ($breakOnError) {
                    out += ' if (' + ($nextValid) + ') { ';
                }
            }
            else {
                $async = $refVal.$async === true || (it.async && $refVal.$async !== false);
                $refCode = $refVal.code;
            }
        }
        if ($refCode) {
            var $$outStack = $$outStack || [];
            $$outStack.push(out);
            out = '';
            if (it.opts.passContext) {
                out += ' ' + ($refCode) + '.call(this, ';
            }
            else {
                out += ' ' + ($refCode) + '( ';
            }
            out += ' ' + ($data) + ', (dataPath || \'\')';
            if (it.errorPath != '""') {
                out += ' + ' + (it.errorPath);
            }
            var $parentData = $dataLvl ? 'data' + (($dataLvl - 1) || '') : 'parentData', $parentDataProperty = $dataLvl ? it.dataPathArr[$dataLvl] : 'parentDataProperty';
            out += ' , ' + ($parentData) + ' , ' + ($parentDataProperty) + ', rootData)  ';
            var __callValidate = out;
            out = $$outStack.pop();
            if ($async) {
                if (!it.async)
                    throw new Error('async schema referenced by sync schema');
                if ($breakOnError) {
                    out += ' var ' + ($valid) + '; ';
                }
                out += ' try { await ' + (__callValidate) + '; ';
                if ($breakOnError) {
                    out += ' ' + ($valid) + ' = true; ';
                }
                out += ' } catch (e) { if (!(e instanceof ValidationError)) throw e; if (vErrors === null) vErrors = e.errors; else vErrors = vErrors.concat(e.errors); errors = vErrors.length; ';
                if ($breakOnError) {
                    out += ' ' + ($valid) + ' = false; ';
                }
                out += ' } ';
                if ($breakOnError) {
                    out += ' if (' + ($valid) + ') { ';
                }
            }
            else {
                out += ' if (!' + (__callValidate) + ') { if (vErrors === null) vErrors = ' + ($refCode) + '.errors; else vErrors = vErrors.concat(' + ($refCode) + '.errors); errors = vErrors.length; } ';
                if ($breakOnError) {
                    out += ' else { ';
                }
            }
        }
        return out;
    };
},
/* ajv/lib/dotjs/allOf.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_allOf(it, $keyword, $ruleType) {
        var out = ' ';
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $it = it.util.copy(it);
        var $closingBraces = '';
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        var $currentBaseId = $it.baseId, $allSchemasEmpty = true;
        var arr1 = $schema;
        if (arr1) {
            var $sch, $i = -1, l1 = arr1.length - 1;
            while ($i < l1) {
                $sch = arr1[$i += 1];
                if ((it.opts.strictKeywords ? (typeof $sch == 'object' && Object.keys($sch).length > 0) || $sch === false : it.util.schemaHasRules($sch, it.RULES.all))) {
                    $allSchemasEmpty = false;
                    $it.schema = $sch;
                    $it.schemaPath = $schemaPath + '[' + $i + ']';
                    $it.errSchemaPath = $errSchemaPath + '/' + $i;
                    out += '  ' + (it.validate($it)) + ' ';
                    $it.baseId = $currentBaseId;
                    if ($breakOnError) {
                        out += ' if (' + ($nextValid) + ') { ';
                        $closingBraces += '}';
                    }
                }
            }
        }
        if ($breakOnError) {
            if ($allSchemasEmpty) {
                out += ' if (true) { ';
            }
            else {
                out += ' ' + ($closingBraces.slice(0, -1)) + ' ';
            }
        }
        return out;
    };
},
/* ajv/lib/dotjs/anyOf.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_anyOf(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $errs = 'errs__' + $lvl;
        var $it = it.util.copy(it);
        var $closingBraces = '';
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        var $noEmptySchema = $schema.every(function ($sch) {
            return (it.opts.strictKeywords ? (typeof $sch == 'object' && Object.keys($sch).length > 0) || $sch === false : it.util.schemaHasRules($sch, it.RULES.all));
        });
        if ($noEmptySchema) {
            var $currentBaseId = $it.baseId;
            out += ' var ' + ($errs) + ' = errors; var ' + ($valid) + ' = false;  ';
            var $wasComposite = it.compositeRule;
            it.compositeRule = $it.compositeRule = true;
            var arr1 = $schema;
            if (arr1) {
                var $sch, $i = -1, l1 = arr1.length - 1;
                while ($i < l1) {
                    $sch = arr1[$i += 1];
                    $it.schema = $sch;
                    $it.schemaPath = $schemaPath + '[' + $i + ']';
                    $it.errSchemaPath = $errSchemaPath + '/' + $i;
                    out += '  ' + (it.validate($it)) + ' ';
                    $it.baseId = $currentBaseId;
                    out += ' ' + ($valid) + ' = ' + ($valid) + ' || ' + ($nextValid) + '; if (!' + ($valid) + ') { ';
                    $closingBraces += '}';
                }
            }
            it.compositeRule = $it.compositeRule = $wasComposite;
            out += ' ' + ($closingBraces) + ' if (!' + ($valid) + ') {   var err =   '; /* istanbul ignore else */
            if (it.createErrors !== false) {
                out += ' { keyword: \'' + ('anyOf') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: {} ';
                if (it.opts.messages !== false) {
                    out += ' , message: \'should match some schema in anyOf\' ';
                }
                if (it.opts.verbose) {
                    out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                }
                out += ' } ';
            }
            else {
                out += ' {} ';
            }
            out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
            if (!it.compositeRule && $breakOnError) {
                /* istanbul ignore if */
                if (it.async) {
                    out += ' throw new ValidationError(vErrors); ';
                }
                else {
                    out += ' validate.errors = vErrors; return false; ';
                }
            }
            out += ' } else {  errors = ' + ($errs) + '; if (vErrors !== null) { if (' + ($errs) + ') vErrors.length = ' + ($errs) + '; else vErrors = null; } ';
            if (it.opts.allErrors) {
                out += ' } ';
            }
        }
        else {
            if ($breakOnError) {
                out += ' if (true) { ';
            }
        }
        return out;
    };
},
/* ajv/lib/dotjs/comment.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_comment(it, $keyword, $ruleType) {
        var out = ' ';
        var $schema = it.schema[$keyword];
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $comment = it.util.toQuotedString($schema);
        if (it.opts.$comment === true) {
            out += ' console.log(' + ($comment) + ');';
        }
        else if (typeof it.opts.$comment == 'function') {
            out += ' self._opts.$comment(' + ($comment) + ', ' + (it.util.toQuotedString($errSchemaPath)) + ', validate.root.schema);';
        }
        return out;
    };
},
/* ajv/lib/dotjs/const.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_const(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        if (!$isData) {
            out += ' var schema' + ($lvl) + ' = validate.schema' + ($schemaPath) + ';';
        }
        out += 'var ' + ($valid) + ' = equal(' + ($data) + ', schema' + ($lvl) + '); if (!' + ($valid) + ') {   ';
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ('const') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { allowedValue: schema' + ($lvl) + ' } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should be equal to constant\' ';
            }
            if (it.opts.verbose) {
                out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += ' }';
        if ($breakOnError) {
            out += ' else { ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/contains.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_contains(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $errs = 'errs__' + $lvl;
        var $it = it.util.copy(it);
        var $closingBraces = '';
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        var $idx = 'i' + $lvl, $dataNxt = $it.dataLevel = it.dataLevel + 1, $nextData = 'data' + $dataNxt, $currentBaseId = it.baseId, $nonEmptySchema = (it.opts.strictKeywords ? (typeof $schema == 'object' && Object.keys($schema).length > 0) || $schema === false : it.util.schemaHasRules($schema, it.RULES.all));
        out += 'var ' + ($errs) + ' = errors;var ' + ($valid) + ';';
        if ($nonEmptySchema) {
            var $wasComposite = it.compositeRule;
            it.compositeRule = $it.compositeRule = true;
            $it.schema = $schema;
            $it.schemaPath = $schemaPath;
            $it.errSchemaPath = $errSchemaPath;
            out += ' var ' + ($nextValid) + ' = false; for (var ' + ($idx) + ' = 0; ' + ($idx) + ' < ' + ($data) + '.length; ' + ($idx) + '++) { ';
            $it.errorPath = it.util.getPathExpr(it.errorPath, $idx, it.opts.jsonPointers, true);
            var $passData = $data + '[' + $idx + ']';
            $it.dataPathArr[$dataNxt] = $idx;
            var $code = it.validate($it);
            $it.baseId = $currentBaseId;
            if (it.util.varOccurences($code, $nextData) < 2) {
                out += ' ' + (it.util.varReplace($code, $nextData, $passData)) + ' ';
            }
            else {
                out += ' var ' + ($nextData) + ' = ' + ($passData) + '; ' + ($code) + ' ';
            }
            out += ' if (' + ($nextValid) + ') break; }  ';
            it.compositeRule = $it.compositeRule = $wasComposite;
            out += ' ' + ($closingBraces) + ' if (!' + ($nextValid) + ') {';
        }
        else {
            out += ' if (' + ($data) + '.length == 0) {';
        }
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ('contains') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: {} ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should contain a valid item\' ';
            }
            if (it.opts.verbose) {
                out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += ' } else { ';
        if ($nonEmptySchema) {
            out += '  errors = ' + ($errs) + '; if (vErrors !== null) { if (' + ($errs) + ') vErrors.length = ' + ($errs) + '; else vErrors = null; } ';
        }
        if (it.opts.allErrors) {
            out += ' } ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/dependencies.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_dependencies(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $errs = 'errs__' + $lvl;
        var $it = it.util.copy(it);
        var $closingBraces = '';
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        var $schemaDeps = {}, $propertyDeps = {}, $ownProperties = it.opts.ownProperties;
        for ($property in $schema) {
            if ($property == '__proto__')
                continue;
            var $sch = $schema[$property];
            var $deps = Array.isArray($sch) ? $propertyDeps : $schemaDeps;
            $deps[$property] = $sch;
        }
        out += 'var ' + ($errs) + ' = errors;';
        var $currentErrorPath = it.errorPath;
        out += 'var missing' + ($lvl) + ';';
        for (var $property in $propertyDeps) {
            $deps = $propertyDeps[$property];
            if ($deps.length) {
                out += ' if ( ' + ($data) + (it.util.getProperty($property)) + ' !== undefined ';
                if ($ownProperties) {
                    out += ' && Object.prototype.hasOwnProperty.call(' + ($data) + ', \'' + (it.util.escapeQuotes($property)) + '\') ';
                }
                if ($breakOnError) {
                    out += ' && ( ';
                    var arr1 = $deps;
                    if (arr1) {
                        var $propertyKey, $i = -1, l1 = arr1.length - 1;
                        while ($i < l1) {
                            $propertyKey = arr1[$i += 1];
                            if ($i) {
                                out += ' || ';
                            }
                            var $prop = it.util.getProperty($propertyKey), $useData = $data + $prop;
                            out += ' ( ( ' + ($useData) + ' === undefined ';
                            if ($ownProperties) {
                                out += ' || ! Object.prototype.hasOwnProperty.call(' + ($data) + ', \'' + (it.util.escapeQuotes($propertyKey)) + '\') ';
                            }
                            out += ') && (missing' + ($lvl) + ' = ' + (it.util.toQuotedString(it.opts.jsonPointers ? $propertyKey : $prop)) + ') ) ';
                        }
                    }
                    out += ')) {  ';
                    var $propertyPath = 'missing' + $lvl, $missingProperty = '\' + ' + $propertyPath + ' + \'';
                    if (it.opts._errorDataPathProperty) {
                        it.errorPath = it.opts.jsonPointers ? it.util.getPathExpr($currentErrorPath, $propertyPath, true) : $currentErrorPath + ' + ' + $propertyPath;
                    }
                    var $$outStack = $$outStack || [];
                    $$outStack.push(out);
                    out = ''; /* istanbul ignore else */
                    if (it.createErrors !== false) {
                        out += ' { keyword: \'' + ('dependencies') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { property: \'' + (it.util.escapeQuotes($property)) + '\', missingProperty: \'' + ($missingProperty) + '\', depsCount: ' + ($deps.length) + ', deps: \'' + (it.util.escapeQuotes($deps.length == 1 ? $deps[0] : $deps.join(", "))) + '\' } ';
                        if (it.opts.messages !== false) {
                            out += ' , message: \'should have ';
                            if ($deps.length == 1) {
                                out += 'property ' + (it.util.escapeQuotes($deps[0]));
                            }
                            else {
                                out += 'properties ' + (it.util.escapeQuotes($deps.join(", ")));
                            }
                            out += ' when property ' + (it.util.escapeQuotes($property)) + ' is present\' ';
                        }
                        if (it.opts.verbose) {
                            out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                        }
                        out += ' } ';
                    }
                    else {
                        out += ' {} ';
                    }
                    var __err = out;
                    out = $$outStack.pop();
                    if (!it.compositeRule && $breakOnError) {
                        /* istanbul ignore if */
                        if (it.async) {
                            out += ' throw new ValidationError([' + (__err) + ']); ';
                        }
                        else {
                            out += ' validate.errors = [' + (__err) + ']; return false; ';
                        }
                    }
                    else {
                        out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                    }
                }
                else {
                    out += ' ) { ';
                    var arr2 = $deps;
                    if (arr2) {
                        var $propertyKey, i2 = -1, l2 = arr2.length - 1;
                        while (i2 < l2) {
                            $propertyKey = arr2[i2 += 1];
                            var $prop = it.util.getProperty($propertyKey), $missingProperty = it.util.escapeQuotes($propertyKey), $useData = $data + $prop;
                            if (it.opts._errorDataPathProperty) {
                                it.errorPath = it.util.getPath($currentErrorPath, $propertyKey, it.opts.jsonPointers);
                            }
                            out += ' if ( ' + ($useData) + ' === undefined ';
                            if ($ownProperties) {
                                out += ' || ! Object.prototype.hasOwnProperty.call(' + ($data) + ', \'' + (it.util.escapeQuotes($propertyKey)) + '\') ';
                            }
                            out += ') {  var err =   '; /* istanbul ignore else */
                            if (it.createErrors !== false) {
                                out += ' { keyword: \'' + ('dependencies') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { property: \'' + (it.util.escapeQuotes($property)) + '\', missingProperty: \'' + ($missingProperty) + '\', depsCount: ' + ($deps.length) + ', deps: \'' + (it.util.escapeQuotes($deps.length == 1 ? $deps[0] : $deps.join(", "))) + '\' } ';
                                if (it.opts.messages !== false) {
                                    out += ' , message: \'should have ';
                                    if ($deps.length == 1) {
                                        out += 'property ' + (it.util.escapeQuotes($deps[0]));
                                    }
                                    else {
                                        out += 'properties ' + (it.util.escapeQuotes($deps.join(", ")));
                                    }
                                    out += ' when property ' + (it.util.escapeQuotes($property)) + ' is present\' ';
                                }
                                if (it.opts.verbose) {
                                    out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                                }
                                out += ' } ';
                            }
                            else {
                                out += ' {} ';
                            }
                            out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; } ';
                        }
                    }
                }
                out += ' }   ';
                if ($breakOnError) {
                    $closingBraces += '}';
                    out += ' else { ';
                }
            }
        }
        it.errorPath = $currentErrorPath;
        var $currentBaseId = $it.baseId;
        for (var $property in $schemaDeps) {
            var $sch = $schemaDeps[$property];
            if ((it.opts.strictKeywords ? (typeof $sch == 'object' && Object.keys($sch).length > 0) || $sch === false : it.util.schemaHasRules($sch, it.RULES.all))) {
                out += ' ' + ($nextValid) + ' = true; if ( ' + ($data) + (it.util.getProperty($property)) + ' !== undefined ';
                if ($ownProperties) {
                    out += ' && Object.prototype.hasOwnProperty.call(' + ($data) + ', \'' + (it.util.escapeQuotes($property)) + '\') ';
                }
                out += ') { ';
                $it.schema = $sch;
                $it.schemaPath = $schemaPath + it.util.getProperty($property);
                $it.errSchemaPath = $errSchemaPath + '/' + it.util.escapeFragment($property);
                out += '  ' + (it.validate($it)) + ' ';
                $it.baseId = $currentBaseId;
                out += ' }  ';
                if ($breakOnError) {
                    out += ' if (' + ($nextValid) + ') { ';
                    $closingBraces += '}';
                }
            }
        }
        if ($breakOnError) {
            out += '   ' + ($closingBraces) + ' if (' + ($errs) + ' == errors) {';
        }
        return out;
    };
},
/* ajv/lib/dotjs/enum.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_enum(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        var $i = 'i' + $lvl, $vSchema = 'schema' + $lvl;
        if (!$isData) {
            out += ' var ' + ($vSchema) + ' = validate.schema' + ($schemaPath) + ';';
        }
        out += 'var ' + ($valid) + ';';
        if ($isData) {
            out += ' if (schema' + ($lvl) + ' === undefined) ' + ($valid) + ' = true; else if (!Array.isArray(schema' + ($lvl) + ')) ' + ($valid) + ' = false; else {';
        }
        out += '' + ($valid) + ' = false;for (var ' + ($i) + '=0; ' + ($i) + '<' + ($vSchema) + '.length; ' + ($i) + '++) if (equal(' + ($data) + ', ' + ($vSchema) + '[' + ($i) + '])) { ' + ($valid) + ' = true; break; }';
        if ($isData) {
            out += '  }  ';
        }
        out += ' if (!' + ($valid) + ') {   ';
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ('enum') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { allowedValues: schema' + ($lvl) + ' } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should be equal to one of the allowed values\' ';
            }
            if (it.opts.verbose) {
                out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += ' }';
        if ($breakOnError) {
            out += ' else { ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/format.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_format(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        if (it.opts.format === false) {
            if ($breakOnError) {
                out += ' if (true) { ';
            }
            return out;
        }
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        var $unknownFormats = it.opts.unknownFormats, $allowUnknown = Array.isArray($unknownFormats);
        if ($isData) {
            var $format = 'format' + $lvl, $isObject = 'isObject' + $lvl, $formatType = 'formatType' + $lvl;
            out += ' var ' + ($format) + ' = formats[' + ($schemaValue) + ']; var ' + ($isObject) + ' = typeof ' + ($format) + ' == \'object\' && !(' + ($format) + ' instanceof RegExp) && ' + ($format) + '.validate; var ' + ($formatType) + ' = ' + ($isObject) + ' && ' + ($format) + '.type || \'string\'; if (' + ($isObject) + ') { ';
            if (it.async) {
                out += ' var async' + ($lvl) + ' = ' + ($format) + '.async; ';
            }
            out += ' ' + ($format) + ' = ' + ($format) + '.validate; } if (  ';
            if ($isData) {
                out += ' (' + ($schemaValue) + ' !== undefined && typeof ' + ($schemaValue) + ' != \'string\') || ';
            }
            out += ' (';
            if ($unknownFormats != 'ignore') {
                out += ' (' + ($schemaValue) + ' && !' + ($format) + ' ';
                if ($allowUnknown) {
                    out += ' && self._opts.unknownFormats.indexOf(' + ($schemaValue) + ') == -1 ';
                }
                out += ') || ';
            }
            out += ' (' + ($format) + ' && ' + ($formatType) + ' == \'' + ($ruleType) + '\' && !(typeof ' + ($format) + ' == \'function\' ? ';
            if (it.async) {
                out += ' (async' + ($lvl) + ' ? await ' + ($format) + '(' + ($data) + ') : ' + ($format) + '(' + ($data) + ')) ';
            }
            else {
                out += ' ' + ($format) + '(' + ($data) + ') ';
            }
            out += ' : ' + ($format) + '.test(' + ($data) + '))))) {';
        }
        else {
            var $format = it.formats[$schema];
            if (!$format) {
                if ($unknownFormats == 'ignore') {
                    it.logger.warn('unknown format "' + $schema + '" ignored in schema at path "' + it.errSchemaPath + '"');
                    if ($breakOnError) {
                        out += ' if (true) { ';
                    }
                    return out;
                }
                else if ($allowUnknown && $unknownFormats.indexOf($schema) >= 0) {
                    if ($breakOnError) {
                        out += ' if (true) { ';
                    }
                    return out;
                }
                else {
                    throw new Error('unknown format "' + $schema + '" is used in schema at path "' + it.errSchemaPath + '"');
                }
            }
            var $isObject = typeof $format == 'object' && !($format instanceof RegExp) && $format.validate;
            var $formatType = $isObject && $format.type || 'string';
            if ($isObject) {
                var $async = $format.async === true;
                $format = $format.validate;
            }
            if ($formatType != $ruleType) {
                if ($breakOnError) {
                    out += ' if (true) { ';
                }
                return out;
            }
            if ($async) {
                if (!it.async)
                    throw new Error('async format in sync schema');
                var $formatRef = 'formats' + it.util.getProperty($schema) + '.validate';
                out += ' if (!(await ' + ($formatRef) + '(' + ($data) + '))) { ';
            }
            else {
                out += ' if (! ';
                var $formatRef = 'formats' + it.util.getProperty($schema);
                if ($isObject)
                    $formatRef += '.validate';
                if (typeof $format == 'function') {
                    out += ' ' + ($formatRef) + '(' + ($data) + ') ';
                }
                else {
                    out += ' ' + ($formatRef) + '.test(' + ($data) + ') ';
                }
                out += ') { ';
            }
        }
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ('format') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { format:  ';
            if ($isData) {
                out += '' + ($schemaValue);
            }
            else {
                out += '' + (it.util.toQuotedString($schema));
            }
            out += '  } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should match format "';
                if ($isData) {
                    out += '\' + ' + ($schemaValue) + ' + \'';
                }
                else {
                    out += '' + (it.util.escapeQuotes($schema));
                }
                out += '"\' ';
            }
            if (it.opts.verbose) {
                out += ' , schema:  ';
                if ($isData) {
                    out += 'validate.schema' + ($schemaPath);
                }
                else {
                    out += '' + (it.util.toQuotedString($schema));
                }
                out += '         , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += ' } ';
        if ($breakOnError) {
            out += ' else { ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/if.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_if(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $errs = 'errs__' + $lvl;
        var $it = it.util.copy(it);
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        var $thenSch = it.schema['then'], $elseSch = it.schema['else'], $thenPresent = $thenSch !== undefined && (it.opts.strictKeywords ? (typeof $thenSch == 'object' && Object.keys($thenSch).length > 0) || $thenSch === false : it.util.schemaHasRules($thenSch, it.RULES.all)), $elsePresent = $elseSch !== undefined && (it.opts.strictKeywords ? (typeof $elseSch == 'object' && Object.keys($elseSch).length > 0) || $elseSch === false : it.util.schemaHasRules($elseSch, it.RULES.all)), $currentBaseId = $it.baseId;
        if ($thenPresent || $elsePresent) {
            var $ifClause;
            $it.createErrors = false;
            $it.schema = $schema;
            $it.schemaPath = $schemaPath;
            $it.errSchemaPath = $errSchemaPath;
            out += ' var ' + ($errs) + ' = errors; var ' + ($valid) + ' = true;  ';
            var $wasComposite = it.compositeRule;
            it.compositeRule = $it.compositeRule = true;
            out += '  ' + (it.validate($it)) + ' ';
            $it.baseId = $currentBaseId;
            $it.createErrors = true;
            out += '  errors = ' + ($errs) + '; if (vErrors !== null) { if (' + ($errs) + ') vErrors.length = ' + ($errs) + '; else vErrors = null; }  ';
            it.compositeRule = $it.compositeRule = $wasComposite;
            if ($thenPresent) {
                out += ' if (' + ($nextValid) + ') {  ';
                $it.schema = it.schema['then'];
                $it.schemaPath = it.schemaPath + '.then';
                $it.errSchemaPath = it.errSchemaPath + '/then';
                out += '  ' + (it.validate($it)) + ' ';
                $it.baseId = $currentBaseId;
                out += ' ' + ($valid) + ' = ' + ($nextValid) + '; ';
                if ($thenPresent && $elsePresent) {
                    $ifClause = 'ifClause' + $lvl;
                    out += ' var ' + ($ifClause) + ' = \'then\'; ';
                }
                else {
                    $ifClause = '\'then\'';
                }
                out += ' } ';
                if ($elsePresent) {
                    out += ' else { ';
                }
            }
            else {
                out += ' if (!' + ($nextValid) + ') { ';
            }
            if ($elsePresent) {
                $it.schema = it.schema['else'];
                $it.schemaPath = it.schemaPath + '.else';
                $it.errSchemaPath = it.errSchemaPath + '/else';
                out += '  ' + (it.validate($it)) + ' ';
                $it.baseId = $currentBaseId;
                out += ' ' + ($valid) + ' = ' + ($nextValid) + '; ';
                if ($thenPresent && $elsePresent) {
                    $ifClause = 'ifClause' + $lvl;
                    out += ' var ' + ($ifClause) + ' = \'else\'; ';
                }
                else {
                    $ifClause = '\'else\'';
                }
                out += ' } ';
            }
            out += ' if (!' + ($valid) + ') {   var err =   '; /* istanbul ignore else */
            if (it.createErrors !== false) {
                out += ' { keyword: \'' + ('if') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { failingKeyword: ' + ($ifClause) + ' } ';
                if (it.opts.messages !== false) {
                    out += ' , message: \'should match "\' + ' + ($ifClause) + ' + \'" schema\' ';
                }
                if (it.opts.verbose) {
                    out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                }
                out += ' } ';
            }
            else {
                out += ' {} ';
            }
            out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
            if (!it.compositeRule && $breakOnError) {
                /* istanbul ignore if */
                if (it.async) {
                    out += ' throw new ValidationError(vErrors); ';
                }
                else {
                    out += ' validate.errors = vErrors; return false; ';
                }
            }
            out += ' }   ';
            if ($breakOnError) {
                out += ' else { ';
            }
        }
        else {
            if ($breakOnError) {
                out += ' if (true) { ';
            }
        }
        return out;
    };
},
/* ajv/lib/dotjs/items.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_items(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $errs = 'errs__' + $lvl;
        var $it = it.util.copy(it);
        var $closingBraces = '';
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        var $idx = 'i' + $lvl, $dataNxt = $it.dataLevel = it.dataLevel + 1, $nextData = 'data' + $dataNxt, $currentBaseId = it.baseId;
        out += 'var ' + ($errs) + ' = errors;var ' + ($valid) + ';';
        if (Array.isArray($schema)) {
            var $additionalItems = it.schema.additionalItems;
            if ($additionalItems === false) {
                out += ' ' + ($valid) + ' = ' + ($data) + '.length <= ' + ($schema.length) + '; ';
                var $currErrSchemaPath = $errSchemaPath;
                $errSchemaPath = it.errSchemaPath + '/additionalItems';
                out += '  if (!' + ($valid) + ') {   ';
                var $$outStack = $$outStack || [];
                $$outStack.push(out);
                out = ''; /* istanbul ignore else */
                if (it.createErrors !== false) {
                    out += ' { keyword: \'' + ('additionalItems') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { limit: ' + ($schema.length) + ' } ';
                    if (it.opts.messages !== false) {
                        out += ' , message: \'should NOT have more than ' + ($schema.length) + ' items\' ';
                    }
                    if (it.opts.verbose) {
                        out += ' , schema: false , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                    }
                    out += ' } ';
                }
                else {
                    out += ' {} ';
                }
                var __err = out;
                out = $$outStack.pop();
                if (!it.compositeRule && $breakOnError) {
                    /* istanbul ignore if */
                    if (it.async) {
                        out += ' throw new ValidationError([' + (__err) + ']); ';
                    }
                    else {
                        out += ' validate.errors = [' + (__err) + ']; return false; ';
                    }
                }
                else {
                    out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                }
                out += ' } ';
                $errSchemaPath = $currErrSchemaPath;
                if ($breakOnError) {
                    $closingBraces += '}';
                    out += ' else { ';
                }
            }
            var arr1 = $schema;
            if (arr1) {
                var $sch, $i = -1, l1 = arr1.length - 1;
                while ($i < l1) {
                    $sch = arr1[$i += 1];
                    if ((it.opts.strictKeywords ? (typeof $sch == 'object' && Object.keys($sch).length > 0) || $sch === false : it.util.schemaHasRules($sch, it.RULES.all))) {
                        out += ' ' + ($nextValid) + ' = true; if (' + ($data) + '.length > ' + ($i) + ') { ';
                        var $passData = $data + '[' + $i + ']';
                        $it.schema = $sch;
                        $it.schemaPath = $schemaPath + '[' + $i + ']';
                        $it.errSchemaPath = $errSchemaPath + '/' + $i;
                        $it.errorPath = it.util.getPathExpr(it.errorPath, $i, it.opts.jsonPointers, true);
                        $it.dataPathArr[$dataNxt] = $i;
                        var $code = it.validate($it);
                        $it.baseId = $currentBaseId;
                        if (it.util.varOccurences($code, $nextData) < 2) {
                            out += ' ' + (it.util.varReplace($code, $nextData, $passData)) + ' ';
                        }
                        else {
                            out += ' var ' + ($nextData) + ' = ' + ($passData) + '; ' + ($code) + ' ';
                        }
                        out += ' }  ';
                        if ($breakOnError) {
                            out += ' if (' + ($nextValid) + ') { ';
                            $closingBraces += '}';
                        }
                    }
                }
            }
            if (typeof $additionalItems == 'object' && (it.opts.strictKeywords ? (typeof $additionalItems == 'object' && Object.keys($additionalItems).length > 0) || $additionalItems === false : it.util.schemaHasRules($additionalItems, it.RULES.all))) {
                $it.schema = $additionalItems;
                $it.schemaPath = it.schemaPath + '.additionalItems';
                $it.errSchemaPath = it.errSchemaPath + '/additionalItems';
                out += ' ' + ($nextValid) + ' = true; if (' + ($data) + '.length > ' + ($schema.length) + ') {  for (var ' + ($idx) + ' = ' + ($schema.length) + '; ' + ($idx) + ' < ' + ($data) + '.length; ' + ($idx) + '++) { ';
                $it.errorPath = it.util.getPathExpr(it.errorPath, $idx, it.opts.jsonPointers, true);
                var $passData = $data + '[' + $idx + ']';
                $it.dataPathArr[$dataNxt] = $idx;
                var $code = it.validate($it);
                $it.baseId = $currentBaseId;
                if (it.util.varOccurences($code, $nextData) < 2) {
                    out += ' ' + (it.util.varReplace($code, $nextData, $passData)) + ' ';
                }
                else {
                    out += ' var ' + ($nextData) + ' = ' + ($passData) + '; ' + ($code) + ' ';
                }
                if ($breakOnError) {
                    out += ' if (!' + ($nextValid) + ') break; ';
                }
                out += ' } }  ';
                if ($breakOnError) {
                    out += ' if (' + ($nextValid) + ') { ';
                    $closingBraces += '}';
                }
            }
        }
        else if ((it.opts.strictKeywords ? (typeof $schema == 'object' && Object.keys($schema).length > 0) || $schema === false : it.util.schemaHasRules($schema, it.RULES.all))) {
            $it.schema = $schema;
            $it.schemaPath = $schemaPath;
            $it.errSchemaPath = $errSchemaPath;
            out += '  for (var ' + ($idx) + ' = ' + (0) + '; ' + ($idx) + ' < ' + ($data) + '.length; ' + ($idx) + '++) { ';
            $it.errorPath = it.util.getPathExpr(it.errorPath, $idx, it.opts.jsonPointers, true);
            var $passData = $data + '[' + $idx + ']';
            $it.dataPathArr[$dataNxt] = $idx;
            var $code = it.validate($it);
            $it.baseId = $currentBaseId;
            if (it.util.varOccurences($code, $nextData) < 2) {
                out += ' ' + (it.util.varReplace($code, $nextData, $passData)) + ' ';
            }
            else {
                out += ' var ' + ($nextData) + ' = ' + ($passData) + '; ' + ($code) + ' ';
            }
            if ($breakOnError) {
                out += ' if (!' + ($nextValid) + ') break; ';
            }
            out += ' }';
        }
        if ($breakOnError) {
            out += ' ' + ($closingBraces) + ' if (' + ($errs) + ' == errors) {';
        }
        return out;
    };
},
/* ajv/lib/dotjs/_limit.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate__limit(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $errorKeyword;
        var $data = 'data' + ($dataLvl || '');
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        var $isMax = $keyword == 'maximum', $exclusiveKeyword = $isMax ? 'exclusiveMaximum' : 'exclusiveMinimum', $schemaExcl = it.schema[$exclusiveKeyword], $isDataExcl = it.opts.$data && $schemaExcl && $schemaExcl.$data, $op = $isMax ? '<' : '>', $notOp = $isMax ? '>' : '<', $errorKeyword = undefined;
        if (!($isData || typeof $schema == 'number' || $schema === undefined)) {
            throw new Error($keyword + ' must be number');
        }
        if (!($isDataExcl || $schemaExcl === undefined || typeof $schemaExcl == 'number' || typeof $schemaExcl == 'boolean')) {
            throw new Error($exclusiveKeyword + ' must be number or boolean');
        }
        if ($isDataExcl) {
            var $schemaValueExcl = it.util.getData($schemaExcl.$data, $dataLvl, it.dataPathArr), $exclusive = 'exclusive' + $lvl, $exclType = 'exclType' + $lvl, $exclIsNumber = 'exclIsNumber' + $lvl, $opExpr = 'op' + $lvl, $opStr = '\' + ' + $opExpr + ' + \'';
            out += ' var schemaExcl' + ($lvl) + ' = ' + ($schemaValueExcl) + '; ';
            $schemaValueExcl = 'schemaExcl' + $lvl;
            out += ' var ' + ($exclusive) + '; var ' + ($exclType) + ' = typeof ' + ($schemaValueExcl) + '; if (' + ($exclType) + ' != \'boolean\' && ' + ($exclType) + ' != \'undefined\' && ' + ($exclType) + ' != \'number\') { ';
            var $errorKeyword = $exclusiveKeyword;
            var $$outStack = $$outStack || [];
            $$outStack.push(out);
            out = ''; /* istanbul ignore else */
            if (it.createErrors !== false) {
                out += ' { keyword: \'' + ($errorKeyword || '_exclusiveLimit') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: {} ';
                if (it.opts.messages !== false) {
                    out += ' , message: \'' + ($exclusiveKeyword) + ' should be boolean\' ';
                }
                if (it.opts.verbose) {
                    out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                }
                out += ' } ';
            }
            else {
                out += ' {} ';
            }
            var __err = out;
            out = $$outStack.pop();
            if (!it.compositeRule && $breakOnError) {
                /* istanbul ignore if */
                if (it.async) {
                    out += ' throw new ValidationError([' + (__err) + ']); ';
                }
                else {
                    out += ' validate.errors = [' + (__err) + ']; return false; ';
                }
            }
            else {
                out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
            }
            out += ' } else if ( ';
            if ($isData) {
                out += ' (' + ($schemaValue) + ' !== undefined && typeof ' + ($schemaValue) + ' != \'number\') || ';
            }
            out += ' ' + ($exclType) + ' == \'number\' ? ( (' + ($exclusive) + ' = ' + ($schemaValue) + ' === undefined || ' + ($schemaValueExcl) + ' ' + ($op) + '= ' + ($schemaValue) + ') ? ' + ($data) + ' ' + ($notOp) + '= ' + ($schemaValueExcl) + ' : ' + ($data) + ' ' + ($notOp) + ' ' + ($schemaValue) + ' ) : ( (' + ($exclusive) + ' = ' + ($schemaValueExcl) + ' === true) ? ' + ($data) + ' ' + ($notOp) + '= ' + ($schemaValue) + ' : ' + ($data) + ' ' + ($notOp) + ' ' + ($schemaValue) + ' ) || ' + ($data) + ' !== ' + ($data) + ') { var op' + ($lvl) + ' = ' + ($exclusive) + ' ? \'' + ($op) + '\' : \'' + ($op) + '=\'; ';
            if ($schema === undefined) {
                $errorKeyword = $exclusiveKeyword;
                $errSchemaPath = it.errSchemaPath + '/' + $exclusiveKeyword;
                $schemaValue = $schemaValueExcl;
                $isData = $isDataExcl;
            }
        }
        else {
            var $exclIsNumber = typeof $schemaExcl == 'number', $opStr = $op;
            if ($exclIsNumber && $isData) {
                var $opExpr = '\'' + $opStr + '\'';
                out += ' if ( ';
                if ($isData) {
                    out += ' (' + ($schemaValue) + ' !== undefined && typeof ' + ($schemaValue) + ' != \'number\') || ';
                }
                out += ' ( ' + ($schemaValue) + ' === undefined || ' + ($schemaExcl) + ' ' + ($op) + '= ' + ($schemaValue) + ' ? ' + ($data) + ' ' + ($notOp) + '= ' + ($schemaExcl) + ' : ' + ($data) + ' ' + ($notOp) + ' ' + ($schemaValue) + ' ) || ' + ($data) + ' !== ' + ($data) + ') { ';
            }
            else {
                if ($exclIsNumber && $schema === undefined) {
                    $exclusive = true;
                    $errorKeyword = $exclusiveKeyword;
                    $errSchemaPath = it.errSchemaPath + '/' + $exclusiveKeyword;
                    $schemaValue = $schemaExcl;
                    $notOp += '=';
                }
                else {
                    if ($exclIsNumber)
                        $schemaValue = Math[$isMax ? 'min' : 'max']($schemaExcl, $schema);
                    if ($schemaExcl === ($exclIsNumber ? $schemaValue : true)) {
                        $exclusive = true;
                        $errorKeyword = $exclusiveKeyword;
                        $errSchemaPath = it.errSchemaPath + '/' + $exclusiveKeyword;
                        $notOp += '=';
                    }
                    else {
                        $exclusive = false;
                        $opStr += '=';
                    }
                }
                var $opExpr = '\'' + $opStr + '\'';
                out += ' if ( ';
                if ($isData) {
                    out += ' (' + ($schemaValue) + ' !== undefined && typeof ' + ($schemaValue) + ' != \'number\') || ';
                }
                out += ' ' + ($data) + ' ' + ($notOp) + ' ' + ($schemaValue) + ' || ' + ($data) + ' !== ' + ($data) + ') { ';
            }
        }
        $errorKeyword = $errorKeyword || $keyword;
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ($errorKeyword || '_limit') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { comparison: ' + ($opExpr) + ', limit: ' + ($schemaValue) + ', exclusive: ' + ($exclusive) + ' } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should be ' + ($opStr) + ' ';
                if ($isData) {
                    out += '\' + ' + ($schemaValue);
                }
                else {
                    out += '' + ($schemaValue) + '\'';
                }
            }
            if (it.opts.verbose) {
                out += ' , schema:  ';
                if ($isData) {
                    out += 'validate.schema' + ($schemaPath);
                }
                else {
                    out += '' + ($schema);
                }
                out += '         , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += ' } ';
        if ($breakOnError) {
            out += ' else { ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/_limitItems.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate__limitItems(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $errorKeyword;
        var $data = 'data' + ($dataLvl || '');
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        if (!($isData || typeof $schema == 'number')) {
            throw new Error($keyword + ' must be number');
        }
        var $op = $keyword == 'maxItems' ? '>' : '<';
        out += 'if ( ';
        if ($isData) {
            out += ' (' + ($schemaValue) + ' !== undefined && typeof ' + ($schemaValue) + ' != \'number\') || ';
        }
        out += ' ' + ($data) + '.length ' + ($op) + ' ' + ($schemaValue) + ') { ';
        var $errorKeyword = $keyword;
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ($errorKeyword || '_limitItems') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { limit: ' + ($schemaValue) + ' } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should NOT have ';
                if ($keyword == 'maxItems') {
                    out += 'more';
                }
                else {
                    out += 'fewer';
                }
                out += ' than ';
                if ($isData) {
                    out += '\' + ' + ($schemaValue) + ' + \'';
                }
                else {
                    out += '' + ($schema);
                }
                out += ' items\' ';
            }
            if (it.opts.verbose) {
                out += ' , schema:  ';
                if ($isData) {
                    out += 'validate.schema' + ($schemaPath);
                }
                else {
                    out += '' + ($schema);
                }
                out += '         , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += '} ';
        if ($breakOnError) {
            out += ' else { ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/_limitLength.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate__limitLength(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $errorKeyword;
        var $data = 'data' + ($dataLvl || '');
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        if (!($isData || typeof $schema == 'number')) {
            throw new Error($keyword + ' must be number');
        }
        var $op = $keyword == 'maxLength' ? '>' : '<';
        out += 'if ( ';
        if ($isData) {
            out += ' (' + ($schemaValue) + ' !== undefined && typeof ' + ($schemaValue) + ' != \'number\') || ';
        }
        if (it.opts.unicode === false) {
            out += ' ' + ($data) + '.length ';
        }
        else {
            out += ' ucs2length(' + ($data) + ') ';
        }
        out += ' ' + ($op) + ' ' + ($schemaValue) + ') { ';
        var $errorKeyword = $keyword;
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ($errorKeyword || '_limitLength') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { limit: ' + ($schemaValue) + ' } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should NOT be ';
                if ($keyword == 'maxLength') {
                    out += 'longer';
                }
                else {
                    out += 'shorter';
                }
                out += ' than ';
                if ($isData) {
                    out += '\' + ' + ($schemaValue) + ' + \'';
                }
                else {
                    out += '' + ($schema);
                }
                out += ' characters\' ';
            }
            if (it.opts.verbose) {
                out += ' , schema:  ';
                if ($isData) {
                    out += 'validate.schema' + ($schemaPath);
                }
                else {
                    out += '' + ($schema);
                }
                out += '         , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += '} ';
        if ($breakOnError) {
            out += ' else { ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/_limitProperties.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate__limitProperties(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $errorKeyword;
        var $data = 'data' + ($dataLvl || '');
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        if (!($isData || typeof $schema == 'number')) {
            throw new Error($keyword + ' must be number');
        }
        var $op = $keyword == 'maxProperties' ? '>' : '<';
        out += 'if ( ';
        if ($isData) {
            out += ' (' + ($schemaValue) + ' !== undefined && typeof ' + ($schemaValue) + ' != \'number\') || ';
        }
        out += ' Object.keys(' + ($data) + ').length ' + ($op) + ' ' + ($schemaValue) + ') { ';
        var $errorKeyword = $keyword;
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ($errorKeyword || '_limitProperties') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { limit: ' + ($schemaValue) + ' } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should NOT have ';
                if ($keyword == 'maxProperties') {
                    out += 'more';
                }
                else {
                    out += 'fewer';
                }
                out += ' than ';
                if ($isData) {
                    out += '\' + ' + ($schemaValue) + ' + \'';
                }
                else {
                    out += '' + ($schema);
                }
                out += ' properties\' ';
            }
            if (it.opts.verbose) {
                out += ' , schema:  ';
                if ($isData) {
                    out += 'validate.schema' + ($schemaPath);
                }
                else {
                    out += '' + ($schema);
                }
                out += '         , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += '} ';
        if ($breakOnError) {
            out += ' else { ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/multipleOf.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_multipleOf(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        if (!($isData || typeof $schema == 'number')) {
            throw new Error($keyword + ' must be number');
        }
        out += 'var division' + ($lvl) + ';if (';
        if ($isData) {
            out += ' ' + ($schemaValue) + ' !== undefined && ( typeof ' + ($schemaValue) + ' != \'number\' || ';
        }
        out += ' (division' + ($lvl) + ' = ' + ($data) + ' / ' + ($schemaValue) + ', ';
        if (it.opts.multipleOfPrecision) {
            out += ' Math.abs(Math.round(division' + ($lvl) + ') - division' + ($lvl) + ') > 1e-' + (it.opts.multipleOfPrecision) + ' ';
        }
        else {
            out += ' division' + ($lvl) + ' !== parseInt(division' + ($lvl) + ') ';
        }
        out += ' ) ';
        if ($isData) {
            out += '  )  ';
        }
        out += ' ) {   ';
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ('multipleOf') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { multipleOf: ' + ($schemaValue) + ' } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should be multiple of ';
                if ($isData) {
                    out += '\' + ' + ($schemaValue);
                }
                else {
                    out += '' + ($schemaValue) + '\'';
                }
            }
            if (it.opts.verbose) {
                out += ' , schema:  ';
                if ($isData) {
                    out += 'validate.schema' + ($schemaPath);
                }
                else {
                    out += '' + ($schema);
                }
                out += '         , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += '} ';
        if ($breakOnError) {
            out += ' else { ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/not.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_not(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $errs = 'errs__' + $lvl;
        var $it = it.util.copy(it);
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        if ((it.opts.strictKeywords ? (typeof $schema == 'object' && Object.keys($schema).length > 0) || $schema === false : it.util.schemaHasRules($schema, it.RULES.all))) {
            $it.schema = $schema;
            $it.schemaPath = $schemaPath;
            $it.errSchemaPath = $errSchemaPath;
            out += ' var ' + ($errs) + ' = errors;  ';
            var $wasComposite = it.compositeRule;
            it.compositeRule = $it.compositeRule = true;
            $it.createErrors = false;
            var $allErrorsOption;
            if ($it.opts.allErrors) {
                $allErrorsOption = $it.opts.allErrors;
                $it.opts.allErrors = false;
            }
            out += ' ' + (it.validate($it)) + ' ';
            $it.createErrors = true;
            if ($allErrorsOption)
                $it.opts.allErrors = $allErrorsOption;
            it.compositeRule = $it.compositeRule = $wasComposite;
            out += ' if (' + ($nextValid) + ') {   ';
            var $$outStack = $$outStack || [];
            $$outStack.push(out);
            out = ''; /* istanbul ignore else */
            if (it.createErrors !== false) {
                out += ' { keyword: \'' + ('not') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: {} ';
                if (it.opts.messages !== false) {
                    out += ' , message: \'should NOT be valid\' ';
                }
                if (it.opts.verbose) {
                    out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                }
                out += ' } ';
            }
            else {
                out += ' {} ';
            }
            var __err = out;
            out = $$outStack.pop();
            if (!it.compositeRule && $breakOnError) {
                /* istanbul ignore if */
                if (it.async) {
                    out += ' throw new ValidationError([' + (__err) + ']); ';
                }
                else {
                    out += ' validate.errors = [' + (__err) + ']; return false; ';
                }
            }
            else {
                out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
            }
            out += ' } else {  errors = ' + ($errs) + '; if (vErrors !== null) { if (' + ($errs) + ') vErrors.length = ' + ($errs) + '; else vErrors = null; } ';
            if (it.opts.allErrors) {
                out += ' } ';
            }
        }
        else {
            out += '  var err =   '; /* istanbul ignore else */
            if (it.createErrors !== false) {
                out += ' { keyword: \'' + ('not') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: {} ';
                if (it.opts.messages !== false) {
                    out += ' , message: \'should NOT be valid\' ';
                }
                if (it.opts.verbose) {
                    out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                }
                out += ' } ';
            }
            else {
                out += ' {} ';
            }
            out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
            if ($breakOnError) {
                out += ' if (false) { ';
            }
        }
        return out;
    };
},
/* ajv/lib/dotjs/oneOf.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_oneOf(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $errs = 'errs__' + $lvl;
        var $it = it.util.copy(it);
        var $closingBraces = '';
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        var $currentBaseId = $it.baseId, $prevValid = 'prevValid' + $lvl, $passingSchemas = 'passingSchemas' + $lvl;
        out += 'var ' + ($errs) + ' = errors , ' + ($prevValid) + ' = false , ' + ($valid) + ' = false , ' + ($passingSchemas) + ' = null; ';
        var $wasComposite = it.compositeRule;
        it.compositeRule = $it.compositeRule = true;
        var arr1 = $schema;
        if (arr1) {
            var $sch, $i = -1, l1 = arr1.length - 1;
            while ($i < l1) {
                $sch = arr1[$i += 1];
                if ((it.opts.strictKeywords ? (typeof $sch == 'object' && Object.keys($sch).length > 0) || $sch === false : it.util.schemaHasRules($sch, it.RULES.all))) {
                    $it.schema = $sch;
                    $it.schemaPath = $schemaPath + '[' + $i + ']';
                    $it.errSchemaPath = $errSchemaPath + '/' + $i;
                    out += '  ' + (it.validate($it)) + ' ';
                    $it.baseId = $currentBaseId;
                }
                else {
                    out += ' var ' + ($nextValid) + ' = true; ';
                }
                if ($i) {
                    out += ' if (' + ($nextValid) + ' && ' + ($prevValid) + ') { ' + ($valid) + ' = false; ' + ($passingSchemas) + ' = [' + ($passingSchemas) + ', ' + ($i) + ']; } else { ';
                    $closingBraces += '}';
                }
                out += ' if (' + ($nextValid) + ') { ' + ($valid) + ' = ' + ($prevValid) + ' = true; ' + ($passingSchemas) + ' = ' + ($i) + '; }';
            }
        }
        it.compositeRule = $it.compositeRule = $wasComposite;
        out += '' + ($closingBraces) + 'if (!' + ($valid) + ') {   var err =   '; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ('oneOf') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { passingSchemas: ' + ($passingSchemas) + ' } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should match exactly one schema in oneOf\' ';
            }
            if (it.opts.verbose) {
                out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError(vErrors); ';
            }
            else {
                out += ' validate.errors = vErrors; return false; ';
            }
        }
        out += '} else {  errors = ' + ($errs) + '; if (vErrors !== null) { if (' + ($errs) + ') vErrors.length = ' + ($errs) + '; else vErrors = null; }';
        if (it.opts.allErrors) {
            out += ' } ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/pattern.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_pattern(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        var $regexp = $isData ? '(new RegExp(' + $schemaValue + '))' : it.usePattern($schema);
        out += 'if ( ';
        if ($isData) {
            out += ' (' + ($schemaValue) + ' !== undefined && typeof ' + ($schemaValue) + ' != \'string\') || ';
        }
        out += ' !' + ($regexp) + '.test(' + ($data) + ') ) {   ';
        var $$outStack = $$outStack || [];
        $$outStack.push(out);
        out = ''; /* istanbul ignore else */
        if (it.createErrors !== false) {
            out += ' { keyword: \'' + ('pattern') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { pattern:  ';
            if ($isData) {
                out += '' + ($schemaValue);
            }
            else {
                out += '' + (it.util.toQuotedString($schema));
            }
            out += '  } ';
            if (it.opts.messages !== false) {
                out += ' , message: \'should match pattern "';
                if ($isData) {
                    out += '\' + ' + ($schemaValue) + ' + \'';
                }
                else {
                    out += '' + (it.util.escapeQuotes($schema));
                }
                out += '"\' ';
            }
            if (it.opts.verbose) {
                out += ' , schema:  ';
                if ($isData) {
                    out += 'validate.schema' + ($schemaPath);
                }
                else {
                    out += '' + (it.util.toQuotedString($schema));
                }
                out += '         , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
            }
            out += ' } ';
        }
        else {
            out += ' {} ';
        }
        var __err = out;
        out = $$outStack.pop();
        if (!it.compositeRule && $breakOnError) {
            /* istanbul ignore if */
            if (it.async) {
                out += ' throw new ValidationError([' + (__err) + ']); ';
            }
            else {
                out += ' validate.errors = [' + (__err) + ']; return false; ';
            }
        }
        else {
            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
        }
        out += '} ';
        if ($breakOnError) {
            out += ' else { ';
        }
        return out;
    };
},
/* ajv/lib/dotjs/properties.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_properties(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $errs = 'errs__' + $lvl;
        var $it = it.util.copy(it);
        var $closingBraces = '';
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        var $key = 'key' + $lvl, $idx = 'idx' + $lvl, $dataNxt = $it.dataLevel = it.dataLevel + 1, $nextData = 'data' + $dataNxt, $dataProperties = 'dataProperties' + $lvl;
        var $schemaKeys = Object.keys($schema || {}).filter(notProto), $pProperties = it.schema.patternProperties || {}, $pPropertyKeys = Object.keys($pProperties).filter(notProto), $aProperties = it.schema.additionalProperties, $someProperties = $schemaKeys.length || $pPropertyKeys.length, $noAdditional = $aProperties === false, $additionalIsSchema = typeof $aProperties == 'object' && Object.keys($aProperties).length, $removeAdditional = it.opts.removeAdditional, $checkAdditional = $noAdditional || $additionalIsSchema || $removeAdditional, $ownProperties = it.opts.ownProperties, $currentBaseId = it.baseId;
        var $required = it.schema.required;
        if ($required && !(it.opts.$data && $required.$data) && $required.length < it.opts.loopRequired) {
            var $requiredHash = it.util.toHash($required);
        }
        function notProto(p) {
            return p !== '__proto__';
        }
        out += 'var ' + ($errs) + ' = errors;var ' + ($nextValid) + ' = true;';
        if ($ownProperties) {
            out += ' var ' + ($dataProperties) + ' = undefined;';
        }
        if ($checkAdditional) {
            if ($ownProperties) {
                out += ' ' + ($dataProperties) + ' = ' + ($dataProperties) + ' || Object.keys(' + ($data) + '); for (var ' + ($idx) + '=0; ' + ($idx) + '<' + ($dataProperties) + '.length; ' + ($idx) + '++) { var ' + ($key) + ' = ' + ($dataProperties) + '[' + ($idx) + ']; ';
            }
            else {
                out += ' for (var ' + ($key) + ' in ' + ($data) + ') { ';
            }
            if ($someProperties) {
                out += ' var isAdditional' + ($lvl) + ' = !(false ';
                if ($schemaKeys.length) {
                    if ($schemaKeys.length > 8) {
                        out += ' || validate.schema' + ($schemaPath) + '.hasOwnProperty(' + ($key) + ') ';
                    }
                    else {
                        var arr1 = $schemaKeys;
                        if (arr1) {
                            var $propertyKey, i1 = -1, l1 = arr1.length - 1;
                            while (i1 < l1) {
                                $propertyKey = arr1[i1 += 1];
                                out += ' || ' + ($key) + ' == ' + (it.util.toQuotedString($propertyKey)) + ' ';
                            }
                        }
                    }
                }
                if ($pPropertyKeys.length) {
                    var arr2 = $pPropertyKeys;
                    if (arr2) {
                        var $pProperty, $i = -1, l2 = arr2.length - 1;
                        while ($i < l2) {
                            $pProperty = arr2[$i += 1];
                            out += ' || ' + (it.usePattern($pProperty)) + '.test(' + ($key) + ') ';
                        }
                    }
                }
                out += ' ); if (isAdditional' + ($lvl) + ') { ';
            }
            if ($removeAdditional == 'all') {
                out += ' delete ' + ($data) + '[' + ($key) + ']; ';
            }
            else {
                var $currentErrorPath = it.errorPath;
                var $additionalProperty = '\' + ' + $key + ' + \'';
                if (it.opts._errorDataPathProperty) {
                    it.errorPath = it.util.getPathExpr(it.errorPath, $key, it.opts.jsonPointers);
                }
                if ($noAdditional) {
                    if ($removeAdditional) {
                        out += ' delete ' + ($data) + '[' + ($key) + ']; ';
                    }
                    else {
                        out += ' ' + ($nextValid) + ' = false; ';
                        var $currErrSchemaPath = $errSchemaPath;
                        $errSchemaPath = it.errSchemaPath + '/additionalProperties';
                        var $$outStack = $$outStack || [];
                        $$outStack.push(out);
                        out = ''; /* istanbul ignore else */
                        if (it.createErrors !== false) {
                            out += ' { keyword: \'' + ('additionalProperties') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { additionalProperty: \'' + ($additionalProperty) + '\' } ';
                            if (it.opts.messages !== false) {
                                out += ' , message: \'';
                                if (it.opts._errorDataPathProperty) {
                                    out += 'is an invalid additional property';
                                }
                                else {
                                    out += 'should NOT have additional properties';
                                }
                                out += '\' ';
                            }
                            if (it.opts.verbose) {
                                out += ' , schema: false , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                            }
                            out += ' } ';
                        }
                        else {
                            out += ' {} ';
                        }
                        var __err = out;
                        out = $$outStack.pop();
                        if (!it.compositeRule && $breakOnError) {
                            /* istanbul ignore if */
                            if (it.async) {
                                out += ' throw new ValidationError([' + (__err) + ']); ';
                            }
                            else {
                                out += ' validate.errors = [' + (__err) + ']; return false; ';
                            }
                        }
                        else {
                            out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                        }
                        $errSchemaPath = $currErrSchemaPath;
                        if ($breakOnError) {
                            out += ' break; ';
                        }
                    }
                }
                else if ($additionalIsSchema) {
                    if ($removeAdditional == 'failing') {
                        out += ' var ' + ($errs) + ' = errors;  ';
                        var $wasComposite = it.compositeRule;
                        it.compositeRule = $it.compositeRule = true;
                        $it.schema = $aProperties;
                        $it.schemaPath = it.schemaPath + '.additionalProperties';
                        $it.errSchemaPath = it.errSchemaPath + '/additionalProperties';
                        $it.errorPath = it.opts._errorDataPathProperty ? it.errorPath : it.util.getPathExpr(it.errorPath, $key, it.opts.jsonPointers);
                        var $passData = $data + '[' + $key + ']';
                        $it.dataPathArr[$dataNxt] = $key;
                        var $code = it.validate($it);
                        $it.baseId = $currentBaseId;
                        if (it.util.varOccurences($code, $nextData) < 2) {
                            out += ' ' + (it.util.varReplace($code, $nextData, $passData)) + ' ';
                        }
                        else {
                            out += ' var ' + ($nextData) + ' = ' + ($passData) + '; ' + ($code) + ' ';
                        }
                        out += ' if (!' + ($nextValid) + ') { errors = ' + ($errs) + '; if (validate.errors !== null) { if (errors) validate.errors.length = errors; else validate.errors = null; } delete ' + ($data) + '[' + ($key) + ']; }  ';
                        it.compositeRule = $it.compositeRule = $wasComposite;
                    }
                    else {
                        $it.schema = $aProperties;
                        $it.schemaPath = it.schemaPath + '.additionalProperties';
                        $it.errSchemaPath = it.errSchemaPath + '/additionalProperties';
                        $it.errorPath = it.opts._errorDataPathProperty ? it.errorPath : it.util.getPathExpr(it.errorPath, $key, it.opts.jsonPointers);
                        var $passData = $data + '[' + $key + ']';
                        $it.dataPathArr[$dataNxt] = $key;
                        var $code = it.validate($it);
                        $it.baseId = $currentBaseId;
                        if (it.util.varOccurences($code, $nextData) < 2) {
                            out += ' ' + (it.util.varReplace($code, $nextData, $passData)) + ' ';
                        }
                        else {
                            out += ' var ' + ($nextData) + ' = ' + ($passData) + '; ' + ($code) + ' ';
                        }
                        if ($breakOnError) {
                            out += ' if (!' + ($nextValid) + ') break; ';
                        }
                    }
                }
                it.errorPath = $currentErrorPath;
            }
            if ($someProperties) {
                out += ' } ';
            }
            out += ' }  ';
            if ($breakOnError) {
                out += ' if (' + ($nextValid) + ') { ';
                $closingBraces += '}';
            }
        }
        var $useDefaults = it.opts.useDefaults && !it.compositeRule;
        if ($schemaKeys.length) {
            var arr3 = $schemaKeys;
            if (arr3) {
                var $propertyKey, i3 = -1, l3 = arr3.length - 1;
                while (i3 < l3) {
                    $propertyKey = arr3[i3 += 1];
                    var $sch = $schema[$propertyKey];
                    if ((it.opts.strictKeywords ? (typeof $sch == 'object' && Object.keys($sch).length > 0) || $sch === false : it.util.schemaHasRules($sch, it.RULES.all))) {
                        var $prop = it.util.getProperty($propertyKey), $passData = $data + $prop, $hasDefault = $useDefaults && $sch.default !== undefined;
                        $it.schema = $sch;
                        $it.schemaPath = $schemaPath + $prop;
                        $it.errSchemaPath = $errSchemaPath + '/' + it.util.escapeFragment($propertyKey);
                        $it.errorPath = it.util.getPath(it.errorPath, $propertyKey, it.opts.jsonPointers);
                        $it.dataPathArr[$dataNxt] = it.util.toQuotedString($propertyKey);
                        var $code = it.validate($it);
                        $it.baseId = $currentBaseId;
                        if (it.util.varOccurences($code, $nextData) < 2) {
                            $code = it.util.varReplace($code, $nextData, $passData);
                            var $useData = $passData;
                        }
                        else {
                            var $useData = $nextData;
                            out += ' var ' + ($nextData) + ' = ' + ($passData) + '; ';
                        }
                        if ($hasDefault) {
                            out += ' ' + ($code) + ' ';
                        }
                        else {
                            if ($requiredHash && $requiredHash[$propertyKey]) {
                                out += ' if ( ' + ($useData) + ' === undefined ';
                                if ($ownProperties) {
                                    out += ' || ! Object.prototype.hasOwnProperty.call(' + ($data) + ', \'' + (it.util.escapeQuotes($propertyKey)) + '\') ';
                                }
                                out += ') { ' + ($nextValid) + ' = false; ';
                                var $currentErrorPath = it.errorPath, $currErrSchemaPath = $errSchemaPath, $missingProperty = it.util.escapeQuotes($propertyKey);
                                if (it.opts._errorDataPathProperty) {
                                    it.errorPath = it.util.getPath($currentErrorPath, $propertyKey, it.opts.jsonPointers);
                                }
                                $errSchemaPath = it.errSchemaPath + '/required';
                                var $$outStack = $$outStack || [];
                                $$outStack.push(out);
                                out = ''; /* istanbul ignore else */
                                if (it.createErrors !== false) {
                                    out += ' { keyword: \'' + ('required') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { missingProperty: \'' + ($missingProperty) + '\' } ';
                                    if (it.opts.messages !== false) {
                                        out += ' , message: \'';
                                        if (it.opts._errorDataPathProperty) {
                                            out += 'is a required property';
                                        }
                                        else {
                                            out += 'should have required property \\\'' + ($missingProperty) + '\\\'';
                                        }
                                        out += '\' ';
                                    }
                                    if (it.opts.verbose) {
                                        out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                                    }
                                    out += ' } ';
                                }
                                else {
                                    out += ' {} ';
                                }
                                var __err = out;
                                out = $$outStack.pop();
                                if (!it.compositeRule && $breakOnError) {
                                    /* istanbul ignore if */
                                    if (it.async) {
                                        out += ' throw new ValidationError([' + (__err) + ']); ';
                                    }
                                    else {
                                        out += ' validate.errors = [' + (__err) + ']; return false; ';
                                    }
                                }
                                else {
                                    out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                                }
                                $errSchemaPath = $currErrSchemaPath;
                                it.errorPath = $currentErrorPath;
                                out += ' } else { ';
                            }
                            else {
                                if ($breakOnError) {
                                    out += ' if ( ' + ($useData) + ' === undefined ';
                                    if ($ownProperties) {
                                        out += ' || ! Object.prototype.hasOwnProperty.call(' + ($data) + ', \'' + (it.util.escapeQuotes($propertyKey)) + '\') ';
                                    }
                                    out += ') { ' + ($nextValid) + ' = true; } else { ';
                                }
                                else {
                                    out += ' if (' + ($useData) + ' !== undefined ';
                                    if ($ownProperties) {
                                        out += ' &&   Object.prototype.hasOwnProperty.call(' + ($data) + ', \'' + (it.util.escapeQuotes($propertyKey)) + '\') ';
                                    }
                                    out += ' ) { ';
                                }
                            }
                            out += ' ' + ($code) + ' } ';
                        }
                    }
                    if ($breakOnError) {
                        out += ' if (' + ($nextValid) + ') { ';
                        $closingBraces += '}';
                    }
                }
            }
        }
        if ($pPropertyKeys.length) {
            var arr4 = $pPropertyKeys;
            if (arr4) {
                var $pProperty, i4 = -1, l4 = arr4.length - 1;
                while (i4 < l4) {
                    $pProperty = arr4[i4 += 1];
                    var $sch = $pProperties[$pProperty];
                    if ((it.opts.strictKeywords ? (typeof $sch == 'object' && Object.keys($sch).length > 0) || $sch === false : it.util.schemaHasRules($sch, it.RULES.all))) {
                        $it.schema = $sch;
                        $it.schemaPath = it.schemaPath + '.patternProperties' + it.util.getProperty($pProperty);
                        $it.errSchemaPath = it.errSchemaPath + '/patternProperties/' + it.util.escapeFragment($pProperty);
                        if ($ownProperties) {
                            out += ' ' + ($dataProperties) + ' = ' + ($dataProperties) + ' || Object.keys(' + ($data) + '); for (var ' + ($idx) + '=0; ' + ($idx) + '<' + ($dataProperties) + '.length; ' + ($idx) + '++) { var ' + ($key) + ' = ' + ($dataProperties) + '[' + ($idx) + ']; ';
                        }
                        else {
                            out += ' for (var ' + ($key) + ' in ' + ($data) + ') { ';
                        }
                        out += ' if (' + (it.usePattern($pProperty)) + '.test(' + ($key) + ')) { ';
                        $it.errorPath = it.util.getPathExpr(it.errorPath, $key, it.opts.jsonPointers);
                        var $passData = $data + '[' + $key + ']';
                        $it.dataPathArr[$dataNxt] = $key;
                        var $code = it.validate($it);
                        $it.baseId = $currentBaseId;
                        if (it.util.varOccurences($code, $nextData) < 2) {
                            out += ' ' + (it.util.varReplace($code, $nextData, $passData)) + ' ';
                        }
                        else {
                            out += ' var ' + ($nextData) + ' = ' + ($passData) + '; ' + ($code) + ' ';
                        }
                        if ($breakOnError) {
                            out += ' if (!' + ($nextValid) + ') break; ';
                        }
                        out += ' } ';
                        if ($breakOnError) {
                            out += ' else ' + ($nextValid) + ' = true; ';
                        }
                        out += ' }  ';
                        if ($breakOnError) {
                            out += ' if (' + ($nextValid) + ') { ';
                            $closingBraces += '}';
                        }
                    }
                }
            }
        }
        if ($breakOnError) {
            out += ' ' + ($closingBraces) + ' if (' + ($errs) + ' == errors) {';
        }
        return out;
    };
},
/* ajv/lib/dotjs/propertyNames.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_propertyNames(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $errs = 'errs__' + $lvl;
        var $it = it.util.copy(it);
        var $closingBraces = '';
        $it.level++;
        var $nextValid = 'valid' + $it.level;
        out += 'var ' + ($errs) + ' = errors;';
        if ((it.opts.strictKeywords ? (typeof $schema == 'object' && Object.keys($schema).length > 0) || $schema === false : it.util.schemaHasRules($schema, it.RULES.all))) {
            $it.schema = $schema;
            $it.schemaPath = $schemaPath;
            $it.errSchemaPath = $errSchemaPath;
            var $key = 'key' + $lvl, $idx = 'idx' + $lvl, $i = 'i' + $lvl, $invalidName = '\' + ' + $key + ' + \'', $dataNxt = $it.dataLevel = it.dataLevel + 1, $nextData = 'data' + $dataNxt, $dataProperties = 'dataProperties' + $lvl, $ownProperties = it.opts.ownProperties, $currentBaseId = it.baseId;
            if ($ownProperties) {
                out += ' var ' + ($dataProperties) + ' = undefined; ';
            }
            if ($ownProperties) {
                out += ' ' + ($dataProperties) + ' = ' + ($dataProperties) + ' || Object.keys(' + ($data) + '); for (var ' + ($idx) + '=0; ' + ($idx) + '<' + ($dataProperties) + '.length; ' + ($idx) + '++) { var ' + ($key) + ' = ' + ($dataProperties) + '[' + ($idx) + ']; ';
            }
            else {
                out += ' for (var ' + ($key) + ' in ' + ($data) + ') { ';
            }
            out += ' var startErrs' + ($lvl) + ' = errors; ';
            var $passData = $key;
            var $wasComposite = it.compositeRule;
            it.compositeRule = $it.compositeRule = true;
            var $code = it.validate($it);
            $it.baseId = $currentBaseId;
            if (it.util.varOccurences($code, $nextData) < 2) {
                out += ' ' + (it.util.varReplace($code, $nextData, $passData)) + ' ';
            }
            else {
                out += ' var ' + ($nextData) + ' = ' + ($passData) + '; ' + ($code) + ' ';
            }
            it.compositeRule = $it.compositeRule = $wasComposite;
            out += ' if (!' + ($nextValid) + ') { for (var ' + ($i) + '=startErrs' + ($lvl) + '; ' + ($i) + '<errors; ' + ($i) + '++) { vErrors[' + ($i) + '].propertyName = ' + ($key) + '; }   var err =   '; /* istanbul ignore else */
            if (it.createErrors !== false) {
                out += ' { keyword: \'' + ('propertyNames') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { propertyName: \'' + ($invalidName) + '\' } ';
                if (it.opts.messages !== false) {
                    out += ' , message: \'property name \\\'' + ($invalidName) + '\\\' is invalid\' ';
                }
                if (it.opts.verbose) {
                    out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                }
                out += ' } ';
            }
            else {
                out += ' {} ';
            }
            out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
            if (!it.compositeRule && $breakOnError) {
                /* istanbul ignore if */
                if (it.async) {
                    out += ' throw new ValidationError(vErrors); ';
                }
                else {
                    out += ' validate.errors = vErrors; return false; ';
                }
            }
            if ($breakOnError) {
                out += ' break; ';
            }
            out += ' } }';
        }
        if ($breakOnError) {
            out += ' ' + ($closingBraces) + ' if (' + ($errs) + ' == errors) {';
        }
        return out;
    };
},
/* ajv/lib/dotjs/required.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_required(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        var $vSchema = 'schema' + $lvl;
        if (!$isData) {
            if ($schema.length < it.opts.loopRequired && it.schema.properties && Object.keys(it.schema.properties).length) {
                var $required = [];
                var arr1 = $schema;
                if (arr1) {
                    var $property, i1 = -1, l1 = arr1.length - 1;
                    while (i1 < l1) {
                        $property = arr1[i1 += 1];
                        var $propertySch = it.schema.properties[$property];
                        if (!($propertySch && (it.opts.strictKeywords ? (typeof $propertySch == 'object' && Object.keys($propertySch).length > 0) || $propertySch === false : it.util.schemaHasRules($propertySch, it.RULES.all)))) {
                            $required[$required.length] = $property;
                        }
                    }
                }
            }
            else {
                var $required = $schema;
            }
        }
        if ($isData || $required.length) {
            var $currentErrorPath = it.errorPath, $loopRequired = $isData || $required.length >= it.opts.loopRequired, $ownProperties = it.opts.ownProperties;
            if ($breakOnError) {
                out += ' var missing' + ($lvl) + '; ';
                if ($loopRequired) {
                    if (!$isData) {
                        out += ' var ' + ($vSchema) + ' = validate.schema' + ($schemaPath) + '; ';
                    }
                    var $i = 'i' + $lvl, $propertyPath = 'schema' + $lvl + '[' + $i + ']', $missingProperty = '\' + ' + $propertyPath + ' + \'';
                    if (it.opts._errorDataPathProperty) {
                        it.errorPath = it.util.getPathExpr($currentErrorPath, $propertyPath, it.opts.jsonPointers);
                    }
                    out += ' var ' + ($valid) + ' = true; ';
                    if ($isData) {
                        out += ' if (schema' + ($lvl) + ' === undefined) ' + ($valid) + ' = true; else if (!Array.isArray(schema' + ($lvl) + ')) ' + ($valid) + ' = false; else {';
                    }
                    out += ' for (var ' + ($i) + ' = 0; ' + ($i) + ' < ' + ($vSchema) + '.length; ' + ($i) + '++) { ' + ($valid) + ' = ' + ($data) + '[' + ($vSchema) + '[' + ($i) + ']] !== undefined ';
                    if ($ownProperties) {
                        out += ' &&   Object.prototype.hasOwnProperty.call(' + ($data) + ', ' + ($vSchema) + '[' + ($i) + ']) ';
                    }
                    out += '; if (!' + ($valid) + ') break; } ';
                    if ($isData) {
                        out += '  }  ';
                    }
                    out += '  if (!' + ($valid) + ') {   ';
                    var $$outStack = $$outStack || [];
                    $$outStack.push(out);
                    out = ''; /* istanbul ignore else */
                    if (it.createErrors !== false) {
                        out += ' { keyword: \'' + ('required') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { missingProperty: \'' + ($missingProperty) + '\' } ';
                        if (it.opts.messages !== false) {
                            out += ' , message: \'';
                            if (it.opts._errorDataPathProperty) {
                                out += 'is a required property';
                            }
                            else {
                                out += 'should have required property \\\'' + ($missingProperty) + '\\\'';
                            }
                            out += '\' ';
                        }
                        if (it.opts.verbose) {
                            out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                        }
                        out += ' } ';
                    }
                    else {
                        out += ' {} ';
                    }
                    var __err = out;
                    out = $$outStack.pop();
                    if (!it.compositeRule && $breakOnError) {
                        /* istanbul ignore if */
                        if (it.async) {
                            out += ' throw new ValidationError([' + (__err) + ']); ';
                        }
                        else {
                            out += ' validate.errors = [' + (__err) + ']; return false; ';
                        }
                    }
                    else {
                        out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                    }
                    out += ' } else { ';
                }
                else {
                    out += ' if ( ';
                    var arr2 = $required;
                    if (arr2) {
                        var $propertyKey, $i = -1, l2 = arr2.length - 1;
                        while ($i < l2) {
                            $propertyKey = arr2[$i += 1];
                            if ($i) {
                                out += ' || ';
                            }
                            var $prop = it.util.getProperty($propertyKey), $useData = $data + $prop;
                            out += ' ( ( ' + ($useData) + ' === undefined ';
                            if ($ownProperties) {
                                out += ' || ! Object.prototype.hasOwnProperty.call(' + ($data) + ', \'' + (it.util.escapeQuotes($propertyKey)) + '\') ';
                            }
                            out += ') && (missing' + ($lvl) + ' = ' + (it.util.toQuotedString(it.opts.jsonPointers ? $propertyKey : $prop)) + ') ) ';
                        }
                    }
                    out += ') {  ';
                    var $propertyPath = 'missing' + $lvl, $missingProperty = '\' + ' + $propertyPath + ' + \'';
                    if (it.opts._errorDataPathProperty) {
                        it.errorPath = it.opts.jsonPointers ? it.util.getPathExpr($currentErrorPath, $propertyPath, true) : $currentErrorPath + ' + ' + $propertyPath;
                    }
                    var $$outStack = $$outStack || [];
                    $$outStack.push(out);
                    out = ''; /* istanbul ignore else */
                    if (it.createErrors !== false) {
                        out += ' { keyword: \'' + ('required') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { missingProperty: \'' + ($missingProperty) + '\' } ';
                        if (it.opts.messages !== false) {
                            out += ' , message: \'';
                            if (it.opts._errorDataPathProperty) {
                                out += 'is a required property';
                            }
                            else {
                                out += 'should have required property \\\'' + ($missingProperty) + '\\\'';
                            }
                            out += '\' ';
                        }
                        if (it.opts.verbose) {
                            out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                        }
                        out += ' } ';
                    }
                    else {
                        out += ' {} ';
                    }
                    var __err = out;
                    out = $$outStack.pop();
                    if (!it.compositeRule && $breakOnError) {
                        /* istanbul ignore if */
                        if (it.async) {
                            out += ' throw new ValidationError([' + (__err) + ']); ';
                        }
                        else {
                            out += ' validate.errors = [' + (__err) + ']; return false; ';
                        }
                    }
                    else {
                        out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                    }
                    out += ' } else { ';
                }
            }
            else {
                if ($loopRequired) {
                    if (!$isData) {
                        out += ' var ' + ($vSchema) + ' = validate.schema' + ($schemaPath) + '; ';
                    }
                    var $i = 'i' + $lvl, $propertyPath = 'schema' + $lvl + '[' + $i + ']', $missingProperty = '\' + ' + $propertyPath + ' + \'';
                    if (it.opts._errorDataPathProperty) {
                        it.errorPath = it.util.getPathExpr($currentErrorPath, $propertyPath, it.opts.jsonPointers);
                    }
                    if ($isData) {
                        out += ' if (' + ($vSchema) + ' && !Array.isArray(' + ($vSchema) + ')) {  var err =   '; /* istanbul ignore else */
                        if (it.createErrors !== false) {
                            out += ' { keyword: \'' + ('required') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { missingProperty: \'' + ($missingProperty) + '\' } ';
                            if (it.opts.messages !== false) {
                                out += ' , message: \'';
                                if (it.opts._errorDataPathProperty) {
                                    out += 'is a required property';
                                }
                                else {
                                    out += 'should have required property \\\'' + ($missingProperty) + '\\\'';
                                }
                                out += '\' ';
                            }
                            if (it.opts.verbose) {
                                out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                            }
                            out += ' } ';
                        }
                        else {
                            out += ' {} ';
                        }
                        out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; } else if (' + ($vSchema) + ' !== undefined) { ';
                    }
                    out += ' for (var ' + ($i) + ' = 0; ' + ($i) + ' < ' + ($vSchema) + '.length; ' + ($i) + '++) { if (' + ($data) + '[' + ($vSchema) + '[' + ($i) + ']] === undefined ';
                    if ($ownProperties) {
                        out += ' || ! Object.prototype.hasOwnProperty.call(' + ($data) + ', ' + ($vSchema) + '[' + ($i) + ']) ';
                    }
                    out += ') {  var err =   '; /* istanbul ignore else */
                    if (it.createErrors !== false) {
                        out += ' { keyword: \'' + ('required') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { missingProperty: \'' + ($missingProperty) + '\' } ';
                        if (it.opts.messages !== false) {
                            out += ' , message: \'';
                            if (it.opts._errorDataPathProperty) {
                                out += 'is a required property';
                            }
                            else {
                                out += 'should have required property \\\'' + ($missingProperty) + '\\\'';
                            }
                            out += '\' ';
                        }
                        if (it.opts.verbose) {
                            out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                        }
                        out += ' } ';
                    }
                    else {
                        out += ' {} ';
                    }
                    out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; } } ';
                    if ($isData) {
                        out += '  }  ';
                    }
                }
                else {
                    var arr3 = $required;
                    if (arr3) {
                        var $propertyKey, i3 = -1, l3 = arr3.length - 1;
                        while (i3 < l3) {
                            $propertyKey = arr3[i3 += 1];
                            var $prop = it.util.getProperty($propertyKey), $missingProperty = it.util.escapeQuotes($propertyKey), $useData = $data + $prop;
                            if (it.opts._errorDataPathProperty) {
                                it.errorPath = it.util.getPath($currentErrorPath, $propertyKey, it.opts.jsonPointers);
                            }
                            out += ' if ( ' + ($useData) + ' === undefined ';
                            if ($ownProperties) {
                                out += ' || ! Object.prototype.hasOwnProperty.call(' + ($data) + ', \'' + (it.util.escapeQuotes($propertyKey)) + '\') ';
                            }
                            out += ') {  var err =   '; /* istanbul ignore else */
                            if (it.createErrors !== false) {
                                out += ' { keyword: \'' + ('required') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { missingProperty: \'' + ($missingProperty) + '\' } ';
                                if (it.opts.messages !== false) {
                                    out += ' , message: \'';
                                    if (it.opts._errorDataPathProperty) {
                                        out += 'is a required property';
                                    }
                                    else {
                                        out += 'should have required property \\\'' + ($missingProperty) + '\\\'';
                                    }
                                    out += '\' ';
                                }
                                if (it.opts.verbose) {
                                    out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                                }
                                out += ' } ';
                            }
                            else {
                                out += ' {} ';
                            }
                            out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; } ';
                        }
                    }
                }
            }
            it.errorPath = $currentErrorPath;
        }
        else if ($breakOnError) {
            out += ' if (true) {';
        }
        return out;
    };
},
/* ajv/lib/dotjs/uniqueItems.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_uniqueItems(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        if (($schema || $isData) && it.opts.uniqueItems !== false) {
            if ($isData) {
                out += ' var ' + ($valid) + '; if (' + ($schemaValue) + ' === false || ' + ($schemaValue) + ' === undefined) ' + ($valid) + ' = true; else if (typeof ' + ($schemaValue) + ' != \'boolean\') ' + ($valid) + ' = false; else { ';
            }
            out += ' var i = ' + ($data) + '.length , ' + ($valid) + ' = true , j; if (i > 1) { ';
            var $itemType = it.schema.items && it.schema.items.type, $typeIsArray = Array.isArray($itemType);
            if (!$itemType || $itemType == 'object' || $itemType == 'array' || ($typeIsArray && ($itemType.indexOf('object') >= 0 || $itemType.indexOf('array') >= 0))) {
                out += ' outer: for (;i--;) { for (j = i; j--;) { if (equal(' + ($data) + '[i], ' + ($data) + '[j])) { ' + ($valid) + ' = false; break outer; } } } ';
            }
            else {
                out += ' var itemIndices = {}, item; for (;i--;) { var item = ' + ($data) + '[i]; ';
                var $method = 'checkDataType' + ($typeIsArray ? 's' : '');
                out += ' if (' + (it.util[$method]($itemType, 'item', it.opts.strictNumbers, true)) + ') continue; ';
                if ($typeIsArray) {
                    out += ' if (typeof item == \'string\') item = \'"\' + item; ';
                }
                out += ' if (typeof itemIndices[item] == \'number\') { ' + ($valid) + ' = false; j = itemIndices[item]; break; } itemIndices[item] = i; } ';
            }
            out += ' } ';
            if ($isData) {
                out += '  }  ';
            }
            out += ' if (!' + ($valid) + ') {   ';
            var $$outStack = $$outStack || [];
            $$outStack.push(out);
            out = ''; /* istanbul ignore else */
            if (it.createErrors !== false) {
                out += ' { keyword: \'' + ('uniqueItems') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { i: i, j: j } ';
                if (it.opts.messages !== false) {
                    out += ' , message: \'should NOT have duplicate items (items ## \' + j + \' and \' + i + \' are identical)\' ';
                }
                if (it.opts.verbose) {
                    out += ' , schema:  ';
                    if ($isData) {
                        out += 'validate.schema' + ($schemaPath);
                    }
                    else {
                        out += '' + ($schema);
                    }
                    out += '         , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                }
                out += ' } ';
            }
            else {
                out += ' {} ';
            }
            var __err = out;
            out = $$outStack.pop();
            if (!it.compositeRule && $breakOnError) {
                /* istanbul ignore if */
                if (it.async) {
                    out += ' throw new ValidationError([' + (__err) + ']); ';
                }
                else {
                    out += ' validate.errors = [' + (__err) + ']; return false; ';
                }
            }
            else {
                out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
            }
            out += ' } ';
            if ($breakOnError) {
                out += ' else { ';
            }
        }
        else {
            if ($breakOnError) {
                out += ' if (true) { ';
            }
        }
        return out;
    };
},
/* ajv/lib/data.js */ function _(require, module, exports, __esModule, __esExport) {
    var KEYWORDS = [
        'multipleOf',
        'maximum',
        'exclusiveMaximum',
        'minimum',
        'exclusiveMinimum',
        'maxLength',
        'minLength',
        'pattern',
        'additionalItems',
        'maxItems',
        'minItems',
        'uniqueItems',
        'maxProperties',
        'minProperties',
        'required',
        'additionalProperties',
        'enum',
        'format',
        'const'
    ];
    module.exports = function (metaSchema, keywordsJsonPointers) {
        for (var i = 0; i < keywordsJsonPointers.length; i++) {
            metaSchema = JSON.parse(JSON.stringify(metaSchema));
            var segments = keywordsJsonPointers[i].split('/');
            var keywords = metaSchema;
            var j;
            for (j = 1; j < segments.length; j++)
                keywords = keywords[segments[j]];
            for (j = 0; j < KEYWORDS.length; j++) {
                var key = KEYWORDS[j];
                var schema = keywords[key];
                if (schema) {
                    keywords[key] = {
                        anyOf: [
                            schema,
                            { $ref: 'https://raw.githubusercontent.com/ajv-validator/ajv/master/lib/refs/data.json#' }
                        ]
                    };
                }
            }
        }
        return metaSchema;
    };
},
/* ajv/lib/compile/async.js */ function _(require, module, exports, __esModule, __esExport) {
    var MissingRefError = require(232) /* ./error_classes */.MissingRef;
    module.exports = compileAsync;
    /**
     * Creates validating function for passed schema with asynchronous loading of missing schemas.
     * `loadSchema` option should be a function that accepts schema uri and returns promise that resolves with the schema.
     * @this  Ajv
     * @param {Object}   schema schema object
     * @param {Boolean}  meta optional true to compile meta-schema; this parameter can be skipped
     * @param {Function} callback an optional node-style callback, it is called with 2 parameters: error (or null) and validating function.
     * @return {Promise} promise that resolves with a validating function.
     */
    function compileAsync(schema, meta, callback) {
        /* eslint no-shadow: 0 */
        /* global Promise */
        /* jshint validthis: true */
        var self = this;
        if (typeof this._opts.loadSchema != 'function')
            throw new Error('options.loadSchema should be a function');
        if (typeof meta == 'function') {
            callback = meta;
            meta = undefined;
        }
        var p = loadMetaSchemaOf(schema).then(function () {
            var schemaObj = self._addSchema(schema, undefined, meta);
            return schemaObj.validate || _compileAsync(schemaObj);
        });
        if (callback) {
            p.then(function (v) { callback(null, v); }, callback);
        }
        return p;
        function loadMetaSchemaOf(sch) {
            var $schema = sch.$schema;
            return $schema && !self.getSchema($schema)
                ? compileAsync.call(self, { $ref: $schema }, true)
                : Promise.resolve();
        }
        function _compileAsync(schemaObj) {
            try {
                return self._compile(schemaObj);
            }
            catch (e) {
                if (e instanceof MissingRefError)
                    return loadMissingSchema(e);
                throw e;
            }
            function loadMissingSchema(e) {
                var ref = e.missingSchema;
                if (added(ref))
                    throw new Error('Schema ' + ref + ' is loaded but ' + e.missingRef + ' cannot be resolved');
                var schemaPromise = self._loadingSchemas[ref];
                if (!schemaPromise) {
                    schemaPromise = self._loadingSchemas[ref] = self._opts.loadSchema(ref);
                    schemaPromise.then(removePromise, removePromise);
                }
                return schemaPromise.then(function (sch) {
                    if (!added(ref)) {
                        return loadMetaSchemaOf(sch).then(function () {
                            if (!added(ref))
                                self.addSchema(sch, ref, undefined, meta);
                        });
                    }
                }).then(function () {
                    return _compileAsync(schemaObj);
                });
                function removePromise() {
                    delete self._loadingSchemas[ref];
                }
                function added(ref) {
                    return self._refs[ref] || self._schemas[ref];
                }
            }
        }
    }
},
/* ajv/lib/keyword.js */ function _(require, module, exports, __esModule, __esExport) {
    var IDENTIFIER = /^[a-z_$][a-z0-9_$-]*$/i;
    var customRuleCode = require(265) /* ./dotjs/custom */;
    var definitionSchema = require(266) /* ./definition_schema */;
    module.exports = {
        add: addKeyword,
        get: getKeyword,
        remove: removeKeyword,
        validate: validateKeyword
    };
    /**
     * Define custom keyword
     * @this  Ajv
     * @param {String} keyword custom keyword, should be unique (including different from all standard, custom and macro keywords).
     * @param {Object} definition keyword definition object with properties `type` (type(s) which the keyword applies to), `validate` or `compile`.
     * @return {Ajv} this for method chaining
     */
    function addKeyword(keyword, definition) {
        /* jshint validthis: true */
        /* eslint no-shadow: 0 */
        var RULES = this.RULES;
        if (RULES.keywords[keyword])
            throw new Error('Keyword ' + keyword + ' is already defined');
        if (!IDENTIFIER.test(keyword))
            throw new Error('Keyword ' + keyword + ' is not a valid identifier');
        if (definition) {
            this.validateKeyword(definition, true);
            var dataType = definition.type;
            if (Array.isArray(dataType)) {
                for (var i = 0; i < dataType.length; i++)
                    _addRule(keyword, dataType[i], definition);
            }
            else {
                _addRule(keyword, dataType, definition);
            }
            var metaSchema = definition.metaSchema;
            if (metaSchema) {
                if (definition.$data && this._opts.$data) {
                    metaSchema = {
                        anyOf: [
                            metaSchema,
                            { '$ref': 'https://raw.githubusercontent.com/ajv-validator/ajv/master/lib/refs/data.json#' }
                        ]
                    };
                }
                definition.validateSchema = this.compile(metaSchema, true);
            }
        }
        RULES.keywords[keyword] = RULES.all[keyword] = true;
        function _addRule(keyword, dataType, definition) {
            var ruleGroup;
            for (var i = 0; i < RULES.length; i++) {
                var rg = RULES[i];
                if (rg.type == dataType) {
                    ruleGroup = rg;
                    break;
                }
            }
            if (!ruleGroup) {
                ruleGroup = { type: dataType, rules: [] };
                RULES.push(ruleGroup);
            }
            var rule = {
                keyword: keyword,
                definition: definition,
                custom: true,
                code: customRuleCode,
                implements: definition.implements
            };
            ruleGroup.rules.push(rule);
            RULES.custom[keyword] = rule;
        }
        return this;
    }
    /**
     * Get keyword
     * @this  Ajv
     * @param {String} keyword pre-defined or custom keyword.
     * @return {Object|Boolean} custom keyword definition, `true` if it is a predefined keyword, `false` otherwise.
     */
    function getKeyword(keyword) {
        /* jshint validthis: true */
        var rule = this.RULES.custom[keyword];
        return rule ? rule.definition : this.RULES.keywords[keyword] || false;
    }
    /**
     * Remove keyword
     * @this  Ajv
     * @param {String} keyword pre-defined or custom keyword.
     * @return {Ajv} this for method chaining
     */
    function removeKeyword(keyword) {
        /* jshint validthis: true */
        var RULES = this.RULES;
        delete RULES.keywords[keyword];
        delete RULES.all[keyword];
        delete RULES.custom[keyword];
        for (var i = 0; i < RULES.length; i++) {
            var rules = RULES[i].rules;
            for (var j = 0; j < rules.length; j++) {
                if (rules[j].keyword == keyword) {
                    rules.splice(j, 1);
                    break;
                }
            }
        }
        return this;
    }
    /**
     * Validate keyword definition
     * @this  Ajv
     * @param {Object} definition keyword definition object.
     * @param {Boolean} throwError true to throw exception if definition is invalid
     * @return {boolean} validation result
     */
    function validateKeyword(definition, throwError) {
        validateKeyword.errors = null;
        var v = this._validateKeyword = this._validateKeyword
            || this.compile(definitionSchema, true);
        if (v(definition))
            return true;
        validateKeyword.errors = v.errors;
        if (throwError)
            throw new Error('custom keyword definition is invalid: ' + this.errorsText(v.errors));
        else
            return false;
    }
},
/* ajv/lib/dotjs/custom.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function generate_custom(it, $keyword, $ruleType) {
        var out = ' ';
        var $lvl = it.level;
        var $dataLvl = it.dataLevel;
        var $schema = it.schema[$keyword];
        var $schemaPath = it.schemaPath + it.util.getProperty($keyword);
        var $errSchemaPath = it.errSchemaPath + '/' + $keyword;
        var $breakOnError = !it.opts.allErrors;
        var $errorKeyword;
        var $data = 'data' + ($dataLvl || '');
        var $valid = 'valid' + $lvl;
        var $errs = 'errs__' + $lvl;
        var $isData = it.opts.$data && $schema && $schema.$data, $schemaValue;
        if ($isData) {
            out += ' var schema' + ($lvl) + ' = ' + (it.util.getData($schema.$data, $dataLvl, it.dataPathArr)) + '; ';
            $schemaValue = 'schema' + $lvl;
        }
        else {
            $schemaValue = $schema;
        }
        var $rule = this, $definition = 'definition' + $lvl, $rDef = $rule.definition, $closingBraces = '';
        var $compile, $inline, $macro, $ruleValidate, $validateCode;
        if ($isData && $rDef.$data) {
            $validateCode = 'keywordValidate' + $lvl;
            var $validateSchema = $rDef.validateSchema;
            out += ' var ' + ($definition) + ' = RULES.custom[\'' + ($keyword) + '\'].definition; var ' + ($validateCode) + ' = ' + ($definition) + '.validate;';
        }
        else {
            $ruleValidate = it.useCustomRule($rule, $schema, it.schema, it);
            if (!$ruleValidate)
                return;
            $schemaValue = 'validate.schema' + $schemaPath;
            $validateCode = $ruleValidate.code;
            $compile = $rDef.compile;
            $inline = $rDef.inline;
            $macro = $rDef.macro;
        }
        var $ruleErrs = $validateCode + '.errors', $i = 'i' + $lvl, $ruleErr = 'ruleErr' + $lvl, $asyncKeyword = $rDef.async;
        if ($asyncKeyword && !it.async)
            throw new Error('async keyword in sync schema');
        if (!($inline || $macro)) {
            out += '' + ($ruleErrs) + ' = null;';
        }
        out += 'var ' + ($errs) + ' = errors;var ' + ($valid) + ';';
        if ($isData && $rDef.$data) {
            $closingBraces += '}';
            out += ' if (' + ($schemaValue) + ' === undefined) { ' + ($valid) + ' = true; } else { ';
            if ($validateSchema) {
                $closingBraces += '}';
                out += ' ' + ($valid) + ' = ' + ($definition) + '.validateSchema(' + ($schemaValue) + '); if (' + ($valid) + ') { ';
            }
        }
        if ($inline) {
            if ($rDef.statements) {
                out += ' ' + ($ruleValidate.validate) + ' ';
            }
            else {
                out += ' ' + ($valid) + ' = ' + ($ruleValidate.validate) + '; ';
            }
        }
        else if ($macro) {
            var $it = it.util.copy(it);
            var $closingBraces = '';
            $it.level++;
            var $nextValid = 'valid' + $it.level;
            $it.schema = $ruleValidate.validate;
            $it.schemaPath = '';
            var $wasComposite = it.compositeRule;
            it.compositeRule = $it.compositeRule = true;
            var $code = it.validate($it).replace(/validate\.schema/g, $validateCode);
            it.compositeRule = $it.compositeRule = $wasComposite;
            out += ' ' + ($code);
        }
        else {
            var $$outStack = $$outStack || [];
            $$outStack.push(out);
            out = '';
            out += '  ' + ($validateCode) + '.call( ';
            if (it.opts.passContext) {
                out += 'this';
            }
            else {
                out += 'self';
            }
            if ($compile || $rDef.schema === false) {
                out += ' , ' + ($data) + ' ';
            }
            else {
                out += ' , ' + ($schemaValue) + ' , ' + ($data) + ' , validate.schema' + (it.schemaPath) + ' ';
            }
            out += ' , (dataPath || \'\')';
            if (it.errorPath != '""') {
                out += ' + ' + (it.errorPath);
            }
            var $parentData = $dataLvl ? 'data' + (($dataLvl - 1) || '') : 'parentData', $parentDataProperty = $dataLvl ? it.dataPathArr[$dataLvl] : 'parentDataProperty';
            out += ' , ' + ($parentData) + ' , ' + ($parentDataProperty) + ' , rootData )  ';
            var def_callRuleValidate = out;
            out = $$outStack.pop();
            if ($rDef.errors === false) {
                out += ' ' + ($valid) + ' = ';
                if ($asyncKeyword) {
                    out += 'await ';
                }
                out += '' + (def_callRuleValidate) + '; ';
            }
            else {
                if ($asyncKeyword) {
                    $ruleErrs = 'customErrors' + $lvl;
                    out += ' var ' + ($ruleErrs) + ' = null; try { ' + ($valid) + ' = await ' + (def_callRuleValidate) + '; } catch (e) { ' + ($valid) + ' = false; if (e instanceof ValidationError) ' + ($ruleErrs) + ' = e.errors; else throw e; } ';
                }
                else {
                    out += ' ' + ($ruleErrs) + ' = null; ' + ($valid) + ' = ' + (def_callRuleValidate) + '; ';
                }
            }
        }
        if ($rDef.modifying) {
            out += ' if (' + ($parentData) + ') ' + ($data) + ' = ' + ($parentData) + '[' + ($parentDataProperty) + '];';
        }
        out += '' + ($closingBraces);
        if ($rDef.valid) {
            if ($breakOnError) {
                out += ' if (true) { ';
            }
        }
        else {
            out += ' if ( ';
            if ($rDef.valid === undefined) {
                out += ' !';
                if ($macro) {
                    out += '' + ($nextValid);
                }
                else {
                    out += '' + ($valid);
                }
            }
            else {
                out += ' ' + (!$rDef.valid) + ' ';
            }
            out += ') { ';
            $errorKeyword = $rule.keyword;
            var $$outStack = $$outStack || [];
            $$outStack.push(out);
            out = '';
            var $$outStack = $$outStack || [];
            $$outStack.push(out);
            out = ''; /* istanbul ignore else */
            if (it.createErrors !== false) {
                out += ' { keyword: \'' + ($errorKeyword || 'custom') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { keyword: \'' + ($rule.keyword) + '\' } ';
                if (it.opts.messages !== false) {
                    out += ' , message: \'should pass "' + ($rule.keyword) + '" keyword validation\' ';
                }
                if (it.opts.verbose) {
                    out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                }
                out += ' } ';
            }
            else {
                out += ' {} ';
            }
            var __err = out;
            out = $$outStack.pop();
            if (!it.compositeRule && $breakOnError) {
                /* istanbul ignore if */
                if (it.async) {
                    out += ' throw new ValidationError([' + (__err) + ']); ';
                }
                else {
                    out += ' validate.errors = [' + (__err) + ']; return false; ';
                }
            }
            else {
                out += ' var err = ' + (__err) + ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
            }
            var def_customError = out;
            out = $$outStack.pop();
            if ($inline) {
                if ($rDef.errors) {
                    if ($rDef.errors != 'full') {
                        out += '  for (var ' + ($i) + '=' + ($errs) + '; ' + ($i) + '<errors; ' + ($i) + '++) { var ' + ($ruleErr) + ' = vErrors[' + ($i) + ']; if (' + ($ruleErr) + '.dataPath === undefined) ' + ($ruleErr) + '.dataPath = (dataPath || \'\') + ' + (it.errorPath) + '; if (' + ($ruleErr) + '.schemaPath === undefined) { ' + ($ruleErr) + '.schemaPath = "' + ($errSchemaPath) + '"; } ';
                        if (it.opts.verbose) {
                            out += ' ' + ($ruleErr) + '.schema = ' + ($schemaValue) + '; ' + ($ruleErr) + '.data = ' + ($data) + '; ';
                        }
                        out += ' } ';
                    }
                }
                else {
                    if ($rDef.errors === false) {
                        out += ' ' + (def_customError) + ' ';
                    }
                    else {
                        out += ' if (' + ($errs) + ' == errors) { ' + (def_customError) + ' } else {  for (var ' + ($i) + '=' + ($errs) + '; ' + ($i) + '<errors; ' + ($i) + '++) { var ' + ($ruleErr) + ' = vErrors[' + ($i) + ']; if (' + ($ruleErr) + '.dataPath === undefined) ' + ($ruleErr) + '.dataPath = (dataPath || \'\') + ' + (it.errorPath) + '; if (' + ($ruleErr) + '.schemaPath === undefined) { ' + ($ruleErr) + '.schemaPath = "' + ($errSchemaPath) + '"; } ';
                        if (it.opts.verbose) {
                            out += ' ' + ($ruleErr) + '.schema = ' + ($schemaValue) + '; ' + ($ruleErr) + '.data = ' + ($data) + '; ';
                        }
                        out += ' } } ';
                    }
                }
            }
            else if ($macro) {
                out += '   var err =   '; /* istanbul ignore else */
                if (it.createErrors !== false) {
                    out += ' { keyword: \'' + ($errorKeyword || 'custom') + '\' , dataPath: (dataPath || \'\') + ' + (it.errorPath) + ' , schemaPath: ' + (it.util.toQuotedString($errSchemaPath)) + ' , params: { keyword: \'' + ($rule.keyword) + '\' } ';
                    if (it.opts.messages !== false) {
                        out += ' , message: \'should pass "' + ($rule.keyword) + '" keyword validation\' ';
                    }
                    if (it.opts.verbose) {
                        out += ' , schema: validate.schema' + ($schemaPath) + ' , parentSchema: validate.schema' + (it.schemaPath) + ' , data: ' + ($data) + ' ';
                    }
                    out += ' } ';
                }
                else {
                    out += ' {} ';
                }
                out += ';  if (vErrors === null) vErrors = [err]; else vErrors.push(err); errors++; ';
                if (!it.compositeRule && $breakOnError) {
                    /* istanbul ignore if */
                    if (it.async) {
                        out += ' throw new ValidationError(vErrors); ';
                    }
                    else {
                        out += ' validate.errors = vErrors; return false; ';
                    }
                }
            }
            else {
                if ($rDef.errors === false) {
                    out += ' ' + (def_customError) + ' ';
                }
                else {
                    out += ' if (Array.isArray(' + ($ruleErrs) + ')) { if (vErrors === null) vErrors = ' + ($ruleErrs) + '; else vErrors = vErrors.concat(' + ($ruleErrs) + '); errors = vErrors.length;  for (var ' + ($i) + '=' + ($errs) + '; ' + ($i) + '<errors; ' + ($i) + '++) { var ' + ($ruleErr) + ' = vErrors[' + ($i) + ']; if (' + ($ruleErr) + '.dataPath === undefined) ' + ($ruleErr) + '.dataPath = (dataPath || \'\') + ' + (it.errorPath) + ';  ' + ($ruleErr) + '.schemaPath = "' + ($errSchemaPath) + '";  ';
                    if (it.opts.verbose) {
                        out += ' ' + ($ruleErr) + '.schema = ' + ($schemaValue) + '; ' + ($ruleErr) + '.data = ' + ($data) + '; ';
                    }
                    out += ' } } else { ' + (def_customError) + ' } ';
                }
            }
            out += ' } ';
            if ($breakOnError) {
                out += ' else { ';
            }
        }
        return out;
    };
},
/* ajv/lib/definition_schema.js */ function _(require, module, exports, __esModule, __esExport) {
    var metaSchema = require(267) /* ./refs/json-schema-draft-07.json */;
    module.exports = {
        $id: 'https://github.com/ajv-validator/ajv/blob/master/lib/definition_schema.js',
        definitions: {
            simpleTypes: metaSchema.definitions.simpleTypes
        },
        type: 'object',
        dependencies: {
            schema: ['validate'],
            $data: ['validate'],
            statements: ['inline'],
            valid: { not: { required: ['macro'] } }
        },
        properties: {
            type: metaSchema.properties.type,
            schema: { type: 'boolean' },
            statements: { type: 'boolean' },
            dependencies: {
                type: 'array',
                items: { type: 'string' }
            },
            metaSchema: { type: 'object' },
            modifying: { type: 'boolean' },
            valid: { type: 'boolean' },
            $data: { type: 'boolean' },
            async: { type: 'boolean' },
            errors: {
                anyOf: [
                    { type: 'boolean' },
                    { const: 'full' }
                ]
            }
        }
    };
},
/* ajv/lib/refs/json-schema-draft-07.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://json-schema.org/draft-07/schema#",
        "title": "Core schema meta-schema",
        "definitions": {
            "schemaArray": {
                "type": "array",
                "minItems": 1,
                "items": { "$ref": "#" }
            },
            "nonNegativeInteger": {
                "type": "integer",
                "minimum": 0
            },
            "nonNegativeIntegerDefault0": {
                "allOf": [
                    { "$ref": "#/definitions/nonNegativeInteger" },
                    { "default": 0 }
                ]
            },
            "simpleTypes": {
                "enum": [
                    "array",
                    "boolean",
                    "integer",
                    "null",
                    "number",
                    "object",
                    "string"
                ]
            },
            "stringArray": {
                "type": "array",
                "items": { "type": "string" },
                "uniqueItems": true,
                "default": []
            }
        },
        "type": ["object", "boolean"],
        "properties": {
            "$id": {
                "type": "string",
                "format": "uri-reference"
            },
            "$schema": {
                "type": "string",
                "format": "uri"
            },
            "$ref": {
                "type": "string",
                "format": "uri-reference"
            },
            "$comment": {
                "type": "string"
            },
            "title": {
                "type": "string"
            },
            "description": {
                "type": "string"
            },
            "default": true,
            "readOnly": {
                "type": "boolean",
                "default": false
            },
            "examples": {
                "type": "array",
                "items": true
            },
            "multipleOf": {
                "type": "number",
                "exclusiveMinimum": 0
            },
            "maximum": {
                "type": "number"
            },
            "exclusiveMaximum": {
                "type": "number"
            },
            "minimum": {
                "type": "number"
            },
            "exclusiveMinimum": {
                "type": "number"
            },
            "maxLength": { "$ref": "#/definitions/nonNegativeInteger" },
            "minLength": { "$ref": "#/definitions/nonNegativeIntegerDefault0" },
            "pattern": {
                "type": "string",
                "format": "regex"
            },
            "additionalItems": { "$ref": "#" },
            "items": {
                "anyOf": [
                    { "$ref": "#" },
                    { "$ref": "#/definitions/schemaArray" }
                ],
                "default": true
            },
            "maxItems": { "$ref": "#/definitions/nonNegativeInteger" },
            "minItems": { "$ref": "#/definitions/nonNegativeIntegerDefault0" },
            "uniqueItems": {
                "type": "boolean",
                "default": false
            },
            "contains": { "$ref": "#" },
            "maxProperties": { "$ref": "#/definitions/nonNegativeInteger" },
            "minProperties": { "$ref": "#/definitions/nonNegativeIntegerDefault0" },
            "required": { "$ref": "#/definitions/stringArray" },
            "additionalProperties": { "$ref": "#" },
            "definitions": {
                "type": "object",
                "additionalProperties": { "$ref": "#" },
                "default": {}
            },
            "properties": {
                "type": "object",
                "additionalProperties": { "$ref": "#" },
                "default": {}
            },
            "patternProperties": {
                "type": "object",
                "additionalProperties": { "$ref": "#" },
                "propertyNames": { "format": "regex" },
                "default": {}
            },
            "dependencies": {
                "type": "object",
                "additionalProperties": {
                    "anyOf": [
                        { "$ref": "#" },
                        { "$ref": "#/definitions/stringArray" }
                    ]
                }
            },
            "propertyNames": { "$ref": "#" },
            "const": true,
            "enum": {
                "type": "array",
                "items": true,
                "minItems": 1,
                "uniqueItems": true
            },
            "type": {
                "anyOf": [
                    { "$ref": "#/definitions/simpleTypes" },
                    {
                        "type": "array",
                        "items": { "$ref": "#/definitions/simpleTypes" },
                        "minItems": 1,
                        "uniqueItems": true
                    }
                ]
            },
            "format": { "type": "string" },
            "contentMediaType": { "type": "string" },
            "contentEncoding": { "type": "string" },
            "if": { "$ref": "#" },
            "then": { "$ref": "#" },
            "else": { "$ref": "#" },
            "allOf": { "$ref": "#/definitions/schemaArray" },
            "anyOf": { "$ref": "#/definitions/schemaArray" },
            "oneOf": { "$ref": "#/definitions/schemaArray" },
            "not": { "$ref": "#" }
        },
        "default": true
    };
    module.exports = json;
},
/* ajv/lib/refs/data.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "https://raw.githubusercontent.com/ajv-validator/ajv/master/lib/refs/data.json#",
        "description": "Meta-schema for $data reference (JSON Schema extension proposal)",
        "type": "object",
        "required": ["$data"],
        "properties": {
            "$data": {
                "type": "string",
                "anyOf": [
                    { "format": "relative-json-pointer" },
                    { "format": "json-pointer" }
                ]
            }
        },
        "additionalProperties": false
    };
    module.exports = json;
},
/* ajv/lib/refs/json-schema-draft-04.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "id": "http://json-schema.org/draft-04/schema#",
        "$schema": "http://json-schema.org/draft-04/schema#",
        "description": "Core schema meta-schema",
        "definitions": {
            "schemaArray": {
                "type": "array",
                "minItems": 1,
                "items": { "$ref": "#" }
            },
            "positiveInteger": {
                "type": "integer",
                "minimum": 0
            },
            "positiveIntegerDefault0": {
                "allOf": [{ "$ref": "#/definitions/positiveInteger" }, { "default": 0 }]
            },
            "simpleTypes": {
                "enum": ["array", "boolean", "integer", "null", "number", "object", "string"]
            },
            "stringArray": {
                "type": "array",
                "items": { "type": "string" },
                "minItems": 1,
                "uniqueItems": true
            }
        },
        "type": "object",
        "properties": {
            "id": {
                "type": "string"
            },
            "$schema": {
                "type": "string"
            },
            "title": {
                "type": "string"
            },
            "description": {
                "type": "string"
            },
            "default": {},
            "multipleOf": {
                "type": "number",
                "minimum": 0,
                "exclusiveMinimum": true
            },
            "maximum": {
                "type": "number"
            },
            "exclusiveMaximum": {
                "type": "boolean",
                "default": false
            },
            "minimum": {
                "type": "number"
            },
            "exclusiveMinimum": {
                "type": "boolean",
                "default": false
            },
            "maxLength": { "$ref": "#/definitions/positiveInteger" },
            "minLength": { "$ref": "#/definitions/positiveIntegerDefault0" },
            "pattern": {
                "type": "string",
                "format": "regex"
            },
            "additionalItems": {
                "anyOf": [
                    { "type": "boolean" },
                    { "$ref": "#" }
                ],
                "default": {}
            },
            "items": {
                "anyOf": [
                    { "$ref": "#" },
                    { "$ref": "#/definitions/schemaArray" }
                ],
                "default": {}
            },
            "maxItems": { "$ref": "#/definitions/positiveInteger" },
            "minItems": { "$ref": "#/definitions/positiveIntegerDefault0" },
            "uniqueItems": {
                "type": "boolean",
                "default": false
            },
            "maxProperties": { "$ref": "#/definitions/positiveInteger" },
            "minProperties": { "$ref": "#/definitions/positiveIntegerDefault0" },
            "required": { "$ref": "#/definitions/stringArray" },
            "additionalProperties": {
                "anyOf": [
                    { "type": "boolean" },
                    { "$ref": "#" }
                ],
                "default": {}
            },
            "definitions": {
                "type": "object",
                "additionalProperties": { "$ref": "#" },
                "default": {}
            },
            "properties": {
                "type": "object",
                "additionalProperties": { "$ref": "#" },
                "default": {}
            },
            "patternProperties": {
                "type": "object",
                "additionalProperties": { "$ref": "#" },
                "default": {}
            },
            "dependencies": {
                "type": "object",
                "additionalProperties": {
                    "anyOf": [
                        { "$ref": "#" },
                        { "$ref": "#/definitions/stringArray" }
                    ]
                }
            },
            "enum": {
                "type": "array",
                "minItems": 1,
                "uniqueItems": true
            },
            "type": {
                "anyOf": [
                    { "$ref": "#/definitions/simpleTypes" },
                    {
                        "type": "array",
                        "items": { "$ref": "#/definitions/simpleTypes" },
                        "minItems": 1,
                        "uniqueItems": true
                    }
                ]
            },
            "format": { "type": "string" },
            "allOf": { "$ref": "#/definitions/schemaArray" },
            "anyOf": { "$ref": "#/definitions/schemaArray" },
            "oneOf": { "$ref": "#/definitions/schemaArray" },
            "not": { "$ref": "#" }
        },
        "dependencies": {
            "exclusiveMaximum": ["maximum"],
            "exclusiveMinimum": ["minimum"]
        },
        "default": {}
    };
    module.exports = json;
},
/* @eslint/eslintrc/lib/shared/naming.js */ function _(require, module, exports, __esModule, __esExport) {
    const NAMESPACE_REGEX = /^@.*\//iu;
    /**
     * Brings package name to correct format based on prefix
     * @param {string} name The name of the package.
     * @param {string} prefix Can be either "eslint-plugin", "eslint-config" or "eslint-formatter"
     * @returns {string} Normalized name of the package
     * @private
     */
    function normalizePackageName(name, prefix) {
        let normalizedName = name;
        /**
         * On Windows, name can come in with Windows slashes instead of Unix slashes.
         * Normalize to Unix first to avoid errors later on.
         * https://github.com/eslint/eslint/issues/5644
         */
        if (normalizedName.includes("\\")) {
            normalizedName = normalizedName.replace(/\\/gu, "/");
        }
        if (normalizedName.charAt(0) === "@") {
            /**
             * it's a scoped package
             * package name is the prefix, or just a username
             */
            const scopedPackageShortcutRegex = new RegExp(`^(@[^/]+)(?:/(?:${prefix})?)?$`, "u"), scopedPackageNameRegex = new RegExp(`^${prefix}(-|$)`, "u");
            if (scopedPackageShortcutRegex.test(normalizedName)) {
                normalizedName = normalizedName.replace(scopedPackageShortcutRegex, `$1/${prefix}`);
            }
            else if (!scopedPackageNameRegex.test(normalizedName.split("/")[1])) {
                /**
                 * for scoped packages, insert the prefix after the first / unless
                 * the path is already @scope/eslint or @scope/eslint-xxx-yyy
                 */
                normalizedName = normalizedName.replace(/^@([^/]+)\/(.*)$/u, `@$1/${prefix}-$2`);
            }
        }
        else if (!normalizedName.startsWith(`${prefix}-`)) {
            normalizedName = `${prefix}-${normalizedName}`;
        }
        return normalizedName;
    }
    /**
     * Removes the prefix from a fullname.
     * @param {string} fullname The term which may have the prefix.
     * @param {string} prefix The prefix to remove.
     * @returns {string} The term without prefix.
     */
    function getShorthandName(fullname, prefix) {
        if (fullname[0] === "@") {
            let matchResult = new RegExp(`^(@[^/]+)/${prefix}$`, "u").exec(fullname);
            if (matchResult) {
                return matchResult[1];
            }
            matchResult = new RegExp(`^(@[^/]+)/${prefix}-(.+)$`, "u").exec(fullname);
            if (matchResult) {
                return `${matchResult[1]}/${matchResult[2]}`;
            }
        }
        else if (fullname.startsWith(`${prefix}-`)) {
            return fullname.slice(prefix.length + 1);
        }
        return fullname;
    }
    /**
     * Gets the scope (namespace) of a term.
     * @param {string} term The term which may have the namespace.
     * @returns {string} The namespace of the term if it has one.
     */
    function getNamespaceFromTerm(term) {
        const match = term.match(NAMESPACE_REGEX);
        return match ? match[0] : "";
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = {
        normalizePackageName,
        getShorthandName,
        getNamespaceFromTerm
    };
},
/* @eslint/eslintrc/lib/shared/relative-module-resolver.js */ function _(require, module, exports, __esModule, __esExport) {
    const Module = require("module");
    /*
     * `Module.createRequire` is added in v12.2.0. It supports URL as well.
     * We only support the case where the argument is a filepath, not a URL.
     */
    // eslint-disable-next-line node/no-unsupported-features/node-builtins, node/no-deprecated-api
    const createRequire = Module.createRequire || Module.createRequireFromPath;
    module.exports = {
        /**
         * Resolves a Node module relative to another module
         * @param {string} moduleName The name of a Node module, or a path to a Node module.
         * @param {string} relativeToPath An absolute path indicating the module that `moduleName` should be resolved relative to. This must be
         * a file rather than a directory, but the file need not actually exist.
         * @returns {string} The absolute path that would result from calling `require.resolve(moduleName)` in a file located at `relativeToPath`
         */
        resolve(moduleName, relativeToPath) {
            try {
                return createRequire(relativeToPath).resolve(moduleName);
            }
            catch (error) {
                // This `if` block is for older Node.js than 12.0.0. We can remove this block in the future.
                if (typeof error === "object" &&
                    error !== null &&
                    error.code === "MODULE_NOT_FOUND" &&
                    !error.requireStack &&
                    error.message.includes(moduleName)) {
                    error.message += `\nRequire stack:\n- ${relativeToPath}`;
                }
                throw error;
            }
        }
    };
},
/* @eslint/eslintrc/lib/config-array/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const { ConfigArray, getUsedExtractedConfigs } = require(273) /* ./config-array */;
    const { ConfigDependency } = require(277) /* ./config-dependency */;
    const { ExtractedConfig } = require(274) /* ./extracted-config */;
    const { IgnorePattern } = require(275) /* ./ignore-pattern */;
    const { OverrideTester } = require(278) /* ./override-tester */;
    module.exports = {
        ConfigArray,
        ConfigDependency,
        ExtractedConfig,
        IgnorePattern,
        OverrideTester,
        getUsedExtractedConfigs
    };
},
/* @eslint/eslintrc/lib/config-array/config-array.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { ExtractedConfig } = require(274) /* ./extracted-config */;
    const { IgnorePattern } = require(275) /* ./ignore-pattern */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    // Define types for VSCode IntelliSense.
    /** @typedef {import("../../shared/types").Environment} Environment */
    /** @typedef {import("../../shared/types").GlobalConf} GlobalConf */
    /** @typedef {import("../../shared/types").RuleConf} RuleConf */
    /** @typedef {import("../../shared/types").Rule} Rule */
    /** @typedef {import("../../shared/types").Plugin} Plugin */
    /** @typedef {import("../../shared/types").Processor} Processor */
    /** @typedef {import("./config-dependency").DependentParser} DependentParser */
    /** @typedef {import("./config-dependency").DependentPlugin} DependentPlugin */
    /** @typedef {import("./override-tester")["OverrideTester"]} OverrideTester */
    /**
     * @typedef {Object} ConfigArrayElement
     * @property {string} name The name of this config element.
     * @property {string} filePath The path to the source file of this config element.
     * @property {InstanceType<OverrideTester>|null} criteria The tester for the `files` and `excludedFiles` of this config element.
     * @property {Record<string, boolean>|undefined} env The environment settings.
     * @property {Record<string, GlobalConf>|undefined} globals The global variable settings.
     * @property {IgnorePattern|undefined} ignorePattern The ignore patterns.
     * @property {boolean|undefined} noInlineConfig The flag that disables directive comments.
     * @property {DependentParser|undefined} parser The parser loader.
     * @property {Object|undefined} parserOptions The parser options.
     * @property {Record<string, DependentPlugin>|undefined} plugins The plugin loaders.
     * @property {string|undefined} processor The processor name to refer plugin's processor.
     * @property {boolean|undefined} reportUnusedDisableDirectives The flag to report unused `eslint-disable` comments.
     * @property {boolean|undefined} root The flag to express root.
     * @property {Record<string, RuleConf>|undefined} rules The rule settings
     * @property {Object|undefined} settings The shared settings.
     * @property {"config" | "ignore" | "implicit-processor"} type The element type.
     */
    /**
     * @typedef {Object} ConfigArrayInternalSlots
     * @property {Map<string, ExtractedConfig>} cache The cache to extract configs.
     * @property {ReadonlyMap<string, Environment>|null} envMap The map from environment ID to environment definition.
     * @property {ReadonlyMap<string, Processor>|null} processorMap The map from processor ID to environment definition.
     * @property {ReadonlyMap<string, Rule>|null} ruleMap The map from rule ID to rule definition.
     */
    /** @type {WeakMap<ConfigArray, ConfigArrayInternalSlots>} */
    const internalSlotsMap = new class extends WeakMap {
        get(key) {
            let value = super.get(key);
            if (!value) {
                value = {
                    cache: new Map(),
                    envMap: null,
                    processorMap: null,
                    ruleMap: null
                };
                super.set(key, value);
            }
            return value;
        }
    }();
    /**
     * Get the indices which are matched to a given file.
     * @param {ConfigArrayElement[]} elements The elements.
     * @param {string} filePath The path to a target file.
     * @returns {number[]} The indices.
     */
    function getMatchedIndices(elements, filePath) {
        const indices = [];
        for (let i = elements.length - 1; i >= 0; --i) {
            const element = elements[i];
            if (!element.criteria || (filePath && element.criteria.test(filePath))) {
                indices.push(i);
            }
        }
        return indices;
    }
    /**
     * Check if a value is a non-null object.
     * @param {any} x The value to check.
     * @returns {boolean} `true` if the value is a non-null object.
     */
    function isNonNullObject(x) {
        return typeof x === "object" && x !== null;
    }
    /**
     * Merge two objects.
     *
     * Assign every property values of `y` to `x` if `x` doesn't have the property.
     * If `x`'s property value is an object, it does recursive.
     * @param {Object} target The destination to merge
     * @param {Object|undefined} source The source to merge.
     * @returns {void}
     */
    function mergeWithoutOverwrite(target, source) {
        if (!isNonNullObject(source)) {
            return;
        }
        for (const key of Object.keys(source)) {
            if (key === "__proto__") {
                continue;
            }
            if (isNonNullObject(target[key])) {
                mergeWithoutOverwrite(target[key], source[key]);
            }
            else if (target[key] === void 0) {
                if (isNonNullObject(source[key])) {
                    target[key] = Array.isArray(source[key]) ? [] : {};
                    mergeWithoutOverwrite(target[key], source[key]);
                }
                else if (source[key] !== void 0) {
                    target[key] = source[key];
                }
            }
        }
    }
    /**
     * The error for plugin conflicts.
     */
    class PluginConflictError extends Error {
        /**
         * Initialize this error object.
         * @param {string} pluginId The plugin ID.
         * @param {{filePath:string, importerName:string}[]} plugins The resolved plugins.
         */
        constructor(pluginId, plugins) {
            super(`Plugin "${pluginId}" was conflicted between ${plugins.map(p => `"${p.importerName}"`).join(" and ")}.`);
            this.messageTemplate = "plugin-conflict";
            this.messageData = { pluginId, plugins };
        }
    }
    /**
     * Merge plugins.
     * `target`'s definition is prior to `source`'s.
     * @param {Record<string, DependentPlugin>} target The destination to merge
     * @param {Record<string, DependentPlugin>|undefined} source The source to merge.
     * @returns {void}
     */
    function mergePlugins(target, source) {
        if (!isNonNullObject(source)) {
            return;
        }
        for (const key of Object.keys(source)) {
            if (key === "__proto__") {
                continue;
            }
            const targetValue = target[key];
            const sourceValue = source[key];
            // Adopt the plugin which was found at first.
            if (targetValue === void 0) {
                if (sourceValue.error) {
                    throw sourceValue.error;
                }
                target[key] = sourceValue;
            }
            else if (sourceValue.filePath !== targetValue.filePath) {
                throw new PluginConflictError(key, [
                    {
                        filePath: targetValue.filePath,
                        importerName: targetValue.importerName
                    },
                    {
                        filePath: sourceValue.filePath,
                        importerName: sourceValue.importerName
                    }
                ]);
            }
        }
    }
    /**
     * Merge rule configs.
     * `target`'s definition is prior to `source`'s.
     * @param {Record<string, Array>} target The destination to merge
     * @param {Record<string, RuleConf>|undefined} source The source to merge.
     * @returns {void}
     */
    function mergeRuleConfigs(target, source) {
        if (!isNonNullObject(source)) {
            return;
        }
        for (const key of Object.keys(source)) {
            if (key === "__proto__") {
                continue;
            }
            const targetDef = target[key];
            const sourceDef = source[key];
            // Adopt the rule config which was found at first.
            if (targetDef === void 0) {
                if (Array.isArray(sourceDef)) {
                    target[key] = [...sourceDef];
                }
                else {
                    target[key] = [sourceDef];
                }
                /*
                 * If the first found rule config is severity only and the current rule
                 * config has options, merge the severity and the options.
                 */
            }
            else if (targetDef.length === 1 &&
                Array.isArray(sourceDef) &&
                sourceDef.length >= 2) {
                targetDef.push(...sourceDef.slice(1));
            }
        }
    }
    /**
     * Create the extracted config.
     * @param {ConfigArray} instance The config elements.
     * @param {number[]} indices The indices to use.
     * @returns {ExtractedConfig} The extracted config.
     */
    function createConfig(instance, indices) {
        const config = new ExtractedConfig();
        const ignorePatterns = [];
        // Merge elements.
        for (const index of indices) {
            const element = instance[index];
            // Adopt the parser which was found at first.
            if (!config.parser && element.parser) {
                if (element.parser.error) {
                    throw element.parser.error;
                }
                config.parser = element.parser;
            }
            // Adopt the processor which was found at first.
            if (!config.processor && element.processor) {
                config.processor = element.processor;
            }
            // Adopt the noInlineConfig which was found at first.
            if (config.noInlineConfig === void 0 && element.noInlineConfig !== void 0) {
                config.noInlineConfig = element.noInlineConfig;
                config.configNameOfNoInlineConfig = element.name;
            }
            // Adopt the reportUnusedDisableDirectives which was found at first.
            if (config.reportUnusedDisableDirectives === void 0 && element.reportUnusedDisableDirectives !== void 0) {
                config.reportUnusedDisableDirectives = element.reportUnusedDisableDirectives;
            }
            // Collect ignorePatterns
            if (element.ignorePattern) {
                ignorePatterns.push(element.ignorePattern);
            }
            // Merge others.
            mergeWithoutOverwrite(config.env, element.env);
            mergeWithoutOverwrite(config.globals, element.globals);
            mergeWithoutOverwrite(config.parserOptions, element.parserOptions);
            mergeWithoutOverwrite(config.settings, element.settings);
            mergePlugins(config.plugins, element.plugins);
            mergeRuleConfigs(config.rules, element.rules);
        }
        // Create the predicate function for ignore patterns.
        if (ignorePatterns.length > 0) {
            config.ignores = IgnorePattern.createIgnore(ignorePatterns.reverse());
        }
        return config;
    }
    /**
     * Collect definitions.
     * @template T, U
     * @param {string} pluginId The plugin ID for prefix.
     * @param {Record<string,T>} defs The definitions to collect.
     * @param {Map<string, U>} map The map to output.
     * @param {function(T): U} [normalize] The normalize function for each value.
     * @returns {void}
     */
    function collect(pluginId, defs, map, normalize) {
        if (defs) {
            const prefix = pluginId && `${pluginId}/`;
            for (const [key, value] of Object.entries(defs)) {
                map.set(`${prefix}${key}`, normalize ? normalize(value) : value);
            }
        }
    }
    /**
     * Normalize a rule definition.
     * @param {Function|Rule} rule The rule definition to normalize.
     * @returns {Rule} The normalized rule definition.
     */
    function normalizePluginRule(rule) {
        return typeof rule === "function" ? { create: rule } : rule;
    }
    /**
     * Delete the mutation methods from a given map.
     * @param {Map<any, any>} map The map object to delete.
     * @returns {void}
     */
    function deleteMutationMethods(map) {
        Object.defineProperties(map, {
            clear: { configurable: true, value: void 0 },
            delete: { configurable: true, value: void 0 },
            set: { configurable: true, value: void 0 }
        });
    }
    /**
     * Create `envMap`, `processorMap`, `ruleMap` with the plugins in the config array.
     * @param {ConfigArrayElement[]} elements The config elements.
     * @param {ConfigArrayInternalSlots} slots The internal slots.
     * @returns {void}
     */
    function initPluginMemberMaps(elements, slots) {
        const processed = new Set();
        slots.envMap = new Map();
        slots.processorMap = new Map();
        slots.ruleMap = new Map();
        for (const element of elements) {
            if (!element.plugins) {
                continue;
            }
            for (const [pluginId, value] of Object.entries(element.plugins)) {
                const plugin = value.definition;
                if (!plugin || processed.has(pluginId)) {
                    continue;
                }
                processed.add(pluginId);
                collect(pluginId, plugin.environments, slots.envMap);
                collect(pluginId, plugin.processors, slots.processorMap);
                collect(pluginId, plugin.rules, slots.ruleMap, normalizePluginRule);
            }
        }
        deleteMutationMethods(slots.envMap);
        deleteMutationMethods(slots.processorMap);
        deleteMutationMethods(slots.ruleMap);
    }
    /**
     * Create `envMap`, `processorMap`, `ruleMap` with the plugins in the config array.
     * @param {ConfigArray} instance The config elements.
     * @returns {ConfigArrayInternalSlots} The extracted config.
     */
    function ensurePluginMemberMaps(instance) {
        const slots = internalSlotsMap.get(instance);
        if (!slots.ruleMap) {
            initPluginMemberMaps(instance, slots);
        }
        return slots;
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * The Config Array.
     *
     * `ConfigArray` instance contains all settings, parsers, and plugins.
     * You need to call `ConfigArray#extractConfig(filePath)` method in order to
     * extract, merge and get only the config data which is related to an arbitrary
     * file.
     * @extends {Array<ConfigArrayElement>}
     */
    class ConfigArray extends Array {
        /**
         * Get the plugin environments.
         * The returned map cannot be mutated.
         * @type {ReadonlyMap<string, Environment>} The plugin environments.
         */
        get pluginEnvironments() {
            return ensurePluginMemberMaps(this).envMap;
        }
        /**
         * Get the plugin processors.
         * The returned map cannot be mutated.
         * @type {ReadonlyMap<string, Processor>} The plugin processors.
         */
        get pluginProcessors() {
            return ensurePluginMemberMaps(this).processorMap;
        }
        /**
         * Get the plugin rules.
         * The returned map cannot be mutated.
         * @returns {ReadonlyMap<string, Rule>} The plugin rules.
         */
        get pluginRules() {
            return ensurePluginMemberMaps(this).ruleMap;
        }
        /**
         * Check if this config has `root` flag.
         * @returns {boolean} `true` if this config array is root.
         */
        isRoot() {
            for (let i = this.length - 1; i >= 0; --i) {
                const root = this[i].root;
                if (typeof root === "boolean") {
                    return root;
                }
            }
            return false;
        }
        /**
         * Extract the config data which is related to a given file.
         * @param {string} filePath The absolute path to the target file.
         * @returns {ExtractedConfig} The extracted config data.
         */
        extractConfig(filePath) {
            const { cache } = internalSlotsMap.get(this);
            const indices = getMatchedIndices(this, filePath);
            const cacheKey = indices.join(",");
            if (!cache.has(cacheKey)) {
                cache.set(cacheKey, createConfig(this, indices));
            }
            return cache.get(cacheKey);
        }
        /**
         * Check if a given path is an additional lint target.
         * @param {string} filePath The absolute path to the target file.
         * @returns {boolean} `true` if the file is an additional lint target.
         */
        isAdditionalTargetPath(filePath) {
            for (const { criteria, type } of this) {
                if (type === "config" &&
                    criteria &&
                    !criteria.endsWithWildcard &&
                    criteria.test(filePath)) {
                    return true;
                }
            }
            return false;
        }
    }
    const exportObject = {
        ConfigArray,
        /**
         * Get the used extracted configs.
         * CLIEngine will use this method to collect used deprecated rules.
         * @param {ConfigArray} instance The config array object to get.
         * @returns {ExtractedConfig[]} The used extracted configs.
         * @private
         */
        getUsedExtractedConfigs(instance) {
            const { cache } = internalSlotsMap.get(instance);
            return Array.from(cache.values());
        }
    };
    module.exports = exportObject;
},
/* @eslint/eslintrc/lib/config-array/extracted-config.js */ function _(require, module, exports, __esModule, __esExport) {
    const { IgnorePattern } = require(275) /* ./ignore-pattern */;
    // For VSCode intellisense
    /** @typedef {import("../../shared/types").ConfigData} ConfigData */
    /** @typedef {import("../../shared/types").GlobalConf} GlobalConf */
    /** @typedef {import("../../shared/types").SeverityConf} SeverityConf */
    /** @typedef {import("./config-dependency").DependentParser} DependentParser */
    /** @typedef {import("./config-dependency").DependentPlugin} DependentPlugin */
    /**
     * Check if `xs` starts with `ys`.
     * @template T
     * @param {T[]} xs The array to check.
     * @param {T[]} ys The array that may be the first part of `xs`.
     * @returns {boolean} `true` if `xs` starts with `ys`.
     */
    function startsWith(xs, ys) {
        return xs.length >= ys.length && ys.every((y, i) => y === xs[i]);
    }
    /**
     * The class for extracted config data.
     */
    class ExtractedConfig {
        constructor() {
            /**
             * The config name what `noInlineConfig` setting came from.
             * @type {string}
             */
            this.configNameOfNoInlineConfig = "";
            /**
             * Environments.
             * @type {Record<string, boolean>}
             */
            this.env = {};
            /**
             * Global variables.
             * @type {Record<string, GlobalConf>}
             */
            this.globals = {};
            /**
             * The glob patterns that ignore to lint.
             * @type {(((filePath:string, dot?:boolean) => boolean) & { basePath:string; patterns:string[] }) | undefined}
             */
            this.ignores = void 0;
            /**
             * The flag that disables directive comments.
             * @type {boolean|undefined}
             */
            this.noInlineConfig = void 0;
            /**
             * Parser definition.
             * @type {DependentParser|null}
             */
            this.parser = null;
            /**
             * Options for the parser.
             * @type {Object}
             */
            this.parserOptions = {};
            /**
             * Plugin definitions.
             * @type {Record<string, DependentPlugin>}
             */
            this.plugins = {};
            /**
             * Processor ID.
             * @type {string|null}
             */
            this.processor = null;
            /**
             * The flag that reports unused `eslint-disable` directive comments.
             * @type {boolean|undefined}
             */
            this.reportUnusedDisableDirectives = void 0;
            /**
             * Rule settings.
             * @type {Record<string, [SeverityConf, ...any[]]>}
             */
            this.rules = {};
            /**
             * Shared settings.
             * @type {Object}
             */
            this.settings = {};
        }
        /**
         * Convert this config to the compatible object as a config file content.
         * @returns {ConfigData} The converted object.
         */
        toCompatibleObjectAsConfigFileContent() {
            const { 
            /* eslint-disable no-unused-vars */
            configNameOfNoInlineConfig: _ignore1, processor: _ignore2, 
            /* eslint-enable no-unused-vars */
            ignores, ...config } = this;
            config.parser = config.parser && config.parser.filePath;
            config.plugins = Object.keys(config.plugins).filter(Boolean).reverse();
            config.ignorePatterns = ignores ? ignores.patterns : [];
            // Strip the default patterns from `ignorePatterns`.
            if (startsWith(config.ignorePatterns, IgnorePattern.DefaultPatterns)) {
                config.ignorePatterns =
                    config.ignorePatterns.slice(IgnorePattern.DefaultPatterns.length);
            }
            return config;
        }
    }
    module.exports = { ExtractedConfig };
},
/* @eslint/eslintrc/lib/config-array/ignore-pattern.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const assert = require("assert");
    const path = require("path");
    const ignore = require(276) /* ignore */;
    const debug = require(72) /* debug */("eslintrc:ignore-pattern");
    /** @typedef {ReturnType<import("ignore").default>} Ignore */
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Get the path to the common ancestor directory of given paths.
     * @param {string[]} sourcePaths The paths to calculate the common ancestor.
     * @returns {string} The path to the common ancestor directory.
     */
    function getCommonAncestorPath(sourcePaths) {
        let result = sourcePaths[0];
        for (let i = 1; i < sourcePaths.length; ++i) {
            const a = result;
            const b = sourcePaths[i];
            // Set the shorter one (it's the common ancestor if one includes the other).
            result = a.length < b.length ? a : b;
            // Set the common ancestor.
            for (let j = 0, lastSepPos = 0; j < a.length && j < b.length; ++j) {
                if (a[j] !== b[j]) {
                    result = a.slice(0, lastSepPos);
                    break;
                }
                if (a[j] === path.sep) {
                    lastSepPos = j;
                }
            }
        }
        let resolvedResult = result || path.sep;
        // if Windows common ancestor is root of drive must have trailing slash to be absolute.
        if (resolvedResult && resolvedResult.endsWith(":") && process.platform === "win32") {
            resolvedResult += path.sep;
        }
        return resolvedResult;
    }
    /**
     * Make relative path.
     * @param {string} from The source path to get relative path.
     * @param {string} to The destination path to get relative path.
     * @returns {string} The relative path.
     */
    function relative(from, to) {
        const relPath = path.relative(from, to);
        if (path.sep === "/") {
            return relPath;
        }
        return relPath.split(path.sep).join("/");
    }
    /**
     * Get the trailing slash if existed.
     * @param {string} filePath The path to check.
     * @returns {string} The trailing slash if existed.
     */
    function dirSuffix(filePath) {
        const isDir = (filePath.endsWith(path.sep) ||
            (process.platform === "win32" && filePath.endsWith("/")));
        return isDir ? "/" : "";
    }
    const DefaultPatterns = Object.freeze(["/**/node_modules/*"]);
    const DotPatterns = Object.freeze([".*", "!.eslintrc.*", "!../"]);
    //------------------------------------------------------------------------------
    // Public
    //------------------------------------------------------------------------------
    class IgnorePattern {
        /**
         * The default patterns.
         * @type {string[]}
         */
        static get DefaultPatterns() {
            return DefaultPatterns;
        }
        /**
         * Create the default predicate function.
         * @param {string} cwd The current working directory.
         * @returns {((filePath:string, dot:boolean) => boolean) & {basePath:string; patterns:string[]}}
         * The preficate function.
         * The first argument is an absolute path that is checked.
         * The second argument is the flag to not ignore dotfiles.
         * If the predicate function returned `true`, it means the path should be ignored.
         */
        static createDefaultIgnore(cwd) {
            return this.createIgnore([new IgnorePattern(DefaultPatterns, cwd)]);
        }
        /**
         * Create the predicate function from multiple `IgnorePattern` objects.
         * @param {IgnorePattern[]} ignorePatterns The list of ignore patterns.
         * @returns {((filePath:string, dot?:boolean) => boolean) & {basePath:string; patterns:string[]}}
         * The preficate function.
         * The first argument is an absolute path that is checked.
         * The second argument is the flag to not ignore dotfiles.
         * If the predicate function returned `true`, it means the path should be ignored.
         */
        static createIgnore(ignorePatterns) {
            debug("Create with: %o", ignorePatterns);
            const basePath = getCommonAncestorPath(ignorePatterns.map(p => p.basePath));
            const patterns = [].concat(...ignorePatterns.map(p => p.getPatternsRelativeTo(basePath)));
            const ig = ignore().add([...DotPatterns, ...patterns]);
            const dotIg = ignore().add(patterns);
            debug("  processed: %o", { basePath, patterns });
            return Object.assign((filePath, dot = false) => {
                assert(path.isAbsolute(filePath), "'filePath' should be an absolute path.");
                const relPathRaw = relative(basePath, filePath);
                const relPath = relPathRaw && (relPathRaw + dirSuffix(filePath));
                const adoptedIg = dot ? dotIg : ig;
                const result = relPath !== "" && adoptedIg.ignores(relPath);
                debug("Check", { filePath, dot, relativePath: relPath, result });
                return result;
            }, { basePath, patterns });
        }
        /**
         * Initialize a new `IgnorePattern` instance.
         * @param {string[]} patterns The glob patterns that ignore to lint.
         * @param {string} basePath The base path of `patterns`.
         */
        constructor(patterns, basePath) {
            assert(path.isAbsolute(basePath), "'basePath' should be an absolute path.");
            /**
             * The glob patterns that ignore to lint.
             * @type {string[]}
             */
            this.patterns = patterns;
            /**
             * The base path of `patterns`.
             * @type {string}
             */
            this.basePath = basePath;
            /**
             * If `true` then patterns which don't start with `/` will match the paths to the outside of `basePath`. Defaults to `false`.
             *
             * It's set `true` for `.eslintignore`, `package.json`, and `--ignore-path` for backward compatibility.
             * It's `false` as-is for `ignorePatterns` property in config files.
             * @type {boolean}
             */
            this.loose = false;
        }
        /**
         * Get `patterns` as modified for a given base path. It modifies the
         * absolute paths in the patterns as prepending the difference of two base
         * paths.
         * @param {string} newBasePath The base path.
         * @returns {string[]} Modifired patterns.
         */
        getPatternsRelativeTo(newBasePath) {
            assert(path.isAbsolute(newBasePath), "'newBasePath' should be an absolute path.");
            const { basePath, loose, patterns } = this;
            if (newBasePath === basePath) {
                return patterns;
            }
            const prefix = `/${relative(newBasePath, basePath)}`;
            return patterns.map(pattern => {
                const negative = pattern.startsWith("!");
                const head = negative ? "!" : "";
                const body = negative ? pattern.slice(1) : pattern;
                if (body.startsWith("/") || body.startsWith("../")) {
                    return `${head}${prefix}${body}`;
                }
                return loose ? pattern : `${head}${prefix}/**/${body}`;
            });
        }
    }
    module.exports = { IgnorePattern };
},
/* ignore/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // A simple implementation of make-array
    function make_array(subject) {
        return Array.isArray(subject)
            ? subject
            : [subject];
    }
    const REGEX_BLANK_LINE = /^\s+$/;
    const REGEX_LEADING_EXCAPED_EXCLAMATION = /^\\!/;
    const REGEX_LEADING_EXCAPED_HASH = /^\\#/;
    const SLASH = '/';
    const KEY_IGNORE = typeof Symbol !== 'undefined'
        ? Symbol.for('node-ignore')
        /* istanbul ignore next */
        : 'node-ignore';
    const define = (object, key, value) => Object.defineProperty(object, key, { value });
    const REGEX_REGEXP_RANGE = /([0-z])-([0-z])/g;
    // Sanitize the range of a regular expression
    // The cases are complicated, see test cases for details
    const sanitizeRange = range => range.replace(REGEX_REGEXP_RANGE, (match, from, to) => from.charCodeAt(0) <= to.charCodeAt(0)
        ? match
        // Invalid range (out of order) which is ok for gitignore rules but
        //   fatal for JavaScript regular expression, so eliminate it.
        : '');
    // > If the pattern ends with a slash,
    // > it is removed for the purpose of the following description,
    // > but it would only find a match with a directory.
    // > In other words, foo/ will match a directory foo and paths underneath it,
    // > but will not match a regular file or a symbolic link foo
    // >  (this is consistent with the way how pathspec works in general in Git).
    // '`foo/`' will not match regular file '`foo`' or symbolic link '`foo`'
    // -> ignore-rules will not deal with it, because it costs extra `fs.stat` call
    //      you could use option `mark: true` with `glob`
    // '`foo/`' should not continue with the '`..`'
    const DEFAULT_REPLACER_PREFIX = [
        // > Trailing spaces are ignored unless they are quoted with backslash ("\")
        [
            // (a\ ) -> (a )
            // (a  ) -> (a)
            // (a \ ) -> (a  )
            /\\?\s+$/,
            match => match.indexOf('\\') === 0
                ? ' '
                : ''
        ],
        // replace (\ ) with ' '
        [
            /\\\s/g,
            () => ' '
        ],
        // Escape metacharacters
        // which is written down by users but means special for regular expressions.
        // > There are 12 characters with special meanings:
        // > - the backslash \,
        // > - the caret ^,
        // > - the dollar sign $,
        // > - the period or dot .,
        // > - the vertical bar or pipe symbol |,
        // > - the question mark ?,
        // > - the asterisk or star *,
        // > - the plus sign +,
        // > - the opening parenthesis (,
        // > - the closing parenthesis ),
        // > - and the opening square bracket [,
        // > - the opening curly brace {,
        // > These special characters are often called "metacharacters".
        [
            /[\\^$.|*+(){]/g,
            match => `\\${match}`
        ],
        [
            // > [abc] matches any character inside the brackets
            // >    (in this case a, b, or c);
            /\[([^\]/]*)($|\])/g,
            (match, p1, p2) => p2 === ']'
                ? `[${sanitizeRange(p1)}]`
                : `\\${match}`
        ],
        [
            // > a question mark (?) matches a single character
            /(?!\\)\?/g,
            () => '[^/]'
        ],
        // leading slash
        [
            // > A leading slash matches the beginning of the pathname.
            // > For example, "/*.c" matches "cat-file.c" but not "mozilla-sha1/sha1.c".
            // A leading slash matches the beginning of the pathname
            /^\//,
            () => '^'
        ],
        // replace special metacharacter slash after the leading slash
        [
            /\//g,
            () => '\\/'
        ],
        [
            // > A leading "**" followed by a slash means match in all directories.
            // > For example, "**/foo" matches file or directory "foo" anywhere,
            // > the same as pattern "foo".
            // > "**/foo/bar" matches file or directory "bar" anywhere that is directly
            // >   under directory "foo".
            // Notice that the '*'s have been replaced as '\\*'
            /^\^*\\\*\\\*\\\//,
            // '**/foo' <-> 'foo'
            () => '^(?:.*\\/)?'
        ]
    ];
    const DEFAULT_REPLACER_SUFFIX = [
        // starting
        [
            // there will be no leading '/'
            //   (which has been replaced by section "leading slash")
            // If starts with '**', adding a '^' to the regular expression also works
            /^(?=[^^])/,
            function startingReplacer() {
                return !/\/(?!$)/.test(this)
                    // > If the pattern does not contain a slash /,
                    // >   Git treats it as a shell glob pattern
                    // Actually, if there is only a trailing slash,
                    //   git also treats it as a shell glob pattern
                    ? '(?:^|\\/)'
                    // > Otherwise, Git treats the pattern as a shell glob suitable for
                    // >   consumption by fnmatch(3)
                    : '^';
            }
        ],
        // two globstars
        [
            // Use lookahead assertions so that we could match more than one `'/**'`
            /\\\/\\\*\\\*(?=\\\/|$)/g,
            // Zero, one or several directories
            // should not use '*', or it will be replaced by the next replacer
            // Check if it is not the last `'/**'`
            (match, index, str) => index + 6 < str.length
                // case: /**/
                // > A slash followed by two consecutive asterisks then a slash matches
                // >   zero or more directories.
                // > For example, "a/**/b" matches "a/b", "a/x/b", "a/x/y/b" and so on.
                // '/**/'
                ? '(?:\\/[^\\/]+)*'
                // case: /**
                // > A trailing `"/**"` matches everything inside.
                // #21: everything inside but it should not include the current folder
                : '\\/.+'
        ],
        // intermediate wildcards
        [
            // Never replace escaped '*'
            // ignore rule '\*' will match the path '*'
            // 'abc.*/' -> go
            // 'abc.*'  -> skip this rule
            /(^|[^\\]+)\\\*(?=.+)/g,
            // '*.js' matches '.js'
            // '*.js' doesn't match 'abc'
            (match, p1) => `${p1}[^\\/]*`
        ],
        // trailing wildcard
        [
            /(\^|\\\/)?\\\*$/,
            (match, p1) => {
                const prefix = p1
                    // '\^':
                    // '/*' does not match ''
                    // '/*' does not match everything
                    // '\\\/':
                    // 'abc/*' does not match 'abc/'
                    ? `${p1}[^/]+`
                    // 'a*' matches 'a'
                    // 'a*' matches 'aa'
                    : '[^/]*';
                return `${prefix}(?=$|\\/$)`;
            }
        ],
        [
            // unescape
            /\\\\\\/g,
            () => '\\'
        ]
    ];
    const POSITIVE_REPLACERS = [
        ...DEFAULT_REPLACER_PREFIX,
        // 'f'
        // matches
        // - /f(end)
        // - /f/
        // - (start)f(end)
        // - (start)f/
        // doesn't match
        // - oof
        // - foo
        // pseudo:
        // -> (^|/)f(/|$)
        // ending
        [
            // 'js' will not match 'js.'
            // 'ab' will not match 'abc'
            /(?:[^*/])$/,
            // 'js*' will not match 'a.js'
            // 'js/' will not match 'a.js'
            // 'js' will match 'a.js' and 'a.js/'
            match => `${match}(?=$|\\/)`
        ],
        ...DEFAULT_REPLACER_SUFFIX
    ];
    const NEGATIVE_REPLACERS = [
        ...DEFAULT_REPLACER_PREFIX,
        // #24, #38
        // The MISSING rule of [gitignore docs](https://git-scm.com/docs/gitignore)
        // A negative pattern without a trailing wildcard should not
        // re-include the things inside that directory.
        // eg:
        // ['node_modules/*', '!node_modules']
        // should ignore `node_modules/a.js`
        [
            /(?:[^*])$/,
            match => `${match}(?=$|\\/$)`
        ],
        ...DEFAULT_REPLACER_SUFFIX
    ];
    // A simple cache, because an ignore rule only has only one certain meaning
    const cache = Object.create(null);
    // @param {pattern}
    const make_regex = (pattern, negative, ignorecase) => {
        const r = cache[pattern];
        if (r) {
            return r;
        }
        const replacers = negative
            ? NEGATIVE_REPLACERS
            : POSITIVE_REPLACERS;
        const source = replacers.reduce((prev, current) => prev.replace(current[0], current[1].bind(pattern)), pattern);
        return cache[pattern] = ignorecase
            ? new RegExp(source, 'i')
            : new RegExp(source);
    };
    // > A blank line matches no files, so it can serve as a separator for readability.
    const checkPattern = pattern => pattern
        && typeof pattern === 'string'
        && !REGEX_BLANK_LINE.test(pattern)
        // > A line starting with # serves as a comment.
        && pattern.indexOf('#') !== 0;
    const createRule = (pattern, ignorecase) => {
        const origin = pattern;
        let negative = false;
        // > An optional prefix "!" which negates the pattern;
        if (pattern.indexOf('!') === 0) {
            negative = true;
            pattern = pattern.substr(1);
        }
        pattern = pattern
            // > Put a backslash ("\") in front of the first "!" for patterns that
            // >   begin with a literal "!", for example, `"\!important!.txt"`.
            .replace(REGEX_LEADING_EXCAPED_EXCLAMATION, '!')
            // > Put a backslash ("\") in front of the first hash for patterns that
            // >   begin with a hash.
            .replace(REGEX_LEADING_EXCAPED_HASH, '#');
        const regex = make_regex(pattern, negative, ignorecase);
        return {
            origin,
            pattern,
            negative,
            regex
        };
    };
    class IgnoreBase {
        constructor({ ignorecase = true } = {}) {
            this._rules = [];
            this._ignorecase = ignorecase;
            define(this, KEY_IGNORE, true);
            this._initCache();
        }
        _initCache() {
            this._cache = Object.create(null);
        }
        // @param {Array.<string>|string|Ignore} pattern
        add(pattern) {
            this._added = false;
            if (typeof pattern === 'string') {
                pattern = pattern.split(/\r?\n/g);
            }
            make_array(pattern).forEach(this._addPattern, this);
            // Some rules have just added to the ignore,
            // making the behavior changed.
            if (this._added) {
                this._initCache();
            }
            return this;
        }
        // legacy
        addPattern(pattern) {
            return this.add(pattern);
        }
        _addPattern(pattern) {
            // #32
            if (pattern && pattern[KEY_IGNORE]) {
                this._rules = this._rules.concat(pattern._rules);
                this._added = true;
                return;
            }
            if (checkPattern(pattern)) {
                const rule = createRule(pattern, this._ignorecase);
                this._added = true;
                this._rules.push(rule);
            }
        }
        filter(paths) {
            return make_array(paths).filter(path => this._filter(path));
        }
        createFilter() {
            return path => this._filter(path);
        }
        ignores(path) {
            return !this._filter(path);
        }
        // @returns `Boolean` true if the `path` is NOT ignored
        _filter(path, slices) {
            if (!path) {
                return false;
            }
            if (path in this._cache) {
                return this._cache[path];
            }
            if (!slices) {
                // path/to/a.js
                // ['path', 'to', 'a.js']
                slices = path.split(SLASH);
            }
            slices.pop();
            return this._cache[path] = slices.length
                // > It is not possible to re-include a file if a parent directory of
                // >   that file is excluded.
                // If the path contains a parent directory, check the parent first
                ? this._filter(slices.join(SLASH) + SLASH, slices)
                    && this._test(path)
                // Or only test the path
                : this._test(path);
        }
        // @returns {Boolean} true if a file is NOT ignored
        _test(path) {
            // Explicitly define variable type by setting matched to `0`
            let matched = 0;
            this._rules.forEach(rule => {
                // if matched = true, then we only test negative rules
                // if matched = false, then we test non-negative rules
                if (!(matched ^ rule.negative)) {
                    matched = rule.negative ^ rule.regex.test(path);
                }
            });
            return !matched;
        }
    }
    // Windows
    // --------------------------------------------------------------
    /* istanbul ignore if  */
    if (
    // Detect `process` so that it can run in browsers.
    typeof process !== 'undefined'
        && (process.env && process.env.IGNORE_TEST_WIN32
            || process.platform === 'win32')) {
        const filter = IgnoreBase.prototype._filter;
        /* eslint no-control-regex: "off" */
        const make_posix = str => /^\\\\\?\\/.test(str)
            || /[^\x00-\x80]+/.test(str)
            ? str
            : str.replace(/\\/g, '/');
        IgnoreBase.prototype._filter = function filterWin32(path, slices) {
            path = make_posix(path);
            return filter.call(this, path, slices);
        };
    }
    module.exports = options => new IgnoreBase(options);
},
/* @eslint/eslintrc/lib/config-array/config-dependency.js */ function _(require, module, exports, __esModule, __esExport) {
    const util = require("util");
    /**
     * The class is to store parsers or plugins.
     * This class hides the loaded object from `JSON.stringify()` and `console.log`.
     * @template T
     */
    class ConfigDependency {
        /**
         * Initialize this instance.
         * @param {Object} data The dependency data.
         * @param {T} [data.definition] The dependency if the loading succeeded.
         * @param {Error} [data.error] The error object if the loading failed.
         * @param {string} [data.filePath] The actual path to the dependency if the loading succeeded.
         * @param {string} data.id The ID of this dependency.
         * @param {string} data.importerName The name of the config file which loads this dependency.
         * @param {string} data.importerPath The path to the config file which loads this dependency.
         */
        constructor({ definition = null, error = null, filePath = null, id, importerName, importerPath }) {
            /**
             * The loaded dependency if the loading succeeded.
             * @type {T|null}
             */
            this.definition = definition;
            /**
             * The error object if the loading failed.
             * @type {Error|null}
             */
            this.error = error;
            /**
             * The loaded dependency if the loading succeeded.
             * @type {string|null}
             */
            this.filePath = filePath;
            /**
             * The ID of this dependency.
             * @type {string}
             */
            this.id = id;
            /**
             * The name of the config file which loads this dependency.
             * @type {string}
             */
            this.importerName = importerName;
            /**
             * The path to the config file which loads this dependency.
             * @type {string}
             */
            this.importerPath = importerPath;
        }
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @returns {Object} a JSON compatible object.
         */
        toJSON() {
            const obj = this[util.inspect.custom]();
            // Display `error.message` (`Error#message` is unenumerable).
            if (obj.error instanceof Error) {
                obj.error = { ...obj.error, message: obj.error.message };
            }
            return obj;
        }
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @returns {Object} an object to display by `console.log()`.
         */
        [util.inspect.custom]() {
            const { definition: _ignore, // eslint-disable-line no-unused-vars
            ...obj } = this;
            return obj;
        }
    }
    /** @typedef {ConfigDependency<import("../../shared/types").Parser>} DependentParser */
    /** @typedef {ConfigDependency<import("../../shared/types").Plugin>} DependentPlugin */
    module.exports = { ConfigDependency };
},
/* @eslint/eslintrc/lib/config-array/override-tester.js */ function _(require, module, exports, __esModule, __esExport) {
    const assert = require("assert");
    const path = require("path");
    const util = require("util");
    const { Minimatch } = require(279) /* minimatch */;
    const minimatchOpts = { dot: true, matchBase: true };
    /**
     * @typedef {Object} Pattern
     * @property {InstanceType<Minimatch>[] | null} includes The positive matchers.
     * @property {InstanceType<Minimatch>[] | null} excludes The negative matchers.
     */
    /**
     * Normalize a given pattern to an array.
     * @param {string|string[]|undefined} patterns A glob pattern or an array of glob patterns.
     * @returns {string[]|null} Normalized patterns.
     * @private
     */
    function normalizePatterns(patterns) {
        if (Array.isArray(patterns)) {
            return patterns.filter(Boolean);
        }
        if (typeof patterns === "string" && patterns) {
            return [patterns];
        }
        return [];
    }
    /**
     * Create the matchers of given patterns.
     * @param {string[]} patterns The patterns.
     * @returns {InstanceType<Minimatch>[] | null} The matchers.
     */
    function toMatcher(patterns) {
        if (patterns.length === 0) {
            return null;
        }
        return patterns.map(pattern => {
            if (/^\.[/\\]/u.test(pattern)) {
                return new Minimatch(pattern.slice(2), 
                // `./*.js` should not match with `subdir/foo.js`
                { ...minimatchOpts, matchBase: false });
            }
            return new Minimatch(pattern, minimatchOpts);
        });
    }
    /**
     * Convert a given matcher to string.
     * @param {Pattern} matchers The matchers.
     * @returns {string} The string expression of the matcher.
     */
    function patternToJson({ includes, excludes }) {
        return {
            includes: includes && includes.map(m => m.pattern),
            excludes: excludes && excludes.map(m => m.pattern)
        };
    }
    /**
     * The class to test given paths are matched by the patterns.
     */
    class OverrideTester {
        /**
         * Create a tester with given criteria.
         * If there are no criteria, returns `null`.
         * @param {string|string[]} files The glob patterns for included files.
         * @param {string|string[]} excludedFiles The glob patterns for excluded files.
         * @param {string} basePath The path to the base directory to test paths.
         * @returns {OverrideTester|null} The created instance or `null`.
         */
        static create(files, excludedFiles, basePath) {
            const includePatterns = normalizePatterns(files);
            const excludePatterns = normalizePatterns(excludedFiles);
            let endsWithWildcard = false;
            if (includePatterns.length === 0) {
                return null;
            }
            // Rejects absolute paths or relative paths to parents.
            for (const pattern of includePatterns) {
                if (path.isAbsolute(pattern) || pattern.includes("..")) {
                    throw new Error(`Invalid override pattern (expected relative path not containing '..'): ${pattern}`);
                }
                if (pattern.endsWith("*")) {
                    endsWithWildcard = true;
                }
            }
            for (const pattern of excludePatterns) {
                if (path.isAbsolute(pattern) || pattern.includes("..")) {
                    throw new Error(`Invalid override pattern (expected relative path not containing '..'): ${pattern}`);
                }
            }
            const includes = toMatcher(includePatterns);
            const excludes = toMatcher(excludePatterns);
            return new OverrideTester([{ includes, excludes }], basePath, endsWithWildcard);
        }
        /**
         * Combine two testers by logical and.
         * If either of the testers was `null`, returns the other tester.
         * The `basePath` property of the two must be the same value.
         * @param {OverrideTester|null} a A tester.
         * @param {OverrideTester|null} b Another tester.
         * @returns {OverrideTester|null} Combined tester.
         */
        static and(a, b) {
            if (!b) {
                return a && new OverrideTester(a.patterns, a.basePath, a.endsWithWildcard);
            }
            if (!a) {
                return new OverrideTester(b.patterns, b.basePath, b.endsWithWildcard);
            }
            assert.strictEqual(a.basePath, b.basePath);
            return new OverrideTester(a.patterns.concat(b.patterns), a.basePath, a.endsWithWildcard || b.endsWithWildcard);
        }
        /**
         * Initialize this instance.
         * @param {Pattern[]} patterns The matchers.
         * @param {string} basePath The base path.
         * @param {boolean} endsWithWildcard If `true` then a pattern ends with `*`.
         */
        constructor(patterns, basePath, endsWithWildcard = false) {
            /** @type {Pattern[]} */
            this.patterns = patterns;
            /** @type {string} */
            this.basePath = basePath;
            /** @type {boolean} */
            this.endsWithWildcard = endsWithWildcard;
        }
        /**
         * Test if a given path is matched or not.
         * @param {string} filePath The absolute path to the target file.
         * @returns {boolean} `true` if the path was matched.
         */
        test(filePath) {
            if (typeof filePath !== "string" || !path.isAbsolute(filePath)) {
                throw new Error(`'filePath' should be an absolute path, but got ${filePath}.`);
            }
            const relativePath = path.relative(this.basePath, filePath);
            return this.patterns.every(({ includes, excludes }) => ((!includes || includes.some(m => m.match(relativePath))) &&
                (!excludes || !excludes.some(m => m.match(relativePath)))));
        }
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @returns {Object} a JSON compatible object.
         */
        toJSON() {
            if (this.patterns.length === 1) {
                return {
                    ...patternToJson(this.patterns[0]),
                    basePath: this.basePath
                };
            }
            return {
                AND: this.patterns.map(patternToJson),
                basePath: this.basePath
            };
        }
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @returns {Object} an object to display by `console.log()`.
         */
        [util.inspect.custom]() {
            return this.toJSON();
        }
    }
    module.exports = { OverrideTester };
},
/* minimatch/minimatch.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = minimatch;
    minimatch.Minimatch = Minimatch;
    var path = { sep: '/' };
    try {
        path = require('path');
    }
    catch (er) { }
    var GLOBSTAR = minimatch.GLOBSTAR = Minimatch.GLOBSTAR = {};
    var expand = require(280) /* brace-expansion */;
    var plTypes = {
        '!': { open: '(?:(?!(?:', close: '))[^/]*?)' },
        '?': { open: '(?:', close: ')?' },
        '+': { open: '(?:', close: ')+' },
        '*': { open: '(?:', close: ')*' },
        '@': { open: '(?:', close: ')' }
    };
    // any single thing other than /
    // don't need to escape / when using new RegExp()
    var qmark = '[^/]';
    // * => any number of characters
    var star = qmark + '*?';
    // ** when dots are allowed.  Anything goes, except .. and .
    // not (^ or / followed by one or two dots followed by $ or /),
    // followed by anything, any number of times.
    var twoStarDot = '(?:(?!(?:\\\/|^)(?:\\.{1,2})($|\\\/)).)*?';
    // not a ^ or / followed by a dot,
    // followed by anything, any number of times.
    var twoStarNoDot = '(?:(?!(?:\\\/|^)\\.).)*?';
    // characters that need to be escaped in RegExp.
    var reSpecials = charSet('().*{}+?[]^$\\!');
    // "abc" -> { a:true, b:true, c:true }
    function charSet(s) {
        return s.split('').reduce(function (set, c) {
            set[c] = true;
            return set;
        }, {});
    }
    // normalizes slashes.
    var slashSplit = /\/+/;
    minimatch.filter = filter;
    function filter(pattern, options) {
        options = options || {};
        return function (p, i, list) {
            return minimatch(p, pattern, options);
        };
    }
    function ext(a, b) {
        a = a || {};
        b = b || {};
        var t = {};
        Object.keys(b).forEach(function (k) {
            t[k] = b[k];
        });
        Object.keys(a).forEach(function (k) {
            t[k] = a[k];
        });
        return t;
    }
    minimatch.defaults = function (def) {
        if (!def || !Object.keys(def).length)
            return minimatch;
        var orig = minimatch;
        var m = function minimatch(p, pattern, options) {
            return orig.minimatch(p, pattern, ext(def, options));
        };
        m.Minimatch = function Minimatch(pattern, options) {
            return new orig.Minimatch(pattern, ext(def, options));
        };
        return m;
    };
    Minimatch.defaults = function (def) {
        if (!def || !Object.keys(def).length)
            return Minimatch;
        return minimatch.defaults(def).Minimatch;
    };
    function minimatch(p, pattern, options) {
        if (typeof pattern !== 'string') {
            throw new TypeError('glob pattern string required');
        }
        if (!options)
            options = {};
        // shortcut: comments match nothing.
        if (!options.nocomment && pattern.charAt(0) === '#') {
            return false;
        }
        // "" only matches ""
        if (pattern.trim() === '')
            return p === '';
        return new Minimatch(pattern, options).match(p);
    }
    function Minimatch(pattern, options) {
        if (!(this instanceof Minimatch)) {
            return new Minimatch(pattern, options);
        }
        if (typeof pattern !== 'string') {
            throw new TypeError('glob pattern string required');
        }
        if (!options)
            options = {};
        pattern = pattern.trim();
        // windows support: need to use /, not \
        if (path.sep !== '/') {
            pattern = pattern.split(path.sep).join('/');
        }
        this.options = options;
        this.set = [];
        this.pattern = pattern;
        this.regexp = null;
        this.negate = false;
        this.comment = false;
        this.empty = false;
        // make the set of regexps etc.
        this.make();
    }
    Minimatch.prototype.debug = function () { };
    Minimatch.prototype.make = make;
    function make() {
        // don't do it more than once.
        if (this._made)
            return;
        var pattern = this.pattern;
        var options = this.options;
        // empty patterns and comments match nothing.
        if (!options.nocomment && pattern.charAt(0) === '#') {
            this.comment = true;
            return;
        }
        if (!pattern) {
            this.empty = true;
            return;
        }
        // step 1: figure out negation, etc.
        this.parseNegate();
        // step 2: expand braces
        var set = this.globSet = this.braceExpand();
        if (options.debug)
            this.debug = console.error;
        this.debug(this.pattern, set);
        // step 3: now we have a set, so turn each one into a series of path-portion
        // matching patterns.
        // These will be regexps, except in the case of "**", which is
        // set to the GLOBSTAR object for globstar behavior,
        // and will not contain any / characters
        set = this.globParts = set.map(function (s) {
            return s.split(slashSplit);
        });
        this.debug(this.pattern, set);
        // glob --> regexps
        set = set.map(function (s, si, set) {
            return s.map(this.parse, this);
        }, this);
        this.debug(this.pattern, set);
        // filter out everything that didn't compile properly.
        set = set.filter(function (s) {
            return s.indexOf(false) === -1;
        });
        this.debug(this.pattern, set);
        this.set = set;
    }
    Minimatch.prototype.parseNegate = parseNegate;
    function parseNegate() {
        var pattern = this.pattern;
        var negate = false;
        var options = this.options;
        var negateOffset = 0;
        if (options.nonegate)
            return;
        for (var i = 0, l = pattern.length; i < l && pattern.charAt(i) === '!'; i++) {
            negate = !negate;
            negateOffset++;
        }
        if (negateOffset)
            this.pattern = pattern.substr(negateOffset);
        this.negate = negate;
    }
    // Brace expansion:
    // a{b,c}d -> abd acd
    // a{b,}c -> abc ac
    // a{0..3}d -> a0d a1d a2d a3d
    // a{b,c{d,e}f}g -> abg acdfg acefg
    // a{b,c}d{e,f}g -> abdeg acdeg abdeg abdfg
    //
    // Invalid sets are not expanded.
    // a{2..}b -> a{2..}b
    // a{b}c -> a{b}c
    minimatch.braceExpand = function (pattern, options) {
        return braceExpand(pattern, options);
    };
    Minimatch.prototype.braceExpand = braceExpand;
    function braceExpand(pattern, options) {
        if (!options) {
            if (this instanceof Minimatch) {
                options = this.options;
            }
            else {
                options = {};
            }
        }
        pattern = typeof pattern === 'undefined'
            ? this.pattern : pattern;
        if (typeof pattern === 'undefined') {
            throw new TypeError('undefined pattern');
        }
        if (options.nobrace ||
            !pattern.match(/\{.*\}/)) {
            // shortcut. no need to expand.
            return [pattern];
        }
        return expand(pattern);
    }
    // parse a component of the expanded set.
    // At this point, no pattern may contain "/" in it
    // so we're going to return a 2d array, where each entry is the full
    // pattern, split on '/', and then turned into a regular expression.
    // A regexp is made at the end which joins each array with an
    // escaped /, and another full one which joins each regexp with |.
    //
    // Following the lead of Bash 4.1, note that "**" only has special meaning
    // when it is the *only* thing in a path portion.  Otherwise, any series
    // of * is equivalent to a single *.  Globstar behavior is enabled by
    // default, and can be disabled by setting options.noglobstar.
    Minimatch.prototype.parse = parse;
    var SUBPARSE = {};
    function parse(pattern, isSub) {
        if (pattern.length > 1024 * 64) {
            throw new TypeError('pattern is too long');
        }
        var options = this.options;
        // shortcuts
        if (!options.noglobstar && pattern === '**')
            return GLOBSTAR;
        if (pattern === '')
            return '';
        var re = '';
        var hasMagic = !!options.nocase;
        var escaping = false;
        // ? => one single character
        var patternListStack = [];
        var negativeLists = [];
        var stateChar;
        var inClass = false;
        var reClassStart = -1;
        var classStart = -1;
        // . and .. never match anything that doesn't start with .,
        // even when options.dot is set.
        var patternStart = pattern.charAt(0) === '.' ? '' // anything
            // not (start or / followed by . or .. followed by / or end)
            : options.dot ? '(?!(?:^|\\\/)\\.{1,2}(?:$|\\\/))'
                : '(?!\\.)';
        var self = this;
        function clearStateChar() {
            if (stateChar) {
                // we had some state-tracking character
                // that wasn't consumed by this pass.
                switch (stateChar) {
                    case '*':
                        re += star;
                        hasMagic = true;
                        break;
                    case '?':
                        re += qmark;
                        hasMagic = true;
                        break;
                    default:
                        re += '\\' + stateChar;
                        break;
                }
                self.debug('clearStateChar %j %j', stateChar, re);
                stateChar = false;
            }
        }
        for (var i = 0, len = pattern.length, c; (i < len) && (c = pattern.charAt(i)); i++) {
            this.debug('%s\t%s %s %j', pattern, i, re, c);
            // skip over any that are escaped.
            if (escaping && reSpecials[c]) {
                re += '\\' + c;
                escaping = false;
                continue;
            }
            switch (c) {
                case '/':
                    // completely not allowed, even escaped.
                    // Should already be path-split by now.
                    return false;
                case '\\':
                    clearStateChar();
                    escaping = true;
                    continue;
                // the various stateChar values
                // for the "extglob" stuff.
                case '?':
                case '*':
                case '+':
                case '@':
                case '!':
                    this.debug('%s\t%s %s %j <-- stateChar', pattern, i, re, c);
                    // all of those are literals inside a class, except that
                    // the glob [!a] means [^a] in regexp
                    if (inClass) {
                        this.debug('  in class');
                        if (c === '!' && i === classStart + 1)
                            c = '^';
                        re += c;
                        continue;
                    }
                    // if we already have a stateChar, then it means
                    // that there was something like ** or +? in there.
                    // Handle the stateChar, then proceed with this one.
                    self.debug('call clearStateChar %j', stateChar);
                    clearStateChar();
                    stateChar = c;
                    // if extglob is disabled, then +(asdf|foo) isn't a thing.
                    // just clear the statechar *now*, rather than even diving into
                    // the patternList stuff.
                    if (options.noext)
                        clearStateChar();
                    continue;
                case '(':
                    if (inClass) {
                        re += '(';
                        continue;
                    }
                    if (!stateChar) {
                        re += '\\(';
                        continue;
                    }
                    patternListStack.push({
                        type: stateChar,
                        start: i - 1,
                        reStart: re.length,
                        open: plTypes[stateChar].open,
                        close: plTypes[stateChar].close
                    });
                    // negation is (?:(?!js)[^/]*)
                    re += stateChar === '!' ? '(?:(?!(?:' : '(?:';
                    this.debug('plType %j %j', stateChar, re);
                    stateChar = false;
                    continue;
                case ')':
                    if (inClass || !patternListStack.length) {
                        re += '\\)';
                        continue;
                    }
                    clearStateChar();
                    hasMagic = true;
                    var pl = patternListStack.pop();
                    // negation is (?:(?!js)[^/]*)
                    // The others are (?:<pattern>)<type>
                    re += pl.close;
                    if (pl.type === '!') {
                        negativeLists.push(pl);
                    }
                    pl.reEnd = re.length;
                    continue;
                case '|':
                    if (inClass || !patternListStack.length || escaping) {
                        re += '\\|';
                        escaping = false;
                        continue;
                    }
                    clearStateChar();
                    re += '|';
                    continue;
                // these are mostly the same in regexp and glob
                case '[':
                    // swallow any state-tracking char before the [
                    clearStateChar();
                    if (inClass) {
                        re += '\\' + c;
                        continue;
                    }
                    inClass = true;
                    classStart = i;
                    reClassStart = re.length;
                    re += c;
                    continue;
                case ']':
                    //  a right bracket shall lose its special
                    //  meaning and represent itself in
                    //  a bracket expression if it occurs
                    //  first in the list.  -- POSIX.2 2.8.3.2
                    if (i === classStart + 1 || !inClass) {
                        re += '\\' + c;
                        escaping = false;
                        continue;
                    }
                    // handle the case where we left a class open.
                    // "[z-a]" is valid, equivalent to "\[z-a\]"
                    if (inClass) {
                        // split where the last [ was, make sure we don't have
                        // an invalid re. if so, re-walk the contents of the
                        // would-be class to re-translate any characters that
                        // were passed through as-is
                        // TODO: It would probably be faster to determine this
                        // without a try/catch and a new RegExp, but it's tricky
                        // to do safely.  For now, this is safe and works.
                        var cs = pattern.substring(classStart + 1, i);
                        try {
                            RegExp('[' + cs + ']');
                        }
                        catch (er) {
                            // not a valid class!
                            var sp = this.parse(cs, SUBPARSE);
                            re = re.substr(0, reClassStart) + '\\[' + sp[0] + '\\]';
                            hasMagic = hasMagic || sp[1];
                            inClass = false;
                            continue;
                        }
                    }
                    // finish up the class.
                    hasMagic = true;
                    inClass = false;
                    re += c;
                    continue;
                default:
                    // swallow any state char that wasn't consumed
                    clearStateChar();
                    if (escaping) {
                        // no need
                        escaping = false;
                    }
                    else if (reSpecials[c]
                        && !(c === '^' && inClass)) {
                        re += '\\';
                    }
                    re += c;
            } // switch
        } // for
        // handle the case where we left a class open.
        // "[abc" is valid, equivalent to "\[abc"
        if (inClass) {
            // split where the last [ was, and escape it
            // this is a huge pita.  We now have to re-walk
            // the contents of the would-be class to re-translate
            // any characters that were passed through as-is
            cs = pattern.substr(classStart + 1);
            sp = this.parse(cs, SUBPARSE);
            re = re.substr(0, reClassStart) + '\\[' + sp[0];
            hasMagic = hasMagic || sp[1];
        }
        // handle the case where we had a +( thing at the *end*
        // of the pattern.
        // each pattern list stack adds 3 chars, and we need to go through
        // and escape any | chars that were passed through as-is for the regexp.
        // Go through and escape them, taking care not to double-escape any
        // | chars that were already escaped.
        for (pl = patternListStack.pop(); pl; pl = patternListStack.pop()) {
            var tail = re.slice(pl.reStart + pl.open.length);
            this.debug('setting tail', re, pl);
            // maybe some even number of \, then maybe 1 \, followed by a |
            tail = tail.replace(/((?:\\{2}){0,64})(\\?)\|/g, function (_, $1, $2) {
                if (!$2) {
                    // the | isn't already escaped, so escape it.
                    $2 = '\\';
                }
                // need to escape all those slashes *again*, without escaping the
                // one that we need for escaping the | character.  As it works out,
                // escaping an even number of slashes can be done by simply repeating
                // it exactly after itself.  That's why this trick works.
                //
                // I am sorry that you have to see this.
                return $1 + $1 + $2 + '|';
            });
            this.debug('tail=%j\n   %s', tail, tail, pl, re);
            var t = pl.type === '*' ? star
                : pl.type === '?' ? qmark
                    : '\\' + pl.type;
            hasMagic = true;
            re = re.slice(0, pl.reStart) + t + '\\(' + tail;
        }
        // handle trailing things that only matter at the very end.
        clearStateChar();
        if (escaping) {
            // trailing \\
            re += '\\\\';
        }
        // only need to apply the nodot start if the re starts with
        // something that could conceivably capture a dot
        var addPatternStart = false;
        switch (re.charAt(0)) {
            case '.':
            case '[':
            case '(': addPatternStart = true;
        }
        // Hack to work around lack of negative lookbehind in JS
        // A pattern like: *.!(x).!(y|z) needs to ensure that a name
        // like 'a.xyz.yz' doesn't match.  So, the first negative
        // lookahead, has to look ALL the way ahead, to the end of
        // the pattern.
        for (var n = negativeLists.length - 1; n > -1; n--) {
            var nl = negativeLists[n];
            var nlBefore = re.slice(0, nl.reStart);
            var nlFirst = re.slice(nl.reStart, nl.reEnd - 8);
            var nlLast = re.slice(nl.reEnd - 8, nl.reEnd);
            var nlAfter = re.slice(nl.reEnd);
            nlLast += nlAfter;
            // Handle nested stuff like *(*.js|!(*.json)), where open parens
            // mean that we should *not* include the ) in the bit that is considered
            // "after" the negated section.
            var openParensBefore = nlBefore.split('(').length - 1;
            var cleanAfter = nlAfter;
            for (i = 0; i < openParensBefore; i++) {
                cleanAfter = cleanAfter.replace(/\)[+*?]?/, '');
            }
            nlAfter = cleanAfter;
            var dollar = '';
            if (nlAfter === '' && isSub !== SUBPARSE) {
                dollar = '$';
            }
            var newRe = nlBefore + nlFirst + nlAfter + dollar + nlLast;
            re = newRe;
        }
        // if the re is not "" at this point, then we need to make sure
        // it doesn't match against an empty path part.
        // Otherwise a/* will match a/, which it should not.
        if (re !== '' && hasMagic) {
            re = '(?=.)' + re;
        }
        if (addPatternStart) {
            re = patternStart + re;
        }
        // parsing just a piece of a larger pattern.
        if (isSub === SUBPARSE) {
            return [re, hasMagic];
        }
        // skip the regexp for non-magical patterns
        // unescape anything in it, though, so that it'll be
        // an exact match against a file etc.
        if (!hasMagic) {
            return globUnescape(pattern);
        }
        var flags = options.nocase ? 'i' : '';
        try {
            var regExp = new RegExp('^' + re + '$', flags);
        }
        catch (er) {
            // If it was an invalid regular expression, then it can't match
            // anything.  This trick looks for a character after the end of
            // the string, which is of course impossible, except in multi-line
            // mode, but it's not a /m regex.
            return new RegExp('$.');
        }
        regExp._glob = pattern;
        regExp._src = re;
        return regExp;
    }
    minimatch.makeRe = function (pattern, options) {
        return new Minimatch(pattern, options || {}).makeRe();
    };
    Minimatch.prototype.makeRe = makeRe;
    function makeRe() {
        if (this.regexp || this.regexp === false)
            return this.regexp;
        // at this point, this.set is a 2d array of partial
        // pattern strings, or "**".
        //
        // It's better to use .match().  This function shouldn't
        // be used, really, but it's pretty convenient sometimes,
        // when you just want to work with a regex.
        var set = this.set;
        if (!set.length) {
            this.regexp = false;
            return this.regexp;
        }
        var options = this.options;
        var twoStar = options.noglobstar ? star
            : options.dot ? twoStarDot
                : twoStarNoDot;
        var flags = options.nocase ? 'i' : '';
        var re = set.map(function (pattern) {
            return pattern.map(function (p) {
                return (p === GLOBSTAR) ? twoStar
                    : (typeof p === 'string') ? regExpEscape(p)
                        : p._src;
            }).join('\\\/');
        }).join('|');
        // must match entire pattern
        // ending in a * or ** will make it less strict.
        re = '^(?:' + re + ')$';
        // can match anything, as long as it's not this.
        if (this.negate)
            re = '^(?!' + re + ').*$';
        try {
            this.regexp = new RegExp(re, flags);
        }
        catch (ex) {
            this.regexp = false;
        }
        return this.regexp;
    }
    minimatch.match = function (list, pattern, options) {
        options = options || {};
        var mm = new Minimatch(pattern, options);
        list = list.filter(function (f) {
            return mm.match(f);
        });
        if (mm.options.nonull && !list.length) {
            list.push(pattern);
        }
        return list;
    };
    Minimatch.prototype.match = match;
    function match(f, partial) {
        this.debug('match', f, this.pattern);
        // short-circuit in the case of busted things.
        // comments, etc.
        if (this.comment)
            return false;
        if (this.empty)
            return f === '';
        if (f === '/' && partial)
            return true;
        var options = this.options;
        // windows: need to use /, not \
        if (path.sep !== '/') {
            f = f.split(path.sep).join('/');
        }
        // treat the test path as a set of pathparts.
        f = f.split(slashSplit);
        this.debug(this.pattern, 'split', f);
        // just ONE of the pattern sets in this.set needs to match
        // in order for it to be valid.  If negating, then just one
        // match means that we have failed.
        // Either way, return on the first hit.
        var set = this.set;
        this.debug(this.pattern, 'set', set);
        // Find the basename of the path by looking for the last non-empty segment
        var filename;
        var i;
        for (i = f.length - 1; i >= 0; i--) {
            filename = f[i];
            if (filename)
                break;
        }
        for (i = 0; i < set.length; i++) {
            var pattern = set[i];
            var file = f;
            if (options.matchBase && pattern.length === 1) {
                file = [filename];
            }
            var hit = this.matchOne(file, pattern, partial);
            if (hit) {
                if (options.flipNegate)
                    return true;
                return !this.negate;
            }
        }
        // didn't get any hits.  this is success if it's a negative
        // pattern, failure otherwise.
        if (options.flipNegate)
            return false;
        return this.negate;
    }
    // set partial to true to test if, for example,
    // "/a/b" matches the start of "/*/b/*/d"
    // Partial means, if you run out of file before you run
    // out of pattern, then that's fine, as long as all
    // the parts match.
    Minimatch.prototype.matchOne = function (file, pattern, partial) {
        var options = this.options;
        this.debug('matchOne', { 'this': this, file: file, pattern: pattern });
        this.debug('matchOne', file.length, pattern.length);
        for (var fi = 0, pi = 0, fl = file.length, pl = pattern.length; (fi < fl) && (pi < pl); fi++, pi++) {
            this.debug('matchOne loop');
            var p = pattern[pi];
            var f = file[fi];
            this.debug(pattern, p, f);
            // should be impossible.
            // some invalid regexp stuff in the set.
            if (p === false)
                return false;
            if (p === GLOBSTAR) {
                this.debug('GLOBSTAR', [pattern, p, f]);
                // "**"
                // a/**/b/**/c would match the following:
                // a/b/x/y/z/c
                // a/x/y/z/b/c
                // a/b/x/b/x/c
                // a/b/c
                // To do this, take the rest of the pattern after
                // the **, and see if it would match the file remainder.
                // If so, return success.
                // If not, the ** "swallows" a segment, and try again.
                // This is recursively awful.
                //
                // a/**/b/**/c matching a/b/x/y/z/c
                // - a matches a
                // - doublestar
                //   - matchOne(b/x/y/z/c, b/**/c)
                //     - b matches b
                //     - doublestar
                //       - matchOne(x/y/z/c, c) -> no
                //       - matchOne(y/z/c, c) -> no
                //       - matchOne(z/c, c) -> no
                //       - matchOne(c, c) yes, hit
                var fr = fi;
                var pr = pi + 1;
                if (pr === pl) {
                    this.debug('** at the end');
                    // a ** at the end will just swallow the rest.
                    // We have found a match.
                    // however, it will not swallow /.x, unless
                    // options.dot is set.
                    // . and .. are *never* matched by **, for explosively
                    // exponential reasons.
                    for (; fi < fl; fi++) {
                        if (file[fi] === '.' || file[fi] === '..' ||
                            (!options.dot && file[fi].charAt(0) === '.'))
                            return false;
                    }
                    return true;
                }
                // ok, let's see if we can swallow whatever we can.
                while (fr < fl) {
                    var swallowee = file[fr];
                    this.debug('\nglobstar while', file, fr, pattern, pr, swallowee);
                    // XXX remove this slice.  Just pass the start index.
                    if (this.matchOne(file.slice(fr), pattern.slice(pr), partial)) {
                        this.debug('globstar found match!', fr, fl, swallowee);
                        // found a match.
                        return true;
                    }
                    else {
                        // can't swallow "." or ".." ever.
                        // can only swallow ".foo" when explicitly asked.
                        if (swallowee === '.' || swallowee === '..' ||
                            (!options.dot && swallowee.charAt(0) === '.')) {
                            this.debug('dot detected!', file, fr, pattern, pr);
                            break;
                        }
                        // ** swallows a segment, and continue.
                        this.debug('globstar swallow a segment, and continue');
                        fr++;
                    }
                }
                // no match was found.
                // However, in partial mode, we can't say this is necessarily over.
                // If there's more *pattern* left, then
                if (partial) {
                    // ran out of file
                    this.debug('\n>>> no match, partial?', file, fr, pattern, pr);
                    if (fr === fl)
                        return true;
                }
                return false;
            }
            // something other than **
            // non-magic patterns just have to match exactly
            // patterns with magic have been turned into regexps.
            var hit;
            if (typeof p === 'string') {
                if (options.nocase) {
                    hit = f.toLowerCase() === p.toLowerCase();
                }
                else {
                    hit = f === p;
                }
                this.debug('string match', p, f, hit);
            }
            else {
                hit = f.match(p);
                this.debug('pattern match', p, f, hit);
            }
            if (!hit)
                return false;
        }
        // Note: ending in / means that we'll get a final ""
        // at the end of the pattern.  This can only match a
        // corresponding "" at the end of the file.
        // If the file ends in /, then it can only match a
        // a pattern that ends in /, unless the pattern just
        // doesn't have any more for it. But, a/b/ should *not*
        // match "a/b/*", even though "" matches against the
        // [^/]*? pattern, except in partial mode, where it might
        // simply not be reached yet.
        // However, a/b/ should still satisfy a/*
        // now either we fell off the end of the pattern, or we're done.
        if (fi === fl && pi === pl) {
            // ran out of pattern and filename at the same time.
            // an exact hit!
            return true;
        }
        else if (fi === fl) {
            // ran out of file, but still had pattern left.
            // this is ok if we're doing the match as part of
            // a glob fs traversal.
            return partial;
        }
        else if (pi === pl) {
            // ran out of pattern, still have file left.
            // this is only acceptable if we're on the very last
            // empty segment of a file with a trailing slash.
            // a/* should match a/b/
            var emptyFileEnd = (fi === fl - 1) && (file[fi] === '');
            return emptyFileEnd;
        }
        // should be unreachable.
        throw new Error('wtf?');
    };
    // replace stuff like \* with *
    function globUnescape(s) {
        return s.replace(/\\(.)/g, '$1');
    }
    function regExpEscape(s) {
        return s.replace(/[-[\]{}()*+?.,\\^$|#\s]/g, '\\$&');
    }
},
/* brace-expansion/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var concatMap = require(281) /* concat-map */;
    var balanced = require(282) /* balanced-match */;
    module.exports = expandTop;
    var escSlash = '\0SLASH' + Math.random() + '\0';
    var escOpen = '\0OPEN' + Math.random() + '\0';
    var escClose = '\0CLOSE' + Math.random() + '\0';
    var escComma = '\0COMMA' + Math.random() + '\0';
    var escPeriod = '\0PERIOD' + Math.random() + '\0';
    function numeric(str) {
        return parseInt(str, 10) == str
            ? parseInt(str, 10)
            : str.charCodeAt(0);
    }
    function escapeBraces(str) {
        return str.split('\\\\').join(escSlash)
            .split('\\{').join(escOpen)
            .split('\\}').join(escClose)
            .split('\\,').join(escComma)
            .split('\\.').join(escPeriod);
    }
    function unescapeBraces(str) {
        return str.split(escSlash).join('\\')
            .split(escOpen).join('{')
            .split(escClose).join('}')
            .split(escComma).join(',')
            .split(escPeriod).join('.');
    }
    // Basically just str.split(","), but handling cases
    // where we have nested braced sections, which should be
    // treated as individual members, like {a,{b,c},d}
    function parseCommaParts(str) {
        if (!str)
            return [''];
        var parts = [];
        var m = balanced('{', '}', str);
        if (!m)
            return str.split(',');
        var pre = m.pre;
        var body = m.body;
        var post = m.post;
        var p = pre.split(',');
        p[p.length - 1] += '{' + body + '}';
        var postParts = parseCommaParts(post);
        if (post.length) {
            p[p.length - 1] += postParts.shift();
            p.push.apply(p, postParts);
        }
        parts.push.apply(parts, p);
        return parts;
    }
    function expandTop(str) {
        if (!str)
            return [];
        // I don't know why Bash 4.3 does this, but it does.
        // Anything starting with {} will have the first two bytes preserved
        // but *only* at the top level, so {},a}b will not expand to anything,
        // but a{},b}c will be expanded to [a}c,abc].
        // One could argue that this is a bug in Bash, but since the goal of
        // this module is to match Bash's rules, we escape a leading {}
        if (str.substr(0, 2) === '{}') {
            str = '\\{\\}' + str.substr(2);
        }
        return expand(escapeBraces(str), true).map(unescapeBraces);
    }
    function identity(e) {
        return e;
    }
    function embrace(str) {
        return '{' + str + '}';
    }
    function isPadded(el) {
        return /^-?0\d/.test(el);
    }
    function lte(i, y) {
        return i <= y;
    }
    function gte(i, y) {
        return i >= y;
    }
    function expand(str, isTop) {
        var expansions = [];
        var m = balanced('{', '}', str);
        if (!m || /\$$/.test(m.pre))
            return [str];
        var isNumericSequence = /^-?\d+\.\.-?\d+(?:\.\.-?\d+)?$/.test(m.body);
        var isAlphaSequence = /^[a-zA-Z]\.\.[a-zA-Z](?:\.\.-?\d+)?$/.test(m.body);
        var isSequence = isNumericSequence || isAlphaSequence;
        var isOptions = m.body.indexOf(',') >= 0;
        if (!isSequence && !isOptions) {
            // {a},b}
            if (m.post.match(/,.*\}/)) {
                str = m.pre + '{' + m.body + escClose + m.post;
                return expand(str);
            }
            return [str];
        }
        var n;
        if (isSequence) {
            n = m.body.split(/\.\./);
        }
        else {
            n = parseCommaParts(m.body);
            if (n.length === 1) {
                // x{{a,b}}y ==> x{a}y x{b}y
                n = expand(n[0], false).map(embrace);
                if (n.length === 1) {
                    var post = m.post.length
                        ? expand(m.post, false)
                        : [''];
                    return post.map(function (p) {
                        return m.pre + n[0] + p;
                    });
                }
            }
        }
        // at this point, n is the parts, and we know it's not a comma set
        // with a single entry.
        // no need to expand pre, since it is guaranteed to be free of brace-sets
        var pre = m.pre;
        var post = m.post.length
            ? expand(m.post, false)
            : [''];
        var N;
        if (isSequence) {
            var x = numeric(n[0]);
            var y = numeric(n[1]);
            var width = Math.max(n[0].length, n[1].length);
            var incr = n.length == 3
                ? Math.abs(numeric(n[2]))
                : 1;
            var test = lte;
            var reverse = y < x;
            if (reverse) {
                incr *= -1;
                test = gte;
            }
            var pad = n.some(isPadded);
            N = [];
            for (var i = x; test(i, y); i += incr) {
                var c;
                if (isAlphaSequence) {
                    c = String.fromCharCode(i);
                    if (c === '\\')
                        c = '';
                }
                else {
                    c = String(i);
                    if (pad) {
                        var need = width - c.length;
                        if (need > 0) {
                            var z = new Array(need + 1).join('0');
                            if (i < 0)
                                c = '-' + z + c.slice(1);
                            else
                                c = z + c;
                        }
                    }
                }
                N.push(c);
            }
        }
        else {
            N = concatMap(n, function (el) { return expand(el, false); });
        }
        for (var j = 0; j < N.length; j++) {
            for (var k = 0; k < post.length; k++) {
                var expansion = pre + N[j] + post[k];
                if (!isTop || isSequence || expansion)
                    expansions.push(expansion);
            }
        }
        return expansions;
    }
},
/* concat-map/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function (xs, fn) {
        var res = [];
        for (var i = 0; i < xs.length; i++) {
            var x = fn(xs[i], i);
            if (isArray(x))
                res.push.apply(res, x);
            else
                res.push(x);
        }
        return res;
    };
    var isArray = Array.isArray || function (xs) {
        return Object.prototype.toString.call(xs) === '[object Array]';
    };
},
/* balanced-match/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = balanced;
    function balanced(a, b, str) {
        if (a instanceof RegExp)
            a = maybeMatch(a, str);
        if (b instanceof RegExp)
            b = maybeMatch(b, str);
        var r = range(a, b, str);
        return r && {
            start: r[0],
            end: r[1],
            pre: str.slice(0, r[0]),
            body: str.slice(r[0] + a.length, r[1]),
            post: str.slice(r[1] + b.length)
        };
    }
    function maybeMatch(reg, str) {
        var m = str.match(reg);
        return m ? m[0] : null;
    }
    balanced.range = range;
    function range(a, b, str) {
        var begs, beg, left, right, result;
        var ai = str.indexOf(a);
        var bi = str.indexOf(b, ai + 1);
        var i = ai;
        if (ai >= 0 && bi > 0) {
            if (a === b) {
                return [ai, bi];
            }
            begs = [];
            left = str.length;
            while (i >= 0 && !result) {
                if (i == ai) {
                    begs.push(i);
                    ai = str.indexOf(a, i + 1);
                }
                else if (begs.length == 1) {
                    result = [begs.pop(), bi];
                }
                else {
                    beg = begs.pop();
                    if (beg < left) {
                        left = beg;
                        right = bi;
                    }
                    bi = str.indexOf(b, i + 1);
                }
                i = ai < bi && ai >= 0 ? ai : bi;
            }
            if (begs.length) {
                result = [left, right];
            }
        }
        return result;
    }
},
/* js-yaml/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var yaml = require(284) /* ./lib/js-yaml.js */;
    module.exports = yaml;
},
/* js-yaml/lib/js-yaml.js */ function _(require, module, exports, __esModule, __esExport) {
    var loader = require(285) /* ./js-yaml/loader */;
    var dumper = require(312) /* ./js-yaml/dumper */;
    function deprecated(name) {
        return function () {
            throw new Error('Function ' + name + ' is deprecated and cannot be used.');
        };
    }
    module.exports.Type = require(291) /* ./js-yaml/type */;
    module.exports.Schema = require(290) /* ./js-yaml/schema */;
    module.exports.FAILSAFE_SCHEMA = require(294) /* ./js-yaml/schema/failsafe */;
    module.exports.JSON_SCHEMA = require(293) /* ./js-yaml/schema/json */;
    module.exports.CORE_SCHEMA = require(292) /* ./js-yaml/schema/core */;
    module.exports.DEFAULT_SAFE_SCHEMA = require(289) /* ./js-yaml/schema/default_safe */;
    module.exports.DEFAULT_FULL_SCHEMA = require(308) /* ./js-yaml/schema/default_full */;
    module.exports.load = loader.load;
    module.exports.loadAll = loader.loadAll;
    module.exports.safeLoad = loader.safeLoad;
    module.exports.safeLoadAll = loader.safeLoadAll;
    module.exports.dump = dumper.dump;
    module.exports.safeDump = dumper.safeDump;
    module.exports.YAMLException = require(287) /* ./js-yaml/exception */;
    // Deprecated schema names from JS-YAML 2.0.x
    module.exports.MINIMAL_SCHEMA = require(294) /* ./js-yaml/schema/failsafe */;
    module.exports.SAFE_SCHEMA = require(289) /* ./js-yaml/schema/default_safe */;
    module.exports.DEFAULT_SCHEMA = require(308) /* ./js-yaml/schema/default_full */;
    // Deprecated functions from JS-YAML 1.x.x
    module.exports.scan = deprecated('scan');
    module.exports.parse = deprecated('parse');
    module.exports.compose = deprecated('compose');
    module.exports.addConstructor = deprecated('addConstructor');
},
/* js-yaml/lib/js-yaml/loader.js */ function _(require, module, exports, __esModule, __esExport) {
    /*eslint-disable max-len,no-use-before-define*/
    var common = require(286) /* ./common */;
    var YAMLException = require(287) /* ./exception */;
    var Mark = require(288) /* ./mark */;
    var DEFAULT_SAFE_SCHEMA = require(289) /* ./schema/default_safe */;
    var DEFAULT_FULL_SCHEMA = require(308) /* ./schema/default_full */;
    var _hasOwnProperty = Object.prototype.hasOwnProperty;
    var CONTEXT_FLOW_IN = 1;
    var CONTEXT_FLOW_OUT = 2;
    var CONTEXT_BLOCK_IN = 3;
    var CONTEXT_BLOCK_OUT = 4;
    var CHOMPING_CLIP = 1;
    var CHOMPING_STRIP = 2;
    var CHOMPING_KEEP = 3;
    var PATTERN_NON_PRINTABLE = /[\x00-\x08\x0B\x0C\x0E-\x1F\x7F-\x84\x86-\x9F\uFFFE\uFFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF]/;
    var PATTERN_NON_ASCII_LINE_BREAKS = /[\x85\u2028\u2029]/;
    var PATTERN_FLOW_INDICATORS = /[,\[\]\{\}]/;
    var PATTERN_TAG_HANDLE = /^(?:!|!!|![a-z\-]+!)$/i;
    var PATTERN_TAG_URI = /^(?:!|[^,\[\]\{\}])(?:%[0-9a-f]{2}|[0-9a-z\-#;\/\?:@&=\+\$,_\.!~\*'\(\)\[\]])*$/i;
    function _class(obj) { return Object.prototype.toString.call(obj); }
    function is_EOL(c) {
        return (c === 0x0A /* LF */) || (c === 0x0D /* CR */);
    }
    function is_WHITE_SPACE(c) {
        return (c === 0x09 /* Tab */) || (c === 0x20 /* Space */);
    }
    function is_WS_OR_EOL(c) {
        return (c === 0x09 /* Tab */) ||
            (c === 0x20 /* Space */) ||
            (c === 0x0A /* LF */) ||
            (c === 0x0D /* CR */);
    }
    function is_FLOW_INDICATOR(c) {
        return c === 0x2C /* , */ ||
            c === 0x5B /* [ */ ||
            c === 0x5D /* ] */ ||
            c === 0x7B /* { */ ||
            c === 0x7D /* } */;
    }
    function fromHexCode(c) {
        var lc;
        if ((0x30 /* 0 */ <= c) && (c <= 0x39 /* 9 */)) {
            return c - 0x30;
        }
        /*eslint-disable no-bitwise*/
        lc = c | 0x20;
        if ((0x61 /* a */ <= lc) && (lc <= 0x66 /* f */)) {
            return lc - 0x61 + 10;
        }
        return -1;
    }
    function escapedHexLen(c) {
        if (c === 0x78 /* x */) {
            return 2;
        }
        if (c === 0x75 /* u */) {
            return 4;
        }
        if (c === 0x55 /* U */) {
            return 8;
        }
        return 0;
    }
    function fromDecimalCode(c) {
        if ((0x30 /* 0 */ <= c) && (c <= 0x39 /* 9 */)) {
            return c - 0x30;
        }
        return -1;
    }
    function simpleEscapeSequence(c) {
        /* eslint-disable indent */
        return (c === 0x30 /* 0 */) ? '\x00' :
            (c === 0x61 /* a */) ? '\x07' :
                (c === 0x62 /* b */) ? '\x08' :
                    (c === 0x74 /* t */) ? '\x09' :
                        (c === 0x09 /* Tab */) ? '\x09' :
                            (c === 0x6E /* n */) ? '\x0A' :
                                (c === 0x76 /* v */) ? '\x0B' :
                                    (c === 0x66 /* f */) ? '\x0C' :
                                        (c === 0x72 /* r */) ? '\x0D' :
                                            (c === 0x65 /* e */) ? '\x1B' :
                                                (c === 0x20 /* Space */) ? ' ' :
                                                    (c === 0x22 /* " */) ? '\x22' :
                                                        (c === 0x2F /* / */) ? '/' :
                                                            (c === 0x5C /* \ */) ? '\x5C' :
                                                                (c === 0x4E /* N */) ? '\x85' :
                                                                    (c === 0x5F /* _ */) ? '\xA0' :
                                                                        (c === 0x4C /* L */) ? '\u2028' :
                                                                            (c === 0x50 /* P */) ? '\u2029' : '';
    }
    function charFromCodepoint(c) {
        if (c <= 0xFFFF) {
            return String.fromCharCode(c);
        }
        // Encode UTF-16 surrogate pair
        // https://en.wikipedia.org/wiki/UTF-16#Code_points_U.2B010000_to_U.2B10FFFF
        return String.fromCharCode(((c - 0x010000) >> 10) + 0xD800, ((c - 0x010000) & 0x03FF) + 0xDC00);
    }
    var simpleEscapeCheck = new Array(256); // integer, for fast access
    var simpleEscapeMap = new Array(256);
    for (var i = 0; i < 256; i++) {
        simpleEscapeCheck[i] = simpleEscapeSequence(i) ? 1 : 0;
        simpleEscapeMap[i] = simpleEscapeSequence(i);
    }
    function State(input, options) {
        this.input = input;
        this.filename = options['filename'] || null;
        this.schema = options['schema'] || DEFAULT_FULL_SCHEMA;
        this.onWarning = options['onWarning'] || null;
        this.legacy = options['legacy'] || false;
        this.json = options['json'] || false;
        this.listener = options['listener'] || null;
        this.implicitTypes = this.schema.compiledImplicit;
        this.typeMap = this.schema.compiledTypeMap;
        this.length = input.length;
        this.position = 0;
        this.line = 0;
        this.lineStart = 0;
        this.lineIndent = 0;
        this.documents = [];
        /*
        this.version;
        this.checkLineBreaks;
        this.tagMap;
        this.anchorMap;
        this.tag;
        this.anchor;
        this.kind;
        this.result;*/
    }
    function generateError(state, message) {
        return new YAMLException(message, new Mark(state.filename, state.input, state.position, state.line, (state.position - state.lineStart)));
    }
    function throwError(state, message) {
        throw generateError(state, message);
    }
    function throwWarning(state, message) {
        if (state.onWarning) {
            state.onWarning.call(null, generateError(state, message));
        }
    }
    var directiveHandlers = {
        YAML: function handleYamlDirective(state, name, args) {
            var match, major, minor;
            if (state.version !== null) {
                throwError(state, 'duplication of %YAML directive');
            }
            if (args.length !== 1) {
                throwError(state, 'YAML directive accepts exactly one argument');
            }
            match = /^([0-9]+)\.([0-9]+)$/.exec(args[0]);
            if (match === null) {
                throwError(state, 'ill-formed argument of the YAML directive');
            }
            major = parseInt(match[1], 10);
            minor = parseInt(match[2], 10);
            if (major !== 1) {
                throwError(state, 'unacceptable YAML version of the document');
            }
            state.version = args[0];
            state.checkLineBreaks = (minor < 2);
            if (minor !== 1 && minor !== 2) {
                throwWarning(state, 'unsupported YAML version of the document');
            }
        },
        TAG: function handleTagDirective(state, name, args) {
            var handle, prefix;
            if (args.length !== 2) {
                throwError(state, 'TAG directive accepts exactly two arguments');
            }
            handle = args[0];
            prefix = args[1];
            if (!PATTERN_TAG_HANDLE.test(handle)) {
                throwError(state, 'ill-formed tag handle (first argument) of the TAG directive');
            }
            if (_hasOwnProperty.call(state.tagMap, handle)) {
                throwError(state, 'there is a previously declared suffix for "' + handle + '" tag handle');
            }
            if (!PATTERN_TAG_URI.test(prefix)) {
                throwError(state, 'ill-formed tag prefix (second argument) of the TAG directive');
            }
            state.tagMap[handle] = prefix;
        }
    };
    function captureSegment(state, start, end, checkJson) {
        var _position, _length, _character, _result;
        if (start < end) {
            _result = state.input.slice(start, end);
            if (checkJson) {
                for (_position = 0, _length = _result.length; _position < _length; _position += 1) {
                    _character = _result.charCodeAt(_position);
                    if (!(_character === 0x09 ||
                        (0x20 <= _character && _character <= 0x10FFFF))) {
                        throwError(state, 'expected valid JSON character');
                    }
                }
            }
            else if (PATTERN_NON_PRINTABLE.test(_result)) {
                throwError(state, 'the stream contains non-printable characters');
            }
            state.result += _result;
        }
    }
    function mergeMappings(state, destination, source, overridableKeys) {
        var sourceKeys, key, index, quantity;
        if (!common.isObject(source)) {
            throwError(state, 'cannot merge mappings; the provided source object is unacceptable');
        }
        sourceKeys = Object.keys(source);
        for (index = 0, quantity = sourceKeys.length; index < quantity; index += 1) {
            key = sourceKeys[index];
            if (!_hasOwnProperty.call(destination, key)) {
                destination[key] = source[key];
                overridableKeys[key] = true;
            }
        }
    }
    function storeMappingPair(state, _result, overridableKeys, keyTag, keyNode, valueNode, startLine, startPos) {
        var index, quantity;
        // The output is a plain object here, so keys can only be strings.
        // We need to convert keyNode to a string, but doing so can hang the process
        // (deeply nested arrays that explode exponentially using aliases).
        if (Array.isArray(keyNode)) {
            keyNode = Array.prototype.slice.call(keyNode);
            for (index = 0, quantity = keyNode.length; index < quantity; index += 1) {
                if (Array.isArray(keyNode[index])) {
                    throwError(state, 'nested arrays are not supported inside keys');
                }
                if (typeof keyNode === 'object' && _class(keyNode[index]) === '[object Object]') {
                    keyNode[index] = '[object Object]';
                }
            }
        }
        // Avoid code execution in load() via toString property
        // (still use its own toString for arrays, timestamps,
        // and whatever user schema extensions happen to have @@toStringTag)
        if (typeof keyNode === 'object' && _class(keyNode) === '[object Object]') {
            keyNode = '[object Object]';
        }
        keyNode = String(keyNode);
        if (_result === null) {
            _result = {};
        }
        if (keyTag === 'tag:yaml.org,2002:merge') {
            if (Array.isArray(valueNode)) {
                for (index = 0, quantity = valueNode.length; index < quantity; index += 1) {
                    mergeMappings(state, _result, valueNode[index], overridableKeys);
                }
            }
            else {
                mergeMappings(state, _result, valueNode, overridableKeys);
            }
        }
        else {
            if (!state.json &&
                !_hasOwnProperty.call(overridableKeys, keyNode) &&
                _hasOwnProperty.call(_result, keyNode)) {
                state.line = startLine || state.line;
                state.position = startPos || state.position;
                throwError(state, 'duplicated mapping key');
            }
            _result[keyNode] = valueNode;
            delete overridableKeys[keyNode];
        }
        return _result;
    }
    function readLineBreak(state) {
        var ch;
        ch = state.input.charCodeAt(state.position);
        if (ch === 0x0A /* LF */) {
            state.position++;
        }
        else if (ch === 0x0D /* CR */) {
            state.position++;
            if (state.input.charCodeAt(state.position) === 0x0A /* LF */) {
                state.position++;
            }
        }
        else {
            throwError(state, 'a line break is expected');
        }
        state.line += 1;
        state.lineStart = state.position;
    }
    function skipSeparationSpace(state, allowComments, checkIndent) {
        var lineBreaks = 0, ch = state.input.charCodeAt(state.position);
        while (ch !== 0) {
            while (is_WHITE_SPACE(ch)) {
                ch = state.input.charCodeAt(++state.position);
            }
            if (allowComments && ch === 0x23 /* # */) {
                do {
                    ch = state.input.charCodeAt(++state.position);
                } while (ch !== 0x0A /* LF */ && ch !== 0x0D /* CR */ && ch !== 0);
            }
            if (is_EOL(ch)) {
                readLineBreak(state);
                ch = state.input.charCodeAt(state.position);
                lineBreaks++;
                state.lineIndent = 0;
                while (ch === 0x20 /* Space */) {
                    state.lineIndent++;
                    ch = state.input.charCodeAt(++state.position);
                }
            }
            else {
                break;
            }
        }
        if (checkIndent !== -1 && lineBreaks !== 0 && state.lineIndent < checkIndent) {
            throwWarning(state, 'deficient indentation');
        }
        return lineBreaks;
    }
    function testDocumentSeparator(state) {
        var _position = state.position, ch;
        ch = state.input.charCodeAt(_position);
        // Condition state.position === state.lineStart is tested
        // in parent on each call, for efficiency. No needs to test here again.
        if ((ch === 0x2D /* - */ || ch === 0x2E /* . */) &&
            ch === state.input.charCodeAt(_position + 1) &&
            ch === state.input.charCodeAt(_position + 2)) {
            _position += 3;
            ch = state.input.charCodeAt(_position);
            if (ch === 0 || is_WS_OR_EOL(ch)) {
                return true;
            }
        }
        return false;
    }
    function writeFoldedLines(state, count) {
        if (count === 1) {
            state.result += ' ';
        }
        else if (count > 1) {
            state.result += common.repeat('\n', count - 1);
        }
    }
    function readPlainScalar(state, nodeIndent, withinFlowCollection) {
        var preceding, following, captureStart, captureEnd, hasPendingContent, _line, _lineStart, _lineIndent, _kind = state.kind, _result = state.result, ch;
        ch = state.input.charCodeAt(state.position);
        if (is_WS_OR_EOL(ch) ||
            is_FLOW_INDICATOR(ch) ||
            ch === 0x23 /* # */ ||
            ch === 0x26 /* & */ ||
            ch === 0x2A /* * */ ||
            ch === 0x21 /* ! */ ||
            ch === 0x7C /* | */ ||
            ch === 0x3E /* > */ ||
            ch === 0x27 /* ' */ ||
            ch === 0x22 /* " */ ||
            ch === 0x25 /* % */ ||
            ch === 0x40 /* @ */ ||
            ch === 0x60 /* ` */) {
            return false;
        }
        if (ch === 0x3F /* ? */ || ch === 0x2D /* - */) {
            following = state.input.charCodeAt(state.position + 1);
            if (is_WS_OR_EOL(following) ||
                withinFlowCollection && is_FLOW_INDICATOR(following)) {
                return false;
            }
        }
        state.kind = 'scalar';
        state.result = '';
        captureStart = captureEnd = state.position;
        hasPendingContent = false;
        while (ch !== 0) {
            if (ch === 0x3A /* : */) {
                following = state.input.charCodeAt(state.position + 1);
                if (is_WS_OR_EOL(following) ||
                    withinFlowCollection && is_FLOW_INDICATOR(following)) {
                    break;
                }
            }
            else if (ch === 0x23 /* # */) {
                preceding = state.input.charCodeAt(state.position - 1);
                if (is_WS_OR_EOL(preceding)) {
                    break;
                }
            }
            else if ((state.position === state.lineStart && testDocumentSeparator(state)) ||
                withinFlowCollection && is_FLOW_INDICATOR(ch)) {
                break;
            }
            else if (is_EOL(ch)) {
                _line = state.line;
                _lineStart = state.lineStart;
                _lineIndent = state.lineIndent;
                skipSeparationSpace(state, false, -1);
                if (state.lineIndent >= nodeIndent) {
                    hasPendingContent = true;
                    ch = state.input.charCodeAt(state.position);
                    continue;
                }
                else {
                    state.position = captureEnd;
                    state.line = _line;
                    state.lineStart = _lineStart;
                    state.lineIndent = _lineIndent;
                    break;
                }
            }
            if (hasPendingContent) {
                captureSegment(state, captureStart, captureEnd, false);
                writeFoldedLines(state, state.line - _line);
                captureStart = captureEnd = state.position;
                hasPendingContent = false;
            }
            if (!is_WHITE_SPACE(ch)) {
                captureEnd = state.position + 1;
            }
            ch = state.input.charCodeAt(++state.position);
        }
        captureSegment(state, captureStart, captureEnd, false);
        if (state.result) {
            return true;
        }
        state.kind = _kind;
        state.result = _result;
        return false;
    }
    function readSingleQuotedScalar(state, nodeIndent) {
        var ch, captureStart, captureEnd;
        ch = state.input.charCodeAt(state.position);
        if (ch !== 0x27 /* ' */) {
            return false;
        }
        state.kind = 'scalar';
        state.result = '';
        state.position++;
        captureStart = captureEnd = state.position;
        while ((ch = state.input.charCodeAt(state.position)) !== 0) {
            if (ch === 0x27 /* ' */) {
                captureSegment(state, captureStart, state.position, true);
                ch = state.input.charCodeAt(++state.position);
                if (ch === 0x27 /* ' */) {
                    captureStart = state.position;
                    state.position++;
                    captureEnd = state.position;
                }
                else {
                    return true;
                }
            }
            else if (is_EOL(ch)) {
                captureSegment(state, captureStart, captureEnd, true);
                writeFoldedLines(state, skipSeparationSpace(state, false, nodeIndent));
                captureStart = captureEnd = state.position;
            }
            else if (state.position === state.lineStart && testDocumentSeparator(state)) {
                throwError(state, 'unexpected end of the document within a single quoted scalar');
            }
            else {
                state.position++;
                captureEnd = state.position;
            }
        }
        throwError(state, 'unexpected end of the stream within a single quoted scalar');
    }
    function readDoubleQuotedScalar(state, nodeIndent) {
        var captureStart, captureEnd, hexLength, hexResult, tmp, ch;
        ch = state.input.charCodeAt(state.position);
        if (ch !== 0x22 /* " */) {
            return false;
        }
        state.kind = 'scalar';
        state.result = '';
        state.position++;
        captureStart = captureEnd = state.position;
        while ((ch = state.input.charCodeAt(state.position)) !== 0) {
            if (ch === 0x22 /* " */) {
                captureSegment(state, captureStart, state.position, true);
                state.position++;
                return true;
            }
            else if (ch === 0x5C /* \ */) {
                captureSegment(state, captureStart, state.position, true);
                ch = state.input.charCodeAt(++state.position);
                if (is_EOL(ch)) {
                    skipSeparationSpace(state, false, nodeIndent);
                    // TODO: rework to inline fn with no type cast?
                }
                else if (ch < 256 && simpleEscapeCheck[ch]) {
                    state.result += simpleEscapeMap[ch];
                    state.position++;
                }
                else if ((tmp = escapedHexLen(ch)) > 0) {
                    hexLength = tmp;
                    hexResult = 0;
                    for (; hexLength > 0; hexLength--) {
                        ch = state.input.charCodeAt(++state.position);
                        if ((tmp = fromHexCode(ch)) >= 0) {
                            hexResult = (hexResult << 4) + tmp;
                        }
                        else {
                            throwError(state, 'expected hexadecimal character');
                        }
                    }
                    state.result += charFromCodepoint(hexResult);
                    state.position++;
                }
                else {
                    throwError(state, 'unknown escape sequence');
                }
                captureStart = captureEnd = state.position;
            }
            else if (is_EOL(ch)) {
                captureSegment(state, captureStart, captureEnd, true);
                writeFoldedLines(state, skipSeparationSpace(state, false, nodeIndent));
                captureStart = captureEnd = state.position;
            }
            else if (state.position === state.lineStart && testDocumentSeparator(state)) {
                throwError(state, 'unexpected end of the document within a double quoted scalar');
            }
            else {
                state.position++;
                captureEnd = state.position;
            }
        }
        throwError(state, 'unexpected end of the stream within a double quoted scalar');
    }
    function readFlowCollection(state, nodeIndent) {
        var readNext = true, _line, _tag = state.tag, _result, _anchor = state.anchor, following, terminator, isPair, isExplicitPair, isMapping, overridableKeys = {}, keyNode, keyTag, valueNode, ch;
        ch = state.input.charCodeAt(state.position);
        if (ch === 0x5B /* [ */) {
            terminator = 0x5D; /* ] */
            isMapping = false;
            _result = [];
        }
        else if (ch === 0x7B /* { */) {
            terminator = 0x7D; /* } */
            isMapping = true;
            _result = {};
        }
        else {
            return false;
        }
        if (state.anchor !== null) {
            state.anchorMap[state.anchor] = _result;
        }
        ch = state.input.charCodeAt(++state.position);
        while (ch !== 0) {
            skipSeparationSpace(state, true, nodeIndent);
            ch = state.input.charCodeAt(state.position);
            if (ch === terminator) {
                state.position++;
                state.tag = _tag;
                state.anchor = _anchor;
                state.kind = isMapping ? 'mapping' : 'sequence';
                state.result = _result;
                return true;
            }
            else if (!readNext) {
                throwError(state, 'missed comma between flow collection entries');
            }
            keyTag = keyNode = valueNode = null;
            isPair = isExplicitPair = false;
            if (ch === 0x3F /* ? */) {
                following = state.input.charCodeAt(state.position + 1);
                if (is_WS_OR_EOL(following)) {
                    isPair = isExplicitPair = true;
                    state.position++;
                    skipSeparationSpace(state, true, nodeIndent);
                }
            }
            _line = state.line;
            composeNode(state, nodeIndent, CONTEXT_FLOW_IN, false, true);
            keyTag = state.tag;
            keyNode = state.result;
            skipSeparationSpace(state, true, nodeIndent);
            ch = state.input.charCodeAt(state.position);
            if ((isExplicitPair || state.line === _line) && ch === 0x3A /* : */) {
                isPair = true;
                ch = state.input.charCodeAt(++state.position);
                skipSeparationSpace(state, true, nodeIndent);
                composeNode(state, nodeIndent, CONTEXT_FLOW_IN, false, true);
                valueNode = state.result;
            }
            if (isMapping) {
                storeMappingPair(state, _result, overridableKeys, keyTag, keyNode, valueNode);
            }
            else if (isPair) {
                _result.push(storeMappingPair(state, null, overridableKeys, keyTag, keyNode, valueNode));
            }
            else {
                _result.push(keyNode);
            }
            skipSeparationSpace(state, true, nodeIndent);
            ch = state.input.charCodeAt(state.position);
            if (ch === 0x2C /* , */) {
                readNext = true;
                ch = state.input.charCodeAt(++state.position);
            }
            else {
                readNext = false;
            }
        }
        throwError(state, 'unexpected end of the stream within a flow collection');
    }
    function readBlockScalar(state, nodeIndent) {
        var captureStart, folding, chomping = CHOMPING_CLIP, didReadContent = false, detectedIndent = false, textIndent = nodeIndent, emptyLines = 0, atMoreIndented = false, tmp, ch;
        ch = state.input.charCodeAt(state.position);
        if (ch === 0x7C /* | */) {
            folding = false;
        }
        else if (ch === 0x3E /* > */) {
            folding = true;
        }
        else {
            return false;
        }
        state.kind = 'scalar';
        state.result = '';
        while (ch !== 0) {
            ch = state.input.charCodeAt(++state.position);
            if (ch === 0x2B /* + */ || ch === 0x2D /* - */) {
                if (CHOMPING_CLIP === chomping) {
                    chomping = (ch === 0x2B /* + */) ? CHOMPING_KEEP : CHOMPING_STRIP;
                }
                else {
                    throwError(state, 'repeat of a chomping mode identifier');
                }
            }
            else if ((tmp = fromDecimalCode(ch)) >= 0) {
                if (tmp === 0) {
                    throwError(state, 'bad explicit indentation width of a block scalar; it cannot be less than one');
                }
                else if (!detectedIndent) {
                    textIndent = nodeIndent + tmp - 1;
                    detectedIndent = true;
                }
                else {
                    throwError(state, 'repeat of an indentation width identifier');
                }
            }
            else {
                break;
            }
        }
        if (is_WHITE_SPACE(ch)) {
            do {
                ch = state.input.charCodeAt(++state.position);
            } while (is_WHITE_SPACE(ch));
            if (ch === 0x23 /* # */) {
                do {
                    ch = state.input.charCodeAt(++state.position);
                } while (!is_EOL(ch) && (ch !== 0));
            }
        }
        while (ch !== 0) {
            readLineBreak(state);
            state.lineIndent = 0;
            ch = state.input.charCodeAt(state.position);
            while ((!detectedIndent || state.lineIndent < textIndent) &&
                (ch === 0x20 /* Space */)) {
                state.lineIndent++;
                ch = state.input.charCodeAt(++state.position);
            }
            if (!detectedIndent && state.lineIndent > textIndent) {
                textIndent = state.lineIndent;
            }
            if (is_EOL(ch)) {
                emptyLines++;
                continue;
            }
            // End of the scalar.
            if (state.lineIndent < textIndent) {
                // Perform the chomping.
                if (chomping === CHOMPING_KEEP) {
                    state.result += common.repeat('\n', didReadContent ? 1 + emptyLines : emptyLines);
                }
                else if (chomping === CHOMPING_CLIP) {
                    if (didReadContent) { // i.e. only if the scalar is not empty.
                        state.result += '\n';
                    }
                }
                // Break this `while` cycle and go to the funciton's epilogue.
                break;
            }
            // Folded style: use fancy rules to handle line breaks.
            if (folding) {
                // Lines starting with white space characters (more-indented lines) are not folded.
                if (is_WHITE_SPACE(ch)) {
                    atMoreIndented = true;
                    // except for the first content line (cf. Example 8.1)
                    state.result += common.repeat('\n', didReadContent ? 1 + emptyLines : emptyLines);
                    // End of more-indented block.
                }
                else if (atMoreIndented) {
                    atMoreIndented = false;
                    state.result += common.repeat('\n', emptyLines + 1);
                    // Just one line break - perceive as the same line.
                }
                else if (emptyLines === 0) {
                    if (didReadContent) { // i.e. only if we have already read some scalar content.
                        state.result += ' ';
                    }
                    // Several line breaks - perceive as different lines.
                }
                else {
                    state.result += common.repeat('\n', emptyLines);
                }
                // Literal style: just add exact number of line breaks between content lines.
            }
            else {
                // Keep all line breaks except the header line break.
                state.result += common.repeat('\n', didReadContent ? 1 + emptyLines : emptyLines);
            }
            didReadContent = true;
            detectedIndent = true;
            emptyLines = 0;
            captureStart = state.position;
            while (!is_EOL(ch) && (ch !== 0)) {
                ch = state.input.charCodeAt(++state.position);
            }
            captureSegment(state, captureStart, state.position, false);
        }
        return true;
    }
    function readBlockSequence(state, nodeIndent) {
        var _line, _tag = state.tag, _anchor = state.anchor, _result = [], following, detected = false, ch;
        if (state.anchor !== null) {
            state.anchorMap[state.anchor] = _result;
        }
        ch = state.input.charCodeAt(state.position);
        while (ch !== 0) {
            if (ch !== 0x2D /* - */) {
                break;
            }
            following = state.input.charCodeAt(state.position + 1);
            if (!is_WS_OR_EOL(following)) {
                break;
            }
            detected = true;
            state.position++;
            if (skipSeparationSpace(state, true, -1)) {
                if (state.lineIndent <= nodeIndent) {
                    _result.push(null);
                    ch = state.input.charCodeAt(state.position);
                    continue;
                }
            }
            _line = state.line;
            composeNode(state, nodeIndent, CONTEXT_BLOCK_IN, false, true);
            _result.push(state.result);
            skipSeparationSpace(state, true, -1);
            ch = state.input.charCodeAt(state.position);
            if ((state.line === _line || state.lineIndent > nodeIndent) && (ch !== 0)) {
                throwError(state, 'bad indentation of a sequence entry');
            }
            else if (state.lineIndent < nodeIndent) {
                break;
            }
        }
        if (detected) {
            state.tag = _tag;
            state.anchor = _anchor;
            state.kind = 'sequence';
            state.result = _result;
            return true;
        }
        return false;
    }
    function readBlockMapping(state, nodeIndent, flowIndent) {
        var following, allowCompact, _line, _pos, _tag = state.tag, _anchor = state.anchor, _result = {}, overridableKeys = {}, keyTag = null, keyNode = null, valueNode = null, atExplicitKey = false, detected = false, ch;
        if (state.anchor !== null) {
            state.anchorMap[state.anchor] = _result;
        }
        ch = state.input.charCodeAt(state.position);
        while (ch !== 0) {
            following = state.input.charCodeAt(state.position + 1);
            _line = state.line; // Save the current line.
            _pos = state.position;
            //
            // Explicit notation case. There are two separate blocks:
            // first for the key (denoted by "?") and second for the value (denoted by ":")
            //
            if ((ch === 0x3F /* ? */ || ch === 0x3A /* : */) && is_WS_OR_EOL(following)) {
                if (ch === 0x3F /* ? */) {
                    if (atExplicitKey) {
                        storeMappingPair(state, _result, overridableKeys, keyTag, keyNode, null);
                        keyTag = keyNode = valueNode = null;
                    }
                    detected = true;
                    atExplicitKey = true;
                    allowCompact = true;
                }
                else if (atExplicitKey) {
                    // i.e. 0x3A/* : */ === character after the explicit key.
                    atExplicitKey = false;
                    allowCompact = true;
                }
                else {
                    throwError(state, 'incomplete explicit mapping pair; a key node is missed; or followed by a non-tabulated empty line');
                }
                state.position += 1;
                ch = following;
                //
                // Implicit notation case. Flow-style node as the key first, then ":", and the value.
                //
            }
            else if (composeNode(state, flowIndent, CONTEXT_FLOW_OUT, false, true)) {
                if (state.line === _line) {
                    ch = state.input.charCodeAt(state.position);
                    while (is_WHITE_SPACE(ch)) {
                        ch = state.input.charCodeAt(++state.position);
                    }
                    if (ch === 0x3A /* : */) {
                        ch = state.input.charCodeAt(++state.position);
                        if (!is_WS_OR_EOL(ch)) {
                            throwError(state, 'a whitespace character is expected after the key-value separator within a block mapping');
                        }
                        if (atExplicitKey) {
                            storeMappingPair(state, _result, overridableKeys, keyTag, keyNode, null);
                            keyTag = keyNode = valueNode = null;
                        }
                        detected = true;
                        atExplicitKey = false;
                        allowCompact = false;
                        keyTag = state.tag;
                        keyNode = state.result;
                    }
                    else if (detected) {
                        throwError(state, 'can not read an implicit mapping pair; a colon is missed');
                    }
                    else {
                        state.tag = _tag;
                        state.anchor = _anchor;
                        return true; // Keep the result of `composeNode`.
                    }
                }
                else if (detected) {
                    throwError(state, 'can not read a block mapping entry; a multiline key may not be an implicit key');
                }
                else {
                    state.tag = _tag;
                    state.anchor = _anchor;
                    return true; // Keep the result of `composeNode`.
                }
            }
            else {
                break; // Reading is done. Go to the epilogue.
            }
            //
            // Common reading code for both explicit and implicit notations.
            //
            if (state.line === _line || state.lineIndent > nodeIndent) {
                if (composeNode(state, nodeIndent, CONTEXT_BLOCK_OUT, true, allowCompact)) {
                    if (atExplicitKey) {
                        keyNode = state.result;
                    }
                    else {
                        valueNode = state.result;
                    }
                }
                if (!atExplicitKey) {
                    storeMappingPair(state, _result, overridableKeys, keyTag, keyNode, valueNode, _line, _pos);
                    keyTag = keyNode = valueNode = null;
                }
                skipSeparationSpace(state, true, -1);
                ch = state.input.charCodeAt(state.position);
            }
            if (state.lineIndent > nodeIndent && (ch !== 0)) {
                throwError(state, 'bad indentation of a mapping entry');
            }
            else if (state.lineIndent < nodeIndent) {
                break;
            }
        }
        //
        // Epilogue.
        //
        // Special case: last mapping's node contains only the key in explicit notation.
        if (atExplicitKey) {
            storeMappingPair(state, _result, overridableKeys, keyTag, keyNode, null);
        }
        // Expose the resulting mapping.
        if (detected) {
            state.tag = _tag;
            state.anchor = _anchor;
            state.kind = 'mapping';
            state.result = _result;
        }
        return detected;
    }
    function readTagProperty(state) {
        var _position, isVerbatim = false, isNamed = false, tagHandle, tagName, ch;
        ch = state.input.charCodeAt(state.position);
        if (ch !== 0x21 /* ! */)
            return false;
        if (state.tag !== null) {
            throwError(state, 'duplication of a tag property');
        }
        ch = state.input.charCodeAt(++state.position);
        if (ch === 0x3C /* < */) {
            isVerbatim = true;
            ch = state.input.charCodeAt(++state.position);
        }
        else if (ch === 0x21 /* ! */) {
            isNamed = true;
            tagHandle = '!!';
            ch = state.input.charCodeAt(++state.position);
        }
        else {
            tagHandle = '!';
        }
        _position = state.position;
        if (isVerbatim) {
            do {
                ch = state.input.charCodeAt(++state.position);
            } while (ch !== 0 && ch !== 0x3E /* > */);
            if (state.position < state.length) {
                tagName = state.input.slice(_position, state.position);
                ch = state.input.charCodeAt(++state.position);
            }
            else {
                throwError(state, 'unexpected end of the stream within a verbatim tag');
            }
        }
        else {
            while (ch !== 0 && !is_WS_OR_EOL(ch)) {
                if (ch === 0x21 /* ! */) {
                    if (!isNamed) {
                        tagHandle = state.input.slice(_position - 1, state.position + 1);
                        if (!PATTERN_TAG_HANDLE.test(tagHandle)) {
                            throwError(state, 'named tag handle cannot contain such characters');
                        }
                        isNamed = true;
                        _position = state.position + 1;
                    }
                    else {
                        throwError(state, 'tag suffix cannot contain exclamation marks');
                    }
                }
                ch = state.input.charCodeAt(++state.position);
            }
            tagName = state.input.slice(_position, state.position);
            if (PATTERN_FLOW_INDICATORS.test(tagName)) {
                throwError(state, 'tag suffix cannot contain flow indicator characters');
            }
        }
        if (tagName && !PATTERN_TAG_URI.test(tagName)) {
            throwError(state, 'tag name cannot contain such characters: ' + tagName);
        }
        if (isVerbatim) {
            state.tag = tagName;
        }
        else if (_hasOwnProperty.call(state.tagMap, tagHandle)) {
            state.tag = state.tagMap[tagHandle] + tagName;
        }
        else if (tagHandle === '!') {
            state.tag = '!' + tagName;
        }
        else if (tagHandle === '!!') {
            state.tag = 'tag:yaml.org,2002:' + tagName;
        }
        else {
            throwError(state, 'undeclared tag handle "' + tagHandle + '"');
        }
        return true;
    }
    function readAnchorProperty(state) {
        var _position, ch;
        ch = state.input.charCodeAt(state.position);
        if (ch !== 0x26 /* & */)
            return false;
        if (state.anchor !== null) {
            throwError(state, 'duplication of an anchor property');
        }
        ch = state.input.charCodeAt(++state.position);
        _position = state.position;
        while (ch !== 0 && !is_WS_OR_EOL(ch) && !is_FLOW_INDICATOR(ch)) {
            ch = state.input.charCodeAt(++state.position);
        }
        if (state.position === _position) {
            throwError(state, 'name of an anchor node must contain at least one character');
        }
        state.anchor = state.input.slice(_position, state.position);
        return true;
    }
    function readAlias(state) {
        var _position, alias, ch;
        ch = state.input.charCodeAt(state.position);
        if (ch !== 0x2A /* * */)
            return false;
        ch = state.input.charCodeAt(++state.position);
        _position = state.position;
        while (ch !== 0 && !is_WS_OR_EOL(ch) && !is_FLOW_INDICATOR(ch)) {
            ch = state.input.charCodeAt(++state.position);
        }
        if (state.position === _position) {
            throwError(state, 'name of an alias node must contain at least one character');
        }
        alias = state.input.slice(_position, state.position);
        if (!_hasOwnProperty.call(state.anchorMap, alias)) {
            throwError(state, 'unidentified alias "' + alias + '"');
        }
        state.result = state.anchorMap[alias];
        skipSeparationSpace(state, true, -1);
        return true;
    }
    function composeNode(state, parentIndent, nodeContext, allowToSeek, allowCompact) {
        var allowBlockStyles, allowBlockScalars, allowBlockCollections, indentStatus = 1, // 1: this>parent, 0: this=parent, -1: this<parent
        atNewLine = false, hasContent = false, typeIndex, typeQuantity, type, flowIndent, blockIndent;
        if (state.listener !== null) {
            state.listener('open', state);
        }
        state.tag = null;
        state.anchor = null;
        state.kind = null;
        state.result = null;
        allowBlockStyles = allowBlockScalars = allowBlockCollections =
            CONTEXT_BLOCK_OUT === nodeContext ||
                CONTEXT_BLOCK_IN === nodeContext;
        if (allowToSeek) {
            if (skipSeparationSpace(state, true, -1)) {
                atNewLine = true;
                if (state.lineIndent > parentIndent) {
                    indentStatus = 1;
                }
                else if (state.lineIndent === parentIndent) {
                    indentStatus = 0;
                }
                else if (state.lineIndent < parentIndent) {
                    indentStatus = -1;
                }
            }
        }
        if (indentStatus === 1) {
            while (readTagProperty(state) || readAnchorProperty(state)) {
                if (skipSeparationSpace(state, true, -1)) {
                    atNewLine = true;
                    allowBlockCollections = allowBlockStyles;
                    if (state.lineIndent > parentIndent) {
                        indentStatus = 1;
                    }
                    else if (state.lineIndent === parentIndent) {
                        indentStatus = 0;
                    }
                    else if (state.lineIndent < parentIndent) {
                        indentStatus = -1;
                    }
                }
                else {
                    allowBlockCollections = false;
                }
            }
        }
        if (allowBlockCollections) {
            allowBlockCollections = atNewLine || allowCompact;
        }
        if (indentStatus === 1 || CONTEXT_BLOCK_OUT === nodeContext) {
            if (CONTEXT_FLOW_IN === nodeContext || CONTEXT_FLOW_OUT === nodeContext) {
                flowIndent = parentIndent;
            }
            else {
                flowIndent = parentIndent + 1;
            }
            blockIndent = state.position - state.lineStart;
            if (indentStatus === 1) {
                if (allowBlockCollections &&
                    (readBlockSequence(state, blockIndent) ||
                        readBlockMapping(state, blockIndent, flowIndent)) ||
                    readFlowCollection(state, flowIndent)) {
                    hasContent = true;
                }
                else {
                    if ((allowBlockScalars && readBlockScalar(state, flowIndent)) ||
                        readSingleQuotedScalar(state, flowIndent) ||
                        readDoubleQuotedScalar(state, flowIndent)) {
                        hasContent = true;
                    }
                    else if (readAlias(state)) {
                        hasContent = true;
                        if (state.tag !== null || state.anchor !== null) {
                            throwError(state, 'alias node should not have any properties');
                        }
                    }
                    else if (readPlainScalar(state, flowIndent, CONTEXT_FLOW_IN === nodeContext)) {
                        hasContent = true;
                        if (state.tag === null) {
                            state.tag = '?';
                        }
                    }
                    if (state.anchor !== null) {
                        state.anchorMap[state.anchor] = state.result;
                    }
                }
            }
            else if (indentStatus === 0) {
                // Special case: block sequences are allowed to have same indentation level as the parent.
                // http://www.yaml.org/spec/1.2/spec.html#id2799784
                hasContent = allowBlockCollections && readBlockSequence(state, blockIndent);
            }
        }
        if (state.tag !== null && state.tag !== '!') {
            if (state.tag === '?') {
                // Implicit resolving is not allowed for non-scalar types, and '?'
                // non-specific tag is only automatically assigned to plain scalars.
                //
                // We only need to check kind conformity in case user explicitly assigns '?'
                // tag, for example like this: "!<?> [0]"
                //
                if (state.result !== null && state.kind !== 'scalar') {
                    throwError(state, 'unacceptable node kind for !<?> tag; it should be "scalar", not "' + state.kind + '"');
                }
                for (typeIndex = 0, typeQuantity = state.implicitTypes.length; typeIndex < typeQuantity; typeIndex += 1) {
                    type = state.implicitTypes[typeIndex];
                    if (type.resolve(state.result)) { // `state.result` updated in resolver if matched
                        state.result = type.construct(state.result);
                        state.tag = type.tag;
                        if (state.anchor !== null) {
                            state.anchorMap[state.anchor] = state.result;
                        }
                        break;
                    }
                }
            }
            else if (_hasOwnProperty.call(state.typeMap[state.kind || 'fallback'], state.tag)) {
                type = state.typeMap[state.kind || 'fallback'][state.tag];
                if (state.result !== null && type.kind !== state.kind) {
                    throwError(state, 'unacceptable node kind for !<' + state.tag + '> tag; it should be "' + type.kind + '", not "' + state.kind + '"');
                }
                if (!type.resolve(state.result)) { // `state.result` updated in resolver if matched
                    throwError(state, 'cannot resolve a node with !<' + state.tag + '> explicit tag');
                }
                else {
                    state.result = type.construct(state.result);
                    if (state.anchor !== null) {
                        state.anchorMap[state.anchor] = state.result;
                    }
                }
            }
            else {
                throwError(state, 'unknown tag !<' + state.tag + '>');
            }
        }
        if (state.listener !== null) {
            state.listener('close', state);
        }
        return state.tag !== null || state.anchor !== null || hasContent;
    }
    function readDocument(state) {
        var documentStart = state.position, _position, directiveName, directiveArgs, hasDirectives = false, ch;
        state.version = null;
        state.checkLineBreaks = state.legacy;
        state.tagMap = {};
        state.anchorMap = {};
        while ((ch = state.input.charCodeAt(state.position)) !== 0) {
            skipSeparationSpace(state, true, -1);
            ch = state.input.charCodeAt(state.position);
            if (state.lineIndent > 0 || ch !== 0x25 /* % */) {
                break;
            }
            hasDirectives = true;
            ch = state.input.charCodeAt(++state.position);
            _position = state.position;
            while (ch !== 0 && !is_WS_OR_EOL(ch)) {
                ch = state.input.charCodeAt(++state.position);
            }
            directiveName = state.input.slice(_position, state.position);
            directiveArgs = [];
            if (directiveName.length < 1) {
                throwError(state, 'directive name must not be less than one character in length');
            }
            while (ch !== 0) {
                while (is_WHITE_SPACE(ch)) {
                    ch = state.input.charCodeAt(++state.position);
                }
                if (ch === 0x23 /* # */) {
                    do {
                        ch = state.input.charCodeAt(++state.position);
                    } while (ch !== 0 && !is_EOL(ch));
                    break;
                }
                if (is_EOL(ch))
                    break;
                _position = state.position;
                while (ch !== 0 && !is_WS_OR_EOL(ch)) {
                    ch = state.input.charCodeAt(++state.position);
                }
                directiveArgs.push(state.input.slice(_position, state.position));
            }
            if (ch !== 0)
                readLineBreak(state);
            if (_hasOwnProperty.call(directiveHandlers, directiveName)) {
                directiveHandlers[directiveName](state, directiveName, directiveArgs);
            }
            else {
                throwWarning(state, 'unknown document directive "' + directiveName + '"');
            }
        }
        skipSeparationSpace(state, true, -1);
        if (state.lineIndent === 0 &&
            state.input.charCodeAt(state.position) === 0x2D /* - */ &&
            state.input.charCodeAt(state.position + 1) === 0x2D /* - */ &&
            state.input.charCodeAt(state.position + 2) === 0x2D /* - */) {
            state.position += 3;
            skipSeparationSpace(state, true, -1);
        }
        else if (hasDirectives) {
            throwError(state, 'directives end mark is expected');
        }
        composeNode(state, state.lineIndent - 1, CONTEXT_BLOCK_OUT, false, true);
        skipSeparationSpace(state, true, -1);
        if (state.checkLineBreaks &&
            PATTERN_NON_ASCII_LINE_BREAKS.test(state.input.slice(documentStart, state.position))) {
            throwWarning(state, 'non-ASCII line breaks are interpreted as content');
        }
        state.documents.push(state.result);
        if (state.position === state.lineStart && testDocumentSeparator(state)) {
            if (state.input.charCodeAt(state.position) === 0x2E /* . */) {
                state.position += 3;
                skipSeparationSpace(state, true, -1);
            }
            return;
        }
        if (state.position < (state.length - 1)) {
            throwError(state, 'end of the stream or a document separator is expected');
        }
        else {
            return;
        }
    }
    function loadDocuments(input, options) {
        input = String(input);
        options = options || {};
        if (input.length !== 0) {
            // Add tailing `\n` if not exists
            if (input.charCodeAt(input.length - 1) !== 0x0A /* LF */ &&
                input.charCodeAt(input.length - 1) !== 0x0D /* CR */) {
                input += '\n';
            }
            // Strip BOM
            if (input.charCodeAt(0) === 0xFEFF) {
                input = input.slice(1);
            }
        }
        var state = new State(input, options);
        var nullpos = input.indexOf('\0');
        if (nullpos !== -1) {
            state.position = nullpos;
            throwError(state, 'null byte is not allowed in input');
        }
        // Use 0 as string terminator. That significantly simplifies bounds check.
        state.input += '\0';
        while (state.input.charCodeAt(state.position) === 0x20 /* Space */) {
            state.lineIndent += 1;
            state.position += 1;
        }
        while (state.position < (state.length - 1)) {
            readDocument(state);
        }
        return state.documents;
    }
    function loadAll(input, iterator, options) {
        if (iterator !== null && typeof iterator === 'object' && typeof options === 'undefined') {
            options = iterator;
            iterator = null;
        }
        var documents = loadDocuments(input, options);
        if (typeof iterator !== 'function') {
            return documents;
        }
        for (var index = 0, length = documents.length; index < length; index += 1) {
            iterator(documents[index]);
        }
    }
    function load(input, options) {
        var documents = loadDocuments(input, options);
        if (documents.length === 0) {
            /*eslint-disable no-undefined*/
            return undefined;
        }
        else if (documents.length === 1) {
            return documents[0];
        }
        throw new YAMLException('expected a single document in the stream, but found more');
    }
    function safeLoadAll(input, iterator, options) {
        if (typeof iterator === 'object' && iterator !== null && typeof options === 'undefined') {
            options = iterator;
            iterator = null;
        }
        return loadAll(input, iterator, common.extend({ schema: DEFAULT_SAFE_SCHEMA }, options));
    }
    function safeLoad(input, options) {
        return load(input, common.extend({ schema: DEFAULT_SAFE_SCHEMA }, options));
    }
    module.exports.loadAll = loadAll;
    module.exports.load = load;
    module.exports.safeLoadAll = safeLoadAll;
    module.exports.safeLoad = safeLoad;
},
/* js-yaml/lib/js-yaml/common.js */ function _(require, module, exports, __esModule, __esExport) {
    function isNothing(subject) {
        return (typeof subject === 'undefined') || (subject === null);
    }
    function isObject(subject) {
        return (typeof subject === 'object') && (subject !== null);
    }
    function toArray(sequence) {
        if (Array.isArray(sequence))
            return sequence;
        else if (isNothing(sequence))
            return [];
        return [sequence];
    }
    function extend(target, source) {
        var index, length, key, sourceKeys;
        if (source) {
            sourceKeys = Object.keys(source);
            for (index = 0, length = sourceKeys.length; index < length; index += 1) {
                key = sourceKeys[index];
                target[key] = source[key];
            }
        }
        return target;
    }
    function repeat(string, count) {
        var result = '', cycle;
        for (cycle = 0; cycle < count; cycle += 1) {
            result += string;
        }
        return result;
    }
    function isNegativeZero(number) {
        return (number === 0) && (Number.NEGATIVE_INFINITY === 1 / number);
    }
    module.exports.isNothing = isNothing;
    module.exports.isObject = isObject;
    module.exports.toArray = toArray;
    module.exports.repeat = repeat;
    module.exports.isNegativeZero = isNegativeZero;
    module.exports.extend = extend;
},
/* js-yaml/lib/js-yaml/exception.js */ function _(require, module, exports, __esModule, __esExport) {
    function YAMLException(reason, mark) {
        // Super constructor
        Error.call(this);
        this.name = 'YAMLException';
        this.reason = reason;
        this.mark = mark;
        this.message = (this.reason || '(unknown reason)') + (this.mark ? ' ' + this.mark.toString() : '');
        // Include stack trace in error object
        if (Error.captureStackTrace) {
            // Chrome and NodeJS
            Error.captureStackTrace(this, this.constructor);
        }
        else {
            // FF, IE 10+ and Safari 6+. Fallback for others
            this.stack = (new Error()).stack || '';
        }
    }
    // Inherit from Error
    YAMLException.prototype = Object.create(Error.prototype);
    YAMLException.prototype.constructor = YAMLException;
    YAMLException.prototype.toString = function toString(compact) {
        var result = this.name + ': ';
        result += this.reason || '(unknown reason)';
        if (!compact && this.mark) {
            result += ' ' + this.mark.toString();
        }
        return result;
    };
    module.exports = YAMLException;
},
/* js-yaml/lib/js-yaml/mark.js */ function _(require, module, exports, __esModule, __esExport) {
    var common = require(286) /* ./common */;
    function Mark(name, buffer, position, line, column) {
        this.name = name;
        this.buffer = buffer;
        this.position = position;
        this.line = line;
        this.column = column;
    }
    Mark.prototype.getSnippet = function getSnippet(indent, maxLength) {
        var head, start, tail, end, snippet;
        if (!this.buffer)
            return null;
        indent = indent || 4;
        maxLength = maxLength || 75;
        head = '';
        start = this.position;
        while (start > 0 && '\x00\r\n\x85\u2028\u2029'.indexOf(this.buffer.charAt(start - 1)) === -1) {
            start -= 1;
            if (this.position - start > (maxLength / 2 - 1)) {
                head = ' ... ';
                start += 5;
                break;
            }
        }
        tail = '';
        end = this.position;
        while (end < this.buffer.length && '\x00\r\n\x85\u2028\u2029'.indexOf(this.buffer.charAt(end)) === -1) {
            end += 1;
            if (end - this.position > (maxLength / 2 - 1)) {
                tail = ' ... ';
                end -= 5;
                break;
            }
        }
        snippet = this.buffer.slice(start, end);
        return common.repeat(' ', indent) + head + snippet + tail + '\n' +
            common.repeat(' ', indent + this.position - start + head.length) + '^';
    };
    Mark.prototype.toString = function toString(compact) {
        var snippet, where = '';
        if (this.name) {
            where += 'in "' + this.name + '" ';
        }
        where += 'at line ' + (this.line + 1) + ', column ' + (this.column + 1);
        if (!compact) {
            snippet = this.getSnippet();
            if (snippet) {
                where += ':\n' + snippet;
            }
        }
        return where;
    };
    module.exports = Mark;
},
/* js-yaml/lib/js-yaml/schema/default_safe.js */ function _(require, module, exports, __esModule, __esExport) {
    var Schema = require(290) /* ../schema */;
    module.exports = new Schema({
        include: [
            require(292) /* ./core */
        ],
        implicit: [
            require(302) /* ../type/timestamp */,
            require(303) /* ../type/merge */
        ],
        explicit: [
            require(304) /* ../type/binary */,
            require(305) /* ../type/omap */,
            require(306) /* ../type/pairs */,
            require(307) /* ../type/set */
        ]
    });
},
/* js-yaml/lib/js-yaml/schema.js */ function _(require, module, exports, __esModule, __esExport) {
    /*eslint-disable max-len*/
    var common = require(286) /* ./common */;
    var YAMLException = require(287) /* ./exception */;
    var Type = require(291) /* ./type */;
    function compileList(schema, name, result) {
        var exclude = [];
        schema.include.forEach(function (includedSchema) {
            result = compileList(includedSchema, name, result);
        });
        schema[name].forEach(function (currentType) {
            result.forEach(function (previousType, previousIndex) {
                if (previousType.tag === currentType.tag && previousType.kind === currentType.kind) {
                    exclude.push(previousIndex);
                }
            });
            result.push(currentType);
        });
        return result.filter(function (type, index) {
            return exclude.indexOf(index) === -1;
        });
    }
    function compileMap( /* lists... */) {
        var result = {
            scalar: {},
            sequence: {},
            mapping: {},
            fallback: {}
        }, index, length;
        function collectType(type) {
            result[type.kind][type.tag] = result['fallback'][type.tag] = type;
        }
        for (index = 0, length = arguments.length; index < length; index += 1) {
            arguments[index].forEach(collectType);
        }
        return result;
    }
    function Schema(definition) {
        this.include = definition.include || [];
        this.implicit = definition.implicit || [];
        this.explicit = definition.explicit || [];
        this.implicit.forEach(function (type) {
            if (type.loadKind && type.loadKind !== 'scalar') {
                throw new YAMLException('There is a non-scalar type in the implicit list of a schema. Implicit resolving of such types is not supported.');
            }
        });
        this.compiledImplicit = compileList(this, 'implicit', []);
        this.compiledExplicit = compileList(this, 'explicit', []);
        this.compiledTypeMap = compileMap(this.compiledImplicit, this.compiledExplicit);
    }
    Schema.DEFAULT = null;
    Schema.create = function createSchema() {
        var schemas, types;
        switch (arguments.length) {
            case 1:
                schemas = Schema.DEFAULT;
                types = arguments[0];
                break;
            case 2:
                schemas = arguments[0];
                types = arguments[1];
                break;
            default:
                throw new YAMLException('Wrong number of arguments for Schema.create function');
        }
        schemas = common.toArray(schemas);
        types = common.toArray(types);
        if (!schemas.every(function (schema) { return schema instanceof Schema; })) {
            throw new YAMLException('Specified list of super schemas (or a single Schema object) contains a non-Schema object.');
        }
        if (!types.every(function (type) { return type instanceof Type; })) {
            throw new YAMLException('Specified list of YAML types (or a single Type object) contains a non-Type object.');
        }
        return new Schema({
            include: schemas,
            explicit: types
        });
    };
    module.exports = Schema;
},
/* js-yaml/lib/js-yaml/type.js */ function _(require, module, exports, __esModule, __esExport) {
    var YAMLException = require(287) /* ./exception */;
    var TYPE_CONSTRUCTOR_OPTIONS = [
        'kind',
        'resolve',
        'construct',
        'instanceOf',
        'predicate',
        'represent',
        'defaultStyle',
        'styleAliases'
    ];
    var YAML_NODE_KINDS = [
        'scalar',
        'sequence',
        'mapping'
    ];
    function compileStyleAliases(map) {
        var result = {};
        if (map !== null) {
            Object.keys(map).forEach(function (style) {
                map[style].forEach(function (alias) {
                    result[String(alias)] = style;
                });
            });
        }
        return result;
    }
    function Type(tag, options) {
        options = options || {};
        Object.keys(options).forEach(function (name) {
            if (TYPE_CONSTRUCTOR_OPTIONS.indexOf(name) === -1) {
                throw new YAMLException('Unknown option "' + name + '" is met in definition of "' + tag + '" YAML type.');
            }
        });
        // TODO: Add tag format check.
        this.tag = tag;
        this.kind = options['kind'] || null;
        this.resolve = options['resolve'] || function () { return true; };
        this.construct = options['construct'] || function (data) { return data; };
        this.instanceOf = options['instanceOf'] || null;
        this.predicate = options['predicate'] || null;
        this.represent = options['represent'] || null;
        this.defaultStyle = options['defaultStyle'] || null;
        this.styleAliases = compileStyleAliases(options['styleAliases'] || null);
        if (YAML_NODE_KINDS.indexOf(this.kind) === -1) {
            throw new YAMLException('Unknown kind "' + this.kind + '" is specified for "' + tag + '" YAML type.');
        }
    }
    module.exports = Type;
},
/* js-yaml/lib/js-yaml/schema/core.js */ function _(require, module, exports, __esModule, __esExport) {
    var Schema = require(290) /* ../schema */;
    module.exports = new Schema({
        include: [
            require(293) /* ./json */
        ]
    });
},
/* js-yaml/lib/js-yaml/schema/json.js */ function _(require, module, exports, __esModule, __esExport) {
    var Schema = require(290) /* ../schema */;
    module.exports = new Schema({
        include: [
            require(294) /* ./failsafe */
        ],
        implicit: [
            require(298) /* ../type/null */,
            require(299) /* ../type/bool */,
            require(300) /* ../type/int */,
            require(301) /* ../type/float */
        ]
    });
},
/* js-yaml/lib/js-yaml/schema/failsafe.js */ function _(require, module, exports, __esModule, __esExport) {
    var Schema = require(290) /* ../schema */;
    module.exports = new Schema({
        explicit: [
            require(295) /* ../type/str */,
            require(296) /* ../type/seq */,
            require(297) /* ../type/map */
        ]
    });
},
/* js-yaml/lib/js-yaml/type/str.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    module.exports = new Type('tag:yaml.org,2002:str', {
        kind: 'scalar',
        construct: function (data) { return data !== null ? data : ''; }
    });
},
/* js-yaml/lib/js-yaml/type/seq.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    module.exports = new Type('tag:yaml.org,2002:seq', {
        kind: 'sequence',
        construct: function (data) { return data !== null ? data : []; }
    });
},
/* js-yaml/lib/js-yaml/type/map.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    module.exports = new Type('tag:yaml.org,2002:map', {
        kind: 'mapping',
        construct: function (data) { return data !== null ? data : {}; }
    });
},
/* js-yaml/lib/js-yaml/type/null.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    function resolveYamlNull(data) {
        if (data === null)
            return true;
        var max = data.length;
        return (max === 1 && data === '~') ||
            (max === 4 && (data === 'null' || data === 'Null' || data === 'NULL'));
    }
    function constructYamlNull() {
        return null;
    }
    function isNull(object) {
        return object === null;
    }
    module.exports = new Type('tag:yaml.org,2002:null', {
        kind: 'scalar',
        resolve: resolveYamlNull,
        construct: constructYamlNull,
        predicate: isNull,
        represent: {
            canonical: function () { return '~'; },
            lowercase: function () { return 'null'; },
            uppercase: function () { return 'NULL'; },
            camelcase: function () { return 'Null'; }
        },
        defaultStyle: 'lowercase'
    });
},
/* js-yaml/lib/js-yaml/type/bool.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    function resolveYamlBoolean(data) {
        if (data === null)
            return false;
        var max = data.length;
        return (max === 4 && (data === 'true' || data === 'True' || data === 'TRUE')) ||
            (max === 5 && (data === 'false' || data === 'False' || data === 'FALSE'));
    }
    function constructYamlBoolean(data) {
        return data === 'true' ||
            data === 'True' ||
            data === 'TRUE';
    }
    function isBoolean(object) {
        return Object.prototype.toString.call(object) === '[object Boolean]';
    }
    module.exports = new Type('tag:yaml.org,2002:bool', {
        kind: 'scalar',
        resolve: resolveYamlBoolean,
        construct: constructYamlBoolean,
        predicate: isBoolean,
        represent: {
            lowercase: function (object) { return object ? 'true' : 'false'; },
            uppercase: function (object) { return object ? 'TRUE' : 'FALSE'; },
            camelcase: function (object) { return object ? 'True' : 'False'; }
        },
        defaultStyle: 'lowercase'
    });
},
/* js-yaml/lib/js-yaml/type/int.js */ function _(require, module, exports, __esModule, __esExport) {
    var common = require(286) /* ../common */;
    var Type = require(291) /* ../type */;
    function isHexCode(c) {
        return ((0x30 /* 0 */ <= c) && (c <= 0x39 /* 9 */)) ||
            ((0x41 /* A */ <= c) && (c <= 0x46 /* F */)) ||
            ((0x61 /* a */ <= c) && (c <= 0x66 /* f */));
    }
    function isOctCode(c) {
        return ((0x30 /* 0 */ <= c) && (c <= 0x37 /* 7 */));
    }
    function isDecCode(c) {
        return ((0x30 /* 0 */ <= c) && (c <= 0x39 /* 9 */));
    }
    function resolveYamlInteger(data) {
        if (data === null)
            return false;
        var max = data.length, index = 0, hasDigits = false, ch;
        if (!max)
            return false;
        ch = data[index];
        // sign
        if (ch === '-' || ch === '+') {
            ch = data[++index];
        }
        if (ch === '0') {
            // 0
            if (index + 1 === max)
                return true;
            ch = data[++index];
            // base 2, base 8, base 16
            if (ch === 'b') {
                // base 2
                index++;
                for (; index < max; index++) {
                    ch = data[index];
                    if (ch === '_')
                        continue;
                    if (ch !== '0' && ch !== '1')
                        return false;
                    hasDigits = true;
                }
                return hasDigits && ch !== '_';
            }
            if (ch === 'x') {
                // base 16
                index++;
                for (; index < max; index++) {
                    ch = data[index];
                    if (ch === '_')
                        continue;
                    if (!isHexCode(data.charCodeAt(index)))
                        return false;
                    hasDigits = true;
                }
                return hasDigits && ch !== '_';
            }
            // base 8
            for (; index < max; index++) {
                ch = data[index];
                if (ch === '_')
                    continue;
                if (!isOctCode(data.charCodeAt(index)))
                    return false;
                hasDigits = true;
            }
            return hasDigits && ch !== '_';
        }
        // base 10 (except 0) or base 60
        // value should not start with `_`;
        if (ch === '_')
            return false;
        for (; index < max; index++) {
            ch = data[index];
            if (ch === '_')
                continue;
            if (ch === ':')
                break;
            if (!isDecCode(data.charCodeAt(index))) {
                return false;
            }
            hasDigits = true;
        }
        // Should have digits and should not end with `_`
        if (!hasDigits || ch === '_')
            return false;
        // if !base60 - done;
        if (ch !== ':')
            return true;
        // base60 almost not used, no needs to optimize
        return /^(:[0-5]?[0-9])+$/.test(data.slice(index));
    }
    function constructYamlInteger(data) {
        var value = data, sign = 1, ch, base, digits = [];
        if (value.indexOf('_') !== -1) {
            value = value.replace(/_/g, '');
        }
        ch = value[0];
        if (ch === '-' || ch === '+') {
            if (ch === '-')
                sign = -1;
            value = value.slice(1);
            ch = value[0];
        }
        if (value === '0')
            return 0;
        if (ch === '0') {
            if (value[1] === 'b')
                return sign * parseInt(value.slice(2), 2);
            if (value[1] === 'x')
                return sign * parseInt(value, 16);
            return sign * parseInt(value, 8);
        }
        if (value.indexOf(':') !== -1) {
            value.split(':').forEach(function (v) {
                digits.unshift(parseInt(v, 10));
            });
            value = 0;
            base = 1;
            digits.forEach(function (d) {
                value += (d * base);
                base *= 60;
            });
            return sign * value;
        }
        return sign * parseInt(value, 10);
    }
    function isInteger(object) {
        return (Object.prototype.toString.call(object)) === '[object Number]' &&
            (object % 1 === 0 && !common.isNegativeZero(object));
    }
    module.exports = new Type('tag:yaml.org,2002:int', {
        kind: 'scalar',
        resolve: resolveYamlInteger,
        construct: constructYamlInteger,
        predicate: isInteger,
        represent: {
            binary: function (obj) { return obj >= 0 ? '0b' + obj.toString(2) : '-0b' + obj.toString(2).slice(1); },
            octal: function (obj) { return obj >= 0 ? '0' + obj.toString(8) : '-0' + obj.toString(8).slice(1); },
            decimal: function (obj) { return obj.toString(10); },
            /* eslint-disable max-len */
            hexadecimal: function (obj) { return obj >= 0 ? '0x' + obj.toString(16).toUpperCase() : '-0x' + obj.toString(16).toUpperCase().slice(1); }
        },
        defaultStyle: 'decimal',
        styleAliases: {
            binary: [2, 'bin'],
            octal: [8, 'oct'],
            decimal: [10, 'dec'],
            hexadecimal: [16, 'hex']
        }
    });
},
/* js-yaml/lib/js-yaml/type/float.js */ function _(require, module, exports, __esModule, __esExport) {
    var common = require(286) /* ../common */;
    var Type = require(291) /* ../type */;
    var YAML_FLOAT_PATTERN = new RegExp(
    // 2.5e4, 2.5 and integers
    '^(?:[-+]?(?:0|[1-9][0-9_]*)(?:\\.[0-9_]*)?(?:[eE][-+]?[0-9]+)?' +
        // .2e4, .2
        // special case, seems not from spec
        '|\\.[0-9_]+(?:[eE][-+]?[0-9]+)?' +
        // 20:59
        '|[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*' +
        // .inf
        '|[-+]?\\.(?:inf|Inf|INF)' +
        // .nan
        '|\\.(?:nan|NaN|NAN))$');
    function resolveYamlFloat(data) {
        if (data === null)
            return false;
        if (!YAML_FLOAT_PATTERN.test(data) ||
            // Quick hack to not allow integers end with `_`
            // Probably should update regexp & check speed
            data[data.length - 1] === '_') {
            return false;
        }
        return true;
    }
    function constructYamlFloat(data) {
        var value, sign, base, digits;
        value = data.replace(/_/g, '').toLowerCase();
        sign = value[0] === '-' ? -1 : 1;
        digits = [];
        if ('+-'.indexOf(value[0]) >= 0) {
            value = value.slice(1);
        }
        if (value === '.inf') {
            return (sign === 1) ? Number.POSITIVE_INFINITY : Number.NEGATIVE_INFINITY;
        }
        else if (value === '.nan') {
            return NaN;
        }
        else if (value.indexOf(':') >= 0) {
            value.split(':').forEach(function (v) {
                digits.unshift(parseFloat(v, 10));
            });
            value = 0.0;
            base = 1;
            digits.forEach(function (d) {
                value += d * base;
                base *= 60;
            });
            return sign * value;
        }
        return sign * parseFloat(value, 10);
    }
    var SCIENTIFIC_WITHOUT_DOT = /^[-+]?[0-9]+e/;
    function representYamlFloat(object, style) {
        var res;
        if (isNaN(object)) {
            switch (style) {
                case 'lowercase': return '.nan';
                case 'uppercase': return '.NAN';
                case 'camelcase': return '.NaN';
            }
        }
        else if (Number.POSITIVE_INFINITY === object) {
            switch (style) {
                case 'lowercase': return '.inf';
                case 'uppercase': return '.INF';
                case 'camelcase': return '.Inf';
            }
        }
        else if (Number.NEGATIVE_INFINITY === object) {
            switch (style) {
                case 'lowercase': return '-.inf';
                case 'uppercase': return '-.INF';
                case 'camelcase': return '-.Inf';
            }
        }
        else if (common.isNegativeZero(object)) {
            return '-0.0';
        }
        res = object.toString(10);
        // JS stringifier can build scientific format without dots: 5e-100,
        // while YAML requres dot: 5.e-100. Fix it with simple hack
        return SCIENTIFIC_WITHOUT_DOT.test(res) ? res.replace('e', '.e') : res;
    }
    function isFloat(object) {
        return (Object.prototype.toString.call(object) === '[object Number]') &&
            (object % 1 !== 0 || common.isNegativeZero(object));
    }
    module.exports = new Type('tag:yaml.org,2002:float', {
        kind: 'scalar',
        resolve: resolveYamlFloat,
        construct: constructYamlFloat,
        predicate: isFloat,
        represent: representYamlFloat,
        defaultStyle: 'lowercase'
    });
},
/* js-yaml/lib/js-yaml/type/timestamp.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    var YAML_DATE_REGEXP = new RegExp('^([0-9][0-9][0-9][0-9])' + // [1] year
        '-([0-9][0-9])' + // [2] month
        '-([0-9][0-9])$'); // [3] day
    var YAML_TIMESTAMP_REGEXP = new RegExp('^([0-9][0-9][0-9][0-9])' + // [1] year
        '-([0-9][0-9]?)' + // [2] month
        '-([0-9][0-9]?)' + // [3] day
        '(?:[Tt]|[ \\t]+)' + // ...
        '([0-9][0-9]?)' + // [4] hour
        ':([0-9][0-9])' + // [5] minute
        ':([0-9][0-9])' + // [6] second
        '(?:\\.([0-9]*))?' + // [7] fraction
        '(?:[ \\t]*(Z|([-+])([0-9][0-9]?)' + // [8] tz [9] tz_sign [10] tz_hour
        '(?::([0-9][0-9]))?))?$'); // [11] tz_minute
    function resolveYamlTimestamp(data) {
        if (data === null)
            return false;
        if (YAML_DATE_REGEXP.exec(data) !== null)
            return true;
        if (YAML_TIMESTAMP_REGEXP.exec(data) !== null)
            return true;
        return false;
    }
    function constructYamlTimestamp(data) {
        var match, year, month, day, hour, minute, second, fraction = 0, delta = null, tz_hour, tz_minute, date;
        match = YAML_DATE_REGEXP.exec(data);
        if (match === null)
            match = YAML_TIMESTAMP_REGEXP.exec(data);
        if (match === null)
            throw new Error('Date resolve error');
        // match: [1] year [2] month [3] day
        year = +(match[1]);
        month = +(match[2]) - 1; // JS month starts with 0
        day = +(match[3]);
        if (!match[4]) { // no hour
            return new Date(Date.UTC(year, month, day));
        }
        // match: [4] hour [5] minute [6] second [7] fraction
        hour = +(match[4]);
        minute = +(match[5]);
        second = +(match[6]);
        if (match[7]) {
            fraction = match[7].slice(0, 3);
            while (fraction.length < 3) { // milli-seconds
                fraction += '0';
            }
            fraction = +fraction;
        }
        // match: [8] tz [9] tz_sign [10] tz_hour [11] tz_minute
        if (match[9]) {
            tz_hour = +(match[10]);
            tz_minute = +(match[11] || 0);
            delta = (tz_hour * 60 + tz_minute) * 60000; // delta in mili-seconds
            if (match[9] === '-')
                delta = -delta;
        }
        date = new Date(Date.UTC(year, month, day, hour, minute, second, fraction));
        if (delta)
            date.setTime(date.getTime() - delta);
        return date;
    }
    function representYamlTimestamp(object /*, style*/) {
        return object.toISOString();
    }
    module.exports = new Type('tag:yaml.org,2002:timestamp', {
        kind: 'scalar',
        resolve: resolveYamlTimestamp,
        construct: constructYamlTimestamp,
        instanceOf: Date,
        represent: representYamlTimestamp
    });
},
/* js-yaml/lib/js-yaml/type/merge.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    function resolveYamlMerge(data) {
        return data === '<<' || data === null;
    }
    module.exports = new Type('tag:yaml.org,2002:merge', {
        kind: 'scalar',
        resolve: resolveYamlMerge
    });
},
/* js-yaml/lib/js-yaml/type/binary.js */ function _(require, module, exports, __esModule, __esExport) {
    /*eslint-disable no-bitwise*/
    var NodeBuffer;
    try {
        // A trick for browserified version, to not include `Buffer` shim
        var _require = require;
        NodeBuffer = _require('buffer').Buffer;
    }
    catch (__) { }
    var Type = require(291) /* ../type */;
    // [ 64, 65, 66 ] -> [ padding, CR, LF ]
    var BASE64_MAP = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=\n\r';
    function resolveYamlBinary(data) {
        if (data === null)
            return false;
        var code, idx, bitlen = 0, max = data.length, map = BASE64_MAP;
        // Convert one by one.
        for (idx = 0; idx < max; idx++) {
            code = map.indexOf(data.charAt(idx));
            // Skip CR/LF
            if (code > 64)
                continue;
            // Fail on illegal characters
            if (code < 0)
                return false;
            bitlen += 6;
        }
        // If there are any bits left, source was corrupted
        return (bitlen % 8) === 0;
    }
    function constructYamlBinary(data) {
        var idx, tailbits, input = data.replace(/[\r\n=]/g, ''), // remove CR/LF & padding to simplify scan
        max = input.length, map = BASE64_MAP, bits = 0, result = [];
        // Collect by 6*4 bits (3 bytes)
        for (idx = 0; idx < max; idx++) {
            if ((idx % 4 === 0) && idx) {
                result.push((bits >> 16) & 0xFF);
                result.push((bits >> 8) & 0xFF);
                result.push(bits & 0xFF);
            }
            bits = (bits << 6) | map.indexOf(input.charAt(idx));
        }
        // Dump tail
        tailbits = (max % 4) * 6;
        if (tailbits === 0) {
            result.push((bits >> 16) & 0xFF);
            result.push((bits >> 8) & 0xFF);
            result.push(bits & 0xFF);
        }
        else if (tailbits === 18) {
            result.push((bits >> 10) & 0xFF);
            result.push((bits >> 2) & 0xFF);
        }
        else if (tailbits === 12) {
            result.push((bits >> 4) & 0xFF);
        }
        // Wrap into Buffer for NodeJS and leave Array for browser
        if (NodeBuffer) {
            // Support node 6.+ Buffer API when available
            return NodeBuffer.from ? NodeBuffer.from(result) : new NodeBuffer(result);
        }
        return result;
    }
    function representYamlBinary(object /*, style*/) {
        var result = '', bits = 0, idx, tail, max = object.length, map = BASE64_MAP;
        // Convert every three bytes to 4 ASCII characters.
        for (idx = 0; idx < max; idx++) {
            if ((idx % 3 === 0) && idx) {
                result += map[(bits >> 18) & 0x3F];
                result += map[(bits >> 12) & 0x3F];
                result += map[(bits >> 6) & 0x3F];
                result += map[bits & 0x3F];
            }
            bits = (bits << 8) + object[idx];
        }
        // Dump tail
        tail = max % 3;
        if (tail === 0) {
            result += map[(bits >> 18) & 0x3F];
            result += map[(bits >> 12) & 0x3F];
            result += map[(bits >> 6) & 0x3F];
            result += map[bits & 0x3F];
        }
        else if (tail === 2) {
            result += map[(bits >> 10) & 0x3F];
            result += map[(bits >> 4) & 0x3F];
            result += map[(bits << 2) & 0x3F];
            result += map[64];
        }
        else if (tail === 1) {
            result += map[(bits >> 2) & 0x3F];
            result += map[(bits << 4) & 0x3F];
            result += map[64];
            result += map[64];
        }
        return result;
    }
    function isBinary(object) {
        return NodeBuffer && NodeBuffer.isBuffer(object);
    }
    module.exports = new Type('tag:yaml.org,2002:binary', {
        kind: 'scalar',
        resolve: resolveYamlBinary,
        construct: constructYamlBinary,
        predicate: isBinary,
        represent: representYamlBinary
    });
},
/* js-yaml/lib/js-yaml/type/omap.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    var _hasOwnProperty = Object.prototype.hasOwnProperty;
    var _toString = Object.prototype.toString;
    function resolveYamlOmap(data) {
        if (data === null)
            return true;
        var objectKeys = [], index, length, pair, pairKey, pairHasKey, object = data;
        for (index = 0, length = object.length; index < length; index += 1) {
            pair = object[index];
            pairHasKey = false;
            if (_toString.call(pair) !== '[object Object]')
                return false;
            for (pairKey in pair) {
                if (_hasOwnProperty.call(pair, pairKey)) {
                    if (!pairHasKey)
                        pairHasKey = true;
                    else
                        return false;
                }
            }
            if (!pairHasKey)
                return false;
            if (objectKeys.indexOf(pairKey) === -1)
                objectKeys.push(pairKey);
            else
                return false;
        }
        return true;
    }
    function constructYamlOmap(data) {
        return data !== null ? data : [];
    }
    module.exports = new Type('tag:yaml.org,2002:omap', {
        kind: 'sequence',
        resolve: resolveYamlOmap,
        construct: constructYamlOmap
    });
},
/* js-yaml/lib/js-yaml/type/pairs.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    var _toString = Object.prototype.toString;
    function resolveYamlPairs(data) {
        if (data === null)
            return true;
        var index, length, pair, keys, result, object = data;
        result = new Array(object.length);
        for (index = 0, length = object.length; index < length; index += 1) {
            pair = object[index];
            if (_toString.call(pair) !== '[object Object]')
                return false;
            keys = Object.keys(pair);
            if (keys.length !== 1)
                return false;
            result[index] = [keys[0], pair[keys[0]]];
        }
        return true;
    }
    function constructYamlPairs(data) {
        if (data === null)
            return [];
        var index, length, pair, keys, result, object = data;
        result = new Array(object.length);
        for (index = 0, length = object.length; index < length; index += 1) {
            pair = object[index];
            keys = Object.keys(pair);
            result[index] = [keys[0], pair[keys[0]]];
        }
        return result;
    }
    module.exports = new Type('tag:yaml.org,2002:pairs', {
        kind: 'sequence',
        resolve: resolveYamlPairs,
        construct: constructYamlPairs
    });
},
/* js-yaml/lib/js-yaml/type/set.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../type */;
    var _hasOwnProperty = Object.prototype.hasOwnProperty;
    function resolveYamlSet(data) {
        if (data === null)
            return true;
        var key, object = data;
        for (key in object) {
            if (_hasOwnProperty.call(object, key)) {
                if (object[key] !== null)
                    return false;
            }
        }
        return true;
    }
    function constructYamlSet(data) {
        return data !== null ? data : {};
    }
    module.exports = new Type('tag:yaml.org,2002:set', {
        kind: 'mapping',
        resolve: resolveYamlSet,
        construct: constructYamlSet
    });
},
/* js-yaml/lib/js-yaml/schema/default_full.js */ function _(require, module, exports, __esModule, __esExport) {
    var Schema = require(290) /* ../schema */;
    module.exports = Schema.DEFAULT = new Schema({
        include: [
            require(289) /* ./default_safe */
        ],
        explicit: [
            require(309) /* ../type/js/undefined */,
            require(310) /* ../type/js/regexp */,
            require(311) /* ../type/js/function */
        ]
    });
},
/* js-yaml/lib/js-yaml/type/js/undefined.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../../type */;
    function resolveJavascriptUndefined() {
        return true;
    }
    function constructJavascriptUndefined() {
        /*eslint-disable no-undefined*/
        return undefined;
    }
    function representJavascriptUndefined() {
        return '';
    }
    function isUndefined(object) {
        return typeof object === 'undefined';
    }
    module.exports = new Type('tag:yaml.org,2002:js/undefined', {
        kind: 'scalar',
        resolve: resolveJavascriptUndefined,
        construct: constructJavascriptUndefined,
        predicate: isUndefined,
        represent: representJavascriptUndefined
    });
},
/* js-yaml/lib/js-yaml/type/js/regexp.js */ function _(require, module, exports, __esModule, __esExport) {
    var Type = require(291) /* ../../type */;
    function resolveJavascriptRegExp(data) {
        if (data === null)
            return false;
        if (data.length === 0)
            return false;
        var regexp = data, tail = /\/([gim]*)$/.exec(data), modifiers = '';
        // if regexp starts with '/' it can have modifiers and must be properly closed
        // `/foo/gim` - modifiers tail can be maximum 3 chars
        if (regexp[0] === '/') {
            if (tail)
                modifiers = tail[1];
            if (modifiers.length > 3)
                return false;
            // if expression starts with /, is should be properly terminated
            if (regexp[regexp.length - modifiers.length - 1] !== '/')
                return false;
        }
        return true;
    }
    function constructJavascriptRegExp(data) {
        var regexp = data, tail = /\/([gim]*)$/.exec(data), modifiers = '';
        // `/foo/gim` - tail can be maximum 4 chars
        if (regexp[0] === '/') {
            if (tail)
                modifiers = tail[1];
            regexp = regexp.slice(1, regexp.length - modifiers.length - 1);
        }
        return new RegExp(regexp, modifiers);
    }
    function representJavascriptRegExp(object /*, style*/) {
        var result = '/' + object.source + '/';
        if (object.global)
            result += 'g';
        if (object.multiline)
            result += 'm';
        if (object.ignoreCase)
            result += 'i';
        return result;
    }
    function isRegExp(object) {
        return Object.prototype.toString.call(object) === '[object RegExp]';
    }
    module.exports = new Type('tag:yaml.org,2002:js/regexp', {
        kind: 'scalar',
        resolve: resolveJavascriptRegExp,
        construct: constructJavascriptRegExp,
        predicate: isRegExp,
        represent: representJavascriptRegExp
    });
},
/* js-yaml/lib/js-yaml/type/js/function.js */ function _(require, module, exports, __esModule, __esExport) {
    var esprima;
    // Browserified version does not have esprima
    //
    // 1. For node.js just require module as deps
    // 2. For browser try to require mudule via external AMD system.
    //    If not found - try to fallback to window.esprima. If not
    //    found too - then fail to parse.
    //
    try {
        // workaround to exclude package from browserify list.
        var _require = require;
        esprima = _require('esprima');
    }
    catch (_) {
        /* eslint-disable no-redeclare */
        /* global window */
        if (typeof window !== 'undefined')
            esprima = window.esprima;
    }
    var Type = require(291) /* ../../type */;
    function resolveJavascriptFunction(data) {
        if (data === null)
            return false;
        try {
            var source = '(' + data + ')', ast = esprima.parse(source, { range: true });
            if (ast.type !== 'Program' ||
                ast.body.length !== 1 ||
                ast.body[0].type !== 'ExpressionStatement' ||
                (ast.body[0].expression.type !== 'ArrowFunctionExpression' &&
                    ast.body[0].expression.type !== 'FunctionExpression')) {
                return false;
            }
            return true;
        }
        catch (err) {
            return false;
        }
    }
    function constructJavascriptFunction(data) {
        /*jslint evil:true*/
        var source = '(' + data + ')', ast = esprima.parse(source, { range: true }), params = [], body;
        if (ast.type !== 'Program' ||
            ast.body.length !== 1 ||
            ast.body[0].type !== 'ExpressionStatement' ||
            (ast.body[0].expression.type !== 'ArrowFunctionExpression' &&
                ast.body[0].expression.type !== 'FunctionExpression')) {
            throw new Error('Failed to resolve function');
        }
        ast.body[0].expression.params.forEach(function (param) {
            params.push(param.name);
        });
        body = ast.body[0].expression.body.range;
        // Esprima's ranges include the first '{' and the last '}' characters on
        // function expressions. So cut them out.
        if (ast.body[0].expression.body.type === 'BlockStatement') {
            /*eslint-disable no-new-func*/
            return new Function(params, source.slice(body[0] + 1, body[1] - 1));
        }
        // ES6 arrow functions can omit the BlockStatement. In that case, just return
        // the body.
        /*eslint-disable no-new-func*/
        return new Function(params, 'return ' + source.slice(body[0], body[1]));
    }
    function representJavascriptFunction(object /*, style*/) {
        return object.toString();
    }
    function isFunction(object) {
        return Object.prototype.toString.call(object) === '[object Function]';
    }
    module.exports = new Type('tag:yaml.org,2002:js/function', {
        kind: 'scalar',
        resolve: resolveJavascriptFunction,
        construct: constructJavascriptFunction,
        predicate: isFunction,
        represent: representJavascriptFunction
    });
},
/* js-yaml/lib/js-yaml/dumper.js */ function _(require, module, exports, __esModule, __esExport) {
    /*eslint-disable no-use-before-define*/
    var common = require(286) /* ./common */;
    var YAMLException = require(287) /* ./exception */;
    var DEFAULT_FULL_SCHEMA = require(308) /* ./schema/default_full */;
    var DEFAULT_SAFE_SCHEMA = require(289) /* ./schema/default_safe */;
    var _toString = Object.prototype.toString;
    var _hasOwnProperty = Object.prototype.hasOwnProperty;
    var CHAR_TAB = 0x09; /* Tab */
    var CHAR_LINE_FEED = 0x0A; /* LF */
    var CHAR_CARRIAGE_RETURN = 0x0D; /* CR */
    var CHAR_SPACE = 0x20; /* Space */
    var CHAR_EXCLAMATION = 0x21; /* ! */
    var CHAR_DOUBLE_QUOTE = 0x22; /* " */
    var CHAR_SHARP = 0x23; /* # */
    var CHAR_PERCENT = 0x25; /* % */
    var CHAR_AMPERSAND = 0x26; /* & */
    var CHAR_SINGLE_QUOTE = 0x27; /* ' */
    var CHAR_ASTERISK = 0x2A; /* * */
    var CHAR_COMMA = 0x2C; /* , */
    var CHAR_MINUS = 0x2D; /* - */
    var CHAR_COLON = 0x3A; /* : */
    var CHAR_EQUALS = 0x3D; /* = */
    var CHAR_GREATER_THAN = 0x3E; /* > */
    var CHAR_QUESTION = 0x3F; /* ? */
    var CHAR_COMMERCIAL_AT = 0x40; /* @ */
    var CHAR_LEFT_SQUARE_BRACKET = 0x5B; /* [ */
    var CHAR_RIGHT_SQUARE_BRACKET = 0x5D; /* ] */
    var CHAR_GRAVE_ACCENT = 0x60; /* ` */
    var CHAR_LEFT_CURLY_BRACKET = 0x7B; /* { */
    var CHAR_VERTICAL_LINE = 0x7C; /* | */
    var CHAR_RIGHT_CURLY_BRACKET = 0x7D; /* } */
    var ESCAPE_SEQUENCES = {};
    ESCAPE_SEQUENCES[0x00] = '\\0';
    ESCAPE_SEQUENCES[0x07] = '\\a';
    ESCAPE_SEQUENCES[0x08] = '\\b';
    ESCAPE_SEQUENCES[0x09] = '\\t';
    ESCAPE_SEQUENCES[0x0A] = '\\n';
    ESCAPE_SEQUENCES[0x0B] = '\\v';
    ESCAPE_SEQUENCES[0x0C] = '\\f';
    ESCAPE_SEQUENCES[0x0D] = '\\r';
    ESCAPE_SEQUENCES[0x1B] = '\\e';
    ESCAPE_SEQUENCES[0x22] = '\\"';
    ESCAPE_SEQUENCES[0x5C] = '\\\\';
    ESCAPE_SEQUENCES[0x85] = '\\N';
    ESCAPE_SEQUENCES[0xA0] = '\\_';
    ESCAPE_SEQUENCES[0x2028] = '\\L';
    ESCAPE_SEQUENCES[0x2029] = '\\P';
    var DEPRECATED_BOOLEANS_SYNTAX = [
        'y', 'Y', 'yes', 'Yes', 'YES', 'on', 'On', 'ON',
        'n', 'N', 'no', 'No', 'NO', 'off', 'Off', 'OFF'
    ];
    function compileStyleMap(schema, map) {
        var result, keys, index, length, tag, style, type;
        if (map === null)
            return {};
        result = {};
        keys = Object.keys(map);
        for (index = 0, length = keys.length; index < length; index += 1) {
            tag = keys[index];
            style = String(map[tag]);
            if (tag.slice(0, 2) === '!!') {
                tag = 'tag:yaml.org,2002:' + tag.slice(2);
            }
            type = schema.compiledTypeMap['fallback'][tag];
            if (type && _hasOwnProperty.call(type.styleAliases, style)) {
                style = type.styleAliases[style];
            }
            result[tag] = style;
        }
        return result;
    }
    function encodeHex(character) {
        var string, handle, length;
        string = character.toString(16).toUpperCase();
        if (character <= 0xFF) {
            handle = 'x';
            length = 2;
        }
        else if (character <= 0xFFFF) {
            handle = 'u';
            length = 4;
        }
        else if (character <= 0xFFFFFFFF) {
            handle = 'U';
            length = 8;
        }
        else {
            throw new YAMLException('code point within a string may not be greater than 0xFFFFFFFF');
        }
        return '\\' + handle + common.repeat('0', length - string.length) + string;
    }
    function State(options) {
        this.schema = options['schema'] || DEFAULT_FULL_SCHEMA;
        this.indent = Math.max(1, (options['indent'] || 2));
        this.noArrayIndent = options['noArrayIndent'] || false;
        this.skipInvalid = options['skipInvalid'] || false;
        this.flowLevel = (common.isNothing(options['flowLevel']) ? -1 : options['flowLevel']);
        this.styleMap = compileStyleMap(this.schema, options['styles'] || null);
        this.sortKeys = options['sortKeys'] || false;
        this.lineWidth = options['lineWidth'] || 80;
        this.noRefs = options['noRefs'] || false;
        this.noCompatMode = options['noCompatMode'] || false;
        this.condenseFlow = options['condenseFlow'] || false;
        this.implicitTypes = this.schema.compiledImplicit;
        this.explicitTypes = this.schema.compiledExplicit;
        this.tag = null;
        this.result = '';
        this.duplicates = [];
        this.usedDuplicates = null;
    }
    // Indents every line in a string. Empty lines (\n only) are not indented.
    function indentString(string, spaces) {
        var ind = common.repeat(' ', spaces), position = 0, next = -1, result = '', line, length = string.length;
        while (position < length) {
            next = string.indexOf('\n', position);
            if (next === -1) {
                line = string.slice(position);
                position = length;
            }
            else {
                line = string.slice(position, next + 1);
                position = next + 1;
            }
            if (line.length && line !== '\n')
                result += ind;
            result += line;
        }
        return result;
    }
    function generateNextLine(state, level) {
        return '\n' + common.repeat(' ', state.indent * level);
    }
    function testImplicitResolving(state, str) {
        var index, length, type;
        for (index = 0, length = state.implicitTypes.length; index < length; index += 1) {
            type = state.implicitTypes[index];
            if (type.resolve(str)) {
                return true;
            }
        }
        return false;
    }
    // [33] s-white ::= s-space | s-tab
    function isWhitespace(c) {
        return c === CHAR_SPACE || c === CHAR_TAB;
    }
    // Returns true if the character can be printed without escaping.
    // From YAML 1.2: "any allowed characters known to be non-printable
    // should also be escaped. [However,] This isn’t mandatory"
    // Derived from nb-char - \t - #x85 - #xA0 - #x2028 - #x2029.
    function isPrintable(c) {
        return (0x00020 <= c && c <= 0x00007E)
            || ((0x000A1 <= c && c <= 0x00D7FF) && c !== 0x2028 && c !== 0x2029)
            || ((0x0E000 <= c && c <= 0x00FFFD) && c !== 0xFEFF /* BOM */)
            || (0x10000 <= c && c <= 0x10FFFF);
    }
    // [34] ns-char ::= nb-char - s-white
    // [27] nb-char ::= c-printable - b-char - c-byte-order-mark
    // [26] b-char  ::= b-line-feed | b-carriage-return
    // [24] b-line-feed       ::=     #xA    /* LF */
    // [25] b-carriage-return ::=     #xD    /* CR */
    // [3]  c-byte-order-mark ::=     #xFEFF
    function isNsChar(c) {
        return isPrintable(c) && !isWhitespace(c)
            // byte-order-mark
            && c !== 0xFEFF
            // b-char
            && c !== CHAR_CARRIAGE_RETURN
            && c !== CHAR_LINE_FEED;
    }
    // Simplified test for values allowed after the first character in plain style.
    function isPlainSafe(c, prev) {
        // Uses a subset of nb-char - c-flow-indicator - ":" - "#"
        // where nb-char ::= c-printable - b-char - c-byte-order-mark.
        return isPrintable(c) && c !== 0xFEFF
            // - c-flow-indicator
            && c !== CHAR_COMMA
            && c !== CHAR_LEFT_SQUARE_BRACKET
            && c !== CHAR_RIGHT_SQUARE_BRACKET
            && c !== CHAR_LEFT_CURLY_BRACKET
            && c !== CHAR_RIGHT_CURLY_BRACKET
            // - ":" - "#"
            // /* An ns-char preceding */ "#"
            && c !== CHAR_COLON
            && ((c !== CHAR_SHARP) || (prev && isNsChar(prev)));
    }
    // Simplified test for values allowed as the first character in plain style.
    function isPlainSafeFirst(c) {
        // Uses a subset of ns-char - c-indicator
        // where ns-char = nb-char - s-white.
        return isPrintable(c) && c !== 0xFEFF
            && !isWhitespace(c) // - s-white
            // - (c-indicator ::=
            // “-” | “?” | “:” | “,” | “[” | “]” | “{” | “}”
            && c !== CHAR_MINUS
            && c !== CHAR_QUESTION
            && c !== CHAR_COLON
            && c !== CHAR_COMMA
            && c !== CHAR_LEFT_SQUARE_BRACKET
            && c !== CHAR_RIGHT_SQUARE_BRACKET
            && c !== CHAR_LEFT_CURLY_BRACKET
            && c !== CHAR_RIGHT_CURLY_BRACKET
            // | “#” | “&” | “*” | “!” | “|” | “=” | “>” | “'” | “"”
            && c !== CHAR_SHARP
            && c !== CHAR_AMPERSAND
            && c !== CHAR_ASTERISK
            && c !== CHAR_EXCLAMATION
            && c !== CHAR_VERTICAL_LINE
            && c !== CHAR_EQUALS
            && c !== CHAR_GREATER_THAN
            && c !== CHAR_SINGLE_QUOTE
            && c !== CHAR_DOUBLE_QUOTE
            // | “%” | “@” | “`”)
            && c !== CHAR_PERCENT
            && c !== CHAR_COMMERCIAL_AT
            && c !== CHAR_GRAVE_ACCENT;
    }
    // Determines whether block indentation indicator is required.
    function needIndentIndicator(string) {
        var leadingSpaceRe = /^\n* /;
        return leadingSpaceRe.test(string);
    }
    var STYLE_PLAIN = 1, STYLE_SINGLE = 2, STYLE_LITERAL = 3, STYLE_FOLDED = 4, STYLE_DOUBLE = 5;
    // Determines which scalar styles are possible and returns the preferred style.
    // lineWidth = -1 => no limit.
    // Pre-conditions: str.length > 0.
    // Post-conditions:
    //    STYLE_PLAIN or STYLE_SINGLE => no \n are in the string.
    //    STYLE_LITERAL => no lines are suitable for folding (or lineWidth is -1).
    //    STYLE_FOLDED => a line > lineWidth and can be folded (and lineWidth != -1).
    function chooseScalarStyle(string, singleLineOnly, indentPerLevel, lineWidth, testAmbiguousType) {
        var i;
        var char, prev_char;
        var hasLineBreak = false;
        var hasFoldableLine = false; // only checked if shouldTrackWidth
        var shouldTrackWidth = lineWidth !== -1;
        var previousLineBreak = -1; // count the first line correctly
        var plain = isPlainSafeFirst(string.charCodeAt(0))
            && !isWhitespace(string.charCodeAt(string.length - 1));
        if (singleLineOnly) {
            // Case: no block styles.
            // Check for disallowed characters to rule out plain and single.
            for (i = 0; i < string.length; i++) {
                char = string.charCodeAt(i);
                if (!isPrintable(char)) {
                    return STYLE_DOUBLE;
                }
                prev_char = i > 0 ? string.charCodeAt(i - 1) : null;
                plain = plain && isPlainSafe(char, prev_char);
            }
        }
        else {
            // Case: block styles permitted.
            for (i = 0; i < string.length; i++) {
                char = string.charCodeAt(i);
                if (char === CHAR_LINE_FEED) {
                    hasLineBreak = true;
                    // Check if any line can be folded.
                    if (shouldTrackWidth) {
                        hasFoldableLine = hasFoldableLine ||
                            // Foldable line = too long, and not more-indented.
                            (i - previousLineBreak - 1 > lineWidth &&
                                string[previousLineBreak + 1] !== ' ');
                        previousLineBreak = i;
                    }
                }
                else if (!isPrintable(char)) {
                    return STYLE_DOUBLE;
                }
                prev_char = i > 0 ? string.charCodeAt(i - 1) : null;
                plain = plain && isPlainSafe(char, prev_char);
            }
            // in case the end is missing a \n
            hasFoldableLine = hasFoldableLine || (shouldTrackWidth &&
                (i - previousLineBreak - 1 > lineWidth &&
                    string[previousLineBreak + 1] !== ' '));
        }
        // Although every style can represent \n without escaping, prefer block styles
        // for multiline, since they're more readable and they don't add empty lines.
        // Also prefer folding a super-long line.
        if (!hasLineBreak && !hasFoldableLine) {
            // Strings interpretable as another type have to be quoted;
            // e.g. the string 'true' vs. the boolean true.
            return plain && !testAmbiguousType(string)
                ? STYLE_PLAIN : STYLE_SINGLE;
        }
        // Edge case: block indentation indicator can only have one digit.
        if (indentPerLevel > 9 && needIndentIndicator(string)) {
            return STYLE_DOUBLE;
        }
        // At this point we know block styles are valid.
        // Prefer literal style unless we want to fold.
        return hasFoldableLine ? STYLE_FOLDED : STYLE_LITERAL;
    }
    // Note: line breaking/folding is implemented for only the folded style.
    // NB. We drop the last trailing newline (if any) of a returned block scalar
    //  since the dumper adds its own newline. This always works:
    //    • No ending newline => unaffected; already using strip "-" chomping.
    //    • Ending newline    => removed then restored.
    //  Importantly, this keeps the "+" chomp indicator from gaining an extra line.
    function writeScalar(state, string, level, iskey) {
        state.dump = (function () {
            if (string.length === 0) {
                return "''";
            }
            if (!state.noCompatMode &&
                DEPRECATED_BOOLEANS_SYNTAX.indexOf(string) !== -1) {
                return "'" + string + "'";
            }
            var indent = state.indent * Math.max(1, level); // no 0-indent scalars
            // As indentation gets deeper, let the width decrease monotonically
            // to the lower bound min(state.lineWidth, 40).
            // Note that this implies
            //  state.lineWidth ≤ 40 + state.indent: width is fixed at the lower bound.
            //  state.lineWidth > 40 + state.indent: width decreases until the lower bound.
            // This behaves better than a constant minimum width which disallows narrower options,
            // or an indent threshold which causes the width to suddenly increase.
            var lineWidth = state.lineWidth === -1
                ? -1 : Math.max(Math.min(state.lineWidth, 40), state.lineWidth - indent);
            // Without knowing if keys are implicit/explicit, assume implicit for safety.
            var singleLineOnly = iskey
                // No block styles in flow mode.
                || (state.flowLevel > -1 && level >= state.flowLevel);
            function testAmbiguity(string) {
                return testImplicitResolving(state, string);
            }
            switch (chooseScalarStyle(string, singleLineOnly, state.indent, lineWidth, testAmbiguity)) {
                case STYLE_PLAIN:
                    return string;
                case STYLE_SINGLE:
                    return "'" + string.replace(/'/g, "''") + "'";
                case STYLE_LITERAL:
                    return '|' + blockHeader(string, state.indent)
                        + dropEndingNewline(indentString(string, indent));
                case STYLE_FOLDED:
                    return '>' + blockHeader(string, state.indent)
                        + dropEndingNewline(indentString(foldString(string, lineWidth), indent));
                case STYLE_DOUBLE:
                    return '"' + escapeString(string, lineWidth) + '"';
                default:
                    throw new YAMLException('impossible error: invalid scalar style');
            }
        }());
    }
    // Pre-conditions: string is valid for a block scalar, 1 <= indentPerLevel <= 9.
    function blockHeader(string, indentPerLevel) {
        var indentIndicator = needIndentIndicator(string) ? String(indentPerLevel) : '';
        // note the special case: the string '\n' counts as a "trailing" empty line.
        var clip = string[string.length - 1] === '\n';
        var keep = clip && (string[string.length - 2] === '\n' || string === '\n');
        var chomp = keep ? '+' : (clip ? '' : '-');
        return indentIndicator + chomp + '\n';
    }
    // (See the note for writeScalar.)
    function dropEndingNewline(string) {
        return string[string.length - 1] === '\n' ? string.slice(0, -1) : string;
    }
    // Note: a long line without a suitable break point will exceed the width limit.
    // Pre-conditions: every char in str isPrintable, str.length > 0, width > 0.
    function foldString(string, width) {
        // In folded style, $k$ consecutive newlines output as $k+1$ newlines—
        // unless they're before or after a more-indented line, or at the very
        // beginning or end, in which case $k$ maps to $k$.
        // Therefore, parse each chunk as newline(s) followed by a content line.
        var lineRe = /(\n+)([^\n]*)/g;
        // first line (possibly an empty line)
        var result = (function () {
            var nextLF = string.indexOf('\n');
            nextLF = nextLF !== -1 ? nextLF : string.length;
            lineRe.lastIndex = nextLF;
            return foldLine(string.slice(0, nextLF), width);
        }());
        // If we haven't reached the first content line yet, don't add an extra \n.
        var prevMoreIndented = string[0] === '\n' || string[0] === ' ';
        var moreIndented;
        // rest of the lines
        var match;
        while ((match = lineRe.exec(string))) {
            var prefix = match[1], line = match[2];
            moreIndented = (line[0] === ' ');
            result += prefix
                + (!prevMoreIndented && !moreIndented && line !== ''
                    ? '\n' : '')
                + foldLine(line, width);
            prevMoreIndented = moreIndented;
        }
        return result;
    }
    // Greedy line breaking.
    // Picks the longest line under the limit each time,
    // otherwise settles for the shortest line over the limit.
    // NB. More-indented lines *cannot* be folded, as that would add an extra \n.
    function foldLine(line, width) {
        if (line === '' || line[0] === ' ')
            return line;
        // Since a more-indented line adds a \n, breaks can't be followed by a space.
        var breakRe = / [^ ]/g; // note: the match index will always be <= length-2.
        var match;
        // start is an inclusive index. end, curr, and next are exclusive.
        var start = 0, end, curr = 0, next = 0;
        var result = '';
        // Invariants: 0 <= start <= length-1.
        //   0 <= curr <= next <= max(0, length-2). curr - start <= width.
        // Inside the loop:
        //   A match implies length >= 2, so curr and next are <= length-2.
        while ((match = breakRe.exec(line))) {
            next = match.index;
            // maintain invariant: curr - start <= width
            if (next - start > width) {
                end = (curr > start) ? curr : next; // derive end <= length-2
                result += '\n' + line.slice(start, end);
                // skip the space that was output as \n
                start = end + 1; // derive start <= length-1
            }
            curr = next;
        }
        // By the invariants, start <= length-1, so there is something left over.
        // It is either the whole string or a part starting from non-whitespace.
        result += '\n';
        // Insert a break if the remainder is too long and there is a break available.
        if (line.length - start > width && curr > start) {
            result += line.slice(start, curr) + '\n' + line.slice(curr + 1);
        }
        else {
            result += line.slice(start);
        }
        return result.slice(1); // drop extra \n joiner
    }
    // Escapes a double-quoted string.
    function escapeString(string) {
        var result = '';
        var char, nextChar;
        var escapeSeq;
        for (var i = 0; i < string.length; i++) {
            char = string.charCodeAt(i);
            // Check for surrogate pairs (reference Unicode 3.0 section "3.7 Surrogates").
            if (char >= 0xD800 && char <= 0xDBFF /* high surrogate */) {
                nextChar = string.charCodeAt(i + 1);
                if (nextChar >= 0xDC00 && nextChar <= 0xDFFF /* low surrogate */) {
                    // Combine the surrogate pair and store it escaped.
                    result += encodeHex((char - 0xD800) * 0x400 + nextChar - 0xDC00 + 0x10000);
                    // Advance index one extra since we already used that char here.
                    i++;
                    continue;
                }
            }
            escapeSeq = ESCAPE_SEQUENCES[char];
            result += !escapeSeq && isPrintable(char)
                ? string[i]
                : escapeSeq || encodeHex(char);
        }
        return result;
    }
    function writeFlowSequence(state, level, object) {
        var _result = '', _tag = state.tag, index, length;
        for (index = 0, length = object.length; index < length; index += 1) {
            // Write only valid elements.
            if (writeNode(state, level, object[index], false, false)) {
                if (index !== 0)
                    _result += ',' + (!state.condenseFlow ? ' ' : '');
                _result += state.dump;
            }
        }
        state.tag = _tag;
        state.dump = '[' + _result + ']';
    }
    function writeBlockSequence(state, level, object, compact) {
        var _result = '', _tag = state.tag, index, length;
        for (index = 0, length = object.length; index < length; index += 1) {
            // Write only valid elements.
            if (writeNode(state, level + 1, object[index], true, true)) {
                if (!compact || index !== 0) {
                    _result += generateNextLine(state, level);
                }
                if (state.dump && CHAR_LINE_FEED === state.dump.charCodeAt(0)) {
                    _result += '-';
                }
                else {
                    _result += '- ';
                }
                _result += state.dump;
            }
        }
        state.tag = _tag;
        state.dump = _result || '[]'; // Empty sequence if no valid values.
    }
    function writeFlowMapping(state, level, object) {
        var _result = '', _tag = state.tag, objectKeyList = Object.keys(object), index, length, objectKey, objectValue, pairBuffer;
        for (index = 0, length = objectKeyList.length; index < length; index += 1) {
            pairBuffer = '';
            if (index !== 0)
                pairBuffer += ', ';
            if (state.condenseFlow)
                pairBuffer += '"';
            objectKey = objectKeyList[index];
            objectValue = object[objectKey];
            if (!writeNode(state, level, objectKey, false, false)) {
                continue; // Skip this pair because of invalid key;
            }
            if (state.dump.length > 1024)
                pairBuffer += '? ';
            pairBuffer += state.dump + (state.condenseFlow ? '"' : '') + ':' + (state.condenseFlow ? '' : ' ');
            if (!writeNode(state, level, objectValue, false, false)) {
                continue; // Skip this pair because of invalid value.
            }
            pairBuffer += state.dump;
            // Both key and value are valid.
            _result += pairBuffer;
        }
        state.tag = _tag;
        state.dump = '{' + _result + '}';
    }
    function writeBlockMapping(state, level, object, compact) {
        var _result = '', _tag = state.tag, objectKeyList = Object.keys(object), index, length, objectKey, objectValue, explicitPair, pairBuffer;
        // Allow sorting keys so that the output file is deterministic
        if (state.sortKeys === true) {
            // Default sorting
            objectKeyList.sort();
        }
        else if (typeof state.sortKeys === 'function') {
            // Custom sort function
            objectKeyList.sort(state.sortKeys);
        }
        else if (state.sortKeys) {
            // Something is wrong
            throw new YAMLException('sortKeys must be a boolean or a function');
        }
        for (index = 0, length = objectKeyList.length; index < length; index += 1) {
            pairBuffer = '';
            if (!compact || index !== 0) {
                pairBuffer += generateNextLine(state, level);
            }
            objectKey = objectKeyList[index];
            objectValue = object[objectKey];
            if (!writeNode(state, level + 1, objectKey, true, true, true)) {
                continue; // Skip this pair because of invalid key.
            }
            explicitPair = (state.tag !== null && state.tag !== '?') ||
                (state.dump && state.dump.length > 1024);
            if (explicitPair) {
                if (state.dump && CHAR_LINE_FEED === state.dump.charCodeAt(0)) {
                    pairBuffer += '?';
                }
                else {
                    pairBuffer += '? ';
                }
            }
            pairBuffer += state.dump;
            if (explicitPair) {
                pairBuffer += generateNextLine(state, level);
            }
            if (!writeNode(state, level + 1, objectValue, true, explicitPair)) {
                continue; // Skip this pair because of invalid value.
            }
            if (state.dump && CHAR_LINE_FEED === state.dump.charCodeAt(0)) {
                pairBuffer += ':';
            }
            else {
                pairBuffer += ': ';
            }
            pairBuffer += state.dump;
            // Both key and value are valid.
            _result += pairBuffer;
        }
        state.tag = _tag;
        state.dump = _result || '{}'; // Empty mapping if no valid pairs.
    }
    function detectType(state, object, explicit) {
        var _result, typeList, index, length, type, style;
        typeList = explicit ? state.explicitTypes : state.implicitTypes;
        for (index = 0, length = typeList.length; index < length; index += 1) {
            type = typeList[index];
            if ((type.instanceOf || type.predicate) &&
                (!type.instanceOf || ((typeof object === 'object') && (object instanceof type.instanceOf))) &&
                (!type.predicate || type.predicate(object))) {
                state.tag = explicit ? type.tag : '?';
                if (type.represent) {
                    style = state.styleMap[type.tag] || type.defaultStyle;
                    if (_toString.call(type.represent) === '[object Function]') {
                        _result = type.represent(object, style);
                    }
                    else if (_hasOwnProperty.call(type.represent, style)) {
                        _result = type.represent[style](object, style);
                    }
                    else {
                        throw new YAMLException('!<' + type.tag + '> tag resolver accepts not "' + style + '" style');
                    }
                    state.dump = _result;
                }
                return true;
            }
        }
        return false;
    }
    // Serializes `object` and writes it to global `result`.
    // Returns true on success, or false on invalid object.
    //
    function writeNode(state, level, object, block, compact, iskey) {
        state.tag = null;
        state.dump = object;
        if (!detectType(state, object, false)) {
            detectType(state, object, true);
        }
        var type = _toString.call(state.dump);
        if (block) {
            block = (state.flowLevel < 0 || state.flowLevel > level);
        }
        var objectOrArray = type === '[object Object]' || type === '[object Array]', duplicateIndex, duplicate;
        if (objectOrArray) {
            duplicateIndex = state.duplicates.indexOf(object);
            duplicate = duplicateIndex !== -1;
        }
        if ((state.tag !== null && state.tag !== '?') || duplicate || (state.indent !== 2 && level > 0)) {
            compact = false;
        }
        if (duplicate && state.usedDuplicates[duplicateIndex]) {
            state.dump = '*ref_' + duplicateIndex;
        }
        else {
            if (objectOrArray && duplicate && !state.usedDuplicates[duplicateIndex]) {
                state.usedDuplicates[duplicateIndex] = true;
            }
            if (type === '[object Object]') {
                if (block && (Object.keys(state.dump).length !== 0)) {
                    writeBlockMapping(state, level, state.dump, compact);
                    if (duplicate) {
                        state.dump = '&ref_' + duplicateIndex + state.dump;
                    }
                }
                else {
                    writeFlowMapping(state, level, state.dump);
                    if (duplicate) {
                        state.dump = '&ref_' + duplicateIndex + ' ' + state.dump;
                    }
                }
            }
            else if (type === '[object Array]') {
                var arrayLevel = (state.noArrayIndent && (level > 0)) ? level - 1 : level;
                if (block && (state.dump.length !== 0)) {
                    writeBlockSequence(state, arrayLevel, state.dump, compact);
                    if (duplicate) {
                        state.dump = '&ref_' + duplicateIndex + state.dump;
                    }
                }
                else {
                    writeFlowSequence(state, arrayLevel, state.dump);
                    if (duplicate) {
                        state.dump = '&ref_' + duplicateIndex + ' ' + state.dump;
                    }
                }
            }
            else if (type === '[object String]') {
                if (state.tag !== '?') {
                    writeScalar(state, state.dump, level, iskey);
                }
            }
            else {
                if (state.skipInvalid)
                    return false;
                throw new YAMLException('unacceptable kind of an object to dump ' + type);
            }
            if (state.tag !== null && state.tag !== '?') {
                state.dump = '!<' + state.tag + '> ' + state.dump;
            }
        }
        return true;
    }
    function getDuplicateReferences(object, state) {
        var objects = [], duplicatesIndexes = [], index, length;
        inspectNode(object, objects, duplicatesIndexes);
        for (index = 0, length = duplicatesIndexes.length; index < length; index += 1) {
            state.duplicates.push(objects[duplicatesIndexes[index]]);
        }
        state.usedDuplicates = new Array(length);
    }
    function inspectNode(object, objects, duplicatesIndexes) {
        var objectKeyList, index, length;
        if (object !== null && typeof object === 'object') {
            index = objects.indexOf(object);
            if (index !== -1) {
                if (duplicatesIndexes.indexOf(index) === -1) {
                    duplicatesIndexes.push(index);
                }
            }
            else {
                objects.push(object);
                if (Array.isArray(object)) {
                    for (index = 0, length = object.length; index < length; index += 1) {
                        inspectNode(object[index], objects, duplicatesIndexes);
                    }
                }
                else {
                    objectKeyList = Object.keys(object);
                    for (index = 0, length = objectKeyList.length; index < length; index += 1) {
                        inspectNode(object[objectKeyList[index]], objects, duplicatesIndexes);
                    }
                }
            }
        }
    }
    function dump(input, options) {
        options = options || {};
        var state = new State(options);
        if (!state.noRefs)
            getDuplicateReferences(input, state);
        if (writeNode(state, 0, input, true, true))
            return state.dump + '\n';
        return '';
    }
    function safeDump(input, options) {
        return dump(input, common.extend({ schema: DEFAULT_SAFE_SCHEMA }, options));
    }
    module.exports.dump = dump;
    module.exports.safeDump = safeDump;
},
/* espree/espree.js */ function _(require, module, exports, __esModule, __esExport) {
    const acorn = require(46) /* acorn */;
    const jsx = require(314) /* acorn-jsx */;
    const astNodeTypes = require(316) /* ./lib/ast-node-types */;
    const espree = require(317) /* ./lib/espree */;
    const { getLatestEcmaVersion, getSupportedEcmaVersions } = require(319) /* ./lib/options */;
    // To initialize lazily.
    const parsers = {
        _regular: null,
        _jsx: null,
        get regular() {
            if (this._regular === null) {
                this._regular = acorn.Parser.extend(espree());
            }
            return this._regular;
        },
        get jsx() {
            if (this._jsx === null) {
                this._jsx = acorn.Parser.extend(jsx(), espree());
            }
            return this._jsx;
        },
        get(options) {
            const useJsx = Boolean(options &&
                options.ecmaFeatures &&
                options.ecmaFeatures.jsx);
            return useJsx ? this.jsx : this.regular;
        }
    };
    //------------------------------------------------------------------------------
    // Tokenizer
    //------------------------------------------------------------------------------
    /**
     * Tokenizes the given code.
     * @param {string} code The code to tokenize.
     * @param {Object} options Options defining how to tokenize.
     * @returns {Token[]} An array of tokens.
     * @throws {SyntaxError} If the input code is invalid.
     * @private
     */
    function tokenize(code, options) {
        const Parser = parsers.get(options);
        // Ensure to collect tokens.
        if (!options || options.tokens !== true) {
            options = Object.assign({}, options, { tokens: true }); // eslint-disable-line no-param-reassign
        }
        return new Parser(options, code).tokenize();
    }
    //------------------------------------------------------------------------------
    // Parser
    //------------------------------------------------------------------------------
    /**
     * Parses the given code.
     * @param {string} code The code to tokenize.
     * @param {Object} options Options defining how to tokenize.
     * @returns {ASTNode} The "Program" AST node.
     * @throws {SyntaxError} If the input code is invalid.
     */
    function parse(code, options) {
        const Parser = parsers.get(options);
        return new Parser(options, code).parse();
    }
    //------------------------------------------------------------------------------
    // Public
    //------------------------------------------------------------------------------
    exports.version = require(320) /* ./package.json */.version;
    exports.tokenize = tokenize;
    exports.parse = parse;
    // Deep copy.
    /* istanbul ignore next */
    exports.Syntax = (function () {
        let name, types = {};
        if (typeof Object.create === "function") {
            types = Object.create(null);
        }
        for (name in astNodeTypes) {
            if (Object.hasOwnProperty.call(astNodeTypes, name)) {
                types[name] = astNodeTypes[name];
            }
        }
        if (typeof Object.freeze === "function") {
            Object.freeze(types);
        }
        return types;
    }());
    /* istanbul ignore next */
    exports.VisitorKeys = (function () {
        return require(321) /* eslint-visitor-keys */.KEYS;
    }());
    exports.latestEcmaVersion = getLatestEcmaVersion();
    exports.supportedEcmaVersions = getSupportedEcmaVersions();
},
/* espree/node_modules/acorn-jsx/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const XHTMLEntities = require(315) /* ./xhtml */;
    const hexNumber = /^[\da-fA-F]+$/;
    const decimalNumber = /^\d+$/;
    // The map to `acorn-jsx` tokens from `acorn` namespace objects.
    const acornJsxMap = new WeakMap();
    // Get the original tokens for the given `acorn` namespace object.
    function getJsxTokens(acorn) {
        acorn = acorn.Parser.acorn || acorn;
        let acornJsx = acornJsxMap.get(acorn);
        if (!acornJsx) {
            const tt = acorn.tokTypes;
            const TokContext = acorn.TokContext;
            const TokenType = acorn.TokenType;
            const tc_oTag = new TokContext('<tag', false);
            const tc_cTag = new TokContext('</tag', false);
            const tc_expr = new TokContext('<tag>...</tag>', true, true);
            const tokContexts = {
                tc_oTag: tc_oTag,
                tc_cTag: tc_cTag,
                tc_expr: tc_expr
            };
            const tokTypes = {
                jsxName: new TokenType('jsxName'),
                jsxText: new TokenType('jsxText', { beforeExpr: true }),
                jsxTagStart: new TokenType('jsxTagStart', { startsExpr: true }),
                jsxTagEnd: new TokenType('jsxTagEnd')
            };
            tokTypes.jsxTagStart.updateContext = function () {
                this.context.push(tc_expr); // treat as beginning of JSX expression
                this.context.push(tc_oTag); // start opening tag context
                this.exprAllowed = false;
            };
            tokTypes.jsxTagEnd.updateContext = function (prevType) {
                let out = this.context.pop();
                if (out === tc_oTag && prevType === tt.slash || out === tc_cTag) {
                    this.context.pop();
                    this.exprAllowed = this.curContext() === tc_expr;
                }
                else {
                    this.exprAllowed = true;
                }
            };
            acornJsx = { tokContexts: tokContexts, tokTypes: tokTypes };
            acornJsxMap.set(acorn, acornJsx);
        }
        return acornJsx;
    }
    // Transforms JSX element name to string.
    function getQualifiedJSXName(object) {
        if (!object)
            return object;
        if (object.type === 'JSXIdentifier')
            return object.name;
        if (object.type === 'JSXNamespacedName')
            return object.namespace.name + ':' + object.name.name;
        if (object.type === 'JSXMemberExpression')
            return getQualifiedJSXName(object.object) + '.' +
                getQualifiedJSXName(object.property);
    }
    module.exports = function (options) {
        options = options || {};
        return function (Parser) {
            return plugin({
                allowNamespaces: options.allowNamespaces !== false,
                allowNamespacedObjects: !!options.allowNamespacedObjects
            }, Parser);
        };
    };
    // This is `tokTypes` of the peer dep.
    // This can be different instances from the actual `tokTypes` this plugin uses.
    Object.defineProperty(module.exports, "tokTypes", {
        get: function get_tokTypes() {
            return getJsxTokens(require(46) /* acorn */).tokTypes;
        },
        configurable: true,
        enumerable: true
    });
    function plugin(options, Parser) {
        const acorn = Parser.acorn || require(46) /* acorn */;
        const acornJsx = getJsxTokens(acorn);
        const tt = acorn.tokTypes;
        const tok = acornJsx.tokTypes;
        const tokContexts = acorn.tokContexts;
        const tc_oTag = acornJsx.tokContexts.tc_oTag;
        const tc_cTag = acornJsx.tokContexts.tc_cTag;
        const tc_expr = acornJsx.tokContexts.tc_expr;
        const isNewLine = acorn.isNewLine;
        const isIdentifierStart = acorn.isIdentifierStart;
        const isIdentifierChar = acorn.isIdentifierChar;
        return class extends Parser {
            // Expose actual `tokTypes` and `tokContexts` to other plugins.
            static get acornJsx() {
                return acornJsx;
            }
            // Reads inline JSX contents token.
            jsx_readToken() {
                let out = '', chunkStart = this.pos;
                for (;;) {
                    if (this.pos >= this.input.length)
                        this.raise(this.start, 'Unterminated JSX contents');
                    let ch = this.input.charCodeAt(this.pos);
                    switch (ch) {
                        case 60: // '<'
                        case 123: // '{'
                            if (this.pos === this.start) {
                                if (ch === 60 && this.exprAllowed) {
                                    ++this.pos;
                                    return this.finishToken(tok.jsxTagStart);
                                }
                                return this.getTokenFromCode(ch);
                            }
                            out += this.input.slice(chunkStart, this.pos);
                            return this.finishToken(tok.jsxText, out);
                        case 38: // '&'
                            out += this.input.slice(chunkStart, this.pos);
                            out += this.jsx_readEntity();
                            chunkStart = this.pos;
                            break;
                        case 62: // '>'
                        case 125: // '}'
                            this.raise(this.pos, "Unexpected token `" + this.input[this.pos] + "`. Did you mean `" +
                                (ch === 62 ? "&gt;" : "&rbrace;") + "` or " + "`{\"" + this.input[this.pos] + "\"}" + "`?");
                        default:
                            if (isNewLine(ch)) {
                                out += this.input.slice(chunkStart, this.pos);
                                out += this.jsx_readNewLine(true);
                                chunkStart = this.pos;
                            }
                            else {
                                ++this.pos;
                            }
                    }
                }
            }
            jsx_readNewLine(normalizeCRLF) {
                let ch = this.input.charCodeAt(this.pos);
                let out;
                ++this.pos;
                if (ch === 13 && this.input.charCodeAt(this.pos) === 10) {
                    ++this.pos;
                    out = normalizeCRLF ? '\n' : '\r\n';
                }
                else {
                    out = String.fromCharCode(ch);
                }
                if (this.options.locations) {
                    ++this.curLine;
                    this.lineStart = this.pos;
                }
                return out;
            }
            jsx_readString(quote) {
                let out = '', chunkStart = ++this.pos;
                for (;;) {
                    if (this.pos >= this.input.length)
                        this.raise(this.start, 'Unterminated string constant');
                    let ch = this.input.charCodeAt(this.pos);
                    if (ch === quote)
                        break;
                    if (ch === 38) { // '&'
                        out += this.input.slice(chunkStart, this.pos);
                        out += this.jsx_readEntity();
                        chunkStart = this.pos;
                    }
                    else if (isNewLine(ch)) {
                        out += this.input.slice(chunkStart, this.pos);
                        out += this.jsx_readNewLine(false);
                        chunkStart = this.pos;
                    }
                    else {
                        ++this.pos;
                    }
                }
                out += this.input.slice(chunkStart, this.pos++);
                return this.finishToken(tt.string, out);
            }
            jsx_readEntity() {
                let str = '', count = 0, entity;
                let ch = this.input[this.pos];
                if (ch !== '&')
                    this.raise(this.pos, 'Entity must start with an ampersand');
                let startPos = ++this.pos;
                while (this.pos < this.input.length && count++ < 10) {
                    ch = this.input[this.pos++];
                    if (ch === ';') {
                        if (str[0] === '#') {
                            if (str[1] === 'x') {
                                str = str.substr(2);
                                if (hexNumber.test(str))
                                    entity = String.fromCharCode(parseInt(str, 16));
                            }
                            else {
                                str = str.substr(1);
                                if (decimalNumber.test(str))
                                    entity = String.fromCharCode(parseInt(str, 10));
                            }
                        }
                        else {
                            entity = XHTMLEntities[str];
                        }
                        break;
                    }
                    str += ch;
                }
                if (!entity) {
                    this.pos = startPos;
                    return '&';
                }
                return entity;
            }
            // Read a JSX identifier (valid tag or attribute name).
            //
            // Optimized version since JSX identifiers can't contain
            // escape characters and so can be read as single slice.
            // Also assumes that first character was already checked
            // by isIdentifierStart in readToken.
            jsx_readWord() {
                let ch, start = this.pos;
                do {
                    ch = this.input.charCodeAt(++this.pos);
                } while (isIdentifierChar(ch) || ch === 45); // '-'
                return this.finishToken(tok.jsxName, this.input.slice(start, this.pos));
            }
            // Parse next token as JSX identifier
            jsx_parseIdentifier() {
                let node = this.startNode();
                if (this.type === tok.jsxName)
                    node.name = this.value;
                else if (this.type.keyword)
                    node.name = this.type.keyword;
                else
                    this.unexpected();
                this.next();
                return this.finishNode(node, 'JSXIdentifier');
            }
            // Parse namespaced identifier.
            jsx_parseNamespacedName() {
                let startPos = this.start, startLoc = this.startLoc;
                let name = this.jsx_parseIdentifier();
                if (!options.allowNamespaces || !this.eat(tt.colon))
                    return name;
                var node = this.startNodeAt(startPos, startLoc);
                node.namespace = name;
                node.name = this.jsx_parseIdentifier();
                return this.finishNode(node, 'JSXNamespacedName');
            }
            // Parses element name in any form - namespaced, member
            // or single identifier.
            jsx_parseElementName() {
                if (this.type === tok.jsxTagEnd)
                    return '';
                let startPos = this.start, startLoc = this.startLoc;
                let node = this.jsx_parseNamespacedName();
                if (this.type === tt.dot && node.type === 'JSXNamespacedName' && !options.allowNamespacedObjects) {
                    this.unexpected();
                }
                while (this.eat(tt.dot)) {
                    let newNode = this.startNodeAt(startPos, startLoc);
                    newNode.object = node;
                    newNode.property = this.jsx_parseIdentifier();
                    node = this.finishNode(newNode, 'JSXMemberExpression');
                }
                return node;
            }
            // Parses any type of JSX attribute value.
            jsx_parseAttributeValue() {
                switch (this.type) {
                    case tt.braceL:
                        let node = this.jsx_parseExpressionContainer();
                        if (node.expression.type === 'JSXEmptyExpression')
                            this.raise(node.start, 'JSX attributes must only be assigned a non-empty expression');
                        return node;
                    case tok.jsxTagStart:
                    case tt.string:
                        return this.parseExprAtom();
                    default:
                        this.raise(this.start, 'JSX value should be either an expression or a quoted JSX text');
                }
            }
            // JSXEmptyExpression is unique type since it doesn't actually parse anything,
            // and so it should start at the end of last read token (left brace) and finish
            // at the beginning of the next one (right brace).
            jsx_parseEmptyExpression() {
                let node = this.startNodeAt(this.lastTokEnd, this.lastTokEndLoc);
                return this.finishNodeAt(node, 'JSXEmptyExpression', this.start, this.startLoc);
            }
            // Parses JSX expression enclosed into curly brackets.
            jsx_parseExpressionContainer() {
                let node = this.startNode();
                this.next();
                node.expression = this.type === tt.braceR
                    ? this.jsx_parseEmptyExpression()
                    : this.parseExpression();
                this.expect(tt.braceR);
                return this.finishNode(node, 'JSXExpressionContainer');
            }
            // Parses following JSX attribute name-value pair.
            jsx_parseAttribute() {
                let node = this.startNode();
                if (this.eat(tt.braceL)) {
                    this.expect(tt.ellipsis);
                    node.argument = this.parseMaybeAssign();
                    this.expect(tt.braceR);
                    return this.finishNode(node, 'JSXSpreadAttribute');
                }
                node.name = this.jsx_parseNamespacedName();
                node.value = this.eat(tt.eq) ? this.jsx_parseAttributeValue() : null;
                return this.finishNode(node, 'JSXAttribute');
            }
            // Parses JSX opening tag starting after '<'.
            jsx_parseOpeningElementAt(startPos, startLoc) {
                let node = this.startNodeAt(startPos, startLoc);
                node.attributes = [];
                let nodeName = this.jsx_parseElementName();
                if (nodeName)
                    node.name = nodeName;
                while (this.type !== tt.slash && this.type !== tok.jsxTagEnd)
                    node.attributes.push(this.jsx_parseAttribute());
                node.selfClosing = this.eat(tt.slash);
                this.expect(tok.jsxTagEnd);
                return this.finishNode(node, nodeName ? 'JSXOpeningElement' : 'JSXOpeningFragment');
            }
            // Parses JSX closing tag starting after '</'.
            jsx_parseClosingElementAt(startPos, startLoc) {
                let node = this.startNodeAt(startPos, startLoc);
                let nodeName = this.jsx_parseElementName();
                if (nodeName)
                    node.name = nodeName;
                this.expect(tok.jsxTagEnd);
                return this.finishNode(node, nodeName ? 'JSXClosingElement' : 'JSXClosingFragment');
            }
            // Parses entire JSX element, including it's opening tag
            // (starting after '<'), attributes, contents and closing tag.
            jsx_parseElementAt(startPos, startLoc) {
                let node = this.startNodeAt(startPos, startLoc);
                let children = [];
                let openingElement = this.jsx_parseOpeningElementAt(startPos, startLoc);
                let closingElement = null;
                if (!openingElement.selfClosing) {
                    contents: for (;;) {
                        switch (this.type) {
                            case tok.jsxTagStart:
                                startPos = this.start;
                                startLoc = this.startLoc;
                                this.next();
                                if (this.eat(tt.slash)) {
                                    closingElement = this.jsx_parseClosingElementAt(startPos, startLoc);
                                    break contents;
                                }
                                children.push(this.jsx_parseElementAt(startPos, startLoc));
                                break;
                            case tok.jsxText:
                                children.push(this.parseExprAtom());
                                break;
                            case tt.braceL:
                                children.push(this.jsx_parseExpressionContainer());
                                break;
                            default:
                                this.unexpected();
                        }
                    }
                    if (getQualifiedJSXName(closingElement.name) !== getQualifiedJSXName(openingElement.name)) {
                        this.raise(closingElement.start, 'Expected corresponding JSX closing tag for <' + getQualifiedJSXName(openingElement.name) + '>');
                    }
                }
                let fragmentOrElement = openingElement.name ? 'Element' : 'Fragment';
                node['opening' + fragmentOrElement] = openingElement;
                node['closing' + fragmentOrElement] = closingElement;
                node.children = children;
                if (this.type === tt.relational && this.value === "<") {
                    this.raise(this.start, "Adjacent JSX elements must be wrapped in an enclosing tag");
                }
                return this.finishNode(node, 'JSX' + fragmentOrElement);
            }
            // Parse JSX text
            jsx_parseText() {
                let node = this.parseLiteral(this.value);
                node.type = "JSXText";
                return node;
            }
            // Parses entire JSX element from current position.
            jsx_parseElement() {
                let startPos = this.start, startLoc = this.startLoc;
                this.next();
                return this.jsx_parseElementAt(startPos, startLoc);
            }
            parseExprAtom(refShortHandDefaultPos) {
                if (this.type === tok.jsxText)
                    return this.jsx_parseText();
                else if (this.type === tok.jsxTagStart)
                    return this.jsx_parseElement();
                else
                    return super.parseExprAtom(refShortHandDefaultPos);
            }
            readToken(code) {
                let context = this.curContext();
                if (context === tc_expr)
                    return this.jsx_readToken();
                if (context === tc_oTag || context === tc_cTag) {
                    if (isIdentifierStart(code))
                        return this.jsx_readWord();
                    if (code == 62) {
                        ++this.pos;
                        return this.finishToken(tok.jsxTagEnd);
                    }
                    if ((code === 34 || code === 39) && context == tc_oTag)
                        return this.jsx_readString(code);
                }
                if (code === 60 && this.exprAllowed && this.input.charCodeAt(this.pos + 1) !== 33) {
                    ++this.pos;
                    return this.finishToken(tok.jsxTagStart);
                }
                return super.readToken(code);
            }
            updateContext(prevType) {
                if (this.type == tt.braceL) {
                    var curContext = this.curContext();
                    if (curContext == tc_oTag)
                        this.context.push(tokContexts.b_expr);
                    else if (curContext == tc_expr)
                        this.context.push(tokContexts.b_tmpl);
                    else
                        super.updateContext(prevType);
                    this.exprAllowed = true;
                }
                else if (this.type === tt.slash && prevType === tok.jsxTagStart) {
                    this.context.length -= 2; // do not consider JSX expr -> JSX open tag -> ... anymore
                    this.context.push(tc_cTag); // reconsider as closing tag context
                    this.exprAllowed = false;
                }
                else {
                    return super.updateContext(prevType);
                }
            }
        };
    }
},
/* espree/node_modules/acorn-jsx/xhtml.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        quot: '\u0022',
        amp: '&',
        apos: '\u0027',
        lt: '<',
        gt: '>',
        nbsp: '\u00A0',
        iexcl: '\u00A1',
        cent: '\u00A2',
        pound: '\u00A3',
        curren: '\u00A4',
        yen: '\u00A5',
        brvbar: '\u00A6',
        sect: '\u00A7',
        uml: '\u00A8',
        copy: '\u00A9',
        ordf: '\u00AA',
        laquo: '\u00AB',
        not: '\u00AC',
        shy: '\u00AD',
        reg: '\u00AE',
        macr: '\u00AF',
        deg: '\u00B0',
        plusmn: '\u00B1',
        sup2: '\u00B2',
        sup3: '\u00B3',
        acute: '\u00B4',
        micro: '\u00B5',
        para: '\u00B6',
        middot: '\u00B7',
        cedil: '\u00B8',
        sup1: '\u00B9',
        ordm: '\u00BA',
        raquo: '\u00BB',
        frac14: '\u00BC',
        frac12: '\u00BD',
        frac34: '\u00BE',
        iquest: '\u00BF',
        Agrave: '\u00C0',
        Aacute: '\u00C1',
        Acirc: '\u00C2',
        Atilde: '\u00C3',
        Auml: '\u00C4',
        Aring: '\u00C5',
        AElig: '\u00C6',
        Ccedil: '\u00C7',
        Egrave: '\u00C8',
        Eacute: '\u00C9',
        Ecirc: '\u00CA',
        Euml: '\u00CB',
        Igrave: '\u00CC',
        Iacute: '\u00CD',
        Icirc: '\u00CE',
        Iuml: '\u00CF',
        ETH: '\u00D0',
        Ntilde: '\u00D1',
        Ograve: '\u00D2',
        Oacute: '\u00D3',
        Ocirc: '\u00D4',
        Otilde: '\u00D5',
        Ouml: '\u00D6',
        times: '\u00D7',
        Oslash: '\u00D8',
        Ugrave: '\u00D9',
        Uacute: '\u00DA',
        Ucirc: '\u00DB',
        Uuml: '\u00DC',
        Yacute: '\u00DD',
        THORN: '\u00DE',
        szlig: '\u00DF',
        agrave: '\u00E0',
        aacute: '\u00E1',
        acirc: '\u00E2',
        atilde: '\u00E3',
        auml: '\u00E4',
        aring: '\u00E5',
        aelig: '\u00E6',
        ccedil: '\u00E7',
        egrave: '\u00E8',
        eacute: '\u00E9',
        ecirc: '\u00EA',
        euml: '\u00EB',
        igrave: '\u00EC',
        iacute: '\u00ED',
        icirc: '\u00EE',
        iuml: '\u00EF',
        eth: '\u00F0',
        ntilde: '\u00F1',
        ograve: '\u00F2',
        oacute: '\u00F3',
        ocirc: '\u00F4',
        otilde: '\u00F5',
        ouml: '\u00F6',
        divide: '\u00F7',
        oslash: '\u00F8',
        ugrave: '\u00F9',
        uacute: '\u00FA',
        ucirc: '\u00FB',
        uuml: '\u00FC',
        yacute: '\u00FD',
        thorn: '\u00FE',
        yuml: '\u00FF',
        OElig: '\u0152',
        oelig: '\u0153',
        Scaron: '\u0160',
        scaron: '\u0161',
        Yuml: '\u0178',
        fnof: '\u0192',
        circ: '\u02C6',
        tilde: '\u02DC',
        Alpha: '\u0391',
        Beta: '\u0392',
        Gamma: '\u0393',
        Delta: '\u0394',
        Epsilon: '\u0395',
        Zeta: '\u0396',
        Eta: '\u0397',
        Theta: '\u0398',
        Iota: '\u0399',
        Kappa: '\u039A',
        Lambda: '\u039B',
        Mu: '\u039C',
        Nu: '\u039D',
        Xi: '\u039E',
        Omicron: '\u039F',
        Pi: '\u03A0',
        Rho: '\u03A1',
        Sigma: '\u03A3',
        Tau: '\u03A4',
        Upsilon: '\u03A5',
        Phi: '\u03A6',
        Chi: '\u03A7',
        Psi: '\u03A8',
        Omega: '\u03A9',
        alpha: '\u03B1',
        beta: '\u03B2',
        gamma: '\u03B3',
        delta: '\u03B4',
        epsilon: '\u03B5',
        zeta: '\u03B6',
        eta: '\u03B7',
        theta: '\u03B8',
        iota: '\u03B9',
        kappa: '\u03BA',
        lambda: '\u03BB',
        mu: '\u03BC',
        nu: '\u03BD',
        xi: '\u03BE',
        omicron: '\u03BF',
        pi: '\u03C0',
        rho: '\u03C1',
        sigmaf: '\u03C2',
        sigma: '\u03C3',
        tau: '\u03C4',
        upsilon: '\u03C5',
        phi: '\u03C6',
        chi: '\u03C7',
        psi: '\u03C8',
        omega: '\u03C9',
        thetasym: '\u03D1',
        upsih: '\u03D2',
        piv: '\u03D6',
        ensp: '\u2002',
        emsp: '\u2003',
        thinsp: '\u2009',
        zwnj: '\u200C',
        zwj: '\u200D',
        lrm: '\u200E',
        rlm: '\u200F',
        ndash: '\u2013',
        mdash: '\u2014',
        lsquo: '\u2018',
        rsquo: '\u2019',
        sbquo: '\u201A',
        ldquo: '\u201C',
        rdquo: '\u201D',
        bdquo: '\u201E',
        dagger: '\u2020',
        Dagger: '\u2021',
        bull: '\u2022',
        hellip: '\u2026',
        permil: '\u2030',
        prime: '\u2032',
        Prime: '\u2033',
        lsaquo: '\u2039',
        rsaquo: '\u203A',
        oline: '\u203E',
        frasl: '\u2044',
        euro: '\u20AC',
        image: '\u2111',
        weierp: '\u2118',
        real: '\u211C',
        trade: '\u2122',
        alefsym: '\u2135',
        larr: '\u2190',
        uarr: '\u2191',
        rarr: '\u2192',
        darr: '\u2193',
        harr: '\u2194',
        crarr: '\u21B5',
        lArr: '\u21D0',
        uArr: '\u21D1',
        rArr: '\u21D2',
        dArr: '\u21D3',
        hArr: '\u21D4',
        forall: '\u2200',
        part: '\u2202',
        exist: '\u2203',
        empty: '\u2205',
        nabla: '\u2207',
        isin: '\u2208',
        notin: '\u2209',
        ni: '\u220B',
        prod: '\u220F',
        sum: '\u2211',
        minus: '\u2212',
        lowast: '\u2217',
        radic: '\u221A',
        prop: '\u221D',
        infin: '\u221E',
        ang: '\u2220',
        and: '\u2227',
        or: '\u2228',
        cap: '\u2229',
        cup: '\u222A',
        'int': '\u222B',
        there4: '\u2234',
        sim: '\u223C',
        cong: '\u2245',
        asymp: '\u2248',
        ne: '\u2260',
        equiv: '\u2261',
        le: '\u2264',
        ge: '\u2265',
        sub: '\u2282',
        sup: '\u2283',
        nsub: '\u2284',
        sube: '\u2286',
        supe: '\u2287',
        oplus: '\u2295',
        otimes: '\u2297',
        perp: '\u22A5',
        sdot: '\u22C5',
        lceil: '\u2308',
        rceil: '\u2309',
        lfloor: '\u230A',
        rfloor: '\u230B',
        lang: '\u2329',
        rang: '\u232A',
        loz: '\u25CA',
        spades: '\u2660',
        clubs: '\u2663',
        hearts: '\u2665',
        diams: '\u2666'
    };
},
/* espree/lib/ast-node-types.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    // None!
    //------------------------------------------------------------------------------
    // Public
    //------------------------------------------------------------------------------
    module.exports = {
        AssignmentExpression: "AssignmentExpression",
        AssignmentPattern: "AssignmentPattern",
        ArrayExpression: "ArrayExpression",
        ArrayPattern: "ArrayPattern",
        ArrowFunctionExpression: "ArrowFunctionExpression",
        AwaitExpression: "AwaitExpression",
        BlockStatement: "BlockStatement",
        BinaryExpression: "BinaryExpression",
        BreakStatement: "BreakStatement",
        CallExpression: "CallExpression",
        CatchClause: "CatchClause",
        ClassBody: "ClassBody",
        ClassDeclaration: "ClassDeclaration",
        ClassExpression: "ClassExpression",
        ConditionalExpression: "ConditionalExpression",
        ContinueStatement: "ContinueStatement",
        DoWhileStatement: "DoWhileStatement",
        DebuggerStatement: "DebuggerStatement",
        EmptyStatement: "EmptyStatement",
        ExpressionStatement: "ExpressionStatement",
        ForStatement: "ForStatement",
        ForInStatement: "ForInStatement",
        ForOfStatement: "ForOfStatement",
        FunctionDeclaration: "FunctionDeclaration",
        FunctionExpression: "FunctionExpression",
        Identifier: "Identifier",
        IfStatement: "IfStatement",
        Literal: "Literal",
        LabeledStatement: "LabeledStatement",
        LogicalExpression: "LogicalExpression",
        MemberExpression: "MemberExpression",
        MetaProperty: "MetaProperty",
        MethodDefinition: "MethodDefinition",
        NewExpression: "NewExpression",
        ObjectExpression: "ObjectExpression",
        ObjectPattern: "ObjectPattern",
        Program: "Program",
        Property: "Property",
        RestElement: "RestElement",
        ReturnStatement: "ReturnStatement",
        SequenceExpression: "SequenceExpression",
        SpreadElement: "SpreadElement",
        Super: "Super",
        SwitchCase: "SwitchCase",
        SwitchStatement: "SwitchStatement",
        TaggedTemplateExpression: "TaggedTemplateExpression",
        TemplateElement: "TemplateElement",
        TemplateLiteral: "TemplateLiteral",
        ThisExpression: "ThisExpression",
        ThrowStatement: "ThrowStatement",
        TryStatement: "TryStatement",
        UnaryExpression: "UnaryExpression",
        UpdateExpression: "UpdateExpression",
        VariableDeclaration: "VariableDeclaration",
        VariableDeclarator: "VariableDeclarator",
        WhileStatement: "WhileStatement",
        WithStatement: "WithStatement",
        YieldExpression: "YieldExpression",
        JSXIdentifier: "JSXIdentifier",
        JSXNamespacedName: "JSXNamespacedName",
        JSXMemberExpression: "JSXMemberExpression",
        JSXEmptyExpression: "JSXEmptyExpression",
        JSXExpressionContainer: "JSXExpressionContainer",
        JSXElement: "JSXElement",
        JSXClosingElement: "JSXClosingElement",
        JSXOpeningElement: "JSXOpeningElement",
        JSXAttribute: "JSXAttribute",
        JSXSpreadAttribute: "JSXSpreadAttribute",
        JSXText: "JSXText",
        ExportDefaultDeclaration: "ExportDefaultDeclaration",
        ExportNamedDeclaration: "ExportNamedDeclaration",
        ExportAllDeclaration: "ExportAllDeclaration",
        ExportSpecifier: "ExportSpecifier",
        ImportDeclaration: "ImportDeclaration",
        ImportSpecifier: "ImportSpecifier",
        ImportDefaultSpecifier: "ImportDefaultSpecifier",
        ImportNamespaceSpecifier: "ImportNamespaceSpecifier"
    };
},
/* espree/lib/espree.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint-disable no-param-reassign*/
    const TokenTranslator = require(318) /* ./token-translator */;
    const { normalizeOptions } = require(319) /* ./options */;
    const STATE = Symbol("espree's internal state");
    const ESPRIMA_FINISH_NODE = Symbol("espree's esprimaFinishNode");
    /**
     * Converts an Acorn comment to a Esprima comment.
     * @param {boolean} block True if it's a block comment, false if not.
     * @param {string} text The text of the comment.
     * @param {int} start The index at which the comment starts.
     * @param {int} end The index at which the comment ends.
     * @param {Location} startLoc The location at which the comment starts.
     * @param {Location} endLoc The location at which the comment ends.
     * @returns {Object} The comment object.
     * @private
     */
    function convertAcornCommentToEsprimaComment(block, text, start, end, startLoc, endLoc) {
        const comment = {
            type: block ? "Block" : "Line",
            value: text
        };
        if (typeof start === "number") {
            comment.start = start;
            comment.end = end;
            comment.range = [start, end];
        }
        if (typeof startLoc === "object") {
            comment.loc = {
                start: startLoc,
                end: endLoc
            };
        }
        return comment;
    }
    module.exports = () => Parser => {
        const tokTypes = Object.assign({}, Parser.acorn.tokTypes);
        if (Parser.acornJsx) {
            Object.assign(tokTypes, Parser.acornJsx.tokTypes);
        }
        return class Espree extends Parser {
            constructor(opts, code) {
                if (typeof opts !== "object" || opts === null) {
                    opts = {};
                }
                if (typeof code !== "string" && !(code instanceof String)) {
                    code = String(code);
                }
                const options = normalizeOptions(opts);
                const ecmaFeatures = options.ecmaFeatures || {};
                const tokenTranslator = options.tokens === true
                    ? new TokenTranslator(tokTypes, code)
                    : null;
                // Initialize acorn parser.
                super({
                    // TODO: use {...options} when spread is supported(Node.js >= 8.3.0).
                    ecmaVersion: options.ecmaVersion,
                    sourceType: options.sourceType,
                    ranges: options.ranges,
                    locations: options.locations,
                    // Truthy value is true for backward compatibility.
                    allowReturnOutsideFunction: Boolean(ecmaFeatures.globalReturn),
                    // Collect tokens
                    onToken: token => {
                        if (tokenTranslator) {
                            // Use `tokens`, `ecmaVersion`, and `jsxAttrValueToken` in the state.
                            tokenTranslator.onToken(token, this[STATE]);
                        }
                        if (token.type !== tokTypes.eof) {
                            this[STATE].lastToken = token;
                        }
                    },
                    // Collect comments
                    onComment: (block, text, start, end, startLoc, endLoc) => {
                        if (this[STATE].comments) {
                            const comment = convertAcornCommentToEsprimaComment(block, text, start, end, startLoc, endLoc);
                            this[STATE].comments.push(comment);
                        }
                    }
                }, code);
                // Initialize internal state.
                this[STATE] = {
                    tokens: tokenTranslator ? [] : null,
                    comments: options.comment === true ? [] : null,
                    impliedStrict: ecmaFeatures.impliedStrict === true && this.options.ecmaVersion >= 5,
                    ecmaVersion: this.options.ecmaVersion,
                    jsxAttrValueToken: false,
                    lastToken: null
                };
            }
            tokenize() {
                do {
                    this.next();
                } while (this.type !== tokTypes.eof);
                // Consume the final eof token
                this.next();
                const extra = this[STATE];
                const tokens = extra.tokens;
                if (extra.comments) {
                    tokens.comments = extra.comments;
                }
                return tokens;
            }
            finishNode(...args) {
                const result = super.finishNode(...args);
                return this[ESPRIMA_FINISH_NODE](result);
            }
            finishNodeAt(...args) {
                const result = super.finishNodeAt(...args);
                return this[ESPRIMA_FINISH_NODE](result);
            }
            parse() {
                const extra = this[STATE];
                const program = super.parse();
                program.sourceType = this.options.sourceType;
                if (extra.comments) {
                    program.comments = extra.comments;
                }
                if (extra.tokens) {
                    program.tokens = extra.tokens;
                }
                /*
                 * Adjust opening and closing position of program to match Esprima.
                 * Acorn always starts programs at range 0 whereas Esprima starts at the
                 * first AST node's start (the only real difference is when there's leading
                 * whitespace or leading comments). Acorn also counts trailing whitespace
                 * as part of the program whereas Esprima only counts up to the last token.
                 */
                if (program.range) {
                    program.range[0] = program.body.length ? program.body[0].range[0] : program.range[0];
                    program.range[1] = extra.lastToken ? extra.lastToken.range[1] : program.range[1];
                }
                if (program.loc) {
                    program.loc.start = program.body.length ? program.body[0].loc.start : program.loc.start;
                    program.loc.end = extra.lastToken ? extra.lastToken.loc.end : program.loc.end;
                }
                return program;
            }
            parseTopLevel(node) {
                if (this[STATE].impliedStrict) {
                    this.strict = true;
                }
                return super.parseTopLevel(node);
            }
            /**
             * Overwrites the default raise method to throw Esprima-style errors.
             * @param {int} pos The position of the error.
             * @param {string} message The error message.
             * @throws {SyntaxError} A syntax error.
             * @returns {void}
             */
            raise(pos, message) {
                const loc = Parser.acorn.getLineInfo(this.input, pos);
                const err = new SyntaxError(message);
                err.index = pos;
                err.lineNumber = loc.line;
                err.column = loc.column + 1; // acorn uses 0-based columns
                throw err;
            }
            /**
             * Overwrites the default raise method to throw Esprima-style errors.
             * @param {int} pos The position of the error.
             * @param {string} message The error message.
             * @throws {SyntaxError} A syntax error.
             * @returns {void}
             */
            raiseRecoverable(pos, message) {
                this.raise(pos, message);
            }
            /**
             * Overwrites the default unexpected method to throw Esprima-style errors.
             * @param {int} pos The position of the error.
             * @throws {SyntaxError} A syntax error.
             * @returns {void}
             */
            unexpected(pos) {
                let message = "Unexpected token";
                if (pos !== null && pos !== void 0) {
                    this.pos = pos;
                    if (this.options.locations) {
                        while (this.pos < this.lineStart) {
                            this.lineStart = this.input.lastIndexOf("\n", this.lineStart - 2) + 1;
                            --this.curLine;
                        }
                    }
                    this.nextToken();
                }
                if (this.end > this.start) {
                    message += ` ${this.input.slice(this.start, this.end)}`;
                }
                this.raise(this.start, message);
            }
            /*
            * Esprima-FB represents JSX strings as tokens called "JSXText", but Acorn-JSX
            * uses regular tt.string without any distinction between this and regular JS
            * strings. As such, we intercept an attempt to read a JSX string and set a flag
            * on extra so that when tokens are converted, the next token will be switched
            * to JSXText via onToken.
            */
            jsx_readString(quote) {
                const result = super.jsx_readString(quote);
                if (this.type === tokTypes.string) {
                    this[STATE].jsxAttrValueToken = true;
                }
                return result;
            }
            /**
             * Performs last-minute Esprima-specific compatibility checks and fixes.
             * @param {ASTNode} result The node to check.
             * @returns {ASTNode} The finished node.
             */
            [ESPRIMA_FINISH_NODE](result) {
                // Acorn doesn't count the opening and closing backticks as part of templates
                // so we have to adjust ranges/locations appropriately.
                if (result.type === "TemplateElement") {
                    // additional adjustment needed if ${ is the last token
                    const terminalDollarBraceL = this.input.slice(result.end, result.end + 2) === "${";
                    if (result.range) {
                        result.range[0]--;
                        result.range[1] += (terminalDollarBraceL ? 2 : 1);
                    }
                    if (result.loc) {
                        result.loc.start.column--;
                        result.loc.end.column += (terminalDollarBraceL ? 2 : 1);
                    }
                }
                if (result.type.indexOf("Function") > -1 && !result.generator) {
                    result.generator = false;
                }
                return result;
            }
        };
    };
},
/* espree/lib/token-translator.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    // none!
    //------------------------------------------------------------------------------
    // Private
    //------------------------------------------------------------------------------
    // Esprima Token Types
    const Token = {
        Boolean: "Boolean",
        EOF: "<end>",
        Identifier: "Identifier",
        Keyword: "Keyword",
        Null: "Null",
        Numeric: "Numeric",
        Punctuator: "Punctuator",
        String: "String",
        RegularExpression: "RegularExpression",
        Template: "Template",
        JSXIdentifier: "JSXIdentifier",
        JSXText: "JSXText"
    };
    /**
     * Converts part of a template into an Esprima token.
     * @param {AcornToken[]} tokens The Acorn tokens representing the template.
     * @param {string} code The source code.
     * @returns {EsprimaToken} The Esprima equivalent of the template token.
     * @private
     */
    function convertTemplatePart(tokens, code) {
        const firstToken = tokens[0], lastTemplateToken = tokens[tokens.length - 1];
        const token = {
            type: Token.Template,
            value: code.slice(firstToken.start, lastTemplateToken.end)
        };
        if (firstToken.loc) {
            token.loc = {
                start: firstToken.loc.start,
                end: lastTemplateToken.loc.end
            };
        }
        if (firstToken.range) {
            token.start = firstToken.range[0];
            token.end = lastTemplateToken.range[1];
            token.range = [token.start, token.end];
        }
        return token;
    }
    /**
     * Contains logic to translate Acorn tokens into Esprima tokens.
     * @param {Object} acornTokTypes The Acorn token types.
     * @param {string} code The source code Acorn is parsing. This is necessary
     *      to correct the "value" property of some tokens.
     * @constructor
     */
    function TokenTranslator(acornTokTypes, code) {
        // token types
        this._acornTokTypes = acornTokTypes;
        // token buffer for templates
        this._tokens = [];
        // track the last curly brace
        this._curlyBrace = null;
        // the source code
        this._code = code;
    }
    TokenTranslator.prototype = {
        constructor: TokenTranslator,
        /**
         * Translates a single Esprima token to a single Acorn token. This may be
         * inaccurate due to how templates are handled differently in Esprima and
         * Acorn, but should be accurate for all other tokens.
         * @param {AcornToken} token The Acorn token to translate.
         * @param {Object} extra Espree extra object.
         * @returns {EsprimaToken} The Esprima version of the token.
         */
        translate(token, extra) {
            const type = token.type, tt = this._acornTokTypes;
            if (type === tt.name) {
                token.type = Token.Identifier;
                // TODO: See if this is an Acorn bug
                if (token.value === "static") {
                    token.type = Token.Keyword;
                }
                if (extra.ecmaVersion > 5 && (token.value === "yield" || token.value === "let")) {
                    token.type = Token.Keyword;
                }
            }
            else if (type === tt.semi || type === tt.comma ||
                type === tt.parenL || type === tt.parenR ||
                type === tt.braceL || type === tt.braceR ||
                type === tt.dot || type === tt.bracketL ||
                type === tt.colon || type === tt.question ||
                type === tt.bracketR || type === tt.ellipsis ||
                type === tt.arrow || type === tt.jsxTagStart ||
                type === tt.incDec || type === tt.starstar ||
                type === tt.jsxTagEnd || type === tt.prefix ||
                type === tt.questionDot ||
                (type.binop && !type.keyword) ||
                type.isAssign) {
                token.type = Token.Punctuator;
                token.value = this._code.slice(token.start, token.end);
            }
            else if (type === tt.jsxName) {
                token.type = Token.JSXIdentifier;
            }
            else if (type.label === "jsxText" || type === tt.jsxAttrValueToken) {
                token.type = Token.JSXText;
            }
            else if (type.keyword) {
                if (type.keyword === "true" || type.keyword === "false") {
                    token.type = Token.Boolean;
                }
                else if (type.keyword === "null") {
                    token.type = Token.Null;
                }
                else {
                    token.type = Token.Keyword;
                }
            }
            else if (type === tt.num) {
                token.type = Token.Numeric;
                token.value = this._code.slice(token.start, token.end);
            }
            else if (type === tt.string) {
                if (extra.jsxAttrValueToken) {
                    extra.jsxAttrValueToken = false;
                    token.type = Token.JSXText;
                }
                else {
                    token.type = Token.String;
                }
                token.value = this._code.slice(token.start, token.end);
            }
            else if (type === tt.regexp) {
                token.type = Token.RegularExpression;
                const value = token.value;
                token.regex = {
                    flags: value.flags,
                    pattern: value.pattern
                };
                token.value = `/${value.pattern}/${value.flags}`;
            }
            return token;
        },
        /**
         * Function to call during Acorn's onToken handler.
         * @param {AcornToken} token The Acorn token.
         * @param {Object} extra The Espree extra object.
         * @returns {void}
         */
        onToken(token, extra) {
            const that = this, tt = this._acornTokTypes, tokens = extra.tokens, templateTokens = this._tokens;
            /**
             * Flushes the buffered template tokens and resets the template
             * tracking.
             * @returns {void}
             * @private
             */
            function translateTemplateTokens() {
                tokens.push(convertTemplatePart(that._tokens, that._code));
                that._tokens = [];
            }
            if (token.type === tt.eof) {
                // might be one last curlyBrace
                if (this._curlyBrace) {
                    tokens.push(this.translate(this._curlyBrace, extra));
                }
                return;
            }
            if (token.type === tt.backQuote) {
                // if there's already a curly, it's not part of the template
                if (this._curlyBrace) {
                    tokens.push(this.translate(this._curlyBrace, extra));
                    this._curlyBrace = null;
                }
                templateTokens.push(token);
                // it's the end
                if (templateTokens.length > 1) {
                    translateTemplateTokens();
                }
                return;
            }
            if (token.type === tt.dollarBraceL) {
                templateTokens.push(token);
                translateTemplateTokens();
                return;
            }
            if (token.type === tt.braceR) {
                // if there's already a curly, it's not part of the template
                if (this._curlyBrace) {
                    tokens.push(this.translate(this._curlyBrace, extra));
                }
                // store new curly for later
                this._curlyBrace = token;
                return;
            }
            if (token.type === tt.template || token.type === tt.invalidTemplate) {
                if (this._curlyBrace) {
                    templateTokens.push(this._curlyBrace);
                    this._curlyBrace = null;
                }
                templateTokens.push(token);
                return;
            }
            if (this._curlyBrace) {
                tokens.push(this.translate(this._curlyBrace, extra));
                this._curlyBrace = null;
            }
            tokens.push(this.translate(token, extra));
        }
    };
    //------------------------------------------------------------------------------
    // Public
    //------------------------------------------------------------------------------
    module.exports = TokenTranslator;
},
/* espree/lib/options.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const DEFAULT_ECMA_VERSION = 5;
    const SUPPORTED_VERSIONS = [
        3,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12
    ];
    /**
     * Normalize ECMAScript version from the initial config
     * @param {number} ecmaVersion ECMAScript version from the initial config
     * @throws {Error} throws an error if the ecmaVersion is invalid.
     * @returns {number} normalized ECMAScript version
     */
    function normalizeEcmaVersion(ecmaVersion = DEFAULT_ECMA_VERSION) {
        if (typeof ecmaVersion !== "number") {
            throw new Error(`ecmaVersion must be a number. Received value of type ${typeof ecmaVersion} instead.`);
        }
        let version = ecmaVersion;
        // Calculate ECMAScript edition number from official year version starting with
        // ES2015, which corresponds with ES6 (or a difference of 2009).
        if (version >= 2015) {
            version -= 2009;
        }
        if (!SUPPORTED_VERSIONS.includes(version)) {
            throw new Error("Invalid ecmaVersion.");
        }
        return version;
    }
    /**
     * Normalize sourceType from the initial config
     * @param {string} sourceType to normalize
     * @throws {Error} throw an error if sourceType is invalid
     * @returns {string} normalized sourceType
     */
    function normalizeSourceType(sourceType = "script") {
        if (sourceType === "script" || sourceType === "module") {
            return sourceType;
        }
        throw new Error("Invalid sourceType.");
    }
    /**
     * Normalize parserOptions
     * @param {Object} options the parser options to normalize
     * @throws {Error} throw an error if found invalid option.
     * @returns {Object} normalized options
     */
    function normalizeOptions(options) {
        const ecmaVersion = normalizeEcmaVersion(options.ecmaVersion);
        const sourceType = normalizeSourceType(options.sourceType);
        const ranges = options.range === true;
        const locations = options.loc === true;
        if (sourceType === "module" && ecmaVersion < 6) {
            throw new Error("sourceType 'module' is not supported when ecmaVersion < 2015. Consider adding `{ ecmaVersion: 2015 }` to the parser options.");
        }
        return Object.assign({}, options, { ecmaVersion, sourceType, ranges, locations });
    }
    /**
     * Get the latest ECMAScript version supported by Espree.
     * @returns {number} The latest ECMAScript version.
     */
    function getLatestEcmaVersion() {
        return SUPPORTED_VERSIONS[SUPPORTED_VERSIONS.length - 1];
    }
    /**
     * Get the list of ECMAScript versions supported by Espree.
     * @returns {number[]} An array containing the supported ECMAScript versions.
     */
    function getSupportedEcmaVersions() {
        return [...SUPPORTED_VERSIONS];
    }
    //------------------------------------------------------------------------------
    // Public
    //------------------------------------------------------------------------------
    module.exports = {
        normalizeOptions,
        getLatestEcmaVersion,
        getSupportedEcmaVersions
    };
},
/* espree/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "espree",
        "description": "An Esprima-compatible JavaScript parser built on Acorn",
        "author": "Nicholas C. Zakas <nicholas+npm@nczconsulting.com>",
        "homepage": "https://github.com/eslint/espree",
        "main": "espree.js",
        "version": "7.3.1",
        "files": [
            "lib",
            "espree.js"
        ],
        "engines": {
            "node": "^10.12.0 || >=12.0.0"
        },
        "repository": "eslint/espree",
        "bugs": {
            "url": "http://github.com/eslint/espree.git"
        },
        "license": "BSD-2-Clause",
        "dependencies": {
            "acorn": "^7.4.0",
            "acorn-jsx": "^5.3.1",
            "eslint-visitor-keys": "^1.3.0"
        },
        "devDependencies": {
            "browserify": "^16.5.0",
            "chai": "^4.2.0",
            "eslint": "^6.0.1",
            "eslint-config-eslint": "^5.0.1",
            "eslint-plugin-node": "^9.1.0",
            "eslint-release": "^1.0.0",
            "esprima": "latest",
            "esprima-fb": "^8001.2001.0-dev-harmony-fb",
            "json-diff": "^0.5.4",
            "leche": "^2.3.0",
            "mocha": "^6.2.0",
            "nyc": "^14.1.1",
            "regenerate": "^1.4.0",
            "shelljs": "^0.3.0",
            "shelljs-nodecli": "^0.1.1",
            "unicode-6.3.0": "^0.7.5"
        },
        "keywords": [
            "ast",
            "ecmascript",
            "javascript",
            "parser",
            "syntax",
            "acorn"
        ],
        "scripts": {
            "generate-regex": "node tools/generate-identifier-regex.js",
            "test": "npm run-script lint && node Makefile.js test",
            "lint": "node Makefile.js lint",
            "fixlint": "node Makefile.js lint --fix",
            "sync-docs": "node Makefile.js docs",
            "browserify": "node Makefile.js browserify",
            "generate-release": "eslint-generate-release",
            "generate-alpharelease": "eslint-generate-prerelease alpha",
            "generate-betarelease": "eslint-generate-prerelease beta",
            "generate-rcrelease": "eslint-generate-prerelease rc",
            "publish-release": "eslint-publish-release"
        }
    };
    module.exports = json;
},
/* eslint-visitor-keys/lib/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const KEYS = require(322) /* ./visitor-keys.json */;
    // Types.
    const NODE_TYPES = Object.freeze(Object.keys(KEYS));
    // Freeze the keys.
    for (const type of NODE_TYPES) {
        Object.freeze(KEYS[type]);
    }
    Object.freeze(KEYS);
    // List to ignore keys.
    const KEY_BLACKLIST = new Set([
        "parent",
        "leadingComments",
        "trailingComments"
    ]);
    /**
     * Check whether a given key should be used or not.
     * @param {string} key The key to check.
     * @returns {boolean} `true` if the key should be used.
     */
    function filterKey(key) {
        return !KEY_BLACKLIST.has(key) && key[0] !== "_";
    }
    //------------------------------------------------------------------------------
    // Public interfaces
    //------------------------------------------------------------------------------
    module.exports = Object.freeze({
        /**
         * Visitor keys.
         * @type {{ [type: string]: string[] | undefined }}
         */
        KEYS,
        /**
         * Get visitor keys of a given node.
         * @param {Object} node The AST node to get keys.
         * @returns {string[]} Visitor keys of the node.
         */
        getKeys(node) {
            return Object.keys(node).filter(filterKey);
        },
        // Disable valid-jsdoc rule because it reports syntax error on the type of @returns.
        // eslint-disable-next-line valid-jsdoc
        /**
         * Make the union set with `KEYS` and given keys.
         * @param {Object} additionalKeys The additional keys.
         * @returns {{ [type: string]: string[] | undefined }} The union set.
         */
        unionWith(additionalKeys) {
            const retv = Object.assign({}, KEYS);
            for (const type of Object.keys(additionalKeys)) {
                if (retv.hasOwnProperty(type)) {
                    const keys = new Set(additionalKeys[type]);
                    for (const key of retv[type]) {
                        keys.add(key);
                    }
                    retv[type] = Object.freeze(Array.from(keys));
                }
                else {
                    retv[type] = Object.freeze(Array.from(additionalKeys[type]));
                }
            }
            return Object.freeze(retv);
        }
    });
},
/* eslint-visitor-keys/lib/visitor-keys.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "AssignmentExpression": [
            "left",
            "right"
        ],
        "AssignmentPattern": [
            "left",
            "right"
        ],
        "ArrayExpression": [
            "elements"
        ],
        "ArrayPattern": [
            "elements"
        ],
        "ArrowFunctionExpression": [
            "params",
            "body"
        ],
        "AwaitExpression": [
            "argument"
        ],
        "BlockStatement": [
            "body"
        ],
        "BinaryExpression": [
            "left",
            "right"
        ],
        "BreakStatement": [
            "label"
        ],
        "CallExpression": [
            "callee",
            "arguments"
        ],
        "CatchClause": [
            "param",
            "body"
        ],
        "ChainExpression": [
            "expression"
        ],
        "ClassBody": [
            "body"
        ],
        "ClassDeclaration": [
            "id",
            "superClass",
            "body"
        ],
        "ClassExpression": [
            "id",
            "superClass",
            "body"
        ],
        "ConditionalExpression": [
            "test",
            "consequent",
            "alternate"
        ],
        "ContinueStatement": [
            "label"
        ],
        "DebuggerStatement": [],
        "DoWhileStatement": [
            "body",
            "test"
        ],
        "EmptyStatement": [],
        "ExportAllDeclaration": [
            "exported",
            "source"
        ],
        "ExportDefaultDeclaration": [
            "declaration"
        ],
        "ExportNamedDeclaration": [
            "declaration",
            "specifiers",
            "source"
        ],
        "ExportSpecifier": [
            "exported",
            "local"
        ],
        "ExpressionStatement": [
            "expression"
        ],
        "ExperimentalRestProperty": [
            "argument"
        ],
        "ExperimentalSpreadProperty": [
            "argument"
        ],
        "ForStatement": [
            "init",
            "test",
            "update",
            "body"
        ],
        "ForInStatement": [
            "left",
            "right",
            "body"
        ],
        "ForOfStatement": [
            "left",
            "right",
            "body"
        ],
        "FunctionDeclaration": [
            "id",
            "params",
            "body"
        ],
        "FunctionExpression": [
            "id",
            "params",
            "body"
        ],
        "Identifier": [],
        "IfStatement": [
            "test",
            "consequent",
            "alternate"
        ],
        "ImportDeclaration": [
            "specifiers",
            "source"
        ],
        "ImportDefaultSpecifier": [
            "local"
        ],
        "ImportExpression": [
            "source"
        ],
        "ImportNamespaceSpecifier": [
            "local"
        ],
        "ImportSpecifier": [
            "imported",
            "local"
        ],
        "JSXAttribute": [
            "name",
            "value"
        ],
        "JSXClosingElement": [
            "name"
        ],
        "JSXElement": [
            "openingElement",
            "children",
            "closingElement"
        ],
        "JSXEmptyExpression": [],
        "JSXExpressionContainer": [
            "expression"
        ],
        "JSXIdentifier": [],
        "JSXMemberExpression": [
            "object",
            "property"
        ],
        "JSXNamespacedName": [
            "namespace",
            "name"
        ],
        "JSXOpeningElement": [
            "name",
            "attributes"
        ],
        "JSXSpreadAttribute": [
            "argument"
        ],
        "JSXText": [],
        "JSXFragment": [
            "openingFragment",
            "children",
            "closingFragment"
        ],
        "Literal": [],
        "LabeledStatement": [
            "label",
            "body"
        ],
        "LogicalExpression": [
            "left",
            "right"
        ],
        "MemberExpression": [
            "object",
            "property"
        ],
        "MetaProperty": [
            "meta",
            "property"
        ],
        "MethodDefinition": [
            "key",
            "value"
        ],
        "NewExpression": [
            "callee",
            "arguments"
        ],
        "ObjectExpression": [
            "properties"
        ],
        "ObjectPattern": [
            "properties"
        ],
        "PrivateIdentifier": [],
        "Program": [
            "body"
        ],
        "Property": [
            "key",
            "value"
        ],
        "PropertyDefinition": [
            "key",
            "value"
        ],
        "RestElement": [
            "argument"
        ],
        "ReturnStatement": [
            "argument"
        ],
        "SequenceExpression": [
            "expressions"
        ],
        "SpreadElement": [
            "argument"
        ],
        "Super": [],
        "SwitchStatement": [
            "discriminant",
            "cases"
        ],
        "SwitchCase": [
            "test",
            "consequent"
        ],
        "TaggedTemplateExpression": [
            "tag",
            "quasi"
        ],
        "TemplateElement": [],
        "TemplateLiteral": [
            "quasis",
            "expressions"
        ],
        "ThisExpression": [],
        "ThrowStatement": [
            "argument"
        ],
        "TryStatement": [
            "block",
            "handler",
            "finalizer"
        ],
        "UnaryExpression": [
            "argument"
        ],
        "UpdateExpression": [
            "argument"
        ],
        "VariableDeclaration": [
            "declarations"
        ],
        "VariableDeclarator": [
            "id",
            "init"
        ],
        "WhileStatement": [
            "test",
            "body"
        ],
        "WithStatement": [
            "object",
            "body"
        ],
        "YieldExpression": [
            "argument"
        ]
    };
    module.exports = json;
},
/* @eslint/eslintrc/lib/cascading-config-array-factory.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const os = require("os");
    const path = require("path");
    const ConfigValidator = require(215) /* ./shared/config-validator */;
    const { emitDeprecationWarning } = require(221) /* ./shared/deprecation-warnings */;
    const { ConfigArrayFactory } = require(209) /* ./config-array-factory */;
    const { ConfigArray, ConfigDependency, IgnorePattern } = require(272) /* ./config-array */;
    const debug = require(72) /* debug */("eslintrc:cascading-config-array-factory");
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    // Define types for VSCode IntelliSense.
    /** @typedef {import("./shared/types").ConfigData} ConfigData */
    /** @typedef {import("./shared/types").Parser} Parser */
    /** @typedef {import("./shared/types").Plugin} Plugin */
    /** @typedef {import("./shared/types").Rule} Rule */
    /** @typedef {ReturnType<ConfigArrayFactory["create"]>} ConfigArray */
    /**
     * @typedef {Object} CascadingConfigArrayFactoryOptions
     * @property {Map<string,Plugin>} [additionalPluginPool] The map for additional plugins.
     * @property {ConfigData} [baseConfig] The config by `baseConfig` option.
     * @property {ConfigData} [cliConfig] The config by CLI options (`--env`, `--global`, `--ignore-pattern`, `--parser`, `--parser-options`, `--plugin`, and `--rule`). CLI options overwrite the setting in config files.
     * @property {string} [cwd] The base directory to start lookup.
     * @property {string} [ignorePath] The path to the alternative file of `.eslintignore`.
     * @property {string[]} [rulePaths] The value of `--rulesdir` option.
     * @property {string} [specificConfigPath] The value of `--config` option.
     * @property {boolean} [useEslintrc] if `false` then it doesn't load config files.
     * @property {Function} loadRules The function to use to load rules.
     * @property {Map<string,Rule>} builtInRules The rules that are built in to ESLint.
     * @property {Object} [resolver=ModuleResolver] The module resolver object.
     * @property {string} eslintAllPath The path to the definitions for eslint:all.
     * @property {string} eslintRecommendedPath The path to the definitions for eslint:recommended.
     */
    /**
     * @typedef {Object} CascadingConfigArrayFactoryInternalSlots
     * @property {ConfigArray} baseConfigArray The config array of `baseConfig` option.
     * @property {ConfigData} baseConfigData The config data of `baseConfig` option. This is used to reset `baseConfigArray`.
     * @property {ConfigArray} cliConfigArray The config array of CLI options.
     * @property {ConfigData} cliConfigData The config data of CLI options. This is used to reset `cliConfigArray`.
     * @property {ConfigArrayFactory} configArrayFactory The factory for config arrays.
     * @property {Map<string, ConfigArray>} configCache The cache from directory paths to config arrays.
     * @property {string} cwd The base directory to start lookup.
     * @property {WeakMap<ConfigArray, ConfigArray>} finalizeCache The cache from config arrays to finalized config arrays.
     * @property {string} [ignorePath] The path to the alternative file of `.eslintignore`.
     * @property {string[]|null} rulePaths The value of `--rulesdir` option. This is used to reset `baseConfigArray`.
     * @property {string|null} specificConfigPath The value of `--config` option. This is used to reset `cliConfigArray`.
     * @property {boolean} useEslintrc if `false` then it doesn't load config files.
     * @property {Function} loadRules The function to use to load rules.
     * @property {Map<string,Rule>} builtInRules The rules that are built in to ESLint.
     * @property {Object} [resolver=ModuleResolver] The module resolver object.
     * @property {string} eslintAllPath The path to the definitions for eslint:all.
     * @property {string} eslintRecommendedPath The path to the definitions for eslint:recommended.
     */
    /** @type {WeakMap<CascadingConfigArrayFactory, CascadingConfigArrayFactoryInternalSlots>} */
    const internalSlotsMap = new WeakMap();
    /**
     * Create the config array from `baseConfig` and `rulePaths`.
     * @param {CascadingConfigArrayFactoryInternalSlots} slots The slots.
     * @returns {ConfigArray} The config array of the base configs.
     */
    function createBaseConfigArray({ configArrayFactory, baseConfigData, rulePaths, cwd, loadRules }) {
        const baseConfigArray = configArrayFactory.create(baseConfigData, { name: "BaseConfig" });
        /*
         * Create the config array element for the default ignore patterns.
         * This element has `ignorePattern` property that ignores the default
         * patterns in the current working directory.
         */
        baseConfigArray.unshift(configArrayFactory.create({ ignorePatterns: IgnorePattern.DefaultPatterns }, { name: "DefaultIgnorePattern" })[0]);
        /*
         * Load rules `--rulesdir` option as a pseudo plugin.
         * Use a pseudo plugin to define rules of `--rulesdir`, so we can validate
         * the rule's options with only information in the config array.
         */
        if (rulePaths && rulePaths.length > 0) {
            baseConfigArray.push({
                type: "config",
                name: "--rulesdir",
                filePath: "",
                plugins: {
                    "": new ConfigDependency({
                        definition: {
                            rules: rulePaths.reduce((map, rulesPath) => Object.assign(map, loadRules(rulesPath, cwd)), {})
                        },
                        filePath: "",
                        id: "",
                        importerName: "--rulesdir",
                        importerPath: ""
                    })
                }
            });
        }
        return baseConfigArray;
    }
    /**
     * Create the config array from CLI options.
     * @param {CascadingConfigArrayFactoryInternalSlots} slots The slots.
     * @returns {ConfigArray} The config array of the base configs.
     */
    function createCLIConfigArray({ cliConfigData, configArrayFactory, cwd, ignorePath, specificConfigPath }) {
        const cliConfigArray = configArrayFactory.create(cliConfigData, { name: "CLIOptions" });
        cliConfigArray.unshift(...(ignorePath
            ? configArrayFactory.loadESLintIgnore(ignorePath)
            : configArrayFactory.loadDefaultESLintIgnore()));
        if (specificConfigPath) {
            cliConfigArray.unshift(...configArrayFactory.loadFile(specificConfigPath, { name: "--config", basePath: cwd }));
        }
        return cliConfigArray;
    }
    /**
     * The error type when there are files matched by a glob, but all of them have been ignored.
     */
    class ConfigurationNotFoundError extends Error {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {string} directoryPath The directory path.
         */
        constructor(directoryPath) {
            super(`No ESLint configuration found in ${directoryPath}.`);
            this.messageTemplate = "no-config-found";
            this.messageData = { directoryPath };
        }
    }
    /**
     * This class provides the functionality that enumerates every file which is
     * matched by given glob patterns and that configuration.
     */
    class CascadingConfigArrayFactory {
        /**
         * Initialize this enumerator.
         * @param {CascadingConfigArrayFactoryOptions} options The options.
         */
        constructor({ additionalPluginPool = new Map(), baseConfig: baseConfigData = null, cliConfig: cliConfigData = null, cwd = process.cwd(), ignorePath, resolvePluginsRelativeTo, rulePaths = [], specificConfigPath = null, useEslintrc = true, builtInRules = new Map(), loadRules, resolver, eslintRecommendedPath, eslintAllPath } = {}) {
            const configArrayFactory = new ConfigArrayFactory({
                additionalPluginPool,
                cwd,
                resolvePluginsRelativeTo,
                builtInRules,
                resolver,
                eslintRecommendedPath,
                eslintAllPath
            });
            internalSlotsMap.set(this, {
                baseConfigArray: createBaseConfigArray({
                    baseConfigData,
                    configArrayFactory,
                    cwd,
                    rulePaths,
                    loadRules,
                    resolver
                }),
                baseConfigData,
                cliConfigArray: createCLIConfigArray({
                    cliConfigData,
                    configArrayFactory,
                    cwd,
                    ignorePath,
                    specificConfigPath
                }),
                cliConfigData,
                configArrayFactory,
                configCache: new Map(),
                cwd,
                finalizeCache: new WeakMap(),
                ignorePath,
                rulePaths,
                specificConfigPath,
                useEslintrc,
                builtInRules,
                loadRules
            });
        }
        /**
         * The path to the current working directory.
         * This is used by tests.
         * @type {string}
         */
        get cwd() {
            const { cwd } = internalSlotsMap.get(this);
            return cwd;
        }
        /**
         * Get the config array of a given file.
         * If `filePath` was not given, it returns the config which contains only
         * `baseConfigData` and `cliConfigData`.
         * @param {string} [filePath] The file path to a file.
         * @param {Object} [options] The options.
         * @param {boolean} [options.ignoreNotFoundError] If `true` then it doesn't throw `ConfigurationNotFoundError`.
         * @returns {ConfigArray} The config array of the file.
         */
        getConfigArrayForFile(filePath, { ignoreNotFoundError = false } = {}) {
            const { baseConfigArray, cliConfigArray, cwd } = internalSlotsMap.get(this);
            if (!filePath) {
                return new ConfigArray(...baseConfigArray, ...cliConfigArray);
            }
            const directoryPath = path.dirname(path.resolve(cwd, filePath));
            debug(`Load config files for ${directoryPath}.`);
            return this._finalizeConfigArray(this._loadConfigInAncestors(directoryPath), directoryPath, ignoreNotFoundError);
        }
        /**
         * Set the config data to override all configs.
         * Require to call `clearCache()` method after this method is called.
         * @param {ConfigData} configData The config data to override all configs.
         * @returns {void}
         */
        setOverrideConfig(configData) {
            const slots = internalSlotsMap.get(this);
            slots.cliConfigData = configData;
        }
        /**
         * Clear config cache.
         * @returns {void}
         */
        clearCache() {
            const slots = internalSlotsMap.get(this);
            slots.baseConfigArray = createBaseConfigArray(slots);
            slots.cliConfigArray = createCLIConfigArray(slots);
            slots.configCache.clear();
        }
        /**
         * Load and normalize config files from the ancestor directories.
         * @param {string} directoryPath The path to a leaf directory.
         * @param {boolean} configsExistInSubdirs `true` if configurations exist in subdirectories.
         * @returns {ConfigArray} The loaded config.
         * @private
         */
        _loadConfigInAncestors(directoryPath, configsExistInSubdirs = false) {
            const { baseConfigArray, configArrayFactory, configCache, cwd, useEslintrc } = internalSlotsMap.get(this);
            if (!useEslintrc) {
                return baseConfigArray;
            }
            let configArray = configCache.get(directoryPath);
            // Hit cache.
            if (configArray) {
                debug(`Cache hit: ${directoryPath}.`);
                return configArray;
            }
            debug(`No cache found: ${directoryPath}.`);
            const homePath = os.homedir();
            // Consider this is root.
            if (directoryPath === homePath && cwd !== homePath) {
                debug("Stop traversing because of considered root.");
                if (configsExistInSubdirs) {
                    const filePath = ConfigArrayFactory.getPathToConfigFileInDirectory(directoryPath);
                    if (filePath) {
                        emitDeprecationWarning(filePath, "ESLINT_PERSONAL_CONFIG_SUPPRESS");
                    }
                }
                return this._cacheConfig(directoryPath, baseConfigArray);
            }
            // Load the config on this directory.
            try {
                configArray = configArrayFactory.loadInDirectory(directoryPath);
            }
            catch (error) {
                /* istanbul ignore next */
                if (error.code === "EACCES") {
                    debug("Stop traversing because of 'EACCES' error.");
                    return this._cacheConfig(directoryPath, baseConfigArray);
                }
                throw error;
            }
            if (configArray.length > 0 && configArray.isRoot()) {
                debug("Stop traversing because of 'root:true'.");
                configArray.unshift(...baseConfigArray);
                return this._cacheConfig(directoryPath, configArray);
            }
            // Load from the ancestors and merge it.
            const parentPath = path.dirname(directoryPath);
            const parentConfigArray = parentPath && parentPath !== directoryPath
                ? this._loadConfigInAncestors(parentPath, configsExistInSubdirs || configArray.length > 0)
                : baseConfigArray;
            if (configArray.length > 0) {
                configArray.unshift(...parentConfigArray);
            }
            else {
                configArray = parentConfigArray;
            }
            // Cache and return.
            return this._cacheConfig(directoryPath, configArray);
        }
        /**
         * Freeze and cache a given config.
         * @param {string} directoryPath The path to a directory as a cache key.
         * @param {ConfigArray} configArray The config array as a cache value.
         * @returns {ConfigArray} The `configArray` (frozen).
         */
        _cacheConfig(directoryPath, configArray) {
            const { configCache } = internalSlotsMap.get(this);
            Object.freeze(configArray);
            configCache.set(directoryPath, configArray);
            return configArray;
        }
        /**
         * Finalize a given config array.
         * Concatenate `--config` and other CLI options.
         * @param {ConfigArray} configArray The parent config array.
         * @param {string} directoryPath The path to the leaf directory to find config files.
         * @param {boolean} ignoreNotFoundError If `true` then it doesn't throw `ConfigurationNotFoundError`.
         * @returns {ConfigArray} The loaded config.
         * @private
         */
        _finalizeConfigArray(configArray, directoryPath, ignoreNotFoundError) {
            const { cliConfigArray, configArrayFactory, finalizeCache, useEslintrc, builtInRules } = internalSlotsMap.get(this);
            let finalConfigArray = finalizeCache.get(configArray);
            if (!finalConfigArray) {
                finalConfigArray = configArray;
                // Load the personal config if there are no regular config files.
                if (useEslintrc &&
                    configArray.every(c => !c.filePath) &&
                    cliConfigArray.every(c => !c.filePath) // `--config` option can be a file.
                ) {
                    const homePath = os.homedir();
                    debug("Loading the config file of the home directory:", homePath);
                    const personalConfigArray = configArrayFactory.loadInDirectory(homePath, { name: "PersonalConfig" });
                    if (personalConfigArray.length > 0 &&
                        !directoryPath.startsWith(homePath)) {
                        const lastElement = personalConfigArray[personalConfigArray.length - 1];
                        emitDeprecationWarning(lastElement.filePath, "ESLINT_PERSONAL_CONFIG_LOAD");
                    }
                    finalConfigArray = finalConfigArray.concat(personalConfigArray);
                }
                // Apply CLI options.
                if (cliConfigArray.length > 0) {
                    finalConfigArray = finalConfigArray.concat(cliConfigArray);
                }
                // Validate rule settings and environments.
                const validator = new ConfigValidator({
                    builtInRules
                });
                validator.validateConfigArray(finalConfigArray);
                // Cache it.
                Object.freeze(finalConfigArray);
                finalizeCache.set(configArray, finalConfigArray);
                debug("Configuration was determined: %o on %s", finalConfigArray, directoryPath);
            }
            // At least one element (the default ignore patterns) exists.
            if (!ignoreNotFoundError && useEslintrc && finalConfigArray.length <= 1) {
                throw new ConfigurationNotFoundError(directoryPath);
            }
            return finalConfigArray;
        }
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = { CascadingConfigArrayFactory };
},
/* @eslint/eslintrc/lib/flat-compat.js */ function _(require, module, exports, __esModule, __esExport) {
    //-----------------------------------------------------------------------------
    // Requirements
    //-----------------------------------------------------------------------------
    const path = require("path");
    const environments = require(217) /* ../conf/environments */;
    const createDebug = require(72) /* debug */;
    const { ConfigArrayFactory } = require(209) /* ./config-array-factory */;
    //-----------------------------------------------------------------------------
    // Helpers
    //-----------------------------------------------------------------------------
    /** @typedef {import("../../shared/types").Environment} Environment */
    /** @typedef {import("../../shared/types").Processor} Processor */
    const debug = createDebug("eslintrc:flat-compat");
    const cafactory = Symbol("cafactory");
    /**
     * Translates an ESLintRC-style config object into a flag-config-style config
     * object.
     * @param {Object} eslintrcConfig An ESLintRC-style config object.
     * @param {Object} options Options to help translate the config.
     * @param {string} options.resolveConfigRelativeTo To the directory to resolve
     *      configs from.
     * @param {string} options.resolvePluginsRelativeTo The directory to resolve
     *      plugins from.
     * @param {ReadOnlyMap<string,Environment>} options.pluginEnvironments A map of plugin environment
     *      names to objects.
     * @param {ReadOnlyMap<string,Processor>} options.pluginProcessors A map of plugin processor
     *      names to objects.
     * @returns {Object} A flag-config-style config object.
     */
    function translateESLintRC(eslintrcConfig, { resolveConfigRelativeTo, resolvePluginsRelativeTo, pluginEnvironments, pluginProcessors }) {
        const flatConfig = {};
        const configs = [];
        const languageOptions = {};
        const linterOptions = {};
        const keysToCopy = ["settings", "rules", "processor"];
        const languageOptionsKeysToCopy = ["globals", "parser", "parserOptions"];
        const linterOptionsKeysToCopy = ["noInlineConfig", "reportUnusedDisableDirectives"];
        // check for special settings for eslint:all and eslint:recommended:
        if (eslintrcConfig.settings) {
            if (eslintrcConfig.settings["eslint:all"] === true) {
                return ["eslint:all"];
            }
            if (eslintrcConfig.settings["eslint:recommended"] === true) {
                return ["eslint:recommended"];
            }
        }
        // copy over simple translations
        for (const key of keysToCopy) {
            if (key in eslintrcConfig && typeof eslintrcConfig[key] !== "undefined") {
                flatConfig[key] = eslintrcConfig[key];
            }
        }
        // copy over languageOptions
        for (const key of languageOptionsKeysToCopy) {
            if (key in eslintrcConfig && typeof eslintrcConfig[key] !== "undefined") {
                // create the languageOptions key in the flat config
                flatConfig.languageOptions = languageOptions;
                if (key === "parser") {
                    debug(`Resolving parser '${languageOptions[key]}' relative to ${resolveConfigRelativeTo}`);
                    if (eslintrcConfig[key].error) {
                        throw eslintrcConfig[key].error;
                    }
                    languageOptions[key] = eslintrcConfig[key].definition;
                    continue;
                }
                // clone any object values that are in the eslintrc config
                if (eslintrcConfig[key] && typeof eslintrcConfig[key] === "object") {
                    languageOptions[key] = {
                        ...eslintrcConfig[key]
                    };
                }
                else {
                    languageOptions[key] = eslintrcConfig[key];
                }
            }
        }
        // copy over linterOptions
        for (const key of linterOptionsKeysToCopy) {
            if (key in eslintrcConfig && typeof eslintrcConfig[key] !== "undefined") {
                flatConfig.linterOptions = linterOptions;
                linterOptions[key] = eslintrcConfig[key];
            }
        }
        // move ecmaVersion a level up
        if (languageOptions.parserOptions) {
            if ("ecmaVersion" in languageOptions.parserOptions) {
                languageOptions.ecmaVersion = languageOptions.parserOptions.ecmaVersion;
                delete languageOptions.parserOptions.ecmaVersion;
            }
            if ("sourceType" in languageOptions.parserOptions) {
                languageOptions.sourceType = languageOptions.parserOptions.sourceType;
                delete languageOptions.parserOptions.sourceType;
            }
            // check to see if we even need parserOptions anymore and remove it if not
            if (Object.keys(languageOptions.parserOptions).length === 0) {
                delete languageOptions.parserOptions;
            }
        }
        // overrides
        if (eslintrcConfig.criteria) {
            flatConfig.files = [absoluteFilePath => eslintrcConfig.criteria.test(absoluteFilePath)];
        }
        // translate plugins
        if (eslintrcConfig.plugins && typeof eslintrcConfig.plugins === "object") {
            debug(`Translating plugins: ${eslintrcConfig.plugins}`);
            flatConfig.plugins = {};
            for (const pluginName of Object.keys(eslintrcConfig.plugins)) {
                debug(`Translating plugin: ${pluginName}`);
                debug(`Resolving plugin '${pluginName} relative to ${resolvePluginsRelativeTo}`);
                const { definition: plugin, error } = eslintrcConfig.plugins[pluginName];
                if (error) {
                    throw error;
                }
                flatConfig.plugins[pluginName] = plugin;
                // create a config for any processors
                if (plugin.processors) {
                    for (const processorName of Object.keys(plugin.processors)) {
                        if (processorName.startsWith(".")) {
                            debug(`Assigning processor: ${pluginName}/${processorName}`);
                            configs.unshift({
                                files: [`**/*${processorName}`],
                                processor: pluginProcessors.get(`${pluginName}/${processorName}`)
                            });
                        }
                    }
                }
            }
        }
        // translate env - must come after plugins
        if (eslintrcConfig.env && typeof eslintrcConfig.env === "object") {
            for (const envName of Object.keys(eslintrcConfig.env)) {
                // only add environments that are true
                if (eslintrcConfig.env[envName]) {
                    debug(`Translating environment: ${envName}`);
                    if (environments.has(envName)) {
                        // built-in environments should be defined first
                        configs.unshift(...translateESLintRC(environments.get(envName), {
                            resolveConfigRelativeTo,
                            resolvePluginsRelativeTo
                        }));
                    }
                    else if (pluginEnvironments.has(envName)) {
                        // if the environment comes from a plugin, it should come after the plugin config
                        configs.push(...translateESLintRC(pluginEnvironments.get(envName), {
                            resolveConfigRelativeTo,
                            resolvePluginsRelativeTo
                        }));
                    }
                }
            }
        }
        // only add if there are actually keys in the config
        if (Object.keys(flatConfig).length > 0) {
            configs.push(flatConfig);
        }
        return configs;
    }
    //-----------------------------------------------------------------------------
    // Exports
    //-----------------------------------------------------------------------------
    /**
     * A compatibility class for working with configs.
     */
    class FlatCompat {
        constructor({ baseDirectory = process.cwd(), resolvePluginsRelativeTo = baseDirectory } = {}) {
            this.baseDirectory = baseDirectory;
            this.resolvePluginsRelativeTo = resolvePluginsRelativeTo;
            this[cafactory] = new ConfigArrayFactory({
                cwd: baseDirectory,
                resolvePluginsRelativeTo,
                eslintAllPath: path.resolve(__dirname, "../conf/eslint-all.js"),
                eslintRecommendedPath: path.resolve(__dirname, "../conf/eslint-recommended.js")
            });
        }
        /**
         * Translates an ESLintRC-style config into a flag-config-style config.
         * @param {Object} eslintrcConfig The ESLintRC-style config object.
         * @returns {Object} A flag-config-style config object.
         */
        config(eslintrcConfig) {
            const eslintrcArray = this[cafactory].create(eslintrcConfig, {
                basePath: this.baseDirectory
            });
            const flatArray = [];
            let hasIgnorePatterns = false;
            eslintrcArray.forEach(configData => {
                if (configData.type === "config") {
                    hasIgnorePatterns = hasIgnorePatterns || configData.ignorePattern;
                    flatArray.push(...translateESLintRC(configData, {
                        resolveConfigRelativeTo: path.join(this.baseDirectory, "__placeholder.js"),
                        resolvePluginsRelativeTo: path.join(this.resolvePluginsRelativeTo, "__placeholder.js"),
                        pluginEnvironments: eslintrcArray.pluginEnvironments,
                        pluginProcessors: eslintrcArray.pluginProcessors
                    }));
                }
            });
            // combine ignorePatterns to emulate ESLintRC behavior better
            if (hasIgnorePatterns) {
                flatArray.unshift({
                    ignores: [filePath => {
                            // Compute the final config for this file.
                            // This filters config array elements by `files`/`excludedFiles` then merges the elements.
                            const finalConfig = eslintrcArray.extractConfig(filePath);
                            // Test the `ignorePattern` properties of the final config.
                            return Boolean(finalConfig.ignores) && finalConfig.ignores(filePath);
                        }]
                });
            }
            return flatArray;
        }
        /**
         * Translates the `env` section of an ESLintRC-style config.
         * @param {Object} envConfig The `env` section of an ESLintRC config.
         * @returns {Object} A flag-config object representing the environments.
         */
        env(envConfig) {
            return this.config({
                env: envConfig
            });
        }
        /**
         * Translates the `extends` section of an ESLintRC-style config.
         * @param {...string} configsToExtend The names of the configs to load.
         * @returns {Object} A flag-config object representing the config.
         */
        extends(...configsToExtend) {
            return this.config({
                extends: configsToExtend
            });
        }
        /**
         * Translates the `plugins` section of an ESLintRC-style config.
         * @param {...string} plugins The names of the plugins to load.
         * @returns {Object} A flag-config object representing the plugins.
         */
        plugins(...plugins) {
            return this.config({
                plugins
            });
        }
    }
    exports.FlatCompat = FlatCompat;
},
/* eslint/lib/cli-engine/file-enumerator.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const fs = require("fs");
    const path = require("path");
    const getGlobParent = require(326) /* glob-parent */;
    const isGlob = require(327) /* is-glob */;
    const escapeRegExp = require(329) /* escape-string-regexp */;
    const { Minimatch } = require(279) /* minimatch */;
    const { Legacy: { IgnorePattern, CascadingConfigArrayFactory } } = require(208) /* @eslint/eslintrc */;
    const debug = require(72) /* debug */("eslint:file-enumerator");
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const minimatchOpts = { dot: true, matchBase: true };
    const dotfilesPattern = /(?:(?:^\.)|(?:[/\\]\.))[^/\\.].*/u;
    const NONE = 0;
    const IGNORED_SILENTLY = 1;
    const IGNORED = 2;
    // For VSCode intellisense
    /** @typedef {ReturnType<CascadingConfigArrayFactory["getConfigArrayForFile"]>} ConfigArray */
    /**
     * @typedef {Object} FileEnumeratorOptions
     * @property {CascadingConfigArrayFactory} [configArrayFactory] The factory for config arrays.
     * @property {string} [cwd] The base directory to start lookup.
     * @property {string[]} [extensions] The extensions to match files for directory patterns.
     * @property {boolean} [globInputPaths] Set to false to skip glob resolution of input file paths to lint (default: true). If false, each input file paths is assumed to be a non-glob path to an existing file.
     * @property {boolean} [ignore] The flag to check ignored files.
     * @property {string[]} [rulePaths] The value of `--rulesdir` option.
     */
    /**
     * @typedef {Object} FileAndConfig
     * @property {string} filePath The path to a target file.
     * @property {ConfigArray} config The config entries of that file.
     * @property {boolean} ignored If `true` then this file should be ignored and warned because it was directly specified.
     */
    /**
     * @typedef {Object} FileEntry
     * @property {string} filePath The path to a target file.
     * @property {ConfigArray} config The config entries of that file.
     * @property {NONE|IGNORED_SILENTLY|IGNORED} flag The flag.
     * - `NONE` means the file is a target file.
     * - `IGNORED_SILENTLY` means the file should be ignored silently.
     * - `IGNORED` means the file should be ignored and warned because it was directly specified.
     */
    /**
     * @typedef {Object} FileEnumeratorInternalSlots
     * @property {CascadingConfigArrayFactory} configArrayFactory The factory for config arrays.
     * @property {string} cwd The base directory to start lookup.
     * @property {RegExp|null} extensionRegExp The RegExp to test if a string ends with specific file extensions.
     * @property {boolean} globInputPaths Set to false to skip glob resolution of input file paths to lint (default: true). If false, each input file paths is assumed to be a non-glob path to an existing file.
     * @property {boolean} ignoreFlag The flag to check ignored files.
     * @property {(filePath:string, dot:boolean) => boolean} defaultIgnores The default predicate function to ignore files.
     */
    /** @type {WeakMap<FileEnumerator, FileEnumeratorInternalSlots>} */
    const internalSlotsMap = new WeakMap();
    /**
     * Check if a string is a glob pattern or not.
     * @param {string} pattern A glob pattern.
     * @returns {boolean} `true` if the string is a glob pattern.
     */
    function isGlobPattern(pattern) {
        return isGlob(path.sep === "\\" ? pattern.replace(/\\/gu, "/") : pattern);
    }
    /**
     * Get stats of a given path.
     * @param {string} filePath The path to target file.
     * @returns {fs.Stats|null} The stats.
     * @private
     */
    function statSafeSync(filePath) {
        try {
            return fs.statSync(filePath);
        }
        catch (error) {
            /* istanbul ignore next */
            if (error.code !== "ENOENT") {
                throw error;
            }
            return null;
        }
    }
    /**
     * Get filenames in a given path to a directory.
     * @param {string} directoryPath The path to target directory.
     * @returns {import("fs").Dirent[]} The filenames.
     * @private
     */
    function readdirSafeSync(directoryPath) {
        try {
            return fs.readdirSync(directoryPath, { withFileTypes: true });
        }
        catch (error) {
            /* istanbul ignore next */
            if (error.code !== "ENOENT") {
                throw error;
            }
            return [];
        }
    }
    /**
     * Create a `RegExp` object to detect extensions.
     * @param {string[] | null} extensions The extensions to create.
     * @returns {RegExp | null} The created `RegExp` object or null.
     */
    function createExtensionRegExp(extensions) {
        if (extensions) {
            const normalizedExts = extensions.map(ext => escapeRegExp(ext.startsWith(".")
                ? ext.slice(1)
                : ext));
            return new RegExp(`.\\.(?:${normalizedExts.join("|")})$`, "u");
        }
        return null;
    }
    /**
     * The error type when no files match a glob.
     */
    class NoFilesFoundError extends Error {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {string} pattern The glob pattern which was not found.
         * @param {boolean} globDisabled If `true` then the pattern was a glob pattern, but glob was disabled.
         */
        constructor(pattern, globDisabled) {
            super(`No files matching '${pattern}' were found${globDisabled ? " (glob was disabled)" : ""}.`);
            this.messageTemplate = "file-not-found";
            this.messageData = { pattern, globDisabled };
        }
    }
    /**
     * The error type when there are files matched by a glob, but all of them have been ignored.
     */
    class AllFilesIgnoredError extends Error {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {string} pattern The glob pattern which was not found.
         */
        constructor(pattern) {
            super(`All files matched by '${pattern}' are ignored.`);
            this.messageTemplate = "all-files-ignored";
            this.messageData = { pattern };
        }
    }
    /**
     * This class provides the functionality that enumerates every file which is
     * matched by given glob patterns and that configuration.
     */
    class FileEnumerator {
        /**
         * Initialize this enumerator.
         * @param {FileEnumeratorOptions} options The options.
         */
        constructor({ cwd = process.cwd(), configArrayFactory = new CascadingConfigArrayFactory({
            cwd,
            eslintRecommendedPath: path.resolve(__dirname, "../../conf/eslint-recommended.js"),
            eslintAllPath: path.resolve(__dirname, "../../conf/eslint-all.js")
        }), extensions = null, globInputPaths = true, errorOnUnmatchedPattern = true, ignore = true } = {}) {
            internalSlotsMap.set(this, {
                configArrayFactory,
                cwd,
                defaultIgnores: IgnorePattern.createDefaultIgnore(cwd),
                extensionRegExp: createExtensionRegExp(extensions),
                globInputPaths,
                errorOnUnmatchedPattern,
                ignoreFlag: ignore
            });
        }
        /**
         * Check if a given file is target or not.
         * @param {string} filePath The path to a candidate file.
         * @param {ConfigArray} [providedConfig] Optional. The configuration for the file.
         * @returns {boolean} `true` if the file is a target.
         */
        isTargetPath(filePath, providedConfig) {
            const { configArrayFactory, extensionRegExp } = internalSlotsMap.get(this);
            // If `--ext` option is present, use it.
            if (extensionRegExp) {
                return extensionRegExp.test(filePath);
            }
            // `.js` file is target by default.
            if (filePath.endsWith(".js")) {
                return true;
            }
            // use `overrides[].files` to check additional targets.
            const config = providedConfig ||
                configArrayFactory.getConfigArrayForFile(filePath, { ignoreNotFoundError: true });
            return config.isAdditionalTargetPath(filePath);
        }
        /**
         * Iterate files which are matched by given glob patterns.
         * @param {string|string[]} patternOrPatterns The glob patterns to iterate files.
         * @returns {IterableIterator<FileAndConfig>} The found files.
         */
        *iterateFiles(patternOrPatterns) {
            const { globInputPaths, errorOnUnmatchedPattern } = internalSlotsMap.get(this);
            const patterns = Array.isArray(patternOrPatterns)
                ? patternOrPatterns
                : [patternOrPatterns];
            debug("Start to iterate files: %o", patterns);
            // The set of paths to remove duplicate.
            const set = new Set();
            for (const pattern of patterns) {
                let foundRegardlessOfIgnored = false;
                let found = false;
                // Skip empty string.
                if (!pattern) {
                    continue;
                }
                // Iterate files of this pattern.
                for (const { config, filePath, flag } of this._iterateFiles(pattern)) {
                    foundRegardlessOfIgnored = true;
                    if (flag === IGNORED_SILENTLY) {
                        continue;
                    }
                    found = true;
                    // Remove duplicate paths while yielding paths.
                    if (!set.has(filePath)) {
                        set.add(filePath);
                        yield {
                            config,
                            filePath,
                            ignored: flag === IGNORED
                        };
                    }
                }
                // Raise an error if any files were not found.
                if (errorOnUnmatchedPattern) {
                    if (!foundRegardlessOfIgnored) {
                        throw new NoFilesFoundError(pattern, !globInputPaths && isGlob(pattern));
                    }
                    if (!found) {
                        throw new AllFilesIgnoredError(pattern);
                    }
                }
            }
            debug(`Complete iterating files: ${JSON.stringify(patterns)}`);
        }
        /**
         * Iterate files which are matched by a given glob pattern.
         * @param {string} pattern The glob pattern to iterate files.
         * @returns {IterableIterator<FileEntry>} The found files.
         */
        _iterateFiles(pattern) {
            const { cwd, globInputPaths } = internalSlotsMap.get(this);
            const absolutePath = path.resolve(cwd, pattern);
            const isDot = dotfilesPattern.test(pattern);
            const stat = statSafeSync(absolutePath);
            if (stat && stat.isDirectory()) {
                return this._iterateFilesWithDirectory(absolutePath, isDot);
            }
            if (stat && stat.isFile()) {
                return this._iterateFilesWithFile(absolutePath);
            }
            if (globInputPaths && isGlobPattern(pattern)) {
                return this._iterateFilesWithGlob(absolutePath, isDot);
            }
            return [];
        }
        /**
         * Iterate a file which is matched by a given path.
         * @param {string} filePath The path to the target file.
         * @returns {IterableIterator<FileEntry>} The found files.
         * @private
         */
        _iterateFilesWithFile(filePath) {
            debug(`File: ${filePath}`);
            const { configArrayFactory } = internalSlotsMap.get(this);
            const config = configArrayFactory.getConfigArrayForFile(filePath);
            const ignored = this._isIgnoredFile(filePath, { config, direct: true });
            const flag = ignored ? IGNORED : NONE;
            return [{ config, filePath, flag }];
        }
        /**
         * Iterate files in a given path.
         * @param {string} directoryPath The path to the target directory.
         * @param {boolean} dotfiles If `true` then it doesn't skip dot files by default.
         * @returns {IterableIterator<FileEntry>} The found files.
         * @private
         */
        _iterateFilesWithDirectory(directoryPath, dotfiles) {
            debug(`Directory: ${directoryPath}`);
            return this._iterateFilesRecursive(directoryPath, { dotfiles, recursive: true, selector: null });
        }
        /**
         * Iterate files which are matched by a given glob pattern.
         * @param {string} pattern The glob pattern to iterate files.
         * @param {boolean} dotfiles If `true` then it doesn't skip dot files by default.
         * @returns {IterableIterator<FileEntry>} The found files.
         * @private
         */
        _iterateFilesWithGlob(pattern, dotfiles) {
            debug(`Glob: ${pattern}`);
            const directoryPath = path.resolve(getGlobParent(pattern));
            const globPart = pattern.slice(directoryPath.length + 1);
            /*
             * recursive if there are `**` or path separators in the glob part.
             * Otherwise, patterns such as `src/*.js`, it doesn't need recursive.
             */
            const recursive = /\*\*|\/|\\/u.test(globPart);
            const selector = new Minimatch(pattern, minimatchOpts);
            debug(`recursive? ${recursive}`);
            return this._iterateFilesRecursive(directoryPath, { dotfiles, recursive, selector });
        }
        /**
         * Iterate files in a given path.
         * @param {string} directoryPath The path to the target directory.
         * @param {Object} options The options to iterate files.
         * @param {boolean} [options.dotfiles] If `true` then it doesn't skip dot files by default.
         * @param {boolean} [options.recursive] If `true` then it dives into sub directories.
         * @param {InstanceType<Minimatch>} [options.selector] The matcher to choose files.
         * @returns {IterableIterator<FileEntry>} The found files.
         * @private
         */
        *_iterateFilesRecursive(directoryPath, options) {
            debug(`Enter the directory: ${directoryPath}`);
            const { configArrayFactory } = internalSlotsMap.get(this);
            /** @type {ConfigArray|null} */
            let config = null;
            // Enumerate the files of this directory.
            for (const entry of readdirSafeSync(directoryPath)) {
                const filePath = path.join(directoryPath, entry.name);
                const fileInfo = entry.isSymbolicLink() ? statSafeSync(filePath) : entry;
                if (!fileInfo) {
                    continue;
                }
                // Check if the file is matched.
                if (fileInfo.isFile()) {
                    if (!config) {
                        config = configArrayFactory.getConfigArrayForFile(filePath, 
                        /*
                         * We must ignore `ConfigurationNotFoundError` at this
                         * point because we don't know if target files exist in
                         * this directory.
                         */
                        { ignoreNotFoundError: true });
                    }
                    const matched = options.selector
                        // Started with a glob pattern; choose by the pattern.
                        ? options.selector.match(filePath)
                        // Started with a directory path; choose by file extensions.
                        : this.isTargetPath(filePath, config);
                    if (matched) {
                        const ignored = this._isIgnoredFile(filePath, { ...options, config });
                        const flag = ignored ? IGNORED_SILENTLY : NONE;
                        debug(`Yield: ${entry.name}${ignored ? " but ignored" : ""}`);
                        yield {
                            config: configArrayFactory.getConfigArrayForFile(filePath),
                            filePath,
                            flag
                        };
                    }
                    else {
                        debug(`Didn't match: ${entry.name}`);
                    }
                    // Dive into the sub directory.
                }
                else if (options.recursive && fileInfo.isDirectory()) {
                    if (!config) {
                        config = configArrayFactory.getConfigArrayForFile(filePath, { ignoreNotFoundError: true });
                    }
                    const ignored = this._isIgnoredFile(filePath + path.sep, { ...options, config });
                    if (!ignored) {
                        yield* this._iterateFilesRecursive(filePath, options);
                    }
                }
            }
            debug(`Leave the directory: ${directoryPath}`);
        }
        /**
         * Check if a given file should be ignored.
         * @param {string} filePath The path to a file to check.
         * @param {Object} options Options
         * @param {ConfigArray} [options.config] The config for this file.
         * @param {boolean} [options.dotfiles] If `true` then this is not ignore dot files by default.
         * @param {boolean} [options.direct] If `true` then this is a direct specified file.
         * @returns {boolean} `true` if the file should be ignored.
         * @private
         */
        _isIgnoredFile(filePath, { config: providedConfig, dotfiles = false, direct = false }) {
            const { configArrayFactory, defaultIgnores, ignoreFlag } = internalSlotsMap.get(this);
            if (ignoreFlag) {
                const config = providedConfig ||
                    configArrayFactory.getConfigArrayForFile(filePath, { ignoreNotFoundError: true });
                const ignores = config.extractConfig(filePath).ignores || defaultIgnores;
                return ignores(filePath, dotfiles);
            }
            return !direct && defaultIgnores(filePath, dotfiles);
        }
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = { FileEnumerator };
},
/* glob-parent/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var isGlob = require(327) /* is-glob */;
    var pathPosixDirname = require('path').posix.dirname;
    var isWin32 = require('os').platform() === 'win32';
    var slash = '/';
    var backslash = /\\/g;
    var enclosure = /[\{\[].*[\}\]]$/;
    var globby = /(^|[^\\])([\{\[]|\([^\)]+$)/;
    var escaped = /\\([\!\*\?\|\[\]\(\)\{\}])/g;
    /**
     * @param {string} str
     * @param {Object} opts
     * @param {boolean} [opts.flipBackslashes=true]
     * @returns {string}
     */
    module.exports = function globParent(str, opts) {
        var options = Object.assign({ flipBackslashes: true }, opts);
        // flip windows path separators
        if (options.flipBackslashes && isWin32 && str.indexOf(slash) < 0) {
            str = str.replace(backslash, slash);
        }
        // special case for strings ending in enclosure containing path separator
        if (enclosure.test(str)) {
            str += slash;
        }
        // preserves full path in case of trailing path separator
        str += 'a';
        // remove path parts that are globby
        do {
            str = pathPosixDirname(str);
        } while (isGlob(str) || globby.test(str));
        // remove escape chars and return result
        return str.replace(escaped, '$1');
    };
},
/* is-glob/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /*!
     * is-glob <https://github.com/jonschlinkert/is-glob>
     *
     * Copyright (c) 2014-2017, Jon Schlinkert.
     * Released under the MIT License.
     */
    var isExtglob = require(328) /* is-extglob */;
    var chars = { '{': '}', '(': ')', '[': ']' };
    var strictRegex = /\\(.)|(^!|\*|[\].+)]\?|\[[^\\\]]+\]|\{[^\\}]+\}|\(\?[:!=][^\\)]+\)|\([^|]+\|[^\\)]+\))/;
    var relaxedRegex = /\\(.)|(^!|[*?{}()[\]]|\(\?)/;
    module.exports = function isGlob(str, options) {
        if (typeof str !== 'string' || str === '') {
            return false;
        }
        if (isExtglob(str)) {
            return true;
        }
        var regex = strictRegex;
        var match;
        // optionally relax regex
        if (options && options.strict === false) {
            regex = relaxedRegex;
        }
        while ((match = regex.exec(str))) {
            if (match[2])
                return true;
            var idx = match.index + match[0].length;
            // if an open bracket/brace/paren is escaped,
            // set the index to the next closing character
            var open = match[1];
            var close = open ? chars[open] : null;
            if (open && close) {
                var n = str.indexOf(close, idx);
                if (n !== -1) {
                    idx = n + 1;
                }
            }
            str = str.slice(idx);
        }
        return false;
    };
},
/* is-extglob/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /*!
     * is-extglob <https://github.com/jonschlinkert/is-extglob>
     *
     * Copyright (c) 2014-2016, Jon Schlinkert.
     * Licensed under the MIT License.
     */
    module.exports = function isExtglob(str) {
        if (typeof str !== 'string' || str === '') {
            return false;
        }
        var match;
        while ((match = /(\\).|([@?!+*]\(.*\))/g.exec(str))) {
            if (match[2])
                return true;
            str = str.slice(match.index + match[0].length);
        }
        return false;
    };
},
/* escape-string-regexp/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = string => {
        if (typeof string !== 'string') {
            throw new TypeError('Expected a string');
        }
        // Escape characters with special meaning either inside or outside character sets.
        // Use a simple backslash escape when it’s always valid, and a \unnnn escape when the simpler form would be disallowed by Unicode patterns’ stricter grammar.
        return string
            .replace(/[|\\{}()[\]^$+*?.]/g, '\\$&')
            .replace(/-/g, '\\x2d');
    };
},
/* eslint/lib/linter/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const { Linter } = require(331) /* ./linter */;
    const interpolate = require(389) /* ./interpolate */;
    const SourceCodeFixer = require(699) /* ./source-code-fixer */;
    module.exports = {
        Linter,
        // For testers.
        SourceCodeFixer,
        interpolate
    };
},
/* eslint/lib/linter/linter.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const path = require("path"), eslintScope = require(332) /* eslint-scope */, evk = require(321) /* eslint-visitor-keys */, espree = require(313) /* espree */, merge = require(345) /* lodash.merge */, BuiltInEnvironments = require(217) /* @eslint/eslintrc/conf/environments */, pkg = require(207) /* ../../package.json */, astUtils = require(346) /* ../shared/ast-utils */, ConfigOps = require(220) /* @eslint/eslintrc/lib/shared/config-ops */, ConfigValidator = require(215) /* @eslint/eslintrc/lib/shared/config-validator */, Traverser = require(347) /* ../shared/traverser */, { SourceCode } = require(348) /* ../source-code */, CodePathAnalyzer = require(364) /* ./code-path-analysis/code-path-analyzer */, applyDisableDirectives = require(371) /* ./apply-disable-directives */, ConfigCommentParser = require(372) /* ./config-comment-parser */, NodeEventGenerator = require(385) /* ./node-event-generator */, createReportTranslator = require(387) /* ./report-translator */, Rules = require(390) /* ./rules */, createEmitter = require(698) /* ./safe-emitter */, SourceCodeFixer = require(699) /* ./source-code-fixer */, timing = require(700) /* ./timing */, ruleReplacements = require(701) /* ../../conf/replacements.json */;
    const debug = require(72) /* debug */("eslint:linter");
    const MAX_AUTOFIX_PASSES = 10;
    const DEFAULT_PARSER_NAME = "espree";
    const DEFAULT_ECMA_VERSION = 5;
    const commentParser = new ConfigCommentParser();
    const DEFAULT_ERROR_LOC = { start: { line: 1, column: 0 }, end: { line: 1, column: 1 } };
    const parserSymbol = Symbol.for("eslint.RuleTester.parser");
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    /** @typedef {InstanceType<import("../cli-engine/config-array")["ConfigArray"]>} ConfigArray */
    /** @typedef {InstanceType<import("../cli-engine/config-array")["ExtractedConfig"]>} ExtractedConfig */
    /** @typedef {import("../shared/types").ConfigData} ConfigData */
    /** @typedef {import("../shared/types").Environment} Environment */
    /** @typedef {import("../shared/types").GlobalConf} GlobalConf */
    /** @typedef {import("../shared/types").LintMessage} LintMessage */
    /** @typedef {import("../shared/types").ParserOptions} ParserOptions */
    /** @typedef {import("../shared/types").Processor} Processor */
    /** @typedef {import("../shared/types").Rule} Rule */
    /**
     * @template T
     * @typedef {{ [P in keyof T]-?: T[P] }} Required
     */
    /**
     * @typedef {Object} DisableDirective
     * @property {("disable"|"enable"|"disable-line"|"disable-next-line")} type
     * @property {number} line
     * @property {number} column
     * @property {(string|null)} ruleId
     */
    /**
     * The private data for `Linter` instance.
     * @typedef {Object} LinterInternalSlots
     * @property {ConfigArray|null} lastConfigArray The `ConfigArray` instance that the last `verify()` call used.
     * @property {SourceCode|null} lastSourceCode The `SourceCode` instance that the last `verify()` call used.
     * @property {Map<string, Parser>} parserMap The loaded parsers.
     * @property {Rules} ruleMap The loaded rules.
     */
    /**
     * @typedef {Object} VerifyOptions
     * @property {boolean} [allowInlineConfig] Allow/disallow inline comments' ability
     *      to change config once it is set. Defaults to true if not supplied.
     *      Useful if you want to validate JS without comments overriding rules.
     * @property {boolean} [disableFixes] if `true` then the linter doesn't make `fix`
     *      properties into the lint result.
     * @property {string} [filename] the filename of the source code.
     * @property {boolean | "off" | "warn" | "error"} [reportUnusedDisableDirectives] Adds reported errors for
     *      unused `eslint-disable` directives.
     */
    /**
     * @typedef {Object} ProcessorOptions
     * @property {(filename:string, text:string) => boolean} [filterCodeBlock] the
     *      predicate function that selects adopt code blocks.
     * @property {Processor["postprocess"]} [postprocess] postprocessor for report
     *      messages. If provided, this should accept an array of the message lists
     *      for each code block returned from the preprocessor, apply a mapping to
     *      the messages as appropriate, and return a one-dimensional array of
     *      messages.
     * @property {Processor["preprocess"]} [preprocess] preprocessor for source text.
     *      If provided, this should accept a string of source text, and return an
     *      array of code blocks to lint.
     */
    /**
     * @typedef {Object} FixOptions
     * @property {boolean | ((message: LintMessage) => boolean)} [fix] Determines
     *      whether fixes should be applied.
     */
    /**
     * @typedef {Object} InternalOptions
     * @property {string | null} warnInlineConfig The config name what `noInlineConfig` setting came from. If `noInlineConfig` setting didn't exist, this is null. If this is a config name, then the linter warns directive comments.
     * @property {"off" | "warn" | "error"} reportUnusedDisableDirectives (boolean values were normalized)
     */
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Ensures that variables representing built-in properties of the Global Object,
     * and any globals declared by special block comments, are present in the global
     * scope.
     * @param {Scope} globalScope The global scope.
     * @param {Object} configGlobals The globals declared in configuration
     * @param {{exportedVariables: Object, enabledGlobals: Object}} commentDirectives Directives from comment configuration
     * @returns {void}
     */
    function addDeclaredGlobals(globalScope, configGlobals, { exportedVariables, enabledGlobals }) {
        // Define configured global variables.
        for (const id of new Set([...Object.keys(configGlobals), ...Object.keys(enabledGlobals)])) {
            /*
             * `ConfigOps.normalizeConfigGlobal` will throw an error if a configured global value is invalid. However, these errors would
             * typically be caught when validating a config anyway (validity for inline global comments is checked separately).
             */
            const configValue = configGlobals[id] === void 0 ? void 0 : ConfigOps.normalizeConfigGlobal(configGlobals[id]);
            const commentValue = enabledGlobals[id] && enabledGlobals[id].value;
            const value = commentValue || configValue;
            const sourceComments = enabledGlobals[id] && enabledGlobals[id].comments;
            if (value === "off") {
                continue;
            }
            let variable = globalScope.set.get(id);
            if (!variable) {
                variable = new eslintScope.Variable(id, globalScope);
                globalScope.variables.push(variable);
                globalScope.set.set(id, variable);
            }
            variable.eslintImplicitGlobalSetting = configValue;
            variable.eslintExplicitGlobal = sourceComments !== void 0;
            variable.eslintExplicitGlobalComments = sourceComments;
            variable.writeable = (value === "writable");
        }
        // mark all exported variables as such
        Object.keys(exportedVariables).forEach(name => {
            const variable = globalScope.set.get(name);
            if (variable) {
                variable.eslintUsed = true;
            }
        });
        /*
         * "through" contains all references which definitions cannot be found.
         * Since we augment the global scope using configuration, we need to update
         * references and remove the ones that were added by configuration.
         */
        globalScope.through = globalScope.through.filter(reference => {
            const name = reference.identifier.name;
            const variable = globalScope.set.get(name);
            if (variable) {
                /*
                 * Links the variable and the reference.
                 * And this reference is removed from `Scope#through`.
                 */
                reference.resolved = variable;
                variable.references.push(reference);
                return false;
            }
            return true;
        });
    }
    /**
     * creates a missing-rule message.
     * @param {string} ruleId the ruleId to create
     * @returns {string} created error message
     * @private
     */
    function createMissingRuleMessage(ruleId) {
        return Object.prototype.hasOwnProperty.call(ruleReplacements.rules, ruleId)
            ? `Rule '${ruleId}' was removed and replaced by: ${ruleReplacements.rules[ruleId].join(", ")}`
            : `Definition for rule '${ruleId}' was not found.`;
    }
    /**
     * creates a linting problem
     * @param {Object} options to create linting error
     * @param {string} [options.ruleId] the ruleId to report
     * @param {Object} [options.loc] the loc to report
     * @param {string} [options.message] the error message to report
     * @param {string} [options.severity] the error message to report
     * @returns {LintMessage} created problem, returns a missing-rule problem if only provided ruleId.
     * @private
     */
    function createLintingProblem(options) {
        const { ruleId = null, loc = DEFAULT_ERROR_LOC, message = createMissingRuleMessage(options.ruleId), severity = 2 } = options;
        return {
            ruleId,
            message,
            line: loc.start.line,
            column: loc.start.column + 1,
            endLine: loc.end.line,
            endColumn: loc.end.column + 1,
            severity,
            nodeType: null
        };
    }
    /**
     * Creates a collection of disable directives from a comment
     * @param {Object} options to create disable directives
     * @param {("disable"|"enable"|"disable-line"|"disable-next-line")} options.type The type of directive comment
     * @param {{line: number, column: number}} options.loc The 0-based location of the comment token
     * @param {string} options.value The value after the directive in the comment
     * comment specified no specific rules, so it applies to all rules (e.g. `eslint-disable`)
     * @param {function(string): {create: Function}} options.ruleMapper A map from rule IDs to defined rules
     * @returns {Object} Directives and problems from the comment
     */
    function createDisableDirectives(options) {
        const { type, loc, value, ruleMapper } = options;
        const ruleIds = Object.keys(commentParser.parseListConfig(value));
        const directiveRules = ruleIds.length ? ruleIds : [null];
        const result = {
            directives: [],
            directiveProblems: [] // problems in directives
        };
        for (const ruleId of directiveRules) {
            // push to directives, if the rule is defined(including null, e.g. /*eslint enable*/)
            if (ruleId === null || ruleMapper(ruleId) !== null) {
                result.directives.push({ type, line: loc.start.line, column: loc.start.column + 1, ruleId });
            }
            else {
                result.directiveProblems.push(createLintingProblem({ ruleId, loc }));
            }
        }
        return result;
    }
    /**
     * Remove the ignored part from a given directive comment and trim it.
     * @param {string} value The comment text to strip.
     * @returns {string} The stripped text.
     */
    function stripDirectiveComment(value) {
        return value.split(/\s-{2,}\s/u)[0].trim();
    }
    /**
     * Parses comments in file to extract file-specific config of rules, globals
     * and environments and merges them with global config; also code blocks
     * where reporting is disabled or enabled and merges them with reporting config.
     * @param {string} filename The file being checked.
     * @param {ASTNode} ast The top node of the AST.
     * @param {function(string): {create: Function}} ruleMapper A map from rule IDs to defined rules
     * @param {string|null} warnInlineConfig If a string then it should warn directive comments as disabled. The string value is the config name what the setting came from.
     * @returns {{configuredRules: Object, enabledGlobals: {value:string,comment:Token}[], exportedVariables: Object, problems: Problem[], disableDirectives: DisableDirective[]}}
     * A collection of the directive comments that were found, along with any problems that occurred when parsing
     */
    function getDirectiveComments(filename, ast, ruleMapper, warnInlineConfig) {
        const configuredRules = {};
        const enabledGlobals = Object.create(null);
        const exportedVariables = {};
        const problems = [];
        const disableDirectives = [];
        const validator = new ConfigValidator({
            builtInRules: Rules
        });
        ast.comments.filter(token => token.type !== "Shebang").forEach(comment => {
            const trimmedCommentText = stripDirectiveComment(comment.value);
            const match = /^(eslint(?:-env|-enable|-disable(?:(?:-next)?-line)?)?|exported|globals?)(?:\s|$)/u.exec(trimmedCommentText);
            if (!match) {
                return;
            }
            const directiveText = match[1];
            const lineCommentSupported = /^eslint-disable-(next-)?line$/u.test(directiveText);
            if (comment.type === "Line" && !lineCommentSupported) {
                return;
            }
            if (warnInlineConfig) {
                const kind = comment.type === "Block" ? `/*${directiveText}*/` : `//${directiveText}`;
                problems.push(createLintingProblem({
                    ruleId: null,
                    message: `'${kind}' has no effect because you have 'noInlineConfig' setting in ${warnInlineConfig}.`,
                    loc: comment.loc,
                    severity: 1
                }));
                return;
            }
            if (lineCommentSupported && comment.loc.start.line !== comment.loc.end.line) {
                const message = `${directiveText} comment should not span multiple lines.`;
                problems.push(createLintingProblem({
                    ruleId: null,
                    message,
                    loc: comment.loc
                }));
                return;
            }
            const directiveValue = trimmedCommentText.slice(match.index + directiveText.length);
            switch (directiveText) {
                case "eslint-disable":
                case "eslint-enable":
                case "eslint-disable-next-line":
                case "eslint-disable-line": {
                    const directiveType = directiveText.slice("eslint-".length);
                    const options = { type: directiveType, loc: comment.loc, value: directiveValue, ruleMapper };
                    const { directives, directiveProblems } = createDisableDirectives(options);
                    disableDirectives.push(...directives);
                    problems.push(...directiveProblems);
                    break;
                }
                case "exported":
                    Object.assign(exportedVariables, commentParser.parseStringConfig(directiveValue, comment));
                    break;
                case "globals":
                case "global":
                    for (const [id, { value }] of Object.entries(commentParser.parseStringConfig(directiveValue, comment))) {
                        let normalizedValue;
                        try {
                            normalizedValue = ConfigOps.normalizeConfigGlobal(value);
                        }
                        catch (err) {
                            problems.push(createLintingProblem({
                                ruleId: null,
                                loc: comment.loc,
                                message: err.message
                            }));
                            continue;
                        }
                        if (enabledGlobals[id]) {
                            enabledGlobals[id].comments.push(comment);
                            enabledGlobals[id].value = normalizedValue;
                        }
                        else {
                            enabledGlobals[id] = {
                                comments: [comment],
                                value: normalizedValue
                            };
                        }
                    }
                    break;
                case "eslint": {
                    const parseResult = commentParser.parseJsonConfig(directiveValue, comment.loc);
                    if (parseResult.success) {
                        Object.keys(parseResult.config).forEach(name => {
                            const rule = ruleMapper(name);
                            const ruleValue = parseResult.config[name];
                            if (rule === null) {
                                problems.push(createLintingProblem({ ruleId: name, loc: comment.loc }));
                                return;
                            }
                            try {
                                validator.validateRuleOptions(rule, name, ruleValue);
                            }
                            catch (err) {
                                problems.push(createLintingProblem({
                                    ruleId: name,
                                    message: err.message,
                                    loc: comment.loc
                                }));
                                // do not apply the config, if found invalid options.
                                return;
                            }
                            configuredRules[name] = ruleValue;
                        });
                    }
                    else {
                        problems.push(parseResult.error);
                    }
                    break;
                }
                // no default
            }
        });
        return {
            configuredRules,
            enabledGlobals,
            exportedVariables,
            problems,
            disableDirectives
        };
    }
    /**
     * Normalize ECMAScript version from the initial config
     * @param {Parser} parser The parser which uses this options.
     * @param {number} ecmaVersion ECMAScript version from the initial config
     * @returns {number} normalized ECMAScript version
     */
    function normalizeEcmaVersion(parser, ecmaVersion) {
        if ((parser[parserSymbol] || parser) === espree) {
            if (ecmaVersion === "latest") {
                return espree.latestEcmaVersion;
            }
        }
        /*
         * Calculate ECMAScript edition number from official year version starting with
         * ES2015, which corresponds with ES6 (or a difference of 2009).
         */
        return ecmaVersion >= 2015 ? ecmaVersion - 2009 : ecmaVersion;
    }
    const eslintEnvPattern = /\/\*\s*eslint-env\s(.+?)\*\//gsu;
    /**
     * Checks whether or not there is a comment which has "eslint-env *" in a given text.
     * @param {string} text A source code text to check.
     * @returns {Object|null} A result of parseListConfig() with "eslint-env *" comment.
     */
    function findEslintEnv(text) {
        let match, retv;
        eslintEnvPattern.lastIndex = 0;
        while ((match = eslintEnvPattern.exec(text)) !== null) {
            retv = Object.assign(retv || {}, commentParser.parseListConfig(stripDirectiveComment(match[1])));
        }
        return retv;
    }
    /**
     * Convert "/path/to/<text>" to "<text>".
     * `CLIEngine#executeOnText()` method gives "/path/to/<text>" if the filename
     * was omitted because `configArray.extractConfig()` requires an absolute path.
     * But the linter should pass `<text>` to `RuleContext#getFilename()` in that
     * case.
     * Also, code blocks can have their virtual filename. If the parent filename was
     * `<text>`, the virtual filename is `<text>/0_foo.js` or something like (i.e.,
     * it's not an absolute path).
     * @param {string} filename The filename to normalize.
     * @returns {string} The normalized filename.
     */
    function normalizeFilename(filename) {
        const parts = filename.split(path.sep);
        const index = parts.lastIndexOf("<text>");
        return index === -1 ? filename : parts.slice(index).join(path.sep);
    }
    /**
     * Normalizes the possible options for `linter.verify` and `linter.verifyAndFix` to a
     * consistent shape.
     * @param {VerifyOptions} providedOptions Options
     * @param {ConfigData} config Config.
     * @returns {Required<VerifyOptions> & InternalOptions} Normalized options
     */
    function normalizeVerifyOptions(providedOptions, config) {
        const disableInlineConfig = config.noInlineConfig === true;
        const ignoreInlineConfig = providedOptions.allowInlineConfig === false;
        const configNameOfNoInlineConfig = config.configNameOfNoInlineConfig
            ? ` (${config.configNameOfNoInlineConfig})`
            : "";
        let reportUnusedDisableDirectives = providedOptions.reportUnusedDisableDirectives;
        if (typeof reportUnusedDisableDirectives === "boolean") {
            reportUnusedDisableDirectives = reportUnusedDisableDirectives ? "error" : "off";
        }
        if (typeof reportUnusedDisableDirectives !== "string") {
            reportUnusedDisableDirectives = config.reportUnusedDisableDirectives ? "warn" : "off";
        }
        return {
            filename: normalizeFilename(providedOptions.filename || "<input>"),
            allowInlineConfig: !ignoreInlineConfig,
            warnInlineConfig: disableInlineConfig && !ignoreInlineConfig
                ? `your config${configNameOfNoInlineConfig}`
                : null,
            reportUnusedDisableDirectives,
            disableFixes: Boolean(providedOptions.disableFixes)
        };
    }
    /**
     * Combines the provided parserOptions with the options from environments
     * @param {Parser} parser The parser which uses this options.
     * @param {ParserOptions} providedOptions The provided 'parserOptions' key in a config
     * @param {Environment[]} enabledEnvironments The environments enabled in configuration and with inline comments
     * @returns {ParserOptions} Resulting parser options after merge
     */
    function resolveParserOptions(parser, providedOptions, enabledEnvironments) {
        const parserOptionsFromEnv = enabledEnvironments
            .filter(env => env.parserOptions)
            .reduce((parserOptions, env) => merge(parserOptions, env.parserOptions), {});
        const mergedParserOptions = merge(parserOptionsFromEnv, providedOptions || {});
        const isModule = mergedParserOptions.sourceType === "module";
        if (isModule) {
            /*
             * can't have global return inside of modules
             * TODO: espree validate parserOptions.globalReturn when sourceType is setting to module.(@aladdin-add)
             */
            mergedParserOptions.ecmaFeatures = Object.assign({}, mergedParserOptions.ecmaFeatures, { globalReturn: false });
        }
        mergedParserOptions.ecmaVersion = normalizeEcmaVersion(parser, mergedParserOptions.ecmaVersion);
        return mergedParserOptions;
    }
    /**
     * Combines the provided globals object with the globals from environments
     * @param {Record<string, GlobalConf>} providedGlobals The 'globals' key in a config
     * @param {Environment[]} enabledEnvironments The environments enabled in configuration and with inline comments
     * @returns {Record<string, GlobalConf>} The resolved globals object
     */
    function resolveGlobals(providedGlobals, enabledEnvironments) {
        return Object.assign({}, ...enabledEnvironments.filter(env => env.globals).map(env => env.globals), providedGlobals);
    }
    /**
     * Strips Unicode BOM from a given text.
     * @param {string} text A text to strip.
     * @returns {string} The stripped text.
     */
    function stripUnicodeBOM(text) {
        /*
         * Check Unicode BOM.
         * In JavaScript, string data is stored as UTF-16, so BOM is 0xFEFF.
         * http://www.ecma-international.org/ecma-262/6.0/#sec-unicode-format-control-characters
         */
        if (text.charCodeAt(0) === 0xFEFF) {
            return text.slice(1);
        }
        return text;
    }
    /**
     * Get the options for a rule (not including severity), if any
     * @param {Array|number} ruleConfig rule configuration
     * @returns {Array} of rule options, empty Array if none
     */
    function getRuleOptions(ruleConfig) {
        if (Array.isArray(ruleConfig)) {
            return ruleConfig.slice(1);
        }
        return [];
    }
    /**
     * Analyze scope of the given AST.
     * @param {ASTNode} ast The `Program` node to analyze.
     * @param {ParserOptions} parserOptions The parser options.
     * @param {Record<string, string[]>} visitorKeys The visitor keys.
     * @returns {ScopeManager} The analysis result.
     */
    function analyzeScope(ast, parserOptions, visitorKeys) {
        const ecmaFeatures = parserOptions.ecmaFeatures || {};
        const ecmaVersion = parserOptions.ecmaVersion || DEFAULT_ECMA_VERSION;
        return eslintScope.analyze(ast, {
            ignoreEval: true,
            nodejsScope: ecmaFeatures.globalReturn,
            impliedStrict: ecmaFeatures.impliedStrict,
            ecmaVersion,
            sourceType: parserOptions.sourceType || "script",
            childVisitorKeys: visitorKeys || evk.KEYS,
            fallback: Traverser.getKeys
        });
    }
    /**
     * Parses text into an AST. Moved out here because the try-catch prevents
     * optimization of functions, so it's best to keep the try-catch as isolated
     * as possible
     * @param {string} text The text to parse.
     * @param {Parser} parser The parser to parse.
     * @param {ParserOptions} providedParserOptions Options to pass to the parser
     * @param {string} filePath The path to the file being parsed.
     * @returns {{success: false, error: Problem}|{success: true, sourceCode: SourceCode}}
     * An object containing the AST and parser services if parsing was successful, or the error if parsing failed
     * @private
     */
    function parse(text, parser, providedParserOptions, filePath) {
        const textToParse = stripUnicodeBOM(text).replace(astUtils.shebangPattern, (match, captured) => `//${captured}`);
        const parserOptions = Object.assign({}, providedParserOptions, {
            loc: true,
            range: true,
            raw: true,
            tokens: true,
            comment: true,
            eslintVisitorKeys: true,
            eslintScopeManager: true,
            filePath
        });
        /*
         * Check for parsing errors first. If there's a parsing error, nothing
         * else can happen. However, a parsing error does not throw an error
         * from this method - it's just considered a fatal error message, a
         * problem that ESLint identified just like any other.
         */
        try {
            const parseResult = (typeof parser.parseForESLint === "function")
                ? parser.parseForESLint(textToParse, parserOptions)
                : { ast: parser.parse(textToParse, parserOptions) };
            const ast = parseResult.ast;
            const parserServices = parseResult.services || {};
            const visitorKeys = parseResult.visitorKeys || evk.KEYS;
            const scopeManager = parseResult.scopeManager || analyzeScope(ast, parserOptions, visitorKeys);
            return {
                success: true,
                /*
                 * Save all values that `parseForESLint()` returned.
                 * If a `SourceCode` object is given as the first parameter instead of source code text,
                 * linter skips the parsing process and reuses the source code object.
                 * In that case, linter needs all the values that `parseForESLint()` returned.
                 */
                sourceCode: new SourceCode({
                    text,
                    ast,
                    parserServices,
                    scopeManager,
                    visitorKeys
                })
            };
        }
        catch (ex) {
            // If the message includes a leading line number, strip it:
            const message = `Parsing error: ${ex.message.replace(/^line \d+:/iu, "").trim()}`;
            debug("%s\n%s", message, ex.stack);
            return {
                success: false,
                error: {
                    ruleId: null,
                    fatal: true,
                    severity: 2,
                    message,
                    line: ex.lineNumber,
                    column: ex.column
                }
            };
        }
    }
    /**
     * Gets the scope for the current node
     * @param {ScopeManager} scopeManager The scope manager for this AST
     * @param {ASTNode} currentNode The node to get the scope of
     * @returns {eslint-scope.Scope} The scope information for this node
     */
    function getScope(scopeManager, currentNode) {
        // On Program node, get the outermost scope to avoid return Node.js special function scope or ES modules scope.
        const inner = currentNode.type !== "Program";
        for (let node = currentNode; node; node = node.parent) {
            const scope = scopeManager.acquire(node, inner);
            if (scope) {
                if (scope.type === "function-expression-name") {
                    return scope.childScopes[0];
                }
                return scope;
            }
        }
        return scopeManager.scopes[0];
    }
    /**
     * Marks a variable as used in the current scope
     * @param {ScopeManager} scopeManager The scope manager for this AST. The scope may be mutated by this function.
     * @param {ASTNode} currentNode The node currently being traversed
     * @param {Object} parserOptions The options used to parse this text
     * @param {string} name The name of the variable that should be marked as used.
     * @returns {boolean} True if the variable was found and marked as used, false if not.
     */
    function markVariableAsUsed(scopeManager, currentNode, parserOptions, name) {
        const hasGlobalReturn = parserOptions.ecmaFeatures && parserOptions.ecmaFeatures.globalReturn;
        const specialScope = hasGlobalReturn || parserOptions.sourceType === "module";
        const currentScope = getScope(scopeManager, currentNode);
        // Special Node.js scope means we need to start one level deeper
        const initialScope = currentScope.type === "global" && specialScope ? currentScope.childScopes[0] : currentScope;
        for (let scope = initialScope; scope; scope = scope.upper) {
            const variable = scope.variables.find(scopeVar => scopeVar.name === name);
            if (variable) {
                variable.eslintUsed = true;
                return true;
            }
        }
        return false;
    }
    /**
     * Runs a rule, and gets its listeners
     * @param {Rule} rule A normalized rule with a `create` method
     * @param {Context} ruleContext The context that should be passed to the rule
     * @returns {Object} A map of selector listeners provided by the rule
     */
    function createRuleListeners(rule, ruleContext) {
        try {
            return rule.create(ruleContext);
        }
        catch (ex) {
            ex.message = `Error while loading rule '${ruleContext.id}': ${ex.message}`;
            throw ex;
        }
    }
    /**
     * Gets all the ancestors of a given node
     * @param {ASTNode} node The node
     * @returns {ASTNode[]} All the ancestor nodes in the AST, not including the provided node, starting
     * from the root node and going inwards to the parent node.
     */
    function getAncestors(node) {
        const ancestorsStartingAtParent = [];
        for (let ancestor = node.parent; ancestor; ancestor = ancestor.parent) {
            ancestorsStartingAtParent.push(ancestor);
        }
        return ancestorsStartingAtParent.reverse();
    }
    // methods that exist on SourceCode object
    const DEPRECATED_SOURCECODE_PASSTHROUGHS = {
        getSource: "getText",
        getSourceLines: "getLines",
        getAllComments: "getAllComments",
        getNodeByRangeIndex: "getNodeByRangeIndex",
        getComments: "getComments",
        getCommentsBefore: "getCommentsBefore",
        getCommentsAfter: "getCommentsAfter",
        getCommentsInside: "getCommentsInside",
        getJSDocComment: "getJSDocComment",
        getFirstToken: "getFirstToken",
        getFirstTokens: "getFirstTokens",
        getLastToken: "getLastToken",
        getLastTokens: "getLastTokens",
        getTokenAfter: "getTokenAfter",
        getTokenBefore: "getTokenBefore",
        getTokenByRangeStart: "getTokenByRangeStart",
        getTokens: "getTokens",
        getTokensAfter: "getTokensAfter",
        getTokensBefore: "getTokensBefore",
        getTokensBetween: "getTokensBetween"
    };
    const BASE_TRAVERSAL_CONTEXT = Object.freeze(Object.keys(DEPRECATED_SOURCECODE_PASSTHROUGHS).reduce((contextInfo, methodName) => Object.assign(contextInfo, {
        [methodName](...args) {
            return this.getSourceCode()[DEPRECATED_SOURCECODE_PASSTHROUGHS[methodName]](...args);
        }
    }), {}));
    /**
     * Runs the given rules on the given SourceCode object
     * @param {SourceCode} sourceCode A SourceCode object for the given text
     * @param {Object} configuredRules The rules configuration
     * @param {function(string): Rule} ruleMapper A mapper function from rule names to rules
     * @param {Object} parserOptions The options that were passed to the parser
     * @param {string} parserName The name of the parser in the config
     * @param {Object} settings The settings that were enabled in the config
     * @param {string} filename The reported filename of the code
     * @param {boolean} disableFixes If true, it doesn't make `fix` properties.
     * @param {string | undefined} cwd cwd of the cli
     * @param {string} physicalFilename The full path of the file on disk without any code block information
     * @returns {Problem[]} An array of reported problems
     */
    function runRules(sourceCode, configuredRules, ruleMapper, parserOptions, parserName, settings, filename, disableFixes, cwd, physicalFilename) {
        const emitter = createEmitter();
        const nodeQueue = [];
        let currentNode = sourceCode.ast;
        Traverser.traverse(sourceCode.ast, {
            enter(node, parent) {
                node.parent = parent;
                nodeQueue.push({ isEntering: true, node });
            },
            leave(node) {
                nodeQueue.push({ isEntering: false, node });
            },
            visitorKeys: sourceCode.visitorKeys
        });
        /*
         * Create a frozen object with the ruleContext properties and methods that are shared by all rules.
         * All rule contexts will inherit from this object. This avoids the performance penalty of copying all the
         * properties once for each rule.
         */
        const sharedTraversalContext = Object.freeze(Object.assign(Object.create(BASE_TRAVERSAL_CONTEXT), {
            getAncestors: () => getAncestors(currentNode),
            getDeclaredVariables: sourceCode.scopeManager.getDeclaredVariables.bind(sourceCode.scopeManager),
            getCwd: () => cwd,
            getFilename: () => filename,
            getPhysicalFilename: () => physicalFilename || filename,
            getScope: () => getScope(sourceCode.scopeManager, currentNode),
            getSourceCode: () => sourceCode,
            markVariableAsUsed: name => markVariableAsUsed(sourceCode.scopeManager, currentNode, parserOptions, name),
            parserOptions,
            parserPath: parserName,
            parserServices: sourceCode.parserServices,
            settings
        }));
        const lintingProblems = [];
        Object.keys(configuredRules).forEach(ruleId => {
            const severity = ConfigOps.getRuleSeverity(configuredRules[ruleId]);
            // not load disabled rules
            if (severity === 0) {
                return;
            }
            const rule = ruleMapper(ruleId);
            if (rule === null) {
                lintingProblems.push(createLintingProblem({ ruleId }));
                return;
            }
            const messageIds = rule.meta && rule.meta.messages;
            let reportTranslator = null;
            const ruleContext = Object.freeze(Object.assign(Object.create(sharedTraversalContext), {
                id: ruleId,
                options: getRuleOptions(configuredRules[ruleId]),
                report(...args) {
                    /*
                     * Create a report translator lazily.
                     * In a vast majority of cases, any given rule reports zero errors on a given
                     * piece of code. Creating a translator lazily avoids the performance cost of
                     * creating a new translator function for each rule that usually doesn't get
                     * called.
                     *
                     * Using lazy report translators improves end-to-end performance by about 3%
                     * with Node 8.4.0.
                     */
                    if (reportTranslator === null) {
                        reportTranslator = createReportTranslator({
                            ruleId,
                            severity,
                            sourceCode,
                            messageIds,
                            disableFixes
                        });
                    }
                    const problem = reportTranslator(...args);
                    if (problem.fix && rule.meta && !rule.meta.fixable) {
                        throw new Error("Fixable rules should export a `meta.fixable` property.");
                    }
                    lintingProblems.push(problem);
                }
            }));
            const ruleListeners = createRuleListeners(rule, ruleContext);
            // add all the selectors from the rule as listeners
            Object.keys(ruleListeners).forEach(selector => {
                emitter.on(selector, timing.enabled
                    ? timing.time(ruleId, ruleListeners[selector])
                    : ruleListeners[selector]);
            });
        });
        // only run code path analyzer if the top level node is "Program", skip otherwise
        const eventGenerator = nodeQueue[0].node.type === "Program"
            ? new CodePathAnalyzer(new NodeEventGenerator(emitter, { visitorKeys: sourceCode.visitorKeys, fallback: Traverser.getKeys }))
            : new NodeEventGenerator(emitter, { visitorKeys: sourceCode.visitorKeys, fallback: Traverser.getKeys });
        nodeQueue.forEach(traversalInfo => {
            currentNode = traversalInfo.node;
            try {
                if (traversalInfo.isEntering) {
                    eventGenerator.enterNode(currentNode);
                }
                else {
                    eventGenerator.leaveNode(currentNode);
                }
            }
            catch (err) {
                err.currentNode = currentNode;
                throw err;
            }
        });
        return lintingProblems;
    }
    /**
     * Ensure the source code to be a string.
     * @param {string|SourceCode} textOrSourceCode The text or source code object.
     * @returns {string} The source code text.
     */
    function ensureText(textOrSourceCode) {
        if (typeof textOrSourceCode === "object") {
            const { hasBOM, text } = textOrSourceCode;
            const bom = hasBOM ? "\uFEFF" : "";
            return bom + text;
        }
        return String(textOrSourceCode);
    }
    /**
     * Get an environment.
     * @param {LinterInternalSlots} slots The internal slots of Linter.
     * @param {string} envId The environment ID to get.
     * @returns {Environment|null} The environment.
     */
    function getEnv(slots, envId) {
        return ((slots.lastConfigArray && slots.lastConfigArray.pluginEnvironments.get(envId)) ||
            BuiltInEnvironments.get(envId) ||
            null);
    }
    /**
     * Get a rule.
     * @param {LinterInternalSlots} slots The internal slots of Linter.
     * @param {string} ruleId The rule ID to get.
     * @returns {Rule|null} The rule.
     */
    function getRule(slots, ruleId) {
        return ((slots.lastConfigArray && slots.lastConfigArray.pluginRules.get(ruleId)) ||
            slots.ruleMap.get(ruleId));
    }
    /**
     * Normalize the value of the cwd
     * @param {string | undefined} cwd raw value of the cwd, path to a directory that should be considered as the current working directory, can be undefined.
     * @returns {string | undefined} normalized cwd
     */
    function normalizeCwd(cwd) {
        if (cwd) {
            return cwd;
        }
        if (typeof process === "object") {
            return process.cwd();
        }
        // It's more explicit to assign the undefined
        // eslint-disable-next-line no-undefined
        return undefined;
    }
    /**
     * The map to store private data.
     * @type {WeakMap<Linter, LinterInternalSlots>}
     */
    const internalSlotsMap = new WeakMap();
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * Object that is responsible for verifying JavaScript text
     * @name eslint
     */
    class Linter {
        /**
         * Initialize the Linter.
         * @param {Object} [config] the config object
         * @param {string} [config.cwd]  path to a directory that should be considered as the current working directory, can be undefined.
         */
        constructor({ cwd } = {}) {
            internalSlotsMap.set(this, {
                cwd: normalizeCwd(cwd),
                lastConfigArray: null,
                lastSourceCode: null,
                parserMap: new Map([["espree", espree]]),
                ruleMap: new Rules()
            });
            this.version = pkg.version;
        }
        /**
         * Getter for package version.
         * @static
         * @returns {string} The version from package.json.
         */
        static get version() {
            return pkg.version;
        }
        /**
         * Same as linter.verify, except without support for processors.
         * @param {string|SourceCode} textOrSourceCode The text to parse or a SourceCode object.
         * @param {ConfigData} providedConfig An ESLintConfig instance to configure everything.
         * @param {VerifyOptions} [providedOptions] The optional filename of the file being checked.
         * @returns {LintMessage[]} The results as an array of messages or an empty array if no messages.
         */
        _verifyWithoutProcessors(textOrSourceCode, providedConfig, providedOptions) {
            const slots = internalSlotsMap.get(this);
            const config = providedConfig || {};
            const options = normalizeVerifyOptions(providedOptions, config);
            let text;
            // evaluate arguments
            if (typeof textOrSourceCode === "string") {
                slots.lastSourceCode = null;
                text = textOrSourceCode;
            }
            else {
                slots.lastSourceCode = textOrSourceCode;
                text = textOrSourceCode.text;
            }
            // Resolve parser.
            let parserName = DEFAULT_PARSER_NAME;
            let parser = espree;
            if (typeof config.parser === "object" && config.parser !== null) {
                parserName = config.parser.filePath;
                parser = config.parser.definition;
            }
            else if (typeof config.parser === "string") {
                if (!slots.parserMap.has(config.parser)) {
                    return [{
                            ruleId: null,
                            fatal: true,
                            severity: 2,
                            message: `Configured parser '${config.parser}' was not found.`,
                            line: 0,
                            column: 0
                        }];
                }
                parserName = config.parser;
                parser = slots.parserMap.get(config.parser);
            }
            // search and apply "eslint-env *".
            const envInFile = options.allowInlineConfig && !options.warnInlineConfig
                ? findEslintEnv(text)
                : {};
            const resolvedEnvConfig = Object.assign({ builtin: true }, config.env, envInFile);
            const enabledEnvs = Object.keys(resolvedEnvConfig)
                .filter(envName => resolvedEnvConfig[envName])
                .map(envName => getEnv(slots, envName))
                .filter(env => env);
            const parserOptions = resolveParserOptions(parser, config.parserOptions || {}, enabledEnvs);
            const configuredGlobals = resolveGlobals(config.globals || {}, enabledEnvs);
            const settings = config.settings || {};
            if (!slots.lastSourceCode) {
                const parseResult = parse(text, parser, parserOptions, options.filename);
                if (!parseResult.success) {
                    return [parseResult.error];
                }
                slots.lastSourceCode = parseResult.sourceCode;
            }
            else {
                /*
                 * If the given source code object as the first argument does not have scopeManager, analyze the scope.
                 * This is for backward compatibility (SourceCode is frozen so it cannot rebind).
                 */
                if (!slots.lastSourceCode.scopeManager) {
                    slots.lastSourceCode = new SourceCode({
                        text: slots.lastSourceCode.text,
                        ast: slots.lastSourceCode.ast,
                        parserServices: slots.lastSourceCode.parserServices,
                        visitorKeys: slots.lastSourceCode.visitorKeys,
                        scopeManager: analyzeScope(slots.lastSourceCode.ast, parserOptions)
                    });
                }
            }
            const sourceCode = slots.lastSourceCode;
            const commentDirectives = options.allowInlineConfig
                ? getDirectiveComments(options.filename, sourceCode.ast, ruleId => getRule(slots, ruleId), options.warnInlineConfig)
                : { configuredRules: {}, enabledGlobals: {}, exportedVariables: {}, problems: [], disableDirectives: [] };
            // augment global scope with declared global variables
            addDeclaredGlobals(sourceCode.scopeManager.scopes[0], configuredGlobals, { exportedVariables: commentDirectives.exportedVariables, enabledGlobals: commentDirectives.enabledGlobals });
            const configuredRules = Object.assign({}, config.rules, commentDirectives.configuredRules);
            let lintingProblems;
            try {
                lintingProblems = runRules(sourceCode, configuredRules, ruleId => getRule(slots, ruleId), parserOptions, parserName, settings, options.filename, options.disableFixes, slots.cwd, providedOptions.physicalFilename);
            }
            catch (err) {
                err.message += `\nOccurred while linting ${options.filename}`;
                debug("An error occurred while traversing");
                debug("Filename:", options.filename);
                if (err.currentNode) {
                    const { line } = err.currentNode.loc.start;
                    debug("Line:", line);
                    err.message += `:${line}`;
                }
                debug("Parser Options:", parserOptions);
                debug("Parser Path:", parserName);
                debug("Settings:", settings);
                throw err;
            }
            return applyDisableDirectives({
                directives: commentDirectives.disableDirectives,
                problems: lintingProblems
                    .concat(commentDirectives.problems)
                    .sort((problemA, problemB) => problemA.line - problemB.line || problemA.column - problemB.column),
                reportUnusedDisableDirectives: options.reportUnusedDisableDirectives
            });
        }
        /**
         * Verifies the text against the rules specified by the second argument.
         * @param {string|SourceCode} textOrSourceCode The text to parse or a SourceCode object.
         * @param {ConfigData|ConfigArray} config An ESLintConfig instance to configure everything.
         * @param {(string|(VerifyOptions&ProcessorOptions))} [filenameOrOptions] The optional filename of the file being checked.
         *      If this is not set, the filename will default to '<input>' in the rule context. If
         *      an object, then it has "filename", "allowInlineConfig", and some properties.
         * @returns {LintMessage[]} The results as an array of messages or an empty array if no messages.
         */
        verify(textOrSourceCode, config, filenameOrOptions) {
            debug("Verify");
            const options = typeof filenameOrOptions === "string"
                ? { filename: filenameOrOptions }
                : filenameOrOptions || {};
            // CLIEngine passes a `ConfigArray` object.
            if (config && typeof config.extractConfig === "function") {
                return this._verifyWithConfigArray(textOrSourceCode, config, options);
            }
            /*
             * `Linter` doesn't support `overrides` property in configuration.
             * So we cannot apply multiple processors.
             */
            if (options.preprocess || options.postprocess) {
                return this._verifyWithProcessor(textOrSourceCode, config, options);
            }
            return this._verifyWithoutProcessors(textOrSourceCode, config, options);
        }
        /**
         * Verify a given code with `ConfigArray`.
         * @param {string|SourceCode} textOrSourceCode The source code.
         * @param {ConfigArray} configArray The config array.
         * @param {VerifyOptions&ProcessorOptions} options The options.
         * @returns {LintMessage[]} The found problems.
         */
        _verifyWithConfigArray(textOrSourceCode, configArray, options) {
            debug("With ConfigArray: %s", options.filename);
            // Store the config array in order to get plugin envs and rules later.
            internalSlotsMap.get(this).lastConfigArray = configArray;
            // Extract the final config for this file.
            const config = configArray.extractConfig(options.filename);
            const processor = config.processor &&
                configArray.pluginProcessors.get(config.processor);
            // Verify.
            if (processor) {
                debug("Apply the processor: %o", config.processor);
                const { preprocess, postprocess, supportsAutofix } = processor;
                const disableFixes = options.disableFixes || !supportsAutofix;
                return this._verifyWithProcessor(textOrSourceCode, config, { ...options, disableFixes, postprocess, preprocess }, configArray);
            }
            return this._verifyWithoutProcessors(textOrSourceCode, config, options);
        }
        /**
         * Verify with a processor.
         * @param {string|SourceCode} textOrSourceCode The source code.
         * @param {ConfigData|ExtractedConfig} config The config array.
         * @param {VerifyOptions&ProcessorOptions} options The options.
         * @param {ConfigArray} [configForRecursive] The `ConfigArray` object to apply multiple processors recursively.
         * @returns {LintMessage[]} The found problems.
         */
        _verifyWithProcessor(textOrSourceCode, config, options, configForRecursive) {
            const filename = options.filename || "<input>";
            const filenameToExpose = normalizeFilename(filename);
            const physicalFilename = options.physicalFilename || filenameToExpose;
            const text = ensureText(textOrSourceCode);
            const preprocess = options.preprocess || (rawText => [rawText]);
            // TODO(stephenwade): Replace this with array.flat() when we drop support for Node v10
            const postprocess = options.postprocess || (array => [].concat(...array));
            const filterCodeBlock = options.filterCodeBlock ||
                (blockFilename => blockFilename.endsWith(".js"));
            const originalExtname = path.extname(filename);
            const messageLists = preprocess(text, filenameToExpose).map((block, i) => {
                debug("A code block was found: %o", block.filename || "(unnamed)");
                // Keep the legacy behavior.
                if (typeof block === "string") {
                    return this._verifyWithoutProcessors(block, config, options);
                }
                const blockText = block.text;
                const blockName = path.join(filename, `${i}_${block.filename}`);
                // Skip this block if filtered.
                if (!filterCodeBlock(blockName, blockText)) {
                    debug("This code block was skipped.");
                    return [];
                }
                // Resolve configuration again if the file content or extension was changed.
                if (configForRecursive && (text !== blockText || path.extname(blockName) !== originalExtname)) {
                    debug("Resolving configuration again because the file content or extension was changed.");
                    return this._verifyWithConfigArray(blockText, configForRecursive, { ...options, filename: blockName, physicalFilename });
                }
                // Does lint.
                return this._verifyWithoutProcessors(blockText, config, { ...options, filename: blockName, physicalFilename });
            });
            return postprocess(messageLists, filenameToExpose);
        }
        /**
         * Gets the SourceCode object representing the parsed source.
         * @returns {SourceCode} The SourceCode object.
         */
        getSourceCode() {
            return internalSlotsMap.get(this).lastSourceCode;
        }
        /**
         * Defines a new linting rule.
         * @param {string} ruleId A unique rule identifier
         * @param {Function | Rule} ruleModule Function from context to object mapping AST node types to event handlers
         * @returns {void}
         */
        defineRule(ruleId, ruleModule) {
            internalSlotsMap.get(this).ruleMap.define(ruleId, ruleModule);
        }
        /**
         * Defines many new linting rules.
         * @param {Record<string, Function | Rule>} rulesToDefine map from unique rule identifier to rule
         * @returns {void}
         */
        defineRules(rulesToDefine) {
            Object.getOwnPropertyNames(rulesToDefine).forEach(ruleId => {
                this.defineRule(ruleId, rulesToDefine[ruleId]);
            });
        }
        /**
         * Gets an object with all loaded rules.
         * @returns {Map<string, Rule>} All loaded rules
         */
        getRules() {
            const { lastConfigArray, ruleMap } = internalSlotsMap.get(this);
            return new Map(function* () {
                yield* ruleMap;
                if (lastConfigArray) {
                    yield* lastConfigArray.pluginRules;
                }
            }());
        }
        /**
         * Define a new parser module
         * @param {string} parserId Name of the parser
         * @param {Parser} parserModule The parser object
         * @returns {void}
         */
        defineParser(parserId, parserModule) {
            internalSlotsMap.get(this).parserMap.set(parserId, parserModule);
        }
        /**
         * Performs multiple autofix passes over the text until as many fixes as possible
         * have been applied.
         * @param {string} text The source text to apply fixes to.
         * @param {ConfigData|ConfigArray} config The ESLint config object to use.
         * @param {VerifyOptions&ProcessorOptions&FixOptions} options The ESLint options object to use.
         * @returns {{fixed:boolean,messages:LintMessage[],output:string}} The result of the fix operation as returned from the
         *      SourceCodeFixer.
         */
        verifyAndFix(text, config, options) {
            let messages = [], fixedResult, fixed = false, passNumber = 0, currentText = text;
            const debugTextDescription = options && options.filename || `${text.slice(0, 10)}...`;
            const shouldFix = options && typeof options.fix !== "undefined" ? options.fix : true;
            /**
             * This loop continues until one of the following is true:
             *
             * 1. No more fixes have been applied.
             * 2. Ten passes have been made.
             *
             * That means anytime a fix is successfully applied, there will be another pass.
             * Essentially, guaranteeing a minimum of two passes.
             */
            do {
                passNumber++;
                debug(`Linting code for ${debugTextDescription} (pass ${passNumber})`);
                messages = this.verify(currentText, config, options);
                debug(`Generating fixed text for ${debugTextDescription} (pass ${passNumber})`);
                fixedResult = SourceCodeFixer.applyFixes(currentText, messages, shouldFix);
                /*
                 * stop if there are any syntax errors.
                 * 'fixedResult.output' is a empty string.
                 */
                if (messages.length === 1 && messages[0].fatal) {
                    break;
                }
                // keep track if any fixes were ever applied - important for return value
                fixed = fixed || fixedResult.fixed;
                // update to use the fixed output instead of the original text
                currentText = fixedResult.output;
            } while (fixedResult.fixed &&
                passNumber < MAX_AUTOFIX_PASSES);
            /*
             * If the last result had fixes, we need to lint again to be sure we have
             * the most up-to-date information.
             */
            if (fixedResult.fixed) {
                fixedResult.messages = this.verify(currentText, config, options);
            }
            // ensure the last result properly reflects if fixes were done
            fixedResult.fixed = fixed;
            fixedResult.output = currentText;
            return fixedResult;
        }
    }
    module.exports = {
        Linter,
        /**
         * Get the internal slots of a given Linter instance for tests.
         * @param {Linter} instance The Linter instance to get.
         * @returns {LinterInternalSlots} The internal slots.
         */
        getLinterInternalSlots(instance) {
            return internalSlotsMap.get(instance);
        }
    };
},
/* eslint-scope/lib/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint no-underscore-dangle: ["error", { "allow": ["__currentScope"] }] */
    const assert = require("assert");
    const ScopeManager = require(333) /* ./scope-manager */;
    const Referencer = require(340) /* ./referencer */;
    const Reference = require(337) /* ./reference */;
    const Variable = require(338) /* ./variable */;
    const Scope = require(334) /* ./scope */.Scope;
    const version = require(344) /* ../package.json */.version;
    /**
     * Set the default options
     * @returns {Object} options
     */
    function defaultOptions() {
        return {
            optimistic: false,
            directive: false,
            nodejsScope: false,
            impliedStrict: false,
            sourceType: "script",
            ecmaVersion: 5,
            childVisitorKeys: null,
            fallback: "iteration"
        };
    }
    /**
     * Preform deep update on option object
     * @param {Object} target - Options
     * @param {Object} override - Updates
     * @returns {Object} Updated options
     */
    function updateDeeply(target, override) {
        /**
         * Is hash object
         * @param {Object} value - Test value
         * @returns {boolean} Result
         */
        function isHashObject(value) {
            return typeof value === "object" && value instanceof Object && !(value instanceof Array) && !(value instanceof RegExp);
        }
        for (const key in override) {
            if (Object.prototype.hasOwnProperty.call(override, key)) {
                const val = override[key];
                if (isHashObject(val)) {
                    if (isHashObject(target[key])) {
                        updateDeeply(target[key], val);
                    }
                    else {
                        target[key] = updateDeeply({}, val);
                    }
                }
                else {
                    target[key] = val;
                }
            }
        }
        return target;
    }
    /**
     * Main interface function. Takes an Espree syntax tree and returns the
     * analyzed scopes.
     * @function analyze
     * @param {espree.Tree} tree - Abstract Syntax Tree
     * @param {Object} providedOptions - Options that tailor the scope analysis
     * @param {boolean} [providedOptions.optimistic=false] - the optimistic flag
     * @param {boolean} [providedOptions.directive=false]- the directive flag
     * @param {boolean} [providedOptions.ignoreEval=false]- whether to check 'eval()' calls
     * @param {boolean} [providedOptions.nodejsScope=false]- whether the whole
     * script is executed under node.js environment. When enabled, escope adds
     * a function scope immediately following the global scope.
     * @param {boolean} [providedOptions.impliedStrict=false]- implied strict mode
     * (if ecmaVersion >= 5).
     * @param {string} [providedOptions.sourceType='script']- the source type of the script. one of 'script' and 'module'
     * @param {number} [providedOptions.ecmaVersion=5]- which ECMAScript version is considered
     * @param {Object} [providedOptions.childVisitorKeys=null] - Additional known visitor keys. See [esrecurse](https://github.com/estools/esrecurse)'s the `childVisitorKeys` option.
     * @param {string} [providedOptions.fallback='iteration'] - A kind of the fallback in order to encounter with unknown node. See [esrecurse](https://github.com/estools/esrecurse)'s the `fallback` option.
     * @returns {ScopeManager} ScopeManager
     */
    function analyze(tree, providedOptions) {
        const options = updateDeeply(defaultOptions(), providedOptions);
        const scopeManager = new ScopeManager(options);
        const referencer = new Referencer(options, scopeManager);
        referencer.visit(tree);
        assert(scopeManager.__currentScope === null, "currentScope should be null.");
        return scopeManager;
    }
    module.exports = {
        /** @name module:escope.version */
        version,
        /** @name module:escope.Reference */
        Reference,
        /** @name module:escope.Variable */
        Variable,
        /** @name module:escope.Scope */
        Scope,
        /** @name module:escope.ScopeManager */
        ScopeManager,
        analyze
    };
},
/* eslint-scope/lib/scope-manager.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint-disable no-underscore-dangle */
    const Scope = require(334) /* ./scope */;
    const assert = require("assert");
    const GlobalScope = Scope.GlobalScope;
    const CatchScope = Scope.CatchScope;
    const WithScope = Scope.WithScope;
    const ModuleScope = Scope.ModuleScope;
    const ClassScope = Scope.ClassScope;
    const SwitchScope = Scope.SwitchScope;
    const FunctionScope = Scope.FunctionScope;
    const ForScope = Scope.ForScope;
    const FunctionExpressionNameScope = Scope.FunctionExpressionNameScope;
    const BlockScope = Scope.BlockScope;
    /**
     * @class ScopeManager
     */
    class ScopeManager {
        constructor(options) {
            this.scopes = [];
            this.globalScope = null;
            this.__nodeToScope = new WeakMap();
            this.__currentScope = null;
            this.__options = options;
            this.__declaredVariables = new WeakMap();
        }
        __useDirective() {
            return this.__options.directive;
        }
        __isOptimistic() {
            return this.__options.optimistic;
        }
        __ignoreEval() {
            return this.__options.ignoreEval;
        }
        __isNodejsScope() {
            return this.__options.nodejsScope;
        }
        isModule() {
            return this.__options.sourceType === "module";
        }
        isImpliedStrict() {
            return this.__options.impliedStrict;
        }
        isStrictModeSupported() {
            return this.__options.ecmaVersion >= 5;
        }
        // Returns appropriate scope for this node.
        __get(node) {
            return this.__nodeToScope.get(node);
        }
        /**
         * Get variables that are declared by the node.
         *
         * "are declared by the node" means the node is same as `Variable.defs[].node` or `Variable.defs[].parent`.
         * If the node declares nothing, this method returns an empty array.
         * CAUTION: This API is experimental. See https://github.com/estools/escope/pull/69 for more details.
         *
         * @param {Espree.Node} node - a node to get.
         * @returns {Variable[]} variables that declared by the node.
         */
        getDeclaredVariables(node) {
            return this.__declaredVariables.get(node) || [];
        }
        /**
         * acquire scope from node.
         * @method ScopeManager#acquire
         * @param {Espree.Node} node - node for the acquired scope.
         * @param {boolean=} inner - look up the most inner scope, default value is false.
         * @returns {Scope?} Scope from node
         */
        acquire(node, inner) {
            /**
             * predicate
             * @param {Scope} testScope - scope to test
             * @returns {boolean} predicate
             */
            function predicate(testScope) {
                if (testScope.type === "function" && testScope.functionExpressionScope) {
                    return false;
                }
                return true;
            }
            const scopes = this.__get(node);
            if (!scopes || scopes.length === 0) {
                return null;
            }
            // Heuristic selection from all scopes.
            // If you would like to get all scopes, please use ScopeManager#acquireAll.
            if (scopes.length === 1) {
                return scopes[0];
            }
            if (inner) {
                for (let i = scopes.length - 1; i >= 0; --i) {
                    const scope = scopes[i];
                    if (predicate(scope)) {
                        return scope;
                    }
                }
            }
            else {
                for (let i = 0, iz = scopes.length; i < iz; ++i) {
                    const scope = scopes[i];
                    if (predicate(scope)) {
                        return scope;
                    }
                }
            }
            return null;
        }
        /**
         * acquire all scopes from node.
         * @method ScopeManager#acquireAll
         * @param {Espree.Node} node - node for the acquired scope.
         * @returns {Scopes?} Scope array
         */
        acquireAll(node) {
            return this.__get(node);
        }
        /**
         * release the node.
         * @method ScopeManager#release
         * @param {Espree.Node} node - releasing node.
         * @param {boolean=} inner - look up the most inner scope, default value is false.
         * @returns {Scope?} upper scope for the node.
         */
        release(node, inner) {
            const scopes = this.__get(node);
            if (scopes && scopes.length) {
                const scope = scopes[0].upper;
                if (!scope) {
                    return null;
                }
                return this.acquire(scope.block, inner);
            }
            return null;
        }
        attach() { } // eslint-disable-line class-methods-use-this
        detach() { } // eslint-disable-line class-methods-use-this
        __nestScope(scope) {
            if (scope instanceof GlobalScope) {
                assert(this.__currentScope === null);
                this.globalScope = scope;
            }
            this.__currentScope = scope;
            return scope;
        }
        __nestGlobalScope(node) {
            return this.__nestScope(new GlobalScope(this, node));
        }
        __nestBlockScope(node) {
            return this.__nestScope(new BlockScope(this, this.__currentScope, node));
        }
        __nestFunctionScope(node, isMethodDefinition) {
            return this.__nestScope(new FunctionScope(this, this.__currentScope, node, isMethodDefinition));
        }
        __nestForScope(node) {
            return this.__nestScope(new ForScope(this, this.__currentScope, node));
        }
        __nestCatchScope(node) {
            return this.__nestScope(new CatchScope(this, this.__currentScope, node));
        }
        __nestWithScope(node) {
            return this.__nestScope(new WithScope(this, this.__currentScope, node));
        }
        __nestClassScope(node) {
            return this.__nestScope(new ClassScope(this, this.__currentScope, node));
        }
        __nestSwitchScope(node) {
            return this.__nestScope(new SwitchScope(this, this.__currentScope, node));
        }
        __nestModuleScope(node) {
            return this.__nestScope(new ModuleScope(this, this.__currentScope, node));
        }
        __nestFunctionExpressionNameScope(node) {
            return this.__nestScope(new FunctionExpressionNameScope(this, this.__currentScope, node));
        }
        __isES6() {
            return this.__options.ecmaVersion >= 6;
        }
    }
    module.exports = ScopeManager;
},
/* eslint-scope/lib/scope.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint-disable no-underscore-dangle */
    /* eslint-disable no-undefined */
    const Syntax = require(335) /* estraverse */.Syntax;
    const Reference = require(337) /* ./reference */;
    const Variable = require(338) /* ./variable */;
    const Definition = require(339) /* ./definition */.Definition;
    const assert = require("assert");
    /**
     * Test if scope is struct
     * @param {Scope} scope - scope
     * @param {Block} block - block
     * @param {boolean} isMethodDefinition - is method definition
     * @param {boolean} useDirective - use directive
     * @returns {boolean} is strict scope
     */
    function isStrictScope(scope, block, isMethodDefinition, useDirective) {
        let body;
        // When upper scope is exists and strict, inner scope is also strict.
        if (scope.upper && scope.upper.isStrict) {
            return true;
        }
        if (isMethodDefinition) {
            return true;
        }
        if (scope.type === "class" || scope.type === "module") {
            return true;
        }
        if (scope.type === "block" || scope.type === "switch") {
            return false;
        }
        if (scope.type === "function") {
            if (block.type === Syntax.ArrowFunctionExpression && block.body.type !== Syntax.BlockStatement) {
                return false;
            }
            if (block.type === Syntax.Program) {
                body = block;
            }
            else {
                body = block.body;
            }
            if (!body) {
                return false;
            }
        }
        else if (scope.type === "global") {
            body = block;
        }
        else {
            return false;
        }
        // Search 'use strict' directive.
        if (useDirective) {
            for (let i = 0, iz = body.body.length; i < iz; ++i) {
                const stmt = body.body[i];
                if (stmt.type !== Syntax.DirectiveStatement) {
                    break;
                }
                if (stmt.raw === "\"use strict\"" || stmt.raw === "'use strict'") {
                    return true;
                }
            }
        }
        else {
            for (let i = 0, iz = body.body.length; i < iz; ++i) {
                const stmt = body.body[i];
                if (stmt.type !== Syntax.ExpressionStatement) {
                    break;
                }
                const expr = stmt.expression;
                if (expr.type !== Syntax.Literal || typeof expr.value !== "string") {
                    break;
                }
                if (expr.raw !== null && expr.raw !== undefined) {
                    if (expr.raw === "\"use strict\"" || expr.raw === "'use strict'") {
                        return true;
                    }
                }
                else {
                    if (expr.value === "use strict") {
                        return true;
                    }
                }
            }
        }
        return false;
    }
    /**
     * Register scope
     * @param {ScopeManager} scopeManager - scope manager
     * @param {Scope} scope - scope
     * @returns {void}
     */
    function registerScope(scopeManager, scope) {
        scopeManager.scopes.push(scope);
        const scopes = scopeManager.__nodeToScope.get(scope.block);
        if (scopes) {
            scopes.push(scope);
        }
        else {
            scopeManager.__nodeToScope.set(scope.block, [scope]);
        }
    }
    /**
     * Should be statically
     * @param {Object} def - def
     * @returns {boolean} should be statically
     */
    function shouldBeStatically(def) {
        return ((def.type === Variable.ClassName) ||
            (def.type === Variable.Variable && def.parent.kind !== "var"));
    }
    /**
     * @class Scope
     */
    class Scope {
        constructor(scopeManager, type, upperScope, block, isMethodDefinition) {
            /**
             * One of 'module', 'block', 'switch', 'function', 'catch', 'with', 'function', 'class', 'global'.
             * @member {String} Scope#type
             */
            this.type = type;
            /**
             * The scoped {@link Variable}s of this scope, as <code>{ Variable.name
             * : Variable }</code>.
             * @member {Map} Scope#set
             */
            this.set = new Map();
            /**
             * The tainted variables of this scope, as <code>{ Variable.name :
             * boolean }</code>.
             * @member {Map} Scope#taints */
            this.taints = new Map();
            /**
             * Generally, through the lexical scoping of JS you can always know
             * which variable an identifier in the source code refers to. There are
             * a few exceptions to this rule. With 'global' and 'with' scopes you
             * can only decide at runtime which variable a reference refers to.
             * Moreover, if 'eval()' is used in a scope, it might introduce new
             * bindings in this or its parent scopes.
             * All those scopes are considered 'dynamic'.
             * @member {boolean} Scope#dynamic
             */
            this.dynamic = this.type === "global" || this.type === "with";
            /**
             * A reference to the scope-defining syntax node.
             * @member {espree.Node} Scope#block
             */
            this.block = block;
            /**
             * The {@link Reference|references} that are not resolved with this scope.
             * @member {Reference[]} Scope#through
             */
            this.through = [];
            /**
             * The scoped {@link Variable}s of this scope. In the case of a
             * 'function' scope this includes the automatic argument <em>arguments</em> as
             * its first element, as well as all further formal arguments.
             * @member {Variable[]} Scope#variables
             */
            this.variables = [];
            /**
             * Any variable {@link Reference|reference} found in this scope. This
             * includes occurrences of local variables as well as variables from
             * parent scopes (including the global scope). For local variables
             * this also includes defining occurrences (like in a 'var' statement).
             * In a 'function' scope this does not include the occurrences of the
             * formal parameter in the parameter list.
             * @member {Reference[]} Scope#references
             */
            this.references = [];
            /**
             * For 'global' and 'function' scopes, this is a self-reference. For
             * other scope types this is the <em>variableScope</em> value of the
             * parent scope.
             * @member {Scope} Scope#variableScope
             */
            this.variableScope =
                (this.type === "global" || this.type === "function" || this.type === "module") ? this : upperScope.variableScope;
            /**
             * Whether this scope is created by a FunctionExpression.
             * @member {boolean} Scope#functionExpressionScope
             */
            this.functionExpressionScope = false;
            /**
             * Whether this is a scope that contains an 'eval()' invocation.
             * @member {boolean} Scope#directCallToEvalScope
             */
            this.directCallToEvalScope = false;
            /**
             * @member {boolean} Scope#thisFound
             */
            this.thisFound = false;
            this.__left = [];
            /**
             * Reference to the parent {@link Scope|scope}.
             * @member {Scope} Scope#upper
             */
            this.upper = upperScope;
            /**
             * Whether 'use strict' is in effect in this scope.
             * @member {boolean} Scope#isStrict
             */
            this.isStrict = isStrictScope(this, block, isMethodDefinition, scopeManager.__useDirective());
            /**
             * List of nested {@link Scope}s.
             * @member {Scope[]} Scope#childScopes
             */
            this.childScopes = [];
            if (this.upper) {
                this.upper.childScopes.push(this);
            }
            this.__declaredVariables = scopeManager.__declaredVariables;
            registerScope(scopeManager, this);
        }
        __shouldStaticallyClose(scopeManager) {
            return (!this.dynamic || scopeManager.__isOptimistic());
        }
        __shouldStaticallyCloseForGlobal(ref) {
            // On global scope, let/const/class declarations should be resolved statically.
            const name = ref.identifier.name;
            if (!this.set.has(name)) {
                return false;
            }
            const variable = this.set.get(name);
            const defs = variable.defs;
            return defs.length > 0 && defs.every(shouldBeStatically);
        }
        __staticCloseRef(ref) {
            if (!this.__resolve(ref)) {
                this.__delegateToUpperScope(ref);
            }
        }
        __dynamicCloseRef(ref) {
            // notify all names are through to global
            let current = this;
            do {
                current.through.push(ref);
                current = current.upper;
            } while (current);
        }
        __globalCloseRef(ref) {
            // let/const/class declarations should be resolved statically.
            // others should be resolved dynamically.
            if (this.__shouldStaticallyCloseForGlobal(ref)) {
                this.__staticCloseRef(ref);
            }
            else {
                this.__dynamicCloseRef(ref);
            }
        }
        __close(scopeManager) {
            let closeRef;
            if (this.__shouldStaticallyClose(scopeManager)) {
                closeRef = this.__staticCloseRef;
            }
            else if (this.type !== "global") {
                closeRef = this.__dynamicCloseRef;
            }
            else {
                closeRef = this.__globalCloseRef;
            }
            // Try Resolving all references in this scope.
            for (let i = 0, iz = this.__left.length; i < iz; ++i) {
                const ref = this.__left[i];
                closeRef.call(this, ref);
            }
            this.__left = null;
            return this.upper;
        }
        // To override by function scopes.
        // References in default parameters isn't resolved to variables which are in their function body.
        __isValidResolution(ref, variable) {
            return true;
        }
        __resolve(ref) {
            const name = ref.identifier.name;
            if (!this.set.has(name)) {
                return false;
            }
            const variable = this.set.get(name);
            if (!this.__isValidResolution(ref, variable)) {
                return false;
            }
            variable.references.push(ref);
            variable.stack = variable.stack && ref.from.variableScope === this.variableScope;
            if (ref.tainted) {
                variable.tainted = true;
                this.taints.set(variable.name, true);
            }
            ref.resolved = variable;
            return true;
        }
        __delegateToUpperScope(ref) {
            if (this.upper) {
                this.upper.__left.push(ref);
            }
            this.through.push(ref);
        }
        __addDeclaredVariablesOfNode(variable, node) {
            if (node === null || node === undefined) {
                return;
            }
            let variables = this.__declaredVariables.get(node);
            if (variables === null || variables === undefined) {
                variables = [];
                this.__declaredVariables.set(node, variables);
            }
            if (variables.indexOf(variable) === -1) {
                variables.push(variable);
            }
        }
        __defineGeneric(name, set, variables, node, def) {
            let variable;
            variable = set.get(name);
            if (!variable) {
                variable = new Variable(name, this);
                set.set(name, variable);
                variables.push(variable);
            }
            if (def) {
                variable.defs.push(def);
                this.__addDeclaredVariablesOfNode(variable, def.node);
                this.__addDeclaredVariablesOfNode(variable, def.parent);
            }
            if (node) {
                variable.identifiers.push(node);
            }
        }
        __define(node, def) {
            if (node && node.type === Syntax.Identifier) {
                this.__defineGeneric(node.name, this.set, this.variables, node, def);
            }
        }
        __referencing(node, assign, writeExpr, maybeImplicitGlobal, partial, init) {
            // because Array element may be null
            if (!node || node.type !== Syntax.Identifier) {
                return;
            }
            // Specially handle like `this`.
            if (node.name === "super") {
                return;
            }
            const ref = new Reference(node, this, assign || Reference.READ, writeExpr, maybeImplicitGlobal, !!partial, !!init);
            this.references.push(ref);
            this.__left.push(ref);
        }
        __detectEval() {
            let current = this;
            this.directCallToEvalScope = true;
            do {
                current.dynamic = true;
                current = current.upper;
            } while (current);
        }
        __detectThis() {
            this.thisFound = true;
        }
        __isClosed() {
            return this.__left === null;
        }
        /**
         * returns resolved {Reference}
         * @method Scope#resolve
         * @param {Espree.Identifier} ident - identifier to be resolved.
         * @returns {Reference} reference
         */
        resolve(ident) {
            let ref, i, iz;
            assert(this.__isClosed(), "Scope should be closed.");
            assert(ident.type === Syntax.Identifier, "Target should be identifier.");
            for (i = 0, iz = this.references.length; i < iz; ++i) {
                ref = this.references[i];
                if (ref.identifier === ident) {
                    return ref;
                }
            }
            return null;
        }
        /**
         * returns this scope is static
         * @method Scope#isStatic
         * @returns {boolean} static
         */
        isStatic() {
            return !this.dynamic;
        }
        /**
         * returns this scope has materialized arguments
         * @method Scope#isArgumentsMaterialized
         * @returns {boolean} arguemnts materialized
         */
        isArgumentsMaterialized() {
            return true;
        }
        /**
         * returns this scope has materialized `this` reference
         * @method Scope#isThisMaterialized
         * @returns {boolean} this materialized
         */
        isThisMaterialized() {
            return true;
        }
        isUsedName(name) {
            if (this.set.has(name)) {
                return true;
            }
            for (let i = 0, iz = this.through.length; i < iz; ++i) {
                if (this.through[i].identifier.name === name) {
                    return true;
                }
            }
            return false;
        }
    }
    class GlobalScope extends Scope {
        constructor(scopeManager, block) {
            super(scopeManager, "global", null, block, false);
            this.implicit = {
                set: new Map(),
                variables: [],
                /**
                * List of {@link Reference}s that are left to be resolved (i.e. which
                * need to be linked to the variable they refer to).
                * @member {Reference[]} Scope#implicit#left
                */
                left: []
            };
        }
        __close(scopeManager) {
            const implicit = [];
            for (let i = 0, iz = this.__left.length; i < iz; ++i) {
                const ref = this.__left[i];
                if (ref.__maybeImplicitGlobal && !this.set.has(ref.identifier.name)) {
                    implicit.push(ref.__maybeImplicitGlobal);
                }
            }
            // create an implicit global variable from assignment expression
            for (let i = 0, iz = implicit.length; i < iz; ++i) {
                const info = implicit[i];
                this.__defineImplicit(info.pattern, new Definition(Variable.ImplicitGlobalVariable, info.pattern, info.node, null, null, null));
            }
            this.implicit.left = this.__left;
            return super.__close(scopeManager);
        }
        __defineImplicit(node, def) {
            if (node && node.type === Syntax.Identifier) {
                this.__defineGeneric(node.name, this.implicit.set, this.implicit.variables, node, def);
            }
        }
    }
    class ModuleScope extends Scope {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, "module", upperScope, block, false);
        }
    }
    class FunctionExpressionNameScope extends Scope {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, "function-expression-name", upperScope, block, false);
            this.__define(block.id, new Definition(Variable.FunctionName, block.id, block, null, null, null));
            this.functionExpressionScope = true;
        }
    }
    class CatchScope extends Scope {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, "catch", upperScope, block, false);
        }
    }
    class WithScope extends Scope {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, "with", upperScope, block, false);
        }
        __close(scopeManager) {
            if (this.__shouldStaticallyClose(scopeManager)) {
                return super.__close(scopeManager);
            }
            for (let i = 0, iz = this.__left.length; i < iz; ++i) {
                const ref = this.__left[i];
                ref.tainted = true;
                this.__delegateToUpperScope(ref);
            }
            this.__left = null;
            return this.upper;
        }
    }
    class BlockScope extends Scope {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, "block", upperScope, block, false);
        }
    }
    class SwitchScope extends Scope {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, "switch", upperScope, block, false);
        }
    }
    class FunctionScope extends Scope {
        constructor(scopeManager, upperScope, block, isMethodDefinition) {
            super(scopeManager, "function", upperScope, block, isMethodDefinition);
            // section 9.2.13, FunctionDeclarationInstantiation.
            // NOTE Arrow functions never have an arguments objects.
            if (this.block.type !== Syntax.ArrowFunctionExpression) {
                this.__defineArguments();
            }
        }
        isArgumentsMaterialized() {
            // TODO(Constellation)
            // We can more aggressive on this condition like this.
            //
            // function t() {
            //     // arguments of t is always hidden.
            //     function arguments() {
            //     }
            // }
            if (this.block.type === Syntax.ArrowFunctionExpression) {
                return false;
            }
            if (!this.isStatic()) {
                return true;
            }
            const variable = this.set.get("arguments");
            assert(variable, "Always have arguments variable.");
            return variable.tainted || variable.references.length !== 0;
        }
        isThisMaterialized() {
            if (!this.isStatic()) {
                return true;
            }
            return this.thisFound;
        }
        __defineArguments() {
            this.__defineGeneric("arguments", this.set, this.variables, null, null);
            this.taints.set("arguments", true);
        }
        // References in default parameters isn't resolved to variables which are in their function body.
        //     const x = 1
        //     function f(a = x) { // This `x` is resolved to the `x` in the outer scope.
        //         const x = 2
        //         console.log(a)
        //     }
        __isValidResolution(ref, variable) {
            // If `options.nodejsScope` is true, `this.block` becomes a Program node.
            if (this.block.type === "Program") {
                return true;
            }
            const bodyStart = this.block.body.range[0];
            // It's invalid resolution in the following case:
            return !(variable.scope === this &&
                ref.identifier.range[0] < bodyStart && // the reference is in the parameter part.
                variable.defs.every(d => d.name.range[0] >= bodyStart) // the variable is in the body.
            );
        }
    }
    class ForScope extends Scope {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, "for", upperScope, block, false);
        }
    }
    class ClassScope extends Scope {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, "class", upperScope, block, false);
        }
    }
    module.exports = {
        Scope,
        GlobalScope,
        ModuleScope,
        FunctionExpressionNameScope,
        CatchScope,
        WithScope,
        BlockScope,
        SwitchScope,
        FunctionScope,
        ForScope,
        ClassScope
    };
},
/* estraverse/estraverse.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
      Copyright (C) 2012-2013 Yusuke Suzuki <utatane.tea@gmail.com>
      Copyright (C) 2012 Ariya Hidayat <ariya.hidayat@gmail.com>
    
      Redistribution and use in source and binary forms, with or without
      modification, are permitted provided that the following conditions are met:
    
        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
    
      THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
      AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
      IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
      ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
      DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
      (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
      LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
      ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
      (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
      THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    */
    /*jslint vars:false, bitwise:true*/
    /*jshint indent:4*/
    /*global exports:true*/
    (function clone(exports) {
        'use strict';
        var Syntax, VisitorOption, VisitorKeys, BREAK, SKIP, REMOVE;
        function deepCopy(obj) {
            var ret = {}, key, val;
            for (key in obj) {
                if (obj.hasOwnProperty(key)) {
                    val = obj[key];
                    if (typeof val === 'object' && val !== null) {
                        ret[key] = deepCopy(val);
                    }
                    else {
                        ret[key] = val;
                    }
                }
            }
            return ret;
        }
        // based on LLVM libc++ upper_bound / lower_bound
        // MIT License
        function upperBound(array, func) {
            var diff, len, i, current;
            len = array.length;
            i = 0;
            while (len) {
                diff = len >>> 1;
                current = i + diff;
                if (func(array[current])) {
                    len = diff;
                }
                else {
                    i = current + 1;
                    len -= diff + 1;
                }
            }
            return i;
        }
        Syntax = {
            AssignmentExpression: 'AssignmentExpression',
            AssignmentPattern: 'AssignmentPattern',
            ArrayExpression: 'ArrayExpression',
            ArrayPattern: 'ArrayPattern',
            ArrowFunctionExpression: 'ArrowFunctionExpression',
            AwaitExpression: 'AwaitExpression',
            BlockStatement: 'BlockStatement',
            BinaryExpression: 'BinaryExpression',
            BreakStatement: 'BreakStatement',
            CallExpression: 'CallExpression',
            CatchClause: 'CatchClause',
            ClassBody: 'ClassBody',
            ClassDeclaration: 'ClassDeclaration',
            ClassExpression: 'ClassExpression',
            ComprehensionBlock: 'ComprehensionBlock',
            ComprehensionExpression: 'ComprehensionExpression',
            ConditionalExpression: 'ConditionalExpression',
            ContinueStatement: 'ContinueStatement',
            DebuggerStatement: 'DebuggerStatement',
            DirectiveStatement: 'DirectiveStatement',
            DoWhileStatement: 'DoWhileStatement',
            EmptyStatement: 'EmptyStatement',
            ExportAllDeclaration: 'ExportAllDeclaration',
            ExportDefaultDeclaration: 'ExportDefaultDeclaration',
            ExportNamedDeclaration: 'ExportNamedDeclaration',
            ExportSpecifier: 'ExportSpecifier',
            ExpressionStatement: 'ExpressionStatement',
            ForStatement: 'ForStatement',
            ForInStatement: 'ForInStatement',
            ForOfStatement: 'ForOfStatement',
            FunctionDeclaration: 'FunctionDeclaration',
            FunctionExpression: 'FunctionExpression',
            GeneratorExpression: 'GeneratorExpression',
            Identifier: 'Identifier',
            IfStatement: 'IfStatement',
            ImportExpression: 'ImportExpression',
            ImportDeclaration: 'ImportDeclaration',
            ImportDefaultSpecifier: 'ImportDefaultSpecifier',
            ImportNamespaceSpecifier: 'ImportNamespaceSpecifier',
            ImportSpecifier: 'ImportSpecifier',
            Literal: 'Literal',
            LabeledStatement: 'LabeledStatement',
            LogicalExpression: 'LogicalExpression',
            MemberExpression: 'MemberExpression',
            MetaProperty: 'MetaProperty',
            MethodDefinition: 'MethodDefinition',
            ModuleSpecifier: 'ModuleSpecifier',
            NewExpression: 'NewExpression',
            ObjectExpression: 'ObjectExpression',
            ObjectPattern: 'ObjectPattern',
            Program: 'Program',
            Property: 'Property',
            RestElement: 'RestElement',
            ReturnStatement: 'ReturnStatement',
            SequenceExpression: 'SequenceExpression',
            SpreadElement: 'SpreadElement',
            Super: 'Super',
            SwitchStatement: 'SwitchStatement',
            SwitchCase: 'SwitchCase',
            TaggedTemplateExpression: 'TaggedTemplateExpression',
            TemplateElement: 'TemplateElement',
            TemplateLiteral: 'TemplateLiteral',
            ThisExpression: 'ThisExpression',
            ThrowStatement: 'ThrowStatement',
            TryStatement: 'TryStatement',
            UnaryExpression: 'UnaryExpression',
            UpdateExpression: 'UpdateExpression',
            VariableDeclaration: 'VariableDeclaration',
            VariableDeclarator: 'VariableDeclarator',
            WhileStatement: 'WhileStatement',
            WithStatement: 'WithStatement',
            YieldExpression: 'YieldExpression'
        };
        VisitorKeys = {
            AssignmentExpression: ['left', 'right'],
            AssignmentPattern: ['left', 'right'],
            ArrayExpression: ['elements'],
            ArrayPattern: ['elements'],
            ArrowFunctionExpression: ['params', 'body'],
            AwaitExpression: ['argument'],
            BlockStatement: ['body'],
            BinaryExpression: ['left', 'right'],
            BreakStatement: ['label'],
            CallExpression: ['callee', 'arguments'],
            CatchClause: ['param', 'body'],
            ClassBody: ['body'],
            ClassDeclaration: ['id', 'superClass', 'body'],
            ClassExpression: ['id', 'superClass', 'body'],
            ComprehensionBlock: ['left', 'right'],
            ComprehensionExpression: ['blocks', 'filter', 'body'],
            ConditionalExpression: ['test', 'consequent', 'alternate'],
            ContinueStatement: ['label'],
            DebuggerStatement: [],
            DirectiveStatement: [],
            DoWhileStatement: ['body', 'test'],
            EmptyStatement: [],
            ExportAllDeclaration: ['source'],
            ExportDefaultDeclaration: ['declaration'],
            ExportNamedDeclaration: ['declaration', 'specifiers', 'source'],
            ExportSpecifier: ['exported', 'local'],
            ExpressionStatement: ['expression'],
            ForStatement: ['init', 'test', 'update', 'body'],
            ForInStatement: ['left', 'right', 'body'],
            ForOfStatement: ['left', 'right', 'body'],
            FunctionDeclaration: ['id', 'params', 'body'],
            FunctionExpression: ['id', 'params', 'body'],
            GeneratorExpression: ['blocks', 'filter', 'body'],
            Identifier: [],
            IfStatement: ['test', 'consequent', 'alternate'],
            ImportExpression: ['source'],
            ImportDeclaration: ['specifiers', 'source'],
            ImportDefaultSpecifier: ['local'],
            ImportNamespaceSpecifier: ['local'],
            ImportSpecifier: ['imported', 'local'],
            Literal: [],
            LabeledStatement: ['label', 'body'],
            LogicalExpression: ['left', 'right'],
            MemberExpression: ['object', 'property'],
            MetaProperty: ['meta', 'property'],
            MethodDefinition: ['key', 'value'],
            ModuleSpecifier: [],
            NewExpression: ['callee', 'arguments'],
            ObjectExpression: ['properties'],
            ObjectPattern: ['properties'],
            Program: ['body'],
            Property: ['key', 'value'],
            RestElement: ['argument'],
            ReturnStatement: ['argument'],
            SequenceExpression: ['expressions'],
            SpreadElement: ['argument'],
            Super: [],
            SwitchStatement: ['discriminant', 'cases'],
            SwitchCase: ['test', 'consequent'],
            TaggedTemplateExpression: ['tag', 'quasi'],
            TemplateElement: [],
            TemplateLiteral: ['quasis', 'expressions'],
            ThisExpression: [],
            ThrowStatement: ['argument'],
            TryStatement: ['block', 'handler', 'finalizer'],
            UnaryExpression: ['argument'],
            UpdateExpression: ['argument'],
            VariableDeclaration: ['declarations'],
            VariableDeclarator: ['id', 'init'],
            WhileStatement: ['test', 'body'],
            WithStatement: ['object', 'body'],
            YieldExpression: ['argument']
        };
        // unique id
        BREAK = {};
        SKIP = {};
        REMOVE = {};
        VisitorOption = {
            Break: BREAK,
            Skip: SKIP,
            Remove: REMOVE
        };
        function Reference(parent, key) {
            this.parent = parent;
            this.key = key;
        }
        Reference.prototype.replace = function replace(node) {
            this.parent[this.key] = node;
        };
        Reference.prototype.remove = function remove() {
            if (Array.isArray(this.parent)) {
                this.parent.splice(this.key, 1);
                return true;
            }
            else {
                this.replace(null);
                return false;
            }
        };
        function Element(node, path, wrap, ref) {
            this.node = node;
            this.path = path;
            this.wrap = wrap;
            this.ref = ref;
        }
        function Controller() { }
        // API:
        // return property path array from root to current node
        Controller.prototype.path = function path() {
            var i, iz, j, jz, result, element;
            function addToPath(result, path) {
                if (Array.isArray(path)) {
                    for (j = 0, jz = path.length; j < jz; ++j) {
                        result.push(path[j]);
                    }
                }
                else {
                    result.push(path);
                }
            }
            // root node
            if (!this.__current.path) {
                return null;
            }
            // first node is sentinel, second node is root element
            result = [];
            for (i = 2, iz = this.__leavelist.length; i < iz; ++i) {
                element = this.__leavelist[i];
                addToPath(result, element.path);
            }
            addToPath(result, this.__current.path);
            return result;
        };
        // API:
        // return type of current node
        Controller.prototype.type = function () {
            var node = this.current();
            return node.type || this.__current.wrap;
        };
        // API:
        // return array of parent elements
        Controller.prototype.parents = function parents() {
            var i, iz, result;
            // first node is sentinel
            result = [];
            for (i = 1, iz = this.__leavelist.length; i < iz; ++i) {
                result.push(this.__leavelist[i].node);
            }
            return result;
        };
        // API:
        // return current node
        Controller.prototype.current = function current() {
            return this.__current.node;
        };
        Controller.prototype.__execute = function __execute(callback, element) {
            var previous, result;
            result = undefined;
            previous = this.__current;
            this.__current = element;
            this.__state = null;
            if (callback) {
                result = callback.call(this, element.node, this.__leavelist[this.__leavelist.length - 1].node);
            }
            this.__current = previous;
            return result;
        };
        // API:
        // notify control skip / break
        Controller.prototype.notify = function notify(flag) {
            this.__state = flag;
        };
        // API:
        // skip child nodes of current node
        Controller.prototype.skip = function () {
            this.notify(SKIP);
        };
        // API:
        // break traversals
        Controller.prototype['break'] = function () {
            this.notify(BREAK);
        };
        // API:
        // remove node
        Controller.prototype.remove = function () {
            this.notify(REMOVE);
        };
        Controller.prototype.__initialize = function (root, visitor) {
            this.visitor = visitor;
            this.root = root;
            this.__worklist = [];
            this.__leavelist = [];
            this.__current = null;
            this.__state = null;
            this.__fallback = null;
            if (visitor.fallback === 'iteration') {
                this.__fallback = Object.keys;
            }
            else if (typeof visitor.fallback === 'function') {
                this.__fallback = visitor.fallback;
            }
            this.__keys = VisitorKeys;
            if (visitor.keys) {
                this.__keys = Object.assign(Object.create(this.__keys), visitor.keys);
            }
        };
        function isNode(node) {
            if (node == null) {
                return false;
            }
            return typeof node === 'object' && typeof node.type === 'string';
        }
        function isProperty(nodeType, key) {
            return (nodeType === Syntax.ObjectExpression || nodeType === Syntax.ObjectPattern) && 'properties' === key;
        }
        Controller.prototype.traverse = function traverse(root, visitor) {
            var worklist, leavelist, element, node, nodeType, ret, key, current, current2, candidates, candidate, sentinel;
            this.__initialize(root, visitor);
            sentinel = {};
            // reference
            worklist = this.__worklist;
            leavelist = this.__leavelist;
            // initialize
            worklist.push(new Element(root, null, null, null));
            leavelist.push(new Element(null, null, null, null));
            while (worklist.length) {
                element = worklist.pop();
                if (element === sentinel) {
                    element = leavelist.pop();
                    ret = this.__execute(visitor.leave, element);
                    if (this.__state === BREAK || ret === BREAK) {
                        return;
                    }
                    continue;
                }
                if (element.node) {
                    ret = this.__execute(visitor.enter, element);
                    if (this.__state === BREAK || ret === BREAK) {
                        return;
                    }
                    worklist.push(sentinel);
                    leavelist.push(element);
                    if (this.__state === SKIP || ret === SKIP) {
                        continue;
                    }
                    node = element.node;
                    nodeType = node.type || element.wrap;
                    candidates = this.__keys[nodeType];
                    if (!candidates) {
                        if (this.__fallback) {
                            candidates = this.__fallback(node);
                        }
                        else {
                            throw new Error('Unknown node type ' + nodeType + '.');
                        }
                    }
                    current = candidates.length;
                    while ((current -= 1) >= 0) {
                        key = candidates[current];
                        candidate = node[key];
                        if (!candidate) {
                            continue;
                        }
                        if (Array.isArray(candidate)) {
                            current2 = candidate.length;
                            while ((current2 -= 1) >= 0) {
                                if (!candidate[current2]) {
                                    continue;
                                }
                                if (isProperty(nodeType, candidates[current])) {
                                    element = new Element(candidate[current2], [key, current2], 'Property', null);
                                }
                                else if (isNode(candidate[current2])) {
                                    element = new Element(candidate[current2], [key, current2], null, null);
                                }
                                else {
                                    continue;
                                }
                                worklist.push(element);
                            }
                        }
                        else if (isNode(candidate)) {
                            worklist.push(new Element(candidate, key, null, null));
                        }
                    }
                }
            }
        };
        Controller.prototype.replace = function replace(root, visitor) {
            var worklist, leavelist, node, nodeType, target, element, current, current2, candidates, candidate, sentinel, outer, key;
            function removeElem(element) {
                var i, key, nextElem, parent;
                if (element.ref.remove()) {
                    // When the reference is an element of an array.
                    key = element.ref.key;
                    parent = element.ref.parent;
                    // If removed from array, then decrease following items' keys.
                    i = worklist.length;
                    while (i--) {
                        nextElem = worklist[i];
                        if (nextElem.ref && nextElem.ref.parent === parent) {
                            if (nextElem.ref.key < key) {
                                break;
                            }
                            --nextElem.ref.key;
                        }
                    }
                }
            }
            this.__initialize(root, visitor);
            sentinel = {};
            // reference
            worklist = this.__worklist;
            leavelist = this.__leavelist;
            // initialize
            outer = {
                root: root
            };
            element = new Element(root, null, null, new Reference(outer, 'root'));
            worklist.push(element);
            leavelist.push(element);
            while (worklist.length) {
                element = worklist.pop();
                if (element === sentinel) {
                    element = leavelist.pop();
                    target = this.__execute(visitor.leave, element);
                    // node may be replaced with null,
                    // so distinguish between undefined and null in this place
                    if (target !== undefined && target !== BREAK && target !== SKIP && target !== REMOVE) {
                        // replace
                        element.ref.replace(target);
                    }
                    if (this.__state === REMOVE || target === REMOVE) {
                        removeElem(element);
                    }
                    if (this.__state === BREAK || target === BREAK) {
                        return outer.root;
                    }
                    continue;
                }
                target = this.__execute(visitor.enter, element);
                // node may be replaced with null,
                // so distinguish between undefined and null in this place
                if (target !== undefined && target !== BREAK && target !== SKIP && target !== REMOVE) {
                    // replace
                    element.ref.replace(target);
                    element.node = target;
                }
                if (this.__state === REMOVE || target === REMOVE) {
                    removeElem(element);
                    element.node = null;
                }
                if (this.__state === BREAK || target === BREAK) {
                    return outer.root;
                }
                // node may be null
                node = element.node;
                if (!node) {
                    continue;
                }
                worklist.push(sentinel);
                leavelist.push(element);
                if (this.__state === SKIP || target === SKIP) {
                    continue;
                }
                nodeType = node.type || element.wrap;
                candidates = this.__keys[nodeType];
                if (!candidates) {
                    if (this.__fallback) {
                        candidates = this.__fallback(node);
                    }
                    else {
                        throw new Error('Unknown node type ' + nodeType + '.');
                    }
                }
                current = candidates.length;
                while ((current -= 1) >= 0) {
                    key = candidates[current];
                    candidate = node[key];
                    if (!candidate) {
                        continue;
                    }
                    if (Array.isArray(candidate)) {
                        current2 = candidate.length;
                        while ((current2 -= 1) >= 0) {
                            if (!candidate[current2]) {
                                continue;
                            }
                            if (isProperty(nodeType, candidates[current])) {
                                element = new Element(candidate[current2], [key, current2], 'Property', new Reference(candidate, current2));
                            }
                            else if (isNode(candidate[current2])) {
                                element = new Element(candidate[current2], [key, current2], null, new Reference(candidate, current2));
                            }
                            else {
                                continue;
                            }
                            worklist.push(element);
                        }
                    }
                    else if (isNode(candidate)) {
                        worklist.push(new Element(candidate, key, null, new Reference(node, key)));
                    }
                }
            }
            return outer.root;
        };
        function traverse(root, visitor) {
            var controller = new Controller();
            return controller.traverse(root, visitor);
        }
        function replace(root, visitor) {
            var controller = new Controller();
            return controller.replace(root, visitor);
        }
        function extendCommentRange(comment, tokens) {
            var target;
            target = upperBound(tokens, function search(token) {
                return token.range[0] > comment.range[0];
            });
            comment.extendedRange = [comment.range[0], comment.range[1]];
            if (target !== tokens.length) {
                comment.extendedRange[1] = tokens[target].range[0];
            }
            target -= 1;
            if (target >= 0) {
                comment.extendedRange[0] = tokens[target].range[1];
            }
            return comment;
        }
        function attachComments(tree, providedComments, tokens) {
            // At first, we should calculate extended comment ranges.
            var comments = [], comment, len, i, cursor;
            if (!tree.range) {
                throw new Error('attachComments needs range information');
            }
            // tokens array is empty, we attach comments to tree as 'leadingComments'
            if (!tokens.length) {
                if (providedComments.length) {
                    for (i = 0, len = providedComments.length; i < len; i += 1) {
                        comment = deepCopy(providedComments[i]);
                        comment.extendedRange = [0, tree.range[0]];
                        comments.push(comment);
                    }
                    tree.leadingComments = comments;
                }
                return tree;
            }
            for (i = 0, len = providedComments.length; i < len; i += 1) {
                comments.push(extendCommentRange(deepCopy(providedComments[i]), tokens));
            }
            // This is based on John Freeman's implementation.
            cursor = 0;
            traverse(tree, {
                enter: function (node) {
                    var comment;
                    while (cursor < comments.length) {
                        comment = comments[cursor];
                        if (comment.extendedRange[1] > node.range[0]) {
                            break;
                        }
                        if (comment.extendedRange[1] === node.range[0]) {
                            if (!node.leadingComments) {
                                node.leadingComments = [];
                            }
                            node.leadingComments.push(comment);
                            comments.splice(cursor, 1);
                        }
                        else {
                            cursor += 1;
                        }
                    }
                    // already out of owned node
                    if (cursor === comments.length) {
                        return VisitorOption.Break;
                    }
                    if (comments[cursor].extendedRange[0] > node.range[1]) {
                        return VisitorOption.Skip;
                    }
                }
            });
            cursor = 0;
            traverse(tree, {
                leave: function (node) {
                    var comment;
                    while (cursor < comments.length) {
                        comment = comments[cursor];
                        if (node.range[1] < comment.extendedRange[0]) {
                            break;
                        }
                        if (node.range[1] === comment.extendedRange[0]) {
                            if (!node.trailingComments) {
                                node.trailingComments = [];
                            }
                            node.trailingComments.push(comment);
                            comments.splice(cursor, 1);
                        }
                        else {
                            cursor += 1;
                        }
                    }
                    // already out of owned node
                    if (cursor === comments.length) {
                        return VisitorOption.Break;
                    }
                    if (comments[cursor].extendedRange[0] > node.range[1]) {
                        return VisitorOption.Skip;
                    }
                }
            });
            return tree;
        }
        exports.version = require(336) /* ./package.json */.version;
        exports.Syntax = Syntax;
        exports.traverse = traverse;
        exports.replace = replace;
        exports.attachComments = attachComments;
        exports.VisitorKeys = VisitorKeys;
        exports.VisitorOption = VisitorOption;
        exports.Controller = Controller;
        exports.cloneEnvironment = function () { return clone({}); };
        return exports;
    }(exports));
},
/* estraverse/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "estraverse",
        "description": "ECMAScript JS AST traversal functions",
        "homepage": "https://github.com/estools/estraverse",
        "main": "estraverse.js",
        "version": "4.3.0",
        "engines": {
            "node": ">=4.0"
        },
        "maintainers": [
            {
                "name": "Yusuke Suzuki",
                "email": "utatane.tea@gmail.com",
                "web": "http://github.com/Constellation"
            }
        ],
        "repository": {
            "type": "git",
            "url": "http://github.com/estools/estraverse.git"
        },
        "devDependencies": {
            "babel-preset-env": "^1.6.1",
            "babel-register": "^6.3.13",
            "chai": "^2.1.1",
            "espree": "^1.11.0",
            "gulp": "^3.8.10",
            "gulp-bump": "^0.2.2",
            "gulp-filter": "^2.0.0",
            "gulp-git": "^1.0.1",
            "gulp-tag-version": "^1.3.0",
            "jshint": "^2.5.6",
            "mocha": "^2.1.0"
        },
        "license": "BSD-2-Clause",
        "scripts": {
            "test": "npm run-script lint && npm run-script unit-test",
            "lint": "jshint estraverse.js",
            "unit-test": "mocha --compilers js:babel-register"
        }
    };
    module.exports = json;
},
/* eslint-scope/lib/reference.js */ function _(require, module, exports, __esModule, __esExport) {
    const READ = 0x1;
    const WRITE = 0x2;
    const RW = READ | WRITE;
    /**
     * A Reference represents a single occurrence of an identifier in code.
     * @class Reference
     */
    class Reference {
        constructor(ident, scope, flag, writeExpr, maybeImplicitGlobal, partial, init) {
            /**
             * Identifier syntax node.
             * @member {espreeIdentifier} Reference#identifier
             */
            this.identifier = ident;
            /**
             * Reference to the enclosing Scope.
             * @member {Scope} Reference#from
             */
            this.from = scope;
            /**
             * Whether the reference comes from a dynamic scope (such as 'eval',
             * 'with', etc.), and may be trapped by dynamic scopes.
             * @member {boolean} Reference#tainted
             */
            this.tainted = false;
            /**
             * The variable this reference is resolved with.
             * @member {Variable} Reference#resolved
             */
            this.resolved = null;
            /**
             * The read-write mode of the reference. (Value is one of {@link
             * Reference.READ}, {@link Reference.RW}, {@link Reference.WRITE}).
             * @member {number} Reference#flag
             * @private
             */
            this.flag = flag;
            if (this.isWrite()) {
                /**
                 * If reference is writeable, this is the tree being written to it.
                 * @member {espreeNode} Reference#writeExpr
                 */
                this.writeExpr = writeExpr;
                /**
                 * Whether the Reference might refer to a partial value of writeExpr.
                 * @member {boolean} Reference#partial
                 */
                this.partial = partial;
                /**
                 * Whether the Reference is to write of initialization.
                 * @member {boolean} Reference#init
                 */
                this.init = init;
            }
            this.__maybeImplicitGlobal = maybeImplicitGlobal;
        }
        /**
         * Whether the reference is static.
         * @method Reference#isStatic
         * @returns {boolean} static
         */
        isStatic() {
            return !this.tainted && this.resolved && this.resolved.scope.isStatic();
        }
        /**
         * Whether the reference is writeable.
         * @method Reference#isWrite
         * @returns {boolean} write
         */
        isWrite() {
            return !!(this.flag & Reference.WRITE);
        }
        /**
         * Whether the reference is readable.
         * @method Reference#isRead
         * @returns {boolean} read
         */
        isRead() {
            return !!(this.flag & Reference.READ);
        }
        /**
         * Whether the reference is read-only.
         * @method Reference#isReadOnly
         * @returns {boolean} read only
         */
        isReadOnly() {
            return this.flag === Reference.READ;
        }
        /**
         * Whether the reference is write-only.
         * @method Reference#isWriteOnly
         * @returns {boolean} write only
         */
        isWriteOnly() {
            return this.flag === Reference.WRITE;
        }
        /**
         * Whether the reference is read-write.
         * @method Reference#isReadWrite
         * @returns {boolean} read write
         */
        isReadWrite() {
            return this.flag === Reference.RW;
        }
    }
    /**
     * @constant Reference.READ
     * @private
     */
    Reference.READ = READ;
    /**
     * @constant Reference.WRITE
     * @private
     */
    Reference.WRITE = WRITE;
    /**
     * @constant Reference.RW
     * @private
     */
    Reference.RW = RW;
    module.exports = Reference;
},
/* eslint-scope/lib/variable.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * A Variable represents a locally scoped identifier. These include arguments to
     * functions.
     * @class Variable
     */
    class Variable {
        constructor(name, scope) {
            /**
             * The variable name, as given in the source code.
             * @member {String} Variable#name
             */
            this.name = name;
            /**
             * List of defining occurrences of this variable (like in 'var ...'
             * statements or as parameter), as AST nodes.
             * @member {espree.Identifier[]} Variable#identifiers
             */
            this.identifiers = [];
            /**
             * List of {@link Reference|references} of this variable (excluding parameter entries)
             * in its defining scope and all nested scopes. For defining
             * occurrences only see {@link Variable#defs}.
             * @member {Reference[]} Variable#references
             */
            this.references = [];
            /**
             * List of defining occurrences of this variable (like in 'var ...'
             * statements or as parameter), as custom objects.
             * @member {Definition[]} Variable#defs
             */
            this.defs = [];
            this.tainted = false;
            /**
             * Whether this is a stack variable.
             * @member {boolean} Variable#stack
             */
            this.stack = true;
            /**
             * Reference to the enclosing Scope.
             * @member {Scope} Variable#scope
             */
            this.scope = scope;
        }
    }
    Variable.CatchClause = "CatchClause";
    Variable.Parameter = "Parameter";
    Variable.FunctionName = "FunctionName";
    Variable.ClassName = "ClassName";
    Variable.Variable = "Variable";
    Variable.ImportBinding = "ImportBinding";
    Variable.ImplicitGlobalVariable = "ImplicitGlobalVariable";
    module.exports = Variable;
},
/* eslint-scope/lib/definition.js */ function _(require, module, exports, __esModule, __esExport) {
    const Variable = require(338) /* ./variable */;
    /**
     * @class Definition
     */
    class Definition {
        constructor(type, name, node, parent, index, kind) {
            /**
             * @member {String} Definition#type - type of the occurrence (e.g. "Parameter", "Variable", ...).
             */
            this.type = type;
            /**
             * @member {espree.Identifier} Definition#name - the identifier AST node of the occurrence.
             */
            this.name = name;
            /**
             * @member {espree.Node} Definition#node - the enclosing node of the identifier.
             */
            this.node = node;
            /**
             * @member {espree.Node?} Definition#parent - the enclosing statement node of the identifier.
             */
            this.parent = parent;
            /**
             * @member {Number?} Definition#index - the index in the declaration statement.
             */
            this.index = index;
            /**
             * @member {String?} Definition#kind - the kind of the declaration statement.
             */
            this.kind = kind;
        }
    }
    /**
     * @class ParameterDefinition
     */
    class ParameterDefinition extends Definition {
        constructor(name, node, index, rest) {
            super(Variable.Parameter, name, node, null, index, null);
            /**
             * Whether the parameter definition is a part of a rest parameter.
             * @member {boolean} ParameterDefinition#rest
             */
            this.rest = rest;
        }
    }
    module.exports = {
        ParameterDefinition,
        Definition
    };
},
/* eslint-scope/lib/referencer.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint-disable no-underscore-dangle */
    /* eslint-disable no-undefined */
    const Syntax = require(335) /* estraverse */.Syntax;
    const esrecurse = require(341) /* esrecurse */;
    const Reference = require(337) /* ./reference */;
    const Variable = require(338) /* ./variable */;
    const PatternVisitor = require(343) /* ./pattern-visitor */;
    const definition = require(339) /* ./definition */;
    const assert = require("assert");
    const ParameterDefinition = definition.ParameterDefinition;
    const Definition = definition.Definition;
    /**
     * Traverse identifier in pattern
     * @param {Object} options - options
     * @param {pattern} rootPattern - root pattern
     * @param {Refencer} referencer - referencer
     * @param {callback} callback - callback
     * @returns {void}
     */
    function traverseIdentifierInPattern(options, rootPattern, referencer, callback) {
        // Call the callback at left hand identifier nodes, and Collect right hand nodes.
        const visitor = new PatternVisitor(options, rootPattern, callback);
        visitor.visit(rootPattern);
        // Process the right hand nodes recursively.
        if (referencer !== null && referencer !== undefined) {
            visitor.rightHandNodes.forEach(referencer.visit, referencer);
        }
    }
    // Importing ImportDeclaration.
    // http://people.mozilla.org/~jorendorff/es6-draft.html#sec-moduledeclarationinstantiation
    // https://github.com/estree/estree/blob/master/es6.md#importdeclaration
    // FIXME: Now, we don't create module environment, because the context is
    // implementation dependent.
    class Importer extends esrecurse.Visitor {
        constructor(declaration, referencer) {
            super(null, referencer.options);
            this.declaration = declaration;
            this.referencer = referencer;
        }
        visitImport(id, specifier) {
            this.referencer.visitPattern(id, pattern => {
                this.referencer.currentScope().__define(pattern, new Definition(Variable.ImportBinding, pattern, specifier, this.declaration, null, null));
            });
        }
        ImportNamespaceSpecifier(node) {
            const local = (node.local || node.id);
            if (local) {
                this.visitImport(local, node);
            }
        }
        ImportDefaultSpecifier(node) {
            const local = (node.local || node.id);
            this.visitImport(local, node);
        }
        ImportSpecifier(node) {
            const local = (node.local || node.id);
            if (node.name) {
                this.visitImport(node.name, node);
            }
            else {
                this.visitImport(local, node);
            }
        }
    }
    // Referencing variables and creating bindings.
    class Referencer extends esrecurse.Visitor {
        constructor(options, scopeManager) {
            super(null, options);
            this.options = options;
            this.scopeManager = scopeManager;
            this.parent = null;
            this.isInnerMethodDefinition = false;
        }
        currentScope() {
            return this.scopeManager.__currentScope;
        }
        close(node) {
            while (this.currentScope() && node === this.currentScope().block) {
                this.scopeManager.__currentScope = this.currentScope().__close(this.scopeManager);
            }
        }
        pushInnerMethodDefinition(isInnerMethodDefinition) {
            const previous = this.isInnerMethodDefinition;
            this.isInnerMethodDefinition = isInnerMethodDefinition;
            return previous;
        }
        popInnerMethodDefinition(isInnerMethodDefinition) {
            this.isInnerMethodDefinition = isInnerMethodDefinition;
        }
        referencingDefaultValue(pattern, assignments, maybeImplicitGlobal, init) {
            const scope = this.currentScope();
            assignments.forEach(assignment => {
                scope.__referencing(pattern, Reference.WRITE, assignment.right, maybeImplicitGlobal, pattern !== assignment.left, init);
            });
        }
        visitPattern(node, options, callback) {
            let visitPatternOptions = options;
            let visitPatternCallback = callback;
            if (typeof options === "function") {
                visitPatternCallback = options;
                visitPatternOptions = { processRightHandNodes: false };
            }
            traverseIdentifierInPattern(this.options, node, visitPatternOptions.processRightHandNodes ? this : null, visitPatternCallback);
        }
        visitFunction(node) {
            let i, iz;
            // FunctionDeclaration name is defined in upper scope
            // NOTE: Not referring variableScope. It is intended.
            // Since
            //  in ES5, FunctionDeclaration should be in FunctionBody.
            //  in ES6, FunctionDeclaration should be block scoped.
            if (node.type === Syntax.FunctionDeclaration) {
                // id is defined in upper scope
                this.currentScope().__define(node.id, new Definition(Variable.FunctionName, node.id, node, null, null, null));
            }
            // FunctionExpression with name creates its special scope;
            // FunctionExpressionNameScope.
            if (node.type === Syntax.FunctionExpression && node.id) {
                this.scopeManager.__nestFunctionExpressionNameScope(node);
            }
            // Consider this function is in the MethodDefinition.
            this.scopeManager.__nestFunctionScope(node, this.isInnerMethodDefinition);
            const that = this;
            /**
             * Visit pattern callback
             * @param {pattern} pattern - pattern
             * @param {Object} info - info
             * @returns {void}
             */
            function visitPatternCallback(pattern, info) {
                that.currentScope().__define(pattern, new ParameterDefinition(pattern, node, i, info.rest));
                that.referencingDefaultValue(pattern, info.assignments, null, true);
            }
            // Process parameter declarations.
            for (i = 0, iz = node.params.length; i < iz; ++i) {
                this.visitPattern(node.params[i], { processRightHandNodes: true }, visitPatternCallback);
            }
            // if there's a rest argument, add that
            if (node.rest) {
                this.visitPattern({
                    type: "RestElement",
                    argument: node.rest
                }, pattern => {
                    this.currentScope().__define(pattern, new ParameterDefinition(pattern, node, node.params.length, true));
                });
            }
            // In TypeScript there are a number of function-like constructs which have no body,
            // so check it exists before traversing
            if (node.body) {
                // Skip BlockStatement to prevent creating BlockStatement scope.
                if (node.body.type === Syntax.BlockStatement) {
                    this.visitChildren(node.body);
                }
                else {
                    this.visit(node.body);
                }
            }
            this.close(node);
        }
        visitClass(node) {
            if (node.type === Syntax.ClassDeclaration) {
                this.currentScope().__define(node.id, new Definition(Variable.ClassName, node.id, node, null, null, null));
            }
            this.visit(node.superClass);
            this.scopeManager.__nestClassScope(node);
            if (node.id) {
                this.currentScope().__define(node.id, new Definition(Variable.ClassName, node.id, node));
            }
            this.visit(node.body);
            this.close(node);
        }
        visitProperty(node) {
            let previous;
            if (node.computed) {
                this.visit(node.key);
            }
            const isMethodDefinition = node.type === Syntax.MethodDefinition;
            if (isMethodDefinition) {
                previous = this.pushInnerMethodDefinition(true);
            }
            this.visit(node.value);
            if (isMethodDefinition) {
                this.popInnerMethodDefinition(previous);
            }
        }
        visitForIn(node) {
            if (node.left.type === Syntax.VariableDeclaration && node.left.kind !== "var") {
                this.scopeManager.__nestForScope(node);
            }
            if (node.left.type === Syntax.VariableDeclaration) {
                this.visit(node.left);
                this.visitPattern(node.left.declarations[0].id, pattern => {
                    this.currentScope().__referencing(pattern, Reference.WRITE, node.right, null, true, true);
                });
            }
            else {
                this.visitPattern(node.left, { processRightHandNodes: true }, (pattern, info) => {
                    let maybeImplicitGlobal = null;
                    if (!this.currentScope().isStrict) {
                        maybeImplicitGlobal = {
                            pattern,
                            node
                        };
                    }
                    this.referencingDefaultValue(pattern, info.assignments, maybeImplicitGlobal, false);
                    this.currentScope().__referencing(pattern, Reference.WRITE, node.right, maybeImplicitGlobal, true, false);
                });
            }
            this.visit(node.right);
            this.visit(node.body);
            this.close(node);
        }
        visitVariableDeclaration(variableTargetScope, type, node, index) {
            const decl = node.declarations[index];
            const init = decl.init;
            this.visitPattern(decl.id, { processRightHandNodes: true }, (pattern, info) => {
                variableTargetScope.__define(pattern, new Definition(type, pattern, decl, node, index, node.kind));
                this.referencingDefaultValue(pattern, info.assignments, null, true);
                if (init) {
                    this.currentScope().__referencing(pattern, Reference.WRITE, init, null, !info.topLevel, true);
                }
            });
        }
        AssignmentExpression(node) {
            if (PatternVisitor.isPattern(node.left)) {
                if (node.operator === "=") {
                    this.visitPattern(node.left, { processRightHandNodes: true }, (pattern, info) => {
                        let maybeImplicitGlobal = null;
                        if (!this.currentScope().isStrict) {
                            maybeImplicitGlobal = {
                                pattern,
                                node
                            };
                        }
                        this.referencingDefaultValue(pattern, info.assignments, maybeImplicitGlobal, false);
                        this.currentScope().__referencing(pattern, Reference.WRITE, node.right, maybeImplicitGlobal, !info.topLevel, false);
                    });
                }
                else {
                    this.currentScope().__referencing(node.left, Reference.RW, node.right);
                }
            }
            else {
                this.visit(node.left);
            }
            this.visit(node.right);
        }
        CatchClause(node) {
            this.scopeManager.__nestCatchScope(node);
            this.visitPattern(node.param, { processRightHandNodes: true }, (pattern, info) => {
                this.currentScope().__define(pattern, new Definition(Variable.CatchClause, node.param, node, null, null, null));
                this.referencingDefaultValue(pattern, info.assignments, null, true);
            });
            this.visit(node.body);
            this.close(node);
        }
        Program(node) {
            this.scopeManager.__nestGlobalScope(node);
            if (this.scopeManager.__isNodejsScope()) {
                // Force strictness of GlobalScope to false when using node.js scope.
                this.currentScope().isStrict = false;
                this.scopeManager.__nestFunctionScope(node, false);
            }
            if (this.scopeManager.__isES6() && this.scopeManager.isModule()) {
                this.scopeManager.__nestModuleScope(node);
            }
            if (this.scopeManager.isStrictModeSupported() && this.scopeManager.isImpliedStrict()) {
                this.currentScope().isStrict = true;
            }
            this.visitChildren(node);
            this.close(node);
        }
        Identifier(node) {
            this.currentScope().__referencing(node);
        }
        UpdateExpression(node) {
            if (PatternVisitor.isPattern(node.argument)) {
                this.currentScope().__referencing(node.argument, Reference.RW, null);
            }
            else {
                this.visitChildren(node);
            }
        }
        MemberExpression(node) {
            this.visit(node.object);
            if (node.computed) {
                this.visit(node.property);
            }
        }
        Property(node) {
            this.visitProperty(node);
        }
        MethodDefinition(node) {
            this.visitProperty(node);
        }
        BreakStatement() { } // eslint-disable-line class-methods-use-this
        ContinueStatement() { } // eslint-disable-line class-methods-use-this
        LabeledStatement(node) {
            this.visit(node.body);
        }
        ForStatement(node) {
            // Create ForStatement declaration.
            // NOTE: In ES6, ForStatement dynamically generates
            // per iteration environment. However, escope is
            // a static analyzer, we only generate one scope for ForStatement.
            if (node.init && node.init.type === Syntax.VariableDeclaration && node.init.kind !== "var") {
                this.scopeManager.__nestForScope(node);
            }
            this.visitChildren(node);
            this.close(node);
        }
        ClassExpression(node) {
            this.visitClass(node);
        }
        ClassDeclaration(node) {
            this.visitClass(node);
        }
        CallExpression(node) {
            // Check this is direct call to eval
            if (!this.scopeManager.__ignoreEval() && node.callee.type === Syntax.Identifier && node.callee.name === "eval") {
                // NOTE: This should be `variableScope`. Since direct eval call always creates Lexical environment and
                // let / const should be enclosed into it. Only VariableDeclaration affects on the caller's environment.
                this.currentScope().variableScope.__detectEval();
            }
            this.visitChildren(node);
        }
        BlockStatement(node) {
            if (this.scopeManager.__isES6()) {
                this.scopeManager.__nestBlockScope(node);
            }
            this.visitChildren(node);
            this.close(node);
        }
        ThisExpression() {
            this.currentScope().variableScope.__detectThis();
        }
        WithStatement(node) {
            this.visit(node.object);
            // Then nest scope for WithStatement.
            this.scopeManager.__nestWithScope(node);
            this.visit(node.body);
            this.close(node);
        }
        VariableDeclaration(node) {
            const variableTargetScope = (node.kind === "var") ? this.currentScope().variableScope : this.currentScope();
            for (let i = 0, iz = node.declarations.length; i < iz; ++i) {
                const decl = node.declarations[i];
                this.visitVariableDeclaration(variableTargetScope, Variable.Variable, node, i);
                if (decl.init) {
                    this.visit(decl.init);
                }
            }
        }
        // sec 13.11.8
        SwitchStatement(node) {
            this.visit(node.discriminant);
            if (this.scopeManager.__isES6()) {
                this.scopeManager.__nestSwitchScope(node);
            }
            for (let i = 0, iz = node.cases.length; i < iz; ++i) {
                this.visit(node.cases[i]);
            }
            this.close(node);
        }
        FunctionDeclaration(node) {
            this.visitFunction(node);
        }
        FunctionExpression(node) {
            this.visitFunction(node);
        }
        ForOfStatement(node) {
            this.visitForIn(node);
        }
        ForInStatement(node) {
            this.visitForIn(node);
        }
        ArrowFunctionExpression(node) {
            this.visitFunction(node);
        }
        ImportDeclaration(node) {
            assert(this.scopeManager.__isES6() && this.scopeManager.isModule(), "ImportDeclaration should appear when the mode is ES6 and in the module context.");
            const importer = new Importer(node, this);
            importer.visit(node);
        }
        visitExportDeclaration(node) {
            if (node.source) {
                return;
            }
            if (node.declaration) {
                this.visit(node.declaration);
                return;
            }
            this.visitChildren(node);
        }
        // TODO: ExportDeclaration doesn't exist. for bc?
        ExportDeclaration(node) {
            this.visitExportDeclaration(node);
        }
        ExportAllDeclaration(node) {
            this.visitExportDeclaration(node);
        }
        ExportDefaultDeclaration(node) {
            this.visitExportDeclaration(node);
        }
        ExportNamedDeclaration(node) {
            this.visitExportDeclaration(node);
        }
        ExportSpecifier(node) {
            // TODO: `node.id` doesn't exist. for bc?
            const local = (node.id || node.local);
            this.visit(local);
        }
        MetaProperty() {
            // do nothing.
        }
    }
    module.exports = Referencer;
},
/* esrecurse/esrecurse.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
      Copyright (C) 2014 Yusuke Suzuki <utatane.tea@gmail.com>
    
      Redistribution and use in source and binary forms, with or without
      modification, are permitted provided that the following conditions are met:
    
        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
    
      THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
      AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
      IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
      ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
      DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
      (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
      LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
      ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
      (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
      THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    */
    (function () {
        'use strict';
        var estraverse = require(335) /* estraverse */;
        function isNode(node) {
            if (node == null) {
                return false;
            }
            return typeof node === 'object' && typeof node.type === 'string';
        }
        function isProperty(nodeType, key) {
            return (nodeType === estraverse.Syntax.ObjectExpression || nodeType === estraverse.Syntax.ObjectPattern) && key === 'properties';
        }
        function Visitor(visitor, options) {
            options = options || {};
            this.__visitor = visitor || this;
            this.__childVisitorKeys = options.childVisitorKeys
                ? Object.assign({}, estraverse.VisitorKeys, options.childVisitorKeys)
                : estraverse.VisitorKeys;
            if (options.fallback === 'iteration') {
                this.__fallback = Object.keys;
            }
            else if (typeof options.fallback === 'function') {
                this.__fallback = options.fallback;
            }
        }
        /* Default method for visiting children.
         * When you need to call default visiting operation inside custom visiting
         * operation, you can use it with `this.visitChildren(node)`.
         */
        Visitor.prototype.visitChildren = function (node) {
            var type, children, i, iz, j, jz, child;
            if (node == null) {
                return;
            }
            type = node.type || estraverse.Syntax.Property;
            children = this.__childVisitorKeys[type];
            if (!children) {
                if (this.__fallback) {
                    children = this.__fallback(node);
                }
                else {
                    throw new Error('Unknown node type ' + type + '.');
                }
            }
            for (i = 0, iz = children.length; i < iz; ++i) {
                child = node[children[i]];
                if (child) {
                    if (Array.isArray(child)) {
                        for (j = 0, jz = child.length; j < jz; ++j) {
                            if (child[j]) {
                                if (isNode(child[j]) || isProperty(type, children[i])) {
                                    this.visit(child[j]);
                                }
                            }
                        }
                    }
                    else if (isNode(child)) {
                        this.visit(child);
                    }
                }
            }
        };
        /* Dispatching node. */
        Visitor.prototype.visit = function (node) {
            var type;
            if (node == null) {
                return;
            }
            type = node.type || estraverse.Syntax.Property;
            if (this.__visitor[type]) {
                this.__visitor[type].call(this, node);
                return;
            }
            this.visitChildren(node);
        };
        exports.version = require(342) /* ./package.json */.version;
        exports.Visitor = Visitor;
        exports.visit = function (node, visitor, options) {
            var v = new Visitor(visitor, options);
            v.visit(node);
        };
    }());
},
/* esrecurse/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "esrecurse",
        "description": "ECMAScript AST recursive visitor",
        "homepage": "https://github.com/estools/esrecurse",
        "main": "esrecurse.js",
        "version": "4.3.0",
        "engines": {
            "node": ">=4.0"
        },
        "maintainers": [
            {
                "name": "Yusuke Suzuki",
                "email": "utatane.tea@gmail.com",
                "web": "https://github.com/Constellation"
            }
        ],
        "repository": {
            "type": "git",
            "url": "https://github.com/estools/esrecurse.git"
        },
        "dependencies": {
            "estraverse": "^5.2.0"
        },
        "devDependencies": {
            "babel-cli": "^6.24.1",
            "babel-eslint": "^7.2.3",
            "babel-preset-es2015": "^6.24.1",
            "babel-register": "^6.24.1",
            "chai": "^4.0.2",
            "esprima": "^4.0.0",
            "gulp": "^3.9.0",
            "gulp-bump": "^2.7.0",
            "gulp-eslint": "^4.0.0",
            "gulp-filter": "^5.0.0",
            "gulp-git": "^2.4.1",
            "gulp-mocha": "^4.3.1",
            "gulp-tag-version": "^1.2.1",
            "jsdoc": "^3.3.0-alpha10",
            "minimist": "^1.1.0"
        },
        "license": "BSD-2-Clause",
        "scripts": {
            "test": "gulp travis",
            "unit-test": "gulp test",
            "lint": "gulp lint"
        },
        "babel": {
            "presets": [
                "es2015"
            ]
        }
    };
    module.exports = json;
},
/* eslint-scope/lib/pattern-visitor.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint-disable no-undefined */
    const Syntax = require(335) /* estraverse */.Syntax;
    const esrecurse = require(341) /* esrecurse */;
    /**
     * Get last array element
     * @param {array} xs - array
     * @returns {any} Last elment
     */
    function getLast(xs) {
        return xs[xs.length - 1] || null;
    }
    class PatternVisitor extends esrecurse.Visitor {
        static isPattern(node) {
            const nodeType = node.type;
            return (nodeType === Syntax.Identifier ||
                nodeType === Syntax.ObjectPattern ||
                nodeType === Syntax.ArrayPattern ||
                nodeType === Syntax.SpreadElement ||
                nodeType === Syntax.RestElement ||
                nodeType === Syntax.AssignmentPattern);
        }
        constructor(options, rootPattern, callback) {
            super(null, options);
            this.rootPattern = rootPattern;
            this.callback = callback;
            this.assignments = [];
            this.rightHandNodes = [];
            this.restElements = [];
        }
        Identifier(pattern) {
            const lastRestElement = getLast(this.restElements);
            this.callback(pattern, {
                topLevel: pattern === this.rootPattern,
                rest: lastRestElement !== null && lastRestElement !== undefined && lastRestElement.argument === pattern,
                assignments: this.assignments
            });
        }
        Property(property) {
            // Computed property's key is a right hand node.
            if (property.computed) {
                this.rightHandNodes.push(property.key);
            }
            // If it's shorthand, its key is same as its value.
            // If it's shorthand and has its default value, its key is same as its value.left (the value is AssignmentPattern).
            // If it's not shorthand, the name of new variable is its value's.
            this.visit(property.value);
        }
        ArrayPattern(pattern) {
            for (let i = 0, iz = pattern.elements.length; i < iz; ++i) {
                const element = pattern.elements[i];
                this.visit(element);
            }
        }
        AssignmentPattern(pattern) {
            this.assignments.push(pattern);
            this.visit(pattern.left);
            this.rightHandNodes.push(pattern.right);
            this.assignments.pop();
        }
        RestElement(pattern) {
            this.restElements.push(pattern);
            this.visit(pattern.argument);
            this.restElements.pop();
        }
        MemberExpression(node) {
            // Computed property's key is a right hand node.
            if (node.computed) {
                this.rightHandNodes.push(node.property);
            }
            // the object is only read, write to its property.
            this.rightHandNodes.push(node.object);
        }
        //
        // ForInStatement.left and AssignmentExpression.left are LeftHandSideExpression.
        // By spec, LeftHandSideExpression is Pattern or MemberExpression.
        //   (see also: https://github.com/estree/estree/pull/20#issuecomment-74584758)
        // But espree 2.0 parses to ArrayExpression, ObjectExpression, etc...
        //
        SpreadElement(node) {
            this.visit(node.argument);
        }
        ArrayExpression(node) {
            node.elements.forEach(this.visit, this);
        }
        AssignmentExpression(node) {
            this.assignments.push(node);
            this.visit(node.left);
            this.rightHandNodes.push(node.right);
            this.assignments.pop();
        }
        CallExpression(node) {
            // arguments are right hand nodes.
            node.arguments.forEach(a => {
                this.rightHandNodes.push(a);
            });
            this.visit(node.callee);
        }
    }
    module.exports = PatternVisitor;
},
/* eslint-scope/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "eslint-scope",
        "description": "ECMAScript scope analyzer for ESLint",
        "homepage": "http://github.com/eslint/eslint-scope",
        "main": "lib/index.js",
        "version": "5.1.1",
        "engines": {
            "node": ">=8.0.0"
        },
        "repository": "eslint/eslint-scope",
        "bugs": {
            "url": "https://github.com/eslint/eslint-scope/issues"
        },
        "license": "BSD-2-Clause",
        "scripts": {
            "test": "node Makefile.js test",
            "lint": "node Makefile.js lint",
            "generate-release": "eslint-generate-release",
            "generate-alpharelease": "eslint-generate-prerelease alpha",
            "generate-betarelease": "eslint-generate-prerelease beta",
            "generate-rcrelease": "eslint-generate-prerelease rc",
            "publish-release": "eslint-publish-release"
        },
        "files": [
            "LICENSE",
            "README.md",
            "lib"
        ],
        "dependencies": {
            "esrecurse": "^4.3.0",
            "estraverse": "^4.1.1"
        },
        "devDependencies": {
            "@typescript-eslint/parser": "^1.11.0",
            "chai": "^4.2.0",
            "eslint": "^6.0.1",
            "eslint-config-eslint": "^5.0.1",
            "eslint-plugin-node": "^9.1.0",
            "eslint-release": "^1.0.0",
            "eslint-visitor-keys": "^1.2.0",
            "espree": "^7.1.0",
            "istanbul": "^0.4.5",
            "mocha": "^6.1.4",
            "npm-license": "^0.3.3",
            "shelljs": "^0.8.3",
            "typescript": "^3.5.2"
        }
    };
    module.exports = json;
},
/* lodash.merge/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Lodash (Custom Build) <https://lodash.com/>
     * Build: `lodash modularize exports="npm" -o ./`
     * Copyright OpenJS Foundation and other contributors <https://openjsf.org/>
     * Released under MIT license <https://lodash.com/license>
     * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
     * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
     */
    /** Used as the size to enable large array optimizations. */
    var LARGE_ARRAY_SIZE = 200;
    /** Used to stand-in for `undefined` hash values. */
    var HASH_UNDEFINED = '__lodash_hash_undefined__';
    /** Used to detect hot functions by number of calls within a span of milliseconds. */
    var HOT_COUNT = 800, HOT_SPAN = 16;
    /** Used as references for various `Number` constants. */
    var MAX_SAFE_INTEGER = 9007199254740991;
    /** `Object#toString` result references. */
    var argsTag = '[object Arguments]', arrayTag = '[object Array]', asyncTag = '[object AsyncFunction]', boolTag = '[object Boolean]', dateTag = '[object Date]', errorTag = '[object Error]', funcTag = '[object Function]', genTag = '[object GeneratorFunction]', mapTag = '[object Map]', numberTag = '[object Number]', nullTag = '[object Null]', objectTag = '[object Object]', proxyTag = '[object Proxy]', regexpTag = '[object RegExp]', setTag = '[object Set]', stringTag = '[object String]', undefinedTag = '[object Undefined]', weakMapTag = '[object WeakMap]';
    var arrayBufferTag = '[object ArrayBuffer]', dataViewTag = '[object DataView]', float32Tag = '[object Float32Array]', float64Tag = '[object Float64Array]', int8Tag = '[object Int8Array]', int16Tag = '[object Int16Array]', int32Tag = '[object Int32Array]', uint8Tag = '[object Uint8Array]', uint8ClampedTag = '[object Uint8ClampedArray]', uint16Tag = '[object Uint16Array]', uint32Tag = '[object Uint32Array]';
    /**
     * Used to match `RegExp`
     * [syntax characters](http://ecma-international.org/ecma-262/7.0/#sec-patterns).
     */
    var reRegExpChar = /[\\^$.*+?()[\]{}|]/g;
    /** Used to detect host constructors (Safari). */
    var reIsHostCtor = /^\[object .+?Constructor\]$/;
    /** Used to detect unsigned integer values. */
    var reIsUint = /^(?:0|[1-9]\d*)$/;
    /** Used to identify `toStringTag` values of typed arrays. */
    var typedArrayTags = {};
    typedArrayTags[float32Tag] = typedArrayTags[float64Tag] =
        typedArrayTags[int8Tag] = typedArrayTags[int16Tag] =
            typedArrayTags[int32Tag] = typedArrayTags[uint8Tag] =
                typedArrayTags[uint8ClampedTag] = typedArrayTags[uint16Tag] =
                    typedArrayTags[uint32Tag] = true;
    typedArrayTags[argsTag] = typedArrayTags[arrayTag] =
        typedArrayTags[arrayBufferTag] = typedArrayTags[boolTag] =
            typedArrayTags[dataViewTag] = typedArrayTags[dateTag] =
                typedArrayTags[errorTag] = typedArrayTags[funcTag] =
                    typedArrayTags[mapTag] = typedArrayTags[numberTag] =
                        typedArrayTags[objectTag] = typedArrayTags[regexpTag] =
                            typedArrayTags[setTag] = typedArrayTags[stringTag] =
                                typedArrayTags[weakMapTag] = false;
    /** Detect free variable `global` from Node.js. */
    var freeGlobal = typeof global == 'object' && global && global.Object === Object && global;
    /** Detect free variable `self`. */
    var freeSelf = typeof self == 'object' && self && self.Object === Object && self;
    /** Used as a reference to the global object. */
    var root = freeGlobal || freeSelf || Function('return this')();
    /** Detect free variable `exports`. */
    var freeExports = typeof exports == 'object' && exports && !exports.nodeType && exports;
    /** Detect free variable `module`. */
    var freeModule = freeExports && typeof module == 'object' && module && !module.nodeType && module;
    /** Detect the popular CommonJS extension `module.exports`. */
    var moduleExports = freeModule && freeModule.exports === freeExports;
    /** Detect free variable `process` from Node.js. */
    var freeProcess = moduleExports && freeGlobal.process;
    /** Used to access faster Node.js helpers. */
    var nodeUtil = (function () {
        try {
            // Use `util.types` for Node.js 10+.
            var types = freeModule && freeModule.require && freeModule.require('util').types;
            if (types) {
                return types;
            }
            // Legacy `process.binding('util')` for Node.js < 10.
            return freeProcess && freeProcess.binding && freeProcess.binding('util');
        }
        catch (e) { }
    }());
    /* Node.js helper references. */
    var nodeIsTypedArray = nodeUtil && nodeUtil.isTypedArray;
    /**
     * A faster alternative to `Function#apply`, this function invokes `func`
     * with the `this` binding of `thisArg` and the arguments of `args`.
     *
     * @private
     * @param {Function} func The function to invoke.
     * @param {*} thisArg The `this` binding of `func`.
     * @param {Array} args The arguments to invoke `func` with.
     * @returns {*} Returns the result of `func`.
     */
    function apply(func, thisArg, args) {
        switch (args.length) {
            case 0: return func.call(thisArg);
            case 1: return func.call(thisArg, args[0]);
            case 2: return func.call(thisArg, args[0], args[1]);
            case 3: return func.call(thisArg, args[0], args[1], args[2]);
        }
        return func.apply(thisArg, args);
    }
    /**
     * The base implementation of `_.times` without support for iteratee shorthands
     * or max array length checks.
     *
     * @private
     * @param {number} n The number of times to invoke `iteratee`.
     * @param {Function} iteratee The function invoked per iteration.
     * @returns {Array} Returns the array of results.
     */
    function baseTimes(n, iteratee) {
        var index = -1, result = Array(n);
        while (++index < n) {
            result[index] = iteratee(index);
        }
        return result;
    }
    /**
     * The base implementation of `_.unary` without support for storing metadata.
     *
     * @private
     * @param {Function} func The function to cap arguments for.
     * @returns {Function} Returns the new capped function.
     */
    function baseUnary(func) {
        return function (value) {
            return func(value);
        };
    }
    /**
     * Gets the value at `key` of `object`.
     *
     * @private
     * @param {Object} [object] The object to query.
     * @param {string} key The key of the property to get.
     * @returns {*} Returns the property value.
     */
    function getValue(object, key) {
        return object == null ? undefined : object[key];
    }
    /**
     * Creates a unary function that invokes `func` with its argument transformed.
     *
     * @private
     * @param {Function} func The function to wrap.
     * @param {Function} transform The argument transform.
     * @returns {Function} Returns the new function.
     */
    function overArg(func, transform) {
        return function (arg) {
            return func(transform(arg));
        };
    }
    /** Used for built-in method references. */
    var arrayProto = Array.prototype, funcProto = Function.prototype, objectProto = Object.prototype;
    /** Used to detect overreaching core-js shims. */
    var coreJsData = root['__core-js_shared__'];
    /** Used to resolve the decompiled source of functions. */
    var funcToString = funcProto.toString;
    /** Used to check objects for own properties. */
    var hasOwnProperty = objectProto.hasOwnProperty;
    /** Used to detect methods masquerading as native. */
    var maskSrcKey = (function () {
        var uid = /[^.]+$/.exec(coreJsData && coreJsData.keys && coreJsData.keys.IE_PROTO || '');
        return uid ? ('Symbol(src)_1.' + uid) : '';
    }());
    /**
     * Used to resolve the
     * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
     * of values.
     */
    var nativeObjectToString = objectProto.toString;
    /** Used to infer the `Object` constructor. */
    var objectCtorString = funcToString.call(Object);
    /** Used to detect if a method is native. */
    var reIsNative = RegExp('^' +
        funcToString.call(hasOwnProperty).replace(reRegExpChar, '\\$&')
            .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$');
    /** Built-in value references. */
    var Buffer = moduleExports ? root.Buffer : undefined, Symbol = root.Symbol, Uint8Array = root.Uint8Array, allocUnsafe = Buffer ? Buffer.allocUnsafe : undefined, getPrototype = overArg(Object.getPrototypeOf, Object), objectCreate = Object.create, propertyIsEnumerable = objectProto.propertyIsEnumerable, splice = arrayProto.splice, symToStringTag = Symbol ? Symbol.toStringTag : undefined;
    var defineProperty = (function () {
        try {
            var func = getNative(Object, 'defineProperty');
            func({}, '', {});
            return func;
        }
        catch (e) { }
    }());
    /* Built-in method references for those with the same name as other `lodash` methods. */
    var nativeIsBuffer = Buffer ? Buffer.isBuffer : undefined, nativeMax = Math.max, nativeNow = Date.now;
    /* Built-in method references that are verified to be native. */
    var Map = getNative(root, 'Map'), nativeCreate = getNative(Object, 'create');
    /**
     * The base implementation of `_.create` without support for assigning
     * properties to the created object.
     *
     * @private
     * @param {Object} proto The object to inherit from.
     * @returns {Object} Returns the new object.
     */
    var baseCreate = (function () {
        function object() { }
        return function (proto) {
            if (!isObject(proto)) {
                return {};
            }
            if (objectCreate) {
                return objectCreate(proto);
            }
            object.prototype = proto;
            var result = new object;
            object.prototype = undefined;
            return result;
        };
    }());
    /**
     * Creates a hash object.
     *
     * @private
     * @constructor
     * @param {Array} [entries] The key-value pairs to cache.
     */
    function Hash(entries) {
        var index = -1, length = entries == null ? 0 : entries.length;
        this.clear();
        while (++index < length) {
            var entry = entries[index];
            this.set(entry[0], entry[1]);
        }
    }
    /**
     * Removes all key-value entries from the hash.
     *
     * @private
     * @name clear
     * @memberOf Hash
     */
    function hashClear() {
        this.__data__ = nativeCreate ? nativeCreate(null) : {};
        this.size = 0;
    }
    /**
     * Removes `key` and its value from the hash.
     *
     * @private
     * @name delete
     * @memberOf Hash
     * @param {Object} hash The hash to modify.
     * @param {string} key The key of the value to remove.
     * @returns {boolean} Returns `true` if the entry was removed, else `false`.
     */
    function hashDelete(key) {
        var result = this.has(key) && delete this.__data__[key];
        this.size -= result ? 1 : 0;
        return result;
    }
    /**
     * Gets the hash value for `key`.
     *
     * @private
     * @name get
     * @memberOf Hash
     * @param {string} key The key of the value to get.
     * @returns {*} Returns the entry value.
     */
    function hashGet(key) {
        var data = this.__data__;
        if (nativeCreate) {
            var result = data[key];
            return result === HASH_UNDEFINED ? undefined : result;
        }
        return hasOwnProperty.call(data, key) ? data[key] : undefined;
    }
    /**
     * Checks if a hash value for `key` exists.
     *
     * @private
     * @name has
     * @memberOf Hash
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function hashHas(key) {
        var data = this.__data__;
        return nativeCreate ? (data[key] !== undefined) : hasOwnProperty.call(data, key);
    }
    /**
     * Sets the hash `key` to `value`.
     *
     * @private
     * @name set
     * @memberOf Hash
     * @param {string} key The key of the value to set.
     * @param {*} value The value to set.
     * @returns {Object} Returns the hash instance.
     */
    function hashSet(key, value) {
        var data = this.__data__;
        this.size += this.has(key) ? 0 : 1;
        data[key] = (nativeCreate && value === undefined) ? HASH_UNDEFINED : value;
        return this;
    }
    // Add methods to `Hash`.
    Hash.prototype.clear = hashClear;
    Hash.prototype['delete'] = hashDelete;
    Hash.prototype.get = hashGet;
    Hash.prototype.has = hashHas;
    Hash.prototype.set = hashSet;
    /**
     * Creates an list cache object.
     *
     * @private
     * @constructor
     * @param {Array} [entries] The key-value pairs to cache.
     */
    function ListCache(entries) {
        var index = -1, length = entries == null ? 0 : entries.length;
        this.clear();
        while (++index < length) {
            var entry = entries[index];
            this.set(entry[0], entry[1]);
        }
    }
    /**
     * Removes all key-value entries from the list cache.
     *
     * @private
     * @name clear
     * @memberOf ListCache
     */
    function listCacheClear() {
        this.__data__ = [];
        this.size = 0;
    }
    /**
     * Removes `key` and its value from the list cache.
     *
     * @private
     * @name delete
     * @memberOf ListCache
     * @param {string} key The key of the value to remove.
     * @returns {boolean} Returns `true` if the entry was removed, else `false`.
     */
    function listCacheDelete(key) {
        var data = this.__data__, index = assocIndexOf(data, key);
        if (index < 0) {
            return false;
        }
        var lastIndex = data.length - 1;
        if (index == lastIndex) {
            data.pop();
        }
        else {
            splice.call(data, index, 1);
        }
        --this.size;
        return true;
    }
    /**
     * Gets the list cache value for `key`.
     *
     * @private
     * @name get
     * @memberOf ListCache
     * @param {string} key The key of the value to get.
     * @returns {*} Returns the entry value.
     */
    function listCacheGet(key) {
        var data = this.__data__, index = assocIndexOf(data, key);
        return index < 0 ? undefined : data[index][1];
    }
    /**
     * Checks if a list cache value for `key` exists.
     *
     * @private
     * @name has
     * @memberOf ListCache
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function listCacheHas(key) {
        return assocIndexOf(this.__data__, key) > -1;
    }
    /**
     * Sets the list cache `key` to `value`.
     *
     * @private
     * @name set
     * @memberOf ListCache
     * @param {string} key The key of the value to set.
     * @param {*} value The value to set.
     * @returns {Object} Returns the list cache instance.
     */
    function listCacheSet(key, value) {
        var data = this.__data__, index = assocIndexOf(data, key);
        if (index < 0) {
            ++this.size;
            data.push([key, value]);
        }
        else {
            data[index][1] = value;
        }
        return this;
    }
    // Add methods to `ListCache`.
    ListCache.prototype.clear = listCacheClear;
    ListCache.prototype['delete'] = listCacheDelete;
    ListCache.prototype.get = listCacheGet;
    ListCache.prototype.has = listCacheHas;
    ListCache.prototype.set = listCacheSet;
    /**
     * Creates a map cache object to store key-value pairs.
     *
     * @private
     * @constructor
     * @param {Array} [entries] The key-value pairs to cache.
     */
    function MapCache(entries) {
        var index = -1, length = entries == null ? 0 : entries.length;
        this.clear();
        while (++index < length) {
            var entry = entries[index];
            this.set(entry[0], entry[1]);
        }
    }
    /**
     * Removes all key-value entries from the map.
     *
     * @private
     * @name clear
     * @memberOf MapCache
     */
    function mapCacheClear() {
        this.size = 0;
        this.__data__ = {
            'hash': new Hash,
            'map': new (Map || ListCache),
            'string': new Hash
        };
    }
    /**
     * Removes `key` and its value from the map.
     *
     * @private
     * @name delete
     * @memberOf MapCache
     * @param {string} key The key of the value to remove.
     * @returns {boolean} Returns `true` if the entry was removed, else `false`.
     */
    function mapCacheDelete(key) {
        var result = getMapData(this, key)['delete'](key);
        this.size -= result ? 1 : 0;
        return result;
    }
    /**
     * Gets the map value for `key`.
     *
     * @private
     * @name get
     * @memberOf MapCache
     * @param {string} key The key of the value to get.
     * @returns {*} Returns the entry value.
     */
    function mapCacheGet(key) {
        return getMapData(this, key).get(key);
    }
    /**
     * Checks if a map value for `key` exists.
     *
     * @private
     * @name has
     * @memberOf MapCache
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function mapCacheHas(key) {
        return getMapData(this, key).has(key);
    }
    /**
     * Sets the map `key` to `value`.
     *
     * @private
     * @name set
     * @memberOf MapCache
     * @param {string} key The key of the value to set.
     * @param {*} value The value to set.
     * @returns {Object} Returns the map cache instance.
     */
    function mapCacheSet(key, value) {
        var data = getMapData(this, key), size = data.size;
        data.set(key, value);
        this.size += data.size == size ? 0 : 1;
        return this;
    }
    // Add methods to `MapCache`.
    MapCache.prototype.clear = mapCacheClear;
    MapCache.prototype['delete'] = mapCacheDelete;
    MapCache.prototype.get = mapCacheGet;
    MapCache.prototype.has = mapCacheHas;
    MapCache.prototype.set = mapCacheSet;
    /**
     * Creates a stack cache object to store key-value pairs.
     *
     * @private
     * @constructor
     * @param {Array} [entries] The key-value pairs to cache.
     */
    function Stack(entries) {
        var data = this.__data__ = new ListCache(entries);
        this.size = data.size;
    }
    /**
     * Removes all key-value entries from the stack.
     *
     * @private
     * @name clear
     * @memberOf Stack
     */
    function stackClear() {
        this.__data__ = new ListCache;
        this.size = 0;
    }
    /**
     * Removes `key` and its value from the stack.
     *
     * @private
     * @name delete
     * @memberOf Stack
     * @param {string} key The key of the value to remove.
     * @returns {boolean} Returns `true` if the entry was removed, else `false`.
     */
    function stackDelete(key) {
        var data = this.__data__, result = data['delete'](key);
        this.size = data.size;
        return result;
    }
    /**
     * Gets the stack value for `key`.
     *
     * @private
     * @name get
     * @memberOf Stack
     * @param {string} key The key of the value to get.
     * @returns {*} Returns the entry value.
     */
    function stackGet(key) {
        return this.__data__.get(key);
    }
    /**
     * Checks if a stack value for `key` exists.
     *
     * @private
     * @name has
     * @memberOf Stack
     * @param {string} key The key of the entry to check.
     * @returns {boolean} Returns `true` if an entry for `key` exists, else `false`.
     */
    function stackHas(key) {
        return this.__data__.has(key);
    }
    /**
     * Sets the stack `key` to `value`.
     *
     * @private
     * @name set
     * @memberOf Stack
     * @param {string} key The key of the value to set.
     * @param {*} value The value to set.
     * @returns {Object} Returns the stack cache instance.
     */
    function stackSet(key, value) {
        var data = this.__data__;
        if (data instanceof ListCache) {
            var pairs = data.__data__;
            if (!Map || (pairs.length < LARGE_ARRAY_SIZE - 1)) {
                pairs.push([key, value]);
                this.size = ++data.size;
                return this;
            }
            data = this.__data__ = new MapCache(pairs);
        }
        data.set(key, value);
        this.size = data.size;
        return this;
    }
    // Add methods to `Stack`.
    Stack.prototype.clear = stackClear;
    Stack.prototype['delete'] = stackDelete;
    Stack.prototype.get = stackGet;
    Stack.prototype.has = stackHas;
    Stack.prototype.set = stackSet;
    /**
     * Creates an array of the enumerable property names of the array-like `value`.
     *
     * @private
     * @param {*} value The value to query.
     * @param {boolean} inherited Specify returning inherited property names.
     * @returns {Array} Returns the array of property names.
     */
    function arrayLikeKeys(value, inherited) {
        var isArr = isArray(value), isArg = !isArr && isArguments(value), isBuff = !isArr && !isArg && isBuffer(value), isType = !isArr && !isArg && !isBuff && isTypedArray(value), skipIndexes = isArr || isArg || isBuff || isType, result = skipIndexes ? baseTimes(value.length, String) : [], length = result.length;
        for (var key in value) {
            if ((inherited || hasOwnProperty.call(value, key)) &&
                !(skipIndexes && (
                // Safari 9 has enumerable `arguments.length` in strict mode.
                key == 'length' ||
                    // Node.js 0.10 has enumerable non-index properties on buffers.
                    (isBuff && (key == 'offset' || key == 'parent')) ||
                    // PhantomJS 2 has enumerable non-index properties on typed arrays.
                    (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
                    // Skip index properties.
                    isIndex(key, length)))) {
                result.push(key);
            }
        }
        return result;
    }
    /**
     * This function is like `assignValue` except that it doesn't assign
     * `undefined` values.
     *
     * @private
     * @param {Object} object The object to modify.
     * @param {string} key The key of the property to assign.
     * @param {*} value The value to assign.
     */
    function assignMergeValue(object, key, value) {
        if ((value !== undefined && !eq(object[key], value)) ||
            (value === undefined && !(key in object))) {
            baseAssignValue(object, key, value);
        }
    }
    /**
     * Assigns `value` to `key` of `object` if the existing value is not equivalent
     * using [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * for equality comparisons.
     *
     * @private
     * @param {Object} object The object to modify.
     * @param {string} key The key of the property to assign.
     * @param {*} value The value to assign.
     */
    function assignValue(object, key, value) {
        var objValue = object[key];
        if (!(hasOwnProperty.call(object, key) && eq(objValue, value)) ||
            (value === undefined && !(key in object))) {
            baseAssignValue(object, key, value);
        }
    }
    /**
     * Gets the index at which the `key` is found in `array` of key-value pairs.
     *
     * @private
     * @param {Array} array The array to inspect.
     * @param {*} key The key to search for.
     * @returns {number} Returns the index of the matched value, else `-1`.
     */
    function assocIndexOf(array, key) {
        var length = array.length;
        while (length--) {
            if (eq(array[length][0], key)) {
                return length;
            }
        }
        return -1;
    }
    /**
     * The base implementation of `assignValue` and `assignMergeValue` without
     * value checks.
     *
     * @private
     * @param {Object} object The object to modify.
     * @param {string} key The key of the property to assign.
     * @param {*} value The value to assign.
     */
    function baseAssignValue(object, key, value) {
        if (key == '__proto__' && defineProperty) {
            defineProperty(object, key, {
                'configurable': true,
                'enumerable': true,
                'value': value,
                'writable': true
            });
        }
        else {
            object[key] = value;
        }
    }
    /**
     * The base implementation of `baseForOwn` which iterates over `object`
     * properties returned by `keysFunc` and invokes `iteratee` for each property.
     * Iteratee functions may exit iteration early by explicitly returning `false`.
     *
     * @private
     * @param {Object} object The object to iterate over.
     * @param {Function} iteratee The function invoked per iteration.
     * @param {Function} keysFunc The function to get the keys of `object`.
     * @returns {Object} Returns `object`.
     */
    var baseFor = createBaseFor();
    /**
     * The base implementation of `getTag` without fallbacks for buggy environments.
     *
     * @private
     * @param {*} value The value to query.
     * @returns {string} Returns the `toStringTag`.
     */
    function baseGetTag(value) {
        if (value == null) {
            return value === undefined ? undefinedTag : nullTag;
        }
        return (symToStringTag && symToStringTag in Object(value))
            ? getRawTag(value)
            : objectToString(value);
    }
    /**
     * The base implementation of `_.isArguments`.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an `arguments` object,
     */
    function baseIsArguments(value) {
        return isObjectLike(value) && baseGetTag(value) == argsTag;
    }
    /**
     * The base implementation of `_.isNative` without bad shim checks.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a native function,
     *  else `false`.
     */
    function baseIsNative(value) {
        if (!isObject(value) || isMasked(value)) {
            return false;
        }
        var pattern = isFunction(value) ? reIsNative : reIsHostCtor;
        return pattern.test(toSource(value));
    }
    /**
     * The base implementation of `_.isTypedArray` without Node.js optimizations.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
     */
    function baseIsTypedArray(value) {
        return isObjectLike(value) &&
            isLength(value.length) && !!typedArrayTags[baseGetTag(value)];
    }
    /**
     * The base implementation of `_.keysIn` which doesn't treat sparse arrays as dense.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names.
     */
    function baseKeysIn(object) {
        if (!isObject(object)) {
            return nativeKeysIn(object);
        }
        var isProto = isPrototype(object), result = [];
        for (var key in object) {
            if (!(key == 'constructor' && (isProto || !hasOwnProperty.call(object, key)))) {
                result.push(key);
            }
        }
        return result;
    }
    /**
     * The base implementation of `_.merge` without support for multiple sources.
     *
     * @private
     * @param {Object} object The destination object.
     * @param {Object} source The source object.
     * @param {number} srcIndex The index of `source`.
     * @param {Function} [customizer] The function to customize merged values.
     * @param {Object} [stack] Tracks traversed source values and their merged
     *  counterparts.
     */
    function baseMerge(object, source, srcIndex, customizer, stack) {
        if (object === source) {
            return;
        }
        baseFor(source, function (srcValue, key) {
            stack || (stack = new Stack);
            if (isObject(srcValue)) {
                baseMergeDeep(object, source, key, srcIndex, baseMerge, customizer, stack);
            }
            else {
                var newValue = customizer
                    ? customizer(safeGet(object, key), srcValue, (key + ''), object, source, stack)
                    : undefined;
                if (newValue === undefined) {
                    newValue = srcValue;
                }
                assignMergeValue(object, key, newValue);
            }
        }, keysIn);
    }
    /**
     * A specialized version of `baseMerge` for arrays and objects which performs
     * deep merges and tracks traversed objects enabling objects with circular
     * references to be merged.
     *
     * @private
     * @param {Object} object The destination object.
     * @param {Object} source The source object.
     * @param {string} key The key of the value to merge.
     * @param {number} srcIndex The index of `source`.
     * @param {Function} mergeFunc The function to merge values.
     * @param {Function} [customizer] The function to customize assigned values.
     * @param {Object} [stack] Tracks traversed source values and their merged
     *  counterparts.
     */
    function baseMergeDeep(object, source, key, srcIndex, mergeFunc, customizer, stack) {
        var objValue = safeGet(object, key), srcValue = safeGet(source, key), stacked = stack.get(srcValue);
        if (stacked) {
            assignMergeValue(object, key, stacked);
            return;
        }
        var newValue = customizer
            ? customizer(objValue, srcValue, (key + ''), object, source, stack)
            : undefined;
        var isCommon = newValue === undefined;
        if (isCommon) {
            var isArr = isArray(srcValue), isBuff = !isArr && isBuffer(srcValue), isTyped = !isArr && !isBuff && isTypedArray(srcValue);
            newValue = srcValue;
            if (isArr || isBuff || isTyped) {
                if (isArray(objValue)) {
                    newValue = objValue;
                }
                else if (isArrayLikeObject(objValue)) {
                    newValue = copyArray(objValue);
                }
                else if (isBuff) {
                    isCommon = false;
                    newValue = cloneBuffer(srcValue, true);
                }
                else if (isTyped) {
                    isCommon = false;
                    newValue = cloneTypedArray(srcValue, true);
                }
                else {
                    newValue = [];
                }
            }
            else if (isPlainObject(srcValue) || isArguments(srcValue)) {
                newValue = objValue;
                if (isArguments(objValue)) {
                    newValue = toPlainObject(objValue);
                }
                else if (!isObject(objValue) || isFunction(objValue)) {
                    newValue = initCloneObject(srcValue);
                }
            }
            else {
                isCommon = false;
            }
        }
        if (isCommon) {
            // Recursively merge objects and arrays (susceptible to call stack limits).
            stack.set(srcValue, newValue);
            mergeFunc(newValue, srcValue, srcIndex, customizer, stack);
            stack['delete'](srcValue);
        }
        assignMergeValue(object, key, newValue);
    }
    /**
     * The base implementation of `_.rest` which doesn't validate or coerce arguments.
     *
     * @private
     * @param {Function} func The function to apply a rest parameter to.
     * @param {number} [start=func.length-1] The start position of the rest parameter.
     * @returns {Function} Returns the new function.
     */
    function baseRest(func, start) {
        return setToString(overRest(func, start, identity), func + '');
    }
    /**
     * The base implementation of `setToString` without support for hot loop shorting.
     *
     * @private
     * @param {Function} func The function to modify.
     * @param {Function} string The `toString` result.
     * @returns {Function} Returns `func`.
     */
    var baseSetToString = !defineProperty ? identity : function (func, string) {
        return defineProperty(func, 'toString', {
            'configurable': true,
            'enumerable': false,
            'value': constant(string),
            'writable': true
        });
    };
    /**
     * Creates a clone of  `buffer`.
     *
     * @private
     * @param {Buffer} buffer The buffer to clone.
     * @param {boolean} [isDeep] Specify a deep clone.
     * @returns {Buffer} Returns the cloned buffer.
     */
    function cloneBuffer(buffer, isDeep) {
        if (isDeep) {
            return buffer.slice();
        }
        var length = buffer.length, result = allocUnsafe ? allocUnsafe(length) : new buffer.constructor(length);
        buffer.copy(result);
        return result;
    }
    /**
     * Creates a clone of `arrayBuffer`.
     *
     * @private
     * @param {ArrayBuffer} arrayBuffer The array buffer to clone.
     * @returns {ArrayBuffer} Returns the cloned array buffer.
     */
    function cloneArrayBuffer(arrayBuffer) {
        var result = new arrayBuffer.constructor(arrayBuffer.byteLength);
        new Uint8Array(result).set(new Uint8Array(arrayBuffer));
        return result;
    }
    /**
     * Creates a clone of `typedArray`.
     *
     * @private
     * @param {Object} typedArray The typed array to clone.
     * @param {boolean} [isDeep] Specify a deep clone.
     * @returns {Object} Returns the cloned typed array.
     */
    function cloneTypedArray(typedArray, isDeep) {
        var buffer = isDeep ? cloneArrayBuffer(typedArray.buffer) : typedArray.buffer;
        return new typedArray.constructor(buffer, typedArray.byteOffset, typedArray.length);
    }
    /**
     * Copies the values of `source` to `array`.
     *
     * @private
     * @param {Array} source The array to copy values from.
     * @param {Array} [array=[]] The array to copy values to.
     * @returns {Array} Returns `array`.
     */
    function copyArray(source, array) {
        var index = -1, length = source.length;
        array || (array = Array(length));
        while (++index < length) {
            array[index] = source[index];
        }
        return array;
    }
    /**
     * Copies properties of `source` to `object`.
     *
     * @private
     * @param {Object} source The object to copy properties from.
     * @param {Array} props The property identifiers to copy.
     * @param {Object} [object={}] The object to copy properties to.
     * @param {Function} [customizer] The function to customize copied values.
     * @returns {Object} Returns `object`.
     */
    function copyObject(source, props, object, customizer) {
        var isNew = !object;
        object || (object = {});
        var index = -1, length = props.length;
        while (++index < length) {
            var key = props[index];
            var newValue = customizer
                ? customizer(object[key], source[key], key, object, source)
                : undefined;
            if (newValue === undefined) {
                newValue = source[key];
            }
            if (isNew) {
                baseAssignValue(object, key, newValue);
            }
            else {
                assignValue(object, key, newValue);
            }
        }
        return object;
    }
    /**
     * Creates a function like `_.assign`.
     *
     * @private
     * @param {Function} assigner The function to assign values.
     * @returns {Function} Returns the new assigner function.
     */
    function createAssigner(assigner) {
        return baseRest(function (object, sources) {
            var index = -1, length = sources.length, customizer = length > 1 ? sources[length - 1] : undefined, guard = length > 2 ? sources[2] : undefined;
            customizer = (assigner.length > 3 && typeof customizer == 'function')
                ? (length--, customizer)
                : undefined;
            if (guard && isIterateeCall(sources[0], sources[1], guard)) {
                customizer = length < 3 ? undefined : customizer;
                length = 1;
            }
            object = Object(object);
            while (++index < length) {
                var source = sources[index];
                if (source) {
                    assigner(object, source, index, customizer);
                }
            }
            return object;
        });
    }
    /**
     * Creates a base function for methods like `_.forIn` and `_.forOwn`.
     *
     * @private
     * @param {boolean} [fromRight] Specify iterating from right to left.
     * @returns {Function} Returns the new base function.
     */
    function createBaseFor(fromRight) {
        return function (object, iteratee, keysFunc) {
            var index = -1, iterable = Object(object), props = keysFunc(object), length = props.length;
            while (length--) {
                var key = props[fromRight ? length : ++index];
                if (iteratee(iterable[key], key, iterable) === false) {
                    break;
                }
            }
            return object;
        };
    }
    /**
     * Gets the data for `map`.
     *
     * @private
     * @param {Object} map The map to query.
     * @param {string} key The reference key.
     * @returns {*} Returns the map data.
     */
    function getMapData(map, key) {
        var data = map.__data__;
        return isKeyable(key)
            ? data[typeof key == 'string' ? 'string' : 'hash']
            : data.map;
    }
    /**
     * Gets the native function at `key` of `object`.
     *
     * @private
     * @param {Object} object The object to query.
     * @param {string} key The key of the method to get.
     * @returns {*} Returns the function if it's native, else `undefined`.
     */
    function getNative(object, key) {
        var value = getValue(object, key);
        return baseIsNative(value) ? value : undefined;
    }
    /**
     * A specialized version of `baseGetTag` which ignores `Symbol.toStringTag` values.
     *
     * @private
     * @param {*} value The value to query.
     * @returns {string} Returns the raw `toStringTag`.
     */
    function getRawTag(value) {
        var isOwn = hasOwnProperty.call(value, symToStringTag), tag = value[symToStringTag];
        try {
            value[symToStringTag] = undefined;
            var unmasked = true;
        }
        catch (e) { }
        var result = nativeObjectToString.call(value);
        if (unmasked) {
            if (isOwn) {
                value[symToStringTag] = tag;
            }
            else {
                delete value[symToStringTag];
            }
        }
        return result;
    }
    /**
     * Initializes an object clone.
     *
     * @private
     * @param {Object} object The object to clone.
     * @returns {Object} Returns the initialized clone.
     */
    function initCloneObject(object) {
        return (typeof object.constructor == 'function' && !isPrototype(object))
            ? baseCreate(getPrototype(object))
            : {};
    }
    /**
     * Checks if `value` is a valid array-like index.
     *
     * @private
     * @param {*} value The value to check.
     * @param {number} [length=MAX_SAFE_INTEGER] The upper bounds of a valid index.
     * @returns {boolean} Returns `true` if `value` is a valid index, else `false`.
     */
    function isIndex(value, length) {
        var type = typeof value;
        length = length == null ? MAX_SAFE_INTEGER : length;
        return !!length &&
            (type == 'number' ||
                (type != 'symbol' && reIsUint.test(value))) &&
            (value > -1 && value % 1 == 0 && value < length);
    }
    /**
     * Checks if the given arguments are from an iteratee call.
     *
     * @private
     * @param {*} value The potential iteratee value argument.
     * @param {*} index The potential iteratee index or key argument.
     * @param {*} object The potential iteratee object argument.
     * @returns {boolean} Returns `true` if the arguments are from an iteratee call,
     *  else `false`.
     */
    function isIterateeCall(value, index, object) {
        if (!isObject(object)) {
            return false;
        }
        var type = typeof index;
        if (type == 'number'
            ? (isArrayLike(object) && isIndex(index, object.length))
            : (type == 'string' && index in object)) {
            return eq(object[index], value);
        }
        return false;
    }
    /**
     * Checks if `value` is suitable for use as unique object key.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is suitable, else `false`.
     */
    function isKeyable(value) {
        var type = typeof value;
        return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
            ? (value !== '__proto__')
            : (value === null);
    }
    /**
     * Checks if `func` has its source masked.
     *
     * @private
     * @param {Function} func The function to check.
     * @returns {boolean} Returns `true` if `func` is masked, else `false`.
     */
    function isMasked(func) {
        return !!maskSrcKey && (maskSrcKey in func);
    }
    /**
     * Checks if `value` is likely a prototype object.
     *
     * @private
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a prototype, else `false`.
     */
    function isPrototype(value) {
        var Ctor = value && value.constructor, proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto;
        return value === proto;
    }
    /**
     * This function is like
     * [`Object.keys`](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
     * except that it includes inherited enumerable properties.
     *
     * @private
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names.
     */
    function nativeKeysIn(object) {
        var result = [];
        if (object != null) {
            for (var key in Object(object)) {
                result.push(key);
            }
        }
        return result;
    }
    /**
     * Converts `value` to a string using `Object.prototype.toString`.
     *
     * @private
     * @param {*} value The value to convert.
     * @returns {string} Returns the converted string.
     */
    function objectToString(value) {
        return nativeObjectToString.call(value);
    }
    /**
     * A specialized version of `baseRest` which transforms the rest array.
     *
     * @private
     * @param {Function} func The function to apply a rest parameter to.
     * @param {number} [start=func.length-1] The start position of the rest parameter.
     * @param {Function} transform The rest array transform.
     * @returns {Function} Returns the new function.
     */
    function overRest(func, start, transform) {
        start = nativeMax(start === undefined ? (func.length - 1) : start, 0);
        return function () {
            var args = arguments, index = -1, length = nativeMax(args.length - start, 0), array = Array(length);
            while (++index < length) {
                array[index] = args[start + index];
            }
            index = -1;
            var otherArgs = Array(start + 1);
            while (++index < start) {
                otherArgs[index] = args[index];
            }
            otherArgs[start] = transform(array);
            return apply(func, this, otherArgs);
        };
    }
    /**
     * Gets the value at `key`, unless `key` is "__proto__" or "constructor".
     *
     * @private
     * @param {Object} object The object to query.
     * @param {string} key The key of the property to get.
     * @returns {*} Returns the property value.
     */
    function safeGet(object, key) {
        if (key === 'constructor' && typeof object[key] === 'function') {
            return;
        }
        if (key == '__proto__') {
            return;
        }
        return object[key];
    }
    /**
     * Sets the `toString` method of `func` to return `string`.
     *
     * @private
     * @param {Function} func The function to modify.
     * @param {Function} string The `toString` result.
     * @returns {Function} Returns `func`.
     */
    var setToString = shortOut(baseSetToString);
    /**
     * Creates a function that'll short out and invoke `identity` instead
     * of `func` when it's called `HOT_COUNT` or more times in `HOT_SPAN`
     * milliseconds.
     *
     * @private
     * @param {Function} func The function to restrict.
     * @returns {Function} Returns the new shortable function.
     */
    function shortOut(func) {
        var count = 0, lastCalled = 0;
        return function () {
            var stamp = nativeNow(), remaining = HOT_SPAN - (stamp - lastCalled);
            lastCalled = stamp;
            if (remaining > 0) {
                if (++count >= HOT_COUNT) {
                    return arguments[0];
                }
            }
            else {
                count = 0;
            }
            return func.apply(undefined, arguments);
        };
    }
    /**
     * Converts `func` to its source code.
     *
     * @private
     * @param {Function} func The function to convert.
     * @returns {string} Returns the source code.
     */
    function toSource(func) {
        if (func != null) {
            try {
                return funcToString.call(func);
            }
            catch (e) { }
            try {
                return (func + '');
            }
            catch (e) { }
        }
        return '';
    }
    /**
     * Performs a
     * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
     * comparison between two values to determine if they are equivalent.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to compare.
     * @param {*} other The other value to compare.
     * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
     * @example
     *
     * var object = { 'a': 1 };
     * var other = { 'a': 1 };
     *
     * _.eq(object, object);
     * // => true
     *
     * _.eq(object, other);
     * // => false
     *
     * _.eq('a', 'a');
     * // => true
     *
     * _.eq('a', Object('a'));
     * // => false
     *
     * _.eq(NaN, NaN);
     * // => true
     */
    function eq(value, other) {
        return value === other || (value !== value && other !== other);
    }
    /**
     * Checks if `value` is likely an `arguments` object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an `arguments` object,
     *  else `false`.
     * @example
     *
     * _.isArguments(function() { return arguments; }());
     * // => true
     *
     * _.isArguments([1, 2, 3]);
     * // => false
     */
    var isArguments = baseIsArguments(function () { return arguments; }()) ? baseIsArguments : function (value) {
        return isObjectLike(value) && hasOwnProperty.call(value, 'callee') &&
            !propertyIsEnumerable.call(value, 'callee');
    };
    /**
     * Checks if `value` is classified as an `Array` object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an array, else `false`.
     * @example
     *
     * _.isArray([1, 2, 3]);
     * // => true
     *
     * _.isArray(document.body.children);
     * // => false
     *
     * _.isArray('abc');
     * // => false
     *
     * _.isArray(_.noop);
     * // => false
     */
    var isArray = Array.isArray;
    /**
     * Checks if `value` is array-like. A value is considered array-like if it's
     * not a function and has a `value.length` that's an integer greater than or
     * equal to `0` and less than or equal to `Number.MAX_SAFE_INTEGER`.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is array-like, else `false`.
     * @example
     *
     * _.isArrayLike([1, 2, 3]);
     * // => true
     *
     * _.isArrayLike(document.body.children);
     * // => true
     *
     * _.isArrayLike('abc');
     * // => true
     *
     * _.isArrayLike(_.noop);
     * // => false
     */
    function isArrayLike(value) {
        return value != null && isLength(value.length) && !isFunction(value);
    }
    /**
     * This method is like `_.isArrayLike` except that it also checks if `value`
     * is an object.
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an array-like object,
     *  else `false`.
     * @example
     *
     * _.isArrayLikeObject([1, 2, 3]);
     * // => true
     *
     * _.isArrayLikeObject(document.body.children);
     * // => true
     *
     * _.isArrayLikeObject('abc');
     * // => false
     *
     * _.isArrayLikeObject(_.noop);
     * // => false
     */
    function isArrayLikeObject(value) {
        return isObjectLike(value) && isArrayLike(value);
    }
    /**
     * Checks if `value` is a buffer.
     *
     * @static
     * @memberOf _
     * @since 4.3.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a buffer, else `false`.
     * @example
     *
     * _.isBuffer(new Buffer(2));
     * // => true
     *
     * _.isBuffer(new Uint8Array(2));
     * // => false
     */
    var isBuffer = nativeIsBuffer || stubFalse;
    /**
     * Checks if `value` is classified as a `Function` object.
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a function, else `false`.
     * @example
     *
     * _.isFunction(_);
     * // => true
     *
     * _.isFunction(/abc/);
     * // => false
     */
    function isFunction(value) {
        if (!isObject(value)) {
            return false;
        }
        // The use of `Object#toString` avoids issues with the `typeof` operator
        // in Safari 9 which returns 'object' for typed arrays and other constructors.
        var tag = baseGetTag(value);
        return tag == funcTag || tag == genTag || tag == asyncTag || tag == proxyTag;
    }
    /**
     * Checks if `value` is a valid array-like length.
     *
     * **Note:** This method is loosely based on
     * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a valid length, else `false`.
     * @example
     *
     * _.isLength(3);
     * // => true
     *
     * _.isLength(Number.MIN_VALUE);
     * // => false
     *
     * _.isLength(Infinity);
     * // => false
     *
     * _.isLength('3');
     * // => false
     */
    function isLength(value) {
        return typeof value == 'number' &&
            value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER;
    }
    /**
     * Checks if `value` is the
     * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
     * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
     *
     * @static
     * @memberOf _
     * @since 0.1.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is an object, else `false`.
     * @example
     *
     * _.isObject({});
     * // => true
     *
     * _.isObject([1, 2, 3]);
     * // => true
     *
     * _.isObject(_.noop);
     * // => true
     *
     * _.isObject(null);
     * // => false
     */
    function isObject(value) {
        var type = typeof value;
        return value != null && (type == 'object' || type == 'function');
    }
    /**
     * Checks if `value` is object-like. A value is object-like if it's not `null`
     * and has a `typeof` result of "object".
     *
     * @static
     * @memberOf _
     * @since 4.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
     * @example
     *
     * _.isObjectLike({});
     * // => true
     *
     * _.isObjectLike([1, 2, 3]);
     * // => true
     *
     * _.isObjectLike(_.noop);
     * // => false
     *
     * _.isObjectLike(null);
     * // => false
     */
    function isObjectLike(value) {
        return value != null && typeof value == 'object';
    }
    /**
     * Checks if `value` is a plain object, that is, an object created by the
     * `Object` constructor or one with a `[[Prototype]]` of `null`.
     *
     * @static
     * @memberOf _
     * @since 0.8.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a plain object, else `false`.
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     * }
     *
     * _.isPlainObject(new Foo);
     * // => false
     *
     * _.isPlainObject([1, 2, 3]);
     * // => false
     *
     * _.isPlainObject({ 'x': 0, 'y': 0 });
     * // => true
     *
     * _.isPlainObject(Object.create(null));
     * // => true
     */
    function isPlainObject(value) {
        if (!isObjectLike(value) || baseGetTag(value) != objectTag) {
            return false;
        }
        var proto = getPrototype(value);
        if (proto === null) {
            return true;
        }
        var Ctor = hasOwnProperty.call(proto, 'constructor') && proto.constructor;
        return typeof Ctor == 'function' && Ctor instanceof Ctor &&
            funcToString.call(Ctor) == objectCtorString;
    }
    /**
     * Checks if `value` is classified as a typed array.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Lang
     * @param {*} value The value to check.
     * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
     * @example
     *
     * _.isTypedArray(new Uint8Array);
     * // => true
     *
     * _.isTypedArray([]);
     * // => false
     */
    var isTypedArray = nodeIsTypedArray ? baseUnary(nodeIsTypedArray) : baseIsTypedArray;
    /**
     * Converts `value` to a plain object flattening inherited enumerable string
     * keyed properties of `value` to own properties of the plain object.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Lang
     * @param {*} value The value to convert.
     * @returns {Object} Returns the converted plain object.
     * @example
     *
     * function Foo() {
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.assign({ 'a': 1 }, new Foo);
     * // => { 'a': 1, 'b': 2 }
     *
     * _.assign({ 'a': 1 }, _.toPlainObject(new Foo));
     * // => { 'a': 1, 'b': 2, 'c': 3 }
     */
    function toPlainObject(value) {
        return copyObject(value, keysIn(value));
    }
    /**
     * Creates an array of the own and inherited enumerable property names of `object`.
     *
     * **Note:** Non-object values are coerced to objects.
     *
     * @static
     * @memberOf _
     * @since 3.0.0
     * @category Object
     * @param {Object} object The object to query.
     * @returns {Array} Returns the array of property names.
     * @example
     *
     * function Foo() {
     *   this.a = 1;
     *   this.b = 2;
     * }
     *
     * Foo.prototype.c = 3;
     *
     * _.keysIn(new Foo);
     * // => ['a', 'b', 'c'] (iteration order is not guaranteed)
     */
    function keysIn(object) {
        return isArrayLike(object) ? arrayLikeKeys(object, true) : baseKeysIn(object);
    }
    /**
     * This method is like `_.assign` except that it recursively merges own and
     * inherited enumerable string keyed properties of source objects into the
     * destination object. Source properties that resolve to `undefined` are
     * skipped if a destination value exists. Array and plain object properties
     * are merged recursively. Other objects and value types are overridden by
     * assignment. Source objects are applied from left to right. Subsequent
     * sources overwrite property assignments of previous sources.
     *
     * **Note:** This method mutates `object`.
     *
     * @static
     * @memberOf _
     * @since 0.5.0
     * @category Object
     * @param {Object} object The destination object.
     * @param {...Object} [sources] The source objects.
     * @returns {Object} Returns `object`.
     * @example
     *
     * var object = {
     *   'a': [{ 'b': 2 }, { 'd': 4 }]
     * };
     *
     * var other = {
     *   'a': [{ 'c': 3 }, { 'e': 5 }]
     * };
     *
     * _.merge(object, other);
     * // => { 'a': [{ 'b': 2, 'c': 3 }, { 'd': 4, 'e': 5 }] }
     */
    var merge = createAssigner(function (object, source, srcIndex) {
        baseMerge(object, source, srcIndex);
    });
    /**
     * Creates a function that returns `value`.
     *
     * @static
     * @memberOf _
     * @since 2.4.0
     * @category Util
     * @param {*} value The value to return from the new function.
     * @returns {Function} Returns the new constant function.
     * @example
     *
     * var objects = _.times(2, _.constant({ 'a': 1 }));
     *
     * console.log(objects);
     * // => [{ 'a': 1 }, { 'a': 1 }]
     *
     * console.log(objects[0] === objects[1]);
     * // => true
     */
    function constant(value) {
        return function () {
            return value;
        };
    }
    /**
     * This method returns the first argument it receives.
     *
     * @static
     * @since 0.1.0
     * @memberOf _
     * @category Util
     * @param {*} value Any value.
     * @returns {*} Returns `value`.
     * @example
     *
     * var object = { 'a': 1 };
     *
     * console.log(_.identity(object) === object);
     * // => true
     */
    function identity(value) {
        return value;
    }
    /**
     * This method returns `false`.
     *
     * @static
     * @memberOf _
     * @since 4.13.0
     * @category Util
     * @returns {boolean} Returns `false`.
     * @example
     *
     * _.times(2, _.stubFalse);
     * // => [false, false]
     */
    function stubFalse() {
        return false;
    }
    module.exports = merge;
},
/* eslint/lib/shared/ast-utils.js */ function _(require, module, exports, __esModule, __esExport) {
    const breakableTypePattern = /^(?:(?:Do)?While|For(?:In|Of)?|Switch)Statement$/u;
    const lineBreakPattern = /\r\n|[\r\n\u2028\u2029]/u;
    const shebangPattern = /^#!([^\r\n]+)/u;
    /**
     * Creates a version of the `lineBreakPattern` regex with the global flag.
     * Global regexes are mutable, so this needs to be a function instead of a constant.
     * @returns {RegExp} A global regular expression that matches line terminators
     */
    function createGlobalLinebreakMatcher() {
        return new RegExp(lineBreakPattern.source, "gu");
    }
    module.exports = {
        breakableTypePattern,
        lineBreakPattern,
        createGlobalLinebreakMatcher,
        shebangPattern
    };
},
/* eslint/lib/shared/traverser.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const vk = require(321) /* eslint-visitor-keys */;
    const debug = require(72) /* debug */("eslint:traverser");
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Do nothing.
     * @returns {void}
     */
    function noop() {
        // do nothing.
    }
    /**
     * Check whether the given value is an ASTNode or not.
     * @param {any} x The value to check.
     * @returns {boolean} `true` if the value is an ASTNode.
     */
    function isNode(x) {
        return x !== null && typeof x === "object" && typeof x.type === "string";
    }
    /**
     * Get the visitor keys of a given node.
     * @param {Object} visitorKeys The map of visitor keys.
     * @param {ASTNode} node The node to get their visitor keys.
     * @returns {string[]} The visitor keys of the node.
     */
    function getVisitorKeys(visitorKeys, node) {
        let keys = visitorKeys[node.type];
        if (!keys) {
            keys = vk.getKeys(node);
            debug("Unknown node type \"%s\": Estimated visitor keys %j", node.type, keys);
        }
        return keys;
    }
    /**
     * The traverser class to traverse AST trees.
     */
    class Traverser {
        constructor() {
            this._current = null;
            this._parents = [];
            this._skipped = false;
            this._broken = false;
            this._visitorKeys = null;
            this._enter = null;
            this._leave = null;
        }
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @returns {ASTNode} The current node.
         */
        current() {
            return this._current;
        }
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @returns {ASTNode[]} The ancestor nodes.
         */
        parents() {
            return this._parents.slice(0);
        }
        /**
         * Break the current traversal.
         * @returns {void}
         */
        break() {
            this._broken = true;
        }
        /**
         * Skip child nodes for the current traversal.
         * @returns {void}
         */
        skip() {
            this._skipped = true;
        }
        /**
         * Traverse the given AST tree.
         * @param {ASTNode} node The root node to traverse.
         * @param {Object} options The option object.
         * @param {Object} [options.visitorKeys=DEFAULT_VISITOR_KEYS] The keys of each node types to traverse child nodes. Default is `./default-visitor-keys.json`.
         * @param {Function} [options.enter=noop] The callback function which is called on entering each node.
         * @param {Function} [options.leave=noop] The callback function which is called on leaving each node.
         * @returns {void}
         */
        traverse(node, options) {
            this._current = null;
            this._parents = [];
            this._skipped = false;
            this._broken = false;
            this._visitorKeys = options.visitorKeys || vk.KEYS;
            this._enter = options.enter || noop;
            this._leave = options.leave || noop;
            this._traverse(node, null);
        }
        /**
         * Traverse the given AST tree recursively.
         * @param {ASTNode} node The current node.
         * @param {ASTNode|null} parent The parent node.
         * @returns {void}
         * @private
         */
        _traverse(node, parent) {
            if (!isNode(node)) {
                return;
            }
            this._current = node;
            this._skipped = false;
            this._enter(node, parent);
            if (!this._skipped && !this._broken) {
                const keys = getVisitorKeys(this._visitorKeys, node);
                if (keys.length >= 1) {
                    this._parents.push(node);
                    for (let i = 0; i < keys.length && !this._broken; ++i) {
                        const child = node[keys[i]];
                        if (Array.isArray(child)) {
                            for (let j = 0; j < child.length && !this._broken; ++j) {
                                this._traverse(child[j], node);
                            }
                        }
                        else {
                            this._traverse(child, node);
                        }
                    }
                    this._parents.pop();
                }
            }
            if (!this._broken) {
                this._leave(node, parent);
            }
            this._current = parent;
        }
        /**
         * Calculates the keys to use for traversal.
         * @param {ASTNode} node The node to read keys from.
         * @returns {string[]} An array of keys to visit on the node.
         * @private
         */
        static getKeys(node) {
            return vk.getKeys(node);
        }
        /**
         * Traverse the given AST tree.
         * @param {ASTNode} node The root node to traverse.
         * @param {Object} options The option object.
         * @param {Object} [options.visitorKeys=DEFAULT_VISITOR_KEYS] The keys of each node types to traverse child nodes. Default is `./default-visitor-keys.json`.
         * @param {Function} [options.enter=noop] The callback function which is called on entering each node.
         * @param {Function} [options.leave=noop] The callback function which is called on leaving each node.
         * @returns {void}
         */
        static traverse(node, options) {
            new Traverser().traverse(node, options);
        }
        /**
         * The default visitor keys.
         * @type {Object}
         */
        static get DEFAULT_VISITOR_KEYS() {
            return vk.KEYS;
        }
    }
    module.exports = Traverser;
},
/* eslint/lib/source-code/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        SourceCode: require(349) /* ./source-code */
    };
},
/* eslint/lib/source-code/source-code.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { isCommentToken } = require(350) /* eslint-utils */, TokenStore = require(351) /* ./token-store */, astUtils = require(346) /* ../shared/ast-utils */, Traverser = require(347) /* ../shared/traverser */;
    //------------------------------------------------------------------------------
    // Private
    //------------------------------------------------------------------------------
    /**
     * Validates that the given AST has the required information.
     * @param {ASTNode} ast The Program node of the AST to check.
     * @throws {Error} If the AST doesn't contain the correct information.
     * @returns {void}
     * @private
     */
    function validate(ast) {
        if (!ast.tokens) {
            throw new Error("AST is missing the tokens array.");
        }
        if (!ast.comments) {
            throw new Error("AST is missing the comments array.");
        }
        if (!ast.loc) {
            throw new Error("AST is missing location information.");
        }
        if (!ast.range) {
            throw new Error("AST is missing range information");
        }
    }
    /**
     * Check to see if its a ES6 export declaration.
     * @param {ASTNode} astNode An AST node.
     * @returns {boolean} whether the given node represents an export declaration.
     * @private
     */
    function looksLikeExport(astNode) {
        return astNode.type === "ExportDefaultDeclaration" || astNode.type === "ExportNamedDeclaration" ||
            astNode.type === "ExportAllDeclaration" || astNode.type === "ExportSpecifier";
    }
    /**
     * Merges two sorted lists into a larger sorted list in O(n) time.
     * @param {Token[]} tokens The list of tokens.
     * @param {Token[]} comments The list of comments.
     * @returns {Token[]} A sorted list of tokens and comments.
     * @private
     */
    function sortedMerge(tokens, comments) {
        const result = [];
        let tokenIndex = 0;
        let commentIndex = 0;
        while (tokenIndex < tokens.length || commentIndex < comments.length) {
            if (commentIndex >= comments.length || tokenIndex < tokens.length && tokens[tokenIndex].range[0] < comments[commentIndex].range[0]) {
                result.push(tokens[tokenIndex++]);
            }
            else {
                result.push(comments[commentIndex++]);
            }
        }
        return result;
    }
    /**
     * Determines if two nodes or tokens overlap.
     * @param {ASTNode|Token} first The first node or token to check.
     * @param {ASTNode|Token} second The second node or token to check.
     * @returns {boolean} True if the two nodes or tokens overlap.
     * @private
     */
    function nodesOrTokensOverlap(first, second) {
        return (first.range[0] <= second.range[0] && first.range[1] >= second.range[0]) ||
            (second.range[0] <= first.range[0] && second.range[1] >= first.range[0]);
    }
    /**
     * Determines if two nodes or tokens have at least one whitespace character
     * between them. Order does not matter. Returns false if the given nodes or
     * tokens overlap.
     * @param {SourceCode} sourceCode The source code object.
     * @param {ASTNode|Token} first The first node or token to check between.
     * @param {ASTNode|Token} second The second node or token to check between.
     * @param {boolean} checkInsideOfJSXText If `true` is present, check inside of JSXText tokens for backward compatibility.
     * @returns {boolean} True if there is a whitespace character between
     * any of the tokens found between the two given nodes or tokens.
     * @public
     */
    function isSpaceBetween(sourceCode, first, second, checkInsideOfJSXText) {
        if (nodesOrTokensOverlap(first, second)) {
            return false;
        }
        const [startingNodeOrToken, endingNodeOrToken] = first.range[1] <= second.range[0]
            ? [first, second]
            : [second, first];
        const firstToken = sourceCode.getLastToken(startingNodeOrToken) || startingNodeOrToken;
        const finalToken = sourceCode.getFirstToken(endingNodeOrToken) || endingNodeOrToken;
        let currentToken = firstToken;
        while (currentToken !== finalToken) {
            const nextToken = sourceCode.getTokenAfter(currentToken, { includeComments: true });
            if (currentToken.range[1] !== nextToken.range[0] ||
                /*
                 * For backward compatibility, check spaces in JSXText.
                 * https://github.com/eslint/eslint/issues/12614
                 */
                (checkInsideOfJSXText &&
                    nextToken !== finalToken &&
                    nextToken.type === "JSXText" &&
                    /\s/u.test(nextToken.value))) {
                return true;
            }
            currentToken = nextToken;
        }
        return false;
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    class SourceCode extends TokenStore {
        /**
         * Represents parsed source code.
         * @param {string|Object} textOrConfig The source code text or config object.
         * @param {string} textOrConfig.text The source code text.
         * @param {ASTNode} textOrConfig.ast The Program node of the AST representing the code. This AST should be created from the text that BOM was stripped.
         * @param {Object|null} textOrConfig.parserServices The parser services.
         * @param {ScopeManager|null} textOrConfig.scopeManager The scope of this source code.
         * @param {Object|null} textOrConfig.visitorKeys The visitor keys to traverse AST.
         * @param {ASTNode} [astIfNoConfig] The Program node of the AST representing the code. This AST should be created from the text that BOM was stripped.
         */
        constructor(textOrConfig, astIfNoConfig) {
            let text, ast, parserServices, scopeManager, visitorKeys;
            // Process overloading.
            if (typeof textOrConfig === "string") {
                text = textOrConfig;
                ast = astIfNoConfig;
            }
            else if (typeof textOrConfig === "object" && textOrConfig !== null) {
                text = textOrConfig.text;
                ast = textOrConfig.ast;
                parserServices = textOrConfig.parserServices;
                scopeManager = textOrConfig.scopeManager;
                visitorKeys = textOrConfig.visitorKeys;
            }
            validate(ast);
            super(ast.tokens, ast.comments);
            /**
             * The flag to indicate that the source code has Unicode BOM.
             * @type boolean
             */
            this.hasBOM = (text.charCodeAt(0) === 0xFEFF);
            /**
             * The original text source code.
             * BOM was stripped from this text.
             * @type string
             */
            this.text = (this.hasBOM ? text.slice(1) : text);
            /**
             * The parsed AST for the source code.
             * @type ASTNode
             */
            this.ast = ast;
            /**
             * The parser services of this source code.
             * @type {Object}
             */
            this.parserServices = parserServices || {};
            /**
             * The scope of this source code.
             * @type {ScopeManager|null}
             */
            this.scopeManager = scopeManager || null;
            /**
             * The visitor keys to traverse AST.
             * @type {Object}
             */
            this.visitorKeys = visitorKeys || Traverser.DEFAULT_VISITOR_KEYS;
            // Check the source text for the presence of a shebang since it is parsed as a standard line comment.
            const shebangMatched = this.text.match(astUtils.shebangPattern);
            const hasShebang = shebangMatched && ast.comments.length && ast.comments[0].value === shebangMatched[1];
            if (hasShebang) {
                ast.comments[0].type = "Shebang";
            }
            this.tokensAndComments = sortedMerge(ast.tokens, ast.comments);
            /**
             * The source code split into lines according to ECMA-262 specification.
             * This is done to avoid each rule needing to do so separately.
             * @type string[]
             */
            this.lines = [];
            this.lineStartIndices = [0];
            const lineEndingPattern = astUtils.createGlobalLinebreakMatcher();
            let match;
            /*
             * Previously, this was implemented using a regex that
             * matched a sequence of non-linebreak characters followed by a
             * linebreak, then adding the lengths of the matches. However,
             * this caused a catastrophic backtracking issue when the end
             * of a file contained a large number of non-newline characters.
             * To avoid this, the current implementation just matches newlines
             * and uses match.index to get the correct line start indices.
             */
            while ((match = lineEndingPattern.exec(this.text))) {
                this.lines.push(this.text.slice(this.lineStartIndices[this.lineStartIndices.length - 1], match.index));
                this.lineStartIndices.push(match.index + match[0].length);
            }
            this.lines.push(this.text.slice(this.lineStartIndices[this.lineStartIndices.length - 1]));
            // Cache for comments found using getComments().
            this._commentCache = new WeakMap();
            // don't allow modification of this object
            Object.freeze(this);
            Object.freeze(this.lines);
        }
        /**
         * Split the source code into multiple lines based on the line delimiters.
         * @param {string} text Source code as a string.
         * @returns {string[]} Array of source code lines.
         * @public
         */
        static splitLines(text) {
            return text.split(astUtils.createGlobalLinebreakMatcher());
        }
        /**
         * Gets the source code for the given node.
         * @param {ASTNode} [node] The AST node to get the text for.
         * @param {int} [beforeCount] The number of characters before the node to retrieve.
         * @param {int} [afterCount] The number of characters after the node to retrieve.
         * @returns {string} The text representing the AST node.
         * @public
         */
        getText(node, beforeCount, afterCount) {
            if (node) {
                return this.text.slice(Math.max(node.range[0] - (beforeCount || 0), 0), node.range[1] + (afterCount || 0));
            }
            return this.text;
        }
        /**
         * Gets the entire source text split into an array of lines.
         * @returns {Array} The source text as an array of lines.
         * @public
         */
        getLines() {
            return this.lines;
        }
        /**
         * Retrieves an array containing all comments in the source code.
         * @returns {ASTNode[]} An array of comment nodes.
         * @public
         */
        getAllComments() {
            return this.ast.comments;
        }
        /**
         * Gets all comments for the given node.
         * @param {ASTNode} node The AST node to get the comments for.
         * @returns {Object} An object containing a leading and trailing array
         *      of comments indexed by their position.
         * @public
         * @deprecated replaced by getCommentsBefore(), getCommentsAfter(), and getCommentsInside().
         */
        getComments(node) {
            if (this._commentCache.has(node)) {
                return this._commentCache.get(node);
            }
            const comments = {
                leading: [],
                trailing: []
            };
            /*
             * Return all comments as leading comments of the Program node when
             * there is no executable code.
             */
            if (node.type === "Program") {
                if (node.body.length === 0) {
                    comments.leading = node.comments;
                }
            }
            else {
                /*
                 * Return comments as trailing comments of nodes that only contain
                 * comments (to mimic the comment attachment behavior present in Espree).
                 */
                if ((node.type === "BlockStatement" || node.type === "ClassBody") && node.body.length === 0 ||
                    node.type === "ObjectExpression" && node.properties.length === 0 ||
                    node.type === "ArrayExpression" && node.elements.length === 0 ||
                    node.type === "SwitchStatement" && node.cases.length === 0) {
                    comments.trailing = this.getTokens(node, {
                        includeComments: true,
                        filter: isCommentToken
                    });
                }
                /*
                 * Iterate over tokens before and after node and collect comment tokens.
                 * Do not include comments that exist outside of the parent node
                 * to avoid duplication.
                 */
                let currentToken = this.getTokenBefore(node, { includeComments: true });
                while (currentToken && isCommentToken(currentToken)) {
                    if (node.parent && node.parent.type !== "Program" && (currentToken.start < node.parent.start)) {
                        break;
                    }
                    comments.leading.push(currentToken);
                    currentToken = this.getTokenBefore(currentToken, { includeComments: true });
                }
                comments.leading.reverse();
                currentToken = this.getTokenAfter(node, { includeComments: true });
                while (currentToken && isCommentToken(currentToken)) {
                    if (node.parent && node.parent.type !== "Program" && (currentToken.end > node.parent.end)) {
                        break;
                    }
                    comments.trailing.push(currentToken);
                    currentToken = this.getTokenAfter(currentToken, { includeComments: true });
                }
            }
            this._commentCache.set(node, comments);
            return comments;
        }
        /**
         * Retrieves the JSDoc comment for a given node.
         * @param {ASTNode} node The AST node to get the comment for.
         * @returns {Token|null} The Block comment token containing the JSDoc comment
         *      for the given node or null if not found.
         * @public
         * @deprecated
         */
        getJSDocComment(node) {
            /**
             * Checks for the presence of a JSDoc comment for the given node and returns it.
             * @param {ASTNode} astNode The AST node to get the comment for.
             * @returns {Token|null} The Block comment token containing the JSDoc comment
             *      for the given node or null if not found.
             * @private
             */
            const findJSDocComment = astNode => {
                const tokenBefore = this.getTokenBefore(astNode, { includeComments: true });
                if (tokenBefore &&
                    isCommentToken(tokenBefore) &&
                    tokenBefore.type === "Block" &&
                    tokenBefore.value.charAt(0) === "*" &&
                    astNode.loc.start.line - tokenBefore.loc.end.line <= 1) {
                    return tokenBefore;
                }
                return null;
            };
            let parent = node.parent;
            switch (node.type) {
                case "ClassDeclaration":
                case "FunctionDeclaration":
                    return findJSDocComment(looksLikeExport(parent) ? parent : node);
                case "ClassExpression":
                    return findJSDocComment(parent.parent);
                case "ArrowFunctionExpression":
                case "FunctionExpression":
                    if (parent.type !== "CallExpression" && parent.type !== "NewExpression") {
                        while (!this.getCommentsBefore(parent).length &&
                            !/Function/u.test(parent.type) &&
                            parent.type !== "MethodDefinition" &&
                            parent.type !== "Property") {
                            parent = parent.parent;
                            if (!parent) {
                                break;
                            }
                        }
                        if (parent && parent.type !== "FunctionDeclaration" && parent.type !== "Program") {
                            return findJSDocComment(parent);
                        }
                    }
                    return findJSDocComment(node);
                // falls through
                default:
                    return null;
            }
        }
        /**
         * Gets the deepest node containing a range index.
         * @param {int} index Range index of the desired node.
         * @returns {ASTNode} The node if found or null if not found.
         * @public
         */
        getNodeByRangeIndex(index) {
            let result = null;
            Traverser.traverse(this.ast, {
                visitorKeys: this.visitorKeys,
                enter(node) {
                    if (node.range[0] <= index && index < node.range[1]) {
                        result = node;
                    }
                    else {
                        this.skip();
                    }
                },
                leave(node) {
                    if (node === result) {
                        this.break();
                    }
                }
            });
            return result;
        }
        /**
         * Determines if two nodes or tokens have at least one whitespace character
         * between them. Order does not matter. Returns false if the given nodes or
         * tokens overlap.
         * @param {ASTNode|Token} first The first node or token to check between.
         * @param {ASTNode|Token} second The second node or token to check between.
         * @returns {boolean} True if there is a whitespace character between
         * any of the tokens found between the two given nodes or tokens.
         * @public
         */
        isSpaceBetween(first, second) {
            return isSpaceBetween(this, first, second, false);
        }
        /**
         * Determines if two nodes or tokens have at least one whitespace character
         * between them. Order does not matter. Returns false if the given nodes or
         * tokens overlap.
         * For backward compatibility, this method returns true if there are
         * `JSXText` tokens that contain whitespaces between the two.
         * @param {ASTNode|Token} first The first node or token to check between.
         * @param {ASTNode|Token} second The second node or token to check between.
         * @returns {boolean} True if there is a whitespace character between
         * any of the tokens found between the two given nodes or tokens.
         * @deprecated in favor of isSpaceBetween().
         * @public
         */
        isSpaceBetweenTokens(first, second) {
            return isSpaceBetween(this, first, second, true);
        }
        /**
         * Converts a source text index into a (line, column) pair.
         * @param {number} index The index of a character in a file
         * @returns {Object} A {line, column} location object with a 0-indexed column
         * @public
         */
        getLocFromIndex(index) {
            if (typeof index !== "number") {
                throw new TypeError("Expected `index` to be a number.");
            }
            if (index < 0 || index > this.text.length) {
                throw new RangeError(`Index out of range (requested index ${index}, but source text has length ${this.text.length}).`);
            }
            /*
             * For an argument of this.text.length, return the location one "spot" past the last character
             * of the file. If the last character is a linebreak, the location will be column 0 of the next
             * line; otherwise, the location will be in the next column on the same line.
             *
             * See getIndexFromLoc for the motivation for this special case.
             */
            if (index === this.text.length) {
                return { line: this.lines.length, column: this.lines[this.lines.length - 1].length };
            }
            /*
             * To figure out which line index is on, determine the last place at which index could
             * be inserted into lineStartIndices to keep the list sorted.
             */
            const lineNumber = index >= this.lineStartIndices[this.lineStartIndices.length - 1]
                ? this.lineStartIndices.length
                : this.lineStartIndices.findIndex(el => index < el);
            return { line: lineNumber, column: index - this.lineStartIndices[lineNumber - 1] };
        }
        /**
         * Converts a (line, column) pair into a range index.
         * @param {Object} loc A line/column location
         * @param {number} loc.line The line number of the location (1-indexed)
         * @param {number} loc.column The column number of the location (0-indexed)
         * @returns {number} The range index of the location in the file.
         * @public
         */
        getIndexFromLoc(loc) {
            if (typeof loc !== "object" || typeof loc.line !== "number" || typeof loc.column !== "number") {
                throw new TypeError("Expected `loc` to be an object with numeric `line` and `column` properties.");
            }
            if (loc.line <= 0) {
                throw new RangeError(`Line number out of range (line ${loc.line} requested). Line numbers should be 1-based.`);
            }
            if (loc.line > this.lineStartIndices.length) {
                throw new RangeError(`Line number out of range (line ${loc.line} requested, but only ${this.lineStartIndices.length} lines present).`);
            }
            const lineStartIndex = this.lineStartIndices[loc.line - 1];
            const lineEndIndex = loc.line === this.lineStartIndices.length ? this.text.length : this.lineStartIndices[loc.line];
            const positionIndex = lineStartIndex + loc.column;
            /*
             * By design, getIndexFromLoc({ line: lineNum, column: 0 }) should return the start index of
             * the given line, provided that the line number is valid element of this.lines. Since the
             * last element of this.lines is an empty string for files with trailing newlines, add a
             * special case where getting the index for the first location after the end of the file
             * will return the length of the file, rather than throwing an error. This allows rules to
             * use getIndexFromLoc consistently without worrying about edge cases at the end of a file.
             */
            if (loc.line === this.lineStartIndices.length && positionIndex > lineEndIndex ||
                loc.line < this.lineStartIndices.length && positionIndex >= lineEndIndex) {
                throw new RangeError(`Column number out of range (column ${loc.column} requested, but the length of line ${loc.line} is ${lineEndIndex - lineStartIndex}).`);
            }
            return positionIndex;
        }
    }
    module.exports = SourceCode;
},
/* eslint-utils/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function _interopDefault(ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }
    var evk = _interopDefault(require(321) /* eslint-visitor-keys */);
    /**
     * Get the innermost scope which contains a given location.
     * @param {Scope} initialScope The initial scope to search.
     * @param {Node} node The location to search.
     * @returns {Scope} The innermost scope.
     */
    function getInnermostScope(initialScope, node) {
        const location = node.range[0];
        let scope = initialScope;
        let found = false;
        do {
            found = false;
            for (const childScope of scope.childScopes) {
                const range = childScope.block.range;
                if (range[0] <= location && location < range[1]) {
                    scope = childScope;
                    found = true;
                    break;
                }
            }
        } while (found);
        return scope;
    }
    /**
     * Find the variable of a given name.
     * @param {Scope} initialScope The scope to start finding.
     * @param {string|Node} nameOrNode The variable name to find. If this is a Node object then it should be an Identifier node.
     * @returns {Variable|null} The found variable or null.
     */
    function findVariable(initialScope, nameOrNode) {
        let name = "";
        let scope = initialScope;
        if (typeof nameOrNode === "string") {
            name = nameOrNode;
        }
        else {
            name = nameOrNode.name;
            scope = getInnermostScope(scope, nameOrNode);
        }
        while (scope != null) {
            const variable = scope.set.get(name);
            if (variable != null) {
                return variable;
            }
            scope = scope.upper;
        }
        return null;
    }
    /**
     * Negate the result of `this` calling.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the result of `this(token)` is `false`.
     */
    function negate0(token) {
        return !this(token); //eslint-disable-line no-invalid-this
    }
    /**
     * Creates the negate function of the given function.
     * @param {function(Token):boolean} f - The function to negate.
     * @returns {function(Token):boolean} Negated function.
     */
    function negate(f) {
        return negate0.bind(f);
    }
    /**
     * Checks if the given token is a PunctuatorToken with the given value
     * @param {Token} token - The token to check.
     * @param {string} value - The value to check.
     * @returns {boolean} `true` if the token is a PunctuatorToken with the given value.
     */
    function isPunctuatorTokenWithValue(token, value) {
        return token.type === "Punctuator" && token.value === value;
    }
    /**
     * Checks if the given token is an arrow token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is an arrow token.
     */
    function isArrowToken(token) {
        return isPunctuatorTokenWithValue(token, "=>");
    }
    /**
     * Checks if the given token is a comma token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is a comma token.
     */
    function isCommaToken(token) {
        return isPunctuatorTokenWithValue(token, ",");
    }
    /**
     * Checks if the given token is a semicolon token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is a semicolon token.
     */
    function isSemicolonToken(token) {
        return isPunctuatorTokenWithValue(token, ";");
    }
    /**
     * Checks if the given token is a colon token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is a colon token.
     */
    function isColonToken(token) {
        return isPunctuatorTokenWithValue(token, ":");
    }
    /**
     * Checks if the given token is an opening parenthesis token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is an opening parenthesis token.
     */
    function isOpeningParenToken(token) {
        return isPunctuatorTokenWithValue(token, "(");
    }
    /**
     * Checks if the given token is a closing parenthesis token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is a closing parenthesis token.
     */
    function isClosingParenToken(token) {
        return isPunctuatorTokenWithValue(token, ")");
    }
    /**
     * Checks if the given token is an opening square bracket token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is an opening square bracket token.
     */
    function isOpeningBracketToken(token) {
        return isPunctuatorTokenWithValue(token, "[");
    }
    /**
     * Checks if the given token is a closing square bracket token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is a closing square bracket token.
     */
    function isClosingBracketToken(token) {
        return isPunctuatorTokenWithValue(token, "]");
    }
    /**
     * Checks if the given token is an opening brace token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is an opening brace token.
     */
    function isOpeningBraceToken(token) {
        return isPunctuatorTokenWithValue(token, "{");
    }
    /**
     * Checks if the given token is a closing brace token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is a closing brace token.
     */
    function isClosingBraceToken(token) {
        return isPunctuatorTokenWithValue(token, "}");
    }
    /**
     * Checks if the given token is a comment token or not.
     * @param {Token} token - The token to check.
     * @returns {boolean} `true` if the token is a comment token.
     */
    function isCommentToken(token) {
        return ["Block", "Line", "Shebang"].includes(token.type);
    }
    const isNotArrowToken = negate(isArrowToken);
    const isNotCommaToken = negate(isCommaToken);
    const isNotSemicolonToken = negate(isSemicolonToken);
    const isNotColonToken = negate(isColonToken);
    const isNotOpeningParenToken = negate(isOpeningParenToken);
    const isNotClosingParenToken = negate(isClosingParenToken);
    const isNotOpeningBracketToken = negate(isOpeningBracketToken);
    const isNotClosingBracketToken = negate(isClosingBracketToken);
    const isNotOpeningBraceToken = negate(isOpeningBraceToken);
    const isNotClosingBraceToken = negate(isClosingBraceToken);
    const isNotCommentToken = negate(isCommentToken);
    /**
     * Get the `(` token of the given function node.
     * @param {Node} node - The function node to get.
     * @param {SourceCode} sourceCode - The source code object to get tokens.
     * @returns {Token} `(` token.
     */
    function getOpeningParenOfParams(node, sourceCode) {
        return node.id
            ? sourceCode.getTokenAfter(node.id, isOpeningParenToken)
            : sourceCode.getFirstToken(node, isOpeningParenToken);
    }
    /**
     * Get the location of the given function node for reporting.
     * @param {Node} node - The function node to get.
     * @param {SourceCode} sourceCode - The source code object to get tokens.
     * @returns {string} The location of the function node for reporting.
     */
    function getFunctionHeadLocation(node, sourceCode) {
        const parent = node.parent;
        let start = null;
        let end = null;
        if (node.type === "ArrowFunctionExpression") {
            const arrowToken = sourceCode.getTokenBefore(node.body, isArrowToken);
            start = arrowToken.loc.start;
            end = arrowToken.loc.end;
        }
        else if (parent.type === "Property" ||
            parent.type === "MethodDefinition" ||
            parent.type === "PropertyDefinition") {
            start = parent.loc.start;
            end = getOpeningParenOfParams(node, sourceCode).loc.start;
        }
        else {
            start = node.loc.start;
            end = getOpeningParenOfParams(node, sourceCode).loc.start;
        }
        return {
            start: { ...start },
            end: { ...end },
        };
    }
    /* globals globalThis, global, self, window */
    const globalObject = typeof globalThis !== "undefined"
        ? globalThis
        : typeof self !== "undefined"
            ? self
            : typeof window !== "undefined"
                ? window
                : typeof global !== "undefined"
                    ? global
                    : {};
    const builtinNames = Object.freeze(new Set([
        "Array",
        "ArrayBuffer",
        "BigInt",
        "BigInt64Array",
        "BigUint64Array",
        "Boolean",
        "DataView",
        "Date",
        "decodeURI",
        "decodeURIComponent",
        "encodeURI",
        "encodeURIComponent",
        "escape",
        "Float32Array",
        "Float64Array",
        "Function",
        "Infinity",
        "Int16Array",
        "Int32Array",
        "Int8Array",
        "isFinite",
        "isNaN",
        "isPrototypeOf",
        "JSON",
        "Map",
        "Math",
        "NaN",
        "Number",
        "Object",
        "parseFloat",
        "parseInt",
        "Promise",
        "Proxy",
        "Reflect",
        "RegExp",
        "Set",
        "String",
        "Symbol",
        "Uint16Array",
        "Uint32Array",
        "Uint8Array",
        "Uint8ClampedArray",
        "undefined",
        "unescape",
        "WeakMap",
        "WeakSet",
    ]));
    const callAllowed = new Set([
        Array.isArray,
        typeof BigInt === "function" ? BigInt : undefined,
        Boolean,
        Date,
        Date.parse,
        decodeURI,
        decodeURIComponent,
        encodeURI,
        encodeURIComponent,
        escape,
        isFinite,
        isNaN,
        isPrototypeOf,
        ...Object.getOwnPropertyNames(Math)
            .map((k) => Math[k])
            .filter((f) => typeof f === "function"),
        Number,
        Number.isFinite,
        Number.isNaN,
        Number.parseFloat,
        Number.parseInt,
        Object,
        Object.entries,
        Object.is,
        Object.isExtensible,
        Object.isFrozen,
        Object.isSealed,
        Object.keys,
        Object.values,
        parseFloat,
        parseInt,
        RegExp,
        String,
        String.fromCharCode,
        String.fromCodePoint,
        String.raw,
        Symbol.for,
        Symbol.keyFor,
        unescape,
    ].filter((f) => typeof f === "function"));
    const callPassThrough = new Set([
        Object.freeze,
        Object.preventExtensions,
        Object.seal,
    ]);
    /**
     * Get the property descriptor.
     * @param {object} object The object to get.
     * @param {string|number|symbol} name The property name to get.
     */
    function getPropertyDescriptor(object, name) {
        let x = object;
        while ((typeof x === "object" || typeof x === "function") && x !== null) {
            const d = Object.getOwnPropertyDescriptor(x, name);
            if (d) {
                return d;
            }
            x = Object.getPrototypeOf(x);
        }
        return null;
    }
    /**
     * Check if a property is getter or not.
     * @param {object} object The object to check.
     * @param {string|number|symbol} name The property name to check.
     */
    function isGetter(object, name) {
        const d = getPropertyDescriptor(object, name);
        return d != null && d.get != null;
    }
    /**
     * Get the element values of a given node list.
     * @param {Node[]} nodeList The node list to get values.
     * @param {Scope|undefined} initialScope The initial scope to find variables.
     * @returns {any[]|null} The value list if all nodes are constant. Otherwise, null.
     */
    function getElementValues(nodeList, initialScope) {
        const valueList = [];
        for (let i = 0; i < nodeList.length; ++i) {
            const elementNode = nodeList[i];
            if (elementNode == null) {
                valueList.length = i + 1;
            }
            else if (elementNode.type === "SpreadElement") {
                const argument = getStaticValueR(elementNode.argument, initialScope);
                if (argument == null) {
                    return null;
                }
                valueList.push(...argument.value);
            }
            else {
                const element = getStaticValueR(elementNode, initialScope);
                if (element == null) {
                    return null;
                }
                valueList.push(element.value);
            }
        }
        return valueList;
    }
    const operations = Object.freeze({
        ArrayExpression(node, initialScope) {
            const elements = getElementValues(node.elements, initialScope);
            return elements != null ? { value: elements } : null;
        },
        AssignmentExpression(node, initialScope) {
            if (node.operator === "=") {
                return getStaticValueR(node.right, initialScope);
            }
            return null;
        },
        //eslint-disable-next-line complexity
        BinaryExpression(node, initialScope) {
            if (node.operator === "in" || node.operator === "instanceof") {
                // Not supported.
                return null;
            }
            const left = getStaticValueR(node.left, initialScope);
            const right = getStaticValueR(node.right, initialScope);
            if (left != null && right != null) {
                switch (node.operator) {
                    case "==":
                        return { value: left.value == right.value }; //eslint-disable-line eqeqeq
                    case "!=":
                        return { value: left.value != right.value }; //eslint-disable-line eqeqeq
                    case "===":
                        return { value: left.value === right.value };
                    case "!==":
                        return { value: left.value !== right.value };
                    case "<":
                        return { value: left.value < right.value };
                    case "<=":
                        return { value: left.value <= right.value };
                    case ">":
                        return { value: left.value > right.value };
                    case ">=":
                        return { value: left.value >= right.value };
                    case "<<":
                        return { value: left.value << right.value };
                    case ">>":
                        return { value: left.value >> right.value };
                    case ">>>":
                        return { value: left.value >>> right.value };
                    case "+":
                        return { value: left.value + right.value };
                    case "-":
                        return { value: left.value - right.value };
                    case "*":
                        return { value: left.value * right.value };
                    case "/":
                        return { value: left.value / right.value };
                    case "%":
                        return { value: left.value % right.value };
                    case "**":
                        return { value: Math.pow(left.value, right.value) };
                    case "|":
                        return { value: left.value | right.value };
                    case "^":
                        return { value: left.value ^ right.value };
                    case "&":
                        return { value: left.value & right.value };
                    // no default
                }
            }
            return null;
        },
        CallExpression(node, initialScope) {
            const calleeNode = node.callee;
            const args = getElementValues(node.arguments, initialScope);
            if (args != null) {
                if (calleeNode.type === "MemberExpression") {
                    if (calleeNode.property.type === "PrivateIdentifier") {
                        return null;
                    }
                    const object = getStaticValueR(calleeNode.object, initialScope);
                    if (object != null) {
                        if (object.value == null &&
                            (object.optional || node.optional)) {
                            return { value: undefined, optional: true };
                        }
                        const property = getStaticPropertyNameValue(calleeNode, initialScope);
                        if (property != null) {
                            const receiver = object.value;
                            const methodName = property.value;
                            if (callAllowed.has(receiver[methodName])) {
                                return { value: receiver[methodName](...args) };
                            }
                            if (callPassThrough.has(receiver[methodName])) {
                                return { value: args[0] };
                            }
                        }
                    }
                }
                else {
                    const callee = getStaticValueR(calleeNode, initialScope);
                    if (callee != null) {
                        if (callee.value == null && node.optional) {
                            return { value: undefined, optional: true };
                        }
                        const func = callee.value;
                        if (callAllowed.has(func)) {
                            return { value: func(...args) };
                        }
                        if (callPassThrough.has(func)) {
                            return { value: args[0] };
                        }
                    }
                }
            }
            return null;
        },
        ConditionalExpression(node, initialScope) {
            const test = getStaticValueR(node.test, initialScope);
            if (test != null) {
                return test.value
                    ? getStaticValueR(node.consequent, initialScope)
                    : getStaticValueR(node.alternate, initialScope);
            }
            return null;
        },
        ExpressionStatement(node, initialScope) {
            return getStaticValueR(node.expression, initialScope);
        },
        Identifier(node, initialScope) {
            if (initialScope != null) {
                const variable = findVariable(initialScope, node);
                // Built-in globals.
                if (variable != null &&
                    variable.defs.length === 0 &&
                    builtinNames.has(variable.name) &&
                    variable.name in globalObject) {
                    return { value: globalObject[variable.name] };
                }
                // Constants.
                if (variable != null && variable.defs.length === 1) {
                    const def = variable.defs[0];
                    if (def.parent &&
                        def.parent.kind === "const" &&
                        // TODO(mysticatea): don't support destructuring here.
                        def.node.id.type === "Identifier") {
                        return getStaticValueR(def.node.init, initialScope);
                    }
                }
            }
            return null;
        },
        Literal(node) {
            //istanbul ignore if : this is implementation-specific behavior.
            if ((node.regex != null || node.bigint != null) && node.value == null) {
                // It was a RegExp/BigInt literal, but Node.js didn't support it.
                return null;
            }
            return { value: node.value };
        },
        LogicalExpression(node, initialScope) {
            const left = getStaticValueR(node.left, initialScope);
            if (left != null) {
                if ((node.operator === "||" && Boolean(left.value) === true) ||
                    (node.operator === "&&" && Boolean(left.value) === false) ||
                    (node.operator === "??" && left.value != null)) {
                    return left;
                }
                const right = getStaticValueR(node.right, initialScope);
                if (right != null) {
                    return right;
                }
            }
            return null;
        },
        MemberExpression(node, initialScope) {
            if (node.property.type === "PrivateIdentifier") {
                return null;
            }
            const object = getStaticValueR(node.object, initialScope);
            if (object != null) {
                if (object.value == null && (object.optional || node.optional)) {
                    return { value: undefined, optional: true };
                }
                const property = getStaticPropertyNameValue(node, initialScope);
                if (property != null && !isGetter(object.value, property.value)) {
                    return { value: object.value[property.value] };
                }
            }
            return null;
        },
        ChainExpression(node, initialScope) {
            const expression = getStaticValueR(node.expression, initialScope);
            if (expression != null) {
                return { value: expression.value };
            }
            return null;
        },
        NewExpression(node, initialScope) {
            const callee = getStaticValueR(node.callee, initialScope);
            const args = getElementValues(node.arguments, initialScope);
            if (callee != null && args != null) {
                const Func = callee.value;
                if (callAllowed.has(Func)) {
                    return { value: new Func(...args) };
                }
            }
            return null;
        },
        ObjectExpression(node, initialScope) {
            const object = {};
            for (const propertyNode of node.properties) {
                if (propertyNode.type === "Property") {
                    if (propertyNode.kind !== "init") {
                        return null;
                    }
                    const key = getStaticPropertyNameValue(propertyNode, initialScope);
                    const value = getStaticValueR(propertyNode.value, initialScope);
                    if (key == null || value == null) {
                        return null;
                    }
                    object[key.value] = value.value;
                }
                else if (propertyNode.type === "SpreadElement" ||
                    propertyNode.type === "ExperimentalSpreadProperty") {
                    const argument = getStaticValueR(propertyNode.argument, initialScope);
                    if (argument == null) {
                        return null;
                    }
                    Object.assign(object, argument.value);
                }
                else {
                    return null;
                }
            }
            return { value: object };
        },
        SequenceExpression(node, initialScope) {
            const last = node.expressions[node.expressions.length - 1];
            return getStaticValueR(last, initialScope);
        },
        TaggedTemplateExpression(node, initialScope) {
            const tag = getStaticValueR(node.tag, initialScope);
            const expressions = getElementValues(node.quasi.expressions, initialScope);
            if (tag != null && expressions != null) {
                const func = tag.value;
                const strings = node.quasi.quasis.map((q) => q.value.cooked);
                strings.raw = node.quasi.quasis.map((q) => q.value.raw);
                if (func === String.raw) {
                    return { value: func(strings, ...expressions) };
                }
            }
            return null;
        },
        TemplateLiteral(node, initialScope) {
            const expressions = getElementValues(node.expressions, initialScope);
            if (expressions != null) {
                let value = node.quasis[0].value.cooked;
                for (let i = 0; i < expressions.length; ++i) {
                    value += expressions[i];
                    value += node.quasis[i + 1].value.cooked;
                }
                return { value };
            }
            return null;
        },
        UnaryExpression(node, initialScope) {
            if (node.operator === "delete") {
                // Not supported.
                return null;
            }
            if (node.operator === "void") {
                return { value: undefined };
            }
            const arg = getStaticValueR(node.argument, initialScope);
            if (arg != null) {
                switch (node.operator) {
                    case "-":
                        return { value: -arg.value };
                    case "+":
                        return { value: +arg.value }; //eslint-disable-line no-implicit-coercion
                    case "!":
                        return { value: !arg.value };
                    case "~":
                        return { value: ~arg.value };
                    case "typeof":
                        return { value: typeof arg.value };
                    // no default
                }
            }
            return null;
        },
    });
    /**
     * Get the value of a given node if it's a static value.
     * @param {Node} node The node to get.
     * @param {Scope|undefined} initialScope The scope to start finding variable.
     * @returns {{value:any}|{value:undefined,optional?:true}|null} The static value of the node, or `null`.
     */
    function getStaticValueR(node, initialScope) {
        if (node != null && Object.hasOwnProperty.call(operations, node.type)) {
            return operations[node.type](node, initialScope);
        }
        return null;
    }
    /**
     * Get the static value of property name from a MemberExpression node or a Property node.
     * @param {Node} node The node to get.
     * @param {Scope} [initialScope] The scope to start finding variable. Optional. If the node is a computed property node and this scope was given, this checks the computed property name by the `getStringIfConstant` function with the scope, and returns the value of it.
     * @returns {{value:any}|{value:undefined,optional?:true}|null} The static value of the property name of the node, or `null`.
     */
    function getStaticPropertyNameValue(node, initialScope) {
        const nameNode = node.type === "Property" ? node.key : node.property;
        if (node.computed) {
            return getStaticValueR(nameNode, initialScope);
        }
        if (nameNode.type === "Identifier") {
            return { value: nameNode.name };
        }
        if (nameNode.type === "Literal") {
            if (nameNode.bigint) {
                return { value: nameNode.bigint };
            }
            return { value: String(nameNode.value) };
        }
        return null;
    }
    /**
     * Get the value of a given node if it's a static value.
     * @param {Node} node The node to get.
     * @param {Scope} [initialScope] The scope to start finding variable. Optional. If this scope was given, this tries to resolve identifier references which are in the given node as much as possible.
     * @returns {{value:any}|{value:undefined,optional?:true}|null} The static value of the node, or `null`.
     */
    function getStaticValue(node, initialScope = null) {
        try {
            return getStaticValueR(node, initialScope);
        }
        catch (_error) {
            return null;
        }
    }
    /**
     * Get the value of a given node if it's a literal or a template literal.
     * @param {Node} node The node to get.
     * @param {Scope} [initialScope] The scope to start finding variable. Optional. If the node is an Identifier node and this scope was given, this checks the variable of the identifier, and returns the value of it if the variable is a constant.
     * @returns {string|null} The value of the node, or `null`.
     */
    function getStringIfConstant(node, initialScope = null) {
        // Handle the literals that the platform doesn't support natively.
        if (node && node.type === "Literal" && node.value === null) {
            if (node.regex) {
                return `/${node.regex.pattern}/${node.regex.flags}`;
            }
            if (node.bigint) {
                return node.bigint;
            }
        }
        const evaluated = getStaticValue(node, initialScope);
        return evaluated && String(evaluated.value);
    }
    /**
     * Get the property name from a MemberExpression node or a Property node.
     * @param {Node} node The node to get.
     * @param {Scope} [initialScope] The scope to start finding variable. Optional. If the node is a computed property node and this scope was given, this checks the computed property name by the `getStringIfConstant` function with the scope, and returns the value of it.
     * @returns {string|null} The property name of the node.
     */
    function getPropertyName(node, initialScope) {
        switch (node.type) {
            case "MemberExpression":
                if (node.computed) {
                    return getStringIfConstant(node.property, initialScope);
                }
                if (node.property.type === "PrivateIdentifier") {
                    return null;
                }
                return node.property.name;
            case "Property":
            case "MethodDefinition":
            case "PropertyDefinition":
                if (node.computed) {
                    return getStringIfConstant(node.key, initialScope);
                }
                if (node.key.type === "Literal") {
                    return String(node.key.value);
                }
                if (node.key.type === "PrivateIdentifier") {
                    return null;
                }
                return node.key.name;
            // no default
        }
        return null;
    }
    /**
     * Get the name and kind of the given function node.
     * @param {ASTNode} node - The function node to get.
     * @param {SourceCode} [sourceCode] The source code object to get the code of computed property keys.
     * @returns {string} The name and kind of the function node.
     */
    // eslint-disable-next-line complexity
    function getFunctionNameWithKind(node, sourceCode) {
        const parent = node.parent;
        const tokens = [];
        const isObjectMethod = parent.type === "Property" && parent.value === node;
        const isClassMethod = parent.type === "MethodDefinition" && parent.value === node;
        const isClassFieldMethod = parent.type === "PropertyDefinition" && parent.value === node;
        // Modifiers.
        if (isClassMethod || isClassFieldMethod) {
            if (parent.static) {
                tokens.push("static");
            }
            if (parent.key.type === "PrivateIdentifier") {
                tokens.push("private");
            }
        }
        if (node.async) {
            tokens.push("async");
        }
        if (node.generator) {
            tokens.push("generator");
        }
        // Kinds.
        if (isObjectMethod || isClassMethod) {
            if (parent.kind === "constructor") {
                return "constructor";
            }
            if (parent.kind === "get") {
                tokens.push("getter");
            }
            else if (parent.kind === "set") {
                tokens.push("setter");
            }
            else {
                tokens.push("method");
            }
        }
        else if (isClassFieldMethod) {
            tokens.push("method");
        }
        else {
            if (node.type === "ArrowFunctionExpression") {
                tokens.push("arrow");
            }
            tokens.push("function");
        }
        // Names.
        if (isObjectMethod || isClassMethod || isClassFieldMethod) {
            if (parent.key.type === "PrivateIdentifier") {
                tokens.push(`#${parent.key.name}`);
            }
            else {
                const name = getPropertyName(parent);
                if (name) {
                    tokens.push(`'${name}'`);
                }
                else if (sourceCode) {
                    const keyText = sourceCode.getText(parent.key);
                    if (!keyText.includes("\n")) {
                        tokens.push(`[${keyText}]`);
                    }
                }
            }
        }
        else if (node.id) {
            tokens.push(`'${node.id.name}'`);
        }
        else if (parent.type === "VariableDeclarator" &&
            parent.id &&
            parent.id.type === "Identifier") {
            tokens.push(`'${parent.id.name}'`);
        }
        else if ((parent.type === "AssignmentExpression" ||
            parent.type === "AssignmentPattern") &&
            parent.left &&
            parent.left.type === "Identifier") {
            tokens.push(`'${parent.left.name}'`);
        }
        return tokens.join(" ");
    }
    const typeConversionBinaryOps = Object.freeze(new Set([
        "==",
        "!=",
        "<",
        "<=",
        ">",
        ">=",
        "<<",
        ">>",
        ">>>",
        "+",
        "-",
        "*",
        "/",
        "%",
        "|",
        "^",
        "&",
        "in",
    ]));
    const typeConversionUnaryOps = Object.freeze(new Set(["-", "+", "!", "~"]));
    /**
     * Check whether the given value is an ASTNode or not.
     * @param {any} x The value to check.
     * @returns {boolean} `true` if the value is an ASTNode.
     */
    function isNode(x) {
        return x !== null && typeof x === "object" && typeof x.type === "string";
    }
    const visitor = Object.freeze(Object.assign(Object.create(null), {
        $visit(node, options, visitorKeys) {
            const { type } = node;
            if (typeof this[type] === "function") {
                return this[type](node, options, visitorKeys);
            }
            return this.$visitChildren(node, options, visitorKeys);
        },
        $visitChildren(node, options, visitorKeys) {
            const { type } = node;
            for (const key of visitorKeys[type] || evk.getKeys(node)) {
                const value = node[key];
                if (Array.isArray(value)) {
                    for (const element of value) {
                        if (isNode(element) &&
                            this.$visit(element, options, visitorKeys)) {
                            return true;
                        }
                    }
                }
                else if (isNode(value) &&
                    this.$visit(value, options, visitorKeys)) {
                    return true;
                }
            }
            return false;
        },
        ArrowFunctionExpression() {
            return false;
        },
        AssignmentExpression() {
            return true;
        },
        AwaitExpression() {
            return true;
        },
        BinaryExpression(node, options, visitorKeys) {
            if (options.considerImplicitTypeConversion &&
                typeConversionBinaryOps.has(node.operator) &&
                (node.left.type !== "Literal" || node.right.type !== "Literal")) {
                return true;
            }
            return this.$visitChildren(node, options, visitorKeys);
        },
        CallExpression() {
            return true;
        },
        FunctionExpression() {
            return false;
        },
        ImportExpression() {
            return true;
        },
        MemberExpression(node, options, visitorKeys) {
            if (options.considerGetters) {
                return true;
            }
            if (options.considerImplicitTypeConversion &&
                node.computed &&
                node.property.type !== "Literal") {
                return true;
            }
            return this.$visitChildren(node, options, visitorKeys);
        },
        MethodDefinition(node, options, visitorKeys) {
            if (options.considerImplicitTypeConversion &&
                node.computed &&
                node.key.type !== "Literal") {
                return true;
            }
            return this.$visitChildren(node, options, visitorKeys);
        },
        NewExpression() {
            return true;
        },
        Property(node, options, visitorKeys) {
            if (options.considerImplicitTypeConversion &&
                node.computed &&
                node.key.type !== "Literal") {
                return true;
            }
            return this.$visitChildren(node, options, visitorKeys);
        },
        PropertyDefinition(node, options, visitorKeys) {
            if (options.considerImplicitTypeConversion &&
                node.computed &&
                node.key.type !== "Literal") {
                return true;
            }
            return this.$visitChildren(node, options, visitorKeys);
        },
        UnaryExpression(node, options, visitorKeys) {
            if (node.operator === "delete") {
                return true;
            }
            if (options.considerImplicitTypeConversion &&
                typeConversionUnaryOps.has(node.operator) &&
                node.argument.type !== "Literal") {
                return true;
            }
            return this.$visitChildren(node, options, visitorKeys);
        },
        UpdateExpression() {
            return true;
        },
        YieldExpression() {
            return true;
        },
    }));
    /**
     * Check whether a given node has any side effect or not.
     * @param {Node} node The node to get.
     * @param {SourceCode} sourceCode The source code object.
     * @param {object} [options] The option object.
     * @param {boolean} [options.considerGetters=false] If `true` then it considers member accesses as the node which has side effects.
     * @param {boolean} [options.considerImplicitTypeConversion=false] If `true` then it considers implicit type conversion as the node which has side effects.
     * @param {object} [options.visitorKeys=evk.KEYS] The keys to traverse nodes. Use `context.getSourceCode().visitorKeys`.
     * @returns {boolean} `true` if the node has a certain side effect.
     */
    function hasSideEffect(node, sourceCode, { considerGetters = false, considerImplicitTypeConversion = false } = {}) {
        return visitor.$visit(node, { considerGetters, considerImplicitTypeConversion }, sourceCode.visitorKeys || evk.KEYS);
    }
    /**
     * Get the left parenthesis of the parent node syntax if it exists.
     * E.g., `if (a) {}` then the `(`.
     * @param {Node} node The AST node to check.
     * @param {SourceCode} sourceCode The source code object to get tokens.
     * @returns {Token|null} The left parenthesis of the parent node syntax
     */
    function getParentSyntaxParen(node, sourceCode) {
        const parent = node.parent;
        switch (parent.type) {
            case "CallExpression":
            case "NewExpression":
                if (parent.arguments.length === 1 && parent.arguments[0] === node) {
                    return sourceCode.getTokenAfter(parent.callee, isOpeningParenToken);
                }
                return null;
            case "DoWhileStatement":
                if (parent.test === node) {
                    return sourceCode.getTokenAfter(parent.body, isOpeningParenToken);
                }
                return null;
            case "IfStatement":
            case "WhileStatement":
                if (parent.test === node) {
                    return sourceCode.getFirstToken(parent, 1);
                }
                return null;
            case "ImportExpression":
                if (parent.source === node) {
                    return sourceCode.getFirstToken(parent, 1);
                }
                return null;
            case "SwitchStatement":
                if (parent.discriminant === node) {
                    return sourceCode.getFirstToken(parent, 1);
                }
                return null;
            case "WithStatement":
                if (parent.object === node) {
                    return sourceCode.getFirstToken(parent, 1);
                }
                return null;
            default:
                return null;
        }
    }
    /**
     * Check whether a given node is parenthesized or not.
     * @param {number} times The number of parantheses.
     * @param {Node} node The AST node to check.
     * @param {SourceCode} sourceCode The source code object to get tokens.
     * @returns {boolean} `true` if the node is parenthesized the given times.
     */
    /**
     * Check whether a given node is parenthesized or not.
     * @param {Node} node The AST node to check.
     * @param {SourceCode} sourceCode The source code object to get tokens.
     * @returns {boolean} `true` if the node is parenthesized.
     */
    function isParenthesized(timesOrNode, nodeOrSourceCode, optionalSourceCode) {
        let times, node, sourceCode, maybeLeftParen, maybeRightParen;
        if (typeof timesOrNode === "number") {
            times = timesOrNode | 0;
            node = nodeOrSourceCode;
            sourceCode = optionalSourceCode;
            if (!(times >= 1)) {
                throw new TypeError("'times' should be a positive integer.");
            }
        }
        else {
            times = 1;
            node = timesOrNode;
            sourceCode = nodeOrSourceCode;
        }
        if (node == null ||
            // `CatchClause.param` can't be parenthesized, example `try {} catch (error) {}`
            (node.parent.type === "CatchClause" && node.parent.param === node)) {
            return false;
        }
        maybeLeftParen = maybeRightParen = node;
        do {
            maybeLeftParen = sourceCode.getTokenBefore(maybeLeftParen);
            maybeRightParen = sourceCode.getTokenAfter(maybeRightParen);
        } while (maybeLeftParen != null &&
            maybeRightParen != null &&
            isOpeningParenToken(maybeLeftParen) &&
            isClosingParenToken(maybeRightParen) &&
            // Avoid false positive such as `if (a) {}`
            maybeLeftParen !== getParentSyntaxParen(node, sourceCode) &&
            --times > 0);
        return times === 0;
    }
    /**
     * @author Toru Nagashima <https://github.com/mysticatea>
     * See LICENSE file in root directory for full license.
     */
    const placeholder = /\$(?:[$&`']|[1-9][0-9]?)/gu;
    /** @type {WeakMap<PatternMatcher, {pattern:RegExp,escaped:boolean}>} */
    const internal = new WeakMap();
    /**
     * Check whether a given character is escaped or not.
     * @param {string} str The string to check.
     * @param {number} index The location of the character to check.
     * @returns {boolean} `true` if the character is escaped.
     */
    function isEscaped(str, index) {
        let escaped = false;
        for (let i = index - 1; i >= 0 && str.charCodeAt(i) === 0x5c; --i) {
            escaped = !escaped;
        }
        return escaped;
    }
    /**
     * Replace a given string by a given matcher.
     * @param {PatternMatcher} matcher The pattern matcher.
     * @param {string} str The string to be replaced.
     * @param {string} replacement The new substring to replace each matched part.
     * @returns {string} The replaced string.
     */
    function replaceS(matcher, str, replacement) {
        const chunks = [];
        let index = 0;
        /** @type {RegExpExecArray} */
        let match = null;
        /**
         * @param {string} key The placeholder.
         * @returns {string} The replaced string.
         */
        function replacer(key) {
            switch (key) {
                case "$$":
                    return "$";
                case "$&":
                    return match[0];
                case "$`":
                    return str.slice(0, match.index);
                case "$'":
                    return str.slice(match.index + match[0].length);
                default: {
                    const i = key.slice(1);
                    if (i in match) {
                        return match[i];
                    }
                    return key;
                }
            }
        }
        for (match of matcher.execAll(str)) {
            chunks.push(str.slice(index, match.index));
            chunks.push(replacement.replace(placeholder, replacer));
            index = match.index + match[0].length;
        }
        chunks.push(str.slice(index));
        return chunks.join("");
    }
    /**
     * Replace a given string by a given matcher.
     * @param {PatternMatcher} matcher The pattern matcher.
     * @param {string} str The string to be replaced.
     * @param {(...strs[])=>string} replace The function to replace each matched part.
     * @returns {string} The replaced string.
     */
    function replaceF(matcher, str, replace) {
        const chunks = [];
        let index = 0;
        for (const match of matcher.execAll(str)) {
            chunks.push(str.slice(index, match.index));
            chunks.push(String(replace(...match, match.index, match.input)));
            index = match.index + match[0].length;
        }
        chunks.push(str.slice(index));
        return chunks.join("");
    }
    /**
     * The class to find patterns as considering escape sequences.
     */
    class PatternMatcher {
        /**
         * Initialize this matcher.
         * @param {RegExp} pattern The pattern to match.
         * @param {{escaped:boolean}} options The options.
         */
        constructor(pattern, { escaped = false } = {}) {
            if (!(pattern instanceof RegExp)) {
                throw new TypeError("'pattern' should be a RegExp instance.");
            }
            if (!pattern.flags.includes("g")) {
                throw new Error("'pattern' should contains 'g' flag.");
            }
            internal.set(this, {
                pattern: new RegExp(pattern.source, pattern.flags),
                escaped: Boolean(escaped),
            });
        }
        /**
         * Find the pattern in a given string.
         * @param {string} str The string to find.
         * @returns {IterableIterator<RegExpExecArray>} The iterator which iterate the matched information.
         */
        *execAll(str) {
            const { pattern, escaped } = internal.get(this);
            let match = null;
            let lastIndex = 0;
            pattern.lastIndex = 0;
            while ((match = pattern.exec(str)) != null) {
                if (escaped || !isEscaped(str, match.index)) {
                    lastIndex = pattern.lastIndex;
                    yield match;
                    pattern.lastIndex = lastIndex;
                }
            }
        }
        /**
         * Check whether the pattern is found in a given string.
         * @param {string} str The string to check.
         * @returns {boolean} `true` if the pattern was found in the string.
         */
        test(str) {
            const it = this.execAll(str);
            const ret = it.next();
            return !ret.done;
        }
        /**
         * Replace a given string.
         * @param {string} str The string to be replaced.
         * @param {(string|((...strs:string[])=>string))} replacer The string or function to replace. This is the same as the 2nd argument of `String.prototype.replace`.
         * @returns {string} The replaced string.
         */
        [Symbol.replace](str, replacer) {
            return typeof replacer === "function"
                ? replaceF(this, String(str), replacer)
                : replaceS(this, String(str), String(replacer));
        }
    }
    const IMPORT_TYPE = /^(?:Import|Export(?:All|Default|Named))Declaration$/u;
    const has = Function.call.bind(Object.hasOwnProperty);
    const READ = Symbol("read");
    const CALL = Symbol("call");
    const CONSTRUCT = Symbol("construct");
    const ESM = Symbol("esm");
    const requireCall = { require: { [CALL]: true } };
    /**
     * Check whether a given variable is modified or not.
     * @param {Variable} variable The variable to check.
     * @returns {boolean} `true` if the variable is modified.
     */
    function isModifiedGlobal(variable) {
        return (variable == null ||
            variable.defs.length !== 0 ||
            variable.references.some((r) => r.isWrite()));
    }
    /**
     * Check if the value of a given node is passed through to the parent syntax as-is.
     * For example, `a` and `b` in (`a || b` and `c ? a : b`) are passed through.
     * @param {Node} node A node to check.
     * @returns {boolean} `true` if the node is passed through.
     */
    function isPassThrough(node) {
        const parent = node.parent;
        switch (parent && parent.type) {
            case "ConditionalExpression":
                return parent.consequent === node || parent.alternate === node;
            case "LogicalExpression":
                return true;
            case "SequenceExpression":
                return parent.expressions[parent.expressions.length - 1] === node;
            case "ChainExpression":
                return true;
            default:
                return false;
        }
    }
    /**
     * The reference tracker.
     */
    class ReferenceTracker {
        /**
         * Initialize this tracker.
         * @param {Scope} globalScope The global scope.
         * @param {object} [options] The options.
         * @param {"legacy"|"strict"} [options.mode="strict"] The mode to determine the ImportDeclaration's behavior for CJS modules.
         * @param {string[]} [options.globalObjectNames=["global","globalThis","self","window"]] The variable names for Global Object.
         */
        constructor(globalScope, { mode = "strict", globalObjectNames = ["global", "globalThis", "self", "window"], } = {}) {
            this.variableStack = [];
            this.globalScope = globalScope;
            this.mode = mode;
            this.globalObjectNames = globalObjectNames.slice(0);
        }
        /**
         * Iterate the references of global variables.
         * @param {object} traceMap The trace map.
         * @returns {IterableIterator<{node:Node,path:string[],type:symbol,info:any}>} The iterator to iterate references.
         */
        *iterateGlobalReferences(traceMap) {
            for (const key of Object.keys(traceMap)) {
                const nextTraceMap = traceMap[key];
                const path = [key];
                const variable = this.globalScope.set.get(key);
                if (isModifiedGlobal(variable)) {
                    continue;
                }
                yield* this._iterateVariableReferences(variable, path, nextTraceMap, true);
            }
            for (const key of this.globalObjectNames) {
                const path = [];
                const variable = this.globalScope.set.get(key);
                if (isModifiedGlobal(variable)) {
                    continue;
                }
                yield* this._iterateVariableReferences(variable, path, traceMap, false);
            }
        }
        /**
         * Iterate the references of CommonJS modules.
         * @param {object} traceMap The trace map.
         * @returns {IterableIterator<{node:Node,path:string[],type:symbol,info:any}>} The iterator to iterate references.
         */
        *iterateCjsReferences(traceMap) {
            for (const { node } of this.iterateGlobalReferences(requireCall)) {
                const key = getStringIfConstant(node.arguments[0]);
                if (key == null || !has(traceMap, key)) {
                    continue;
                }
                const nextTraceMap = traceMap[key];
                const path = [key];
                if (nextTraceMap[READ]) {
                    yield {
                        node,
                        path,
                        type: READ,
                        info: nextTraceMap[READ],
                    };
                }
                yield* this._iteratePropertyReferences(node, path, nextTraceMap);
            }
        }
        /**
         * Iterate the references of ES modules.
         * @param {object} traceMap The trace map.
         * @returns {IterableIterator<{node:Node,path:string[],type:symbol,info:any}>} The iterator to iterate references.
         */
        *iterateEsmReferences(traceMap) {
            const programNode = this.globalScope.block;
            for (const node of programNode.body) {
                if (!IMPORT_TYPE.test(node.type) || node.source == null) {
                    continue;
                }
                const moduleId = node.source.value;
                if (!has(traceMap, moduleId)) {
                    continue;
                }
                const nextTraceMap = traceMap[moduleId];
                const path = [moduleId];
                if (nextTraceMap[READ]) {
                    yield { node, path, type: READ, info: nextTraceMap[READ] };
                }
                if (node.type === "ExportAllDeclaration") {
                    for (const key of Object.keys(nextTraceMap)) {
                        const exportTraceMap = nextTraceMap[key];
                        if (exportTraceMap[READ]) {
                            yield {
                                node,
                                path: path.concat(key),
                                type: READ,
                                info: exportTraceMap[READ],
                            };
                        }
                    }
                }
                else {
                    for (const specifier of node.specifiers) {
                        const esm = has(nextTraceMap, ESM);
                        const it = this._iterateImportReferences(specifier, path, esm
                            ? nextTraceMap
                            : this.mode === "legacy"
                                ? { default: nextTraceMap, ...nextTraceMap }
                                : { default: nextTraceMap });
                        if (esm) {
                            yield* it;
                        }
                        else {
                            for (const report of it) {
                                report.path = report.path.filter(exceptDefault);
                                if (report.path.length >= 2 ||
                                    report.type !== READ) {
                                    yield report;
                                }
                            }
                        }
                    }
                }
            }
        }
        /**
         * Iterate the references for a given variable.
         * @param {Variable} variable The variable to iterate that references.
         * @param {string[]} path The current path.
         * @param {object} traceMap The trace map.
         * @param {boolean} shouldReport = The flag to report those references.
         * @returns {IterableIterator<{node:Node,path:string[],type:symbol,info:any}>} The iterator to iterate references.
         */
        *_iterateVariableReferences(variable, path, traceMap, shouldReport) {
            if (this.variableStack.includes(variable)) {
                return;
            }
            this.variableStack.push(variable);
            try {
                for (const reference of variable.references) {
                    if (!reference.isRead()) {
                        continue;
                    }
                    const node = reference.identifier;
                    if (shouldReport && traceMap[READ]) {
                        yield { node, path, type: READ, info: traceMap[READ] };
                    }
                    yield* this._iteratePropertyReferences(node, path, traceMap);
                }
            }
            finally {
                this.variableStack.pop();
            }
        }
        /**
         * Iterate the references for a given AST node.
         * @param rootNode The AST node to iterate references.
         * @param {string[]} path The current path.
         * @param {object} traceMap The trace map.
         * @returns {IterableIterator<{node:Node,path:string[],type:symbol,info:any}>} The iterator to iterate references.
         */
        //eslint-disable-next-line complexity
        *_iteratePropertyReferences(rootNode, path, traceMap) {
            let node = rootNode;
            while (isPassThrough(node)) {
                node = node.parent;
            }
            const parent = node.parent;
            if (parent.type === "MemberExpression") {
                if (parent.object === node) {
                    const key = getPropertyName(parent);
                    if (key == null || !has(traceMap, key)) {
                        return;
                    }
                    path = path.concat(key); //eslint-disable-line no-param-reassign
                    const nextTraceMap = traceMap[key];
                    if (nextTraceMap[READ]) {
                        yield {
                            node: parent,
                            path,
                            type: READ,
                            info: nextTraceMap[READ],
                        };
                    }
                    yield* this._iteratePropertyReferences(parent, path, nextTraceMap);
                }
                return;
            }
            if (parent.type === "CallExpression") {
                if (parent.callee === node && traceMap[CALL]) {
                    yield { node: parent, path, type: CALL, info: traceMap[CALL] };
                }
                return;
            }
            if (parent.type === "NewExpression") {
                if (parent.callee === node && traceMap[CONSTRUCT]) {
                    yield {
                        node: parent,
                        path,
                        type: CONSTRUCT,
                        info: traceMap[CONSTRUCT],
                    };
                }
                return;
            }
            if (parent.type === "AssignmentExpression") {
                if (parent.right === node) {
                    yield* this._iterateLhsReferences(parent.left, path, traceMap);
                    yield* this._iteratePropertyReferences(parent, path, traceMap);
                }
                return;
            }
            if (parent.type === "AssignmentPattern") {
                if (parent.right === node) {
                    yield* this._iterateLhsReferences(parent.left, path, traceMap);
                }
                return;
            }
            if (parent.type === "VariableDeclarator") {
                if (parent.init === node) {
                    yield* this._iterateLhsReferences(parent.id, path, traceMap);
                }
            }
        }
        /**
         * Iterate the references for a given Pattern node.
         * @param {Node} patternNode The Pattern node to iterate references.
         * @param {string[]} path The current path.
         * @param {object} traceMap The trace map.
         * @returns {IterableIterator<{node:Node,path:string[],type:symbol,info:any}>} The iterator to iterate references.
         */
        *_iterateLhsReferences(patternNode, path, traceMap) {
            if (patternNode.type === "Identifier") {
                const variable = findVariable(this.globalScope, patternNode);
                if (variable != null) {
                    yield* this._iterateVariableReferences(variable, path, traceMap, false);
                }
                return;
            }
            if (patternNode.type === "ObjectPattern") {
                for (const property of patternNode.properties) {
                    const key = getPropertyName(property);
                    if (key == null || !has(traceMap, key)) {
                        continue;
                    }
                    const nextPath = path.concat(key);
                    const nextTraceMap = traceMap[key];
                    if (nextTraceMap[READ]) {
                        yield {
                            node: property,
                            path: nextPath,
                            type: READ,
                            info: nextTraceMap[READ],
                        };
                    }
                    yield* this._iterateLhsReferences(property.value, nextPath, nextTraceMap);
                }
                return;
            }
            if (patternNode.type === "AssignmentPattern") {
                yield* this._iterateLhsReferences(patternNode.left, path, traceMap);
            }
        }
        /**
         * Iterate the references for a given ModuleSpecifier node.
         * @param {Node} specifierNode The ModuleSpecifier node to iterate references.
         * @param {string[]} path The current path.
         * @param {object} traceMap The trace map.
         * @returns {IterableIterator<{node:Node,path:string[],type:symbol,info:any}>} The iterator to iterate references.
         */
        *_iterateImportReferences(specifierNode, path, traceMap) {
            const type = specifierNode.type;
            if (type === "ImportSpecifier" || type === "ImportDefaultSpecifier") {
                const key = type === "ImportDefaultSpecifier"
                    ? "default"
                    : specifierNode.imported.name;
                if (!has(traceMap, key)) {
                    return;
                }
                path = path.concat(key); //eslint-disable-line no-param-reassign
                const nextTraceMap = traceMap[key];
                if (nextTraceMap[READ]) {
                    yield {
                        node: specifierNode,
                        path,
                        type: READ,
                        info: nextTraceMap[READ],
                    };
                }
                yield* this._iterateVariableReferences(findVariable(this.globalScope, specifierNode.local), path, nextTraceMap, false);
                return;
            }
            if (type === "ImportNamespaceSpecifier") {
                yield* this._iterateVariableReferences(findVariable(this.globalScope, specifierNode.local), path, traceMap, false);
                return;
            }
            if (type === "ExportSpecifier") {
                const key = specifierNode.local.name;
                if (!has(traceMap, key)) {
                    return;
                }
                path = path.concat(key); //eslint-disable-line no-param-reassign
                const nextTraceMap = traceMap[key];
                if (nextTraceMap[READ]) {
                    yield {
                        node: specifierNode,
                        path,
                        type: READ,
                        info: nextTraceMap[READ],
                    };
                }
            }
        }
    }
    ReferenceTracker.READ = READ;
    ReferenceTracker.CALL = CALL;
    ReferenceTracker.CONSTRUCT = CONSTRUCT;
    ReferenceTracker.ESM = ESM;
    /**
     * This is a predicate function for Array#filter.
     * @param {string} name A name part.
     * @param {number} index The index of the name.
     * @returns {boolean} `false` if it's default.
     */
    function exceptDefault(name, index) {
        return !(index === 1 && name === "default");
    }
    var index = {
        CALL,
        CONSTRUCT,
        ESM,
        findVariable,
        getFunctionHeadLocation,
        getFunctionNameWithKind,
        getInnermostScope,
        getPropertyName,
        getStaticValue,
        getStringIfConstant,
        hasSideEffect,
        isArrowToken,
        isClosingBraceToken,
        isClosingBracketToken,
        isClosingParenToken,
        isColonToken,
        isCommaToken,
        isCommentToken,
        isNotArrowToken,
        isNotClosingBraceToken,
        isNotClosingBracketToken,
        isNotClosingParenToken,
        isNotColonToken,
        isNotCommaToken,
        isNotCommentToken,
        isNotOpeningBraceToken,
        isNotOpeningBracketToken,
        isNotOpeningParenToken,
        isNotSemicolonToken,
        isOpeningBraceToken,
        isOpeningBracketToken,
        isOpeningParenToken,
        isParenthesized,
        isSemicolonToken,
        PatternMatcher,
        READ,
        ReferenceTracker,
    };
    exports.CALL = CALL;
    exports.CONSTRUCT = CONSTRUCT;
    exports.ESM = ESM;
    exports.PatternMatcher = PatternMatcher;
    exports.READ = READ;
    exports.ReferenceTracker = ReferenceTracker;
    exports.default = index;
    exports.findVariable = findVariable;
    exports.getFunctionHeadLocation = getFunctionHeadLocation;
    exports.getFunctionNameWithKind = getFunctionNameWithKind;
    exports.getInnermostScope = getInnermostScope;
    exports.getPropertyName = getPropertyName;
    exports.getStaticValue = getStaticValue;
    exports.getStringIfConstant = getStringIfConstant;
    exports.hasSideEffect = hasSideEffect;
    exports.isArrowToken = isArrowToken;
    exports.isClosingBraceToken = isClosingBraceToken;
    exports.isClosingBracketToken = isClosingBracketToken;
    exports.isClosingParenToken = isClosingParenToken;
    exports.isColonToken = isColonToken;
    exports.isCommaToken = isCommaToken;
    exports.isCommentToken = isCommentToken;
    exports.isNotArrowToken = isNotArrowToken;
    exports.isNotClosingBraceToken = isNotClosingBraceToken;
    exports.isNotClosingBracketToken = isNotClosingBracketToken;
    exports.isNotClosingParenToken = isNotClosingParenToken;
    exports.isNotColonToken = isNotColonToken;
    exports.isNotCommaToken = isNotCommaToken;
    exports.isNotCommentToken = isNotCommentToken;
    exports.isNotOpeningBraceToken = isNotOpeningBraceToken;
    exports.isNotOpeningBracketToken = isNotOpeningBracketToken;
    exports.isNotOpeningParenToken = isNotOpeningParenToken;
    exports.isNotSemicolonToken = isNotSemicolonToken;
    exports.isOpeningBraceToken = isOpeningBraceToken;
    exports.isOpeningBracketToken = isOpeningBracketToken;
    exports.isOpeningParenToken = isOpeningParenToken;
    exports.isParenthesized = isParenthesized;
    exports.isSemicolonToken = isSemicolonToken;
},
/* eslint/lib/source-code/token-store/index.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const assert = require("assert");
    const { isCommentToken } = require(350) /* eslint-utils */;
    const cursors = require(352) /* ./cursors */;
    const ForwardTokenCursor = require(360) /* ./forward-token-cursor */;
    const PaddedTokenCursor = require(363) /* ./padded-token-cursor */;
    const utils = require(355) /* ./utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const TOKENS = Symbol("tokens");
    const COMMENTS = Symbol("comments");
    const INDEX_MAP = Symbol("indexMap");
    /**
     * Creates the map from locations to indices in `tokens`.
     *
     * The first/last location of tokens is mapped to the index of the token.
     * The first/last location of comments is mapped to the index of the next token of each comment.
     * @param {Token[]} tokens The array of tokens.
     * @param {Comment[]} comments The array of comments.
     * @returns {Object} The map from locations to indices in `tokens`.
     * @private
     */
    function createIndexMap(tokens, comments) {
        const map = Object.create(null);
        let tokenIndex = 0;
        let commentIndex = 0;
        let nextStart = 0;
        let range = null;
        while (tokenIndex < tokens.length || commentIndex < comments.length) {
            nextStart = (commentIndex < comments.length) ? comments[commentIndex].range[0] : Number.MAX_SAFE_INTEGER;
            while (tokenIndex < tokens.length && (range = tokens[tokenIndex].range)[0] < nextStart) {
                map[range[0]] = tokenIndex;
                map[range[1] - 1] = tokenIndex;
                tokenIndex += 1;
            }
            nextStart = (tokenIndex < tokens.length) ? tokens[tokenIndex].range[0] : Number.MAX_SAFE_INTEGER;
            while (commentIndex < comments.length && (range = comments[commentIndex].range)[0] < nextStart) {
                map[range[0]] = tokenIndex;
                map[range[1] - 1] = tokenIndex;
                commentIndex += 1;
            }
        }
        return map;
    }
    /**
     * Creates the cursor iterates tokens with options.
     * @param {CursorFactory} factory The cursor factory to initialize cursor.
     * @param {Token[]} tokens The array of tokens.
     * @param {Comment[]} comments The array of comments.
     * @param {Object} indexMap The map from locations to indices in `tokens`.
     * @param {number} startLoc The start location of the iteration range.
     * @param {number} endLoc The end location of the iteration range.
     * @param {number|Function|Object} [opts=0] The option object. If this is a number then it's `opts.skip`. If this is a function then it's `opts.filter`.
     * @param {boolean} [opts.includeComments=false] The flag to iterate comments as well.
     * @param {Function|null} [opts.filter=null] The predicate function to choose tokens.
     * @param {number} [opts.skip=0] The count of tokens the cursor skips.
     * @returns {Cursor} The created cursor.
     * @private
     */
    function createCursorWithSkip(factory, tokens, comments, indexMap, startLoc, endLoc, opts) {
        let includeComments = false;
        let skip = 0;
        let filter = null;
        if (typeof opts === "number") {
            skip = opts | 0;
        }
        else if (typeof opts === "function") {
            filter = opts;
        }
        else if (opts) {
            includeComments = !!opts.includeComments;
            skip = opts.skip | 0;
            filter = opts.filter || null;
        }
        assert(skip >= 0, "options.skip should be zero or a positive integer.");
        assert(!filter || typeof filter === "function", "options.filter should be a function.");
        return factory.createCursor(tokens, comments, indexMap, startLoc, endLoc, includeComments, filter, skip, -1);
    }
    /**
     * Creates the cursor iterates tokens with options.
     * @param {CursorFactory} factory The cursor factory to initialize cursor.
     * @param {Token[]} tokens The array of tokens.
     * @param {Comment[]} comments The array of comments.
     * @param {Object} indexMap The map from locations to indices in `tokens`.
     * @param {number} startLoc The start location of the iteration range.
     * @param {number} endLoc The end location of the iteration range.
     * @param {number|Function|Object} [opts=0] The option object. If this is a number then it's `opts.count`. If this is a function then it's `opts.filter`.
     * @param {boolean} [opts.includeComments] The flag to iterate comments as well.
     * @param {Function|null} [opts.filter=null] The predicate function to choose tokens.
     * @param {number} [opts.count=0] The maximum count of tokens the cursor iterates. Zero is no iteration for backward compatibility.
     * @returns {Cursor} The created cursor.
     * @private
     */
    function createCursorWithCount(factory, tokens, comments, indexMap, startLoc, endLoc, opts) {
        let includeComments = false;
        let count = 0;
        let countExists = false;
        let filter = null;
        if (typeof opts === "number") {
            count = opts | 0;
            countExists = true;
        }
        else if (typeof opts === "function") {
            filter = opts;
        }
        else if (opts) {
            includeComments = !!opts.includeComments;
            count = opts.count | 0;
            countExists = typeof opts.count === "number";
            filter = opts.filter || null;
        }
        assert(count >= 0, "options.count should be zero or a positive integer.");
        assert(!filter || typeof filter === "function", "options.filter should be a function.");
        return factory.createCursor(tokens, comments, indexMap, startLoc, endLoc, includeComments, filter, 0, countExists ? count : -1);
    }
    /**
     * Creates the cursor iterates tokens with options.
     * This is overload function of the below.
     * @param {Token[]} tokens The array of tokens.
     * @param {Comment[]} comments The array of comments.
     * @param {Object} indexMap The map from locations to indices in `tokens`.
     * @param {number} startLoc The start location of the iteration range.
     * @param {number} endLoc The end location of the iteration range.
     * @param {Function|Object} opts The option object. If this is a function then it's `opts.filter`.
     * @param {boolean} [opts.includeComments] The flag to iterate comments as well.
     * @param {Function|null} [opts.filter=null] The predicate function to choose tokens.
     * @param {number} [opts.count=0] The maximum count of tokens the cursor iterates. Zero is no iteration for backward compatibility.
     * @returns {Cursor} The created cursor.
     * @private
     */
    /**
     * Creates the cursor iterates tokens with options.
     * @param {Token[]} tokens The array of tokens.
     * @param {Comment[]} comments The array of comments.
     * @param {Object} indexMap The map from locations to indices in `tokens`.
     * @param {number} startLoc The start location of the iteration range.
     * @param {number} endLoc The end location of the iteration range.
     * @param {number} [beforeCount=0] The number of tokens before the node to retrieve.
     * @param {boolean} [afterCount=0] The number of tokens after the node to retrieve.
     * @returns {Cursor} The created cursor.
     * @private
     */
    function createCursorWithPadding(tokens, comments, indexMap, startLoc, endLoc, beforeCount, afterCount) {
        if (typeof beforeCount === "undefined" && typeof afterCount === "undefined") {
            return new ForwardTokenCursor(tokens, comments, indexMap, startLoc, endLoc);
        }
        if (typeof beforeCount === "number" || typeof beforeCount === "undefined") {
            return new PaddedTokenCursor(tokens, comments, indexMap, startLoc, endLoc, beforeCount | 0, afterCount | 0);
        }
        return createCursorWithCount(cursors.forward, tokens, comments, indexMap, startLoc, endLoc, beforeCount);
    }
    /**
     * Gets comment tokens that are adjacent to the current cursor position.
     * @param {Cursor} cursor A cursor instance.
     * @returns {Array} An array of comment tokens adjacent to the current cursor position.
     * @private
     */
    function getAdjacentCommentTokensFromCursor(cursor) {
        const tokens = [];
        let currentToken = cursor.getOneToken();
        while (currentToken && isCommentToken(currentToken)) {
            tokens.push(currentToken);
            currentToken = cursor.getOneToken();
        }
        return tokens;
    }
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The token store.
     *
     * This class provides methods to get tokens by locations as fast as possible.
     * The methods are a part of public API, so we should be careful if it changes this class.
     *
     * People can get tokens in O(1) by the hash map which is mapping from the location of tokens/comments to tokens.
     * Also people can get a mix of tokens and comments in O(log k), the k is the number of comments.
     * Assuming that comments to be much fewer than tokens, this does not make hash map from token's locations to comments to reduce memory cost.
     * This uses binary-searching instead for comments.
     */
    module.exports = class TokenStore {
        /**
         * Initializes this token store.
         * @param {Token[]} tokens The array of tokens.
         * @param {Comment[]} comments The array of comments.
         */
        constructor(tokens, comments) {
            this[TOKENS] = tokens;
            this[COMMENTS] = comments;
            this[INDEX_MAP] = createIndexMap(tokens, comments);
        }
        //--------------------------------------------------------------------------
        // Gets single token.
        //--------------------------------------------------------------------------
        /**
         * Gets the token starting at the specified index.
         * @param {number} offset Index of the start of the token's range.
         * @param {Object} [options=0] The option object.
         * @param {boolean} [options.includeComments=false] The flag to iterate comments as well.
         * @returns {Token|null} The token starting at index, or null if no such token.
         */
        getTokenByRangeStart(offset, options) {
            const includeComments = options && options.includeComments;
            const token = cursors.forward.createBaseCursor(this[TOKENS], this[COMMENTS], this[INDEX_MAP], offset, -1, includeComments).getOneToken();
            if (token && token.range[0] === offset) {
                return token;
            }
            return null;
        }
        /**
         * Gets the first token of the given node.
         * @param {ASTNode} node The AST node.
         * @param {number|Function|Object} [options=0] The option object. If this is a number then it's `options.skip`. If this is a function then it's `options.filter`.
         * @param {boolean} [options.includeComments=false] The flag to iterate comments as well.
         * @param {Function|null} [options.filter=null] The predicate function to choose tokens.
         * @param {number} [options.skip=0] The count of tokens the cursor skips.
         * @returns {Token|null} An object representing the token.
         */
        getFirstToken(node, options) {
            return createCursorWithSkip(cursors.forward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], node.range[0], node.range[1], options).getOneToken();
        }
        /**
         * Gets the last token of the given node.
         * @param {ASTNode} node The AST node.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstToken()
         * @returns {Token|null} An object representing the token.
         */
        getLastToken(node, options) {
            return createCursorWithSkip(cursors.backward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], node.range[0], node.range[1], options).getOneToken();
        }
        /**
         * Gets the token that precedes a given node or token.
         * @param {ASTNode|Token|Comment} node The AST node or token.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstToken()
         * @returns {Token|null} An object representing the token.
         */
        getTokenBefore(node, options) {
            return createCursorWithSkip(cursors.backward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], -1, node.range[0], options).getOneToken();
        }
        /**
         * Gets the token that follows a given node or token.
         * @param {ASTNode|Token|Comment} node The AST node or token.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstToken()
         * @returns {Token|null} An object representing the token.
         */
        getTokenAfter(node, options) {
            return createCursorWithSkip(cursors.forward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], node.range[1], -1, options).getOneToken();
        }
        /**
         * Gets the first token between two non-overlapping nodes.
         * @param {ASTNode|Token|Comment} left Node before the desired token range.
         * @param {ASTNode|Token|Comment} right Node after the desired token range.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstToken()
         * @returns {Token|null} An object representing the token.
         */
        getFirstTokenBetween(left, right, options) {
            return createCursorWithSkip(cursors.forward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], left.range[1], right.range[0], options).getOneToken();
        }
        /**
         * Gets the last token between two non-overlapping nodes.
         * @param {ASTNode|Token|Comment} left Node before the desired token range.
         * @param {ASTNode|Token|Comment} right Node after the desired token range.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstToken()
         * @returns {Token|null} An object representing the token.
         */
        getLastTokenBetween(left, right, options) {
            return createCursorWithSkip(cursors.backward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], left.range[1], right.range[0], options).getOneToken();
        }
        /**
         * Gets the token that precedes a given node or token in the token stream.
         * This is defined for backward compatibility. Use `includeComments` option instead.
         * TODO: We have a plan to remove this in a future major version.
         * @param {ASTNode|Token|Comment} node The AST node or token.
         * @param {number} [skip=0] A number of tokens to skip.
         * @returns {Token|null} An object representing the token.
         * @deprecated
         */
        getTokenOrCommentBefore(node, skip) {
            return this.getTokenBefore(node, { includeComments: true, skip });
        }
        /**
         * Gets the token that follows a given node or token in the token stream.
         * This is defined for backward compatibility. Use `includeComments` option instead.
         * TODO: We have a plan to remove this in a future major version.
         * @param {ASTNode|Token|Comment} node The AST node or token.
         * @param {number} [skip=0] A number of tokens to skip.
         * @returns {Token|null} An object representing the token.
         * @deprecated
         */
        getTokenOrCommentAfter(node, skip) {
            return this.getTokenAfter(node, { includeComments: true, skip });
        }
        //--------------------------------------------------------------------------
        // Gets multiple tokens.
        //--------------------------------------------------------------------------
        /**
         * Gets the first `count` tokens of the given node.
         * @param {ASTNode} node The AST node.
         * @param {number|Function|Object} [options=0] The option object. If this is a number then it's `options.count`. If this is a function then it's `options.filter`.
         * @param {boolean} [options.includeComments=false] The flag to iterate comments as well.
         * @param {Function|null} [options.filter=null] The predicate function to choose tokens.
         * @param {number} [options.count=0] The maximum count of tokens the cursor iterates.
         * @returns {Token[]} Tokens.
         */
        getFirstTokens(node, options) {
            return createCursorWithCount(cursors.forward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], node.range[0], node.range[1], options).getAllTokens();
        }
        /**
         * Gets the last `count` tokens of the given node.
         * @param {ASTNode} node The AST node.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstTokens()
         * @returns {Token[]} Tokens.
         */
        getLastTokens(node, options) {
            return createCursorWithCount(cursors.backward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], node.range[0], node.range[1], options).getAllTokens().reverse();
        }
        /**
         * Gets the `count` tokens that precedes a given node or token.
         * @param {ASTNode|Token|Comment} node The AST node or token.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstTokens()
         * @returns {Token[]} Tokens.
         */
        getTokensBefore(node, options) {
            return createCursorWithCount(cursors.backward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], -1, node.range[0], options).getAllTokens().reverse();
        }
        /**
         * Gets the `count` tokens that follows a given node or token.
         * @param {ASTNode|Token|Comment} node The AST node or token.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstTokens()
         * @returns {Token[]} Tokens.
         */
        getTokensAfter(node, options) {
            return createCursorWithCount(cursors.forward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], node.range[1], -1, options).getAllTokens();
        }
        /**
         * Gets the first `count` tokens between two non-overlapping nodes.
         * @param {ASTNode|Token|Comment} left Node before the desired token range.
         * @param {ASTNode|Token|Comment} right Node after the desired token range.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstTokens()
         * @returns {Token[]} Tokens between left and right.
         */
        getFirstTokensBetween(left, right, options) {
            return createCursorWithCount(cursors.forward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], left.range[1], right.range[0], options).getAllTokens();
        }
        /**
         * Gets the last `count` tokens between two non-overlapping nodes.
         * @param {ASTNode|Token|Comment} left Node before the desired token range.
         * @param {ASTNode|Token|Comment} right Node after the desired token range.
         * @param {number|Function|Object} [options=0] The option object. Same options as getFirstTokens()
         * @returns {Token[]} Tokens between left and right.
         */
        getLastTokensBetween(left, right, options) {
            return createCursorWithCount(cursors.backward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], left.range[1], right.range[0], options).getAllTokens().reverse();
        }
        /**
         * Gets all tokens that are related to the given node.
         * @param {ASTNode} node The AST node.
         * @param {Function|Object} options The option object. If this is a function then it's `options.filter`.
         * @param {boolean} [options.includeComments=false] The flag to iterate comments as well.
         * @param {Function|null} [options.filter=null] The predicate function to choose tokens.
         * @param {number} [options.count=0] The maximum count of tokens the cursor iterates.
         * @returns {Token[]} Array of objects representing tokens.
         */
        /**
         * Gets all tokens that are related to the given node.
         * @param {ASTNode} node The AST node.
         * @param {int} [beforeCount=0] The number of tokens before the node to retrieve.
         * @param {int} [afterCount=0] The number of tokens after the node to retrieve.
         * @returns {Token[]} Array of objects representing tokens.
         */
        getTokens(node, beforeCount, afterCount) {
            return createCursorWithPadding(this[TOKENS], this[COMMENTS], this[INDEX_MAP], node.range[0], node.range[1], beforeCount, afterCount).getAllTokens();
        }
        /**
         * Gets all of the tokens between two non-overlapping nodes.
         * @param {ASTNode|Token|Comment} left Node before the desired token range.
         * @param {ASTNode|Token|Comment} right Node after the desired token range.
         * @param {Function|Object} options The option object. If this is a function then it's `options.filter`.
         * @param {boolean} [options.includeComments=false] The flag to iterate comments as well.
         * @param {Function|null} [options.filter=null] The predicate function to choose tokens.
         * @param {number} [options.count=0] The maximum count of tokens the cursor iterates.
         * @returns {Token[]} Tokens between left and right.
         */
        /**
         * Gets all of the tokens between two non-overlapping nodes.
         * @param {ASTNode|Token|Comment} left Node before the desired token range.
         * @param {ASTNode|Token|Comment} right Node after the desired token range.
         * @param {int} [padding=0] Number of extra tokens on either side of center.
         * @returns {Token[]} Tokens between left and right.
         */
        getTokensBetween(left, right, padding) {
            return createCursorWithPadding(this[TOKENS], this[COMMENTS], this[INDEX_MAP], left.range[1], right.range[0], padding, padding).getAllTokens();
        }
        //--------------------------------------------------------------------------
        // Others.
        //--------------------------------------------------------------------------
        /**
         * Checks whether any comments exist or not between the given 2 nodes.
         * @param {ASTNode} left The node to check.
         * @param {ASTNode} right The node to check.
         * @returns {boolean} `true` if one or more comments exist.
         */
        commentsExistBetween(left, right) {
            const index = utils.search(this[COMMENTS], left.range[1]);
            return (index < this[COMMENTS].length &&
                this[COMMENTS][index].range[1] <= right.range[0]);
        }
        /**
         * Gets all comment tokens directly before the given node or token.
         * @param {ASTNode|token} nodeOrToken The AST node or token to check for adjacent comment tokens.
         * @returns {Array} An array of comments in occurrence order.
         */
        getCommentsBefore(nodeOrToken) {
            const cursor = createCursorWithCount(cursors.backward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], -1, nodeOrToken.range[0], { includeComments: true });
            return getAdjacentCommentTokensFromCursor(cursor).reverse();
        }
        /**
         * Gets all comment tokens directly after the given node or token.
         * @param {ASTNode|token} nodeOrToken The AST node or token to check for adjacent comment tokens.
         * @returns {Array} An array of comments in occurrence order.
         */
        getCommentsAfter(nodeOrToken) {
            const cursor = createCursorWithCount(cursors.forward, this[TOKENS], this[COMMENTS], this[INDEX_MAP], nodeOrToken.range[1], -1, { includeComments: true });
            return getAdjacentCommentTokensFromCursor(cursor);
        }
        /**
         * Gets all comment tokens inside the given node.
         * @param {ASTNode} node The AST node to get the comments for.
         * @returns {Array} An array of comments in occurrence order.
         */
        getCommentsInside(node) {
            return this.getTokens(node, {
                includeComments: true,
                filter: isCommentToken
            });
        }
    };
},
/* eslint/lib/source-code/token-store/cursors.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const BackwardTokenCommentCursor = require(353) /* ./backward-token-comment-cursor */;
    const BackwardTokenCursor = require(356) /* ./backward-token-cursor */;
    const FilterCursor = require(357) /* ./filter-cursor */;
    const ForwardTokenCommentCursor = require(359) /* ./forward-token-comment-cursor */;
    const ForwardTokenCursor = require(360) /* ./forward-token-cursor */;
    const LimitCursor = require(361) /* ./limit-cursor */;
    const SkipCursor = require(362) /* ./skip-cursor */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * The cursor factory.
     * @private
     */
    class CursorFactory {
        /**
         * Initializes this cursor.
         * @param {Function} TokenCursor The class of the cursor which iterates tokens only.
         * @param {Function} TokenCommentCursor The class of the cursor which iterates the mix of tokens and comments.
         */
        constructor(TokenCursor, TokenCommentCursor) {
            this.TokenCursor = TokenCursor;
            this.TokenCommentCursor = TokenCommentCursor;
        }
        /**
         * Creates a base cursor instance that can be decorated by createCursor.
         * @param {Token[]} tokens The array of tokens.
         * @param {Comment[]} comments The array of comments.
         * @param {Object} indexMap The map from locations to indices in `tokens`.
         * @param {number} startLoc The start location of the iteration range.
         * @param {number} endLoc The end location of the iteration range.
         * @param {boolean} includeComments The flag to iterate comments as well.
         * @returns {Cursor} The created base cursor.
         */
        createBaseCursor(tokens, comments, indexMap, startLoc, endLoc, includeComments) {
            const Cursor = includeComments ? this.TokenCommentCursor : this.TokenCursor;
            return new Cursor(tokens, comments, indexMap, startLoc, endLoc);
        }
        /**
         * Creates a cursor that iterates tokens with normalized options.
         * @param {Token[]} tokens The array of tokens.
         * @param {Comment[]} comments The array of comments.
         * @param {Object} indexMap The map from locations to indices in `tokens`.
         * @param {number} startLoc The start location of the iteration range.
         * @param {number} endLoc The end location of the iteration range.
         * @param {boolean} includeComments The flag to iterate comments as well.
         * @param {Function|null} filter The predicate function to choose tokens.
         * @param {number} skip The count of tokens the cursor skips.
         * @param {number} count The maximum count of tokens the cursor iterates. Zero is no iteration for backward compatibility.
         * @returns {Cursor} The created cursor.
         */
        createCursor(tokens, comments, indexMap, startLoc, endLoc, includeComments, filter, skip, count) {
            let cursor = this.createBaseCursor(tokens, comments, indexMap, startLoc, endLoc, includeComments);
            if (filter) {
                cursor = new FilterCursor(cursor, filter);
            }
            if (skip >= 1) {
                cursor = new SkipCursor(cursor, skip);
            }
            if (count >= 0) {
                cursor = new LimitCursor(cursor, count);
            }
            return cursor;
        }
    }
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    exports.forward = new CursorFactory(ForwardTokenCursor, ForwardTokenCommentCursor);
    exports.backward = new CursorFactory(BackwardTokenCursor, BackwardTokenCommentCursor);
},
/* eslint/lib/source-code/token-store/backward-token-comment-cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const Cursor = require(354) /* ./cursor */;
    const utils = require(355) /* ./utils */;
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The cursor which iterates tokens and comments in reverse.
     */
    module.exports = class BackwardTokenCommentCursor extends Cursor {
        /**
         * Initializes this cursor.
         * @param {Token[]} tokens The array of tokens.
         * @param {Comment[]} comments The array of comments.
         * @param {Object} indexMap The map from locations to indices in `tokens`.
         * @param {number} startLoc The start location of the iteration range.
         * @param {number} endLoc The end location of the iteration range.
         */
        constructor(tokens, comments, indexMap, startLoc, endLoc) {
            super();
            this.tokens = tokens;
            this.comments = comments;
            this.tokenIndex = utils.getLastIndex(tokens, indexMap, endLoc);
            this.commentIndex = utils.search(comments, endLoc) - 1;
            this.border = startLoc;
        }
        /** @inheritdoc */
        moveNext() {
            const token = (this.tokenIndex >= 0) ? this.tokens[this.tokenIndex] : null;
            const comment = (this.commentIndex >= 0) ? this.comments[this.commentIndex] : null;
            if (token && (!comment || token.range[1] > comment.range[1])) {
                this.current = token;
                this.tokenIndex -= 1;
            }
            else if (comment) {
                this.current = comment;
                this.commentIndex -= 1;
            }
            else {
                this.current = null;
            }
            return Boolean(this.current) && (this.border === -1 || this.current.range[0] >= this.border);
        }
    };
},
/* eslint/lib/source-code/token-store/cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The abstract class about cursors which iterate tokens.
     *
     * This class has 2 abstract methods.
     *
     * - `current: Token | Comment | null` ... The current token.
     * - `moveNext(): boolean` ... Moves this cursor to the next token. If the next token didn't exist, it returns `false`.
     *
     * This is similar to ES2015 Iterators.
     * However, Iterators were slow (at 2017-01), so I created this class as similar to C# IEnumerable.
     *
     * There are the following known sub classes.
     *
     * - ForwardTokenCursor .......... The cursor which iterates tokens only.
     * - BackwardTokenCursor ......... The cursor which iterates tokens only in reverse.
     * - ForwardTokenCommentCursor ... The cursor which iterates tokens and comments.
     * - BackwardTokenCommentCursor .. The cursor which iterates tokens and comments in reverse.
     * - DecorativeCursor
     *     - FilterCursor ............ The cursor which ignores the specified tokens.
     *     - SkipCursor .............. The cursor which ignores the first few tokens.
     *     - LimitCursor ............. The cursor which limits the count of tokens.
     *
     */
    module.exports = class Cursor {
        /**
         * Initializes this cursor.
         */
        constructor() {
            this.current = null;
        }
        /**
         * Gets the first token.
         * This consumes this cursor.
         * @returns {Token|Comment} The first token or null.
         */
        getOneToken() {
            return this.moveNext() ? this.current : null;
        }
        /**
         * Gets the first tokens.
         * This consumes this cursor.
         * @returns {(Token|Comment)[]} All tokens.
         */
        getAllTokens() {
            const tokens = [];
            while (this.moveNext()) {
                tokens.push(this.current);
            }
            return tokens;
        }
        /**
         * Moves this cursor to the next token.
         * @returns {boolean} `true` if the next token exists.
         * @abstract
         */
        /* istanbul ignore next */
        moveNext() {
            throw new Error("Not implemented.");
        }
    };
},
/* eslint/lib/source-code/token-store/utils.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Gets `token.range[0]` from the given token.
     * @param {Node|Token|Comment} token The token to get.
     * @returns {number} The start location.
     * @private
     */
    function getStartLocation(token) {
        return token.range[0];
    }
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * Finds the index of the first token which is after the given location.
     * If it was not found, this returns `tokens.length`.
     * @param {(Token|Comment)[]} tokens It searches the token in this list.
     * @param {number} location The location to search.
     * @returns {number} The found index or `tokens.length`.
     */
    exports.search = function search(tokens, location) {
        const index = tokens.findIndex(el => location <= getStartLocation(el));
        return index === -1 ? tokens.length : index;
    };
    /**
     * Gets the index of the `startLoc` in `tokens`.
     * `startLoc` can be the value of `node.range[1]`, so this checks about `startLoc - 1` as well.
     * @param {(Token|Comment)[]} tokens The tokens to find an index.
     * @param {Object} indexMap The map from locations to indices.
     * @param {number} startLoc The location to get an index.
     * @returns {number} The index.
     */
    exports.getFirstIndex = function getFirstIndex(tokens, indexMap, startLoc) {
        if (startLoc in indexMap) {
            return indexMap[startLoc];
        }
        if ((startLoc - 1) in indexMap) {
            const index = indexMap[startLoc - 1];
            const token = (index >= 0 && index < tokens.length) ? tokens[index] : null;
            /*
             * For the map of "comment's location -> token's index", it points the next token of a comment.
             * In that case, +1 is unnecessary.
             */
            if (token && token.range[0] >= startLoc) {
                return index;
            }
            return index + 1;
        }
        return 0;
    };
    /**
     * Gets the index of the `endLoc` in `tokens`.
     * The information of end locations are recorded at `endLoc - 1` in `indexMap`, so this checks about `endLoc - 1` as well.
     * @param {(Token|Comment)[]} tokens The tokens to find an index.
     * @param {Object} indexMap The map from locations to indices.
     * @param {number} endLoc The location to get an index.
     * @returns {number} The index.
     */
    exports.getLastIndex = function getLastIndex(tokens, indexMap, endLoc) {
        if (endLoc in indexMap) {
            return indexMap[endLoc] - 1;
        }
        if ((endLoc - 1) in indexMap) {
            const index = indexMap[endLoc - 1];
            const token = (index >= 0 && index < tokens.length) ? tokens[index] : null;
            /*
             * For the map of "comment's location -> token's index", it points the next token of a comment.
             * In that case, -1 is necessary.
             */
            if (token && token.range[1] > endLoc) {
                return index - 1;
            }
            return index;
        }
        return tokens.length - 1;
    };
},
/* eslint/lib/source-code/token-store/backward-token-cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const Cursor = require(354) /* ./cursor */;
    const utils = require(355) /* ./utils */;
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The cursor which iterates tokens only in reverse.
     */
    module.exports = class BackwardTokenCursor extends Cursor {
        /**
         * Initializes this cursor.
         * @param {Token[]} tokens The array of tokens.
         * @param {Comment[]} comments The array of comments.
         * @param {Object} indexMap The map from locations to indices in `tokens`.
         * @param {number} startLoc The start location of the iteration range.
         * @param {number} endLoc The end location of the iteration range.
         */
        constructor(tokens, comments, indexMap, startLoc, endLoc) {
            super();
            this.tokens = tokens;
            this.index = utils.getLastIndex(tokens, indexMap, endLoc);
            this.indexEnd = utils.getFirstIndex(tokens, indexMap, startLoc);
        }
        /** @inheritdoc */
        moveNext() {
            if (this.index >= this.indexEnd) {
                this.current = this.tokens[this.index];
                this.index -= 1;
                return true;
            }
            return false;
        }
        /*
         *
         * Shorthand for performance.
         *
         */
        /** @inheritdoc */
        getOneToken() {
            return (this.index >= this.indexEnd) ? this.tokens[this.index] : null;
        }
    };
},
/* eslint/lib/source-code/token-store/filter-cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const DecorativeCursor = require(358) /* ./decorative-cursor */;
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The decorative cursor which ignores specified tokens.
     */
    module.exports = class FilterCursor extends DecorativeCursor {
        /**
         * Initializes this cursor.
         * @param {Cursor} cursor The cursor to be decorated.
         * @param {Function} predicate The predicate function to decide tokens this cursor iterates.
         */
        constructor(cursor, predicate) {
            super(cursor);
            this.predicate = predicate;
        }
        /** @inheritdoc */
        moveNext() {
            const predicate = this.predicate;
            while (super.moveNext()) {
                if (predicate(this.current)) {
                    return true;
                }
            }
            return false;
        }
    };
},
/* eslint/lib/source-code/token-store/decorative-cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const Cursor = require(354) /* ./cursor */;
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The abstract class about cursors which manipulate another cursor.
     */
    module.exports = class DecorativeCursor extends Cursor {
        /**
         * Initializes this cursor.
         * @param {Cursor} cursor The cursor to be decorated.
         */
        constructor(cursor) {
            super();
            this.cursor = cursor;
        }
        /** @inheritdoc */
        moveNext() {
            const retv = this.cursor.moveNext();
            this.current = this.cursor.current;
            return retv;
        }
    };
},
/* eslint/lib/source-code/token-store/forward-token-comment-cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const Cursor = require(354) /* ./cursor */;
    const utils = require(355) /* ./utils */;
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The cursor which iterates tokens and comments.
     */
    module.exports = class ForwardTokenCommentCursor extends Cursor {
        /**
         * Initializes this cursor.
         * @param {Token[]} tokens The array of tokens.
         * @param {Comment[]} comments The array of comments.
         * @param {Object} indexMap The map from locations to indices in `tokens`.
         * @param {number} startLoc The start location of the iteration range.
         * @param {number} endLoc The end location of the iteration range.
         */
        constructor(tokens, comments, indexMap, startLoc, endLoc) {
            super();
            this.tokens = tokens;
            this.comments = comments;
            this.tokenIndex = utils.getFirstIndex(tokens, indexMap, startLoc);
            this.commentIndex = utils.search(comments, startLoc);
            this.border = endLoc;
        }
        /** @inheritdoc */
        moveNext() {
            const token = (this.tokenIndex < this.tokens.length) ? this.tokens[this.tokenIndex] : null;
            const comment = (this.commentIndex < this.comments.length) ? this.comments[this.commentIndex] : null;
            if (token && (!comment || token.range[0] < comment.range[0])) {
                this.current = token;
                this.tokenIndex += 1;
            }
            else if (comment) {
                this.current = comment;
                this.commentIndex += 1;
            }
            else {
                this.current = null;
            }
            return Boolean(this.current) && (this.border === -1 || this.current.range[1] <= this.border);
        }
    };
},
/* eslint/lib/source-code/token-store/forward-token-cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const Cursor = require(354) /* ./cursor */;
    const utils = require(355) /* ./utils */;
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The cursor which iterates tokens only.
     */
    module.exports = class ForwardTokenCursor extends Cursor {
        /**
         * Initializes this cursor.
         * @param {Token[]} tokens The array of tokens.
         * @param {Comment[]} comments The array of comments.
         * @param {Object} indexMap The map from locations to indices in `tokens`.
         * @param {number} startLoc The start location of the iteration range.
         * @param {number} endLoc The end location of the iteration range.
         */
        constructor(tokens, comments, indexMap, startLoc, endLoc) {
            super();
            this.tokens = tokens;
            this.index = utils.getFirstIndex(tokens, indexMap, startLoc);
            this.indexEnd = utils.getLastIndex(tokens, indexMap, endLoc);
        }
        /** @inheritdoc */
        moveNext() {
            if (this.index <= this.indexEnd) {
                this.current = this.tokens[this.index];
                this.index += 1;
                return true;
            }
            return false;
        }
        /*
         *
         * Shorthand for performance.
         *
         */
        /** @inheritdoc */
        getOneToken() {
            return (this.index <= this.indexEnd) ? this.tokens[this.index] : null;
        }
        /** @inheritdoc */
        getAllTokens() {
            return this.tokens.slice(this.index, this.indexEnd + 1);
        }
    };
},
/* eslint/lib/source-code/token-store/limit-cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const DecorativeCursor = require(358) /* ./decorative-cursor */;
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The decorative cursor which limits the number of tokens.
     */
    module.exports = class LimitCursor extends DecorativeCursor {
        /**
         * Initializes this cursor.
         * @param {Cursor} cursor The cursor to be decorated.
         * @param {number} count The count of tokens this cursor iterates.
         */
        constructor(cursor, count) {
            super(cursor);
            this.count = count;
        }
        /** @inheritdoc */
        moveNext() {
            if (this.count > 0) {
                this.count -= 1;
                return super.moveNext();
            }
            return false;
        }
    };
},
/* eslint/lib/source-code/token-store/skip-cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const DecorativeCursor = require(358) /* ./decorative-cursor */;
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The decorative cursor which ignores the first few tokens.
     */
    module.exports = class SkipCursor extends DecorativeCursor {
        /**
         * Initializes this cursor.
         * @param {Cursor} cursor The cursor to be decorated.
         * @param {number} count The count of tokens this cursor skips.
         */
        constructor(cursor, count) {
            super(cursor);
            this.count = count;
        }
        /** @inheritdoc */
        moveNext() {
            while (this.count > 0) {
                this.count -= 1;
                if (!super.moveNext()) {
                    return false;
                }
            }
            return super.moveNext();
        }
    };
},
/* eslint/lib/source-code/token-store/padded-token-cursor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const ForwardTokenCursor = require(360) /* ./forward-token-cursor */;
    //------------------------------------------------------------------------------
    // Exports
    //------------------------------------------------------------------------------
    /**
     * The cursor which iterates tokens only, with inflated range.
     * This is for the backward compatibility of padding options.
     */
    module.exports = class PaddedTokenCursor extends ForwardTokenCursor {
        /**
         * Initializes this cursor.
         * @param {Token[]} tokens The array of tokens.
         * @param {Comment[]} comments The array of comments.
         * @param {Object} indexMap The map from locations to indices in `tokens`.
         * @param {number} startLoc The start location of the iteration range.
         * @param {number} endLoc The end location of the iteration range.
         * @param {number} beforeCount The number of tokens this cursor iterates before start.
         * @param {number} afterCount The number of tokens this cursor iterates after end.
         */
        constructor(tokens, comments, indexMap, startLoc, endLoc, beforeCount, afterCount) {
            super(tokens, comments, indexMap, startLoc, endLoc);
            this.index = Math.max(0, this.index - beforeCount);
            this.indexEnd = Math.min(tokens.length - 1, this.indexEnd + afterCount);
        }
    };
},
/* eslint/lib/linter/code-path-analysis/code-path-analyzer.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const assert = require("assert"), { breakableTypePattern } = require(346) /* ../../shared/ast-utils */, CodePath = require(365) /* ./code-path */, CodePathSegment = require(367) /* ./code-path-segment */, IdGenerator = require(370) /* ./id-generator */, debug = require(368) /* ./debug-helpers */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given node is a `case` node (not `default` node).
     * @param {ASTNode} node A `SwitchCase` node to check.
     * @returns {boolean} `true` if the node is a `case` node (not `default` node).
     */
    function isCaseNode(node) {
        return Boolean(node.test);
    }
    /**
     * Checks whether the given logical operator is taken into account for the code
     * path analysis.
     * @param {string} operator The operator found in the LogicalExpression node
     * @returns {boolean} `true` if the operator is "&&" or "||" or "??"
     */
    function isHandledLogicalOperator(operator) {
        return operator === "&&" || operator === "||" || operator === "??";
    }
    /**
     * Checks whether the given assignment operator is a logical assignment operator.
     * Logical assignments are taken into account for the code path analysis
     * because of their short-circuiting semantics.
     * @param {string} operator The operator found in the AssignmentExpression node
     * @returns {boolean} `true` if the operator is "&&=" or "||=" or "??="
     */
    function isLogicalAssignmentOperator(operator) {
        return operator === "&&=" || operator === "||=" || operator === "??=";
    }
    /**
     * Gets the label if the parent node of a given node is a LabeledStatement.
     * @param {ASTNode} node A node to get.
     * @returns {string|null} The label or `null`.
     */
    function getLabel(node) {
        if (node.parent.type === "LabeledStatement") {
            return node.parent.label.name;
        }
        return null;
    }
    /**
     * Checks whether or not a given logical expression node goes different path
     * between the `true` case and the `false` case.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node is a test of a choice statement.
     */
    function isForkingByTrueOrFalse(node) {
        const parent = node.parent;
        switch (parent.type) {
            case "ConditionalExpression":
            case "IfStatement":
            case "WhileStatement":
            case "DoWhileStatement":
            case "ForStatement":
                return parent.test === node;
            case "LogicalExpression":
                return isHandledLogicalOperator(parent.operator);
            case "AssignmentExpression":
                return isLogicalAssignmentOperator(parent.operator);
            default:
                return false;
        }
    }
    /**
     * Gets the boolean value of a given literal node.
     *
     * This is used to detect infinity loops (e.g. `while (true) {}`).
     * Statements preceded by an infinity loop are unreachable if the loop didn't
     * have any `break` statement.
     * @param {ASTNode} node A node to get.
     * @returns {boolean|undefined} a boolean value if the node is a Literal node,
     *   otherwise `undefined`.
     */
    function getBooleanValueIfSimpleConstant(node) {
        if (node.type === "Literal") {
            return Boolean(node.value);
        }
        return void 0;
    }
    /**
     * Checks that a given identifier node is a reference or not.
     *
     * This is used to detect the first throwable node in a `try` block.
     * @param {ASTNode} node An Identifier node to check.
     * @returns {boolean} `true` if the node is a reference.
     */
    function isIdentifierReference(node) {
        const parent = node.parent;
        switch (parent.type) {
            case "LabeledStatement":
            case "BreakStatement":
            case "ContinueStatement":
            case "ArrayPattern":
            case "RestElement":
            case "ImportSpecifier":
            case "ImportDefaultSpecifier":
            case "ImportNamespaceSpecifier":
            case "CatchClause":
                return false;
            case "FunctionDeclaration":
            case "FunctionExpression":
            case "ArrowFunctionExpression":
            case "ClassDeclaration":
            case "ClassExpression":
            case "VariableDeclarator":
                return parent.id !== node;
            case "Property":
            case "MethodDefinition":
                return (parent.key !== node ||
                    parent.computed ||
                    parent.shorthand);
            case "AssignmentPattern":
                return parent.key !== node;
            default:
                return true;
        }
    }
    /**
     * Updates the current segment with the head segment.
     * This is similar to local branches and tracking branches of git.
     *
     * To separate the current and the head is in order to not make useless segments.
     *
     * In this process, both "onCodePathSegmentStart" and "onCodePathSegmentEnd"
     * events are fired.
     * @param {CodePathAnalyzer} analyzer The instance.
     * @param {ASTNode} node The current AST node.
     * @returns {void}
     */
    function forwardCurrentToHead(analyzer, node) {
        const codePath = analyzer.codePath;
        const state = CodePath.getState(codePath);
        const currentSegments = state.currentSegments;
        const headSegments = state.headSegments;
        const end = Math.max(currentSegments.length, headSegments.length);
        let i, currentSegment, headSegment;
        // Fires leaving events.
        for (i = 0; i < end; ++i) {
            currentSegment = currentSegments[i];
            headSegment = headSegments[i];
            if (currentSegment !== headSegment && currentSegment) {
                debug.dump(`onCodePathSegmentEnd ${currentSegment.id}`);
                if (currentSegment.reachable) {
                    analyzer.emitter.emit("onCodePathSegmentEnd", currentSegment, node);
                }
            }
        }
        // Update state.
        state.currentSegments = headSegments;
        // Fires entering events.
        for (i = 0; i < end; ++i) {
            currentSegment = currentSegments[i];
            headSegment = headSegments[i];
            if (currentSegment !== headSegment && headSegment) {
                debug.dump(`onCodePathSegmentStart ${headSegment.id}`);
                CodePathSegment.markUsed(headSegment);
                if (headSegment.reachable) {
                    analyzer.emitter.emit("onCodePathSegmentStart", headSegment, node);
                }
            }
        }
    }
    /**
     * Updates the current segment with empty.
     * This is called at the last of functions or the program.
     * @param {CodePathAnalyzer} analyzer The instance.
     * @param {ASTNode} node The current AST node.
     * @returns {void}
     */
    function leaveFromCurrentSegment(analyzer, node) {
        const state = CodePath.getState(analyzer.codePath);
        const currentSegments = state.currentSegments;
        for (let i = 0; i < currentSegments.length; ++i) {
            const currentSegment = currentSegments[i];
            debug.dump(`onCodePathSegmentEnd ${currentSegment.id}`);
            if (currentSegment.reachable) {
                analyzer.emitter.emit("onCodePathSegmentEnd", currentSegment, node);
            }
        }
        state.currentSegments = [];
    }
    /**
     * Updates the code path due to the position of a given node in the parent node
     * thereof.
     *
     * For example, if the node is `parent.consequent`, this creates a fork from the
     * current path.
     * @param {CodePathAnalyzer} analyzer The instance.
     * @param {ASTNode} node The current AST node.
     * @returns {void}
     */
    function preprocess(analyzer, node) {
        const codePath = analyzer.codePath;
        const state = CodePath.getState(codePath);
        const parent = node.parent;
        switch (parent.type) {
            // The `arguments.length == 0` case is in `postprocess` function.
            case "CallExpression":
                if (parent.optional === true && parent.arguments.length >= 1 && parent.arguments[0] === node) {
                    state.makeOptionalRight();
                }
                break;
            case "MemberExpression":
                if (parent.optional === true && parent.property === node) {
                    state.makeOptionalRight();
                }
                break;
            case "LogicalExpression":
                if (parent.right === node &&
                    isHandledLogicalOperator(parent.operator)) {
                    state.makeLogicalRight();
                }
                break;
            case "AssignmentExpression":
                if (parent.right === node &&
                    isLogicalAssignmentOperator(parent.operator)) {
                    state.makeLogicalRight();
                }
                break;
            case "ConditionalExpression":
            case "IfStatement":
                /*
                 * Fork if this node is at `consequent`/`alternate`.
                 * `popForkContext()` exists at `IfStatement:exit` and
                 * `ConditionalExpression:exit`.
                 */
                if (parent.consequent === node) {
                    state.makeIfConsequent();
                }
                else if (parent.alternate === node) {
                    state.makeIfAlternate();
                }
                break;
            case "SwitchCase":
                if (parent.consequent[0] === node) {
                    state.makeSwitchCaseBody(false, !parent.test);
                }
                break;
            case "TryStatement":
                if (parent.handler === node) {
                    state.makeCatchBlock();
                }
                else if (parent.finalizer === node) {
                    state.makeFinallyBlock();
                }
                break;
            case "WhileStatement":
                if (parent.test === node) {
                    state.makeWhileTest(getBooleanValueIfSimpleConstant(node));
                }
                else {
                    assert(parent.body === node);
                    state.makeWhileBody();
                }
                break;
            case "DoWhileStatement":
                if (parent.body === node) {
                    state.makeDoWhileBody();
                }
                else {
                    assert(parent.test === node);
                    state.makeDoWhileTest(getBooleanValueIfSimpleConstant(node));
                }
                break;
            case "ForStatement":
                if (parent.test === node) {
                    state.makeForTest(getBooleanValueIfSimpleConstant(node));
                }
                else if (parent.update === node) {
                    state.makeForUpdate();
                }
                else if (parent.body === node) {
                    state.makeForBody();
                }
                break;
            case "ForInStatement":
            case "ForOfStatement":
                if (parent.left === node) {
                    state.makeForInOfLeft();
                }
                else if (parent.right === node) {
                    state.makeForInOfRight();
                }
                else {
                    assert(parent.body === node);
                    state.makeForInOfBody();
                }
                break;
            case "AssignmentPattern":
                /*
                 * Fork if this node is at `right`.
                 * `left` is executed always, so it uses the current path.
                 * `popForkContext()` exists at `AssignmentPattern:exit`.
                 */
                if (parent.right === node) {
                    state.pushForkContext();
                    state.forkBypassPath();
                    state.forkPath();
                }
                break;
            default:
                break;
        }
    }
    /**
     * Updates the code path due to the type of a given node in entering.
     * @param {CodePathAnalyzer} analyzer The instance.
     * @param {ASTNode} node The current AST node.
     * @returns {void}
     */
    function processCodePathToEnter(analyzer, node) {
        let codePath = analyzer.codePath;
        let state = codePath && CodePath.getState(codePath);
        const parent = node.parent;
        switch (node.type) {
            case "Program":
            case "FunctionDeclaration":
            case "FunctionExpression":
            case "ArrowFunctionExpression":
                if (codePath) {
                    // Emits onCodePathSegmentStart events if updated.
                    forwardCurrentToHead(analyzer, node);
                    debug.dumpState(node, state, false);
                }
                // Create the code path of this scope.
                codePath = analyzer.codePath = new CodePath(analyzer.idGenerator.next(), codePath, analyzer.onLooped);
                state = CodePath.getState(codePath);
                // Emits onCodePathStart events.
                debug.dump(`onCodePathStart ${codePath.id}`);
                analyzer.emitter.emit("onCodePathStart", codePath, node);
                break;
            case "ChainExpression":
                state.pushChainContext();
                break;
            case "CallExpression":
                if (node.optional === true) {
                    state.makeOptionalNode();
                }
                break;
            case "MemberExpression":
                if (node.optional === true) {
                    state.makeOptionalNode();
                }
                break;
            case "LogicalExpression":
                if (isHandledLogicalOperator(node.operator)) {
                    state.pushChoiceContext(node.operator, isForkingByTrueOrFalse(node));
                }
                break;
            case "AssignmentExpression":
                if (isLogicalAssignmentOperator(node.operator)) {
                    state.pushChoiceContext(node.operator.slice(0, -1), // removes `=` from the end
                    isForkingByTrueOrFalse(node));
                }
                break;
            case "ConditionalExpression":
            case "IfStatement":
                state.pushChoiceContext("test", false);
                break;
            case "SwitchStatement":
                state.pushSwitchContext(node.cases.some(isCaseNode), getLabel(node));
                break;
            case "TryStatement":
                state.pushTryContext(Boolean(node.finalizer));
                break;
            case "SwitchCase":
                /*
                 * Fork if this node is after the 2st node in `cases`.
                 * It's similar to `else` blocks.
                 * The next `test` node is processed in this path.
                 */
                if (parent.discriminant !== node && parent.cases[0] !== node) {
                    state.forkPath();
                }
                break;
            case "WhileStatement":
            case "DoWhileStatement":
            case "ForStatement":
            case "ForInStatement":
            case "ForOfStatement":
                state.pushLoopContext(node.type, getLabel(node));
                break;
            case "LabeledStatement":
                if (!breakableTypePattern.test(node.body.type)) {
                    state.pushBreakContext(false, node.label.name);
                }
                break;
            default:
                break;
        }
        // Emits onCodePathSegmentStart events if updated.
        forwardCurrentToHead(analyzer, node);
        debug.dumpState(node, state, false);
    }
    /**
     * Updates the code path due to the type of a given node in leaving.
     * @param {CodePathAnalyzer} analyzer The instance.
     * @param {ASTNode} node The current AST node.
     * @returns {void}
     */
    function processCodePathToExit(analyzer, node) {
        const codePath = analyzer.codePath;
        const state = CodePath.getState(codePath);
        let dontForward = false;
        switch (node.type) {
            case "ChainExpression":
                state.popChainContext();
                break;
            case "IfStatement":
            case "ConditionalExpression":
                state.popChoiceContext();
                break;
            case "LogicalExpression":
                if (isHandledLogicalOperator(node.operator)) {
                    state.popChoiceContext();
                }
                break;
            case "AssignmentExpression":
                if (isLogicalAssignmentOperator(node.operator)) {
                    state.popChoiceContext();
                }
                break;
            case "SwitchStatement":
                state.popSwitchContext();
                break;
            case "SwitchCase":
                /*
                 * This is the same as the process at the 1st `consequent` node in
                 * `preprocess` function.
                 * Must do if this `consequent` is empty.
                 */
                if (node.consequent.length === 0) {
                    state.makeSwitchCaseBody(true, !node.test);
                }
                if (state.forkContext.reachable) {
                    dontForward = true;
                }
                break;
            case "TryStatement":
                state.popTryContext();
                break;
            case "BreakStatement":
                forwardCurrentToHead(analyzer, node);
                state.makeBreak(node.label && node.label.name);
                dontForward = true;
                break;
            case "ContinueStatement":
                forwardCurrentToHead(analyzer, node);
                state.makeContinue(node.label && node.label.name);
                dontForward = true;
                break;
            case "ReturnStatement":
                forwardCurrentToHead(analyzer, node);
                state.makeReturn();
                dontForward = true;
                break;
            case "ThrowStatement":
                forwardCurrentToHead(analyzer, node);
                state.makeThrow();
                dontForward = true;
                break;
            case "Identifier":
                if (isIdentifierReference(node)) {
                    state.makeFirstThrowablePathInTryBlock();
                    dontForward = true;
                }
                break;
            case "CallExpression":
            case "ImportExpression":
            case "MemberExpression":
            case "NewExpression":
            case "YieldExpression":
                state.makeFirstThrowablePathInTryBlock();
                break;
            case "WhileStatement":
            case "DoWhileStatement":
            case "ForStatement":
            case "ForInStatement":
            case "ForOfStatement":
                state.popLoopContext();
                break;
            case "AssignmentPattern":
                state.popForkContext();
                break;
            case "LabeledStatement":
                if (!breakableTypePattern.test(node.body.type)) {
                    state.popBreakContext();
                }
                break;
            default:
                break;
        }
        // Emits onCodePathSegmentStart events if updated.
        if (!dontForward) {
            forwardCurrentToHead(analyzer, node);
        }
        debug.dumpState(node, state, true);
    }
    /**
     * Updates the code path to finalize the current code path.
     * @param {CodePathAnalyzer} analyzer The instance.
     * @param {ASTNode} node The current AST node.
     * @returns {void}
     */
    function postprocess(analyzer, node) {
        switch (node.type) {
            case "Program":
            case "FunctionDeclaration":
            case "FunctionExpression":
            case "ArrowFunctionExpression": {
                let codePath = analyzer.codePath;
                // Mark the current path as the final node.
                CodePath.getState(codePath).makeFinal();
                // Emits onCodePathSegmentEnd event of the current segments.
                leaveFromCurrentSegment(analyzer, node);
                // Emits onCodePathEnd event of this code path.
                debug.dump(`onCodePathEnd ${codePath.id}`);
                analyzer.emitter.emit("onCodePathEnd", codePath, node);
                debug.dumpDot(codePath);
                codePath = analyzer.codePath = analyzer.codePath.upper;
                if (codePath) {
                    debug.dumpState(node, CodePath.getState(codePath), true);
                }
                break;
            }
            // The `arguments.length >= 1` case is in `preprocess` function.
            case "CallExpression":
                if (node.optional === true && node.arguments.length === 0) {
                    CodePath.getState(analyzer.codePath).makeOptionalRight();
                }
                break;
            default:
                break;
        }
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * The class to analyze code paths.
     * This class implements the EventGenerator interface.
     */
    class CodePathAnalyzer {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {EventGenerator} eventGenerator An event generator to wrap.
         */
        constructor(eventGenerator) {
            this.original = eventGenerator;
            this.emitter = eventGenerator.emitter;
            this.codePath = null;
            this.idGenerator = new IdGenerator("s");
            this.currentNode = null;
            this.onLooped = this.onLooped.bind(this);
        }
        /**
         * Does the process to enter a given AST node.
         * This updates state of analysis and calls `enterNode` of the wrapped.
         * @param {ASTNode} node A node which is entering.
         * @returns {void}
         */
        enterNode(node) {
            this.currentNode = node;
            // Updates the code path due to node's position in its parent node.
            if (node.parent) {
                preprocess(this, node);
            }
            /*
             * Updates the code path.
             * And emits onCodePathStart/onCodePathSegmentStart events.
             */
            processCodePathToEnter(this, node);
            // Emits node events.
            this.original.enterNode(node);
            this.currentNode = null;
        }
        /**
         * Does the process to leave a given AST node.
         * This updates state of analysis and calls `leaveNode` of the wrapped.
         * @param {ASTNode} node A node which is leaving.
         * @returns {void}
         */
        leaveNode(node) {
            this.currentNode = node;
            /*
             * Updates the code path.
             * And emits onCodePathStart/onCodePathSegmentStart events.
             */
            processCodePathToExit(this, node);
            // Emits node events.
            this.original.leaveNode(node);
            // Emits the last onCodePathStart/onCodePathSegmentStart events.
            postprocess(this, node);
            this.currentNode = null;
        }
        /**
         * This is called on a code path looped.
         * Then this raises a looped event.
         * @param {CodePathSegment} fromSegment A segment of prev.
         * @param {CodePathSegment} toSegment A segment of next.
         * @returns {void}
         */
        onLooped(fromSegment, toSegment) {
            if (fromSegment.reachable && toSegment.reachable) {
                debug.dump(`onCodePathSegmentLoop ${fromSegment.id} -> ${toSegment.id}`);
                this.emitter.emit("onCodePathSegmentLoop", fromSegment, toSegment, this.currentNode);
            }
        }
    }
    module.exports = CodePathAnalyzer;
},
/* eslint/lib/linter/code-path-analysis/code-path.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const CodePathState = require(366) /* ./code-path-state */;
    const IdGenerator = require(370) /* ./id-generator */;
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * A code path.
     */
    class CodePath {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {string} id An identifier.
         * @param {CodePath|null} upper The code path of the upper function scope.
         * @param {Function} onLooped A callback function to notify looping.
         */
        constructor(id, upper, onLooped) {
            /**
             * The identifier of this code path.
             * Rules use it to store additional information of each rule.
             * @type {string}
             */
            this.id = id;
            /**
             * The code path of the upper function scope.
             * @type {CodePath|null}
             */
            this.upper = upper;
            /**
             * The code paths of nested function scopes.
             * @type {CodePath[]}
             */
            this.childCodePaths = [];
            // Initializes internal state.
            Object.defineProperty(this, "internal", { value: new CodePathState(new IdGenerator(`${id}_`), onLooped) });
            // Adds this into `childCodePaths` of `upper`.
            if (upper) {
                upper.childCodePaths.push(this);
            }
        }
        /**
         * Gets the state of a given code path.
         * @param {CodePath} codePath A code path to get.
         * @returns {CodePathState} The state of the code path.
         */
        static getState(codePath) {
            return codePath.internal;
        }
        /**
         * The initial code path segment.
         * @type {CodePathSegment}
         */
        get initialSegment() {
            return this.internal.initialSegment;
        }
        /**
         * Final code path segments.
         * This array is a mix of `returnedSegments` and `thrownSegments`.
         * @type {CodePathSegment[]}
         */
        get finalSegments() {
            return this.internal.finalSegments;
        }
        /**
         * Final code path segments which is with `return` statements.
         * This array contains the last path segment if it's reachable.
         * Since the reachable last path returns `undefined`.
         * @type {CodePathSegment[]}
         */
        get returnedSegments() {
            return this.internal.returnedForkContext;
        }
        /**
         * Final code path segments which is with `throw` statements.
         * @type {CodePathSegment[]}
         */
        get thrownSegments() {
            return this.internal.thrownForkContext;
        }
        /**
         * Current code path segments.
         * @type {CodePathSegment[]}
         */
        get currentSegments() {
            return this.internal.currentSegments;
        }
        /**
         * Traverses all segments in this code path.
         *
         *     codePath.traverseSegments(function(segment, controller) {
         *         // do something.
         *     });
         *
         * This method enumerates segments in order from the head.
         *
         * The `controller` object has two methods.
         *
         * - `controller.skip()` - Skip the following segments in this branch.
         * - `controller.break()` - Skip all following segments.
         * @param {Object} [options] Omittable.
         * @param {CodePathSegment} [options.first] The first segment to traverse.
         * @param {CodePathSegment} [options.last] The last segment to traverse.
         * @param {Function} callback A callback function.
         * @returns {void}
         */
        traverseSegments(options, callback) {
            let resolvedOptions;
            let resolvedCallback;
            if (typeof options === "function") {
                resolvedCallback = options;
                resolvedOptions = {};
            }
            else {
                resolvedOptions = options || {};
                resolvedCallback = callback;
            }
            const startSegment = resolvedOptions.first || this.internal.initialSegment;
            const lastSegment = resolvedOptions.last;
            let item = null;
            let index = 0;
            let end = 0;
            let segment = null;
            const visited = Object.create(null);
            const stack = [[startSegment, 0]];
            let skippedSegment = null;
            let broken = false;
            const controller = {
                skip() {
                    if (stack.length <= 1) {
                        broken = true;
                    }
                    else {
                        skippedSegment = stack[stack.length - 2][0];
                    }
                },
                break() {
                    broken = true;
                }
            };
            /**
             * Checks a given previous segment has been visited.
             * @param {CodePathSegment} prevSegment A previous segment to check.
             * @returns {boolean} `true` if the segment has been visited.
             */
            function isVisited(prevSegment) {
                return (visited[prevSegment.id] ||
                    segment.isLoopedPrevSegment(prevSegment));
            }
            while (stack.length > 0) {
                item = stack[stack.length - 1];
                segment = item[0];
                index = item[1];
                if (index === 0) {
                    // Skip if this segment has been visited already.
                    if (visited[segment.id]) {
                        stack.pop();
                        continue;
                    }
                    // Skip if all previous segments have not been visited.
                    if (segment !== startSegment &&
                        segment.prevSegments.length > 0 &&
                        !segment.prevSegments.every(isVisited)) {
                        stack.pop();
                        continue;
                    }
                    // Reset the flag of skipping if all branches have been skipped.
                    if (skippedSegment && segment.prevSegments.indexOf(skippedSegment) !== -1) {
                        skippedSegment = null;
                    }
                    visited[segment.id] = true;
                    // Call the callback when the first time.
                    if (!skippedSegment) {
                        resolvedCallback.call(this, segment, controller);
                        if (segment === lastSegment) {
                            controller.skip();
                        }
                        if (broken) {
                            break;
                        }
                    }
                }
                // Update the stack.
                end = segment.nextSegments.length - 1;
                if (index < end) {
                    item[1] += 1;
                    stack.push([segment.nextSegments[index], 0]);
                }
                else if (index === end) {
                    item[0] = segment.nextSegments[index];
                    item[1] = 0;
                }
                else {
                    stack.pop();
                }
            }
        }
    }
    module.exports = CodePath;
},
/* eslint/lib/linter/code-path-analysis/code-path-state.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const CodePathSegment = require(367) /* ./code-path-segment */, ForkContext = require(369) /* ./fork-context */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Adds given segments into the `dest` array.
     * If the `others` array does not includes the given segments, adds to the `all`
     * array as well.
     *
     * This adds only reachable and used segments.
     * @param {CodePathSegment[]} dest A destination array (`returnedSegments` or `thrownSegments`).
     * @param {CodePathSegment[]} others Another destination array (`returnedSegments` or `thrownSegments`).
     * @param {CodePathSegment[]} all The unified destination array (`finalSegments`).
     * @param {CodePathSegment[]} segments Segments to add.
     * @returns {void}
     */
    function addToReturnedOrThrown(dest, others, all, segments) {
        for (let i = 0; i < segments.length; ++i) {
            const segment = segments[i];
            dest.push(segment);
            if (others.indexOf(segment) === -1) {
                all.push(segment);
            }
        }
    }
    /**
     * Gets a loop-context for a `continue` statement.
     * @param {CodePathState} state A state to get.
     * @param {string} label The label of a `continue` statement.
     * @returns {LoopContext} A loop-context for a `continue` statement.
     */
    function getContinueContext(state, label) {
        if (!label) {
            return state.loopContext;
        }
        let context = state.loopContext;
        while (context) {
            if (context.label === label) {
                return context;
            }
            context = context.upper;
        }
        /* istanbul ignore next: foolproof (syntax error) */
        return null;
    }
    /**
     * Gets a context for a `break` statement.
     * @param {CodePathState} state A state to get.
     * @param {string} label The label of a `break` statement.
     * @returns {LoopContext|SwitchContext} A context for a `break` statement.
     */
    function getBreakContext(state, label) {
        let context = state.breakContext;
        while (context) {
            if (label ? context.label === label : context.breakable) {
                return context;
            }
            context = context.upper;
        }
        /* istanbul ignore next: foolproof (syntax error) */
        return null;
    }
    /**
     * Gets a context for a `return` statement.
     * @param {CodePathState} state A state to get.
     * @returns {TryContext|CodePathState} A context for a `return` statement.
     */
    function getReturnContext(state) {
        let context = state.tryContext;
        while (context) {
            if (context.hasFinalizer && context.position !== "finally") {
                return context;
            }
            context = context.upper;
        }
        return state;
    }
    /**
     * Gets a context for a `throw` statement.
     * @param {CodePathState} state A state to get.
     * @returns {TryContext|CodePathState} A context for a `throw` statement.
     */
    function getThrowContext(state) {
        let context = state.tryContext;
        while (context) {
            if (context.position === "try" ||
                (context.hasFinalizer && context.position === "catch")) {
                return context;
            }
            context = context.upper;
        }
        return state;
    }
    /**
     * Removes a given element from a given array.
     * @param {any[]} xs An array to remove the specific element.
     * @param {any} x An element to be removed.
     * @returns {void}
     */
    function remove(xs, x) {
        xs.splice(xs.indexOf(x), 1);
    }
    /**
     * Disconnect given segments.
     *
     * This is used in a process for switch statements.
     * If there is the "default" chunk before other cases, the order is different
     * between node's and running's.
     * @param {CodePathSegment[]} prevSegments Forward segments to disconnect.
     * @param {CodePathSegment[]} nextSegments Backward segments to disconnect.
     * @returns {void}
     */
    function removeConnection(prevSegments, nextSegments) {
        for (let i = 0; i < prevSegments.length; ++i) {
            const prevSegment = prevSegments[i];
            const nextSegment = nextSegments[i];
            remove(prevSegment.nextSegments, nextSegment);
            remove(prevSegment.allNextSegments, nextSegment);
            remove(nextSegment.prevSegments, prevSegment);
            remove(nextSegment.allPrevSegments, prevSegment);
        }
    }
    /**
     * Creates looping path.
     * @param {CodePathState} state The instance.
     * @param {CodePathSegment[]} unflattenedFromSegments Segments which are source.
     * @param {CodePathSegment[]} unflattenedToSegments Segments which are destination.
     * @returns {void}
     */
    function makeLooped(state, unflattenedFromSegments, unflattenedToSegments) {
        const fromSegments = CodePathSegment.flattenUnusedSegments(unflattenedFromSegments);
        const toSegments = CodePathSegment.flattenUnusedSegments(unflattenedToSegments);
        const end = Math.min(fromSegments.length, toSegments.length);
        for (let i = 0; i < end; ++i) {
            const fromSegment = fromSegments[i];
            const toSegment = toSegments[i];
            if (toSegment.reachable) {
                fromSegment.nextSegments.push(toSegment);
            }
            if (fromSegment.reachable) {
                toSegment.prevSegments.push(fromSegment);
            }
            fromSegment.allNextSegments.push(toSegment);
            toSegment.allPrevSegments.push(fromSegment);
            if (toSegment.allPrevSegments.length >= 2) {
                CodePathSegment.markPrevSegmentAsLooped(toSegment, fromSegment);
            }
            state.notifyLooped(fromSegment, toSegment);
        }
    }
    /**
     * Finalizes segments of `test` chunk of a ForStatement.
     *
     * - Adds `false` paths to paths which are leaving from the loop.
     * - Sets `true` paths to paths which go to the body.
     * @param {LoopContext} context A loop context to modify.
     * @param {ChoiceContext} choiceContext A choice context of this loop.
     * @param {CodePathSegment[]} head The current head paths.
     * @returns {void}
     */
    function finalizeTestSegmentsOfFor(context, choiceContext, head) {
        if (!choiceContext.processed) {
            choiceContext.trueForkContext.add(head);
            choiceContext.falseForkContext.add(head);
            choiceContext.qqForkContext.add(head);
        }
        if (context.test !== true) {
            context.brokenForkContext.addAll(choiceContext.falseForkContext);
        }
        context.endOfTestSegments = choiceContext.trueForkContext.makeNext(0, -1);
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * A class which manages state to analyze code paths.
     */
    class CodePathState {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {IdGenerator} idGenerator An id generator to generate id for code
         *   path segments.
         * @param {Function} onLooped A callback function to notify looping.
         */
        constructor(idGenerator, onLooped) {
            this.idGenerator = idGenerator;
            this.notifyLooped = onLooped;
            this.forkContext = ForkContext.newRoot(idGenerator);
            this.choiceContext = null;
            this.switchContext = null;
            this.tryContext = null;
            this.loopContext = null;
            this.breakContext = null;
            this.chainContext = null;
            this.currentSegments = [];
            this.initialSegment = this.forkContext.head[0];
            // returnedSegments and thrownSegments push elements into finalSegments also.
            const final = this.finalSegments = [];
            const returned = this.returnedForkContext = [];
            const thrown = this.thrownForkContext = [];
            returned.add = addToReturnedOrThrown.bind(null, returned, thrown, final);
            thrown.add = addToReturnedOrThrown.bind(null, thrown, returned, final);
        }
        /**
         * The head segments.
         * @type {CodePathSegment[]}
         */
        get headSegments() {
            return this.forkContext.head;
        }
        /**
         * The parent forking context.
         * This is used for the root of new forks.
         * @type {ForkContext}
         */
        get parentForkContext() {
            const current = this.forkContext;
            return current && current.upper;
        }
        /**
         * Creates and stacks new forking context.
         * @param {boolean} forkLeavingPath A flag which shows being in a
         *   "finally" block.
         * @returns {ForkContext} The created context.
         */
        pushForkContext(forkLeavingPath) {
            this.forkContext = ForkContext.newEmpty(this.forkContext, forkLeavingPath);
            return this.forkContext;
        }
        /**
         * Pops and merges the last forking context.
         * @returns {ForkContext} The last context.
         */
        popForkContext() {
            const lastContext = this.forkContext;
            this.forkContext = lastContext.upper;
            this.forkContext.replaceHead(lastContext.makeNext(0, -1));
            return lastContext;
        }
        /**
         * Creates a new path.
         * @returns {void}
         */
        forkPath() {
            this.forkContext.add(this.parentForkContext.makeNext(-1, -1));
        }
        /**
         * Creates a bypass path.
         * This is used for such as IfStatement which does not have "else" chunk.
         * @returns {void}
         */
        forkBypassPath() {
            this.forkContext.add(this.parentForkContext.head);
        }
        //--------------------------------------------------------------------------
        // ConditionalExpression, LogicalExpression, IfStatement
        //--------------------------------------------------------------------------
        /**
         * Creates a context for ConditionalExpression, LogicalExpression, AssignmentExpression (logical assignments only),
         * IfStatement, WhileStatement, DoWhileStatement, or ForStatement.
         *
         * LogicalExpressions have cases that it goes different paths between the
         * `true` case and the `false` case.
         *
         * For Example:
         *
         *     if (a || b) {
         *         foo();
         *     } else {
         *         bar();
         *     }
         *
         * In this case, `b` is evaluated always in the code path of the `else`
         * block, but it's not so in the code path of the `if` block.
         * So there are 3 paths.
         *
         *     a -> foo();
         *     a -> b -> foo();
         *     a -> b -> bar();
         * @param {string} kind A kind string.
         *   If the new context is LogicalExpression's or AssignmentExpression's, this is `"&&"` or `"||"` or `"??"`.
         *   If it's IfStatement's or ConditionalExpression's, this is `"test"`.
         *   Otherwise, this is `"loop"`.
         * @param {boolean} isForkingAsResult A flag that shows that goes different
         *   paths between `true` and `false`.
         * @returns {void}
         */
        pushChoiceContext(kind, isForkingAsResult) {
            this.choiceContext = {
                upper: this.choiceContext,
                kind,
                isForkingAsResult,
                trueForkContext: ForkContext.newEmpty(this.forkContext),
                falseForkContext: ForkContext.newEmpty(this.forkContext),
                qqForkContext: ForkContext.newEmpty(this.forkContext),
                processed: false
            };
        }
        /**
         * Pops the last choice context and finalizes it.
         * @returns {ChoiceContext} The popped context.
         */
        popChoiceContext() {
            const context = this.choiceContext;
            this.choiceContext = context.upper;
            const forkContext = this.forkContext;
            const headSegments = forkContext.head;
            switch (context.kind) {
                case "&&":
                case "||":
                case "??":
                    /*
                     * If any result were not transferred from child contexts,
                     * this sets the head segments to both cases.
                     * The head segments are the path of the right-hand operand.
                     */
                    if (!context.processed) {
                        context.trueForkContext.add(headSegments);
                        context.falseForkContext.add(headSegments);
                        context.qqForkContext.add(headSegments);
                    }
                    /*
                     * Transfers results to upper context if this context is in
                     * test chunk.
                     */
                    if (context.isForkingAsResult) {
                        const parentContext = this.choiceContext;
                        parentContext.trueForkContext.addAll(context.trueForkContext);
                        parentContext.falseForkContext.addAll(context.falseForkContext);
                        parentContext.qqForkContext.addAll(context.qqForkContext);
                        parentContext.processed = true;
                        return context;
                    }
                    break;
                case "test":
                    if (!context.processed) {
                        /*
                         * The head segments are the path of the `if` block here.
                         * Updates the `true` path with the end of the `if` block.
                         */
                        context.trueForkContext.clear();
                        context.trueForkContext.add(headSegments);
                    }
                    else {
                        /*
                         * The head segments are the path of the `else` block here.
                         * Updates the `false` path with the end of the `else`
                         * block.
                         */
                        context.falseForkContext.clear();
                        context.falseForkContext.add(headSegments);
                    }
                    break;
                case "loop":
                    /*
                     * Loops are addressed in popLoopContext().
                     * This is called from popLoopContext().
                     */
                    return context;
                /* istanbul ignore next */
                default:
                    throw new Error("unreachable");
            }
            // Merges all paths.
            const prevForkContext = context.trueForkContext;
            prevForkContext.addAll(context.falseForkContext);
            forkContext.replaceHead(prevForkContext.makeNext(0, -1));
            return context;
        }
        /**
         * Makes a code path segment of the right-hand operand of a logical
         * expression.
         * @returns {void}
         */
        makeLogicalRight() {
            const context = this.choiceContext;
            const forkContext = this.forkContext;
            if (context.processed) {
                /*
                 * This got segments already from the child choice context.
                 * Creates the next path from own true/false fork context.
                 */
                let prevForkContext;
                switch (context.kind) {
                    case "&&": // if true then go to the right-hand side.
                        prevForkContext = context.trueForkContext;
                        break;
                    case "||": // if false then go to the right-hand side.
                        prevForkContext = context.falseForkContext;
                        break;
                    case "??": // Both true/false can short-circuit, so needs the third path to go to the right-hand side. That's qqForkContext.
                        prevForkContext = context.qqForkContext;
                        break;
                    default:
                        throw new Error("unreachable");
                }
                forkContext.replaceHead(prevForkContext.makeNext(0, -1));
                prevForkContext.clear();
                context.processed = false;
            }
            else {
                /*
                 * This did not get segments from the child choice context.
                 * So addresses the head segments.
                 * The head segments are the path of the left-hand operand.
                 */
                switch (context.kind) {
                    case "&&": // the false path can short-circuit.
                        context.falseForkContext.add(forkContext.head);
                        break;
                    case "||": // the true path can short-circuit.
                        context.trueForkContext.add(forkContext.head);
                        break;
                    case "??": // both can short-circuit.
                        context.trueForkContext.add(forkContext.head);
                        context.falseForkContext.add(forkContext.head);
                        break;
                    default:
                        throw new Error("unreachable");
                }
                forkContext.replaceHead(forkContext.makeNext(-1, -1));
            }
        }
        /**
         * Makes a code path segment of the `if` block.
         * @returns {void}
         */
        makeIfConsequent() {
            const context = this.choiceContext;
            const forkContext = this.forkContext;
            /*
             * If any result were not transferred from child contexts,
             * this sets the head segments to both cases.
             * The head segments are the path of the test expression.
             */
            if (!context.processed) {
                context.trueForkContext.add(forkContext.head);
                context.falseForkContext.add(forkContext.head);
                context.qqForkContext.add(forkContext.head);
            }
            context.processed = false;
            // Creates new path from the `true` case.
            forkContext.replaceHead(context.trueForkContext.makeNext(0, -1));
        }
        /**
         * Makes a code path segment of the `else` block.
         * @returns {void}
         */
        makeIfAlternate() {
            const context = this.choiceContext;
            const forkContext = this.forkContext;
            /*
             * The head segments are the path of the `if` block.
             * Updates the `true` path with the end of the `if` block.
             */
            context.trueForkContext.clear();
            context.trueForkContext.add(forkContext.head);
            context.processed = true;
            // Creates new path from the `false` case.
            forkContext.replaceHead(context.falseForkContext.makeNext(0, -1));
        }
        //--------------------------------------------------------------------------
        // ChainExpression
        //--------------------------------------------------------------------------
        /**
         * Push a new `ChainExpression` context to the stack.
         * This method is called on entering to each `ChainExpression` node.
         * This context is used to count forking in the optional chain then merge them on the exiting from the `ChainExpression` node.
         * @returns {void}
         */
        pushChainContext() {
            this.chainContext = {
                upper: this.chainContext,
                countChoiceContexts: 0
            };
        }
        /**
         * Pop a `ChainExpression` context from the stack.
         * This method is called on exiting from each `ChainExpression` node.
         * This merges all forks of the last optional chaining.
         * @returns {void}
         */
        popChainContext() {
            const context = this.chainContext;
            this.chainContext = context.upper;
            // pop all choice contexts of this.
            for (let i = context.countChoiceContexts; i > 0; --i) {
                this.popChoiceContext();
            }
        }
        /**
         * Create a choice context for optional access.
         * This method is called on entering to each `(Call|Member)Expression[optional=true]` node.
         * This creates a choice context as similar to `LogicalExpression[operator="??"]` node.
         * @returns {void}
         */
        makeOptionalNode() {
            if (this.chainContext) {
                this.chainContext.countChoiceContexts += 1;
                this.pushChoiceContext("??", false);
            }
        }
        /**
         * Create a fork.
         * This method is called on entering to the `arguments|property` property of each `(Call|Member)Expression` node.
         * @returns {void}
         */
        makeOptionalRight() {
            if (this.chainContext) {
                this.makeLogicalRight();
            }
        }
        //--------------------------------------------------------------------------
        // SwitchStatement
        //--------------------------------------------------------------------------
        /**
         * Creates a context object of SwitchStatement and stacks it.
         * @param {boolean} hasCase `true` if the switch statement has one or more
         *   case parts.
         * @param {string|null} label The label text.
         * @returns {void}
         */
        pushSwitchContext(hasCase, label) {
            this.switchContext = {
                upper: this.switchContext,
                hasCase,
                defaultSegments: null,
                defaultBodySegments: null,
                foundDefault: false,
                lastIsDefault: false,
                countForks: 0
            };
            this.pushBreakContext(true, label);
        }
        /**
         * Pops the last context of SwitchStatement and finalizes it.
         *
         * - Disposes all forking stack for `case` and `default`.
         * - Creates the next code path segment from `context.brokenForkContext`.
         * - If the last `SwitchCase` node is not a `default` part, creates a path
         *   to the `default` body.
         * @returns {void}
         */
        popSwitchContext() {
            const context = this.switchContext;
            this.switchContext = context.upper;
            const forkContext = this.forkContext;
            const brokenForkContext = this.popBreakContext().brokenForkContext;
            if (context.countForks === 0) {
                /*
                 * When there is only one `default` chunk and there is one or more
                 * `break` statements, even if forks are nothing, it needs to merge
                 * those.
                 */
                if (!brokenForkContext.empty) {
                    brokenForkContext.add(forkContext.makeNext(-1, -1));
                    forkContext.replaceHead(brokenForkContext.makeNext(0, -1));
                }
                return;
            }
            const lastSegments = forkContext.head;
            this.forkBypassPath();
            const lastCaseSegments = forkContext.head;
            /*
             * `brokenForkContext` is used to make the next segment.
             * It must add the last segment into `brokenForkContext`.
             */
            brokenForkContext.add(lastSegments);
            /*
             * A path which is failed in all case test should be connected to path
             * of `default` chunk.
             */
            if (!context.lastIsDefault) {
                if (context.defaultBodySegments) {
                    /*
                     * Remove a link from `default` label to its chunk.
                     * It's false route.
                     */
                    removeConnection(context.defaultSegments, context.defaultBodySegments);
                    makeLooped(this, lastCaseSegments, context.defaultBodySegments);
                }
                else {
                    /*
                     * It handles the last case body as broken if `default` chunk
                     * does not exist.
                     */
                    brokenForkContext.add(lastCaseSegments);
                }
            }
            // Pops the segment context stack until the entry segment.
            for (let i = 0; i < context.countForks; ++i) {
                this.forkContext = this.forkContext.upper;
            }
            /*
             * Creates a path from all brokenForkContext paths.
             * This is a path after switch statement.
             */
            this.forkContext.replaceHead(brokenForkContext.makeNext(0, -1));
        }
        /**
         * Makes a code path segment for a `SwitchCase` node.
         * @param {boolean} isEmpty `true` if the body is empty.
         * @param {boolean} isDefault `true` if the body is the default case.
         * @returns {void}
         */
        makeSwitchCaseBody(isEmpty, isDefault) {
            const context = this.switchContext;
            if (!context.hasCase) {
                return;
            }
            /*
             * Merge forks.
             * The parent fork context has two segments.
             * Those are from the current case and the body of the previous case.
             */
            const parentForkContext = this.forkContext;
            const forkContext = this.pushForkContext();
            forkContext.add(parentForkContext.makeNext(0, -1));
            /*
             * Save `default` chunk info.
             * If the `default` label is not at the last, we must make a path from
             * the last `case` to the `default` chunk.
             */
            if (isDefault) {
                context.defaultSegments = parentForkContext.head;
                if (isEmpty) {
                    context.foundDefault = true;
                }
                else {
                    context.defaultBodySegments = forkContext.head;
                }
            }
            else {
                if (!isEmpty && context.foundDefault) {
                    context.foundDefault = false;
                    context.defaultBodySegments = forkContext.head;
                }
            }
            context.lastIsDefault = isDefault;
            context.countForks += 1;
        }
        //--------------------------------------------------------------------------
        // TryStatement
        //--------------------------------------------------------------------------
        /**
         * Creates a context object of TryStatement and stacks it.
         * @param {boolean} hasFinalizer `true` if the try statement has a
         *   `finally` block.
         * @returns {void}
         */
        pushTryContext(hasFinalizer) {
            this.tryContext = {
                upper: this.tryContext,
                position: "try",
                hasFinalizer,
                returnedForkContext: hasFinalizer
                    ? ForkContext.newEmpty(this.forkContext)
                    : null,
                thrownForkContext: ForkContext.newEmpty(this.forkContext),
                lastOfTryIsReachable: false,
                lastOfCatchIsReachable: false
            };
        }
        /**
         * Pops the last context of TryStatement and finalizes it.
         * @returns {void}
         */
        popTryContext() {
            const context = this.tryContext;
            this.tryContext = context.upper;
            if (context.position === "catch") {
                // Merges two paths from the `try` block and `catch` block merely.
                this.popForkContext();
                return;
            }
            /*
             * The following process is executed only when there is the `finally`
             * block.
             */
            const returned = context.returnedForkContext;
            const thrown = context.thrownForkContext;
            if (returned.empty && thrown.empty) {
                return;
            }
            // Separate head to normal paths and leaving paths.
            const headSegments = this.forkContext.head;
            this.forkContext = this.forkContext.upper;
            const normalSegments = headSegments.slice(0, headSegments.length / 2 | 0);
            const leavingSegments = headSegments.slice(headSegments.length / 2 | 0);
            // Forwards the leaving path to upper contexts.
            if (!returned.empty) {
                getReturnContext(this).returnedForkContext.add(leavingSegments);
            }
            if (!thrown.empty) {
                getThrowContext(this).thrownForkContext.add(leavingSegments);
            }
            // Sets the normal path as the next.
            this.forkContext.replaceHead(normalSegments);
            /*
             * If both paths of the `try` block and the `catch` block are
             * unreachable, the next path becomes unreachable as well.
             */
            if (!context.lastOfTryIsReachable && !context.lastOfCatchIsReachable) {
                this.forkContext.makeUnreachable();
            }
        }
        /**
         * Makes a code path segment for a `catch` block.
         * @returns {void}
         */
        makeCatchBlock() {
            const context = this.tryContext;
            const forkContext = this.forkContext;
            const thrown = context.thrownForkContext;
            // Update state.
            context.position = "catch";
            context.thrownForkContext = ForkContext.newEmpty(forkContext);
            context.lastOfTryIsReachable = forkContext.reachable;
            // Merge thrown paths.
            thrown.add(forkContext.head);
            const thrownSegments = thrown.makeNext(0, -1);
            // Fork to a bypass and the merged thrown path.
            this.pushForkContext();
            this.forkBypassPath();
            this.forkContext.add(thrownSegments);
        }
        /**
         * Makes a code path segment for a `finally` block.
         *
         * In the `finally` block, parallel paths are created. The parallel paths
         * are used as leaving-paths. The leaving-paths are paths from `return`
         * statements and `throw` statements in a `try` block or a `catch` block.
         * @returns {void}
         */
        makeFinallyBlock() {
            const context = this.tryContext;
            let forkContext = this.forkContext;
            const returned = context.returnedForkContext;
            const thrown = context.thrownForkContext;
            const headOfLeavingSegments = forkContext.head;
            // Update state.
            if (context.position === "catch") {
                // Merges two paths from the `try` block and `catch` block.
                this.popForkContext();
                forkContext = this.forkContext;
                context.lastOfCatchIsReachable = forkContext.reachable;
            }
            else {
                context.lastOfTryIsReachable = forkContext.reachable;
            }
            context.position = "finally";
            if (returned.empty && thrown.empty) {
                // This path does not leave.
                return;
            }
            /*
             * Create a parallel segment from merging returned and thrown.
             * This segment will leave at the end of this finally block.
             */
            const segments = forkContext.makeNext(-1, -1);
            for (let i = 0; i < forkContext.count; ++i) {
                const prevSegsOfLeavingSegment = [headOfLeavingSegments[i]];
                for (let j = 0; j < returned.segmentsList.length; ++j) {
                    prevSegsOfLeavingSegment.push(returned.segmentsList[j][i]);
                }
                for (let j = 0; j < thrown.segmentsList.length; ++j) {
                    prevSegsOfLeavingSegment.push(thrown.segmentsList[j][i]);
                }
                segments.push(CodePathSegment.newNext(this.idGenerator.next(), prevSegsOfLeavingSegment));
            }
            this.pushForkContext(true);
            this.forkContext.add(segments);
        }
        /**
         * Makes a code path segment from the first throwable node to the `catch`
         * block or the `finally` block.
         * @returns {void}
         */
        makeFirstThrowablePathInTryBlock() {
            const forkContext = this.forkContext;
            if (!forkContext.reachable) {
                return;
            }
            const context = getThrowContext(this);
            if (context === this ||
                context.position !== "try" ||
                !context.thrownForkContext.empty) {
                return;
            }
            context.thrownForkContext.add(forkContext.head);
            forkContext.replaceHead(forkContext.makeNext(-1, -1));
        }
        //--------------------------------------------------------------------------
        // Loop Statements
        //--------------------------------------------------------------------------
        /**
         * Creates a context object of a loop statement and stacks it.
         * @param {string} type The type of the node which was triggered. One of
         *   `WhileStatement`, `DoWhileStatement`, `ForStatement`, `ForInStatement`,
         *   and `ForStatement`.
         * @param {string|null} label A label of the node which was triggered.
         * @returns {void}
         */
        pushLoopContext(type, label) {
            const forkContext = this.forkContext;
            const breakContext = this.pushBreakContext(true, label);
            switch (type) {
                case "WhileStatement":
                    this.pushChoiceContext("loop", false);
                    this.loopContext = {
                        upper: this.loopContext,
                        type,
                        label,
                        test: void 0,
                        continueDestSegments: null,
                        brokenForkContext: breakContext.brokenForkContext
                    };
                    break;
                case "DoWhileStatement":
                    this.pushChoiceContext("loop", false);
                    this.loopContext = {
                        upper: this.loopContext,
                        type,
                        label,
                        test: void 0,
                        entrySegments: null,
                        continueForkContext: ForkContext.newEmpty(forkContext),
                        brokenForkContext: breakContext.brokenForkContext
                    };
                    break;
                case "ForStatement":
                    this.pushChoiceContext("loop", false);
                    this.loopContext = {
                        upper: this.loopContext,
                        type,
                        label,
                        test: void 0,
                        endOfInitSegments: null,
                        testSegments: null,
                        endOfTestSegments: null,
                        updateSegments: null,
                        endOfUpdateSegments: null,
                        continueDestSegments: null,
                        brokenForkContext: breakContext.brokenForkContext
                    };
                    break;
                case "ForInStatement":
                case "ForOfStatement":
                    this.loopContext = {
                        upper: this.loopContext,
                        type,
                        label,
                        prevSegments: null,
                        leftSegments: null,
                        endOfLeftSegments: null,
                        continueDestSegments: null,
                        brokenForkContext: breakContext.brokenForkContext
                    };
                    break;
                /* istanbul ignore next */
                default:
                    throw new Error(`unknown type: "${type}"`);
            }
        }
        /**
         * Pops the last context of a loop statement and finalizes it.
         * @returns {void}
         */
        popLoopContext() {
            const context = this.loopContext;
            this.loopContext = context.upper;
            const forkContext = this.forkContext;
            const brokenForkContext = this.popBreakContext().brokenForkContext;
            // Creates a looped path.
            switch (context.type) {
                case "WhileStatement":
                case "ForStatement":
                    this.popChoiceContext();
                    makeLooped(this, forkContext.head, context.continueDestSegments);
                    break;
                case "DoWhileStatement": {
                    const choiceContext = this.popChoiceContext();
                    if (!choiceContext.processed) {
                        choiceContext.trueForkContext.add(forkContext.head);
                        choiceContext.falseForkContext.add(forkContext.head);
                    }
                    if (context.test !== true) {
                        brokenForkContext.addAll(choiceContext.falseForkContext);
                    }
                    // `true` paths go to looping.
                    const segmentsList = choiceContext.trueForkContext.segmentsList;
                    for (let i = 0; i < segmentsList.length; ++i) {
                        makeLooped(this, segmentsList[i], context.entrySegments);
                    }
                    break;
                }
                case "ForInStatement":
                case "ForOfStatement":
                    brokenForkContext.add(forkContext.head);
                    makeLooped(this, forkContext.head, context.leftSegments);
                    break;
                /* istanbul ignore next */
                default:
                    throw new Error("unreachable");
            }
            // Go next.
            if (brokenForkContext.empty) {
                forkContext.replaceHead(forkContext.makeUnreachable(-1, -1));
            }
            else {
                forkContext.replaceHead(brokenForkContext.makeNext(0, -1));
            }
        }
        /**
         * Makes a code path segment for the test part of a WhileStatement.
         * @param {boolean|undefined} test The test value (only when constant).
         * @returns {void}
         */
        makeWhileTest(test) {
            const context = this.loopContext;
            const forkContext = this.forkContext;
            const testSegments = forkContext.makeNext(0, -1);
            // Update state.
            context.test = test;
            context.continueDestSegments = testSegments;
            forkContext.replaceHead(testSegments);
        }
        /**
         * Makes a code path segment for the body part of a WhileStatement.
         * @returns {void}
         */
        makeWhileBody() {
            const context = this.loopContext;
            const choiceContext = this.choiceContext;
            const forkContext = this.forkContext;
            if (!choiceContext.processed) {
                choiceContext.trueForkContext.add(forkContext.head);
                choiceContext.falseForkContext.add(forkContext.head);
            }
            // Update state.
            if (context.test !== true) {
                context.brokenForkContext.addAll(choiceContext.falseForkContext);
            }
            forkContext.replaceHead(choiceContext.trueForkContext.makeNext(0, -1));
        }
        /**
         * Makes a code path segment for the body part of a DoWhileStatement.
         * @returns {void}
         */
        makeDoWhileBody() {
            const context = this.loopContext;
            const forkContext = this.forkContext;
            const bodySegments = forkContext.makeNext(-1, -1);
            // Update state.
            context.entrySegments = bodySegments;
            forkContext.replaceHead(bodySegments);
        }
        /**
         * Makes a code path segment for the test part of a DoWhileStatement.
         * @param {boolean|undefined} test The test value (only when constant).
         * @returns {void}
         */
        makeDoWhileTest(test) {
            const context = this.loopContext;
            const forkContext = this.forkContext;
            context.test = test;
            // Creates paths of `continue` statements.
            if (!context.continueForkContext.empty) {
                context.continueForkContext.add(forkContext.head);
                const testSegments = context.continueForkContext.makeNext(0, -1);
                forkContext.replaceHead(testSegments);
            }
        }
        /**
         * Makes a code path segment for the test part of a ForStatement.
         * @param {boolean|undefined} test The test value (only when constant).
         * @returns {void}
         */
        makeForTest(test) {
            const context = this.loopContext;
            const forkContext = this.forkContext;
            const endOfInitSegments = forkContext.head;
            const testSegments = forkContext.makeNext(-1, -1);
            // Update state.
            context.test = test;
            context.endOfInitSegments = endOfInitSegments;
            context.continueDestSegments = context.testSegments = testSegments;
            forkContext.replaceHead(testSegments);
        }
        /**
         * Makes a code path segment for the update part of a ForStatement.
         * @returns {void}
         */
        makeForUpdate() {
            const context = this.loopContext;
            const choiceContext = this.choiceContext;
            const forkContext = this.forkContext;
            // Make the next paths of the test.
            if (context.testSegments) {
                finalizeTestSegmentsOfFor(context, choiceContext, forkContext.head);
            }
            else {
                context.endOfInitSegments = forkContext.head;
            }
            // Update state.
            const updateSegments = forkContext.makeDisconnected(-1, -1);
            context.continueDestSegments = context.updateSegments = updateSegments;
            forkContext.replaceHead(updateSegments);
        }
        /**
         * Makes a code path segment for the body part of a ForStatement.
         * @returns {void}
         */
        makeForBody() {
            const context = this.loopContext;
            const choiceContext = this.choiceContext;
            const forkContext = this.forkContext;
            // Update state.
            if (context.updateSegments) {
                context.endOfUpdateSegments = forkContext.head;
                // `update` -> `test`
                if (context.testSegments) {
                    makeLooped(this, context.endOfUpdateSegments, context.testSegments);
                }
            }
            else if (context.testSegments) {
                finalizeTestSegmentsOfFor(context, choiceContext, forkContext.head);
            }
            else {
                context.endOfInitSegments = forkContext.head;
            }
            let bodySegments = context.endOfTestSegments;
            if (!bodySegments) {
                /*
                 * If there is not the `test` part, the `body` path comes from the
                 * `init` part and the `update` part.
                 */
                const prevForkContext = ForkContext.newEmpty(forkContext);
                prevForkContext.add(context.endOfInitSegments);
                if (context.endOfUpdateSegments) {
                    prevForkContext.add(context.endOfUpdateSegments);
                }
                bodySegments = prevForkContext.makeNext(0, -1);
            }
            context.continueDestSegments = context.continueDestSegments || bodySegments;
            forkContext.replaceHead(bodySegments);
        }
        /**
         * Makes a code path segment for the left part of a ForInStatement and a
         * ForOfStatement.
         * @returns {void}
         */
        makeForInOfLeft() {
            const context = this.loopContext;
            const forkContext = this.forkContext;
            const leftSegments = forkContext.makeDisconnected(-1, -1);
            // Update state.
            context.prevSegments = forkContext.head;
            context.leftSegments = context.continueDestSegments = leftSegments;
            forkContext.replaceHead(leftSegments);
        }
        /**
         * Makes a code path segment for the right part of a ForInStatement and a
         * ForOfStatement.
         * @returns {void}
         */
        makeForInOfRight() {
            const context = this.loopContext;
            const forkContext = this.forkContext;
            const temp = ForkContext.newEmpty(forkContext);
            temp.add(context.prevSegments);
            const rightSegments = temp.makeNext(-1, -1);
            // Update state.
            context.endOfLeftSegments = forkContext.head;
            forkContext.replaceHead(rightSegments);
        }
        /**
         * Makes a code path segment for the body part of a ForInStatement and a
         * ForOfStatement.
         * @returns {void}
         */
        makeForInOfBody() {
            const context = this.loopContext;
            const forkContext = this.forkContext;
            const temp = ForkContext.newEmpty(forkContext);
            temp.add(context.endOfLeftSegments);
            const bodySegments = temp.makeNext(-1, -1);
            // Make a path: `right` -> `left`.
            makeLooped(this, forkContext.head, context.leftSegments);
            // Update state.
            context.brokenForkContext.add(forkContext.head);
            forkContext.replaceHead(bodySegments);
        }
        //--------------------------------------------------------------------------
        // Control Statements
        //--------------------------------------------------------------------------
        /**
         * Creates new context for BreakStatement.
         * @param {boolean} breakable The flag to indicate it can break by
         *      an unlabeled BreakStatement.
         * @param {string|null} label The label of this context.
         * @returns {Object} The new context.
         */
        pushBreakContext(breakable, label) {
            this.breakContext = {
                upper: this.breakContext,
                breakable,
                label,
                brokenForkContext: ForkContext.newEmpty(this.forkContext)
            };
            return this.breakContext;
        }
        /**
         * Removes the top item of the break context stack.
         * @returns {Object} The removed context.
         */
        popBreakContext() {
            const context = this.breakContext;
            const forkContext = this.forkContext;
            this.breakContext = context.upper;
            // Process this context here for other than switches and loops.
            if (!context.breakable) {
                const brokenForkContext = context.brokenForkContext;
                if (!brokenForkContext.empty) {
                    brokenForkContext.add(forkContext.head);
                    forkContext.replaceHead(brokenForkContext.makeNext(0, -1));
                }
            }
            return context;
        }
        /**
         * Makes a path for a `break` statement.
         *
         * It registers the head segment to a context of `break`.
         * It makes new unreachable segment, then it set the head with the segment.
         * @param {string} label A label of the break statement.
         * @returns {void}
         */
        makeBreak(label) {
            const forkContext = this.forkContext;
            if (!forkContext.reachable) {
                return;
            }
            const context = getBreakContext(this, label);
            /* istanbul ignore else: foolproof (syntax error) */
            if (context) {
                context.brokenForkContext.add(forkContext.head);
            }
            forkContext.replaceHead(forkContext.makeUnreachable(-1, -1));
        }
        /**
         * Makes a path for a `continue` statement.
         *
         * It makes a looping path.
         * It makes new unreachable segment, then it set the head with the segment.
         * @param {string} label A label of the continue statement.
         * @returns {void}
         */
        makeContinue(label) {
            const forkContext = this.forkContext;
            if (!forkContext.reachable) {
                return;
            }
            const context = getContinueContext(this, label);
            /* istanbul ignore else: foolproof (syntax error) */
            if (context) {
                if (context.continueDestSegments) {
                    makeLooped(this, forkContext.head, context.continueDestSegments);
                    // If the context is a for-in/of loop, this effects a break also.
                    if (context.type === "ForInStatement" ||
                        context.type === "ForOfStatement") {
                        context.brokenForkContext.add(forkContext.head);
                    }
                }
                else {
                    context.continueForkContext.add(forkContext.head);
                }
            }
            forkContext.replaceHead(forkContext.makeUnreachable(-1, -1));
        }
        /**
         * Makes a path for a `return` statement.
         *
         * It registers the head segment to a context of `return`.
         * It makes new unreachable segment, then it set the head with the segment.
         * @returns {void}
         */
        makeReturn() {
            const forkContext = this.forkContext;
            if (forkContext.reachable) {
                getReturnContext(this).returnedForkContext.add(forkContext.head);
                forkContext.replaceHead(forkContext.makeUnreachable(-1, -1));
            }
        }
        /**
         * Makes a path for a `throw` statement.
         *
         * It registers the head segment to a context of `throw`.
         * It makes new unreachable segment, then it set the head with the segment.
         * @returns {void}
         */
        makeThrow() {
            const forkContext = this.forkContext;
            if (forkContext.reachable) {
                getThrowContext(this).thrownForkContext.add(forkContext.head);
                forkContext.replaceHead(forkContext.makeUnreachable(-1, -1));
            }
        }
        /**
         * Makes the final path.
         * @returns {void}
         */
        makeFinal() {
            const segments = this.currentSegments;
            if (segments.length > 0 && segments[0].reachable) {
                this.returnedForkContext.add(segments);
            }
        }
    }
    module.exports = CodePathState;
},
/* eslint/lib/linter/code-path-analysis/code-path-segment.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const debug = require(368) /* ./debug-helpers */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given segment is reachable.
     * @param {CodePathSegment} segment A segment to check.
     * @returns {boolean} `true` if the segment is reachable.
     */
    function isReachable(segment) {
        return segment.reachable;
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * A code path segment.
     */
    class CodePathSegment {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {string} id An identifier.
         * @param {CodePathSegment[]} allPrevSegments An array of the previous segments.
         *   This array includes unreachable segments.
         * @param {boolean} reachable A flag which shows this is reachable.
         */
        constructor(id, allPrevSegments, reachable) {
            /**
             * The identifier of this code path.
             * Rules use it to store additional information of each rule.
             * @type {string}
             */
            this.id = id;
            /**
             * An array of the next segments.
             * @type {CodePathSegment[]}
             */
            this.nextSegments = [];
            /**
             * An array of the previous segments.
             * @type {CodePathSegment[]}
             */
            this.prevSegments = allPrevSegments.filter(isReachable);
            /**
             * An array of the next segments.
             * This array includes unreachable segments.
             * @type {CodePathSegment[]}
             */
            this.allNextSegments = [];
            /**
             * An array of the previous segments.
             * This array includes unreachable segments.
             * @type {CodePathSegment[]}
             */
            this.allPrevSegments = allPrevSegments;
            /**
             * A flag which shows this is reachable.
             * @type {boolean}
             */
            this.reachable = reachable;
            // Internal data.
            Object.defineProperty(this, "internal", {
                value: {
                    used: false,
                    loopedPrevSegments: []
                }
            });
            /* istanbul ignore if */
            if (debug.enabled) {
                this.internal.nodes = [];
            }
        }
        /**
         * Checks a given previous segment is coming from the end of a loop.
         * @param {CodePathSegment} segment A previous segment to check.
         * @returns {boolean} `true` if the segment is coming from the end of a loop.
         */
        isLoopedPrevSegment(segment) {
            return this.internal.loopedPrevSegments.indexOf(segment) !== -1;
        }
        /**
         * Creates the root segment.
         * @param {string} id An identifier.
         * @returns {CodePathSegment} The created segment.
         */
        static newRoot(id) {
            return new CodePathSegment(id, [], true);
        }
        /**
         * Creates a segment that follows given segments.
         * @param {string} id An identifier.
         * @param {CodePathSegment[]} allPrevSegments An array of the previous segments.
         * @returns {CodePathSegment} The created segment.
         */
        static newNext(id, allPrevSegments) {
            return new CodePathSegment(id, CodePathSegment.flattenUnusedSegments(allPrevSegments), allPrevSegments.some(isReachable));
        }
        /**
         * Creates an unreachable segment that follows given segments.
         * @param {string} id An identifier.
         * @param {CodePathSegment[]} allPrevSegments An array of the previous segments.
         * @returns {CodePathSegment} The created segment.
         */
        static newUnreachable(id, allPrevSegments) {
            const segment = new CodePathSegment(id, CodePathSegment.flattenUnusedSegments(allPrevSegments), false);
            /*
             * In `if (a) return a; foo();` case, the unreachable segment preceded by
             * the return statement is not used but must not be remove.
             */
            CodePathSegment.markUsed(segment);
            return segment;
        }
        /**
         * Creates a segment that follows given segments.
         * This factory method does not connect with `allPrevSegments`.
         * But this inherits `reachable` flag.
         * @param {string} id An identifier.
         * @param {CodePathSegment[]} allPrevSegments An array of the previous segments.
         * @returns {CodePathSegment} The created segment.
         */
        static newDisconnected(id, allPrevSegments) {
            return new CodePathSegment(id, [], allPrevSegments.some(isReachable));
        }
        /**
         * Makes a given segment being used.
         *
         * And this function registers the segment into the previous segments as a next.
         * @param {CodePathSegment} segment A segment to mark.
         * @returns {void}
         */
        static markUsed(segment) {
            if (segment.internal.used) {
                return;
            }
            segment.internal.used = true;
            let i;
            if (segment.reachable) {
                for (i = 0; i < segment.allPrevSegments.length; ++i) {
                    const prevSegment = segment.allPrevSegments[i];
                    prevSegment.allNextSegments.push(segment);
                    prevSegment.nextSegments.push(segment);
                }
            }
            else {
                for (i = 0; i < segment.allPrevSegments.length; ++i) {
                    segment.allPrevSegments[i].allNextSegments.push(segment);
                }
            }
        }
        /**
         * Marks a previous segment as looped.
         * @param {CodePathSegment} segment A segment.
         * @param {CodePathSegment} prevSegment A previous segment to mark.
         * @returns {void}
         */
        static markPrevSegmentAsLooped(segment, prevSegment) {
            segment.internal.loopedPrevSegments.push(prevSegment);
        }
        /**
         * Replaces unused segments with the previous segments of each unused segment.
         * @param {CodePathSegment[]} segments An array of segments to replace.
         * @returns {CodePathSegment[]} The replaced array.
         */
        static flattenUnusedSegments(segments) {
            const done = Object.create(null);
            const retv = [];
            for (let i = 0; i < segments.length; ++i) {
                const segment = segments[i];
                // Ignores duplicated.
                if (done[segment.id]) {
                    continue;
                }
                // Use previous segments if unused.
                if (!segment.internal.used) {
                    for (let j = 0; j < segment.allPrevSegments.length; ++j) {
                        const prevSegment = segment.allPrevSegments[j];
                        if (!done[prevSegment.id]) {
                            done[prevSegment.id] = true;
                            retv.push(prevSegment);
                        }
                    }
                }
                else {
                    done[segment.id] = true;
                    retv.push(segment);
                }
            }
            return retv;
        }
    }
    module.exports = CodePathSegment;
},
/* eslint/lib/linter/code-path-analysis/debug-helpers.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const debug = require(72) /* debug */("eslint:code-path");
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Gets id of a given segment.
     * @param {CodePathSegment} segment A segment to get.
     * @returns {string} Id of the segment.
     */
    /* istanbul ignore next */
    function getId(segment) {
        return segment.id + (segment.reachable ? "" : "!");
    }
    /**
     * Get string for the given node and operation.
     * @param {ASTNode} node The node to convert.
     * @param {"enter" | "exit" | undefined} label The operation label.
     * @returns {string} The string representation.
     */
    function nodeToString(node, label) {
        const suffix = label ? `:${label}` : "";
        switch (node.type) {
            case "Identifier": return `${node.type}${suffix} (${node.name})`;
            case "Literal": return `${node.type}${suffix} (${node.value})`;
            default: return `${node.type}${suffix}`;
        }
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = {
        /**
         * A flag that debug dumping is enabled or not.
         * @type {boolean}
         */
        enabled: debug.enabled,
        /**
         * Dumps given objects.
         * @param {...any} args objects to dump.
         * @returns {void}
         */
        dump: debug,
        /**
         * Dumps the current analyzing state.
         * @param {ASTNode} node A node to dump.
         * @param {CodePathState} state A state to dump.
         * @param {boolean} leaving A flag whether or not it's leaving
         * @returns {void}
         */
        dumpState: !debug.enabled ? debug : /* istanbul ignore next */ function (node, state, leaving) {
            for (let i = 0; i < state.currentSegments.length; ++i) {
                const segInternal = state.currentSegments[i].internal;
                if (leaving) {
                    const last = segInternal.nodes.length - 1;
                    if (last >= 0 && segInternal.nodes[last] === nodeToString(node, "enter")) {
                        segInternal.nodes[last] = nodeToString(node, void 0);
                    }
                    else {
                        segInternal.nodes.push(nodeToString(node, "exit"));
                    }
                }
                else {
                    segInternal.nodes.push(nodeToString(node, "enter"));
                }
            }
            debug([
                `${state.currentSegments.map(getId).join(",")})`,
                `${node.type}${leaving ? ":exit" : ""}`
            ].join(" "));
        },
        /**
         * Dumps a DOT code of a given code path.
         * The DOT code can be visualized with Graphvis.
         * @param {CodePath} codePath A code path to dump.
         * @returns {void}
         * @see http://www.graphviz.org
         * @see http://www.webgraphviz.com
         */
        dumpDot: !debug.enabled ? debug : /* istanbul ignore next */ function (codePath) {
            let text = "\n" +
                "digraph {\n" +
                "node[shape=box,style=\"rounded,filled\",fillcolor=white];\n" +
                "initial[label=\"\",shape=circle,style=filled,fillcolor=black,width=0.25,height=0.25];\n";
            if (codePath.returnedSegments.length > 0) {
                text += "final[label=\"\",shape=doublecircle,style=filled,fillcolor=black,width=0.25,height=0.25];\n";
            }
            if (codePath.thrownSegments.length > 0) {
                text += "thrown[label=\"✘\",shape=circle,width=0.3,height=0.3,fixedsize];\n";
            }
            const traceMap = Object.create(null);
            const arrows = this.makeDotArrows(codePath, traceMap);
            for (const id in traceMap) { // eslint-disable-line guard-for-in
                const segment = traceMap[id];
                text += `${id}[`;
                if (segment.reachable) {
                    text += "label=\"";
                }
                else {
                    text += "style=\"rounded,dashed,filled\",fillcolor=\"#FF9800\",label=\"<<unreachable>>\\n";
                }
                if (segment.internal.nodes.length > 0) {
                    text += segment.internal.nodes.join("\\n");
                }
                else {
                    text += "????";
                }
                text += "\"];\n";
            }
            text += `${arrows}\n`;
            text += "}";
            debug("DOT", text);
        },
        /**
         * Makes a DOT code of a given code path.
         * The DOT code can be visualized with Graphvis.
         * @param {CodePath} codePath A code path to make DOT.
         * @param {Object} traceMap Optional. A map to check whether or not segments had been done.
         * @returns {string} A DOT code of the code path.
         */
        makeDotArrows(codePath, traceMap) {
            const stack = [[codePath.initialSegment, 0]];
            const done = traceMap || Object.create(null);
            let lastId = codePath.initialSegment.id;
            let text = `initial->${codePath.initialSegment.id}`;
            while (stack.length > 0) {
                const item = stack.pop();
                const segment = item[0];
                const index = item[1];
                if (done[segment.id] && index === 0) {
                    continue;
                }
                done[segment.id] = segment;
                const nextSegment = segment.allNextSegments[index];
                if (!nextSegment) {
                    continue;
                }
                if (lastId === segment.id) {
                    text += `->${nextSegment.id}`;
                }
                else {
                    text += `;\n${segment.id}->${nextSegment.id}`;
                }
                lastId = nextSegment.id;
                stack.unshift([segment, 1 + index]);
                stack.push([nextSegment, 0]);
            }
            codePath.returnedSegments.forEach(finalSegment => {
                if (lastId === finalSegment.id) {
                    text += "->final";
                }
                else {
                    text += `;\n${finalSegment.id}->final`;
                }
                lastId = null;
            });
            codePath.thrownSegments.forEach(finalSegment => {
                if (lastId === finalSegment.id) {
                    text += "->thrown";
                }
                else {
                    text += `;\n${finalSegment.id}->thrown`;
                }
                lastId = null;
            });
            return `${text};`;
        }
    };
},
/* eslint/lib/linter/code-path-analysis/fork-context.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const assert = require("assert"), CodePathSegment = require(367) /* ./code-path-segment */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Gets whether or not a given segment is reachable.
     * @param {CodePathSegment} segment A segment to get.
     * @returns {boolean} `true` if the segment is reachable.
     */
    function isReachable(segment) {
        return segment.reachable;
    }
    /**
     * Creates new segments from the specific range of `context.segmentsList`.
     *
     * When `context.segmentsList` is `[[a, b], [c, d], [e, f]]`, `begin` is `0`, and
     * `end` is `-1`, this creates `[g, h]`. This `g` is from `a`, `c`, and `e`.
     * This `h` is from `b`, `d`, and `f`.
     * @param {ForkContext} context An instance.
     * @param {number} begin The first index of the previous segments.
     * @param {number} end The last index of the previous segments.
     * @param {Function} create A factory function of new segments.
     * @returns {CodePathSegment[]} New segments.
     */
    function makeSegments(context, begin, end, create) {
        const list = context.segmentsList;
        const normalizedBegin = begin >= 0 ? begin : list.length + begin;
        const normalizedEnd = end >= 0 ? end : list.length + end;
        const segments = [];
        for (let i = 0; i < context.count; ++i) {
            const allPrevSegments = [];
            for (let j = normalizedBegin; j <= normalizedEnd; ++j) {
                allPrevSegments.push(list[j][i]);
            }
            segments.push(create(context.idGenerator.next(), allPrevSegments));
        }
        return segments;
    }
    /**
     * `segments` becomes doubly in a `finally` block. Then if a code path exits by a
     * control statement (such as `break`, `continue`) from the `finally` block, the
     * destination's segments may be half of the source segments. In that case, this
     * merges segments.
     * @param {ForkContext} context An instance.
     * @param {CodePathSegment[]} segments Segments to merge.
     * @returns {CodePathSegment[]} The merged segments.
     */
    function mergeExtraSegments(context, segments) {
        let currentSegments = segments;
        while (currentSegments.length > context.count) {
            const merged = [];
            for (let i = 0, length = currentSegments.length / 2 | 0; i < length; ++i) {
                merged.push(CodePathSegment.newNext(context.idGenerator.next(), [currentSegments[i], currentSegments[i + length]]));
            }
            currentSegments = merged;
        }
        return currentSegments;
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * A class to manage forking.
     */
    class ForkContext {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {IdGenerator} idGenerator An identifier generator for segments.
         * @param {ForkContext|null} upper An upper fork context.
         * @param {number} count A number of parallel segments.
         */
        constructor(idGenerator, upper, count) {
            this.idGenerator = idGenerator;
            this.upper = upper;
            this.count = count;
            this.segmentsList = [];
        }
        /**
         * The head segments.
         * @type {CodePathSegment[]}
         */
        get head() {
            const list = this.segmentsList;
            return list.length === 0 ? [] : list[list.length - 1];
        }
        /**
         * A flag which shows empty.
         * @type {boolean}
         */
        get empty() {
            return this.segmentsList.length === 0;
        }
        /**
         * A flag which shows reachable.
         * @type {boolean}
         */
        get reachable() {
            const segments = this.head;
            return segments.length > 0 && segments.some(isReachable);
        }
        /**
         * Creates new segments from this context.
         * @param {number} begin The first index of previous segments.
         * @param {number} end The last index of previous segments.
         * @returns {CodePathSegment[]} New segments.
         */
        makeNext(begin, end) {
            return makeSegments(this, begin, end, CodePathSegment.newNext);
        }
        /**
         * Creates new segments from this context.
         * The new segments is always unreachable.
         * @param {number} begin The first index of previous segments.
         * @param {number} end The last index of previous segments.
         * @returns {CodePathSegment[]} New segments.
         */
        makeUnreachable(begin, end) {
            return makeSegments(this, begin, end, CodePathSegment.newUnreachable);
        }
        /**
         * Creates new segments from this context.
         * The new segments don't have connections for previous segments.
         * But these inherit the reachable flag from this context.
         * @param {number} begin The first index of previous segments.
         * @param {number} end The last index of previous segments.
         * @returns {CodePathSegment[]} New segments.
         */
        makeDisconnected(begin, end) {
            return makeSegments(this, begin, end, CodePathSegment.newDisconnected);
        }
        /**
         * Adds segments into this context.
         * The added segments become the head.
         * @param {CodePathSegment[]} segments Segments to add.
         * @returns {void}
         */
        add(segments) {
            assert(segments.length >= this.count, `${segments.length} >= ${this.count}`);
            this.segmentsList.push(mergeExtraSegments(this, segments));
        }
        /**
         * Replaces the head segments with given segments.
         * The current head segments are removed.
         * @param {CodePathSegment[]} segments Segments to add.
         * @returns {void}
         */
        replaceHead(segments) {
            assert(segments.length >= this.count, `${segments.length} >= ${this.count}`);
            this.segmentsList.splice(-1, 1, mergeExtraSegments(this, segments));
        }
        /**
         * Adds all segments of a given fork context into this context.
         * @param {ForkContext} context A fork context to add.
         * @returns {void}
         */
        addAll(context) {
            assert(context.count === this.count);
            const source = context.segmentsList;
            for (let i = 0; i < source.length; ++i) {
                this.segmentsList.push(source[i]);
            }
        }
        /**
         * Clears all segments in this context.
         * @returns {void}
         */
        clear() {
            this.segmentsList = [];
        }
        /**
         * Creates the root fork context.
         * @param {IdGenerator} idGenerator An identifier generator for segments.
         * @returns {ForkContext} New fork context.
         */
        static newRoot(idGenerator) {
            const context = new ForkContext(idGenerator, null, 1);
            context.add([CodePathSegment.newRoot(idGenerator.next())]);
            return context;
        }
        /**
         * Creates an empty fork context preceded by a given context.
         * @param {ForkContext} parentContext The parent fork context.
         * @param {boolean} forkLeavingPath A flag which shows inside of `finally` block.
         * @returns {ForkContext} New fork context.
         */
        static newEmpty(parentContext, forkLeavingPath) {
            return new ForkContext(parentContext.idGenerator, parentContext, (forkLeavingPath ? 2 : 1) * parentContext.count);
        }
    }
    module.exports = ForkContext;
},
/* eslint/lib/linter/code-path-analysis/id-generator.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * A generator for unique ids.
     */
    class IdGenerator {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {string} prefix Optional. A prefix of generated ids.
         */
        constructor(prefix) {
            this.prefix = String(prefix);
            this.n = 0;
        }
        /**
         * Generates id.
         * @returns {string} A generated id.
         */
        next() {
            this.n = 1 + this.n | 0;
            /* istanbul ignore if */
            if (this.n < 0) {
                this.n = 1;
            }
            return this.prefix + this.n;
        }
    }
    module.exports = IdGenerator;
},
/* eslint/lib/linter/apply-disable-directives.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Compares the locations of two objects in a source file
     * @param {{line: number, column: number}} itemA The first object
     * @param {{line: number, column: number}} itemB The second object
     * @returns {number} A value less than 1 if itemA appears before itemB in the source file, greater than 1 if
     * itemA appears after itemB in the source file, or 0 if itemA and itemB have the same location.
     */
    function compareLocations(itemA, itemB) {
        return itemA.line - itemB.line || itemA.column - itemB.column;
    }
    /**
     * This is the same as the exported function, except that it
     * doesn't handle disable-line and disable-next-line directives, and it always reports unused
     * disable directives.
     * @param {Object} options options for applying directives. This is the same as the options
     * for the exported function, except that `reportUnusedDisableDirectives` is not supported
     * (this function always reports unused disable directives).
     * @returns {{problems: Problem[], unusedDisableDirectives: Problem[]}} An object with a list
     * of filtered problems and unused eslint-disable directives
     */
    function applyDirectives(options) {
        const problems = [];
        let nextDirectiveIndex = 0;
        let currentGlobalDisableDirective = null;
        const disabledRuleMap = new Map();
        // enabledRules is only used when there is a current global disable directive.
        const enabledRules = new Set();
        const usedDisableDirectives = new Set();
        for (const problem of options.problems) {
            while (nextDirectiveIndex < options.directives.length &&
                compareLocations(options.directives[nextDirectiveIndex], problem) <= 0) {
                const directive = options.directives[nextDirectiveIndex++];
                switch (directive.type) {
                    case "disable":
                        if (directive.ruleId === null) {
                            currentGlobalDisableDirective = directive;
                            disabledRuleMap.clear();
                            enabledRules.clear();
                        }
                        else if (currentGlobalDisableDirective) {
                            enabledRules.delete(directive.ruleId);
                            disabledRuleMap.set(directive.ruleId, directive);
                        }
                        else {
                            disabledRuleMap.set(directive.ruleId, directive);
                        }
                        break;
                    case "enable":
                        if (directive.ruleId === null) {
                            currentGlobalDisableDirective = null;
                            disabledRuleMap.clear();
                        }
                        else if (currentGlobalDisableDirective) {
                            enabledRules.add(directive.ruleId);
                            disabledRuleMap.delete(directive.ruleId);
                        }
                        else {
                            disabledRuleMap.delete(directive.ruleId);
                        }
                        break;
                    // no default
                }
            }
            if (disabledRuleMap.has(problem.ruleId)) {
                usedDisableDirectives.add(disabledRuleMap.get(problem.ruleId));
            }
            else if (currentGlobalDisableDirective && !enabledRules.has(problem.ruleId)) {
                usedDisableDirectives.add(currentGlobalDisableDirective);
            }
            else {
                problems.push(problem);
            }
        }
        const unusedDisableDirectives = options.directives
            .filter(directive => directive.type === "disable" && !usedDisableDirectives.has(directive))
            .map(directive => ({
            ruleId: null,
            message: directive.ruleId
                ? `Unused eslint-disable directive (no problems were reported from '${directive.ruleId}').`
                : "Unused eslint-disable directive (no problems were reported).",
            line: directive.unprocessedDirective.line,
            column: directive.unprocessedDirective.column,
            severity: options.reportUnusedDisableDirectives === "warn" ? 1 : 2,
            nodeType: null
        }));
        return { problems, unusedDisableDirectives };
    }
    /**
     * Given a list of directive comments (i.e. metadata about eslint-disable and eslint-enable comments) and a list
     * of reported problems, determines which problems should be reported.
     * @param {Object} options Information about directives and problems
     * @param {{
     *      type: ("disable"|"enable"|"disable-line"|"disable-next-line"),
     *      ruleId: (string|null),
     *      line: number,
     *      column: number
     * }} options.directives Directive comments found in the file, with one-based columns.
     * Two directive comments can only have the same location if they also have the same type (e.g. a single eslint-disable
     * comment for two different rules is represented as two directives).
     * @param {{ruleId: (string|null), line: number, column: number}[]} options.problems
     * A list of problems reported by rules, sorted by increasing location in the file, with one-based columns.
     * @param {"off" | "warn" | "error"} options.reportUnusedDisableDirectives If `"warn"` or `"error"`, adds additional problems for unused directives
     * @returns {{ruleId: (string|null), line: number, column: number}[]}
     * A list of reported problems that were not disabled by the directive comments.
     */
    module.exports = ({ directives, problems, reportUnusedDisableDirectives = "off" }) => {
        const blockDirectives = directives
            .filter(directive => directive.type === "disable" || directive.type === "enable")
            .map(directive => Object.assign({}, directive, { unprocessedDirective: directive }))
            .sort(compareLocations);
        /**
         * Returns a new array formed by applying a given callback function to each element of the array, and then flattening the result by one level.
         * TODO(stephenwade): Replace this with array.flatMap when we drop support for Node v10
         * @param {any[]} array The array to process
         * @param {Function} fn The function to use
         * @returns {any[]} The result array
         */
        function flatMap(array, fn) {
            const mapped = array.map(fn);
            const flattened = [].concat(...mapped);
            return flattened;
        }
        const lineDirectives = flatMap(directives, directive => {
            switch (directive.type) {
                case "disable":
                case "enable":
                    return [];
                case "disable-line":
                    return [
                        { type: "disable", line: directive.line, column: 1, ruleId: directive.ruleId, unprocessedDirective: directive },
                        { type: "enable", line: directive.line + 1, column: 0, ruleId: directive.ruleId, unprocessedDirective: directive }
                    ];
                case "disable-next-line":
                    return [
                        { type: "disable", line: directive.line + 1, column: 1, ruleId: directive.ruleId, unprocessedDirective: directive },
                        { type: "enable", line: directive.line + 2, column: 0, ruleId: directive.ruleId, unprocessedDirective: directive }
                    ];
                default:
                    throw new TypeError(`Unrecognized directive type '${directive.type}'`);
            }
        }).sort(compareLocations);
        const blockDirectivesResult = applyDirectives({
            problems,
            directives: blockDirectives,
            reportUnusedDisableDirectives
        });
        const lineDirectivesResult = applyDirectives({
            problems: blockDirectivesResult.problems,
            directives: lineDirectives,
            reportUnusedDisableDirectives
        });
        return reportUnusedDisableDirectives !== "off"
            ? lineDirectivesResult.problems
                .concat(blockDirectivesResult.unusedDisableDirectives)
                .concat(lineDirectivesResult.unusedDisableDirectives)
                .sort(compareLocations)
            : lineDirectivesResult.problems;
    };
},
/* eslint/lib/linter/config-comment-parser.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const levn = require(373) /* levn */, ConfigOps = require(220) /* @eslint/eslintrc/lib/shared/config-ops */;
    const debug = require(72) /* debug */("eslint:config-comment-parser");
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * Object to parse ESLint configuration comments inside JavaScript files.
     * @name ConfigCommentParser
     */
    module.exports = class ConfigCommentParser {
        /**
         * Parses a list of "name:string_value" or/and "name" options divided by comma or
         * whitespace. Used for "global" and "exported" comments.
         * @param {string} string The string to parse.
         * @param {Comment} comment The comment node which has the string.
         * @returns {Object} Result map object of names and string values, or null values if no value was provided
         */
        parseStringConfig(string, comment) {
            debug("Parsing String config");
            const items = {};
            // Collapse whitespace around `:` and `,` to make parsing easier
            const trimmedString = string.replace(/\s*([:,])\s*/gu, "$1");
            trimmedString.split(/\s|,+/u).forEach(name => {
                if (!name) {
                    return;
                }
                // value defaults to null (if not provided), e.g: "foo" => ["foo", null]
                const [key, value = null] = name.split(":");
                items[key] = { value, comment };
            });
            return items;
        }
        /**
         * Parses a JSON-like config.
         * @param {string} string The string to parse.
         * @param {Object} location Start line and column of comments for potential error message.
         * @returns {({success: true, config: Object}|{success: false, error: Problem})} Result map object
         */
        parseJsonConfig(string, location) {
            debug("Parsing JSON config");
            let items = {};
            // Parses a JSON-like comment by the same way as parsing CLI option.
            try {
                items = levn.parse("Object", string) || {};
                // Some tests say that it should ignore invalid comments such as `/*eslint no-alert:abc*/`.
                // Also, commaless notations have invalid severity:
                //     "no-alert: 2 no-console: 2" --> {"no-alert": "2 no-console: 2"}
                // Should ignore that case as well.
                if (ConfigOps.isEverySeverityValid(items)) {
                    return {
                        success: true,
                        config: items
                    };
                }
            }
            catch {
                debug("Levn parsing failed; falling back to manual parsing.");
                // ignore to parse the string by a fallback.
            }
            /*
             * Optionator cannot parse commaless notations.
             * But we are supporting that. So this is a fallback for that.
             */
            items = {};
            const normalizedString = string.replace(/([-a-zA-Z0-9/]+):/gu, "\"$1\":").replace(/(\]|[0-9])\s+(?=")/u, "$1,");
            try {
                items = JSON.parse(`{${normalizedString}}`);
            }
            catch (ex) {
                debug("Manual parsing failed.");
                return {
                    success: false,
                    error: {
                        ruleId: null,
                        fatal: true,
                        severity: 2,
                        message: `Failed to parse JSON from '${normalizedString}': ${ex.message}`,
                        line: location.start.line,
                        column: location.start.column + 1
                    }
                };
            }
            return {
                success: true,
                config: items
            };
        }
        /**
         * Parses a config of values separated by comma.
         * @param {string} string The string to parse.
         * @returns {Object} Result map of values and true values
         */
        parseListConfig(string) {
            debug("Parsing list config");
            const items = {};
            // Collapse whitespace around commas
            string.replace(/\s*,\s*/gu, ",").split(/,+/u).forEach(name => {
                const trimmedName = name.trim();
                if (trimmedName) {
                    items[trimmedName] = true;
                }
            });
            return items;
        }
    };
},
/* levn/lib/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    (function () {
        var parseString, cast, parseType, VERSION, parsedTypeParse, parse;
        parseString = require(374) /* ./parse-string */;
        cast = require(381) /* ./cast */;
        parseType = require(382) /* type-check */.parseType;
        VERSION = '0.4.1';
        parsedTypeParse = function (parsedType, string, options) {
            options == null && (options = {});
            options.explicit == null && (options.explicit = false);
            options.customTypes == null && (options.customTypes = {});
            return cast(parseString(parsedType, string, options), parsedType, options);
        };
        parse = function (type, string, options) {
            return parsedTypeParse(parseType(type), string, options);
        };
        module.exports = {
            VERSION: VERSION,
            parse: parse,
            parsedTypeParse: parsedTypeParse
        };
    }).call(this);
},
/* levn/lib/parse-string.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    (function () {
        var reject, special, tokenRegex;
        reject = require(375) /* prelude-ls */.reject;
        function consumeOp(tokens, op) {
            if (tokens[0] === op) {
                return tokens.shift();
            }
            else {
                throw new Error("Expected '" + op + "', but got '" + tokens[0] + "' instead in " + JSON.stringify(tokens) + ".");
            }
        }
        function maybeConsumeOp(tokens, op) {
            if (tokens[0] === op) {
                return tokens.shift();
            }
        }
        function consumeList(tokens, arg$, hasDelimiters) {
            var open, close, result, untilTest;
            open = arg$[0], close = arg$[1];
            if (hasDelimiters) {
                consumeOp(tokens, open);
            }
            result = [];
            untilTest = "," + (hasDelimiters ? close : '');
            while (tokens.length && (hasDelimiters && tokens[0] !== close)) {
                result.push(consumeElement(tokens, untilTest));
                maybeConsumeOp(tokens, ',');
            }
            if (hasDelimiters) {
                consumeOp(tokens, close);
            }
            return result;
        }
        function consumeArray(tokens, hasDelimiters) {
            return consumeList(tokens, ['[', ']'], hasDelimiters);
        }
        function consumeTuple(tokens, hasDelimiters) {
            return consumeList(tokens, ['(', ')'], hasDelimiters);
        }
        function consumeFields(tokens, hasDelimiters) {
            var result, untilTest, key;
            if (hasDelimiters) {
                consumeOp(tokens, '{');
            }
            result = {};
            untilTest = "," + (hasDelimiters ? '}' : '');
            while (tokens.length && (!hasDelimiters || tokens[0] !== '}')) {
                key = consumeValue(tokens, ':');
                consumeOp(tokens, ':');
                result[key] = consumeElement(tokens, untilTest);
                maybeConsumeOp(tokens, ',');
            }
            if (hasDelimiters) {
                consumeOp(tokens, '}');
            }
            return result;
        }
        function consumeValue(tokens, untilTest) {
            var out;
            untilTest == null && (untilTest = '');
            out = '';
            while (tokens.length && -1 === untilTest.indexOf(tokens[0])) {
                out += tokens.shift();
            }
            return out;
        }
        function consumeElement(tokens, untilTest) {
            switch (tokens[0]) {
                case '[':
                    return consumeArray(tokens, true);
                case '(':
                    return consumeTuple(tokens, true);
                case '{':
                    return consumeFields(tokens, true);
                default:
                    return consumeValue(tokens, untilTest);
            }
        }
        function consumeTopLevel(tokens, types, options) {
            var ref$, type, structure, origTokens, result, finalResult, x$, y$;
            ref$ = types[0], type = ref$.type, structure = ref$.structure;
            origTokens = tokens.concat();
            if (!options.explicit && types.length === 1 && ((!type && structure) || (type === 'Array' || type === 'Object'))) {
                result = structure === 'array' || type === 'Array'
                    ? consumeArray(tokens, tokens[0] === '[')
                    : structure === 'tuple'
                        ? consumeTuple(tokens, tokens[0] === '(')
                        : consumeFields(tokens, tokens[0] === '{');
                finalResult = tokens.length ? consumeElement(structure === 'array' || type === 'Array'
                    ? (x$ = origTokens, x$.unshift('['), x$.push(']'), x$)
                    : (y$ = origTokens, y$.unshift('('), y$.push(')'), y$)) : result;
            }
            else {
                finalResult = consumeElement(tokens);
            }
            return finalResult;
        }
        special = /\[\]\(\)}{:,/.source;
        tokenRegex = RegExp('("(?:\\\\"|[^"])*")|(\'(?:\\\\\'|[^\'])*\')|(/(?:\\\\/|[^/])*/[a-zA-Z]*)|(#.*#)|([' + special + '])|([^\\s' + special + '](?:\\s*[^\\s' + special + ']+)*)|\\s*');
        module.exports = function (types, string, options) {
            var tokens, node;
            options == null && (options = {});
            if (!options.explicit && types.length === 1 && types[0].type === 'String') {
                return string;
            }
            tokens = reject(not$, string.split(tokenRegex));
            node = consumeTopLevel(tokens, types, options);
            if (!node) {
                throw new Error("Error parsing '" + string + "'.");
            }
            return node;
        };
        function not$(x) { return !x; }
    }).call(this);
},
/* prelude-ls/lib/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    var Func, List, Obj, Str, Num, id, isType, replicate, prelude, toString$ = {}.toString;
    Func = require(376) /* ./Func.js */;
    List = require(377) /* ./List.js */;
    Obj = require(378) /* ./Obj.js */;
    Str = require(379) /* ./Str.js */;
    Num = require(380) /* ./Num.js */;
    id = function (x) {
        return x;
    };
    isType = curry$(function (type, x) {
        return toString$.call(x).slice(8, -1) === type;
    });
    replicate = curry$(function (n, x) {
        var i$, results$ = [];
        for (i$ = 0; i$ < n; ++i$) {
            results$.push(x);
        }
        return results$;
    });
    Str.empty = List.empty;
    Str.slice = List.slice;
    Str.take = List.take;
    Str.drop = List.drop;
    Str.splitAt = List.splitAt;
    Str.takeWhile = List.takeWhile;
    Str.dropWhile = List.dropWhile;
    Str.span = List.span;
    Str.breakStr = List.breakList;
    prelude = {
        Func: Func,
        List: List,
        Obj: Obj,
        Str: Str,
        Num: Num,
        id: id,
        isType: isType,
        replicate: replicate
    };
    prelude.each = List.each;
    prelude.map = List.map;
    prelude.filter = List.filter;
    prelude.compact = List.compact;
    prelude.reject = List.reject;
    prelude.partition = List.partition;
    prelude.find = List.find;
    prelude.head = List.head;
    prelude.first = List.first;
    prelude.tail = List.tail;
    prelude.last = List.last;
    prelude.initial = List.initial;
    prelude.empty = List.empty;
    prelude.reverse = List.reverse;
    prelude.difference = List.difference;
    prelude.intersection = List.intersection;
    prelude.union = List.union;
    prelude.countBy = List.countBy;
    prelude.groupBy = List.groupBy;
    prelude.fold = List.fold;
    prelude.foldl = List.foldl;
    prelude.fold1 = List.fold1;
    prelude.foldl1 = List.foldl1;
    prelude.foldr = List.foldr;
    prelude.foldr1 = List.foldr1;
    prelude.unfoldr = List.unfoldr;
    prelude.andList = List.andList;
    prelude.orList = List.orList;
    prelude.any = List.any;
    prelude.all = List.all;
    prelude.unique = List.unique;
    prelude.uniqueBy = List.uniqueBy;
    prelude.sort = List.sort;
    prelude.sortWith = List.sortWith;
    prelude.sortBy = List.sortBy;
    prelude.sum = List.sum;
    prelude.product = List.product;
    prelude.mean = List.mean;
    prelude.average = List.average;
    prelude.concat = List.concat;
    prelude.concatMap = List.concatMap;
    prelude.flatten = List.flatten;
    prelude.maximum = List.maximum;
    prelude.minimum = List.minimum;
    prelude.maximumBy = List.maximumBy;
    prelude.minimumBy = List.minimumBy;
    prelude.scan = List.scan;
    prelude.scanl = List.scanl;
    prelude.scan1 = List.scan1;
    prelude.scanl1 = List.scanl1;
    prelude.scanr = List.scanr;
    prelude.scanr1 = List.scanr1;
    prelude.slice = List.slice;
    prelude.take = List.take;
    prelude.drop = List.drop;
    prelude.splitAt = List.splitAt;
    prelude.takeWhile = List.takeWhile;
    prelude.dropWhile = List.dropWhile;
    prelude.span = List.span;
    prelude.breakList = List.breakList;
    prelude.zip = List.zip;
    prelude.zipWith = List.zipWith;
    prelude.zipAll = List.zipAll;
    prelude.zipAllWith = List.zipAllWith;
    prelude.at = List.at;
    prelude.elemIndex = List.elemIndex;
    prelude.elemIndices = List.elemIndices;
    prelude.findIndex = List.findIndex;
    prelude.findIndices = List.findIndices;
    prelude.apply = Func.apply;
    prelude.curry = Func.curry;
    prelude.flip = Func.flip;
    prelude.fix = Func.fix;
    prelude.over = Func.over;
    prelude.split = Str.split;
    prelude.join = Str.join;
    prelude.lines = Str.lines;
    prelude.unlines = Str.unlines;
    prelude.words = Str.words;
    prelude.unwords = Str.unwords;
    prelude.chars = Str.chars;
    prelude.unchars = Str.unchars;
    prelude.repeat = Str.repeat;
    prelude.capitalize = Str.capitalize;
    prelude.camelize = Str.camelize;
    prelude.dasherize = Str.dasherize;
    prelude.values = Obj.values;
    prelude.keys = Obj.keys;
    prelude.pairsToObj = Obj.pairsToObj;
    prelude.objToPairs = Obj.objToPairs;
    prelude.listsToObj = Obj.listsToObj;
    prelude.objToLists = Obj.objToLists;
    prelude.max = Num.max;
    prelude.min = Num.min;
    prelude.negate = Num.negate;
    prelude.abs = Num.abs;
    prelude.signum = Num.signum;
    prelude.quot = Num.quot;
    prelude.rem = Num.rem;
    prelude.div = Num.div;
    prelude.mod = Num.mod;
    prelude.recip = Num.recip;
    prelude.pi = Num.pi;
    prelude.tau = Num.tau;
    prelude.exp = Num.exp;
    prelude.sqrt = Num.sqrt;
    prelude.ln = Num.ln;
    prelude.pow = Num.pow;
    prelude.sin = Num.sin;
    prelude.tan = Num.tan;
    prelude.cos = Num.cos;
    prelude.acos = Num.acos;
    prelude.asin = Num.asin;
    prelude.atan = Num.atan;
    prelude.atan2 = Num.atan2;
    prelude.truncate = Num.truncate;
    prelude.round = Num.round;
    prelude.ceiling = Num.ceiling;
    prelude.floor = Num.floor;
    prelude.isItNaN = Num.isItNaN;
    prelude.even = Num.even;
    prelude.odd = Num.odd;
    prelude.gcd = Num.gcd;
    prelude.lcm = Num.lcm;
    prelude.VERSION = '1.2.1';
    module.exports = prelude;
    function curry$(f, bound) {
        var context, _curry = function (args) {
            return f.length > 1 ? function () {
                var params = args ? args.concat() : [];
                context = bound ? context || this : this;
                return params.push.apply(params, arguments) <
                    f.length && arguments.length ?
                    _curry.call(context, params) : f.apply(context, params);
            } : f;
        };
        return _curry();
    }
},
/* prelude-ls/lib/Func.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    var apply, curry, flip, fix, over, memoize, toString$ = {}.toString;
    apply = curry$(function (f, list) {
        return f.apply(null, list);
    });
    curry = function (f) {
        return curry$(f);
    };
    flip = curry$(function (f, x, y) {
        return f(y, x);
    });
    fix = function (f) {
        return function (g) {
            return function () {
                return f(g(g)).apply(null, arguments);
            };
        }(function (g) {
            return function () {
                return f(g(g)).apply(null, arguments);
            };
        });
    };
    over = curry$(function (f, g, x, y) {
        return f(g(x), g(y));
    });
    memoize = function (f) {
        var memo;
        memo = {};
        return function () {
            var args, res$, i$, to$, key, arg;
            res$ = [];
            for (i$ = 0, to$ = arguments.length; i$ < to$; ++i$) {
                res$.push(arguments[i$]);
            }
            args = res$;
            key = (function () {
                var i$, ref$, len$, results$ = [];
                for (i$ = 0, len$ = (ref$ = args).length; i$ < len$; ++i$) {
                    arg = ref$[i$];
                    results$.push(arg + toString$.call(arg).slice(8, -1));
                }
                return results$;
            }()).join('');
            return memo[key] = key in memo
                ? memo[key]
                : f.apply(null, args);
        };
    };
    module.exports = {
        curry: curry,
        flip: flip,
        fix: fix,
        apply: apply,
        over: over,
        memoize: memoize
    };
    function curry$(f, bound) {
        var context, _curry = function (args) {
            return f.length > 1 ? function () {
                var params = args ? args.concat() : [];
                context = bound ? context || this : this;
                return params.push.apply(params, arguments) <
                    f.length && arguments.length ?
                    _curry.call(context, params) : f.apply(context, params);
            } : f;
        };
        return _curry();
    }
},
/* prelude-ls/lib/List.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    var each, map, compact, filter, reject, remove, partition, find, head, first, tail, last, initial, empty, reverse, unique, uniqueBy, fold, foldl, fold1, foldl1, foldr, foldr1, unfoldr, concat, concatMap, flatten, difference, intersection, union, countBy, groupBy, andList, orList, any, all, sort, sortWith, sortBy, sum, product, mean, average, maximum, minimum, maximumBy, minimumBy, scan, scanl, scan1, scanl1, scanr, scanr1, slice, take, drop, splitAt, takeWhile, dropWhile, span, breakList, zip, zipWith, zipAll, zipAllWith, at, elemIndex, elemIndices, findIndex, findIndices, toString$ = {}.toString;
    each = curry$(function (f, xs) {
        var i$, len$, x;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            f(x);
        }
        return xs;
    });
    map = curry$(function (f, xs) {
        var i$, len$, x, results$ = [];
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            results$.push(f(x));
        }
        return results$;
    });
    compact = function (xs) {
        var i$, len$, x, results$ = [];
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            if (x) {
                results$.push(x);
            }
        }
        return results$;
    };
    filter = curry$(function (f, xs) {
        var i$, len$, x, results$ = [];
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            if (f(x)) {
                results$.push(x);
            }
        }
        return results$;
    });
    reject = curry$(function (f, xs) {
        var i$, len$, x, results$ = [];
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            if (!f(x)) {
                results$.push(x);
            }
        }
        return results$;
    });
    remove = curry$(function (el, xs) {
        var i, x$;
        i = elemIndex(el, xs);
        x$ = xs.slice();
        if (i != null) {
            x$.splice(i, 1);
        }
        return x$;
    });
    partition = curry$(function (f, xs) {
        var passed, failed, i$, len$, x;
        passed = [];
        failed = [];
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            (f(x) ? passed : failed).push(x);
        }
        return [passed, failed];
    });
    find = curry$(function (f, xs) {
        var i$, len$, x;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            if (f(x)) {
                return x;
            }
        }
    });
    head = first = function (xs) {
        return xs[0];
    };
    tail = function (xs) {
        if (!xs.length) {
            return;
        }
        return xs.slice(1);
    };
    last = function (xs) {
        return xs[xs.length - 1];
    };
    initial = function (xs) {
        if (!xs.length) {
            return;
        }
        return xs.slice(0, -1);
    };
    empty = function (xs) {
        return !xs.length;
    };
    reverse = function (xs) {
        return xs.concat().reverse();
    };
    unique = function (xs) {
        var result, i$, len$, x;
        result = [];
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            if (!in$(x, result)) {
                result.push(x);
            }
        }
        return result;
    };
    uniqueBy = curry$(function (f, xs) {
        var seen, i$, len$, x, val, results$ = [];
        seen = [];
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            val = f(x);
            if (in$(val, seen)) {
                continue;
            }
            seen.push(val);
            results$.push(x);
        }
        return results$;
    });
    fold = foldl = curry$(function (f, memo, xs) {
        var i$, len$, x;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            memo = f(memo, x);
        }
        return memo;
    });
    fold1 = foldl1 = curry$(function (f, xs) {
        return fold(f, xs[0], xs.slice(1));
    });
    foldr = curry$(function (f, memo, xs) {
        var i$, x;
        for (i$ = xs.length - 1; i$ >= 0; --i$) {
            x = xs[i$];
            memo = f(x, memo);
        }
        return memo;
    });
    foldr1 = curry$(function (f, xs) {
        return foldr(f, xs[xs.length - 1], xs.slice(0, -1));
    });
    unfoldr = curry$(function (f, b) {
        var result, x, that;
        result = [];
        x = b;
        while ((that = f(x)) != null) {
            result.push(that[0]);
            x = that[1];
        }
        return result;
    });
    concat = function (xss) {
        return [].concat.apply([], xss);
    };
    concatMap = curry$(function (f, xs) {
        var x;
        return [].concat.apply([], (function () {
            var i$, ref$, len$, results$ = [];
            for (i$ = 0, len$ = (ref$ = xs).length; i$ < len$; ++i$) {
                x = ref$[i$];
                results$.push(f(x));
            }
            return results$;
        }()));
    });
    flatten = function (xs) {
        var x;
        return [].concat.apply([], (function () {
            var i$, ref$, len$, results$ = [];
            for (i$ = 0, len$ = (ref$ = xs).length; i$ < len$; ++i$) {
                x = ref$[i$];
                if (toString$.call(x).slice(8, -1) === 'Array') {
                    results$.push(flatten(x));
                }
                else {
                    results$.push(x);
                }
            }
            return results$;
        }()));
    };
    difference = function (xs) {
        var yss, res$, i$, to$, results, len$, x, j$, len1$, ys;
        res$ = [];
        for (i$ = 1, to$ = arguments.length; i$ < to$; ++i$) {
            res$.push(arguments[i$]);
        }
        yss = res$;
        results = [];
        outer: for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            for (j$ = 0, len1$ = yss.length; j$ < len1$; ++j$) {
                ys = yss[j$];
                if (in$(x, ys)) {
                    continue outer;
                }
            }
            results.push(x);
        }
        return results;
    };
    intersection = function (xs) {
        var yss, res$, i$, to$, results, len$, x, j$, len1$, ys;
        res$ = [];
        for (i$ = 1, to$ = arguments.length; i$ < to$; ++i$) {
            res$.push(arguments[i$]);
        }
        yss = res$;
        results = [];
        outer: for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            for (j$ = 0, len1$ = yss.length; j$ < len1$; ++j$) {
                ys = yss[j$];
                if (!in$(x, ys)) {
                    continue outer;
                }
            }
            results.push(x);
        }
        return results;
    };
    union = function () {
        var xss, res$, i$, to$, results, len$, xs, j$, len1$, x;
        res$ = [];
        for (i$ = 0, to$ = arguments.length; i$ < to$; ++i$) {
            res$.push(arguments[i$]);
        }
        xss = res$;
        results = [];
        for (i$ = 0, len$ = xss.length; i$ < len$; ++i$) {
            xs = xss[i$];
            for (j$ = 0, len1$ = xs.length; j$ < len1$; ++j$) {
                x = xs[j$];
                if (!in$(x, results)) {
                    results.push(x);
                }
            }
        }
        return results;
    };
    countBy = curry$(function (f, xs) {
        var results, i$, len$, x, key;
        results = {};
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            key = f(x);
            if (key in results) {
                results[key] += 1;
            }
            else {
                results[key] = 1;
            }
        }
        return results;
    });
    groupBy = curry$(function (f, xs) {
        var results, i$, len$, x, key;
        results = {};
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            key = f(x);
            if (key in results) {
                results[key].push(x);
            }
            else {
                results[key] = [x];
            }
        }
        return results;
    });
    andList = function (xs) {
        var i$, len$, x;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            if (!x) {
                return false;
            }
        }
        return true;
    };
    orList = function (xs) {
        var i$, len$, x;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            if (x) {
                return true;
            }
        }
        return false;
    };
    any = curry$(function (f, xs) {
        var i$, len$, x;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            if (f(x)) {
                return true;
            }
        }
        return false;
    });
    all = curry$(function (f, xs) {
        var i$, len$, x;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            if (!f(x)) {
                return false;
            }
        }
        return true;
    });
    sort = function (xs) {
        return xs.concat().sort(function (x, y) {
            if (x > y) {
                return 1;
            }
            else if (x < y) {
                return -1;
            }
            else {
                return 0;
            }
        });
    };
    sortWith = curry$(function (f, xs) {
        return xs.concat().sort(f);
    });
    sortBy = curry$(function (f, xs) {
        return xs.concat().sort(function (x, y) {
            if (f(x) > f(y)) {
                return 1;
            }
            else if (f(x) < f(y)) {
                return -1;
            }
            else {
                return 0;
            }
        });
    });
    sum = function (xs) {
        var result, i$, len$, x;
        result = 0;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            result += x;
        }
        return result;
    };
    product = function (xs) {
        var result, i$, len$, x;
        result = 1;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            result *= x;
        }
        return result;
    };
    mean = average = function (xs) {
        var sum, i$, len$, x;
        sum = 0;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            x = xs[i$];
            sum += x;
        }
        return sum / xs.length;
    };
    maximum = function (xs) {
        var max, i$, ref$, len$, x;
        max = xs[0];
        for (i$ = 0, len$ = (ref$ = xs.slice(1)).length; i$ < len$; ++i$) {
            x = ref$[i$];
            if (x > max) {
                max = x;
            }
        }
        return max;
    };
    minimum = function (xs) {
        var min, i$, ref$, len$, x;
        min = xs[0];
        for (i$ = 0, len$ = (ref$ = xs.slice(1)).length; i$ < len$; ++i$) {
            x = ref$[i$];
            if (x < min) {
                min = x;
            }
        }
        return min;
    };
    maximumBy = curry$(function (f, xs) {
        var max, i$, ref$, len$, x;
        max = xs[0];
        for (i$ = 0, len$ = (ref$ = xs.slice(1)).length; i$ < len$; ++i$) {
            x = ref$[i$];
            if (f(x) > f(max)) {
                max = x;
            }
        }
        return max;
    });
    minimumBy = curry$(function (f, xs) {
        var min, i$, ref$, len$, x;
        min = xs[0];
        for (i$ = 0, len$ = (ref$ = xs.slice(1)).length; i$ < len$; ++i$) {
            x = ref$[i$];
            if (f(x) < f(min)) {
                min = x;
            }
        }
        return min;
    });
    scan = scanl = curry$(function (f, memo, xs) {
        var last, x;
        last = memo;
        return [memo].concat((function () {
            var i$, ref$, len$, results$ = [];
            for (i$ = 0, len$ = (ref$ = xs).length; i$ < len$; ++i$) {
                x = ref$[i$];
                results$.push(last = f(last, x));
            }
            return results$;
        }()));
    });
    scan1 = scanl1 = curry$(function (f, xs) {
        if (!xs.length) {
            return;
        }
        return scan(f, xs[0], xs.slice(1));
    });
    scanr = curry$(function (f, memo, xs) {
        xs = xs.concat().reverse();
        return scan(f, memo, xs).reverse();
    });
    scanr1 = curry$(function (f, xs) {
        if (!xs.length) {
            return;
        }
        xs = xs.concat().reverse();
        return scan(f, xs[0], xs.slice(1)).reverse();
    });
    slice = curry$(function (x, y, xs) {
        return xs.slice(x, y);
    });
    take = curry$(function (n, xs) {
        if (n <= 0) {
            return xs.slice(0, 0);
        }
        else {
            return xs.slice(0, n);
        }
    });
    drop = curry$(function (n, xs) {
        if (n <= 0) {
            return xs;
        }
        else {
            return xs.slice(n);
        }
    });
    splitAt = curry$(function (n, xs) {
        return [take(n, xs), drop(n, xs)];
    });
    takeWhile = curry$(function (p, xs) {
        var len, i;
        len = xs.length;
        if (!len) {
            return xs;
        }
        i = 0;
        while (i < len && p(xs[i])) {
            i += 1;
        }
        return xs.slice(0, i);
    });
    dropWhile = curry$(function (p, xs) {
        var len, i;
        len = xs.length;
        if (!len) {
            return xs;
        }
        i = 0;
        while (i < len && p(xs[i])) {
            i += 1;
        }
        return xs.slice(i);
    });
    span = curry$(function (p, xs) {
        return [takeWhile(p, xs), dropWhile(p, xs)];
    });
    breakList = curry$(function (p, xs) {
        return span(compose$(p, not$), xs);
    });
    zip = curry$(function (xs, ys) {
        var result, len, i$, len$, i, x;
        result = [];
        len = ys.length;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            i = i$;
            x = xs[i$];
            if (i === len) {
                break;
            }
            result.push([x, ys[i]]);
        }
        return result;
    });
    zipWith = curry$(function (f, xs, ys) {
        var result, len, i$, len$, i, x;
        result = [];
        len = ys.length;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            i = i$;
            x = xs[i$];
            if (i === len) {
                break;
            }
            result.push(f(x, ys[i]));
        }
        return result;
    });
    zipAll = function () {
        var xss, res$, i$, to$, minLength, len$, xs, ref$, i, lresult$, j$, results$ = [];
        res$ = [];
        for (i$ = 0, to$ = arguments.length; i$ < to$; ++i$) {
            res$.push(arguments[i$]);
        }
        xss = res$;
        minLength = undefined;
        for (i$ = 0, len$ = xss.length; i$ < len$; ++i$) {
            xs = xss[i$];
            minLength <= (ref$ = xs.length) || (minLength = ref$);
        }
        for (i$ = 0; i$ < minLength; ++i$) {
            i = i$;
            lresult$ = [];
            for (j$ = 0, len$ = xss.length; j$ < len$; ++j$) {
                xs = xss[j$];
                lresult$.push(xs[i]);
            }
            results$.push(lresult$);
        }
        return results$;
    };
    zipAllWith = function (f) {
        var xss, res$, i$, to$, minLength, len$, xs, ref$, i, results$ = [];
        res$ = [];
        for (i$ = 1, to$ = arguments.length; i$ < to$; ++i$) {
            res$.push(arguments[i$]);
        }
        xss = res$;
        minLength = undefined;
        for (i$ = 0, len$ = xss.length; i$ < len$; ++i$) {
            xs = xss[i$];
            minLength <= (ref$ = xs.length) || (minLength = ref$);
        }
        for (i$ = 0; i$ < minLength; ++i$) {
            i = i$;
            results$.push(f.apply(null, (fn$())));
        }
        return results$;
        function fn$() {
            var i$, ref$, len$, results$ = [];
            for (i$ = 0, len$ = (ref$ = xss).length; i$ < len$; ++i$) {
                xs = ref$[i$];
                results$.push(xs[i]);
            }
            return results$;
        }
    };
    at = curry$(function (n, xs) {
        if (n < 0) {
            return xs[xs.length + n];
        }
        else {
            return xs[n];
        }
    });
    elemIndex = curry$(function (el, xs) {
        var i$, len$, i, x;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            i = i$;
            x = xs[i$];
            if (x === el) {
                return i;
            }
        }
    });
    elemIndices = curry$(function (el, xs) {
        var i$, len$, i, x, results$ = [];
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            i = i$;
            x = xs[i$];
            if (x === el) {
                results$.push(i);
            }
        }
        return results$;
    });
    findIndex = curry$(function (f, xs) {
        var i$, len$, i, x;
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            i = i$;
            x = xs[i$];
            if (f(x)) {
                return i;
            }
        }
    });
    findIndices = curry$(function (f, xs) {
        var i$, len$, i, x, results$ = [];
        for (i$ = 0, len$ = xs.length; i$ < len$; ++i$) {
            i = i$;
            x = xs[i$];
            if (f(x)) {
                results$.push(i);
            }
        }
        return results$;
    });
    module.exports = {
        each: each,
        map: map,
        filter: filter,
        compact: compact,
        reject: reject,
        remove: remove,
        partition: partition,
        find: find,
        head: head,
        first: first,
        tail: tail,
        last: last,
        initial: initial,
        empty: empty,
        reverse: reverse,
        difference: difference,
        intersection: intersection,
        union: union,
        countBy: countBy,
        groupBy: groupBy,
        fold: fold,
        fold1: fold1,
        foldl: foldl,
        foldl1: foldl1,
        foldr: foldr,
        foldr1: foldr1,
        unfoldr: unfoldr,
        andList: andList,
        orList: orList,
        any: any,
        all: all,
        unique: unique,
        uniqueBy: uniqueBy,
        sort: sort,
        sortWith: sortWith,
        sortBy: sortBy,
        sum: sum,
        product: product,
        mean: mean,
        average: average,
        concat: concat,
        concatMap: concatMap,
        flatten: flatten,
        maximum: maximum,
        minimum: minimum,
        maximumBy: maximumBy,
        minimumBy: minimumBy,
        scan: scan,
        scan1: scan1,
        scanl: scanl,
        scanl1: scanl1,
        scanr: scanr,
        scanr1: scanr1,
        slice: slice,
        take: take,
        drop: drop,
        splitAt: splitAt,
        takeWhile: takeWhile,
        dropWhile: dropWhile,
        span: span,
        breakList: breakList,
        zip: zip,
        zipWith: zipWith,
        zipAll: zipAll,
        zipAllWith: zipAllWith,
        at: at,
        elemIndex: elemIndex,
        elemIndices: elemIndices,
        findIndex: findIndex,
        findIndices: findIndices
    };
    function curry$(f, bound) {
        var context, _curry = function (args) {
            return f.length > 1 ? function () {
                var params = args ? args.concat() : [];
                context = bound ? context || this : this;
                return params.push.apply(params, arguments) <
                    f.length && arguments.length ?
                    _curry.call(context, params) : f.apply(context, params);
            } : f;
        };
        return _curry();
    }
    function in$(x, xs) {
        var i = -1, l = xs.length >>> 0;
        while (++i < l)
            if (x === xs[i])
                return true;
        return false;
    }
    function compose$() {
        var functions = arguments;
        return function () {
            var i, result;
            result = functions[0].apply(this, arguments);
            for (i = 1; i < functions.length; ++i) {
                result = functions[i](result);
            }
            return result;
        };
    }
    function not$(x) { return !x; }
},
/* prelude-ls/lib/Obj.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    var values, keys, pairsToObj, objToPairs, listsToObj, objToLists, empty, each, map, compact, filter, reject, partition, find;
    values = function (object) {
        var i$, x, results$ = [];
        for (i$ in object) {
            x = object[i$];
            results$.push(x);
        }
        return results$;
    };
    keys = function (object) {
        var x, results$ = [];
        for (x in object) {
            results$.push(x);
        }
        return results$;
    };
    pairsToObj = function (object) {
        var i$, len$, x, resultObj$ = {};
        for (i$ = 0, len$ = object.length; i$ < len$; ++i$) {
            x = object[i$];
            resultObj$[x[0]] = x[1];
        }
        return resultObj$;
    };
    objToPairs = function (object) {
        var key, value, results$ = [];
        for (key in object) {
            value = object[key];
            results$.push([key, value]);
        }
        return results$;
    };
    listsToObj = curry$(function (keys, values) {
        var i$, len$, i, key, resultObj$ = {};
        for (i$ = 0, len$ = keys.length; i$ < len$; ++i$) {
            i = i$;
            key = keys[i$];
            resultObj$[key] = values[i];
        }
        return resultObj$;
    });
    objToLists = function (object) {
        var keys, values, key, value;
        keys = [];
        values = [];
        for (key in object) {
            value = object[key];
            keys.push(key);
            values.push(value);
        }
        return [keys, values];
    };
    empty = function (object) {
        var x;
        for (x in object) {
            return false;
        }
        return true;
    };
    each = curry$(function (f, object) {
        var i$, x;
        for (i$ in object) {
            x = object[i$];
            f(x);
        }
        return object;
    });
    map = curry$(function (f, object) {
        var k, x, resultObj$ = {};
        for (k in object) {
            x = object[k];
            resultObj$[k] = f(x);
        }
        return resultObj$;
    });
    compact = function (object) {
        var k, x, resultObj$ = {};
        for (k in object) {
            x = object[k];
            if (x) {
                resultObj$[k] = x;
            }
        }
        return resultObj$;
    };
    filter = curry$(function (f, object) {
        var k, x, resultObj$ = {};
        for (k in object) {
            x = object[k];
            if (f(x)) {
                resultObj$[k] = x;
            }
        }
        return resultObj$;
    });
    reject = curry$(function (f, object) {
        var k, x, resultObj$ = {};
        for (k in object) {
            x = object[k];
            if (!f(x)) {
                resultObj$[k] = x;
            }
        }
        return resultObj$;
    });
    partition = curry$(function (f, object) {
        var passed, failed, k, x;
        passed = {};
        failed = {};
        for (k in object) {
            x = object[k];
            (f(x) ? passed : failed)[k] = x;
        }
        return [passed, failed];
    });
    find = curry$(function (f, object) {
        var i$, x;
        for (i$ in object) {
            x = object[i$];
            if (f(x)) {
                return x;
            }
        }
    });
    module.exports = {
        values: values,
        keys: keys,
        pairsToObj: pairsToObj,
        objToPairs: objToPairs,
        listsToObj: listsToObj,
        objToLists: objToLists,
        empty: empty,
        each: each,
        map: map,
        filter: filter,
        compact: compact,
        reject: reject,
        partition: partition,
        find: find
    };
    function curry$(f, bound) {
        var context, _curry = function (args) {
            return f.length > 1 ? function () {
                var params = args ? args.concat() : [];
                context = bound ? context || this : this;
                return params.push.apply(params, arguments) <
                    f.length && arguments.length ?
                    _curry.call(context, params) : f.apply(context, params);
            } : f;
        };
        return _curry();
    }
},
/* prelude-ls/lib/Str.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    var split, join, lines, unlines, words, unwords, chars, unchars, reverse, repeat, capitalize, camelize, dasherize;
    split = curry$(function (sep, str) {
        return str.split(sep);
    });
    join = curry$(function (sep, xs) {
        return xs.join(sep);
    });
    lines = function (str) {
        if (!str.length) {
            return [];
        }
        return str.split('\n');
    };
    unlines = function (it) {
        return it.join('\n');
    };
    words = function (str) {
        if (!str.length) {
            return [];
        }
        return str.split(/[ ]+/);
    };
    unwords = function (it) {
        return it.join(' ');
    };
    chars = function (it) {
        return it.split('');
    };
    unchars = function (it) {
        return it.join('');
    };
    reverse = function (str) {
        return str.split('').reverse().join('');
    };
    repeat = curry$(function (n, str) {
        var result, i$;
        result = '';
        for (i$ = 0; i$ < n; ++i$) {
            result += str;
        }
        return result;
    });
    capitalize = function (str) {
        return str.charAt(0).toUpperCase() + str.slice(1);
    };
    camelize = function (it) {
        return it.replace(/[-_]+(.)?/g, function (arg$, c) {
            return (c != null ? c : '').toUpperCase();
        });
    };
    dasherize = function (str) {
        return str.replace(/([^-A-Z])([A-Z]+)/g, function (arg$, lower, upper) {
            return lower + "-" + (upper.length > 1
                ? upper
                : upper.toLowerCase());
        }).replace(/^([A-Z]+)/, function (arg$, upper) {
            if (upper.length > 1) {
                return upper + "-";
            }
            else {
                return upper.toLowerCase();
            }
        });
    };
    module.exports = {
        split: split,
        join: join,
        lines: lines,
        unlines: unlines,
        words: words,
        unwords: unwords,
        chars: chars,
        unchars: unchars,
        reverse: reverse,
        repeat: repeat,
        capitalize: capitalize,
        camelize: camelize,
        dasherize: dasherize
    };
    function curry$(f, bound) {
        var context, _curry = function (args) {
            return f.length > 1 ? function () {
                var params = args ? args.concat() : [];
                context = bound ? context || this : this;
                return params.push.apply(params, arguments) <
                    f.length && arguments.length ?
                    _curry.call(context, params) : f.apply(context, params);
            } : f;
        };
        return _curry();
    }
},
/* prelude-ls/lib/Num.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    var max, min, negate, abs, signum, quot, rem, div, mod, recip, pi, tau, exp, sqrt, ln, pow, sin, tan, cos, asin, acos, atan, atan2, truncate, round, ceiling, floor, isItNaN, even, odd, gcd, lcm;
    max = curry$(function (x$, y$) {
        return x$ > y$ ? x$ : y$;
    });
    min = curry$(function (x$, y$) {
        return x$ < y$ ? x$ : y$;
    });
    negate = function (x) {
        return -x;
    };
    abs = Math.abs;
    signum = function (x) {
        if (x < 0) {
            return -1;
        }
        else if (x > 0) {
            return 1;
        }
        else {
            return 0;
        }
    };
    quot = curry$(function (x, y) {
        return ~~(x / y);
    });
    rem = curry$(function (x$, y$) {
        return x$ % y$;
    });
    div = curry$(function (x, y) {
        return Math.floor(x / y);
    });
    mod = curry$(function (x$, y$) {
        var ref$;
        return ((x$) % (ref$ = y$) + ref$) % ref$;
    });
    recip = (function (it) {
        return 1 / it;
    });
    pi = Math.PI;
    tau = pi * 2;
    exp = Math.exp;
    sqrt = Math.sqrt;
    ln = Math.log;
    pow = curry$(function (x$, y$) {
        return Math.pow(x$, y$);
    });
    sin = Math.sin;
    tan = Math.tan;
    cos = Math.cos;
    asin = Math.asin;
    acos = Math.acos;
    atan = Math.atan;
    atan2 = curry$(function (x, y) {
        return Math.atan2(x, y);
    });
    truncate = function (x) {
        return ~~x;
    };
    round = Math.round;
    ceiling = Math.ceil;
    floor = Math.floor;
    isItNaN = function (x) {
        return x !== x;
    };
    even = function (x) {
        return x % 2 === 0;
    };
    odd = function (x) {
        return x % 2 !== 0;
    };
    gcd = curry$(function (x, y) {
        var z;
        x = Math.abs(x);
        y = Math.abs(y);
        while (y !== 0) {
            z = x % y;
            x = y;
            y = z;
        }
        return x;
    });
    lcm = curry$(function (x, y) {
        return Math.abs(Math.floor(x / gcd(x, y) * y));
    });
    module.exports = {
        max: max,
        min: min,
        negate: negate,
        abs: abs,
        signum: signum,
        quot: quot,
        rem: rem,
        div: div,
        mod: mod,
        recip: recip,
        pi: pi,
        tau: tau,
        exp: exp,
        sqrt: sqrt,
        ln: ln,
        pow: pow,
        sin: sin,
        tan: tan,
        cos: cos,
        acos: acos,
        asin: asin,
        atan: atan,
        atan2: atan2,
        truncate: truncate,
        round: round,
        ceiling: ceiling,
        floor: floor,
        isItNaN: isItNaN,
        even: even,
        odd: odd,
        gcd: gcd,
        lcm: lcm
    };
    function curry$(f, bound) {
        var context, _curry = function (args) {
            return f.length > 1 ? function () {
                var params = args ? args.concat() : [];
                context = bound ? context || this : this;
                return params.push.apply(params, arguments) <
                    f.length && arguments.length ?
                    _curry.call(context, params) : f.apply(context, params);
            } : f;
        };
        return _curry();
    }
},
/* levn/lib/cast.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    (function () {
        var parsedTypeCheck, types, toString$ = {}.toString;
        parsedTypeCheck = require(382) /* type-check */.parsedTypeCheck;
        types = {
            '*': function (value, options) {
                switch (toString$.call(value).slice(8, -1)) {
                    case 'Array':
                        return typeCast(value, {
                            type: 'Array'
                        }, options);
                    case 'Object':
                        return typeCast(value, {
                            type: 'Object'
                        }, options);
                    default:
                        return {
                            type: 'Just',
                            value: typesCast(value, [
                                {
                                    type: 'Undefined'
                                }, {
                                    type: 'Null'
                                }, {
                                    type: 'NaN'
                                }, {
                                    type: 'Boolean'
                                }, {
                                    type: 'Number'
                                }, {
                                    type: 'Date'
                                }, {
                                    type: 'RegExp'
                                }, {
                                    type: 'Array'
                                }, {
                                    type: 'Object'
                                }, {
                                    type: 'String'
                                }
                            ], (options.explicit = true, options))
                        };
                }
            },
            Undefined: function (it) {
                if (it === 'undefined' || it === void 8) {
                    return {
                        type: 'Just',
                        value: void 8
                    };
                }
                else {
                    return {
                        type: 'Nothing'
                    };
                }
            },
            Null: function (it) {
                if (it === 'null') {
                    return {
                        type: 'Just',
                        value: null
                    };
                }
                else {
                    return {
                        type: 'Nothing'
                    };
                }
            },
            NaN: function (it) {
                if (it === 'NaN') {
                    return {
                        type: 'Just',
                        value: NaN
                    };
                }
                else {
                    return {
                        type: 'Nothing'
                    };
                }
            },
            Boolean: function (it) {
                if (it === 'true') {
                    return {
                        type: 'Just',
                        value: true
                    };
                }
                else if (it === 'false') {
                    return {
                        type: 'Just',
                        value: false
                    };
                }
                else {
                    return {
                        type: 'Nothing'
                    };
                }
            },
            Number: function (it) {
                return {
                    type: 'Just',
                    value: +it
                };
            },
            Int: function (it) {
                return {
                    type: 'Just',
                    value: +it
                };
            },
            Float: function (it) {
                return {
                    type: 'Just',
                    value: +it
                };
            },
            Date: function (value, options) {
                var that;
                if (that = /^\#([\s\S]*)\#$/.exec(value)) {
                    return {
                        type: 'Just',
                        value: new Date(+that[1] || that[1])
                    };
                }
                else if (options.explicit) {
                    return {
                        type: 'Nothing'
                    };
                }
                else {
                    return {
                        type: 'Just',
                        value: new Date(+value || value)
                    };
                }
            },
            RegExp: function (value, options) {
                var that;
                if (that = /^\/([\s\S]*)\/([gimy]*)$/.exec(value)) {
                    return {
                        type: 'Just',
                        value: new RegExp(that[1], that[2])
                    };
                }
                else if (options.explicit) {
                    return {
                        type: 'Nothing'
                    };
                }
                else {
                    return {
                        type: 'Just',
                        value: new RegExp(value)
                    };
                }
            },
            Array: function (value, options) {
                return castArray(value, {
                    of: [{
                            type: '*'
                        }]
                }, options);
            },
            Object: function (value, options) {
                return castFields(value, {
                    of: {}
                }, options);
            },
            String: function (it) {
                var replace, that;
                if (toString$.call(it).slice(8, -1) !== 'String') {
                    return {
                        type: 'Nothing'
                    };
                }
                replace = function (value, quote) {
                    return value.replace(/\\([^u]|u[0-9a-fA-F]{4})/g, function (all, escaped) {
                        switch (escaped[0]) {
                            case quote:
                                return quote;
                            case '\\':
                                return '\\';
                            case 'b':
                                return '\b';
                            case 'f':
                                return '\f';
                            case 'n':
                                return '\n';
                            case 'r':
                                return '\r';
                            case 't':
                                return '\t';
                            case 'u':
                                return JSON.parse("\"" + all + "\"");
                            default:
                                return escaped;
                        }
                    });
                };
                if (that = it.match(/^'([\s\S]*)'$/)) {
                    return {
                        type: 'Just',
                        value: replace(that[1], "'")
                    };
                }
                else if (that = it.match(/^"([\s\S]*)"$/)) {
                    return {
                        type: 'Just',
                        value: replace(that[1], '"')
                    };
                }
                else {
                    return {
                        type: 'Just',
                        value: it
                    };
                }
            }
        };
        function castArray(node, type, options) {
            var typeOf, element;
            if (toString$.call(node).slice(8, -1) !== 'Array') {
                return {
                    type: 'Nothing'
                };
            }
            typeOf = type.of;
            return {
                type: 'Just',
                value: (function () {
                    var i$, ref$, len$, results$ = [];
                    for (i$ = 0, len$ = (ref$ = node).length; i$ < len$; ++i$) {
                        element = ref$[i$];
                        results$.push(typesCast(element, typeOf, options));
                    }
                    return results$;
                }())
            };
        }
        function castTuple(node, type, options) {
            var result, i, i$, ref$, len$, types, cast;
            if (toString$.call(node).slice(8, -1) !== 'Array') {
                return {
                    type: 'Nothing'
                };
            }
            result = [];
            i = 0;
            for (i$ = 0, len$ = (ref$ = type.of).length; i$ < len$; ++i$) {
                types = ref$[i$];
                cast = typesCast(node[i], types, options);
                if (toString$.call(cast).slice(8, -1) !== 'Undefined') {
                    result.push(cast);
                }
                i++;
            }
            if (node.length <= i) {
                return {
                    type: 'Just',
                    value: result
                };
            }
            else {
                return {
                    type: 'Nothing'
                };
            }
        }
        function castFields(node, type, options) {
            var typeOf, key, value;
            if (toString$.call(node).slice(8, -1) !== 'Object') {
                return {
                    type: 'Nothing'
                };
            }
            typeOf = type.of;
            return {
                type: 'Just',
                value: (function () {
                    var ref$, resultObj$ = {};
                    for (key in ref$ = node) {
                        value = ref$[key];
                        resultObj$[typesCast(key, [{
                                type: 'String'
                            }], options)] = typesCast(value, typeOf[key] || [{
                                type: '*'
                            }], options);
                    }
                    return resultObj$;
                }())
            };
        }
        function typeCast(node, typeObj, options) {
            var type, structure, castFunc, ref$;
            type = typeObj.type, structure = typeObj.structure;
            if (type) {
                castFunc = ((ref$ = options.customTypes[type]) != null ? ref$.cast : void 8) || types[type];
                if (!castFunc) {
                    throw new Error("Type not defined: " + type + ".");
                }
                return castFunc(node, options, typesCast);
            }
            else {
                switch (structure) {
                    case 'array':
                        return castArray(node, typeObj, options);
                    case 'tuple':
                        return castTuple(node, typeObj, options);
                    case 'fields':
                        return castFields(node, typeObj, options);
                }
            }
        }
        function typesCast(node, types, options) {
            var i$, len$, type, ref$, valueType, value;
            for (i$ = 0, len$ = types.length; i$ < len$; ++i$) {
                type = types[i$];
                ref$ = typeCast(node, type, options), valueType = ref$.type, value = ref$.value;
                if (valueType === 'Nothing') {
                    continue;
                }
                if (parsedTypeCheck([type], value, {
                    customTypes: options.customTypes
                })) {
                    return value;
                }
            }
            throw new Error("Value " + JSON.stringify(node) + " does not type check against " + JSON.stringify(types) + ".");
        }
        module.exports = function (node, types, options) {
            if (!options.explicit && types.length === 1 && types[0].type === 'String') {
                return node;
            }
            return typesCast(node, types, options);
        };
    }).call(this);
},
/* type-check/lib/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    (function () {
        var VERSION, parseType, parsedTypeCheck, typeCheck;
        VERSION = '0.4.0';
        parseType = require(383) /* ./parse-type */;
        parsedTypeCheck = require(384) /* ./check */;
        typeCheck = function (type, input, options) {
            return parsedTypeCheck(parseType(type), input, options);
        };
        module.exports = {
            VERSION: VERSION,
            typeCheck: typeCheck,
            parsedTypeCheck: parsedTypeCheck,
            parseType: parseType
        };
    }).call(this);
},
/* type-check/lib/parse-type.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    (function () {
        var identifierRegex, tokenRegex;
        identifierRegex = /[\$\w]+/;
        function peek(tokens) {
            var token;
            token = tokens[0];
            if (token == null) {
                throw new Error('Unexpected end of input.');
            }
            return token;
        }
        function consumeIdent(tokens) {
            var token;
            token = peek(tokens);
            if (!identifierRegex.test(token)) {
                throw new Error("Expected text, got '" + token + "' instead.");
            }
            return tokens.shift();
        }
        function consumeOp(tokens, op) {
            var token;
            token = peek(tokens);
            if (token !== op) {
                throw new Error("Expected '" + op + "', got '" + token + "' instead.");
            }
            return tokens.shift();
        }
        function maybeConsumeOp(tokens, op) {
            var token;
            token = tokens[0];
            if (token === op) {
                return tokens.shift();
            }
            else {
                return null;
            }
        }
        function consumeArray(tokens) {
            var types;
            consumeOp(tokens, '[');
            if (peek(tokens) === ']') {
                throw new Error("Must specify type of Array - eg. [Type], got [] instead.");
            }
            types = consumeTypes(tokens);
            consumeOp(tokens, ']');
            return {
                structure: 'array',
                of: types
            };
        }
        function consumeTuple(tokens) {
            var components;
            components = [];
            consumeOp(tokens, '(');
            if (peek(tokens) === ')') {
                throw new Error("Tuple must be of at least length 1 - eg. (Type), got () instead.");
            }
            for (;;) {
                components.push(consumeTypes(tokens));
                maybeConsumeOp(tokens, ',');
                if (')' === peek(tokens)) {
                    break;
                }
            }
            consumeOp(tokens, ')');
            return {
                structure: 'tuple',
                of: components
            };
        }
        function consumeFields(tokens) {
            var fields, subset, ref$, key, types;
            fields = {};
            consumeOp(tokens, '{');
            subset = false;
            for (;;) {
                if (maybeConsumeOp(tokens, '...')) {
                    subset = true;
                    break;
                }
                ref$ = consumeField(tokens), key = ref$[0], types = ref$[1];
                fields[key] = types;
                maybeConsumeOp(tokens, ',');
                if ('}' === peek(tokens)) {
                    break;
                }
            }
            consumeOp(tokens, '}');
            return {
                structure: 'fields',
                of: fields,
                subset: subset
            };
        }
        function consumeField(tokens) {
            var key, types;
            key = consumeIdent(tokens);
            consumeOp(tokens, ':');
            types = consumeTypes(tokens);
            return [key, types];
        }
        function maybeConsumeStructure(tokens) {
            switch (tokens[0]) {
                case '[':
                    return consumeArray(tokens);
                case '(':
                    return consumeTuple(tokens);
                case '{':
                    return consumeFields(tokens);
            }
        }
        function consumeType(tokens) {
            var token, wildcard, type, structure;
            token = peek(tokens);
            wildcard = token === '*';
            if (wildcard || identifierRegex.test(token)) {
                type = wildcard
                    ? consumeOp(tokens, '*')
                    : consumeIdent(tokens);
                structure = maybeConsumeStructure(tokens);
                if (structure) {
                    return structure.type = type, structure;
                }
                else {
                    return {
                        type: type
                    };
                }
            }
            else {
                structure = maybeConsumeStructure(tokens);
                if (!structure) {
                    throw new Error("Unexpected character: " + token);
                }
                return structure;
            }
        }
        function consumeTypes(tokens) {
            var lookahead, types, typesSoFar, typeObj, type, structure;
            if ('::' === peek(tokens)) {
                throw new Error("No comment before comment separator '::' found.");
            }
            lookahead = tokens[1];
            if (lookahead != null && lookahead === '::') {
                tokens.shift();
                tokens.shift();
            }
            types = [];
            typesSoFar = {};
            if ('Maybe' === peek(tokens)) {
                tokens.shift();
                types = [
                    {
                        type: 'Undefined'
                    }, {
                        type: 'Null'
                    }
                ];
                typesSoFar = {
                    Undefined: true,
                    Null: true
                };
            }
            for (;;) {
                typeObj = consumeType(tokens), type = typeObj.type, structure = typeObj.structure;
                if (!typesSoFar[type]) {
                    types.push(typeObj);
                }
                if (structure == null) {
                    typesSoFar[type] = true;
                }
                if (!maybeConsumeOp(tokens, '|')) {
                    break;
                }
            }
            return types;
        }
        tokenRegex = RegExp('\\.\\.\\.|::|->|' + identifierRegex.source + '|\\S', 'g');
        module.exports = function (input) {
            var tokens, e;
            if (!input.length) {
                throw new Error('No type specified.');
            }
            tokens = input.match(tokenRegex) || [];
            if (in$('->', tokens)) {
                throw new Error("Function types are not supported.\ To validate that something is a function, you may use 'Function'.");
            }
            try {
                return consumeTypes(tokens);
            }
            catch (e$) {
                e = e$;
                throw new Error(e.message + " - Remaining tokens: " + JSON.stringify(tokens) + " - Initial input: '" + input + "'");
            }
        };
        function in$(x, xs) {
            var i = -1, l = xs.length >>> 0;
            while (++i < l)
                if (x === xs[i])
                    return true;
            return false;
        }
    }).call(this);
},
/* type-check/lib/check.js */ function _(require, module, exports, __esModule, __esExport) {
    // Generated by LiveScript 1.6.0
    (function () {
        var ref$, any, all, isItNaN, types, defaultType, toString$ = {}.toString;
        ref$ = require(375) /* prelude-ls */, any = ref$.any, all = ref$.all, isItNaN = ref$.isItNaN;
        types = {
            Number: {
                typeOf: 'Number',
                validate: function (it) {
                    return !isItNaN(it);
                }
            },
            NaN: {
                typeOf: 'Number',
                validate: isItNaN
            },
            Int: {
                typeOf: 'Number',
                validate: function (it) {
                    return !isItNaN(it) && it % 1 === 0;
                }
            },
            Float: {
                typeOf: 'Number',
                validate: function (it) {
                    return !isItNaN(it);
                }
            },
            Date: {
                typeOf: 'Date',
                validate: function (it) {
                    return !isItNaN(it.getTime());
                }
            }
        };
        defaultType = {
            array: 'Array',
            tuple: 'Array'
        };
        function checkArray(input, type, options) {
            return all(function (it) {
                return checkMultiple(it, type.of, options);
            }, input);
        }
        function checkTuple(input, type, options) {
            var i, i$, ref$, len$, types;
            i = 0;
            for (i$ = 0, len$ = (ref$ = type.of).length; i$ < len$; ++i$) {
                types = ref$[i$];
                if (!checkMultiple(input[i], types, options)) {
                    return false;
                }
                i++;
            }
            return input.length <= i;
        }
        function checkFields(input, type, options) {
            var inputKeys, numInputKeys, k, numOfKeys, key, ref$, types;
            inputKeys = {};
            numInputKeys = 0;
            for (k in input) {
                inputKeys[k] = true;
                numInputKeys++;
            }
            numOfKeys = 0;
            for (key in ref$ = type.of) {
                types = ref$[key];
                if (!checkMultiple(input[key], types, options)) {
                    return false;
                }
                if (inputKeys[key]) {
                    numOfKeys++;
                }
            }
            return type.subset || numInputKeys === numOfKeys;
        }
        function checkStructure(input, type, options) {
            if (!(input instanceof Object)) {
                return false;
            }
            switch (type.structure) {
                case 'fields':
                    return checkFields(input, type, options);
                case 'array':
                    return checkArray(input, type, options);
                case 'tuple':
                    return checkTuple(input, type, options);
            }
        }
        function check(input, typeObj, options) {
            var type, structure, setting, that;
            type = typeObj.type, structure = typeObj.structure;
            if (type) {
                if (type === '*') {
                    return true;
                }
                setting = options.customTypes[type] || types[type];
                if (setting) {
                    return (setting.typeOf === void 8 || setting.typeOf === toString$.call(input).slice(8, -1)) && setting.validate(input);
                }
                else {
                    return type === toString$.call(input).slice(8, -1) && (!structure || checkStructure(input, typeObj, options));
                }
            }
            else if (structure) {
                if (that = defaultType[structure]) {
                    if (that !== toString$.call(input).slice(8, -1)) {
                        return false;
                    }
                }
                return checkStructure(input, typeObj, options);
            }
            else {
                throw new Error("No type defined. Input: " + input + ".");
            }
        }
        function checkMultiple(input, types, options) {
            if (toString$.call(types).slice(8, -1) !== 'Array') {
                throw new Error("Types must be in an array. Input: " + input + ".");
            }
            return any(function (it) {
                return check(input, it, options);
            }, types);
        }
        module.exports = function (parsedType, input, options) {
            options == null && (options = {});
            if (options.customTypes == null) {
                options.customTypes = {};
            }
            return checkMultiple(input, parsedType, options);
        };
    }).call(this);
},
/* eslint/lib/linter/node-event-generator.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const esquery = require(386) /* esquery */;
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    /**
     * An object describing an AST selector
     * @typedef {Object} ASTSelector
     * @property {string} rawSelector The string that was parsed into this selector
     * @property {boolean} isExit `true` if this should be emitted when exiting the node rather than when entering
     * @property {Object} parsedSelector An object (from esquery) describing the matching behavior of the selector
     * @property {string[]|null} listenerTypes A list of node types that could possibly cause the selector to match,
     * or `null` if all node types could cause a match
     * @property {number} attributeCount The total number of classes, pseudo-classes, and attribute queries in this selector
     * @property {number} identifierCount The total number of identifier queries in this selector
     */
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Computes the union of one or more arrays
     * @param {...any[]} arrays One or more arrays to union
     * @returns {any[]} The union of the input arrays
     */
    function union(...arrays) {
        // TODO(stephenwade): Replace this with arrays.flat() when we drop support for Node v10
        return [...new Set([].concat(...arrays))];
    }
    /**
     * Computes the intersection of one or more arrays
     * @param {...any[]} arrays One or more arrays to intersect
     * @returns {any[]} The intersection of the input arrays
     */
    function intersection(...arrays) {
        if (arrays.length === 0) {
            return [];
        }
        let result = [...new Set(arrays[0])];
        for (const array of arrays.slice(1)) {
            result = result.filter(x => array.includes(x));
        }
        return result;
    }
    /**
     * Gets the possible types of a selector
     * @param {Object} parsedSelector An object (from esquery) describing the matching behavior of the selector
     * @returns {string[]|null} The node types that could possibly trigger this selector, or `null` if all node types could trigger it
     */
    function getPossibleTypes(parsedSelector) {
        switch (parsedSelector.type) {
            case "identifier":
                return [parsedSelector.value];
            case "matches": {
                const typesForComponents = parsedSelector.selectors.map(getPossibleTypes);
                if (typesForComponents.every(Boolean)) {
                    return union(...typesForComponents);
                }
                return null;
            }
            case "compound": {
                const typesForComponents = parsedSelector.selectors.map(getPossibleTypes).filter(typesForComponent => typesForComponent);
                // If all of the components could match any type, then the compound could also match any type.
                if (!typesForComponents.length) {
                    return null;
                }
                /*
                 * If at least one of the components could only match a particular type, the compound could only match
                 * the intersection of those types.
                 */
                return intersection(...typesForComponents);
            }
            case "child":
            case "descendant":
            case "sibling":
            case "adjacent":
                return getPossibleTypes(parsedSelector.right);
            default:
                return null;
        }
    }
    /**
     * Counts the number of class, pseudo-class, and attribute queries in this selector
     * @param {Object} parsedSelector An object (from esquery) describing the selector's matching behavior
     * @returns {number} The number of class, pseudo-class, and attribute queries in this selector
     */
    function countClassAttributes(parsedSelector) {
        switch (parsedSelector.type) {
            case "child":
            case "descendant":
            case "sibling":
            case "adjacent":
                return countClassAttributes(parsedSelector.left) + countClassAttributes(parsedSelector.right);
            case "compound":
            case "not":
            case "matches":
                return parsedSelector.selectors.reduce((sum, childSelector) => sum + countClassAttributes(childSelector), 0);
            case "attribute":
            case "field":
            case "nth-child":
            case "nth-last-child":
                return 1;
            default:
                return 0;
        }
    }
    /**
     * Counts the number of identifier queries in this selector
     * @param {Object} parsedSelector An object (from esquery) describing the selector's matching behavior
     * @returns {number} The number of identifier queries
     */
    function countIdentifiers(parsedSelector) {
        switch (parsedSelector.type) {
            case "child":
            case "descendant":
            case "sibling":
            case "adjacent":
                return countIdentifiers(parsedSelector.left) + countIdentifiers(parsedSelector.right);
            case "compound":
            case "not":
            case "matches":
                return parsedSelector.selectors.reduce((sum, childSelector) => sum + countIdentifiers(childSelector), 0);
            case "identifier":
                return 1;
            default:
                return 0;
        }
    }
    /**
     * Compares the specificity of two selector objects, with CSS-like rules.
     * @param {ASTSelector} selectorA An AST selector descriptor
     * @param {ASTSelector} selectorB Another AST selector descriptor
     * @returns {number}
     * a value less than 0 if selectorA is less specific than selectorB
     * a value greater than 0 if selectorA is more specific than selectorB
     * a value less than 0 if selectorA and selectorB have the same specificity, and selectorA <= selectorB alphabetically
     * a value greater than 0 if selectorA and selectorB have the same specificity, and selectorA > selectorB alphabetically
     */
    function compareSpecificity(selectorA, selectorB) {
        return selectorA.attributeCount - selectorB.attributeCount ||
            selectorA.identifierCount - selectorB.identifierCount ||
            (selectorA.rawSelector <= selectorB.rawSelector ? -1 : 1);
    }
    /**
     * Parses a raw selector string, and throws a useful error if parsing fails.
     * @param {string} rawSelector A raw AST selector
     * @returns {Object} An object (from esquery) describing the matching behavior of this selector
     * @throws {Error} An error if the selector is invalid
     */
    function tryParseSelector(rawSelector) {
        try {
            return esquery.parse(rawSelector.replace(/:exit$/u, ""));
        }
        catch (err) {
            if (err.location && err.location.start && typeof err.location.start.offset === "number") {
                throw new SyntaxError(`Syntax error in selector "${rawSelector}" at position ${err.location.start.offset}: ${err.message}`);
            }
            throw err;
        }
    }
    const selectorCache = new Map();
    /**
     * Parses a raw selector string, and returns the parsed selector along with specificity and type information.
     * @param {string} rawSelector A raw AST selector
     * @returns {ASTSelector} A selector descriptor
     */
    function parseSelector(rawSelector) {
        if (selectorCache.has(rawSelector)) {
            return selectorCache.get(rawSelector);
        }
        const parsedSelector = tryParseSelector(rawSelector);
        const result = {
            rawSelector,
            isExit: rawSelector.endsWith(":exit"),
            parsedSelector,
            listenerTypes: getPossibleTypes(parsedSelector),
            attributeCount: countClassAttributes(parsedSelector),
            identifierCount: countIdentifiers(parsedSelector)
        };
        selectorCache.set(rawSelector, result);
        return result;
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * The event generator for AST nodes.
     * This implements below interface.
     *
     * ```ts
     * interface EventGenerator {
     *     emitter: SafeEmitter;
     *     enterNode(node: ASTNode): void;
     *     leaveNode(node: ASTNode): void;
     * }
     * ```
     */
    class NodeEventGenerator {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {SafeEmitter} emitter
         * An SafeEmitter which is the destination of events. This emitter must already
         * have registered listeners for all of the events that it needs to listen for.
         * (See lib/linter/safe-emitter.js for more details on `SafeEmitter`.)
         * @param {ESQueryOptions} esqueryOptions `esquery` options for traversing custom nodes.
         * @returns {NodeEventGenerator} new instance
         */
        constructor(emitter, esqueryOptions) {
            this.emitter = emitter;
            this.esqueryOptions = esqueryOptions;
            this.currentAncestry = [];
            this.enterSelectorsByNodeType = new Map();
            this.exitSelectorsByNodeType = new Map();
            this.anyTypeEnterSelectors = [];
            this.anyTypeExitSelectors = [];
            emitter.eventNames().forEach(rawSelector => {
                const selector = parseSelector(rawSelector);
                if (selector.listenerTypes) {
                    const typeMap = selector.isExit ? this.exitSelectorsByNodeType : this.enterSelectorsByNodeType;
                    selector.listenerTypes.forEach(nodeType => {
                        if (!typeMap.has(nodeType)) {
                            typeMap.set(nodeType, []);
                        }
                        typeMap.get(nodeType).push(selector);
                    });
                    return;
                }
                const selectors = selector.isExit ? this.anyTypeExitSelectors : this.anyTypeEnterSelectors;
                selectors.push(selector);
            });
            this.anyTypeEnterSelectors.sort(compareSpecificity);
            this.anyTypeExitSelectors.sort(compareSpecificity);
            this.enterSelectorsByNodeType.forEach(selectorList => selectorList.sort(compareSpecificity));
            this.exitSelectorsByNodeType.forEach(selectorList => selectorList.sort(compareSpecificity));
        }
        /**
         * Checks a selector against a node, and emits it if it matches
         * @param {ASTNode} node The node to check
         * @param {ASTSelector} selector An AST selector descriptor
         * @returns {void}
         */
        applySelector(node, selector) {
            if (esquery.matches(node, selector.parsedSelector, this.currentAncestry, this.esqueryOptions)) {
                this.emitter.emit(selector.rawSelector, node);
            }
        }
        /**
         * Applies all appropriate selectors to a node, in specificity order
         * @param {ASTNode} node The node to check
         * @param {boolean} isExit `false` if the node is currently being entered, `true` if it's currently being exited
         * @returns {void}
         */
        applySelectors(node, isExit) {
            const selectorsByNodeType = (isExit ? this.exitSelectorsByNodeType : this.enterSelectorsByNodeType).get(node.type) || [];
            const anyTypeSelectors = isExit ? this.anyTypeExitSelectors : this.anyTypeEnterSelectors;
            /*
             * selectorsByNodeType and anyTypeSelectors were already sorted by specificity in the constructor.
             * Iterate through each of them, applying selectors in the right order.
             */
            let selectorsByTypeIndex = 0;
            let anyTypeSelectorsIndex = 0;
            while (selectorsByTypeIndex < selectorsByNodeType.length || anyTypeSelectorsIndex < anyTypeSelectors.length) {
                if (selectorsByTypeIndex >= selectorsByNodeType.length ||
                    anyTypeSelectorsIndex < anyTypeSelectors.length &&
                        compareSpecificity(anyTypeSelectors[anyTypeSelectorsIndex], selectorsByNodeType[selectorsByTypeIndex]) < 0) {
                    this.applySelector(node, anyTypeSelectors[anyTypeSelectorsIndex++]);
                }
                else {
                    this.applySelector(node, selectorsByNodeType[selectorsByTypeIndex++]);
                }
            }
        }
        /**
         * Emits an event of entering AST node.
         * @param {ASTNode} node A node which was entered.
         * @returns {void}
         */
        enterNode(node) {
            if (node.parent) {
                this.currentAncestry.unshift(node.parent);
            }
            this.applySelectors(node, false);
        }
        /**
         * Emits an event of leaving AST node.
         * @param {ASTNode} node A node which was left.
         * @returns {void}
         */
        leaveNode(node) {
            this.applySelectors(node, true);
            this.currentAncestry.shift();
        }
    }
    module.exports = NodeEventGenerator;
},
/* esquery/dist/esquery.min.js */ function _(require, module, exports, __esModule, __esExport) {
    !function (e, t) { "object" == typeof exports && "undefined" != typeof module ? module.exports = t() : "function" == typeof define && define.amd ? define(t) : (e = e || self).esquery = t(); }(this, (function () {
        "use strict";
        function e(t) { return (e = "function" == typeof Symbol && "symbol" == typeof Symbol.iterator ? function (e) { return typeof e; } : function (e) { return e && "function" == typeof Symbol && e.constructor === Symbol && e !== Symbol.prototype ? "symbol" : typeof e; })(t); }
        function t(e, t) { return function (e) { if (Array.isArray(e))
            return e; }(e) || function (e, t) { if ("undefined" == typeof Symbol || !(Symbol.iterator in Object(e)))
            return; var r = [], n = !0, o = !1, a = void 0; try {
            for (var i, s = e[Symbol.iterator](); !(n = (i = s.next()).done) && (r.push(i.value), !t || r.length !== t); n = !0)
                ;
        }
        catch (e) {
            o = !0, a = e;
        }
        finally {
            try {
                n || null == s.return || s.return();
            }
            finally {
                if (o)
                    throw a;
            }
        } return r; }(e, t) || n(e, t) || function () { throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); }(); }
        function r(e) { return function (e) { if (Array.isArray(e))
            return o(e); }(e) || function (e) { if ("undefined" != typeof Symbol && Symbol.iterator in Object(e))
            return Array.from(e); }(e) || n(e) || function () { throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); }(); }
        function n(e, t) { if (e) {
            if ("string" == typeof e)
                return o(e, t);
            var r = Object.prototype.toString.call(e).slice(8, -1);
            return "Object" === r && e.constructor && (r = e.constructor.name), "Map" === r || "Set" === r ? Array.from(e) : "Arguments" === r || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(r) ? o(e, t) : void 0;
        } }
        function o(e, t) { (null == t || t > e.length) && (t = e.length); for (var r = 0, n = new Array(t); r < t; r++)
            n[r] = e[r]; return n; }
        function a(e, t) { var r; if ("undefined" == typeof Symbol || null == e[Symbol.iterator]) {
            if (Array.isArray(e) || (r = n(e)) || t && e && "number" == typeof e.length) {
                r && (e = r);
                var o = 0, a = function () { };
                return { s: a, n: function () { return o >= e.length ? { done: !0 } : { done: !1, value: e[o++] }; }, e: function (e) { throw e; }, f: a };
            }
            throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
        } var i, s = !0, l = !1; return { s: function () { r = e[Symbol.iterator](); }, n: function () { var e = r.next(); return s = e.done, e; }, e: function (e) { l = !0, i = e; }, f: function () { try {
                s || null == r.return || r.return();
            }
            finally {
                if (l)
                    throw i;
            } } }; }
        "undefined" != typeof globalThis ? globalThis : "undefined" != typeof window ? window : "undefined" != typeof global ? global : "undefined" != typeof self && self;
        function i(e, t) { return e(t = { exports: {} }, t.exports), t.exports; }
        var s = i((function (e, t) { !function e(t) { var r, n, o, a, i, s; function l(e) { var t, r, n = {}; for (t in e)
            e.hasOwnProperty(t) && (r = e[t], n[t] = "object" == typeof r && null !== r ? l(r) : r); return n; } function u(e, t) { this.parent = e, this.key = t; } function c(e, t, r, n) { this.node = e, this.path = t, this.wrap = r, this.ref = n; } function f() { } function p(e) { return null != e && ("object" == typeof e && "string" == typeof e.type); } function h(e, t) { return (e === r.ObjectExpression || e === r.ObjectPattern) && "properties" === t; } function y(e, t) { for (var r = e.length - 1; r >= 0; --r)
            if (e[r].node === t)
                return !0; return !1; } function d(e, t) { return (new f).traverse(e, t); } function m(e, t) { var r; return r = function (e, t) { var r, n, o, a; for (n = e.length, o = 0; n;)
            t(e[a = o + (r = n >>> 1)]) ? n = r : (o = a + 1, n -= r + 1); return o; }(t, (function (t) { return t.range[0] > e.range[0]; })), e.extendedRange = [e.range[0], e.range[1]], r !== t.length && (e.extendedRange[1] = t[r].range[0]), (r -= 1) >= 0 && (e.extendedRange[0] = t[r].range[1]), e; } return r = { AssignmentExpression: "AssignmentExpression", AssignmentPattern: "AssignmentPattern", ArrayExpression: "ArrayExpression", ArrayPattern: "ArrayPattern", ArrowFunctionExpression: "ArrowFunctionExpression", AwaitExpression: "AwaitExpression", BlockStatement: "BlockStatement", BinaryExpression: "BinaryExpression", BreakStatement: "BreakStatement", CallExpression: "CallExpression", CatchClause: "CatchClause", ChainExpression: "ChainExpression", ClassBody: "ClassBody", ClassDeclaration: "ClassDeclaration", ClassExpression: "ClassExpression", ComprehensionBlock: "ComprehensionBlock", ComprehensionExpression: "ComprehensionExpression", ConditionalExpression: "ConditionalExpression", ContinueStatement: "ContinueStatement", DebuggerStatement: "DebuggerStatement", DirectiveStatement: "DirectiveStatement", DoWhileStatement: "DoWhileStatement", EmptyStatement: "EmptyStatement", ExportAllDeclaration: "ExportAllDeclaration", ExportDefaultDeclaration: "ExportDefaultDeclaration", ExportNamedDeclaration: "ExportNamedDeclaration", ExportSpecifier: "ExportSpecifier", ExpressionStatement: "ExpressionStatement", ForStatement: "ForStatement", ForInStatement: "ForInStatement", ForOfStatement: "ForOfStatement", FunctionDeclaration: "FunctionDeclaration", FunctionExpression: "FunctionExpression", GeneratorExpression: "GeneratorExpression", Identifier: "Identifier", IfStatement: "IfStatement", ImportExpression: "ImportExpression", ImportDeclaration: "ImportDeclaration", ImportDefaultSpecifier: "ImportDefaultSpecifier", ImportNamespaceSpecifier: "ImportNamespaceSpecifier", ImportSpecifier: "ImportSpecifier", Literal: "Literal", LabeledStatement: "LabeledStatement", LogicalExpression: "LogicalExpression", MemberExpression: "MemberExpression", MetaProperty: "MetaProperty", MethodDefinition: "MethodDefinition", ModuleSpecifier: "ModuleSpecifier", NewExpression: "NewExpression", ObjectExpression: "ObjectExpression", ObjectPattern: "ObjectPattern", Program: "Program", Property: "Property", RestElement: "RestElement", ReturnStatement: "ReturnStatement", SequenceExpression: "SequenceExpression", SpreadElement: "SpreadElement", Super: "Super", SwitchStatement: "SwitchStatement", SwitchCase: "SwitchCase", TaggedTemplateExpression: "TaggedTemplateExpression", TemplateElement: "TemplateElement", TemplateLiteral: "TemplateLiteral", ThisExpression: "ThisExpression", ThrowStatement: "ThrowStatement", TryStatement: "TryStatement", UnaryExpression: "UnaryExpression", UpdateExpression: "UpdateExpression", VariableDeclaration: "VariableDeclaration", VariableDeclarator: "VariableDeclarator", WhileStatement: "WhileStatement", WithStatement: "WithStatement", YieldExpression: "YieldExpression" }, o = { AssignmentExpression: ["left", "right"], AssignmentPattern: ["left", "right"], ArrayExpression: ["elements"], ArrayPattern: ["elements"], ArrowFunctionExpression: ["params", "body"], AwaitExpression: ["argument"], BlockStatement: ["body"], BinaryExpression: ["left", "right"], BreakStatement: ["label"], CallExpression: ["callee", "arguments"], CatchClause: ["param", "body"], ChainExpression: ["expression"], ClassBody: ["body"], ClassDeclaration: ["id", "superClass", "body"], ClassExpression: ["id", "superClass", "body"], ComprehensionBlock: ["left", "right"], ComprehensionExpression: ["blocks", "filter", "body"], ConditionalExpression: ["test", "consequent", "alternate"], ContinueStatement: ["label"], DebuggerStatement: [], DirectiveStatement: [], DoWhileStatement: ["body", "test"], EmptyStatement: [], ExportAllDeclaration: ["source"], ExportDefaultDeclaration: ["declaration"], ExportNamedDeclaration: ["declaration", "specifiers", "source"], ExportSpecifier: ["exported", "local"], ExpressionStatement: ["expression"], ForStatement: ["init", "test", "update", "body"], ForInStatement: ["left", "right", "body"], ForOfStatement: ["left", "right", "body"], FunctionDeclaration: ["id", "params", "body"], FunctionExpression: ["id", "params", "body"], GeneratorExpression: ["blocks", "filter", "body"], Identifier: [], IfStatement: ["test", "consequent", "alternate"], ImportExpression: ["source"], ImportDeclaration: ["specifiers", "source"], ImportDefaultSpecifier: ["local"], ImportNamespaceSpecifier: ["local"], ImportSpecifier: ["imported", "local"], Literal: [], LabeledStatement: ["label", "body"], LogicalExpression: ["left", "right"], MemberExpression: ["object", "property"], MetaProperty: ["meta", "property"], MethodDefinition: ["key", "value"], ModuleSpecifier: [], NewExpression: ["callee", "arguments"], ObjectExpression: ["properties"], ObjectPattern: ["properties"], Program: ["body"], Property: ["key", "value"], RestElement: ["argument"], ReturnStatement: ["argument"], SequenceExpression: ["expressions"], SpreadElement: ["argument"], Super: [], SwitchStatement: ["discriminant", "cases"], SwitchCase: ["test", "consequent"], TaggedTemplateExpression: ["tag", "quasi"], TemplateElement: [], TemplateLiteral: ["quasis", "expressions"], ThisExpression: [], ThrowStatement: ["argument"], TryStatement: ["block", "handler", "finalizer"], UnaryExpression: ["argument"], UpdateExpression: ["argument"], VariableDeclaration: ["declarations"], VariableDeclarator: ["id", "init"], WhileStatement: ["test", "body"], WithStatement: ["object", "body"], YieldExpression: ["argument"] }, n = { Break: a = {}, Skip: i = {}, Remove: s = {} }, u.prototype.replace = function (e) { this.parent[this.key] = e; }, u.prototype.remove = function () { return Array.isArray(this.parent) ? (this.parent.splice(this.key, 1), !0) : (this.replace(null), !1); }, f.prototype.path = function () { var e, t, r, n, o; function a(e, t) { if (Array.isArray(t))
            for (r = 0, n = t.length; r < n; ++r)
                e.push(t[r]);
        else
            e.push(t); } if (!this.__current.path)
            return null; for (o = [], e = 2, t = this.__leavelist.length; e < t; ++e)
            a(o, this.__leavelist[e].path); return a(o, this.__current.path), o; }, f.prototype.type = function () { return this.current().type || this.__current.wrap; }, f.prototype.parents = function () { var e, t, r; for (r = [], e = 1, t = this.__leavelist.length; e < t; ++e)
            r.push(this.__leavelist[e].node); return r; }, f.prototype.current = function () { return this.__current.node; }, f.prototype.__execute = function (e, t) { var r, n; return n = void 0, r = this.__current, this.__current = t, this.__state = null, e && (n = e.call(this, t.node, this.__leavelist[this.__leavelist.length - 1].node)), this.__current = r, n; }, f.prototype.notify = function (e) { this.__state = e; }, f.prototype.skip = function () { this.notify(i); }, f.prototype.break = function () { this.notify(a); }, f.prototype.remove = function () { this.notify(s); }, f.prototype.__initialize = function (e, t) { this.visitor = t, this.root = e, this.__worklist = [], this.__leavelist = [], this.__current = null, this.__state = null, this.__fallback = null, "iteration" === t.fallback ? this.__fallback = Object.keys : "function" == typeof t.fallback && (this.__fallback = t.fallback), this.__keys = o, t.keys && (this.__keys = Object.assign(Object.create(this.__keys), t.keys)); }, f.prototype.traverse = function (e, t) { var r, n, o, s, l, u, f, d, m, x, v, g; for (this.__initialize(e, t), g = {}, r = this.__worklist, n = this.__leavelist, r.push(new c(e, null, null, null)), n.push(new c(null, null, null, null)); r.length;)
            if ((o = r.pop()) !== g) {
                if (o.node) {
                    if (u = this.__execute(t.enter, o), this.__state === a || u === a)
                        return;
                    if (r.push(g), n.push(o), this.__state === i || u === i)
                        continue;
                    if (l = (s = o.node).type || o.wrap, !(x = this.__keys[l])) {
                        if (!this.__fallback)
                            throw new Error("Unknown node type " + l + ".");
                        x = this.__fallback(s);
                    }
                    for (d = x.length; (d -= 1) >= 0;)
                        if (v = s[f = x[d]])
                            if (Array.isArray(v)) {
                                for (m = v.length; (m -= 1) >= 0;)
                                    if (v[m] && !y(n, v[m])) {
                                        if (h(l, x[d]))
                                            o = new c(v[m], [f, m], "Property", null);
                                        else {
                                            if (!p(v[m]))
                                                continue;
                                            o = new c(v[m], [f, m], null, null);
                                        }
                                        r.push(o);
                                    }
                            }
                            else if (p(v)) {
                                if (y(n, v))
                                    continue;
                                r.push(new c(v, f, null, null));
                            }
                }
            }
            else if (o = n.pop(), u = this.__execute(t.leave, o), this.__state === a || u === a)
                return; }, f.prototype.replace = function (e, t) { var r, n, o, l, f, y, d, m, x, v, g, b, A; function E(e) { var t, n, o, a; if (e.ref.remove())
            for (n = e.ref.key, a = e.ref.parent, t = r.length; t--;)
                if ((o = r[t]).ref && o.ref.parent === a) {
                    if (o.ref.key < n)
                        break;
                    --o.ref.key;
                } } for (this.__initialize(e, t), g = {}, r = this.__worklist, n = this.__leavelist, y = new c(e, null, null, new u(b = { root: e }, "root")), r.push(y), n.push(y); r.length;)
            if ((y = r.pop()) !== g) {
                if (void 0 !== (f = this.__execute(t.enter, y)) && f !== a && f !== i && f !== s && (y.ref.replace(f), y.node = f), this.__state !== s && f !== s || (E(y), y.node = null), this.__state === a || f === a)
                    return b.root;
                if ((o = y.node) && (r.push(g), n.push(y), this.__state !== i && f !== i)) {
                    if (l = o.type || y.wrap, !(x = this.__keys[l])) {
                        if (!this.__fallback)
                            throw new Error("Unknown node type " + l + ".");
                        x = this.__fallback(o);
                    }
                    for (d = x.length; (d -= 1) >= 0;)
                        if (v = o[A = x[d]])
                            if (Array.isArray(v)) {
                                for (m = v.length; (m -= 1) >= 0;)
                                    if (v[m]) {
                                        if (h(l, x[d]))
                                            y = new c(v[m], [A, m], "Property", new u(v, m));
                                        else {
                                            if (!p(v[m]))
                                                continue;
                                            y = new c(v[m], [A, m], null, new u(v, m));
                                        }
                                        r.push(y);
                                    }
                            }
                            else
                                p(v) && r.push(new c(v, A, null, new u(o, A)));
                }
            }
            else if (y = n.pop(), void 0 !== (f = this.__execute(t.leave, y)) && f !== a && f !== i && f !== s && y.ref.replace(f), this.__state !== s && f !== s || E(y), this.__state === a || f === a)
                return b.root; return b.root; }, t.Syntax = r, t.traverse = d, t.replace = function (e, t) { return (new f).replace(e, t); }, t.attachComments = function (e, t, r) { var o, a, i, s, u = []; if (!e.range)
            throw new Error("attachComments needs range information"); if (!r.length) {
            if (t.length) {
                for (i = 0, a = t.length; i < a; i += 1)
                    (o = l(t[i])).extendedRange = [0, e.range[0]], u.push(o);
                e.leadingComments = u;
            }
            return e;
        } for (i = 0, a = t.length; i < a; i += 1)
            u.push(m(l(t[i]), r)); return s = 0, d(e, { enter: function (e) { for (var t; s < u.length && !((t = u[s]).extendedRange[1] > e.range[0]);)
                t.extendedRange[1] === e.range[0] ? (e.leadingComments || (e.leadingComments = []), e.leadingComments.push(t), u.splice(s, 1)) : s += 1; return s === u.length ? n.Break : u[s].extendedRange[0] > e.range[1] ? n.Skip : void 0; } }), s = 0, d(e, { leave: function (e) { for (var t; s < u.length && (t = u[s], !(e.range[1] < t.extendedRange[0]));)
                e.range[1] === t.extendedRange[0] ? (e.trailingComments || (e.trailingComments = []), e.trailingComments.push(t), u.splice(s, 1)) : s += 1; return s === u.length ? n.Break : u[s].extendedRange[0] > e.range[1] ? n.Skip : void 0; } }), e; }, t.VisitorKeys = o, t.VisitorOption = n, t.Controller = f, t.cloneEnvironment = function () { return e({}); }, t; }(t); })), l = i((function (e) { e.exports && (e.exports = function () { function e(t, r, n, o) { this.message = t, this.expected = r, this.found = n, this.location = o, this.name = "SyntaxError", "function" == typeof Error.captureStackTrace && Error.captureStackTrace(this, e); } return function (e, t) { function r() { this.constructor = e; } r.prototype = t.prototype, e.prototype = new r; }(e, Error), e.buildMessage = function (e, t) { var r = { literal: function (e) { return '"' + o(e.text) + '"'; }, class: function (e) { var t, r = ""; for (t = 0; t < e.parts.length; t++)
                r += e.parts[t] instanceof Array ? a(e.parts[t][0]) + "-" + a(e.parts[t][1]) : a(e.parts[t]); return "[" + (e.inverted ? "^" : "") + r + "]"; }, any: function (e) { return "any character"; }, end: function (e) { return "end of input"; }, other: function (e) { return e.description; } }; function n(e) { return e.charCodeAt(0).toString(16).toUpperCase(); } function o(e) { return e.replace(/\\/g, "\\\\").replace(/"/g, '\\"').replace(/\0/g, "\\0").replace(/\t/g, "\\t").replace(/\n/g, "\\n").replace(/\r/g, "\\r").replace(/[\x00-\x0F]/g, (function (e) { return "\\x0" + n(e); })).replace(/[\x10-\x1F\x7F-\x9F]/g, (function (e) { return "\\x" + n(e); })); } function a(e) { return e.replace(/\\/g, "\\\\").replace(/\]/g, "\\]").replace(/\^/g, "\\^").replace(/-/g, "\\-").replace(/\0/g, "\\0").replace(/\t/g, "\\t").replace(/\n/g, "\\n").replace(/\r/g, "\\r").replace(/[\x00-\x0F]/g, (function (e) { return "\\x0" + n(e); })).replace(/[\x10-\x1F\x7F-\x9F]/g, (function (e) { return "\\x" + n(e); })); } return "Expected " + function (e) { var t, n, o, a = new Array(e.length); for (t = 0; t < e.length; t++)
            a[t] = (o = e[t], r[o.type](o)); if (a.sort(), a.length > 0) {
            for (t = 1, n = 1; t < a.length; t++)
                a[t - 1] !== a[t] && (a[n] = a[t], n++);
            a.length = n;
        } switch (a.length) {
            case 1: return a[0];
            case 2: return a[0] + " or " + a[1];
            default: return a.slice(0, -1).join(", ") + ", or " + a[a.length - 1];
        } }(e) + " but " + function (e) { return e ? '"' + o(e) + '"' : "end of input"; }(t) + " found."; }, { SyntaxError: e, parse: function (t, r) { r = void 0 !== r ? r : {}; var n, o, a, i, s = {}, l = { start: be }, u = be, c = de(" ", !1), f = /^[^ [\],():#!=><~+.]/, p = me([" ", "[", "]", ",", "(", ")", ":", "#", "!", "=", ">", "<", "~", "+", "."], !0, !1), h = de(">", !1), y = de("~", !1), d = de("+", !1), m = de(",", !1), x = de("!", !1), v = de("*", !1), g = de("#", !1), b = de("[", !1), A = de("]", !1), E = /^[><!]/, S = me([">", "<", "!"], !1, !1), _ = de("=", !1), w = function (e) { return (e || "") + "="; }, C = /^[><]/, P = me([">", "<"], !1, !1), k = de(".", !1), D = function (e, t, r) { return { type: "attribute", name: e, operator: t, value: r }; }, j = de('"', !1), I = /^[^\\"]/, T = me(["\\", '"'], !0, !1), F = de("\\", !1), L = { type: "any" }, O = function (e, t) { return e + t; }, R = function (e) { return { type: "literal", value: (t = e.join(""), t.replace(/\\(.)/g, (function (e, t) { switch (t) {
                    case "b": return "\b";
                    case "f": return "\f";
                    case "n": return "\n";
                    case "r": return "\r";
                    case "t": return "\t";
                    case "v": return "\v";
                    default: return t;
                } }))) }; var t; }, B = de("'", !1), M = /^[^\\']/, U = me(["\\", "'"], !0, !1), q = /^[0-9]/, V = me([["0", "9"]], !1, !1), N = de("type(", !1), W = /^[^ )]/, K = me([" ", ")"], !0, !1), G = de(")", !1), z = /^[imsu]/, H = me(["i", "m", "s", "u"], !1, !1), Y = de("/", !1), $ = /^[^\/]/, J = me(["/"], !0, !1), Q = de(":not(", !1), X = de(":matches(", !1), Z = de(":has(", !1), ee = de(":first-child", !1), te = de(":last-child", !1), re = de(":nth-child(", !1), ne = de(":nth-last-child(", !1), oe = de(":", !1), ae = de("statement", !0), ie = de("expression", !0), se = de("declaration", !0), le = de("function", !0), ue = de("pattern", !0), ce = 0, fe = [{ line: 1, column: 1 }], pe = 0, he = [], ye = {}; if ("startRule" in r) {
                if (!(r.startRule in l))
                    throw new Error("Can't start parsing from rule \"" + r.startRule + '".');
                u = l[r.startRule];
            } function de(e, t) { return { type: "literal", text: e, ignoreCase: t }; } function me(e, t, r) { return { type: "class", parts: e, inverted: t, ignoreCase: r }; } function xe(e) { var r, n = fe[e]; if (n)
                return n; for (r = e - 1; !fe[r];)
                r--; for (n = { line: (n = fe[r]).line, column: n.column }; r < e;)
                10 === t.charCodeAt(r) ? (n.line++, n.column = 1) : n.column++, r++; return fe[e] = n, n; } function ve(e, t) { var r = xe(e), n = xe(t); return { start: { offset: e, line: r.line, column: r.column }, end: { offset: t, line: n.line, column: n.column } }; } function ge(e) { ce < pe || (ce > pe && (pe = ce, he = []), he.push(e)); } function be() { var e, t, r, n, o = 30 * ce + 0, a = ye[o]; return a ? (ce = a.nextPos, a.result) : (e = ce, (t = Ae()) !== s && (r = _e()) !== s && Ae() !== s ? e = t = 1 === (n = r).length ? n[0] : { type: "matches", selectors: n } : (ce = e, e = s), e === s && (e = ce, (t = Ae()) !== s && (t = void 0), e = t), ye[o] = { nextPos: ce, result: e }, e); } function Ae() { var e, r, n = 30 * ce + 1, o = ye[n]; if (o)
                return ce = o.nextPos, o.result; for (e = [], 32 === t.charCodeAt(ce) ? (r = " ", ce++) : (r = s, ge(c)); r !== s;)
                e.push(r), 32 === t.charCodeAt(ce) ? (r = " ", ce++) : (r = s, ge(c)); return ye[n] = { nextPos: ce, result: e }, e; } function Ee() { var e, r, n, o = 30 * ce + 2, a = ye[o]; if (a)
                return ce = a.nextPos, a.result; if (r = [], f.test(t.charAt(ce)) ? (n = t.charAt(ce), ce++) : (n = s, ge(p)), n !== s)
                for (; n !== s;)
                    r.push(n), f.test(t.charAt(ce)) ? (n = t.charAt(ce), ce++) : (n = s, ge(p));
            else
                r = s; return r !== s && (r = r.join("")), e = r, ye[o] = { nextPos: ce, result: e }, e; } function Se() { var e, r, n, o = 30 * ce + 3, a = ye[o]; return a ? (ce = a.nextPos, a.result) : (e = ce, (r = Ae()) !== s ? (62 === t.charCodeAt(ce) ? (n = ">", ce++) : (n = s, ge(h)), n !== s && Ae() !== s ? e = r = "child" : (ce = e, e = s)) : (ce = e, e = s), e === s && (e = ce, (r = Ae()) !== s ? (126 === t.charCodeAt(ce) ? (n = "~", ce++) : (n = s, ge(y)), n !== s && Ae() !== s ? e = r = "sibling" : (ce = e, e = s)) : (ce = e, e = s), e === s && (e = ce, (r = Ae()) !== s ? (43 === t.charCodeAt(ce) ? (n = "+", ce++) : (n = s, ge(d)), n !== s && Ae() !== s ? e = r = "adjacent" : (ce = e, e = s)) : (ce = e, e = s), e === s && (e = ce, 32 === t.charCodeAt(ce) ? (r = " ", ce++) : (r = s, ge(c)), r !== s && (n = Ae()) !== s ? e = r = "descendant" : (ce = e, e = s)))), ye[o] = { nextPos: ce, result: e }, e); } function _e() { var e, r, n, o, a, i, l, u, c = 30 * ce + 4, f = ye[c]; if (f)
                return ce = f.nextPos, f.result; if (e = ce, (r = we()) !== s) {
                for (n = [], o = ce, (a = Ae()) !== s ? (44 === t.charCodeAt(ce) ? (i = ",", ce++) : (i = s, ge(m)), i !== s && (l = Ae()) !== s && (u = we()) !== s ? o = a = [a, i, l, u] : (ce = o, o = s)) : (ce = o, o = s); o !== s;)
                    n.push(o), o = ce, (a = Ae()) !== s ? (44 === t.charCodeAt(ce) ? (i = ",", ce++) : (i = s, ge(m)), i !== s && (l = Ae()) !== s && (u = we()) !== s ? o = a = [a, i, l, u] : (ce = o, o = s)) : (ce = o, o = s);
                n !== s ? e = r = [r].concat(n.map((function (e) { return e[3]; }))) : (ce = e, e = s);
            }
            else
                ce = e, e = s; return ye[c] = { nextPos: ce, result: e }, e; } function we() { var e, t, r, n, o, a, i, l = 30 * ce + 5, u = ye[l]; if (u)
                return ce = u.nextPos, u.result; if (e = ce, (t = Ce()) !== s) {
                for (r = [], n = ce, (o = Se()) !== s && (a = Ce()) !== s ? n = o = [o, a] : (ce = n, n = s); n !== s;)
                    r.push(n), n = ce, (o = Se()) !== s && (a = Ce()) !== s ? n = o = [o, a] : (ce = n, n = s);
                r !== s ? (i = t, e = t = r.reduce((function (e, t) { return { type: t[0], left: e, right: t[1] }; }), i)) : (ce = e, e = s);
            }
            else
                ce = e, e = s; return ye[l] = { nextPos: ce, result: e }, e; } function Ce() { var e, r, n, o, a, i, l, u = 30 * ce + 6, c = ye[u]; if (c)
                return ce = c.nextPos, c.result; if (e = ce, 33 === t.charCodeAt(ce) ? (r = "!", ce++) : (r = s, ge(x)), r === s && (r = null), r !== s) {
                if (n = [], (o = Pe()) !== s)
                    for (; o !== s;)
                        n.push(o), o = Pe();
                else
                    n = s;
                n !== s ? (a = r, l = 1 === (i = n).length ? i[0] : { type: "compound", selectors: i }, a && (l.subject = !0), e = r = l) : (ce = e, e = s);
            }
            else
                ce = e, e = s; return ye[u] = { nextPos: ce, result: e }, e; } function Pe() { var e, r = 30 * ce + 7, n = ye[r]; return n ? (ce = n.nextPos, n.result) : ((e = function () { var e, r, n = 30 * ce + 8, o = ye[n]; return o ? (ce = o.nextPos, o.result) : (42 === t.charCodeAt(ce) ? (r = "*", ce++) : (r = s, ge(v)), r !== s && (r = { type: "wildcard", value: r }), e = r, ye[n] = { nextPos: ce, result: e }, e); }()) === s && (e = function () { var e, r, n, o = 30 * ce + 9, a = ye[o]; return a ? (ce = a.nextPos, a.result) : (e = ce, 35 === t.charCodeAt(ce) ? (r = "#", ce++) : (r = s, ge(g)), r === s && (r = null), r !== s && (n = Ee()) !== s ? e = r = { type: "identifier", value: n } : (ce = e, e = s), ye[o] = { nextPos: ce, result: e }, e); }()) === s && (e = function () { var e, r, n, o, a = 30 * ce + 10, i = ye[a]; return i ? (ce = i.nextPos, i.result) : (e = ce, 91 === t.charCodeAt(ce) ? (r = "[", ce++) : (r = s, ge(b)), r !== s && Ae() !== s && (n = function () { var e, r, n, o, a = 30 * ce + 14, i = ye[a]; return i ? (ce = i.nextPos, i.result) : (e = ce, (r = ke()) !== s && Ae() !== s && (n = function () { var e, r, n, o = 30 * ce + 12, a = ye[o]; return a ? (ce = a.nextPos, a.result) : (e = ce, 33 === t.charCodeAt(ce) ? (r = "!", ce++) : (r = s, ge(x)), r === s && (r = null), r !== s ? (61 === t.charCodeAt(ce) ? (n = "=", ce++) : (n = s, ge(_)), n !== s ? (r = w(r), e = r) : (ce = e, e = s)) : (ce = e, e = s), ye[o] = { nextPos: ce, result: e }, e); }()) !== s && Ae() !== s ? ((o = function () { var e, r, n, o, a, i = 30 * ce + 18, l = ye[i]; if (l)
                return ce = l.nextPos, l.result; if (e = ce, "type(" === t.substr(ce, 5) ? (r = "type(", ce += 5) : (r = s, ge(N)), r !== s)
                if (Ae() !== s) {
                    if (n = [], W.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(K)), o !== s)
                        for (; o !== s;)
                            n.push(o), W.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(K));
                    else
                        n = s;
                    n !== s && (o = Ae()) !== s ? (41 === t.charCodeAt(ce) ? (a = ")", ce++) : (a = s, ge(G)), a !== s ? (r = { type: "type", value: n.join("") }, e = r) : (ce = e, e = s)) : (ce = e, e = s);
                }
                else
                    ce = e, e = s;
            else
                ce = e, e = s; return ye[i] = { nextPos: ce, result: e }, e; }()) === s && (o = function () { var e, r, n, o, a, i, l = 30 * ce + 20, u = ye[l]; if (u)
                return ce = u.nextPos, u.result; if (e = ce, 47 === t.charCodeAt(ce) ? (r = "/", ce++) : (r = s, ge(Y)), r !== s) {
                if (n = [], $.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(J)), o !== s)
                    for (; o !== s;)
                        n.push(o), $.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(J));
                else
                    n = s;
                n !== s ? (47 === t.charCodeAt(ce) ? (o = "/", ce++) : (o = s, ge(Y)), o !== s ? ((a = function () { var e, r, n = 30 * ce + 19, o = ye[n]; if (o)
                    return ce = o.nextPos, o.result; if (e = [], z.test(t.charAt(ce)) ? (r = t.charAt(ce), ce++) : (r = s, ge(H)), r !== s)
                    for (; r !== s;)
                        e.push(r), z.test(t.charAt(ce)) ? (r = t.charAt(ce), ce++) : (r = s, ge(H));
                else
                    e = s; return ye[n] = { nextPos: ce, result: e }, e; }()) === s && (a = null), a !== s ? (i = a, r = { type: "regexp", value: new RegExp(n.join(""), i ? i.join("") : "") }, e = r) : (ce = e, e = s)) : (ce = e, e = s)) : (ce = e, e = s);
            }
            else
                ce = e, e = s; return ye[l] = { nextPos: ce, result: e }, e; }()), o !== s ? (r = D(r, n, o), e = r) : (ce = e, e = s)) : (ce = e, e = s), e === s && (e = ce, (r = ke()) !== s && Ae() !== s && (n = function () { var e, r, n, o = 30 * ce + 11, a = ye[o]; return a ? (ce = a.nextPos, a.result) : (e = ce, E.test(t.charAt(ce)) ? (r = t.charAt(ce), ce++) : (r = s, ge(S)), r === s && (r = null), r !== s ? (61 === t.charCodeAt(ce) ? (n = "=", ce++) : (n = s, ge(_)), n !== s ? (r = w(r), e = r) : (ce = e, e = s)) : (ce = e, e = s), e === s && (C.test(t.charAt(ce)) ? (e = t.charAt(ce), ce++) : (e = s, ge(P))), ye[o] = { nextPos: ce, result: e }, e); }()) !== s && Ae() !== s ? ((o = function () { var e, r, n, o, a, i, l = 30 * ce + 15, u = ye[l]; if (u)
                return ce = u.nextPos, u.result; if (e = ce, 34 === t.charCodeAt(ce) ? (r = '"', ce++) : (r = s, ge(j)), r !== s) {
                for (n = [], I.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(T)), o === s && (o = ce, 92 === t.charCodeAt(ce) ? (a = "\\", ce++) : (a = s, ge(F)), a !== s ? (t.length > ce ? (i = t.charAt(ce), ce++) : (i = s, ge(L)), i !== s ? (a = O(a, i), o = a) : (ce = o, o = s)) : (ce = o, o = s)); o !== s;)
                    n.push(o), I.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(T)), o === s && (o = ce, 92 === t.charCodeAt(ce) ? (a = "\\", ce++) : (a = s, ge(F)), a !== s ? (t.length > ce ? (i = t.charAt(ce), ce++) : (i = s, ge(L)), i !== s ? (a = O(a, i), o = a) : (ce = o, o = s)) : (ce = o, o = s));
                n !== s ? (34 === t.charCodeAt(ce) ? (o = '"', ce++) : (o = s, ge(j)), o !== s ? (r = R(n), e = r) : (ce = e, e = s)) : (ce = e, e = s);
            }
            else
                ce = e, e = s; if (e === s)
                if (e = ce, 39 === t.charCodeAt(ce) ? (r = "'", ce++) : (r = s, ge(B)), r !== s) {
                    for (n = [], M.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(U)), o === s && (o = ce, 92 === t.charCodeAt(ce) ? (a = "\\", ce++) : (a = s, ge(F)), a !== s ? (t.length > ce ? (i = t.charAt(ce), ce++) : (i = s, ge(L)), i !== s ? (a = O(a, i), o = a) : (ce = o, o = s)) : (ce = o, o = s)); o !== s;)
                        n.push(o), M.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(U)), o === s && (o = ce, 92 === t.charCodeAt(ce) ? (a = "\\", ce++) : (a = s, ge(F)), a !== s ? (t.length > ce ? (i = t.charAt(ce), ce++) : (i = s, ge(L)), i !== s ? (a = O(a, i), o = a) : (ce = o, o = s)) : (ce = o, o = s));
                    n !== s ? (39 === t.charCodeAt(ce) ? (o = "'", ce++) : (o = s, ge(B)), o !== s ? (r = R(n), e = r) : (ce = e, e = s)) : (ce = e, e = s);
                }
                else
                    ce = e, e = s; return ye[l] = { nextPos: ce, result: e }, e; }()) === s && (o = function () { var e, r, n, o, a, i, l, u = 30 * ce + 16, c = ye[u]; if (c)
                return ce = c.nextPos, c.result; for (e = ce, r = ce, n = [], q.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(V)); o !== s;)
                n.push(o), q.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(V)); if (n !== s ? (46 === t.charCodeAt(ce) ? (o = ".", ce++) : (o = s, ge(k)), o !== s ? r = n = [n, o] : (ce = r, r = s)) : (ce = r, r = s), r === s && (r = null), r !== s) {
                if (n = [], q.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(V)), o !== s)
                    for (; o !== s;)
                        n.push(o), q.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(V));
                else
                    n = s;
                n !== s ? (i = n, l = (a = r) ? [].concat.apply([], a).join("") : "", r = { type: "literal", value: parseFloat(l + i.join("")) }, e = r) : (ce = e, e = s);
            }
            else
                ce = e, e = s; return ye[u] = { nextPos: ce, result: e }, e; }()) === s && (o = function () { var e, t, r = 30 * ce + 17, n = ye[r]; return n ? (ce = n.nextPos, n.result) : ((t = Ee()) !== s && (t = { type: "literal", value: t }), e = t, ye[r] = { nextPos: ce, result: e }, e); }()), o !== s ? (r = D(r, n, o), e = r) : (ce = e, e = s)) : (ce = e, e = s), e === s && (e = ce, (r = ke()) !== s && (r = { type: "attribute", name: r }), e = r)), ye[a] = { nextPos: ce, result: e }, e); }()) !== s && Ae() !== s ? (93 === t.charCodeAt(ce) ? (o = "]", ce++) : (o = s, ge(A)), o !== s ? e = r = n : (ce = e, e = s)) : (ce = e, e = s), ye[a] = { nextPos: ce, result: e }, e); }()) === s && (e = function () { var e, r, n, o, a, i, l, u, c = 30 * ce + 21, f = ye[c]; if (f)
                return ce = f.nextPos, f.result; if (e = ce, 46 === t.charCodeAt(ce) ? (r = ".", ce++) : (r = s, ge(k)), r !== s)
                if ((n = Ee()) !== s) {
                    for (o = [], a = ce, 46 === t.charCodeAt(ce) ? (i = ".", ce++) : (i = s, ge(k)), i !== s && (l = Ee()) !== s ? a = i = [i, l] : (ce = a, a = s); a !== s;)
                        o.push(a), a = ce, 46 === t.charCodeAt(ce) ? (i = ".", ce++) : (i = s, ge(k)), i !== s && (l = Ee()) !== s ? a = i = [i, l] : (ce = a, a = s);
                    o !== s ? (u = n, r = { type: "field", name: o.reduce((function (e, t) { return e + t[0] + t[1]; }), u) }, e = r) : (ce = e, e = s);
                }
                else
                    ce = e, e = s;
            else
                ce = e, e = s; return ye[c] = { nextPos: ce, result: e }, e; }()) === s && (e = function () { var e, r, n, o, a = 30 * ce + 22, i = ye[a]; return i ? (ce = i.nextPos, i.result) : (e = ce, ":not(" === t.substr(ce, 5) ? (r = ":not(", ce += 5) : (r = s, ge(Q)), r !== s && Ae() !== s && (n = _e()) !== s && Ae() !== s ? (41 === t.charCodeAt(ce) ? (o = ")", ce++) : (o = s, ge(G)), o !== s ? e = r = { type: "not", selectors: n } : (ce = e, e = s)) : (ce = e, e = s), ye[a] = { nextPos: ce, result: e }, e); }()) === s && (e = function () { var e, r, n, o, a = 30 * ce + 23, i = ye[a]; return i ? (ce = i.nextPos, i.result) : (e = ce, ":matches(" === t.substr(ce, 9) ? (r = ":matches(", ce += 9) : (r = s, ge(X)), r !== s && Ae() !== s && (n = _e()) !== s && Ae() !== s ? (41 === t.charCodeAt(ce) ? (o = ")", ce++) : (o = s, ge(G)), o !== s ? e = r = { type: "matches", selectors: n } : (ce = e, e = s)) : (ce = e, e = s), ye[a] = { nextPos: ce, result: e }, e); }()) === s && (e = function () { var e, r, n, o, a = 30 * ce + 24, i = ye[a]; return i ? (ce = i.nextPos, i.result) : (e = ce, ":has(" === t.substr(ce, 5) ? (r = ":has(", ce += 5) : (r = s, ge(Z)), r !== s && Ae() !== s && (n = _e()) !== s && Ae() !== s ? (41 === t.charCodeAt(ce) ? (o = ")", ce++) : (o = s, ge(G)), o !== s ? e = r = { type: "has", selectors: n } : (ce = e, e = s)) : (ce = e, e = s), ye[a] = { nextPos: ce, result: e }, e); }()) === s && (e = function () { var e, r, n = 30 * ce + 25, o = ye[n]; return o ? (ce = o.nextPos, o.result) : (":first-child" === t.substr(ce, 12) ? (r = ":first-child", ce += 12) : (r = s, ge(ee)), r !== s && (r = De(1)), e = r, ye[n] = { nextPos: ce, result: e }, e); }()) === s && (e = function () { var e, r, n = 30 * ce + 26, o = ye[n]; return o ? (ce = o.nextPos, o.result) : (":last-child" === t.substr(ce, 11) ? (r = ":last-child", ce += 11) : (r = s, ge(te)), r !== s && (r = je(1)), e = r, ye[n] = { nextPos: ce, result: e }, e); }()) === s && (e = function () { var e, r, n, o, a, i = 30 * ce + 27, l = ye[i]; if (l)
                return ce = l.nextPos, l.result; if (e = ce, ":nth-child(" === t.substr(ce, 11) ? (r = ":nth-child(", ce += 11) : (r = s, ge(re)), r !== s)
                if (Ae() !== s) {
                    if (n = [], q.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(V)), o !== s)
                        for (; o !== s;)
                            n.push(o), q.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(V));
                    else
                        n = s;
                    n !== s && (o = Ae()) !== s ? (41 === t.charCodeAt(ce) ? (a = ")", ce++) : (a = s, ge(G)), a !== s ? (r = De(parseInt(n.join(""), 10)), e = r) : (ce = e, e = s)) : (ce = e, e = s);
                }
                else
                    ce = e, e = s;
            else
                ce = e, e = s; return ye[i] = { nextPos: ce, result: e }, e; }()) === s && (e = function () { var e, r, n, o, a, i = 30 * ce + 28, l = ye[i]; if (l)
                return ce = l.nextPos, l.result; if (e = ce, ":nth-last-child(" === t.substr(ce, 16) ? (r = ":nth-last-child(", ce += 16) : (r = s, ge(ne)), r !== s)
                if (Ae() !== s) {
                    if (n = [], q.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(V)), o !== s)
                        for (; o !== s;)
                            n.push(o), q.test(t.charAt(ce)) ? (o = t.charAt(ce), ce++) : (o = s, ge(V));
                    else
                        n = s;
                    n !== s && (o = Ae()) !== s ? (41 === t.charCodeAt(ce) ? (a = ")", ce++) : (a = s, ge(G)), a !== s ? (r = je(parseInt(n.join(""), 10)), e = r) : (ce = e, e = s)) : (ce = e, e = s);
                }
                else
                    ce = e, e = s;
            else
                ce = e, e = s; return ye[i] = { nextPos: ce, result: e }, e; }()) === s && (e = function () { var e, r, n, o = 30 * ce + 29, a = ye[o]; return a ? (ce = a.nextPos, a.result) : (e = ce, 58 === t.charCodeAt(ce) ? (r = ":", ce++) : (r = s, ge(oe)), r !== s ? ("statement" === t.substr(ce, 9).toLowerCase() ? (n = t.substr(ce, 9), ce += 9) : (n = s, ge(ae)), n === s && ("expression" === t.substr(ce, 10).toLowerCase() ? (n = t.substr(ce, 10), ce += 10) : (n = s, ge(ie)), n === s && ("declaration" === t.substr(ce, 11).toLowerCase() ? (n = t.substr(ce, 11), ce += 11) : (n = s, ge(se)), n === s && ("function" === t.substr(ce, 8).toLowerCase() ? (n = t.substr(ce, 8), ce += 8) : (n = s, ge(le)), n === s && ("pattern" === t.substr(ce, 7).toLowerCase() ? (n = t.substr(ce, 7), ce += 7) : (n = s, ge(ue)))))), n !== s ? e = r = { type: "class", name: n } : (ce = e, e = s)) : (ce = e, e = s), ye[o] = { nextPos: ce, result: e }, e); }()), ye[r] = { nextPos: ce, result: e }, e); } function ke() { var e, r, n, o, a, i, l, u, c = 30 * ce + 13, f = ye[c]; if (f)
                return ce = f.nextPos, f.result; if (e = ce, (r = Ee()) !== s) {
                for (n = [], o = ce, 46 === t.charCodeAt(ce) ? (a = ".", ce++) : (a = s, ge(k)), a !== s && (i = Ee()) !== s ? o = a = [a, i] : (ce = o, o = s); o !== s;)
                    n.push(o), o = ce, 46 === t.charCodeAt(ce) ? (a = ".", ce++) : (a = s, ge(k)), a !== s && (i = Ee()) !== s ? o = a = [a, i] : (ce = o, o = s);
                n !== s ? (l = r, u = n, e = r = [].concat.apply([l], u).join("")) : (ce = e, e = s);
            }
            else
                ce = e, e = s; return ye[c] = { nextPos: ce, result: e }, e; } function De(e) { return { type: "nth-child", index: { type: "literal", value: e } }; } function je(e) { return { type: "nth-last-child", index: { type: "literal", value: e } }; } if ((n = u()) !== s && ce === t.length)
                return n; throw n !== s && ce < t.length && ge({ type: "end" }), o = he, a = pe < t.length ? t.charAt(pe) : null, i = pe < t.length ? ve(pe, pe + 1) : ve(pe, pe), new e(e.buildMessage(o, a), o, a, i); } }; }()); }));
        function u(t, r, n, o) { if (!r)
            return !0; if (!t)
            return !1; switch (n || (n = []), r.type) {
            case "wildcard": return !0;
            case "identifier": return r.value.toLowerCase() === t.type.toLowerCase();
            case "field":
                var i = r.name.split("."), l = n[i.length - 1];
                return function e(t, r, n) { if (0 === n.length)
                    return t === r; if (null == r)
                    return !1; var o = r[n[0]], i = n.slice(1); if (Array.isArray(o)) {
                    var s, l = a(o);
                    try {
                        for (l.s(); !(s = l.n()).done;) {
                            if (e(t, s.value, i))
                                return !0;
                        }
                    }
                    catch (e) {
                        l.e(e);
                    }
                    finally {
                        l.f();
                    }
                    return !1;
                } return e(t, o, i); }(t, l, i);
            case "matches":
                var c, f = a(r.selectors);
                try {
                    for (f.s(); !(c = f.n()).done;) {
                        var d = c.value;
                        if (u(t, d, n, o))
                            return !0;
                    }
                }
                catch (e) {
                    f.e(e);
                }
                finally {
                    f.f();
                }
                return !1;
            case "compound":
                var m, x = a(r.selectors);
                try {
                    for (x.s(); !(m = x.n()).done;) {
                        var v = m.value;
                        if (!u(t, v, n, o))
                            return !1;
                    }
                }
                catch (e) {
                    x.e(e);
                }
                finally {
                    x.f();
                }
                return !0;
            case "not":
                var g, b = a(r.selectors);
                try {
                    for (b.s(); !(g = b.n()).done;) {
                        var A = g.value;
                        if (u(t, A, n, o))
                            return !1;
                    }
                }
                catch (e) {
                    b.e(e);
                }
                finally {
                    b.f();
                }
                return !0;
            case "has":
                var E = function () { var e, n = [], i = a(r.selectors); try {
                    var l = function () { var r = e.value, a = []; s.traverse(t, { enter: function (e, t) { null != t && a.unshift(t), u(e, r, a, o) && n.push(e); }, leave: function () { a.shift(); }, keys: o && o.visitorKeys, fallback: o && o.fallback || "iteration" }); };
                    for (i.s(); !(e = i.n()).done;)
                        l();
                }
                catch (e) {
                    i.e(e);
                }
                finally {
                    i.f();
                } return { v: 0 !== n.length }; }();
                if ("object" === e(E))
                    return E.v;
            case "child": return !!u(t, r.right, n, o) && u(n[0], r.left, n.slice(1), o);
            case "descendant":
                if (u(t, r.right, n, o))
                    for (var S = 0, _ = n.length; S < _; ++S)
                        if (u(n[S], r.left, n.slice(S + 1), o))
                            return !0;
                return !1;
            case "attribute":
                var w = function (e, t) { var r, n = a(t.split(".")); try {
                    for (n.s(); !(r = n.n()).done;) {
                        var o = r.value;
                        if (null == e)
                            return e;
                        e = e[o];
                    }
                }
                catch (e) {
                    n.e(e);
                }
                finally {
                    n.f();
                } return e; }(t, r.name);
                switch (r.operator) {
                    case void 0: return null != w;
                    case "=":
                        switch (r.value.type) {
                            case "regexp": return "string" == typeof w && r.value.value.test(w);
                            case "literal": return "".concat(r.value.value) === "".concat(w);
                            case "type": return r.value.value === e(w);
                        }
                        throw new Error("Unknown selector value type: ".concat(r.value.type));
                    case "!=":
                        switch (r.value.type) {
                            case "regexp": return !r.value.value.test(w);
                            case "literal": return "".concat(r.value.value) !== "".concat(w);
                            case "type": return r.value.value !== e(w);
                        }
                        throw new Error("Unknown selector value type: ".concat(r.value.type));
                    case "<=": return w <= r.value.value;
                    case "<": return w < r.value.value;
                    case ">": return w > r.value.value;
                    case ">=": return w >= r.value.value;
                }
                throw new Error("Unknown operator: ".concat(r.operator));
            case "sibling": return u(t, r.right, n, o) && p(t, r.left, n, "LEFT_SIDE", o) || r.left.subject && u(t, r.left, n, o) && p(t, r.right, n, "RIGHT_SIDE", o);
            case "adjacent": return u(t, r.right, n, o) && h(t, r.left, n, "LEFT_SIDE", o) || r.right.subject && u(t, r.left, n, o) && h(t, r.right, n, "RIGHT_SIDE", o);
            case "nth-child": return u(t, r.right, n, o) && y(t, n, (function () { return r.index.value - 1; }), o);
            case "nth-last-child": return u(t, r.right, n, o) && y(t, n, (function (e) { return e - r.index.value; }), o);
            case "class":
                switch (r.name.toLowerCase()) {
                    case "statement": if ("Statement" === t.type.slice(-9))
                        return !0;
                    case "declaration": return "Declaration" === t.type.slice(-11);
                    case "pattern": if ("Pattern" === t.type.slice(-7))
                        return !0;
                    case "expression": return "Expression" === t.type.slice(-10) || "Literal" === t.type.slice(-7) || "Identifier" === t.type && (0 === n.length || "MetaProperty" !== n[0].type) || "MetaProperty" === t.type;
                    case "function": return "FunctionDeclaration" === t.type || "FunctionExpression" === t.type || "ArrowFunctionExpression" === t.type;
                }
                throw new Error("Unknown class name: ".concat(r.name));
        } throw new Error("Unknown selector type: ".concat(r.type)); }
        function c(e, t) { var r = e.type; return t && t.visitorKeys && t.visitorKeys[r] ? t.visitorKeys[r] : s.VisitorKeys[r] ? s.VisitorKeys[r] : t && "function" == typeof t.fallback ? t.fallback(e) : Object.keys(e).filter((function (e) { return "type" !== e; })); }
        function f(t) { return null !== t && "object" === e(t) && "string" == typeof t.type; }
        function p(e, r, n, o, i) { var s = t(n, 1)[0]; if (!s)
            return !1; var l, p = a(c(s, i)); try {
            for (p.s(); !(l = p.n()).done;) {
                var h = s[l.value];
                if (Array.isArray(h)) {
                    var y = h.indexOf(e);
                    if (y < 0)
                        continue;
                    var d = void 0, m = void 0;
                    "LEFT_SIDE" === o ? (d = 0, m = y) : (d = y + 1, m = h.length);
                    for (var x = d; x < m; ++x)
                        if (f(h[x]) && u(h[x], r, n, i))
                            return !0;
                }
            }
        }
        catch (e) {
            p.e(e);
        }
        finally {
            p.f();
        } return !1; }
        function h(e, r, n, o, i) { var s = t(n, 1)[0]; if (!s)
            return !1; var l, p = a(c(s, i)); try {
            for (p.s(); !(l = p.n()).done;) {
                var h = s[l.value];
                if (Array.isArray(h)) {
                    var y = h.indexOf(e);
                    if (y < 0)
                        continue;
                    if ("LEFT_SIDE" === o && y > 0 && f(h[y - 1]) && u(h[y - 1], r, n, i))
                        return !0;
                    if ("RIGHT_SIDE" === o && y < h.length - 1 && f(h[y + 1]) && u(h[y + 1], r, n, i))
                        return !0;
                }
            }
        }
        catch (e) {
            p.e(e);
        }
        finally {
            p.f();
        } return !1; }
        function y(e, r, n, o) { var i = t(r, 1)[0]; if (!i)
            return !1; var s, l = a(c(i, o)); try {
            for (l.s(); !(s = l.n()).done;) {
                var u = i[s.value];
                if (Array.isArray(u)) {
                    var f = u.indexOf(e);
                    if (f >= 0 && f === n(u.length))
                        return !0;
                }
            }
        }
        catch (e) {
            l.e(e);
        }
        finally {
            l.f();
        } return !1; }
        function d(n, o) { if (null == n || "object" != e(n))
            return []; null == o && (o = n); for (var a = n.subject ? [o] : [], i = 0, s = function (e) { for (var t = [], r = Object.keys(e), n = 0; n < r.length; n++)
            t.push([r[n], e[r[n]]]); return t; }(n); i < s.length; i++) {
            var l = t(s[i], 2), u = l[0], c = l[1];
            a.push.apply(a, r(d(c, "left" === u ? c : o)));
        } return a; }
        function m(e, t, r, n) { if (t) {
            var o = [], a = d(t);
            s.traverse(e, { enter: function (e, i) { if (null != i && o.unshift(i), u(e, t, o, n))
                    if (a.length)
                        for (var s = 0, l = a.length; s < l; ++s) {
                            u(e, a[s], o, n) && r(e, i, o);
                            for (var c = 0, f = o.length; c < f; ++c) {
                                var p = o.slice(c + 1);
                                u(o[c], a[s], p, n) && r(o[c], i, p);
                            }
                        }
                    else
                        r(e, i, o); }, leave: function () { o.shift(); }, keys: n && n.visitorKeys, fallback: n && n.fallback || "iteration" });
        } }
        function x(e, t, r) { var n = []; return m(e, t, (function (e) { n.push(e); }), r), n; }
        function v(e) { return l.parse(e); }
        function g(e, t, r) { return x(e, v(t), r); }
        return g.parse = v, g.match = x, g.traverse = m, g.matches = u, g.query = g, g;
    }));
},
/* eslint/lib/linter/report-translator.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const assert = require("assert");
    const ruleFixer = require(388) /* ./rule-fixer */;
    const interpolate = require(389) /* ./interpolate */;
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    /**
     * An error message description
     * @typedef {Object} MessageDescriptor
     * @property {ASTNode} [node] The reported node
     * @property {Location} loc The location of the problem.
     * @property {string} message The problem message.
     * @property {Object} [data] Optional data to use to fill in placeholders in the
     *      message.
     * @property {Function} [fix] The function to call that creates a fix command.
     * @property {Array<{desc?: string, messageId?: string, fix: Function}>} suggest Suggestion descriptions and functions to create a the associated fixes.
     */
    /**
     * Information about the report
     * @typedef {Object} ReportInfo
     * @property {string} ruleId
     * @property {(0|1|2)} severity
     * @property {(string|undefined)} message
     * @property {(string|undefined)} [messageId]
     * @property {number} line
     * @property {number} column
     * @property {(number|undefined)} [endLine]
     * @property {(number|undefined)} [endColumn]
     * @property {(string|null)} nodeType
     * @property {string} source
     * @property {({text: string, range: (number[]|null)}|null)} [fix]
     * @property {Array<{text: string, range: (number[]|null)}|null>} [suggestions]
     */
    //------------------------------------------------------------------------------
    // Module Definition
    //------------------------------------------------------------------------------
    /**
     * Translates a multi-argument context.report() call into a single object argument call
     * @param {...*} args A list of arguments passed to `context.report`
     * @returns {MessageDescriptor} A normalized object containing report information
     */
    function normalizeMultiArgReportCall(...args) {
        // If there is one argument, it is considered to be a new-style call already.
        if (args.length === 1) {
            // Shallow clone the object to avoid surprises if reusing the descriptor
            return Object.assign({}, args[0]);
        }
        // If the second argument is a string, the arguments are interpreted as [node, message, data, fix].
        if (typeof args[1] === "string") {
            return {
                node: args[0],
                message: args[1],
                data: args[2],
                fix: args[3]
            };
        }
        // Otherwise, the arguments are interpreted as [node, loc, message, data, fix].
        return {
            node: args[0],
            loc: args[1],
            message: args[2],
            data: args[3],
            fix: args[4]
        };
    }
    /**
     * Asserts that either a loc or a node was provided, and the node is valid if it was provided.
     * @param {MessageDescriptor} descriptor A descriptor to validate
     * @returns {void}
     * @throws AssertionError if neither a node nor a loc was provided, or if the node is not an object
     */
    function assertValidNodeInfo(descriptor) {
        if (descriptor.node) {
            assert(typeof descriptor.node === "object", "Node must be an object");
        }
        else {
            assert(descriptor.loc, "Node must be provided when reporting error if location is not provided");
        }
    }
    /**
     * Normalizes a MessageDescriptor to always have a `loc` with `start` and `end` properties
     * @param {MessageDescriptor} descriptor A descriptor for the report from a rule.
     * @returns {{start: Location, end: (Location|null)}} An updated location that infers the `start` and `end` properties
     * from the `node` of the original descriptor, or infers the `start` from the `loc` of the original descriptor.
     */
    function normalizeReportLoc(descriptor) {
        if (descriptor.loc) {
            if (descriptor.loc.start) {
                return descriptor.loc;
            }
            return { start: descriptor.loc, end: null };
        }
        return descriptor.node.loc;
    }
    /**
     * Check that a fix has a valid range.
     * @param {Fix|null} fix The fix to validate.
     * @returns {void}
     */
    function assertValidFix(fix) {
        if (fix) {
            assert(fix.range && typeof fix.range[0] === "number" && typeof fix.range[1] === "number", `Fix has invalid range: ${JSON.stringify(fix, null, 2)}`);
        }
    }
    /**
     * Compares items in a fixes array by range.
     * @param {Fix} a The first message.
     * @param {Fix} b The second message.
     * @returns {int} -1 if a comes before b, 1 if a comes after b, 0 if equal.
     * @private
     */
    function compareFixesByRange(a, b) {
        return a.range[0] - b.range[0] || a.range[1] - b.range[1];
    }
    /**
     * Merges the given fixes array into one.
     * @param {Fix[]} fixes The fixes to merge.
     * @param {SourceCode} sourceCode The source code object to get the text between fixes.
     * @returns {{text: string, range: number[]}} The merged fixes
     */
    function mergeFixes(fixes, sourceCode) {
        for (const fix of fixes) {
            assertValidFix(fix);
        }
        if (fixes.length === 0) {
            return null;
        }
        if (fixes.length === 1) {
            return fixes[0];
        }
        fixes.sort(compareFixesByRange);
        const originalText = sourceCode.text;
        const start = fixes[0].range[0];
        const end = fixes[fixes.length - 1].range[1];
        let text = "";
        let lastPos = Number.MIN_SAFE_INTEGER;
        for (const fix of fixes) {
            assert(fix.range[0] >= lastPos, "Fix objects must not be overlapped in a report.");
            if (fix.range[0] >= 0) {
                text += originalText.slice(Math.max(0, start, lastPos), fix.range[0]);
            }
            text += fix.text;
            lastPos = fix.range[1];
        }
        text += originalText.slice(Math.max(0, start, lastPos), end);
        return { range: [start, end], text };
    }
    /**
     * Gets one fix object from the given descriptor.
     * If the descriptor retrieves multiple fixes, this merges those to one.
     * @param {MessageDescriptor} descriptor The report descriptor.
     * @param {SourceCode} sourceCode The source code object to get text between fixes.
     * @returns {({text: string, range: number[]}|null)} The fix for the descriptor
     */
    function normalizeFixes(descriptor, sourceCode) {
        if (typeof descriptor.fix !== "function") {
            return null;
        }
        // @type {null | Fix | Fix[] | IterableIterator<Fix>}
        const fix = descriptor.fix(ruleFixer);
        // Merge to one.
        if (fix && Symbol.iterator in fix) {
            return mergeFixes(Array.from(fix), sourceCode);
        }
        assertValidFix(fix);
        return fix;
    }
    /**
     * Gets an array of suggestion objects from the given descriptor.
     * @param {MessageDescriptor} descriptor The report descriptor.
     * @param {SourceCode} sourceCode The source code object to get text between fixes.
     * @param {Object} messages Object of meta messages for the rule.
     * @returns {Array<SuggestionResult>} The suggestions for the descriptor
     */
    function mapSuggestions(descriptor, sourceCode, messages) {
        if (!descriptor.suggest || !Array.isArray(descriptor.suggest)) {
            return [];
        }
        return descriptor.suggest
            .map(suggestInfo => {
            const computedDesc = suggestInfo.desc || messages[suggestInfo.messageId];
            return {
                ...suggestInfo,
                desc: interpolate(computedDesc, suggestInfo.data),
                fix: normalizeFixes(suggestInfo, sourceCode)
            };
        })
            // Remove suggestions that didn't provide a fix
            .filter(({ fix }) => fix);
    }
    /**
     * Creates information about the report from a descriptor
     * @param {Object} options Information about the problem
     * @param {string} options.ruleId Rule ID
     * @param {(0|1|2)} options.severity Rule severity
     * @param {(ASTNode|null)} options.node Node
     * @param {string} options.message Error message
     * @param {string} [options.messageId] The error message ID.
     * @param {{start: SourceLocation, end: (SourceLocation|null)}} options.loc Start and end location
     * @param {{text: string, range: (number[]|null)}} options.fix The fix object
     * @param {Array<{text: string, range: (number[]|null)}>} options.suggestions The array of suggestions objects
     * @returns {function(...args): ReportInfo} Function that returns information about the report
     */
    function createProblem(options) {
        const problem = {
            ruleId: options.ruleId,
            severity: options.severity,
            message: options.message,
            line: options.loc.start.line,
            column: options.loc.start.column + 1,
            nodeType: options.node && options.node.type || null
        };
        /*
         * If this isn’t in the conditional, some of the tests fail
         * because `messageId` is present in the problem object
         */
        if (options.messageId) {
            problem.messageId = options.messageId;
        }
        if (options.loc.end) {
            problem.endLine = options.loc.end.line;
            problem.endColumn = options.loc.end.column + 1;
        }
        if (options.fix) {
            problem.fix = options.fix;
        }
        if (options.suggestions && options.suggestions.length > 0) {
            problem.suggestions = options.suggestions;
        }
        return problem;
    }
    /**
     * Validates that suggestions are properly defined. Throws if an error is detected.
     * @param {Array<{ desc?: string, messageId?: string }>} suggest The incoming suggest data.
     * @param {Object} messages Object of meta messages for the rule.
     * @returns {void}
     */
    function validateSuggestions(suggest, messages) {
        if (suggest && Array.isArray(suggest)) {
            suggest.forEach(suggestion => {
                if (suggestion.messageId) {
                    const { messageId } = suggestion;
                    if (!messages) {
                        throw new TypeError(`context.report() called with a suggest option with a messageId '${messageId}', but no messages were present in the rule metadata.`);
                    }
                    if (!messages[messageId]) {
                        throw new TypeError(`context.report() called with a suggest option with a messageId '${messageId}' which is not present in the 'messages' config: ${JSON.stringify(messages, null, 2)}`);
                    }
                    if (suggestion.desc) {
                        throw new TypeError("context.report() called with a suggest option that defines both a 'messageId' and an 'desc'. Please only pass one.");
                    }
                }
                else if (!suggestion.desc) {
                    throw new TypeError("context.report() called with a suggest option that doesn't have either a `desc` or `messageId`");
                }
                if (typeof suggestion.fix !== "function") {
                    throw new TypeError(`context.report() called with a suggest option without a fix function. See: ${suggestion}`);
                }
            });
        }
    }
    /**
     * Returns a function that converts the arguments of a `context.report` call from a rule into a reported
     * problem for the Node.js API.
     * @param {{ruleId: string, severity: number, sourceCode: SourceCode, messageIds: Object, disableFixes: boolean}} metadata Metadata for the reported problem
     * @param {SourceCode} sourceCode The `SourceCode` instance for the text being linted
     * @returns {function(...args): ReportInfo} Function that returns information about the report
     */
    module.exports = function createReportTranslator(metadata) {
        /*
         * `createReportTranslator` gets called once per enabled rule per file. It needs to be very performant.
         * The report translator itself (i.e. the function that `createReportTranslator` returns) gets
         * called every time a rule reports a problem, which happens much less frequently (usually, the vast
         * majority of rules don't report any problems for a given file).
         */
        return (...args) => {
            const descriptor = normalizeMultiArgReportCall(...args);
            const messages = metadata.messageIds;
            assertValidNodeInfo(descriptor);
            let computedMessage;
            if (descriptor.messageId) {
                if (!messages) {
                    throw new TypeError("context.report() called with a messageId, but no messages were present in the rule metadata.");
                }
                const id = descriptor.messageId;
                if (descriptor.message) {
                    throw new TypeError("context.report() called with a message and a messageId. Please only pass one.");
                }
                if (!messages || !Object.prototype.hasOwnProperty.call(messages, id)) {
                    throw new TypeError(`context.report() called with a messageId of '${id}' which is not present in the 'messages' config: ${JSON.stringify(messages, null, 2)}`);
                }
                computedMessage = messages[id];
            }
            else if (descriptor.message) {
                computedMessage = descriptor.message;
            }
            else {
                throw new TypeError("Missing `message` property in report() call; add a message that describes the linting problem.");
            }
            validateSuggestions(descriptor.suggest, messages);
            return createProblem({
                ruleId: metadata.ruleId,
                severity: metadata.severity,
                node: descriptor.node,
                message: interpolate(computedMessage, descriptor.data),
                messageId: descriptor.messageId,
                loc: normalizeReportLoc(descriptor),
                fix: metadata.disableFixes ? null : normalizeFixes(descriptor, metadata.sourceCode),
                suggestions: metadata.disableFixes ? [] : mapSuggestions(descriptor, metadata.sourceCode, messages)
            });
        };
    };
},
/* eslint/lib/linter/rule-fixer.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    // none!
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Creates a fix command that inserts text at the specified index in the source text.
     * @param {int} index The 0-based index at which to insert the new text.
     * @param {string} text The text to insert.
     * @returns {Object} The fix command.
     * @private
     */
    function insertTextAt(index, text) {
        return {
            range: [index, index],
            text
        };
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * Creates code fixing commands for rules.
     */
    const ruleFixer = Object.freeze({
        /**
         * Creates a fix command that inserts text after the given node or token.
         * The fix is not applied until applyFixes() is called.
         * @param {ASTNode|Token} nodeOrToken The node or token to insert after.
         * @param {string} text The text to insert.
         * @returns {Object} The fix command.
         */
        insertTextAfter(nodeOrToken, text) {
            return this.insertTextAfterRange(nodeOrToken.range, text);
        },
        /**
         * Creates a fix command that inserts text after the specified range in the source text.
         * The fix is not applied until applyFixes() is called.
         * @param {int[]} range The range to replace, first item is start of range, second
         *      is end of range.
         * @param {string} text The text to insert.
         * @returns {Object} The fix command.
         */
        insertTextAfterRange(range, text) {
            return insertTextAt(range[1], text);
        },
        /**
         * Creates a fix command that inserts text before the given node or token.
         * The fix is not applied until applyFixes() is called.
         * @param {ASTNode|Token} nodeOrToken The node or token to insert before.
         * @param {string} text The text to insert.
         * @returns {Object} The fix command.
         */
        insertTextBefore(nodeOrToken, text) {
            return this.insertTextBeforeRange(nodeOrToken.range, text);
        },
        /**
         * Creates a fix command that inserts text before the specified range in the source text.
         * The fix is not applied until applyFixes() is called.
         * @param {int[]} range The range to replace, first item is start of range, second
         *      is end of range.
         * @param {string} text The text to insert.
         * @returns {Object} The fix command.
         */
        insertTextBeforeRange(range, text) {
            return insertTextAt(range[0], text);
        },
        /**
         * Creates a fix command that replaces text at the node or token.
         * The fix is not applied until applyFixes() is called.
         * @param {ASTNode|Token} nodeOrToken The node or token to remove.
         * @param {string} text The text to insert.
         * @returns {Object} The fix command.
         */
        replaceText(nodeOrToken, text) {
            return this.replaceTextRange(nodeOrToken.range, text);
        },
        /**
         * Creates a fix command that replaces text at the specified range in the source text.
         * The fix is not applied until applyFixes() is called.
         * @param {int[]} range The range to replace, first item is start of range, second
         *      is end of range.
         * @param {string} text The text to insert.
         * @returns {Object} The fix command.
         */
        replaceTextRange(range, text) {
            return {
                range,
                text
            };
        },
        /**
         * Creates a fix command that removes the node or token from the source.
         * The fix is not applied until applyFixes() is called.
         * @param {ASTNode|Token} nodeOrToken The node or token to remove.
         * @returns {Object} The fix command.
         */
        remove(nodeOrToken) {
            return this.removeRange(nodeOrToken.range);
        },
        /**
         * Creates a fix command that removes the specified range of text from the source.
         * The fix is not applied until applyFixes() is called.
         * @param {int[]} range The range to remove, first item is start of range, second
         *      is end of range.
         * @returns {Object} The fix command.
         */
        removeRange(range) {
            return {
                range,
                text: ""
            };
        }
    });
    module.exports = ruleFixer;
},
/* eslint/lib/linter/interpolate.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = (text, data) => {
        if (!data) {
            return text;
        }
        // Substitution content for any {{ }} markers.
        return text.replace(/\{\{([^{}]+?)\}\}/gu, (fullMatch, termWithWhitespace) => {
            const term = termWithWhitespace.trim();
            if (term in data) {
                return data[term];
            }
            // Preserve old behavior: If parameter name not provided, don't replace it.
            return fullMatch;
        });
    };
},
/* eslint/lib/linter/rules.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const builtInRules = require(391) /* ../rules */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Normalizes a rule module to the new-style API
     * @param {(Function|{create: Function})} rule A rule object, which can either be a function
     * ("old-style") or an object with a `create` method ("new-style")
     * @returns {{create: Function}} A new-style rule.
     */
    function normalizeRule(rule) {
        return typeof rule === "function" ? Object.assign({ create: rule }, rule) : rule;
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    class Rules {
        constructor() {
            this._rules = Object.create(null);
        }
        /**
         * Registers a rule module for rule id in storage.
         * @param {string} ruleId Rule id (file name).
         * @param {Function} ruleModule Rule handler.
         * @returns {void}
         */
        define(ruleId, ruleModule) {
            this._rules[ruleId] = normalizeRule(ruleModule);
        }
        /**
         * Access rule handler by id (file name).
         * @param {string} ruleId Rule id (file name).
         * @returns {{create: Function, schema: JsonSchema[]}}
         * A rule. This is normalized to always have the new-style shape with a `create` method.
         */
        get(ruleId) {
            if (typeof this._rules[ruleId] === "string") {
                this.define(ruleId, require(this._rules[ruleId]));
            }
            if (this._rules[ruleId]) {
                return this._rules[ruleId];
            }
            if (builtInRules.has(ruleId)) {
                return builtInRules.get(ruleId);
            }
            return null;
        }
        *[Symbol.iterator]() {
            yield* builtInRules;
            for (const ruleId of Object.keys(this._rules)) {
                yield [ruleId, this.get(ruleId)];
            }
        }
    }
    module.exports = Rules;
},
/* eslint/lib/rules/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint sort-keys: ["error", "asc"] */
    const { LazyLoadingRuleMap } = require(392) /* ./utils/lazy-loading-rule-map */;
    /** @type {Map<string, import("../shared/types").Rule>} */
    module.exports = new LazyLoadingRuleMap(Object.entries({
        "accessor-pairs": () => require(393) /* ./accessor-pairs */,
        "array-bracket-newline": () => require(399) /* ./array-bracket-newline */,
        "array-bracket-spacing": () => require(400) /* ./array-bracket-spacing */,
        "array-callback-return": () => require(401) /* ./array-callback-return */,
        "array-element-newline": () => require(402) /* ./array-element-newline */,
        "arrow-body-style": () => require(403) /* ./arrow-body-style */,
        "arrow-parens": () => require(404) /* ./arrow-parens */,
        "arrow-spacing": () => require(405) /* ./arrow-spacing */,
        "block-scoped-var": () => require(406) /* ./block-scoped-var */,
        "block-spacing": () => require(407) /* ./block-spacing */,
        "brace-style": () => require(408) /* ./brace-style */,
        "callback-return": () => require(409) /* ./callback-return */,
        camelcase: () => require(410) /* ./camelcase */,
        "capitalized-comments": () => require(411) /* ./capitalized-comments */,
        "class-methods-use-this": () => require(413) /* ./class-methods-use-this */,
        "comma-dangle": () => require(414) /* ./comma-dangle */,
        "comma-spacing": () => require(415) /* ./comma-spacing */,
        "comma-style": () => require(416) /* ./comma-style */,
        complexity: () => require(417) /* ./complexity */,
        "computed-property-spacing": () => require(419) /* ./computed-property-spacing */,
        "consistent-return": () => require(420) /* ./consistent-return */,
        "consistent-this": () => require(421) /* ./consistent-this */,
        "constructor-super": () => require(422) /* ./constructor-super */,
        curly: () => require(423) /* ./curly */,
        "default-case": () => require(424) /* ./default-case */,
        "default-case-last": () => require(425) /* ./default-case-last */,
        "default-param-last": () => require(426) /* ./default-param-last */,
        "dot-location": () => require(427) /* ./dot-location */,
        "dot-notation": () => require(428) /* ./dot-notation */,
        "eol-last": () => require(430) /* ./eol-last */,
        eqeqeq: () => require(431) /* ./eqeqeq */,
        "for-direction": () => require(432) /* ./for-direction */,
        "func-call-spacing": () => require(433) /* ./func-call-spacing */,
        "func-name-matching": () => require(434) /* ./func-name-matching */,
        "func-names": () => require(435) /* ./func-names */,
        "func-style": () => require(436) /* ./func-style */,
        "function-call-argument-newline": () => require(437) /* ./function-call-argument-newline */,
        "function-paren-newline": () => require(438) /* ./function-paren-newline */,
        "generator-star-spacing": () => require(439) /* ./generator-star-spacing */,
        "getter-return": () => require(440) /* ./getter-return */,
        "global-require": () => require(441) /* ./global-require */,
        "grouped-accessor-pairs": () => require(442) /* ./grouped-accessor-pairs */,
        "guard-for-in": () => require(443) /* ./guard-for-in */,
        "handle-callback-err": () => require(444) /* ./handle-callback-err */,
        "id-blacklist": () => require(445) /* ./id-blacklist */,
        "id-denylist": () => require(446) /* ./id-denylist */,
        "id-length": () => require(447) /* ./id-length */,
        "id-match": () => require(448) /* ./id-match */,
        "implicit-arrow-linebreak": () => require(449) /* ./implicit-arrow-linebreak */,
        indent: () => require(450) /* ./indent */,
        "indent-legacy": () => require(452) /* ./indent-legacy */,
        "init-declarations": () => require(453) /* ./init-declarations */,
        "jsx-quotes": () => require(454) /* ./jsx-quotes */,
        "key-spacing": () => require(455) /* ./key-spacing */,
        "keyword-spacing": () => require(456) /* ./keyword-spacing */,
        "line-comment-position": () => require(457) /* ./line-comment-position */,
        "linebreak-style": () => require(458) /* ./linebreak-style */,
        "lines-around-comment": () => require(459) /* ./lines-around-comment */,
        "lines-around-directive": () => require(460) /* ./lines-around-directive */,
        "lines-between-class-members": () => require(461) /* ./lines-between-class-members */,
        "max-classes-per-file": () => require(462) /* ./max-classes-per-file */,
        "max-depth": () => require(463) /* ./max-depth */,
        "max-len": () => require(464) /* ./max-len */,
        "max-lines": () => require(465) /* ./max-lines */,
        "max-lines-per-function": () => require(466) /* ./max-lines-per-function */,
        "max-nested-callbacks": () => require(467) /* ./max-nested-callbacks */,
        "max-params": () => require(468) /* ./max-params */,
        "max-statements": () => require(469) /* ./max-statements */,
        "max-statements-per-line": () => require(470) /* ./max-statements-per-line */,
        "multiline-comment-style": () => require(471) /* ./multiline-comment-style */,
        "multiline-ternary": () => require(472) /* ./multiline-ternary */,
        "new-cap": () => require(473) /* ./new-cap */,
        "new-parens": () => require(474) /* ./new-parens */,
        "newline-after-var": () => require(475) /* ./newline-after-var */,
        "newline-before-return": () => require(476) /* ./newline-before-return */,
        "newline-per-chained-call": () => require(477) /* ./newline-per-chained-call */,
        "no-alert": () => require(478) /* ./no-alert */,
        "no-array-constructor": () => require(479) /* ./no-array-constructor */,
        "no-async-promise-executor": () => require(480) /* ./no-async-promise-executor */,
        "no-await-in-loop": () => require(481) /* ./no-await-in-loop */,
        "no-bitwise": () => require(482) /* ./no-bitwise */,
        "no-buffer-constructor": () => require(483) /* ./no-buffer-constructor */,
        "no-caller": () => require(484) /* ./no-caller */,
        "no-case-declarations": () => require(485) /* ./no-case-declarations */,
        "no-catch-shadow": () => require(486) /* ./no-catch-shadow */,
        "no-class-assign": () => require(487) /* ./no-class-assign */,
        "no-compare-neg-zero": () => require(488) /* ./no-compare-neg-zero */,
        "no-cond-assign": () => require(489) /* ./no-cond-assign */,
        "no-confusing-arrow": () => require(490) /* ./no-confusing-arrow */,
        "no-console": () => require(491) /* ./no-console */,
        "no-const-assign": () => require(492) /* ./no-const-assign */,
        "no-constant-condition": () => require(493) /* ./no-constant-condition */,
        "no-constructor-return": () => require(494) /* ./no-constructor-return */,
        "no-continue": () => require(495) /* ./no-continue */,
        "no-control-regex": () => require(496) /* ./no-control-regex */,
        "no-debugger": () => require(498) /* ./no-debugger */,
        "no-delete-var": () => require(499) /* ./no-delete-var */,
        "no-div-regex": () => require(500) /* ./no-div-regex */,
        "no-dupe-args": () => require(501) /* ./no-dupe-args */,
        "no-dupe-class-members": () => require(502) /* ./no-dupe-class-members */,
        "no-dupe-else-if": () => require(503) /* ./no-dupe-else-if */,
        "no-dupe-keys": () => require(504) /* ./no-dupe-keys */,
        "no-duplicate-case": () => require(505) /* ./no-duplicate-case */,
        "no-duplicate-imports": () => require(506) /* ./no-duplicate-imports */,
        "no-else-return": () => require(507) /* ./no-else-return */,
        "no-empty": () => require(509) /* ./no-empty */,
        "no-empty-character-class": () => require(510) /* ./no-empty-character-class */,
        "no-empty-function": () => require(511) /* ./no-empty-function */,
        "no-empty-pattern": () => require(512) /* ./no-empty-pattern */,
        "no-eq-null": () => require(513) /* ./no-eq-null */,
        "no-eval": () => require(514) /* ./no-eval */,
        "no-ex-assign": () => require(515) /* ./no-ex-assign */,
        "no-extend-native": () => require(516) /* ./no-extend-native */,
        "no-extra-bind": () => require(517) /* ./no-extra-bind */,
        "no-extra-boolean-cast": () => require(518) /* ./no-extra-boolean-cast */,
        "no-extra-label": () => require(519) /* ./no-extra-label */,
        "no-extra-parens": () => require(520) /* ./no-extra-parens */,
        "no-extra-semi": () => require(521) /* ./no-extra-semi */,
        "no-fallthrough": () => require(522) /* ./no-fallthrough */,
        "no-floating-decimal": () => require(523) /* ./no-floating-decimal */,
        "no-func-assign": () => require(524) /* ./no-func-assign */,
        "no-global-assign": () => require(525) /* ./no-global-assign */,
        "no-implicit-coercion": () => require(526) /* ./no-implicit-coercion */,
        "no-implicit-globals": () => require(527) /* ./no-implicit-globals */,
        "no-implied-eval": () => require(528) /* ./no-implied-eval */,
        "no-import-assign": () => require(529) /* ./no-import-assign */,
        "no-inline-comments": () => require(530) /* ./no-inline-comments */,
        "no-inner-declarations": () => require(531) /* ./no-inner-declarations */,
        "no-invalid-regexp": () => require(532) /* ./no-invalid-regexp */,
        "no-invalid-this": () => require(533) /* ./no-invalid-this */,
        "no-irregular-whitespace": () => require(534) /* ./no-irregular-whitespace */,
        "no-iterator": () => require(535) /* ./no-iterator */,
        "no-label-var": () => require(536) /* ./no-label-var */,
        "no-labels": () => require(537) /* ./no-labels */,
        "no-lone-blocks": () => require(538) /* ./no-lone-blocks */,
        "no-lonely-if": () => require(539) /* ./no-lonely-if */,
        "no-loop-func": () => require(540) /* ./no-loop-func */,
        "no-loss-of-precision": () => require(541) /* ./no-loss-of-precision */,
        "no-magic-numbers": () => require(542) /* ./no-magic-numbers */,
        "no-misleading-character-class": () => require(543) /* ./no-misleading-character-class */,
        "no-mixed-operators": () => require(549) /* ./no-mixed-operators */,
        "no-mixed-requires": () => require(550) /* ./no-mixed-requires */,
        "no-mixed-spaces-and-tabs": () => require(551) /* ./no-mixed-spaces-and-tabs */,
        "no-multi-assign": () => require(552) /* ./no-multi-assign */,
        "no-multi-spaces": () => require(553) /* ./no-multi-spaces */,
        "no-multi-str": () => require(554) /* ./no-multi-str */,
        "no-multiple-empty-lines": () => require(555) /* ./no-multiple-empty-lines */,
        "no-native-reassign": () => require(556) /* ./no-native-reassign */,
        "no-negated-condition": () => require(557) /* ./no-negated-condition */,
        "no-negated-in-lhs": () => require(558) /* ./no-negated-in-lhs */,
        "no-nested-ternary": () => require(559) /* ./no-nested-ternary */,
        "no-new": () => require(560) /* ./no-new */,
        "no-new-func": () => require(561) /* ./no-new-func */,
        "no-new-object": () => require(562) /* ./no-new-object */,
        "no-new-require": () => require(563) /* ./no-new-require */,
        "no-new-symbol": () => require(564) /* ./no-new-symbol */,
        "no-new-wrappers": () => require(565) /* ./no-new-wrappers */,
        "no-nonoctal-decimal-escape": () => require(566) /* ./no-nonoctal-decimal-escape */,
        "no-obj-calls": () => require(567) /* ./no-obj-calls */,
        "no-octal": () => require(568) /* ./no-octal */,
        "no-octal-escape": () => require(569) /* ./no-octal-escape */,
        "no-param-reassign": () => require(570) /* ./no-param-reassign */,
        "no-path-concat": () => require(571) /* ./no-path-concat */,
        "no-plusplus": () => require(572) /* ./no-plusplus */,
        "no-process-env": () => require(573) /* ./no-process-env */,
        "no-process-exit": () => require(574) /* ./no-process-exit */,
        "no-promise-executor-return": () => require(575) /* ./no-promise-executor-return */,
        "no-proto": () => require(576) /* ./no-proto */,
        "no-prototype-builtins": () => require(577) /* ./no-prototype-builtins */,
        "no-redeclare": () => require(578) /* ./no-redeclare */,
        "no-regex-spaces": () => require(579) /* ./no-regex-spaces */,
        "no-restricted-exports": () => require(580) /* ./no-restricted-exports */,
        "no-restricted-globals": () => require(581) /* ./no-restricted-globals */,
        "no-restricted-imports": () => require(582) /* ./no-restricted-imports */,
        "no-restricted-modules": () => require(583) /* ./no-restricted-modules */,
        "no-restricted-properties": () => require(584) /* ./no-restricted-properties */,
        "no-restricted-syntax": () => require(585) /* ./no-restricted-syntax */,
        "no-return-assign": () => require(586) /* ./no-return-assign */,
        "no-return-await": () => require(587) /* ./no-return-await */,
        "no-script-url": () => require(588) /* ./no-script-url */,
        "no-self-assign": () => require(589) /* ./no-self-assign */,
        "no-self-compare": () => require(590) /* ./no-self-compare */,
        "no-sequences": () => require(591) /* ./no-sequences */,
        "no-setter-return": () => require(592) /* ./no-setter-return */,
        "no-shadow": () => require(593) /* ./no-shadow */,
        "no-shadow-restricted-names": () => require(594) /* ./no-shadow-restricted-names */,
        "no-spaced-func": () => require(595) /* ./no-spaced-func */,
        "no-sparse-arrays": () => require(596) /* ./no-sparse-arrays */,
        "no-sync": () => require(597) /* ./no-sync */,
        "no-tabs": () => require(598) /* ./no-tabs */,
        "no-template-curly-in-string": () => require(599) /* ./no-template-curly-in-string */,
        "no-ternary": () => require(600) /* ./no-ternary */,
        "no-this-before-super": () => require(601) /* ./no-this-before-super */,
        "no-throw-literal": () => require(602) /* ./no-throw-literal */,
        "no-trailing-spaces": () => require(603) /* ./no-trailing-spaces */,
        "no-undef": () => require(604) /* ./no-undef */,
        "no-undef-init": () => require(605) /* ./no-undef-init */,
        "no-undefined": () => require(606) /* ./no-undefined */,
        "no-underscore-dangle": () => require(607) /* ./no-underscore-dangle */,
        "no-unexpected-multiline": () => require(608) /* ./no-unexpected-multiline */,
        "no-unmodified-loop-condition": () => require(609) /* ./no-unmodified-loop-condition */,
        "no-unneeded-ternary": () => require(610) /* ./no-unneeded-ternary */,
        "no-unreachable": () => require(611) /* ./no-unreachable */,
        "no-unreachable-loop": () => require(612) /* ./no-unreachable-loop */,
        "no-unsafe-finally": () => require(613) /* ./no-unsafe-finally */,
        "no-unsafe-negation": () => require(614) /* ./no-unsafe-negation */,
        "no-unsafe-optional-chaining": () => require(615) /* ./no-unsafe-optional-chaining */,
        "no-unused-expressions": () => require(616) /* ./no-unused-expressions */,
        "no-unused-labels": () => require(617) /* ./no-unused-labels */,
        "no-unused-vars": () => require(618) /* ./no-unused-vars */,
        "no-use-before-define": () => require(619) /* ./no-use-before-define */,
        "no-useless-backreference": () => require(620) /* ./no-useless-backreference */,
        "no-useless-call": () => require(621) /* ./no-useless-call */,
        "no-useless-catch": () => require(622) /* ./no-useless-catch */,
        "no-useless-computed-key": () => require(623) /* ./no-useless-computed-key */,
        "no-useless-concat": () => require(624) /* ./no-useless-concat */,
        "no-useless-constructor": () => require(625) /* ./no-useless-constructor */,
        "no-useless-escape": () => require(626) /* ./no-useless-escape */,
        "no-useless-rename": () => require(627) /* ./no-useless-rename */,
        "no-useless-return": () => require(628) /* ./no-useless-return */,
        "no-var": () => require(629) /* ./no-var */,
        "no-void": () => require(630) /* ./no-void */,
        "no-warning-comments": () => require(631) /* ./no-warning-comments */,
        "no-whitespace-before-property": () => require(632) /* ./no-whitespace-before-property */,
        "no-with": () => require(633) /* ./no-with */,
        "nonblock-statement-body-position": () => require(634) /* ./nonblock-statement-body-position */,
        "object-curly-newline": () => require(635) /* ./object-curly-newline */,
        "object-curly-spacing": () => require(636) /* ./object-curly-spacing */,
        "object-property-newline": () => require(637) /* ./object-property-newline */,
        "object-shorthand": () => require(638) /* ./object-shorthand */,
        "one-var": () => require(639) /* ./one-var */,
        "one-var-declaration-per-line": () => require(640) /* ./one-var-declaration-per-line */,
        "operator-assignment": () => require(641) /* ./operator-assignment */,
        "operator-linebreak": () => require(642) /* ./operator-linebreak */,
        "padded-blocks": () => require(643) /* ./padded-blocks */,
        "padding-line-between-statements": () => require(644) /* ./padding-line-between-statements */,
        "prefer-arrow-callback": () => require(645) /* ./prefer-arrow-callback */,
        "prefer-const": () => require(646) /* ./prefer-const */,
        "prefer-destructuring": () => require(647) /* ./prefer-destructuring */,
        "prefer-exponentiation-operator": () => require(648) /* ./prefer-exponentiation-operator */,
        "prefer-named-capture-group": () => require(649) /* ./prefer-named-capture-group */,
        "prefer-numeric-literals": () => require(650) /* ./prefer-numeric-literals */,
        "prefer-object-spread": () => require(651) /* ./prefer-object-spread */,
        "prefer-promise-reject-errors": () => require(652) /* ./prefer-promise-reject-errors */,
        "prefer-reflect": () => require(653) /* ./prefer-reflect */,
        "prefer-regex-literals": () => require(654) /* ./prefer-regex-literals */,
        "prefer-rest-params": () => require(655) /* ./prefer-rest-params */,
        "prefer-spread": () => require(656) /* ./prefer-spread */,
        "prefer-template": () => require(657) /* ./prefer-template */,
        "quote-props": () => require(658) /* ./quote-props */,
        quotes: () => require(659) /* ./quotes */,
        radix: () => require(660) /* ./radix */,
        "require-atomic-updates": () => require(661) /* ./require-atomic-updates */,
        "require-await": () => require(662) /* ./require-await */,
        "require-jsdoc": () => require(663) /* ./require-jsdoc */,
        "require-unicode-regexp": () => require(664) /* ./require-unicode-regexp */,
        "require-yield": () => require(665) /* ./require-yield */,
        "rest-spread-spacing": () => require(666) /* ./rest-spread-spacing */,
        semi: () => require(667) /* ./semi */,
        "semi-spacing": () => require(668) /* ./semi-spacing */,
        "semi-style": () => require(669) /* ./semi-style */,
        "sort-imports": () => require(670) /* ./sort-imports */,
        "sort-keys": () => require(671) /* ./sort-keys */,
        "sort-vars": () => require(673) /* ./sort-vars */,
        "space-before-blocks": () => require(674) /* ./space-before-blocks */,
        "space-before-function-paren": () => require(675) /* ./space-before-function-paren */,
        "space-in-parens": () => require(676) /* ./space-in-parens */,
        "space-infix-ops": () => require(677) /* ./space-infix-ops */,
        "space-unary-ops": () => require(678) /* ./space-unary-ops */,
        "spaced-comment": () => require(679) /* ./spaced-comment */,
        strict: () => require(680) /* ./strict */,
        "switch-colon-spacing": () => require(681) /* ./switch-colon-spacing */,
        "symbol-description": () => require(682) /* ./symbol-description */,
        "template-curly-spacing": () => require(683) /* ./template-curly-spacing */,
        "template-tag-spacing": () => require(684) /* ./template-tag-spacing */,
        "unicode-bom": () => require(685) /* ./unicode-bom */,
        "use-isnan": () => require(686) /* ./use-isnan */,
        "valid-jsdoc": () => require(687) /* ./valid-jsdoc */,
        "valid-typeof": () => require(692) /* ./valid-typeof */,
        "vars-on-top": () => require(693) /* ./vars-on-top */,
        "wrap-iife": () => require(694) /* ./wrap-iife */,
        "wrap-regex": () => require(695) /* ./wrap-regex */,
        "yield-star-spacing": () => require(696) /* ./yield-star-spacing */,
        yoda: () => require(697) /* ./yoda */
    }));
},
/* eslint/lib/rules/utils/lazy-loading-rule-map.js */ function _(require, module, exports, __esModule, __esExport) {
    const debug = require(72) /* debug */("eslint:rules");
    /** @typedef {import("./types").Rule} Rule */
    /**
     * The `Map` object that loads each rule when it's accessed.
     * @example
     * const rules = new LazyLoadingRuleMap([
     *     ["eqeqeq", () => require("eqeqeq")],
     *     ["semi", () => require("semi")],
     *     ["no-unused-vars", () => require("no-unused-vars")],
     * ])
     *
     * rules.get("semi") // call `() => require("semi")` here.
     *
     * @extends {Map<string, () => Rule>}
     */
    class LazyLoadingRuleMap extends Map {
        /**
         * Initialize this map.
         * @param {Array<[string, function(): Rule]>} loaders The rule loaders.
         */
        constructor(loaders) {
            let remaining = loaders.length;
            super(debug.enabled
                ? loaders.map(([ruleId, load]) => {
                    let cache = null;
                    return [
                        ruleId,
                        () => {
                            if (!cache) {
                                debug("Loading rule %o (remaining=%d)", ruleId, --remaining);
                                cache = load();
                            }
                            return cache;
                        }
                    ];
                })
                : loaders);
            // `super(...iterable)` uses `this.set()`, so disable it here.
            Object.defineProperty(LazyLoadingRuleMap.prototype, "set", {
                configurable: true,
                value: void 0
            });
        }
        /**
         * Get a rule.
         * Each rule will be loaded on the first access.
         * @param {string} ruleId The rule ID to get.
         * @returns {Rule|undefined} The rule.
         */
        get(ruleId) {
            const load = super.get(ruleId);
            return load && load();
        }
        /**
         * Iterate rules.
         * @returns {IterableIterator<Rule>} Rules.
         */
        *values() {
            for (const load of super.values()) {
                yield load();
            }
        }
        /**
         * Iterate rules.
         * @returns {IterableIterator<[string, Rule]>} Rules.
         */
        *entries() {
            for (const [ruleId, load] of super.entries()) {
                yield [ruleId, load()];
            }
        }
        /**
         * Call a function with each rule.
         * @param {Function} callbackFn The callback function.
         * @param {any} [thisArg] The object to pass to `this` of the callback function.
         * @returns {void}
         */
        forEach(callbackFn, thisArg) {
            for (const [ruleId, load] of super.entries()) {
                callbackFn.call(thisArg, load(), ruleId, this);
            }
        }
    }
    // Forbid mutation.
    Object.defineProperties(LazyLoadingRuleMap.prototype, {
        clear: { configurable: true, value: void 0 },
        delete: { configurable: true, value: void 0 },
        [Symbol.iterator]: {
            configurable: true,
            writable: true,
            value: LazyLoadingRuleMap.prototype.entries
        }
    });
    module.exports = { LazyLoadingRuleMap };
},
/* eslint/lib/rules/accessor-pairs.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    /**
     * Property name if it can be computed statically, otherwise the list of the tokens of the key node.
     * @typedef {string|Token[]} Key
     */
    /**
     * Accessor nodes with the same key.
     * @typedef {Object} AccessorData
     * @property {Key} key Accessor's key
     * @property {ASTNode[]} getters List of getter nodes.
     * @property {ASTNode[]} setters List of setter nodes.
     */
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not the given lists represent the equal tokens in the same order.
     * Tokens are compared by their properties, not by instance.
     * @param {Token[]} left First list of tokens.
     * @param {Token[]} right Second list of tokens.
     * @returns {boolean} `true` if the lists have same tokens.
     */
    function areEqualTokenLists(left, right) {
        if (left.length !== right.length) {
            return false;
        }
        for (let i = 0; i < left.length; i++) {
            const leftToken = left[i], rightToken = right[i];
            if (leftToken.type !== rightToken.type || leftToken.value !== rightToken.value) {
                return false;
            }
        }
        return true;
    }
    /**
     * Checks whether or not the given keys are equal.
     * @param {Key} left First key.
     * @param {Key} right Second key.
     * @returns {boolean} `true` if the keys are equal.
     */
    function areEqualKeys(left, right) {
        if (typeof left === "string" && typeof right === "string") {
            // Statically computed names.
            return left === right;
        }
        if (Array.isArray(left) && Array.isArray(right)) {
            // Token lists.
            return areEqualTokenLists(left, right);
        }
        return false;
    }
    /**
     * Checks whether or not a given node is of an accessor kind ('get' or 'set').
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node is of an accessor kind.
     */
    function isAccessorKind(node) {
        return node.kind === "get" || node.kind === "set";
    }
    /**
     * Checks whether or not a given node is an argument of a specified method call.
     * @param {ASTNode} node A node to check.
     * @param {number} index An expected index of the node in arguments.
     * @param {string} object An expected name of the object of the method.
     * @param {string} property An expected name of the method.
     * @returns {boolean} `true` if the node is an argument of the specified method call.
     */
    function isArgumentOfMethodCall(node, index, object, property) {
        const parent = node.parent;
        return (parent.type === "CallExpression" &&
            astUtils.isSpecificMemberAccess(parent.callee, object, property) &&
            parent.arguments[index] === node);
    }
    /**
     * Checks whether or not a given node is a property descriptor.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node is a property descriptor.
     */
    function isPropertyDescriptor(node) {
        // Object.defineProperty(obj, "foo", {set: ...})
        if (isArgumentOfMethodCall(node, 2, "Object", "defineProperty") ||
            isArgumentOfMethodCall(node, 2, "Reflect", "defineProperty")) {
            return true;
        }
        /*
         * Object.defineProperties(obj, {foo: {set: ...}})
         * Object.create(proto, {foo: {set: ...}})
         */
        const grandparent = node.parent.parent;
        return grandparent.type === "ObjectExpression" && (isArgumentOfMethodCall(grandparent, 1, "Object", "create") ||
            isArgumentOfMethodCall(grandparent, 1, "Object", "defineProperties"));
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce getter and setter pairs in objects and classes",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/accessor-pairs"
            },
            schema: [{
                    type: "object",
                    properties: {
                        getWithoutSet: {
                            type: "boolean",
                            default: false
                        },
                        setWithoutGet: {
                            type: "boolean",
                            default: true
                        },
                        enforceForClassMembers: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                missingGetterInPropertyDescriptor: "Getter is not present in property descriptor.",
                missingSetterInPropertyDescriptor: "Setter is not present in property descriptor.",
                missingGetterInObjectLiteral: "Getter is not present for {{ name }}.",
                missingSetterInObjectLiteral: "Setter is not present for {{ name }}.",
                missingGetterInClass: "Getter is not present for class {{ name }}.",
                missingSetterInClass: "Setter is not present for class {{ name }}."
            }
        },
        create(context) {
            const config = context.options[0] || {};
            const checkGetWithoutSet = config.getWithoutSet === true;
            const checkSetWithoutGet = config.setWithoutGet !== false;
            const enforceForClassMembers = config.enforceForClassMembers !== false;
            const sourceCode = context.getSourceCode();
            /**
             * Reports the given node.
             * @param {ASTNode} node The node to report.
             * @param {string} messageKind "missingGetter" or "missingSetter".
             * @returns {void}
             * @private
             */
            function report(node, messageKind) {
                if (node.type === "Property") {
                    context.report({
                        node,
                        messageId: `${messageKind}InObjectLiteral`,
                        loc: astUtils.getFunctionHeadLoc(node.value, sourceCode),
                        data: { name: astUtils.getFunctionNameWithKind(node.value) }
                    });
                }
                else if (node.type === "MethodDefinition") {
                    context.report({
                        node,
                        messageId: `${messageKind}InClass`,
                        loc: astUtils.getFunctionHeadLoc(node.value, sourceCode),
                        data: { name: astUtils.getFunctionNameWithKind(node.value) }
                    });
                }
                else {
                    context.report({
                        node,
                        messageId: `${messageKind}InPropertyDescriptor`
                    });
                }
            }
            /**
             * Reports each of the nodes in the given list using the same messageId.
             * @param {ASTNode[]} nodes Nodes to report.
             * @param {string} messageKind "missingGetter" or "missingSetter".
             * @returns {void}
             * @private
             */
            function reportList(nodes, messageKind) {
                for (const node of nodes) {
                    report(node, messageKind);
                }
            }
            /**
             * Creates a new `AccessorData` object for the given getter or setter node.
             * @param {ASTNode} node A getter or setter node.
             * @returns {AccessorData} New `AccessorData` object that contains the given node.
             * @private
             */
            function createAccessorData(node) {
                const name = astUtils.getStaticPropertyName(node);
                const key = (name !== null) ? name : sourceCode.getTokens(node.key);
                return {
                    key,
                    getters: node.kind === "get" ? [node] : [],
                    setters: node.kind === "set" ? [node] : []
                };
            }
            /**
             * Merges the given `AccessorData` object into the given accessors list.
             * @param {AccessorData[]} accessors The list to merge into.
             * @param {AccessorData} accessorData The object to merge.
             * @returns {AccessorData[]} The same instance with the merged object.
             * @private
             */
            function mergeAccessorData(accessors, accessorData) {
                const equalKeyElement = accessors.find(a => areEqualKeys(a.key, accessorData.key));
                if (equalKeyElement) {
                    equalKeyElement.getters.push(...accessorData.getters);
                    equalKeyElement.setters.push(...accessorData.setters);
                }
                else {
                    accessors.push(accessorData);
                }
                return accessors;
            }
            /**
             * Checks accessor pairs in the given list of nodes.
             * @param {ASTNode[]} nodes The list to check.
             * @returns {void}
             * @private
             */
            function checkList(nodes) {
                const accessors = nodes
                    .filter(isAccessorKind)
                    .map(createAccessorData)
                    .reduce(mergeAccessorData, []);
                for (const { getters, setters } of accessors) {
                    if (checkSetWithoutGet && setters.length && !getters.length) {
                        reportList(setters, "missingGetter");
                    }
                    if (checkGetWithoutSet && getters.length && !setters.length) {
                        reportList(getters, "missingSetter");
                    }
                }
            }
            /**
             * Checks accessor pairs in an object literal.
             * @param {ASTNode} node `ObjectExpression` node to check.
             * @returns {void}
             * @private
             */
            function checkObjectLiteral(node) {
                checkList(node.properties.filter(p => p.type === "Property"));
            }
            /**
             * Checks accessor pairs in a property descriptor.
             * @param {ASTNode} node Property descriptor `ObjectExpression` node to check.
             * @returns {void}
             * @private
             */
            function checkPropertyDescriptor(node) {
                const namesToCheck = node.properties
                    .filter(p => p.type === "Property" && p.kind === "init" && !p.computed)
                    .map(({ key }) => key.name);
                const hasGetter = namesToCheck.includes("get");
                const hasSetter = namesToCheck.includes("set");
                if (checkSetWithoutGet && hasSetter && !hasGetter) {
                    report(node, "missingGetter");
                }
                if (checkGetWithoutSet && hasGetter && !hasSetter) {
                    report(node, "missingSetter");
                }
            }
            /**
             * Checks the given object expression as an object literal and as a possible property descriptor.
             * @param {ASTNode} node `ObjectExpression` node to check.
             * @returns {void}
             * @private
             */
            function checkObjectExpression(node) {
                checkObjectLiteral(node);
                if (isPropertyDescriptor(node)) {
                    checkPropertyDescriptor(node);
                }
            }
            /**
             * Checks the given class body.
             * @param {ASTNode} node `ClassBody` node to check.
             * @returns {void}
             * @private
             */
            function checkClassBody(node) {
                const methodDefinitions = node.body.filter(m => m.type === "MethodDefinition");
                checkList(methodDefinitions.filter(m => m.static));
                checkList(methodDefinitions.filter(m => !m.static));
            }
            const listeners = {};
            if (checkSetWithoutGet || checkGetWithoutSet) {
                listeners.ObjectExpression = checkObjectExpression;
                if (enforceForClassMembers) {
                    listeners.ClassBody = checkClassBody;
                }
            }
            return listeners;
        }
    };
},
/* eslint/lib/rules/utils/ast-utils.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const esutils = require(395) /* esutils */;
    const espree = require(313) /* espree */;
    const escapeRegExp = require(329) /* escape-string-regexp */;
    const { breakableTypePattern, createGlobalLinebreakMatcher, lineBreakPattern, shebangPattern } = require(346) /* ../../shared/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const anyFunctionPattern = /^(?:Function(?:Declaration|Expression)|ArrowFunctionExpression)$/u;
    const anyLoopPattern = /^(?:DoWhile|For|ForIn|ForOf|While)Statement$/u;
    const arrayOrTypedArrayPattern = /Array$/u;
    const arrayMethodPattern = /^(?:every|filter|find|findIndex|forEach|map|some)$/u;
    const bindOrCallOrApplyPattern = /^(?:bind|call|apply)$/u;
    const thisTagPattern = /^[\s*]*@this/mu;
    const COMMENTS_IGNORE_PATTERN = /^\s*(?:eslint|jshint\s+|jslint\s+|istanbul\s+|globals?\s+|exported\s+|jscs)/u;
    const LINEBREAKS = new Set(["\r\n", "\r", "\n", "\u2028", "\u2029"]);
    // A set of node types that can contain a list of statements
    const STATEMENT_LIST_PARENTS = new Set(["Program", "BlockStatement", "SwitchCase"]);
    const DECIMAL_INTEGER_PATTERN = /^(?:0|0[0-7]*[89]\d*|[1-9](?:_?\d)*)$/u;
    // Tests the presence of at least one LegacyOctalEscapeSequence or NonOctalDecimalEscapeSequence in a raw string
    const OCTAL_OR_NON_OCTAL_DECIMAL_ESCAPE_PATTERN = /^(?:[^\\]|\\.)*\\(?:[1-9]|0[0-9])/su;
    const LOGICAL_ASSIGNMENT_OPERATORS = new Set(["&&=", "||=", "??="]);
    /**
     * Checks reference if is non initializer and writable.
     * @param {Reference} reference A reference to check.
     * @param {int} index The index of the reference in the references.
     * @param {Reference[]} references The array that the reference belongs to.
     * @returns {boolean} Success/Failure
     * @private
     */
    function isModifyingReference(reference, index, references) {
        const identifier = reference.identifier;
        /*
         * Destructuring assignments can have multiple default value, so
         * possibly there are multiple writeable references for the same
         * identifier.
         */
        const modifyingDifferentIdentifier = index === 0 ||
            references[index - 1].identifier !== identifier;
        return (identifier &&
            reference.init === false &&
            reference.isWrite() &&
            modifyingDifferentIdentifier);
    }
    /**
     * Checks whether the given string starts with uppercase or not.
     * @param {string} s The string to check.
     * @returns {boolean} `true` if the string starts with uppercase.
     */
    function startsWithUpperCase(s) {
        return s[0] !== s[0].toLocaleLowerCase();
    }
    /**
     * Checks whether or not a node is a constructor.
     * @param {ASTNode} node A function node to check.
     * @returns {boolean} Whether or not a node is a constructor.
     */
    function isES5Constructor(node) {
        return (node.id && startsWithUpperCase(node.id.name));
    }
    /**
     * Finds a function node from ancestors of a node.
     * @param {ASTNode} node A start node to find.
     * @returns {Node|null} A found function node.
     */
    function getUpperFunction(node) {
        for (let currentNode = node; currentNode; currentNode = currentNode.parent) {
            if (anyFunctionPattern.test(currentNode.type)) {
                return currentNode;
            }
        }
        return null;
    }
    /**
     * Checks whether a given node is a function node or not.
     * The following types are function nodes:
     *
     * - ArrowFunctionExpression
     * - FunctionDeclaration
     * - FunctionExpression
     * @param {ASTNode|null} node A node to check.
     * @returns {boolean} `true` if the node is a function node.
     */
    function isFunction(node) {
        return Boolean(node && anyFunctionPattern.test(node.type));
    }
    /**
     * Checks whether a given node is a loop node or not.
     * The following types are loop nodes:
     *
     * - DoWhileStatement
     * - ForInStatement
     * - ForOfStatement
     * - ForStatement
     * - WhileStatement
     * @param {ASTNode|null} node A node to check.
     * @returns {boolean} `true` if the node is a loop node.
     */
    function isLoop(node) {
        return Boolean(node && anyLoopPattern.test(node.type));
    }
    /**
     * Checks whether the given node is in a loop or not.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is in a loop.
     */
    function isInLoop(node) {
        for (let currentNode = node; currentNode && !isFunction(currentNode); currentNode = currentNode.parent) {
            if (isLoop(currentNode)) {
                return true;
            }
        }
        return false;
    }
    /**
     * Determines whether the given node is a `null` literal.
     * @param {ASTNode} node The node to check
     * @returns {boolean} `true` if the node is a `null` literal
     */
    function isNullLiteral(node) {
        /*
         * Checking `node.value === null` does not guarantee that a literal is a null literal.
         * When parsing values that cannot be represented in the current environment (e.g. unicode
         * regexes in Node 4), `node.value` is set to `null` because it wouldn't be possible to
         * set `node.value` to a unicode regex. To make sure a literal is actually `null`, check
         * `node.regex` instead. Also see: https://github.com/eslint/eslint/issues/8020
         */
        return node.type === "Literal" && node.value === null && !node.regex && !node.bigint;
    }
    /**
     * Checks whether or not a node is `null` or `undefined`.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} Whether or not the node is a `null` or `undefined`.
     * @public
     */
    function isNullOrUndefined(node) {
        return (isNullLiteral(node) ||
            (node.type === "Identifier" && node.name === "undefined") ||
            (node.type === "UnaryExpression" && node.operator === "void"));
    }
    /**
     * Checks whether or not a node is callee.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} Whether or not the node is callee.
     */
    function isCallee(node) {
        return node.parent.type === "CallExpression" && node.parent.callee === node;
    }
    /**
     * Returns the result of the string conversion applied to the evaluated value of the given expression node,
     * if it can be determined statically.
     *
     * This function returns a `string` value for all `Literal` nodes and simple `TemplateLiteral` nodes only.
     * In all other cases, this function returns `null`.
     * @param {ASTNode} node Expression node.
     * @returns {string|null} String value if it can be determined. Otherwise, `null`.
     */
    function getStaticStringValue(node) {
        switch (node.type) {
            case "Literal":
                if (node.value === null) {
                    if (isNullLiteral(node)) {
                        return String(node.value); // "null"
                    }
                    if (node.regex) {
                        return `/${node.regex.pattern}/${node.regex.flags}`;
                    }
                    if (node.bigint) {
                        return node.bigint;
                    }
                    // Otherwise, this is an unknown literal. The function will return null.
                }
                else {
                    return String(node.value);
                }
                break;
            case "TemplateLiteral":
                if (node.expressions.length === 0 && node.quasis.length === 1) {
                    return node.quasis[0].value.cooked;
                }
                break;
            // no default
        }
        return null;
    }
    /**
     * Gets the property name of a given node.
     * The node can be a MemberExpression, a Property, or a MethodDefinition.
     *
     * If the name is dynamic, this returns `null`.
     *
     * For examples:
     *
     *     a.b           // => "b"
     *     a["b"]        // => "b"
     *     a['b']        // => "b"
     *     a[`b`]        // => "b"
     *     a[100]        // => "100"
     *     a[b]          // => null
     *     a["a" + "b"]  // => null
     *     a[tag`b`]     // => null
     *     a[`${b}`]     // => null
     *
     *     let a = {b: 1}            // => "b"
     *     let a = {["b"]: 1}        // => "b"
     *     let a = {['b']: 1}        // => "b"
     *     let a = {[`b`]: 1}        // => "b"
     *     let a = {[100]: 1}        // => "100"
     *     let a = {[b]: 1}          // => null
     *     let a = {["a" + "b"]: 1}  // => null
     *     let a = {[tag`b`]: 1}     // => null
     *     let a = {[`${b}`]: 1}     // => null
     * @param {ASTNode} node The node to get.
     * @returns {string|null} The property name if static. Otherwise, null.
     */
    function getStaticPropertyName(node) {
        let prop;
        switch (node && node.type) {
            case "ChainExpression":
                return getStaticPropertyName(node.expression);
            case "Property":
            case "MethodDefinition":
                prop = node.key;
                break;
            case "MemberExpression":
                prop = node.property;
                break;
            // no default
        }
        if (prop) {
            if (prop.type === "Identifier" && !node.computed) {
                return prop.name;
            }
            return getStaticStringValue(prop);
        }
        return null;
    }
    /**
     * Retrieve `ChainExpression#expression` value if the given node a `ChainExpression` node. Otherwise, pass through it.
     * @param {ASTNode} node The node to address.
     * @returns {ASTNode} The `ChainExpression#expression` value if the node is a `ChainExpression` node. Otherwise, the node.
     */
    function skipChainExpression(node) {
        return node && node.type === "ChainExpression" ? node.expression : node;
    }
    /**
     * Check if the `actual` is an expected value.
     * @param {string} actual The string value to check.
     * @param {string | RegExp} expected The expected string value or pattern.
     * @returns {boolean} `true` if the `actual` is an expected value.
     */
    function checkText(actual, expected) {
        return typeof expected === "string"
            ? actual === expected
            : expected.test(actual);
    }
    /**
     * Check if a given node is an Identifier node with a given name.
     * @param {ASTNode} node The node to check.
     * @param {string | RegExp} name The expected name or the expected pattern of the object name.
     * @returns {boolean} `true` if the node is an Identifier node with the name.
     */
    function isSpecificId(node, name) {
        return node.type === "Identifier" && checkText(node.name, name);
    }
    /**
     * Check if a given node is member access with a given object name and property name pair.
     * This is regardless of optional or not.
     * @param {ASTNode} node The node to check.
     * @param {string | RegExp | null} objectName The expected name or the expected pattern of the object name. If this is nullish, this method doesn't check object.
     * @param {string | RegExp | null} propertyName The expected name or the expected pattern of the property name. If this is nullish, this method doesn't check property.
     * @returns {boolean} `true` if the node is member access with the object name and property name pair.
     * The node is a `MemberExpression` or `ChainExpression`.
     */
    function isSpecificMemberAccess(node, objectName, propertyName) {
        const checkNode = skipChainExpression(node);
        if (checkNode.type !== "MemberExpression") {
            return false;
        }
        if (objectName && !isSpecificId(checkNode.object, objectName)) {
            return false;
        }
        if (propertyName) {
            const actualPropertyName = getStaticPropertyName(checkNode);
            if (typeof actualPropertyName !== "string" || !checkText(actualPropertyName, propertyName)) {
                return false;
            }
        }
        return true;
    }
    /**
     * Check if two literal nodes are the same value.
     * @param {ASTNode} left The Literal node to compare.
     * @param {ASTNode} right The other Literal node to compare.
     * @returns {boolean} `true` if the two literal nodes are the same value.
     */
    function equalLiteralValue(left, right) {
        // RegExp literal.
        if (left.regex || right.regex) {
            return Boolean(left.regex &&
                right.regex &&
                left.regex.pattern === right.regex.pattern &&
                left.regex.flags === right.regex.flags);
        }
        // BigInt literal.
        if (left.bigint || right.bigint) {
            return left.bigint === right.bigint;
        }
        return left.value === right.value;
    }
    /**
     * Check if two expressions reference the same value. For example:
     *     a = a
     *     a.b = a.b
     *     a[0] = a[0]
     *     a['b'] = a['b']
     * @param {ASTNode} left The left side of the comparison.
     * @param {ASTNode} right The right side of the comparison.
     * @param {boolean} [disableStaticComputedKey] Don't address `a.b` and `a["b"]` are the same if `true`. For backward compatibility.
     * @returns {boolean} `true` if both sides match and reference the same value.
     */
    function isSameReference(left, right, disableStaticComputedKey = false) {
        if (left.type !== right.type) {
            // Handle `a.b` and `a?.b` are samely.
            if (left.type === "ChainExpression") {
                return isSameReference(left.expression, right, disableStaticComputedKey);
            }
            if (right.type === "ChainExpression") {
                return isSameReference(left, right.expression, disableStaticComputedKey);
            }
            return false;
        }
        switch (left.type) {
            case "Super":
            case "ThisExpression":
                return true;
            case "Identifier":
                return left.name === right.name;
            case "Literal":
                return equalLiteralValue(left, right);
            case "ChainExpression":
                return isSameReference(left.expression, right.expression, disableStaticComputedKey);
            case "MemberExpression": {
                if (!disableStaticComputedKey) {
                    const nameA = getStaticPropertyName(left);
                    // x.y = x["y"]
                    if (nameA !== null) {
                        return (isSameReference(left.object, right.object, disableStaticComputedKey) &&
                            nameA === getStaticPropertyName(right));
                    }
                }
                /*
                 * x[0] = x[0]
                 * x[y] = x[y]
                 * x.y = x.y
                 */
                return (left.computed === right.computed &&
                    isSameReference(left.object, right.object, disableStaticComputedKey) &&
                    isSameReference(left.property, right.property, disableStaticComputedKey));
            }
            default:
                return false;
        }
    }
    /**
     * Checks whether or not a node is `Reflect.apply`.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} Whether or not the node is a `Reflect.apply`.
     */
    function isReflectApply(node) {
        return isSpecificMemberAccess(node, "Reflect", "apply");
    }
    /**
     * Checks whether or not a node is `Array.from`.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} Whether or not the node is a `Array.from`.
     */
    function isArrayFromMethod(node) {
        return isSpecificMemberAccess(node, arrayOrTypedArrayPattern, "from");
    }
    /**
     * Checks whether or not a node is a method which has `thisArg`.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} Whether or not the node is a method which has `thisArg`.
     */
    function isMethodWhichHasThisArg(node) {
        return isSpecificMemberAccess(node, null, arrayMethodPattern);
    }
    /**
     * Creates the negate function of the given function.
     * @param {Function} f The function to negate.
     * @returns {Function} Negated function.
     */
    function negate(f) {
        return token => !f(token);
    }
    /**
     * Checks whether or not a node has a `@this` tag in its comments.
     * @param {ASTNode} node A node to check.
     * @param {SourceCode} sourceCode A SourceCode instance to get comments.
     * @returns {boolean} Whether or not the node has a `@this` tag in its comments.
     */
    function hasJSDocThisTag(node, sourceCode) {
        const jsdocComment = sourceCode.getJSDocComment(node);
        if (jsdocComment && thisTagPattern.test(jsdocComment.value)) {
            return true;
        }
        // Checks `@this` in its leading comments for callbacks,
        // because callbacks don't have its JSDoc comment.
        // e.g.
        //     sinon.test(/* @this sinon.Sandbox */function() { this.spy(); });
        return sourceCode.getCommentsBefore(node).some(comment => thisTagPattern.test(comment.value));
    }
    /**
     * Determines if a node is surrounded by parentheses.
     * @param {SourceCode} sourceCode The ESLint source code object
     * @param {ASTNode} node The node to be checked.
     * @returns {boolean} True if the node is parenthesised.
     * @private
     */
    function isParenthesised(sourceCode, node) {
        const previousToken = sourceCode.getTokenBefore(node), nextToken = sourceCode.getTokenAfter(node);
        return Boolean(previousToken && nextToken) &&
            previousToken.value === "(" && previousToken.range[1] <= node.range[0] &&
            nextToken.value === ")" && nextToken.range[0] >= node.range[1];
    }
    /**
     * Checks if the given token is an arrow token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is an arrow token.
     */
    function isArrowToken(token) {
        return token.value === "=>" && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is a comma token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a comma token.
     */
    function isCommaToken(token) {
        return token.value === "," && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is a dot token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a dot token.
     */
    function isDotToken(token) {
        return token.value === "." && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is a `?.` token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a `?.` token.
     */
    function isQuestionDotToken(token) {
        return token.value === "?." && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is a semicolon token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a semicolon token.
     */
    function isSemicolonToken(token) {
        return token.value === ";" && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is a colon token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a colon token.
     */
    function isColonToken(token) {
        return token.value === ":" && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is an opening parenthesis token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is an opening parenthesis token.
     */
    function isOpeningParenToken(token) {
        return token.value === "(" && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is a closing parenthesis token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a closing parenthesis token.
     */
    function isClosingParenToken(token) {
        return token.value === ")" && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is an opening square bracket token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is an opening square bracket token.
     */
    function isOpeningBracketToken(token) {
        return token.value === "[" && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is a closing square bracket token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a closing square bracket token.
     */
    function isClosingBracketToken(token) {
        return token.value === "]" && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is an opening brace token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is an opening brace token.
     */
    function isOpeningBraceToken(token) {
        return token.value === "{" && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is a closing brace token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a closing brace token.
     */
    function isClosingBraceToken(token) {
        return token.value === "}" && token.type === "Punctuator";
    }
    /**
     * Checks if the given token is a comment token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a comment token.
     */
    function isCommentToken(token) {
        return token.type === "Line" || token.type === "Block" || token.type === "Shebang";
    }
    /**
     * Checks if the given token is a keyword token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a keyword token.
     */
    function isKeywordToken(token) {
        return token.type === "Keyword";
    }
    /**
     * Gets the `(` token of the given function node.
     * @param {ASTNode} node The function node to get.
     * @param {SourceCode} sourceCode The source code object to get tokens.
     * @returns {Token} `(` token.
     */
    function getOpeningParenOfParams(node, sourceCode) {
        return node.id
            ? sourceCode.getTokenAfter(node.id, isOpeningParenToken)
            : sourceCode.getFirstToken(node, isOpeningParenToken);
    }
    /**
     * Checks whether or not the tokens of two given nodes are same.
     * @param {ASTNode} left A node 1 to compare.
     * @param {ASTNode} right A node 2 to compare.
     * @param {SourceCode} sourceCode The ESLint source code object.
     * @returns {boolean} the source code for the given node.
     */
    function equalTokens(left, right, sourceCode) {
        const tokensL = sourceCode.getTokens(left);
        const tokensR = sourceCode.getTokens(right);
        if (tokensL.length !== tokensR.length) {
            return false;
        }
        for (let i = 0; i < tokensL.length; ++i) {
            if (tokensL[i].type !== tokensR[i].type ||
                tokensL[i].value !== tokensR[i].value) {
                return false;
            }
        }
        return true;
    }
    /**
     * Check if the given node is a true logical expression or not.
     *
     * The three binary expressions logical-or (`||`), logical-and (`&&`), and
     * coalesce (`??`) are known as `ShortCircuitExpression`.
     * But ESTree represents those by `LogicalExpression` node.
     *
     * This function rejects coalesce expressions of `LogicalExpression` node.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is `&&` or `||`.
     * @see https://tc39.es/ecma262/#prod-ShortCircuitExpression
     */
    function isLogicalExpression(node) {
        return (node.type === "LogicalExpression" &&
            (node.operator === "&&" || node.operator === "||"));
    }
    /**
     * Check if the given node is a nullish coalescing expression or not.
     *
     * The three binary expressions logical-or (`||`), logical-and (`&&`), and
     * coalesce (`??`) are known as `ShortCircuitExpression`.
     * But ESTree represents those by `LogicalExpression` node.
     *
     * This function finds only coalesce expressions of `LogicalExpression` node.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is `??`.
     */
    function isCoalesceExpression(node) {
        return node.type === "LogicalExpression" && node.operator === "??";
    }
    /**
     * Check if given two nodes are the pair of a logical expression and a coalesce expression.
     * @param {ASTNode} left A node to check.
     * @param {ASTNode} right Another node to check.
     * @returns {boolean} `true` if the two nodes are the pair of a logical expression and a coalesce expression.
     */
    function isMixedLogicalAndCoalesceExpressions(left, right) {
        return ((isLogicalExpression(left) && isCoalesceExpression(right)) ||
            (isCoalesceExpression(left) && isLogicalExpression(right)));
    }
    /**
     * Checks if the given operator is a logical assignment operator.
     * @param {string} operator The operator to check.
     * @returns {boolean} `true` if the operator is a logical assignment operator.
     */
    function isLogicalAssignmentOperator(operator) {
        return LOGICAL_ASSIGNMENT_OPERATORS.has(operator);
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = {
        COMMENTS_IGNORE_PATTERN,
        LINEBREAKS,
        LINEBREAK_MATCHER: lineBreakPattern,
        SHEBANG_MATCHER: shebangPattern,
        STATEMENT_LIST_PARENTS,
        /**
         * Determines whether two adjacent tokens are on the same line.
         * @param {Object} left The left token object.
         * @param {Object} right The right token object.
         * @returns {boolean} Whether or not the tokens are on the same line.
         * @public
         */
        isTokenOnSameLine(left, right) {
            return left.loc.end.line === right.loc.start.line;
        },
        isNullOrUndefined,
        isCallee,
        isES5Constructor,
        getUpperFunction,
        isFunction,
        isLoop,
        isInLoop,
        isArrayFromMethod,
        isParenthesised,
        createGlobalLinebreakMatcher,
        equalTokens,
        isArrowToken,
        isClosingBraceToken,
        isClosingBracketToken,
        isClosingParenToken,
        isColonToken,
        isCommaToken,
        isCommentToken,
        isDotToken,
        isQuestionDotToken,
        isKeywordToken,
        isNotClosingBraceToken: negate(isClosingBraceToken),
        isNotClosingBracketToken: negate(isClosingBracketToken),
        isNotClosingParenToken: negate(isClosingParenToken),
        isNotColonToken: negate(isColonToken),
        isNotCommaToken: negate(isCommaToken),
        isNotDotToken: negate(isDotToken),
        isNotQuestionDotToken: negate(isQuestionDotToken),
        isNotOpeningBraceToken: negate(isOpeningBraceToken),
        isNotOpeningBracketToken: negate(isOpeningBracketToken),
        isNotOpeningParenToken: negate(isOpeningParenToken),
        isNotSemicolonToken: negate(isSemicolonToken),
        isOpeningBraceToken,
        isOpeningBracketToken,
        isOpeningParenToken,
        isSemicolonToken,
        /**
         * Checks whether or not a given node is a string literal.
         * @param {ASTNode} node A node to check.
         * @returns {boolean} `true` if the node is a string literal.
         */
        isStringLiteral(node) {
            return ((node.type === "Literal" && typeof node.value === "string") ||
                node.type === "TemplateLiteral");
        },
        /**
         * Checks whether a given node is a breakable statement or not.
         * The node is breakable if the node is one of the following type:
         *
         * - DoWhileStatement
         * - ForInStatement
         * - ForOfStatement
         * - ForStatement
         * - SwitchStatement
         * - WhileStatement
         * @param {ASTNode} node A node to check.
         * @returns {boolean} `true` if the node is breakable.
         */
        isBreakableStatement(node) {
            return breakableTypePattern.test(node.type);
        },
        /**
         * Gets references which are non initializer and writable.
         * @param {Reference[]} references An array of references.
         * @returns {Reference[]} An array of only references which are non initializer and writable.
         * @public
         */
        getModifyingReferences(references) {
            return references.filter(isModifyingReference);
        },
        /**
         * Validate that a string passed in is surrounded by the specified character
         * @param  {string} val The text to check.
         * @param  {string} character The character to see if it's surrounded by.
         * @returns {boolean} True if the text is surrounded by the character, false if not.
         * @private
         */
        isSurroundedBy(val, character) {
            return val[0] === character && val[val.length - 1] === character;
        },
        /**
         * Returns whether the provided node is an ESLint directive comment or not
         * @param {Line|Block} node The comment token to be checked
         * @returns {boolean} `true` if the node is an ESLint directive comment
         */
        isDirectiveComment(node) {
            const comment = node.value.trim();
            return (node.type === "Line" && comment.indexOf("eslint-") === 0 ||
                node.type === "Block" && (comment.indexOf("global ") === 0 ||
                    comment.indexOf("eslint ") === 0 ||
                    comment.indexOf("eslint-") === 0));
        },
        /**
         * Gets the trailing statement of a given node.
         *
         *     if (code)
         *         consequent;
         *
         * When taking this `IfStatement`, returns `consequent;` statement.
         * @param {ASTNode} A node to get.
         * @returns {ASTNode|null} The trailing statement's node.
         */
        getTrailingStatement: esutils.ast.trailingStatement,
        /**
         * Finds the variable by a given name in a given scope and its upper scopes.
         * @param {eslint-scope.Scope} initScope A scope to start find.
         * @param {string} name A variable name to find.
         * @returns {eslint-scope.Variable|null} A found variable or `null`.
         */
        getVariableByName(initScope, name) {
            let scope = initScope;
            while (scope) {
                const variable = scope.set.get(name);
                if (variable) {
                    return variable;
                }
                scope = scope.upper;
            }
            return null;
        },
        /**
         * Checks whether or not a given function node is the default `this` binding.
         *
         * First, this checks the node:
         *
         * - The function name does not start with uppercase. It's a convention to capitalize the names
         *   of constructor functions. This check is not performed if `capIsConstructor` is set to `false`.
         * - The function does not have a JSDoc comment that has a @this tag.
         *
         * Next, this checks the location of the node.
         * If the location is below, this judges `this` is valid.
         *
         * - The location is not on an object literal.
         * - The location is not assigned to a variable which starts with an uppercase letter. Applies to anonymous
         *   functions only, as the name of the variable is considered to be the name of the function in this case.
         *   This check is not performed if `capIsConstructor` is set to `false`.
         * - The location is not on an ES2015 class.
         * - Its `bind`/`call`/`apply` method is not called directly.
         * - The function is not a callback of array methods (such as `.forEach()`) if `thisArg` is given.
         * @param {ASTNode} node A function node to check.
         * @param {SourceCode} sourceCode A SourceCode instance to get comments.
         * @param {boolean} [capIsConstructor = true] `false` disables the assumption that functions which name starts
         * with an uppercase or are assigned to a variable which name starts with an uppercase are constructors.
         * @returns {boolean} The function node is the default `this` binding.
         */
        isDefaultThisBinding(node, sourceCode, { capIsConstructor = true } = {}) {
            if ((capIsConstructor && isES5Constructor(node)) ||
                hasJSDocThisTag(node, sourceCode)) {
                return false;
            }
            const isAnonymous = node.id === null;
            let currentNode = node;
            while (currentNode) {
                const parent = currentNode.parent;
                switch (parent.type) {
                    /*
                     * Looks up the destination.
                     * e.g., obj.foo = nativeFoo || function foo() { ... };
                     */
                    case "LogicalExpression":
                    case "ConditionalExpression":
                    case "ChainExpression":
                        currentNode = parent;
                        break;
                    /*
                     * If the upper function is IIFE, checks the destination of the return value.
                     * e.g.
                     *   obj.foo = (function() {
                     *     // setup...
                     *     return function foo() { ... };
                     *   })();
                     *   obj.foo = (() =>
                     *     function foo() { ... }
                     *   )();
                     */
                    case "ReturnStatement": {
                        const func = getUpperFunction(parent);
                        if (func === null || !isCallee(func)) {
                            return true;
                        }
                        currentNode = func.parent;
                        break;
                    }
                    case "ArrowFunctionExpression":
                        if (currentNode !== parent.body || !isCallee(parent)) {
                            return true;
                        }
                        currentNode = parent.parent;
                        break;
                    /*
                     * e.g.
                     *   var obj = { foo() { ... } };
                     *   var obj = { foo: function() { ... } };
                     *   class A { constructor() { ... } }
                     *   class A { foo() { ... } }
                     *   class A { get foo() { ... } }
                     *   class A { set foo() { ... } }
                     *   class A { static foo() { ... } }
                     */
                    case "Property":
                    case "MethodDefinition":
                        return parent.value !== currentNode;
                    /*
                     * e.g.
                     *   obj.foo = function foo() { ... };
                     *   Foo = function() { ... };
                     *   [obj.foo = function foo() { ... }] = a;
                     *   [Foo = function() { ... }] = a;
                     */
                    case "AssignmentExpression":
                    case "AssignmentPattern":
                        if (parent.left.type === "MemberExpression") {
                            return false;
                        }
                        if (capIsConstructor &&
                            isAnonymous &&
                            parent.left.type === "Identifier" &&
                            startsWithUpperCase(parent.left.name)) {
                            return false;
                        }
                        return true;
                    /*
                     * e.g.
                     *   var Foo = function() { ... };
                     */
                    case "VariableDeclarator":
                        return !(capIsConstructor &&
                            isAnonymous &&
                            parent.init === currentNode &&
                            parent.id.type === "Identifier" &&
                            startsWithUpperCase(parent.id.name));
                    /*
                     * e.g.
                     *   var foo = function foo() { ... }.bind(obj);
                     *   (function foo() { ... }).call(obj);
                     *   (function foo() { ... }).apply(obj, []);
                     */
                    case "MemberExpression":
                        if (parent.object === currentNode &&
                            isSpecificMemberAccess(parent, null, bindOrCallOrApplyPattern)) {
                            const maybeCalleeNode = parent.parent.type === "ChainExpression"
                                ? parent.parent
                                : parent;
                            return !(isCallee(maybeCalleeNode) &&
                                maybeCalleeNode.parent.arguments.length >= 1 &&
                                !isNullOrUndefined(maybeCalleeNode.parent.arguments[0]));
                        }
                        return true;
                    /*
                     * e.g.
                     *   Reflect.apply(function() {}, obj, []);
                     *   Array.from([], function() {}, obj);
                     *   list.forEach(function() {}, obj);
                     */
                    case "CallExpression":
                        if (isReflectApply(parent.callee)) {
                            return (parent.arguments.length !== 3 ||
                                parent.arguments[0] !== currentNode ||
                                isNullOrUndefined(parent.arguments[1]));
                        }
                        if (isArrayFromMethod(parent.callee)) {
                            return (parent.arguments.length !== 3 ||
                                parent.arguments[1] !== currentNode ||
                                isNullOrUndefined(parent.arguments[2]));
                        }
                        if (isMethodWhichHasThisArg(parent.callee)) {
                            return (parent.arguments.length !== 2 ||
                                parent.arguments[0] !== currentNode ||
                                isNullOrUndefined(parent.arguments[1]));
                        }
                        return true;
                    // Otherwise `this` is default.
                    default:
                        return true;
                }
            }
            /* istanbul ignore next */
            return true;
        },
        /**
         * Get the precedence level based on the node type
         * @param {ASTNode} node node to evaluate
         * @returns {int} precedence level
         * @private
         */
        getPrecedence(node) {
            switch (node.type) {
                case "SequenceExpression":
                    return 0;
                case "AssignmentExpression":
                case "ArrowFunctionExpression":
                case "YieldExpression":
                    return 1;
                case "ConditionalExpression":
                    return 3;
                case "LogicalExpression":
                    switch (node.operator) {
                        case "||":
                        case "??":
                            return 4;
                        case "&&":
                            return 5;
                        // no default
                    }
                /* falls through */
                case "BinaryExpression":
                    switch (node.operator) {
                        case "|":
                            return 6;
                        case "^":
                            return 7;
                        case "&":
                            return 8;
                        case "==":
                        case "!=":
                        case "===":
                        case "!==":
                            return 9;
                        case "<":
                        case "<=":
                        case ">":
                        case ">=":
                        case "in":
                        case "instanceof":
                            return 10;
                        case "<<":
                        case ">>":
                        case ">>>":
                            return 11;
                        case "+":
                        case "-":
                            return 12;
                        case "*":
                        case "/":
                        case "%":
                            return 13;
                        case "**":
                            return 15;
                        // no default
                    }
                /* falls through */
                case "UnaryExpression":
                case "AwaitExpression":
                    return 16;
                case "UpdateExpression":
                    return 17;
                case "CallExpression":
                case "ChainExpression":
                case "ImportExpression":
                    return 18;
                case "NewExpression":
                    return 19;
                default:
                    return 20;
            }
        },
        /**
         * Checks whether the given node is an empty block node or not.
         * @param {ASTNode|null} node The node to check.
         * @returns {boolean} `true` if the node is an empty block.
         */
        isEmptyBlock(node) {
            return Boolean(node && node.type === "BlockStatement" && node.body.length === 0);
        },
        /**
         * Checks whether the given node is an empty function node or not.
         * @param {ASTNode|null} node The node to check.
         * @returns {boolean} `true` if the node is an empty function.
         */
        isEmptyFunction(node) {
            return isFunction(node) && module.exports.isEmptyBlock(node.body);
        },
        /**
         * Get directives from directive prologue of a Program or Function node.
         * @param {ASTNode} node The node to check.
         * @returns {ASTNode[]} The directives found in the directive prologue.
         */
        getDirectivePrologue(node) {
            const directives = [];
            // Directive prologues only occur at the top of files or functions.
            if (node.type === "Program" ||
                node.type === "FunctionDeclaration" ||
                node.type === "FunctionExpression" ||
                /*
                 * Do not check arrow functions with implicit return.
                 * `() => "use strict";` returns the string `"use strict"`.
                 */
                (node.type === "ArrowFunctionExpression" && node.body.type === "BlockStatement")) {
                const statements = node.type === "Program" ? node.body : node.body.body;
                for (const statement of statements) {
                    if (statement.type === "ExpressionStatement" &&
                        statement.expression.type === "Literal") {
                        directives.push(statement);
                    }
                    else {
                        break;
                    }
                }
            }
            return directives;
        },
        /**
         * Determines whether this node is a decimal integer literal. If a node is a decimal integer literal, a dot added
         * after the node will be parsed as a decimal point, rather than a property-access dot.
         * @param {ASTNode} node The node to check.
         * @returns {boolean} `true` if this node is a decimal integer.
         * @example
         *
         * 0         // true
         * 5         // true
         * 50        // true
         * 5_000     // true
         * 1_234_56  // true
         * 08        // true
         * 0192      // true
         * 5.        // false
         * .5        // false
         * 5.0       // false
         * 5.00_00   // false
         * 05        // false
         * 0x5       // false
         * 0b101     // false
         * 0b11_01   // false
         * 0o5       // false
         * 5e0       // false
         * 5e1_000   // false
         * 5n        // false
         * 1_000n    // false
         * '5'       // false
         */
        isDecimalInteger(node) {
            return node.type === "Literal" && typeof node.value === "number" &&
                DECIMAL_INTEGER_PATTERN.test(node.raw);
        },
        /**
         * Determines whether this token is a decimal integer numeric token.
         * This is similar to isDecimalInteger(), but for tokens.
         * @param {Token} token The token to check.
         * @returns {boolean} `true` if this token is a decimal integer.
         */
        isDecimalIntegerNumericToken(token) {
            return token.type === "Numeric" && DECIMAL_INTEGER_PATTERN.test(token.value);
        },
        /**
         * Gets the name and kind of the given function node.
         *
         * - `function foo() {}`  .................... `function 'foo'`
         * - `(function foo() {})`  .................. `function 'foo'`
         * - `(function() {})`  ...................... `function`
         * - `function* foo() {}`  ................... `generator function 'foo'`
         * - `(function* foo() {})`  ................. `generator function 'foo'`
         * - `(function*() {})`  ..................... `generator function`
         * - `() => {}`  ............................. `arrow function`
         * - `async () => {}`  ....................... `async arrow function`
         * - `({ foo: function foo() {} })`  ......... `method 'foo'`
         * - `({ foo: function() {} })`  ............. `method 'foo'`
         * - `({ ['foo']: function() {} })`  ......... `method 'foo'`
         * - `({ [foo]: function() {} })`  ........... `method`
         * - `({ foo() {} })`  ....................... `method 'foo'`
         * - `({ foo: function* foo() {} })`  ........ `generator method 'foo'`
         * - `({ foo: function*() {} })`  ............ `generator method 'foo'`
         * - `({ ['foo']: function*() {} })`  ........ `generator method 'foo'`
         * - `({ [foo]: function*() {} })`  .......... `generator method`
         * - `({ *foo() {} })`  ...................... `generator method 'foo'`
         * - `({ foo: async function foo() {} })`  ... `async method 'foo'`
         * - `({ foo: async function() {} })`  ....... `async method 'foo'`
         * - `({ ['foo']: async function() {} })`  ... `async method 'foo'`
         * - `({ [foo]: async function() {} })`  ..... `async method`
         * - `({ async foo() {} })`  ................. `async method 'foo'`
         * - `({ get foo() {} })`  ................... `getter 'foo'`
         * - `({ set foo(a) {} })`  .................. `setter 'foo'`
         * - `class A { constructor() {} }`  ......... `constructor`
         * - `class A { foo() {} }`  ................. `method 'foo'`
         * - `class A { *foo() {} }`  ................ `generator method 'foo'`
         * - `class A { async foo() {} }`  ........... `async method 'foo'`
         * - `class A { ['foo']() {} }`  ............. `method 'foo'`
         * - `class A { *['foo']() {} }`  ............ `generator method 'foo'`
         * - `class A { async ['foo']() {} }`  ....... `async method 'foo'`
         * - `class A { [foo]() {} }`  ............... `method`
         * - `class A { *[foo]() {} }`  .............. `generator method`
         * - `class A { async [foo]() {} }`  ......... `async method`
         * - `class A { get foo() {} }`  ............. `getter 'foo'`
         * - `class A { set foo(a) {} }`  ............ `setter 'foo'`
         * - `class A { static foo() {} }`  .......... `static method 'foo'`
         * - `class A { static *foo() {} }`  ......... `static generator method 'foo'`
         * - `class A { static async foo() {} }`  .... `static async method 'foo'`
         * - `class A { static get foo() {} }`  ...... `static getter 'foo'`
         * - `class A { static set foo(a) {} }`  ..... `static setter 'foo'`
         * @param {ASTNode} node The function node to get.
         * @returns {string} The name and kind of the function node.
         */
        getFunctionNameWithKind(node) {
            const parent = node.parent;
            const tokens = [];
            if (parent.type === "MethodDefinition" && parent.static) {
                tokens.push("static");
            }
            if (node.async) {
                tokens.push("async");
            }
            if (node.generator) {
                tokens.push("generator");
            }
            if (node.type === "ArrowFunctionExpression") {
                tokens.push("arrow", "function");
            }
            else if (parent.type === "Property" || parent.type === "MethodDefinition") {
                if (parent.kind === "constructor") {
                    return "constructor";
                }
                if (parent.kind === "get") {
                    tokens.push("getter");
                }
                else if (parent.kind === "set") {
                    tokens.push("setter");
                }
                else {
                    tokens.push("method");
                }
            }
            else {
                tokens.push("function");
            }
            if (node.id) {
                tokens.push(`'${node.id.name}'`);
            }
            else {
                const name = getStaticPropertyName(parent);
                if (name !== null) {
                    tokens.push(`'${name}'`);
                }
            }
            return tokens.join(" ");
        },
        /**
         * Gets the location of the given function node for reporting.
         *
         * - `function foo() {}`
         *    ^^^^^^^^^^^^
         * - `(function foo() {})`
         *     ^^^^^^^^^^^^
         * - `(function() {})`
         *     ^^^^^^^^
         * - `function* foo() {}`
         *    ^^^^^^^^^^^^^
         * - `(function* foo() {})`
         *     ^^^^^^^^^^^^^
         * - `(function*() {})`
         *     ^^^^^^^^^
         * - `() => {}`
         *       ^^
         * - `async () => {}`
         *             ^^
         * - `({ foo: function foo() {} })`
         *       ^^^^^^^^^^^^^^^^^
         * - `({ foo: function() {} })`
         *       ^^^^^^^^^^^^^
         * - `({ ['foo']: function() {} })`
         *       ^^^^^^^^^^^^^^^^^
         * - `({ [foo]: function() {} })`
         *       ^^^^^^^^^^^^^^^
         * - `({ foo() {} })`
         *       ^^^
         * - `({ foo: function* foo() {} })`
         *       ^^^^^^^^^^^^^^^^^^
         * - `({ foo: function*() {} })`
         *       ^^^^^^^^^^^^^^
         * - `({ ['foo']: function*() {} })`
         *       ^^^^^^^^^^^^^^^^^^
         * - `({ [foo]: function*() {} })`
         *       ^^^^^^^^^^^^^^^^
         * - `({ *foo() {} })`
         *       ^^^^
         * - `({ foo: async function foo() {} })`
         *       ^^^^^^^^^^^^^^^^^^^^^^^
         * - `({ foo: async function() {} })`
         *       ^^^^^^^^^^^^^^^^^^^
         * - `({ ['foo']: async function() {} })`
         *       ^^^^^^^^^^^^^^^^^^^^^^^
         * - `({ [foo]: async function() {} })`
         *       ^^^^^^^^^^^^^^^^^^^^^
         * - `({ async foo() {} })`
         *       ^^^^^^^^^
         * - `({ get foo() {} })`
         *       ^^^^^^^
         * - `({ set foo(a) {} })`
         *       ^^^^^^^
         * - `class A { constructor() {} }`
         *              ^^^^^^^^^^^
         * - `class A { foo() {} }`
         *              ^^^
         * - `class A { *foo() {} }`
         *              ^^^^
         * - `class A { async foo() {} }`
         *              ^^^^^^^^^
         * - `class A { ['foo']() {} }`
         *              ^^^^^^^
         * - `class A { *['foo']() {} }`
         *              ^^^^^^^^
         * - `class A { async ['foo']() {} }`
         *              ^^^^^^^^^^^^^
         * - `class A { [foo]() {} }`
         *              ^^^^^
         * - `class A { *[foo]() {} }`
         *              ^^^^^^
         * - `class A { async [foo]() {} }`
         *              ^^^^^^^^^^^
         * - `class A { get foo() {} }`
         *              ^^^^^^^
         * - `class A { set foo(a) {} }`
         *              ^^^^^^^
         * - `class A { static foo() {} }`
         *              ^^^^^^^^^^
         * - `class A { static *foo() {} }`
         *              ^^^^^^^^^^^
         * - `class A { static async foo() {} }`
         *              ^^^^^^^^^^^^^^^^
         * - `class A { static get foo() {} }`
         *              ^^^^^^^^^^^^^^
         * - `class A { static set foo(a) {} }`
         *              ^^^^^^^^^^^^^^
         * @param {ASTNode} node The function node to get.
         * @param {SourceCode} sourceCode The source code object to get tokens.
         * @returns {string} The location of the function node for reporting.
         */
        getFunctionHeadLoc(node, sourceCode) {
            const parent = node.parent;
            let start = null;
            let end = null;
            if (node.type === "ArrowFunctionExpression") {
                const arrowToken = sourceCode.getTokenBefore(node.body, isArrowToken);
                start = arrowToken.loc.start;
                end = arrowToken.loc.end;
            }
            else if (parent.type === "Property" || parent.type === "MethodDefinition") {
                start = parent.loc.start;
                end = getOpeningParenOfParams(node, sourceCode).loc.start;
            }
            else {
                start = node.loc.start;
                end = getOpeningParenOfParams(node, sourceCode).loc.start;
            }
            return {
                start: Object.assign({}, start),
                end: Object.assign({}, end)
            };
        },
        /**
         * Gets next location when the result is not out of bound, otherwise returns null.
         *
         * Assumptions:
         *
         * - The given location represents a valid location in the given source code.
         * - Columns are 0-based.
         * - Lines are 1-based.
         * - Column immediately after the last character in a line (not incl. linebreaks) is considered to be a valid location.
         * - If the source code ends with a linebreak, `sourceCode.lines` array will have an extra element (empty string) at the end.
         *   The start (column 0) of that extra line is considered to be a valid location.
         *
         * Examples of successive locations (line, column):
         *
         * code: foo
         * locations: (1, 0) -> (1, 1) -> (1, 2) -> (1, 3) -> null
         *
         * code: foo<LF>
         * locations: (1, 0) -> (1, 1) -> (1, 2) -> (1, 3) -> (2, 0) -> null
         *
         * code: foo<CR><LF>
         * locations: (1, 0) -> (1, 1) -> (1, 2) -> (1, 3) -> (2, 0) -> null
         *
         * code: a<LF>b
         * locations: (1, 0) -> (1, 1) -> (2, 0) -> (2, 1) -> null
         *
         * code: a<LF>b<LF>
         * locations: (1, 0) -> (1, 1) -> (2, 0) -> (2, 1) -> (3, 0) -> null
         *
         * code: a<CR><LF>b<CR><LF>
         * locations: (1, 0) -> (1, 1) -> (2, 0) -> (2, 1) -> (3, 0) -> null
         *
         * code: a<LF><LF>
         * locations: (1, 0) -> (1, 1) -> (2, 0) -> (3, 0) -> null
         *
         * code: <LF>
         * locations: (1, 0) -> (2, 0) -> null
         *
         * code:
         * locations: (1, 0) -> null
         * @param {SourceCode} sourceCode The sourceCode
         * @param {{line: number, column: number}} location The location
         * @returns {{line: number, column: number} | null} Next location
         */
        getNextLocation(sourceCode, { line, column }) {
            if (column < sourceCode.lines[line - 1].length) {
                return {
                    line,
                    column: column + 1
                };
            }
            if (line < sourceCode.lines.length) {
                return {
                    line: line + 1,
                    column: 0
                };
            }
            return null;
        },
        /**
         * Gets the parenthesized text of a node. This is similar to sourceCode.getText(node), but it also includes any parentheses
         * surrounding the node.
         * @param {SourceCode} sourceCode The source code object
         * @param {ASTNode} node An expression node
         * @returns {string} The text representing the node, with all surrounding parentheses included
         */
        getParenthesisedText(sourceCode, node) {
            let leftToken = sourceCode.getFirstToken(node);
            let rightToken = sourceCode.getLastToken(node);
            while (sourceCode.getTokenBefore(leftToken) &&
                sourceCode.getTokenBefore(leftToken).type === "Punctuator" &&
                sourceCode.getTokenBefore(leftToken).value === "(" &&
                sourceCode.getTokenAfter(rightToken) &&
                sourceCode.getTokenAfter(rightToken).type === "Punctuator" &&
                sourceCode.getTokenAfter(rightToken).value === ")") {
                leftToken = sourceCode.getTokenBefore(leftToken);
                rightToken = sourceCode.getTokenAfter(rightToken);
            }
            return sourceCode.getText().slice(leftToken.range[0], rightToken.range[1]);
        },
        /*
         * Determine if a node has a possibility to be an Error object
         * @param  {ASTNode} node  ASTNode to check
         * @returns {boolean} True if there is a chance it contains an Error obj
         */
        couldBeError(node) {
            switch (node.type) {
                case "Identifier":
                case "CallExpression":
                case "NewExpression":
                case "MemberExpression":
                case "TaggedTemplateExpression":
                case "YieldExpression":
                case "AwaitExpression":
                case "ChainExpression":
                    return true; // possibly an error object.
                case "AssignmentExpression":
                    if (["=", "&&="].includes(node.operator)) {
                        return module.exports.couldBeError(node.right);
                    }
                    if (["||=", "??="].includes(node.operator)) {
                        return module.exports.couldBeError(node.left) || module.exports.couldBeError(node.right);
                    }
                    /**
                     * All other assignment operators are mathematical assignment operators (arithmetic or bitwise).
                     * An assignment expression with a mathematical operator can either evaluate to a primitive value,
                     * or throw, depending on the operands. Thus, it cannot evaluate to an `Error` object.
                     */
                    return false;
                case "SequenceExpression": {
                    const exprs = node.expressions;
                    return exprs.length !== 0 && module.exports.couldBeError(exprs[exprs.length - 1]);
                }
                case "LogicalExpression":
                    /*
                     * If the && operator short-circuits, the left side was falsy and therefore not an error, and if it
                     * doesn't short-circuit, it takes the value from the right side, so the right side must always be
                     * a plausible error. A future improvement could verify that the left side could be truthy by
                     * excluding falsy literals.
                     */
                    if (node.operator === "&&") {
                        return module.exports.couldBeError(node.right);
                    }
                    return module.exports.couldBeError(node.left) || module.exports.couldBeError(node.right);
                case "ConditionalExpression":
                    return module.exports.couldBeError(node.consequent) || module.exports.couldBeError(node.alternate);
                default:
                    return false;
            }
        },
        /**
         * Check if a given node is a numeric literal or not.
         * @param {ASTNode} node The node to check.
         * @returns {boolean} `true` if the node is a number or bigint literal.
         */
        isNumericLiteral(node) {
            return (node.type === "Literal" &&
                (typeof node.value === "number" || Boolean(node.bigint)));
        },
        /**
         * Determines whether two tokens can safely be placed next to each other without merging into a single token
         * @param {Token|string} leftValue The left token. If this is a string, it will be tokenized and the last token will be used.
         * @param {Token|string} rightValue The right token. If this is a string, it will be tokenized and the first token will be used.
         * @returns {boolean} If the tokens cannot be safely placed next to each other, returns `false`. If the tokens can be placed
         * next to each other, behavior is undefined (although it should return `true` in most cases).
         */
        canTokensBeAdjacent(leftValue, rightValue) {
            const espreeOptions = {
                ecmaVersion: espree.latestEcmaVersion,
                comment: true,
                range: true
            };
            let leftToken;
            if (typeof leftValue === "string") {
                let tokens;
                try {
                    tokens = espree.tokenize(leftValue, espreeOptions);
                }
                catch {
                    return false;
                }
                const comments = tokens.comments;
                leftToken = tokens[tokens.length - 1];
                if (comments.length) {
                    const lastComment = comments[comments.length - 1];
                    if (lastComment.range[0] > leftToken.range[0]) {
                        leftToken = lastComment;
                    }
                }
            }
            else {
                leftToken = leftValue;
            }
            if (leftToken.type === "Shebang") {
                return false;
            }
            let rightToken;
            if (typeof rightValue === "string") {
                let tokens;
                try {
                    tokens = espree.tokenize(rightValue, espreeOptions);
                }
                catch {
                    return false;
                }
                const comments = tokens.comments;
                rightToken = tokens[0];
                if (comments.length) {
                    const firstComment = comments[0];
                    if (firstComment.range[0] < rightToken.range[0]) {
                        rightToken = firstComment;
                    }
                }
            }
            else {
                rightToken = rightValue;
            }
            if (leftToken.type === "Punctuator" || rightToken.type === "Punctuator") {
                if (leftToken.type === "Punctuator" && rightToken.type === "Punctuator") {
                    const PLUS_TOKENS = new Set(["+", "++"]);
                    const MINUS_TOKENS = new Set(["-", "--"]);
                    return !(PLUS_TOKENS.has(leftToken.value) && PLUS_TOKENS.has(rightToken.value) ||
                        MINUS_TOKENS.has(leftToken.value) && MINUS_TOKENS.has(rightToken.value));
                }
                if (leftToken.type === "Punctuator" && leftToken.value === "/") {
                    return !["Block", "Line", "RegularExpression"].includes(rightToken.type);
                }
                return true;
            }
            if (leftToken.type === "String" || rightToken.type === "String" ||
                leftToken.type === "Template" || rightToken.type === "Template") {
                return true;
            }
            if (leftToken.type !== "Numeric" && rightToken.type === "Numeric" && rightToken.value.startsWith(".")) {
                return true;
            }
            if (leftToken.type === "Block" || rightToken.type === "Block" || rightToken.type === "Line") {
                return true;
            }
            return false;
        },
        /**
         * Get the `loc` object of a given name in a `/*globals` directive comment.
         * @param {SourceCode} sourceCode The source code to convert index to loc.
         * @param {Comment} comment The `/*globals` directive comment which include the name.
         * @param {string} name The name to find.
         * @returns {SourceLocation} The `loc` object.
         */
        getNameLocationInGlobalDirectiveComment(sourceCode, comment, name) {
            const namePattern = new RegExp(`[\\s,]${escapeRegExp(name)}(?:$|[\\s,:])`, "gu");
            // To ignore the first text "global".
            namePattern.lastIndex = comment.value.indexOf("global") + 6;
            // Search a given variable name.
            const match = namePattern.exec(comment.value);
            // Convert the index to loc.
            const start = sourceCode.getLocFromIndex(comment.range[0] +
                "/*".length +
                (match ? match.index + 1 : 0));
            const end = {
                line: start.line,
                column: start.column + (match ? name.length : 1)
            };
            return { start, end };
        },
        /**
         * Determines whether the given raw string contains an octal escape sequence
         * or a non-octal decimal escape sequence ("\8", "\9").
         *
         * "\1", "\2" ... "\7", "\8", "\9"
         * "\00", "\01" ... "\07", "\08", "\09"
         *
         * "\0", when not followed by a digit, is not an octal escape sequence.
         * @param {string} rawString A string in its raw representation.
         * @returns {boolean} `true` if the string contains at least one octal escape sequence
         * or at least one non-octal decimal escape sequence.
         */
        hasOctalOrNonOctalDecimalEscapeSequence(rawString) {
            return OCTAL_OR_NON_OCTAL_DECIMAL_ESCAPE_PATTERN.test(rawString);
        },
        isLogicalExpression,
        isCoalesceExpression,
        isMixedLogicalAndCoalesceExpressions,
        isNullLiteral,
        getStaticStringValue,
        getStaticPropertyName,
        skipChainExpression,
        isSpecificId,
        isSpecificMemberAccess,
        equalLiteralValue,
        isSameReference,
        isLogicalAssignmentOperator
    };
},
/* esutils/lib/utils.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
      Copyright (C) 2013 Yusuke Suzuki <utatane.tea@gmail.com>
    
      Redistribution and use in source and binary forms, with or without
      modification, are permitted provided that the following conditions are met:
    
        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
    
      THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
      AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
      IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
      ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
      DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
      (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
      LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
      ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
      (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
      THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    */
    (function () {
        'use strict';
        exports.ast = require(396) /* ./ast */;
        exports.code = require(397) /* ./code */;
        exports.keyword = require(398) /* ./keyword */;
    }());
},
/* esutils/lib/ast.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
      Copyright (C) 2013 Yusuke Suzuki <utatane.tea@gmail.com>
    
      Redistribution and use in source and binary forms, with or without
      modification, are permitted provided that the following conditions are met:
    
        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
    
      THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 'AS IS'
      AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
      IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
      ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
      DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
      (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
      LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
      ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
      (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
      THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    */
    (function () {
        'use strict';
        function isExpression(node) {
            if (node == null) {
                return false;
            }
            switch (node.type) {
                case 'ArrayExpression':
                case 'AssignmentExpression':
                case 'BinaryExpression':
                case 'CallExpression':
                case 'ConditionalExpression':
                case 'FunctionExpression':
                case 'Identifier':
                case 'Literal':
                case 'LogicalExpression':
                case 'MemberExpression':
                case 'NewExpression':
                case 'ObjectExpression':
                case 'SequenceExpression':
                case 'ThisExpression':
                case 'UnaryExpression':
                case 'UpdateExpression':
                    return true;
            }
            return false;
        }
        function isIterationStatement(node) {
            if (node == null) {
                return false;
            }
            switch (node.type) {
                case 'DoWhileStatement':
                case 'ForInStatement':
                case 'ForStatement':
                case 'WhileStatement':
                    return true;
            }
            return false;
        }
        function isStatement(node) {
            if (node == null) {
                return false;
            }
            switch (node.type) {
                case 'BlockStatement':
                case 'BreakStatement':
                case 'ContinueStatement':
                case 'DebuggerStatement':
                case 'DoWhileStatement':
                case 'EmptyStatement':
                case 'ExpressionStatement':
                case 'ForInStatement':
                case 'ForStatement':
                case 'IfStatement':
                case 'LabeledStatement':
                case 'ReturnStatement':
                case 'SwitchStatement':
                case 'ThrowStatement':
                case 'TryStatement':
                case 'VariableDeclaration':
                case 'WhileStatement':
                case 'WithStatement':
                    return true;
            }
            return false;
        }
        function isSourceElement(node) {
            return isStatement(node) || node != null && node.type === 'FunctionDeclaration';
        }
        function trailingStatement(node) {
            switch (node.type) {
                case 'IfStatement':
                    if (node.alternate != null) {
                        return node.alternate;
                    }
                    return node.consequent;
                case 'LabeledStatement':
                case 'ForStatement':
                case 'ForInStatement':
                case 'WhileStatement':
                case 'WithStatement':
                    return node.body;
            }
            return null;
        }
        function isProblematicIfStatement(node) {
            var current;
            if (node.type !== 'IfStatement') {
                return false;
            }
            if (node.alternate == null) {
                return false;
            }
            current = node.consequent;
            do {
                if (current.type === 'IfStatement') {
                    if (current.alternate == null) {
                        return true;
                    }
                }
                current = trailingStatement(current);
            } while (current);
            return false;
        }
        module.exports = {
            isExpression: isExpression,
            isStatement: isStatement,
            isIterationStatement: isIterationStatement,
            isSourceElement: isSourceElement,
            isProblematicIfStatement: isProblematicIfStatement,
            trailingStatement: trailingStatement
        };
    }());
},
/* esutils/lib/code.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
      Copyright (C) 2013-2014 Yusuke Suzuki <utatane.tea@gmail.com>
      Copyright (C) 2014 Ivan Nikulin <ifaaan@gmail.com>
    
      Redistribution and use in source and binary forms, with or without
      modification, are permitted provided that the following conditions are met:
    
        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
    
      THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
      AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
      IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
      ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
      DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
      (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
      LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
      ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
      (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
      THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    */
    (function () {
        'use strict';
        var ES6Regex, ES5Regex, NON_ASCII_WHITESPACES, IDENTIFIER_START, IDENTIFIER_PART, ch;
        // See `tools/generate-identifier-regex.js`.
        ES5Regex = {
            // ECMAScript 5.1/Unicode v9.0.0 NonAsciiIdentifierStart:
            NonAsciiIdentifierStart: /[\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u08A0-\u08B4\u08B6-\u08BD\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0AF9\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D\u0C58-\u0C5A\u0C60\u0C61\u0C80\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D54-\u0D56\u0D5F-\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u1884\u1887-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191E\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1C80-\u1C88\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2119-\u211D\u2124\u2126\u2128\u212A-\u212D\u212F-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2E2F\u3005-\u3007\u3021-\u3029\u3031-\u3035\u3038-\u303C\u3041-\u3096\u309D-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA69D\uA6A0-\uA6EF\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AE\uA7B0-\uA7B7\uA7F7-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA8FD\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uA9E0-\uA9E4\uA9E6-\uA9EF\uA9FA-\uA9FE\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA7E-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]/,
            // ECMAScript 5.1/Unicode v9.0.0 NonAsciiIdentifierPart:
            NonAsciiIdentifierPart: /[\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0300-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u0483-\u0487\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u0591-\u05BD\u05BF\u05C1\u05C2\u05C4\u05C5\u05C7\u05D0-\u05EA\u05F0-\u05F2\u0610-\u061A\u0620-\u0669\u066E-\u06D3\u06D5-\u06DC\u06DF-\u06E8\u06EA-\u06FC\u06FF\u0710-\u074A\u074D-\u07B1\u07C0-\u07F5\u07FA\u0800-\u082D\u0840-\u085B\u08A0-\u08B4\u08B6-\u08BD\u08D4-\u08E1\u08E3-\u0963\u0966-\u096F\u0971-\u0983\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BC-\u09C4\u09C7\u09C8\u09CB-\u09CE\u09D7\u09DC\u09DD\u09DF-\u09E3\u09E6-\u09F1\u0A01-\u0A03\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A3C\u0A3E-\u0A42\u0A47\u0A48\u0A4B-\u0A4D\u0A51\u0A59-\u0A5C\u0A5E\u0A66-\u0A75\u0A81-\u0A83\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABC-\u0AC5\u0AC7-\u0AC9\u0ACB-\u0ACD\u0AD0\u0AE0-\u0AE3\u0AE6-\u0AEF\u0AF9\u0B01-\u0B03\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3C-\u0B44\u0B47\u0B48\u0B4B-\u0B4D\u0B56\u0B57\u0B5C\u0B5D\u0B5F-\u0B63\u0B66-\u0B6F\u0B71\u0B82\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BBE-\u0BC2\u0BC6-\u0BC8\u0BCA-\u0BCD\u0BD0\u0BD7\u0BE6-\u0BEF\u0C00-\u0C03\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D-\u0C44\u0C46-\u0C48\u0C4A-\u0C4D\u0C55\u0C56\u0C58-\u0C5A\u0C60-\u0C63\u0C66-\u0C6F\u0C80-\u0C83\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBC-\u0CC4\u0CC6-\u0CC8\u0CCA-\u0CCD\u0CD5\u0CD6\u0CDE\u0CE0-\u0CE3\u0CE6-\u0CEF\u0CF1\u0CF2\u0D01-\u0D03\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D-\u0D44\u0D46-\u0D48\u0D4A-\u0D4E\u0D54-\u0D57\u0D5F-\u0D63\u0D66-\u0D6F\u0D7A-\u0D7F\u0D82\u0D83\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0DCA\u0DCF-\u0DD4\u0DD6\u0DD8-\u0DDF\u0DE6-\u0DEF\u0DF2\u0DF3\u0E01-\u0E3A\u0E40-\u0E4E\u0E50-\u0E59\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB9\u0EBB-\u0EBD\u0EC0-\u0EC4\u0EC6\u0EC8-\u0ECD\u0ED0-\u0ED9\u0EDC-\u0EDF\u0F00\u0F18\u0F19\u0F20-\u0F29\u0F35\u0F37\u0F39\u0F3E-\u0F47\u0F49-\u0F6C\u0F71-\u0F84\u0F86-\u0F97\u0F99-\u0FBC\u0FC6\u1000-\u1049\u1050-\u109D\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u135D-\u135F\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1714\u1720-\u1734\u1740-\u1753\u1760-\u176C\u176E-\u1770\u1772\u1773\u1780-\u17D3\u17D7\u17DC\u17DD\u17E0-\u17E9\u180B-\u180D\u1810-\u1819\u1820-\u1877\u1880-\u18AA\u18B0-\u18F5\u1900-\u191E\u1920-\u192B\u1930-\u193B\u1946-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u19D0-\u19D9\u1A00-\u1A1B\u1A20-\u1A5E\u1A60-\u1A7C\u1A7F-\u1A89\u1A90-\u1A99\u1AA7\u1AB0-\u1ABD\u1B00-\u1B4B\u1B50-\u1B59\u1B6B-\u1B73\u1B80-\u1BF3\u1C00-\u1C37\u1C40-\u1C49\u1C4D-\u1C7D\u1C80-\u1C88\u1CD0-\u1CD2\u1CD4-\u1CF6\u1CF8\u1CF9\u1D00-\u1DF5\u1DFB-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u200C\u200D\u203F\u2040\u2054\u2071\u207F\u2090-\u209C\u20D0-\u20DC\u20E1\u20E5-\u20F0\u2102\u2107\u210A-\u2113\u2115\u2119-\u211D\u2124\u2126\u2128\u212A-\u212D\u212F-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D7F-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2DE0-\u2DFF\u2E2F\u3005-\u3007\u3021-\u302F\u3031-\u3035\u3038-\u303C\u3041-\u3096\u3099\u309A\u309D-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA62B\uA640-\uA66F\uA674-\uA67D\uA67F-\uA6F1\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AE\uA7B0-\uA7B7\uA7F7-\uA827\uA840-\uA873\uA880-\uA8C5\uA8D0-\uA8D9\uA8E0-\uA8F7\uA8FB\uA8FD\uA900-\uA92D\uA930-\uA953\uA960-\uA97C\uA980-\uA9C0\uA9CF-\uA9D9\uA9E0-\uA9FE\uAA00-\uAA36\uAA40-\uAA4D\uAA50-\uAA59\uAA60-\uAA76\uAA7A-\uAAC2\uAADB-\uAADD\uAAE0-\uAAEF\uAAF2-\uAAF6\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABEA\uABEC\uABED\uABF0-\uABF9\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE00-\uFE0F\uFE20-\uFE2F\uFE33\uFE34\uFE4D-\uFE4F\uFE70-\uFE74\uFE76-\uFEFC\uFF10-\uFF19\uFF21-\uFF3A\uFF3F\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]/
        };
        ES6Regex = {
            // ECMAScript 6/Unicode v9.0.0 NonAsciiIdentifierStart:
            NonAsciiIdentifierStart: /[\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u08A0-\u08B4\u08B6-\u08BD\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0AF9\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D\u0C58-\u0C5A\u0C60\u0C61\u0C80\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D54-\u0D56\u0D5F-\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191E\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1C80-\u1C88\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2118-\u211D\u2124\u2126\u2128\u212A-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u3005-\u3007\u3021-\u3029\u3031-\u3035\u3038-\u303C\u3041-\u3096\u309B-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA69D\uA6A0-\uA6EF\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AE\uA7B0-\uA7B7\uA7F7-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA8FD\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uA9E0-\uA9E4\uA9E6-\uA9EF\uA9FA-\uA9FE\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA7E-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDD40-\uDD74\uDE80-\uDE9C\uDEA0-\uDED0\uDF00-\uDF1F\uDF30-\uDF4A\uDF50-\uDF75\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF\uDFD1-\uDFD5]|\uD801[\uDC00-\uDC9D\uDCB0-\uDCD3\uDCD8-\uDCFB\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00\uDE10-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE4\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC03-\uDC37\uDC83-\uDCAF\uDCD0-\uDCE8\uDD03-\uDD26\uDD50-\uDD72\uDD76\uDD83-\uDDB2\uDDC1-\uDDC4\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE2B\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEDE\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3D\uDF50\uDF5D-\uDF61]|\uD805[\uDC00-\uDC34\uDC47-\uDC4A\uDC80-\uDCAF\uDCC4\uDCC5\uDCC7\uDD80-\uDDAE\uDDD8-\uDDDB\uDE00-\uDE2F\uDE44\uDE80-\uDEAA\uDF00-\uDF19]|\uD806[\uDCA0-\uDCDF\uDCFF\uDEC0-\uDEF8]|\uD807[\uDC00-\uDC08\uDC0A-\uDC2E\uDC40\uDC72-\uDC8F]|\uD808[\uDC00-\uDF99]|\uD809[\uDC00-\uDC6E\uDC80-\uDD43]|[\uD80C\uD81C-\uD820\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDED0-\uDEED\uDF00-\uDF2F\uDF40-\uDF43\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50\uDF93-\uDF9F\uDFE0]|\uD821[\uDC00-\uDFEC]|\uD822[\uDC00-\uDEF2]|\uD82C[\uDC00\uDC01]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB]|\uD83A[\uDC00-\uDCC4\uDD00-\uDD43]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1]|\uD87E[\uDC00-\uDE1D]/,
            // ECMAScript 6/Unicode v9.0.0 NonAsciiIdentifierPart:
            NonAsciiIdentifierPart: /[\xAA\xB5\xB7\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0300-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u0483-\u0487\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u0591-\u05BD\u05BF\u05C1\u05C2\u05C4\u05C5\u05C7\u05D0-\u05EA\u05F0-\u05F2\u0610-\u061A\u0620-\u0669\u066E-\u06D3\u06D5-\u06DC\u06DF-\u06E8\u06EA-\u06FC\u06FF\u0710-\u074A\u074D-\u07B1\u07C0-\u07F5\u07FA\u0800-\u082D\u0840-\u085B\u08A0-\u08B4\u08B6-\u08BD\u08D4-\u08E1\u08E3-\u0963\u0966-\u096F\u0971-\u0983\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BC-\u09C4\u09C7\u09C8\u09CB-\u09CE\u09D7\u09DC\u09DD\u09DF-\u09E3\u09E6-\u09F1\u0A01-\u0A03\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A3C\u0A3E-\u0A42\u0A47\u0A48\u0A4B-\u0A4D\u0A51\u0A59-\u0A5C\u0A5E\u0A66-\u0A75\u0A81-\u0A83\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABC-\u0AC5\u0AC7-\u0AC9\u0ACB-\u0ACD\u0AD0\u0AE0-\u0AE3\u0AE6-\u0AEF\u0AF9\u0B01-\u0B03\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3C-\u0B44\u0B47\u0B48\u0B4B-\u0B4D\u0B56\u0B57\u0B5C\u0B5D\u0B5F-\u0B63\u0B66-\u0B6F\u0B71\u0B82\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BBE-\u0BC2\u0BC6-\u0BC8\u0BCA-\u0BCD\u0BD0\u0BD7\u0BE6-\u0BEF\u0C00-\u0C03\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D-\u0C44\u0C46-\u0C48\u0C4A-\u0C4D\u0C55\u0C56\u0C58-\u0C5A\u0C60-\u0C63\u0C66-\u0C6F\u0C80-\u0C83\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBC-\u0CC4\u0CC6-\u0CC8\u0CCA-\u0CCD\u0CD5\u0CD6\u0CDE\u0CE0-\u0CE3\u0CE6-\u0CEF\u0CF1\u0CF2\u0D01-\u0D03\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D-\u0D44\u0D46-\u0D48\u0D4A-\u0D4E\u0D54-\u0D57\u0D5F-\u0D63\u0D66-\u0D6F\u0D7A-\u0D7F\u0D82\u0D83\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0DCA\u0DCF-\u0DD4\u0DD6\u0DD8-\u0DDF\u0DE6-\u0DEF\u0DF2\u0DF3\u0E01-\u0E3A\u0E40-\u0E4E\u0E50-\u0E59\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB9\u0EBB-\u0EBD\u0EC0-\u0EC4\u0EC6\u0EC8-\u0ECD\u0ED0-\u0ED9\u0EDC-\u0EDF\u0F00\u0F18\u0F19\u0F20-\u0F29\u0F35\u0F37\u0F39\u0F3E-\u0F47\u0F49-\u0F6C\u0F71-\u0F84\u0F86-\u0F97\u0F99-\u0FBC\u0FC6\u1000-\u1049\u1050-\u109D\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u135D-\u135F\u1369-\u1371\u1380-\u138F\u13A0-\u13F5\u13F8-\u13FD\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16EE-\u16F8\u1700-\u170C\u170E-\u1714\u1720-\u1734\u1740-\u1753\u1760-\u176C\u176E-\u1770\u1772\u1773\u1780-\u17D3\u17D7\u17DC\u17DD\u17E0-\u17E9\u180B-\u180D\u1810-\u1819\u1820-\u1877\u1880-\u18AA\u18B0-\u18F5\u1900-\u191E\u1920-\u192B\u1930-\u193B\u1946-\u196D\u1970-\u1974\u1980-\u19AB\u19B0-\u19C9\u19D0-\u19DA\u1A00-\u1A1B\u1A20-\u1A5E\u1A60-\u1A7C\u1A7F-\u1A89\u1A90-\u1A99\u1AA7\u1AB0-\u1ABD\u1B00-\u1B4B\u1B50-\u1B59\u1B6B-\u1B73\u1B80-\u1BF3\u1C00-\u1C37\u1C40-\u1C49\u1C4D-\u1C7D\u1C80-\u1C88\u1CD0-\u1CD2\u1CD4-\u1CF6\u1CF8\u1CF9\u1D00-\u1DF5\u1DFB-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u200C\u200D\u203F\u2040\u2054\u2071\u207F\u2090-\u209C\u20D0-\u20DC\u20E1\u20E5-\u20F0\u2102\u2107\u210A-\u2113\u2115\u2118-\u211D\u2124\u2126\u2128\u212A-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2160-\u2188\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D7F-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2DE0-\u2DFF\u3005-\u3007\u3021-\u302F\u3031-\u3035\u3038-\u303C\u3041-\u3096\u3099-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FD5\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA62B\uA640-\uA66F\uA674-\uA67D\uA67F-\uA6F1\uA717-\uA71F\uA722-\uA788\uA78B-\uA7AE\uA7B0-\uA7B7\uA7F7-\uA827\uA840-\uA873\uA880-\uA8C5\uA8D0-\uA8D9\uA8E0-\uA8F7\uA8FB\uA8FD\uA900-\uA92D\uA930-\uA953\uA960-\uA97C\uA980-\uA9C0\uA9CF-\uA9D9\uA9E0-\uA9FE\uAA00-\uAA36\uAA40-\uAA4D\uAA50-\uAA59\uAA60-\uAA76\uAA7A-\uAAC2\uAADB-\uAADD\uAAE0-\uAAEF\uAAF2-\uAAF6\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB65\uAB70-\uABEA\uABEC\uABED\uABF0-\uABF9\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE00-\uFE0F\uFE20-\uFE2F\uFE33\uFE34\uFE4D-\uFE4F\uFE70-\uFE74\uFE76-\uFEFC\uFF10-\uFF19\uFF21-\uFF3A\uFF3F\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDD40-\uDD74\uDDFD\uDE80-\uDE9C\uDEA0-\uDED0\uDEE0\uDF00-\uDF1F\uDF30-\uDF4A\uDF50-\uDF7A\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF\uDFD1-\uDFD5]|\uD801[\uDC00-\uDC9D\uDCA0-\uDCA9\uDCB0-\uDCD3\uDCD8-\uDCFB\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDCE0-\uDCF2\uDCF4\uDCF5\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00-\uDE03\uDE05\uDE06\uDE0C-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE38-\uDE3A\uDE3F\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE6\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48\uDC80-\uDCB2\uDCC0-\uDCF2]|\uD804[\uDC00-\uDC46\uDC66-\uDC6F\uDC7F-\uDCBA\uDCD0-\uDCE8\uDCF0-\uDCF9\uDD00-\uDD34\uDD36-\uDD3F\uDD50-\uDD73\uDD76\uDD80-\uDDC4\uDDCA-\uDDCC\uDDD0-\uDDDA\uDDDC\uDE00-\uDE11\uDE13-\uDE37\uDE3E\uDE80-\uDE86\uDE88\uDE8A-\uDE8D\uDE8F-\uDE9D\uDE9F-\uDEA8\uDEB0-\uDEEA\uDEF0-\uDEF9\uDF00-\uDF03\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3C-\uDF44\uDF47\uDF48\uDF4B-\uDF4D\uDF50\uDF57\uDF5D-\uDF63\uDF66-\uDF6C\uDF70-\uDF74]|\uD805[\uDC00-\uDC4A\uDC50-\uDC59\uDC80-\uDCC5\uDCC7\uDCD0-\uDCD9\uDD80-\uDDB5\uDDB8-\uDDC0\uDDD8-\uDDDD\uDE00-\uDE40\uDE44\uDE50-\uDE59\uDE80-\uDEB7\uDEC0-\uDEC9\uDF00-\uDF19\uDF1D-\uDF2B\uDF30-\uDF39]|\uD806[\uDCA0-\uDCE9\uDCFF\uDEC0-\uDEF8]|\uD807[\uDC00-\uDC08\uDC0A-\uDC36\uDC38-\uDC40\uDC50-\uDC59\uDC72-\uDC8F\uDC92-\uDCA7\uDCA9-\uDCB6]|\uD808[\uDC00-\uDF99]|\uD809[\uDC00-\uDC6E\uDC80-\uDD43]|[\uD80C\uD81C-\uD820\uD840-\uD868\uD86A-\uD86C\uD86F-\uD872][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD811[\uDC00-\uDE46]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDE60-\uDE69\uDED0-\uDEED\uDEF0-\uDEF4\uDF00-\uDF36\uDF40-\uDF43\uDF50-\uDF59\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50-\uDF7E\uDF8F-\uDF9F\uDFE0]|\uD821[\uDC00-\uDFEC]|\uD822[\uDC00-\uDEF2]|\uD82C[\uDC00\uDC01]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99\uDC9D\uDC9E]|\uD834[\uDD65-\uDD69\uDD6D-\uDD72\uDD7B-\uDD82\uDD85-\uDD8B\uDDAA-\uDDAD\uDE42-\uDE44]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB\uDFCE-\uDFFF]|\uD836[\uDE00-\uDE36\uDE3B-\uDE6C\uDE75\uDE84\uDE9B-\uDE9F\uDEA1-\uDEAF]|\uD838[\uDC00-\uDC06\uDC08-\uDC18\uDC1B-\uDC21\uDC23\uDC24\uDC26-\uDC2A]|\uD83A[\uDC00-\uDCC4\uDCD0-\uDCD6\uDD00-\uDD4A\uDD50-\uDD59]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D\uDC20-\uDFFF]|\uD873[\uDC00-\uDEA1]|\uD87E[\uDC00-\uDE1D]|\uDB40[\uDD00-\uDDEF]/
        };
        function isDecimalDigit(ch) {
            return 0x30 <= ch && ch <= 0x39; // 0..9
        }
        function isHexDigit(ch) {
            return 0x30 <= ch && ch <= 0x39 || // 0..9
                0x61 <= ch && ch <= 0x66 || // a..f
                0x41 <= ch && ch <= 0x46; // A..F
        }
        function isOctalDigit(ch) {
            return ch >= 0x30 && ch <= 0x37; // 0..7
        }
        // 7.2 White Space
        NON_ASCII_WHITESPACES = [
            0x1680,
            0x2000, 0x2001, 0x2002, 0x2003, 0x2004, 0x2005, 0x2006, 0x2007, 0x2008, 0x2009, 0x200A,
            0x202F, 0x205F,
            0x3000,
            0xFEFF
        ];
        function isWhiteSpace(ch) {
            return ch === 0x20 || ch === 0x09 || ch === 0x0B || ch === 0x0C || ch === 0xA0 ||
                ch >= 0x1680 && NON_ASCII_WHITESPACES.indexOf(ch) >= 0;
        }
        // 7.3 Line Terminators
        function isLineTerminator(ch) {
            return ch === 0x0A || ch === 0x0D || ch === 0x2028 || ch === 0x2029;
        }
        // 7.6 Identifier Names and Identifiers
        function fromCodePoint(cp) {
            if (cp <= 0xFFFF) {
                return String.fromCharCode(cp);
            }
            var cu1 = String.fromCharCode(Math.floor((cp - 0x10000) / 0x400) + 0xD800);
            var cu2 = String.fromCharCode(((cp - 0x10000) % 0x400) + 0xDC00);
            return cu1 + cu2;
        }
        IDENTIFIER_START = new Array(0x80);
        for (ch = 0; ch < 0x80; ++ch) {
            IDENTIFIER_START[ch] =
                ch >= 0x61 && ch <= 0x7A || // a..z
                    ch >= 0x41 && ch <= 0x5A || // A..Z
                    ch === 0x24 || ch === 0x5F; // $ (dollar) and _ (underscore)
        }
        IDENTIFIER_PART = new Array(0x80);
        for (ch = 0; ch < 0x80; ++ch) {
            IDENTIFIER_PART[ch] =
                ch >= 0x61 && ch <= 0x7A || // a..z
                    ch >= 0x41 && ch <= 0x5A || // A..Z
                    ch >= 0x30 && ch <= 0x39 || // 0..9
                    ch === 0x24 || ch === 0x5F; // $ (dollar) and _ (underscore)
        }
        function isIdentifierStartES5(ch) {
            return ch < 0x80 ? IDENTIFIER_START[ch] : ES5Regex.NonAsciiIdentifierStart.test(fromCodePoint(ch));
        }
        function isIdentifierPartES5(ch) {
            return ch < 0x80 ? IDENTIFIER_PART[ch] : ES5Regex.NonAsciiIdentifierPart.test(fromCodePoint(ch));
        }
        function isIdentifierStartES6(ch) {
            return ch < 0x80 ? IDENTIFIER_START[ch] : ES6Regex.NonAsciiIdentifierStart.test(fromCodePoint(ch));
        }
        function isIdentifierPartES6(ch) {
            return ch < 0x80 ? IDENTIFIER_PART[ch] : ES6Regex.NonAsciiIdentifierPart.test(fromCodePoint(ch));
        }
        module.exports = {
            isDecimalDigit: isDecimalDigit,
            isHexDigit: isHexDigit,
            isOctalDigit: isOctalDigit,
            isWhiteSpace: isWhiteSpace,
            isLineTerminator: isLineTerminator,
            isIdentifierStartES5: isIdentifierStartES5,
            isIdentifierPartES5: isIdentifierPartES5,
            isIdentifierStartES6: isIdentifierStartES6,
            isIdentifierPartES6: isIdentifierPartES6
        };
    }());
},
/* esutils/lib/keyword.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
      Copyright (C) 2013 Yusuke Suzuki <utatane.tea@gmail.com>
    
      Redistribution and use in source and binary forms, with or without
      modification, are permitted provided that the following conditions are met:
    
        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
    
      THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
      AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
      IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
      ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
      DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
      (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
      LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
      ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
      (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
      THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    */
    (function () {
        'use strict';
        var code = require(397) /* ./code */;
        function isStrictModeReservedWordES6(id) {
            switch (id) {
                case 'implements':
                case 'interface':
                case 'package':
                case 'private':
                case 'protected':
                case 'public':
                case 'static':
                case 'let':
                    return true;
                default:
                    return false;
            }
        }
        function isKeywordES5(id, strict) {
            // yield should not be treated as keyword under non-strict mode.
            if (!strict && id === 'yield') {
                return false;
            }
            return isKeywordES6(id, strict);
        }
        function isKeywordES6(id, strict) {
            if (strict && isStrictModeReservedWordES6(id)) {
                return true;
            }
            switch (id.length) {
                case 2:
                    return (id === 'if') || (id === 'in') || (id === 'do');
                case 3:
                    return (id === 'var') || (id === 'for') || (id === 'new') || (id === 'try');
                case 4:
                    return (id === 'this') || (id === 'else') || (id === 'case') ||
                        (id === 'void') || (id === 'with') || (id === 'enum');
                case 5:
                    return (id === 'while') || (id === 'break') || (id === 'catch') ||
                        (id === 'throw') || (id === 'const') || (id === 'yield') ||
                        (id === 'class') || (id === 'super');
                case 6:
                    return (id === 'return') || (id === 'typeof') || (id === 'delete') ||
                        (id === 'switch') || (id === 'export') || (id === 'import');
                case 7:
                    return (id === 'default') || (id === 'finally') || (id === 'extends');
                case 8:
                    return (id === 'function') || (id === 'continue') || (id === 'debugger');
                case 10:
                    return (id === 'instanceof');
                default:
                    return false;
            }
        }
        function isReservedWordES5(id, strict) {
            return id === 'null' || id === 'true' || id === 'false' || isKeywordES5(id, strict);
        }
        function isReservedWordES6(id, strict) {
            return id === 'null' || id === 'true' || id === 'false' || isKeywordES6(id, strict);
        }
        function isRestrictedWord(id) {
            return id === 'eval' || id === 'arguments';
        }
        function isIdentifierNameES5(id) {
            var i, iz, ch;
            if (id.length === 0) {
                return false;
            }
            ch = id.charCodeAt(0);
            if (!code.isIdentifierStartES5(ch)) {
                return false;
            }
            for (i = 1, iz = id.length; i < iz; ++i) {
                ch = id.charCodeAt(i);
                if (!code.isIdentifierPartES5(ch)) {
                    return false;
                }
            }
            return true;
        }
        function decodeUtf16(lead, trail) {
            return (lead - 0xD800) * 0x400 + (trail - 0xDC00) + 0x10000;
        }
        function isIdentifierNameES6(id) {
            var i, iz, ch, lowCh, check;
            if (id.length === 0) {
                return false;
            }
            check = code.isIdentifierStartES6;
            for (i = 0, iz = id.length; i < iz; ++i) {
                ch = id.charCodeAt(i);
                if (0xD800 <= ch && ch <= 0xDBFF) {
                    ++i;
                    if (i >= iz) {
                        return false;
                    }
                    lowCh = id.charCodeAt(i);
                    if (!(0xDC00 <= lowCh && lowCh <= 0xDFFF)) {
                        return false;
                    }
                    ch = decodeUtf16(ch, lowCh);
                }
                if (!check(ch)) {
                    return false;
                }
                check = code.isIdentifierPartES6;
            }
            return true;
        }
        function isIdentifierES5(id, strict) {
            return isIdentifierNameES5(id) && !isReservedWordES5(id, strict);
        }
        function isIdentifierES6(id, strict) {
            return isIdentifierNameES6(id) && !isReservedWordES6(id, strict);
        }
        module.exports = {
            isKeywordES5: isKeywordES5,
            isKeywordES6: isKeywordES6,
            isReservedWordES5: isReservedWordES5,
            isReservedWordES6: isReservedWordES6,
            isRestrictedWord: isRestrictedWord,
            isIdentifierNameES5: isIdentifierNameES5,
            isIdentifierNameES6: isIdentifierNameES6,
            isIdentifierES5: isIdentifierES5,
            isIdentifierES6: isIdentifierES6
        };
    }());
},
/* eslint/lib/rules/array-bracket-newline.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce linebreaks after opening and before closing array brackets",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/array-bracket-newline"
            },
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["always", "never", "consistent"]
                        },
                        {
                            type: "object",
                            properties: {
                                multiline: {
                                    type: "boolean"
                                },
                                minItems: {
                                    type: ["integer", "null"],
                                    minimum: 0
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                unexpectedOpeningLinebreak: "There should be no linebreak after '['.",
                unexpectedClosingLinebreak: "There should be no linebreak before ']'.",
                missingOpeningLinebreak: "A linebreak is required after '['.",
                missingClosingLinebreak: "A linebreak is required before ']'."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            //----------------------------------------------------------------------
            // Helpers
            //----------------------------------------------------------------------
            /**
             * Normalizes a given option value.
             * @param {string|Object|undefined} option An option value to parse.
             * @returns {{multiline: boolean, minItems: number}} Normalized option object.
             */
            function normalizeOptionValue(option) {
                let consistent = false;
                let multiline = false;
                let minItems = 0;
                if (option) {
                    if (option === "consistent") {
                        consistent = true;
                        minItems = Number.POSITIVE_INFINITY;
                    }
                    else if (option === "always" || option.minItems === 0) {
                        minItems = 0;
                    }
                    else if (option === "never") {
                        minItems = Number.POSITIVE_INFINITY;
                    }
                    else {
                        multiline = Boolean(option.multiline);
                        minItems = option.minItems || Number.POSITIVE_INFINITY;
                    }
                }
                else {
                    consistent = false;
                    multiline = true;
                    minItems = Number.POSITIVE_INFINITY;
                }
                return { consistent, multiline, minItems };
            }
            /**
             * Normalizes a given option value.
             * @param {string|Object|undefined} options An option value to parse.
             * @returns {{ArrayExpression: {multiline: boolean, minItems: number}, ArrayPattern: {multiline: boolean, minItems: number}}} Normalized option object.
             */
            function normalizeOptions(options) {
                const value = normalizeOptionValue(options);
                return { ArrayExpression: value, ArrayPattern: value };
            }
            /**
             * Reports that there shouldn't be a linebreak after the first token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportNoBeginningLinebreak(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "unexpectedOpeningLinebreak",
                    fix(fixer) {
                        const nextToken = sourceCode.getTokenAfter(token, { includeComments: true });
                        if (astUtils.isCommentToken(nextToken)) {
                            return null;
                        }
                        return fixer.removeRange([token.range[1], nextToken.range[0]]);
                    }
                });
            }
            /**
             * Reports that there shouldn't be a linebreak before the last token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportNoEndingLinebreak(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "unexpectedClosingLinebreak",
                    fix(fixer) {
                        const previousToken = sourceCode.getTokenBefore(token, { includeComments: true });
                        if (astUtils.isCommentToken(previousToken)) {
                            return null;
                        }
                        return fixer.removeRange([previousToken.range[1], token.range[0]]);
                    }
                });
            }
            /**
             * Reports that there should be a linebreak after the first token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportRequiredBeginningLinebreak(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "missingOpeningLinebreak",
                    fix(fixer) {
                        return fixer.insertTextAfter(token, "\n");
                    }
                });
            }
            /**
             * Reports that there should be a linebreak before the last token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportRequiredEndingLinebreak(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "missingClosingLinebreak",
                    fix(fixer) {
                        return fixer.insertTextBefore(token, "\n");
                    }
                });
            }
            /**
             * Reports a given node if it violated this rule.
             * @param {ASTNode} node A node to check. This is an ArrayExpression node or an ArrayPattern node.
             * @returns {void}
             */
            function check(node) {
                const elements = node.elements;
                const normalizedOptions = normalizeOptions(context.options[0]);
                const options = normalizedOptions[node.type];
                const openBracket = sourceCode.getFirstToken(node);
                const closeBracket = sourceCode.getLastToken(node);
                const firstIncComment = sourceCode.getTokenAfter(openBracket, { includeComments: true });
                const lastIncComment = sourceCode.getTokenBefore(closeBracket, { includeComments: true });
                const first = sourceCode.getTokenAfter(openBracket);
                const last = sourceCode.getTokenBefore(closeBracket);
                const needsLinebreaks = (elements.length >= options.minItems ||
                    (options.multiline &&
                        elements.length > 0 &&
                        firstIncComment.loc.start.line !== lastIncComment.loc.end.line) ||
                    (elements.length === 0 &&
                        firstIncComment.type === "Block" &&
                        firstIncComment.loc.start.line !== lastIncComment.loc.end.line &&
                        firstIncComment === lastIncComment) ||
                    (options.consistent &&
                        openBracket.loc.end.line !== first.loc.start.line));
                /*
                 * Use tokens or comments to check multiline or not.
                 * But use only tokens to check whether linebreaks are needed.
                 * This allows:
                 *     var arr = [ // eslint-disable-line foo
                 *         'a'
                 *     ]
                 */
                if (needsLinebreaks) {
                    if (astUtils.isTokenOnSameLine(openBracket, first)) {
                        reportRequiredBeginningLinebreak(node, openBracket);
                    }
                    if (astUtils.isTokenOnSameLine(last, closeBracket)) {
                        reportRequiredEndingLinebreak(node, closeBracket);
                    }
                }
                else {
                    if (!astUtils.isTokenOnSameLine(openBracket, first)) {
                        reportNoBeginningLinebreak(node, openBracket);
                    }
                    if (!astUtils.isTokenOnSameLine(last, closeBracket)) {
                        reportNoEndingLinebreak(node, closeBracket);
                    }
                }
            }
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                ArrayPattern: check,
                ArrayExpression: check
            };
        }
    };
},
/* eslint/lib/rules/array-bracket-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing inside array brackets",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/array-bracket-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never"]
                },
                {
                    type: "object",
                    properties: {
                        singleValue: {
                            type: "boolean"
                        },
                        objectsInArrays: {
                            type: "boolean"
                        },
                        arraysInArrays: {
                            type: "boolean"
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedSpaceAfter: "There should be no space after '{{tokenValue}}'.",
                unexpectedSpaceBefore: "There should be no space before '{{tokenValue}}'.",
                missingSpaceAfter: "A space is required after '{{tokenValue}}'.",
                missingSpaceBefore: "A space is required before '{{tokenValue}}'."
            }
        },
        create(context) {
            const spaced = context.options[0] === "always", sourceCode = context.getSourceCode();
            /**
             * Determines whether an option is set, relative to the spacing option.
             * If spaced is "always", then check whether option is set to false.
             * If spaced is "never", then check whether option is set to true.
             * @param {Object} option The option to exclude.
             * @returns {boolean} Whether or not the property is excluded.
             */
            function isOptionSet(option) {
                return context.options[1] ? context.options[1][option] === !spaced : false;
            }
            const options = {
                spaced,
                singleElementException: isOptionSet("singleValue"),
                objectsInArraysException: isOptionSet("objectsInArrays"),
                arraysInArraysException: isOptionSet("arraysInArrays")
            };
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Reports that there shouldn't be a space after the first token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportNoBeginningSpace(node, token) {
                const nextToken = sourceCode.getTokenAfter(token);
                context.report({
                    node,
                    loc: { start: token.loc.end, end: nextToken.loc.start },
                    messageId: "unexpectedSpaceAfter",
                    data: {
                        tokenValue: token.value
                    },
                    fix(fixer) {
                        return fixer.removeRange([token.range[1], nextToken.range[0]]);
                    }
                });
            }
            /**
             * Reports that there shouldn't be a space before the last token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportNoEndingSpace(node, token) {
                const previousToken = sourceCode.getTokenBefore(token);
                context.report({
                    node,
                    loc: { start: previousToken.loc.end, end: token.loc.start },
                    messageId: "unexpectedSpaceBefore",
                    data: {
                        tokenValue: token.value
                    },
                    fix(fixer) {
                        return fixer.removeRange([previousToken.range[1], token.range[0]]);
                    }
                });
            }
            /**
             * Reports that there should be a space after the first token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportRequiredBeginningSpace(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "missingSpaceAfter",
                    data: {
                        tokenValue: token.value
                    },
                    fix(fixer) {
                        return fixer.insertTextAfter(token, " ");
                    }
                });
            }
            /**
             * Reports that there should be a space before the last token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportRequiredEndingSpace(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "missingSpaceBefore",
                    data: {
                        tokenValue: token.value
                    },
                    fix(fixer) {
                        return fixer.insertTextBefore(token, " ");
                    }
                });
            }
            /**
             * Determines if a node is an object type
             * @param {ASTNode} node The node to check.
             * @returns {boolean} Whether or not the node is an object type.
             */
            function isObjectType(node) {
                return node && (node.type === "ObjectExpression" || node.type === "ObjectPattern");
            }
            /**
             * Determines if a node is an array type
             * @param {ASTNode} node The node to check.
             * @returns {boolean} Whether or not the node is an array type.
             */
            function isArrayType(node) {
                return node && (node.type === "ArrayExpression" || node.type === "ArrayPattern");
            }
            /**
             * Validates the spacing around array brackets
             * @param {ASTNode} node The node we're checking for spacing
             * @returns {void}
             */
            function validateArraySpacing(node) {
                if (options.spaced && node.elements.length === 0) {
                    return;
                }
                const first = sourceCode.getFirstToken(node), second = sourceCode.getFirstToken(node, 1), last = node.typeAnnotation
                    ? sourceCode.getTokenBefore(node.typeAnnotation)
                    : sourceCode.getLastToken(node), penultimate = sourceCode.getTokenBefore(last), firstElement = node.elements[0], lastElement = node.elements[node.elements.length - 1];
                const openingBracketMustBeSpaced = options.objectsInArraysException && isObjectType(firstElement) ||
                    options.arraysInArraysException && isArrayType(firstElement) ||
                    options.singleElementException && node.elements.length === 1
                    ? !options.spaced : options.spaced;
                const closingBracketMustBeSpaced = options.objectsInArraysException && isObjectType(lastElement) ||
                    options.arraysInArraysException && isArrayType(lastElement) ||
                    options.singleElementException && node.elements.length === 1
                    ? !options.spaced : options.spaced;
                if (astUtils.isTokenOnSameLine(first, second)) {
                    if (openingBracketMustBeSpaced && !sourceCode.isSpaceBetweenTokens(first, second)) {
                        reportRequiredBeginningSpace(node, first);
                    }
                    if (!openingBracketMustBeSpaced && sourceCode.isSpaceBetweenTokens(first, second)) {
                        reportNoBeginningSpace(node, first);
                    }
                }
                if (first !== penultimate && astUtils.isTokenOnSameLine(penultimate, last)) {
                    if (closingBracketMustBeSpaced && !sourceCode.isSpaceBetweenTokens(penultimate, last)) {
                        reportRequiredEndingSpace(node, last);
                    }
                    if (!closingBracketMustBeSpaced && sourceCode.isSpaceBetweenTokens(penultimate, last)) {
                        reportNoEndingSpace(node, last);
                    }
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                ArrayPattern: validateArraySpacing,
                ArrayExpression: validateArraySpacing
            };
        }
    };
},
/* eslint/lib/rules/array-callback-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const TARGET_NODE_TYPE = /^(?:Arrow)?FunctionExpression$/u;
    const TARGET_METHODS = /^(?:every|filter|find(?:Index)?|flatMap|forEach|map|reduce(?:Right)?|some|sort)$/u;
    /**
     * Checks a given code path segment is reachable.
     * @param {CodePathSegment} segment A segment to check.
     * @returns {boolean} `true` if the segment is reachable.
     */
    function isReachable(segment) {
        return segment.reachable;
    }
    /**
     * Checks a given node is a member access which has the specified name's
     * property.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node is a member access which has
     *      the specified name's property. The node may be a `(Chain|Member)Expression` node.
     */
    function isTargetMethod(node) {
        return astUtils.isSpecificMemberAccess(node, null, TARGET_METHODS);
    }
    /**
     * Returns a human-legible description of an array method
     * @param {string} arrayMethodName A method name to fully qualify
     * @returns {string} the method name prefixed with `Array.` if it is a class method,
     *      or else `Array.prototype.` if it is an instance method.
     */
    function fullMethodName(arrayMethodName) {
        if (["from", "of", "isArray"].includes(arrayMethodName)) {
            return "Array.".concat(arrayMethodName);
        }
        return "Array.prototype.".concat(arrayMethodName);
    }
    /**
     * Checks whether or not a given node is a function expression which is the
     * callback of an array method, returning the method name.
     * @param {ASTNode} node A node to check. This is one of
     *      FunctionExpression or ArrowFunctionExpression.
     * @returns {string} The method name if the node is a callback method,
     *      null otherwise.
     */
    function getArrayMethodName(node) {
        let currentNode = node;
        while (currentNode) {
            const parent = currentNode.parent;
            switch (parent.type) {
                /*
                 * Looks up the destination. e.g.,
                 * foo.every(nativeFoo || function foo() { ... });
                 */
                case "LogicalExpression":
                case "ConditionalExpression":
                case "ChainExpression":
                    currentNode = parent;
                    break;
                /*
                 * If the upper function is IIFE, checks the destination of the return value.
                 * e.g.
                 *   foo.every((function() {
                 *     // setup...
                 *     return function callback() { ... };
                 *   })());
                 */
                case "ReturnStatement": {
                    const func = astUtils.getUpperFunction(parent);
                    if (func === null || !astUtils.isCallee(func)) {
                        return null;
                    }
                    currentNode = func.parent;
                    break;
                }
                /*
                 * e.g.
                 *   Array.from([], function() {});
                 *   list.every(function() {});
                 */
                case "CallExpression":
                    if (astUtils.isArrayFromMethod(parent.callee)) {
                        if (parent.arguments.length >= 2 &&
                            parent.arguments[1] === currentNode) {
                            return "from";
                        }
                    }
                    if (isTargetMethod(parent.callee)) {
                        if (parent.arguments.length >= 1 &&
                            parent.arguments[0] === currentNode) {
                            return astUtils.getStaticPropertyName(parent.callee);
                        }
                    }
                    return null;
                // Otherwise this node is not target.
                default:
                    return null;
            }
        }
        /* istanbul ignore next: unreachable */
        return null;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "enforce `return` statements in callbacks of array methods",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/array-callback-return"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowImplicit: {
                            type: "boolean",
                            default: false
                        },
                        checkForEach: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                expectedAtEnd: "{{arrayMethodName}}() expects a value to be returned at the end of {{name}}.",
                expectedInside: "{{arrayMethodName}}() expects a return value from {{name}}.",
                expectedReturnValue: "{{arrayMethodName}}() expects a return value from {{name}}.",
                expectedNoReturnValue: "{{arrayMethodName}}() expects no useless return value from {{name}}."
            }
        },
        create(context) {
            const options = context.options[0] || { allowImplicit: false, checkForEach: false };
            const sourceCode = context.getSourceCode();
            let funcInfo = {
                arrayMethodName: null,
                upper: null,
                codePath: null,
                hasReturn: false,
                shouldCheck: false,
                node: null
            };
            /**
             * Checks whether or not the last code path segment is reachable.
             * Then reports this function if the segment is reachable.
             *
             * If the last code path segment is reachable, there are paths which are not
             * returned or thrown.
             * @param {ASTNode} node A node to check.
             * @returns {void}
             */
            function checkLastSegment(node) {
                if (!funcInfo.shouldCheck) {
                    return;
                }
                let messageId = null;
                if (funcInfo.arrayMethodName === "forEach") {
                    if (options.checkForEach && node.type === "ArrowFunctionExpression" && node.expression) {
                        messageId = "expectedNoReturnValue";
                    }
                }
                else {
                    if (node.body.type === "BlockStatement" && funcInfo.codePath.currentSegments.some(isReachable)) {
                        messageId = funcInfo.hasReturn ? "expectedAtEnd" : "expectedInside";
                    }
                }
                if (messageId) {
                    const name = astUtils.getFunctionNameWithKind(node);
                    context.report({
                        node,
                        loc: astUtils.getFunctionHeadLoc(node, sourceCode),
                        messageId,
                        data: { name, arrayMethodName: fullMethodName(funcInfo.arrayMethodName) }
                    });
                }
            }
            return {
                // Stacks this function's information.
                onCodePathStart(codePath, node) {
                    let methodName = null;
                    if (TARGET_NODE_TYPE.test(node.type)) {
                        methodName = getArrayMethodName(node);
                    }
                    funcInfo = {
                        arrayMethodName: methodName,
                        upper: funcInfo,
                        codePath,
                        hasReturn: false,
                        shouldCheck: methodName &&
                            !node.async &&
                            !node.generator,
                        node
                    };
                },
                // Pops this function's information.
                onCodePathEnd() {
                    funcInfo = funcInfo.upper;
                },
                // Checks the return statement is valid.
                ReturnStatement(node) {
                    if (!funcInfo.shouldCheck) {
                        return;
                    }
                    funcInfo.hasReturn = true;
                    let messageId = null;
                    if (funcInfo.arrayMethodName === "forEach") {
                        // if checkForEach: true, returning a value at any path inside a forEach is not allowed
                        if (options.checkForEach && node.argument) {
                            messageId = "expectedNoReturnValue";
                        }
                    }
                    else {
                        // if allowImplicit: false, should also check node.argument
                        if (!options.allowImplicit && !node.argument) {
                            messageId = "expectedReturnValue";
                        }
                    }
                    if (messageId) {
                        context.report({
                            node,
                            messageId,
                            data: {
                                name: astUtils.getFunctionNameWithKind(funcInfo.node),
                                arrayMethodName: fullMethodName(funcInfo.arrayMethodName)
                            }
                        });
                    }
                },
                // Reports a given function if the last path is reachable.
                "FunctionExpression:exit": checkLastSegment,
                "ArrowFunctionExpression:exit": checkLastSegment
            };
        }
    };
},
/* eslint/lib/rules/array-element-newline.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce line breaks after each array element",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/array-element-newline"
            },
            fixable: "whitespace",
            schema: {
                definitions: {
                    basicConfig: {
                        oneOf: [
                            {
                                enum: ["always", "never", "consistent"]
                            },
                            {
                                type: "object",
                                properties: {
                                    multiline: {
                                        type: "boolean"
                                    },
                                    minItems: {
                                        type: ["integer", "null"],
                                        minimum: 0
                                    }
                                },
                                additionalProperties: false
                            }
                        ]
                    }
                },
                items: [
                    {
                        oneOf: [
                            {
                                $ref: "#/definitions/basicConfig"
                            },
                            {
                                type: "object",
                                properties: {
                                    ArrayExpression: {
                                        $ref: "#/definitions/basicConfig"
                                    },
                                    ArrayPattern: {
                                        $ref: "#/definitions/basicConfig"
                                    }
                                },
                                additionalProperties: false,
                                minProperties: 1
                            }
                        ]
                    }
                ]
            },
            messages: {
                unexpectedLineBreak: "There should be no linebreak here.",
                missingLineBreak: "There should be a linebreak after this element."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            //----------------------------------------------------------------------
            // Helpers
            //----------------------------------------------------------------------
            /**
             * Normalizes a given option value.
             * @param {string|Object|undefined} providedOption An option value to parse.
             * @returns {{multiline: boolean, minItems: number}} Normalized option object.
             */
            function normalizeOptionValue(providedOption) {
                let consistent = false;
                let multiline = false;
                let minItems;
                const option = providedOption || "always";
                if (!option || option === "always" || option.minItems === 0) {
                    minItems = 0;
                }
                else if (option === "never") {
                    minItems = Number.POSITIVE_INFINITY;
                }
                else if (option === "consistent") {
                    consistent = true;
                    minItems = Number.POSITIVE_INFINITY;
                }
                else {
                    multiline = Boolean(option.multiline);
                    minItems = option.minItems || Number.POSITIVE_INFINITY;
                }
                return { consistent, multiline, minItems };
            }
            /**
             * Normalizes a given option value.
             * @param {string|Object|undefined} options An option value to parse.
             * @returns {{ArrayExpression: {multiline: boolean, minItems: number}, ArrayPattern: {multiline: boolean, minItems: number}}} Normalized option object.
             */
            function normalizeOptions(options) {
                if (options && (options.ArrayExpression || options.ArrayPattern)) {
                    let expressionOptions, patternOptions;
                    if (options.ArrayExpression) {
                        expressionOptions = normalizeOptionValue(options.ArrayExpression);
                    }
                    if (options.ArrayPattern) {
                        patternOptions = normalizeOptionValue(options.ArrayPattern);
                    }
                    return { ArrayExpression: expressionOptions, ArrayPattern: patternOptions };
                }
                const value = normalizeOptionValue(options);
                return { ArrayExpression: value, ArrayPattern: value };
            }
            /**
             * Reports that there shouldn't be a line break after the first token
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportNoLineBreak(token) {
                const tokenBefore = sourceCode.getTokenBefore(token, { includeComments: true });
                context.report({
                    loc: {
                        start: tokenBefore.loc.end,
                        end: token.loc.start
                    },
                    messageId: "unexpectedLineBreak",
                    fix(fixer) {
                        if (astUtils.isCommentToken(tokenBefore)) {
                            return null;
                        }
                        if (!astUtils.isTokenOnSameLine(tokenBefore, token)) {
                            return fixer.replaceTextRange([tokenBefore.range[1], token.range[0]], " ");
                        }
                        /*
                         * This will check if the comma is on the same line as the next element
                         * Following array:
                         * [
                         *     1
                         *     , 2
                         *     , 3
                         * ]
                         *
                         * will be fixed to:
                         * [
                         *     1, 2, 3
                         * ]
                         */
                        const twoTokensBefore = sourceCode.getTokenBefore(tokenBefore, { includeComments: true });
                        if (astUtils.isCommentToken(twoTokensBefore)) {
                            return null;
                        }
                        return fixer.replaceTextRange([twoTokensBefore.range[1], tokenBefore.range[0]], "");
                    }
                });
            }
            /**
             * Reports that there should be a line break after the first token
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportRequiredLineBreak(token) {
                const tokenBefore = sourceCode.getTokenBefore(token, { includeComments: true });
                context.report({
                    loc: {
                        start: tokenBefore.loc.end,
                        end: token.loc.start
                    },
                    messageId: "missingLineBreak",
                    fix(fixer) {
                        return fixer.replaceTextRange([tokenBefore.range[1], token.range[0]], "\n");
                    }
                });
            }
            /**
             * Reports a given node if it violated this rule.
             * @param {ASTNode} node A node to check. This is an ObjectExpression node or an ObjectPattern node.
             * @returns {void}
             */
            function check(node) {
                const elements = node.elements;
                const normalizedOptions = normalizeOptions(context.options[0]);
                const options = normalizedOptions[node.type];
                if (!options) {
                    return;
                }
                let elementBreak = false;
                /*
                 * MULTILINE: true
                 * loop through every element and check
                 * if at least one element has linebreaks inside
                 * this ensures that following is not valid (due to elements are on the same line):
                 *
                 * [
                 *      1,
                 *      2,
                 *      3
                 * ]
                 */
                if (options.multiline) {
                    elementBreak = elements
                        .filter(element => element !== null)
                        .some(element => element.loc.start.line !== element.loc.end.line);
                }
                const linebreaksCount = node.elements.map((element, i) => {
                    const previousElement = elements[i - 1];
                    if (i === 0 || element === null || previousElement === null) {
                        return false;
                    }
                    const commaToken = sourceCode.getFirstTokenBetween(previousElement, element, astUtils.isCommaToken);
                    const lastTokenOfPreviousElement = sourceCode.getTokenBefore(commaToken);
                    const firstTokenOfCurrentElement = sourceCode.getTokenAfter(commaToken);
                    return !astUtils.isTokenOnSameLine(lastTokenOfPreviousElement, firstTokenOfCurrentElement);
                }).filter(isBreak => isBreak === true).length;
                const needsLinebreaks = (elements.length >= options.minItems ||
                    (options.multiline &&
                        elementBreak) ||
                    (options.consistent &&
                        linebreaksCount > 0 &&
                        linebreaksCount < node.elements.length));
                elements.forEach((element, i) => {
                    const previousElement = elements[i - 1];
                    if (i === 0 || element === null || previousElement === null) {
                        return;
                    }
                    const commaToken = sourceCode.getFirstTokenBetween(previousElement, element, astUtils.isCommaToken);
                    const lastTokenOfPreviousElement = sourceCode.getTokenBefore(commaToken);
                    const firstTokenOfCurrentElement = sourceCode.getTokenAfter(commaToken);
                    if (needsLinebreaks) {
                        if (astUtils.isTokenOnSameLine(lastTokenOfPreviousElement, firstTokenOfCurrentElement)) {
                            reportRequiredLineBreak(firstTokenOfCurrentElement);
                        }
                    }
                    else {
                        if (!astUtils.isTokenOnSameLine(lastTokenOfPreviousElement, firstTokenOfCurrentElement)) {
                            reportNoLineBreak(firstTokenOfCurrentElement);
                        }
                    }
                });
            }
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                ArrayPattern: check,
                ArrayExpression: check
            };
        }
    };
},
/* eslint/lib/rules/arrow-body-style.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require braces around arrow function bodies",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/arrow-body-style"
            },
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always", "never"]
                            }
                        ],
                        minItems: 0,
                        maxItems: 1
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["as-needed"]
                            },
                            {
                                type: "object",
                                properties: {
                                    requireReturnForObjectLiteral: { type: "boolean" }
                                },
                                additionalProperties: false
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    }
                ]
            },
            fixable: "code",
            messages: {
                unexpectedOtherBlock: "Unexpected block statement surrounding arrow body.",
                unexpectedEmptyBlock: "Unexpected block statement surrounding arrow body; put a value of `undefined` immediately after the `=>`.",
                unexpectedObjectBlock: "Unexpected block statement surrounding arrow body; parenthesize the returned value and move it immediately after the `=>`.",
                unexpectedSingleBlock: "Unexpected block statement surrounding arrow body; move the returned value immediately after the `=>`.",
                expectedBlock: "Expected block statement surrounding arrow body."
            }
        },
        create(context) {
            const options = context.options;
            const always = options[0] === "always";
            const asNeeded = !options[0] || options[0] === "as-needed";
            const never = options[0] === "never";
            const requireReturnForObjectLiteral = options[1] && options[1].requireReturnForObjectLiteral;
            const sourceCode = context.getSourceCode();
            let funcInfo = null;
            /**
             * Checks whether the given node has ASI problem or not.
             * @param {Token} token The token to check.
             * @returns {boolean} `true` if it changes semantics if `;` or `}` followed by the token are removed.
             */
            function hasASIProblem(token) {
                return token && token.type === "Punctuator" && /^[([/`+-]/u.test(token.value);
            }
            /**
             * Gets the closing parenthesis by the given node.
             * @param {ASTNode} node first node after an opening parenthesis.
             * @returns {Token} The found closing parenthesis token.
             */
            function findClosingParen(node) {
                let nodeToCheck = node;
                while (!astUtils.isParenthesised(sourceCode, nodeToCheck)) {
                    nodeToCheck = nodeToCheck.parent;
                }
                return sourceCode.getTokenAfter(nodeToCheck);
            }
            /**
             * Check whether the node is inside of a for loop's init
             * @param {ASTNode} node node is inside for loop
             * @returns {boolean} `true` if the node is inside of a for loop, else `false`
             */
            function isInsideForLoopInitializer(node) {
                if (node && node.parent) {
                    if (node.parent.type === "ForStatement" && node.parent.init === node) {
                        return true;
                    }
                    return isInsideForLoopInitializer(node.parent);
                }
                return false;
            }
            /**
             * Determines whether a arrow function body needs braces
             * @param {ASTNode} node The arrow function node.
             * @returns {void}
             */
            function validate(node) {
                const arrowBody = node.body;
                if (arrowBody.type === "BlockStatement") {
                    const blockBody = arrowBody.body;
                    if (blockBody.length !== 1 && !never) {
                        return;
                    }
                    if (asNeeded && requireReturnForObjectLiteral && blockBody[0].type === "ReturnStatement" &&
                        blockBody[0].argument && blockBody[0].argument.type === "ObjectExpression") {
                        return;
                    }
                    if (never || asNeeded && blockBody[0].type === "ReturnStatement") {
                        let messageId;
                        if (blockBody.length === 0) {
                            messageId = "unexpectedEmptyBlock";
                        }
                        else if (blockBody.length > 1) {
                            messageId = "unexpectedOtherBlock";
                        }
                        else if (blockBody[0].argument === null) {
                            messageId = "unexpectedSingleBlock";
                        }
                        else if (astUtils.isOpeningBraceToken(sourceCode.getFirstToken(blockBody[0], { skip: 1 }))) {
                            messageId = "unexpectedObjectBlock";
                        }
                        else {
                            messageId = "unexpectedSingleBlock";
                        }
                        context.report({
                            node,
                            loc: arrowBody.loc,
                            messageId,
                            fix(fixer) {
                                const fixes = [];
                                if (blockBody.length !== 1 ||
                                    blockBody[0].type !== "ReturnStatement" ||
                                    !blockBody[0].argument ||
                                    hasASIProblem(sourceCode.getTokenAfter(arrowBody))) {
                                    return fixes;
                                }
                                const openingBrace = sourceCode.getFirstToken(arrowBody);
                                const closingBrace = sourceCode.getLastToken(arrowBody);
                                const firstValueToken = sourceCode.getFirstToken(blockBody[0], 1);
                                const lastValueToken = sourceCode.getLastToken(blockBody[0]);
                                const commentsExist = sourceCode.commentsExistBetween(openingBrace, firstValueToken) ||
                                    sourceCode.commentsExistBetween(lastValueToken, closingBrace);
                                /*
                                 * Remove tokens around the return value.
                                 * If comments don't exist, remove extra spaces as well.
                                 */
                                if (commentsExist) {
                                    fixes.push(fixer.remove(openingBrace), fixer.remove(closingBrace), fixer.remove(sourceCode.getTokenAfter(openingBrace)) // return keyword
                                    );
                                }
                                else {
                                    fixes.push(fixer.removeRange([openingBrace.range[0], firstValueToken.range[0]]), fixer.removeRange([lastValueToken.range[1], closingBrace.range[1]]));
                                }
                                /*
                                 * If the first token of the return value is `{` or the return value is a sequence expression,
                                 * enclose the return value by parentheses to avoid syntax error.
                                 */
                                if (astUtils.isOpeningBraceToken(firstValueToken) || blockBody[0].argument.type === "SequenceExpression" || (funcInfo.hasInOperator && isInsideForLoopInitializer(node))) {
                                    if (!astUtils.isParenthesised(sourceCode, blockBody[0].argument)) {
                                        fixes.push(fixer.insertTextBefore(firstValueToken, "("), fixer.insertTextAfter(lastValueToken, ")"));
                                    }
                                }
                                /*
                                 * If the last token of the return statement is semicolon, remove it.
                                 * Non-block arrow body is an expression, not a statement.
                                 */
                                if (astUtils.isSemicolonToken(lastValueToken)) {
                                    fixes.push(fixer.remove(lastValueToken));
                                }
                                return fixes;
                            }
                        });
                    }
                }
                else {
                    if (always || (asNeeded && requireReturnForObjectLiteral && arrowBody.type === "ObjectExpression")) {
                        context.report({
                            node,
                            loc: arrowBody.loc,
                            messageId: "expectedBlock",
                            fix(fixer) {
                                const fixes = [];
                                const arrowToken = sourceCode.getTokenBefore(arrowBody, astUtils.isArrowToken);
                                const [firstTokenAfterArrow, secondTokenAfterArrow] = sourceCode.getTokensAfter(arrowToken, { count: 2 });
                                const lastToken = sourceCode.getLastToken(node);
                                let parenthesisedObjectLiteral = null;
                                if (astUtils.isOpeningParenToken(firstTokenAfterArrow) &&
                                    astUtils.isOpeningBraceToken(secondTokenAfterArrow)) {
                                    const braceNode = sourceCode.getNodeByRangeIndex(secondTokenAfterArrow.range[0]);
                                    if (braceNode.type === "ObjectExpression") {
                                        parenthesisedObjectLiteral = braceNode;
                                    }
                                }
                                // If the value is object literal, remove parentheses which were forced by syntax.
                                if (parenthesisedObjectLiteral) {
                                    const openingParenToken = firstTokenAfterArrow;
                                    const openingBraceToken = secondTokenAfterArrow;
                                    if (astUtils.isTokenOnSameLine(openingParenToken, openingBraceToken)) {
                                        fixes.push(fixer.replaceText(openingParenToken, "{return "));
                                    }
                                    else {
                                        // Avoid ASI
                                        fixes.push(fixer.replaceText(openingParenToken, "{"), fixer.insertTextBefore(openingBraceToken, "return "));
                                    }
                                    // Closing paren for the object doesn't have to be lastToken, e.g.: () => ({}).foo()
                                    fixes.push(fixer.remove(findClosingParen(parenthesisedObjectLiteral)));
                                    fixes.push(fixer.insertTextAfter(lastToken, "}"));
                                }
                                else {
                                    fixes.push(fixer.insertTextBefore(firstTokenAfterArrow, "{return "));
                                    fixes.push(fixer.insertTextAfter(lastToken, "}"));
                                }
                                return fixes;
                            }
                        });
                    }
                }
            }
            return {
                "BinaryExpression[operator='in']"() {
                    let info = funcInfo;
                    while (info) {
                        info.hasInOperator = true;
                        info = info.upper;
                    }
                },
                ArrowFunctionExpression() {
                    funcInfo = {
                        upper: funcInfo,
                        hasInOperator: false
                    };
                },
                "ArrowFunctionExpression:exit"(node) {
                    validate(node);
                    funcInfo = funcInfo.upper;
                }
            };
        }
    };
},
/* eslint/lib/rules/arrow-parens.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Determines if the given arrow function has block body.
     * @param {ASTNode} node `ArrowFunctionExpression` node.
     * @returns {boolean} `true` if the function has block body.
     */
    function hasBlockBody(node) {
        return node.body.type === "BlockStatement";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require parentheses around arrow function arguments",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/arrow-parens"
            },
            fixable: "code",
            schema: [
                {
                    enum: ["always", "as-needed"]
                },
                {
                    type: "object",
                    properties: {
                        requireForBlockBody: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedParens: "Unexpected parentheses around single function argument.",
                expectedParens: "Expected parentheses around arrow function argument.",
                unexpectedParensInline: "Unexpected parentheses around single function argument having a body with no curly braces.",
                expectedParensBlock: "Expected parentheses around arrow function argument having a body with curly braces."
            }
        },
        create(context) {
            const asNeeded = context.options[0] === "as-needed";
            const requireForBlockBody = asNeeded && context.options[1] && context.options[1].requireForBlockBody === true;
            const sourceCode = context.getSourceCode();
            /**
             * Finds opening paren of parameters for the given arrow function, if it exists.
             * It is assumed that the given arrow function has exactly one parameter.
             * @param {ASTNode} node `ArrowFunctionExpression` node.
             * @returns {Token|null} the opening paren, or `null` if the given arrow function doesn't have parens of parameters.
             */
            function findOpeningParenOfParams(node) {
                const tokenBeforeParams = sourceCode.getTokenBefore(node.params[0]);
                if (tokenBeforeParams &&
                    astUtils.isOpeningParenToken(tokenBeforeParams) &&
                    node.range[0] <= tokenBeforeParams.range[0]) {
                    return tokenBeforeParams;
                }
                return null;
            }
            /**
             * Finds closing paren of parameters for the given arrow function.
             * It is assumed that the given arrow function has parens of parameters and that it has exactly one parameter.
             * @param {ASTNode} node `ArrowFunctionExpression` node.
             * @returns {Token} the closing paren of parameters.
             */
            function getClosingParenOfParams(node) {
                return sourceCode.getTokenAfter(node.params[0], astUtils.isClosingParenToken);
            }
            /**
             * Determines whether the given arrow function has comments inside parens of parameters.
             * It is assumed that the given arrow function has parens of parameters.
             * @param {ASTNode} node `ArrowFunctionExpression` node.
             * @param {Token} openingParen Opening paren of parameters.
             * @returns {boolean} `true` if the function has at least one comment inside of parens of parameters.
             */
            function hasCommentsInParensOfParams(node, openingParen) {
                return sourceCode.commentsExistBetween(openingParen, getClosingParenOfParams(node));
            }
            /**
             * Determines whether the given arrow function has unexpected tokens before opening paren of parameters,
             * in which case it will be assumed that the existing parens of parameters are necessary.
             * Only tokens within the range of the arrow function (tokens that are part of the arrow function) are taken into account.
             * Example: <T>(a) => b
             * @param {ASTNode} node `ArrowFunctionExpression` node.
             * @param {Token} openingParen Opening paren of parameters.
             * @returns {boolean} `true` if the function has at least one unexpected token.
             */
            function hasUnexpectedTokensBeforeOpeningParen(node, openingParen) {
                const expectedCount = node.async ? 1 : 0;
                return sourceCode.getFirstToken(node, { skip: expectedCount }) !== openingParen;
            }
            return {
                "ArrowFunctionExpression[params.length=1]"(node) {
                    const shouldHaveParens = !asNeeded || requireForBlockBody && hasBlockBody(node);
                    const openingParen = findOpeningParenOfParams(node);
                    const hasParens = openingParen !== null;
                    const [param] = node.params;
                    if (shouldHaveParens && !hasParens) {
                        context.report({
                            node,
                            messageId: requireForBlockBody ? "expectedParensBlock" : "expectedParens",
                            loc: param.loc,
                            *fix(fixer) {
                                yield fixer.insertTextBefore(param, "(");
                                yield fixer.insertTextAfter(param, ")");
                            }
                        });
                    }
                    if (!shouldHaveParens &&
                        hasParens &&
                        param.type === "Identifier" &&
                        !param.typeAnnotation &&
                        !node.returnType &&
                        !hasCommentsInParensOfParams(node, openingParen) &&
                        !hasUnexpectedTokensBeforeOpeningParen(node, openingParen)) {
                        context.report({
                            node,
                            messageId: requireForBlockBody ? "unexpectedParensInline" : "unexpectedParens",
                            loc: param.loc,
                            *fix(fixer) {
                                const tokenBeforeOpeningParen = sourceCode.getTokenBefore(openingParen);
                                const closingParen = getClosingParenOfParams(node);
                                if (tokenBeforeOpeningParen &&
                                    tokenBeforeOpeningParen.range[1] === openingParen.range[0] &&
                                    !astUtils.canTokensBeAdjacent(tokenBeforeOpeningParen, sourceCode.getFirstToken(param))) {
                                    yield fixer.insertTextBefore(openingParen, " ");
                                }
                                // remove parens, whitespace inside parens, and possible trailing comma
                                yield fixer.removeRange([openingParen.range[0], param.range[0]]);
                                yield fixer.removeRange([param.range[1], closingParen.range[1]]);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/arrow-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing before and after the arrow in arrow functions",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/arrow-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        before: {
                            type: "boolean",
                            default: true
                        },
                        after: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                expectedBefore: "Missing space before =>.",
                unexpectedBefore: "Unexpected space before =>.",
                expectedAfter: "Missing space after =>.",
                unexpectedAfter: "Unexpected space after =>."
            }
        },
        create(context) {
            // merge rules with default
            const rule = Object.assign({}, context.options[0]);
            rule.before = rule.before !== false;
            rule.after = rule.after !== false;
            const sourceCode = context.getSourceCode();
            /**
             * Get tokens of arrow(`=>`) and before/after arrow.
             * @param {ASTNode} node The arrow function node.
             * @returns {Object} Tokens of arrow and before/after arrow.
             */
            function getTokens(node) {
                const arrow = sourceCode.getTokenBefore(node.body, astUtils.isArrowToken);
                return {
                    before: sourceCode.getTokenBefore(arrow),
                    arrow,
                    after: sourceCode.getTokenAfter(arrow)
                };
            }
            /**
             * Count spaces before/after arrow(`=>`) token.
             * @param {Object} tokens Tokens before/after arrow.
             * @returns {Object} count of space before/after arrow.
             */
            function countSpaces(tokens) {
                const before = tokens.arrow.range[0] - tokens.before.range[1];
                const after = tokens.after.range[0] - tokens.arrow.range[1];
                return { before, after };
            }
            /**
             * Determines whether space(s) before after arrow(`=>`) is satisfy rule.
             * if before/after value is `true`, there should be space(s).
             * if before/after value is `false`, there should be no space.
             * @param {ASTNode} node The arrow function node.
             * @returns {void}
             */
            function spaces(node) {
                const tokens = getTokens(node);
                const countSpace = countSpaces(tokens);
                if (rule.before) {
                    // should be space(s) before arrow
                    if (countSpace.before === 0) {
                        context.report({
                            node: tokens.before,
                            messageId: "expectedBefore",
                            fix(fixer) {
                                return fixer.insertTextBefore(tokens.arrow, " ");
                            }
                        });
                    }
                }
                else {
                    // should be no space before arrow
                    if (countSpace.before > 0) {
                        context.report({
                            node: tokens.before,
                            messageId: "unexpectedBefore",
                            fix(fixer) {
                                return fixer.removeRange([tokens.before.range[1], tokens.arrow.range[0]]);
                            }
                        });
                    }
                }
                if (rule.after) {
                    // should be space(s) after arrow
                    if (countSpace.after === 0) {
                        context.report({
                            node: tokens.after,
                            messageId: "expectedAfter",
                            fix(fixer) {
                                return fixer.insertTextAfter(tokens.arrow, " ");
                            }
                        });
                    }
                }
                else {
                    // should be no space after arrow
                    if (countSpace.after > 0) {
                        context.report({
                            node: tokens.after,
                            messageId: "unexpectedAfter",
                            fix(fixer) {
                                return fixer.removeRange([tokens.arrow.range[1], tokens.after.range[0]]);
                            }
                        });
                    }
                }
            }
            return {
                ArrowFunctionExpression: spaces
            };
        }
    };
},
/* eslint/lib/rules/block-scoped-var.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce the use of variables within the scope they are defined",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/block-scoped-var"
            },
            schema: [],
            messages: {
                outOfScope: "'{{name}}' used outside of binding context."
            }
        },
        create(context) {
            let stack = [];
            /**
             * Makes a block scope.
             * @param {ASTNode} node A node of a scope.
             * @returns {void}
             */
            function enterScope(node) {
                stack.push(node.range);
            }
            /**
             * Pops the last block scope.
             * @returns {void}
             */
            function exitScope() {
                stack.pop();
            }
            /**
             * Reports a given reference.
             * @param {eslint-scope.Reference} reference A reference to report.
             * @returns {void}
             */
            function report(reference) {
                const identifier = reference.identifier;
                context.report({ node: identifier, messageId: "outOfScope", data: { name: identifier.name } });
            }
            /**
             * Finds and reports references which are outside of valid scopes.
             * @param {ASTNode} node A node to get variables.
             * @returns {void}
             */
            function checkForVariables(node) {
                if (node.kind !== "var") {
                    return;
                }
                // Defines a predicate to check whether or not a given reference is outside of valid scope.
                const scopeRange = stack[stack.length - 1];
                /**
                 * Check if a reference is out of scope
                 * @param {ASTNode} reference node to examine
                 * @returns {boolean} True is its outside the scope
                 * @private
                 */
                function isOutsideOfScope(reference) {
                    const idRange = reference.identifier.range;
                    return idRange[0] < scopeRange[0] || idRange[1] > scopeRange[1];
                }
                // Gets declared variables, and checks its references.
                const variables = context.getDeclaredVariables(node);
                for (let i = 0; i < variables.length; ++i) {
                    // Reports.
                    variables[i]
                        .references
                        .filter(isOutsideOfScope)
                        .forEach(report);
                }
            }
            return {
                Program(node) {
                    stack = [node.range];
                },
                // Manages scopes.
                BlockStatement: enterScope,
                "BlockStatement:exit": exitScope,
                ForStatement: enterScope,
                "ForStatement:exit": exitScope,
                ForInStatement: enterScope,
                "ForInStatement:exit": exitScope,
                ForOfStatement: enterScope,
                "ForOfStatement:exit": exitScope,
                SwitchStatement: enterScope,
                "SwitchStatement:exit": exitScope,
                CatchClause: enterScope,
                "CatchClause:exit": exitScope,
                // Finds and reports references which are outside of valid scope.
                VariableDeclaration: checkForVariables
            };
        }
    };
},
/* eslint/lib/rules/block-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    const util = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "disallow or enforce spaces inside of blocks after opening block and before closing block",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/block-spacing"
            },
            fixable: "whitespace",
            schema: [
                { enum: ["always", "never"] }
            ],
            messages: {
                missing: "Requires a space {{location}} '{{token}}'.",
                extra: "Unexpected space(s) {{location}} '{{token}}'."
            }
        },
        create(context) {
            const always = (context.options[0] !== "never"), messageId = always ? "missing" : "extra", sourceCode = context.getSourceCode();
            /**
             * Gets the open brace token from a given node.
             * @param {ASTNode} node A BlockStatement/SwitchStatement node to get.
             * @returns {Token} The token of the open brace.
             */
            function getOpenBrace(node) {
                if (node.type === "SwitchStatement") {
                    if (node.cases.length > 0) {
                        return sourceCode.getTokenBefore(node.cases[0]);
                    }
                    return sourceCode.getLastToken(node, 1);
                }
                return sourceCode.getFirstToken(node);
            }
            /**
             * Checks whether or not:
             *   - given tokens are on same line.
             *   - there is/isn't a space between given tokens.
             * @param {Token} left A token to check.
             * @param {Token} right The token which is next to `left`.
             * @returns {boolean}
             *    When the option is `"always"`, `true` if there are one or more spaces between given tokens.
             *    When the option is `"never"`, `true` if there are not any spaces between given tokens.
             *    If given tokens are not on same line, it's always `true`.
             */
            function isValid(left, right) {
                return (!util.isTokenOnSameLine(left, right) ||
                    sourceCode.isSpaceBetweenTokens(left, right) === always);
            }
            /**
             * Reports invalid spacing style inside braces.
             * @param {ASTNode} node A BlockStatement/SwitchStatement node to get.
             * @returns {void}
             */
            function checkSpacingInsideBraces(node) {
                // Gets braces and the first/last token of content.
                const openBrace = getOpenBrace(node);
                const closeBrace = sourceCode.getLastToken(node);
                const firstToken = sourceCode.getTokenAfter(openBrace, { includeComments: true });
                const lastToken = sourceCode.getTokenBefore(closeBrace, { includeComments: true });
                // Skip if the node is invalid or empty.
                if (openBrace.type !== "Punctuator" ||
                    openBrace.value !== "{" ||
                    closeBrace.type !== "Punctuator" ||
                    closeBrace.value !== "}" ||
                    firstToken === closeBrace) {
                    return;
                }
                // Skip line comments for option never
                if (!always && firstToken.type === "Line") {
                    return;
                }
                // Check.
                if (!isValid(openBrace, firstToken)) {
                    let loc = openBrace.loc;
                    if (messageId === "extra") {
                        loc = {
                            start: openBrace.loc.end,
                            end: firstToken.loc.start
                        };
                    }
                    context.report({
                        node,
                        loc,
                        messageId,
                        data: {
                            location: "after",
                            token: openBrace.value
                        },
                        fix(fixer) {
                            if (always) {
                                return fixer.insertTextBefore(firstToken, " ");
                            }
                            return fixer.removeRange([openBrace.range[1], firstToken.range[0]]);
                        }
                    });
                }
                if (!isValid(lastToken, closeBrace)) {
                    let loc = closeBrace.loc;
                    if (messageId === "extra") {
                        loc = {
                            start: lastToken.loc.end,
                            end: closeBrace.loc.start
                        };
                    }
                    context.report({
                        node,
                        loc,
                        messageId,
                        data: {
                            location: "before",
                            token: closeBrace.value
                        },
                        fix(fixer) {
                            if (always) {
                                return fixer.insertTextAfter(lastToken, " ");
                            }
                            return fixer.removeRange([lastToken.range[1], closeBrace.range[0]]);
                        }
                    });
                }
            }
            return {
                BlockStatement: checkSpacingInsideBraces,
                SwitchStatement: checkSpacingInsideBraces
            };
        }
    };
},
/* eslint/lib/rules/brace-style.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent brace style for blocks",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/brace-style"
            },
            schema: [
                {
                    enum: ["1tbs", "stroustrup", "allman"]
                },
                {
                    type: "object",
                    properties: {
                        allowSingleLine: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "whitespace",
            messages: {
                nextLineOpen: "Opening curly brace does not appear on the same line as controlling statement.",
                sameLineOpen: "Opening curly brace appears on the same line as controlling statement.",
                blockSameLine: "Statement inside of curly braces should be on next line.",
                nextLineClose: "Closing curly brace does not appear on the same line as the subsequent block.",
                singleLineClose: "Closing curly brace should be on the same line as opening curly brace or on the line after the previous block.",
                sameLineClose: "Closing curly brace appears on the same line as the subsequent block."
            }
        },
        create(context) {
            const style = context.options[0] || "1tbs", params = context.options[1] || {}, sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Fixes a place where a newline unexpectedly appears
             * @param {Token} firstToken The token before the unexpected newline
             * @param {Token} secondToken The token after the unexpected newline
             * @returns {Function} A fixer function to remove the newlines between the tokens
             */
            function removeNewlineBetween(firstToken, secondToken) {
                const textRange = [firstToken.range[1], secondToken.range[0]];
                const textBetween = sourceCode.text.slice(textRange[0], textRange[1]);
                // Don't do a fix if there is a comment between the tokens
                if (textBetween.trim()) {
                    return null;
                }
                return fixer => fixer.replaceTextRange(textRange, " ");
            }
            /**
             * Validates a pair of curly brackets based on the user's config
             * @param {Token} openingCurly The opening curly bracket
             * @param {Token} closingCurly The closing curly bracket
             * @returns {void}
             */
            function validateCurlyPair(openingCurly, closingCurly) {
                const tokenBeforeOpeningCurly = sourceCode.getTokenBefore(openingCurly);
                const tokenAfterOpeningCurly = sourceCode.getTokenAfter(openingCurly);
                const tokenBeforeClosingCurly = sourceCode.getTokenBefore(closingCurly);
                const singleLineException = params.allowSingleLine && astUtils.isTokenOnSameLine(openingCurly, closingCurly);
                if (style !== "allman" && !astUtils.isTokenOnSameLine(tokenBeforeOpeningCurly, openingCurly)) {
                    context.report({
                        node: openingCurly,
                        messageId: "nextLineOpen",
                        fix: removeNewlineBetween(tokenBeforeOpeningCurly, openingCurly)
                    });
                }
                if (style === "allman" && astUtils.isTokenOnSameLine(tokenBeforeOpeningCurly, openingCurly) && !singleLineException) {
                    context.report({
                        node: openingCurly,
                        messageId: "sameLineOpen",
                        fix: fixer => fixer.insertTextBefore(openingCurly, "\n")
                    });
                }
                if (astUtils.isTokenOnSameLine(openingCurly, tokenAfterOpeningCurly) && tokenAfterOpeningCurly !== closingCurly && !singleLineException) {
                    context.report({
                        node: openingCurly,
                        messageId: "blockSameLine",
                        fix: fixer => fixer.insertTextAfter(openingCurly, "\n")
                    });
                }
                if (tokenBeforeClosingCurly !== openingCurly && !singleLineException && astUtils.isTokenOnSameLine(tokenBeforeClosingCurly, closingCurly)) {
                    context.report({
                        node: closingCurly,
                        messageId: "singleLineClose",
                        fix: fixer => fixer.insertTextBefore(closingCurly, "\n")
                    });
                }
            }
            /**
             * Validates the location of a token that appears before a keyword (e.g. a newline before `else`)
             * @param {Token} curlyToken The closing curly token. This is assumed to precede a keyword token (such as `else` or `finally`).
             * @returns {void}
             */
            function validateCurlyBeforeKeyword(curlyToken) {
                const keywordToken = sourceCode.getTokenAfter(curlyToken);
                if (style === "1tbs" && !astUtils.isTokenOnSameLine(curlyToken, keywordToken)) {
                    context.report({
                        node: curlyToken,
                        messageId: "nextLineClose",
                        fix: removeNewlineBetween(curlyToken, keywordToken)
                    });
                }
                if (style !== "1tbs" && astUtils.isTokenOnSameLine(curlyToken, keywordToken)) {
                    context.report({
                        node: curlyToken,
                        messageId: "sameLineClose",
                        fix: fixer => fixer.insertTextAfter(curlyToken, "\n")
                    });
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                BlockStatement(node) {
                    if (!astUtils.STATEMENT_LIST_PARENTS.has(node.parent.type)) {
                        validateCurlyPair(sourceCode.getFirstToken(node), sourceCode.getLastToken(node));
                    }
                },
                ClassBody(node) {
                    validateCurlyPair(sourceCode.getFirstToken(node), sourceCode.getLastToken(node));
                },
                SwitchStatement(node) {
                    const closingCurly = sourceCode.getLastToken(node);
                    const openingCurly = sourceCode.getTokenBefore(node.cases.length ? node.cases[0] : closingCurly);
                    validateCurlyPair(openingCurly, closingCurly);
                },
                IfStatement(node) {
                    if (node.consequent.type === "BlockStatement" && node.alternate) {
                        // Handle the keyword after the `if` block (before `else`)
                        validateCurlyBeforeKeyword(sourceCode.getLastToken(node.consequent));
                    }
                },
                TryStatement(node) {
                    // Handle the keyword after the `try` block (before `catch` or `finally`)
                    validateCurlyBeforeKeyword(sourceCode.getLastToken(node.block));
                    if (node.handler && node.finalizer) {
                        // Handle the keyword after the `catch` block (before `finally`)
                        validateCurlyBeforeKeyword(sourceCode.getLastToken(node.handler.body));
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/callback-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "require `return` statements after callbacks",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/callback-return"
            },
            schema: [{
                    type: "array",
                    items: { type: "string" }
                }],
            messages: {
                missingReturn: "Expected return with your callback function."
            }
        },
        create(context) {
            const callbacks = context.options[0] || ["callback", "cb", "next"], sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Find the closest parent matching a list of types.
             * @param {ASTNode} node The node whose parents we are searching
             * @param {Array} types The node types to match
             * @returns {ASTNode} The matched node or undefined.
             */
            function findClosestParentOfType(node, types) {
                if (!node.parent) {
                    return null;
                }
                if (types.indexOf(node.parent.type) === -1) {
                    return findClosestParentOfType(node.parent, types);
                }
                return node.parent;
            }
            /**
             * Check to see if a node contains only identifiers
             * @param {ASTNode} node The node to check
             * @returns {boolean} Whether or not the node contains only identifiers
             */
            function containsOnlyIdentifiers(node) {
                if (node.type === "Identifier") {
                    return true;
                }
                if (node.type === "MemberExpression") {
                    if (node.object.type === "Identifier") {
                        return true;
                    }
                    if (node.object.type === "MemberExpression") {
                        return containsOnlyIdentifiers(node.object);
                    }
                }
                return false;
            }
            /**
             * Check to see if a CallExpression is in our callback list.
             * @param {ASTNode} node The node to check against our callback names list.
             * @returns {boolean} Whether or not this function matches our callback name.
             */
            function isCallback(node) {
                return containsOnlyIdentifiers(node.callee) && callbacks.indexOf(sourceCode.getText(node.callee)) > -1;
            }
            /**
             * Determines whether or not the callback is part of a callback expression.
             * @param {ASTNode} node The callback node
             * @param {ASTNode} parentNode The expression node
             * @returns {boolean} Whether or not this is part of a callback expression
             */
            function isCallbackExpression(node, parentNode) {
                // ensure the parent node exists and is an expression
                if (!parentNode || parentNode.type !== "ExpressionStatement") {
                    return false;
                }
                // cb()
                if (parentNode.expression === node) {
                    return true;
                }
                // special case for cb && cb() and similar
                if (parentNode.expression.type === "BinaryExpression" || parentNode.expression.type === "LogicalExpression") {
                    if (parentNode.expression.right === node) {
                        return true;
                    }
                }
                return false;
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                CallExpression(node) {
                    // if we're not a callback we can return
                    if (!isCallback(node)) {
                        return;
                    }
                    // find the closest block, return or loop
                    const closestBlock = findClosestParentOfType(node, ["BlockStatement", "ReturnStatement", "ArrowFunctionExpression"]) || {};
                    // if our parent is a return we know we're ok
                    if (closestBlock.type === "ReturnStatement") {
                        return;
                    }
                    // arrow functions don't always have blocks and implicitly return
                    if (closestBlock.type === "ArrowFunctionExpression") {
                        return;
                    }
                    // block statements are part of functions and most if statements
                    if (closestBlock.type === "BlockStatement") {
                        // find the last item in the block
                        const lastItem = closestBlock.body[closestBlock.body.length - 1];
                        // if the callback is the last thing in a block that might be ok
                        if (isCallbackExpression(node, lastItem)) {
                            const parentType = closestBlock.parent.type;
                            // but only if the block is part of a function
                            if (parentType === "FunctionExpression" ||
                                parentType === "FunctionDeclaration" ||
                                parentType === "ArrowFunctionExpression") {
                                return;
                            }
                        }
                        // ending a block with a return is also ok
                        if (lastItem.type === "ReturnStatement") {
                            // but only if the callback is immediately before
                            if (isCallbackExpression(node, closestBlock.body[closestBlock.body.length - 2])) {
                                return;
                            }
                        }
                    }
                    // as long as you're the child of a function at this point you should be asked to return
                    if (findClosestParentOfType(node, ["FunctionDeclaration", "FunctionExpression", "ArrowFunctionExpression"])) {
                        context.report({ node, messageId: "missingReturn" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/camelcase.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce camelcase naming convention",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/camelcase"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        ignoreDestructuring: {
                            type: "boolean",
                            default: false
                        },
                        ignoreImports: {
                            type: "boolean",
                            default: false
                        },
                        ignoreGlobals: {
                            type: "boolean",
                            default: false
                        },
                        properties: {
                            enum: ["always", "never"]
                        },
                        allow: {
                            type: "array",
                            items: [
                                {
                                    type: "string"
                                }
                            ],
                            minItems: 0,
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                notCamelCase: "Identifier '{{name}}' is not in camel case."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            let properties = options.properties || "";
            const ignoreDestructuring = options.ignoreDestructuring;
            const ignoreImports = options.ignoreImports;
            const ignoreGlobals = options.ignoreGlobals;
            const allow = options.allow || [];
            let globalScope;
            if (properties !== "always" && properties !== "never") {
                properties = "always";
            }
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            // contains reported nodes to avoid reporting twice on destructuring with shorthand notation
            const reported = [];
            const ALLOWED_PARENT_TYPES = new Set(["CallExpression", "NewExpression"]);
            /**
             * Checks if a string contains an underscore and isn't all upper-case
             * @param {string} name The string to check.
             * @returns {boolean} if the string is underscored
             * @private
             */
            function isUnderscored(name) {
                // if there's an underscore, it might be A_CONSTANT, which is okay
                return name.includes("_") && name !== name.toUpperCase();
            }
            /**
             * Checks if a string match the ignore list
             * @param {string} name The string to check.
             * @returns {boolean} if the string is ignored
             * @private
             */
            function isAllowed(name) {
                return allow.some(entry => name === entry || name.match(new RegExp(entry, "u")));
            }
            /**
             * Checks if a parent of a node is an ObjectPattern.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} if the node is inside an ObjectPattern
             * @private
             */
            function isInsideObjectPattern(node) {
                let current = node;
                while (current) {
                    const parent = current.parent;
                    if (parent && parent.type === "Property" && parent.computed && parent.key === current) {
                        return false;
                    }
                    if (current.type === "ObjectPattern") {
                        return true;
                    }
                    current = parent;
                }
                return false;
            }
            /**
             * Checks whether the given node represents assignment target property in destructuring.
             *
             * For examples:
             *    ({a: b.foo} = c);  // => true for `foo`
             *    ([a.foo] = b);     // => true for `foo`
             *    ([a.foo = 1] = b); // => true for `foo`
             *    ({...a.foo} = b);  // => true for `foo`
             * @param {ASTNode} node An Identifier node to check
             * @returns {boolean} True if the node is an assignment target property in destructuring.
             */
            function isAssignmentTargetPropertyInDestructuring(node) {
                if (node.parent.type === "MemberExpression" &&
                    node.parent.property === node &&
                    !node.parent.computed) {
                    const effectiveParent = node.parent.parent;
                    return (effectiveParent.type === "Property" &&
                        effectiveParent.value === node.parent &&
                        effectiveParent.parent.type === "ObjectPattern" ||
                        effectiveParent.type === "ArrayPattern" ||
                        effectiveParent.type === "RestElement" ||
                        (effectiveParent.type === "AssignmentPattern" &&
                            effectiveParent.left === node.parent));
                }
                return false;
            }
            /**
             * Checks whether the given node represents a reference to a global variable that is not declared in the source code.
             * These identifiers will be allowed, as it is assumed that user has no control over the names of external global variables.
             * @param {ASTNode} node `Identifier` node to check.
             * @returns {boolean} `true` if the node is a reference to a global variable.
             */
            function isReferenceToGlobalVariable(node) {
                const variable = globalScope.set.get(node.name);
                return variable && variable.defs.length === 0 &&
                    variable.references.some(ref => ref.identifier === node);
            }
            /**
             * Checks whether the given node represents a reference to a property of an object in an object literal expression.
             * This allows to differentiate between a global variable that is allowed to be used as a reference, and the key
             * of the expressed object (which shouldn't be allowed).
             * @param {ASTNode} node `Identifier` node to check.
             * @returns {boolean} `true` if the node is a property name of an object literal expression
             */
            function isPropertyNameInObjectLiteral(node) {
                const parent = node.parent;
                return (parent.type === "Property" &&
                    parent.parent.type === "ObjectExpression" &&
                    !parent.computed &&
                    parent.key === node);
            }
            /**
             * Reports an AST node as a rule violation.
             * @param {ASTNode} node The node to report.
             * @returns {void}
             * @private
             */
            function report(node) {
                if (!reported.includes(node)) {
                    reported.push(node);
                    context.report({ node, messageId: "notCamelCase", data: { name: node.name } });
                }
            }
            return {
                Program() {
                    globalScope = context.getScope();
                },
                Identifier(node) {
                    /*
                     * Leading and trailing underscores are commonly used to flag
                     * private/protected identifiers, strip them before checking if underscored
                     */
                    const name = node.name, nameIsUnderscored = isUnderscored(name.replace(/^_+|_+$/gu, "")), effectiveParent = (node.parent.type === "MemberExpression") ? node.parent.parent : node.parent;
                    // First, we ignore the node if it match the ignore list
                    if (isAllowed(name)) {
                        return;
                    }
                    // Check if it's a global variable
                    if (ignoreGlobals && isReferenceToGlobalVariable(node) && !isPropertyNameInObjectLiteral(node)) {
                        return;
                    }
                    // MemberExpressions get special rules
                    if (node.parent.type === "MemberExpression") {
                        // "never" check properties
                        if (properties === "never") {
                            return;
                        }
                        // Always report underscored object names
                        if (node.parent.object.type === "Identifier" && node.parent.object.name === node.name && nameIsUnderscored) {
                            report(node);
                            // Report AssignmentExpressions only if they are the left side of the assignment
                        }
                        else if (effectiveParent.type === "AssignmentExpression" && nameIsUnderscored && (effectiveParent.right.type !== "MemberExpression" || effectiveParent.left.type === "MemberExpression" && effectiveParent.left.property.name === node.name)) {
                            report(node);
                        }
                        else if (isAssignmentTargetPropertyInDestructuring(node) && nameIsUnderscored) {
                            report(node);
                        }
                        /*
                         * Properties have their own rules, and
                         * AssignmentPattern nodes can be treated like Properties:
                         * e.g.: const { no_camelcased = false } = bar;
                         */
                    }
                    else if (node.parent.type === "Property" || node.parent.type === "AssignmentPattern") {
                        if (node.parent.parent && node.parent.parent.type === "ObjectPattern") {
                            if (node.parent.shorthand && node.parent.value.left && nameIsUnderscored) {
                                report(node);
                            }
                            const assignmentKeyEqualsValue = node.parent.key.name === node.parent.value.name;
                            if (nameIsUnderscored && node.parent.computed) {
                                report(node);
                            }
                            // prevent checking righthand side of destructured object
                            if (node.parent.key === node && node.parent.value !== node) {
                                return;
                            }
                            const valueIsUnderscored = node.parent.value.name && nameIsUnderscored;
                            // ignore destructuring if the option is set, unless a new identifier is created
                            if (valueIsUnderscored && !(assignmentKeyEqualsValue && ignoreDestructuring)) {
                                report(node);
                            }
                        }
                        // "never" check properties or always ignore destructuring
                        if (properties === "never" || (ignoreDestructuring && isInsideObjectPattern(node))) {
                            return;
                        }
                        // don't check right hand side of AssignmentExpression to prevent duplicate warnings
                        if (nameIsUnderscored && !ALLOWED_PARENT_TYPES.has(effectiveParent.type) && !(node.parent.right === node)) {
                            report(node);
                        }
                        // Check if it's an import specifier
                    }
                    else if (["ImportSpecifier", "ImportNamespaceSpecifier", "ImportDefaultSpecifier"].includes(node.parent.type)) {
                        if (node.parent.type === "ImportSpecifier" && ignoreImports) {
                            return;
                        }
                        // Report only if the local imported identifier is underscored
                        if (node.parent.local &&
                            node.parent.local.name === node.name &&
                            nameIsUnderscored) {
                            report(node);
                        }
                        // Report anything that is underscored that isn't a CallExpression
                    }
                    else if (nameIsUnderscored && !ALLOWED_PARENT_TYPES.has(effectiveParent.type)) {
                        report(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/capitalized-comments.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const LETTER_PATTERN = require(412) /* ./utils/patterns/letters */;
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const DEFAULT_IGNORE_PATTERN = astUtils.COMMENTS_IGNORE_PATTERN, WHITESPACE = /\s/gu, MAYBE_URL = /^\s*[^:/?#\s]+:\/\/[^?#]/u; // TODO: Combine w/ max-len pattern?
    /*
     * Base schema body for defining the basic capitalization rule, ignorePattern,
     * and ignoreInlineComments values.
     * This can be used in a few different ways in the actual schema.
     */
    const SCHEMA_BODY = {
        type: "object",
        properties: {
            ignorePattern: {
                type: "string"
            },
            ignoreInlineComments: {
                type: "boolean"
            },
            ignoreConsecutiveComments: {
                type: "boolean"
            }
        },
        additionalProperties: false
    };
    const DEFAULTS = {
        ignorePattern: "",
        ignoreInlineComments: false,
        ignoreConsecutiveComments: false
    };
    /**
     * Get normalized options for either block or line comments from the given
     * user-provided options.
     * - If the user-provided options is just a string, returns a normalized
     *   set of options using default values for all other options.
     * - If the user-provided options is an object, then a normalized option
     *   set is returned. Options specified in overrides will take priority
     *   over options specified in the main options object, which will in
     *   turn take priority over the rule's defaults.
     * @param {Object|string} rawOptions The user-provided options.
     * @param {string} which Either "line" or "block".
     * @returns {Object} The normalized options.
     */
    function getNormalizedOptions(rawOptions, which) {
        return Object.assign({}, DEFAULTS, rawOptions[which] || rawOptions);
    }
    /**
     * Get normalized options for block and line comments.
     * @param {Object|string} rawOptions The user-provided options.
     * @returns {Object} An object with "Line" and "Block" keys and corresponding
     * normalized options objects.
     */
    function getAllNormalizedOptions(rawOptions = {}) {
        return {
            Line: getNormalizedOptions(rawOptions, "line"),
            Block: getNormalizedOptions(rawOptions, "block")
        };
    }
    /**
     * Creates a regular expression for each ignorePattern defined in the rule
     * options.
     *
     * This is done in order to avoid invoking the RegExp constructor repeatedly.
     * @param {Object} normalizedOptions The normalized rule options.
     * @returns {void}
     */
    function createRegExpForIgnorePatterns(normalizedOptions) {
        Object.keys(normalizedOptions).forEach(key => {
            const ignorePatternStr = normalizedOptions[key].ignorePattern;
            if (ignorePatternStr) {
                const regExp = RegExp(`^\\s*(?:${ignorePatternStr})`, "u");
                normalizedOptions[key].ignorePatternRegExp = regExp;
            }
        });
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce or disallow capitalization of the first letter of a comment",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/capitalized-comments"
            },
            fixable: "code",
            schema: [
                { enum: ["always", "never"] },
                {
                    oneOf: [
                        SCHEMA_BODY,
                        {
                            type: "object",
                            properties: {
                                line: SCHEMA_BODY,
                                block: SCHEMA_BODY
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                unexpectedLowercaseComment: "Comments should not begin with a lowercase character.",
                unexpectedUppercaseComment: "Comments should not begin with an uppercase character."
            }
        },
        create(context) {
            const capitalize = context.options[0] || "always", normalizedOptions = getAllNormalizedOptions(context.options[1]), sourceCode = context.getSourceCode();
            createRegExpForIgnorePatterns(normalizedOptions);
            //----------------------------------------------------------------------
            // Helpers
            //----------------------------------------------------------------------
            /**
             * Checks whether a comment is an inline comment.
             *
             * For the purpose of this rule, a comment is inline if:
             * 1. The comment is preceded by a token on the same line; and
             * 2. The command is followed by a token on the same line.
             *
             * Note that the comment itself need not be single-line!
             *
             * Also, it follows from this definition that only block comments can
             * be considered as possibly inline. This is because line comments
             * would consume any following tokens on the same line as the comment.
             * @param {ASTNode} comment The comment node to check.
             * @returns {boolean} True if the comment is an inline comment, false
             * otherwise.
             */
            function isInlineComment(comment) {
                const previousToken = sourceCode.getTokenBefore(comment, { includeComments: true }), nextToken = sourceCode.getTokenAfter(comment, { includeComments: true });
                return Boolean(previousToken &&
                    nextToken &&
                    comment.loc.start.line === previousToken.loc.end.line &&
                    comment.loc.end.line === nextToken.loc.start.line);
            }
            /**
             * Determine if a comment follows another comment.
             * @param {ASTNode} comment The comment to check.
             * @returns {boolean} True if the comment follows a valid comment.
             */
            function isConsecutiveComment(comment) {
                const previousTokenOrComment = sourceCode.getTokenBefore(comment, { includeComments: true });
                return Boolean(previousTokenOrComment &&
                    ["Block", "Line"].indexOf(previousTokenOrComment.type) !== -1);
            }
            /**
             * Check a comment to determine if it is valid for this rule.
             * @param {ASTNode} comment The comment node to process.
             * @param {Object} options The options for checking this comment.
             * @returns {boolean} True if the comment is valid, false otherwise.
             */
            function isCommentValid(comment, options) {
                // 1. Check for default ignore pattern.
                if (DEFAULT_IGNORE_PATTERN.test(comment.value)) {
                    return true;
                }
                // 2. Check for custom ignore pattern.
                const commentWithoutAsterisks = comment.value
                    .replace(/\*/gu, "");
                if (options.ignorePatternRegExp && options.ignorePatternRegExp.test(commentWithoutAsterisks)) {
                    return true;
                }
                // 3. Check for inline comments.
                if (options.ignoreInlineComments && isInlineComment(comment)) {
                    return true;
                }
                // 4. Is this a consecutive comment (and are we tolerating those)?
                if (options.ignoreConsecutiveComments && isConsecutiveComment(comment)) {
                    return true;
                }
                // 5. Does the comment start with a possible URL?
                if (MAYBE_URL.test(commentWithoutAsterisks)) {
                    return true;
                }
                // 6. Is the initial word character a letter?
                const commentWordCharsOnly = commentWithoutAsterisks
                    .replace(WHITESPACE, "");
                if (commentWordCharsOnly.length === 0) {
                    return true;
                }
                const firstWordChar = commentWordCharsOnly[0];
                if (!LETTER_PATTERN.test(firstWordChar)) {
                    return true;
                }
                // 7. Check the case of the initial word character.
                const isUppercase = firstWordChar !== firstWordChar.toLocaleLowerCase(), isLowercase = firstWordChar !== firstWordChar.toLocaleUpperCase();
                if (capitalize === "always" && isLowercase) {
                    return false;
                }
                if (capitalize === "never" && isUppercase) {
                    return false;
                }
                return true;
            }
            /**
             * Process a comment to determine if it needs to be reported.
             * @param {ASTNode} comment The comment node to process.
             * @returns {void}
             */
            function processComment(comment) {
                const options = normalizedOptions[comment.type], commentValid = isCommentValid(comment, options);
                if (!commentValid) {
                    const messageId = capitalize === "always"
                        ? "unexpectedLowercaseComment"
                        : "unexpectedUppercaseComment";
                    context.report({
                        node: null,
                        loc: comment.loc,
                        messageId,
                        fix(fixer) {
                            const match = comment.value.match(LETTER_PATTERN);
                            return fixer.replaceTextRange(
                            // Offset match.index by 2 to account for the first 2 characters that start the comment (// or /*)
                            [comment.range[0] + match.index + 2, comment.range[0] + match.index + 3], capitalize === "always" ? match[0].toLocaleUpperCase() : match[0].toLocaleLowerCase());
                        }
                    });
                }
            }
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                Program() {
                    const comments = sourceCode.getAllComments();
                    comments.filter(token => token.type !== "Shebang").forEach(processComment);
                }
            };
        }
    };
},
/* eslint/lib/rules/utils/patterns/letters.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = /[A-Za-z\xAA\xB5\xBA\xC0-\xD6\xD8-\xF6\xF8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u037F\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u052F\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u08A0-\u08B2\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0980\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C39\u0C3D\u0C58\u0C59\u0C60\u0C61\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D60\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F4\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u16F1-\u16F8\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191E\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19C1-\u19C7\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2119-\u211D\u2124\u2126\u2128\u212A-\u212D\u212F-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2183\u2184\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2E2F\u3005\u3006\u3031-\u3035\u303B\u303C\u3041-\u3096\u309D-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FCC\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA69D\uA6A0-\uA6E5\uA717-\uA71F\uA722-\uA788\uA78B-\uA78E\uA790-\uA7AD\uA7B0\uA7B1\uA7F7-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uA9E0-\uA9E4\uA9E6-\uA9EF\uA9FA-\uA9FE\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA7E-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uAB30-\uAB5A\uAB5C-\uAB5F\uAB64\uAB65\uABC0-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]|\uD800[\uDC00-\uDC0B\uDC0D-\uDC26\uDC28-\uDC3A\uDC3C\uDC3D\uDC3F-\uDC4D\uDC50-\uDC5D\uDC80-\uDCFA\uDE80-\uDE9C\uDEA0-\uDED0\uDF00-\uDF1F\uDF30-\uDF40\uDF42-\uDF49\uDF50-\uDF75\uDF80-\uDF9D\uDFA0-\uDFC3\uDFC8-\uDFCF]|\uD801[\uDC00-\uDC9D\uDD00-\uDD27\uDD30-\uDD63\uDE00-\uDF36\uDF40-\uDF55\uDF60-\uDF67]|\uD802[\uDC00-\uDC05\uDC08\uDC0A-\uDC35\uDC37\uDC38\uDC3C\uDC3F-\uDC55\uDC60-\uDC76\uDC80-\uDC9E\uDD00-\uDD15\uDD20-\uDD39\uDD80-\uDDB7\uDDBE\uDDBF\uDE00\uDE10-\uDE13\uDE15-\uDE17\uDE19-\uDE33\uDE60-\uDE7C\uDE80-\uDE9C\uDEC0-\uDEC7\uDEC9-\uDEE4\uDF00-\uDF35\uDF40-\uDF55\uDF60-\uDF72\uDF80-\uDF91]|\uD803[\uDC00-\uDC48]|\uD804[\uDC03-\uDC37\uDC83-\uDCAF\uDCD0-\uDCE8\uDD03-\uDD26\uDD50-\uDD72\uDD76\uDD83-\uDDB2\uDDC1-\uDDC4\uDDDA\uDE00-\uDE11\uDE13-\uDE2B\uDEB0-\uDEDE\uDF05-\uDF0C\uDF0F\uDF10\uDF13-\uDF28\uDF2A-\uDF30\uDF32\uDF33\uDF35-\uDF39\uDF3D\uDF5D-\uDF61]|\uD805[\uDC80-\uDCAF\uDCC4\uDCC5\uDCC7\uDD80-\uDDAE\uDE00-\uDE2F\uDE44\uDE80-\uDEAA]|\uD806[\uDCA0-\uDCDF\uDCFF\uDEC0-\uDEF8]|\uD808[\uDC00-\uDF98]|[\uD80C\uD840-\uD868\uD86A-\uD86C][\uDC00-\uDFFF]|\uD80D[\uDC00-\uDC2E]|\uD81A[\uDC00-\uDE38\uDE40-\uDE5E\uDED0-\uDEED\uDF00-\uDF2F\uDF40-\uDF43\uDF63-\uDF77\uDF7D-\uDF8F]|\uD81B[\uDF00-\uDF44\uDF50\uDF93-\uDF9F]|\uD82C[\uDC00\uDC01]|\uD82F[\uDC00-\uDC6A\uDC70-\uDC7C\uDC80-\uDC88\uDC90-\uDC99]|\uD835[\uDC00-\uDC54\uDC56-\uDC9C\uDC9E\uDC9F\uDCA2\uDCA5\uDCA6\uDCA9-\uDCAC\uDCAE-\uDCB9\uDCBB\uDCBD-\uDCC3\uDCC5-\uDD05\uDD07-\uDD0A\uDD0D-\uDD14\uDD16-\uDD1C\uDD1E-\uDD39\uDD3B-\uDD3E\uDD40-\uDD44\uDD46\uDD4A-\uDD50\uDD52-\uDEA5\uDEA8-\uDEC0\uDEC2-\uDEDA\uDEDC-\uDEFA\uDEFC-\uDF14\uDF16-\uDF34\uDF36-\uDF4E\uDF50-\uDF6E\uDF70-\uDF88\uDF8A-\uDFA8\uDFAA-\uDFC2\uDFC4-\uDFCB]|\uD83A[\uDC00-\uDCC4]|\uD83B[\uDE00-\uDE03\uDE05-\uDE1F\uDE21\uDE22\uDE24\uDE27\uDE29-\uDE32\uDE34-\uDE37\uDE39\uDE3B\uDE42\uDE47\uDE49\uDE4B\uDE4D-\uDE4F\uDE51\uDE52\uDE54\uDE57\uDE59\uDE5B\uDE5D\uDE5F\uDE61\uDE62\uDE64\uDE67-\uDE6A\uDE6C-\uDE72\uDE74-\uDE77\uDE79-\uDE7C\uDE7E\uDE80-\uDE89\uDE8B-\uDE9B\uDEA1-\uDEA3\uDEA5-\uDEA9\uDEAB-\uDEBB]|\uD869[\uDC00-\uDED6\uDF00-\uDFFF]|\uD86D[\uDC00-\uDF34\uDF40-\uDFFF]|\uD86E[\uDC00-\uDC1D]|\uD87E[\uDC00-\uDE1D]/u;
},
/* eslint/lib/rules/class-methods-use-this.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce that class methods utilize `this`",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/class-methods-use-this"
            },
            schema: [{
                    type: "object",
                    properties: {
                        exceptMethods: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                missingThis: "Expected 'this' to be used by class {{name}}."
            }
        },
        create(context) {
            const config = Object.assign({}, context.options[0]);
            const exceptMethods = new Set(config.exceptMethods || []);
            const stack = [];
            /**
             * Initializes the current context to false and pushes it onto the stack.
             * These booleans represent whether 'this' has been used in the context.
             * @returns {void}
             * @private
             */
            function enterFunction() {
                stack.push(false);
            }
            /**
             * Check if the node is an instance method
             * @param {ASTNode} node node to check
             * @returns {boolean} True if its an instance method
             * @private
             */
            function isInstanceMethod(node) {
                return !node.static && node.kind !== "constructor" && node.type === "MethodDefinition";
            }
            /**
             * Check if the node is an instance method not excluded by config
             * @param {ASTNode} node node to check
             * @returns {boolean} True if it is an instance method, and not excluded by config
             * @private
             */
            function isIncludedInstanceMethod(node) {
                return isInstanceMethod(node) &&
                    (node.computed || !exceptMethods.has(node.key.name));
            }
            /**
             * Checks if we are leaving a function that is a method, and reports if 'this' has not been used.
             * Static methods and the constructor are exempt.
             * Then pops the context off the stack.
             * @param {ASTNode} node A function node that was entered.
             * @returns {void}
             * @private
             */
            function exitFunction(node) {
                const methodUsesThis = stack.pop();
                if (isIncludedInstanceMethod(node.parent) && !methodUsesThis) {
                    context.report({
                        node,
                        messageId: "missingThis",
                        data: {
                            name: astUtils.getFunctionNameWithKind(node)
                        }
                    });
                }
            }
            /**
             * Mark the current context as having used 'this'.
             * @returns {void}
             * @private
             */
            function markThisUsed() {
                if (stack.length) {
                    stack[stack.length - 1] = true;
                }
            }
            return {
                FunctionDeclaration: enterFunction,
                "FunctionDeclaration:exit": exitFunction,
                FunctionExpression: enterFunction,
                "FunctionExpression:exit": exitFunction,
                ThisExpression: markThisUsed,
                Super: markThisUsed
            };
        }
    };
},
/* eslint/lib/rules/comma-dangle.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const DEFAULT_OPTIONS = Object.freeze({
        arrays: "never",
        objects: "never",
        imports: "never",
        exports: "never",
        functions: "never"
    });
    /**
     * Checks whether or not a trailing comma is allowed in a given node.
     * If the `lastItem` is `RestElement` or `RestProperty`, it disallows trailing commas.
     * @param {ASTNode} lastItem The node of the last element in the given node.
     * @returns {boolean} `true` if a trailing comma is allowed.
     */
    function isTrailingCommaAllowed(lastItem) {
        return !(lastItem.type === "RestElement" ||
            lastItem.type === "RestProperty" ||
            lastItem.type === "ExperimentalRestProperty");
    }
    /**
     * Normalize option value.
     * @param {string|Object|undefined} optionValue The 1st option value to normalize.
     * @param {number} ecmaVersion The normalized ECMAScript version.
     * @returns {Object} The normalized option value.
     */
    function normalizeOptions(optionValue, ecmaVersion) {
        if (typeof optionValue === "string") {
            return {
                arrays: optionValue,
                objects: optionValue,
                imports: optionValue,
                exports: optionValue,
                functions: (!ecmaVersion || ecmaVersion < 8) ? "ignore" : optionValue
            };
        }
        if (typeof optionValue === "object" && optionValue !== null) {
            return {
                arrays: optionValue.arrays || DEFAULT_OPTIONS.arrays,
                objects: optionValue.objects || DEFAULT_OPTIONS.objects,
                imports: optionValue.imports || DEFAULT_OPTIONS.imports,
                exports: optionValue.exports || DEFAULT_OPTIONS.exports,
                functions: optionValue.functions || DEFAULT_OPTIONS.functions
            };
        }
        return DEFAULT_OPTIONS;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow trailing commas",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/comma-dangle"
            },
            fixable: "code",
            schema: {
                definitions: {
                    value: {
                        enum: [
                            "always-multiline",
                            "always",
                            "never",
                            "only-multiline"
                        ]
                    },
                    valueWithIgnore: {
                        enum: [
                            "always-multiline",
                            "always",
                            "ignore",
                            "never",
                            "only-multiline"
                        ]
                    }
                },
                type: "array",
                items: [
                    {
                        oneOf: [
                            {
                                $ref: "#/definitions/value"
                            },
                            {
                                type: "object",
                                properties: {
                                    arrays: { $ref: "#/definitions/valueWithIgnore" },
                                    objects: { $ref: "#/definitions/valueWithIgnore" },
                                    imports: { $ref: "#/definitions/valueWithIgnore" },
                                    exports: { $ref: "#/definitions/valueWithIgnore" },
                                    functions: { $ref: "#/definitions/valueWithIgnore" }
                                },
                                additionalProperties: false
                            }
                        ]
                    }
                ]
            },
            messages: {
                unexpected: "Unexpected trailing comma.",
                missing: "Missing trailing comma."
            }
        },
        create(context) {
            const options = normalizeOptions(context.options[0], context.parserOptions.ecmaVersion);
            const sourceCode = context.getSourceCode();
            /**
             * Gets the last item of the given node.
             * @param {ASTNode} node The node to get.
             * @returns {ASTNode|null} The last node or null.
             */
            function getLastItem(node) {
                /**
                 * Returns the last element of an array
                 * @param {any[]} array The input array
                 * @returns {any} The last element
                 */
                function last(array) {
                    return array[array.length - 1];
                }
                switch (node.type) {
                    case "ObjectExpression":
                    case "ObjectPattern":
                        return last(node.properties);
                    case "ArrayExpression":
                    case "ArrayPattern":
                        return last(node.elements);
                    case "ImportDeclaration":
                    case "ExportNamedDeclaration":
                        return last(node.specifiers);
                    case "FunctionDeclaration":
                    case "FunctionExpression":
                    case "ArrowFunctionExpression":
                        return last(node.params);
                    case "CallExpression":
                    case "NewExpression":
                        return last(node.arguments);
                    default:
                        return null;
                }
            }
            /**
             * Gets the trailing comma token of the given node.
             * If the trailing comma does not exist, this returns the token which is
             * the insertion point of the trailing comma token.
             * @param {ASTNode} node The node to get.
             * @param {ASTNode} lastItem The last item of the node.
             * @returns {Token} The trailing comma token or the insertion point.
             */
            function getTrailingToken(node, lastItem) {
                switch (node.type) {
                    case "ObjectExpression":
                    case "ArrayExpression":
                    case "CallExpression":
                    case "NewExpression":
                        return sourceCode.getLastToken(node, 1);
                    default: {
                        const nextToken = sourceCode.getTokenAfter(lastItem);
                        if (astUtils.isCommaToken(nextToken)) {
                            return nextToken;
                        }
                        return sourceCode.getLastToken(lastItem);
                    }
                }
            }
            /**
             * Checks whether or not a given node is multiline.
             * This rule handles a given node as multiline when the closing parenthesis
             * and the last element are not on the same line.
             * @param {ASTNode} node A node to check.
             * @returns {boolean} `true` if the node is multiline.
             */
            function isMultiline(node) {
                const lastItem = getLastItem(node);
                if (!lastItem) {
                    return false;
                }
                const penultimateToken = getTrailingToken(node, lastItem);
                const lastToken = sourceCode.getTokenAfter(penultimateToken);
                return lastToken.loc.end.line !== penultimateToken.loc.end.line;
            }
            /**
             * Reports a trailing comma if it exists.
             * @param {ASTNode} node A node to check. Its type is one of
             *   ObjectExpression, ObjectPattern, ArrayExpression, ArrayPattern,
             *   ImportDeclaration, and ExportNamedDeclaration.
             * @returns {void}
             */
            function forbidTrailingComma(node) {
                const lastItem = getLastItem(node);
                if (!lastItem || (node.type === "ImportDeclaration" && lastItem.type !== "ImportSpecifier")) {
                    return;
                }
                const trailingToken = getTrailingToken(node, lastItem);
                if (astUtils.isCommaToken(trailingToken)) {
                    context.report({
                        node: lastItem,
                        loc: trailingToken.loc,
                        messageId: "unexpected",
                        fix(fixer) {
                            return fixer.remove(trailingToken);
                        }
                    });
                }
            }
            /**
             * Reports the last element of a given node if it does not have a trailing
             * comma.
             *
             * If a given node is `ArrayPattern` which has `RestElement`, the trailing
             * comma is disallowed, so report if it exists.
             * @param {ASTNode} node A node to check. Its type is one of
             *   ObjectExpression, ObjectPattern, ArrayExpression, ArrayPattern,
             *   ImportDeclaration, and ExportNamedDeclaration.
             * @returns {void}
             */
            function forceTrailingComma(node) {
                const lastItem = getLastItem(node);
                if (!lastItem || (node.type === "ImportDeclaration" && lastItem.type !== "ImportSpecifier")) {
                    return;
                }
                if (!isTrailingCommaAllowed(lastItem)) {
                    forbidTrailingComma(node);
                    return;
                }
                const trailingToken = getTrailingToken(node, lastItem);
                if (trailingToken.value !== ",") {
                    context.report({
                        node: lastItem,
                        loc: {
                            start: trailingToken.loc.end,
                            end: astUtils.getNextLocation(sourceCode, trailingToken.loc.end)
                        },
                        messageId: "missing",
                        fix(fixer) {
                            return fixer.insertTextAfter(trailingToken, ",");
                        }
                    });
                }
            }
            /**
             * If a given node is multiline, reports the last element of a given node
             * when it does not have a trailing comma.
             * Otherwise, reports a trailing comma if it exists.
             * @param {ASTNode} node A node to check. Its type is one of
             *   ObjectExpression, ObjectPattern, ArrayExpression, ArrayPattern,
             *   ImportDeclaration, and ExportNamedDeclaration.
             * @returns {void}
             */
            function forceTrailingCommaIfMultiline(node) {
                if (isMultiline(node)) {
                    forceTrailingComma(node);
                }
                else {
                    forbidTrailingComma(node);
                }
            }
            /**
             * Only if a given node is not multiline, reports the last element of a given node
             * when it does not have a trailing comma.
             * Otherwise, reports a trailing comma if it exists.
             * @param {ASTNode} node A node to check. Its type is one of
             *   ObjectExpression, ObjectPattern, ArrayExpression, ArrayPattern,
             *   ImportDeclaration, and ExportNamedDeclaration.
             * @returns {void}
             */
            function allowTrailingCommaIfMultiline(node) {
                if (!isMultiline(node)) {
                    forbidTrailingComma(node);
                }
            }
            const predicate = {
                always: forceTrailingComma,
                "always-multiline": forceTrailingCommaIfMultiline,
                "only-multiline": allowTrailingCommaIfMultiline,
                never: forbidTrailingComma,
                ignore: () => { }
            };
            return {
                ObjectExpression: predicate[options.objects],
                ObjectPattern: predicate[options.objects],
                ArrayExpression: predicate[options.arrays],
                ArrayPattern: predicate[options.arrays],
                ImportDeclaration: predicate[options.imports],
                ExportNamedDeclaration: predicate[options.exports],
                FunctionDeclaration: predicate[options.functions],
                FunctionExpression: predicate[options.functions],
                ArrowFunctionExpression: predicate[options.functions],
                CallExpression: predicate[options.functions],
                NewExpression: predicate[options.functions]
            };
        }
    };
},
/* eslint/lib/rules/comma-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing before and after commas",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/comma-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        before: {
                            type: "boolean",
                            default: false
                        },
                        after: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                missing: "A space is required {{loc}} ','.",
                unexpected: "There should be no space {{loc}} ','."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const tokensAndComments = sourceCode.tokensAndComments;
            const options = {
                before: context.options[0] ? context.options[0].before : false,
                after: context.options[0] ? context.options[0].after : true
            };
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            // list of comma tokens to ignore for the check of leading whitespace
            const commaTokensToIgnore = [];
            /**
             * Reports a spacing error with an appropriate message.
             * @param {ASTNode} node The binary expression node to report.
             * @param {string} loc Is the error "before" or "after" the comma?
             * @param {ASTNode} otherNode The node at the left or right of `node`
             * @returns {void}
             * @private
             */
            function report(node, loc, otherNode) {
                context.report({
                    node,
                    fix(fixer) {
                        if (options[loc]) {
                            if (loc === "before") {
                                return fixer.insertTextBefore(node, " ");
                            }
                            return fixer.insertTextAfter(node, " ");
                        }
                        let start, end;
                        const newText = "";
                        if (loc === "before") {
                            start = otherNode.range[1];
                            end = node.range[0];
                        }
                        else {
                            start = node.range[1];
                            end = otherNode.range[0];
                        }
                        return fixer.replaceTextRange([start, end], newText);
                    },
                    messageId: options[loc] ? "missing" : "unexpected",
                    data: {
                        loc
                    }
                });
            }
            /**
             * Validates the spacing around a comma token.
             * @param {Object} tokens The tokens to be validated.
             * @param {Token} tokens.comma The token representing the comma.
             * @param {Token} [tokens.left] The last token before the comma.
             * @param {Token} [tokens.right] The first token after the comma.
             * @param {Token|ASTNode} reportItem The item to use when reporting an error.
             * @returns {void}
             * @private
             */
            function validateCommaItemSpacing(tokens, reportItem) {
                if (tokens.left && astUtils.isTokenOnSameLine(tokens.left, tokens.comma) &&
                    (options.before !== sourceCode.isSpaceBetweenTokens(tokens.left, tokens.comma))) {
                    report(reportItem, "before", tokens.left);
                }
                if (tokens.right && astUtils.isClosingParenToken(tokens.right)) {
                    return;
                }
                if (tokens.right && !options.after && tokens.right.type === "Line") {
                    return;
                }
                if (tokens.right && astUtils.isTokenOnSameLine(tokens.comma, tokens.right) &&
                    (options.after !== sourceCode.isSpaceBetweenTokens(tokens.comma, tokens.right))) {
                    report(reportItem, "after", tokens.right);
                }
            }
            /**
             * Adds null elements of the given ArrayExpression or ArrayPattern node to the ignore list.
             * @param {ASTNode} node An ArrayExpression or ArrayPattern node.
             * @returns {void}
             */
            function addNullElementsToIgnoreList(node) {
                let previousToken = sourceCode.getFirstToken(node);
                node.elements.forEach(element => {
                    let token;
                    if (element === null) {
                        token = sourceCode.getTokenAfter(previousToken);
                        if (astUtils.isCommaToken(token)) {
                            commaTokensToIgnore.push(token);
                        }
                    }
                    else {
                        token = sourceCode.getTokenAfter(element);
                    }
                    previousToken = token;
                });
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                "Program:exit"() {
                    tokensAndComments.forEach((token, i) => {
                        if (!astUtils.isCommaToken(token)) {
                            return;
                        }
                        if (token && token.type === "JSXText") {
                            return;
                        }
                        const previousToken = tokensAndComments[i - 1];
                        const nextToken = tokensAndComments[i + 1];
                        validateCommaItemSpacing({
                            comma: token,
                            left: astUtils.isCommaToken(previousToken) || commaTokensToIgnore.includes(token) ? null : previousToken,
                            right: astUtils.isCommaToken(nextToken) ? null : nextToken
                        }, token);
                    });
                },
                ArrayExpression: addNullElementsToIgnoreList,
                ArrayPattern: addNullElementsToIgnoreList
            };
        }
    };
},
/* eslint/lib/rules/comma-style.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent comma style",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/comma-style"
            },
            fixable: "code",
            schema: [
                {
                    enum: ["first", "last"]
                },
                {
                    type: "object",
                    properties: {
                        exceptions: {
                            type: "object",
                            additionalProperties: {
                                type: "boolean"
                            }
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedLineBeforeAndAfterComma: "Bad line breaking before and after ','.",
                expectedCommaFirst: "',' should be placed first.",
                expectedCommaLast: "',' should be placed last."
            }
        },
        create(context) {
            const style = context.options[0] || "last", sourceCode = context.getSourceCode();
            const exceptions = {
                ArrayPattern: true,
                ArrowFunctionExpression: true,
                CallExpression: true,
                FunctionDeclaration: true,
                FunctionExpression: true,
                ImportDeclaration: true,
                ObjectPattern: true,
                NewExpression: true
            };
            if (context.options.length === 2 && Object.prototype.hasOwnProperty.call(context.options[1], "exceptions")) {
                const keys = Object.keys(context.options[1].exceptions);
                for (let i = 0; i < keys.length; i++) {
                    exceptions[keys[i]] = context.options[1].exceptions[keys[i]];
                }
            }
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Modified text based on the style
             * @param {string} styleType Style type
             * @param {string} text Source code text
             * @returns {string} modified text
             * @private
             */
            function getReplacedText(styleType, text) {
                switch (styleType) {
                    case "between":
                        return `,${text.replace(astUtils.LINEBREAK_MATCHER, "")}`;
                    case "first":
                        return `${text},`;
                    case "last":
                        return `,${text}`;
                    default:
                        return "";
                }
            }
            /**
             * Determines the fixer function for a given style.
             * @param {string} styleType comma style
             * @param {ASTNode} previousItemToken The token to check.
             * @param {ASTNode} commaToken The token to check.
             * @param {ASTNode} currentItemToken The token to check.
             * @returns {Function} Fixer function
             * @private
             */
            function getFixerFunction(styleType, previousItemToken, commaToken, currentItemToken) {
                const text = sourceCode.text.slice(previousItemToken.range[1], commaToken.range[0]) +
                    sourceCode.text.slice(commaToken.range[1], currentItemToken.range[0]);
                const range = [previousItemToken.range[1], currentItemToken.range[0]];
                return function (fixer) {
                    return fixer.replaceTextRange(range, getReplacedText(styleType, text));
                };
            }
            /**
             * Validates the spacing around single items in lists.
             * @param {Token} previousItemToken The last token from the previous item.
             * @param {Token} commaToken The token representing the comma.
             * @param {Token} currentItemToken The first token of the current item.
             * @param {Token} reportItem The item to use when reporting an error.
             * @returns {void}
             * @private
             */
            function validateCommaItemSpacing(previousItemToken, commaToken, currentItemToken, reportItem) {
                // if single line
                if (astUtils.isTokenOnSameLine(commaToken, currentItemToken) &&
                    astUtils.isTokenOnSameLine(previousItemToken, commaToken)) {
                    // do nothing.
                }
                else if (!astUtils.isTokenOnSameLine(commaToken, currentItemToken) &&
                    !astUtils.isTokenOnSameLine(previousItemToken, commaToken)) {
                    const comment = sourceCode.getCommentsAfter(commaToken)[0];
                    const styleType = comment && comment.type === "Block" && astUtils.isTokenOnSameLine(commaToken, comment)
                        ? style
                        : "between";
                    // lone comma
                    context.report({
                        node: reportItem,
                        loc: commaToken.loc,
                        messageId: "unexpectedLineBeforeAndAfterComma",
                        fix: getFixerFunction(styleType, previousItemToken, commaToken, currentItemToken)
                    });
                }
                else if (style === "first" && !astUtils.isTokenOnSameLine(commaToken, currentItemToken)) {
                    context.report({
                        node: reportItem,
                        loc: commaToken.loc,
                        messageId: "expectedCommaFirst",
                        fix: getFixerFunction(style, previousItemToken, commaToken, currentItemToken)
                    });
                }
                else if (style === "last" && astUtils.isTokenOnSameLine(commaToken, currentItemToken)) {
                    context.report({
                        node: reportItem,
                        loc: commaToken.loc,
                        messageId: "expectedCommaLast",
                        fix: getFixerFunction(style, previousItemToken, commaToken, currentItemToken)
                    });
                }
            }
            /**
             * Checks the comma placement with regards to a declaration/property/element
             * @param {ASTNode} node The binary expression node to check
             * @param {string} property The property of the node containing child nodes.
             * @private
             * @returns {void}
             */
            function validateComma(node, property) {
                const items = node[property], arrayLiteral = (node.type === "ArrayExpression" || node.type === "ArrayPattern");
                if (items.length > 1 || arrayLiteral) {
                    // seed as opening [
                    let previousItemToken = sourceCode.getFirstToken(node);
                    items.forEach(item => {
                        const commaToken = item ? sourceCode.getTokenBefore(item) : previousItemToken, currentItemToken = item ? sourceCode.getFirstToken(item) : sourceCode.getTokenAfter(commaToken), reportItem = item || currentItemToken;
                        /*
                         * This works by comparing three token locations:
                         * - previousItemToken is the last token of the previous item
                         * - commaToken is the location of the comma before the current item
                         * - currentItemToken is the first token of the current item
                         *
                         * These values get switched around if item is undefined.
                         * previousItemToken will refer to the last token not belonging
                         * to the current item, which could be a comma or an opening
                         * square bracket. currentItemToken could be a comma.
                         *
                         * All comparisons are done based on these tokens directly, so
                         * they are always valid regardless of an undefined item.
                         */
                        if (astUtils.isCommaToken(commaToken)) {
                            validateCommaItemSpacing(previousItemToken, commaToken, currentItemToken, reportItem);
                        }
                        if (item) {
                            const tokenAfterItem = sourceCode.getTokenAfter(item, astUtils.isNotClosingParenToken);
                            previousItemToken = tokenAfterItem
                                ? sourceCode.getTokenBefore(tokenAfterItem)
                                : sourceCode.ast.tokens[sourceCode.ast.tokens.length - 1];
                        }
                        else {
                            previousItemToken = currentItemToken;
                        }
                    });
                    /*
                     * Special case for array literals that have empty last items, such
                     * as [ 1, 2, ]. These arrays only have two items show up in the
                     * AST, so we need to look at the token to verify that there's no
                     * dangling comma.
                     */
                    if (arrayLiteral) {
                        const lastToken = sourceCode.getLastToken(node), nextToLastToken = sourceCode.getTokenBefore(lastToken);
                        if (astUtils.isCommaToken(nextToLastToken)) {
                            validateCommaItemSpacing(sourceCode.getTokenBefore(nextToLastToken), nextToLastToken, lastToken, lastToken);
                        }
                    }
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            const nodes = {};
            if (!exceptions.VariableDeclaration) {
                nodes.VariableDeclaration = function (node) {
                    validateComma(node, "declarations");
                };
            }
            if (!exceptions.ObjectExpression) {
                nodes.ObjectExpression = function (node) {
                    validateComma(node, "properties");
                };
            }
            if (!exceptions.ObjectPattern) {
                nodes.ObjectPattern = function (node) {
                    validateComma(node, "properties");
                };
            }
            if (!exceptions.ArrayExpression) {
                nodes.ArrayExpression = function (node) {
                    validateComma(node, "elements");
                };
            }
            if (!exceptions.ArrayPattern) {
                nodes.ArrayPattern = function (node) {
                    validateComma(node, "elements");
                };
            }
            if (!exceptions.FunctionDeclaration) {
                nodes.FunctionDeclaration = function (node) {
                    validateComma(node, "params");
                };
            }
            if (!exceptions.FunctionExpression) {
                nodes.FunctionExpression = function (node) {
                    validateComma(node, "params");
                };
            }
            if (!exceptions.ArrowFunctionExpression) {
                nodes.ArrowFunctionExpression = function (node) {
                    validateComma(node, "params");
                };
            }
            if (!exceptions.CallExpression) {
                nodes.CallExpression = function (node) {
                    validateComma(node, "arguments");
                };
            }
            if (!exceptions.ImportDeclaration) {
                nodes.ImportDeclaration = function (node) {
                    validateComma(node, "specifiers");
                };
            }
            if (!exceptions.NewExpression) {
                nodes.NewExpression = function (node) {
                    validateComma(node, "arguments");
                };
            }
            return nodes;
        }
    };
},
/* eslint/lib/rules/complexity.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const { upperCaseFirst } = require(418) /* ../shared/string-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce a maximum cyclomatic complexity allowed in a program",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/complexity"
            },
            schema: [
                {
                    oneOf: [
                        {
                            type: "integer",
                            minimum: 0
                        },
                        {
                            type: "object",
                            properties: {
                                maximum: {
                                    type: "integer",
                                    minimum: 0
                                },
                                max: {
                                    type: "integer",
                                    minimum: 0
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                complex: "{{name}} has a complexity of {{complexity}}. Maximum allowed is {{max}}."
            }
        },
        create(context) {
            const option = context.options[0];
            let THRESHOLD = 20;
            if (typeof option === "object" &&
                (Object.prototype.hasOwnProperty.call(option, "maximum") || Object.prototype.hasOwnProperty.call(option, "max"))) {
                THRESHOLD = option.maximum || option.max;
            }
            else if (typeof option === "number") {
                THRESHOLD = option;
            }
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            // Using a stack to store complexity (handling nested functions)
            const fns = [];
            /**
             * When parsing a new function, store it in our function stack
             * @returns {void}
             * @private
             */
            function startFunction() {
                fns.push(1);
            }
            /**
             * Evaluate the node at the end of function
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function endFunction(node) {
                const name = upperCaseFirst(astUtils.getFunctionNameWithKind(node));
                const complexity = fns.pop();
                if (complexity > THRESHOLD) {
                    context.report({
                        node,
                        messageId: "complex",
                        data: { name, complexity, max: THRESHOLD }
                    });
                }
            }
            /**
             * Increase the complexity of the function in context
             * @returns {void}
             * @private
             */
            function increaseComplexity() {
                if (fns.length) {
                    fns[fns.length - 1]++;
                }
            }
            /**
             * Increase the switch complexity in context
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function increaseSwitchComplexity(node) {
                // Avoiding `default`
                if (node.test) {
                    increaseComplexity();
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                FunctionDeclaration: startFunction,
                FunctionExpression: startFunction,
                ArrowFunctionExpression: startFunction,
                "FunctionDeclaration:exit": endFunction,
                "FunctionExpression:exit": endFunction,
                "ArrowFunctionExpression:exit": endFunction,
                CatchClause: increaseComplexity,
                ConditionalExpression: increaseComplexity,
                LogicalExpression: increaseComplexity,
                ForStatement: increaseComplexity,
                ForInStatement: increaseComplexity,
                ForOfStatement: increaseComplexity,
                IfStatement: increaseComplexity,
                SwitchCase: increaseSwitchComplexity,
                WhileStatement: increaseComplexity,
                DoWhileStatement: increaseComplexity,
                AssignmentExpression(node) {
                    if (astUtils.isLogicalAssignmentOperator(node.operator)) {
                        increaseComplexity();
                    }
                }
            };
        }
    };
},
/* eslint/lib/shared/string-utils.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Converts the first letter of a string to uppercase.
     * @param {string} string The string to operate on
     * @returns {string} The converted string
     */
    function upperCaseFirst(string) {
        if (string.length <= 1) {
            return string.toUpperCase();
        }
        return string[0].toUpperCase() + string.slice(1);
    }
    module.exports = {
        upperCaseFirst
    };
},
/* eslint/lib/rules/computed-property-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing inside computed property brackets",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/computed-property-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never"]
                },
                {
                    type: "object",
                    properties: {
                        enforceForClassMembers: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedSpaceBefore: "There should be no space before '{{tokenValue}}'.",
                unexpectedSpaceAfter: "There should be no space after '{{tokenValue}}'.",
                missingSpaceBefore: "A space is required before '{{tokenValue}}'.",
                missingSpaceAfter: "A space is required after '{{tokenValue}}'."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const propertyNameMustBeSpaced = context.options[0] === "always"; // default is "never"
            const enforceForClassMembers = !context.options[1] || context.options[1].enforceForClassMembers;
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Reports that there shouldn't be a space after the first token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @param {Token} tokenAfter The token after `token`.
             * @returns {void}
             */
            function reportNoBeginningSpace(node, token, tokenAfter) {
                context.report({
                    node,
                    loc: { start: token.loc.end, end: tokenAfter.loc.start },
                    messageId: "unexpectedSpaceAfter",
                    data: {
                        tokenValue: token.value
                    },
                    fix(fixer) {
                        return fixer.removeRange([token.range[1], tokenAfter.range[0]]);
                    }
                });
            }
            /**
             * Reports that there shouldn't be a space before the last token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @param {Token} tokenBefore The token before `token`.
             * @returns {void}
             */
            function reportNoEndingSpace(node, token, tokenBefore) {
                context.report({
                    node,
                    loc: { start: tokenBefore.loc.end, end: token.loc.start },
                    messageId: "unexpectedSpaceBefore",
                    data: {
                        tokenValue: token.value
                    },
                    fix(fixer) {
                        return fixer.removeRange([tokenBefore.range[1], token.range[0]]);
                    }
                });
            }
            /**
             * Reports that there should be a space after the first token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportRequiredBeginningSpace(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "missingSpaceAfter",
                    data: {
                        tokenValue: token.value
                    },
                    fix(fixer) {
                        return fixer.insertTextAfter(token, " ");
                    }
                });
            }
            /**
             * Reports that there should be a space before the last token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportRequiredEndingSpace(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "missingSpaceBefore",
                    data: {
                        tokenValue: token.value
                    },
                    fix(fixer) {
                        return fixer.insertTextBefore(token, " ");
                    }
                });
            }
            /**
             * Returns a function that checks the spacing of a node on the property name
             * that was passed in.
             * @param {string} propertyName The property on the node to check for spacing
             * @returns {Function} A function that will check spacing on a node
             */
            function checkSpacing(propertyName) {
                return function (node) {
                    if (!node.computed) {
                        return;
                    }
                    const property = node[propertyName];
                    const before = sourceCode.getTokenBefore(property, astUtils.isOpeningBracketToken), first = sourceCode.getTokenAfter(before, { includeComments: true }), after = sourceCode.getTokenAfter(property, astUtils.isClosingBracketToken), last = sourceCode.getTokenBefore(after, { includeComments: true });
                    if (astUtils.isTokenOnSameLine(before, first)) {
                        if (propertyNameMustBeSpaced) {
                            if (!sourceCode.isSpaceBetweenTokens(before, first) && astUtils.isTokenOnSameLine(before, first)) {
                                reportRequiredBeginningSpace(node, before);
                            }
                        }
                        else {
                            if (sourceCode.isSpaceBetweenTokens(before, first)) {
                                reportNoBeginningSpace(node, before, first);
                            }
                        }
                    }
                    if (astUtils.isTokenOnSameLine(last, after)) {
                        if (propertyNameMustBeSpaced) {
                            if (!sourceCode.isSpaceBetweenTokens(last, after) && astUtils.isTokenOnSameLine(last, after)) {
                                reportRequiredEndingSpace(node, after);
                            }
                        }
                        else {
                            if (sourceCode.isSpaceBetweenTokens(last, after)) {
                                reportNoEndingSpace(node, after, last);
                            }
                        }
                    }
                };
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            const listeners = {
                Property: checkSpacing("key"),
                MemberExpression: checkSpacing("property")
            };
            if (enforceForClassMembers) {
                listeners.MethodDefinition = checkSpacing("key");
            }
            return listeners;
        }
    };
},
/* eslint/lib/rules/consistent-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const { upperCaseFirst } = require(418) /* ../shared/string-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given code path segment is unreachable.
     * @param {CodePathSegment} segment A CodePathSegment to check.
     * @returns {boolean} `true` if the segment is unreachable.
     */
    function isUnreachable(segment) {
        return !segment.reachable;
    }
    /**
     * Checks whether a given node is a `constructor` method in an ES6 class
     * @param {ASTNode} node A node to check
     * @returns {boolean} `true` if the node is a `constructor` method
     */
    function isClassConstructor(node) {
        return node.type === "FunctionExpression" &&
            node.parent &&
            node.parent.type === "MethodDefinition" &&
            node.parent.kind === "constructor";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require `return` statements to either always or never specify values",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/consistent-return"
            },
            schema: [{
                    type: "object",
                    properties: {
                        treatUndefinedAsUnspecified: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                missingReturn: "Expected to return a value at the end of {{name}}.",
                missingReturnValue: "{{name}} expected a return value.",
                unexpectedReturnValue: "{{name}} expected no return value."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const treatUndefinedAsUnspecified = options.treatUndefinedAsUnspecified === true;
            let funcInfo = null;
            /**
             * Checks whether of not the implicit returning is consistent if the last
             * code path segment is reachable.
             * @param {ASTNode} node A program/function node to check.
             * @returns {void}
             */
            function checkLastSegment(node) {
                let loc, name;
                /*
                 * Skip if it expected no return value or unreachable.
                 * When unreachable, all paths are returned or thrown.
                 */
                if (!funcInfo.hasReturnValue ||
                    funcInfo.codePath.currentSegments.every(isUnreachable) ||
                    astUtils.isES5Constructor(node) ||
                    isClassConstructor(node)) {
                    return;
                }
                // Adjust a location and a message.
                if (node.type === "Program") {
                    // The head of program.
                    loc = { line: 1, column: 0 };
                    name = "program";
                }
                else if (node.type === "ArrowFunctionExpression") {
                    // `=>` token
                    loc = context.getSourceCode().getTokenBefore(node.body, astUtils.isArrowToken).loc;
                }
                else if (node.parent.type === "MethodDefinition" ||
                    (node.parent.type === "Property" && node.parent.method)) {
                    // Method name.
                    loc = node.parent.key.loc;
                }
                else {
                    // Function name or `function` keyword.
                    loc = (node.id || context.getSourceCode().getFirstToken(node)).loc;
                }
                if (!name) {
                    name = astUtils.getFunctionNameWithKind(node);
                }
                // Reports.
                context.report({
                    node,
                    loc,
                    messageId: "missingReturn",
                    data: { name }
                });
            }
            return {
                // Initializes/Disposes state of each code path.
                onCodePathStart(codePath, node) {
                    funcInfo = {
                        upper: funcInfo,
                        codePath,
                        hasReturn: false,
                        hasReturnValue: false,
                        messageId: "",
                        node
                    };
                },
                onCodePathEnd() {
                    funcInfo = funcInfo.upper;
                },
                // Reports a given return statement if it's inconsistent.
                ReturnStatement(node) {
                    const argument = node.argument;
                    let hasReturnValue = Boolean(argument);
                    if (treatUndefinedAsUnspecified && hasReturnValue) {
                        hasReturnValue = !astUtils.isSpecificId(argument, "undefined") && argument.operator !== "void";
                    }
                    if (!funcInfo.hasReturn) {
                        funcInfo.hasReturn = true;
                        funcInfo.hasReturnValue = hasReturnValue;
                        funcInfo.messageId = hasReturnValue ? "missingReturnValue" : "unexpectedReturnValue";
                        funcInfo.data = {
                            name: funcInfo.node.type === "Program"
                                ? "Program"
                                : upperCaseFirst(astUtils.getFunctionNameWithKind(funcInfo.node))
                        };
                    }
                    else if (funcInfo.hasReturnValue !== hasReturnValue) {
                        context.report({
                            node,
                            messageId: funcInfo.messageId,
                            data: funcInfo.data
                        });
                    }
                },
                // Reports a given program/function if the implicit returning is not consistent.
                "Program:exit": checkLastSegment,
                "FunctionDeclaration:exit": checkLastSegment,
                "FunctionExpression:exit": checkLastSegment,
                "ArrowFunctionExpression:exit": checkLastSegment
            };
        }
    };
},
/* eslint/lib/rules/consistent-this.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce consistent naming when capturing the current execution context",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/consistent-this"
            },
            schema: {
                type: "array",
                items: {
                    type: "string",
                    minLength: 1
                },
                uniqueItems: true
            },
            messages: {
                aliasNotAssignedToThis: "Designated alias '{{name}}' is not assigned to 'this'.",
                unexpectedAlias: "Unexpected alias '{{name}}' for 'this'."
            }
        },
        create(context) {
            let aliases = [];
            if (context.options.length === 0) {
                aliases.push("that");
            }
            else {
                aliases = context.options;
            }
            /**
             * Reports that a variable declarator or assignment expression is assigning
             * a non-'this' value to the specified alias.
             * @param {ASTNode} node The assigning node.
             * @param {string}  name the name of the alias that was incorrectly used.
             * @returns {void}
             */
            function reportBadAssignment(node, name) {
                context.report({ node, messageId: "aliasNotAssignedToThis", data: { name } });
            }
            /**
             * Checks that an assignment to an identifier only assigns 'this' to the
             * appropriate alias, and the alias is only assigned to 'this'.
             * @param {ASTNode} node The assigning node.
             * @param {Identifier} name The name of the variable assigned to.
             * @param {Expression} value The value of the assignment.
             * @returns {void}
             */
            function checkAssignment(node, name, value) {
                const isThis = value.type === "ThisExpression";
                if (aliases.indexOf(name) !== -1) {
                    if (!isThis || node.operator && node.operator !== "=") {
                        reportBadAssignment(node, name);
                    }
                }
                else if (isThis) {
                    context.report({ node, messageId: "unexpectedAlias", data: { name } });
                }
            }
            /**
             * Ensures that a variable declaration of the alias in a program or function
             * is assigned to the correct value.
             * @param {string} alias alias the check the assignment of.
             * @param {Object} scope scope of the current code we are checking.
             * @private
             * @returns {void}
             */
            function checkWasAssigned(alias, scope) {
                const variable = scope.set.get(alias);
                if (!variable) {
                    return;
                }
                if (variable.defs.some(def => def.node.type === "VariableDeclarator" &&
                    def.node.init !== null)) {
                    return;
                }
                /*
                 * The alias has been declared and not assigned: check it was
                 * assigned later in the same scope.
                 */
                if (!variable.references.some(reference => {
                    const write = reference.writeExpr;
                    return (reference.from === scope &&
                        write && write.type === "ThisExpression" &&
                        write.parent.operator === "=");
                })) {
                    variable.defs.map(def => def.node).forEach(node => {
                        reportBadAssignment(node, alias);
                    });
                }
            }
            /**
             * Check each alias to ensure that is was assigned to the correct value.
             * @returns {void}
             */
            function ensureWasAssigned() {
                const scope = context.getScope();
                aliases.forEach(alias => {
                    checkWasAssigned(alias, scope);
                });
            }
            return {
                "Program:exit": ensureWasAssigned,
                "FunctionExpression:exit": ensureWasAssigned,
                "FunctionDeclaration:exit": ensureWasAssigned,
                VariableDeclarator(node) {
                    const id = node.id;
                    const isDestructuring = id.type === "ArrayPattern" || id.type === "ObjectPattern";
                    if (node.init !== null && !isDestructuring) {
                        checkAssignment(node, id.name, node.init);
                    }
                },
                AssignmentExpression(node) {
                    if (node.left.type === "Identifier") {
                        checkAssignment(node, node.left.name, node.right);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/constructor-super.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether a given code path segment is reachable or not.
     * @param {CodePathSegment} segment A code path segment to check.
     * @returns {boolean} `true` if the segment is reachable.
     */
    function isReachable(segment) {
        return segment.reachable;
    }
    /**
     * Checks whether or not a given node is a constructor.
     * @param {ASTNode} node A node to check. This node type is one of
     *   `Program`, `FunctionDeclaration`, `FunctionExpression`, and
     *   `ArrowFunctionExpression`.
     * @returns {boolean} `true` if the node is a constructor.
     */
    function isConstructorFunction(node) {
        return (node.type === "FunctionExpression" &&
            node.parent.type === "MethodDefinition" &&
            node.parent.kind === "constructor");
    }
    /**
     * Checks whether a given node can be a constructor or not.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node can be a constructor.
     */
    function isPossibleConstructor(node) {
        if (!node) {
            return false;
        }
        switch (node.type) {
            case "ClassExpression":
            case "FunctionExpression":
            case "ThisExpression":
            case "MemberExpression":
            case "CallExpression":
            case "NewExpression":
            case "ChainExpression":
            case "YieldExpression":
            case "TaggedTemplateExpression":
            case "MetaProperty":
                return true;
            case "Identifier":
                return node.name !== "undefined";
            case "AssignmentExpression":
                if (["=", "&&="].includes(node.operator)) {
                    return isPossibleConstructor(node.right);
                }
                if (["||=", "??="].includes(node.operator)) {
                    return (isPossibleConstructor(node.left) ||
                        isPossibleConstructor(node.right));
                }
                /**
                 * All other assignment operators are mathematical assignment operators (arithmetic or bitwise).
                 * An assignment expression with a mathematical operator can either evaluate to a primitive value,
                 * or throw, depending on the operands. Thus, it cannot evaluate to a constructor function.
                 */
                return false;
            case "LogicalExpression":
                /*
                 * If the && operator short-circuits, the left side was falsy and therefore not a constructor, and if
                 * it doesn't short-circuit, it takes the value from the right side, so the right side must always be a
                 * possible constructor. A future improvement could verify that the left side could be truthy by
                 * excluding falsy literals.
                 */
                if (node.operator === "&&") {
                    return isPossibleConstructor(node.right);
                }
                return (isPossibleConstructor(node.left) ||
                    isPossibleConstructor(node.right));
            case "ConditionalExpression":
                return (isPossibleConstructor(node.alternate) ||
                    isPossibleConstructor(node.consequent));
            case "SequenceExpression": {
                const lastExpression = node.expressions[node.expressions.length - 1];
                return isPossibleConstructor(lastExpression);
            }
            default:
                return false;
        }
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "require `super()` calls in constructors",
                category: "ECMAScript 6",
                recommended: true,
                url: "https://eslint.org/docs/rules/constructor-super"
            },
            schema: [],
            messages: {
                missingSome: "Lacked a call of 'super()' in some code paths.",
                missingAll: "Expected to call 'super()'.",
                duplicate: "Unexpected duplicate 'super()'.",
                badSuper: "Unexpected 'super()' because 'super' is not a constructor.",
                unexpected: "Unexpected 'super()'."
            }
        },
        create(context) {
            /*
             * {{hasExtends: boolean, scope: Scope, codePath: CodePath}[]}
             * Information for each constructor.
             * - upper:      Information of the upper constructor.
             * - hasExtends: A flag which shows whether own class has a valid `extends`
             *               part.
             * - scope:      The scope of own class.
             * - codePath:   The code path object of the constructor.
             */
            let funcInfo = null;
            /*
             * {Map<string, {calledInSomePaths: boolean, calledInEveryPaths: boolean}>}
             * Information for each code path segment.
             * - calledInSomePaths:  A flag of be called `super()` in some code paths.
             * - calledInEveryPaths: A flag of be called `super()` in all code paths.
             * - validNodes:
             */
            let segInfoMap = Object.create(null);
            /**
             * Gets the flag which shows `super()` is called in some paths.
             * @param {CodePathSegment} segment A code path segment to get.
             * @returns {boolean} The flag which shows `super()` is called in some paths
             */
            function isCalledInSomePath(segment) {
                return segment.reachable && segInfoMap[segment.id].calledInSomePaths;
            }
            /**
             * Gets the flag which shows `super()` is called in all paths.
             * @param {CodePathSegment} segment A code path segment to get.
             * @returns {boolean} The flag which shows `super()` is called in all paths.
             */
            function isCalledInEveryPath(segment) {
                /*
                 * If specific segment is the looped segment of the current segment,
                 * skip the segment.
                 * If not skipped, this never becomes true after a loop.
                 */
                if (segment.nextSegments.length === 1 &&
                    segment.nextSegments[0].isLoopedPrevSegment(segment)) {
                    return true;
                }
                return segment.reachable && segInfoMap[segment.id].calledInEveryPaths;
            }
            return {
                /**
                 * Stacks a constructor information.
                 * @param {CodePath} codePath A code path which was started.
                 * @param {ASTNode} node The current node.
                 * @returns {void}
                 */
                onCodePathStart(codePath, node) {
                    if (isConstructorFunction(node)) {
                        // Class > ClassBody > MethodDefinition > FunctionExpression
                        const classNode = node.parent.parent.parent;
                        const superClass = classNode.superClass;
                        funcInfo = {
                            upper: funcInfo,
                            isConstructor: true,
                            hasExtends: Boolean(superClass),
                            superIsConstructor: isPossibleConstructor(superClass),
                            codePath
                        };
                    }
                    else {
                        funcInfo = {
                            upper: funcInfo,
                            isConstructor: false,
                            hasExtends: false,
                            superIsConstructor: false,
                            codePath
                        };
                    }
                },
                /**
                 * Pops a constructor information.
                 * And reports if `super()` lacked.
                 * @param {CodePath} codePath A code path which was ended.
                 * @param {ASTNode} node The current node.
                 * @returns {void}
                 */
                onCodePathEnd(codePath, node) {
                    const hasExtends = funcInfo.hasExtends;
                    // Pop.
                    funcInfo = funcInfo.upper;
                    if (!hasExtends) {
                        return;
                    }
                    // Reports if `super()` lacked.
                    const segments = codePath.returnedSegments;
                    const calledInEveryPaths = segments.every(isCalledInEveryPath);
                    const calledInSomePaths = segments.some(isCalledInSomePath);
                    if (!calledInEveryPaths) {
                        context.report({
                            messageId: calledInSomePaths
                                ? "missingSome"
                                : "missingAll",
                            node: node.parent
                        });
                    }
                },
                /**
                 * Initialize information of a given code path segment.
                 * @param {CodePathSegment} segment A code path segment to initialize.
                 * @returns {void}
                 */
                onCodePathSegmentStart(segment) {
                    if (!(funcInfo && funcInfo.isConstructor && funcInfo.hasExtends)) {
                        return;
                    }
                    // Initialize info.
                    const info = segInfoMap[segment.id] = {
                        calledInSomePaths: false,
                        calledInEveryPaths: false,
                        validNodes: []
                    };
                    // When there are previous segments, aggregates these.
                    const prevSegments = segment.prevSegments;
                    if (prevSegments.length > 0) {
                        info.calledInSomePaths = prevSegments.some(isCalledInSomePath);
                        info.calledInEveryPaths = prevSegments.every(isCalledInEveryPath);
                    }
                },
                /**
                 * Update information of the code path segment when a code path was
                 * looped.
                 * @param {CodePathSegment} fromSegment The code path segment of the
                 *      end of a loop.
                 * @param {CodePathSegment} toSegment A code path segment of the head
                 *      of a loop.
                 * @returns {void}
                 */
                onCodePathSegmentLoop(fromSegment, toSegment) {
                    if (!(funcInfo && funcInfo.isConstructor && funcInfo.hasExtends)) {
                        return;
                    }
                    // Update information inside of the loop.
                    const isRealLoop = toSegment.prevSegments.length >= 2;
                    funcInfo.codePath.traverseSegments({ first: toSegment, last: fromSegment }, segment => {
                        const info = segInfoMap[segment.id];
                        const prevSegments = segment.prevSegments;
                        // Updates flags.
                        info.calledInSomePaths = prevSegments.some(isCalledInSomePath);
                        info.calledInEveryPaths = prevSegments.every(isCalledInEveryPath);
                        // If flags become true anew, reports the valid nodes.
                        if (info.calledInSomePaths || isRealLoop) {
                            const nodes = info.validNodes;
                            info.validNodes = [];
                            for (let i = 0; i < nodes.length; ++i) {
                                const node = nodes[i];
                                context.report({
                                    messageId: "duplicate",
                                    node
                                });
                            }
                        }
                    });
                },
                /**
                 * Checks for a call of `super()`.
                 * @param {ASTNode} node A CallExpression node to check.
                 * @returns {void}
                 */
                "CallExpression:exit"(node) {
                    if (!(funcInfo && funcInfo.isConstructor)) {
                        return;
                    }
                    // Skips except `super()`.
                    if (node.callee.type !== "Super") {
                        return;
                    }
                    // Reports if needed.
                    if (funcInfo.hasExtends) {
                        const segments = funcInfo.codePath.currentSegments;
                        let duplicate = false;
                        let info = null;
                        for (let i = 0; i < segments.length; ++i) {
                            const segment = segments[i];
                            if (segment.reachable) {
                                info = segInfoMap[segment.id];
                                duplicate = duplicate || info.calledInSomePaths;
                                info.calledInSomePaths = info.calledInEveryPaths = true;
                            }
                        }
                        if (info) {
                            if (duplicate) {
                                context.report({
                                    messageId: "duplicate",
                                    node
                                });
                            }
                            else if (!funcInfo.superIsConstructor) {
                                context.report({
                                    messageId: "badSuper",
                                    node
                                });
                            }
                            else {
                                info.validNodes.push(node);
                            }
                        }
                    }
                    else if (funcInfo.codePath.currentSegments.some(isReachable)) {
                        context.report({
                            messageId: "unexpected",
                            node
                        });
                    }
                },
                /**
                 * Set the mark to the returned path as `super()` was called.
                 * @param {ASTNode} node A ReturnStatement node to check.
                 * @returns {void}
                 */
                ReturnStatement(node) {
                    if (!(funcInfo && funcInfo.isConstructor && funcInfo.hasExtends)) {
                        return;
                    }
                    // Skips if no argument.
                    if (!node.argument) {
                        return;
                    }
                    // Returning argument is a substitute of 'super()'.
                    const segments = funcInfo.codePath.currentSegments;
                    for (let i = 0; i < segments.length; ++i) {
                        const segment = segments[i];
                        if (segment.reachable) {
                            const info = segInfoMap[segment.id];
                            info.calledInSomePaths = info.calledInEveryPaths = true;
                        }
                    }
                },
                /**
                 * Resets state.
                 * @returns {void}
                 */
                "Program:exit"() {
                    segInfoMap = Object.create(null);
                }
            };
        }
    };
},
/* eslint/lib/rules/curly.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce consistent brace style for all control statements",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/curly"
            },
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["all"]
                            }
                        ],
                        minItems: 0,
                        maxItems: 1
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["multi", "multi-line", "multi-or-nest"]
                            },
                            {
                                enum: ["consistent"]
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    }
                ]
            },
            fixable: "code",
            messages: {
                missingCurlyAfter: "Expected { after '{{name}}'.",
                missingCurlyAfterCondition: "Expected { after '{{name}}' condition.",
                unexpectedCurlyAfter: "Unnecessary { after '{{name}}'.",
                unexpectedCurlyAfterCondition: "Unnecessary { after '{{name}}' condition."
            }
        },
        create(context) {
            const multiOnly = (context.options[0] === "multi");
            const multiLine = (context.options[0] === "multi-line");
            const multiOrNest = (context.options[0] === "multi-or-nest");
            const consistent = (context.options[1] === "consistent");
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Determines if a given node is a one-liner that's on the same line as it's preceding code.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} True if the node is a one-liner that's on the same line as it's preceding code.
             * @private
             */
            function isCollapsedOneLiner(node) {
                const before = sourceCode.getTokenBefore(node);
                const last = sourceCode.getLastToken(node);
                const lastExcludingSemicolon = astUtils.isSemicolonToken(last) ? sourceCode.getTokenBefore(last) : last;
                return before.loc.start.line === lastExcludingSemicolon.loc.end.line;
            }
            /**
             * Determines if a given node is a one-liner.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} True if the node is a one-liner.
             * @private
             */
            function isOneLiner(node) {
                if (node.type === "EmptyStatement") {
                    return true;
                }
                const first = sourceCode.getFirstToken(node);
                const last = sourceCode.getLastToken(node);
                const lastExcludingSemicolon = astUtils.isSemicolonToken(last) ? sourceCode.getTokenBefore(last) : last;
                return first.loc.start.line === lastExcludingSemicolon.loc.end.line;
            }
            /**
             * Determines if the given node is a lexical declaration (let, const, function, or class)
             * @param {ASTNode} node The node to check
             * @returns {boolean} True if the node is a lexical declaration
             * @private
             */
            function isLexicalDeclaration(node) {
                if (node.type === "VariableDeclaration") {
                    return node.kind === "const" || node.kind === "let";
                }
                return node.type === "FunctionDeclaration" || node.type === "ClassDeclaration";
            }
            /**
             * Checks if the given token is an `else` token or not.
             * @param {Token} token The token to check.
             * @returns {boolean} `true` if the token is an `else` token.
             */
            function isElseKeywordToken(token) {
                return token.value === "else" && token.type === "Keyword";
            }
            /**
             * Determines whether the given node has an `else` keyword token as the first token after.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} `true` if the node is followed by an `else` keyword token.
             */
            function isFollowedByElseKeyword(node) {
                const nextToken = sourceCode.getTokenAfter(node);
                return Boolean(nextToken) && isElseKeywordToken(nextToken);
            }
            /**
             * Determines if a semicolon needs to be inserted after removing a set of curly brackets, in order to avoid a SyntaxError.
             * @param {Token} closingBracket The } token
             * @returns {boolean} `true` if a semicolon needs to be inserted after the last statement in the block.
             */
            function needsSemicolon(closingBracket) {
                const tokenBefore = sourceCode.getTokenBefore(closingBracket);
                const tokenAfter = sourceCode.getTokenAfter(closingBracket);
                const lastBlockNode = sourceCode.getNodeByRangeIndex(tokenBefore.range[0]);
                if (astUtils.isSemicolonToken(tokenBefore)) {
                    // If the last statement already has a semicolon, don't add another one.
                    return false;
                }
                if (!tokenAfter) {
                    // If there are no statements after this block, there is no need to add a semicolon.
                    return false;
                }
                if (lastBlockNode.type === "BlockStatement" && lastBlockNode.parent.type !== "FunctionExpression" && lastBlockNode.parent.type !== "ArrowFunctionExpression") {
                    /*
                     * If the last node surrounded by curly brackets is a BlockStatement (other than a FunctionExpression or an ArrowFunctionExpression),
                     * don't insert a semicolon. Otherwise, the semicolon would be parsed as a separate statement, which would cause
                     * a SyntaxError if it was followed by `else`.
                     */
                    return false;
                }
                if (tokenBefore.loc.end.line === tokenAfter.loc.start.line) {
                    // If the next token is on the same line, insert a semicolon.
                    return true;
                }
                if (/^[([/`+-]/u.test(tokenAfter.value)) {
                    // If the next token starts with a character that would disrupt ASI, insert a semicolon.
                    return true;
                }
                if (tokenBefore.type === "Punctuator" && (tokenBefore.value === "++" || tokenBefore.value === "--")) {
                    // If the last token is ++ or --, insert a semicolon to avoid disrupting ASI.
                    return true;
                }
                // Otherwise, do not insert a semicolon.
                return false;
            }
            /**
             * Determines whether the code represented by the given node contains an `if` statement
             * that would become associated with an `else` keyword directly appended to that code.
             *
             * Examples where it returns `true`:
             *
             *    if (a)
             *        foo();
             *
             *    if (a) {
             *        foo();
             *    }
             *
             *    if (a)
             *        foo();
             *    else if (b)
             *        bar();
             *
             *    while (a)
             *        if (b)
             *            if(c)
             *                foo();
             *            else
             *                bar();
             *
             * Examples where it returns `false`:
             *
             *    if (a)
             *        foo();
             *    else
             *        bar();
             *
             *    while (a) {
             *        if (b)
             *            if(c)
             *                foo();
             *            else
             *                bar();
             *    }
             *
             *    while (a)
             *        if (b) {
             *            if(c)
             *                foo();
             *        }
             *        else
             *            bar();
             * @param {ASTNode} node Node representing the code to check.
             * @returns {boolean} `true` if an `if` statement within the code would become associated with an `else` appended to that code.
             */
            function hasUnsafeIf(node) {
                switch (node.type) {
                    case "IfStatement":
                        if (!node.alternate) {
                            return true;
                        }
                        return hasUnsafeIf(node.alternate);
                    case "ForStatement":
                    case "ForInStatement":
                    case "ForOfStatement":
                    case "LabeledStatement":
                    case "WithStatement":
                    case "WhileStatement":
                        return hasUnsafeIf(node.body);
                    default:
                        return false;
                }
            }
            /**
             * Determines whether the existing curly braces around the single statement are necessary to preserve the semantics of the code.
             * The braces, which make the given block body, are necessary in either of the following situations:
             *
             * 1. The statement is a lexical declaration.
             * 2. Without the braces, an `if` within the statement would become associated with an `else` after the closing brace:
             *
             *     if (a) {
             *         if (b)
             *             foo();
             *     }
             *     else
             *         bar();
             *
             *     if (a)
             *         while (b)
             *             while (c) {
             *                 while (d)
             *                     if (e)
             *                         while(f)
             *                             foo();
             *            }
             *     else
             *         bar();
             * @param {ASTNode} node `BlockStatement` body with exactly one statement directly inside. The statement can have its own nested statements.
             * @returns {boolean} `true` if the braces are necessary - removing them (replacing the given `BlockStatement` body with its single statement content)
             * would change the semantics of the code or produce a syntax error.
             */
            function areBracesNecessary(node) {
                const statement = node.body[0];
                return isLexicalDeclaration(statement) ||
                    hasUnsafeIf(statement) && isFollowedByElseKeyword(node);
            }
            /**
             * Prepares to check the body of a node to see if it's a block statement.
             * @param {ASTNode} node The node to report if there's a problem.
             * @param {ASTNode} body The body node to check for blocks.
             * @param {string} name The name to report if there's a problem.
             * @param {{ condition: boolean }} opts Options to pass to the report functions
             * @returns {Object} a prepared check object, with "actual", "expected", "check" properties.
             *   "actual" will be `true` or `false` whether the body is already a block statement.
             *   "expected" will be `true` or `false` if the body should be a block statement or not, or
             *   `null` if it doesn't matter, depending on the rule options. It can be modified to change
             *   the final behavior of "check".
             *   "check" will be a function reporting appropriate problems depending on the other
             *   properties.
             */
            function prepareCheck(node, body, name, opts) {
                const hasBlock = (body.type === "BlockStatement");
                let expected = null;
                if (hasBlock && (body.body.length !== 1 || areBracesNecessary(body))) {
                    expected = true;
                }
                else if (multiOnly) {
                    expected = false;
                }
                else if (multiLine) {
                    if (!isCollapsedOneLiner(body)) {
                        expected = true;
                    }
                    // otherwise, the body is allowed to have braces or not to have braces
                }
                else if (multiOrNest) {
                    if (hasBlock) {
                        const statement = body.body[0];
                        const leadingCommentsInBlock = sourceCode.getCommentsBefore(statement);
                        expected = !isOneLiner(statement) || leadingCommentsInBlock.length > 0;
                    }
                    else {
                        expected = !isOneLiner(body);
                    }
                }
                else {
                    // default "all"
                    expected = true;
                }
                return {
                    actual: hasBlock,
                    expected,
                    check() {
                        if (this.expected !== null && this.expected !== this.actual) {
                            if (this.expected) {
                                context.report({
                                    node,
                                    loc: body.loc,
                                    messageId: opts && opts.condition ? "missingCurlyAfterCondition" : "missingCurlyAfter",
                                    data: {
                                        name
                                    },
                                    fix: fixer => fixer.replaceText(body, `{${sourceCode.getText(body)}}`)
                                });
                            }
                            else {
                                context.report({
                                    node,
                                    loc: body.loc,
                                    messageId: opts && opts.condition ? "unexpectedCurlyAfterCondition" : "unexpectedCurlyAfter",
                                    data: {
                                        name
                                    },
                                    fix(fixer) {
                                        /*
                                         * `do while` expressions sometimes need a space to be inserted after `do`.
                                         * e.g. `do{foo()} while (bar)` should be corrected to `do foo() while (bar)`
                                         */
                                        const needsPrecedingSpace = node.type === "DoWhileStatement" &&
                                            sourceCode.getTokenBefore(body).range[1] === body.range[0] &&
                                            !astUtils.canTokensBeAdjacent("do", sourceCode.getFirstToken(body, { skip: 1 }));
                                        const openingBracket = sourceCode.getFirstToken(body);
                                        const closingBracket = sourceCode.getLastToken(body);
                                        const lastTokenInBlock = sourceCode.getTokenBefore(closingBracket);
                                        if (needsSemicolon(closingBracket)) {
                                            /*
                                             * If removing braces would cause a SyntaxError due to multiple statements on the same line (or
                                             * change the semantics of the code due to ASI), don't perform a fix.
                                             */
                                            return null;
                                        }
                                        const resultingBodyText = sourceCode.getText().slice(openingBracket.range[1], lastTokenInBlock.range[0]) +
                                            sourceCode.getText(lastTokenInBlock) +
                                            sourceCode.getText().slice(lastTokenInBlock.range[1], closingBracket.range[0]);
                                        return fixer.replaceText(body, (needsPrecedingSpace ? " " : "") + resultingBodyText);
                                    }
                                });
                            }
                        }
                    }
                };
            }
            /**
             * Prepares to check the bodies of a "if", "else if" and "else" chain.
             * @param {ASTNode} node The first IfStatement node of the chain.
             * @returns {Object[]} prepared checks for each body of the chain. See `prepareCheck` for more
             *   information.
             */
            function prepareIfChecks(node) {
                const preparedChecks = [];
                for (let currentNode = node; currentNode; currentNode = currentNode.alternate) {
                    preparedChecks.push(prepareCheck(currentNode, currentNode.consequent, "if", { condition: true }));
                    if (currentNode.alternate && currentNode.alternate.type !== "IfStatement") {
                        preparedChecks.push(prepareCheck(currentNode, currentNode.alternate, "else"));
                        break;
                    }
                }
                if (consistent) {
                    /*
                     * If any node should have or already have braces, make sure they
                     * all have braces.
                     * If all nodes shouldn't have braces, make sure they don't.
                     */
                    const expected = preparedChecks.some(preparedCheck => {
                        if (preparedCheck.expected !== null) {
                            return preparedCheck.expected;
                        }
                        return preparedCheck.actual;
                    });
                    preparedChecks.forEach(preparedCheck => {
                        preparedCheck.expected = expected;
                    });
                }
                return preparedChecks;
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                IfStatement(node) {
                    const parent = node.parent;
                    const isElseIf = parent.type === "IfStatement" && parent.alternate === node;
                    if (!isElseIf) {
                        // This is a top `if`, check the whole `if-else-if` chain
                        prepareIfChecks(node).forEach(preparedCheck => {
                            preparedCheck.check();
                        });
                    }
                    // Skip `else if`, it's already checked (when the top `if` was visited)
                },
                WhileStatement(node) {
                    prepareCheck(node, node.body, "while", { condition: true }).check();
                },
                DoWhileStatement(node) {
                    prepareCheck(node, node.body, "do").check();
                },
                ForStatement(node) {
                    prepareCheck(node, node.body, "for", { condition: true }).check();
                },
                ForInStatement(node) {
                    prepareCheck(node, node.body, "for-in").check();
                },
                ForOfStatement(node) {
                    prepareCheck(node, node.body, "for-of").check();
                }
            };
        }
    };
},
/* eslint/lib/rules/default-case.js */ function _(require, module, exports, __esModule, __esExport) {
    const DEFAULT_COMMENT_PATTERN = /^no default$/iu;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require `default` cases in `switch` statements",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/default-case"
            },
            schema: [{
                    type: "object",
                    properties: {
                        commentPattern: {
                            type: "string"
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                missingDefaultCase: "Expected a default case."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const commentPattern = options.commentPattern
                ? new RegExp(options.commentPattern, "u")
                : DEFAULT_COMMENT_PATTERN;
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Shortcut to get last element of array
             * @param  {*[]} collection Array
             * @returns {*} Last element
             */
            function last(collection) {
                return collection[collection.length - 1];
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                SwitchStatement(node) {
                    if (!node.cases.length) {
                        /*
                         * skip check of empty switch because there is no easy way
                         * to extract comments inside it now
                         */
                        return;
                    }
                    const hasDefault = node.cases.some(v => v.test === null);
                    if (!hasDefault) {
                        let comment;
                        const lastCase = last(node.cases);
                        const comments = sourceCode.getCommentsAfter(lastCase);
                        if (comments.length) {
                            comment = last(comments);
                        }
                        if (!comment || !commentPattern.test(comment.value.trim())) {
                            context.report({ node, messageId: "missingDefaultCase" });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/default-case-last.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce default clauses in switch statements to be last",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/default-case-last"
            },
            schema: [],
            messages: {
                notLast: "Default clause should be the last clause."
            }
        },
        create(context) {
            return {
                SwitchStatement(node) {
                    const cases = node.cases, indexOfDefault = cases.findIndex(c => c.test === null);
                    if (indexOfDefault !== -1 && indexOfDefault !== cases.length - 1) {
                        const defaultClause = cases[indexOfDefault];
                        context.report({ node: defaultClause, messageId: "notLast" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/default-param-last.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce default parameters to be last",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/default-param-last"
            },
            schema: [],
            messages: {
                shouldBeLast: "Default parameters should be last."
            }
        },
        create(context) {
            // eslint-disable-next-line jsdoc/require-description
            /**
             * @param {ASTNode} node function node
             * @returns {void}
             */
            function handleFunction(node) {
                let hasSeenPlainParam = false;
                for (let i = node.params.length - 1; i >= 0; i -= 1) {
                    const param = node.params[i];
                    if (param.type !== "AssignmentPattern" &&
                        param.type !== "RestElement") {
                        hasSeenPlainParam = true;
                        continue;
                    }
                    if (hasSeenPlainParam && param.type === "AssignmentPattern") {
                        context.report({
                            node: param,
                            messageId: "shouldBeLast"
                        });
                    }
                }
            }
            return {
                FunctionDeclaration: handleFunction,
                FunctionExpression: handleFunction,
                ArrowFunctionExpression: handleFunction
            };
        }
    };
},
/* eslint/lib/rules/dot-location.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent newlines before and after dots",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/dot-location"
            },
            schema: [
                {
                    enum: ["object", "property"]
                }
            ],
            fixable: "code",
            messages: {
                expectedDotAfterObject: "Expected dot to be on same line as object.",
                expectedDotBeforeProperty: "Expected dot to be on same line as property."
            }
        },
        create(context) {
            const config = context.options[0];
            // default to onObject if no preference is passed
            const onObject = config === "object" || !config;
            const sourceCode = context.getSourceCode();
            /**
             * Reports if the dot between object and property is on the correct location.
             * @param {ASTNode} node The `MemberExpression` node.
             * @returns {void}
             */
            function checkDotLocation(node) {
                const property = node.property;
                const dotToken = sourceCode.getTokenBefore(property);
                if (onObject) {
                    // `obj` expression can be parenthesized, but those paren tokens are not a part of the `obj` node.
                    const tokenBeforeDot = sourceCode.getTokenBefore(dotToken);
                    if (!astUtils.isTokenOnSameLine(tokenBeforeDot, dotToken)) {
                        context.report({
                            node,
                            loc: dotToken.loc,
                            messageId: "expectedDotAfterObject",
                            *fix(fixer) {
                                if (dotToken.value.startsWith(".") && astUtils.isDecimalIntegerNumericToken(tokenBeforeDot)) {
                                    yield fixer.insertTextAfter(tokenBeforeDot, ` ${dotToken.value}`);
                                }
                                else {
                                    yield fixer.insertTextAfter(tokenBeforeDot, dotToken.value);
                                }
                                yield fixer.remove(dotToken);
                            }
                        });
                    }
                }
                else if (!astUtils.isTokenOnSameLine(dotToken, property)) {
                    context.report({
                        node,
                        loc: dotToken.loc,
                        messageId: "expectedDotBeforeProperty",
                        *fix(fixer) {
                            yield fixer.remove(dotToken);
                            yield fixer.insertTextBefore(property, dotToken.value);
                        }
                    });
                }
            }
            /**
             * Checks the spacing of the dot within a member expression.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             */
            function checkNode(node) {
                if (!node.computed) {
                    checkDotLocation(node);
                }
            }
            return {
                MemberExpression: checkNode
            };
        }
    };
},
/* eslint/lib/rules/dot-notation.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const keywords = require(429) /* ./utils/keywords */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const validIdentifier = /^[a-zA-Z_$][a-zA-Z0-9_$]*$/u;
    // `null` literal must be handled separately.
    const literalTypesToCheck = new Set(["string", "boolean"]);
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce dot notation whenever possible",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/dot-notation"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowKeywords: {
                            type: "boolean",
                            default: true
                        },
                        allowPattern: {
                            type: "string",
                            default: ""
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "code",
            messages: {
                useDot: "[{{key}}] is better written in dot notation.",
                useBrackets: ".{{key}} is a syntax error."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const allowKeywords = options.allowKeywords === void 0 || options.allowKeywords;
            const sourceCode = context.getSourceCode();
            let allowPattern;
            if (options.allowPattern) {
                allowPattern = new RegExp(options.allowPattern, "u");
            }
            /**
             * Check if the property is valid dot notation
             * @param {ASTNode} node The dot notation node
             * @param {string} value Value which is to be checked
             * @returns {void}
             */
            function checkComputedProperty(node, value) {
                if (validIdentifier.test(value) &&
                    (allowKeywords || keywords.indexOf(String(value)) === -1) &&
                    !(allowPattern && allowPattern.test(value))) {
                    const formattedValue = node.property.type === "Literal" ? JSON.stringify(value) : `\`${value}\``;
                    context.report({
                        node: node.property,
                        messageId: "useDot",
                        data: {
                            key: formattedValue
                        },
                        *fix(fixer) {
                            const leftBracket = sourceCode.getTokenAfter(node.object, astUtils.isOpeningBracketToken);
                            const rightBracket = sourceCode.getLastToken(node);
                            const nextToken = sourceCode.getTokenAfter(node);
                            // Don't perform any fixes if there are comments inside the brackets.
                            if (sourceCode.commentsExistBetween(leftBracket, rightBracket)) {
                                return;
                            }
                            // Replace the brackets by an identifier.
                            if (!node.optional) {
                                yield fixer.insertTextBefore(leftBracket, astUtils.isDecimalInteger(node.object) ? " ." : ".");
                            }
                            yield fixer.replaceTextRange([leftBracket.range[0], rightBracket.range[1]], value);
                            // Insert a space after the property if it will be connected to the next token.
                            if (nextToken &&
                                rightBracket.range[1] === nextToken.range[0] &&
                                !astUtils.canTokensBeAdjacent(String(value), nextToken)) {
                                yield fixer.insertTextAfter(node, " ");
                            }
                        }
                    });
                }
            }
            return {
                MemberExpression(node) {
                    if (node.computed &&
                        node.property.type === "Literal" &&
                        (literalTypesToCheck.has(typeof node.property.value) || astUtils.isNullLiteral(node.property))) {
                        checkComputedProperty(node, node.property.value);
                    }
                    if (node.computed &&
                        node.property.type === "TemplateLiteral" &&
                        node.property.expressions.length === 0) {
                        checkComputedProperty(node, node.property.quasis[0].value.cooked);
                    }
                    if (!allowKeywords &&
                        !node.computed &&
                        keywords.indexOf(String(node.property.name)) !== -1) {
                        context.report({
                            node: node.property,
                            messageId: "useBrackets",
                            data: {
                                key: node.property.name
                            },
                            *fix(fixer) {
                                const dotToken = sourceCode.getTokenBefore(node.property);
                                // A statement that starts with `let[` is parsed as a destructuring variable declaration, not a MemberExpression.
                                if (node.object.type === "Identifier" && node.object.name === "let" && !node.optional) {
                                    return;
                                }
                                // Don't perform any fixes if there are comments between the dot and the property name.
                                if (sourceCode.commentsExistBetween(dotToken, node.property)) {
                                    return;
                                }
                                // Replace the identifier to brackets.
                                if (!node.optional) {
                                    yield fixer.remove(dotToken);
                                }
                                yield fixer.replaceText(node.property, `["${node.property.name}"]`);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/utils/keywords.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = [
        "abstract",
        "boolean",
        "break",
        "byte",
        "case",
        "catch",
        "char",
        "class",
        "const",
        "continue",
        "debugger",
        "default",
        "delete",
        "do",
        "double",
        "else",
        "enum",
        "export",
        "extends",
        "false",
        "final",
        "finally",
        "float",
        "for",
        "function",
        "goto",
        "if",
        "implements",
        "import",
        "in",
        "instanceof",
        "int",
        "interface",
        "long",
        "native",
        "new",
        "null",
        "package",
        "private",
        "protected",
        "public",
        "return",
        "short",
        "static",
        "super",
        "switch",
        "synchronized",
        "this",
        "throw",
        "throws",
        "transient",
        "true",
        "try",
        "typeof",
        "var",
        "void",
        "volatile",
        "while",
        "with"
    ];
},
/* eslint/lib/rules/eol-last.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow newline at the end of files",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/eol-last"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never", "unix", "windows"]
                }
            ],
            messages: {
                missing: "Newline required at end of file but not found.",
                unexpected: "Newline not allowed at end of file."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program: function checkBadEOF(node) {
                    const sourceCode = context.getSourceCode(), src = sourceCode.getText(), lastLine = sourceCode.lines[sourceCode.lines.length - 1], location = {
                        column: lastLine.length,
                        line: sourceCode.lines.length
                    }, LF = "\n", CRLF = `\r${LF}`, endsWithNewline = src.endsWith(LF);
                    /*
                     * Empty source is always valid: No content in file so we don't
                     * need to lint for a newline on the last line of content.
                     */
                    if (!src.length) {
                        return;
                    }
                    let mode = context.options[0] || "always", appendCRLF = false;
                    if (mode === "unix") {
                        // `"unix"` should behave exactly as `"always"`
                        mode = "always";
                    }
                    if (mode === "windows") {
                        // `"windows"` should behave exactly as `"always"`, but append CRLF in the fixer for backwards compatibility
                        mode = "always";
                        appendCRLF = true;
                    }
                    if (mode === "always" && !endsWithNewline) {
                        // File is not newline-terminated, but should be
                        context.report({
                            node,
                            loc: location,
                            messageId: "missing",
                            fix(fixer) {
                                return fixer.insertTextAfterRange([0, src.length], appendCRLF ? CRLF : LF);
                            }
                        });
                    }
                    else if (mode === "never" && endsWithNewline) {
                        // File is newline-terminated, but shouldn't be
                        context.report({
                            node,
                            loc: location,
                            messageId: "unexpected",
                            fix(fixer) {
                                const finalEOLs = /(?:\r?\n)+$/u, match = finalEOLs.exec(sourceCode.text), start = match.index, end = sourceCode.text.length;
                                return fixer.replaceTextRange([start, end], "");
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/eqeqeq.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require the use of `===` and `!==`",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/eqeqeq"
            },
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always"]
                            },
                            {
                                type: "object",
                                properties: {
                                    null: {
                                        enum: ["always", "never", "ignore"]
                                    }
                                },
                                additionalProperties: false
                            }
                        ],
                        additionalItems: false
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["smart", "allow-null"]
                            }
                        ],
                        additionalItems: false
                    }
                ]
            },
            fixable: "code",
            messages: {
                unexpected: "Expected '{{expectedOperator}}' and instead saw '{{actualOperator}}'."
            }
        },
        create(context) {
            const config = context.options[0] || "always";
            const options = context.options[1] || {};
            const sourceCode = context.getSourceCode();
            const nullOption = (config === "always")
                ? options.null || "always"
                : "ignore";
            const enforceRuleForNull = (nullOption === "always");
            const enforceInverseRuleForNull = (nullOption === "never");
            /**
             * Checks if an expression is a typeof expression
             * @param  {ASTNode} node The node to check
             * @returns {boolean} if the node is a typeof expression
             */
            function isTypeOf(node) {
                return node.type === "UnaryExpression" && node.operator === "typeof";
            }
            /**
             * Checks if either operand of a binary expression is a typeof operation
             * @param {ASTNode} node The node to check
             * @returns {boolean} if one of the operands is typeof
             * @private
             */
            function isTypeOfBinary(node) {
                return isTypeOf(node.left) || isTypeOf(node.right);
            }
            /**
             * Checks if operands are literals of the same type (via typeof)
             * @param {ASTNode} node The node to check
             * @returns {boolean} if operands are of same type
             * @private
             */
            function areLiteralsAndSameType(node) {
                return node.left.type === "Literal" && node.right.type === "Literal" &&
                    typeof node.left.value === typeof node.right.value;
            }
            /**
             * Checks if one of the operands is a literal null
             * @param {ASTNode} node The node to check
             * @returns {boolean} if operands are null
             * @private
             */
            function isNullCheck(node) {
                return astUtils.isNullLiteral(node.right) || astUtils.isNullLiteral(node.left);
            }
            /**
             * Reports a message for this rule.
             * @param {ASTNode} node The binary expression node that was checked
             * @param {string} expectedOperator The operator that was expected (either '==', '!=', '===', or '!==')
             * @returns {void}
             * @private
             */
            function report(node, expectedOperator) {
                const operatorToken = sourceCode.getFirstTokenBetween(node.left, node.right, token => token.value === node.operator);
                context.report({
                    node,
                    loc: operatorToken.loc,
                    messageId: "unexpected",
                    data: { expectedOperator, actualOperator: node.operator },
                    fix(fixer) {
                        // If the comparison is a `typeof` comparison or both sides are literals with the same type, then it's safe to fix.
                        if (isTypeOfBinary(node) || areLiteralsAndSameType(node)) {
                            return fixer.replaceText(operatorToken, expectedOperator);
                        }
                        return null;
                    }
                });
            }
            return {
                BinaryExpression(node) {
                    const isNull = isNullCheck(node);
                    if (node.operator !== "==" && node.operator !== "!=") {
                        if (enforceInverseRuleForNull && isNull) {
                            report(node, node.operator.slice(0, -1));
                        }
                        return;
                    }
                    if (config === "smart" && (isTypeOfBinary(node) ||
                        areLiteralsAndSameType(node) || isNull)) {
                        return;
                    }
                    if (!enforceRuleForNull && isNull) {
                        return;
                    }
                    report(node, `${node.operator}=`);
                }
            };
        }
    };
},
/* eslint/lib/rules/for-direction.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "enforce \"for\" loop update clause moving the counter in the right direction.",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/for-direction"
            },
            fixable: null,
            schema: [],
            messages: {
                incorrectDirection: "The update clause in this loop moves the variable in the wrong direction."
            }
        },
        create(context) {
            /**
             * report an error.
             * @param {ASTNode} node the node to report.
             * @returns {void}
             */
            function report(node) {
                context.report({
                    node,
                    messageId: "incorrectDirection"
                });
            }
            /**
             * check the right side of the assignment
             * @param {ASTNode} update UpdateExpression to check
             * @param {int} dir expected direction that could either be turned around or invalidated
             * @returns {int} return dir, the negated dir or zero if it's not clear for identifiers
             */
            function getRightDirection(update, dir) {
                if (update.right.type === "UnaryExpression") {
                    if (update.right.operator === "-") {
                        return -dir;
                    }
                }
                else if (update.right.type === "Identifier") {
                    return 0;
                }
                return dir;
            }
            /**
             * check UpdateExpression add/sub the counter
             * @param {ASTNode} update UpdateExpression to check
             * @param {string} counter variable name to check
             * @returns {int} if add return 1, if sub return -1, if nochange, return 0
             */
            function getUpdateDirection(update, counter) {
                if (update.argument.type === "Identifier" && update.argument.name === counter) {
                    if (update.operator === "++") {
                        return 1;
                    }
                    if (update.operator === "--") {
                        return -1;
                    }
                }
                return 0;
            }
            /**
             * check AssignmentExpression add/sub the counter
             * @param {ASTNode} update AssignmentExpression to check
             * @param {string} counter variable name to check
             * @returns {int} if add return 1, if sub return -1, if nochange, return 0
             */
            function getAssignmentDirection(update, counter) {
                if (update.left.name === counter) {
                    if (update.operator === "+=") {
                        return getRightDirection(update, 1);
                    }
                    if (update.operator === "-=") {
                        return getRightDirection(update, -1);
                    }
                }
                return 0;
            }
            return {
                ForStatement(node) {
                    if (node.test && node.test.type === "BinaryExpression" && node.test.left.type === "Identifier" && node.update) {
                        const counter = node.test.left.name;
                        const operator = node.test.operator;
                        const update = node.update;
                        let wrongDirection;
                        if (operator === "<" || operator === "<=") {
                            wrongDirection = -1;
                        }
                        else if (operator === ">" || operator === ">=") {
                            wrongDirection = 1;
                        }
                        else {
                            return;
                        }
                        if (update.type === "UpdateExpression") {
                            if (getUpdateDirection(update, counter) === wrongDirection) {
                                report(node);
                            }
                        }
                        else if (update.type === "AssignmentExpression" && getAssignmentDirection(update, counter) === wrongDirection) {
                            report(node);
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/func-call-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow spacing between function identifiers and their invocations",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/func-call-spacing"
            },
            fixable: "whitespace",
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["never"]
                            }
                        ],
                        minItems: 0,
                        maxItems: 1
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always"]
                            },
                            {
                                type: "object",
                                properties: {
                                    allowNewlines: {
                                        type: "boolean"
                                    }
                                },
                                additionalProperties: false
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    }
                ]
            },
            messages: {
                unexpectedWhitespace: "Unexpected whitespace between function name and paren.",
                unexpectedNewline: "Unexpected newline between function name and paren.",
                missing: "Missing space between function name and paren."
            }
        },
        create(context) {
            const never = context.options[0] !== "always";
            const allowNewlines = !never && context.options[1] && context.options[1].allowNewlines;
            const sourceCode = context.getSourceCode();
            const text = sourceCode.getText();
            /**
             * Check if open space is present in a function name
             * @param {ASTNode} node node to evaluate
             * @param {Token} leftToken The last token of the callee. This may be the closing parenthesis that encloses the callee.
             * @param {Token} rightToken Tha first token of the arguments. this is the opening parenthesis that encloses the arguments.
             * @returns {void}
             * @private
             */
            function checkSpacing(node, leftToken, rightToken) {
                const textBetweenTokens = text.slice(leftToken.range[1], rightToken.range[0]).replace(/\/\*.*?\*\//gu, "");
                const hasWhitespace = /\s/u.test(textBetweenTokens);
                const hasNewline = hasWhitespace && astUtils.LINEBREAK_MATCHER.test(textBetweenTokens);
                /*
                 * never allowNewlines hasWhitespace hasNewline message
                 * F     F             F             F          Missing space between function name and paren.
                 * F     F             F             T          (Invalid `!hasWhitespace && hasNewline`)
                 * F     F             T             T          Unexpected newline between function name and paren.
                 * F     F             T             F          (OK)
                 * F     T             T             F          (OK)
                 * F     T             T             T          (OK)
                 * F     T             F             T          (Invalid `!hasWhitespace && hasNewline`)
                 * F     T             F             F          Missing space between function name and paren.
                 * T     T             F             F          (Invalid `never && allowNewlines`)
                 * T     T             F             T          (Invalid `!hasWhitespace && hasNewline`)
                 * T     T             T             T          (Invalid `never && allowNewlines`)
                 * T     T             T             F          (Invalid `never && allowNewlines`)
                 * T     F             T             F          Unexpected space between function name and paren.
                 * T     F             T             T          Unexpected space between function name and paren.
                 * T     F             F             T          (Invalid `!hasWhitespace && hasNewline`)
                 * T     F             F             F          (OK)
                 *
                 * T                   T                        Unexpected space between function name and paren.
                 * F                   F                        Missing space between function name and paren.
                 * F     F                           T          Unexpected newline between function name and paren.
                 */
                if (never && hasWhitespace) {
                    context.report({
                        node,
                        loc: {
                            start: leftToken.loc.end,
                            end: {
                                line: rightToken.loc.start.line,
                                column: rightToken.loc.start.column - 1
                            }
                        },
                        messageId: "unexpectedWhitespace",
                        fix(fixer) {
                            // Don't remove comments.
                            if (sourceCode.commentsExistBetween(leftToken, rightToken)) {
                                return null;
                            }
                            // If `?.` exists, it doesn't hide no-unexpected-multiline errors
                            if (node.optional) {
                                return fixer.replaceTextRange([leftToken.range[1], rightToken.range[0]], "?.");
                            }
                            /*
                             * Only autofix if there is no newline
                             * https://github.com/eslint/eslint/issues/7787
                             */
                            if (hasNewline) {
                                return null;
                            }
                            return fixer.removeRange([leftToken.range[1], rightToken.range[0]]);
                        }
                    });
                }
                else if (!never && !hasWhitespace) {
                    context.report({
                        node,
                        loc: {
                            start: {
                                line: leftToken.loc.end.line,
                                column: leftToken.loc.end.column - 1
                            },
                            end: rightToken.loc.start
                        },
                        messageId: "missing",
                        fix(fixer) {
                            if (node.optional) {
                                return null; // Not sure if inserting a space to either before/after `?.` token.
                            }
                            return fixer.insertTextBefore(rightToken, " ");
                        }
                    });
                }
                else if (!never && !allowNewlines && hasNewline) {
                    context.report({
                        node,
                        loc: {
                            start: leftToken.loc.end,
                            end: rightToken.loc.start
                        },
                        messageId: "unexpectedNewline",
                        fix(fixer) {
                            /*
                             * Only autofix if there is no newline
                             * https://github.com/eslint/eslint/issues/7787
                             * But if `?.` exists, it doesn't hide no-unexpected-multiline errors
                             */
                            if (!node.optional) {
                                return null;
                            }
                            // Don't remove comments.
                            if (sourceCode.commentsExistBetween(leftToken, rightToken)) {
                                return null;
                            }
                            const range = [leftToken.range[1], rightToken.range[0]];
                            const qdToken = sourceCode.getTokenAfter(leftToken);
                            if (qdToken.range[0] === leftToken.range[1]) {
                                return fixer.replaceTextRange(range, "?. ");
                            }
                            if (qdToken.range[1] === rightToken.range[0]) {
                                return fixer.replaceTextRange(range, " ?.");
                            }
                            return fixer.replaceTextRange(range, " ?. ");
                        }
                    });
                }
            }
            return {
                "CallExpression, NewExpression"(node) {
                    const lastToken = sourceCode.getLastToken(node);
                    const lastCalleeToken = sourceCode.getLastToken(node.callee);
                    const parenToken = sourceCode.getFirstTokenBetween(lastCalleeToken, lastToken, astUtils.isOpeningParenToken);
                    const prevToken = parenToken && sourceCode.getTokenBefore(parenToken, astUtils.isNotQuestionDotToken);
                    // Parens in NewExpression are optional
                    if (!(parenToken && parenToken.range[1] < node.range[1])) {
                        return;
                    }
                    checkSpacing(node, prevToken, parenToken);
                },
                ImportExpression(node) {
                    const leftToken = sourceCode.getFirstToken(node);
                    const rightToken = sourceCode.getTokenAfter(leftToken);
                    checkSpacing(node, leftToken, rightToken);
                }
            };
        }
    };
},
/* eslint/lib/rules/func-name-matching.js */ function _(require, module, exports, __esModule, __esExport) {
    //--------------------------------------------------------------------------
    // Requirements
    //--------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const esutils = require(395) /* esutils */;
    //--------------------------------------------------------------------------
    // Helpers
    //--------------------------------------------------------------------------
    /**
     * Determines if a pattern is `module.exports` or `module["exports"]`
     * @param {ASTNode} pattern The left side of the AssignmentExpression
     * @returns {boolean} True if the pattern is `module.exports` or `module["exports"]`
     */
    function isModuleExports(pattern) {
        if (pattern.type === "MemberExpression" && pattern.object.type === "Identifier" && pattern.object.name === "module") {
            // module.exports
            if (pattern.property.type === "Identifier" && pattern.property.name === "exports") {
                return true;
            }
            // module["exports"]
            if (pattern.property.type === "Literal" && pattern.property.value === "exports") {
                return true;
            }
        }
        return false;
    }
    /**
     * Determines if a string name is a valid identifier
     * @param {string} name The string to be checked
     * @param {int} ecmaVersion The ECMAScript version if specified in the parserOptions config
     * @returns {boolean} True if the string is a valid identifier
     */
    function isIdentifier(name, ecmaVersion) {
        if (ecmaVersion >= 6) {
            return esutils.keyword.isIdentifierES6(name);
        }
        return esutils.keyword.isIdentifierES5(name);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const alwaysOrNever = { enum: ["always", "never"] };
    const optionsObject = {
        type: "object",
        properties: {
            considerPropertyDescriptor: {
                type: "boolean"
            },
            includeCommonJSModuleExports: {
                type: "boolean"
            }
        },
        additionalProperties: false
    };
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require function names to match the name of the variable or property to which they are assigned",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/func-name-matching"
            },
            schema: {
                anyOf: [{
                        type: "array",
                        additionalItems: false,
                        items: [alwaysOrNever, optionsObject]
                    }, {
                        type: "array",
                        additionalItems: false,
                        items: [optionsObject]
                    }]
            },
            messages: {
                matchProperty: "Function name `{{funcName}}` should match property name `{{name}}`.",
                matchVariable: "Function name `{{funcName}}` should match variable name `{{name}}`.",
                notMatchProperty: "Function name `{{funcName}}` should not match property name `{{name}}`.",
                notMatchVariable: "Function name `{{funcName}}` should not match variable name `{{name}}`."
            }
        },
        create(context) {
            const options = (typeof context.options[0] === "object" ? context.options[0] : context.options[1]) || {};
            const nameMatches = typeof context.options[0] === "string" ? context.options[0] : "always";
            const considerPropertyDescriptor = options.considerPropertyDescriptor;
            const includeModuleExports = options.includeCommonJSModuleExports;
            const ecmaVersion = context.parserOptions && context.parserOptions.ecmaVersion ? context.parserOptions.ecmaVersion : 5;
            /**
             * Check whether node is a certain CallExpression.
             * @param {string} objName object name
             * @param {string} funcName function name
             * @param {ASTNode} node The node to check
             * @returns {boolean} `true` if node matches CallExpression
             */
            function isPropertyCall(objName, funcName, node) {
                if (!node) {
                    return false;
                }
                return node.type === "CallExpression" && astUtils.isSpecificMemberAccess(node.callee, objName, funcName);
            }
            /**
             * Compares identifiers based on the nameMatches option
             * @param {string} x the first identifier
             * @param {string} y the second identifier
             * @returns {boolean} whether the two identifiers should warn.
             */
            function shouldWarn(x, y) {
                return (nameMatches === "always" && x !== y) || (nameMatches === "never" && x === y);
            }
            /**
             * Reports
             * @param {ASTNode} node The node to report
             * @param {string} name The variable or property name
             * @param {string} funcName The function name
             * @param {boolean} isProp True if the reported node is a property assignment
             * @returns {void}
             */
            function report(node, name, funcName, isProp) {
                let messageId;
                if (nameMatches === "always" && isProp) {
                    messageId = "matchProperty";
                }
                else if (nameMatches === "always") {
                    messageId = "matchVariable";
                }
                else if (isProp) {
                    messageId = "notMatchProperty";
                }
                else {
                    messageId = "notMatchVariable";
                }
                context.report({
                    node,
                    messageId,
                    data: {
                        name,
                        funcName
                    }
                });
            }
            /**
             * Determines whether a given node is a string literal
             * @param {ASTNode} node The node to check
             * @returns {boolean} `true` if the node is a string literal
             */
            function isStringLiteral(node) {
                return node.type === "Literal" && typeof node.value === "string";
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                VariableDeclarator(node) {
                    if (!node.init || node.init.type !== "FunctionExpression" || node.id.type !== "Identifier") {
                        return;
                    }
                    if (node.init.id && shouldWarn(node.id.name, node.init.id.name)) {
                        report(node, node.id.name, node.init.id.name, false);
                    }
                },
                AssignmentExpression(node) {
                    if (node.right.type !== "FunctionExpression" ||
                        (node.left.computed && node.left.property.type !== "Literal") ||
                        (!includeModuleExports && isModuleExports(node.left)) ||
                        (node.left.type !== "Identifier" && node.left.type !== "MemberExpression")) {
                        return;
                    }
                    const isProp = node.left.type === "MemberExpression";
                    const name = isProp ? astUtils.getStaticPropertyName(node.left) : node.left.name;
                    if (node.right.id && isIdentifier(name) && shouldWarn(name, node.right.id.name)) {
                        report(node, name, node.right.id.name, isProp);
                    }
                },
                Property(node) {
                    if (node.value.type !== "FunctionExpression" || !node.value.id || node.computed && !isStringLiteral(node.key)) {
                        return;
                    }
                    if (node.key.type === "Identifier") {
                        const functionName = node.value.id.name;
                        let propertyName = node.key.name;
                        if (considerPropertyDescriptor && propertyName === "value") {
                            if (isPropertyCall("Object", "defineProperty", node.parent.parent) || isPropertyCall("Reflect", "defineProperty", node.parent.parent)) {
                                const property = node.parent.parent.arguments[1];
                                if (isStringLiteral(property) && shouldWarn(property.value, functionName)) {
                                    report(node, property.value, functionName, true);
                                }
                            }
                            else if (isPropertyCall("Object", "defineProperties", node.parent.parent.parent.parent)) {
                                propertyName = node.parent.parent.key.name;
                                if (!node.parent.parent.computed && shouldWarn(propertyName, functionName)) {
                                    report(node, propertyName, functionName, true);
                                }
                            }
                            else if (isPropertyCall("Object", "create", node.parent.parent.parent.parent)) {
                                propertyName = node.parent.parent.key.name;
                                if (!node.parent.parent.computed && shouldWarn(propertyName, functionName)) {
                                    report(node, propertyName, functionName, true);
                                }
                            }
                            else if (shouldWarn(propertyName, functionName)) {
                                report(node, propertyName, functionName, true);
                            }
                        }
                        else if (shouldWarn(propertyName, functionName)) {
                            report(node, propertyName, functionName, true);
                        }
                        return;
                    }
                    if (isStringLiteral(node.key) &&
                        isIdentifier(node.key.value, ecmaVersion) &&
                        shouldWarn(node.key.value, node.value.id.name)) {
                        report(node, node.key.value, node.value.id.name, true);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/func-names.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    /**
     * Checks whether or not a given variable is a function name.
     * @param {eslint-scope.Variable} variable A variable to check.
     * @returns {boolean} `true` if the variable is a function name.
     */
    function isFunctionName(variable) {
        return variable && variable.defs[0].type === "FunctionName";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require or disallow named `function` expressions",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/func-names"
            },
            schema: {
                definitions: {
                    value: {
                        enum: [
                            "always",
                            "as-needed",
                            "never"
                        ]
                    }
                },
                items: [
                    {
                        $ref: "#/definitions/value"
                    },
                    {
                        type: "object",
                        properties: {
                            generators: {
                                $ref: "#/definitions/value"
                            }
                        },
                        additionalProperties: false
                    }
                ]
            },
            messages: {
                unnamed: "Unexpected unnamed {{name}}.",
                named: "Unexpected named {{name}}."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Returns the config option for the given node.
             * @param {ASTNode} node A node to get the config for.
             * @returns {string} The config option.
             */
            function getConfigForNode(node) {
                if (node.generator &&
                    context.options.length > 1 &&
                    context.options[1].generators) {
                    return context.options[1].generators;
                }
                return context.options[0] || "always";
            }
            /**
             * Determines whether the current FunctionExpression node is a get, set, or
             * shorthand method in an object literal or a class.
             * @param {ASTNode} node A node to check.
             * @returns {boolean} True if the node is a get, set, or shorthand method.
             */
            function isObjectOrClassMethod(node) {
                const parent = node.parent;
                return (parent.type === "MethodDefinition" || (parent.type === "Property" && (parent.method ||
                    parent.kind === "get" ||
                    parent.kind === "set")));
            }
            /**
             * Determines whether the current FunctionExpression node has a name that would be
             * inferred from context in a conforming ES6 environment.
             * @param {ASTNode} node A node to check.
             * @returns {boolean} True if the node would have a name assigned automatically.
             */
            function hasInferredName(node) {
                const parent = node.parent;
                return isObjectOrClassMethod(node) ||
                    (parent.type === "VariableDeclarator" && parent.id.type === "Identifier" && parent.init === node) ||
                    (parent.type === "Property" && parent.value === node) ||
                    (parent.type === "AssignmentExpression" && parent.left.type === "Identifier" && parent.right === node) ||
                    (parent.type === "AssignmentPattern" && parent.left.type === "Identifier" && parent.right === node);
            }
            /**
             * Reports that an unnamed function should be named
             * @param {ASTNode} node The node to report in the event of an error.
             * @returns {void}
             */
            function reportUnexpectedUnnamedFunction(node) {
                context.report({
                    node,
                    messageId: "unnamed",
                    loc: astUtils.getFunctionHeadLoc(node, sourceCode),
                    data: { name: astUtils.getFunctionNameWithKind(node) }
                });
            }
            /**
             * Reports that a named function should be unnamed
             * @param {ASTNode} node The node to report in the event of an error.
             * @returns {void}
             */
            function reportUnexpectedNamedFunction(node) {
                context.report({
                    node,
                    messageId: "named",
                    loc: astUtils.getFunctionHeadLoc(node, sourceCode),
                    data: { name: astUtils.getFunctionNameWithKind(node) }
                });
            }
            /**
             * The listener for function nodes.
             * @param {ASTNode} node function node
             * @returns {void}
             */
            function handleFunction(node) {
                // Skip recursive functions.
                const nameVar = context.getDeclaredVariables(node)[0];
                if (isFunctionName(nameVar) && nameVar.references.length > 0) {
                    return;
                }
                const hasName = Boolean(node.id && node.id.name);
                const config = getConfigForNode(node);
                if (config === "never") {
                    if (hasName && node.type !== "FunctionDeclaration") {
                        reportUnexpectedNamedFunction(node);
                    }
                }
                else if (config === "as-needed") {
                    if (!hasName && !hasInferredName(node)) {
                        reportUnexpectedUnnamedFunction(node);
                    }
                }
                else {
                    if (!hasName && !isObjectOrClassMethod(node)) {
                        reportUnexpectedUnnamedFunction(node);
                    }
                }
            }
            return {
                "FunctionExpression:exit": handleFunction,
                "ExportDefaultDeclaration > FunctionDeclaration": handleFunction
            };
        }
    };
},
/* eslint/lib/rules/func-style.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce the consistent use of either `function` declarations or expressions",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/func-style"
            },
            schema: [
                {
                    enum: ["declaration", "expression"]
                },
                {
                    type: "object",
                    properties: {
                        allowArrowFunctions: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                expression: "Expected a function expression.",
                declaration: "Expected a function declaration."
            }
        },
        create(context) {
            const style = context.options[0], allowArrowFunctions = context.options[1] && context.options[1].allowArrowFunctions, enforceDeclarations = (style === "declaration"), stack = [];
            const nodesToCheck = {
                FunctionDeclaration(node) {
                    stack.push(false);
                    if (!enforceDeclarations && node.parent.type !== "ExportDefaultDeclaration") {
                        context.report({ node, messageId: "expression" });
                    }
                },
                "FunctionDeclaration:exit"() {
                    stack.pop();
                },
                FunctionExpression(node) {
                    stack.push(false);
                    if (enforceDeclarations && node.parent.type === "VariableDeclarator") {
                        context.report({ node: node.parent, messageId: "declaration" });
                    }
                },
                "FunctionExpression:exit"() {
                    stack.pop();
                },
                ThisExpression() {
                    if (stack.length > 0) {
                        stack[stack.length - 1] = true;
                    }
                }
            };
            if (!allowArrowFunctions) {
                nodesToCheck.ArrowFunctionExpression = function () {
                    stack.push(false);
                };
                nodesToCheck["ArrowFunctionExpression:exit"] = function (node) {
                    const hasThisExpr = stack.pop();
                    if (enforceDeclarations && !hasThisExpr && node.parent.type === "VariableDeclarator") {
                        context.report({ node: node.parent, messageId: "declaration" });
                    }
                };
            }
            return nodesToCheck;
        }
    };
},
/* eslint/lib/rules/function-call-argument-newline.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce line breaks between arguments of a function call",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/function-call-argument-newline"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never", "consistent"]
                }
            ],
            messages: {
                unexpectedLineBreak: "There should be no line break here.",
                missingLineBreak: "There should be a line break after this argument."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const checkers = {
                unexpected: {
                    messageId: "unexpectedLineBreak",
                    check: (prevToken, currentToken) => prevToken.loc.end.line !== currentToken.loc.start.line,
                    createFix: (token, tokenBefore) => fixer => fixer.replaceTextRange([tokenBefore.range[1], token.range[0]], " ")
                },
                missing: {
                    messageId: "missingLineBreak",
                    check: (prevToken, currentToken) => prevToken.loc.end.line === currentToken.loc.start.line,
                    createFix: (token, tokenBefore) => fixer => fixer.replaceTextRange([tokenBefore.range[1], token.range[0]], "\n")
                }
            };
            /**
             * Check all arguments for line breaks in the CallExpression
             * @param {CallExpression} node node to evaluate
             * @param {{ messageId: string, check: Function }} checker selected checker
             * @returns {void}
             * @private
             */
            function checkArguments(node, checker) {
                for (let i = 1; i < node.arguments.length; i++) {
                    const prevArgToken = sourceCode.getLastToken(node.arguments[i - 1]);
                    const currentArgToken = sourceCode.getFirstToken(node.arguments[i]);
                    if (checker.check(prevArgToken, currentArgToken)) {
                        const tokenBefore = sourceCode.getTokenBefore(currentArgToken, { includeComments: true });
                        const hasLineCommentBefore = tokenBefore.type === "Line";
                        context.report({
                            node,
                            loc: {
                                start: tokenBefore.loc.end,
                                end: currentArgToken.loc.start
                            },
                            messageId: checker.messageId,
                            fix: hasLineCommentBefore ? null : checker.createFix(currentArgToken, tokenBefore)
                        });
                    }
                }
            }
            /**
             * Check if open space is present in a function name
             * @param {CallExpression} node node to evaluate
             * @returns {void}
             * @private
             */
            function check(node) {
                if (node.arguments.length < 2) {
                    return;
                }
                const option = context.options[0] || "always";
                if (option === "never") {
                    checkArguments(node, checkers.unexpected);
                }
                else if (option === "always") {
                    checkArguments(node, checkers.missing);
                }
                else if (option === "consistent") {
                    const firstArgToken = sourceCode.getLastToken(node.arguments[0]);
                    const secondArgToken = sourceCode.getFirstToken(node.arguments[1]);
                    if (firstArgToken.loc.end.line === secondArgToken.loc.start.line) {
                        checkArguments(node, checkers.unexpected);
                    }
                    else {
                        checkArguments(node, checkers.missing);
                    }
                }
            }
            return {
                CallExpression: check,
                NewExpression: check
            };
        }
    };
},
/* eslint/lib/rules/function-paren-newline.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent line breaks inside function parentheses",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/function-paren-newline"
            },
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["always", "never", "consistent", "multiline", "multiline-arguments"]
                        },
                        {
                            type: "object",
                            properties: {
                                minItems: {
                                    type: "integer",
                                    minimum: 0
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                expectedBefore: "Expected newline before ')'.",
                expectedAfter: "Expected newline after '('.",
                expectedBetween: "Expected newline between arguments/params.",
                unexpectedBefore: "Unexpected newline before ')'.",
                unexpectedAfter: "Unexpected newline after '('."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const rawOption = context.options[0] || "multiline";
            const multilineOption = rawOption === "multiline";
            const multilineArgumentsOption = rawOption === "multiline-arguments";
            const consistentOption = rawOption === "consistent";
            let minItems;
            if (typeof rawOption === "object") {
                minItems = rawOption.minItems;
            }
            else if (rawOption === "always") {
                minItems = 0;
            }
            else if (rawOption === "never") {
                minItems = Infinity;
            }
            else {
                minItems = null;
            }
            //----------------------------------------------------------------------
            // Helpers
            //----------------------------------------------------------------------
            /**
             * Determines whether there should be newlines inside function parens
             * @param {ASTNode[]} elements The arguments or parameters in the list
             * @param {boolean} hasLeftNewline `true` if the left paren has a newline in the current code.
             * @returns {boolean} `true` if there should be newlines inside the function parens
             */
            function shouldHaveNewlines(elements, hasLeftNewline) {
                if (multilineArgumentsOption && elements.length === 1) {
                    return hasLeftNewline;
                }
                if (multilineOption || multilineArgumentsOption) {
                    return elements.some((element, index) => index !== elements.length - 1 && element.loc.end.line !== elements[index + 1].loc.start.line);
                }
                if (consistentOption) {
                    return hasLeftNewline;
                }
                return elements.length >= minItems;
            }
            /**
             * Validates parens
             * @param {Object} parens An object with keys `leftParen` for the left paren token, and `rightParen` for the right paren token
             * @param {ASTNode[]} elements The arguments or parameters in the list
             * @returns {void}
             */
            function validateParens(parens, elements) {
                const leftParen = parens.leftParen;
                const rightParen = parens.rightParen;
                const tokenAfterLeftParen = sourceCode.getTokenAfter(leftParen);
                const tokenBeforeRightParen = sourceCode.getTokenBefore(rightParen);
                const hasLeftNewline = !astUtils.isTokenOnSameLine(leftParen, tokenAfterLeftParen);
                const hasRightNewline = !astUtils.isTokenOnSameLine(tokenBeforeRightParen, rightParen);
                const needsNewlines = shouldHaveNewlines(elements, hasLeftNewline);
                if (hasLeftNewline && !needsNewlines) {
                    context.report({
                        node: leftParen,
                        messageId: "unexpectedAfter",
                        fix(fixer) {
                            return sourceCode.getText().slice(leftParen.range[1], tokenAfterLeftParen.range[0]).trim()
                                // If there is a comment between the ( and the first element, don't do a fix.
                                ? null
                                : fixer.removeRange([leftParen.range[1], tokenAfterLeftParen.range[0]]);
                        }
                    });
                }
                else if (!hasLeftNewline && needsNewlines) {
                    context.report({
                        node: leftParen,
                        messageId: "expectedAfter",
                        fix: fixer => fixer.insertTextAfter(leftParen, "\n")
                    });
                }
                if (hasRightNewline && !needsNewlines) {
                    context.report({
                        node: rightParen,
                        messageId: "unexpectedBefore",
                        fix(fixer) {
                            return sourceCode.getText().slice(tokenBeforeRightParen.range[1], rightParen.range[0]).trim()
                                // If there is a comment between the last element and the ), don't do a fix.
                                ? null
                                : fixer.removeRange([tokenBeforeRightParen.range[1], rightParen.range[0]]);
                        }
                    });
                }
                else if (!hasRightNewline && needsNewlines) {
                    context.report({
                        node: rightParen,
                        messageId: "expectedBefore",
                        fix: fixer => fixer.insertTextBefore(rightParen, "\n")
                    });
                }
            }
            /**
             * Validates a list of arguments or parameters
             * @param {Object} parens An object with keys `leftParen` for the left paren token, and `rightParen` for the right paren token
             * @param {ASTNode[]} elements The arguments or parameters in the list
             * @returns {void}
             */
            function validateArguments(parens, elements) {
                const leftParen = parens.leftParen;
                const tokenAfterLeftParen = sourceCode.getTokenAfter(leftParen);
                const hasLeftNewline = !astUtils.isTokenOnSameLine(leftParen, tokenAfterLeftParen);
                const needsNewlines = shouldHaveNewlines(elements, hasLeftNewline);
                for (let i = 0; i <= elements.length - 2; i++) {
                    const currentElement = elements[i];
                    const nextElement = elements[i + 1];
                    const hasNewLine = currentElement.loc.end.line !== nextElement.loc.start.line;
                    if (!hasNewLine && needsNewlines) {
                        context.report({
                            node: currentElement,
                            messageId: "expectedBetween",
                            fix: fixer => fixer.insertTextBefore(nextElement, "\n")
                        });
                    }
                }
            }
            /**
             * Gets the left paren and right paren tokens of a node.
             * @param {ASTNode} node The node with parens
             * @returns {Object} An object with keys `leftParen` for the left paren token, and `rightParen` for the right paren token.
             * Can also return `null` if an expression has no parens (e.g. a NewExpression with no arguments, or an ArrowFunctionExpression
             * with a single parameter)
             */
            function getParenTokens(node) {
                switch (node.type) {
                    case "NewExpression":
                        if (!node.arguments.length && !(astUtils.isOpeningParenToken(sourceCode.getLastToken(node, { skip: 1 })) &&
                            astUtils.isClosingParenToken(sourceCode.getLastToken(node)))) {
                            // If the NewExpression does not have parens (e.g. `new Foo`), return null.
                            return null;
                        }
                    // falls through
                    case "CallExpression":
                        return {
                            leftParen: sourceCode.getTokenAfter(node.callee, astUtils.isOpeningParenToken),
                            rightParen: sourceCode.getLastToken(node)
                        };
                    case "FunctionDeclaration":
                    case "FunctionExpression": {
                        const leftParen = sourceCode.getFirstToken(node, astUtils.isOpeningParenToken);
                        const rightParen = node.params.length
                            ? sourceCode.getTokenAfter(node.params[node.params.length - 1], astUtils.isClosingParenToken)
                            : sourceCode.getTokenAfter(leftParen);
                        return { leftParen, rightParen };
                    }
                    case "ArrowFunctionExpression": {
                        const firstToken = sourceCode.getFirstToken(node, { skip: (node.async ? 1 : 0) });
                        if (!astUtils.isOpeningParenToken(firstToken)) {
                            // If the ArrowFunctionExpression has a single param without parens, return null.
                            return null;
                        }
                        return {
                            leftParen: firstToken,
                            rightParen: sourceCode.getTokenBefore(node.body, astUtils.isClosingParenToken)
                        };
                    }
                    case "ImportExpression": {
                        const leftParen = sourceCode.getFirstToken(node, 1);
                        const rightParen = sourceCode.getLastToken(node);
                        return { leftParen, rightParen };
                    }
                    default:
                        throw new TypeError(`unexpected node with type ${node.type}`);
                }
            }
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                [[
                    "ArrowFunctionExpression",
                    "CallExpression",
                    "FunctionDeclaration",
                    "FunctionExpression",
                    "ImportExpression",
                    "NewExpression"
                ]](node) {
                    const parens = getParenTokens(node);
                    let params;
                    if (node.type === "ImportExpression") {
                        params = [node.source];
                    }
                    else if (astUtils.isFunction(node)) {
                        params = node.params;
                    }
                    else {
                        params = node.arguments;
                    }
                    if (parens) {
                        validateParens(parens, params);
                        if (multilineArgumentsOption) {
                            validateArguments(parens, params);
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/generator-star-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const OVERRIDE_SCHEMA = {
        oneOf: [
            {
                enum: ["before", "after", "both", "neither"]
            },
            {
                type: "object",
                properties: {
                    before: { type: "boolean" },
                    after: { type: "boolean" }
                },
                additionalProperties: false
            }
        ]
    };
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing around `*` operators in generator functions",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/generator-star-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["before", "after", "both", "neither"]
                        },
                        {
                            type: "object",
                            properties: {
                                before: { type: "boolean" },
                                after: { type: "boolean" },
                                named: OVERRIDE_SCHEMA,
                                anonymous: OVERRIDE_SCHEMA,
                                method: OVERRIDE_SCHEMA
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                missingBefore: "Missing space before *.",
                missingAfter: "Missing space after *.",
                unexpectedBefore: "Unexpected space before *.",
                unexpectedAfter: "Unexpected space after *."
            }
        },
        create(context) {
            const optionDefinitions = {
                before: { before: true, after: false },
                after: { before: false, after: true },
                both: { before: true, after: true },
                neither: { before: false, after: false }
            };
            /**
             * Returns resolved option definitions based on an option and defaults
             * @param {any} option The option object or string value
             * @param {Object} defaults The defaults to use if options are not present
             * @returns {Object} the resolved object definition
             */
            function optionToDefinition(option, defaults) {
                if (!option) {
                    return defaults;
                }
                return typeof option === "string"
                    ? optionDefinitions[option]
                    : Object.assign({}, defaults, option);
            }
            const modes = (function (option) {
                const defaults = optionToDefinition(option, optionDefinitions.before);
                return {
                    named: optionToDefinition(option.named, defaults),
                    anonymous: optionToDefinition(option.anonymous, defaults),
                    method: optionToDefinition(option.method, defaults)
                };
            }(context.options[0] || {}));
            const sourceCode = context.getSourceCode();
            /**
             * Checks if the given token is a star token or not.
             * @param {Token} token The token to check.
             * @returns {boolean} `true` if the token is a star token.
             */
            function isStarToken(token) {
                return token.value === "*" && token.type === "Punctuator";
            }
            /**
             * Gets the generator star token of the given function node.
             * @param {ASTNode} node The function node to get.
             * @returns {Token} Found star token.
             */
            function getStarToken(node) {
                return sourceCode.getFirstToken((node.parent.method || node.parent.type === "MethodDefinition") ? node.parent : node, isStarToken);
            }
            /**
             * capitalize a given string.
             * @param {string} str the given string.
             * @returns {string} the capitalized string.
             */
            function capitalize(str) {
                return str[0].toUpperCase() + str.slice(1);
            }
            /**
             * Checks the spacing between two tokens before or after the star token.
             * @param {string} kind Either "named", "anonymous", or "method"
             * @param {string} side Either "before" or "after".
             * @param {Token} leftToken `function` keyword token if side is "before", or
             *     star token if side is "after".
             * @param {Token} rightToken Star token if side is "before", or identifier
             *     token if side is "after".
             * @returns {void}
             */
            function checkSpacing(kind, side, leftToken, rightToken) {
                if (!!(rightToken.range[0] - leftToken.range[1]) !== modes[kind][side]) {
                    const after = leftToken.value === "*";
                    const spaceRequired = modes[kind][side];
                    const node = after ? leftToken : rightToken;
                    const messageId = `${spaceRequired ? "missing" : "unexpected"}${capitalize(side)}`;
                    context.report({
                        node,
                        messageId,
                        fix(fixer) {
                            if (spaceRequired) {
                                if (after) {
                                    return fixer.insertTextAfter(node, " ");
                                }
                                return fixer.insertTextBefore(node, " ");
                            }
                            return fixer.removeRange([leftToken.range[1], rightToken.range[0]]);
                        }
                    });
                }
            }
            /**
             * Enforces the spacing around the star if node is a generator function.
             * @param {ASTNode} node A function expression or declaration node.
             * @returns {void}
             */
            function checkFunction(node) {
                if (!node.generator) {
                    return;
                }
                const starToken = getStarToken(node);
                const prevToken = sourceCode.getTokenBefore(starToken);
                const nextToken = sourceCode.getTokenAfter(starToken);
                let kind = "named";
                if (node.parent.type === "MethodDefinition" || (node.parent.type === "Property" && node.parent.method)) {
                    kind = "method";
                }
                else if (!node.id) {
                    kind = "anonymous";
                }
                // Only check before when preceded by `function`|`static` keyword
                if (!(kind === "method" && starToken === sourceCode.getFirstToken(node.parent))) {
                    checkSpacing(kind, "before", prevToken, starToken);
                }
                checkSpacing(kind, "after", starToken, nextToken);
            }
            return {
                FunctionDeclaration: checkFunction,
                FunctionExpression: checkFunction
            };
        }
    };
},
/* eslint/lib/rules/getter-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const TARGET_NODE_TYPE = /^(?:Arrow)?FunctionExpression$/u;
    /**
     * Checks a given code path segment is reachable.
     * @param {CodePathSegment} segment A segment to check.
     * @returns {boolean} `true` if the segment is reachable.
     */
    function isReachable(segment) {
        return segment.reachable;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "enforce `return` statements in getters",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/getter-return"
            },
            fixable: null,
            schema: [
                {
                    type: "object",
                    properties: {
                        allowImplicit: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                expected: "Expected to return a value in {{name}}.",
                expectedAlways: "Expected {{name}} to always return a value."
            }
        },
        create(context) {
            const options = context.options[0] || { allowImplicit: false };
            const sourceCode = context.getSourceCode();
            let funcInfo = {
                upper: null,
                codePath: null,
                hasReturn: false,
                shouldCheck: false,
                node: null
            };
            /**
             * Checks whether or not the last code path segment is reachable.
             * Then reports this function if the segment is reachable.
             *
             * If the last code path segment is reachable, there are paths which are not
             * returned or thrown.
             * @param {ASTNode} node A node to check.
             * @returns {void}
             */
            function checkLastSegment(node) {
                if (funcInfo.shouldCheck &&
                    funcInfo.codePath.currentSegments.some(isReachable)) {
                    context.report({
                        node,
                        loc: astUtils.getFunctionHeadLoc(node, sourceCode),
                        messageId: funcInfo.hasReturn ? "expectedAlways" : "expected",
                        data: {
                            name: astUtils.getFunctionNameWithKind(funcInfo.node)
                        }
                    });
                }
            }
            /**
             * Checks whether a node means a getter function.
             * @param {ASTNode} node a node to check.
             * @returns {boolean} if node means a getter, return true; else return false.
             */
            function isGetter(node) {
                const parent = node.parent;
                if (TARGET_NODE_TYPE.test(node.type) && node.body.type === "BlockStatement") {
                    if (parent.kind === "get") {
                        return true;
                    }
                    if (parent.type === "Property" && astUtils.getStaticPropertyName(parent) === "get" && parent.parent.type === "ObjectExpression") {
                        // Object.defineProperty()
                        if (parent.parent.parent.type === "CallExpression" &&
                            astUtils.getStaticPropertyName(parent.parent.parent.callee) === "defineProperty") {
                            return true;
                        }
                        // Object.defineProperties()
                        if (parent.parent.parent.type === "Property" &&
                            parent.parent.parent.parent.type === "ObjectExpression" &&
                            parent.parent.parent.parent.parent.type === "CallExpression" &&
                            astUtils.getStaticPropertyName(parent.parent.parent.parent.parent.callee) === "defineProperties") {
                            return true;
                        }
                    }
                }
                return false;
            }
            return {
                // Stacks this function's information.
                onCodePathStart(codePath, node) {
                    funcInfo = {
                        upper: funcInfo,
                        codePath,
                        hasReturn: false,
                        shouldCheck: isGetter(node),
                        node
                    };
                },
                // Pops this function's information.
                onCodePathEnd() {
                    funcInfo = funcInfo.upper;
                },
                // Checks the return statement is valid.
                ReturnStatement(node) {
                    if (funcInfo.shouldCheck) {
                        funcInfo.hasReturn = true;
                        // if allowImplicit: false, should also check node.argument
                        if (!options.allowImplicit && !node.argument) {
                            context.report({
                                node,
                                messageId: "expected",
                                data: {
                                    name: astUtils.getFunctionNameWithKind(funcInfo.node)
                                }
                            });
                        }
                    }
                },
                // Reports a given function if the last path is reachable.
                "FunctionExpression:exit": checkLastSegment,
                "ArrowFunctionExpression:exit": checkLastSegment
            };
        }
    };
},
/* eslint/lib/rules/global-require.js */ function _(require, module, exports, __esModule, __esExport) {
    const ACCEPTABLE_PARENTS = [
        "AssignmentExpression",
        "VariableDeclarator",
        "MemberExpression",
        "ExpressionStatement",
        "CallExpression",
        "ConditionalExpression",
        "Program",
        "VariableDeclaration",
        "ChainExpression"
    ];
    /**
     * Finds the eslint-scope reference in the given scope.
     * @param {Object} scope The scope to search.
     * @param {ASTNode} node The identifier node.
     * @returns {Reference|null} Returns the found reference or null if none were found.
     */
    function findReference(scope, node) {
        const references = scope.references.filter(reference => reference.identifier.range[0] === node.range[0] &&
            reference.identifier.range[1] === node.range[1]);
        /* istanbul ignore else: correctly returns null */
        if (references.length === 1) {
            return references[0];
        }
        return null;
    }
    /**
     * Checks if the given identifier node is shadowed in the given scope.
     * @param {Object} scope The current scope.
     * @param {ASTNode} node The identifier node to check.
     * @returns {boolean} Whether or not the name is shadowed.
     */
    function isShadowed(scope, node) {
        const reference = findReference(scope, node);
        return reference && reference.resolved && reference.resolved.defs.length > 0;
    }
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "require `require()` calls to be placed at top-level module scope",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/global-require"
            },
            schema: [],
            messages: {
                unexpected: "Unexpected require()."
            }
        },
        create(context) {
            return {
                CallExpression(node) {
                    const currentScope = context.getScope();
                    if (node.callee.name === "require" && !isShadowed(currentScope, node.callee)) {
                        const isGoodRequire = context.getAncestors().every(parent => ACCEPTABLE_PARENTS.indexOf(parent.type) > -1);
                        if (!isGoodRequire) {
                            context.report({ node, messageId: "unexpected" });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/grouped-accessor-pairs.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    /**
     * Property name if it can be computed statically, otherwise the list of the tokens of the key node.
     * @typedef {string|Token[]} Key
     */
    /**
     * Accessor nodes with the same key.
     * @typedef {Object} AccessorData
     * @property {Key} key Accessor's key
     * @property {ASTNode[]} getters List of getter nodes.
     * @property {ASTNode[]} setters List of setter nodes.
     */
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not the given lists represent the equal tokens in the same order.
     * Tokens are compared by their properties, not by instance.
     * @param {Token[]} left First list of tokens.
     * @param {Token[]} right Second list of tokens.
     * @returns {boolean} `true` if the lists have same tokens.
     */
    function areEqualTokenLists(left, right) {
        if (left.length !== right.length) {
            return false;
        }
        for (let i = 0; i < left.length; i++) {
            const leftToken = left[i], rightToken = right[i];
            if (leftToken.type !== rightToken.type || leftToken.value !== rightToken.value) {
                return false;
            }
        }
        return true;
    }
    /**
     * Checks whether or not the given keys are equal.
     * @param {Key} left First key.
     * @param {Key} right Second key.
     * @returns {boolean} `true` if the keys are equal.
     */
    function areEqualKeys(left, right) {
        if (typeof left === "string" && typeof right === "string") {
            // Statically computed names.
            return left === right;
        }
        if (Array.isArray(left) && Array.isArray(right)) {
            // Token lists.
            return areEqualTokenLists(left, right);
        }
        return false;
    }
    /**
     * Checks whether or not a given node is of an accessor kind ('get' or 'set').
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node is of an accessor kind.
     */
    function isAccessorKind(node) {
        return node.kind === "get" || node.kind === "set";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require grouped accessor pairs in object literals and classes",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/grouped-accessor-pairs"
            },
            schema: [
                {
                    enum: ["anyOrder", "getBeforeSet", "setBeforeGet"]
                }
            ],
            messages: {
                notGrouped: "Accessor pair {{ formerName }} and {{ latterName }} should be grouped.",
                invalidOrder: "Expected {{ latterName }} to be before {{ formerName }}."
            }
        },
        create(context) {
            const order = context.options[0] || "anyOrder";
            const sourceCode = context.getSourceCode();
            /**
             * Reports the given accessor pair.
             * @param {string} messageId messageId to report.
             * @param {ASTNode} formerNode getter/setter node that is defined before `latterNode`.
             * @param {ASTNode} latterNode getter/setter node that is defined after `formerNode`.
             * @returns {void}
             * @private
             */
            function report(messageId, formerNode, latterNode) {
                context.report({
                    node: latterNode,
                    messageId,
                    loc: astUtils.getFunctionHeadLoc(latterNode.value, sourceCode),
                    data: {
                        formerName: astUtils.getFunctionNameWithKind(formerNode.value),
                        latterName: astUtils.getFunctionNameWithKind(latterNode.value)
                    }
                });
            }
            /**
             * Creates a new `AccessorData` object for the given getter or setter node.
             * @param {ASTNode} node A getter or setter node.
             * @returns {AccessorData} New `AccessorData` object that contains the given node.
             * @private
             */
            function createAccessorData(node) {
                const name = astUtils.getStaticPropertyName(node);
                const key = (name !== null) ? name : sourceCode.getTokens(node.key);
                return {
                    key,
                    getters: node.kind === "get" ? [node] : [],
                    setters: node.kind === "set" ? [node] : []
                };
            }
            /**
             * Merges the given `AccessorData` object into the given accessors list.
             * @param {AccessorData[]} accessors The list to merge into.
             * @param {AccessorData} accessorData The object to merge.
             * @returns {AccessorData[]} The same instance with the merged object.
             * @private
             */
            function mergeAccessorData(accessors, accessorData) {
                const equalKeyElement = accessors.find(a => areEqualKeys(a.key, accessorData.key));
                if (equalKeyElement) {
                    equalKeyElement.getters.push(...accessorData.getters);
                    equalKeyElement.setters.push(...accessorData.setters);
                }
                else {
                    accessors.push(accessorData);
                }
                return accessors;
            }
            /**
             * Checks accessor pairs in the given list of nodes.
             * @param {ASTNode[]} nodes The list to check.
             * @param {Function} shouldCheck – Predicate that returns `true` if the node should be checked.
             * @returns {void}
             * @private
             */
            function checkList(nodes, shouldCheck) {
                const accessors = nodes
                    .filter(shouldCheck)
                    .filter(isAccessorKind)
                    .map(createAccessorData)
                    .reduce(mergeAccessorData, []);
                for (const { getters, setters } of accessors) {
                    // Don't report accessor properties that have duplicate getters or setters.
                    if (getters.length === 1 && setters.length === 1) {
                        const [getter] = getters, [setter] = setters, getterIndex = nodes.indexOf(getter), setterIndex = nodes.indexOf(setter), formerNode = getterIndex < setterIndex ? getter : setter, latterNode = getterIndex < setterIndex ? setter : getter;
                        if (Math.abs(getterIndex - setterIndex) > 1) {
                            report("notGrouped", formerNode, latterNode);
                        }
                        else if ((order === "getBeforeSet" && getterIndex > setterIndex) ||
                            (order === "setBeforeGet" && getterIndex < setterIndex)) {
                            report("invalidOrder", formerNode, latterNode);
                        }
                    }
                }
            }
            return {
                ObjectExpression(node) {
                    checkList(node.properties, n => n.type === "Property");
                },
                ClassBody(node) {
                    checkList(node.body, n => n.type === "MethodDefinition" && !n.static);
                    checkList(node.body, n => n.type === "MethodDefinition" && n.static);
                }
            };
        }
    };
},
/* eslint/lib/rules/guard-for-in.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require `for-in` loops to include an `if` statement",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/guard-for-in"
            },
            schema: [],
            messages: {
                wrap: "The body of a for-in should be wrapped in an if statement to filter unwanted properties from the prototype."
            }
        },
        create(context) {
            return {
                ForInStatement(node) {
                    const body = node.body;
                    // empty statement
                    if (body.type === "EmptyStatement") {
                        return;
                    }
                    // if statement
                    if (body.type === "IfStatement") {
                        return;
                    }
                    // empty block
                    if (body.type === "BlockStatement" && body.body.length === 0) {
                        return;
                    }
                    // block with just if statement
                    if (body.type === "BlockStatement" && body.body.length === 1 && body.body[0].type === "IfStatement") {
                        return;
                    }
                    // block that starts with if statement
                    if (body.type === "BlockStatement" && body.body.length >= 1 && body.body[0].type === "IfStatement") {
                        const i = body.body[0];
                        // ... whose consequent is a continue
                        if (i.consequent.type === "ContinueStatement") {
                            return;
                        }
                        // ... whose consequent is a block that contains only a continue
                        if (i.consequent.type === "BlockStatement" && i.consequent.body.length === 1 && i.consequent.body[0].type === "ContinueStatement") {
                            return;
                        }
                    }
                    context.report({ node, messageId: "wrap" });
                }
            };
        }
    };
},
/* eslint/lib/rules/handle-callback-err.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "require error handling in callbacks",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/handle-callback-err"
            },
            schema: [
                {
                    type: "string"
                }
            ],
            messages: {
                expected: "Expected error to be handled."
            }
        },
        create(context) {
            const errorArgument = context.options[0] || "err";
            /**
             * Checks if the given argument should be interpreted as a regexp pattern.
             * @param {string} stringToCheck The string which should be checked.
             * @returns {boolean} Whether or not the string should be interpreted as a pattern.
             */
            function isPattern(stringToCheck) {
                const firstChar = stringToCheck[0];
                return firstChar === "^";
            }
            /**
             * Checks if the given name matches the configured error argument.
             * @param {string} name The name which should be compared.
             * @returns {boolean} Whether or not the given name matches the configured error variable name.
             */
            function matchesConfiguredErrorName(name) {
                if (isPattern(errorArgument)) {
                    const regexp = new RegExp(errorArgument, "u");
                    return regexp.test(name);
                }
                return name === errorArgument;
            }
            /**
             * Get the parameters of a given function scope.
             * @param {Object} scope The function scope.
             * @returns {Array} All parameters of the given scope.
             */
            function getParameters(scope) {
                return scope.variables.filter(variable => variable.defs[0] && variable.defs[0].type === "Parameter");
            }
            /**
             * Check to see if we're handling the error object properly.
             * @param {ASTNode} node The AST node to check.
             * @returns {void}
             */
            function checkForError(node) {
                const scope = context.getScope(), parameters = getParameters(scope), firstParameter = parameters[0];
                if (firstParameter && matchesConfiguredErrorName(firstParameter.name)) {
                    if (firstParameter.references.length === 0) {
                        context.report({ node, messageId: "expected" });
                    }
                }
            }
            return {
                FunctionDeclaration: checkForError,
                FunctionExpression: checkForError,
                ArrowFunctionExpression: checkForError
            };
        }
    };
},
/* eslint/lib/rules/id-blacklist.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether the given node represents assignment target in a normal assignment or destructuring.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is assignment target.
     */
    function isAssignmentTarget(node) {
        const parent = node.parent;
        return (
        // normal assignment
        (parent.type === "AssignmentExpression" &&
            parent.left === node) ||
            // destructuring
            parent.type === "ArrayPattern" ||
            parent.type === "RestElement" ||
            (parent.type === "Property" &&
                parent.value === node &&
                parent.parent.type === "ObjectPattern") ||
            (parent.type === "AssignmentPattern" &&
                parent.left === node));
    }
    /**
     * Checks whether the given node represents an imported name that is renamed in the same import/export specifier.
     *
     * Examples:
     * import { a as b } from 'mod'; // node `a` is renamed import
     * export { a as b } from 'mod'; // node `a` is renamed import
     * @param {ASTNode} node `Identifier` node to check.
     * @returns {boolean} `true` if the node is a renamed import.
     */
    function isRenamedImport(node) {
        const parent = node.parent;
        return ((parent.type === "ImportSpecifier" &&
            parent.imported !== parent.local &&
            parent.imported === node) ||
            (parent.type === "ExportSpecifier" &&
                parent.parent.source && // re-export
                parent.local !== parent.exported &&
                parent.local === node));
    }
    /**
     * Checks whether the given node is a renamed identifier node in an ObjectPattern destructuring.
     *
     * Examples:
     * const { a : b } = foo; // node `a` is renamed node.
     * @param {ASTNode} node `Identifier` node to check.
     * @returns {boolean} `true` if the node is a renamed node in an ObjectPattern destructuring.
     */
    function isRenamedInDestructuring(node) {
        const parent = node.parent;
        return ((!parent.computed &&
            parent.type === "Property" &&
            parent.parent.type === "ObjectPattern" &&
            parent.value !== node &&
            parent.key === node));
    }
    /**
     * Checks whether the given node represents shorthand definition of a property in an object literal.
     * @param {ASTNode} node `Identifier` node to check.
     * @returns {boolean} `true` if the node is a shorthand property definition.
     */
    function isShorthandPropertyDefinition(node) {
        const parent = node.parent;
        return (parent.type === "Property" &&
            parent.parent.type === "ObjectExpression" &&
            parent.shorthand);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: ["id-denylist"],
            type: "suggestion",
            docs: {
                description: "disallow specified identifiers",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/id-blacklist"
            },
            schema: {
                type: "array",
                items: {
                    type: "string"
                },
                uniqueItems: true
            },
            messages: {
                restricted: "Identifier '{{name}}' is restricted."
            }
        },
        create(context) {
            const denyList = new Set(context.options);
            const reportedNodes = new Set();
            let globalScope;
            /**
             * Checks whether the given name is restricted.
             * @param {string} name The name to check.
             * @returns {boolean} `true` if the name is restricted.
             * @private
             */
            function isRestricted(name) {
                return denyList.has(name);
            }
            /**
             * Checks whether the given node represents a reference to a global variable that is not declared in the source code.
             * These identifiers will be allowed, as it is assumed that user has no control over the names of external global variables.
             * @param {ASTNode} node `Identifier` node to check.
             * @returns {boolean} `true` if the node is a reference to a global variable.
             */
            function isReferenceToGlobalVariable(node) {
                const variable = globalScope.set.get(node.name);
                return variable && variable.defs.length === 0 &&
                    variable.references.some(ref => ref.identifier === node);
            }
            /**
             * Determines whether the given node should be checked.
             * @param {ASTNode} node `Identifier` node.
             * @returns {boolean} `true` if the node should be checked.
             */
            function shouldCheck(node) {
                const parent = node.parent;
                /*
                 * Member access has special rules for checking property names.
                 * Read access to a property with a restricted name is allowed, because it can be on an object that user has no control over.
                 * Write access isn't allowed, because it potentially creates a new property with a restricted name.
                 */
                if (parent.type === "MemberExpression" &&
                    parent.property === node &&
                    !parent.computed) {
                    return isAssignmentTarget(parent);
                }
                return (parent.type !== "CallExpression" &&
                    parent.type !== "NewExpression" &&
                    !isRenamedImport(node) &&
                    !isRenamedInDestructuring(node) &&
                    !(isReferenceToGlobalVariable(node) &&
                        !isShorthandPropertyDefinition(node)));
            }
            /**
             * Reports an AST node as a rule violation.
             * @param {ASTNode} node The node to report.
             * @returns {void}
             * @private
             */
            function report(node) {
                if (!reportedNodes.has(node)) {
                    context.report({
                        node,
                        messageId: "restricted",
                        data: {
                            name: node.name
                        }
                    });
                    reportedNodes.add(node);
                }
            }
            return {
                Program() {
                    globalScope = context.getScope();
                },
                Identifier(node) {
                    if (isRestricted(node.name) && shouldCheck(node)) {
                        report(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/id-denylist.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether the given node represents assignment target in a normal assignment or destructuring.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is assignment target.
     */
    function isAssignmentTarget(node) {
        const parent = node.parent;
        return (
        // normal assignment
        (parent.type === "AssignmentExpression" &&
            parent.left === node) ||
            // destructuring
            parent.type === "ArrayPattern" ||
            parent.type === "RestElement" ||
            (parent.type === "Property" &&
                parent.value === node &&
                parent.parent.type === "ObjectPattern") ||
            (parent.type === "AssignmentPattern" &&
                parent.left === node));
    }
    /**
     * Checks whether the given node represents an imported name that is renamed in the same import/export specifier.
     *
     * Examples:
     * import { a as b } from 'mod'; // node `a` is renamed import
     * export { a as b } from 'mod'; // node `a` is renamed import
     * @param {ASTNode} node `Identifier` node to check.
     * @returns {boolean} `true` if the node is a renamed import.
     */
    function isRenamedImport(node) {
        const parent = node.parent;
        return ((parent.type === "ImportSpecifier" &&
            parent.imported !== parent.local &&
            parent.imported === node) ||
            (parent.type === "ExportSpecifier" &&
                parent.parent.source && // re-export
                parent.local !== parent.exported &&
                parent.local === node));
    }
    /**
     * Checks whether the given node is a renamed identifier node in an ObjectPattern destructuring.
     *
     * Examples:
     * const { a : b } = foo; // node `a` is renamed node.
     * @param {ASTNode} node `Identifier` node to check.
     * @returns {boolean} `true` if the node is a renamed node in an ObjectPattern destructuring.
     */
    function isRenamedInDestructuring(node) {
        const parent = node.parent;
        return ((!parent.computed &&
            parent.type === "Property" &&
            parent.parent.type === "ObjectPattern" &&
            parent.value !== node &&
            parent.key === node));
    }
    /**
     * Checks whether the given node represents shorthand definition of a property in an object literal.
     * @param {ASTNode} node `Identifier` node to check.
     * @returns {boolean} `true` if the node is a shorthand property definition.
     */
    function isShorthandPropertyDefinition(node) {
        const parent = node.parent;
        return (parent.type === "Property" &&
            parent.parent.type === "ObjectExpression" &&
            parent.shorthand);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow specified identifiers",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/id-denylist"
            },
            schema: {
                type: "array",
                items: {
                    type: "string"
                },
                uniqueItems: true
            },
            messages: {
                restricted: "Identifier '{{name}}' is restricted."
            }
        },
        create(context) {
            const denyList = new Set(context.options);
            const reportedNodes = new Set();
            let globalScope;
            /**
             * Checks whether the given name is restricted.
             * @param {string} name The name to check.
             * @returns {boolean} `true` if the name is restricted.
             * @private
             */
            function isRestricted(name) {
                return denyList.has(name);
            }
            /**
             * Checks whether the given node represents a reference to a global variable that is not declared in the source code.
             * These identifiers will be allowed, as it is assumed that user has no control over the names of external global variables.
             * @param {ASTNode} node `Identifier` node to check.
             * @returns {boolean} `true` if the node is a reference to a global variable.
             */
            function isReferenceToGlobalVariable(node) {
                const variable = globalScope.set.get(node.name);
                return variable && variable.defs.length === 0 &&
                    variable.references.some(ref => ref.identifier === node);
            }
            /**
             * Determines whether the given node should be checked.
             * @param {ASTNode} node `Identifier` node.
             * @returns {boolean} `true` if the node should be checked.
             */
            function shouldCheck(node) {
                const parent = node.parent;
                /*
                 * Member access has special rules for checking property names.
                 * Read access to a property with a restricted name is allowed, because it can be on an object that user has no control over.
                 * Write access isn't allowed, because it potentially creates a new property with a restricted name.
                 */
                if (parent.type === "MemberExpression" &&
                    parent.property === node &&
                    !parent.computed) {
                    return isAssignmentTarget(parent);
                }
                return (parent.type !== "CallExpression" &&
                    parent.type !== "NewExpression" &&
                    !isRenamedImport(node) &&
                    !isRenamedInDestructuring(node) &&
                    !(isReferenceToGlobalVariable(node) &&
                        !isShorthandPropertyDefinition(node)));
            }
            /**
             * Reports an AST node as a rule violation.
             * @param {ASTNode} node The node to report.
             * @returns {void}
             * @private
             */
            function report(node) {
                if (!reportedNodes.has(node)) {
                    context.report({
                        node,
                        messageId: "restricted",
                        data: {
                            name: node.name
                        }
                    });
                    reportedNodes.add(node);
                }
            }
            return {
                Program() {
                    globalScope = context.getScope();
                },
                Identifier(node) {
                    if (isRestricted(node.name) && shouldCheck(node)) {
                        report(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/id-length.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce minimum and maximum identifier lengths",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/id-length"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        min: {
                            type: "integer",
                            default: 2
                        },
                        max: {
                            type: "integer"
                        },
                        exceptions: {
                            type: "array",
                            uniqueItems: true,
                            items: {
                                type: "string"
                            }
                        },
                        exceptionPatterns: {
                            type: "array",
                            uniqueItems: true,
                            items: {
                                type: "string"
                            }
                        },
                        properties: {
                            enum: ["always", "never"]
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                tooShort: "Identifier name '{{name}}' is too short (< {{min}}).",
                tooLong: "Identifier name '{{name}}' is too long (> {{max}})."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const minLength = typeof options.min !== "undefined" ? options.min : 2;
            const maxLength = typeof options.max !== "undefined" ? options.max : Infinity;
            const properties = options.properties !== "never";
            const exceptions = new Set(options.exceptions);
            const exceptionPatterns = (options.exceptionPatterns || []).map(pattern => new RegExp(pattern, "u"));
            const reportedNode = new Set();
            /**
             * Checks if a string matches the provided exception patterns
             * @param {string} name The string to check.
             * @returns {boolean} if the string is a match
             * @private
             */
            function matchesExceptionPattern(name) {
                return exceptionPatterns.some(pattern => pattern.test(name));
            }
            const SUPPORTED_EXPRESSIONS = {
                MemberExpression: properties && function (parent) {
                    return !parent.computed && (
                    // regular property assignment
                    (parent.parent.left === parent && parent.parent.type === "AssignmentExpression" ||
                        // or the last identifier in an ObjectPattern destructuring
                        parent.parent.type === "Property" && parent.parent.value === parent &&
                            parent.parent.parent.type === "ObjectPattern" && parent.parent.parent.parent.left === parent.parent.parent));
                },
                AssignmentPattern(parent, node) {
                    return parent.left === node;
                },
                VariableDeclarator(parent, node) {
                    return parent.id === node;
                },
                Property(parent, node) {
                    if (parent.parent.type === "ObjectPattern") {
                        return (parent.value !== parent.key && parent.value === node ||
                            parent.value === parent.key && parent.key === node && properties);
                    }
                    return properties && !parent.computed && parent.key === node;
                },
                ImportDefaultSpecifier: true,
                RestElement: true,
                FunctionExpression: true,
                ArrowFunctionExpression: true,
                ClassDeclaration: true,
                FunctionDeclaration: true,
                MethodDefinition: true,
                CatchClause: true,
                ArrayPattern: true
            };
            return {
                Identifier(node) {
                    const name = node.name;
                    const parent = node.parent;
                    const isShort = name.length < minLength;
                    const isLong = name.length > maxLength;
                    if (!(isShort || isLong) || exceptions.has(name) || matchesExceptionPattern(name)) {
                        return; // Nothing to report
                    }
                    const isValidExpression = SUPPORTED_EXPRESSIONS[parent.type];
                    if (isValidExpression && !reportedNode.has(node) && (isValidExpression === true || isValidExpression(parent, node))) {
                        reportedNode.add(node);
                        context.report({
                            node,
                            messageId: isShort ? "tooShort" : "tooLong",
                            data: { name, min: minLength, max: maxLength }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/id-match.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require identifiers to match a specified regular expression",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/id-match"
            },
            schema: [
                {
                    type: "string"
                },
                {
                    type: "object",
                    properties: {
                        properties: {
                            type: "boolean",
                            default: false
                        },
                        onlyDeclarations: {
                            type: "boolean",
                            default: false
                        },
                        ignoreDestructuring: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                notMatch: "Identifier '{{name}}' does not match the pattern '{{pattern}}'."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Options
            //--------------------------------------------------------------------------
            const pattern = context.options[0] || "^.+$", regexp = new RegExp(pattern, "u");
            const options = context.options[1] || {}, properties = !!options.properties, onlyDeclarations = !!options.onlyDeclarations, ignoreDestructuring = !!options.ignoreDestructuring;
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            // contains reported nodes to avoid reporting twice on destructuring with shorthand notation
            const reported = new Map();
            const ALLOWED_PARENT_TYPES = new Set(["CallExpression", "NewExpression"]);
            const DECLARATION_TYPES = new Set(["FunctionDeclaration", "VariableDeclarator"]);
            const IMPORT_TYPES = new Set(["ImportSpecifier", "ImportNamespaceSpecifier", "ImportDefaultSpecifier"]);
            /**
             * Checks if a string matches the provided pattern
             * @param {string} name The string to check.
             * @returns {boolean} if the string is a match
             * @private
             */
            function isInvalid(name) {
                return !regexp.test(name);
            }
            /**
             * Checks if a parent of a node is an ObjectPattern.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} if the node is inside an ObjectPattern
             * @private
             */
            function isInsideObjectPattern(node) {
                let { parent } = node;
                while (parent) {
                    if (parent.type === "ObjectPattern") {
                        return true;
                    }
                    parent = parent.parent;
                }
                return false;
            }
            /**
             * Verifies if we should report an error or not based on the effective
             * parent node and the identifier name.
             * @param {ASTNode} effectiveParent The effective parent node of the node to be reported
             * @param {string} name The identifier name of the identifier node
             * @returns {boolean} whether an error should be reported or not
             */
            function shouldReport(effectiveParent, name) {
                return (!onlyDeclarations || DECLARATION_TYPES.has(effectiveParent.type)) &&
                    !ALLOWED_PARENT_TYPES.has(effectiveParent.type) && isInvalid(name);
            }
            /**
             * Reports an AST node as a rule violation.
             * @param {ASTNode} node The node to report.
             * @returns {void}
             * @private
             */
            function report(node) {
                if (!reported.has(node)) {
                    context.report({
                        node,
                        messageId: "notMatch",
                        data: {
                            name: node.name,
                            pattern
                        }
                    });
                    reported.set(node, true);
                }
            }
            return {
                Identifier(node) {
                    const name = node.name, parent = node.parent, effectiveParent = (parent.type === "MemberExpression") ? parent.parent : parent;
                    if (parent.type === "MemberExpression") {
                        if (!properties) {
                            return;
                        }
                        // Always check object names
                        if (parent.object.type === "Identifier" &&
                            parent.object.name === name) {
                            if (isInvalid(name)) {
                                report(node);
                            }
                            // Report AssignmentExpressions left side's assigned variable id
                        }
                        else if (effectiveParent.type === "AssignmentExpression" &&
                            effectiveParent.left.type === "MemberExpression" &&
                            effectiveParent.left.property.name === node.name) {
                            if (isInvalid(name)) {
                                report(node);
                            }
                            // Report AssignmentExpressions only if they are the left side of the assignment
                        }
                        else if (effectiveParent.type === "AssignmentExpression" && effectiveParent.right.type !== "MemberExpression") {
                            if (isInvalid(name)) {
                                report(node);
                            }
                        }
                        /*
                         * Properties have their own rules, and
                         * AssignmentPattern nodes can be treated like Properties:
                         * e.g.: const { no_camelcased = false } = bar;
                         */
                    }
                    else if (parent.type === "Property" || parent.type === "AssignmentPattern") {
                        if (parent.parent && parent.parent.type === "ObjectPattern") {
                            if (parent.shorthand && parent.value.left && isInvalid(name)) {
                                report(node);
                            }
                            const assignmentKeyEqualsValue = parent.key.name === parent.value.name;
                            // prevent checking righthand side of destructured object
                            if (!assignmentKeyEqualsValue && parent.key === node) {
                                return;
                            }
                            const valueIsInvalid = parent.value.name && isInvalid(name);
                            // ignore destructuring if the option is set, unless a new identifier is created
                            if (valueIsInvalid && !(assignmentKeyEqualsValue && ignoreDestructuring)) {
                                report(node);
                            }
                        }
                        // never check properties or always ignore destructuring
                        if (!properties || (ignoreDestructuring && isInsideObjectPattern(node))) {
                            return;
                        }
                        // don't check right hand side of AssignmentExpression to prevent duplicate warnings
                        if (parent.right !== node && shouldReport(effectiveParent, name)) {
                            report(node);
                        }
                        // Check if it's an import specifier
                    }
                    else if (IMPORT_TYPES.has(parent.type)) {
                        // Report only if the local imported identifier is invalid
                        if (parent.local && parent.local.name === node.name && isInvalid(name)) {
                            report(node);
                        }
                        // Report anything that is invalid that isn't a CallExpression
                    }
                    else if (shouldReport(effectiveParent, name)) {
                        report(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/implicit-arrow-linebreak.js */ function _(require, module, exports, __esModule, __esExport) {
    const { isCommentToken, isNotOpeningParenToken } = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce the location of arrow function bodies",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/implicit-arrow-linebreak"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["beside", "below"]
                }
            ],
            messages: {
                expected: "Expected a linebreak before this expression.",
                unexpected: "Expected no linebreak before this expression."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const option = context.options[0] || "beside";
            /**
             * Validates the location of an arrow function body
             * @param {ASTNode} node The arrow function body
             * @returns {void}
             */
            function validateExpression(node) {
                if (node.body.type === "BlockStatement") {
                    return;
                }
                const arrowToken = sourceCode.getTokenBefore(node.body, isNotOpeningParenToken);
                const firstTokenOfBody = sourceCode.getTokenAfter(arrowToken);
                if (arrowToken.loc.end.line === firstTokenOfBody.loc.start.line && option === "below") {
                    context.report({
                        node: firstTokenOfBody,
                        messageId: "expected",
                        fix: fixer => fixer.insertTextBefore(firstTokenOfBody, "\n")
                    });
                }
                else if (arrowToken.loc.end.line !== firstTokenOfBody.loc.start.line && option === "beside") {
                    context.report({
                        node: firstTokenOfBody,
                        messageId: "unexpected",
                        fix(fixer) {
                            if (sourceCode.getFirstTokenBetween(arrowToken, firstTokenOfBody, { includeComments: true, filter: isCommentToken })) {
                                return null;
                            }
                            return fixer.replaceTextRange([arrowToken.range[1], firstTokenOfBody.range[0]], " ");
                        }
                    });
                }
            }
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                ArrowFunctionExpression: node => validateExpression(node)
            };
        }
    };
},
/* eslint/lib/rules/indent.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const createTree = require(451) /* functional-red-black-tree */;
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const KNOWN_NODES = new Set([
        "AssignmentExpression",
        "AssignmentPattern",
        "ArrayExpression",
        "ArrayPattern",
        "ArrowFunctionExpression",
        "AwaitExpression",
        "BlockStatement",
        "BinaryExpression",
        "BreakStatement",
        "CallExpression",
        "CatchClause",
        "ChainExpression",
        "ClassBody",
        "ClassDeclaration",
        "ClassExpression",
        "ConditionalExpression",
        "ContinueStatement",
        "DoWhileStatement",
        "DebuggerStatement",
        "EmptyStatement",
        "ExperimentalRestProperty",
        "ExperimentalSpreadProperty",
        "ExpressionStatement",
        "ForStatement",
        "ForInStatement",
        "ForOfStatement",
        "FunctionDeclaration",
        "FunctionExpression",
        "Identifier",
        "IfStatement",
        "Literal",
        "LabeledStatement",
        "LogicalExpression",
        "MemberExpression",
        "MetaProperty",
        "MethodDefinition",
        "NewExpression",
        "ObjectExpression",
        "ObjectPattern",
        "Program",
        "Property",
        "RestElement",
        "ReturnStatement",
        "SequenceExpression",
        "SpreadElement",
        "Super",
        "SwitchCase",
        "SwitchStatement",
        "TaggedTemplateExpression",
        "TemplateElement",
        "TemplateLiteral",
        "ThisExpression",
        "ThrowStatement",
        "TryStatement",
        "UnaryExpression",
        "UpdateExpression",
        "VariableDeclaration",
        "VariableDeclarator",
        "WhileStatement",
        "WithStatement",
        "YieldExpression",
        "JSXFragment",
        "JSXOpeningFragment",
        "JSXClosingFragment",
        "JSXIdentifier",
        "JSXNamespacedName",
        "JSXMemberExpression",
        "JSXEmptyExpression",
        "JSXExpressionContainer",
        "JSXElement",
        "JSXClosingElement",
        "JSXOpeningElement",
        "JSXAttribute",
        "JSXSpreadAttribute",
        "JSXText",
        "ExportDefaultDeclaration",
        "ExportNamedDeclaration",
        "ExportAllDeclaration",
        "ExportSpecifier",
        "ImportDeclaration",
        "ImportSpecifier",
        "ImportDefaultSpecifier",
        "ImportNamespaceSpecifier",
        "ImportExpression"
    ]);
    /*
     * General rule strategy:
     * 1. An OffsetStorage instance stores a map of desired offsets, where each token has a specified offset from another
     *    specified token or to the first column.
     * 2. As the AST is traversed, modify the desired offsets of tokens accordingly. For example, when entering a
     *    BlockStatement, offset all of the tokens in the BlockStatement by 1 indent level from the opening curly
     *    brace of the BlockStatement.
     * 3. After traversing the AST, calculate the expected indentation levels of every token according to the
     *    OffsetStorage container.
     * 4. For each line, compare the expected indentation of the first token to the actual indentation in the file,
     *    and report the token if the two values are not equal.
     */
    /**
     * A mutable balanced binary search tree that stores (key, value) pairs. The keys are numeric, and must be unique.
     * This is intended to be a generic wrapper around a balanced binary search tree library, so that the underlying implementation
     * can easily be swapped out.
     */
    class BinarySearchTree {
        /**
         * Creates an empty tree
         */
        constructor() {
            this._rbTree = createTree();
        }
        /**
         * Inserts an entry into the tree.
         * @param {number} key The entry's key
         * @param {*} value The entry's value
         * @returns {void}
         */
        insert(key, value) {
            const iterator = this._rbTree.find(key);
            if (iterator.valid) {
                this._rbTree = iterator.update(value);
            }
            else {
                this._rbTree = this._rbTree.insert(key, value);
            }
        }
        /**
         * Finds the entry with the largest key less than or equal to the provided key
         * @param {number} key The provided key
         * @returns {{key: number, value: *}|null} The found entry, or null if no such entry exists.
         */
        findLe(key) {
            const iterator = this._rbTree.le(key);
            return iterator && { key: iterator.key, value: iterator.value };
        }
        /**
         * Deletes all of the keys in the interval [start, end)
         * @param {number} start The start of the range
         * @param {number} end The end of the range
         * @returns {void}
         */
        deleteRange(start, end) {
            // Exit without traversing the tree if the range has zero size.
            if (start === end) {
                return;
            }
            const iterator = this._rbTree.ge(start);
            while (iterator.valid && iterator.key < end) {
                this._rbTree = this._rbTree.remove(iterator.key);
                iterator.next();
            }
        }
    }
    /**
     * A helper class to get token-based info related to indentation
     */
    class TokenInfo {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {SourceCode} sourceCode A SourceCode object
         */
        constructor(sourceCode) {
            this.sourceCode = sourceCode;
            this.firstTokensByLineNumber = sourceCode.tokensAndComments.reduce((map, token) => {
                if (!map.has(token.loc.start.line)) {
                    map.set(token.loc.start.line, token);
                }
                if (!map.has(token.loc.end.line) && sourceCode.text.slice(token.range[1] - token.loc.end.column, token.range[1]).trim()) {
                    map.set(token.loc.end.line, token);
                }
                return map;
            }, new Map());
        }
        /**
         * Gets the first token on a given token's line
         * @param {Token|ASTNode} token a node or token
         * @returns {Token} The first token on the given line
         */
        getFirstTokenOfLine(token) {
            return this.firstTokensByLineNumber.get(token.loc.start.line);
        }
        /**
         * Determines whether a token is the first token in its line
         * @param {Token} token The token
         * @returns {boolean} `true` if the token is the first on its line
         */
        isFirstTokenOfLine(token) {
            return this.getFirstTokenOfLine(token) === token;
        }
        /**
         * Get the actual indent of a token
         * @param {Token} token Token to examine. This should be the first token on its line.
         * @returns {string} The indentation characters that precede the token
         */
        getTokenIndent(token) {
            return this.sourceCode.text.slice(token.range[0] - token.loc.start.column, token.range[0]);
        }
    }
    /**
     * A class to store information on desired offsets of tokens from each other
     */
    class OffsetStorage {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {TokenInfo} tokenInfo a TokenInfo instance
         * @param {number} indentSize The desired size of each indentation level
         * @param {string} indentType The indentation character
         */
        constructor(tokenInfo, indentSize, indentType) {
            this._tokenInfo = tokenInfo;
            this._indentSize = indentSize;
            this._indentType = indentType;
            this._tree = new BinarySearchTree();
            this._tree.insert(0, { offset: 0, from: null, force: false });
            this._lockedFirstTokens = new WeakMap();
            this._desiredIndentCache = new WeakMap();
            this._ignoredTokens = new WeakSet();
        }
        _getOffsetDescriptor(token) {
            return this._tree.findLe(token.range[0]).value;
        }
        /**
         * Sets the offset column of token B to match the offset column of token A.
         * **WARNING**: This matches a *column*, even if baseToken is not the first token on its line. In
         * most cases, `setDesiredOffset` should be used instead.
         * @param {Token} baseToken The first token
         * @param {Token} offsetToken The second token, whose offset should be matched to the first token
         * @returns {void}
         */
        matchOffsetOf(baseToken, offsetToken) {
            /*
             * lockedFirstTokens is a map from a token whose indentation is controlled by the "first" option to
             * the token that it depends on. For example, with the `ArrayExpression: first` option, the first
             * token of each element in the array after the first will be mapped to the first token of the first
             * element. The desired indentation of each of these tokens is computed based on the desired indentation
             * of the "first" element, rather than through the normal offset mechanism.
             */
            this._lockedFirstTokens.set(offsetToken, baseToken);
        }
        /**
         * Sets the desired offset of a token.
         *
         * This uses a line-based offset collapsing behavior to handle tokens on the same line.
         * For example, consider the following two cases:
         *
         * (
         *     [
         *         bar
         *     ]
         * )
         *
         * ([
         *     bar
         * ])
         *
         * Based on the first case, it's clear that the `bar` token needs to have an offset of 1 indent level (4 spaces) from
         * the `[` token, and the `[` token has to have an offset of 1 indent level from the `(` token. Since the `(` token is
         * the first on its line (with an indent of 0 spaces), the `bar` token needs to be offset by 2 indent levels (8 spaces)
         * from the start of its line.
         *
         * However, in the second case `bar` should only be indented by 4 spaces. This is because the offset of 1 indent level
         * between the `(` and the `[` tokens gets "collapsed" because the two tokens are on the same line. As a result, the
         * `(` token is mapped to the `[` token with an offset of 0, and the rule correctly decides that `bar` should be indented
         * by 1 indent level from the start of the line.
         *
         * This is useful because rule listeners can usually just call `setDesiredOffset` for all the tokens in the node,
         * without needing to check which lines those tokens are on.
         *
         * Note that since collapsing only occurs when two tokens are on the same line, there are a few cases where non-intuitive
         * behavior can occur. For example, consider the following cases:
         *
         * foo(
         * ).
         *     bar(
         *         baz
         *     )
         *
         * foo(
         * ).bar(
         *     baz
         * )
         *
         * Based on the first example, it would seem that `bar` should be offset by 1 indent level from `foo`, and `baz`
         * should be offset by 1 indent level from `bar`. However, this is not correct, because it would result in `baz`
         * being indented by 2 indent levels in the second case (since `foo`, `bar`, and `baz` are all on separate lines, no
         * collapsing would occur).
         *
         * Instead, the correct way would be to offset `baz` by 1 level from `bar`, offset `bar` by 1 level from the `)`, and
         * offset the `)` by 0 levels from `foo`. This ensures that the offset between `bar` and the `)` are correctly collapsed
         * in the second case.
         * @param {Token} token The token
         * @param {Token} fromToken The token that `token` should be offset from
         * @param {number} offset The desired indent level
         * @returns {void}
         */
        setDesiredOffset(token, fromToken, offset) {
            return this.setDesiredOffsets(token.range, fromToken, offset);
        }
        /**
         * Sets the desired offset of all tokens in a range
         * It's common for node listeners in this file to need to apply the same offset to a large, contiguous range of tokens.
         * Moreover, the offset of any given token is usually updated multiple times (roughly once for each node that contains
         * it). This means that the offset of each token is updated O(AST depth) times.
         * It would not be performant to store and update the offsets for each token independently, because the rule would end
         * up having a time complexity of O(number of tokens * AST depth), which is quite slow for large files.
         *
         * Instead, the offset tree is represented as a collection of contiguous offset ranges in a file. For example, the following
         * list could represent the state of the offset tree at a given point:
         *
         * * Tokens starting in the interval [0, 15) are aligned with the beginning of the file
         * * Tokens starting in the interval [15, 30) are offset by 1 indent level from the `bar` token
         * * Tokens starting in the interval [30, 43) are offset by 1 indent level from the `foo` token
         * * Tokens starting in the interval [43, 820) are offset by 2 indent levels from the `bar` token
         * * Tokens starting in the interval [820, ∞) are offset by 1 indent level from the `baz` token
         *
         * The `setDesiredOffsets` methods inserts ranges like the ones above. The third line above would be inserted by using:
         * `setDesiredOffsets([30, 43], fooToken, 1);`
         * @param {[number, number]} range A [start, end] pair. All tokens with range[0] <= token.start < range[1] will have the offset applied.
         * @param {Token} fromToken The token that this is offset from
         * @param {number} offset The desired indent level
         * @param {boolean} force `true` if this offset should not use the normal collapsing behavior. This should almost always be false.
         * @returns {void}
         */
        setDesiredOffsets(range, fromToken, offset, force) {
            /*
             * Offset ranges are stored as a collection of nodes, where each node maps a numeric key to an offset
             * descriptor. The tree for the example above would have the following nodes:
             *
             * * key: 0, value: { offset: 0, from: null }
             * * key: 15, value: { offset: 1, from: barToken }
             * * key: 30, value: { offset: 1, from: fooToken }
             * * key: 43, value: { offset: 2, from: barToken }
             * * key: 820, value: { offset: 1, from: bazToken }
             *
             * To find the offset descriptor for any given token, one needs to find the node with the largest key
             * which is <= token.start. To make this operation fast, the nodes are stored in a balanced binary
             * search tree indexed by key.
             */
            const descriptorToInsert = { offset, from: fromToken, force };
            const descriptorAfterRange = this._tree.findLe(range[1]).value;
            const fromTokenIsInRange = fromToken && fromToken.range[0] >= range[0] && fromToken.range[1] <= range[1];
            const fromTokenDescriptor = fromTokenIsInRange && this._getOffsetDescriptor(fromToken);
            // First, remove any existing nodes in the range from the tree.
            this._tree.deleteRange(range[0] + 1, range[1]);
            // Insert a new node into the tree for this range
            this._tree.insert(range[0], descriptorToInsert);
            /*
             * To avoid circular offset dependencies, keep the `fromToken` token mapped to whatever it was mapped to previously,
             * even if it's in the current range.
             */
            if (fromTokenIsInRange) {
                this._tree.insert(fromToken.range[0], fromTokenDescriptor);
                this._tree.insert(fromToken.range[1], descriptorToInsert);
            }
            /*
             * To avoid modifying the offset of tokens after the range, insert another node to keep the offset of the following
             * tokens the same as it was before.
             */
            this._tree.insert(range[1], descriptorAfterRange);
        }
        /**
         * Gets the desired indent of a token
         * @param {Token} token The token
         * @returns {string} The desired indent of the token
         */
        getDesiredIndent(token) {
            if (!this._desiredIndentCache.has(token)) {
                if (this._ignoredTokens.has(token)) {
                    /*
                     * If the token is ignored, use the actual indent of the token as the desired indent.
                     * This ensures that no errors are reported for this token.
                     */
                    this._desiredIndentCache.set(token, this._tokenInfo.getTokenIndent(token));
                }
                else if (this._lockedFirstTokens.has(token)) {
                    const firstToken = this._lockedFirstTokens.get(token);
                    this._desiredIndentCache.set(token, 
                    // (indentation for the first element's line)
                    this.getDesiredIndent(this._tokenInfo.getFirstTokenOfLine(firstToken)) +
                        // (space between the start of the first element's line and the first element)
                        this._indentType.repeat(firstToken.loc.start.column - this._tokenInfo.getFirstTokenOfLine(firstToken).loc.start.column));
                }
                else {
                    const offsetInfo = this._getOffsetDescriptor(token);
                    const offset = (offsetInfo.from &&
                        offsetInfo.from.loc.start.line === token.loc.start.line &&
                        !/^\s*?\n/u.test(token.value) &&
                        !offsetInfo.force) ? 0 : offsetInfo.offset * this._indentSize;
                    this._desiredIndentCache.set(token, (offsetInfo.from ? this.getDesiredIndent(offsetInfo.from) : "") + this._indentType.repeat(offset));
                }
            }
            return this._desiredIndentCache.get(token);
        }
        /**
         * Ignores a token, preventing it from being reported.
         * @param {Token} token The token
         * @returns {void}
         */
        ignoreToken(token) {
            if (this._tokenInfo.isFirstTokenOfLine(token)) {
                this._ignoredTokens.add(token);
            }
        }
        /**
         * Gets the first token that the given token's indentation is dependent on
         * @param {Token} token The token
         * @returns {Token} The token that the given token depends on, or `null` if the given token is at the top level
         */
        getFirstDependency(token) {
            return this._getOffsetDescriptor(token).from;
        }
    }
    const ELEMENT_LIST_SCHEMA = {
        oneOf: [
            {
                type: "integer",
                minimum: 0
            },
            {
                enum: ["first", "off"]
            }
        ]
    };
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent indentation",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/indent"
            },
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["tab"]
                        },
                        {
                            type: "integer",
                            minimum: 0
                        }
                    ]
                },
                {
                    type: "object",
                    properties: {
                        SwitchCase: {
                            type: "integer",
                            minimum: 0,
                            default: 0
                        },
                        VariableDeclarator: {
                            oneOf: [
                                ELEMENT_LIST_SCHEMA,
                                {
                                    type: "object",
                                    properties: {
                                        var: ELEMENT_LIST_SCHEMA,
                                        let: ELEMENT_LIST_SCHEMA,
                                        const: ELEMENT_LIST_SCHEMA
                                    },
                                    additionalProperties: false
                                }
                            ]
                        },
                        outerIIFEBody: {
                            oneOf: [
                                {
                                    type: "integer",
                                    minimum: 0
                                },
                                {
                                    enum: ["off"]
                                }
                            ]
                        },
                        MemberExpression: {
                            oneOf: [
                                {
                                    type: "integer",
                                    minimum: 0
                                },
                                {
                                    enum: ["off"]
                                }
                            ]
                        },
                        FunctionDeclaration: {
                            type: "object",
                            properties: {
                                parameters: ELEMENT_LIST_SCHEMA,
                                body: {
                                    type: "integer",
                                    minimum: 0
                                }
                            },
                            additionalProperties: false
                        },
                        FunctionExpression: {
                            type: "object",
                            properties: {
                                parameters: ELEMENT_LIST_SCHEMA,
                                body: {
                                    type: "integer",
                                    minimum: 0
                                }
                            },
                            additionalProperties: false
                        },
                        CallExpression: {
                            type: "object",
                            properties: {
                                arguments: ELEMENT_LIST_SCHEMA
                            },
                            additionalProperties: false
                        },
                        ArrayExpression: ELEMENT_LIST_SCHEMA,
                        ObjectExpression: ELEMENT_LIST_SCHEMA,
                        ImportDeclaration: ELEMENT_LIST_SCHEMA,
                        flatTernaryExpressions: {
                            type: "boolean",
                            default: false
                        },
                        offsetTernaryExpressions: {
                            type: "boolean",
                            default: false
                        },
                        ignoredNodes: {
                            type: "array",
                            items: {
                                type: "string",
                                not: {
                                    pattern: ":exit$"
                                }
                            }
                        },
                        ignoreComments: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                wrongIndentation: "Expected indentation of {{expected}} but found {{actual}}."
            }
        },
        create(context) {
            const DEFAULT_VARIABLE_INDENT = 1;
            const DEFAULT_PARAMETER_INDENT = 1;
            const DEFAULT_FUNCTION_BODY_INDENT = 1;
            let indentType = "space";
            let indentSize = 4;
            const options = {
                SwitchCase: 0,
                VariableDeclarator: {
                    var: DEFAULT_VARIABLE_INDENT,
                    let: DEFAULT_VARIABLE_INDENT,
                    const: DEFAULT_VARIABLE_INDENT
                },
                outerIIFEBody: 1,
                FunctionDeclaration: {
                    parameters: DEFAULT_PARAMETER_INDENT,
                    body: DEFAULT_FUNCTION_BODY_INDENT
                },
                FunctionExpression: {
                    parameters: DEFAULT_PARAMETER_INDENT,
                    body: DEFAULT_FUNCTION_BODY_INDENT
                },
                CallExpression: {
                    arguments: DEFAULT_PARAMETER_INDENT
                },
                MemberExpression: 1,
                ArrayExpression: 1,
                ObjectExpression: 1,
                ImportDeclaration: 1,
                flatTernaryExpressions: false,
                ignoredNodes: [],
                ignoreComments: false
            };
            if (context.options.length) {
                if (context.options[0] === "tab") {
                    indentSize = 1;
                    indentType = "tab";
                }
                else {
                    indentSize = context.options[0];
                    indentType = "space";
                }
                if (context.options[1]) {
                    Object.assign(options, context.options[1]);
                    if (typeof options.VariableDeclarator === "number" || options.VariableDeclarator === "first") {
                        options.VariableDeclarator = {
                            var: options.VariableDeclarator,
                            let: options.VariableDeclarator,
                            const: options.VariableDeclarator
                        };
                    }
                }
            }
            const sourceCode = context.getSourceCode();
            const tokenInfo = new TokenInfo(sourceCode);
            const offsets = new OffsetStorage(tokenInfo, indentSize, indentType === "space" ? " " : "\t");
            const parameterParens = new WeakSet();
            /**
             * Creates an error message for a line, given the expected/actual indentation.
             * @param {int} expectedAmount The expected amount of indentation characters for this line
             * @param {int} actualSpaces The actual number of indentation spaces that were found on this line
             * @param {int} actualTabs The actual number of indentation tabs that were found on this line
             * @returns {string} An error message for this line
             */
            function createErrorMessageData(expectedAmount, actualSpaces, actualTabs) {
                const expectedStatement = `${expectedAmount} ${indentType}${expectedAmount === 1 ? "" : "s"}`; // e.g. "2 tabs"
                const foundSpacesWord = `space${actualSpaces === 1 ? "" : "s"}`; // e.g. "space"
                const foundTabsWord = `tab${actualTabs === 1 ? "" : "s"}`; // e.g. "tabs"
                let foundStatement;
                if (actualSpaces > 0) {
                    /*
                     * Abbreviate the message if the expected indentation is also spaces.
                     * e.g. 'Expected 4 spaces but found 2' rather than 'Expected 4 spaces but found 2 spaces'
                     */
                    foundStatement = indentType === "space" ? actualSpaces : `${actualSpaces} ${foundSpacesWord}`;
                }
                else if (actualTabs > 0) {
                    foundStatement = indentType === "tab" ? actualTabs : `${actualTabs} ${foundTabsWord}`;
                }
                else {
                    foundStatement = "0";
                }
                return {
                    expected: expectedStatement,
                    actual: foundStatement
                };
            }
            /**
             * Reports a given indent violation
             * @param {Token} token Token violating the indent rule
             * @param {string} neededIndent Expected indentation string
             * @returns {void}
             */
            function report(token, neededIndent) {
                const actualIndent = Array.from(tokenInfo.getTokenIndent(token));
                const numSpaces = actualIndent.filter(char => char === " ").length;
                const numTabs = actualIndent.filter(char => char === "\t").length;
                context.report({
                    node: token,
                    messageId: "wrongIndentation",
                    data: createErrorMessageData(neededIndent.length, numSpaces, numTabs),
                    loc: {
                        start: { line: token.loc.start.line, column: 0 },
                        end: { line: token.loc.start.line, column: token.loc.start.column }
                    },
                    fix(fixer) {
                        const range = [token.range[0] - token.loc.start.column, token.range[0]];
                        const newText = neededIndent;
                        return fixer.replaceTextRange(range, newText);
                    }
                });
            }
            /**
             * Checks if a token's indentation is correct
             * @param {Token} token Token to examine
             * @param {string} desiredIndent Desired indentation of the string
             * @returns {boolean} `true` if the token's indentation is correct
             */
            function validateTokenIndent(token, desiredIndent) {
                const indentation = tokenInfo.getTokenIndent(token);
                return indentation === desiredIndent ||
                    // To avoid conflicts with no-mixed-spaces-and-tabs, don't report mixed spaces and tabs.
                    indentation.includes(" ") && indentation.includes("\t");
            }
            /**
             * Check to see if the node is a file level IIFE
             * @param {ASTNode} node The function node to check.
             * @returns {boolean} True if the node is the outer IIFE
             */
            function isOuterIIFE(node) {
                /*
                 * Verify that the node is an IIFE
                 */
                if (!node.parent || node.parent.type !== "CallExpression" || node.parent.callee !== node) {
                    return false;
                }
                /*
                 * Navigate legal ancestors to determine whether this IIFE is outer.
                 * A "legal ancestor" is an expression or statement that causes the function to get executed immediately.
                 * For example, `!(function(){})()` is an outer IIFE even though it is preceded by a ! operator.
                 */
                let statement = node.parent && node.parent.parent;
                while (statement.type === "UnaryExpression" && ["!", "~", "+", "-"].indexOf(statement.operator) > -1 ||
                    statement.type === "AssignmentExpression" ||
                    statement.type === "LogicalExpression" ||
                    statement.type === "SequenceExpression" ||
                    statement.type === "VariableDeclarator") {
                    statement = statement.parent;
                }
                return (statement.type === "ExpressionStatement" || statement.type === "VariableDeclaration") && statement.parent.type === "Program";
            }
            /**
             * Counts the number of linebreaks that follow the last non-whitespace character in a string
             * @param {string} string The string to check
             * @returns {number} The number of JavaScript linebreaks that follow the last non-whitespace character,
             * or the total number of linebreaks if the string is all whitespace.
             */
            function countTrailingLinebreaks(string) {
                const trailingWhitespace = string.match(/\s*$/u)[0];
                const linebreakMatches = trailingWhitespace.match(astUtils.createGlobalLinebreakMatcher());
                return linebreakMatches === null ? 0 : linebreakMatches.length;
            }
            /**
             * Check indentation for lists of elements (arrays, objects, function params)
             * @param {ASTNode[]} elements List of elements that should be offset
             * @param {Token} startToken The start token of the list that element should be aligned against, e.g. '['
             * @param {Token} endToken The end token of the list, e.g. ']'
             * @param {number|string} offset The amount that the elements should be offset
             * @returns {void}
             */
            function addElementListIndent(elements, startToken, endToken, offset) {
                /**
                 * Gets the first token of a given element, including surrounding parentheses.
                 * @param {ASTNode} element A node in the `elements` list
                 * @returns {Token} The first token of this element
                 */
                function getFirstToken(element) {
                    let token = sourceCode.getTokenBefore(element);
                    while (astUtils.isOpeningParenToken(token) && token !== startToken) {
                        token = sourceCode.getTokenBefore(token);
                    }
                    return sourceCode.getTokenAfter(token);
                }
                // Run through all the tokens in the list, and offset them by one indent level (mainly for comments, other things will end up overridden)
                offsets.setDesiredOffsets([startToken.range[1], endToken.range[0]], startToken, typeof offset === "number" ? offset : 1);
                offsets.setDesiredOffset(endToken, startToken, 0);
                // If the preference is "first" but there is no first element (e.g. sparse arrays w/ empty first slot), fall back to 1 level.
                if (offset === "first" && elements.length && !elements[0]) {
                    return;
                }
                elements.forEach((element, index) => {
                    if (!element) {
                        // Skip holes in arrays
                        return;
                    }
                    if (offset === "off") {
                        // Ignore the first token of every element if the "off" option is used
                        offsets.ignoreToken(getFirstToken(element));
                    }
                    // Offset the following elements correctly relative to the first element
                    if (index === 0) {
                        return;
                    }
                    if (offset === "first" && tokenInfo.isFirstTokenOfLine(getFirstToken(element))) {
                        offsets.matchOffsetOf(getFirstToken(elements[0]), getFirstToken(element));
                    }
                    else {
                        const previousElement = elements[index - 1];
                        const firstTokenOfPreviousElement = previousElement && getFirstToken(previousElement);
                        const previousElementLastToken = previousElement && sourceCode.getLastToken(previousElement);
                        if (previousElement &&
                            previousElementLastToken.loc.end.line - countTrailingLinebreaks(previousElementLastToken.value) > startToken.loc.end.line) {
                            offsets.setDesiredOffsets([previousElement.range[1], element.range[1]], firstTokenOfPreviousElement, 0);
                        }
                    }
                });
            }
            /**
             * Check and decide whether to check for indentation for blockless nodes
             * Scenarios are for or while statements without braces around them
             * @param {ASTNode} node node to examine
             * @returns {void}
             */
            function addBlocklessNodeIndent(node) {
                if (node.type !== "BlockStatement") {
                    const lastParentToken = sourceCode.getTokenBefore(node, astUtils.isNotOpeningParenToken);
                    let firstBodyToken = sourceCode.getFirstToken(node);
                    let lastBodyToken = sourceCode.getLastToken(node);
                    while (astUtils.isOpeningParenToken(sourceCode.getTokenBefore(firstBodyToken)) &&
                        astUtils.isClosingParenToken(sourceCode.getTokenAfter(lastBodyToken))) {
                        firstBodyToken = sourceCode.getTokenBefore(firstBodyToken);
                        lastBodyToken = sourceCode.getTokenAfter(lastBodyToken);
                    }
                    offsets.setDesiredOffsets([firstBodyToken.range[0], lastBodyToken.range[1]], lastParentToken, 1);
                    /*
                     * For blockless nodes with semicolon-first style, don't indent the semicolon.
                     * e.g.
                     * if (foo) bar()
                     * ; [1, 2, 3].map(foo)
                     */
                    const lastToken = sourceCode.getLastToken(node);
                    if (node.type !== "EmptyStatement" && astUtils.isSemicolonToken(lastToken)) {
                        offsets.setDesiredOffset(lastToken, lastParentToken, 0);
                    }
                }
            }
            /**
             * Checks the indentation for nodes that are like function calls (`CallExpression` and `NewExpression`)
             * @param {ASTNode} node A CallExpression or NewExpression node
             * @returns {void}
             */
            function addFunctionCallIndent(node) {
                let openingParen;
                if (node.arguments.length) {
                    openingParen = sourceCode.getFirstTokenBetween(node.callee, node.arguments[0], astUtils.isOpeningParenToken);
                }
                else {
                    openingParen = sourceCode.getLastToken(node, 1);
                }
                const closingParen = sourceCode.getLastToken(node);
                parameterParens.add(openingParen);
                parameterParens.add(closingParen);
                /*
                 * If `?.` token exists, set desired offset for that.
                 * This logic is copied from `MemberExpression`'s.
                 */
                if (node.optional) {
                    const dotToken = sourceCode.getTokenAfter(node.callee, astUtils.isQuestionDotToken);
                    const calleeParenCount = sourceCode.getTokensBetween(node.callee, dotToken, { filter: astUtils.isClosingParenToken }).length;
                    const firstTokenOfCallee = calleeParenCount
                        ? sourceCode.getTokenBefore(node.callee, { skip: calleeParenCount - 1 })
                        : sourceCode.getFirstToken(node.callee);
                    const lastTokenOfCallee = sourceCode.getTokenBefore(dotToken);
                    const offsetBase = lastTokenOfCallee.loc.end.line === openingParen.loc.start.line
                        ? lastTokenOfCallee
                        : firstTokenOfCallee;
                    offsets.setDesiredOffset(dotToken, offsetBase, 1);
                }
                const offsetAfterToken = node.callee.type === "TaggedTemplateExpression" ? sourceCode.getFirstToken(node.callee.quasi) : openingParen;
                const offsetToken = sourceCode.getTokenBefore(offsetAfterToken);
                offsets.setDesiredOffset(openingParen, offsetToken, 0);
                addElementListIndent(node.arguments, openingParen, closingParen, options.CallExpression.arguments);
            }
            /**
             * Checks the indentation of parenthesized values, given a list of tokens in a program
             * @param {Token[]} tokens A list of tokens
             * @returns {void}
             */
            function addParensIndent(tokens) {
                const parenStack = [];
                const parenPairs = [];
                tokens.forEach(nextToken => {
                    // Accumulate a list of parenthesis pairs
                    if (astUtils.isOpeningParenToken(nextToken)) {
                        parenStack.push(nextToken);
                    }
                    else if (astUtils.isClosingParenToken(nextToken)) {
                        parenPairs.unshift({ left: parenStack.pop(), right: nextToken });
                    }
                });
                parenPairs.forEach(pair => {
                    const leftParen = pair.left;
                    const rightParen = pair.right;
                    // We only want to handle parens around expressions, so exclude parentheses that are in function parameters and function call arguments.
                    if (!parameterParens.has(leftParen) && !parameterParens.has(rightParen)) {
                        const parenthesizedTokens = new Set(sourceCode.getTokensBetween(leftParen, rightParen));
                        parenthesizedTokens.forEach(token => {
                            if (!parenthesizedTokens.has(offsets.getFirstDependency(token))) {
                                offsets.setDesiredOffset(token, leftParen, 1);
                            }
                        });
                    }
                    offsets.setDesiredOffset(rightParen, leftParen, 0);
                });
            }
            /**
             * Ignore all tokens within an unknown node whose offset do not depend
             * on another token's offset within the unknown node
             * @param {ASTNode} node Unknown Node
             * @returns {void}
             */
            function ignoreNode(node) {
                const unknownNodeTokens = new Set(sourceCode.getTokens(node, { includeComments: true }));
                unknownNodeTokens.forEach(token => {
                    if (!unknownNodeTokens.has(offsets.getFirstDependency(token))) {
                        const firstTokenOfLine = tokenInfo.getFirstTokenOfLine(token);
                        if (token === firstTokenOfLine) {
                            offsets.ignoreToken(token);
                        }
                        else {
                            offsets.setDesiredOffset(token, firstTokenOfLine, 0);
                        }
                    }
                });
            }
            /**
             * Check whether the given token is on the first line of a statement.
             * @param {Token} token The token to check.
             * @param {ASTNode} leafNode The expression node that the token belongs directly.
             * @returns {boolean} `true` if the token is on the first line of a statement.
             */
            function isOnFirstLineOfStatement(token, leafNode) {
                let node = leafNode;
                while (node.parent && !node.parent.type.endsWith("Statement") && !node.parent.type.endsWith("Declaration")) {
                    node = node.parent;
                }
                node = node.parent;
                return !node || node.loc.start.line === token.loc.start.line;
            }
            /**
             * Check whether there are any blank (whitespace-only) lines between
             * two tokens on separate lines.
             * @param {Token} firstToken The first token.
             * @param {Token} secondToken The second token.
             * @returns {boolean} `true` if the tokens are on separate lines and
             *   there exists a blank line between them, `false` otherwise.
             */
            function hasBlankLinesBetween(firstToken, secondToken) {
                const firstTokenLine = firstToken.loc.end.line;
                const secondTokenLine = secondToken.loc.start.line;
                if (firstTokenLine === secondTokenLine || firstTokenLine === secondTokenLine - 1) {
                    return false;
                }
                for (let line = firstTokenLine + 1; line < secondTokenLine; ++line) {
                    if (!tokenInfo.firstTokensByLineNumber.has(line)) {
                        return true;
                    }
                }
                return false;
            }
            const ignoredNodeFirstTokens = new Set();
            const baseOffsetListeners = {
                "ArrayExpression, ArrayPattern"(node) {
                    const openingBracket = sourceCode.getFirstToken(node);
                    const closingBracket = sourceCode.getTokenAfter([...node.elements].reverse().find(_ => _) || openingBracket, astUtils.isClosingBracketToken);
                    addElementListIndent(node.elements, openingBracket, closingBracket, options.ArrayExpression);
                },
                "ObjectExpression, ObjectPattern"(node) {
                    const openingCurly = sourceCode.getFirstToken(node);
                    const closingCurly = sourceCode.getTokenAfter(node.properties.length ? node.properties[node.properties.length - 1] : openingCurly, astUtils.isClosingBraceToken);
                    addElementListIndent(node.properties, openingCurly, closingCurly, options.ObjectExpression);
                },
                ArrowFunctionExpression(node) {
                    const maybeOpeningParen = sourceCode.getFirstToken(node, { skip: node.async ? 1 : 0 });
                    if (astUtils.isOpeningParenToken(maybeOpeningParen)) {
                        const openingParen = maybeOpeningParen;
                        const closingParen = sourceCode.getTokenBefore(node.body, astUtils.isClosingParenToken);
                        parameterParens.add(openingParen);
                        parameterParens.add(closingParen);
                        addElementListIndent(node.params, openingParen, closingParen, options.FunctionExpression.parameters);
                    }
                    addBlocklessNodeIndent(node.body);
                },
                AssignmentExpression(node) {
                    const operator = sourceCode.getFirstTokenBetween(node.left, node.right, token => token.value === node.operator);
                    offsets.setDesiredOffsets([operator.range[0], node.range[1]], sourceCode.getLastToken(node.left), 1);
                    offsets.ignoreToken(operator);
                    offsets.ignoreToken(sourceCode.getTokenAfter(operator));
                },
                "BinaryExpression, LogicalExpression"(node) {
                    const operator = sourceCode.getFirstTokenBetween(node.left, node.right, token => token.value === node.operator);
                    /*
                     * For backwards compatibility, don't check BinaryExpression indents, e.g.
                     * var foo = bar &&
                     *                   baz;
                     */
                    const tokenAfterOperator = sourceCode.getTokenAfter(operator);
                    offsets.ignoreToken(operator);
                    offsets.ignoreToken(tokenAfterOperator);
                    offsets.setDesiredOffset(tokenAfterOperator, operator, 0);
                },
                "BlockStatement, ClassBody"(node) {
                    let blockIndentLevel;
                    if (node.parent && isOuterIIFE(node.parent)) {
                        blockIndentLevel = options.outerIIFEBody;
                    }
                    else if (node.parent && (node.parent.type === "FunctionExpression" || node.parent.type === "ArrowFunctionExpression")) {
                        blockIndentLevel = options.FunctionExpression.body;
                    }
                    else if (node.parent && node.parent.type === "FunctionDeclaration") {
                        blockIndentLevel = options.FunctionDeclaration.body;
                    }
                    else {
                        blockIndentLevel = 1;
                    }
                    /*
                     * For blocks that aren't lone statements, ensure that the opening curly brace
                     * is aligned with the parent.
                     */
                    if (!astUtils.STATEMENT_LIST_PARENTS.has(node.parent.type)) {
                        offsets.setDesiredOffset(sourceCode.getFirstToken(node), sourceCode.getFirstToken(node.parent), 0);
                    }
                    addElementListIndent(node.body, sourceCode.getFirstToken(node), sourceCode.getLastToken(node), blockIndentLevel);
                },
                CallExpression: addFunctionCallIndent,
                "ClassDeclaration[superClass], ClassExpression[superClass]"(node) {
                    const classToken = sourceCode.getFirstToken(node);
                    const extendsToken = sourceCode.getTokenBefore(node.superClass, astUtils.isNotOpeningParenToken);
                    offsets.setDesiredOffsets([extendsToken.range[0], node.body.range[0]], classToken, 1);
                },
                ConditionalExpression(node) {
                    const firstToken = sourceCode.getFirstToken(node);
                    // `flatTernaryExpressions` option is for the following style:
                    // var a =
                    //     foo > 0 ? bar :
                    //     foo < 0 ? baz :
                    //     /*else*/ qiz ;
                    if (!options.flatTernaryExpressions ||
                        !astUtils.isTokenOnSameLine(node.test, node.consequent) ||
                        isOnFirstLineOfStatement(firstToken, node)) {
                        const questionMarkToken = sourceCode.getFirstTokenBetween(node.test, node.consequent, token => token.type === "Punctuator" && token.value === "?");
                        const colonToken = sourceCode.getFirstTokenBetween(node.consequent, node.alternate, token => token.type === "Punctuator" && token.value === ":");
                        const firstConsequentToken = sourceCode.getTokenAfter(questionMarkToken);
                        const lastConsequentToken = sourceCode.getTokenBefore(colonToken);
                        const firstAlternateToken = sourceCode.getTokenAfter(colonToken);
                        offsets.setDesiredOffset(questionMarkToken, firstToken, 1);
                        offsets.setDesiredOffset(colonToken, firstToken, 1);
                        offsets.setDesiredOffset(firstConsequentToken, firstToken, firstConsequentToken.type === "Punctuator" &&
                            options.offsetTernaryExpressions ? 2 : 1);
                        /*
                         * The alternate and the consequent should usually have the same indentation.
                         * If they share part of a line, align the alternate against the first token of the consequent.
                         * This allows the alternate to be indented correctly in cases like this:
                         * foo ? (
                         *   bar
                         * ) : ( // this '(' is aligned with the '(' above, so it's considered to be aligned with `foo`
                         *   baz // as a result, `baz` is offset by 1 rather than 2
                         * )
                         */
                        if (lastConsequentToken.loc.end.line === firstAlternateToken.loc.start.line) {
                            offsets.setDesiredOffset(firstAlternateToken, firstConsequentToken, 0);
                        }
                        else {
                            /**
                             * If the alternate and consequent do not share part of a line, offset the alternate from the first
                             * token of the conditional expression. For example:
                             * foo ? bar
                             *   : baz
                             *
                             * If `baz` were aligned with `bar` rather than being offset by 1 from `foo`, `baz` would end up
                             * having no expected indentation.
                             */
                            offsets.setDesiredOffset(firstAlternateToken, firstToken, firstAlternateToken.type === "Punctuator" &&
                                options.offsetTernaryExpressions ? 2 : 1);
                        }
                    }
                },
                "DoWhileStatement, WhileStatement, ForInStatement, ForOfStatement": node => addBlocklessNodeIndent(node.body),
                ExportNamedDeclaration(node) {
                    if (node.declaration === null) {
                        const closingCurly = sourceCode.getLastToken(node, astUtils.isClosingBraceToken);
                        // Indent the specifiers in `export {foo, bar, baz}`
                        addElementListIndent(node.specifiers, sourceCode.getFirstToken(node, { skip: 1 }), closingCurly, 1);
                        if (node.source) {
                            // Indent everything after and including the `from` token in `export {foo, bar, baz} from 'qux'`
                            offsets.setDesiredOffsets([closingCurly.range[1], node.range[1]], sourceCode.getFirstToken(node), 1);
                        }
                    }
                },
                ForStatement(node) {
                    const forOpeningParen = sourceCode.getFirstToken(node, 1);
                    if (node.init) {
                        offsets.setDesiredOffsets(node.init.range, forOpeningParen, 1);
                    }
                    if (node.test) {
                        offsets.setDesiredOffsets(node.test.range, forOpeningParen, 1);
                    }
                    if (node.update) {
                        offsets.setDesiredOffsets(node.update.range, forOpeningParen, 1);
                    }
                    addBlocklessNodeIndent(node.body);
                },
                "FunctionDeclaration, FunctionExpression"(node) {
                    const closingParen = sourceCode.getTokenBefore(node.body);
                    const openingParen = sourceCode.getTokenBefore(node.params.length ? node.params[0] : closingParen);
                    parameterParens.add(openingParen);
                    parameterParens.add(closingParen);
                    addElementListIndent(node.params, openingParen, closingParen, options[node.type].parameters);
                },
                IfStatement(node) {
                    addBlocklessNodeIndent(node.consequent);
                    if (node.alternate && node.alternate.type !== "IfStatement") {
                        addBlocklessNodeIndent(node.alternate);
                    }
                },
                ImportDeclaration(node) {
                    if (node.specifiers.some(specifier => specifier.type === "ImportSpecifier")) {
                        const openingCurly = sourceCode.getFirstToken(node, astUtils.isOpeningBraceToken);
                        const closingCurly = sourceCode.getLastToken(node, astUtils.isClosingBraceToken);
                        addElementListIndent(node.specifiers.filter(specifier => specifier.type === "ImportSpecifier"), openingCurly, closingCurly, options.ImportDeclaration);
                    }
                    const fromToken = sourceCode.getLastToken(node, token => token.type === "Identifier" && token.value === "from");
                    const sourceToken = sourceCode.getLastToken(node, token => token.type === "String");
                    const semiToken = sourceCode.getLastToken(node, token => token.type === "Punctuator" && token.value === ";");
                    if (fromToken) {
                        const end = semiToken && semiToken.range[1] === sourceToken.range[1] ? node.range[1] : sourceToken.range[1];
                        offsets.setDesiredOffsets([fromToken.range[0], end], sourceCode.getFirstToken(node), 1);
                    }
                },
                ImportExpression(node) {
                    const openingParen = sourceCode.getFirstToken(node, 1);
                    const closingParen = sourceCode.getLastToken(node);
                    parameterParens.add(openingParen);
                    parameterParens.add(closingParen);
                    offsets.setDesiredOffset(openingParen, sourceCode.getTokenBefore(openingParen), 0);
                    addElementListIndent([node.source], openingParen, closingParen, options.CallExpression.arguments);
                },
                "MemberExpression, JSXMemberExpression, MetaProperty"(node) {
                    const object = node.type === "MetaProperty" ? node.meta : node.object;
                    const firstNonObjectToken = sourceCode.getFirstTokenBetween(object, node.property, astUtils.isNotClosingParenToken);
                    const secondNonObjectToken = sourceCode.getTokenAfter(firstNonObjectToken);
                    const objectParenCount = sourceCode.getTokensBetween(object, node.property, { filter: astUtils.isClosingParenToken }).length;
                    const firstObjectToken = objectParenCount
                        ? sourceCode.getTokenBefore(object, { skip: objectParenCount - 1 })
                        : sourceCode.getFirstToken(object);
                    const lastObjectToken = sourceCode.getTokenBefore(firstNonObjectToken);
                    const firstPropertyToken = node.computed ? firstNonObjectToken : secondNonObjectToken;
                    if (node.computed) {
                        // For computed MemberExpressions, match the closing bracket with the opening bracket.
                        offsets.setDesiredOffset(sourceCode.getLastToken(node), firstNonObjectToken, 0);
                        offsets.setDesiredOffsets(node.property.range, firstNonObjectToken, 1);
                    }
                    /*
                     * If the object ends on the same line that the property starts, match against the last token
                     * of the object, to ensure that the MemberExpression is not indented.
                     *
                     * Otherwise, match against the first token of the object, e.g.
                     * foo
                     *   .bar
                     *   .baz // <-- offset by 1 from `foo`
                     */
                    const offsetBase = lastObjectToken.loc.end.line === firstPropertyToken.loc.start.line
                        ? lastObjectToken
                        : firstObjectToken;
                    if (typeof options.MemberExpression === "number") {
                        // Match the dot (for non-computed properties) or the opening bracket (for computed properties) against the object.
                        offsets.setDesiredOffset(firstNonObjectToken, offsetBase, options.MemberExpression);
                        /*
                         * For computed MemberExpressions, match the first token of the property against the opening bracket.
                         * Otherwise, match the first token of the property against the object.
                         */
                        offsets.setDesiredOffset(secondNonObjectToken, node.computed ? firstNonObjectToken : offsetBase, options.MemberExpression);
                    }
                    else {
                        // If the MemberExpression option is off, ignore the dot and the first token of the property.
                        offsets.ignoreToken(firstNonObjectToken);
                        offsets.ignoreToken(secondNonObjectToken);
                        // To ignore the property indentation, ensure that the property tokens depend on the ignored tokens.
                        offsets.setDesiredOffset(firstNonObjectToken, offsetBase, 0);
                        offsets.setDesiredOffset(secondNonObjectToken, firstNonObjectToken, 0);
                    }
                },
                NewExpression(node) {
                    // Only indent the arguments if the NewExpression has parens (e.g. `new Foo(bar)` or `new Foo()`, but not `new Foo`
                    if (node.arguments.length > 0 ||
                        astUtils.isClosingParenToken(sourceCode.getLastToken(node)) &&
                            astUtils.isOpeningParenToken(sourceCode.getLastToken(node, 1))) {
                        addFunctionCallIndent(node);
                    }
                },
                Property(node) {
                    if (!node.shorthand && !node.method && node.kind === "init") {
                        const colon = sourceCode.getFirstTokenBetween(node.key, node.value, astUtils.isColonToken);
                        offsets.ignoreToken(sourceCode.getTokenAfter(colon));
                    }
                },
                SwitchStatement(node) {
                    const openingCurly = sourceCode.getTokenAfter(node.discriminant, astUtils.isOpeningBraceToken);
                    const closingCurly = sourceCode.getLastToken(node);
                    offsets.setDesiredOffsets([openingCurly.range[1], closingCurly.range[0]], openingCurly, options.SwitchCase);
                    if (node.cases.length) {
                        sourceCode.getTokensBetween(node.cases[node.cases.length - 1], closingCurly, { includeComments: true, filter: astUtils.isCommentToken }).forEach(token => offsets.ignoreToken(token));
                    }
                },
                SwitchCase(node) {
                    if (!(node.consequent.length === 1 && node.consequent[0].type === "BlockStatement")) {
                        const caseKeyword = sourceCode.getFirstToken(node);
                        const tokenAfterCurrentCase = sourceCode.getTokenAfter(node);
                        offsets.setDesiredOffsets([caseKeyword.range[1], tokenAfterCurrentCase.range[0]], caseKeyword, 1);
                    }
                },
                TemplateLiteral(node) {
                    node.expressions.forEach((expression, index) => {
                        const previousQuasi = node.quasis[index];
                        const nextQuasi = node.quasis[index + 1];
                        const tokenToAlignFrom = previousQuasi.loc.start.line === previousQuasi.loc.end.line
                            ? sourceCode.getFirstToken(previousQuasi)
                            : null;
                        offsets.setDesiredOffsets([previousQuasi.range[1], nextQuasi.range[0]], tokenToAlignFrom, 1);
                        offsets.setDesiredOffset(sourceCode.getFirstToken(nextQuasi), tokenToAlignFrom, 0);
                    });
                },
                VariableDeclaration(node) {
                    let variableIndent = Object.prototype.hasOwnProperty.call(options.VariableDeclarator, node.kind)
                        ? options.VariableDeclarator[node.kind]
                        : DEFAULT_VARIABLE_INDENT;
                    const firstToken = sourceCode.getFirstToken(node), lastToken = sourceCode.getLastToken(node);
                    if (options.VariableDeclarator[node.kind] === "first") {
                        if (node.declarations.length > 1) {
                            addElementListIndent(node.declarations, firstToken, lastToken, "first");
                            return;
                        }
                        variableIndent = DEFAULT_VARIABLE_INDENT;
                    }
                    if (node.declarations[node.declarations.length - 1].loc.start.line > node.loc.start.line) {
                        /*
                         * VariableDeclarator indentation is a bit different from other forms of indentation, in that the
                         * indentation of an opening bracket sometimes won't match that of a closing bracket. For example,
                         * the following indentations are correct:
                         *
                         * var foo = {
                         *   ok: true
                         * };
                         *
                         * var foo = {
                         *     ok: true,
                         *   },
                         *   bar = 1;
                         *
                         * Account for when exiting the AST (after indentations have already been set for the nodes in
                         * the declaration) by manually increasing the indentation level of the tokens in this declarator
                         * on the same line as the start of the declaration, provided that there are declarators that
                         * follow this one.
                         */
                        offsets.setDesiredOffsets(node.range, firstToken, variableIndent, true);
                    }
                    else {
                        offsets.setDesiredOffsets(node.range, firstToken, variableIndent);
                    }
                    if (astUtils.isSemicolonToken(lastToken)) {
                        offsets.ignoreToken(lastToken);
                    }
                },
                VariableDeclarator(node) {
                    if (node.init) {
                        const equalOperator = sourceCode.getTokenBefore(node.init, astUtils.isNotOpeningParenToken);
                        const tokenAfterOperator = sourceCode.getTokenAfter(equalOperator);
                        offsets.ignoreToken(equalOperator);
                        offsets.ignoreToken(tokenAfterOperator);
                        offsets.setDesiredOffsets([tokenAfterOperator.range[0], node.range[1]], equalOperator, 1);
                        offsets.setDesiredOffset(equalOperator, sourceCode.getLastToken(node.id), 0);
                    }
                },
                "JSXAttribute[value]"(node) {
                    const equalsToken = sourceCode.getFirstTokenBetween(node.name, node.value, token => token.type === "Punctuator" && token.value === "=");
                    offsets.setDesiredOffsets([equalsToken.range[0], node.value.range[1]], sourceCode.getFirstToken(node.name), 1);
                },
                JSXElement(node) {
                    if (node.closingElement) {
                        addElementListIndent(node.children, sourceCode.getFirstToken(node.openingElement), sourceCode.getFirstToken(node.closingElement), 1);
                    }
                },
                JSXOpeningElement(node) {
                    const firstToken = sourceCode.getFirstToken(node);
                    let closingToken;
                    if (node.selfClosing) {
                        closingToken = sourceCode.getLastToken(node, { skip: 1 });
                        offsets.setDesiredOffset(sourceCode.getLastToken(node), closingToken, 0);
                    }
                    else {
                        closingToken = sourceCode.getLastToken(node);
                    }
                    offsets.setDesiredOffsets(node.name.range, sourceCode.getFirstToken(node));
                    addElementListIndent(node.attributes, firstToken, closingToken, 1);
                },
                JSXClosingElement(node) {
                    const firstToken = sourceCode.getFirstToken(node);
                    offsets.setDesiredOffsets(node.name.range, firstToken, 1);
                },
                JSXFragment(node) {
                    const firstOpeningToken = sourceCode.getFirstToken(node.openingFragment);
                    const firstClosingToken = sourceCode.getFirstToken(node.closingFragment);
                    addElementListIndent(node.children, firstOpeningToken, firstClosingToken, 1);
                },
                JSXOpeningFragment(node) {
                    const firstToken = sourceCode.getFirstToken(node);
                    const closingToken = sourceCode.getLastToken(node);
                    offsets.setDesiredOffsets(node.range, firstToken, 1);
                    offsets.matchOffsetOf(firstToken, closingToken);
                },
                JSXClosingFragment(node) {
                    const firstToken = sourceCode.getFirstToken(node);
                    const slashToken = sourceCode.getLastToken(node, { skip: 1 });
                    const closingToken = sourceCode.getLastToken(node);
                    const tokenToMatch = astUtils.isTokenOnSameLine(slashToken, closingToken) ? slashToken : closingToken;
                    offsets.setDesiredOffsets(node.range, firstToken, 1);
                    offsets.matchOffsetOf(firstToken, tokenToMatch);
                },
                JSXExpressionContainer(node) {
                    const openingCurly = sourceCode.getFirstToken(node);
                    const closingCurly = sourceCode.getLastToken(node);
                    offsets.setDesiredOffsets([openingCurly.range[1], closingCurly.range[0]], openingCurly, 1);
                },
                JSXSpreadAttribute(node) {
                    const openingCurly = sourceCode.getFirstToken(node);
                    const closingCurly = sourceCode.getLastToken(node);
                    offsets.setDesiredOffsets([openingCurly.range[1], closingCurly.range[0]], openingCurly, 1);
                },
                "*"(node) {
                    const firstToken = sourceCode.getFirstToken(node);
                    // Ensure that the children of every node are indented at least as much as the first token.
                    if (firstToken && !ignoredNodeFirstTokens.has(firstToken)) {
                        offsets.setDesiredOffsets(node.range, firstToken, 0);
                    }
                }
            };
            const listenerCallQueue = [];
            /*
             * To ignore the indentation of a node:
             * 1. Don't call the node's listener when entering it (if it has a listener)
             * 2. Don't set any offsets against the first token of the node.
             * 3. Call `ignoreNode` on the node sometime after exiting it and before validating offsets.
             */
            const offsetListeners = {};
            for (const [selector, listener] of Object.entries(baseOffsetListeners)) {
                /*
                 * Offset listener calls are deferred until traversal is finished, and are called as
                 * part of the final `Program:exit` listener. This is necessary because a node might
                 * be matched by multiple selectors.
                 *
                 * Example: Suppose there is an offset listener for `Identifier`, and the user has
                 * specified in configuration that `MemberExpression > Identifier` should be ignored.
                 * Due to selector specificity rules, the `Identifier` listener will get called first. However,
                 * if a given Identifier node is supposed to be ignored, then the `Identifier` offset listener
                 * should not have been called at all. Without doing extra selector matching, we don't know
                 * whether the Identifier matches the `MemberExpression > Identifier` selector until the
                 * `MemberExpression > Identifier` listener is called.
                 *
                 * To avoid this, the `Identifier` listener isn't called until traversal finishes and all
                 * ignored nodes are known.
                 */
                offsetListeners[selector] = node => listenerCallQueue.push({ listener, node });
            }
            // For each ignored node selector, set up a listener to collect it into the `ignoredNodes` set.
            const ignoredNodes = new Set();
            /**
             * Ignores a node
             * @param {ASTNode} node The node to ignore
             * @returns {void}
             */
            function addToIgnoredNodes(node) {
                ignoredNodes.add(node);
                ignoredNodeFirstTokens.add(sourceCode.getFirstToken(node));
            }
            const ignoredNodeListeners = options.ignoredNodes.reduce((listeners, ignoredSelector) => Object.assign(listeners, { [ignoredSelector]: addToIgnoredNodes }), {});
            /*
             * Join the listeners, and add a listener to verify that all tokens actually have the correct indentation
             * at the end.
             *
             * Using Object.assign will cause some offset listeners to be overwritten if the same selector also appears
             * in `ignoredNodeListeners`. This isn't a problem because all of the matching nodes will be ignored,
             * so those listeners wouldn't be called anyway.
             */
            return Object.assign(offsetListeners, ignoredNodeListeners, {
                "*:exit"(node) {
                    // If a node's type is nonstandard, we can't tell how its children should be offset, so ignore it.
                    if (!KNOWN_NODES.has(node.type)) {
                        addToIgnoredNodes(node);
                    }
                },
                "Program:exit"() {
                    // If ignoreComments option is enabled, ignore all comment tokens.
                    if (options.ignoreComments) {
                        sourceCode.getAllComments()
                            .forEach(comment => offsets.ignoreToken(comment));
                    }
                    // Invoke the queued offset listeners for the nodes that aren't ignored.
                    listenerCallQueue
                        .filter(nodeInfo => !ignoredNodes.has(nodeInfo.node))
                        .forEach(nodeInfo => nodeInfo.listener(nodeInfo.node));
                    // Update the offsets for ignored nodes to prevent their child tokens from being reported.
                    ignoredNodes.forEach(ignoreNode);
                    addParensIndent(sourceCode.ast.tokens);
                    /*
                     * Create a Map from (tokenOrComment) => (precedingToken).
                     * This is necessary because sourceCode.getTokenBefore does not handle a comment as an argument correctly.
                     */
                    const precedingTokens = sourceCode.ast.comments.reduce((commentMap, comment) => {
                        const tokenOrCommentBefore = sourceCode.getTokenBefore(comment, { includeComments: true });
                        return commentMap.set(comment, commentMap.has(tokenOrCommentBefore) ? commentMap.get(tokenOrCommentBefore) : tokenOrCommentBefore);
                    }, new WeakMap());
                    sourceCode.lines.forEach((line, lineIndex) => {
                        const lineNumber = lineIndex + 1;
                        if (!tokenInfo.firstTokensByLineNumber.has(lineNumber)) {
                            // Don't check indentation on blank lines
                            return;
                        }
                        const firstTokenOfLine = tokenInfo.firstTokensByLineNumber.get(lineNumber);
                        if (firstTokenOfLine.loc.start.line !== lineNumber) {
                            // Don't check the indentation of multi-line tokens (e.g. template literals or block comments) twice.
                            return;
                        }
                        if (astUtils.isCommentToken(firstTokenOfLine)) {
                            const tokenBefore = precedingTokens.get(firstTokenOfLine);
                            const tokenAfter = tokenBefore ? sourceCode.getTokenAfter(tokenBefore) : sourceCode.ast.tokens[0];
                            const mayAlignWithBefore = tokenBefore && !hasBlankLinesBetween(tokenBefore, firstTokenOfLine);
                            const mayAlignWithAfter = tokenAfter && !hasBlankLinesBetween(firstTokenOfLine, tokenAfter);
                            /*
                             * If a comment precedes a line that begins with a semicolon token, align to that token, i.e.
                             *
                             * let foo
                             * // comment
                             * ;(async () => {})()
                             */
                            if (tokenAfter && astUtils.isSemicolonToken(tokenAfter) && !astUtils.isTokenOnSameLine(firstTokenOfLine, tokenAfter)) {
                                offsets.setDesiredOffset(firstTokenOfLine, tokenAfter, 0);
                            }
                            // If a comment matches the expected indentation of the token immediately before or after, don't report it.
                            if (mayAlignWithBefore && validateTokenIndent(firstTokenOfLine, offsets.getDesiredIndent(tokenBefore)) ||
                                mayAlignWithAfter && validateTokenIndent(firstTokenOfLine, offsets.getDesiredIndent(tokenAfter))) {
                                return;
                            }
                        }
                        // If the token matches the expected indentation, don't report it.
                        if (validateTokenIndent(firstTokenOfLine, offsets.getDesiredIndent(firstTokenOfLine))) {
                            return;
                        }
                        // Otherwise, report the token/comment.
                        report(firstTokenOfLine, offsets.getDesiredIndent(firstTokenOfLine));
                    });
                }
            });
        }
    };
},
/* functional-red-black-tree/rbtree.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = createRBTree;
    var RED = 0;
    var BLACK = 1;
    function RBNode(color, key, value, left, right, count) {
        this._color = color;
        this.key = key;
        this.value = value;
        this.left = left;
        this.right = right;
        this._count = count;
    }
    function cloneNode(node) {
        return new RBNode(node._color, node.key, node.value, node.left, node.right, node._count);
    }
    function repaint(color, node) {
        return new RBNode(color, node.key, node.value, node.left, node.right, node._count);
    }
    function recount(node) {
        node._count = 1 + (node.left ? node.left._count : 0) + (node.right ? node.right._count : 0);
    }
    function RedBlackTree(compare, root) {
        this._compare = compare;
        this.root = root;
    }
    var proto = RedBlackTree.prototype;
    Object.defineProperty(proto, "keys", {
        get: function () {
            var result = [];
            this.forEach(function (k, v) {
                result.push(k);
            });
            return result;
        }
    });
    Object.defineProperty(proto, "values", {
        get: function () {
            var result = [];
            this.forEach(function (k, v) {
                result.push(v);
            });
            return result;
        }
    });
    //Returns the number of nodes in the tree
    Object.defineProperty(proto, "length", {
        get: function () {
            if (this.root) {
                return this.root._count;
            }
            return 0;
        }
    });
    //Insert a new item into the tree
    proto.insert = function (key, value) {
        var cmp = this._compare;
        //Find point to insert new node at
        var n = this.root;
        var n_stack = [];
        var d_stack = [];
        while (n) {
            var d = cmp(key, n.key);
            n_stack.push(n);
            d_stack.push(d);
            if (d <= 0) {
                n = n.left;
            }
            else {
                n = n.right;
            }
        }
        //Rebuild path to leaf node
        n_stack.push(new RBNode(RED, key, value, null, null, 1));
        for (var s = n_stack.length - 2; s >= 0; --s) {
            var n = n_stack[s];
            if (d_stack[s] <= 0) {
                n_stack[s] = new RBNode(n._color, n.key, n.value, n_stack[s + 1], n.right, n._count + 1);
            }
            else {
                n_stack[s] = new RBNode(n._color, n.key, n.value, n.left, n_stack[s + 1], n._count + 1);
            }
        }
        //Rebalance tree using rotations
        //console.log("start insert", key, d_stack)
        for (var s = n_stack.length - 1; s > 1; --s) {
            var p = n_stack[s - 1];
            var n = n_stack[s];
            if (p._color === BLACK || n._color === BLACK) {
                break;
            }
            var pp = n_stack[s - 2];
            if (pp.left === p) {
                if (p.left === n) {
                    var y = pp.right;
                    if (y && y._color === RED) {
                        //console.log("LLr")
                        p._color = BLACK;
                        pp.right = repaint(BLACK, y);
                        pp._color = RED;
                        s -= 1;
                    }
                    else {
                        //console.log("LLb")
                        pp._color = RED;
                        pp.left = p.right;
                        p._color = BLACK;
                        p.right = pp;
                        n_stack[s - 2] = p;
                        n_stack[s - 1] = n;
                        recount(pp);
                        recount(p);
                        if (s >= 3) {
                            var ppp = n_stack[s - 3];
                            if (ppp.left === pp) {
                                ppp.left = p;
                            }
                            else {
                                ppp.right = p;
                            }
                        }
                        break;
                    }
                }
                else {
                    var y = pp.right;
                    if (y && y._color === RED) {
                        //console.log("LRr")
                        p._color = BLACK;
                        pp.right = repaint(BLACK, y);
                        pp._color = RED;
                        s -= 1;
                    }
                    else {
                        //console.log("LRb")
                        p.right = n.left;
                        pp._color = RED;
                        pp.left = n.right;
                        n._color = BLACK;
                        n.left = p;
                        n.right = pp;
                        n_stack[s - 2] = n;
                        n_stack[s - 1] = p;
                        recount(pp);
                        recount(p);
                        recount(n);
                        if (s >= 3) {
                            var ppp = n_stack[s - 3];
                            if (ppp.left === pp) {
                                ppp.left = n;
                            }
                            else {
                                ppp.right = n;
                            }
                        }
                        break;
                    }
                }
            }
            else {
                if (p.right === n) {
                    var y = pp.left;
                    if (y && y._color === RED) {
                        //console.log("RRr", y.key)
                        p._color = BLACK;
                        pp.left = repaint(BLACK, y);
                        pp._color = RED;
                        s -= 1;
                    }
                    else {
                        //console.log("RRb")
                        pp._color = RED;
                        pp.right = p.left;
                        p._color = BLACK;
                        p.left = pp;
                        n_stack[s - 2] = p;
                        n_stack[s - 1] = n;
                        recount(pp);
                        recount(p);
                        if (s >= 3) {
                            var ppp = n_stack[s - 3];
                            if (ppp.right === pp) {
                                ppp.right = p;
                            }
                            else {
                                ppp.left = p;
                            }
                        }
                        break;
                    }
                }
                else {
                    var y = pp.left;
                    if (y && y._color === RED) {
                        //console.log("RLr")
                        p._color = BLACK;
                        pp.left = repaint(BLACK, y);
                        pp._color = RED;
                        s -= 1;
                    }
                    else {
                        //console.log("RLb")
                        p.left = n.right;
                        pp._color = RED;
                        pp.right = n.left;
                        n._color = BLACK;
                        n.right = p;
                        n.left = pp;
                        n_stack[s - 2] = n;
                        n_stack[s - 1] = p;
                        recount(pp);
                        recount(p);
                        recount(n);
                        if (s >= 3) {
                            var ppp = n_stack[s - 3];
                            if (ppp.right === pp) {
                                ppp.right = n;
                            }
                            else {
                                ppp.left = n;
                            }
                        }
                        break;
                    }
                }
            }
        }
        //Return new tree
        n_stack[0]._color = BLACK;
        return new RedBlackTree(cmp, n_stack[0]);
    };
    //Visit all nodes inorder
    function doVisitFull(visit, node) {
        if (node.left) {
            var v = doVisitFull(visit, node.left);
            if (v) {
                return v;
            }
        }
        var v = visit(node.key, node.value);
        if (v) {
            return v;
        }
        if (node.right) {
            return doVisitFull(visit, node.right);
        }
    }
    //Visit half nodes in order
    function doVisitHalf(lo, compare, visit, node) {
        var l = compare(lo, node.key);
        if (l <= 0) {
            if (node.left) {
                var v = doVisitHalf(lo, compare, visit, node.left);
                if (v) {
                    return v;
                }
            }
            var v = visit(node.key, node.value);
            if (v) {
                return v;
            }
        }
        if (node.right) {
            return doVisitHalf(lo, compare, visit, node.right);
        }
    }
    //Visit all nodes within a range
    function doVisit(lo, hi, compare, visit, node) {
        var l = compare(lo, node.key);
        var h = compare(hi, node.key);
        var v;
        if (l <= 0) {
            if (node.left) {
                v = doVisit(lo, hi, compare, visit, node.left);
                if (v) {
                    return v;
                }
            }
            if (h > 0) {
                v = visit(node.key, node.value);
                if (v) {
                    return v;
                }
            }
        }
        if (h > 0 && node.right) {
            return doVisit(lo, hi, compare, visit, node.right);
        }
    }
    proto.forEach = function rbTreeForEach(visit, lo, hi) {
        if (!this.root) {
            return;
        }
        switch (arguments.length) {
            case 1:
                return doVisitFull(visit, this.root);
                break;
            case 2:
                return doVisitHalf(lo, this._compare, visit, this.root);
                break;
            case 3:
                if (this._compare(lo, hi) >= 0) {
                    return;
                }
                return doVisit(lo, hi, this._compare, visit, this.root);
                break;
        }
    };
    //First item in list
    Object.defineProperty(proto, "begin", {
        get: function () {
            var stack = [];
            var n = this.root;
            while (n) {
                stack.push(n);
                n = n.left;
            }
            return new RedBlackTreeIterator(this, stack);
        }
    });
    //Last item in list
    Object.defineProperty(proto, "end", {
        get: function () {
            var stack = [];
            var n = this.root;
            while (n) {
                stack.push(n);
                n = n.right;
            }
            return new RedBlackTreeIterator(this, stack);
        }
    });
    //Find the ith item in the tree
    proto.at = function (idx) {
        if (idx < 0) {
            return new RedBlackTreeIterator(this, []);
        }
        var n = this.root;
        var stack = [];
        while (true) {
            stack.push(n);
            if (n.left) {
                if (idx < n.left._count) {
                    n = n.left;
                    continue;
                }
                idx -= n.left._count;
            }
            if (!idx) {
                return new RedBlackTreeIterator(this, stack);
            }
            idx -= 1;
            if (n.right) {
                if (idx >= n.right._count) {
                    break;
                }
                n = n.right;
            }
            else {
                break;
            }
        }
        return new RedBlackTreeIterator(this, []);
    };
    proto.ge = function (key) {
        var cmp = this._compare;
        var n = this.root;
        var stack = [];
        var last_ptr = 0;
        while (n) {
            var d = cmp(key, n.key);
            stack.push(n);
            if (d <= 0) {
                last_ptr = stack.length;
            }
            if (d <= 0) {
                n = n.left;
            }
            else {
                n = n.right;
            }
        }
        stack.length = last_ptr;
        return new RedBlackTreeIterator(this, stack);
    };
    proto.gt = function (key) {
        var cmp = this._compare;
        var n = this.root;
        var stack = [];
        var last_ptr = 0;
        while (n) {
            var d = cmp(key, n.key);
            stack.push(n);
            if (d < 0) {
                last_ptr = stack.length;
            }
            if (d < 0) {
                n = n.left;
            }
            else {
                n = n.right;
            }
        }
        stack.length = last_ptr;
        return new RedBlackTreeIterator(this, stack);
    };
    proto.lt = function (key) {
        var cmp = this._compare;
        var n = this.root;
        var stack = [];
        var last_ptr = 0;
        while (n) {
            var d = cmp(key, n.key);
            stack.push(n);
            if (d > 0) {
                last_ptr = stack.length;
            }
            if (d <= 0) {
                n = n.left;
            }
            else {
                n = n.right;
            }
        }
        stack.length = last_ptr;
        return new RedBlackTreeIterator(this, stack);
    };
    proto.le = function (key) {
        var cmp = this._compare;
        var n = this.root;
        var stack = [];
        var last_ptr = 0;
        while (n) {
            var d = cmp(key, n.key);
            stack.push(n);
            if (d >= 0) {
                last_ptr = stack.length;
            }
            if (d < 0) {
                n = n.left;
            }
            else {
                n = n.right;
            }
        }
        stack.length = last_ptr;
        return new RedBlackTreeIterator(this, stack);
    };
    //Finds the item with key if it exists
    proto.find = function (key) {
        var cmp = this._compare;
        var n = this.root;
        var stack = [];
        while (n) {
            var d = cmp(key, n.key);
            stack.push(n);
            if (d === 0) {
                return new RedBlackTreeIterator(this, stack);
            }
            if (d <= 0) {
                n = n.left;
            }
            else {
                n = n.right;
            }
        }
        return new RedBlackTreeIterator(this, []);
    };
    //Removes item with key from tree
    proto.remove = function (key) {
        var iter = this.find(key);
        if (iter) {
            return iter.remove();
        }
        return this;
    };
    //Returns the item at `key`
    proto.get = function (key) {
        var cmp = this._compare;
        var n = this.root;
        while (n) {
            var d = cmp(key, n.key);
            if (d === 0) {
                return n.value;
            }
            if (d <= 0) {
                n = n.left;
            }
            else {
                n = n.right;
            }
        }
        return;
    };
    //Iterator for red black tree
    function RedBlackTreeIterator(tree, stack) {
        this.tree = tree;
        this._stack = stack;
    }
    var iproto = RedBlackTreeIterator.prototype;
    //Test if iterator is valid
    Object.defineProperty(iproto, "valid", {
        get: function () {
            return this._stack.length > 0;
        }
    });
    //Node of the iterator
    Object.defineProperty(iproto, "node", {
        get: function () {
            if (this._stack.length > 0) {
                return this._stack[this._stack.length - 1];
            }
            return null;
        },
        enumerable: true
    });
    //Makes a copy of an iterator
    iproto.clone = function () {
        return new RedBlackTreeIterator(this.tree, this._stack.slice());
    };
    //Swaps two nodes
    function swapNode(n, v) {
        n.key = v.key;
        n.value = v.value;
        n.left = v.left;
        n.right = v.right;
        n._color = v._color;
        n._count = v._count;
    }
    //Fix up a double black node in a tree
    function fixDoubleBlack(stack) {
        var n, p, s, z;
        for (var i = stack.length - 1; i >= 0; --i) {
            n = stack[i];
            if (i === 0) {
                n._color = BLACK;
                return;
            }
            //console.log("visit node:", n.key, i, stack[i].key, stack[i-1].key)
            p = stack[i - 1];
            if (p.left === n) {
                //console.log("left child")
                s = p.right;
                if (s.right && s.right._color === RED) {
                    //console.log("case 1: right sibling child red")
                    s = p.right = cloneNode(s);
                    z = s.right = cloneNode(s.right);
                    p.right = s.left;
                    s.left = p;
                    s.right = z;
                    s._color = p._color;
                    n._color = BLACK;
                    p._color = BLACK;
                    z._color = BLACK;
                    recount(p);
                    recount(s);
                    if (i > 1) {
                        var pp = stack[i - 2];
                        if (pp.left === p) {
                            pp.left = s;
                        }
                        else {
                            pp.right = s;
                        }
                    }
                    stack[i - 1] = s;
                    return;
                }
                else if (s.left && s.left._color === RED) {
                    //console.log("case 1: left sibling child red")
                    s = p.right = cloneNode(s);
                    z = s.left = cloneNode(s.left);
                    p.right = z.left;
                    s.left = z.right;
                    z.left = p;
                    z.right = s;
                    z._color = p._color;
                    p._color = BLACK;
                    s._color = BLACK;
                    n._color = BLACK;
                    recount(p);
                    recount(s);
                    recount(z);
                    if (i > 1) {
                        var pp = stack[i - 2];
                        if (pp.left === p) {
                            pp.left = z;
                        }
                        else {
                            pp.right = z;
                        }
                    }
                    stack[i - 1] = z;
                    return;
                }
                if (s._color === BLACK) {
                    if (p._color === RED) {
                        //console.log("case 2: black sibling, red parent", p.right.value)
                        p._color = BLACK;
                        p.right = repaint(RED, s);
                        return;
                    }
                    else {
                        //console.log("case 2: black sibling, black parent", p.right.value)
                        p.right = repaint(RED, s);
                        continue;
                    }
                }
                else {
                    //console.log("case 3: red sibling")
                    s = cloneNode(s);
                    p.right = s.left;
                    s.left = p;
                    s._color = p._color;
                    p._color = RED;
                    recount(p);
                    recount(s);
                    if (i > 1) {
                        var pp = stack[i - 2];
                        if (pp.left === p) {
                            pp.left = s;
                        }
                        else {
                            pp.right = s;
                        }
                    }
                    stack[i - 1] = s;
                    stack[i] = p;
                    if (i + 1 < stack.length) {
                        stack[i + 1] = n;
                    }
                    else {
                        stack.push(n);
                    }
                    i = i + 2;
                }
            }
            else {
                //console.log("right child")
                s = p.left;
                if (s.left && s.left._color === RED) {
                    //console.log("case 1: left sibling child red", p.value, p._color)
                    s = p.left = cloneNode(s);
                    z = s.left = cloneNode(s.left);
                    p.left = s.right;
                    s.right = p;
                    s.left = z;
                    s._color = p._color;
                    n._color = BLACK;
                    p._color = BLACK;
                    z._color = BLACK;
                    recount(p);
                    recount(s);
                    if (i > 1) {
                        var pp = stack[i - 2];
                        if (pp.right === p) {
                            pp.right = s;
                        }
                        else {
                            pp.left = s;
                        }
                    }
                    stack[i - 1] = s;
                    return;
                }
                else if (s.right && s.right._color === RED) {
                    //console.log("case 1: right sibling child red")
                    s = p.left = cloneNode(s);
                    z = s.right = cloneNode(s.right);
                    p.left = z.right;
                    s.right = z.left;
                    z.right = p;
                    z.left = s;
                    z._color = p._color;
                    p._color = BLACK;
                    s._color = BLACK;
                    n._color = BLACK;
                    recount(p);
                    recount(s);
                    recount(z);
                    if (i > 1) {
                        var pp = stack[i - 2];
                        if (pp.right === p) {
                            pp.right = z;
                        }
                        else {
                            pp.left = z;
                        }
                    }
                    stack[i - 1] = z;
                    return;
                }
                if (s._color === BLACK) {
                    if (p._color === RED) {
                        //console.log("case 2: black sibling, red parent")
                        p._color = BLACK;
                        p.left = repaint(RED, s);
                        return;
                    }
                    else {
                        //console.log("case 2: black sibling, black parent")
                        p.left = repaint(RED, s);
                        continue;
                    }
                }
                else {
                    //console.log("case 3: red sibling")
                    s = cloneNode(s);
                    p.left = s.right;
                    s.right = p;
                    s._color = p._color;
                    p._color = RED;
                    recount(p);
                    recount(s);
                    if (i > 1) {
                        var pp = stack[i - 2];
                        if (pp.right === p) {
                            pp.right = s;
                        }
                        else {
                            pp.left = s;
                        }
                    }
                    stack[i - 1] = s;
                    stack[i] = p;
                    if (i + 1 < stack.length) {
                        stack[i + 1] = n;
                    }
                    else {
                        stack.push(n);
                    }
                    i = i + 2;
                }
            }
        }
    }
    //Removes item at iterator from tree
    iproto.remove = function () {
        var stack = this._stack;
        if (stack.length === 0) {
            return this.tree;
        }
        //First copy path to node
        var cstack = new Array(stack.length);
        var n = stack[stack.length - 1];
        cstack[cstack.length - 1] = new RBNode(n._color, n.key, n.value, n.left, n.right, n._count);
        for (var i = stack.length - 2; i >= 0; --i) {
            var n = stack[i];
            if (n.left === stack[i + 1]) {
                cstack[i] = new RBNode(n._color, n.key, n.value, cstack[i + 1], n.right, n._count);
            }
            else {
                cstack[i] = new RBNode(n._color, n.key, n.value, n.left, cstack[i + 1], n._count);
            }
        }
        //Get node
        n = cstack[cstack.length - 1];
        //console.log("start remove: ", n.value)
        //If not leaf, then swap with previous node
        if (n.left && n.right) {
            //console.log("moving to leaf")
            //First walk to previous leaf
            var split = cstack.length;
            n = n.left;
            while (n.right) {
                cstack.push(n);
                n = n.right;
            }
            //Copy path to leaf
            var v = cstack[split - 1];
            cstack.push(new RBNode(n._color, v.key, v.value, n.left, n.right, n._count));
            cstack[split - 1].key = n.key;
            cstack[split - 1].value = n.value;
            //Fix up stack
            for (var i = cstack.length - 2; i >= split; --i) {
                n = cstack[i];
                cstack[i] = new RBNode(n._color, n.key, n.value, n.left, cstack[i + 1], n._count);
            }
            cstack[split - 1].left = cstack[split];
        }
        //console.log("stack=", cstack.map(function(v) { return v.value }))
        //Remove leaf node
        n = cstack[cstack.length - 1];
        if (n._color === RED) {
            //Easy case: removing red leaf
            //console.log("RED leaf")
            var p = cstack[cstack.length - 2];
            if (p.left === n) {
                p.left = null;
            }
            else if (p.right === n) {
                p.right = null;
            }
            cstack.pop();
            for (var i = 0; i < cstack.length; ++i) {
                cstack[i]._count--;
            }
            return new RedBlackTree(this.tree._compare, cstack[0]);
        }
        else {
            if (n.left || n.right) {
                //Second easy case:  Single child black parent
                //console.log("BLACK single child")
                if (n.left) {
                    swapNode(n, n.left);
                }
                else if (n.right) {
                    swapNode(n, n.right);
                }
                //Child must be red, so repaint it black to balance color
                n._color = BLACK;
                for (var i = 0; i < cstack.length - 1; ++i) {
                    cstack[i]._count--;
                }
                return new RedBlackTree(this.tree._compare, cstack[0]);
            }
            else if (cstack.length === 1) {
                //Third easy case: root
                //console.log("ROOT")
                return new RedBlackTree(this.tree._compare, null);
            }
            else {
                //Hard case: Repaint n, and then do some nasty stuff
                //console.log("BLACK leaf no children")
                for (var i = 0; i < cstack.length; ++i) {
                    cstack[i]._count--;
                }
                var parent = cstack[cstack.length - 2];
                fixDoubleBlack(cstack);
                //Fix up links
                if (parent.left === n) {
                    parent.left = null;
                }
                else {
                    parent.right = null;
                }
            }
        }
        return new RedBlackTree(this.tree._compare, cstack[0]);
    };
    //Returns key
    Object.defineProperty(iproto, "key", {
        get: function () {
            if (this._stack.length > 0) {
                return this._stack[this._stack.length - 1].key;
            }
            return;
        },
        enumerable: true
    });
    //Returns value
    Object.defineProperty(iproto, "value", {
        get: function () {
            if (this._stack.length > 0) {
                return this._stack[this._stack.length - 1].value;
            }
            return;
        },
        enumerable: true
    });
    //Returns the position of this iterator in the sorted list
    Object.defineProperty(iproto, "index", {
        get: function () {
            var idx = 0;
            var stack = this._stack;
            if (stack.length === 0) {
                var r = this.tree.root;
                if (r) {
                    return r._count;
                }
                return 0;
            }
            else if (stack[stack.length - 1].left) {
                idx = stack[stack.length - 1].left._count;
            }
            for (var s = stack.length - 2; s >= 0; --s) {
                if (stack[s + 1] === stack[s].right) {
                    ++idx;
                    if (stack[s].left) {
                        idx += stack[s].left._count;
                    }
                }
            }
            return idx;
        },
        enumerable: true
    });
    //Advances iterator to next element in list
    iproto.next = function () {
        var stack = this._stack;
        if (stack.length === 0) {
            return;
        }
        var n = stack[stack.length - 1];
        if (n.right) {
            n = n.right;
            while (n) {
                stack.push(n);
                n = n.left;
            }
        }
        else {
            stack.pop();
            while (stack.length > 0 && stack[stack.length - 1].right === n) {
                n = stack[stack.length - 1];
                stack.pop();
            }
        }
    };
    //Checks if iterator is at end of tree
    Object.defineProperty(iproto, "hasNext", {
        get: function () {
            var stack = this._stack;
            if (stack.length === 0) {
                return false;
            }
            if (stack[stack.length - 1].right) {
                return true;
            }
            for (var s = stack.length - 1; s > 0; --s) {
                if (stack[s - 1].left === stack[s]) {
                    return true;
                }
            }
            return false;
        }
    });
    //Update value
    iproto.update = function (value) {
        var stack = this._stack;
        if (stack.length === 0) {
            throw new Error("Can't update empty node!");
        }
        var cstack = new Array(stack.length);
        var n = stack[stack.length - 1];
        cstack[cstack.length - 1] = new RBNode(n._color, n.key, value, n.left, n.right, n._count);
        for (var i = stack.length - 2; i >= 0; --i) {
            n = stack[i];
            if (n.left === stack[i + 1]) {
                cstack[i] = new RBNode(n._color, n.key, n.value, cstack[i + 1], n.right, n._count);
            }
            else {
                cstack[i] = new RBNode(n._color, n.key, n.value, n.left, cstack[i + 1], n._count);
            }
        }
        return new RedBlackTree(this.tree._compare, cstack[0]);
    };
    //Moves iterator backward one element
    iproto.prev = function () {
        var stack = this._stack;
        if (stack.length === 0) {
            return;
        }
        var n = stack[stack.length - 1];
        if (n.left) {
            n = n.left;
            while (n) {
                stack.push(n);
                n = n.right;
            }
        }
        else {
            stack.pop();
            while (stack.length > 0 && stack[stack.length - 1].left === n) {
                n = stack[stack.length - 1];
                stack.pop();
            }
        }
    };
    //Checks if iterator is at start of tree
    Object.defineProperty(iproto, "hasPrev", {
        get: function () {
            var stack = this._stack;
            if (stack.length === 0) {
                return false;
            }
            if (stack[stack.length - 1].left) {
                return true;
            }
            for (var s = stack.length - 1; s > 0; --s) {
                if (stack[s - 1].right === stack[s]) {
                    return true;
                }
            }
            return false;
        }
    });
    //Default comparison function
    function defaultCompare(a, b) {
        if (a < b) {
            return -1;
        }
        if (a > b) {
            return 1;
        }
        return 0;
    }
    //Build a tree
    function createRBTree(compare) {
        return new RedBlackTree(compare || defaultCompare, null);
    }
},
/* eslint/lib/rules/indent-legacy.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    /* istanbul ignore next: this rule has known coverage issues, but it's deprecated and shouldn't be updated in the future anyway. */
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent indentation",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/indent-legacy"
            },
            deprecated: true,
            replacedBy: ["indent"],
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["tab"]
                        },
                        {
                            type: "integer",
                            minimum: 0
                        }
                    ]
                },
                {
                    type: "object",
                    properties: {
                        SwitchCase: {
                            type: "integer",
                            minimum: 0
                        },
                        VariableDeclarator: {
                            oneOf: [
                                {
                                    type: "integer",
                                    minimum: 0
                                },
                                {
                                    type: "object",
                                    properties: {
                                        var: {
                                            type: "integer",
                                            minimum: 0
                                        },
                                        let: {
                                            type: "integer",
                                            minimum: 0
                                        },
                                        const: {
                                            type: "integer",
                                            minimum: 0
                                        }
                                    }
                                }
                            ]
                        },
                        outerIIFEBody: {
                            type: "integer",
                            minimum: 0
                        },
                        MemberExpression: {
                            type: "integer",
                            minimum: 0
                        },
                        FunctionDeclaration: {
                            type: "object",
                            properties: {
                                parameters: {
                                    oneOf: [
                                        {
                                            type: "integer",
                                            minimum: 0
                                        },
                                        {
                                            enum: ["first"]
                                        }
                                    ]
                                },
                                body: {
                                    type: "integer",
                                    minimum: 0
                                }
                            }
                        },
                        FunctionExpression: {
                            type: "object",
                            properties: {
                                parameters: {
                                    oneOf: [
                                        {
                                            type: "integer",
                                            minimum: 0
                                        },
                                        {
                                            enum: ["first"]
                                        }
                                    ]
                                },
                                body: {
                                    type: "integer",
                                    minimum: 0
                                }
                            }
                        },
                        CallExpression: {
                            type: "object",
                            properties: {
                                parameters: {
                                    oneOf: [
                                        {
                                            type: "integer",
                                            minimum: 0
                                        },
                                        {
                                            enum: ["first"]
                                        }
                                    ]
                                }
                            }
                        },
                        ArrayExpression: {
                            oneOf: [
                                {
                                    type: "integer",
                                    minimum: 0
                                },
                                {
                                    enum: ["first"]
                                }
                            ]
                        },
                        ObjectExpression: {
                            oneOf: [
                                {
                                    type: "integer",
                                    minimum: 0
                                },
                                {
                                    enum: ["first"]
                                }
                            ]
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                expected: "Expected indentation of {{expected}} but found {{actual}}."
            }
        },
        create(context) {
            const DEFAULT_VARIABLE_INDENT = 1;
            const DEFAULT_PARAMETER_INDENT = null; // For backwards compatibility, don't check parameter indentation unless specified in the config
            const DEFAULT_FUNCTION_BODY_INDENT = 1;
            let indentType = "space";
            let indentSize = 4;
            const options = {
                SwitchCase: 0,
                VariableDeclarator: {
                    var: DEFAULT_VARIABLE_INDENT,
                    let: DEFAULT_VARIABLE_INDENT,
                    const: DEFAULT_VARIABLE_INDENT
                },
                outerIIFEBody: null,
                FunctionDeclaration: {
                    parameters: DEFAULT_PARAMETER_INDENT,
                    body: DEFAULT_FUNCTION_BODY_INDENT
                },
                FunctionExpression: {
                    parameters: DEFAULT_PARAMETER_INDENT,
                    body: DEFAULT_FUNCTION_BODY_INDENT
                },
                CallExpression: {
                    arguments: DEFAULT_PARAMETER_INDENT
                },
                ArrayExpression: 1,
                ObjectExpression: 1
            };
            const sourceCode = context.getSourceCode();
            if (context.options.length) {
                if (context.options[0] === "tab") {
                    indentSize = 1;
                    indentType = "tab";
                }
                else /* istanbul ignore else : this will be caught by options validation */ if (typeof context.options[0] === "number") {
                    indentSize = context.options[0];
                    indentType = "space";
                }
                if (context.options[1]) {
                    const opts = context.options[1];
                    options.SwitchCase = opts.SwitchCase || 0;
                    const variableDeclaratorRules = opts.VariableDeclarator;
                    if (typeof variableDeclaratorRules === "number") {
                        options.VariableDeclarator = {
                            var: variableDeclaratorRules,
                            let: variableDeclaratorRules,
                            const: variableDeclaratorRules
                        };
                    }
                    else if (typeof variableDeclaratorRules === "object") {
                        Object.assign(options.VariableDeclarator, variableDeclaratorRules);
                    }
                    if (typeof opts.outerIIFEBody === "number") {
                        options.outerIIFEBody = opts.outerIIFEBody;
                    }
                    if (typeof opts.MemberExpression === "number") {
                        options.MemberExpression = opts.MemberExpression;
                    }
                    if (typeof opts.FunctionDeclaration === "object") {
                        Object.assign(options.FunctionDeclaration, opts.FunctionDeclaration);
                    }
                    if (typeof opts.FunctionExpression === "object") {
                        Object.assign(options.FunctionExpression, opts.FunctionExpression);
                    }
                    if (typeof opts.CallExpression === "object") {
                        Object.assign(options.CallExpression, opts.CallExpression);
                    }
                    if (typeof opts.ArrayExpression === "number" || typeof opts.ArrayExpression === "string") {
                        options.ArrayExpression = opts.ArrayExpression;
                    }
                    if (typeof opts.ObjectExpression === "number" || typeof opts.ObjectExpression === "string") {
                        options.ObjectExpression = opts.ObjectExpression;
                    }
                }
            }
            const caseIndentStore = {};
            /**
             * Creates an error message for a line, given the expected/actual indentation.
             * @param {int} expectedAmount The expected amount of indentation characters for this line
             * @param {int} actualSpaces The actual number of indentation spaces that were found on this line
             * @param {int} actualTabs The actual number of indentation tabs that were found on this line
             * @returns {string} An error message for this line
             */
            function createErrorMessageData(expectedAmount, actualSpaces, actualTabs) {
                const expectedStatement = `${expectedAmount} ${indentType}${expectedAmount === 1 ? "" : "s"}`; // e.g. "2 tabs"
                const foundSpacesWord = `space${actualSpaces === 1 ? "" : "s"}`; // e.g. "space"
                const foundTabsWord = `tab${actualTabs === 1 ? "" : "s"}`; // e.g. "tabs"
                let foundStatement;
                if (actualSpaces > 0 && actualTabs > 0) {
                    foundStatement = `${actualSpaces} ${foundSpacesWord} and ${actualTabs} ${foundTabsWord}`; // e.g. "1 space and 2 tabs"
                }
                else if (actualSpaces > 0) {
                    /*
                     * Abbreviate the message if the expected indentation is also spaces.
                     * e.g. 'Expected 4 spaces but found 2' rather than 'Expected 4 spaces but found 2 spaces'
                     */
                    foundStatement = indentType === "space" ? actualSpaces : `${actualSpaces} ${foundSpacesWord}`;
                }
                else if (actualTabs > 0) {
                    foundStatement = indentType === "tab" ? actualTabs : `${actualTabs} ${foundTabsWord}`;
                }
                else {
                    foundStatement = "0";
                }
                return {
                    expected: expectedStatement,
                    actual: foundStatement
                };
            }
            /**
             * Reports a given indent violation
             * @param {ASTNode} node Node violating the indent rule
             * @param {int} needed Expected indentation character count
             * @param {int} gottenSpaces Indentation space count in the actual node/code
             * @param {int} gottenTabs Indentation tab count in the actual node/code
             * @param {Object} [loc] Error line and column location
             * @param {boolean} isLastNodeCheck Is the error for last node check
             * @returns {void}
             */
            function report(node, needed, gottenSpaces, gottenTabs, loc, isLastNodeCheck) {
                if (gottenSpaces && gottenTabs) {
                    // To avoid conflicts with `no-mixed-spaces-and-tabs`, don't report lines that have both spaces and tabs.
                    return;
                }
                const desiredIndent = (indentType === "space" ? " " : "\t").repeat(needed);
                const textRange = isLastNodeCheck
                    ? [node.range[1] - node.loc.end.column, node.range[1] - node.loc.end.column + gottenSpaces + gottenTabs]
                    : [node.range[0] - node.loc.start.column, node.range[0] - node.loc.start.column + gottenSpaces + gottenTabs];
                context.report({
                    node,
                    loc,
                    messageId: "expected",
                    data: createErrorMessageData(needed, gottenSpaces, gottenTabs),
                    fix: fixer => fixer.replaceTextRange(textRange, desiredIndent)
                });
            }
            /**
             * Get the actual indent of node
             * @param {ASTNode|Token} node Node to examine
             * @param {boolean} [byLastLine=false] get indent of node's last line
             * @returns {Object} The node's indent. Contains keys `space` and `tab`, representing the indent of each character. Also
             * contains keys `goodChar` and `badChar`, where `goodChar` is the amount of the user's desired indentation character, and
             * `badChar` is the amount of the other indentation character.
             */
            function getNodeIndent(node, byLastLine) {
                const token = byLastLine ? sourceCode.getLastToken(node) : sourceCode.getFirstToken(node);
                const srcCharsBeforeNode = sourceCode.getText(token, token.loc.start.column).split("");
                const indentChars = srcCharsBeforeNode.slice(0, srcCharsBeforeNode.findIndex(char => char !== " " && char !== "\t"));
                const spaces = indentChars.filter(char => char === " ").length;
                const tabs = indentChars.filter(char => char === "\t").length;
                return {
                    space: spaces,
                    tab: tabs,
                    goodChar: indentType === "space" ? spaces : tabs,
                    badChar: indentType === "space" ? tabs : spaces
                };
            }
            /**
             * Checks node is the first in its own start line. By default it looks by start line.
             * @param {ASTNode} node The node to check
             * @param {boolean} [byEndLocation=false] Lookup based on start position or end
             * @returns {boolean} true if its the first in the its start line
             */
            function isNodeFirstInLine(node, byEndLocation) {
                const firstToken = byEndLocation === true ? sourceCode.getLastToken(node, 1) : sourceCode.getTokenBefore(node), startLine = byEndLocation === true ? node.loc.end.line : node.loc.start.line, endLine = firstToken ? firstToken.loc.end.line : -1;
                return startLine !== endLine;
            }
            /**
             * Check indent for node
             * @param {ASTNode} node Node to check
             * @param {int} neededIndent needed indent
             * @returns {void}
             */
            function checkNodeIndent(node, neededIndent) {
                const actualIndent = getNodeIndent(node, false);
                if (node.type !== "ArrayExpression" &&
                    node.type !== "ObjectExpression" &&
                    (actualIndent.goodChar !== neededIndent || actualIndent.badChar !== 0) &&
                    isNodeFirstInLine(node)) {
                    report(node, neededIndent, actualIndent.space, actualIndent.tab);
                }
                if (node.type === "IfStatement" && node.alternate) {
                    const elseToken = sourceCode.getTokenBefore(node.alternate);
                    checkNodeIndent(elseToken, neededIndent);
                    if (!isNodeFirstInLine(node.alternate)) {
                        checkNodeIndent(node.alternate, neededIndent);
                    }
                }
                if (node.type === "TryStatement" && node.handler) {
                    const catchToken = sourceCode.getFirstToken(node.handler);
                    checkNodeIndent(catchToken, neededIndent);
                }
                if (node.type === "TryStatement" && node.finalizer) {
                    const finallyToken = sourceCode.getTokenBefore(node.finalizer);
                    checkNodeIndent(finallyToken, neededIndent);
                }
                if (node.type === "DoWhileStatement") {
                    const whileToken = sourceCode.getTokenAfter(node.body);
                    checkNodeIndent(whileToken, neededIndent);
                }
            }
            /**
             * Check indent for nodes list
             * @param {ASTNode[]} nodes list of node objects
             * @param {int} indent needed indent
             * @returns {void}
             */
            function checkNodesIndent(nodes, indent) {
                nodes.forEach(node => checkNodeIndent(node, indent));
            }
            /**
             * Check last node line indent this detects, that block closed correctly
             * @param {ASTNode} node Node to examine
             * @param {int} lastLineIndent needed indent
             * @returns {void}
             */
            function checkLastNodeLineIndent(node, lastLineIndent) {
                const lastToken = sourceCode.getLastToken(node);
                const endIndent = getNodeIndent(lastToken, true);
                if ((endIndent.goodChar !== lastLineIndent || endIndent.badChar !== 0) && isNodeFirstInLine(node, true)) {
                    report(node, lastLineIndent, endIndent.space, endIndent.tab, { line: lastToken.loc.start.line, column: lastToken.loc.start.column }, true);
                }
            }
            /**
             * Check last node line indent this detects, that block closed correctly
             * This function for more complicated return statement case, where closing parenthesis may be followed by ';'
             * @param {ASTNode} node Node to examine
             * @param {int} firstLineIndent first line needed indent
             * @returns {void}
             */
            function checkLastReturnStatementLineIndent(node, firstLineIndent) {
                /*
                 * in case if return statement ends with ');' we have traverse back to ')'
                 * otherwise we'll measure indent for ';' and replace ')'
                 */
                const lastToken = sourceCode.getLastToken(node, astUtils.isClosingParenToken);
                const textBeforeClosingParenthesis = sourceCode.getText(lastToken, lastToken.loc.start.column).slice(0, -1);
                if (textBeforeClosingParenthesis.trim()) {
                    // There are tokens before the closing paren, don't report this case
                    return;
                }
                const endIndent = getNodeIndent(lastToken, true);
                if (endIndent.goodChar !== firstLineIndent) {
                    report(node, firstLineIndent, endIndent.space, endIndent.tab, { line: lastToken.loc.start.line, column: lastToken.loc.start.column }, true);
                }
            }
            /**
             * Check first node line indent is correct
             * @param {ASTNode} node Node to examine
             * @param {int} firstLineIndent needed indent
             * @returns {void}
             */
            function checkFirstNodeLineIndent(node, firstLineIndent) {
                const startIndent = getNodeIndent(node, false);
                if ((startIndent.goodChar !== firstLineIndent || startIndent.badChar !== 0) && isNodeFirstInLine(node)) {
                    report(node, firstLineIndent, startIndent.space, startIndent.tab, { line: node.loc.start.line, column: node.loc.start.column });
                }
            }
            /**
             * Returns a parent node of given node based on a specified type
             * if not present then return null
             * @param {ASTNode} node node to examine
             * @param {string} type type that is being looked for
             * @param {string} stopAtList end points for the evaluating code
             * @returns {ASTNode|void} if found then node otherwise null
             */
            function getParentNodeByType(node, type, stopAtList) {
                let parent = node.parent;
                const stopAtSet = new Set(stopAtList || ["Program"]);
                while (parent.type !== type && !stopAtSet.has(parent.type) && parent.type !== "Program") {
                    parent = parent.parent;
                }
                return parent.type === type ? parent : null;
            }
            /**
             * Returns the VariableDeclarator based on the current node
             * if not present then return null
             * @param {ASTNode} node node to examine
             * @returns {ASTNode|void} if found then node otherwise null
             */
            function getVariableDeclaratorNode(node) {
                return getParentNodeByType(node, "VariableDeclarator");
            }
            /**
             * Check to see if the node is part of the multi-line variable declaration.
             * Also if its on the same line as the varNode
             * @param {ASTNode} node node to check
             * @param {ASTNode} varNode variable declaration node to check against
             * @returns {boolean} True if all the above condition satisfy
             */
            function isNodeInVarOnTop(node, varNode) {
                return varNode &&
                    varNode.parent.loc.start.line === node.loc.start.line &&
                    varNode.parent.declarations.length > 1;
            }
            /**
             * Check to see if the argument before the callee node is multi-line and
             * there should only be 1 argument before the callee node
             * @param {ASTNode} node node to check
             * @returns {boolean} True if arguments are multi-line
             */
            function isArgBeforeCalleeNodeMultiline(node) {
                const parent = node.parent;
                if (parent.arguments.length >= 2 && parent.arguments[1] === node) {
                    return parent.arguments[0].loc.end.line > parent.arguments[0].loc.start.line;
                }
                return false;
            }
            /**
             * Check to see if the node is a file level IIFE
             * @param {ASTNode} node The function node to check.
             * @returns {boolean} True if the node is the outer IIFE
             */
            function isOuterIIFE(node) {
                const parent = node.parent;
                let stmt = parent.parent;
                /*
                 * Verify that the node is an IIEF
                 */
                if (parent.type !== "CallExpression" ||
                    parent.callee !== node) {
                    return false;
                }
                /*
                 * Navigate legal ancestors to determine whether this IIEF is outer
                 */
                while (stmt.type === "UnaryExpression" && (stmt.operator === "!" ||
                    stmt.operator === "~" ||
                    stmt.operator === "+" ||
                    stmt.operator === "-") ||
                    stmt.type === "AssignmentExpression" ||
                    stmt.type === "LogicalExpression" ||
                    stmt.type === "SequenceExpression" ||
                    stmt.type === "VariableDeclarator") {
                    stmt = stmt.parent;
                }
                return ((stmt.type === "ExpressionStatement" ||
                    stmt.type === "VariableDeclaration") &&
                    stmt.parent && stmt.parent.type === "Program");
            }
            /**
             * Check indent for function block content
             * @param {ASTNode} node A BlockStatement node that is inside of a function.
             * @returns {void}
             */
            function checkIndentInFunctionBlock(node) {
                /*
                 * Search first caller in chain.
                 * Ex.:
                 *
                 * Models <- Identifier
                 *   .User
                 *   .find()
                 *   .exec(function() {
                 *   // function body
                 * });
                 *
                 * Looks for 'Models'
                 */
                const calleeNode = node.parent; // FunctionExpression
                let indent;
                if (calleeNode.parent &&
                    (calleeNode.parent.type === "Property" ||
                        calleeNode.parent.type === "ArrayExpression")) {
                    // If function is part of array or object, comma can be put at left
                    indent = getNodeIndent(calleeNode, false).goodChar;
                }
                else {
                    // If function is standalone, simple calculate indent
                    indent = getNodeIndent(calleeNode).goodChar;
                }
                if (calleeNode.parent.type === "CallExpression") {
                    const calleeParent = calleeNode.parent;
                    if (calleeNode.type !== "FunctionExpression" && calleeNode.type !== "ArrowFunctionExpression") {
                        if (calleeParent && calleeParent.loc.start.line < node.loc.start.line) {
                            indent = getNodeIndent(calleeParent).goodChar;
                        }
                    }
                    else {
                        if (isArgBeforeCalleeNodeMultiline(calleeNode) &&
                            calleeParent.callee.loc.start.line === calleeParent.callee.loc.end.line &&
                            !isNodeFirstInLine(calleeNode)) {
                            indent = getNodeIndent(calleeParent).goodChar;
                        }
                    }
                }
                /*
                 * function body indent should be indent + indent size, unless this
                 * is a FunctionDeclaration, FunctionExpression, or outer IIFE and the corresponding options are enabled.
                 */
                let functionOffset = indentSize;
                if (options.outerIIFEBody !== null && isOuterIIFE(calleeNode)) {
                    functionOffset = options.outerIIFEBody * indentSize;
                }
                else if (calleeNode.type === "FunctionExpression") {
                    functionOffset = options.FunctionExpression.body * indentSize;
                }
                else if (calleeNode.type === "FunctionDeclaration") {
                    functionOffset = options.FunctionDeclaration.body * indentSize;
                }
                indent += functionOffset;
                // check if the node is inside a variable
                const parentVarNode = getVariableDeclaratorNode(node);
                if (parentVarNode && isNodeInVarOnTop(node, parentVarNode)) {
                    indent += indentSize * options.VariableDeclarator[parentVarNode.parent.kind];
                }
                if (node.body.length > 0) {
                    checkNodesIndent(node.body, indent);
                }
                checkLastNodeLineIndent(node, indent - functionOffset);
            }
            /**
             * Checks if the given node starts and ends on the same line
             * @param {ASTNode} node The node to check
             * @returns {boolean} Whether or not the block starts and ends on the same line.
             */
            function isSingleLineNode(node) {
                const lastToken = sourceCode.getLastToken(node), startLine = node.loc.start.line, endLine = lastToken.loc.end.line;
                return startLine === endLine;
            }
            /**
             * Check indent for array block content or object block content
             * @param {ASTNode} node node to examine
             * @returns {void}
             */
            function checkIndentInArrayOrObjectBlock(node) {
                // Skip inline
                if (isSingleLineNode(node)) {
                    return;
                }
                let elements = (node.type === "ArrayExpression") ? node.elements : node.properties;
                // filter out empty elements example would be [ , 2] so remove first element as espree considers it as null
                elements = elements.filter(elem => elem !== null);
                let nodeIndent;
                let elementsIndent;
                const parentVarNode = getVariableDeclaratorNode(node);
                // TODO - come up with a better strategy in future
                if (isNodeFirstInLine(node)) {
                    const parent = node.parent;
                    nodeIndent = getNodeIndent(parent).goodChar;
                    if (!parentVarNode || parentVarNode.loc.start.line !== node.loc.start.line) {
                        if (parent.type !== "VariableDeclarator" || parentVarNode === parentVarNode.parent.declarations[0]) {
                            if (parent.type === "VariableDeclarator" && parentVarNode.loc.start.line === parent.loc.start.line) {
                                nodeIndent += (indentSize * options.VariableDeclarator[parentVarNode.parent.kind]);
                            }
                            else if (parent.type === "ObjectExpression" || parent.type === "ArrayExpression") {
                                const parentElements = node.parent.type === "ObjectExpression" ? node.parent.properties : node.parent.elements;
                                if (parentElements[0] &&
                                    parentElements[0].loc.start.line === parent.loc.start.line &&
                                    parentElements[0].loc.end.line !== parent.loc.start.line) {
                                    /*
                                     * If the first element of the array spans multiple lines, don't increase the expected indentation of the rest.
                                     * e.g. [{
                                     *        foo: 1
                                     *      },
                                     *      {
                                     *        bar: 1
                                     *      }]
                                     * the second object is not indented.
                                     */
                                }
                                else if (typeof options[parent.type] === "number") {
                                    nodeIndent += options[parent.type] * indentSize;
                                }
                                else {
                                    nodeIndent = parentElements[0].loc.start.column;
                                }
                            }
                            else if (parent.type === "CallExpression" || parent.type === "NewExpression") {
                                if (typeof options.CallExpression.arguments === "number") {
                                    nodeIndent += options.CallExpression.arguments * indentSize;
                                }
                                else if (options.CallExpression.arguments === "first") {
                                    if (parent.arguments.indexOf(node) !== -1) {
                                        nodeIndent = parent.arguments[0].loc.start.column;
                                    }
                                }
                                else {
                                    nodeIndent += indentSize;
                                }
                            }
                            else if (parent.type === "LogicalExpression" || parent.type === "ArrowFunctionExpression") {
                                nodeIndent += indentSize;
                            }
                        }
                    }
                    checkFirstNodeLineIndent(node, nodeIndent);
                }
                else {
                    nodeIndent = getNodeIndent(node).goodChar;
                }
                if (options[node.type] === "first") {
                    elementsIndent = elements.length ? elements[0].loc.start.column : 0; // If there are no elements, elementsIndent doesn't matter.
                }
                else {
                    elementsIndent = nodeIndent + indentSize * options[node.type];
                }
                /*
                 * Check if the node is a multiple variable declaration; if so, then
                 * make sure indentation takes that into account.
                 */
                if (isNodeInVarOnTop(node, parentVarNode)) {
                    elementsIndent += indentSize * options.VariableDeclarator[parentVarNode.parent.kind];
                }
                checkNodesIndent(elements, elementsIndent);
                if (elements.length > 0) {
                    // Skip last block line check if last item in same line
                    if (elements[elements.length - 1].loc.end.line === node.loc.end.line) {
                        return;
                    }
                }
                checkLastNodeLineIndent(node, nodeIndent +
                    (isNodeInVarOnTop(node, parentVarNode) ? options.VariableDeclarator[parentVarNode.parent.kind] * indentSize : 0));
            }
            /**
             * Check if the node or node body is a BlockStatement or not
             * @param {ASTNode} node node to test
             * @returns {boolean} True if it or its body is a block statement
             */
            function isNodeBodyBlock(node) {
                return node.type === "BlockStatement" || node.type === "ClassBody" || (node.body && node.body.type === "BlockStatement") ||
                    (node.consequent && node.consequent.type === "BlockStatement");
            }
            /**
             * Check indentation for blocks
             * @param {ASTNode} node node to check
             * @returns {void}
             */
            function blockIndentationCheck(node) {
                // Skip inline blocks
                if (isSingleLineNode(node)) {
                    return;
                }
                if (node.parent && (node.parent.type === "FunctionExpression" ||
                    node.parent.type === "FunctionDeclaration" ||
                    node.parent.type === "ArrowFunctionExpression")) {
                    checkIndentInFunctionBlock(node);
                    return;
                }
                let indent;
                let nodesToCheck = [];
                /*
                 * For this statements we should check indent from statement beginning,
                 * not from the beginning of the block.
                 */
                const statementsWithProperties = [
                    "IfStatement", "WhileStatement", "ForStatement", "ForInStatement", "ForOfStatement", "DoWhileStatement", "ClassDeclaration", "TryStatement"
                ];
                if (node.parent && statementsWithProperties.indexOf(node.parent.type) !== -1 && isNodeBodyBlock(node)) {
                    indent = getNodeIndent(node.parent).goodChar;
                }
                else if (node.parent && node.parent.type === "CatchClause") {
                    indent = getNodeIndent(node.parent.parent).goodChar;
                }
                else {
                    indent = getNodeIndent(node).goodChar;
                }
                if (node.type === "IfStatement" && node.consequent.type !== "BlockStatement") {
                    nodesToCheck = [node.consequent];
                }
                else if (Array.isArray(node.body)) {
                    nodesToCheck = node.body;
                }
                else {
                    nodesToCheck = [node.body];
                }
                if (nodesToCheck.length > 0) {
                    checkNodesIndent(nodesToCheck, indent + indentSize);
                }
                if (node.type === "BlockStatement") {
                    checkLastNodeLineIndent(node, indent);
                }
            }
            /**
             * Filter out the elements which are on the same line of each other or the node.
             * basically have only 1 elements from each line except the variable declaration line.
             * @param {ASTNode} node Variable declaration node
             * @returns {ASTNode[]} Filtered elements
             */
            function filterOutSameLineVars(node) {
                return node.declarations.reduce((finalCollection, elem) => {
                    const lastElem = finalCollection[finalCollection.length - 1];
                    if ((elem.loc.start.line !== node.loc.start.line && !lastElem) ||
                        (lastElem && lastElem.loc.start.line !== elem.loc.start.line)) {
                        finalCollection.push(elem);
                    }
                    return finalCollection;
                }, []);
            }
            /**
             * Check indentation for variable declarations
             * @param {ASTNode} node node to examine
             * @returns {void}
             */
            function checkIndentInVariableDeclarations(node) {
                const elements = filterOutSameLineVars(node);
                const nodeIndent = getNodeIndent(node).goodChar;
                const lastElement = elements[elements.length - 1];
                const elementsIndent = nodeIndent + indentSize * options.VariableDeclarator[node.kind];
                checkNodesIndent(elements, elementsIndent);
                // Only check the last line if there is any token after the last item
                if (sourceCode.getLastToken(node).loc.end.line <= lastElement.loc.end.line) {
                    return;
                }
                const tokenBeforeLastElement = sourceCode.getTokenBefore(lastElement);
                if (tokenBeforeLastElement.value === ",") {
                    // Special case for comma-first syntax where the semicolon is indented
                    checkLastNodeLineIndent(node, getNodeIndent(tokenBeforeLastElement).goodChar);
                }
                else {
                    checkLastNodeLineIndent(node, elementsIndent - indentSize);
                }
            }
            /**
             * Check and decide whether to check for indentation for blockless nodes
             * Scenarios are for or while statements without braces around them
             * @param {ASTNode} node node to examine
             * @returns {void}
             */
            function blockLessNodes(node) {
                if (node.body.type !== "BlockStatement") {
                    blockIndentationCheck(node);
                }
            }
            /**
             * Returns the expected indentation for the case statement
             * @param {ASTNode} node node to examine
             * @param {int} [providedSwitchIndent] indent for switch statement
             * @returns {int} indent size
             */
            function expectedCaseIndent(node, providedSwitchIndent) {
                const switchNode = (node.type === "SwitchStatement") ? node : node.parent;
                const switchIndent = typeof providedSwitchIndent === "undefined"
                    ? getNodeIndent(switchNode).goodChar
                    : providedSwitchIndent;
                let caseIndent;
                if (caseIndentStore[switchNode.loc.start.line]) {
                    return caseIndentStore[switchNode.loc.start.line];
                }
                if (switchNode.cases.length > 0 && options.SwitchCase === 0) {
                    caseIndent = switchIndent;
                }
                else {
                    caseIndent = switchIndent + (indentSize * options.SwitchCase);
                }
                caseIndentStore[switchNode.loc.start.line] = caseIndent;
                return caseIndent;
            }
            /**
             * Checks whether a return statement is wrapped in ()
             * @param {ASTNode} node node to examine
             * @returns {boolean} the result
             */
            function isWrappedInParenthesis(node) {
                const regex = /^return\s*?\(\s*?\);*?/u;
                const statementWithoutArgument = sourceCode.getText(node).replace(sourceCode.getText(node.argument), "");
                return regex.test(statementWithoutArgument);
            }
            return {
                Program(node) {
                    if (node.body.length > 0) {
                        // Root nodes should have no indent
                        checkNodesIndent(node.body, getNodeIndent(node).goodChar);
                    }
                },
                ClassBody: blockIndentationCheck,
                BlockStatement: blockIndentationCheck,
                WhileStatement: blockLessNodes,
                ForStatement: blockLessNodes,
                ForInStatement: blockLessNodes,
                ForOfStatement: blockLessNodes,
                DoWhileStatement: blockLessNodes,
                IfStatement(node) {
                    if (node.consequent.type !== "BlockStatement" && node.consequent.loc.start.line > node.loc.start.line) {
                        blockIndentationCheck(node);
                    }
                },
                VariableDeclaration(node) {
                    if (node.declarations[node.declarations.length - 1].loc.start.line > node.declarations[0].loc.start.line) {
                        checkIndentInVariableDeclarations(node);
                    }
                },
                ObjectExpression(node) {
                    checkIndentInArrayOrObjectBlock(node);
                },
                ArrayExpression(node) {
                    checkIndentInArrayOrObjectBlock(node);
                },
                MemberExpression(node) {
                    if (typeof options.MemberExpression === "undefined") {
                        return;
                    }
                    if (isSingleLineNode(node)) {
                        return;
                    }
                    /*
                     * The typical layout of variable declarations and assignments
                     * alter the expectation of correct indentation. Skip them.
                     * TODO: Add appropriate configuration options for variable
                     * declarations and assignments.
                     */
                    if (getParentNodeByType(node, "VariableDeclarator", ["FunctionExpression", "ArrowFunctionExpression"])) {
                        return;
                    }
                    if (getParentNodeByType(node, "AssignmentExpression", ["FunctionExpression"])) {
                        return;
                    }
                    const propertyIndent = getNodeIndent(node).goodChar + indentSize * options.MemberExpression;
                    const checkNodes = [node.property];
                    const dot = sourceCode.getTokenBefore(node.property);
                    if (dot.type === "Punctuator" && dot.value === ".") {
                        checkNodes.push(dot);
                    }
                    checkNodesIndent(checkNodes, propertyIndent);
                },
                SwitchStatement(node) {
                    // Switch is not a 'BlockStatement'
                    const switchIndent = getNodeIndent(node).goodChar;
                    const caseIndent = expectedCaseIndent(node, switchIndent);
                    checkNodesIndent(node.cases, caseIndent);
                    checkLastNodeLineIndent(node, switchIndent);
                },
                SwitchCase(node) {
                    // Skip inline cases
                    if (isSingleLineNode(node)) {
                        return;
                    }
                    const caseIndent = expectedCaseIndent(node);
                    checkNodesIndent(node.consequent, caseIndent + indentSize);
                },
                FunctionDeclaration(node) {
                    if (isSingleLineNode(node)) {
                        return;
                    }
                    if (options.FunctionDeclaration.parameters === "first" && node.params.length) {
                        checkNodesIndent(node.params.slice(1), node.params[0].loc.start.column);
                    }
                    else if (options.FunctionDeclaration.parameters !== null) {
                        checkNodesIndent(node.params, getNodeIndent(node).goodChar + indentSize * options.FunctionDeclaration.parameters);
                    }
                },
                FunctionExpression(node) {
                    if (isSingleLineNode(node)) {
                        return;
                    }
                    if (options.FunctionExpression.parameters === "first" && node.params.length) {
                        checkNodesIndent(node.params.slice(1), node.params[0].loc.start.column);
                    }
                    else if (options.FunctionExpression.parameters !== null) {
                        checkNodesIndent(node.params, getNodeIndent(node).goodChar + indentSize * options.FunctionExpression.parameters);
                    }
                },
                ReturnStatement(node) {
                    if (isSingleLineNode(node)) {
                        return;
                    }
                    const firstLineIndent = getNodeIndent(node).goodChar;
                    // in case if return statement is wrapped in parenthesis
                    if (isWrappedInParenthesis(node)) {
                        checkLastReturnStatementLineIndent(node, firstLineIndent);
                    }
                    else {
                        checkNodeIndent(node, firstLineIndent);
                    }
                },
                CallExpression(node) {
                    if (isSingleLineNode(node)) {
                        return;
                    }
                    if (options.CallExpression.arguments === "first" && node.arguments.length) {
                        checkNodesIndent(node.arguments.slice(1), node.arguments[0].loc.start.column);
                    }
                    else if (options.CallExpression.arguments !== null) {
                        checkNodesIndent(node.arguments, getNodeIndent(node).goodChar + indentSize * options.CallExpression.arguments);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/init-declarations.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given node is a for loop.
     * @param {ASTNode} block A node to check.
     * @returns {boolean} `true` when the node is a for loop.
     */
    function isForLoop(block) {
        return block.type === "ForInStatement" ||
            block.type === "ForOfStatement" ||
            block.type === "ForStatement";
    }
    /**
     * Checks whether or not a given declarator node has its initializer.
     * @param {ASTNode} node A declarator node to check.
     * @returns {boolean} `true` when the node has its initializer.
     */
    function isInitialized(node) {
        const declaration = node.parent;
        const block = declaration.parent;
        if (isForLoop(block)) {
            if (block.type === "ForStatement") {
                return block.init === declaration;
            }
            return block.left === declaration;
        }
        return Boolean(node.init);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require or disallow initialization in variable declarations",
                category: "Variables",
                recommended: false,
                url: "https://eslint.org/docs/rules/init-declarations"
            },
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always"]
                            }
                        ],
                        minItems: 0,
                        maxItems: 1
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["never"]
                            },
                            {
                                type: "object",
                                properties: {
                                    ignoreForLoopInit: {
                                        type: "boolean"
                                    }
                                },
                                additionalProperties: false
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    }
                ]
            },
            messages: {
                initialized: "Variable '{{idName}}' should be initialized on declaration.",
                notInitialized: "Variable '{{idName}}' should not be initialized on declaration."
            }
        },
        create(context) {
            const MODE_ALWAYS = "always", MODE_NEVER = "never";
            const mode = context.options[0] || MODE_ALWAYS;
            const params = context.options[1] || {};
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                "VariableDeclaration:exit"(node) {
                    const kind = node.kind, declarations = node.declarations;
                    for (let i = 0; i < declarations.length; ++i) {
                        const declaration = declarations[i], id = declaration.id, initialized = isInitialized(declaration), isIgnoredForLoop = params.ignoreForLoopInit && isForLoop(node.parent);
                        let messageId = "";
                        if (mode === MODE_ALWAYS && !initialized) {
                            messageId = "initialized";
                        }
                        else if (mode === MODE_NEVER && kind !== "const" && initialized && !isIgnoredForLoop) {
                            messageId = "notInitialized";
                        }
                        if (id.type === "Identifier" && messageId) {
                            context.report({
                                node: declaration,
                                messageId,
                                data: {
                                    idName: id.name
                                }
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/jsx-quotes.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Constants
    //------------------------------------------------------------------------------
    const QUOTE_SETTINGS = {
        "prefer-double": {
            quote: "\"",
            description: "singlequote",
            convert(str) {
                return str.replace(/'/gu, "\"");
            }
        },
        "prefer-single": {
            quote: "'",
            description: "doublequote",
            convert(str) {
                return str.replace(/"/gu, "'");
            }
        }
    };
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce the consistent use of either double or single quotes in JSX attributes",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/jsx-quotes"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["prefer-single", "prefer-double"]
                }
            ],
            messages: {
                unexpected: "Unexpected usage of {{description}}."
            }
        },
        create(context) {
            const quoteOption = context.options[0] || "prefer-double", setting = QUOTE_SETTINGS[quoteOption];
            /**
             * Checks if the given string literal node uses the expected quotes
             * @param {ASTNode} node A string literal node.
             * @returns {boolean} Whether or not the string literal used the expected quotes.
             * @public
             */
            function usesExpectedQuotes(node) {
                return node.value.indexOf(setting.quote) !== -1 || astUtils.isSurroundedBy(node.raw, setting.quote);
            }
            return {
                JSXAttribute(node) {
                    const attributeValue = node.value;
                    if (attributeValue && astUtils.isStringLiteral(attributeValue) && !usesExpectedQuotes(attributeValue)) {
                        context.report({
                            node: attributeValue,
                            messageId: "unexpected",
                            data: {
                                description: setting.description
                            },
                            fix(fixer) {
                                return fixer.replaceText(attributeValue, setting.convert(attributeValue.raw));
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/key-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether a string contains a line terminator as defined in
     * http://www.ecma-international.org/ecma-262/5.1/#sec-7.3
     * @param {string} str String to test.
     * @returns {boolean} True if str contains a line terminator.
     */
    function containsLineTerminator(str) {
        return astUtils.LINEBREAK_MATCHER.test(str);
    }
    /**
     * Gets the last element of an array.
     * @param {Array} arr An array.
     * @returns {any} Last element of arr.
     */
    function last(arr) {
        return arr[arr.length - 1];
    }
    /**
     * Checks whether a node is contained on a single line.
     * @param {ASTNode} node AST Node being evaluated.
     * @returns {boolean} True if the node is a single line.
     */
    function isSingleLine(node) {
        return (node.loc.end.line === node.loc.start.line);
    }
    /**
     * Checks whether the properties on a single line.
     * @param {ASTNode[]} properties List of Property AST nodes.
     * @returns {boolean} True if all properties is on a single line.
     */
    function isSingleLineProperties(properties) {
        const [firstProp] = properties, lastProp = last(properties);
        return firstProp.loc.start.line === lastProp.loc.end.line;
    }
    /**
     * Initializes a single option property from the configuration with defaults for undefined values
     * @param {Object} toOptions Object to be initialized
     * @param {Object} fromOptions Object to be initialized from
     * @returns {Object} The object with correctly initialized options and values
     */
    function initOptionProperty(toOptions, fromOptions) {
        toOptions.mode = fromOptions.mode || "strict";
        // Set value of beforeColon
        if (typeof fromOptions.beforeColon !== "undefined") {
            toOptions.beforeColon = +fromOptions.beforeColon;
        }
        else {
            toOptions.beforeColon = 0;
        }
        // Set value of afterColon
        if (typeof fromOptions.afterColon !== "undefined") {
            toOptions.afterColon = +fromOptions.afterColon;
        }
        else {
            toOptions.afterColon = 1;
        }
        // Set align if exists
        if (typeof fromOptions.align !== "undefined") {
            if (typeof fromOptions.align === "object") {
                toOptions.align = fromOptions.align;
            }
            else { // "string"
                toOptions.align = {
                    on: fromOptions.align,
                    mode: toOptions.mode,
                    beforeColon: toOptions.beforeColon,
                    afterColon: toOptions.afterColon
                };
            }
        }
        return toOptions;
    }
    /**
     * Initializes all the option values (singleLine, multiLine and align) from the configuration with defaults for undefined values
     * @param {Object} toOptions Object to be initialized
     * @param {Object} fromOptions Object to be initialized from
     * @returns {Object} The object with correctly initialized options and values
     */
    function initOptions(toOptions, fromOptions) {
        if (typeof fromOptions.align === "object") {
            // Initialize the alignment configuration
            toOptions.align = initOptionProperty({}, fromOptions.align);
            toOptions.align.on = fromOptions.align.on || "colon";
            toOptions.align.mode = fromOptions.align.mode || "strict";
            toOptions.multiLine = initOptionProperty({}, (fromOptions.multiLine || fromOptions));
            toOptions.singleLine = initOptionProperty({}, (fromOptions.singleLine || fromOptions));
        }
        else { // string or undefined
            toOptions.multiLine = initOptionProperty({}, (fromOptions.multiLine || fromOptions));
            toOptions.singleLine = initOptionProperty({}, (fromOptions.singleLine || fromOptions));
            // If alignment options are defined in multiLine, pull them out into the general align configuration
            if (toOptions.multiLine.align) {
                toOptions.align = {
                    on: toOptions.multiLine.align.on,
                    mode: toOptions.multiLine.align.mode || toOptions.multiLine.mode,
                    beforeColon: toOptions.multiLine.align.beforeColon,
                    afterColon: toOptions.multiLine.align.afterColon
                };
            }
        }
        return toOptions;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing between keys and values in object literal properties",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/key-spacing"
            },
            fixable: "whitespace",
            schema: [{
                    anyOf: [
                        {
                            type: "object",
                            properties: {
                                align: {
                                    anyOf: [
                                        {
                                            enum: ["colon", "value"]
                                        },
                                        {
                                            type: "object",
                                            properties: {
                                                mode: {
                                                    enum: ["strict", "minimum"]
                                                },
                                                on: {
                                                    enum: ["colon", "value"]
                                                },
                                                beforeColon: {
                                                    type: "boolean"
                                                },
                                                afterColon: {
                                                    type: "boolean"
                                                }
                                            },
                                            additionalProperties: false
                                        }
                                    ]
                                },
                                mode: {
                                    enum: ["strict", "minimum"]
                                },
                                beforeColon: {
                                    type: "boolean"
                                },
                                afterColon: {
                                    type: "boolean"
                                }
                            },
                            additionalProperties: false
                        },
                        {
                            type: "object",
                            properties: {
                                singleLine: {
                                    type: "object",
                                    properties: {
                                        mode: {
                                            enum: ["strict", "minimum"]
                                        },
                                        beforeColon: {
                                            type: "boolean"
                                        },
                                        afterColon: {
                                            type: "boolean"
                                        }
                                    },
                                    additionalProperties: false
                                },
                                multiLine: {
                                    type: "object",
                                    properties: {
                                        align: {
                                            anyOf: [
                                                {
                                                    enum: ["colon", "value"]
                                                },
                                                {
                                                    type: "object",
                                                    properties: {
                                                        mode: {
                                                            enum: ["strict", "minimum"]
                                                        },
                                                        on: {
                                                            enum: ["colon", "value"]
                                                        },
                                                        beforeColon: {
                                                            type: "boolean"
                                                        },
                                                        afterColon: {
                                                            type: "boolean"
                                                        }
                                                    },
                                                    additionalProperties: false
                                                }
                                            ]
                                        },
                                        mode: {
                                            enum: ["strict", "minimum"]
                                        },
                                        beforeColon: {
                                            type: "boolean"
                                        },
                                        afterColon: {
                                            type: "boolean"
                                        }
                                    },
                                    additionalProperties: false
                                }
                            },
                            additionalProperties: false
                        },
                        {
                            type: "object",
                            properties: {
                                singleLine: {
                                    type: "object",
                                    properties: {
                                        mode: {
                                            enum: ["strict", "minimum"]
                                        },
                                        beforeColon: {
                                            type: "boolean"
                                        },
                                        afterColon: {
                                            type: "boolean"
                                        }
                                    },
                                    additionalProperties: false
                                },
                                multiLine: {
                                    type: "object",
                                    properties: {
                                        mode: {
                                            enum: ["strict", "minimum"]
                                        },
                                        beforeColon: {
                                            type: "boolean"
                                        },
                                        afterColon: {
                                            type: "boolean"
                                        }
                                    },
                                    additionalProperties: false
                                },
                                align: {
                                    type: "object",
                                    properties: {
                                        mode: {
                                            enum: ["strict", "minimum"]
                                        },
                                        on: {
                                            enum: ["colon", "value"]
                                        },
                                        beforeColon: {
                                            type: "boolean"
                                        },
                                        afterColon: {
                                            type: "boolean"
                                        }
                                    },
                                    additionalProperties: false
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }],
            messages: {
                extraKey: "Extra space after {{computed}}key '{{key}}'.",
                extraValue: "Extra space before value for {{computed}}key '{{key}}'.",
                missingKey: "Missing space after {{computed}}key '{{key}}'.",
                missingValue: "Missing space before value for {{computed}}key '{{key}}'."
            }
        },
        create(context) {
            /**
             * OPTIONS
             * "key-spacing": [2, {
             *     beforeColon: false,
             *     afterColon: true,
             *     align: "colon" // Optional, or "value"
             * }
             */
            const options = context.options[0] || {}, ruleOptions = initOptions({}, options), multiLineOptions = ruleOptions.multiLine, singleLineOptions = ruleOptions.singleLine, alignmentOptions = ruleOptions.align || null;
            const sourceCode = context.getSourceCode();
            /**
             * Checks whether a property is a member of the property group it follows.
             * @param {ASTNode} lastMember The last Property known to be in the group.
             * @param {ASTNode} candidate The next Property that might be in the group.
             * @returns {boolean} True if the candidate property is part of the group.
             */
            function continuesPropertyGroup(lastMember, candidate) {
                const groupEndLine = lastMember.loc.start.line, candidateStartLine = candidate.loc.start.line;
                if (candidateStartLine - groupEndLine <= 1) {
                    return true;
                }
                /*
                 * Check that the first comment is adjacent to the end of the group, the
                 * last comment is adjacent to the candidate property, and that successive
                 * comments are adjacent to each other.
                 */
                const leadingComments = sourceCode.getCommentsBefore(candidate);
                if (leadingComments.length &&
                    leadingComments[0].loc.start.line - groupEndLine <= 1 &&
                    candidateStartLine - last(leadingComments).loc.end.line <= 1) {
                    for (let i = 1; i < leadingComments.length; i++) {
                        if (leadingComments[i].loc.start.line - leadingComments[i - 1].loc.end.line > 1) {
                            return false;
                        }
                    }
                    return true;
                }
                return false;
            }
            /**
             * Determines if the given property is key-value property.
             * @param {ASTNode} property Property node to check.
             * @returns {boolean} Whether the property is a key-value property.
             */
            function isKeyValueProperty(property) {
                return !((property.method ||
                    property.shorthand ||
                    property.kind !== "init" || property.type !== "Property") // Could be "ExperimentalSpreadProperty" or "SpreadElement"
                );
            }
            /**
             * Starting from the given a node (a property.key node here) looks forward
             * until it finds the last token before a colon punctuator and returns it.
             * @param {ASTNode} node The node to start looking from.
             * @returns {ASTNode} The last token before a colon punctuator.
             */
            function getLastTokenBeforeColon(node) {
                const colonToken = sourceCode.getTokenAfter(node, astUtils.isColonToken);
                return sourceCode.getTokenBefore(colonToken);
            }
            /**
             * Starting from the given a node (a property.key node here) looks forward
             * until it finds the colon punctuator and returns it.
             * @param {ASTNode} node The node to start looking from.
             * @returns {ASTNode} The colon punctuator.
             */
            function getNextColon(node) {
                return sourceCode.getTokenAfter(node, astUtils.isColonToken);
            }
            /**
             * Gets an object literal property's key as the identifier name or string value.
             * @param {ASTNode} property Property node whose key to retrieve.
             * @returns {string} The property's key.
             */
            function getKey(property) {
                const key = property.key;
                if (property.computed) {
                    return sourceCode.getText().slice(key.range[0], key.range[1]);
                }
                return astUtils.getStaticPropertyName(property);
            }
            /**
             * Reports an appropriately-formatted error if spacing is incorrect on one
             * side of the colon.
             * @param {ASTNode} property Key-value pair in an object literal.
             * @param {string} side Side being verified - either "key" or "value".
             * @param {string} whitespace Actual whitespace string.
             * @param {int} expected Expected whitespace length.
             * @param {string} mode Value of the mode as "strict" or "minimum"
             * @returns {void}
             */
            function report(property, side, whitespace, expected, mode) {
                const diff = whitespace.length - expected, nextColon = getNextColon(property.key), tokenBeforeColon = sourceCode.getTokenBefore(nextColon, { includeComments: true }), tokenAfterColon = sourceCode.getTokenAfter(nextColon, { includeComments: true }), isKeySide = side === "key", isExtra = diff > 0, diffAbs = Math.abs(diff), spaces = Array(diffAbs + 1).join(" ");
                const locStart = isKeySide ? tokenBeforeColon.loc.end : nextColon.loc.start;
                const locEnd = isKeySide ? nextColon.loc.start : tokenAfterColon.loc.start;
                const missingLoc = isKeySide ? tokenBeforeColon.loc : tokenAfterColon.loc;
                const loc = isExtra ? { start: locStart, end: locEnd } : missingLoc;
                if ((diff && mode === "strict" ||
                    diff < 0 && mode === "minimum" ||
                    diff > 0 && !expected && mode === "minimum") &&
                    !(expected && containsLineTerminator(whitespace))) {
                    let fix;
                    if (isExtra) {
                        let range;
                        // Remove whitespace
                        if (isKeySide) {
                            range = [tokenBeforeColon.range[1], tokenBeforeColon.range[1] + diffAbs];
                        }
                        else {
                            range = [tokenAfterColon.range[0] - diffAbs, tokenAfterColon.range[0]];
                        }
                        fix = function (fixer) {
                            return fixer.removeRange(range);
                        };
                    }
                    else {
                        // Add whitespace
                        if (isKeySide) {
                            fix = function (fixer) {
                                return fixer.insertTextAfter(tokenBeforeColon, spaces);
                            };
                        }
                        else {
                            fix = function (fixer) {
                                return fixer.insertTextBefore(tokenAfterColon, spaces);
                            };
                        }
                    }
                    let messageId = "";
                    if (isExtra) {
                        messageId = side === "key" ? "extraKey" : "extraValue";
                    }
                    else {
                        messageId = side === "key" ? "missingKey" : "missingValue";
                    }
                    context.report({
                        node: property[side],
                        loc,
                        messageId,
                        data: {
                            computed: property.computed ? "computed " : "",
                            key: getKey(property)
                        },
                        fix
                    });
                }
            }
            /**
             * Gets the number of characters in a key, including quotes around string
             * keys and braces around computed property keys.
             * @param {ASTNode} property Property of on object literal.
             * @returns {int} Width of the key.
             */
            function getKeyWidth(property) {
                const startToken = sourceCode.getFirstToken(property);
                const endToken = getLastTokenBeforeColon(property.key);
                return endToken.range[1] - startToken.range[0];
            }
            /**
             * Gets the whitespace around the colon in an object literal property.
             * @param {ASTNode} property Property node from an object literal.
             * @returns {Object} Whitespace before and after the property's colon.
             */
            function getPropertyWhitespace(property) {
                const whitespace = /(\s*):(\s*)/u.exec(sourceCode.getText().slice(property.key.range[1], property.value.range[0]));
                if (whitespace) {
                    return {
                        beforeColon: whitespace[1],
                        afterColon: whitespace[2]
                    };
                }
                return null;
            }
            /**
             * Creates groups of properties.
             * @param  {ASTNode} node ObjectExpression node being evaluated.
             * @returns {Array.<ASTNode[]>} Groups of property AST node lists.
             */
            function createGroups(node) {
                if (node.properties.length === 1) {
                    return [node.properties];
                }
                return node.properties.reduce((groups, property) => {
                    const currentGroup = last(groups), prev = last(currentGroup);
                    if (!prev || continuesPropertyGroup(prev, property)) {
                        currentGroup.push(property);
                    }
                    else {
                        groups.push([property]);
                    }
                    return groups;
                }, [
                    []
                ]);
            }
            /**
             * Verifies correct vertical alignment of a group of properties.
             * @param {ASTNode[]} properties List of Property AST nodes.
             * @returns {void}
             */
            function verifyGroupAlignment(properties) {
                const length = properties.length, widths = properties.map(getKeyWidth), // Width of keys, including quotes
                align = alignmentOptions.on; // "value" or "colon"
                let targetWidth = Math.max(...widths), beforeColon, afterColon, mode;
                if (alignmentOptions && length > 1) { // When aligning values within a group, use the alignment configuration.
                    beforeColon = alignmentOptions.beforeColon;
                    afterColon = alignmentOptions.afterColon;
                    mode = alignmentOptions.mode;
                }
                else {
                    beforeColon = multiLineOptions.beforeColon;
                    afterColon = multiLineOptions.afterColon;
                    mode = alignmentOptions.mode;
                }
                // Conditionally include one space before or after colon
                targetWidth += (align === "colon" ? beforeColon : afterColon);
                for (let i = 0; i < length; i++) {
                    const property = properties[i];
                    const whitespace = getPropertyWhitespace(property);
                    if (whitespace) { // Object literal getters/setters lack a colon
                        const width = widths[i];
                        if (align === "value") {
                            report(property, "key", whitespace.beforeColon, beforeColon, mode);
                            report(property, "value", whitespace.afterColon, targetWidth - width, mode);
                        }
                        else { // align = "colon"
                            report(property, "key", whitespace.beforeColon, targetWidth - width, mode);
                            report(property, "value", whitespace.afterColon, afterColon, mode);
                        }
                    }
                }
            }
            /**
             * Verifies spacing of property conforms to specified options.
             * @param  {ASTNode} node Property node being evaluated.
             * @param {Object} lineOptions Configured singleLine or multiLine options
             * @returns {void}
             */
            function verifySpacing(node, lineOptions) {
                const actual = getPropertyWhitespace(node);
                if (actual) { // Object literal getters/setters lack colons
                    report(node, "key", actual.beforeColon, lineOptions.beforeColon, lineOptions.mode);
                    report(node, "value", actual.afterColon, lineOptions.afterColon, lineOptions.mode);
                }
            }
            /**
             * Verifies spacing of each property in a list.
             * @param {ASTNode[]} properties List of Property AST nodes.
             * @param {Object} lineOptions Configured singleLine or multiLine options
             * @returns {void}
             */
            function verifyListSpacing(properties, lineOptions) {
                const length = properties.length;
                for (let i = 0; i < length; i++) {
                    verifySpacing(properties[i], lineOptions);
                }
            }
            /**
             * Verifies vertical alignment, taking into account groups of properties.
             * @param  {ASTNode} node ObjectExpression node being evaluated.
             * @returns {void}
             */
            function verifyAlignment(node) {
                createGroups(node).forEach(group => {
                    const properties = group.filter(isKeyValueProperty);
                    if (properties.length > 0 && isSingleLineProperties(properties)) {
                        verifyListSpacing(properties, multiLineOptions);
                    }
                    else {
                        verifyGroupAlignment(properties);
                    }
                });
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            if (alignmentOptions) { // Verify vertical alignment
                return {
                    ObjectExpression(node) {
                        if (isSingleLine(node)) {
                            verifyListSpacing(node.properties.filter(isKeyValueProperty), singleLineOptions);
                        }
                        else {
                            verifyAlignment(node);
                        }
                    }
                };
            }
            // Obey beforeColon and afterColon in each property as configured
            return {
                Property(node) {
                    verifySpacing(node, isSingleLine(node.parent) ? singleLineOptions : multiLineOptions);
                }
            };
        }
    };
},
/* eslint/lib/rules/keyword-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */, keywords = require(429) /* ./utils/keywords */;
    //------------------------------------------------------------------------------
    // Constants
    //------------------------------------------------------------------------------
    const PREV_TOKEN = /^[)\]}>]$/u;
    const NEXT_TOKEN = /^(?:[([{<~!]|\+\+?|--?)$/u;
    const PREV_TOKEN_M = /^[)\]}>*]$/u;
    const NEXT_TOKEN_M = /^[{*]$/u;
    const TEMPLATE_OPEN_PAREN = /\$\{$/u;
    const TEMPLATE_CLOSE_PAREN = /^\}/u;
    const CHECK_TYPE = /^(?:JSXElement|RegularExpression|String|Template)$/u;
    const KEYS = keywords.concat(["as", "async", "await", "from", "get", "let", "of", "set", "yield"]);
    // check duplications.
    (function () {
        KEYS.sort();
        for (let i = 1; i < KEYS.length; ++i) {
            if (KEYS[i] === KEYS[i - 1]) {
                throw new Error(`Duplication was found in the keyword list: ${KEYS[i]}`);
            }
        }
    }());
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given token is a "Template" token ends with "${".
     * @param {Token} token A token to check.
     * @returns {boolean} `true` if the token is a "Template" token ends with "${".
     */
    function isOpenParenOfTemplate(token) {
        return token.type === "Template" && TEMPLATE_OPEN_PAREN.test(token.value);
    }
    /**
     * Checks whether or not a given token is a "Template" token starts with "}".
     * @param {Token} token A token to check.
     * @returns {boolean} `true` if the token is a "Template" token starts with "}".
     */
    function isCloseParenOfTemplate(token) {
        return token.type === "Template" && TEMPLATE_CLOSE_PAREN.test(token.value);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing before and after keywords",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/keyword-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        before: { type: "boolean", default: true },
                        after: { type: "boolean", default: true },
                        overrides: {
                            type: "object",
                            properties: KEYS.reduce((retv, key) => {
                                retv[key] = {
                                    type: "object",
                                    properties: {
                                        before: { type: "boolean" },
                                        after: { type: "boolean" }
                                    },
                                    additionalProperties: false
                                };
                                return retv;
                            }, {}),
                            additionalProperties: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                expectedBefore: "Expected space(s) before \"{{value}}\".",
                expectedAfter: "Expected space(s) after \"{{value}}\".",
                unexpectedBefore: "Unexpected space(s) before \"{{value}}\".",
                unexpectedAfter: "Unexpected space(s) after \"{{value}}\"."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Reports a given token if there are not space(s) before the token.
             * @param {Token} token A token to report.
             * @param {RegExp} pattern A pattern of the previous token to check.
             * @returns {void}
             */
            function expectSpaceBefore(token, pattern) {
                const prevToken = sourceCode.getTokenBefore(token);
                if (prevToken &&
                    (CHECK_TYPE.test(prevToken.type) || pattern.test(prevToken.value)) &&
                    !isOpenParenOfTemplate(prevToken) &&
                    astUtils.isTokenOnSameLine(prevToken, token) &&
                    !sourceCode.isSpaceBetweenTokens(prevToken, token)) {
                    context.report({
                        loc: token.loc,
                        messageId: "expectedBefore",
                        data: token,
                        fix(fixer) {
                            return fixer.insertTextBefore(token, " ");
                        }
                    });
                }
            }
            /**
             * Reports a given token if there are space(s) before the token.
             * @param {Token} token A token to report.
             * @param {RegExp} pattern A pattern of the previous token to check.
             * @returns {void}
             */
            function unexpectSpaceBefore(token, pattern) {
                const prevToken = sourceCode.getTokenBefore(token);
                if (prevToken &&
                    (CHECK_TYPE.test(prevToken.type) || pattern.test(prevToken.value)) &&
                    !isOpenParenOfTemplate(prevToken) &&
                    astUtils.isTokenOnSameLine(prevToken, token) &&
                    sourceCode.isSpaceBetweenTokens(prevToken, token)) {
                    context.report({
                        loc: { start: prevToken.loc.end, end: token.loc.start },
                        messageId: "unexpectedBefore",
                        data: token,
                        fix(fixer) {
                            return fixer.removeRange([prevToken.range[1], token.range[0]]);
                        }
                    });
                }
            }
            /**
             * Reports a given token if there are not space(s) after the token.
             * @param {Token} token A token to report.
             * @param {RegExp} pattern A pattern of the next token to check.
             * @returns {void}
             */
            function expectSpaceAfter(token, pattern) {
                const nextToken = sourceCode.getTokenAfter(token);
                if (nextToken &&
                    (CHECK_TYPE.test(nextToken.type) || pattern.test(nextToken.value)) &&
                    !isCloseParenOfTemplate(nextToken) &&
                    astUtils.isTokenOnSameLine(token, nextToken) &&
                    !sourceCode.isSpaceBetweenTokens(token, nextToken)) {
                    context.report({
                        loc: token.loc,
                        messageId: "expectedAfter",
                        data: token,
                        fix(fixer) {
                            return fixer.insertTextAfter(token, " ");
                        }
                    });
                }
            }
            /**
             * Reports a given token if there are space(s) after the token.
             * @param {Token} token A token to report.
             * @param {RegExp} pattern A pattern of the next token to check.
             * @returns {void}
             */
            function unexpectSpaceAfter(token, pattern) {
                const nextToken = sourceCode.getTokenAfter(token);
                if (nextToken &&
                    (CHECK_TYPE.test(nextToken.type) || pattern.test(nextToken.value)) &&
                    !isCloseParenOfTemplate(nextToken) &&
                    astUtils.isTokenOnSameLine(token, nextToken) &&
                    sourceCode.isSpaceBetweenTokens(token, nextToken)) {
                    context.report({
                        loc: { start: token.loc.end, end: nextToken.loc.start },
                        messageId: "unexpectedAfter",
                        data: token,
                        fix(fixer) {
                            return fixer.removeRange([token.range[1], nextToken.range[0]]);
                        }
                    });
                }
            }
            /**
             * Parses the option object and determines check methods for each keyword.
             * @param {Object|undefined} options The option object to parse.
             * @returns {Object} - Normalized option object.
             *      Keys are keywords (there are for every keyword).
             *      Values are instances of `{"before": function, "after": function}`.
             */
            function parseOptions(options = {}) {
                const before = options.before !== false;
                const after = options.after !== false;
                const defaultValue = {
                    before: before ? expectSpaceBefore : unexpectSpaceBefore,
                    after: after ? expectSpaceAfter : unexpectSpaceAfter
                };
                const overrides = (options && options.overrides) || {};
                const retv = Object.create(null);
                for (let i = 0; i < KEYS.length; ++i) {
                    const key = KEYS[i];
                    const override = overrides[key];
                    if (override) {
                        const thisBefore = ("before" in override) ? override.before : before;
                        const thisAfter = ("after" in override) ? override.after : after;
                        retv[key] = {
                            before: thisBefore ? expectSpaceBefore : unexpectSpaceBefore,
                            after: thisAfter ? expectSpaceAfter : unexpectSpaceAfter
                        };
                    }
                    else {
                        retv[key] = defaultValue;
                    }
                }
                return retv;
            }
            const checkMethodMap = parseOptions(context.options[0]);
            /**
             * Reports a given token if usage of spacing followed by the token is
             * invalid.
             * @param {Token} token A token to report.
             * @param {RegExp} [pattern] Optional. A pattern of the previous
             *      token to check.
             * @returns {void}
             */
            function checkSpacingBefore(token, pattern) {
                checkMethodMap[token.value].before(token, pattern || PREV_TOKEN);
            }
            /**
             * Reports a given token if usage of spacing preceded by the token is
             * invalid.
             * @param {Token} token A token to report.
             * @param {RegExp} [pattern] Optional. A pattern of the next
             *      token to check.
             * @returns {void}
             */
            function checkSpacingAfter(token, pattern) {
                checkMethodMap[token.value].after(token, pattern || NEXT_TOKEN);
            }
            /**
             * Reports a given token if usage of spacing around the token is invalid.
             * @param {Token} token A token to report.
             * @returns {void}
             */
            function checkSpacingAround(token) {
                checkSpacingBefore(token);
                checkSpacingAfter(token);
            }
            /**
             * Reports the first token of a given node if the first token is a keyword
             * and usage of spacing around the token is invalid.
             * @param {ASTNode|null} node A node to report.
             * @returns {void}
             */
            function checkSpacingAroundFirstToken(node) {
                const firstToken = node && sourceCode.getFirstToken(node);
                if (firstToken && firstToken.type === "Keyword") {
                    checkSpacingAround(firstToken);
                }
            }
            /**
             * Reports the first token of a given node if the first token is a keyword
             * and usage of spacing followed by the token is invalid.
             *
             * This is used for unary operators (e.g. `typeof`), `function`, and `super`.
             * Other rules are handling usage of spacing preceded by those keywords.
             * @param {ASTNode|null} node A node to report.
             * @returns {void}
             */
            function checkSpacingBeforeFirstToken(node) {
                const firstToken = node && sourceCode.getFirstToken(node);
                if (firstToken && firstToken.type === "Keyword") {
                    checkSpacingBefore(firstToken);
                }
            }
            /**
             * Reports the previous token of a given node if the token is a keyword and
             * usage of spacing around the token is invalid.
             * @param {ASTNode|null} node A node to report.
             * @returns {void}
             */
            function checkSpacingAroundTokenBefore(node) {
                if (node) {
                    const token = sourceCode.getTokenBefore(node, astUtils.isKeywordToken);
                    checkSpacingAround(token);
                }
            }
            /**
             * Reports `async` or `function` keywords of a given node if usage of
             * spacing around those keywords is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForFunction(node) {
                const firstToken = node && sourceCode.getFirstToken(node);
                if (firstToken &&
                    ((firstToken.type === "Keyword" && firstToken.value === "function") ||
                        firstToken.value === "async")) {
                    checkSpacingBefore(firstToken);
                }
            }
            /**
             * Reports `class` and `extends` keywords of a given node if usage of
             * spacing around those keywords is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForClass(node) {
                checkSpacingAroundFirstToken(node);
                checkSpacingAroundTokenBefore(node.superClass);
            }
            /**
             * Reports `if` and `else` keywords of a given node if usage of spacing
             * around those keywords is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForIfStatement(node) {
                checkSpacingAroundFirstToken(node);
                checkSpacingAroundTokenBefore(node.alternate);
            }
            /**
             * Reports `try`, `catch`, and `finally` keywords of a given node if usage
             * of spacing around those keywords is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForTryStatement(node) {
                checkSpacingAroundFirstToken(node);
                checkSpacingAroundFirstToken(node.handler);
                checkSpacingAroundTokenBefore(node.finalizer);
            }
            /**
             * Reports `do` and `while` keywords of a given node if usage of spacing
             * around those keywords is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForDoWhileStatement(node) {
                checkSpacingAroundFirstToken(node);
                checkSpacingAroundTokenBefore(node.test);
            }
            /**
             * Reports `for` and `in` keywords of a given node if usage of spacing
             * around those keywords is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForForInStatement(node) {
                checkSpacingAroundFirstToken(node);
                checkSpacingAroundTokenBefore(node.right);
            }
            /**
             * Reports `for` and `of` keywords of a given node if usage of spacing
             * around those keywords is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForForOfStatement(node) {
                if (node.await) {
                    checkSpacingBefore(sourceCode.getFirstToken(node, 0));
                    checkSpacingAfter(sourceCode.getFirstToken(node, 1));
                }
                else {
                    checkSpacingAroundFirstToken(node);
                }
                checkSpacingAround(sourceCode.getTokenBefore(node.right, astUtils.isNotOpeningParenToken));
            }
            /**
             * Reports `import`, `export`, `as`, and `from` keywords of a given node if
             * usage of spacing around those keywords is invalid.
             *
             * This rule handles the `*` token in module declarations.
             *
             *     import*as A from "./a"; /*error Expected space(s) after "import".
             *                               error Expected space(s) before "as".
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForModuleDeclaration(node) {
                const firstToken = sourceCode.getFirstToken(node);
                checkSpacingBefore(firstToken, PREV_TOKEN_M);
                checkSpacingAfter(firstToken, NEXT_TOKEN_M);
                if (node.type === "ExportDefaultDeclaration") {
                    checkSpacingAround(sourceCode.getTokenAfter(firstToken));
                }
                if (node.type === "ExportAllDeclaration" && node.exported) {
                    const asToken = sourceCode.getTokenBefore(node.exported);
                    checkSpacingBefore(asToken, PREV_TOKEN_M);
                }
                if (node.source) {
                    const fromToken = sourceCode.getTokenBefore(node.source);
                    checkSpacingBefore(fromToken, PREV_TOKEN_M);
                    checkSpacingAfter(fromToken, NEXT_TOKEN_M);
                }
            }
            /**
             * Reports `as` keyword of a given node if usage of spacing around this
             * keyword is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForImportNamespaceSpecifier(node) {
                const asToken = sourceCode.getFirstToken(node, 1);
                checkSpacingBefore(asToken, PREV_TOKEN_M);
            }
            /**
             * Reports `static`, `get`, and `set` keywords of a given node if usage of
             * spacing around those keywords is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForProperty(node) {
                if (node.static) {
                    checkSpacingAroundFirstToken(node);
                }
                if (node.kind === "get" ||
                    node.kind === "set" ||
                    ((node.method || node.type === "MethodDefinition") &&
                        node.value.async)) {
                    const token = sourceCode.getTokenBefore(node.key, tok => {
                        switch (tok.value) {
                            case "get":
                            case "set":
                            case "async":
                                return true;
                            default:
                                return false;
                        }
                    });
                    if (!token) {
                        throw new Error("Failed to find token get, set, or async beside method name");
                    }
                    checkSpacingAround(token);
                }
            }
            /**
             * Reports `await` keyword of a given node if usage of spacing before
             * this keyword is invalid.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function checkSpacingForAwaitExpression(node) {
                checkSpacingBefore(sourceCode.getFirstToken(node));
            }
            return {
                // Statements
                DebuggerStatement: checkSpacingAroundFirstToken,
                WithStatement: checkSpacingAroundFirstToken,
                // Statements - Control flow
                BreakStatement: checkSpacingAroundFirstToken,
                ContinueStatement: checkSpacingAroundFirstToken,
                ReturnStatement: checkSpacingAroundFirstToken,
                ThrowStatement: checkSpacingAroundFirstToken,
                TryStatement: checkSpacingForTryStatement,
                // Statements - Choice
                IfStatement: checkSpacingForIfStatement,
                SwitchStatement: checkSpacingAroundFirstToken,
                SwitchCase: checkSpacingAroundFirstToken,
                // Statements - Loops
                DoWhileStatement: checkSpacingForDoWhileStatement,
                ForInStatement: checkSpacingForForInStatement,
                ForOfStatement: checkSpacingForForOfStatement,
                ForStatement: checkSpacingAroundFirstToken,
                WhileStatement: checkSpacingAroundFirstToken,
                // Statements - Declarations
                ClassDeclaration: checkSpacingForClass,
                ExportNamedDeclaration: checkSpacingForModuleDeclaration,
                ExportDefaultDeclaration: checkSpacingForModuleDeclaration,
                ExportAllDeclaration: checkSpacingForModuleDeclaration,
                FunctionDeclaration: checkSpacingForFunction,
                ImportDeclaration: checkSpacingForModuleDeclaration,
                VariableDeclaration: checkSpacingAroundFirstToken,
                // Expressions
                ArrowFunctionExpression: checkSpacingForFunction,
                AwaitExpression: checkSpacingForAwaitExpression,
                ClassExpression: checkSpacingForClass,
                FunctionExpression: checkSpacingForFunction,
                NewExpression: checkSpacingBeforeFirstToken,
                Super: checkSpacingBeforeFirstToken,
                ThisExpression: checkSpacingBeforeFirstToken,
                UnaryExpression: checkSpacingBeforeFirstToken,
                YieldExpression: checkSpacingBeforeFirstToken,
                // Others
                ImportNamespaceSpecifier: checkSpacingForImportNamespaceSpecifier,
                MethodDefinition: checkSpacingForProperty,
                Property: checkSpacingForProperty
            };
        }
    };
},
/* eslint/lib/rules/line-comment-position.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce position of line comments",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/line-comment-position"
            },
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["above", "beside"]
                        },
                        {
                            type: "object",
                            properties: {
                                position: {
                                    enum: ["above", "beside"]
                                },
                                ignorePattern: {
                                    type: "string"
                                },
                                applyDefaultPatterns: {
                                    type: "boolean"
                                },
                                applyDefaultIgnorePatterns: {
                                    type: "boolean"
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                above: "Expected comment to be above code.",
                beside: "Expected comment to be beside code."
            }
        },
        create(context) {
            const options = context.options[0];
            let above, ignorePattern, applyDefaultIgnorePatterns = true;
            if (!options || typeof options === "string") {
                above = !options || options === "above";
            }
            else {
                above = !options.position || options.position === "above";
                ignorePattern = options.ignorePattern;
                if (Object.prototype.hasOwnProperty.call(options, "applyDefaultIgnorePatterns")) {
                    applyDefaultIgnorePatterns = options.applyDefaultIgnorePatterns;
                }
                else {
                    applyDefaultIgnorePatterns = options.applyDefaultPatterns !== false;
                }
            }
            const defaultIgnoreRegExp = astUtils.COMMENTS_IGNORE_PATTERN;
            const fallThroughRegExp = /^\s*falls?\s?through/u;
            const customIgnoreRegExp = new RegExp(ignorePattern, "u");
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program() {
                    const comments = sourceCode.getAllComments();
                    comments.filter(token => token.type === "Line").forEach(node => {
                        if (applyDefaultIgnorePatterns && (defaultIgnoreRegExp.test(node.value) || fallThroughRegExp.test(node.value))) {
                            return;
                        }
                        if (ignorePattern && customIgnoreRegExp.test(node.value)) {
                            return;
                        }
                        const previous = sourceCode.getTokenBefore(node, { includeComments: true });
                        const isOnSameLine = previous && previous.loc.end.line === node.loc.start.line;
                        if (above) {
                            if (isOnSameLine) {
                                context.report({
                                    node,
                                    messageId: "above"
                                });
                            }
                        }
                        else {
                            if (!isOnSameLine) {
                                context.report({
                                    node,
                                    messageId: "beside"
                                });
                            }
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/linebreak-style.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent linebreak style",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/linebreak-style"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["unix", "windows"]
                }
            ],
            messages: {
                expectedLF: "Expected linebreaks to be 'LF' but found 'CRLF'.",
                expectedCRLF: "Expected linebreaks to be 'CRLF' but found 'LF'."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Builds a fix function that replaces text at the specified range in the source text.
             * @param {int[]} range The range to replace
             * @param {string} text The text to insert.
             * @returns {Function} Fixer function
             * @private
             */
            function createFix(range, text) {
                return function (fixer) {
                    return fixer.replaceTextRange(range, text);
                };
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program: function checkForLinebreakStyle(node) {
                    const linebreakStyle = context.options[0] || "unix", expectedLF = linebreakStyle === "unix", expectedLFChars = expectedLF ? "\n" : "\r\n", source = sourceCode.getText(), pattern = astUtils.createGlobalLinebreakMatcher();
                    let match;
                    let i = 0;
                    while ((match = pattern.exec(source)) !== null) {
                        i++;
                        if (match[0] === expectedLFChars) {
                            continue;
                        }
                        const index = match.index;
                        const range = [index, index + match[0].length];
                        context.report({
                            node,
                            loc: {
                                start: {
                                    line: i,
                                    column: sourceCode.lines[i - 1].length
                                },
                                end: {
                                    line: i + 1,
                                    column: 0
                                }
                            },
                            messageId: expectedLF ? "expectedLF" : "expectedCRLF",
                            fix: createFix(range, expectedLFChars)
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/lines-around-comment.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Return an array with with any line numbers that are empty.
     * @param {Array} lines An array of each line of the file.
     * @returns {Array} An array of line numbers.
     */
    function getEmptyLineNums(lines) {
        const emptyLines = lines.map((line, i) => ({
            code: line.trim(),
            num: i + 1
        })).filter(line => !line.code).map(line => line.num);
        return emptyLines;
    }
    /**
     * Return an array with with any line numbers that contain comments.
     * @param {Array} comments An array of comment tokens.
     * @returns {Array} An array of line numbers.
     */
    function getCommentLineNums(comments) {
        const lines = [];
        comments.forEach(token => {
            const start = token.loc.start.line;
            const end = token.loc.end.line;
            lines.push(start, end);
        });
        return lines;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require empty lines around comments",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/lines-around-comment"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        beforeBlockComment: {
                            type: "boolean",
                            default: true
                        },
                        afterBlockComment: {
                            type: "boolean",
                            default: false
                        },
                        beforeLineComment: {
                            type: "boolean",
                            default: false
                        },
                        afterLineComment: {
                            type: "boolean",
                            default: false
                        },
                        allowBlockStart: {
                            type: "boolean",
                            default: false
                        },
                        allowBlockEnd: {
                            type: "boolean",
                            default: false
                        },
                        allowClassStart: {
                            type: "boolean"
                        },
                        allowClassEnd: {
                            type: "boolean"
                        },
                        allowObjectStart: {
                            type: "boolean"
                        },
                        allowObjectEnd: {
                            type: "boolean"
                        },
                        allowArrayStart: {
                            type: "boolean"
                        },
                        allowArrayEnd: {
                            type: "boolean"
                        },
                        ignorePattern: {
                            type: "string"
                        },
                        applyDefaultIgnorePatterns: {
                            type: "boolean"
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                after: "Expected line after comment.",
                before: "Expected line before comment."
            }
        },
        create(context) {
            const options = Object.assign({}, context.options[0]);
            const ignorePattern = options.ignorePattern;
            const defaultIgnoreRegExp = astUtils.COMMENTS_IGNORE_PATTERN;
            const customIgnoreRegExp = new RegExp(ignorePattern, "u");
            const applyDefaultIgnorePatterns = options.applyDefaultIgnorePatterns !== false;
            options.beforeBlockComment = typeof options.beforeBlockComment !== "undefined" ? options.beforeBlockComment : true;
            const sourceCode = context.getSourceCode();
            const lines = sourceCode.lines, numLines = lines.length + 1, comments = sourceCode.getAllComments(), commentLines = getCommentLineNums(comments), emptyLines = getEmptyLineNums(lines), commentAndEmptyLines = commentLines.concat(emptyLines);
            /**
             * Returns whether or not comments are on lines starting with or ending with code
             * @param {token} token The comment token to check.
             * @returns {boolean} True if the comment is not alone.
             */
            function codeAroundComment(token) {
                let currentToken = token;
                do {
                    currentToken = sourceCode.getTokenBefore(currentToken, { includeComments: true });
                } while (currentToken && astUtils.isCommentToken(currentToken));
                if (currentToken && astUtils.isTokenOnSameLine(currentToken, token)) {
                    return true;
                }
                currentToken = token;
                do {
                    currentToken = sourceCode.getTokenAfter(currentToken, { includeComments: true });
                } while (currentToken && astUtils.isCommentToken(currentToken));
                if (currentToken && astUtils.isTokenOnSameLine(token, currentToken)) {
                    return true;
                }
                return false;
            }
            /**
             * Returns whether or not comments are inside a node type or not.
             * @param {ASTNode} parent The Comment parent node.
             * @param {string} nodeType The parent type to check against.
             * @returns {boolean} True if the comment is inside nodeType.
             */
            function isParentNodeType(parent, nodeType) {
                return parent.type === nodeType ||
                    (parent.body && parent.body.type === nodeType) ||
                    (parent.consequent && parent.consequent.type === nodeType);
            }
            /**
             * Returns the parent node that contains the given token.
             * @param {token} token The token to check.
             * @returns {ASTNode} The parent node that contains the given token.
             */
            function getParentNodeOfToken(token) {
                return sourceCode.getNodeByRangeIndex(token.range[0]);
            }
            /**
             * Returns whether or not comments are at the parent start or not.
             * @param {token} token The Comment token.
             * @param {string} nodeType The parent type to check against.
             * @returns {boolean} True if the comment is at parent start.
             */
            function isCommentAtParentStart(token, nodeType) {
                const parent = getParentNodeOfToken(token);
                return parent && isParentNodeType(parent, nodeType) &&
                    token.loc.start.line - parent.loc.start.line === 1;
            }
            /**
             * Returns whether or not comments are at the parent end or not.
             * @param {token} token The Comment token.
             * @param {string} nodeType The parent type to check against.
             * @returns {boolean} True if the comment is at parent end.
             */
            function isCommentAtParentEnd(token, nodeType) {
                const parent = getParentNodeOfToken(token);
                return parent && isParentNodeType(parent, nodeType) &&
                    parent.loc.end.line - token.loc.end.line === 1;
            }
            /**
             * Returns whether or not comments are at the block start or not.
             * @param {token} token The Comment token.
             * @returns {boolean} True if the comment is at block start.
             */
            function isCommentAtBlockStart(token) {
                return isCommentAtParentStart(token, "ClassBody") || isCommentAtParentStart(token, "BlockStatement") || isCommentAtParentStart(token, "SwitchCase");
            }
            /**
             * Returns whether or not comments are at the block end or not.
             * @param {token} token The Comment token.
             * @returns {boolean} True if the comment is at block end.
             */
            function isCommentAtBlockEnd(token) {
                return isCommentAtParentEnd(token, "ClassBody") || isCommentAtParentEnd(token, "BlockStatement") || isCommentAtParentEnd(token, "SwitchCase") || isCommentAtParentEnd(token, "SwitchStatement");
            }
            /**
             * Returns whether or not comments are at the class start or not.
             * @param {token} token The Comment token.
             * @returns {boolean} True if the comment is at class start.
             */
            function isCommentAtClassStart(token) {
                return isCommentAtParentStart(token, "ClassBody");
            }
            /**
             * Returns whether or not comments are at the class end or not.
             * @param {token} token The Comment token.
             * @returns {boolean} True if the comment is at class end.
             */
            function isCommentAtClassEnd(token) {
                return isCommentAtParentEnd(token, "ClassBody");
            }
            /**
             * Returns whether or not comments are at the object start or not.
             * @param {token} token The Comment token.
             * @returns {boolean} True if the comment is at object start.
             */
            function isCommentAtObjectStart(token) {
                return isCommentAtParentStart(token, "ObjectExpression") || isCommentAtParentStart(token, "ObjectPattern");
            }
            /**
             * Returns whether or not comments are at the object end or not.
             * @param {token} token The Comment token.
             * @returns {boolean} True if the comment is at object end.
             */
            function isCommentAtObjectEnd(token) {
                return isCommentAtParentEnd(token, "ObjectExpression") || isCommentAtParentEnd(token, "ObjectPattern");
            }
            /**
             * Returns whether or not comments are at the array start or not.
             * @param {token} token The Comment token.
             * @returns {boolean} True if the comment is at array start.
             */
            function isCommentAtArrayStart(token) {
                return isCommentAtParentStart(token, "ArrayExpression") || isCommentAtParentStart(token, "ArrayPattern");
            }
            /**
             * Returns whether or not comments are at the array end or not.
             * @param {token} token The Comment token.
             * @returns {boolean} True if the comment is at array end.
             */
            function isCommentAtArrayEnd(token) {
                return isCommentAtParentEnd(token, "ArrayExpression") || isCommentAtParentEnd(token, "ArrayPattern");
            }
            /**
             * Checks if a comment token has lines around it (ignores inline comments)
             * @param {token} token The Comment token.
             * @param {Object} opts Options to determine the newline.
             * @param {boolean} opts.after Should have a newline after this line.
             * @param {boolean} opts.before Should have a newline before this line.
             * @returns {void}
             */
            function checkForEmptyLine(token, opts) {
                if (applyDefaultIgnorePatterns && defaultIgnoreRegExp.test(token.value)) {
                    return;
                }
                if (ignorePattern && customIgnoreRegExp.test(token.value)) {
                    return;
                }
                let after = opts.after, before = opts.before;
                const prevLineNum = token.loc.start.line - 1, nextLineNum = token.loc.end.line + 1, commentIsNotAlone = codeAroundComment(token);
                const blockStartAllowed = options.allowBlockStart &&
                    isCommentAtBlockStart(token) &&
                    !(options.allowClassStart === false &&
                        isCommentAtClassStart(token)), blockEndAllowed = options.allowBlockEnd && isCommentAtBlockEnd(token) && !(options.allowClassEnd === false && isCommentAtClassEnd(token)), classStartAllowed = options.allowClassStart && isCommentAtClassStart(token), classEndAllowed = options.allowClassEnd && isCommentAtClassEnd(token), objectStartAllowed = options.allowObjectStart && isCommentAtObjectStart(token), objectEndAllowed = options.allowObjectEnd && isCommentAtObjectEnd(token), arrayStartAllowed = options.allowArrayStart && isCommentAtArrayStart(token), arrayEndAllowed = options.allowArrayEnd && isCommentAtArrayEnd(token);
                const exceptionStartAllowed = blockStartAllowed || classStartAllowed || objectStartAllowed || arrayStartAllowed;
                const exceptionEndAllowed = blockEndAllowed || classEndAllowed || objectEndAllowed || arrayEndAllowed;
                // ignore top of the file and bottom of the file
                if (prevLineNum < 1) {
                    before = false;
                }
                if (nextLineNum >= numLines) {
                    after = false;
                }
                // we ignore all inline comments
                if (commentIsNotAlone) {
                    return;
                }
                const previousTokenOrComment = sourceCode.getTokenBefore(token, { includeComments: true });
                const nextTokenOrComment = sourceCode.getTokenAfter(token, { includeComments: true });
                // check for newline before
                if (!exceptionStartAllowed && before && !commentAndEmptyLines.includes(prevLineNum) &&
                    !(astUtils.isCommentToken(previousTokenOrComment) && astUtils.isTokenOnSameLine(previousTokenOrComment, token))) {
                    const lineStart = token.range[0] - token.loc.start.column;
                    const range = [lineStart, lineStart];
                    context.report({
                        node: token,
                        messageId: "before",
                        fix(fixer) {
                            return fixer.insertTextBeforeRange(range, "\n");
                        }
                    });
                }
                // check for newline after
                if (!exceptionEndAllowed && after && !commentAndEmptyLines.includes(nextLineNum) &&
                    !(astUtils.isCommentToken(nextTokenOrComment) && astUtils.isTokenOnSameLine(token, nextTokenOrComment))) {
                    context.report({
                        node: token,
                        messageId: "after",
                        fix(fixer) {
                            return fixer.insertTextAfter(token, "\n");
                        }
                    });
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program() {
                    comments.forEach(token => {
                        if (token.type === "Line") {
                            if (options.beforeLineComment || options.afterLineComment) {
                                checkForEmptyLine(token, {
                                    after: options.afterLineComment,
                                    before: options.beforeLineComment
                                });
                            }
                        }
                        else if (token.type === "Block") {
                            if (options.beforeBlockComment || options.afterBlockComment) {
                                checkForEmptyLine(token, {
                                    after: options.afterBlockComment,
                                    before: options.beforeBlockComment
                                });
                            }
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/lines-around-directive.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow newlines around directives",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/lines-around-directive"
            },
            schema: [{
                    oneOf: [
                        {
                            enum: ["always", "never"]
                        },
                        {
                            type: "object",
                            properties: {
                                before: {
                                    enum: ["always", "never"]
                                },
                                after: {
                                    enum: ["always", "never"]
                                }
                            },
                            additionalProperties: false,
                            minProperties: 2
                        }
                    ]
                }],
            fixable: "whitespace",
            messages: {
                expected: "Expected newline {{location}} \"{{value}}\" directive.",
                unexpected: "Unexpected newline {{location}} \"{{value}}\" directive."
            },
            deprecated: true,
            replacedBy: ["padding-line-between-statements"]
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const config = context.options[0] || "always";
            const expectLineBefore = typeof config === "string" ? config : config.before;
            const expectLineAfter = typeof config === "string" ? config : config.after;
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Check if node is preceded by a blank newline.
             * @param {ASTNode} node Node to check.
             * @returns {boolean} Whether or not the passed in node is preceded by a blank newline.
             */
            function hasNewlineBefore(node) {
                const tokenBefore = sourceCode.getTokenBefore(node, { includeComments: true });
                const tokenLineBefore = tokenBefore ? tokenBefore.loc.end.line : 0;
                return node.loc.start.line - tokenLineBefore >= 2;
            }
            /**
             * Gets the last token of a node that is on the same line as the rest of the node.
             * This will usually be the last token of the node, but it will be the second-to-last token if the node has a trailing
             * semicolon on a different line.
             * @param {ASTNode} node A directive node
             * @returns {Token} The last token of the node on the line
             */
            function getLastTokenOnLine(node) {
                const lastToken = sourceCode.getLastToken(node);
                const secondToLastToken = sourceCode.getTokenBefore(lastToken);
                return astUtils.isSemicolonToken(lastToken) && lastToken.loc.start.line > secondToLastToken.loc.end.line
                    ? secondToLastToken
                    : lastToken;
            }
            /**
             * Check if node is followed by a blank newline.
             * @param {ASTNode} node Node to check.
             * @returns {boolean} Whether or not the passed in node is followed by a blank newline.
             */
            function hasNewlineAfter(node) {
                const lastToken = getLastTokenOnLine(node);
                const tokenAfter = sourceCode.getTokenAfter(lastToken, { includeComments: true });
                return tokenAfter.loc.start.line - lastToken.loc.end.line >= 2;
            }
            /**
             * Report errors for newlines around directives.
             * @param {ASTNode} node Node to check.
             * @param {string} location Whether the error was found before or after the directive.
             * @param {boolean} expected Whether or not a newline was expected or unexpected.
             * @returns {void}
             */
            function reportError(node, location, expected) {
                context.report({
                    node,
                    messageId: expected ? "expected" : "unexpected",
                    data: {
                        value: node.expression.value,
                        location
                    },
                    fix(fixer) {
                        const lastToken = getLastTokenOnLine(node);
                        if (expected) {
                            return location === "before" ? fixer.insertTextBefore(node, "\n") : fixer.insertTextAfter(lastToken, "\n");
                        }
                        return fixer.removeRange(location === "before" ? [node.range[0] - 1, node.range[0]] : [lastToken.range[1], lastToken.range[1] + 1]);
                    }
                });
            }
            /**
             * Check lines around directives in node
             * @param {ASTNode} node node to check
             * @returns {void}
             */
            function checkDirectives(node) {
                const directives = astUtils.getDirectivePrologue(node);
                if (!directives.length) {
                    return;
                }
                const firstDirective = directives[0];
                const leadingComments = sourceCode.getCommentsBefore(firstDirective);
                /*
                 * Only check before the first directive if it is preceded by a comment or if it is at the top of
                 * the file and expectLineBefore is set to "never". This is to not force a newline at the top of
                 * the file if there are no comments as well as for compatibility with padded-blocks.
                 */
                if (leadingComments.length) {
                    if (expectLineBefore === "always" && !hasNewlineBefore(firstDirective)) {
                        reportError(firstDirective, "before", true);
                    }
                    if (expectLineBefore === "never" && hasNewlineBefore(firstDirective)) {
                        reportError(firstDirective, "before", false);
                    }
                }
                else if (node.type === "Program" &&
                    expectLineBefore === "never" &&
                    !leadingComments.length &&
                    hasNewlineBefore(firstDirective)) {
                    reportError(firstDirective, "before", false);
                }
                const lastDirective = directives[directives.length - 1];
                const statements = node.type === "Program" ? node.body : node.body.body;
                /*
                 * Do not check after the last directive if the body only
                 * contains a directive prologue and isn't followed by a comment to ensure
                 * this rule behaves well with padded-blocks.
                 */
                if (lastDirective === statements[statements.length - 1] && !lastDirective.trailingComments) {
                    return;
                }
                if (expectLineAfter === "always" && !hasNewlineAfter(lastDirective)) {
                    reportError(lastDirective, "after", true);
                }
                if (expectLineAfter === "never" && hasNewlineAfter(lastDirective)) {
                    reportError(lastDirective, "after", false);
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program: checkDirectives,
                FunctionDeclaration: checkDirectives,
                FunctionExpression: checkDirectives,
                ArrowFunctionExpression: checkDirectives
            };
        }
    };
},
/* eslint/lib/rules/lines-between-class-members.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow an empty line between class members",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/lines-between-class-members"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never"]
                },
                {
                    type: "object",
                    properties: {
                        exceptAfterSingleLine: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                never: "Unexpected blank line between class members.",
                always: "Expected blank line between class members."
            }
        },
        create(context) {
            const options = [];
            options[0] = context.options[0] || "always";
            options[1] = context.options[1] || { exceptAfterSingleLine: false };
            const sourceCode = context.getSourceCode();
            /**
             * Return the last token among the consecutive tokens that have no exceed max line difference in between, before the first token in the next member.
             * @param {Token} prevLastToken The last token in the previous member node.
             * @param {Token} nextFirstToken The first token in the next member node.
             * @param {number} maxLine The maximum number of allowed line difference between consecutive tokens.
             * @returns {Token} The last token among the consecutive tokens.
             */
            function findLastConsecutiveTokenAfter(prevLastToken, nextFirstToken, maxLine) {
                const after = sourceCode.getTokenAfter(prevLastToken, { includeComments: true });
                if (after !== nextFirstToken && after.loc.start.line - prevLastToken.loc.end.line <= maxLine) {
                    return findLastConsecutiveTokenAfter(after, nextFirstToken, maxLine);
                }
                return prevLastToken;
            }
            /**
             * Return the first token among the consecutive tokens that have no exceed max line difference in between, after the last token in the previous member.
             * @param {Token} nextFirstToken The first token in the next member node.
             * @param {Token} prevLastToken The last token in the previous member node.
             * @param {number} maxLine The maximum number of allowed line difference between consecutive tokens.
             * @returns {Token} The first token among the consecutive tokens.
             */
            function findFirstConsecutiveTokenBefore(nextFirstToken, prevLastToken, maxLine) {
                const before = sourceCode.getTokenBefore(nextFirstToken, { includeComments: true });
                if (before !== prevLastToken && nextFirstToken.loc.start.line - before.loc.end.line <= maxLine) {
                    return findFirstConsecutiveTokenBefore(before, prevLastToken, maxLine);
                }
                return nextFirstToken;
            }
            /**
             * Checks if there is a token or comment between two tokens.
             * @param {Token} before The token before.
             * @param {Token} after The token after.
             * @returns {boolean} True if there is a token or comment between two tokens.
             */
            function hasTokenOrCommentBetween(before, after) {
                return sourceCode.getTokensBetween(before, after, { includeComments: true }).length !== 0;
            }
            return {
                ClassBody(node) {
                    const body = node.body;
                    for (let i = 0; i < body.length - 1; i++) {
                        const curFirst = sourceCode.getFirstToken(body[i]);
                        const curLast = sourceCode.getLastToken(body[i]);
                        const nextFirst = sourceCode.getFirstToken(body[i + 1]);
                        const isMulti = !astUtils.isTokenOnSameLine(curFirst, curLast);
                        const skip = !isMulti && options[1].exceptAfterSingleLine;
                        const beforePadding = findLastConsecutiveTokenAfter(curLast, nextFirst, 1);
                        const afterPadding = findFirstConsecutiveTokenBefore(nextFirst, curLast, 1);
                        const isPadded = afterPadding.loc.start.line - beforePadding.loc.end.line > 1;
                        const hasTokenInPadding = hasTokenOrCommentBetween(beforePadding, afterPadding);
                        const curLineLastToken = findLastConsecutiveTokenAfter(curLast, nextFirst, 0);
                        if ((options[0] === "always" && !skip && !isPadded) ||
                            (options[0] === "never" && isPadded)) {
                            context.report({
                                node: body[i + 1],
                                messageId: isPadded ? "never" : "always",
                                fix(fixer) {
                                    if (hasTokenInPadding) {
                                        return null;
                                    }
                                    return isPadded
                                        ? fixer.replaceTextRange([beforePadding.range[1], afterPadding.range[0]], "\n")
                                        : fixer.insertTextAfter(curLineLastToken, "\n");
                                }
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/max-classes-per-file.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce a maximum number of classes per file",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/max-classes-per-file"
            },
            schema: [
                {
                    type: "integer",
                    minimum: 1
                }
            ],
            messages: {
                maximumExceeded: "File has too many classes ({{ classCount }}). Maximum allowed is {{ max }}."
            }
        },
        create(context) {
            const maxClasses = context.options[0] || 1;
            let classCount = 0;
            return {
                Program() {
                    classCount = 0;
                },
                "Program:exit"(node) {
                    if (classCount > maxClasses) {
                        context.report({
                            node,
                            messageId: "maximumExceeded",
                            data: {
                                classCount,
                                max: maxClasses
                            }
                        });
                    }
                },
                "ClassDeclaration, ClassExpression"() {
                    classCount++;
                }
            };
        }
    };
},
/* eslint/lib/rules/max-depth.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce a maximum depth that blocks can be nested",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/max-depth"
            },
            schema: [
                {
                    oneOf: [
                        {
                            type: "integer",
                            minimum: 0
                        },
                        {
                            type: "object",
                            properties: {
                                maximum: {
                                    type: "integer",
                                    minimum: 0
                                },
                                max: {
                                    type: "integer",
                                    minimum: 0
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                tooDeeply: "Blocks are nested too deeply ({{depth}}). Maximum allowed is {{maxDepth}}."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            const functionStack = [], option = context.options[0];
            let maxDepth = 4;
            if (typeof option === "object" &&
                (Object.prototype.hasOwnProperty.call(option, "maximum") || Object.prototype.hasOwnProperty.call(option, "max"))) {
                maxDepth = option.maximum || option.max;
            }
            if (typeof option === "number") {
                maxDepth = option;
            }
            /**
             * When parsing a new function, store it in our function stack
             * @returns {void}
             * @private
             */
            function startFunction() {
                functionStack.push(0);
            }
            /**
             * When parsing is done then pop out the reference
             * @returns {void}
             * @private
             */
            function endFunction() {
                functionStack.pop();
            }
            /**
             * Save the block and Evaluate the node
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function pushBlock(node) {
                const len = ++functionStack[functionStack.length - 1];
                if (len > maxDepth) {
                    context.report({ node, messageId: "tooDeeply", data: { depth: len, maxDepth } });
                }
            }
            /**
             * Pop the saved block
             * @returns {void}
             * @private
             */
            function popBlock() {
                functionStack[functionStack.length - 1]--;
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                Program: startFunction,
                FunctionDeclaration: startFunction,
                FunctionExpression: startFunction,
                ArrowFunctionExpression: startFunction,
                IfStatement(node) {
                    if (node.parent.type !== "IfStatement") {
                        pushBlock(node);
                    }
                },
                SwitchStatement: pushBlock,
                TryStatement: pushBlock,
                DoWhileStatement: pushBlock,
                WhileStatement: pushBlock,
                WithStatement: pushBlock,
                ForStatement: pushBlock,
                ForInStatement: pushBlock,
                ForOfStatement: pushBlock,
                "IfStatement:exit": popBlock,
                "SwitchStatement:exit": popBlock,
                "TryStatement:exit": popBlock,
                "DoWhileStatement:exit": popBlock,
                "WhileStatement:exit": popBlock,
                "WithStatement:exit": popBlock,
                "ForStatement:exit": popBlock,
                "ForInStatement:exit": popBlock,
                "ForOfStatement:exit": popBlock,
                "FunctionDeclaration:exit": endFunction,
                "FunctionExpression:exit": endFunction,
                "ArrowFunctionExpression:exit": endFunction,
                "Program:exit": endFunction
            };
        }
    };
},
/* eslint/lib/rules/max-len.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Constants
    //------------------------------------------------------------------------------
    const OPTIONS_SCHEMA = {
        type: "object",
        properties: {
            code: {
                type: "integer",
                minimum: 0
            },
            comments: {
                type: "integer",
                minimum: 0
            },
            tabWidth: {
                type: "integer",
                minimum: 0
            },
            ignorePattern: {
                type: "string"
            },
            ignoreComments: {
                type: "boolean"
            },
            ignoreStrings: {
                type: "boolean"
            },
            ignoreUrls: {
                type: "boolean"
            },
            ignoreTemplateLiterals: {
                type: "boolean"
            },
            ignoreRegExpLiterals: {
                type: "boolean"
            },
            ignoreTrailingComments: {
                type: "boolean"
            }
        },
        additionalProperties: false
    };
    const OPTIONS_OR_INTEGER_SCHEMA = {
        anyOf: [
            OPTIONS_SCHEMA,
            {
                type: "integer",
                minimum: 0
            }
        ]
    };
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce a maximum line length",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/max-len"
            },
            schema: [
                OPTIONS_OR_INTEGER_SCHEMA,
                OPTIONS_OR_INTEGER_SCHEMA,
                OPTIONS_SCHEMA
            ],
            messages: {
                max: "This line has a length of {{lineLength}}. Maximum allowed is {{maxLength}}.",
                maxComment: "This line has a comment length of {{lineLength}}. Maximum allowed is {{maxCommentLength}}."
            }
        },
        create(context) {
            /*
             * Inspired by http://tools.ietf.org/html/rfc3986#appendix-B, however:
             * - They're matching an entire string that we know is a URI
             * - We're matching part of a string where we think there *might* be a URL
             * - We're only concerned about URLs, as picking out any URI would cause
             *   too many false positives
             * - We don't care about matching the entire URL, any small segment is fine
             */
            const URL_REGEXP = /[^:/?#]:\/\/[^?#]/u;
            const sourceCode = context.getSourceCode();
            /**
             * Computes the length of a line that may contain tabs. The width of each
             * tab will be the number of spaces to the next tab stop.
             * @param {string} line The line.
             * @param {int} tabWidth The width of each tab stop in spaces.
             * @returns {int} The computed line length.
             * @private
             */
            function computeLineLength(line, tabWidth) {
                let extraCharacterCount = 0;
                line.replace(/\t/gu, (match, offset) => {
                    const totalOffset = offset + extraCharacterCount, previousTabStopOffset = tabWidth ? totalOffset % tabWidth : 0, spaceCount = tabWidth - previousTabStopOffset;
                    extraCharacterCount += spaceCount - 1; // -1 for the replaced tab
                });
                return Array.from(line).length + extraCharacterCount;
            }
            // The options object must be the last option specified…
            const options = Object.assign({}, context.options[context.options.length - 1]);
            // …but max code length…
            if (typeof context.options[0] === "number") {
                options.code = context.options[0];
            }
            // …and tabWidth can be optionally specified directly as integers.
            if (typeof context.options[1] === "number") {
                options.tabWidth = context.options[1];
            }
            const maxLength = typeof options.code === "number" ? options.code : 80, tabWidth = typeof options.tabWidth === "number" ? options.tabWidth : 4, ignoreComments = !!options.ignoreComments, ignoreStrings = !!options.ignoreStrings, ignoreTemplateLiterals = !!options.ignoreTemplateLiterals, ignoreRegExpLiterals = !!options.ignoreRegExpLiterals, ignoreTrailingComments = !!options.ignoreTrailingComments || !!options.ignoreComments, ignoreUrls = !!options.ignoreUrls, maxCommentLength = options.comments;
            let ignorePattern = options.ignorePattern || null;
            if (ignorePattern) {
                ignorePattern = new RegExp(ignorePattern, "u");
            }
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Tells if a given comment is trailing: it starts on the current line and
             * extends to or past the end of the current line.
             * @param {string} line The source line we want to check for a trailing comment on
             * @param {number} lineNumber The one-indexed line number for line
             * @param {ASTNode} comment The comment to inspect
             * @returns {boolean} If the comment is trailing on the given line
             */
            function isTrailingComment(line, lineNumber, comment) {
                return comment &&
                    (comment.loc.start.line === lineNumber && lineNumber <= comment.loc.end.line) &&
                    (comment.loc.end.line > lineNumber || comment.loc.end.column === line.length);
            }
            /**
             * Tells if a comment encompasses the entire line.
             * @param {string} line The source line with a trailing comment
             * @param {number} lineNumber The one-indexed line number this is on
             * @param {ASTNode} comment The comment to remove
             * @returns {boolean} If the comment covers the entire line
             */
            function isFullLineComment(line, lineNumber, comment) {
                const start = comment.loc.start, end = comment.loc.end, isFirstTokenOnLine = !line.slice(0, comment.loc.start.column).trim();
                return comment &&
                    (start.line < lineNumber || (start.line === lineNumber && isFirstTokenOnLine)) &&
                    (end.line > lineNumber || (end.line === lineNumber && end.column === line.length));
            }
            /**
             * Check if a node is a JSXEmptyExpression contained in a single line JSXExpressionContainer.
             * @param {ASTNode} node A node to check.
             * @returns {boolean} True if the node is a JSXEmptyExpression contained in a single line JSXExpressionContainer.
             */
            function isJSXEmptyExpressionInSingleLineContainer(node) {
                if (!node || !node.parent || node.type !== "JSXEmptyExpression" || node.parent.type !== "JSXExpressionContainer") {
                    return false;
                }
                const parent = node.parent;
                return parent.loc.start.line === parent.loc.end.line;
            }
            /**
             * Gets the line after the comment and any remaining trailing whitespace is
             * stripped.
             * @param {string} line The source line with a trailing comment
             * @param {ASTNode} comment The comment to remove
             * @returns {string} Line without comment and trailing whitespace
             */
            function stripTrailingComment(line, comment) {
                // loc.column is zero-indexed
                return line.slice(0, comment.loc.start.column).replace(/\s+$/u, "");
            }
            /**
             * Ensure that an array exists at [key] on `object`, and add `value` to it.
             * @param {Object} object the object to mutate
             * @param {string} key the object's key
             * @param {*} value the value to add
             * @returns {void}
             * @private
             */
            function ensureArrayAndPush(object, key, value) {
                if (!Array.isArray(object[key])) {
                    object[key] = [];
                }
                object[key].push(value);
            }
            /**
             * Retrieves an array containing all strings (" or ') in the source code.
             * @returns {ASTNode[]} An array of string nodes.
             */
            function getAllStrings() {
                return sourceCode.ast.tokens.filter(token => (token.type === "String" ||
                    (token.type === "JSXText" && sourceCode.getNodeByRangeIndex(token.range[0] - 1).type === "JSXAttribute")));
            }
            /**
             * Retrieves an array containing all template literals in the source code.
             * @returns {ASTNode[]} An array of template literal nodes.
             */
            function getAllTemplateLiterals() {
                return sourceCode.ast.tokens.filter(token => token.type === "Template");
            }
            /**
             * Retrieves an array containing all RegExp literals in the source code.
             * @returns {ASTNode[]} An array of RegExp literal nodes.
             */
            function getAllRegExpLiterals() {
                return sourceCode.ast.tokens.filter(token => token.type === "RegularExpression");
            }
            /**
             * A reducer to group an AST node by line number, both start and end.
             * @param {Object} acc the accumulator
             * @param {ASTNode} node the AST node in question
             * @returns {Object} the modified accumulator
             * @private
             */
            function groupByLineNumber(acc, node) {
                for (let i = node.loc.start.line; i <= node.loc.end.line; ++i) {
                    ensureArrayAndPush(acc, i, node);
                }
                return acc;
            }
            /**
             * Returns an array of all comments in the source code.
             * If the element in the array is a JSXEmptyExpression contained with a single line JSXExpressionContainer,
             * the element is changed with JSXExpressionContainer node.
             * @returns {ASTNode[]} An array of comment nodes
             */
            function getAllComments() {
                const comments = [];
                sourceCode.getAllComments()
                    .forEach(commentNode => {
                    const containingNode = sourceCode.getNodeByRangeIndex(commentNode.range[0]);
                    if (isJSXEmptyExpressionInSingleLineContainer(containingNode)) {
                        // push a unique node only
                        if (comments[comments.length - 1] !== containingNode.parent) {
                            comments.push(containingNode.parent);
                        }
                    }
                    else {
                        comments.push(commentNode);
                    }
                });
                return comments;
            }
            /**
             * Check the program for max length
             * @param {ASTNode} node Node to examine
             * @returns {void}
             * @private
             */
            function checkProgramForMaxLength(node) {
                // split (honors line-ending)
                const lines = sourceCode.lines, 
                // list of comments to ignore
                comments = ignoreComments || maxCommentLength || ignoreTrailingComments ? getAllComments() : [];
                // we iterate over comments in parallel with the lines
                let commentsIndex = 0;
                const strings = getAllStrings();
                const stringsByLine = strings.reduce(groupByLineNumber, {});
                const templateLiterals = getAllTemplateLiterals();
                const templateLiteralsByLine = templateLiterals.reduce(groupByLineNumber, {});
                const regExpLiterals = getAllRegExpLiterals();
                const regExpLiteralsByLine = regExpLiterals.reduce(groupByLineNumber, {});
                lines.forEach((line, i) => {
                    // i is zero-indexed, line numbers are one-indexed
                    const lineNumber = i + 1;
                    /*
                     * if we're checking comment length; we need to know whether this
                     * line is a comment
                     */
                    let lineIsComment = false;
                    let textToMeasure;
                    /*
                     * We can short-circuit the comment checks if we're already out of
                     * comments to check.
                     */
                    if (commentsIndex < comments.length) {
                        let comment = null;
                        // iterate over comments until we find one past the current line
                        do {
                            comment = comments[++commentsIndex];
                        } while (comment && comment.loc.start.line <= lineNumber);
                        // and step back by one
                        comment = comments[--commentsIndex];
                        if (isFullLineComment(line, lineNumber, comment)) {
                            lineIsComment = true;
                            textToMeasure = line;
                        }
                        else if (ignoreTrailingComments && isTrailingComment(line, lineNumber, comment)) {
                            textToMeasure = stripTrailingComment(line, comment);
                            // ignore multiple trailing comments in the same line
                            let lastIndex = commentsIndex;
                            while (isTrailingComment(textToMeasure, lineNumber, comments[--lastIndex])) {
                                textToMeasure = stripTrailingComment(textToMeasure, comments[lastIndex]);
                            }
                        }
                        else {
                            textToMeasure = line;
                        }
                    }
                    else {
                        textToMeasure = line;
                    }
                    if (ignorePattern && ignorePattern.test(textToMeasure) ||
                        ignoreUrls && URL_REGEXP.test(textToMeasure) ||
                        ignoreStrings && stringsByLine[lineNumber] ||
                        ignoreTemplateLiterals && templateLiteralsByLine[lineNumber] ||
                        ignoreRegExpLiterals && regExpLiteralsByLine[lineNumber]) {
                        // ignore this line
                        return;
                    }
                    const lineLength = computeLineLength(textToMeasure, tabWidth);
                    const commentLengthApplies = lineIsComment && maxCommentLength;
                    if (lineIsComment && ignoreComments) {
                        return;
                    }
                    const loc = {
                        start: {
                            line: lineNumber,
                            column: 0
                        },
                        end: {
                            line: lineNumber,
                            column: textToMeasure.length
                        }
                    };
                    if (commentLengthApplies) {
                        if (lineLength > maxCommentLength) {
                            context.report({
                                node,
                                loc,
                                messageId: "maxComment",
                                data: {
                                    lineLength,
                                    maxCommentLength
                                }
                            });
                        }
                    }
                    else if (lineLength > maxLength) {
                        context.report({
                            node,
                            loc,
                            messageId: "max",
                            data: {
                                lineLength,
                                maxLength
                            }
                        });
                    }
                });
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                Program: checkProgramForMaxLength
            };
        }
    };
},
/* eslint/lib/rules/max-lines.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Creates an array of numbers from `start` up to, but not including, `end`
     * @param {number} start The start of the range
     * @param {number} end The end of the range
     * @returns {number[]} The range of numbers
     */
    function range(start, end) {
        return [...Array(end - start).keys()].map(x => x + start);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce a maximum number of lines per file",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/max-lines"
            },
            schema: [
                {
                    oneOf: [
                        {
                            type: "integer",
                            minimum: 0
                        },
                        {
                            type: "object",
                            properties: {
                                max: {
                                    type: "integer",
                                    minimum: 0
                                },
                                skipComments: {
                                    type: "boolean"
                                },
                                skipBlankLines: {
                                    type: "boolean"
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                exceed: "File has too many lines ({{actual}}). Maximum allowed is {{max}}."
            }
        },
        create(context) {
            const option = context.options[0];
            let max = 300;
            if (typeof option === "object" &&
                Object.prototype.hasOwnProperty.call(option, "max")) {
                max = option.max;
            }
            else if (typeof option === "number") {
                max = option;
            }
            const skipComments = option && option.skipComments;
            const skipBlankLines = option && option.skipBlankLines;
            const sourceCode = context.getSourceCode();
            /**
             * Returns whether or not a token is a comment node type
             * @param {Token} token The token to check
             * @returns {boolean} True if the token is a comment node
             */
            function isCommentNodeType(token) {
                return token && (token.type === "Block" || token.type === "Line");
            }
            /**
             * Returns the line numbers of a comment that don't have any code on the same line
             * @param {Node} comment The comment node to check
             * @returns {number[]} The line numbers
             */
            function getLinesWithoutCode(comment) {
                let start = comment.loc.start.line;
                let end = comment.loc.end.line;
                let token;
                token = comment;
                do {
                    token = sourceCode.getTokenBefore(token, {
                        includeComments: true
                    });
                } while (isCommentNodeType(token));
                if (token && astUtils.isTokenOnSameLine(token, comment)) {
                    start += 1;
                }
                token = comment;
                do {
                    token = sourceCode.getTokenAfter(token, {
                        includeComments: true
                    });
                } while (isCommentNodeType(token));
                if (token && astUtils.isTokenOnSameLine(comment, token)) {
                    end -= 1;
                }
                if (start <= end) {
                    return range(start, end + 1);
                }
                return [];
            }
            /**
             * Returns a new array formed by applying a given callback function to each element of the array, and then flattening the result by one level.
             * TODO(stephenwade): Replace this with array.flatMap when we drop support for Node v10
             * @param {any[]} array The array to process
             * @param {Function} fn The function to use
             * @returns {any[]} The result array
             */
            function flatMap(array, fn) {
                const mapped = array.map(fn);
                const flattened = [].concat(...mapped);
                return flattened;
            }
            return {
                "Program:exit"() {
                    let lines = sourceCode.lines.map((text, i) => ({
                        lineNumber: i + 1,
                        text
                    }));
                    /*
                     * If file ends with a linebreak, `sourceCode.lines` will have one extra empty line at the end.
                     * That isn't a real line, so we shouldn't count it.
                     */
                    if (lines.length > 1 && lines[lines.length - 1].text === "") {
                        lines.pop();
                    }
                    if (skipBlankLines) {
                        lines = lines.filter(l => l.text.trim() !== "");
                    }
                    if (skipComments) {
                        const comments = sourceCode.getAllComments();
                        const commentLines = flatMap(comments, comment => getLinesWithoutCode(comment));
                        lines = lines.filter(l => !commentLines.includes(l.lineNumber));
                    }
                    if (lines.length > max) {
                        const loc = {
                            start: {
                                line: lines[max].lineNumber,
                                column: 0
                            },
                            end: {
                                line: sourceCode.lines.length,
                                column: sourceCode.lines[sourceCode.lines.length - 1].length
                            }
                        };
                        context.report({
                            loc,
                            messageId: "exceed",
                            data: {
                                max,
                                actual: lines.length
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/max-lines-per-function.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const { upperCaseFirst } = require(418) /* ../shared/string-utils */;
    //------------------------------------------------------------------------------
    // Constants
    //------------------------------------------------------------------------------
    const OPTIONS_SCHEMA = {
        type: "object",
        properties: {
            max: {
                type: "integer",
                minimum: 0
            },
            skipComments: {
                type: "boolean"
            },
            skipBlankLines: {
                type: "boolean"
            },
            IIFEs: {
                type: "boolean"
            }
        },
        additionalProperties: false
    };
    const OPTIONS_OR_INTEGER_SCHEMA = {
        oneOf: [
            OPTIONS_SCHEMA,
            {
                type: "integer",
                minimum: 1
            }
        ]
    };
    /**
     * Given a list of comment nodes, return a map with numeric keys (source code line numbers) and comment token values.
     * @param {Array} comments An array of comment nodes.
     * @returns {Map.<string,Node>} A map with numeric keys (source code line numbers) and comment token values.
     */
    function getCommentLineNumbers(comments) {
        const map = new Map();
        comments.forEach(comment => {
            for (let i = comment.loc.start.line; i <= comment.loc.end.line; i++) {
                map.set(i, comment);
            }
        });
        return map;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce a maximum number of lines of code in a function",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/max-lines-per-function"
            },
            schema: [
                OPTIONS_OR_INTEGER_SCHEMA
            ],
            messages: {
                exceed: "{{name}} has too many lines ({{lineCount}}). Maximum allowed is {{maxLines}}."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const lines = sourceCode.lines;
            const option = context.options[0];
            let maxLines = 50;
            let skipComments = false;
            let skipBlankLines = false;
            let IIFEs = false;
            if (typeof option === "object") {
                maxLines = typeof option.max === "number" ? option.max : 50;
                skipComments = !!option.skipComments;
                skipBlankLines = !!option.skipBlankLines;
                IIFEs = !!option.IIFEs;
            }
            else if (typeof option === "number") {
                maxLines = option;
            }
            const commentLineNumbers = getCommentLineNumbers(sourceCode.getAllComments());
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Tells if a comment encompasses the entire line.
             * @param {string} line The source line with a trailing comment
             * @param {number} lineNumber The one-indexed line number this is on
             * @param {ASTNode} comment The comment to remove
             * @returns {boolean} If the comment covers the entire line
             */
            function isFullLineComment(line, lineNumber, comment) {
                const start = comment.loc.start, end = comment.loc.end, isFirstTokenOnLine = start.line === lineNumber && !line.slice(0, start.column).trim(), isLastTokenOnLine = end.line === lineNumber && !line.slice(end.column).trim();
                return comment &&
                    (start.line < lineNumber || isFirstTokenOnLine) &&
                    (end.line > lineNumber || isLastTokenOnLine);
            }
            /**
             * Identifies is a node is a FunctionExpression which is part of an IIFE
             * @param {ASTNode} node Node to test
             * @returns {boolean} True if it's an IIFE
             */
            function isIIFE(node) {
                return (node.type === "FunctionExpression" || node.type === "ArrowFunctionExpression") && node.parent && node.parent.type === "CallExpression" && node.parent.callee === node;
            }
            /**
             * Identifies is a node is a FunctionExpression which is embedded within a MethodDefinition or Property
             * @param {ASTNode} node Node to test
             * @returns {boolean} True if it's a FunctionExpression embedded within a MethodDefinition or Property
             */
            function isEmbedded(node) {
                if (!node.parent) {
                    return false;
                }
                if (node !== node.parent.value) {
                    return false;
                }
                if (node.parent.type === "MethodDefinition") {
                    return true;
                }
                if (node.parent.type === "Property") {
                    return node.parent.method === true || node.parent.kind === "get" || node.parent.kind === "set";
                }
                return false;
            }
            /**
             * Count the lines in the function
             * @param {ASTNode} funcNode Function AST node
             * @returns {void}
             * @private
             */
            function processFunction(funcNode) {
                const node = isEmbedded(funcNode) ? funcNode.parent : funcNode;
                if (!IIFEs && isIIFE(node)) {
                    return;
                }
                let lineCount = 0;
                for (let i = node.loc.start.line - 1; i < node.loc.end.line; ++i) {
                    const line = lines[i];
                    if (skipComments) {
                        if (commentLineNumbers.has(i + 1) && isFullLineComment(line, i + 1, commentLineNumbers.get(i + 1))) {
                            continue;
                        }
                    }
                    if (skipBlankLines) {
                        if (line.match(/^\s*$/u)) {
                            continue;
                        }
                    }
                    lineCount++;
                }
                if (lineCount > maxLines) {
                    const name = upperCaseFirst(astUtils.getFunctionNameWithKind(funcNode));
                    context.report({
                        node,
                        messageId: "exceed",
                        data: { name, lineCount, maxLines }
                    });
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                FunctionDeclaration: processFunction,
                FunctionExpression: processFunction,
                ArrowFunctionExpression: processFunction
            };
        }
    };
},
/* eslint/lib/rules/max-nested-callbacks.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce a maximum depth that callbacks can be nested",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/max-nested-callbacks"
            },
            schema: [
                {
                    oneOf: [
                        {
                            type: "integer",
                            minimum: 0
                        },
                        {
                            type: "object",
                            properties: {
                                maximum: {
                                    type: "integer",
                                    minimum: 0
                                },
                                max: {
                                    type: "integer",
                                    minimum: 0
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                exceed: "Too many nested callbacks ({{num}}). Maximum allowed is {{max}}."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Constants
            //--------------------------------------------------------------------------
            const option = context.options[0];
            let THRESHOLD = 10;
            if (typeof option === "object" &&
                (Object.prototype.hasOwnProperty.call(option, "maximum") || Object.prototype.hasOwnProperty.call(option, "max"))) {
                THRESHOLD = option.maximum || option.max;
            }
            else if (typeof option === "number") {
                THRESHOLD = option;
            }
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            const callbackStack = [];
            /**
             * Checks a given function node for too many callbacks.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             * @private
             */
            function checkFunction(node) {
                const parent = node.parent;
                if (parent.type === "CallExpression") {
                    callbackStack.push(node);
                }
                if (callbackStack.length > THRESHOLD) {
                    const opts = { num: callbackStack.length, max: THRESHOLD };
                    context.report({ node, messageId: "exceed", data: opts });
                }
            }
            /**
             * Pops the call stack.
             * @returns {void}
             * @private
             */
            function popStack() {
                callbackStack.pop();
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                ArrowFunctionExpression: checkFunction,
                "ArrowFunctionExpression:exit": popStack,
                FunctionExpression: checkFunction,
                "FunctionExpression:exit": popStack
            };
        }
    };
},
/* eslint/lib/rules/max-params.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const { upperCaseFirst } = require(418) /* ../shared/string-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce a maximum number of parameters in function definitions",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/max-params"
            },
            schema: [
                {
                    oneOf: [
                        {
                            type: "integer",
                            minimum: 0
                        },
                        {
                            type: "object",
                            properties: {
                                maximum: {
                                    type: "integer",
                                    minimum: 0
                                },
                                max: {
                                    type: "integer",
                                    minimum: 0
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                exceed: "{{name}} has too many parameters ({{count}}). Maximum allowed is {{max}}."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const option = context.options[0];
            let numParams = 3;
            if (typeof option === "object" &&
                (Object.prototype.hasOwnProperty.call(option, "maximum") || Object.prototype.hasOwnProperty.call(option, "max"))) {
                numParams = option.maximum || option.max;
            }
            if (typeof option === "number") {
                numParams = option;
            }
            /**
             * Checks a function to see if it has too many parameters.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             * @private
             */
            function checkFunction(node) {
                if (node.params.length > numParams) {
                    context.report({
                        loc: astUtils.getFunctionHeadLoc(node, sourceCode),
                        node,
                        messageId: "exceed",
                        data: {
                            name: upperCaseFirst(astUtils.getFunctionNameWithKind(node)),
                            count: node.params.length,
                            max: numParams
                        }
                    });
                }
            }
            return {
                FunctionDeclaration: checkFunction,
                ArrowFunctionExpression: checkFunction,
                FunctionExpression: checkFunction
            };
        }
    };
},
/* eslint/lib/rules/max-statements.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const { upperCaseFirst } = require(418) /* ../shared/string-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce a maximum number of statements allowed in function blocks",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/max-statements"
            },
            schema: [
                {
                    oneOf: [
                        {
                            type: "integer",
                            minimum: 0
                        },
                        {
                            type: "object",
                            properties: {
                                maximum: {
                                    type: "integer",
                                    minimum: 0
                                },
                                max: {
                                    type: "integer",
                                    minimum: 0
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                },
                {
                    type: "object",
                    properties: {
                        ignoreTopLevelFunctions: {
                            type: "boolean"
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                exceed: "{{name}} has too many statements ({{count}}). Maximum allowed is {{max}}."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            const functionStack = [], option = context.options[0], ignoreTopLevelFunctions = context.options[1] && context.options[1].ignoreTopLevelFunctions || false, topLevelFunctions = [];
            let maxStatements = 10;
            if (typeof option === "object" &&
                (Object.prototype.hasOwnProperty.call(option, "maximum") || Object.prototype.hasOwnProperty.call(option, "max"))) {
                maxStatements = option.maximum || option.max;
            }
            else if (typeof option === "number") {
                maxStatements = option;
            }
            /**
             * Reports a node if it has too many statements
             * @param {ASTNode} node node to evaluate
             * @param {int} count Number of statements in node
             * @param {int} max Maximum number of statements allowed
             * @returns {void}
             * @private
             */
            function reportIfTooManyStatements(node, count, max) {
                if (count > max) {
                    const name = upperCaseFirst(astUtils.getFunctionNameWithKind(node));
                    context.report({
                        node,
                        messageId: "exceed",
                        data: { name, count, max }
                    });
                }
            }
            /**
             * When parsing a new function, store it in our function stack
             * @returns {void}
             * @private
             */
            function startFunction() {
                functionStack.push(0);
            }
            /**
             * Evaluate the node at the end of function
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function endFunction(node) {
                const count = functionStack.pop();
                if (ignoreTopLevelFunctions && functionStack.length === 0) {
                    topLevelFunctions.push({ node, count });
                }
                else {
                    reportIfTooManyStatements(node, count, maxStatements);
                }
            }
            /**
             * Increment the count of the functions
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function countStatements(node) {
                functionStack[functionStack.length - 1] += node.body.length;
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                FunctionDeclaration: startFunction,
                FunctionExpression: startFunction,
                ArrowFunctionExpression: startFunction,
                BlockStatement: countStatements,
                "FunctionDeclaration:exit": endFunction,
                "FunctionExpression:exit": endFunction,
                "ArrowFunctionExpression:exit": endFunction,
                "Program:exit"() {
                    if (topLevelFunctions.length === 1) {
                        return;
                    }
                    topLevelFunctions.forEach(element => {
                        const count = element.count;
                        const node = element.node;
                        reportIfTooManyStatements(node, count, maxStatements);
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/max-statements-per-line.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce a maximum number of statements allowed per line",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/max-statements-per-line"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        max: {
                            type: "integer",
                            minimum: 1,
                            default: 1
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                exceed: "This line has {{numberOfStatementsOnThisLine}} {{statements}}. Maximum allowed is {{maxStatementsPerLine}}."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode(), options = context.options[0] || {}, maxStatementsPerLine = typeof options.max !== "undefined" ? options.max : 1;
            let lastStatementLine = 0, numberOfStatementsOnThisLine = 0, firstExtraStatement;
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            const SINGLE_CHILD_ALLOWED = /^(?:(?:DoWhile|For|ForIn|ForOf|If|Labeled|While)Statement|Export(?:Default|Named)Declaration)$/u;
            /**
             * Reports with the first extra statement, and clears it.
             * @returns {void}
             */
            function reportFirstExtraStatementAndClear() {
                if (firstExtraStatement) {
                    context.report({
                        node: firstExtraStatement,
                        messageId: "exceed",
                        data: {
                            numberOfStatementsOnThisLine,
                            maxStatementsPerLine,
                            statements: numberOfStatementsOnThisLine === 1 ? "statement" : "statements"
                        }
                    });
                }
                firstExtraStatement = null;
            }
            /**
             * Gets the actual last token of a given node.
             * @param {ASTNode} node A node to get. This is a node except EmptyStatement.
             * @returns {Token} The actual last token.
             */
            function getActualLastToken(node) {
                return sourceCode.getLastToken(node, astUtils.isNotSemicolonToken);
            }
            /**
             * Addresses a given node.
             * It updates the state of this rule, then reports the node if the node violated this rule.
             * @param {ASTNode} node A node to check.
             * @returns {void}
             */
            function enterStatement(node) {
                const line = node.loc.start.line;
                /*
                 * Skip to allow non-block statements if this is direct child of control statements.
                 * `if (a) foo();` is counted as 1.
                 * But `if (a) foo(); else foo();` should be counted as 2.
                 */
                if (SINGLE_CHILD_ALLOWED.test(node.parent.type) &&
                    node.parent.alternate !== node) {
                    return;
                }
                // Update state.
                if (line === lastStatementLine) {
                    numberOfStatementsOnThisLine += 1;
                }
                else {
                    reportFirstExtraStatementAndClear();
                    numberOfStatementsOnThisLine = 1;
                    lastStatementLine = line;
                }
                // Reports if the node violated this rule.
                if (numberOfStatementsOnThisLine === maxStatementsPerLine + 1) {
                    firstExtraStatement = firstExtraStatement || node;
                }
            }
            /**
             * Updates the state of this rule with the end line of leaving node to check with the next statement.
             * @param {ASTNode} node A node to check.
             * @returns {void}
             */
            function leaveStatement(node) {
                const line = getActualLastToken(node).loc.end.line;
                // Update state.
                if (line !== lastStatementLine) {
                    reportFirstExtraStatementAndClear();
                    numberOfStatementsOnThisLine = 1;
                    lastStatementLine = line;
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                BreakStatement: enterStatement,
                ClassDeclaration: enterStatement,
                ContinueStatement: enterStatement,
                DebuggerStatement: enterStatement,
                DoWhileStatement: enterStatement,
                ExpressionStatement: enterStatement,
                ForInStatement: enterStatement,
                ForOfStatement: enterStatement,
                ForStatement: enterStatement,
                FunctionDeclaration: enterStatement,
                IfStatement: enterStatement,
                ImportDeclaration: enterStatement,
                LabeledStatement: enterStatement,
                ReturnStatement: enterStatement,
                SwitchStatement: enterStatement,
                ThrowStatement: enterStatement,
                TryStatement: enterStatement,
                VariableDeclaration: enterStatement,
                WhileStatement: enterStatement,
                WithStatement: enterStatement,
                ExportNamedDeclaration: enterStatement,
                ExportDefaultDeclaration: enterStatement,
                ExportAllDeclaration: enterStatement,
                "BreakStatement:exit": leaveStatement,
                "ClassDeclaration:exit": leaveStatement,
                "ContinueStatement:exit": leaveStatement,
                "DebuggerStatement:exit": leaveStatement,
                "DoWhileStatement:exit": leaveStatement,
                "ExpressionStatement:exit": leaveStatement,
                "ForInStatement:exit": leaveStatement,
                "ForOfStatement:exit": leaveStatement,
                "ForStatement:exit": leaveStatement,
                "FunctionDeclaration:exit": leaveStatement,
                "IfStatement:exit": leaveStatement,
                "ImportDeclaration:exit": leaveStatement,
                "LabeledStatement:exit": leaveStatement,
                "ReturnStatement:exit": leaveStatement,
                "SwitchStatement:exit": leaveStatement,
                "ThrowStatement:exit": leaveStatement,
                "TryStatement:exit": leaveStatement,
                "VariableDeclaration:exit": leaveStatement,
                "WhileStatement:exit": leaveStatement,
                "WithStatement:exit": leaveStatement,
                "ExportNamedDeclaration:exit": leaveStatement,
                "ExportDefaultDeclaration:exit": leaveStatement,
                "ExportAllDeclaration:exit": leaveStatement,
                "Program:exit": reportFirstExtraStatementAndClear
            };
        }
    };
},
/* eslint/lib/rules/multiline-comment-style.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce a particular style for multiline comments",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/multiline-comment-style"
            },
            fixable: "whitespace",
            schema: [{ enum: ["starred-block", "separate-lines", "bare-block"] }],
            messages: {
                expectedBlock: "Expected a block comment instead of consecutive line comments.",
                expectedBareBlock: "Expected a block comment without padding stars.",
                startNewline: "Expected a linebreak after '/*'.",
                endNewline: "Expected a linebreak before '*/'.",
                missingStar: "Expected a '*' at the start of this line.",
                alignment: "Expected this line to be aligned with the start of the comment.",
                expectedLines: "Expected multiple line comments instead of a block comment."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const option = context.options[0] || "starred-block";
            //----------------------------------------------------------------------
            // Helpers
            //----------------------------------------------------------------------
            /**
             * Checks if a comment line is starred.
             * @param {string} line A string representing a comment line.
             * @returns {boolean} Whether or not the comment line is starred.
             */
            function isStarredCommentLine(line) {
                return /^\s*\*/u.test(line);
            }
            /**
             * Checks if a comment group is in starred-block form.
             * @param {Token[]} commentGroup A group of comments, containing either multiple line comments or a single block comment.
             * @returns {boolean} Whether or not the comment group is in starred block form.
             */
            function isStarredBlockComment([firstComment]) {
                if (firstComment.type !== "Block") {
                    return false;
                }
                const lines = firstComment.value.split(astUtils.LINEBREAK_MATCHER);
                // The first and last lines can only contain whitespace.
                return lines.length > 0 && lines.every((line, i) => (i === 0 || i === lines.length - 1 ? /^\s*$/u : /^\s*\*/u).test(line));
            }
            /**
             * Checks if a comment group is in JSDoc form.
             * @param {Token[]} commentGroup A group of comments, containing either multiple line comments or a single block comment.
             * @returns {boolean} Whether or not the comment group is in JSDoc form.
             */
            function isJSDocComment([firstComment]) {
                if (firstComment.type !== "Block") {
                    return false;
                }
                const lines = firstComment.value.split(astUtils.LINEBREAK_MATCHER);
                return /^\*\s*$/u.test(lines[0]) &&
                    lines.slice(1, -1).every(line => /^\s* /u.test(line)) &&
                    /^\s*$/u.test(lines[lines.length - 1]);
            }
            /**
             * Processes a comment group that is currently in separate-line form, calculating the offset for each line.
             * @param {Token[]} commentGroup A group of comments containing multiple line comments.
             * @returns {string[]} An array of the processed lines.
             */
            function processSeparateLineComments(commentGroup) {
                const allLinesHaveLeadingSpace = commentGroup
                    .map(({ value }) => value)
                    .filter(line => line.trim().length)
                    .every(line => line.startsWith(" "));
                return commentGroup.map(({ value }) => (allLinesHaveLeadingSpace ? value.replace(/^ /u, "") : value));
            }
            /**
             * Processes a comment group that is currently in starred-block form, calculating the offset for each line.
             * @param {Token} comment A single block comment token in starred-block form.
             * @returns {string[]} An array of the processed lines.
             */
            function processStarredBlockComment(comment) {
                const lines = comment.value.split(astUtils.LINEBREAK_MATCHER)
                    .filter((line, i, linesArr) => !(i === 0 || i === linesArr.length - 1))
                    .map(line => line.replace(/^\s*$/u, ""));
                const allLinesHaveLeadingSpace = lines
                    .map(line => line.replace(/\s*\*/u, ""))
                    .filter(line => line.trim().length)
                    .every(line => line.startsWith(" "));
                return lines.map(line => line.replace(allLinesHaveLeadingSpace ? /\s*\* ?/u : /\s*\*/u, ""));
            }
            /**
             * Processes a comment group that is currently in bare-block form, calculating the offset for each line.
             * @param {Token} comment A single block comment token in bare-block form.
             * @returns {string[]} An array of the processed lines.
             */
            function processBareBlockComment(comment) {
                const lines = comment.value.split(astUtils.LINEBREAK_MATCHER).map(line => line.replace(/^\s*$/u, ""));
                const leadingWhitespace = `${sourceCode.text.slice(comment.range[0] - comment.loc.start.column, comment.range[0])}   `;
                let offset = "";
                /*
                 * Calculate the offset of the least indented line and use that as the basis for offsetting all the lines.
                 * The first line should not be checked because it is inline with the opening block comment delimiter.
                 */
                for (const [i, line] of lines.entries()) {
                    if (!line.trim().length || i === 0) {
                        continue;
                    }
                    const [, lineOffset] = line.match(/^(\s*\*?\s*)/u);
                    if (lineOffset.length < leadingWhitespace.length) {
                        const newOffset = leadingWhitespace.slice(lineOffset.length - leadingWhitespace.length);
                        if (newOffset.length > offset.length) {
                            offset = newOffset;
                        }
                    }
                }
                return lines.map(line => {
                    const match = line.match(/^(\s*\*?\s*)(.*)/u);
                    const [, lineOffset, lineContents] = match;
                    if (lineOffset.length > leadingWhitespace.length) {
                        return `${lineOffset.slice(leadingWhitespace.length - (offset.length + lineOffset.length))}${lineContents}`;
                    }
                    if (lineOffset.length < leadingWhitespace.length) {
                        return `${lineOffset.slice(leadingWhitespace.length)}${lineContents}`;
                    }
                    return lineContents;
                });
            }
            /**
             * Gets a list of comment lines in a group, formatting leading whitespace as necessary.
             * @param {Token[]} commentGroup A group of comments containing either multiple line comments or a single block comment.
             * @returns {string[]} A list of comment lines.
             */
            function getCommentLines(commentGroup) {
                const [firstComment] = commentGroup;
                if (firstComment.type === "Line") {
                    return processSeparateLineComments(commentGroup);
                }
                if (isStarredBlockComment(commentGroup)) {
                    return processStarredBlockComment(firstComment);
                }
                return processBareBlockComment(firstComment);
            }
            /**
             * Gets the initial offset (whitespace) from the beginning of a line to a given comment token.
             * @param {Token} comment The token to check.
             * @returns {string} The offset from the beginning of a line to the token.
             */
            function getInitialOffset(comment) {
                return sourceCode.text.slice(comment.range[0] - comment.loc.start.column, comment.range[0]);
            }
            /**
             * Converts a comment into starred-block form
             * @param {Token} firstComment The first comment of the group being converted
             * @param {string[]} commentLinesList A list of lines to appear in the new starred-block comment
             * @returns {string} A representation of the comment value in starred-block form, excluding start and end markers
             */
            function convertToStarredBlock(firstComment, commentLinesList) {
                const initialOffset = getInitialOffset(firstComment);
                return `/*\n${commentLinesList.map(line => `${initialOffset} * ${line}`).join("\n")}\n${initialOffset} */`;
            }
            /**
             * Converts a comment into separate-line form
             * @param {Token} firstComment The first comment of the group being converted
             * @param {string[]} commentLinesList A list of lines to appear in the new starred-block comment
             * @returns {string} A representation of the comment value in separate-line form
             */
            function convertToSeparateLines(firstComment, commentLinesList) {
                return commentLinesList.map(line => `// ${line}`).join(`\n${getInitialOffset(firstComment)}`);
            }
            /**
             * Converts a comment into bare-block form
             * @param {Token} firstComment The first comment of the group being converted
             * @param {string[]} commentLinesList A list of lines to appear in the new starred-block comment
             * @returns {string} A representation of the comment value in bare-block form
             */
            function convertToBlock(firstComment, commentLinesList) {
                return `/* ${commentLinesList.join(`\n${getInitialOffset(firstComment)}   `)} */`;
            }
            /**
             * Each method checks a group of comments to see if it's valid according to the given option.
             * @param {Token[]} commentGroup A list of comments that appear together. This will either contain a single
             * block comment or multiple line comments.
             * @returns {void}
             */
            const commentGroupCheckers = {
                "starred-block"(commentGroup) {
                    const [firstComment] = commentGroup;
                    const commentLines = getCommentLines(commentGroup);
                    if (commentLines.some(value => value.includes("*/"))) {
                        return;
                    }
                    if (commentGroup.length > 1) {
                        context.report({
                            loc: {
                                start: firstComment.loc.start,
                                end: commentGroup[commentGroup.length - 1].loc.end
                            },
                            messageId: "expectedBlock",
                            fix(fixer) {
                                const range = [firstComment.range[0], commentGroup[commentGroup.length - 1].range[1]];
                                return commentLines.some(value => value.startsWith("/"))
                                    ? null
                                    : fixer.replaceTextRange(range, convertToStarredBlock(firstComment, commentLines));
                            }
                        });
                    }
                    else {
                        const lines = firstComment.value.split(astUtils.LINEBREAK_MATCHER);
                        const expectedLeadingWhitespace = getInitialOffset(firstComment);
                        const expectedLinePrefix = `${expectedLeadingWhitespace} *`;
                        if (!/^\*?\s*$/u.test(lines[0])) {
                            const start = firstComment.value.startsWith("*") ? firstComment.range[0] + 1 : firstComment.range[0];
                            context.report({
                                loc: {
                                    start: firstComment.loc.start,
                                    end: { line: firstComment.loc.start.line, column: firstComment.loc.start.column + 2 }
                                },
                                messageId: "startNewline",
                                fix: fixer => fixer.insertTextAfterRange([start, start + 2], `\n${expectedLinePrefix}`)
                            });
                        }
                        if (!/^\s*$/u.test(lines[lines.length - 1])) {
                            context.report({
                                loc: {
                                    start: { line: firstComment.loc.end.line, column: firstComment.loc.end.column - 2 },
                                    end: firstComment.loc.end
                                },
                                messageId: "endNewline",
                                fix: fixer => fixer.replaceTextRange([firstComment.range[1] - 2, firstComment.range[1]], `\n${expectedLinePrefix}/`)
                            });
                        }
                        for (let lineNumber = firstComment.loc.start.line + 1; lineNumber <= firstComment.loc.end.line; lineNumber++) {
                            const lineText = sourceCode.lines[lineNumber - 1];
                            const errorType = isStarredCommentLine(lineText)
                                ? "alignment"
                                : "missingStar";
                            if (!lineText.startsWith(expectedLinePrefix)) {
                                context.report({
                                    loc: {
                                        start: { line: lineNumber, column: 0 },
                                        end: { line: lineNumber, column: lineText.length }
                                    },
                                    messageId: errorType,
                                    fix(fixer) {
                                        const lineStartIndex = sourceCode.getIndexFromLoc({ line: lineNumber, column: 0 });
                                        if (errorType === "alignment") {
                                            const [, commentTextPrefix = ""] = lineText.match(/^(\s*\*)/u) || [];
                                            const commentTextStartIndex = lineStartIndex + commentTextPrefix.length;
                                            return fixer.replaceTextRange([lineStartIndex, commentTextStartIndex], expectedLinePrefix);
                                        }
                                        const [, commentTextPrefix = ""] = lineText.match(/^(\s*)/u) || [];
                                        const commentTextStartIndex = lineStartIndex + commentTextPrefix.length;
                                        let offset;
                                        for (const [idx, line] of lines.entries()) {
                                            if (!/\S+/u.test(line)) {
                                                continue;
                                            }
                                            const lineTextToAlignWith = sourceCode.lines[firstComment.loc.start.line - 1 + idx];
                                            const [, prefix = "", initialOffset = ""] = lineTextToAlignWith.match(/^(\s*(?:\/?\*)?(\s*))/u) || [];
                                            offset = `${commentTextPrefix.slice(prefix.length)}${initialOffset}`;
                                            if (/^\s*\//u.test(lineText) && offset.length === 0) {
                                                offset += " ";
                                            }
                                            break;
                                        }
                                        return fixer.replaceTextRange([lineStartIndex, commentTextStartIndex], `${expectedLinePrefix}${offset}`);
                                    }
                                });
                            }
                        }
                    }
                },
                "separate-lines"(commentGroup) {
                    const [firstComment] = commentGroup;
                    if (firstComment.type !== "Block" || isJSDocComment(commentGroup)) {
                        return;
                    }
                    const commentLines = getCommentLines(commentGroup);
                    const tokenAfter = sourceCode.getTokenAfter(firstComment, { includeComments: true });
                    if (tokenAfter && firstComment.loc.end.line === tokenAfter.loc.start.line) {
                        return;
                    }
                    context.report({
                        loc: {
                            start: firstComment.loc.start,
                            end: { line: firstComment.loc.start.line, column: firstComment.loc.start.column + 2 }
                        },
                        messageId: "expectedLines",
                        fix(fixer) {
                            return fixer.replaceText(firstComment, convertToSeparateLines(firstComment, commentLines));
                        }
                    });
                },
                "bare-block"(commentGroup) {
                    if (isJSDocComment(commentGroup)) {
                        return;
                    }
                    const [firstComment] = commentGroup;
                    const commentLines = getCommentLines(commentGroup);
                    // Disallows consecutive line comments in favor of using a block comment.
                    if (firstComment.type === "Line" && commentLines.length > 1 &&
                        !commentLines.some(value => value.includes("*/"))) {
                        context.report({
                            loc: {
                                start: firstComment.loc.start,
                                end: commentGroup[commentGroup.length - 1].loc.end
                            },
                            messageId: "expectedBlock",
                            fix(fixer) {
                                return fixer.replaceTextRange([firstComment.range[0], commentGroup[commentGroup.length - 1].range[1]], convertToBlock(firstComment, commentLines));
                            }
                        });
                    }
                    // Prohibits block comments from having a * at the beginning of each line.
                    if (isStarredBlockComment(commentGroup)) {
                        context.report({
                            loc: {
                                start: firstComment.loc.start,
                                end: { line: firstComment.loc.start.line, column: firstComment.loc.start.column + 2 }
                            },
                            messageId: "expectedBareBlock",
                            fix(fixer) {
                                return fixer.replaceText(firstComment, convertToBlock(firstComment, commentLines));
                            }
                        });
                    }
                }
            };
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                Program() {
                    return sourceCode.getAllComments()
                        .filter(comment => comment.type !== "Shebang")
                        .filter(comment => !astUtils.COMMENTS_IGNORE_PATTERN.test(comment.value))
                        .filter(comment => {
                        const tokenBefore = sourceCode.getTokenBefore(comment, { includeComments: true });
                        return !tokenBefore || tokenBefore.loc.end.line < comment.loc.start.line;
                    })
                        .reduce((commentGroups, comment, index, commentList) => {
                        const tokenBefore = sourceCode.getTokenBefore(comment, { includeComments: true });
                        if (comment.type === "Line" &&
                            index && commentList[index - 1].type === "Line" &&
                            tokenBefore && tokenBefore.loc.end.line === comment.loc.start.line - 1 &&
                            tokenBefore === commentList[index - 1]) {
                            commentGroups[commentGroups.length - 1].push(comment);
                        }
                        else {
                            commentGroups.push([comment]);
                        }
                        return commentGroups;
                    }, [])
                        .filter(commentGroup => !(commentGroup.length === 1 && commentGroup[0].loc.start.line === commentGroup[0].loc.end.line))
                        .forEach(commentGroupCheckers[option]);
                }
            };
        }
    };
},
/* eslint/lib/rules/multiline-ternary.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce newlines between operands of ternary expressions",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/multiline-ternary"
            },
            schema: [
                {
                    enum: ["always", "always-multiline", "never"]
                }
            ],
            messages: {
                expectedTestCons: "Expected newline between test and consequent of ternary expression.",
                expectedConsAlt: "Expected newline between consequent and alternate of ternary expression.",
                unexpectedTestCons: "Unexpected newline between test and consequent of ternary expression.",
                unexpectedConsAlt: "Unexpected newline between consequent and alternate of ternary expression."
            },
            fixable: "whitespace"
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const option = context.options[0];
            const multiline = option !== "never";
            const allowSingleLine = option === "always-multiline";
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                ConditionalExpression(node) {
                    const questionToken = sourceCode.getTokenAfter(node.test, astUtils.isNotClosingParenToken);
                    const colonToken = sourceCode.getTokenAfter(node.consequent, astUtils.isNotClosingParenToken);
                    const firstTokenOfTest = sourceCode.getFirstToken(node);
                    const lastTokenOfTest = sourceCode.getTokenBefore(questionToken);
                    const firstTokenOfConsequent = sourceCode.getTokenAfter(questionToken);
                    const lastTokenOfConsequent = sourceCode.getTokenBefore(colonToken);
                    const firstTokenOfAlternate = sourceCode.getTokenAfter(colonToken);
                    const areTestAndConsequentOnSameLine = astUtils.isTokenOnSameLine(lastTokenOfTest, firstTokenOfConsequent);
                    const areConsequentAndAlternateOnSameLine = astUtils.isTokenOnSameLine(lastTokenOfConsequent, firstTokenOfAlternate);
                    const hasComments = !!sourceCode.getCommentsInside(node).length;
                    if (!multiline) {
                        if (!areTestAndConsequentOnSameLine) {
                            context.report({
                                node: node.test,
                                loc: {
                                    start: firstTokenOfTest.loc.start,
                                    end: lastTokenOfTest.loc.end
                                },
                                messageId: "unexpectedTestCons",
                                fix: fixer => {
                                    if (hasComments) {
                                        return null;
                                    }
                                    const fixers = [];
                                    const areTestAndQuestionOnSameLine = astUtils.isTokenOnSameLine(lastTokenOfTest, questionToken);
                                    const areQuestionAndConsOnSameLine = astUtils.isTokenOnSameLine(questionToken, firstTokenOfConsequent);
                                    if (!areTestAndQuestionOnSameLine) {
                                        fixers.push(fixer.removeRange([lastTokenOfTest.range[1], questionToken.range[0]]));
                                    }
                                    if (!areQuestionAndConsOnSameLine) {
                                        fixers.push(fixer.removeRange([questionToken.range[1], firstTokenOfConsequent.range[0]]));
                                    }
                                    return fixers;
                                }
                            });
                        }
                        if (!areConsequentAndAlternateOnSameLine) {
                            context.report({
                                node: node.consequent,
                                loc: {
                                    start: firstTokenOfConsequent.loc.start,
                                    end: lastTokenOfConsequent.loc.end
                                },
                                messageId: "unexpectedConsAlt",
                                fix: fixer => {
                                    if (hasComments) {
                                        return null;
                                    }
                                    const fixers = [];
                                    const areConsAndColonOnSameLine = astUtils.isTokenOnSameLine(lastTokenOfConsequent, colonToken);
                                    const areColonAndAltOnSameLine = astUtils.isTokenOnSameLine(colonToken, firstTokenOfAlternate);
                                    if (!areConsAndColonOnSameLine) {
                                        fixers.push(fixer.removeRange([lastTokenOfConsequent.range[1], colonToken.range[0]]));
                                    }
                                    if (!areColonAndAltOnSameLine) {
                                        fixers.push(fixer.removeRange([colonToken.range[1], firstTokenOfAlternate.range[0]]));
                                    }
                                    return fixers;
                                }
                            });
                        }
                    }
                    else {
                        if (allowSingleLine && node.loc.start.line === node.loc.end.line) {
                            return;
                        }
                        if (areTestAndConsequentOnSameLine) {
                            context.report({
                                node: node.test,
                                loc: {
                                    start: firstTokenOfTest.loc.start,
                                    end: lastTokenOfTest.loc.end
                                },
                                messageId: "expectedTestCons",
                                fix: fixer => (hasComments ? null : (fixer.replaceTextRange([
                                    lastTokenOfTest.range[1],
                                    questionToken.range[0]
                                ], "\n")))
                            });
                        }
                        if (areConsequentAndAlternateOnSameLine) {
                            context.report({
                                node: node.consequent,
                                loc: {
                                    start: firstTokenOfConsequent.loc.start,
                                    end: lastTokenOfConsequent.loc.end
                                },
                                messageId: "expectedConsAlt",
                                fix: (fixer => (hasComments ? null : (fixer.replaceTextRange([
                                    lastTokenOfConsequent.range[1],
                                    colonToken.range[0]
                                ], "\n"))))
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/new-cap.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const CAPS_ALLOWED = [
        "Array",
        "Boolean",
        "Date",
        "Error",
        "Function",
        "Number",
        "Object",
        "RegExp",
        "String",
        "Symbol",
        "BigInt"
    ];
    /**
     * Ensure that if the key is provided, it must be an array.
     * @param {Object} obj Object to check with `key`.
     * @param {string} key Object key to check on `obj`.
     * @param {*} fallback If obj[key] is not present, this will be returned.
     * @returns {string[]} Returns obj[key] if it's an Array, otherwise `fallback`
     */
    function checkArray(obj, key, fallback) {
        /* istanbul ignore if */
        if (Object.prototype.hasOwnProperty.call(obj, key) && !Array.isArray(obj[key])) {
            throw new TypeError(`${key}, if provided, must be an Array`);
        }
        return obj[key] || fallback;
    }
    /**
     * A reducer function to invert an array to an Object mapping the string form of the key, to `true`.
     * @param {Object} map Accumulator object for the reduce.
     * @param {string} key Object key to set to `true`.
     * @returns {Object} Returns the updated Object for further reduction.
     */
    function invert(map, key) {
        map[key] = true;
        return map;
    }
    /**
     * Creates an object with the cap is new exceptions as its keys and true as their values.
     * @param {Object} config Rule configuration
     * @returns {Object} Object with cap is new exceptions.
     */
    function calculateCapIsNewExceptions(config) {
        let capIsNewExceptions = checkArray(config, "capIsNewExceptions", CAPS_ALLOWED);
        if (capIsNewExceptions !== CAPS_ALLOWED) {
            capIsNewExceptions = capIsNewExceptions.concat(CAPS_ALLOWED);
        }
        return capIsNewExceptions.reduce(invert, {});
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require constructor names to begin with a capital letter",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/new-cap"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        newIsCap: {
                            type: "boolean",
                            default: true
                        },
                        capIsNew: {
                            type: "boolean",
                            default: true
                        },
                        newIsCapExceptions: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        },
                        newIsCapExceptionPattern: {
                            type: "string"
                        },
                        capIsNewExceptions: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        },
                        capIsNewExceptionPattern: {
                            type: "string"
                        },
                        properties: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                upper: "A function with a name starting with an uppercase letter should only be used as a constructor.",
                lower: "A constructor name should not start with a lowercase letter."
            }
        },
        create(context) {
            const config = Object.assign({}, context.options[0]);
            config.newIsCap = config.newIsCap !== false;
            config.capIsNew = config.capIsNew !== false;
            const skipProperties = config.properties === false;
            const newIsCapExceptions = checkArray(config, "newIsCapExceptions", []).reduce(invert, {});
            const newIsCapExceptionPattern = config.newIsCapExceptionPattern ? new RegExp(config.newIsCapExceptionPattern, "u") : null;
            const capIsNewExceptions = calculateCapIsNewExceptions(config);
            const capIsNewExceptionPattern = config.capIsNewExceptionPattern ? new RegExp(config.capIsNewExceptionPattern, "u") : null;
            const listeners = {};
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Get exact callee name from expression
             * @param {ASTNode} node CallExpression or NewExpression node
             * @returns {string} name
             */
            function extractNameFromExpression(node) {
                return node.callee.type === "Identifier"
                    ? node.callee.name
                    : astUtils.getStaticPropertyName(node.callee) || "";
            }
            /**
             * Returns the capitalization state of the string -
             * Whether the first character is uppercase, lowercase, or non-alphabetic
             * @param {string} str String
             * @returns {string} capitalization state: "non-alpha", "lower", or "upper"
             */
            function getCap(str) {
                const firstChar = str.charAt(0);
                const firstCharLower = firstChar.toLowerCase();
                const firstCharUpper = firstChar.toUpperCase();
                if (firstCharLower === firstCharUpper) {
                    // char has no uppercase variant, so it's non-alphabetic
                    return "non-alpha";
                }
                if (firstChar === firstCharLower) {
                    return "lower";
                }
                return "upper";
            }
            /**
             * Check if capitalization is allowed for a CallExpression
             * @param {Object} allowedMap Object mapping calleeName to a Boolean
             * @param {ASTNode} node CallExpression node
             * @param {string} calleeName Capitalized callee name from a CallExpression
             * @param {Object} pattern RegExp object from options pattern
             * @returns {boolean} Returns true if the callee may be capitalized
             */
            function isCapAllowed(allowedMap, node, calleeName, pattern) {
                const sourceText = sourceCode.getText(node.callee);
                if (allowedMap[calleeName] || allowedMap[sourceText]) {
                    return true;
                }
                if (pattern && pattern.test(sourceText)) {
                    return true;
                }
                const callee = astUtils.skipChainExpression(node.callee);
                if (calleeName === "UTC" && callee.type === "MemberExpression") {
                    // allow if callee is Date.UTC
                    return callee.object.type === "Identifier" &&
                        callee.object.name === "Date";
                }
                return skipProperties && callee.type === "MemberExpression";
            }
            /**
             * Reports the given messageId for the given node. The location will be the start of the property or the callee.
             * @param {ASTNode} node CallExpression or NewExpression node.
             * @param {string} messageId The messageId to report.
             * @returns {void}
             */
            function report(node, messageId) {
                let callee = astUtils.skipChainExpression(node.callee);
                if (callee.type === "MemberExpression") {
                    callee = callee.property;
                }
                context.report({ node, loc: callee.loc, messageId });
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            if (config.newIsCap) {
                listeners.NewExpression = function (node) {
                    const constructorName = extractNameFromExpression(node);
                    if (constructorName) {
                        const capitalization = getCap(constructorName);
                        const isAllowed = capitalization !== "lower" || isCapAllowed(newIsCapExceptions, node, constructorName, newIsCapExceptionPattern);
                        if (!isAllowed) {
                            report(node, "lower");
                        }
                    }
                };
            }
            if (config.capIsNew) {
                listeners.CallExpression = function (node) {
                    const calleeName = extractNameFromExpression(node);
                    if (calleeName) {
                        const capitalization = getCap(calleeName);
                        const isAllowed = capitalization !== "upper" || isCapAllowed(capIsNewExceptions, node, calleeName, capIsNewExceptionPattern);
                        if (!isAllowed) {
                            report(node, "upper");
                        }
                    }
                };
            }
            return listeners;
        }
    };
},
/* eslint/lib/rules/new-parens.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce or disallow parentheses when invoking a constructor with no arguments",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/new-parens"
            },
            fixable: "code",
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always", "never"]
                            }
                        ],
                        minItems: 0,
                        maxItems: 1
                    }
                ]
            },
            messages: {
                missing: "Missing '()' invoking a constructor.",
                unnecessary: "Unnecessary '()' invoking a constructor with no arguments."
            }
        },
        create(context) {
            const options = context.options;
            const always = options[0] !== "never"; // Default is always
            const sourceCode = context.getSourceCode();
            return {
                NewExpression(node) {
                    if (node.arguments.length !== 0) {
                        return; // if there are arguments, there have to be parens
                    }
                    const lastToken = sourceCode.getLastToken(node);
                    const hasLastParen = lastToken && astUtils.isClosingParenToken(lastToken);
                    // `hasParens` is true only if the new expression ends with its own parens, e.g., new new foo() does not end with its own parens
                    const hasParens = hasLastParen &&
                        astUtils.isOpeningParenToken(sourceCode.getTokenBefore(lastToken)) &&
                        node.callee.range[1] < node.range[1];
                    if (always) {
                        if (!hasParens) {
                            context.report({
                                node,
                                messageId: "missing",
                                fix: fixer => fixer.insertTextAfter(node, "()")
                            });
                        }
                    }
                    else {
                        if (hasParens) {
                            context.report({
                                node,
                                messageId: "unnecessary",
                                fix: fixer => [
                                    fixer.remove(sourceCode.getTokenBefore(lastToken)),
                                    fixer.remove(lastToken),
                                    fixer.insertTextBefore(node, "("),
                                    fixer.insertTextAfter(node, ")")
                                ]
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/newline-after-var.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow an empty line after variable declarations",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/newline-after-var"
            },
            schema: [
                {
                    enum: ["never", "always"]
                }
            ],
            fixable: "whitespace",
            messages: {
                expected: "Expected blank line after variable declarations.",
                unexpected: "Unexpected blank line after variable declarations."
            },
            deprecated: true,
            replacedBy: ["padding-line-between-statements"]
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            // Default `mode` to "always".
            const mode = context.options[0] === "never" ? "never" : "always";
            // Cache starting and ending line numbers of comments for faster lookup
            const commentEndLine = sourceCode.getAllComments().reduce((result, token) => {
                result[token.loc.start.line] = token.loc.end.line;
                return result;
            }, {});
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Gets a token from the given node to compare line to the next statement.
             *
             * In general, the token is the last token of the node. However, the token is the second last token if the following conditions satisfy.
             *
             * - The last token is semicolon.
             * - The semicolon is on a different line from the previous token of the semicolon.
             *
             * This behavior would address semicolon-less style code. e.g.:
             *
             *     var foo = 1
             *
             *     ;(a || b).doSomething()
             * @param {ASTNode} node The node to get.
             * @returns {Token} The token to compare line to the next statement.
             */
            function getLastToken(node) {
                const lastToken = sourceCode.getLastToken(node);
                if (lastToken.type === "Punctuator" && lastToken.value === ";") {
                    const prevToken = sourceCode.getTokenBefore(lastToken);
                    if (prevToken.loc.end.line !== lastToken.loc.start.line) {
                        return prevToken;
                    }
                }
                return lastToken;
            }
            /**
             * Determine if provided keyword is a variable declaration
             * @private
             * @param {string} keyword keyword to test
             * @returns {boolean} True if `keyword` is a type of var
             */
            function isVar(keyword) {
                return keyword === "var" || keyword === "let" || keyword === "const";
            }
            /**
             * Determine if provided keyword is a variant of for specifiers
             * @private
             * @param {string} keyword keyword to test
             * @returns {boolean} True if `keyword` is a variant of for specifier
             */
            function isForTypeSpecifier(keyword) {
                return keyword === "ForStatement" || keyword === "ForInStatement" || keyword === "ForOfStatement";
            }
            /**
             * Determine if provided keyword is an export specifiers
             * @private
             * @param {string} nodeType nodeType to test
             * @returns {boolean} True if `nodeType` is an export specifier
             */
            function isExportSpecifier(nodeType) {
                return nodeType === "ExportNamedDeclaration" || nodeType === "ExportSpecifier" ||
                    nodeType === "ExportDefaultDeclaration" || nodeType === "ExportAllDeclaration";
            }
            /**
             * Determine if provided node is the last of their parent block.
             * @private
             * @param {ASTNode} node node to test
             * @returns {boolean} True if `node` is last of their parent block.
             */
            function isLastNode(node) {
                const token = sourceCode.getTokenAfter(node);
                return !token || (token.type === "Punctuator" && token.value === "}");
            }
            /**
             * Gets the last line of a group of consecutive comments
             * @param {number} commentStartLine The starting line of the group
             * @returns {number} The number of the last comment line of the group
             */
            function getLastCommentLineOfBlock(commentStartLine) {
                const currentCommentEnd = commentEndLine[commentStartLine];
                return commentEndLine[currentCommentEnd + 1] ? getLastCommentLineOfBlock(currentCommentEnd + 1) : currentCommentEnd;
            }
            /**
             * Determine if a token starts more than one line after a comment ends
             * @param  {token}   token            The token being checked
             * @param {integer}  commentStartLine The line number on which the comment starts
             * @returns {boolean}                 True if `token` does not start immediately after a comment
             */
            function hasBlankLineAfterComment(token, commentStartLine) {
                return token.loc.start.line > getLastCommentLineOfBlock(commentStartLine) + 1;
            }
            /**
             * Checks that a blank line exists after a variable declaration when mode is
             * set to "always", or checks that there is no blank line when mode is set
             * to "never"
             * @private
             * @param {ASTNode} node `VariableDeclaration` node to test
             * @returns {void}
             */
            function checkForBlankLine(node) {
                /*
                 * lastToken is the last token on the node's line. It will usually also be the last token of the node, but it will
                 * sometimes be second-last if there is a semicolon on a different line.
                 */
                const lastToken = getLastToken(node), 
                /*
                 * If lastToken is the last token of the node, nextToken should be the token after the node. Otherwise, nextToken
                 * is the last token of the node.
                 */
                nextToken = lastToken === sourceCode.getLastToken(node) ? sourceCode.getTokenAfter(node) : sourceCode.getLastToken(node), nextLineNum = lastToken.loc.end.line + 1;
                // Ignore if there is no following statement
                if (!nextToken) {
                    return;
                }
                // Ignore if parent of node is a for variant
                if (isForTypeSpecifier(node.parent.type)) {
                    return;
                }
                // Ignore if parent of node is an export specifier
                if (isExportSpecifier(node.parent.type)) {
                    return;
                }
                /*
                 * Some coding styles use multiple `var` statements, so do nothing if
                 * the next token is a `var` statement.
                 */
                if (nextToken.type === "Keyword" && isVar(nextToken.value)) {
                    return;
                }
                // Ignore if it is last statement in a block
                if (isLastNode(node)) {
                    return;
                }
                // Next statement is not a `var`...
                const noNextLineToken = nextToken.loc.start.line > nextLineNum;
                const hasNextLineComment = (typeof commentEndLine[nextLineNum] !== "undefined");
                if (mode === "never" && noNextLineToken && !hasNextLineComment) {
                    context.report({
                        node,
                        messageId: "unexpected",
                        data: { identifier: node.name },
                        fix(fixer) {
                            const linesBetween = sourceCode.getText().slice(lastToken.range[1], nextToken.range[0]).split(astUtils.LINEBREAK_MATCHER);
                            return fixer.replaceTextRange([lastToken.range[1], nextToken.range[0]], `${linesBetween.slice(0, -1).join("")}\n${linesBetween[linesBetween.length - 1]}`);
                        }
                    });
                }
                // Token on the next line, or comment without blank line
                if (mode === "always" && (!noNextLineToken ||
                    hasNextLineComment && !hasBlankLineAfterComment(nextToken, nextLineNum))) {
                    context.report({
                        node,
                        messageId: "expected",
                        data: { identifier: node.name },
                        fix(fixer) {
                            if ((noNextLineToken ? getLastCommentLineOfBlock(nextLineNum) : lastToken.loc.end.line) === nextToken.loc.start.line) {
                                return fixer.insertTextBefore(nextToken, "\n\n");
                            }
                            return fixer.insertTextBeforeRange([nextToken.range[0] - nextToken.loc.start.column, nextToken.range[1]], "\n");
                        }
                    });
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                VariableDeclaration: checkForBlankLine
            };
        }
    };
},
/* eslint/lib/rules/newline-before-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require an empty line before `return` statements",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/newline-before-return"
            },
            fixable: "whitespace",
            schema: [],
            messages: {
                expected: "Expected newline before return statement."
            },
            deprecated: true,
            replacedBy: ["padding-line-between-statements"]
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Tests whether node is preceded by supplied tokens
             * @param {ASTNode} node node to check
             * @param {Array} testTokens array of tokens to test against
             * @returns {boolean} Whether or not the node is preceded by one of the supplied tokens
             * @private
             */
            function isPrecededByTokens(node, testTokens) {
                const tokenBefore = sourceCode.getTokenBefore(node);
                return testTokens.some(token => tokenBefore.value === token);
            }
            /**
             * Checks whether node is the first node after statement or in block
             * @param {ASTNode} node node to check
             * @returns {boolean} Whether or not the node is the first node after statement or in block
             * @private
             */
            function isFirstNode(node) {
                const parentType = node.parent.type;
                if (node.parent.body) {
                    return Array.isArray(node.parent.body)
                        ? node.parent.body[0] === node
                        : node.parent.body === node;
                }
                if (parentType === "IfStatement") {
                    return isPrecededByTokens(node, ["else", ")"]);
                }
                if (parentType === "DoWhileStatement") {
                    return isPrecededByTokens(node, ["do"]);
                }
                if (parentType === "SwitchCase") {
                    return isPrecededByTokens(node, [":"]);
                }
                return isPrecededByTokens(node, [")"]);
            }
            /**
             * Returns the number of lines of comments that precede the node
             * @param {ASTNode} node node to check for overlapping comments
             * @param {number} lineNumTokenBefore line number of previous token, to check for overlapping comments
             * @returns {number} Number of lines of comments that precede the node
             * @private
             */
            function calcCommentLines(node, lineNumTokenBefore) {
                const comments = sourceCode.getCommentsBefore(node);
                let numLinesComments = 0;
                if (!comments.length) {
                    return numLinesComments;
                }
                comments.forEach(comment => {
                    numLinesComments++;
                    if (comment.type === "Block") {
                        numLinesComments += comment.loc.end.line - comment.loc.start.line;
                    }
                    // avoid counting lines with inline comments twice
                    if (comment.loc.start.line === lineNumTokenBefore) {
                        numLinesComments--;
                    }
                    if (comment.loc.end.line === node.loc.start.line) {
                        numLinesComments--;
                    }
                });
                return numLinesComments;
            }
            /**
             * Returns the line number of the token before the node that is passed in as an argument
             * @param {ASTNode} node The node to use as the start of the calculation
             * @returns {number} Line number of the token before `node`
             * @private
             */
            function getLineNumberOfTokenBefore(node) {
                const tokenBefore = sourceCode.getTokenBefore(node);
                let lineNumTokenBefore;
                /**
                 * Global return (at the beginning of a script) is a special case.
                 * If there is no token before `return`, then we expect no line
                 * break before the return. Comments are allowed to occupy lines
                 * before the global return, just no blank lines.
                 * Setting lineNumTokenBefore to zero in that case results in the
                 * desired behavior.
                 */
                if (tokenBefore) {
                    lineNumTokenBefore = tokenBefore.loc.end.line;
                }
                else {
                    lineNumTokenBefore = 0; // global return at beginning of script
                }
                return lineNumTokenBefore;
            }
            /**
             * Checks whether node is preceded by a newline
             * @param {ASTNode} node node to check
             * @returns {boolean} Whether or not the node is preceded by a newline
             * @private
             */
            function hasNewlineBefore(node) {
                const lineNumNode = node.loc.start.line;
                const lineNumTokenBefore = getLineNumberOfTokenBefore(node);
                const commentLines = calcCommentLines(node, lineNumTokenBefore);
                return (lineNumNode - lineNumTokenBefore - commentLines) > 1;
            }
            /**
             * Checks whether it is safe to apply a fix to a given return statement.
             *
             * The fix is not considered safe if the given return statement has leading comments,
             * as we cannot safely determine if the newline should be added before or after the comments.
             * For more information, see: https://github.com/eslint/eslint/issues/5958#issuecomment-222767211
             * @param {ASTNode} node The return statement node to check.
             * @returns {boolean} `true` if it can fix the node.
             * @private
             */
            function canFix(node) {
                const leadingComments = sourceCode.getCommentsBefore(node);
                const lastLeadingComment = leadingComments[leadingComments.length - 1];
                const tokenBefore = sourceCode.getTokenBefore(node);
                if (leadingComments.length === 0) {
                    return true;
                }
                /*
                 * if the last leading comment ends in the same line as the previous token and
                 * does not share a line with the `return` node, we can consider it safe to fix.
                 * Example:
                 * function a() {
                 *     var b; //comment
                 *     return;
                 * }
                 */
                if (lastLeadingComment.loc.end.line === tokenBefore.loc.end.line &&
                    lastLeadingComment.loc.end.line !== node.loc.start.line) {
                    return true;
                }
                return false;
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                ReturnStatement(node) {
                    if (!isFirstNode(node) && !hasNewlineBefore(node)) {
                        context.report({
                            node,
                            messageId: "expected",
                            fix(fixer) {
                                if (canFix(node)) {
                                    const tokenBefore = sourceCode.getTokenBefore(node);
                                    const newlines = node.loc.start.line === tokenBefore.loc.end.line ? "\n\n" : "\n";
                                    return fixer.insertTextBefore(node, newlines);
                                }
                                return null;
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/newline-per-chained-call.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require a newline after each call in a method chain",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/newline-per-chained-call"
            },
            fixable: "whitespace",
            schema: [{
                    type: "object",
                    properties: {
                        ignoreChainWithDepth: {
                            type: "integer",
                            minimum: 1,
                            maximum: 10,
                            default: 2
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                expected: "Expected line break before `{{callee}}`."
            }
        },
        create(context) {
            const options = context.options[0] || {}, ignoreChainWithDepth = options.ignoreChainWithDepth || 2;
            const sourceCode = context.getSourceCode();
            /**
             * Get the prefix of a given MemberExpression node.
             * If the MemberExpression node is a computed value it returns a
             * left bracket. If not it returns a period.
             * @param  {ASTNode} node A MemberExpression node to get
             * @returns {string} The prefix of the node.
             */
            function getPrefix(node) {
                if (node.computed) {
                    if (node.optional) {
                        return "?.[";
                    }
                    return "[";
                }
                if (node.optional) {
                    return "?.";
                }
                return ".";
            }
            /**
             * Gets the property text of a given MemberExpression node.
             * If the text is multiline, this returns only the first line.
             * @param {ASTNode} node A MemberExpression node to get.
             * @returns {string} The property text of the node.
             */
            function getPropertyText(node) {
                const prefix = getPrefix(node);
                const lines = sourceCode.getText(node.property).split(astUtils.LINEBREAK_MATCHER);
                const suffix = node.computed && lines.length === 1 ? "]" : "";
                return prefix + lines[0] + suffix;
            }
            return {
                "CallExpression:exit"(node) {
                    const callee = astUtils.skipChainExpression(node.callee);
                    if (callee.type !== "MemberExpression") {
                        return;
                    }
                    let parent = astUtils.skipChainExpression(callee.object);
                    let depth = 1;
                    while (parent && parent.callee) {
                        depth += 1;
                        parent = astUtils.skipChainExpression(astUtils.skipChainExpression(parent.callee).object);
                    }
                    if (depth > ignoreChainWithDepth && astUtils.isTokenOnSameLine(callee.object, callee.property)) {
                        const firstTokenAfterObject = sourceCode.getTokenAfter(callee.object, astUtils.isNotClosingParenToken);
                        context.report({
                            node: callee.property,
                            loc: {
                                start: firstTokenAfterObject.loc.start,
                                end: callee.loc.end
                            },
                            messageId: "expected",
                            data: {
                                callee: getPropertyText(callee)
                            },
                            fix(fixer) {
                                return fixer.insertTextBefore(firstTokenAfterObject, "\n");
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-alert.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { getStaticPropertyName: getPropertyName, getVariableByName, skipChainExpression } = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks if the given name is a prohibited identifier.
     * @param {string} name The name to check
     * @returns {boolean} Whether or not the name is prohibited.
     */
    function isProhibitedIdentifier(name) {
        return /^(alert|confirm|prompt)$/u.test(name);
    }
    /**
     * Finds the eslint-scope reference in the given scope.
     * @param {Object} scope The scope to search.
     * @param {ASTNode} node The identifier node.
     * @returns {Reference|null} Returns the found reference or null if none were found.
     */
    function findReference(scope, node) {
        const references = scope.references.filter(reference => reference.identifier.range[0] === node.range[0] &&
            reference.identifier.range[1] === node.range[1]);
        if (references.length === 1) {
            return references[0];
        }
        return null;
    }
    /**
     * Checks if the given identifier node is shadowed in the given scope.
     * @param {Object} scope The current scope.
     * @param {string} node The identifier node to check
     * @returns {boolean} Whether or not the name is shadowed.
     */
    function isShadowed(scope, node) {
        const reference = findReference(scope, node);
        return reference && reference.resolved && reference.resolved.defs.length > 0;
    }
    /**
     * Checks if the given identifier node is a ThisExpression in the global scope or the global window property.
     * @param {Object} scope The current scope.
     * @param {string} node The identifier node to check
     * @returns {boolean} Whether or not the node is a reference to the global object.
     */
    function isGlobalThisReferenceOrGlobalWindow(scope, node) {
        if (scope.type === "global" && node.type === "ThisExpression") {
            return true;
        }
        if (node.type === "Identifier" &&
            (node.name === "window" ||
                (node.name === "globalThis" && getVariableByName(scope, "globalThis")))) {
            return !isShadowed(scope, node);
        }
        return false;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the use of `alert`, `confirm`, and `prompt`",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-alert"
            },
            schema: [],
            messages: {
                unexpected: "Unexpected {{name}}."
            }
        },
        create(context) {
            return {
                CallExpression(node) {
                    const callee = skipChainExpression(node.callee), currentScope = context.getScope();
                    // without window.
                    if (callee.type === "Identifier") {
                        const name = callee.name;
                        if (!isShadowed(currentScope, callee) && isProhibitedIdentifier(callee.name)) {
                            context.report({
                                node,
                                messageId: "unexpected",
                                data: { name }
                            });
                        }
                    }
                    else if (callee.type === "MemberExpression" && isGlobalThisReferenceOrGlobalWindow(currentScope, callee.object)) {
                        const name = getPropertyName(callee);
                        if (isProhibitedIdentifier(name)) {
                            context.report({
                                node,
                                messageId: "unexpected",
                                data: { name }
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-array-constructor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `Array` constructors",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-array-constructor"
            },
            schema: [],
            messages: {
                preferLiteral: "The array literal notation [] is preferable."
            }
        },
        create(context) {
            /**
             * Disallow construction of dense arrays using the Array constructor
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function check(node) {
                if (node.arguments.length !== 1 &&
                    node.callee.type === "Identifier" &&
                    node.callee.name === "Array") {
                    context.report({ node, messageId: "preferLiteral" });
                }
            }
            return {
                CallExpression: check,
                NewExpression: check
            };
        }
    };
},
/* eslint/lib/rules/no-async-promise-executor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow using an async function as a Promise executor",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-async-promise-executor"
            },
            fixable: null,
            schema: [],
            messages: {
                async: "Promise executor functions should not be async."
            }
        },
        create(context) {
            return {
                "NewExpression[callee.name='Promise'][arguments.0.async=true]"(node) {
                    context.report({
                        node: context.getSourceCode().getFirstToken(node.arguments[0], token => token.value === "async"),
                        messageId: "async"
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-await-in-loop.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Check whether it should stop traversing ancestors at the given node.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if it should stop traversing.
     */
    function isBoundary(node) {
        const t = node.type;
        return (t === "FunctionDeclaration" ||
            t === "FunctionExpression" ||
            t === "ArrowFunctionExpression" ||
            /*
             * Don't report the await expressions on for-await-of loop since it's
             * asynchronous iteration intentionally.
             */
            (t === "ForOfStatement" && node.await === true));
    }
    /**
     * Check whether the given node is in loop.
     * @param {ASTNode} node A node to check.
     * @param {ASTNode} parent A parent node to check.
     * @returns {boolean} `true` if the node is in loop.
     */
    function isLooped(node, parent) {
        switch (parent.type) {
            case "ForStatement":
                return (node === parent.test ||
                    node === parent.update ||
                    node === parent.body);
            case "ForOfStatement":
            case "ForInStatement":
                return node === parent.body;
            case "WhileStatement":
            case "DoWhileStatement":
                return node === parent.test || node === parent.body;
            default:
                return false;
        }
    }
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow `await` inside of loops",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-await-in-loop"
            },
            schema: [],
            messages: {
                unexpectedAwait: "Unexpected `await` inside a loop."
            }
        },
        create(context) {
            /**
             * Validate an await expression.
             * @param {ASTNode} awaitNode An AwaitExpression or ForOfStatement node to validate.
             * @returns {void}
             */
            function validate(awaitNode) {
                if (awaitNode.type === "ForOfStatement" && !awaitNode.await) {
                    return;
                }
                let node = awaitNode;
                let parent = node.parent;
                while (parent && !isBoundary(parent)) {
                    if (isLooped(node, parent)) {
                        context.report({
                            node: awaitNode,
                            messageId: "unexpectedAwait"
                        });
                        return;
                    }
                    node = parent;
                    parent = parent.parent;
                }
            }
            return {
                AwaitExpression: validate,
                ForOfStatement: validate
            };
        }
    };
},
/* eslint/lib/rules/no-bitwise.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
     *
     * Set of bitwise operators.
     *
     */
    const BITWISE_OPERATORS = [
        "^", "|", "&", "<<", ">>", ">>>",
        "^=", "|=", "&=", "<<=", ">>=", ">>>=",
        "~"
    ];
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow bitwise operators",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-bitwise"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allow: {
                            type: "array",
                            items: {
                                enum: BITWISE_OPERATORS
                            },
                            uniqueItems: true
                        },
                        int32Hint: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpected: "Unexpected use of '{{operator}}'."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const allowed = options.allow || [];
            const int32Hint = options.int32Hint === true;
            /**
             * Reports an unexpected use of a bitwise operator.
             * @param   {ASTNode} node Node which contains the bitwise operator.
             * @returns {void}
             */
            function report(node) {
                context.report({ node, messageId: "unexpected", data: { operator: node.operator } });
            }
            /**
             * Checks if the given node has a bitwise operator.
             * @param   {ASTNode} node The node to check.
             * @returns {boolean} Whether or not the node has a bitwise operator.
             */
            function hasBitwiseOperator(node) {
                return BITWISE_OPERATORS.indexOf(node.operator) !== -1;
            }
            /**
             * Checks if exceptions were provided, e.g. `{ allow: ['~', '|'] }`.
             * @param   {ASTNode} node The node to check.
             * @returns {boolean} Whether or not the node has a bitwise operator.
             */
            function allowedOperator(node) {
                return allowed.indexOf(node.operator) !== -1;
            }
            /**
             * Checks if the given bitwise operator is used for integer typecasting, i.e. "|0"
             * @param   {ASTNode} node The node to check.
             * @returns {boolean} whether the node is used in integer typecasting.
             */
            function isInt32Hint(node) {
                return int32Hint && node.operator === "|" && node.right &&
                    node.right.type === "Literal" && node.right.value === 0;
            }
            /**
             * Report if the given node contains a bitwise operator.
             * @param   {ASTNode} node The node to check.
             * @returns {void}
             */
            function checkNodeForBitwiseOperator(node) {
                if (hasBitwiseOperator(node) && !allowedOperator(node) && !isInt32Hint(node)) {
                    report(node);
                }
            }
            return {
                AssignmentExpression: checkNodeForBitwiseOperator,
                BinaryExpression: checkNodeForBitwiseOperator,
                UnaryExpression: checkNodeForBitwiseOperator
            };
        }
    };
},
/* eslint/lib/rules/no-buffer-constructor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "problem",
            docs: {
                description: "disallow use of the `Buffer()` constructor",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-buffer-constructor"
            },
            schema: [],
            messages: {
                deprecated: "{{expr}} is deprecated. Use Buffer.from(), Buffer.alloc(), or Buffer.allocUnsafe() instead."
            }
        },
        create(context) {
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                "CallExpression[callee.name='Buffer'], NewExpression[callee.name='Buffer']"(node) {
                    context.report({
                        node,
                        messageId: "deprecated",
                        data: { expr: node.type === "CallExpression" ? "Buffer()" : "new Buffer()" }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-caller.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the use of `arguments.caller` or `arguments.callee`",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-caller"
            },
            schema: [],
            messages: {
                unexpected: "Avoid arguments.{{prop}}."
            }
        },
        create(context) {
            return {
                MemberExpression(node) {
                    const objectName = node.object.name, propertyName = node.property.name;
                    if (objectName === "arguments" && !node.computed && propertyName && propertyName.match(/^calle[er]$/u)) {
                        context.report({ node, messageId: "unexpected", data: { prop: propertyName } });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-case-declarations.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow lexical declarations in case clauses",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-case-declarations"
            },
            schema: [],
            messages: {
                unexpected: "Unexpected lexical declaration in case block."
            }
        },
        create(context) {
            /**
             * Checks whether or not a node is a lexical declaration.
             * @param {ASTNode} node A direct child statement of a switch case.
             * @returns {boolean} Whether or not the node is a lexical declaration.
             */
            function isLexicalDeclaration(node) {
                switch (node.type) {
                    case "FunctionDeclaration":
                    case "ClassDeclaration":
                        return true;
                    case "VariableDeclaration":
                        return node.kind !== "var";
                    default:
                        return false;
                }
            }
            return {
                SwitchCase(node) {
                    for (let i = 0; i < node.consequent.length; i++) {
                        const statement = node.consequent[i];
                        if (isLexicalDeclaration(statement)) {
                            context.report({
                                node: statement,
                                messageId: "unexpected"
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-catch-shadow.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `catch` clause parameters from shadowing variables in the outer scope",
                category: "Variables",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-catch-shadow"
            },
            replacedBy: ["no-shadow"],
            deprecated: true,
            schema: [],
            messages: {
                mutable: "Value of '{{name}}' may be overwritten in IE 8 and earlier."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Check if the parameters are been shadowed
             * @param {Object} scope current scope
             * @param {string} name parameter name
             * @returns {boolean} True is its been shadowed
             */
            function paramIsShadowing(scope, name) {
                return astUtils.getVariableByName(scope, name) !== null;
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                "CatchClause[param!=null]"(node) {
                    let scope = context.getScope();
                    /*
                     * When ecmaVersion >= 6, CatchClause creates its own scope
                     * so start from one upper scope to exclude the current node
                     */
                    if (scope.block === node) {
                        scope = scope.upper;
                    }
                    if (paramIsShadowing(scope, node.param.name)) {
                        context.report({ node, messageId: "mutable", data: { name: node.param.name } });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-class-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow reassigning class members",
                category: "ECMAScript 6",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-class-assign"
            },
            schema: [],
            messages: {
                class: "'{{name}}' is a class."
            }
        },
        create(context) {
            /**
             * Finds and reports references that are non initializer and writable.
             * @param {Variable} variable A variable to check.
             * @returns {void}
             */
            function checkVariable(variable) {
                astUtils.getModifyingReferences(variable.references).forEach(reference => {
                    context.report({ node: reference.identifier, messageId: "class", data: { name: reference.identifier.name } });
                });
            }
            /**
             * Finds and reports references that are non initializer and writable.
             * @param {ASTNode} node A ClassDeclaration/ClassExpression node to check.
             * @returns {void}
             */
            function checkForClass(node) {
                context.getDeclaredVariables(node).forEach(checkVariable);
            }
            return {
                ClassDeclaration: checkForClass,
                ClassExpression: checkForClass
            };
        }
    };
},
/* eslint/lib/rules/no-compare-neg-zero.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow comparing against -0",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-compare-neg-zero"
            },
            fixable: null,
            schema: [],
            messages: {
                unexpected: "Do not use the '{{operator}}' operator to compare against -0."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Checks a given node is -0
             * @param {ASTNode} node A node to check.
             * @returns {boolean} `true` if the node is -0.
             */
            function isNegZero(node) {
                return node.type === "UnaryExpression" && node.operator === "-" && node.argument.type === "Literal" && node.argument.value === 0;
            }
            const OPERATORS_TO_CHECK = new Set([">", ">=", "<", "<=", "==", "===", "!=", "!=="]);
            return {
                BinaryExpression(node) {
                    if (OPERATORS_TO_CHECK.has(node.operator)) {
                        if (isNegZero(node.left) || isNegZero(node.right)) {
                            context.report({
                                node,
                                messageId: "unexpected",
                                data: { operator: node.operator }
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-cond-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const TEST_CONDITION_PARENT_TYPES = new Set(["IfStatement", "WhileStatement", "DoWhileStatement", "ForStatement", "ConditionalExpression"]);
    const NODE_DESCRIPTIONS = {
        DoWhileStatement: "a 'do...while' statement",
        ForStatement: "a 'for' statement",
        IfStatement: "an 'if' statement",
        WhileStatement: "a 'while' statement"
    };
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow assignment operators in conditional expressions",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-cond-assign"
            },
            schema: [
                {
                    enum: ["except-parens", "always"]
                }
            ],
            messages: {
                unexpected: "Unexpected assignment within {{type}}.",
                // must match JSHint's error message
                missing: "Expected a conditional expression and instead saw an assignment."
            }
        },
        create(context) {
            const prohibitAssign = (context.options[0] || "except-parens");
            const sourceCode = context.getSourceCode();
            /**
             * Check whether an AST node is the test expression for a conditional statement.
             * @param {!Object} node The node to test.
             * @returns {boolean} `true` if the node is the text expression for a conditional statement; otherwise, `false`.
             */
            function isConditionalTestExpression(node) {
                return node.parent &&
                    TEST_CONDITION_PARENT_TYPES.has(node.parent.type) &&
                    node === node.parent.test;
            }
            /**
             * Given an AST node, perform a bottom-up search for the first ancestor that represents a conditional statement.
             * @param {!Object} node The node to use at the start of the search.
             * @returns {?Object} The closest ancestor node that represents a conditional statement.
             */
            function findConditionalAncestor(node) {
                let currentAncestor = node;
                do {
                    if (isConditionalTestExpression(currentAncestor)) {
                        return currentAncestor.parent;
                    }
                } while ((currentAncestor = currentAncestor.parent) && !astUtils.isFunction(currentAncestor));
                return null;
            }
            /**
             * Check whether the code represented by an AST node is enclosed in two sets of parentheses.
             * @param {!Object} node The node to test.
             * @returns {boolean} `true` if the code is enclosed in two sets of parentheses; otherwise, `false`.
             */
            function isParenthesisedTwice(node) {
                const previousToken = sourceCode.getTokenBefore(node, 1), nextToken = sourceCode.getTokenAfter(node, 1);
                return astUtils.isParenthesised(sourceCode, node) &&
                    previousToken && astUtils.isOpeningParenToken(previousToken) && previousToken.range[1] <= node.range[0] &&
                    astUtils.isClosingParenToken(nextToken) && nextToken.range[0] >= node.range[1];
            }
            /**
             * Check a conditional statement's test expression for top-level assignments that are not enclosed in parentheses.
             * @param {!Object} node The node for the conditional statement.
             * @returns {void}
             */
            function testForAssign(node) {
                if (node.test &&
                    (node.test.type === "AssignmentExpression") &&
                    (node.type === "ForStatement"
                        ? !astUtils.isParenthesised(sourceCode, node.test)
                        : !isParenthesisedTwice(node.test))) {
                    context.report({
                        node: node.test,
                        messageId: "missing"
                    });
                }
            }
            /**
             * Check whether an assignment expression is descended from a conditional statement's test expression.
             * @param {!Object} node The node for the assignment expression.
             * @returns {void}
             */
            function testForConditionalAncestor(node) {
                const ancestor = findConditionalAncestor(node);
                if (ancestor) {
                    context.report({
                        node,
                        messageId: "unexpected",
                        data: {
                            type: NODE_DESCRIPTIONS[ancestor.type] || ancestor.type
                        }
                    });
                }
            }
            if (prohibitAssign === "always") {
                return {
                    AssignmentExpression: testForConditionalAncestor
                };
            }
            return {
                DoWhileStatement: testForAssign,
                ForStatement: testForAssign,
                IfStatement: testForAssign,
                WhileStatement: testForAssign,
                ConditionalExpression: testForAssign
            };
        }
    };
},
/* eslint/lib/rules/no-confusing-arrow.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils.js */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a node is a conditional expression.
     * @param {ASTNode} node node to test
     * @returns {boolean} `true` if the node is a conditional expression.
     */
    function isConditional(node) {
        return node && node.type === "ConditionalExpression";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow arrow functions where they could be confused with comparisons",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-confusing-arrow"
            },
            fixable: "code",
            schema: [{
                    type: "object",
                    properties: {
                        allowParens: { type: "boolean", default: true }
                    },
                    additionalProperties: false
                }],
            messages: {
                confusing: "Arrow function used ambiguously with a conditional expression."
            }
        },
        create(context) {
            const config = context.options[0] || {};
            const allowParens = config.allowParens || (config.allowParens === void 0);
            const sourceCode = context.getSourceCode();
            /**
             * Reports if an arrow function contains an ambiguous conditional.
             * @param {ASTNode} node A node to check and report.
             * @returns {void}
             */
            function checkArrowFunc(node) {
                const body = node.body;
                if (isConditional(body) && !(allowParens && astUtils.isParenthesised(sourceCode, body))) {
                    context.report({
                        node,
                        messageId: "confusing",
                        fix(fixer) {
                            // if `allowParens` is not set to true don't bother wrapping in parens
                            return allowParens && fixer.replaceText(node.body, `(${sourceCode.getText(node.body)})`);
                        }
                    });
                }
            }
            return {
                ArrowFunctionExpression: checkArrowFunc
            };
        }
    };
},
/* eslint/lib/rules/no-console.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the use of `console`",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-console"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allow: {
                            type: "array",
                            items: {
                                type: "string"
                            },
                            minItems: 1,
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpected: "Unexpected console statement."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const allowed = options.allow || [];
            /**
             * Checks whether the given reference is 'console' or not.
             * @param {eslint-scope.Reference} reference The reference to check.
             * @returns {boolean} `true` if the reference is 'console'.
             */
            function isConsole(reference) {
                const id = reference.identifier;
                return id && id.name === "console";
            }
            /**
             * Checks whether the property name of the given MemberExpression node
             * is allowed by options or not.
             * @param {ASTNode} node The MemberExpression node to check.
             * @returns {boolean} `true` if the property name of the node is allowed.
             */
            function isAllowed(node) {
                const propertyName = astUtils.getStaticPropertyName(node);
                return propertyName && allowed.indexOf(propertyName) !== -1;
            }
            /**
             * Checks whether the given reference is a member access which is not
             * allowed by options or not.
             * @param {eslint-scope.Reference} reference The reference to check.
             * @returns {boolean} `true` if the reference is a member access which
             *      is not allowed by options.
             */
            function isMemberAccessExceptAllowed(reference) {
                const node = reference.identifier;
                const parent = node.parent;
                return (parent.type === "MemberExpression" &&
                    parent.object === node &&
                    !isAllowed(parent));
            }
            /**
             * Reports the given reference as a violation.
             * @param {eslint-scope.Reference} reference The reference to report.
             * @returns {void}
             */
            function report(reference) {
                const node = reference.identifier.parent;
                context.report({
                    node,
                    loc: node.loc,
                    messageId: "unexpected"
                });
            }
            return {
                "Program:exit"() {
                    const scope = context.getScope();
                    const consoleVar = astUtils.getVariableByName(scope, "console");
                    const shadowed = consoleVar && consoleVar.defs.length > 0;
                    /*
                     * 'scope.through' includes all references to undefined
                     * variables. If the variable 'console' is not defined, it uses
                     * 'scope.through'.
                     */
                    const references = consoleVar
                        ? consoleVar.references
                        : scope.through.filter(isConsole);
                    if (!shadowed) {
                        references
                            .filter(isMemberAccessExceptAllowed)
                            .forEach(report);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-const-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow reassigning `const` variables",
                category: "ECMAScript 6",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-const-assign"
            },
            schema: [],
            messages: {
                const: "'{{name}}' is constant."
            }
        },
        create(context) {
            /**
             * Finds and reports references that are non initializer and writable.
             * @param {Variable} variable A variable to check.
             * @returns {void}
             */
            function checkVariable(variable) {
                astUtils.getModifyingReferences(variable.references).forEach(reference => {
                    context.report({ node: reference.identifier, messageId: "const", data: { name: reference.identifier.name } });
                });
            }
            return {
                VariableDeclaration(node) {
                    if (node.kind === "const") {
                        context.getDeclaredVariables(node).forEach(checkVariable);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-constant-condition.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow constant expressions in conditions",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-constant-condition"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        checkLoops: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpected: "Unexpected constant condition."
            }
        },
        create(context) {
            const options = context.options[0] || {}, checkLoops = options.checkLoops !== false, loopSetStack = [];
            let loopsInCurrentScope = new Set();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Returns literal's value converted to the Boolean type
             * @param {ASTNode} node any `Literal` node
             * @returns {boolean | null} `true` when node is truthy, `false` when node is falsy,
             *  `null` when it cannot be determined.
             */
            function getBooleanValue(node) {
                if (node.value === null) {
                    /*
                     * it might be a null literal or bigint/regex literal in unsupported environments .
                     * https://github.com/estree/estree/blob/14df8a024956ea289bd55b9c2226a1d5b8a473ee/es5.md#regexpliteral
                     * https://github.com/estree/estree/blob/14df8a024956ea289bd55b9c2226a1d5b8a473ee/es2020.md#bigintliteral
                     */
                    if (node.raw === "null") {
                        return false;
                    }
                    // regex is always truthy
                    if (typeof node.regex === "object") {
                        return true;
                    }
                    return null;
                }
                return !!node.value;
            }
            /**
             * Checks if a branch node of LogicalExpression short circuits the whole condition
             * @param {ASTNode} node The branch of main condition which needs to be checked
             * @param {string} operator The operator of the main LogicalExpression.
             * @returns {boolean} true when condition short circuits whole condition
             */
            function isLogicalIdentity(node, operator) {
                switch (node.type) {
                    case "Literal":
                        return (operator === "||" && getBooleanValue(node) === true) ||
                            (operator === "&&" && getBooleanValue(node) === false);
                    case "UnaryExpression":
                        return (operator === "&&" && node.operator === "void");
                    case "LogicalExpression":
                        /*
                         * handles `a && false || b`
                         * `false` is an identity element of `&&` but not `||`
                         */
                        return operator === node.operator &&
                            (isLogicalIdentity(node.left, operator) ||
                                isLogicalIdentity(node.right, operator));
                    case "AssignmentExpression":
                        return ["||=", "&&="].includes(node.operator) &&
                            operator === node.operator.slice(0, -1) &&
                            isLogicalIdentity(node.right, operator);
                    // no default
                }
                return false;
            }
            /**
             * Checks if a node has a constant truthiness value.
             * @param {ASTNode} node The AST node to check.
             * @param {boolean} inBooleanPosition `false` if checking branch of a condition.
             *  `true` in all other cases
             * @returns {Bool} true when node's truthiness is constant
             * @private
             */
            function isConstant(node, inBooleanPosition) {
                // node.elements can return null values in the case of sparse arrays ex. [,]
                if (!node) {
                    return true;
                }
                switch (node.type) {
                    case "Literal":
                    case "ArrowFunctionExpression":
                    case "FunctionExpression":
                    case "ObjectExpression":
                        return true;
                    case "TemplateLiteral":
                        return (inBooleanPosition && node.quasis.some(quasi => quasi.value.cooked.length)) ||
                            node.expressions.every(exp => isConstant(exp, inBooleanPosition));
                    case "ArrayExpression": {
                        if (node.parent.type === "BinaryExpression" && node.parent.operator === "+") {
                            return node.elements.every(element => isConstant(element, false));
                        }
                        return true;
                    }
                    case "UnaryExpression":
                        if (node.operator === "void" ||
                            node.operator === "typeof" && inBooleanPosition) {
                            return true;
                        }
                        if (node.operator === "!") {
                            return isConstant(node.argument, true);
                        }
                        return isConstant(node.argument, false);
                    case "BinaryExpression":
                        return isConstant(node.left, false) &&
                            isConstant(node.right, false) &&
                            node.operator !== "in";
                    case "LogicalExpression": {
                        const isLeftConstant = isConstant(node.left, inBooleanPosition);
                        const isRightConstant = isConstant(node.right, inBooleanPosition);
                        const isLeftShortCircuit = (isLeftConstant && isLogicalIdentity(node.left, node.operator));
                        const isRightShortCircuit = (inBooleanPosition && isRightConstant && isLogicalIdentity(node.right, node.operator));
                        return (isLeftConstant && isRightConstant) ||
                            isLeftShortCircuit ||
                            isRightShortCircuit;
                    }
                    case "AssignmentExpression":
                        if (node.operator === "=") {
                            return isConstant(node.right, inBooleanPosition);
                        }
                        if (["||=", "&&="].includes(node.operator) && inBooleanPosition) {
                            return isLogicalIdentity(node.right, node.operator.slice(0, -1));
                        }
                        return false;
                    case "SequenceExpression":
                        return isConstant(node.expressions[node.expressions.length - 1], inBooleanPosition);
                    // no default
                }
                return false;
            }
            /**
             * Tracks when the given node contains a constant condition.
             * @param {ASTNode} node The AST node to check.
             * @returns {void}
             * @private
             */
            function trackConstantConditionLoop(node) {
                if (node.test && isConstant(node.test, true)) {
                    loopsInCurrentScope.add(node);
                }
            }
            /**
             * Reports when the set contains the given constant condition node
             * @param {ASTNode} node The AST node to check.
             * @returns {void}
             * @private
             */
            function checkConstantConditionLoopInSet(node) {
                if (loopsInCurrentScope.has(node)) {
                    loopsInCurrentScope.delete(node);
                    context.report({ node: node.test, messageId: "unexpected" });
                }
            }
            /**
             * Reports when the given node contains a constant condition.
             * @param {ASTNode} node The AST node to check.
             * @returns {void}
             * @private
             */
            function reportIfConstant(node) {
                if (node.test && isConstant(node.test, true)) {
                    context.report({ node: node.test, messageId: "unexpected" });
                }
            }
            /**
             * Stores current set of constant loops in loopSetStack temporarily
             * and uses a new set to track constant loops
             * @returns {void}
             * @private
             */
            function enterFunction() {
                loopSetStack.push(loopsInCurrentScope);
                loopsInCurrentScope = new Set();
            }
            /**
             * Reports when the set still contains stored constant conditions
             * @returns {void}
             * @private
             */
            function exitFunction() {
                loopsInCurrentScope = loopSetStack.pop();
            }
            /**
             * Checks node when checkLoops option is enabled
             * @param {ASTNode} node The AST node to check.
             * @returns {void}
             * @private
             */
            function checkLoop(node) {
                if (checkLoops) {
                    trackConstantConditionLoop(node);
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                ConditionalExpression: reportIfConstant,
                IfStatement: reportIfConstant,
                WhileStatement: checkLoop,
                "WhileStatement:exit": checkConstantConditionLoopInSet,
                DoWhileStatement: checkLoop,
                "DoWhileStatement:exit": checkConstantConditionLoopInSet,
                ForStatement: checkLoop,
                "ForStatement > .test": node => checkLoop(node.parent),
                "ForStatement:exit": checkConstantConditionLoopInSet,
                FunctionDeclaration: enterFunction,
                "FunctionDeclaration:exit": exitFunction,
                FunctionExpression: enterFunction,
                "FunctionExpression:exit": exitFunction,
                YieldExpression: () => loopsInCurrentScope.clear()
            };
        }
    };
},
/* eslint/lib/rules/no-constructor-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow returning value from constructor",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-constructor-return"
            },
            schema: {},
            fixable: null,
            messages: {
                unexpected: "Unexpected return statement in constructor."
            }
        },
        create(context) {
            const stack = [];
            return {
                onCodePathStart(_, node) {
                    stack.push(node);
                },
                onCodePathEnd() {
                    stack.pop();
                },
                ReturnStatement(node) {
                    const last = stack[stack.length - 1];
                    if (!last.parent) {
                        return;
                    }
                    if (last.parent.type === "MethodDefinition" &&
                        last.parent.kind === "constructor" &&
                        (node.parent.parent === last || node.argument)) {
                        context.report({
                            node,
                            messageId: "unexpected"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-continue.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `continue` statements",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-continue"
            },
            schema: [],
            messages: {
                unexpected: "Unexpected use of continue statement."
            }
        },
        create(context) {
            return {
                ContinueStatement(node) {
                    context.report({ node, messageId: "unexpected" });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-control-regex.js */ function _(require, module, exports, __esModule, __esExport) {
    const RegExpValidator = require(497) /* regexpp */.RegExpValidator;
    const collector = new (class {
        constructor() {
            this._source = "";
            this._controlChars = [];
            this._validator = new RegExpValidator(this);
        }
        onPatternEnter() {
            this._controlChars = [];
        }
        onCharacter(start, end, cp) {
            if (cp >= 0x00 &&
                cp <= 0x1F &&
                (this._source.codePointAt(start) === cp ||
                    this._source.slice(start, end).startsWith("\\x") ||
                    this._source.slice(start, end).startsWith("\\u"))) {
                this._controlChars.push(`\\x${`0${cp.toString(16)}`.slice(-2)}`);
            }
        }
        collectControlChars(regexpStr) {
            try {
                this._source = regexpStr;
                this._validator.validatePattern(regexpStr); // Call onCharacter hook
            }
            catch {
                // Ignore syntax errors in RegExp.
            }
            return this._controlChars;
        }
    })();
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow control characters in regular expressions",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-control-regex"
            },
            schema: [],
            messages: {
                unexpected: "Unexpected control character(s) in regular expression: {{controlChars}}."
            }
        },
        create(context) {
            /**
             * Get the regex expression
             * @param {ASTNode} node node to evaluate
             * @returns {RegExp|null} Regex if found else null
             * @private
             */
            function getRegExpPattern(node) {
                if (node.regex) {
                    return node.regex.pattern;
                }
                if (typeof node.value === "string" &&
                    (node.parent.type === "NewExpression" || node.parent.type === "CallExpression") &&
                    node.parent.callee.type === "Identifier" &&
                    node.parent.callee.name === "RegExp" &&
                    node.parent.arguments[0] === node) {
                    return node.value;
                }
                return null;
            }
            return {
                Literal(node) {
                    const pattern = getRegExpPattern(node);
                    if (pattern) {
                        const controlCharacters = collector.collectControlChars(pattern);
                        if (controlCharacters.length > 0) {
                            context.report({
                                node,
                                messageId: "unexpected",
                                data: {
                                    controlChars: controlCharacters.join(", ")
                                }
                            });
                        }
                    }
                }
            };
        }
    };
},
/* regexpp/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var ast = /*#__PURE__*/ Object.freeze({});
    let largeIdStartRanges = undefined;
    let largeIdContinueRanges = undefined;
    function isIdStart(cp) {
        if (cp < 0x41)
            return false;
        if (cp < 0x5b)
            return true;
        if (cp < 0x61)
            return false;
        if (cp < 0x7b)
            return true;
        return isLargeIdStart(cp);
    }
    function isIdContinue(cp) {
        if (cp < 0x30)
            return false;
        if (cp < 0x3a)
            return true;
        if (cp < 0x41)
            return false;
        if (cp < 0x5b)
            return true;
        if (cp === 0x5f)
            return true;
        if (cp < 0x61)
            return false;
        if (cp < 0x7b)
            return true;
        return isLargeIdStart(cp) || isLargeIdContinue(cp);
    }
    function isLargeIdStart(cp) {
        return isInRange(cp, largeIdStartRanges || (largeIdStartRanges = initLargeIdStartRanges()));
    }
    function isLargeIdContinue(cp) {
        return isInRange(cp, largeIdContinueRanges ||
            (largeIdContinueRanges = initLargeIdContinueRanges()));
    }
    function initLargeIdStartRanges() {
        return restoreRanges("4q 0 b 0 5 0 6 m 2 u 2 cp 5 b f 4 8 0 2 0 3m 4 2 1 3 3 2 0 7 0 2 2 2 0 2 j 2 2a 2 3u 9 4l 2 11 3 0 7 14 20 q 5 3 1a 16 10 1 2 2q 2 0 g 1 8 1 b 2 3 0 h 0 2 t u 2g c 0 p w a 1 5 0 6 l 5 0 a 0 4 0 o o 8 a 1i k 2 h 1p 1h 4 0 j 0 8 9 g f 5 7 3 1 3 l 2 6 2 0 4 3 4 0 h 0 e 1 2 2 f 1 b 0 9 5 5 1 3 l 2 6 2 1 2 1 2 1 w 3 2 0 k 2 h 8 2 2 2 l 2 6 2 1 2 4 4 0 j 0 g 1 o 0 c 7 3 1 3 l 2 6 2 1 2 4 4 0 v 1 2 2 g 0 i 0 2 5 4 2 2 3 4 1 2 0 2 1 4 1 4 2 4 b n 0 1h 7 2 2 2 m 2 f 4 0 r 2 6 1 v 0 5 7 2 2 2 m 2 9 2 4 4 0 x 0 2 1 g 1 i 8 2 2 2 14 3 0 h 0 6 2 9 2 p 5 6 h 4 n 2 8 2 0 3 6 1n 1b 2 1 d 6 1n 1 2 0 2 4 2 n 2 0 2 9 2 1 a 0 3 4 2 0 m 3 x 0 1s 7 2 z s 4 38 16 l 0 h 5 5 3 4 0 4 1 8 2 5 c d 0 i 11 2 0 6 0 3 16 2 98 2 3 3 6 2 0 2 3 3 14 2 3 3 w 2 3 3 6 2 0 2 3 3 e 2 1k 2 3 3 1u 12 f h 2d 3 5 4 h7 3 g 2 p 6 22 4 a 8 c 2 3 f h f h f c 2 2 g 1f 10 0 5 0 1w 2g 8 14 2 0 6 1x b u 1e t 3 4 c 17 5 p 1j m a 1g 2b 0 2m 1a i 6 1k t e 1 b 17 r z 16 2 b z 3 8 8 16 3 2 16 3 2 5 2 1 4 0 6 5b 1t 7p 3 5 3 11 3 5 3 7 2 0 2 0 2 0 2 u 3 1g 2 6 2 0 4 2 2 6 4 3 3 5 5 c 6 2 2 6 39 0 e 0 h c 2u 0 5 0 3 9 2 0 3 5 7 0 2 0 2 0 2 f 3 3 6 4 5 0 i 14 22g 1a 2 1a 2 3o 7 3 4 1 d 11 2 0 6 0 3 1j 8 0 h m a 6 2 6 2 6 2 6 2 6 2 6 2 6 2 6 fb 2 q 8 8 4 3 4 5 2d 5 4 2 2h 2 3 6 16 2 2l i v 1d f e9 533 1t g70 4 wc 1w 19 3 7g 4 f b 1 l 1a h u 3 27 14 8 3 2u 3 1g 3 8 17 c 2 2 2 3 2 m u 1f f 1d 1r 5 4 0 2 1 c r b m q s 8 1a t 0 h 4 2 9 b 4 2 14 o 2 2 7 l m 4 0 4 1d 2 0 4 1 3 4 3 0 2 0 p 2 3 a 8 2 d 5 3 5 3 5 a 6 2 6 2 16 2 d 7 36 u 8mb d m 5 1c 6it a5 3 2x 13 6 d 4 6 0 2 9 2 c 2 4 2 0 2 1 2 1 2 2z y a2 j 1r 3 1h 15 b 39 4 2 3q 11 p 7 p c 2g 4 5 3 5 3 5 3 2 10 b 2 p 2 i 2 1 2 e 3 d z 3e 1y 1g 7g s 4 1c 1c v e t 6 11 b t 3 z 5 7 2 4 17 4d j z 5 z 5 13 9 1f 4d 8m a l b 7 49 5 3 0 2 17 2 1 4 0 3 m b m a u 1u i 2 1 b l b p 1z 1j 7 1 1t 0 g 3 2 2 2 s 17 s 4 s 10 7 2 r s 1h b l b i e h 33 20 1k 1e e 1e e z 9p 15 7 1 27 s b 0 9 l 2z k s m d 1g 24 18 x o r z u 0 3 0 9 y 4 0 d 1b f 3 m 0 2 0 10 h 2 o 2d 6 2 0 2 3 2 e 2 9 8 1a 13 7 3 1 3 l 2 6 2 1 2 4 4 0 j 0 d 4 4f 1g j 3 l 2 v 1b l 1 2 0 55 1a 16 3 11 1b l 0 1o 16 e 0 20 q 6e 17 39 1r w 7 3 0 3 7 2 1 2 n g 0 2 0 2n 7 3 12 h 0 2 0 t 0 b 13 8 0 m 0 c 19 k 0 z 1k 7c 8 2 10 i 0 1e t 35 6 2 1 2 11 m 0 q 5 2 1 2 v f 0 94 i 5a 0 28 pl 2v 32 i 5f 24d tq 34i g6 6nu fs 8 u 36 t j 1b h 3 w k 6 i j5 1r 3l 22 6 0 1v c 1t 1 2 0 t 4qf 9 yd 17 8 6wo 7y 1e 2 i 3 9 az 1s5 2y 6 c 4 8 8 9 4mf 2c 2 1y 2 1 3 0 3 1 3 3 2 b 2 0 2 6 2 1s 2 3 3 7 2 6 2 r 2 3 2 4 2 0 4 6 2 9f 3 o 2 o 2 u 2 o 2 u 2 o 2 u 2 o 2 u 2 o 2 7 1th 18 b 6 h 0 aa 17 105 5g 1o 1v 8 0 xh 3 2 q 2 1 2 0 3 0 2 9 2 3 2 0 2 0 7 0 5 0 2 0 2 0 2 2 2 1 2 0 3 0 2 0 2 0 2 0 2 0 2 1 2 0 3 3 2 6 2 3 2 3 2 0 2 9 2 g 6 2 2 4 2 g 3et wyl z 378 c 65 3 4g1 f 5rk 2e8 f1 15v 3t6");
    }
    function initLargeIdContinueRanges() {
        return restoreRanges("53 0 g9 33 o 0 70 4 7e 18 2 0 2 1 2 1 2 0 21 a 1d u 7 0 2u 6 3 5 3 1 2 3 3 9 o 0 v q 2k a g 9 y 8 a 0 p 3 2 8 2 2 2 4 18 2 3c e 2 w 1j 2 2 h 2 6 b 1 3 9 i 2 1l 0 2 6 3 1 3 2 a 0 b 1 3 9 f 0 3 2 1l 0 2 4 5 1 3 2 4 0 l b 4 0 c 2 1l 0 2 7 2 2 2 2 l 1 3 9 b 5 2 2 1l 0 2 6 3 1 3 2 8 2 b 1 3 9 j 0 1o 4 4 2 2 3 a 0 f 9 h 4 1m 6 2 2 2 3 8 1 c 1 3 9 i 2 1l 0 2 6 2 2 2 3 8 1 c 1 3 9 h 3 1k 1 2 6 2 2 2 3 a 0 b 1 3 9 i 2 1z 0 5 5 2 0 2 7 7 9 3 1 1q 0 3 6 d 7 2 9 2g 0 3 8 c 5 3 9 1r 1 7 9 c 0 2 0 2 0 5 1 1e j 2 1 6 a 2 z a 0 2t j 2 9 d 3 5 2 2 2 3 6 4 3 e b 2 e jk 2 a 8 pt 2 u 2 u 1 v 1 1t v a 0 3 9 y 2 3 9 40 0 3b b 5 b b 9 3l a 1p 4 1m 9 2 s 3 a 7 9 n d 2 1 1s 4 1c g c 9 i 8 d 2 v c 3 9 19 d 1d j 9 9 7 9 3b 2 2 k 5 0 7 0 3 2 5j 1l 2 4 g0 1 k 0 3g c 5 0 4 b 2db 2 3y 0 2p v ff 5 2y 1 n7q 9 1y 0 5 9 x 1 29 1 7l 0 4 0 5 0 o 4 5 0 2c 1 1f h b 9 7 h e a t 7 q c 19 3 1c d g 9 c 0 b 9 1c d d 0 9 1 3 9 y 2 1f 0 2 2 3 1 6 1 2 0 16 4 6 1 6l 7 2 1 3 9 fmt 0 ki f h f 4 1 p 2 5d 9 12 0 ji 0 6b 0 46 4 86 9 120 2 2 1 6 3 15 2 5 0 4m 1 fy 3 9 9 aa 1 4a a 4w 2 1i e w 9 g 3 1a a 1i 9 7 2 11 d 2 9 6 1 19 0 d 2 1d d 9 3 2 b 2b b 7 0 4h b 6 9 7 3 1k 1 2 6 3 1 3 2 a 0 b 1 3 6 4 4 5d h a 9 5 0 2a j d 9 5y 6 3 8 s 1 2b g g 9 2a c 9 9 2c e 5 9 6r e 4m 9 1z 5 2 1 3 3 2 0 2 1 d 9 3c 6 3 6 4 0 t 9 15 6 2 3 9 0 a a 1b f ba 7 2 7 h 9 1l l 2 d 3f 5 4 0 2 1 2 6 2 0 9 9 1d 4 2 1 2 4 9 9 96 3 ewa 9 3r 4 1o 6 q 9 s6 0 2 1i 8 3 2a 0 c 1 f58 1 43r 4 4 5 9 7 3 6 v 3 45 2 13e 1d e9 1i 5 1d 9 0 f 0 n 4 2 e 11t 6 2 g 3 6 2 1 2 4 7a 6 a 9 bn d 15j 6 32 6 6 9 3o7 9 gvt3 6n");
    }
    function isInRange(cp, ranges) {
        let l = 0, r = (ranges.length / 2) | 0, i = 0, min = 0, max = 0;
        while (l < r) {
            i = ((l + r) / 2) | 0;
            min = ranges[2 * i];
            max = ranges[2 * i + 1];
            if (cp < min) {
                r = i;
            }
            else if (cp > max) {
                l = i + 1;
            }
            else {
                return true;
            }
        }
        return false;
    }
    function restoreRanges(data) {
        let last = 0;
        return data.split(" ").map(s => (last += parseInt(s, 36) | 0));
    }
    class DataSet {
        constructor(raw2018, raw2019, raw2020, raw2021) {
            this._raw2018 = raw2018;
            this._raw2019 = raw2019;
            this._raw2020 = raw2020;
            this._raw2021 = raw2021;
        }
        get es2018() {
            return (this._set2018 || (this._set2018 = new Set(this._raw2018.split(" "))));
        }
        get es2019() {
            return (this._set2019 || (this._set2019 = new Set(this._raw2019.split(" "))));
        }
        get es2020() {
            return (this._set2020 || (this._set2020 = new Set(this._raw2020.split(" "))));
        }
        get es2021() {
            return (this._set2021 || (this._set2021 = new Set(this._raw2021.split(" "))));
        }
    }
    const gcNameSet = new Set(["General_Category", "gc"]);
    const scNameSet = new Set(["Script", "Script_Extensions", "sc", "scx"]);
    const gcValueSets = new DataSet("C Cased_Letter Cc Cf Close_Punctuation Cn Co Combining_Mark Connector_Punctuation Control Cs Currency_Symbol Dash_Punctuation Decimal_Number Enclosing_Mark Final_Punctuation Format Initial_Punctuation L LC Letter Letter_Number Line_Separator Ll Lm Lo Lowercase_Letter Lt Lu M Mark Math_Symbol Mc Me Mn Modifier_Letter Modifier_Symbol N Nd Nl No Nonspacing_Mark Number Open_Punctuation Other Other_Letter Other_Number Other_Punctuation Other_Symbol P Paragraph_Separator Pc Pd Pe Pf Pi Po Private_Use Ps Punctuation S Sc Separator Sk Sm So Space_Separator Spacing_Mark Surrogate Symbol Titlecase_Letter Unassigned Uppercase_Letter Z Zl Zp Zs cntrl digit punct", "", "", "");
    const scValueSets = new DataSet("Adlam Adlm Aghb Ahom Anatolian_Hieroglyphs Arab Arabic Armenian Armi Armn Avestan Avst Bali Balinese Bamu Bamum Bass Bassa_Vah Batak Batk Beng Bengali Bhaiksuki Bhks Bopo Bopomofo Brah Brahmi Brai Braille Bugi Buginese Buhd Buhid Cakm Canadian_Aboriginal Cans Cari Carian Caucasian_Albanian Chakma Cham Cher Cherokee Common Copt Coptic Cprt Cuneiform Cypriot Cyrillic Cyrl Deseret Deva Devanagari Dsrt Dupl Duployan Egyp Egyptian_Hieroglyphs Elba Elbasan Ethi Ethiopic Geor Georgian Glag Glagolitic Gonm Goth Gothic Gran Grantha Greek Grek Gujarati Gujr Gurmukhi Guru Han Hang Hangul Hani Hano Hanunoo Hatr Hatran Hebr Hebrew Hira Hiragana Hluw Hmng Hung Imperial_Aramaic Inherited Inscriptional_Pahlavi Inscriptional_Parthian Ital Java Javanese Kaithi Kali Kana Kannada Katakana Kayah_Li Khar Kharoshthi Khmer Khmr Khoj Khojki Khudawadi Knda Kthi Lana Lao Laoo Latin Latn Lepc Lepcha Limb Limbu Lina Linb Linear_A Linear_B Lisu Lyci Lycian Lydi Lydian Mahajani Mahj Malayalam Mand Mandaic Mani Manichaean Marc Marchen Masaram_Gondi Meetei_Mayek Mend Mende_Kikakui Merc Mero Meroitic_Cursive Meroitic_Hieroglyphs Miao Mlym Modi Mong Mongolian Mro Mroo Mtei Mult Multani Myanmar Mymr Nabataean Narb Nbat New_Tai_Lue Newa Nko Nkoo Nshu Nushu Ogam Ogham Ol_Chiki Olck Old_Hungarian Old_Italic Old_North_Arabian Old_Permic Old_Persian Old_South_Arabian Old_Turkic Oriya Orkh Orya Osage Osge Osma Osmanya Pahawh_Hmong Palm Palmyrene Pau_Cin_Hau Pauc Perm Phag Phags_Pa Phli Phlp Phnx Phoenician Plrd Prti Psalter_Pahlavi Qaac Qaai Rejang Rjng Runic Runr Samaritan Samr Sarb Saur Saurashtra Sgnw Sharada Shavian Shaw Shrd Sidd Siddham SignWriting Sind Sinh Sinhala Sora Sora_Sompeng Soyo Soyombo Sund Sundanese Sylo Syloti_Nagri Syrc Syriac Tagalog Tagb Tagbanwa Tai_Le Tai_Tham Tai_Viet Takr Takri Tale Talu Tamil Taml Tang Tangut Tavt Telu Telugu Tfng Tglg Thaa Thaana Thai Tibetan Tibt Tifinagh Tirh Tirhuta Ugar Ugaritic Vai Vaii Wara Warang_Citi Xpeo Xsux Yi Yiii Zanabazar_Square Zanb Zinh Zyyy", "Dogr Dogra Gong Gunjala_Gondi Hanifi_Rohingya Maka Makasar Medefaidrin Medf Old_Sogdian Rohg Sogd Sogdian Sogo", "Elym Elymaic Hmnp Nand Nandinagari Nyiakeng_Puachue_Hmong Wancho Wcho", "Chorasmian Chrs Diak Dives_Akuru Khitan_Small_Script Kits Yezi Yezidi");
    const binPropertySets = new DataSet("AHex ASCII ASCII_Hex_Digit Alpha Alphabetic Any Assigned Bidi_C Bidi_Control Bidi_M Bidi_Mirrored CI CWCF CWCM CWKCF CWL CWT CWU Case_Ignorable Cased Changes_When_Casefolded Changes_When_Casemapped Changes_When_Lowercased Changes_When_NFKC_Casefolded Changes_When_Titlecased Changes_When_Uppercased DI Dash Default_Ignorable_Code_Point Dep Deprecated Dia Diacritic Emoji Emoji_Component Emoji_Modifier Emoji_Modifier_Base Emoji_Presentation Ext Extender Gr_Base Gr_Ext Grapheme_Base Grapheme_Extend Hex Hex_Digit IDC IDS IDSB IDST IDS_Binary_Operator IDS_Trinary_Operator ID_Continue ID_Start Ideo Ideographic Join_C Join_Control LOE Logical_Order_Exception Lower Lowercase Math NChar Noncharacter_Code_Point Pat_Syn Pat_WS Pattern_Syntax Pattern_White_Space QMark Quotation_Mark RI Radical Regional_Indicator SD STerm Sentence_Terminal Soft_Dotted Term Terminal_Punctuation UIdeo Unified_Ideograph Upper Uppercase VS Variation_Selector White_Space XIDC XIDS XID_Continue XID_Start space", "Extended_Pictographic", "", "EBase EComp EMod EPres ExtPict");
    function isValidUnicodeProperty(version, name, value) {
        if (gcNameSet.has(name)) {
            return version >= 2018 && gcValueSets.es2018.has(value);
        }
        if (scNameSet.has(name)) {
            return ((version >= 2018 && scValueSets.es2018.has(value)) ||
                (version >= 2019 && scValueSets.es2019.has(value)) ||
                (version >= 2020 && scValueSets.es2020.has(value)) ||
                (version >= 2021 && scValueSets.es2021.has(value)));
        }
        return false;
    }
    function isValidLoneUnicodeProperty(version, value) {
        return ((version >= 2018 && binPropertySets.es2018.has(value)) ||
            (version >= 2019 && binPropertySets.es2019.has(value)) ||
            (version >= 2021 && binPropertySets.es2021.has(value)));
    }
    const Backspace = 0x08;
    const CharacterTabulation = 0x09;
    const LineFeed = 0x0a;
    const LineTabulation = 0x0b;
    const FormFeed = 0x0c;
    const CarriageReturn = 0x0d;
    const ExclamationMark = 0x21;
    const DollarSign = 0x24;
    const LeftParenthesis = 0x28;
    const RightParenthesis = 0x29;
    const Asterisk = 0x2a;
    const PlusSign = 0x2b;
    const Comma = 0x2c;
    const HyphenMinus = 0x2d;
    const FullStop = 0x2e;
    const Solidus = 0x2f;
    const DigitZero = 0x30;
    const DigitOne = 0x31;
    const DigitSeven = 0x37;
    const DigitNine = 0x39;
    const Colon = 0x3a;
    const LessThanSign = 0x3c;
    const EqualsSign = 0x3d;
    const GreaterThanSign = 0x3e;
    const QuestionMark = 0x3f;
    const LatinCapitalLetterA = 0x41;
    const LatinCapitalLetterB = 0x42;
    const LatinCapitalLetterD = 0x44;
    const LatinCapitalLetterF = 0x46;
    const LatinCapitalLetterP = 0x50;
    const LatinCapitalLetterS = 0x53;
    const LatinCapitalLetterW = 0x57;
    const LatinCapitalLetterZ = 0x5a;
    const LowLine = 0x5f;
    const LatinSmallLetterA = 0x61;
    const LatinSmallLetterB = 0x62;
    const LatinSmallLetterC = 0x63;
    const LatinSmallLetterD = 0x64;
    const LatinSmallLetterF = 0x66;
    const LatinSmallLetterG = 0x67;
    const LatinSmallLetterI = 0x69;
    const LatinSmallLetterK = 0x6b;
    const LatinSmallLetterM = 0x6d;
    const LatinSmallLetterN = 0x6e;
    const LatinSmallLetterP = 0x70;
    const LatinSmallLetterR = 0x72;
    const LatinSmallLetterS = 0x73;
    const LatinSmallLetterT = 0x74;
    const LatinSmallLetterU = 0x75;
    const LatinSmallLetterV = 0x76;
    const LatinSmallLetterW = 0x77;
    const LatinSmallLetterX = 0x78;
    const LatinSmallLetterY = 0x79;
    const LatinSmallLetterZ = 0x7a;
    const LeftSquareBracket = 0x5b;
    const ReverseSolidus = 0x5c;
    const RightSquareBracket = 0x5d;
    const CircumflexAccent = 0x5e;
    const LeftCurlyBracket = 0x7b;
    const VerticalLine = 0x7c;
    const RightCurlyBracket = 0x7d;
    const ZeroWidthNonJoiner = 0x200c;
    const ZeroWidthJoiner = 0x200d;
    const LineSeparator = 0x2028;
    const ParagraphSeparator = 0x2029;
    const MinCodePoint = 0x00;
    const MaxCodePoint = 0x10ffff;
    function isLatinLetter(code) {
        return ((code >= LatinCapitalLetterA && code <= LatinCapitalLetterZ) ||
            (code >= LatinSmallLetterA && code <= LatinSmallLetterZ));
    }
    function isDecimalDigit(code) {
        return code >= DigitZero && code <= DigitNine;
    }
    function isOctalDigit(code) {
        return code >= DigitZero && code <= DigitSeven;
    }
    function isHexDigit(code) {
        return ((code >= DigitZero && code <= DigitNine) ||
            (code >= LatinCapitalLetterA && code <= LatinCapitalLetterF) ||
            (code >= LatinSmallLetterA && code <= LatinSmallLetterF));
    }
    function isLineTerminator(code) {
        return (code === LineFeed ||
            code === CarriageReturn ||
            code === LineSeparator ||
            code === ParagraphSeparator);
    }
    function isValidUnicode(code) {
        return code >= MinCodePoint && code <= MaxCodePoint;
    }
    function digitToInt(code) {
        if (code >= LatinSmallLetterA && code <= LatinSmallLetterF) {
            return code - LatinSmallLetterA + 10;
        }
        if (code >= LatinCapitalLetterA && code <= LatinCapitalLetterF) {
            return code - LatinCapitalLetterA + 10;
        }
        return code - DigitZero;
    }
    function isLeadSurrogate(code) {
        return code >= 0xd800 && code <= 0xdbff;
    }
    function isTrailSurrogate(code) {
        return code >= 0xdc00 && code <= 0xdfff;
    }
    function combineSurrogatePair(lead, trail) {
        return (lead - 0xd800) * 0x400 + (trail - 0xdc00) + 0x10000;
    }
    const legacyImpl = {
        at(s, end, i) {
            return i < end ? s.charCodeAt(i) : -1;
        },
        width(c) {
            return 1;
        },
    };
    const unicodeImpl = {
        at(s, end, i) {
            return i < end ? s.codePointAt(i) : -1;
        },
        width(c) {
            return c > 0xffff ? 2 : 1;
        },
    };
    class Reader {
        constructor() {
            this._impl = legacyImpl;
            this._s = "";
            this._i = 0;
            this._end = 0;
            this._cp1 = -1;
            this._w1 = 1;
            this._cp2 = -1;
            this._w2 = 1;
            this._cp3 = -1;
            this._w3 = 1;
            this._cp4 = -1;
        }
        get source() {
            return this._s;
        }
        get index() {
            return this._i;
        }
        get currentCodePoint() {
            return this._cp1;
        }
        get nextCodePoint() {
            return this._cp2;
        }
        get nextCodePoint2() {
            return this._cp3;
        }
        get nextCodePoint3() {
            return this._cp4;
        }
        reset(source, start, end, uFlag) {
            this._impl = uFlag ? unicodeImpl : legacyImpl;
            this._s = source;
            this._end = end;
            this.rewind(start);
        }
        rewind(index) {
            const impl = this._impl;
            this._i = index;
            this._cp1 = impl.at(this._s, this._end, index);
            this._w1 = impl.width(this._cp1);
            this._cp2 = impl.at(this._s, this._end, index + this._w1);
            this._w2 = impl.width(this._cp2);
            this._cp3 = impl.at(this._s, this._end, index + this._w1 + this._w2);
            this._w3 = impl.width(this._cp3);
            this._cp4 = impl.at(this._s, this._end, index + this._w1 + this._w2 + this._w3);
        }
        advance() {
            if (this._cp1 !== -1) {
                const impl = this._impl;
                this._i += this._w1;
                this._cp1 = this._cp2;
                this._w1 = this._w2;
                this._cp2 = this._cp3;
                this._w2 = impl.width(this._cp2);
                this._cp3 = this._cp4;
                this._w3 = impl.width(this._cp3);
                this._cp4 = impl.at(this._s, this._end, this._i + this._w1 + this._w2 + this._w3);
            }
        }
        eat(cp) {
            if (this._cp1 === cp) {
                this.advance();
                return true;
            }
            return false;
        }
        eat2(cp1, cp2) {
            if (this._cp1 === cp1 && this._cp2 === cp2) {
                this.advance();
                this.advance();
                return true;
            }
            return false;
        }
        eat3(cp1, cp2, cp3) {
            if (this._cp1 === cp1 && this._cp2 === cp2 && this._cp3 === cp3) {
                this.advance();
                this.advance();
                this.advance();
                return true;
            }
            return false;
        }
    }
    class RegExpSyntaxError extends SyntaxError {
        constructor(source, uFlag, index, message) {
            if (source) {
                if (!source.startsWith("/")) {
                    source = `/${source}/${uFlag ? "u" : ""}`;
                }
                source = `: ${source}`;
            }
            super(`Invalid regular expression${source}: ${message}`);
            this.index = index;
        }
    }
    function isSyntaxCharacter(cp) {
        return (cp === CircumflexAccent ||
            cp === DollarSign ||
            cp === ReverseSolidus ||
            cp === FullStop ||
            cp === Asterisk ||
            cp === PlusSign ||
            cp === QuestionMark ||
            cp === LeftParenthesis ||
            cp === RightParenthesis ||
            cp === LeftSquareBracket ||
            cp === RightSquareBracket ||
            cp === LeftCurlyBracket ||
            cp === RightCurlyBracket ||
            cp === VerticalLine);
    }
    function isRegExpIdentifierStart(cp) {
        return isIdStart(cp) || cp === DollarSign || cp === LowLine;
    }
    function isRegExpIdentifierPart(cp) {
        return (isIdContinue(cp) ||
            cp === DollarSign ||
            cp === LowLine ||
            cp === ZeroWidthNonJoiner ||
            cp === ZeroWidthJoiner);
    }
    function isUnicodePropertyNameCharacter(cp) {
        return isLatinLetter(cp) || cp === LowLine;
    }
    function isUnicodePropertyValueCharacter(cp) {
        return isUnicodePropertyNameCharacter(cp) || isDecimalDigit(cp);
    }
    class RegExpValidator {
        constructor(options) {
            this._reader = new Reader();
            this._uFlag = false;
            this._nFlag = false;
            this._lastIntValue = 0;
            this._lastMinValue = 0;
            this._lastMaxValue = 0;
            this._lastStrValue = "";
            this._lastKeyValue = "";
            this._lastValValue = "";
            this._lastAssertionIsQuantifiable = false;
            this._numCapturingParens = 0;
            this._groupNames = new Set();
            this._backreferenceNames = new Set();
            this._options = options || {};
        }
        validateLiteral(source, start = 0, end = source.length) {
            this._uFlag = this._nFlag = false;
            this.reset(source, start, end);
            this.onLiteralEnter(start);
            if (this.eat(Solidus) && this.eatRegExpBody() && this.eat(Solidus)) {
                const flagStart = this.index;
                const uFlag = source.includes("u", flagStart);
                this.validateFlags(source, flagStart, end);
                this.validatePattern(source, start + 1, flagStart - 1, uFlag);
            }
            else if (start >= end) {
                this.raise("Empty");
            }
            else {
                const c = String.fromCodePoint(this.currentCodePoint);
                this.raise(`Unexpected character '${c}'`);
            }
            this.onLiteralLeave(start, end);
        }
        validateFlags(source, start = 0, end = source.length) {
            const existingFlags = new Set();
            let global = false;
            let ignoreCase = false;
            let multiline = false;
            let sticky = false;
            let unicode = false;
            let dotAll = false;
            let hasIndices = false;
            for (let i = start; i < end; ++i) {
                const flag = source.charCodeAt(i);
                if (existingFlags.has(flag)) {
                    this.raise(`Duplicated flag '${source[i]}'`);
                }
                existingFlags.add(flag);
                if (flag === LatinSmallLetterG) {
                    global = true;
                }
                else if (flag === LatinSmallLetterI) {
                    ignoreCase = true;
                }
                else if (flag === LatinSmallLetterM) {
                    multiline = true;
                }
                else if (flag === LatinSmallLetterU && this.ecmaVersion >= 2015) {
                    unicode = true;
                }
                else if (flag === LatinSmallLetterY && this.ecmaVersion >= 2015) {
                    sticky = true;
                }
                else if (flag === LatinSmallLetterS && this.ecmaVersion >= 2018) {
                    dotAll = true;
                }
                else if (flag === LatinSmallLetterD && this.ecmaVersion >= 2022) {
                    hasIndices = true;
                }
                else {
                    this.raise(`Invalid flag '${source[i]}'`);
                }
            }
            this.onFlags(start, end, global, ignoreCase, multiline, unicode, sticky, dotAll, hasIndices);
        }
        validatePattern(source, start = 0, end = source.length, uFlag = false) {
            this._uFlag = uFlag && this.ecmaVersion >= 2015;
            this._nFlag = uFlag && this.ecmaVersion >= 2018;
            this.reset(source, start, end);
            this.consumePattern();
            if (!this._nFlag &&
                this.ecmaVersion >= 2018 &&
                this._groupNames.size > 0) {
                this._nFlag = true;
                this.rewind(start);
                this.consumePattern();
            }
        }
        get strict() {
            return Boolean(this._options.strict || this._uFlag);
        }
        get ecmaVersion() {
            return this._options.ecmaVersion || 2022;
        }
        onLiteralEnter(start) {
            if (this._options.onLiteralEnter) {
                this._options.onLiteralEnter(start);
            }
        }
        onLiteralLeave(start, end) {
            if (this._options.onLiteralLeave) {
                this._options.onLiteralLeave(start, end);
            }
        }
        onFlags(start, end, global, ignoreCase, multiline, unicode, sticky, dotAll, hasIndices) {
            if (this._options.onFlags) {
                this._options.onFlags(start, end, global, ignoreCase, multiline, unicode, sticky, dotAll, hasIndices);
            }
        }
        onPatternEnter(start) {
            if (this._options.onPatternEnter) {
                this._options.onPatternEnter(start);
            }
        }
        onPatternLeave(start, end) {
            if (this._options.onPatternLeave) {
                this._options.onPatternLeave(start, end);
            }
        }
        onDisjunctionEnter(start) {
            if (this._options.onDisjunctionEnter) {
                this._options.onDisjunctionEnter(start);
            }
        }
        onDisjunctionLeave(start, end) {
            if (this._options.onDisjunctionLeave) {
                this._options.onDisjunctionLeave(start, end);
            }
        }
        onAlternativeEnter(start, index) {
            if (this._options.onAlternativeEnter) {
                this._options.onAlternativeEnter(start, index);
            }
        }
        onAlternativeLeave(start, end, index) {
            if (this._options.onAlternativeLeave) {
                this._options.onAlternativeLeave(start, end, index);
            }
        }
        onGroupEnter(start) {
            if (this._options.onGroupEnter) {
                this._options.onGroupEnter(start);
            }
        }
        onGroupLeave(start, end) {
            if (this._options.onGroupLeave) {
                this._options.onGroupLeave(start, end);
            }
        }
        onCapturingGroupEnter(start, name) {
            if (this._options.onCapturingGroupEnter) {
                this._options.onCapturingGroupEnter(start, name);
            }
        }
        onCapturingGroupLeave(start, end, name) {
            if (this._options.onCapturingGroupLeave) {
                this._options.onCapturingGroupLeave(start, end, name);
            }
        }
        onQuantifier(start, end, min, max, greedy) {
            if (this._options.onQuantifier) {
                this._options.onQuantifier(start, end, min, max, greedy);
            }
        }
        onLookaroundAssertionEnter(start, kind, negate) {
            if (this._options.onLookaroundAssertionEnter) {
                this._options.onLookaroundAssertionEnter(start, kind, negate);
            }
        }
        onLookaroundAssertionLeave(start, end, kind, negate) {
            if (this._options.onLookaroundAssertionLeave) {
                this._options.onLookaroundAssertionLeave(start, end, kind, negate);
            }
        }
        onEdgeAssertion(start, end, kind) {
            if (this._options.onEdgeAssertion) {
                this._options.onEdgeAssertion(start, end, kind);
            }
        }
        onWordBoundaryAssertion(start, end, kind, negate) {
            if (this._options.onWordBoundaryAssertion) {
                this._options.onWordBoundaryAssertion(start, end, kind, negate);
            }
        }
        onAnyCharacterSet(start, end, kind) {
            if (this._options.onAnyCharacterSet) {
                this._options.onAnyCharacterSet(start, end, kind);
            }
        }
        onEscapeCharacterSet(start, end, kind, negate) {
            if (this._options.onEscapeCharacterSet) {
                this._options.onEscapeCharacterSet(start, end, kind, negate);
            }
        }
        onUnicodePropertyCharacterSet(start, end, kind, key, value, negate) {
            if (this._options.onUnicodePropertyCharacterSet) {
                this._options.onUnicodePropertyCharacterSet(start, end, kind, key, value, negate);
            }
        }
        onCharacter(start, end, value) {
            if (this._options.onCharacter) {
                this._options.onCharacter(start, end, value);
            }
        }
        onBackreference(start, end, ref) {
            if (this._options.onBackreference) {
                this._options.onBackreference(start, end, ref);
            }
        }
        onCharacterClassEnter(start, negate) {
            if (this._options.onCharacterClassEnter) {
                this._options.onCharacterClassEnter(start, negate);
            }
        }
        onCharacterClassLeave(start, end, negate) {
            if (this._options.onCharacterClassLeave) {
                this._options.onCharacterClassLeave(start, end, negate);
            }
        }
        onCharacterClassRange(start, end, min, max) {
            if (this._options.onCharacterClassRange) {
                this._options.onCharacterClassRange(start, end, min, max);
            }
        }
        get source() {
            return this._reader.source;
        }
        get index() {
            return this._reader.index;
        }
        get currentCodePoint() {
            return this._reader.currentCodePoint;
        }
        get nextCodePoint() {
            return this._reader.nextCodePoint;
        }
        get nextCodePoint2() {
            return this._reader.nextCodePoint2;
        }
        get nextCodePoint3() {
            return this._reader.nextCodePoint3;
        }
        reset(source, start, end) {
            this._reader.reset(source, start, end, this._uFlag);
        }
        rewind(index) {
            this._reader.rewind(index);
        }
        advance() {
            this._reader.advance();
        }
        eat(cp) {
            return this._reader.eat(cp);
        }
        eat2(cp1, cp2) {
            return this._reader.eat2(cp1, cp2);
        }
        eat3(cp1, cp2, cp3) {
            return this._reader.eat3(cp1, cp2, cp3);
        }
        raise(message) {
            throw new RegExpSyntaxError(this.source, this._uFlag, this.index, message);
        }
        eatRegExpBody() {
            const start = this.index;
            let inClass = false;
            let escaped = false;
            for (;;) {
                const cp = this.currentCodePoint;
                if (cp === -1 || isLineTerminator(cp)) {
                    const kind = inClass ? "character class" : "regular expression";
                    this.raise(`Unterminated ${kind}`);
                }
                if (escaped) {
                    escaped = false;
                }
                else if (cp === ReverseSolidus) {
                    escaped = true;
                }
                else if (cp === LeftSquareBracket) {
                    inClass = true;
                }
                else if (cp === RightSquareBracket) {
                    inClass = false;
                }
                else if ((cp === Solidus && !inClass) ||
                    (cp === Asterisk && this.index === start)) {
                    break;
                }
                this.advance();
            }
            return this.index !== start;
        }
        consumePattern() {
            const start = this.index;
            this._numCapturingParens = this.countCapturingParens();
            this._groupNames.clear();
            this._backreferenceNames.clear();
            this.onPatternEnter(start);
            this.consumeDisjunction();
            const cp = this.currentCodePoint;
            if (this.currentCodePoint !== -1) {
                if (cp === RightParenthesis) {
                    this.raise("Unmatched ')'");
                }
                if (cp === ReverseSolidus) {
                    this.raise("\\ at end of pattern");
                }
                if (cp === RightSquareBracket || cp === RightCurlyBracket) {
                    this.raise("Lone quantifier brackets");
                }
                const c = String.fromCodePoint(cp);
                this.raise(`Unexpected character '${c}'`);
            }
            for (const name of this._backreferenceNames) {
                if (!this._groupNames.has(name)) {
                    this.raise("Invalid named capture referenced");
                }
            }
            this.onPatternLeave(start, this.index);
        }
        countCapturingParens() {
            const start = this.index;
            let inClass = false;
            let escaped = false;
            let count = 0;
            let cp = 0;
            while ((cp = this.currentCodePoint) !== -1) {
                if (escaped) {
                    escaped = false;
                }
                else if (cp === ReverseSolidus) {
                    escaped = true;
                }
                else if (cp === LeftSquareBracket) {
                    inClass = true;
                }
                else if (cp === RightSquareBracket) {
                    inClass = false;
                }
                else if (cp === LeftParenthesis &&
                    !inClass &&
                    (this.nextCodePoint !== QuestionMark ||
                        (this.nextCodePoint2 === LessThanSign &&
                            this.nextCodePoint3 !== EqualsSign &&
                            this.nextCodePoint3 !== ExclamationMark))) {
                    count += 1;
                }
                this.advance();
            }
            this.rewind(start);
            return count;
        }
        consumeDisjunction() {
            const start = this.index;
            let i = 0;
            this.onDisjunctionEnter(start);
            do {
                this.consumeAlternative(i++);
            } while (this.eat(VerticalLine));
            if (this.consumeQuantifier(true)) {
                this.raise("Nothing to repeat");
            }
            if (this.eat(LeftCurlyBracket)) {
                this.raise("Lone quantifier brackets");
            }
            this.onDisjunctionLeave(start, this.index);
        }
        consumeAlternative(i) {
            const start = this.index;
            this.onAlternativeEnter(start, i);
            while (this.currentCodePoint !== -1 && this.consumeTerm()) {
            }
            this.onAlternativeLeave(start, this.index, i);
        }
        consumeTerm() {
            if (this._uFlag || this.strict) {
                return (this.consumeAssertion() ||
                    (this.consumeAtom() && this.consumeOptionalQuantifier()));
            }
            return ((this.consumeAssertion() &&
                (!this._lastAssertionIsQuantifiable ||
                    this.consumeOptionalQuantifier())) ||
                (this.consumeExtendedAtom() && this.consumeOptionalQuantifier()));
        }
        consumeOptionalQuantifier() {
            this.consumeQuantifier();
            return true;
        }
        consumeAssertion() {
            const start = this.index;
            this._lastAssertionIsQuantifiable = false;
            if (this.eat(CircumflexAccent)) {
                this.onEdgeAssertion(start, this.index, "start");
                return true;
            }
            if (this.eat(DollarSign)) {
                this.onEdgeAssertion(start, this.index, "end");
                return true;
            }
            if (this.eat2(ReverseSolidus, LatinCapitalLetterB)) {
                this.onWordBoundaryAssertion(start, this.index, "word", true);
                return true;
            }
            if (this.eat2(ReverseSolidus, LatinSmallLetterB)) {
                this.onWordBoundaryAssertion(start, this.index, "word", false);
                return true;
            }
            if (this.eat2(LeftParenthesis, QuestionMark)) {
                const lookbehind = this.ecmaVersion >= 2018 && this.eat(LessThanSign);
                let negate = false;
                if (this.eat(EqualsSign) || (negate = this.eat(ExclamationMark))) {
                    const kind = lookbehind ? "lookbehind" : "lookahead";
                    this.onLookaroundAssertionEnter(start, kind, negate);
                    this.consumeDisjunction();
                    if (!this.eat(RightParenthesis)) {
                        this.raise("Unterminated group");
                    }
                    this._lastAssertionIsQuantifiable = !lookbehind && !this.strict;
                    this.onLookaroundAssertionLeave(start, this.index, kind, negate);
                    return true;
                }
                this.rewind(start);
            }
            return false;
        }
        consumeQuantifier(noConsume = false) {
            const start = this.index;
            let min = 0;
            let max = 0;
            let greedy = false;
            if (this.eat(Asterisk)) {
                min = 0;
                max = Number.POSITIVE_INFINITY;
            }
            else if (this.eat(PlusSign)) {
                min = 1;
                max = Number.POSITIVE_INFINITY;
            }
            else if (this.eat(QuestionMark)) {
                min = 0;
                max = 1;
            }
            else if (this.eatBracedQuantifier(noConsume)) {
                min = this._lastMinValue;
                max = this._lastMaxValue;
            }
            else {
                return false;
            }
            greedy = !this.eat(QuestionMark);
            if (!noConsume) {
                this.onQuantifier(start, this.index, min, max, greedy);
            }
            return true;
        }
        eatBracedQuantifier(noError) {
            const start = this.index;
            if (this.eat(LeftCurlyBracket)) {
                this._lastMinValue = 0;
                this._lastMaxValue = Number.POSITIVE_INFINITY;
                if (this.eatDecimalDigits()) {
                    this._lastMinValue = this._lastMaxValue = this._lastIntValue;
                    if (this.eat(Comma)) {
                        this._lastMaxValue = this.eatDecimalDigits()
                            ? this._lastIntValue
                            : Number.POSITIVE_INFINITY;
                    }
                    if (this.eat(RightCurlyBracket)) {
                        if (!noError && this._lastMaxValue < this._lastMinValue) {
                            this.raise("numbers out of order in {} quantifier");
                        }
                        return true;
                    }
                }
                if (!noError && (this._uFlag || this.strict)) {
                    this.raise("Incomplete quantifier");
                }
                this.rewind(start);
            }
            return false;
        }
        consumeAtom() {
            return (this.consumePatternCharacter() ||
                this.consumeDot() ||
                this.consumeReverseSolidusAtomEscape() ||
                this.consumeCharacterClass() ||
                this.consumeUncapturingGroup() ||
                this.consumeCapturingGroup());
        }
        consumeDot() {
            if (this.eat(FullStop)) {
                this.onAnyCharacterSet(this.index - 1, this.index, "any");
                return true;
            }
            return false;
        }
        consumeReverseSolidusAtomEscape() {
            const start = this.index;
            if (this.eat(ReverseSolidus)) {
                if (this.consumeAtomEscape()) {
                    return true;
                }
                this.rewind(start);
            }
            return false;
        }
        consumeUncapturingGroup() {
            const start = this.index;
            if (this.eat3(LeftParenthesis, QuestionMark, Colon)) {
                this.onGroupEnter(start);
                this.consumeDisjunction();
                if (!this.eat(RightParenthesis)) {
                    this.raise("Unterminated group");
                }
                this.onGroupLeave(start, this.index);
                return true;
            }
            return false;
        }
        consumeCapturingGroup() {
            const start = this.index;
            if (this.eat(LeftParenthesis)) {
                let name = null;
                if (this.ecmaVersion >= 2018) {
                    if (this.consumeGroupSpecifier()) {
                        name = this._lastStrValue;
                    }
                }
                else if (this.currentCodePoint === QuestionMark) {
                    this.raise("Invalid group");
                }
                this.onCapturingGroupEnter(start, name);
                this.consumeDisjunction();
                if (!this.eat(RightParenthesis)) {
                    this.raise("Unterminated group");
                }
                this.onCapturingGroupLeave(start, this.index, name);
                return true;
            }
            return false;
        }
        consumeExtendedAtom() {
            return (this.consumeDot() ||
                this.consumeReverseSolidusAtomEscape() ||
                this.consumeReverseSolidusFollowedByC() ||
                this.consumeCharacterClass() ||
                this.consumeUncapturingGroup() ||
                this.consumeCapturingGroup() ||
                this.consumeInvalidBracedQuantifier() ||
                this.consumeExtendedPatternCharacter());
        }
        consumeReverseSolidusFollowedByC() {
            const start = this.index;
            if (this.currentCodePoint === ReverseSolidus &&
                this.nextCodePoint === LatinSmallLetterC) {
                this._lastIntValue = this.currentCodePoint;
                this.advance();
                this.onCharacter(start, this.index, ReverseSolidus);
                return true;
            }
            return false;
        }
        consumeInvalidBracedQuantifier() {
            if (this.eatBracedQuantifier(true)) {
                this.raise("Nothing to repeat");
            }
            return false;
        }
        consumePatternCharacter() {
            const start = this.index;
            const cp = this.currentCodePoint;
            if (cp !== -1 && !isSyntaxCharacter(cp)) {
                this.advance();
                this.onCharacter(start, this.index, cp);
                return true;
            }
            return false;
        }
        consumeExtendedPatternCharacter() {
            const start = this.index;
            const cp = this.currentCodePoint;
            if (cp !== -1 &&
                cp !== CircumflexAccent &&
                cp !== DollarSign &&
                cp !== ReverseSolidus &&
                cp !== FullStop &&
                cp !== Asterisk &&
                cp !== PlusSign &&
                cp !== QuestionMark &&
                cp !== LeftParenthesis &&
                cp !== RightParenthesis &&
                cp !== LeftSquareBracket &&
                cp !== VerticalLine) {
                this.advance();
                this.onCharacter(start, this.index, cp);
                return true;
            }
            return false;
        }
        consumeGroupSpecifier() {
            if (this.eat(QuestionMark)) {
                if (this.eatGroupName()) {
                    if (!this._groupNames.has(this._lastStrValue)) {
                        this._groupNames.add(this._lastStrValue);
                        return true;
                    }
                    this.raise("Duplicate capture group name");
                }
                this.raise("Invalid group");
            }
            return false;
        }
        consumeAtomEscape() {
            if (this.consumeBackreference() ||
                this.consumeCharacterClassEscape() ||
                this.consumeCharacterEscape() ||
                (this._nFlag && this.consumeKGroupName())) {
                return true;
            }
            if (this.strict || this._uFlag) {
                this.raise("Invalid escape");
            }
            return false;
        }
        consumeBackreference() {
            const start = this.index;
            if (this.eatDecimalEscape()) {
                const n = this._lastIntValue;
                if (n <= this._numCapturingParens) {
                    this.onBackreference(start - 1, this.index, n);
                    return true;
                }
                if (this.strict || this._uFlag) {
                    this.raise("Invalid escape");
                }
                this.rewind(start);
            }
            return false;
        }
        consumeCharacterClassEscape() {
            const start = this.index;
            if (this.eat(LatinSmallLetterD)) {
                this._lastIntValue = -1;
                this.onEscapeCharacterSet(start - 1, this.index, "digit", false);
                return true;
            }
            if (this.eat(LatinCapitalLetterD)) {
                this._lastIntValue = -1;
                this.onEscapeCharacterSet(start - 1, this.index, "digit", true);
                return true;
            }
            if (this.eat(LatinSmallLetterS)) {
                this._lastIntValue = -1;
                this.onEscapeCharacterSet(start - 1, this.index, "space", false);
                return true;
            }
            if (this.eat(LatinCapitalLetterS)) {
                this._lastIntValue = -1;
                this.onEscapeCharacterSet(start - 1, this.index, "space", true);
                return true;
            }
            if (this.eat(LatinSmallLetterW)) {
                this._lastIntValue = -1;
                this.onEscapeCharacterSet(start - 1, this.index, "word", false);
                return true;
            }
            if (this.eat(LatinCapitalLetterW)) {
                this._lastIntValue = -1;
                this.onEscapeCharacterSet(start - 1, this.index, "word", true);
                return true;
            }
            let negate = false;
            if (this._uFlag &&
                this.ecmaVersion >= 2018 &&
                (this.eat(LatinSmallLetterP) ||
                    (negate = this.eat(LatinCapitalLetterP)))) {
                this._lastIntValue = -1;
                if (this.eat(LeftCurlyBracket) &&
                    this.eatUnicodePropertyValueExpression() &&
                    this.eat(RightCurlyBracket)) {
                    this.onUnicodePropertyCharacterSet(start - 1, this.index, "property", this._lastKeyValue, this._lastValValue || null, negate);
                    return true;
                }
                this.raise("Invalid property name");
            }
            return false;
        }
        consumeCharacterEscape() {
            const start = this.index;
            if (this.eatControlEscape() ||
                this.eatCControlLetter() ||
                this.eatZero() ||
                this.eatHexEscapeSequence() ||
                this.eatRegExpUnicodeEscapeSequence() ||
                (!this.strict &&
                    !this._uFlag &&
                    this.eatLegacyOctalEscapeSequence()) ||
                this.eatIdentityEscape()) {
                this.onCharacter(start - 1, this.index, this._lastIntValue);
                return true;
            }
            return false;
        }
        consumeKGroupName() {
            const start = this.index;
            if (this.eat(LatinSmallLetterK)) {
                if (this.eatGroupName()) {
                    const groupName = this._lastStrValue;
                    this._backreferenceNames.add(groupName);
                    this.onBackreference(start - 1, this.index, groupName);
                    return true;
                }
                this.raise("Invalid named reference");
            }
            return false;
        }
        consumeCharacterClass() {
            const start = this.index;
            if (this.eat(LeftSquareBracket)) {
                const negate = this.eat(CircumflexAccent);
                this.onCharacterClassEnter(start, negate);
                this.consumeClassRanges();
                if (!this.eat(RightSquareBracket)) {
                    this.raise("Unterminated character class");
                }
                this.onCharacterClassLeave(start, this.index, negate);
                return true;
            }
            return false;
        }
        consumeClassRanges() {
            const strict = this.strict || this._uFlag;
            for (;;) {
                const rangeStart = this.index;
                if (!this.consumeClassAtom()) {
                    break;
                }
                const min = this._lastIntValue;
                if (!this.eat(HyphenMinus)) {
                    continue;
                }
                this.onCharacter(this.index - 1, this.index, HyphenMinus);
                if (!this.consumeClassAtom()) {
                    break;
                }
                const max = this._lastIntValue;
                if (min === -1 || max === -1) {
                    if (strict) {
                        this.raise("Invalid character class");
                    }
                    continue;
                }
                if (min > max) {
                    this.raise("Range out of order in character class");
                }
                this.onCharacterClassRange(rangeStart, this.index, min, max);
            }
        }
        consumeClassAtom() {
            const start = this.index;
            const cp = this.currentCodePoint;
            if (cp !== -1 && cp !== ReverseSolidus && cp !== RightSquareBracket) {
                this.advance();
                this._lastIntValue = cp;
                this.onCharacter(start, this.index, this._lastIntValue);
                return true;
            }
            if (this.eat(ReverseSolidus)) {
                if (this.consumeClassEscape()) {
                    return true;
                }
                if (!this.strict && this.currentCodePoint === LatinSmallLetterC) {
                    this._lastIntValue = ReverseSolidus;
                    this.onCharacter(start, this.index, this._lastIntValue);
                    return true;
                }
                if (this.strict || this._uFlag) {
                    this.raise("Invalid escape");
                }
                this.rewind(start);
            }
            return false;
        }
        consumeClassEscape() {
            const start = this.index;
            if (this.eat(LatinSmallLetterB)) {
                this._lastIntValue = Backspace;
                this.onCharacter(start - 1, this.index, this._lastIntValue);
                return true;
            }
            if (this._uFlag && this.eat(HyphenMinus)) {
                this._lastIntValue = HyphenMinus;
                this.onCharacter(start - 1, this.index, this._lastIntValue);
                return true;
            }
            let cp = 0;
            if (!this.strict &&
                !this._uFlag &&
                this.currentCodePoint === LatinSmallLetterC &&
                (isDecimalDigit((cp = this.nextCodePoint)) || cp === LowLine)) {
                this.advance();
                this.advance();
                this._lastIntValue = cp % 0x20;
                this.onCharacter(start - 1, this.index, this._lastIntValue);
                return true;
            }
            return (this.consumeCharacterClassEscape() || this.consumeCharacterEscape());
        }
        eatGroupName() {
            if (this.eat(LessThanSign)) {
                if (this.eatRegExpIdentifierName() && this.eat(GreaterThanSign)) {
                    return true;
                }
                this.raise("Invalid capture group name");
            }
            return false;
        }
        eatRegExpIdentifierName() {
            if (this.eatRegExpIdentifierStart()) {
                this._lastStrValue = String.fromCodePoint(this._lastIntValue);
                while (this.eatRegExpIdentifierPart()) {
                    this._lastStrValue += String.fromCodePoint(this._lastIntValue);
                }
                return true;
            }
            return false;
        }
        eatRegExpIdentifierStart() {
            const start = this.index;
            const forceUFlag = !this._uFlag && this.ecmaVersion >= 2020;
            let cp = this.currentCodePoint;
            this.advance();
            if (cp === ReverseSolidus &&
                this.eatRegExpUnicodeEscapeSequence(forceUFlag)) {
                cp = this._lastIntValue;
            }
            else if (forceUFlag &&
                isLeadSurrogate(cp) &&
                isTrailSurrogate(this.currentCodePoint)) {
                cp = combineSurrogatePair(cp, this.currentCodePoint);
                this.advance();
            }
            if (isRegExpIdentifierStart(cp)) {
                this._lastIntValue = cp;
                return true;
            }
            if (this.index !== start) {
                this.rewind(start);
            }
            return false;
        }
        eatRegExpIdentifierPart() {
            const start = this.index;
            const forceUFlag = !this._uFlag && this.ecmaVersion >= 2020;
            let cp = this.currentCodePoint;
            this.advance();
            if (cp === ReverseSolidus &&
                this.eatRegExpUnicodeEscapeSequence(forceUFlag)) {
                cp = this._lastIntValue;
            }
            else if (forceUFlag &&
                isLeadSurrogate(cp) &&
                isTrailSurrogate(this.currentCodePoint)) {
                cp = combineSurrogatePair(cp, this.currentCodePoint);
                this.advance();
            }
            if (isRegExpIdentifierPart(cp)) {
                this._lastIntValue = cp;
                return true;
            }
            if (this.index !== start) {
                this.rewind(start);
            }
            return false;
        }
        eatCControlLetter() {
            const start = this.index;
            if (this.eat(LatinSmallLetterC)) {
                if (this.eatControlLetter()) {
                    return true;
                }
                this.rewind(start);
            }
            return false;
        }
        eatZero() {
            if (this.currentCodePoint === DigitZero &&
                !isDecimalDigit(this.nextCodePoint)) {
                this._lastIntValue = 0;
                this.advance();
                return true;
            }
            return false;
        }
        eatControlEscape() {
            if (this.eat(LatinSmallLetterF)) {
                this._lastIntValue = FormFeed;
                return true;
            }
            if (this.eat(LatinSmallLetterN)) {
                this._lastIntValue = LineFeed;
                return true;
            }
            if (this.eat(LatinSmallLetterR)) {
                this._lastIntValue = CarriageReturn;
                return true;
            }
            if (this.eat(LatinSmallLetterT)) {
                this._lastIntValue = CharacterTabulation;
                return true;
            }
            if (this.eat(LatinSmallLetterV)) {
                this._lastIntValue = LineTabulation;
                return true;
            }
            return false;
        }
        eatControlLetter() {
            const cp = this.currentCodePoint;
            if (isLatinLetter(cp)) {
                this.advance();
                this._lastIntValue = cp % 0x20;
                return true;
            }
            return false;
        }
        eatRegExpUnicodeEscapeSequence(forceUFlag = false) {
            const start = this.index;
            const uFlag = forceUFlag || this._uFlag;
            if (this.eat(LatinSmallLetterU)) {
                if ((uFlag && this.eatRegExpUnicodeSurrogatePairEscape()) ||
                    this.eatFixedHexDigits(4) ||
                    (uFlag && this.eatRegExpUnicodeCodePointEscape())) {
                    return true;
                }
                if (this.strict || uFlag) {
                    this.raise("Invalid unicode escape");
                }
                this.rewind(start);
            }
            return false;
        }
        eatRegExpUnicodeSurrogatePairEscape() {
            const start = this.index;
            if (this.eatFixedHexDigits(4)) {
                const lead = this._lastIntValue;
                if (isLeadSurrogate(lead) &&
                    this.eat(ReverseSolidus) &&
                    this.eat(LatinSmallLetterU) &&
                    this.eatFixedHexDigits(4)) {
                    const trail = this._lastIntValue;
                    if (isTrailSurrogate(trail)) {
                        this._lastIntValue = combineSurrogatePair(lead, trail);
                        return true;
                    }
                }
                this.rewind(start);
            }
            return false;
        }
        eatRegExpUnicodeCodePointEscape() {
            const start = this.index;
            if (this.eat(LeftCurlyBracket) &&
                this.eatHexDigits() &&
                this.eat(RightCurlyBracket) &&
                isValidUnicode(this._lastIntValue)) {
                return true;
            }
            this.rewind(start);
            return false;
        }
        eatIdentityEscape() {
            const cp = this.currentCodePoint;
            if (this.isValidIdentityEscape(cp)) {
                this._lastIntValue = cp;
                this.advance();
                return true;
            }
            return false;
        }
        isValidIdentityEscape(cp) {
            if (cp === -1) {
                return false;
            }
            if (this._uFlag) {
                return isSyntaxCharacter(cp) || cp === Solidus;
            }
            if (this.strict) {
                return !isIdContinue(cp);
            }
            if (this._nFlag) {
                return !(cp === LatinSmallLetterC || cp === LatinSmallLetterK);
            }
            return cp !== LatinSmallLetterC;
        }
        eatDecimalEscape() {
            this._lastIntValue = 0;
            let cp = this.currentCodePoint;
            if (cp >= DigitOne && cp <= DigitNine) {
                do {
                    this._lastIntValue = 10 * this._lastIntValue + (cp - DigitZero);
                    this.advance();
                } while ((cp = this.currentCodePoint) >= DigitZero &&
                    cp <= DigitNine);
                return true;
            }
            return false;
        }
        eatUnicodePropertyValueExpression() {
            const start = this.index;
            if (this.eatUnicodePropertyName() && this.eat(EqualsSign)) {
                this._lastKeyValue = this._lastStrValue;
                if (this.eatUnicodePropertyValue()) {
                    this._lastValValue = this._lastStrValue;
                    if (isValidUnicodeProperty(this.ecmaVersion, this._lastKeyValue, this._lastValValue)) {
                        return true;
                    }
                    this.raise("Invalid property name");
                }
            }
            this.rewind(start);
            if (this.eatLoneUnicodePropertyNameOrValue()) {
                const nameOrValue = this._lastStrValue;
                if (isValidUnicodeProperty(this.ecmaVersion, "General_Category", nameOrValue)) {
                    this._lastKeyValue = "General_Category";
                    this._lastValValue = nameOrValue;
                    return true;
                }
                if (isValidLoneUnicodeProperty(this.ecmaVersion, nameOrValue)) {
                    this._lastKeyValue = nameOrValue;
                    this._lastValValue = "";
                    return true;
                }
                this.raise("Invalid property name");
            }
            return false;
        }
        eatUnicodePropertyName() {
            this._lastStrValue = "";
            while (isUnicodePropertyNameCharacter(this.currentCodePoint)) {
                this._lastStrValue += String.fromCodePoint(this.currentCodePoint);
                this.advance();
            }
            return this._lastStrValue !== "";
        }
        eatUnicodePropertyValue() {
            this._lastStrValue = "";
            while (isUnicodePropertyValueCharacter(this.currentCodePoint)) {
                this._lastStrValue += String.fromCodePoint(this.currentCodePoint);
                this.advance();
            }
            return this._lastStrValue !== "";
        }
        eatLoneUnicodePropertyNameOrValue() {
            return this.eatUnicodePropertyValue();
        }
        eatHexEscapeSequence() {
            const start = this.index;
            if (this.eat(LatinSmallLetterX)) {
                if (this.eatFixedHexDigits(2)) {
                    return true;
                }
                if (this._uFlag || this.strict) {
                    this.raise("Invalid escape");
                }
                this.rewind(start);
            }
            return false;
        }
        eatDecimalDigits() {
            const start = this.index;
            this._lastIntValue = 0;
            while (isDecimalDigit(this.currentCodePoint)) {
                this._lastIntValue =
                    10 * this._lastIntValue + digitToInt(this.currentCodePoint);
                this.advance();
            }
            return this.index !== start;
        }
        eatHexDigits() {
            const start = this.index;
            this._lastIntValue = 0;
            while (isHexDigit(this.currentCodePoint)) {
                this._lastIntValue =
                    16 * this._lastIntValue + digitToInt(this.currentCodePoint);
                this.advance();
            }
            return this.index !== start;
        }
        eatLegacyOctalEscapeSequence() {
            if (this.eatOctalDigit()) {
                const n1 = this._lastIntValue;
                if (this.eatOctalDigit()) {
                    const n2 = this._lastIntValue;
                    if (n1 <= 3 && this.eatOctalDigit()) {
                        this._lastIntValue = n1 * 64 + n2 * 8 + this._lastIntValue;
                    }
                    else {
                        this._lastIntValue = n1 * 8 + n2;
                    }
                }
                else {
                    this._lastIntValue = n1;
                }
                return true;
            }
            return false;
        }
        eatOctalDigit() {
            const cp = this.currentCodePoint;
            if (isOctalDigit(cp)) {
                this.advance();
                this._lastIntValue = cp - DigitZero;
                return true;
            }
            this._lastIntValue = 0;
            return false;
        }
        eatFixedHexDigits(length) {
            const start = this.index;
            this._lastIntValue = 0;
            for (let i = 0; i < length; ++i) {
                const cp = this.currentCodePoint;
                if (!isHexDigit(cp)) {
                    this.rewind(start);
                    return false;
                }
                this._lastIntValue = 16 * this._lastIntValue + digitToInt(cp);
                this.advance();
            }
            return true;
        }
    }
    const DummyPattern = {};
    const DummyFlags = {};
    const DummyCapturingGroup = {};
    class RegExpParserState {
        constructor(options) {
            this._node = DummyPattern;
            this._flags = DummyFlags;
            this._backreferences = [];
            this._capturingGroups = [];
            this.source = "";
            this.strict = Boolean(options && options.strict);
            this.ecmaVersion = (options && options.ecmaVersion) || 2022;
        }
        get pattern() {
            if (this._node.type !== "Pattern") {
                throw new Error("UnknownError");
            }
            return this._node;
        }
        get flags() {
            if (this._flags.type !== "Flags") {
                throw new Error("UnknownError");
            }
            return this._flags;
        }
        onFlags(start, end, global, ignoreCase, multiline, unicode, sticky, dotAll, hasIndices) {
            this._flags = {
                type: "Flags",
                parent: null,
                start,
                end,
                raw: this.source.slice(start, end),
                global,
                ignoreCase,
                multiline,
                unicode,
                sticky,
                dotAll,
                hasIndices,
            };
        }
        onPatternEnter(start) {
            this._node = {
                type: "Pattern",
                parent: null,
                start,
                end: start,
                raw: "",
                alternatives: [],
            };
            this._backreferences.length = 0;
            this._capturingGroups.length = 0;
        }
        onPatternLeave(start, end) {
            this._node.end = end;
            this._node.raw = this.source.slice(start, end);
            for (const reference of this._backreferences) {
                const ref = reference.ref;
                const group = typeof ref === "number"
                    ? this._capturingGroups[ref - 1]
                    : this._capturingGroups.find(g => g.name === ref);
                reference.resolved = group;
                group.references.push(reference);
            }
        }
        onAlternativeEnter(start) {
            const parent = this._node;
            if (parent.type !== "Assertion" &&
                parent.type !== "CapturingGroup" &&
                parent.type !== "Group" &&
                parent.type !== "Pattern") {
                throw new Error("UnknownError");
            }
            this._node = {
                type: "Alternative",
                parent,
                start,
                end: start,
                raw: "",
                elements: [],
            };
            parent.alternatives.push(this._node);
        }
        onAlternativeLeave(start, end) {
            const node = this._node;
            if (node.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            node.end = end;
            node.raw = this.source.slice(start, end);
            this._node = node.parent;
        }
        onGroupEnter(start) {
            const parent = this._node;
            if (parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            this._node = {
                type: "Group",
                parent,
                start,
                end: start,
                raw: "",
                alternatives: [],
            };
            parent.elements.push(this._node);
        }
        onGroupLeave(start, end) {
            const node = this._node;
            if (node.type !== "Group" || node.parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            node.end = end;
            node.raw = this.source.slice(start, end);
            this._node = node.parent;
        }
        onCapturingGroupEnter(start, name) {
            const parent = this._node;
            if (parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            this._node = {
                type: "CapturingGroup",
                parent,
                start,
                end: start,
                raw: "",
                name,
                alternatives: [],
                references: [],
            };
            parent.elements.push(this._node);
            this._capturingGroups.push(this._node);
        }
        onCapturingGroupLeave(start, end) {
            const node = this._node;
            if (node.type !== "CapturingGroup" ||
                node.parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            node.end = end;
            node.raw = this.source.slice(start, end);
            this._node = node.parent;
        }
        onQuantifier(start, end, min, max, greedy) {
            const parent = this._node;
            if (parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            const element = parent.elements.pop();
            if (element == null ||
                element.type === "Quantifier" ||
                (element.type === "Assertion" && element.kind !== "lookahead")) {
                throw new Error("UnknownError");
            }
            const node = {
                type: "Quantifier",
                parent,
                start: element.start,
                end,
                raw: this.source.slice(element.start, end),
                min,
                max,
                greedy,
                element,
            };
            parent.elements.push(node);
            element.parent = node;
        }
        onLookaroundAssertionEnter(start, kind, negate) {
            const parent = this._node;
            if (parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            const node = (this._node = {
                type: "Assertion",
                parent,
                start,
                end: start,
                raw: "",
                kind,
                negate,
                alternatives: [],
            });
            parent.elements.push(node);
        }
        onLookaroundAssertionLeave(start, end) {
            const node = this._node;
            if (node.type !== "Assertion" || node.parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            node.end = end;
            node.raw = this.source.slice(start, end);
            this._node = node.parent;
        }
        onEdgeAssertion(start, end, kind) {
            const parent = this._node;
            if (parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            parent.elements.push({
                type: "Assertion",
                parent,
                start,
                end,
                raw: this.source.slice(start, end),
                kind,
            });
        }
        onWordBoundaryAssertion(start, end, kind, negate) {
            const parent = this._node;
            if (parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            parent.elements.push({
                type: "Assertion",
                parent,
                start,
                end,
                raw: this.source.slice(start, end),
                kind,
                negate,
            });
        }
        onAnyCharacterSet(start, end, kind) {
            const parent = this._node;
            if (parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            parent.elements.push({
                type: "CharacterSet",
                parent,
                start,
                end,
                raw: this.source.slice(start, end),
                kind,
            });
        }
        onEscapeCharacterSet(start, end, kind, negate) {
            const parent = this._node;
            if (parent.type !== "Alternative" && parent.type !== "CharacterClass") {
                throw new Error("UnknownError");
            }
            parent.elements.push({
                type: "CharacterSet",
                parent,
                start,
                end,
                raw: this.source.slice(start, end),
                kind,
                negate,
            });
        }
        onUnicodePropertyCharacterSet(start, end, kind, key, value, negate) {
            const parent = this._node;
            if (parent.type !== "Alternative" && parent.type !== "CharacterClass") {
                throw new Error("UnknownError");
            }
            parent.elements.push({
                type: "CharacterSet",
                parent,
                start,
                end,
                raw: this.source.slice(start, end),
                kind,
                key,
                value,
                negate,
            });
        }
        onCharacter(start, end, value) {
            const parent = this._node;
            if (parent.type !== "Alternative" && parent.type !== "CharacterClass") {
                throw new Error("UnknownError");
            }
            parent.elements.push({
                type: "Character",
                parent,
                start,
                end,
                raw: this.source.slice(start, end),
                value,
            });
        }
        onBackreference(start, end, ref) {
            const parent = this._node;
            if (parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            const node = {
                type: "Backreference",
                parent,
                start,
                end,
                raw: this.source.slice(start, end),
                ref,
                resolved: DummyCapturingGroup,
            };
            parent.elements.push(node);
            this._backreferences.push(node);
        }
        onCharacterClassEnter(start, negate) {
            const parent = this._node;
            if (parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            this._node = {
                type: "CharacterClass",
                parent,
                start,
                end: start,
                raw: "",
                negate,
                elements: [],
            };
            parent.elements.push(this._node);
        }
        onCharacterClassLeave(start, end) {
            const node = this._node;
            if (node.type !== "CharacterClass" ||
                node.parent.type !== "Alternative") {
                throw new Error("UnknownError");
            }
            node.end = end;
            node.raw = this.source.slice(start, end);
            this._node = node.parent;
        }
        onCharacterClassRange(start, end) {
            const parent = this._node;
            if (parent.type !== "CharacterClass") {
                throw new Error("UnknownError");
            }
            const elements = parent.elements;
            const max = elements.pop();
            const hyphen = elements.pop();
            const min = elements.pop();
            if (!min ||
                !max ||
                !hyphen ||
                min.type !== "Character" ||
                max.type !== "Character" ||
                hyphen.type !== "Character" ||
                hyphen.value !== HyphenMinus) {
                throw new Error("UnknownError");
            }
            const node = {
                type: "CharacterClassRange",
                parent,
                start,
                end,
                raw: this.source.slice(start, end),
                min,
                max,
            };
            min.parent = node;
            max.parent = node;
            elements.push(node);
        }
    }
    class RegExpParser {
        constructor(options) {
            this._state = new RegExpParserState(options);
            this._validator = new RegExpValidator(this._state);
        }
        parseLiteral(source, start = 0, end = source.length) {
            this._state.source = source;
            this._validator.validateLiteral(source, start, end);
            const pattern = this._state.pattern;
            const flags = this._state.flags;
            const literal = {
                type: "RegExpLiteral",
                parent: null,
                start,
                end,
                raw: source,
                pattern,
                flags,
            };
            pattern.parent = literal;
            flags.parent = literal;
            return literal;
        }
        parseFlags(source, start = 0, end = source.length) {
            this._state.source = source;
            this._validator.validateFlags(source, start, end);
            return this._state.flags;
        }
        parsePattern(source, start = 0, end = source.length, uFlag = false) {
            this._state.source = source;
            this._validator.validatePattern(source, start, end, uFlag);
            return this._state.pattern;
        }
    }
    class RegExpVisitor {
        constructor(handlers) {
            this._handlers = handlers;
        }
        visit(node) {
            switch (node.type) {
                case "Alternative":
                    this.visitAlternative(node);
                    break;
                case "Assertion":
                    this.visitAssertion(node);
                    break;
                case "Backreference":
                    this.visitBackreference(node);
                    break;
                case "CapturingGroup":
                    this.visitCapturingGroup(node);
                    break;
                case "Character":
                    this.visitCharacter(node);
                    break;
                case "CharacterClass":
                    this.visitCharacterClass(node);
                    break;
                case "CharacterClassRange":
                    this.visitCharacterClassRange(node);
                    break;
                case "CharacterSet":
                    this.visitCharacterSet(node);
                    break;
                case "Flags":
                    this.visitFlags(node);
                    break;
                case "Group":
                    this.visitGroup(node);
                    break;
                case "Pattern":
                    this.visitPattern(node);
                    break;
                case "Quantifier":
                    this.visitQuantifier(node);
                    break;
                case "RegExpLiteral":
                    this.visitRegExpLiteral(node);
                    break;
                default:
                    throw new Error(`Unknown type: ${node.type}`);
            }
        }
        visitAlternative(node) {
            if (this._handlers.onAlternativeEnter) {
                this._handlers.onAlternativeEnter(node);
            }
            node.elements.forEach(this.visit, this);
            if (this._handlers.onAlternativeLeave) {
                this._handlers.onAlternativeLeave(node);
            }
        }
        visitAssertion(node) {
            if (this._handlers.onAssertionEnter) {
                this._handlers.onAssertionEnter(node);
            }
            if (node.kind === "lookahead" || node.kind === "lookbehind") {
                node.alternatives.forEach(this.visit, this);
            }
            if (this._handlers.onAssertionLeave) {
                this._handlers.onAssertionLeave(node);
            }
        }
        visitBackreference(node) {
            if (this._handlers.onBackreferenceEnter) {
                this._handlers.onBackreferenceEnter(node);
            }
            if (this._handlers.onBackreferenceLeave) {
                this._handlers.onBackreferenceLeave(node);
            }
        }
        visitCapturingGroup(node) {
            if (this._handlers.onCapturingGroupEnter) {
                this._handlers.onCapturingGroupEnter(node);
            }
            node.alternatives.forEach(this.visit, this);
            if (this._handlers.onCapturingGroupLeave) {
                this._handlers.onCapturingGroupLeave(node);
            }
        }
        visitCharacter(node) {
            if (this._handlers.onCharacterEnter) {
                this._handlers.onCharacterEnter(node);
            }
            if (this._handlers.onCharacterLeave) {
                this._handlers.onCharacterLeave(node);
            }
        }
        visitCharacterClass(node) {
            if (this._handlers.onCharacterClassEnter) {
                this._handlers.onCharacterClassEnter(node);
            }
            node.elements.forEach(this.visit, this);
            if (this._handlers.onCharacterClassLeave) {
                this._handlers.onCharacterClassLeave(node);
            }
        }
        visitCharacterClassRange(node) {
            if (this._handlers.onCharacterClassRangeEnter) {
                this._handlers.onCharacterClassRangeEnter(node);
            }
            this.visitCharacter(node.min);
            this.visitCharacter(node.max);
            if (this._handlers.onCharacterClassRangeLeave) {
                this._handlers.onCharacterClassRangeLeave(node);
            }
        }
        visitCharacterSet(node) {
            if (this._handlers.onCharacterSetEnter) {
                this._handlers.onCharacterSetEnter(node);
            }
            if (this._handlers.onCharacterSetLeave) {
                this._handlers.onCharacterSetLeave(node);
            }
        }
        visitFlags(node) {
            if (this._handlers.onFlagsEnter) {
                this._handlers.onFlagsEnter(node);
            }
            if (this._handlers.onFlagsLeave) {
                this._handlers.onFlagsLeave(node);
            }
        }
        visitGroup(node) {
            if (this._handlers.onGroupEnter) {
                this._handlers.onGroupEnter(node);
            }
            node.alternatives.forEach(this.visit, this);
            if (this._handlers.onGroupLeave) {
                this._handlers.onGroupLeave(node);
            }
        }
        visitPattern(node) {
            if (this._handlers.onPatternEnter) {
                this._handlers.onPatternEnter(node);
            }
            node.alternatives.forEach(this.visit, this);
            if (this._handlers.onPatternLeave) {
                this._handlers.onPatternLeave(node);
            }
        }
        visitQuantifier(node) {
            if (this._handlers.onQuantifierEnter) {
                this._handlers.onQuantifierEnter(node);
            }
            this.visit(node.element);
            if (this._handlers.onQuantifierLeave) {
                this._handlers.onQuantifierLeave(node);
            }
        }
        visitRegExpLiteral(node) {
            if (this._handlers.onRegExpLiteralEnter) {
                this._handlers.onRegExpLiteralEnter(node);
            }
            this.visitPattern(node.pattern);
            this.visitFlags(node.flags);
            if (this._handlers.onRegExpLiteralLeave) {
                this._handlers.onRegExpLiteralLeave(node);
            }
        }
    }
    function parseRegExpLiteral(source, options) {
        return new RegExpParser(options).parseLiteral(String(source));
    }
    function validateRegExpLiteral(source, options) {
        return new RegExpValidator(options).validateLiteral(source);
    }
    function visitRegExpAST(node, handlers) {
        new RegExpVisitor(handlers).visit(node);
    }
    exports.AST = ast;
    exports.RegExpParser = RegExpParser;
    exports.RegExpValidator = RegExpValidator;
    exports.parseRegExpLiteral = parseRegExpLiteral;
    exports.validateRegExpLiteral = validateRegExpLiteral;
    exports.visitRegExpAST = visitRegExpAST;
},
/* eslint/lib/rules/no-debugger.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow the use of `debugger`",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-debugger"
            },
            fixable: null,
            schema: [],
            messages: {
                unexpected: "Unexpected 'debugger' statement."
            }
        },
        create(context) {
            return {
                DebuggerStatement(node) {
                    context.report({
                        node,
                        messageId: "unexpected"
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-delete-var.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow deleting variables",
                category: "Variables",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-delete-var"
            },
            schema: [],
            messages: {
                unexpected: "Variables should not be deleted."
            }
        },
        create(context) {
            return {
                UnaryExpression(node) {
                    if (node.operator === "delete" && node.argument.type === "Identifier") {
                        context.report({ node, messageId: "unexpected" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-div-regex.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow division operators explicitly at the beginning of regular expressions",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-div-regex"
            },
            fixable: "code",
            schema: [],
            messages: {
                unexpected: "A regular expression literal can be confused with '/='."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                Literal(node) {
                    const token = sourceCode.getFirstToken(node);
                    if (token.type === "RegularExpression" && token.value[1] === "=") {
                        context.report({
                            node,
                            messageId: "unexpected",
                            fix(fixer) {
                                return fixer.replaceTextRange([token.range[0] + 1, token.range[0] + 2], "[=]");
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-dupe-args.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow duplicate arguments in `function` definitions",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-dupe-args"
            },
            schema: [],
            messages: {
                unexpected: "Duplicate param '{{name}}'."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Checks whether or not a given definition is a parameter's.
             * @param {eslint-scope.DefEntry} def A definition to check.
             * @returns {boolean} `true` if the definition is a parameter's.
             */
            function isParameter(def) {
                return def.type === "Parameter";
            }
            /**
             * Determines if a given node has duplicate parameters.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             * @private
             */
            function checkParams(node) {
                const variables = context.getDeclaredVariables(node);
                for (let i = 0; i < variables.length; ++i) {
                    const variable = variables[i];
                    // Checks and reports duplications.
                    const defs = variable.defs.filter(isParameter);
                    if (defs.length >= 2) {
                        context.report({
                            node,
                            messageId: "unexpected",
                            data: { name: variable.name }
                        });
                    }
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                FunctionDeclaration: checkParams,
                FunctionExpression: checkParams
            };
        }
    };
},
/* eslint/lib/rules/no-dupe-class-members.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow duplicate class members",
                category: "ECMAScript 6",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-dupe-class-members"
            },
            schema: [],
            messages: {
                unexpected: "Duplicate name '{{name}}'."
            }
        },
        create(context) {
            let stack = [];
            /**
             * Gets state of a given member name.
             * @param {string} name A name of a member.
             * @param {boolean} isStatic A flag which specifies that is a static member.
             * @returns {Object} A state of a given member name.
             *   - retv.init {boolean} A flag which shows the name is declared as normal member.
             *   - retv.get {boolean} A flag which shows the name is declared as getter.
             *   - retv.set {boolean} A flag which shows the name is declared as setter.
             */
            function getState(name, isStatic) {
                const stateMap = stack[stack.length - 1];
                const key = `$${name}`; // to avoid "__proto__".
                if (!stateMap[key]) {
                    stateMap[key] = {
                        nonStatic: { init: false, get: false, set: false },
                        static: { init: false, get: false, set: false }
                    };
                }
                return stateMap[key][isStatic ? "static" : "nonStatic"];
            }
            return {
                // Initializes the stack of state of member declarations.
                Program() {
                    stack = [];
                },
                // Initializes state of member declarations for the class.
                ClassBody() {
                    stack.push(Object.create(null));
                },
                // Disposes the state for the class.
                "ClassBody:exit"() {
                    stack.pop();
                },
                // Reports the node if its name has been declared already.
                MethodDefinition(node) {
                    const name = astUtils.getStaticPropertyName(node);
                    if (name === null || node.kind === "constructor") {
                        return;
                    }
                    const state = getState(name, node.static);
                    let isDuplicate = false;
                    if (node.kind === "get") {
                        isDuplicate = (state.init || state.get);
                        state.get = true;
                    }
                    else if (node.kind === "set") {
                        isDuplicate = (state.init || state.set);
                        state.set = true;
                    }
                    else {
                        isDuplicate = (state.init || state.get || state.set);
                        state.init = true;
                    }
                    if (isDuplicate) {
                        context.report({ node, messageId: "unexpected", data: { name } });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-dupe-else-if.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Determines whether the first given array is a subset of the second given array.
     * @param {Function} comparator A function to compare two elements, should return `true` if they are equal.
     * @param {Array} arrA The array to compare from.
     * @param {Array} arrB The array to compare against.
     * @returns {boolean} `true` if the array `arrA` is a subset of the array `arrB`.
     */
    function isSubsetByComparator(comparator, arrA, arrB) {
        return arrA.every(a => arrB.some(b => comparator(a, b)));
    }
    /**
     * Splits the given node by the given logical operator.
     * @param {string} operator Logical operator `||` or `&&`.
     * @param {ASTNode} node The node to split.
     * @returns {ASTNode[]} Array of conditions that makes the node when joined by the operator.
     */
    function splitByLogicalOperator(operator, node) {
        if (node.type === "LogicalExpression" && node.operator === operator) {
            return [...splitByLogicalOperator(operator, node.left), ...splitByLogicalOperator(operator, node.right)];
        }
        return [node];
    }
    const splitByOr = splitByLogicalOperator.bind(null, "||");
    const splitByAnd = splitByLogicalOperator.bind(null, "&&");
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow duplicate conditions in if-else-if chains",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-dupe-else-if"
            },
            schema: [],
            messages: {
                unexpected: "This branch can never execute. Its condition is a duplicate or covered by previous conditions in the if-else-if chain."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Determines whether the two given nodes are considered to be equal. In particular, given that the nodes
             * represent expressions in a boolean context, `||` and `&&` can be considered as commutative operators.
             * @param {ASTNode} a First node.
             * @param {ASTNode} b Second node.
             * @returns {boolean} `true` if the nodes are considered to be equal.
             */
            function equal(a, b) {
                if (a.type !== b.type) {
                    return false;
                }
                if (a.type === "LogicalExpression" &&
                    (a.operator === "||" || a.operator === "&&") &&
                    a.operator === b.operator) {
                    return equal(a.left, b.left) && equal(a.right, b.right) ||
                        equal(a.left, b.right) && equal(a.right, b.left);
                }
                return astUtils.equalTokens(a, b, sourceCode);
            }
            const isSubset = isSubsetByComparator.bind(null, equal);
            return {
                IfStatement(node) {
                    const test = node.test, conditionsToCheck = test.type === "LogicalExpression" && test.operator === "&&"
                        ? [test, ...splitByAnd(test)]
                        : [test];
                    let current = node, listToCheck = conditionsToCheck.map(c => splitByOr(c).map(splitByAnd));
                    while (current.parent && current.parent.type === "IfStatement" && current.parent.alternate === current) {
                        current = current.parent;
                        const currentOrOperands = splitByOr(current.test).map(splitByAnd);
                        listToCheck = listToCheck.map(orOperands => orOperands.filter(orOperand => !currentOrOperands.some(currentOrOperand => isSubset(currentOrOperand, orOperand))));
                        if (listToCheck.some(orOperands => orOperands.length === 0)) {
                            context.report({ node: test, messageId: "unexpected" });
                            break;
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-dupe-keys.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const GET_KIND = /^(?:init|get)$/u;
    const SET_KIND = /^(?:init|set)$/u;
    /**
     * The class which stores properties' information of an object.
     */
    class ObjectInfo {
        // eslint-disable-next-line jsdoc/require-description
        /**
         * @param {ObjectInfo|null} upper The information of the outer object.
         * @param {ASTNode} node The ObjectExpression node of this information.
         */
        constructor(upper, node) {
            this.upper = upper;
            this.node = node;
            this.properties = new Map();
        }
        /**
         * Gets the information of the given Property node.
         * @param {ASTNode} node The Property node to get.
         * @returns {{get: boolean, set: boolean}} The information of the property.
         */
        getPropertyInfo(node) {
            const name = astUtils.getStaticPropertyName(node);
            if (!this.properties.has(name)) {
                this.properties.set(name, { get: false, set: false });
            }
            return this.properties.get(name);
        }
        /**
         * Checks whether the given property has been defined already or not.
         * @param {ASTNode} node The Property node to check.
         * @returns {boolean} `true` if the property has been defined.
         */
        isPropertyDefined(node) {
            const entry = this.getPropertyInfo(node);
            return ((GET_KIND.test(node.kind) && entry.get) ||
                (SET_KIND.test(node.kind) && entry.set));
        }
        /**
         * Defines the given property.
         * @param {ASTNode} node The Property node to define.
         * @returns {void}
         */
        defineProperty(node) {
            const entry = this.getPropertyInfo(node);
            if (GET_KIND.test(node.kind)) {
                entry.get = true;
            }
            if (SET_KIND.test(node.kind)) {
                entry.set = true;
            }
        }
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow duplicate keys in object literals",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-dupe-keys"
            },
            schema: [],
            messages: {
                unexpected: "Duplicate key '{{name}}'."
            }
        },
        create(context) {
            let info = null;
            return {
                ObjectExpression(node) {
                    info = new ObjectInfo(info, node);
                },
                "ObjectExpression:exit"() {
                    info = info.upper;
                },
                Property(node) {
                    const name = astUtils.getStaticPropertyName(node);
                    // Skip destructuring.
                    if (node.parent.type !== "ObjectExpression") {
                        return;
                    }
                    // Skip if the name is not static.
                    if (name === null) {
                        return;
                    }
                    // Reports if the name is defined already.
                    if (info.isPropertyDefined(node)) {
                        context.report({
                            node: info.node,
                            loc: node.key.loc,
                            messageId: "unexpected",
                            data: { name }
                        });
                    }
                    // Update info.
                    info.defineProperty(node);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-duplicate-case.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow duplicate case labels",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-duplicate-case"
            },
            schema: [],
            messages: {
                unexpected: "Duplicate case label."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Determines whether the two given nodes are considered to be equal.
             * @param {ASTNode} a First node.
             * @param {ASTNode} b Second node.
             * @returns {boolean} `true` if the nodes are considered to be equal.
             */
            function equal(a, b) {
                if (a.type !== b.type) {
                    return false;
                }
                return astUtils.equalTokens(a, b, sourceCode);
            }
            return {
                SwitchStatement(node) {
                    const previousTests = [];
                    for (const switchCase of node.cases) {
                        if (switchCase.test) {
                            const test = switchCase.test;
                            if (previousTests.some(previousTest => equal(previousTest, test))) {
                                context.report({ node: switchCase, messageId: "unexpected" });
                            }
                            else {
                                previousTests.push(test);
                            }
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-duplicate-imports.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const NAMED_TYPES = ["ImportSpecifier", "ExportSpecifier"];
    const NAMESPACE_TYPES = [
        "ImportNamespaceSpecifier",
        "ExportNamespaceSpecifier"
    ];
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    /**
     * Check if an import/export type belongs to (ImportSpecifier|ExportSpecifier) or (ImportNamespaceSpecifier|ExportNamespaceSpecifier).
     * @param {string} importExportType An import/export type to check.
     * @param {string} type Can be "named" or "namespace"
     * @returns {boolean} True if import/export type belongs to (ImportSpecifier|ExportSpecifier) or (ImportNamespaceSpecifier|ExportNamespaceSpecifier) and false if it doesn't.
     */
    function isImportExportSpecifier(importExportType, type) {
        const arrayToCheck = type === "named" ? NAMED_TYPES : NAMESPACE_TYPES;
        return arrayToCheck.includes(importExportType);
    }
    /**
     * Return the type of (import|export).
     * @param {ASTNode} node A node to get.
     * @returns {string} The type of the (import|export).
     */
    function getImportExportType(node) {
        if (node.specifiers && node.specifiers.length > 0) {
            const nodeSpecifiers = node.specifiers;
            const index = nodeSpecifiers.findIndex(({ type }) => isImportExportSpecifier(type, "named") ||
                isImportExportSpecifier(type, "namespace"));
            const i = index > -1 ? index : 0;
            return nodeSpecifiers[i].type;
        }
        if (node.type === "ExportAllDeclaration") {
            if (node.exported) {
                return "ExportNamespaceSpecifier";
            }
            return "ExportAll";
        }
        return "SideEffectImport";
    }
    /**
     * Returns a boolean indicates if two (import|export) can be merged
     * @param {ASTNode} node1 A node to check.
     * @param {ASTNode} node2 A node to check.
     * @returns {boolean} True if two (import|export) can be merged, false if they can't.
     */
    function isImportExportCanBeMerged(node1, node2) {
        const importExportType1 = getImportExportType(node1);
        const importExportType2 = getImportExportType(node2);
        if ((importExportType1 === "ExportAll" &&
            importExportType2 !== "ExportAll" &&
            importExportType2 !== "SideEffectImport") ||
            (importExportType1 !== "ExportAll" &&
                importExportType1 !== "SideEffectImport" &&
                importExportType2 === "ExportAll")) {
            return false;
        }
        if ((isImportExportSpecifier(importExportType1, "namespace") &&
            isImportExportSpecifier(importExportType2, "named")) ||
            (isImportExportSpecifier(importExportType2, "namespace") &&
                isImportExportSpecifier(importExportType1, "named"))) {
            return false;
        }
        return true;
    }
    /**
     * Returns a boolean if we should report (import|export).
     * @param {ASTNode} node A node to be reported or not.
     * @param {[ASTNode]} previousNodes An array contains previous nodes of the module imported or exported.
     * @returns {boolean} True if the (import|export) should be reported.
     */
    function shouldReportImportExport(node, previousNodes) {
        let i = 0;
        while (i < previousNodes.length) {
            if (isImportExportCanBeMerged(node, previousNodes[i])) {
                return true;
            }
            i++;
        }
        return false;
    }
    /**
     * Returns array contains only nodes with declarations types equal to type.
     * @param {[{node: ASTNode, declarationType: string}]} nodes An array contains objects, each object contains a node and a declaration type.
     * @param {string} type Declaration type.
     * @returns {[ASTNode]} An array contains only nodes with declarations types equal to type.
     */
    function getNodesByDeclarationType(nodes, type) {
        return nodes
            .filter(({ declarationType }) => declarationType === type)
            .map(({ node }) => node);
    }
    /**
     * Returns the name of the module imported or re-exported.
     * @param {ASTNode} node A node to get.
     * @returns {string} The name of the module, or empty string if no name.
     */
    function getModule(node) {
        if (node && node.source && node.source.value) {
            return node.source.value.trim();
        }
        return "";
    }
    /**
     * Checks if the (import|export) can be merged with at least one import or one export, and reports if so.
     * @param {RuleContext} context The ESLint rule context object.
     * @param {ASTNode} node A node to get.
     * @param {Map} modules A Map object contains as a key a module name and as value an array contains objects, each object contains a node and a declaration type.
     * @param {string} declarationType A declaration type can be an import or export.
     * @param {boolean} includeExports Whether or not to check for exports in addition to imports.
     * @returns {void} No return value.
     */
    function checkAndReport(context, node, modules, declarationType, includeExports) {
        const module = getModule(node);
        if (modules.has(module)) {
            const previousNodes = modules.get(module);
            const messagesIds = [];
            const importNodes = getNodesByDeclarationType(previousNodes, "import");
            let exportNodes;
            if (includeExports) {
                exportNodes = getNodesByDeclarationType(previousNodes, "export");
            }
            if (declarationType === "import") {
                if (shouldReportImportExport(node, importNodes)) {
                    messagesIds.push("import");
                }
                if (includeExports) {
                    if (shouldReportImportExport(node, exportNodes)) {
                        messagesIds.push("importAs");
                    }
                }
            }
            else if (declarationType === "export") {
                if (shouldReportImportExport(node, exportNodes)) {
                    messagesIds.push("export");
                }
                if (shouldReportImportExport(node, importNodes)) {
                    messagesIds.push("exportAs");
                }
            }
            messagesIds.forEach(messageId => context.report({
                node,
                messageId,
                data: {
                    module
                }
            }));
        }
    }
    /**
     * @callback nodeCallback
     * @param {ASTNode} node A node to handle.
     */
    /**
     * Returns a function handling the (imports|exports) of a given file
     * @param {RuleContext} context The ESLint rule context object.
     * @param {Map} modules A Map object contains as a key a module name and as value an array contains objects, each object contains a node and a declaration type.
     * @param {string} declarationType A declaration type can be an import or export.
     * @param {boolean} includeExports Whether or not to check for exports in addition to imports.
     * @returns {nodeCallback} A function passed to ESLint to handle the statement.
     */
    function handleImportsExports(context, modules, declarationType, includeExports) {
        return function (node) {
            const module = getModule(node);
            if (module) {
                checkAndReport(context, node, modules, declarationType, includeExports);
                const currentNode = { node, declarationType };
                let nodes = [currentNode];
                if (modules.has(module)) {
                    const previousNodes = modules.get(module);
                    nodes = [...previousNodes, currentNode];
                }
                modules.set(module, nodes);
            }
        };
    }
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow duplicate module imports",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-duplicate-imports"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        includeExports: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                import: "'{{module}}' import is duplicated.",
                importAs: "'{{module}}' import is duplicated as export.",
                export: "'{{module}}' export is duplicated.",
                exportAs: "'{{module}}' export is duplicated as import."
            }
        },
        create(context) {
            const includeExports = (context.options[0] || {}).includeExports, modules = new Map();
            const handlers = {
                ImportDeclaration: handleImportsExports(context, modules, "import", includeExports)
            };
            if (includeExports) {
                handlers.ExportNamedDeclaration = handleImportsExports(context, modules, "export", includeExports);
                handlers.ExportAllDeclaration = handleImportsExports(context, modules, "export", includeExports);
            }
            return handlers;
        }
    };
},
/* eslint/lib/rules/no-else-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const FixTracker = require(508) /* ./utils/fix-tracker */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `else` blocks after `return` statements in `if` statements",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-else-return"
            },
            schema: [{
                    type: "object",
                    properties: {
                        allowElseIf: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }],
            fixable: "code",
            messages: {
                unexpected: "Unnecessary 'else' after 'return'."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Checks whether the given names can be safely used to declare block-scoped variables
             * in the given scope. Name collisions can produce redeclaration syntax errors,
             * or silently change references and modify behavior of the original code.
             *
             * This is not a generic function. In particular, it is assumed that the scope is a function scope or
             * a function's inner scope, and that the names can be valid identifiers in the given scope.
             * @param {string[]} names Array of variable names.
             * @param {eslint-scope.Scope} scope Function scope or a function's inner scope.
             * @returns {boolean} True if all names can be safely declared, false otherwise.
             */
            function isSafeToDeclare(names, scope) {
                if (names.length === 0) {
                    return true;
                }
                const functionScope = scope.variableScope;
                /*
                 * If this is a function scope, scope.variables will contain parameters, implicit variables such as "arguments",
                 * all function-scoped variables ('var'), and block-scoped variables defined in the scope.
                 * If this is an inner scope, scope.variables will contain block-scoped variables defined in the scope.
                 *
                 * Redeclaring any of these would cause a syntax error, except for the implicit variables.
                 */
                const declaredVariables = scope.variables.filter(({ defs }) => defs.length > 0);
                if (declaredVariables.some(({ name }) => names.includes(name))) {
                    return false;
                }
                // Redeclaring a catch variable would also cause a syntax error.
                if (scope !== functionScope && scope.upper.type === "catch") {
                    if (scope.upper.variables.some(({ name }) => names.includes(name))) {
                        return false;
                    }
                }
                /*
                 * Redeclaring an implicit variable, such as "arguments", would not cause a syntax error.
                 * However, if the variable was used, declaring a new one with the same name would change references
                 * and modify behavior.
                 */
                const usedImplicitVariables = scope.variables.filter(({ defs, references }) => defs.length === 0 && references.length > 0);
                if (usedImplicitVariables.some(({ name }) => names.includes(name))) {
                    return false;
                }
                /*
                 * Declaring a variable with a name that was already used to reference a variable from an upper scope
                 * would change references and modify behavior.
                 */
                if (scope.through.some(t => names.includes(t.identifier.name))) {
                    return false;
                }
                /*
                 * If the scope is an inner scope (not the function scope), an uninitialized `var` variable declared inside
                 * the scope node (directly or in one of its descendants) is neither declared nor 'through' in the scope.
                 *
                 * For example, this would be a syntax error "Identifier 'a' has already been declared":
                 * function foo() { if (bar) { let a; if (baz) { var a; } } }
                 */
                if (scope !== functionScope) {
                    const scopeNodeRange = scope.block.range;
                    const variablesToCheck = functionScope.variables.filter(({ name }) => names.includes(name));
                    if (variablesToCheck.some(v => v.defs.some(({ node: { range } }) => scopeNodeRange[0] <= range[0] && range[1] <= scopeNodeRange[1]))) {
                        return false;
                    }
                }
                return true;
            }
            /**
             * Checks whether the removal of `else` and its braces is safe from variable name collisions.
             * @param {Node} node The 'else' node.
             * @param {eslint-scope.Scope} scope The scope in which the node and the whole 'if' statement is.
             * @returns {boolean} True if it is safe, false otherwise.
             */
            function isSafeFromNameCollisions(node, scope) {
                if (node.type === "FunctionDeclaration") {
                    // Conditional function declaration. Scope and hoisting are unpredictable, different engines work differently.
                    return false;
                }
                if (node.type !== "BlockStatement") {
                    return true;
                }
                const elseBlockScope = scope.childScopes.find(({ block }) => block === node);
                if (!elseBlockScope) {
                    // ecmaVersion < 6, `else` block statement cannot have its own scope, no possible collisions.
                    return true;
                }
                /*
                 * elseBlockScope is supposed to merge into its upper scope. elseBlockScope.variables array contains
                 * only block-scoped variables (such as let and const variables or class and function declarations)
                 * defined directly in the elseBlockScope. These are exactly the only names that could cause collisions.
                 */
                const namesToCheck = elseBlockScope.variables.map(({ name }) => name);
                return isSafeToDeclare(namesToCheck, scope);
            }
            /**
             * Display the context report if rule is violated
             * @param {Node} node The 'else' node
             * @returns {void}
             */
            function displayReport(node) {
                const currentScope = context.getScope();
                context.report({
                    node,
                    messageId: "unexpected",
                    fix: fixer => {
                        if (!isSafeFromNameCollisions(node, currentScope)) {
                            return null;
                        }
                        const sourceCode = context.getSourceCode();
                        const startToken = sourceCode.getFirstToken(node);
                        const elseToken = sourceCode.getTokenBefore(startToken);
                        const source = sourceCode.getText(node);
                        const lastIfToken = sourceCode.getTokenBefore(elseToken);
                        let fixedSource, firstTokenOfElseBlock;
                        if (startToken.type === "Punctuator" && startToken.value === "{") {
                            firstTokenOfElseBlock = sourceCode.getTokenAfter(startToken);
                        }
                        else {
                            firstTokenOfElseBlock = startToken;
                        }
                        /*
                         * If the if block does not have curly braces and does not end in a semicolon
                         * and the else block starts with (, [, /, +, ` or -, then it is not
                         * safe to remove the else keyword, because ASI will not add a semicolon
                         * after the if block
                         */
                        const ifBlockMaybeUnsafe = node.parent.consequent.type !== "BlockStatement" && lastIfToken.value !== ";";
                        const elseBlockUnsafe = /^[([/+`-]/u.test(firstTokenOfElseBlock.value);
                        if (ifBlockMaybeUnsafe && elseBlockUnsafe) {
                            return null;
                        }
                        const endToken = sourceCode.getLastToken(node);
                        const lastTokenOfElseBlock = sourceCode.getTokenBefore(endToken);
                        if (lastTokenOfElseBlock.value !== ";") {
                            const nextToken = sourceCode.getTokenAfter(endToken);
                            const nextTokenUnsafe = nextToken && /^[([/+`-]/u.test(nextToken.value);
                            const nextTokenOnSameLine = nextToken && nextToken.loc.start.line === lastTokenOfElseBlock.loc.start.line;
                            /*
                             * If the else block contents does not end in a semicolon,
                             * and the else block starts with (, [, /, +, ` or -, then it is not
                             * safe to remove the else block, because ASI will not add a semicolon
                             * after the remaining else block contents
                             */
                            if (nextTokenUnsafe || (nextTokenOnSameLine && nextToken.value !== "}")) {
                                return null;
                            }
                        }
                        if (startToken.type === "Punctuator" && startToken.value === "{") {
                            fixedSource = source.slice(1, -1);
                        }
                        else {
                            fixedSource = source;
                        }
                        /*
                         * Extend the replacement range to include the entire
                         * function to avoid conflicting with no-useless-return.
                         * https://github.com/eslint/eslint/issues/8026
                         *
                         * Also, to avoid name collisions between two else blocks.
                         */
                        return new FixTracker(fixer, sourceCode)
                            .retainEnclosingFunction(node)
                            .replaceTextRange([elseToken.range[0], node.range[1]], fixedSource);
                    }
                });
            }
            /**
             * Check to see if the node is a ReturnStatement
             * @param {Node} node The node being evaluated
             * @returns {boolean} True if node is a return
             */
            function checkForReturn(node) {
                return node.type === "ReturnStatement";
            }
            /**
             * Naive return checking, does not iterate through the whole
             * BlockStatement because we make the assumption that the ReturnStatement
             * will be the last node in the body of the BlockStatement.
             * @param {Node} node The consequent/alternate node
             * @returns {boolean} True if it has a return
             */
            function naiveHasReturn(node) {
                if (node.type === "BlockStatement") {
                    const body = node.body, lastChildNode = body[body.length - 1];
                    return lastChildNode && checkForReturn(lastChildNode);
                }
                return checkForReturn(node);
            }
            /**
             * Check to see if the node is valid for evaluation,
             * meaning it has an else.
             * @param {Node} node The node being evaluated
             * @returns {boolean} True if the node is valid
             */
            function hasElse(node) {
                return node.alternate && node.consequent;
            }
            /**
             * If the consequent is an IfStatement, check to see if it has an else
             * and both its consequent and alternate path return, meaning this is
             * a nested case of rule violation.  If-Else not considered currently.
             * @param {Node} node The consequent node
             * @returns {boolean} True if this is a nested rule violation
             */
            function checkForIf(node) {
                return node.type === "IfStatement" && hasElse(node) &&
                    naiveHasReturn(node.alternate) && naiveHasReturn(node.consequent);
            }
            /**
             * Check the consequent/body node to make sure it is not
             * a ReturnStatement or an IfStatement that returns on both
             * code paths.
             * @param {Node} node The consequent or body node
             * @returns {boolean} `true` if it is a Return/If node that always returns.
             */
            function checkForReturnOrIf(node) {
                return checkForReturn(node) || checkForIf(node);
            }
            /**
             * Check whether a node returns in every codepath.
             * @param {Node} node The node to be checked
             * @returns {boolean} `true` if it returns on every codepath.
             */
            function alwaysReturns(node) {
                if (node.type === "BlockStatement") {
                    // If we have a BlockStatement, check each consequent body node.
                    return node.body.some(checkForReturnOrIf);
                }
                /*
                 * If not a block statement, make sure the consequent isn't a
                 * ReturnStatement or an IfStatement with returns on both paths.
                 */
                return checkForReturnOrIf(node);
            }
            /**
             * Check the if statement, but don't catch else-if blocks.
             * @returns {void}
             * @param {Node} node The node for the if statement to check
             * @private
             */
            function checkIfWithoutElse(node) {
                const parent = node.parent;
                /*
                 * Fixing this would require splitting one statement into two, so no error should
                 * be reported if this node is in a position where only one statement is allowed.
                 */
                if (!astUtils.STATEMENT_LIST_PARENTS.has(parent.type)) {
                    return;
                }
                const consequents = [];
                let alternate;
                for (let currentNode = node; currentNode.type === "IfStatement"; currentNode = currentNode.alternate) {
                    if (!currentNode.alternate) {
                        return;
                    }
                    consequents.push(currentNode.consequent);
                    alternate = currentNode.alternate;
                }
                if (consequents.every(alwaysReturns)) {
                    displayReport(alternate);
                }
            }
            /**
             * Check the if statement
             * @returns {void}
             * @param {Node} node The node for the if statement to check
             * @private
             */
            function checkIfWithElse(node) {
                const parent = node.parent;
                /*
                 * Fixing this would require splitting one statement into two, so no error should
                 * be reported if this node is in a position where only one statement is allowed.
                 */
                if (!astUtils.STATEMENT_LIST_PARENTS.has(parent.type)) {
                    return;
                }
                const alternate = node.alternate;
                if (alternate && alwaysReturns(node.consequent)) {
                    displayReport(alternate);
                }
            }
            const allowElseIf = !(context.options[0] && context.options[0].allowElseIf === false);
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                "IfStatement:exit": allowElseIf ? checkIfWithoutElse : checkIfWithElse
            };
        }
    };
},
/* eslint/lib/rules/utils/fix-tracker.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./ast-utils */;
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * A helper class to combine fix options into a fix command. Currently, it
     * exposes some "retain" methods that extend the range of the text being
     * replaced so that other fixes won't touch that region in the same pass.
     */
    class FixTracker {
        /**
         * Create a new FixTracker.
         * @param {ruleFixer} fixer A ruleFixer instance.
         * @param {SourceCode} sourceCode A SourceCode object for the current code.
         */
        constructor(fixer, sourceCode) {
            this.fixer = fixer;
            this.sourceCode = sourceCode;
            this.retainedRange = null;
        }
        /**
         * Mark the given range as "retained", meaning that other fixes may not
         * may not modify this region in the same pass.
         * @param {int[]} range The range to retain.
         * @returns {FixTracker} The same RuleFixer, for chained calls.
         */
        retainRange(range) {
            this.retainedRange = range;
            return this;
        }
        /**
         * Given a node, find the function containing it (or the entire program) and
         * mark it as retained, meaning that other fixes may not modify it in this
         * pass. This is useful for avoiding conflicts in fixes that modify control
         * flow.
         * @param {ASTNode} node The node to use as a starting point.
         * @returns {FixTracker} The same RuleFixer, for chained calls.
         */
        retainEnclosingFunction(node) {
            const functionNode = astUtils.getUpperFunction(node);
            return this.retainRange(functionNode ? functionNode.range : this.sourceCode.ast.range);
        }
        /**
         * Given a node or token, find the token before and afterward, and mark that
         * range as retained, meaning that other fixes may not modify it in this
         * pass. This is useful for avoiding conflicts in fixes that make a small
         * change to the code where the AST should not be changed.
         * @param {ASTNode|Token} nodeOrToken The node or token to use as a starting
         *      point. The token to the left and right are use in the range.
         * @returns {FixTracker} The same RuleFixer, for chained calls.
         */
        retainSurroundingTokens(nodeOrToken) {
            const tokenBefore = this.sourceCode.getTokenBefore(nodeOrToken) || nodeOrToken;
            const tokenAfter = this.sourceCode.getTokenAfter(nodeOrToken) || nodeOrToken;
            return this.retainRange([tokenBefore.range[0], tokenAfter.range[1]]);
        }
        /**
         * Create a fix command that replaces the given range with the given text,
         * accounting for any retained ranges.
         * @param {int[]} range The range to remove in the fix.
         * @param {string} text The text to insert in place of the range.
         * @returns {Object} The fix command.
         */
        replaceTextRange(range, text) {
            let actualRange;
            if (this.retainedRange) {
                actualRange = [
                    Math.min(this.retainedRange[0], range[0]),
                    Math.max(this.retainedRange[1], range[1])
                ];
            }
            else {
                actualRange = range;
            }
            return this.fixer.replaceTextRange(actualRange, this.sourceCode.text.slice(actualRange[0], range[0]) +
                text +
                this.sourceCode.text.slice(range[1], actualRange[1]));
        }
        /**
         * Create a fix command that removes the given node or token, accounting for
         * any retained ranges.
         * @param {ASTNode|Token} nodeOrToken The node or token to remove.
         * @returns {Object} The fix command.
         */
        remove(nodeOrToken) {
            return this.replaceTextRange(nodeOrToken.range, "");
        }
    }
    module.exports = FixTracker;
},
/* eslint/lib/rules/no-empty.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow empty block statements",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-empty"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowEmptyCatch: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpected: "Empty {{type}} statement."
            }
        },
        create(context) {
            const options = context.options[0] || {}, allowEmptyCatch = options.allowEmptyCatch || false;
            const sourceCode = context.getSourceCode();
            return {
                BlockStatement(node) {
                    // if the body is not empty, we can just return immediately
                    if (node.body.length !== 0) {
                        return;
                    }
                    // a function is generally allowed to be empty
                    if (astUtils.isFunction(node.parent)) {
                        return;
                    }
                    if (allowEmptyCatch && node.parent.type === "CatchClause") {
                        return;
                    }
                    // any other block is only allowed to be empty, if it contains a comment
                    if (sourceCode.getCommentsInside(node).length > 0) {
                        return;
                    }
                    context.report({ node, messageId: "unexpected", data: { type: "block" } });
                },
                SwitchStatement(node) {
                    if (typeof node.cases === "undefined" || node.cases.length === 0) {
                        context.report({ node, messageId: "unexpected", data: { type: "switch" } });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-empty-character-class.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /*
     * plain-English description of the following regexp:
     * 0. `^` fix the match at the beginning of the string
     * 1. `\/`: the `/` that begins the regexp
     * 2. `([^\\[]|\\.|\[([^\\\]]|\\.)+\])*`: regexp contents; 0 or more of the following
     * 2.0. `[^\\[]`: any character that's not a `\` or a `[` (anything but escape sequences and character classes)
     * 2.1. `\\.`: an escape sequence
     * 2.2. `\[([^\\\]]|\\.)+\]`: a character class that isn't empty
     * 3. `\/` the `/` that ends the regexp
     * 4. `[gimuy]*`: optional regexp flags
     * 5. `$`: fix the match at the end of the string
     */
    const regex = /^\/([^\\[]|\\.|\[([^\\\]]|\\.)+\])*\/[gimuys]*$/u;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow empty character classes in regular expressions",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-empty-character-class"
            },
            schema: [],
            messages: {
                unexpected: "Empty class."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                Literal(node) {
                    const token = sourceCode.getFirstToken(node);
                    if (token.type === "RegularExpression" && !regex.test(token.value)) {
                        context.report({ node, messageId: "unexpected" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-empty-function.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const ALLOW_OPTIONS = Object.freeze([
        "functions",
        "arrowFunctions",
        "generatorFunctions",
        "methods",
        "generatorMethods",
        "getters",
        "setters",
        "constructors",
        "asyncFunctions",
        "asyncMethods"
    ]);
    /**
     * Gets the kind of a given function node.
     * @param {ASTNode} node A function node to get. This is one of
     *      an ArrowFunctionExpression, a FunctionDeclaration, or a
     *      FunctionExpression.
     * @returns {string} The kind of the function. This is one of "functions",
     *      "arrowFunctions", "generatorFunctions", "asyncFunctions", "methods",
     *      "generatorMethods", "asyncMethods", "getters", "setters", and
     *      "constructors".
     */
    function getKind(node) {
        const parent = node.parent;
        let kind = "";
        if (node.type === "ArrowFunctionExpression") {
            return "arrowFunctions";
        }
        // Detects main kind.
        if (parent.type === "Property") {
            if (parent.kind === "get") {
                return "getters";
            }
            if (parent.kind === "set") {
                return "setters";
            }
            kind = parent.method ? "methods" : "functions";
        }
        else if (parent.type === "MethodDefinition") {
            if (parent.kind === "get") {
                return "getters";
            }
            if (parent.kind === "set") {
                return "setters";
            }
            if (parent.kind === "constructor") {
                return "constructors";
            }
            kind = "methods";
        }
        else {
            kind = "functions";
        }
        // Detects prefix.
        let prefix = "";
        if (node.generator) {
            prefix = "generator";
        }
        else if (node.async) {
            prefix = "async";
        }
        else {
            return kind;
        }
        return prefix + kind[0].toUpperCase() + kind.slice(1);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow empty functions",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-empty-function"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allow: {
                            type: "array",
                            items: { enum: ALLOW_OPTIONS },
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpected: "Unexpected empty {{name}}."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const allowed = options.allow || [];
            const sourceCode = context.getSourceCode();
            /**
             * Reports a given function node if the node matches the following patterns.
             *
             * - Not allowed by options.
             * - The body is empty.
             * - The body doesn't have any comments.
             * @param {ASTNode} node A function node to report. This is one of
             *      an ArrowFunctionExpression, a FunctionDeclaration, or a
             *      FunctionExpression.
             * @returns {void}
             */
            function reportIfEmpty(node) {
                const kind = getKind(node);
                const name = astUtils.getFunctionNameWithKind(node);
                const innerComments = sourceCode.getTokens(node.body, {
                    includeComments: true,
                    filter: astUtils.isCommentToken
                });
                if (allowed.indexOf(kind) === -1 &&
                    node.body.type === "BlockStatement" &&
                    node.body.body.length === 0 &&
                    innerComments.length === 0) {
                    context.report({
                        node,
                        loc: node.body.loc,
                        messageId: "unexpected",
                        data: { name }
                    });
                }
            }
            return {
                ArrowFunctionExpression: reportIfEmpty,
                FunctionDeclaration: reportIfEmpty,
                FunctionExpression: reportIfEmpty
            };
        }
    };
},
/* eslint/lib/rules/no-empty-pattern.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow empty destructuring patterns",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-empty-pattern"
            },
            schema: [],
            messages: {
                unexpected: "Unexpected empty {{type}} pattern."
            }
        },
        create(context) {
            return {
                ObjectPattern(node) {
                    if (node.properties.length === 0) {
                        context.report({ node, messageId: "unexpected", data: { type: "object" } });
                    }
                },
                ArrayPattern(node) {
                    if (node.elements.length === 0) {
                        context.report({ node, messageId: "unexpected", data: { type: "array" } });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-eq-null.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `null` comparisons without type-checking operators",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-eq-null"
            },
            schema: [],
            messages: {
                unexpected: "Use '===' to compare with null."
            }
        },
        create(context) {
            return {
                BinaryExpression(node) {
                    const badOperator = node.operator === "==" || node.operator === "!=";
                    if (node.right.type === "Literal" && node.right.raw === "null" && badOperator ||
                        node.left.type === "Literal" && node.left.raw === "null" && badOperator) {
                        context.report({ node, messageId: "unexpected" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-eval.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const candidatesOfGlobalObject = Object.freeze([
        "global",
        "window",
        "globalThis"
    ]);
    /**
     * Checks a given node is a MemberExpression node which has the specified name's
     * property.
     * @param {ASTNode} node A node to check.
     * @param {string} name A name to check.
     * @returns {boolean} `true` if the node is a MemberExpression node which has
     *      the specified name's property
     */
    function isMember(node, name) {
        return astUtils.isSpecificMemberAccess(node, null, name);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the use of `eval()`",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-eval"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowIndirect: { type: "boolean", default: false }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpected: "eval can be harmful."
            }
        },
        create(context) {
            const allowIndirect = Boolean(context.options[0] &&
                context.options[0].allowIndirect);
            const sourceCode = context.getSourceCode();
            let funcInfo = null;
            /**
             * Pushs a variable scope (Program or Function) information to the stack.
             *
             * This is used in order to check whether or not `this` binding is a
             * reference to the global object.
             * @param {ASTNode} node A node of the scope. This is one of Program,
             *      FunctionDeclaration, FunctionExpression, and ArrowFunctionExpression.
             * @returns {void}
             */
            function enterVarScope(node) {
                const strict = context.getScope().isStrict;
                funcInfo = {
                    upper: funcInfo,
                    node,
                    strict,
                    defaultThis: false,
                    initialized: strict
                };
            }
            /**
             * Pops a variable scope from the stack.
             * @returns {void}
             */
            function exitVarScope() {
                funcInfo = funcInfo.upper;
            }
            /**
             * Reports a given node.
             *
             * `node` is `Identifier` or `MemberExpression`.
             * The parent of `node` might be `CallExpression`.
             *
             * The location of the report is always `eval` `Identifier` (or possibly
             * `Literal`). The type of the report is `CallExpression` if the parent is
             * `CallExpression`. Otherwise, it's the given node type.
             * @param {ASTNode} node A node to report.
             * @returns {void}
             */
            function report(node) {
                const parent = node.parent;
                const locationNode = node.type === "MemberExpression"
                    ? node.property
                    : node;
                const reportNode = parent.type === "CallExpression" && parent.callee === node
                    ? parent
                    : node;
                context.report({
                    node: reportNode,
                    loc: locationNode.loc,
                    messageId: "unexpected"
                });
            }
            /**
             * Reports accesses of `eval` via the global object.
             * @param {eslint-scope.Scope} globalScope The global scope.
             * @returns {void}
             */
            function reportAccessingEvalViaGlobalObject(globalScope) {
                for (let i = 0; i < candidatesOfGlobalObject.length; ++i) {
                    const name = candidatesOfGlobalObject[i];
                    const variable = astUtils.getVariableByName(globalScope, name);
                    if (!variable) {
                        continue;
                    }
                    const references = variable.references;
                    for (let j = 0; j < references.length; ++j) {
                        const identifier = references[j].identifier;
                        let node = identifier.parent;
                        // To detect code like `window.window.eval`.
                        while (isMember(node, name)) {
                            node = node.parent;
                        }
                        // Reports.
                        if (isMember(node, "eval")) {
                            report(node);
                        }
                    }
                }
            }
            /**
             * Reports all accesses of `eval` (excludes direct calls to eval).
             * @param {eslint-scope.Scope} globalScope The global scope.
             * @returns {void}
             */
            function reportAccessingEval(globalScope) {
                const variable = astUtils.getVariableByName(globalScope, "eval");
                if (!variable) {
                    return;
                }
                const references = variable.references;
                for (let i = 0; i < references.length; ++i) {
                    const reference = references[i];
                    const id = reference.identifier;
                    if (id.name === "eval" && !astUtils.isCallee(id)) {
                        // Is accessing to eval (excludes direct calls to eval)
                        report(id);
                    }
                }
            }
            if (allowIndirect) {
                // Checks only direct calls to eval. It's simple!
                return {
                    "CallExpression:exit"(node) {
                        const callee = node.callee;
                        /*
                         * Optional call (`eval?.("code")`) is not direct eval.
                         * The direct eval is only step 6.a.vi of https://tc39.es/ecma262/#sec-function-calls-runtime-semantics-evaluation
                         * But the optional call is https://tc39.es/ecma262/#sec-optional-chaining-chain-evaluation
                         */
                        if (!node.optional && astUtils.isSpecificId(callee, "eval")) {
                            report(callee);
                        }
                    }
                };
            }
            return {
                "CallExpression:exit"(node) {
                    const callee = node.callee;
                    if (astUtils.isSpecificId(callee, "eval")) {
                        report(callee);
                    }
                },
                Program(node) {
                    const scope = context.getScope(), features = context.parserOptions.ecmaFeatures || {}, strict = scope.isStrict ||
                        node.sourceType === "module" ||
                        (features.globalReturn && scope.childScopes[0].isStrict);
                    funcInfo = {
                        upper: null,
                        node,
                        strict,
                        defaultThis: true,
                        initialized: true
                    };
                },
                "Program:exit"() {
                    const globalScope = context.getScope();
                    exitVarScope();
                    reportAccessingEval(globalScope);
                    reportAccessingEvalViaGlobalObject(globalScope);
                },
                FunctionDeclaration: enterVarScope,
                "FunctionDeclaration:exit": exitVarScope,
                FunctionExpression: enterVarScope,
                "FunctionExpression:exit": exitVarScope,
                ArrowFunctionExpression: enterVarScope,
                "ArrowFunctionExpression:exit": exitVarScope,
                ThisExpression(node) {
                    if (!isMember(node.parent, "eval")) {
                        return;
                    }
                    /*
                     * `this.eval` is found.
                     * Checks whether or not the value of `this` is the global object.
                     */
                    if (!funcInfo.initialized) {
                        funcInfo.initialized = true;
                        funcInfo.defaultThis = astUtils.isDefaultThisBinding(funcInfo.node, sourceCode);
                    }
                    if (!funcInfo.strict && funcInfo.defaultThis) {
                        // `this.eval` is possible built-in `eval`.
                        report(node.parent);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-ex-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow reassigning exceptions in `catch` clauses",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-ex-assign"
            },
            schema: [],
            messages: {
                unexpected: "Do not assign to the exception parameter."
            }
        },
        create(context) {
            /**
             * Finds and reports references that are non initializer and writable.
             * @param {Variable} variable A variable to check.
             * @returns {void}
             */
            function checkVariable(variable) {
                astUtils.getModifyingReferences(variable.references).forEach(reference => {
                    context.report({ node: reference.identifier, messageId: "unexpected" });
                });
            }
            return {
                CatchClause(node) {
                    context.getDeclaredVariables(node).forEach(checkVariable);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-extend-native.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const globals = require(218) /* globals */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow extending native types",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-extend-native"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        exceptions: {
                            type: "array",
                            items: {
                                type: "string"
                            },
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpected: "{{builtin}} prototype is read only, properties should not be added."
            }
        },
        create(context) {
            const config = context.options[0] || {};
            const exceptions = new Set(config.exceptions || []);
            const modifiedBuiltins = new Set(Object.keys(globals.builtin)
                .filter(builtin => builtin[0].toUpperCase() === builtin[0])
                .filter(builtin => !exceptions.has(builtin)));
            /**
             * Reports a lint error for the given node.
             * @param {ASTNode} node The node to report.
             * @param {string} builtin The name of the native builtin being extended.
             * @returns {void}
             */
            function reportNode(node, builtin) {
                context.report({
                    node,
                    messageId: "unexpected",
                    data: {
                        builtin
                    }
                });
            }
            /**
             * Check to see if the `prototype` property of the given object
             * identifier node is being accessed.
             * @param {ASTNode} identifierNode The Identifier representing the object
             * to check.
             * @returns {boolean} True if the identifier is the object of a
             * MemberExpression and its `prototype` property is being accessed,
             * false otherwise.
             */
            function isPrototypePropertyAccessed(identifierNode) {
                return Boolean(identifierNode &&
                    identifierNode.parent &&
                    identifierNode.parent.type === "MemberExpression" &&
                    identifierNode.parent.object === identifierNode &&
                    astUtils.getStaticPropertyName(identifierNode.parent) === "prototype");
            }
            /**
             * Check if it's an assignment to the property of the given node.
             * Example: `*.prop = 0` // the `*` is the given node.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} True if an assignment to the property of the node.
             */
            function isAssigningToPropertyOf(node) {
                return (node.parent.type === "MemberExpression" &&
                    node.parent.object === node &&
                    node.parent.parent.type === "AssignmentExpression" &&
                    node.parent.parent.left === node.parent);
            }
            /**
             * Checks if the given node is at the first argument of the method call of `Object.defineProperty()` or `Object.defineProperties()`.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} True if the node is at the first argument of the method call of `Object.defineProperty()` or `Object.defineProperties()`.
             */
            function isInDefinePropertyCall(node) {
                return (node.parent.type === "CallExpression" &&
                    node.parent.arguments[0] === node &&
                    astUtils.isSpecificMemberAccess(node.parent.callee, "Object", /^definePropert(?:y|ies)$/u));
            }
            /**
             * Check to see if object prototype access is part of a prototype
             * extension. There are three ways a prototype can be extended:
             * 1. Assignment to prototype property (Object.prototype.foo = 1)
             * 2. Object.defineProperty()/Object.defineProperties() on a prototype
             * If prototype extension is detected, report the AssignmentExpression
             * or CallExpression node.
             * @param {ASTNode} identifierNode The Identifier representing the object
             * which prototype is being accessed and possibly extended.
             * @returns {void}
             */
            function checkAndReportPrototypeExtension(identifierNode) {
                if (!isPrototypePropertyAccessed(identifierNode)) {
                    return; // This is not `*.prototype` access.
                }
                /*
                 * `identifierNode.parent` is a MemberExpression `*.prototype`.
                 * If it's an optional member access, it may be wrapped by a `ChainExpression` node.
                 */
                const prototypeNode = identifierNode.parent.parent.type === "ChainExpression"
                    ? identifierNode.parent.parent
                    : identifierNode.parent;
                if (isAssigningToPropertyOf(prototypeNode)) {
                    // `*.prototype` -> MemberExpression -> AssignmentExpression
                    reportNode(prototypeNode.parent.parent, identifierNode.name);
                }
                else if (isInDefinePropertyCall(prototypeNode)) {
                    // `*.prototype` -> CallExpression
                    reportNode(prototypeNode.parent, identifierNode.name);
                }
            }
            return {
                "Program:exit"() {
                    const globalScope = context.getScope();
                    modifiedBuiltins.forEach(builtin => {
                        const builtinVar = globalScope.set.get(builtin);
                        if (builtinVar && builtinVar.references) {
                            builtinVar.references
                                .map(ref => ref.identifier)
                                .forEach(checkAndReportPrototypeExtension);
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-extra-bind.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const SIDE_EFFECT_FREE_NODE_TYPES = new Set(["Literal", "Identifier", "ThisExpression", "FunctionExpression"]);
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary calls to `.bind()`",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-extra-bind"
            },
            schema: [],
            fixable: "code",
            messages: {
                unexpected: "The function binding is unnecessary."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            let scopeInfo = null;
            /**
             * Checks if a node is free of side effects.
             *
             * This check is stricter than it needs to be, in order to keep the implementation simple.
             * @param {ASTNode} node A node to check.
             * @returns {boolean} True if the node is known to be side-effect free, false otherwise.
             */
            function isSideEffectFree(node) {
                return SIDE_EFFECT_FREE_NODE_TYPES.has(node.type);
            }
            /**
             * Reports a given function node.
             * @param {ASTNode} node A node to report. This is a FunctionExpression or
             *      an ArrowFunctionExpression.
             * @returns {void}
             */
            function report(node) {
                const memberNode = node.parent;
                const callNode = memberNode.parent.type === "ChainExpression"
                    ? memberNode.parent.parent
                    : memberNode.parent;
                context.report({
                    node: callNode,
                    messageId: "unexpected",
                    loc: memberNode.property.loc,
                    fix(fixer) {
                        if (!isSideEffectFree(callNode.arguments[0])) {
                            return null;
                        }
                        /*
                         * The list of the first/last token pair of a removal range.
                         * This is two parts because closing parentheses may exist between the method name and arguments.
                         * E.g. `(function(){}.bind ) (obj)`
                         *                    ^^^^^   ^^^^^ < removal ranges
                         * E.g. `(function(){}?.['bind'] ) ?.(obj)`
                         *                    ^^^^^^^^^^   ^^^^^^^ < removal ranges
                         */
                        const tokenPairs = [
                            [
                                // `.`, `?.`, or `[` token.
                                sourceCode.getTokenAfter(memberNode.object, astUtils.isNotClosingParenToken),
                                // property name or `]` token.
                                sourceCode.getLastToken(memberNode)
                            ],
                            [
                                // `?.` or `(` token of arguments.
                                sourceCode.getTokenAfter(memberNode, astUtils.isNotClosingParenToken),
                                // `)` token of arguments.
                                sourceCode.getLastToken(callNode)
                            ]
                        ];
                        const firstTokenToRemove = tokenPairs[0][0];
                        const lastTokenToRemove = tokenPairs[1][1];
                        if (sourceCode.commentsExistBetween(firstTokenToRemove, lastTokenToRemove)) {
                            return null;
                        }
                        return tokenPairs.map(([start, end]) => fixer.removeRange([start.range[0], end.range[1]]));
                    }
                });
            }
            /**
             * Checks whether or not a given function node is the callee of `.bind()`
             * method.
             *
             * e.g. `(function() {}.bind(foo))`
             * @param {ASTNode} node A node to report. This is a FunctionExpression or
             *      an ArrowFunctionExpression.
             * @returns {boolean} `true` if the node is the callee of `.bind()` method.
             */
            function isCalleeOfBindMethod(node) {
                if (!astUtils.isSpecificMemberAccess(node.parent, null, "bind")) {
                    return false;
                }
                // The node of `*.bind` member access.
                const bindNode = node.parent.parent.type === "ChainExpression"
                    ? node.parent.parent
                    : node.parent;
                return (bindNode.parent.type === "CallExpression" &&
                    bindNode.parent.callee === bindNode &&
                    bindNode.parent.arguments.length === 1 &&
                    bindNode.parent.arguments[0].type !== "SpreadElement");
            }
            /**
             * Adds a scope information object to the stack.
             * @param {ASTNode} node A node to add. This node is a FunctionExpression
             *      or a FunctionDeclaration node.
             * @returns {void}
             */
            function enterFunction(node) {
                scopeInfo = {
                    isBound: isCalleeOfBindMethod(node),
                    thisFound: false,
                    upper: scopeInfo
                };
            }
            /**
             * Removes the scope information object from the top of the stack.
             * At the same time, this reports the function node if the function has
             * `.bind()` and the `this` keywords found.
             * @param {ASTNode} node A node to remove. This node is a
             *      FunctionExpression or a FunctionDeclaration node.
             * @returns {void}
             */
            function exitFunction(node) {
                if (scopeInfo.isBound && !scopeInfo.thisFound) {
                    report(node);
                }
                scopeInfo = scopeInfo.upper;
            }
            /**
             * Reports a given arrow function if the function is callee of `.bind()`
             * method.
             * @param {ASTNode} node A node to report. This node is an
             *      ArrowFunctionExpression.
             * @returns {void}
             */
            function exitArrowFunction(node) {
                if (isCalleeOfBindMethod(node)) {
                    report(node);
                }
            }
            /**
             * Set the mark as the `this` keyword was found in this scope.
             * @returns {void}
             */
            function markAsThisFound() {
                if (scopeInfo) {
                    scopeInfo.thisFound = true;
                }
            }
            return {
                "ArrowFunctionExpression:exit": exitArrowFunction,
                FunctionDeclaration: enterFunction,
                "FunctionDeclaration:exit": exitFunction,
                FunctionExpression: enterFunction,
                "FunctionExpression:exit": exitFunction,
                ThisExpression: markAsThisFound
            };
        }
    };
},
/* eslint/lib/rules/no-extra-boolean-cast.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const eslintUtils = require(350) /* eslint-utils */;
    const precedence = astUtils.getPrecedence;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary boolean casts",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-extra-boolean-cast"
            },
            schema: [{
                    type: "object",
                    properties: {
                        enforceForLogicalOperands: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }],
            fixable: "code",
            messages: {
                unexpectedCall: "Redundant Boolean call.",
                unexpectedNegation: "Redundant double negation."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            // Node types which have a test which will coerce values to booleans.
            const BOOLEAN_NODE_TYPES = [
                "IfStatement",
                "DoWhileStatement",
                "WhileStatement",
                "ConditionalExpression",
                "ForStatement"
            ];
            /**
             * Check if a node is a Boolean function or constructor.
             * @param {ASTNode} node the node
             * @returns {boolean} If the node is Boolean function or constructor
             */
            function isBooleanFunctionOrConstructorCall(node) {
                // Boolean(<bool>) and new Boolean(<bool>)
                return (node.type === "CallExpression" || node.type === "NewExpression") &&
                    node.callee.type === "Identifier" &&
                    node.callee.name === "Boolean";
            }
            /**
             * Checks whether the node is a logical expression and that the option is enabled
             * @param {ASTNode} node the node
             * @returns {boolean} if the node is a logical expression and option is enabled
             */
            function isLogicalContext(node) {
                return node.type === "LogicalExpression" &&
                    (node.operator === "||" || node.operator === "&&") &&
                    (context.options.length && context.options[0].enforceForLogicalOperands === true);
            }
            /**
             * Check if a node is in a context where its value would be coerced to a boolean at runtime.
             * @param {ASTNode} node The node
             * @returns {boolean} If it is in a boolean context
             */
            function isInBooleanContext(node) {
                return ((isBooleanFunctionOrConstructorCall(node.parent) &&
                    node === node.parent.arguments[0]) ||
                    (BOOLEAN_NODE_TYPES.indexOf(node.parent.type) !== -1 &&
                        node === node.parent.test) ||
                    // !<bool>
                    (node.parent.type === "UnaryExpression" &&
                        node.parent.operator === "!"));
            }
            /**
             * Checks whether the node is a context that should report an error
             * Acts recursively if it is in a logical context
             * @param {ASTNode} node the node
             * @returns {boolean} If the node is in one of the flagged contexts
             */
            function isInFlaggedContext(node) {
                if (node.parent.type === "ChainExpression") {
                    return isInFlaggedContext(node.parent);
                }
                return isInBooleanContext(node) ||
                    (isLogicalContext(node.parent) &&
                        // For nested logical statements
                        isInFlaggedContext(node.parent));
            }
            /**
             * Check if a node has comments inside.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} `true` if it has comments inside.
             */
            function hasCommentsInside(node) {
                return Boolean(sourceCode.getCommentsInside(node).length);
            }
            /**
             * Checks if the given node is wrapped in grouping parentheses. Parentheses for constructs such as if() don't count.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} `true` if the node is parenthesized.
             * @private
             */
            function isParenthesized(node) {
                return eslintUtils.isParenthesized(1, node, sourceCode);
            }
            /**
             * Determines whether the given node needs to be parenthesized when replacing the previous node.
             * It assumes that `previousNode` is the node to be reported by this rule, so it has a limited list
             * of possible parent node types. By the same assumption, the node's role in a particular parent is already known.
             * For example, if the parent is `ConditionalExpression`, `previousNode` must be its `test` child.
             * @param {ASTNode} previousNode Previous node.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} `true` if the node needs to be parenthesized.
             */
            function needsParens(previousNode, node) {
                if (previousNode.parent.type === "ChainExpression") {
                    return needsParens(previousNode.parent, node);
                }
                if (isParenthesized(previousNode)) {
                    // parentheses around the previous node will stay, so there is no need for an additional pair
                    return false;
                }
                // parent of the previous node will become parent of the replacement node
                const parent = previousNode.parent;
                switch (parent.type) {
                    case "CallExpression":
                    case "NewExpression":
                        return node.type === "SequenceExpression";
                    case "IfStatement":
                    case "DoWhileStatement":
                    case "WhileStatement":
                    case "ForStatement":
                        return false;
                    case "ConditionalExpression":
                        return precedence(node) <= precedence(parent);
                    case "UnaryExpression":
                        return precedence(node) < precedence(parent);
                    case "LogicalExpression":
                        if (astUtils.isMixedLogicalAndCoalesceExpressions(node, parent)) {
                            return true;
                        }
                        if (previousNode === parent.left) {
                            return precedence(node) < precedence(parent);
                        }
                        return precedence(node) <= precedence(parent);
                    /* istanbul ignore next */
                    default:
                        throw new Error(`Unexpected parent type: ${parent.type}`);
                }
            }
            return {
                UnaryExpression(node) {
                    const parent = node.parent;
                    // Exit early if it's guaranteed not to match
                    if (node.operator !== "!" ||
                        parent.type !== "UnaryExpression" ||
                        parent.operator !== "!") {
                        return;
                    }
                    if (isInFlaggedContext(parent)) {
                        context.report({
                            node: parent,
                            messageId: "unexpectedNegation",
                            fix(fixer) {
                                if (hasCommentsInside(parent)) {
                                    return null;
                                }
                                if (needsParens(parent, node.argument)) {
                                    return fixer.replaceText(parent, `(${sourceCode.getText(node.argument)})`);
                                }
                                let prefix = "";
                                const tokenBefore = sourceCode.getTokenBefore(parent);
                                const firstReplacementToken = sourceCode.getFirstToken(node.argument);
                                if (tokenBefore &&
                                    tokenBefore.range[1] === parent.range[0] &&
                                    !astUtils.canTokensBeAdjacent(tokenBefore, firstReplacementToken)) {
                                    prefix = " ";
                                }
                                return fixer.replaceText(parent, prefix + sourceCode.getText(node.argument));
                            }
                        });
                    }
                },
                CallExpression(node) {
                    if (node.callee.type !== "Identifier" || node.callee.name !== "Boolean") {
                        return;
                    }
                    if (isInFlaggedContext(node)) {
                        context.report({
                            node,
                            messageId: "unexpectedCall",
                            fix(fixer) {
                                const parent = node.parent;
                                if (node.arguments.length === 0) {
                                    if (parent.type === "UnaryExpression" && parent.operator === "!") {
                                        /*
                                         * !Boolean() -> true
                                         */
                                        if (hasCommentsInside(parent)) {
                                            return null;
                                        }
                                        const replacement = "true";
                                        let prefix = "";
                                        const tokenBefore = sourceCode.getTokenBefore(parent);
                                        if (tokenBefore &&
                                            tokenBefore.range[1] === parent.range[0] &&
                                            !astUtils.canTokensBeAdjacent(tokenBefore, replacement)) {
                                            prefix = " ";
                                        }
                                        return fixer.replaceText(parent, prefix + replacement);
                                    }
                                    /*
                                     * Boolean() -> false
                                     */
                                    if (hasCommentsInside(node)) {
                                        return null;
                                    }
                                    return fixer.replaceText(node, "false");
                                }
                                if (node.arguments.length === 1) {
                                    const argument = node.arguments[0];
                                    if (argument.type === "SpreadElement" || hasCommentsInside(node)) {
                                        return null;
                                    }
                                    /*
                                     * Boolean(expression) -> expression
                                     */
                                    if (needsParens(node, argument)) {
                                        return fixer.replaceText(node, `(${sourceCode.getText(argument)})`);
                                    }
                                    return fixer.replaceText(node, sourceCode.getText(argument));
                                }
                                // two or more arguments
                                return null;
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-extra-label.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary labels",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-extra-label"
            },
            schema: [],
            fixable: "code",
            messages: {
                unexpected: "This label '{{name}}' is unnecessary."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            let scopeInfo = null;
            /**
             * Creates a new scope with a breakable statement.
             * @param {ASTNode} node A node to create. This is a BreakableStatement.
             * @returns {void}
             */
            function enterBreakableStatement(node) {
                scopeInfo = {
                    label: node.parent.type === "LabeledStatement" ? node.parent.label : null,
                    breakable: true,
                    upper: scopeInfo
                };
            }
            /**
             * Removes the top scope of the stack.
             * @returns {void}
             */
            function exitBreakableStatement() {
                scopeInfo = scopeInfo.upper;
            }
            /**
             * Creates a new scope with a labeled statement.
             *
             * This ignores it if the body is a breakable statement.
             * In this case it's handled in the `enterBreakableStatement` function.
             * @param {ASTNode} node A node to create. This is a LabeledStatement.
             * @returns {void}
             */
            function enterLabeledStatement(node) {
                if (!astUtils.isBreakableStatement(node.body)) {
                    scopeInfo = {
                        label: node.label,
                        breakable: false,
                        upper: scopeInfo
                    };
                }
            }
            /**
             * Removes the top scope of the stack.
             *
             * This ignores it if the body is a breakable statement.
             * In this case it's handled in the `exitBreakableStatement` function.
             * @param {ASTNode} node A node. This is a LabeledStatement.
             * @returns {void}
             */
            function exitLabeledStatement(node) {
                if (!astUtils.isBreakableStatement(node.body)) {
                    scopeInfo = scopeInfo.upper;
                }
            }
            /**
             * Reports a given control node if it's unnecessary.
             * @param {ASTNode} node A node. This is a BreakStatement or a
             *      ContinueStatement.
             * @returns {void}
             */
            function reportIfUnnecessary(node) {
                if (!node.label) {
                    return;
                }
                const labelNode = node.label;
                for (let info = scopeInfo; info !== null; info = info.upper) {
                    if (info.breakable || info.label && info.label.name === labelNode.name) {
                        if (info.breakable && info.label && info.label.name === labelNode.name) {
                            context.report({
                                node: labelNode,
                                messageId: "unexpected",
                                data: labelNode,
                                fix(fixer) {
                                    const breakOrContinueToken = sourceCode.getFirstToken(node);
                                    if (sourceCode.commentsExistBetween(breakOrContinueToken, labelNode)) {
                                        return null;
                                    }
                                    return fixer.removeRange([breakOrContinueToken.range[1], labelNode.range[1]]);
                                }
                            });
                        }
                        return;
                    }
                }
            }
            return {
                WhileStatement: enterBreakableStatement,
                "WhileStatement:exit": exitBreakableStatement,
                DoWhileStatement: enterBreakableStatement,
                "DoWhileStatement:exit": exitBreakableStatement,
                ForStatement: enterBreakableStatement,
                "ForStatement:exit": exitBreakableStatement,
                ForInStatement: enterBreakableStatement,
                "ForInStatement:exit": exitBreakableStatement,
                ForOfStatement: enterBreakableStatement,
                "ForOfStatement:exit": exitBreakableStatement,
                SwitchStatement: enterBreakableStatement,
                "SwitchStatement:exit": exitBreakableStatement,
                LabeledStatement: enterLabeledStatement,
                "LabeledStatement:exit": exitLabeledStatement,
                BreakStatement: reportIfUnnecessary,
                ContinueStatement: reportIfUnnecessary
            };
        }
    };
},
/* eslint/lib/rules/no-extra-parens.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const { isParenthesized: isParenthesizedRaw } = require(350) /* eslint-utils */;
    const astUtils = require(394) /* ./utils/ast-utils.js */;
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "disallow unnecessary parentheses",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-extra-parens"
            },
            fixable: "code",
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["functions"]
                            }
                        ],
                        minItems: 0,
                        maxItems: 1
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["all"]
                            },
                            {
                                type: "object",
                                properties: {
                                    conditionalAssign: { type: "boolean" },
                                    nestedBinaryExpressions: { type: "boolean" },
                                    returnAssign: { type: "boolean" },
                                    ignoreJSX: { enum: ["none", "all", "single-line", "multi-line"] },
                                    enforceForArrowConditionals: { type: "boolean" },
                                    enforceForSequenceExpressions: { type: "boolean" },
                                    enforceForNewInMemberExpressions: { type: "boolean" },
                                    enforceForFunctionPrototypeMethods: { type: "boolean" }
                                },
                                additionalProperties: false
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    }
                ]
            },
            messages: {
                unexpected: "Unnecessary parentheses around expression."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const tokensToIgnore = new WeakSet();
            const precedence = astUtils.getPrecedence;
            const ALL_NODES = context.options[0] !== "functions";
            const EXCEPT_COND_ASSIGN = ALL_NODES && context.options[1] && context.options[1].conditionalAssign === false;
            const NESTED_BINARY = ALL_NODES && context.options[1] && context.options[1].nestedBinaryExpressions === false;
            const EXCEPT_RETURN_ASSIGN = ALL_NODES && context.options[1] && context.options[1].returnAssign === false;
            const IGNORE_JSX = ALL_NODES && context.options[1] && context.options[1].ignoreJSX;
            const IGNORE_ARROW_CONDITIONALS = ALL_NODES && context.options[1] &&
                context.options[1].enforceForArrowConditionals === false;
            const IGNORE_SEQUENCE_EXPRESSIONS = ALL_NODES && context.options[1] &&
                context.options[1].enforceForSequenceExpressions === false;
            const IGNORE_NEW_IN_MEMBER_EXPR = ALL_NODES && context.options[1] &&
                context.options[1].enforceForNewInMemberExpressions === false;
            const IGNORE_FUNCTION_PROTOTYPE_METHODS = ALL_NODES && context.options[1] &&
                context.options[1].enforceForFunctionPrototypeMethods === false;
            const PRECEDENCE_OF_ASSIGNMENT_EXPR = precedence({ type: "AssignmentExpression" });
            const PRECEDENCE_OF_UPDATE_EXPR = precedence({ type: "UpdateExpression" });
            let reportsBuffer;
            /**
             * Determines whether the given node is a `call` or `apply` method call, invoked directly on a `FunctionExpression` node.
             * Example: function(){}.call()
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the node is an immediate `call` or `apply` method call.
             * @private
             */
            function isImmediateFunctionPrototypeMethodCall(node) {
                const callNode = astUtils.skipChainExpression(node);
                if (callNode.type !== "CallExpression") {
                    return false;
                }
                const callee = astUtils.skipChainExpression(callNode.callee);
                return (callee.type === "MemberExpression" &&
                    callee.object.type === "FunctionExpression" &&
                    ["call", "apply"].includes(astUtils.getStaticPropertyName(callee)));
            }
            /**
             * Determines if this rule should be enforced for a node given the current configuration.
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the rule should be enforced for this node.
             * @private
             */
            function ruleApplies(node) {
                if (node.type === "JSXElement" || node.type === "JSXFragment") {
                    const isSingleLine = node.loc.start.line === node.loc.end.line;
                    switch (IGNORE_JSX) {
                        // Exclude this JSX element from linting
                        case "all":
                            return false;
                        // Exclude this JSX element if it is multi-line element
                        case "multi-line":
                            return isSingleLine;
                        // Exclude this JSX element if it is single-line element
                        case "single-line":
                            return !isSingleLine;
                        // Nothing special to be done for JSX elements
                        case "none":
                            break;
                        // no default
                    }
                }
                if (node.type === "SequenceExpression" && IGNORE_SEQUENCE_EXPRESSIONS) {
                    return false;
                }
                if (isImmediateFunctionPrototypeMethodCall(node) && IGNORE_FUNCTION_PROTOTYPE_METHODS) {
                    return false;
                }
                return ALL_NODES || node.type === "FunctionExpression" || node.type === "ArrowFunctionExpression";
            }
            /**
             * Determines if a node is surrounded by parentheses.
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the node is parenthesised.
             * @private
             */
            function isParenthesised(node) {
                return isParenthesizedRaw(1, node, sourceCode);
            }
            /**
             * Determines if a node is surrounded by parentheses twice.
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the node is doubly parenthesised.
             * @private
             */
            function isParenthesisedTwice(node) {
                return isParenthesizedRaw(2, node, sourceCode);
            }
            /**
             * Determines if a node is surrounded by (potentially) invalid parentheses.
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the node is incorrectly parenthesised.
             * @private
             */
            function hasExcessParens(node) {
                return ruleApplies(node) && isParenthesised(node);
            }
            /**
             * Determines if a node that is expected to be parenthesised is surrounded by
             * (potentially) invalid extra parentheses.
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the node is has an unexpected extra pair of parentheses.
             * @private
             */
            function hasDoubleExcessParens(node) {
                return ruleApplies(node) && isParenthesisedTwice(node);
            }
            /**
             * Determines if a node that is expected to be parenthesised is surrounded by
             * (potentially) invalid extra parentheses with considering precedence level of the node.
             * If the preference level of the node is not higher or equal to precedence lower limit, it also checks
             * whether the node is surrounded by parentheses twice or not.
             * @param {ASTNode} node The node to be checked.
             * @param {number} precedenceLowerLimit The lower limit of precedence.
             * @returns {boolean} True if the node is has an unexpected extra pair of parentheses.
             * @private
             */
            function hasExcessParensWithPrecedence(node, precedenceLowerLimit) {
                if (ruleApplies(node) && isParenthesised(node)) {
                    if (precedence(node) >= precedenceLowerLimit ||
                        isParenthesisedTwice(node)) {
                        return true;
                    }
                }
                return false;
            }
            /**
             * Determines if a node test expression is allowed to have a parenthesised assignment
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the assignment can be parenthesised.
             * @private
             */
            function isCondAssignException(node) {
                return EXCEPT_COND_ASSIGN && node.test.type === "AssignmentExpression";
            }
            /**
             * Determines if a node is in a return statement
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the node is in a return statement.
             * @private
             */
            function isInReturnStatement(node) {
                for (let currentNode = node; currentNode; currentNode = currentNode.parent) {
                    if (currentNode.type === "ReturnStatement" ||
                        (currentNode.type === "ArrowFunctionExpression" && currentNode.body.type !== "BlockStatement")) {
                        return true;
                    }
                }
                return false;
            }
            /**
             * Determines if a constructor function is newed-up with parens
             * @param {ASTNode} newExpression The NewExpression node to be checked.
             * @returns {boolean} True if the constructor is called with parens.
             * @private
             */
            function isNewExpressionWithParens(newExpression) {
                const lastToken = sourceCode.getLastToken(newExpression);
                const penultimateToken = sourceCode.getTokenBefore(lastToken);
                return newExpression.arguments.length > 0 ||
                    (
                    // The expression should end with its own parens, e.g., new new foo() is not a new expression with parens
                    astUtils.isOpeningParenToken(penultimateToken) &&
                        astUtils.isClosingParenToken(lastToken) &&
                        newExpression.callee.range[1] < newExpression.range[1]);
            }
            /**
             * Determines if a node is or contains an assignment expression
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the node is or contains an assignment expression.
             * @private
             */
            function containsAssignment(node) {
                if (node.type === "AssignmentExpression") {
                    return true;
                }
                if (node.type === "ConditionalExpression" &&
                    (node.consequent.type === "AssignmentExpression" || node.alternate.type === "AssignmentExpression")) {
                    return true;
                }
                if ((node.left && node.left.type === "AssignmentExpression") ||
                    (node.right && node.right.type === "AssignmentExpression")) {
                    return true;
                }
                return false;
            }
            /**
             * Determines if a node is contained by or is itself a return statement and is allowed to have a parenthesised assignment
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the assignment can be parenthesised.
             * @private
             */
            function isReturnAssignException(node) {
                if (!EXCEPT_RETURN_ASSIGN || !isInReturnStatement(node)) {
                    return false;
                }
                if (node.type === "ReturnStatement") {
                    return node.argument && containsAssignment(node.argument);
                }
                if (node.type === "ArrowFunctionExpression" && node.body.type !== "BlockStatement") {
                    return containsAssignment(node.body);
                }
                return containsAssignment(node);
            }
            /**
             * Determines if a node following a [no LineTerminator here] restriction is
             * surrounded by (potentially) invalid extra parentheses.
             * @param {Token} token The token preceding the [no LineTerminator here] restriction.
             * @param {ASTNode} node The node to be checked.
             * @returns {boolean} True if the node is incorrectly parenthesised.
             * @private
             */
            function hasExcessParensNoLineTerminator(token, node) {
                if (token.loc.end.line === node.loc.start.line) {
                    return hasExcessParens(node);
                }
                return hasDoubleExcessParens(node);
            }
            /**
             * Determines whether a node should be preceded by an additional space when removing parens
             * @param {ASTNode} node node to evaluate; must be surrounded by parentheses
             * @returns {boolean} `true` if a space should be inserted before the node
             * @private
             */
            function requiresLeadingSpace(node) {
                const leftParenToken = sourceCode.getTokenBefore(node);
                const tokenBeforeLeftParen = sourceCode.getTokenBefore(leftParenToken, { includeComments: true });
                const tokenAfterLeftParen = sourceCode.getTokenAfter(leftParenToken, { includeComments: true });
                return tokenBeforeLeftParen &&
                    tokenBeforeLeftParen.range[1] === leftParenToken.range[0] &&
                    leftParenToken.range[1] === tokenAfterLeftParen.range[0] &&
                    !astUtils.canTokensBeAdjacent(tokenBeforeLeftParen, tokenAfterLeftParen);
            }
            /**
             * Determines whether a node should be followed by an additional space when removing parens
             * @param {ASTNode} node node to evaluate; must be surrounded by parentheses
             * @returns {boolean} `true` if a space should be inserted after the node
             * @private
             */
            function requiresTrailingSpace(node) {
                const nextTwoTokens = sourceCode.getTokensAfter(node, { count: 2 });
                const rightParenToken = nextTwoTokens[0];
                const tokenAfterRightParen = nextTwoTokens[1];
                const tokenBeforeRightParen = sourceCode.getLastToken(node);
                return rightParenToken && tokenAfterRightParen &&
                    !sourceCode.isSpaceBetweenTokens(rightParenToken, tokenAfterRightParen) &&
                    !astUtils.canTokensBeAdjacent(tokenBeforeRightParen, tokenAfterRightParen);
            }
            /**
             * Determines if a given expression node is an IIFE
             * @param {ASTNode} node The node to check
             * @returns {boolean} `true` if the given node is an IIFE
             */
            function isIIFE(node) {
                const maybeCallNode = astUtils.skipChainExpression(node);
                return maybeCallNode.type === "CallExpression" && maybeCallNode.callee.type === "FunctionExpression";
            }
            /**
             * Determines if the given node can be the assignment target in destructuring or the LHS of an assignment.
             * This is to avoid an autofix that could change behavior because parsers mistakenly allow invalid syntax,
             * such as `(a = b) = c` and `[(a = b) = c] = []`. Ideally, this function shouldn't be necessary.
             * @param {ASTNode} [node] The node to check
             * @returns {boolean} `true` if the given node can be a valid assignment target
             */
            function canBeAssignmentTarget(node) {
                return node && (node.type === "Identifier" || node.type === "MemberExpression");
            }
            /**
             * Report the node
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function report(node) {
                const leftParenToken = sourceCode.getTokenBefore(node);
                const rightParenToken = sourceCode.getTokenAfter(node);
                if (!isParenthesisedTwice(node)) {
                    if (tokensToIgnore.has(sourceCode.getFirstToken(node))) {
                        return;
                    }
                    if (isIIFE(node) && !isParenthesised(node.callee)) {
                        return;
                    }
                }
                /**
                 * Finishes reporting
                 * @returns {void}
                 * @private
                 */
                function finishReport() {
                    context.report({
                        node,
                        loc: leftParenToken.loc,
                        messageId: "unexpected",
                        fix(fixer) {
                            const parenthesizedSource = sourceCode.text.slice(leftParenToken.range[1], rightParenToken.range[0]);
                            return fixer.replaceTextRange([
                                leftParenToken.range[0],
                                rightParenToken.range[1]
                            ], (requiresLeadingSpace(node) ? " " : "") + parenthesizedSource + (requiresTrailingSpace(node) ? " " : ""));
                        }
                    });
                }
                if (reportsBuffer) {
                    reportsBuffer.reports.push({ node, finishReport });
                    return;
                }
                finishReport();
            }
            /**
             * Evaluate a argument of the node.
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkArgumentWithPrecedence(node) {
                if (hasExcessParensWithPrecedence(node.argument, precedence(node))) {
                    report(node.argument);
                }
            }
            /**
             * Check if a member expression contains a call expression
             * @param {ASTNode} node MemberExpression node to evaluate
             * @returns {boolean} true if found, false if not
             */
            function doesMemberExpressionContainCallExpression(node) {
                let currentNode = node.object;
                let currentNodeType = node.object.type;
                while (currentNodeType === "MemberExpression") {
                    currentNode = currentNode.object;
                    currentNodeType = currentNode.type;
                }
                return currentNodeType === "CallExpression";
            }
            /**
             * Evaluate a new call
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkCallNew(node) {
                const callee = node.callee;
                if (hasExcessParensWithPrecedence(callee, precedence(node))) {
                    if (hasDoubleExcessParens(callee) ||
                        !(isIIFE(node) ||
                            // (new A)(); new (new A)();
                            (callee.type === "NewExpression" &&
                                !isNewExpressionWithParens(callee) &&
                                !(node.type === "NewExpression" &&
                                    !isNewExpressionWithParens(node))) ||
                            // new (a().b)(); new (a.b().c);
                            (node.type === "NewExpression" &&
                                callee.type === "MemberExpression" &&
                                doesMemberExpressionContainCallExpression(callee)) ||
                            // (a?.b)(); (a?.())();
                            (!node.optional &&
                                callee.type === "ChainExpression"))) {
                        report(node.callee);
                    }
                }
                node.arguments
                    .filter(arg => hasExcessParensWithPrecedence(arg, PRECEDENCE_OF_ASSIGNMENT_EXPR))
                    .forEach(report);
            }
            /**
             * Evaluate binary logicals
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkBinaryLogical(node) {
                const prec = precedence(node);
                const leftPrecedence = precedence(node.left);
                const rightPrecedence = precedence(node.right);
                const isExponentiation = node.operator === "**";
                const shouldSkipLeft = NESTED_BINARY && (node.left.type === "BinaryExpression" || node.left.type === "LogicalExpression");
                const shouldSkipRight = NESTED_BINARY && (node.right.type === "BinaryExpression" || node.right.type === "LogicalExpression");
                if (!shouldSkipLeft && hasExcessParens(node.left)) {
                    if (!(["AwaitExpression", "UnaryExpression"].includes(node.left.type) && isExponentiation) &&
                        !astUtils.isMixedLogicalAndCoalesceExpressions(node.left, node) &&
                        (leftPrecedence > prec || (leftPrecedence === prec && !isExponentiation)) ||
                        isParenthesisedTwice(node.left)) {
                        report(node.left);
                    }
                }
                if (!shouldSkipRight && hasExcessParens(node.right)) {
                    if (!astUtils.isMixedLogicalAndCoalesceExpressions(node.right, node) &&
                        (rightPrecedence > prec || (rightPrecedence === prec && isExponentiation)) ||
                        isParenthesisedTwice(node.right)) {
                        report(node.right);
                    }
                }
            }
            /**
             * Check the parentheses around the super class of the given class definition.
             * @param {ASTNode} node The node of class declarations to check.
             * @returns {void}
             */
            function checkClass(node) {
                if (!node.superClass) {
                    return;
                }
                /*
                 * If `node.superClass` is a LeftHandSideExpression, parentheses are extra.
                 * Otherwise, parentheses are needed.
                 */
                const hasExtraParens = precedence(node.superClass) > PRECEDENCE_OF_UPDATE_EXPR
                    ? hasExcessParens(node.superClass)
                    : hasDoubleExcessParens(node.superClass);
                if (hasExtraParens) {
                    report(node.superClass);
                }
            }
            /**
             * Check the parentheses around the argument of the given spread operator.
             * @param {ASTNode} node The node of spread elements/properties to check.
             * @returns {void}
             */
            function checkSpreadOperator(node) {
                if (hasExcessParensWithPrecedence(node.argument, PRECEDENCE_OF_ASSIGNMENT_EXPR)) {
                    report(node.argument);
                }
            }
            /**
             * Checks the parentheses for an ExpressionStatement or ExportDefaultDeclaration
             * @param {ASTNode} node The ExpressionStatement.expression or ExportDefaultDeclaration.declaration node
             * @returns {void}
             */
            function checkExpressionOrExportStatement(node) {
                const firstToken = isParenthesised(node) ? sourceCode.getTokenBefore(node) : sourceCode.getFirstToken(node);
                const secondToken = sourceCode.getTokenAfter(firstToken, astUtils.isNotOpeningParenToken);
                const thirdToken = secondToken ? sourceCode.getTokenAfter(secondToken) : null;
                const tokenAfterClosingParens = secondToken ? sourceCode.getTokenAfter(secondToken, astUtils.isNotClosingParenToken) : null;
                if (astUtils.isOpeningParenToken(firstToken) &&
                    (astUtils.isOpeningBraceToken(secondToken) ||
                        secondToken.type === "Keyword" && (secondToken.value === "function" ||
                            secondToken.value === "class" ||
                            secondToken.value === "let" &&
                                tokenAfterClosingParens &&
                                (astUtils.isOpeningBracketToken(tokenAfterClosingParens) ||
                                    tokenAfterClosingParens.type === "Identifier")) ||
                        secondToken && secondToken.type === "Identifier" && secondToken.value === "async" && thirdToken && thirdToken.type === "Keyword" && thirdToken.value === "function")) {
                    tokensToIgnore.add(secondToken);
                }
                const hasExtraParens = node.parent.type === "ExportDefaultDeclaration"
                    ? hasExcessParensWithPrecedence(node, PRECEDENCE_OF_ASSIGNMENT_EXPR)
                    : hasExcessParens(node);
                if (hasExtraParens) {
                    report(node);
                }
            }
            /**
             * Finds the path from the given node to the specified ancestor.
             * @param {ASTNode} node First node in the path.
             * @param {ASTNode} ancestor Last node in the path.
             * @returns {ASTNode[]} Path, including both nodes.
             * @throws {Error} If the given node does not have the specified ancestor.
             */
            function pathToAncestor(node, ancestor) {
                const path = [node];
                let currentNode = node;
                while (currentNode !== ancestor) {
                    currentNode = currentNode.parent;
                    /* istanbul ignore if */
                    if (currentNode === null) {
                        throw new Error("Nodes are not in the ancestor-descendant relationship.");
                    }
                    path.push(currentNode);
                }
                return path;
            }
            /**
             * Finds the path from the given node to the specified descendant.
             * @param {ASTNode} node First node in the path.
             * @param {ASTNode} descendant Last node in the path.
             * @returns {ASTNode[]} Path, including both nodes.
             * @throws {Error} If the given node does not have the specified descendant.
             */
            function pathToDescendant(node, descendant) {
                return pathToAncestor(descendant, node).reverse();
            }
            /**
             * Checks whether the syntax of the given ancestor of an 'in' expression inside a for-loop initializer
             * is preventing the 'in' keyword from being interpreted as a part of an ill-formed for-in loop.
             * @param {ASTNode} node Ancestor of an 'in' expression.
             * @param {ASTNode} child Child of the node, ancestor of the same 'in' expression or the 'in' expression itself.
             * @returns {boolean} True if the keyword 'in' would be interpreted as the 'in' operator, without any parenthesis.
             */
            function isSafelyEnclosingInExpression(node, child) {
                switch (node.type) {
                    case "ArrayExpression":
                    case "ArrayPattern":
                    case "BlockStatement":
                    case "ObjectExpression":
                    case "ObjectPattern":
                    case "TemplateLiteral":
                        return true;
                    case "ArrowFunctionExpression":
                    case "FunctionExpression":
                        return node.params.includes(child);
                    case "CallExpression":
                    case "NewExpression":
                        return node.arguments.includes(child);
                    case "MemberExpression":
                        return node.computed && node.property === child;
                    case "ConditionalExpression":
                        return node.consequent === child;
                    default:
                        return false;
                }
            }
            /**
             * Starts a new reports buffering. Warnings will be stored in a buffer instead of being reported immediately.
             * An additional logic that requires multiple nodes (e.g. a whole subtree) may dismiss some of the stored warnings.
             * @returns {void}
             */
            function startNewReportsBuffering() {
                reportsBuffer = {
                    upper: reportsBuffer,
                    inExpressionNodes: [],
                    reports: []
                };
            }
            /**
             * Ends the current reports buffering.
             * @returns {void}
             */
            function endCurrentReportsBuffering() {
                const { upper, inExpressionNodes, reports } = reportsBuffer;
                if (upper) {
                    upper.inExpressionNodes.push(...inExpressionNodes);
                    upper.reports.push(...reports);
                }
                else {
                    // flush remaining reports
                    reports.forEach(({ finishReport }) => finishReport());
                }
                reportsBuffer = upper;
            }
            /**
             * Checks whether the given node is in the current reports buffer.
             * @param {ASTNode} node Node to check.
             * @returns {boolean} True if the node is in the current buffer, false otherwise.
             */
            function isInCurrentReportsBuffer(node) {
                return reportsBuffer.reports.some(r => r.node === node);
            }
            /**
             * Removes the given node from the current reports buffer.
             * @param {ASTNode} node Node to remove.
             * @returns {void}
             */
            function removeFromCurrentReportsBuffer(node) {
                reportsBuffer.reports = reportsBuffer.reports.filter(r => r.node !== node);
            }
            /**
             * Checks whether a node is a MemberExpression at NewExpression's callee.
             * @param {ASTNode} node node to check.
             * @returns {boolean} True if the node is a MemberExpression at NewExpression's callee. false otherwise.
             */
            function isMemberExpInNewCallee(node) {
                if (node.type === "MemberExpression") {
                    return node.parent.type === "NewExpression" && node.parent.callee === node
                        ? true
                        : node.parent.object === node && isMemberExpInNewCallee(node.parent);
                }
                return false;
            }
            return {
                ArrayExpression(node) {
                    node.elements
                        .filter(e => e && hasExcessParensWithPrecedence(e, PRECEDENCE_OF_ASSIGNMENT_EXPR))
                        .forEach(report);
                },
                ArrayPattern(node) {
                    node.elements
                        .filter(e => canBeAssignmentTarget(e) && hasExcessParens(e))
                        .forEach(report);
                },
                ArrowFunctionExpression(node) {
                    if (isReturnAssignException(node)) {
                        return;
                    }
                    if (node.body.type === "ConditionalExpression" &&
                        IGNORE_ARROW_CONDITIONALS) {
                        return;
                    }
                    if (node.body.type !== "BlockStatement") {
                        const firstBodyToken = sourceCode.getFirstToken(node.body, astUtils.isNotOpeningParenToken);
                        const tokenBeforeFirst = sourceCode.getTokenBefore(firstBodyToken);
                        if (astUtils.isOpeningParenToken(tokenBeforeFirst) && astUtils.isOpeningBraceToken(firstBodyToken)) {
                            tokensToIgnore.add(firstBodyToken);
                        }
                        if (hasExcessParensWithPrecedence(node.body, PRECEDENCE_OF_ASSIGNMENT_EXPR)) {
                            report(node.body);
                        }
                    }
                },
                AssignmentExpression(node) {
                    if (canBeAssignmentTarget(node.left) && hasExcessParens(node.left)) {
                        report(node.left);
                    }
                    if (!isReturnAssignException(node) && hasExcessParensWithPrecedence(node.right, precedence(node))) {
                        report(node.right);
                    }
                },
                BinaryExpression(node) {
                    if (reportsBuffer && node.operator === "in") {
                        reportsBuffer.inExpressionNodes.push(node);
                    }
                    checkBinaryLogical(node);
                },
                CallExpression: checkCallNew,
                ClassBody(node) {
                    node.body
                        .filter(member => member.type === "MethodDefinition" && member.computed && member.key)
                        .filter(member => hasExcessParensWithPrecedence(member.key, PRECEDENCE_OF_ASSIGNMENT_EXPR))
                        .forEach(member => report(member.key));
                },
                ConditionalExpression(node) {
                    if (isReturnAssignException(node)) {
                        return;
                    }
                    if (!isCondAssignException(node) &&
                        hasExcessParensWithPrecedence(node.test, precedence({ type: "LogicalExpression", operator: "||" }))) {
                        report(node.test);
                    }
                    if (hasExcessParensWithPrecedence(node.consequent, PRECEDENCE_OF_ASSIGNMENT_EXPR)) {
                        report(node.consequent);
                    }
                    if (hasExcessParensWithPrecedence(node.alternate, PRECEDENCE_OF_ASSIGNMENT_EXPR)) {
                        report(node.alternate);
                    }
                },
                DoWhileStatement(node) {
                    if (hasExcessParens(node.test) && !isCondAssignException(node)) {
                        report(node.test);
                    }
                },
                ExportDefaultDeclaration: node => checkExpressionOrExportStatement(node.declaration),
                ExpressionStatement: node => checkExpressionOrExportStatement(node.expression),
                ForInStatement(node) {
                    if (node.left.type !== "VariableDeclaration") {
                        const firstLeftToken = sourceCode.getFirstToken(node.left, astUtils.isNotOpeningParenToken);
                        if (firstLeftToken.value === "let" &&
                            astUtils.isOpeningBracketToken(sourceCode.getTokenAfter(firstLeftToken, astUtils.isNotClosingParenToken))) {
                            // ForInStatement#left expression cannot start with `let[`.
                            tokensToIgnore.add(firstLeftToken);
                        }
                    }
                    if (hasExcessParens(node.left)) {
                        report(node.left);
                    }
                    if (hasExcessParens(node.right)) {
                        report(node.right);
                    }
                },
                ForOfStatement(node) {
                    if (node.left.type !== "VariableDeclaration") {
                        const firstLeftToken = sourceCode.getFirstToken(node.left, astUtils.isNotOpeningParenToken);
                        if (firstLeftToken.value === "let") {
                            // ForOfStatement#left expression cannot start with `let`.
                            tokensToIgnore.add(firstLeftToken);
                        }
                    }
                    if (hasExcessParens(node.left)) {
                        report(node.left);
                    }
                    if (hasExcessParensWithPrecedence(node.right, PRECEDENCE_OF_ASSIGNMENT_EXPR)) {
                        report(node.right);
                    }
                },
                ForStatement(node) {
                    if (node.test && hasExcessParens(node.test) && !isCondAssignException(node)) {
                        report(node.test);
                    }
                    if (node.update && hasExcessParens(node.update)) {
                        report(node.update);
                    }
                    if (node.init) {
                        if (node.init.type !== "VariableDeclaration") {
                            const firstToken = sourceCode.getFirstToken(node.init, astUtils.isNotOpeningParenToken);
                            if (firstToken.value === "let" &&
                                astUtils.isOpeningBracketToken(sourceCode.getTokenAfter(firstToken, astUtils.isNotClosingParenToken))) {
                                // ForStatement#init expression cannot start with `let[`.
                                tokensToIgnore.add(firstToken);
                            }
                        }
                        startNewReportsBuffering();
                        if (hasExcessParens(node.init)) {
                            report(node.init);
                        }
                    }
                },
                "ForStatement > *.init:exit"(node) {
                    /*
                     * Removing parentheses around `in` expressions might change semantics and cause errors.
                     *
                     * For example, this valid for loop:
                     *      for (let a = (b in c); ;);
                     * after removing parentheses would be treated as an invalid for-in loop:
                     *      for (let a = b in c; ;);
                     */
                    if (reportsBuffer.reports.length) {
                        reportsBuffer.inExpressionNodes.forEach(inExpressionNode => {
                            const path = pathToDescendant(node, inExpressionNode);
                            let nodeToExclude;
                            for (let i = 0; i < path.length; i++) {
                                const pathNode = path[i];
                                if (i < path.length - 1) {
                                    const nextPathNode = path[i + 1];
                                    if (isSafelyEnclosingInExpression(pathNode, nextPathNode)) {
                                        // The 'in' expression in safely enclosed by the syntax of its ancestor nodes (e.g. by '{}' or '[]').
                                        return;
                                    }
                                }
                                if (isParenthesised(pathNode)) {
                                    if (isInCurrentReportsBuffer(pathNode)) {
                                        // This node was supposed to be reported, but parentheses might be necessary.
                                        if (isParenthesisedTwice(pathNode)) {
                                            /*
                                             * This node is parenthesised twice, it certainly has at least one pair of `extra` parentheses.
                                             * If the --fix option is on, the current fixing iteration will remove only one pair of parentheses.
                                             * The remaining pair is safely enclosing the 'in' expression.
                                             */
                                            return;
                                        }
                                        // Exclude the outermost node only.
                                        if (!nodeToExclude) {
                                            nodeToExclude = pathNode;
                                        }
                                        // Don't break the loop here, there might be some safe nodes or parentheses that will stay inside.
                                    }
                                    else {
                                        // This node will stay parenthesised, the 'in' expression in safely enclosed by '()'.
                                        return;
                                    }
                                }
                            }
                            // Exclude the node from the list (i.e. treat parentheses as necessary)
                            removeFromCurrentReportsBuffer(nodeToExclude);
                        });
                    }
                    endCurrentReportsBuffering();
                },
                IfStatement(node) {
                    if (hasExcessParens(node.test) && !isCondAssignException(node)) {
                        report(node.test);
                    }
                },
                ImportExpression(node) {
                    const { source } = node;
                    if (source.type === "SequenceExpression") {
                        if (hasDoubleExcessParens(source)) {
                            report(source);
                        }
                    }
                    else if (hasExcessParens(source)) {
                        report(source);
                    }
                },
                LogicalExpression: checkBinaryLogical,
                MemberExpression(node) {
                    const shouldAllowWrapOnce = isMemberExpInNewCallee(node) &&
                        doesMemberExpressionContainCallExpression(node);
                    const nodeObjHasExcessParens = shouldAllowWrapOnce
                        ? hasDoubleExcessParens(node.object)
                        : hasExcessParens(node.object) &&
                            !(isImmediateFunctionPrototypeMethodCall(node.parent) &&
                                node.parent.callee === node &&
                                IGNORE_FUNCTION_PROTOTYPE_METHODS);
                    if (nodeObjHasExcessParens &&
                        precedence(node.object) >= precedence(node) &&
                        (node.computed ||
                            !(astUtils.isDecimalInteger(node.object) ||
                                // RegExp literal is allowed to have parens (#1589)
                                (node.object.type === "Literal" && node.object.regex)))) {
                        report(node.object);
                    }
                    if (nodeObjHasExcessParens &&
                        node.object.type === "CallExpression") {
                        report(node.object);
                    }
                    if (nodeObjHasExcessParens &&
                        !IGNORE_NEW_IN_MEMBER_EXPR &&
                        node.object.type === "NewExpression" &&
                        isNewExpressionWithParens(node.object)) {
                        report(node.object);
                    }
                    if (nodeObjHasExcessParens &&
                        node.optional &&
                        node.object.type === "ChainExpression") {
                        report(node.object);
                    }
                    if (node.computed && hasExcessParens(node.property)) {
                        report(node.property);
                    }
                },
                NewExpression: checkCallNew,
                ObjectExpression(node) {
                    node.properties
                        .filter(property => property.value && hasExcessParensWithPrecedence(property.value, PRECEDENCE_OF_ASSIGNMENT_EXPR))
                        .forEach(property => report(property.value));
                },
                ObjectPattern(node) {
                    node.properties
                        .filter(property => {
                        const value = property.value;
                        return canBeAssignmentTarget(value) && hasExcessParens(value);
                    }).forEach(property => report(property.value));
                },
                Property(node) {
                    if (node.computed) {
                        const { key } = node;
                        if (key && hasExcessParensWithPrecedence(key, PRECEDENCE_OF_ASSIGNMENT_EXPR)) {
                            report(key);
                        }
                    }
                },
                RestElement(node) {
                    const argument = node.argument;
                    if (canBeAssignmentTarget(argument) && hasExcessParens(argument)) {
                        report(argument);
                    }
                },
                ReturnStatement(node) {
                    const returnToken = sourceCode.getFirstToken(node);
                    if (isReturnAssignException(node)) {
                        return;
                    }
                    if (node.argument &&
                        hasExcessParensNoLineTerminator(returnToken, node.argument) &&
                        // RegExp literal is allowed to have parens (#1589)
                        !(node.argument.type === "Literal" && node.argument.regex)) {
                        report(node.argument);
                    }
                },
                SequenceExpression(node) {
                    const precedenceOfNode = precedence(node);
                    node.expressions
                        .filter(e => hasExcessParensWithPrecedence(e, precedenceOfNode))
                        .forEach(report);
                },
                SwitchCase(node) {
                    if (node.test && hasExcessParens(node.test)) {
                        report(node.test);
                    }
                },
                SwitchStatement(node) {
                    if (hasExcessParens(node.discriminant)) {
                        report(node.discriminant);
                    }
                },
                ThrowStatement(node) {
                    const throwToken = sourceCode.getFirstToken(node);
                    if (hasExcessParensNoLineTerminator(throwToken, node.argument)) {
                        report(node.argument);
                    }
                },
                UnaryExpression: checkArgumentWithPrecedence,
                UpdateExpression(node) {
                    if (node.prefix) {
                        checkArgumentWithPrecedence(node);
                    }
                    else {
                        const { argument } = node;
                        const operatorToken = sourceCode.getLastToken(node);
                        if (argument.loc.end.line === operatorToken.loc.start.line) {
                            checkArgumentWithPrecedence(node);
                        }
                        else {
                            if (hasDoubleExcessParens(argument)) {
                                report(argument);
                            }
                        }
                    }
                },
                AwaitExpression: checkArgumentWithPrecedence,
                VariableDeclarator(node) {
                    if (node.init && hasExcessParensWithPrecedence(node.init, PRECEDENCE_OF_ASSIGNMENT_EXPR) &&
                        // RegExp literal is allowed to have parens (#1589)
                        !(node.init.type === "Literal" && node.init.regex)) {
                        report(node.init);
                    }
                },
                WhileStatement(node) {
                    if (hasExcessParens(node.test) && !isCondAssignException(node)) {
                        report(node.test);
                    }
                },
                WithStatement(node) {
                    if (hasExcessParens(node.object)) {
                        report(node.object);
                    }
                },
                YieldExpression(node) {
                    if (node.argument) {
                        const yieldToken = sourceCode.getFirstToken(node);
                        if ((precedence(node.argument) >= precedence(node) &&
                            hasExcessParensNoLineTerminator(yieldToken, node.argument)) ||
                            hasDoubleExcessParens(node.argument)) {
                            report(node.argument);
                        }
                    }
                },
                ClassDeclaration: checkClass,
                ClassExpression: checkClass,
                SpreadElement: checkSpreadOperator,
                SpreadProperty: checkSpreadOperator,
                ExperimentalSpreadProperty: checkSpreadOperator,
                TemplateLiteral(node) {
                    node.expressions
                        .filter(e => e && hasExcessParens(e))
                        .forEach(report);
                },
                AssignmentPattern(node) {
                    const { left, right } = node;
                    if (canBeAssignmentTarget(left) && hasExcessParens(left)) {
                        report(left);
                    }
                    if (right && hasExcessParensWithPrecedence(right, PRECEDENCE_OF_ASSIGNMENT_EXPR)) {
                        report(right);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-extra-semi.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const FixTracker = require(508) /* ./utils/fix-tracker */;
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary semicolons",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-extra-semi"
            },
            fixable: "code",
            schema: [],
            messages: {
                unexpected: "Unnecessary semicolon."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Reports an unnecessary semicolon error.
             * @param {Node|Token} nodeOrToken A node or a token to be reported.
             * @returns {void}
             */
            function report(nodeOrToken) {
                context.report({
                    node: nodeOrToken,
                    messageId: "unexpected",
                    fix(fixer) {
                        /*
                         * Expand the replacement range to include the surrounding
                         * tokens to avoid conflicting with semi.
                         * https://github.com/eslint/eslint/issues/7928
                         */
                        return new FixTracker(fixer, context.getSourceCode())
                            .retainSurroundingTokens(nodeOrToken)
                            .remove(nodeOrToken);
                    }
                });
            }
            /**
             * Checks for a part of a class body.
             * This checks tokens from a specified token to a next MethodDefinition or the end of class body.
             * @param {Token} firstToken The first token to check.
             * @returns {void}
             */
            function checkForPartOfClassBody(firstToken) {
                for (let token = firstToken; token.type === "Punctuator" && !astUtils.isClosingBraceToken(token); token = sourceCode.getTokenAfter(token)) {
                    if (astUtils.isSemicolonToken(token)) {
                        report(token);
                    }
                }
            }
            return {
                /**
                 * Reports this empty statement, except if the parent node is a loop.
                 * @param {Node} node A EmptyStatement node to be reported.
                 * @returns {void}
                 */
                EmptyStatement(node) {
                    const parent = node.parent, allowedParentTypes = [
                        "ForStatement",
                        "ForInStatement",
                        "ForOfStatement",
                        "WhileStatement",
                        "DoWhileStatement",
                        "IfStatement",
                        "LabeledStatement",
                        "WithStatement"
                    ];
                    if (allowedParentTypes.indexOf(parent.type) === -1) {
                        report(node);
                    }
                },
                /**
                 * Checks tokens from the head of this class body to the first MethodDefinition or the end of this class body.
                 * @param {Node} node A ClassBody node to check.
                 * @returns {void}
                 */
                ClassBody(node) {
                    checkForPartOfClassBody(sourceCode.getFirstToken(node, 1)); // 0 is `{`.
                },
                /**
                 * Checks tokens from this MethodDefinition to the next MethodDefinition or the end of this class body.
                 * @param {Node} node A MethodDefinition node of the start point.
                 * @returns {void}
                 */
                MethodDefinition(node) {
                    checkForPartOfClassBody(sourceCode.getTokenAfter(node));
                }
            };
        }
    };
},
/* eslint/lib/rules/no-fallthrough.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const DEFAULT_FALLTHROUGH_COMMENT = /falls?\s?through/iu;
    /**
     * Checks whether or not a given case has a fallthrough comment.
     * @param {ASTNode} caseWhichFallsThrough SwitchCase node which falls through.
     * @param {ASTNode} subsequentCase The case after caseWhichFallsThrough.
     * @param {RuleContext} context A rule context which stores comments.
     * @param {RegExp} fallthroughCommentPattern A pattern to match comment to.
     * @returns {boolean} `true` if the case has a valid fallthrough comment.
     */
    function hasFallthroughComment(caseWhichFallsThrough, subsequentCase, context, fallthroughCommentPattern) {
        const sourceCode = context.getSourceCode();
        if (caseWhichFallsThrough.consequent.length === 1 && caseWhichFallsThrough.consequent[0].type === "BlockStatement") {
            const trailingCloseBrace = sourceCode.getLastToken(caseWhichFallsThrough.consequent[0]);
            const commentInBlock = sourceCode.getCommentsBefore(trailingCloseBrace).pop();
            if (commentInBlock && fallthroughCommentPattern.test(commentInBlock.value)) {
                return true;
            }
        }
        const comment = sourceCode.getCommentsBefore(subsequentCase).pop();
        return Boolean(comment && fallthroughCommentPattern.test(comment.value));
    }
    /**
     * Checks whether or not a given code path segment is reachable.
     * @param {CodePathSegment} segment A CodePathSegment to check.
     * @returns {boolean} `true` if the segment is reachable.
     */
    function isReachable(segment) {
        return segment.reachable;
    }
    /**
     * Checks whether a node and a token are separated by blank lines
     * @param {ASTNode} node The node to check
     * @param {Token} token The token to compare against
     * @returns {boolean} `true` if there are blank lines between node and token
     */
    function hasBlankLinesBetween(node, token) {
        return token.loc.start.line > node.loc.end.line + 1;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow fallthrough of `case` statements",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-fallthrough"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        commentPattern: {
                            type: "string",
                            default: ""
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                case: "Expected a 'break' statement before 'case'.",
                default: "Expected a 'break' statement before 'default'."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            let currentCodePath = null;
            const sourceCode = context.getSourceCode();
            /*
             * We need to use leading comments of the next SwitchCase node because
             * trailing comments is wrong if semicolons are omitted.
             */
            let fallthroughCase = null;
            let fallthroughCommentPattern = null;
            if (options.commentPattern) {
                fallthroughCommentPattern = new RegExp(options.commentPattern, "u");
            }
            else {
                fallthroughCommentPattern = DEFAULT_FALLTHROUGH_COMMENT;
            }
            return {
                onCodePathStart(codePath) {
                    currentCodePath = codePath;
                },
                onCodePathEnd() {
                    currentCodePath = currentCodePath.upper;
                },
                SwitchCase(node) {
                    /*
                     * Checks whether or not there is a fallthrough comment.
                     * And reports the previous fallthrough node if that does not exist.
                     */
                    if (fallthroughCase && !hasFallthroughComment(fallthroughCase, node, context, fallthroughCommentPattern)) {
                        context.report({
                            messageId: node.test ? "case" : "default",
                            node
                        });
                    }
                    fallthroughCase = null;
                },
                "SwitchCase:exit"(node) {
                    const nextToken = sourceCode.getTokenAfter(node);
                    /*
                     * `reachable` meant fall through because statements preceded by
                     * `break`, `return`, or `throw` are unreachable.
                     * And allows empty cases and the last case.
                     */
                    if (currentCodePath.currentSegments.some(isReachable) &&
                        (node.consequent.length > 0 || hasBlankLinesBetween(node, nextToken)) &&
                        node.parent.cases[node.parent.cases.length - 1] !== node) {
                        fallthroughCase = node;
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-floating-decimal.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow leading or trailing decimal points in numeric literals",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-floating-decimal"
            },
            schema: [],
            fixable: "code",
            messages: {
                leading: "A leading decimal point can be confused with a dot.",
                trailing: "A trailing decimal point can be confused with a dot."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                Literal(node) {
                    if (typeof node.value === "number") {
                        if (node.raw.startsWith(".")) {
                            context.report({
                                node,
                                messageId: "leading",
                                fix(fixer) {
                                    const tokenBefore = sourceCode.getTokenBefore(node);
                                    const needsSpaceBefore = tokenBefore &&
                                        tokenBefore.range[1] === node.range[0] &&
                                        !astUtils.canTokensBeAdjacent(tokenBefore, `0${node.raw}`);
                                    return fixer.insertTextBefore(node, needsSpaceBefore ? " 0" : "0");
                                }
                            });
                        }
                        if (node.raw.indexOf(".") === node.raw.length - 1) {
                            context.report({
                                node,
                                messageId: "trailing",
                                fix: fixer => fixer.insertTextAfter(node, "0")
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-func-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow reassigning `function` declarations",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-func-assign"
            },
            schema: [],
            messages: {
                isAFunction: "'{{name}}' is a function."
            }
        },
        create(context) {
            /**
             * Reports a reference if is non initializer and writable.
             * @param {References} references Collection of reference to check.
             * @returns {void}
             */
            function checkReference(references) {
                astUtils.getModifyingReferences(references).forEach(reference => {
                    context.report({
                        node: reference.identifier,
                        messageId: "isAFunction",
                        data: {
                            name: reference.identifier.name
                        }
                    });
                });
            }
            /**
             * Finds and reports references that are non initializer and writable.
             * @param {Variable} variable A variable to check.
             * @returns {void}
             */
            function checkVariable(variable) {
                if (variable.defs[0].type === "FunctionName") {
                    checkReference(variable.references);
                }
            }
            /**
             * Checks parameters of a given function node.
             * @param {ASTNode} node A function node to check.
             * @returns {void}
             */
            function checkForFunction(node) {
                context.getDeclaredVariables(node).forEach(checkVariable);
            }
            return {
                FunctionDeclaration: checkForFunction,
                FunctionExpression: checkForFunction
            };
        }
    };
},
/* eslint/lib/rules/no-global-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow assignments to native objects or read-only global variables",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-global-assign"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        exceptions: {
                            type: "array",
                            items: { type: "string" },
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                globalShouldNotBeModified: "Read-only global '{{name}}' should not be modified."
            }
        },
        create(context) {
            const config = context.options[0];
            const exceptions = (config && config.exceptions) || [];
            /**
             * Reports write references.
             * @param {Reference} reference A reference to check.
             * @param {int} index The index of the reference in the references.
             * @param {Reference[]} references The array that the reference belongs to.
             * @returns {void}
             */
            function checkReference(reference, index, references) {
                const identifier = reference.identifier;
                if (reference.init === false &&
                    reference.isWrite() &&
                    /*
                     * Destructuring assignments can have multiple default value,
                     * so possibly there are multiple writeable references for the same identifier.
                     */
                    (index === 0 || references[index - 1].identifier !== identifier)) {
                    context.report({
                        node: identifier,
                        messageId: "globalShouldNotBeModified",
                        data: {
                            name: identifier.name
                        }
                    });
                }
            }
            /**
             * Reports write references if a given variable is read-only builtin.
             * @param {Variable} variable A variable to check.
             * @returns {void}
             */
            function checkVariable(variable) {
                if (variable.writeable === false && exceptions.indexOf(variable.name) === -1) {
                    variable.references.forEach(checkReference);
                }
            }
            return {
                Program() {
                    const globalScope = context.getScope();
                    globalScope.variables.forEach(checkVariable);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-implicit-coercion.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const INDEX_OF_PATTERN = /^(?:i|lastI)ndexOf$/u;
    const ALLOWABLE_OPERATORS = ["~", "!!", "+", "*"];
    /**
     * Parses and normalizes an option object.
     * @param {Object} options An option object to parse.
     * @returns {Object} The parsed and normalized option object.
     */
    function parseOptions(options) {
        return {
            boolean: "boolean" in options ? options.boolean : true,
            number: "number" in options ? options.number : true,
            string: "string" in options ? options.string : true,
            disallowTemplateShorthand: "disallowTemplateShorthand" in options ? options.disallowTemplateShorthand : false,
            allow: options.allow || []
        };
    }
    /**
     * Checks whether or not a node is a double logical nigating.
     * @param {ASTNode} node An UnaryExpression node to check.
     * @returns {boolean} Whether or not the node is a double logical nigating.
     */
    function isDoubleLogicalNegating(node) {
        return (node.operator === "!" &&
            node.argument.type === "UnaryExpression" &&
            node.argument.operator === "!");
    }
    /**
     * Checks whether or not a node is a binary negating of `.indexOf()` method calling.
     * @param {ASTNode} node An UnaryExpression node to check.
     * @returns {boolean} Whether or not the node is a binary negating of `.indexOf()` method calling.
     */
    function isBinaryNegatingOfIndexOf(node) {
        if (node.operator !== "~") {
            return false;
        }
        const callNode = astUtils.skipChainExpression(node.argument);
        return (callNode.type === "CallExpression" &&
            astUtils.isSpecificMemberAccess(callNode.callee, null, INDEX_OF_PATTERN));
    }
    /**
     * Checks whether or not a node is a multiplying by one.
     * @param {BinaryExpression} node A BinaryExpression node to check.
     * @returns {boolean} Whether or not the node is a multiplying by one.
     */
    function isMultiplyByOne(node) {
        return node.operator === "*" && (node.left.type === "Literal" && node.left.value === 1 ||
            node.right.type === "Literal" && node.right.value === 1);
    }
    /**
     * Checks whether the result of a node is numeric or not
     * @param {ASTNode} node The node to test
     * @returns {boolean} true if the node is a number literal or a `Number()`, `parseInt` or `parseFloat` call
     */
    function isNumeric(node) {
        return (node.type === "Literal" && typeof node.value === "number" ||
            node.type === "CallExpression" && (node.callee.name === "Number" ||
                node.callee.name === "parseInt" ||
                node.callee.name === "parseFloat"));
    }
    /**
     * Returns the first non-numeric operand in a BinaryExpression. Designed to be
     * used from bottom to up since it walks up the BinaryExpression trees using
     * node.parent to find the result.
     * @param {BinaryExpression} node The BinaryExpression node to be walked up on
     * @returns {ASTNode|null} The first non-numeric item in the BinaryExpression tree or null
     */
    function getNonNumericOperand(node) {
        const left = node.left, right = node.right;
        if (right.type !== "BinaryExpression" && !isNumeric(right)) {
            return right;
        }
        if (left.type !== "BinaryExpression" && !isNumeric(left)) {
            return left;
        }
        return null;
    }
    /**
     * Checks whether an expression evaluates to a string.
     * @param {ASTNode} node node that represents the expression to check.
     * @returns {boolean} Whether or not the expression evaluates to a string.
     */
    function isStringType(node) {
        return astUtils.isStringLiteral(node) ||
            (node.type === "CallExpression" &&
                node.callee.type === "Identifier" &&
                node.callee.name === "String");
    }
    /**
     * Checks whether a node is an empty string literal or not.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} Whether or not the passed in node is an
     * empty string literal or not.
     */
    function isEmptyString(node) {
        return astUtils.isStringLiteral(node) && (node.value === "" || (node.type === "TemplateLiteral" && node.quasis.length === 1 && node.quasis[0].value.cooked === ""));
    }
    /**
     * Checks whether or not a node is a concatenating with an empty string.
     * @param {ASTNode} node A BinaryExpression node to check.
     * @returns {boolean} Whether or not the node is a concatenating with an empty string.
     */
    function isConcatWithEmptyString(node) {
        return node.operator === "+" && ((isEmptyString(node.left) && !isStringType(node.right)) ||
            (isEmptyString(node.right) && !isStringType(node.left)));
    }
    /**
     * Checks whether or not a node is appended with an empty string.
     * @param {ASTNode} node An AssignmentExpression node to check.
     * @returns {boolean} Whether or not the node is appended with an empty string.
     */
    function isAppendEmptyString(node) {
        return node.operator === "+=" && isEmptyString(node.right);
    }
    /**
     * Returns the operand that is not an empty string from a flagged BinaryExpression.
     * @param {ASTNode} node The flagged BinaryExpression node to check.
     * @returns {ASTNode} The operand that is not an empty string from a flagged BinaryExpression.
     */
    function getNonEmptyOperand(node) {
        return isEmptyString(node.left) ? node.right : node.left;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow shorthand type conversions",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-implicit-coercion"
            },
            fixable: "code",
            schema: [{
                    type: "object",
                    properties: {
                        boolean: {
                            type: "boolean",
                            default: true
                        },
                        number: {
                            type: "boolean",
                            default: true
                        },
                        string: {
                            type: "boolean",
                            default: true
                        },
                        disallowTemplateShorthand: {
                            type: "boolean",
                            default: false
                        },
                        allow: {
                            type: "array",
                            items: {
                                enum: ALLOWABLE_OPERATORS
                            },
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                useRecommendation: "use `{{recommendation}}` instead."
            }
        },
        create(context) {
            const options = parseOptions(context.options[0] || {});
            const sourceCode = context.getSourceCode();
            /**
             * Reports an error and autofixes the node
             * @param {ASTNode} node An ast node to report the error on.
             * @param {string} recommendation The recommended code for the issue
             * @param {bool} shouldFix Whether this report should fix the node
             * @returns {void}
             */
            function report(node, recommendation, shouldFix) {
                context.report({
                    node,
                    messageId: "useRecommendation",
                    data: {
                        recommendation
                    },
                    fix(fixer) {
                        if (!shouldFix) {
                            return null;
                        }
                        const tokenBefore = sourceCode.getTokenBefore(node);
                        if (tokenBefore &&
                            tokenBefore.range[1] === node.range[0] &&
                            !astUtils.canTokensBeAdjacent(tokenBefore, recommendation)) {
                            return fixer.replaceText(node, ` ${recommendation}`);
                        }
                        return fixer.replaceText(node, recommendation);
                    }
                });
            }
            return {
                UnaryExpression(node) {
                    let operatorAllowed;
                    // !!foo
                    operatorAllowed = options.allow.indexOf("!!") >= 0;
                    if (!operatorAllowed && options.boolean && isDoubleLogicalNegating(node)) {
                        const recommendation = `Boolean(${sourceCode.getText(node.argument.argument)})`;
                        report(node, recommendation, true);
                    }
                    // ~foo.indexOf(bar)
                    operatorAllowed = options.allow.indexOf("~") >= 0;
                    if (!operatorAllowed && options.boolean && isBinaryNegatingOfIndexOf(node)) {
                        // `foo?.indexOf(bar) !== -1` will be true (== found) if the `foo` is nullish. So use `>= 0` in that case.
                        const comparison = node.argument.type === "ChainExpression" ? ">= 0" : "!== -1";
                        const recommendation = `${sourceCode.getText(node.argument)} ${comparison}`;
                        report(node, recommendation, false);
                    }
                    // +foo
                    operatorAllowed = options.allow.indexOf("+") >= 0;
                    if (!operatorAllowed && options.number && node.operator === "+" && !isNumeric(node.argument)) {
                        const recommendation = `Number(${sourceCode.getText(node.argument)})`;
                        report(node, recommendation, true);
                    }
                },
                // Use `:exit` to prevent double reporting
                "BinaryExpression:exit"(node) {
                    let operatorAllowed;
                    // 1 * foo
                    operatorAllowed = options.allow.indexOf("*") >= 0;
                    const nonNumericOperand = !operatorAllowed && options.number && isMultiplyByOne(node) && getNonNumericOperand(node);
                    if (nonNumericOperand) {
                        const recommendation = `Number(${sourceCode.getText(nonNumericOperand)})`;
                        report(node, recommendation, true);
                    }
                    // "" + foo
                    operatorAllowed = options.allow.indexOf("+") >= 0;
                    if (!operatorAllowed && options.string && isConcatWithEmptyString(node)) {
                        const recommendation = `String(${sourceCode.getText(getNonEmptyOperand(node))})`;
                        report(node, recommendation, true);
                    }
                },
                AssignmentExpression(node) {
                    // foo += ""
                    const operatorAllowed = options.allow.indexOf("+") >= 0;
                    if (!operatorAllowed && options.string && isAppendEmptyString(node)) {
                        const code = sourceCode.getText(getNonEmptyOperand(node));
                        const recommendation = `${code} = String(${code})`;
                        report(node, recommendation, true);
                    }
                },
                TemplateLiteral(node) {
                    if (!options.disallowTemplateShorthand) {
                        return;
                    }
                    // tag`${foo}`
                    if (node.parent.type === "TaggedTemplateExpression") {
                        return;
                    }
                    // `` or `${foo}${bar}`
                    if (node.expressions.length !== 1) {
                        return;
                    }
                    //  `prefix${foo}`
                    if (node.quasis[0].value.cooked !== "") {
                        return;
                    }
                    //  `${foo}postfix`
                    if (node.quasis[1].value.cooked !== "") {
                        return;
                    }
                    // if the expression is already a string, then this isn't a coercion
                    if (isStringType(node.expressions[0])) {
                        return;
                    }
                    const code = sourceCode.getText(node.expressions[0]);
                    const recommendation = `String(${code})`;
                    report(node, recommendation, true);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-implicit-globals.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow declarations in the global scope",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-implicit-globals"
            },
            schema: [{
                    type: "object",
                    properties: {
                        lexicalBindings: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                globalNonLexicalBinding: "Unexpected {{kind}} declaration in the global scope, wrap in an IIFE for a local variable, assign as global property for a global variable.",
                globalLexicalBinding: "Unexpected {{kind}} declaration in the global scope, wrap in a block or in an IIFE.",
                globalVariableLeak: "Global variable leak, declare the variable if it is intended to be local.",
                assignmentToReadonlyGlobal: "Unexpected assignment to read-only global variable.",
                redeclarationOfReadonlyGlobal: "Unexpected redeclaration of read-only global variable."
            }
        },
        create(context) {
            const checkLexicalBindings = context.options[0] && context.options[0].lexicalBindings === true;
            /**
             * Reports the node.
             * @param {ASTNode} node Node to report.
             * @param {string} messageId Id of the message to report.
             * @param {string|undefined} kind Declaration kind, can be 'var', 'const', 'let', function or class.
             * @returns {void}
             */
            function report(node, messageId, kind) {
                context.report({
                    node,
                    messageId,
                    data: {
                        kind
                    }
                });
            }
            return {
                Program() {
                    const scope = context.getScope();
                    scope.variables.forEach(variable => {
                        // Only ESLint global variables have the `writable` key.
                        const isReadonlyEslintGlobalVariable = variable.writeable === false;
                        const isWritableEslintGlobalVariable = variable.writeable === true;
                        if (isWritableEslintGlobalVariable) {
                            // Everything is allowed with writable ESLint global variables.
                            return;
                        }
                        variable.defs.forEach(def => {
                            const defNode = def.node;
                            if (def.type === "FunctionName" || (def.type === "Variable" && def.parent.kind === "var")) {
                                if (isReadonlyEslintGlobalVariable) {
                                    report(defNode, "redeclarationOfReadonlyGlobal");
                                }
                                else {
                                    report(defNode, "globalNonLexicalBinding", def.type === "FunctionName" ? "function" : `'${def.parent.kind}'`);
                                }
                            }
                            if (checkLexicalBindings) {
                                if (def.type === "ClassName" ||
                                    (def.type === "Variable" && (def.parent.kind === "let" || def.parent.kind === "const"))) {
                                    if (isReadonlyEslintGlobalVariable) {
                                        report(defNode, "redeclarationOfReadonlyGlobal");
                                    }
                                    else {
                                        report(defNode, "globalLexicalBinding", def.type === "ClassName" ? "class" : `'${def.parent.kind}'`);
                                    }
                                }
                            }
                        });
                    });
                    // Undeclared assigned variables.
                    scope.implicit.variables.forEach(variable => {
                        const scopeVariable = scope.set.get(variable.name);
                        let messageId;
                        if (scopeVariable) {
                            // ESLint global variable
                            if (scopeVariable.writeable) {
                                return;
                            }
                            messageId = "assignmentToReadonlyGlobal";
                        }
                        else {
                            // Reference to an unknown variable, possible global leak.
                            messageId = "globalVariableLeak";
                        }
                        // def.node is an AssignmentExpression, ForInStatement or ForOfStatement.
                        variable.defs.forEach(def => {
                            report(def.node, messageId);
                        });
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-implied-eval.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const { getStaticValue } = require(350) /* eslint-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the use of `eval()`-like methods",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-implied-eval"
            },
            schema: [],
            messages: {
                impliedEval: "Implied eval. Consider passing a function instead of a string."
            }
        },
        create(context) {
            const GLOBAL_CANDIDATES = Object.freeze(["global", "window", "globalThis"]);
            const EVAL_LIKE_FUNC_PATTERN = /^(?:set(?:Interval|Timeout)|execScript)$/u;
            /**
             * Checks whether a node is evaluated as a string or not.
             * @param {ASTNode} node A node to check.
             * @returns {boolean} True if the node is evaluated as a string.
             */
            function isEvaluatedString(node) {
                if ((node.type === "Literal" && typeof node.value === "string") ||
                    node.type === "TemplateLiteral") {
                    return true;
                }
                if (node.type === "BinaryExpression" && node.operator === "+") {
                    return isEvaluatedString(node.left) || isEvaluatedString(node.right);
                }
                return false;
            }
            /**
             * Reports if the `CallExpression` node has evaluated argument.
             * @param {ASTNode} node A CallExpression to check.
             * @returns {void}
             */
            function reportImpliedEvalCallExpression(node) {
                const [firstArgument] = node.arguments;
                if (firstArgument) {
                    const staticValue = getStaticValue(firstArgument, context.getScope());
                    const isStaticString = staticValue && typeof staticValue.value === "string";
                    const isString = isStaticString || isEvaluatedString(firstArgument);
                    if (isString) {
                        context.report({
                            node,
                            messageId: "impliedEval"
                        });
                    }
                }
            }
            /**
             * Reports calls of `implied eval` via the global references.
             * @param {Variable} globalVar A global variable to check.
             * @returns {void}
             */
            function reportImpliedEvalViaGlobal(globalVar) {
                const { references, name } = globalVar;
                references.forEach(ref => {
                    const identifier = ref.identifier;
                    let node = identifier.parent;
                    while (astUtils.isSpecificMemberAccess(node, null, name)) {
                        node = node.parent;
                    }
                    if (astUtils.isSpecificMemberAccess(node, null, EVAL_LIKE_FUNC_PATTERN)) {
                        const calleeNode = node.parent.type === "ChainExpression" ? node.parent : node;
                        const parent = calleeNode.parent;
                        if (parent.type === "CallExpression" && parent.callee === calleeNode) {
                            reportImpliedEvalCallExpression(parent);
                        }
                    }
                });
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                CallExpression(node) {
                    if (astUtils.isSpecificId(node.callee, EVAL_LIKE_FUNC_PATTERN)) {
                        reportImpliedEvalCallExpression(node);
                    }
                },
                "Program:exit"() {
                    const globalScope = context.getScope();
                    GLOBAL_CANDIDATES
                        .map(candidate => astUtils.getVariableByName(globalScope, candidate))
                        .filter(globalVar => !!globalVar && globalVar.defs.length === 0)
                        .forEach(reportImpliedEvalViaGlobal);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-import-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const { findVariable } = require(350) /* eslint-utils */;
    const astUtils = require(394) /* ./utils/ast-utils */;
    const WellKnownMutationFunctions = {
        Object: /^(?:assign|definePropert(?:y|ies)|freeze|setPrototypeOf)$/u,
        Reflect: /^(?:(?:define|delete)Property|set(?:PrototypeOf)?)$/u
    };
    /**
     * Check if a given node is LHS of an assignment node.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is LHS.
     */
    function isAssignmentLeft(node) {
        const { parent } = node;
        return ((parent.type === "AssignmentExpression" &&
            parent.left === node) ||
            // Destructuring assignments
            parent.type === "ArrayPattern" ||
            (parent.type === "Property" &&
                parent.value === node &&
                parent.parent.type === "ObjectPattern") ||
            parent.type === "RestElement" ||
            (parent.type === "AssignmentPattern" &&
                parent.left === node));
    }
    /**
     * Check if a given node is the operand of mutation unary operator.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is the operand of mutation unary operator.
     */
    function isOperandOfMutationUnaryOperator(node) {
        const argumentNode = node.parent.type === "ChainExpression"
            ? node.parent
            : node;
        const { parent } = argumentNode;
        return ((parent.type === "UpdateExpression" &&
            parent.argument === argumentNode) ||
            (parent.type === "UnaryExpression" &&
                parent.operator === "delete" &&
                parent.argument === argumentNode));
    }
    /**
     * Check if a given node is the iteration variable of `for-in`/`for-of` syntax.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is the iteration variable.
     */
    function isIterationVariable(node) {
        const { parent } = node;
        return ((parent.type === "ForInStatement" &&
            parent.left === node) ||
            (parent.type === "ForOfStatement" &&
                parent.left === node));
    }
    /**
     * Check if a given node is at the first argument of a well-known mutation function.
     * - `Object.assign`
     * - `Object.defineProperty`
     * - `Object.defineProperties`
     * - `Object.freeze`
     * - `Object.setPrototypeOf`
     * - `Reflect.defineProperty`
     * - `Reflect.deleteProperty`
     * - `Reflect.set`
     * - `Reflect.setPrototypeOf`
     * @param {ASTNode} node The node to check.
     * @param {Scope} scope A `escope.Scope` object to find variable (whichever).
     * @returns {boolean} `true` if the node is at the first argument of a well-known mutation function.
     */
    function isArgumentOfWellKnownMutationFunction(node, scope) {
        const { parent } = node;
        if (parent.type !== "CallExpression" || parent.arguments[0] !== node) {
            return false;
        }
        const callee = astUtils.skipChainExpression(parent.callee);
        if (!astUtils.isSpecificMemberAccess(callee, "Object", WellKnownMutationFunctions.Object) &&
            !astUtils.isSpecificMemberAccess(callee, "Reflect", WellKnownMutationFunctions.Reflect)) {
            return false;
        }
        const variable = findVariable(scope, callee.object);
        return variable !== null && variable.scope.type === "global";
    }
    /**
     * Check if the identifier node is placed at to update members.
     * @param {ASTNode} id The Identifier node to check.
     * @param {Scope} scope A `escope.Scope` object to find variable (whichever).
     * @returns {boolean} `true` if the member of `id` was updated.
     */
    function isMemberWrite(id, scope) {
        const { parent } = id;
        return ((parent.type === "MemberExpression" &&
            parent.object === id &&
            (isAssignmentLeft(parent) ||
                isOperandOfMutationUnaryOperator(parent) ||
                isIterationVariable(parent))) ||
            isArgumentOfWellKnownMutationFunction(id, scope));
    }
    /**
     * Get the mutation node.
     * @param {ASTNode} id The Identifier node to get.
     * @returns {ASTNode} The mutation node.
     */
    function getWriteNode(id) {
        let node = id.parent;
        while (node &&
            node.type !== "AssignmentExpression" &&
            node.type !== "UpdateExpression" &&
            node.type !== "UnaryExpression" &&
            node.type !== "CallExpression" &&
            node.type !== "ForInStatement" &&
            node.type !== "ForOfStatement") {
            node = node.parent;
        }
        return node || id;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow assigning to imported bindings",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-import-assign"
            },
            schema: [],
            messages: {
                readonly: "'{{name}}' is read-only.",
                readonlyMember: "The members of '{{name}}' are read-only."
            }
        },
        create(context) {
            return {
                ImportDeclaration(node) {
                    const scope = context.getScope();
                    for (const variable of context.getDeclaredVariables(node)) {
                        const shouldCheckMembers = variable.defs.some(d => d.node.type === "ImportNamespaceSpecifier");
                        let prevIdNode = null;
                        for (const reference of variable.references) {
                            const idNode = reference.identifier;
                            /*
                             * AssignmentPattern (e.g. `[a = 0] = b`) makes two write
                             * references for the same identifier. This should skip
                             * the one of the two in order to prevent redundant reports.
                             */
                            if (idNode === prevIdNode) {
                                continue;
                            }
                            prevIdNode = idNode;
                            if (reference.isWrite()) {
                                context.report({
                                    node: getWriteNode(idNode),
                                    messageId: "readonly",
                                    data: { name: idNode.name }
                                });
                            }
                            else if (shouldCheckMembers && isMemberWrite(idNode, scope)) {
                                context.report({
                                    node: getWriteNode(idNode),
                                    messageId: "readonlyMember",
                                    data: { name: idNode.name }
                                });
                            }
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-inline-comments.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow inline comments after code",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-inline-comments"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        ignorePattern: {
                            type: "string"
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedInlineComment: "Unexpected comment inline with code."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const options = context.options[0];
            let customIgnoreRegExp;
            if (options && options.ignorePattern) {
                customIgnoreRegExp = new RegExp(options.ignorePattern, "u");
            }
            /**
             * Will check that comments are not on lines starting with or ending with code
             * @param {ASTNode} node The comment node to check
             * @private
             * @returns {void}
             */
            function testCodeAroundComment(node) {
                const startLine = String(sourceCode.lines[node.loc.start.line - 1]), endLine = String(sourceCode.lines[node.loc.end.line - 1]), preamble = startLine.slice(0, node.loc.start.column).trim(), postamble = endLine.slice(node.loc.end.column).trim(), isPreambleEmpty = !preamble, isPostambleEmpty = !postamble;
                // Nothing on both sides
                if (isPreambleEmpty && isPostambleEmpty) {
                    return;
                }
                // Matches the ignore pattern
                if (customIgnoreRegExp && customIgnoreRegExp.test(node.value)) {
                    return;
                }
                // JSX Exception
                if ((isPreambleEmpty || preamble === "{") &&
                    (isPostambleEmpty || postamble === "}")) {
                    const enclosingNode = sourceCode.getNodeByRangeIndex(node.range[0]);
                    if (enclosingNode && enclosingNode.type === "JSXEmptyExpression") {
                        return;
                    }
                }
                // Don't report ESLint directive comments
                if (astUtils.isDirectiveComment(node)) {
                    return;
                }
                context.report({
                    node,
                    messageId: "unexpectedInlineComment"
                });
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program() {
                    sourceCode.getAllComments()
                        .filter(token => token.type !== "Shebang")
                        .forEach(testCodeAroundComment);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-inner-declarations.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const validParent = new Set(["Program", "ExportNamedDeclaration", "ExportDefaultDeclaration"]);
    const validBlockStatementParent = new Set(["FunctionDeclaration", "FunctionExpression", "ArrowFunctionExpression"]);
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow variable or `function` declarations in nested blocks",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-inner-declarations"
            },
            schema: [
                {
                    enum: ["functions", "both"]
                }
            ],
            messages: {
                moveDeclToRoot: "Move {{type}} declaration to {{body}} root."
            }
        },
        create(context) {
            /**
             * Ensure that a given node is at a program or function body's root.
             * @param {ASTNode} node Declaration node to check.
             * @returns {void}
             */
            function check(node) {
                const parent = node.parent;
                if (parent.type === "BlockStatement" && validBlockStatementParent.has(parent.parent.type)) {
                    return;
                }
                if (validParent.has(parent.type)) {
                    return;
                }
                const upperFunction = astUtils.getUpperFunction(parent);
                context.report({
                    node,
                    messageId: "moveDeclToRoot",
                    data: {
                        type: (node.type === "FunctionDeclaration" ? "function" : "variable"),
                        body: (upperFunction === null ? "program" : "function body")
                    }
                });
            }
            return {
                FunctionDeclaration: check,
                VariableDeclaration(node) {
                    if (context.options[0] === "both" && node.kind === "var") {
                        check(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-invalid-regexp.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const RegExpValidator = require(497) /* regexpp */.RegExpValidator;
    const validator = new RegExpValidator();
    const validFlags = /[gimuys]/gu;
    const undefined1 = void 0;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow invalid regular expression strings in `RegExp` constructors",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-invalid-regexp"
            },
            schema: [{
                    type: "object",
                    properties: {
                        allowConstructorFlags: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                regexMessage: "{{message}}."
            }
        },
        create(context) {
            const options = context.options[0];
            let allowedFlags = null;
            if (options && options.allowConstructorFlags) {
                const temp = options.allowConstructorFlags.join("").replace(validFlags, "");
                if (temp) {
                    allowedFlags = new RegExp(`[${temp}]`, "giu");
                }
            }
            /**
             * Check if node is a string
             * @param {ASTNode} node node to evaluate
             * @returns {boolean} True if its a string
             * @private
             */
            function isString(node) {
                return node && node.type === "Literal" && typeof node.value === "string";
            }
            /**
             * Gets flags of a regular expression created by the given `RegExp()` or `new RegExp()` call
             * Examples:
             *     new RegExp(".")         // => ""
             *     new RegExp(".", "gu")   // => "gu"
             *     new RegExp(".", flags)  // => null
             * @param {ASTNode} node `CallExpression` or `NewExpression` node
             * @returns {string|null} flags if they can be determined, `null` otherwise
             * @private
             */
            function getFlags(node) {
                if (node.arguments.length < 2) {
                    return "";
                }
                if (isString(node.arguments[1])) {
                    return node.arguments[1].value;
                }
                return null;
            }
            /**
             * Check syntax error in a given pattern.
             * @param {string} pattern The RegExp pattern to validate.
             * @param {boolean} uFlag The Unicode flag.
             * @returns {string|null} The syntax error.
             */
            function validateRegExpPattern(pattern, uFlag) {
                try {
                    validator.validatePattern(pattern, undefined1, undefined1, uFlag);
                    return null;
                }
                catch (err) {
                    return err.message;
                }
            }
            /**
             * Check syntax error in a given flags.
             * @param {string} flags The RegExp flags to validate.
             * @returns {string|null} The syntax error.
             */
            function validateRegExpFlags(flags) {
                try {
                    validator.validateFlags(flags);
                    return null;
                }
                catch {
                    return `Invalid flags supplied to RegExp constructor '${flags}'`;
                }
            }
            return {
                "CallExpression, NewExpression"(node) {
                    if (node.callee.type !== "Identifier" || node.callee.name !== "RegExp" || !isString(node.arguments[0])) {
                        return;
                    }
                    const pattern = node.arguments[0].value;
                    let flags = getFlags(node);
                    if (flags && allowedFlags) {
                        flags = flags.replace(allowedFlags, "");
                    }
                    const message = (flags && validateRegExpFlags(flags)) ||
                        (
                        // If flags are unknown, report the regex only if its pattern is invalid both with and without the "u" flag
                        flags === null
                            ? validateRegExpPattern(pattern, true) && validateRegExpPattern(pattern, false)
                            : validateRegExpPattern(pattern, flags.includes("u")));
                    if (message) {
                        context.report({
                            node,
                            messageId: "regexMessage",
                            data: { message }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-invalid-this.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `this` keywords outside of classes or class-like objects",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-invalid-this"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        capIsConstructor: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedThis: "Unexpected 'this'."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const capIsConstructor = options.capIsConstructor !== false;
            const stack = [], sourceCode = context.getSourceCode();
            /**
             * Gets the current checking context.
             *
             * The return value has a flag that whether or not `this` keyword is valid.
             * The flag is initialized when got at the first time.
             * @returns {{valid: boolean}}
             *   an object which has a flag that whether or not `this` keyword is valid.
             */
            stack.getCurrent = function () {
                const current = this[this.length - 1];
                if (!current.init) {
                    current.init = true;
                    current.valid = !astUtils.isDefaultThisBinding(current.node, sourceCode, { capIsConstructor });
                }
                return current;
            };
            /**
             * Pushs new checking context into the stack.
             *
             * The checking context is not initialized yet.
             * Because most functions don't have `this` keyword.
             * When `this` keyword was found, the checking context is initialized.
             * @param {ASTNode} node A function node that was entered.
             * @returns {void}
             */
            function enterFunction(node) {
                // `this` can be invalid only under strict mode.
                stack.push({
                    init: !context.getScope().isStrict,
                    node,
                    valid: true
                });
            }
            /**
             * Pops the current checking context from the stack.
             * @returns {void}
             */
            function exitFunction() {
                stack.pop();
            }
            return {
                /*
                 * `this` is invalid only under strict mode.
                 * Modules is always strict mode.
                 */
                Program(node) {
                    const scope = context.getScope(), features = context.parserOptions.ecmaFeatures || {};
                    stack.push({
                        init: true,
                        node,
                        valid: !(scope.isStrict ||
                            node.sourceType === "module" ||
                            (features.globalReturn && scope.childScopes[0].isStrict))
                    });
                },
                "Program:exit"() {
                    stack.pop();
                },
                FunctionDeclaration: enterFunction,
                "FunctionDeclaration:exit": exitFunction,
                FunctionExpression: enterFunction,
                "FunctionExpression:exit": exitFunction,
                // Reports if `this` of the current context is invalid.
                ThisExpression(node) {
                    const current = stack.getCurrent();
                    if (current && !current.valid) {
                        context.report({
                            node,
                            messageId: "unexpectedThis"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-irregular-whitespace.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Constants
    //------------------------------------------------------------------------------
    const ALL_IRREGULARS = /[\f\v\u0085\ufeff\u00a0\u1680\u180e\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a\u200b\u202f\u205f\u3000\u2028\u2029]/u;
    const IRREGULAR_WHITESPACE = /[\f\v\u0085\ufeff\u00a0\u1680\u180e\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a\u200b\u202f\u205f\u3000]+/mgu;
    const IRREGULAR_LINE_TERMINATORS = /[\u2028\u2029]/mgu;
    const LINE_BREAK = astUtils.createGlobalLinebreakMatcher();
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow irregular whitespace",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-irregular-whitespace"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        skipComments: {
                            type: "boolean",
                            default: false
                        },
                        skipStrings: {
                            type: "boolean",
                            default: true
                        },
                        skipTemplates: {
                            type: "boolean",
                            default: false
                        },
                        skipRegExps: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                noIrregularWhitespace: "Irregular whitespace not allowed."
            }
        },
        create(context) {
            // Module store of errors that we have found
            let errors = [];
            // Lookup the `skipComments` option, which defaults to `false`.
            const options = context.options[0] || {};
            const skipComments = !!options.skipComments;
            const skipStrings = options.skipStrings !== false;
            const skipRegExps = !!options.skipRegExps;
            const skipTemplates = !!options.skipTemplates;
            const sourceCode = context.getSourceCode();
            const commentNodes = sourceCode.getAllComments();
            /**
             * Removes errors that occur inside the given node
             * @param {ASTNode} node to check for matching errors.
             * @returns {void}
             * @private
             */
            function removeWhitespaceError(node) {
                const locStart = node.loc.start;
                const locEnd = node.loc.end;
                errors = errors.filter(({ loc: { start: errorLocStart } }) => (errorLocStart.line < locStart.line ||
                    errorLocStart.line === locStart.line && errorLocStart.column < locStart.column ||
                    errorLocStart.line === locEnd.line && errorLocStart.column >= locEnd.column ||
                    errorLocStart.line > locEnd.line));
            }
            /**
             * Checks identifier or literal nodes for errors that we are choosing to ignore and calls the relevant methods to remove the errors
             * @param {ASTNode} node to check for matching errors.
             * @returns {void}
             * @private
             */
            function removeInvalidNodeErrorsInIdentifierOrLiteral(node) {
                const shouldCheckStrings = skipStrings && (typeof node.value === "string");
                const shouldCheckRegExps = skipRegExps && Boolean(node.regex);
                if (shouldCheckStrings || shouldCheckRegExps) {
                    // If we have irregular characters remove them from the errors list
                    if (ALL_IRREGULARS.test(node.raw)) {
                        removeWhitespaceError(node);
                    }
                }
            }
            /**
             * Checks template string literal nodes for errors that we are choosing to ignore and calls the relevant methods to remove the errors
             * @param {ASTNode} node to check for matching errors.
             * @returns {void}
             * @private
             */
            function removeInvalidNodeErrorsInTemplateLiteral(node) {
                if (typeof node.value.raw === "string") {
                    if (ALL_IRREGULARS.test(node.value.raw)) {
                        removeWhitespaceError(node);
                    }
                }
            }
            /**
             * Checks comment nodes for errors that we are choosing to ignore and calls the relevant methods to remove the errors
             * @param {ASTNode} node to check for matching errors.
             * @returns {void}
             * @private
             */
            function removeInvalidNodeErrorsInComment(node) {
                if (ALL_IRREGULARS.test(node.value)) {
                    removeWhitespaceError(node);
                }
            }
            /**
             * Checks the program source for irregular whitespace
             * @param {ASTNode} node The program node
             * @returns {void}
             * @private
             */
            function checkForIrregularWhitespace(node) {
                const sourceLines = sourceCode.lines;
                sourceLines.forEach((sourceLine, lineIndex) => {
                    const lineNumber = lineIndex + 1;
                    let match;
                    while ((match = IRREGULAR_WHITESPACE.exec(sourceLine)) !== null) {
                        errors.push({
                            node,
                            messageId: "noIrregularWhitespace",
                            loc: {
                                start: {
                                    line: lineNumber,
                                    column: match.index
                                },
                                end: {
                                    line: lineNumber,
                                    column: match.index + match[0].length
                                }
                            }
                        });
                    }
                });
            }
            /**
             * Checks the program source for irregular line terminators
             * @param {ASTNode} node The program node
             * @returns {void}
             * @private
             */
            function checkForIrregularLineTerminators(node) {
                const source = sourceCode.getText(), sourceLines = sourceCode.lines, linebreaks = source.match(LINE_BREAK);
                let lastLineIndex = -1, match;
                while ((match = IRREGULAR_LINE_TERMINATORS.exec(source)) !== null) {
                    const lineIndex = linebreaks.indexOf(match[0], lastLineIndex + 1) || 0;
                    errors.push({
                        node,
                        messageId: "noIrregularWhitespace",
                        loc: {
                            start: {
                                line: lineIndex + 1,
                                column: sourceLines[lineIndex].length
                            },
                            end: {
                                line: lineIndex + 2,
                                column: 0
                            }
                        }
                    });
                    lastLineIndex = lineIndex;
                }
            }
            /**
             * A no-op function to act as placeholder for comment accumulation when the `skipComments` option is `false`.
             * @returns {void}
             * @private
             */
            function noop() { }
            const nodes = {};
            if (ALL_IRREGULARS.test(sourceCode.getText())) {
                nodes.Program = function (node) {
                    /*
                     * As we can easily fire warnings for all white space issues with
                     * all the source its simpler to fire them here.
                     * This means we can check all the application code without having
                     * to worry about issues caused in the parser tokens.
                     * When writing this code also evaluating per node was missing out
                     * connecting tokens in some cases.
                     * We can later filter the errors when they are found to be not an
                     * issue in nodes we don't care about.
                     */
                    checkForIrregularWhitespace(node);
                    checkForIrregularLineTerminators(node);
                };
                nodes.Identifier = removeInvalidNodeErrorsInIdentifierOrLiteral;
                nodes.Literal = removeInvalidNodeErrorsInIdentifierOrLiteral;
                nodes.TemplateElement = skipTemplates ? removeInvalidNodeErrorsInTemplateLiteral : noop;
                nodes["Program:exit"] = function () {
                    if (skipComments) {
                        // First strip errors occurring in comment nodes.
                        commentNodes.forEach(removeInvalidNodeErrorsInComment);
                    }
                    // If we have any errors remaining report on them
                    errors.forEach(error => context.report(error));
                };
            }
            else {
                nodes.Program = noop;
            }
            return nodes;
        }
    };
},
/* eslint/lib/rules/no-iterator.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { getStaticPropertyName } = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the use of the `__iterator__` property",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-iterator"
            },
            schema: [],
            messages: {
                noIterator: "Reserved name '__iterator__'."
            }
        },
        create(context) {
            return {
                MemberExpression(node) {
                    if (getStaticPropertyName(node) === "__iterator__") {
                        context.report({
                            node,
                            messageId: "noIterator"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-label-var.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow labels that share a name with a variable",
                category: "Variables",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-label-var"
            },
            schema: [],
            messages: {
                identifierClashWithLabel: "Found identifier with same name as label."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Check if the identifier is present inside current scope
             * @param {Object} scope current scope
             * @param {string} name To evaluate
             * @returns {boolean} True if its present
             * @private
             */
            function findIdentifier(scope, name) {
                return astUtils.getVariableByName(scope, name) !== null;
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                LabeledStatement(node) {
                    // Fetch the innermost scope.
                    const scope = context.getScope();
                    /*
                     * Recursively find the identifier walking up the scope, starting
                     * with the innermost scope.
                     */
                    if (findIdentifier(scope, node.label.name)) {
                        context.report({
                            node,
                            messageId: "identifierClashWithLabel"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-labels.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow labeled statements",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-labels"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowLoop: {
                            type: "boolean",
                            default: false
                        },
                        allowSwitch: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedLabel: "Unexpected labeled statement.",
                unexpectedLabelInBreak: "Unexpected label in break statement.",
                unexpectedLabelInContinue: "Unexpected label in continue statement."
            }
        },
        create(context) {
            const options = context.options[0];
            const allowLoop = options && options.allowLoop;
            const allowSwitch = options && options.allowSwitch;
            let scopeInfo = null;
            /**
             * Gets the kind of a given node.
             * @param {ASTNode} node A node to get.
             * @returns {string} The kind of the node.
             */
            function getBodyKind(node) {
                if (astUtils.isLoop(node)) {
                    return "loop";
                }
                if (node.type === "SwitchStatement") {
                    return "switch";
                }
                return "other";
            }
            /**
             * Checks whether the label of a given kind is allowed or not.
             * @param {string} kind A kind to check.
             * @returns {boolean} `true` if the kind is allowed.
             */
            function isAllowed(kind) {
                switch (kind) {
                    case "loop": return allowLoop;
                    case "switch": return allowSwitch;
                    default: return false;
                }
            }
            /**
             * Checks whether a given name is a label of a loop or not.
             * @param {string} label A name of a label to check.
             * @returns {boolean} `true` if the name is a label of a loop.
             */
            function getKind(label) {
                let info = scopeInfo;
                while (info) {
                    if (info.label === label) {
                        return info.kind;
                    }
                    info = info.upper;
                }
                /* istanbul ignore next: syntax error */
                return "other";
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                LabeledStatement(node) {
                    scopeInfo = {
                        label: node.label.name,
                        kind: getBodyKind(node.body),
                        upper: scopeInfo
                    };
                },
                "LabeledStatement:exit"(node) {
                    if (!isAllowed(scopeInfo.kind)) {
                        context.report({
                            node,
                            messageId: "unexpectedLabel"
                        });
                    }
                    scopeInfo = scopeInfo.upper;
                },
                BreakStatement(node) {
                    if (node.label && !isAllowed(getKind(node.label.name))) {
                        context.report({
                            node,
                            messageId: "unexpectedLabelInBreak"
                        });
                    }
                },
                ContinueStatement(node) {
                    if (node.label && !isAllowed(getKind(node.label.name))) {
                        context.report({
                            node,
                            messageId: "unexpectedLabelInContinue"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-lone-blocks.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary nested blocks",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-lone-blocks"
            },
            schema: [],
            messages: {
                redundantBlock: "Block is redundant.",
                redundantNestedBlock: "Nested block is redundant."
            }
        },
        create(context) {
            // A stack of lone blocks to be checked for block-level bindings
            const loneBlocks = [];
            let ruleDef;
            /**
             * Reports a node as invalid.
             * @param {ASTNode} node The node to be reported.
             * @returns {void}
             */
            function report(node) {
                const messageId = node.parent.type === "BlockStatement" ? "redundantNestedBlock" : "redundantBlock";
                context.report({
                    node,
                    messageId
                });
            }
            /**
             * Checks for any occurrence of a BlockStatement in a place where lists of statements can appear
             * @param {ASTNode} node The node to check
             * @returns {boolean} True if the node is a lone block.
             */
            function isLoneBlock(node) {
                return node.parent.type === "BlockStatement" ||
                    node.parent.type === "Program" ||
                    // Don't report blocks in switch cases if the block is the only statement of the case.
                    node.parent.type === "SwitchCase" && !(node.parent.consequent[0] === node && node.parent.consequent.length === 1);
            }
            /**
             * Checks the enclosing block of the current node for block-level bindings,
             * and "marks it" as valid if any.
             * @returns {void}
             */
            function markLoneBlock() {
                if (loneBlocks.length === 0) {
                    return;
                }
                const block = context.getAncestors().pop();
                if (loneBlocks[loneBlocks.length - 1] === block) {
                    loneBlocks.pop();
                }
            }
            // Default rule definition: report all lone blocks
            ruleDef = {
                BlockStatement(node) {
                    if (isLoneBlock(node)) {
                        report(node);
                    }
                }
            };
            // ES6: report blocks without block-level bindings, or that's only child of another block
            if (context.parserOptions.ecmaVersion >= 6) {
                ruleDef = {
                    BlockStatement(node) {
                        if (isLoneBlock(node)) {
                            loneBlocks.push(node);
                        }
                    },
                    "BlockStatement:exit"(node) {
                        if (loneBlocks.length > 0 && loneBlocks[loneBlocks.length - 1] === node) {
                            loneBlocks.pop();
                            report(node);
                        }
                        else if (node.parent.type === "BlockStatement" &&
                            node.parent.body.length === 1) {
                            report(node);
                        }
                    }
                };
                ruleDef.VariableDeclaration = function (node) {
                    if (node.kind === "let" || node.kind === "const") {
                        markLoneBlock();
                    }
                };
                ruleDef.FunctionDeclaration = function () {
                    if (context.getScope().isStrict) {
                        markLoneBlock();
                    }
                };
                ruleDef.ClassDeclaration = markLoneBlock;
            }
            return ruleDef;
        }
    };
},
/* eslint/lib/rules/no-lonely-if.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `if` statements as the only statement in `else` blocks",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-lonely-if"
            },
            schema: [],
            fixable: "code",
            messages: {
                unexpectedLonelyIf: "Unexpected if as the only statement in an else block."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                IfStatement(node) {
                    const ancestors = context.getAncestors(), parent = ancestors.pop(), grandparent = ancestors.pop();
                    if (parent && parent.type === "BlockStatement" &&
                        parent.body.length === 1 && grandparent &&
                        grandparent.type === "IfStatement" &&
                        parent === grandparent.alternate) {
                        context.report({
                            node,
                            messageId: "unexpectedLonelyIf",
                            fix(fixer) {
                                const openingElseCurly = sourceCode.getFirstToken(parent);
                                const closingElseCurly = sourceCode.getLastToken(parent);
                                const elseKeyword = sourceCode.getTokenBefore(openingElseCurly);
                                const tokenAfterElseBlock = sourceCode.getTokenAfter(closingElseCurly);
                                const lastIfToken = sourceCode.getLastToken(node.consequent);
                                const sourceText = sourceCode.getText();
                                if (sourceText.slice(openingElseCurly.range[1], node.range[0]).trim() || sourceText.slice(node.range[1], closingElseCurly.range[0]).trim()) {
                                    // Don't fix if there are any non-whitespace characters interfering (e.g. comments)
                                    return null;
                                }
                                if (node.consequent.type !== "BlockStatement" && lastIfToken.value !== ";" && tokenAfterElseBlock &&
                                    (node.consequent.loc.end.line === tokenAfterElseBlock.loc.start.line ||
                                        /^[([/+`-]/u.test(tokenAfterElseBlock.value) ||
                                        lastIfToken.value === "++" ||
                                        lastIfToken.value === "--")) {
                                    /*
                                     * If the `if` statement has no block, and is not followed by a semicolon, make sure that fixing
                                     * the issue would not change semantics due to ASI. If this would happen, don't do a fix.
                                     */
                                    return null;
                                }
                                return fixer.replaceTextRange([openingElseCurly.range[0], closingElseCurly.range[1]], (elseKeyword.range[1] === openingElseCurly.range[0] ? " " : "") + sourceCode.getText(node));
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-loop-func.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Gets the containing loop node of a specified node.
     *
     * We don't need to check nested functions, so this ignores those.
     * `Scope.through` contains references of nested functions.
     * @param {ASTNode} node An AST node to get.
     * @returns {ASTNode|null} The containing loop node of the specified node, or
     *      `null`.
     */
    function getContainingLoopNode(node) {
        for (let currentNode = node; currentNode.parent; currentNode = currentNode.parent) {
            const parent = currentNode.parent;
            switch (parent.type) {
                case "WhileStatement":
                case "DoWhileStatement":
                    return parent;
                case "ForStatement":
                    // `init` is outside of the loop.
                    if (parent.init !== currentNode) {
                        return parent;
                    }
                    break;
                case "ForInStatement":
                case "ForOfStatement":
                    // `right` is outside of the loop.
                    if (parent.right !== currentNode) {
                        return parent;
                    }
                    break;
                case "ArrowFunctionExpression":
                case "FunctionExpression":
                case "FunctionDeclaration":
                    // We don't need to check nested functions.
                    return null;
                default:
                    break;
            }
        }
        return null;
    }
    /**
     * Gets the containing loop node of a given node.
     * If the loop was nested, this returns the most outer loop.
     * @param {ASTNode} node A node to get. This is a loop node.
     * @param {ASTNode|null} excludedNode A node that the result node should not
     *      include.
     * @returns {ASTNode} The most outer loop node.
     */
    function getTopLoopNode(node, excludedNode) {
        const border = excludedNode ? excludedNode.range[1] : 0;
        let retv = node;
        let containingLoopNode = node;
        while (containingLoopNode && containingLoopNode.range[0] >= border) {
            retv = containingLoopNode;
            containingLoopNode = getContainingLoopNode(containingLoopNode);
        }
        return retv;
    }
    /**
     * Checks whether a given reference which refers to an upper scope's variable is
     * safe or not.
     * @param {ASTNode} loopNode A containing loop node.
     * @param {eslint-scope.Reference} reference A reference to check.
     * @returns {boolean} `true` if the reference is safe or not.
     */
    function isSafe(loopNode, reference) {
        const variable = reference.resolved;
        const definition = variable && variable.defs[0];
        const declaration = definition && definition.parent;
        const kind = (declaration && declaration.type === "VariableDeclaration")
            ? declaration.kind
            : "";
        // Variables which are declared by `const` is safe.
        if (kind === "const") {
            return true;
        }
        /*
         * Variables which are declared by `let` in the loop is safe.
         * It's a different instance from the next loop step's.
         */
        if (kind === "let" &&
            declaration.range[0] > loopNode.range[0] &&
            declaration.range[1] < loopNode.range[1]) {
            return true;
        }
        /*
         * WriteReferences which exist after this border are unsafe because those
         * can modify the variable.
         */
        const border = getTopLoopNode(loopNode, (kind === "let") ? declaration : null).range[0];
        /**
         * Checks whether a given reference is safe or not.
         * The reference is every reference of the upper scope's variable we are
         * looking now.
         *
         * It's safeafe if the reference matches one of the following condition.
         * - is readonly.
         * - doesn't exist inside a local function and after the border.
         * @param {eslint-scope.Reference} upperRef A reference to check.
         * @returns {boolean} `true` if the reference is safe.
         */
        function isSafeReference(upperRef) {
            const id = upperRef.identifier;
            return (!upperRef.isWrite() ||
                variable.scope.variableScope === upperRef.from.variableScope &&
                    id.range[0] < border);
        }
        return Boolean(variable) && variable.references.every(isSafeReference);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow function declarations that contain unsafe references inside loop statements",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-loop-func"
            },
            schema: [],
            messages: {
                unsafeRefs: "Function declared in a loop contains unsafe references to variable(s) {{ varNames }}."
            }
        },
        create(context) {
            /**
             * Reports functions which match the following condition:
             *
             * - has a loop node in ancestors.
             * - has any references which refers to an unsafe variable.
             * @param {ASTNode} node The AST node to check.
             * @returns {boolean} Whether or not the node is within a loop.
             */
            function checkForLoops(node) {
                const loopNode = getContainingLoopNode(node);
                if (!loopNode) {
                    return;
                }
                const references = context.getScope().through;
                const unsafeRefs = references.filter(r => !isSafe(loopNode, r)).map(r => r.identifier.name);
                if (unsafeRefs.length > 0) {
                    context.report({
                        node,
                        messageId: "unsafeRefs",
                        data: { varNames: `'${unsafeRefs.join("', '")}'` }
                    });
                }
            }
            return {
                ArrowFunctionExpression: checkForLoops,
                FunctionExpression: checkForLoops,
                FunctionDeclaration: checkForLoops
            };
        }
    };
},
/* eslint/lib/rules/no-loss-of-precision.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow literal numbers that lose precision",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-loss-of-precision"
            },
            schema: [],
            messages: {
                noLossOfPrecision: "This number literal will lose precision at runtime."
            }
        },
        create(context) {
            /**
             * Returns whether the node is number literal
             * @param {Node} node the node literal being evaluated
             * @returns {boolean} true if the node is a number literal
             */
            function isNumber(node) {
                return typeof node.value === "number";
            }
            /**
             * Gets the source code of the given number literal. Removes `_` numeric separators from the result.
             * @param {Node} node the number `Literal` node
             * @returns {string} raw source code of the literal, without numeric separators
             */
            function getRaw(node) {
                return node.raw.replace(/_/gu, "");
            }
            /**
             * Checks whether the number is  base ten
             * @param {ASTNode} node the node being evaluated
             * @returns {boolean} true if the node is in base ten
             */
            function isBaseTen(node) {
                const prefixes = ["0x", "0X", "0b", "0B", "0o", "0O"];
                return prefixes.every(prefix => !node.raw.startsWith(prefix)) &&
                    !/^0[0-7]+$/u.test(node.raw);
            }
            /**
             * Checks that the user-intended non-base ten number equals the actual number after is has been converted to the Number type
             * @param {Node} node the node being evaluated
             * @returns {boolean} true if they do not match
             */
            function notBaseTenLosesPrecision(node) {
                const rawString = getRaw(node).toUpperCase();
                let base = 0;
                if (rawString.startsWith("0B")) {
                    base = 2;
                }
                else if (rawString.startsWith("0X")) {
                    base = 16;
                }
                else {
                    base = 8;
                }
                return !rawString.endsWith(node.value.toString(base).toUpperCase());
            }
            /**
             * Adds a decimal point to the numeric string at index 1
             * @param {string} stringNumber the numeric string without any decimal point
             * @returns {string} the numeric string with a decimal point in the proper place
             */
            function addDecimalPointToNumber(stringNumber) {
                return `${stringNumber.slice(0, 1)}.${stringNumber.slice(1)}`;
            }
            /**
             * Returns the number stripped of leading zeros
             * @param {string} numberAsString the string representation of the number
             * @returns {string} the stripped string
             */
            function removeLeadingZeros(numberAsString) {
                return numberAsString.replace(/^0*/u, "");
            }
            /**
             * Returns the number stripped of trailing zeros
             * @param {string} numberAsString the string representation of the number
             * @returns {string} the stripped string
             */
            function removeTrailingZeros(numberAsString) {
                return numberAsString.replace(/0*$/u, "");
            }
            /**
             * Converts an integer to to an object containing the integer's coefficient and order of magnitude
             * @param {string} stringInteger the string representation of the integer being converted
             * @returns {Object} the object containing the integer's coefficient and order of magnitude
             */
            function normalizeInteger(stringInteger) {
                const significantDigits = removeTrailingZeros(removeLeadingZeros(stringInteger));
                return {
                    magnitude: stringInteger.startsWith("0") ? stringInteger.length - 2 : stringInteger.length - 1,
                    coefficient: addDecimalPointToNumber(significantDigits)
                };
            }
            /**
             *
             * Converts a float to to an object containing the floats's coefficient and order of magnitude
             * @param {string} stringFloat the string representation of the float being converted
             * @returns {Object} the object containing the integer's coefficient and order of magnitude
             */
            function normalizeFloat(stringFloat) {
                const trimmedFloat = removeLeadingZeros(stringFloat);
                if (trimmedFloat.startsWith(".")) {
                    const decimalDigits = trimmedFloat.split(".").pop();
                    const significantDigits = removeLeadingZeros(decimalDigits);
                    return {
                        magnitude: significantDigits.length - decimalDigits.length - 1,
                        coefficient: addDecimalPointToNumber(significantDigits)
                    };
                }
                return {
                    magnitude: trimmedFloat.indexOf(".") - 1,
                    coefficient: addDecimalPointToNumber(trimmedFloat.replace(".", ""))
                };
            }
            /**
             * Converts a base ten number to proper scientific notation
             * @param {string} stringNumber the string representation of the base ten number to be converted
             * @returns {string} the number converted to scientific notation
             */
            function convertNumberToScientificNotation(stringNumber) {
                const splitNumber = stringNumber.replace("E", "e").split("e");
                const originalCoefficient = splitNumber[0];
                const normalizedNumber = stringNumber.includes(".") ? normalizeFloat(originalCoefficient)
                    : normalizeInteger(originalCoefficient);
                const normalizedCoefficient = normalizedNumber.coefficient;
                const magnitude = splitNumber.length > 1 ? (parseInt(splitNumber[1], 10) + normalizedNumber.magnitude)
                    : normalizedNumber.magnitude;
                return `${normalizedCoefficient}e${magnitude}`;
            }
            /**
             * Checks that the user-intended base ten number equals the actual number after is has been converted to the Number type
             * @param {Node} node the node being evaluated
             * @returns {boolean} true if they do not match
             */
            function baseTenLosesPrecision(node) {
                const normalizedRawNumber = convertNumberToScientificNotation(getRaw(node));
                const requestedPrecision = normalizedRawNumber.split("e")[0].replace(".", "").length;
                if (requestedPrecision > 100) {
                    return true;
                }
                const storedNumber = node.value.toPrecision(requestedPrecision);
                const normalizedStoredNumber = convertNumberToScientificNotation(storedNumber);
                return normalizedRawNumber !== normalizedStoredNumber;
            }
            /**
             * Checks that the user-intended number equals the actual number after is has been converted to the Number type
             * @param {Node} node the node being evaluated
             * @returns {boolean} true if they do not match
             */
            function losesPrecision(node) {
                return isBaseTen(node) ? baseTenLosesPrecision(node) : notBaseTenLosesPrecision(node);
            }
            return {
                Literal(node) {
                    if (node.value && isNumber(node) && losesPrecision(node)) {
                        context.report({
                            messageId: "noLossOfPrecision",
                            node
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-magic-numbers.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    // Maximum array length by the ECMAScript Specification.
    const MAX_ARRAY_LENGTH = 2 ** 32 - 1;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    /**
     * Convert the value to bigint if it's a string. Otherwise return the value as-is.
     * @param {bigint|number|string} x The value to normalize.
     * @returns {bigint|number} The normalized value.
     */
    function normalizeIgnoreValue(x) {
        if (typeof x === "string") {
            return BigInt(x.slice(0, -1));
        }
        return x;
    }
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow magic numbers",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-magic-numbers"
            },
            schema: [{
                    type: "object",
                    properties: {
                        detectObjects: {
                            type: "boolean",
                            default: false
                        },
                        enforceConst: {
                            type: "boolean",
                            default: false
                        },
                        ignore: {
                            type: "array",
                            items: {
                                anyOf: [
                                    { type: "number" },
                                    { type: "string", pattern: "^[+-]?(?:0|[1-9][0-9]*)n$" }
                                ]
                            },
                            uniqueItems: true
                        },
                        ignoreArrayIndexes: {
                            type: "boolean",
                            default: false
                        },
                        ignoreDefaultValues: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                useConst: "Number constants declarations must use 'const'.",
                noMagic: "No magic number: {{raw}}."
            }
        },
        create(context) {
            const config = context.options[0] || {}, detectObjects = !!config.detectObjects, enforceConst = !!config.enforceConst, ignore = (config.ignore || []).map(normalizeIgnoreValue), ignoreArrayIndexes = !!config.ignoreArrayIndexes, ignoreDefaultValues = !!config.ignoreDefaultValues;
            const okTypes = detectObjects ? [] : ["ObjectExpression", "Property", "AssignmentExpression"];
            /**
             * Returns whether the rule is configured to ignore the given value
             * @param {bigint|number} value The value to check
             * @returns {boolean} true if the value is ignored
             */
            function isIgnoredValue(value) {
                return ignore.indexOf(value) !== -1;
            }
            /**
             * Returns whether the number is a default value assignment.
             * @param {ASTNode} fullNumberNode `Literal` or `UnaryExpression` full number node
             * @returns {boolean} true if the number is a default value
             */
            function isDefaultValue(fullNumberNode) {
                const parent = fullNumberNode.parent;
                return parent.type === "AssignmentPattern" && parent.right === fullNumberNode;
            }
            /**
             * Returns whether the given node is used as a radix within parseInt() or Number.parseInt()
             * @param {ASTNode} fullNumberNode `Literal` or `UnaryExpression` full number node
             * @returns {boolean} true if the node is radix
             */
            function isParseIntRadix(fullNumberNode) {
                const parent = fullNumberNode.parent;
                return parent.type === "CallExpression" && fullNumberNode === parent.arguments[1] &&
                    (astUtils.isSpecificId(parent.callee, "parseInt") ||
                        astUtils.isSpecificMemberAccess(parent.callee, "Number", "parseInt"));
            }
            /**
             * Returns whether the given node is a direct child of a JSX node.
             * In particular, it aims to detect numbers used as prop values in JSX tags.
             * Example: <input maxLength={10} />
             * @param {ASTNode} fullNumberNode `Literal` or `UnaryExpression` full number node
             * @returns {boolean} true if the node is a JSX number
             */
            function isJSXNumber(fullNumberNode) {
                return fullNumberNode.parent.type.indexOf("JSX") === 0;
            }
            /**
             * Returns whether the given node is used as an array index.
             * Value must coerce to a valid array index name: "0", "1", "2" ... "4294967294".
             *
             * All other values, like "-1", "2.5", or "4294967295", are just "normal" object properties,
             * which can be created and accessed on an array in addition to the array index properties,
             * but they don't affect array's length and are not considered by methods such as .map(), .forEach() etc.
             *
             * The maximum array length by the specification is 2 ** 32 - 1 = 4294967295,
             * thus the maximum valid index is 2 ** 32 - 2 = 4294967294.
             *
             * All notations are allowed, as long as the value coerces to one of "0", "1", "2" ... "4294967294".
             *
             * Valid examples:
             * a[0], a[1], a[1.2e1], a[0xAB], a[0n], a[1n]
             * a[-0] (same as a[0] because -0 coerces to "0")
             * a[-0n] (-0n evaluates to 0n)
             *
             * Invalid examples:
             * a[-1], a[-0xAB], a[-1n], a[2.5], a[1.23e1], a[12e-1]
             * a[4294967295] (above the max index, it's an access to a regular property a["4294967295"])
             * a[999999999999999999999] (even if it wasn't above the max index, it would be a["1e+21"])
             * a[1e310] (same as a["Infinity"])
             * @param {ASTNode} fullNumberNode `Literal` or `UnaryExpression` full number node
             * @param {bigint|number} value Value expressed by the fullNumberNode
             * @returns {boolean} true if the node is a valid array index
             */
            function isArrayIndex(fullNumberNode, value) {
                const parent = fullNumberNode.parent;
                return parent.type === "MemberExpression" && parent.property === fullNumberNode &&
                    (Number.isInteger(value) || typeof value === "bigint") &&
                    value >= 0 && value < MAX_ARRAY_LENGTH;
            }
            return {
                Literal(node) {
                    if (!astUtils.isNumericLiteral(node)) {
                        return;
                    }
                    let fullNumberNode;
                    let value;
                    let raw;
                    // Treat unary minus as a part of the number
                    if (node.parent.type === "UnaryExpression" && node.parent.operator === "-") {
                        fullNumberNode = node.parent;
                        value = -node.value;
                        raw = `-${node.raw}`;
                    }
                    else {
                        fullNumberNode = node;
                        value = node.value;
                        raw = node.raw;
                    }
                    const parent = fullNumberNode.parent;
                    // Always allow radix arguments and JSX props
                    if (isIgnoredValue(value) ||
                        (ignoreDefaultValues && isDefaultValue(fullNumberNode)) ||
                        isParseIntRadix(fullNumberNode) ||
                        isJSXNumber(fullNumberNode) ||
                        (ignoreArrayIndexes && isArrayIndex(fullNumberNode, value))) {
                        return;
                    }
                    if (parent.type === "VariableDeclarator") {
                        if (enforceConst && parent.parent.kind !== "const") {
                            context.report({
                                node: fullNumberNode,
                                messageId: "useConst"
                            });
                        }
                    }
                    else if (okTypes.indexOf(parent.type) === -1 ||
                        (parent.type === "AssignmentExpression" && parent.left.type === "Identifier")) {
                        context.report({
                            node: fullNumberNode,
                            messageId: "noMagic",
                            data: {
                                raw
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-misleading-character-class.js */ function _(require, module, exports, __esModule, __esExport) {
    const { CALL, CONSTRUCT, ReferenceTracker, getStringIfConstant } = require(350) /* eslint-utils */;
    const { RegExpParser, visitRegExpAST } = require(497) /* regexpp */;
    const { isCombiningCharacter, isEmojiModifier, isRegionalIndicatorSymbol, isSurrogatePair } = require(544) /* ./utils/unicode */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Iterate character sequences of a given nodes.
     *
     * CharacterClassRange syntax can steal a part of character sequence,
     * so this function reverts CharacterClassRange syntax and restore the sequence.
     * @param {regexpp.AST.CharacterClassElement[]} nodes The node list to iterate character sequences.
     * @returns {IterableIterator<number[]>} The list of character sequences.
     */
    function* iterateCharacterSequence(nodes) {
        let seq = [];
        for (const node of nodes) {
            switch (node.type) {
                case "Character":
                    seq.push(node.value);
                    break;
                case "CharacterClassRange":
                    seq.push(node.min.value);
                    yield seq;
                    seq = [node.max.value];
                    break;
                case "CharacterSet":
                    if (seq.length > 0) {
                        yield seq;
                        seq = [];
                    }
                    break;
                // no default
            }
        }
        if (seq.length > 0) {
            yield seq;
        }
    }
    const hasCharacterSequence = {
        surrogatePairWithoutUFlag(chars) {
            return chars.some((c, i) => i !== 0 && isSurrogatePair(chars[i - 1], c));
        },
        combiningClass(chars) {
            return chars.some((c, i) => (i !== 0 &&
                isCombiningCharacter(c) &&
                !isCombiningCharacter(chars[i - 1])));
        },
        emojiModifier(chars) {
            return chars.some((c, i) => (i !== 0 &&
                isEmojiModifier(c) &&
                !isEmojiModifier(chars[i - 1])));
        },
        regionalIndicatorSymbol(chars) {
            return chars.some((c, i) => (i !== 0 &&
                isRegionalIndicatorSymbol(c) &&
                isRegionalIndicatorSymbol(chars[i - 1])));
        },
        zwj(chars) {
            const lastIndex = chars.length - 1;
            return chars.some((c, i) => (i !== 0 &&
                i !== lastIndex &&
                c === 0x200d &&
                chars[i - 1] !== 0x200d &&
                chars[i + 1] !== 0x200d));
        }
    };
    const kinds = Object.keys(hasCharacterSequence);
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow characters which are made with multiple code points in character class syntax",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-misleading-character-class"
            },
            schema: [],
            messages: {
                surrogatePairWithoutUFlag: "Unexpected surrogate pair in character class. Use 'u' flag.",
                combiningClass: "Unexpected combined character in character class.",
                emojiModifier: "Unexpected modified Emoji in character class.",
                regionalIndicatorSymbol: "Unexpected national flag in character class.",
                zwj: "Unexpected joined character sequence in character class."
            }
        },
        create(context) {
            const parser = new RegExpParser();
            /**
             * Verify a given regular expression.
             * @param {Node} node The node to report.
             * @param {string} pattern The regular expression pattern to verify.
             * @param {string} flags The flags of the regular expression.
             * @returns {void}
             */
            function verify(node, pattern, flags) {
                const has = {
                    surrogatePairWithoutUFlag: false,
                    combiningClass: false,
                    variationSelector: false,
                    emojiModifier: false,
                    regionalIndicatorSymbol: false,
                    zwj: false
                };
                let patternNode;
                try {
                    patternNode = parser.parsePattern(pattern, 0, pattern.length, flags.includes("u"));
                }
                catch {
                    // Ignore regular expressions with syntax errors
                    return;
                }
                visitRegExpAST(patternNode, {
                    onCharacterClassEnter(ccNode) {
                        for (const chars of iterateCharacterSequence(ccNode.elements)) {
                            for (const kind of kinds) {
                                has[kind] = has[kind] || hasCharacterSequence[kind](chars);
                            }
                        }
                    }
                });
                for (const kind of kinds) {
                    if (has[kind]) {
                        context.report({ node, messageId: kind });
                    }
                }
            }
            return {
                "Literal[regex]"(node) {
                    verify(node, node.regex.pattern, node.regex.flags);
                },
                "Program"() {
                    const scope = context.getScope();
                    const tracker = new ReferenceTracker(scope);
                    /*
                     * Iterate calls of RegExp.
                     * E.g., `new RegExp()`, `RegExp()`, `new window.RegExp()`,
                     *       `const {RegExp: a} = window; new a()`, etc...
                     */
                    for (const { node } of tracker.iterateGlobalReferences({
                        RegExp: { [CALL]: true, [CONSTRUCT]: true }
                    })) {
                        const [patternNode, flagsNode] = node.arguments;
                        const pattern = getStringIfConstant(patternNode, scope);
                        const flags = getStringIfConstant(flagsNode, scope);
                        if (typeof pattern === "string") {
                            verify(node, pattern, flags || "");
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/utils/unicode/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        isCombiningCharacter: require(545) /* ./is-combining-character */,
        isEmojiModifier: require(546) /* ./is-emoji-modifier */,
        isRegionalIndicatorSymbol: require(547) /* ./is-regional-indicator-symbol */,
        isSurrogatePair: require(548) /* ./is-surrogate-pair */
    };
},
/* eslint/lib/rules/utils/unicode/is-combining-character.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Check whether a given character is a combining mark or not.
     * @param {number} codePoint The character code to check.
     * @returns {boolean} `true` if the character belongs to the category, any of `Mc`, `Me`, and `Mn`.
     */
    module.exports = function isCombiningCharacter(codePoint) {
        return /^[\p{Mc}\p{Me}\p{Mn}]$/u.test(String.fromCodePoint(codePoint));
    };
},
/* eslint/lib/rules/utils/unicode/is-emoji-modifier.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Check whether a given character is an emoji modifier.
     * @param {number} code The character code to check.
     * @returns {boolean} `true` if the character is an emoji modifier.
     */
    module.exports = function isEmojiModifier(code) {
        return code >= 0x1F3FB && code <= 0x1F3FF;
    };
},
/* eslint/lib/rules/utils/unicode/is-regional-indicator-symbol.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Check whether a given character is a regional indicator symbol.
     * @param {number} code The character code to check.
     * @returns {boolean} `true` if the character is a regional indicator symbol.
     */
    module.exports = function isRegionalIndicatorSymbol(code) {
        return code >= 0x1F1E6 && code <= 0x1F1FF;
    };
},
/* eslint/lib/rules/utils/unicode/is-surrogate-pair.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Check whether given two characters are a surrogate pair.
     * @param {number} lead The code of the lead character.
     * @param {number} tail The code of the tail character.
     * @returns {boolean} `true` if the character pair is a surrogate pair.
     */
    module.exports = function isSurrogatePair(lead, tail) {
        return lead >= 0xD800 && lead < 0xDC00 && tail >= 0xDC00 && tail < 0xE000;
    };
},
/* eslint/lib/rules/no-mixed-operators.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils.js */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const ARITHMETIC_OPERATORS = ["+", "-", "*", "/", "%", "**"];
    const BITWISE_OPERATORS = ["&", "|", "^", "~", "<<", ">>", ">>>"];
    const COMPARISON_OPERATORS = ["==", "!=", "===", "!==", ">", ">=", "<", "<="];
    const LOGICAL_OPERATORS = ["&&", "||"];
    const RELATIONAL_OPERATORS = ["in", "instanceof"];
    const TERNARY_OPERATOR = ["?:"];
    const COALESCE_OPERATOR = ["??"];
    const ALL_OPERATORS = [].concat(ARITHMETIC_OPERATORS, BITWISE_OPERATORS, COMPARISON_OPERATORS, LOGICAL_OPERATORS, RELATIONAL_OPERATORS, TERNARY_OPERATOR, COALESCE_OPERATOR);
    const DEFAULT_GROUPS = [
        ARITHMETIC_OPERATORS,
        BITWISE_OPERATORS,
        COMPARISON_OPERATORS,
        LOGICAL_OPERATORS,
        RELATIONAL_OPERATORS
    ];
    const TARGET_NODE_TYPE = /^(?:Binary|Logical|Conditional)Expression$/u;
    /**
     * Normalizes options.
     * @param {Object|undefined} options A options object to normalize.
     * @returns {Object} Normalized option object.
     */
    function normalizeOptions(options = {}) {
        const hasGroups = options.groups && options.groups.length > 0;
        const groups = hasGroups ? options.groups : DEFAULT_GROUPS;
        const allowSamePrecedence = options.allowSamePrecedence !== false;
        return {
            groups,
            allowSamePrecedence
        };
    }
    /**
     * Checks whether any group which includes both given operator exists or not.
     * @param {Array.<string[]>} groups A list of groups to check.
     * @param {string} left An operator.
     * @param {string} right Another operator.
     * @returns {boolean} `true` if such group existed.
     */
    function includesBothInAGroup(groups, left, right) {
        return groups.some(group => group.indexOf(left) !== -1 && group.indexOf(right) !== -1);
    }
    /**
     * Checks whether the given node is a conditional expression and returns the test node else the left node.
     * @param {ASTNode} node A node which can be a BinaryExpression or a LogicalExpression node.
     * This parent node can be BinaryExpression, LogicalExpression
     *      , or a ConditionalExpression node
     * @returns {ASTNode} node the appropriate node(left or test).
     */
    function getChildNode(node) {
        return node.type === "ConditionalExpression" ? node.test : node.left;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow mixed binary operators",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-mixed-operators"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        groups: {
                            type: "array",
                            items: {
                                type: "array",
                                items: { enum: ALL_OPERATORS },
                                minItems: 2,
                                uniqueItems: true
                            },
                            uniqueItems: true
                        },
                        allowSamePrecedence: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedMixedOperator: "Unexpected mix of '{{leftOperator}}' and '{{rightOperator}}'. Use parentheses to clarify the intended order of operations."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const options = normalizeOptions(context.options[0]);
            /**
             * Checks whether a given node should be ignored by options or not.
             * @param {ASTNode} node A node to check. This is a BinaryExpression
             *      node or a LogicalExpression node. This parent node is one of
             *      them, too.
             * @returns {boolean} `true` if the node should be ignored.
             */
            function shouldIgnore(node) {
                const a = node;
                const b = node.parent;
                return (!includesBothInAGroup(options.groups, a.operator, b.type === "ConditionalExpression" ? "?:" : b.operator) ||
                    (options.allowSamePrecedence &&
                        astUtils.getPrecedence(a) === astUtils.getPrecedence(b)));
            }
            /**
             * Checks whether the operator of a given node is mixed with parent
             * node's operator or not.
             * @param {ASTNode} node A node to check. This is a BinaryExpression
             *      node or a LogicalExpression node. This parent node is one of
             *      them, too.
             * @returns {boolean} `true` if the node was mixed.
             */
            function isMixedWithParent(node) {
                return (node.operator !== node.parent.operator &&
                    !astUtils.isParenthesised(sourceCode, node));
            }
            /**
             * Gets the operator token of a given node.
             * @param {ASTNode} node A node to check. This is a BinaryExpression
             *      node or a LogicalExpression node.
             * @returns {Token} The operator token of the node.
             */
            function getOperatorToken(node) {
                return sourceCode.getTokenAfter(getChildNode(node), astUtils.isNotClosingParenToken);
            }
            /**
             * Reports both the operator of a given node and the operator of the
             * parent node.
             * @param {ASTNode} node A node to check. This is a BinaryExpression
             *      node or a LogicalExpression node. This parent node is one of
             *      them, too.
             * @returns {void}
             */
            function reportBothOperators(node) {
                const parent = node.parent;
                const left = (getChildNode(parent) === node) ? node : parent;
                const right = (getChildNode(parent) !== node) ? node : parent;
                const data = {
                    leftOperator: left.operator || "?:",
                    rightOperator: right.operator || "?:"
                };
                context.report({
                    node: left,
                    loc: getOperatorToken(left).loc,
                    messageId: "unexpectedMixedOperator",
                    data
                });
                context.report({
                    node: right,
                    loc: getOperatorToken(right).loc,
                    messageId: "unexpectedMixedOperator",
                    data
                });
            }
            /**
             * Checks between the operator of this node and the operator of the
             * parent node.
             * @param {ASTNode} node A node to check.
             * @returns {void}
             */
            function check(node) {
                if (TARGET_NODE_TYPE.test(node.parent.type) &&
                    isMixedWithParent(node) &&
                    !shouldIgnore(node)) {
                    reportBothOperators(node);
                }
            }
            return {
                BinaryExpression: check,
                LogicalExpression: check
            };
        }
    };
},
/* eslint/lib/rules/no-mixed-requires.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "disallow `require` calls to be mixed with regular variable declarations",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-mixed-requires"
            },
            schema: [
                {
                    oneOf: [
                        {
                            type: "boolean"
                        },
                        {
                            type: "object",
                            properties: {
                                grouping: {
                                    type: "boolean"
                                },
                                allowCall: {
                                    type: "boolean"
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                noMixRequire: "Do not mix 'require' and other declarations.",
                noMixCoreModuleFileComputed: "Do not mix core, module, file and computed requires."
            }
        },
        create(context) {
            const options = context.options[0];
            let grouping = false, allowCall = false;
            if (typeof options === "object") {
                grouping = options.grouping;
                allowCall = options.allowCall;
            }
            else {
                grouping = !!options;
            }
            /**
             * Returns the list of built-in modules.
             * @returns {string[]} An array of built-in Node.js modules.
             */
            function getBuiltinModules() {
                /*
                 * This list is generated using:
                 * `require("repl")._builtinLibs.concat('repl').sort()`
                 * This particular list is as per nodejs v0.12.2 and iojs v0.7.1
                 */
                return [
                    "assert", "buffer", "child_process", "cluster", "crypto",
                    "dgram", "dns", "domain", "events", "fs", "http", "https",
                    "net", "os", "path", "punycode", "querystring", "readline",
                    "repl", "smalloc", "stream", "string_decoder", "tls", "tty",
                    "url", "util", "v8", "vm", "zlib"
                ];
            }
            const BUILTIN_MODULES = getBuiltinModules();
            const DECL_REQUIRE = "require", DECL_UNINITIALIZED = "uninitialized", DECL_OTHER = "other";
            const REQ_CORE = "core", REQ_FILE = "file", REQ_MODULE = "module", REQ_COMPUTED = "computed";
            /**
             * Determines the type of a declaration statement.
             * @param {ASTNode} initExpression The init node of the VariableDeclarator.
             * @returns {string} The type of declaration represented by the expression.
             */
            function getDeclarationType(initExpression) {
                if (!initExpression) {
                    // "var x;"
                    return DECL_UNINITIALIZED;
                }
                if (initExpression.type === "CallExpression" &&
                    initExpression.callee.type === "Identifier" &&
                    initExpression.callee.name === "require") {
                    // "var x = require('util');"
                    return DECL_REQUIRE;
                }
                if (allowCall &&
                    initExpression.type === "CallExpression" &&
                    initExpression.callee.type === "CallExpression") {
                    // "var x = require('diagnose')('sub-module');"
                    return getDeclarationType(initExpression.callee);
                }
                if (initExpression.type === "MemberExpression") {
                    // "var x = require('glob').Glob;"
                    return getDeclarationType(initExpression.object);
                }
                // "var x = 42;"
                return DECL_OTHER;
            }
            /**
             * Determines the type of module that is loaded via require.
             * @param {ASTNode} initExpression The init node of the VariableDeclarator.
             * @returns {string} The module type.
             */
            function inferModuleType(initExpression) {
                if (initExpression.type === "MemberExpression") {
                    // "var x = require('glob').Glob;"
                    return inferModuleType(initExpression.object);
                }
                if (initExpression.arguments.length === 0) {
                    // "var x = require();"
                    return REQ_COMPUTED;
                }
                const arg = initExpression.arguments[0];
                if (arg.type !== "Literal" || typeof arg.value !== "string") {
                    // "var x = require(42);"
                    return REQ_COMPUTED;
                }
                if (BUILTIN_MODULES.indexOf(arg.value) !== -1) {
                    // "var fs = require('fs');"
                    return REQ_CORE;
                }
                if (/^\.{0,2}\//u.test(arg.value)) {
                    // "var utils = require('./utils');"
                    return REQ_FILE;
                }
                // "var async = require('async');"
                return REQ_MODULE;
            }
            /**
             * Check if the list of variable declarations is mixed, i.e. whether it
             * contains both require and other declarations.
             * @param {ASTNode} declarations The list of VariableDeclarators.
             * @returns {boolean} True if the declarations are mixed, false if not.
             */
            function isMixed(declarations) {
                const contains = {};
                declarations.forEach(declaration => {
                    const type = getDeclarationType(declaration.init);
                    contains[type] = true;
                });
                return !!(contains[DECL_REQUIRE] &&
                    (contains[DECL_UNINITIALIZED] || contains[DECL_OTHER]));
            }
            /**
             * Check if all require declarations in the given list are of the same
             * type.
             * @param {ASTNode} declarations The list of VariableDeclarators.
             * @returns {boolean} True if the declarations are grouped, false if not.
             */
            function isGrouped(declarations) {
                const found = {};
                declarations.forEach(declaration => {
                    if (getDeclarationType(declaration.init) === DECL_REQUIRE) {
                        found[inferModuleType(declaration.init)] = true;
                    }
                });
                return Object.keys(found).length <= 1;
            }
            return {
                VariableDeclaration(node) {
                    if (isMixed(node.declarations)) {
                        context.report({
                            node,
                            messageId: "noMixRequire"
                        });
                    }
                    else if (grouping && !isGrouped(node.declarations)) {
                        context.report({
                            node,
                            messageId: "noMixCoreModuleFileComputed"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-mixed-spaces-and-tabs.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "disallow mixed spaces and tabs for indentation",
                category: "Stylistic Issues",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-mixed-spaces-and-tabs"
            },
            schema: [
                {
                    enum: ["smart-tabs", true, false]
                }
            ],
            messages: {
                mixedSpacesAndTabs: "Mixed spaces and tabs."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            let smartTabs;
            switch (context.options[0]) {
                case true: // Support old syntax, maybe add deprecation warning here
                case "smart-tabs":
                    smartTabs = true;
                    break;
                default:
                    smartTabs = false;
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                "Program:exit"(node) {
                    const lines = sourceCode.lines, comments = sourceCode.getAllComments(), ignoredCommentLines = new Set();
                    // Add all lines except the first ones.
                    comments.forEach(comment => {
                        for (let i = comment.loc.start.line + 1; i <= comment.loc.end.line; i++) {
                            ignoredCommentLines.add(i);
                        }
                    });
                    /*
                     * At least one space followed by a tab
                     * or the reverse before non-tab/-space
                     * characters begin.
                     */
                    let regex = /^(?=( +|\t+))\1(?:\t| )/u;
                    if (smartTabs) {
                        /*
                         * At least one space followed by a tab
                         * before non-tab/-space characters begin.
                         */
                        regex = /^(?=(\t*))\1(?=( +))\2\t/u;
                    }
                    lines.forEach((line, i) => {
                        const match = regex.exec(line);
                        if (match) {
                            const lineNumber = i + 1;
                            const loc = {
                                start: {
                                    line: lineNumber,
                                    column: match[0].length - 2
                                },
                                end: {
                                    line: lineNumber,
                                    column: match[0].length
                                }
                            };
                            if (!ignoredCommentLines.has(lineNumber)) {
                                const containingNode = sourceCode.getNodeByRangeIndex(sourceCode.getIndexFromLoc(loc.start));
                                if (!(containingNode && ["Literal", "TemplateElement"].includes(containingNode.type))) {
                                    context.report({
                                        node,
                                        loc,
                                        messageId: "mixedSpacesAndTabs"
                                    });
                                }
                            }
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-multi-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow use of chained assignment expressions",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-multi-assign"
            },
            schema: [{
                    type: "object",
                    properties: {
                        ignoreNonDeclaration: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                unexpectedChain: "Unexpected chained assignment."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            const options = context.options[0] || {
                ignoreNonDeclaration: false
            };
            const targetParent = options.ignoreNonDeclaration ? ["VariableDeclarator"] : ["AssignmentExpression", "VariableDeclarator"];
            return {
                AssignmentExpression(node) {
                    if (targetParent.indexOf(node.parent.type) !== -1) {
                        context.report({
                            node,
                            messageId: "unexpectedChain"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-multi-spaces.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "disallow multiple spaces",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-multi-spaces"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        exceptions: {
                            type: "object",
                            patternProperties: {
                                "^([A-Z][a-z]*)+$": {
                                    type: "boolean"
                                }
                            },
                            additionalProperties: false
                        },
                        ignoreEOLComments: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                multipleSpaces: "Multiple spaces found before '{{displayValue}}'."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const options = context.options[0] || {};
            const ignoreEOLComments = options.ignoreEOLComments;
            const exceptions = Object.assign({ Property: true }, options.exceptions);
            const hasExceptions = Object.keys(exceptions).filter(key => exceptions[key]).length > 0;
            /**
             * Formats value of given comment token for error message by truncating its length.
             * @param {Token} token comment token
             * @returns {string} formatted value
             * @private
             */
            function formatReportedCommentValue(token) {
                const valueLines = token.value.split("\n");
                const value = valueLines[0];
                const formattedValue = `${value.slice(0, 12)}...`;
                return valueLines.length === 1 && value.length <= 12 ? value : formattedValue;
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program() {
                    sourceCode.tokensAndComments.forEach((leftToken, leftIndex, tokensAndComments) => {
                        if (leftIndex === tokensAndComments.length - 1) {
                            return;
                        }
                        const rightToken = tokensAndComments[leftIndex + 1];
                        // Ignore tokens that don't have 2 spaces between them or are on different lines
                        if (!sourceCode.text.slice(leftToken.range[1], rightToken.range[0]).includes("  ") ||
                            leftToken.loc.end.line < rightToken.loc.start.line) {
                            return;
                        }
                        // Ignore comments that are the last token on their line if `ignoreEOLComments` is active.
                        if (ignoreEOLComments &&
                            astUtils.isCommentToken(rightToken) &&
                            (leftIndex === tokensAndComments.length - 2 ||
                                rightToken.loc.end.line < tokensAndComments[leftIndex + 2].loc.start.line)) {
                            return;
                        }
                        // Ignore tokens that are in a node in the "exceptions" object
                        if (hasExceptions) {
                            const parentNode = sourceCode.getNodeByRangeIndex(rightToken.range[0] - 1);
                            if (parentNode && exceptions[parentNode.type]) {
                                return;
                            }
                        }
                        let displayValue;
                        if (rightToken.type === "Block") {
                            displayValue = `/*${formatReportedCommentValue(rightToken)}*/`;
                        }
                        else if (rightToken.type === "Line") {
                            displayValue = `//${formatReportedCommentValue(rightToken)}`;
                        }
                        else {
                            displayValue = rightToken.value;
                        }
                        context.report({
                            node: rightToken,
                            loc: { start: leftToken.loc.end, end: rightToken.loc.start },
                            messageId: "multipleSpaces",
                            data: { displayValue },
                            fix: fixer => fixer.replaceTextRange([leftToken.range[1], rightToken.range[0]], " ")
                        });
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-multi-str.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow multiline strings",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-multi-str"
            },
            schema: [],
            messages: {
                multilineString: "Multiline support is limited to browsers supporting ES5 only."
            }
        },
        create(context) {
            /**
             * Determines if a given node is part of JSX syntax.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} True if the node is a JSX node, false if not.
             * @private
             */
            function isJSXElement(node) {
                return node.type.indexOf("JSX") === 0;
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                Literal(node) {
                    if (astUtils.LINEBREAK_MATCHER.test(node.raw) && !isJSXElement(node.parent)) {
                        context.report({
                            node,
                            messageId: "multilineString"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-multiple-empty-lines.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "disallow multiple empty lines",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-multiple-empty-lines"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        max: {
                            type: "integer",
                            minimum: 0
                        },
                        maxEOF: {
                            type: "integer",
                            minimum: 0
                        },
                        maxBOF: {
                            type: "integer",
                            minimum: 0
                        }
                    },
                    required: ["max"],
                    additionalProperties: false
                }
            ],
            messages: {
                blankBeginningOfFile: "Too many blank lines at the beginning of file. Max of {{max}} allowed.",
                blankEndOfFile: "Too many blank lines at the end of file. Max of {{max}} allowed.",
                consecutiveBlank: "More than {{max}} blank {{pluralizedLines}} not allowed."
            }
        },
        create(context) {
            // Use options.max or 2 as default
            let max = 2, maxEOF = max, maxBOF = max;
            if (context.options.length) {
                max = context.options[0].max;
                maxEOF = typeof context.options[0].maxEOF !== "undefined" ? context.options[0].maxEOF : max;
                maxBOF = typeof context.options[0].maxBOF !== "undefined" ? context.options[0].maxBOF : max;
            }
            const sourceCode = context.getSourceCode();
            // Swallow the final newline, as some editors add it automatically and we don't want it to cause an issue
            const allLines = sourceCode.lines[sourceCode.lines.length - 1] === "" ? sourceCode.lines.slice(0, -1) : sourceCode.lines;
            const templateLiteralLines = new Set();
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                TemplateLiteral(node) {
                    node.quasis.forEach(literalPart => {
                        // Empty lines have a semantic meaning if they're inside template literals. Don't count these as empty lines.
                        for (let ignoredLine = literalPart.loc.start.line; ignoredLine < literalPart.loc.end.line; ignoredLine++) {
                            templateLiteralLines.add(ignoredLine);
                        }
                    });
                },
                "Program:exit"(node) {
                    return allLines
                        // Given a list of lines, first get a list of line numbers that are non-empty.
                        .reduce((nonEmptyLineNumbers, line, index) => {
                        if (line.trim() || templateLiteralLines.has(index + 1)) {
                            nonEmptyLineNumbers.push(index + 1);
                        }
                        return nonEmptyLineNumbers;
                    }, [])
                        // Add a value at the end to allow trailing empty lines to be checked.
                        .concat(allLines.length + 1)
                        // Given two line numbers of non-empty lines, report the lines between if the difference is too large.
                        .reduce((lastLineNumber, lineNumber) => {
                        let messageId, maxAllowed;
                        if (lastLineNumber === 0) {
                            messageId = "blankBeginningOfFile";
                            maxAllowed = maxBOF;
                        }
                        else if (lineNumber === allLines.length + 1) {
                            messageId = "blankEndOfFile";
                            maxAllowed = maxEOF;
                        }
                        else {
                            messageId = "consecutiveBlank";
                            maxAllowed = max;
                        }
                        if (lineNumber - lastLineNumber - 1 > maxAllowed) {
                            context.report({
                                node,
                                loc: {
                                    start: { line: lastLineNumber + maxAllowed + 1, column: 0 },
                                    end: { line: lineNumber, column: 0 }
                                },
                                messageId,
                                data: {
                                    max: maxAllowed,
                                    pluralizedLines: maxAllowed === 1 ? "line" : "lines"
                                },
                                fix(fixer) {
                                    const rangeStart = sourceCode.getIndexFromLoc({ line: lastLineNumber + 1, column: 0 });
                                    /*
                                     * The end of the removal range is usually the start index of the next line.
                                     * However, at the end of the file there is no next line, so the end of the
                                     * range is just the length of the text.
                                     */
                                    const lineNumberAfterRemovedLines = lineNumber - maxAllowed;
                                    const rangeEnd = lineNumberAfterRemovedLines <= allLines.length
                                        ? sourceCode.getIndexFromLoc({ line: lineNumberAfterRemovedLines, column: 0 })
                                        : sourceCode.text.length;
                                    return fixer.removeRange([rangeStart, rangeEnd]);
                                }
                            });
                        }
                        return lineNumber;
                    }, 0);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-native-reassign.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow assignments to native objects or read-only global variables",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-native-reassign"
            },
            deprecated: true,
            replacedBy: ["no-global-assign"],
            schema: [
                {
                    type: "object",
                    properties: {
                        exceptions: {
                            type: "array",
                            items: { type: "string" },
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                nativeReassign: "Read-only global '{{name}}' should not be modified."
            }
        },
        create(context) {
            const config = context.options[0];
            const exceptions = (config && config.exceptions) || [];
            /**
             * Reports write references.
             * @param {Reference} reference A reference to check.
             * @param {int} index The index of the reference in the references.
             * @param {Reference[]} references The array that the reference belongs to.
             * @returns {void}
             */
            function checkReference(reference, index, references) {
                const identifier = reference.identifier;
                if (reference.init === false &&
                    reference.isWrite() &&
                    /*
                     * Destructuring assignments can have multiple default value,
                     * so possibly there are multiple writeable references for the same identifier.
                     */
                    (index === 0 || references[index - 1].identifier !== identifier)) {
                    context.report({
                        node: identifier,
                        messageId: "nativeReassign",
                        data: identifier
                    });
                }
            }
            /**
             * Reports write references if a given variable is read-only builtin.
             * @param {Variable} variable A variable to check.
             * @returns {void}
             */
            function checkVariable(variable) {
                if (variable.writeable === false && exceptions.indexOf(variable.name) === -1) {
                    variable.references.forEach(checkReference);
                }
            }
            return {
                Program() {
                    const globalScope = context.getScope();
                    globalScope.variables.forEach(checkVariable);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-negated-condition.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow negated conditions",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-negated-condition"
            },
            schema: [],
            messages: {
                unexpectedNegated: "Unexpected negated condition."
            }
        },
        create(context) {
            /**
             * Determines if a given node is an if-else without a condition on the else
             * @param {ASTNode} node The node to check.
             * @returns {boolean} True if the node has an else without an if.
             * @private
             */
            function hasElseWithoutCondition(node) {
                return node.alternate && node.alternate.type !== "IfStatement";
            }
            /**
             * Determines if a given node is a negated unary expression
             * @param {Object} test The test object to check.
             * @returns {boolean} True if the node is a negated unary expression.
             * @private
             */
            function isNegatedUnaryExpression(test) {
                return test.type === "UnaryExpression" && test.operator === "!";
            }
            /**
             * Determines if a given node is a negated binary expression
             * @param {Test} test The test to check.
             * @returns {boolean} True if the node is a negated binary expression.
             * @private
             */
            function isNegatedBinaryExpression(test) {
                return test.type === "BinaryExpression" &&
                    (test.operator === "!=" || test.operator === "!==");
            }
            /**
             * Determines if a given node has a negated if expression
             * @param {ASTNode} node The node to check.
             * @returns {boolean} True if the node has a negated if expression.
             * @private
             */
            function isNegatedIf(node) {
                return isNegatedUnaryExpression(node.test) || isNegatedBinaryExpression(node.test);
            }
            return {
                IfStatement(node) {
                    if (!hasElseWithoutCondition(node)) {
                        return;
                    }
                    if (isNegatedIf(node)) {
                        context.report({
                            node,
                            messageId: "unexpectedNegated"
                        });
                    }
                },
                ConditionalExpression(node) {
                    if (isNegatedIf(node)) {
                        context.report({
                            node,
                            messageId: "unexpectedNegated"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-negated-in-lhs.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow negating the left operand in `in` expressions",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-negated-in-lhs"
            },
            replacedBy: ["no-unsafe-negation"],
            deprecated: true,
            schema: [],
            messages: {
                negatedLHS: "The 'in' expression's left operand is negated."
            }
        },
        create(context) {
            return {
                BinaryExpression(node) {
                    if (node.operator === "in" && node.left.type === "UnaryExpression" && node.left.operator === "!") {
                        context.report({ node, messageId: "negatedLHS" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-nested-ternary.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow nested ternary expressions",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-nested-ternary"
            },
            schema: [],
            messages: {
                noNestedTernary: "Do not nest ternary expressions."
            }
        },
        create(context) {
            return {
                ConditionalExpression(node) {
                    if (node.alternate.type === "ConditionalExpression" ||
                        node.consequent.type === "ConditionalExpression") {
                        context.report({
                            node,
                            messageId: "noNestedTernary"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-new.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `new` operators outside of assignments or comparisons",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-new"
            },
            schema: [],
            messages: {
                noNewStatement: "Do not use 'new' for side effects."
            }
        },
        create(context) {
            return {
                "ExpressionStatement > NewExpression"(node) {
                    context.report({
                        node: node.parent,
                        messageId: "noNewStatement"
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-new-func.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `new` operators with the `Function` object",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-new-func"
            },
            schema: [],
            messages: {
                noFunctionConstructor: "The Function constructor is eval."
            }
        },
        create(context) {
            return {
                "Program:exit"() {
                    const globalScope = context.getScope();
                    const variable = globalScope.set.get("Function");
                    if (variable && variable.defs.length === 0) {
                        variable.references.forEach(ref => {
                            const node = ref.identifier;
                            const { parent } = node;
                            if (parent &&
                                (parent.type === "NewExpression" || parent.type === "CallExpression") &&
                                node === parent.callee) {
                                context.report({
                                    node: parent,
                                    messageId: "noFunctionConstructor"
                                });
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-new-object.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `Object` constructors",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-new-object"
            },
            schema: [],
            messages: {
                preferLiteral: "The object literal notation {} is preferrable."
            }
        },
        create(context) {
            return {
                NewExpression(node) {
                    const variable = astUtils.getVariableByName(context.getScope(), node.callee.name);
                    if (variable && variable.identifiers.length > 0) {
                        return;
                    }
                    if (node.callee.name === "Object") {
                        context.report({
                            node,
                            messageId: "preferLiteral"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-new-require.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "disallow `new` operators with calls to `require`",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-new-require"
            },
            schema: [],
            messages: {
                noNewRequire: "Unexpected use of new with require."
            }
        },
        create(context) {
            return {
                NewExpression(node) {
                    if (node.callee.type === "Identifier" && node.callee.name === "require") {
                        context.report({
                            node,
                            messageId: "noNewRequire"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-new-symbol.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow `new` operators with the `Symbol` object",
                category: "ECMAScript 6",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-new-symbol"
            },
            schema: [],
            messages: {
                noNewSymbol: "`Symbol` cannot be called as a constructor."
            }
        },
        create(context) {
            return {
                "Program:exit"() {
                    const globalScope = context.getScope();
                    const variable = globalScope.set.get("Symbol");
                    if (variable && variable.defs.length === 0) {
                        variable.references.forEach(ref => {
                            const node = ref.identifier;
                            const parent = node.parent;
                            if (parent && parent.type === "NewExpression" && parent.callee === node) {
                                context.report({
                                    node,
                                    messageId: "noNewSymbol"
                                });
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-new-wrappers.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `new` operators with the `String`, `Number`, and `Boolean` objects",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-new-wrappers"
            },
            schema: [],
            messages: {
                noConstructor: "Do not use {{fn}} as a constructor."
            }
        },
        create(context) {
            return {
                NewExpression(node) {
                    const wrapperObjects = ["String", "Number", "Boolean"];
                    if (wrapperObjects.indexOf(node.callee.name) > -1) {
                        context.report({
                            node,
                            messageId: "noConstructor",
                            data: { fn: node.callee.name }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-nonoctal-decimal-escape.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const QUICK_TEST_REGEX = /\\[89]/u;
    /**
     * Returns unicode escape sequence that represents the given character.
     * @param {string} character A single code unit.
     * @returns {string} "\uXXXX" sequence.
     */
    function getUnicodeEscape(character) {
        return `\\u${character.charCodeAt(0).toString(16).padStart(4, "0")}`;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `\\8` and `\\9` escape sequences in string literals",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-nonoctal-decimal-escape",
                suggestion: true
            },
            schema: [],
            messages: {
                decimalEscape: "Don't use '{{decimalEscape}}' escape sequence.",
                // suggestions
                refactor: "Replace '{{original}}' with '{{replacement}}'. This maintains the current functionality.",
                escapeBackslash: "Replace '{{original}}' with '{{replacement}}' to include the actual backslash character."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Creates a new Suggestion object.
             * @param {string} messageId "refactor" or "escapeBackslash".
             * @param {int[]} range The range to replace.
             * @param {string} replacement New text for the range.
             * @returns {Object} Suggestion
             */
            function createSuggestion(messageId, range, replacement) {
                return {
                    messageId,
                    data: {
                        original: sourceCode.getText().slice(...range),
                        replacement
                    },
                    fix(fixer) {
                        return fixer.replaceTextRange(range, replacement);
                    }
                };
            }
            return {
                Literal(node) {
                    if (typeof node.value !== "string") {
                        return;
                    }
                    if (!QUICK_TEST_REGEX.test(node.raw)) {
                        return;
                    }
                    const regex = /(?:[^\\]|(?<previousEscape>\\.))*?(?<decimalEscape>\\[89])/suy;
                    let match;
                    while ((match = regex.exec(node.raw))) {
                        const { previousEscape, decimalEscape } = match.groups;
                        const decimalEscapeRangeEnd = node.range[0] + match.index + match[0].length;
                        const decimalEscapeRangeStart = decimalEscapeRangeEnd - decimalEscape.length;
                        const decimalEscapeRange = [decimalEscapeRangeStart, decimalEscapeRangeEnd];
                        const suggest = [];
                        // When `regex` is matched, `previousEscape` can only capture characters adjacent to `decimalEscape`
                        if (previousEscape === "\\0") {
                            /*
                             * Now we have a NULL escape "\0" immediately followed by a decimal escape, e.g.: "\0\8".
                             * Fixing this to "\08" would turn "\0" into a legacy octal escape. To avoid producing
                             * an octal escape while fixing a decimal escape, we provide different suggestions.
                             */
                            suggest.push(createSuggestion(// "\0\8" -> "\u00008"
                            "refactor", [decimalEscapeRangeStart - previousEscape.length, decimalEscapeRangeEnd], `${getUnicodeEscape("\0")}${decimalEscape[1]}`), createSuggestion(// "\8" -> "\u0038"
                            "refactor", decimalEscapeRange, getUnicodeEscape(decimalEscape[1])));
                        }
                        else {
                            suggest.push(createSuggestion(// "\8" -> "8"
                            "refactor", decimalEscapeRange, decimalEscape[1]));
                        }
                        suggest.push(createSuggestion(// "\8" -> "\\8"
                        "escapeBackslash", decimalEscapeRange, `\\${decimalEscape}`));
                        context.report({
                            node,
                            loc: {
                                start: sourceCode.getLocFromIndex(decimalEscapeRangeStart),
                                end: sourceCode.getLocFromIndex(decimalEscapeRangeEnd)
                            },
                            messageId: "decimalEscape",
                            data: {
                                decimalEscape
                            },
                            suggest
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-obj-calls.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { CALL, CONSTRUCT, ReferenceTracker } = require(350) /* eslint-utils */;
    const getPropertyName = require(394) /* ./utils/ast-utils */.getStaticPropertyName;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const nonCallableGlobals = ["Atomics", "JSON", "Math", "Reflect"];
    /**
     * Returns the name of the node to report
     * @param {ASTNode} node A node to report
     * @returns {string} name to report
     */
    function getReportNodeName(node) {
        if (node.type === "ChainExpression") {
            return getReportNodeName(node.expression);
        }
        if (node.type === "MemberExpression") {
            return getPropertyName(node);
        }
        return node.name;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow calling global object properties as functions",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-obj-calls"
            },
            schema: [],
            messages: {
                unexpectedCall: "'{{name}}' is not a function.",
                unexpectedRefCall: "'{{name}}' is reference to '{{ref}}', which is not a function."
            }
        },
        create(context) {
            return {
                Program() {
                    const scope = context.getScope();
                    const tracker = new ReferenceTracker(scope);
                    const traceMap = {};
                    for (const g of nonCallableGlobals) {
                        traceMap[g] = {
                            [CALL]: true,
                            [CONSTRUCT]: true
                        };
                    }
                    for (const { node, path } of tracker.iterateGlobalReferences(traceMap)) {
                        const name = getReportNodeName(node.callee);
                        const ref = path[0];
                        const messageId = name === ref ? "unexpectedCall" : "unexpectedRefCall";
                        context.report({ node, messageId, data: { name, ref } });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-octal.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow octal literals",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-octal"
            },
            schema: [],
            messages: {
                noOcatal: "Octal literals should not be used."
            }
        },
        create(context) {
            return {
                Literal(node) {
                    if (typeof node.value === "number" && /^0[0-9]/u.test(node.raw)) {
                        context.report({
                            node,
                            messageId: "noOcatal"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-octal-escape.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow octal escape sequences in string literals",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-octal-escape"
            },
            schema: [],
            messages: {
                octalEscapeSequence: "Don't use octal: '\\{{sequence}}'. Use '\\u....' instead."
            }
        },
        create(context) {
            return {
                Literal(node) {
                    if (typeof node.value !== "string") {
                        return;
                    }
                    // \0 represents a valid NULL character if it isn't followed by a digit.
                    const match = node.raw.match(/^(?:[^\\]|\\.)*?\\([0-3][0-7]{1,2}|[4-7][0-7]|0(?=[89])|[1-7])/su);
                    if (match) {
                        context.report({
                            node,
                            messageId: "octalEscapeSequence",
                            data: { sequence: match[1] }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-param-reassign.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const stopNodePattern = /(?:Statement|Declaration|Function(?:Expression)?|Program)$/u;
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow reassigning `function` parameters",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-param-reassign"
            },
            schema: [
                {
                    oneOf: [
                        {
                            type: "object",
                            properties: {
                                props: {
                                    enum: [false]
                                }
                            },
                            additionalProperties: false
                        },
                        {
                            type: "object",
                            properties: {
                                props: {
                                    enum: [true]
                                },
                                ignorePropertyModificationsFor: {
                                    type: "array",
                                    items: {
                                        type: "string"
                                    },
                                    uniqueItems: true
                                },
                                ignorePropertyModificationsForRegex: {
                                    type: "array",
                                    items: {
                                        type: "string"
                                    },
                                    uniqueItems: true
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                assignmentToFunctionParam: "Assignment to function parameter '{{name}}'.",
                assignmentToFunctionParamProp: "Assignment to property of function parameter '{{name}}'."
            }
        },
        create(context) {
            const props = context.options[0] && context.options[0].props;
            const ignoredPropertyAssignmentsFor = context.options[0] && context.options[0].ignorePropertyModificationsFor || [];
            const ignoredPropertyAssignmentsForRegex = context.options[0] && context.options[0].ignorePropertyModificationsForRegex || [];
            /**
             * Checks whether or not the reference modifies properties of its variable.
             * @param {Reference} reference A reference to check.
             * @returns {boolean} Whether or not the reference modifies properties of its variable.
             */
            function isModifyingProp(reference) {
                let node = reference.identifier;
                let parent = node.parent;
                while (parent && (!stopNodePattern.test(parent.type) ||
                    parent.type === "ForInStatement" || parent.type === "ForOfStatement")) {
                    switch (parent.type) {
                        // e.g. foo.a = 0;
                        case "AssignmentExpression":
                            return parent.left === node;
                        // e.g. ++foo.a;
                        case "UpdateExpression":
                            return true;
                        // e.g. delete foo.a;
                        case "UnaryExpression":
                            if (parent.operator === "delete") {
                                return true;
                            }
                            break;
                        // e.g. for (foo.a in b) {}
                        case "ForInStatement":
                        case "ForOfStatement":
                            if (parent.left === node) {
                                return true;
                            }
                            // this is a stop node for parent.right and parent.body
                            return false;
                        // EXCLUDES: e.g. cache.get(foo.a).b = 0;
                        case "CallExpression":
                            if (parent.callee !== node) {
                                return false;
                            }
                            break;
                        // EXCLUDES: e.g. cache[foo.a] = 0;
                        case "MemberExpression":
                            if (parent.property === node) {
                                return false;
                            }
                            break;
                        // EXCLUDES: e.g. ({ [foo]: a }) = bar;
                        case "Property":
                            if (parent.key === node) {
                                return false;
                            }
                            break;
                        // EXCLUDES: e.g. (foo ? a : b).c = bar;
                        case "ConditionalExpression":
                            if (parent.test === node) {
                                return false;
                            }
                            break;
                        // no default
                    }
                    node = parent;
                    parent = node.parent;
                }
                return false;
            }
            /**
             * Tests that an identifier name matches any of the ignored property assignments.
             * First we test strings in ignoredPropertyAssignmentsFor.
             * Then we instantiate and test RegExp objects from ignoredPropertyAssignmentsForRegex strings.
             * @param {string} identifierName A string that describes the name of an identifier to
             * ignore property assignments for.
             * @returns {boolean} Whether the string matches an ignored property assignment regular expression or not.
             */
            function isIgnoredPropertyAssignment(identifierName) {
                return ignoredPropertyAssignmentsFor.includes(identifierName) ||
                    ignoredPropertyAssignmentsForRegex.some(ignored => new RegExp(ignored, "u").test(identifierName));
            }
            /**
             * Reports a reference if is non initializer and writable.
             * @param {Reference} reference A reference to check.
             * @param {int} index The index of the reference in the references.
             * @param {Reference[]} references The array that the reference belongs to.
             * @returns {void}
             */
            function checkReference(reference, index, references) {
                const identifier = reference.identifier;
                if (identifier &&
                    !reference.init &&
                    /*
                     * Destructuring assignments can have multiple default value,
                     * so possibly there are multiple writeable references for the same identifier.
                     */
                    (index === 0 || references[index - 1].identifier !== identifier)) {
                    if (reference.isWrite()) {
                        context.report({
                            node: identifier,
                            messageId: "assignmentToFunctionParam",
                            data: { name: identifier.name }
                        });
                    }
                    else if (props && isModifyingProp(reference) && !isIgnoredPropertyAssignment(identifier.name)) {
                        context.report({
                            node: identifier,
                            messageId: "assignmentToFunctionParamProp",
                            data: { name: identifier.name }
                        });
                    }
                }
            }
            /**
             * Finds and reports references that are non initializer and writable.
             * @param {Variable} variable A variable to check.
             * @returns {void}
             */
            function checkVariable(variable) {
                if (variable.defs[0].type === "Parameter") {
                    variable.references.forEach(checkReference);
                }
            }
            /**
             * Checks parameters of a given function node.
             * @param {ASTNode} node A function node to check.
             * @returns {void}
             */
            function checkForFunction(node) {
                context.getDeclaredVariables(node).forEach(checkVariable);
            }
            return {
                // `:exit` is needed for the `node.parent` property of identifier nodes.
                "FunctionDeclaration:exit": checkForFunction,
                "FunctionExpression:exit": checkForFunction,
                "ArrowFunctionExpression:exit": checkForFunction
            };
        }
    };
},
/* eslint/lib/rules/no-path-concat.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "disallow string concatenation with `__dirname` and `__filename`",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-path-concat"
            },
            schema: [],
            messages: {
                usePathFunctions: "Use path.join() or path.resolve() instead of + to create paths."
            }
        },
        create(context) {
            const MATCHER = /^__(?:dir|file)name$/u;
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                BinaryExpression(node) {
                    const left = node.left, right = node.right;
                    if (node.operator === "+" &&
                        ((left.type === "Identifier" && MATCHER.test(left.name)) ||
                            (right.type === "Identifier" && MATCHER.test(right.name)))) {
                        context.report({
                            node,
                            messageId: "usePathFunctions"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-plusplus.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Determines whether the given node is the update node of a `ForStatement`.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is `ForStatement` update.
     */
    function isForStatementUpdate(node) {
        const parent = node.parent;
        return parent.type === "ForStatement" && parent.update === node;
    }
    /**
     * Determines whether the given node is considered to be a for loop "afterthought" by the logic of this rule.
     * In particular, it returns `true` if the given node is either:
     *   - The update node of a `ForStatement`: for (;; i++) {}
     *   - An operand of a sequence expression that is the update node: for (;; foo(), i++) {}
     *   - An operand of a sequence expression that is child of another sequence expression, etc.,
     *     up to the sequence expression that is the update node: for (;; foo(), (bar(), (baz(), i++))) {}
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is a for loop afterthought.
     */
    function isForLoopAfterthought(node) {
        const parent = node.parent;
        if (parent.type === "SequenceExpression") {
            return isForLoopAfterthought(parent);
        }
        return isForStatementUpdate(node);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the unary operators `++` and `--`",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-plusplus"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowForLoopAfterthoughts: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedUnaryOp: "Unary operator '{{operator}}' used."
            }
        },
        create(context) {
            const config = context.options[0];
            let allowForLoopAfterthoughts = false;
            if (typeof config === "object") {
                allowForLoopAfterthoughts = config.allowForLoopAfterthoughts === true;
            }
            return {
                UpdateExpression(node) {
                    if (allowForLoopAfterthoughts && isForLoopAfterthought(node)) {
                        return;
                    }
                    context.report({
                        node,
                        messageId: "unexpectedUnaryOp",
                        data: {
                            operator: node.operator
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-process-env.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "disallow the use of `process.env`",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-process-env"
            },
            schema: [],
            messages: {
                unexpectedProcessEnv: "Unexpected use of process.env."
            }
        },
        create(context) {
            return {
                MemberExpression(node) {
                    const objectName = node.object.name, propertyName = node.property.name;
                    if (objectName === "process" && !node.computed && propertyName && propertyName === "env") {
                        context.report({ node, messageId: "unexpectedProcessEnv" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-process-exit.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "disallow the use of `process.exit()`",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-process-exit"
            },
            schema: [],
            messages: {
                noProcessExit: "Don't use process.exit(); throw an error instead."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                "CallExpression > MemberExpression.callee[object.name = 'process'][property.name = 'exit']"(node) {
                    context.report({ node: node.parent, messageId: "noProcessExit" });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-promise-executor-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { findVariable } = require(350) /* eslint-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const functionTypesToCheck = new Set(["ArrowFunctionExpression", "FunctionExpression"]);
    /**
     * Determines whether the given identifier node is a reference to a global variable.
     * @param {ASTNode} node `Identifier` node to check.
     * @param {Scope} scope Scope to which the node belongs.
     * @returns {boolean} True if the identifier is a reference to a global variable.
     */
    function isGlobalReference(node, scope) {
        const variable = findVariable(scope, node);
        return variable !== null && variable.scope.type === "global" && variable.defs.length === 0;
    }
    /**
     * Finds function's outer scope.
     * @param {Scope} scope Function's own scope.
     * @returns {Scope} Function's outer scope.
     */
    function getOuterScope(scope) {
        const upper = scope.upper;
        if (upper.type === "function-expression-name") {
            return upper.upper;
        }
        return upper;
    }
    /**
     * Determines whether the given function node is used as a Promise executor.
     * @param {ASTNode} node The node to check.
     * @param {Scope} scope Function's own scope.
     * @returns {boolean} `true` if the node is a Promise executor.
     */
    function isPromiseExecutor(node, scope) {
        const parent = node.parent;
        return parent.type === "NewExpression" &&
            parent.arguments[0] === node &&
            parent.callee.type === "Identifier" &&
            parent.callee.name === "Promise" &&
            isGlobalReference(parent.callee, getOuterScope(scope));
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow returning values from Promise executor functions",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-promise-executor-return"
            },
            schema: [],
            messages: {
                returnsValue: "Return values from promise executor functions cannot be read."
            }
        },
        create(context) {
            let funcInfo = null;
            /**
             * Reports the given node.
             * @param {ASTNode} node Node to report.
             * @returns {void}
             */
            function report(node) {
                context.report({ node, messageId: "returnsValue" });
            }
            return {
                onCodePathStart(_, node) {
                    funcInfo = {
                        upper: funcInfo,
                        shouldCheck: functionTypesToCheck.has(node.type) && isPromiseExecutor(node, context.getScope())
                    };
                    if (funcInfo.shouldCheck && node.type === "ArrowFunctionExpression" && node.expression) {
                        report(node.body);
                    }
                },
                onCodePathEnd() {
                    funcInfo = funcInfo.upper;
                },
                ReturnStatement(node) {
                    if (funcInfo.shouldCheck && node.argument) {
                        report(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-proto.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { getStaticPropertyName } = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the use of the `__proto__` property",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-proto"
            },
            schema: [],
            messages: {
                unexpectedProto: "The '__proto__' property is deprecated."
            }
        },
        create(context) {
            return {
                MemberExpression(node) {
                    if (getStaticPropertyName(node) === "__proto__") {
                        context.report({ node, messageId: "unexpectedProto" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-prototype-builtins.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow calling some `Object.prototype` methods directly on objects",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-prototype-builtins"
            },
            schema: [],
            messages: {
                prototypeBuildIn: "Do not access Object.prototype method '{{prop}}' from target object."
            }
        },
        create(context) {
            const DISALLOWED_PROPS = [
                "hasOwnProperty",
                "isPrototypeOf",
                "propertyIsEnumerable"
            ];
            /**
             * Reports if a disallowed property is used in a CallExpression
             * @param {ASTNode} node The CallExpression node.
             * @returns {void}
             */
            function disallowBuiltIns(node) {
                const callee = astUtils.skipChainExpression(node.callee);
                if (callee.type !== "MemberExpression") {
                    return;
                }
                const propName = astUtils.getStaticPropertyName(callee);
                if (propName !== null && DISALLOWED_PROPS.indexOf(propName) > -1) {
                    context.report({
                        messageId: "prototypeBuildIn",
                        loc: callee.property.loc,
                        data: { prop: propName },
                        node
                    });
                }
            }
            return {
                CallExpression: disallowBuiltIns
            };
        }
    };
},
/* eslint/lib/rules/no-redeclare.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow variable redeclaration",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-redeclare"
            },
            messages: {
                redeclared: "'{{id}}' is already defined.",
                redeclaredAsBuiltin: "'{{id}}' is already defined as a built-in global variable.",
                redeclaredBySyntax: "'{{id}}' is already defined by a variable declaration."
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        builtinGlobals: { type: "boolean", default: true }
                    },
                    additionalProperties: false
                }
            ]
        },
        create(context) {
            const options = {
                builtinGlobals: Boolean(context.options.length === 0 ||
                    context.options[0].builtinGlobals)
            };
            const sourceCode = context.getSourceCode();
            /**
             * Iterate declarations of a given variable.
             * @param {escope.variable} variable The variable object to iterate declarations.
             * @returns {IterableIterator<{type:string,node:ASTNode,loc:SourceLocation}>} The declarations.
             */
            function* iterateDeclarations(variable) {
                if (options.builtinGlobals && (variable.eslintImplicitGlobalSetting === "readonly" ||
                    variable.eslintImplicitGlobalSetting === "writable")) {
                    yield { type: "builtin" };
                }
                for (const id of variable.identifiers) {
                    yield { type: "syntax", node: id, loc: id.loc };
                }
                if (variable.eslintExplicitGlobalComments) {
                    for (const comment of variable.eslintExplicitGlobalComments) {
                        yield {
                            type: "comment",
                            node: comment,
                            loc: astUtils.getNameLocationInGlobalDirectiveComment(sourceCode, comment, variable.name)
                        };
                    }
                }
            }
            /**
             * Find variables in a given scope and flag redeclared ones.
             * @param {Scope} scope An eslint-scope scope object.
             * @returns {void}
             * @private
             */
            function findVariablesInScope(scope) {
                for (const variable of scope.variables) {
                    const [declaration, ...extraDeclarations] = iterateDeclarations(variable);
                    if (extraDeclarations.length === 0) {
                        continue;
                    }
                    /*
                     * If the type of a declaration is different from the type of
                     * the first declaration, it shows the location of the first
                     * declaration.
                     */
                    const detailMessageId = declaration.type === "builtin"
                        ? "redeclaredAsBuiltin"
                        : "redeclaredBySyntax";
                    const data = { id: variable.name };
                    // Report extra declarations.
                    for (const { type, node, loc } of extraDeclarations) {
                        const messageId = type === declaration.type
                            ? "redeclared"
                            : detailMessageId;
                        context.report({ node, loc, messageId, data });
                    }
                }
            }
            /**
             * Find variables in the current scope.
             * @param {ASTNode} node The node of the current scope.
             * @returns {void}
             * @private
             */
            function checkForBlock(node) {
                const scope = context.getScope();
                /*
                 * In ES5, some node type such as `BlockStatement` doesn't have that scope.
                 * `scope.block` is a different node in such a case.
                 */
                if (scope.block === node) {
                    findVariablesInScope(scope);
                }
            }
            return {
                Program() {
                    const scope = context.getScope();
                    findVariablesInScope(scope);
                    // Node.js or ES modules has a special scope.
                    if (scope.type === "global" &&
                        scope.childScopes[0] &&
                        // The special scope's block is the Program node.
                        scope.block === scope.childScopes[0].block) {
                        findVariablesInScope(scope.childScopes[0]);
                    }
                },
                FunctionDeclaration: checkForBlock,
                FunctionExpression: checkForBlock,
                ArrowFunctionExpression: checkForBlock,
                BlockStatement: checkForBlock,
                ForStatement: checkForBlock,
                ForInStatement: checkForBlock,
                ForOfStatement: checkForBlock,
                SwitchStatement: checkForBlock
            };
        }
    };
},
/* eslint/lib/rules/no-regex-spaces.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const regexpp = require(497) /* regexpp */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const regExpParser = new regexpp.RegExpParser();
    const DOUBLE_SPACE = / {2}/u;
    /**
     * Check if node is a string
     * @param {ASTNode} node node to evaluate
     * @returns {boolean} True if its a string
     * @private
     */
    function isString(node) {
        return node && node.type === "Literal" && typeof node.value === "string";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow multiple spaces in regular expressions",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-regex-spaces"
            },
            schema: [],
            fixable: "code",
            messages: {
                multipleSpaces: "Spaces are hard to count. Use {{{length}}}."
            }
        },
        create(context) {
            /**
             * Validate regular expression
             * @param {ASTNode} nodeToReport Node to report.
             * @param {string} pattern Regular expression pattern to validate.
             * @param {string} rawPattern Raw representation of the pattern in the source code.
             * @param {number} rawPatternStartRange Start range of the pattern in the source code.
             * @param {string} flags Regular expression flags.
             * @returns {void}
             * @private
             */
            function checkRegex(nodeToReport, pattern, rawPattern, rawPatternStartRange, flags) {
                // Skip if there are no consecutive spaces in the source code, to avoid reporting e.g., RegExp(' \ ').
                if (!DOUBLE_SPACE.test(rawPattern)) {
                    return;
                }
                const characterClassNodes = [];
                let regExpAST;
                try {
                    regExpAST = regExpParser.parsePattern(pattern, 0, pattern.length, flags.includes("u"));
                }
                catch {
                    // Ignore regular expressions with syntax errors
                    return;
                }
                regexpp.visitRegExpAST(regExpAST, {
                    onCharacterClassEnter(ccNode) {
                        characterClassNodes.push(ccNode);
                    }
                });
                const spacesPattern = /( {2,})(?: [+*{?]|[^+*{?]|$)/gu;
                let match;
                while ((match = spacesPattern.exec(pattern))) {
                    const { 1: { length }, index } = match;
                    // Report only consecutive spaces that are not in character classes.
                    if (characterClassNodes.every(({ start, end }) => index < start || end <= index)) {
                        context.report({
                            node: nodeToReport,
                            messageId: "multipleSpaces",
                            data: { length },
                            fix(fixer) {
                                if (pattern !== rawPattern) {
                                    return null;
                                }
                                return fixer.replaceTextRange([rawPatternStartRange + index, rawPatternStartRange + index + length], ` {${length}}`);
                            }
                        });
                        // Report only the first occurrence of consecutive spaces
                        return;
                    }
                }
            }
            /**
             * Validate regular expression literals
             * @param {ASTNode} node node to validate
             * @returns {void}
             * @private
             */
            function checkLiteral(node) {
                if (node.regex) {
                    const pattern = node.regex.pattern;
                    const rawPattern = node.raw.slice(1, node.raw.lastIndexOf("/"));
                    const rawPatternStartRange = node.range[0] + 1;
                    const flags = node.regex.flags;
                    checkRegex(node, pattern, rawPattern, rawPatternStartRange, flags);
                }
            }
            /**
             * Validate strings passed to the RegExp constructor
             * @param {ASTNode} node node to validate
             * @returns {void}
             * @private
             */
            function checkFunction(node) {
                const scope = context.getScope();
                const regExpVar = astUtils.getVariableByName(scope, "RegExp");
                const shadowed = regExpVar && regExpVar.defs.length > 0;
                const patternNode = node.arguments[0];
                const flagsNode = node.arguments[1];
                if (node.callee.type === "Identifier" && node.callee.name === "RegExp" && isString(patternNode) && !shadowed) {
                    const pattern = patternNode.value;
                    const rawPattern = patternNode.raw.slice(1, -1);
                    const rawPatternStartRange = patternNode.range[0] + 1;
                    const flags = isString(flagsNode) ? flagsNode.value : "";
                    checkRegex(node, pattern, rawPattern, rawPatternStartRange, flags);
                }
            }
            return {
                Literal: checkLiteral,
                CallExpression: checkFunction,
                NewExpression: checkFunction
            };
        }
    };
},
/* eslint/lib/rules/no-restricted-exports.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow specified names in exports",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-restricted-exports"
            },
            schema: [{
                    type: "object",
                    properties: {
                        restrictedNamedExports: {
                            type: "array",
                            items: {
                                type: "string"
                            },
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                restrictedNamed: "'{{name}}' is restricted from being used as an exported name."
            }
        },
        create(context) {
            const restrictedNames = new Set(context.options[0] && context.options[0].restrictedNamedExports);
            /**
             * Checks and reports given exported identifier.
             * @param {ASTNode} node exported `Identifier` node to check.
             * @returns {void}
             */
            function checkExportedName(node) {
                const name = node.name;
                if (restrictedNames.has(name)) {
                    context.report({
                        node,
                        messageId: "restrictedNamed",
                        data: { name }
                    });
                }
            }
            return {
                ExportAllDeclaration(node) {
                    if (node.exported) {
                        checkExportedName(node.exported);
                    }
                },
                ExportNamedDeclaration(node) {
                    const declaration = node.declaration;
                    if (declaration) {
                        if (declaration.type === "FunctionDeclaration" || declaration.type === "ClassDeclaration") {
                            checkExportedName(declaration.id);
                        }
                        else if (declaration.type === "VariableDeclaration") {
                            context.getDeclaredVariables(declaration)
                                .map(v => v.defs.find(d => d.parent === declaration))
                                .map(d => d.name) // Identifier nodes
                                .forEach(checkExportedName);
                        }
                    }
                    else {
                        node.specifiers
                            .map(s => s.exported)
                            .forEach(checkExportedName);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-restricted-globals.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow specified global variables",
                category: "Variables",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-restricted-globals"
            },
            schema: {
                type: "array",
                items: {
                    oneOf: [
                        {
                            type: "string"
                        },
                        {
                            type: "object",
                            properties: {
                                name: { type: "string" },
                                message: { type: "string" }
                            },
                            required: ["name"],
                            additionalProperties: false
                        }
                    ]
                },
                uniqueItems: true,
                minItems: 0
            },
            messages: {
                defaultMessage: "Unexpected use of '{{name}}'.",
                // eslint-disable-next-line eslint-plugin/report-message-format
                customMessage: "Unexpected use of '{{name}}'. {{customMessage}}"
            }
        },
        create(context) {
            // If no globals are restricted, we don't need to do anything
            if (context.options.length === 0) {
                return {};
            }
            const restrictedGlobalMessages = context.options.reduce((memo, option) => {
                if (typeof option === "string") {
                    memo[option] = null;
                }
                else {
                    memo[option.name] = option.message;
                }
                return memo;
            }, {});
            /**
             * Report a variable to be used as a restricted global.
             * @param {Reference} reference the variable reference
             * @returns {void}
             * @private
             */
            function reportReference(reference) {
                const name = reference.identifier.name, customMessage = restrictedGlobalMessages[name], messageId = customMessage
                    ? "customMessage"
                    : "defaultMessage";
                context.report({
                    node: reference.identifier,
                    messageId,
                    data: {
                        name,
                        customMessage
                    }
                });
            }
            /**
             * Check if the given name is a restricted global name.
             * @param {string} name name of a variable
             * @returns {boolean} whether the variable is a restricted global or not
             * @private
             */
            function isRestricted(name) {
                return Object.prototype.hasOwnProperty.call(restrictedGlobalMessages, name);
            }
            return {
                Program() {
                    const scope = context.getScope();
                    // Report variables declared elsewhere (ex: variables defined as "global" by eslint)
                    scope.variables.forEach(variable => {
                        if (!variable.defs.length && isRestricted(variable.name)) {
                            variable.references.forEach(reportReference);
                        }
                    });
                    // Report variables not declared at all
                    scope.through.forEach(reference => {
                        if (isRestricted(reference.identifier.name)) {
                            reportReference(reference);
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-restricted-imports.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const ignore = require(276) /* ignore */;
    const arrayOfStringsOrObjects = {
        type: "array",
        items: {
            anyOf: [
                { type: "string" },
                {
                    type: "object",
                    properties: {
                        name: { type: "string" },
                        message: {
                            type: "string",
                            minLength: 1
                        },
                        importNames: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        }
                    },
                    additionalProperties: false,
                    required: ["name"]
                }
            ]
        },
        uniqueItems: true
    };
    const arrayOfStringsOrObjectPatterns = {
        anyOf: [
            {
                type: "array",
                items: {
                    type: "string"
                },
                uniqueItems: true
            },
            {
                type: "array",
                items: {
                    type: "object",
                    properties: {
                        group: {
                            type: "array",
                            items: {
                                type: "string"
                            },
                            minItems: 1,
                            uniqueItems: true
                        },
                        message: {
                            type: "string",
                            minLength: 1
                        }
                    },
                    additionalProperties: false,
                    required: ["group"]
                },
                uniqueItems: true
            }
        ]
    };
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow specified modules when loaded by `import`",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-restricted-imports"
            },
            messages: {
                path: "'{{importSource}}' import is restricted from being used.",
                // eslint-disable-next-line eslint-plugin/report-message-format
                pathWithCustomMessage: "'{{importSource}}' import is restricted from being used. {{customMessage}}",
                patterns: "'{{importSource}}' import is restricted from being used by a pattern.",
                // eslint-disable-next-line eslint-plugin/report-message-format
                patternWithCustomMessage: "'{{importSource}}' import is restricted from being used by a pattern. {{customMessage}}",
                everything: "* import is invalid because '{{importNames}}' from '{{importSource}}' is restricted.",
                // eslint-disable-next-line eslint-plugin/report-message-format
                everythingWithCustomMessage: "* import is invalid because '{{importNames}}' from '{{importSource}}' is restricted. {{customMessage}}",
                importName: "'{{importName}}' import from '{{importSource}}' is restricted.",
                // eslint-disable-next-line eslint-plugin/report-message-format
                importNameWithCustomMessage: "'{{importName}}' import from '{{importSource}}' is restricted. {{customMessage}}"
            },
            schema: {
                anyOf: [
                    arrayOfStringsOrObjects,
                    {
                        type: "array",
                        items: [{
                                type: "object",
                                properties: {
                                    paths: arrayOfStringsOrObjects,
                                    patterns: arrayOfStringsOrObjectPatterns
                                },
                                additionalProperties: false
                            }],
                        additionalItems: false
                    }
                ]
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const options = Array.isArray(context.options) ? context.options : [];
            const isPathAndPatternsObject = typeof options[0] === "object" &&
                (Object.prototype.hasOwnProperty.call(options[0], "paths") || Object.prototype.hasOwnProperty.call(options[0], "patterns"));
            const restrictedPaths = (isPathAndPatternsObject ? options[0].paths : context.options) || [];
            const restrictedPathMessages = restrictedPaths.reduce((memo, importSource) => {
                if (typeof importSource === "string") {
                    memo[importSource] = { message: null };
                }
                else {
                    memo[importSource.name] = {
                        message: importSource.message,
                        importNames: importSource.importNames
                    };
                }
                return memo;
            }, {});
            // Handle patterns too, either as strings or groups
            const restrictedPatterns = (isPathAndPatternsObject ? options[0].patterns : []) || [];
            const restrictedPatternGroups = restrictedPatterns.length > 0 && typeof restrictedPatterns[0] === "string"
                ? [{ matcher: ignore().add(restrictedPatterns) }]
                : restrictedPatterns.map(({ group, message }) => ({ matcher: ignore().add(group), customMessage: message }));
            // if no imports are restricted we don"t need to check
            if (Object.keys(restrictedPaths).length === 0 && restrictedPatternGroups.length === 0) {
                return {};
            }
            /**
             * Report a restricted path.
             * @param {string} importSource path of the import
             * @param {Map<string,Object[]>} importNames Map of import names that are being imported
             * @param {node} node representing the restricted path reference
             * @returns {void}
             * @private
             */
            function checkRestrictedPathAndReport(importSource, importNames, node) {
                if (!Object.prototype.hasOwnProperty.call(restrictedPathMessages, importSource)) {
                    return;
                }
                const customMessage = restrictedPathMessages[importSource].message;
                const restrictedImportNames = restrictedPathMessages[importSource].importNames;
                if (restrictedImportNames) {
                    if (importNames.has("*")) {
                        const specifierData = importNames.get("*")[0];
                        context.report({
                            node,
                            messageId: customMessage ? "everythingWithCustomMessage" : "everything",
                            loc: specifierData.loc,
                            data: {
                                importSource,
                                importNames: restrictedImportNames,
                                customMessage
                            }
                        });
                    }
                    restrictedImportNames.forEach(importName => {
                        if (importNames.has(importName)) {
                            const specifiers = importNames.get(importName);
                            specifiers.forEach(specifier => {
                                context.report({
                                    node,
                                    messageId: customMessage ? "importNameWithCustomMessage" : "importName",
                                    loc: specifier.loc,
                                    data: {
                                        importSource,
                                        customMessage,
                                        importName
                                    }
                                });
                            });
                        }
                    });
                }
                else {
                    context.report({
                        node,
                        messageId: customMessage ? "pathWithCustomMessage" : "path",
                        data: {
                            importSource,
                            customMessage
                        }
                    });
                }
            }
            /**
             * Report a restricted path specifically for patterns.
             * @param {node} node representing the restricted path reference
             * @param {Object} group contains a Ignore instance for paths, and the customMessage to show if it fails
             * @returns {void}
             * @private
             */
            function reportPathForPatterns(node, group) {
                const importSource = node.source.value.trim();
                context.report({
                    node,
                    messageId: group.customMessage ? "patternWithCustomMessage" : "patterns",
                    data: {
                        importSource,
                        customMessage: group.customMessage
                    }
                });
            }
            /**
             * Check if the given importSource is restricted by a pattern.
             * @param {string} importSource path of the import
             * @param {Object} group contains a Ignore instance for paths, and the customMessage to show if it fails
             * @returns {boolean} whether the variable is a restricted pattern or not
             * @private
             */
            function isRestrictedPattern(importSource, group) {
                return group.matcher.ignores(importSource);
            }
            /**
             * Checks a node to see if any problems should be reported.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             * @private
             */
            function checkNode(node) {
                const importSource = node.source.value.trim();
                const importNames = new Map();
                if (node.type === "ExportAllDeclaration") {
                    const starToken = sourceCode.getFirstToken(node, 1);
                    importNames.set("*", [{ loc: starToken.loc }]);
                }
                else if (node.specifiers) {
                    for (const specifier of node.specifiers) {
                        let name;
                        const specifierData = { loc: specifier.loc };
                        if (specifier.type === "ImportDefaultSpecifier") {
                            name = "default";
                        }
                        else if (specifier.type === "ImportNamespaceSpecifier") {
                            name = "*";
                        }
                        else if (specifier.imported) {
                            name = specifier.imported.name;
                        }
                        else if (specifier.local) {
                            name = specifier.local.name;
                        }
                        if (name) {
                            if (importNames.has(name)) {
                                importNames.get(name).push(specifierData);
                            }
                            else {
                                importNames.set(name, [specifierData]);
                            }
                        }
                    }
                }
                checkRestrictedPathAndReport(importSource, importNames, node);
                restrictedPatternGroups.forEach(group => {
                    if (isRestrictedPattern(importSource, group)) {
                        reportPathForPatterns(node, group);
                    }
                });
            }
            return {
                ImportDeclaration: checkNode,
                ExportNamedDeclaration(node) {
                    if (node.source) {
                        checkNode(node);
                    }
                },
                ExportAllDeclaration: checkNode
            };
        }
    };
},
/* eslint/lib/rules/no-restricted-modules.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const ignore = require(276) /* ignore */;
    const arrayOfStrings = {
        type: "array",
        items: { type: "string" },
        uniqueItems: true
    };
    const arrayOfStringsOrObjects = {
        type: "array",
        items: {
            anyOf: [
                { type: "string" },
                {
                    type: "object",
                    properties: {
                        name: { type: "string" },
                        message: {
                            type: "string",
                            minLength: 1
                        }
                    },
                    additionalProperties: false,
                    required: ["name"]
                }
            ]
        },
        uniqueItems: true
    };
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "disallow specified modules when loaded by `require`",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-restricted-modules"
            },
            schema: {
                anyOf: [
                    arrayOfStringsOrObjects,
                    {
                        type: "array",
                        items: {
                            type: "object",
                            properties: {
                                paths: arrayOfStringsOrObjects,
                                patterns: arrayOfStrings
                            },
                            additionalProperties: false
                        },
                        additionalItems: false
                    }
                ]
            },
            messages: {
                defaultMessage: "'{{name}}' module is restricted from being used.",
                // eslint-disable-next-line eslint-plugin/report-message-format
                customMessage: "'{{name}}' module is restricted from being used. {{customMessage}}",
                patternMessage: "'{{name}}' module is restricted from being used by a pattern."
            }
        },
        create(context) {
            const options = Array.isArray(context.options) ? context.options : [];
            const isPathAndPatternsObject = typeof options[0] === "object" &&
                (Object.prototype.hasOwnProperty.call(options[0], "paths") || Object.prototype.hasOwnProperty.call(options[0], "patterns"));
            const restrictedPaths = (isPathAndPatternsObject ? options[0].paths : context.options) || [];
            const restrictedPatterns = (isPathAndPatternsObject ? options[0].patterns : []) || [];
            const restrictedPathMessages = restrictedPaths.reduce((memo, importName) => {
                if (typeof importName === "string") {
                    memo[importName] = null;
                }
                else {
                    memo[importName.name] = importName.message;
                }
                return memo;
            }, {});
            // if no imports are restricted we don"t need to check
            if (Object.keys(restrictedPaths).length === 0 && restrictedPatterns.length === 0) {
                return {};
            }
            const ig = ignore().add(restrictedPatterns);
            /**
             * Function to check if a node is a string literal.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} If the node is a string literal.
             */
            function isStringLiteral(node) {
                return node && node.type === "Literal" && typeof node.value === "string";
            }
            /**
             * Function to check if a node is a static string template literal.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} If the node is a string template literal.
             */
            function isStaticTemplateLiteral(node) {
                return node && node.type === "TemplateLiteral" && node.expressions.length === 0;
            }
            /**
             * Function to check if a node is a require call.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} If the node is a require call.
             */
            function isRequireCall(node) {
                return node.callee.type === "Identifier" && node.callee.name === "require";
            }
            /**
             * Extract string from Literal or TemplateLiteral node
             * @param {ASTNode} node The node to extract from
             * @returns {string|null} Extracted string or null if node doesn't represent a string
             */
            function getFirstArgumentString(node) {
                if (isStringLiteral(node)) {
                    return node.value.trim();
                }
                if (isStaticTemplateLiteral(node)) {
                    return node.quasis[0].value.cooked.trim();
                }
                return null;
            }
            /**
             * Report a restricted path.
             * @param {node} node representing the restricted path reference
             * @param {string} name restricted path
             * @returns {void}
             * @private
             */
            function reportPath(node, name) {
                const customMessage = restrictedPathMessages[name];
                const messageId = customMessage
                    ? "customMessage"
                    : "defaultMessage";
                context.report({
                    node,
                    messageId,
                    data: {
                        name,
                        customMessage
                    }
                });
            }
            /**
             * Check if the given name is a restricted path name
             * @param {string} name name of a variable
             * @returns {boolean} whether the variable is a restricted path or not
             * @private
             */
            function isRestrictedPath(name) {
                return Object.prototype.hasOwnProperty.call(restrictedPathMessages, name);
            }
            return {
                CallExpression(node) {
                    if (isRequireCall(node)) {
                        // node has arguments
                        if (node.arguments.length) {
                            const name = getFirstArgumentString(node.arguments[0]);
                            // if first argument is a string literal or a static string template literal
                            if (name) {
                                // check if argument value is in restricted modules array
                                if (isRestrictedPath(name)) {
                                    reportPath(node, name);
                                }
                                if (restrictedPatterns.length > 0 && ig.ignores(name)) {
                                    context.report({
                                        node,
                                        messageId: "patternMessage",
                                        data: { name }
                                    });
                                }
                            }
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-restricted-properties.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow certain properties on certain objects",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-restricted-properties"
            },
            schema: {
                type: "array",
                items: {
                    anyOf: [
                        {
                            type: "object",
                            properties: {
                                object: {
                                    type: "string"
                                },
                                property: {
                                    type: "string"
                                },
                                message: {
                                    type: "string"
                                }
                            },
                            additionalProperties: false,
                            required: ["object"]
                        },
                        {
                            type: "object",
                            properties: {
                                object: {
                                    type: "string"
                                },
                                property: {
                                    type: "string"
                                },
                                message: {
                                    type: "string"
                                }
                            },
                            additionalProperties: false,
                            required: ["property"]
                        }
                    ]
                },
                uniqueItems: true
            },
            messages: {
                // eslint-disable-next-line eslint-plugin/report-message-format
                restrictedObjectProperty: "'{{objectName}}.{{propertyName}}' is restricted from being used.{{message}}",
                // eslint-disable-next-line eslint-plugin/report-message-format
                restrictedProperty: "'{{propertyName}}' is restricted from being used.{{message}}"
            }
        },
        create(context) {
            const restrictedCalls = context.options;
            if (restrictedCalls.length === 0) {
                return {};
            }
            const restrictedProperties = new Map();
            const globallyRestrictedObjects = new Map();
            const globallyRestrictedProperties = new Map();
            restrictedCalls.forEach(option => {
                const objectName = option.object;
                const propertyName = option.property;
                if (typeof objectName === "undefined") {
                    globallyRestrictedProperties.set(propertyName, { message: option.message });
                }
                else if (typeof propertyName === "undefined") {
                    globallyRestrictedObjects.set(objectName, { message: option.message });
                }
                else {
                    if (!restrictedProperties.has(objectName)) {
                        restrictedProperties.set(objectName, new Map());
                    }
                    restrictedProperties.get(objectName).set(propertyName, {
                        message: option.message
                    });
                }
            });
            /**
             * Checks to see whether a property access is restricted, and reports it if so.
             * @param {ASTNode} node The node to report
             * @param {string} objectName The name of the object
             * @param {string} propertyName The name of the property
             * @returns {undefined}
             */
            function checkPropertyAccess(node, objectName, propertyName) {
                if (propertyName === null) {
                    return;
                }
                const matchedObject = restrictedProperties.get(objectName);
                const matchedObjectProperty = matchedObject ? matchedObject.get(propertyName) : globallyRestrictedObjects.get(objectName);
                const globalMatchedProperty = globallyRestrictedProperties.get(propertyName);
                if (matchedObjectProperty) {
                    const message = matchedObjectProperty.message ? ` ${matchedObjectProperty.message}` : "";
                    context.report({
                        node,
                        messageId: "restrictedObjectProperty",
                        data: {
                            objectName,
                            propertyName,
                            message
                        }
                    });
                }
                else if (globalMatchedProperty) {
                    const message = globalMatchedProperty.message ? ` ${globalMatchedProperty.message}` : "";
                    context.report({
                        node,
                        messageId: "restrictedProperty",
                        data: {
                            propertyName,
                            message
                        }
                    });
                }
            }
            /**
             * Checks property accesses in a destructuring assignment expression, e.g. `var foo; ({foo} = bar);`
             * @param {ASTNode} node An AssignmentExpression or AssignmentPattern node
             * @returns {undefined}
             */
            function checkDestructuringAssignment(node) {
                if (node.right.type === "Identifier") {
                    const objectName = node.right.name;
                    if (node.left.type === "ObjectPattern") {
                        node.left.properties.forEach(property => {
                            checkPropertyAccess(node.left, objectName, astUtils.getStaticPropertyName(property));
                        });
                    }
                }
            }
            return {
                MemberExpression(node) {
                    checkPropertyAccess(node, node.object && node.object.name, astUtils.getStaticPropertyName(node));
                },
                VariableDeclarator(node) {
                    if (node.init && node.init.type === "Identifier") {
                        const objectName = node.init.name;
                        if (node.id.type === "ObjectPattern") {
                            node.id.properties.forEach(property => {
                                checkPropertyAccess(node.id, objectName, astUtils.getStaticPropertyName(property));
                            });
                        }
                    }
                },
                AssignmentExpression: checkDestructuringAssignment,
                AssignmentPattern: checkDestructuringAssignment
            };
        }
    };
},
/* eslint/lib/rules/no-restricted-syntax.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow specified syntax",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-restricted-syntax"
            },
            schema: {
                type: "array",
                items: {
                    oneOf: [
                        {
                            type: "string"
                        },
                        {
                            type: "object",
                            properties: {
                                selector: { type: "string" },
                                message: { type: "string" }
                            },
                            required: ["selector"],
                            additionalProperties: false
                        }
                    ]
                },
                uniqueItems: true,
                minItems: 0
            },
            messages: {
                // eslint-disable-next-line eslint-plugin/report-message-format
                restrictedSyntax: "{{message}}"
            }
        },
        create(context) {
            return context.options.reduce((result, selectorOrObject) => {
                const isStringFormat = (typeof selectorOrObject === "string");
                const hasCustomMessage = !isStringFormat && Boolean(selectorOrObject.message);
                const selector = isStringFormat ? selectorOrObject : selectorOrObject.selector;
                const message = hasCustomMessage ? selectorOrObject.message : `Using '${selector}' is not allowed.`;
                return Object.assign(result, {
                    [selector](node) {
                        context.report({
                            node,
                            messageId: "restrictedSyntax",
                            data: { message }
                        });
                    }
                });
            }, {});
        }
    };
},
/* eslint/lib/rules/no-return-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const SENTINEL_TYPE = /^(?:[a-zA-Z]+?Statement|ArrowFunctionExpression|FunctionExpression|ClassExpression)$/u;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow assignment operators in `return` statements",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-return-assign"
            },
            schema: [
                {
                    enum: ["except-parens", "always"]
                }
            ],
            messages: {
                returnAssignment: "Return statement should not contain assignment.",
                arrowAssignment: "Arrow function should not return assignment."
            }
        },
        create(context) {
            const always = (context.options[0] || "except-parens") !== "except-parens";
            const sourceCode = context.getSourceCode();
            return {
                AssignmentExpression(node) {
                    if (!always && astUtils.isParenthesised(sourceCode, node)) {
                        return;
                    }
                    let currentChild = node;
                    let parent = currentChild.parent;
                    // Find ReturnStatement or ArrowFunctionExpression in ancestors.
                    while (parent && !SENTINEL_TYPE.test(parent.type)) {
                        currentChild = parent;
                        parent = parent.parent;
                    }
                    // Reports.
                    if (parent && parent.type === "ReturnStatement") {
                        context.report({
                            node: parent,
                            messageId: "returnAssignment"
                        });
                    }
                    else if (parent && parent.type === "ArrowFunctionExpression" && parent.body === currentChild) {
                        context.report({
                            node: parent,
                            messageId: "arrowAssignment"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-return-await.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary `return await`",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-return-await"
            },
            fixable: null,
            schema: [],
            messages: {
                redundantUseOfAwait: "Redundant use of `await` on a return value."
            }
        },
        create(context) {
            /**
             * Reports a found unnecessary `await` expression.
             * @param {ASTNode} node The node representing the `await` expression to report
             * @returns {void}
             */
            function reportUnnecessaryAwait(node) {
                context.report({
                    node: context.getSourceCode().getFirstToken(node),
                    loc: node.loc,
                    messageId: "redundantUseOfAwait"
                });
            }
            /**
             * Determines whether a thrown error from this node will be caught/handled within this function rather than immediately halting
             * this function. For example, a statement in a `try` block will always have an error handler. A statement in
             * a `catch` block will only have an error handler if there is also a `finally` block.
             * @param {ASTNode} node A node representing a location where an could be thrown
             * @returns {boolean} `true` if a thrown error will be caught/handled in this function
             */
            function hasErrorHandler(node) {
                let ancestor = node;
                while (!astUtils.isFunction(ancestor) && ancestor.type !== "Program") {
                    if (ancestor.parent.type === "TryStatement" && (ancestor === ancestor.parent.block || ancestor === ancestor.parent.handler && ancestor.parent.finalizer)) {
                        return true;
                    }
                    ancestor = ancestor.parent;
                }
                return false;
            }
            /**
             * Checks if a node is placed in tail call position. Once `return` arguments (or arrow function expressions) can be a complex expression,
             * an `await` expression could or could not be unnecessary by the definition of this rule. So we're looking for `await` expressions that are in tail position.
             * @param {ASTNode} node A node representing the `await` expression to check
             * @returns {boolean} The checking result
             */
            function isInTailCallPosition(node) {
                if (node.parent.type === "ArrowFunctionExpression") {
                    return true;
                }
                if (node.parent.type === "ReturnStatement") {
                    return !hasErrorHandler(node.parent);
                }
                if (node.parent.type === "ConditionalExpression" && (node === node.parent.consequent || node === node.parent.alternate)) {
                    return isInTailCallPosition(node.parent);
                }
                if (node.parent.type === "LogicalExpression" && node === node.parent.right) {
                    return isInTailCallPosition(node.parent);
                }
                if (node.parent.type === "SequenceExpression" && node === node.parent.expressions[node.parent.expressions.length - 1]) {
                    return isInTailCallPosition(node.parent);
                }
                return false;
            }
            return {
                AwaitExpression(node) {
                    if (isInTailCallPosition(node) && !hasErrorHandler(node)) {
                        reportUnnecessaryAwait(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-script-url.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `javascript:` urls",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-script-url"
            },
            schema: [],
            messages: {
                unexpectedScriptURL: "Script URL is a form of eval."
            }
        },
        create(context) {
            /**
             * Check whether a node's static value starts with "javascript:" or not.
             * And report an error for unexpected script URL.
             * @param {ASTNode} node node to check
             * @returns {void}
             */
            function check(node) {
                const value = astUtils.getStaticStringValue(node);
                if (typeof value === "string" && value.toLowerCase().indexOf("javascript:") === 0) {
                    context.report({ node, messageId: "unexpectedScriptURL" });
                }
            }
            return {
                Literal(node) {
                    if (node.value && typeof node.value === "string") {
                        check(node);
                    }
                },
                TemplateLiteral(node) {
                    if (!(node.parent && node.parent.type === "TaggedTemplateExpression")) {
                        check(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-self-assign.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const SPACES = /\s+/gu;
    /**
     * Traverses 2 Pattern nodes in parallel, then reports self-assignments.
     * @param {ASTNode|null} left A left node to traverse. This is a Pattern or
     *      a Property.
     * @param {ASTNode|null} right A right node to traverse. This is a Pattern or
     *      a Property.
     * @param {boolean} props The flag to check member expressions as well.
     * @param {Function} report A callback function to report.
     * @returns {void}
     */
    function eachSelfAssignment(left, right, props, report) {
        if (!left || !right) {
            // do nothing
        }
        else if (left.type === "Identifier" &&
            right.type === "Identifier" &&
            left.name === right.name) {
            report(right);
        }
        else if (left.type === "ArrayPattern" &&
            right.type === "ArrayExpression") {
            const end = Math.min(left.elements.length, right.elements.length);
            for (let i = 0; i < end; ++i) {
                const leftElement = left.elements[i];
                const rightElement = right.elements[i];
                // Avoid cases such as [...a] = [...a, 1]
                if (leftElement &&
                    leftElement.type === "RestElement" &&
                    i < right.elements.length - 1) {
                    break;
                }
                eachSelfAssignment(leftElement, rightElement, props, report);
                // After a spread element, those indices are unknown.
                if (rightElement && rightElement.type === "SpreadElement") {
                    break;
                }
            }
        }
        else if (left.type === "RestElement" &&
            right.type === "SpreadElement") {
            eachSelfAssignment(left.argument, right.argument, props, report);
        }
        else if (left.type === "ObjectPattern" &&
            right.type === "ObjectExpression" &&
            right.properties.length >= 1) {
            /*
             * Gets the index of the last spread property.
             * It's possible to overwrite properties followed by it.
             */
            let startJ = 0;
            for (let i = right.properties.length - 1; i >= 0; --i) {
                const propType = right.properties[i].type;
                if (propType === "SpreadElement" || propType === "ExperimentalSpreadProperty") {
                    startJ = i + 1;
                    break;
                }
            }
            for (let i = 0; i < left.properties.length; ++i) {
                for (let j = startJ; j < right.properties.length; ++j) {
                    eachSelfAssignment(left.properties[i], right.properties[j], props, report);
                }
            }
        }
        else if (left.type === "Property" &&
            right.type === "Property" &&
            right.kind === "init" &&
            !right.method) {
            const leftName = astUtils.getStaticPropertyName(left);
            if (leftName !== null && leftName === astUtils.getStaticPropertyName(right)) {
                eachSelfAssignment(left.value, right.value, props, report);
            }
        }
        else if (props &&
            astUtils.skipChainExpression(left).type === "MemberExpression" &&
            astUtils.skipChainExpression(right).type === "MemberExpression" &&
            astUtils.isSameReference(left, right)) {
            report(right);
        }
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow assignments where both sides are exactly the same",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-self-assign"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        props: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                selfAssignment: "'{{name}}' is assigned to itself."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const [{ props = true } = {}] = context.options;
            /**
             * Reports a given node as self assignments.
             * @param {ASTNode} node A node to report. This is an Identifier node.
             * @returns {void}
             */
            function report(node) {
                context.report({
                    node,
                    messageId: "selfAssignment",
                    data: {
                        name: sourceCode.getText(node).replace(SPACES, "")
                    }
                });
            }
            return {
                AssignmentExpression(node) {
                    if (node.operator === "=") {
                        eachSelfAssignment(node.left, node.right, props, report);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-self-compare.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow comparisons where both sides are exactly the same",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-self-compare"
            },
            schema: [],
            messages: {
                comparingToSelf: "Comparing to itself is potentially pointless."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Determines whether two nodes are composed of the same tokens.
             * @param {ASTNode} nodeA The first node
             * @param {ASTNode} nodeB The second node
             * @returns {boolean} true if the nodes have identical token representations
             */
            function hasSameTokens(nodeA, nodeB) {
                const tokensA = sourceCode.getTokens(nodeA);
                const tokensB = sourceCode.getTokens(nodeB);
                return tokensA.length === tokensB.length &&
                    tokensA.every((token, index) => token.type === tokensB[index].type && token.value === tokensB[index].value);
            }
            return {
                BinaryExpression(node) {
                    const operators = new Set(["===", "==", "!==", "!=", ">", "<", ">=", "<="]);
                    if (operators.has(node.operator) && hasSameTokens(node.left, node.right)) {
                        context.report({ node, messageId: "comparingToSelf" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-sequences.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const DEFAULT_OPTIONS = {
        allowInParentheses: true
    };
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow comma operators",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-sequences"
            },
            schema: [{
                    properties: {
                        allowInParentheses: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                unexpectedCommaExpression: "Unexpected use of comma operator."
            }
        },
        create(context) {
            const options = Object.assign({}, DEFAULT_OPTIONS, context.options[0]);
            const sourceCode = context.getSourceCode();
            /**
             * Parts of the grammar that are required to have parens.
             */
            const parenthesized = {
                DoWhileStatement: "test",
                IfStatement: "test",
                SwitchStatement: "discriminant",
                WhileStatement: "test",
                WithStatement: "object",
                ArrowFunctionExpression: "body"
                /*
                 * Omitting CallExpression - commas are parsed as argument separators
                 * Omitting NewExpression - commas are parsed as argument separators
                 * Omitting ForInStatement - parts aren't individually parenthesised
                 * Omitting ForStatement - parts aren't individually parenthesised
                 */
            };
            /**
             * Determines whether a node is required by the grammar to be wrapped in
             * parens, e.g. the test of an if statement.
             * @param {ASTNode} node The AST node
             * @returns {boolean} True if parens around node belong to parent node.
             */
            function requiresExtraParens(node) {
                return node.parent && parenthesized[node.parent.type] &&
                    node === node.parent[parenthesized[node.parent.type]];
            }
            /**
             * Check if a node is wrapped in parens.
             * @param {ASTNode} node The AST node
             * @returns {boolean} True if the node has a paren on each side.
             */
            function isParenthesised(node) {
                return astUtils.isParenthesised(sourceCode, node);
            }
            /**
             * Check if a node is wrapped in two levels of parens.
             * @param {ASTNode} node The AST node
             * @returns {boolean} True if two parens surround the node on each side.
             */
            function isParenthesisedTwice(node) {
                const previousToken = sourceCode.getTokenBefore(node, 1), nextToken = sourceCode.getTokenAfter(node, 1);
                return isParenthesised(node) && previousToken && nextToken &&
                    astUtils.isOpeningParenToken(previousToken) && previousToken.range[1] <= node.range[0] &&
                    astUtils.isClosingParenToken(nextToken) && nextToken.range[0] >= node.range[1];
            }
            return {
                SequenceExpression(node) {
                    // Always allow sequences in for statement update
                    if (node.parent.type === "ForStatement" &&
                        (node === node.parent.init || node === node.parent.update)) {
                        return;
                    }
                    // Wrapping a sequence in extra parens indicates intent
                    if (options.allowInParentheses) {
                        if (requiresExtraParens(node)) {
                            if (isParenthesisedTwice(node)) {
                                return;
                            }
                        }
                        else {
                            if (isParenthesised(node)) {
                                return;
                            }
                        }
                    }
                    const firstCommaToken = sourceCode.getTokenAfter(node.expressions[0], astUtils.isCommaToken);
                    context.report({ node, loc: firstCommaToken.loc, messageId: "unexpectedCommaExpression" });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-setter-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const { findVariable } = require(350) /* eslint-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Determines whether the given identifier node is a reference to a global variable.
     * @param {ASTNode} node `Identifier` node to check.
     * @param {Scope} scope Scope to which the node belongs.
     * @returns {boolean} True if the identifier is a reference to a global variable.
     */
    function isGlobalReference(node, scope) {
        const variable = findVariable(scope, node);
        return variable !== null && variable.scope.type === "global" && variable.defs.length === 0;
    }
    /**
     * Determines whether the given node is an argument of the specified global method call, at the given `index` position.
     * E.g., for given `index === 1`, this function checks for `objectName.methodName(foo, node)`, where objectName is a global variable.
     * @param {ASTNode} node The node to check.
     * @param {Scope} scope Scope to which the node belongs.
     * @param {string} objectName Name of the global object.
     * @param {string} methodName Name of the method.
     * @param {number} index The given position.
     * @returns {boolean} `true` if the node is argument at the given position.
     */
    function isArgumentOfGlobalMethodCall(node, scope, objectName, methodName, index) {
        const callNode = node.parent;
        return callNode.type === "CallExpression" &&
            callNode.arguments[index] === node &&
            astUtils.isSpecificMemberAccess(callNode.callee, objectName, methodName) &&
            isGlobalReference(astUtils.skipChainExpression(callNode.callee).object, scope);
    }
    /**
     * Determines whether the given node is used as a property descriptor.
     * @param {ASTNode} node The node to check.
     * @param {Scope} scope Scope to which the node belongs.
     * @returns {boolean} `true` if the node is a property descriptor.
     */
    function isPropertyDescriptor(node, scope) {
        if (isArgumentOfGlobalMethodCall(node, scope, "Object", "defineProperty", 2) ||
            isArgumentOfGlobalMethodCall(node, scope, "Reflect", "defineProperty", 2)) {
            return true;
        }
        const parent = node.parent;
        if (parent.type === "Property" &&
            parent.value === node) {
            const grandparent = parent.parent;
            if (grandparent.type === "ObjectExpression" &&
                (isArgumentOfGlobalMethodCall(grandparent, scope, "Object", "create", 1) ||
                    isArgumentOfGlobalMethodCall(grandparent, scope, "Object", "defineProperties", 1))) {
                return true;
            }
        }
        return false;
    }
    /**
     * Determines whether the given function node is used as a setter function.
     * @param {ASTNode} node The node to check.
     * @param {Scope} scope Scope to which the node belongs.
     * @returns {boolean} `true` if the node is a setter.
     */
    function isSetter(node, scope) {
        const parent = node.parent;
        if (parent.kind === "set" &&
            parent.value === node) {
            // Setter in an object literal or in a class
            return true;
        }
        if (parent.type === "Property" &&
            parent.value === node &&
            astUtils.getStaticPropertyName(parent) === "set" &&
            parent.parent.type === "ObjectExpression" &&
            isPropertyDescriptor(parent.parent, scope)) {
            // Setter in a property descriptor
            return true;
        }
        return false;
    }
    /**
     * Finds function's outer scope.
     * @param {Scope} scope Function's own scope.
     * @returns {Scope} Function's outer scope.
     */
    function getOuterScope(scope) {
        const upper = scope.upper;
        if (upper.type === "function-expression-name") {
            return upper.upper;
        }
        return upper;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow returning values from setters",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-setter-return"
            },
            schema: [],
            messages: {
                returnsValue: "Setter cannot return a value."
            }
        },
        create(context) {
            let funcInfo = null;
            /**
             * Creates and pushes to the stack a function info object for the given function node.
             * @param {ASTNode} node The function node.
             * @returns {void}
             */
            function enterFunction(node) {
                const outerScope = getOuterScope(context.getScope());
                funcInfo = {
                    upper: funcInfo,
                    isSetter: isSetter(node, outerScope)
                };
            }
            /**
             * Pops the current function info object from the stack.
             * @returns {void}
             */
            function exitFunction() {
                funcInfo = funcInfo.upper;
            }
            /**
             * Reports the given node.
             * @param {ASTNode} node Node to report.
             * @returns {void}
             */
            function report(node) {
                context.report({ node, messageId: "returnsValue" });
            }
            return {
                /*
                 * Function declarations cannot be setters, but we still have to track them in the `funcInfo` stack to avoid
                 * false positives, because a ReturnStatement node can belong to a function declaration inside a setter.
                 *
                 * Note: A previously declared function can be referenced and actually used as a setter in a property descriptor,
                 * but that's out of scope for this rule.
                 */
                FunctionDeclaration: enterFunction,
                FunctionExpression: enterFunction,
                ArrowFunctionExpression(node) {
                    enterFunction(node);
                    if (funcInfo.isSetter && node.expression) {
                        // { set: foo => bar } property descriptor. Report implicit return 'bar' as the equivalent for a return statement.
                        report(node.body);
                    }
                },
                "FunctionDeclaration:exit": exitFunction,
                "FunctionExpression:exit": exitFunction,
                "ArrowFunctionExpression:exit": exitFunction,
                ReturnStatement(node) {
                    // Global returns (e.g., at the top level of a Node module) don't have `funcInfo`.
                    if (funcInfo && funcInfo.isSetter && node.argument) {
                        report(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-shadow.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow variable declarations from shadowing variables declared in the outer scope",
                category: "Variables",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-shadow"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        builtinGlobals: { type: "boolean", default: false },
                        hoist: { enum: ["all", "functions", "never"], default: "functions" },
                        allow: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                noShadow: "'{{name}}' is already declared in the upper scope on line {{shadowedLine}} column {{shadowedColumn}}.",
                noShadowGlobal: "'{{name}}' is already a global variable."
            }
        },
        create(context) {
            const options = {
                builtinGlobals: context.options[0] && context.options[0].builtinGlobals,
                hoist: (context.options[0] && context.options[0].hoist) || "functions",
                allow: (context.options[0] && context.options[0].allow) || []
            };
            /**
             * Check if variable name is allowed.
             * @param  {ASTNode} variable The variable to check.
             * @returns {boolean} Whether or not the variable name is allowed.
             */
            function isAllowed(variable) {
                return options.allow.indexOf(variable.name) !== -1;
            }
            /**
             * Checks if a variable of the class name in the class scope of ClassDeclaration.
             *
             * ClassDeclaration creates two variables of its name into its outer scope and its class scope.
             * So we should ignore the variable in the class scope.
             * @param {Object} variable The variable to check.
             * @returns {boolean} Whether or not the variable of the class name in the class scope of ClassDeclaration.
             */
            function isDuplicatedClassNameVariable(variable) {
                const block = variable.scope.block;
                return block.type === "ClassDeclaration" && block.id === variable.identifiers[0];
            }
            /**
             * Checks if a variable is inside the initializer of scopeVar.
             *
             * To avoid reporting at declarations such as `var a = function a() {};`.
             * But it should report `var a = function(a) {};` or `var a = function() { function a() {} };`.
             * @param {Object} variable The variable to check.
             * @param {Object} scopeVar The scope variable to look for.
             * @returns {boolean} Whether or not the variable is inside initializer of scopeVar.
             */
            function isOnInitializer(variable, scopeVar) {
                const outerScope = scopeVar.scope;
                const outerDef = scopeVar.defs[0];
                const outer = outerDef && outerDef.parent && outerDef.parent.range;
                const innerScope = variable.scope;
                const innerDef = variable.defs[0];
                const inner = innerDef && innerDef.name.range;
                return (outer &&
                    inner &&
                    outer[0] < inner[0] &&
                    inner[1] < outer[1] &&
                    ((innerDef.type === "FunctionName" && innerDef.node.type === "FunctionExpression") || innerDef.node.type === "ClassExpression") &&
                    outerScope === innerScope.upper);
            }
            /**
             * Get a range of a variable's identifier node.
             * @param {Object} variable The variable to get.
             * @returns {Array|undefined} The range of the variable's identifier node.
             */
            function getNameRange(variable) {
                const def = variable.defs[0];
                return def && def.name.range;
            }
            /**
             * Get declared line and column of a variable.
             * @param {eslint-scope.Variable} variable The variable to get.
             * @returns {Object} The declared line and column of the variable.
             */
            function getDeclaredLocation(variable) {
                const identifier = variable.identifiers[0];
                let obj;
                if (identifier) {
                    obj = {
                        global: false,
                        line: identifier.loc.start.line,
                        column: identifier.loc.start.column + 1
                    };
                }
                else {
                    obj = {
                        global: true
                    };
                }
                return obj;
            }
            /**
             * Checks if a variable is in TDZ of scopeVar.
             * @param {Object} variable The variable to check.
             * @param {Object} scopeVar The variable of TDZ.
             * @returns {boolean} Whether or not the variable is in TDZ of scopeVar.
             */
            function isInTdz(variable, scopeVar) {
                const outerDef = scopeVar.defs[0];
                const inner = getNameRange(variable);
                const outer = getNameRange(scopeVar);
                return (inner &&
                    outer &&
                    inner[1] < outer[0] &&
                    // Excepts FunctionDeclaration if is {"hoist":"function"}.
                    (options.hoist !== "functions" || !outerDef || outerDef.node.type !== "FunctionDeclaration"));
            }
            /**
             * Checks the current context for shadowed variables.
             * @param {Scope} scope Fixme
             * @returns {void}
             */
            function checkForShadows(scope) {
                const variables = scope.variables;
                for (let i = 0; i < variables.length; ++i) {
                    const variable = variables[i];
                    // Skips "arguments" or variables of a class name in the class scope of ClassDeclaration.
                    if (variable.identifiers.length === 0 ||
                        isDuplicatedClassNameVariable(variable) ||
                        isAllowed(variable)) {
                        continue;
                    }
                    // Gets shadowed variable.
                    const shadowed = astUtils.getVariableByName(scope.upper, variable.name);
                    if (shadowed &&
                        (shadowed.identifiers.length > 0 || (options.builtinGlobals && "writeable" in shadowed)) &&
                        !isOnInitializer(variable, shadowed) &&
                        !(options.hoist !== "all" && isInTdz(variable, shadowed))) {
                        const location = getDeclaredLocation(shadowed);
                        const messageId = location.global ? "noShadowGlobal" : "noShadow";
                        const data = { name: variable.name };
                        if (!location.global) {
                            data.shadowedLine = location.line;
                            data.shadowedColumn = location.column;
                        }
                        context.report({
                            node: variable.identifiers[0],
                            messageId,
                            data
                        });
                    }
                }
            }
            return {
                "Program:exit"() {
                    const globalScope = context.getScope();
                    const stack = globalScope.childScopes.slice();
                    while (stack.length) {
                        const scope = stack.pop();
                        stack.push(...scope.childScopes);
                        checkForShadows(scope);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-shadow-restricted-names.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Determines if a variable safely shadows undefined.
     * This is the case when a variable named `undefined` is never assigned to a value (i.e. it always shares the same value
     * as the global).
     * @param {eslintScope.Variable} variable The variable to check
     * @returns {boolean} true if this variable safely shadows `undefined`
     */
    function safelyShadowsUndefined(variable) {
        return variable.name === "undefined" &&
            variable.references.every(ref => !ref.isWrite()) &&
            variable.defs.every(def => def.node.type === "VariableDeclarator" && def.node.init === null);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow identifiers from shadowing restricted names",
                category: "Variables",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-shadow-restricted-names"
            },
            schema: [],
            messages: {
                shadowingRestrictedName: "Shadowing of global property '{{name}}'."
            }
        },
        create(context) {
            const RESTRICTED = new Set(["undefined", "NaN", "Infinity", "arguments", "eval"]);
            return {
                "VariableDeclaration, :function, CatchClause"(node) {
                    for (const variable of context.getDeclaredVariables(node)) {
                        if (variable.defs.length > 0 && RESTRICTED.has(variable.name) && !safelyShadowsUndefined(variable)) {
                            context.report({
                                node: variable.defs[0].name,
                                messageId: "shadowingRestrictedName",
                                data: {
                                    name: variable.name
                                }
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-spaced-func.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "disallow spacing between function identifiers and their applications (deprecated)",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-spaced-func"
            },
            deprecated: true,
            replacedBy: ["func-call-spacing"],
            fixable: "whitespace",
            schema: [],
            messages: {
                noSpacedFunction: "Unexpected space between function name and paren."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Check if open space is present in a function name
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function detectOpenSpaces(node) {
                const lastCalleeToken = sourceCode.getLastToken(node.callee);
                let prevToken = lastCalleeToken, parenToken = sourceCode.getTokenAfter(lastCalleeToken);
                // advances to an open parenthesis.
                while (parenToken &&
                    parenToken.range[1] < node.range[1] &&
                    parenToken.value !== "(") {
                    prevToken = parenToken;
                    parenToken = sourceCode.getTokenAfter(parenToken);
                }
                // look for a space between the callee and the open paren
                if (parenToken &&
                    parenToken.range[1] < node.range[1] &&
                    sourceCode.isSpaceBetweenTokens(prevToken, parenToken)) {
                    context.report({
                        node,
                        loc: lastCalleeToken.loc.start,
                        messageId: "noSpacedFunction",
                        fix(fixer) {
                            return fixer.removeRange([prevToken.range[1], parenToken.range[0]]);
                        }
                    });
                }
            }
            return {
                CallExpression: detectOpenSpaces,
                NewExpression: detectOpenSpaces
            };
        }
    };
},
/* eslint/lib/rules/no-sparse-arrays.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow sparse arrays",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-sparse-arrays"
            },
            schema: [],
            messages: {
                unexpectedSparseArray: "Unexpected comma in middle of array."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                ArrayExpression(node) {
                    const emptySpot = node.elements.indexOf(null) > -1;
                    if (emptySpot) {
                        context.report({ node, messageId: "unexpectedSparseArray" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-sync.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            deprecated: true,
            replacedBy: [],
            type: "suggestion",
            docs: {
                description: "disallow synchronous methods",
                category: "Node.js and CommonJS",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-sync"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowAtRootLevel: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                noSync: "Unexpected sync method: '{{propertyName}}'."
            }
        },
        create(context) {
            const selector = context.options[0] && context.options[0].allowAtRootLevel
                ? ":function MemberExpression[property.name=/.*Sync$/]"
                : "MemberExpression[property.name=/.*Sync$/]";
            return {
                [selector](node) {
                    context.report({
                        node,
                        messageId: "noSync",
                        data: {
                            propertyName: node.property.name
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-tabs.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const tabRegex = /\t+/gu;
    const anyNonWhitespaceRegex = /\S/u;
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "disallow all tabs",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-tabs"
            },
            schema: [{
                    type: "object",
                    properties: {
                        allowIndentationTabs: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                unexpectedTab: "Unexpected tab character."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const allowIndentationTabs = context.options && context.options[0] && context.options[0].allowIndentationTabs;
            return {
                Program(node) {
                    sourceCode.getLines().forEach((line, index) => {
                        let match;
                        while ((match = tabRegex.exec(line)) !== null) {
                            if (allowIndentationTabs && !anyNonWhitespaceRegex.test(line.slice(0, match.index))) {
                                continue;
                            }
                            context.report({
                                node,
                                loc: {
                                    start: {
                                        line: index + 1,
                                        column: match.index
                                    },
                                    end: {
                                        line: index + 1,
                                        column: match.index + match[0].length
                                    }
                                },
                                messageId: "unexpectedTab"
                            });
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-template-curly-in-string.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow template literal placeholder syntax in regular strings",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-template-curly-in-string"
            },
            schema: [],
            messages: {
                unexpectedTemplateExpression: "Unexpected template string expression."
            }
        },
        create(context) {
            const regex = /\$\{[^}]+\}/u;
            return {
                Literal(node) {
                    if (typeof node.value === "string" && regex.test(node.value)) {
                        context.report({
                            node,
                            messageId: "unexpectedTemplateExpression"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-ternary.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow ternary operators",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-ternary"
            },
            schema: [],
            messages: {
                noTernaryOperator: "Ternary operator used."
            }
        },
        create(context) {
            return {
                ConditionalExpression(node) {
                    context.report({ node, messageId: "noTernaryOperator" });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-this-before-super.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given node is a constructor.
     * @param {ASTNode} node A node to check. This node type is one of
     *   `Program`, `FunctionDeclaration`, `FunctionExpression`, and
     *   `ArrowFunctionExpression`.
     * @returns {boolean} `true` if the node is a constructor.
     */
    function isConstructorFunction(node) {
        return (node.type === "FunctionExpression" &&
            node.parent.type === "MethodDefinition" &&
            node.parent.kind === "constructor");
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow `this`/`super` before calling `super()` in constructors",
                category: "ECMAScript 6",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-this-before-super"
            },
            schema: [],
            messages: {
                noBeforeSuper: "'{{kind}}' is not allowed before 'super()'."
            }
        },
        create(context) {
            /*
             * Information for each constructor.
             * - upper:      Information of the upper constructor.
             * - hasExtends: A flag which shows whether the owner class has a valid
             *   `extends` part.
             * - scope:      The scope of the owner class.
             * - codePath:   The code path of this constructor.
             */
            let funcInfo = null;
            /*
             * Information for each code path segment.
             * Each key is the id of a code path segment.
             * Each value is an object:
             * - superCalled:  The flag which shows `super()` called in all code paths.
             * - invalidNodes: The array of invalid ThisExpression and Super nodes.
             */
            let segInfoMap = Object.create(null);
            /**
             * Gets whether or not `super()` is called in a given code path segment.
             * @param {CodePathSegment} segment A code path segment to get.
             * @returns {boolean} `true` if `super()` is called.
             */
            function isCalled(segment) {
                return !segment.reachable || segInfoMap[segment.id].superCalled;
            }
            /**
             * Checks whether or not this is in a constructor.
             * @returns {boolean} `true` if this is in a constructor.
             */
            function isInConstructorOfDerivedClass() {
                return Boolean(funcInfo && funcInfo.isConstructor && funcInfo.hasExtends);
            }
            /**
             * Checks whether or not this is before `super()` is called.
             * @returns {boolean} `true` if this is before `super()` is called.
             */
            function isBeforeCallOfSuper() {
                return (isInConstructorOfDerivedClass() &&
                    !funcInfo.codePath.currentSegments.every(isCalled));
            }
            /**
             * Sets a given node as invalid.
             * @param {ASTNode} node A node to set as invalid. This is one of
             *      a ThisExpression and a Super.
             * @returns {void}
             */
            function setInvalid(node) {
                const segments = funcInfo.codePath.currentSegments;
                for (let i = 0; i < segments.length; ++i) {
                    const segment = segments[i];
                    if (segment.reachable) {
                        segInfoMap[segment.id].invalidNodes.push(node);
                    }
                }
            }
            /**
             * Sets the current segment as `super` was called.
             * @returns {void}
             */
            function setSuperCalled() {
                const segments = funcInfo.codePath.currentSegments;
                for (let i = 0; i < segments.length; ++i) {
                    const segment = segments[i];
                    if (segment.reachable) {
                        segInfoMap[segment.id].superCalled = true;
                    }
                }
            }
            return {
                /**
                 * Adds information of a constructor into the stack.
                 * @param {CodePath} codePath A code path which was started.
                 * @param {ASTNode} node The current node.
                 * @returns {void}
                 */
                onCodePathStart(codePath, node) {
                    if (isConstructorFunction(node)) {
                        // Class > ClassBody > MethodDefinition > FunctionExpression
                        const classNode = node.parent.parent.parent;
                        funcInfo = {
                            upper: funcInfo,
                            isConstructor: true,
                            hasExtends: Boolean(classNode.superClass &&
                                !astUtils.isNullOrUndefined(classNode.superClass)),
                            codePath
                        };
                    }
                    else {
                        funcInfo = {
                            upper: funcInfo,
                            isConstructor: false,
                            hasExtends: false,
                            codePath
                        };
                    }
                },
                /**
                 * Removes the top of stack item.
                 *
                 * And this traverses all segments of this code path then reports every
                 * invalid node.
                 * @param {CodePath} codePath A code path which was ended.
                 * @returns {void}
                 */
                onCodePathEnd(codePath) {
                    const isDerivedClass = funcInfo.hasExtends;
                    funcInfo = funcInfo.upper;
                    if (!isDerivedClass) {
                        return;
                    }
                    codePath.traverseSegments((segment, controller) => {
                        const info = segInfoMap[segment.id];
                        for (let i = 0; i < info.invalidNodes.length; ++i) {
                            const invalidNode = info.invalidNodes[i];
                            context.report({
                                messageId: "noBeforeSuper",
                                node: invalidNode,
                                data: {
                                    kind: invalidNode.type === "Super" ? "super" : "this"
                                }
                            });
                        }
                        if (info.superCalled) {
                            controller.skip();
                        }
                    });
                },
                /**
                 * Initialize information of a given code path segment.
                 * @param {CodePathSegment} segment A code path segment to initialize.
                 * @returns {void}
                 */
                onCodePathSegmentStart(segment) {
                    if (!isInConstructorOfDerivedClass()) {
                        return;
                    }
                    // Initialize info.
                    segInfoMap[segment.id] = {
                        superCalled: (segment.prevSegments.length > 0 &&
                            segment.prevSegments.every(isCalled)),
                        invalidNodes: []
                    };
                },
                /**
                 * Update information of the code path segment when a code path was
                 * looped.
                 * @param {CodePathSegment} fromSegment The code path segment of the
                 *      end of a loop.
                 * @param {CodePathSegment} toSegment A code path segment of the head
                 *      of a loop.
                 * @returns {void}
                 */
                onCodePathSegmentLoop(fromSegment, toSegment) {
                    if (!isInConstructorOfDerivedClass()) {
                        return;
                    }
                    // Update information inside of the loop.
                    funcInfo.codePath.traverseSegments({ first: toSegment, last: fromSegment }, (segment, controller) => {
                        const info = segInfoMap[segment.id];
                        if (info.superCalled) {
                            info.invalidNodes = [];
                            controller.skip();
                        }
                        else if (segment.prevSegments.length > 0 &&
                            segment.prevSegments.every(isCalled)) {
                            info.superCalled = true;
                            info.invalidNodes = [];
                        }
                    });
                },
                /**
                 * Reports if this is before `super()`.
                 * @param {ASTNode} node A target node.
                 * @returns {void}
                 */
                ThisExpression(node) {
                    if (isBeforeCallOfSuper()) {
                        setInvalid(node);
                    }
                },
                /**
                 * Reports if this is before `super()`.
                 * @param {ASTNode} node A target node.
                 * @returns {void}
                 */
                Super(node) {
                    if (!astUtils.isCallee(node) && isBeforeCallOfSuper()) {
                        setInvalid(node);
                    }
                },
                /**
                 * Marks `super()` called.
                 * @param {ASTNode} node A target node.
                 * @returns {void}
                 */
                "CallExpression:exit"(node) {
                    if (node.callee.type === "Super" && isBeforeCallOfSuper()) {
                        setSuperCalled();
                    }
                },
                /**
                 * Resets state.
                 * @returns {void}
                 */
                "Program:exit"() {
                    segInfoMap = Object.create(null);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-throw-literal.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow throwing literals as exceptions",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-throw-literal"
            },
            schema: [],
            messages: {
                object: "Expected an error object to be thrown.",
                undef: "Do not throw undefined."
            }
        },
        create(context) {
            return {
                ThrowStatement(node) {
                    if (!astUtils.couldBeError(node.argument)) {
                        context.report({ node, messageId: "object" });
                    }
                    else if (node.argument.type === "Identifier") {
                        if (node.argument.name === "undefined") {
                            context.report({ node, messageId: "undef" });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-trailing-spaces.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "disallow trailing whitespace at the end of lines",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-trailing-spaces"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        skipBlankLines: {
                            type: "boolean",
                            default: false
                        },
                        ignoreComments: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                trailingSpace: "Trailing spaces not allowed."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const BLANK_CLASS = "[ \t\u00a0\u2000-\u200b\u3000]", SKIP_BLANK = `^${BLANK_CLASS}*$`, NONBLANK = `${BLANK_CLASS}+$`;
            const options = context.options[0] || {}, skipBlankLines = options.skipBlankLines || false, ignoreComments = options.ignoreComments || false;
            /**
             * Report the error message
             * @param {ASTNode} node node to report
             * @param {int[]} location range information
             * @param {int[]} fixRange Range based on the whole program
             * @returns {void}
             */
            function report(node, location, fixRange) {
                /*
                 * Passing node is a bit dirty, because message data will contain big
                 * text in `source`. But... who cares :) ?
                 * One more kludge will not make worse the bloody wizardry of this
                 * plugin.
                 */
                context.report({
                    node,
                    loc: location,
                    messageId: "trailingSpace",
                    fix(fixer) {
                        return fixer.removeRange(fixRange);
                    }
                });
            }
            /**
             * Given a list of comment nodes, return the line numbers for those comments.
             * @param {Array} comments An array of comment nodes.
             * @returns {number[]} An array of line numbers containing comments.
             */
            function getCommentLineNumbers(comments) {
                const lines = new Set();
                comments.forEach(comment => {
                    const endLine = comment.type === "Block"
                        ? comment.loc.end.line - 1
                        : comment.loc.end.line;
                    for (let i = comment.loc.start.line; i <= endLine; i++) {
                        lines.add(i);
                    }
                });
                return lines;
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program: function checkTrailingSpaces(node) {
                    /*
                     * Let's hack. Since Espree does not return whitespace nodes,
                     * fetch the source code and do matching via regexps.
                     */
                    const re = new RegExp(NONBLANK, "u"), skipMatch = new RegExp(SKIP_BLANK, "u"), lines = sourceCode.lines, linebreaks = sourceCode.getText().match(astUtils.createGlobalLinebreakMatcher()), comments = sourceCode.getAllComments(), commentLineNumbers = getCommentLineNumbers(comments);
                    let totalLength = 0, fixRange = [];
                    for (let i = 0, ii = lines.length; i < ii; i++) {
                        const lineNumber = i + 1;
                        /*
                         * Always add linebreak length to line length to accommodate for line break (\n or \r\n)
                         * Because during the fix time they also reserve one spot in the array.
                         * Usually linebreak length is 2 for \r\n (CRLF) and 1 for \n (LF)
                         */
                        const linebreakLength = linebreaks && linebreaks[i] ? linebreaks[i].length : 1;
                        const lineLength = lines[i].length + linebreakLength;
                        const matches = re.exec(lines[i]);
                        if (matches) {
                            const location = {
                                start: {
                                    line: lineNumber,
                                    column: matches.index
                                },
                                end: {
                                    line: lineNumber,
                                    column: lineLength - linebreakLength
                                }
                            };
                            const rangeStart = totalLength + location.start.column;
                            const rangeEnd = totalLength + location.end.column;
                            const containingNode = sourceCode.getNodeByRangeIndex(rangeStart);
                            if (containingNode && containingNode.type === "TemplateElement" &&
                                rangeStart > containingNode.parent.range[0] &&
                                rangeEnd < containingNode.parent.range[1]) {
                                totalLength += lineLength;
                                continue;
                            }
                            /*
                             * If the line has only whitespace, and skipBlankLines
                             * is true, don't report it
                             */
                            if (skipBlankLines && skipMatch.test(lines[i])) {
                                totalLength += lineLength;
                                continue;
                            }
                            fixRange = [rangeStart, rangeEnd];
                            if (!ignoreComments || !commentLineNumbers.has(lineNumber)) {
                                report(node, location, fixRange);
                            }
                        }
                        totalLength += lineLength;
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-undef.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks if the given node is the argument of a typeof operator.
     * @param {ASTNode} node The AST node being checked.
     * @returns {boolean} Whether or not the node is the argument of a typeof operator.
     */
    function hasTypeOfOperator(node) {
        const parent = node.parent;
        return parent.type === "UnaryExpression" && parent.operator === "typeof";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow the use of undeclared variables unless mentioned in `/*global */` comments",
                category: "Variables",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-undef"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        typeof: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                undef: "'{{name}}' is not defined."
            }
        },
        create(context) {
            const options = context.options[0];
            const considerTypeOf = options && options.typeof === true || false;
            return {
                "Program:exit"( /* node */) {
                    const globalScope = context.getScope();
                    globalScope.through.forEach(ref => {
                        const identifier = ref.identifier;
                        if (!considerTypeOf && hasTypeOfOperator(identifier)) {
                            return;
                        }
                        context.report({
                            node: identifier,
                            messageId: "undef",
                            data: identifier
                        });
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-undef-init.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow initializing variables to `undefined`",
                category: "Variables",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-undef-init"
            },
            schema: [],
            fixable: "code",
            messages: {
                unnecessaryUndefinedInit: "It's not necessary to initialize '{{name}}' to undefined."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                VariableDeclarator(node) {
                    const name = sourceCode.getText(node.id), init = node.init && node.init.name, scope = context.getScope(), undefinedVar = astUtils.getVariableByName(scope, "undefined"), shadowed = undefinedVar && undefinedVar.defs.length > 0, lastToken = sourceCode.getLastToken(node);
                    if (init === "undefined" && node.parent.kind !== "const" && !shadowed) {
                        context.report({
                            node,
                            messageId: "unnecessaryUndefinedInit",
                            data: { name },
                            fix(fixer) {
                                if (node.parent.kind === "var") {
                                    return null;
                                }
                                if (node.id.type === "ArrayPattern" || node.id.type === "ObjectPattern") {
                                    // Don't fix destructuring assignment to `undefined`.
                                    return null;
                                }
                                if (sourceCode.commentsExistBetween(node.id, lastToken)) {
                                    return null;
                                }
                                return fixer.removeRange([node.id.range[1], node.range[1]]);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-undefined.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the use of `undefined` as an identifier",
                category: "Variables",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-undefined"
            },
            schema: [],
            messages: {
                unexpectedUndefined: "Unexpected use of undefined."
            }
        },
        create(context) {
            /**
             * Report an invalid "undefined" identifier node.
             * @param {ASTNode} node The node to report.
             * @returns {void}
             */
            function report(node) {
                context.report({
                    node,
                    messageId: "unexpectedUndefined"
                });
            }
            /**
             * Checks the given scope for references to `undefined` and reports
             * all references found.
             * @param {eslint-scope.Scope} scope The scope to check.
             * @returns {void}
             */
            function checkScope(scope) {
                const undefinedVar = scope.set.get("undefined");
                if (!undefinedVar) {
                    return;
                }
                const references = undefinedVar.references;
                const defs = undefinedVar.defs;
                // Report non-initializing references (those are covered in defs below)
                references
                    .filter(ref => !ref.init)
                    .forEach(ref => report(ref.identifier));
                defs.forEach(def => report(def.name));
            }
            return {
                "Program:exit"() {
                    const globalScope = context.getScope();
                    const stack = [globalScope];
                    while (stack.length) {
                        const scope = stack.pop();
                        stack.push(...scope.childScopes);
                        checkScope(scope);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-underscore-dangle.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow dangling underscores in identifiers",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-underscore-dangle"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allow: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        },
                        allowAfterThis: {
                            type: "boolean",
                            default: false
                        },
                        allowAfterSuper: {
                            type: "boolean",
                            default: false
                        },
                        allowAfterThisConstructor: {
                            type: "boolean",
                            default: false
                        },
                        enforceInMethodNames: {
                            type: "boolean",
                            default: false
                        },
                        allowFunctionParams: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedUnderscore: "Unexpected dangling '_' in '{{identifier}}'."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const ALLOWED_VARIABLES = options.allow ? options.allow : [];
            const allowAfterThis = typeof options.allowAfterThis !== "undefined" ? options.allowAfterThis : false;
            const allowAfterSuper = typeof options.allowAfterSuper !== "undefined" ? options.allowAfterSuper : false;
            const allowAfterThisConstructor = typeof options.allowAfterThisConstructor !== "undefined" ? options.allowAfterThisConstructor : false;
            const enforceInMethodNames = typeof options.enforceInMethodNames !== "undefined" ? options.enforceInMethodNames : false;
            const allowFunctionParams = typeof options.allowFunctionParams !== "undefined" ? options.allowFunctionParams : true;
            //-------------------------------------------------------------------------
            // Helpers
            //-------------------------------------------------------------------------
            /**
             * Check if identifier is present inside the allowed option
             * @param {string} identifier name of the node
             * @returns {boolean} true if its is present
             * @private
             */
            function isAllowed(identifier) {
                return ALLOWED_VARIABLES.some(ident => ident === identifier);
            }
            /**
             * Check if identifier has a dangling underscore
             * @param {string} identifier name of the node
             * @returns {boolean} true if its is present
             * @private
             */
            function hasDanglingUnderscore(identifier) {
                const len = identifier.length;
                return identifier !== "_" && (identifier[0] === "_" || identifier[len - 1] === "_");
            }
            /**
             * Check if identifier is a special case member expression
             * @param {string} identifier name of the node
             * @returns {boolean} true if its is a special case
             * @private
             */
            function isSpecialCaseIdentifierForMemberExpression(identifier) {
                return identifier === "__proto__";
            }
            /**
             * Check if identifier is a special case variable expression
             * @param {string} identifier name of the node
             * @returns {boolean} true if its is a special case
             * @private
             */
            function isSpecialCaseIdentifierInVariableExpression(identifier) {
                // Checks for the underscore library usage here
                return identifier === "_";
            }
            /**
             * Check if a node is a member reference of this.constructor
             * @param {ASTNode} node node to evaluate
             * @returns {boolean} true if it is a reference on this.constructor
             * @private
             */
            function isThisConstructorReference(node) {
                return node.object.type === "MemberExpression" &&
                    node.object.property.name === "constructor" &&
                    node.object.object.type === "ThisExpression";
            }
            /**
             * Check if function parameter has a dangling underscore.
             * @param {ASTNode} node function node to evaluate
             * @returns {void}
             * @private
             */
            function checkForDanglingUnderscoreInFunctionParameters(node) {
                if (!allowFunctionParams) {
                    node.params.forEach(param => {
                        const { type } = param;
                        let nodeToCheck;
                        if (type === "RestElement") {
                            nodeToCheck = param.argument;
                        }
                        else if (type === "AssignmentPattern") {
                            nodeToCheck = param.left;
                        }
                        else {
                            nodeToCheck = param;
                        }
                        if (nodeToCheck.type === "Identifier") {
                            const identifier = nodeToCheck.name;
                            if (hasDanglingUnderscore(identifier) && !isAllowed(identifier)) {
                                context.report({
                                    node: param,
                                    messageId: "unexpectedUnderscore",
                                    data: {
                                        identifier
                                    }
                                });
                            }
                        }
                    });
                }
            }
            /**
             * Check if function has a dangling underscore
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkForDanglingUnderscoreInFunction(node) {
                if (node.type === "FunctionDeclaration" && node.id) {
                    const identifier = node.id.name;
                    if (typeof identifier !== "undefined" && hasDanglingUnderscore(identifier) && !isAllowed(identifier)) {
                        context.report({
                            node,
                            messageId: "unexpectedUnderscore",
                            data: {
                                identifier
                            }
                        });
                    }
                }
                checkForDanglingUnderscoreInFunctionParameters(node);
            }
            /**
             * Check if variable expression has a dangling underscore
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkForDanglingUnderscoreInVariableExpression(node) {
                const identifier = node.id.name;
                if (typeof identifier !== "undefined" && hasDanglingUnderscore(identifier) &&
                    !isSpecialCaseIdentifierInVariableExpression(identifier) && !isAllowed(identifier)) {
                    context.report({
                        node,
                        messageId: "unexpectedUnderscore",
                        data: {
                            identifier
                        }
                    });
                }
            }
            /**
             * Check if member expression has a dangling underscore
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkForDanglingUnderscoreInMemberExpression(node) {
                const identifier = node.property.name, isMemberOfThis = node.object.type === "ThisExpression", isMemberOfSuper = node.object.type === "Super", isMemberOfThisConstructor = isThisConstructorReference(node);
                if (typeof identifier !== "undefined" && hasDanglingUnderscore(identifier) &&
                    !(isMemberOfThis && allowAfterThis) &&
                    !(isMemberOfSuper && allowAfterSuper) &&
                    !(isMemberOfThisConstructor && allowAfterThisConstructor) &&
                    !isSpecialCaseIdentifierForMemberExpression(identifier) && !isAllowed(identifier)) {
                    context.report({
                        node,
                        messageId: "unexpectedUnderscore",
                        data: {
                            identifier
                        }
                    });
                }
            }
            /**
             * Check if method declaration or method property has a dangling underscore
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkForDanglingUnderscoreInMethod(node) {
                const identifier = node.key.name;
                const isMethod = node.type === "MethodDefinition" || node.type === "Property" && node.method;
                if (typeof identifier !== "undefined" && enforceInMethodNames && isMethod && hasDanglingUnderscore(identifier) && !isAllowed(identifier)) {
                    context.report({
                        node,
                        messageId: "unexpectedUnderscore",
                        data: {
                            identifier
                        }
                    });
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                FunctionDeclaration: checkForDanglingUnderscoreInFunction,
                VariableDeclarator: checkForDanglingUnderscoreInVariableExpression,
                MemberExpression: checkForDanglingUnderscoreInMemberExpression,
                MethodDefinition: checkForDanglingUnderscoreInMethod,
                Property: checkForDanglingUnderscoreInMethod,
                FunctionExpression: checkForDanglingUnderscoreInFunction,
                ArrowFunctionExpression: checkForDanglingUnderscoreInFunction
            };
        }
    };
},
/* eslint/lib/rules/no-unexpected-multiline.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow confusing multiline expressions",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-unexpected-multiline"
            },
            schema: [],
            messages: {
                function: "Unexpected newline between function and ( of function call.",
                property: "Unexpected newline between object and [ of property access.",
                taggedTemplate: "Unexpected newline between template tag and template literal.",
                division: "Unexpected newline between numerator and division operator."
            }
        },
        create(context) {
            const REGEX_FLAG_MATCHER = /^[gimsuy]+$/u;
            const sourceCode = context.getSourceCode();
            /**
             * Check to see if there is a newline between the node and the following open bracket
             * line's expression
             * @param {ASTNode} node The node to check.
             * @param {string} messageId The error messageId to use.
             * @returns {void}
             * @private
             */
            function checkForBreakAfter(node, messageId) {
                const openParen = sourceCode.getTokenAfter(node, astUtils.isNotClosingParenToken);
                const nodeExpressionEnd = sourceCode.getTokenBefore(openParen);
                if (openParen.loc.start.line !== nodeExpressionEnd.loc.end.line) {
                    context.report({
                        node,
                        loc: openParen.loc,
                        messageId
                    });
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                MemberExpression(node) {
                    if (!node.computed || node.optional) {
                        return;
                    }
                    checkForBreakAfter(node.object, "property");
                },
                TaggedTemplateExpression(node) {
                    const { quasi } = node;
                    // handles common tags, parenthesized tags, and typescript's generic type arguments
                    const tokenBefore = sourceCode.getTokenBefore(quasi);
                    if (tokenBefore.loc.end.line !== quasi.loc.start.line) {
                        context.report({
                            node,
                            loc: {
                                start: quasi.loc.start,
                                end: {
                                    line: quasi.loc.start.line,
                                    column: quasi.loc.start.column + 1
                                }
                            },
                            messageId: "taggedTemplate"
                        });
                    }
                },
                CallExpression(node) {
                    if (node.arguments.length === 0 || node.optional) {
                        return;
                    }
                    checkForBreakAfter(node.callee, "function");
                },
                "BinaryExpression[operator='/'] > BinaryExpression[operator='/'].left"(node) {
                    const secondSlash = sourceCode.getTokenAfter(node, token => token.value === "/");
                    const tokenAfterOperator = sourceCode.getTokenAfter(secondSlash);
                    if (tokenAfterOperator.type === "Identifier" &&
                        REGEX_FLAG_MATCHER.test(tokenAfterOperator.value) &&
                        secondSlash.range[1] === tokenAfterOperator.range[0]) {
                        checkForBreakAfter(node.left, "division");
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-unmodified-loop-condition.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const Traverser = require(347) /* ../shared/traverser */, astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const SENTINEL_PATTERN = /(?:(?:Call|Class|Function|Member|New|Yield)Expression|Statement|Declaration)$/u;
    const LOOP_PATTERN = /^(?:DoWhile|For|While)Statement$/u; // for-in/of statements don't have `test` property.
    const GROUP_PATTERN = /^(?:BinaryExpression|ConditionalExpression)$/u;
    const SKIP_PATTERN = /^(?:ArrowFunction|Class|Function)Expression$/u;
    const DYNAMIC_PATTERN = /^(?:Call|Member|New|TaggedTemplate|Yield)Expression$/u;
    /**
     * @typedef {Object} LoopConditionInfo
     * @property {eslint-scope.Reference} reference - The reference.
     * @property {ASTNode} group - BinaryExpression or ConditionalExpression nodes
     *      that the reference is belonging to.
     * @property {Function} isInLoop - The predicate which checks a given reference
     *      is in this loop.
     * @property {boolean} modified - The flag that the reference is modified in
     *      this loop.
     */
    /**
     * Checks whether or not a given reference is a write reference.
     * @param {eslint-scope.Reference} reference A reference to check.
     * @returns {boolean} `true` if the reference is a write reference.
     */
    function isWriteReference(reference) {
        if (reference.init) {
            const def = reference.resolved && reference.resolved.defs[0];
            if (!def || def.type !== "Variable" || def.parent.kind !== "var") {
                return false;
            }
        }
        return reference.isWrite();
    }
    /**
     * Checks whether or not a given loop condition info does not have the modified
     * flag.
     * @param {LoopConditionInfo} condition A loop condition info to check.
     * @returns {boolean} `true` if the loop condition info is "unmodified".
     */
    function isUnmodified(condition) {
        return !condition.modified;
    }
    /**
     * Checks whether or not a given loop condition info does not have the modified
     * flag and does not have the group this condition belongs to.
     * @param {LoopConditionInfo} condition A loop condition info to check.
     * @returns {boolean} `true` if the loop condition info is "unmodified".
     */
    function isUnmodifiedAndNotBelongToGroup(condition) {
        return !(condition.modified || condition.group);
    }
    /**
     * Checks whether or not a given reference is inside of a given node.
     * @param {ASTNode} node A node to check.
     * @param {eslint-scope.Reference} reference A reference to check.
     * @returns {boolean} `true` if the reference is inside of the node.
     */
    function isInRange(node, reference) {
        const or = node.range;
        const ir = reference.identifier.range;
        return or[0] <= ir[0] && ir[1] <= or[1];
    }
    /**
     * Checks whether or not a given reference is inside of a loop node's condition.
     * @param {ASTNode} node A node to check.
     * @param {eslint-scope.Reference} reference A reference to check.
     * @returns {boolean} `true` if the reference is inside of the loop node's
     *      condition.
     */
    const isInLoop = {
        WhileStatement: isInRange,
        DoWhileStatement: isInRange,
        ForStatement(node, reference) {
            return (isInRange(node, reference) &&
                !(node.init && isInRange(node.init, reference)));
        }
    };
    /**
     * Gets the function which encloses a given reference.
     * This supports only FunctionDeclaration.
     * @param {eslint-scope.Reference} reference A reference to get.
     * @returns {ASTNode|null} The function node or null.
     */
    function getEncloseFunctionDeclaration(reference) {
        let node = reference.identifier;
        while (node) {
            if (node.type === "FunctionDeclaration") {
                return node.id ? node : null;
            }
            node = node.parent;
        }
        return null;
    }
    /**
     * Updates the "modified" flags of given loop conditions with given modifiers.
     * @param {LoopConditionInfo[]} conditions The loop conditions to be updated.
     * @param {eslint-scope.Reference[]} modifiers The references to update.
     * @returns {void}
     */
    function updateModifiedFlag(conditions, modifiers) {
        for (let i = 0; i < conditions.length; ++i) {
            const condition = conditions[i];
            for (let j = 0; !condition.modified && j < modifiers.length; ++j) {
                const modifier = modifiers[j];
                let funcNode, funcVar;
                /*
                 * Besides checking for the condition being in the loop, we want to
                 * check the function that this modifier is belonging to is called
                 * in the loop.
                 * FIXME: This should probably be extracted to a function.
                 */
                const inLoop = condition.isInLoop(modifier) || Boolean((funcNode = getEncloseFunctionDeclaration(modifier)) &&
                    (funcVar = astUtils.getVariableByName(modifier.from.upper, funcNode.id.name)) &&
                    funcVar.references.some(condition.isInLoop));
                condition.modified = inLoop;
            }
        }
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow unmodified loop conditions",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-unmodified-loop-condition"
            },
            schema: [],
            messages: {
                loopConditionNotModified: "'{{name}}' is not modified in this loop."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            let groupMap = null;
            /**
             * Reports a given condition info.
             * @param {LoopConditionInfo} condition A loop condition info to report.
             * @returns {void}
             */
            function report(condition) {
                const node = condition.reference.identifier;
                context.report({
                    node,
                    messageId: "loopConditionNotModified",
                    data: node
                });
            }
            /**
             * Registers given conditions to the group the condition belongs to.
             * @param {LoopConditionInfo[]} conditions A loop condition info to
             *      register.
             * @returns {void}
             */
            function registerConditionsToGroup(conditions) {
                for (let i = 0; i < conditions.length; ++i) {
                    const condition = conditions[i];
                    if (condition.group) {
                        let group = groupMap.get(condition.group);
                        if (!group) {
                            group = [];
                            groupMap.set(condition.group, group);
                        }
                        group.push(condition);
                    }
                }
            }
            /**
             * Reports references which are inside of unmodified groups.
             * @param {LoopConditionInfo[]} conditions A loop condition info to report.
             * @returns {void}
             */
            function checkConditionsInGroup(conditions) {
                if (conditions.every(isUnmodified)) {
                    conditions.forEach(report);
                }
            }
            /**
             * Checks whether or not a given group node has any dynamic elements.
             * @param {ASTNode} root A node to check.
             *      This node is one of BinaryExpression or ConditionalExpression.
             * @returns {boolean} `true` if the node is dynamic.
             */
            function hasDynamicExpressions(root) {
                let retv = false;
                Traverser.traverse(root, {
                    visitorKeys: sourceCode.visitorKeys,
                    enter(node) {
                        if (DYNAMIC_PATTERN.test(node.type)) {
                            retv = true;
                            this.break();
                        }
                        else if (SKIP_PATTERN.test(node.type)) {
                            this.skip();
                        }
                    }
                });
                return retv;
            }
            /**
             * Creates the loop condition information from a given reference.
             * @param {eslint-scope.Reference} reference A reference to create.
             * @returns {LoopConditionInfo|null} Created loop condition info, or null.
             */
            function toLoopCondition(reference) {
                if (reference.init) {
                    return null;
                }
                let group = null;
                let child = reference.identifier;
                let node = child.parent;
                while (node) {
                    if (SENTINEL_PATTERN.test(node.type)) {
                        if (LOOP_PATTERN.test(node.type) && node.test === child) {
                            // This reference is inside of a loop condition.
                            return {
                                reference,
                                group,
                                isInLoop: isInLoop[node.type].bind(null, node),
                                modified: false
                            };
                        }
                        // This reference is outside of a loop condition.
                        break;
                    }
                    /*
                     * If it's inside of a group, OK if either operand is modified.
                     * So stores the group this reference belongs to.
                     */
                    if (GROUP_PATTERN.test(node.type)) {
                        // If this expression is dynamic, no need to check.
                        if (hasDynamicExpressions(node)) {
                            break;
                        }
                        else {
                            group = node;
                        }
                    }
                    child = node;
                    node = node.parent;
                }
                return null;
            }
            /**
             * Finds unmodified references which are inside of a loop condition.
             * Then reports the references which are outside of groups.
             * @param {eslint-scope.Variable} variable A variable to report.
             * @returns {void}
             */
            function checkReferences(variable) {
                // Gets references that exist in loop conditions.
                const conditions = variable
                    .references
                    .map(toLoopCondition)
                    .filter(Boolean);
                if (conditions.length === 0) {
                    return;
                }
                // Registers the conditions to belonging groups.
                registerConditionsToGroup(conditions);
                // Check the conditions are modified.
                const modifiers = variable.references.filter(isWriteReference);
                if (modifiers.length > 0) {
                    updateModifiedFlag(conditions, modifiers);
                }
                /*
                 * Reports the conditions which are not belonging to groups.
                 * Others will be reported after all variables are done.
                 */
                conditions
                    .filter(isUnmodifiedAndNotBelongToGroup)
                    .forEach(report);
            }
            return {
                "Program:exit"() {
                    const queue = [context.getScope()];
                    groupMap = new Map();
                    let scope;
                    while ((scope = queue.pop())) {
                        queue.push(...scope.childScopes);
                        scope.variables.forEach(checkReferences);
                    }
                    groupMap.forEach(checkConditionsInGroup);
                    groupMap = null;
                }
            };
        }
    };
},
/* eslint/lib/rules/no-unneeded-ternary.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    // Operators that always result in a boolean value
    const BOOLEAN_OPERATORS = new Set(["==", "===", "!=", "!==", ">", ">=", "<", "<=", "in", "instanceof"]);
    const OPERATOR_INVERSES = {
        "==": "!=",
        "!=": "==",
        "===": "!==",
        "!==": "==="
        // Operators like < and >= are not true inverses, since both will return false with NaN.
    };
    const OR_PRECEDENCE = astUtils.getPrecedence({ type: "LogicalExpression", operator: "||" });
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow ternary operators when simpler alternatives exist",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-unneeded-ternary"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        defaultAssignment: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "code",
            messages: {
                unnecessaryConditionalExpression: "Unnecessary use of boolean literals in conditional expression.",
                unnecessaryConditionalAssignment: "Unnecessary use of conditional expression for default assignment."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const defaultAssignment = options.defaultAssignment !== false;
            const sourceCode = context.getSourceCode();
            /**
             * Test if the node is a boolean literal
             * @param {ASTNode} node The node to report.
             * @returns {boolean} True if the its a boolean literal
             * @private
             */
            function isBooleanLiteral(node) {
                return node.type === "Literal" && typeof node.value === "boolean";
            }
            /**
             * Creates an expression that represents the boolean inverse of the expression represented by the original node
             * @param {ASTNode} node A node representing an expression
             * @returns {string} A string representing an inverted expression
             */
            function invertExpression(node) {
                if (node.type === "BinaryExpression" && Object.prototype.hasOwnProperty.call(OPERATOR_INVERSES, node.operator)) {
                    const operatorToken = sourceCode.getFirstTokenBetween(node.left, node.right, token => token.value === node.operator);
                    const text = sourceCode.getText();
                    return text.slice(node.range[0], operatorToken.range[0]) + OPERATOR_INVERSES[node.operator] + text.slice(operatorToken.range[1], node.range[1]);
                }
                if (astUtils.getPrecedence(node) < astUtils.getPrecedence({ type: "UnaryExpression" })) {
                    return `!(${astUtils.getParenthesisedText(sourceCode, node)})`;
                }
                return `!${astUtils.getParenthesisedText(sourceCode, node)}`;
            }
            /**
             * Tests if a given node always evaluates to a boolean value
             * @param {ASTNode} node An expression node
             * @returns {boolean} True if it is determined that the node will always evaluate to a boolean value
             */
            function isBooleanExpression(node) {
                return node.type === "BinaryExpression" && BOOLEAN_OPERATORS.has(node.operator) ||
                    node.type === "UnaryExpression" && node.operator === "!";
            }
            /**
             * Test if the node matches the pattern id ? id : expression
             * @param {ASTNode} node The ConditionalExpression to check.
             * @returns {boolean} True if the pattern is matched, and false otherwise
             * @private
             */
            function matchesDefaultAssignment(node) {
                return node.test.type === "Identifier" &&
                    node.consequent.type === "Identifier" &&
                    node.test.name === node.consequent.name;
            }
            return {
                ConditionalExpression(node) {
                    if (isBooleanLiteral(node.alternate) && isBooleanLiteral(node.consequent)) {
                        context.report({
                            node,
                            messageId: "unnecessaryConditionalExpression",
                            fix(fixer) {
                                if (node.consequent.value === node.alternate.value) {
                                    // Replace `foo ? true : true` with just `true`, but don't replace `foo() ? true : true`
                                    return node.test.type === "Identifier" ? fixer.replaceText(node, node.consequent.value.toString()) : null;
                                }
                                if (node.alternate.value) {
                                    // Replace `foo() ? false : true` with `!(foo())`
                                    return fixer.replaceText(node, invertExpression(node.test));
                                }
                                // Replace `foo ? true : false` with `foo` if `foo` is guaranteed to be a boolean, or `!!foo` otherwise.
                                return fixer.replaceText(node, isBooleanExpression(node.test) ? astUtils.getParenthesisedText(sourceCode, node.test) : `!${invertExpression(node.test)}`);
                            }
                        });
                    }
                    else if (!defaultAssignment && matchesDefaultAssignment(node)) {
                        context.report({
                            node,
                            messageId: "unnecessaryConditionalAssignment",
                            fix: fixer => {
                                const shouldParenthesizeAlternate = (astUtils.getPrecedence(node.alternate) < OR_PRECEDENCE ||
                                    astUtils.isCoalesceExpression(node.alternate)) &&
                                    !astUtils.isParenthesised(sourceCode, node.alternate);
                                const alternateText = shouldParenthesizeAlternate
                                    ? `(${sourceCode.getText(node.alternate)})`
                                    : astUtils.getParenthesisedText(sourceCode, node.alternate);
                                const testText = astUtils.getParenthesisedText(sourceCode, node.test);
                                return fixer.replaceText(node, `${testText} || ${alternateText}`);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-unreachable.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given variable declarator has the initializer.
     * @param {ASTNode} node A VariableDeclarator node to check.
     * @returns {boolean} `true` if the node has the initializer.
     */
    function isInitialized(node) {
        return Boolean(node.init);
    }
    /**
     * Checks whether or not a given code path segment is unreachable.
     * @param {CodePathSegment} segment A CodePathSegment to check.
     * @returns {boolean} `true` if the segment is unreachable.
     */
    function isUnreachable(segment) {
        return !segment.reachable;
    }
    /**
     * The class to distinguish consecutive unreachable statements.
     */
    class ConsecutiveRange {
        constructor(sourceCode) {
            this.sourceCode = sourceCode;
            this.startNode = null;
            this.endNode = null;
        }
        /**
         * The location object of this range.
         * @type {Object}
         */
        get location() {
            return {
                start: this.startNode.loc.start,
                end: this.endNode.loc.end
            };
        }
        /**
         * `true` if this range is empty.
         * @type {boolean}
         */
        get isEmpty() {
            return !(this.startNode && this.endNode);
        }
        /**
         * Checks whether the given node is inside of this range.
         * @param {ASTNode|Token} node The node to check.
         * @returns {boolean} `true` if the node is inside of this range.
         */
        contains(node) {
            return (node.range[0] >= this.startNode.range[0] &&
                node.range[1] <= this.endNode.range[1]);
        }
        /**
         * Checks whether the given node is consecutive to this range.
         * @param {ASTNode} node The node to check.
         * @returns {boolean} `true` if the node is consecutive to this range.
         */
        isConsecutive(node) {
            return this.contains(this.sourceCode.getTokenBefore(node));
        }
        /**
         * Merges the given node to this range.
         * @param {ASTNode} node The node to merge.
         * @returns {void}
         */
        merge(node) {
            this.endNode = node;
        }
        /**
         * Resets this range by the given node or null.
         * @param {ASTNode|null} node The node to reset, or null.
         * @returns {void}
         */
        reset(node) {
            this.startNode = this.endNode = node;
        }
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow unreachable code after `return`, `throw`, `continue`, and `break` statements",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-unreachable"
            },
            schema: [],
            messages: {
                unreachableCode: "Unreachable code."
            }
        },
        create(context) {
            let currentCodePath = null;
            const range = new ConsecutiveRange(context.getSourceCode());
            /**
             * Reports a given node if it's unreachable.
             * @param {ASTNode} node A statement node to report.
             * @returns {void}
             */
            function reportIfUnreachable(node) {
                let nextNode = null;
                if (node && currentCodePath.currentSegments.every(isUnreachable)) {
                    // Store this statement to distinguish consecutive statements.
                    if (range.isEmpty) {
                        range.reset(node);
                        return;
                    }
                    // Skip if this statement is inside of the current range.
                    if (range.contains(node)) {
                        return;
                    }
                    // Merge if this statement is consecutive to the current range.
                    if (range.isConsecutive(node)) {
                        range.merge(node);
                        return;
                    }
                    nextNode = node;
                }
                /*
                 * Report the current range since this statement is reachable or is
                 * not consecutive to the current range.
                 */
                if (!range.isEmpty) {
                    context.report({
                        messageId: "unreachableCode",
                        loc: range.location,
                        node: range.startNode
                    });
                }
                // Update the current range.
                range.reset(nextNode);
            }
            return {
                // Manages the current code path.
                onCodePathStart(codePath) {
                    currentCodePath = codePath;
                },
                onCodePathEnd() {
                    currentCodePath = currentCodePath.upper;
                },
                // Registers for all statement nodes (excludes FunctionDeclaration).
                BlockStatement: reportIfUnreachable,
                BreakStatement: reportIfUnreachable,
                ClassDeclaration: reportIfUnreachable,
                ContinueStatement: reportIfUnreachable,
                DebuggerStatement: reportIfUnreachable,
                DoWhileStatement: reportIfUnreachable,
                ExpressionStatement: reportIfUnreachable,
                ForInStatement: reportIfUnreachable,
                ForOfStatement: reportIfUnreachable,
                ForStatement: reportIfUnreachable,
                IfStatement: reportIfUnreachable,
                ImportDeclaration: reportIfUnreachable,
                LabeledStatement: reportIfUnreachable,
                ReturnStatement: reportIfUnreachable,
                SwitchStatement: reportIfUnreachable,
                ThrowStatement: reportIfUnreachable,
                TryStatement: reportIfUnreachable,
                VariableDeclaration(node) {
                    if (node.kind !== "var" || node.declarations.some(isInitialized)) {
                        reportIfUnreachable(node);
                    }
                },
                WhileStatement: reportIfUnreachable,
                WithStatement: reportIfUnreachable,
                ExportNamedDeclaration: reportIfUnreachable,
                ExportDefaultDeclaration: reportIfUnreachable,
                ExportAllDeclaration: reportIfUnreachable,
                "Program:exit"() {
                    reportIfUnreachable();
                }
            };
        }
    };
},
/* eslint/lib/rules/no-unreachable-loop.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const allLoopTypes = ["WhileStatement", "DoWhileStatement", "ForStatement", "ForInStatement", "ForOfStatement"];
    /**
     * Determines whether the given node is the first node in the code path to which a loop statement
     * 'loops' for the next iteration.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is a looping target.
     */
    function isLoopingTarget(node) {
        const parent = node.parent;
        if (parent) {
            switch (parent.type) {
                case "WhileStatement":
                    return node === parent.test;
                case "DoWhileStatement":
                    return node === parent.body;
                case "ForStatement":
                    return node === (parent.update || parent.test || parent.body);
                case "ForInStatement":
                case "ForOfStatement":
                    return node === parent.left;
                // no default
            }
        }
        return false;
    }
    /**
     * Creates an array with elements from the first given array that are not included in the second given array.
     * @param {Array} arrA The array to compare from.
     * @param {Array} arrB The array to compare against.
     * @returns {Array} a new array that represents `arrA \ arrB`.
     */
    function getDifference(arrA, arrB) {
        return arrA.filter(a => !arrB.includes(a));
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow loops with a body that allows only one iteration",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-unreachable-loop"
            },
            schema: [{
                    type: "object",
                    properties: {
                        ignore: {
                            type: "array",
                            items: {
                                enum: allLoopTypes
                            },
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }],
            messages: {
                invalid: "Invalid loop. Its body allows only one iteration."
            }
        },
        create(context) {
            const ignoredLoopTypes = context.options[0] && context.options[0].ignore || [], loopTypesToCheck = getDifference(allLoopTypes, ignoredLoopTypes), loopSelector = loopTypesToCheck.join(","), loopsByTargetSegments = new Map(), loopsToReport = new Set();
            let currentCodePath = null;
            return {
                onCodePathStart(codePath) {
                    currentCodePath = codePath;
                },
                onCodePathEnd() {
                    currentCodePath = currentCodePath.upper;
                },
                [loopSelector](node) {
                    /**
                     * Ignore unreachable loop statements to avoid unnecessary complexity in the implementation, or false positives otherwise.
                     * For unreachable segments, the code path analysis does not raise events required for this implementation.
                     */
                    if (currentCodePath.currentSegments.some(segment => segment.reachable)) {
                        loopsToReport.add(node);
                    }
                },
                onCodePathSegmentStart(segment, node) {
                    if (isLoopingTarget(node)) {
                        const loop = node.parent;
                        loopsByTargetSegments.set(segment, loop);
                    }
                },
                onCodePathSegmentLoop(_, toSegment, node) {
                    const loop = loopsByTargetSegments.get(toSegment);
                    /**
                     * The second iteration is reachable, meaning that the loop is valid by the logic of this rule,
                     * only if there is at least one loop event with the appropriate target (which has been already
                     * determined in the `loopsByTargetSegments` map), raised from either:
                     *
                     * - the end of the loop's body (in which case `node === loop`)
                     * - a `continue` statement
                     *
                     * This condition skips loop events raised from `ForInStatement > .right` and `ForOfStatement > .right` nodes.
                     */
                    if (node === loop || node.type === "ContinueStatement") {
                        // Removes loop if it exists in the set. Otherwise, `Set#delete` has no effect and doesn't throw.
                        loopsToReport.delete(loop);
                    }
                },
                "Program:exit"() {
                    loopsToReport.forEach(node => context.report({ node, messageId: "invalid" }));
                }
            };
        }
    };
},
/* eslint/lib/rules/no-unsafe-finally.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const SENTINEL_NODE_TYPE_RETURN_THROW = /^(?:Program|(?:Function|Class)(?:Declaration|Expression)|ArrowFunctionExpression)$/u;
    const SENTINEL_NODE_TYPE_BREAK = /^(?:Program|(?:Function|Class)(?:Declaration|Expression)|ArrowFunctionExpression|DoWhileStatement|WhileStatement|ForOfStatement|ForInStatement|ForStatement|SwitchStatement)$/u;
    const SENTINEL_NODE_TYPE_CONTINUE = /^(?:Program|(?:Function|Class)(?:Declaration|Expression)|ArrowFunctionExpression|DoWhileStatement|WhileStatement|ForOfStatement|ForInStatement|ForStatement)$/u;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow control flow statements in `finally` blocks",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-unsafe-finally"
            },
            schema: [],
            messages: {
                unsafeUsage: "Unsafe usage of {{nodeType}}."
            }
        },
        create(context) {
            /**
             * Checks if the node is the finalizer of a TryStatement
             * @param {ASTNode} node node to check.
             * @returns {boolean} - true if the node is the finalizer of a TryStatement
             */
            function isFinallyBlock(node) {
                return node.parent.type === "TryStatement" && node.parent.finalizer === node;
            }
            /**
             * Climbs up the tree if the node is not a sentinel node
             * @param {ASTNode} node node to check.
             * @param {string} label label of the break or continue statement
             * @returns {boolean} - return whether the node is a finally block or a sentinel node
             */
            function isInFinallyBlock(node, label) {
                let labelInside = false;
                let sentinelNodeType;
                if (node.type === "BreakStatement" && !node.label) {
                    sentinelNodeType = SENTINEL_NODE_TYPE_BREAK;
                }
                else if (node.type === "ContinueStatement") {
                    sentinelNodeType = SENTINEL_NODE_TYPE_CONTINUE;
                }
                else {
                    sentinelNodeType = SENTINEL_NODE_TYPE_RETURN_THROW;
                }
                for (let currentNode = node; currentNode && !sentinelNodeType.test(currentNode.type); currentNode = currentNode.parent) {
                    if (currentNode.parent.label && label && (currentNode.parent.label.name === label.name)) {
                        labelInside = true;
                    }
                    if (isFinallyBlock(currentNode)) {
                        if (label && labelInside) {
                            return false;
                        }
                        return true;
                    }
                }
                return false;
            }
            /**
             * Checks whether the possibly-unsafe statement is inside a finally block.
             * @param {ASTNode} node node to check.
             * @returns {void}
             */
            function check(node) {
                if (isInFinallyBlock(node, node.label)) {
                    context.report({
                        messageId: "unsafeUsage",
                        data: {
                            nodeType: node.type
                        },
                        node,
                        line: node.loc.line,
                        column: node.loc.column
                    });
                }
            }
            return {
                ReturnStatement: check,
                ThrowStatement: check,
                BreakStatement: check,
                ContinueStatement: check
            };
        }
    };
},
/* eslint/lib/rules/no-unsafe-negation.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether the given operator is `in` or `instanceof`
     * @param {string} op The operator type to check.
     * @returns {boolean} `true` if the operator is `in` or `instanceof`
     */
    function isInOrInstanceOfOperator(op) {
        return op === "in" || op === "instanceof";
    }
    /**
     * Checks whether the given operator is an ordering relational operator or not.
     * @param {string} op The operator type to check.
     * @returns {boolean} `true` if the operator is an ordering relational operator.
     */
    function isOrderingRelationalOperator(op) {
        return op === "<" || op === ">" || op === ">=" || op === "<=";
    }
    /**
     * Checks whether the given node is a logical negation expression or not.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is a logical negation expression.
     */
    function isNegation(node) {
        return node.type === "UnaryExpression" && node.operator === "!";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow negating the left operand of relational operators",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-unsafe-negation",
                suggestion: true
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        enforceForOrderingRelations: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: null,
            messages: {
                unexpected: "Unexpected negating the left operand of '{{operator}}' operator.",
                suggestNegatedExpression: "Negate '{{operator}}' expression instead of its left operand. This changes the current behavior.",
                suggestParenthesisedNegation: "Wrap negation in '()' to make the intention explicit. This preserves the current behavior."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const options = context.options[0] || {};
            const enforceForOrderingRelations = options.enforceForOrderingRelations === true;
            return {
                BinaryExpression(node) {
                    const operator = node.operator;
                    const orderingRelationRuleApplies = enforceForOrderingRelations && isOrderingRelationalOperator(operator);
                    if ((isInOrInstanceOfOperator(operator) || orderingRelationRuleApplies) &&
                        isNegation(node.left) &&
                        !astUtils.isParenthesised(sourceCode, node.left)) {
                        context.report({
                            node,
                            loc: node.left.loc,
                            messageId: "unexpected",
                            data: { operator },
                            suggest: [
                                {
                                    messageId: "suggestNegatedExpression",
                                    data: { operator },
                                    fix(fixer) {
                                        const negationToken = sourceCode.getFirstToken(node.left);
                                        const fixRange = [negationToken.range[1], node.range[1]];
                                        const text = sourceCode.text.slice(fixRange[0], fixRange[1]);
                                        return fixer.replaceTextRange(fixRange, `(${text})`);
                                    }
                                },
                                {
                                    messageId: "suggestParenthesisedNegation",
                                    fix(fixer) {
                                        return fixer.replaceText(node.left, `(${sourceCode.getText(node.left)})`);
                                    }
                                }
                            ]
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-unsafe-optional-chaining.js */ function _(require, module, exports, __esModule, __esExport) {
    const UNSAFE_ARITHMETIC_OPERATORS = new Set(["+", "-", "/", "*", "%", "**"]);
    const UNSAFE_ASSIGNMENT_OPERATORS = new Set(["+=", "-=", "/=", "*=", "%=", "**="]);
    const UNSAFE_RELATIONAL_OPERATORS = new Set(["in", "instanceof"]);
    /**
     * Checks whether a node is a destructuring pattern or not
     * @param {ASTNode} node node to check
     * @returns {boolean} `true` if a node is a destructuring pattern, otherwise `false`
     */
    function isDestructuringPattern(node) {
        return node.type === "ObjectPattern" || node.type === "ArrayPattern";
    }
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow use of optional chaining in contexts where the `undefined` value is not allowed",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-unsafe-optional-chaining"
            },
            schema: [{
                    type: "object",
                    properties: {
                        disallowArithmeticOperators: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }],
            fixable: null,
            messages: {
                unsafeOptionalChain: "Unsafe usage of optional chaining. If it short-circuits with 'undefined' the evaluation will throw TypeError.",
                unsafeArithmetic: "Unsafe arithmetic operation on optional chaining. It can result in NaN."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const disallowArithmeticOperators = (options.disallowArithmeticOperators) || false;
            /**
             * Reports unsafe usage of optional chaining
             * @param {ASTNode} node node to report
             * @returns {void}
             */
            function reportUnsafeUsage(node) {
                context.report({
                    messageId: "unsafeOptionalChain",
                    node
                });
            }
            /**
             * Reports unsafe arithmetic operation on optional chaining
             * @param {ASTNode} node node to report
             * @returns {void}
             */
            function reportUnsafeArithmetic(node) {
                context.report({
                    messageId: "unsafeArithmetic",
                    node
                });
            }
            /**
             * Checks and reports if a node can short-circuit with `undefined` by optional chaining.
             * @param {ASTNode} [node] node to check
             * @param {Function} reportFunc report function
             * @returns {void}
             */
            function checkUndefinedShortCircuit(node, reportFunc) {
                if (!node) {
                    return;
                }
                switch (node.type) {
                    case "LogicalExpression":
                        if (node.operator === "||" || node.operator === "??") {
                            checkUndefinedShortCircuit(node.right, reportFunc);
                        }
                        else if (node.operator === "&&") {
                            checkUndefinedShortCircuit(node.left, reportFunc);
                            checkUndefinedShortCircuit(node.right, reportFunc);
                        }
                        break;
                    case "SequenceExpression":
                        checkUndefinedShortCircuit(node.expressions[node.expressions.length - 1], reportFunc);
                        break;
                    case "ConditionalExpression":
                        checkUndefinedShortCircuit(node.consequent, reportFunc);
                        checkUndefinedShortCircuit(node.alternate, reportFunc);
                        break;
                    case "AwaitExpression":
                        checkUndefinedShortCircuit(node.argument, reportFunc);
                        break;
                    case "ChainExpression":
                        reportFunc(node);
                        break;
                    default:
                        break;
                }
            }
            /**
             * Checks unsafe usage of optional chaining
             * @param {ASTNode} node node to check
             * @returns {void}
             */
            function checkUnsafeUsage(node) {
                checkUndefinedShortCircuit(node, reportUnsafeUsage);
            }
            /**
             * Checks unsafe arithmetic operations on optional chaining
             * @param {ASTNode} node node to check
             * @returns {void}
             */
            function checkUnsafeArithmetic(node) {
                checkUndefinedShortCircuit(node, reportUnsafeArithmetic);
            }
            return {
                "AssignmentExpression, AssignmentPattern"(node) {
                    if (isDestructuringPattern(node.left)) {
                        checkUnsafeUsage(node.right);
                    }
                },
                "ClassDeclaration, ClassExpression"(node) {
                    checkUnsafeUsage(node.superClass);
                },
                CallExpression(node) {
                    if (!node.optional) {
                        checkUnsafeUsage(node.callee);
                    }
                },
                NewExpression(node) {
                    checkUnsafeUsage(node.callee);
                },
                VariableDeclarator(node) {
                    if (isDestructuringPattern(node.id)) {
                        checkUnsafeUsage(node.init);
                    }
                },
                MemberExpression(node) {
                    if (!node.optional) {
                        checkUnsafeUsage(node.object);
                    }
                },
                TaggedTemplateExpression(node) {
                    checkUnsafeUsage(node.tag);
                },
                ForOfStatement(node) {
                    checkUnsafeUsage(node.right);
                },
                SpreadElement(node) {
                    if (node.parent && node.parent.type !== "ObjectExpression") {
                        checkUnsafeUsage(node.argument);
                    }
                },
                BinaryExpression(node) {
                    if (UNSAFE_RELATIONAL_OPERATORS.has(node.operator)) {
                        checkUnsafeUsage(node.right);
                    }
                    if (disallowArithmeticOperators &&
                        UNSAFE_ARITHMETIC_OPERATORS.has(node.operator)) {
                        checkUnsafeArithmetic(node.right);
                        checkUnsafeArithmetic(node.left);
                    }
                },
                WithStatement(node) {
                    checkUnsafeUsage(node.object);
                },
                UnaryExpression(node) {
                    if (disallowArithmeticOperators &&
                        UNSAFE_ARITHMETIC_OPERATORS.has(node.operator)) {
                        checkUnsafeArithmetic(node.argument);
                    }
                },
                AssignmentExpression(node) {
                    if (disallowArithmeticOperators &&
                        UNSAFE_ASSIGNMENT_OPERATORS.has(node.operator)) {
                        checkUnsafeArithmetic(node.right);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-unused-expressions.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    /**
     * Returns `true`.
     * @returns {boolean} `true`.
     */
    function alwaysTrue() {
        return true;
    }
    /**
     * Returns `false`.
     * @returns {boolean} `false`.
     */
    function alwaysFalse() {
        return false;
    }
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unused expressions",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-unused-expressions"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowShortCircuit: {
                            type: "boolean",
                            default: false
                        },
                        allowTernary: {
                            type: "boolean",
                            default: false
                        },
                        allowTaggedTemplates: {
                            type: "boolean",
                            default: false
                        },
                        enforceForJSX: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unusedExpression: "Expected an assignment or function call and instead saw an expression."
            }
        },
        create(context) {
            const config = context.options[0] || {}, allowShortCircuit = config.allowShortCircuit || false, allowTernary = config.allowTernary || false, allowTaggedTemplates = config.allowTaggedTemplates || false, enforceForJSX = config.enforceForJSX || false;
            // eslint-disable-next-line jsdoc/require-description
            /**
             * @param {ASTNode} node any node
             * @returns {boolean} whether the given node structurally represents a directive
             */
            function looksLikeDirective(node) {
                return node.type === "ExpressionStatement" &&
                    node.expression.type === "Literal" && typeof node.expression.value === "string";
            }
            // eslint-disable-next-line jsdoc/require-description
            /**
             * @param {Function} predicate ([a] -> Boolean) the function used to make the determination
             * @param {a[]} list the input list
             * @returns {a[]} the leading sequence of members in the given list that pass the given predicate
             */
            function takeWhile(predicate, list) {
                for (let i = 0; i < list.length; ++i) {
                    if (!predicate(list[i])) {
                        return list.slice(0, i);
                    }
                }
                return list.slice();
            }
            // eslint-disable-next-line jsdoc/require-description
            /**
             * @param {ASTNode} node a Program or BlockStatement node
             * @returns {ASTNode[]} the leading sequence of directive nodes in the given node's body
             */
            function directives(node) {
                return takeWhile(looksLikeDirective, node.body);
            }
            // eslint-disable-next-line jsdoc/require-description
            /**
             * @param {ASTNode} node any node
             * @param {ASTNode[]} ancestors the given node's ancestors
             * @returns {boolean} whether the given node is considered a directive in its current position
             */
            function isDirective(node, ancestors) {
                const parent = ancestors[ancestors.length - 1], grandparent = ancestors[ancestors.length - 2];
                return (parent.type === "Program" || parent.type === "BlockStatement" &&
                    (/Function/u.test(grandparent.type))) &&
                    directives(parent).indexOf(node) >= 0;
            }
            /**
             * The member functions return `true` if the type has no side-effects.
             * Unknown nodes are handled as `false`, then this rule ignores those.
             */
            const Checker = Object.assign(Object.create(null), {
                isDisallowed(node) {
                    return (Checker[node.type] || alwaysFalse)(node);
                },
                ArrayExpression: alwaysTrue,
                ArrowFunctionExpression: alwaysTrue,
                BinaryExpression: alwaysTrue,
                ChainExpression(node) {
                    return Checker.isDisallowed(node.expression);
                },
                ClassExpression: alwaysTrue,
                ConditionalExpression(node) {
                    if (allowTernary) {
                        return Checker.isDisallowed(node.consequent) || Checker.isDisallowed(node.alternate);
                    }
                    return true;
                },
                FunctionExpression: alwaysTrue,
                Identifier: alwaysTrue,
                JSXElement() {
                    return enforceForJSX;
                },
                JSXFragment() {
                    return enforceForJSX;
                },
                Literal: alwaysTrue,
                LogicalExpression(node) {
                    if (allowShortCircuit) {
                        return Checker.isDisallowed(node.right);
                    }
                    return true;
                },
                MemberExpression: alwaysTrue,
                MetaProperty: alwaysTrue,
                ObjectExpression: alwaysTrue,
                SequenceExpression: alwaysTrue,
                TaggedTemplateExpression() {
                    return !allowTaggedTemplates;
                },
                TemplateLiteral: alwaysTrue,
                ThisExpression: alwaysTrue,
                UnaryExpression(node) {
                    return node.operator !== "void" && node.operator !== "delete";
                }
            });
            return {
                ExpressionStatement(node) {
                    if (Checker.isDisallowed(node.expression) && !isDirective(node, context.getAncestors())) {
                        context.report({ node, messageId: "unusedExpression" });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-unused-labels.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unused labels",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-unused-labels"
            },
            schema: [],
            fixable: "code",
            messages: {
                unused: "'{{name}}:' is defined but never used."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            let scopeInfo = null;
            /**
             * Adds a scope info to the stack.
             * @param {ASTNode} node A node to add. This is a LabeledStatement.
             * @returns {void}
             */
            function enterLabeledScope(node) {
                scopeInfo = {
                    label: node.label.name,
                    used: false,
                    upper: scopeInfo
                };
            }
            /**
             * Removes the top of the stack.
             * At the same time, this reports the label if it's never used.
             * @param {ASTNode} node A node to report. This is a LabeledStatement.
             * @returns {void}
             */
            function exitLabeledScope(node) {
                if (!scopeInfo.used) {
                    context.report({
                        node: node.label,
                        messageId: "unused",
                        data: node.label,
                        fix(fixer) {
                            /*
                             * Only perform a fix if there are no comments between the label and the body. This will be the case
                             * when there is exactly one token/comment (the ":") between the label and the body.
                             */
                            if (sourceCode.getTokenAfter(node.label, { includeComments: true }) ===
                                sourceCode.getTokenBefore(node.body, { includeComments: true })) {
                                return fixer.removeRange([node.range[0], node.body.range[0]]);
                            }
                            return null;
                        }
                    });
                }
                scopeInfo = scopeInfo.upper;
            }
            /**
             * Marks the label of a given node as used.
             * @param {ASTNode} node A node to mark. This is a BreakStatement or
             *      ContinueStatement.
             * @returns {void}
             */
            function markAsUsed(node) {
                if (!node.label) {
                    return;
                }
                const label = node.label.name;
                let info = scopeInfo;
                while (info) {
                    if (info.label === label) {
                        info.used = true;
                        break;
                    }
                    info = info.upper;
                }
            }
            return {
                LabeledStatement: enterLabeledScope,
                "LabeledStatement:exit": exitLabeledScope,
                BreakStatement: markAsUsed,
                ContinueStatement: markAsUsed
            };
        }
    };
},
/* eslint/lib/rules/no-unused-vars.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    /**
     * Bag of data used for formatting the `unusedVar` lint message.
     * @typedef {Object} UnusedVarMessageData
     * @property {string} varName The name of the unused var.
     * @property {'defined'|'assigned a value'} action Description of the vars state.
     * @property {string} additional Any additional info to be appended at the end.
     */
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow unused variables",
                category: "Variables",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-unused-vars"
            },
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["all", "local"]
                        },
                        {
                            type: "object",
                            properties: {
                                vars: {
                                    enum: ["all", "local"]
                                },
                                varsIgnorePattern: {
                                    type: "string"
                                },
                                args: {
                                    enum: ["all", "after-used", "none"]
                                },
                                ignoreRestSiblings: {
                                    type: "boolean"
                                },
                                argsIgnorePattern: {
                                    type: "string"
                                },
                                caughtErrors: {
                                    enum: ["all", "none"]
                                },
                                caughtErrorsIgnorePattern: {
                                    type: "string"
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                unusedVar: "'{{varName}}' is {{action}} but never used{{additional}}."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const REST_PROPERTY_TYPE = /^(?:RestElement|(?:Experimental)?RestProperty)$/u;
            const config = {
                vars: "all",
                args: "after-used",
                ignoreRestSiblings: false,
                caughtErrors: "none"
            };
            const firstOption = context.options[0];
            if (firstOption) {
                if (typeof firstOption === "string") {
                    config.vars = firstOption;
                }
                else {
                    config.vars = firstOption.vars || config.vars;
                    config.args = firstOption.args || config.args;
                    config.ignoreRestSiblings = firstOption.ignoreRestSiblings || config.ignoreRestSiblings;
                    config.caughtErrors = firstOption.caughtErrors || config.caughtErrors;
                    if (firstOption.varsIgnorePattern) {
                        config.varsIgnorePattern = new RegExp(firstOption.varsIgnorePattern, "u");
                    }
                    if (firstOption.argsIgnorePattern) {
                        config.argsIgnorePattern = new RegExp(firstOption.argsIgnorePattern, "u");
                    }
                    if (firstOption.caughtErrorsIgnorePattern) {
                        config.caughtErrorsIgnorePattern = new RegExp(firstOption.caughtErrorsIgnorePattern, "u");
                    }
                }
            }
            /**
             * Generates the message data about the variable being defined and unused,
             * including the ignore pattern if configured.
             * @param {Variable} unusedVar eslint-scope variable object.
             * @returns {UnusedVarMessageData} The message data to be used with this unused variable.
             */
            function getDefinedMessageData(unusedVar) {
                const defType = unusedVar.defs && unusedVar.defs[0] && unusedVar.defs[0].type;
                let type;
                let pattern;
                if (defType === "CatchClause" && config.caughtErrorsIgnorePattern) {
                    type = "args";
                    pattern = config.caughtErrorsIgnorePattern.toString();
                }
                else if (defType === "Parameter" && config.argsIgnorePattern) {
                    type = "args";
                    pattern = config.argsIgnorePattern.toString();
                }
                else if (defType !== "Parameter" && config.varsIgnorePattern) {
                    type = "vars";
                    pattern = config.varsIgnorePattern.toString();
                }
                const additional = type ? `. Allowed unused ${type} must match ${pattern}` : "";
                return {
                    varName: unusedVar.name,
                    action: "defined",
                    additional
                };
            }
            /**
             * Generate the warning message about the variable being
             * assigned and unused, including the ignore pattern if configured.
             * @param {Variable} unusedVar eslint-scope variable object.
             * @returns {UnusedVarMessageData} The message data to be used with this unused variable.
             */
            function getAssignedMessageData(unusedVar) {
                const additional = config.varsIgnorePattern ? `. Allowed unused vars must match ${config.varsIgnorePattern.toString()}` : "";
                return {
                    varName: unusedVar.name,
                    action: "assigned a value",
                    additional
                };
            }
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            const STATEMENT_TYPE = /(?:Statement|Declaration)$/u;
            /**
             * Determines if a given variable is being exported from a module.
             * @param {Variable} variable eslint-scope variable object.
             * @returns {boolean} True if the variable is exported, false if not.
             * @private
             */
            function isExported(variable) {
                const definition = variable.defs[0];
                if (definition) {
                    let node = definition.node;
                    if (node.type === "VariableDeclarator") {
                        node = node.parent;
                    }
                    else if (definition.type === "Parameter") {
                        return false;
                    }
                    return node.parent.type.indexOf("Export") === 0;
                }
                return false;
            }
            /**
             * Checks whether a node is a sibling of the rest property or not.
             * @param {ASTNode} node a node to check
             * @returns {boolean} True if the node is a sibling of the rest property, otherwise false.
             */
            function hasRestSibling(node) {
                return node.type === "Property" &&
                    node.parent.type === "ObjectPattern" &&
                    REST_PROPERTY_TYPE.test(node.parent.properties[node.parent.properties.length - 1].type);
            }
            /**
             * Determines if a variable has a sibling rest property
             * @param {Variable} variable eslint-scope variable object.
             * @returns {boolean} True if the variable is exported, false if not.
             * @private
             */
            function hasRestSpreadSibling(variable) {
                if (config.ignoreRestSiblings) {
                    const hasRestSiblingDefinition = variable.defs.some(def => hasRestSibling(def.name.parent));
                    const hasRestSiblingReference = variable.references.some(ref => hasRestSibling(ref.identifier.parent));
                    return hasRestSiblingDefinition || hasRestSiblingReference;
                }
                return false;
            }
            /**
             * Determines if a reference is a read operation.
             * @param {Reference} ref An eslint-scope Reference
             * @returns {boolean} whether the given reference represents a read operation
             * @private
             */
            function isReadRef(ref) {
                return ref.isRead();
            }
            /**
             * Determine if an identifier is referencing an enclosing function name.
             * @param {Reference} ref The reference to check.
             * @param {ASTNode[]} nodes The candidate function nodes.
             * @returns {boolean} True if it's a self-reference, false if not.
             * @private
             */
            function isSelfReference(ref, nodes) {
                let scope = ref.from;
                while (scope) {
                    if (nodes.indexOf(scope.block) >= 0) {
                        return true;
                    }
                    scope = scope.upper;
                }
                return false;
            }
            /**
             * Gets a list of function definitions for a specified variable.
             * @param {Variable} variable eslint-scope variable object.
             * @returns {ASTNode[]} Function nodes.
             * @private
             */
            function getFunctionDefinitions(variable) {
                const functionDefinitions = [];
                variable.defs.forEach(def => {
                    const { type, node } = def;
                    // FunctionDeclarations
                    if (type === "FunctionName") {
                        functionDefinitions.push(node);
                    }
                    // FunctionExpressions
                    if (type === "Variable" && node.init &&
                        (node.init.type === "FunctionExpression" || node.init.type === "ArrowFunctionExpression")) {
                        functionDefinitions.push(node.init);
                    }
                });
                return functionDefinitions;
            }
            /**
             * Checks the position of given nodes.
             * @param {ASTNode} inner A node which is expected as inside.
             * @param {ASTNode} outer A node which is expected as outside.
             * @returns {boolean} `true` if the `inner` node exists in the `outer` node.
             * @private
             */
            function isInside(inner, outer) {
                return (inner.range[0] >= outer.range[0] &&
                    inner.range[1] <= outer.range[1]);
            }
            /**
             * If a given reference is left-hand side of an assignment, this gets
             * the right-hand side node of the assignment.
             *
             * In the following cases, this returns null.
             *
             * - The reference is not the LHS of an assignment expression.
             * - The reference is inside of a loop.
             * - The reference is inside of a function scope which is different from
             *   the declaration.
             * @param {eslint-scope.Reference} ref A reference to check.
             * @param {ASTNode} prevRhsNode The previous RHS node.
             * @returns {ASTNode|null} The RHS node or null.
             * @private
             */
            function getRhsNode(ref, prevRhsNode) {
                const id = ref.identifier;
                const parent = id.parent;
                const grandparent = parent.parent;
                const refScope = ref.from.variableScope;
                const varScope = ref.resolved.scope.variableScope;
                const canBeUsedLater = refScope !== varScope || astUtils.isInLoop(id);
                /*
                 * Inherits the previous node if this reference is in the node.
                 * This is for `a = a + a`-like code.
                 */
                if (prevRhsNode && isInside(id, prevRhsNode)) {
                    return prevRhsNode;
                }
                if (parent.type === "AssignmentExpression" &&
                    grandparent.type === "ExpressionStatement" &&
                    id === parent.left &&
                    !canBeUsedLater) {
                    return parent.right;
                }
                return null;
            }
            /**
             * Checks whether a given function node is stored to somewhere or not.
             * If the function node is stored, the function can be used later.
             * @param {ASTNode} funcNode A function node to check.
             * @param {ASTNode} rhsNode The RHS node of the previous assignment.
             * @returns {boolean} `true` if under the following conditions:
             *      - the funcNode is assigned to a variable.
             *      - the funcNode is bound as an argument of a function call.
             *      - the function is bound to a property and the object satisfies above conditions.
             * @private
             */
            function isStorableFunction(funcNode, rhsNode) {
                let node = funcNode;
                let parent = funcNode.parent;
                while (parent && isInside(parent, rhsNode)) {
                    switch (parent.type) {
                        case "SequenceExpression":
                            if (parent.expressions[parent.expressions.length - 1] !== node) {
                                return false;
                            }
                            break;
                        case "CallExpression":
                        case "NewExpression":
                            return parent.callee !== node;
                        case "AssignmentExpression":
                        case "TaggedTemplateExpression":
                        case "YieldExpression":
                            return true;
                        default:
                            if (STATEMENT_TYPE.test(parent.type)) {
                                /*
                                 * If it encountered statements, this is a complex pattern.
                                 * Since analyzing complex patterns is hard, this returns `true` to avoid false positive.
                                 */
                                return true;
                            }
                    }
                    node = parent;
                    parent = parent.parent;
                }
                return false;
            }
            /**
             * Checks whether a given Identifier node exists inside of a function node which can be used later.
             *
             * "can be used later" means:
             * - the function is assigned to a variable.
             * - the function is bound to a property and the object can be used later.
             * - the function is bound as an argument of a function call.
             *
             * If a reference exists in a function which can be used later, the reference is read when the function is called.
             * @param {ASTNode} id An Identifier node to check.
             * @param {ASTNode} rhsNode The RHS node of the previous assignment.
             * @returns {boolean} `true` if the `id` node exists inside of a function node which can be used later.
             * @private
             */
            function isInsideOfStorableFunction(id, rhsNode) {
                const funcNode = astUtils.getUpperFunction(id);
                return (funcNode &&
                    isInside(funcNode, rhsNode) &&
                    isStorableFunction(funcNode, rhsNode));
            }
            /**
             * Checks whether a given node is unused expression or not.
             * @param {ASTNode} node The node itself
             * @returns {boolean} The node is an unused expression.
             * @private
             */
            function isUnusedExpression(node) {
                const parent = node.parent;
                if (parent.type === "ExpressionStatement") {
                    return true;
                }
                if (parent.type === "SequenceExpression") {
                    const isLastExpression = parent.expressions[parent.expressions.length - 1] === node;
                    if (!isLastExpression) {
                        return true;
                    }
                    return isUnusedExpression(parent);
                }
                return false;
            }
            /**
             * Checks whether a given reference is a read to update itself or not.
             * @param {eslint-scope.Reference} ref A reference to check.
             * @param {ASTNode} rhsNode The RHS node of the previous assignment.
             * @returns {boolean} The reference is a read to update itself.
             * @private
             */
            function isReadForItself(ref, rhsNode) {
                const id = ref.identifier;
                const parent = id.parent;
                return ref.isRead() && (
                // self update. e.g. `a += 1`, `a++`
                ((parent.type === "AssignmentExpression" &&
                    parent.left === id &&
                    isUnusedExpression(parent)) ||
                    (parent.type === "UpdateExpression" &&
                        isUnusedExpression(parent))) ||
                    // in RHS of an assignment for itself. e.g. `a = a + 1`
                    (rhsNode &&
                        isInside(id, rhsNode) &&
                        !isInsideOfStorableFunction(id, rhsNode)));
            }
            /**
             * Determine if an identifier is used either in for-in loops.
             * @param {Reference} ref The reference to check.
             * @returns {boolean} whether reference is used in the for-in loops
             * @private
             */
            function isForInRef(ref) {
                let target = ref.identifier.parent;
                // "for (var ...) { return; }"
                if (target.type === "VariableDeclarator") {
                    target = target.parent.parent;
                }
                if (target.type !== "ForInStatement") {
                    return false;
                }
                // "for (...) { return; }"
                if (target.body.type === "BlockStatement") {
                    target = target.body.body[0];
                    // "for (...) return;"
                }
                else {
                    target = target.body;
                }
                // For empty loop body
                if (!target) {
                    return false;
                }
                return target.type === "ReturnStatement";
            }
            /**
             * Determines if the variable is used.
             * @param {Variable} variable The variable to check.
             * @returns {boolean} True if the variable is used
             * @private
             */
            function isUsedVariable(variable) {
                const functionNodes = getFunctionDefinitions(variable), isFunctionDefinition = functionNodes.length > 0;
                let rhsNode = null;
                return variable.references.some(ref => {
                    if (isForInRef(ref)) {
                        return true;
                    }
                    const forItself = isReadForItself(ref, rhsNode);
                    rhsNode = getRhsNode(ref, rhsNode);
                    return (isReadRef(ref) &&
                        !forItself &&
                        !(isFunctionDefinition && isSelfReference(ref, functionNodes)));
                });
            }
            /**
             * Checks whether the given variable is after the last used parameter.
             * @param {eslint-scope.Variable} variable The variable to check.
             * @returns {boolean} `true` if the variable is defined after the last
             * used parameter.
             */
            function isAfterLastUsedArg(variable) {
                const def = variable.defs[0];
                const params = context.getDeclaredVariables(def.node);
                const posteriorParams = params.slice(params.indexOf(variable) + 1);
                // If any used parameters occur after this parameter, do not report.
                return !posteriorParams.some(v => v.references.length > 0 || v.eslintUsed);
            }
            /**
             * Gets an array of variables without read references.
             * @param {Scope} scope an eslint-scope Scope object.
             * @param {Variable[]} unusedVars an array that saving result.
             * @returns {Variable[]} unused variables of the scope and descendant scopes.
             * @private
             */
            function collectUnusedVariables(scope, unusedVars) {
                const variables = scope.variables;
                const childScopes = scope.childScopes;
                let i, l;
                if (scope.type !== "global" || config.vars === "all") {
                    for (i = 0, l = variables.length; i < l; ++i) {
                        const variable = variables[i];
                        // skip a variable of class itself name in the class scope
                        if (scope.type === "class" && scope.block.id === variable.identifiers[0]) {
                            continue;
                        }
                        // skip function expression names and variables marked with markVariableAsUsed()
                        if (scope.functionExpressionScope || variable.eslintUsed) {
                            continue;
                        }
                        // skip implicit "arguments" variable
                        if (scope.type === "function" && variable.name === "arguments" && variable.identifiers.length === 0) {
                            continue;
                        }
                        // explicit global variables don't have definitions.
                        const def = variable.defs[0];
                        if (def) {
                            const type = def.type;
                            // skip catch variables
                            if (type === "CatchClause") {
                                if (config.caughtErrors === "none") {
                                    continue;
                                }
                                // skip ignored parameters
                                if (config.caughtErrorsIgnorePattern && config.caughtErrorsIgnorePattern.test(def.name.name)) {
                                    continue;
                                }
                            }
                            if (type === "Parameter") {
                                // skip any setter argument
                                if ((def.node.parent.type === "Property" || def.node.parent.type === "MethodDefinition") && def.node.parent.kind === "set") {
                                    continue;
                                }
                                // if "args" option is "none", skip any parameter
                                if (config.args === "none") {
                                    continue;
                                }
                                // skip ignored parameters
                                if (config.argsIgnorePattern && config.argsIgnorePattern.test(def.name.name)) {
                                    continue;
                                }
                                // if "args" option is "after-used", skip used variables
                                if (config.args === "after-used" && astUtils.isFunction(def.name.parent) && !isAfterLastUsedArg(variable)) {
                                    continue;
                                }
                            }
                            else {
                                // skip ignored variables
                                if (config.varsIgnorePattern && config.varsIgnorePattern.test(def.name.name)) {
                                    continue;
                                }
                            }
                        }
                        if (!isUsedVariable(variable) && !isExported(variable) && !hasRestSpreadSibling(variable)) {
                            unusedVars.push(variable);
                        }
                    }
                }
                for (i = 0, l = childScopes.length; i < l; ++i) {
                    collectUnusedVariables(childScopes[i], unusedVars);
                }
                return unusedVars;
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                "Program:exit"(programNode) {
                    const unusedVars = collectUnusedVariables(context.getScope(), []);
                    for (let i = 0, l = unusedVars.length; i < l; ++i) {
                        const unusedVar = unusedVars[i];
                        // Report the first declaration.
                        if (unusedVar.defs.length > 0) {
                            // report last write reference, https://github.com/eslint/eslint/issues/14324
                            const writeReferences = unusedVar.references.filter(ref => ref.isWrite() && ref.from.variableScope === unusedVar.scope.variableScope);
                            let referenceToReport;
                            if (writeReferences.length > 0) {
                                referenceToReport = writeReferences[writeReferences.length - 1];
                            }
                            context.report({
                                node: referenceToReport ? referenceToReport.identifier : unusedVar.identifiers[0],
                                messageId: "unusedVar",
                                data: unusedVar.references.some(ref => ref.isWrite())
                                    ? getAssignedMessageData(unusedVar)
                                    : getDefinedMessageData(unusedVar)
                            });
                            // If there are no regular declaration, report the first `/*globals*/` comment directive.
                        }
                        else if (unusedVar.eslintExplicitGlobalComments) {
                            const directiveComment = unusedVar.eslintExplicitGlobalComments[0];
                            context.report({
                                node: programNode,
                                loc: astUtils.getNameLocationInGlobalDirectiveComment(sourceCode, directiveComment, unusedVar.name),
                                messageId: "unusedVar",
                                data: getDefinedMessageData(unusedVar)
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-use-before-define.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const SENTINEL_TYPE = /^(?:(?:Function|Class)(?:Declaration|Expression)|ArrowFunctionExpression|CatchClause|ImportDeclaration|ExportNamedDeclaration)$/u;
    const FOR_IN_OF_TYPE = /^For(?:In|Of)Statement$/u;
    /**
     * Parses a given value as options.
     * @param {any} options A value to parse.
     * @returns {Object} The parsed options.
     */
    function parseOptions(options) {
        let functions = true;
        let classes = true;
        let variables = true;
        if (typeof options === "string") {
            functions = (options !== "nofunc");
        }
        else if (typeof options === "object" && options !== null) {
            functions = options.functions !== false;
            classes = options.classes !== false;
            variables = options.variables !== false;
        }
        return { functions, classes, variables };
    }
    /**
     * Checks whether or not a given variable is a function declaration.
     * @param {eslint-scope.Variable} variable A variable to check.
     * @returns {boolean} `true` if the variable is a function declaration.
     */
    function isFunction(variable) {
        return variable.defs[0].type === "FunctionName";
    }
    /**
     * Checks whether or not a given variable is a class declaration in an upper function scope.
     * @param {eslint-scope.Variable} variable A variable to check.
     * @param {eslint-scope.Reference} reference A reference to check.
     * @returns {boolean} `true` if the variable is a class declaration.
     */
    function isOuterClass(variable, reference) {
        return (variable.defs[0].type === "ClassName" &&
            variable.scope.variableScope !== reference.from.variableScope);
    }
    /**
     * Checks whether or not a given variable is a variable declaration in an upper function scope.
     * @param {eslint-scope.Variable} variable A variable to check.
     * @param {eslint-scope.Reference} reference A reference to check.
     * @returns {boolean} `true` if the variable is a variable declaration.
     */
    function isOuterVariable(variable, reference) {
        return (variable.defs[0].type === "Variable" &&
            variable.scope.variableScope !== reference.from.variableScope);
    }
    /**
     * Checks whether or not a given location is inside of the range of a given node.
     * @param {ASTNode} node An node to check.
     * @param {number} location A location to check.
     * @returns {boolean} `true` if the location is inside of the range of the node.
     */
    function isInRange(node, location) {
        return node && node.range[0] <= location && location <= node.range[1];
    }
    /**
     * Checks whether or not a given reference is inside of the initializers of a given variable.
     *
     * This returns `true` in the following cases:
     *
     *     var a = a
     *     var [a = a] = list
     *     var {a = a} = obj
     *     for (var a in a) {}
     *     for (var a of a) {}
     * @param {Variable} variable A variable to check.
     * @param {Reference} reference A reference to check.
     * @returns {boolean} `true` if the reference is inside of the initializers.
     */
    function isInInitializer(variable, reference) {
        if (variable.scope !== reference.from) {
            return false;
        }
        let node = variable.identifiers[0].parent;
        const location = reference.identifier.range[1];
        while (node) {
            if (node.type === "VariableDeclarator") {
                if (isInRange(node.init, location)) {
                    return true;
                }
                if (FOR_IN_OF_TYPE.test(node.parent.parent.type) &&
                    isInRange(node.parent.parent.right, location)) {
                    return true;
                }
                break;
            }
            else if (node.type === "AssignmentPattern") {
                if (isInRange(node.right, location)) {
                    return true;
                }
            }
            else if (SENTINEL_TYPE.test(node.type)) {
                break;
            }
            node = node.parent;
        }
        return false;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow the use of variables before they are defined",
                category: "Variables",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-use-before-define"
            },
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["nofunc"]
                        },
                        {
                            type: "object",
                            properties: {
                                functions: { type: "boolean" },
                                classes: { type: "boolean" },
                                variables: { type: "boolean" }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                usedBeforeDefined: "'{{name}}' was used before it was defined."
            }
        },
        create(context) {
            const options = parseOptions(context.options[0]);
            /**
             * Determines whether a given use-before-define case should be reported according to the options.
             * @param {eslint-scope.Variable} variable The variable that gets used before being defined
             * @param {eslint-scope.Reference} reference The reference to the variable
             * @returns {boolean} `true` if the usage should be reported
             */
            function isForbidden(variable, reference) {
                if (isFunction(variable)) {
                    return options.functions;
                }
                if (isOuterClass(variable, reference)) {
                    return options.classes;
                }
                if (isOuterVariable(variable, reference)) {
                    return options.variables;
                }
                return true;
            }
            /**
             * Finds and validates all variables in a given scope.
             * @param {Scope} scope The scope object.
             * @returns {void}
             * @private
             */
            function findVariablesInScope(scope) {
                scope.references.forEach(reference => {
                    const variable = reference.resolved;
                    /*
                     * Skips when the reference is:
                     * - initialization's.
                     * - referring to an undefined variable.
                     * - referring to a global environment variable (there're no identifiers).
                     * - located preceded by the variable (except in initializers).
                     * - allowed by options.
                     */
                    if (reference.init ||
                        !variable ||
                        variable.identifiers.length === 0 ||
                        (variable.identifiers[0].range[1] < reference.identifier.range[1] && !isInInitializer(variable, reference)) ||
                        !isForbidden(variable, reference)) {
                        return;
                    }
                    // Reports.
                    context.report({
                        node: reference.identifier,
                        messageId: "usedBeforeDefined",
                        data: reference.identifier
                    });
                });
                scope.childScopes.forEach(findVariablesInScope);
            }
            return {
                Program() {
                    findVariablesInScope(context.getScope());
                }
            };
        }
    };
},
/* eslint/lib/rules/no-useless-backreference.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { CALL, CONSTRUCT, ReferenceTracker, getStringIfConstant } = require(350) /* eslint-utils */;
    const { RegExpParser, visitRegExpAST } = require(497) /* regexpp */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const parser = new RegExpParser();
    /**
     * Finds the path from the given `regexpp` AST node to the root node.
     * @param {regexpp.Node} node Node.
     * @returns {regexpp.Node[]} Array that starts with the given node and ends with the root node.
     */
    function getPathToRoot(node) {
        const path = [];
        let current = node;
        do {
            path.push(current);
            current = current.parent;
        } while (current);
        return path;
    }
    /**
     * Determines whether the given `regexpp` AST node is a lookaround node.
     * @param {regexpp.Node} node Node.
     * @returns {boolean} `true` if it is a lookaround node.
     */
    function isLookaround(node) {
        return node.type === "Assertion" &&
            (node.kind === "lookahead" || node.kind === "lookbehind");
    }
    /**
     * Determines whether the given `regexpp` AST node is a negative lookaround node.
     * @param {regexpp.Node} node Node.
     * @returns {boolean} `true` if it is a negative lookaround node.
     */
    function isNegativeLookaround(node) {
        return isLookaround(node) && node.negate;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow useless backreferences in regular expressions",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-useless-backreference"
            },
            schema: [],
            messages: {
                nested: "Backreference '{{ bref }}' will be ignored. It references group '{{ group }}' from within that group.",
                forward: "Backreference '{{ bref }}' will be ignored. It references group '{{ group }}' which appears later in the pattern.",
                backward: "Backreference '{{ bref }}' will be ignored. It references group '{{ group }}' which appears before in the same lookbehind.",
                disjunctive: "Backreference '{{ bref }}' will be ignored. It references group '{{ group }}' which is in another alternative.",
                intoNegativeLookaround: "Backreference '{{ bref }}' will be ignored. It references group '{{ group }}' which is in a negative lookaround."
            }
        },
        create(context) {
            /**
             * Checks and reports useless backreferences in the given regular expression.
             * @param {ASTNode} node Node that represents regular expression. A regex literal or RegExp constructor call.
             * @param {string} pattern Regular expression pattern.
             * @param {string} flags Regular expression flags.
             * @returns {void}
             */
            function checkRegex(node, pattern, flags) {
                let regExpAST;
                try {
                    regExpAST = parser.parsePattern(pattern, 0, pattern.length, flags.includes("u"));
                }
                catch {
                    // Ignore regular expressions with syntax errors
                    return;
                }
                visitRegExpAST(regExpAST, {
                    onBackreferenceEnter(bref) {
                        const group = bref.resolved, brefPath = getPathToRoot(bref), groupPath = getPathToRoot(group);
                        let messageId = null;
                        if (brefPath.includes(group)) {
                            // group is bref's ancestor => bref is nested ('nested reference') => group hasn't matched yet when bref starts to match.
                            messageId = "nested";
                        }
                        else {
                            // Start from the root to find the lowest common ancestor.
                            let i = brefPath.length - 1, j = groupPath.length - 1;
                            do {
                                i--;
                                j--;
                            } while (brefPath[i] === groupPath[j]);
                            const indexOfLowestCommonAncestor = j + 1, groupCut = groupPath.slice(0, indexOfLowestCommonAncestor), commonPath = groupPath.slice(indexOfLowestCommonAncestor), lowestCommonLookaround = commonPath.find(isLookaround), isMatchingBackward = lowestCommonLookaround && lowestCommonLookaround.kind === "lookbehind";
                            if (!isMatchingBackward && bref.end <= group.start) {
                                // bref is left, group is right ('forward reference') => group hasn't matched yet when bref starts to match.
                                messageId = "forward";
                            }
                            else if (isMatchingBackward && group.end <= bref.start) {
                                // the opposite of the previous when the regex is matching backward in a lookbehind context.
                                messageId = "backward";
                            }
                            else if (groupCut[groupCut.length - 1].type === "Alternative") {
                                // group's and bref's ancestor nodes below the lowest common ancestor are sibling alternatives => they're disjunctive.
                                messageId = "disjunctive";
                            }
                            else if (groupCut.some(isNegativeLookaround)) {
                                // group is in a negative lookaround which isn't bref's ancestor => group has already failed when bref starts to match.
                                messageId = "intoNegativeLookaround";
                            }
                        }
                        if (messageId) {
                            context.report({
                                node,
                                messageId,
                                data: {
                                    bref: bref.raw,
                                    group: group.raw
                                }
                            });
                        }
                    }
                });
            }
            return {
                "Literal[regex]"(node) {
                    const { pattern, flags } = node.regex;
                    checkRegex(node, pattern, flags);
                },
                Program() {
                    const scope = context.getScope(), tracker = new ReferenceTracker(scope), traceMap = {
                        RegExp: {
                            [CALL]: true,
                            [CONSTRUCT]: true
                        }
                    };
                    for (const { node } of tracker.iterateGlobalReferences(traceMap)) {
                        const [patternNode, flagsNode] = node.arguments, pattern = getStringIfConstant(patternNode, scope), flags = getStringIfConstant(flagsNode, scope);
                        if (typeof pattern === "string") {
                            checkRegex(node, pattern, flags || "");
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-useless-call.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a node is a `.call()`/`.apply()`.
     * @param {ASTNode} node A CallExpression node to check.
     * @returns {boolean} Whether or not the node is a `.call()`/`.apply()`.
     */
    function isCallOrNonVariadicApply(node) {
        const callee = astUtils.skipChainExpression(node.callee);
        return (callee.type === "MemberExpression" &&
            callee.property.type === "Identifier" &&
            callee.computed === false &&
            ((callee.property.name === "call" && node.arguments.length >= 1) ||
                (callee.property.name === "apply" && node.arguments.length === 2 && node.arguments[1].type === "ArrayExpression")));
    }
    /**
     * Checks whether or not `thisArg` is not changed by `.call()`/`.apply()`.
     * @param {ASTNode|null} expectedThis The node that is the owner of the applied function.
     * @param {ASTNode} thisArg The node that is given to the first argument of the `.call()`/`.apply()`.
     * @param {SourceCode} sourceCode The ESLint source code object.
     * @returns {boolean} Whether or not `thisArg` is not changed by `.call()`/`.apply()`.
     */
    function isValidThisArg(expectedThis, thisArg, sourceCode) {
        if (!expectedThis) {
            return astUtils.isNullOrUndefined(thisArg);
        }
        return astUtils.equalTokens(expectedThis, thisArg, sourceCode);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary calls to `.call()` and `.apply()`",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-useless-call"
            },
            schema: [],
            messages: {
                unnecessaryCall: "Unnecessary '.{{name}}()'."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                CallExpression(node) {
                    if (!isCallOrNonVariadicApply(node)) {
                        return;
                    }
                    const callee = astUtils.skipChainExpression(node.callee);
                    const applied = astUtils.skipChainExpression(callee.object);
                    const expectedThis = (applied.type === "MemberExpression") ? applied.object : null;
                    const thisArg = node.arguments[0];
                    if (isValidThisArg(expectedThis, thisArg, sourceCode)) {
                        context.report({ node, messageId: "unnecessaryCall", data: { name: callee.property.name } });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-useless-catch.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary `catch` clauses",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-useless-catch"
            },
            schema: [],
            messages: {
                unnecessaryCatchClause: "Unnecessary catch clause.",
                unnecessaryCatch: "Unnecessary try/catch wrapper."
            }
        },
        create(context) {
            return {
                CatchClause(node) {
                    if (node.param &&
                        node.param.type === "Identifier" &&
                        node.body.body.length &&
                        node.body.body[0].type === "ThrowStatement" &&
                        node.body.body[0].argument.type === "Identifier" &&
                        node.body.body[0].argument.name === node.param.name) {
                        if (node.parent.finalizer) {
                            context.report({
                                node,
                                messageId: "unnecessaryCatchClause"
                            });
                        }
                        else {
                            context.report({
                                node: node.parent,
                                messageId: "unnecessaryCatch"
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-useless-computed-key.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary computed property keys in objects and classes",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-useless-computed-key"
            },
            schema: [{
                    type: "object",
                    properties: {
                        enforceForClassMembers: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }],
            fixable: "code",
            messages: {
                unnecessarilyComputedProperty: "Unnecessarily computed property [{{property}}] found."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const enforceForClassMembers = context.options[0] && context.options[0].enforceForClassMembers;
            /**
             * Reports a given node if it violated this rule.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             */
            function check(node) {
                if (!node.computed) {
                    return;
                }
                const key = node.key, nodeType = typeof key.value;
                let allowedKey;
                if (node.type === "MethodDefinition") {
                    allowedKey = node.static ? "prototype" : "constructor";
                }
                else {
                    allowedKey = "__proto__";
                }
                if (key.type === "Literal" && (nodeType === "string" || nodeType === "number") && key.value !== allowedKey) {
                    context.report({
                        node,
                        messageId: "unnecessarilyComputedProperty",
                        data: { property: sourceCode.getText(key) },
                        fix(fixer) {
                            const leftSquareBracket = sourceCode.getTokenBefore(key, astUtils.isOpeningBracketToken);
                            const rightSquareBracket = sourceCode.getTokenAfter(key, astUtils.isClosingBracketToken);
                            // If there are comments between the brackets and the property name, don't do a fix.
                            if (sourceCode.commentsExistBetween(leftSquareBracket, rightSquareBracket)) {
                                return null;
                            }
                            const tokenBeforeLeftBracket = sourceCode.getTokenBefore(leftSquareBracket);
                            // Insert a space before the key to avoid changing identifiers, e.g. ({ get[2]() {} }) to ({ get2() {} })
                            const needsSpaceBeforeKey = tokenBeforeLeftBracket.range[1] === leftSquareBracket.range[0] &&
                                !astUtils.canTokensBeAdjacent(tokenBeforeLeftBracket, sourceCode.getFirstToken(key));
                            const replacementKey = (needsSpaceBeforeKey ? " " : "") + key.raw;
                            return fixer.replaceTextRange([leftSquareBracket.range[0], rightSquareBracket.range[1]], replacementKey);
                        }
                    });
                }
            }
            /**
             * A no-op function to act as placeholder for checking a node when the `enforceForClassMembers` option is `false`.
             * @returns {void}
             * @private
             */
            function noop() { }
            return {
                Property: check,
                MethodDefinition: enforceForClassMembers ? check : noop
            };
        }
    };
},
/* eslint/lib/rules/no-useless-concat.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given node is a concatenation.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node is a concatenation.
     */
    function isConcatenation(node) {
        return node.type === "BinaryExpression" && node.operator === "+";
    }
    /**
     * Checks if the given token is a `+` token or not.
     * @param {Token} token The token to check.
     * @returns {boolean} `true` if the token is a `+` token.
     */
    function isConcatOperatorToken(token) {
        return token.value === "+" && token.type === "Punctuator";
    }
    /**
     * Get's the right most node on the left side of a BinaryExpression with + operator.
     * @param {ASTNode} node A BinaryExpression node to check.
     * @returns {ASTNode} node
     */
    function getLeft(node) {
        let left = node.left;
        while (isConcatenation(left)) {
            left = left.right;
        }
        return left;
    }
    /**
     * Get's the left most node on the right side of a BinaryExpression with + operator.
     * @param {ASTNode} node A BinaryExpression node to check.
     * @returns {ASTNode} node
     */
    function getRight(node) {
        let right = node.right;
        while (isConcatenation(right)) {
            right = right.left;
        }
        return right;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary concatenation of literals or template literals",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-useless-concat"
            },
            schema: [],
            messages: {
                unexpectedConcat: "Unexpected string concatenation of literals."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                BinaryExpression(node) {
                    // check if not concatenation
                    if (node.operator !== "+") {
                        return;
                    }
                    // account for the `foo + "a" + "b"` case
                    const left = getLeft(node);
                    const right = getRight(node);
                    if (astUtils.isStringLiteral(left) &&
                        astUtils.isStringLiteral(right) &&
                        astUtils.isTokenOnSameLine(left, right)) {
                        const operatorToken = sourceCode.getFirstTokenBetween(left, right, isConcatOperatorToken);
                        context.report({
                            node,
                            loc: operatorToken.loc,
                            messageId: "unexpectedConcat"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-useless-constructor.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether a given array of statements is a single call of `super`.
     * @param {ASTNode[]} body An array of statements to check.
     * @returns {boolean} `true` if the body is a single call of `super`.
     */
    function isSingleSuperCall(body) {
        return (body.length === 1 &&
            body[0].type === "ExpressionStatement" &&
            body[0].expression.type === "CallExpression" &&
            body[0].expression.callee.type === "Super");
    }
    /**
     * Checks whether a given node is a pattern which doesn't have any side effects.
     * Default parameters and Destructuring parameters can have side effects.
     * @param {ASTNode} node A pattern node.
     * @returns {boolean} `true` if the node doesn't have any side effects.
     */
    function isSimple(node) {
        return node.type === "Identifier" || node.type === "RestElement";
    }
    /**
     * Checks whether a given array of expressions is `...arguments` or not.
     * `super(...arguments)` passes all arguments through.
     * @param {ASTNode[]} superArgs An array of expressions to check.
     * @returns {boolean} `true` if the superArgs is `...arguments`.
     */
    function isSpreadArguments(superArgs) {
        return (superArgs.length === 1 &&
            superArgs[0].type === "SpreadElement" &&
            superArgs[0].argument.type === "Identifier" &&
            superArgs[0].argument.name === "arguments");
    }
    /**
     * Checks whether given 2 nodes are identifiers which have the same name or not.
     * @param {ASTNode} ctorParam A node to check.
     * @param {ASTNode} superArg A node to check.
     * @returns {boolean} `true` if the nodes are identifiers which have the same
     *      name.
     */
    function isValidIdentifierPair(ctorParam, superArg) {
        return (ctorParam.type === "Identifier" &&
            superArg.type === "Identifier" &&
            ctorParam.name === superArg.name);
    }
    /**
     * Checks whether given 2 nodes are a rest/spread pair which has the same values.
     * @param {ASTNode} ctorParam A node to check.
     * @param {ASTNode} superArg A node to check.
     * @returns {boolean} `true` if the nodes are a rest/spread pair which has the
     *      same values.
     */
    function isValidRestSpreadPair(ctorParam, superArg) {
        return (ctorParam.type === "RestElement" &&
            superArg.type === "SpreadElement" &&
            isValidIdentifierPair(ctorParam.argument, superArg.argument));
    }
    /**
     * Checks whether given 2 nodes have the same value or not.
     * @param {ASTNode} ctorParam A node to check.
     * @param {ASTNode} superArg A node to check.
     * @returns {boolean} `true` if the nodes have the same value or not.
     */
    function isValidPair(ctorParam, superArg) {
        return (isValidIdentifierPair(ctorParam, superArg) ||
            isValidRestSpreadPair(ctorParam, superArg));
    }
    /**
     * Checks whether the parameters of a constructor and the arguments of `super()`
     * have the same values or not.
     * @param {ASTNode} ctorParams The parameters of a constructor to check.
     * @param {ASTNode} superArgs The arguments of `super()` to check.
     * @returns {boolean} `true` if those have the same values.
     */
    function isPassingThrough(ctorParams, superArgs) {
        if (ctorParams.length !== superArgs.length) {
            return false;
        }
        for (let i = 0; i < ctorParams.length; ++i) {
            if (!isValidPair(ctorParams[i], superArgs[i])) {
                return false;
            }
        }
        return true;
    }
    /**
     * Checks whether the constructor body is a redundant super call.
     * @param {Array} body constructor body content.
     * @param {Array} ctorParams The params to check against super call.
     * @returns {boolean} true if the constructor body is redundant
     */
    function isRedundantSuperCall(body, ctorParams) {
        return (isSingleSuperCall(body) &&
            ctorParams.every(isSimple) &&
            (isSpreadArguments(body[0].expression.arguments) ||
                isPassingThrough(ctorParams, body[0].expression.arguments)));
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary constructors",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-useless-constructor"
            },
            schema: [],
            messages: {
                noUselessConstructor: "Useless constructor."
            }
        },
        create(context) {
            /**
             * Checks whether a node is a redundant constructor
             * @param {ASTNode} node node to check
             * @returns {void}
             */
            function checkForConstructor(node) {
                if (node.kind !== "constructor") {
                    return;
                }
                /*
                 * Prevent crashing on parsers which do not require class constructor
                 * to have a body, e.g. typescript and flow
                 */
                if (!node.value.body) {
                    return;
                }
                const body = node.value.body.body;
                const ctorParams = node.value.params;
                const superClass = node.parent.parent.superClass;
                if (superClass ? isRedundantSuperCall(body, ctorParams) : (body.length === 0)) {
                    context.report({
                        node,
                        messageId: "noUselessConstructor"
                    });
                }
            }
            return {
                MethodDefinition: checkForConstructor
            };
        }
    };
},
/* eslint/lib/rules/no-useless-escape.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    /**
     * Returns the union of two sets.
     * @param {Set} setA The first set
     * @param {Set} setB The second set
     * @returns {Set} The union of the two sets
     */
    function union(setA, setB) {
        return new Set(function* () {
            yield* setA;
            yield* setB;
        }());
    }
    const VALID_STRING_ESCAPES = union(new Set("\\nrvtbfux"), astUtils.LINEBREAKS);
    const REGEX_GENERAL_ESCAPES = new Set("\\bcdDfnpPrsStvwWxu0123456789]");
    const REGEX_NON_CHARCLASS_ESCAPES = union(REGEX_GENERAL_ESCAPES, new Set("^/.$*+?[{}|()Bk"));
    /**
     * Parses a regular expression into a list of characters with character class info.
     * @param {string} regExpText The raw text used to create the regular expression
     * @returns {Object[]} A list of characters, each with info on escaping and whether they're in a character class.
     * @example
     *
     * parseRegExp('a\\b[cd-]')
     *
     * returns:
     * [
     *   {text: 'a', index: 0, escaped: false, inCharClass: false, startsCharClass: false, endsCharClass: false},
     *   {text: 'b', index: 2, escaped: true, inCharClass: false, startsCharClass: false, endsCharClass: false},
     *   {text: 'c', index: 4, escaped: false, inCharClass: true, startsCharClass: true, endsCharClass: false},
     *   {text: 'd', index: 5, escaped: false, inCharClass: true, startsCharClass: false, endsCharClass: false},
     *   {text: '-', index: 6, escaped: false, inCharClass: true, startsCharClass: false, endsCharClass: false}
     * ]
     */
    function parseRegExp(regExpText) {
        const charList = [];
        regExpText.split("").reduce((state, char, index) => {
            if (!state.escapeNextChar) {
                if (char === "\\") {
                    return Object.assign(state, { escapeNextChar: true });
                }
                if (char === "[" && !state.inCharClass) {
                    return Object.assign(state, { inCharClass: true, startingCharClass: true });
                }
                if (char === "]" && state.inCharClass) {
                    if (charList.length && charList[charList.length - 1].inCharClass) {
                        charList[charList.length - 1].endsCharClass = true;
                    }
                    return Object.assign(state, { inCharClass: false, startingCharClass: false });
                }
            }
            charList.push({
                text: char,
                index,
                escaped: state.escapeNextChar,
                inCharClass: state.inCharClass,
                startsCharClass: state.startingCharClass,
                endsCharClass: false
            });
            return Object.assign(state, { escapeNextChar: false, startingCharClass: false });
        }, { escapeNextChar: false, inCharClass: false, startingCharClass: false });
        return charList;
    }
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow unnecessary escape characters",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-useless-escape",
                suggestion: true
            },
            messages: {
                unnecessaryEscape: "Unnecessary escape character: \\{{character}}.",
                removeEscape: "Remove the `\\`. This maintains the current functionality.",
                escapeBackslash: "Replace the `\\` with `\\\\` to include the actual backslash character."
            },
            schema: []
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Reports a node
             * @param {ASTNode} node The node to report
             * @param {number} startOffset The backslash's offset from the start of the node
             * @param {string} character The uselessly escaped character (not including the backslash)
             * @returns {void}
             */
            function report(node, startOffset, character) {
                const rangeStart = node.range[0] + startOffset;
                const range = [rangeStart, rangeStart + 1];
                const start = sourceCode.getLocFromIndex(rangeStart);
                context.report({
                    node,
                    loc: {
                        start,
                        end: { line: start.line, column: start.column + 1 }
                    },
                    messageId: "unnecessaryEscape",
                    data: { character },
                    suggest: [
                        {
                            messageId: "removeEscape",
                            fix(fixer) {
                                return fixer.removeRange(range);
                            }
                        },
                        {
                            messageId: "escapeBackslash",
                            fix(fixer) {
                                return fixer.insertTextBeforeRange(range, "\\");
                            }
                        }
                    ]
                });
            }
            /**
             * Checks if the escape character in given string slice is unnecessary.
             * @private
             * @param {ASTNode} node node to validate.
             * @param {string} match string slice to validate.
             * @returns {void}
             */
            function validateString(node, match) {
                const isTemplateElement = node.type === "TemplateElement";
                const escapedChar = match[0][1];
                let isUnnecessaryEscape = !VALID_STRING_ESCAPES.has(escapedChar);
                let isQuoteEscape;
                if (isTemplateElement) {
                    isQuoteEscape = escapedChar === "`";
                    if (escapedChar === "$") {
                        // Warn if `\$` is not followed by `{`
                        isUnnecessaryEscape = match.input[match.index + 2] !== "{";
                    }
                    else if (escapedChar === "{") {
                        /*
                         * Warn if `\{` is not preceded by `$`. If preceded by `$`, escaping
                         * is necessary and the rule should not warn. If preceded by `/$`, the rule
                         * will warn for the `/$` instead, as it is the first unnecessarily escaped character.
                         */
                        isUnnecessaryEscape = match.input[match.index - 1] !== "$";
                    }
                }
                else {
                    isQuoteEscape = escapedChar === node.raw[0];
                }
                if (isUnnecessaryEscape && !isQuoteEscape) {
                    report(node, match.index, match[0].slice(1));
                }
            }
            /**
             * Checks if a node has an escape.
             * @param {ASTNode} node node to check.
             * @returns {void}
             */
            function check(node) {
                const isTemplateElement = node.type === "TemplateElement";
                if (isTemplateElement &&
                    node.parent &&
                    node.parent.parent &&
                    node.parent.parent.type === "TaggedTemplateExpression" &&
                    node.parent === node.parent.parent.quasi) {
                    // Don't report tagged template literals, because the backslash character is accessible to the tag function.
                    return;
                }
                if (typeof node.value === "string" || isTemplateElement) {
                    /*
                     * JSXAttribute doesn't have any escape sequence: https://facebook.github.io/jsx/.
                     * In addition, backticks are not supported by JSX yet: https://github.com/facebook/jsx/issues/25.
                     */
                    if (node.parent.type === "JSXAttribute" || node.parent.type === "JSXElement" || node.parent.type === "JSXFragment") {
                        return;
                    }
                    const value = isTemplateElement ? sourceCode.getText(node) : node.raw;
                    const pattern = /\\[^\d]/gu;
                    let match;
                    while ((match = pattern.exec(value))) {
                        validateString(node, match);
                    }
                }
                else if (node.regex) {
                    parseRegExp(node.regex.pattern)
                        /*
                         * The '-' character is a special case, because it's only valid to escape it if it's in a character
                         * class, and is not at either edge of the character class. To account for this, don't consider '-'
                         * characters to be valid in general, and filter out '-' characters that appear in the middle of a
                         * character class.
                         */
                        .filter(charInfo => !(charInfo.text === "-" && charInfo.inCharClass && !charInfo.startsCharClass && !charInfo.endsCharClass))
                        /*
                         * The '^' character is also a special case; it must always be escaped outside of character classes, but
                         * it only needs to be escaped in character classes if it's at the beginning of the character class. To
                         * account for this, consider it to be a valid escape character outside of character classes, and filter
                         * out '^' characters that appear at the start of a character class.
                         */
                        .filter(charInfo => !(charInfo.text === "^" && charInfo.startsCharClass))
                        // Filter out characters that aren't escaped.
                        .filter(charInfo => charInfo.escaped)
                        // Filter out characters that are valid to escape, based on their position in the regular expression.
                        .filter(charInfo => !(charInfo.inCharClass ? REGEX_GENERAL_ESCAPES : REGEX_NON_CHARCLASS_ESCAPES).has(charInfo.text))
                        // Report all the remaining characters.
                        .forEach(charInfo => report(node, charInfo.index, charInfo.text));
                }
            }
            return {
                Literal: check,
                TemplateElement: check
            };
        }
    };
},
/* eslint/lib/rules/no-useless-rename.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow renaming import, export, and destructured assignments to the same name",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-useless-rename"
            },
            fixable: "code",
            schema: [
                {
                    type: "object",
                    properties: {
                        ignoreDestructuring: { type: "boolean", default: false },
                        ignoreImport: { type: "boolean", default: false },
                        ignoreExport: { type: "boolean", default: false }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unnecessarilyRenamed: "{{type}} {{name}} unnecessarily renamed."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode(), options = context.options[0] || {}, ignoreDestructuring = options.ignoreDestructuring === true, ignoreImport = options.ignoreImport === true, ignoreExport = options.ignoreExport === true;
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Reports error for unnecessarily renamed assignments
             * @param {ASTNode} node node to report
             * @param {ASTNode} initial node with initial name value
             * @param {string} type the type of the offending node
             * @returns {void}
             */
            function reportError(node, initial, type) {
                const name = initial.type === "Identifier" ? initial.name : initial.value;
                return context.report({
                    node,
                    messageId: "unnecessarilyRenamed",
                    data: {
                        name,
                        type
                    },
                    fix(fixer) {
                        const replacementNode = node.type === "Property" ? node.value : node.local;
                        if (sourceCode.getCommentsInside(node).length > sourceCode.getCommentsInside(replacementNode).length) {
                            return null;
                        }
                        // Don't autofix code such as `({foo: (foo) = a} = obj);`, parens are not allowed in shorthand properties.
                        if (replacementNode.type === "AssignmentPattern" &&
                            astUtils.isParenthesised(sourceCode, replacementNode.left)) {
                            return null;
                        }
                        return fixer.replaceText(node, sourceCode.getText(replacementNode));
                    }
                });
            }
            /**
             * Checks whether a destructured assignment is unnecessarily renamed
             * @param {ASTNode} node node to check
             * @returns {void}
             */
            function checkDestructured(node) {
                if (ignoreDestructuring) {
                    return;
                }
                for (const property of node.properties) {
                    /**
                     * Properties using shorthand syntax and rest elements can not be renamed.
                     * If the property is computed, we have no idea if a rename is useless or not.
                     */
                    if (property.type !== "Property" || property.shorthand || property.computed) {
                        continue;
                    }
                    const key = (property.key.type === "Identifier" && property.key.name) || (property.key.type === "Literal" && property.key.value);
                    const renamedKey = property.value.type === "AssignmentPattern" ? property.value.left.name : property.value.name;
                    if (key === renamedKey) {
                        reportError(property, property.key, "Destructuring assignment");
                    }
                }
            }
            /**
             * Checks whether an import is unnecessarily renamed
             * @param {ASTNode} node node to check
             * @returns {void}
             */
            function checkImport(node) {
                if (ignoreImport) {
                    return;
                }
                if (node.imported.name === node.local.name &&
                    node.imported.range[0] !== node.local.range[0]) {
                    reportError(node, node.imported, "Import");
                }
            }
            /**
             * Checks whether an export is unnecessarily renamed
             * @param {ASTNode} node node to check
             * @returns {void}
             */
            function checkExport(node) {
                if (ignoreExport) {
                    return;
                }
                if (node.local.name === node.exported.name &&
                    node.local.range[0] !== node.exported.range[0]) {
                    reportError(node, node.local, "Export");
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                ObjectPattern: checkDestructured,
                ImportSpecifier: checkImport,
                ExportSpecifier: checkExport
            };
        }
    };
},
/* eslint/lib/rules/no-useless-return.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */, FixTracker = require(508) /* ./utils/fix-tracker */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Removes the given element from the array.
     * @param {Array} array The source array to remove.
     * @param {any} element The target item to remove.
     * @returns {void}
     */
    function remove(array, element) {
        const index = array.indexOf(element);
        if (index !== -1) {
            array.splice(index, 1);
        }
    }
    /**
     * Checks whether it can remove the given return statement or not.
     * @param {ASTNode} node The return statement node to check.
     * @returns {boolean} `true` if the node is removable.
     */
    function isRemovable(node) {
        return astUtils.STATEMENT_LIST_PARENTS.has(node.parent.type);
    }
    /**
     * Checks whether the given return statement is in a `finally` block or not.
     * @param {ASTNode} node The return statement node to check.
     * @returns {boolean} `true` if the node is in a `finally` block.
     */
    function isInFinally(node) {
        for (let currentNode = node; currentNode && currentNode.parent && !astUtils.isFunction(currentNode); currentNode = currentNode.parent) {
            if (currentNode.parent.type === "TryStatement" && currentNode.parent.finalizer === currentNode) {
                return true;
            }
        }
        return false;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow redundant return statements",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-useless-return"
            },
            fixable: "code",
            schema: [],
            messages: {
                unnecessaryReturn: "Unnecessary return statement."
            }
        },
        create(context) {
            const segmentInfoMap = new WeakMap();
            const usedUnreachableSegments = new WeakSet();
            const sourceCode = context.getSourceCode();
            let scopeInfo = null;
            /**
             * Checks whether the given segment is terminated by a return statement or not.
             * @param {CodePathSegment} segment The segment to check.
             * @returns {boolean} `true` if the segment is terminated by a return statement, or if it's still a part of unreachable.
             */
            function isReturned(segment) {
                const info = segmentInfoMap.get(segment);
                return !info || info.returned;
            }
            /**
             * Collects useless return statements from the given previous segments.
             *
             * A previous segment may be an unreachable segment.
             * In that case, the information object of the unreachable segment is not
             * initialized because `onCodePathSegmentStart` event is not notified for
             * unreachable segments.
             * This goes to the previous segments of the unreachable segment recursively
             * if the unreachable segment was generated by a return statement. Otherwise,
             * this ignores the unreachable segment.
             *
             * This behavior would simulate code paths for the case that the return
             * statement does not exist.
             * @param {ASTNode[]} uselessReturns The collected return statements.
             * @param {CodePathSegment[]} prevSegments The previous segments to traverse.
             * @param {WeakSet<CodePathSegment>} [providedTraversedSegments] A set of segments that have already been traversed in this call
             * @returns {ASTNode[]} `uselessReturns`.
             */
            function getUselessReturns(uselessReturns, prevSegments, providedTraversedSegments) {
                const traversedSegments = providedTraversedSegments || new WeakSet();
                for (const segment of prevSegments) {
                    if (!segment.reachable) {
                        if (!traversedSegments.has(segment)) {
                            traversedSegments.add(segment);
                            getUselessReturns(uselessReturns, segment.allPrevSegments.filter(isReturned), traversedSegments);
                        }
                        continue;
                    }
                    uselessReturns.push(...segmentInfoMap.get(segment).uselessReturns);
                }
                return uselessReturns;
            }
            /**
             * Removes the return statements on the given segment from the useless return
             * statement list.
             *
             * This segment may be an unreachable segment.
             * In that case, the information object of the unreachable segment is not
             * initialized because `onCodePathSegmentStart` event is not notified for
             * unreachable segments.
             * This goes to the previous segments of the unreachable segment recursively
             * if the unreachable segment was generated by a return statement. Otherwise,
             * this ignores the unreachable segment.
             *
             * This behavior would simulate code paths for the case that the return
             * statement does not exist.
             * @param {CodePathSegment} segment The segment to get return statements.
             * @returns {void}
             */
            function markReturnStatementsOnSegmentAsUsed(segment) {
                if (!segment.reachable) {
                    usedUnreachableSegments.add(segment);
                    segment.allPrevSegments
                        .filter(isReturned)
                        .filter(prevSegment => !usedUnreachableSegments.has(prevSegment))
                        .forEach(markReturnStatementsOnSegmentAsUsed);
                    return;
                }
                const info = segmentInfoMap.get(segment);
                for (const node of info.uselessReturns) {
                    remove(scopeInfo.uselessReturns, node);
                }
                info.uselessReturns = [];
            }
            /**
             * Removes the return statements on the current segments from the useless
             * return statement list.
             *
             * This function will be called at every statement except FunctionDeclaration,
             * BlockStatement, and BreakStatement.
             *
             * - FunctionDeclarations are always executed whether it's returned or not.
             * - BlockStatements do nothing.
             * - BreakStatements go the next merely.
             * @returns {void}
             */
            function markReturnStatementsOnCurrentSegmentsAsUsed() {
                scopeInfo
                    .codePath
                    .currentSegments
                    .forEach(markReturnStatementsOnSegmentAsUsed);
            }
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                // Makes and pushs a new scope information.
                onCodePathStart(codePath) {
                    scopeInfo = {
                        upper: scopeInfo,
                        uselessReturns: [],
                        codePath
                    };
                },
                // Reports useless return statements if exist.
                onCodePathEnd() {
                    for (const node of scopeInfo.uselessReturns) {
                        context.report({
                            node,
                            loc: node.loc,
                            messageId: "unnecessaryReturn",
                            fix(fixer) {
                                if (isRemovable(node) && !sourceCode.getCommentsInside(node).length) {
                                    /*
                                     * Extend the replacement range to include the
                                     * entire function to avoid conflicting with
                                     * no-else-return.
                                     * https://github.com/eslint/eslint/issues/8026
                                     */
                                    return new FixTracker(fixer, sourceCode)
                                        .retainEnclosingFunction(node)
                                        .remove(node);
                                }
                                return null;
                            }
                        });
                    }
                    scopeInfo = scopeInfo.upper;
                },
                /*
                 * Initializes segments.
                 * NOTE: This event is notified for only reachable segments.
                 */
                onCodePathSegmentStart(segment) {
                    const info = {
                        uselessReturns: getUselessReturns([], segment.allPrevSegments),
                        returned: false
                    };
                    // Stores the info.
                    segmentInfoMap.set(segment, info);
                },
                // Adds ReturnStatement node to check whether it's useless or not.
                ReturnStatement(node) {
                    if (node.argument) {
                        markReturnStatementsOnCurrentSegmentsAsUsed();
                    }
                    if (node.argument ||
                        astUtils.isInLoop(node) ||
                        isInFinally(node) ||
                        // Ignore `return` statements in unreachable places (https://github.com/eslint/eslint/issues/11647).
                        !scopeInfo.codePath.currentSegments.some(s => s.reachable)) {
                        return;
                    }
                    for (const segment of scopeInfo.codePath.currentSegments) {
                        const info = segmentInfoMap.get(segment);
                        if (info) {
                            info.uselessReturns.push(node);
                            info.returned = true;
                        }
                    }
                    scopeInfo.uselessReturns.push(node);
                },
                /*
                 * Registers for all statement nodes except FunctionDeclaration, BlockStatement, BreakStatement.
                 * Removes return statements of the current segments from the useless return statement list.
                 */
                ClassDeclaration: markReturnStatementsOnCurrentSegmentsAsUsed,
                ContinueStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                DebuggerStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                DoWhileStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                EmptyStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                ExpressionStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                ForInStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                ForOfStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                ForStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                IfStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                ImportDeclaration: markReturnStatementsOnCurrentSegmentsAsUsed,
                LabeledStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                SwitchStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                ThrowStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                TryStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                VariableDeclaration: markReturnStatementsOnCurrentSegmentsAsUsed,
                WhileStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                WithStatement: markReturnStatementsOnCurrentSegmentsAsUsed,
                ExportNamedDeclaration: markReturnStatementsOnCurrentSegmentsAsUsed,
                ExportDefaultDeclaration: markReturnStatementsOnCurrentSegmentsAsUsed,
                ExportAllDeclaration: markReturnStatementsOnCurrentSegmentsAsUsed
            };
        }
    };
},
/* eslint/lib/rules/no-var.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Check whether a given variable is a global variable or not.
     * @param {eslint-scope.Variable} variable The variable to check.
     * @returns {boolean} `true` if the variable is a global variable.
     */
    function isGlobal(variable) {
        return Boolean(variable.scope) && variable.scope.type === "global";
    }
    /**
     * Finds the nearest function scope or global scope walking up the scope
     * hierarchy.
     * @param {eslint-scope.Scope} scope The scope to traverse.
     * @returns {eslint-scope.Scope} a function scope or global scope containing the given
     *      scope.
     */
    function getEnclosingFunctionScope(scope) {
        let currentScope = scope;
        while (currentScope.type !== "function" && currentScope.type !== "global") {
            currentScope = currentScope.upper;
        }
        return currentScope;
    }
    /**
     * Checks whether the given variable has any references from a more specific
     * function expression (i.e. a closure).
     * @param {eslint-scope.Variable} variable A variable to check.
     * @returns {boolean} `true` if the variable is used from a closure.
     */
    function isReferencedInClosure(variable) {
        const enclosingFunctionScope = getEnclosingFunctionScope(variable.scope);
        return variable.references.some(reference => getEnclosingFunctionScope(reference.from) !== enclosingFunctionScope);
    }
    /**
     * Checks whether the given node is the assignee of a loop.
     * @param {ASTNode} node A VariableDeclaration node to check.
     * @returns {boolean} `true` if the declaration is assigned as part of loop
     *      iteration.
     */
    function isLoopAssignee(node) {
        return (node.parent.type === "ForOfStatement" || node.parent.type === "ForInStatement") &&
            node === node.parent.left;
    }
    /**
     * Checks whether the given variable declaration is immediately initialized.
     * @param {ASTNode} node A VariableDeclaration node to check.
     * @returns {boolean} `true` if the declaration has an initializer.
     */
    function isDeclarationInitialized(node) {
        return node.declarations.every(declarator => declarator.init !== null);
    }
    const SCOPE_NODE_TYPE = /^(?:Program|BlockStatement|SwitchStatement|ForStatement|ForInStatement|ForOfStatement)$/u;
    /**
     * Gets the scope node which directly contains a given node.
     * @param {ASTNode} node A node to get. This is a `VariableDeclaration` or
     *      an `Identifier`.
     * @returns {ASTNode} A scope node. This is one of `Program`, `BlockStatement`,
     *      `SwitchStatement`, `ForStatement`, `ForInStatement`, and
     *      `ForOfStatement`.
     */
    function getScopeNode(node) {
        for (let currentNode = node; currentNode; currentNode = currentNode.parent) {
            if (SCOPE_NODE_TYPE.test(currentNode.type)) {
                return currentNode;
            }
        }
        /* istanbul ignore next : unreachable */
        return null;
    }
    /**
     * Checks whether a given variable is redeclared or not.
     * @param {eslint-scope.Variable} variable A variable to check.
     * @returns {boolean} `true` if the variable is redeclared.
     */
    function isRedeclared(variable) {
        return variable.defs.length >= 2;
    }
    /**
     * Checks whether a given variable is used from outside of the specified scope.
     * @param {ASTNode} scopeNode A scope node to check.
     * @returns {Function} The predicate function which checks whether a given
     *      variable is used from outside of the specified scope.
     */
    function isUsedFromOutsideOf(scopeNode) {
        /**
         * Checks whether a given reference is inside of the specified scope or not.
         * @param {eslint-scope.Reference} reference A reference to check.
         * @returns {boolean} `true` if the reference is inside of the specified
         *      scope.
         */
        function isOutsideOfScope(reference) {
            const scope = scopeNode.range;
            const id = reference.identifier.range;
            return id[0] < scope[0] || id[1] > scope[1];
        }
        return function (variable) {
            return variable.references.some(isOutsideOfScope);
        };
    }
    /**
     * Creates the predicate function which checks whether a variable has their references in TDZ.
     *
     * The predicate function would return `true`:
     *
     * - if a reference is before the declarator. E.g. (var a = b, b = 1;)(var {a = b, b} = {};)
     * - if a reference is in the expression of their default value.  E.g. (var {a = a} = {};)
     * - if a reference is in the expression of their initializer.  E.g. (var a = a;)
     * @param {ASTNode} node The initializer node of VariableDeclarator.
     * @returns {Function} The predicate function.
     * @private
     */
    function hasReferenceInTDZ(node) {
        const initStart = node.range[0];
        const initEnd = node.range[1];
        return variable => {
            const id = variable.defs[0].name;
            const idStart = id.range[0];
            const defaultValue = (id.parent.type === "AssignmentPattern" ? id.parent.right : null);
            const defaultStart = defaultValue && defaultValue.range[0];
            const defaultEnd = defaultValue && defaultValue.range[1];
            return variable.references.some(reference => {
                const start = reference.identifier.range[0];
                const end = reference.identifier.range[1];
                return !reference.init && (start < idStart ||
                    (defaultValue !== null && start >= defaultStart && end <= defaultEnd) ||
                    (start >= initStart && end <= initEnd));
            });
        };
    }
    /**
     * Checks whether a given variable has name that is allowed for 'var' declarations,
     * but disallowed for `let` declarations.
     * @param {eslint-scope.Variable} variable The variable to check.
     * @returns {boolean} `true` if the variable has a disallowed name.
     */
    function hasNameDisallowedForLetDeclarations(variable) {
        return variable.name === "let";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require `let` or `const` instead of `var`",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-var"
            },
            schema: [],
            fixable: "code",
            messages: {
                unexpectedVar: "Unexpected var, use let or const instead."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Checks whether the variables which are defined by the given declarator node have their references in TDZ.
             * @param {ASTNode} declarator The VariableDeclarator node to check.
             * @returns {boolean} `true` if one of the variables which are defined by the given declarator node have their references in TDZ.
             */
            function hasSelfReferenceInTDZ(declarator) {
                if (!declarator.init) {
                    return false;
                }
                const variables = context.getDeclaredVariables(declarator);
                return variables.some(hasReferenceInTDZ(declarator.init));
            }
            /**
             * Checks whether it can fix a given variable declaration or not.
             * It cannot fix if the following cases:
             *
             * - A variable is a global variable.
             * - A variable is declared on a SwitchCase node.
             * - A variable is redeclared.
             * - A variable is used from outside the scope.
             * - A variable is used from a closure within a loop.
             * - A variable might be used before it is assigned within a loop.
             * - A variable might be used in TDZ.
             * - A variable is declared in statement position (e.g. a single-line `IfStatement`)
             * - A variable has name that is disallowed for `let` declarations.
             *
             * ## A variable is declared on a SwitchCase node.
             *
             * If this rule modifies 'var' declarations on a SwitchCase node, it
             * would generate the warnings of 'no-case-declarations' rule. And the
             * 'eslint:recommended' preset includes 'no-case-declarations' rule, so
             * this rule doesn't modify those declarations.
             *
             * ## A variable is redeclared.
             *
             * The language spec disallows redeclarations of `let` declarations.
             * Those variables would cause syntax errors.
             *
             * ## A variable is used from outside the scope.
             *
             * The language spec disallows accesses from outside of the scope for
             * `let` declarations. Those variables would cause reference errors.
             *
             * ## A variable is used from a closure within a loop.
             *
             * A `var` declaration within a loop shares the same variable instance
             * across all loop iterations, while a `let` declaration creates a new
             * instance for each iteration. This means if a variable in a loop is
             * referenced by any closure, changing it from `var` to `let` would
             * change the behavior in a way that is generally unsafe.
             *
             * ## A variable might be used before it is assigned within a loop.
             *
             * Within a loop, a `let` declaration without an initializer will be
             * initialized to null, while a `var` declaration will retain its value
             * from the previous iteration, so it is only safe to change `var` to
             * `let` if we can statically determine that the variable is always
             * assigned a value before its first access in the loop body. To keep
             * the implementation simple, we only convert `var` to `let` within
             * loops when the variable is a loop assignee or the declaration has an
             * initializer.
             * @param {ASTNode} node A variable declaration node to check.
             * @returns {boolean} `true` if it can fix the node.
             */
            function canFix(node) {
                const variables = context.getDeclaredVariables(node);
                const scopeNode = getScopeNode(node);
                if (node.parent.type === "SwitchCase" ||
                    node.declarations.some(hasSelfReferenceInTDZ) ||
                    variables.some(isGlobal) ||
                    variables.some(isRedeclared) ||
                    variables.some(isUsedFromOutsideOf(scopeNode)) ||
                    variables.some(hasNameDisallowedForLetDeclarations)) {
                    return false;
                }
                if (astUtils.isInLoop(node)) {
                    if (variables.some(isReferencedInClosure)) {
                        return false;
                    }
                    if (!isLoopAssignee(node) && !isDeclarationInitialized(node)) {
                        return false;
                    }
                }
                if (!isLoopAssignee(node) &&
                    !(node.parent.type === "ForStatement" && node.parent.init === node) &&
                    !astUtils.STATEMENT_LIST_PARENTS.has(node.parent.type)) {
                    // If the declaration is not in a block, e.g. `if (foo) var bar = 1;`, then it can't be fixed.
                    return false;
                }
                return true;
            }
            /**
             * Reports a given variable declaration node.
             * @param {ASTNode} node A variable declaration node to report.
             * @returns {void}
             */
            function report(node) {
                context.report({
                    node,
                    messageId: "unexpectedVar",
                    fix(fixer) {
                        const varToken = sourceCode.getFirstToken(node, { filter: t => t.value === "var" });
                        return canFix(node)
                            ? fixer.replaceText(varToken, "let")
                            : null;
                    }
                });
            }
            return {
                "VariableDeclaration:exit"(node) {
                    if (node.kind === "var") {
                        report(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-void.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `void` operators",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-void"
            },
            messages: {
                noVoid: "Expected 'undefined' and instead saw 'void'."
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowAsStatement: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ]
        },
        create(context) {
            const allowAsStatement = context.options[0] && context.options[0].allowAsStatement;
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                'UnaryExpression[operator="void"]'(node) {
                    if (allowAsStatement &&
                        node.parent &&
                        node.parent.type === "ExpressionStatement") {
                        return;
                    }
                    context.report({
                        node,
                        messageId: "noVoid"
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/no-warning-comments.js */ function _(require, module, exports, __esModule, __esExport) {
    const escapeRegExp = require(329) /* escape-string-regexp */;
    const astUtils = require(394) /* ./utils/ast-utils */;
    const CHAR_LIMIT = 40;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow specified warning terms in comments",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-warning-comments"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        terms: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        },
                        location: {
                            enum: ["start", "anywhere"]
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedComment: "Unexpected '{{matchedTerm}}' comment: '{{comment}}'."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode(), configuration = context.options[0] || {}, warningTerms = configuration.terms || ["todo", "fixme", "xxx"], location = configuration.location || "start", selfConfigRegEx = /\bno-warning-comments\b/u;
            /**
             * Convert a warning term into a RegExp which will match a comment containing that whole word in the specified
             * location ("start" or "anywhere"). If the term starts or ends with non word characters, then the match will not
             * require word boundaries on that side.
             * @param {string} term A term to convert to a RegExp
             * @returns {RegExp} The term converted to a RegExp
             */
            function convertToRegExp(term) {
                const escaped = escapeRegExp(term);
                const wordBoundary = "\\b";
                const eitherOrWordBoundary = `|${wordBoundary}`;
                let prefix;
                /*
                 * If the term ends in a word character (a-z0-9_), ensure a word
                 * boundary at the end, so that substrings do not get falsely
                 * matched. eg "todo" in a string such as "mastodon".
                 * If the term ends in a non-word character, then \b won't match on
                 * the boundary to the next non-word character, which would likely
                 * be a space. For example `/\bFIX!\b/.test('FIX! blah') === false`.
                 * In these cases, use no bounding match. Same applies for the
                 * prefix, handled below.
                 */
                const suffix = /\w$/u.test(term) ? "\\b" : "";
                if (location === "start") {
                    /*
                     * When matching at the start, ignore leading whitespace, and
                     * there's no need to worry about word boundaries.
                     */
                    prefix = "^\\s*";
                }
                else if (/^\w/u.test(term)) {
                    prefix = wordBoundary;
                }
                else {
                    prefix = "";
                }
                if (location === "start") {
                    /*
                     * For location "start" the regex should be
                     * ^\s*TERM\b.  This checks the word boundary
                     * at the beginning of the comment.
                     */
                    return new RegExp(prefix + escaped + suffix, "iu");
                }
                /*
                 * For location "anywhere" the regex should be
                 * \bTERM\b|\bTERM\b, this checks the entire comment
                 * for the term.
                 */
                return new RegExp(prefix +
                    escaped +
                    suffix +
                    eitherOrWordBoundary +
                    term +
                    wordBoundary, "iu");
            }
            const warningRegExps = warningTerms.map(convertToRegExp);
            /**
             * Checks the specified comment for matches of the configured warning terms and returns the matches.
             * @param {string} comment The comment which is checked.
             * @returns {Array} All matched warning terms for this comment.
             */
            function commentContainsWarningTerm(comment) {
                const matches = [];
                warningRegExps.forEach((regex, index) => {
                    if (regex.test(comment)) {
                        matches.push(warningTerms[index]);
                    }
                });
                return matches;
            }
            /**
             * Checks the specified node for matching warning comments and reports them.
             * @param {ASTNode} node The AST node being checked.
             * @returns {void} undefined.
             */
            function checkComment(node) {
                const comment = node.value;
                if (astUtils.isDirectiveComment(node) &&
                    selfConfigRegEx.test(comment)) {
                    return;
                }
                const matches = commentContainsWarningTerm(comment);
                matches.forEach(matchedTerm => {
                    let commentToDisplay = "";
                    let truncated = false;
                    for (const c of comment.trim().split(/\s+/u)) {
                        const tmp = commentToDisplay ? `${commentToDisplay} ${c}` : c;
                        if (tmp.length <= CHAR_LIMIT) {
                            commentToDisplay = tmp;
                        }
                        else {
                            truncated = true;
                            break;
                        }
                    }
                    context.report({
                        node,
                        messageId: "unexpectedComment",
                        data: {
                            matchedTerm,
                            comment: `${commentToDisplay}${truncated ? "..." : ""}`
                        }
                    });
                });
            }
            return {
                Program() {
                    const comments = sourceCode.getAllComments();
                    comments
                        .filter(token => token.type !== "Shebang")
                        .forEach(checkComment);
                }
            };
        }
    };
},
/* eslint/lib/rules/no-whitespace-before-property.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "disallow whitespace before properties",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/no-whitespace-before-property"
            },
            fixable: "whitespace",
            schema: [],
            messages: {
                unexpectedWhitespace: "Unexpected whitespace before property {{propName}}."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Reports whitespace before property token
             * @param {ASTNode} node the node to report in the event of an error
             * @param {Token} leftToken the left token
             * @param {Token} rightToken the right token
             * @returns {void}
             * @private
             */
            function reportError(node, leftToken, rightToken) {
                context.report({
                    node,
                    messageId: "unexpectedWhitespace",
                    data: {
                        propName: sourceCode.getText(node.property)
                    },
                    fix(fixer) {
                        let replacementText = "";
                        if (!node.computed && !node.optional && astUtils.isDecimalInteger(node.object)) {
                            /*
                             * If the object is a number literal, fixing it to something like 5.toString() would cause a SyntaxError.
                             * Don't fix this case.
                             */
                            return null;
                        }
                        // Don't fix if comments exist.
                        if (sourceCode.commentsExistBetween(leftToken, rightToken)) {
                            return null;
                        }
                        if (node.optional) {
                            replacementText = "?.";
                        }
                        else if (!node.computed) {
                            replacementText = ".";
                        }
                        return fixer.replaceTextRange([leftToken.range[1], rightToken.range[0]], replacementText);
                    }
                });
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                MemberExpression(node) {
                    let rightToken;
                    let leftToken;
                    if (!astUtils.isTokenOnSameLine(node.object, node.property)) {
                        return;
                    }
                    if (node.computed) {
                        rightToken = sourceCode.getTokenBefore(node.property, astUtils.isOpeningBracketToken);
                        leftToken = sourceCode.getTokenBefore(rightToken, node.optional ? 1 : 0);
                    }
                    else {
                        rightToken = sourceCode.getFirstToken(node.property);
                        leftToken = sourceCode.getTokenBefore(rightToken, 1);
                    }
                    if (sourceCode.isSpaceBetweenTokens(leftToken, rightToken)) {
                        reportError(node, leftToken, rightToken);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/no-with.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `with` statements",
                category: "Best Practices",
                recommended: true,
                url: "https://eslint.org/docs/rules/no-with"
            },
            schema: [],
            messages: {
                unexpectedWith: "Unexpected use of 'with' statement."
            }
        },
        create(context) {
            return {
                WithStatement(node) {
                    context.report({ node, messageId: "unexpectedWith" });
                }
            };
        }
    };
},
/* eslint/lib/rules/nonblock-statement-body-position.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const POSITION_SCHEMA = { enum: ["beside", "below", "any"] };
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce the location of single-line statements",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/nonblock-statement-body-position"
            },
            fixable: "whitespace",
            schema: [
                POSITION_SCHEMA,
                {
                    properties: {
                        overrides: {
                            properties: {
                                if: POSITION_SCHEMA,
                                else: POSITION_SCHEMA,
                                while: POSITION_SCHEMA,
                                do: POSITION_SCHEMA,
                                for: POSITION_SCHEMA
                            },
                            additionalProperties: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                expectNoLinebreak: "Expected no linebreak before this statement.",
                expectLinebreak: "Expected a linebreak before this statement."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            //----------------------------------------------------------------------
            // Helpers
            //----------------------------------------------------------------------
            /**
             * Gets the applicable preference for a particular keyword
             * @param {string} keywordName The name of a keyword, e.g. 'if'
             * @returns {string} The applicable option for the keyword, e.g. 'beside'
             */
            function getOption(keywordName) {
                return context.options[1] && context.options[1].overrides && context.options[1].overrides[keywordName] ||
                    context.options[0] ||
                    "beside";
            }
            /**
             * Validates the location of a single-line statement
             * @param {ASTNode} node The single-line statement
             * @param {string} keywordName The applicable keyword name for the single-line statement
             * @returns {void}
             */
            function validateStatement(node, keywordName) {
                const option = getOption(keywordName);
                if (node.type === "BlockStatement" || option === "any") {
                    return;
                }
                const tokenBefore = sourceCode.getTokenBefore(node);
                if (tokenBefore.loc.end.line === node.loc.start.line && option === "below") {
                    context.report({
                        node,
                        messageId: "expectLinebreak",
                        fix: fixer => fixer.insertTextBefore(node, "\n")
                    });
                }
                else if (tokenBefore.loc.end.line !== node.loc.start.line && option === "beside") {
                    context.report({
                        node,
                        messageId: "expectNoLinebreak",
                        fix(fixer) {
                            if (sourceCode.getText().slice(tokenBefore.range[1], node.range[0]).trim()) {
                                return null;
                            }
                            return fixer.replaceTextRange([tokenBefore.range[1], node.range[0]], " ");
                        }
                    });
                }
            }
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                IfStatement(node) {
                    validateStatement(node.consequent, "if");
                    // Check the `else` node, but don't check 'else if' statements.
                    if (node.alternate && node.alternate.type !== "IfStatement") {
                        validateStatement(node.alternate, "else");
                    }
                },
                WhileStatement: node => validateStatement(node.body, "while"),
                DoWhileStatement: node => validateStatement(node.body, "do"),
                ForStatement: node => validateStatement(node.body, "for"),
                ForInStatement: node => validateStatement(node.body, "for"),
                ForOfStatement: node => validateStatement(node.body, "for")
            };
        }
    };
},
/* eslint/lib/rules/object-curly-newline.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    // Schema objects.
    const OPTION_VALUE = {
        oneOf: [
            {
                enum: ["always", "never"]
            },
            {
                type: "object",
                properties: {
                    multiline: {
                        type: "boolean"
                    },
                    minProperties: {
                        type: "integer",
                        minimum: 0
                    },
                    consistent: {
                        type: "boolean"
                    }
                },
                additionalProperties: false,
                minProperties: 1
            }
        ]
    };
    /**
     * Normalizes a given option value.
     * @param {string|Object|undefined} value An option value to parse.
     * @returns {{multiline: boolean, minProperties: number, consistent: boolean}} Normalized option object.
     */
    function normalizeOptionValue(value) {
        let multiline = false;
        let minProperties = Number.POSITIVE_INFINITY;
        let consistent = false;
        if (value) {
            if (value === "always") {
                minProperties = 0;
            }
            else if (value === "never") {
                minProperties = Number.POSITIVE_INFINITY;
            }
            else {
                multiline = Boolean(value.multiline);
                minProperties = value.minProperties || Number.POSITIVE_INFINITY;
                consistent = Boolean(value.consistent);
            }
        }
        else {
            consistent = true;
        }
        return { multiline, minProperties, consistent };
    }
    /**
     * Checks if a value is an object.
     * @param {any} value The value to check
     * @returns {boolean} `true` if the value is an object, otherwise `false`
     */
    function isObject(value) {
        return typeof value === "object" && value !== null;
    }
    /**
     * Checks if an option is a node-specific option
     * @param {any} option The option to check
     * @returns {boolean} `true` if the option is node-specific, otherwise `false`
     */
    function isNodeSpecificOption(option) {
        return isObject(option) || typeof option === "string";
    }
    /**
     * Normalizes a given option value.
     * @param {string|Object|undefined} options An option value to parse.
     * @returns {{
     *   ObjectExpression: {multiline: boolean, minProperties: number, consistent: boolean},
     *   ObjectPattern: {multiline: boolean, minProperties: number, consistent: boolean},
     *   ImportDeclaration: {multiline: boolean, minProperties: number, consistent: boolean},
     *   ExportNamedDeclaration : {multiline: boolean, minProperties: number, consistent: boolean}
     * }} Normalized option object.
     */
    function normalizeOptions(options) {
        if (isObject(options) && Object.values(options).some(isNodeSpecificOption)) {
            return {
                ObjectExpression: normalizeOptionValue(options.ObjectExpression),
                ObjectPattern: normalizeOptionValue(options.ObjectPattern),
                ImportDeclaration: normalizeOptionValue(options.ImportDeclaration),
                ExportNamedDeclaration: normalizeOptionValue(options.ExportDeclaration)
            };
        }
        const value = normalizeOptionValue(options);
        return { ObjectExpression: value, ObjectPattern: value, ImportDeclaration: value, ExportNamedDeclaration: value };
    }
    /**
     * Determines if ObjectExpression, ObjectPattern, ImportDeclaration or ExportNamedDeclaration
     * node needs to be checked for missing line breaks
     * @param {ASTNode} node Node under inspection
     * @param {Object} options option specific to node type
     * @param {Token} first First object property
     * @param {Token} last Last object property
     * @returns {boolean} `true` if node needs to be checked for missing line breaks
     */
    function areLineBreaksRequired(node, options, first, last) {
        let objectProperties;
        if (node.type === "ObjectExpression" || node.type === "ObjectPattern") {
            objectProperties = node.properties;
        }
        else {
            // is ImportDeclaration or ExportNamedDeclaration
            objectProperties = node.specifiers
                .filter(s => s.type === "ImportSpecifier" || s.type === "ExportSpecifier");
        }
        return objectProperties.length >= options.minProperties ||
            (options.multiline &&
                objectProperties.length > 0 &&
                first.loc.start.line !== last.loc.end.line);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent line breaks after opening and before closing braces",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/object-curly-newline"
            },
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        OPTION_VALUE,
                        {
                            type: "object",
                            properties: {
                                ObjectExpression: OPTION_VALUE,
                                ObjectPattern: OPTION_VALUE,
                                ImportDeclaration: OPTION_VALUE,
                                ExportDeclaration: OPTION_VALUE
                            },
                            additionalProperties: false,
                            minProperties: 1
                        }
                    ]
                }
            ],
            messages: {
                unexpectedLinebreakBeforeClosingBrace: "Unexpected line break before this closing brace.",
                unexpectedLinebreakAfterOpeningBrace: "Unexpected line break after this opening brace.",
                expectedLinebreakBeforeClosingBrace: "Expected a line break before this closing brace.",
                expectedLinebreakAfterOpeningBrace: "Expected a line break after this opening brace."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const normalizedOptions = normalizeOptions(context.options[0]);
            /**
             * Reports a given node if it violated this rule.
             * @param {ASTNode} node A node to check. This is an ObjectExpression, ObjectPattern, ImportDeclaration or ExportNamedDeclaration node.
             * @returns {void}
             */
            function check(node) {
                const options = normalizedOptions[node.type];
                if ((node.type === "ImportDeclaration" &&
                    !node.specifiers.some(specifier => specifier.type === "ImportSpecifier")) ||
                    (node.type === "ExportNamedDeclaration" &&
                        !node.specifiers.some(specifier => specifier.type === "ExportSpecifier"))) {
                    return;
                }
                const openBrace = sourceCode.getFirstToken(node, token => token.value === "{");
                let closeBrace;
                if (node.typeAnnotation) {
                    closeBrace = sourceCode.getTokenBefore(node.typeAnnotation);
                }
                else {
                    closeBrace = sourceCode.getLastToken(node, token => token.value === "}");
                }
                let first = sourceCode.getTokenAfter(openBrace, { includeComments: true });
                let last = sourceCode.getTokenBefore(closeBrace, { includeComments: true });
                const needsLineBreaks = areLineBreaksRequired(node, options, first, last);
                const hasCommentsFirstToken = astUtils.isCommentToken(first);
                const hasCommentsLastToken = astUtils.isCommentToken(last);
                /*
                 * Use tokens or comments to check multiline or not.
                 * But use only tokens to check whether line breaks are needed.
                 * This allows:
                 *     var obj = { // eslint-disable-line foo
                 *         a: 1
                 *     }
                 */
                first = sourceCode.getTokenAfter(openBrace);
                last = sourceCode.getTokenBefore(closeBrace);
                if (needsLineBreaks) {
                    if (astUtils.isTokenOnSameLine(openBrace, first)) {
                        context.report({
                            messageId: "expectedLinebreakAfterOpeningBrace",
                            node,
                            loc: openBrace.loc,
                            fix(fixer) {
                                if (hasCommentsFirstToken) {
                                    return null;
                                }
                                return fixer.insertTextAfter(openBrace, "\n");
                            }
                        });
                    }
                    if (astUtils.isTokenOnSameLine(last, closeBrace)) {
                        context.report({
                            messageId: "expectedLinebreakBeforeClosingBrace",
                            node,
                            loc: closeBrace.loc,
                            fix(fixer) {
                                if (hasCommentsLastToken) {
                                    return null;
                                }
                                return fixer.insertTextBefore(closeBrace, "\n");
                            }
                        });
                    }
                }
                else {
                    const consistent = options.consistent;
                    const hasLineBreakBetweenOpenBraceAndFirst = !astUtils.isTokenOnSameLine(openBrace, first);
                    const hasLineBreakBetweenCloseBraceAndLast = !astUtils.isTokenOnSameLine(last, closeBrace);
                    if ((!consistent && hasLineBreakBetweenOpenBraceAndFirst) ||
                        (consistent && hasLineBreakBetweenOpenBraceAndFirst && !hasLineBreakBetweenCloseBraceAndLast)) {
                        context.report({
                            messageId: "unexpectedLinebreakAfterOpeningBrace",
                            node,
                            loc: openBrace.loc,
                            fix(fixer) {
                                if (hasCommentsFirstToken) {
                                    return null;
                                }
                                return fixer.removeRange([
                                    openBrace.range[1],
                                    first.range[0]
                                ]);
                            }
                        });
                    }
                    if ((!consistent && hasLineBreakBetweenCloseBraceAndLast) ||
                        (consistent && !hasLineBreakBetweenOpenBraceAndFirst && hasLineBreakBetweenCloseBraceAndLast)) {
                        context.report({
                            messageId: "unexpectedLinebreakBeforeClosingBrace",
                            node,
                            loc: closeBrace.loc,
                            fix(fixer) {
                                if (hasCommentsLastToken) {
                                    return null;
                                }
                                return fixer.removeRange([
                                    last.range[1],
                                    closeBrace.range[0]
                                ]);
                            }
                        });
                    }
                }
            }
            return {
                ObjectExpression: check,
                ObjectPattern: check,
                ImportDeclaration: check,
                ExportNamedDeclaration: check
            };
        }
    };
},
/* eslint/lib/rules/object-curly-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing inside braces",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/object-curly-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never"]
                },
                {
                    type: "object",
                    properties: {
                        arraysInObjects: {
                            type: "boolean"
                        },
                        objectsInObjects: {
                            type: "boolean"
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                requireSpaceBefore: "A space is required before '{{token}}'.",
                requireSpaceAfter: "A space is required after '{{token}}'.",
                unexpectedSpaceBefore: "There should be no space before '{{token}}'.",
                unexpectedSpaceAfter: "There should be no space after '{{token}}'."
            }
        },
        create(context) {
            const spaced = context.options[0] === "always", sourceCode = context.getSourceCode();
            /**
             * Determines whether an option is set, relative to the spacing option.
             * If spaced is "always", then check whether option is set to false.
             * If spaced is "never", then check whether option is set to true.
             * @param {Object} option The option to exclude.
             * @returns {boolean} Whether or not the property is excluded.
             */
            function isOptionSet(option) {
                return context.options[1] ? context.options[1][option] === !spaced : false;
            }
            const options = {
                spaced,
                arraysInObjectsException: isOptionSet("arraysInObjects"),
                objectsInObjectsException: isOptionSet("objectsInObjects")
            };
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Reports that there shouldn't be a space after the first token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportNoBeginningSpace(node, token) {
                const nextToken = context.getSourceCode().getTokenAfter(token, { includeComments: true });
                context.report({
                    node,
                    loc: { start: token.loc.end, end: nextToken.loc.start },
                    messageId: "unexpectedSpaceAfter",
                    data: {
                        token: token.value
                    },
                    fix(fixer) {
                        return fixer.removeRange([token.range[1], nextToken.range[0]]);
                    }
                });
            }
            /**
             * Reports that there shouldn't be a space before the last token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportNoEndingSpace(node, token) {
                const previousToken = context.getSourceCode().getTokenBefore(token, { includeComments: true });
                context.report({
                    node,
                    loc: { start: previousToken.loc.end, end: token.loc.start },
                    messageId: "unexpectedSpaceBefore",
                    data: {
                        token: token.value
                    },
                    fix(fixer) {
                        return fixer.removeRange([previousToken.range[1], token.range[0]]);
                    }
                });
            }
            /**
             * Reports that there should be a space after the first token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportRequiredBeginningSpace(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "requireSpaceAfter",
                    data: {
                        token: token.value
                    },
                    fix(fixer) {
                        return fixer.insertTextAfter(token, " ");
                    }
                });
            }
            /**
             * Reports that there should be a space before the last token
             * @param {ASTNode} node The node to report in the event of an error.
             * @param {Token} token The token to use for the report.
             * @returns {void}
             */
            function reportRequiredEndingSpace(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: "requireSpaceBefore",
                    data: {
                        token: token.value
                    },
                    fix(fixer) {
                        return fixer.insertTextBefore(token, " ");
                    }
                });
            }
            /**
             * Determines if spacing in curly braces is valid.
             * @param {ASTNode} node The AST node to check.
             * @param {Token} first The first token to check (should be the opening brace)
             * @param {Token} second The second token to check (should be first after the opening brace)
             * @param {Token} penultimate The penultimate token to check (should be last before closing brace)
             * @param {Token} last The last token to check (should be closing brace)
             * @returns {void}
             */
            function validateBraceSpacing(node, first, second, penultimate, last) {
                if (astUtils.isTokenOnSameLine(first, second)) {
                    const firstSpaced = sourceCode.isSpaceBetweenTokens(first, second);
                    if (options.spaced && !firstSpaced) {
                        reportRequiredBeginningSpace(node, first);
                    }
                    if (!options.spaced && firstSpaced && second.type !== "Line") {
                        reportNoBeginningSpace(node, first);
                    }
                }
                if (astUtils.isTokenOnSameLine(penultimate, last)) {
                    const shouldCheckPenultimate = (options.arraysInObjectsException && astUtils.isClosingBracketToken(penultimate) ||
                        options.objectsInObjectsException && astUtils.isClosingBraceToken(penultimate));
                    const penultimateType = shouldCheckPenultimate && sourceCode.getNodeByRangeIndex(penultimate.range[0]).type;
                    const closingCurlyBraceMustBeSpaced = (options.arraysInObjectsException && penultimateType === "ArrayExpression" ||
                        options.objectsInObjectsException && (penultimateType === "ObjectExpression" || penultimateType === "ObjectPattern")) ? !options.spaced : options.spaced;
                    const lastSpaced = sourceCode.isSpaceBetweenTokens(penultimate, last);
                    if (closingCurlyBraceMustBeSpaced && !lastSpaced) {
                        reportRequiredEndingSpace(node, last);
                    }
                    if (!closingCurlyBraceMustBeSpaced && lastSpaced) {
                        reportNoEndingSpace(node, last);
                    }
                }
            }
            /**
             * Gets '}' token of an object node.
             *
             * Because the last token of object patterns might be a type annotation,
             * this traverses tokens preceded by the last property, then returns the
             * first '}' token.
             * @param {ASTNode} node The node to get. This node is an
             *      ObjectExpression or an ObjectPattern. And this node has one or
             *      more properties.
             * @returns {Token} '}' token.
             */
            function getClosingBraceOfObject(node) {
                const lastProperty = node.properties[node.properties.length - 1];
                return sourceCode.getTokenAfter(lastProperty, astUtils.isClosingBraceToken);
            }
            /**
             * Reports a given object node if spacing in curly braces is invalid.
             * @param {ASTNode} node An ObjectExpression or ObjectPattern node to check.
             * @returns {void}
             */
            function checkForObject(node) {
                if (node.properties.length === 0) {
                    return;
                }
                const first = sourceCode.getFirstToken(node), last = getClosingBraceOfObject(node), second = sourceCode.getTokenAfter(first, { includeComments: true }), penultimate = sourceCode.getTokenBefore(last, { includeComments: true });
                validateBraceSpacing(node, first, second, penultimate, last);
            }
            /**
             * Reports a given import node if spacing in curly braces is invalid.
             * @param {ASTNode} node An ImportDeclaration node to check.
             * @returns {void}
             */
            function checkForImport(node) {
                if (node.specifiers.length === 0) {
                    return;
                }
                let firstSpecifier = node.specifiers[0];
                const lastSpecifier = node.specifiers[node.specifiers.length - 1];
                if (lastSpecifier.type !== "ImportSpecifier") {
                    return;
                }
                if (firstSpecifier.type !== "ImportSpecifier") {
                    firstSpecifier = node.specifiers[1];
                }
                const first = sourceCode.getTokenBefore(firstSpecifier), last = sourceCode.getTokenAfter(lastSpecifier, astUtils.isNotCommaToken), second = sourceCode.getTokenAfter(first, { includeComments: true }), penultimate = sourceCode.getTokenBefore(last, { includeComments: true });
                validateBraceSpacing(node, first, second, penultimate, last);
            }
            /**
             * Reports a given export node if spacing in curly braces is invalid.
             * @param {ASTNode} node An ExportNamedDeclaration node to check.
             * @returns {void}
             */
            function checkForExport(node) {
                if (node.specifiers.length === 0) {
                    return;
                }
                const firstSpecifier = node.specifiers[0], lastSpecifier = node.specifiers[node.specifiers.length - 1], first = sourceCode.getTokenBefore(firstSpecifier), last = sourceCode.getTokenAfter(lastSpecifier, astUtils.isNotCommaToken), second = sourceCode.getTokenAfter(first, { includeComments: true }), penultimate = sourceCode.getTokenBefore(last, { includeComments: true });
                validateBraceSpacing(node, first, second, penultimate, last);
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                // var {x} = y;
                ObjectPattern: checkForObject,
                // var y = {x: 'y'}
                ObjectExpression: checkForObject,
                // import {y} from 'x';
                ImportDeclaration: checkForImport,
                // export {name} from 'yo';
                ExportNamedDeclaration: checkForExport
            };
        }
    };
},
/* eslint/lib/rules/object-property-newline.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce placing object properties on separate lines",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/object-property-newline"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowAllPropertiesOnSameLine: {
                            type: "boolean",
                            default: false
                        },
                        allowMultiplePropertiesPerLine: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "whitespace",
            messages: {
                propertiesOnNewlineAll: "Object properties must go on a new line if they aren't all on the same line.",
                propertiesOnNewline: "Object properties must go on a new line."
            }
        },
        create(context) {
            const allowSameLine = context.options[0] && ((context.options[0].allowAllPropertiesOnSameLine || context.options[0].allowMultiplePropertiesPerLine /* Deprecated */));
            const messageId = allowSameLine
                ? "propertiesOnNewlineAll"
                : "propertiesOnNewline";
            const sourceCode = context.getSourceCode();
            return {
                ObjectExpression(node) {
                    if (allowSameLine) {
                        if (node.properties.length > 1) {
                            const firstTokenOfFirstProperty = sourceCode.getFirstToken(node.properties[0]);
                            const lastTokenOfLastProperty = sourceCode.getLastToken(node.properties[node.properties.length - 1]);
                            if (firstTokenOfFirstProperty.loc.end.line === lastTokenOfLastProperty.loc.start.line) {
                                // All keys and values are on the same line
                                return;
                            }
                        }
                    }
                    for (let i = 1; i < node.properties.length; i++) {
                        const lastTokenOfPreviousProperty = sourceCode.getLastToken(node.properties[i - 1]);
                        const firstTokenOfCurrentProperty = sourceCode.getFirstToken(node.properties[i]);
                        if (lastTokenOfPreviousProperty.loc.end.line === firstTokenOfCurrentProperty.loc.start.line) {
                            context.report({
                                node,
                                loc: firstTokenOfCurrentProperty.loc,
                                messageId,
                                fix(fixer) {
                                    const comma = sourceCode.getTokenBefore(firstTokenOfCurrentProperty);
                                    const rangeAfterComma = [comma.range[1], firstTokenOfCurrentProperty.range[0]];
                                    // Don't perform a fix if there are any comments between the comma and the next property.
                                    if (sourceCode.text.slice(rangeAfterComma[0], rangeAfterComma[1]).trim()) {
                                        return null;
                                    }
                                    return fixer.replaceTextRange(rangeAfterComma, "\n");
                                }
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/object-shorthand.js */ function _(require, module, exports, __esModule, __esExport) {
    const OPTIONS = {
        always: "always",
        never: "never",
        methods: "methods",
        properties: "properties",
        consistent: "consistent",
        consistentAsNeeded: "consistent-as-needed"
    };
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require or disallow method and property shorthand syntax for object literals",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/object-shorthand"
            },
            fixable: "code",
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always", "methods", "properties", "never", "consistent", "consistent-as-needed"]
                            }
                        ],
                        minItems: 0,
                        maxItems: 1
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always", "methods", "properties"]
                            },
                            {
                                type: "object",
                                properties: {
                                    avoidQuotes: {
                                        type: "boolean"
                                    }
                                },
                                additionalProperties: false
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always", "methods"]
                            },
                            {
                                type: "object",
                                properties: {
                                    ignoreConstructors: {
                                        type: "boolean"
                                    },
                                    avoidQuotes: {
                                        type: "boolean"
                                    },
                                    avoidExplicitReturnArrows: {
                                        type: "boolean"
                                    }
                                },
                                additionalProperties: false
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    }
                ]
            },
            messages: {
                expectedAllPropertiesShorthanded: "Expected shorthand for all properties.",
                expectedLiteralMethodLongform: "Expected longform method syntax for string literal keys.",
                expectedPropertyShorthand: "Expected property shorthand.",
                expectedPropertyLongform: "Expected longform property syntax.",
                expectedMethodShorthand: "Expected method shorthand.",
                expectedMethodLongform: "Expected longform method syntax.",
                unexpectedMix: "Unexpected mix of shorthand and non-shorthand properties."
            }
        },
        create(context) {
            const APPLY = context.options[0] || OPTIONS.always;
            const APPLY_TO_METHODS = APPLY === OPTIONS.methods || APPLY === OPTIONS.always;
            const APPLY_TO_PROPS = APPLY === OPTIONS.properties || APPLY === OPTIONS.always;
            const APPLY_NEVER = APPLY === OPTIONS.never;
            const APPLY_CONSISTENT = APPLY === OPTIONS.consistent;
            const APPLY_CONSISTENT_AS_NEEDED = APPLY === OPTIONS.consistentAsNeeded;
            const PARAMS = context.options[1] || {};
            const IGNORE_CONSTRUCTORS = PARAMS.ignoreConstructors;
            const AVOID_QUOTES = PARAMS.avoidQuotes;
            const AVOID_EXPLICIT_RETURN_ARROWS = !!PARAMS.avoidExplicitReturnArrows;
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            const CTOR_PREFIX_REGEX = /[^_$0-9]/u;
            /**
             * Determines if the first character of the name is a capital letter.
             * @param {string} name The name of the node to evaluate.
             * @returns {boolean} True if the first character of the property name is a capital letter, false if not.
             * @private
             */
            function isConstructor(name) {
                const match = CTOR_PREFIX_REGEX.exec(name);
                // Not a constructor if name has no characters apart from '_', '$' and digits e.g. '_', '$$', '_8'
                if (!match) {
                    return false;
                }
                const firstChar = name.charAt(match.index);
                return firstChar === firstChar.toUpperCase();
            }
            /**
             * Determines if the property can have a shorthand form.
             * @param {ASTNode} property Property AST node
             * @returns {boolean} True if the property can have a shorthand form
             * @private
             *
             */
            function canHaveShorthand(property) {
                return (property.kind !== "set" && property.kind !== "get" && property.type !== "SpreadElement" && property.type !== "SpreadProperty" && property.type !== "ExperimentalSpreadProperty");
            }
            /**
             * Checks whether a node is a string literal.
             * @param   {ASTNode} node Any AST node.
             * @returns {boolean} `true` if it is a string literal.
             */
            function isStringLiteral(node) {
                return node.type === "Literal" && typeof node.value === "string";
            }
            /**
             * Determines if the property is a shorthand or not.
             * @param {ASTNode} property Property AST node
             * @returns {boolean} True if the property is considered shorthand, false if not.
             * @private
             *
             */
            function isShorthand(property) {
                // property.method is true when `{a(){}}`.
                return (property.shorthand || property.method);
            }
            /**
             * Determines if the property's key and method or value are named equally.
             * @param {ASTNode} property Property AST node
             * @returns {boolean} True if the key and value are named equally, false if not.
             * @private
             *
             */
            function isRedundant(property) {
                const value = property.value;
                if (value.type === "FunctionExpression") {
                    return !value.id; // Only anonymous should be shorthand method.
                }
                if (value.type === "Identifier") {
                    return astUtils.getStaticPropertyName(property) === value.name;
                }
                return false;
            }
            /**
             * Ensures that an object's properties are consistently shorthand, or not shorthand at all.
             * @param   {ASTNode} node Property AST node
             * @param   {boolean} checkRedundancy Whether to check longform redundancy
             * @returns {void}
             *
             */
            function checkConsistency(node, checkRedundancy) {
                // We are excluding getters/setters and spread properties as they are considered neither longform nor shorthand.
                const properties = node.properties.filter(canHaveShorthand);
                // Do we still have properties left after filtering the getters and setters?
                if (properties.length > 0) {
                    const shorthandProperties = properties.filter(isShorthand);
                    /*
                     * If we do not have an equal number of longform properties as
                     * shorthand properties, we are using the annotations inconsistently
                     */
                    if (shorthandProperties.length !== properties.length) {
                        // We have at least 1 shorthand property
                        if (shorthandProperties.length > 0) {
                            context.report({ node, messageId: "unexpectedMix" });
                        }
                        else if (checkRedundancy) {
                            /*
                             * If all properties of the object contain a method or value with a name matching it's key,
                             * all the keys are redundant.
                             */
                            const canAlwaysUseShorthand = properties.every(isRedundant);
                            if (canAlwaysUseShorthand) {
                                context.report({ node, messageId: "expectedAllPropertiesShorthanded" });
                            }
                        }
                    }
                }
            }
            /**
             * Fixes a FunctionExpression node by making it into a shorthand property.
             * @param {SourceCodeFixer} fixer The fixer object
             * @param {ASTNode} node A `Property` node that has a `FunctionExpression` or `ArrowFunctionExpression` as its value
             * @returns {Object} A fix for this node
             */
            function makeFunctionShorthand(fixer, node) {
                const firstKeyToken = node.computed
                    ? sourceCode.getFirstToken(node, astUtils.isOpeningBracketToken)
                    : sourceCode.getFirstToken(node.key);
                const lastKeyToken = node.computed
                    ? sourceCode.getFirstTokenBetween(node.key, node.value, astUtils.isClosingBracketToken)
                    : sourceCode.getLastToken(node.key);
                const keyText = sourceCode.text.slice(firstKeyToken.range[0], lastKeyToken.range[1]);
                let keyPrefix = "";
                // key: /* */ () => {}
                if (sourceCode.commentsExistBetween(lastKeyToken, node.value)) {
                    return null;
                }
                if (node.value.async) {
                    keyPrefix += "async ";
                }
                if (node.value.generator) {
                    keyPrefix += "*";
                }
                const fixRange = [firstKeyToken.range[0], node.range[1]];
                const methodPrefix = keyPrefix + keyText;
                if (node.value.type === "FunctionExpression") {
                    const functionToken = sourceCode.getTokens(node.value).find(token => token.type === "Keyword" && token.value === "function");
                    const tokenBeforeParams = node.value.generator ? sourceCode.getTokenAfter(functionToken) : functionToken;
                    return fixer.replaceTextRange(fixRange, methodPrefix + sourceCode.text.slice(tokenBeforeParams.range[1], node.value.range[1]));
                }
                const arrowToken = sourceCode.getTokenBefore(node.value.body, astUtils.isArrowToken);
                const fnBody = sourceCode.text.slice(arrowToken.range[1], node.value.range[1]);
                let shouldAddParensAroundParameters = false;
                let tokenBeforeParams;
                if (node.value.params.length === 0) {
                    tokenBeforeParams = sourceCode.getFirstToken(node.value, astUtils.isOpeningParenToken);
                }
                else {
                    tokenBeforeParams = sourceCode.getTokenBefore(node.value.params[0]);
                }
                if (node.value.params.length === 1) {
                    const hasParen = astUtils.isOpeningParenToken(tokenBeforeParams);
                    const isTokenOutsideNode = tokenBeforeParams.range[0] < node.range[0];
                    shouldAddParensAroundParameters = !hasParen || isTokenOutsideNode;
                }
                const sliceStart = shouldAddParensAroundParameters
                    ? node.value.params[0].range[0]
                    : tokenBeforeParams.range[0];
                const sliceEnd = sourceCode.getTokenBefore(arrowToken).range[1];
                const oldParamText = sourceCode.text.slice(sliceStart, sliceEnd);
                const newParamText = shouldAddParensAroundParameters ? `(${oldParamText})` : oldParamText;
                return fixer.replaceTextRange(fixRange, methodPrefix + newParamText + fnBody);
            }
            /**
             * Fixes a FunctionExpression node by making it into a longform property.
             * @param {SourceCodeFixer} fixer The fixer object
             * @param {ASTNode} node A `Property` node that has a `FunctionExpression` as its value
             * @returns {Object} A fix for this node
             */
            function makeFunctionLongform(fixer, node) {
                const firstKeyToken = node.computed ? sourceCode.getTokens(node).find(token => token.value === "[") : sourceCode.getFirstToken(node.key);
                const lastKeyToken = node.computed ? sourceCode.getTokensBetween(node.key, node.value).find(token => token.value === "]") : sourceCode.getLastToken(node.key);
                const keyText = sourceCode.text.slice(firstKeyToken.range[0], lastKeyToken.range[1]);
                let functionHeader = "function";
                if (node.value.async) {
                    functionHeader = `async ${functionHeader}`;
                }
                if (node.value.generator) {
                    functionHeader = `${functionHeader}*`;
                }
                return fixer.replaceTextRange([node.range[0], lastKeyToken.range[1]], `${keyText}: ${functionHeader}`);
            }
            /*
             * To determine whether a given arrow function has a lexical identifier (`this`, `arguments`, `super`, or `new.target`),
             * create a stack of functions that define these identifiers (i.e. all functions except arrow functions) as the AST is
             * traversed. Whenever a new function is encountered, create a new entry on the stack (corresponding to a different lexical
             * scope of `this`), and whenever a function is exited, pop that entry off the stack. When an arrow function is entered,
             * keep a reference to it on the current stack entry, and remove that reference when the arrow function is exited.
             * When a lexical identifier is encountered, mark all the arrow functions on the current stack entry by adding them
             * to an `arrowsWithLexicalIdentifiers` set. Any arrow function in that set will not be reported by this rule,
             * because converting it into a method would change the value of one of the lexical identifiers.
             */
            const lexicalScopeStack = [];
            const arrowsWithLexicalIdentifiers = new WeakSet();
            const argumentsIdentifiers = new WeakSet();
            /**
             * Enters a function. This creates a new lexical identifier scope, so a new Set of arrow functions is pushed onto the stack.
             * Also, this marks all `arguments` identifiers so that they can be detected later.
             * @returns {void}
             */
            function enterFunction() {
                lexicalScopeStack.unshift(new Set());
                context.getScope().variables.filter(variable => variable.name === "arguments").forEach(variable => {
                    variable.references.map(ref => ref.identifier).forEach(identifier => argumentsIdentifiers.add(identifier));
                });
            }
            /**
             * Exits a function. This pops the current set of arrow functions off the lexical scope stack.
             * @returns {void}
             */
            function exitFunction() {
                lexicalScopeStack.shift();
            }
            /**
             * Marks the current function as having a lexical keyword. This implies that all arrow functions
             * in the current lexical scope contain a reference to this lexical keyword.
             * @returns {void}
             */
            function reportLexicalIdentifier() {
                lexicalScopeStack[0].forEach(arrowFunction => arrowsWithLexicalIdentifiers.add(arrowFunction));
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program: enterFunction,
                FunctionDeclaration: enterFunction,
                FunctionExpression: enterFunction,
                "Program:exit": exitFunction,
                "FunctionDeclaration:exit": exitFunction,
                "FunctionExpression:exit": exitFunction,
                ArrowFunctionExpression(node) {
                    lexicalScopeStack[0].add(node);
                },
                "ArrowFunctionExpression:exit"(node) {
                    lexicalScopeStack[0].delete(node);
                },
                ThisExpression: reportLexicalIdentifier,
                Super: reportLexicalIdentifier,
                MetaProperty(node) {
                    if (node.meta.name === "new" && node.property.name === "target") {
                        reportLexicalIdentifier();
                    }
                },
                Identifier(node) {
                    if (argumentsIdentifiers.has(node)) {
                        reportLexicalIdentifier();
                    }
                },
                ObjectExpression(node) {
                    if (APPLY_CONSISTENT) {
                        checkConsistency(node, false);
                    }
                    else if (APPLY_CONSISTENT_AS_NEEDED) {
                        checkConsistency(node, true);
                    }
                },
                "Property:exit"(node) {
                    const isConciseProperty = node.method || node.shorthand;
                    // Ignore destructuring assignment
                    if (node.parent.type === "ObjectPattern") {
                        return;
                    }
                    // getters and setters are ignored
                    if (node.kind === "get" || node.kind === "set") {
                        return;
                    }
                    // only computed methods can fail the following checks
                    if (node.computed && node.value.type !== "FunctionExpression" && node.value.type !== "ArrowFunctionExpression") {
                        return;
                    }
                    //--------------------------------------------------------------
                    // Checks for property/method shorthand.
                    if (isConciseProperty) {
                        if (node.method && (APPLY_NEVER || AVOID_QUOTES && isStringLiteral(node.key))) {
                            const messageId = APPLY_NEVER ? "expectedMethodLongform" : "expectedLiteralMethodLongform";
                            // { x() {} } should be written as { x: function() {} }
                            context.report({
                                node,
                                messageId,
                                fix: fixer => makeFunctionLongform(fixer, node)
                            });
                        }
                        else if (APPLY_NEVER) {
                            // { x } should be written as { x: x }
                            context.report({
                                node,
                                messageId: "expectedPropertyLongform",
                                fix: fixer => fixer.insertTextAfter(node.key, `: ${node.key.name}`)
                            });
                        }
                    }
                    else if (APPLY_TO_METHODS && !node.value.id && (node.value.type === "FunctionExpression" || node.value.type === "ArrowFunctionExpression")) {
                        if (IGNORE_CONSTRUCTORS && node.key.type === "Identifier" && isConstructor(node.key.name)) {
                            return;
                        }
                        if (AVOID_QUOTES && isStringLiteral(node.key)) {
                            return;
                        }
                        // {[x]: function(){}} should be written as {[x]() {}}
                        if (node.value.type === "FunctionExpression" ||
                            node.value.type === "ArrowFunctionExpression" &&
                                node.value.body.type === "BlockStatement" &&
                                AVOID_EXPLICIT_RETURN_ARROWS &&
                                !arrowsWithLexicalIdentifiers.has(node.value)) {
                            context.report({
                                node,
                                messageId: "expectedMethodShorthand",
                                fix: fixer => makeFunctionShorthand(fixer, node)
                            });
                        }
                    }
                    else if (node.value.type === "Identifier" && node.key.name === node.value.name && APPLY_TO_PROPS) {
                        // {x: x} should be written as {x}
                        context.report({
                            node,
                            messageId: "expectedPropertyShorthand",
                            fix(fixer) {
                                return fixer.replaceText(node, node.value.name);
                            }
                        });
                    }
                    else if (node.value.type === "Identifier" && node.key.type === "Literal" && node.key.value === node.value.name && APPLY_TO_PROPS) {
                        if (AVOID_QUOTES) {
                            return;
                        }
                        // {"x": x} should be written as {x}
                        context.report({
                            node,
                            messageId: "expectedPropertyShorthand",
                            fix(fixer) {
                                return fixer.replaceText(node, node.value.name);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/one-var.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Determines whether the given node is in a statement list.
     * @param {ASTNode} node node to check
     * @returns {boolean} `true` if the given node is in a statement list
     */
    function isInStatementList(node) {
        return astUtils.STATEMENT_LIST_PARENTS.has(node.parent.type);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce variables to be declared either together or separately in functions",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/one-var"
            },
            fixable: "code",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["always", "never", "consecutive"]
                        },
                        {
                            type: "object",
                            properties: {
                                separateRequires: {
                                    type: "boolean"
                                },
                                var: {
                                    enum: ["always", "never", "consecutive"]
                                },
                                let: {
                                    enum: ["always", "never", "consecutive"]
                                },
                                const: {
                                    enum: ["always", "never", "consecutive"]
                                }
                            },
                            additionalProperties: false
                        },
                        {
                            type: "object",
                            properties: {
                                initialized: {
                                    enum: ["always", "never", "consecutive"]
                                },
                                uninitialized: {
                                    enum: ["always", "never", "consecutive"]
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                combineUninitialized: "Combine this with the previous '{{type}}' statement with uninitialized variables.",
                combineInitialized: "Combine this with the previous '{{type}}' statement with initialized variables.",
                splitUninitialized: "Split uninitialized '{{type}}' declarations into multiple statements.",
                splitInitialized: "Split initialized '{{type}}' declarations into multiple statements.",
                splitRequires: "Split requires to be separated into a single block.",
                combine: "Combine this with the previous '{{type}}' statement.",
                split: "Split '{{type}}' declarations into multiple statements."
            }
        },
        create(context) {
            const MODE_ALWAYS = "always";
            const MODE_NEVER = "never";
            const MODE_CONSECUTIVE = "consecutive";
            const mode = context.options[0] || MODE_ALWAYS;
            const options = {};
            if (typeof mode === "string") { // simple options configuration with just a string
                options.var = { uninitialized: mode, initialized: mode };
                options.let = { uninitialized: mode, initialized: mode };
                options.const = { uninitialized: mode, initialized: mode };
            }
            else if (typeof mode === "object") { // options configuration is an object
                options.separateRequires = !!mode.separateRequires;
                options.var = { uninitialized: mode.var, initialized: mode.var };
                options.let = { uninitialized: mode.let, initialized: mode.let };
                options.const = { uninitialized: mode.const, initialized: mode.const };
                if (Object.prototype.hasOwnProperty.call(mode, "uninitialized")) {
                    options.var.uninitialized = mode.uninitialized;
                    options.let.uninitialized = mode.uninitialized;
                    options.const.uninitialized = mode.uninitialized;
                }
                if (Object.prototype.hasOwnProperty.call(mode, "initialized")) {
                    options.var.initialized = mode.initialized;
                    options.let.initialized = mode.initialized;
                    options.const.initialized = mode.initialized;
                }
            }
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            const functionStack = [];
            const blockStack = [];
            /**
             * Increments the blockStack counter.
             * @returns {void}
             * @private
             */
            function startBlock() {
                blockStack.push({
                    let: { initialized: false, uninitialized: false },
                    const: { initialized: false, uninitialized: false }
                });
            }
            /**
             * Increments the functionStack counter.
             * @returns {void}
             * @private
             */
            function startFunction() {
                functionStack.push({ initialized: false, uninitialized: false });
                startBlock();
            }
            /**
             * Decrements the blockStack counter.
             * @returns {void}
             * @private
             */
            function endBlock() {
                blockStack.pop();
            }
            /**
             * Decrements the functionStack counter.
             * @returns {void}
             * @private
             */
            function endFunction() {
                functionStack.pop();
                endBlock();
            }
            /**
             * Check if a variable declaration is a require.
             * @param {ASTNode} decl variable declaration Node
             * @returns {bool} if decl is a require, return true; else return false.
             * @private
             */
            function isRequire(decl) {
                return decl.init && decl.init.type === "CallExpression" && decl.init.callee.name === "require";
            }
            /**
             * Records whether initialized/uninitialized/required variables are defined in current scope.
             * @param {string} statementType node.kind, one of: "var", "let", or "const"
             * @param {ASTNode[]} declarations List of declarations
             * @param {Object} currentScope The scope being investigated
             * @returns {void}
             * @private
             */
            function recordTypes(statementType, declarations, currentScope) {
                for (let i = 0; i < declarations.length; i++) {
                    if (declarations[i].init === null) {
                        if (options[statementType] && options[statementType].uninitialized === MODE_ALWAYS) {
                            currentScope.uninitialized = true;
                        }
                    }
                    else {
                        if (options[statementType] && options[statementType].initialized === MODE_ALWAYS) {
                            if (options.separateRequires && isRequire(declarations[i])) {
                                currentScope.required = true;
                            }
                            else {
                                currentScope.initialized = true;
                            }
                        }
                    }
                }
            }
            /**
             * Determines the current scope (function or block)
             * @param  {string} statementType node.kind, one of: "var", "let", or "const"
             * @returns {Object} The scope associated with statementType
             */
            function getCurrentScope(statementType) {
                let currentScope;
                if (statementType === "var") {
                    currentScope = functionStack[functionStack.length - 1];
                }
                else if (statementType === "let") {
                    currentScope = blockStack[blockStack.length - 1].let;
                }
                else if (statementType === "const") {
                    currentScope = blockStack[blockStack.length - 1].const;
                }
                return currentScope;
            }
            /**
             * Counts the number of initialized and uninitialized declarations in a list of declarations
             * @param {ASTNode[]} declarations List of declarations
             * @returns {Object} Counts of 'uninitialized' and 'initialized' declarations
             * @private
             */
            function countDeclarations(declarations) {
                const counts = { uninitialized: 0, initialized: 0 };
                for (let i = 0; i < declarations.length; i++) {
                    if (declarations[i].init === null) {
                        counts.uninitialized++;
                    }
                    else {
                        counts.initialized++;
                    }
                }
                return counts;
            }
            /**
             * Determines if there is more than one var statement in the current scope.
             * @param {string} statementType node.kind, one of: "var", "let", or "const"
             * @param {ASTNode[]} declarations List of declarations
             * @returns {boolean} Returns true if it is the first var declaration, false if not.
             * @private
             */
            function hasOnlyOneStatement(statementType, declarations) {
                const declarationCounts = countDeclarations(declarations);
                const currentOptions = options[statementType] || {};
                const currentScope = getCurrentScope(statementType);
                const hasRequires = declarations.some(isRequire);
                if (currentOptions.uninitialized === MODE_ALWAYS && currentOptions.initialized === MODE_ALWAYS) {
                    if (currentScope.uninitialized || currentScope.initialized) {
                        if (!hasRequires) {
                            return false;
                        }
                    }
                }
                if (declarationCounts.uninitialized > 0) {
                    if (currentOptions.uninitialized === MODE_ALWAYS && currentScope.uninitialized) {
                        return false;
                    }
                }
                if (declarationCounts.initialized > 0) {
                    if (currentOptions.initialized === MODE_ALWAYS && currentScope.initialized) {
                        if (!hasRequires) {
                            return false;
                        }
                    }
                }
                if (currentScope.required && hasRequires) {
                    return false;
                }
                recordTypes(statementType, declarations, currentScope);
                return true;
            }
            /**
             * Fixer to join VariableDeclaration's into a single declaration
             * @param {VariableDeclarator[]} declarations The `VariableDeclaration` to join
             * @returns {Function} The fixer function
             */
            function joinDeclarations(declarations) {
                const declaration = declarations[0];
                const body = Array.isArray(declaration.parent.parent.body) ? declaration.parent.parent.body : [];
                const currentIndex = body.findIndex(node => node.range[0] === declaration.parent.range[0]);
                const previousNode = body[currentIndex - 1];
                return fixer => {
                    const type = sourceCode.getTokenBefore(declaration);
                    const prevSemi = sourceCode.getTokenBefore(type);
                    const res = [];
                    if (previousNode && previousNode.kind === sourceCode.getText(type)) {
                        if (prevSemi.value === ";") {
                            res.push(fixer.replaceText(prevSemi, ","));
                        }
                        else {
                            res.push(fixer.insertTextAfter(prevSemi, ","));
                        }
                        res.push(fixer.replaceText(type, ""));
                    }
                    return res;
                };
            }
            /**
             * Fixer to split a VariableDeclaration into individual declarations
             * @param {VariableDeclaration} declaration The `VariableDeclaration` to split
             * @returns {Function|null} The fixer function
             */
            function splitDeclarations(declaration) {
                const { parent } = declaration;
                // don't autofix code such as: if (foo) var x, y;
                if (!isInStatementList(parent.type === "ExportNamedDeclaration" ? parent : declaration)) {
                    return null;
                }
                return fixer => declaration.declarations.map(declarator => {
                    const tokenAfterDeclarator = sourceCode.getTokenAfter(declarator);
                    if (tokenAfterDeclarator === null) {
                        return null;
                    }
                    const afterComma = sourceCode.getTokenAfter(tokenAfterDeclarator, { includeComments: true });
                    if (tokenAfterDeclarator.value !== ",") {
                        return null;
                    }
                    const exportPlacement = declaration.parent.type === "ExportNamedDeclaration" ? "export " : "";
                    /*
                     * `var x,y`
                     * tokenAfterDeclarator ^^ afterComma
                     */
                    if (afterComma.range[0] === tokenAfterDeclarator.range[1]) {
                        return fixer.replaceText(tokenAfterDeclarator, `; ${exportPlacement}${declaration.kind} `);
                    }
                    /*
                     * `var x,
                     * tokenAfterDeclarator ^
                     *      y`
                     *      ^ afterComma
                     */
                    if (afterComma.loc.start.line > tokenAfterDeclarator.loc.end.line ||
                        afterComma.type === "Line" ||
                        afterComma.type === "Block") {
                        let lastComment = afterComma;
                        while (lastComment.type === "Line" || lastComment.type === "Block") {
                            lastComment = sourceCode.getTokenAfter(lastComment, { includeComments: true });
                        }
                        return fixer.replaceTextRange([tokenAfterDeclarator.range[0], lastComment.range[0]], `;${sourceCode.text.slice(tokenAfterDeclarator.range[1], lastComment.range[0])}${exportPlacement}${declaration.kind} `);
                    }
                    return fixer.replaceText(tokenAfterDeclarator, `; ${exportPlacement}${declaration.kind}`);
                }).filter(x => x);
            }
            /**
             * Checks a given VariableDeclaration node for errors.
             * @param {ASTNode} node The VariableDeclaration node to check
             * @returns {void}
             * @private
             */
            function checkVariableDeclaration(node) {
                const parent = node.parent;
                const type = node.kind;
                if (!options[type]) {
                    return;
                }
                const declarations = node.declarations;
                const declarationCounts = countDeclarations(declarations);
                const mixedRequires = declarations.some(isRequire) && !declarations.every(isRequire);
                if (options[type].initialized === MODE_ALWAYS) {
                    if (options.separateRequires && mixedRequires) {
                        context.report({
                            node,
                            messageId: "splitRequires"
                        });
                    }
                }
                // consecutive
                const nodeIndex = (parent.body && parent.body.length > 0 && parent.body.indexOf(node)) || 0;
                if (nodeIndex > 0) {
                    const previousNode = parent.body[nodeIndex - 1];
                    const isPreviousNodeDeclaration = previousNode.type === "VariableDeclaration";
                    const declarationsWithPrevious = declarations.concat(previousNode.declarations || []);
                    if (isPreviousNodeDeclaration &&
                        previousNode.kind === type &&
                        !(declarationsWithPrevious.some(isRequire) && !declarationsWithPrevious.every(isRequire))) {
                        const previousDeclCounts = countDeclarations(previousNode.declarations);
                        if (options[type].initialized === MODE_CONSECUTIVE && options[type].uninitialized === MODE_CONSECUTIVE) {
                            context.report({
                                node,
                                messageId: "combine",
                                data: {
                                    type
                                },
                                fix: joinDeclarations(declarations)
                            });
                        }
                        else if (options[type].initialized === MODE_CONSECUTIVE && declarationCounts.initialized > 0 && previousDeclCounts.initialized > 0) {
                            context.report({
                                node,
                                messageId: "combineInitialized",
                                data: {
                                    type
                                },
                                fix: joinDeclarations(declarations)
                            });
                        }
                        else if (options[type].uninitialized === MODE_CONSECUTIVE &&
                            declarationCounts.uninitialized > 0 &&
                            previousDeclCounts.uninitialized > 0) {
                            context.report({
                                node,
                                messageId: "combineUninitialized",
                                data: {
                                    type
                                },
                                fix: joinDeclarations(declarations)
                            });
                        }
                    }
                }
                // always
                if (!hasOnlyOneStatement(type, declarations)) {
                    if (options[type].initialized === MODE_ALWAYS && options[type].uninitialized === MODE_ALWAYS) {
                        context.report({
                            node,
                            messageId: "combine",
                            data: {
                                type
                            },
                            fix: joinDeclarations(declarations)
                        });
                    }
                    else {
                        if (options[type].initialized === MODE_ALWAYS && declarationCounts.initialized > 0) {
                            context.report({
                                node,
                                messageId: "combineInitialized",
                                data: {
                                    type
                                },
                                fix: joinDeclarations(declarations)
                            });
                        }
                        if (options[type].uninitialized === MODE_ALWAYS && declarationCounts.uninitialized > 0) {
                            if (node.parent.left === node && (node.parent.type === "ForInStatement" || node.parent.type === "ForOfStatement")) {
                                return;
                            }
                            context.report({
                                node,
                                messageId: "combineUninitialized",
                                data: {
                                    type
                                },
                                fix: joinDeclarations(declarations)
                            });
                        }
                    }
                }
                // never
                if (parent.type !== "ForStatement" || parent.init !== node) {
                    const totalDeclarations = declarationCounts.uninitialized + declarationCounts.initialized;
                    if (totalDeclarations > 1) {
                        if (options[type].initialized === MODE_NEVER && options[type].uninitialized === MODE_NEVER) {
                            // both initialized and uninitialized
                            context.report({
                                node,
                                messageId: "split",
                                data: {
                                    type
                                },
                                fix: splitDeclarations(node)
                            });
                        }
                        else if (options[type].initialized === MODE_NEVER && declarationCounts.initialized > 0) {
                            // initialized
                            context.report({
                                node,
                                messageId: "splitInitialized",
                                data: {
                                    type
                                },
                                fix: splitDeclarations(node)
                            });
                        }
                        else if (options[type].uninitialized === MODE_NEVER && declarationCounts.uninitialized > 0) {
                            // uninitialized
                            context.report({
                                node,
                                messageId: "splitUninitialized",
                                data: {
                                    type
                                },
                                fix: splitDeclarations(node)
                            });
                        }
                    }
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                Program: startFunction,
                FunctionDeclaration: startFunction,
                FunctionExpression: startFunction,
                ArrowFunctionExpression: startFunction,
                BlockStatement: startBlock,
                ForStatement: startBlock,
                ForInStatement: startBlock,
                ForOfStatement: startBlock,
                SwitchStatement: startBlock,
                VariableDeclaration: checkVariableDeclaration,
                "ForStatement:exit": endBlock,
                "ForOfStatement:exit": endBlock,
                "ForInStatement:exit": endBlock,
                "SwitchStatement:exit": endBlock,
                "BlockStatement:exit": endBlock,
                "Program:exit": endFunction,
                "FunctionDeclaration:exit": endFunction,
                "FunctionExpression:exit": endFunction,
                "ArrowFunctionExpression:exit": endFunction
            };
        }
    };
},
/* eslint/lib/rules/one-var-declaration-per-line.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require or disallow newlines around variable declarations",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/one-var-declaration-per-line"
            },
            schema: [
                {
                    enum: ["always", "initializations"]
                }
            ],
            fixable: "whitespace",
            messages: {
                expectVarOnNewline: "Expected variable declaration to be on a new line."
            }
        },
        create(context) {
            const always = context.options[0] === "always";
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Determine if provided keyword is a variant of for specifiers
             * @private
             * @param {string} keyword keyword to test
             * @returns {boolean} True if `keyword` is a variant of for specifier
             */
            function isForTypeSpecifier(keyword) {
                return keyword === "ForStatement" || keyword === "ForInStatement" || keyword === "ForOfStatement";
            }
            /**
             * Checks newlines around variable declarations.
             * @private
             * @param {ASTNode} node `VariableDeclaration` node to test
             * @returns {void}
             */
            function checkForNewLine(node) {
                if (isForTypeSpecifier(node.parent.type)) {
                    return;
                }
                const declarations = node.declarations;
                let prev;
                declarations.forEach(current => {
                    if (prev && prev.loc.end.line === current.loc.start.line) {
                        if (always || prev.init || current.init) {
                            context.report({
                                node,
                                messageId: "expectVarOnNewline",
                                loc: current.loc,
                                fix: fixer => fixer.insertTextBefore(current, "\n")
                            });
                        }
                    }
                    prev = current;
                });
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                VariableDeclaration: checkForNewLine
            };
        }
    };
},
/* eslint/lib/rules/operator-assignment.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether an operator is commutative and has an operator assignment
     * shorthand form.
     * @param   {string}  operator Operator to check.
     * @returns {boolean}          True if the operator is commutative and has a
     *     shorthand form.
     */
    function isCommutativeOperatorWithShorthand(operator) {
        return ["*", "&", "^", "|"].indexOf(operator) >= 0;
    }
    /**
     * Checks whether an operator is not commutative and has an operator assignment
     * shorthand form.
     * @param   {string}  operator Operator to check.
     * @returns {boolean}          True if the operator is not commutative and has
     *     a shorthand form.
     */
    function isNonCommutativeOperatorWithShorthand(operator) {
        return ["+", "-", "/", "%", "<<", ">>", ">>>", "**"].indexOf(operator) >= 0;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    /**
     * Determines if the left side of a node can be safely fixed (i.e. if it activates the same getters/setters and)
     * toString calls regardless of whether assignment shorthand is used)
     * @param {ASTNode} node The node on the left side of the expression
     * @returns {boolean} `true` if the node can be fixed
     */
    function canBeFixed(node) {
        return (node.type === "Identifier" ||
            (node.type === "MemberExpression" &&
                (node.object.type === "Identifier" || node.object.type === "ThisExpression") &&
                (!node.computed || node.property.type === "Literal")));
    }
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require or disallow assignment operator shorthand where possible",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/operator-assignment"
            },
            schema: [
                {
                    enum: ["always", "never"]
                }
            ],
            fixable: "code",
            messages: {
                replaced: "Assignment (=) can be replaced with operator assignment ({{operator}}=).",
                unexpected: "Unexpected operator assignment ({{operator}}=) shorthand."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Returns the operator token of an AssignmentExpression or BinaryExpression
             * @param {ASTNode} node An AssignmentExpression or BinaryExpression node
             * @returns {Token} The operator token in the node
             */
            function getOperatorToken(node) {
                return sourceCode.getFirstTokenBetween(node.left, node.right, token => token.value === node.operator);
            }
            /**
             * Ensures that an assignment uses the shorthand form where possible.
             * @param   {ASTNode} node An AssignmentExpression node.
             * @returns {void}
             */
            function verify(node) {
                if (node.operator !== "=" || node.right.type !== "BinaryExpression") {
                    return;
                }
                const left = node.left;
                const expr = node.right;
                const operator = expr.operator;
                if (isCommutativeOperatorWithShorthand(operator) || isNonCommutativeOperatorWithShorthand(operator)) {
                    if (astUtils.isSameReference(left, expr.left, true)) {
                        context.report({
                            node,
                            messageId: "replaced",
                            data: { operator },
                            fix(fixer) {
                                if (canBeFixed(left) && canBeFixed(expr.left)) {
                                    const equalsToken = getOperatorToken(node);
                                    const operatorToken = getOperatorToken(expr);
                                    const leftText = sourceCode.getText().slice(node.range[0], equalsToken.range[0]);
                                    const rightText = sourceCode.getText().slice(operatorToken.range[1], node.right.range[1]);
                                    // Check for comments that would be removed.
                                    if (sourceCode.commentsExistBetween(equalsToken, operatorToken)) {
                                        return null;
                                    }
                                    return fixer.replaceText(node, `${leftText}${expr.operator}=${rightText}`);
                                }
                                return null;
                            }
                        });
                    }
                    else if (astUtils.isSameReference(left, expr.right, true) && isCommutativeOperatorWithShorthand(operator)) {
                        /*
                         * This case can't be fixed safely.
                         * If `a` and `b` both have custom valueOf() behavior, then fixing `a = b * a` to `a *= b` would
                         * change the execution order of the valueOf() functions.
                         */
                        context.report({
                            node,
                            messageId: "replaced",
                            data: { operator }
                        });
                    }
                }
            }
            /**
             * Warns if an assignment expression uses operator assignment shorthand.
             * @param   {ASTNode} node An AssignmentExpression node.
             * @returns {void}
             */
            function prohibit(node) {
                if (node.operator !== "=" && !astUtils.isLogicalAssignmentOperator(node.operator)) {
                    context.report({
                        node,
                        messageId: "unexpected",
                        data: { operator: node.operator },
                        fix(fixer) {
                            if (canBeFixed(node.left)) {
                                const firstToken = sourceCode.getFirstToken(node);
                                const operatorToken = getOperatorToken(node);
                                const leftText = sourceCode.getText().slice(node.range[0], operatorToken.range[0]);
                                const newOperator = node.operator.slice(0, -1);
                                let rightText;
                                // Check for comments that would be duplicated.
                                if (sourceCode.commentsExistBetween(firstToken, operatorToken)) {
                                    return null;
                                }
                                // If this change would modify precedence (e.g. `foo *= bar + 1` => `foo = foo * (bar + 1)`), parenthesize the right side.
                                if (astUtils.getPrecedence(node.right) <= astUtils.getPrecedence({ type: "BinaryExpression", operator: newOperator }) &&
                                    !astUtils.isParenthesised(sourceCode, node.right)) {
                                    rightText = `${sourceCode.text.slice(operatorToken.range[1], node.right.range[0])}(${sourceCode.getText(node.right)})`;
                                }
                                else {
                                    const tokenAfterOperator = sourceCode.getTokenAfter(operatorToken, { includeComments: true });
                                    let rightTextPrefix = "";
                                    if (operatorToken.range[1] === tokenAfterOperator.range[0] &&
                                        !astUtils.canTokensBeAdjacent({ type: "Punctuator", value: newOperator }, tokenAfterOperator)) {
                                        rightTextPrefix = " "; // foo+=+bar -> foo= foo+ +bar
                                    }
                                    rightText = `${rightTextPrefix}${sourceCode.text.slice(operatorToken.range[1], node.range[1])}`;
                                }
                                return fixer.replaceText(node, `${leftText}= ${leftText}${newOperator}${rightText}`);
                            }
                            return null;
                        }
                    });
                }
            }
            return {
                AssignmentExpression: context.options[0] !== "never" ? verify : prohibit
            };
        }
    };
},
/* eslint/lib/rules/operator-linebreak.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent linebreak style for operators",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/operator-linebreak"
            },
            schema: [
                {
                    enum: ["after", "before", "none", null]
                },
                {
                    type: "object",
                    properties: {
                        overrides: {
                            type: "object",
                            additionalProperties: {
                                enum: ["after", "before", "none", "ignore"]
                            }
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "code",
            messages: {
                operatorAtBeginning: "'{{operator}}' should be placed at the beginning of the line.",
                operatorAtEnd: "'{{operator}}' should be placed at the end of the line.",
                badLinebreak: "Bad line breaking before and after '{{operator}}'.",
                noLinebreak: "There should be no line break before or after '{{operator}}'."
            }
        },
        create(context) {
            const usedDefaultGlobal = !context.options[0];
            const globalStyle = context.options[0] || "after";
            const options = context.options[1] || {};
            const styleOverrides = options.overrides ? Object.assign({}, options.overrides) : {};
            if (usedDefaultGlobal && !styleOverrides["?"]) {
                styleOverrides["?"] = "before";
            }
            if (usedDefaultGlobal && !styleOverrides[":"]) {
                styleOverrides[":"] = "before";
            }
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Gets a fixer function to fix rule issues
             * @param {Token} operatorToken The operator token of an expression
             * @param {string} desiredStyle The style for the rule. One of 'before', 'after', 'none'
             * @returns {Function} A fixer function
             */
            function getFixer(operatorToken, desiredStyle) {
                return fixer => {
                    const tokenBefore = sourceCode.getTokenBefore(operatorToken);
                    const tokenAfter = sourceCode.getTokenAfter(operatorToken);
                    const textBefore = sourceCode.text.slice(tokenBefore.range[1], operatorToken.range[0]);
                    const textAfter = sourceCode.text.slice(operatorToken.range[1], tokenAfter.range[0]);
                    const hasLinebreakBefore = !astUtils.isTokenOnSameLine(tokenBefore, operatorToken);
                    const hasLinebreakAfter = !astUtils.isTokenOnSameLine(operatorToken, tokenAfter);
                    let newTextBefore, newTextAfter;
                    if (hasLinebreakBefore !== hasLinebreakAfter && desiredStyle !== "none") {
                        // If there is a comment before and after the operator, don't do a fix.
                        if (sourceCode.getTokenBefore(operatorToken, { includeComments: true }) !== tokenBefore &&
                            sourceCode.getTokenAfter(operatorToken, { includeComments: true }) !== tokenAfter) {
                            return null;
                        }
                        /*
                         * If there is only one linebreak and it's on the wrong side of the operator, swap the text before and after the operator.
                         * foo &&
                         *           bar
                         * would get fixed to
                         * foo
                         *        && bar
                         */
                        newTextBefore = textAfter;
                        newTextAfter = textBefore;
                    }
                    else {
                        const LINEBREAK_REGEX = astUtils.createGlobalLinebreakMatcher();
                        // Otherwise, if no linebreak is desired and no comments interfere, replace the linebreaks with empty strings.
                        newTextBefore = desiredStyle === "before" || textBefore.trim() ? textBefore : textBefore.replace(LINEBREAK_REGEX, "");
                        newTextAfter = desiredStyle === "after" || textAfter.trim() ? textAfter : textAfter.replace(LINEBREAK_REGEX, "");
                        // If there was no change (due to interfering comments), don't output a fix.
                        if (newTextBefore === textBefore && newTextAfter === textAfter) {
                            return null;
                        }
                    }
                    if (newTextAfter === "" && tokenAfter.type === "Punctuator" && "+-".includes(operatorToken.value) && tokenAfter.value === operatorToken.value) {
                        // To avoid accidentally creating a ++ or -- operator, insert a space if the operator is a +/- and the following token is a unary +/-.
                        newTextAfter += " ";
                    }
                    return fixer.replaceTextRange([tokenBefore.range[1], tokenAfter.range[0]], newTextBefore + operatorToken.value + newTextAfter);
                };
            }
            /**
             * Checks the operator placement
             * @param {ASTNode} node The node to check
             * @param {ASTNode} leftSide The node that comes before the operator in `node`
             * @private
             * @returns {void}
             */
            function validateNode(node, leftSide) {
                /*
                 * When the left part of a binary expression is a single expression wrapped in
                 * parentheses (ex: `(a) + b`), leftToken will be the last token of the expression
                 * and operatorToken will be the closing parenthesis.
                 * The leftToken should be the last closing parenthesis, and the operatorToken
                 * should be the token right after that.
                 */
                const operatorToken = sourceCode.getTokenAfter(leftSide, astUtils.isNotClosingParenToken);
                const leftToken = sourceCode.getTokenBefore(operatorToken);
                const rightToken = sourceCode.getTokenAfter(operatorToken);
                const operator = operatorToken.value;
                const operatorStyleOverride = styleOverrides[operator];
                const style = operatorStyleOverride || globalStyle;
                const fix = getFixer(operatorToken, style);
                // if single line
                if (astUtils.isTokenOnSameLine(leftToken, operatorToken) &&
                    astUtils.isTokenOnSameLine(operatorToken, rightToken)) {
                    // do nothing.
                }
                else if (operatorStyleOverride !== "ignore" && !astUtils.isTokenOnSameLine(leftToken, operatorToken) &&
                    !astUtils.isTokenOnSameLine(operatorToken, rightToken)) {
                    // lone operator
                    context.report({
                        node,
                        loc: operatorToken.loc,
                        messageId: "badLinebreak",
                        data: {
                            operator
                        },
                        fix
                    });
                }
                else if (style === "before" && astUtils.isTokenOnSameLine(leftToken, operatorToken)) {
                    context.report({
                        node,
                        loc: operatorToken.loc,
                        messageId: "operatorAtBeginning",
                        data: {
                            operator
                        },
                        fix
                    });
                }
                else if (style === "after" && astUtils.isTokenOnSameLine(operatorToken, rightToken)) {
                    context.report({
                        node,
                        loc: operatorToken.loc,
                        messageId: "operatorAtEnd",
                        data: {
                            operator
                        },
                        fix
                    });
                }
                else if (style === "none") {
                    context.report({
                        node,
                        loc: operatorToken.loc,
                        messageId: "noLinebreak",
                        data: {
                            operator
                        },
                        fix
                    });
                }
            }
            /**
             * Validates a binary expression using `validateNode`
             * @param {BinaryExpression|LogicalExpression|AssignmentExpression} node node to be validated
             * @returns {void}
             */
            function validateBinaryExpression(node) {
                validateNode(node, node.left);
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                BinaryExpression: validateBinaryExpression,
                LogicalExpression: validateBinaryExpression,
                AssignmentExpression: validateBinaryExpression,
                VariableDeclarator(node) {
                    if (node.init) {
                        validateNode(node, node.id);
                    }
                },
                ConditionalExpression(node) {
                    validateNode(node, node.test);
                    validateNode(node, node.consequent);
                }
            };
        }
    };
},
/* eslint/lib/rules/padded-blocks.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow padding within blocks",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/padded-blocks"
            },
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["always", "never"]
                        },
                        {
                            type: "object",
                            properties: {
                                blocks: {
                                    enum: ["always", "never"]
                                },
                                switches: {
                                    enum: ["always", "never"]
                                },
                                classes: {
                                    enum: ["always", "never"]
                                }
                            },
                            additionalProperties: false,
                            minProperties: 1
                        }
                    ]
                },
                {
                    type: "object",
                    properties: {
                        allowSingleLineBlocks: {
                            type: "boolean"
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                alwaysPadBlock: "Block must be padded by blank lines.",
                neverPadBlock: "Block must not be padded by blank lines."
            }
        },
        create(context) {
            const options = {};
            const typeOptions = context.options[0] || "always";
            const exceptOptions = context.options[1] || {};
            if (typeof typeOptions === "string") {
                const shouldHavePadding = typeOptions === "always";
                options.blocks = shouldHavePadding;
                options.switches = shouldHavePadding;
                options.classes = shouldHavePadding;
            }
            else {
                if (Object.prototype.hasOwnProperty.call(typeOptions, "blocks")) {
                    options.blocks = typeOptions.blocks === "always";
                }
                if (Object.prototype.hasOwnProperty.call(typeOptions, "switches")) {
                    options.switches = typeOptions.switches === "always";
                }
                if (Object.prototype.hasOwnProperty.call(typeOptions, "classes")) {
                    options.classes = typeOptions.classes === "always";
                }
            }
            if (Object.prototype.hasOwnProperty.call(exceptOptions, "allowSingleLineBlocks")) {
                options.allowSingleLineBlocks = exceptOptions.allowSingleLineBlocks === true;
            }
            const sourceCode = context.getSourceCode();
            /**
             * Gets the open brace token from a given node.
             * @param {ASTNode} node A BlockStatement or SwitchStatement node from which to get the open brace.
             * @returns {Token} The token of the open brace.
             */
            function getOpenBrace(node) {
                if (node.type === "SwitchStatement") {
                    return sourceCode.getTokenBefore(node.cases[0]);
                }
                return sourceCode.getFirstToken(node);
            }
            /**
             * Checks if the given parameter is a comment node
             * @param {ASTNode|Token} node An AST node or token
             * @returns {boolean} True if node is a comment
             */
            function isComment(node) {
                return node.type === "Line" || node.type === "Block";
            }
            /**
             * Checks if there is padding between two tokens
             * @param {Token} first The first token
             * @param {Token} second The second token
             * @returns {boolean} True if there is at least a line between the tokens
             */
            function isPaddingBetweenTokens(first, second) {
                return second.loc.start.line - first.loc.end.line >= 2;
            }
            /**
             * Checks if the given token has a blank line after it.
             * @param {Token} token The token to check.
             * @returns {boolean} Whether or not the token is followed by a blank line.
             */
            function getFirstBlockToken(token) {
                let prev, first = token;
                do {
                    prev = first;
                    first = sourceCode.getTokenAfter(first, { includeComments: true });
                } while (isComment(first) && first.loc.start.line === prev.loc.end.line);
                return first;
            }
            /**
             * Checks if the given token is preceded by a blank line.
             * @param {Token} token The token to check
             * @returns {boolean} Whether or not the token is preceded by a blank line
             */
            function getLastBlockToken(token) {
                let last = token, next;
                do {
                    next = last;
                    last = sourceCode.getTokenBefore(last, { includeComments: true });
                } while (isComment(last) && last.loc.end.line === next.loc.start.line);
                return last;
            }
            /**
             * Checks if a node should be padded, according to the rule config.
             * @param {ASTNode} node The AST node to check.
             * @returns {boolean} True if the node should be padded, false otherwise.
             */
            function requirePaddingFor(node) {
                switch (node.type) {
                    case "BlockStatement":
                        return options.blocks;
                    case "SwitchStatement":
                        return options.switches;
                    case "ClassBody":
                        return options.classes;
                    /* istanbul ignore next */
                    default:
                        throw new Error("unreachable");
                }
            }
            /**
             * Checks the given BlockStatement node to be padded if the block is not empty.
             * @param {ASTNode} node The AST node of a BlockStatement.
             * @returns {void} undefined.
             */
            function checkPadding(node) {
                const openBrace = getOpenBrace(node), firstBlockToken = getFirstBlockToken(openBrace), tokenBeforeFirst = sourceCode.getTokenBefore(firstBlockToken, { includeComments: true }), closeBrace = sourceCode.getLastToken(node), lastBlockToken = getLastBlockToken(closeBrace), tokenAfterLast = sourceCode.getTokenAfter(lastBlockToken, { includeComments: true }), blockHasTopPadding = isPaddingBetweenTokens(tokenBeforeFirst, firstBlockToken), blockHasBottomPadding = isPaddingBetweenTokens(lastBlockToken, tokenAfterLast);
                if (options.allowSingleLineBlocks && astUtils.isTokenOnSameLine(tokenBeforeFirst, tokenAfterLast)) {
                    return;
                }
                if (requirePaddingFor(node)) {
                    if (!blockHasTopPadding) {
                        context.report({
                            node,
                            loc: {
                                start: tokenBeforeFirst.loc.start,
                                end: firstBlockToken.loc.start
                            },
                            fix(fixer) {
                                return fixer.insertTextAfter(tokenBeforeFirst, "\n");
                            },
                            messageId: "alwaysPadBlock"
                        });
                    }
                    if (!blockHasBottomPadding) {
                        context.report({
                            node,
                            loc: {
                                end: tokenAfterLast.loc.start,
                                start: lastBlockToken.loc.end
                            },
                            fix(fixer) {
                                return fixer.insertTextBefore(tokenAfterLast, "\n");
                            },
                            messageId: "alwaysPadBlock"
                        });
                    }
                }
                else {
                    if (blockHasTopPadding) {
                        context.report({
                            node,
                            loc: {
                                start: tokenBeforeFirst.loc.start,
                                end: firstBlockToken.loc.start
                            },
                            fix(fixer) {
                                return fixer.replaceTextRange([tokenBeforeFirst.range[1], firstBlockToken.range[0] - firstBlockToken.loc.start.column], "\n");
                            },
                            messageId: "neverPadBlock"
                        });
                    }
                    if (blockHasBottomPadding) {
                        context.report({
                            node,
                            loc: {
                                end: tokenAfterLast.loc.start,
                                start: lastBlockToken.loc.end
                            },
                            messageId: "neverPadBlock",
                            fix(fixer) {
                                return fixer.replaceTextRange([lastBlockToken.range[1], tokenAfterLast.range[0] - tokenAfterLast.loc.start.column], "\n");
                            }
                        });
                    }
                }
            }
            const rule = {};
            if (Object.prototype.hasOwnProperty.call(options, "switches")) {
                rule.SwitchStatement = function (node) {
                    if (node.cases.length === 0) {
                        return;
                    }
                    checkPadding(node);
                };
            }
            if (Object.prototype.hasOwnProperty.call(options, "blocks")) {
                rule.BlockStatement = function (node) {
                    if (node.body.length === 0) {
                        return;
                    }
                    checkPadding(node);
                };
            }
            if (Object.prototype.hasOwnProperty.call(options, "classes")) {
                rule.ClassBody = function (node) {
                    if (node.body.length === 0) {
                        return;
                    }
                    checkPadding(node);
                };
            }
            return rule;
        }
    };
},
/* eslint/lib/rules/padding-line-between-statements.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const LT = `[${Array.from(astUtils.LINEBREAKS).join("")}]`;
    const PADDING_LINE_SEQUENCE = new RegExp(String.raw `^(\s*?${LT})\s*${LT}(\s*;?)$`, "u");
    const CJS_EXPORT = /^(?:module\s*\.\s*)?exports(?:\s*\.|\s*\[|$)/u;
    const CJS_IMPORT = /^require\(/u;
    /**
     * Creates tester which check if a node starts with specific keyword.
     * @param {string} keyword The keyword to test.
     * @returns {Object} the created tester.
     * @private
     */
    function newKeywordTester(keyword) {
        return {
            test: (node, sourceCode) => sourceCode.getFirstToken(node).value === keyword
        };
    }
    /**
     * Creates tester which check if a node starts with specific keyword and spans a single line.
     * @param {string} keyword The keyword to test.
     * @returns {Object} the created tester.
     * @private
     */
    function newSinglelineKeywordTester(keyword) {
        return {
            test: (node, sourceCode) => node.loc.start.line === node.loc.end.line &&
                sourceCode.getFirstToken(node).value === keyword
        };
    }
    /**
     * Creates tester which check if a node starts with specific keyword and spans multiple lines.
     * @param {string} keyword The keyword to test.
     * @returns {Object} the created tester.
     * @private
     */
    function newMultilineKeywordTester(keyword) {
        return {
            test: (node, sourceCode) => node.loc.start.line !== node.loc.end.line &&
                sourceCode.getFirstToken(node).value === keyword
        };
    }
    /**
     * Creates tester which check if a node is specific type.
     * @param {string} type The node type to test.
     * @returns {Object} the created tester.
     * @private
     */
    function newNodeTypeTester(type) {
        return {
            test: node => node.type === type
        };
    }
    /**
     * Checks the given node is an expression statement of IIFE.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is an expression statement of IIFE.
     * @private
     */
    function isIIFEStatement(node) {
        if (node.type === "ExpressionStatement") {
            let call = astUtils.skipChainExpression(node.expression);
            if (call.type === "UnaryExpression") {
                call = astUtils.skipChainExpression(call.argument);
            }
            return call.type === "CallExpression" && astUtils.isFunction(call.callee);
        }
        return false;
    }
    /**
     * Checks whether the given node is a block-like statement.
     * This checks the last token of the node is the closing brace of a block.
     * @param {SourceCode} sourceCode The source code to get tokens.
     * @param {ASTNode} node The node to check.
     * @returns {boolean} `true` if the node is a block-like statement.
     * @private
     */
    function isBlockLikeStatement(sourceCode, node) {
        // do-while with a block is a block-like statement.
        if (node.type === "DoWhileStatement" && node.body.type === "BlockStatement") {
            return true;
        }
        /*
         * IIFE is a block-like statement specially from
         * JSCS#disallowPaddingNewLinesAfterBlocks.
         */
        if (isIIFEStatement(node)) {
            return true;
        }
        // Checks the last token is a closing brace of blocks.
        const lastToken = sourceCode.getLastToken(node, astUtils.isNotSemicolonToken);
        const belongingNode = lastToken && astUtils.isClosingBraceToken(lastToken)
            ? sourceCode.getNodeByRangeIndex(lastToken.range[0])
            : null;
        return Boolean(belongingNode) && (belongingNode.type === "BlockStatement" ||
            belongingNode.type === "SwitchStatement");
    }
    /**
     * Check whether the given node is a directive or not.
     * @param {ASTNode} node The node to check.
     * @param {SourceCode} sourceCode The source code object to get tokens.
     * @returns {boolean} `true` if the node is a directive.
     */
    function isDirective(node, sourceCode) {
        return (node.type === "ExpressionStatement" &&
            (node.parent.type === "Program" ||
                (node.parent.type === "BlockStatement" &&
                    astUtils.isFunction(node.parent.parent))) &&
            node.expression.type === "Literal" &&
            typeof node.expression.value === "string" &&
            !astUtils.isParenthesised(sourceCode, node.expression));
    }
    /**
     * Check whether the given node is a part of directive prologue or not.
     * @param {ASTNode} node The node to check.
     * @param {SourceCode} sourceCode The source code object to get tokens.
     * @returns {boolean} `true` if the node is a part of directive prologue.
     */
    function isDirectivePrologue(node, sourceCode) {
        if (isDirective(node, sourceCode)) {
            for (const sibling of node.parent.body) {
                if (sibling === node) {
                    break;
                }
                if (!isDirective(sibling, sourceCode)) {
                    return false;
                }
            }
            return true;
        }
        return false;
    }
    /**
     * Gets the actual last token.
     *
     * If a semicolon is semicolon-less style's semicolon, this ignores it.
     * For example:
     *
     *     foo()
     *     ;[1, 2, 3].forEach(bar)
     * @param {SourceCode} sourceCode The source code to get tokens.
     * @param {ASTNode} node The node to get.
     * @returns {Token} The actual last token.
     * @private
     */
    function getActualLastToken(sourceCode, node) {
        const semiToken = sourceCode.getLastToken(node);
        const prevToken = sourceCode.getTokenBefore(semiToken);
        const nextToken = sourceCode.getTokenAfter(semiToken);
        const isSemicolonLessStyle = Boolean(prevToken &&
            nextToken &&
            prevToken.range[0] >= node.range[0] &&
            astUtils.isSemicolonToken(semiToken) &&
            semiToken.loc.start.line !== prevToken.loc.end.line &&
            semiToken.loc.end.line === nextToken.loc.start.line);
        return isSemicolonLessStyle ? prevToken : semiToken;
    }
    /**
     * This returns the concatenation of the first 2 captured strings.
     * @param {string} _ Unused. Whole matched string.
     * @param {string} trailingSpaces The trailing spaces of the first line.
     * @param {string} indentSpaces The indentation spaces of the last line.
     * @returns {string} The concatenation of trailingSpaces and indentSpaces.
     * @private
     */
    function replacerToRemovePaddingLines(_, trailingSpaces, indentSpaces) {
        return trailingSpaces + indentSpaces;
    }
    /**
     * Check and report statements for `any` configuration.
     * It does nothing.
     * @returns {void}
     * @private
     */
    function verifyForAny() {
    }
    /**
     * Check and report statements for `never` configuration.
     * This autofix removes blank lines between the given 2 statements.
     * However, if comments exist between 2 blank lines, it does not remove those
     * blank lines automatically.
     * @param {RuleContext} context The rule context to report.
     * @param {ASTNode} _ Unused. The previous node to check.
     * @param {ASTNode} nextNode The next node to check.
     * @param {Array<Token[]>} paddingLines The array of token pairs that blank
     * lines exist between the pair.
     * @returns {void}
     * @private
     */
    function verifyForNever(context, _, nextNode, paddingLines) {
        if (paddingLines.length === 0) {
            return;
        }
        context.report({
            node: nextNode,
            messageId: "unexpectedBlankLine",
            fix(fixer) {
                if (paddingLines.length >= 2) {
                    return null;
                }
                const prevToken = paddingLines[0][0];
                const nextToken = paddingLines[0][1];
                const start = prevToken.range[1];
                const end = nextToken.range[0];
                const text = context.getSourceCode().text
                    .slice(start, end)
                    .replace(PADDING_LINE_SEQUENCE, replacerToRemovePaddingLines);
                return fixer.replaceTextRange([start, end], text);
            }
        });
    }
    /**
     * Check and report statements for `always` configuration.
     * This autofix inserts a blank line between the given 2 statements.
     * If the `prevNode` has trailing comments, it inserts a blank line after the
     * trailing comments.
     * @param {RuleContext} context The rule context to report.
     * @param {ASTNode} prevNode The previous node to check.
     * @param {ASTNode} nextNode The next node to check.
     * @param {Array<Token[]>} paddingLines The array of token pairs that blank
     * lines exist between the pair.
     * @returns {void}
     * @private
     */
    function verifyForAlways(context, prevNode, nextNode, paddingLines) {
        if (paddingLines.length > 0) {
            return;
        }
        context.report({
            node: nextNode,
            messageId: "expectedBlankLine",
            fix(fixer) {
                const sourceCode = context.getSourceCode();
                let prevToken = getActualLastToken(sourceCode, prevNode);
                const nextToken = sourceCode.getFirstTokenBetween(prevToken, nextNode, {
                    includeComments: true,
                    /**
                     * Skip the trailing comments of the previous node.
                     * This inserts a blank line after the last trailing comment.
                     *
                     * For example:
                     *
                     *     foo(); // trailing comment.
                     *     // comment.
                     *     bar();
                     *
                     * Get fixed to:
                     *
                     *     foo(); // trailing comment.
                     *
                     *     // comment.
                     *     bar();
                     * @param {Token} token The token to check.
                     * @returns {boolean} `true` if the token is not a trailing comment.
                     * @private
                     */
                    filter(token) {
                        if (astUtils.isTokenOnSameLine(prevToken, token)) {
                            prevToken = token;
                            return false;
                        }
                        return true;
                    }
                }) || nextNode;
                const insertText = astUtils.isTokenOnSameLine(prevToken, nextToken)
                    ? "\n\n"
                    : "\n";
                return fixer.insertTextAfter(prevToken, insertText);
            }
        });
    }
    /**
     * Types of blank lines.
     * `any`, `never`, and `always` are defined.
     * Those have `verify` method to check and report statements.
     * @private
     */
    const PaddingTypes = {
        any: { verify: verifyForAny },
        never: { verify: verifyForNever },
        always: { verify: verifyForAlways }
    };
    /**
     * Types of statements.
     * Those have `test` method to check it matches to the given statement.
     * @private
     */
    const StatementTypes = {
        "*": { test: () => true },
        "block-like": {
            test: (node, sourceCode) => isBlockLikeStatement(sourceCode, node)
        },
        "cjs-export": {
            test: (node, sourceCode) => node.type === "ExpressionStatement" &&
                node.expression.type === "AssignmentExpression" &&
                CJS_EXPORT.test(sourceCode.getText(node.expression.left))
        },
        "cjs-import": {
            test: (node, sourceCode) => node.type === "VariableDeclaration" &&
                node.declarations.length > 0 &&
                Boolean(node.declarations[0].init) &&
                CJS_IMPORT.test(sourceCode.getText(node.declarations[0].init))
        },
        directive: {
            test: isDirectivePrologue
        },
        expression: {
            test: (node, sourceCode) => node.type === "ExpressionStatement" &&
                !isDirectivePrologue(node, sourceCode)
        },
        iife: {
            test: isIIFEStatement
        },
        "multiline-block-like": {
            test: (node, sourceCode) => node.loc.start.line !== node.loc.end.line &&
                isBlockLikeStatement(sourceCode, node)
        },
        "multiline-expression": {
            test: (node, sourceCode) => node.loc.start.line !== node.loc.end.line &&
                node.type === "ExpressionStatement" &&
                !isDirectivePrologue(node, sourceCode)
        },
        "multiline-const": newMultilineKeywordTester("const"),
        "multiline-let": newMultilineKeywordTester("let"),
        "multiline-var": newMultilineKeywordTester("var"),
        "singleline-const": newSinglelineKeywordTester("const"),
        "singleline-let": newSinglelineKeywordTester("let"),
        "singleline-var": newSinglelineKeywordTester("var"),
        block: newNodeTypeTester("BlockStatement"),
        empty: newNodeTypeTester("EmptyStatement"),
        function: newNodeTypeTester("FunctionDeclaration"),
        break: newKeywordTester("break"),
        case: newKeywordTester("case"),
        class: newKeywordTester("class"),
        const: newKeywordTester("const"),
        continue: newKeywordTester("continue"),
        debugger: newKeywordTester("debugger"),
        default: newKeywordTester("default"),
        do: newKeywordTester("do"),
        export: newKeywordTester("export"),
        for: newKeywordTester("for"),
        if: newKeywordTester("if"),
        import: newKeywordTester("import"),
        let: newKeywordTester("let"),
        return: newKeywordTester("return"),
        switch: newKeywordTester("switch"),
        throw: newKeywordTester("throw"),
        try: newKeywordTester("try"),
        var: newKeywordTester("var"),
        while: newKeywordTester("while"),
        with: newKeywordTester("with")
    };
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow padding lines between statements",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/padding-line-between-statements"
            },
            fixable: "whitespace",
            schema: {
                definitions: {
                    paddingType: {
                        enum: Object.keys(PaddingTypes)
                    },
                    statementType: {
                        anyOf: [
                            { enum: Object.keys(StatementTypes) },
                            {
                                type: "array",
                                items: { enum: Object.keys(StatementTypes) },
                                minItems: 1,
                                uniqueItems: true,
                                additionalItems: false
                            }
                        ]
                    }
                },
                type: "array",
                items: {
                    type: "object",
                    properties: {
                        blankLine: { $ref: "#/definitions/paddingType" },
                        prev: { $ref: "#/definitions/statementType" },
                        next: { $ref: "#/definitions/statementType" }
                    },
                    additionalProperties: false,
                    required: ["blankLine", "prev", "next"]
                },
                additionalItems: false
            },
            messages: {
                unexpectedBlankLine: "Unexpected blank line before this statement.",
                expectedBlankLine: "Expected blank line before this statement."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const configureList = context.options || [];
            let scopeInfo = null;
            /**
             * Processes to enter to new scope.
             * This manages the current previous statement.
             * @returns {void}
             * @private
             */
            function enterScope() {
                scopeInfo = {
                    upper: scopeInfo,
                    prevNode: null
                };
            }
            /**
             * Processes to exit from the current scope.
             * @returns {void}
             * @private
             */
            function exitScope() {
                scopeInfo = scopeInfo.upper;
            }
            /**
             * Checks whether the given node matches the given type.
             * @param {ASTNode} node The statement node to check.
             * @param {string|string[]} type The statement type to check.
             * @returns {boolean} `true` if the statement node matched the type.
             * @private
             */
            function match(node, type) {
                let innerStatementNode = node;
                while (innerStatementNode.type === "LabeledStatement") {
                    innerStatementNode = innerStatementNode.body;
                }
                if (Array.isArray(type)) {
                    return type.some(match.bind(null, innerStatementNode));
                }
                return StatementTypes[type].test(innerStatementNode, sourceCode);
            }
            /**
             * Finds the last matched configure from configureList.
             * @param {ASTNode} prevNode The previous statement to match.
             * @param {ASTNode} nextNode The current statement to match.
             * @returns {Object} The tester of the last matched configure.
             * @private
             */
            function getPaddingType(prevNode, nextNode) {
                for (let i = configureList.length - 1; i >= 0; --i) {
                    const configure = configureList[i];
                    const matched = match(prevNode, configure.prev) &&
                        match(nextNode, configure.next);
                    if (matched) {
                        return PaddingTypes[configure.blankLine];
                    }
                }
                return PaddingTypes.any;
            }
            /**
             * Gets padding line sequences between the given 2 statements.
             * Comments are separators of the padding line sequences.
             * @param {ASTNode} prevNode The previous statement to count.
             * @param {ASTNode} nextNode The current statement to count.
             * @returns {Array<Token[]>} The array of token pairs.
             * @private
             */
            function getPaddingLineSequences(prevNode, nextNode) {
                const pairs = [];
                let prevToken = getActualLastToken(sourceCode, prevNode);
                if (nextNode.loc.start.line - prevToken.loc.end.line >= 2) {
                    do {
                        const token = sourceCode.getTokenAfter(prevToken, { includeComments: true });
                        if (token.loc.start.line - prevToken.loc.end.line >= 2) {
                            pairs.push([prevToken, token]);
                        }
                        prevToken = token;
                    } while (prevToken.range[0] < nextNode.range[0]);
                }
                return pairs;
            }
            /**
             * Verify padding lines between the given node and the previous node.
             * @param {ASTNode} node The node to verify.
             * @returns {void}
             * @private
             */
            function verify(node) {
                const parentType = node.parent.type;
                const validParent = astUtils.STATEMENT_LIST_PARENTS.has(parentType) ||
                    parentType === "SwitchStatement";
                if (!validParent) {
                    return;
                }
                // Save this node as the current previous statement.
                const prevNode = scopeInfo.prevNode;
                // Verify.
                if (prevNode) {
                    const type = getPaddingType(prevNode, node);
                    const paddingLines = getPaddingLineSequences(prevNode, node);
                    type.verify(context, prevNode, node, paddingLines);
                }
                scopeInfo.prevNode = node;
            }
            /**
             * Verify padding lines between the given node and the previous node.
             * Then process to enter to new scope.
             * @param {ASTNode} node The node to verify.
             * @returns {void}
             * @private
             */
            function verifyThenEnterScope(node) {
                verify(node);
                enterScope();
            }
            return {
                Program: enterScope,
                BlockStatement: enterScope,
                SwitchStatement: enterScope,
                "Program:exit": exitScope,
                "BlockStatement:exit": exitScope,
                "SwitchStatement:exit": exitScope,
                ":statement": verify,
                SwitchCase: verifyThenEnterScope,
                "SwitchCase:exit": exitScope
            };
        }
    };
},
/* eslint/lib/rules/prefer-arrow-callback.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given variable is a function name.
     * @param {eslint-scope.Variable} variable A variable to check.
     * @returns {boolean} `true` if the variable is a function name.
     */
    function isFunctionName(variable) {
        return variable && variable.defs[0].type === "FunctionName";
    }
    /**
     * Checks whether or not a given MetaProperty node equals to a given value.
     * @param {ASTNode} node A MetaProperty node to check.
     * @param {string} metaName The name of `MetaProperty.meta`.
     * @param {string} propertyName The name of `MetaProperty.property`.
     * @returns {boolean} `true` if the node is the specific value.
     */
    function checkMetaProperty(node, metaName, propertyName) {
        return node.meta.name === metaName && node.property.name === propertyName;
    }
    /**
     * Gets the variable object of `arguments` which is defined implicitly.
     * @param {eslint-scope.Scope} scope A scope to get.
     * @returns {eslint-scope.Variable} The found variable object.
     */
    function getVariableOfArguments(scope) {
        const variables = scope.variables;
        for (let i = 0; i < variables.length; ++i) {
            const variable = variables[i];
            if (variable.name === "arguments") {
                /*
                 * If there was a parameter which is named "arguments", the
                 * implicit "arguments" is not defined.
                 * So does fast return with null.
                 */
                return (variable.identifiers.length === 0) ? variable : null;
            }
        }
        /* istanbul ignore next */
        return null;
    }
    /**
     * Checks whether or not a given node is a callback.
     * @param {ASTNode} node A node to check.
     * @returns {Object}
     *   {boolean} retv.isCallback - `true` if the node is a callback.
     *   {boolean} retv.isLexicalThis - `true` if the node is with `.bind(this)`.
     */
    function getCallbackInfo(node) {
        const retv = { isCallback: false, isLexicalThis: false };
        let currentNode = node;
        let parent = node.parent;
        let bound = false;
        while (currentNode) {
            switch (parent.type) {
                // Checks parents recursively.
                case "LogicalExpression":
                case "ChainExpression":
                case "ConditionalExpression":
                    break;
                // Checks whether the parent node is `.bind(this)` call.
                case "MemberExpression":
                    if (parent.object === currentNode &&
                        !parent.property.computed &&
                        parent.property.type === "Identifier" &&
                        parent.property.name === "bind") {
                        const maybeCallee = parent.parent.type === "ChainExpression"
                            ? parent.parent
                            : parent;
                        if (astUtils.isCallee(maybeCallee)) {
                            if (!bound) {
                                bound = true; // Use only the first `.bind()` to make `isLexicalThis` value.
                                retv.isLexicalThis = (maybeCallee.parent.arguments.length === 1 &&
                                    maybeCallee.parent.arguments[0].type === "ThisExpression");
                            }
                            parent = maybeCallee.parent;
                        }
                        else {
                            return retv;
                        }
                    }
                    else {
                        return retv;
                    }
                    break;
                // Checks whether the node is a callback.
                case "CallExpression":
                case "NewExpression":
                    if (parent.callee !== currentNode) {
                        retv.isCallback = true;
                    }
                    return retv;
                default:
                    return retv;
            }
            currentNode = parent;
            parent = parent.parent;
        }
        /* istanbul ignore next */
        throw new Error("unreachable");
    }
    /**
     * Checks whether a simple list of parameters contains any duplicates. This does not handle complex
     * parameter lists (e.g. with destructuring), since complex parameter lists are a SyntaxError with duplicate
     * parameter names anyway. Instead, it always returns `false` for complex parameter lists.
     * @param {ASTNode[]} paramsList The list of parameters for a function
     * @returns {boolean} `true` if the list of parameters contains any duplicates
     */
    function hasDuplicateParams(paramsList) {
        return paramsList.every(param => param.type === "Identifier") && paramsList.length !== new Set(paramsList.map(param => param.name)).size;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require using arrow functions for callbacks",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-arrow-callback"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        allowNamedFunctions: {
                            type: "boolean",
                            default: false
                        },
                        allowUnboundThis: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "code",
            messages: {
                preferArrowCallback: "Unexpected function expression."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const allowUnboundThis = options.allowUnboundThis !== false; // default to true
            const allowNamedFunctions = options.allowNamedFunctions;
            const sourceCode = context.getSourceCode();
            /*
             * {Array<{this: boolean, super: boolean, meta: boolean}>}
             * - this - A flag which shows there are one or more ThisExpression.
             * - super - A flag which shows there are one or more Super.
             * - meta - A flag which shows there are one or more MethProperty.
             */
            let stack = [];
            /**
             * Pushes new function scope with all `false` flags.
             * @returns {void}
             */
            function enterScope() {
                stack.push({ this: false, super: false, meta: false });
            }
            /**
             * Pops a function scope from the stack.
             * @returns {{this: boolean, super: boolean, meta: boolean}} The information of the last scope.
             */
            function exitScope() {
                return stack.pop();
            }
            return {
                // Reset internal state.
                Program() {
                    stack = [];
                },
                // If there are below, it cannot replace with arrow functions merely.
                ThisExpression() {
                    const info = stack[stack.length - 1];
                    if (info) {
                        info.this = true;
                    }
                },
                Super() {
                    const info = stack[stack.length - 1];
                    if (info) {
                        info.super = true;
                    }
                },
                MetaProperty(node) {
                    const info = stack[stack.length - 1];
                    if (info && checkMetaProperty(node, "new", "target")) {
                        info.meta = true;
                    }
                },
                // To skip nested scopes.
                FunctionDeclaration: enterScope,
                "FunctionDeclaration:exit": exitScope,
                // Main.
                FunctionExpression: enterScope,
                "FunctionExpression:exit"(node) {
                    const scopeInfo = exitScope();
                    // Skip named function expressions
                    if (allowNamedFunctions && node.id && node.id.name) {
                        return;
                    }
                    // Skip generators.
                    if (node.generator) {
                        return;
                    }
                    // Skip recursive functions.
                    const nameVar = context.getDeclaredVariables(node)[0];
                    if (isFunctionName(nameVar) && nameVar.references.length > 0) {
                        return;
                    }
                    // Skip if it's using arguments.
                    const variable = getVariableOfArguments(context.getScope());
                    if (variable && variable.references.length > 0) {
                        return;
                    }
                    // Reports if it's a callback which can replace with arrows.
                    const callbackInfo = getCallbackInfo(node);
                    if (callbackInfo.isCallback &&
                        (!allowUnboundThis || !scopeInfo.this || callbackInfo.isLexicalThis) &&
                        !scopeInfo.super &&
                        !scopeInfo.meta) {
                        context.report({
                            node,
                            messageId: "preferArrowCallback",
                            *fix(fixer) {
                                if ((!callbackInfo.isLexicalThis && scopeInfo.this) || hasDuplicateParams(node.params)) {
                                    /*
                                     * If the callback function does not have .bind(this) and contains a reference to `this`, there
                                     * is no way to determine what `this` should be, so don't perform any fixes.
                                     * If the callback function has duplicates in its list of parameters (possible in sloppy mode),
                                     * don't replace it with an arrow function, because this is a SyntaxError with arrow functions.
                                     */
                                    return;
                                }
                                // Remove `.bind(this)` if exists.
                                if (callbackInfo.isLexicalThis) {
                                    const memberNode = node.parent;
                                    /*
                                     * If `.bind(this)` exists but the parent is not `.bind(this)`, don't remove it automatically.
                                     * E.g. `(foo || function(){}).bind(this)`
                                     */
                                    if (memberNode.type !== "MemberExpression") {
                                        return;
                                    }
                                    const callNode = memberNode.parent;
                                    const firstTokenToRemove = sourceCode.getTokenAfter(memberNode.object, astUtils.isNotClosingParenToken);
                                    const lastTokenToRemove = sourceCode.getLastToken(callNode);
                                    /*
                                     * If the member expression is parenthesized, don't remove the right paren.
                                     * E.g. `(function(){}.bind)(this)`
                                     *                    ^^^^^^^^^^^^
                                     */
                                    if (astUtils.isParenthesised(sourceCode, memberNode)) {
                                        return;
                                    }
                                    // If comments exist in the `.bind(this)`, don't remove those.
                                    if (sourceCode.commentsExistBetween(firstTokenToRemove, lastTokenToRemove)) {
                                        return;
                                    }
                                    yield fixer.removeRange([firstTokenToRemove.range[0], lastTokenToRemove.range[1]]);
                                }
                                // Convert the function expression to an arrow function.
                                const functionToken = sourceCode.getFirstToken(node, node.async ? 1 : 0);
                                const leftParenToken = sourceCode.getTokenAfter(functionToken, astUtils.isOpeningParenToken);
                                if (sourceCode.commentsExistBetween(functionToken, leftParenToken)) {
                                    // Remove only extra tokens to keep comments.
                                    yield fixer.remove(functionToken);
                                    if (node.id) {
                                        yield fixer.remove(node.id);
                                    }
                                }
                                else {
                                    // Remove extra tokens and spaces.
                                    yield fixer.removeRange([functionToken.range[0], leftParenToken.range[0]]);
                                }
                                yield fixer.insertTextBefore(node.body, "=> ");
                                // Get the node that will become the new arrow function.
                                let replacedNode = callbackInfo.isLexicalThis ? node.parent.parent : node;
                                if (replacedNode.type === "ChainExpression") {
                                    replacedNode = replacedNode.parent;
                                }
                                /*
                                 * If the replaced node is part of a BinaryExpression, LogicalExpression, or MemberExpression, then
                                 * the arrow function needs to be parenthesized, because `foo || () => {}` is invalid syntax even
                                 * though `foo || function() {}` is valid.
                                 */
                                if (replacedNode.parent.type !== "CallExpression" &&
                                    replacedNode.parent.type !== "ConditionalExpression" &&
                                    !astUtils.isParenthesised(sourceCode, replacedNode) &&
                                    !astUtils.isParenthesised(sourceCode, node)) {
                                    yield fixer.insertTextBefore(replacedNode, "(");
                                    yield fixer.insertTextAfter(replacedNode, ")");
                                }
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-const.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const FixTracker = require(508) /* ./utils/fix-tracker */;
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const PATTERN_TYPE = /^(?:.+?Pattern|RestElement|SpreadProperty|ExperimentalRestProperty|Property)$/u;
    const DECLARATION_HOST_TYPE = /^(?:Program|BlockStatement|SwitchCase)$/u;
    const DESTRUCTURING_HOST_TYPE = /^(?:VariableDeclarator|AssignmentExpression)$/u;
    /**
     * Checks whether a given node is located at `ForStatement.init` or not.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node is located at `ForStatement.init`.
     */
    function isInitOfForStatement(node) {
        return node.parent.type === "ForStatement" && node.parent.init === node;
    }
    /**
     * Checks whether a given Identifier node becomes a VariableDeclaration or not.
     * @param {ASTNode} identifier An Identifier node to check.
     * @returns {boolean} `true` if the node can become a VariableDeclaration.
     */
    function canBecomeVariableDeclaration(identifier) {
        let node = identifier.parent;
        while (PATTERN_TYPE.test(node.type)) {
            node = node.parent;
        }
        return (node.type === "VariableDeclarator" ||
            (node.type === "AssignmentExpression" &&
                node.parent.type === "ExpressionStatement" &&
                DECLARATION_HOST_TYPE.test(node.parent.parent.type)));
    }
    /**
     * Checks if an property or element is from outer scope or function parameters
     * in destructing pattern.
     * @param {string} name A variable name to be checked.
     * @param {eslint-scope.Scope} initScope A scope to start find.
     * @returns {boolean} Indicates if the variable is from outer scope or function parameters.
     */
    function isOuterVariableInDestructing(name, initScope) {
        if (initScope.through.find(ref => ref.resolved && ref.resolved.name === name)) {
            return true;
        }
        const variable = astUtils.getVariableByName(initScope, name);
        if (variable !== null) {
            return variable.defs.some(def => def.type === "Parameter");
        }
        return false;
    }
    /**
     * Gets the VariableDeclarator/AssignmentExpression node that a given reference
     * belongs to.
     * This is used to detect a mix of reassigned and never reassigned in a
     * destructuring.
     * @param {eslint-scope.Reference} reference A reference to get.
     * @returns {ASTNode|null} A VariableDeclarator/AssignmentExpression node or
     *      null.
     */
    function getDestructuringHost(reference) {
        if (!reference.isWrite()) {
            return null;
        }
        let node = reference.identifier.parent;
        while (PATTERN_TYPE.test(node.type)) {
            node = node.parent;
        }
        if (!DESTRUCTURING_HOST_TYPE.test(node.type)) {
            return null;
        }
        return node;
    }
    /**
     * Determines if a destructuring assignment node contains
     * any MemberExpression nodes. This is used to determine if a
     * variable that is only written once using destructuring can be
     * safely converted into a const declaration.
     * @param {ASTNode} node The ObjectPattern or ArrayPattern node to check.
     * @returns {boolean} True if the destructuring pattern contains
     *      a MemberExpression, false if not.
     */
    function hasMemberExpressionAssignment(node) {
        switch (node.type) {
            case "ObjectPattern":
                return node.properties.some(prop => {
                    if (prop) {
                        /*
                         * Spread elements have an argument property while
                         * others have a value property. Because different
                         * parsers use different node types for spread elements,
                         * we just check if there is an argument property.
                         */
                        return hasMemberExpressionAssignment(prop.argument || prop.value);
                    }
                    return false;
                });
            case "ArrayPattern":
                return node.elements.some(element => {
                    if (element) {
                        return hasMemberExpressionAssignment(element);
                    }
                    return false;
                });
            case "AssignmentPattern":
                return hasMemberExpressionAssignment(node.left);
            case "MemberExpression":
                return true;
            // no default
        }
        return false;
    }
    /**
     * Gets an identifier node of a given variable.
     *
     * If the initialization exists or one or more reading references exist before
     * the first assignment, the identifier node is the node of the declaration.
     * Otherwise, the identifier node is the node of the first assignment.
     *
     * If the variable should not change to const, this function returns null.
     * - If the variable is reassigned.
     * - If the variable is never initialized nor assigned.
     * - If the variable is initialized in a different scope from the declaration.
     * - If the unique assignment of the variable cannot change to a declaration.
     *   e.g. `if (a) b = 1` / `return (b = 1)`
     * - If the variable is declared in the global scope and `eslintUsed` is `true`.
     *   `/*exported foo` directive comment makes such variables. This rule does not
     *   warn such variables because this rule cannot distinguish whether the
     *   exported variables are reassigned or not.
     * @param {eslint-scope.Variable} variable A variable to get.
     * @param {boolean} ignoreReadBeforeAssign
     *      The value of `ignoreReadBeforeAssign` option.
     * @returns {ASTNode|null}
     *      An Identifier node if the variable should change to const.
     *      Otherwise, null.
     */
    function getIdentifierIfShouldBeConst(variable, ignoreReadBeforeAssign) {
        if (variable.eslintUsed && variable.scope.type === "global") {
            return null;
        }
        // Finds the unique WriteReference.
        let writer = null;
        let isReadBeforeInit = false;
        const references = variable.references;
        for (let i = 0; i < references.length; ++i) {
            const reference = references[i];
            if (reference.isWrite()) {
                const isReassigned = (writer !== null &&
                    writer.identifier !== reference.identifier);
                if (isReassigned) {
                    return null;
                }
                const destructuringHost = getDestructuringHost(reference);
                if (destructuringHost !== null && destructuringHost.left !== void 0) {
                    const leftNode = destructuringHost.left;
                    let hasOuterVariables = false, hasNonIdentifiers = false;
                    if (leftNode.type === "ObjectPattern") {
                        const properties = leftNode.properties;
                        hasOuterVariables = properties
                            .filter(prop => prop.value)
                            .map(prop => prop.value.name)
                            .some(name => isOuterVariableInDestructing(name, variable.scope));
                        hasNonIdentifiers = hasMemberExpressionAssignment(leftNode);
                    }
                    else if (leftNode.type === "ArrayPattern") {
                        const elements = leftNode.elements;
                        hasOuterVariables = elements
                            .map(element => element && element.name)
                            .some(name => isOuterVariableInDestructing(name, variable.scope));
                        hasNonIdentifiers = hasMemberExpressionAssignment(leftNode);
                    }
                    if (hasOuterVariables || hasNonIdentifiers) {
                        return null;
                    }
                }
                writer = reference;
            }
            else if (reference.isRead() && writer === null) {
                if (ignoreReadBeforeAssign) {
                    return null;
                }
                isReadBeforeInit = true;
            }
        }
        /*
         * If the assignment is from a different scope, ignore it.
         * If the assignment cannot change to a declaration, ignore it.
         */
        const shouldBeConst = (writer !== null &&
            writer.from === variable.scope &&
            canBecomeVariableDeclaration(writer.identifier));
        if (!shouldBeConst) {
            return null;
        }
        if (isReadBeforeInit) {
            return variable.defs[0].name;
        }
        return writer.identifier;
    }
    /**
     * Groups by the VariableDeclarator/AssignmentExpression node that each
     * reference of given variables belongs to.
     * This is used to detect a mix of reassigned and never reassigned in a
     * destructuring.
     * @param {eslint-scope.Variable[]} variables Variables to group by destructuring.
     * @param {boolean} ignoreReadBeforeAssign
     *      The value of `ignoreReadBeforeAssign` option.
     * @returns {Map<ASTNode, ASTNode[]>} Grouped identifier nodes.
     */
    function groupByDestructuring(variables, ignoreReadBeforeAssign) {
        const identifierMap = new Map();
        for (let i = 0; i < variables.length; ++i) {
            const variable = variables[i];
            const references = variable.references;
            const identifier = getIdentifierIfShouldBeConst(variable, ignoreReadBeforeAssign);
            let prevId = null;
            for (let j = 0; j < references.length; ++j) {
                const reference = references[j];
                const id = reference.identifier;
                /*
                 * Avoid counting a reference twice or more for default values of
                 * destructuring.
                 */
                if (id === prevId) {
                    continue;
                }
                prevId = id;
                // Add the identifier node into the destructuring group.
                const group = getDestructuringHost(reference);
                if (group) {
                    if (identifierMap.has(group)) {
                        identifierMap.get(group).push(identifier);
                    }
                    else {
                        identifierMap.set(group, [identifier]);
                    }
                }
            }
        }
        return identifierMap;
    }
    /**
     * Finds the nearest parent of node with a given type.
     * @param {ASTNode} node The node to search from.
     * @param {string} type The type field of the parent node.
     * @param {Function} shouldStop A predicate that returns true if the traversal should stop, and false otherwise.
     * @returns {ASTNode} The closest ancestor with the specified type; null if no such ancestor exists.
     */
    function findUp(node, type, shouldStop) {
        if (!node || shouldStop(node)) {
            return null;
        }
        if (node.type === type) {
            return node;
        }
        return findUp(node.parent, type, shouldStop);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require `const` declarations for variables that are never reassigned after declared",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-const"
            },
            fixable: "code",
            schema: [
                {
                    type: "object",
                    properties: {
                        destructuring: { enum: ["any", "all"], default: "any" },
                        ignoreReadBeforeAssign: { type: "boolean", default: false }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                useConst: "'{{name}}' is never reassigned. Use 'const' instead."
            }
        },
        create(context) {
            const options = context.options[0] || {};
            const sourceCode = context.getSourceCode();
            const shouldMatchAnyDestructuredVariable = options.destructuring !== "all";
            const ignoreReadBeforeAssign = options.ignoreReadBeforeAssign === true;
            const variables = [];
            let reportCount = 0;
            let checkedId = null;
            let checkedName = "";
            /**
             * Reports given identifier nodes if all of the nodes should be declared
             * as const.
             *
             * The argument 'nodes' is an array of Identifier nodes.
             * This node is the result of 'getIdentifierIfShouldBeConst()', so it's
             * nullable. In simple declaration or assignment cases, the length of
             * the array is 1. In destructuring cases, the length of the array can
             * be 2 or more.
             * @param {(eslint-scope.Reference|null)[]} nodes
             *      References which are grouped by destructuring to report.
             * @returns {void}
             */
            function checkGroup(nodes) {
                const nodesToReport = nodes.filter(Boolean);
                if (nodes.length && (shouldMatchAnyDestructuredVariable || nodesToReport.length === nodes.length)) {
                    const varDeclParent = findUp(nodes[0], "VariableDeclaration", parentNode => parentNode.type.endsWith("Statement"));
                    const isVarDecParentNull = varDeclParent === null;
                    if (!isVarDecParentNull && varDeclParent.declarations.length > 0) {
                        const firstDeclaration = varDeclParent.declarations[0];
                        if (firstDeclaration.init) {
                            const firstDecParent = firstDeclaration.init.parent;
                            /*
                             * First we check the declaration type and then depending on
                             * if the type is a "VariableDeclarator" or its an "ObjectPattern"
                             * we compare the name and id from the first identifier, if the names are different
                             * we assign the new name, id and reset the count of reportCount and nodeCount in
                             * order to check each block for the number of reported errors and base our fix
                             * based on comparing nodes.length and nodesToReport.length.
                             */
                            if (firstDecParent.type === "VariableDeclarator") {
                                if (firstDecParent.id.name !== checkedName) {
                                    checkedName = firstDecParent.id.name;
                                    reportCount = 0;
                                }
                                if (firstDecParent.id.type === "ObjectPattern") {
                                    if (firstDecParent.init.name !== checkedName) {
                                        checkedName = firstDecParent.init.name;
                                        reportCount = 0;
                                    }
                                }
                                if (firstDecParent.id !== checkedId) {
                                    checkedId = firstDecParent.id;
                                    reportCount = 0;
                                }
                            }
                        }
                    }
                    let shouldFix = varDeclParent &&
                        // Don't do a fix unless all variables in the declarations are initialized (or it's in a for-in or for-of loop)
                        (varDeclParent.parent.type === "ForInStatement" || varDeclParent.parent.type === "ForOfStatement" ||
                            varDeclParent.declarations.every(declaration => declaration.init)) &&
                        /*
                         * If options.destructuring is "all", then this warning will not occur unless
                         * every assignment in the destructuring should be const. In that case, it's safe
                         * to apply the fix.
                         */
                        nodesToReport.length === nodes.length;
                    if (!isVarDecParentNull && varDeclParent.declarations && varDeclParent.declarations.length !== 1) {
                        if (varDeclParent && varDeclParent.declarations && varDeclParent.declarations.length >= 1) {
                            /*
                             * Add nodesToReport.length to a count, then comparing the count to the length
                             * of the declarations in the current block.
                             */
                            reportCount += nodesToReport.length;
                            shouldFix = shouldFix && (reportCount === varDeclParent.declarations.length);
                        }
                    }
                    nodesToReport.forEach(node => {
                        context.report({
                            node,
                            messageId: "useConst",
                            data: node,
                            fix: shouldFix
                                ? fixer => {
                                    const letKeywordToken = sourceCode.getFirstToken(varDeclParent, t => t.value === varDeclParent.kind);
                                    /**
                                     * Extend the replacement range to the whole declaration,
                                     * in order to prevent other fixes in the same pass
                                     * https://github.com/eslint/eslint/issues/13899
                                     */
                                    return new FixTracker(fixer, sourceCode)
                                        .retainRange(varDeclParent.range)
                                        .replaceTextRange(letKeywordToken.range, "const");
                                }
                                : null
                        });
                    });
                }
            }
            return {
                "Program:exit"() {
                    groupByDestructuring(variables, ignoreReadBeforeAssign).forEach(checkGroup);
                },
                VariableDeclaration(node) {
                    if (node.kind === "let" && !isInitOfForStatement(node)) {
                        variables.push(...context.getDeclaredVariables(node));
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-destructuring.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const PRECEDENCE_OF_ASSIGNMENT_EXPR = astUtils.getPrecedence({ type: "AssignmentExpression" });
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require destructuring from arrays and/or objects",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-destructuring"
            },
            fixable: "code",
            schema: [
                {
                    /*
                     * old support {array: Boolean, object: Boolean}
                     * new support {VariableDeclarator: {}, AssignmentExpression: {}}
                     */
                    oneOf: [
                        {
                            type: "object",
                            properties: {
                                VariableDeclarator: {
                                    type: "object",
                                    properties: {
                                        array: {
                                            type: "boolean"
                                        },
                                        object: {
                                            type: "boolean"
                                        }
                                    },
                                    additionalProperties: false
                                },
                                AssignmentExpression: {
                                    type: "object",
                                    properties: {
                                        array: {
                                            type: "boolean"
                                        },
                                        object: {
                                            type: "boolean"
                                        }
                                    },
                                    additionalProperties: false
                                }
                            },
                            additionalProperties: false
                        },
                        {
                            type: "object",
                            properties: {
                                array: {
                                    type: "boolean"
                                },
                                object: {
                                    type: "boolean"
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                },
                {
                    type: "object",
                    properties: {
                        enforceForRenamedProperties: {
                            type: "boolean"
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                preferDestructuring: "Use {{type}} destructuring."
            }
        },
        create(context) {
            const enabledTypes = context.options[0];
            const enforceForRenamedProperties = context.options[1] && context.options[1].enforceForRenamedProperties;
            let normalizedOptions = {
                VariableDeclarator: { array: true, object: true },
                AssignmentExpression: { array: true, object: true }
            };
            if (enabledTypes) {
                normalizedOptions = typeof enabledTypes.array !== "undefined" || typeof enabledTypes.object !== "undefined"
                    ? { VariableDeclarator: enabledTypes, AssignmentExpression: enabledTypes }
                    : enabledTypes;
            }
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            // eslint-disable-next-line jsdoc/require-description
            /**
             * @param {string} nodeType "AssignmentExpression" or "VariableDeclarator"
             * @param {string} destructuringType "array" or "object"
             * @returns {boolean} `true` if the destructuring type should be checked for the given node
             */
            function shouldCheck(nodeType, destructuringType) {
                return normalizedOptions &&
                    normalizedOptions[nodeType] &&
                    normalizedOptions[nodeType][destructuringType];
            }
            /**
             * Determines if the given node is accessing an array index
             *
             * This is used to differentiate array index access from object property
             * access.
             * @param {ASTNode} node the node to evaluate
             * @returns {boolean} whether or not the node is an integer
             */
            function isArrayIndexAccess(node) {
                return Number.isInteger(node.property.value);
            }
            /**
             * Report that the given node should use destructuring
             * @param {ASTNode} reportNode the node to report
             * @param {string} type the type of destructuring that should have been done
             * @param {Function|null} fix the fix function or null to pass to context.report
             * @returns {void}
             */
            function report(reportNode, type, fix) {
                context.report({
                    node: reportNode,
                    messageId: "preferDestructuring",
                    data: { type },
                    fix
                });
            }
            /**
             * Determines if a node should be fixed into object destructuring
             *
             * The fixer only fixes the simplest case of object destructuring,
             * like: `let x = a.x`;
             *
             * Assignment expression is not fixed.
             * Array destructuring is not fixed.
             * Renamed property is not fixed.
             * @param {ASTNode} node the node to evaluate
             * @returns {boolean} whether or not the node should be fixed
             */
            function shouldFix(node) {
                return node.type === "VariableDeclarator" &&
                    node.id.type === "Identifier" &&
                    node.init.type === "MemberExpression" &&
                    !node.init.computed &&
                    node.init.property.type === "Identifier" &&
                    node.id.name === node.init.property.name;
            }
            /**
             * Fix a node into object destructuring.
             * This function only handles the simplest case of object destructuring,
             * see {@link shouldFix}.
             * @param {SourceCodeFixer} fixer the fixer object
             * @param {ASTNode} node the node to be fixed.
             * @returns {Object} a fix for the node
             */
            function fixIntoObjectDestructuring(fixer, node) {
                const rightNode = node.init;
                const sourceCode = context.getSourceCode();
                // Don't fix if that would remove any comments. Only comments inside `rightNode.object` can be preserved.
                if (sourceCode.getCommentsInside(node).length > sourceCode.getCommentsInside(rightNode.object).length) {
                    return null;
                }
                let objectText = sourceCode.getText(rightNode.object);
                if (astUtils.getPrecedence(rightNode.object) < PRECEDENCE_OF_ASSIGNMENT_EXPR) {
                    objectText = `(${objectText})`;
                }
                return fixer.replaceText(node, `{${rightNode.property.name}} = ${objectText}`);
            }
            /**
             * Check that the `prefer-destructuring` rules are followed based on the
             * given left- and right-hand side of the assignment.
             *
             * Pulled out into a separate method so that VariableDeclarators and
             * AssignmentExpressions can share the same verification logic.
             * @param {ASTNode} leftNode the left-hand side of the assignment
             * @param {ASTNode} rightNode the right-hand side of the assignment
             * @param {ASTNode} reportNode the node to report the error on
             * @returns {void}
             */
            function performCheck(leftNode, rightNode, reportNode) {
                if (rightNode.type !== "MemberExpression" || rightNode.object.type === "Super") {
                    return;
                }
                if (isArrayIndexAccess(rightNode)) {
                    if (shouldCheck(reportNode.type, "array")) {
                        report(reportNode, "array", null);
                    }
                    return;
                }
                const fix = shouldFix(reportNode)
                    ? fixer => fixIntoObjectDestructuring(fixer, reportNode)
                    : null;
                if (shouldCheck(reportNode.type, "object") && enforceForRenamedProperties) {
                    report(reportNode, "object", fix);
                    return;
                }
                if (shouldCheck(reportNode.type, "object")) {
                    const property = rightNode.property;
                    if ((property.type === "Literal" && leftNode.name === property.value) ||
                        (property.type === "Identifier" && leftNode.name === property.name && !rightNode.computed)) {
                        report(reportNode, "object", fix);
                    }
                }
            }
            /**
             * Check if a given variable declarator is coming from an property access
             * that should be using destructuring instead
             * @param {ASTNode} node the variable declarator to check
             * @returns {void}
             */
            function checkVariableDeclarator(node) {
                // Skip if variable is declared without assignment
                if (!node.init) {
                    return;
                }
                // We only care about member expressions past this point
                if (node.init.type !== "MemberExpression") {
                    return;
                }
                performCheck(node.id, node.init, node);
            }
            /**
             * Run the `prefer-destructuring` check on an AssignmentExpression
             * @param {ASTNode} node the AssignmentExpression node
             * @returns {void}
             */
            function checkAssignmentExpression(node) {
                if (node.operator === "=") {
                    performCheck(node.left, node.right, node);
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                VariableDeclarator: checkVariableDeclarator,
                AssignmentExpression: checkAssignmentExpression
            };
        }
    };
},
/* eslint/lib/rules/prefer-exponentiation-operator.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const { CALL, ReferenceTracker } = require(350) /* eslint-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const PRECEDENCE_OF_EXPONENTIATION_EXPR = astUtils.getPrecedence({ type: "BinaryExpression", operator: "**" });
    /**
     * Determines whether the given node needs parens if used as the base in an exponentiation binary expression.
     * @param {ASTNode} base The node to check.
     * @returns {boolean} `true` if the node needs to be parenthesised.
     */
    function doesBaseNeedParens(base) {
        return (
        // '**' is right-associative, parens are needed when Math.pow(a ** b, c) is converted to (a ** b) ** c
        astUtils.getPrecedence(base) <= PRECEDENCE_OF_EXPONENTIATION_EXPR ||
            // An unary operator cannot be used immediately before an exponentiation expression
            base.type === "AwaitExpression" ||
            base.type === "UnaryExpression");
    }
    /**
     * Determines whether the given node needs parens if used as the exponent in an exponentiation binary expression.
     * @param {ASTNode} exponent The node to check.
     * @returns {boolean} `true` if the node needs to be parenthesised.
     */
    function doesExponentNeedParens(exponent) {
        // '**' is right-associative, there is no need for parens when Math.pow(a, b ** c) is converted to a ** b ** c
        return astUtils.getPrecedence(exponent) < PRECEDENCE_OF_EXPONENTIATION_EXPR;
    }
    /**
     * Determines whether an exponentiation binary expression at the place of the given node would need parens.
     * @param {ASTNode} node A node that would be replaced by an exponentiation binary expression.
     * @param {SourceCode} sourceCode A SourceCode object.
     * @returns {boolean} `true` if the expression needs to be parenthesised.
     */
    function doesExponentiationExpressionNeedParens(node, sourceCode) {
        const parent = node.parent.type === "ChainExpression" ? node.parent.parent : node.parent;
        const needsParens = (parent.type === "ClassDeclaration" ||
            (parent.type.endsWith("Expression") &&
                astUtils.getPrecedence(parent) >= PRECEDENCE_OF_EXPONENTIATION_EXPR &&
                !(parent.type === "BinaryExpression" && parent.operator === "**" && parent.right === node) &&
                !((parent.type === "CallExpression" || parent.type === "NewExpression") && parent.arguments.includes(node)) &&
                !(parent.type === "MemberExpression" && parent.computed && parent.property === node) &&
                !(parent.type === "ArrayExpression")));
        return needsParens && !astUtils.isParenthesised(sourceCode, node);
    }
    /**
     * Optionally parenthesizes given text.
     * @param {string} text The text to parenthesize.
     * @param {boolean} shouldParenthesize If `true`, the text will be parenthesised.
     * @returns {string} parenthesised or unchanged text.
     */
    function parenthesizeIfShould(text, shouldParenthesize) {
        return shouldParenthesize ? `(${text})` : text;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow the use of `Math.pow` in favor of the `**` operator",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-exponentiation-operator"
            },
            schema: [],
            fixable: "code",
            messages: {
                useExponentiation: "Use the '**' operator instead of 'Math.pow'."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Reports the given node.
             * @param {ASTNode} node 'Math.pow()' node to report.
             * @returns {void}
             */
            function report(node) {
                context.report({
                    node,
                    messageId: "useExponentiation",
                    fix(fixer) {
                        if (node.arguments.length !== 2 ||
                            node.arguments.some(arg => arg.type === "SpreadElement") ||
                            sourceCode.getCommentsInside(node).length > 0) {
                            return null;
                        }
                        const base = node.arguments[0], exponent = node.arguments[1], baseText = sourceCode.getText(base), exponentText = sourceCode.getText(exponent), shouldParenthesizeBase = doesBaseNeedParens(base), shouldParenthesizeExponent = doesExponentNeedParens(exponent), shouldParenthesizeAll = doesExponentiationExpressionNeedParens(node, sourceCode);
                        let prefix = "", suffix = "";
                        if (!shouldParenthesizeAll) {
                            if (!shouldParenthesizeBase) {
                                const firstReplacementToken = sourceCode.getFirstToken(base), tokenBefore = sourceCode.getTokenBefore(node);
                                if (tokenBefore &&
                                    tokenBefore.range[1] === node.range[0] &&
                                    !astUtils.canTokensBeAdjacent(tokenBefore, firstReplacementToken)) {
                                    prefix = " "; // a+Math.pow(++b, c) -> a+ ++b**c
                                }
                            }
                            if (!shouldParenthesizeExponent) {
                                const lastReplacementToken = sourceCode.getLastToken(exponent), tokenAfter = sourceCode.getTokenAfter(node);
                                if (tokenAfter &&
                                    node.range[1] === tokenAfter.range[0] &&
                                    !astUtils.canTokensBeAdjacent(lastReplacementToken, tokenAfter)) {
                                    suffix = " "; // Math.pow(a, b)in c -> a**b in c
                                }
                            }
                        }
                        const baseReplacement = parenthesizeIfShould(baseText, shouldParenthesizeBase), exponentReplacement = parenthesizeIfShould(exponentText, shouldParenthesizeExponent), replacement = parenthesizeIfShould(`${baseReplacement}**${exponentReplacement}`, shouldParenthesizeAll);
                        return fixer.replaceText(node, `${prefix}${replacement}${suffix}`);
                    }
                });
            }
            return {
                Program() {
                    const scope = context.getScope();
                    const tracker = new ReferenceTracker(scope);
                    const trackMap = {
                        Math: {
                            pow: { [CALL]: true }
                        }
                    };
                    for (const { node } of tracker.iterateGlobalReferences(trackMap)) {
                        report(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-named-capture-group.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { CALL, CONSTRUCT, ReferenceTracker, getStringIfConstant } = require(350) /* eslint-utils */;
    const regexpp = require(497) /* regexpp */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const parser = new regexpp.RegExpParser();
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce using named capture group in regular expression",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-named-capture-group"
            },
            schema: [],
            messages: {
                required: "Capture group '{{group}}' should be converted to a named or non-capturing group."
            }
        },
        create(context) {
            /**
             * Function to check regular expression.
             * @param {string} pattern The regular expression pattern to be check.
             * @param {ASTNode} node AST node which contains regular expression.
             * @param {boolean} uFlag Flag indicates whether unicode mode is enabled or not.
             * @returns {void}
             */
            function checkRegex(pattern, node, uFlag) {
                let ast;
                try {
                    ast = parser.parsePattern(pattern, 0, pattern.length, uFlag);
                }
                catch {
                    // ignore regex syntax errors
                    return;
                }
                regexpp.visitRegExpAST(ast, {
                    onCapturingGroupEnter(group) {
                        if (!group.name) {
                            context.report({
                                node,
                                messageId: "required",
                                data: {
                                    group: group.raw
                                }
                            });
                        }
                    }
                });
            }
            return {
                Literal(node) {
                    if (node.regex) {
                        checkRegex(node.regex.pattern, node, node.regex.flags.includes("u"));
                    }
                },
                Program() {
                    const scope = context.getScope();
                    const tracker = new ReferenceTracker(scope);
                    const traceMap = {
                        RegExp: {
                            [CALL]: true,
                            [CONSTRUCT]: true
                        }
                    };
                    for (const { node } of tracker.iterateGlobalReferences(traceMap)) {
                        const regex = getStringIfConstant(node.arguments[0]);
                        const flags = getStringIfConstant(node.arguments[1]);
                        if (regex) {
                            checkRegex(regex, node, flags && flags.includes("u"));
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-numeric-literals.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const radixMap = new Map([
        [2, { system: "binary", literalPrefix: "0b" }],
        [8, { system: "octal", literalPrefix: "0o" }],
        [16, { system: "hexadecimal", literalPrefix: "0x" }]
    ]);
    /**
     * Checks to see if a CallExpression's callee node is `parseInt` or
     * `Number.parseInt`.
     * @param {ASTNode} calleeNode The callee node to evaluate.
     * @returns {boolean} True if the callee is `parseInt` or `Number.parseInt`,
     * false otherwise.
     */
    function isParseInt(calleeNode) {
        return (astUtils.isSpecificId(calleeNode, "parseInt") ||
            astUtils.isSpecificMemberAccess(calleeNode, "Number", "parseInt"));
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow `parseInt()` and `Number.parseInt()` in favor of binary, octal, and hexadecimal literals",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-numeric-literals"
            },
            schema: [],
            messages: {
                useLiteral: "Use {{system}} literals instead of {{functionName}}()."
            },
            fixable: "code"
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                "CallExpression[arguments.length=2]"(node) {
                    const [strNode, radixNode] = node.arguments, str = astUtils.getStaticStringValue(strNode), radix = radixNode.value;
                    if (str !== null &&
                        astUtils.isStringLiteral(strNode) &&
                        radixNode.type === "Literal" &&
                        typeof radix === "number" &&
                        radixMap.has(radix) &&
                        isParseInt(node.callee)) {
                        const { system, literalPrefix } = radixMap.get(radix);
                        context.report({
                            node,
                            messageId: "useLiteral",
                            data: {
                                system,
                                functionName: sourceCode.getText(node.callee)
                            },
                            fix(fixer) {
                                if (sourceCode.getCommentsInside(node).length) {
                                    return null;
                                }
                                const replacement = `${literalPrefix}${str}`;
                                if (+replacement !== parseInt(str, radix)) {
                                    /*
                                     * If the newly-produced literal would be invalid, (e.g. 0b1234),
                                     * or it would yield an incorrect parseInt result for some other reason, don't make a fix.
                                     *
                                     * If `str` had numeric separators, `+replacement` will evaluate to `NaN` because unary `+`
                                     * per the specification doesn't support numeric separators. Thus, the above condition will be `true`
                                     * (`NaN !== anything` is always `true`) regardless of the `parseInt(str, radix)` value.
                                     * Consequently, no autofixes will be made. This is correct behavior because `parseInt` also
                                     * doesn't support numeric separators, but it does parse part of the string before the first `_`,
                                     * so the autofix would be invalid:
                                     *
                                     *   parseInt("1_1", 2) // === 1
                                     *   0b1_1 // === 3
                                     */
                                    return null;
                                }
                                const tokenBefore = sourceCode.getTokenBefore(node), tokenAfter = sourceCode.getTokenAfter(node);
                                let prefix = "", suffix = "";
                                if (tokenBefore &&
                                    tokenBefore.range[1] === node.range[0] &&
                                    !astUtils.canTokensBeAdjacent(tokenBefore, replacement)) {
                                    prefix = " ";
                                }
                                if (tokenAfter &&
                                    node.range[1] === tokenAfter.range[0] &&
                                    !astUtils.canTokensBeAdjacent(replacement, tokenAfter)) {
                                    suffix = " ";
                                }
                                return fixer.replaceText(node, `${prefix}${replacement}${suffix}`);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-object-spread.js */ function _(require, module, exports, __esModule, __esExport) {
    const { CALL, ReferenceTracker } = require(350) /* eslint-utils */;
    const { isCommaToken, isOpeningParenToken, isClosingParenToken, isParenthesised } = require(394) /* ./utils/ast-utils */;
    const ANY_SPACE = /\s/u;
    /**
     * Helper that checks if the Object.assign call has array spread
     * @param {ASTNode} node The node that the rule warns on
     * @returns {boolean} - Returns true if the Object.assign call has array spread
     */
    function hasArraySpread(node) {
        return node.arguments.some(arg => arg.type === "SpreadElement");
    }
    /**
     * Determines whether the given node is an accessor property (getter/setter).
     * @param {ASTNode} node Node to check.
     * @returns {boolean} `true` if the node is a getter or a setter.
     */
    function isAccessorProperty(node) {
        return node.type === "Property" &&
            (node.kind === "get" || node.kind === "set");
    }
    /**
     * Determines whether the given object expression node has accessor properties (getters/setters).
     * @param {ASTNode} node `ObjectExpression` node to check.
     * @returns {boolean} `true` if the node has at least one getter/setter.
     */
    function hasAccessors(node) {
        return node.properties.some(isAccessorProperty);
    }
    /**
     * Determines whether the given call expression node has object expression arguments with accessor properties (getters/setters).
     * @param {ASTNode} node `CallExpression` node to check.
     * @returns {boolean} `true` if the node has at least one argument that is an object expression with at least one getter/setter.
     */
    function hasArgumentsWithAccessors(node) {
        return node.arguments
            .filter(arg => arg.type === "ObjectExpression")
            .some(hasAccessors);
    }
    /**
     * Helper that checks if the node needs parentheses to be valid JS.
     * The default is to wrap the node in parentheses to avoid parsing errors.
     * @param {ASTNode} node The node that the rule warns on
     * @param {Object} sourceCode in context sourcecode object
     * @returns {boolean} - Returns true if the node needs parentheses
     */
    function needsParens(node, sourceCode) {
        const parent = node.parent;
        switch (parent.type) {
            case "VariableDeclarator":
            case "ArrayExpression":
            case "ReturnStatement":
            case "CallExpression":
            case "Property":
                return false;
            case "AssignmentExpression":
                return parent.left === node && !isParenthesised(sourceCode, node);
            default:
                return !isParenthesised(sourceCode, node);
        }
    }
    /**
     * Determines if an argument needs parentheses. The default is to not add parens.
     * @param {ASTNode} node The node to be checked.
     * @param {Object} sourceCode in context sourcecode object
     * @returns {boolean} True if the node needs parentheses
     */
    function argNeedsParens(node, sourceCode) {
        switch (node.type) {
            case "AssignmentExpression":
            case "ArrowFunctionExpression":
            case "ConditionalExpression":
                return !isParenthesised(sourceCode, node);
            default:
                return false;
        }
    }
    /**
     * Get the parenthesis tokens of a given ObjectExpression node.
     * This includes the braces of the object literal and enclosing parentheses.
     * @param {ASTNode} node The node to get.
     * @param {Token} leftArgumentListParen The opening paren token of the argument list.
     * @param {SourceCode} sourceCode The source code object to get tokens.
     * @returns {Token[]} The parenthesis tokens of the node. This is sorted by the location.
     */
    function getParenTokens(node, leftArgumentListParen, sourceCode) {
        const parens = [sourceCode.getFirstToken(node), sourceCode.getLastToken(node)];
        let leftNext = sourceCode.getTokenBefore(node);
        let rightNext = sourceCode.getTokenAfter(node);
        // Note: don't include the parens of the argument list.
        while (leftNext &&
            rightNext &&
            leftNext.range[0] > leftArgumentListParen.range[0] &&
            isOpeningParenToken(leftNext) &&
            isClosingParenToken(rightNext)) {
            parens.push(leftNext, rightNext);
            leftNext = sourceCode.getTokenBefore(leftNext);
            rightNext = sourceCode.getTokenAfter(rightNext);
        }
        return parens.sort((a, b) => a.range[0] - b.range[0]);
    }
    /**
     * Get the range of a given token and around whitespaces.
     * @param {Token} token The token to get range.
     * @param {SourceCode} sourceCode The source code object to get tokens.
     * @returns {number} The end of the range of the token and around whitespaces.
     */
    function getStartWithSpaces(token, sourceCode) {
        const text = sourceCode.text;
        let start = token.range[0];
        // If the previous token is a line comment then skip this step to avoid commenting this token out.
        {
            const prevToken = sourceCode.getTokenBefore(token, { includeComments: true });
            if (prevToken && prevToken.type === "Line") {
                return start;
            }
        }
        // Detect spaces before the token.
        while (ANY_SPACE.test(text[start - 1] || "")) {
            start -= 1;
        }
        return start;
    }
    /**
     * Get the range of a given token and around whitespaces.
     * @param {Token} token The token to get range.
     * @param {SourceCode} sourceCode The source code object to get tokens.
     * @returns {number} The start of the range of the token and around whitespaces.
     */
    function getEndWithSpaces(token, sourceCode) {
        const text = sourceCode.text;
        let end = token.range[1];
        // Detect spaces after the token.
        while (ANY_SPACE.test(text[end] || "")) {
            end += 1;
        }
        return end;
    }
    /**
     * Autofixes the Object.assign call to use an object spread instead.
     * @param {ASTNode|null} node The node that the rule warns on, i.e. the Object.assign call
     * @param {string} sourceCode sourceCode of the Object.assign call
     * @returns {Function} autofixer - replaces the Object.assign with a spread object.
     */
    function defineFixer(node, sourceCode) {
        return function* (fixer) {
            const leftParen = sourceCode.getTokenAfter(node.callee, isOpeningParenToken);
            const rightParen = sourceCode.getLastToken(node);
            // Remove everything before the opening paren: callee `Object.assign`, type arguments, and whitespace between the callee and the paren.
            yield fixer.removeRange([node.range[0], leftParen.range[0]]);
            // Replace the parens of argument list to braces.
            if (needsParens(node, sourceCode)) {
                yield fixer.replaceText(leftParen, "({");
                yield fixer.replaceText(rightParen, "})");
            }
            else {
                yield fixer.replaceText(leftParen, "{");
                yield fixer.replaceText(rightParen, "}");
            }
            // Process arguments.
            for (const argNode of node.arguments) {
                const innerParens = getParenTokens(argNode, leftParen, sourceCode);
                const left = innerParens.shift();
                const right = innerParens.pop();
                if (argNode.type === "ObjectExpression") {
                    const maybeTrailingComma = sourceCode.getLastToken(argNode, 1);
                    const maybeArgumentComma = sourceCode.getTokenAfter(right);
                    /*
                     * Make bare this object literal.
                     * And remove spaces inside of the braces for better formatting.
                     */
                    for (const innerParen of innerParens) {
                        yield fixer.remove(innerParen);
                    }
                    const leftRange = [left.range[0], getEndWithSpaces(left, sourceCode)];
                    const rightRange = [
                        Math.max(getStartWithSpaces(right, sourceCode), leftRange[1]),
                        right.range[1]
                    ];
                    yield fixer.removeRange(leftRange);
                    yield fixer.removeRange(rightRange);
                    // Remove the comma of this argument if it's duplication.
                    if ((argNode.properties.length === 0 || isCommaToken(maybeTrailingComma)) &&
                        isCommaToken(maybeArgumentComma)) {
                        yield fixer.remove(maybeArgumentComma);
                    }
                }
                else {
                    // Make spread.
                    if (argNeedsParens(argNode, sourceCode)) {
                        yield fixer.insertTextBefore(left, "...(");
                        yield fixer.insertTextAfter(right, ")");
                    }
                    else {
                        yield fixer.insertTextBefore(left, "...");
                    }
                }
            }
        };
    }
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow using Object.assign with an object literal as the first argument and prefer the use of object spread instead.",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-object-spread"
            },
            schema: [],
            fixable: "code",
            messages: {
                useSpreadMessage: "Use an object spread instead of `Object.assign` eg: `{ ...foo }`.",
                useLiteralMessage: "Use an object literal instead of `Object.assign`. eg: `{ foo: bar }`."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                Program() {
                    const scope = context.getScope();
                    const tracker = new ReferenceTracker(scope);
                    const trackMap = {
                        Object: {
                            assign: { [CALL]: true }
                        }
                    };
                    // Iterate all calls of `Object.assign` (only of the global variable `Object`).
                    for (const { node } of tracker.iterateGlobalReferences(trackMap)) {
                        if (node.arguments.length >= 1 &&
                            node.arguments[0].type === "ObjectExpression" &&
                            !hasArraySpread(node) &&
                            !(node.arguments.length > 1 &&
                                hasArgumentsWithAccessors(node))) {
                            const messageId = node.arguments.length === 1
                                ? "useLiteralMessage"
                                : "useSpreadMessage";
                            const fix = defineFixer(node, sourceCode);
                            context.report({ node, messageId, fix });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-promise-reject-errors.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require using Error objects as Promise rejection reasons",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-promise-reject-errors"
            },
            fixable: null,
            schema: [
                {
                    type: "object",
                    properties: {
                        allowEmptyReject: { type: "boolean", default: false }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                rejectAnError: "Expected the Promise rejection reason to be an Error."
            }
        },
        create(context) {
            const ALLOW_EMPTY_REJECT = context.options.length && context.options[0].allowEmptyReject;
            //----------------------------------------------------------------------
            // Helpers
            //----------------------------------------------------------------------
            /**
             * Checks the argument of a reject() or Promise.reject() CallExpression, and reports it if it can't be an Error
             * @param {ASTNode} callExpression A CallExpression node which is used to reject a Promise
             * @returns {void}
             */
            function checkRejectCall(callExpression) {
                if (!callExpression.arguments.length && ALLOW_EMPTY_REJECT) {
                    return;
                }
                if (!callExpression.arguments.length ||
                    !astUtils.couldBeError(callExpression.arguments[0]) ||
                    callExpression.arguments[0].type === "Identifier" && callExpression.arguments[0].name === "undefined") {
                    context.report({
                        node: callExpression,
                        messageId: "rejectAnError"
                    });
                }
            }
            /**
             * Determines whether a function call is a Promise.reject() call
             * @param {ASTNode} node A CallExpression node
             * @returns {boolean} `true` if the call is a Promise.reject() call
             */
            function isPromiseRejectCall(node) {
                return astUtils.isSpecificMemberAccess(node.callee, "Promise", "reject");
            }
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                // Check `Promise.reject(value)` calls.
                CallExpression(node) {
                    if (isPromiseRejectCall(node)) {
                        checkRejectCall(node);
                    }
                },
                /*
                 * Check for `new Promise((resolve, reject) => {})`, and check for reject() calls.
                 * This function is run on "NewExpression:exit" instead of "NewExpression" to ensure that
                 * the nodes in the expression already have the `parent` property.
                 */
                "NewExpression:exit"(node) {
                    if (node.callee.type === "Identifier" && node.callee.name === "Promise" &&
                        node.arguments.length && astUtils.isFunction(node.arguments[0]) &&
                        node.arguments[0].params.length > 1 && node.arguments[0].params[1].type === "Identifier") {
                        context.getDeclaredVariables(node.arguments[0])
                            /*
                             * Find the first variable that matches the second parameter's name.
                             * If the first parameter has the same name as the second parameter, then the variable will actually
                             * be "declared" when the first parameter is evaluated, but then it will be immediately overwritten
                             * by the second parameter. It's not possible for an expression with the variable to be evaluated before
                             * the variable is overwritten, because functions with duplicate parameters cannot have destructuring or
                             * default assignments in their parameter lists. Therefore, it's not necessary to explicitly account for
                             * this case.
                             */
                            .find(variable => variable.name === node.arguments[0].params[1].name)
                            // Get the references to that variable.
                            .references
                            // Only check the references that read the parameter's value.
                            .filter(ref => ref.isRead())
                            // Only check the references that are used as the callee in a function call, e.g. `reject(foo)`.
                            .filter(ref => ref.identifier.parent.type === "CallExpression" && ref.identifier === ref.identifier.parent.callee)
                            // Check the argument of the function call to determine whether it's an Error.
                            .forEach(ref => checkRejectCall(ref.identifier.parent));
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-reflect.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require `Reflect` methods where applicable",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-reflect"
            },
            deprecated: true,
            replacedBy: [],
            schema: [
                {
                    type: "object",
                    properties: {
                        exceptions: {
                            type: "array",
                            items: {
                                enum: [
                                    "apply",
                                    "call",
                                    "delete",
                                    "defineProperty",
                                    "getOwnPropertyDescriptor",
                                    "getPrototypeOf",
                                    "setPrototypeOf",
                                    "isExtensible",
                                    "getOwnPropertyNames",
                                    "preventExtensions"
                                ]
                            },
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                preferReflect: "Avoid using {{existing}}, instead use {{substitute}}."
            }
        },
        create(context) {
            const existingNames = {
                apply: "Function.prototype.apply",
                call: "Function.prototype.call",
                defineProperty: "Object.defineProperty",
                getOwnPropertyDescriptor: "Object.getOwnPropertyDescriptor",
                getPrototypeOf: "Object.getPrototypeOf",
                setPrototypeOf: "Object.setPrototypeOf",
                isExtensible: "Object.isExtensible",
                getOwnPropertyNames: "Object.getOwnPropertyNames",
                preventExtensions: "Object.preventExtensions"
            };
            const reflectSubstitutes = {
                apply: "Reflect.apply",
                call: "Reflect.apply",
                defineProperty: "Reflect.defineProperty",
                getOwnPropertyDescriptor: "Reflect.getOwnPropertyDescriptor",
                getPrototypeOf: "Reflect.getPrototypeOf",
                setPrototypeOf: "Reflect.setPrototypeOf",
                isExtensible: "Reflect.isExtensible",
                getOwnPropertyNames: "Reflect.getOwnPropertyNames",
                preventExtensions: "Reflect.preventExtensions"
            };
            const exceptions = (context.options[0] || {}).exceptions || [];
            /**
             * Reports the Reflect violation based on the `existing` and `substitute`
             * @param {Object} node The node that violates the rule.
             * @param {string} existing The existing method name that has been used.
             * @param {string} substitute The Reflect substitute that should be used.
             * @returns {void}
             */
            function report(node, existing, substitute) {
                context.report({
                    node,
                    messageId: "preferReflect",
                    data: {
                        existing,
                        substitute
                    }
                });
            }
            return {
                CallExpression(node) {
                    const methodName = (node.callee.property || {}).name;
                    const isReflectCall = (node.callee.object || {}).name === "Reflect";
                    const hasReflectSubstitute = Object.prototype.hasOwnProperty.call(reflectSubstitutes, methodName);
                    const userConfiguredException = exceptions.indexOf(methodName) !== -1;
                    if (hasReflectSubstitute && !isReflectCall && !userConfiguredException) {
                        report(node, existingNames[methodName], reflectSubstitutes[methodName]);
                    }
                },
                UnaryExpression(node) {
                    const isDeleteOperator = node.operator === "delete";
                    const targetsIdentifier = node.argument.type === "Identifier";
                    const userConfiguredException = exceptions.indexOf("delete") !== -1;
                    if (isDeleteOperator && !targetsIdentifier && !userConfiguredException) {
                        report(node, "the delete keyword", "Reflect.deleteProperty");
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-regex-literals.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const { CALL, CONSTRUCT, ReferenceTracker, findVariable } = require(350) /* eslint-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Determines whether the given node is a string literal.
     * @param {ASTNode} node Node to check.
     * @returns {boolean} True if the node is a string literal.
     */
    function isStringLiteral(node) {
        return node.type === "Literal" && typeof node.value === "string";
    }
    /**
     * Determines whether the given node is a regex literal.
     * @param {ASTNode} node Node to check.
     * @returns {boolean} True if the node is a regex literal.
     */
    function isRegexLiteral(node) {
        return node.type === "Literal" && Object.prototype.hasOwnProperty.call(node, "regex");
    }
    /**
     * Determines whether the given node is a template literal without expressions.
     * @param {ASTNode} node Node to check.
     * @returns {boolean} True if the node is a template literal without expressions.
     */
    function isStaticTemplateLiteral(node) {
        return node.type === "TemplateLiteral" && node.expressions.length === 0;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow use of the `RegExp` constructor in favor of regular expression literals",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-regex-literals"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        disallowRedundantWrapping: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedRegExp: "Use a regular expression literal instead of the 'RegExp' constructor.",
                unexpectedRedundantRegExp: "Regular expression literal is unnecessarily wrapped within a 'RegExp' constructor.",
                unexpectedRedundantRegExpWithFlags: "Use regular expression literal with flags instead of the 'RegExp' constructor."
            }
        },
        create(context) {
            const [{ disallowRedundantWrapping = false } = {}] = context.options;
            /**
             * Determines whether the given identifier node is a reference to a global variable.
             * @param {ASTNode} node `Identifier` node to check.
             * @returns {boolean} True if the identifier is a reference to a global variable.
             */
            function isGlobalReference(node) {
                const scope = context.getScope();
                const variable = findVariable(scope, node);
                return variable !== null && variable.scope.type === "global" && variable.defs.length === 0;
            }
            /**
             * Determines whether the given node is a String.raw`` tagged template expression
             * with a static template literal.
             * @param {ASTNode} node Node to check.
             * @returns {boolean} True if the node is String.raw`` with a static template.
             */
            function isStringRawTaggedStaticTemplateLiteral(node) {
                return node.type === "TaggedTemplateExpression" &&
                    astUtils.isSpecificMemberAccess(node.tag, "String", "raw") &&
                    isGlobalReference(astUtils.skipChainExpression(node.tag).object) &&
                    isStaticTemplateLiteral(node.quasi);
            }
            /**
             * Determines whether the given node is considered to be a static string by the logic of this rule.
             * @param {ASTNode} node Node to check.
             * @returns {boolean} True if the node is a static string.
             */
            function isStaticString(node) {
                return isStringLiteral(node) ||
                    isStaticTemplateLiteral(node) ||
                    isStringRawTaggedStaticTemplateLiteral(node);
            }
            /**
             * Determines whether the relevant arguments of the given are all static string literals.
             * @param {ASTNode} node Node to check.
             * @returns {boolean} True if all arguments are static strings.
             */
            function hasOnlyStaticStringArguments(node) {
                const args = node.arguments;
                if ((args.length === 1 || args.length === 2) && args.every(isStaticString)) {
                    return true;
                }
                return false;
            }
            /**
             * Determines whether the arguments of the given node indicate that a regex literal is unnecessarily wrapped.
             * @param {ASTNode} node Node to check.
             * @returns {boolean} True if the node already contains a regex literal argument.
             */
            function isUnnecessarilyWrappedRegexLiteral(node) {
                const args = node.arguments;
                if (args.length === 1 && isRegexLiteral(args[0])) {
                    return true;
                }
                if (args.length === 2 && isRegexLiteral(args[0]) && isStaticString(args[1])) {
                    return true;
                }
                return false;
            }
            return {
                Program() {
                    const scope = context.getScope();
                    const tracker = new ReferenceTracker(scope);
                    const traceMap = {
                        RegExp: {
                            [CALL]: true,
                            [CONSTRUCT]: true
                        }
                    };
                    for (const { node } of tracker.iterateGlobalReferences(traceMap)) {
                        if (disallowRedundantWrapping && isUnnecessarilyWrappedRegexLiteral(node)) {
                            if (node.arguments.length === 2) {
                                context.report({ node, messageId: "unexpectedRedundantRegExpWithFlags" });
                            }
                            else {
                                context.report({ node, messageId: "unexpectedRedundantRegExp" });
                            }
                        }
                        else if (hasOnlyStaticStringArguments(node)) {
                            context.report({ node, messageId: "unexpectedRegExp" });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-rest-params.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Gets the variable object of `arguments` which is defined implicitly.
     * @param {eslint-scope.Scope} scope A scope to get.
     * @returns {eslint-scope.Variable} The found variable object.
     */
    function getVariableOfArguments(scope) {
        const variables = scope.variables;
        for (let i = 0; i < variables.length; ++i) {
            const variable = variables[i];
            if (variable.name === "arguments") {
                /*
                 * If there was a parameter which is named "arguments", the implicit "arguments" is not defined.
                 * So does fast return with null.
                 */
                return (variable.identifiers.length === 0) ? variable : null;
            }
        }
        /* istanbul ignore next : unreachable */
        return null;
    }
    /**
     * Checks if the given reference is not normal member access.
     *
     * - arguments         .... true    // not member access
     * - arguments[i]      .... true    // computed member access
     * - arguments[0]      .... true    // computed member access
     * - arguments.length  .... false   // normal member access
     * @param {eslint-scope.Reference} reference The reference to check.
     * @returns {boolean} `true` if the reference is not normal member access.
     */
    function isNotNormalMemberAccess(reference) {
        const id = reference.identifier;
        const parent = id.parent;
        return !(parent.type === "MemberExpression" &&
            parent.object === id &&
            !parent.computed);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require rest parameters instead of `arguments`",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-rest-params"
            },
            schema: [],
            messages: {
                preferRestParams: "Use the rest parameters instead of 'arguments'."
            }
        },
        create(context) {
            /**
             * Reports a given reference.
             * @param {eslint-scope.Reference} reference A reference to report.
             * @returns {void}
             */
            function report(reference) {
                context.report({
                    node: reference.identifier,
                    loc: reference.identifier.loc,
                    messageId: "preferRestParams"
                });
            }
            /**
             * Reports references of the implicit `arguments` variable if exist.
             * @returns {void}
             */
            function checkForArguments() {
                const argumentsVar = getVariableOfArguments(context.getScope());
                if (argumentsVar) {
                    argumentsVar
                        .references
                        .filter(isNotNormalMemberAccess)
                        .forEach(report);
                }
            }
            return {
                "FunctionDeclaration:exit": checkForArguments,
                "FunctionExpression:exit": checkForArguments
            };
        }
    };
},
/* eslint/lib/rules/prefer-spread.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a node is a `.apply()` for variadic.
     * @param {ASTNode} node A CallExpression node to check.
     * @returns {boolean} Whether or not the node is a `.apply()` for variadic.
     */
    function isVariadicApplyCalling(node) {
        return (astUtils.isSpecificMemberAccess(node.callee, null, "apply") &&
            node.arguments.length === 2 &&
            node.arguments[1].type !== "ArrayExpression" &&
            node.arguments[1].type !== "SpreadElement");
    }
    /**
     * Checks whether or not `thisArg` is not changed by `.apply()`.
     * @param {ASTNode|null} expectedThis The node that is the owner of the applied function.
     * @param {ASTNode} thisArg The node that is given to the first argument of the `.apply()`.
     * @param {RuleContext} context The ESLint rule context object.
     * @returns {boolean} Whether or not `thisArg` is not changed by `.apply()`.
     */
    function isValidThisArg(expectedThis, thisArg, context) {
        if (!expectedThis) {
            return astUtils.isNullOrUndefined(thisArg);
        }
        return astUtils.equalTokens(expectedThis, thisArg, context);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require spread operators instead of `.apply()`",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-spread"
            },
            schema: [],
            fixable: null,
            messages: {
                preferSpread: "Use the spread operator instead of '.apply()'."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                CallExpression(node) {
                    if (!isVariadicApplyCalling(node)) {
                        return;
                    }
                    const applied = astUtils.skipChainExpression(astUtils.skipChainExpression(node.callee).object);
                    const expectedThis = (applied.type === "MemberExpression") ? applied.object : null;
                    const thisArg = node.arguments[0];
                    if (isValidThisArg(expectedThis, thisArg, sourceCode)) {
                        context.report({
                            node,
                            messageId: "preferSpread"
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/prefer-template.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether or not a given node is a concatenation.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node is a concatenation.
     */
    function isConcatenation(node) {
        return node.type === "BinaryExpression" && node.operator === "+";
    }
    /**
     * Gets the top binary expression node for concatenation in parents of a given node.
     * @param {ASTNode} node A node to get.
     * @returns {ASTNode} the top binary expression node in parents of a given node.
     */
    function getTopConcatBinaryExpression(node) {
        let currentNode = node;
        while (isConcatenation(currentNode.parent)) {
            currentNode = currentNode.parent;
        }
        return currentNode;
    }
    /**
     * Checks whether or not a node contains a string literal with an octal or non-octal decimal escape sequence
     * @param {ASTNode} node A node to check
     * @returns {boolean} `true` if at least one string literal within the node contains
     * an octal or non-octal decimal escape sequence
     */
    function hasOctalOrNonOctalDecimalEscapeSequence(node) {
        if (isConcatenation(node)) {
            return (hasOctalOrNonOctalDecimalEscapeSequence(node.left) ||
                hasOctalOrNonOctalDecimalEscapeSequence(node.right));
        }
        // No need to check TemplateLiterals – would throw parsing error
        if (node.type === "Literal" && typeof node.value === "string") {
            return astUtils.hasOctalOrNonOctalDecimalEscapeSequence(node.raw);
        }
        return false;
    }
    /**
     * Checks whether or not a given binary expression has string literals.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node has string literals.
     */
    function hasStringLiteral(node) {
        if (isConcatenation(node)) {
            // `left` is deeper than `right` normally.
            return hasStringLiteral(node.right) || hasStringLiteral(node.left);
        }
        return astUtils.isStringLiteral(node);
    }
    /**
     * Checks whether or not a given binary expression has non string literals.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node has non string literals.
     */
    function hasNonStringLiteral(node) {
        if (isConcatenation(node)) {
            // `left` is deeper than `right` normally.
            return hasNonStringLiteral(node.right) || hasNonStringLiteral(node.left);
        }
        return !astUtils.isStringLiteral(node);
    }
    /**
     * Determines whether a given node will start with a template curly expression (`${}`) when being converted to a template literal.
     * @param {ASTNode} node The node that will be fixed to a template literal
     * @returns {boolean} `true` if the node will start with a template curly.
     */
    function startsWithTemplateCurly(node) {
        if (node.type === "BinaryExpression") {
            return startsWithTemplateCurly(node.left);
        }
        if (node.type === "TemplateLiteral") {
            return node.expressions.length && node.quasis.length && node.quasis[0].range[0] === node.quasis[0].range[1];
        }
        return node.type !== "Literal" || typeof node.value !== "string";
    }
    /**
     * Determines whether a given node end with a template curly expression (`${}`) when being converted to a template literal.
     * @param {ASTNode} node The node that will be fixed to a template literal
     * @returns {boolean} `true` if the node will end with a template curly.
     */
    function endsWithTemplateCurly(node) {
        if (node.type === "BinaryExpression") {
            return startsWithTemplateCurly(node.right);
        }
        if (node.type === "TemplateLiteral") {
            return node.expressions.length && node.quasis.length && node.quasis[node.quasis.length - 1].range[0] === node.quasis[node.quasis.length - 1].range[1];
        }
        return node.type !== "Literal" || typeof node.value !== "string";
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require template literals instead of string concatenation",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/prefer-template"
            },
            schema: [],
            fixable: "code",
            messages: {
                unexpectedStringConcatenation: "Unexpected string concatenation."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            let done = Object.create(null);
            /**
             * Gets the non-token text between two nodes, ignoring any other tokens that appear between the two tokens.
             * @param {ASTNode} node1 The first node
             * @param {ASTNode} node2 The second node
             * @returns {string} The text between the nodes, excluding other tokens
             */
            function getTextBetween(node1, node2) {
                const allTokens = [node1].concat(sourceCode.getTokensBetween(node1, node2)).concat(node2);
                const sourceText = sourceCode.getText();
                return allTokens.slice(0, -1).reduce((accumulator, token, index) => accumulator + sourceText.slice(token.range[1], allTokens[index + 1].range[0]), "");
            }
            /**
             * Returns a template literal form of the given node.
             * @param {ASTNode} currentNode A node that should be converted to a template literal
             * @param {string} textBeforeNode Text that should appear before the node
             * @param {string} textAfterNode Text that should appear after the node
             * @returns {string} A string form of this node, represented as a template literal
             */
            function getTemplateLiteral(currentNode, textBeforeNode, textAfterNode) {
                if (currentNode.type === "Literal" && typeof currentNode.value === "string") {
                    /*
                     * If the current node is a string literal, escape any instances of ${ or ` to prevent them from being interpreted
                     * as a template placeholder. However, if the code already contains a backslash before the ${ or `
                     * for some reason, don't add another backslash, because that would change the meaning of the code (it would cause
                     * an actual backslash character to appear before the dollar sign).
                     */
                    return `\`${currentNode.raw.slice(1, -1).replace(/\\*(\$\{|`)/gu, matched => {
                        if (matched.lastIndexOf("\\") % 2) {
                            return `\\${matched}`;
                        }
                        return matched;
                        // Unescape any quotes that appear in the original Literal that no longer need to be escaped.
                    }).replace(new RegExp(`\\\\${currentNode.raw[0]}`, "gu"), currentNode.raw[0])}\``;
                }
                if (currentNode.type === "TemplateLiteral") {
                    return sourceCode.getText(currentNode);
                }
                if (isConcatenation(currentNode) && hasStringLiteral(currentNode) && hasNonStringLiteral(currentNode)) {
                    const plusSign = sourceCode.getFirstTokenBetween(currentNode.left, currentNode.right, token => token.value === "+");
                    const textBeforePlus = getTextBetween(currentNode.left, plusSign);
                    const textAfterPlus = getTextBetween(plusSign, currentNode.right);
                    const leftEndsWithCurly = endsWithTemplateCurly(currentNode.left);
                    const rightStartsWithCurly = startsWithTemplateCurly(currentNode.right);
                    if (leftEndsWithCurly) {
                        // If the left side of the expression ends with a template curly, add the extra text to the end of the curly bracket.
                        // `foo${bar}` /* comment */ + 'baz' --> `foo${bar /* comment */  }${baz}`
                        return getTemplateLiteral(currentNode.left, textBeforeNode, textBeforePlus + textAfterPlus).slice(0, -1) +
                            getTemplateLiteral(currentNode.right, null, textAfterNode).slice(1);
                    }
                    if (rightStartsWithCurly) {
                        // Otherwise, if the right side of the expression starts with a template curly, add the text there.
                        // 'foo' /* comment */ + `${bar}baz` --> `foo${ /* comment */  bar}baz`
                        return getTemplateLiteral(currentNode.left, textBeforeNode, null).slice(0, -1) +
                            getTemplateLiteral(currentNode.right, textBeforePlus + textAfterPlus, textAfterNode).slice(1);
                    }
                    /*
                     * Otherwise, these nodes should not be combined into a template curly, since there is nowhere to put
                     * the text between them.
                     */
                    return `${getTemplateLiteral(currentNode.left, textBeforeNode, null)}${textBeforePlus}+${textAfterPlus}${getTemplateLiteral(currentNode.right, textAfterNode, null)}`;
                }
                return `\`\${${textBeforeNode || ""}${sourceCode.getText(currentNode)}${textAfterNode || ""}}\``;
            }
            /**
             * Returns a fixer object that converts a non-string binary expression to a template literal
             * @param {SourceCodeFixer} fixer The fixer object
             * @param {ASTNode} node A node that should be converted to a template literal
             * @returns {Object} A fix for this binary expression
             */
            function fixNonStringBinaryExpression(fixer, node) {
                const topBinaryExpr = getTopConcatBinaryExpression(node.parent);
                if (hasOctalOrNonOctalDecimalEscapeSequence(topBinaryExpr)) {
                    return null;
                }
                return fixer.replaceText(topBinaryExpr, getTemplateLiteral(topBinaryExpr, null, null));
            }
            /**
             * Reports if a given node is string concatenation with non string literals.
             * @param {ASTNode} node A node to check.
             * @returns {void}
             */
            function checkForStringConcat(node) {
                if (!astUtils.isStringLiteral(node) || !isConcatenation(node.parent)) {
                    return;
                }
                const topBinaryExpr = getTopConcatBinaryExpression(node.parent);
                // Checks whether or not this node had been checked already.
                if (done[topBinaryExpr.range[0]]) {
                    return;
                }
                done[topBinaryExpr.range[0]] = true;
                if (hasNonStringLiteral(topBinaryExpr)) {
                    context.report({
                        node: topBinaryExpr,
                        messageId: "unexpectedStringConcatenation",
                        fix: fixer => fixNonStringBinaryExpression(fixer, node)
                    });
                }
            }
            return {
                Program() {
                    done = Object.create(null);
                },
                Literal: checkForStringConcat,
                TemplateLiteral: checkForStringConcat
            };
        }
    };
},
/* eslint/lib/rules/quote-props.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const espree = require(313) /* espree */;
    const astUtils = require(394) /* ./utils/ast-utils */;
    const keywords = require(429) /* ./utils/keywords */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require quotes around object literal property names",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/quote-props"
            },
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always", "as-needed", "consistent", "consistent-as-needed"]
                            }
                        ],
                        minItems: 0,
                        maxItems: 1
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always", "as-needed", "consistent", "consistent-as-needed"]
                            },
                            {
                                type: "object",
                                properties: {
                                    keywords: {
                                        type: "boolean"
                                    },
                                    unnecessary: {
                                        type: "boolean"
                                    },
                                    numbers: {
                                        type: "boolean"
                                    }
                                },
                                additionalProperties: false
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    }
                ]
            },
            fixable: "code",
            messages: {
                requireQuotesDueToReservedWord: "Properties should be quoted as '{{property}}' is a reserved word.",
                inconsistentlyQuotedProperty: "Inconsistently quoted property '{{key}}' found.",
                unnecessarilyQuotedProperty: "Unnecessarily quoted property '{{property}}' found.",
                unquotedReservedProperty: "Unquoted reserved word '{{property}}' used as key.",
                unquotedNumericProperty: "Unquoted number literal '{{property}}' used as key.",
                unquotedPropertyFound: "Unquoted property '{{property}}' found.",
                redundantQuoting: "Properties shouldn't be quoted as all quotes are redundant."
            }
        },
        create(context) {
            const MODE = context.options[0], KEYWORDS = context.options[1] && context.options[1].keywords, CHECK_UNNECESSARY = !context.options[1] || context.options[1].unnecessary !== false, NUMBERS = context.options[1] && context.options[1].numbers, sourceCode = context.getSourceCode();
            /**
             * Checks whether a certain string constitutes an ES3 token
             * @param   {string} tokenStr The string to be checked.
             * @returns {boolean} `true` if it is an ES3 token.
             */
            function isKeyword(tokenStr) {
                return keywords.indexOf(tokenStr) >= 0;
            }
            /**
             * Checks if an espree-tokenized key has redundant quotes (i.e. whether quotes are unnecessary)
             * @param   {string} rawKey The raw key value from the source
             * @param   {espreeTokens} tokens The espree-tokenized node key
             * @param   {boolean} [skipNumberLiterals=false] Indicates whether number literals should be checked
             * @returns {boolean} Whether or not a key has redundant quotes.
             * @private
             */
            function areQuotesRedundant(rawKey, tokens, skipNumberLiterals) {
                return tokens.length === 1 && tokens[0].start === 0 && tokens[0].end === rawKey.length &&
                    (["Identifier", "Keyword", "Null", "Boolean"].indexOf(tokens[0].type) >= 0 ||
                        (tokens[0].type === "Numeric" && !skipNumberLiterals && String(+tokens[0].value) === tokens[0].value));
            }
            /**
             * Returns a string representation of a property node with quotes removed
             * @param {ASTNode} key Key AST Node, which may or may not be quoted
             * @returns {string} A replacement string for this property
             */
            function getUnquotedKey(key) {
                return key.type === "Identifier" ? key.name : key.value;
            }
            /**
             * Returns a string representation of a property node with quotes added
             * @param {ASTNode} key Key AST Node, which may or may not be quoted
             * @returns {string} A replacement string for this property
             */
            function getQuotedKey(key) {
                if (key.type === "Literal" && typeof key.value === "string") {
                    // If the key is already a string literal, don't replace the quotes with double quotes.
                    return sourceCode.getText(key);
                }
                // Otherwise, the key is either an identifier or a number literal.
                return `"${key.type === "Identifier" ? key.name : key.value}"`;
            }
            /**
             * Ensures that a property's key is quoted only when necessary
             * @param   {ASTNode} node Property AST node
             * @returns {void}
             */
            function checkUnnecessaryQuotes(node) {
                const key = node.key;
                if (node.method || node.computed || node.shorthand) {
                    return;
                }
                if (key.type === "Literal" && typeof key.value === "string") {
                    let tokens;
                    try {
                        tokens = espree.tokenize(key.value);
                    }
                    catch {
                        return;
                    }
                    if (tokens.length !== 1) {
                        return;
                    }
                    const isKeywordToken = isKeyword(tokens[0].value);
                    if (isKeywordToken && KEYWORDS) {
                        return;
                    }
                    if (CHECK_UNNECESSARY && areQuotesRedundant(key.value, tokens, NUMBERS)) {
                        context.report({
                            node,
                            messageId: "unnecessarilyQuotedProperty",
                            data: { property: key.value },
                            fix: fixer => fixer.replaceText(key, getUnquotedKey(key))
                        });
                    }
                }
                else if (KEYWORDS && key.type === "Identifier" && isKeyword(key.name)) {
                    context.report({
                        node,
                        messageId: "unquotedReservedProperty",
                        data: { property: key.name },
                        fix: fixer => fixer.replaceText(key, getQuotedKey(key))
                    });
                }
                else if (NUMBERS && key.type === "Literal" && astUtils.isNumericLiteral(key)) {
                    context.report({
                        node,
                        messageId: "unquotedNumericProperty",
                        data: { property: key.value },
                        fix: fixer => fixer.replaceText(key, getQuotedKey(key))
                    });
                }
            }
            /**
             * Ensures that a property's key is quoted
             * @param   {ASTNode} node Property AST node
             * @returns {void}
             */
            function checkOmittedQuotes(node) {
                const key = node.key;
                if (!node.method && !node.computed && !node.shorthand && !(key.type === "Literal" && typeof key.value === "string")) {
                    context.report({
                        node,
                        messageId: "unquotedPropertyFound",
                        data: { property: key.name || key.value },
                        fix: fixer => fixer.replaceText(key, getQuotedKey(key))
                    });
                }
            }
            /**
             * Ensures that an object's keys are consistently quoted, optionally checks for redundancy of quotes
             * @param   {ASTNode} node Property AST node
             * @param   {boolean} checkQuotesRedundancy Whether to check quotes' redundancy
             * @returns {void}
             */
            function checkConsistency(node, checkQuotesRedundancy) {
                const quotedProps = [], unquotedProps = [];
                let keywordKeyName = null, necessaryQuotes = false;
                node.properties.forEach(property => {
                    const key = property.key;
                    if (!key || property.method || property.computed || property.shorthand) {
                        return;
                    }
                    if (key.type === "Literal" && typeof key.value === "string") {
                        quotedProps.push(property);
                        if (checkQuotesRedundancy) {
                            let tokens;
                            try {
                                tokens = espree.tokenize(key.value);
                            }
                            catch {
                                necessaryQuotes = true;
                                return;
                            }
                            necessaryQuotes = necessaryQuotes || !areQuotesRedundant(key.value, tokens) || KEYWORDS && isKeyword(tokens[0].value);
                        }
                    }
                    else if (KEYWORDS && checkQuotesRedundancy && key.type === "Identifier" && isKeyword(key.name)) {
                        unquotedProps.push(property);
                        necessaryQuotes = true;
                        keywordKeyName = key.name;
                    }
                    else {
                        unquotedProps.push(property);
                    }
                });
                if (checkQuotesRedundancy && quotedProps.length && !necessaryQuotes) {
                    quotedProps.forEach(property => {
                        context.report({
                            node: property,
                            messageId: "redundantQuoting",
                            fix: fixer => fixer.replaceText(property.key, getUnquotedKey(property.key))
                        });
                    });
                }
                else if (unquotedProps.length && keywordKeyName) {
                    unquotedProps.forEach(property => {
                        context.report({
                            node: property,
                            messageId: "requireQuotesDueToReservedWord",
                            data: { property: keywordKeyName },
                            fix: fixer => fixer.replaceText(property.key, getQuotedKey(property.key))
                        });
                    });
                }
                else if (quotedProps.length && unquotedProps.length) {
                    unquotedProps.forEach(property => {
                        context.report({
                            node: property,
                            messageId: "inconsistentlyQuotedProperty",
                            data: { key: property.key.name || property.key.value },
                            fix: fixer => fixer.replaceText(property.key, getQuotedKey(property.key))
                        });
                    });
                }
            }
            return {
                Property(node) {
                    if (MODE === "always" || !MODE) {
                        checkOmittedQuotes(node);
                    }
                    if (MODE === "as-needed") {
                        checkUnnecessaryQuotes(node);
                    }
                },
                ObjectExpression(node) {
                    if (MODE === "consistent") {
                        checkConsistency(node, false);
                    }
                    if (MODE === "consistent-as-needed") {
                        checkConsistency(node, true);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/quotes.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Constants
    //------------------------------------------------------------------------------
    const QUOTE_SETTINGS = {
        double: {
            quote: "\"",
            alternateQuote: "'",
            description: "doublequote"
        },
        single: {
            quote: "'",
            alternateQuote: "\"",
            description: "singlequote"
        },
        backtick: {
            quote: "`",
            alternateQuote: "\"",
            description: "backtick"
        }
    };
    // An unescaped newline is a newline preceded by an even number of backslashes.
    const UNESCAPED_LINEBREAK_PATTERN = new RegExp(String.raw `(^|[^\\])(\\\\)*[${Array.from(astUtils.LINEBREAKS).join("")}]`, "u");
    /**
     * Switches quoting of javascript string between ' " and `
     * escaping and unescaping as necessary.
     * Only escaping of the minimal set of characters is changed.
     * Note: escaping of newlines when switching from backtick to other quotes is not handled.
     * @param {string} str A string to convert.
     * @returns {string} The string with changed quotes.
     * @private
     */
    QUOTE_SETTINGS.double.convert =
        QUOTE_SETTINGS.single.convert =
            QUOTE_SETTINGS.backtick.convert = function (str) {
                const newQuote = this.quote;
                const oldQuote = str[0];
                if (newQuote === oldQuote) {
                    return str;
                }
                return newQuote + str.slice(1, -1).replace(/\\(\$\{|\r\n?|\n|.)|["'`]|\$\{|(\r\n?|\n)/gu, (match, escaped, newline) => {
                    if (escaped === oldQuote || oldQuote === "`" && escaped === "${") {
                        return escaped; // unescape
                    }
                    if (match === newQuote || newQuote === "`" && match === "${") {
                        return `\\${match}`; // escape
                    }
                    if (newline && oldQuote === "`") {
                        return "\\n"; // escape newlines
                    }
                    return match;
                }) + newQuote;
            };
    const AVOID_ESCAPE = "avoid-escape";
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce the consistent use of either backticks, double, or single quotes",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/quotes"
            },
            fixable: "code",
            schema: [
                {
                    enum: ["single", "double", "backtick"]
                },
                {
                    anyOf: [
                        {
                            enum: ["avoid-escape"]
                        },
                        {
                            type: "object",
                            properties: {
                                avoidEscape: {
                                    type: "boolean"
                                },
                                allowTemplateLiterals: {
                                    type: "boolean"
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                wrongQuotes: "Strings must use {{description}}."
            }
        },
        create(context) {
            const quoteOption = context.options[0], settings = QUOTE_SETTINGS[quoteOption || "double"], options = context.options[1], allowTemplateLiterals = options && options.allowTemplateLiterals === true, sourceCode = context.getSourceCode();
            let avoidEscape = options && options.avoidEscape === true;
            // deprecated
            if (options === AVOID_ESCAPE) {
                avoidEscape = true;
            }
            /**
             * Determines if a given node is part of JSX syntax.
             *
             * This function returns `true` in the following cases:
             *
             * - `<div className="foo"></div>` ... If the literal is an attribute value, the parent of the literal is `JSXAttribute`.
             * - `<div>foo</div>` ... If the literal is a text content, the parent of the literal is `JSXElement`.
             * - `<>foo</>` ... If the literal is a text content, the parent of the literal is `JSXFragment`.
             *
             * In particular, this function returns `false` in the following cases:
             *
             * - `<div className={"foo"}></div>`
             * - `<div>{"foo"}</div>`
             *
             * In both cases, inside of the braces is handled as normal JavaScript.
             * The braces are `JSXExpressionContainer` nodes.
             * @param {ASTNode} node The Literal node to check.
             * @returns {boolean} True if the node is a part of JSX, false if not.
             * @private
             */
            function isJSXLiteral(node) {
                return node.parent.type === "JSXAttribute" || node.parent.type === "JSXElement" || node.parent.type === "JSXFragment";
            }
            /**
             * Checks whether or not a given node is a directive.
             * The directive is a `ExpressionStatement` which has only a string literal.
             * @param {ASTNode} node A node to check.
             * @returns {boolean} Whether or not the node is a directive.
             * @private
             */
            function isDirective(node) {
                return (node.type === "ExpressionStatement" &&
                    node.expression.type === "Literal" &&
                    typeof node.expression.value === "string");
            }
            /**
             * Checks whether or not a given node is a part of directive prologues.
             * See also: http://www.ecma-international.org/ecma-262/6.0/#sec-directive-prologues-and-the-use-strict-directive
             * @param {ASTNode} node A node to check.
             * @returns {boolean} Whether or not the node is a part of directive prologues.
             * @private
             */
            function isPartOfDirectivePrologue(node) {
                const block = node.parent.parent;
                if (block.type !== "Program" && (block.type !== "BlockStatement" || !astUtils.isFunction(block.parent))) {
                    return false;
                }
                // Check the node is at a prologue.
                for (let i = 0; i < block.body.length; ++i) {
                    const statement = block.body[i];
                    if (statement === node.parent) {
                        return true;
                    }
                    if (!isDirective(statement)) {
                        break;
                    }
                }
                return false;
            }
            /**
             * Checks whether or not a given node is allowed as non backtick.
             * @param {ASTNode} node A node to check.
             * @returns {boolean} Whether or not the node is allowed as non backtick.
             * @private
             */
            function isAllowedAsNonBacktick(node) {
                const parent = node.parent;
                switch (parent.type) {
                    // Directive Prologues.
                    case "ExpressionStatement":
                        return isPartOfDirectivePrologue(node);
                    // LiteralPropertyName.
                    case "Property":
                    case "MethodDefinition":
                        return parent.key === node && !parent.computed;
                    // ModuleSpecifier.
                    case "ImportDeclaration":
                    case "ExportNamedDeclaration":
                    case "ExportAllDeclaration":
                        return parent.source === node;
                    // Others don't allow.
                    default:
                        return false;
                }
            }
            /**
             * Checks whether or not a given TemplateLiteral node is actually using any of the special features provided by template literal strings.
             * @param {ASTNode} node A TemplateLiteral node to check.
             * @returns {boolean} Whether or not the TemplateLiteral node is using any of the special features provided by template literal strings.
             * @private
             */
            function isUsingFeatureOfTemplateLiteral(node) {
                const hasTag = node.parent.type === "TaggedTemplateExpression" && node === node.parent.quasi;
                if (hasTag) {
                    return true;
                }
                const hasStringInterpolation = node.expressions.length > 0;
                if (hasStringInterpolation) {
                    return true;
                }
                const isMultilineString = node.quasis.length >= 1 && UNESCAPED_LINEBREAK_PATTERN.test(node.quasis[0].value.raw);
                if (isMultilineString) {
                    return true;
                }
                return false;
            }
            return {
                Literal(node) {
                    const val = node.value, rawVal = node.raw;
                    if (settings && typeof val === "string") {
                        let isValid = (quoteOption === "backtick" && isAllowedAsNonBacktick(node)) ||
                            isJSXLiteral(node) ||
                            astUtils.isSurroundedBy(rawVal, settings.quote);
                        if (!isValid && avoidEscape) {
                            isValid = astUtils.isSurroundedBy(rawVal, settings.alternateQuote) && rawVal.indexOf(settings.quote) >= 0;
                        }
                        if (!isValid) {
                            context.report({
                                node,
                                messageId: "wrongQuotes",
                                data: {
                                    description: settings.description
                                },
                                fix(fixer) {
                                    if (quoteOption === "backtick" && astUtils.hasOctalOrNonOctalDecimalEscapeSequence(rawVal)) {
                                        /*
                                         * An octal or non-octal decimal escape sequence in a template literal would
                                         * produce syntax error, even in non-strict mode.
                                         */
                                        return null;
                                    }
                                    return fixer.replaceText(node, settings.convert(node.raw));
                                }
                            });
                        }
                    }
                },
                TemplateLiteral(node) {
                    // Don't throw an error if backticks are expected or a template literal feature is in use.
                    if (allowTemplateLiterals ||
                        quoteOption === "backtick" ||
                        isUsingFeatureOfTemplateLiteral(node)) {
                        return;
                    }
                    context.report({
                        node,
                        messageId: "wrongQuotes",
                        data: {
                            description: settings.description
                        },
                        fix(fixer) {
                            if (isPartOfDirectivePrologue(node)) {
                                /*
                                 * TemplateLiterals in a directive prologue aren't actually directives, but if they're
                                 * in the directive prologue, then fixing them might turn them into directives and change
                                 * the behavior of the code.
                                 */
                                return null;
                            }
                            return fixer.replaceText(node, settings.convert(sourceCode.getText(node)));
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/radix.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const MODE_ALWAYS = "always", MODE_AS_NEEDED = "as-needed";
    const validRadixValues = new Set(Array.from({ length: 37 - 2 }, (_, index) => index + 2));
    /**
     * Checks whether a given variable is shadowed or not.
     * @param {eslint-scope.Variable} variable A variable to check.
     * @returns {boolean} `true` if the variable is shadowed.
     */
    function isShadowed(variable) {
        return variable.defs.length >= 1;
    }
    /**
     * Checks whether a given node is a MemberExpression of `parseInt` method or not.
     * @param {ASTNode} node A node to check.
     * @returns {boolean} `true` if the node is a MemberExpression of `parseInt`
     *      method.
     */
    function isParseIntMethod(node) {
        return (node.type === "MemberExpression" &&
            !node.computed &&
            node.property.type === "Identifier" &&
            node.property.name === "parseInt");
    }
    /**
     * Checks whether a given node is a valid value of radix or not.
     *
     * The following values are invalid.
     *
     * - A literal except integers between 2 and 36.
     * - undefined.
     * @param {ASTNode} radix A node of radix to check.
     * @returns {boolean} `true` if the node is valid.
     */
    function isValidRadix(radix) {
        return !((radix.type === "Literal" && !validRadixValues.has(radix.value)) ||
            (radix.type === "Identifier" && radix.name === "undefined"));
    }
    /**
     * Checks whether a given node is a default value of radix or not.
     * @param {ASTNode} radix A node of radix to check.
     * @returns {boolean} `true` if the node is the literal node of `10`.
     */
    function isDefaultRadix(radix) {
        return radix.type === "Literal" && radix.value === 10;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce the consistent use of the radix argument when using `parseInt()`",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/radix",
                suggestion: true
            },
            schema: [
                {
                    enum: ["always", "as-needed"]
                }
            ],
            messages: {
                missingParameters: "Missing parameters.",
                redundantRadix: "Redundant radix parameter.",
                missingRadix: "Missing radix parameter.",
                invalidRadix: "Invalid radix parameter, must be an integer between 2 and 36.",
                addRadixParameter10: "Add radix parameter `10` for parsing decimal numbers."
            }
        },
        create(context) {
            const mode = context.options[0] || MODE_ALWAYS;
            /**
             * Checks the arguments of a given CallExpression node and reports it if it
             * offends this rule.
             * @param {ASTNode} node A CallExpression node to check.
             * @returns {void}
             */
            function checkArguments(node) {
                const args = node.arguments;
                switch (args.length) {
                    case 0:
                        context.report({
                            node,
                            messageId: "missingParameters"
                        });
                        break;
                    case 1:
                        if (mode === MODE_ALWAYS) {
                            context.report({
                                node,
                                messageId: "missingRadix",
                                suggest: [
                                    {
                                        messageId: "addRadixParameter10",
                                        fix(fixer) {
                                            const sourceCode = context.getSourceCode();
                                            const tokens = sourceCode.getTokens(node);
                                            const lastToken = tokens[tokens.length - 1]; // Parenthesis.
                                            const secondToLastToken = tokens[tokens.length - 2]; // May or may not be a comma.
                                            const hasTrailingComma = secondToLastToken.type === "Punctuator" && secondToLastToken.value === ",";
                                            return fixer.insertTextBefore(lastToken, hasTrailingComma ? " 10," : ", 10");
                                        }
                                    }
                                ]
                            });
                        }
                        break;
                    default:
                        if (mode === MODE_AS_NEEDED && isDefaultRadix(args[1])) {
                            context.report({
                                node,
                                messageId: "redundantRadix"
                            });
                        }
                        else if (!isValidRadix(args[1])) {
                            context.report({
                                node,
                                messageId: "invalidRadix"
                            });
                        }
                        break;
                }
            }
            return {
                "Program:exit"() {
                    const scope = context.getScope();
                    let variable;
                    // Check `parseInt()`
                    variable = astUtils.getVariableByName(scope, "parseInt");
                    if (variable && !isShadowed(variable)) {
                        variable.references.forEach(reference => {
                            const node = reference.identifier;
                            if (astUtils.isCallee(node)) {
                                checkArguments(node.parent);
                            }
                        });
                    }
                    // Check `Number.parseInt()`
                    variable = astUtils.getVariableByName(scope, "Number");
                    if (variable && !isShadowed(variable)) {
                        variable.references.forEach(reference => {
                            const node = reference.identifier.parent;
                            const maybeCallee = node.parent.type === "ChainExpression"
                                ? node.parent
                                : node;
                            if (isParseIntMethod(node) && astUtils.isCallee(maybeCallee)) {
                                checkArguments(maybeCallee.parent);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/require-atomic-updates.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Make the map from identifiers to each reference.
     * @param {escope.Scope} scope The scope to get references.
     * @param {Map<Identifier, escope.Reference>} [outReferenceMap] The map from identifier nodes to each reference object.
     * @returns {Map<Identifier, escope.Reference>} `referenceMap`.
     */
    function createReferenceMap(scope, outReferenceMap = new Map()) {
        for (const reference of scope.references) {
            if (reference.resolved === null) {
                continue;
            }
            outReferenceMap.set(reference.identifier, reference);
        }
        for (const childScope of scope.childScopes) {
            if (childScope.type !== "function") {
                createReferenceMap(childScope, outReferenceMap);
            }
        }
        return outReferenceMap;
    }
    /**
     * Get `reference.writeExpr` of a given reference.
     * If it's the read reference of MemberExpression in LHS, returns RHS in order to address `a.b = await a`
     * @param {escope.Reference} reference The reference to get.
     * @returns {Expression|null} The `reference.writeExpr`.
     */
    function getWriteExpr(reference) {
        if (reference.writeExpr) {
            return reference.writeExpr;
        }
        let node = reference.identifier;
        while (node) {
            const t = node.parent.type;
            if (t === "AssignmentExpression" && node.parent.left === node) {
                return node.parent.right;
            }
            if (t === "MemberExpression" && node.parent.object === node) {
                node = node.parent;
                continue;
            }
            break;
        }
        return null;
    }
    /**
     * Checks if an expression is a variable that can only be observed within the given function.
     * @param {Variable|null} variable The variable to check
     * @param {boolean} isMemberAccess If `true` then this is a member access.
     * @returns {boolean} `true` if the variable is local to the given function, and is never referenced in a closure.
     */
    function isLocalVariableWithoutEscape(variable, isMemberAccess) {
        if (!variable) {
            return false; // A global variable which was not defined.
        }
        // If the reference is a property access and the variable is a parameter, it handles the variable is not local.
        if (isMemberAccess && variable.defs.some(d => d.type === "Parameter")) {
            return false;
        }
        const functionScope = variable.scope.variableScope;
        return variable.references.every(reference => reference.from.variableScope === functionScope);
    }
    class SegmentInfo {
        constructor() {
            this.info = new WeakMap();
        }
        /**
         * Initialize the segment information.
         * @param {PathSegment} segment The segment to initialize.
         * @returns {void}
         */
        initialize(segment) {
            const outdatedReadVariables = new Set();
            const freshReadVariables = new Set();
            for (const prevSegment of segment.prevSegments) {
                const info = this.info.get(prevSegment);
                if (info) {
                    info.outdatedReadVariables.forEach(Set.prototype.add, outdatedReadVariables);
                    info.freshReadVariables.forEach(Set.prototype.add, freshReadVariables);
                }
            }
            this.info.set(segment, { outdatedReadVariables, freshReadVariables });
        }
        /**
         * Mark a given variable as read on given segments.
         * @param {PathSegment[]} segments The segments that it read the variable on.
         * @param {Variable} variable The variable to be read.
         * @returns {void}
         */
        markAsRead(segments, variable) {
            for (const segment of segments) {
                const info = this.info.get(segment);
                if (info) {
                    info.freshReadVariables.add(variable);
                    // If a variable is freshly read again, then it's no more out-dated.
                    info.outdatedReadVariables.delete(variable);
                }
            }
        }
        /**
         * Move `freshReadVariables` to `outdatedReadVariables`.
         * @param {PathSegment[]} segments The segments to process.
         * @returns {void}
         */
        makeOutdated(segments) {
            for (const segment of segments) {
                const info = this.info.get(segment);
                if (info) {
                    info.freshReadVariables.forEach(Set.prototype.add, info.outdatedReadVariables);
                    info.freshReadVariables.clear();
                }
            }
        }
        /**
         * Check if a given variable is outdated on the current segments.
         * @param {PathSegment[]} segments The current segments.
         * @param {Variable} variable The variable to check.
         * @returns {boolean} `true` if the variable is outdated on the segments.
         */
        isOutdated(segments, variable) {
            for (const segment of segments) {
                const info = this.info.get(segment);
                if (info && info.outdatedReadVariables.has(variable)) {
                    return true;
                }
            }
            return false;
        }
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "disallow assignments that can lead to race conditions due to usage of `await` or `yield`",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/require-atomic-updates"
            },
            fixable: null,
            schema: [],
            messages: {
                nonAtomicUpdate: "Possible race condition: `{{value}}` might be reassigned based on an outdated value of `{{value}}`."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const assignmentReferences = new Map();
            const segmentInfo = new SegmentInfo();
            let stack = null;
            return {
                onCodePathStart(codePath) {
                    const scope = context.getScope();
                    const shouldVerify = scope.type === "function" &&
                        (scope.block.async || scope.block.generator);
                    stack = {
                        upper: stack,
                        codePath,
                        referenceMap: shouldVerify ? createReferenceMap(scope) : null
                    };
                },
                onCodePathEnd() {
                    stack = stack.upper;
                },
                // Initialize the segment information.
                onCodePathSegmentStart(segment) {
                    segmentInfo.initialize(segment);
                },
                // Handle references to prepare verification.
                Identifier(node) {
                    const { codePath, referenceMap } = stack;
                    const reference = referenceMap && referenceMap.get(node);
                    // Ignore if this is not a valid variable reference.
                    if (!reference) {
                        return;
                    }
                    const variable = reference.resolved;
                    const writeExpr = getWriteExpr(reference);
                    const isMemberAccess = reference.identifier.parent.type === "MemberExpression";
                    // Add a fresh read variable.
                    if (reference.isRead() && !(writeExpr && writeExpr.parent.operator === "=")) {
                        segmentInfo.markAsRead(codePath.currentSegments, variable);
                    }
                    /*
                     * Register the variable to verify after ESLint traversed the `writeExpr` node
                     * if this reference is an assignment to a variable which is referred from other closure.
                     */
                    if (writeExpr &&
                        writeExpr.parent.right === writeExpr && // ← exclude variable declarations.
                        !isLocalVariableWithoutEscape(variable, isMemberAccess)) {
                        let refs = assignmentReferences.get(writeExpr);
                        if (!refs) {
                            refs = [];
                            assignmentReferences.set(writeExpr, refs);
                        }
                        refs.push(reference);
                    }
                },
                /*
                 * Verify assignments.
                 * If the reference exists in `outdatedReadVariables` list, report it.
                 */
                ":expression:exit"(node) {
                    const { codePath, referenceMap } = stack;
                    // referenceMap exists if this is in a resumable function scope.
                    if (!referenceMap) {
                        return;
                    }
                    // Mark the read variables on this code path as outdated.
                    if (node.type === "AwaitExpression" || node.type === "YieldExpression") {
                        segmentInfo.makeOutdated(codePath.currentSegments);
                    }
                    // Verify.
                    const references = assignmentReferences.get(node);
                    if (references) {
                        assignmentReferences.delete(node);
                        for (const reference of references) {
                            const variable = reference.resolved;
                            if (segmentInfo.isOutdated(codePath.currentSegments, variable)) {
                                context.report({
                                    node: node.parent,
                                    messageId: "nonAtomicUpdate",
                                    data: {
                                        value: sourceCode.getText(node.parent.left)
                                    }
                                });
                            }
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/require-await.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Capitalize the 1st letter of the given text.
     * @param {string} text The text to capitalize.
     * @returns {string} The text that the 1st letter was capitalized.
     */
    function capitalizeFirstLetter(text) {
        return text[0].toUpperCase() + text.slice(1);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "disallow async functions which have no `await` expression",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/require-await"
            },
            schema: [],
            messages: {
                missingAwait: "{{name}} has no 'await' expression."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            let scopeInfo = null;
            /**
             * Push the scope info object to the stack.
             * @returns {void}
             */
            function enterFunction() {
                scopeInfo = {
                    upper: scopeInfo,
                    hasAwait: false
                };
            }
            /**
             * Pop the top scope info object from the stack.
             * Also, it reports the function if needed.
             * @param {ASTNode} node The node to report.
             * @returns {void}
             */
            function exitFunction(node) {
                if (!node.generator && node.async && !scopeInfo.hasAwait && !astUtils.isEmptyFunction(node)) {
                    context.report({
                        node,
                        loc: astUtils.getFunctionHeadLoc(node, sourceCode),
                        messageId: "missingAwait",
                        data: {
                            name: capitalizeFirstLetter(astUtils.getFunctionNameWithKind(node))
                        }
                    });
                }
                scopeInfo = scopeInfo.upper;
            }
            return {
                FunctionDeclaration: enterFunction,
                FunctionExpression: enterFunction,
                ArrowFunctionExpression: enterFunction,
                "FunctionDeclaration:exit": exitFunction,
                "FunctionExpression:exit": exitFunction,
                "ArrowFunctionExpression:exit": exitFunction,
                AwaitExpression() {
                    if (!scopeInfo) {
                        return;
                    }
                    scopeInfo.hasAwait = true;
                },
                ForOfStatement(node) {
                    if (!scopeInfo) {
                        return;
                    }
                    if (node.await) {
                        scopeInfo.hasAwait = true;
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/require-jsdoc.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require JSDoc comments",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/require-jsdoc"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        require: {
                            type: "object",
                            properties: {
                                ClassDeclaration: {
                                    type: "boolean",
                                    default: false
                                },
                                MethodDefinition: {
                                    type: "boolean",
                                    default: false
                                },
                                FunctionDeclaration: {
                                    type: "boolean",
                                    default: true
                                },
                                ArrowFunctionExpression: {
                                    type: "boolean",
                                    default: false
                                },
                                FunctionExpression: {
                                    type: "boolean",
                                    default: false
                                }
                            },
                            additionalProperties: false,
                            default: {}
                        }
                    },
                    additionalProperties: false
                }
            ],
            deprecated: true,
            replacedBy: [],
            messages: {
                missingJSDocComment: "Missing JSDoc comment."
            }
        },
        create(context) {
            const source = context.getSourceCode();
            const DEFAULT_OPTIONS = {
                FunctionDeclaration: true,
                MethodDefinition: false,
                ClassDeclaration: false,
                ArrowFunctionExpression: false,
                FunctionExpression: false
            };
            const options = Object.assign(DEFAULT_OPTIONS, context.options[0] && context.options[0].require);
            /**
             * Report the error message
             * @param {ASTNode} node node to report
             * @returns {void}
             */
            function report(node) {
                context.report({ node, messageId: "missingJSDocComment" });
            }
            /**
             * Check if the jsdoc comment is present or not.
             * @param {ASTNode} node node to examine
             * @returns {void}
             */
            function checkJsDoc(node) {
                const jsdocComment = source.getJSDocComment(node);
                if (!jsdocComment) {
                    report(node);
                }
            }
            return {
                FunctionDeclaration(node) {
                    if (options.FunctionDeclaration) {
                        checkJsDoc(node);
                    }
                },
                FunctionExpression(node) {
                    if ((options.MethodDefinition && node.parent.type === "MethodDefinition") ||
                        (options.FunctionExpression && (node.parent.type === "VariableDeclarator" || (node.parent.type === "Property" && node === node.parent.value)))) {
                        checkJsDoc(node);
                    }
                },
                ClassDeclaration(node) {
                    if (options.ClassDeclaration) {
                        checkJsDoc(node);
                    }
                },
                ArrowFunctionExpression(node) {
                    if (options.ArrowFunctionExpression && node.parent.type === "VariableDeclarator") {
                        checkJsDoc(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/require-unicode-regexp.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const { CALL, CONSTRUCT, ReferenceTracker, getStringIfConstant } = require(350) /* eslint-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce the use of `u` flag on RegExp",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/require-unicode-regexp"
            },
            messages: {
                requireUFlag: "Use the 'u' flag."
            },
            schema: []
        },
        create(context) {
            return {
                "Literal[regex]"(node) {
                    const flags = node.regex.flags || "";
                    if (!flags.includes("u")) {
                        context.report({ node, messageId: "requireUFlag" });
                    }
                },
                Program() {
                    const scope = context.getScope();
                    const tracker = new ReferenceTracker(scope);
                    const trackMap = {
                        RegExp: { [CALL]: true, [CONSTRUCT]: true }
                    };
                    for (const { node } of tracker.iterateGlobalReferences(trackMap)) {
                        const flagsNode = node.arguments[1];
                        const flags = getStringIfConstant(flagsNode, scope);
                        if (!flagsNode || (typeof flags === "string" && !flags.includes("u"))) {
                            context.report({ node, messageId: "requireUFlag" });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/require-yield.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require generator functions to contain `yield`",
                category: "ECMAScript 6",
                recommended: true,
                url: "https://eslint.org/docs/rules/require-yield"
            },
            schema: [],
            messages: {
                missingYield: "This generator function does not have 'yield'."
            }
        },
        create(context) {
            const stack = [];
            /**
             * If the node is a generator function, start counting `yield` keywords.
             * @param {Node} node A function node to check.
             * @returns {void}
             */
            function beginChecking(node) {
                if (node.generator) {
                    stack.push(0);
                }
            }
            /**
             * If the node is a generator function, end counting `yield` keywords, then
             * reports result.
             * @param {Node} node A function node to check.
             * @returns {void}
             */
            function endChecking(node) {
                if (!node.generator) {
                    return;
                }
                const countYield = stack.pop();
                if (countYield === 0 && node.body.body.length > 0) {
                    context.report({ node, messageId: "missingYield" });
                }
            }
            return {
                FunctionDeclaration: beginChecking,
                "FunctionDeclaration:exit": endChecking,
                FunctionExpression: beginChecking,
                "FunctionExpression:exit": endChecking,
                // Increases the count of `yield` keyword.
                YieldExpression() {
                    /* istanbul ignore else */
                    if (stack.length > 0) {
                        stack[stack.length - 1] += 1;
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/rest-spread-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce spacing between rest and spread operators and their expressions",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/rest-spread-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never"]
                }
            ],
            messages: {
                unexpectedWhitespace: "Unexpected whitespace after {{type}} operator.",
                expectedWhitespace: "Expected whitespace after {{type}} operator."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode(), alwaysSpace = context.options[0] === "always";
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Checks whitespace between rest/spread operators and their expressions
             * @param {ASTNode} node The node to check
             * @returns {void}
             */
            function checkWhiteSpace(node) {
                const operator = sourceCode.getFirstToken(node), nextToken = sourceCode.getTokenAfter(operator), hasWhitespace = sourceCode.isSpaceBetweenTokens(operator, nextToken);
                let type;
                switch (node.type) {
                    case "SpreadElement":
                        type = "spread";
                        if (node.parent.type === "ObjectExpression") {
                            type += " property";
                        }
                        break;
                    case "RestElement":
                        type = "rest";
                        if (node.parent.type === "ObjectPattern") {
                            type += " property";
                        }
                        break;
                    case "ExperimentalSpreadProperty":
                        type = "spread property";
                        break;
                    case "ExperimentalRestProperty":
                        type = "rest property";
                        break;
                    default:
                        return;
                }
                if (alwaysSpace && !hasWhitespace) {
                    context.report({
                        node,
                        loc: operator.loc,
                        messageId: "expectedWhitespace",
                        data: {
                            type
                        },
                        fix(fixer) {
                            return fixer.replaceTextRange([operator.range[1], nextToken.range[0]], " ");
                        }
                    });
                }
                else if (!alwaysSpace && hasWhitespace) {
                    context.report({
                        node,
                        loc: {
                            start: operator.loc.end,
                            end: nextToken.loc.start
                        },
                        messageId: "unexpectedWhitespace",
                        data: {
                            type
                        },
                        fix(fixer) {
                            return fixer.removeRange([operator.range[1], nextToken.range[0]]);
                        }
                    });
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                SpreadElement: checkWhiteSpace,
                RestElement: checkWhiteSpace,
                ExperimentalSpreadProperty: checkWhiteSpace,
                ExperimentalRestProperty: checkWhiteSpace
            };
        }
    };
},
/* eslint/lib/rules/semi.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const FixTracker = require(508) /* ./utils/fix-tracker */;
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow semicolons instead of ASI",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/semi"
            },
            fixable: "code",
            schema: {
                anyOf: [
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["never"]
                            },
                            {
                                type: "object",
                                properties: {
                                    beforeStatementContinuationChars: {
                                        enum: ["always", "any", "never"]
                                    }
                                },
                                additionalProperties: false
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    },
                    {
                        type: "array",
                        items: [
                            {
                                enum: ["always"]
                            },
                            {
                                type: "object",
                                properties: {
                                    omitLastInOneLineBlock: { type: "boolean" }
                                },
                                additionalProperties: false
                            }
                        ],
                        minItems: 0,
                        maxItems: 2
                    }
                ]
            },
            messages: {
                missingSemi: "Missing semicolon.",
                extraSemi: "Extra semicolon."
            }
        },
        create(context) {
            const OPT_OUT_PATTERN = /^[-[(/+`]/u; // One of [(/+-`
            const options = context.options[1];
            const never = context.options[0] === "never";
            const exceptOneLine = Boolean(options && options.omitLastInOneLineBlock);
            const beforeStatementContinuationChars = options && options.beforeStatementContinuationChars || "any";
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Reports a semicolon error with appropriate location and message.
             * @param {ASTNode} node The node with an extra or missing semicolon.
             * @param {boolean} missing True if the semicolon is missing.
             * @returns {void}
             */
            function report(node, missing) {
                const lastToken = sourceCode.getLastToken(node);
                let messageId, fix, loc;
                if (!missing) {
                    messageId = "missingSemi";
                    loc = {
                        start: lastToken.loc.end,
                        end: astUtils.getNextLocation(sourceCode, lastToken.loc.end)
                    };
                    fix = function (fixer) {
                        return fixer.insertTextAfter(lastToken, ";");
                    };
                }
                else {
                    messageId = "extraSemi";
                    loc = lastToken.loc;
                    fix = function (fixer) {
                        /*
                         * Expand the replacement range to include the surrounding
                         * tokens to avoid conflicting with no-extra-semi.
                         * https://github.com/eslint/eslint/issues/7928
                         */
                        return new FixTracker(fixer, sourceCode)
                            .retainSurroundingTokens(lastToken)
                            .remove(lastToken);
                    };
                }
                context.report({
                    node,
                    loc,
                    messageId,
                    fix
                });
            }
            /**
             * Check whether a given semicolon token is redundant.
             * @param {Token} semiToken A semicolon token to check.
             * @returns {boolean} `true` if the next token is `;` or `}`.
             */
            function isRedundantSemi(semiToken) {
                const nextToken = sourceCode.getTokenAfter(semiToken);
                return (!nextToken ||
                    astUtils.isClosingBraceToken(nextToken) ||
                    astUtils.isSemicolonToken(nextToken));
            }
            /**
             * Check whether a given token is the closing brace of an arrow function.
             * @param {Token} lastToken A token to check.
             * @returns {boolean} `true` if the token is the closing brace of an arrow function.
             */
            function isEndOfArrowBlock(lastToken) {
                if (!astUtils.isClosingBraceToken(lastToken)) {
                    return false;
                }
                const node = sourceCode.getNodeByRangeIndex(lastToken.range[0]);
                return (node.type === "BlockStatement" &&
                    node.parent.type === "ArrowFunctionExpression");
            }
            /**
             * Check whether a given node is on the same line with the next token.
             * @param {Node} node A statement node to check.
             * @returns {boolean} `true` if the node is on the same line with the next token.
             */
            function isOnSameLineWithNextToken(node) {
                const prevToken = sourceCode.getLastToken(node, 1);
                const nextToken = sourceCode.getTokenAfter(node);
                return !!nextToken && astUtils.isTokenOnSameLine(prevToken, nextToken);
            }
            /**
             * Check whether a given node can connect the next line if the next line is unreliable.
             * @param {Node} node A statement node to check.
             * @returns {boolean} `true` if the node can connect the next line.
             */
            function maybeAsiHazardAfter(node) {
                const t = node.type;
                if (t === "DoWhileStatement" ||
                    t === "BreakStatement" ||
                    t === "ContinueStatement" ||
                    t === "DebuggerStatement" ||
                    t === "ImportDeclaration" ||
                    t === "ExportAllDeclaration") {
                    return false;
                }
                if (t === "ReturnStatement") {
                    return Boolean(node.argument);
                }
                if (t === "ExportNamedDeclaration") {
                    return Boolean(node.declaration);
                }
                if (isEndOfArrowBlock(sourceCode.getLastToken(node, 1))) {
                    return false;
                }
                return true;
            }
            /**
             * Check whether a given token can connect the previous statement.
             * @param {Token} token A token to check.
             * @returns {boolean} `true` if the token is one of `[`, `(`, `/`, `+`, `-`, ```, `++`, and `--`.
             */
            function maybeAsiHazardBefore(token) {
                return (Boolean(token) &&
                    OPT_OUT_PATTERN.test(token.value) &&
                    token.value !== "++" &&
                    token.value !== "--");
            }
            /**
             * Check if the semicolon of a given node is unnecessary, only true if:
             *   - next token is a valid statement divider (`;` or `}`).
             *   - next token is on a new line and the node is not connectable to the new line.
             * @param {Node} node A statement node to check.
             * @returns {boolean} whether the semicolon is unnecessary.
             */
            function canRemoveSemicolon(node) {
                if (isRedundantSemi(sourceCode.getLastToken(node))) {
                    return true; // `;;` or `;}`
                }
                if (isOnSameLineWithNextToken(node)) {
                    return false; // One liner.
                }
                if (beforeStatementContinuationChars === "never" && !maybeAsiHazardAfter(node)) {
                    return true; // ASI works. This statement doesn't connect to the next.
                }
                if (!maybeAsiHazardBefore(sourceCode.getTokenAfter(node))) {
                    return true; // ASI works. The next token doesn't connect to this statement.
                }
                return false;
            }
            /**
             * Checks a node to see if it's in a one-liner block statement.
             * @param {ASTNode} node The node to check.
             * @returns {boolean} whether the node is in a one-liner block statement.
             */
            function isOneLinerBlock(node) {
                const parent = node.parent;
                const nextToken = sourceCode.getTokenAfter(node);
                if (!nextToken || nextToken.value !== "}") {
                    return false;
                }
                return (!!parent &&
                    parent.type === "BlockStatement" &&
                    parent.loc.start.line === parent.loc.end.line);
            }
            /**
             * Checks a node to see if it's followed by a semicolon.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             */
            function checkForSemicolon(node) {
                const isSemi = astUtils.isSemicolonToken(sourceCode.getLastToken(node));
                if (never) {
                    if (isSemi && canRemoveSemicolon(node)) {
                        report(node, true);
                    }
                    else if (!isSemi && beforeStatementContinuationChars === "always" && maybeAsiHazardBefore(sourceCode.getTokenAfter(node))) {
                        report(node);
                    }
                }
                else {
                    const oneLinerBlock = (exceptOneLine && isOneLinerBlock(node));
                    if (isSemi && oneLinerBlock) {
                        report(node, true);
                    }
                    else if (!isSemi && !oneLinerBlock) {
                        report(node);
                    }
                }
            }
            /**
             * Checks to see if there's a semicolon after a variable declaration.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             */
            function checkForSemicolonForVariableDeclaration(node) {
                const parent = node.parent;
                if ((parent.type !== "ForStatement" || parent.init !== node) &&
                    (!/^For(?:In|Of)Statement/u.test(parent.type) || parent.left !== node)) {
                    checkForSemicolon(node);
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                VariableDeclaration: checkForSemicolonForVariableDeclaration,
                ExpressionStatement: checkForSemicolon,
                ReturnStatement: checkForSemicolon,
                ThrowStatement: checkForSemicolon,
                DoWhileStatement: checkForSemicolon,
                DebuggerStatement: checkForSemicolon,
                BreakStatement: checkForSemicolon,
                ContinueStatement: checkForSemicolon,
                ImportDeclaration: checkForSemicolon,
                ExportAllDeclaration: checkForSemicolon,
                ExportNamedDeclaration(node) {
                    if (!node.declaration) {
                        checkForSemicolon(node);
                    }
                },
                ExportDefaultDeclaration(node) {
                    if (!/(?:Class|Function)Declaration/u.test(node.declaration.type)) {
                        checkForSemicolon(node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/semi-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing before and after semicolons",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/semi-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        before: {
                            type: "boolean",
                            default: false
                        },
                        after: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedWhitespaceBefore: "Unexpected whitespace before semicolon.",
                unexpectedWhitespaceAfter: "Unexpected whitespace after semicolon.",
                missingWhitespaceBefore: "Missing whitespace before semicolon.",
                missingWhitespaceAfter: "Missing whitespace after semicolon."
            }
        },
        create(context) {
            const config = context.options[0], sourceCode = context.getSourceCode();
            let requireSpaceBefore = false, requireSpaceAfter = true;
            if (typeof config === "object") {
                requireSpaceBefore = config.before;
                requireSpaceAfter = config.after;
            }
            /**
             * Checks if a given token has leading whitespace.
             * @param {Object} token The token to check.
             * @returns {boolean} True if the given token has leading space, false if not.
             */
            function hasLeadingSpace(token) {
                const tokenBefore = sourceCode.getTokenBefore(token);
                return tokenBefore && astUtils.isTokenOnSameLine(tokenBefore, token) && sourceCode.isSpaceBetweenTokens(tokenBefore, token);
            }
            /**
             * Checks if a given token has trailing whitespace.
             * @param {Object} token The token to check.
             * @returns {boolean} True if the given token has trailing space, false if not.
             */
            function hasTrailingSpace(token) {
                const tokenAfter = sourceCode.getTokenAfter(token);
                return tokenAfter && astUtils.isTokenOnSameLine(token, tokenAfter) && sourceCode.isSpaceBetweenTokens(token, tokenAfter);
            }
            /**
             * Checks if the given token is the last token in its line.
             * @param {Token} token The token to check.
             * @returns {boolean} Whether or not the token is the last in its line.
             */
            function isLastTokenInCurrentLine(token) {
                const tokenAfter = sourceCode.getTokenAfter(token);
                return !(tokenAfter && astUtils.isTokenOnSameLine(token, tokenAfter));
            }
            /**
             * Checks if the given token is the first token in its line
             * @param {Token} token The token to check.
             * @returns {boolean} Whether or not the token is the first in its line.
             */
            function isFirstTokenInCurrentLine(token) {
                const tokenBefore = sourceCode.getTokenBefore(token);
                return !(tokenBefore && astUtils.isTokenOnSameLine(token, tokenBefore));
            }
            /**
             * Checks if the next token of a given token is a closing parenthesis.
             * @param {Token} token The token to check.
             * @returns {boolean} Whether or not the next token of a given token is a closing parenthesis.
             */
            function isBeforeClosingParen(token) {
                const nextToken = sourceCode.getTokenAfter(token);
                return (nextToken && astUtils.isClosingBraceToken(nextToken) || astUtils.isClosingParenToken(nextToken));
            }
            /**
             * Report location example :
             *
             * for unexpected space `before`
             *
             * var a = 'b'   ;
             *            ^^^
             *
             * for unexpected space `after`
             *
             * var a = 'b';  c = 10;
             *             ^^
             *
             * Reports if the given token has invalid spacing.
             * @param {Token} token The semicolon token to check.
             * @param {ASTNode} node The corresponding node of the token.
             * @returns {void}
             */
            function checkSemicolonSpacing(token, node) {
                if (astUtils.isSemicolonToken(token)) {
                    if (hasLeadingSpace(token)) {
                        if (!requireSpaceBefore) {
                            const tokenBefore = sourceCode.getTokenBefore(token);
                            const loc = {
                                start: tokenBefore.loc.end,
                                end: token.loc.start
                            };
                            context.report({
                                node,
                                loc,
                                messageId: "unexpectedWhitespaceBefore",
                                fix(fixer) {
                                    return fixer.removeRange([tokenBefore.range[1], token.range[0]]);
                                }
                            });
                        }
                    }
                    else {
                        if (requireSpaceBefore) {
                            const loc = token.loc;
                            context.report({
                                node,
                                loc,
                                messageId: "missingWhitespaceBefore",
                                fix(fixer) {
                                    return fixer.insertTextBefore(token, " ");
                                }
                            });
                        }
                    }
                    if (!isFirstTokenInCurrentLine(token) && !isLastTokenInCurrentLine(token) && !isBeforeClosingParen(token)) {
                        if (hasTrailingSpace(token)) {
                            if (!requireSpaceAfter) {
                                const tokenAfter = sourceCode.getTokenAfter(token);
                                const loc = {
                                    start: token.loc.end,
                                    end: tokenAfter.loc.start
                                };
                                context.report({
                                    node,
                                    loc,
                                    messageId: "unexpectedWhitespaceAfter",
                                    fix(fixer) {
                                        return fixer.removeRange([token.range[1], tokenAfter.range[0]]);
                                    }
                                });
                            }
                        }
                        else {
                            if (requireSpaceAfter) {
                                const loc = token.loc;
                                context.report({
                                    node,
                                    loc,
                                    messageId: "missingWhitespaceAfter",
                                    fix(fixer) {
                                        return fixer.insertTextAfter(token, " ");
                                    }
                                });
                            }
                        }
                    }
                }
            }
            /**
             * Checks the spacing of the semicolon with the assumption that the last token is the semicolon.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             */
            function checkNode(node) {
                const token = sourceCode.getLastToken(node);
                checkSemicolonSpacing(token, node);
            }
            return {
                VariableDeclaration: checkNode,
                ExpressionStatement: checkNode,
                BreakStatement: checkNode,
                ContinueStatement: checkNode,
                DebuggerStatement: checkNode,
                DoWhileStatement: checkNode,
                ReturnStatement: checkNode,
                ThrowStatement: checkNode,
                ImportDeclaration: checkNode,
                ExportNamedDeclaration: checkNode,
                ExportAllDeclaration: checkNode,
                ExportDefaultDeclaration: checkNode,
                ForStatement(node) {
                    if (node.init) {
                        checkSemicolonSpacing(sourceCode.getTokenAfter(node.init), node);
                    }
                    if (node.test) {
                        checkSemicolonSpacing(sourceCode.getTokenAfter(node.test), node);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/semi-style.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    const SELECTOR = `:matches(${[
        "BreakStatement", "ContinueStatement", "DebuggerStatement",
        "DoWhileStatement", "ExportAllDeclaration",
        "ExportDefaultDeclaration", "ExportNamedDeclaration",
        "ExpressionStatement", "ImportDeclaration", "ReturnStatement",
        "ThrowStatement", "VariableDeclaration"
    ].join(",")})`;
    /**
     * Get the child node list of a given node.
     * This returns `Program#body`, `BlockStatement#body`, or `SwitchCase#consequent`.
     * This is used to check whether a node is the first/last child.
     * @param {Node} node A node to get child node list.
     * @returns {Node[]|null} The child node list.
     */
    function getChildren(node) {
        const t = node.type;
        if (t === "BlockStatement" || t === "Program") {
            return node.body;
        }
        if (t === "SwitchCase") {
            return node.consequent;
        }
        return null;
    }
    /**
     * Check whether a given node is the last statement in the parent block.
     * @param {Node} node A node to check.
     * @returns {boolean} `true` if the node is the last statement in the parent block.
     */
    function isLastChild(node) {
        const t = node.parent.type;
        if (t === "IfStatement" && node.parent.consequent === node && node.parent.alternate) { // before `else` keyword.
            return true;
        }
        if (t === "DoWhileStatement") { // before `while` keyword.
            return true;
        }
        const nodeList = getChildren(node.parent);
        return nodeList !== null && nodeList[nodeList.length - 1] === node; // before `}` or etc.
    }
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce location of semicolons",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/semi-style"
            },
            schema: [{ enum: ["last", "first"] }],
            fixable: "whitespace",
            messages: {
                expectedSemiColon: "Expected this semicolon to be at {{pos}}."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const option = context.options[0] || "last";
            /**
             * Check the given semicolon token.
             * @param {Token} semiToken The semicolon token to check.
             * @param {"first"|"last"} expected The expected location to check.
             * @returns {void}
             */
            function check(semiToken, expected) {
                const prevToken = sourceCode.getTokenBefore(semiToken);
                const nextToken = sourceCode.getTokenAfter(semiToken);
                const prevIsSameLine = !prevToken || astUtils.isTokenOnSameLine(prevToken, semiToken);
                const nextIsSameLine = !nextToken || astUtils.isTokenOnSameLine(semiToken, nextToken);
                if ((expected === "last" && !prevIsSameLine) || (expected === "first" && !nextIsSameLine)) {
                    context.report({
                        loc: semiToken.loc,
                        messageId: "expectedSemiColon",
                        data: {
                            pos: (expected === "last")
                                ? "the end of the previous line"
                                : "the beginning of the next line"
                        },
                        fix(fixer) {
                            if (prevToken && nextToken && sourceCode.commentsExistBetween(prevToken, nextToken)) {
                                return null;
                            }
                            const start = prevToken ? prevToken.range[1] : semiToken.range[0];
                            const end = nextToken ? nextToken.range[0] : semiToken.range[1];
                            const text = (expected === "last") ? ";\n" : "\n;";
                            return fixer.replaceTextRange([start, end], text);
                        }
                    });
                }
            }
            return {
                [SELECTOR](node) {
                    if (option === "first" && isLastChild(node)) {
                        return;
                    }
                    const lastToken = sourceCode.getLastToken(node);
                    if (astUtils.isSemicolonToken(lastToken)) {
                        check(lastToken, option);
                    }
                },
                ForStatement(node) {
                    const firstSemi = node.init && sourceCode.getTokenAfter(node.init, astUtils.isSemicolonToken);
                    const secondSemi = node.test && sourceCode.getTokenAfter(node.test, astUtils.isSemicolonToken);
                    if (firstSemi) {
                        check(firstSemi, "last");
                    }
                    if (secondSemi) {
                        check(secondSemi, "last");
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/sort-imports.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce sorted import declarations within modules",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/sort-imports"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        ignoreCase: {
                            type: "boolean",
                            default: false
                        },
                        memberSyntaxSortOrder: {
                            type: "array",
                            items: {
                                enum: ["none", "all", "multiple", "single"]
                            },
                            uniqueItems: true,
                            minItems: 4,
                            maxItems: 4
                        },
                        ignoreDeclarationSort: {
                            type: "boolean",
                            default: false
                        },
                        ignoreMemberSort: {
                            type: "boolean",
                            default: false
                        },
                        allowSeparatedGroups: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "code",
            messages: {
                sortImportsAlphabetically: "Imports should be sorted alphabetically.",
                sortMembersAlphabetically: "Member '{{memberName}}' of the import declaration should be sorted alphabetically.",
                unexpectedSyntaxOrder: "Expected '{{syntaxA}}' syntax before '{{syntaxB}}' syntax."
            }
        },
        create(context) {
            const configuration = context.options[0] || {}, ignoreCase = configuration.ignoreCase || false, ignoreDeclarationSort = configuration.ignoreDeclarationSort || false, ignoreMemberSort = configuration.ignoreMemberSort || false, memberSyntaxSortOrder = configuration.memberSyntaxSortOrder || ["none", "all", "multiple", "single"], allowSeparatedGroups = configuration.allowSeparatedGroups || false, sourceCode = context.getSourceCode();
            let previousDeclaration = null;
            /**
             * Gets the used member syntax style.
             *
             * import "my-module.js" --> none
             * import * as myModule from "my-module.js" --> all
             * import {myMember} from "my-module.js" --> single
             * import {foo, bar} from  "my-module.js" --> multiple
             * @param {ASTNode} node the ImportDeclaration node.
             * @returns {string} used member parameter style, ["all", "multiple", "single"]
             */
            function usedMemberSyntax(node) {
                if (node.specifiers.length === 0) {
                    return "none";
                }
                if (node.specifiers[0].type === "ImportNamespaceSpecifier") {
                    return "all";
                }
                if (node.specifiers.length === 1) {
                    return "single";
                }
                return "multiple";
            }
            /**
             * Gets the group by member parameter index for given declaration.
             * @param {ASTNode} node the ImportDeclaration node.
             * @returns {number} the declaration group by member index.
             */
            function getMemberParameterGroupIndex(node) {
                return memberSyntaxSortOrder.indexOf(usedMemberSyntax(node));
            }
            /**
             * Gets the local name of the first imported module.
             * @param {ASTNode} node the ImportDeclaration node.
             * @returns {?string} the local name of the first imported module.
             */
            function getFirstLocalMemberName(node) {
                if (node.specifiers[0]) {
                    return node.specifiers[0].local.name;
                }
                return null;
            }
            /**
             * Calculates number of lines between two nodes. It is assumed that the given `left` node appears before
             * the given `right` node in the source code. Lines are counted from the end of the `left` node till the
             * start of the `right` node. If the given nodes are on the same line, it returns `0`, same as if they were
             * on two consecutive lines.
             * @param {ASTNode} left node that appears before the given `right` node.
             * @param {ASTNode} right node that appears after the given `left` node.
             * @returns {number} number of lines between nodes.
             */
            function getNumberOfLinesBetween(left, right) {
                return Math.max(right.loc.start.line - left.loc.end.line - 1, 0);
            }
            return {
                ImportDeclaration(node) {
                    if (!ignoreDeclarationSort) {
                        if (previousDeclaration &&
                            allowSeparatedGroups &&
                            getNumberOfLinesBetween(previousDeclaration, node) > 0) {
                            // reset declaration sort
                            previousDeclaration = null;
                        }
                        if (previousDeclaration) {
                            const currentMemberSyntaxGroupIndex = getMemberParameterGroupIndex(node), previousMemberSyntaxGroupIndex = getMemberParameterGroupIndex(previousDeclaration);
                            let currentLocalMemberName = getFirstLocalMemberName(node), previousLocalMemberName = getFirstLocalMemberName(previousDeclaration);
                            if (ignoreCase) {
                                previousLocalMemberName = previousLocalMemberName && previousLocalMemberName.toLowerCase();
                                currentLocalMemberName = currentLocalMemberName && currentLocalMemberName.toLowerCase();
                            }
                            /*
                             * When the current declaration uses a different member syntax,
                             * then check if the ordering is correct.
                             * Otherwise, make a default string compare (like rule sort-vars to be consistent) of the first used local member name.
                             */
                            if (currentMemberSyntaxGroupIndex !== previousMemberSyntaxGroupIndex) {
                                if (currentMemberSyntaxGroupIndex < previousMemberSyntaxGroupIndex) {
                                    context.report({
                                        node,
                                        messageId: "unexpectedSyntaxOrder",
                                        data: {
                                            syntaxA: memberSyntaxSortOrder[currentMemberSyntaxGroupIndex],
                                            syntaxB: memberSyntaxSortOrder[previousMemberSyntaxGroupIndex]
                                        }
                                    });
                                }
                            }
                            else {
                                if (previousLocalMemberName &&
                                    currentLocalMemberName &&
                                    currentLocalMemberName < previousLocalMemberName) {
                                    context.report({
                                        node,
                                        messageId: "sortImportsAlphabetically"
                                    });
                                }
                            }
                        }
                        previousDeclaration = node;
                    }
                    if (!ignoreMemberSort) {
                        const importSpecifiers = node.specifiers.filter(specifier => specifier.type === "ImportSpecifier");
                        const getSortableName = ignoreCase ? specifier => specifier.local.name.toLowerCase() : specifier => specifier.local.name;
                        const firstUnsortedIndex = importSpecifiers.map(getSortableName).findIndex((name, index, array) => array[index - 1] > name);
                        if (firstUnsortedIndex !== -1) {
                            context.report({
                                node: importSpecifiers[firstUnsortedIndex],
                                messageId: "sortMembersAlphabetically",
                                data: { memberName: importSpecifiers[firstUnsortedIndex].local.name },
                                fix(fixer) {
                                    if (importSpecifiers.some(specifier => sourceCode.getCommentsBefore(specifier).length || sourceCode.getCommentsAfter(specifier).length)) {
                                        // If there are comments in the ImportSpecifier list, don't rearrange the specifiers.
                                        return null;
                                    }
                                    return fixer.replaceTextRange([importSpecifiers[0].range[0], importSpecifiers[importSpecifiers.length - 1].range[1]], importSpecifiers
                                        // Clone the importSpecifiers array to avoid mutating it
                                        .slice()
                                        // Sort the array into the desired order
                                        .sort((specifierA, specifierB) => {
                                        const aName = getSortableName(specifierA);
                                        const bName = getSortableName(specifierB);
                                        return aName > bName ? 1 : -1;
                                    })
                                        // Build a string out of the sorted list of import specifiers and the text between the originals
                                        .reduce((sourceText, specifier, index) => {
                                        const textAfterSpecifier = index === importSpecifiers.length - 1
                                            ? ""
                                            : sourceCode.getText().slice(importSpecifiers[index].range[1], importSpecifiers[index + 1].range[0]);
                                        return sourceText + sourceCode.getText(specifier) + textAfterSpecifier;
                                    }, ""));
                                }
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/sort-keys.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */, naturalCompare = require(672) /* natural-compare */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Gets the property name of the given `Property` node.
     *
     * - If the property's key is an `Identifier` node, this returns the key's name
     *   whether it's a computed property or not.
     * - If the property has a static name, this returns the static name.
     * - Otherwise, this returns null.
     * @param {ASTNode} node The `Property` node to get.
     * @returns {string|null} The property name or null.
     * @private
     */
    function getPropertyName(node) {
        const staticName = astUtils.getStaticPropertyName(node);
        if (staticName !== null) {
            return staticName;
        }
        return node.key.name || null;
    }
    /**
     * Functions which check that the given 2 names are in specific order.
     *
     * Postfix `I` is meant insensitive.
     * Postfix `N` is meant natural.
     * @private
     */
    const isValidOrders = {
        asc(a, b) {
            return a <= b;
        },
        ascI(a, b) {
            return a.toLowerCase() <= b.toLowerCase();
        },
        ascN(a, b) {
            return naturalCompare(a, b) <= 0;
        },
        ascIN(a, b) {
            return naturalCompare(a.toLowerCase(), b.toLowerCase()) <= 0;
        },
        desc(a, b) {
            return isValidOrders.asc(b, a);
        },
        descI(a, b) {
            return isValidOrders.ascI(b, a);
        },
        descN(a, b) {
            return isValidOrders.ascN(b, a);
        },
        descIN(a, b) {
            return isValidOrders.ascIN(b, a);
        }
    };
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require object keys to be sorted",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/sort-keys"
            },
            schema: [
                {
                    enum: ["asc", "desc"]
                },
                {
                    type: "object",
                    properties: {
                        caseSensitive: {
                            type: "boolean",
                            default: true
                        },
                        natural: {
                            type: "boolean",
                            default: false
                        },
                        minKeys: {
                            type: "integer",
                            minimum: 2,
                            default: 2
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                sortKeys: "Expected object keys to be in {{natural}}{{insensitive}}{{order}}ending order. '{{thisName}}' should be before '{{prevName}}'."
            }
        },
        create(context) {
            // Parse options.
            const order = context.options[0] || "asc";
            const options = context.options[1];
            const insensitive = options && options.caseSensitive === false;
            const natural = options && options.natural;
            const minKeys = options && options.minKeys;
            const isValidOrder = isValidOrders[order + (insensitive ? "I" : "") + (natural ? "N" : "")];
            // The stack to save the previous property's name for each object literals.
            let stack = null;
            return {
                ObjectExpression(node) {
                    stack = {
                        upper: stack,
                        prevName: null,
                        numKeys: node.properties.length
                    };
                },
                "ObjectExpression:exit"() {
                    stack = stack.upper;
                },
                SpreadElement(node) {
                    if (node.parent.type === "ObjectExpression") {
                        stack.prevName = null;
                    }
                },
                Property(node) {
                    if (node.parent.type === "ObjectPattern") {
                        return;
                    }
                    const prevName = stack.prevName;
                    const numKeys = stack.numKeys;
                    const thisName = getPropertyName(node);
                    if (thisName !== null) {
                        stack.prevName = thisName;
                    }
                    if (prevName === null || thisName === null || numKeys < minKeys) {
                        return;
                    }
                    if (!isValidOrder(prevName, thisName)) {
                        context.report({
                            node,
                            loc: node.key.loc,
                            messageId: "sortKeys",
                            data: {
                                thisName,
                                prevName,
                                order,
                                insensitive: insensitive ? "insensitive " : "",
                                natural: natural ? "natural " : ""
                            }
                        });
                    }
                }
            };
        }
    };
},
/* natural-compare/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
     * @version    1.4.0
     * @date       2015-10-26
     * @stability  3 - Stable
     * @author     Lauri Rooden (https://github.com/litejs/natural-compare-lite)
     * @license    MIT License
     */
    var naturalCompare = function (a, b) {
        var i, codeA, codeB = 1, posA = 0, posB = 0, alphabet = String.alphabet;
        function getCode(str, pos, code) {
            if (code) {
                for (i = pos; code = getCode(str, i), code < 76 && code > 65;)
                    ++i;
                return +str.slice(pos - 1, i);
            }
            code = alphabet && alphabet.indexOf(str.charAt(pos));
            return code > -1 ? code + 76 : ((code = str.charCodeAt(pos) || 0), code < 45 || code > 127) ? code
                : code < 46 ? 65 // -
                    : code < 48 ? code - 1
                        : code < 58 ? code + 18 // 0-9
                            : code < 65 ? code - 11
                                : code < 91 ? code + 11 // A-Z
                                    : code < 97 ? code - 37
                                        : code < 123 ? code + 5 // a-z
                                            : code - 63;
        }
        if ((a += "") != (b += ""))
            for (; codeB;) {
                codeA = getCode(a, posA++);
                codeB = getCode(b, posB++);
                if (codeA < 76 && codeB < 76 && codeA > 66 && codeB > 66) {
                    codeA = getCode(a, posA, posA);
                    codeB = getCode(b, posB, posA = i);
                    posB = i;
                }
                if (codeA != codeB)
                    return (codeA < codeB) ? -1 : 1;
            }
        return 0;
    };
    try {
        module.exports = naturalCompare;
    }
    catch (e) {
        String.naturalCompare = naturalCompare;
    }
},
/* eslint/lib/rules/sort-vars.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require variables within the same declaration block to be sorted",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/sort-vars"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        ignoreCase: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "code",
            messages: {
                sortVars: "Variables within the same declaration block should be sorted alphabetically."
            }
        },
        create(context) {
            const configuration = context.options[0] || {}, ignoreCase = configuration.ignoreCase || false, sourceCode = context.getSourceCode();
            return {
                VariableDeclaration(node) {
                    const idDeclarations = node.declarations.filter(decl => decl.id.type === "Identifier");
                    const getSortableName = ignoreCase ? decl => decl.id.name.toLowerCase() : decl => decl.id.name;
                    const unfixable = idDeclarations.some(decl => decl.init !== null && decl.init.type !== "Literal");
                    let fixed = false;
                    idDeclarations.slice(1).reduce((memo, decl) => {
                        const lastVariableName = getSortableName(memo), currentVariableName = getSortableName(decl);
                        if (currentVariableName < lastVariableName) {
                            context.report({
                                node: decl,
                                messageId: "sortVars",
                                fix(fixer) {
                                    if (unfixable || fixed) {
                                        return null;
                                    }
                                    return fixer.replaceTextRange([idDeclarations[0].range[0], idDeclarations[idDeclarations.length - 1].range[1]], idDeclarations
                                        // Clone the idDeclarations array to avoid mutating it
                                        .slice()
                                        // Sort the array into the desired order
                                        .sort((declA, declB) => {
                                        const aName = getSortableName(declA);
                                        const bName = getSortableName(declB);
                                        return aName > bName ? 1 : -1;
                                    })
                                        // Build a string out of the sorted list of identifier declarations and the text between the originals
                                        .reduce((sourceText, identifier, index) => {
                                        const textAfterIdentifier = index === idDeclarations.length - 1
                                            ? ""
                                            : sourceCode.getText().slice(idDeclarations[index].range[1], idDeclarations[index + 1].range[0]);
                                        return sourceText + sourceCode.getText(identifier) + textAfterIdentifier;
                                    }, ""));
                                }
                            });
                            fixed = true;
                            return memo;
                        }
                        return decl;
                    }, idDeclarations[0]);
                }
            };
        }
    };
},
/* eslint/lib/rules/space-before-blocks.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Checks whether the given node represents the body of a function.
     * @param {ASTNode} node the node to check.
     * @returns {boolean} `true` if the node is function body.
     */
    function isFunctionBody(node) {
        const parent = node.parent;
        return (node.type === "BlockStatement" &&
            astUtils.isFunction(parent) &&
            parent.body === node);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing before blocks",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/space-before-blocks"
            },
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["always", "never"]
                        },
                        {
                            type: "object",
                            properties: {
                                keywords: {
                                    enum: ["always", "never", "off"]
                                },
                                functions: {
                                    enum: ["always", "never", "off"]
                                },
                                classes: {
                                    enum: ["always", "never", "off"]
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                unexpectedSpace: "Unexpected space before opening brace.",
                missingSpace: "Missing space before opening brace."
            }
        },
        create(context) {
            const config = context.options[0], sourceCode = context.getSourceCode();
            let alwaysFunctions = true, alwaysKeywords = true, alwaysClasses = true, neverFunctions = false, neverKeywords = false, neverClasses = false;
            if (typeof config === "object") {
                alwaysFunctions = config.functions === "always";
                alwaysKeywords = config.keywords === "always";
                alwaysClasses = config.classes === "always";
                neverFunctions = config.functions === "never";
                neverKeywords = config.keywords === "never";
                neverClasses = config.classes === "never";
            }
            else if (config === "never") {
                alwaysFunctions = false;
                alwaysKeywords = false;
                alwaysClasses = false;
                neverFunctions = true;
                neverKeywords = true;
                neverClasses = true;
            }
            /**
             * Checks whether the spacing before the given block is already controlled by another rule:
             * - `arrow-spacing` checks spaces after `=>`.
             * - `keyword-spacing` checks spaces after keywords in certain contexts.
             * @param {Token} precedingToken first token before the block.
             * @param {ASTNode|Token} node `BlockStatement` node or `{` token of a `SwitchStatement` node.
             * @returns {boolean} `true` if requiring or disallowing spaces before the given block could produce conflicts with other rules.
             */
            function isConflicted(precedingToken, node) {
                return astUtils.isArrowToken(precedingToken) ||
                    astUtils.isKeywordToken(precedingToken) && !isFunctionBody(node);
            }
            /**
             * Checks the given BlockStatement node has a preceding space if it doesn’t start on a new line.
             * @param {ASTNode|Token} node The AST node of a BlockStatement.
             * @returns {void} undefined.
             */
            function checkPrecedingSpace(node) {
                const precedingToken = sourceCode.getTokenBefore(node);
                if (precedingToken && !isConflicted(precedingToken, node) && astUtils.isTokenOnSameLine(precedingToken, node)) {
                    const hasSpace = sourceCode.isSpaceBetweenTokens(precedingToken, node);
                    let requireSpace;
                    let requireNoSpace;
                    if (isFunctionBody(node)) {
                        requireSpace = alwaysFunctions;
                        requireNoSpace = neverFunctions;
                    }
                    else if (node.type === "ClassBody") {
                        requireSpace = alwaysClasses;
                        requireNoSpace = neverClasses;
                    }
                    else {
                        requireSpace = alwaysKeywords;
                        requireNoSpace = neverKeywords;
                    }
                    if (requireSpace && !hasSpace) {
                        context.report({
                            node,
                            messageId: "missingSpace",
                            fix(fixer) {
                                return fixer.insertTextBefore(node, " ");
                            }
                        });
                    }
                    else if (requireNoSpace && hasSpace) {
                        context.report({
                            node,
                            messageId: "unexpectedSpace",
                            fix(fixer) {
                                return fixer.removeRange([precedingToken.range[1], node.range[0]]);
                            }
                        });
                    }
                }
            }
            /**
             * Checks if the CaseBlock of an given SwitchStatement node has a preceding space.
             * @param {ASTNode} node The node of a SwitchStatement.
             * @returns {void} undefined.
             */
            function checkSpaceBeforeCaseBlock(node) {
                const cases = node.cases;
                let openingBrace;
                if (cases.length > 0) {
                    openingBrace = sourceCode.getTokenBefore(cases[0]);
                }
                else {
                    openingBrace = sourceCode.getLastToken(node, 1);
                }
                checkPrecedingSpace(openingBrace);
            }
            return {
                BlockStatement: checkPrecedingSpace,
                ClassBody: checkPrecedingSpace,
                SwitchStatement: checkSpaceBeforeCaseBlock
            };
        }
    };
},
/* eslint/lib/rules/space-before-function-paren.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing before `function` definition opening parenthesis",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/space-before-function-paren"
            },
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["always", "never"]
                        },
                        {
                            type: "object",
                            properties: {
                                anonymous: {
                                    enum: ["always", "never", "ignore"]
                                },
                                named: {
                                    enum: ["always", "never", "ignore"]
                                },
                                asyncArrow: {
                                    enum: ["always", "never", "ignore"]
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                unexpectedSpace: "Unexpected space before function parentheses.",
                missingSpace: "Missing space before function parentheses."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const baseConfig = typeof context.options[0] === "string" ? context.options[0] : "always";
            const overrideConfig = typeof context.options[0] === "object" ? context.options[0] : {};
            /**
             * Determines whether a function has a name.
             * @param {ASTNode} node The function node.
             * @returns {boolean} Whether the function has a name.
             */
            function isNamedFunction(node) {
                if (node.id) {
                    return true;
                }
                const parent = node.parent;
                return parent.type === "MethodDefinition" ||
                    (parent.type === "Property" &&
                        (parent.kind === "get" ||
                            parent.kind === "set" ||
                            parent.method));
            }
            /**
             * Gets the config for a given function
             * @param {ASTNode} node The function node
             * @returns {string} "always", "never", or "ignore"
             */
            function getConfigForFunction(node) {
                if (node.type === "ArrowFunctionExpression") {
                    // Always ignore non-async functions and arrow functions without parens, e.g. async foo => bar
                    if (node.async && astUtils.isOpeningParenToken(sourceCode.getFirstToken(node, { skip: 1 }))) {
                        return overrideConfig.asyncArrow || baseConfig;
                    }
                }
                else if (isNamedFunction(node)) {
                    return overrideConfig.named || baseConfig;
                    // `generator-star-spacing` should warn anonymous generators. E.g. `function* () {}`
                }
                else if (!node.generator) {
                    return overrideConfig.anonymous || baseConfig;
                }
                return "ignore";
            }
            /**
             * Checks the parens of a function node
             * @param {ASTNode} node A function node
             * @returns {void}
             */
            function checkFunction(node) {
                const functionConfig = getConfigForFunction(node);
                if (functionConfig === "ignore") {
                    return;
                }
                const rightToken = sourceCode.getFirstToken(node, astUtils.isOpeningParenToken);
                const leftToken = sourceCode.getTokenBefore(rightToken);
                const hasSpacing = sourceCode.isSpaceBetweenTokens(leftToken, rightToken);
                if (hasSpacing && functionConfig === "never") {
                    context.report({
                        node,
                        loc: {
                            start: leftToken.loc.end,
                            end: rightToken.loc.start
                        },
                        messageId: "unexpectedSpace",
                        fix(fixer) {
                            const comments = sourceCode.getCommentsBefore(rightToken);
                            // Don't fix anything if there's a single line comment between the left and the right token
                            if (comments.some(comment => comment.type === "Line")) {
                                return null;
                            }
                            return fixer.replaceTextRange([leftToken.range[1], rightToken.range[0]], comments.reduce((text, comment) => text + sourceCode.getText(comment), ""));
                        }
                    });
                }
                else if (!hasSpacing && functionConfig === "always") {
                    context.report({
                        node,
                        loc: rightToken.loc,
                        messageId: "missingSpace",
                        fix: fixer => fixer.insertTextAfter(leftToken, " ")
                    });
                }
            }
            return {
                ArrowFunctionExpression: checkFunction,
                FunctionDeclaration: checkFunction,
                FunctionExpression: checkFunction
            };
        }
    };
},
/* eslint/lib/rules/space-in-parens.js */ function _(require, module, exports, __esModule, __esExport) {
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing inside parentheses",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/space-in-parens"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never"]
                },
                {
                    type: "object",
                    properties: {
                        exceptions: {
                            type: "array",
                            items: {
                                enum: ["{}", "[]", "()", "empty"]
                            },
                            uniqueItems: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                missingOpeningSpace: "There must be a space after this paren.",
                missingClosingSpace: "There must be a space before this paren.",
                rejectedOpeningSpace: "There should be no space after this paren.",
                rejectedClosingSpace: "There should be no space before this paren."
            }
        },
        create(context) {
            const ALWAYS = context.options[0] === "always", exceptionsArrayOptions = (context.options[1] && context.options[1].exceptions) || [], options = {};
            let exceptions;
            if (exceptionsArrayOptions.length) {
                options.braceException = exceptionsArrayOptions.includes("{}");
                options.bracketException = exceptionsArrayOptions.includes("[]");
                options.parenException = exceptionsArrayOptions.includes("()");
                options.empty = exceptionsArrayOptions.includes("empty");
            }
            /**
             * Produces an object with the opener and closer exception values
             * @returns {Object} `openers` and `closers` exception values
             * @private
             */
            function getExceptions() {
                const openers = [], closers = [];
                if (options.braceException) {
                    openers.push("{");
                    closers.push("}");
                }
                if (options.bracketException) {
                    openers.push("[");
                    closers.push("]");
                }
                if (options.parenException) {
                    openers.push("(");
                    closers.push(")");
                }
                if (options.empty) {
                    openers.push(")");
                    closers.push("(");
                }
                return {
                    openers,
                    closers
                };
            }
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            const sourceCode = context.getSourceCode();
            /**
             * Determines if a token is one of the exceptions for the opener paren
             * @param {Object} token The token to check
             * @returns {boolean} True if the token is one of the exceptions for the opener paren
             */
            function isOpenerException(token) {
                return exceptions.openers.includes(token.value);
            }
            /**
             * Determines if a token is one of the exceptions for the closer paren
             * @param {Object} token The token to check
             * @returns {boolean} True if the token is one of the exceptions for the closer paren
             */
            function isCloserException(token) {
                return exceptions.closers.includes(token.value);
            }
            /**
             * Determines if an opening paren is immediately followed by a required space
             * @param {Object} openingParenToken The paren token
             * @param {Object} tokenAfterOpeningParen The token after it
             * @returns {boolean} True if the opening paren is missing a required space
             */
            function openerMissingSpace(openingParenToken, tokenAfterOpeningParen) {
                if (sourceCode.isSpaceBetweenTokens(openingParenToken, tokenAfterOpeningParen)) {
                    return false;
                }
                if (!options.empty && astUtils.isClosingParenToken(tokenAfterOpeningParen)) {
                    return false;
                }
                if (ALWAYS) {
                    return !isOpenerException(tokenAfterOpeningParen);
                }
                return isOpenerException(tokenAfterOpeningParen);
            }
            /**
             * Determines if an opening paren is immediately followed by a disallowed space
             * @param {Object} openingParenToken The paren token
             * @param {Object} tokenAfterOpeningParen The token after it
             * @returns {boolean} True if the opening paren has a disallowed space
             */
            function openerRejectsSpace(openingParenToken, tokenAfterOpeningParen) {
                if (!astUtils.isTokenOnSameLine(openingParenToken, tokenAfterOpeningParen)) {
                    return false;
                }
                if (tokenAfterOpeningParen.type === "Line") {
                    return false;
                }
                if (!sourceCode.isSpaceBetweenTokens(openingParenToken, tokenAfterOpeningParen)) {
                    return false;
                }
                if (ALWAYS) {
                    return isOpenerException(tokenAfterOpeningParen);
                }
                return !isOpenerException(tokenAfterOpeningParen);
            }
            /**
             * Determines if a closing paren is immediately preceded by a required space
             * @param {Object} tokenBeforeClosingParen The token before the paren
             * @param {Object} closingParenToken The paren token
             * @returns {boolean} True if the closing paren is missing a required space
             */
            function closerMissingSpace(tokenBeforeClosingParen, closingParenToken) {
                if (sourceCode.isSpaceBetweenTokens(tokenBeforeClosingParen, closingParenToken)) {
                    return false;
                }
                if (!options.empty && astUtils.isOpeningParenToken(tokenBeforeClosingParen)) {
                    return false;
                }
                if (ALWAYS) {
                    return !isCloserException(tokenBeforeClosingParen);
                }
                return isCloserException(tokenBeforeClosingParen);
            }
            /**
             * Determines if a closer paren is immediately preceded by a disallowed space
             * @param {Object} tokenBeforeClosingParen The token before the paren
             * @param {Object} closingParenToken The paren token
             * @returns {boolean} True if the closing paren has a disallowed space
             */
            function closerRejectsSpace(tokenBeforeClosingParen, closingParenToken) {
                if (!astUtils.isTokenOnSameLine(tokenBeforeClosingParen, closingParenToken)) {
                    return false;
                }
                if (!sourceCode.isSpaceBetweenTokens(tokenBeforeClosingParen, closingParenToken)) {
                    return false;
                }
                if (ALWAYS) {
                    return isCloserException(tokenBeforeClosingParen);
                }
                return !isCloserException(tokenBeforeClosingParen);
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program: function checkParenSpaces(node) {
                    exceptions = getExceptions();
                    const tokens = sourceCode.tokensAndComments;
                    tokens.forEach((token, i) => {
                        const prevToken = tokens[i - 1];
                        const nextToken = tokens[i + 1];
                        // if token is not an opening or closing paren token, do nothing
                        if (!astUtils.isOpeningParenToken(token) && !astUtils.isClosingParenToken(token)) {
                            return;
                        }
                        // if token is an opening paren and is not followed by a required space
                        if (token.value === "(" && openerMissingSpace(token, nextToken)) {
                            context.report({
                                node,
                                loc: token.loc,
                                messageId: "missingOpeningSpace",
                                fix(fixer) {
                                    return fixer.insertTextAfter(token, " ");
                                }
                            });
                        }
                        // if token is an opening paren and is followed by a disallowed space
                        if (token.value === "(" && openerRejectsSpace(token, nextToken)) {
                            context.report({
                                node,
                                loc: { start: token.loc.end, end: nextToken.loc.start },
                                messageId: "rejectedOpeningSpace",
                                fix(fixer) {
                                    return fixer.removeRange([token.range[1], nextToken.range[0]]);
                                }
                            });
                        }
                        // if token is a closing paren and is not preceded by a required space
                        if (token.value === ")" && closerMissingSpace(prevToken, token)) {
                            context.report({
                                node,
                                loc: token.loc,
                                messageId: "missingClosingSpace",
                                fix(fixer) {
                                    return fixer.insertTextBefore(token, " ");
                                }
                            });
                        }
                        // if token is a closing paren and is preceded by a disallowed space
                        if (token.value === ")" && closerRejectsSpace(prevToken, token)) {
                            context.report({
                                node,
                                loc: { start: prevToken.loc.end, end: token.loc.start },
                                messageId: "rejectedClosingSpace",
                                fix(fixer) {
                                    return fixer.removeRange([prevToken.range[1], token.range[0]]);
                                }
                            });
                        }
                    });
                }
            };
        }
    };
},
/* eslint/lib/rules/space-infix-ops.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require spacing around infix operators",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/space-infix-ops"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        int32Hint: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                missingSpace: "Operator '{{operator}}' must be spaced."
            }
        },
        create(context) {
            const int32Hint = context.options[0] ? context.options[0].int32Hint === true : false;
            const sourceCode = context.getSourceCode();
            /**
             * Returns the first token which violates the rule
             * @param {ASTNode} left The left node of the main node
             * @param {ASTNode} right The right node of the main node
             * @param {string} op The operator of the main node
             * @returns {Object} The violator token or null
             * @private
             */
            function getFirstNonSpacedToken(left, right, op) {
                const operator = sourceCode.getFirstTokenBetween(left, right, token => token.value === op);
                const prev = sourceCode.getTokenBefore(operator);
                const next = sourceCode.getTokenAfter(operator);
                if (!sourceCode.isSpaceBetweenTokens(prev, operator) || !sourceCode.isSpaceBetweenTokens(operator, next)) {
                    return operator;
                }
                return null;
            }
            /**
             * Reports an AST node as a rule violation
             * @param {ASTNode} mainNode The node to report
             * @param {Object} culpritToken The token which has a problem
             * @returns {void}
             * @private
             */
            function report(mainNode, culpritToken) {
                context.report({
                    node: mainNode,
                    loc: culpritToken.loc,
                    messageId: "missingSpace",
                    data: {
                        operator: culpritToken.value
                    },
                    fix(fixer) {
                        const previousToken = sourceCode.getTokenBefore(culpritToken);
                        const afterToken = sourceCode.getTokenAfter(culpritToken);
                        let fixString = "";
                        if (culpritToken.range[0] - previousToken.range[1] === 0) {
                            fixString = " ";
                        }
                        fixString += culpritToken.value;
                        if (afterToken.range[0] - culpritToken.range[1] === 0) {
                            fixString += " ";
                        }
                        return fixer.replaceText(culpritToken, fixString);
                    }
                });
            }
            /**
             * Check if the node is binary then report
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkBinary(node) {
                const leftNode = (node.left.typeAnnotation) ? node.left.typeAnnotation : node.left;
                const rightNode = node.right;
                // search for = in AssignmentPattern nodes
                const operator = node.operator || "=";
                const nonSpacedNode = getFirstNonSpacedToken(leftNode, rightNode, operator);
                if (nonSpacedNode) {
                    if (!(int32Hint && sourceCode.getText(node).endsWith("|0"))) {
                        report(node, nonSpacedNode);
                    }
                }
            }
            /**
             * Check if the node is conditional
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkConditional(node) {
                const nonSpacedConsequentNode = getFirstNonSpacedToken(node.test, node.consequent, "?");
                const nonSpacedAlternateNode = getFirstNonSpacedToken(node.consequent, node.alternate, ":");
                if (nonSpacedConsequentNode) {
                    report(node, nonSpacedConsequentNode);
                }
                if (nonSpacedAlternateNode) {
                    report(node, nonSpacedAlternateNode);
                }
            }
            /**
             * Check if the node is a variable
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkVar(node) {
                const leftNode = (node.id.typeAnnotation) ? node.id.typeAnnotation : node.id;
                const rightNode = node.init;
                if (rightNode) {
                    const nonSpacedNode = getFirstNonSpacedToken(leftNode, rightNode, "=");
                    if (nonSpacedNode) {
                        report(node, nonSpacedNode);
                    }
                }
            }
            return {
                AssignmentExpression: checkBinary,
                AssignmentPattern: checkBinary,
                BinaryExpression: checkBinary,
                LogicalExpression: checkBinary,
                ConditionalExpression: checkConditional,
                VariableDeclarator: checkVar
            };
        }
    };
},
/* eslint/lib/rules/space-unary-ops.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce consistent spacing before or after unary operators",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/space-unary-ops"
            },
            fixable: "whitespace",
            schema: [
                {
                    type: "object",
                    properties: {
                        words: {
                            type: "boolean",
                            default: true
                        },
                        nonwords: {
                            type: "boolean",
                            default: false
                        },
                        overrides: {
                            type: "object",
                            additionalProperties: {
                                type: "boolean"
                            }
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedBefore: "Unexpected space before unary operator '{{operator}}'.",
                unexpectedAfter: "Unexpected space after unary operator '{{operator}}'.",
                unexpectedAfterWord: "Unexpected space after unary word operator '{{word}}'.",
                wordOperator: "Unary word operator '{{word}}' must be followed by whitespace.",
                operator: "Unary operator '{{operator}}' must be followed by whitespace.",
                beforeUnaryExpressions: "Space is required before unary expressions '{{token}}'."
            }
        },
        create(context) {
            const options = context.options[0] || { words: true, nonwords: false };
            const sourceCode = context.getSourceCode();
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Check if the node is the first "!" in a "!!" convert to Boolean expression
             * @param {ASTnode} node AST node
             * @returns {boolean} Whether or not the node is first "!" in "!!"
             */
            function isFirstBangInBangBangExpression(node) {
                return node && node.type === "UnaryExpression" && node.argument.operator === "!" &&
                    node.argument && node.argument.type === "UnaryExpression" && node.argument.operator === "!";
            }
            /**
             * Checks if an override exists for a given operator.
             * @param {string} operator Operator
             * @returns {boolean} Whether or not an override has been provided for the operator
             */
            function overrideExistsForOperator(operator) {
                return options.overrides && Object.prototype.hasOwnProperty.call(options.overrides, operator);
            }
            /**
             * Gets the value that the override was set to for this operator
             * @param {string} operator Operator
             * @returns {boolean} Whether or not an override enforces a space with this operator
             */
            function overrideEnforcesSpaces(operator) {
                return options.overrides[operator];
            }
            /**
             * Verify Unary Word Operator has spaces after the word operator
             * @param {ASTnode} node AST node
             * @param {Object} firstToken first token from the AST node
             * @param {Object} secondToken second token from the AST node
             * @param {string} word The word to be used for reporting
             * @returns {void}
             */
            function verifyWordHasSpaces(node, firstToken, secondToken, word) {
                if (secondToken.range[0] === firstToken.range[1]) {
                    context.report({
                        node,
                        messageId: "wordOperator",
                        data: {
                            word
                        },
                        fix(fixer) {
                            return fixer.insertTextAfter(firstToken, " ");
                        }
                    });
                }
            }
            /**
             * Verify Unary Word Operator doesn't have spaces after the word operator
             * @param {ASTnode} node AST node
             * @param {Object} firstToken first token from the AST node
             * @param {Object} secondToken second token from the AST node
             * @param {string} word The word to be used for reporting
             * @returns {void}
             */
            function verifyWordDoesntHaveSpaces(node, firstToken, secondToken, word) {
                if (astUtils.canTokensBeAdjacent(firstToken, secondToken)) {
                    if (secondToken.range[0] > firstToken.range[1]) {
                        context.report({
                            node,
                            messageId: "unexpectedAfterWord",
                            data: {
                                word
                            },
                            fix(fixer) {
                                return fixer.removeRange([firstToken.range[1], secondToken.range[0]]);
                            }
                        });
                    }
                }
            }
            /**
             * Check Unary Word Operators for spaces after the word operator
             * @param {ASTnode} node AST node
             * @param {Object} firstToken first token from the AST node
             * @param {Object} secondToken second token from the AST node
             * @param {string} word The word to be used for reporting
             * @returns {void}
             */
            function checkUnaryWordOperatorForSpaces(node, firstToken, secondToken, word) {
                if (overrideExistsForOperator(word)) {
                    if (overrideEnforcesSpaces(word)) {
                        verifyWordHasSpaces(node, firstToken, secondToken, word);
                    }
                    else {
                        verifyWordDoesntHaveSpaces(node, firstToken, secondToken, word);
                    }
                }
                else if (options.words) {
                    verifyWordHasSpaces(node, firstToken, secondToken, word);
                }
                else {
                    verifyWordDoesntHaveSpaces(node, firstToken, secondToken, word);
                }
            }
            /**
             * Verifies YieldExpressions satisfy spacing requirements
             * @param {ASTnode} node AST node
             * @returns {void}
             */
            function checkForSpacesAfterYield(node) {
                const tokens = sourceCode.getFirstTokens(node, 3), word = "yield";
                if (!node.argument || node.delegate) {
                    return;
                }
                checkUnaryWordOperatorForSpaces(node, tokens[0], tokens[1], word);
            }
            /**
             * Verifies AwaitExpressions satisfy spacing requirements
             * @param {ASTNode} node AwaitExpression AST node
             * @returns {void}
             */
            function checkForSpacesAfterAwait(node) {
                const tokens = sourceCode.getFirstTokens(node, 3);
                checkUnaryWordOperatorForSpaces(node, tokens[0], tokens[1], "await");
            }
            /**
             * Verifies UnaryExpression, UpdateExpression and NewExpression have spaces before or after the operator
             * @param {ASTnode} node AST node
             * @param {Object} firstToken First token in the expression
             * @param {Object} secondToken Second token in the expression
             * @returns {void}
             */
            function verifyNonWordsHaveSpaces(node, firstToken, secondToken) {
                if (node.prefix) {
                    if (isFirstBangInBangBangExpression(node)) {
                        return;
                    }
                    if (firstToken.range[1] === secondToken.range[0]) {
                        context.report({
                            node,
                            messageId: "operator",
                            data: {
                                operator: firstToken.value
                            },
                            fix(fixer) {
                                return fixer.insertTextAfter(firstToken, " ");
                            }
                        });
                    }
                }
                else {
                    if (firstToken.range[1] === secondToken.range[0]) {
                        context.report({
                            node,
                            messageId: "beforeUnaryExpressions",
                            data: {
                                token: secondToken.value
                            },
                            fix(fixer) {
                                return fixer.insertTextBefore(secondToken, " ");
                            }
                        });
                    }
                }
            }
            /**
             * Verifies UnaryExpression, UpdateExpression and NewExpression don't have spaces before or after the operator
             * @param {ASTnode} node AST node
             * @param {Object} firstToken First token in the expression
             * @param {Object} secondToken Second token in the expression
             * @returns {void}
             */
            function verifyNonWordsDontHaveSpaces(node, firstToken, secondToken) {
                if (node.prefix) {
                    if (secondToken.range[0] > firstToken.range[1]) {
                        context.report({
                            node,
                            messageId: "unexpectedAfter",
                            data: {
                                operator: firstToken.value
                            },
                            fix(fixer) {
                                if (astUtils.canTokensBeAdjacent(firstToken, secondToken)) {
                                    return fixer.removeRange([firstToken.range[1], secondToken.range[0]]);
                                }
                                return null;
                            }
                        });
                    }
                }
                else {
                    if (secondToken.range[0] > firstToken.range[1]) {
                        context.report({
                            node,
                            messageId: "unexpectedBefore",
                            data: {
                                operator: secondToken.value
                            },
                            fix(fixer) {
                                return fixer.removeRange([firstToken.range[1], secondToken.range[0]]);
                            }
                        });
                    }
                }
            }
            /**
             * Verifies UnaryExpression, UpdateExpression and NewExpression satisfy spacing requirements
             * @param {ASTnode} node AST node
             * @returns {void}
             */
            function checkForSpaces(node) {
                const tokens = node.type === "UpdateExpression" && !node.prefix
                    ? sourceCode.getLastTokens(node, 2)
                    : sourceCode.getFirstTokens(node, 2);
                const firstToken = tokens[0];
                const secondToken = tokens[1];
                if ((node.type === "NewExpression" || node.prefix) && firstToken.type === "Keyword") {
                    checkUnaryWordOperatorForSpaces(node, firstToken, secondToken, firstToken.value);
                    return;
                }
                const operator = node.prefix ? tokens[0].value : tokens[1].value;
                if (overrideExistsForOperator(operator)) {
                    if (overrideEnforcesSpaces(operator)) {
                        verifyNonWordsHaveSpaces(node, firstToken, secondToken);
                    }
                    else {
                        verifyNonWordsDontHaveSpaces(node, firstToken, secondToken);
                    }
                }
                else if (options.nonwords) {
                    verifyNonWordsHaveSpaces(node, firstToken, secondToken);
                }
                else {
                    verifyNonWordsDontHaveSpaces(node, firstToken, secondToken);
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                UnaryExpression: checkForSpaces,
                UpdateExpression: checkForSpaces,
                NewExpression: checkForSpaces,
                YieldExpression: checkForSpacesAfterYield,
                AwaitExpression: checkForSpacesAfterAwait
            };
        }
    };
},
/* eslint/lib/rules/spaced-comment.js */ function _(require, module, exports, __esModule, __esExport) {
    const escapeRegExp = require(329) /* escape-string-regexp */;
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Escapes the control characters of a given string.
     * @param {string} s A string to escape.
     * @returns {string} An escaped string.
     */
    function escape(s) {
        return `(?:${escapeRegExp(s)})`;
    }
    /**
     * Escapes the control characters of a given string.
     * And adds a repeat flag.
     * @param {string} s A string to escape.
     * @returns {string} An escaped string.
     */
    function escapeAndRepeat(s) {
        return `${escape(s)}+`;
    }
    /**
     * Parses `markers` option.
     * If markers don't include `"*"`, this adds `"*"` to allow JSDoc comments.
     * @param {string[]} [markers] A marker list.
     * @returns {string[]} A marker list.
     */
    function parseMarkersOption(markers) {
        // `*` is a marker for JSDoc comments.
        if (markers.indexOf("*") === -1) {
            return markers.concat("*");
        }
        return markers;
    }
    /**
     * Creates string pattern for exceptions.
     * Generated pattern:
     *
     * 1. A space or an exception pattern sequence.
     * @param {string[]} exceptions An exception pattern list.
     * @returns {string} A regular expression string for exceptions.
     */
    function createExceptionsPattern(exceptions) {
        let pattern = "";
        /*
         * A space or an exception pattern sequence.
         * []                 ==> "\s"
         * ["-"]              ==> "(?:\s|\-+$)"
         * ["-", "="]         ==> "(?:\s|(?:\-+|=+)$)"
         * ["-", "=", "--=="] ==> "(?:\s|(?:\-+|=+|(?:\-\-==)+)$)" ==> https://jex.im/regulex/#!embed=false&flags=&re=(%3F%3A%5Cs%7C(%3F%3A%5C-%2B%7C%3D%2B%7C(%3F%3A%5C-%5C-%3D%3D)%2B)%24)
         */
        if (exceptions.length === 0) {
            // a space.
            pattern += "\\s";
        }
        else {
            // a space or...
            pattern += "(?:\\s|";
            if (exceptions.length === 1) {
                // a sequence of the exception pattern.
                pattern += escapeAndRepeat(exceptions[0]);
            }
            else {
                // a sequence of one of the exception patterns.
                pattern += "(?:";
                pattern += exceptions.map(escapeAndRepeat).join("|");
                pattern += ")";
            }
            pattern += `(?:$|[${Array.from(astUtils.LINEBREAKS).join("")}]))`;
        }
        return pattern;
    }
    /**
     * Creates RegExp object for `always` mode.
     * Generated pattern for beginning of comment:
     *
     * 1. First, a marker or nothing.
     * 2. Next, a space or an exception pattern sequence.
     * @param {string[]} markers A marker list.
     * @param {string[]} exceptions An exception pattern list.
     * @returns {RegExp} A RegExp object for the beginning of a comment in `always` mode.
     */
    function createAlwaysStylePattern(markers, exceptions) {
        let pattern = "^";
        /*
         * A marker or nothing.
         * ["*"]            ==> "\*?"
         * ["*", "!"]       ==> "(?:\*|!)?"
         * ["*", "/", "!<"] ==> "(?:\*|\/|(?:!<))?" ==> https://jex.im/regulex/#!embed=false&flags=&re=(%3F%3A%5C*%7C%5C%2F%7C(%3F%3A!%3C))%3F
         */
        if (markers.length === 1) {
            // the marker.
            pattern += escape(markers[0]);
        }
        else {
            // one of markers.
            pattern += "(?:";
            pattern += markers.map(escape).join("|");
            pattern += ")";
        }
        pattern += "?"; // or nothing.
        pattern += createExceptionsPattern(exceptions);
        return new RegExp(pattern, "u");
    }
    /**
     * Creates RegExp object for `never` mode.
     * Generated pattern for beginning of comment:
     *
     * 1. First, a marker or nothing (captured).
     * 2. Next, a space or a tab.
     * @param {string[]} markers A marker list.
     * @returns {RegExp} A RegExp object for `never` mode.
     */
    function createNeverStylePattern(markers) {
        const pattern = `^(${markers.map(escape).join("|")})?[ \t]+`;
        return new RegExp(pattern, "u");
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce consistent spacing after the `//` or `/*` in a comment",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/spaced-comment"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never"]
                },
                {
                    type: "object",
                    properties: {
                        exceptions: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        },
                        markers: {
                            type: "array",
                            items: {
                                type: "string"
                            }
                        },
                        line: {
                            type: "object",
                            properties: {
                                exceptions: {
                                    type: "array",
                                    items: {
                                        type: "string"
                                    }
                                },
                                markers: {
                                    type: "array",
                                    items: {
                                        type: "string"
                                    }
                                }
                            },
                            additionalProperties: false
                        },
                        block: {
                            type: "object",
                            properties: {
                                exceptions: {
                                    type: "array",
                                    items: {
                                        type: "string"
                                    }
                                },
                                markers: {
                                    type: "array",
                                    items: {
                                        type: "string"
                                    }
                                },
                                balanced: {
                                    type: "boolean",
                                    default: false
                                }
                            },
                            additionalProperties: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                unexpectedSpaceAfterMarker: "Unexpected space or tab after marker ({{refChar}}) in comment.",
                expectedExceptionAfter: "Expected exception block, space or tab after '{{refChar}}' in comment.",
                unexpectedSpaceBefore: "Unexpected space or tab before '*/' in comment.",
                unexpectedSpaceAfter: "Unexpected space or tab after '{{refChar}}' in comment.",
                expectedSpaceBefore: "Expected space or tab before '*/' in comment.",
                expectedSpaceAfter: "Expected space or tab after '{{refChar}}' in comment."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            // Unless the first option is never, require a space
            const requireSpace = context.options[0] !== "never";
            /*
             * Parse the second options.
             * If markers don't include `"*"`, it's added automatically for JSDoc
             * comments.
             */
            const config = context.options[1] || {};
            const balanced = config.block && config.block.balanced;
            const styleRules = ["block", "line"].reduce((rule, type) => {
                const markers = parseMarkersOption(config[type] && config[type].markers || config.markers || []);
                const exceptions = config[type] && config[type].exceptions || config.exceptions || [];
                const endNeverPattern = "[ \t]+$";
                // Create RegExp object for valid patterns.
                rule[type] = {
                    beginRegex: requireSpace ? createAlwaysStylePattern(markers, exceptions) : createNeverStylePattern(markers),
                    endRegex: balanced && requireSpace ? new RegExp(`${createExceptionsPattern(exceptions)}$`, "u") : new RegExp(endNeverPattern, "u"),
                    hasExceptions: exceptions.length > 0,
                    captureMarker: new RegExp(`^(${markers.map(escape).join("|")})`, "u"),
                    markers: new Set(markers)
                };
                return rule;
            }, {});
            /**
             * Reports a beginning spacing error with an appropriate message.
             * @param {ASTNode} node A comment node to check.
             * @param {string} messageId An error message to report.
             * @param {Array} match An array of match results for markers.
             * @param {string} refChar Character used for reference in the error message.
             * @returns {void}
             */
            function reportBegin(node, messageId, match, refChar) {
                const type = node.type.toLowerCase(), commentIdentifier = type === "block" ? "/*" : "//";
                context.report({
                    node,
                    fix(fixer) {
                        const start = node.range[0];
                        let end = start + 2;
                        if (requireSpace) {
                            if (match) {
                                end += match[0].length;
                            }
                            return fixer.insertTextAfterRange([start, end], " ");
                        }
                        end += match[0].length;
                        return fixer.replaceTextRange([start, end], commentIdentifier + (match[1] ? match[1] : ""));
                    },
                    messageId,
                    data: { refChar }
                });
            }
            /**
             * Reports an ending spacing error with an appropriate message.
             * @param {ASTNode} node A comment node to check.
             * @param {string} messageId An error message to report.
             * @param {string} match An array of the matched whitespace characters.
             * @returns {void}
             */
            function reportEnd(node, messageId, match) {
                context.report({
                    node,
                    fix(fixer) {
                        if (requireSpace) {
                            return fixer.insertTextAfterRange([node.range[0], node.range[1] - 2], " ");
                        }
                        const end = node.range[1] - 2, start = end - match[0].length;
                        return fixer.replaceTextRange([start, end], "");
                    },
                    messageId
                });
            }
            /**
             * Reports a given comment if it's invalid.
             * @param {ASTNode} node a comment node to check.
             * @returns {void}
             */
            function checkCommentForSpace(node) {
                const type = node.type.toLowerCase(), rule = styleRules[type], commentIdentifier = type === "block" ? "/*" : "//";
                // Ignores empty comments and comments that consist only of a marker.
                if (node.value.length === 0 || rule.markers.has(node.value)) {
                    return;
                }
                const beginMatch = rule.beginRegex.exec(node.value);
                const endMatch = rule.endRegex.exec(node.value);
                // Checks.
                if (requireSpace) {
                    if (!beginMatch) {
                        const hasMarker = rule.captureMarker.exec(node.value);
                        const marker = hasMarker ? commentIdentifier + hasMarker[0] : commentIdentifier;
                        if (rule.hasExceptions) {
                            reportBegin(node, "expectedExceptionAfter", hasMarker, marker);
                        }
                        else {
                            reportBegin(node, "expectedSpaceAfter", hasMarker, marker);
                        }
                    }
                    if (balanced && type === "block" && !endMatch) {
                        reportEnd(node, "expectedSpaceBefore");
                    }
                }
                else {
                    if (beginMatch) {
                        if (!beginMatch[1]) {
                            reportBegin(node, "unexpectedSpaceAfter", beginMatch, commentIdentifier);
                        }
                        else {
                            reportBegin(node, "unexpectedSpaceAfterMarker", beginMatch, beginMatch[1]);
                        }
                    }
                    if (balanced && type === "block" && endMatch) {
                        reportEnd(node, "unexpectedSpaceBefore", endMatch);
                    }
                }
            }
            return {
                Program() {
                    const comments = sourceCode.getAllComments();
                    comments.filter(token => token.type !== "Shebang").forEach(checkCommentForSpace);
                }
            };
        }
    };
},
/* eslint/lib/rules/strict.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Gets all of the Use Strict Directives in the Directive Prologue of a group of
     * statements.
     * @param {ASTNode[]} statements Statements in the program or function body.
     * @returns {ASTNode[]} All of the Use Strict Directives.
     */
    function getUseStrictDirectives(statements) {
        const directives = [];
        for (let i = 0; i < statements.length; i++) {
            const statement = statements[i];
            if (statement.type === "ExpressionStatement" &&
                statement.expression.type === "Literal" &&
                statement.expression.value === "use strict") {
                directives[i] = statement;
            }
            else {
                break;
            }
        }
        return directives;
    }
    /**
     * Checks whether a given parameter is a simple parameter.
     * @param {ASTNode} node A pattern node to check.
     * @returns {boolean} `true` if the node is an Identifier node.
     */
    function isSimpleParameter(node) {
        return node.type === "Identifier";
    }
    /**
     * Checks whether a given parameter list is a simple parameter list.
     * @param {ASTNode[]} params A parameter list to check.
     * @returns {boolean} `true` if the every parameter is an Identifier node.
     */
    function isSimpleParameterList(params) {
        return params.every(isSimpleParameter);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require or disallow strict mode directives",
                category: "Strict Mode",
                recommended: false,
                url: "https://eslint.org/docs/rules/strict"
            },
            schema: [
                {
                    enum: ["never", "global", "function", "safe"]
                }
            ],
            fixable: "code",
            messages: {
                function: "Use the function form of 'use strict'.",
                global: "Use the global form of 'use strict'.",
                multiple: "Multiple 'use strict' directives.",
                never: "Strict mode is not permitted.",
                unnecessary: "Unnecessary 'use strict' directive.",
                module: "'use strict' is unnecessary inside of modules.",
                implied: "'use strict' is unnecessary when implied strict mode is enabled.",
                unnecessaryInClasses: "'use strict' is unnecessary inside of classes.",
                nonSimpleParameterList: "'use strict' directive inside a function with non-simple parameter list throws a syntax error since ES2016.",
                wrap: "Wrap {{name}} in a function with 'use strict' directive."
            }
        },
        create(context) {
            const ecmaFeatures = context.parserOptions.ecmaFeatures || {}, scopes = [], classScopes = [];
            let mode = context.options[0] || "safe";
            if (ecmaFeatures.impliedStrict) {
                mode = "implied";
            }
            else if (mode === "safe") {
                mode = ecmaFeatures.globalReturn ? "global" : "function";
            }
            /**
             * Determines whether a reported error should be fixed, depending on the error type.
             * @param {string} errorType The type of error
             * @returns {boolean} `true` if the reported error should be fixed
             */
            function shouldFix(errorType) {
                return errorType === "multiple" || errorType === "unnecessary" || errorType === "module" || errorType === "implied" || errorType === "unnecessaryInClasses";
            }
            /**
             * Gets a fixer function to remove a given 'use strict' directive.
             * @param {ASTNode} node The directive that should be removed
             * @returns {Function} A fixer function
             */
            function getFixFunction(node) {
                return fixer => fixer.remove(node);
            }
            /**
             * Report a slice of an array of nodes with a given message.
             * @param {ASTNode[]} nodes Nodes.
             * @param {string} start Index to start from.
             * @param {string} end Index to end before.
             * @param {string} messageId Message to display.
             * @param {boolean} fix `true` if the directive should be fixed (i.e. removed)
             * @returns {void}
             */
            function reportSlice(nodes, start, end, messageId, fix) {
                nodes.slice(start, end).forEach(node => {
                    context.report({ node, messageId, fix: fix ? getFixFunction(node) : null });
                });
            }
            /**
             * Report all nodes in an array with a given message.
             * @param {ASTNode[]} nodes Nodes.
             * @param {string} messageId Message id to display.
             * @param {boolean} fix `true` if the directive should be fixed (i.e. removed)
             * @returns {void}
             */
            function reportAll(nodes, messageId, fix) {
                reportSlice(nodes, 0, nodes.length, messageId, fix);
            }
            /**
             * Report all nodes in an array, except the first, with a given message.
             * @param {ASTNode[]} nodes Nodes.
             * @param {string} messageId Message id to display.
             * @param {boolean} fix `true` if the directive should be fixed (i.e. removed)
             * @returns {void}
             */
            function reportAllExceptFirst(nodes, messageId, fix) {
                reportSlice(nodes, 1, nodes.length, messageId, fix);
            }
            /**
             * Entering a function in 'function' mode pushes a new nested scope onto the
             * stack. The new scope is true if the nested function is strict mode code.
             * @param {ASTNode} node The function declaration or expression.
             * @param {ASTNode[]} useStrictDirectives The Use Strict Directives of the node.
             * @returns {void}
             */
            function enterFunctionInFunctionMode(node, useStrictDirectives) {
                const isInClass = classScopes.length > 0, isParentGlobal = scopes.length === 0 && classScopes.length === 0, isParentStrict = scopes.length > 0 && scopes[scopes.length - 1], isStrict = useStrictDirectives.length > 0;
                if (isStrict) {
                    if (!isSimpleParameterList(node.params)) {
                        context.report({ node: useStrictDirectives[0], messageId: "nonSimpleParameterList" });
                    }
                    else if (isParentStrict) {
                        context.report({ node: useStrictDirectives[0], messageId: "unnecessary", fix: getFixFunction(useStrictDirectives[0]) });
                    }
                    else if (isInClass) {
                        context.report({ node: useStrictDirectives[0], messageId: "unnecessaryInClasses", fix: getFixFunction(useStrictDirectives[0]) });
                    }
                    reportAllExceptFirst(useStrictDirectives, "multiple", true);
                }
                else if (isParentGlobal) {
                    if (isSimpleParameterList(node.params)) {
                        context.report({ node, messageId: "function" });
                    }
                    else {
                        context.report({
                            node,
                            messageId: "wrap",
                            data: { name: astUtils.getFunctionNameWithKind(node) }
                        });
                    }
                }
                scopes.push(isParentStrict || isStrict);
            }
            /**
             * Exiting a function in 'function' mode pops its scope off the stack.
             * @returns {void}
             */
            function exitFunctionInFunctionMode() {
                scopes.pop();
            }
            /**
             * Enter a function and either:
             * - Push a new nested scope onto the stack (in 'function' mode).
             * - Report all the Use Strict Directives (in the other modes).
             * @param {ASTNode} node The function declaration or expression.
             * @returns {void}
             */
            function enterFunction(node) {
                const isBlock = node.body.type === "BlockStatement", useStrictDirectives = isBlock
                    ? getUseStrictDirectives(node.body.body) : [];
                if (mode === "function") {
                    enterFunctionInFunctionMode(node, useStrictDirectives);
                }
                else if (useStrictDirectives.length > 0) {
                    if (isSimpleParameterList(node.params)) {
                        reportAll(useStrictDirectives, mode, shouldFix(mode));
                    }
                    else {
                        context.report({ node: useStrictDirectives[0], messageId: "nonSimpleParameterList" });
                        reportAllExceptFirst(useStrictDirectives, "multiple", true);
                    }
                }
            }
            const rule = {
                Program(node) {
                    const useStrictDirectives = getUseStrictDirectives(node.body);
                    if (node.sourceType === "module") {
                        mode = "module";
                    }
                    if (mode === "global") {
                        if (node.body.length > 0 && useStrictDirectives.length === 0) {
                            context.report({ node, messageId: "global" });
                        }
                        reportAllExceptFirst(useStrictDirectives, "multiple", true);
                    }
                    else {
                        reportAll(useStrictDirectives, mode, shouldFix(mode));
                    }
                },
                FunctionDeclaration: enterFunction,
                FunctionExpression: enterFunction,
                ArrowFunctionExpression: enterFunction
            };
            if (mode === "function") {
                Object.assign(rule, {
                    // Inside of class bodies are always strict mode.
                    ClassBody() {
                        classScopes.push(true);
                    },
                    "ClassBody:exit"() {
                        classScopes.pop();
                    },
                    "FunctionDeclaration:exit": exitFunctionInFunctionMode,
                    "FunctionExpression:exit": exitFunctionInFunctionMode,
                    "ArrowFunctionExpression:exit": exitFunctionInFunctionMode
                });
            }
            return rule;
        }
    };
},
/* eslint/lib/rules/switch-colon-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "enforce spacing around colons of switch statements",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/switch-colon-spacing"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        before: { type: "boolean", default: false },
                        after: { type: "boolean", default: true }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "whitespace",
            messages: {
                expectedBefore: "Expected space(s) before this colon.",
                expectedAfter: "Expected space(s) after this colon.",
                unexpectedBefore: "Unexpected space(s) before this colon.",
                unexpectedAfter: "Unexpected space(s) after this colon."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const options = context.options[0] || {};
            const beforeSpacing = options.before === true; // false by default
            const afterSpacing = options.after !== false; // true by default
            /**
             * Get the colon token of the given SwitchCase node.
             * @param {ASTNode} node The SwitchCase node to get.
             * @returns {Token} The colon token of the node.
             */
            function getColonToken(node) {
                if (node.test) {
                    return sourceCode.getTokenAfter(node.test, astUtils.isColonToken);
                }
                return sourceCode.getFirstToken(node, 1);
            }
            /**
             * Check whether the spacing between the given 2 tokens is valid or not.
             * @param {Token} left The left token to check.
             * @param {Token} right The right token to check.
             * @param {boolean} expected The expected spacing to check. `true` if there should be a space.
             * @returns {boolean} `true` if the spacing between the tokens is valid.
             */
            function isValidSpacing(left, right, expected) {
                return (astUtils.isClosingBraceToken(right) ||
                    !astUtils.isTokenOnSameLine(left, right) ||
                    sourceCode.isSpaceBetweenTokens(left, right) === expected);
            }
            /**
             * Check whether comments exist between the given 2 tokens.
             * @param {Token} left The left token to check.
             * @param {Token} right The right token to check.
             * @returns {boolean} `true` if comments exist between the given 2 tokens.
             */
            function commentsExistBetween(left, right) {
                return sourceCode.getFirstTokenBetween(left, right, {
                    includeComments: true,
                    filter: astUtils.isCommentToken
                }) !== null;
            }
            /**
             * Fix the spacing between the given 2 tokens.
             * @param {RuleFixer} fixer The fixer to fix.
             * @param {Token} left The left token of fix range.
             * @param {Token} right The right token of fix range.
             * @param {boolean} spacing The spacing style. `true` if there should be a space.
             * @returns {Fix|null} The fix object.
             */
            function fix(fixer, left, right, spacing) {
                if (commentsExistBetween(left, right)) {
                    return null;
                }
                if (spacing) {
                    return fixer.insertTextAfter(left, " ");
                }
                return fixer.removeRange([left.range[1], right.range[0]]);
            }
            return {
                SwitchCase(node) {
                    const colonToken = getColonToken(node);
                    const beforeToken = sourceCode.getTokenBefore(colonToken);
                    const afterToken = sourceCode.getTokenAfter(colonToken);
                    if (!isValidSpacing(beforeToken, colonToken, beforeSpacing)) {
                        context.report({
                            node,
                            loc: colonToken.loc,
                            messageId: beforeSpacing ? "expectedBefore" : "unexpectedBefore",
                            fix: fixer => fix(fixer, beforeToken, colonToken, beforeSpacing)
                        });
                    }
                    if (!isValidSpacing(colonToken, afterToken, afterSpacing)) {
                        context.report({
                            node,
                            loc: colonToken.loc,
                            messageId: afterSpacing ? "expectedAfter" : "unexpectedAfter",
                            fix: fixer => fix(fixer, colonToken, afterToken, afterSpacing)
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/symbol-description.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require symbol descriptions",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/symbol-description"
            },
            fixable: null,
            schema: [],
            messages: {
                expected: "Expected Symbol to have a description."
            }
        },
        create(context) {
            /**
             * Reports if node does not conform the rule in case rule is set to
             * report missing description
             * @param {ASTNode} node A CallExpression node to check.
             * @returns {void}
             */
            function checkArgument(node) {
                if (node.arguments.length === 0) {
                    context.report({
                        node,
                        messageId: "expected"
                    });
                }
            }
            return {
                "Program:exit"() {
                    const scope = context.getScope();
                    const variable = astUtils.getVariableByName(scope, "Symbol");
                    if (variable && variable.defs.length === 0) {
                        variable.references.forEach(reference => {
                            const node = reference.identifier;
                            if (astUtils.isCallee(node)) {
                                checkArgument(node.parent);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/template-curly-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow spacing around embedded expressions of template strings",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/template-curly-spacing"
            },
            fixable: "whitespace",
            schema: [
                { enum: ["always", "never"] }
            ],
            messages: {
                expectedBefore: "Expected space(s) before '}'.",
                expectedAfter: "Expected space(s) after '${'.",
                unexpectedBefore: "Unexpected space(s) before '}'.",
                unexpectedAfter: "Unexpected space(s) after '${'."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const always = context.options[0] === "always";
            /**
             * Checks spacing before `}` of a given token.
             * @param {Token} token A token to check. This is a Template token.
             * @returns {void}
             */
            function checkSpacingBefore(token) {
                if (!token.value.startsWith("}")) {
                    return; // starts with a backtick, this is the first template element in the template literal
                }
                const prevToken = sourceCode.getTokenBefore(token, { includeComments: true }), hasSpace = sourceCode.isSpaceBetween(prevToken, token);
                if (!astUtils.isTokenOnSameLine(prevToken, token)) {
                    return;
                }
                if (always && !hasSpace) {
                    context.report({
                        loc: {
                            start: token.loc.start,
                            end: {
                                line: token.loc.start.line,
                                column: token.loc.start.column + 1
                            }
                        },
                        messageId: "expectedBefore",
                        fix: fixer => fixer.insertTextBefore(token, " ")
                    });
                }
                if (!always && hasSpace) {
                    context.report({
                        loc: {
                            start: prevToken.loc.end,
                            end: token.loc.start
                        },
                        messageId: "unexpectedBefore",
                        fix: fixer => fixer.removeRange([prevToken.range[1], token.range[0]])
                    });
                }
            }
            /**
             * Checks spacing after `${` of a given token.
             * @param {Token} token A token to check. This is a Template token.
             * @returns {void}
             */
            function checkSpacingAfter(token) {
                if (!token.value.endsWith("${")) {
                    return; // ends with a backtick, this is the last template element in the template literal
                }
                const nextToken = sourceCode.getTokenAfter(token, { includeComments: true }), hasSpace = sourceCode.isSpaceBetween(token, nextToken);
                if (!astUtils.isTokenOnSameLine(token, nextToken)) {
                    return;
                }
                if (always && !hasSpace) {
                    context.report({
                        loc: {
                            start: {
                                line: token.loc.end.line,
                                column: token.loc.end.column - 2
                            },
                            end: token.loc.end
                        },
                        messageId: "expectedAfter",
                        fix: fixer => fixer.insertTextAfter(token, " ")
                    });
                }
                if (!always && hasSpace) {
                    context.report({
                        loc: {
                            start: token.loc.end,
                            end: nextToken.loc.start
                        },
                        messageId: "unexpectedAfter",
                        fix: fixer => fixer.removeRange([token.range[1], nextToken.range[0]])
                    });
                }
            }
            return {
                TemplateElement(node) {
                    const token = sourceCode.getFirstToken(node);
                    checkSpacingBefore(token);
                    checkSpacingAfter(token);
                }
            };
        }
    };
},
/* eslint/lib/rules/template-tag-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow spacing between template tags and their literals",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/template-tag-spacing"
            },
            fixable: "whitespace",
            schema: [
                { enum: ["always", "never"] }
            ],
            messages: {
                unexpected: "Unexpected space between template tag and template literal.",
                missing: "Missing space between template tag and template literal."
            }
        },
        create(context) {
            const never = context.options[0] !== "always";
            const sourceCode = context.getSourceCode();
            /**
             * Check if a space is present between a template tag and its literal
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkSpacing(node) {
                const tagToken = sourceCode.getTokenBefore(node.quasi);
                const literalToken = sourceCode.getFirstToken(node.quasi);
                const hasWhitespace = sourceCode.isSpaceBetweenTokens(tagToken, literalToken);
                if (never && hasWhitespace) {
                    context.report({
                        node,
                        loc: {
                            start: tagToken.loc.end,
                            end: literalToken.loc.start
                        },
                        messageId: "unexpected",
                        fix(fixer) {
                            const comments = sourceCode.getCommentsBefore(node.quasi);
                            // Don't fix anything if there's a single line comment after the template tag
                            if (comments.some(comment => comment.type === "Line")) {
                                return null;
                            }
                            return fixer.replaceTextRange([tagToken.range[1], literalToken.range[0]], comments.reduce((text, comment) => text + sourceCode.getText(comment), ""));
                        }
                    });
                }
                else if (!never && !hasWhitespace) {
                    context.report({
                        node,
                        loc: {
                            start: node.loc.start,
                            end: literalToken.loc.start
                        },
                        messageId: "missing",
                        fix(fixer) {
                            return fixer.insertTextAfter(tagToken, " ");
                        }
                    });
                }
            }
            return {
                TaggedTemplateExpression: checkSpacing
            };
        }
    };
},
/* eslint/lib/rules/unicode-bom.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow Unicode byte order mark (BOM)",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/unicode-bom"
            },
            fixable: "whitespace",
            schema: [
                {
                    enum: ["always", "never"]
                }
            ],
            messages: {
                expected: "Expected Unicode BOM (Byte Order Mark).",
                unexpected: "Unexpected Unicode BOM (Byte Order Mark)."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                Program: function checkUnicodeBOM(node) {
                    const sourceCode = context.getSourceCode(), location = { column: 0, line: 1 }, requireBOM = context.options[0] || "never";
                    if (!sourceCode.hasBOM && (requireBOM === "always")) {
                        context.report({
                            node,
                            loc: location,
                            messageId: "expected",
                            fix(fixer) {
                                return fixer.insertTextBeforeRange([0, 1], "\uFEFF");
                            }
                        });
                    }
                    else if (sourceCode.hasBOM && (requireBOM === "never")) {
                        context.report({
                            node,
                            loc: location,
                            messageId: "unexpected",
                            fix(fixer) {
                                return fixer.removeRange([-1, 0]);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/use-isnan.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /**
     * Determines if the given node is a NaN `Identifier` node.
     * @param {ASTNode|null} node The node to check.
     * @returns {boolean} `true` if the node is 'NaN' identifier.
     */
    function isNaNIdentifier(node) {
        return Boolean(node) && (astUtils.isSpecificId(node, "NaN") ||
            astUtils.isSpecificMemberAccess(node, "Number", "NaN"));
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "require calls to `isNaN()` when checking for `NaN`",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/use-isnan"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        enforceForSwitchCase: {
                            type: "boolean",
                            default: true
                        },
                        enforceForIndexOf: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                comparisonWithNaN: "Use the isNaN function to compare with NaN.",
                switchNaN: "'switch(NaN)' can never match a case clause. Use Number.isNaN instead of the switch.",
                caseNaN: "'case NaN' can never match. Use Number.isNaN before the switch.",
                indexOfNaN: "Array prototype method '{{ methodName }}' cannot find NaN."
            }
        },
        create(context) {
            const enforceForSwitchCase = !context.options[0] || context.options[0].enforceForSwitchCase;
            const enforceForIndexOf = context.options[0] && context.options[0].enforceForIndexOf;
            /**
             * Checks the given `BinaryExpression` node for `foo === NaN` and other comparisons.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             */
            function checkBinaryExpression(node) {
                if (/^(?:[<>]|[!=]=)=?$/u.test(node.operator) &&
                    (isNaNIdentifier(node.left) || isNaNIdentifier(node.right))) {
                    context.report({ node, messageId: "comparisonWithNaN" });
                }
            }
            /**
             * Checks the discriminant and all case clauses of the given `SwitchStatement` node for `switch(NaN)` and `case NaN:`
             * @param {ASTNode} node The node to check.
             * @returns {void}
             */
            function checkSwitchStatement(node) {
                if (isNaNIdentifier(node.discriminant)) {
                    context.report({ node, messageId: "switchNaN" });
                }
                for (const switchCase of node.cases) {
                    if (isNaNIdentifier(switchCase.test)) {
                        context.report({ node: switchCase, messageId: "caseNaN" });
                    }
                }
            }
            /**
             * Checks the given `CallExpression` node for `.indexOf(NaN)` and `.lastIndexOf(NaN)`.
             * @param {ASTNode} node The node to check.
             * @returns {void}
             */
            function checkCallExpression(node) {
                const callee = astUtils.skipChainExpression(node.callee);
                if (callee.type === "MemberExpression") {
                    const methodName = astUtils.getStaticPropertyName(callee);
                    if ((methodName === "indexOf" || methodName === "lastIndexOf") &&
                        node.arguments.length === 1 &&
                        isNaNIdentifier(node.arguments[0])) {
                        context.report({ node, messageId: "indexOfNaN", data: { methodName } });
                    }
                }
            }
            const listeners = {
                BinaryExpression: checkBinaryExpression
            };
            if (enforceForSwitchCase) {
                listeners.SwitchStatement = checkSwitchStatement;
            }
            if (enforceForIndexOf) {
                listeners.CallExpression = checkCallExpression;
            }
            return listeners;
        }
    };
},
/* eslint/lib/rules/valid-jsdoc.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const doctrine = require(688) /* doctrine */;
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "enforce valid JSDoc comments",
                category: "Possible Errors",
                recommended: false,
                url: "https://eslint.org/docs/rules/valid-jsdoc"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        prefer: {
                            type: "object",
                            additionalProperties: {
                                type: "string"
                            }
                        },
                        preferType: {
                            type: "object",
                            additionalProperties: {
                                type: "string"
                            }
                        },
                        requireReturn: {
                            type: "boolean",
                            default: true
                        },
                        requireParamDescription: {
                            type: "boolean",
                            default: true
                        },
                        requireReturnDescription: {
                            type: "boolean",
                            default: true
                        },
                        matchDescription: {
                            type: "string"
                        },
                        requireReturnType: {
                            type: "boolean",
                            default: true
                        },
                        requireParamType: {
                            type: "boolean",
                            default: true
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "code",
            messages: {
                unexpectedTag: "Unexpected @{{title}} tag; function has no return statement.",
                expected: "Expected JSDoc for '{{name}}' but found '{{jsdocName}}'.",
                use: "Use @{{name}} instead.",
                useType: "Use '{{expectedTypeName}}' instead of '{{currentTypeName}}'.",
                syntaxError: "JSDoc syntax error.",
                missingBrace: "JSDoc type missing brace.",
                missingParamDesc: "Missing JSDoc parameter description for '{{name}}'.",
                missingParamType: "Missing JSDoc parameter type for '{{name}}'.",
                missingReturnType: "Missing JSDoc return type.",
                missingReturnDesc: "Missing JSDoc return description.",
                missingReturn: "Missing JSDoc @{{returns}} for function.",
                missingParam: "Missing JSDoc for parameter '{{name}}'.",
                duplicateParam: "Duplicate JSDoc parameter '{{name}}'.",
                unsatisfiedDesc: "JSDoc description does not satisfy the regex pattern."
            },
            deprecated: true,
            replacedBy: []
        },
        create(context) {
            const options = context.options[0] || {}, prefer = options.prefer || {}, sourceCode = context.getSourceCode(), 
            // these both default to true, so you have to explicitly make them false
            requireReturn = options.requireReturn !== false, requireParamDescription = options.requireParamDescription !== false, requireReturnDescription = options.requireReturnDescription !== false, requireReturnType = options.requireReturnType !== false, requireParamType = options.requireParamType !== false, preferType = options.preferType || {}, checkPreferType = Object.keys(preferType).length !== 0;
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            // Using a stack to store if a function returns or not (handling nested functions)
            const fns = [];
            /**
             * Check if node type is a Class
             * @param {ASTNode} node node to check.
             * @returns {boolean} True is its a class
             * @private
             */
            function isTypeClass(node) {
                return node.type === "ClassExpression" || node.type === "ClassDeclaration";
            }
            /**
             * When parsing a new function, store it in our function stack.
             * @param {ASTNode} node A function node to check.
             * @returns {void}
             * @private
             */
            function startFunction(node) {
                fns.push({
                    returnPresent: (node.type === "ArrowFunctionExpression" && node.body.type !== "BlockStatement") ||
                        isTypeClass(node) || node.async
                });
            }
            /**
             * Indicate that return has been found in the current function.
             * @param {ASTNode} node The return node.
             * @returns {void}
             * @private
             */
            function addReturn(node) {
                const functionState = fns[fns.length - 1];
                if (functionState && node.argument !== null) {
                    functionState.returnPresent = true;
                }
            }
            /**
             * Check if return tag type is void or undefined
             * @param {Object} tag JSDoc tag
             * @returns {boolean} True if its of type void or undefined
             * @private
             */
            function isValidReturnType(tag) {
                return tag.type === null || tag.type.name === "void" || tag.type.type === "UndefinedLiteral";
            }
            /**
             * Check if type should be validated based on some exceptions
             * @param {Object} type JSDoc tag
             * @returns {boolean} True if it can be validated
             * @private
             */
            function canTypeBeValidated(type) {
                return type !== "UndefinedLiteral" && // {undefined} as there is no name property available.
                    type !== "NullLiteral" && // {null}
                    type !== "NullableLiteral" && // {?}
                    type !== "FunctionType" && // {function(a)}
                    type !== "AllLiteral"; // {*}
            }
            /**
             * Extract the current and expected type based on the input type object
             * @param {Object} type JSDoc tag
             * @returns {{currentType: Doctrine.Type, expectedTypeName: string}} The current type annotation and
             * the expected name of the annotation
             * @private
             */
            function getCurrentExpectedTypes(type) {
                let currentType;
                if (type.name) {
                    currentType = type;
                }
                else if (type.expression) {
                    currentType = type.expression;
                }
                return {
                    currentType,
                    expectedTypeName: currentType && preferType[currentType.name]
                };
            }
            /**
             * Gets the location of a JSDoc node in a file
             * @param {Token} jsdocComment The comment that this node is parsed from
             * @param {{range: number[]}} parsedJsdocNode A tag or other node which was parsed from this comment
             * @returns {{start: SourceLocation, end: SourceLocation}} The 0-based source location for the tag
             */
            function getAbsoluteRange(jsdocComment, parsedJsdocNode) {
                return {
                    start: sourceCode.getLocFromIndex(jsdocComment.range[0] + 2 + parsedJsdocNode.range[0]),
                    end: sourceCode.getLocFromIndex(jsdocComment.range[0] + 2 + parsedJsdocNode.range[1])
                };
            }
            /**
             * Validate type for a given JSDoc node
             * @param {Object} jsdocNode JSDoc node
             * @param {Object} type JSDoc tag
             * @returns {void}
             * @private
             */
            function validateType(jsdocNode, type) {
                if (!type || !canTypeBeValidated(type.type)) {
                    return;
                }
                const typesToCheck = [];
                let elements = [];
                switch (type.type) {
                    case "TypeApplication": // {Array.<String>}
                        elements = type.applications[0].type === "UnionType" ? type.applications[0].elements : type.applications;
                        typesToCheck.push(getCurrentExpectedTypes(type));
                        break;
                    case "RecordType": // {{20:String}}
                        elements = type.fields;
                        break;
                    case "UnionType": // {String|number|Test}
                    case "ArrayType": // {[String, number, Test]}
                        elements = type.elements;
                        break;
                    case "FieldType": // Array.<{count: number, votes: number}>
                        if (type.value) {
                            typesToCheck.push(getCurrentExpectedTypes(type.value));
                        }
                        break;
                    default:
                        typesToCheck.push(getCurrentExpectedTypes(type));
                }
                elements.forEach(validateType.bind(null, jsdocNode));
                typesToCheck.forEach(typeToCheck => {
                    if (typeToCheck.expectedTypeName &&
                        typeToCheck.expectedTypeName !== typeToCheck.currentType.name) {
                        context.report({
                            node: jsdocNode,
                            messageId: "useType",
                            loc: getAbsoluteRange(jsdocNode, typeToCheck.currentType),
                            data: {
                                currentTypeName: typeToCheck.currentType.name,
                                expectedTypeName: typeToCheck.expectedTypeName
                            },
                            fix(fixer) {
                                return fixer.replaceTextRange(typeToCheck.currentType.range.map(indexInComment => jsdocNode.range[0] + 2 + indexInComment), typeToCheck.expectedTypeName);
                            }
                        });
                    }
                });
            }
            /**
             * Validate the JSDoc node and output warnings if anything is wrong.
             * @param {ASTNode} node The AST node to check.
             * @returns {void}
             * @private
             */
            function checkJSDoc(node) {
                const jsdocNode = sourceCode.getJSDocComment(node), functionData = fns.pop(), paramTagsByName = Object.create(null), paramTags = [];
                let hasReturns = false, returnsTag, hasConstructor = false, isInterface = false, isOverride = false, isAbstract = false;
                // make sure only to validate JSDoc comments
                if (jsdocNode) {
                    let jsdoc;
                    try {
                        jsdoc = doctrine.parse(jsdocNode.value, {
                            strict: true,
                            unwrap: true,
                            sloppy: true,
                            range: true
                        });
                    }
                    catch (ex) {
                        if (/braces/iu.test(ex.message)) {
                            context.report({ node: jsdocNode, messageId: "missingBrace" });
                        }
                        else {
                            context.report({ node: jsdocNode, messageId: "syntaxError" });
                        }
                        return;
                    }
                    jsdoc.tags.forEach(tag => {
                        switch (tag.title.toLowerCase()) {
                            case "param":
                            case "arg":
                            case "argument":
                                paramTags.push(tag);
                                break;
                            case "return":
                            case "returns":
                                hasReturns = true;
                                returnsTag = tag;
                                break;
                            case "constructor":
                            case "class":
                                hasConstructor = true;
                                break;
                            case "override":
                            case "inheritdoc":
                                isOverride = true;
                                break;
                            case "abstract":
                            case "virtual":
                                isAbstract = true;
                                break;
                            case "interface":
                                isInterface = true;
                                break;
                            // no default
                        }
                        // check tag preferences
                        if (Object.prototype.hasOwnProperty.call(prefer, tag.title) && tag.title !== prefer[tag.title]) {
                            const entireTagRange = getAbsoluteRange(jsdocNode, tag);
                            context.report({
                                node: jsdocNode,
                                messageId: "use",
                                loc: {
                                    start: entireTagRange.start,
                                    end: {
                                        line: entireTagRange.start.line,
                                        column: entireTagRange.start.column + `@${tag.title}`.length
                                    }
                                },
                                data: { name: prefer[tag.title] },
                                fix(fixer) {
                                    return fixer.replaceTextRange([
                                        jsdocNode.range[0] + tag.range[0] + 3,
                                        jsdocNode.range[0] + tag.range[0] + tag.title.length + 3
                                    ], prefer[tag.title]);
                                }
                            });
                        }
                        // validate the types
                        if (checkPreferType && tag.type) {
                            validateType(jsdocNode, tag.type);
                        }
                    });
                    paramTags.forEach(param => {
                        if (requireParamType && !param.type) {
                            context.report({
                                node: jsdocNode,
                                messageId: "missingParamType",
                                loc: getAbsoluteRange(jsdocNode, param),
                                data: { name: param.name }
                            });
                        }
                        if (!param.description && requireParamDescription) {
                            context.report({
                                node: jsdocNode,
                                messageId: "missingParamDesc",
                                loc: getAbsoluteRange(jsdocNode, param),
                                data: { name: param.name }
                            });
                        }
                        if (paramTagsByName[param.name]) {
                            context.report({
                                node: jsdocNode,
                                messageId: "duplicateParam",
                                loc: getAbsoluteRange(jsdocNode, param),
                                data: { name: param.name }
                            });
                        }
                        else if (param.name.indexOf(".") === -1) {
                            paramTagsByName[param.name] = param;
                        }
                    });
                    if (hasReturns) {
                        if (!requireReturn && !functionData.returnPresent && (returnsTag.type === null || !isValidReturnType(returnsTag)) && !isAbstract) {
                            context.report({
                                node: jsdocNode,
                                messageId: "unexpectedTag",
                                loc: getAbsoluteRange(jsdocNode, returnsTag),
                                data: {
                                    title: returnsTag.title
                                }
                            });
                        }
                        else {
                            if (requireReturnType && !returnsTag.type) {
                                context.report({ node: jsdocNode, messageId: "missingReturnType" });
                            }
                            if (!isValidReturnType(returnsTag) && !returnsTag.description && requireReturnDescription) {
                                context.report({ node: jsdocNode, messageId: "missingReturnDesc" });
                            }
                        }
                    }
                    // check for functions missing @returns
                    if (!isOverride && !hasReturns && !hasConstructor && !isInterface &&
                        node.parent.kind !== "get" && node.parent.kind !== "constructor" &&
                        node.parent.kind !== "set" && !isTypeClass(node)) {
                        if (requireReturn || (functionData.returnPresent && !node.async)) {
                            context.report({
                                node: jsdocNode,
                                messageId: "missingReturn",
                                data: {
                                    returns: prefer.returns || "returns"
                                }
                            });
                        }
                    }
                    // check the parameters
                    const jsdocParamNames = Object.keys(paramTagsByName);
                    if (node.params) {
                        node.params.forEach((param, paramsIndex) => {
                            const bindingParam = param.type === "AssignmentPattern"
                                ? param.left
                                : param;
                            // TODO(nzakas): Figure out logical things to do with destructured, default, rest params
                            if (bindingParam.type === "Identifier") {
                                const name = bindingParam.name;
                                if (jsdocParamNames[paramsIndex] && (name !== jsdocParamNames[paramsIndex])) {
                                    context.report({
                                        node: jsdocNode,
                                        messageId: "expected",
                                        loc: getAbsoluteRange(jsdocNode, paramTagsByName[jsdocParamNames[paramsIndex]]),
                                        data: {
                                            name,
                                            jsdocName: jsdocParamNames[paramsIndex]
                                        }
                                    });
                                }
                                else if (!paramTagsByName[name] && !isOverride) {
                                    context.report({
                                        node: jsdocNode,
                                        messageId: "missingParam",
                                        data: {
                                            name
                                        }
                                    });
                                }
                            }
                        });
                    }
                    if (options.matchDescription) {
                        const regex = new RegExp(options.matchDescription, "u");
                        if (!regex.test(jsdoc.description)) {
                            context.report({ node: jsdocNode, messageId: "unsatisfiedDesc" });
                        }
                    }
                }
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                ArrowFunctionExpression: startFunction,
                FunctionExpression: startFunction,
                FunctionDeclaration: startFunction,
                ClassExpression: startFunction,
                ClassDeclaration: startFunction,
                "ArrowFunctionExpression:exit": checkJSDoc,
                "FunctionExpression:exit": checkJSDoc,
                "FunctionDeclaration:exit": checkJSDoc,
                "ClassExpression:exit": checkJSDoc,
                "ClassDeclaration:exit": checkJSDoc,
                ReturnStatement: addReturn
            };
        }
    };
},
/* doctrine/lib/doctrine.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
     * @fileoverview Main Doctrine object
     * @author Yusuke Suzuki <utatane.tea@gmail.com>
     * @author Dan Tao <daniel.tao@gmail.com>
     * @author Andrew Eisenberg <andrew@eisenberg.as>
     */
    (function () {
        'use strict';
        var typed, utility, jsdoc, esutils, hasOwnProperty;
        esutils = require(395) /* esutils */;
        typed = require(689) /* ./typed */;
        utility = require(690) /* ./utility */;
        function sliceSource(source, index, last) {
            return source.slice(index, last);
        }
        hasOwnProperty = (function () {
            var func = Object.prototype.hasOwnProperty;
            return function hasOwnProperty(obj, name) {
                return func.call(obj, name);
            };
        }());
        function shallowCopy(obj) {
            var ret = {}, key;
            for (key in obj) {
                if (obj.hasOwnProperty(key)) {
                    ret[key] = obj[key];
                }
            }
            return ret;
        }
        function isASCIIAlphanumeric(ch) {
            return (ch >= 0x61 /* 'a' */ && ch <= 0x7A /* 'z' */) ||
                (ch >= 0x41 /* 'A' */ && ch <= 0x5A /* 'Z' */) ||
                (ch >= 0x30 /* '0' */ && ch <= 0x39 /* '9' */);
        }
        function isParamTitle(title) {
            return title === 'param' || title === 'argument' || title === 'arg';
        }
        function isReturnTitle(title) {
            return title === 'return' || title === 'returns';
        }
        function isProperty(title) {
            return title === 'property' || title === 'prop';
        }
        function isNameParameterRequired(title) {
            return isParamTitle(title) || isProperty(title) ||
                title === 'alias' || title === 'this' || title === 'mixes' || title === 'requires';
        }
        function isAllowedName(title) {
            return isNameParameterRequired(title) || title === 'const' || title === 'constant';
        }
        function isAllowedNested(title) {
            return isProperty(title) || isParamTitle(title);
        }
        function isAllowedOptional(title) {
            return isProperty(title) || isParamTitle(title);
        }
        function isTypeParameterRequired(title) {
            return isParamTitle(title) || isReturnTitle(title) ||
                title === 'define' || title === 'enum' ||
                title === 'implements' || title === 'this' ||
                title === 'type' || title === 'typedef' || isProperty(title);
        }
        // Consider deprecation instead using 'isTypeParameterRequired' and 'Rules' declaration to pick when a type is optional/required
        // This would require changes to 'parseType'
        function isAllowedType(title) {
            return isTypeParameterRequired(title) || title === 'throws' || title === 'const' || title === 'constant' ||
                title === 'namespace' || title === 'member' || title === 'var' || title === 'module' ||
                title === 'constructor' || title === 'class' || title === 'extends' || title === 'augments' ||
                title === 'public' || title === 'private' || title === 'protected';
        }
        // A regex character class that contains all whitespace except linebreak characters (\r, \n, \u2028, \u2029)
        var WHITESPACE = '[ \\f\\t\\v\\u00a0\\u1680\\u180e\\u2000-\\u200a\\u202f\\u205f\\u3000\\ufeff]';
        var STAR_MATCHER = '(' + WHITESPACE + '*(?:\\*' + WHITESPACE + '?)?)(.+|[\r\n\u2028\u2029])';
        function unwrapComment(doc) {
            // JSDoc comment is following form
            //   /**
            //    * .......
            //    */
            return doc.
                // remove /**
                replace(/^\/\*\*?/, '').
                // remove */
                replace(/\*\/$/, '').
                // remove ' * ' at the beginning of a line
                replace(new RegExp(STAR_MATCHER, 'g'), '$2').
                // remove trailing whitespace
                replace(/\s*$/, '');
        }
        /**
         * Converts an index in an "unwrapped" JSDoc comment to the corresponding index in the original "wrapped" version
         * @param {string} originalSource The original wrapped comment
         * @param {number} unwrappedIndex The index of a character in the unwrapped string
         * @returns {number} The index of the corresponding character in the original wrapped string
         */
        function convertUnwrappedCommentIndex(originalSource, unwrappedIndex) {
            var replacedSource = originalSource.replace(/^\/\*\*?/, '');
            var numSkippedChars = 0;
            var matcher = new RegExp(STAR_MATCHER, 'g');
            var match;
            while ((match = matcher.exec(replacedSource))) {
                numSkippedChars += match[1].length;
                if (match.index + match[0].length > unwrappedIndex + numSkippedChars) {
                    return unwrappedIndex + numSkippedChars + originalSource.length - replacedSource.length;
                }
            }
            return originalSource.replace(/\*\/$/, '').replace(/\s*$/, '').length;
        }
        // JSDoc Tag Parser
        (function (exports) {
            var Rules, index, lineNumber, length, source, originalSource, recoverable, sloppy, strict;
            function advance() {
                var ch = source.charCodeAt(index);
                index += 1;
                if (esutils.code.isLineTerminator(ch) && !(ch === 0x0D /* '\r' */ && source.charCodeAt(index) === 0x0A /* '\n' */)) {
                    lineNumber += 1;
                }
                return String.fromCharCode(ch);
            }
            function scanTitle() {
                var title = '';
                // waste '@'
                advance();
                while (index < length && isASCIIAlphanumeric(source.charCodeAt(index))) {
                    title += advance();
                }
                return title;
            }
            function seekContent() {
                var ch, waiting, last = index;
                waiting = false;
                while (last < length) {
                    ch = source.charCodeAt(last);
                    if (esutils.code.isLineTerminator(ch) && !(ch === 0x0D /* '\r' */ && source.charCodeAt(last + 1) === 0x0A /* '\n' */)) {
                        waiting = true;
                    }
                    else if (waiting) {
                        if (ch === 0x40 /* '@' */) {
                            break;
                        }
                        if (!esutils.code.isWhiteSpace(ch)) {
                            waiting = false;
                        }
                    }
                    last += 1;
                }
                return last;
            }
            // type expression may have nest brace, such as,
            // { { ok: string } }
            //
            // therefore, scanning type expression with balancing braces.
            function parseType(title, last, addRange) {
                var ch, brace, type, startIndex, direct = false;
                // search '{'
                while (index < last) {
                    ch = source.charCodeAt(index);
                    if (esutils.code.isWhiteSpace(ch)) {
                        advance();
                    }
                    else if (ch === 0x7B /* '{' */) {
                        advance();
                        break;
                    }
                    else {
                        // this is direct pattern
                        direct = true;
                        break;
                    }
                }
                if (direct) {
                    return null;
                }
                // type expression { is found
                brace = 1;
                type = '';
                while (index < last) {
                    ch = source.charCodeAt(index);
                    if (esutils.code.isLineTerminator(ch)) {
                        advance();
                    }
                    else {
                        if (ch === 0x7D /* '}' */) {
                            brace -= 1;
                            if (brace === 0) {
                                advance();
                                break;
                            }
                        }
                        else if (ch === 0x7B /* '{' */) {
                            brace += 1;
                        }
                        if (type === '') {
                            startIndex = index;
                        }
                        type += advance();
                    }
                }
                if (brace !== 0) {
                    // braces is not balanced
                    return utility.throwError('Braces are not balanced');
                }
                if (isAllowedOptional(title)) {
                    return typed.parseParamType(type, { startIndex: convertIndex(startIndex), range: addRange });
                }
                return typed.parseType(type, { startIndex: convertIndex(startIndex), range: addRange });
            }
            function scanIdentifier(last) {
                var identifier;
                if (!esutils.code.isIdentifierStartES5(source.charCodeAt(index)) && !source[index].match(/[0-9]/)) {
                    return null;
                }
                identifier = advance();
                while (index < last && esutils.code.isIdentifierPartES5(source.charCodeAt(index))) {
                    identifier += advance();
                }
                return identifier;
            }
            function skipWhiteSpace(last) {
                while (index < last && (esutils.code.isWhiteSpace(source.charCodeAt(index)) || esutils.code.isLineTerminator(source.charCodeAt(index)))) {
                    advance();
                }
            }
            function parseName(last, allowBrackets, allowNestedParams) {
                var name = '', useBrackets, insideString;
                skipWhiteSpace(last);
                if (index >= last) {
                    return null;
                }
                if (source.charCodeAt(index) === 0x5B /* '[' */) {
                    if (allowBrackets) {
                        useBrackets = true;
                        name = advance();
                    }
                    else {
                        return null;
                    }
                }
                name += scanIdentifier(last);
                if (allowNestedParams) {
                    if (source.charCodeAt(index) === 0x3A /* ':' */ && (name === 'module' ||
                        name === 'external' ||
                        name === 'event')) {
                        name += advance();
                        name += scanIdentifier(last);
                    }
                    if (source.charCodeAt(index) === 0x5B /* '[' */ && source.charCodeAt(index + 1) === 0x5D /* ']' */) {
                        name += advance();
                        name += advance();
                    }
                    while (source.charCodeAt(index) === 0x2E /* '.' */ ||
                        source.charCodeAt(index) === 0x2F /* '/' */ ||
                        source.charCodeAt(index) === 0x23 /* '#' */ ||
                        source.charCodeAt(index) === 0x2D /* '-' */ ||
                        source.charCodeAt(index) === 0x7E /* '~' */) {
                        name += advance();
                        name += scanIdentifier(last);
                    }
                }
                if (useBrackets) {
                    skipWhiteSpace(last);
                    // do we have a default value for this?
                    if (source.charCodeAt(index) === 0x3D /* '=' */) {
                        // consume the '='' symbol
                        name += advance();
                        skipWhiteSpace(last);
                        var ch;
                        var bracketDepth = 1;
                        // scan in the default value
                        while (index < last) {
                            ch = source.charCodeAt(index);
                            if (esutils.code.isWhiteSpace(ch)) {
                                if (!insideString) {
                                    skipWhiteSpace(last);
                                    ch = source.charCodeAt(index);
                                }
                            }
                            if (ch === 0x27 /* ''' */) {
                                if (!insideString) {
                                    insideString = '\'';
                                }
                                else {
                                    if (insideString === '\'') {
                                        insideString = '';
                                    }
                                }
                            }
                            if (ch === 0x22 /* '"' */) {
                                if (!insideString) {
                                    insideString = '"';
                                }
                                else {
                                    if (insideString === '"') {
                                        insideString = '';
                                    }
                                }
                            }
                            if (ch === 0x5B /* '[' */) {
                                bracketDepth++;
                            }
                            else if (ch === 0x5D /* ']' */ &&
                                --bracketDepth === 0) {
                                break;
                            }
                            name += advance();
                        }
                    }
                    skipWhiteSpace(last);
                    if (index >= last || source.charCodeAt(index) !== 0x5D /* ']' */) {
                        // we never found a closing ']'
                        return null;
                    }
                    // collect the last ']'
                    name += advance();
                }
                return name;
            }
            function skipToTag() {
                while (index < length && source.charCodeAt(index) !== 0x40 /* '@' */) {
                    advance();
                }
                if (index >= length) {
                    return false;
                }
                utility.assert(source.charCodeAt(index) === 0x40 /* '@' */);
                return true;
            }
            function convertIndex(rangeIndex) {
                if (source === originalSource) {
                    return rangeIndex;
                }
                return convertUnwrappedCommentIndex(originalSource, rangeIndex);
            }
            function TagParser(options, title) {
                this._options = options;
                this._title = title.toLowerCase();
                this._tag = {
                    title: title,
                    description: null
                };
                if (this._options.lineNumbers) {
                    this._tag.lineNumber = lineNumber;
                }
                this._first = index - title.length - 1;
                this._last = 0;
                // space to save special information for title parsers.
                this._extra = {};
            }
            // addError(err, ...)
            TagParser.prototype.addError = function addError(errorText) {
                var args = Array.prototype.slice.call(arguments, 1), msg = errorText.replace(/%(\d)/g, function (whole, index) {
                    utility.assert(index < args.length, 'Message reference must be in range');
                    return args[index];
                });
                if (!this._tag.errors) {
                    this._tag.errors = [];
                }
                if (strict) {
                    utility.throwError(msg);
                }
                this._tag.errors.push(msg);
                return recoverable;
            };
            TagParser.prototype.parseType = function () {
                // type required titles
                if (isTypeParameterRequired(this._title)) {
                    try {
                        this._tag.type = parseType(this._title, this._last, this._options.range);
                        if (!this._tag.type) {
                            if (!isParamTitle(this._title) && !isReturnTitle(this._title)) {
                                if (!this.addError('Missing or invalid tag type')) {
                                    return false;
                                }
                            }
                        }
                    }
                    catch (error) {
                        this._tag.type = null;
                        if (!this.addError(error.message)) {
                            return false;
                        }
                    }
                }
                else if (isAllowedType(this._title)) {
                    // optional types
                    try {
                        this._tag.type = parseType(this._title, this._last, this._options.range);
                    }
                    catch (e) {
                        //For optional types, lets drop the thrown error when we hit the end of the file
                    }
                }
                return true;
            };
            TagParser.prototype._parseNamePath = function (optional) {
                var name;
                name = parseName(this._last, sloppy && isAllowedOptional(this._title), true);
                if (!name) {
                    if (!optional) {
                        if (!this.addError('Missing or invalid tag name')) {
                            return false;
                        }
                    }
                }
                this._tag.name = name;
                return true;
            };
            TagParser.prototype.parseNamePath = function () {
                return this._parseNamePath(false);
            };
            TagParser.prototype.parseNamePathOptional = function () {
                return this._parseNamePath(true);
            };
            TagParser.prototype.parseName = function () {
                var assign, name;
                // param, property requires name
                if (isAllowedName(this._title)) {
                    this._tag.name = parseName(this._last, sloppy && isAllowedOptional(this._title), isAllowedNested(this._title));
                    if (!this._tag.name) {
                        if (!isNameParameterRequired(this._title)) {
                            return true;
                        }
                        // it's possible the name has already been parsed but interpreted as a type
                        // it's also possible this is a sloppy declaration, in which case it will be
                        // fixed at the end
                        if (isParamTitle(this._title) && this._tag.type && this._tag.type.name) {
                            this._extra.name = this._tag.type;
                            this._tag.name = this._tag.type.name;
                            this._tag.type = null;
                        }
                        else {
                            if (!this.addError('Missing or invalid tag name')) {
                                return false;
                            }
                        }
                    }
                    else {
                        name = this._tag.name;
                        if (name.charAt(0) === '[' && name.charAt(name.length - 1) === ']') {
                            // extract the default value if there is one
                            // example: @param {string} [somebody=John Doe] description
                            assign = name.substring(1, name.length - 1).split('=');
                            if (assign.length > 1) {
                                this._tag['default'] = assign.slice(1).join('=');
                            }
                            this._tag.name = assign[0];
                            // convert to an optional type
                            if (this._tag.type && this._tag.type.type !== 'OptionalType') {
                                this._tag.type = {
                                    type: 'OptionalType',
                                    expression: this._tag.type
                                };
                            }
                        }
                    }
                }
                return true;
            };
            TagParser.prototype.parseDescription = function parseDescription() {
                var description = sliceSource(source, index, this._last).trim();
                if (description) {
                    if ((/^-\s+/).test(description)) {
                        description = description.substring(2);
                    }
                    this._tag.description = description;
                }
                return true;
            };
            TagParser.prototype.parseCaption = function parseDescription() {
                var description = sliceSource(source, index, this._last).trim();
                var captionStartTag = '<caption>';
                var captionEndTag = '</caption>';
                var captionStart = description.indexOf(captionStartTag);
                var captionEnd = description.indexOf(captionEndTag);
                if (captionStart >= 0 && captionEnd >= 0) {
                    this._tag.caption = description.substring(captionStart + captionStartTag.length, captionEnd).trim();
                    this._tag.description = description.substring(captionEnd + captionEndTag.length).trim();
                }
                else {
                    this._tag.description = description;
                }
                return true;
            };
            TagParser.prototype.parseKind = function parseKind() {
                var kind, kinds;
                kinds = {
                    'class': true,
                    'constant': true,
                    'event': true,
                    'external': true,
                    'file': true,
                    'function': true,
                    'member': true,
                    'mixin': true,
                    'module': true,
                    'namespace': true,
                    'typedef': true
                };
                kind = sliceSource(source, index, this._last).trim();
                this._tag.kind = kind;
                if (!hasOwnProperty(kinds, kind)) {
                    if (!this.addError('Invalid kind name \'%0\'', kind)) {
                        return false;
                    }
                }
                return true;
            };
            TagParser.prototype.parseAccess = function parseAccess() {
                var access;
                access = sliceSource(source, index, this._last).trim();
                this._tag.access = access;
                if (access !== 'private' && access !== 'protected' && access !== 'public') {
                    if (!this.addError('Invalid access name \'%0\'', access)) {
                        return false;
                    }
                }
                return true;
            };
            TagParser.prototype.parseThis = function parseThis() {
                // this name may be a name expression (e.g. {foo.bar}),
                // an union (e.g. {foo.bar|foo.baz}) or a name path (e.g. foo.bar)
                var value = sliceSource(source, index, this._last).trim();
                if (value && value.charAt(0) === '{') {
                    var gotType = this.parseType();
                    if (gotType && this._tag.type.type === 'NameExpression' || this._tag.type.type === 'UnionType') {
                        this._tag.name = this._tag.type.name;
                        return true;
                    }
                    else {
                        return this.addError('Invalid name for this');
                    }
                }
                else {
                    return this.parseNamePath();
                }
            };
            TagParser.prototype.parseVariation = function parseVariation() {
                var variation, text;
                text = sliceSource(source, index, this._last).trim();
                variation = parseFloat(text, 10);
                this._tag.variation = variation;
                if (isNaN(variation)) {
                    if (!this.addError('Invalid variation \'%0\'', text)) {
                        return false;
                    }
                }
                return true;
            };
            TagParser.prototype.ensureEnd = function () {
                var shouldBeEmpty = sliceSource(source, index, this._last).trim();
                if (shouldBeEmpty) {
                    if (!this.addError('Unknown content \'%0\'', shouldBeEmpty)) {
                        return false;
                    }
                }
                return true;
            };
            TagParser.prototype.epilogue = function epilogue() {
                var description;
                description = this._tag.description;
                // un-fix potentially sloppy declaration
                if (isAllowedOptional(this._title) && !this._tag.type && description && description.charAt(0) === '[') {
                    this._tag.type = this._extra.name;
                    if (!this._tag.name) {
                        this._tag.name = undefined;
                    }
                    if (!sloppy) {
                        if (!this.addError('Missing or invalid tag name')) {
                            return false;
                        }
                    }
                }
                return true;
            };
            Rules = {
                // http://usejsdoc.org/tags-access.html
                'access': ['parseAccess'],
                // http://usejsdoc.org/tags-alias.html
                'alias': ['parseNamePath', 'ensureEnd'],
                // http://usejsdoc.org/tags-augments.html
                'augments': ['parseType', 'parseNamePathOptional', 'ensureEnd'],
                // http://usejsdoc.org/tags-constructor.html
                'constructor': ['parseType', 'parseNamePathOptional', 'ensureEnd'],
                // Synonym: http://usejsdoc.org/tags-constructor.html
                'class': ['parseType', 'parseNamePathOptional', 'ensureEnd'],
                // Synonym: http://usejsdoc.org/tags-extends.html
                'extends': ['parseType', 'parseNamePathOptional', 'ensureEnd'],
                // http://usejsdoc.org/tags-example.html
                'example': ['parseCaption'],
                // http://usejsdoc.org/tags-deprecated.html
                'deprecated': ['parseDescription'],
                // http://usejsdoc.org/tags-global.html
                'global': ['ensureEnd'],
                // http://usejsdoc.org/tags-inner.html
                'inner': ['ensureEnd'],
                // http://usejsdoc.org/tags-instance.html
                'instance': ['ensureEnd'],
                // http://usejsdoc.org/tags-kind.html
                'kind': ['parseKind'],
                // http://usejsdoc.org/tags-mixes.html
                'mixes': ['parseNamePath', 'ensureEnd'],
                // http://usejsdoc.org/tags-mixin.html
                'mixin': ['parseNamePathOptional', 'ensureEnd'],
                // http://usejsdoc.org/tags-member.html
                'member': ['parseType', 'parseNamePathOptional', 'ensureEnd'],
                // http://usejsdoc.org/tags-method.html
                'method': ['parseNamePathOptional', 'ensureEnd'],
                // http://usejsdoc.org/tags-module.html
                'module': ['parseType', 'parseNamePathOptional', 'ensureEnd'],
                // Synonym: http://usejsdoc.org/tags-method.html
                'func': ['parseNamePathOptional', 'ensureEnd'],
                // Synonym: http://usejsdoc.org/tags-method.html
                'function': ['parseNamePathOptional', 'ensureEnd'],
                // Synonym: http://usejsdoc.org/tags-member.html
                'var': ['parseType', 'parseNamePathOptional', 'ensureEnd'],
                // http://usejsdoc.org/tags-name.html
                'name': ['parseNamePath', 'ensureEnd'],
                // http://usejsdoc.org/tags-namespace.html
                'namespace': ['parseType', 'parseNamePathOptional', 'ensureEnd'],
                // http://usejsdoc.org/tags-private.html
                'private': ['parseType', 'parseDescription'],
                // http://usejsdoc.org/tags-protected.html
                'protected': ['parseType', 'parseDescription'],
                // http://usejsdoc.org/tags-public.html
                'public': ['parseType', 'parseDescription'],
                // http://usejsdoc.org/tags-readonly.html
                'readonly': ['ensureEnd'],
                // http://usejsdoc.org/tags-requires.html
                'requires': ['parseNamePath', 'ensureEnd'],
                // http://usejsdoc.org/tags-since.html
                'since': ['parseDescription'],
                // http://usejsdoc.org/tags-static.html
                'static': ['ensureEnd'],
                // http://usejsdoc.org/tags-summary.html
                'summary': ['parseDescription'],
                // http://usejsdoc.org/tags-this.html
                'this': ['parseThis', 'ensureEnd'],
                // http://usejsdoc.org/tags-todo.html
                'todo': ['parseDescription'],
                // http://usejsdoc.org/tags-typedef.html
                'typedef': ['parseType', 'parseNamePathOptional'],
                // http://usejsdoc.org/tags-variation.html
                'variation': ['parseVariation'],
                // http://usejsdoc.org/tags-version.html
                'version': ['parseDescription']
            };
            TagParser.prototype.parse = function parse() {
                var i, iz, sequences, method;
                // empty title
                if (!this._title) {
                    if (!this.addError('Missing or invalid title')) {
                        return null;
                    }
                }
                // Seek to content last index.
                this._last = seekContent(this._title);
                if (this._options.range) {
                    this._tag.range = [this._first, source.slice(0, this._last).replace(/\s*$/, '').length].map(convertIndex);
                }
                if (hasOwnProperty(Rules, this._title)) {
                    sequences = Rules[this._title];
                }
                else {
                    // default sequences
                    sequences = ['parseType', 'parseName', 'parseDescription', 'epilogue'];
                }
                for (i = 0, iz = sequences.length; i < iz; ++i) {
                    method = sequences[i];
                    if (!this[method]()) {
                        return null;
                    }
                }
                return this._tag;
            };
            function parseTag(options) {
                var title, parser, tag;
                // skip to tag
                if (!skipToTag()) {
                    return null;
                }
                // scan title
                title = scanTitle();
                // construct tag parser
                parser = new TagParser(options, title);
                tag = parser.parse();
                // Seek global index to end of this tag.
                while (index < parser._last) {
                    advance();
                }
                return tag;
            }
            //
            // Parse JSDoc
            //
            function scanJSDocDescription(preserveWhitespace) {
                var description = '', ch, atAllowed;
                atAllowed = true;
                while (index < length) {
                    ch = source.charCodeAt(index);
                    if (atAllowed && ch === 0x40 /* '@' */) {
                        break;
                    }
                    if (esutils.code.isLineTerminator(ch)) {
                        atAllowed = true;
                    }
                    else if (atAllowed && !esutils.code.isWhiteSpace(ch)) {
                        atAllowed = false;
                    }
                    description += advance();
                }
                return preserveWhitespace ? description : description.trim();
            }
            function parse(comment, options) {
                var tags = [], tag, description, interestingTags, i, iz;
                if (options === undefined) {
                    options = {};
                }
                if (typeof options.unwrap === 'boolean' && options.unwrap) {
                    source = unwrapComment(comment);
                }
                else {
                    source = comment;
                }
                originalSource = comment;
                // array of relevant tags
                if (options.tags) {
                    if (Array.isArray(options.tags)) {
                        interestingTags = {};
                        for (i = 0, iz = options.tags.length; i < iz; i++) {
                            if (typeof options.tags[i] === 'string') {
                                interestingTags[options.tags[i]] = true;
                            }
                            else {
                                utility.throwError('Invalid "tags" parameter: ' + options.tags);
                            }
                        }
                    }
                    else {
                        utility.throwError('Invalid "tags" parameter: ' + options.tags);
                    }
                }
                length = source.length;
                index = 0;
                lineNumber = 0;
                recoverable = options.recoverable;
                sloppy = options.sloppy;
                strict = options.strict;
                description = scanJSDocDescription(options.preserveWhitespace);
                while (true) {
                    tag = parseTag(options);
                    if (!tag) {
                        break;
                    }
                    if (!interestingTags || interestingTags.hasOwnProperty(tag.title)) {
                        tags.push(tag);
                    }
                }
                return {
                    description: description,
                    tags: tags
                };
            }
            exports.parse = parse;
        }(jsdoc = {}));
        exports.version = utility.VERSION;
        exports.parse = jsdoc.parse;
        exports.parseType = typed.parseType;
        exports.parseParamType = typed.parseParamType;
        exports.unwrapComment = unwrapComment;
        exports.Syntax = shallowCopy(typed.Syntax);
        exports.Error = utility.DoctrineError;
        exports.type = {
            Syntax: exports.Syntax,
            parseType: typed.parseType,
            parseParamType: typed.parseParamType,
            stringify: typed.stringify
        };
    }());
},
/* doctrine/lib/typed.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
     * @fileoverview Type expression parser.
     * @author Yusuke Suzuki <utatane.tea@gmail.com>
     * @author Dan Tao <daniel.tao@gmail.com>
     * @author Andrew Eisenberg <andrew@eisenberg.as>
     */
    // "typed", the Type Expression Parser for doctrine.
    (function () {
        'use strict';
        var Syntax, Token, source, length, index, previous, token, value, esutils, utility, rangeOffset, addRange;
        esutils = require(395) /* esutils */;
        utility = require(690) /* ./utility */;
        Syntax = {
            NullableLiteral: 'NullableLiteral',
            AllLiteral: 'AllLiteral',
            NullLiteral: 'NullLiteral',
            UndefinedLiteral: 'UndefinedLiteral',
            VoidLiteral: 'VoidLiteral',
            UnionType: 'UnionType',
            ArrayType: 'ArrayType',
            RecordType: 'RecordType',
            FieldType: 'FieldType',
            FunctionType: 'FunctionType',
            ParameterType: 'ParameterType',
            RestType: 'RestType',
            NonNullableType: 'NonNullableType',
            OptionalType: 'OptionalType',
            NullableType: 'NullableType',
            NameExpression: 'NameExpression',
            TypeApplication: 'TypeApplication',
            StringLiteralType: 'StringLiteralType',
            NumericLiteralType: 'NumericLiteralType',
            BooleanLiteralType: 'BooleanLiteralType'
        };
        Token = {
            ILLEGAL: 0,
            DOT_LT: 1,
            REST: 2,
            LT: 3,
            GT: 4,
            LPAREN: 5,
            RPAREN: 6,
            LBRACE: 7,
            RBRACE: 8,
            LBRACK: 9,
            RBRACK: 10,
            COMMA: 11,
            COLON: 12,
            STAR: 13,
            PIPE: 14,
            QUESTION: 15,
            BANG: 16,
            EQUAL: 17,
            NAME: 18,
            STRING: 19,
            NUMBER: 20,
            EOF: 21
        };
        function isTypeName(ch) {
            return '><(){}[],:*|?!='.indexOf(String.fromCharCode(ch)) === -1 && !esutils.code.isWhiteSpace(ch) && !esutils.code.isLineTerminator(ch);
        }
        function Context(previous, index, token, value) {
            this._previous = previous;
            this._index = index;
            this._token = token;
            this._value = value;
        }
        Context.prototype.restore = function () {
            previous = this._previous;
            index = this._index;
            token = this._token;
            value = this._value;
        };
        Context.save = function () {
            return new Context(previous, index, token, value);
        };
        function maybeAddRange(node, range) {
            if (addRange) {
                node.range = [range[0] + rangeOffset, range[1] + rangeOffset];
            }
            return node;
        }
        function advance() {
            var ch = source.charAt(index);
            index += 1;
            return ch;
        }
        function scanHexEscape(prefix) {
            var i, len, ch, code = 0;
            len = (prefix === 'u') ? 4 : 2;
            for (i = 0; i < len; ++i) {
                if (index < length && esutils.code.isHexDigit(source.charCodeAt(index))) {
                    ch = advance();
                    code = code * 16 + '0123456789abcdef'.indexOf(ch.toLowerCase());
                }
                else {
                    return '';
                }
            }
            return String.fromCharCode(code);
        }
        function scanString() {
            var str = '', quote, ch, code, unescaped, restore; //TODO review removal octal = false
            quote = source.charAt(index);
            ++index;
            while (index < length) {
                ch = advance();
                if (ch === quote) {
                    quote = '';
                    break;
                }
                else if (ch === '\\') {
                    ch = advance();
                    if (!esutils.code.isLineTerminator(ch.charCodeAt(0))) {
                        switch (ch) {
                            case 'n':
                                str += '\n';
                                break;
                            case 'r':
                                str += '\r';
                                break;
                            case 't':
                                str += '\t';
                                break;
                            case 'u':
                            case 'x':
                                restore = index;
                                unescaped = scanHexEscape(ch);
                                if (unescaped) {
                                    str += unescaped;
                                }
                                else {
                                    index = restore;
                                    str += ch;
                                }
                                break;
                            case 'b':
                                str += '\b';
                                break;
                            case 'f':
                                str += '\f';
                                break;
                            case 'v':
                                str += '\v';
                                break;
                            default:
                                if (esutils.code.isOctalDigit(ch.charCodeAt(0))) {
                                    code = '01234567'.indexOf(ch);
                                    // \0 is not octal escape sequence
                                    // Deprecating unused code. TODO review removal
                                    //if (code !== 0) {
                                    //    octal = true;
                                    //}
                                    if (index < length && esutils.code.isOctalDigit(source.charCodeAt(index))) {
                                        //TODO Review Removal octal = true;
                                        code = code * 8 + '01234567'.indexOf(advance());
                                        // 3 digits are only allowed when string starts
                                        // with 0, 1, 2, 3
                                        if ('0123'.indexOf(ch) >= 0 &&
                                            index < length &&
                                            esutils.code.isOctalDigit(source.charCodeAt(index))) {
                                            code = code * 8 + '01234567'.indexOf(advance());
                                        }
                                    }
                                    str += String.fromCharCode(code);
                                }
                                else {
                                    str += ch;
                                }
                                break;
                        }
                    }
                    else {
                        if (ch === '\r' && source.charCodeAt(index) === 0x0A /* '\n' */) {
                            ++index;
                        }
                    }
                }
                else if (esutils.code.isLineTerminator(ch.charCodeAt(0))) {
                    break;
                }
                else {
                    str += ch;
                }
            }
            if (quote !== '') {
                utility.throwError('unexpected quote');
            }
            value = str;
            return Token.STRING;
        }
        function scanNumber() {
            var number, ch;
            number = '';
            ch = source.charCodeAt(index);
            if (ch !== 0x2E /* '.' */) {
                number = advance();
                ch = source.charCodeAt(index);
                if (number === '0') {
                    if (ch === 0x78 /* 'x' */ || ch === 0x58 /* 'X' */) {
                        number += advance();
                        while (index < length) {
                            ch = source.charCodeAt(index);
                            if (!esutils.code.isHexDigit(ch)) {
                                break;
                            }
                            number += advance();
                        }
                        if (number.length <= 2) {
                            // only 0x
                            utility.throwError('unexpected token');
                        }
                        if (index < length) {
                            ch = source.charCodeAt(index);
                            if (esutils.code.isIdentifierStartES5(ch)) {
                                utility.throwError('unexpected token');
                            }
                        }
                        value = parseInt(number, 16);
                        return Token.NUMBER;
                    }
                    if (esutils.code.isOctalDigit(ch)) {
                        number += advance();
                        while (index < length) {
                            ch = source.charCodeAt(index);
                            if (!esutils.code.isOctalDigit(ch)) {
                                break;
                            }
                            number += advance();
                        }
                        if (index < length) {
                            ch = source.charCodeAt(index);
                            if (esutils.code.isIdentifierStartES5(ch) || esutils.code.isDecimalDigit(ch)) {
                                utility.throwError('unexpected token');
                            }
                        }
                        value = parseInt(number, 8);
                        return Token.NUMBER;
                    }
                    if (esutils.code.isDecimalDigit(ch)) {
                        utility.throwError('unexpected token');
                    }
                }
                while (index < length) {
                    ch = source.charCodeAt(index);
                    if (!esutils.code.isDecimalDigit(ch)) {
                        break;
                    }
                    number += advance();
                }
            }
            if (ch === 0x2E /* '.' */) {
                number += advance();
                while (index < length) {
                    ch = source.charCodeAt(index);
                    if (!esutils.code.isDecimalDigit(ch)) {
                        break;
                    }
                    number += advance();
                }
            }
            if (ch === 0x65 /* 'e' */ || ch === 0x45 /* 'E' */) {
                number += advance();
                ch = source.charCodeAt(index);
                if (ch === 0x2B /* '+' */ || ch === 0x2D /* '-' */) {
                    number += advance();
                }
                ch = source.charCodeAt(index);
                if (esutils.code.isDecimalDigit(ch)) {
                    number += advance();
                    while (index < length) {
                        ch = source.charCodeAt(index);
                        if (!esutils.code.isDecimalDigit(ch)) {
                            break;
                        }
                        number += advance();
                    }
                }
                else {
                    utility.throwError('unexpected token');
                }
            }
            if (index < length) {
                ch = source.charCodeAt(index);
                if (esutils.code.isIdentifierStartES5(ch)) {
                    utility.throwError('unexpected token');
                }
            }
            value = parseFloat(number);
            return Token.NUMBER;
        }
        function scanTypeName() {
            var ch, ch2;
            value = advance();
            while (index < length && isTypeName(source.charCodeAt(index))) {
                ch = source.charCodeAt(index);
                if (ch === 0x2E /* '.' */) {
                    if ((index + 1) >= length) {
                        return Token.ILLEGAL;
                    }
                    ch2 = source.charCodeAt(index + 1);
                    if (ch2 === 0x3C /* '<' */) {
                        break;
                    }
                }
                value += advance();
            }
            return Token.NAME;
        }
        function next() {
            var ch;
            previous = index;
            while (index < length && esutils.code.isWhiteSpace(source.charCodeAt(index))) {
                advance();
            }
            if (index >= length) {
                token = Token.EOF;
                return token;
            }
            ch = source.charCodeAt(index);
            switch (ch) {
                case 0x27: /* ''' */
                case 0x22: /* '"' */
                    token = scanString();
                    return token;
                case 0x3A: /* ':' */
                    advance();
                    token = Token.COLON;
                    return token;
                case 0x2C: /* ',' */
                    advance();
                    token = Token.COMMA;
                    return token;
                case 0x28: /* '(' */
                    advance();
                    token = Token.LPAREN;
                    return token;
                case 0x29: /* ')' */
                    advance();
                    token = Token.RPAREN;
                    return token;
                case 0x5B: /* '[' */
                    advance();
                    token = Token.LBRACK;
                    return token;
                case 0x5D: /* ']' */
                    advance();
                    token = Token.RBRACK;
                    return token;
                case 0x7B: /* '{' */
                    advance();
                    token = Token.LBRACE;
                    return token;
                case 0x7D: /* '}' */
                    advance();
                    token = Token.RBRACE;
                    return token;
                case 0x2E: /* '.' */
                    if (index + 1 < length) {
                        ch = source.charCodeAt(index + 1);
                        if (ch === 0x3C /* '<' */) {
                            advance(); // '.'
                            advance(); // '<'
                            token = Token.DOT_LT;
                            return token;
                        }
                        if (ch === 0x2E /* '.' */ && index + 2 < length && source.charCodeAt(index + 2) === 0x2E /* '.' */) {
                            advance(); // '.'
                            advance(); // '.'
                            advance(); // '.'
                            token = Token.REST;
                            return token;
                        }
                        if (esutils.code.isDecimalDigit(ch)) {
                            token = scanNumber();
                            return token;
                        }
                    }
                    token = Token.ILLEGAL;
                    return token;
                case 0x3C: /* '<' */
                    advance();
                    token = Token.LT;
                    return token;
                case 0x3E: /* '>' */
                    advance();
                    token = Token.GT;
                    return token;
                case 0x2A: /* '*' */
                    advance();
                    token = Token.STAR;
                    return token;
                case 0x7C: /* '|' */
                    advance();
                    token = Token.PIPE;
                    return token;
                case 0x3F: /* '?' */
                    advance();
                    token = Token.QUESTION;
                    return token;
                case 0x21: /* '!' */
                    advance();
                    token = Token.BANG;
                    return token;
                case 0x3D: /* '=' */
                    advance();
                    token = Token.EQUAL;
                    return token;
                case 0x2D: /* '-' */
                    token = scanNumber();
                    return token;
                default:
                    if (esutils.code.isDecimalDigit(ch)) {
                        token = scanNumber();
                        return token;
                    }
                    // type string permits following case,
                    //
                    // namespace.module.MyClass
                    //
                    // this reduced 1 token TK_NAME
                    utility.assert(isTypeName(ch));
                    token = scanTypeName();
                    return token;
            }
        }
        function consume(target, text) {
            utility.assert(token === target, text || 'consumed token not matched');
            next();
        }
        function expect(target, message) {
            if (token !== target) {
                utility.throwError(message || 'unexpected token');
            }
            next();
        }
        // UnionType := '(' TypeUnionList ')'
        //
        // TypeUnionList :=
        //     <<empty>>
        //   | NonemptyTypeUnionList
        //
        // NonemptyTypeUnionList :=
        //     TypeExpression
        //   | TypeExpression '|' NonemptyTypeUnionList
        function parseUnionType() {
            var elements, startIndex = index - 1;
            consume(Token.LPAREN, 'UnionType should start with (');
            elements = [];
            if (token !== Token.RPAREN) {
                while (true) {
                    elements.push(parseTypeExpression());
                    if (token === Token.RPAREN) {
                        break;
                    }
                    expect(Token.PIPE);
                }
            }
            consume(Token.RPAREN, 'UnionType should end with )');
            return maybeAddRange({
                type: Syntax.UnionType,
                elements: elements
            }, [startIndex, previous]);
        }
        // ArrayType := '[' ElementTypeList ']'
        //
        // ElementTypeList :=
        //     <<empty>>
        //  | TypeExpression
        //  | '...' TypeExpression
        //  | TypeExpression ',' ElementTypeList
        function parseArrayType() {
            var elements, startIndex = index - 1, restStartIndex;
            consume(Token.LBRACK, 'ArrayType should start with [');
            elements = [];
            while (token !== Token.RBRACK) {
                if (token === Token.REST) {
                    restStartIndex = index - 3;
                    consume(Token.REST);
                    elements.push(maybeAddRange({
                        type: Syntax.RestType,
                        expression: parseTypeExpression()
                    }, [restStartIndex, previous]));
                    break;
                }
                else {
                    elements.push(parseTypeExpression());
                }
                if (token !== Token.RBRACK) {
                    expect(Token.COMMA);
                }
            }
            expect(Token.RBRACK);
            return maybeAddRange({
                type: Syntax.ArrayType,
                elements: elements
            }, [startIndex, previous]);
        }
        function parseFieldName() {
            var v = value;
            if (token === Token.NAME || token === Token.STRING) {
                next();
                return v;
            }
            if (token === Token.NUMBER) {
                consume(Token.NUMBER);
                return String(v);
            }
            utility.throwError('unexpected token');
        }
        // FieldType :=
        //     FieldName
        //   | FieldName ':' TypeExpression
        //
        // FieldName :=
        //     NameExpression
        //   | StringLiteral
        //   | NumberLiteral
        //   | ReservedIdentifier
        function parseFieldType() {
            var key, rangeStart = previous;
            key = parseFieldName();
            if (token === Token.COLON) {
                consume(Token.COLON);
                return maybeAddRange({
                    type: Syntax.FieldType,
                    key: key,
                    value: parseTypeExpression()
                }, [rangeStart, previous]);
            }
            return maybeAddRange({
                type: Syntax.FieldType,
                key: key,
                value: null
            }, [rangeStart, previous]);
        }
        // RecordType := '{' FieldTypeList '}'
        //
        // FieldTypeList :=
        //     <<empty>>
        //   | FieldType
        //   | FieldType ',' FieldTypeList
        function parseRecordType() {
            var fields, rangeStart = index - 1, rangeEnd;
            consume(Token.LBRACE, 'RecordType should start with {');
            fields = [];
            if (token === Token.COMMA) {
                consume(Token.COMMA);
            }
            else {
                while (token !== Token.RBRACE) {
                    fields.push(parseFieldType());
                    if (token !== Token.RBRACE) {
                        expect(Token.COMMA);
                    }
                }
            }
            rangeEnd = index;
            expect(Token.RBRACE);
            return maybeAddRange({
                type: Syntax.RecordType,
                fields: fields
            }, [rangeStart, rangeEnd]);
        }
        // NameExpression :=
        //    Identifier
        //  | TagIdentifier ':' Identifier
        //
        // Tag identifier is one of "module", "external" or "event"
        // Identifier is the same as Token.NAME, including any dots, something like
        // namespace.module.MyClass
        function parseNameExpression() {
            var name = value, rangeStart = index - name.length;
            expect(Token.NAME);
            if (token === Token.COLON && (name === 'module' ||
                name === 'external' ||
                name === 'event')) {
                consume(Token.COLON);
                name += ':' + value;
                expect(Token.NAME);
            }
            return maybeAddRange({
                type: Syntax.NameExpression,
                name: name
            }, [rangeStart, previous]);
        }
        // TypeExpressionList :=
        //     TopLevelTypeExpression
        //   | TopLevelTypeExpression ',' TypeExpressionList
        function parseTypeExpressionList() {
            var elements = [];
            elements.push(parseTop());
            while (token === Token.COMMA) {
                consume(Token.COMMA);
                elements.push(parseTop());
            }
            return elements;
        }
        // TypeName :=
        //     NameExpression
        //   | NameExpression TypeApplication
        //
        // TypeApplication :=
        //     '.<' TypeExpressionList '>'
        //   | '<' TypeExpressionList '>'   // this is extension of doctrine
        function parseTypeName() {
            var expr, applications, startIndex = index - value.length;
            expr = parseNameExpression();
            if (token === Token.DOT_LT || token === Token.LT) {
                next();
                applications = parseTypeExpressionList();
                expect(Token.GT);
                return maybeAddRange({
                    type: Syntax.TypeApplication,
                    expression: expr,
                    applications: applications
                }, [startIndex, previous]);
            }
            return expr;
        }
        // ResultType :=
        //     <<empty>>
        //   | ':' void
        //   | ':' TypeExpression
        //
        // BNF is above
        // but, we remove <<empty>> pattern, so token is always TypeToken::COLON
        function parseResultType() {
            consume(Token.COLON, 'ResultType should start with :');
            if (token === Token.NAME && value === 'void') {
                consume(Token.NAME);
                return {
                    type: Syntax.VoidLiteral
                };
            }
            return parseTypeExpression();
        }
        // ParametersType :=
        //     RestParameterType
        //   | NonRestParametersType
        //   | NonRestParametersType ',' RestParameterType
        //
        // RestParameterType :=
        //     '...'
        //     '...' Identifier
        //
        // NonRestParametersType :=
        //     ParameterType ',' NonRestParametersType
        //   | ParameterType
        //   | OptionalParametersType
        //
        // OptionalParametersType :=
        //     OptionalParameterType
        //   | OptionalParameterType, OptionalParametersType
        //
        // OptionalParameterType := ParameterType=
        //
        // ParameterType := TypeExpression | Identifier ':' TypeExpression
        //
        // Identifier is "new" or "this"
        function parseParametersType() {
            var params = [], optionalSequence = false, expr, rest = false, startIndex, restStartIndex = index - 3, nameStartIndex;
            while (token !== Token.RPAREN) {
                if (token === Token.REST) {
                    // RestParameterType
                    consume(Token.REST);
                    rest = true;
                }
                startIndex = previous;
                expr = parseTypeExpression();
                if (expr.type === Syntax.NameExpression && token === Token.COLON) {
                    nameStartIndex = previous - expr.name.length;
                    // Identifier ':' TypeExpression
                    consume(Token.COLON);
                    expr = maybeAddRange({
                        type: Syntax.ParameterType,
                        name: expr.name,
                        expression: parseTypeExpression()
                    }, [nameStartIndex, previous]);
                }
                if (token === Token.EQUAL) {
                    consume(Token.EQUAL);
                    expr = maybeAddRange({
                        type: Syntax.OptionalType,
                        expression: expr
                    }, [startIndex, previous]);
                    optionalSequence = true;
                }
                else {
                    if (optionalSequence) {
                        utility.throwError('unexpected token');
                    }
                }
                if (rest) {
                    expr = maybeAddRange({
                        type: Syntax.RestType,
                        expression: expr
                    }, [restStartIndex, previous]);
                }
                params.push(expr);
                if (token !== Token.RPAREN) {
                    expect(Token.COMMA);
                }
            }
            return params;
        }
        // FunctionType := 'function' FunctionSignatureType
        //
        // FunctionSignatureType :=
        //   | TypeParameters '(' ')' ResultType
        //   | TypeParameters '(' ParametersType ')' ResultType
        //   | TypeParameters '(' 'this' ':' TypeName ')' ResultType
        //   | TypeParameters '(' 'this' ':' TypeName ',' ParametersType ')' ResultType
        function parseFunctionType() {
            var isNew, thisBinding, params, result, fnType, startIndex = index - value.length;
            utility.assert(token === Token.NAME && value === 'function', 'FunctionType should start with \'function\'');
            consume(Token.NAME);
            // Google Closure Compiler is not implementing TypeParameters.
            // So we do not. if we don't get '(', we see it as error.
            expect(Token.LPAREN);
            isNew = false;
            params = [];
            thisBinding = null;
            if (token !== Token.RPAREN) {
                // ParametersType or 'this'
                if (token === Token.NAME &&
                    (value === 'this' || value === 'new')) {
                    // 'this' or 'new'
                    // 'new' is Closure Compiler extension
                    isNew = value === 'new';
                    consume(Token.NAME);
                    expect(Token.COLON);
                    thisBinding = parseTypeName();
                    if (token === Token.COMMA) {
                        consume(Token.COMMA);
                        params = parseParametersType();
                    }
                }
                else {
                    params = parseParametersType();
                }
            }
            expect(Token.RPAREN);
            result = null;
            if (token === Token.COLON) {
                result = parseResultType();
            }
            fnType = maybeAddRange({
                type: Syntax.FunctionType,
                params: params,
                result: result
            }, [startIndex, previous]);
            if (thisBinding) {
                // avoid adding null 'new' and 'this' properties
                fnType['this'] = thisBinding;
                if (isNew) {
                    fnType['new'] = true;
                }
            }
            return fnType;
        }
        // BasicTypeExpression :=
        //     '*'
        //   | 'null'
        //   | 'undefined'
        //   | TypeName
        //   | FunctionType
        //   | UnionType
        //   | RecordType
        //   | ArrayType
        function parseBasicTypeExpression() {
            var context, startIndex;
            switch (token) {
                case Token.STAR:
                    consume(Token.STAR);
                    return maybeAddRange({
                        type: Syntax.AllLiteral
                    }, [previous - 1, previous]);
                case Token.LPAREN:
                    return parseUnionType();
                case Token.LBRACK:
                    return parseArrayType();
                case Token.LBRACE:
                    return parseRecordType();
                case Token.NAME:
                    startIndex = index - value.length;
                    if (value === 'null') {
                        consume(Token.NAME);
                        return maybeAddRange({
                            type: Syntax.NullLiteral
                        }, [startIndex, previous]);
                    }
                    if (value === 'undefined') {
                        consume(Token.NAME);
                        return maybeAddRange({
                            type: Syntax.UndefinedLiteral
                        }, [startIndex, previous]);
                    }
                    if (value === 'true' || value === 'false') {
                        consume(Token.NAME);
                        return maybeAddRange({
                            type: Syntax.BooleanLiteralType,
                            value: value === 'true'
                        }, [startIndex, previous]);
                    }
                    context = Context.save();
                    if (value === 'function') {
                        try {
                            return parseFunctionType();
                        }
                        catch (e) {
                            context.restore();
                        }
                    }
                    return parseTypeName();
                case Token.STRING:
                    next();
                    return maybeAddRange({
                        type: Syntax.StringLiteralType,
                        value: value
                    }, [previous - value.length - 2, previous]);
                case Token.NUMBER:
                    next();
                    return maybeAddRange({
                        type: Syntax.NumericLiteralType,
                        value: value
                    }, [previous - String(value).length, previous]);
                default:
                    utility.throwError('unexpected token');
            }
        }
        // TypeExpression :=
        //     BasicTypeExpression
        //   | '?' BasicTypeExpression
        //   | '!' BasicTypeExpression
        //   | BasicTypeExpression '?'
        //   | BasicTypeExpression '!'
        //   | '?'
        //   | BasicTypeExpression '[]'
        function parseTypeExpression() {
            var expr, rangeStart;
            if (token === Token.QUESTION) {
                rangeStart = index - 1;
                consume(Token.QUESTION);
                if (token === Token.COMMA || token === Token.EQUAL || token === Token.RBRACE ||
                    token === Token.RPAREN || token === Token.PIPE || token === Token.EOF ||
                    token === Token.RBRACK || token === Token.GT) {
                    return maybeAddRange({
                        type: Syntax.NullableLiteral
                    }, [rangeStart, previous]);
                }
                return maybeAddRange({
                    type: Syntax.NullableType,
                    expression: parseBasicTypeExpression(),
                    prefix: true
                }, [rangeStart, previous]);
            }
            else if (token === Token.BANG) {
                rangeStart = index - 1;
                consume(Token.BANG);
                return maybeAddRange({
                    type: Syntax.NonNullableType,
                    expression: parseBasicTypeExpression(),
                    prefix: true
                }, [rangeStart, previous]);
            }
            else {
                rangeStart = previous;
            }
            expr = parseBasicTypeExpression();
            if (token === Token.BANG) {
                consume(Token.BANG);
                return maybeAddRange({
                    type: Syntax.NonNullableType,
                    expression: expr,
                    prefix: false
                }, [rangeStart, previous]);
            }
            if (token === Token.QUESTION) {
                consume(Token.QUESTION);
                return maybeAddRange({
                    type: Syntax.NullableType,
                    expression: expr,
                    prefix: false
                }, [rangeStart, previous]);
            }
            if (token === Token.LBRACK) {
                consume(Token.LBRACK);
                expect(Token.RBRACK, 'expected an array-style type declaration (' + value + '[])');
                return maybeAddRange({
                    type: Syntax.TypeApplication,
                    expression: maybeAddRange({
                        type: Syntax.NameExpression,
                        name: 'Array'
                    }, [rangeStart, previous]),
                    applications: [expr]
                }, [rangeStart, previous]);
            }
            return expr;
        }
        // TopLevelTypeExpression :=
        //      TypeExpression
        //    | TypeUnionList
        //
        // This rule is Google Closure Compiler extension, not ES4
        // like,
        //   { number | string }
        // If strict to ES4, we should write it as
        //   { (number|string) }
        function parseTop() {
            var expr, elements;
            expr = parseTypeExpression();
            if (token !== Token.PIPE) {
                return expr;
            }
            elements = [expr];
            consume(Token.PIPE);
            while (true) {
                elements.push(parseTypeExpression());
                if (token !== Token.PIPE) {
                    break;
                }
                consume(Token.PIPE);
            }
            return maybeAddRange({
                type: Syntax.UnionType,
                elements: elements
            }, [0, index]);
        }
        function parseTopParamType() {
            var expr;
            if (token === Token.REST) {
                consume(Token.REST);
                return maybeAddRange({
                    type: Syntax.RestType,
                    expression: parseTop()
                }, [0, index]);
            }
            expr = parseTop();
            if (token === Token.EQUAL) {
                consume(Token.EQUAL);
                return maybeAddRange({
                    type: Syntax.OptionalType,
                    expression: expr
                }, [0, index]);
            }
            return expr;
        }
        function parseType(src, opt) {
            var expr;
            source = src;
            length = source.length;
            index = 0;
            previous = 0;
            addRange = opt && opt.range;
            rangeOffset = opt && opt.startIndex || 0;
            next();
            expr = parseTop();
            if (opt && opt.midstream) {
                return {
                    expression: expr,
                    index: previous
                };
            }
            if (token !== Token.EOF) {
                utility.throwError('not reach to EOF');
            }
            return expr;
        }
        function parseParamType(src, opt) {
            var expr;
            source = src;
            length = source.length;
            index = 0;
            previous = 0;
            addRange = opt && opt.range;
            rangeOffset = opt && opt.startIndex || 0;
            next();
            expr = parseTopParamType();
            if (opt && opt.midstream) {
                return {
                    expression: expr,
                    index: previous
                };
            }
            if (token !== Token.EOF) {
                utility.throwError('not reach to EOF');
            }
            return expr;
        }
        function stringifyImpl(node, compact, topLevel) {
            var result, i, iz;
            switch (node.type) {
                case Syntax.NullableLiteral:
                    result = '?';
                    break;
                case Syntax.AllLiteral:
                    result = '*';
                    break;
                case Syntax.NullLiteral:
                    result = 'null';
                    break;
                case Syntax.UndefinedLiteral:
                    result = 'undefined';
                    break;
                case Syntax.VoidLiteral:
                    result = 'void';
                    break;
                case Syntax.UnionType:
                    if (!topLevel) {
                        result = '(';
                    }
                    else {
                        result = '';
                    }
                    for (i = 0, iz = node.elements.length; i < iz; ++i) {
                        result += stringifyImpl(node.elements[i], compact);
                        if ((i + 1) !== iz) {
                            result += compact ? '|' : ' | ';
                        }
                    }
                    if (!topLevel) {
                        result += ')';
                    }
                    break;
                case Syntax.ArrayType:
                    result = '[';
                    for (i = 0, iz = node.elements.length; i < iz; ++i) {
                        result += stringifyImpl(node.elements[i], compact);
                        if ((i + 1) !== iz) {
                            result += compact ? ',' : ', ';
                        }
                    }
                    result += ']';
                    break;
                case Syntax.RecordType:
                    result = '{';
                    for (i = 0, iz = node.fields.length; i < iz; ++i) {
                        result += stringifyImpl(node.fields[i], compact);
                        if ((i + 1) !== iz) {
                            result += compact ? ',' : ', ';
                        }
                    }
                    result += '}';
                    break;
                case Syntax.FieldType:
                    if (node.value) {
                        result = node.key + (compact ? ':' : ': ') + stringifyImpl(node.value, compact);
                    }
                    else {
                        result = node.key;
                    }
                    break;
                case Syntax.FunctionType:
                    result = compact ? 'function(' : 'function (';
                    if (node['this']) {
                        if (node['new']) {
                            result += (compact ? 'new:' : 'new: ');
                        }
                        else {
                            result += (compact ? 'this:' : 'this: ');
                        }
                        result += stringifyImpl(node['this'], compact);
                        if (node.params.length !== 0) {
                            result += compact ? ',' : ', ';
                        }
                    }
                    for (i = 0, iz = node.params.length; i < iz; ++i) {
                        result += stringifyImpl(node.params[i], compact);
                        if ((i + 1) !== iz) {
                            result += compact ? ',' : ', ';
                        }
                    }
                    result += ')';
                    if (node.result) {
                        result += (compact ? ':' : ': ') + stringifyImpl(node.result, compact);
                    }
                    break;
                case Syntax.ParameterType:
                    result = node.name + (compact ? ':' : ': ') + stringifyImpl(node.expression, compact);
                    break;
                case Syntax.RestType:
                    result = '...';
                    if (node.expression) {
                        result += stringifyImpl(node.expression, compact);
                    }
                    break;
                case Syntax.NonNullableType:
                    if (node.prefix) {
                        result = '!' + stringifyImpl(node.expression, compact);
                    }
                    else {
                        result = stringifyImpl(node.expression, compact) + '!';
                    }
                    break;
                case Syntax.OptionalType:
                    result = stringifyImpl(node.expression, compact) + '=';
                    break;
                case Syntax.NullableType:
                    if (node.prefix) {
                        result = '?' + stringifyImpl(node.expression, compact);
                    }
                    else {
                        result = stringifyImpl(node.expression, compact) + '?';
                    }
                    break;
                case Syntax.NameExpression:
                    result = node.name;
                    break;
                case Syntax.TypeApplication:
                    result = stringifyImpl(node.expression, compact) + '.<';
                    for (i = 0, iz = node.applications.length; i < iz; ++i) {
                        result += stringifyImpl(node.applications[i], compact);
                        if ((i + 1) !== iz) {
                            result += compact ? ',' : ', ';
                        }
                    }
                    result += '>';
                    break;
                case Syntax.StringLiteralType:
                    result = '"' + node.value + '"';
                    break;
                case Syntax.NumericLiteralType:
                    result = String(node.value);
                    break;
                case Syntax.BooleanLiteralType:
                    result = String(node.value);
                    break;
                default:
                    utility.throwError('Unknown type ' + node.type);
            }
            return result;
        }
        function stringify(node, options) {
            if (options == null) {
                options = {};
            }
            return stringifyImpl(node, options.compact, options.topLevel);
        }
        exports.parseType = parseType;
        exports.parseParamType = parseParamType;
        exports.stringify = stringify;
        exports.Syntax = Syntax;
    }());
},
/* doctrine/lib/utility.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
     * @fileoverview Utilities for Doctrine
     * @author Yusuke Suzuki <utatane.tea@gmail.com>
     */
    (function () {
        'use strict';
        var VERSION;
        VERSION = require(691) /* ../package.json */.version;
        exports.VERSION = VERSION;
        function DoctrineError(message) {
            this.name = 'DoctrineError';
            this.message = message;
        }
        DoctrineError.prototype = (function () {
            var Middle = function () { };
            Middle.prototype = Error.prototype;
            return new Middle();
        }());
        DoctrineError.prototype.constructor = DoctrineError;
        exports.DoctrineError = DoctrineError;
        function throwError(message) {
            throw new DoctrineError(message);
        }
        exports.throwError = throwError;
        exports.assert = require('assert');
    }());
},
/* doctrine/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "doctrine",
        "description": "JSDoc parser",
        "homepage": "https://github.com/eslint/doctrine",
        "main": "lib/doctrine.js",
        "version": "3.0.0",
        "engines": {
            "node": ">=6.0.0"
        },
        "directories": {
            "lib": "./lib"
        },
        "files": [
            "lib"
        ],
        "maintainers": [
            {
                "name": "Nicholas C. Zakas",
                "email": "nicholas+npm@nczconsulting.com",
                "web": "https://www.nczonline.net"
            },
            {
                "name": "Yusuke Suzuki",
                "email": "utatane.tea@gmail.com",
                "web": "https://github.com/Constellation"
            }
        ],
        "repository": "eslint/doctrine",
        "devDependencies": {
            "coveralls": "^3.0.1",
            "dateformat": "^1.0.11",
            "eslint": "^1.10.3",
            "eslint-release": "^1.0.0",
            "linefix": "^0.1.1",
            "mocha": "^3.4.2",
            "npm-license": "^0.3.1",
            "nyc": "^10.3.2",
            "semver": "^5.0.3",
            "shelljs": "^0.5.3",
            "shelljs-nodecli": "^0.1.1",
            "should": "^5.0.1"
        },
        "license": "Apache-2.0",
        "scripts": {
            "pretest": "npm run lint",
            "test": "nyc mocha",
            "coveralls": "nyc report --reporter=text-lcov | coveralls",
            "lint": "eslint lib/",
            "generate-release": "eslint-generate-release",
            "generate-alpharelease": "eslint-generate-prerelease alpha",
            "generate-betarelease": "eslint-generate-prerelease beta",
            "generate-rcrelease": "eslint-generate-prerelease rc",
            "publish-release": "eslint-publish-release"
        },
        "dependencies": {
            "esutils": "^2.0.2"
        }
    };
    module.exports = json;
},
/* eslint/lib/rules/valid-typeof.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "problem",
            docs: {
                description: "enforce comparing `typeof` expressions against valid strings",
                category: "Possible Errors",
                recommended: true,
                url: "https://eslint.org/docs/rules/valid-typeof"
            },
            schema: [
                {
                    type: "object",
                    properties: {
                        requireStringLiterals: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            messages: {
                invalidValue: "Invalid typeof comparison value.",
                notString: "Typeof comparisons should be to string literals."
            }
        },
        create(context) {
            const VALID_TYPES = ["symbol", "undefined", "object", "boolean", "number", "string", "function", "bigint"], OPERATORS = ["==", "===", "!=", "!=="];
            const requireStringLiterals = context.options[0] && context.options[0].requireStringLiterals;
            /**
             * Determines whether a node is a typeof expression.
             * @param {ASTNode} node The node
             * @returns {boolean} `true` if the node is a typeof expression
             */
            function isTypeofExpression(node) {
                return node.type === "UnaryExpression" && node.operator === "typeof";
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                UnaryExpression(node) {
                    if (isTypeofExpression(node)) {
                        const parent = context.getAncestors().pop();
                        if (parent.type === "BinaryExpression" && OPERATORS.indexOf(parent.operator) !== -1) {
                            const sibling = parent.left === node ? parent.right : parent.left;
                            if (sibling.type === "Literal" || sibling.type === "TemplateLiteral" && !sibling.expressions.length) {
                                const value = sibling.type === "Literal" ? sibling.value : sibling.quasis[0].value.cooked;
                                if (VALID_TYPES.indexOf(value) === -1) {
                                    context.report({ node: sibling, messageId: "invalidValue" });
                                }
                            }
                            else if (requireStringLiterals && !isTypeofExpression(sibling)) {
                                context.report({ node: sibling, messageId: "notString" });
                            }
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/vars-on-top.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: "require `var` declarations be placed at the top of their containing scope",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/vars-on-top"
            },
            schema: [],
            messages: {
                top: "All 'var' declarations must be at the top of the function scope."
            }
        },
        create(context) {
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            // eslint-disable-next-line jsdoc/require-description
            /**
             * @param {ASTNode} node any node
             * @returns {boolean} whether the given node structurally represents a directive
             */
            function looksLikeDirective(node) {
                return node.type === "ExpressionStatement" &&
                    node.expression.type === "Literal" && typeof node.expression.value === "string";
            }
            /**
             * Check to see if its a ES6 import declaration
             * @param {ASTNode} node any node
             * @returns {boolean} whether the given node represents a import declaration
             */
            function looksLikeImport(node) {
                return node.type === "ImportDeclaration" || node.type === "ImportSpecifier" ||
                    node.type === "ImportDefaultSpecifier" || node.type === "ImportNamespaceSpecifier";
            }
            /**
             * Checks whether a given node is a variable declaration or not.
             * @param {ASTNode} node any node
             * @returns {boolean} `true` if the node is a variable declaration.
             */
            function isVariableDeclaration(node) {
                return (node.type === "VariableDeclaration" ||
                    (node.type === "ExportNamedDeclaration" &&
                        node.declaration &&
                        node.declaration.type === "VariableDeclaration"));
            }
            /**
             * Checks whether this variable is on top of the block body
             * @param {ASTNode} node The node to check
             * @param {ASTNode[]} statements collection of ASTNodes for the parent node block
             * @returns {boolean} True if var is on top otherwise false
             */
            function isVarOnTop(node, statements) {
                const l = statements.length;
                let i = 0;
                // skip over directives
                for (; i < l; ++i) {
                    if (!looksLikeDirective(statements[i]) && !looksLikeImport(statements[i])) {
                        break;
                    }
                }
                for (; i < l; ++i) {
                    if (!isVariableDeclaration(statements[i])) {
                        return false;
                    }
                    if (statements[i] === node) {
                        return true;
                    }
                }
                return false;
            }
            /**
             * Checks whether variable is on top at the global level
             * @param {ASTNode} node The node to check
             * @param {ASTNode} parent Parent of the node
             * @returns {void}
             */
            function globalVarCheck(node, parent) {
                if (!isVarOnTop(node, parent.body)) {
                    context.report({ node, messageId: "top" });
                }
            }
            /**
             * Checks whether variable is on top at functional block scope level
             * @param {ASTNode} node The node to check
             * @param {ASTNode} parent Parent of the node
             * @param {ASTNode} grandParent Parent of the node's parent
             * @returns {void}
             */
            function blockScopeVarCheck(node, parent, grandParent) {
                if (!(/Function/u.test(grandParent.type) &&
                    parent.type === "BlockStatement" &&
                    isVarOnTop(node, parent.body))) {
                    context.report({ node, messageId: "top" });
                }
            }
            //--------------------------------------------------------------------------
            // Public API
            //--------------------------------------------------------------------------
            return {
                "VariableDeclaration[kind='var']"(node) {
                    if (node.parent.type === "ExportNamedDeclaration") {
                        globalVarCheck(node.parent, node.parent.parent);
                    }
                    else if (node.parent.type === "Program") {
                        globalVarCheck(node, node.parent);
                    }
                    else {
                        blockScopeVarCheck(node, node.parent, node.parent.parent);
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/wrap-iife.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    const eslintUtils = require(350) /* eslint-utils */;
    //----------------------------------------------------------------------
    // Helpers
    //----------------------------------------------------------------------
    /**
     * Check if the given node is callee of a `NewExpression` node
     * @param {ASTNode} node node to check
     * @returns {boolean} True if the node is callee of a `NewExpression` node
     * @private
     */
    function isCalleeOfNewExpression(node) {
        const maybeCallee = node.parent.type === "ChainExpression"
            ? node.parent
            : node;
        return (maybeCallee.parent.type === "NewExpression" &&
            maybeCallee.parent.callee === maybeCallee);
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require parentheses around immediate `function` invocations",
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/wrap-iife"
            },
            schema: [
                {
                    enum: ["outside", "inside", "any"]
                },
                {
                    type: "object",
                    properties: {
                        functionPrototypeMethods: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "code",
            messages: {
                wrapInvocation: "Wrap an immediate function invocation in parentheses.",
                wrapExpression: "Wrap only the function expression in parens.",
                moveInvocation: "Move the invocation into the parens that contain the function."
            }
        },
        create(context) {
            const style = context.options[0] || "outside";
            const includeFunctionPrototypeMethods = context.options[1] && context.options[1].functionPrototypeMethods;
            const sourceCode = context.getSourceCode();
            /**
             * Check if the node is wrapped in any (). All parens count: grouping parens and parens for constructs such as if()
             * @param {ASTNode} node node to evaluate
             * @returns {boolean} True if it is wrapped in any parens
             * @private
             */
            function isWrappedInAnyParens(node) {
                return astUtils.isParenthesised(sourceCode, node);
            }
            /**
             * Check if the node is wrapped in grouping (). Parens for constructs such as if() don't count
             * @param {ASTNode} node node to evaluate
             * @returns {boolean} True if it is wrapped in grouping parens
             * @private
             */
            function isWrappedInGroupingParens(node) {
                return eslintUtils.isParenthesized(1, node, sourceCode);
            }
            /**
             * Get the function node from an IIFE
             * @param {ASTNode} node node to evaluate
             * @returns {ASTNode} node that is the function expression of the given IIFE, or null if none exist
             */
            function getFunctionNodeFromIIFE(node) {
                const callee = astUtils.skipChainExpression(node.callee);
                if (callee.type === "FunctionExpression") {
                    return callee;
                }
                if (includeFunctionPrototypeMethods &&
                    callee.type === "MemberExpression" &&
                    callee.object.type === "FunctionExpression" &&
                    (astUtils.getStaticPropertyName(callee) === "call" || astUtils.getStaticPropertyName(callee) === "apply")) {
                    return callee.object;
                }
                return null;
            }
            return {
                CallExpression(node) {
                    const innerNode = getFunctionNodeFromIIFE(node);
                    if (!innerNode) {
                        return;
                    }
                    const isCallExpressionWrapped = isWrappedInAnyParens(node), isFunctionExpressionWrapped = isWrappedInAnyParens(innerNode);
                    if (!isCallExpressionWrapped && !isFunctionExpressionWrapped) {
                        context.report({
                            node,
                            messageId: "wrapInvocation",
                            fix(fixer) {
                                const nodeToSurround = style === "inside" ? innerNode : node;
                                return fixer.replaceText(nodeToSurround, `(${sourceCode.getText(nodeToSurround)})`);
                            }
                        });
                    }
                    else if (style === "inside" && !isFunctionExpressionWrapped) {
                        context.report({
                            node,
                            messageId: "wrapExpression",
                            fix(fixer) {
                                // The outer call expression will always be wrapped at this point.
                                if (isWrappedInGroupingParens(node) && !isCalleeOfNewExpression(node)) {
                                    /*
                                     * Parenthesize the function expression and remove unnecessary grouping parens around the call expression.
                                     * Replace the range between the end of the function expression and the end of the call expression.
                                     * for example, in `(function(foo) {}(bar))`, the range `(bar))` should get replaced with `)(bar)`.
                                     */
                                    const parenAfter = sourceCode.getTokenAfter(node);
                                    return fixer.replaceTextRange([innerNode.range[1], parenAfter.range[1]], `)${sourceCode.getText().slice(innerNode.range[1], parenAfter.range[0])}`);
                                }
                                /*
                                 * Call expression is wrapped in mandatory parens such as if(), or in necessary grouping parens.
                                 * These parens cannot be removed, so just parenthesize the function expression.
                                 */
                                return fixer.replaceText(innerNode, `(${sourceCode.getText(innerNode)})`);
                            }
                        });
                    }
                    else if (style === "outside" && !isCallExpressionWrapped) {
                        context.report({
                            node,
                            messageId: "moveInvocation",
                            fix(fixer) {
                                /*
                                 * The inner function expression will always be wrapped at this point.
                                 * It's only necessary to replace the range between the end of the function expression
                                 * and the call expression. For example, in `(function(foo) {})(bar)`, the range `)(bar)`
                                 * should get replaced with `(bar))`.
                                 */
                                const parenAfter = sourceCode.getTokenAfter(innerNode);
                                return fixer.replaceTextRange([parenAfter.range[0], node.range[1]], `${sourceCode.getText().slice(parenAfter.range[1], node.range[1])})`);
                            }
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/wrap-regex.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require parenthesis around regex literals",
                category: "Stylistic Issues",
                recommended: false,
                url: "https://eslint.org/docs/rules/wrap-regex"
            },
            schema: [],
            fixable: "code",
            messages: {
                requireParens: "Wrap the regexp literal in parens to disambiguate the slash."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                Literal(node) {
                    const token = sourceCode.getFirstToken(node), nodeType = token.type;
                    if (nodeType === "RegularExpression") {
                        const beforeToken = sourceCode.getTokenBefore(node);
                        const afterToken = sourceCode.getTokenAfter(node);
                        const ancestors = context.getAncestors();
                        const grandparent = ancestors[ancestors.length - 1];
                        if (grandparent.type === "MemberExpression" && grandparent.object === node &&
                            !(beforeToken && beforeToken.value === "(" && afterToken && afterToken.value === ")")) {
                            context.report({
                                node,
                                messageId: "requireParens",
                                fix: fixer => fixer.replaceText(node, `(${sourceCode.getText(node)})`)
                            });
                        }
                    }
                }
            };
        }
    };
},
/* eslint/lib/rules/yield-star-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "layout",
            docs: {
                description: "require or disallow spacing around the `*` in `yield*` expressions",
                category: "ECMAScript 6",
                recommended: false,
                url: "https://eslint.org/docs/rules/yield-star-spacing"
            },
            fixable: "whitespace",
            schema: [
                {
                    oneOf: [
                        {
                            enum: ["before", "after", "both", "neither"]
                        },
                        {
                            type: "object",
                            properties: {
                                before: { type: "boolean" },
                                after: { type: "boolean" }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            ],
            messages: {
                missingBefore: "Missing space before *.",
                missingAfter: "Missing space after *.",
                unexpectedBefore: "Unexpected space before *.",
                unexpectedAfter: "Unexpected space after *."
            }
        },
        create(context) {
            const sourceCode = context.getSourceCode();
            const mode = (function (option) {
                if (!option || typeof option === "string") {
                    return {
                        before: { before: true, after: false },
                        after: { before: false, after: true },
                        both: { before: true, after: true },
                        neither: { before: false, after: false }
                    }[option || "after"];
                }
                return option;
            }(context.options[0]));
            /**
             * Checks the spacing between two tokens before or after the star token.
             * @param {string} side Either "before" or "after".
             * @param {Token} leftToken `function` keyword token if side is "before", or
             *     star token if side is "after".
             * @param {Token} rightToken Star token if side is "before", or identifier
             *     token if side is "after".
             * @returns {void}
             */
            function checkSpacing(side, leftToken, rightToken) {
                if (sourceCode.isSpaceBetweenTokens(leftToken, rightToken) !== mode[side]) {
                    const after = leftToken.value === "*";
                    const spaceRequired = mode[side];
                    const node = after ? leftToken : rightToken;
                    let messageId = "";
                    if (spaceRequired) {
                        messageId = side === "before" ? "missingBefore" : "missingAfter";
                    }
                    else {
                        messageId = side === "before" ? "unexpectedBefore" : "unexpectedAfter";
                    }
                    context.report({
                        node,
                        messageId,
                        fix(fixer) {
                            if (spaceRequired) {
                                if (after) {
                                    return fixer.insertTextAfter(node, " ");
                                }
                                return fixer.insertTextBefore(node, " ");
                            }
                            return fixer.removeRange([leftToken.range[1], rightToken.range[0]]);
                        }
                    });
                }
            }
            /**
             * Enforces the spacing around the star if node is a yield* expression.
             * @param {ASTNode} node A yield expression node.
             * @returns {void}
             */
            function checkExpression(node) {
                if (!node.delegate) {
                    return;
                }
                const tokens = sourceCode.getFirstTokens(node, 3);
                const yieldToken = tokens[0];
                const starToken = tokens[1];
                const nextToken = tokens[2];
                checkSpacing("before", yieldToken, starToken);
                checkSpacing("after", starToken, nextToken);
            }
            return {
                YieldExpression: checkExpression
            };
        }
    };
},
/* eslint/lib/rules/yoda.js */ function _(require, module, exports, __esModule, __esExport) {
    //--------------------------------------------------------------------------
    // Requirements
    //--------------------------------------------------------------------------
    const astUtils = require(394) /* ./utils/ast-utils */;
    //--------------------------------------------------------------------------
    // Helpers
    //--------------------------------------------------------------------------
    /**
     * Determines whether an operator is a comparison operator.
     * @param {string} operator The operator to check.
     * @returns {boolean} Whether or not it is a comparison operator.
     */
    function isComparisonOperator(operator) {
        return /^(==|===|!=|!==|<|>|<=|>=)$/u.test(operator);
    }
    /**
     * Determines whether an operator is an equality operator.
     * @param {string} operator The operator to check.
     * @returns {boolean} Whether or not it is an equality operator.
     */
    function isEqualityOperator(operator) {
        return /^(==|===)$/u.test(operator);
    }
    /**
     * Determines whether an operator is one used in a range test.
     * Allowed operators are `<` and `<=`.
     * @param {string} operator The operator to check.
     * @returns {boolean} Whether the operator is used in range tests.
     */
    function isRangeTestOperator(operator) {
        return ["<", "<="].indexOf(operator) >= 0;
    }
    /**
     * Determines whether a non-Literal node is a negative number that should be
     * treated as if it were a single Literal node.
     * @param {ASTNode} node Node to test.
     * @returns {boolean} True if the node is a negative number that looks like a
     *                    real literal and should be treated as such.
     */
    function isNegativeNumericLiteral(node) {
        return (node.type === "UnaryExpression" &&
            node.operator === "-" &&
            node.prefix &&
            astUtils.isNumericLiteral(node.argument));
    }
    /**
     * Determines whether a node is a Template Literal which can be determined statically.
     * @param {ASTNode} node Node to test
     * @returns {boolean} True if the node is a Template Literal without expression.
     */
    function isStaticTemplateLiteral(node) {
        return node.type === "TemplateLiteral" && node.expressions.length === 0;
    }
    /**
     * Determines whether a non-Literal node should be treated as a single Literal node.
     * @param {ASTNode} node Node to test
     * @returns {boolean} True if the node should be treated as a single Literal node.
     */
    function looksLikeLiteral(node) {
        return isNegativeNumericLiteral(node) || isStaticTemplateLiteral(node);
    }
    /**
     * Attempts to derive a Literal node from nodes that are treated like literals.
     * @param {ASTNode} node Node to normalize.
     * @returns {ASTNode} One of the following options.
     *  1. The original node if the node is already a Literal
     *  2. A normalized Literal node with the negative number as the value if the
     *     node represents a negative number literal.
     *  3. A normalized Literal node with the string as the value if the node is
     *     a Template Literal without expression.
     *  4. Otherwise `null`.
     */
    function getNormalizedLiteral(node) {
        if (node.type === "Literal") {
            return node;
        }
        if (isNegativeNumericLiteral(node)) {
            return {
                type: "Literal",
                value: -node.argument.value,
                raw: `-${node.argument.value}`
            };
        }
        if (isStaticTemplateLiteral(node)) {
            return {
                type: "Literal",
                value: node.quasis[0].value.cooked,
                raw: node.quasis[0].value.raw
            };
        }
        return null;
    }
    //------------------------------------------------------------------------------
    // Rule Definition
    //------------------------------------------------------------------------------
    module.exports = {
        meta: {
            type: "suggestion",
            docs: {
                description: 'require or disallow "Yoda" conditions',
                category: "Best Practices",
                recommended: false,
                url: "https://eslint.org/docs/rules/yoda"
            },
            schema: [
                {
                    enum: ["always", "never"]
                },
                {
                    type: "object",
                    properties: {
                        exceptRange: {
                            type: "boolean",
                            default: false
                        },
                        onlyEquality: {
                            type: "boolean",
                            default: false
                        }
                    },
                    additionalProperties: false
                }
            ],
            fixable: "code",
            messages: {
                expected: "Expected literal to be on the {{expectedSide}} side of {{operator}}."
            }
        },
        create(context) {
            // Default to "never" (!always) if no option
            const always = context.options[0] === "always";
            const exceptRange = context.options[1] && context.options[1].exceptRange;
            const onlyEquality = context.options[1] && context.options[1].onlyEquality;
            const sourceCode = context.getSourceCode();
            /**
             * Determines whether node represents a range test.
             * A range test is a "between" test like `(0 <= x && x < 1)` or an "outside"
             * test like `(x < 0 || 1 <= x)`. It must be wrapped in parentheses, and
             * both operators must be `<` or `<=`. Finally, the literal on the left side
             * must be less than or equal to the literal on the right side so that the
             * test makes any sense.
             * @param {ASTNode} node LogicalExpression node to test.
             * @returns {boolean} Whether node is a range test.
             */
            function isRangeTest(node) {
                const left = node.left, right = node.right;
                /**
                 * Determines whether node is of the form `0 <= x && x < 1`.
                 * @returns {boolean} Whether node is a "between" range test.
                 */
                function isBetweenTest() {
                    if (node.operator === "&&" && astUtils.isSameReference(left.right, right.left)) {
                        const leftLiteral = getNormalizedLiteral(left.left);
                        const rightLiteral = getNormalizedLiteral(right.right);
                        if (leftLiteral === null && rightLiteral === null) {
                            return false;
                        }
                        if (rightLiteral === null || leftLiteral === null) {
                            return true;
                        }
                        if (leftLiteral.value <= rightLiteral.value) {
                            return true;
                        }
                    }
                    return false;
                }
                /**
                 * Determines whether node is of the form `x < 0 || 1 <= x`.
                 * @returns {boolean} Whether node is an "outside" range test.
                 */
                function isOutsideTest() {
                    if (node.operator === "||" && astUtils.isSameReference(left.left, right.right)) {
                        const leftLiteral = getNormalizedLiteral(left.right);
                        const rightLiteral = getNormalizedLiteral(right.left);
                        if (leftLiteral === null && rightLiteral === null) {
                            return false;
                        }
                        if (rightLiteral === null || leftLiteral === null) {
                            return true;
                        }
                        if (leftLiteral.value <= rightLiteral.value) {
                            return true;
                        }
                    }
                    return false;
                }
                /**
                 * Determines whether node is wrapped in parentheses.
                 * @returns {boolean} Whether node is preceded immediately by an open
                 *                    paren token and followed immediately by a close
                 *                    paren token.
                 */
                function isParenWrapped() {
                    return astUtils.isParenthesised(sourceCode, node);
                }
                return (node.type === "LogicalExpression" &&
                    left.type === "BinaryExpression" &&
                    right.type === "BinaryExpression" &&
                    isRangeTestOperator(left.operator) &&
                    isRangeTestOperator(right.operator) &&
                    (isBetweenTest() || isOutsideTest()) &&
                    isParenWrapped());
            }
            const OPERATOR_FLIP_MAP = {
                "===": "===",
                "!==": "!==",
                "==": "==",
                "!=": "!=",
                "<": ">",
                ">": "<",
                "<=": ">=",
                ">=": "<="
            };
            /**
             * Returns a string representation of a BinaryExpression node with its sides/operator flipped around.
             * @param {ASTNode} node The BinaryExpression node
             * @returns {string} A string representation of the node with the sides and operator flipped
             */
            function getFlippedString(node) {
                const operatorToken = sourceCode.getFirstTokenBetween(node.left, node.right, token => token.value === node.operator);
                const lastLeftToken = sourceCode.getTokenBefore(operatorToken);
                const firstRightToken = sourceCode.getTokenAfter(operatorToken);
                const source = sourceCode.getText();
                const leftText = source.slice(node.range[0], lastLeftToken.range[1]);
                const textBeforeOperator = source.slice(lastLeftToken.range[1], operatorToken.range[0]);
                const textAfterOperator = source.slice(operatorToken.range[1], firstRightToken.range[0]);
                const rightText = source.slice(firstRightToken.range[0], node.range[1]);
                const tokenBefore = sourceCode.getTokenBefore(node);
                const tokenAfter = sourceCode.getTokenAfter(node);
                let prefix = "";
                let suffix = "";
                if (tokenBefore &&
                    tokenBefore.range[1] === node.range[0] &&
                    !astUtils.canTokensBeAdjacent(tokenBefore, firstRightToken)) {
                    prefix = " ";
                }
                if (tokenAfter &&
                    node.range[1] === tokenAfter.range[0] &&
                    !astUtils.canTokensBeAdjacent(lastLeftToken, tokenAfter)) {
                    suffix = " ";
                }
                return (prefix +
                    rightText +
                    textBeforeOperator +
                    OPERATOR_FLIP_MAP[operatorToken.value] +
                    textAfterOperator +
                    leftText +
                    suffix);
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            return {
                BinaryExpression(node) {
                    const expectedLiteral = always ? node.left : node.right;
                    const expectedNonLiteral = always ? node.right : node.left;
                    // If `expectedLiteral` is not a literal, and `expectedNonLiteral` is a literal, raise an error.
                    if ((expectedNonLiteral.type === "Literal" ||
                        looksLikeLiteral(expectedNonLiteral)) &&
                        !(expectedLiteral.type === "Literal" ||
                            looksLikeLiteral(expectedLiteral)) &&
                        !(!isEqualityOperator(node.operator) && onlyEquality) &&
                        isComparisonOperator(node.operator) &&
                        !(exceptRange && isRangeTest(context.getAncestors().pop()))) {
                        context.report({
                            node,
                            messageId: "expected",
                            data: {
                                operator: node.operator,
                                expectedSide: always ? "left" : "right"
                            },
                            fix: fixer => fixer.replaceText(node, getFlippedString(node))
                        });
                    }
                }
            };
        }
    };
},
/* eslint/lib/linter/safe-emitter.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    /**
     * An event emitter
     * @typedef {Object} SafeEmitter
     * @property {function(eventName: string, listenerFunc: Function): void} on Adds a listener for a given event name
     * @property {function(eventName: string, arg1?: any, arg2?: any, arg3?: any)} emit Emits an event with a given name.
     * This calls all the listeners that were listening for that name, with `arg1`, `arg2`, and `arg3` as arguments.
     * @property {function(): string[]} eventNames Gets the list of event names that have registered listeners.
     */
    /**
     * Creates an object which can listen for and emit events.
     * This is similar to the EventEmitter API in Node's standard library, but it has a few differences.
     * The goal is to allow multiple modules to attach arbitrary listeners to the same emitter, without
     * letting the modules know about each other at all.
     * 1. It has no special keys like `error` and `newListener`, which would allow modules to detect when
     * another module throws an error or registers a listener.
     * 2. It calls listener functions without any `this` value. (`EventEmitter` calls listeners with a
     * `this` value of the emitter instance, which would give listeners access to other listeners.)
     * @returns {SafeEmitter} An emitter
     */
    module.exports = () => {
        const listeners = Object.create(null);
        return Object.freeze({
            on(eventName, listener) {
                if (eventName in listeners) {
                    listeners[eventName].push(listener);
                }
                else {
                    listeners[eventName] = [listener];
                }
            },
            emit(eventName, ...args) {
                if (eventName in listeners) {
                    listeners[eventName].forEach(listener => listener(...args));
                }
            },
            eventNames() {
                return Object.keys(listeners);
            }
        });
    };
},
/* eslint/lib/linter/source-code-fixer.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const debug = require(72) /* debug */("eslint:source-code-fixer");
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const BOM = "\uFEFF";
    /**
     * Compares items in a messages array by range.
     * @param {Message} a The first message.
     * @param {Message} b The second message.
     * @returns {int} -1 if a comes before b, 1 if a comes after b, 0 if equal.
     * @private
     */
    function compareMessagesByFixRange(a, b) {
        return a.fix.range[0] - b.fix.range[0] || a.fix.range[1] - b.fix.range[1];
    }
    /**
     * Compares items in a messages array by line and column.
     * @param {Message} a The first message.
     * @param {Message} b The second message.
     * @returns {int} -1 if a comes before b, 1 if a comes after b, 0 if equal.
     * @private
     */
    function compareMessagesByLocation(a, b) {
        return a.line - b.line || a.column - b.column;
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * Utility for apply fixes to source code.
     * @constructor
     */
    function SourceCodeFixer() {
        Object.freeze(this);
    }
    /**
     * Applies the fixes specified by the messages to the given text. Tries to be
     * smart about the fixes and won't apply fixes over the same area in the text.
     * @param {string} sourceText The text to apply the changes to.
     * @param {Message[]} messages The array of messages reported by ESLint.
     * @param {boolean|Function} [shouldFix=true] Determines whether each message should be fixed
     * @returns {Object} An object containing the fixed text and any unfixed messages.
     */
    SourceCodeFixer.applyFixes = function (sourceText, messages, shouldFix) {
        debug("Applying fixes");
        if (shouldFix === false) {
            debug("shouldFix parameter was false, not attempting fixes");
            return {
                fixed: false,
                messages,
                output: sourceText
            };
        }
        // clone the array
        const remainingMessages = [], fixes = [], bom = sourceText.startsWith(BOM) ? BOM : "", text = bom ? sourceText.slice(1) : sourceText;
        let lastPos = Number.NEGATIVE_INFINITY, output = bom;
        /**
         * Try to use the 'fix' from a problem.
         * @param   {Message} problem The message object to apply fixes from
         * @returns {boolean}         Whether fix was successfully applied
         */
        function attemptFix(problem) {
            const fix = problem.fix;
            const start = fix.range[0];
            const end = fix.range[1];
            // Remain it as a problem if it's overlapped or it's a negative range
            if (lastPos >= start || start > end) {
                remainingMessages.push(problem);
                return false;
            }
            // Remove BOM.
            if ((start < 0 && end >= 0) || (start === 0 && fix.text.startsWith(BOM))) {
                output = "";
            }
            // Make output to this fix.
            output += text.slice(Math.max(0, lastPos), Math.max(0, start));
            output += fix.text;
            lastPos = end;
            return true;
        }
        messages.forEach(problem => {
            if (Object.prototype.hasOwnProperty.call(problem, "fix")) {
                fixes.push(problem);
            }
            else {
                remainingMessages.push(problem);
            }
        });
        if (fixes.length) {
            debug("Found fixes to apply");
            let fixesWereApplied = false;
            for (const problem of fixes.sort(compareMessagesByFixRange)) {
                if (typeof shouldFix !== "function" || shouldFix(problem)) {
                    attemptFix(problem);
                    /*
                     * The only time attemptFix will fail is if a previous fix was
                     * applied which conflicts with it.  So we can mark this as true.
                     */
                    fixesWereApplied = true;
                }
                else {
                    remainingMessages.push(problem);
                }
            }
            output += text.slice(Math.max(0, lastPos));
            return {
                fixed: fixesWereApplied,
                messages: remainingMessages.sort(compareMessagesByLocation),
                output
            };
        }
        debug("No fixes to apply");
        return {
            fixed: false,
            messages,
            output: bom + text
        };
    };
    module.exports = SourceCodeFixer;
},
/* eslint/lib/linter/timing.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    /* istanbul ignore next */
    /**
     * Align the string to left
     * @param {string} str string to evaluate
     * @param {int} len length of the string
     * @param {string} ch delimiter character
     * @returns {string} modified string
     * @private
     */
    function alignLeft(str, len, ch) {
        return str + new Array(len - str.length + 1).join(ch || " ");
    }
    /* istanbul ignore next */
    /**
     * Align the string to right
     * @param {string} str string to evaluate
     * @param {int} len length of the string
     * @param {string} ch delimiter character
     * @returns {string} modified string
     * @private
     */
    function alignRight(str, len, ch) {
        return new Array(len - str.length + 1).join(ch || " ") + str;
    }
    //------------------------------------------------------------------------------
    // Module definition
    //------------------------------------------------------------------------------
    const enabled = !!process.env.TIMING;
    const HEADERS = ["Rule", "Time (ms)", "Relative"];
    const ALIGN = [alignLeft, alignRight, alignRight];
    /**
     * Decide how many rules to show in the output list.
     * @returns {number} the number of rules to show
     */
    function getListSize() {
        const MINIMUM_SIZE = 10;
        if (typeof process.env.TIMING !== "string") {
            return MINIMUM_SIZE;
        }
        if (process.env.TIMING.toLowerCase() === "all") {
            return Number.POSITIVE_INFINITY;
        }
        const TIMING_ENV_VAR_AS_INTEGER = Number.parseInt(process.env.TIMING, 10);
        return TIMING_ENV_VAR_AS_INTEGER > 10 ? TIMING_ENV_VAR_AS_INTEGER : MINIMUM_SIZE;
    }
    /* istanbul ignore next */
    /**
     * display the data
     * @param {Object} data Data object to be displayed
     * @returns {void} prints modified string with console.log
     * @private
     */
    function display(data) {
        let total = 0;
        const rows = Object.keys(data)
            .map(key => {
            const time = data[key];
            total += time;
            return [key, time];
        })
            .sort((a, b) => b[1] - a[1])
            .slice(0, getListSize());
        rows.forEach(row => {
            row.push(`${(row[1] * 100 / total).toFixed(1)}%`);
            row[1] = row[1].toFixed(3);
        });
        rows.unshift(HEADERS);
        const widths = [];
        rows.forEach(row => {
            const len = row.length;
            for (let i = 0; i < len; i++) {
                const n = row[i].length;
                if (!widths[i] || n > widths[i]) {
                    widths[i] = n;
                }
            }
        });
        const table = rows.map(row => (row
            .map((cell, index) => ALIGN[index](cell, widths[index]))
            .join(" | ")));
        table.splice(1, 0, widths.map((width, index) => {
            const extraAlignment = index !== 0 && index !== widths.length - 1 ? 2 : 1;
            return ALIGN[index](":", width + extraAlignment, "-");
        }).join("|"));
        console.log(table.join("\n")); // eslint-disable-line no-console
    }
    /* istanbul ignore next */
    module.exports = (function () {
        const data = Object.create(null);
        /**
         * Time the run
         * @param {*} key key from the data object
         * @param {Function} fn function to be called
         * @returns {Function} function to be executed
         * @private
         */
        function time(key, fn) {
            if (typeof data[key] === "undefined") {
                data[key] = 0;
            }
            return function (...args) {
                let t = process.hrtime();
                fn(...args);
                t = process.hrtime(t);
                data[key] += t[0] * 1e3 + t[1] / 1e6;
            };
        }
        if (enabled) {
            process.on("exit", () => {
                display(data);
            });
        }
        return {
            time,
            enabled,
            getListSize
        };
    }());
},
/* eslint/conf/replacements.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "rules": {
            "generator-star": ["generator-star-spacing"],
            "global-strict": ["strict"],
            "no-arrow-condition": ["no-confusing-arrow", "no-constant-condition"],
            "no-comma-dangle": ["comma-dangle"],
            "no-empty-class": ["no-empty-character-class"],
            "no-empty-label": ["no-labels"],
            "no-extra-strict": ["strict"],
            "no-reserved-keys": ["quote-props"],
            "no-space-before-semi": ["semi-spacing"],
            "no-wrap-func": ["no-extra-parens"],
            "space-after-function-name": ["space-before-function-paren"],
            "space-after-keywords": ["keyword-spacing"],
            "space-before-function-parentheses": ["space-before-function-paren"],
            "space-before-keywords": ["keyword-spacing"],
            "space-in-brackets": ["object-curly-spacing", "array-bracket-spacing", "computed-property-spacing"],
            "space-return-throw-case": ["keyword-spacing"],
            "space-unary-word-ops": ["space-unary-ops"],
            "spaced-line-comment": ["spaced-comment"]
        }
    };
    module.exports = json;
},
/* eslint/lib/cli-engine/load-rules.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const fs = require("fs"), path = require("path");
    const rulesDirCache = {};
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    /**
     * Load all rule modules from specified directory.
     * @param {string} relativeRulesDir Path to rules directory, may be relative.
     * @param {string} cwd Current working directory
     * @returns {Object} Loaded rule modules.
     */
    module.exports = function (relativeRulesDir, cwd) {
        const rulesDir = path.resolve(cwd, relativeRulesDir);
        // cache will help performance as IO operation are expensive
        if (rulesDirCache[rulesDir]) {
            return rulesDirCache[rulesDir];
        }
        const rules = Object.create(null);
        fs.readdirSync(rulesDir).forEach(file => {
            if (path.extname(file) !== ".js") {
                return;
            }
            rules[file.slice(0, -3)] = require(path.join(rulesDir, file));
        });
        rulesDirCache[rulesDir] = rules;
        return rules;
    };
},
/* eslint/lib/cli-engine/hash.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const murmur = require(704) /* imurmurhash */;
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    //------------------------------------------------------------------------------
    // Private
    //------------------------------------------------------------------------------
    /**
     * hash the given string
     * @param  {string} str the string to hash
     * @returns {string}    the hash
     */
    function hash(str) {
        return murmur(str).result().toString(36);
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = hash;
},
/* imurmurhash/imurmurhash.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * @preserve
     * JS Implementation of incremental MurmurHash3 (r150) (as of May 10, 2013)
     *
     * @author <a href="mailto:jensyt@gmail.com">Jens Taylor</a>
     * @see http://github.com/homebrewing/brauhaus-diff
     * @author <a href="mailto:gary.court@gmail.com">Gary Court</a>
     * @see http://github.com/garycourt/murmurhash-js
     * @author <a href="mailto:aappleby@gmail.com">Austin Appleby</a>
     * @see http://sites.google.com/site/murmurhash/
     */
    (function () {
        var cache;
        // Call this function without `new` to use the cached object (good for
        // single-threaded environments), or with `new` to create a new object.
        //
        // @param {string} key A UTF-16 or ASCII string
        // @param {number} seed An optional positive integer
        // @return {object} A MurmurHash3 object for incremental hashing
        function MurmurHash3(key, seed) {
            var m = this instanceof MurmurHash3 ? this : cache;
            m.reset(seed);
            if (typeof key === 'string' && key.length > 0) {
                m.hash(key);
            }
            if (m !== this) {
                return m;
            }
        }
        ;
        // Incrementally add a string to this hash
        //
        // @param {string} key A UTF-16 or ASCII string
        // @return {object} this
        MurmurHash3.prototype.hash = function (key) {
            var h1, k1, i, top, len;
            len = key.length;
            this.len += len;
            k1 = this.k1;
            i = 0;
            switch (this.rem) {
                case 0: k1 ^= len > i ? (key.charCodeAt(i++) & 0xffff) : 0;
                case 1: k1 ^= len > i ? (key.charCodeAt(i++) & 0xffff) << 8 : 0;
                case 2: k1 ^= len > i ? (key.charCodeAt(i++) & 0xffff) << 16 : 0;
                case 3:
                    k1 ^= len > i ? (key.charCodeAt(i) & 0xff) << 24 : 0;
                    k1 ^= len > i ? (key.charCodeAt(i++) & 0xff00) >> 8 : 0;
            }
            this.rem = (len + this.rem) & 3; // & 3 is same as % 4
            len -= this.rem;
            if (len > 0) {
                h1 = this.h1;
                while (1) {
                    k1 = (k1 * 0x2d51 + (k1 & 0xffff) * 0xcc9e0000) & 0xffffffff;
                    k1 = (k1 << 15) | (k1 >>> 17);
                    k1 = (k1 * 0x3593 + (k1 & 0xffff) * 0x1b870000) & 0xffffffff;
                    h1 ^= k1;
                    h1 = (h1 << 13) | (h1 >>> 19);
                    h1 = (h1 * 5 + 0xe6546b64) & 0xffffffff;
                    if (i >= len) {
                        break;
                    }
                    k1 = ((key.charCodeAt(i++) & 0xffff)) ^
                        ((key.charCodeAt(i++) & 0xffff) << 8) ^
                        ((key.charCodeAt(i++) & 0xffff) << 16);
                    top = key.charCodeAt(i++);
                    k1 ^= ((top & 0xff) << 24) ^
                        ((top & 0xff00) >> 8);
                }
                k1 = 0;
                switch (this.rem) {
                    case 3: k1 ^= (key.charCodeAt(i + 2) & 0xffff) << 16;
                    case 2: k1 ^= (key.charCodeAt(i + 1) & 0xffff) << 8;
                    case 1: k1 ^= (key.charCodeAt(i) & 0xffff);
                }
                this.h1 = h1;
            }
            this.k1 = k1;
            return this;
        };
        // Get the result of this hash
        //
        // @return {number} The 32-bit hash
        MurmurHash3.prototype.result = function () {
            var k1, h1;
            k1 = this.k1;
            h1 = this.h1;
            if (k1 > 0) {
                k1 = (k1 * 0x2d51 + (k1 & 0xffff) * 0xcc9e0000) & 0xffffffff;
                k1 = (k1 << 15) | (k1 >>> 17);
                k1 = (k1 * 0x3593 + (k1 & 0xffff) * 0x1b870000) & 0xffffffff;
                h1 ^= k1;
            }
            h1 ^= this.len;
            h1 ^= h1 >>> 16;
            h1 = (h1 * 0xca6b + (h1 & 0xffff) * 0x85eb0000) & 0xffffffff;
            h1 ^= h1 >>> 13;
            h1 = (h1 * 0xae35 + (h1 & 0xffff) * 0xc2b20000) & 0xffffffff;
            h1 ^= h1 >>> 16;
            return h1 >>> 0;
        };
        // Reset the hash object for reuse
        //
        // @param {number} seed An optional positive integer
        MurmurHash3.prototype.reset = function (seed) {
            this.h1 = typeof seed === 'number' ? seed : 0;
            this.rem = this.k1 = this.len = 0;
            return this;
        };
        // A cached object to use. This can be safely used if you're in a single-
        // threaded environment, otherwise you need to create new hashes to use.
        cache = new MurmurHash3();
        if (typeof (module) != 'undefined') {
            module.exports = MurmurHash3;
        }
        else {
            this.MurmurHash3 = MurmurHash3;
        }
    }());
},
/* eslint/lib/cli-engine/lint-result-cache.js */ function _(require, module, exports, __esModule, __esExport) {
    //-----------------------------------------------------------------------------
    // Requirements
    //-----------------------------------------------------------------------------
    const assert = require("assert");
    const fs = require("fs");
    const fileEntryCache = require(706) /* file-entry-cache */;
    const stringify = require(723) /* json-stable-stringify-without-jsonify */;
    const pkg = require(207) /* ../../package.json */;
    const hash = require(703) /* ./hash */;
    const debug = require(72) /* debug */("eslint:lint-result-cache");
    //-----------------------------------------------------------------------------
    // Helpers
    //-----------------------------------------------------------------------------
    const configHashCache = new WeakMap();
    const nodeVersion = process && process.version;
    const validCacheStrategies = ["metadata", "content"];
    const invalidCacheStrategyErrorMessage = `Cache strategy must be one of: ${validCacheStrategies
        .map(strategy => `"${strategy}"`)
        .join(", ")}`;
    /**
     * Tests whether a provided cacheStrategy is valid
     * @param {string} cacheStrategy The cache strategy to use
     * @returns {boolean} true if `cacheStrategy` is one of `validCacheStrategies`; false otherwise
     */
    function isValidCacheStrategy(cacheStrategy) {
        return (validCacheStrategies.indexOf(cacheStrategy) !== -1);
    }
    /**
     * Calculates the hash of the config
     * @param {ConfigArray} config The config.
     * @returns {string} The hash of the config
     */
    function hashOfConfigFor(config) {
        if (!configHashCache.has(config)) {
            configHashCache.set(config, hash(`${pkg.version}_${nodeVersion}_${stringify(config)}`));
        }
        return configHashCache.get(config);
    }
    //-----------------------------------------------------------------------------
    // Public Interface
    //-----------------------------------------------------------------------------
    /**
     * Lint result cache. This wraps around the file-entry-cache module,
     * transparently removing properties that are difficult or expensive to
     * serialize and adding them back in on retrieval.
     */
    class LintResultCache {
        /**
         * Creates a new LintResultCache instance.
         * @param {string} cacheFileLocation The cache file location.
         * @param {"metadata" | "content"} cacheStrategy The cache strategy to use.
         */
        constructor(cacheFileLocation, cacheStrategy) {
            assert(cacheFileLocation, "Cache file location is required");
            assert(cacheStrategy, "Cache strategy is required");
            assert(isValidCacheStrategy(cacheStrategy), invalidCacheStrategyErrorMessage);
            debug(`Caching results to ${cacheFileLocation}`);
            const useChecksum = cacheStrategy === "content";
            debug(`Using "${cacheStrategy}" strategy to detect changes`);
            this.fileEntryCache = fileEntryCache.create(cacheFileLocation, void 0, useChecksum);
            this.cacheFileLocation = cacheFileLocation;
        }
        /**
         * Retrieve cached lint results for a given file path, if present in the
         * cache. If the file is present and has not been changed, rebuild any
         * missing result information.
         * @param {string} filePath The file for which to retrieve lint results.
         * @param {ConfigArray} config The config of the file.
         * @returns {Object|null} The rebuilt lint results, or null if the file is
         *   changed or not in the filesystem.
         */
        getCachedLintResults(filePath, config) {
            /*
             * Cached lint results are valid if and only if:
             * 1. The file is present in the filesystem
             * 2. The file has not changed since the time it was previously linted
             * 3. The ESLint configuration has not changed since the time the file
             *    was previously linted
             * If any of these are not true, we will not reuse the lint results.
             */
            const fileDescriptor = this.fileEntryCache.getFileDescriptor(filePath);
            const hashOfConfig = hashOfConfigFor(config);
            const changed = fileDescriptor.changed ||
                fileDescriptor.meta.hashOfConfig !== hashOfConfig;
            if (fileDescriptor.notFound) {
                debug(`File not found on the file system: ${filePath}`);
                return null;
            }
            if (changed) {
                debug(`Cache entry not found or no longer valid: ${filePath}`);
                return null;
            }
            // If source is present but null, need to reread the file from the filesystem.
            if (fileDescriptor.meta.results &&
                fileDescriptor.meta.results.source === null) {
                debug(`Rereading cached result source from filesystem: ${filePath}`);
                fileDescriptor.meta.results.source = fs.readFileSync(filePath, "utf-8");
            }
            return fileDescriptor.meta.results;
        }
        /**
         * Set the cached lint results for a given file path, after removing any
         * information that will be both unnecessary and difficult to serialize.
         * Avoids caching results with an "output" property (meaning fixes were
         * applied), to prevent potentially incorrect results if fixes are not
         * written to disk.
         * @param {string} filePath The file for which to set lint results.
         * @param {ConfigArray} config The config of the file.
         * @param {Object} result The lint result to be set for the file.
         * @returns {void}
         */
        setCachedLintResults(filePath, config, result) {
            if (result && Object.prototype.hasOwnProperty.call(result, "output")) {
                return;
            }
            const fileDescriptor = this.fileEntryCache.getFileDescriptor(filePath);
            if (fileDescriptor && !fileDescriptor.notFound) {
                debug(`Updating cached result: ${filePath}`);
                // Serialize the result, except that we want to remove the file source if present.
                const resultToSerialize = Object.assign({}, result);
                /*
                 * Set result.source to null.
                 * In `getCachedLintResults`, if source is explicitly null, we will
                 * read the file from the filesystem to set the value again.
                 */
                if (Object.prototype.hasOwnProperty.call(resultToSerialize, "source")) {
                    resultToSerialize.source = null;
                }
                fileDescriptor.meta.results = resultToSerialize;
                fileDescriptor.meta.hashOfConfig = hashOfConfigFor(config);
            }
        }
        /**
         * Persists the in-memory cache to disk.
         * @returns {void}
         */
        reconcile() {
            debug(`Persisting cached results: ${this.cacheFileLocation}`);
            this.fileEntryCache.reconcile();
        }
    }
    module.exports = LintResultCache;
},
/* file-entry-cache/cache.js */ function _(require, module, exports, __esModule, __esExport) {
    var path = require('path');
    var crypto = require('crypto');
    module.exports = {
        createFromFile: function (filePath, useChecksum) {
            var fname = path.basename(filePath);
            var dir = path.dirname(filePath);
            return this.create(fname, dir, useChecksum);
        },
        create: function (cacheId, _path, useChecksum) {
            var fs = require('fs');
            var flatCache = require(707) /* flat-cache */;
            var cache = flatCache.load(cacheId, _path);
            var normalizedEntries = {};
            var removeNotFoundFiles = function removeNotFoundFiles() {
                const cachedEntries = cache.keys();
                // remove not found entries
                cachedEntries.forEach(function remover(fPath) {
                    try {
                        fs.statSync(fPath);
                    }
                    catch (err) {
                        if (err.code === 'ENOENT') {
                            cache.removeKey(fPath);
                        }
                    }
                });
            };
            removeNotFoundFiles();
            return {
                /**
                 * the flat cache storage used to persist the metadata of the `files
                 * @type {Object}
                 */
                cache: cache,
                /**
                 * Given a buffer, calculate md5 hash of its content.
                 * @method getHash
                 * @param  {Buffer} buffer   buffer to calculate hash on
                 * @return {String}          content hash digest
                 */
                getHash: function (buffer) {
                    return crypto.createHash('md5').update(buffer).digest('hex');
                },
                /**
                 * Return whether or not a file has changed since last time reconcile was called.
                 * @method hasFileChanged
                 * @param  {String}  file  the filepath to check
                 * @return {Boolean}       wheter or not the file has changed
                 */
                hasFileChanged: function (file) {
                    return this.getFileDescriptor(file).changed;
                },
                /**
                 * given an array of file paths it return and object with three arrays:
                 *  - changedFiles: Files that changed since previous run
                 *  - notChangedFiles: Files that haven't change
                 *  - notFoundFiles: Files that were not found, probably deleted
                 *
                 * @param  {Array} files the files to analyze and compare to the previous seen files
                 * @return {[type]}       [description]
                 */
                analyzeFiles: function (files) {
                    var me = this;
                    files = files || [];
                    var res = {
                        changedFiles: [],
                        notFoundFiles: [],
                        notChangedFiles: [],
                    };
                    me.normalizeEntries(files).forEach(function (entry) {
                        if (entry.changed) {
                            res.changedFiles.push(entry.key);
                            return;
                        }
                        if (entry.notFound) {
                            res.notFoundFiles.push(entry.key);
                            return;
                        }
                        res.notChangedFiles.push(entry.key);
                    });
                    return res;
                },
                getFileDescriptor: function (file) {
                    var fstat;
                    try {
                        fstat = fs.statSync(file);
                    }
                    catch (ex) {
                        this.removeEntry(file);
                        return { key: file, notFound: true, err: ex };
                    }
                    if (useChecksum) {
                        return this._getFileDescriptorUsingChecksum(file);
                    }
                    return this._getFileDescriptorUsingMtimeAndSize(file, fstat);
                },
                _getFileDescriptorUsingMtimeAndSize: function (file, fstat) {
                    var meta = cache.getKey(file);
                    var cacheExists = !!meta;
                    var cSize = fstat.size;
                    var cTime = fstat.mtime.getTime();
                    var isDifferentDate;
                    var isDifferentSize;
                    if (!meta) {
                        meta = { size: cSize, mtime: cTime };
                    }
                    else {
                        isDifferentDate = cTime !== meta.mtime;
                        isDifferentSize = cSize !== meta.size;
                    }
                    var nEntry = (normalizedEntries[file] = {
                        key: file,
                        changed: !cacheExists || isDifferentDate || isDifferentSize,
                        meta: meta,
                    });
                    return nEntry;
                },
                _getFileDescriptorUsingChecksum: function (file) {
                    var meta = cache.getKey(file);
                    var cacheExists = !!meta;
                    var contentBuffer;
                    try {
                        contentBuffer = fs.readFileSync(file);
                    }
                    catch (ex) {
                        contentBuffer = '';
                    }
                    var isDifferent = true;
                    var hash = this.getHash(contentBuffer);
                    if (!meta) {
                        meta = { hash: hash };
                    }
                    else {
                        isDifferent = hash !== meta.hash;
                    }
                    var nEntry = (normalizedEntries[file] = {
                        key: file,
                        changed: !cacheExists || isDifferent,
                        meta: meta,
                    });
                    return nEntry;
                },
                /**
                 * Return the list o the files that changed compared
                 * against the ones stored in the cache
                 *
                 * @method getUpdated
                 * @param files {Array} the array of files to compare against the ones in the cache
                 * @returns {Array}
                 */
                getUpdatedFiles: function (files) {
                    var me = this;
                    files = files || [];
                    return me
                        .normalizeEntries(files)
                        .filter(function (entry) {
                        return entry.changed;
                    })
                        .map(function (entry) {
                        return entry.key;
                    });
                },
                /**
                 * return the list of files
                 * @method normalizeEntries
                 * @param files
                 * @returns {*}
                 */
                normalizeEntries: function (files) {
                    files = files || [];
                    var me = this;
                    var nEntries = files.map(function (file) {
                        return me.getFileDescriptor(file);
                    });
                    //normalizeEntries = nEntries;
                    return nEntries;
                },
                /**
                 * Remove an entry from the file-entry-cache. Useful to force the file to still be considered
                 * modified the next time the process is run
                 *
                 * @method removeEntry
                 * @param entryName
                 */
                removeEntry: function (entryName) {
                    delete normalizedEntries[entryName];
                    cache.removeKey(entryName);
                },
                /**
                 * Delete the cache file from the disk
                 * @method deleteCacheFile
                 */
                deleteCacheFile: function () {
                    cache.removeCacheFile();
                },
                /**
                 * remove the cache from the file and clear the memory cache
                 */
                destroy: function () {
                    normalizedEntries = {};
                    cache.destroy();
                },
                _getMetaForFileUsingCheckSum: function (cacheEntry) {
                    var contentBuffer = fs.readFileSync(cacheEntry.key);
                    var hash = this.getHash(contentBuffer);
                    var meta = Object.assign(cacheEntry.meta, { hash: hash });
                    delete meta.size;
                    delete meta.mtime;
                    return meta;
                },
                _getMetaForFileUsingMtimeAndSize: function (cacheEntry) {
                    var stat = fs.statSync(cacheEntry.key);
                    var meta = Object.assign(cacheEntry.meta, {
                        size: stat.size,
                        mtime: stat.mtime.getTime(),
                    });
                    delete meta.hash;
                    return meta;
                },
                /**
                 * Sync the files and persist them to the cache
                 * @method reconcile
                 */
                reconcile: function (noPrune) {
                    removeNotFoundFiles();
                    noPrune = typeof noPrune === 'undefined' ? true : noPrune;
                    var entries = normalizedEntries;
                    var keys = Object.keys(entries);
                    if (keys.length === 0) {
                        return;
                    }
                    var me = this;
                    keys.forEach(function (entryName) {
                        var cacheEntry = entries[entryName];
                        try {
                            var meta = useChecksum
                                ? me._getMetaForFileUsingCheckSum(cacheEntry)
                                : me._getMetaForFileUsingMtimeAndSize(cacheEntry);
                            cache.setKey(entryName, meta);
                        }
                        catch (err) {
                            // if the file does not exists we don't save it
                            // other errors are just thrown
                            if (err.code !== 'ENOENT') {
                                throw err;
                            }
                        }
                    });
                    cache.save(noPrune);
                },
            };
        },
    };
},
/* flat-cache/src/cache.js */ function _(require, module, exports, __esModule, __esExport) {
    var path = require('path');
    var fs = require('fs');
    var utils = require(708) /* ./utils */;
    var del = require(710) /* ./del */;
    var writeJSON = utils.writeJSON;
    var cache = {
        /**
         * Load a cache identified by the given Id. If the element does not exists, then initialize an empty
         * cache storage. If specified `cacheDir` will be used as the directory to persist the data to. If omitted
         * then the cache module directory `./cache` will be used instead
         *
         * @method load
         * @param docId {String} the id of the cache, would also be used as the name of the file cache
         * @param [cacheDir] {String} directory for the cache entry
         */
        load: function (docId, cacheDir) {
            var me = this;
            me._visited = {};
            me._persisted = {};
            me._pathToFile = cacheDir ? path.resolve(cacheDir, docId) : path.resolve(__dirname, '../.cache/', docId);
            if (fs.existsSync(me._pathToFile)) {
                me._persisted = utils.tryParse(me._pathToFile, {});
            }
        },
        /**
         * Load the cache from the provided file
         * @method loadFile
         * @param  {String} pathToFile the path to the file containing the info for the cache
         */
        loadFile: function (pathToFile) {
            var me = this;
            var dir = path.dirname(pathToFile);
            var fName = path.basename(pathToFile);
            me.load(fName, dir);
        },
        /**
         * Returns the entire persisted object
         * @method all
         * @returns {*}
         */
        all: function () {
            return this._persisted;
        },
        keys: function () {
            return Object.keys(this._persisted);
        },
        /**
         * sets a key to a given value
         * @method setKey
         * @param key {string} the key to set
         * @param value {object} the value of the key. Could be any object that can be serialized with JSON.stringify
         */
        setKey: function (key, value) {
            this._visited[key] = true;
            this._persisted[key] = value;
        },
        /**
         * remove a given key from the cache
         * @method removeKey
         * @param key {String} the key to remove from the object
         */
        removeKey: function (key) {
            delete this._visited[key]; // esfmt-ignore-line
            delete this._persisted[key]; // esfmt-ignore-line
        },
        /**
         * Return the value of the provided key
         * @method getKey
         * @param key {String} the name of the key to retrieve
         * @returns {*} the value from the key
         */
        getKey: function (key) {
            this._visited[key] = true;
            return this._persisted[key];
        },
        /**
         * Remove keys that were not accessed/set since the
         * last time the `prune` method was called.
         * @method _prune
         * @private
         */
        _prune: function () {
            var me = this;
            var obj = {};
            var keys = Object.keys(me._visited);
            // no keys visited for either get or set value
            if (keys.length === 0) {
                return;
            }
            keys.forEach(function (key) {
                obj[key] = me._persisted[key];
            });
            me._visited = {};
            me._persisted = obj;
        },
        /**
         * Save the state of the cache identified by the docId to disk
         * as a JSON structure
         * @param [noPrune=false] {Boolean} whether to remove from cache the non visited files
         * @method save
         */
        save: function (noPrune) {
            var me = this;
            !noPrune && me._prune();
            writeJSON(me._pathToFile, me._persisted);
        },
        /**
         * remove the file where the cache is persisted
         * @method removeCacheFile
         * @return {Boolean} true or false if the file was successfully deleted
         */
        removeCacheFile: function () {
            return del(this._pathToFile);
        },
        /**
         * Destroy the file cache and cache content.
         * @method destroy
         */
        destroy: function () {
            var me = this;
            me._visited = {};
            me._persisted = {};
            me.removeCacheFile();
        },
    };
    module.exports = {
        /**
         * Alias for create. Should be considered depreacted. Will be removed in next releases
         *
         * @method load
         * @param docId {String} the id of the cache, would also be used as the name of the file cache
         * @param [cacheDir] {String} directory for the cache entry
         * @returns {cache} cache instance
         */
        load: function (docId, cacheDir) {
            return this.create(docId, cacheDir);
        },
        /**
         * Load a cache identified by the given Id. If the element does not exists, then initialize an empty
         * cache storage.
         *
         * @method create
         * @param docId {String} the id of the cache, would also be used as the name of the file cache
         * @param [cacheDir] {String} directory for the cache entry
         * @returns {cache} cache instance
         */
        create: function (docId, cacheDir) {
            var obj = Object.create(cache);
            obj.load(docId, cacheDir);
            return obj;
        },
        createFromFile: function (filePath) {
            var obj = Object.create(cache);
            obj.loadFile(filePath);
            return obj;
        },
        /**
         * Clear the cache identified by the given id. Caches stored in a different cache directory can be deleted directly
         *
         * @method clearCache
         * @param docId {String} the id of the cache, would also be used as the name of the file cache
         * @param cacheDir {String} the directory where the cache file was written
         * @returns {Boolean} true if the cache folder was deleted. False otherwise
         */
        clearCacheById: function (docId, cacheDir) {
            var filePath = cacheDir ? path.resolve(cacheDir, docId) : path.resolve(__dirname, '../.cache/', docId);
            return del(filePath);
        },
        /**
         * Remove all cache stored in the cache directory
         * @method clearAll
         * @returns {Boolean} true if the cache folder was deleted. False otherwise
         */
        clearAll: function (cacheDir) {
            var filePath = cacheDir ? path.resolve(cacheDir) : path.resolve(__dirname, '../.cache/');
            return del(filePath);
        },
    };
},
/* flat-cache/src/utils.js */ function _(require, module, exports, __esModule, __esExport) {
    var fs = require('fs');
    var path = require('path');
    var flatted = require(709) /* flatted */;
    module.exports = {
        tryParse: function (filePath, defaultValue) {
            var result;
            try {
                result = this.readJSON(filePath);
            }
            catch (ex) {
                result = defaultValue;
            }
            return result;
        },
        /**
         * Read json file synchronously using flatted
         *
         * @method readJSON
         * @param  {String} filePath Json filepath
         * @returns {*} parse result
         */
        readJSON: function (filePath) {
            return flatted.parse(fs.readFileSync(filePath, {
                encoding: 'utf8',
            }));
        },
        /**
         * Write json file synchronously using circular-json
         *
         * @method writeJSON
         * @param  {String} filePath Json filepath
         * @param  {*} data Object to serialize
         */
        writeJSON: function (filePath, data) {
            fs.mkdirSync(path.dirname(filePath), {
                recursive: true,
            });
            fs.writeFileSync(filePath, flatted.stringify(data));
        },
    };
},
/* flatted/cjs/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /*! (c) 2020 Andrea Giammarchi */
    const { parse: $parse, stringify: $stringify } = JSON;
    const { keys } = Object;
    const Primitive = String; // it could be Number
    const primitive = 'string'; // it could be 'number'
    const ignore = {};
    const object = 'object';
    const noop = (_, value) => value;
    const primitives = value => (value instanceof Primitive ? Primitive(value) : value);
    const Primitives = (_, value) => (typeof value === primitive ? new Primitive(value) : value);
    const revive = (input, parsed, output, $) => {
        const lazy = [];
        for (let ke = keys(output), { length } = ke, y = 0; y < length; y++) {
            const k = ke[y];
            const value = output[k];
            if (value instanceof Primitive) {
                const tmp = input[value];
                if (typeof tmp === object && !parsed.has(tmp)) {
                    parsed.add(tmp);
                    output[k] = ignore;
                    lazy.push({ k, a: [input, parsed, tmp, $] });
                }
                else
                    output[k] = $.call(output, k, tmp);
            }
            else if (output[k] !== ignore)
                output[k] = $.call(output, k, value);
        }
        for (let { length } = lazy, i = 0; i < length; i++) {
            const { k, a } = lazy[i];
            output[k] = $.call(output, k, revive.apply(null, a));
        }
        return output;
    };
    const set = (known, input, value) => {
        const index = Primitive(input.push(value) - 1);
        known.set(value, index);
        return index;
    };
    const parse = (text, reviver) => {
        const input = $parse(text, Primitives).map(primitives);
        const value = input[0];
        const $ = reviver || noop;
        const tmp = typeof value === object && value ?
            revive(input, new Set, value, $) :
            value;
        return $.call({ '': tmp }, '', tmp);
    };
    exports.parse = parse;
    const stringify = (value, replacer, space) => {
        const $ = replacer && typeof replacer === object ?
            (k, v) => (k === '' || -1 < replacer.indexOf(k) ? v : void 0) :
            (replacer || noop);
        const known = new Map;
        const input = [];
        const output = [];
        let i = +set(known, input, $.call({ '': value }, '', value));
        let firstRun = !i;
        while (i < input.length) {
            firstRun = true;
            output[i] = $stringify(input[i++], replace, space);
        }
        return '[' + output.join(',') + ']';
        function replace(key, value) {
            if (firstRun) {
                firstRun = !firstRun;
                return value;
            }
            const after = $.call(this, key, value);
            switch (typeof after) {
                case object:
                    if (after === null)
                        return after;
                case primitive:
                    return known.get(after) || set(known, input, after);
            }
            return after;
        }
    };
    exports.stringify = stringify;
},
/* flat-cache/src/del.js */ function _(require, module, exports, __esModule, __esExport) {
    var rimraf = require(711) /* rimraf */.sync;
    var fs = require('fs');
    module.exports = function del(file) {
        if (fs.existsSync(file)) {
            //if rimraf doesn't throw then the file has been deleted or didn't exist
            rimraf(file, {
                glob: false,
            });
            return true;
        }
        return false;
    };
},
/* rimraf/rimraf.js */ function _(require, module, exports, __esModule, __esExport) {
    const assert = require("assert");
    const path = require("path");
    const fs = require("fs");
    let glob = undefined;
    try {
        glob = require(712) /* glob */;
    }
    catch (_err) {
        // treat glob as optional.
    }
    const defaultGlobOpts = {
        nosort: true,
        silent: true
    };
    // for EMFILE handling
    let timeout = 0;
    const isWindows = (process.platform === "win32");
    const defaults = options => {
        const methods = [
            'unlink',
            'chmod',
            'stat',
            'lstat',
            'rmdir',
            'readdir'
        ];
        methods.forEach(m => {
            options[m] = options[m] || fs[m];
            m = m + 'Sync';
            options[m] = options[m] || fs[m];
        });
        options.maxBusyTries = options.maxBusyTries || 3;
        options.emfileWait = options.emfileWait || 1000;
        if (options.glob === false) {
            options.disableGlob = true;
        }
        if (options.disableGlob !== true && glob === undefined) {
            throw Error('glob dependency not found, set `options.disableGlob = true` if intentional');
        }
        options.disableGlob = options.disableGlob || false;
        options.glob = options.glob || defaultGlobOpts;
    };
    const rimraf = (p, options, cb) => {
        if (typeof options === 'function') {
            cb = options;
            options = {};
        }
        assert(p, 'rimraf: missing path');
        assert.equal(typeof p, 'string', 'rimraf: path should be a string');
        assert.equal(typeof cb, 'function', 'rimraf: callback function required');
        assert(options, 'rimraf: invalid options argument provided');
        assert.equal(typeof options, 'object', 'rimraf: options should be object');
        defaults(options);
        let busyTries = 0;
        let errState = null;
        let n = 0;
        const next = (er) => {
            errState = errState || er;
            if (--n === 0)
                cb(errState);
        };
        const afterGlob = (er, results) => {
            if (er)
                return cb(er);
            n = results.length;
            if (n === 0)
                return cb();
            results.forEach(p => {
                const CB = (er) => {
                    if (er) {
                        if ((er.code === "EBUSY" || er.code === "ENOTEMPTY" || er.code === "EPERM") &&
                            busyTries < options.maxBusyTries) {
                            busyTries++;
                            // try again, with the same exact callback as this one.
                            return setTimeout(() => rimraf_(p, options, CB), busyTries * 100);
                        }
                        // this one won't happen if graceful-fs is used.
                        if (er.code === "EMFILE" && timeout < options.emfileWait) {
                            return setTimeout(() => rimraf_(p, options, CB), timeout++);
                        }
                        // already gone
                        if (er.code === "ENOENT")
                            er = null;
                    }
                    timeout = 0;
                    next(er);
                };
                rimraf_(p, options, CB);
            });
        };
        if (options.disableGlob || !glob.hasMagic(p))
            return afterGlob(null, [p]);
        options.lstat(p, (er, stat) => {
            if (!er)
                return afterGlob(null, [p]);
            glob(p, options.glob, afterGlob);
        });
    };
    // Two possible strategies.
    // 1. Assume it's a file.  unlink it, then do the dir stuff on EPERM or EISDIR
    // 2. Assume it's a directory.  readdir, then do the file stuff on ENOTDIR
    //
    // Both result in an extra syscall when you guess wrong.  However, there
    // are likely far more normal files in the world than directories.  This
    // is based on the assumption that a the average number of files per
    // directory is >= 1.
    //
    // If anyone ever complains about this, then I guess the strategy could
    // be made configurable somehow.  But until then, YAGNI.
    const rimraf_ = (p, options, cb) => {
        assert(p);
        assert(options);
        assert(typeof cb === 'function');
        // sunos lets the root user unlink directories, which is... weird.
        // so we have to lstat here and make sure it's not a dir.
        options.lstat(p, (er, st) => {
            if (er && er.code === "ENOENT")
                return cb(null);
            // Windows can EPERM on stat.  Life is suffering.
            if (er && er.code === "EPERM" && isWindows)
                fixWinEPERM(p, options, er, cb);
            if (st && st.isDirectory())
                return rmdir(p, options, er, cb);
            options.unlink(p, er => {
                if (er) {
                    if (er.code === "ENOENT")
                        return cb(null);
                    if (er.code === "EPERM")
                        return (isWindows)
                            ? fixWinEPERM(p, options, er, cb)
                            : rmdir(p, options, er, cb);
                    if (er.code === "EISDIR")
                        return rmdir(p, options, er, cb);
                }
                return cb(er);
            });
        });
    };
    const fixWinEPERM = (p, options, er, cb) => {
        assert(p);
        assert(options);
        assert(typeof cb === 'function');
        options.chmod(p, 0o666, er2 => {
            if (er2)
                cb(er2.code === "ENOENT" ? null : er);
            else
                options.stat(p, (er3, stats) => {
                    if (er3)
                        cb(er3.code === "ENOENT" ? null : er);
                    else if (stats.isDirectory())
                        rmdir(p, options, er, cb);
                    else
                        options.unlink(p, cb);
                });
        });
    };
    const fixWinEPERMSync = (p, options, er) => {
        assert(p);
        assert(options);
        try {
            options.chmodSync(p, 0o666);
        }
        catch (er2) {
            if (er2.code === "ENOENT")
                return;
            else
                throw er;
        }
        let stats;
        try {
            stats = options.statSync(p);
        }
        catch (er3) {
            if (er3.code === "ENOENT")
                return;
            else
                throw er;
        }
        if (stats.isDirectory())
            rmdirSync(p, options, er);
        else
            options.unlinkSync(p);
    };
    const rmdir = (p, options, originalEr, cb) => {
        assert(p);
        assert(options);
        assert(typeof cb === 'function');
        // try to rmdir first, and only readdir on ENOTEMPTY or EEXIST (SunOS)
        // if we guessed wrong, and it's not a directory, then
        // raise the original error.
        options.rmdir(p, er => {
            if (er && (er.code === "ENOTEMPTY" || er.code === "EEXIST" || er.code === "EPERM"))
                rmkids(p, options, cb);
            else if (er && er.code === "ENOTDIR")
                cb(originalEr);
            else
                cb(er);
        });
    };
    const rmkids = (p, options, cb) => {
        assert(p);
        assert(options);
        assert(typeof cb === 'function');
        options.readdir(p, (er, files) => {
            if (er)
                return cb(er);
            let n = files.length;
            if (n === 0)
                return options.rmdir(p, cb);
            let errState;
            files.forEach(f => {
                rimraf(path.join(p, f), options, er => {
                    if (errState)
                        return;
                    if (er)
                        return cb(errState = er);
                    if (--n === 0)
                        options.rmdir(p, cb);
                });
            });
        });
    };
    // this looks simpler, and is strictly *faster*, but will
    // tie up the JavaScript thread and fail on excessively
    // deep directory trees.
    const rimrafSync = (p, options) => {
        options = options || {};
        defaults(options);
        assert(p, 'rimraf: missing path');
        assert.equal(typeof p, 'string', 'rimraf: path should be a string');
        assert(options, 'rimraf: missing options');
        assert.equal(typeof options, 'object', 'rimraf: options should be object');
        let results;
        if (options.disableGlob || !glob.hasMagic(p)) {
            results = [p];
        }
        else {
            try {
                options.lstatSync(p);
                results = [p];
            }
            catch (er) {
                results = glob.sync(p, options.glob);
            }
        }
        if (!results.length)
            return;
        for (let i = 0; i < results.length; i++) {
            const p = results[i];
            let st;
            try {
                st = options.lstatSync(p);
            }
            catch (er) {
                if (er.code === "ENOENT")
                    return;
                // Windows can EPERM on stat.  Life is suffering.
                if (er.code === "EPERM" && isWindows)
                    fixWinEPERMSync(p, options, er);
            }
            try {
                // sunos lets the root user unlink directories, which is... weird.
                if (st && st.isDirectory())
                    rmdirSync(p, options, null);
                else
                    options.unlinkSync(p);
            }
            catch (er) {
                if (er.code === "ENOENT")
                    return;
                if (er.code === "EPERM")
                    return isWindows ? fixWinEPERMSync(p, options, er) : rmdirSync(p, options, er);
                if (er.code !== "EISDIR")
                    throw er;
                rmdirSync(p, options, er);
            }
        }
    };
    const rmdirSync = (p, options, originalEr) => {
        assert(p);
        assert(options);
        try {
            options.rmdirSync(p);
        }
        catch (er) {
            if (er.code === "ENOENT")
                return;
            if (er.code === "ENOTDIR")
                throw originalEr;
            if (er.code === "ENOTEMPTY" || er.code === "EEXIST" || er.code === "EPERM")
                rmkidsSync(p, options);
        }
    };
    const rmkidsSync = (p, options) => {
        assert(p);
        assert(options);
        options.readdirSync(p).forEach(f => rimrafSync(path.join(p, f), options));
        // We only end up here once we got ENOTEMPTY at least once, and
        // at this point, we are guaranteed to have removed all the kids.
        // So, we know that it won't be ENOENT or ENOTDIR or anything else.
        // try really hard to delete stuff on windows, because it has a
        // PROFOUNDLY annoying habit of not closing handles promptly when
        // files are deleted, resulting in spurious ENOTEMPTY errors.
        const retries = isWindows ? 100 : 1;
        let i = 0;
        do {
            let threw = true;
            try {
                const ret = options.rmdirSync(p, options);
                threw = false;
                return ret;
            }
            finally {
                if (++i < retries && threw)
                    continue;
            }
        } while (true);
    };
    module.exports = rimraf;
    rimraf.sync = rimrafSync;
},
/* glob/glob.js */ function _(require, module, exports, __esModule, __esExport) {
    // Approach:
    //
    // 1. Get the minimatch set
    // 2. For each pattern in the set, PROCESS(pattern, false)
    // 3. Store matches per-set, then uniq them
    //
    // PROCESS(pattern, inGlobStar)
    // Get the first [n] items from pattern that are all strings
    // Join these together.  This is PREFIX.
    //   If there is no more remaining, then stat(PREFIX) and
    //   add to matches if it succeeds.  END.
    //
    // If inGlobStar and PREFIX is symlink and points to dir
    //   set ENTRIES = []
    // else readdir(PREFIX) as ENTRIES
    //   If fail, END
    //
    // with ENTRIES
    //   If pattern[n] is GLOBSTAR
    //     // handle the case where the globstar match is empty
    //     // by pruning it out, and testing the resulting pattern
    //     PROCESS(pattern[0..n] + pattern[n+1 .. $], false)
    //     // handle other cases.
    //     for ENTRY in ENTRIES (not dotfiles)
    //       // attach globstar + tail onto the entry
    //       // Mark that this entry is a globstar match
    //       PROCESS(pattern[0..n] + ENTRY + pattern[n .. $], true)
    //
    //   else // not globstar
    //     for ENTRY in ENTRIES (not dotfiles, unless pattern[n] is dot)
    //       Test ENTRY against pattern[n]
    //       If fails, continue
    //       If passes, PROCESS(pattern[0..n] + item + pattern[n+1 .. $])
    //
    // Caveat:
    //   Cache all stats and readdirs results to minimize syscall.  Since all
    //   we ever care about is existence and directory-ness, we can just keep
    //   `true` for files, and [children,...] for directories, or `false` for
    //   things that don't exist.
    module.exports = glob;
    var fs = require('fs');
    var rp = require(713) /* fs.realpath */;
    var minimatch = require(279) /* minimatch */;
    var Minimatch = minimatch.Minimatch;
    var inherits = require(715) /* inherits */;
    var EE = require('events').EventEmitter;
    var path = require('path');
    var assert = require('assert');
    var isAbsolute = require(717) /* path-is-absolute */;
    var globSync = require(718) /* ./sync.js */;
    var common = require(719) /* ./common.js */;
    var setopts = common.setopts;
    var ownProp = common.ownProp;
    var inflight = require(720) /* inflight */;
    var util = require('util');
    var childrenIgnored = common.childrenIgnored;
    var isIgnored = common.isIgnored;
    var once = require(722) /* once */;
    function glob(pattern, options, cb) {
        if (typeof options === 'function')
            cb = options, options = {};
        if (!options)
            options = {};
        if (options.sync) {
            if (cb)
                throw new TypeError('callback provided to sync glob');
            return globSync(pattern, options);
        }
        return new Glob(pattern, options, cb);
    }
    glob.sync = globSync;
    var GlobSync = glob.GlobSync = globSync.GlobSync;
    // old api surface
    glob.glob = glob;
    function extend(origin, add) {
        if (add === null || typeof add !== 'object') {
            return origin;
        }
        var keys = Object.keys(add);
        var i = keys.length;
        while (i--) {
            origin[keys[i]] = add[keys[i]];
        }
        return origin;
    }
    glob.hasMagic = function (pattern, options_) {
        var options = extend({}, options_);
        options.noprocess = true;
        var g = new Glob(pattern, options);
        var set = g.minimatch.set;
        if (!pattern)
            return false;
        if (set.length > 1)
            return true;
        for (var j = 0; j < set[0].length; j++) {
            if (typeof set[0][j] !== 'string')
                return true;
        }
        return false;
    };
    glob.Glob = Glob;
    inherits(Glob, EE);
    function Glob(pattern, options, cb) {
        if (typeof options === 'function') {
            cb = options;
            options = null;
        }
        if (options && options.sync) {
            if (cb)
                throw new TypeError('callback provided to sync glob');
            return new GlobSync(pattern, options);
        }
        if (!(this instanceof Glob))
            return new Glob(pattern, options, cb);
        setopts(this, pattern, options);
        this._didRealPath = false;
        // process each pattern in the minimatch set
        var n = this.minimatch.set.length;
        // The matches are stored as {<filename>: true,...} so that
        // duplicates are automagically pruned.
        // Later, we do an Object.keys() on these.
        // Keep them as a list so we can fill in when nonull is set.
        this.matches = new Array(n);
        if (typeof cb === 'function') {
            cb = once(cb);
            this.on('error', cb);
            this.on('end', function (matches) {
                cb(null, matches);
            });
        }
        var self = this;
        this._processing = 0;
        this._emitQueue = [];
        this._processQueue = [];
        this.paused = false;
        if (this.noprocess)
            return this;
        if (n === 0)
            return done();
        var sync = true;
        for (var i = 0; i < n; i++) {
            this._process(this.minimatch.set[i], i, false, done);
        }
        sync = false;
        function done() {
            --self._processing;
            if (self._processing <= 0) {
                if (sync) {
                    process.nextTick(function () {
                        self._finish();
                    });
                }
                else {
                    self._finish();
                }
            }
        }
    }
    Glob.prototype._finish = function () {
        assert(this instanceof Glob);
        if (this.aborted)
            return;
        if (this.realpath && !this._didRealpath)
            return this._realpath();
        common.finish(this);
        this.emit('end', this.found);
    };
    Glob.prototype._realpath = function () {
        if (this._didRealpath)
            return;
        this._didRealpath = true;
        var n = this.matches.length;
        if (n === 0)
            return this._finish();
        var self = this;
        for (var i = 0; i < this.matches.length; i++)
            this._realpathSet(i, next);
        function next() {
            if (--n === 0)
                self._finish();
        }
    };
    Glob.prototype._realpathSet = function (index, cb) {
        var matchset = this.matches[index];
        if (!matchset)
            return cb();
        var found = Object.keys(matchset);
        var self = this;
        var n = found.length;
        if (n === 0)
            return cb();
        var set = this.matches[index] = Object.create(null);
        found.forEach(function (p, i) {
            // If there's a problem with the stat, then it means that
            // one or more of the links in the realpath couldn't be
            // resolved.  just return the abs value in that case.
            p = self._makeAbs(p);
            rp.realpath(p, self.realpathCache, function (er, real) {
                if (!er)
                    set[real] = true;
                else if (er.syscall === 'stat')
                    set[p] = true;
                else
                    self.emit('error', er); // srsly wtf right here
                if (--n === 0) {
                    self.matches[index] = set;
                    cb();
                }
            });
        });
    };
    Glob.prototype._mark = function (p) {
        return common.mark(this, p);
    };
    Glob.prototype._makeAbs = function (f) {
        return common.makeAbs(this, f);
    };
    Glob.prototype.abort = function () {
        this.aborted = true;
        this.emit('abort');
    };
    Glob.prototype.pause = function () {
        if (!this.paused) {
            this.paused = true;
            this.emit('pause');
        }
    };
    Glob.prototype.resume = function () {
        if (this.paused) {
            this.emit('resume');
            this.paused = false;
            if (this._emitQueue.length) {
                var eq = this._emitQueue.slice(0);
                this._emitQueue.length = 0;
                for (var i = 0; i < eq.length; i++) {
                    var e = eq[i];
                    this._emitMatch(e[0], e[1]);
                }
            }
            if (this._processQueue.length) {
                var pq = this._processQueue.slice(0);
                this._processQueue.length = 0;
                for (var i = 0; i < pq.length; i++) {
                    var p = pq[i];
                    this._processing--;
                    this._process(p[0], p[1], p[2], p[3]);
                }
            }
        }
    };
    Glob.prototype._process = function (pattern, index, inGlobStar, cb) {
        assert(this instanceof Glob);
        assert(typeof cb === 'function');
        if (this.aborted)
            return;
        this._processing++;
        if (this.paused) {
            this._processQueue.push([pattern, index, inGlobStar, cb]);
            return;
        }
        //console.error('PROCESS %d', this._processing, pattern)
        // Get the first [n] parts of pattern that are all strings.
        var n = 0;
        while (typeof pattern[n] === 'string') {
            n++;
        }
        // now n is the index of the first one that is *not* a string.
        // see if there's anything else
        var prefix;
        switch (n) {
            // if not, then this is rather simple
            case pattern.length:
                this._processSimple(pattern.join('/'), index, cb);
                return;
            case 0:
                // pattern *starts* with some non-trivial item.
                // going to readdir(cwd), but not include the prefix in matches.
                prefix = null;
                break;
            default:
                // pattern has some string bits in the front.
                // whatever it starts with, whether that's 'absolute' like /foo/bar,
                // or 'relative' like '../baz'
                prefix = pattern.slice(0, n).join('/');
                break;
        }
        var remain = pattern.slice(n);
        // get the list of entries.
        var read;
        if (prefix === null)
            read = '.';
        else if (isAbsolute(prefix) || isAbsolute(pattern.join('/'))) {
            if (!prefix || !isAbsolute(prefix))
                prefix = '/' + prefix;
            read = prefix;
        }
        else
            read = prefix;
        var abs = this._makeAbs(read);
        //if ignored, skip _processing
        if (childrenIgnored(this, read))
            return cb();
        var isGlobStar = remain[0] === minimatch.GLOBSTAR;
        if (isGlobStar)
            this._processGlobStar(prefix, read, abs, remain, index, inGlobStar, cb);
        else
            this._processReaddir(prefix, read, abs, remain, index, inGlobStar, cb);
    };
    Glob.prototype._processReaddir = function (prefix, read, abs, remain, index, inGlobStar, cb) {
        var self = this;
        this._readdir(abs, inGlobStar, function (er, entries) {
            return self._processReaddir2(prefix, read, abs, remain, index, inGlobStar, entries, cb);
        });
    };
    Glob.prototype._processReaddir2 = function (prefix, read, abs, remain, index, inGlobStar, entries, cb) {
        // if the abs isn't a dir, then nothing can match!
        if (!entries)
            return cb();
        // It will only match dot entries if it starts with a dot, or if
        // dot is set.  Stuff like @(.foo|.bar) isn't allowed.
        var pn = remain[0];
        var negate = !!this.minimatch.negate;
        var rawGlob = pn._glob;
        var dotOk = this.dot || rawGlob.charAt(0) === '.';
        var matchedEntries = [];
        for (var i = 0; i < entries.length; i++) {
            var e = entries[i];
            if (e.charAt(0) !== '.' || dotOk) {
                var m;
                if (negate && !prefix) {
                    m = !e.match(pn);
                }
                else {
                    m = e.match(pn);
                }
                if (m)
                    matchedEntries.push(e);
            }
        }
        //console.error('prd2', prefix, entries, remain[0]._glob, matchedEntries)
        var len = matchedEntries.length;
        // If there are no matched entries, then nothing matches.
        if (len === 0)
            return cb();
        // if this is the last remaining pattern bit, then no need for
        // an additional stat *unless* the user has specified mark or
        // stat explicitly.  We know they exist, since readdir returned
        // them.
        if (remain.length === 1 && !this.mark && !this.stat) {
            if (!this.matches[index])
                this.matches[index] = Object.create(null);
            for (var i = 0; i < len; i++) {
                var e = matchedEntries[i];
                if (prefix) {
                    if (prefix !== '/')
                        e = prefix + '/' + e;
                    else
                        e = prefix + e;
                }
                if (e.charAt(0) === '/' && !this.nomount) {
                    e = path.join(this.root, e);
                }
                this._emitMatch(index, e);
            }
            // This was the last one, and no stats were needed
            return cb();
        }
        // now test all matched entries as stand-ins for that part
        // of the pattern.
        remain.shift();
        for (var i = 0; i < len; i++) {
            var e = matchedEntries[i];
            var newPattern;
            if (prefix) {
                if (prefix !== '/')
                    e = prefix + '/' + e;
                else
                    e = prefix + e;
            }
            this._process([e].concat(remain), index, inGlobStar, cb);
        }
        cb();
    };
    Glob.prototype._emitMatch = function (index, e) {
        if (this.aborted)
            return;
        if (isIgnored(this, e))
            return;
        if (this.paused) {
            this._emitQueue.push([index, e]);
            return;
        }
        var abs = isAbsolute(e) ? e : this._makeAbs(e);
        if (this.mark)
            e = this._mark(e);
        if (this.absolute)
            e = abs;
        if (this.matches[index][e])
            return;
        if (this.nodir) {
            var c = this.cache[abs];
            if (c === 'DIR' || Array.isArray(c))
                return;
        }
        this.matches[index][e] = true;
        var st = this.statCache[abs];
        if (st)
            this.emit('stat', e, st);
        this.emit('match', e);
    };
    Glob.prototype._readdirInGlobStar = function (abs, cb) {
        if (this.aborted)
            return;
        // follow all symlinked directories forever
        // just proceed as if this is a non-globstar situation
        if (this.follow)
            return this._readdir(abs, false, cb);
        var lstatkey = 'lstat\0' + abs;
        var self = this;
        var lstatcb = inflight(lstatkey, lstatcb_);
        if (lstatcb)
            fs.lstat(abs, lstatcb);
        function lstatcb_(er, lstat) {
            if (er && er.code === 'ENOENT')
                return cb();
            var isSym = lstat && lstat.isSymbolicLink();
            self.symlinks[abs] = isSym;
            // If it's not a symlink or a dir, then it's definitely a regular file.
            // don't bother doing a readdir in that case.
            if (!isSym && lstat && !lstat.isDirectory()) {
                self.cache[abs] = 'FILE';
                cb();
            }
            else
                self._readdir(abs, false, cb);
        }
    };
    Glob.prototype._readdir = function (abs, inGlobStar, cb) {
        if (this.aborted)
            return;
        cb = inflight('readdir\0' + abs + '\0' + inGlobStar, cb);
        if (!cb)
            return;
        //console.error('RD %j %j', +inGlobStar, abs)
        if (inGlobStar && !ownProp(this.symlinks, abs))
            return this._readdirInGlobStar(abs, cb);
        if (ownProp(this.cache, abs)) {
            var c = this.cache[abs];
            if (!c || c === 'FILE')
                return cb();
            if (Array.isArray(c))
                return cb(null, c);
        }
        var self = this;
        fs.readdir(abs, readdirCb(this, abs, cb));
    };
    function readdirCb(self, abs, cb) {
        return function (er, entries) {
            if (er)
                self._readdirError(abs, er, cb);
            else
                self._readdirEntries(abs, entries, cb);
        };
    }
    Glob.prototype._readdirEntries = function (abs, entries, cb) {
        if (this.aborted)
            return;
        // if we haven't asked to stat everything, then just
        // assume that everything in there exists, so we can avoid
        // having to stat it a second time.
        if (!this.mark && !this.stat) {
            for (var i = 0; i < entries.length; i++) {
                var e = entries[i];
                if (abs === '/')
                    e = abs + e;
                else
                    e = abs + '/' + e;
                this.cache[e] = true;
            }
        }
        this.cache[abs] = entries;
        return cb(null, entries);
    };
    Glob.prototype._readdirError = function (f, er, cb) {
        if (this.aborted)
            return;
        // handle errors, and cache the information
        switch (er.code) {
            case 'ENOTSUP': // https://github.com/isaacs/node-glob/issues/205
            case 'ENOTDIR': // totally normal. means it *does* exist.
                var abs = this._makeAbs(f);
                this.cache[abs] = 'FILE';
                if (abs === this.cwdAbs) {
                    var error = new Error(er.code + ' invalid cwd ' + this.cwd);
                    error.path = this.cwd;
                    error.code = er.code;
                    this.emit('error', error);
                    this.abort();
                }
                break;
            case 'ENOENT': // not terribly unusual
            case 'ELOOP':
            case 'ENAMETOOLONG':
            case 'UNKNOWN':
                this.cache[this._makeAbs(f)] = false;
                break;
            default: // some unusual error.  Treat as failure.
                this.cache[this._makeAbs(f)] = false;
                if (this.strict) {
                    this.emit('error', er);
                    // If the error is handled, then we abort
                    // if not, we threw out of here
                    this.abort();
                }
                if (!this.silent)
                    console.error('glob error', er);
                break;
        }
        return cb();
    };
    Glob.prototype._processGlobStar = function (prefix, read, abs, remain, index, inGlobStar, cb) {
        var self = this;
        this._readdir(abs, inGlobStar, function (er, entries) {
            self._processGlobStar2(prefix, read, abs, remain, index, inGlobStar, entries, cb);
        });
    };
    Glob.prototype._processGlobStar2 = function (prefix, read, abs, remain, index, inGlobStar, entries, cb) {
        //console.error('pgs2', prefix, remain[0], entries)
        // no entries means not a dir, so it can never have matches
        // foo.txt/** doesn't match foo.txt
        if (!entries)
            return cb();
        // test without the globstar, and with every child both below
        // and replacing the globstar.
        var remainWithoutGlobStar = remain.slice(1);
        var gspref = prefix ? [prefix] : [];
        var noGlobStar = gspref.concat(remainWithoutGlobStar);
        // the noGlobStar pattern exits the inGlobStar state
        this._process(noGlobStar, index, false, cb);
        var isSym = this.symlinks[abs];
        var len = entries.length;
        // If it's a symlink, and we're in a globstar, then stop
        if (isSym && inGlobStar)
            return cb();
        for (var i = 0; i < len; i++) {
            var e = entries[i];
            if (e.charAt(0) === '.' && !this.dot)
                continue;
            // these two cases enter the inGlobStar state
            var instead = gspref.concat(entries[i], remainWithoutGlobStar);
            this._process(instead, index, true, cb);
            var below = gspref.concat(entries[i], remain);
            this._process(below, index, true, cb);
        }
        cb();
    };
    Glob.prototype._processSimple = function (prefix, index, cb) {
        // XXX review this.  Shouldn't it be doing the mounting etc
        // before doing stat?  kinda weird?
        var self = this;
        this._stat(prefix, function (er, exists) {
            self._processSimple2(prefix, index, er, exists, cb);
        });
    };
    Glob.prototype._processSimple2 = function (prefix, index, er, exists, cb) {
        //console.error('ps2', prefix, exists)
        if (!this.matches[index])
            this.matches[index] = Object.create(null);
        // If it doesn't exist, then just mark the lack of results
        if (!exists)
            return cb();
        if (prefix && isAbsolute(prefix) && !this.nomount) {
            var trail = /[\/\\]$/.test(prefix);
            if (prefix.charAt(0) === '/') {
                prefix = path.join(this.root, prefix);
            }
            else {
                prefix = path.resolve(this.root, prefix);
                if (trail)
                    prefix += '/';
            }
        }
        if (process.platform === 'win32')
            prefix = prefix.replace(/\\/g, '/');
        // Mark this as a match
        this._emitMatch(index, prefix);
        cb();
    };
    // Returns either 'DIR', 'FILE', or false
    Glob.prototype._stat = function (f, cb) {
        var abs = this._makeAbs(f);
        var needDir = f.slice(-1) === '/';
        if (f.length > this.maxLength)
            return cb();
        if (!this.stat && ownProp(this.cache, abs)) {
            var c = this.cache[abs];
            if (Array.isArray(c))
                c = 'DIR';
            // It exists, but maybe not how we need it
            if (!needDir || c === 'DIR')
                return cb(null, c);
            if (needDir && c === 'FILE')
                return cb();
            // otherwise we have to stat, because maybe c=true
            // if we know it exists, but not what it is.
        }
        var exists;
        var stat = this.statCache[abs];
        if (stat !== undefined) {
            if (stat === false)
                return cb(null, stat);
            else {
                var type = stat.isDirectory() ? 'DIR' : 'FILE';
                if (needDir && type === 'FILE')
                    return cb();
                else
                    return cb(null, type, stat);
            }
        }
        var self = this;
        var statcb = inflight('stat\0' + abs, lstatcb_);
        if (statcb)
            fs.lstat(abs, statcb);
        function lstatcb_(er, lstat) {
            if (lstat && lstat.isSymbolicLink()) {
                // If it's a symlink, then treat it as the target, unless
                // the target does not exist, then treat it as a file.
                return fs.stat(abs, function (er, stat) {
                    if (er)
                        self._stat2(f, abs, null, lstat, cb);
                    else
                        self._stat2(f, abs, er, stat, cb);
                });
            }
            else {
                self._stat2(f, abs, er, lstat, cb);
            }
        }
    };
    Glob.prototype._stat2 = function (f, abs, er, stat, cb) {
        if (er && (er.code === 'ENOENT' || er.code === 'ENOTDIR')) {
            this.statCache[abs] = false;
            return cb();
        }
        var needDir = f.slice(-1) === '/';
        this.statCache[abs] = stat;
        if (abs.slice(-1) === '/' && stat && !stat.isDirectory())
            return cb(null, false, stat);
        var c = true;
        if (stat)
            c = stat.isDirectory() ? 'DIR' : 'FILE';
        this.cache[abs] = this.cache[abs] || c;
        if (needDir && c === 'FILE')
            return cb();
        return cb(null, c, stat);
    };
},
/* fs.realpath/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = realpath;
    realpath.realpath = realpath;
    realpath.sync = realpathSync;
    realpath.realpathSync = realpathSync;
    realpath.monkeypatch = monkeypatch;
    realpath.unmonkeypatch = unmonkeypatch;
    var fs = require('fs');
    var origRealpath = fs.realpath;
    var origRealpathSync = fs.realpathSync;
    var version = process.version;
    var ok = /^v[0-5]\./.test(version);
    var old = require(714) /* ./old.js */;
    function newError(er) {
        return er && er.syscall === 'realpath' && (er.code === 'ELOOP' ||
            er.code === 'ENOMEM' ||
            er.code === 'ENAMETOOLONG');
    }
    function realpath(p, cache, cb) {
        if (ok) {
            return origRealpath(p, cache, cb);
        }
        if (typeof cache === 'function') {
            cb = cache;
            cache = null;
        }
        origRealpath(p, cache, function (er, result) {
            if (newError(er)) {
                old.realpath(p, cache, cb);
            }
            else {
                cb(er, result);
            }
        });
    }
    function realpathSync(p, cache) {
        if (ok) {
            return origRealpathSync(p, cache);
        }
        try {
            return origRealpathSync(p, cache);
        }
        catch (er) {
            if (newError(er)) {
                return old.realpathSync(p, cache);
            }
            else {
                throw er;
            }
        }
    }
    function monkeypatch() {
        fs.realpath = realpath;
        fs.realpathSync = realpathSync;
    }
    function unmonkeypatch() {
        fs.realpath = origRealpath;
        fs.realpathSync = origRealpathSync;
    }
},
/* fs.realpath/old.js */ function _(require, module, exports, __esModule, __esExport) {
    // Copyright Joyent, Inc. and other Node contributors.
    //
    // Permission is hereby granted, free of charge, to any person obtaining a
    // copy of this software and associated documentation files (the
    // "Software"), to deal in the Software without restriction, including
    // without limitation the rights to use, copy, modify, merge, publish,
    // distribute, sublicense, and/or sell copies of the Software, and to permit
    // persons to whom the Software is furnished to do so, subject to the
    // following conditions:
    //
    // The above copyright notice and this permission notice shall be included
    // in all copies or substantial portions of the Software.
    //
    // THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    // OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    // MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    // NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
    // DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
    // USE OR OTHER DEALINGS IN THE SOFTWARE.
    var pathModule = require('path');
    var isWindows = process.platform === 'win32';
    var fs = require('fs');
    // JavaScript implementation of realpath, ported from node pre-v6
    var DEBUG = process.env.NODE_DEBUG && /fs/.test(process.env.NODE_DEBUG);
    function rethrow() {
        // Only enable in debug mode. A backtrace uses ~1000 bytes of heap space and
        // is fairly slow to generate.
        var callback;
        if (DEBUG) {
            var backtrace = new Error;
            callback = debugCallback;
        }
        else
            callback = missingCallback;
        return callback;
        function debugCallback(err) {
            if (err) {
                backtrace.message = err.message;
                err = backtrace;
                missingCallback(err);
            }
        }
        function missingCallback(err) {
            if (err) {
                if (process.throwDeprecation)
                    throw err; // Forgot a callback but don't know where? Use NODE_DEBUG=fs
                else if (!process.noDeprecation) {
                    var msg = 'fs: missing callback ' + (err.stack || err.message);
                    if (process.traceDeprecation)
                        console.trace(msg);
                    else
                        console.error(msg);
                }
            }
        }
    }
    function maybeCallback(cb) {
        return typeof cb === 'function' ? cb : rethrow();
    }
    var normalize = pathModule.normalize;
    // Regexp that finds the next partion of a (partial) path
    // result is [base_with_slash, base], e.g. ['somedir/', 'somedir']
    if (isWindows) {
        var nextPartRe = /(.*?)(?:[\/\\]+|$)/g;
    }
    else {
        var nextPartRe = /(.*?)(?:[\/]+|$)/g;
    }
    // Regex to find the device root, including trailing slash. E.g. 'c:\\'.
    if (isWindows) {
        var splitRootRe = /^(?:[a-zA-Z]:|[\\\/]{2}[^\\\/]+[\\\/][^\\\/]+)?[\\\/]*/;
    }
    else {
        var splitRootRe = /^[\/]*/;
    }
    exports.realpathSync = function realpathSync(p, cache) {
        // make p is absolute
        p = pathModule.resolve(p);
        if (cache && Object.prototype.hasOwnProperty.call(cache, p)) {
            return cache[p];
        }
        var original = p, seenLinks = {}, knownHard = {};
        // current character position in p
        var pos;
        // the partial path so far, including a trailing slash if any
        var current;
        // the partial path without a trailing slash (except when pointing at a root)
        var base;
        // the partial path scanned in the previous round, with slash
        var previous;
        start();
        function start() {
            // Skip over roots
            var m = splitRootRe.exec(p);
            pos = m[0].length;
            current = m[0];
            base = m[0];
            previous = '';
            // On windows, check that the root exists. On unix there is no need.
            if (isWindows && !knownHard[base]) {
                fs.lstatSync(base);
                knownHard[base] = true;
            }
        }
        // walk down the path, swapping out linked pathparts for their real
        // values
        // NB: p.length changes.
        while (pos < p.length) {
            // find the next part
            nextPartRe.lastIndex = pos;
            var result = nextPartRe.exec(p);
            previous = current;
            current += result[0];
            base = previous + result[1];
            pos = nextPartRe.lastIndex;
            // continue if not a symlink
            if (knownHard[base] || (cache && cache[base] === base)) {
                continue;
            }
            var resolvedLink;
            if (cache && Object.prototype.hasOwnProperty.call(cache, base)) {
                // some known symbolic link.  no need to stat again.
                resolvedLink = cache[base];
            }
            else {
                var stat = fs.lstatSync(base);
                if (!stat.isSymbolicLink()) {
                    knownHard[base] = true;
                    if (cache)
                        cache[base] = base;
                    continue;
                }
                // read the link if it wasn't read before
                // dev/ino always return 0 on windows, so skip the check.
                var linkTarget = null;
                if (!isWindows) {
                    var id = stat.dev.toString(32) + ':' + stat.ino.toString(32);
                    if (seenLinks.hasOwnProperty(id)) {
                        linkTarget = seenLinks[id];
                    }
                }
                if (linkTarget === null) {
                    fs.statSync(base);
                    linkTarget = fs.readlinkSync(base);
                }
                resolvedLink = pathModule.resolve(previous, linkTarget);
                // track this, if given a cache.
                if (cache)
                    cache[base] = resolvedLink;
                if (!isWindows)
                    seenLinks[id] = linkTarget;
            }
            // resolve the link, then start over
            p = pathModule.resolve(resolvedLink, p.slice(pos));
            start();
        }
        if (cache)
            cache[original] = p;
        return p;
    };
    exports.realpath = function realpath(p, cache, cb) {
        if (typeof cb !== 'function') {
            cb = maybeCallback(cache);
            cache = null;
        }
        // make p is absolute
        p = pathModule.resolve(p);
        if (cache && Object.prototype.hasOwnProperty.call(cache, p)) {
            return process.nextTick(cb.bind(null, null, cache[p]));
        }
        var original = p, seenLinks = {}, knownHard = {};
        // current character position in p
        var pos;
        // the partial path so far, including a trailing slash if any
        var current;
        // the partial path without a trailing slash (except when pointing at a root)
        var base;
        // the partial path scanned in the previous round, with slash
        var previous;
        start();
        function start() {
            // Skip over roots
            var m = splitRootRe.exec(p);
            pos = m[0].length;
            current = m[0];
            base = m[0];
            previous = '';
            // On windows, check that the root exists. On unix there is no need.
            if (isWindows && !knownHard[base]) {
                fs.lstat(base, function (err) {
                    if (err)
                        return cb(err);
                    knownHard[base] = true;
                    LOOP();
                });
            }
            else {
                process.nextTick(LOOP);
            }
        }
        // walk down the path, swapping out linked pathparts for their real
        // values
        function LOOP() {
            // stop if scanned past end of path
            if (pos >= p.length) {
                if (cache)
                    cache[original] = p;
                return cb(null, p);
            }
            // find the next part
            nextPartRe.lastIndex = pos;
            var result = nextPartRe.exec(p);
            previous = current;
            current += result[0];
            base = previous + result[1];
            pos = nextPartRe.lastIndex;
            // continue if not a symlink
            if (knownHard[base] || (cache && cache[base] === base)) {
                return process.nextTick(LOOP);
            }
            if (cache && Object.prototype.hasOwnProperty.call(cache, base)) {
                // known symbolic link.  no need to stat again.
                return gotResolvedLink(cache[base]);
            }
            return fs.lstat(base, gotStat);
        }
        function gotStat(err, stat) {
            if (err)
                return cb(err);
            // if not a symlink, skip to the next path part
            if (!stat.isSymbolicLink()) {
                knownHard[base] = true;
                if (cache)
                    cache[base] = base;
                return process.nextTick(LOOP);
            }
            // stat & read the link if not read before
            // call gotTarget as soon as the link target is known
            // dev/ino always return 0 on windows, so skip the check.
            if (!isWindows) {
                var id = stat.dev.toString(32) + ':' + stat.ino.toString(32);
                if (seenLinks.hasOwnProperty(id)) {
                    return gotTarget(null, seenLinks[id], base);
                }
            }
            fs.stat(base, function (err) {
                if (err)
                    return cb(err);
                fs.readlink(base, function (err, target) {
                    if (!isWindows)
                        seenLinks[id] = target;
                    gotTarget(err, target);
                });
            });
        }
        function gotTarget(err, target, base) {
            if (err)
                return cb(err);
            var resolvedLink = pathModule.resolve(previous, target);
            if (cache)
                cache[base] = resolvedLink;
            gotResolvedLink(resolvedLink);
        }
        function gotResolvedLink(resolvedLink) {
            // resolve the link, then start over
            p = pathModule.resolve(resolvedLink, p.slice(pos));
            start();
        }
    };
},
/* inherits/inherits.js */ function _(require, module, exports, __esModule, __esExport) {
    try {
        var util = require('util');
        /* istanbul ignore next */
        if (typeof util.inherits !== 'function')
            throw '';
        module.exports = util.inherits;
    }
    catch (e) {
        /* istanbul ignore next */
        module.exports = require(716) /* ./inherits_browser.js */;
    }
},
/* inherits/inherits_browser.js */ function _(require, module, exports, __esModule, __esExport) {
    if (typeof Object.create === 'function') {
        // implementation from standard node.js 'util' module
        module.exports = function inherits(ctor, superCtor) {
            if (superCtor) {
                ctor.super_ = superCtor;
                ctor.prototype = Object.create(superCtor.prototype, {
                    constructor: {
                        value: ctor,
                        enumerable: false,
                        writable: true,
                        configurable: true
                    }
                });
            }
        };
    }
    else {
        // old school shim for old browsers
        module.exports = function inherits(ctor, superCtor) {
            if (superCtor) {
                ctor.super_ = superCtor;
                var TempCtor = function () { };
                TempCtor.prototype = superCtor.prototype;
                ctor.prototype = new TempCtor();
                ctor.prototype.constructor = ctor;
            }
        };
    }
},
/* path-is-absolute/index.js */ function _(require, module, exports, __esModule, __esExport) {
    function posix(path) {
        return path.charAt(0) === '/';
    }
    function win32(path) {
        // https://github.com/nodejs/node/blob/b3fcc245fb25539909ef1d5eaa01dbf92e168633/lib/path.js#L56
        var splitDeviceRe = /^([a-zA-Z]:|[\\\/]{2}[^\\\/]+[\\\/]+[^\\\/]+)?([\\\/])?([\s\S]*?)$/;
        var result = splitDeviceRe.exec(path);
        var device = result[1] || '';
        var isUnc = Boolean(device && device.charAt(1) !== ':');
        // UNC paths are always absolute
        return Boolean(result[2] || isUnc);
    }
    module.exports = process.platform === 'win32' ? win32 : posix;
    module.exports.posix = posix;
    module.exports.win32 = win32;
},
/* glob/sync.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = globSync;
    globSync.GlobSync = GlobSync;
    var fs = require('fs');
    var rp = require(713) /* fs.realpath */;
    var minimatch = require(279) /* minimatch */;
    var Minimatch = minimatch.Minimatch;
    var Glob = require(712) /* ./glob.js */.Glob;
    var util = require('util');
    var path = require('path');
    var assert = require('assert');
    var isAbsolute = require(717) /* path-is-absolute */;
    var common = require(719) /* ./common.js */;
    var setopts = common.setopts;
    var ownProp = common.ownProp;
    var childrenIgnored = common.childrenIgnored;
    var isIgnored = common.isIgnored;
    function globSync(pattern, options) {
        if (typeof options === 'function' || arguments.length === 3)
            throw new TypeError('callback provided to sync glob\n' +
                'See: https://github.com/isaacs/node-glob/issues/167');
        return new GlobSync(pattern, options).found;
    }
    function GlobSync(pattern, options) {
        if (!pattern)
            throw new Error('must provide pattern');
        if (typeof options === 'function' || arguments.length === 3)
            throw new TypeError('callback provided to sync glob\n' +
                'See: https://github.com/isaacs/node-glob/issues/167');
        if (!(this instanceof GlobSync))
            return new GlobSync(pattern, options);
        setopts(this, pattern, options);
        if (this.noprocess)
            return this;
        var n = this.minimatch.set.length;
        this.matches = new Array(n);
        for (var i = 0; i < n; i++) {
            this._process(this.minimatch.set[i], i, false);
        }
        this._finish();
    }
    GlobSync.prototype._finish = function () {
        assert(this instanceof GlobSync);
        if (this.realpath) {
            var self = this;
            this.matches.forEach(function (matchset, index) {
                var set = self.matches[index] = Object.create(null);
                for (var p in matchset) {
                    try {
                        p = self._makeAbs(p);
                        var real = rp.realpathSync(p, self.realpathCache);
                        set[real] = true;
                    }
                    catch (er) {
                        if (er.syscall === 'stat')
                            set[self._makeAbs(p)] = true;
                        else
                            throw er;
                    }
                }
            });
        }
        common.finish(this);
    };
    GlobSync.prototype._process = function (pattern, index, inGlobStar) {
        assert(this instanceof GlobSync);
        // Get the first [n] parts of pattern that are all strings.
        var n = 0;
        while (typeof pattern[n] === 'string') {
            n++;
        }
        // now n is the index of the first one that is *not* a string.
        // See if there's anything else
        var prefix;
        switch (n) {
            // if not, then this is rather simple
            case pattern.length:
                this._processSimple(pattern.join('/'), index);
                return;
            case 0:
                // pattern *starts* with some non-trivial item.
                // going to readdir(cwd), but not include the prefix in matches.
                prefix = null;
                break;
            default:
                // pattern has some string bits in the front.
                // whatever it starts with, whether that's 'absolute' like /foo/bar,
                // or 'relative' like '../baz'
                prefix = pattern.slice(0, n).join('/');
                break;
        }
        var remain = pattern.slice(n);
        // get the list of entries.
        var read;
        if (prefix === null)
            read = '.';
        else if (isAbsolute(prefix) || isAbsolute(pattern.join('/'))) {
            if (!prefix || !isAbsolute(prefix))
                prefix = '/' + prefix;
            read = prefix;
        }
        else
            read = prefix;
        var abs = this._makeAbs(read);
        //if ignored, skip processing
        if (childrenIgnored(this, read))
            return;
        var isGlobStar = remain[0] === minimatch.GLOBSTAR;
        if (isGlobStar)
            this._processGlobStar(prefix, read, abs, remain, index, inGlobStar);
        else
            this._processReaddir(prefix, read, abs, remain, index, inGlobStar);
    };
    GlobSync.prototype._processReaddir = function (prefix, read, abs, remain, index, inGlobStar) {
        var entries = this._readdir(abs, inGlobStar);
        // if the abs isn't a dir, then nothing can match!
        if (!entries)
            return;
        // It will only match dot entries if it starts with a dot, or if
        // dot is set.  Stuff like @(.foo|.bar) isn't allowed.
        var pn = remain[0];
        var negate = !!this.minimatch.negate;
        var rawGlob = pn._glob;
        var dotOk = this.dot || rawGlob.charAt(0) === '.';
        var matchedEntries = [];
        for (var i = 0; i < entries.length; i++) {
            var e = entries[i];
            if (e.charAt(0) !== '.' || dotOk) {
                var m;
                if (negate && !prefix) {
                    m = !e.match(pn);
                }
                else {
                    m = e.match(pn);
                }
                if (m)
                    matchedEntries.push(e);
            }
        }
        var len = matchedEntries.length;
        // If there are no matched entries, then nothing matches.
        if (len === 0)
            return;
        // if this is the last remaining pattern bit, then no need for
        // an additional stat *unless* the user has specified mark or
        // stat explicitly.  We know they exist, since readdir returned
        // them.
        if (remain.length === 1 && !this.mark && !this.stat) {
            if (!this.matches[index])
                this.matches[index] = Object.create(null);
            for (var i = 0; i < len; i++) {
                var e = matchedEntries[i];
                if (prefix) {
                    if (prefix.slice(-1) !== '/')
                        e = prefix + '/' + e;
                    else
                        e = prefix + e;
                }
                if (e.charAt(0) === '/' && !this.nomount) {
                    e = path.join(this.root, e);
                }
                this._emitMatch(index, e);
            }
            // This was the last one, and no stats were needed
            return;
        }
        // now test all matched entries as stand-ins for that part
        // of the pattern.
        remain.shift();
        for (var i = 0; i < len; i++) {
            var e = matchedEntries[i];
            var newPattern;
            if (prefix)
                newPattern = [prefix, e];
            else
                newPattern = [e];
            this._process(newPattern.concat(remain), index, inGlobStar);
        }
    };
    GlobSync.prototype._emitMatch = function (index, e) {
        if (isIgnored(this, e))
            return;
        var abs = this._makeAbs(e);
        if (this.mark)
            e = this._mark(e);
        if (this.absolute) {
            e = abs;
        }
        if (this.matches[index][e])
            return;
        if (this.nodir) {
            var c = this.cache[abs];
            if (c === 'DIR' || Array.isArray(c))
                return;
        }
        this.matches[index][e] = true;
        if (this.stat)
            this._stat(e);
    };
    GlobSync.prototype._readdirInGlobStar = function (abs) {
        // follow all symlinked directories forever
        // just proceed as if this is a non-globstar situation
        if (this.follow)
            return this._readdir(abs, false);
        var entries;
        var lstat;
        var stat;
        try {
            lstat = fs.lstatSync(abs);
        }
        catch (er) {
            if (er.code === 'ENOENT') {
                // lstat failed, doesn't exist
                return null;
            }
        }
        var isSym = lstat && lstat.isSymbolicLink();
        this.symlinks[abs] = isSym;
        // If it's not a symlink or a dir, then it's definitely a regular file.
        // don't bother doing a readdir in that case.
        if (!isSym && lstat && !lstat.isDirectory())
            this.cache[abs] = 'FILE';
        else
            entries = this._readdir(abs, false);
        return entries;
    };
    GlobSync.prototype._readdir = function (abs, inGlobStar) {
        var entries;
        if (inGlobStar && !ownProp(this.symlinks, abs))
            return this._readdirInGlobStar(abs);
        if (ownProp(this.cache, abs)) {
            var c = this.cache[abs];
            if (!c || c === 'FILE')
                return null;
            if (Array.isArray(c))
                return c;
        }
        try {
            return this._readdirEntries(abs, fs.readdirSync(abs));
        }
        catch (er) {
            this._readdirError(abs, er);
            return null;
        }
    };
    GlobSync.prototype._readdirEntries = function (abs, entries) {
        // if we haven't asked to stat everything, then just
        // assume that everything in there exists, so we can avoid
        // having to stat it a second time.
        if (!this.mark && !this.stat) {
            for (var i = 0; i < entries.length; i++) {
                var e = entries[i];
                if (abs === '/')
                    e = abs + e;
                else
                    e = abs + '/' + e;
                this.cache[e] = true;
            }
        }
        this.cache[abs] = entries;
        // mark and cache dir-ness
        return entries;
    };
    GlobSync.prototype._readdirError = function (f, er) {
        // handle errors, and cache the information
        switch (er.code) {
            case 'ENOTSUP': // https://github.com/isaacs/node-glob/issues/205
            case 'ENOTDIR': // totally normal. means it *does* exist.
                var abs = this._makeAbs(f);
                this.cache[abs] = 'FILE';
                if (abs === this.cwdAbs) {
                    var error = new Error(er.code + ' invalid cwd ' + this.cwd);
                    error.path = this.cwd;
                    error.code = er.code;
                    throw error;
                }
                break;
            case 'ENOENT': // not terribly unusual
            case 'ELOOP':
            case 'ENAMETOOLONG':
            case 'UNKNOWN':
                this.cache[this._makeAbs(f)] = false;
                break;
            default: // some unusual error.  Treat as failure.
                this.cache[this._makeAbs(f)] = false;
                if (this.strict)
                    throw er;
                if (!this.silent)
                    console.error('glob error', er);
                break;
        }
    };
    GlobSync.prototype._processGlobStar = function (prefix, read, abs, remain, index, inGlobStar) {
        var entries = this._readdir(abs, inGlobStar);
        // no entries means not a dir, so it can never have matches
        // foo.txt/** doesn't match foo.txt
        if (!entries)
            return;
        // test without the globstar, and with every child both below
        // and replacing the globstar.
        var remainWithoutGlobStar = remain.slice(1);
        var gspref = prefix ? [prefix] : [];
        var noGlobStar = gspref.concat(remainWithoutGlobStar);
        // the noGlobStar pattern exits the inGlobStar state
        this._process(noGlobStar, index, false);
        var len = entries.length;
        var isSym = this.symlinks[abs];
        // If it's a symlink, and we're in a globstar, then stop
        if (isSym && inGlobStar)
            return;
        for (var i = 0; i < len; i++) {
            var e = entries[i];
            if (e.charAt(0) === '.' && !this.dot)
                continue;
            // these two cases enter the inGlobStar state
            var instead = gspref.concat(entries[i], remainWithoutGlobStar);
            this._process(instead, index, true);
            var below = gspref.concat(entries[i], remain);
            this._process(below, index, true);
        }
    };
    GlobSync.prototype._processSimple = function (prefix, index) {
        // XXX review this.  Shouldn't it be doing the mounting etc
        // before doing stat?  kinda weird?
        var exists = this._stat(prefix);
        if (!this.matches[index])
            this.matches[index] = Object.create(null);
        // If it doesn't exist, then just mark the lack of results
        if (!exists)
            return;
        if (prefix && isAbsolute(prefix) && !this.nomount) {
            var trail = /[\/\\]$/.test(prefix);
            if (prefix.charAt(0) === '/') {
                prefix = path.join(this.root, prefix);
            }
            else {
                prefix = path.resolve(this.root, prefix);
                if (trail)
                    prefix += '/';
            }
        }
        if (process.platform === 'win32')
            prefix = prefix.replace(/\\/g, '/');
        // Mark this as a match
        this._emitMatch(index, prefix);
    };
    // Returns either 'DIR', 'FILE', or false
    GlobSync.prototype._stat = function (f) {
        var abs = this._makeAbs(f);
        var needDir = f.slice(-1) === '/';
        if (f.length > this.maxLength)
            return false;
        if (!this.stat && ownProp(this.cache, abs)) {
            var c = this.cache[abs];
            if (Array.isArray(c))
                c = 'DIR';
            // It exists, but maybe not how we need it
            if (!needDir || c === 'DIR')
                return c;
            if (needDir && c === 'FILE')
                return false;
            // otherwise we have to stat, because maybe c=true
            // if we know it exists, but not what it is.
        }
        var exists;
        var stat = this.statCache[abs];
        if (!stat) {
            var lstat;
            try {
                lstat = fs.lstatSync(abs);
            }
            catch (er) {
                if (er && (er.code === 'ENOENT' || er.code === 'ENOTDIR')) {
                    this.statCache[abs] = false;
                    return false;
                }
            }
            if (lstat && lstat.isSymbolicLink()) {
                try {
                    stat = fs.statSync(abs);
                }
                catch (er) {
                    stat = lstat;
                }
            }
            else {
                stat = lstat;
            }
        }
        this.statCache[abs] = stat;
        var c = true;
        if (stat)
            c = stat.isDirectory() ? 'DIR' : 'FILE';
        this.cache[abs] = this.cache[abs] || c;
        if (needDir && c === 'FILE')
            return false;
        return c;
    };
    GlobSync.prototype._mark = function (p) {
        return common.mark(this, p);
    };
    GlobSync.prototype._makeAbs = function (f) {
        return common.makeAbs(this, f);
    };
},
/* glob/common.js */ function _(require, module, exports, __esModule, __esExport) {
    exports.setopts = setopts;
    exports.ownProp = ownProp;
    exports.makeAbs = makeAbs;
    exports.finish = finish;
    exports.mark = mark;
    exports.isIgnored = isIgnored;
    exports.childrenIgnored = childrenIgnored;
    function ownProp(obj, field) {
        return Object.prototype.hasOwnProperty.call(obj, field);
    }
    var path = require("path");
    var minimatch = require(279) /* minimatch */;
    var isAbsolute = require(717) /* path-is-absolute */;
    var Minimatch = minimatch.Minimatch;
    function alphasort(a, b) {
        return a.localeCompare(b, 'en');
    }
    function setupIgnores(self, options) {
        self.ignore = options.ignore || [];
        if (!Array.isArray(self.ignore))
            self.ignore = [self.ignore];
        if (self.ignore.length) {
            self.ignore = self.ignore.map(ignoreMap);
        }
    }
    // ignore patterns are always in dot:true mode.
    function ignoreMap(pattern) {
        var gmatcher = null;
        if (pattern.slice(-3) === '/**') {
            var gpattern = pattern.replace(/(\/\*\*)+$/, '');
            gmatcher = new Minimatch(gpattern, { dot: true });
        }
        return {
            matcher: new Minimatch(pattern, { dot: true }),
            gmatcher: gmatcher
        };
    }
    function setopts(self, pattern, options) {
        if (!options)
            options = {};
        // base-matching: just use globstar for that.
        if (options.matchBase && -1 === pattern.indexOf("/")) {
            if (options.noglobstar) {
                throw new Error("base matching requires globstar");
            }
            pattern = "**/" + pattern;
        }
        self.silent = !!options.silent;
        self.pattern = pattern;
        self.strict = options.strict !== false;
        self.realpath = !!options.realpath;
        self.realpathCache = options.realpathCache || Object.create(null);
        self.follow = !!options.follow;
        self.dot = !!options.dot;
        self.mark = !!options.mark;
        self.nodir = !!options.nodir;
        if (self.nodir)
            self.mark = true;
        self.sync = !!options.sync;
        self.nounique = !!options.nounique;
        self.nonull = !!options.nonull;
        self.nosort = !!options.nosort;
        self.nocase = !!options.nocase;
        self.stat = !!options.stat;
        self.noprocess = !!options.noprocess;
        self.absolute = !!options.absolute;
        self.maxLength = options.maxLength || Infinity;
        self.cache = options.cache || Object.create(null);
        self.statCache = options.statCache || Object.create(null);
        self.symlinks = options.symlinks || Object.create(null);
        setupIgnores(self, options);
        self.changedCwd = false;
        var cwd = process.cwd();
        if (!ownProp(options, "cwd"))
            self.cwd = cwd;
        else {
            self.cwd = path.resolve(options.cwd);
            self.changedCwd = self.cwd !== cwd;
        }
        self.root = options.root || path.resolve(self.cwd, "/");
        self.root = path.resolve(self.root);
        if (process.platform === "win32")
            self.root = self.root.replace(/\\/g, "/");
        // TODO: is an absolute `cwd` supposed to be resolved against `root`?
        // e.g. { cwd: '/test', root: __dirname } === path.join(__dirname, '/test')
        self.cwdAbs = isAbsolute(self.cwd) ? self.cwd : makeAbs(self, self.cwd);
        if (process.platform === "win32")
            self.cwdAbs = self.cwdAbs.replace(/\\/g, "/");
        self.nomount = !!options.nomount;
        // disable comments and negation in Minimatch.
        // Note that they are not supported in Glob itself anyway.
        options.nonegate = true;
        options.nocomment = true;
        self.minimatch = new Minimatch(pattern, options);
        self.options = self.minimatch.options;
    }
    function finish(self) {
        var nou = self.nounique;
        var all = nou ? [] : Object.create(null);
        for (var i = 0, l = self.matches.length; i < l; i++) {
            var matches = self.matches[i];
            if (!matches || Object.keys(matches).length === 0) {
                if (self.nonull) {
                    // do like the shell, and spit out the literal glob
                    var literal = self.minimatch.globSet[i];
                    if (nou)
                        all.push(literal);
                    else
                        all[literal] = true;
                }
            }
            else {
                // had matches
                var m = Object.keys(matches);
                if (nou)
                    all.push.apply(all, m);
                else
                    m.forEach(function (m) {
                        all[m] = true;
                    });
            }
        }
        if (!nou)
            all = Object.keys(all);
        if (!self.nosort)
            all = all.sort(alphasort);
        // at *some* point we statted all of these
        if (self.mark) {
            for (var i = 0; i < all.length; i++) {
                all[i] = self._mark(all[i]);
            }
            if (self.nodir) {
                all = all.filter(function (e) {
                    var notDir = !(/\/$/.test(e));
                    var c = self.cache[e] || self.cache[makeAbs(self, e)];
                    if (notDir && c)
                        notDir = c !== 'DIR' && !Array.isArray(c);
                    return notDir;
                });
            }
        }
        if (self.ignore.length)
            all = all.filter(function (m) {
                return !isIgnored(self, m);
            });
        self.found = all;
    }
    function mark(self, p) {
        var abs = makeAbs(self, p);
        var c = self.cache[abs];
        var m = p;
        if (c) {
            var isDir = c === 'DIR' || Array.isArray(c);
            var slash = p.slice(-1) === '/';
            if (isDir && !slash)
                m += '/';
            else if (!isDir && slash)
                m = m.slice(0, -1);
            if (m !== p) {
                var mabs = makeAbs(self, m);
                self.statCache[mabs] = self.statCache[abs];
                self.cache[mabs] = self.cache[abs];
            }
        }
        return m;
    }
    // lotta situps...
    function makeAbs(self, f) {
        var abs = f;
        if (f.charAt(0) === '/') {
            abs = path.join(self.root, f);
        }
        else if (isAbsolute(f) || f === '') {
            abs = f;
        }
        else if (self.changedCwd) {
            abs = path.resolve(self.cwd, f);
        }
        else {
            abs = path.resolve(f);
        }
        if (process.platform === 'win32')
            abs = abs.replace(/\\/g, '/');
        return abs;
    }
    // Return true, if pattern ends with globstar '**', for the accompanying parent directory.
    // Ex:- If node_modules/** is the pattern, add 'node_modules' to ignore list along with it's contents
    function isIgnored(self, path) {
        if (!self.ignore.length)
            return false;
        return self.ignore.some(function (item) {
            return item.matcher.match(path) || !!(item.gmatcher && item.gmatcher.match(path));
        });
    }
    function childrenIgnored(self, path) {
        if (!self.ignore.length)
            return false;
        return self.ignore.some(function (item) {
            return !!(item.gmatcher && item.gmatcher.match(path));
        });
    }
},
/* inflight/inflight.js */ function _(require, module, exports, __esModule, __esExport) {
    var wrappy = require(721) /* wrappy */;
    var reqs = Object.create(null);
    var once = require(722) /* once */;
    module.exports = wrappy(inflight);
    function inflight(key, cb) {
        if (reqs[key]) {
            reqs[key].push(cb);
            return null;
        }
        else {
            reqs[key] = [cb];
            return makeres(key);
        }
    }
    function makeres(key) {
        return once(function RES() {
            var cbs = reqs[key];
            var len = cbs.length;
            var args = slice(arguments);
            // XXX It's somewhat ambiguous whether a new callback added in this
            // pass should be queued for later execution if something in the
            // list of callbacks throws, or if it should just be discarded.
            // However, it's such an edge case that it hardly matters, and either
            // choice is likely as surprising as the other.
            // As it happens, we do go ahead and schedule it for later execution.
            try {
                for (var i = 0; i < len; i++) {
                    cbs[i].apply(null, args);
                }
            }
            finally {
                if (cbs.length > len) {
                    // added more in the interim.
                    // de-zalgo, just in case, but don't call again.
                    cbs.splice(0, len);
                    process.nextTick(function () {
                        RES.apply(null, args);
                    });
                }
                else {
                    delete reqs[key];
                }
            }
        });
    }
    function slice(args) {
        var length = args.length;
        var array = [];
        for (var i = 0; i < length; i++)
            array[i] = args[i];
        return array;
    }
},
/* wrappy/wrappy.js */ function _(require, module, exports, __esModule, __esExport) {
    // Returns a wrapper function that returns a wrapped callback
    // The wrapper function should do some stuff, and return a
    // presumably different callback function.
    // This makes sure that own properties are retained, so that
    // decorations and such are not lost along the way.
    module.exports = wrappy;
    function wrappy(fn, cb) {
        if (fn && cb)
            return wrappy(fn)(cb);
        if (typeof fn !== 'function')
            throw new TypeError('need wrapper function');
        Object.keys(fn).forEach(function (k) {
            wrapper[k] = fn[k];
        });
        return wrapper;
        function wrapper() {
            var args = new Array(arguments.length);
            for (var i = 0; i < args.length; i++) {
                args[i] = arguments[i];
            }
            var ret = fn.apply(this, args);
            var cb = args[args.length - 1];
            if (typeof ret === 'function' && ret !== cb) {
                Object.keys(cb).forEach(function (k) {
                    ret[k] = cb[k];
                });
            }
            return ret;
        }
    }
},
/* once/once.js */ function _(require, module, exports, __esModule, __esExport) {
    var wrappy = require(721) /* wrappy */;
    module.exports = wrappy(once);
    module.exports.strict = wrappy(onceStrict);
    once.proto = once(function () {
        Object.defineProperty(Function.prototype, 'once', {
            value: function () {
                return once(this);
            },
            configurable: true
        });
        Object.defineProperty(Function.prototype, 'onceStrict', {
            value: function () {
                return onceStrict(this);
            },
            configurable: true
        });
    });
    function once(fn) {
        var f = function () {
            if (f.called)
                return f.value;
            f.called = true;
            return f.value = fn.apply(this, arguments);
        };
        f.called = false;
        return f;
    }
    function onceStrict(fn) {
        var f = function () {
            if (f.called)
                throw new Error(f.onceError);
            f.called = true;
            return f.value = fn.apply(this, arguments);
        };
        var name = fn.name || 'Function wrapped with `once`';
        f.onceError = name + " shouldn't be called more than once";
        f.called = false;
        return f;
    }
},
/* json-stable-stringify-without-jsonify/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function (obj, opts) {
        if (!opts)
            opts = {};
        if (typeof opts === 'function')
            opts = { cmp: opts };
        var space = opts.space || '';
        if (typeof space === 'number')
            space = Array(space + 1).join(' ');
        var cycles = (typeof opts.cycles === 'boolean') ? opts.cycles : false;
        var replacer = opts.replacer || function (key, value) { return value; };
        var cmp = opts.cmp && (function (f) {
            return function (node) {
                return function (a, b) {
                    var aobj = { key: a, value: node[a] };
                    var bobj = { key: b, value: node[b] };
                    return f(aobj, bobj);
                };
            };
        })(opts.cmp);
        var seen = [];
        return (function stringify(parent, key, node, level) {
            var indent = space ? ('\n' + new Array(level + 1).join(space)) : '';
            var colonSeparator = space ? ': ' : ':';
            if (node && node.toJSON && typeof node.toJSON === 'function') {
                node = node.toJSON();
            }
            node = replacer.call(parent, key, node);
            if (node === undefined) {
                return;
            }
            if (typeof node !== 'object' || node === null) {
                return JSON.stringify(node);
            }
            if (isArray(node)) {
                var out = [];
                for (var i = 0; i < node.length; i++) {
                    var item = stringify(node, i, node[i], level + 1) || JSON.stringify(null);
                    out.push(indent + space + item);
                }
                return '[' + out.join(',') + indent + ']';
            }
            else {
                if (seen.indexOf(node) !== -1) {
                    if (cycles)
                        return JSON.stringify('__cycle__');
                    throw new TypeError('Converting circular structure to JSON');
                }
                else
                    seen.push(node);
                var keys = objectKeys(node).sort(cmp && cmp(node));
                var out = [];
                for (var i = 0; i < keys.length; i++) {
                    var key = keys[i];
                    var value = stringify(node, key, node[key], level + 1);
                    if (!value)
                        continue;
                    var keyValue = JSON.stringify(key)
                        + colonSeparator
                        + value;
                    ;
                    out.push(indent + space + keyValue);
                }
                seen.splice(seen.indexOf(node), 1);
                return '{' + out.join(',') + indent + '}';
            }
        })({ '': obj }, '', obj, 0);
    };
    var isArray = Array.isArray || function (x) {
        return {}.toString.call(x) === '[object Array]';
    };
    var objectKeys = Object.keys || function (obj) {
        var has = Object.prototype.hasOwnProperty || function () { return true; };
        var keys = [];
        for (var key in obj) {
            if (has.call(obj, key))
                keys.push(key);
        }
        return keys;
    };
},
/* eslint/lib/eslint/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const { ESLint } = require(725) /* ./eslint */;
    module.exports = {
        ESLint
    };
},
/* eslint/lib/eslint/eslint.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const path = require("path");
    const fs = require("fs");
    const { promisify } = require("util");
    const { CLIEngine, getCLIEngineInternalSlots } = require(205) /* ../cli-engine/cli-engine */;
    const BuiltinRules = require(391) /* ../rules */;
    const { Legacy: { ConfigOps: { getRuleSeverity } } } = require(208) /* @eslint/eslintrc */;
    const { version } = require(207) /* ../../package.json */;
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    /** @typedef {import("../cli-engine/cli-engine").LintReport} CLIEngineLintReport */
    /** @typedef {import("../shared/types").DeprecatedRuleInfo} DeprecatedRuleInfo */
    /** @typedef {import("../shared/types").ConfigData} ConfigData */
    /** @typedef {import("../shared/types").LintMessage} LintMessage */
    /** @typedef {import("../shared/types").Plugin} Plugin */
    /** @typedef {import("../shared/types").Rule} Rule */
    /** @typedef {import("./load-formatter").Formatter} Formatter */
    /**
     * The options with which to configure the ESLint instance.
     * @typedef {Object} ESLintOptions
     * @property {boolean} [allowInlineConfig] Enable or disable inline configuration comments.
     * @property {ConfigData} [baseConfig] Base config object, extended by all configs used with this instance
     * @property {boolean} [cache] Enable result caching.
     * @property {string} [cacheLocation] The cache file to use instead of .eslintcache.
     * @property {"metadata" | "content"} [cacheStrategy] The strategy used to detect changed files.
     * @property {string} [cwd] The value to use for the current working directory.
     * @property {boolean} [errorOnUnmatchedPattern] If `false` then `ESLint#lintFiles()` doesn't throw even if no target files found. Defaults to `true`.
     * @property {string[]} [extensions] An array of file extensions to check.
     * @property {boolean|Function} [fix] Execute in autofix mode. If a function, should return a boolean.
     * @property {string[]} [fixTypes] Array of rule types to apply fixes for.
     * @property {boolean} [globInputPaths] Set to false to skip glob resolution of input file paths to lint (default: true). If false, each input file paths is assumed to be a non-glob path to an existing file.
     * @property {boolean} [ignore] False disables use of .eslintignore.
     * @property {string} [ignorePath] The ignore file to use instead of .eslintignore.
     * @property {ConfigData} [overrideConfig] Override config object, overrides all configs used with this instance
     * @property {string} [overrideConfigFile] The configuration file to use.
     * @property {Record<string,Plugin>} [plugins] An array of plugin implementations.
     * @property {"error" | "warn" | "off"} [reportUnusedDisableDirectives] the severity to report unused eslint-disable directives.
     * @property {string} [resolvePluginsRelativeTo] The folder where plugins should be resolved from, defaulting to the CWD.
     * @property {string[]} [rulePaths] An array of directories to load custom rules from.
     * @property {boolean} [useEslintrc] False disables looking for .eslintrc.* files.
     */
    /**
     * A rules metadata object.
     * @typedef {Object} RulesMeta
     * @property {string} id The plugin ID.
     * @property {Object} definition The plugin definition.
     */
    /**
     * A linting result.
     * @typedef {Object} LintResult
     * @property {string} filePath The path to the file that was linted.
     * @property {LintMessage[]} messages All of the messages for the result.
     * @property {number} errorCount Number of errors for the result.
     * @property {number} warningCount Number of warnings for the result.
     * @property {number} fixableErrorCount Number of fixable errors for the result.
     * @property {number} fixableWarningCount Number of fixable warnings for the result.
     * @property {string} [source] The source code of the file that was linted.
     * @property {string} [output] The source code of the file that was linted, with as many fixes applied as possible.
     * @property {DeprecatedRuleInfo[]} usedDeprecatedRules The list of used deprecated rules.
     */
    /**
     * Private members for the `ESLint` instance.
     * @typedef {Object} ESLintPrivateMembers
     * @property {CLIEngine} cliEngine The wrapped CLIEngine instance.
     * @property {ESLintOptions} options The options used to instantiate the ESLint instance.
     */
    //------------------------------------------------------------------------------
    // Helpers
    //------------------------------------------------------------------------------
    const writeFile = promisify(fs.writeFile);
    /**
     * The map with which to store private class members.
     * @type {WeakMap<ESLint, ESLintPrivateMembers>}
     */
    const privateMembersMap = new WeakMap();
    /**
     * Check if a given value is a non-empty string or not.
     * @param {any} x The value to check.
     * @returns {boolean} `true` if `x` is a non-empty string.
     */
    function isNonEmptyString(x) {
        return typeof x === "string" && x.trim() !== "";
    }
    /**
     * Check if a given value is an array of non-empty stringss or not.
     * @param {any} x The value to check.
     * @returns {boolean} `true` if `x` is an array of non-empty stringss.
     */
    function isArrayOfNonEmptyString(x) {
        return Array.isArray(x) && x.every(isNonEmptyString);
    }
    /**
     * Check if a given value is a valid fix type or not.
     * @param {any} x The value to check.
     * @returns {boolean} `true` if `x` is valid fix type.
     */
    function isFixType(x) {
        return x === "problem" || x === "suggestion" || x === "layout";
    }
    /**
     * Check if a given value is an array of fix types or not.
     * @param {any} x The value to check.
     * @returns {boolean} `true` if `x` is an array of fix types.
     */
    function isFixTypeArray(x) {
        return Array.isArray(x) && x.every(isFixType);
    }
    /**
     * The error for invalid options.
     */
    class ESLintInvalidOptionsError extends Error {
        constructor(messages) {
            super(`Invalid Options:\n- ${messages.join("\n- ")}`);
            this.code = "ESLINT_INVALID_OPTIONS";
            Error.captureStackTrace(this, ESLintInvalidOptionsError);
        }
    }
    /**
     * Validates and normalizes options for the wrapped CLIEngine instance.
     * @param {ESLintOptions} options The options to process.
     * @returns {ESLintOptions} The normalized options.
     */
    function processOptions({ allowInlineConfig = true, // ← we cannot use `overrideConfig.noInlineConfig` instead because `allowInlineConfig` has side-effect that suppress warnings that show inline configs are ignored.
    baseConfig = null, cache = false, cacheLocation = ".eslintcache", cacheStrategy = "metadata", cwd = process.cwd(), errorOnUnmatchedPattern = true, extensions = null, // ← should be null by default because if it's an array then it suppresses RFC20 feature.
    fix = false, fixTypes = null, // ← should be null by default because if it's an array then it suppresses rules that don't have the `meta.type` property.
    globInputPaths = true, ignore = true, ignorePath = null, // ← should be null by default because if it's a string then it may throw ENOENT.
    overrideConfig = null, overrideConfigFile = null, plugins = {}, reportUnusedDisableDirectives = null, // ← should be null by default because if it's a string then it overrides the 'reportUnusedDisableDirectives' setting in config files. And we cannot use `overrideConfig.reportUnusedDisableDirectives` instead because we cannot configure the `error` severity with that.
    resolvePluginsRelativeTo = null, // ← should be null by default because if it's a string then it suppresses RFC47 feature.
    rulePaths = [], useEslintrc = true, ...unknownOptions }) {
        const errors = [];
        const unknownOptionKeys = Object.keys(unknownOptions);
        if (unknownOptionKeys.length >= 1) {
            errors.push(`Unknown options: ${unknownOptionKeys.join(", ")}`);
            if (unknownOptionKeys.includes("cacheFile")) {
                errors.push("'cacheFile' has been removed. Please use the 'cacheLocation' option instead.");
            }
            if (unknownOptionKeys.includes("configFile")) {
                errors.push("'configFile' has been removed. Please use the 'overrideConfigFile' option instead.");
            }
            if (unknownOptionKeys.includes("envs")) {
                errors.push("'envs' has been removed. Please use the 'overrideConfig.env' option instead.");
            }
            if (unknownOptionKeys.includes("globals")) {
                errors.push("'globals' has been removed. Please use the 'overrideConfig.globals' option instead.");
            }
            if (unknownOptionKeys.includes("ignorePattern")) {
                errors.push("'ignorePattern' has been removed. Please use the 'overrideConfig.ignorePatterns' option instead.");
            }
            if (unknownOptionKeys.includes("parser")) {
                errors.push("'parser' has been removed. Please use the 'overrideConfig.parser' option instead.");
            }
            if (unknownOptionKeys.includes("parserOptions")) {
                errors.push("'parserOptions' has been removed. Please use the 'overrideConfig.parserOptions' option instead.");
            }
            if (unknownOptionKeys.includes("rules")) {
                errors.push("'rules' has been removed. Please use the 'overrideConfig.rules' option instead.");
            }
        }
        if (typeof allowInlineConfig !== "boolean") {
            errors.push("'allowInlineConfig' must be a boolean.");
        }
        if (typeof baseConfig !== "object") {
            errors.push("'baseConfig' must be an object or null.");
        }
        if (typeof cache !== "boolean") {
            errors.push("'cache' must be a boolean.");
        }
        if (!isNonEmptyString(cacheLocation)) {
            errors.push("'cacheLocation' must be a non-empty string.");
        }
        if (cacheStrategy !== "metadata" &&
            cacheStrategy !== "content") {
            errors.push("'cacheStrategy' must be any of \"metadata\", \"content\".");
        }
        if (!isNonEmptyString(cwd) || !path.isAbsolute(cwd)) {
            errors.push("'cwd' must be an absolute path.");
        }
        if (typeof errorOnUnmatchedPattern !== "boolean") {
            errors.push("'errorOnUnmatchedPattern' must be a boolean.");
        }
        if (!isArrayOfNonEmptyString(extensions) && extensions !== null) {
            errors.push("'extensions' must be an array of non-empty strings or null.");
        }
        if (typeof fix !== "boolean" && typeof fix !== "function") {
            errors.push("'fix' must be a boolean or a function.");
        }
        if (fixTypes !== null && !isFixTypeArray(fixTypes)) {
            errors.push("'fixTypes' must be an array of any of \"problem\", \"suggestion\", and \"layout\".");
        }
        if (typeof globInputPaths !== "boolean") {
            errors.push("'globInputPaths' must be a boolean.");
        }
        if (typeof ignore !== "boolean") {
            errors.push("'ignore' must be a boolean.");
        }
        if (!isNonEmptyString(ignorePath) && ignorePath !== null) {
            errors.push("'ignorePath' must be a non-empty string or null.");
        }
        if (typeof overrideConfig !== "object") {
            errors.push("'overrideConfig' must be an object or null.");
        }
        if (!isNonEmptyString(overrideConfigFile) && overrideConfigFile !== null) {
            errors.push("'overrideConfigFile' must be a non-empty string or null.");
        }
        if (typeof plugins !== "object") {
            errors.push("'plugins' must be an object or null.");
        }
        else if (plugins !== null && Object.keys(plugins).includes("")) {
            errors.push("'plugins' must not include an empty string.");
        }
        if (Array.isArray(plugins)) {
            errors.push("'plugins' doesn't add plugins to configuration to load. Please use the 'overrideConfig.plugins' option instead.");
        }
        if (reportUnusedDisableDirectives !== "error" &&
            reportUnusedDisableDirectives !== "warn" &&
            reportUnusedDisableDirectives !== "off" &&
            reportUnusedDisableDirectives !== null) {
            errors.push("'reportUnusedDisableDirectives' must be any of \"error\", \"warn\", \"off\", and null.");
        }
        if (!isNonEmptyString(resolvePluginsRelativeTo) &&
            resolvePluginsRelativeTo !== null) {
            errors.push("'resolvePluginsRelativeTo' must be a non-empty string or null.");
        }
        if (!isArrayOfNonEmptyString(rulePaths)) {
            errors.push("'rulePaths' must be an array of non-empty strings.");
        }
        if (typeof useEslintrc !== "boolean") {
            errors.push("'useEslintrc' must be a boolean.");
        }
        if (errors.length > 0) {
            throw new ESLintInvalidOptionsError(errors);
        }
        return {
            allowInlineConfig,
            baseConfig,
            cache,
            cacheLocation,
            cacheStrategy,
            configFile: overrideConfigFile,
            cwd,
            errorOnUnmatchedPattern,
            extensions,
            fix,
            fixTypes,
            globInputPaths,
            ignore,
            ignorePath,
            reportUnusedDisableDirectives,
            resolvePluginsRelativeTo,
            rulePaths,
            useEslintrc
        };
    }
    /**
     * Check if a value has one or more properties and that value is not undefined.
     * @param {any} obj The value to check.
     * @returns {boolean} `true` if `obj` has one or more properties that that value is not undefined.
     */
    function hasDefinedProperty(obj) {
        if (typeof obj === "object" && obj !== null) {
            for (const key in obj) {
                if (typeof obj[key] !== "undefined") {
                    return true;
                }
            }
        }
        return false;
    }
    /**
     * Create rulesMeta object.
     * @param {Map<string,Rule>} rules a map of rules from which to generate the object.
     * @returns {Object} metadata for all enabled rules.
     */
    function createRulesMeta(rules) {
        return Array.from(rules).reduce((retVal, [id, rule]) => {
            retVal[id] = rule.meta;
            return retVal;
        }, {});
    }
    /** @type {WeakMap<ExtractedConfig, DeprecatedRuleInfo[]>} */
    const usedDeprecatedRulesCache = new WeakMap();
    /**
     * Create used deprecated rule list.
     * @param {CLIEngine} cliEngine The CLIEngine instance.
     * @param {string} maybeFilePath The absolute path to a lint target file or `"<text>"`.
     * @returns {DeprecatedRuleInfo[]} The used deprecated rule list.
     */
    function getOrFindUsedDeprecatedRules(cliEngine, maybeFilePath) {
        const { configArrayFactory, options: { cwd } } = getCLIEngineInternalSlots(cliEngine);
        const filePath = path.isAbsolute(maybeFilePath)
            ? maybeFilePath
            : path.join(cwd, "__placeholder__.js");
        const configArray = configArrayFactory.getConfigArrayForFile(filePath);
        const config = configArray.extractConfig(filePath);
        // Most files use the same config, so cache it.
        if (!usedDeprecatedRulesCache.has(config)) {
            const pluginRules = configArray.pluginRules;
            const retv = [];
            for (const [ruleId, ruleConf] of Object.entries(config.rules)) {
                if (getRuleSeverity(ruleConf) === 0) {
                    continue;
                }
                const rule = pluginRules.get(ruleId) || BuiltinRules.get(ruleId);
                const meta = rule && rule.meta;
                if (meta && meta.deprecated) {
                    retv.push({ ruleId, replacedBy: meta.replacedBy || [] });
                }
            }
            usedDeprecatedRulesCache.set(config, Object.freeze(retv));
        }
        return usedDeprecatedRulesCache.get(config);
    }
    /**
     * Processes the linting results generated by a CLIEngine linting report to
     * match the ESLint class's API.
     * @param {CLIEngine} cliEngine The CLIEngine instance.
     * @param {CLIEngineLintReport} report The CLIEngine linting report to process.
     * @returns {LintResult[]} The processed linting results.
     */
    function processCLIEngineLintReport(cliEngine, { results }) {
        const descriptor = {
            configurable: true,
            enumerable: true,
            get() {
                return getOrFindUsedDeprecatedRules(cliEngine, this.filePath);
            }
        };
        for (const result of results) {
            Object.defineProperty(result, "usedDeprecatedRules", descriptor);
        }
        return results;
    }
    /**
     * An Array.prototype.sort() compatible compare function to order results by their file path.
     * @param {LintResult} a The first lint result.
     * @param {LintResult} b The second lint result.
     * @returns {number} An integer representing the order in which the two results should occur.
     */
    function compareResultsByFilePath(a, b) {
        if (a.filePath < b.filePath) {
            return -1;
        }
        if (a.filePath > b.filePath) {
            return 1;
        }
        return 0;
    }
    class ESLint {
        /**
         * Creates a new instance of the main ESLint API.
         * @param {ESLintOptions} options The options for this instance.
         */
        constructor(options = {}) {
            const processedOptions = processOptions(options);
            const cliEngine = new CLIEngine(processedOptions);
            const { additionalPluginPool, configArrayFactory, lastConfigArrays } = getCLIEngineInternalSlots(cliEngine);
            let updated = false;
            /*
             * Address `plugins` to add plugin implementations.
             * Operate the `additionalPluginPool` internal slot directly to avoid
             * using `addPlugin(id, plugin)` method that resets cache everytime.
             */
            if (options.plugins) {
                for (const [id, plugin] of Object.entries(options.plugins)) {
                    additionalPluginPool.set(id, plugin);
                    updated = true;
                }
            }
            /*
             * Address `overrideConfig` to set override config.
             * Operate the `configArrayFactory` internal slot directly because this
             * functionality doesn't exist as the public API of CLIEngine.
             */
            if (hasDefinedProperty(options.overrideConfig)) {
                configArrayFactory.setOverrideConfig(options.overrideConfig);
                updated = true;
            }
            // Update caches.
            if (updated) {
                configArrayFactory.clearCache();
                lastConfigArrays[0] = configArrayFactory.getConfigArrayForFile();
            }
            // Initialize private properties.
            privateMembersMap.set(this, {
                cliEngine,
                options: processedOptions
            });
        }
        /**
         * The version text.
         * @type {string}
         */
        static get version() {
            return version;
        }
        /**
         * Outputs fixes from the given results to files.
         * @param {LintResult[]} results The lint results.
         * @returns {Promise<void>} Returns a promise that is used to track side effects.
         */
        static async outputFixes(results) {
            if (!Array.isArray(results)) {
                throw new Error("'results' must be an array");
            }
            await Promise.all(results
                .filter(result => {
                if (typeof result !== "object" || result === null) {
                    throw new Error("'results' must include only objects");
                }
                return (typeof result.output === "string" &&
                    path.isAbsolute(result.filePath));
            })
                .map(r => writeFile(r.filePath, r.output)));
        }
        /**
         * Returns results that only contains errors.
         * @param {LintResult[]} results The results to filter.
         * @returns {LintResult[]} The filtered results.
         */
        static getErrorResults(results) {
            return CLIEngine.getErrorResults(results);
        }
        /**
         * Returns meta objects for each rule represented in the lint results.
         * @param {LintResult[]} results The results to fetch rules meta for.
         * @returns {Object} A mapping of ruleIds to rule meta objects.
         */
        getRulesMetaForResults(results) {
            const resultRuleIds = new Set();
            // first gather all ruleIds from all results
            for (const result of results) {
                for (const { ruleId } of result.messages) {
                    resultRuleIds.add(ruleId);
                }
            }
            // create a map of all rules in the results
            const { cliEngine } = privateMembersMap.get(this);
            const rules = cliEngine.getRules();
            const resultRules = new Map();
            for (const [ruleId, rule] of rules) {
                if (resultRuleIds.has(ruleId)) {
                    resultRules.set(ruleId, rule);
                }
            }
            return createRulesMeta(resultRules);
        }
        /**
         * Executes the current configuration on an array of file and directory names.
         * @param {string[]} patterns An array of file and directory names.
         * @returns {Promise<LintResult[]>} The results of linting the file patterns given.
         */
        async lintFiles(patterns) {
            if (!isNonEmptyString(patterns) && !isArrayOfNonEmptyString(patterns)) {
                throw new Error("'patterns' must be a non-empty string or an array of non-empty strings");
            }
            const { cliEngine } = privateMembersMap.get(this);
            return processCLIEngineLintReport(cliEngine, cliEngine.executeOnFiles(patterns));
        }
        /**
         * Executes the current configuration on text.
         * @param {string} code A string of JavaScript code to lint.
         * @param {Object} [options] The options.
         * @param {string} [options.filePath] The path to the file of the source code.
         * @param {boolean} [options.warnIgnored] When set to true, warn if given filePath is an ignored path.
         * @returns {Promise<LintResult[]>} The results of linting the string of code given.
         */
        async lintText(code, options = {}) {
            if (typeof code !== "string") {
                throw new Error("'code' must be a string");
            }
            if (typeof options !== "object") {
                throw new Error("'options' must be an object, null, or undefined");
            }
            const { filePath, warnIgnored = false, ...unknownOptions } = options || {};
            const unknownOptionKeys = Object.keys(unknownOptions);
            if (unknownOptionKeys.length > 0) {
                throw new Error(`'options' must not include the unknown option(s): ${unknownOptionKeys.join(", ")}`);
            }
            if (filePath !== void 0 && !isNonEmptyString(filePath)) {
                throw new Error("'options.filePath' must be a non-empty string or undefined");
            }
            if (typeof warnIgnored !== "boolean") {
                throw new Error("'options.warnIgnored' must be a boolean or undefined");
            }
            const { cliEngine } = privateMembersMap.get(this);
            return processCLIEngineLintReport(cliEngine, cliEngine.executeOnText(code, filePath, warnIgnored));
        }
        /**
         * Returns the formatter representing the given formatter name.
         * @param {string} [name] The name of the formatter to load.
         * The following values are allowed:
         * - `undefined` ... Load `stylish` builtin formatter.
         * - A builtin formatter name ... Load the builtin formatter.
         * - A thirdparty formatter name:
         *   - `foo` → `eslint-formatter-foo`
         *   - `@foo` → `@foo/eslint-formatter`
         *   - `@foo/bar` → `@foo/eslint-formatter-bar`
         * - A file path ... Load the file.
         * @returns {Promise<Formatter>} A promise resolving to the formatter object.
         * This promise will be rejected if the given formatter was not found or not
         * a function.
         */
        async loadFormatter(name = "stylish") {
            if (typeof name !== "string") {
                throw new Error("'name' must be a string");
            }
            const { cliEngine } = privateMembersMap.get(this);
            const formatter = cliEngine.getFormatter(name);
            if (typeof formatter !== "function") {
                throw new Error(`Formatter must be a function, but got a ${typeof formatter}.`);
            }
            return {
                /**
                 * The main formatter method.
                 * @param {LintResults[]} results The lint results to format.
                 * @returns {string} The formatted lint results.
                 */
                format(results) {
                    let rulesMeta = null;
                    results.sort(compareResultsByFilePath);
                    return formatter(results, {
                        get rulesMeta() {
                            if (!rulesMeta) {
                                rulesMeta = createRulesMeta(cliEngine.getRules());
                            }
                            return rulesMeta;
                        }
                    });
                }
            };
        }
        /**
         * Returns a configuration object for the given file based on the CLI options.
         * This is the same logic used by the ESLint CLI executable to determine
         * configuration for each file it processes.
         * @param {string} filePath The path of the file to retrieve a config object for.
         * @returns {Promise<ConfigData>} A configuration object for the file.
         */
        async calculateConfigForFile(filePath) {
            if (!isNonEmptyString(filePath)) {
                throw new Error("'filePath' must be a non-empty string");
            }
            const { cliEngine } = privateMembersMap.get(this);
            return cliEngine.getConfigForFile(filePath);
        }
        /**
         * Checks if a given path is ignored by ESLint.
         * @param {string} filePath The path of the file to check.
         * @returns {Promise<boolean>} Whether or not the given path is ignored.
         */
        async isPathIgnored(filePath) {
            if (!isNonEmptyString(filePath)) {
                throw new Error("'filePath' must be a non-empty string");
            }
            const { cliEngine } = privateMembersMap.get(this);
            return cliEngine.isPathIgnored(filePath);
        }
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = {
        ESLint,
        /**
         * Get the private class members of a given ESLint instance for tests.
         * @param {ESLint} instance The ESLint instance to get.
         * @returns {ESLintPrivateMembers} The instance's private class members.
         */
        getESLintPrivateMembers(instance) {
            return privateMembersMap.get(instance);
        }
    };
},
/* eslint/lib/rule-tester/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        RuleTester: require(727) /* ./rule-tester */
    };
},
/* eslint/lib/rule-tester/rule-tester.js */ function _(require, module, exports, __esModule, __esExport) {
    /* global describe, it */
    /*
     * This is a wrapper around mocha to allow for DRY unittests for eslint
     * Format:
     * RuleTester.run("{ruleName}", {
     *      valid: [
     *          "{code}",
     *          { code: "{code}", options: {options}, globals: {globals}, parser: "{parser}", settings: {settings} }
     *      ],
     *      invalid: [
     *          { code: "{code}", errors: {numErrors} },
     *          { code: "{code}", errors: ["{errorMessage}"] },
     *          { code: "{code}", options: {options}, globals: {globals}, parser: "{parser}", settings: {settings}, errors: [{ message: "{errorMessage}", type: "{errorNodeType}"}] }
     *      ]
     *  });
     *
     * Variables:
     * {code} - String that represents the code to be tested
     * {options} - Arguments that are passed to the configurable rules.
     * {globals} - An object representing a list of variables that are
     *             registered as globals
     * {parser} - String representing the parser to use
     * {settings} - An object representing global settings for all rules
     * {numErrors} - If failing case doesn't need to check error message,
     *               this integer will specify how many errors should be
     *               received
     * {errorMessage} - Message that is returned by the rule on failure
     * {errorNodeType} - AST node type that is returned by they rule as
     *                   a cause of the failure.
     */
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const assert = require("assert"), path = require("path"), util = require("util"), merge = require(345) /* lodash.merge */, equal = require(227) /* fast-deep-equal */, Traverser = require(347) /* ../../lib/shared/traverser */, { getRuleOptionsSchema, validate } = require(728) /* ../shared/config-validator */, { Linter, SourceCodeFixer, interpolate } = require(330) /* ../linter */;
    const ajv = require(731) /* ../shared/ajv */({ strictDefaults: true });
    const espreePath = require.resolve("espree");
    const parserSymbol = Symbol.for("eslint.RuleTester.parser");
    //------------------------------------------------------------------------------
    // Typedefs
    //------------------------------------------------------------------------------
    /** @typedef {import("../shared/types").Parser} Parser */
    /**
     * A test case that is expected to pass lint.
     * @typedef {Object} ValidTestCase
     * @property {string} code Code for the test case.
     * @property {any[]} [options] Options for the test case.
     * @property {{ [name: string]: any }} [settings] Settings for the test case.
     * @property {string} [filename] The fake filename for the test case. Useful for rules that make assertion about filenames.
     * @property {string} [parser] The absolute path for the parser.
     * @property {{ [name: string]: any }} [parserOptions] Options for the parser.
     * @property {{ [name: string]: "readonly" | "writable" | "off" }} [globals] The additional global variables.
     * @property {{ [name: string]: boolean }} [env] Environments for the test case.
     * @property {boolean} [only] Run only this test case or the subset of test cases with this property.
     */
    /**
     * A test case that is expected to fail lint.
     * @typedef {Object} InvalidTestCase
     * @property {string} code Code for the test case.
     * @property {number | Array<TestCaseError | string | RegExp>} errors Expected errors.
     * @property {string | null} [output] The expected code after autofixes are applied. If set to `null`, the test runner will assert that no autofix is suggested.
     * @property {any[]} [options] Options for the test case.
     * @property {{ [name: string]: any }} [settings] Settings for the test case.
     * @property {string} [filename] The fake filename for the test case. Useful for rules that make assertion about filenames.
     * @property {string} [parser] The absolute path for the parser.
     * @property {{ [name: string]: any }} [parserOptions] Options for the parser.
     * @property {{ [name: string]: "readonly" | "writable" | "off" }} [globals] The additional global variables.
     * @property {{ [name: string]: boolean }} [env] Environments for the test case.
     * @property {boolean} [only] Run only this test case or the subset of test cases with this property.
     */
    /**
     * A description of a reported error used in a rule tester test.
     * @typedef {Object} TestCaseError
     * @property {string | RegExp} [message] Message.
     * @property {string} [messageId] Message ID.
     * @property {string} [type] The type of the reported AST node.
     * @property {{ [name: string]: string }} [data] The data used to fill the message template.
     * @property {number} [line] The 1-based line number of the reported start location.
     * @property {number} [column] The 1-based column number of the reported start location.
     * @property {number} [endLine] The 1-based line number of the reported end location.
     * @property {number} [endColumn] The 1-based column number of the reported end location.
     */
    //------------------------------------------------------------------------------
    // Private Members
    //------------------------------------------------------------------------------
    /*
     * testerDefaultConfig must not be modified as it allows to reset the tester to
     * the initial default configuration
     */
    const testerDefaultConfig = { rules: {} };
    let defaultConfig = { rules: {} };
    /*
     * List every parameters possible on a test case that are not related to eslint
     * configuration
     */
    const RuleTesterParameters = [
        "code",
        "filename",
        "options",
        "errors",
        "output",
        "only"
    ];
    /*
     * All allowed property names in error objects.
     */
    const errorObjectParameters = new Set([
        "message",
        "messageId",
        "data",
        "type",
        "line",
        "column",
        "endLine",
        "endColumn",
        "suggestions"
    ]);
    const friendlyErrorObjectParameterList = `[${[...errorObjectParameters].map(key => `'${key}'`).join(", ")}]`;
    /*
     * All allowed property names in suggestion objects.
     */
    const suggestionObjectParameters = new Set([
        "desc",
        "messageId",
        "data",
        "output"
    ]);
    const friendlySuggestionObjectParameterList = `[${[...suggestionObjectParameters].map(key => `'${key}'`).join(", ")}]`;
    const hasOwnProperty = Function.call.bind(Object.hasOwnProperty);
    /**
     * Clones a given value deeply.
     * Note: This ignores `parent` property.
     * @param {any} x A value to clone.
     * @returns {any} A cloned value.
     */
    function cloneDeeplyExcludesParent(x) {
        if (typeof x === "object" && x !== null) {
            if (Array.isArray(x)) {
                return x.map(cloneDeeplyExcludesParent);
            }
            const retv = {};
            for (const key in x) {
                if (key !== "parent" && hasOwnProperty(x, key)) {
                    retv[key] = cloneDeeplyExcludesParent(x[key]);
                }
            }
            return retv;
        }
        return x;
    }
    /**
     * Freezes a given value deeply.
     * @param {any} x A value to freeze.
     * @returns {void}
     */
    function freezeDeeply(x) {
        if (typeof x === "object" && x !== null) {
            if (Array.isArray(x)) {
                x.forEach(freezeDeeply);
            }
            else {
                for (const key in x) {
                    if (key !== "parent" && hasOwnProperty(x, key)) {
                        freezeDeeply(x[key]);
                    }
                }
            }
            Object.freeze(x);
        }
    }
    /**
     * Replace control characters by `\u00xx` form.
     * @param {string} text The text to sanitize.
     * @returns {string} The sanitized text.
     */
    function sanitize(text) {
        return text.replace(/[\u0000-\u0009\u000b-\u001a]/gu, // eslint-disable-line no-control-regex
        // eslint-disable-line no-control-regex
        c => `\\u${c.codePointAt(0).toString(16).padStart(4, "0")}`);
    }
    /**
     * Define `start`/`end` properties as throwing error.
     * @param {string} objName Object name used for error messages.
     * @param {ASTNode} node The node to define.
     * @returns {void}
     */
    function defineStartEndAsError(objName, node) {
        Object.defineProperties(node, {
            start: {
                get() {
                    throw new Error(`Use ${objName}.range[0] instead of ${objName}.start`);
                },
                configurable: true,
                enumerable: false
            },
            end: {
                get() {
                    throw new Error(`Use ${objName}.range[1] instead of ${objName}.end`);
                },
                configurable: true,
                enumerable: false
            }
        });
    }
    /**
     * Define `start`/`end` properties of all nodes of the given AST as throwing error.
     * @param {ASTNode} ast The root node to errorize `start`/`end` properties.
     * @param {Object} [visitorKeys] Visitor keys to be used for traversing the given ast.
     * @returns {void}
     */
    function defineStartEndAsErrorInTree(ast, visitorKeys) {
        Traverser.traverse(ast, { visitorKeys, enter: defineStartEndAsError.bind(null, "node") });
        ast.tokens.forEach(defineStartEndAsError.bind(null, "token"));
        ast.comments.forEach(defineStartEndAsError.bind(null, "token"));
    }
    /**
     * Wraps the given parser in order to intercept and modify return values from the `parse` and `parseForESLint` methods, for test purposes.
     * In particular, to modify ast nodes, tokens and comments to throw on access to their `start` and `end` properties.
     * @param {Parser} parser Parser object.
     * @returns {Parser} Wrapped parser object.
     */
    function wrapParser(parser) {
        if (typeof parser.parseForESLint === "function") {
            return {
                [parserSymbol]: parser,
                parseForESLint(...args) {
                    const ret = parser.parseForESLint(...args);
                    defineStartEndAsErrorInTree(ret.ast, ret.visitorKeys);
                    return ret;
                }
            };
        }
        return {
            [parserSymbol]: parser,
            parse(...args) {
                const ast = parser.parse(...args);
                defineStartEndAsErrorInTree(ast);
                return ast;
            }
        };
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    // default separators for testing
    const DESCRIBE = Symbol("describe");
    const IT = Symbol("it");
    const IT_ONLY = Symbol("itOnly");
    /**
     * This is `it` default handler if `it` don't exist.
     * @this {Mocha}
     * @param {string} text The description of the test case.
     * @param {Function} method The logic of the test case.
     * @returns {any} Returned value of `method`.
     */
    function itDefaultHandler(text, method) {
        try {
            return method.call(this);
        }
        catch (err) {
            if (err instanceof assert.AssertionError) {
                err.message += ` (${util.inspect(err.actual)} ${err.operator} ${util.inspect(err.expected)})`;
            }
            throw err;
        }
    }
    /**
     * This is `describe` default handler if `describe` don't exist.
     * @this {Mocha}
     * @param {string} text The description of the test case.
     * @param {Function} method The logic of the test case.
     * @returns {any} Returned value of `method`.
     */
    function describeDefaultHandler(text, method) {
        return method.call(this);
    }
    class RuleTester {
        /**
         * Creates a new instance of RuleTester.
         * @param {Object} [testerConfig] Optional, extra configuration for the tester
         */
        constructor(testerConfig) {
            /**
             * The configuration to use for this tester. Combination of the tester
             * configuration and the default configuration.
             * @type {Object}
             */
            this.testerConfig = merge({}, defaultConfig, testerConfig, { rules: { "rule-tester/validate-ast": "error" } });
            /**
             * Rule definitions to define before tests.
             * @type {Object}
             */
            this.rules = {};
            this.linter = new Linter();
        }
        /**
         * Set the configuration to use for all future tests
         * @param {Object} config the configuration to use.
         * @returns {void}
         */
        static setDefaultConfig(config) {
            if (typeof config !== "object") {
                throw new TypeError("RuleTester.setDefaultConfig: config must be an object");
            }
            defaultConfig = config;
            // Make sure the rules object exists since it is assumed to exist later
            defaultConfig.rules = defaultConfig.rules || {};
        }
        /**
         * Get the current configuration used for all tests
         * @returns {Object} the current configuration
         */
        static getDefaultConfig() {
            return defaultConfig;
        }
        /**
         * Reset the configuration to the initial configuration of the tester removing
         * any changes made until now.
         * @returns {void}
         */
        static resetDefaultConfig() {
            defaultConfig = merge({}, testerDefaultConfig);
        }
        /*
         * If people use `mocha test.js --watch` command, `describe` and `it` function
         * instances are different for each execution. So `describe` and `it` should get fresh instance
         * always.
         */
        static get describe() {
            return (this[DESCRIBE] ||
                (typeof describe === "function" ? describe : describeDefaultHandler));
        }
        static set describe(value) {
            this[DESCRIBE] = value;
        }
        static get it() {
            return (this[IT] ||
                (typeof it === "function" ? it : itDefaultHandler));
        }
        static set it(value) {
            this[IT] = value;
        }
        /**
         * Adds the `only` property to a test to run it in isolation.
         * @param {string | ValidTestCase | InvalidTestCase} item A single test to run by itself.
         * @returns {ValidTestCase | InvalidTestCase} The test with `only` set.
         */
        static only(item) {
            if (typeof item === "string") {
                return { code: item, only: true };
            }
            return { ...item, only: true };
        }
        static get itOnly() {
            if (typeof this[IT_ONLY] === "function") {
                return this[IT_ONLY];
            }
            if (typeof this[IT] === "function" && typeof this[IT].only === "function") {
                return Function.bind.call(this[IT].only, this[IT]);
            }
            if (typeof it === "function" && typeof it.only === "function") {
                return Function.bind.call(it.only, it);
            }
            if (typeof this[DESCRIBE] === "function" || typeof this[IT] === "function") {
                throw new Error("Set `RuleTester.itOnly` to use `only` with a custom test framework.\n" +
                    "See https://eslint.org/docs/developer-guide/nodejs-api#customizing-ruletester for more.");
            }
            if (typeof it === "function") {
                throw new Error("The current test framework does not support exclusive tests with `only`.");
            }
            throw new Error("To use `only`, use RuleTester with a test framework that provides `it.only()` like Mocha.");
        }
        static set itOnly(value) {
            this[IT_ONLY] = value;
        }
        /**
         * Define a rule for one particular run of tests.
         * @param {string} name The name of the rule to define.
         * @param {Function} rule The rule definition.
         * @returns {void}
         */
        defineRule(name, rule) {
            this.rules[name] = rule;
        }
        /**
         * Adds a new rule test to execute.
         * @param {string} ruleName The name of the rule to run.
         * @param {Function} rule The rule to test.
         * @param {{
         *   valid: (ValidTestCase | string)[],
         *   invalid: InvalidTestCase[]
         * }} test The collection of tests to run.
         * @returns {void}
         */
        run(ruleName, rule, test) {
            const testerConfig = this.testerConfig, requiredScenarios = ["valid", "invalid"], scenarioErrors = [], linter = this.linter;
            if (!test || typeof test !== "object") {
                throw new TypeError(`Test Scenarios for rule ${ruleName} : Could not find test scenario object`);
            }
            requiredScenarios.forEach(scenarioType => {
                if (!test[scenarioType]) {
                    scenarioErrors.push(`Could not find any ${scenarioType} test scenarios`);
                }
            });
            if (scenarioErrors.length > 0) {
                throw new Error([
                    `Test Scenarios for rule ${ruleName} is invalid:`
                ].concat(scenarioErrors).join("\n"));
            }
            linter.defineRule(ruleName, Object.assign({}, rule, {
                // Create a wrapper rule that freezes the `context` properties.
                create(context) {
                    freezeDeeply(context.options);
                    freezeDeeply(context.settings);
                    freezeDeeply(context.parserOptions);
                    return (typeof rule === "function" ? rule : rule.create)(context);
                }
            }));
            linter.defineRules(this.rules);
            /**
             * Run the rule for the given item
             * @param {string|Object} item Item to run the rule against
             * @returns {Object} Eslint run result
             * @private
             */
            function runRuleForItem(item) {
                let config = merge({}, testerConfig), code, filename, output, beforeAST, afterAST;
                if (typeof item === "string") {
                    code = item;
                }
                else {
                    code = item.code;
                    /*
                     * Assumes everything on the item is a config except for the
                     * parameters used by this tester
                     */
                    const itemConfig = { ...item };
                    for (const parameter of RuleTesterParameters) {
                        delete itemConfig[parameter];
                    }
                    /*
                     * Create the config object from the tester config and this item
                     * specific configurations.
                     */
                    config = merge(config, itemConfig);
                }
                if (item.filename) {
                    filename = item.filename;
                }
                if (hasOwnProperty(item, "options")) {
                    assert(Array.isArray(item.options), "options must be an array");
                    config.rules[ruleName] = [1].concat(item.options);
                }
                else {
                    config.rules[ruleName] = 1;
                }
                const schema = getRuleOptionsSchema(rule);
                /*
                 * Setup AST getters.
                 * The goal is to check whether or not AST was modified when
                 * running the rule under test.
                 */
                linter.defineRule("rule-tester/validate-ast", () => ({
                    Program(node) {
                        beforeAST = cloneDeeplyExcludesParent(node);
                    },
                    "Program:exit"(node) {
                        afterAST = node;
                    }
                }));
                if (typeof config.parser === "string") {
                    assert(path.isAbsolute(config.parser), "Parsers provided as strings to RuleTester must be absolute paths");
                }
                else {
                    config.parser = espreePath;
                }
                linter.defineParser(config.parser, wrapParser(require(config.parser)));
                if (schema) {
                    ajv.validateSchema(schema);
                    if (ajv.errors) {
                        const errors = ajv.errors.map(error => {
                            const field = error.dataPath[0] === "." ? error.dataPath.slice(1) : error.dataPath;
                            return `\t${field}: ${error.message}`;
                        }).join("\n");
                        throw new Error([`Schema for rule ${ruleName} is invalid:`, errors]);
                    }
                    /*
                     * `ajv.validateSchema` checks for errors in the structure of the schema (by comparing the schema against a "meta-schema"),
                     * and it reports those errors individually. However, there are other types of schema errors that only occur when compiling
                     * the schema (e.g. using invalid defaults in a schema), and only one of these errors can be reported at a time. As a result,
                     * the schema is compiled here separately from checking for `validateSchema` errors.
                     */
                    try {
                        ajv.compile(schema);
                    }
                    catch (err) {
                        throw new Error(`Schema for rule ${ruleName} is invalid: ${err.message}`);
                    }
                }
                validate(config, "rule-tester", id => (id === ruleName ? rule : null));
                // Verify the code.
                const messages = linter.verify(code, config, filename);
                const fatalErrorMessage = messages.find(m => m.fatal);
                assert(!fatalErrorMessage, `A fatal parsing error occurred: ${fatalErrorMessage && fatalErrorMessage.message}`);
                // Verify if autofix makes a syntax error or not.
                if (messages.some(m => m.fix)) {
                    output = SourceCodeFixer.applyFixes(code, messages).output;
                    const errorMessageInFix = linter.verify(output, config, filename).find(m => m.fatal);
                    assert(!errorMessageInFix, [
                        "A fatal parsing error occurred in autofix.",
                        `Error: ${errorMessageInFix && errorMessageInFix.message}`,
                        "Autofix output:",
                        output
                    ].join("\n"));
                }
                else {
                    output = code;
                }
                return {
                    messages,
                    output,
                    beforeAST,
                    afterAST: cloneDeeplyExcludesParent(afterAST)
                };
            }
            /**
             * Check if the AST was changed
             * @param {ASTNode} beforeAST AST node before running
             * @param {ASTNode} afterAST AST node after running
             * @returns {void}
             * @private
             */
            function assertASTDidntChange(beforeAST, afterAST) {
                if (!equal(beforeAST, afterAST)) {
                    assert.fail("Rule should not modify AST.");
                }
            }
            /**
             * Check if the template is valid or not
             * all valid cases go through this
             * @param {string|Object} item Item to run the rule against
             * @returns {void}
             * @private
             */
            function testValidTemplate(item) {
                const result = runRuleForItem(item);
                const messages = result.messages;
                assert.strictEqual(messages.length, 0, util.format("Should have no errors but had %d: %s", messages.length, util.inspect(messages)));
                assertASTDidntChange(result.beforeAST, result.afterAST);
            }
            /**
             * Asserts that the message matches its expected value. If the expected
             * value is a regular expression, it is checked against the actual
             * value.
             * @param {string} actual Actual value
             * @param {string|RegExp} expected Expected value
             * @returns {void}
             * @private
             */
            function assertMessageMatches(actual, expected) {
                if (expected instanceof RegExp) {
                    // assert.js doesn't have a built-in RegExp match function
                    assert.ok(expected.test(actual), `Expected '${actual}' to match ${expected}`);
                }
                else {
                    assert.strictEqual(actual, expected);
                }
            }
            /**
             * Check if the template is invalid or not
             * all invalid cases go through this.
             * @param {string|Object} item Item to run the rule against
             * @returns {void}
             * @private
             */
            function testInvalidTemplate(item) {
                assert.ok(item.errors || item.errors === 0, `Did not specify errors for an invalid test of ${ruleName}`);
                if (Array.isArray(item.errors) && item.errors.length === 0) {
                    assert.fail("Invalid cases must have at least one error");
                }
                const ruleHasMetaMessages = hasOwnProperty(rule, "meta") && hasOwnProperty(rule.meta, "messages");
                const friendlyIDList = ruleHasMetaMessages ? `[${Object.keys(rule.meta.messages).map(key => `'${key}'`).join(", ")}]` : null;
                const result = runRuleForItem(item);
                const messages = result.messages;
                if (typeof item.errors === "number") {
                    if (item.errors === 0) {
                        assert.fail("Invalid cases must have 'error' value greater than 0");
                    }
                    assert.strictEqual(messages.length, item.errors, util.format("Should have %d error%s but had %d: %s", item.errors, item.errors === 1 ? "" : "s", messages.length, util.inspect(messages)));
                }
                else {
                    assert.strictEqual(messages.length, item.errors.length, util.format("Should have %d error%s but had %d: %s", item.errors.length, item.errors.length === 1 ? "" : "s", messages.length, util.inspect(messages)));
                    const hasMessageOfThisRule = messages.some(m => m.ruleId === ruleName);
                    for (let i = 0, l = item.errors.length; i < l; i++) {
                        const error = item.errors[i];
                        const message = messages[i];
                        assert(hasMessageOfThisRule, "Error rule name should be the same as the name of the rule being tested");
                        if (typeof error === "string" || error instanceof RegExp) {
                            // Just an error message.
                            assertMessageMatches(message.message, error);
                        }
                        else if (typeof error === "object" && error !== null) {
                            /*
                             * Error object.
                             * This may have a message, messageId, data, node type, line, and/or
                             * column.
                             */
                            Object.keys(error).forEach(propertyName => {
                                assert.ok(errorObjectParameters.has(propertyName), `Invalid error property name '${propertyName}'. Expected one of ${friendlyErrorObjectParameterList}.`);
                            });
                            if (hasOwnProperty(error, "message")) {
                                assert.ok(!hasOwnProperty(error, "messageId"), "Error should not specify both 'message' and a 'messageId'.");
                                assert.ok(!hasOwnProperty(error, "data"), "Error should not specify both 'data' and 'message'.");
                                assertMessageMatches(message.message, error.message);
                            }
                            else if (hasOwnProperty(error, "messageId")) {
                                assert.ok(ruleHasMetaMessages, "Error can not use 'messageId' if rule under test doesn't define 'meta.messages'.");
                                if (!hasOwnProperty(rule.meta.messages, error.messageId)) {
                                    assert(false, `Invalid messageId '${error.messageId}'. Expected one of ${friendlyIDList}.`);
                                }
                                assert.strictEqual(message.messageId, error.messageId, `messageId '${message.messageId}' does not match expected messageId '${error.messageId}'.`);
                                if (hasOwnProperty(error, "data")) {
                                    /*
                                     *  if data was provided, then directly compare the returned message to a synthetic
                                     *  interpolated message using the same message ID and data provided in the test.
                                     *  See https://github.com/eslint/eslint/issues/9890 for context.
                                     */
                                    const unformattedOriginalMessage = rule.meta.messages[error.messageId];
                                    const rehydratedMessage = interpolate(unformattedOriginalMessage, error.data);
                                    assert.strictEqual(message.message, rehydratedMessage, `Hydrated message "${rehydratedMessage}" does not match "${message.message}"`);
                                }
                            }
                            assert.ok(hasOwnProperty(error, "data") ? hasOwnProperty(error, "messageId") : true, "Error must specify 'messageId' if 'data' is used.");
                            if (error.type) {
                                assert.strictEqual(message.nodeType, error.type, `Error type should be ${error.type}, found ${message.nodeType}`);
                            }
                            if (hasOwnProperty(error, "line")) {
                                assert.strictEqual(message.line, error.line, `Error line should be ${error.line}`);
                            }
                            if (hasOwnProperty(error, "column")) {
                                assert.strictEqual(message.column, error.column, `Error column should be ${error.column}`);
                            }
                            if (hasOwnProperty(error, "endLine")) {
                                assert.strictEqual(message.endLine, error.endLine, `Error endLine should be ${error.endLine}`);
                            }
                            if (hasOwnProperty(error, "endColumn")) {
                                assert.strictEqual(message.endColumn, error.endColumn, `Error endColumn should be ${error.endColumn}`);
                            }
                            if (hasOwnProperty(error, "suggestions")) {
                                // Support asserting there are no suggestions
                                if (!error.suggestions || (Array.isArray(error.suggestions) && error.suggestions.length === 0)) {
                                    if (Array.isArray(message.suggestions) && message.suggestions.length > 0) {
                                        assert.fail(`Error should have no suggestions on error with message: "${message.message}"`);
                                    }
                                }
                                else {
                                    assert.strictEqual(Array.isArray(message.suggestions), true, `Error should have an array of suggestions. Instead received "${message.suggestions}" on error with message: "${message.message}"`);
                                    assert.strictEqual(message.suggestions.length, error.suggestions.length, `Error should have ${error.suggestions.length} suggestions. Instead found ${message.suggestions.length} suggestions`);
                                    error.suggestions.forEach((expectedSuggestion, index) => {
                                        assert.ok(typeof expectedSuggestion === "object" && expectedSuggestion !== null, "Test suggestion in 'suggestions' array must be an object.");
                                        Object.keys(expectedSuggestion).forEach(propertyName => {
                                            assert.ok(suggestionObjectParameters.has(propertyName), `Invalid suggestion property name '${propertyName}'. Expected one of ${friendlySuggestionObjectParameterList}.`);
                                        });
                                        const actualSuggestion = message.suggestions[index];
                                        const suggestionPrefix = `Error Suggestion at index ${index} :`;
                                        if (hasOwnProperty(expectedSuggestion, "desc")) {
                                            assert.ok(!hasOwnProperty(expectedSuggestion, "data"), `${suggestionPrefix} Test should not specify both 'desc' and 'data'.`);
                                            assert.strictEqual(actualSuggestion.desc, expectedSuggestion.desc, `${suggestionPrefix} desc should be "${expectedSuggestion.desc}" but got "${actualSuggestion.desc}" instead.`);
                                        }
                                        if (hasOwnProperty(expectedSuggestion, "messageId")) {
                                            assert.ok(ruleHasMetaMessages, `${suggestionPrefix} Test can not use 'messageId' if rule under test doesn't define 'meta.messages'.`);
                                            assert.ok(hasOwnProperty(rule.meta.messages, expectedSuggestion.messageId), `${suggestionPrefix} Test has invalid messageId '${expectedSuggestion.messageId}', the rule under test allows only one of ${friendlyIDList}.`);
                                            assert.strictEqual(actualSuggestion.messageId, expectedSuggestion.messageId, `${suggestionPrefix} messageId should be '${expectedSuggestion.messageId}' but got '${actualSuggestion.messageId}' instead.`);
                                            if (hasOwnProperty(expectedSuggestion, "data")) {
                                                const unformattedMetaMessage = rule.meta.messages[expectedSuggestion.messageId];
                                                const rehydratedDesc = interpolate(unformattedMetaMessage, expectedSuggestion.data);
                                                assert.strictEqual(actualSuggestion.desc, rehydratedDesc, `${suggestionPrefix} Hydrated test desc "${rehydratedDesc}" does not match received desc "${actualSuggestion.desc}".`);
                                            }
                                        }
                                        else {
                                            assert.ok(!hasOwnProperty(expectedSuggestion, "data"), `${suggestionPrefix} Test must specify 'messageId' if 'data' is used.`);
                                        }
                                        if (hasOwnProperty(expectedSuggestion, "output")) {
                                            const codeWithAppliedSuggestion = SourceCodeFixer.applyFixes(item.code, [actualSuggestion]).output;
                                            assert.strictEqual(codeWithAppliedSuggestion, expectedSuggestion.output, `Expected the applied suggestion fix to match the test suggestion output for suggestion at index: ${index} on error with message: "${message.message}"`);
                                        }
                                    });
                                }
                            }
                        }
                        else {
                            // Message was an unexpected type
                            assert.fail(`Error should be a string, object, or RegExp, but found (${util.inspect(message)})`);
                        }
                    }
                }
                if (hasOwnProperty(item, "output")) {
                    if (item.output === null) {
                        assert.strictEqual(result.output, item.code, "Expected no autofixes to be suggested");
                    }
                    else {
                        assert.strictEqual(result.output, item.output, "Output is incorrect.");
                    }
                }
                else {
                    assert.strictEqual(result.output, item.code, "The rule fixed the code. Please add 'output' property.");
                }
                // Rules that produce fixes must have `meta.fixable` property.
                if (result.output !== item.code) {
                    assert.ok(hasOwnProperty(rule, "meta"), "Fixable rules should export a `meta.fixable` property.");
                    // Linter throws if a rule that produced a fix has `meta` but doesn't have `meta.fixable`.
                }
                assertASTDidntChange(result.beforeAST, result.afterAST);
            }
            /*
             * This creates a mocha test suite and pipes all supplied info through
             * one of the templates above.
             */
            RuleTester.describe(ruleName, () => {
                RuleTester.describe("valid", () => {
                    test.valid.forEach(valid => {
                        RuleTester[valid.only ? "itOnly" : "it"](sanitize(typeof valid === "object" ? valid.code : valid), () => {
                            testValidTemplate(valid);
                        });
                    });
                });
                RuleTester.describe("invalid", () => {
                    test.invalid.forEach(invalid => {
                        RuleTester[invalid.only ? "itOnly" : "it"](sanitize(invalid.code), () => {
                            testInvalidTemplate(invalid);
                        });
                    });
                });
            });
        }
    }
    RuleTester[DESCRIBE] = RuleTester[IT] = RuleTester[IT_ONLY] = null;
    module.exports = RuleTester;
},
/* eslint/lib/shared/config-validator.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const util = require("util"), configSchema = require(729) /* ../../conf/config-schema */, BuiltInEnvironments = require(217) /* @eslint/eslintrc/conf/environments */, BuiltInRules = require(391) /* ../rules */, ConfigOps = require(220) /* @eslint/eslintrc/lib/shared/config-ops */, { emitDeprecationWarning } = require(730) /* ./deprecation-warnings */;
    const ajv = require(731) /* ./ajv */();
    const ruleValidators = new WeakMap();
    const noop = Function.prototype;
    //------------------------------------------------------------------------------
    // Private
    //------------------------------------------------------------------------------
    let validateSchema;
    const severityMap = {
        error: 2,
        warn: 1,
        off: 0
    };
    /**
     * Gets a complete options schema for a rule.
     * @param {{create: Function, schema: (Array|null)}} rule A new-style rule object
     * @returns {Object} JSON Schema for the rule's options.
     */
    function getRuleOptionsSchema(rule) {
        if (!rule) {
            return null;
        }
        const schema = rule.schema || rule.meta && rule.meta.schema;
        // Given a tuple of schemas, insert warning level at the beginning
        if (Array.isArray(schema)) {
            if (schema.length) {
                return {
                    type: "array",
                    items: schema,
                    minItems: 0,
                    maxItems: schema.length
                };
            }
            return {
                type: "array",
                minItems: 0,
                maxItems: 0
            };
        }
        // Given a full schema, leave it alone
        return schema || null;
    }
    /**
     * Validates a rule's severity and returns the severity value. Throws an error if the severity is invalid.
     * @param {options} options The given options for the rule.
     * @returns {number|string} The rule's severity value
     */
    function validateRuleSeverity(options) {
        const severity = Array.isArray(options) ? options[0] : options;
        const normSeverity = typeof severity === "string" ? severityMap[severity.toLowerCase()] : severity;
        if (normSeverity === 0 || normSeverity === 1 || normSeverity === 2) {
            return normSeverity;
        }
        throw new Error(`\tSeverity should be one of the following: 0 = off, 1 = warn, 2 = error (you passed '${util.inspect(severity).replace(/'/gu, "\"").replace(/\n/gu, "")}').\n`);
    }
    /**
     * Validates the non-severity options passed to a rule, based on its schema.
     * @param {{create: Function}} rule The rule to validate
     * @param {Array} localOptions The options for the rule, excluding severity
     * @returns {void}
     */
    function validateRuleSchema(rule, localOptions) {
        if (!ruleValidators.has(rule)) {
            const schema = getRuleOptionsSchema(rule);
            if (schema) {
                ruleValidators.set(rule, ajv.compile(schema));
            }
        }
        const validateRule = ruleValidators.get(rule);
        if (validateRule) {
            validateRule(localOptions);
            if (validateRule.errors) {
                throw new Error(validateRule.errors.map(error => `\tValue ${JSON.stringify(error.data)} ${error.message}.\n`).join(""));
            }
        }
    }
    /**
     * Validates a rule's options against its schema.
     * @param {{create: Function}|null} rule The rule that the config is being validated for
     * @param {string} ruleId The rule's unique name.
     * @param {Array|number} options The given options for the rule.
     * @param {string|null} source The name of the configuration source to report in any errors. If null or undefined,
     * no source is prepended to the message.
     * @returns {void}
     */
    function validateRuleOptions(rule, ruleId, options, source = null) {
        try {
            const severity = validateRuleSeverity(options);
            if (severity !== 0) {
                validateRuleSchema(rule, Array.isArray(options) ? options.slice(1) : []);
            }
        }
        catch (err) {
            const enhancedMessage = `Configuration for rule "${ruleId}" is invalid:\n${err.message}`;
            if (typeof source === "string") {
                throw new Error(`${source}:\n\t${enhancedMessage}`);
            }
            else {
                throw new Error(enhancedMessage);
            }
        }
    }
    /**
     * Validates an environment object
     * @param {Object} environment The environment config object to validate.
     * @param {string} source The name of the configuration source to report in any errors.
     * @param {function(envId:string): Object} [getAdditionalEnv] A map from strings to loaded environments.
     * @returns {void}
     */
    function validateEnvironment(environment, source, getAdditionalEnv = noop) {
        // not having an environment is ok
        if (!environment) {
            return;
        }
        Object.keys(environment).forEach(id => {
            const env = getAdditionalEnv(id) || BuiltInEnvironments.get(id) || null;
            if (!env) {
                const message = `${source}:\n\tEnvironment key "${id}" is unknown\n`;
                throw new Error(message);
            }
        });
    }
    /**
     * Validates a rules config object
     * @param {Object} rulesConfig The rules config object to validate.
     * @param {string} source The name of the configuration source to report in any errors.
     * @param {function(ruleId:string): Object} getAdditionalRule A map from strings to loaded rules
     * @returns {void}
     */
    function validateRules(rulesConfig, source, getAdditionalRule = noop) {
        if (!rulesConfig) {
            return;
        }
        Object.keys(rulesConfig).forEach(id => {
            const rule = getAdditionalRule(id) || BuiltInRules.get(id) || null;
            validateRuleOptions(rule, id, rulesConfig[id], source);
        });
    }
    /**
     * Validates a `globals` section of a config file
     * @param {Object} globalsConfig The `globals` section
     * @param {string|null} source The name of the configuration source to report in the event of an error.
     * @returns {void}
     */
    function validateGlobals(globalsConfig, source = null) {
        if (!globalsConfig) {
            return;
        }
        Object.entries(globalsConfig)
            .forEach(([configuredGlobal, configuredValue]) => {
            try {
                ConfigOps.normalizeConfigGlobal(configuredValue);
            }
            catch (err) {
                throw new Error(`ESLint configuration of global '${configuredGlobal}' in ${source} is invalid:\n${err.message}`);
            }
        });
    }
    /**
     * Validate `processor` configuration.
     * @param {string|undefined} processorName The processor name.
     * @param {string} source The name of config file.
     * @param {function(id:string): Processor} getProcessor The getter of defined processors.
     * @returns {void}
     */
    function validateProcessor(processorName, source, getProcessor) {
        if (processorName && !getProcessor(processorName)) {
            throw new Error(`ESLint configuration of processor in '${source}' is invalid: '${processorName}' was not found.`);
        }
    }
    /**
     * Formats an array of schema validation errors.
     * @param {Array} errors An array of error messages to format.
     * @returns {string} Formatted error message
     */
    function formatErrors(errors) {
        return errors.map(error => {
            if (error.keyword === "additionalProperties") {
                const formattedPropertyPath = error.dataPath.length ? `${error.dataPath.slice(1)}.${error.params.additionalProperty}` : error.params.additionalProperty;
                return `Unexpected top-level property "${formattedPropertyPath}"`;
            }
            if (error.keyword === "type") {
                const formattedField = error.dataPath.slice(1);
                const formattedExpectedType = Array.isArray(error.schema) ? error.schema.join("/") : error.schema;
                const formattedValue = JSON.stringify(error.data);
                return `Property "${formattedField}" is the wrong type (expected ${formattedExpectedType} but got \`${formattedValue}\`)`;
            }
            const field = error.dataPath[0] === "." ? error.dataPath.slice(1) : error.dataPath;
            return `"${field}" ${error.message}. Value: ${JSON.stringify(error.data)}`;
        }).map(message => `\t- ${message}.\n`).join("");
    }
    /**
     * Validates the top level properties of the config object.
     * @param {Object} config The config object to validate.
     * @param {string} source The name of the configuration source to report in any errors.
     * @returns {void}
     */
    function validateConfigSchema(config, source = null) {
        validateSchema = validateSchema || ajv.compile(configSchema);
        if (!validateSchema(config)) {
            throw new Error(`ESLint configuration in ${source} is invalid:\n${formatErrors(validateSchema.errors)}`);
        }
        if (Object.hasOwnProperty.call(config, "ecmaFeatures")) {
            emitDeprecationWarning(source, "ESLINT_LEGACY_ECMAFEATURES");
        }
    }
    /**
     * Validates an entire config object.
     * @param {Object} config The config object to validate.
     * @param {string} source The name of the configuration source to report in any errors.
     * @param {function(ruleId:string): Object} [getAdditionalRule] A map from strings to loaded rules.
     * @param {function(envId:string): Object} [getAdditionalEnv] A map from strings to loaded envs.
     * @returns {void}
     */
    function validate(config, source, getAdditionalRule, getAdditionalEnv) {
        validateConfigSchema(config, source);
        validateRules(config.rules, source, getAdditionalRule);
        validateEnvironment(config.env, source, getAdditionalEnv);
        validateGlobals(config.globals, source);
        for (const override of config.overrides || []) {
            validateRules(override.rules, source, getAdditionalRule);
            validateEnvironment(override.env, source, getAdditionalEnv);
            validateGlobals(config.globals, source);
        }
    }
    const validated = new WeakSet();
    /**
     * Validate config array object.
     * @param {ConfigArray} configArray The config array to validate.
     * @returns {void}
     */
    function validateConfigArray(configArray) {
        const getPluginEnv = Map.prototype.get.bind(configArray.pluginEnvironments);
        const getPluginProcessor = Map.prototype.get.bind(configArray.pluginProcessors);
        const getPluginRule = Map.prototype.get.bind(configArray.pluginRules);
        // Validate.
        for (const element of configArray) {
            if (validated.has(element)) {
                continue;
            }
            validated.add(element);
            validateEnvironment(element.env, element.name, getPluginEnv);
            validateGlobals(element.globals, element.name);
            validateProcessor(element.processor, element.name, getPluginProcessor);
            validateRules(element.rules, element.name, getPluginRule);
        }
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = {
        getRuleOptionsSchema,
        validate,
        validateConfigArray,
        validateConfigSchema,
        validateRuleOptions
    };
},
/* eslint/conf/config-schema.js */ function _(require, module, exports, __esModule, __esExport) {
    const baseConfigProperties = {
        $schema: { type: "string" },
        env: { type: "object" },
        extends: { $ref: "#/definitions/stringOrStrings" },
        globals: { type: "object" },
        overrides: {
            type: "array",
            items: { $ref: "#/definitions/overrideConfig" },
            additionalItems: false
        },
        parser: { type: ["string", "null"] },
        parserOptions: { type: "object" },
        plugins: { type: "array" },
        processor: { type: "string" },
        rules: { type: "object" },
        settings: { type: "object" },
        noInlineConfig: { type: "boolean" },
        reportUnusedDisableDirectives: { type: "boolean" },
        ecmaFeatures: { type: "object" } // deprecated; logs a warning when used
    };
    const configSchema = {
        definitions: {
            stringOrStrings: {
                oneOf: [
                    { type: "string" },
                    {
                        type: "array",
                        items: { type: "string" },
                        additionalItems: false
                    }
                ]
            },
            stringOrStringsRequired: {
                oneOf: [
                    { type: "string" },
                    {
                        type: "array",
                        items: { type: "string" },
                        additionalItems: false,
                        minItems: 1
                    }
                ]
            },
            // Config at top-level.
            objectConfig: {
                type: "object",
                properties: {
                    root: { type: "boolean" },
                    ignorePatterns: { $ref: "#/definitions/stringOrStrings" },
                    ...baseConfigProperties
                },
                additionalProperties: false
            },
            // Config in `overrides`.
            overrideConfig: {
                type: "object",
                properties: {
                    excludedFiles: { $ref: "#/definitions/stringOrStrings" },
                    files: { $ref: "#/definitions/stringOrStringsRequired" },
                    ...baseConfigProperties
                },
                required: ["files"],
                additionalProperties: false
            }
        },
        $ref: "#/definitions/objectConfig"
    };
    module.exports = configSchema;
},
/* eslint/lib/shared/deprecation-warnings.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const path = require("path");
    //------------------------------------------------------------------------------
    // Private
    //------------------------------------------------------------------------------
    // Definitions for deprecation warnings.
    const deprecationWarningMessages = {
        ESLINT_LEGACY_ECMAFEATURES: "The 'ecmaFeatures' config file property is deprecated and has no effect.",
        ESLINT_PERSONAL_CONFIG_LOAD: "'~/.eslintrc.*' config files have been deprecated. " +
            "Please use a config file per project or the '--config' option.",
        ESLINT_PERSONAL_CONFIG_SUPPRESS: "'~/.eslintrc.*' config files have been deprecated. " +
            "Please remove it or add 'root:true' to the config files in your " +
            "projects in order to avoid loading '~/.eslintrc.*' accidentally."
    };
    const sourceFileErrorCache = new Set();
    /**
     * Emits a deprecation warning containing a given filepath. A new deprecation warning is emitted
     * for each unique file path, but repeated invocations with the same file path have no effect.
     * No warnings are emitted if the `--no-deprecation` or `--no-warnings` Node runtime flags are active.
     * @param {string} source The name of the configuration source to report the warning for.
     * @param {string} errorCode The warning message to show.
     * @returns {void}
     */
    function emitDeprecationWarning(source, errorCode) {
        const cacheKey = JSON.stringify({ source, errorCode });
        if (sourceFileErrorCache.has(cacheKey)) {
            return;
        }
        sourceFileErrorCache.add(cacheKey);
        const rel = path.relative(process.cwd(), source);
        const message = deprecationWarningMessages[errorCode];
        process.emitWarning(`${message} (found in "${rel}")`, "DeprecationWarning", errorCode);
    }
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = {
        emitDeprecationWarning
    };
},
/* eslint/lib/shared/ajv.js */ function _(require, module, exports, __esModule, __esExport) {
    //------------------------------------------------------------------------------
    // Requirements
    //------------------------------------------------------------------------------
    const Ajv = require(223) /* ajv */, metaSchema = require(269) /* ajv/lib/refs/json-schema-draft-04.json */;
    //------------------------------------------------------------------------------
    // Public Interface
    //------------------------------------------------------------------------------
    module.exports = (additionalOptions = {}) => {
        const ajv = new Ajv({
            meta: false,
            useDefaults: true,
            validateSchema: false,
            missingRefs: "ignore",
            verbose: true,
            schemaId: "auto",
            ...additionalOptions
        });
        ajv.addMetaSchema(metaSchema);
        // eslint-disable-next-line no-underscore-dangle
        ajv._opts.defaultMeta = metaSchema.id;
        return ajv;
    };
},
/* @typescript-eslint/eslint-plugin/dist/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    const rules_1 = __importDefault(require(733) /* ./rules */);
    const all_1 = __importDefault(require(1112) /* ./configs/all */);
    const base_1 = __importDefault(require(1113) /* ./configs/base */);
    const recommended_1 = __importDefault(require(1114) /* ./configs/recommended */);
    const recommended_requiring_type_checking_1 = __importDefault(require(1115) /* ./configs/recommended-requiring-type-checking */);
    const eslint_recommended_1 = __importDefault(require(1116) /* ./configs/eslint-recommended */);
    module.exports = {
        rules: rules_1.default,
        configs: {
            all: all_1.default,
            base: base_1.default,
            recommended: recommended_1.default,
            'eslint-recommended': eslint_recommended_1.default,
            'recommended-requiring-type-checking': recommended_requiring_type_checking_1.default,
        },
    };
},
/* @typescript-eslint/eslint-plugin/dist/rules/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const adjacent_overload_signatures_1 = __importDefault(require(734) /* ./adjacent-overload-signatures */);
    const array_type_1 = __importDefault(require(941) /* ./array-type */);
    const await_thenable_1 = __importDefault(require(942) /* ./await-thenable */);
    const ban_ts_comment_1 = __importDefault(require(943) /* ./ban-ts-comment */);
    const ban_tslint_comment_1 = __importDefault(require(944) /* ./ban-tslint-comment */);
    const ban_types_1 = __importDefault(require(945) /* ./ban-types */);
    const brace_style_1 = __importDefault(require(946) /* ./brace-style */);
    const class_literal_property_style_1 = __importDefault(require(947) /* ./class-literal-property-style */);
    const comma_dangle_1 = __importDefault(require(948) /* ./comma-dangle */);
    const comma_spacing_1 = __importDefault(require(949) /* ./comma-spacing */);
    const consistent_indexed_object_style_1 = __importDefault(require(950) /* ./consistent-indexed-object-style */);
    const consistent_type_assertions_1 = __importDefault(require(951) /* ./consistent-type-assertions */);
    const consistent_type_definitions_1 = __importDefault(require(952) /* ./consistent-type-definitions */);
    const consistent_type_imports_1 = __importDefault(require(953) /* ./consistent-type-imports */);
    const default_param_last_1 = __importDefault(require(954) /* ./default-param-last */);
    const dot_notation_1 = __importDefault(require(955) /* ./dot-notation */);
    const explicit_function_return_type_1 = __importDefault(require(956) /* ./explicit-function-return-type */);
    const explicit_member_accessibility_1 = __importDefault(require(958) /* ./explicit-member-accessibility */);
    const explicit_module_boundary_types_1 = __importDefault(require(959) /* ./explicit-module-boundary-types */);
    const func_call_spacing_1 = __importDefault(require(960) /* ./func-call-spacing */);
    const indent_1 = __importDefault(require(961) /* ./indent */);
    const init_declarations_1 = __importDefault(require(962) /* ./init-declarations */);
    const keyword_spacing_1 = __importDefault(require(963) /* ./keyword-spacing */);
    const lines_between_class_members_1 = __importDefault(require(964) /* ./lines-between-class-members */);
    const member_delimiter_style_1 = __importDefault(require(965) /* ./member-delimiter-style */);
    const member_ordering_1 = __importDefault(require(966) /* ./member-ordering */);
    const method_signature_style_1 = __importDefault(require(967) /* ./method-signature-style */);
    const naming_convention_1 = __importDefault(require(968) /* ./naming-convention */);
    const no_array_constructor_1 = __importDefault(require(976) /* ./no-array-constructor */);
    const no_base_to_string_1 = __importDefault(require(977) /* ./no-base-to-string */);
    const no_confusing_non_null_assertion_1 = __importDefault(require(978) /* ./no-confusing-non-null-assertion */);
    const no_confusing_void_expression_1 = __importDefault(require(979) /* ./no-confusing-void-expression */);
    const no_dupe_class_members_1 = __importDefault(require(980) /* ./no-dupe-class-members */);
    const no_duplicate_imports_1 = __importDefault(require(981) /* ./no-duplicate-imports */);
    const no_dynamic_delete_1 = __importDefault(require(982) /* ./no-dynamic-delete */);
    const no_empty_function_1 = __importDefault(require(983) /* ./no-empty-function */);
    const no_empty_interface_1 = __importDefault(require(984) /* ./no-empty-interface */);
    const no_explicit_any_1 = __importDefault(require(985) /* ./no-explicit-any */);
    const no_extra_non_null_assertion_1 = __importDefault(require(986) /* ./no-extra-non-null-assertion */);
    const no_extra_parens_1 = __importDefault(require(987) /* ./no-extra-parens */);
    const no_extra_semi_1 = __importDefault(require(988) /* ./no-extra-semi */);
    const no_extraneous_class_1 = __importDefault(require(989) /* ./no-extraneous-class */);
    const no_floating_promises_1 = __importDefault(require(990) /* ./no-floating-promises */);
    const no_for_in_array_1 = __importDefault(require(991) /* ./no-for-in-array */);
    const no_implicit_any_catch_1 = __importDefault(require(992) /* ./no-implicit-any-catch */);
    const no_implied_eval_1 = __importDefault(require(993) /* ./no-implied-eval */);
    const no_inferrable_types_1 = __importDefault(require(994) /* ./no-inferrable-types */);
    const no_invalid_this_1 = __importDefault(require(995) /* ./no-invalid-this */);
    const no_invalid_void_type_1 = __importDefault(require(996) /* ./no-invalid-void-type */);
    const no_loop_func_1 = __importDefault(require(997) /* ./no-loop-func */);
    const no_loss_of_precision_1 = __importDefault(require(998) /* ./no-loss-of-precision */);
    const no_magic_numbers_1 = __importDefault(require(999) /* ./no-magic-numbers */);
    const no_misused_new_1 = __importDefault(require(1000) /* ./no-misused-new */);
    const no_misused_promises_1 = __importDefault(require(1001) /* ./no-misused-promises */);
    const no_namespace_1 = __importDefault(require(1002) /* ./no-namespace */);
    const no_non_null_asserted_optional_chain_1 = __importDefault(require(1003) /* ./no-non-null-asserted-optional-chain */);
    const no_non_null_assertion_1 = __importDefault(require(1051) /* ./no-non-null-assertion */);
    const no_parameter_properties_1 = __importDefault(require(1052) /* ./no-parameter-properties */);
    const no_redeclare_1 = __importDefault(require(1053) /* ./no-redeclare */);
    const no_require_imports_1 = __importDefault(require(1054) /* ./no-require-imports */);
    const no_shadow_1 = __importDefault(require(1055) /* ./no-shadow */);
    const no_this_alias_1 = __importDefault(require(1056) /* ./no-this-alias */);
    const no_throw_literal_1 = __importDefault(require(1057) /* ./no-throw-literal */);
    const no_type_alias_1 = __importDefault(require(1058) /* ./no-type-alias */);
    const no_unnecessary_boolean_literal_compare_1 = __importDefault(require(1059) /* ./no-unnecessary-boolean-literal-compare */);
    const no_unnecessary_condition_1 = __importDefault(require(1060) /* ./no-unnecessary-condition */);
    const no_unnecessary_qualifier_1 = __importDefault(require(1061) /* ./no-unnecessary-qualifier */);
    const no_unnecessary_type_arguments_1 = __importDefault(require(1062) /* ./no-unnecessary-type-arguments */);
    const no_unnecessary_type_assertion_1 = __importDefault(require(1063) /* ./no-unnecessary-type-assertion */);
    const no_unnecessary_type_constraint_1 = __importDefault(require(1064) /* ./no-unnecessary-type-constraint */);
    const no_unsafe_argument_1 = __importDefault(require(1065) /* ./no-unsafe-argument */);
    const no_unsafe_assignment_1 = __importDefault(require(1066) /* ./no-unsafe-assignment */);
    const no_unsafe_call_1 = __importDefault(require(1067) /* ./no-unsafe-call */);
    const no_unsafe_member_access_1 = __importDefault(require(1068) /* ./no-unsafe-member-access */);
    const no_unsafe_return_1 = __importDefault(require(1069) /* ./no-unsafe-return */);
    const no_unused_expressions_1 = __importDefault(require(1070) /* ./no-unused-expressions */);
    const no_unused_vars_1 = __importDefault(require(1071) /* ./no-unused-vars */);
    const no_unused_vars_experimental_1 = __importDefault(require(1072) /* ./no-unused-vars-experimental */);
    const no_use_before_define_1 = __importDefault(require(1073) /* ./no-use-before-define */);
    const no_useless_constructor_1 = __importDefault(require(1074) /* ./no-useless-constructor */);
    const no_var_requires_1 = __importDefault(require(1075) /* ./no-var-requires */);
    const non_nullable_type_assertion_style_1 = __importDefault(require(1076) /* ./non-nullable-type-assertion-style */);
    const object_curly_spacing_1 = __importDefault(require(1077) /* ./object-curly-spacing */);
    const prefer_as_const_1 = __importDefault(require(1078) /* ./prefer-as-const */);
    const prefer_enum_initializers_1 = __importDefault(require(1079) /* ./prefer-enum-initializers */);
    const prefer_for_of_1 = __importDefault(require(1080) /* ./prefer-for-of */);
    const prefer_function_type_1 = __importDefault(require(1081) /* ./prefer-function-type */);
    const prefer_includes_1 = __importDefault(require(1082) /* ./prefer-includes */);
    const prefer_literal_enum_member_1 = __importDefault(require(1083) /* ./prefer-literal-enum-member */);
    const prefer_namespace_keyword_1 = __importDefault(require(1084) /* ./prefer-namespace-keyword */);
    const prefer_nullish_coalescing_1 = __importDefault(require(1085) /* ./prefer-nullish-coalescing */);
    const prefer_optional_chain_1 = __importDefault(require(1086) /* ./prefer-optional-chain */);
    const prefer_readonly_1 = __importDefault(require(1087) /* ./prefer-readonly */);
    const prefer_readonly_parameter_types_1 = __importDefault(require(1088) /* ./prefer-readonly-parameter-types */);
    const prefer_reduce_type_parameter_1 = __importDefault(require(1089) /* ./prefer-reduce-type-parameter */);
    const prefer_regexp_exec_1 = __importDefault(require(1090) /* ./prefer-regexp-exec */);
    const prefer_return_this_type_1 = __importDefault(require(1091) /* ./prefer-return-this-type */);
    const prefer_string_starts_ends_with_1 = __importDefault(require(1092) /* ./prefer-string-starts-ends-with */);
    const prefer_ts_expect_error_1 = __importDefault(require(1093) /* ./prefer-ts-expect-error */);
    const promise_function_async_1 = __importDefault(require(1094) /* ./promise-function-async */);
    const quotes_1 = __importDefault(require(1095) /* ./quotes */);
    const require_array_sort_compare_1 = __importDefault(require(1096) /* ./require-array-sort-compare */);
    const require_await_1 = __importDefault(require(1097) /* ./require-await */);
    const restrict_plus_operands_1 = __importDefault(require(1098) /* ./restrict-plus-operands */);
    const restrict_template_expressions_1 = __importDefault(require(1099) /* ./restrict-template-expressions */);
    const return_await_1 = __importDefault(require(1100) /* ./return-await */);
    const semi_1 = __importDefault(require(1101) /* ./semi */);
    const sort_type_union_intersection_members_1 = __importDefault(require(1102) /* ./sort-type-union-intersection-members */);
    const space_before_function_paren_1 = __importDefault(require(1103) /* ./space-before-function-paren */);
    const space_infix_ops_1 = __importDefault(require(1104) /* ./space-infix-ops */);
    const strict_boolean_expressions_1 = __importDefault(require(1105) /* ./strict-boolean-expressions */);
    const switch_exhaustiveness_check_1 = __importDefault(require(1106) /* ./switch-exhaustiveness-check */);
    const triple_slash_reference_1 = __importDefault(require(1107) /* ./triple-slash-reference */);
    const type_annotation_spacing_1 = __importDefault(require(1108) /* ./type-annotation-spacing */);
    const typedef_1 = __importDefault(require(1109) /* ./typedef */);
    const unbound_method_1 = __importDefault(require(1110) /* ./unbound-method */);
    const unified_signatures_1 = __importDefault(require(1111) /* ./unified-signatures */);
    exports.default = {
        'adjacent-overload-signatures': adjacent_overload_signatures_1.default,
        'array-type': array_type_1.default,
        'await-thenable': await_thenable_1.default,
        'ban-ts-comment': ban_ts_comment_1.default,
        'ban-tslint-comment': ban_tslint_comment_1.default,
        'ban-types': ban_types_1.default,
        'brace-style': brace_style_1.default,
        'class-literal-property-style': class_literal_property_style_1.default,
        'comma-dangle': comma_dangle_1.default,
        'comma-spacing': comma_spacing_1.default,
        'consistent-indexed-object-style': consistent_indexed_object_style_1.default,
        'consistent-type-assertions': consistent_type_assertions_1.default,
        'consistent-type-definitions': consistent_type_definitions_1.default,
        'consistent-type-imports': consistent_type_imports_1.default,
        'default-param-last': default_param_last_1.default,
        'dot-notation': dot_notation_1.default,
        'explicit-function-return-type': explicit_function_return_type_1.default,
        'explicit-member-accessibility': explicit_member_accessibility_1.default,
        'explicit-module-boundary-types': explicit_module_boundary_types_1.default,
        'func-call-spacing': func_call_spacing_1.default,
        indent: indent_1.default,
        'init-declarations': init_declarations_1.default,
        'keyword-spacing': keyword_spacing_1.default,
        'lines-between-class-members': lines_between_class_members_1.default,
        'member-delimiter-style': member_delimiter_style_1.default,
        'member-ordering': member_ordering_1.default,
        'method-signature-style': method_signature_style_1.default,
        'naming-convention': naming_convention_1.default,
        'no-array-constructor': no_array_constructor_1.default,
        'no-base-to-string': no_base_to_string_1.default,
        'no-confusing-non-null-assertion': no_confusing_non_null_assertion_1.default,
        'no-confusing-void-expression': no_confusing_void_expression_1.default,
        'no-dupe-class-members': no_dupe_class_members_1.default,
        'no-duplicate-imports': no_duplicate_imports_1.default,
        'no-dynamic-delete': no_dynamic_delete_1.default,
        'no-empty-function': no_empty_function_1.default,
        'no-empty-interface': no_empty_interface_1.default,
        'no-explicit-any': no_explicit_any_1.default,
        'no-extra-non-null-assertion': no_extra_non_null_assertion_1.default,
        'no-extra-parens': no_extra_parens_1.default,
        'no-extra-semi': no_extra_semi_1.default,
        'no-extraneous-class': no_extraneous_class_1.default,
        'no-floating-promises': no_floating_promises_1.default,
        'no-for-in-array': no_for_in_array_1.default,
        'no-implicit-any-catch': no_implicit_any_catch_1.default,
        'no-implied-eval': no_implied_eval_1.default,
        'no-inferrable-types': no_inferrable_types_1.default,
        'no-invalid-this': no_invalid_this_1.default,
        'no-invalid-void-type': no_invalid_void_type_1.default,
        'no-loop-func': no_loop_func_1.default,
        'no-loss-of-precision': no_loss_of_precision_1.default,
        'no-magic-numbers': no_magic_numbers_1.default,
        'no-misused-new': no_misused_new_1.default,
        'no-misused-promises': no_misused_promises_1.default,
        'no-namespace': no_namespace_1.default,
        'no-non-null-asserted-optional-chain': no_non_null_asserted_optional_chain_1.default,
        'no-non-null-assertion': no_non_null_assertion_1.default,
        'no-parameter-properties': no_parameter_properties_1.default,
        'no-redeclare': no_redeclare_1.default,
        'no-require-imports': no_require_imports_1.default,
        'no-shadow': no_shadow_1.default,
        'no-this-alias': no_this_alias_1.default,
        'no-throw-literal': no_throw_literal_1.default,
        'no-type-alias': no_type_alias_1.default,
        'no-unnecessary-boolean-literal-compare': no_unnecessary_boolean_literal_compare_1.default,
        'no-unnecessary-condition': no_unnecessary_condition_1.default,
        'no-unnecessary-qualifier': no_unnecessary_qualifier_1.default,
        'no-unnecessary-type-arguments': no_unnecessary_type_arguments_1.default,
        'no-unnecessary-type-assertion': no_unnecessary_type_assertion_1.default,
        'no-unnecessary-type-constraint': no_unnecessary_type_constraint_1.default,
        'no-unsafe-argument': no_unsafe_argument_1.default,
        'no-unsafe-assignment': no_unsafe_assignment_1.default,
        'no-unsafe-call': no_unsafe_call_1.default,
        'no-unsafe-member-access': no_unsafe_member_access_1.default,
        'no-unsafe-return': no_unsafe_return_1.default,
        'no-unused-expressions': no_unused_expressions_1.default,
        'no-unused-vars': no_unused_vars_1.default,
        'no-unused-vars-experimental': no_unused_vars_experimental_1.default,
        'no-use-before-define': no_use_before_define_1.default,
        'no-useless-constructor': no_useless_constructor_1.default,
        'no-var-requires': no_var_requires_1.default,
        'non-nullable-type-assertion-style': non_nullable_type_assertion_style_1.default,
        'object-curly-spacing': object_curly_spacing_1.default,
        'prefer-as-const': prefer_as_const_1.default,
        'prefer-enum-initializers': prefer_enum_initializers_1.default,
        'prefer-for-of': prefer_for_of_1.default,
        'prefer-function-type': prefer_function_type_1.default,
        'prefer-includes': prefer_includes_1.default,
        'prefer-literal-enum-member': prefer_literal_enum_member_1.default,
        'prefer-namespace-keyword': prefer_namespace_keyword_1.default,
        'prefer-nullish-coalescing': prefer_nullish_coalescing_1.default,
        'prefer-optional-chain': prefer_optional_chain_1.default,
        'prefer-readonly': prefer_readonly_1.default,
        'prefer-readonly-parameter-types': prefer_readonly_parameter_types_1.default,
        'prefer-reduce-type-parameter': prefer_reduce_type_parameter_1.default,
        'prefer-regexp-exec': prefer_regexp_exec_1.default,
        'prefer-return-this-type': prefer_return_this_type_1.default,
        'prefer-string-starts-ends-with': prefer_string_starts_ends_with_1.default,
        'prefer-ts-expect-error': prefer_ts_expect_error_1.default,
        'promise-function-async': promise_function_async_1.default,
        quotes: quotes_1.default,
        'require-array-sort-compare': require_array_sort_compare_1.default,
        'require-await': require_await_1.default,
        'restrict-plus-operands': restrict_plus_operands_1.default,
        'restrict-template-expressions': restrict_template_expressions_1.default,
        'return-await': return_await_1.default,
        semi: semi_1.default,
        'sort-type-union-intersection-members': sort_type_union_intersection_members_1.default,
        'space-before-function-paren': space_before_function_paren_1.default,
        'space-infix-ops': space_infix_ops_1.default,
        'strict-boolean-expressions': strict_boolean_expressions_1.default,
        'switch-exhaustiveness-check': switch_exhaustiveness_check_1.default,
        'triple-slash-reference': triple_slash_reference_1.default,
        'type-annotation-spacing': type_annotation_spacing_1.default,
        typedef: typedef_1.default,
        'unbound-method': unbound_method_1.default,
        'unified-signatures': unified_signatures_1.default,
    };
},
/* @typescript-eslint/eslint-plugin/dist/rules/adjacent-overload-signatures.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'adjacent-overload-signatures',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Require that member overloads be consecutive',
                category: 'Best Practices',
                recommended: 'error',
            },
            schema: [],
            messages: {
                adjacentSignature: "All '{{name}}' signatures should be adjacent.",
            },
        },
        defaultOptions: [],
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Gets the name and attribute of the member being processed.
             * @param member the member being processed.
             * @returns the name and attribute of the member or null if it's a member not relevant to the rule.
             */
            function getMemberMethod(member) {
                var _a, _b;
                if (!member) {
                    return null;
                }
                const isStatic = 'static' in member && !!member.static;
                switch (member.type) {
                    case experimental_utils_1.AST_NODE_TYPES.ExportDefaultDeclaration:
                    case experimental_utils_1.AST_NODE_TYPES.ExportNamedDeclaration: {
                        // export statements (e.g. export { a };)
                        // have no declarations, so ignore them
                        if (!member.declaration) {
                            return null;
                        }
                        return getMemberMethod(member.declaration);
                    }
                    case experimental_utils_1.AST_NODE_TYPES.TSDeclareFunction:
                    case experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration: {
                        const name = (_b = (_a = member.id) === null || _a === void 0 ? void 0 : _a.name) !== null && _b !== void 0 ? _b : null;
                        if (name === null) {
                            return null;
                        }
                        return {
                            name,
                            static: isStatic,
                            callSignature: false,
                        };
                    }
                    case experimental_utils_1.AST_NODE_TYPES.TSMethodSignature:
                        return {
                            name: util.getNameFromMember(member, sourceCode),
                            static: isStatic,
                            callSignature: false,
                        };
                    case experimental_utils_1.AST_NODE_TYPES.TSCallSignatureDeclaration:
                        return {
                            name: 'call',
                            static: isStatic,
                            callSignature: true,
                        };
                    case experimental_utils_1.AST_NODE_TYPES.TSConstructSignatureDeclaration:
                        return {
                            name: 'new',
                            static: isStatic,
                            callSignature: false,
                        };
                    case experimental_utils_1.AST_NODE_TYPES.MethodDefinition:
                        return {
                            name: util.getNameFromMember(member, sourceCode),
                            static: isStatic,
                            callSignature: false,
                        };
                }
                return null;
            }
            function isSameMethod(method1, method2) {
                return (!!method2 &&
                    method1.name === method2.name &&
                    method1.static === method2.static &&
                    method1.callSignature === method2.callSignature);
            }
            function getMembers(node) {
                switch (node.type) {
                    case experimental_utils_1.AST_NODE_TYPES.ClassBody:
                    case experimental_utils_1.AST_NODE_TYPES.Program:
                    case experimental_utils_1.AST_NODE_TYPES.TSModuleBlock:
                    case experimental_utils_1.AST_NODE_TYPES.TSInterfaceBody:
                        return node.body;
                    case experimental_utils_1.AST_NODE_TYPES.TSTypeLiteral:
                        return node.members;
                }
            }
            /**
             * Check the body for overload methods.
             * @param {ASTNode} node the body to be inspected.
             */
            function checkBodyForOverloadMethods(node) {
                const members = getMembers(node);
                if (members) {
                    let lastMethod = null;
                    const seenMethods = [];
                    members.forEach(member => {
                        const method = getMemberMethod(member);
                        if (method === null) {
                            lastMethod = null;
                            return;
                        }
                        const index = seenMethods.findIndex(seenMethod => isSameMethod(method, seenMethod));
                        if (index > -1 && !isSameMethod(method, lastMethod)) {
                            context.report({
                                node: member,
                                messageId: 'adjacentSignature',
                                data: {
                                    name: (method.static ? 'static ' : '') + method.name,
                                },
                            });
                        }
                        else if (index === -1) {
                            seenMethods.push(method);
                        }
                        lastMethod = method;
                    });
                }
            }
            return {
                ClassBody: checkBodyForOverloadMethods,
                Program: checkBodyForOverloadMethods,
                TSModuleBlock: checkBodyForOverloadMethods,
                TSTypeLiteral: checkBodyForOverloadMethods,
                TSInterfaceBody: checkBodyForOverloadMethods,
            };
        },
    });
},
/* @typescript-eslint/experimental-utils/dist/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    const ASTUtils = __importStar(require(736) /* ./ast-utils */);
    exports.ASTUtils = ASTUtils;
    const ESLintUtils = __importStar(require(751) /* ./eslint-utils */);
    exports.ESLintUtils = ESLintUtils;
    const JSONSchema = __importStar(require(895) /* ./json-schema */);
    exports.JSONSchema = JSONSchema;
    const TSESLint = __importStar(require(759) /* ./ts-eslint */);
    exports.TSESLint = TSESLint;
    const TSESLintScope = __importStar(require(896) /* ./ts-eslint-scope */);
    exports.TSESLintScope = TSESLintScope;
    __exportStar(require(739) /* ./ts-estree */, exports);
},
/* @typescript-eslint/experimental-utils/dist/ast-utils/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    __exportStar(require(737) /* ./misc */, exports);
    __exportStar(require(738) /* ./predicates */, exports);
    __exportStar(require(745) /* ./eslint-utils */, exports);
},
/* @typescript-eslint/experimental-utils/dist/ast-utils/misc.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const LINEBREAK_MATCHER = /\r\n|[\r\n\u2028\u2029]/;
    exports.LINEBREAK_MATCHER = LINEBREAK_MATCHER;
    /**
     * Determines whether two adjacent tokens are on the same line
     */
    function isTokenOnSameLine(left, right) {
        return left.loc.end.line === right.loc.start.line;
    }
    exports.isTokenOnSameLine = isTokenOnSameLine;
},
/* @typescript-eslint/experimental-utils/dist/ast-utils/predicates.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ts_estree_1 = require(739) /* ../ts-estree */;
    function isOptionalChainPunctuator(token) {
        return token.type === ts_estree_1.AST_TOKEN_TYPES.Punctuator && token.value === '?.';
    }
    exports.isOptionalChainPunctuator = isOptionalChainPunctuator;
    function isNotOptionalChainPunctuator(token) {
        return !isOptionalChainPunctuator(token);
    }
    exports.isNotOptionalChainPunctuator = isNotOptionalChainPunctuator;
    function isNonNullAssertionPunctuator(token) {
        return token.type === ts_estree_1.AST_TOKEN_TYPES.Punctuator && token.value === '!';
    }
    exports.isNonNullAssertionPunctuator = isNonNullAssertionPunctuator;
    function isNotNonNullAssertionPunctuator(token) {
        return !isNonNullAssertionPunctuator(token);
    }
    exports.isNotNonNullAssertionPunctuator = isNotNonNullAssertionPunctuator;
    /**
     * Returns true if and only if the node represents: foo?.() or foo.bar?.()
     */
    function isOptionalCallExpression(node) {
        return (node.type === ts_estree_1.AST_NODE_TYPES.CallExpression &&
            // this flag means the call expression itself is option
            // i.e. it is foo.bar?.() and not foo?.bar()
            node.optional);
    }
    exports.isOptionalCallExpression = isOptionalCallExpression;
    /**
     * Returns true if and only if the node represents logical OR
     */
    function isLogicalOrOperator(node) {
        return (node.type === ts_estree_1.AST_NODE_TYPES.LogicalExpression && node.operator === '||');
    }
    exports.isLogicalOrOperator = isLogicalOrOperator;
    /**
     * Checks if a node is a type assertion:
     * ```
     * x as foo
     * <foo>x
     * ```
     */
    function isTypeAssertion(node) {
        if (!node) {
            return false;
        }
        return (node.type === ts_estree_1.AST_NODE_TYPES.TSAsExpression ||
            node.type === ts_estree_1.AST_NODE_TYPES.TSTypeAssertion);
    }
    exports.isTypeAssertion = isTypeAssertion;
    function isVariableDeclarator(node) {
        return (node === null || node === void 0 ? void 0 : node.type) === ts_estree_1.AST_NODE_TYPES.VariableDeclarator;
    }
    exports.isVariableDeclarator = isVariableDeclarator;
    function isFunction(node) {
        if (!node) {
            return false;
        }
        return [
            ts_estree_1.AST_NODE_TYPES.ArrowFunctionExpression,
            ts_estree_1.AST_NODE_TYPES.FunctionDeclaration,
            ts_estree_1.AST_NODE_TYPES.FunctionExpression,
        ].includes(node.type);
    }
    exports.isFunction = isFunction;
    function isFunctionType(node) {
        if (!node) {
            return false;
        }
        return [
            ts_estree_1.AST_NODE_TYPES.TSCallSignatureDeclaration,
            ts_estree_1.AST_NODE_TYPES.TSConstructorType,
            ts_estree_1.AST_NODE_TYPES.TSConstructSignatureDeclaration,
            ts_estree_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression,
            ts_estree_1.AST_NODE_TYPES.TSFunctionType,
            ts_estree_1.AST_NODE_TYPES.TSMethodSignature,
        ].includes(node.type);
    }
    exports.isFunctionType = isFunctionType;
    function isFunctionOrFunctionType(node) {
        return isFunction(node) || isFunctionType(node);
    }
    exports.isFunctionOrFunctionType = isFunctionOrFunctionType;
    function isTSFunctionType(node) {
        return (node === null || node === void 0 ? void 0 : node.type) === ts_estree_1.AST_NODE_TYPES.TSFunctionType;
    }
    exports.isTSFunctionType = isTSFunctionType;
    function isTSConstructorType(node) {
        return (node === null || node === void 0 ? void 0 : node.type) === ts_estree_1.AST_NODE_TYPES.TSConstructorType;
    }
    exports.isTSConstructorType = isTSConstructorType;
    function isClassOrTypeElement(node) {
        if (!node) {
            return false;
        }
        return [
            // ClassElement
            ts_estree_1.AST_NODE_TYPES.ClassProperty,
            ts_estree_1.AST_NODE_TYPES.FunctionExpression,
            ts_estree_1.AST_NODE_TYPES.MethodDefinition,
            ts_estree_1.AST_NODE_TYPES.TSAbstractClassProperty,
            ts_estree_1.AST_NODE_TYPES.TSAbstractMethodDefinition,
            ts_estree_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression,
            ts_estree_1.AST_NODE_TYPES.TSIndexSignature,
            // TypeElement
            ts_estree_1.AST_NODE_TYPES.TSCallSignatureDeclaration,
            ts_estree_1.AST_NODE_TYPES.TSConstructSignatureDeclaration,
            // AST_NODE_TYPES.TSIndexSignature,
            ts_estree_1.AST_NODE_TYPES.TSMethodSignature,
            ts_estree_1.AST_NODE_TYPES.TSPropertySignature,
        ].includes(node.type);
    }
    exports.isClassOrTypeElement = isClassOrTypeElement;
    /**
     * Checks if a node is a constructor method.
     */
    function isConstructor(node) {
        return ((node === null || node === void 0 ? void 0 : node.type) === ts_estree_1.AST_NODE_TYPES.MethodDefinition &&
            node.kind === 'constructor');
    }
    exports.isConstructor = isConstructor;
    /**
     * Checks if a node is a setter method.
     */
    function isSetter(node) {
        return (!!node &&
            (node.type === ts_estree_1.AST_NODE_TYPES.MethodDefinition ||
                node.type === ts_estree_1.AST_NODE_TYPES.Property) &&
            node.kind === 'set');
    }
    exports.isSetter = isSetter;
    function isIdentifier(node) {
        return (node === null || node === void 0 ? void 0 : node.type) === ts_estree_1.AST_NODE_TYPES.Identifier;
    }
    exports.isIdentifier = isIdentifier;
    /**
     * Checks if a node represents an `await …` expression.
     */
    function isAwaitExpression(node) {
        return (node === null || node === void 0 ? void 0 : node.type) === ts_estree_1.AST_NODE_TYPES.AwaitExpression;
    }
    exports.isAwaitExpression = isAwaitExpression;
    /**
     * Checks if a possible token is the `await` keyword.
     */
    function isAwaitKeyword(node) {
        return (node === null || node === void 0 ? void 0 : node.type) === ts_estree_1.AST_TOKEN_TYPES.Identifier && node.value === 'await';
    }
    exports.isAwaitKeyword = isAwaitKeyword;
    function isLoop(node) {
        if (!node) {
            return false;
        }
        return (node.type === ts_estree_1.AST_NODE_TYPES.DoWhileStatement ||
            node.type === ts_estree_1.AST_NODE_TYPES.ForStatement ||
            node.type === ts_estree_1.AST_NODE_TYPES.ForInStatement ||
            node.type === ts_estree_1.AST_NODE_TYPES.ForOfStatement ||
            node.type === ts_estree_1.AST_NODE_TYPES.WhileStatement);
    }
    exports.isLoop = isLoop;
},
/* @typescript-eslint/experimental-utils/dist/ts-estree.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var types_1 = require(740) /* @typescript-eslint/types */;
    __esExport("AST_NODE_TYPES", types_1.AST_NODE_TYPES);
    __esExport("AST_TOKEN_TYPES", types_1.AST_TOKEN_TYPES);
    __esExport("TSESTree", types_1.TSESTree);
},
/* @typescript-eslint/types/dist/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    var ast_spec_1 = require(741) /* ./ast-spec */;
    __esExport("AST_NODE_TYPES", ast_spec_1.AST_NODE_TYPES);
    var ast_spec_2 = require(741) /* ./ast-spec */;
    __esExport("AST_TOKEN_TYPES", ast_spec_2.AST_TOKEN_TYPES);
    __exportStar(require(742) /* ./lib */, exports);
    __exportStar(require(743) /* ./parser-options */, exports);
    __exportStar(require(744) /* ./ts-estree */, exports);
},
/* @typescript-eslint/types/dist/ast-spec.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var AST_NODE_TYPES;
    (function (AST_NODE_TYPES) {
        AST_NODE_TYPES["ArrayExpression"] = "ArrayExpression";
        AST_NODE_TYPES["ArrayPattern"] = "ArrayPattern";
        AST_NODE_TYPES["ArrowFunctionExpression"] = "ArrowFunctionExpression";
        AST_NODE_TYPES["AssignmentExpression"] = "AssignmentExpression";
        AST_NODE_TYPES["AssignmentPattern"] = "AssignmentPattern";
        AST_NODE_TYPES["AwaitExpression"] = "AwaitExpression";
        AST_NODE_TYPES["BinaryExpression"] = "BinaryExpression";
        AST_NODE_TYPES["BlockStatement"] = "BlockStatement";
        AST_NODE_TYPES["BreakStatement"] = "BreakStatement";
        AST_NODE_TYPES["CallExpression"] = "CallExpression";
        AST_NODE_TYPES["CatchClause"] = "CatchClause";
        AST_NODE_TYPES["ChainExpression"] = "ChainExpression";
        AST_NODE_TYPES["ClassBody"] = "ClassBody";
        AST_NODE_TYPES["ClassDeclaration"] = "ClassDeclaration";
        AST_NODE_TYPES["ClassExpression"] = "ClassExpression";
        AST_NODE_TYPES["ClassProperty"] = "ClassProperty";
        AST_NODE_TYPES["ConditionalExpression"] = "ConditionalExpression";
        AST_NODE_TYPES["ContinueStatement"] = "ContinueStatement";
        AST_NODE_TYPES["DebuggerStatement"] = "DebuggerStatement";
        AST_NODE_TYPES["Decorator"] = "Decorator";
        AST_NODE_TYPES["DoWhileStatement"] = "DoWhileStatement";
        AST_NODE_TYPES["EmptyStatement"] = "EmptyStatement";
        AST_NODE_TYPES["ExportAllDeclaration"] = "ExportAllDeclaration";
        AST_NODE_TYPES["ExportDefaultDeclaration"] = "ExportDefaultDeclaration";
        AST_NODE_TYPES["ExportNamedDeclaration"] = "ExportNamedDeclaration";
        AST_NODE_TYPES["ExportSpecifier"] = "ExportSpecifier";
        AST_NODE_TYPES["ExpressionStatement"] = "ExpressionStatement";
        AST_NODE_TYPES["ForInStatement"] = "ForInStatement";
        AST_NODE_TYPES["ForOfStatement"] = "ForOfStatement";
        AST_NODE_TYPES["ForStatement"] = "ForStatement";
        AST_NODE_TYPES["FunctionDeclaration"] = "FunctionDeclaration";
        AST_NODE_TYPES["FunctionExpression"] = "FunctionExpression";
        AST_NODE_TYPES["Identifier"] = "Identifier";
        AST_NODE_TYPES["IfStatement"] = "IfStatement";
        AST_NODE_TYPES["ImportDeclaration"] = "ImportDeclaration";
        AST_NODE_TYPES["ImportDefaultSpecifier"] = "ImportDefaultSpecifier";
        AST_NODE_TYPES["ImportExpression"] = "ImportExpression";
        AST_NODE_TYPES["ImportNamespaceSpecifier"] = "ImportNamespaceSpecifier";
        AST_NODE_TYPES["ImportSpecifier"] = "ImportSpecifier";
        AST_NODE_TYPES["JSXAttribute"] = "JSXAttribute";
        AST_NODE_TYPES["JSXClosingElement"] = "JSXClosingElement";
        AST_NODE_TYPES["JSXClosingFragment"] = "JSXClosingFragment";
        AST_NODE_TYPES["JSXElement"] = "JSXElement";
        AST_NODE_TYPES["JSXEmptyExpression"] = "JSXEmptyExpression";
        AST_NODE_TYPES["JSXExpressionContainer"] = "JSXExpressionContainer";
        AST_NODE_TYPES["JSXFragment"] = "JSXFragment";
        AST_NODE_TYPES["JSXIdentifier"] = "JSXIdentifier";
        AST_NODE_TYPES["JSXMemberExpression"] = "JSXMemberExpression";
        AST_NODE_TYPES["JSXNamespacedName"] = "JSXNamespacedName";
        AST_NODE_TYPES["JSXOpeningElement"] = "JSXOpeningElement";
        AST_NODE_TYPES["JSXOpeningFragment"] = "JSXOpeningFragment";
        AST_NODE_TYPES["JSXSpreadAttribute"] = "JSXSpreadAttribute";
        AST_NODE_TYPES["JSXSpreadChild"] = "JSXSpreadChild";
        AST_NODE_TYPES["JSXText"] = "JSXText";
        AST_NODE_TYPES["LabeledStatement"] = "LabeledStatement";
        AST_NODE_TYPES["Literal"] = "Literal";
        AST_NODE_TYPES["LogicalExpression"] = "LogicalExpression";
        AST_NODE_TYPES["MemberExpression"] = "MemberExpression";
        AST_NODE_TYPES["MetaProperty"] = "MetaProperty";
        AST_NODE_TYPES["MethodDefinition"] = "MethodDefinition";
        AST_NODE_TYPES["NewExpression"] = "NewExpression";
        AST_NODE_TYPES["ObjectExpression"] = "ObjectExpression";
        AST_NODE_TYPES["ObjectPattern"] = "ObjectPattern";
        AST_NODE_TYPES["Program"] = "Program";
        AST_NODE_TYPES["Property"] = "Property";
        AST_NODE_TYPES["RestElement"] = "RestElement";
        AST_NODE_TYPES["ReturnStatement"] = "ReturnStatement";
        AST_NODE_TYPES["SequenceExpression"] = "SequenceExpression";
        AST_NODE_TYPES["SpreadElement"] = "SpreadElement";
        AST_NODE_TYPES["Super"] = "Super";
        AST_NODE_TYPES["SwitchCase"] = "SwitchCase";
        AST_NODE_TYPES["SwitchStatement"] = "SwitchStatement";
        AST_NODE_TYPES["TaggedTemplateExpression"] = "TaggedTemplateExpression";
        AST_NODE_TYPES["TemplateElement"] = "TemplateElement";
        AST_NODE_TYPES["TemplateLiteral"] = "TemplateLiteral";
        AST_NODE_TYPES["ThisExpression"] = "ThisExpression";
        AST_NODE_TYPES["ThrowStatement"] = "ThrowStatement";
        AST_NODE_TYPES["TryStatement"] = "TryStatement";
        AST_NODE_TYPES["UnaryExpression"] = "UnaryExpression";
        AST_NODE_TYPES["UpdateExpression"] = "UpdateExpression";
        AST_NODE_TYPES["VariableDeclaration"] = "VariableDeclaration";
        AST_NODE_TYPES["VariableDeclarator"] = "VariableDeclarator";
        AST_NODE_TYPES["WhileStatement"] = "WhileStatement";
        AST_NODE_TYPES["WithStatement"] = "WithStatement";
        AST_NODE_TYPES["YieldExpression"] = "YieldExpression";
        /**
         * TS-prefixed nodes
         */
        AST_NODE_TYPES["TSAbstractClassProperty"] = "TSAbstractClassProperty";
        AST_NODE_TYPES["TSAbstractKeyword"] = "TSAbstractKeyword";
        AST_NODE_TYPES["TSAbstractMethodDefinition"] = "TSAbstractMethodDefinition";
        AST_NODE_TYPES["TSAnyKeyword"] = "TSAnyKeyword";
        AST_NODE_TYPES["TSArrayType"] = "TSArrayType";
        AST_NODE_TYPES["TSAsExpression"] = "TSAsExpression";
        AST_NODE_TYPES["TSAsyncKeyword"] = "TSAsyncKeyword";
        AST_NODE_TYPES["TSBigIntKeyword"] = "TSBigIntKeyword";
        AST_NODE_TYPES["TSBooleanKeyword"] = "TSBooleanKeyword";
        AST_NODE_TYPES["TSCallSignatureDeclaration"] = "TSCallSignatureDeclaration";
        AST_NODE_TYPES["TSClassImplements"] = "TSClassImplements";
        AST_NODE_TYPES["TSConditionalType"] = "TSConditionalType";
        AST_NODE_TYPES["TSConstructorType"] = "TSConstructorType";
        AST_NODE_TYPES["TSConstructSignatureDeclaration"] = "TSConstructSignatureDeclaration";
        AST_NODE_TYPES["TSDeclareFunction"] = "TSDeclareFunction";
        AST_NODE_TYPES["TSDeclareKeyword"] = "TSDeclareKeyword";
        AST_NODE_TYPES["TSEmptyBodyFunctionExpression"] = "TSEmptyBodyFunctionExpression";
        AST_NODE_TYPES["TSEnumDeclaration"] = "TSEnumDeclaration";
        AST_NODE_TYPES["TSEnumMember"] = "TSEnumMember";
        AST_NODE_TYPES["TSExportAssignment"] = "TSExportAssignment";
        AST_NODE_TYPES["TSExportKeyword"] = "TSExportKeyword";
        AST_NODE_TYPES["TSExternalModuleReference"] = "TSExternalModuleReference";
        AST_NODE_TYPES["TSFunctionType"] = "TSFunctionType";
        AST_NODE_TYPES["TSImportEqualsDeclaration"] = "TSImportEqualsDeclaration";
        AST_NODE_TYPES["TSImportType"] = "TSImportType";
        AST_NODE_TYPES["TSIndexedAccessType"] = "TSIndexedAccessType";
        AST_NODE_TYPES["TSIndexSignature"] = "TSIndexSignature";
        AST_NODE_TYPES["TSInferType"] = "TSInferType";
        AST_NODE_TYPES["TSInterfaceBody"] = "TSInterfaceBody";
        AST_NODE_TYPES["TSInterfaceDeclaration"] = "TSInterfaceDeclaration";
        AST_NODE_TYPES["TSInterfaceHeritage"] = "TSInterfaceHeritage";
        AST_NODE_TYPES["TSIntersectionType"] = "TSIntersectionType";
        AST_NODE_TYPES["TSIntrinsicKeyword"] = "TSIntrinsicKeyword";
        AST_NODE_TYPES["TSLiteralType"] = "TSLiteralType";
        AST_NODE_TYPES["TSMappedType"] = "TSMappedType";
        AST_NODE_TYPES["TSMethodSignature"] = "TSMethodSignature";
        AST_NODE_TYPES["TSModuleBlock"] = "TSModuleBlock";
        AST_NODE_TYPES["TSModuleDeclaration"] = "TSModuleDeclaration";
        AST_NODE_TYPES["TSNamedTupleMember"] = "TSNamedTupleMember";
        AST_NODE_TYPES["TSNamespaceExportDeclaration"] = "TSNamespaceExportDeclaration";
        AST_NODE_TYPES["TSNeverKeyword"] = "TSNeverKeyword";
        AST_NODE_TYPES["TSNonNullExpression"] = "TSNonNullExpression";
        AST_NODE_TYPES["TSNullKeyword"] = "TSNullKeyword";
        AST_NODE_TYPES["TSNumberKeyword"] = "TSNumberKeyword";
        AST_NODE_TYPES["TSObjectKeyword"] = "TSObjectKeyword";
        AST_NODE_TYPES["TSOptionalType"] = "TSOptionalType";
        AST_NODE_TYPES["TSParameterProperty"] = "TSParameterProperty";
        AST_NODE_TYPES["TSParenthesizedType"] = "TSParenthesizedType";
        AST_NODE_TYPES["TSPrivateKeyword"] = "TSPrivateKeyword";
        AST_NODE_TYPES["TSPropertySignature"] = "TSPropertySignature";
        AST_NODE_TYPES["TSProtectedKeyword"] = "TSProtectedKeyword";
        AST_NODE_TYPES["TSPublicKeyword"] = "TSPublicKeyword";
        AST_NODE_TYPES["TSQualifiedName"] = "TSQualifiedName";
        AST_NODE_TYPES["TSReadonlyKeyword"] = "TSReadonlyKeyword";
        AST_NODE_TYPES["TSRestType"] = "TSRestType";
        AST_NODE_TYPES["TSStaticKeyword"] = "TSStaticKeyword";
        AST_NODE_TYPES["TSStringKeyword"] = "TSStringKeyword";
        AST_NODE_TYPES["TSSymbolKeyword"] = "TSSymbolKeyword";
        AST_NODE_TYPES["TSTemplateLiteralType"] = "TSTemplateLiteralType";
        AST_NODE_TYPES["TSThisType"] = "TSThisType";
        AST_NODE_TYPES["TSTupleType"] = "TSTupleType";
        AST_NODE_TYPES["TSTypeAliasDeclaration"] = "TSTypeAliasDeclaration";
        AST_NODE_TYPES["TSTypeAnnotation"] = "TSTypeAnnotation";
        AST_NODE_TYPES["TSTypeAssertion"] = "TSTypeAssertion";
        AST_NODE_TYPES["TSTypeLiteral"] = "TSTypeLiteral";
        AST_NODE_TYPES["TSTypeOperator"] = "TSTypeOperator";
        AST_NODE_TYPES["TSTypeParameter"] = "TSTypeParameter";
        AST_NODE_TYPES["TSTypeParameterDeclaration"] = "TSTypeParameterDeclaration";
        AST_NODE_TYPES["TSTypeParameterInstantiation"] = "TSTypeParameterInstantiation";
        AST_NODE_TYPES["TSTypePredicate"] = "TSTypePredicate";
        AST_NODE_TYPES["TSTypeQuery"] = "TSTypeQuery";
        AST_NODE_TYPES["TSTypeReference"] = "TSTypeReference";
        AST_NODE_TYPES["TSUndefinedKeyword"] = "TSUndefinedKeyword";
        AST_NODE_TYPES["TSUnionType"] = "TSUnionType";
        AST_NODE_TYPES["TSUnknownKeyword"] = "TSUnknownKeyword";
        AST_NODE_TYPES["TSVoidKeyword"] = "TSVoidKeyword";
    })(AST_NODE_TYPES = exports.AST_NODE_TYPES || (exports.AST_NODE_TYPES = {}));
    var AST_TOKEN_TYPES;
    (function (AST_TOKEN_TYPES) {
        AST_TOKEN_TYPES["Boolean"] = "Boolean";
        AST_TOKEN_TYPES["Identifier"] = "Identifier";
        AST_TOKEN_TYPES["JSXIdentifier"] = "JSXIdentifier";
        AST_TOKEN_TYPES["JSXText"] = "JSXText";
        AST_TOKEN_TYPES["Keyword"] = "Keyword";
        AST_TOKEN_TYPES["Null"] = "Null";
        AST_TOKEN_TYPES["Numeric"] = "Numeric";
        AST_TOKEN_TYPES["Punctuator"] = "Punctuator";
        AST_TOKEN_TYPES["RegularExpression"] = "RegularExpression";
        AST_TOKEN_TYPES["String"] = "String";
        AST_TOKEN_TYPES["Template"] = "Template";
        AST_TOKEN_TYPES["Block"] = "Block";
        AST_TOKEN_TYPES["Line"] = "Line";
    })(AST_TOKEN_TYPES = exports.AST_TOKEN_TYPES || (exports.AST_TOKEN_TYPES = {}));
},
/* @typescript-eslint/types/dist/lib.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/types/dist/parser-options.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/types/dist/ts-estree.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    exports.TSESTree = __importStar(require(741) /* ./ast-spec */);
},
/* @typescript-eslint/experimental-utils/dist/ast-utils/eslint-utils/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    __exportStar(require(746) /* ./astUtilities */, exports);
    __exportStar(require(747) /* ./PatternMatcher */, exports);
    __exportStar(require(748) /* ./predicates */, exports);
    __exportStar(require(749) /* ./ReferenceTracker */, exports);
    __exportStar(require(750) /* ./scopeAnalysis */, exports);
},
/* @typescript-eslint/experimental-utils/dist/ast-utils/eslint-utils/astUtilities.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const eslintUtils = __importStar(require(350) /* eslint-utils */);
    /**
     * Get the proper location of a given function node to report.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/ast-utils.html#getfunctionheadlocation}
     */
    const getFunctionHeadLocation = eslintUtils.getFunctionHeadLocation;
    exports.getFunctionHeadLocation = getFunctionHeadLocation;
    /**
     * Get the name and kind of a given function node.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/ast-utils.html#getfunctionnamewithkind}
     */
    const getFunctionNameWithKind = eslintUtils.getFunctionNameWithKind;
    exports.getFunctionNameWithKind = getFunctionNameWithKind;
    /**
     * Get the property name of a given property node.
     * If the node is a computed property, this tries to compute the property name by the getStringIfConstant function.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/ast-utils.html#getpropertyname}
     * @returns The property name of the node. If the property name is not constant then it returns `null`.
     */
    const getPropertyName = eslintUtils.getPropertyName;
    exports.getPropertyName = getPropertyName;
    /**
     * Get the value of a given node if it can decide the value statically.
     * If the 2nd parameter `initialScope` was given, this function tries to resolve identifier references which are in the
     * given node as much as possible. In the resolving way, it does on the assumption that built-in global objects have
     * not been modified.
     * For example, it considers `Symbol.iterator`, ` String.raw``hello`` `, and `Object.freeze({a: 1}).a` as static.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/ast-utils.html#getstaticvalue}
     * @returns The `{ value: any }` shaped object. The `value` property is the static value. If it couldn't compute the
     * static value of the node, it returns `null`.
     */
    const getStaticValue = eslintUtils.getStaticValue;
    exports.getStaticValue = getStaticValue;
    /**
     * Get the string value of a given node.
     * This function is a tiny wrapper of the getStaticValue function.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/ast-utils.html#getstringifconstant}
     */
    const getStringIfConstant = eslintUtils.getStringIfConstant;
    exports.getStringIfConstant = getStringIfConstant;
    /**
     * Check whether a given node has any side effect or not.
     * The side effect means that it may modify a certain variable or object member. This function considers the node which
     * contains the following types as the node which has side effects:
     * - `AssignmentExpression`
     * - `AwaitExpression`
     * - `CallExpression`
     * - `ImportExpression`
     * - `NewExpression`
     * - `UnaryExpression([operator = "delete"])`
     * - `UpdateExpression`
     * - `YieldExpression`
     * - When `options.considerGetters` is `true`:
     *   - `MemberExpression`
     * - When `options.considerImplicitTypeConversion` is `true`:
     *   - `BinaryExpression([operator = "==" | "!=" | "<" | "<=" | ">" | ">=" | "<<" | ">>" | ">>>" | "+" | "-" | "*" | "/" | "%" | "|" | "^" | "&" | "in"])`
     *   - `MemberExpression([computed = true])`
     *   - `MethodDefinition([computed = true])`
     *   - `Property([computed = true])`
     *   - `UnaryExpression([operator = "-" | "+" | "!" | "~"])`
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/ast-utils.html#hassideeffect}
     */
    const hasSideEffect = eslintUtils.hasSideEffect;
    exports.hasSideEffect = hasSideEffect;
    /**
     * Check whether a given node is parenthesized or not.
     * This function detects it correctly even if it's parenthesized by specific syntax.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/ast-utils.html#isparenthesized}
     * @returns `true` if the node is parenthesized.
     * If `times` was given, it returns `true` only if the node is parenthesized the `times` times.
     * For example, `isParenthesized(2, node, sourceCode)` returns true for `((foo))`, but not for `(foo)`.
     */
    const isParenthesized = eslintUtils.isParenthesized;
    exports.isParenthesized = isParenthesized;
},
/* @typescript-eslint/experimental-utils/dist/ast-utils/eslint-utils/PatternMatcher.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const eslintUtils = __importStar(require(350) /* eslint-utils */);
    /**
     * The class to find a pattern in strings as handling escape sequences.
     * It ignores the found pattern if it's escaped with `\`.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/ast-utils.html#patternmatcher-class}
     */
    const PatternMatcher = eslintUtils.PatternMatcher;
    exports.PatternMatcher = PatternMatcher;
},
/* @typescript-eslint/experimental-utils/dist/ast-utils/eslint-utils/predicates.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const eslintUtils = __importStar(require(350) /* eslint-utils */);
    const isArrowToken = eslintUtils.isArrowToken;
    exports.isArrowToken = isArrowToken;
    const isNotArrowToken = eslintUtils.isNotArrowToken;
    exports.isNotArrowToken = isNotArrowToken;
    const isClosingBraceToken = eslintUtils.isClosingBraceToken;
    exports.isClosingBraceToken = isClosingBraceToken;
    const isNotClosingBraceToken = eslintUtils.isNotClosingBraceToken;
    exports.isNotClosingBraceToken = isNotClosingBraceToken;
    const isClosingBracketToken = eslintUtils.isClosingBracketToken;
    exports.isClosingBracketToken = isClosingBracketToken;
    const isNotClosingBracketToken = eslintUtils.isNotClosingBracketToken;
    exports.isNotClosingBracketToken = isNotClosingBracketToken;
    const isClosingParenToken = eslintUtils.isClosingParenToken;
    exports.isClosingParenToken = isClosingParenToken;
    const isNotClosingParenToken = eslintUtils.isNotClosingParenToken;
    exports.isNotClosingParenToken = isNotClosingParenToken;
    const isColonToken = eslintUtils.isColonToken;
    exports.isColonToken = isColonToken;
    const isNotColonToken = eslintUtils.isNotColonToken;
    exports.isNotColonToken = isNotColonToken;
    const isCommaToken = eslintUtils.isCommaToken;
    exports.isCommaToken = isCommaToken;
    const isNotCommaToken = eslintUtils.isNotCommaToken;
    exports.isNotCommaToken = isNotCommaToken;
    const isCommentToken = eslintUtils.isCommentToken;
    exports.isCommentToken = isCommentToken;
    const isNotCommentToken = eslintUtils.isNotCommentToken;
    exports.isNotCommentToken = isNotCommentToken;
    const isOpeningBraceToken = eslintUtils.isOpeningBraceToken;
    exports.isOpeningBraceToken = isOpeningBraceToken;
    const isNotOpeningBraceToken = eslintUtils.isNotOpeningBraceToken;
    exports.isNotOpeningBraceToken = isNotOpeningBraceToken;
    const isOpeningBracketToken = eslintUtils.isOpeningBracketToken;
    exports.isOpeningBracketToken = isOpeningBracketToken;
    const isNotOpeningBracketToken = eslintUtils.isNotOpeningBracketToken;
    exports.isNotOpeningBracketToken = isNotOpeningBracketToken;
    const isOpeningParenToken = eslintUtils.isOpeningParenToken;
    exports.isOpeningParenToken = isOpeningParenToken;
    const isNotOpeningParenToken = eslintUtils.isNotOpeningParenToken;
    exports.isNotOpeningParenToken = isNotOpeningParenToken;
    const isSemicolonToken = eslintUtils.isSemicolonToken;
    exports.isSemicolonToken = isSemicolonToken;
    const isNotSemicolonToken = eslintUtils.isNotSemicolonToken;
    exports.isNotSemicolonToken = isNotSemicolonToken;
},
/* @typescript-eslint/experimental-utils/dist/ast-utils/eslint-utils/ReferenceTracker.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    /* eslint-disable @typescript-eslint/no-namespace */
    const eslintUtils = __importStar(require(350) /* eslint-utils */);
    const ReferenceTrackerREAD = eslintUtils.ReferenceTracker.READ;
    const ReferenceTrackerCALL = eslintUtils.ReferenceTracker.CALL;
    const ReferenceTrackerCONSTRUCT = eslintUtils.ReferenceTracker.CONSTRUCT;
    const ReferenceTrackerESM = eslintUtils.ReferenceTracker.ESM;
    /**
     * The tracker for references. This provides reference tracking for global variables, CommonJS modules, and ES modules.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/scope-utils.html#referencetracker-class}
     */
    const ReferenceTracker = eslintUtils.ReferenceTracker;
    exports.ReferenceTracker = ReferenceTracker;
},
/* @typescript-eslint/experimental-utils/dist/ast-utils/eslint-utils/scopeAnalysis.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const eslintUtils = __importStar(require(350) /* eslint-utils */);
    /**
     * Get the variable of a given name.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/scope-utils.html#findvariable}
     */
    const findVariable = eslintUtils.findVariable;
    exports.findVariable = findVariable;
    /**
     * Get the innermost scope which contains a given node.
     *
     * @see {@link https://eslint-utils.mysticatea.dev/api/scope-utils.html#getinnermostscope}
     * @returns The innermost scope which contains the given node.
     * If such scope doesn't exist then it returns the 1st argument `initialScope`.
     */
    const getInnermostScope = eslintUtils.getInnermostScope;
    exports.getInnermostScope = getInnermostScope;
},
/* @typescript-eslint/experimental-utils/dist/eslint-utils/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    __exportStar(require(752) /* ./applyDefault */, exports);
    __exportStar(require(754) /* ./batchedSingleLineTests */, exports);
    __exportStar(require(755) /* ./getParserServices */, exports);
    __exportStar(require(756) /* ./InferTypesFromRule */, exports);
    __exportStar(require(757) /* ./RuleCreator */, exports);
    __exportStar(require(758) /* ./RuleTester */, exports);
    __exportStar(require(753) /* ./deepMerge */, exports);
},
/* @typescript-eslint/experimental-utils/dist/eslint-utils/applyDefault.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const deepMerge_1 = require(753) /* ./deepMerge */;
    /**
     * Pure function - doesn't mutate either parameter!
     * Uses the default options and overrides with the options provided by the user
     * @param defaultOptions the defaults
     * @param userOptions the user opts
     * @returns the options with defaults
     */
    function applyDefault(defaultOptions, userOptions) {
        // clone defaults
        const options = JSON.parse(JSON.stringify(defaultOptions));
        if (userOptions === null || userOptions === undefined) {
            return options;
        }
        options.forEach((opt, i) => {
            if (userOptions[i] !== undefined) {
                const userOpt = userOptions[i];
                if (deepMerge_1.isObjectNotArray(userOpt) && deepMerge_1.isObjectNotArray(opt)) {
                    options[i] = deepMerge_1.deepMerge(opt, userOpt);
                }
                else {
                    options[i] = userOpt;
                }
            }
        });
        return options;
    }
    exports.applyDefault = applyDefault;
},
/* @typescript-eslint/experimental-utils/dist/eslint-utils/deepMerge.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    /**
     * Check if the variable contains an object strictly rejecting arrays
     * @param obj an object
     * @returns `true` if obj is an object
     */
    function isObjectNotArray(obj) {
        return typeof obj === 'object' && !Array.isArray(obj);
    }
    exports.isObjectNotArray = isObjectNotArray;
    /**
     * Pure function - doesn't mutate either parameter!
     * Merges two objects together deeply, overwriting the properties in first with the properties in second
     * @param first The first object
     * @param second The second object
     * @returns a new object
     */
    function deepMerge(first = {}, second = {}) {
        // get the unique set of keys across both objects
        const keys = new Set(Object.keys(first).concat(Object.keys(second)));
        return Array.from(keys).reduce((acc, key) => {
            const firstHasKey = key in first;
            const secondHasKey = key in second;
            const firstValue = first[key];
            const secondValue = second[key];
            if (firstHasKey && secondHasKey) {
                if (isObjectNotArray(firstValue) && isObjectNotArray(secondValue)) {
                    // object type
                    acc[key] = deepMerge(firstValue, secondValue);
                }
                else {
                    // value type
                    acc[key] = secondValue;
                }
            }
            else if (firstHasKey) {
                acc[key] = firstValue;
            }
            else {
                acc[key] = secondValue;
            }
            return acc;
        }, {});
    }
    exports.deepMerge = deepMerge;
},
/* @typescript-eslint/experimental-utils/dist/eslint-utils/batchedSingleLineTests.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function batchedSingleLineTests(options) {
        // eslint counts lines from 1
        const lineOffset = options.code.startsWith('\n') ? 2 : 1;
        const output = 'output' in options && options.output
            ? options.output.trim().split('\n')
            : null;
        return options.code
            .trim()
            .split('\n')
            .map((code, i) => {
            const lineNum = i + lineOffset;
            const errors = 'errors' in options
                ? options.errors.filter(e => e.line === lineNum)
                : [];
            const returnVal = Object.assign(Object.assign({}, options), { code, errors: errors.map(e => (Object.assign(Object.assign({}, e), { line: 1 }))) });
            if (output === null || output === void 0 ? void 0 : output[i]) {
                return Object.assign(Object.assign({}, returnVal), { output: output[i] });
            }
            return returnVal;
        });
    }
    exports.batchedSingleLineTests = batchedSingleLineTests;
},
/* @typescript-eslint/experimental-utils/dist/eslint-utils/getParserServices.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ERROR_MESSAGE = 'You have used a rule which requires parserServices to be generated. You must therefore provide a value for the "parserOptions.project" property for @typescript-eslint/parser.';
    /**
     * Try to retrieve typescript parser service from context
     */
    function getParserServices(context, allowWithoutFullTypeInformation = false) {
        var _a;
        // backwards compatibility check
        // old versions of the parser would not return any parserServices unless parserOptions.project was set
        if (!context.parserServices ||
            !context.parserServices.program ||
            !context.parserServices.esTreeNodeToTSNodeMap ||
            !context.parserServices.tsNodeToESTreeNodeMap) {
            throw new Error(ERROR_MESSAGE);
        }
        const hasFullTypeInformation = (_a = context.parserServices.hasFullTypeInformation) !== null && _a !== void 0 ? _a :
            /* backwards compatible */ true;
        // if a rule requires full type information, then hard fail if it doesn't exist
        // this forces the user to supply parserOptions.project
        if (!hasFullTypeInformation && !allowWithoutFullTypeInformation) {
            throw new Error(ERROR_MESSAGE);
        }
        return context.parserServices;
    }
    exports.getParserServices = getParserServices;
},
/* @typescript-eslint/experimental-utils/dist/eslint-utils/InferTypesFromRule.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/experimental-utils/dist/eslint-utils/RuleCreator.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const applyDefault_1 = require(752) /* ./applyDefault */;
    function RuleCreator(urlCreator) {
        // This function will get much easier to call when this is merged https://github.com/Microsoft/TypeScript/pull/26349
        // TODO - when the above PR lands; add type checking for the context.report `data` property
        return function createRule({ name, meta, defaultOptions, create, }) {
            return {
                meta: Object.assign(Object.assign({}, meta), { docs: Object.assign(Object.assign({}, meta.docs), { url: urlCreator(name) }) }),
                create(context) {
                    const optionsWithDefault = applyDefault_1.applyDefault(defaultOptions, context.options);
                    return create(context, optionsWithDefault);
                },
            };
        };
    }
    exports.RuleCreator = RuleCreator;
},
/* @typescript-eslint/experimental-utils/dist/eslint-utils/RuleTester.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _RuleTester_options;
    __esModule();
    const path = __importStar(require("path"));
    const TSESLint = __importStar(require(759) /* ../ts-eslint */);
    const parser = '@typescript-eslint/parser';
    class RuleTester extends TSESLint.RuleTester {
        // as of eslint 6 you have to provide an absolute path to the parser
        // but that's not as clean to type, this saves us trying to manually enforce
        // that contributors require.resolve everything
        constructor(options) {
            var _a, _b;
            super(Object.assign(Object.assign({}, options), { parserOptions: Object.assign(Object.assign({}, options.parserOptions), { warnOnUnsupportedTypeScriptVersion: (_b = (_a = options.parserOptions) === null || _a === void 0 ? void 0 : _a.warnOnUnsupportedTypeScriptVersion) !== null && _b !== void 0 ? _b : false }), parser: require.resolve(options.parser) }));
            _RuleTester_options.set(this, void 0);
            __classPrivateFieldSet(this, _RuleTester_options, options, "f");
            // make sure that the parser doesn't hold onto file handles between tests
            // on linux (i.e. our CI env), there can be very a limited number of watch handles available
            afterAll(() => {
                try {
                    // instead of creating a hard dependency, just use a soft require
                    // a bit weird, but if they're using this tooling, it'll be installed
                    // eslint-disable-next-line @typescript-eslint/no-unsafe-call, @typescript-eslint/no-unsafe-member-access
                    require(parser).clearCaches();
                }
                catch (_a) {
                    // ignored
                }
            });
        }
        getFilename(options) {
            var _a;
            if (options) {
                const filename = `file.ts${((_a = options.ecmaFeatures) === null || _a === void 0 ? void 0 : _a.jsx) ? 'x' : ''}`;
                if (options.project) {
                    return path.join(options.tsconfigRootDir != null
                        ? options.tsconfigRootDir
                        : process.cwd(), filename);
                }
                return filename;
            }
            else if (__classPrivateFieldGet(this, _RuleTester_options, "f").parserOptions) {
                return this.getFilename(__classPrivateFieldGet(this, _RuleTester_options, "f").parserOptions);
            }
            return 'file.ts';
        }
        // as of eslint 6 you have to provide an absolute path to the parser
        // If you don't do that at the test level, the test will fail somewhat cryptically...
        // This is a lot more explicit
        run(name, rule, testsReadonly) {
            const errorMessage = `Do not set the parser at the test level unless you want to use a parser other than ${parser}`;
            const tests = Object.assign({}, testsReadonly);
            // standardize the valid tests as objects
            tests.valid = tests.valid.map(test => {
                if (typeof test === 'string') {
                    return {
                        code: test,
                    };
                }
                return test;
            });
            tests.valid = tests.valid.map(test => {
                if (typeof test !== 'string') {
                    if (test.parser === parser) {
                        throw new Error(errorMessage);
                    }
                    if (!test.filename) {
                        return Object.assign(Object.assign({}, test), { filename: this.getFilename(test.parserOptions) });
                    }
                }
                return test;
            });
            tests.invalid = tests.invalid.map(test => {
                if (test.parser === parser) {
                    throw new Error(errorMessage);
                }
                if (!test.filename) {
                    return Object.assign(Object.assign({}, test), { filename: this.getFilename(test.parserOptions) });
                }
                return test;
            });
            super.run(name, rule, tests);
        }
    }
    exports.RuleTester = RuleTester;
    _RuleTester_options = new WeakMap();
    /**
     * Simple no-op tag to mark code samples as "should not format with prettier"
     *   for the internal/plugin-test-formatting lint rule
     */
    function noFormat(strings, ...keys) {
        const lastIndex = strings.length - 1;
        return (strings.slice(0, lastIndex).reduce((p, s, i) => p + s + keys[i], '') +
            strings[lastIndex]);
    }
    exports.noFormat = noFormat;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    __exportStar(require(760) /* ./AST */, exports);
    __exportStar(require(761) /* ./CLIEngine */, exports);
    __exportStar(require(762) /* ./ESLint */, exports);
    __exportStar(require(763) /* ./Linter */, exports);
    __exportStar(require(764) /* ./ParserOptions */, exports);
    __exportStar(require(765) /* ./Rule */, exports);
    __exportStar(require(766) /* ./RuleTester */, exports);
    __exportStar(require(767) /* ./Scope */, exports);
    __exportStar(require(894) /* ./SourceCode */, exports);
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/AST.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/CLIEngine.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const eslint_1 = require(203) /* eslint */;
    /**
     * The underlying utility that runs the ESLint command line interface. This object will read the filesystem for
     * configuration and file information but will not output any results. Instead, it allows you direct access to the
     * important information so you can deal with the output yourself.
     * @deprecated use the ESLint class instead
     */
    class CLIEngine extends eslint_1.CLIEngine {
    }
    exports.CLIEngine = CLIEngine;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/ESLint.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const eslint_1 = require(203) /* eslint */;
    // We want to export this class always so it's easy for end users to consume.
    // However on ESLint v6, this class will not exist, so we provide a fallback to make it clear
    // The only users of this should be users scripting ESLint locally, so _they_ should have the correct version installed.
    const _ESLint = (eslint_1.ESLint !== null && eslint_1.ESLint !== void 0 ? eslint_1.ESLint : function () {
        throw new Error('Attempted to construct an ESLint instance on less than ESLint v7.0.0');
    });
    /**
     * The ESLint class is the primary class to use in Node.js applications.
     * This class depends on the Node.js fs module and the file system, so you cannot use it in browsers.
     *
     * If you want to lint code on browsers, use the Linter class instead.
     *
     * @since 7.0.0
     */
    class ESLint extends _ESLint {
    }
    exports.ESLint = ESLint;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/Linter.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const eslint_1 = require(203) /* eslint */;
    /**
     * The Linter object does the actual evaluation of the JavaScript code. It doesn't do any filesystem operations, it
     * simply parses and reports on the code. In particular, the Linter object does not process configuration objects
     * or files.
     */
    class Linter extends eslint_1.Linter {
    }
    exports.Linter = Linter;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/ParserOptions.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/Rule.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/RuleTester.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const eslint_1 = require(203) /* eslint */;
    class RuleTester extends eslint_1.RuleTester {
    }
    exports.RuleTester = RuleTester;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/Scope.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint-disable @typescript-eslint/no-namespace */
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const scopeManager = __importStar(require(768) /* @typescript-eslint/scope-manager */);
    var Scope;
    (function (Scope) {
        Scope.ScopeType = scopeManager.ScopeType;
        Scope.DefinitionType = scopeManager.DefinitionType;
    })(Scope || (Scope = {}));
    exports.Scope = Scope;
},
/* @typescript-eslint/scope-manager/dist/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    var analyze_1 = require(769) /* ./analyze */;
    __esExport("analyze", analyze_1.analyze);
    __exportStar(require(776) /* ./definition */, exports);
    var Reference_1 = require(801) /* ./referencer/Reference */;
    __esExport("Reference", Reference_1.Reference);
    var Visitor_1 = require(793) /* ./referencer/Visitor */;
    __esExport("Visitor", Visitor_1.Visitor);
    var PatternVisitor_1 = require(795) /* ./referencer/PatternVisitor */;
    __esExport("PatternVisitor", PatternVisitor_1.PatternVisitor);
    __exportStar(require(796) /* ./scope */, exports);
    var ScopeManager_1 = require(893) /* ./ScopeManager */;
    __esExport("ScopeManager", ScopeManager_1.ScopeManager);
    __exportStar(require(802) /* ./variable */, exports);
},
/* @typescript-eslint/scope-manager/dist/analyze.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const visitor_keys_1 = require(770) /* @typescript-eslint/visitor-keys */;
    const referencer_1 = require(773) /* ./referencer */;
    const ScopeManager_1 = require(893) /* ./ScopeManager */;
    const lib_1 = require(825) /* ./lib */;
    const DEFAULT_OPTIONS = {
        childVisitorKeys: visitor_keys_1.visitorKeys,
        ecmaVersion: 2018,
        globalReturn: false,
        impliedStrict: false,
        jsxPragma: 'React',
        jsxFragmentName: null,
        lib: ['es2018'],
        sourceType: 'script',
        emitDecoratorMetadata: false,
    };
    function mapEcmaVersion(version) {
        if (version == null || version === 3 || version === 5) {
            return 'es5';
        }
        const year = version > 2000 ? version : 2015 + (version - 6);
        const lib = `es${year}`;
        return lib in lib_1.lib ? lib : year > 2020 ? 'esnext' : 'es5';
    }
    /**
     * Takes an AST and returns the analyzed scopes.
     */
    function analyze(tree, providedOptions) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        const ecmaVersion = (_a = providedOptions === null || providedOptions === void 0 ? void 0 : providedOptions.ecmaVersion) !== null && _a !== void 0 ? _a : DEFAULT_OPTIONS.ecmaVersion;
        const options = {
            childVisitorKeys: (_b = providedOptions === null || providedOptions === void 0 ? void 0 : providedOptions.childVisitorKeys) !== null && _b !== void 0 ? _b : DEFAULT_OPTIONS.childVisitorKeys,
            ecmaVersion,
            globalReturn: (_c = providedOptions === null || providedOptions === void 0 ? void 0 : providedOptions.globalReturn) !== null && _c !== void 0 ? _c : DEFAULT_OPTIONS.globalReturn,
            impliedStrict: (_d = providedOptions === null || providedOptions === void 0 ? void 0 : providedOptions.impliedStrict) !== null && _d !== void 0 ? _d : DEFAULT_OPTIONS.impliedStrict,
            jsxPragma: (providedOptions === null || providedOptions === void 0 ? void 0 : providedOptions.jsxPragma) === undefined
                ? DEFAULT_OPTIONS.jsxPragma
                : providedOptions.jsxPragma,
            jsxFragmentName: (_e = providedOptions === null || providedOptions === void 0 ? void 0 : providedOptions.jsxFragmentName) !== null && _e !== void 0 ? _e : DEFAULT_OPTIONS.jsxFragmentName,
            sourceType: (_f = providedOptions === null || providedOptions === void 0 ? void 0 : providedOptions.sourceType) !== null && _f !== void 0 ? _f : DEFAULT_OPTIONS.sourceType,
            lib: (_g = providedOptions === null || providedOptions === void 0 ? void 0 : providedOptions.lib) !== null && _g !== void 0 ? _g : [mapEcmaVersion(ecmaVersion)],
            emitDecoratorMetadata: (_h = providedOptions === null || providedOptions === void 0 ? void 0 : providedOptions.emitDecoratorMetadata) !== null && _h !== void 0 ? _h : DEFAULT_OPTIONS.emitDecoratorMetadata,
        };
        // ensure the option is lower cased
        options.lib = options.lib.map(l => l.toLowerCase());
        const scopeManager = new ScopeManager_1.ScopeManager(options);
        const referencer = new referencer_1.Referencer(options, scopeManager);
        referencer.visit(tree);
        return scopeManager;
    }
    exports.analyze = analyze;
},
/* @typescript-eslint/visitor-keys/dist/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var get_keys_1 = require(771) /* ./get-keys */;
    __esExport("getKeys", get_keys_1.getKeys);
    var visitor_keys_1 = require(772) /* ./visitor-keys */;
    __esExport("visitorKeys", visitor_keys_1.visitorKeys);
},
/* @typescript-eslint/visitor-keys/dist/get-keys.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const eslint_visitor_keys_1 = require(321) /* eslint-visitor-keys */;
    const getKeys = eslint_visitor_keys_1.getKeys;
    exports.getKeys = getKeys;
},
/* @typescript-eslint/visitor-keys/dist/visitor-keys.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const eslintVisitorKeys = __importStar(require(321) /* eslint-visitor-keys */);
    const additionalKeys = {
        // ES2020
        ImportExpression: ['source'],
        // Additional Properties.
        ArrayPattern: ['decorators', 'elements', 'typeAnnotation'],
        ArrowFunctionExpression: ['typeParameters', 'params', 'returnType', 'body'],
        AssignmentPattern: ['decorators', 'left', 'right', 'typeAnnotation'],
        CallExpression: ['callee', 'typeParameters', 'arguments'],
        ClassDeclaration: [
            'decorators',
            'id',
            'typeParameters',
            'superClass',
            'superTypeParameters',
            'implements',
            'body',
        ],
        ClassExpression: [
            'decorators',
            'id',
            'typeParameters',
            'superClass',
            'superTypeParameters',
            'implements',
            'body',
        ],
        FunctionDeclaration: ['id', 'typeParameters', 'params', 'returnType', 'body'],
        FunctionExpression: ['id', 'typeParameters', 'params', 'returnType', 'body'],
        Identifier: ['decorators', 'typeAnnotation'],
        MethodDefinition: ['decorators', 'key', 'value'],
        NewExpression: ['callee', 'typeParameters', 'arguments'],
        ObjectPattern: ['decorators', 'properties', 'typeAnnotation'],
        RestElement: ['decorators', 'argument', 'typeAnnotation'],
        TaggedTemplateExpression: ['tag', 'typeParameters', 'quasi'],
        // JSX
        JSXOpeningElement: ['name', 'typeParameters', 'attributes'],
        JSXClosingFragment: [],
        JSXOpeningFragment: [],
        JSXSpreadChild: ['expression'],
        // Additional Nodes.
        ClassProperty: ['decorators', 'key', 'typeAnnotation', 'value'],
        Decorator: ['expression'],
        // TS-prefixed nodes
        TSAbstractClassProperty: ['decorators', 'key', 'typeAnnotation', 'value'],
        TSAbstractKeyword: [],
        TSAbstractMethodDefinition: ['key', 'value'],
        TSAnyKeyword: [],
        TSArrayType: ['elementType'],
        TSAsExpression: ['expression', 'typeAnnotation'],
        TSAsyncKeyword: [],
        TSBigIntKeyword: [],
        TSBooleanKeyword: [],
        TSCallSignatureDeclaration: ['typeParameters', 'params', 'returnType'],
        TSClassImplements: ['expression', 'typeParameters'],
        TSConditionalType: ['checkType', 'extendsType', 'trueType', 'falseType'],
        TSConstructorType: ['typeParameters', 'params', 'returnType'],
        TSConstructSignatureDeclaration: ['typeParameters', 'params', 'returnType'],
        TSDeclareFunction: ['id', 'typeParameters', 'params', 'returnType', 'body'],
        TSDeclareKeyword: [],
        TSEmptyBodyFunctionExpression: [
            'id',
            'typeParameters',
            'params',
            'returnType',
        ],
        TSEnumDeclaration: ['id', 'members'],
        TSEnumMember: ['id', 'initializer'],
        TSExportAssignment: ['expression'],
        TSExportKeyword: [],
        TSExternalModuleReference: ['expression'],
        TSFunctionType: ['typeParameters', 'params', 'returnType'],
        TSImportEqualsDeclaration: ['id', 'moduleReference'],
        TSImportType: ['parameter', 'qualifier', 'typeParameters'],
        TSIndexedAccessType: ['indexType', 'objectType'],
        TSIndexSignature: ['parameters', 'typeAnnotation'],
        TSInferType: ['typeParameter'],
        TSInterfaceBody: ['body'],
        TSInterfaceDeclaration: ['id', 'typeParameters', 'extends', 'body'],
        TSInterfaceHeritage: ['expression', 'typeParameters'],
        TSIntersectionType: ['types'],
        TSIntrinsicKeyword: [],
        TSLiteralType: ['literal'],
        TSMappedType: ['nameType', 'typeParameter', 'typeAnnotation'],
        TSMethodSignature: ['typeParameters', 'key', 'params', 'returnType'],
        TSModuleBlock: ['body'],
        TSModuleDeclaration: ['id', 'body'],
        TSNamedTupleMember: ['elementType'],
        TSNamespaceExportDeclaration: ['id'],
        TSNeverKeyword: [],
        TSNonNullExpression: ['expression'],
        TSNullKeyword: [],
        TSNumberKeyword: [],
        TSObjectKeyword: [],
        TSOptionalType: ['typeAnnotation'],
        TSParameterProperty: ['decorators', 'parameter'],
        TSParenthesizedType: ['typeAnnotation'],
        TSPrivateKeyword: [],
        TSPropertySignature: ['typeAnnotation', 'key', 'initializer'],
        TSProtectedKeyword: [],
        TSPublicKeyword: [],
        TSQualifiedName: ['left', 'right'],
        TSReadonlyKeyword: [],
        TSRestType: ['typeAnnotation'],
        TSStaticKeyword: [],
        TSStringKeyword: [],
        TSSymbolKeyword: [],
        TSTemplateLiteralType: ['quasis', 'types'],
        TSThisType: [],
        TSTupleType: ['elementTypes'],
        TSTypeAliasDeclaration: ['id', 'typeParameters', 'typeAnnotation'],
        TSTypeAnnotation: ['typeAnnotation'],
        TSTypeAssertion: ['typeAnnotation', 'expression'],
        TSTypeLiteral: ['members'],
        TSTypeOperator: ['typeAnnotation'],
        TSTypeParameter: ['name', 'constraint', 'default'],
        TSTypeParameterDeclaration: ['params'],
        TSTypeParameterInstantiation: ['params'],
        TSTypePredicate: ['typeAnnotation', 'parameterName'],
        TSTypeQuery: ['exprName'],
        TSTypeReference: ['typeName', 'typeParameters'],
        TSUndefinedKeyword: [],
        TSUnionType: ['types'],
        TSUnknownKeyword: [],
        TSVoidKeyword: [],
    };
    const visitorKeys = eslintVisitorKeys.unionWith(additionalKeys);
    exports.visitorKeys = visitorKeys;
},
/* @typescript-eslint/scope-manager/dist/referencer/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var Referencer_1 = require(774) /* ./Referencer */;
    __esExport("Referencer", Referencer_1.Referencer);
},
/* @typescript-eslint/scope-manager/dist/referencer/Referencer.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _Referencer_jsxPragma, _Referencer_jsxFragmentName, _Referencer_hasReferencedJsxFactory, _Referencer_hasReferencedJsxFragmentFactory, _Referencer_lib, _Referencer_emitDecoratorMetadata;
    __esModule();
    const types_1 = require(740) /* @typescript-eslint/types */;
    const ClassVisitor_1 = require(775) /* ./ClassVisitor */;
    const ExportVisitor_1 = require(823) /* ./ExportVisitor */;
    const ImportVisitor_1 = require(824) /* ./ImportVisitor */;
    const PatternVisitor_1 = require(795) /* ./PatternVisitor */;
    const Reference_1 = require(801) /* ./Reference */;
    const TypeVisitor_1 = require(792) /* ./TypeVisitor */;
    const Visitor_1 = require(793) /* ./Visitor */;
    const assert_1 = require(800) /* ../assert */;
    const definition_1 = require(776) /* ../definition */;
    const lib_1 = require(825) /* ../lib */;
    // Referencing variables and creating bindings.
    class Referencer extends Visitor_1.Visitor {
        constructor(options, scopeManager) {
            super(options);
            _Referencer_jsxPragma.set(this, void 0);
            _Referencer_jsxFragmentName.set(this, void 0);
            _Referencer_hasReferencedJsxFactory.set(this, false);
            _Referencer_hasReferencedJsxFragmentFactory.set(this, false);
            _Referencer_lib.set(this, void 0);
            _Referencer_emitDecoratorMetadata.set(this, void 0);
            this.scopeManager = scopeManager;
            __classPrivateFieldSet(this, _Referencer_jsxPragma, options.jsxPragma, "f");
            __classPrivateFieldSet(this, _Referencer_jsxFragmentName, options.jsxFragmentName, "f");
            __classPrivateFieldSet(this, _Referencer_lib, options.lib, "f");
            __classPrivateFieldSet(this, _Referencer_emitDecoratorMetadata, options.emitDecoratorMetadata, "f");
        }
        currentScope(dontThrowOnNull) {
            if (!dontThrowOnNull) {
                assert_1.assert(this.scopeManager.currentScope, 'aaa');
            }
            return this.scopeManager.currentScope;
        }
        close(node) {
            while (this.currentScope(true) && node === this.currentScope().block) {
                this.scopeManager.currentScope = this.currentScope().close(this.scopeManager);
            }
        }
        referencingDefaultValue(pattern, assignments, maybeImplicitGlobal, init) {
            assignments.forEach(assignment => {
                this.currentScope().referenceValue(pattern, Reference_1.ReferenceFlag.Write, assignment.right, maybeImplicitGlobal, init);
            });
        }
        populateGlobalsFromLib(globalScope) {
            for (const lib of __classPrivateFieldGet(this, _Referencer_lib, "f")) {
                const variables = lib_1.lib[lib];
                /* istanbul ignore if */ if (!variables) {
                    throw new Error(`Invalid value for lib provided: ${lib}`);
                }
                for (const [name, variable] of Object.entries(variables)) {
                    globalScope.defineImplicitVariable(name, variable);
                }
            }
            // for const assertions (`{} as const` / `<const>{}`)
            globalScope.defineImplicitVariable('const', {
                eslintImplicitGlobalSetting: 'readonly',
                isTypeVariable: true,
                isValueVariable: false,
            });
        }
        /**
         * Searches for a variable named "name" in the upper scopes and adds a pseudo-reference from itself to itself
         */
        referenceInSomeUpperScope(name) {
            let scope = this.scopeManager.currentScope;
            while (scope) {
                const variable = scope.set.get(name);
                if (!variable) {
                    scope = scope.upper;
                    continue;
                }
                scope.referenceValue(variable.identifiers[0]);
                return true;
            }
            return false;
        }
        referenceJsxPragma() {
            if (__classPrivateFieldGet(this, _Referencer_jsxPragma, "f") === null || __classPrivateFieldGet(this, _Referencer_hasReferencedJsxFactory, "f")) {
                return;
            }
            __classPrivateFieldSet(this, _Referencer_hasReferencedJsxFactory, this.referenceInSomeUpperScope(__classPrivateFieldGet(this, _Referencer_jsxPragma, "f")), "f");
        }
        referenceJsxFragment() {
            if (__classPrivateFieldGet(this, _Referencer_jsxFragmentName, "f") === null ||
                __classPrivateFieldGet(this, _Referencer_hasReferencedJsxFragmentFactory, "f")) {
                return;
            }
            __classPrivateFieldSet(this, _Referencer_hasReferencedJsxFragmentFactory, this.referenceInSomeUpperScope(__classPrivateFieldGet(this, _Referencer_jsxFragmentName, "f")), "f");
        }
        ///////////////////
        // Visit helpers //
        ///////////////////
        visitClass(node) {
            ClassVisitor_1.ClassVisitor.visit(this, node, __classPrivateFieldGet(this, _Referencer_emitDecoratorMetadata, "f"));
        }
        visitForIn(node) {
            if (node.left.type === types_1.AST_NODE_TYPES.VariableDeclaration &&
                node.left.kind !== 'var') {
                this.scopeManager.nestForScope(node);
            }
            if (node.left.type === types_1.AST_NODE_TYPES.VariableDeclaration) {
                this.visit(node.left);
                this.visitPattern(node.left.declarations[0].id, pattern => {
                    this.currentScope().referenceValue(pattern, Reference_1.ReferenceFlag.Write, node.right, null, true);
                });
            }
            else {
                this.visitPattern(node.left, (pattern, info) => {
                    const maybeImplicitGlobal = !this.currentScope().isStrict
                        ? {
                            pattern,
                            node,
                        }
                        : null;
                    this.referencingDefaultValue(pattern, info.assignments, maybeImplicitGlobal, false);
                    this.currentScope().referenceValue(pattern, Reference_1.ReferenceFlag.Write, node.right, maybeImplicitGlobal, false);
                }, { processRightHandNodes: true });
            }
            this.visit(node.right);
            this.visit(node.body);
            this.close(node);
        }
        visitFunctionParameterTypeAnnotation(node) {
            if ('typeAnnotation' in node) {
                this.visitType(node.typeAnnotation);
            }
            else if (node.type === types_1.AST_NODE_TYPES.AssignmentPattern) {
                this.visitType(node.left.typeAnnotation);
            }
            else if (node.type === types_1.AST_NODE_TYPES.TSParameterProperty) {
                this.visitFunctionParameterTypeAnnotation(node.parameter);
            }
        }
        visitFunction(node) {
            // FunctionDeclaration name is defined in upper scope
            // NOTE: Not referring variableScope. It is intended.
            // Since
            //  in ES5, FunctionDeclaration should be in FunctionBody.
            //  in ES6, FunctionDeclaration should be block scoped.
            var _a;
            if (node.type === types_1.AST_NODE_TYPES.FunctionExpression) {
                if (node.id) {
                    // FunctionExpression with name creates its special scope;
                    // FunctionExpressionNameScope.
                    this.scopeManager.nestFunctionExpressionNameScope(node);
                }
            }
            else if (node.id) {
                // id is defined in upper scope
                this.currentScope().defineIdentifier(node.id, new definition_1.FunctionNameDefinition(node.id, node));
            }
            // Consider this function is in the MethodDefinition.
            this.scopeManager.nestFunctionScope(node, false);
            // Process parameter declarations.
            for (const param of node.params) {
                this.visitPattern(param, (pattern, info) => {
                    this.currentScope().defineIdentifier(pattern, new definition_1.ParameterDefinition(pattern, node, info.rest));
                    this.referencingDefaultValue(pattern, info.assignments, null, true);
                }, { processRightHandNodes: true });
                this.visitFunctionParameterTypeAnnotation(param);
                (_a = param.decorators) === null || _a === void 0 ? void 0 : _a.forEach(d => this.visit(d));
            }
            this.visitType(node.returnType);
            this.visitType(node.typeParameters);
            // In TypeScript there are a number of function-like constructs which have no body,
            // so check it exists before traversing
            if (node.body) {
                // Skip BlockStatement to prevent creating BlockStatement scope.
                if (node.body.type === types_1.AST_NODE_TYPES.BlockStatement) {
                    this.visitChildren(node.body);
                }
                else {
                    this.visit(node.body);
                }
            }
            this.close(node);
        }
        visitProperty(node) {
            if (node.computed) {
                this.visit(node.key);
            }
            this.visit(node.value);
        }
        visitType(node) {
            if (!node) {
                return;
            }
            TypeVisitor_1.TypeVisitor.visit(this, node);
        }
        visitTypeAssertion(node) {
            this.visit(node.expression);
            this.visitType(node.typeAnnotation);
        }
        /////////////////////
        // Visit selectors //
        /////////////////////
        ArrowFunctionExpression(node) {
            this.visitFunction(node);
        }
        AssignmentExpression(node) {
            let left = node.left;
            switch (left.type) {
                case types_1.AST_NODE_TYPES.TSAsExpression:
                case types_1.AST_NODE_TYPES.TSTypeAssertion:
                    // explicitly visit the type annotation
                    this.visitType(left.typeAnnotation);
                // intentional fallthrough
                case types_1.AST_NODE_TYPES.TSNonNullExpression:
                    // unwrap the expression
                    left = left.expression;
            }
            if (PatternVisitor_1.PatternVisitor.isPattern(left)) {
                if (node.operator === '=') {
                    this.visitPattern(left, (pattern, info) => {
                        const maybeImplicitGlobal = !this.currentScope().isStrict
                            ? {
                                pattern,
                                node,
                            }
                            : null;
                        this.referencingDefaultValue(pattern, info.assignments, maybeImplicitGlobal, false);
                        this.currentScope().referenceValue(pattern, Reference_1.ReferenceFlag.Write, node.right, maybeImplicitGlobal, false);
                    }, { processRightHandNodes: true });
                }
                else if (left.type === types_1.AST_NODE_TYPES.Identifier) {
                    this.currentScope().referenceValue(left, Reference_1.ReferenceFlag.ReadWrite, node.right);
                }
            }
            else {
                this.visit(left);
            }
            this.visit(node.right);
        }
        BlockStatement(node) {
            if (this.scopeManager.isES6()) {
                this.scopeManager.nestBlockScope(node);
            }
            this.visitChildren(node);
            this.close(node);
        }
        BreakStatement() {
            // don't reference the break statement's label
        }
        CallExpression(node) {
            this.visitChildren(node, ['typeParameters']);
            this.visitType(node.typeParameters);
        }
        CatchClause(node) {
            this.scopeManager.nestCatchScope(node);
            if (node.param) {
                const param = node.param;
                this.visitPattern(param, (pattern, info) => {
                    this.currentScope().defineIdentifier(pattern, new definition_1.CatchClauseDefinition(param, node));
                    this.referencingDefaultValue(pattern, info.assignments, null, true);
                }, { processRightHandNodes: true });
            }
            this.visit(node.body);
            this.close(node);
        }
        ClassExpression(node) {
            this.visitClass(node);
        }
        ClassDeclaration(node) {
            this.visitClass(node);
        }
        ContinueStatement() {
            // don't reference the continue statement's label
        }
        ExportAllDeclaration() {
            // this defines no local variables
        }
        ExportDefaultDeclaration(node) {
            if (node.declaration.type === types_1.AST_NODE_TYPES.Identifier) {
                ExportVisitor_1.ExportVisitor.visit(this, node);
            }
            else {
                this.visit(node.declaration);
            }
        }
        ExportNamedDeclaration(node) {
            if (node.declaration) {
                this.visit(node.declaration);
            }
            else {
                ExportVisitor_1.ExportVisitor.visit(this, node);
            }
        }
        ForInStatement(node) {
            this.visitForIn(node);
        }
        ForOfStatement(node) {
            this.visitForIn(node);
        }
        ForStatement(node) {
            // Create ForStatement declaration.
            // NOTE: In ES6, ForStatement dynamically generates per iteration environment. However, this is
            // a static analyzer, we only generate one scope for ForStatement.
            if (node.init &&
                node.init.type === types_1.AST_NODE_TYPES.VariableDeclaration &&
                node.init.kind !== 'var') {
                this.scopeManager.nestForScope(node);
            }
            this.visitChildren(node);
            this.close(node);
        }
        FunctionDeclaration(node) {
            this.visitFunction(node);
        }
        FunctionExpression(node) {
            this.visitFunction(node);
        }
        Identifier(node) {
            this.currentScope().referenceValue(node);
            this.visitType(node.typeAnnotation);
        }
        ImportDeclaration(node) {
            assert_1.assert(this.scopeManager.isES6() && this.scopeManager.isModule(), 'ImportDeclaration should appear when the mode is ES6 and in the module context.');
            ImportVisitor_1.ImportVisitor.visit(this, node);
        }
        JSXAttribute(node) {
            this.visit(node.value);
        }
        JSXClosingElement() {
            // should not be counted as a reference
        }
        JSXFragment(node) {
            this.referenceJsxPragma();
            this.referenceJsxFragment();
            this.visitChildren(node);
        }
        JSXIdentifier(node) {
            this.currentScope().referenceValue(node);
        }
        JSXMemberExpression(node) {
            this.visit(node.object);
            // we don't ever reference the property as it's always going to be a property on the thing
        }
        JSXOpeningElement(node) {
            this.referenceJsxPragma();
            if (node.name.type === types_1.AST_NODE_TYPES.JSXIdentifier) {
                if (node.name.name[0].toUpperCase() === node.name.name[0]) {
                    // lower cased component names are always treated as "intrinsic" names, and are converted to a string,
                    // not a variable by JSX transforms:
                    // <div /> => React.createElement("div", null)
                    this.visit(node.name);
                }
            }
            else {
                this.visit(node.name);
            }
            this.visitType(node.typeParameters);
            for (const attr of node.attributes) {
                this.visit(attr);
            }
        }
        LabeledStatement(node) {
            this.visit(node.body);
        }
        MemberExpression(node) {
            this.visit(node.object);
            if (node.computed) {
                this.visit(node.property);
            }
        }
        MetaProperty() {
            // meta properties all builtin globals
        }
        NewExpression(node) {
            this.visitChildren(node, ['typeParameters']);
            this.visitType(node.typeParameters);
        }
        Program(node) {
            const globalScope = this.scopeManager.nestGlobalScope(node);
            this.populateGlobalsFromLib(globalScope);
            if (this.scopeManager.isGlobalReturn()) {
                // Force strictness of GlobalScope to false when using node.js scope.
                this.currentScope().isStrict = false;
                this.scopeManager.nestFunctionScope(node, false);
            }
            if (this.scopeManager.isES6() && this.scopeManager.isModule()) {
                this.scopeManager.nestModuleScope(node);
            }
            if (this.scopeManager.isStrictModeSupported() &&
                this.scopeManager.isImpliedStrict()) {
                this.currentScope().isStrict = true;
            }
            this.visitChildren(node);
            this.close(node);
        }
        Property(node) {
            this.visitProperty(node);
        }
        SwitchStatement(node) {
            this.visit(node.discriminant);
            if (this.scopeManager.isES6()) {
                this.scopeManager.nestSwitchScope(node);
            }
            for (const switchCase of node.cases) {
                this.visit(switchCase);
            }
            this.close(node);
        }
        TaggedTemplateExpression(node) {
            this.visit(node.tag);
            this.visit(node.quasi);
            this.visitType(node.typeParameters);
        }
        TSAsExpression(node) {
            this.visitTypeAssertion(node);
        }
        TSDeclareFunction(node) {
            this.visitFunction(node);
        }
        TSImportEqualsDeclaration(node) {
            this.currentScope().defineIdentifier(node.id, new definition_1.ImportBindingDefinition(node.id, node, node));
            if (node.moduleReference.type === types_1.AST_NODE_TYPES.TSQualifiedName) {
                this.visit(node.moduleReference.left);
            }
            else {
                this.visit(node.moduleReference);
            }
        }
        TSEmptyBodyFunctionExpression(node) {
            this.visitFunction(node);
        }
        TSEnumDeclaration(node) {
            this.currentScope().defineIdentifier(node.id, new definition_1.TSEnumNameDefinition(node.id, node));
            // enum members can be referenced within the enum body
            this.scopeManager.nestTSEnumScope(node);
            // define the enum name again inside the new enum scope
            // references to the enum should not resolve directly to the enum
            this.currentScope().defineIdentifier(node.id, new definition_1.TSEnumNameDefinition(node.id, node));
            for (const member of node.members) {
                // TS resolves literal named members to be actual names
                // enum Foo {
                //   'a' = 1,
                //   b = a, // this references the 'a' member
                // }
                if (member.id.type === types_1.AST_NODE_TYPES.Literal &&
                    typeof member.id.value === 'string') {
                    const name = member.id;
                    this.currentScope().defineLiteralIdentifier(name, new definition_1.TSEnumMemberDefinition(name, member));
                }
                else if (!member.computed &&
                    member.id.type === types_1.AST_NODE_TYPES.Identifier) {
                    this.currentScope().defineIdentifier(member.id, new definition_1.TSEnumMemberDefinition(member.id, member));
                }
                this.visit(member.initializer);
            }
            this.close(node);
        }
        TSInterfaceDeclaration(node) {
            this.visitType(node);
        }
        TSModuleDeclaration(node) {
            if (node.id.type === types_1.AST_NODE_TYPES.Identifier && !node.global) {
                this.currentScope().defineIdentifier(node.id, new definition_1.TSModuleNameDefinition(node.id, node));
            }
            this.scopeManager.nestTSModuleScope(node);
            this.visit(node.body);
            this.close(node);
        }
        TSTypeAliasDeclaration(node) {
            this.visitType(node);
        }
        TSTypeAssertion(node) {
            this.visitTypeAssertion(node);
        }
        UpdateExpression(node) {
            if (PatternVisitor_1.PatternVisitor.isPattern(node.argument)) {
                this.visitPattern(node.argument, pattern => {
                    this.currentScope().referenceValue(pattern, Reference_1.ReferenceFlag.ReadWrite, null);
                });
            }
            else {
                this.visitChildren(node);
            }
        }
        VariableDeclaration(node) {
            const variableTargetScope = node.kind === 'var'
                ? this.currentScope().variableScope
                : this.currentScope();
            for (const decl of node.declarations) {
                const init = decl.init;
                this.visitPattern(decl.id, (pattern, info) => {
                    variableTargetScope.defineIdentifier(pattern, new definition_1.VariableDefinition(pattern, decl, node));
                    this.referencingDefaultValue(pattern, info.assignments, null, true);
                    if (init) {
                        this.currentScope().referenceValue(pattern, Reference_1.ReferenceFlag.Write, init, null, true);
                    }
                }, { processRightHandNodes: true });
                if (decl.init) {
                    this.visit(decl.init);
                }
                if ('typeAnnotation' in decl.id) {
                    this.visitType(decl.id.typeAnnotation);
                }
            }
        }
        WithStatement(node) {
            this.visit(node.object);
            // Then nest scope for WithStatement.
            this.scopeManager.nestWithScope(node);
            this.visit(node.body);
            this.close(node);
        }
    }
    exports.Referencer = Referencer;
    _Referencer_jsxPragma = new WeakMap(), _Referencer_jsxFragmentName = new WeakMap(), _Referencer_hasReferencedJsxFactory = new WeakMap(), _Referencer_hasReferencedJsxFragmentFactory = new WeakMap(), _Referencer_lib = new WeakMap(), _Referencer_emitDecoratorMetadata = new WeakMap();
},
/* @typescript-eslint/scope-manager/dist/referencer/ClassVisitor.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _ClassVisitor_classNode, _ClassVisitor_referencer, _ClassVisitor_emitDecoratorMetadata;
    __esModule();
    const types_1 = require(740) /* @typescript-eslint/types */;
    const definition_1 = require(776) /* ../definition */;
    const TypeVisitor_1 = require(792) /* ./TypeVisitor */;
    const Visitor_1 = require(793) /* ./Visitor */;
    class ClassVisitor extends Visitor_1.Visitor {
        constructor(referencer, node, emitDecoratorMetadata) {
            super(referencer);
            _ClassVisitor_classNode.set(this, void 0);
            _ClassVisitor_referencer.set(this, void 0);
            _ClassVisitor_emitDecoratorMetadata.set(this, void 0);
            __classPrivateFieldSet(this, _ClassVisitor_referencer, referencer, "f");
            __classPrivateFieldSet(this, _ClassVisitor_classNode, node, "f");
            __classPrivateFieldSet(this, _ClassVisitor_emitDecoratorMetadata, emitDecoratorMetadata, "f");
        }
        static visit(referencer, node, emitDecoratorMetadata) {
            const classVisitor = new ClassVisitor(referencer, node, emitDecoratorMetadata);
            classVisitor.visitClass(node);
        }
        visit(node) {
            // make sure we only handle the nodes we are designed to handle
            if (node && node.type in this) {
                super.visit(node);
            }
            else {
                __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(node);
            }
        }
        ///////////////////
        // Visit helpers //
        ///////////////////
        visitClass(node) {
            var _a, _b;
            if (node.type === types_1.AST_NODE_TYPES.ClassDeclaration && node.id) {
                __classPrivateFieldGet(this, _ClassVisitor_referencer, "f")
                    .currentScope()
                    .defineIdentifier(node.id, new definition_1.ClassNameDefinition(node.id, node));
            }
            (_a = node.decorators) === null || _a === void 0 ? void 0 : _a.forEach(d => __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(d));
            __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").scopeManager.nestClassScope(node);
            if (node.id) {
                // define the class name again inside the new scope
                // references to the class should not resolve directly to the parent class
                __classPrivateFieldGet(this, _ClassVisitor_referencer, "f")
                    .currentScope()
                    .defineIdentifier(node.id, new definition_1.ClassNameDefinition(node.id, node));
            }
            __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(node.superClass);
            // visit the type param declarations
            this.visitType(node.typeParameters);
            // then the usages
            this.visitType(node.superTypeParameters);
            (_b = node.implements) === null || _b === void 0 ? void 0 : _b.forEach(imp => this.visitType(imp));
            this.visit(node.body);
            __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").close(node);
        }
        visitClassProperty(node) {
            this.visitProperty(node);
            /**
             * class A {
             *   @meta     // <--- check this
             *   foo: Type;
             * }
             */
            this.visitMetadataType(node.typeAnnotation, !!node.decorators);
        }
        visitFunctionParameterTypeAnnotation(node, withDecorators) {
            if ('typeAnnotation' in node) {
                this.visitMetadataType(node.typeAnnotation, withDecorators);
            }
            else if (node.type === types_1.AST_NODE_TYPES.AssignmentPattern) {
                this.visitMetadataType(node.left.typeAnnotation, withDecorators);
            }
            else if (node.type === types_1.AST_NODE_TYPES.TSParameterProperty) {
                this.visitFunctionParameterTypeAnnotation(node.parameter, withDecorators);
            }
        }
        visitMethodFunction(node, methodNode) {
            var _a, _b;
            if (node.id) {
                // FunctionExpression with name creates its special scope;
                // FunctionExpressionNameScope.
                __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").scopeManager.nestFunctionExpressionNameScope(node);
            }
            // Consider this function is in the MethodDefinition.
            __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").scopeManager.nestFunctionScope(node, true);
            /**
             * class A {
             *   @meta     // <--- check this
             *   foo(a: Type) {}
             *
             *   @meta     // <--- check this
             *   foo(): Type {}
             * }
             */
            let withMethodDecorators = !!methodNode.decorators;
            /**
             * class A {
             *   foo(
             *     @meta    // <--- check this
             *     a: Type
             *   ) {}
             *
             *   set foo(
             *     @meta    // <--- EXCEPT this. TS do nothing for this
             *     a: Type
             *   ) {}
             * }
             */
            withMethodDecorators =
                withMethodDecorators ||
                    (methodNode.kind !== 'set' &&
                        node.params.some(param => param.decorators));
            if (!withMethodDecorators && methodNode.kind === 'set') {
                const keyName = getLiteralMethodKeyName(methodNode);
                /**
                 * class A {
                 *   @meta      // <--- check this
                 *   get a() {}
                 *   set ['a'](v: Type) {}
                 * }
                 */
                if (keyName !== null &&
                    ((_a = __classPrivateFieldGet(this, _ClassVisitor_classNode, "f").body.body.find((node) => node !== methodNode &&
                        node.type === types_1.AST_NODE_TYPES.MethodDefinition &&
                        // Node must both be static or not
                        node.static === methodNode.static &&
                        getLiteralMethodKeyName(node) === keyName)) === null || _a === void 0 ? void 0 : _a.decorators)) {
                    withMethodDecorators = true;
                }
            }
            /**
             * @meta      // <--- check this
             * class A {
             *   constructor(a: Type) {}
             * }
             */
            if (!withMethodDecorators &&
                methodNode.kind === 'constructor' &&
                __classPrivateFieldGet(this, _ClassVisitor_classNode, "f").decorators) {
                withMethodDecorators = true;
            }
            // Process parameter declarations.
            for (const param of node.params) {
                this.visitPattern(param, (pattern, info) => {
                    __classPrivateFieldGet(this, _ClassVisitor_referencer, "f")
                        .currentScope()
                        .defineIdentifier(pattern, new definition_1.ParameterDefinition(pattern, node, info.rest));
                    __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").referencingDefaultValue(pattern, info.assignments, null, true);
                }, { processRightHandNodes: true });
                this.visitFunctionParameterTypeAnnotation(param, withMethodDecorators);
                (_b = param.decorators) === null || _b === void 0 ? void 0 : _b.forEach(d => this.visit(d));
            }
            this.visitMetadataType(node.returnType, withMethodDecorators);
            this.visitType(node.typeParameters);
            // In TypeScript there are a number of function-like constructs which have no body,
            // so check it exists before traversing
            if (node.body) {
                // Skip BlockStatement to prevent creating BlockStatement scope.
                if (node.body.type === types_1.AST_NODE_TYPES.BlockStatement) {
                    __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visitChildren(node.body);
                }
                else {
                    __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(node.body);
                }
            }
            __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").close(node);
        }
        visitProperty(node) {
            var _a;
            if (node.computed) {
                __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(node.key);
            }
            __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(node.value);
            if ('decorators' in node) {
                (_a = node.decorators) === null || _a === void 0 ? void 0 : _a.forEach(d => __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(d));
            }
        }
        visitMethod(node) {
            var _a;
            if (node.computed) {
                __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(node.key);
            }
            if (node.value.type === types_1.AST_NODE_TYPES.FunctionExpression) {
                this.visitMethodFunction(node.value, node);
            }
            else {
                __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(node.value);
            }
            if ('decorators' in node) {
                (_a = node.decorators) === null || _a === void 0 ? void 0 : _a.forEach(d => __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(d));
            }
        }
        visitType(node) {
            if (!node) {
                return;
            }
            TypeVisitor_1.TypeVisitor.visit(__classPrivateFieldGet(this, _ClassVisitor_referencer, "f"), node);
        }
        visitMetadataType(node, withDecorators) {
            if (!node) {
                return;
            }
            // emit decorators metadata only work for TSTypeReference in ClassDeclaration
            if (__classPrivateFieldGet(this, _ClassVisitor_classNode, "f").type === types_1.AST_NODE_TYPES.ClassDeclaration &&
                !__classPrivateFieldGet(this, _ClassVisitor_classNode, "f").declare &&
                node.typeAnnotation.type === types_1.AST_NODE_TYPES.TSTypeReference &&
                __classPrivateFieldGet(this, _ClassVisitor_emitDecoratorMetadata, "f")) {
                let identifier;
                if (node.typeAnnotation.typeName.type === types_1.AST_NODE_TYPES.TSQualifiedName) {
                    let iter = node.typeAnnotation.typeName;
                    while (iter.left.type === types_1.AST_NODE_TYPES.TSQualifiedName) {
                        iter = iter.left;
                    }
                    identifier = iter.left;
                }
                else {
                    identifier = node.typeAnnotation.typeName;
                }
                if (withDecorators) {
                    __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").currentScope().referenceDualValueType(identifier);
                    if (node.typeAnnotation.typeParameters) {
                        this.visitType(node.typeAnnotation.typeParameters);
                    }
                    // everything is handled now
                    return;
                }
            }
            this.visitType(node);
        }
        /////////////////////
        // Visit selectors //
        /////////////////////
        ClassBody(node) {
            // this is here on purpose so that this visitor explicitly declares visitors
            // for all nodes it cares about (see the instance visit method above)
            this.visitChildren(node);
        }
        ClassProperty(node) {
            this.visitClassProperty(node);
        }
        MethodDefinition(node) {
            this.visitMethod(node);
        }
        TSAbstractClassProperty(node) {
            this.visitClassProperty(node);
        }
        TSAbstractMethodDefinition(node) {
            this.visitProperty(node);
        }
        Identifier(node) {
            __classPrivateFieldGet(this, _ClassVisitor_referencer, "f").visit(node);
        }
    }
    exports.ClassVisitor = ClassVisitor;
    _ClassVisitor_classNode = new WeakMap(), _ClassVisitor_referencer = new WeakMap(), _ClassVisitor_emitDecoratorMetadata = new WeakMap();
    /**
     * Only if key is one of [identifier, string, number], ts will combine metadata of accessors .
     * class A {
     *   get a() {}
     *   set ['a'](v: Type) {}
     *
     *   get [1]() {}
     *   set [1](v: Type) {}
     *
     *   // Following won't be combined
     *   get [key]() {}
     *   set [key](v: Type) {}
     *
     *   get [true]() {}
     *   set [true](v: Type) {}
     *
     *   get ['a'+'b']() {}
     *   set ['a'+'b']() {}
     * }
     */
    function getLiteralMethodKeyName(node) {
        if (node.computed && node.key.type === types_1.AST_NODE_TYPES.Literal) {
            if (typeof node.key.value === 'string' ||
                typeof node.key.value === 'number') {
                return node.key.value;
            }
        }
        else if (!node.computed && node.key.type === types_1.AST_NODE_TYPES.Identifier) {
            return node.key.name;
        }
        return null;
    }
},
/* @typescript-eslint/scope-manager/dist/definition/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    __exportStar(require(777) /* ./CatchClauseDefinition */, exports);
    __exportStar(require(781) /* ./ClassNameDefinition */, exports);
    __exportStar(require(782) /* ./Definition */, exports);
    __exportStar(require(778) /* ./DefinitionType */, exports);
    __exportStar(require(783) /* ./FunctionNameDefinition */, exports);
    __exportStar(require(784) /* ./ImplicitGlobalVariableDefinition */, exports);
    __exportStar(require(785) /* ./ImportBindingDefinition */, exports);
    __exportStar(require(786) /* ./ParameterDefinition */, exports);
    __exportStar(require(787) /* ./TSEnumMemberDefinition */, exports);
    __exportStar(require(788) /* ./TSEnumNameDefinition */, exports);
    __exportStar(require(789) /* ./TSModuleNameDefinition */, exports);
    __exportStar(require(790) /* ./TypeDefinition */, exports);
    __exportStar(require(791) /* ./VariableDefinition */, exports);
},
/* @typescript-eslint/scope-manager/dist/definition/CatchClauseDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class CatchClauseDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node) {
            super(DefinitionType_1.DefinitionType.CatchClause, name, node, null);
            this.isTypeDefinition = false;
            this.isVariableDefinition = true;
        }
    }
    exports.CatchClauseDefinition = CatchClauseDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/DefinitionType.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var DefinitionType;
    (function (DefinitionType) {
        // eslint-disable-next-line @typescript-eslint/internal/prefer-ast-types-enum
        DefinitionType["CatchClause"] = "CatchClause";
        DefinitionType["ClassName"] = "ClassName";
        DefinitionType["FunctionName"] = "FunctionName";
        DefinitionType["ImplicitGlobalVariable"] = "ImplicitGlobalVariable";
        DefinitionType["ImportBinding"] = "ImportBinding";
        DefinitionType["Parameter"] = "Parameter";
        DefinitionType["TSEnumName"] = "TSEnumName";
        DefinitionType["TSEnumMember"] = "TSEnumMemberName";
        DefinitionType["TSModuleName"] = "TSModuleName";
        DefinitionType["Type"] = "Type";
        DefinitionType["Variable"] = "Variable";
    })(DefinitionType || (DefinitionType = {}));
    exports.DefinitionType = DefinitionType;
},
/* @typescript-eslint/scope-manager/dist/definition/DefinitionBase.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ID_1 = require(780) /* ../ID */;
    const generator = ID_1.createIdGenerator();
    class DefinitionBase {
        constructor(type, name, node, parent) {
            /**
             * A unique ID for this instance - primarily used to help debugging and testing
             */
            this.$id = generator();
            this.type = type;
            this.name = name;
            this.node = node;
            this.parent = parent;
        }
    }
    exports.DefinitionBase = DefinitionBase;
},
/* @typescript-eslint/scope-manager/dist/ID.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ID_CACHE = new Map();
    let NEXT_KEY = 0;
    function createIdGenerator() {
        const key = (NEXT_KEY += 1);
        ID_CACHE.set(key, 0);
        return () => {
            var _a;
            const current = (_a = ID_CACHE.get(key)) !== null && _a !== void 0 ? _a : 0;
            const next = current + 1;
            ID_CACHE.set(key, next);
            return next;
        };
    }
    exports.createIdGenerator = createIdGenerator;
    function resetIds() {
        ID_CACHE.clear();
    }
    exports.resetIds = resetIds;
},
/* @typescript-eslint/scope-manager/dist/definition/ClassNameDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class ClassNameDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node) {
            super(DefinitionType_1.DefinitionType.ClassName, name, node, null);
            this.isTypeDefinition = true;
            this.isVariableDefinition = true;
        }
    }
    exports.ClassNameDefinition = ClassNameDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/Definition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/scope-manager/dist/definition/FunctionNameDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class FunctionNameDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node) {
            super(DefinitionType_1.DefinitionType.FunctionName, name, node, null);
            this.isTypeDefinition = false;
            this.isVariableDefinition = true;
        }
    }
    exports.FunctionNameDefinition = FunctionNameDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/ImplicitGlobalVariableDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class ImplicitGlobalVariableDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node) {
            super(DefinitionType_1.DefinitionType.ImplicitGlobalVariable, name, node, null);
            this.isTypeDefinition = false;
            this.isVariableDefinition = true;
        }
    }
    exports.ImplicitGlobalVariableDefinition = ImplicitGlobalVariableDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/ImportBindingDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class ImportBindingDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node, decl) {
            super(DefinitionType_1.DefinitionType.ImportBinding, name, node, decl);
            this.isTypeDefinition = true;
            this.isVariableDefinition = true;
        }
    }
    exports.ImportBindingDefinition = ImportBindingDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/ParameterDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class ParameterDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node, rest) {
            super(DefinitionType_1.DefinitionType.Parameter, name, node, null);
            this.isTypeDefinition = false;
            this.isVariableDefinition = true;
            this.rest = rest;
        }
    }
    exports.ParameterDefinition = ParameterDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/TSEnumMemberDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class TSEnumMemberDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node) {
            super(DefinitionType_1.DefinitionType.TSEnumMember, name, node, null);
            this.isTypeDefinition = true;
            this.isVariableDefinition = true;
        }
    }
    exports.TSEnumMemberDefinition = TSEnumMemberDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/TSEnumNameDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class TSEnumNameDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node) {
            super(DefinitionType_1.DefinitionType.TSEnumName, name, node, null);
            this.isTypeDefinition = true;
            this.isVariableDefinition = true;
        }
    }
    exports.TSEnumNameDefinition = TSEnumNameDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/TSModuleNameDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class TSModuleNameDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node) {
            super(DefinitionType_1.DefinitionType.TSModuleName, name, node, null);
            this.isTypeDefinition = true;
            this.isVariableDefinition = true;
        }
    }
    exports.TSModuleNameDefinition = TSModuleNameDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/TypeDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class TypeDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node) {
            super(DefinitionType_1.DefinitionType.Type, name, node, null);
            this.isTypeDefinition = true;
            this.isVariableDefinition = false;
        }
    }
    exports.TypeDefinition = TypeDefinition;
},
/* @typescript-eslint/scope-manager/dist/definition/VariableDefinition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const DefinitionType_1 = require(778) /* ./DefinitionType */;
    const DefinitionBase_1 = require(779) /* ./DefinitionBase */;
    class VariableDefinition extends DefinitionBase_1.DefinitionBase {
        constructor(name, node, decl) {
            super(DefinitionType_1.DefinitionType.Variable, name, node, decl);
            this.isTypeDefinition = false;
            this.isVariableDefinition = true;
        }
    }
    exports.VariableDefinition = VariableDefinition;
},
/* @typescript-eslint/scope-manager/dist/referencer/TypeVisitor.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _TypeVisitor_referencer;
    __esModule();
    const types_1 = require(740) /* @typescript-eslint/types */;
    const Visitor_1 = require(793) /* ./Visitor */;
    const definition_1 = require(776) /* ../definition */;
    const scope_1 = require(796) /* ../scope */;
    class TypeVisitor extends Visitor_1.Visitor {
        constructor(referencer) {
            super(referencer);
            _TypeVisitor_referencer.set(this, void 0);
            __classPrivateFieldSet(this, _TypeVisitor_referencer, referencer, "f");
        }
        static visit(referencer, node) {
            const typeReferencer = new TypeVisitor(referencer);
            typeReferencer.visit(node);
        }
        ///////////////////
        // Visit helpers //
        ///////////////////
        visitFunctionType(node) {
            // arguments and type parameters can only be referenced from within the function
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").scopeManager.nestFunctionTypeScope(node);
            this.visit(node.typeParameters);
            for (const param of node.params) {
                let didVisitAnnotation = false;
                this.visitPattern(param, (pattern, info) => {
                    // a parameter name creates a value type variable which can be referenced later via typeof arg
                    __classPrivateFieldGet(this, _TypeVisitor_referencer, "f")
                        .currentScope()
                        .defineIdentifier(pattern, new definition_1.ParameterDefinition(pattern, node, info.rest));
                    if (pattern.typeAnnotation) {
                        this.visit(pattern.typeAnnotation);
                        didVisitAnnotation = true;
                    }
                });
                // there are a few special cases where the type annotation is owned by the parameter, not the pattern
                if (!didVisitAnnotation && 'typeAnnotation' in param) {
                    this.visit(param.typeAnnotation);
                }
            }
            this.visit(node.returnType);
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").close(node);
        }
        visitPropertyKey(node) {
            if (!node.computed) {
                return;
            }
            // computed members are treated as value references, and TS expects they have a literal type
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").visit(node.key);
        }
        /////////////////////
        // Visit selectors //
        /////////////////////
        Identifier(node) {
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").currentScope().referenceType(node);
        }
        MemberExpression(node) {
            this.visit(node.object);
            // don't visit the property
        }
        TSCallSignatureDeclaration(node) {
            this.visitFunctionType(node);
        }
        TSConditionalType(node) {
            // conditional types can define inferred type parameters
            // which are only accessible from inside the conditional parameter
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").scopeManager.nestConditionalTypeScope(node);
            // type parameters inferred in the condition clause are not accessible within the false branch
            this.visitChildren(node, ['falseType']);
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").close(node);
            this.visit(node.falseType);
        }
        TSConstructorType(node) {
            this.visitFunctionType(node);
        }
        TSConstructSignatureDeclaration(node) {
            this.visitFunctionType(node);
        }
        TSFunctionType(node) {
            this.visitFunctionType(node);
        }
        TSImportType(node) {
            // the TS parser allows any type to be the parameter, but it's a syntax error - so we can ignore it
            this.visit(node.typeParameters);
            // the qualifier is just part of a standard EntityName, so it should not be visited
        }
        TSIndexSignature(node) {
            for (const param of node.parameters) {
                if (param.type === types_1.AST_NODE_TYPES.Identifier) {
                    this.visit(param.typeAnnotation);
                }
            }
            this.visit(node.typeAnnotation);
        }
        TSInferType(node) {
            const typeParameter = node.typeParameter;
            let scope = __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").currentScope();
            /*
            In cases where there is a sub-type scope created within a conditional type, then the generic should be defined in the
            conditional type's scope, not the child type scope.
            If we define it within the child type's scope then it won't be able to be referenced outside the child type
            */
            if (scope.type === scope_1.ScopeType.functionType ||
                scope.type === scope_1.ScopeType.mappedType) {
                // search up the scope tree to figure out if we're in a nested type scope
                let currentScope = scope.upper;
                while (currentScope) {
                    if (currentScope.type === scope_1.ScopeType.functionType ||
                        currentScope.type === scope_1.ScopeType.mappedType) {
                        // ensure valid type parents only
                        currentScope = currentScope.upper;
                        continue;
                    }
                    if (currentScope.type === scope_1.ScopeType.conditionalType) {
                        scope = currentScope;
                        break;
                    }
                    break;
                }
            }
            scope.defineIdentifier(typeParameter.name, new definition_1.TypeDefinition(typeParameter.name, typeParameter));
        }
        TSInterfaceDeclaration(node) {
            var _a, _b;
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f")
                .currentScope()
                .defineIdentifier(node.id, new definition_1.TypeDefinition(node.id, node));
            if (node.typeParameters) {
                // type parameters cannot be referenced from outside their current scope
                __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").scopeManager.nestTypeScope(node);
                this.visit(node.typeParameters);
            }
            (_a = node.extends) === null || _a === void 0 ? void 0 : _a.forEach(this.visit, this);
            (_b = node.implements) === null || _b === void 0 ? void 0 : _b.forEach(this.visit, this);
            this.visit(node.body);
            if (node.typeParameters) {
                __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").close(node);
            }
        }
        TSMappedType(node) {
            // mapped types key can only be referenced within their return value
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").scopeManager.nestMappedTypeScope(node);
            this.visitChildren(node);
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").close(node);
        }
        TSMethodSignature(node) {
            this.visitPropertyKey(node);
            this.visitFunctionType(node);
        }
        TSNamedTupleMember(node) {
            this.visit(node.elementType);
            // we don't visit the label as the label only exists for the purposes of documentation
        }
        TSPropertySignature(node) {
            this.visitPropertyKey(node);
            this.visit(node.typeAnnotation);
        }
        TSQualifiedName(node) {
            this.visit(node.left);
            // we don't visit the right as it a name on the thing, not a name to reference
        }
        TSTypeAliasDeclaration(node) {
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f")
                .currentScope()
                .defineIdentifier(node.id, new definition_1.TypeDefinition(node.id, node));
            if (node.typeParameters) {
                // type parameters cannot be referenced from outside their current scope
                __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").scopeManager.nestTypeScope(node);
                this.visit(node.typeParameters);
            }
            this.visit(node.typeAnnotation);
            if (node.typeParameters) {
                __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").close(node);
            }
        }
        TSTypeParameter(node) {
            __classPrivateFieldGet(this, _TypeVisitor_referencer, "f")
                .currentScope()
                .defineIdentifier(node.name, new definition_1.TypeDefinition(node.name, node));
            this.visit(node.constraint);
            this.visit(node.default);
        }
        TSTypePredicate(node) {
            if (node.parameterName.type !== types_1.AST_NODE_TYPES.TSThisType) {
                __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").currentScope().referenceValue(node.parameterName);
            }
            this.visit(node.typeAnnotation);
        }
        // a type query `typeof foo` is a special case that references a _non-type_ variable,
        TSTypeQuery(node) {
            if (node.exprName.type === types_1.AST_NODE_TYPES.Identifier) {
                __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").currentScope().referenceValue(node.exprName);
            }
            else {
                let expr = node.exprName.left;
                while (expr.type !== types_1.AST_NODE_TYPES.Identifier) {
                    expr = expr.left;
                }
                __classPrivateFieldGet(this, _TypeVisitor_referencer, "f").currentScope().referenceValue(expr);
            }
        }
        TSTypeAnnotation(node) {
            // check
            this.visitChildren(node);
        }
    }
    exports.TypeVisitor = TypeVisitor;
    _TypeVisitor_referencer = new WeakMap();
},
/* @typescript-eslint/scope-manager/dist/referencer/Visitor.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var _Visitor_options;
    __esModule();
    const VisitorBase_1 = require(794) /* ./VisitorBase */;
    __esExport("VisitorBase", VisitorBase_1.VisitorBase);
    const PatternVisitor_1 = require(795) /* ./PatternVisitor */;
    class Visitor extends VisitorBase_1.VisitorBase {
        constructor(optionsOrVisitor) {
            super(optionsOrVisitor instanceof Visitor
                ? __classPrivateFieldGet(optionsOrVisitor, _Visitor_options, "f")
                : optionsOrVisitor);
            _Visitor_options.set(this, void 0);
            __classPrivateFieldSet(this, _Visitor_options, optionsOrVisitor instanceof Visitor
                ? __classPrivateFieldGet(optionsOrVisitor, _Visitor_options, "f")
                : optionsOrVisitor, "f");
        }
        visitPattern(node, callback, options = { processRightHandNodes: false }) {
            // Call the callback at left hand identifier nodes, and Collect right hand nodes.
            const visitor = new PatternVisitor_1.PatternVisitor(__classPrivateFieldGet(this, _Visitor_options, "f"), node, callback);
            visitor.visit(node);
            // Process the right hand nodes recursively.
            if (options.processRightHandNodes) {
                visitor.rightHandNodes.forEach(this.visit, this);
            }
        }
    }
    exports.Visitor = Visitor;
    _Visitor_options = new WeakMap();
},
/* @typescript-eslint/scope-manager/dist/referencer/VisitorBase.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _VisitorBase_childVisitorKeys, _VisitorBase_visitChildrenEvenIfSelectorExists;
    __esModule();
    const visitor_keys_1 = require(770) /* @typescript-eslint/visitor-keys */;
    function isObject(obj) {
        return typeof obj === 'object' && obj != null;
    }
    function isNode(node) {
        return isObject(node) && typeof node.type === 'string';
    }
    class VisitorBase {
        constructor(options) {
            var _a, _b;
            _VisitorBase_childVisitorKeys.set(this, void 0);
            _VisitorBase_visitChildrenEvenIfSelectorExists.set(this, void 0);
            __classPrivateFieldSet(this, _VisitorBase_childVisitorKeys, (_a = options.childVisitorKeys) !== null && _a !== void 0 ? _a : visitor_keys_1.visitorKeys, "f");
            __classPrivateFieldSet(this, _VisitorBase_visitChildrenEvenIfSelectorExists, (_b = options.visitChildrenEvenIfSelectorExists) !== null && _b !== void 0 ? _b : false, "f");
        }
        /**
         * Default method for visiting children.
         * @param node the node whose children should be visited
         * @param exclude a list of keys to not visit
         */
        visitChildren(node, excludeArr = []) {
            var _a;
            if (node == null || node.type == null) {
                return;
            }
            const exclude = new Set(excludeArr.concat(['parent']));
            const children = (_a = __classPrivateFieldGet(this, _VisitorBase_childVisitorKeys, "f")[node.type]) !== null && _a !== void 0 ? _a : Object.keys(node);
            for (const key of children) {
                if (exclude.has(key)) {
                    continue;
                }
                const child = node[key];
                if (!child) {
                    continue;
                }
                if (Array.isArray(child)) {
                    for (const subChild of child) {
                        if (isNode(subChild)) {
                            this.visit(subChild);
                        }
                    }
                }
                else if (isNode(child)) {
                    this.visit(child);
                }
            }
        }
        /**
         * Dispatching node.
         */
        visit(node) {
            if (node == null || node.type == null) {
                return;
            }
            const visitor = this[node.type];
            if (visitor) {
                visitor.call(this, node);
                if (!__classPrivateFieldGet(this, _VisitorBase_visitChildrenEvenIfSelectorExists, "f")) {
                    return;
                }
            }
            this.visitChildren(node);
        }
    }
    exports.VisitorBase = VisitorBase;
    _VisitorBase_childVisitorKeys = new WeakMap(), _VisitorBase_visitChildrenEvenIfSelectorExists = new WeakMap();
},
/* @typescript-eslint/scope-manager/dist/referencer/PatternVisitor.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _PatternVisitor_rootPattern, _PatternVisitor_callback, _PatternVisitor_assignments, _PatternVisitor_restElements;
    __esModule();
    const types_1 = require(740) /* @typescript-eslint/types */;
    const Visitor_1 = require(793) /* ./Visitor */;
    class PatternVisitor extends Visitor_1.VisitorBase {
        constructor(options, rootPattern, callback) {
            super(options);
            _PatternVisitor_rootPattern.set(this, void 0);
            _PatternVisitor_callback.set(this, void 0);
            _PatternVisitor_assignments.set(this, []);
            this.rightHandNodes = [];
            _PatternVisitor_restElements.set(this, []);
            __classPrivateFieldSet(this, _PatternVisitor_rootPattern, rootPattern, "f");
            __classPrivateFieldSet(this, _PatternVisitor_callback, callback, "f");
        }
        static isPattern(node) {
            const nodeType = node.type;
            return (nodeType === types_1.AST_NODE_TYPES.Identifier ||
                nodeType === types_1.AST_NODE_TYPES.ObjectPattern ||
                nodeType === types_1.AST_NODE_TYPES.ArrayPattern ||
                nodeType === types_1.AST_NODE_TYPES.SpreadElement ||
                nodeType === types_1.AST_NODE_TYPES.RestElement ||
                nodeType === types_1.AST_NODE_TYPES.AssignmentPattern);
        }
        ArrayExpression(node) {
            node.elements.forEach(this.visit, this);
        }
        ArrayPattern(pattern) {
            for (const element of pattern.elements) {
                this.visit(element);
            }
        }
        AssignmentExpression(node) {
            __classPrivateFieldGet(this, _PatternVisitor_assignments, "f").push(node);
            this.visit(node.left);
            this.rightHandNodes.push(node.right);
            __classPrivateFieldGet(this, _PatternVisitor_assignments, "f").pop();
        }
        AssignmentPattern(pattern) {
            __classPrivateFieldGet(this, _PatternVisitor_assignments, "f").push(pattern);
            this.visit(pattern.left);
            this.rightHandNodes.push(pattern.right);
            __classPrivateFieldGet(this, _PatternVisitor_assignments, "f").pop();
        }
        CallExpression(node) {
            // arguments are right hand nodes.
            node.arguments.forEach(a => {
                this.rightHandNodes.push(a);
            });
            this.visit(node.callee);
        }
        Decorator() {
            // don't visit any decorators when exploring a pattern
        }
        Identifier(pattern) {
            var _a;
            const lastRestElement = (_a = __classPrivateFieldGet(this, _PatternVisitor_restElements, "f")[__classPrivateFieldGet(this, _PatternVisitor_restElements, "f").length - 1]) !== null && _a !== void 0 ? _a : null;
            __classPrivateFieldGet(this, _PatternVisitor_callback, "f").call(this, pattern, {
                topLevel: pattern === __classPrivateFieldGet(this, _PatternVisitor_rootPattern, "f"),
                rest: lastRestElement !== null &&
                    lastRestElement !== undefined &&
                    lastRestElement.argument === pattern,
                assignments: __classPrivateFieldGet(this, _PatternVisitor_assignments, "f"),
            });
        }
        MemberExpression(node) {
            // Computed property's key is a right hand node.
            if (node.computed) {
                this.rightHandNodes.push(node.property);
            }
            // the object is only read, write to its property.
            this.rightHandNodes.push(node.object);
        }
        Property(property) {
            // Computed property's key is a right hand node.
            if (property.computed) {
                this.rightHandNodes.push(property.key);
            }
            // If it's shorthand, its key is same as its value.
            // If it's shorthand and has its default value, its key is same as its value.left (the value is AssignmentPattern).
            // If it's not shorthand, the name of new variable is its value's.
            this.visit(property.value);
        }
        RestElement(pattern) {
            __classPrivateFieldGet(this, _PatternVisitor_restElements, "f").push(pattern);
            this.visit(pattern.argument);
            __classPrivateFieldGet(this, _PatternVisitor_restElements, "f").pop();
        }
        SpreadElement(node) {
            this.visit(node.argument);
        }
        TSTypeAnnotation() {
            // we don't want to visit types
        }
    }
    exports.PatternVisitor = PatternVisitor;
    _PatternVisitor_rootPattern = new WeakMap(), _PatternVisitor_callback = new WeakMap(), _PatternVisitor_assignments = new WeakMap(), _PatternVisitor_restElements = new WeakMap();
},
/* @typescript-eslint/scope-manager/dist/scope/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    __exportStar(require(797) /* ./BlockScope */, exports);
    __exportStar(require(807) /* ./CatchScope */, exports);
    __exportStar(require(808) /* ./ClassScope */, exports);
    __exportStar(require(809) /* ./ConditionalTypeScope */, exports);
    __exportStar(require(810) /* ./ForScope */, exports);
    __exportStar(require(811) /* ./FunctionExpressionNameScope */, exports);
    __exportStar(require(812) /* ./FunctionScope */, exports);
    __exportStar(require(813) /* ./FunctionTypeScope */, exports);
    __exportStar(require(814) /* ./GlobalScope */, exports);
    __exportStar(require(815) /* ./MappedTypeScope */, exports);
    __exportStar(require(816) /* ./ModuleScope */, exports);
    __exportStar(require(817) /* ./Scope */, exports);
    __exportStar(require(799) /* ./ScopeType */, exports);
    __exportStar(require(818) /* ./SwitchScope */, exports);
    __exportStar(require(819) /* ./TSEnumScope */, exports);
    __exportStar(require(820) /* ./TSModuleScope */, exports);
    __exportStar(require(821) /* ./TypeScope */, exports);
    __exportStar(require(822) /* ./WithScope */, exports);
},
/* @typescript-eslint/scope-manager/dist/scope/BlockScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class BlockScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.block, upperScope, block, false);
        }
    }
    exports.BlockScope = BlockScope;
},
/* @typescript-eslint/scope-manager/dist/scope/ScopeBase.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _ScopeBase_declaredVariables, _ScopeBase_dynamic, _ScopeBase_staticCloseRef, _ScopeBase_dynamicCloseRef, _ScopeBase_globalCloseRef;
    __esModule();
    const types_1 = require(740) /* @typescript-eslint/types */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    const assert_1 = require(800) /* ../assert */;
    const definition_1 = require(776) /* ../definition */;
    const ID_1 = require(780) /* ../ID */;
    const Reference_1 = require(801) /* ../referencer/Reference */;
    const variable_1 = require(802) /* ../variable */;
    /**
     * Test if scope is strict
     */
    function isStrictScope(scope, block, isMethodDefinition) {
        var _a;
        let body;
        // When upper scope is exists and strict, inner scope is also strict.
        if ((_a = scope.upper) === null || _a === void 0 ? void 0 : _a.isStrict) {
            return true;
        }
        if (isMethodDefinition) {
            return true;
        }
        if (scope.type === ScopeType_1.ScopeType.class ||
            scope.type === ScopeType_1.ScopeType.conditionalType ||
            scope.type === ScopeType_1.ScopeType.functionType ||
            scope.type === ScopeType_1.ScopeType.mappedType ||
            scope.type === ScopeType_1.ScopeType.module ||
            scope.type === ScopeType_1.ScopeType.tsEnum ||
            scope.type === ScopeType_1.ScopeType.tsModule ||
            scope.type === ScopeType_1.ScopeType.type) {
            return true;
        }
        if (scope.type === ScopeType_1.ScopeType.block || scope.type === ScopeType_1.ScopeType.switch) {
            return false;
        }
        if (scope.type === ScopeType_1.ScopeType.function) {
            const functionBody = block;
            switch (functionBody.type) {
                case types_1.AST_NODE_TYPES.ArrowFunctionExpression:
                    if (functionBody.body.type !== types_1.AST_NODE_TYPES.BlockStatement) {
                        return false;
                    }
                    body = functionBody.body;
                    break;
                case types_1.AST_NODE_TYPES.Program:
                    body = functionBody;
                    break;
                default:
                    body = functionBody.body;
            }
            if (!body) {
                return false;
            }
        }
        else if (scope.type === ScopeType_1.ScopeType.global) {
            body = block;
        }
        else {
            return false;
        }
        // Search 'use strict' directive.
        for (let i = 0; i < body.body.length; ++i) {
            const stmt = body.body[i];
            if (stmt.type !== types_1.AST_NODE_TYPES.ExpressionStatement) {
                break;
            }
            if (stmt.directive === 'use strict') {
                return true;
            }
            const expr = stmt.expression;
            if (expr.type !== types_1.AST_NODE_TYPES.Literal) {
                break;
            }
            if (expr.raw === '"use strict"' || expr.raw === "'use strict'") {
                return true;
            }
            if (expr.value === 'use strict') {
                return true;
            }
        }
        return false;
    }
    /**
     * Register scope
     */
    function registerScope(scopeManager, scope) {
        scopeManager.scopes.push(scope);
        const scopes = scopeManager.nodeToScope.get(scope.block);
        if (scopes) {
            scopes.push(scope);
        }
        else {
            scopeManager.nodeToScope.set(scope.block, [scope]);
        }
    }
    const generator = ID_1.createIdGenerator();
    const VARIABLE_SCOPE_TYPES = new Set([
        ScopeType_1.ScopeType.global,
        ScopeType_1.ScopeType.function,
        ScopeType_1.ScopeType.module,
        ScopeType_1.ScopeType.tsModule,
    ]);
    class ScopeBase {
        constructor(scopeManager, type, upperScope, block, isMethodDefinition) {
            /**
             * A unique ID for this instance - primarily used to help debugging and testing
             */
            this.$id = generator();
            /**
             * The array of child scopes. This does not include grandchild scopes.
             * @public
             */
            this.childScopes = [];
            /**
             * A map of the variables for each node in this scope.
             * This is map is a pointer to the one in the parent ScopeManager instance
             */
            _ScopeBase_declaredVariables.set(this, void 0);
            /**
             * Generally, through the lexical scoping of JS you can always know which variable an identifier in the source code
             * refers to. There are a few exceptions to this rule. With `global` and `with` scopes you can only decide at runtime
             * which variable a reference refers to.
             * All those scopes are considered "dynamic".
             */
            _ScopeBase_dynamic.set(this, void 0);
            /**
             * Whether this scope is created by a FunctionExpression.
             * @public
             */
            this.functionExpressionScope = false;
            /**
             * List of {@link Reference}s that are left to be resolved (i.e. which
             * need to be linked to the variable they refer to).
             */
            this.leftToResolve = [];
            /**
             * Any variable {@link Reference} found in this scope.
             * This includes occurrences of local variables as well as variables from parent scopes (including the global scope).
             * For local variables this also includes defining occurrences (like in a 'var' statement).
             * In a 'function' scope this does not include the occurrences of the formal parameter in the parameter list.
             * @public
             */
            this.references = [];
            /**
             * The map from variable names to variable objects.
             * @public
             */
            this.set = new Map();
            /**
             * The {@link Reference}s that are not resolved with this scope.
             * @public
             */
            this.through = [];
            /**
             * The scoped {@link Variable}s of this scope.
             * In the case of a 'function' scope this includes the automatic argument `arguments` as its first element, as well
             * as all further formal arguments.
             * This does not include variables which are defined in child scopes.
             * @public
             */
            this.variables = [];
            _ScopeBase_staticCloseRef.set(this, (ref) => {
                const resolve = () => {
                    const name = ref.identifier.name;
                    const variable = this.set.get(name);
                    if (!variable) {
                        return false;
                    }
                    if (!this.isValidResolution(ref, variable)) {
                        return false;
                    }
                    // make sure we don't match a type reference to a value variable
                    const isValidTypeReference = ref.isTypeReference && variable.isTypeVariable;
                    const isValidValueReference = ref.isValueReference && variable.isValueVariable;
                    if (!isValidTypeReference && !isValidValueReference) {
                        return false;
                    }
                    variable.references.push(ref);
                    ref.resolved = variable;
                    return true;
                };
                if (!resolve()) {
                    this.delegateToUpperScope(ref);
                }
            });
            _ScopeBase_dynamicCloseRef.set(this, (ref) => {
                // notify all names are through to global
                let current = this;
                do {
                    current.through.push(ref);
                    current = current.upper;
                } while (current);
            });
            _ScopeBase_globalCloseRef.set(this, (ref, scopeManager) => {
                // let/const/class declarations should be resolved statically.
                // others should be resolved dynamically.
                if (this.shouldStaticallyCloseForGlobal(ref, scopeManager)) {
                    __classPrivateFieldGet(this, _ScopeBase_staticCloseRef, "f").call(this, ref);
                }
                else {
                    __classPrivateFieldGet(this, _ScopeBase_dynamicCloseRef, "f").call(this, ref);
                }
            });
            const upperScopeAsScopeBase = upperScope;
            this.type = type;
            __classPrivateFieldSet(this, _ScopeBase_dynamic, this.type === ScopeType_1.ScopeType.global || this.type === ScopeType_1.ScopeType.with, "f");
            this.block = block;
            this.variableScope = this.isVariableScope()
                ? this
                : upperScopeAsScopeBase.variableScope;
            this.upper = upperScope;
            /**
             * Whether 'use strict' is in effect in this scope.
             * @member {boolean} Scope#isStrict
             */
            this.isStrict = isStrictScope(this, block, isMethodDefinition);
            if (upperScopeAsScopeBase) {
                // this is guaranteed to be correct at runtime
                upperScopeAsScopeBase.childScopes.push(this);
            }
            __classPrivateFieldSet(this, _ScopeBase_declaredVariables, scopeManager.declaredVariables, "f");
            registerScope(scopeManager, this);
        }
        isVariableScope() {
            return VARIABLE_SCOPE_TYPES.has(this.type);
        }
        shouldStaticallyClose() {
            return !__classPrivateFieldGet(this, _ScopeBase_dynamic, "f");
        }
        shouldStaticallyCloseForGlobal(ref, scopeManager) {
            // On global scope, let/const/class declarations should be resolved statically.
            const name = ref.identifier.name;
            const variable = this.set.get(name);
            if (!variable) {
                return false;
            }
            // variable exists on the scope
            // in module mode, we can statically resolve everything, regardless of its decl type
            if (scopeManager.isModule()) {
                return true;
            }
            // in script mode, only certain cases should be statically resolved
            // Example:
            // a `var` decl is ignored by the runtime if it clashes with a global name
            // this means that we should not resolve the reference to the variable
            const defs = variable.defs;
            return (defs.length > 0 &&
                defs.every(def => {
                    var _a;
                    if (def.type === definition_1.DefinitionType.Variable &&
                        ((_a = def.parent) === null || _a === void 0 ? void 0 : _a.type) === types_1.AST_NODE_TYPES.VariableDeclaration &&
                        def.parent.kind === 'var') {
                        return false;
                    }
                    return true;
                }));
        }
        close(scopeManager) {
            let closeRef;
            if (this.shouldStaticallyClose()) {
                closeRef = __classPrivateFieldGet(this, _ScopeBase_staticCloseRef, "f");
            }
            else if (this.type !== 'global') {
                closeRef = __classPrivateFieldGet(this, _ScopeBase_dynamicCloseRef, "f");
            }
            else {
                closeRef = __classPrivateFieldGet(this, _ScopeBase_globalCloseRef, "f");
            }
            // Try Resolving all references in this scope.
            assert_1.assert(this.leftToResolve);
            for (let i = 0; i < this.leftToResolve.length; ++i) {
                const ref = this.leftToResolve[i];
                closeRef(ref, scopeManager);
            }
            this.leftToResolve = null;
            return this.upper;
        }
        /**
         * To override by function scopes.
         * References in default parameters isn't resolved to variables which are in their function body.
         */
        isValidResolution(_ref, _variable) {
            return true;
        }
        delegateToUpperScope(ref) {
            const upper = this.upper;
            if (upper === null || upper === void 0 ? void 0 : upper.leftToResolve) {
                upper.leftToResolve.push(ref);
            }
            this.through.push(ref);
        }
        addDeclaredVariablesOfNode(variable, node) {
            if (node == null) {
                return;
            }
            let variables = __classPrivateFieldGet(this, _ScopeBase_declaredVariables, "f").get(node);
            if (variables == null) {
                variables = [];
                __classPrivateFieldGet(this, _ScopeBase_declaredVariables, "f").set(node, variables);
            }
            if (!variables.includes(variable)) {
                variables.push(variable);
            }
        }
        defineVariable(nameOrVariable, set, variables, node, def) {
            const name = typeof nameOrVariable === 'string' ? nameOrVariable : nameOrVariable.name;
            let variable = set.get(name);
            if (!variable) {
                variable =
                    typeof nameOrVariable === 'string'
                        ? new variable_1.Variable(name, this)
                        : nameOrVariable;
                set.set(name, variable);
                variables.push(variable);
            }
            if (def) {
                variable.defs.push(def);
                this.addDeclaredVariablesOfNode(variable, def.node);
                this.addDeclaredVariablesOfNode(variable, def.parent);
            }
            if (node) {
                variable.identifiers.push(node);
            }
        }
        defineIdentifier(node, def) {
            this.defineVariable(node.name, this.set, this.variables, node, def);
        }
        defineLiteralIdentifier(node, def) {
            this.defineVariable(node.value, this.set, this.variables, null, def);
        }
        referenceValue(node, assign = Reference_1.ReferenceFlag.Read, writeExpr, maybeImplicitGlobal, init = false) {
            var _a;
            const ref = new Reference_1.Reference(node, this, assign, writeExpr, maybeImplicitGlobal, init, Reference_1.ReferenceTypeFlag.Value);
            this.references.push(ref);
            (_a = this.leftToResolve) === null || _a === void 0 ? void 0 : _a.push(ref);
        }
        referenceType(node) {
            var _a;
            const ref = new Reference_1.Reference(node, this, Reference_1.ReferenceFlag.Read, null, null, false, Reference_1.ReferenceTypeFlag.Type);
            this.references.push(ref);
            (_a = this.leftToResolve) === null || _a === void 0 ? void 0 : _a.push(ref);
        }
        referenceDualValueType(node) {
            var _a;
            const ref = new Reference_1.Reference(node, this, Reference_1.ReferenceFlag.Read, null, null, false, Reference_1.ReferenceTypeFlag.Type | Reference_1.ReferenceTypeFlag.Value);
            this.references.push(ref);
            (_a = this.leftToResolve) === null || _a === void 0 ? void 0 : _a.push(ref);
        }
    }
    exports.ScopeBase = ScopeBase;
    _ScopeBase_declaredVariables = new WeakMap(), _ScopeBase_dynamic = new WeakMap(), _ScopeBase_staticCloseRef = new WeakMap(), _ScopeBase_dynamicCloseRef = new WeakMap(), _ScopeBase_globalCloseRef = new WeakMap();
},
/* @typescript-eslint/scope-manager/dist/scope/ScopeType.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var ScopeType;
    (function (ScopeType) {
        ScopeType["block"] = "block";
        ScopeType["catch"] = "catch";
        ScopeType["class"] = "class";
        ScopeType["conditionalType"] = "conditionalType";
        ScopeType["for"] = "for";
        ScopeType["function"] = "function";
        ScopeType["functionExpressionName"] = "function-expression-name";
        ScopeType["functionType"] = "functionType";
        ScopeType["global"] = "global";
        ScopeType["mappedType"] = "mappedType";
        ScopeType["module"] = "module";
        ScopeType["switch"] = "switch";
        ScopeType["tsEnum"] = "tsEnum";
        ScopeType["tsModule"] = "tsModule";
        ScopeType["type"] = "type";
        ScopeType["with"] = "with";
    })(ScopeType || (ScopeType = {}));
    exports.ScopeType = ScopeType;
},
/* @typescript-eslint/scope-manager/dist/assert.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    // the base assert module doesn't use ts assertion syntax
    function assert(value, message) {
        if (value == null) {
            throw new Error(message);
        }
    }
    exports.assert = assert;
},
/* @typescript-eslint/scope-manager/dist/referencer/Reference.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _Reference_flag, _Reference_referenceType;
    __esModule();
    const ID_1 = require(780) /* ../ID */;
    var ReferenceFlag;
    (function (ReferenceFlag) {
        ReferenceFlag[ReferenceFlag["Read"] = 1] = "Read";
        ReferenceFlag[ReferenceFlag["Write"] = 2] = "Write";
        ReferenceFlag[ReferenceFlag["ReadWrite"] = 3] = "ReadWrite";
    })(ReferenceFlag || (ReferenceFlag = {}));
    exports.ReferenceFlag = ReferenceFlag;
    const generator = ID_1.createIdGenerator();
    var ReferenceTypeFlag;
    (function (ReferenceTypeFlag) {
        ReferenceTypeFlag[ReferenceTypeFlag["Value"] = 1] = "Value";
        ReferenceTypeFlag[ReferenceTypeFlag["Type"] = 2] = "Type";
    })(ReferenceTypeFlag || (ReferenceTypeFlag = {}));
    exports.ReferenceTypeFlag = ReferenceTypeFlag;
    /**
     * A Reference represents a single occurrence of an identifier in code.
     */
    class Reference {
        constructor(identifier, scope, flag, writeExpr, maybeImplicitGlobal, init, referenceType = ReferenceTypeFlag.Value) {
            /**
             * A unique ID for this instance - primarily used to help debugging and testing
             */
            this.$id = generator();
            /**
             * The read-write mode of the reference.
             */
            _Reference_flag.set(this, void 0);
            /**
             * In some cases, a reference may be a type, value or both a type and value reference.
             */
            _Reference_referenceType.set(this, void 0);
            this.identifier = identifier;
            this.from = scope;
            this.resolved = null;
            __classPrivateFieldSet(this, _Reference_flag, flag, "f");
            if (this.isWrite()) {
                this.writeExpr = writeExpr;
                this.init = init;
            }
            this.maybeImplicitGlobal = maybeImplicitGlobal;
            __classPrivateFieldSet(this, _Reference_referenceType, referenceType, "f");
        }
        /**
         * True if this reference can reference types
         */
        get isTypeReference() {
            return (__classPrivateFieldGet(this, _Reference_referenceType, "f") & ReferenceTypeFlag.Type) !== 0;
        }
        /**
         * True if this reference can reference values
         */
        get isValueReference() {
            return (__classPrivateFieldGet(this, _Reference_referenceType, "f") & ReferenceTypeFlag.Value) !== 0;
        }
        /**
         * Whether the reference is writeable.
         * @public
         */
        isWrite() {
            return !!(__classPrivateFieldGet(this, _Reference_flag, "f") & ReferenceFlag.Write);
        }
        /**
         * Whether the reference is readable.
         * @public
         */
        isRead() {
            return !!(__classPrivateFieldGet(this, _Reference_flag, "f") & ReferenceFlag.Read);
        }
        /**
         * Whether the reference is read-only.
         * @public
         */
        isReadOnly() {
            return __classPrivateFieldGet(this, _Reference_flag, "f") === ReferenceFlag.Read;
        }
        /**
         * Whether the reference is write-only.
         * @public
         */
        isWriteOnly() {
            return __classPrivateFieldGet(this, _Reference_flag, "f") === ReferenceFlag.Write;
        }
        /**
         * Whether the reference is read-write.
         * @public
         */
        isReadWrite() {
            return __classPrivateFieldGet(this, _Reference_flag, "f") === ReferenceFlag.ReadWrite;
        }
    }
    exports.Reference = Reference;
    _Reference_flag = new WeakMap(), _Reference_referenceType = new WeakMap();
},
/* @typescript-eslint/scope-manager/dist/variable/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var ESLintScopeVariable_1 = require(803) /* ./ESLintScopeVariable */;
    __esExport("ESLintScopeVariable", ESLintScopeVariable_1.ESLintScopeVariable);
    var ImplicitLibVariable_1 = require(805) /* ./ImplicitLibVariable */;
    __esExport("ImplicitLibVariable", ImplicitLibVariable_1.ImplicitLibVariable);
    var Variable_1 = require(806) /* ./Variable */;
    __esExport("Variable", Variable_1.Variable);
},
/* @typescript-eslint/scope-manager/dist/variable/ESLintScopeVariable.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const VariableBase_1 = require(804) /* ./VariableBase */;
    /**
     * ESLint defines global variables using the eslint-scope Variable class
     * This is declared her for consumers to use
     */
    class ESLintScopeVariable extends VariableBase_1.VariableBase {
    }
    exports.ESLintScopeVariable = ESLintScopeVariable;
},
/* @typescript-eslint/scope-manager/dist/variable/VariableBase.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ID_1 = require(780) /* ../ID */;
    const generator = ID_1.createIdGenerator();
    class VariableBase {
        constructor(name, scope) {
            /**
             * A unique ID for this instance - primarily used to help debugging and testing
             */
            this.$id = generator();
            /**
             * The array of the definitions of this variable.
             * @public
             */
            this.defs = [];
            /**
             * True if the variable is considered used for the purposes of `no-unused-vars`, false otherwise.
             * @public
             */
            this.eslintUsed = false;
            /**
             * The array of `Identifier` nodes which define this variable.
             * If this variable is redeclared, this array includes two or more nodes.
             * @public
             */
            this.identifiers = [];
            /**
             * List of {@link Reference} of this variable (excluding parameter entries)  in its defining scope and all nested scopes.
             * For defining occurrences only see {@link Variable#defs}.
             * @public
             */
            this.references = [];
            this.name = name;
            this.scope = scope;
        }
    }
    exports.VariableBase = VariableBase;
},
/* @typescript-eslint/scope-manager/dist/variable/ImplicitLibVariable.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ESLintScopeVariable_1 = require(803) /* ./ESLintScopeVariable */;
    /**
     * An variable implicitly defined by the TS Lib
     */
    class ImplicitLibVariable extends ESLintScopeVariable_1.ESLintScopeVariable {
        constructor(scope, name, { isTypeVariable, isValueVariable, writeable, eslintImplicitGlobalSetting, }) {
            super(name, scope);
            this.isTypeVariable = isTypeVariable !== null && isTypeVariable !== void 0 ? isTypeVariable : false;
            this.isValueVariable = isValueVariable !== null && isValueVariable !== void 0 ? isValueVariable : false;
            this.writeable = writeable !== null && writeable !== void 0 ? writeable : false;
            this.eslintImplicitGlobalSetting =
                eslintImplicitGlobalSetting !== null && eslintImplicitGlobalSetting !== void 0 ? eslintImplicitGlobalSetting : 'readonly';
        }
    }
    exports.ImplicitLibVariable = ImplicitLibVariable;
},
/* @typescript-eslint/scope-manager/dist/variable/Variable.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const VariableBase_1 = require(804) /* ./VariableBase */;
    /**
     * A Variable represents a locally scoped identifier. These include arguments to functions.
     */
    class Variable extends VariableBase_1.VariableBase {
        /**
         * `true` if the variable is valid in a type context, false otherwise
         * @public
         */
        get isTypeVariable() {
            if (this.defs.length === 0) {
                // we don't statically know whether this is a type or a value
                return true;
            }
            return this.defs.some(def => def.isTypeDefinition);
        }
        /**
         * `true` if the variable is valid in a value context, false otherwise
         * @public
         */
        get isValueVariable() {
            if (this.defs.length === 0) {
                // we don't statically know whether this is a type or a value
                return true;
            }
            return this.defs.some(def => def.isVariableDefinition);
        }
    }
    exports.Variable = Variable;
},
/* @typescript-eslint/scope-manager/dist/scope/CatchScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class CatchScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.catch, upperScope, block, false);
        }
    }
    exports.CatchScope = CatchScope;
},
/* @typescript-eslint/scope-manager/dist/scope/ClassScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class ClassScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.class, upperScope, block, false);
        }
    }
    exports.ClassScope = ClassScope;
},
/* @typescript-eslint/scope-manager/dist/scope/ConditionalTypeScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class ConditionalTypeScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.conditionalType, upperScope, block, false);
        }
    }
    exports.ConditionalTypeScope = ConditionalTypeScope;
},
/* @typescript-eslint/scope-manager/dist/scope/ForScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class ForScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.for, upperScope, block, false);
        }
    }
    exports.ForScope = ForScope;
},
/* @typescript-eslint/scope-manager/dist/scope/FunctionExpressionNameScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    const definition_1 = require(776) /* ../definition */;
    class FunctionExpressionNameScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.functionExpressionName, upperScope, block, false);
            if (block.id) {
                this.defineIdentifier(block.id, new definition_1.FunctionNameDefinition(block.id, block));
            }
            this.functionExpressionScope = true;
        }
    }
    exports.FunctionExpressionNameScope = FunctionExpressionNameScope;
},
/* @typescript-eslint/scope-manager/dist/scope/FunctionScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const types_1 = require(740) /* @typescript-eslint/types */;
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class FunctionScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block, isMethodDefinition) {
            super(scopeManager, ScopeType_1.ScopeType.function, upperScope, block, isMethodDefinition);
            // section 9.2.13, FunctionDeclarationInstantiation.
            // NOTE Arrow functions never have an arguments objects.
            if (this.block.type !== types_1.AST_NODE_TYPES.ArrowFunctionExpression) {
                this.defineVariable('arguments', this.set, this.variables, null, null);
            }
        }
        // References in default parameters isn't resolved to variables which are in their function body.
        //     const x = 1
        //     function f(a = x) { // This `x` is resolved to the `x` in the outer scope.
        //         const x = 2
        //         console.log(a)
        //     }
        isValidResolution(ref, variable) {
            var _a, _b;
            // If `options.gloablReturn` is true, `this.block` becomes a Program node.
            if (this.block.type === types_1.AST_NODE_TYPES.Program) {
                return true;
            }
            const bodyStart = (_b = (_a = this.block.body) === null || _a === void 0 ? void 0 : _a.range[0]) !== null && _b !== void 0 ? _b : -1;
            // It's invalid resolution in the following case:
            return !((variable.scope === this &&
                ref.identifier.range[0] < bodyStart && // the reference is in the parameter part.
                variable.defs.every(d => d.name.range[0] >= bodyStart)) // the variable is in the body.
            );
        }
    }
    exports.FunctionScope = FunctionScope;
},
/* @typescript-eslint/scope-manager/dist/scope/FunctionTypeScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class FunctionTypeScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.functionType, upperScope, block, false);
        }
    }
    exports.FunctionTypeScope = FunctionTypeScope;
},
/* @typescript-eslint/scope-manager/dist/scope/GlobalScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const types_1 = require(740) /* @typescript-eslint/types */;
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    const assert_1 = require(800) /* ../assert */;
    const ImplicitGlobalVariableDefinition_1 = require(784) /* ../definition/ImplicitGlobalVariableDefinition */;
    const variable_1 = require(802) /* ../variable */;
    class GlobalScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, block) {
            super(scopeManager, ScopeType_1.ScopeType.global, null, block, false);
            this.implicit = {
                set: new Map(),
                variables: [],
                leftToBeResolved: [],
            };
        }
        defineImplicitVariable(name, options) {
            this.defineVariable(new variable_1.ImplicitLibVariable(this, name, options), this.set, this.variables, null, null);
        }
        close(scopeManager) {
            assert_1.assert(this.leftToResolve);
            for (const ref of this.leftToResolve) {
                if (ref.maybeImplicitGlobal && !this.set.has(ref.identifier.name)) {
                    // create an implicit global variable from assignment expression
                    const info = ref.maybeImplicitGlobal;
                    const node = info.pattern;
                    if (node && node.type === types_1.AST_NODE_TYPES.Identifier) {
                        this.defineVariable(node.name, this.implicit.set, this.implicit.variables, node, new ImplicitGlobalVariableDefinition_1.ImplicitGlobalVariableDefinition(info.pattern, info.node));
                    }
                }
            }
            this.implicit.leftToBeResolved = this.leftToResolve;
            return super.close(scopeManager);
        }
    }
    exports.GlobalScope = GlobalScope;
},
/* @typescript-eslint/scope-manager/dist/scope/MappedTypeScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class MappedTypeScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.mappedType, upperScope, block, false);
        }
    }
    exports.MappedTypeScope = MappedTypeScope;
},
/* @typescript-eslint/scope-manager/dist/scope/ModuleScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class ModuleScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.module, upperScope, block, false);
        }
    }
    exports.ModuleScope = ModuleScope;
},
/* @typescript-eslint/scope-manager/dist/scope/Scope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/scope-manager/dist/scope/SwitchScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class SwitchScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.switch, upperScope, block, false);
        }
    }
    exports.SwitchScope = SwitchScope;
},
/* @typescript-eslint/scope-manager/dist/scope/TSEnumScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class TSEnumScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.tsEnum, upperScope, block, false);
        }
    }
    exports.TSEnumScope = TSEnumScope;
},
/* @typescript-eslint/scope-manager/dist/scope/TSModuleScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class TSModuleScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.tsModule, upperScope, block, false);
        }
    }
    exports.TSModuleScope = TSModuleScope;
},
/* @typescript-eslint/scope-manager/dist/scope/TypeScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    class TypeScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.type, upperScope, block, false);
        }
    }
    exports.TypeScope = TypeScope;
},
/* @typescript-eslint/scope-manager/dist/scope/WithScope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ScopeBase_1 = require(798) /* ./ScopeBase */;
    const ScopeType_1 = require(799) /* ./ScopeType */;
    const assert_1 = require(800) /* ../assert */;
    class WithScope extends ScopeBase_1.ScopeBase {
        constructor(scopeManager, upperScope, block) {
            super(scopeManager, ScopeType_1.ScopeType.with, upperScope, block, false);
        }
        close(scopeManager) {
            if (this.shouldStaticallyClose()) {
                return super.close(scopeManager);
            }
            assert_1.assert(this.leftToResolve);
            for (let i = 0; i < this.leftToResolve.length; ++i) {
                const ref = this.leftToResolve[i];
                this.delegateToUpperScope(ref);
            }
            this.leftToResolve = null;
            return this.upper;
        }
    }
    exports.WithScope = WithScope;
},
/* @typescript-eslint/scope-manager/dist/referencer/ExportVisitor.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _ExportVisitor_referencer, _ExportVisitor_exportNode;
    __esModule();
    const types_1 = require(740) /* @typescript-eslint/types */;
    const Visitor_1 = require(793) /* ./Visitor */;
    class ExportVisitor extends Visitor_1.Visitor {
        constructor(node, referencer) {
            super(referencer);
            _ExportVisitor_referencer.set(this, void 0);
            _ExportVisitor_exportNode.set(this, void 0);
            __classPrivateFieldSet(this, _ExportVisitor_exportNode, node, "f");
            __classPrivateFieldSet(this, _ExportVisitor_referencer, referencer, "f");
        }
        static visit(referencer, node) {
            const exportReferencer = new ExportVisitor(node, referencer);
            exportReferencer.visit(node);
        }
        Identifier(node) {
            if (__classPrivateFieldGet(this, _ExportVisitor_exportNode, "f").exportKind === 'type') {
                // type exports can only reference types
                __classPrivateFieldGet(this, _ExportVisitor_referencer, "f").currentScope().referenceType(node);
            }
            else {
                __classPrivateFieldGet(this, _ExportVisitor_referencer, "f").currentScope().referenceDualValueType(node);
            }
        }
        ExportDefaultDeclaration(node) {
            if (node.declaration.type === types_1.AST_NODE_TYPES.Identifier) {
                // export default A;
                // this could be a type or a variable
                this.visit(node.declaration);
            }
            else {
                // export const a = 1;
                // export something();
                // etc
                // these not included in the scope of this visitor as they are all guaranteed to be values or declare variables
            }
        }
        ExportNamedDeclaration(node) {
            if (node.source) {
                // export ... from 'foo';
                // these are external identifiers so there shouldn't be references or defs
                return;
            }
            if (!node.declaration) {
                // export { x };
                this.visitChildren(node);
            }
            else {
                // export const x = 1;
                // this is not included in the scope of this visitor as it creates a variable
            }
        }
        ExportSpecifier(node) {
            this.visit(node.local);
        }
    }
    exports.ExportVisitor = ExportVisitor;
    _ExportVisitor_referencer = new WeakMap(), _ExportVisitor_exportNode = new WeakMap();
},
/* @typescript-eslint/scope-manager/dist/referencer/ImportVisitor.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _ImportVisitor_declaration, _ImportVisitor_referencer;
    __esModule();
    const definition_1 = require(776) /* ../definition */;
    const Visitor_1 = require(793) /* ./Visitor */;
    class ImportVisitor extends Visitor_1.Visitor {
        constructor(declaration, referencer) {
            super(referencer);
            _ImportVisitor_declaration.set(this, void 0);
            _ImportVisitor_referencer.set(this, void 0);
            __classPrivateFieldSet(this, _ImportVisitor_declaration, declaration, "f");
            __classPrivateFieldSet(this, _ImportVisitor_referencer, referencer, "f");
        }
        static visit(referencer, declaration) {
            const importReferencer = new ImportVisitor(declaration, referencer);
            importReferencer.visit(declaration);
        }
        visitImport(id, specifier) {
            __classPrivateFieldGet(this, _ImportVisitor_referencer, "f")
                .currentScope()
                .defineIdentifier(id, new definition_1.ImportBindingDefinition(id, specifier, __classPrivateFieldGet(this, _ImportVisitor_declaration, "f")));
        }
        ImportNamespaceSpecifier(node) {
            const local = node.local;
            this.visitImport(local, node);
        }
        ImportDefaultSpecifier(node) {
            const local = node.local;
            this.visitImport(local, node);
        }
        ImportSpecifier(node) {
            const local = node.local;
            this.visitImport(local, node);
        }
    }
    exports.ImportVisitor = ImportVisitor;
    _ImportVisitor_declaration = new WeakMap(), _ImportVisitor_referencer = new WeakMap();
},
/* @typescript-eslint/scope-manager/dist/lib/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es5_1 = require(826) /* ./es5 */;
    const es6_1 = require(828) /* ./es6 */;
    const es2015_1 = require(838) /* ./es2015 */;
    const es7_1 = require(839) /* ./es7 */;
    const es2016_1 = require(841) /* ./es2016 */;
    const es2017_1 = require(842) /* ./es2017 */;
    const es2018_1 = require(848) /* ./es2018 */;
    const es2019_1 = require(854) /* ./es2019 */;
    const es2020_1 = require(859) /* ./es2020 */;
    const es2021_1 = require(866) /* ./es2021 */;
    const esnext_1 = require(870) /* ./esnext */;
    const dom_1 = require(872) /* ./dom */;
    const dom_iterable_1 = require(873) /* ./dom.iterable */;
    const webworker_1 = require(874) /* ./webworker */;
    const webworker_importscripts_1 = require(875) /* ./webworker.importscripts */;
    const webworker_iterable_1 = require(876) /* ./webworker.iterable */;
    const scripthost_1 = require(877) /* ./scripthost */;
    const es2015_core_1 = require(829) /* ./es2015.core */;
    const es2015_collection_1 = require(830) /* ./es2015.collection */;
    const es2015_generator_1 = require(833) /* ./es2015.generator */;
    const es2015_iterable_1 = require(831) /* ./es2015.iterable */;
    const es2015_promise_1 = require(834) /* ./es2015.promise */;
    const es2015_proxy_1 = require(835) /* ./es2015.proxy */;
    const es2015_reflect_1 = require(836) /* ./es2015.reflect */;
    const es2015_symbol_1 = require(832) /* ./es2015.symbol */;
    const es2015_symbol_wellknown_1 = require(837) /* ./es2015.symbol.wellknown */;
    const es2016_array_include_1 = require(840) /* ./es2016.array.include */;
    const es2017_object_1 = require(843) /* ./es2017.object */;
    const es2017_sharedmemory_1 = require(844) /* ./es2017.sharedmemory */;
    const es2017_string_1 = require(845) /* ./es2017.string */;
    const es2017_intl_1 = require(846) /* ./es2017.intl */;
    const es2017_typedarrays_1 = require(847) /* ./es2017.typedarrays */;
    const es2018_asyncgenerator_1 = require(850) /* ./es2018.asyncgenerator */;
    const es2018_asynciterable_1 = require(849) /* ./es2018.asynciterable */;
    const es2018_intl_1 = require(853) /* ./es2018.intl */;
    const es2018_promise_1 = require(851) /* ./es2018.promise */;
    const es2018_regexp_1 = require(852) /* ./es2018.regexp */;
    const es2019_array_1 = require(855) /* ./es2019.array */;
    const es2019_object_1 = require(856) /* ./es2019.object */;
    const es2019_string_1 = require(857) /* ./es2019.string */;
    const es2019_symbol_1 = require(858) /* ./es2019.symbol */;
    const es2020_bigint_1 = require(860) /* ./es2020.bigint */;
    const es2020_promise_1 = require(861) /* ./es2020.promise */;
    const es2020_sharedmemory_1 = require(862) /* ./es2020.sharedmemory */;
    const es2020_string_1 = require(863) /* ./es2020.string */;
    const es2020_symbol_wellknown_1 = require(864) /* ./es2020.symbol.wellknown */;
    const es2020_intl_1 = require(865) /* ./es2020.intl */;
    const es2021_promise_1 = require(867) /* ./es2021.promise */;
    const es2021_string_1 = require(868) /* ./es2021.string */;
    const es2021_weakref_1 = require(869) /* ./es2021.weakref */;
    const esnext_array_1 = require(878) /* ./esnext.array */;
    const esnext_symbol_1 = require(879) /* ./esnext.symbol */;
    const esnext_asynciterable_1 = require(880) /* ./esnext.asynciterable */;
    const esnext_intl_1 = require(871) /* ./esnext.intl */;
    const esnext_bigint_1 = require(881) /* ./esnext.bigint */;
    const esnext_string_1 = require(882) /* ./esnext.string */;
    const esnext_promise_1 = require(883) /* ./esnext.promise */;
    const esnext_weakref_1 = require(884) /* ./esnext.weakref */;
    const es2016_full_1 = require(885) /* ./es2016.full */;
    const es2017_full_1 = require(886) /* ./es2017.full */;
    const es2018_full_1 = require(887) /* ./es2018.full */;
    const es2019_full_1 = require(888) /* ./es2019.full */;
    const es2020_full_1 = require(889) /* ./es2020.full */;
    const es2021_full_1 = require(890) /* ./es2021.full */;
    const esnext_full_1 = require(891) /* ./esnext.full */;
    const lib_1 = require(892) /* ./lib */;
    const lib = {
        es5: es5_1.es5,
        es6: es6_1.es6,
        es2015: es2015_1.es2015,
        es7: es7_1.es7,
        es2016: es2016_1.es2016,
        es2017: es2017_1.es2017,
        es2018: es2018_1.es2018,
        es2019: es2019_1.es2019,
        es2020: es2020_1.es2020,
        es2021: es2021_1.es2021,
        esnext: esnext_1.esnext,
        dom: dom_1.dom,
        'dom.iterable': dom_iterable_1.dom_iterable,
        webworker: webworker_1.webworker,
        'webworker.importscripts': webworker_importscripts_1.webworker_importscripts,
        'webworker.iterable': webworker_iterable_1.webworker_iterable,
        scripthost: scripthost_1.scripthost,
        'es2015.core': es2015_core_1.es2015_core,
        'es2015.collection': es2015_collection_1.es2015_collection,
        'es2015.generator': es2015_generator_1.es2015_generator,
        'es2015.iterable': es2015_iterable_1.es2015_iterable,
        'es2015.promise': es2015_promise_1.es2015_promise,
        'es2015.proxy': es2015_proxy_1.es2015_proxy,
        'es2015.reflect': es2015_reflect_1.es2015_reflect,
        'es2015.symbol': es2015_symbol_1.es2015_symbol,
        'es2015.symbol.wellknown': es2015_symbol_wellknown_1.es2015_symbol_wellknown,
        'es2016.array.include': es2016_array_include_1.es2016_array_include,
        'es2017.object': es2017_object_1.es2017_object,
        'es2017.sharedmemory': es2017_sharedmemory_1.es2017_sharedmemory,
        'es2017.string': es2017_string_1.es2017_string,
        'es2017.intl': es2017_intl_1.es2017_intl,
        'es2017.typedarrays': es2017_typedarrays_1.es2017_typedarrays,
        'es2018.asyncgenerator': es2018_asyncgenerator_1.es2018_asyncgenerator,
        'es2018.asynciterable': es2018_asynciterable_1.es2018_asynciterable,
        'es2018.intl': es2018_intl_1.es2018_intl,
        'es2018.promise': es2018_promise_1.es2018_promise,
        'es2018.regexp': es2018_regexp_1.es2018_regexp,
        'es2019.array': es2019_array_1.es2019_array,
        'es2019.object': es2019_object_1.es2019_object,
        'es2019.string': es2019_string_1.es2019_string,
        'es2019.symbol': es2019_symbol_1.es2019_symbol,
        'es2020.bigint': es2020_bigint_1.es2020_bigint,
        'es2020.promise': es2020_promise_1.es2020_promise,
        'es2020.sharedmemory': es2020_sharedmemory_1.es2020_sharedmemory,
        'es2020.string': es2020_string_1.es2020_string,
        'es2020.symbol.wellknown': es2020_symbol_wellknown_1.es2020_symbol_wellknown,
        'es2020.intl': es2020_intl_1.es2020_intl,
        'es2021.promise': es2021_promise_1.es2021_promise,
        'es2021.string': es2021_string_1.es2021_string,
        'es2021.weakref': es2021_weakref_1.es2021_weakref,
        'esnext.array': esnext_array_1.esnext_array,
        'esnext.symbol': esnext_symbol_1.esnext_symbol,
        'esnext.asynciterable': esnext_asynciterable_1.esnext_asynciterable,
        'esnext.intl': esnext_intl_1.esnext_intl,
        'esnext.bigint': esnext_bigint_1.esnext_bigint,
        'esnext.string': esnext_string_1.esnext_string,
        'esnext.promise': esnext_promise_1.esnext_promise,
        'esnext.weakref': esnext_weakref_1.esnext_weakref,
        'es2016.full': es2016_full_1.es2016_full,
        'es2017.full': es2017_full_1.es2017_full,
        'es2018.full': es2018_full_1.es2018_full,
        'es2019.full': es2019_full_1.es2019_full,
        'es2020.full': es2020_full_1.es2020_full,
        'es2021.full': es2021_full_1.es2021_full,
        'esnext.full': esnext_full_1.esnext_full,
        lib: lib_1.lib,
    };
    exports.lib = lib;
},
/* @typescript-eslint/scope-manager/dist/lib/es5.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es5 = {
        Symbol: base_config_1.TYPE,
        PropertyKey: base_config_1.TYPE,
        PropertyDescriptor: base_config_1.TYPE,
        PropertyDescriptorMap: base_config_1.TYPE,
        Object: base_config_1.TYPE_VALUE,
        ObjectConstructor: base_config_1.TYPE,
        Function: base_config_1.TYPE_VALUE,
        FunctionConstructor: base_config_1.TYPE,
        ThisParameterType: base_config_1.TYPE,
        OmitThisParameter: base_config_1.TYPE,
        CallableFunction: base_config_1.TYPE,
        NewableFunction: base_config_1.TYPE,
        IArguments: base_config_1.TYPE,
        String: base_config_1.TYPE_VALUE,
        StringConstructor: base_config_1.TYPE,
        Boolean: base_config_1.TYPE_VALUE,
        BooleanConstructor: base_config_1.TYPE,
        Number: base_config_1.TYPE_VALUE,
        NumberConstructor: base_config_1.TYPE,
        TemplateStringsArray: base_config_1.TYPE,
        ImportMeta: base_config_1.TYPE,
        Math: base_config_1.TYPE_VALUE,
        Date: base_config_1.TYPE_VALUE,
        DateConstructor: base_config_1.TYPE,
        RegExpMatchArray: base_config_1.TYPE,
        RegExpExecArray: base_config_1.TYPE,
        RegExp: base_config_1.TYPE_VALUE,
        RegExpConstructor: base_config_1.TYPE,
        Error: base_config_1.TYPE_VALUE,
        ErrorConstructor: base_config_1.TYPE,
        EvalError: base_config_1.TYPE_VALUE,
        EvalErrorConstructor: base_config_1.TYPE,
        RangeError: base_config_1.TYPE_VALUE,
        RangeErrorConstructor: base_config_1.TYPE,
        ReferenceError: base_config_1.TYPE_VALUE,
        ReferenceErrorConstructor: base_config_1.TYPE,
        SyntaxError: base_config_1.TYPE_VALUE,
        SyntaxErrorConstructor: base_config_1.TYPE,
        TypeError: base_config_1.TYPE_VALUE,
        TypeErrorConstructor: base_config_1.TYPE,
        URIError: base_config_1.TYPE_VALUE,
        URIErrorConstructor: base_config_1.TYPE,
        JSON: base_config_1.TYPE_VALUE,
        ReadonlyArray: base_config_1.TYPE,
        ConcatArray: base_config_1.TYPE,
        Array: base_config_1.TYPE_VALUE,
        ArrayConstructor: base_config_1.TYPE,
        TypedPropertyDescriptor: base_config_1.TYPE,
        ClassDecorator: base_config_1.TYPE,
        PropertyDecorator: base_config_1.TYPE,
        MethodDecorator: base_config_1.TYPE,
        ParameterDecorator: base_config_1.TYPE,
        PromiseConstructorLike: base_config_1.TYPE,
        PromiseLike: base_config_1.TYPE,
        Promise: base_config_1.TYPE,
        ArrayLike: base_config_1.TYPE,
        Partial: base_config_1.TYPE,
        Required: base_config_1.TYPE,
        Readonly: base_config_1.TYPE,
        Pick: base_config_1.TYPE,
        Record: base_config_1.TYPE,
        Exclude: base_config_1.TYPE,
        Extract: base_config_1.TYPE,
        Omit: base_config_1.TYPE,
        NonNullable: base_config_1.TYPE,
        Parameters: base_config_1.TYPE,
        ConstructorParameters: base_config_1.TYPE,
        ReturnType: base_config_1.TYPE,
        InstanceType: base_config_1.TYPE,
        Uppercase: base_config_1.TYPE,
        Lowercase: base_config_1.TYPE,
        Capitalize: base_config_1.TYPE,
        Uncapitalize: base_config_1.TYPE,
        ThisType: base_config_1.TYPE,
        ArrayBuffer: base_config_1.TYPE_VALUE,
        ArrayBufferTypes: base_config_1.TYPE,
        ArrayBufferLike: base_config_1.TYPE,
        ArrayBufferConstructor: base_config_1.TYPE,
        ArrayBufferView: base_config_1.TYPE,
        DataView: base_config_1.TYPE_VALUE,
        DataViewConstructor: base_config_1.TYPE,
        Int8Array: base_config_1.TYPE_VALUE,
        Int8ArrayConstructor: base_config_1.TYPE,
        Uint8Array: base_config_1.TYPE_VALUE,
        Uint8ArrayConstructor: base_config_1.TYPE,
        Uint8ClampedArray: base_config_1.TYPE_VALUE,
        Uint8ClampedArrayConstructor: base_config_1.TYPE,
        Int16Array: base_config_1.TYPE_VALUE,
        Int16ArrayConstructor: base_config_1.TYPE,
        Uint16Array: base_config_1.TYPE_VALUE,
        Uint16ArrayConstructor: base_config_1.TYPE,
        Int32Array: base_config_1.TYPE_VALUE,
        Int32ArrayConstructor: base_config_1.TYPE,
        Uint32Array: base_config_1.TYPE_VALUE,
        Uint32ArrayConstructor: base_config_1.TYPE,
        Float32Array: base_config_1.TYPE_VALUE,
        Float32ArrayConstructor: base_config_1.TYPE,
        Float64Array: base_config_1.TYPE_VALUE,
        Float64ArrayConstructor: base_config_1.TYPE,
        Intl: base_config_1.TYPE_VALUE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/base-config.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.TYPE = Object.freeze({
        eslintImplicitGlobalSetting: 'readonly',
        isTypeVariable: true,
        isValueVariable: false,
    });
    exports.VALUE = Object.freeze({
        eslintImplicitGlobalSetting: 'readonly',
        isTypeVariable: false,
        isValueVariable: true,
    });
    exports.TYPE_VALUE = Object.freeze({
        eslintImplicitGlobalSetting: 'readonly',
        isTypeVariable: true,
        isValueVariable: true,
    });
},
/* @typescript-eslint/scope-manager/dist/lib/es6.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es5_1 = require(826) /* ./es5 */;
    const es2015_core_1 = require(829) /* ./es2015.core */;
    const es2015_collection_1 = require(830) /* ./es2015.collection */;
    const es2015_iterable_1 = require(831) /* ./es2015.iterable */;
    const es2015_generator_1 = require(833) /* ./es2015.generator */;
    const es2015_promise_1 = require(834) /* ./es2015.promise */;
    const es2015_proxy_1 = require(835) /* ./es2015.proxy */;
    const es2015_reflect_1 = require(836) /* ./es2015.reflect */;
    const es2015_symbol_1 = require(832) /* ./es2015.symbol */;
    const es2015_symbol_wellknown_1 = require(837) /* ./es2015.symbol.wellknown */;
    exports.es6 = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es5_1.es5), es2015_core_1.es2015_core), es2015_collection_1.es2015_collection), es2015_iterable_1.es2015_iterable), es2015_generator_1.es2015_generator), es2015_promise_1.es2015_promise), es2015_proxy_1.es2015_proxy), es2015_reflect_1.es2015_reflect), es2015_symbol_1.es2015_symbol), es2015_symbol_wellknown_1.es2015_symbol_wellknown);
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.core.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2015_core = {
        Array: base_config_1.TYPE,
        ArrayConstructor: base_config_1.TYPE,
        DateConstructor: base_config_1.TYPE,
        Function: base_config_1.TYPE,
        Math: base_config_1.TYPE,
        NumberConstructor: base_config_1.TYPE,
        ObjectConstructor: base_config_1.TYPE,
        ReadonlyArray: base_config_1.TYPE,
        RegExp: base_config_1.TYPE,
        RegExpConstructor: base_config_1.TYPE,
        String: base_config_1.TYPE,
        StringConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.collection.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2015_collection = {
        Map: base_config_1.TYPE_VALUE,
        MapConstructor: base_config_1.TYPE,
        ReadonlyMap: base_config_1.TYPE,
        WeakMap: base_config_1.TYPE_VALUE,
        WeakMapConstructor: base_config_1.TYPE,
        Set: base_config_1.TYPE_VALUE,
        SetConstructor: base_config_1.TYPE,
        ReadonlySet: base_config_1.TYPE,
        WeakSet: base_config_1.TYPE_VALUE,
        WeakSetConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.iterable.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_symbol_1 = require(832) /* ./es2015.symbol */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2015_iterable = Object.assign(Object.assign({}, es2015_symbol_1.es2015_symbol), { SymbolConstructor: base_config_1.TYPE, IteratorYieldResult: base_config_1.TYPE, IteratorReturnResult: base_config_1.TYPE, IteratorResult: base_config_1.TYPE, Iterator: base_config_1.TYPE, Iterable: base_config_1.TYPE, IterableIterator: base_config_1.TYPE, Array: base_config_1.TYPE, ArrayConstructor: base_config_1.TYPE, ReadonlyArray: base_config_1.TYPE, IArguments: base_config_1.TYPE, Map: base_config_1.TYPE, ReadonlyMap: base_config_1.TYPE, MapConstructor: base_config_1.TYPE, WeakMap: base_config_1.TYPE, WeakMapConstructor: base_config_1.TYPE, Set: base_config_1.TYPE, ReadonlySet: base_config_1.TYPE, SetConstructor: base_config_1.TYPE, WeakSet: base_config_1.TYPE, WeakSetConstructor: base_config_1.TYPE, Promise: base_config_1.TYPE, PromiseConstructor: base_config_1.TYPE, String: base_config_1.TYPE, Int8Array: base_config_1.TYPE, Int8ArrayConstructor: base_config_1.TYPE, Uint8Array: base_config_1.TYPE, Uint8ArrayConstructor: base_config_1.TYPE, Uint8ClampedArray: base_config_1.TYPE, Uint8ClampedArrayConstructor: base_config_1.TYPE, Int16Array: base_config_1.TYPE, Int16ArrayConstructor: base_config_1.TYPE, Uint16Array: base_config_1.TYPE, Uint16ArrayConstructor: base_config_1.TYPE, Int32Array: base_config_1.TYPE, Int32ArrayConstructor: base_config_1.TYPE, Uint32Array: base_config_1.TYPE, Uint32ArrayConstructor: base_config_1.TYPE, Float32Array: base_config_1.TYPE, Float32ArrayConstructor: base_config_1.TYPE, Float64Array: base_config_1.TYPE, Float64ArrayConstructor: base_config_1.TYPE });
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.symbol.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2015_symbol = {
        SymbolConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.generator.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_iterable_1 = require(831) /* ./es2015.iterable */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2015_generator = Object.assign(Object.assign({}, es2015_iterable_1.es2015_iterable), { Generator: base_config_1.TYPE, GeneratorFunction: base_config_1.TYPE, GeneratorFunctionConstructor: base_config_1.TYPE });
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.promise.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2015_promise = {
        PromiseConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.proxy.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2015_proxy = {
        ProxyHandler: base_config_1.TYPE,
        ProxyConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.reflect.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2015_reflect = {
        Reflect: base_config_1.TYPE_VALUE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.symbol.wellknown.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_symbol_1 = require(832) /* ./es2015.symbol */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2015_symbol_wellknown = Object.assign(Object.assign({}, es2015_symbol_1.es2015_symbol), { SymbolConstructor: base_config_1.TYPE, Symbol: base_config_1.TYPE, Array: base_config_1.TYPE, Date: base_config_1.TYPE, Map: base_config_1.TYPE, WeakMap: base_config_1.TYPE, Set: base_config_1.TYPE, WeakSet: base_config_1.TYPE, JSON: base_config_1.TYPE, Function: base_config_1.TYPE, GeneratorFunction: base_config_1.TYPE, Math: base_config_1.TYPE, Promise: base_config_1.TYPE, PromiseConstructor: base_config_1.TYPE, RegExp: base_config_1.TYPE, RegExpConstructor: base_config_1.TYPE, String: base_config_1.TYPE, ArrayBuffer: base_config_1.TYPE, DataView: base_config_1.TYPE, Int8Array: base_config_1.TYPE, Uint8Array: base_config_1.TYPE, Uint8ClampedArray: base_config_1.TYPE, Int16Array: base_config_1.TYPE, Uint16Array: base_config_1.TYPE, Int32Array: base_config_1.TYPE, Uint32Array: base_config_1.TYPE, Float32Array: base_config_1.TYPE, Float64Array: base_config_1.TYPE, ArrayConstructor: base_config_1.TYPE, MapConstructor: base_config_1.TYPE, SetConstructor: base_config_1.TYPE, ArrayBufferConstructor: base_config_1.TYPE });
},
/* @typescript-eslint/scope-manager/dist/lib/es2015.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es5_1 = require(826) /* ./es5 */;
    const es2015_core_1 = require(829) /* ./es2015.core */;
    const es2015_collection_1 = require(830) /* ./es2015.collection */;
    const es2015_iterable_1 = require(831) /* ./es2015.iterable */;
    const es2015_generator_1 = require(833) /* ./es2015.generator */;
    const es2015_promise_1 = require(834) /* ./es2015.promise */;
    const es2015_proxy_1 = require(835) /* ./es2015.proxy */;
    const es2015_reflect_1 = require(836) /* ./es2015.reflect */;
    const es2015_symbol_1 = require(832) /* ./es2015.symbol */;
    const es2015_symbol_wellknown_1 = require(837) /* ./es2015.symbol.wellknown */;
    exports.es2015 = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es5_1.es5), es2015_core_1.es2015_core), es2015_collection_1.es2015_collection), es2015_iterable_1.es2015_iterable), es2015_generator_1.es2015_generator), es2015_promise_1.es2015_promise), es2015_proxy_1.es2015_proxy), es2015_reflect_1.es2015_reflect), es2015_symbol_1.es2015_symbol), es2015_symbol_wellknown_1.es2015_symbol_wellknown);
},
/* @typescript-eslint/scope-manager/dist/lib/es7.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_1 = require(838) /* ./es2015 */;
    const es2016_array_include_1 = require(840) /* ./es2016.array.include */;
    exports.es7 = Object.assign(Object.assign({}, es2015_1.es2015), es2016_array_include_1.es2016_array_include);
},
/* @typescript-eslint/scope-manager/dist/lib/es2016.array.include.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2016_array_include = {
        Array: base_config_1.TYPE,
        ReadonlyArray: base_config_1.TYPE,
        Int8Array: base_config_1.TYPE,
        Uint8Array: base_config_1.TYPE,
        Uint8ClampedArray: base_config_1.TYPE,
        Int16Array: base_config_1.TYPE,
        Uint16Array: base_config_1.TYPE,
        Int32Array: base_config_1.TYPE,
        Uint32Array: base_config_1.TYPE,
        Float32Array: base_config_1.TYPE,
        Float64Array: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2016.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_1 = require(838) /* ./es2015 */;
    const es2016_array_include_1 = require(840) /* ./es2016.array.include */;
    exports.es2016 = Object.assign(Object.assign({}, es2015_1.es2015), es2016_array_include_1.es2016_array_include);
},
/* @typescript-eslint/scope-manager/dist/lib/es2017.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2016_1 = require(841) /* ./es2016 */;
    const es2017_object_1 = require(843) /* ./es2017.object */;
    const es2017_sharedmemory_1 = require(844) /* ./es2017.sharedmemory */;
    const es2017_string_1 = require(845) /* ./es2017.string */;
    const es2017_intl_1 = require(846) /* ./es2017.intl */;
    const es2017_typedarrays_1 = require(847) /* ./es2017.typedarrays */;
    exports.es2017 = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2016_1.es2016), es2017_object_1.es2017_object), es2017_sharedmemory_1.es2017_sharedmemory), es2017_string_1.es2017_string), es2017_intl_1.es2017_intl), es2017_typedarrays_1.es2017_typedarrays);
},
/* @typescript-eslint/scope-manager/dist/lib/es2017.object.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2017_object = {
        ObjectConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2017.sharedmemory.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_symbol_1 = require(832) /* ./es2015.symbol */;
    const es2015_symbol_wellknown_1 = require(837) /* ./es2015.symbol.wellknown */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2017_sharedmemory = Object.assign(Object.assign(Object.assign({}, es2015_symbol_1.es2015_symbol), es2015_symbol_wellknown_1.es2015_symbol_wellknown), { SharedArrayBuffer: base_config_1.TYPE_VALUE, SharedArrayBufferConstructor: base_config_1.TYPE, ArrayBufferTypes: base_config_1.TYPE, Atomics: base_config_1.TYPE_VALUE });
},
/* @typescript-eslint/scope-manager/dist/lib/es2017.string.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2017_string = {
        String: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2017.intl.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2017_intl = {
        Intl: base_config_1.TYPE_VALUE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2017.typedarrays.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2017_typedarrays = {
        Int8ArrayConstructor: base_config_1.TYPE,
        Uint8ArrayConstructor: base_config_1.TYPE,
        Uint8ClampedArrayConstructor: base_config_1.TYPE,
        Int16ArrayConstructor: base_config_1.TYPE,
        Uint16ArrayConstructor: base_config_1.TYPE,
        Int32ArrayConstructor: base_config_1.TYPE,
        Uint32ArrayConstructor: base_config_1.TYPE,
        Float32ArrayConstructor: base_config_1.TYPE,
        Float64ArrayConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2018.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2017_1 = require(842) /* ./es2017 */;
    const es2018_asynciterable_1 = require(849) /* ./es2018.asynciterable */;
    const es2018_asyncgenerator_1 = require(850) /* ./es2018.asyncgenerator */;
    const es2018_promise_1 = require(851) /* ./es2018.promise */;
    const es2018_regexp_1 = require(852) /* ./es2018.regexp */;
    const es2018_intl_1 = require(853) /* ./es2018.intl */;
    exports.es2018 = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2017_1.es2017), es2018_asynciterable_1.es2018_asynciterable), es2018_asyncgenerator_1.es2018_asyncgenerator), es2018_promise_1.es2018_promise), es2018_regexp_1.es2018_regexp), es2018_intl_1.es2018_intl);
},
/* @typescript-eslint/scope-manager/dist/lib/es2018.asynciterable.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_symbol_1 = require(832) /* ./es2015.symbol */;
    const es2015_iterable_1 = require(831) /* ./es2015.iterable */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2018_asynciterable = Object.assign(Object.assign(Object.assign({}, es2015_symbol_1.es2015_symbol), es2015_iterable_1.es2015_iterable), { SymbolConstructor: base_config_1.TYPE, AsyncIterator: base_config_1.TYPE, AsyncIterable: base_config_1.TYPE, AsyncIterableIterator: base_config_1.TYPE });
},
/* @typescript-eslint/scope-manager/dist/lib/es2018.asyncgenerator.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2018_asynciterable_1 = require(849) /* ./es2018.asynciterable */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2018_asyncgenerator = Object.assign(Object.assign({}, es2018_asynciterable_1.es2018_asynciterable), { AsyncGenerator: base_config_1.TYPE, AsyncGeneratorFunction: base_config_1.TYPE, AsyncGeneratorFunctionConstructor: base_config_1.TYPE });
},
/* @typescript-eslint/scope-manager/dist/lib/es2018.promise.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2018_promise = {
        Promise: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2018.regexp.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2018_regexp = {
        RegExpMatchArray: base_config_1.TYPE,
        RegExpExecArray: base_config_1.TYPE,
        RegExp: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2018.intl.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2018_intl = {
        Intl: base_config_1.TYPE_VALUE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2019.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2018_1 = require(848) /* ./es2018 */;
    const es2019_array_1 = require(855) /* ./es2019.array */;
    const es2019_object_1 = require(856) /* ./es2019.object */;
    const es2019_string_1 = require(857) /* ./es2019.string */;
    const es2019_symbol_1 = require(858) /* ./es2019.symbol */;
    exports.es2019 = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2018_1.es2018), es2019_array_1.es2019_array), es2019_object_1.es2019_object), es2019_string_1.es2019_string), es2019_symbol_1.es2019_symbol);
},
/* @typescript-eslint/scope-manager/dist/lib/es2019.array.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2019_array = {
        FlatArray: base_config_1.TYPE,
        ReadonlyArray: base_config_1.TYPE,
        Array: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2019.object.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_iterable_1 = require(831) /* ./es2015.iterable */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2019_object = Object.assign(Object.assign({}, es2015_iterable_1.es2015_iterable), { ObjectConstructor: base_config_1.TYPE });
},
/* @typescript-eslint/scope-manager/dist/lib/es2019.string.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2019_string = {
        String: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2019.symbol.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2019_symbol = {
        Symbol: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2020.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2019_1 = require(854) /* ./es2019 */;
    const es2020_bigint_1 = require(860) /* ./es2020.bigint */;
    const es2020_promise_1 = require(861) /* ./es2020.promise */;
    const es2020_sharedmemory_1 = require(862) /* ./es2020.sharedmemory */;
    const es2020_string_1 = require(863) /* ./es2020.string */;
    const es2020_symbol_wellknown_1 = require(864) /* ./es2020.symbol.wellknown */;
    const es2020_intl_1 = require(865) /* ./es2020.intl */;
    exports.es2020 = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2019_1.es2019), es2020_bigint_1.es2020_bigint), es2020_promise_1.es2020_promise), es2020_sharedmemory_1.es2020_sharedmemory), es2020_string_1.es2020_string), es2020_symbol_wellknown_1.es2020_symbol_wellknown), es2020_intl_1.es2020_intl);
},
/* @typescript-eslint/scope-manager/dist/lib/es2020.bigint.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2020_bigint = {
        BigIntToLocaleStringOptions: base_config_1.TYPE,
        BigInt: base_config_1.TYPE_VALUE,
        BigIntConstructor: base_config_1.TYPE,
        BigInt64Array: base_config_1.TYPE_VALUE,
        BigInt64ArrayConstructor: base_config_1.TYPE,
        BigUint64Array: base_config_1.TYPE_VALUE,
        BigUint64ArrayConstructor: base_config_1.TYPE,
        DataView: base_config_1.TYPE,
        Intl: base_config_1.TYPE_VALUE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2020.promise.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2020_promise = {
        PromiseFulfilledResult: base_config_1.TYPE,
        PromiseRejectedResult: base_config_1.TYPE,
        PromiseSettledResult: base_config_1.TYPE,
        PromiseConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2020.sharedmemory.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2020_sharedmemory = {
        Atomics: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2020.string.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_iterable_1 = require(831) /* ./es2015.iterable */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2020_string = Object.assign(Object.assign({}, es2015_iterable_1.es2015_iterable), { String: base_config_1.TYPE });
},
/* @typescript-eslint/scope-manager/dist/lib/es2020.symbol.wellknown.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_iterable_1 = require(831) /* ./es2015.iterable */;
    const es2015_symbol_1 = require(832) /* ./es2015.symbol */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2020_symbol_wellknown = Object.assign(Object.assign(Object.assign({}, es2015_iterable_1.es2015_iterable), es2015_symbol_1.es2015_symbol), { SymbolConstructor: base_config_1.TYPE, RegExp: base_config_1.TYPE });
},
/* @typescript-eslint/scope-manager/dist/lib/es2020.intl.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2020_intl = {
        Intl: base_config_1.TYPE_VALUE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2021.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2020_1 = require(859) /* ./es2020 */;
    const es2021_promise_1 = require(867) /* ./es2021.promise */;
    const es2021_string_1 = require(868) /* ./es2021.string */;
    const es2021_weakref_1 = require(869) /* ./es2021.weakref */;
    exports.es2021 = Object.assign(Object.assign(Object.assign(Object.assign({}, es2020_1.es2020), es2021_promise_1.es2021_promise), es2021_string_1.es2021_string), es2021_weakref_1.es2021_weakref);
},
/* @typescript-eslint/scope-manager/dist/lib/es2021.promise.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2021_promise = {
        AggregateError: base_config_1.TYPE_VALUE,
        AggregateErrorConstructor: base_config_1.TYPE,
        PromiseConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2021.string.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2021_string = {
        String: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2021.weakref.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.es2021_weakref = {
        WeakRef: base_config_1.TYPE_VALUE,
        WeakRefConstructor: base_config_1.TYPE,
        FinalizationRegistry: base_config_1.TYPE_VALUE,
        FinalizationRegistryConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2021_1 = require(866) /* ./es2021 */;
    const esnext_intl_1 = require(871) /* ./esnext.intl */;
    exports.esnext = Object.assign(Object.assign({}, es2021_1.es2021), esnext_intl_1.esnext_intl);
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.intl.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.esnext_intl = {
        Intl: base_config_1.TYPE_VALUE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/dom.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.dom = {
        AddEventListenerOptions: base_config_1.TYPE,
        AddressErrors: base_config_1.TYPE,
        AesCbcParams: base_config_1.TYPE,
        AesCtrParams: base_config_1.TYPE,
        AesDerivedKeyParams: base_config_1.TYPE,
        AesGcmParams: base_config_1.TYPE,
        AesKeyAlgorithm: base_config_1.TYPE,
        AesKeyGenParams: base_config_1.TYPE,
        Algorithm: base_config_1.TYPE,
        AnalyserOptions: base_config_1.TYPE,
        AnimationEventInit: base_config_1.TYPE,
        AnimationPlaybackEventInit: base_config_1.TYPE,
        AssignedNodesOptions: base_config_1.TYPE,
        AudioBufferOptions: base_config_1.TYPE,
        AudioBufferSourceOptions: base_config_1.TYPE,
        AudioContextInfo: base_config_1.TYPE,
        AudioContextOptions: base_config_1.TYPE,
        AudioNodeOptions: base_config_1.TYPE,
        AudioParamDescriptor: base_config_1.TYPE,
        AudioProcessingEventInit: base_config_1.TYPE,
        AudioTimestamp: base_config_1.TYPE,
        AudioWorkletNodeOptions: base_config_1.TYPE,
        AuthenticationExtensionsClientInputs: base_config_1.TYPE,
        AuthenticationExtensionsClientOutputs: base_config_1.TYPE,
        AuthenticatorSelectionCriteria: base_config_1.TYPE,
        BiquadFilterOptions: base_config_1.TYPE,
        BlobPropertyBag: base_config_1.TYPE,
        ByteLengthChunk: base_config_1.TYPE,
        CacheQueryOptions: base_config_1.TYPE,
        CanvasRenderingContext2DSettings: base_config_1.TYPE,
        ChannelMergerOptions: base_config_1.TYPE,
        ChannelSplitterOptions: base_config_1.TYPE,
        ClientQueryOptions: base_config_1.TYPE,
        ClipboardEventInit: base_config_1.TYPE,
        CloseEventInit: base_config_1.TYPE,
        CompositionEventInit: base_config_1.TYPE,
        ComputedEffectTiming: base_config_1.TYPE,
        ComputedKeyframe: base_config_1.TYPE,
        ConfirmSiteSpecificExceptionsInformation: base_config_1.TYPE,
        ConstantSourceOptions: base_config_1.TYPE,
        ConstrainBooleanParameters: base_config_1.TYPE,
        ConstrainDOMStringParameters: base_config_1.TYPE,
        ConstrainDoubleRange: base_config_1.TYPE,
        ConstrainULongRange: base_config_1.TYPE,
        ConstrainVideoFacingModeParameters: base_config_1.TYPE,
        ConvolverOptions: base_config_1.TYPE,
        CredentialCreationOptions: base_config_1.TYPE,
        CredentialPropertiesOutput: base_config_1.TYPE,
        CredentialRequestOptions: base_config_1.TYPE,
        CustomEventInit: base_config_1.TYPE,
        DOMMatrix2DInit: base_config_1.TYPE,
        DOMMatrixInit: base_config_1.TYPE,
        DOMPointInit: base_config_1.TYPE,
        DOMQuadInit: base_config_1.TYPE,
        DOMRectInit: base_config_1.TYPE,
        DelayOptions: base_config_1.TYPE,
        DeviceMotionEventAccelerationInit: base_config_1.TYPE,
        DeviceMotionEventInit: base_config_1.TYPE,
        DeviceMotionEventRotationRateInit: base_config_1.TYPE,
        DeviceOrientationEventInit: base_config_1.TYPE,
        DevicePermissionDescriptor: base_config_1.TYPE,
        DocumentTimelineOptions: base_config_1.TYPE,
        DoubleRange: base_config_1.TYPE,
        DragEventInit: base_config_1.TYPE,
        DynamicsCompressorOptions: base_config_1.TYPE,
        EcKeyAlgorithm: base_config_1.TYPE,
        EcKeyGenParams: base_config_1.TYPE,
        EcKeyImportParams: base_config_1.TYPE,
        EcdhKeyDeriveParams: base_config_1.TYPE,
        EcdsaParams: base_config_1.TYPE,
        EffectTiming: base_config_1.TYPE,
        ElementCreationOptions: base_config_1.TYPE,
        ElementDefinitionOptions: base_config_1.TYPE,
        ErrorEventInit: base_config_1.TYPE,
        EventInit: base_config_1.TYPE,
        EventListenerOptions: base_config_1.TYPE,
        EventModifierInit: base_config_1.TYPE,
        EventSourceInit: base_config_1.TYPE,
        ExceptionInformation: base_config_1.TYPE,
        FilePropertyBag: base_config_1.TYPE,
        FocusEventInit: base_config_1.TYPE,
        FocusNavigationEventInit: base_config_1.TYPE,
        FocusNavigationOrigin: base_config_1.TYPE,
        FocusOptions: base_config_1.TYPE,
        FullscreenOptions: base_config_1.TYPE,
        GainOptions: base_config_1.TYPE,
        GamepadEventInit: base_config_1.TYPE,
        GetNotificationOptions: base_config_1.TYPE,
        GetRootNodeOptions: base_config_1.TYPE,
        HashChangeEventInit: base_config_1.TYPE,
        HkdfParams: base_config_1.TYPE,
        HmacImportParams: base_config_1.TYPE,
        HmacKeyAlgorithm: base_config_1.TYPE,
        HmacKeyGenParams: base_config_1.TYPE,
        IDBIndexParameters: base_config_1.TYPE,
        IDBObjectStoreParameters: base_config_1.TYPE,
        IDBVersionChangeEventInit: base_config_1.TYPE,
        IIRFilterOptions: base_config_1.TYPE,
        ImageBitmapOptions: base_config_1.TYPE,
        ImageBitmapRenderingContextSettings: base_config_1.TYPE,
        ImageEncodeOptions: base_config_1.TYPE,
        ImportMeta: base_config_1.TYPE,
        InputEventInit: base_config_1.TYPE,
        IntersectionObserverEntryInit: base_config_1.TYPE,
        IntersectionObserverInit: base_config_1.TYPE,
        JsonWebKey: base_config_1.TYPE,
        KeyAlgorithm: base_config_1.TYPE,
        KeyboardEventInit: base_config_1.TYPE,
        Keyframe: base_config_1.TYPE,
        KeyframeAnimationOptions: base_config_1.TYPE,
        KeyframeEffectOptions: base_config_1.TYPE,
        MediaElementAudioSourceOptions: base_config_1.TYPE,
        MediaEncryptedEventInit: base_config_1.TYPE,
        MediaKeyMessageEventInit: base_config_1.TYPE,
        MediaKeySystemConfiguration: base_config_1.TYPE,
        MediaKeySystemMediaCapability: base_config_1.TYPE,
        MediaQueryListEventInit: base_config_1.TYPE,
        MediaStreamAudioSourceOptions: base_config_1.TYPE,
        MediaStreamConstraints: base_config_1.TYPE,
        MediaStreamErrorEventInit: base_config_1.TYPE,
        MediaStreamEventInit: base_config_1.TYPE,
        MediaStreamTrackAudioSourceOptions: base_config_1.TYPE,
        MediaStreamTrackEventInit: base_config_1.TYPE,
        MediaTrackCapabilities: base_config_1.TYPE,
        MediaTrackConstraintSet: base_config_1.TYPE,
        MediaTrackConstraints: base_config_1.TYPE,
        MediaTrackSettings: base_config_1.TYPE,
        MediaTrackSupportedConstraints: base_config_1.TYPE,
        MessageEventInit: base_config_1.TYPE,
        MidiPermissionDescriptor: base_config_1.TYPE,
        MouseEventInit: base_config_1.TYPE,
        MultiCacheQueryOptions: base_config_1.TYPE,
        MutationObserverInit: base_config_1.TYPE,
        NavigationPreloadState: base_config_1.TYPE,
        NotificationAction: base_config_1.TYPE,
        NotificationOptions: base_config_1.TYPE,
        OfflineAudioCompletionEventInit: base_config_1.TYPE,
        OfflineAudioContextOptions: base_config_1.TYPE,
        OptionalEffectTiming: base_config_1.TYPE,
        OscillatorOptions: base_config_1.TYPE,
        PageTransitionEventInit: base_config_1.TYPE,
        PannerOptions: base_config_1.TYPE,
        PayerErrors: base_config_1.TYPE,
        PaymentCurrencyAmount: base_config_1.TYPE,
        PaymentDetailsBase: base_config_1.TYPE,
        PaymentDetailsInit: base_config_1.TYPE,
        PaymentDetailsModifier: base_config_1.TYPE,
        PaymentDetailsUpdate: base_config_1.TYPE,
        PaymentItem: base_config_1.TYPE,
        PaymentMethodChangeEventInit: base_config_1.TYPE,
        PaymentMethodData: base_config_1.TYPE,
        PaymentOptions: base_config_1.TYPE,
        PaymentRequestUpdateEventInit: base_config_1.TYPE,
        PaymentShippingOption: base_config_1.TYPE,
        PaymentValidationErrors: base_config_1.TYPE,
        Pbkdf2Params: base_config_1.TYPE,
        PerformanceObserverInit: base_config_1.TYPE,
        PeriodicWaveConstraints: base_config_1.TYPE,
        PeriodicWaveOptions: base_config_1.TYPE,
        PermissionDescriptor: base_config_1.TYPE,
        PointerEventInit: base_config_1.TYPE,
        PopStateEventInit: base_config_1.TYPE,
        PositionOptions: base_config_1.TYPE,
        PostMessageOptions: base_config_1.TYPE,
        ProgressEventInit: base_config_1.TYPE,
        PromiseRejectionEventInit: base_config_1.TYPE,
        PropertyIndexedKeyframes: base_config_1.TYPE,
        PublicKeyCredentialCreationOptions: base_config_1.TYPE,
        PublicKeyCredentialDescriptor: base_config_1.TYPE,
        PublicKeyCredentialEntity: base_config_1.TYPE,
        PublicKeyCredentialParameters: base_config_1.TYPE,
        PublicKeyCredentialRequestOptions: base_config_1.TYPE,
        PublicKeyCredentialRpEntity: base_config_1.TYPE,
        PublicKeyCredentialUserEntity: base_config_1.TYPE,
        PushPermissionDescriptor: base_config_1.TYPE,
        PushSubscriptionJSON: base_config_1.TYPE,
        PushSubscriptionOptionsInit: base_config_1.TYPE,
        QueuingStrategy: base_config_1.TYPE,
        QueuingStrategyInit: base_config_1.TYPE,
        RTCAnswerOptions: base_config_1.TYPE,
        RTCCertificateExpiration: base_config_1.TYPE,
        RTCConfiguration: base_config_1.TYPE,
        RTCDTMFToneChangeEventInit: base_config_1.TYPE,
        RTCDataChannelEventInit: base_config_1.TYPE,
        RTCDataChannelInit: base_config_1.TYPE,
        RTCDtlsFingerprint: base_config_1.TYPE,
        RTCDtlsParameters: base_config_1.TYPE,
        RTCErrorEventInit: base_config_1.TYPE,
        RTCErrorInit: base_config_1.TYPE,
        RTCIceCandidateAttributes: base_config_1.TYPE,
        RTCIceCandidateComplete: base_config_1.TYPE,
        RTCIceCandidateDictionary: base_config_1.TYPE,
        RTCIceCandidateInit: base_config_1.TYPE,
        RTCIceCandidatePair: base_config_1.TYPE,
        RTCIceCandidatePairStats: base_config_1.TYPE,
        RTCIceGatherOptions: base_config_1.TYPE,
        RTCIceParameters: base_config_1.TYPE,
        RTCIceServer: base_config_1.TYPE,
        RTCIdentityProviderOptions: base_config_1.TYPE,
        RTCInboundRTPStreamStats: base_config_1.TYPE,
        RTCLocalSessionDescriptionInit: base_config_1.TYPE,
        RTCMediaStreamTrackStats: base_config_1.TYPE,
        RTCOfferAnswerOptions: base_config_1.TYPE,
        RTCOfferOptions: base_config_1.TYPE,
        RTCOutboundRTPStreamStats: base_config_1.TYPE,
        RTCPeerConnectionIceErrorEventInit: base_config_1.TYPE,
        RTCPeerConnectionIceEventInit: base_config_1.TYPE,
        RTCRTPStreamStats: base_config_1.TYPE,
        RTCRtcpFeedback: base_config_1.TYPE,
        RTCRtcpParameters: base_config_1.TYPE,
        RTCRtpCapabilities: base_config_1.TYPE,
        RTCRtpCodecCapability: base_config_1.TYPE,
        RTCRtpCodecParameters: base_config_1.TYPE,
        RTCRtpCodingParameters: base_config_1.TYPE,
        RTCRtpContributingSource: base_config_1.TYPE,
        RTCRtpEncodingParameters: base_config_1.TYPE,
        RTCRtpFecParameters: base_config_1.TYPE,
        RTCRtpHeaderExtension: base_config_1.TYPE,
        RTCRtpHeaderExtensionCapability: base_config_1.TYPE,
        RTCRtpHeaderExtensionParameters: base_config_1.TYPE,
        RTCRtpParameters: base_config_1.TYPE,
        RTCRtpReceiveParameters: base_config_1.TYPE,
        RTCRtpRtxParameters: base_config_1.TYPE,
        RTCRtpSendParameters: base_config_1.TYPE,
        RTCRtpSynchronizationSource: base_config_1.TYPE,
        RTCRtpTransceiverInit: base_config_1.TYPE,
        RTCRtpUnhandled: base_config_1.TYPE,
        RTCSessionDescriptionInit: base_config_1.TYPE,
        RTCSrtpKeyParam: base_config_1.TYPE,
        RTCSrtpSdesParameters: base_config_1.TYPE,
        RTCSsrcRange: base_config_1.TYPE,
        RTCStats: base_config_1.TYPE,
        RTCStatsReport: base_config_1.TYPE_VALUE,
        RTCTrackEventInit: base_config_1.TYPE,
        RTCTransportStats: base_config_1.TYPE,
        ReadableStreamDefaultReadDoneResult: base_config_1.TYPE,
        ReadableStreamDefaultReadValueResult: base_config_1.TYPE,
        ReadableWritablePair: base_config_1.TYPE,
        RegistrationOptions: base_config_1.TYPE,
        RequestInit: base_config_1.TYPE,
        ResizeObserverOptions: base_config_1.TYPE,
        ResponseInit: base_config_1.TYPE,
        RsaHashedImportParams: base_config_1.TYPE,
        RsaHashedKeyAlgorithm: base_config_1.TYPE,
        RsaHashedKeyGenParams: base_config_1.TYPE,
        RsaKeyAlgorithm: base_config_1.TYPE,
        RsaKeyGenParams: base_config_1.TYPE,
        RsaOaepParams: base_config_1.TYPE,
        RsaOtherPrimesInfo: base_config_1.TYPE,
        RsaPssParams: base_config_1.TYPE,
        SVGBoundingBoxOptions: base_config_1.TYPE,
        ScrollIntoViewOptions: base_config_1.TYPE,
        ScrollOptions: base_config_1.TYPE,
        ScrollToOptions: base_config_1.TYPE,
        SecurityPolicyViolationEventInit: base_config_1.TYPE,
        ShadowRootInit: base_config_1.TYPE,
        ShareData: base_config_1.TYPE,
        SpeechRecognitionErrorEventInit: base_config_1.TYPE,
        SpeechRecognitionEventInit: base_config_1.TYPE,
        SpeechSynthesisErrorEventInit: base_config_1.TYPE,
        SpeechSynthesisEventInit: base_config_1.TYPE,
        StaticRangeInit: base_config_1.TYPE,
        StereoPannerOptions: base_config_1.TYPE,
        StorageEstimate: base_config_1.TYPE,
        StorageEventInit: base_config_1.TYPE,
        StoreExceptionsInformation: base_config_1.TYPE,
        StoreSiteSpecificExceptionsInformation: base_config_1.TYPE,
        StreamPipeOptions: base_config_1.TYPE,
        TextDecodeOptions: base_config_1.TYPE,
        TextDecoderOptions: base_config_1.TYPE,
        TextEncoderEncodeIntoResult: base_config_1.TYPE,
        TouchEventInit: base_config_1.TYPE,
        TouchInit: base_config_1.TYPE,
        TrackEventInit: base_config_1.TYPE,
        Transformer: base_config_1.TYPE,
        TransitionEventInit: base_config_1.TYPE,
        UIEventInit: base_config_1.TYPE,
        ULongRange: base_config_1.TYPE,
        UnderlyingSink: base_config_1.TYPE,
        UnderlyingSource: base_config_1.TYPE,
        VRDisplayEventInit: base_config_1.TYPE,
        VRLayer: base_config_1.TYPE,
        VRStageParameters: base_config_1.TYPE,
        WaveShaperOptions: base_config_1.TYPE,
        WebGLContextAttributes: base_config_1.TYPE,
        WebGLContextEventInit: base_config_1.TYPE,
        WheelEventInit: base_config_1.TYPE,
        WorkerOptions: base_config_1.TYPE,
        WorkletOptions: base_config_1.TYPE,
        EventListener: base_config_1.TYPE,
        XPathNSResolver: base_config_1.TYPE,
        ANGLE_instanced_arrays: base_config_1.TYPE,
        AbortController: base_config_1.TYPE_VALUE,
        AbortSignalEventMap: base_config_1.TYPE,
        AbortSignal: base_config_1.TYPE_VALUE,
        AbstractRange: base_config_1.TYPE_VALUE,
        AbstractWorkerEventMap: base_config_1.TYPE,
        AbstractWorker: base_config_1.TYPE,
        AesCfbParams: base_config_1.TYPE,
        AesCmacParams: base_config_1.TYPE,
        AnalyserNode: base_config_1.TYPE_VALUE,
        Animatable: base_config_1.TYPE,
        AnimationEventMap: base_config_1.TYPE,
        Animation: base_config_1.TYPE_VALUE,
        AnimationEffect: base_config_1.TYPE_VALUE,
        AnimationEvent: base_config_1.TYPE_VALUE,
        AnimationFrameProvider: base_config_1.TYPE,
        AnimationPlaybackEvent: base_config_1.TYPE_VALUE,
        AnimationTimeline: base_config_1.TYPE_VALUE,
        ApplicationCacheEventMap: base_config_1.TYPE,
        ApplicationCache: base_config_1.TYPE_VALUE,
        Attr: base_config_1.TYPE_VALUE,
        AudioBuffer: base_config_1.TYPE_VALUE,
        AudioBufferSourceNode: base_config_1.TYPE_VALUE,
        AudioContext: base_config_1.TYPE_VALUE,
        AudioDestinationNode: base_config_1.TYPE_VALUE,
        AudioListener: base_config_1.TYPE_VALUE,
        AudioNode: base_config_1.TYPE_VALUE,
        AudioParam: base_config_1.TYPE_VALUE,
        AudioParamMap: base_config_1.TYPE_VALUE,
        AudioProcessingEvent: base_config_1.TYPE_VALUE,
        AudioScheduledSourceNodeEventMap: base_config_1.TYPE,
        AudioScheduledSourceNode: base_config_1.TYPE_VALUE,
        AudioWorklet: base_config_1.TYPE_VALUE,
        AudioWorkletNodeEventMap: base_config_1.TYPE,
        AudioWorkletNode: base_config_1.TYPE_VALUE,
        AuthenticatorAssertionResponse: base_config_1.TYPE_VALUE,
        AuthenticatorAttestationResponse: base_config_1.TYPE_VALUE,
        AuthenticatorResponse: base_config_1.TYPE_VALUE,
        BarProp: base_config_1.TYPE_VALUE,
        BaseAudioContextEventMap: base_config_1.TYPE,
        BaseAudioContext: base_config_1.TYPE_VALUE,
        BeforeUnloadEvent: base_config_1.TYPE_VALUE,
        BhxBrowser: base_config_1.TYPE_VALUE,
        BiquadFilterNode: base_config_1.TYPE_VALUE,
        Blob: base_config_1.TYPE_VALUE,
        Body: base_config_1.TYPE,
        BroadcastChannelEventMap: base_config_1.TYPE,
        BroadcastChannel: base_config_1.TYPE_VALUE,
        ByteLengthQueuingStrategy: base_config_1.TYPE_VALUE,
        CDATASection: base_config_1.TYPE_VALUE,
        CSSConditionRule: base_config_1.TYPE_VALUE,
        CSSFontFaceRule: base_config_1.TYPE_VALUE,
        CSSGroupingRule: base_config_1.TYPE_VALUE,
        CSSImportRule: base_config_1.TYPE_VALUE,
        CSSKeyframeRule: base_config_1.TYPE_VALUE,
        CSSKeyframesRule: base_config_1.TYPE_VALUE,
        CSSMediaRule: base_config_1.TYPE_VALUE,
        CSSNamespaceRule: base_config_1.TYPE_VALUE,
        CSSPageRule: base_config_1.TYPE_VALUE,
        CSSRule: base_config_1.TYPE_VALUE,
        CSSRuleList: base_config_1.TYPE_VALUE,
        CSSStyleDeclaration: base_config_1.TYPE_VALUE,
        CSSStyleRule: base_config_1.TYPE_VALUE,
        CSSStyleSheet: base_config_1.TYPE_VALUE,
        CSSSupportsRule: base_config_1.TYPE_VALUE,
        Cache: base_config_1.TYPE_VALUE,
        CacheStorage: base_config_1.TYPE_VALUE,
        CanvasCompositing: base_config_1.TYPE,
        CanvasDrawImage: base_config_1.TYPE,
        CanvasDrawPath: base_config_1.TYPE,
        CanvasFillStrokeStyles: base_config_1.TYPE,
        CanvasFilters: base_config_1.TYPE,
        CanvasGradient: base_config_1.TYPE_VALUE,
        CanvasImageData: base_config_1.TYPE,
        CanvasImageSmoothing: base_config_1.TYPE,
        CanvasPath: base_config_1.TYPE,
        CanvasPathDrawingStyles: base_config_1.TYPE,
        CanvasPattern: base_config_1.TYPE_VALUE,
        CanvasRect: base_config_1.TYPE,
        CanvasRenderingContext2D: base_config_1.TYPE_VALUE,
        CanvasShadowStyles: base_config_1.TYPE,
        CanvasState: base_config_1.TYPE,
        CanvasText: base_config_1.TYPE,
        CanvasTextDrawingStyles: base_config_1.TYPE,
        CanvasTransform: base_config_1.TYPE,
        CanvasUserInterface: base_config_1.TYPE,
        CaretPosition: base_config_1.TYPE_VALUE,
        ChannelMergerNode: base_config_1.TYPE_VALUE,
        ChannelSplitterNode: base_config_1.TYPE_VALUE,
        CharacterData: base_config_1.TYPE_VALUE,
        ChildNode: base_config_1.TYPE,
        ClientRect: base_config_1.TYPE_VALUE,
        ClientRectList: base_config_1.TYPE_VALUE,
        Clipboard: base_config_1.TYPE_VALUE,
        ClipboardEvent: base_config_1.TYPE_VALUE,
        CloseEvent: base_config_1.TYPE_VALUE,
        Comment: base_config_1.TYPE_VALUE,
        CompositionEvent: base_config_1.TYPE_VALUE,
        ConcatParams: base_config_1.TYPE,
        ConstantSourceNode: base_config_1.TYPE_VALUE,
        ConvolverNode: base_config_1.TYPE_VALUE,
        CountQueuingStrategy: base_config_1.TYPE_VALUE,
        Credential: base_config_1.TYPE_VALUE,
        CredentialsContainer: base_config_1.TYPE_VALUE,
        Crypto: base_config_1.TYPE_VALUE,
        CryptoKey: base_config_1.TYPE_VALUE,
        CryptoKeyPair: base_config_1.TYPE_VALUE,
        CustomElementRegistry: base_config_1.TYPE_VALUE,
        CustomEvent: base_config_1.TYPE_VALUE,
        DOMError: base_config_1.TYPE_VALUE,
        DOMException: base_config_1.TYPE_VALUE,
        DOMImplementation: base_config_1.TYPE_VALUE,
        DOMMatrix: base_config_1.TYPE_VALUE,
        SVGMatrix: base_config_1.TYPE_VALUE,
        WebKitCSSMatrix: base_config_1.TYPE_VALUE,
        DOMMatrixReadOnly: base_config_1.TYPE_VALUE,
        DOMParser: base_config_1.TYPE_VALUE,
        DOMPoint: base_config_1.TYPE_VALUE,
        SVGPoint: base_config_1.TYPE_VALUE,
        DOMPointReadOnly: base_config_1.TYPE_VALUE,
        DOMQuad: base_config_1.TYPE_VALUE,
        DOMRect: base_config_1.TYPE_VALUE,
        SVGRect: base_config_1.TYPE_VALUE,
        DOMRectList: base_config_1.TYPE_VALUE,
        DOMRectReadOnly: base_config_1.TYPE_VALUE,
        DOMSettableTokenList: base_config_1.TYPE_VALUE,
        DOMStringList: base_config_1.TYPE_VALUE,
        DOMStringMap: base_config_1.TYPE_VALUE,
        DOMTokenList: base_config_1.TYPE_VALUE,
        DataCue: base_config_1.TYPE_VALUE,
        DataTransfer: base_config_1.TYPE_VALUE,
        DataTransferItem: base_config_1.TYPE_VALUE,
        DataTransferItemList: base_config_1.TYPE_VALUE,
        DeferredPermissionRequest: base_config_1.TYPE_VALUE,
        DelayNode: base_config_1.TYPE_VALUE,
        DeviceAcceleration: base_config_1.TYPE_VALUE,
        DeviceMotionEvent: base_config_1.TYPE_VALUE,
        DeviceMotionEventAcceleration: base_config_1.TYPE,
        DeviceMotionEventRotationRate: base_config_1.TYPE,
        DeviceOrientationEvent: base_config_1.TYPE_VALUE,
        DeviceRotationRate: base_config_1.TYPE_VALUE,
        DhImportKeyParams: base_config_1.TYPE,
        DhKeyAlgorithm: base_config_1.TYPE,
        DhKeyDeriveParams: base_config_1.TYPE,
        DhKeyGenParams: base_config_1.TYPE,
        DocumentEventMap: base_config_1.TYPE,
        Document: base_config_1.TYPE_VALUE,
        DocumentAndElementEventHandlersEventMap: base_config_1.TYPE,
        DocumentAndElementEventHandlers: base_config_1.TYPE,
        DocumentEvent: base_config_1.TYPE,
        DocumentFragment: base_config_1.TYPE_VALUE,
        DocumentOrShadowRoot: base_config_1.TYPE,
        DocumentTimeline: base_config_1.TYPE_VALUE,
        DocumentType: base_config_1.TYPE_VALUE,
        DragEvent: base_config_1.TYPE_VALUE,
        DynamicsCompressorNode: base_config_1.TYPE_VALUE,
        EXT_blend_minmax: base_config_1.TYPE,
        EXT_frag_depth: base_config_1.TYPE,
        EXT_sRGB: base_config_1.TYPE,
        EXT_shader_texture_lod: base_config_1.TYPE,
        EXT_texture_filter_anisotropic: base_config_1.TYPE,
        ElementEventMap: base_config_1.TYPE,
        Element: base_config_1.TYPE_VALUE,
        ElementCSSInlineStyle: base_config_1.TYPE,
        ElementContentEditable: base_config_1.TYPE,
        ErrorEvent: base_config_1.TYPE_VALUE,
        Event: base_config_1.TYPE_VALUE,
        EventListenerObject: base_config_1.TYPE,
        EventSourceEventMap: base_config_1.TYPE,
        EventSource: base_config_1.TYPE_VALUE,
        EventTarget: base_config_1.TYPE_VALUE,
        ExtensionScriptApis: base_config_1.TYPE_VALUE,
        External: base_config_1.TYPE_VALUE,
        File: base_config_1.TYPE_VALUE,
        FileList: base_config_1.TYPE_VALUE,
        FileReaderEventMap: base_config_1.TYPE,
        FileReader: base_config_1.TYPE_VALUE,
        FocusEvent: base_config_1.TYPE_VALUE,
        FocusNavigationEvent: base_config_1.TYPE_VALUE,
        FormData: base_config_1.TYPE_VALUE,
        GainNode: base_config_1.TYPE_VALUE,
        Gamepad: base_config_1.TYPE_VALUE,
        GamepadButton: base_config_1.TYPE_VALUE,
        GamepadEvent: base_config_1.TYPE_VALUE,
        GamepadHapticActuator: base_config_1.TYPE_VALUE,
        GamepadPose: base_config_1.TYPE_VALUE,
        GenericTransformStream: base_config_1.TYPE,
        Geolocation: base_config_1.TYPE_VALUE,
        GeolocationCoordinates: base_config_1.TYPE_VALUE,
        GeolocationPosition: base_config_1.TYPE_VALUE,
        GeolocationPositionError: base_config_1.TYPE_VALUE,
        GlobalEventHandlersEventMap: base_config_1.TYPE,
        GlobalEventHandlers: base_config_1.TYPE,
        HTMLAllCollection: base_config_1.TYPE_VALUE,
        HTMLAnchorElement: base_config_1.TYPE_VALUE,
        HTMLAppletElement: base_config_1.TYPE_VALUE,
        HTMLAreaElement: base_config_1.TYPE_VALUE,
        HTMLAudioElement: base_config_1.TYPE_VALUE,
        HTMLBRElement: base_config_1.TYPE_VALUE,
        HTMLBaseElement: base_config_1.TYPE_VALUE,
        HTMLBodyElementEventMap: base_config_1.TYPE,
        HTMLBodyElement: base_config_1.TYPE_VALUE,
        HTMLButtonElement: base_config_1.TYPE_VALUE,
        HTMLCanvasElement: base_config_1.TYPE_VALUE,
        HTMLCollectionBase: base_config_1.TYPE,
        HTMLCollection: base_config_1.TYPE_VALUE,
        HTMLCollectionOf: base_config_1.TYPE,
        HTMLDListElement: base_config_1.TYPE_VALUE,
        HTMLDataElement: base_config_1.TYPE_VALUE,
        HTMLDataListElement: base_config_1.TYPE_VALUE,
        HTMLDetailsElement: base_config_1.TYPE_VALUE,
        HTMLDialogElement: base_config_1.TYPE_VALUE,
        HTMLDirectoryElement: base_config_1.TYPE_VALUE,
        HTMLDivElement: base_config_1.TYPE_VALUE,
        HTMLDocument: base_config_1.TYPE_VALUE,
        HTMLElementEventMap: base_config_1.TYPE,
        HTMLElement: base_config_1.TYPE_VALUE,
        HTMLEmbedElement: base_config_1.TYPE_VALUE,
        HTMLFieldSetElement: base_config_1.TYPE_VALUE,
        HTMLFontElement: base_config_1.TYPE_VALUE,
        HTMLFormControlsCollection: base_config_1.TYPE_VALUE,
        HTMLFormElement: base_config_1.TYPE_VALUE,
        HTMLFrameElement: base_config_1.TYPE_VALUE,
        HTMLFrameSetElementEventMap: base_config_1.TYPE,
        HTMLFrameSetElement: base_config_1.TYPE_VALUE,
        HTMLHRElement: base_config_1.TYPE_VALUE,
        HTMLHeadElement: base_config_1.TYPE_VALUE,
        HTMLHeadingElement: base_config_1.TYPE_VALUE,
        HTMLHtmlElement: base_config_1.TYPE_VALUE,
        HTMLHyperlinkElementUtils: base_config_1.TYPE,
        HTMLIFrameElement: base_config_1.TYPE_VALUE,
        HTMLImageElement: base_config_1.TYPE_VALUE,
        HTMLInputElement: base_config_1.TYPE_VALUE,
        HTMLLIElement: base_config_1.TYPE_VALUE,
        HTMLLabelElement: base_config_1.TYPE_VALUE,
        HTMLLegendElement: base_config_1.TYPE_VALUE,
        HTMLLinkElement: base_config_1.TYPE_VALUE,
        HTMLMapElement: base_config_1.TYPE_VALUE,
        HTMLMarqueeElementEventMap: base_config_1.TYPE,
        HTMLMarqueeElement: base_config_1.TYPE_VALUE,
        HTMLMediaElementEventMap: base_config_1.TYPE,
        HTMLMediaElement: base_config_1.TYPE_VALUE,
        HTMLMenuElement: base_config_1.TYPE_VALUE,
        HTMLMetaElement: base_config_1.TYPE_VALUE,
        HTMLMeterElement: base_config_1.TYPE_VALUE,
        HTMLModElement: base_config_1.TYPE_VALUE,
        HTMLOListElement: base_config_1.TYPE_VALUE,
        HTMLObjectElement: base_config_1.TYPE_VALUE,
        HTMLOptGroupElement: base_config_1.TYPE_VALUE,
        HTMLOptionElement: base_config_1.TYPE_VALUE,
        HTMLOptionsCollection: base_config_1.TYPE_VALUE,
        HTMLOrSVGElement: base_config_1.TYPE,
        HTMLOutputElement: base_config_1.TYPE_VALUE,
        HTMLParagraphElement: base_config_1.TYPE_VALUE,
        HTMLParamElement: base_config_1.TYPE_VALUE,
        HTMLPictureElement: base_config_1.TYPE_VALUE,
        HTMLPreElement: base_config_1.TYPE_VALUE,
        HTMLProgressElement: base_config_1.TYPE_VALUE,
        HTMLQuoteElement: base_config_1.TYPE_VALUE,
        HTMLScriptElement: base_config_1.TYPE_VALUE,
        HTMLSelectElement: base_config_1.TYPE_VALUE,
        HTMLSlotElement: base_config_1.TYPE_VALUE,
        HTMLSourceElement: base_config_1.TYPE_VALUE,
        HTMLSpanElement: base_config_1.TYPE_VALUE,
        HTMLStyleElement: base_config_1.TYPE_VALUE,
        HTMLTableCaptionElement: base_config_1.TYPE_VALUE,
        HTMLTableCellElement: base_config_1.TYPE_VALUE,
        HTMLTableColElement: base_config_1.TYPE_VALUE,
        HTMLTableDataCellElement: base_config_1.TYPE_VALUE,
        HTMLTableElement: base_config_1.TYPE_VALUE,
        HTMLTableHeaderCellElement: base_config_1.TYPE_VALUE,
        HTMLTableRowElement: base_config_1.TYPE_VALUE,
        HTMLTableSectionElement: base_config_1.TYPE_VALUE,
        HTMLTemplateElement: base_config_1.TYPE_VALUE,
        HTMLTextAreaElement: base_config_1.TYPE_VALUE,
        HTMLTimeElement: base_config_1.TYPE_VALUE,
        HTMLTitleElement: base_config_1.TYPE_VALUE,
        HTMLTrackElement: base_config_1.TYPE_VALUE,
        HTMLUListElement: base_config_1.TYPE_VALUE,
        HTMLUnknownElement: base_config_1.TYPE_VALUE,
        HTMLVideoElement: base_config_1.TYPE_VALUE,
        HashChangeEvent: base_config_1.TYPE_VALUE,
        Headers: base_config_1.TYPE_VALUE,
        History: base_config_1.TYPE_VALUE,
        IDBArrayKey: base_config_1.TYPE,
        IDBCursor: base_config_1.TYPE_VALUE,
        IDBCursorWithValue: base_config_1.TYPE_VALUE,
        IDBDatabaseEventMap: base_config_1.TYPE,
        IDBDatabase: base_config_1.TYPE_VALUE,
        IDBFactory: base_config_1.TYPE_VALUE,
        IDBIndex: base_config_1.TYPE_VALUE,
        IDBKeyRange: base_config_1.TYPE_VALUE,
        IDBObjectStore: base_config_1.TYPE_VALUE,
        IDBOpenDBRequestEventMap: base_config_1.TYPE,
        IDBOpenDBRequest: base_config_1.TYPE_VALUE,
        IDBRequestEventMap: base_config_1.TYPE,
        IDBRequest: base_config_1.TYPE_VALUE,
        IDBTransactionEventMap: base_config_1.TYPE,
        IDBTransaction: base_config_1.TYPE_VALUE,
        IDBVersionChangeEvent: base_config_1.TYPE_VALUE,
        IIRFilterNode: base_config_1.TYPE_VALUE,
        ImageBitmap: base_config_1.TYPE_VALUE,
        ImageBitmapRenderingContext: base_config_1.TYPE_VALUE,
        ImageData: base_config_1.TYPE_VALUE,
        InnerHTML: base_config_1.TYPE,
        InputDeviceInfo: base_config_1.TYPE_VALUE,
        InputEvent: base_config_1.TYPE_VALUE,
        IntersectionObserver: base_config_1.TYPE_VALUE,
        IntersectionObserverEntry: base_config_1.TYPE_VALUE,
        KeyboardEvent: base_config_1.TYPE_VALUE,
        KeyframeEffect: base_config_1.TYPE_VALUE,
        LinkStyle: base_config_1.TYPE,
        ListeningStateChangedEvent: base_config_1.TYPE_VALUE,
        Location: base_config_1.TYPE_VALUE,
        MSAssertion: base_config_1.TYPE_VALUE,
        MSBlobBuilder: base_config_1.TYPE_VALUE,
        MSFIDOCredentialAssertion: base_config_1.TYPE_VALUE,
        MSFIDOSignature: base_config_1.TYPE_VALUE,
        MSFIDOSignatureAssertion: base_config_1.TYPE_VALUE,
        MSFileSaver: base_config_1.TYPE,
        MSGesture: base_config_1.TYPE_VALUE,
        MSGraphicsTrust: base_config_1.TYPE_VALUE,
        MSInputMethodContextEventMap: base_config_1.TYPE,
        MSInputMethodContext: base_config_1.TYPE_VALUE,
        MSMediaKeyError: base_config_1.TYPE_VALUE,
        MSMediaKeySession: base_config_1.TYPE_VALUE,
        MSMediaKeys: base_config_1.TYPE_VALUE,
        MSNavigatorDoNotTrack: base_config_1.TYPE,
        MSStream: base_config_1.TYPE_VALUE,
        MediaDeviceInfo: base_config_1.TYPE_VALUE,
        MediaDevicesEventMap: base_config_1.TYPE,
        MediaDevices: base_config_1.TYPE_VALUE,
        MediaElementAudioSourceNode: base_config_1.TYPE_VALUE,
        MediaEncryptedEvent: base_config_1.TYPE_VALUE,
        MediaError: base_config_1.TYPE_VALUE,
        MediaKeyMessageEvent: base_config_1.TYPE_VALUE,
        MediaKeySessionEventMap: base_config_1.TYPE,
        MediaKeySession: base_config_1.TYPE_VALUE,
        MediaKeyStatusMap: base_config_1.TYPE_VALUE,
        MediaKeySystemAccess: base_config_1.TYPE_VALUE,
        MediaKeys: base_config_1.TYPE_VALUE,
        MediaList: base_config_1.TYPE_VALUE,
        MediaQueryListEventMap: base_config_1.TYPE,
        MediaQueryList: base_config_1.TYPE_VALUE,
        MediaQueryListEvent: base_config_1.TYPE_VALUE,
        MediaSourceEventMap: base_config_1.TYPE,
        MediaSource: base_config_1.TYPE_VALUE,
        MediaStreamEventMap: base_config_1.TYPE,
        MediaStream: base_config_1.TYPE_VALUE,
        MediaStreamAudioDestinationNode: base_config_1.TYPE_VALUE,
        MediaStreamAudioSourceNode: base_config_1.TYPE_VALUE,
        MediaStreamError: base_config_1.TYPE_VALUE,
        MediaStreamErrorEvent: base_config_1.TYPE_VALUE,
        MediaStreamEvent: base_config_1.TYPE_VALUE,
        MediaStreamTrackEventMap: base_config_1.TYPE,
        MediaStreamTrack: base_config_1.TYPE_VALUE,
        MediaStreamTrackAudioSourceNode: base_config_1.TYPE_VALUE,
        MediaStreamTrackEvent: base_config_1.TYPE_VALUE,
        MessageChannel: base_config_1.TYPE_VALUE,
        MessageEvent: base_config_1.TYPE_VALUE,
        MessagePortEventMap: base_config_1.TYPE,
        MessagePort: base_config_1.TYPE_VALUE,
        MimeType: base_config_1.TYPE_VALUE,
        MimeTypeArray: base_config_1.TYPE_VALUE,
        MouseEvent: base_config_1.TYPE_VALUE,
        MutationEvent: base_config_1.TYPE_VALUE,
        MutationObserver: base_config_1.TYPE_VALUE,
        MutationRecord: base_config_1.TYPE_VALUE,
        NamedNodeMap: base_config_1.TYPE_VALUE,
        NavigationPreloadManager: base_config_1.TYPE_VALUE,
        Navigator: base_config_1.TYPE_VALUE,
        NavigatorAutomationInformation: base_config_1.TYPE,
        NavigatorBeacon: base_config_1.TYPE,
        NavigatorConcurrentHardware: base_config_1.TYPE,
        NavigatorContentUtils: base_config_1.TYPE,
        NavigatorCookies: base_config_1.TYPE,
        NavigatorID: base_config_1.TYPE,
        NavigatorLanguage: base_config_1.TYPE,
        NavigatorOnLine: base_config_1.TYPE,
        NavigatorPlugins: base_config_1.TYPE,
        NavigatorStorage: base_config_1.TYPE,
        Node: base_config_1.TYPE_VALUE,
        NodeFilter: base_config_1.TYPE_VALUE,
        NodeIterator: base_config_1.TYPE_VALUE,
        NodeList: base_config_1.TYPE_VALUE,
        NodeListOf: base_config_1.TYPE,
        NonDocumentTypeChildNode: base_config_1.TYPE,
        NonElementParentNode: base_config_1.TYPE,
        NotificationEventMap: base_config_1.TYPE,
        Notification: base_config_1.TYPE_VALUE,
        OES_element_index_uint: base_config_1.TYPE,
        OES_standard_derivatives: base_config_1.TYPE,
        OES_texture_float: base_config_1.TYPE,
        OES_texture_float_linear: base_config_1.TYPE,
        OES_texture_half_float: base_config_1.TYPE,
        OES_texture_half_float_linear: base_config_1.TYPE,
        OES_vertex_array_object: base_config_1.TYPE,
        OfflineAudioCompletionEvent: base_config_1.TYPE_VALUE,
        OfflineAudioContextEventMap: base_config_1.TYPE,
        OfflineAudioContext: base_config_1.TYPE_VALUE,
        OffscreenCanvas: base_config_1.TYPE_VALUE,
        OffscreenCanvasRenderingContext2D: base_config_1.TYPE_VALUE,
        OscillatorNode: base_config_1.TYPE_VALUE,
        OverconstrainedError: base_config_1.TYPE_VALUE,
        OverflowEvent: base_config_1.TYPE_VALUE,
        PageTransitionEvent: base_config_1.TYPE_VALUE,
        PannerNode: base_config_1.TYPE_VALUE,
        ParentNode: base_config_1.TYPE,
        Path2D: base_config_1.TYPE_VALUE,
        PaymentAddress: base_config_1.TYPE_VALUE,
        PaymentMethodChangeEvent: base_config_1.TYPE_VALUE,
        PaymentRequestEventMap: base_config_1.TYPE,
        PaymentRequest: base_config_1.TYPE_VALUE,
        PaymentRequestUpdateEvent: base_config_1.TYPE_VALUE,
        PaymentResponseEventMap: base_config_1.TYPE,
        PaymentResponse: base_config_1.TYPE_VALUE,
        PerfWidgetExternal: base_config_1.TYPE_VALUE,
        PerformanceEventMap: base_config_1.TYPE,
        Performance: base_config_1.TYPE_VALUE,
        PerformanceEntry: base_config_1.TYPE_VALUE,
        PerformanceMark: base_config_1.TYPE_VALUE,
        PerformanceMeasure: base_config_1.TYPE_VALUE,
        PerformanceNavigation: base_config_1.TYPE_VALUE,
        PerformanceNavigationTiming: base_config_1.TYPE_VALUE,
        PerformanceObserver: base_config_1.TYPE_VALUE,
        PerformanceObserverEntryList: base_config_1.TYPE_VALUE,
        PerformanceResourceTiming: base_config_1.TYPE_VALUE,
        PerformanceTiming: base_config_1.TYPE_VALUE,
        PeriodicWave: base_config_1.TYPE_VALUE,
        PermissionRequest: base_config_1.TYPE_VALUE,
        PermissionRequestedEvent: base_config_1.TYPE_VALUE,
        PermissionStatusEventMap: base_config_1.TYPE,
        PermissionStatus: base_config_1.TYPE_VALUE,
        Permissions: base_config_1.TYPE_VALUE,
        Plugin: base_config_1.TYPE_VALUE,
        PluginArray: base_config_1.TYPE_VALUE,
        PointerEvent: base_config_1.TYPE_VALUE,
        PopStateEvent: base_config_1.TYPE_VALUE,
        ProcessingInstruction: base_config_1.TYPE_VALUE,
        ProgressEvent: base_config_1.TYPE_VALUE,
        PromiseRejectionEvent: base_config_1.TYPE_VALUE,
        PublicKeyCredential: base_config_1.TYPE_VALUE,
        PushManager: base_config_1.TYPE_VALUE,
        PushSubscription: base_config_1.TYPE_VALUE,
        PushSubscriptionOptions: base_config_1.TYPE_VALUE,
        RTCCertificate: base_config_1.TYPE_VALUE,
        RTCDTMFSenderEventMap: base_config_1.TYPE,
        RTCDTMFSender: base_config_1.TYPE_VALUE,
        RTCDTMFToneChangeEvent: base_config_1.TYPE_VALUE,
        RTCDataChannelEventMap: base_config_1.TYPE,
        RTCDataChannel: base_config_1.TYPE_VALUE,
        RTCDataChannelEvent: base_config_1.TYPE_VALUE,
        RTCDtlsTransportEventMap: base_config_1.TYPE,
        RTCDtlsTransport: base_config_1.TYPE_VALUE,
        RTCDtlsTransportStateChangedEvent: base_config_1.TYPE_VALUE,
        RTCDtmfSenderEventMap: base_config_1.TYPE,
        RTCDtmfSender: base_config_1.TYPE_VALUE,
        RTCError: base_config_1.TYPE_VALUE,
        RTCErrorEvent: base_config_1.TYPE_VALUE,
        RTCIceCandidate: base_config_1.TYPE_VALUE,
        RTCIceCandidatePairChangedEvent: base_config_1.TYPE_VALUE,
        RTCIceGathererEventMap: base_config_1.TYPE,
        RTCIceGatherer: base_config_1.TYPE_VALUE,
        RTCIceGathererEvent: base_config_1.TYPE_VALUE,
        RTCIceTransportEventMap: base_config_1.TYPE,
        RTCIceTransport: base_config_1.TYPE_VALUE,
        RTCIceTransportStateChangedEvent: base_config_1.TYPE_VALUE,
        RTCIdentityAssertion: base_config_1.TYPE_VALUE,
        RTCPeerConnectionEventMap: base_config_1.TYPE,
        RTCPeerConnection: base_config_1.TYPE_VALUE,
        RTCPeerConnectionIceErrorEvent: base_config_1.TYPE_VALUE,
        RTCPeerConnectionIceEvent: base_config_1.TYPE_VALUE,
        RTCRtpReceiver: base_config_1.TYPE_VALUE,
        RTCRtpSender: base_config_1.TYPE_VALUE,
        RTCRtpTransceiver: base_config_1.TYPE_VALUE,
        RTCSctpTransportEventMap: base_config_1.TYPE,
        RTCSctpTransport: base_config_1.TYPE_VALUE,
        RTCSessionDescription: base_config_1.TYPE_VALUE,
        RTCSrtpSdesTransportEventMap: base_config_1.TYPE,
        RTCSrtpSdesTransport: base_config_1.TYPE_VALUE,
        RTCSsrcConflictEvent: base_config_1.TYPE_VALUE,
        RTCStatsProvider: base_config_1.TYPE_VALUE,
        RTCTrackEvent: base_config_1.TYPE_VALUE,
        RadioNodeList: base_config_1.TYPE_VALUE,
        RandomSource: base_config_1.TYPE_VALUE,
        Range: base_config_1.TYPE_VALUE,
        ReadableStream: base_config_1.TYPE_VALUE,
        ReadableStreamDefaultController: base_config_1.TYPE_VALUE,
        ReadableStreamDefaultReader: base_config_1.TYPE_VALUE,
        ReadableStreamGenericReader: base_config_1.TYPE,
        Request: base_config_1.TYPE_VALUE,
        ResizeObserver: base_config_1.TYPE_VALUE,
        ResizeObserverEntry: base_config_1.TYPE_VALUE,
        ResizeObserverSize: base_config_1.TYPE_VALUE,
        Response: base_config_1.TYPE_VALUE,
        SVGAElement: base_config_1.TYPE_VALUE,
        SVGAngle: base_config_1.TYPE_VALUE,
        SVGAnimateElement: base_config_1.TYPE_VALUE,
        SVGAnimateMotionElement: base_config_1.TYPE_VALUE,
        SVGAnimateTransformElement: base_config_1.TYPE_VALUE,
        SVGAnimatedAngle: base_config_1.TYPE_VALUE,
        SVGAnimatedBoolean: base_config_1.TYPE_VALUE,
        SVGAnimatedEnumeration: base_config_1.TYPE_VALUE,
        SVGAnimatedInteger: base_config_1.TYPE_VALUE,
        SVGAnimatedLength: base_config_1.TYPE_VALUE,
        SVGAnimatedLengthList: base_config_1.TYPE_VALUE,
        SVGAnimatedNumber: base_config_1.TYPE_VALUE,
        SVGAnimatedNumberList: base_config_1.TYPE_VALUE,
        SVGAnimatedPoints: base_config_1.TYPE,
        SVGAnimatedPreserveAspectRatio: base_config_1.TYPE_VALUE,
        SVGAnimatedRect: base_config_1.TYPE_VALUE,
        SVGAnimatedString: base_config_1.TYPE_VALUE,
        SVGAnimatedTransformList: base_config_1.TYPE_VALUE,
        SVGAnimationElement: base_config_1.TYPE_VALUE,
        SVGCircleElement: base_config_1.TYPE_VALUE,
        SVGClipPathElement: base_config_1.TYPE_VALUE,
        SVGComponentTransferFunctionElement: base_config_1.TYPE_VALUE,
        SVGCursorElement: base_config_1.TYPE_VALUE,
        SVGDefsElement: base_config_1.TYPE_VALUE,
        SVGDescElement: base_config_1.TYPE_VALUE,
        SVGElementEventMap: base_config_1.TYPE,
        SVGElement: base_config_1.TYPE_VALUE,
        SVGElementInstance: base_config_1.TYPE_VALUE,
        SVGElementInstanceList: base_config_1.TYPE_VALUE,
        SVGEllipseElement: base_config_1.TYPE_VALUE,
        SVGFEBlendElement: base_config_1.TYPE_VALUE,
        SVGFEColorMatrixElement: base_config_1.TYPE_VALUE,
        SVGFEComponentTransferElement: base_config_1.TYPE_VALUE,
        SVGFECompositeElement: base_config_1.TYPE_VALUE,
        SVGFEConvolveMatrixElement: base_config_1.TYPE_VALUE,
        SVGFEDiffuseLightingElement: base_config_1.TYPE_VALUE,
        SVGFEDisplacementMapElement: base_config_1.TYPE_VALUE,
        SVGFEDistantLightElement: base_config_1.TYPE_VALUE,
        SVGFEDropShadowElement: base_config_1.TYPE_VALUE,
        SVGFEFloodElement: base_config_1.TYPE_VALUE,
        SVGFEFuncAElement: base_config_1.TYPE_VALUE,
        SVGFEFuncBElement: base_config_1.TYPE_VALUE,
        SVGFEFuncGElement: base_config_1.TYPE_VALUE,
        SVGFEFuncRElement: base_config_1.TYPE_VALUE,
        SVGFEGaussianBlurElement: base_config_1.TYPE_VALUE,
        SVGFEImageElement: base_config_1.TYPE_VALUE,
        SVGFEMergeElement: base_config_1.TYPE_VALUE,
        SVGFEMergeNodeElement: base_config_1.TYPE_VALUE,
        SVGFEMorphologyElement: base_config_1.TYPE_VALUE,
        SVGFEOffsetElement: base_config_1.TYPE_VALUE,
        SVGFEPointLightElement: base_config_1.TYPE_VALUE,
        SVGFESpecularLightingElement: base_config_1.TYPE_VALUE,
        SVGFESpotLightElement: base_config_1.TYPE_VALUE,
        SVGFETileElement: base_config_1.TYPE_VALUE,
        SVGFETurbulenceElement: base_config_1.TYPE_VALUE,
        SVGFilterElement: base_config_1.TYPE_VALUE,
        SVGFilterPrimitiveStandardAttributes: base_config_1.TYPE,
        SVGFitToViewBox: base_config_1.TYPE,
        SVGForeignObjectElement: base_config_1.TYPE_VALUE,
        SVGGElement: base_config_1.TYPE_VALUE,
        SVGGeometryElement: base_config_1.TYPE_VALUE,
        SVGGradientElement: base_config_1.TYPE_VALUE,
        SVGGraphicsElement: base_config_1.TYPE_VALUE,
        SVGImageElement: base_config_1.TYPE_VALUE,
        SVGLength: base_config_1.TYPE_VALUE,
        SVGLengthList: base_config_1.TYPE_VALUE,
        SVGLineElement: base_config_1.TYPE_VALUE,
        SVGLinearGradientElement: base_config_1.TYPE_VALUE,
        SVGMarkerElement: base_config_1.TYPE_VALUE,
        SVGMaskElement: base_config_1.TYPE_VALUE,
        SVGMetadataElement: base_config_1.TYPE_VALUE,
        SVGNumber: base_config_1.TYPE_VALUE,
        SVGNumberList: base_config_1.TYPE_VALUE,
        SVGPathElement: base_config_1.TYPE_VALUE,
        SVGPathSeg: base_config_1.TYPE_VALUE,
        SVGPathSegArcAbs: base_config_1.TYPE_VALUE,
        SVGPathSegArcRel: base_config_1.TYPE_VALUE,
        SVGPathSegClosePath: base_config_1.TYPE_VALUE,
        SVGPathSegCurvetoCubicAbs: base_config_1.TYPE_VALUE,
        SVGPathSegCurvetoCubicRel: base_config_1.TYPE_VALUE,
        SVGPathSegCurvetoCubicSmoothAbs: base_config_1.TYPE_VALUE,
        SVGPathSegCurvetoCubicSmoothRel: base_config_1.TYPE_VALUE,
        SVGPathSegCurvetoQuadraticAbs: base_config_1.TYPE_VALUE,
        SVGPathSegCurvetoQuadraticRel: base_config_1.TYPE_VALUE,
        SVGPathSegCurvetoQuadraticSmoothAbs: base_config_1.TYPE_VALUE,
        SVGPathSegCurvetoQuadraticSmoothRel: base_config_1.TYPE_VALUE,
        SVGPathSegLinetoAbs: base_config_1.TYPE_VALUE,
        SVGPathSegLinetoHorizontalAbs: base_config_1.TYPE_VALUE,
        SVGPathSegLinetoHorizontalRel: base_config_1.TYPE_VALUE,
        SVGPathSegLinetoRel: base_config_1.TYPE_VALUE,
        SVGPathSegLinetoVerticalAbs: base_config_1.TYPE_VALUE,
        SVGPathSegLinetoVerticalRel: base_config_1.TYPE_VALUE,
        SVGPathSegList: base_config_1.TYPE_VALUE,
        SVGPathSegMovetoAbs: base_config_1.TYPE_VALUE,
        SVGPathSegMovetoRel: base_config_1.TYPE_VALUE,
        SVGPatternElement: base_config_1.TYPE_VALUE,
        SVGPointList: base_config_1.TYPE_VALUE,
        SVGPolygonElement: base_config_1.TYPE_VALUE,
        SVGPolylineElement: base_config_1.TYPE_VALUE,
        SVGPreserveAspectRatio: base_config_1.TYPE_VALUE,
        SVGRadialGradientElement: base_config_1.TYPE_VALUE,
        SVGRectElement: base_config_1.TYPE_VALUE,
        SVGSVGElementEventMap: base_config_1.TYPE,
        SVGSVGElement: base_config_1.TYPE_VALUE,
        SVGScriptElement: base_config_1.TYPE_VALUE,
        SVGStopElement: base_config_1.TYPE_VALUE,
        SVGStringList: base_config_1.TYPE_VALUE,
        SVGStyleElement: base_config_1.TYPE_VALUE,
        SVGSwitchElement: base_config_1.TYPE_VALUE,
        SVGSymbolElement: base_config_1.TYPE_VALUE,
        SVGTSpanElement: base_config_1.TYPE_VALUE,
        SVGTests: base_config_1.TYPE,
        SVGTextContentElement: base_config_1.TYPE_VALUE,
        SVGTextElement: base_config_1.TYPE_VALUE,
        SVGTextPathElement: base_config_1.TYPE_VALUE,
        SVGTextPositioningElement: base_config_1.TYPE_VALUE,
        SVGTitleElement: base_config_1.TYPE_VALUE,
        SVGTransform: base_config_1.TYPE_VALUE,
        SVGTransformList: base_config_1.TYPE_VALUE,
        SVGURIReference: base_config_1.TYPE,
        SVGUnitTypes: base_config_1.TYPE_VALUE,
        SVGUseElement: base_config_1.TYPE_VALUE,
        SVGViewElement: base_config_1.TYPE_VALUE,
        SVGZoomAndPan: base_config_1.TYPE,
        SVGZoomEvent: base_config_1.TYPE_VALUE,
        Screen: base_config_1.TYPE_VALUE,
        ScreenOrientationEventMap: base_config_1.TYPE,
        ScreenOrientation: base_config_1.TYPE_VALUE,
        ScriptProcessorNodeEventMap: base_config_1.TYPE,
        ScriptProcessorNode: base_config_1.TYPE_VALUE,
        SecurityPolicyViolationEvent: base_config_1.TYPE_VALUE,
        Selection: base_config_1.TYPE_VALUE,
        ServiceUIFrameContext: base_config_1.TYPE_VALUE,
        ServiceWorkerEventMap: base_config_1.TYPE,
        ServiceWorker: base_config_1.TYPE_VALUE,
        ServiceWorkerContainerEventMap: base_config_1.TYPE,
        ServiceWorkerContainer: base_config_1.TYPE_VALUE,
        ServiceWorkerRegistrationEventMap: base_config_1.TYPE,
        ServiceWorkerRegistration: base_config_1.TYPE_VALUE,
        ShadowRoot: base_config_1.TYPE_VALUE,
        SharedWorker: base_config_1.TYPE_VALUE,
        Slottable: base_config_1.TYPE,
        SourceBufferEventMap: base_config_1.TYPE,
        SourceBuffer: base_config_1.TYPE_VALUE,
        SourceBufferListEventMap: base_config_1.TYPE,
        SourceBufferList: base_config_1.TYPE_VALUE,
        SpeechGrammar: base_config_1.TYPE_VALUE,
        SpeechGrammarList: base_config_1.TYPE_VALUE,
        SpeechRecognitionEventMap: base_config_1.TYPE,
        SpeechRecognition: base_config_1.TYPE_VALUE,
        SpeechRecognitionAlternative: base_config_1.TYPE_VALUE,
        SpeechRecognitionErrorEvent: base_config_1.TYPE_VALUE,
        SpeechRecognitionEvent: base_config_1.TYPE_VALUE,
        SpeechRecognitionResult: base_config_1.TYPE_VALUE,
        SpeechRecognitionResultList: base_config_1.TYPE_VALUE,
        SpeechSynthesisEventMap: base_config_1.TYPE,
        SpeechSynthesis: base_config_1.TYPE_VALUE,
        SpeechSynthesisErrorEvent: base_config_1.TYPE_VALUE,
        SpeechSynthesisEvent: base_config_1.TYPE_VALUE,
        SpeechSynthesisUtteranceEventMap: base_config_1.TYPE,
        SpeechSynthesisUtterance: base_config_1.TYPE_VALUE,
        SpeechSynthesisVoice: base_config_1.TYPE_VALUE,
        StaticRange: base_config_1.TYPE_VALUE,
        StereoPannerNode: base_config_1.TYPE_VALUE,
        Storage: base_config_1.TYPE_VALUE,
        StorageEvent: base_config_1.TYPE_VALUE,
        StorageManager: base_config_1.TYPE_VALUE,
        StyleMedia: base_config_1.TYPE_VALUE,
        StyleSheet: base_config_1.TYPE_VALUE,
        StyleSheetList: base_config_1.TYPE_VALUE,
        SubtleCrypto: base_config_1.TYPE_VALUE,
        SyncManager: base_config_1.TYPE_VALUE,
        Text: base_config_1.TYPE_VALUE,
        TextDecoder: base_config_1.TYPE_VALUE,
        TextDecoderCommon: base_config_1.TYPE,
        TextDecoderStream: base_config_1.TYPE_VALUE,
        TextEncoder: base_config_1.TYPE_VALUE,
        TextEncoderCommon: base_config_1.TYPE,
        TextEncoderStream: base_config_1.TYPE_VALUE,
        TextEvent: base_config_1.TYPE_VALUE,
        TextMetrics: base_config_1.TYPE_VALUE,
        TextTrackEventMap: base_config_1.TYPE,
        TextTrack: base_config_1.TYPE_VALUE,
        TextTrackCueEventMap: base_config_1.TYPE,
        TextTrackCue: base_config_1.TYPE_VALUE,
        TextTrackCueList: base_config_1.TYPE_VALUE,
        TextTrackListEventMap: base_config_1.TYPE,
        TextTrackList: base_config_1.TYPE_VALUE,
        TimeRanges: base_config_1.TYPE_VALUE,
        Touch: base_config_1.TYPE_VALUE,
        TouchEvent: base_config_1.TYPE_VALUE,
        TouchList: base_config_1.TYPE_VALUE,
        TrackEvent: base_config_1.TYPE_VALUE,
        TransformStream: base_config_1.TYPE_VALUE,
        TransformStreamDefaultController: base_config_1.TYPE_VALUE,
        TransitionEvent: base_config_1.TYPE_VALUE,
        TreeWalker: base_config_1.TYPE_VALUE,
        UIEvent: base_config_1.TYPE_VALUE,
        URL: base_config_1.TYPE_VALUE,
        webkitURL: base_config_1.TYPE_VALUE,
        URLSearchParams: base_config_1.TYPE_VALUE,
        VRDisplay: base_config_1.TYPE_VALUE,
        VRDisplayCapabilities: base_config_1.TYPE_VALUE,
        VRDisplayEvent: base_config_1.TYPE_VALUE,
        VREyeParameters: base_config_1.TYPE_VALUE,
        VRFieldOfView: base_config_1.TYPE_VALUE,
        VRFrameData: base_config_1.TYPE_VALUE,
        VRPose: base_config_1.TYPE_VALUE,
        VTTCue: base_config_1.TYPE_VALUE,
        VTTRegion: base_config_1.TYPE_VALUE,
        ValidityState: base_config_1.TYPE_VALUE,
        VideoPlaybackQuality: base_config_1.TYPE_VALUE,
        VisualViewportEventMap: base_config_1.TYPE,
        VisualViewport: base_config_1.TYPE_VALUE,
        WEBGL_color_buffer_float: base_config_1.TYPE,
        WEBGL_compressed_texture_astc: base_config_1.TYPE,
        WEBGL_compressed_texture_s3tc: base_config_1.TYPE,
        WEBGL_compressed_texture_s3tc_srgb: base_config_1.TYPE,
        WEBGL_debug_renderer_info: base_config_1.TYPE,
        WEBGL_debug_shaders: base_config_1.TYPE,
        WEBGL_depth_texture: base_config_1.TYPE,
        WEBGL_draw_buffers: base_config_1.TYPE,
        WEBGL_lose_context: base_config_1.TYPE,
        WaveShaperNode: base_config_1.TYPE_VALUE,
        WebGL2RenderingContext: base_config_1.TYPE_VALUE,
        WebGL2RenderingContextBase: base_config_1.TYPE,
        WebGL2RenderingContextOverloads: base_config_1.TYPE,
        WebGLActiveInfo: base_config_1.TYPE_VALUE,
        WebGLBuffer: base_config_1.TYPE_VALUE,
        WebGLContextEvent: base_config_1.TYPE_VALUE,
        WebGLFramebuffer: base_config_1.TYPE_VALUE,
        WebGLProgram: base_config_1.TYPE_VALUE,
        WebGLQuery: base_config_1.TYPE_VALUE,
        WebGLRenderbuffer: base_config_1.TYPE_VALUE,
        WebGLRenderingContext: base_config_1.TYPE_VALUE,
        WebGLRenderingContextBase: base_config_1.TYPE,
        WebGLRenderingContextOverloads: base_config_1.TYPE,
        WebGLSampler: base_config_1.TYPE_VALUE,
        WebGLShader: base_config_1.TYPE_VALUE,
        WebGLShaderPrecisionFormat: base_config_1.TYPE_VALUE,
        WebGLSync: base_config_1.TYPE_VALUE,
        WebGLTexture: base_config_1.TYPE_VALUE,
        WebGLTransformFeedback: base_config_1.TYPE_VALUE,
        WebGLUniformLocation: base_config_1.TYPE_VALUE,
        WebGLVertexArrayObject: base_config_1.TYPE_VALUE,
        WebGLVertexArrayObjectOES: base_config_1.TYPE,
        WebKitPoint: base_config_1.TYPE_VALUE,
        WebSocketEventMap: base_config_1.TYPE,
        WebSocket: base_config_1.TYPE_VALUE,
        WheelEvent: base_config_1.TYPE_VALUE,
        WindowEventMap: base_config_1.TYPE,
        Window: base_config_1.TYPE_VALUE,
        WindowEventHandlersEventMap: base_config_1.TYPE,
        WindowEventHandlers: base_config_1.TYPE,
        WindowLocalStorage: base_config_1.TYPE,
        WindowOrWorkerGlobalScope: base_config_1.TYPE,
        WindowSessionStorage: base_config_1.TYPE,
        WorkerEventMap: base_config_1.TYPE,
        Worker: base_config_1.TYPE_VALUE,
        Worklet: base_config_1.TYPE_VALUE,
        WritableStream: base_config_1.TYPE_VALUE,
        WritableStreamDefaultController: base_config_1.TYPE_VALUE,
        WritableStreamDefaultWriter: base_config_1.TYPE_VALUE,
        XMLDocument: base_config_1.TYPE_VALUE,
        XMLHttpRequestEventMap: base_config_1.TYPE,
        XMLHttpRequest: base_config_1.TYPE_VALUE,
        XMLHttpRequestEventTargetEventMap: base_config_1.TYPE,
        XMLHttpRequestEventTarget: base_config_1.TYPE_VALUE,
        XMLHttpRequestUpload: base_config_1.TYPE_VALUE,
        XMLSerializer: base_config_1.TYPE_VALUE,
        XPathEvaluator: base_config_1.TYPE_VALUE,
        XPathEvaluatorBase: base_config_1.TYPE,
        XPathExpression: base_config_1.TYPE_VALUE,
        XPathResult: base_config_1.TYPE_VALUE,
        XSLTProcessor: base_config_1.TYPE_VALUE,
        webkitRTCPeerConnection: base_config_1.TYPE_VALUE,
        EventListenerOrEventListenerObject: base_config_1.TYPE,
        Console: base_config_1.TYPE,
        CSS: base_config_1.TYPE_VALUE,
        WebAssembly: base_config_1.TYPE_VALUE,
        BlobCallback: base_config_1.TYPE,
        CustomElementConstructor: base_config_1.TYPE,
        DecodeErrorCallback: base_config_1.TYPE,
        DecodeSuccessCallback: base_config_1.TYPE,
        EventHandlerNonNull: base_config_1.TYPE,
        ForEachCallback: base_config_1.TYPE,
        FrameRequestCallback: base_config_1.TYPE,
        FunctionStringCallback: base_config_1.TYPE,
        IntersectionObserverCallback: base_config_1.TYPE,
        MSLaunchUriCallback: base_config_1.TYPE,
        MutationCallback: base_config_1.TYPE,
        NavigatorUserMediaErrorCallback: base_config_1.TYPE,
        NavigatorUserMediaSuccessCallback: base_config_1.TYPE,
        NotificationPermissionCallback: base_config_1.TYPE,
        OnBeforeUnloadEventHandlerNonNull: base_config_1.TYPE,
        OnErrorEventHandlerNonNull: base_config_1.TYPE,
        PerformanceObserverCallback: base_config_1.TYPE,
        PositionCallback: base_config_1.TYPE,
        PositionErrorCallback: base_config_1.TYPE,
        QueuingStrategySize: base_config_1.TYPE,
        RTCPeerConnectionErrorCallback: base_config_1.TYPE,
        RTCSessionDescriptionCallback: base_config_1.TYPE,
        RTCStatsCallback: base_config_1.TYPE,
        ResizeObserverCallback: base_config_1.TYPE,
        TransformerFlushCallback: base_config_1.TYPE,
        TransformerStartCallback: base_config_1.TYPE,
        TransformerTransformCallback: base_config_1.TYPE,
        UnderlyingSinkAbortCallback: base_config_1.TYPE,
        UnderlyingSinkCloseCallback: base_config_1.TYPE,
        UnderlyingSinkStartCallback: base_config_1.TYPE,
        UnderlyingSinkWriteCallback: base_config_1.TYPE,
        UnderlyingSourceCancelCallback: base_config_1.TYPE,
        UnderlyingSourcePullCallback: base_config_1.TYPE,
        UnderlyingSourceStartCallback: base_config_1.TYPE,
        VoidFunction: base_config_1.TYPE,
        HTMLElementTagNameMap: base_config_1.TYPE,
        HTMLElementDeprecatedTagNameMap: base_config_1.TYPE,
        SVGElementTagNameMap: base_config_1.TYPE,
        ElementTagNameMap: base_config_1.TYPE,
        HeadersInit: base_config_1.TYPE,
        BodyInit: base_config_1.TYPE,
        RequestInfo: base_config_1.TYPE,
        BlobPart: base_config_1.TYPE,
        DOMHighResTimeStamp: base_config_1.TYPE,
        RenderingContext: base_config_1.TYPE,
        HTMLOrSVGImageElement: base_config_1.TYPE,
        CanvasImageSource: base_config_1.TYPE,
        OffscreenRenderingContext: base_config_1.TYPE,
        MessageEventSource: base_config_1.TYPE,
        HTMLOrSVGScriptElement: base_config_1.TYPE,
        ImageBitmapSource: base_config_1.TYPE,
        MediaProvider: base_config_1.TYPE,
        OnErrorEventHandler: base_config_1.TYPE,
        OnBeforeUnloadEventHandler: base_config_1.TYPE,
        TimerHandler: base_config_1.TYPE,
        ConstrainULong: base_config_1.TYPE,
        ConstrainDouble: base_config_1.TYPE,
        ConstrainBoolean: base_config_1.TYPE,
        ConstrainDOMString: base_config_1.TYPE,
        PerformanceEntryList: base_config_1.TYPE,
        ReadableStreamReader: base_config_1.TYPE,
        ReadableStreamController: base_config_1.TYPE,
        VibratePattern: base_config_1.TYPE,
        COSEAlgorithmIdentifier: base_config_1.TYPE,
        UvmEntry: base_config_1.TYPE,
        UvmEntries: base_config_1.TYPE,
        AlgorithmIdentifier: base_config_1.TYPE,
        HashAlgorithmIdentifier: base_config_1.TYPE,
        BigInteger: base_config_1.TYPE,
        NamedCurve: base_config_1.TYPE,
        GLenum: base_config_1.TYPE,
        GLboolean: base_config_1.TYPE,
        GLbitfield: base_config_1.TYPE,
        GLint: base_config_1.TYPE,
        GLsizei: base_config_1.TYPE,
        GLintptr: base_config_1.TYPE,
        GLsizeiptr: base_config_1.TYPE,
        GLuint: base_config_1.TYPE,
        GLfloat: base_config_1.TYPE,
        GLclampf: base_config_1.TYPE,
        TexImageSource: base_config_1.TYPE,
        Float32List: base_config_1.TYPE,
        Int32List: base_config_1.TYPE,
        GLint64: base_config_1.TYPE,
        GLuint64: base_config_1.TYPE,
        Uint32List: base_config_1.TYPE,
        BufferSource: base_config_1.TYPE,
        DOMTimeStamp: base_config_1.TYPE,
        LineAndPositionSetting: base_config_1.TYPE,
        FormDataEntryValue: base_config_1.TYPE,
        InsertPosition: base_config_1.TYPE,
        IDBValidKey: base_config_1.TYPE,
        MutationRecordType: base_config_1.TYPE,
        IDBKeyPath: base_config_1.TYPE,
        Transferable: base_config_1.TYPE,
        RTCIceGatherCandidate: base_config_1.TYPE,
        RTCTransport: base_config_1.TYPE,
        MouseWheelEvent: base_config_1.TYPE,
        WindowProxy: base_config_1.TYPE,
        ReadableStreamDefaultReadResult: base_config_1.TYPE,
        AlignSetting: base_config_1.TYPE,
        AnimationPlayState: base_config_1.TYPE,
        AppendMode: base_config_1.TYPE,
        AttestationConveyancePreference: base_config_1.TYPE,
        AudioContextLatencyCategory: base_config_1.TYPE,
        AudioContextState: base_config_1.TYPE,
        AuthenticatorAttachment: base_config_1.TYPE,
        AuthenticatorTransport: base_config_1.TYPE,
        AutoKeyword: base_config_1.TYPE,
        AutomationRate: base_config_1.TYPE,
        BinaryType: base_config_1.TYPE,
        BiquadFilterType: base_config_1.TYPE,
        CanPlayTypeResult: base_config_1.TYPE,
        CanvasDirection: base_config_1.TYPE,
        CanvasFillRule: base_config_1.TYPE,
        CanvasLineCap: base_config_1.TYPE,
        CanvasLineJoin: base_config_1.TYPE,
        CanvasTextAlign: base_config_1.TYPE,
        CanvasTextBaseline: base_config_1.TYPE,
        ChannelCountMode: base_config_1.TYPE,
        ChannelInterpretation: base_config_1.TYPE,
        ClientTypes: base_config_1.TYPE,
        ColorSpaceConversion: base_config_1.TYPE,
        CompositeOperation: base_config_1.TYPE,
        CompositeOperationOrAuto: base_config_1.TYPE,
        CredentialMediationRequirement: base_config_1.TYPE,
        DOMParserSupportedType: base_config_1.TYPE,
        DirectionSetting: base_config_1.TYPE,
        DisplayCaptureSurfaceType: base_config_1.TYPE,
        DistanceModelType: base_config_1.TYPE,
        DocumentReadyState: base_config_1.TYPE,
        EndOfStreamError: base_config_1.TYPE,
        EndingType: base_config_1.TYPE,
        FillMode: base_config_1.TYPE,
        FullscreenNavigationUI: base_config_1.TYPE,
        GamepadHand: base_config_1.TYPE,
        GamepadHapticActuatorType: base_config_1.TYPE,
        GamepadMappingType: base_config_1.TYPE,
        IDBCursorDirection: base_config_1.TYPE,
        IDBRequestReadyState: base_config_1.TYPE,
        IDBTransactionMode: base_config_1.TYPE,
        ImageOrientation: base_config_1.TYPE,
        ImageSmoothingQuality: base_config_1.TYPE,
        IterationCompositeOperation: base_config_1.TYPE,
        KeyFormat: base_config_1.TYPE,
        KeyType: base_config_1.TYPE,
        KeyUsage: base_config_1.TYPE,
        LineAlignSetting: base_config_1.TYPE,
        ListeningState: base_config_1.TYPE,
        MSCredentialType: base_config_1.TYPE,
        MSTransportType: base_config_1.TYPE,
        MSWebViewPermissionState: base_config_1.TYPE,
        MSWebViewPermissionType: base_config_1.TYPE,
        MediaDeviceKind: base_config_1.TYPE,
        MediaKeyMessageType: base_config_1.TYPE,
        MediaKeySessionType: base_config_1.TYPE,
        MediaKeyStatus: base_config_1.TYPE,
        MediaKeysRequirement: base_config_1.TYPE,
        MediaStreamTrackState: base_config_1.TYPE,
        NavigationReason: base_config_1.TYPE,
        NavigationType: base_config_1.TYPE,
        NotificationDirection: base_config_1.TYPE,
        NotificationPermission: base_config_1.TYPE,
        OffscreenRenderingContextId: base_config_1.TYPE,
        OrientationLockType: base_config_1.TYPE,
        OrientationType: base_config_1.TYPE,
        OscillatorType: base_config_1.TYPE,
        OverSampleType: base_config_1.TYPE,
        PanningModelType: base_config_1.TYPE,
        PaymentComplete: base_config_1.TYPE,
        PaymentShippingType: base_config_1.TYPE,
        PermissionName: base_config_1.TYPE,
        PermissionState: base_config_1.TYPE,
        PlaybackDirection: base_config_1.TYPE,
        PositionAlignSetting: base_config_1.TYPE,
        PremultiplyAlpha: base_config_1.TYPE,
        PublicKeyCredentialType: base_config_1.TYPE,
        PushEncryptionKeyName: base_config_1.TYPE,
        PushPermissionState: base_config_1.TYPE,
        RTCBundlePolicy: base_config_1.TYPE,
        RTCDataChannelState: base_config_1.TYPE,
        RTCDtlsRole: base_config_1.TYPE,
        RTCDtlsTransportState: base_config_1.TYPE,
        RTCErrorDetailType: base_config_1.TYPE,
        RTCIceCandidateType: base_config_1.TYPE,
        RTCIceComponent: base_config_1.TYPE,
        RTCIceConnectionState: base_config_1.TYPE,
        RTCIceCredentialType: base_config_1.TYPE,
        RTCIceGatherPolicy: base_config_1.TYPE,
        RTCIceGathererState: base_config_1.TYPE,
        RTCIceGatheringState: base_config_1.TYPE,
        RTCIceProtocol: base_config_1.TYPE,
        RTCIceRole: base_config_1.TYPE,
        RTCIceTcpCandidateType: base_config_1.TYPE,
        RTCIceTransportPolicy: base_config_1.TYPE,
        RTCIceTransportState: base_config_1.TYPE,
        RTCPeerConnectionState: base_config_1.TYPE,
        RTCRtcpMuxPolicy: base_config_1.TYPE,
        RTCRtpTransceiverDirection: base_config_1.TYPE,
        RTCSctpTransportState: base_config_1.TYPE,
        RTCSdpType: base_config_1.TYPE,
        RTCSignalingState: base_config_1.TYPE,
        RTCStatsIceCandidatePairState: base_config_1.TYPE,
        RTCStatsIceCandidateType: base_config_1.TYPE,
        RTCStatsType: base_config_1.TYPE,
        ReadyState: base_config_1.TYPE,
        ReferrerPolicy: base_config_1.TYPE,
        RequestCache: base_config_1.TYPE,
        RequestCredentials: base_config_1.TYPE,
        RequestDestination: base_config_1.TYPE,
        RequestMode: base_config_1.TYPE,
        RequestRedirect: base_config_1.TYPE,
        ResidentKeyRequirement: base_config_1.TYPE,
        ResizeObserverBoxOptions: base_config_1.TYPE,
        ResizeQuality: base_config_1.TYPE,
        ResponseType: base_config_1.TYPE,
        ScrollBehavior: base_config_1.TYPE,
        ScrollLogicalPosition: base_config_1.TYPE,
        ScrollRestoration: base_config_1.TYPE,
        ScrollSetting: base_config_1.TYPE,
        SelectionMode: base_config_1.TYPE,
        ServiceWorkerState: base_config_1.TYPE,
        ServiceWorkerUpdateViaCache: base_config_1.TYPE,
        ShadowRootMode: base_config_1.TYPE,
        SpeechRecognitionErrorCode: base_config_1.TYPE,
        SpeechSynthesisErrorCode: base_config_1.TYPE,
        TextTrackKind: base_config_1.TYPE,
        TextTrackMode: base_config_1.TYPE,
        TouchType: base_config_1.TYPE,
        UserVerificationRequirement: base_config_1.TYPE,
        VRDisplayEventReason: base_config_1.TYPE,
        VideoFacingModeEnum: base_config_1.TYPE,
        VisibilityState: base_config_1.TYPE,
        WebGLPowerPreference: base_config_1.TYPE,
        WorkerType: base_config_1.TYPE,
        XMLHttpRequestResponseType: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/dom.iterable.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.dom_iterable = {
        AudioParam: base_config_1.TYPE,
        AudioParamMap: base_config_1.TYPE,
        BaseAudioContext: base_config_1.TYPE,
        CSSRuleList: base_config_1.TYPE,
        CSSStyleDeclaration: base_config_1.TYPE,
        Cache: base_config_1.TYPE,
        CanvasPathDrawingStyles: base_config_1.TYPE,
        ClientRectList: base_config_1.TYPE,
        DOMRectList: base_config_1.TYPE,
        DOMStringList: base_config_1.TYPE,
        DOMTokenList: base_config_1.TYPE,
        DataTransferItemList: base_config_1.TYPE,
        FileList: base_config_1.TYPE,
        FormData: base_config_1.TYPE,
        HTMLAllCollection: base_config_1.TYPE,
        HTMLCollectionBase: base_config_1.TYPE,
        HTMLCollectionOf: base_config_1.TYPE,
        HTMLFormElement: base_config_1.TYPE,
        HTMLSelectElement: base_config_1.TYPE,
        Headers: base_config_1.TYPE,
        IDBDatabase: base_config_1.TYPE,
        IDBObjectStore: base_config_1.TYPE,
        MediaKeyStatusMap: base_config_1.TYPE,
        MediaList: base_config_1.TYPE,
        MimeTypeArray: base_config_1.TYPE,
        NamedNodeMap: base_config_1.TYPE,
        Navigator: base_config_1.TYPE,
        NodeList: base_config_1.TYPE,
        NodeListOf: base_config_1.TYPE,
        Plugin: base_config_1.TYPE,
        PluginArray: base_config_1.TYPE,
        RTCRtpTransceiver: base_config_1.TYPE,
        RTCStatsReport: base_config_1.TYPE,
        SVGLengthList: base_config_1.TYPE,
        SVGNumberList: base_config_1.TYPE,
        SVGPointList: base_config_1.TYPE,
        SVGStringList: base_config_1.TYPE,
        SourceBufferList: base_config_1.TYPE,
        SpeechGrammarList: base_config_1.TYPE,
        SpeechRecognitionResult: base_config_1.TYPE,
        SpeechRecognitionResultList: base_config_1.TYPE,
        StyleSheetList: base_config_1.TYPE,
        TextTrackCueList: base_config_1.TYPE,
        TextTrackList: base_config_1.TYPE,
        TouchList: base_config_1.TYPE,
        URLSearchParams: base_config_1.TYPE,
        VRDisplay: base_config_1.TYPE,
        WEBGL_draw_buffers: base_config_1.TYPE,
        WebGL2RenderingContextBase: base_config_1.TYPE,
        WebGL2RenderingContextOverloads: base_config_1.TYPE,
        WebGLRenderingContextBase: base_config_1.TYPE,
        WebGLRenderingContextOverloads: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/webworker.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.webworker = {
        AddEventListenerOptions: base_config_1.TYPE,
        AesCbcParams: base_config_1.TYPE,
        AesCtrParams: base_config_1.TYPE,
        AesDerivedKeyParams: base_config_1.TYPE,
        AesGcmParams: base_config_1.TYPE,
        AesKeyAlgorithm: base_config_1.TYPE,
        AesKeyGenParams: base_config_1.TYPE,
        Algorithm: base_config_1.TYPE,
        BlobPropertyBag: base_config_1.TYPE,
        CacheQueryOptions: base_config_1.TYPE,
        CanvasRenderingContext2DSettings: base_config_1.TYPE,
        ClientQueryOptions: base_config_1.TYPE,
        CloseEventInit: base_config_1.TYPE,
        CryptoKeyPair: base_config_1.TYPE,
        CustomEventInit: base_config_1.TYPE,
        DOMMatrix2DInit: base_config_1.TYPE,
        DOMMatrixInit: base_config_1.TYPE,
        DOMPointInit: base_config_1.TYPE,
        DOMQuadInit: base_config_1.TYPE,
        DOMRectInit: base_config_1.TYPE,
        DevicePermissionDescriptor: base_config_1.TYPE,
        EcKeyGenParams: base_config_1.TYPE,
        EcKeyImportParams: base_config_1.TYPE,
        EcdhKeyDeriveParams: base_config_1.TYPE,
        EcdsaParams: base_config_1.TYPE,
        ErrorEventInit: base_config_1.TYPE,
        EventInit: base_config_1.TYPE,
        EventListenerOptions: base_config_1.TYPE,
        EventSourceInit: base_config_1.TYPE,
        ExtendableEventInit: base_config_1.TYPE,
        ExtendableMessageEventInit: base_config_1.TYPE,
        FetchEventInit: base_config_1.TYPE,
        FilePropertyBag: base_config_1.TYPE,
        GetNotificationOptions: base_config_1.TYPE,
        HkdfParams: base_config_1.TYPE,
        HmacImportParams: base_config_1.TYPE,
        HmacKeyGenParams: base_config_1.TYPE,
        IDBIndexParameters: base_config_1.TYPE,
        IDBObjectStoreParameters: base_config_1.TYPE,
        IDBVersionChangeEventInit: base_config_1.TYPE,
        ImageBitmapOptions: base_config_1.TYPE,
        ImageBitmapRenderingContextSettings: base_config_1.TYPE,
        ImageEncodeOptions: base_config_1.TYPE,
        ImportMeta: base_config_1.TYPE,
        JsonWebKey: base_config_1.TYPE,
        KeyAlgorithm: base_config_1.TYPE,
        MessageEventInit: base_config_1.TYPE,
        MidiPermissionDescriptor: base_config_1.TYPE,
        MultiCacheQueryOptions: base_config_1.TYPE,
        NavigationPreloadState: base_config_1.TYPE,
        NotificationAction: base_config_1.TYPE,
        NotificationEventInit: base_config_1.TYPE,
        NotificationOptions: base_config_1.TYPE,
        Pbkdf2Params: base_config_1.TYPE,
        PerformanceObserverInit: base_config_1.TYPE,
        PermissionDescriptor: base_config_1.TYPE,
        PostMessageOptions: base_config_1.TYPE,
        ProgressEventInit: base_config_1.TYPE,
        PromiseRejectionEventInit: base_config_1.TYPE,
        PushEventInit: base_config_1.TYPE,
        PushPermissionDescriptor: base_config_1.TYPE,
        PushSubscriptionJSON: base_config_1.TYPE,
        PushSubscriptionOptionsInit: base_config_1.TYPE,
        QueuingStrategy: base_config_1.TYPE,
        QueuingStrategyInit: base_config_1.TYPE,
        ReadableStreamDefaultReadDoneResult: base_config_1.TYPE,
        ReadableStreamDefaultReadValueResult: base_config_1.TYPE,
        ReadableWritablePair: base_config_1.TYPE,
        RegistrationOptions: base_config_1.TYPE,
        RequestInit: base_config_1.TYPE,
        ResponseInit: base_config_1.TYPE,
        RsaHashedImportParams: base_config_1.TYPE,
        RsaHashedKeyGenParams: base_config_1.TYPE,
        RsaKeyGenParams: base_config_1.TYPE,
        RsaOaepParams: base_config_1.TYPE,
        RsaOtherPrimesInfo: base_config_1.TYPE,
        RsaPssParams: base_config_1.TYPE,
        StorageEstimate: base_config_1.TYPE,
        StreamPipeOptions: base_config_1.TYPE,
        SyncEventInit: base_config_1.TYPE,
        TextDecodeOptions: base_config_1.TYPE,
        TextDecoderOptions: base_config_1.TYPE,
        TextEncoderEncodeIntoResult: base_config_1.TYPE,
        Transformer: base_config_1.TYPE,
        UnderlyingSink: base_config_1.TYPE,
        UnderlyingSource: base_config_1.TYPE,
        WebGLContextAttributes: base_config_1.TYPE,
        WebGLContextEventInit: base_config_1.TYPE,
        WorkerOptions: base_config_1.TYPE,
        EventListener: base_config_1.TYPE,
        ANGLE_instanced_arrays: base_config_1.TYPE,
        AbortController: base_config_1.TYPE_VALUE,
        AbortSignalEventMap: base_config_1.TYPE,
        AbortSignal: base_config_1.TYPE_VALUE,
        AbstractWorkerEventMap: base_config_1.TYPE,
        AbstractWorker: base_config_1.TYPE,
        AesCfbParams: base_config_1.TYPE,
        AesCmacParams: base_config_1.TYPE,
        AnimationFrameProvider: base_config_1.TYPE,
        Blob: base_config_1.TYPE_VALUE,
        Body: base_config_1.TYPE,
        BroadcastChannelEventMap: base_config_1.TYPE,
        BroadcastChannel: base_config_1.TYPE_VALUE,
        ByteLengthQueuingStrategy: base_config_1.TYPE_VALUE,
        Cache: base_config_1.TYPE_VALUE,
        CacheStorage: base_config_1.TYPE_VALUE,
        CanvasCompositing: base_config_1.TYPE,
        CanvasDrawImage: base_config_1.TYPE,
        CanvasDrawPath: base_config_1.TYPE,
        CanvasFillStrokeStyles: base_config_1.TYPE,
        CanvasFilters: base_config_1.TYPE,
        CanvasGradient: base_config_1.TYPE_VALUE,
        CanvasImageData: base_config_1.TYPE,
        CanvasImageSmoothing: base_config_1.TYPE,
        CanvasPath: base_config_1.TYPE,
        CanvasPathDrawingStyles: base_config_1.TYPE,
        CanvasPattern: base_config_1.TYPE_VALUE,
        CanvasRect: base_config_1.TYPE,
        CanvasShadowStyles: base_config_1.TYPE,
        CanvasState: base_config_1.TYPE,
        CanvasText: base_config_1.TYPE,
        CanvasTextDrawingStyles: base_config_1.TYPE,
        CanvasTransform: base_config_1.TYPE,
        Client: base_config_1.TYPE_VALUE,
        Clients: base_config_1.TYPE_VALUE,
        CloseEvent: base_config_1.TYPE_VALUE,
        ConcatParams: base_config_1.TYPE,
        CountQueuingStrategy: base_config_1.TYPE_VALUE,
        Crypto: base_config_1.TYPE_VALUE,
        CryptoKey: base_config_1.TYPE_VALUE,
        CustomEvent: base_config_1.TYPE_VALUE,
        DOMException: base_config_1.TYPE_VALUE,
        DOMMatrix: base_config_1.TYPE_VALUE,
        DOMMatrixReadOnly: base_config_1.TYPE_VALUE,
        DOMPoint: base_config_1.TYPE_VALUE,
        DOMPointReadOnly: base_config_1.TYPE_VALUE,
        DOMQuad: base_config_1.TYPE_VALUE,
        DOMRect: base_config_1.TYPE_VALUE,
        DOMRectReadOnly: base_config_1.TYPE_VALUE,
        DOMStringList: base_config_1.TYPE_VALUE,
        DedicatedWorkerGlobalScopeEventMap: base_config_1.TYPE,
        DedicatedWorkerGlobalScope: base_config_1.TYPE_VALUE,
        DhImportKeyParams: base_config_1.TYPE,
        DhKeyAlgorithm: base_config_1.TYPE,
        DhKeyDeriveParams: base_config_1.TYPE,
        DhKeyGenParams: base_config_1.TYPE,
        EXT_blend_minmax: base_config_1.TYPE,
        EXT_frag_depth: base_config_1.TYPE,
        EXT_sRGB: base_config_1.TYPE,
        EXT_shader_texture_lod: base_config_1.TYPE,
        EXT_texture_filter_anisotropic: base_config_1.TYPE,
        ErrorEvent: base_config_1.TYPE_VALUE,
        Event: base_config_1.TYPE_VALUE,
        EventListenerObject: base_config_1.TYPE,
        EventSourceEventMap: base_config_1.TYPE,
        EventSource: base_config_1.TYPE_VALUE,
        EventTarget: base_config_1.TYPE_VALUE,
        ExtendableEvent: base_config_1.TYPE_VALUE,
        ExtendableMessageEvent: base_config_1.TYPE_VALUE,
        FetchEvent: base_config_1.TYPE_VALUE,
        File: base_config_1.TYPE_VALUE,
        FileList: base_config_1.TYPE_VALUE,
        FileReaderEventMap: base_config_1.TYPE,
        FileReader: base_config_1.TYPE_VALUE,
        FileReaderSync: base_config_1.TYPE_VALUE,
        FormData: base_config_1.TYPE_VALUE,
        GenericTransformStream: base_config_1.TYPE,
        Headers: base_config_1.TYPE_VALUE,
        IDBArrayKey: base_config_1.TYPE,
        IDBCursor: base_config_1.TYPE_VALUE,
        IDBCursorWithValue: base_config_1.TYPE_VALUE,
        IDBDatabaseEventMap: base_config_1.TYPE,
        IDBDatabase: base_config_1.TYPE_VALUE,
        IDBFactory: base_config_1.TYPE_VALUE,
        IDBIndex: base_config_1.TYPE_VALUE,
        IDBKeyRange: base_config_1.TYPE_VALUE,
        IDBObjectStore: base_config_1.TYPE_VALUE,
        IDBOpenDBRequestEventMap: base_config_1.TYPE,
        IDBOpenDBRequest: base_config_1.TYPE_VALUE,
        IDBRequestEventMap: base_config_1.TYPE,
        IDBRequest: base_config_1.TYPE_VALUE,
        IDBTransactionEventMap: base_config_1.TYPE,
        IDBTransaction: base_config_1.TYPE_VALUE,
        IDBVersionChangeEvent: base_config_1.TYPE_VALUE,
        ImageBitmap: base_config_1.TYPE_VALUE,
        ImageBitmapRenderingContext: base_config_1.TYPE_VALUE,
        ImageData: base_config_1.TYPE_VALUE,
        MessageChannel: base_config_1.TYPE_VALUE,
        MessageEvent: base_config_1.TYPE_VALUE,
        MessagePortEventMap: base_config_1.TYPE,
        MessagePort: base_config_1.TYPE_VALUE,
        NavigationPreloadManager: base_config_1.TYPE_VALUE,
        NavigatorConcurrentHardware: base_config_1.TYPE,
        NavigatorID: base_config_1.TYPE,
        NavigatorLanguage: base_config_1.TYPE,
        NavigatorOnLine: base_config_1.TYPE,
        NavigatorStorage: base_config_1.TYPE,
        NotificationEventMap: base_config_1.TYPE,
        Notification: base_config_1.TYPE_VALUE,
        NotificationEvent: base_config_1.TYPE_VALUE,
        OES_element_index_uint: base_config_1.TYPE,
        OES_standard_derivatives: base_config_1.TYPE,
        OES_texture_float: base_config_1.TYPE,
        OES_texture_float_linear: base_config_1.TYPE,
        OES_texture_half_float: base_config_1.TYPE,
        OES_texture_half_float_linear: base_config_1.TYPE,
        OES_vertex_array_object: base_config_1.TYPE,
        OffscreenCanvas: base_config_1.TYPE_VALUE,
        OffscreenCanvasRenderingContext2D: base_config_1.TYPE_VALUE,
        Path2D: base_config_1.TYPE_VALUE,
        PerformanceEventMap: base_config_1.TYPE,
        Performance: base_config_1.TYPE_VALUE,
        PerformanceEntry: base_config_1.TYPE_VALUE,
        PerformanceMark: base_config_1.TYPE_VALUE,
        PerformanceMeasure: base_config_1.TYPE_VALUE,
        PerformanceObserver: base_config_1.TYPE_VALUE,
        PerformanceObserverEntryList: base_config_1.TYPE_VALUE,
        PerformanceResourceTiming: base_config_1.TYPE_VALUE,
        PermissionStatusEventMap: base_config_1.TYPE,
        PermissionStatus: base_config_1.TYPE_VALUE,
        Permissions: base_config_1.TYPE_VALUE,
        ProgressEvent: base_config_1.TYPE_VALUE,
        PromiseRejectionEvent: base_config_1.TYPE_VALUE,
        PushEvent: base_config_1.TYPE_VALUE,
        PushManager: base_config_1.TYPE_VALUE,
        PushMessageData: base_config_1.TYPE_VALUE,
        PushSubscription: base_config_1.TYPE_VALUE,
        PushSubscriptionOptions: base_config_1.TYPE_VALUE,
        ReadableStream: base_config_1.TYPE_VALUE,
        ReadableStreamDefaultController: base_config_1.TYPE_VALUE,
        ReadableStreamDefaultReader: base_config_1.TYPE_VALUE,
        ReadableStreamGenericReader: base_config_1.TYPE,
        Request: base_config_1.TYPE_VALUE,
        Response: base_config_1.TYPE_VALUE,
        ServiceWorkerEventMap: base_config_1.TYPE,
        ServiceWorker: base_config_1.TYPE_VALUE,
        ServiceWorkerContainerEventMap: base_config_1.TYPE,
        ServiceWorkerContainer: base_config_1.TYPE_VALUE,
        ServiceWorkerGlobalScopeEventMap: base_config_1.TYPE,
        ServiceWorkerGlobalScope: base_config_1.TYPE_VALUE,
        ServiceWorkerRegistrationEventMap: base_config_1.TYPE,
        ServiceWorkerRegistration: base_config_1.TYPE_VALUE,
        SharedWorker: base_config_1.TYPE_VALUE,
        SharedWorkerGlobalScopeEventMap: base_config_1.TYPE,
        SharedWorkerGlobalScope: base_config_1.TYPE_VALUE,
        StorageManager: base_config_1.TYPE_VALUE,
        SubtleCrypto: base_config_1.TYPE_VALUE,
        SyncEvent: base_config_1.TYPE_VALUE,
        SyncManager: base_config_1.TYPE_VALUE,
        TextDecoder: base_config_1.TYPE_VALUE,
        TextDecoderCommon: base_config_1.TYPE,
        TextDecoderStream: base_config_1.TYPE_VALUE,
        TextEncoder: base_config_1.TYPE_VALUE,
        TextEncoderCommon: base_config_1.TYPE,
        TextEncoderStream: base_config_1.TYPE_VALUE,
        TextMetrics: base_config_1.TYPE_VALUE,
        TransformStream: base_config_1.TYPE_VALUE,
        TransformStreamDefaultController: base_config_1.TYPE_VALUE,
        URL: base_config_1.TYPE_VALUE,
        URLSearchParams: base_config_1.TYPE_VALUE,
        WEBGL_color_buffer_float: base_config_1.TYPE,
        WEBGL_compressed_texture_astc: base_config_1.TYPE,
        WEBGL_compressed_texture_s3tc: base_config_1.TYPE,
        WEBGL_compressed_texture_s3tc_srgb: base_config_1.TYPE,
        WEBGL_debug_renderer_info: base_config_1.TYPE,
        WEBGL_debug_shaders: base_config_1.TYPE,
        WEBGL_depth_texture: base_config_1.TYPE,
        WEBGL_draw_buffers: base_config_1.TYPE,
        WEBGL_lose_context: base_config_1.TYPE,
        WebGL2RenderingContext: base_config_1.TYPE_VALUE,
        WebGL2RenderingContextBase: base_config_1.TYPE,
        WebGL2RenderingContextOverloads: base_config_1.TYPE,
        WebGLActiveInfo: base_config_1.TYPE_VALUE,
        WebGLBuffer: base_config_1.TYPE_VALUE,
        WebGLContextEvent: base_config_1.TYPE_VALUE,
        WebGLFramebuffer: base_config_1.TYPE_VALUE,
        WebGLProgram: base_config_1.TYPE_VALUE,
        WebGLQuery: base_config_1.TYPE_VALUE,
        WebGLRenderbuffer: base_config_1.TYPE_VALUE,
        WebGLRenderingContext: base_config_1.TYPE_VALUE,
        WebGLRenderingContextBase: base_config_1.TYPE,
        WebGLRenderingContextOverloads: base_config_1.TYPE,
        WebGLSampler: base_config_1.TYPE_VALUE,
        WebGLShader: base_config_1.TYPE_VALUE,
        WebGLShaderPrecisionFormat: base_config_1.TYPE_VALUE,
        WebGLSync: base_config_1.TYPE_VALUE,
        WebGLTexture: base_config_1.TYPE_VALUE,
        WebGLTransformFeedback: base_config_1.TYPE_VALUE,
        WebGLUniformLocation: base_config_1.TYPE_VALUE,
        WebGLVertexArrayObject: base_config_1.TYPE_VALUE,
        WebGLVertexArrayObjectOES: base_config_1.TYPE,
        WebSocketEventMap: base_config_1.TYPE,
        WebSocket: base_config_1.TYPE_VALUE,
        WindowClient: base_config_1.TYPE_VALUE,
        WindowOrWorkerGlobalScope: base_config_1.TYPE,
        WorkerEventMap: base_config_1.TYPE,
        Worker: base_config_1.TYPE_VALUE,
        WorkerGlobalScopeEventMap: base_config_1.TYPE,
        WorkerGlobalScope: base_config_1.TYPE_VALUE,
        WorkerLocation: base_config_1.TYPE_VALUE,
        WorkerNavigator: base_config_1.TYPE_VALUE,
        WritableStream: base_config_1.TYPE_VALUE,
        WritableStreamDefaultController: base_config_1.TYPE_VALUE,
        WritableStreamDefaultWriter: base_config_1.TYPE_VALUE,
        XMLHttpRequestEventMap: base_config_1.TYPE,
        XMLHttpRequest: base_config_1.TYPE_VALUE,
        XMLHttpRequestEventTargetEventMap: base_config_1.TYPE,
        XMLHttpRequestEventTarget: base_config_1.TYPE_VALUE,
        XMLHttpRequestUpload: base_config_1.TYPE_VALUE,
        EventListenerOrEventListenerObject: base_config_1.TYPE,
        Console: base_config_1.TYPE,
        WebAssembly: base_config_1.TYPE_VALUE,
        FrameRequestCallback: base_config_1.TYPE,
        OnErrorEventHandlerNonNull: base_config_1.TYPE,
        PerformanceObserverCallback: base_config_1.TYPE,
        QueuingStrategySize: base_config_1.TYPE,
        TransformerFlushCallback: base_config_1.TYPE,
        TransformerStartCallback: base_config_1.TYPE,
        TransformerTransformCallback: base_config_1.TYPE,
        UnderlyingSinkAbortCallback: base_config_1.TYPE,
        UnderlyingSinkCloseCallback: base_config_1.TYPE,
        UnderlyingSinkStartCallback: base_config_1.TYPE,
        UnderlyingSinkWriteCallback: base_config_1.TYPE,
        UnderlyingSourceCancelCallback: base_config_1.TYPE,
        UnderlyingSourcePullCallback: base_config_1.TYPE,
        UnderlyingSourceStartCallback: base_config_1.TYPE,
        VoidFunction: base_config_1.TYPE,
        HeadersInit: base_config_1.TYPE,
        BodyInit: base_config_1.TYPE,
        RequestInfo: base_config_1.TYPE,
        BlobPart: base_config_1.TYPE,
        DOMHighResTimeStamp: base_config_1.TYPE,
        CanvasImageSource: base_config_1.TYPE,
        OffscreenRenderingContext: base_config_1.TYPE,
        MessageEventSource: base_config_1.TYPE,
        ImageBitmapSource: base_config_1.TYPE,
        OnErrorEventHandler: base_config_1.TYPE,
        TimerHandler: base_config_1.TYPE,
        PerformanceEntryList: base_config_1.TYPE,
        PushMessageDataInit: base_config_1.TYPE,
        ReadableStreamReader: base_config_1.TYPE,
        ReadableStreamController: base_config_1.TYPE,
        VibratePattern: base_config_1.TYPE,
        AlgorithmIdentifier: base_config_1.TYPE,
        HashAlgorithmIdentifier: base_config_1.TYPE,
        BigInteger: base_config_1.TYPE,
        NamedCurve: base_config_1.TYPE,
        GLenum: base_config_1.TYPE,
        GLboolean: base_config_1.TYPE,
        GLbitfield: base_config_1.TYPE,
        GLint: base_config_1.TYPE,
        GLsizei: base_config_1.TYPE,
        GLintptr: base_config_1.TYPE,
        GLsizeiptr: base_config_1.TYPE,
        GLuint: base_config_1.TYPE,
        GLfloat: base_config_1.TYPE,
        GLclampf: base_config_1.TYPE,
        TexImageSource: base_config_1.TYPE,
        Float32List: base_config_1.TYPE,
        Int32List: base_config_1.TYPE,
        GLint64: base_config_1.TYPE,
        GLuint64: base_config_1.TYPE,
        Uint32List: base_config_1.TYPE,
        BufferSource: base_config_1.TYPE,
        DOMTimeStamp: base_config_1.TYPE,
        FormDataEntryValue: base_config_1.TYPE,
        IDBValidKey: base_config_1.TYPE,
        Transferable: base_config_1.TYPE,
        ReadableStreamDefaultReadResult: base_config_1.TYPE,
        BinaryType: base_config_1.TYPE,
        CanvasDirection: base_config_1.TYPE,
        CanvasFillRule: base_config_1.TYPE,
        CanvasLineCap: base_config_1.TYPE,
        CanvasLineJoin: base_config_1.TYPE,
        CanvasTextAlign: base_config_1.TYPE,
        CanvasTextBaseline: base_config_1.TYPE,
        ClientTypes: base_config_1.TYPE,
        ColorSpaceConversion: base_config_1.TYPE,
        EndingType: base_config_1.TYPE,
        FrameType: base_config_1.TYPE,
        IDBCursorDirection: base_config_1.TYPE,
        IDBRequestReadyState: base_config_1.TYPE,
        IDBTransactionMode: base_config_1.TYPE,
        ImageOrientation: base_config_1.TYPE,
        ImageSmoothingQuality: base_config_1.TYPE,
        KeyFormat: base_config_1.TYPE,
        KeyType: base_config_1.TYPE,
        KeyUsage: base_config_1.TYPE,
        NotificationDirection: base_config_1.TYPE,
        NotificationPermission: base_config_1.TYPE,
        OffscreenRenderingContextId: base_config_1.TYPE,
        PermissionName: base_config_1.TYPE,
        PermissionState: base_config_1.TYPE,
        PremultiplyAlpha: base_config_1.TYPE,
        PushEncryptionKeyName: base_config_1.TYPE,
        PushPermissionState: base_config_1.TYPE,
        ReferrerPolicy: base_config_1.TYPE,
        RequestCache: base_config_1.TYPE,
        RequestCredentials: base_config_1.TYPE,
        RequestDestination: base_config_1.TYPE,
        RequestMode: base_config_1.TYPE,
        RequestRedirect: base_config_1.TYPE,
        ResizeQuality: base_config_1.TYPE,
        ResponseType: base_config_1.TYPE,
        ServiceWorkerState: base_config_1.TYPE,
        ServiceWorkerUpdateViaCache: base_config_1.TYPE,
        VisibilityState: base_config_1.TYPE,
        WebGLPowerPreference: base_config_1.TYPE,
        WorkerType: base_config_1.TYPE,
        XMLHttpRequestResponseType: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/webworker.importscripts.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.webworker_importscripts = {};
},
/* @typescript-eslint/scope-manager/dist/lib/webworker.iterable.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.webworker_iterable = {
        Cache: base_config_1.TYPE,
        CanvasPathDrawingStyles: base_config_1.TYPE,
        DOMStringList: base_config_1.TYPE,
        FileList: base_config_1.TYPE,
        FormData: base_config_1.TYPE,
        Headers: base_config_1.TYPE,
        IDBDatabase: base_config_1.TYPE,
        IDBObjectStore: base_config_1.TYPE,
        URLSearchParams: base_config_1.TYPE,
        WEBGL_draw_buffers: base_config_1.TYPE,
        WebGL2RenderingContextBase: base_config_1.TYPE,
        WebGL2RenderingContextOverloads: base_config_1.TYPE,
        WebGLRenderingContextBase: base_config_1.TYPE,
        WebGLRenderingContextOverloads: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/scripthost.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.scripthost = {
        ActiveXObject: base_config_1.TYPE_VALUE,
        ITextWriter: base_config_1.TYPE,
        TextStreamBase: base_config_1.TYPE,
        TextStreamWriter: base_config_1.TYPE,
        TextStreamReader: base_config_1.TYPE,
        SafeArray: base_config_1.TYPE_VALUE,
        Enumerator: base_config_1.TYPE_VALUE,
        EnumeratorConstructor: base_config_1.TYPE,
        VBArray: base_config_1.TYPE_VALUE,
        VBArrayConstructor: base_config_1.TYPE,
        VarDate: base_config_1.TYPE_VALUE,
        DateConstructor: base_config_1.TYPE,
        Date: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.array.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.esnext_array = {
        FlatArray: base_config_1.TYPE,
        ReadonlyArray: base_config_1.TYPE,
        Array: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.symbol.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.esnext_symbol = {
        Symbol: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.asynciterable.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2015_symbol_1 = require(832) /* ./es2015.symbol */;
    const es2015_iterable_1 = require(831) /* ./es2015.iterable */;
    const base_config_1 = require(827) /* ./base-config */;
    exports.esnext_asynciterable = Object.assign(Object.assign(Object.assign({}, es2015_symbol_1.es2015_symbol), es2015_iterable_1.es2015_iterable), { SymbolConstructor: base_config_1.TYPE, AsyncIterator: base_config_1.TYPE, AsyncIterable: base_config_1.TYPE, AsyncIterableIterator: base_config_1.TYPE });
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.bigint.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.esnext_bigint = {
        BigIntToLocaleStringOptions: base_config_1.TYPE,
        BigInt: base_config_1.TYPE_VALUE,
        BigIntConstructor: base_config_1.TYPE,
        BigInt64Array: base_config_1.TYPE_VALUE,
        BigInt64ArrayConstructor: base_config_1.TYPE,
        BigUint64Array: base_config_1.TYPE_VALUE,
        BigUint64ArrayConstructor: base_config_1.TYPE,
        DataView: base_config_1.TYPE,
        Intl: base_config_1.TYPE_VALUE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.string.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.esnext_string = {
        String: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.promise.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.esnext_promise = {
        AggregateError: base_config_1.TYPE_VALUE,
        AggregateErrorConstructor: base_config_1.TYPE,
        PromiseConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.weakref.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const base_config_1 = require(827) /* ./base-config */;
    exports.esnext_weakref = {
        WeakRef: base_config_1.TYPE_VALUE,
        WeakRefConstructor: base_config_1.TYPE,
        FinalizationRegistry: base_config_1.TYPE_VALUE,
        FinalizationRegistryConstructor: base_config_1.TYPE,
    };
},
/* @typescript-eslint/scope-manager/dist/lib/es2016.full.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2016_1 = require(841) /* ./es2016 */;
    const dom_1 = require(872) /* ./dom */;
    const webworker_importscripts_1 = require(875) /* ./webworker.importscripts */;
    const scripthost_1 = require(877) /* ./scripthost */;
    const dom_iterable_1 = require(873) /* ./dom.iterable */;
    exports.es2016_full = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2016_1.es2016), dom_1.dom), webworker_importscripts_1.webworker_importscripts), scripthost_1.scripthost), dom_iterable_1.dom_iterable);
},
/* @typescript-eslint/scope-manager/dist/lib/es2017.full.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2017_1 = require(842) /* ./es2017 */;
    const dom_1 = require(872) /* ./dom */;
    const webworker_importscripts_1 = require(875) /* ./webworker.importscripts */;
    const scripthost_1 = require(877) /* ./scripthost */;
    const dom_iterable_1 = require(873) /* ./dom.iterable */;
    exports.es2017_full = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2017_1.es2017), dom_1.dom), webworker_importscripts_1.webworker_importscripts), scripthost_1.scripthost), dom_iterable_1.dom_iterable);
},
/* @typescript-eslint/scope-manager/dist/lib/es2018.full.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2018_1 = require(848) /* ./es2018 */;
    const dom_1 = require(872) /* ./dom */;
    const webworker_importscripts_1 = require(875) /* ./webworker.importscripts */;
    const scripthost_1 = require(877) /* ./scripthost */;
    const dom_iterable_1 = require(873) /* ./dom.iterable */;
    exports.es2018_full = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2018_1.es2018), dom_1.dom), webworker_importscripts_1.webworker_importscripts), scripthost_1.scripthost), dom_iterable_1.dom_iterable);
},
/* @typescript-eslint/scope-manager/dist/lib/es2019.full.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2019_1 = require(854) /* ./es2019 */;
    const dom_1 = require(872) /* ./dom */;
    const webworker_importscripts_1 = require(875) /* ./webworker.importscripts */;
    const scripthost_1 = require(877) /* ./scripthost */;
    const dom_iterable_1 = require(873) /* ./dom.iterable */;
    exports.es2019_full = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2019_1.es2019), dom_1.dom), webworker_importscripts_1.webworker_importscripts), scripthost_1.scripthost), dom_iterable_1.dom_iterable);
},
/* @typescript-eslint/scope-manager/dist/lib/es2020.full.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2020_1 = require(859) /* ./es2020 */;
    const dom_1 = require(872) /* ./dom */;
    const webworker_importscripts_1 = require(875) /* ./webworker.importscripts */;
    const scripthost_1 = require(877) /* ./scripthost */;
    const dom_iterable_1 = require(873) /* ./dom.iterable */;
    exports.es2020_full = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2020_1.es2020), dom_1.dom), webworker_importscripts_1.webworker_importscripts), scripthost_1.scripthost), dom_iterable_1.dom_iterable);
},
/* @typescript-eslint/scope-manager/dist/lib/es2021.full.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es2021_1 = require(866) /* ./es2021 */;
    const dom_1 = require(872) /* ./dom */;
    const webworker_importscripts_1 = require(875) /* ./webworker.importscripts */;
    const scripthost_1 = require(877) /* ./scripthost */;
    const dom_iterable_1 = require(873) /* ./dom.iterable */;
    exports.es2021_full = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, es2021_1.es2021), dom_1.dom), webworker_importscripts_1.webworker_importscripts), scripthost_1.scripthost), dom_iterable_1.dom_iterable);
},
/* @typescript-eslint/scope-manager/dist/lib/esnext.full.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const esnext_1 = require(870) /* ./esnext */;
    const dom_1 = require(872) /* ./dom */;
    const webworker_importscripts_1 = require(875) /* ./webworker.importscripts */;
    const scripthost_1 = require(877) /* ./scripthost */;
    const dom_iterable_1 = require(873) /* ./dom.iterable */;
    exports.esnext_full = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, esnext_1.esnext), dom_1.dom), webworker_importscripts_1.webworker_importscripts), scripthost_1.scripthost), dom_iterable_1.dom_iterable);
},
/* @typescript-eslint/scope-manager/dist/lib/lib.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const es5_1 = require(826) /* ./es5 */;
    const dom_1 = require(872) /* ./dom */;
    const webworker_importscripts_1 = require(875) /* ./webworker.importscripts */;
    const scripthost_1 = require(877) /* ./scripthost */;
    exports.lib = Object.assign(Object.assign(Object.assign(Object.assign({}, es5_1.es5), dom_1.dom), webworker_importscripts_1.webworker_importscripts), scripthost_1.scripthost);
},
/* @typescript-eslint/scope-manager/dist/ScopeManager.js */ function _(require, module, exports, __esModule, __esExport) {
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _ScopeManager_options;
    __esModule();
    const assert_1 = require(800) /* ./assert */;
    const scope_1 = require(796) /* ./scope */;
    class ScopeManager {
        constructor(options) {
            _ScopeManager_options.set(this, void 0);
            this.scopes = [];
            this.globalScope = null;
            this.nodeToScope = new WeakMap();
            this.currentScope = null;
            __classPrivateFieldSet(this, _ScopeManager_options, options, "f");
            this.declaredVariables = new WeakMap();
        }
        get variables() {
            const variables = new Set();
            function recurse(scope) {
                scope.variables.forEach(v => variables.add(v));
                scope.childScopes.forEach(recurse);
            }
            this.scopes.forEach(recurse);
            return Array.from(variables).sort((a, b) => a.$id - b.$id);
        }
        isGlobalReturn() {
            return __classPrivateFieldGet(this, _ScopeManager_options, "f").globalReturn === true;
        }
        isModule() {
            return __classPrivateFieldGet(this, _ScopeManager_options, "f").sourceType === 'module';
        }
        isImpliedStrict() {
            return __classPrivateFieldGet(this, _ScopeManager_options, "f").impliedStrict === true;
        }
        isStrictModeSupported() {
            return __classPrivateFieldGet(this, _ScopeManager_options, "f").ecmaVersion != null && __classPrivateFieldGet(this, _ScopeManager_options, "f").ecmaVersion >= 5;
        }
        isES6() {
            return __classPrivateFieldGet(this, _ScopeManager_options, "f").ecmaVersion != null && __classPrivateFieldGet(this, _ScopeManager_options, "f").ecmaVersion >= 6;
        }
        /**
         * Get the variables that a given AST node defines. The gotten variables' `def[].node`/`def[].parent` property is the node.
         * If the node does not define any variable, this returns an empty array.
         * @param node An AST node to get their variables.
         * @public
         */
        getDeclaredVariables(node) {
            var _a;
            return (_a = this.declaredVariables.get(node)) !== null && _a !== void 0 ? _a : [];
        }
        /**
         * Get the scope of a given AST node. The gotten scope's `block` property is the node.
         * This method never returns `function-expression-name` scope. If the node does not have their scope, this returns `null`.
         *
         * @param node An AST node to get their scope.
         * @param inner If the node has multiple scopes, this returns the outermost scope normally.
         *                If `inner` is `true` then this returns the innermost scope.
         * @public
         */
        acquire(node, inner = false) {
            function predicate(testScope) {
                if (testScope.type === 'function' && testScope.functionExpressionScope) {
                    return false;
                }
                return true;
            }
            const scopes = this.nodeToScope.get(node);
            if (!scopes || scopes.length === 0) {
                return null;
            }
            // Heuristic selection from all scopes.
            // If you would like to get all scopes, please use ScopeManager#acquireAll.
            if (scopes.length === 1) {
                return scopes[0];
            }
            if (inner) {
                for (let i = scopes.length - 1; i >= 0; --i) {
                    const scope = scopes[i];
                    if (predicate(scope)) {
                        return scope;
                    }
                }
            }
            else {
                for (let i = 0; i < scopes.length; ++i) {
                    const scope = scopes[i];
                    if (predicate(scope)) {
                        return scope;
                    }
                }
            }
            return null;
        }
        nestScope(scope) {
            if (scope instanceof scope_1.GlobalScope) {
                assert_1.assert(this.currentScope === null);
                this.globalScope = scope;
            }
            this.currentScope = scope;
            return scope;
        }
        nestBlockScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.BlockScope(this, this.currentScope, node));
        }
        nestCatchScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.CatchScope(this, this.currentScope, node));
        }
        nestClassScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.ClassScope(this, this.currentScope, node));
        }
        nestConditionalTypeScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.ConditionalTypeScope(this, this.currentScope, node));
        }
        nestForScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.ForScope(this, this.currentScope, node));
        }
        nestFunctionExpressionNameScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.FunctionExpressionNameScope(this, this.currentScope, node));
        }
        nestFunctionScope(node, isMethodDefinition) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.FunctionScope(this, this.currentScope, node, isMethodDefinition));
        }
        nestFunctionTypeScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.FunctionTypeScope(this, this.currentScope, node));
        }
        nestGlobalScope(node) {
            return this.nestScope(new scope_1.GlobalScope(this, node));
        }
        nestMappedTypeScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.MappedTypeScope(this, this.currentScope, node));
        }
        nestModuleScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.ModuleScope(this, this.currentScope, node));
        }
        nestSwitchScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.SwitchScope(this, this.currentScope, node));
        }
        nestTSEnumScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.TSEnumScope(this, this.currentScope, node));
        }
        nestTSModuleScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.TSModuleScope(this, this.currentScope, node));
        }
        nestTypeScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.TypeScope(this, this.currentScope, node));
        }
        nestWithScope(node) {
            assert_1.assert(this.currentScope);
            return this.nestScope(new scope_1.WithScope(this, this.currentScope, node));
        }
    }
    exports.ScopeManager = ScopeManager;
    _ScopeManager_options = new WeakMap();
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint/SourceCode.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const eslint_1 = require(203) /* eslint */;
    class SourceCode extends eslint_1.SourceCode {
    }
    exports.SourceCode = SourceCode;
},
/* @typescript-eslint/experimental-utils/dist/json-schema.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    const eslint_scope_1 = require(332) /* eslint-scope */;
    __exportStar(require(897) /* ./analyze */, exports);
    __exportStar(require(898) /* ./Definition */, exports);
    __exportStar(require(899) /* ./Options */, exports);
    __exportStar(require(900) /* ./PatternVisitor */, exports);
    __exportStar(require(901) /* ./Reference */, exports);
    __exportStar(require(902) /* ./Referencer */, exports);
    __exportStar(require(903) /* ./Scope */, exports);
    __exportStar(require(904) /* ./ScopeManager */, exports);
    __exportStar(require(905) /* ./Variable */, exports);
    exports.version = eslint_scope_1.version;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/analyze.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const eslint_scope_1 = require(332) /* eslint-scope */;
    const analyze = eslint_scope_1.analyze;
    exports.analyze = analyze;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/Definition.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const definition_1 = require(339) /* eslint-scope/lib/definition */;
    const Definition = definition_1.Definition;
    exports.Definition = Definition;
    const ParameterDefinition = definition_1.ParameterDefinition;
    exports.ParameterDefinition = ParameterDefinition;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/Options.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/PatternVisitor.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const pattern_visitor_1 = __importDefault(require(343) /* eslint-scope/lib/pattern-visitor */);
    const PatternVisitor = pattern_visitor_1.default;
    exports.PatternVisitor = PatternVisitor;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/Reference.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const reference_1 = __importDefault(require(337) /* eslint-scope/lib/reference */);
    const Reference = reference_1.default;
    exports.Reference = Reference;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/Referencer.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    /* eslint-disable @typescript-eslint/no-explicit-any */
    const referencer_1 = __importDefault(require(340) /* eslint-scope/lib/referencer */);
    const Referencer = referencer_1.default;
    exports.Referencer = Referencer;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/Scope.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const scope_1 = require(334) /* eslint-scope/lib/scope */;
    const Scope = scope_1.Scope;
    exports.Scope = Scope;
    const GlobalScope = scope_1.GlobalScope;
    exports.GlobalScope = GlobalScope;
    const ModuleScope = scope_1.ModuleScope;
    exports.ModuleScope = ModuleScope;
    const FunctionExpressionNameScope = scope_1.FunctionExpressionNameScope;
    exports.FunctionExpressionNameScope = FunctionExpressionNameScope;
    const CatchScope = scope_1.CatchScope;
    exports.CatchScope = CatchScope;
    const WithScope = scope_1.WithScope;
    exports.WithScope = WithScope;
    const BlockScope = scope_1.BlockScope;
    exports.BlockScope = BlockScope;
    const SwitchScope = scope_1.SwitchScope;
    exports.SwitchScope = SwitchScope;
    const FunctionScope = scope_1.FunctionScope;
    exports.FunctionScope = FunctionScope;
    const ForScope = scope_1.ForScope;
    exports.ForScope = ForScope;
    const ClassScope = scope_1.ClassScope;
    exports.ClassScope = ClassScope;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/ScopeManager.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const scope_manager_1 = __importDefault(require(333) /* eslint-scope/lib/scope-manager */);
    const ScopeManager = scope_manager_1.default;
    exports.ScopeManager = ScopeManager;
},
/* @typescript-eslint/experimental-utils/dist/ts-eslint-scope/Variable.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const variable_1 = __importDefault(require(338) /* eslint-scope/lib/variable */);
    const Variable = variable_1.default;
    exports.Variable = Variable;
},
/* @typescript-eslint/eslint-plugin/dist/util/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    __exportStar(require(907) /* ./astUtils */, exports);
    __exportStar(require(909) /* ./collectUnusedVariables */, exports);
    __exportStar(require(910) /* ./createRule */, exports);
    __exportStar(require(912) /* ./getFunctionHeadLoc */, exports);
    __exportStar(require(913) /* ./getThisExpression */, exports);
    __exportStar(require(914) /* ./getWrappingFixer */, exports);
    __exportStar(require(915) /* ./isTypeReadonly */, exports);
    __exportStar(require(935) /* ./misc */, exports);
    __exportStar(require(936) /* ./nullThrows */, exports);
    __exportStar(require(937) /* ./objectIterators */, exports);
    __exportStar(require(938) /* ./propertyTypes */, exports);
    __exportStar(require(939) /* ./requiresQuoting */, exports);
    __exportStar(require(940) /* ./types */, exports);
    // this is done for convenience - saves migrating all of the old rules
    const { applyDefault, deepMerge, isObjectNotArray, getParserServices } = experimental_utils_1.ESLintUtils;
    exports.applyDefault = applyDefault;
    exports.deepMerge = deepMerge;
    exports.isObjectNotArray = isObjectNotArray;
    exports.getParserServices = getParserServices;
},
/* @typescript-eslint/eslint-plugin/dist/util/astUtils.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    const escapeRegExp_1 = require(908) /* ./escapeRegExp */;
    const ts = __importStar(require(21) /* typescript */);
    // deeply re-export, for convenience
    __exportStar(require(736) /* @typescript-eslint/experimental-utils/dist/ast-utils */, exports);
    // The following is copied from `eslint`'s source code since it doesn't exist in eslint@5.
    // https://github.com/eslint/eslint/blob/145aec1ab9052fbca96a44d04927c595951b1536/lib/rules/utils/ast-utils.js#L1751-L1779
    // Could be export { getNameLocationInGlobalDirectiveComment } from 'eslint/lib/rules/utils/ast-utils'
    /**
     * Get the `loc` object of a given name in a `/*globals` directive comment.
     * @param {SourceCode} sourceCode The source code to convert index to loc.
     * @param {Comment} comment The `/*globals` directive comment which include the name.
     * @param {string} name The name to find.
     * @returns {SourceLocation} The `loc` object.
     */
    function getNameLocationInGlobalDirectiveComment(sourceCode, comment, name) {
        const namePattern = new RegExp(`[\\s,]${escapeRegExp_1.escapeRegExp(name)}(?:$|[\\s,:])`, 'gu');
        // To ignore the first text "global".
        namePattern.lastIndex = comment.value.indexOf('global') + 6;
        // Search a given variable name.
        const match = namePattern.exec(comment.value);
        // Convert the index to loc.
        const start = sourceCode.getLocFromIndex(comment.range[0] + '/*'.length + (match ? match.index + 1 : 0));
        const end = {
            line: start.line,
            column: start.column + (match ? name.length : 1),
        };
        return { start, end };
    }
    exports.getNameLocationInGlobalDirectiveComment = getNameLocationInGlobalDirectiveComment;
    // Copied from typescript https://github.com/microsoft/TypeScript/blob/42b0e3c4630c129ca39ce0df9fff5f0d1b4dd348/src/compiler/utilities.ts#L1335
    // Warning: This has the same semantics as the forEach family of functions,
    //          in that traversal terminates in the event that 'visitor' supplies a truthy value.
    function forEachReturnStatement(body, visitor) {
        return traverse(body);
        function traverse(node) {
            switch (node.kind) {
                case ts.SyntaxKind.ReturnStatement:
                    return visitor(node);
                case ts.SyntaxKind.CaseBlock:
                case ts.SyntaxKind.Block:
                case ts.SyntaxKind.IfStatement:
                case ts.SyntaxKind.DoStatement:
                case ts.SyntaxKind.WhileStatement:
                case ts.SyntaxKind.ForStatement:
                case ts.SyntaxKind.ForInStatement:
                case ts.SyntaxKind.ForOfStatement:
                case ts.SyntaxKind.WithStatement:
                case ts.SyntaxKind.SwitchStatement:
                case ts.SyntaxKind.CaseClause:
                case ts.SyntaxKind.DefaultClause:
                case ts.SyntaxKind.LabeledStatement:
                case ts.SyntaxKind.TryStatement:
                case ts.SyntaxKind.CatchClause:
                    return ts.forEachChild(node, traverse);
            }
            return undefined;
        }
    }
    exports.forEachReturnStatement = forEachReturnStatement;
},
/* @typescript-eslint/eslint-plugin/dist/util/escapeRegExp.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    /**
     * Lodash <https://lodash.com/>
     * Released under MIT license <https://lodash.com/license>
     */
    const reRegExpChar = /[\\^$.*+?()[\]{}|]/g;
    const reHasRegExpChar = RegExp(reRegExpChar.source);
    function escapeRegExp(string = '') {
        return string && reHasRegExpChar.test(string)
            ? string.replace(reRegExpChar, '\\$&')
            : string;
    }
    exports.escapeRegExp = escapeRegExp;
},
/* @typescript-eslint/eslint-plugin/dist/util/collectUnusedVariables.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
        if (kind === "m")
            throw new TypeError("Private method is not writable");
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a setter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot write private member to an object whose class did not declare it");
        return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
    };
    var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
        if (kind === "a" && !f)
            throw new TypeError("Private accessor was defined without a getter");
        if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver))
            throw new TypeError("Cannot read private member from an object whose class did not declare it");
        return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
    };
    var _UnusedVarsVisitor_scopeManager;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const scope_manager_1 = require(768) /* @typescript-eslint/scope-manager */;
    const Visitor_1 = require(793) /* @typescript-eslint/scope-manager/dist/referencer/Visitor */;
    const util = __importStar(require(906) /* . */);
    class UnusedVarsVisitor extends Visitor_1.Visitor {
        // readonly #unusedVariables = new Set<TSESLint.Scope.Variable>();
        constructor(context) {
            super({
                visitChildrenEvenIfSelectorExists: true,
            });
            _UnusedVarsVisitor_scopeManager.set(this, void 0);
            //#endregion HELPERS
            //#region VISITORS
            // NOTE - This is a simple visitor - meaning it does not support selectors
            this.ClassDeclaration = this.visitClass;
            this.ClassExpression = this.visitClass;
            this.FunctionDeclaration = this.visitFunction;
            this.FunctionExpression = this.visitFunction;
            this.MethodDefinition = this.visitSetter;
            this.Property = this.visitSetter;
            this.TSCallSignatureDeclaration = this.visitFunctionTypeSignature;
            this.TSConstructorType = this.visitFunctionTypeSignature;
            this.TSConstructSignatureDeclaration = this.visitFunctionTypeSignature;
            this.TSDeclareFunction = this.visitFunctionTypeSignature;
            this.TSEmptyBodyFunctionExpression = this.visitFunctionTypeSignature;
            this.TSFunctionType = this.visitFunctionTypeSignature;
            this.TSMethodSignature = this.visitFunctionTypeSignature;
            __classPrivateFieldSet(this, _UnusedVarsVisitor_scopeManager, util.nullThrows(context.getSourceCode().scopeManager, 'Missing required scope manager'), "f");
        }
        static collectUnusedVariables(context) {
            const program = context.getSourceCode().ast;
            const cached = this.RESULTS_CACHE.get(program);
            if (cached) {
                return cached;
            }
            const visitor = new this(context);
            visitor.visit(program);
            const unusedVars = visitor.collectUnusedVariables(visitor.getScope(program));
            this.RESULTS_CACHE.set(program, unusedVars);
            return unusedVars;
        }
        collectUnusedVariables(scope, unusedVariables = new Set()) {
            for (const variable of scope.variables) {
                if (
                // skip function expression names,
                scope.functionExpressionScope ||
                    // variables marked with markVariableAsUsed(),
                    variable.eslintUsed ||
                    // implicit lib variables (from @typescript-eslint/scope-manager),
                    variable instanceof scope_manager_1.ImplicitLibVariable ||
                    // basic exported variables
                    isExported(variable) ||
                    // variables implicitly exported via a merged declaration
                    isMergableExported(variable) ||
                    // used variables
                    isUsedVariable(variable)) {
                    continue;
                }
                unusedVariables.add(variable);
            }
            for (const childScope of scope.childScopes) {
                this.collectUnusedVariables(childScope, unusedVariables);
            }
            return unusedVariables;
        }
        //#region HELPERS
        getScope(currentNode) {
            // On Program node, get the outermost scope to avoid return Node.js special function scope or ES modules scope.
            const inner = currentNode.type !== experimental_utils_1.AST_NODE_TYPES.Program;
            let node = currentNode;
            while (node) {
                const scope = __classPrivateFieldGet(this, _UnusedVarsVisitor_scopeManager, "f").acquire(node, inner);
                if (scope) {
                    if (scope.type === 'function-expression-name') {
                        return scope.childScopes[0];
                    }
                    return scope;
                }
                node = node.parent;
            }
            return __classPrivateFieldGet(this, _UnusedVarsVisitor_scopeManager, "f").scopes[0];
        }
        markVariableAsUsed(variableOrIdentifierOrName, parent) {
            if (typeof variableOrIdentifierOrName !== 'string' &&
                !('type' in variableOrIdentifierOrName)) {
                variableOrIdentifierOrName.eslintUsed = true;
                return;
            }
            let name;
            let node;
            if (typeof variableOrIdentifierOrName === 'string') {
                name = variableOrIdentifierOrName;
                node = parent;
            }
            else {
                name = variableOrIdentifierOrName.name;
                node = variableOrIdentifierOrName;
            }
            let currentScope = this.getScope(node);
            while (currentScope) {
                const variable = currentScope.variables.find(scopeVar => scopeVar.name === name);
                if (variable) {
                    variable.eslintUsed = true;
                    return;
                }
                currentScope = currentScope.upper;
            }
        }
        visitClass(node) {
            // skip a variable of class itself name in the class scope
            const scope = this.getScope(node);
            for (const variable of scope.variables) {
                if (variable.identifiers[0] === scope.block.id) {
                    this.markVariableAsUsed(variable);
                    return;
                }
            }
        }
        visitFunction(node) {
            const scope = this.getScope(node);
            // skip implicit "arguments" variable
            const variable = scope.set.get('arguments');
            if ((variable === null || variable === void 0 ? void 0 : variable.defs.length) === 0) {
                this.markVariableAsUsed(variable);
            }
        }
        visitFunctionTypeSignature(node) {
            // function type signature params create variables because they can be referenced within the signature,
            // but they obviously aren't unused variables for the purposes of this rule.
            for (const param of node.params) {
                this.visitPattern(param, name => {
                    this.markVariableAsUsed(name);
                });
            }
        }
        visitSetter(node) {
            if (node.kind === 'set') {
                // ignore setter parameters because they're syntactically required to exist
                for (const param of node.value.params) {
                    this.visitPattern(param, id => {
                        this.markVariableAsUsed(id);
                    });
                }
            }
        }
        ForInStatement(node) {
            /**
             * (Brad Zacher): I hate that this has to exist.
             * But it is required for compat with the base ESLint rule.
             *
             * In 2015, ESLint decided to add an exception for these two specific cases
             * ```
             * for (var key in object) return;
             *
             * var key;
             * for (key in object) return;
             * ```
             *
             * I disagree with it, but what are you going to do...
             *
             * https://github.com/eslint/eslint/issues/2342
             */
            let idOrVariable;
            if (node.left.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclaration) {
                const variable = __classPrivateFieldGet(this, _UnusedVarsVisitor_scopeManager, "f").getDeclaredVariables(node.left)[0];
                if (!variable) {
                    return;
                }
                idOrVariable = variable;
            }
            if (node.left.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                idOrVariable = node.left;
            }
            if (idOrVariable == null) {
                return;
            }
            let body = node.body;
            if (node.body.type === experimental_utils_1.AST_NODE_TYPES.BlockStatement) {
                if (node.body.body.length !== 1) {
                    return;
                }
                body = node.body.body[0];
            }
            if (body.type !== experimental_utils_1.AST_NODE_TYPES.ReturnStatement) {
                return;
            }
            this.markVariableAsUsed(idOrVariable);
        }
        Identifier(node) {
            const scope = this.getScope(node);
            if (scope.type === experimental_utils_1.TSESLint.Scope.ScopeType.function &&
                node.name === 'this') {
                // this parameters should always be considered used as they're pseudo-parameters
                if ('params' in scope.block && scope.block.params.includes(node)) {
                    this.markVariableAsUsed(node);
                }
            }
        }
        TSEnumDeclaration(node) {
            // enum members create variables because they can be referenced within the enum,
            // but they obviously aren't unused variables for the purposes of this rule.
            const scope = this.getScope(node);
            for (const variable of scope.variables) {
                this.markVariableAsUsed(variable);
            }
        }
        TSMappedType(node) {
            // mapped types create a variable for their type name, but it's not necessary to reference it,
            // so we shouldn't consider it as unused for the purpose of this rule.
            this.markVariableAsUsed(node.typeParameter.name);
        }
        TSModuleDeclaration(node) {
            // global augmentation can be in any file, and they do not need exports
            if (node.global === true) {
                this.markVariableAsUsed('global', node.parent);
            }
        }
        TSParameterProperty(node) {
            let identifier = null;
            switch (node.parameter.type) {
                case experimental_utils_1.AST_NODE_TYPES.AssignmentPattern:
                    if (node.parameter.left.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                        identifier = node.parameter.left;
                    }
                    break;
                case experimental_utils_1.AST_NODE_TYPES.Identifier:
                    identifier = node.parameter;
                    break;
            }
            if (identifier) {
                this.markVariableAsUsed(identifier);
            }
        }
    }
    _UnusedVarsVisitor_scopeManager = new WeakMap();
    UnusedVarsVisitor.RESULTS_CACHE = new WeakMap();
    //#region private helpers
    /**
     * Checks the position of given nodes.
     * @param inner A node which is expected as inside.
     * @param outer A node which is expected as outside.
     * @returns `true` if the `inner` node exists in the `outer` node.
     */
    function isInside(inner, outer) {
        return inner.range[0] >= outer.range[0] && inner.range[1] <= outer.range[1];
    }
    /**
     * Determine if an identifier is referencing an enclosing name.
     * This only applies to declarations that create their own scope (modules, functions, classes)
     * @param ref The reference to check.
     * @param nodes The candidate function nodes.
     * @returns True if it's a self-reference, false if not.
     */
    function isSelfReference(ref, nodes) {
        let scope = ref.from;
        while (scope) {
            if (nodes.has(scope.block)) {
                return true;
            }
            scope = scope.upper;
        }
        return false;
    }
    const MERGABLE_TYPES = new Set([
        experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration,
        experimental_utils_1.AST_NODE_TYPES.TSTypeAliasDeclaration,
        experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration,
        experimental_utils_1.AST_NODE_TYPES.ClassDeclaration,
        experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration,
    ]);
    /**
     * Determine if the variable is directly exported
     * @param variable the variable to check
     * @param target the type of node that is expected to be exported
     */
    function isMergableExported(variable) {
        var _a, _b;
        // If all of the merged things are of the same type, TS will error if not all of them are exported - so we only need to find one
        for (const def of variable.defs) {
            // parameters can never be exported.
            // their `node` prop points to the function decl, which can be exported
            // so we need to special case them
            if (def.type === experimental_utils_1.TSESLint.Scope.DefinitionType.Parameter) {
                continue;
            }
            if ((MERGABLE_TYPES.has(def.node.type) &&
                ((_a = def.node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.ExportNamedDeclaration) ||
                ((_b = def.node.parent) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.ExportDefaultDeclaration) {
                return true;
            }
        }
        return false;
    }
    /**
     * Determines if a given variable is being exported from a module.
     * @param variable eslint-scope variable object.
     * @returns True if the variable is exported, false if not.
     */
    function isExported(variable) {
        const definition = variable.defs[0];
        if (definition) {
            let node = definition.node;
            if (node.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclarator) {
                node = node.parent;
            }
            else if (definition.type === experimental_utils_1.TSESLint.Scope.DefinitionType.Parameter) {
                return false;
            }
            return node.parent.type.indexOf('Export') === 0;
        }
        return false;
    }
    /**
     * Determines if the variable is used.
     * @param variable The variable to check.
     * @returns True if the variable is used
     */
    function isUsedVariable(variable) {
        /**
         * Gets a list of function definitions for a specified variable.
         * @param variable eslint-scope variable object.
         * @returns Function nodes.
         */
        function getFunctionDefinitions(variable) {
            const functionDefinitions = new Set();
            variable.defs.forEach(def => {
                var _a, _b;
                // FunctionDeclarations
                if (def.type === experimental_utils_1.TSESLint.Scope.DefinitionType.FunctionName) {
                    functionDefinitions.add(def.node);
                }
                // FunctionExpressions
                if (def.type === experimental_utils_1.TSESLint.Scope.DefinitionType.Variable &&
                    (((_a = def.node.init) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.FunctionExpression ||
                        ((_b = def.node.init) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression)) {
                    functionDefinitions.add(def.node.init);
                }
            });
            return functionDefinitions;
        }
        function getTypeDeclarations(variable) {
            const nodes = new Set();
            variable.defs.forEach(def => {
                if (def.node.type === experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration ||
                    def.node.type === experimental_utils_1.AST_NODE_TYPES.TSTypeAliasDeclaration) {
                    nodes.add(def.node);
                }
            });
            return nodes;
        }
        function getModuleDeclarations(variable) {
            const nodes = new Set();
            variable.defs.forEach(def => {
                if (def.node.type === experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration) {
                    nodes.add(def.node);
                }
            });
            return nodes;
        }
        /**
         * Checks if the ref is contained within one of the given nodes
         */
        function isInsideOneOf(ref, nodes) {
            for (const node of nodes) {
                if (isInside(ref.identifier, node)) {
                    return true;
                }
            }
            return false;
        }
        /**
         * If a given reference is left-hand side of an assignment, this gets
         * the right-hand side node of the assignment.
         *
         * In the following cases, this returns null.
         *
         * - The reference is not the LHS of an assignment expression.
         * - The reference is inside of a loop.
         * - The reference is inside of a function scope which is different from
         *   the declaration.
         * @param ref A reference to check.
         * @param prevRhsNode The previous RHS node.
         * @returns The RHS node or null.
         */
        function getRhsNode(ref, prevRhsNode) {
            /**
             * Checks whether the given node is in a loop or not.
             * @param node The node to check.
             * @returns `true` if the node is in a loop.
             */
            function isInLoop(node) {
                let currentNode = node;
                while (currentNode) {
                    if (util.isFunction(currentNode)) {
                        break;
                    }
                    if (util.isLoop(currentNode)) {
                        return true;
                    }
                    currentNode = currentNode.parent;
                }
                return false;
            }
            const id = ref.identifier;
            const parent = id.parent;
            const grandparent = parent.parent;
            const refScope = ref.from.variableScope;
            const varScope = ref.resolved.scope.variableScope;
            const canBeUsedLater = refScope !== varScope || isInLoop(id);
            /*
             * Inherits the previous node if this reference is in the node.
             * This is for `a = a + a`-like code.
             */
            if (prevRhsNode && isInside(id, prevRhsNode)) {
                return prevRhsNode;
            }
            if (parent.type === experimental_utils_1.AST_NODE_TYPES.AssignmentExpression &&
                grandparent.type === experimental_utils_1.AST_NODE_TYPES.ExpressionStatement &&
                id === parent.left &&
                !canBeUsedLater) {
                return parent.right;
            }
            return null;
        }
        /**
         * Checks whether a given reference is a read to update itself or not.
         * @param ref A reference to check.
         * @param rhsNode The RHS node of the previous assignment.
         * @returns The reference is a read to update itself.
         */
        function isReadForItself(ref, rhsNode) {
            /**
             * Checks whether a given Identifier node exists inside of a function node which can be used later.
             *
             * "can be used later" means:
             * - the function is assigned to a variable.
             * - the function is bound to a property and the object can be used later.
             * - the function is bound as an argument of a function call.
             *
             * If a reference exists in a function which can be used later, the reference is read when the function is called.
             * @param id An Identifier node to check.
             * @param rhsNode The RHS node of the previous assignment.
             * @returns `true` if the `id` node exists inside of a function node which can be used later.
             */
            function isInsideOfStorableFunction(id, rhsNode) {
                /**
                 * Finds a function node from ancestors of a node.
                 * @param node A start node to find.
                 * @returns A found function node.
                 */
                function getUpperFunction(node) {
                    let currentNode = node;
                    while (currentNode) {
                        if (util.isFunction(currentNode)) {
                            return currentNode;
                        }
                        currentNode = currentNode.parent;
                    }
                    return null;
                }
                /**
                 * Checks whether a given function node is stored to somewhere or not.
                 * If the function node is stored, the function can be used later.
                 * @param funcNode A function node to check.
                 * @param rhsNode The RHS node of the previous assignment.
                 * @returns `true` if under the following conditions:
                 *      - the funcNode is assigned to a variable.
                 *      - the funcNode is bound as an argument of a function call.
                 *      - the function is bound to a property and the object satisfies above conditions.
                 */
                function isStorableFunction(funcNode, rhsNode) {
                    let node = funcNode;
                    let parent = funcNode.parent;
                    while (parent && isInside(parent, rhsNode)) {
                        switch (parent.type) {
                            case experimental_utils_1.AST_NODE_TYPES.SequenceExpression:
                                if (parent.expressions[parent.expressions.length - 1] !== node) {
                                    return false;
                                }
                                break;
                            case experimental_utils_1.AST_NODE_TYPES.CallExpression:
                            case experimental_utils_1.AST_NODE_TYPES.NewExpression:
                                return parent.callee !== node;
                            case experimental_utils_1.AST_NODE_TYPES.AssignmentExpression:
                            case experimental_utils_1.AST_NODE_TYPES.TaggedTemplateExpression:
                            case experimental_utils_1.AST_NODE_TYPES.YieldExpression:
                                return true;
                            default:
                                if (parent.type.endsWith('Statement') ||
                                    parent.type.endsWith('Declaration')) {
                                    /*
                                     * If it encountered statements, this is a complex pattern.
                                     * Since analyzing complex patterns is hard, this returns `true` to avoid false positive.
                                     */
                                    return true;
                                }
                        }
                        node = parent;
                        parent = parent.parent;
                    }
                    return false;
                }
                const funcNode = getUpperFunction(id);
                return (!!funcNode &&
                    isInside(funcNode, rhsNode) &&
                    isStorableFunction(funcNode, rhsNode));
            }
            const id = ref.identifier;
            const parent = id.parent;
            const grandparent = parent.parent;
            return (ref.isRead() && // in RHS of an assignment for itself. e.g. `a = a + 1`
                // self update. e.g. `a += 1`, `a++`
                ((parent.type === experimental_utils_1.AST_NODE_TYPES.AssignmentExpression &&
                    grandparent.type === experimental_utils_1.AST_NODE_TYPES.ExpressionStatement &&
                    parent.left === id) ||
                    (parent.type === experimental_utils_1.AST_NODE_TYPES.UpdateExpression &&
                        grandparent.type === experimental_utils_1.AST_NODE_TYPES.ExpressionStatement) ||
                    (!!rhsNode &&
                        isInside(id, rhsNode) &&
                        !isInsideOfStorableFunction(id, rhsNode))));
        }
        const functionNodes = getFunctionDefinitions(variable);
        const isFunctionDefinition = functionNodes.size > 0;
        const typeDeclNodes = getTypeDeclarations(variable);
        const isTypeDecl = typeDeclNodes.size > 0;
        const moduleDeclNodes = getModuleDeclarations(variable);
        const isModuleDecl = moduleDeclNodes.size > 0;
        let rhsNode = null;
        return variable.references.some(ref => {
            const forItself = isReadForItself(ref, rhsNode);
            rhsNode = getRhsNode(ref, rhsNode);
            return (ref.isRead() &&
                !forItself &&
                !(isFunctionDefinition && isSelfReference(ref, functionNodes)) &&
                !(isTypeDecl && isInsideOneOf(ref, typeDeclNodes)) &&
                !(isModuleDecl && isSelfReference(ref, moduleDeclNodes)));
        });
    }
    //#endregion private helpers
    /**
     * Collects the set of unused variables for a given context.
     *
     * Due to complexity, this does not take into consideration:
     * - variables within declaration files
     * - variables within ambient module declarations
     */
    function collectUnusedVariables(context) {
        return UnusedVarsVisitor.collectUnusedVariables(context);
    }
    exports.collectUnusedVariables = collectUnusedVariables;
},
/* @typescript-eslint/eslint-plugin/dist/util/createRule.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    // note - cannot migrate this to an import statement because it will make TSC copy the package.json to the dist folder
    // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-member-access
    const version = require(911) /* ../../package.json */.version;
    exports.createRule = experimental_utils_1.ESLintUtils.RuleCreator(name => `https://github.com/typescript-eslint/typescript-eslint/blob/v${version}/packages/eslint-plugin/docs/rules/${name}.md`);
},
/* @typescript-eslint/eslint-plugin/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "@typescript-eslint/eslint-plugin",
        "version": "4.29.1",
        "description": "TypeScript plugin for ESLint",
        "keywords": [
            "eslint",
            "eslintplugin",
            "eslint-plugin",
            "typescript"
        ],
        "engines": {
            "node": "^10.12.0 || >=12.0.0"
        },
        "files": [
            "dist",
            "docs",
            "index.d.ts",
            "package.json",
            "README.md",
            "LICENSE"
        ],
        "repository": {
            "type": "git",
            "url": "https://github.com/typescript-eslint/typescript-eslint.git",
            "directory": "packages/eslint-plugin"
        },
        "bugs": {
            "url": "https://github.com/typescript-eslint/typescript-eslint/issues"
        },
        "license": "MIT",
        "main": "dist/index.js",
        "types": "index.d.ts",
        "scripts": {
            "build": "tsc -b tsconfig.build.json",
            "check-docs": "jest tests/docs.test.ts --runTestsByPath --silent --runInBand",
            "check-configs": "jest tests/configs.test.ts --runTestsByPath --silent --runInBand",
            "clean": "tsc -b tsconfig.build.json --clean",
            "postclean": "rimraf dist && rimraf coverage",
            "format": "prettier --write \"./**/*.{ts,js,json,md}\" --ignore-path ../../.prettierignore",
            "generate:configs": "../../node_modules/.bin/ts-node --files --transpile-only tools/generate-configs.ts",
            "generate:rules-lists": "../../node_modules/.bin/ts-node --files --transpile-only tools/generate-rules-lists.ts",
            "lint": "eslint . --ext .js,.ts --ignore-path ../../.eslintignore",
            "test": "jest --coverage",
            "typecheck": "tsc -p tsconfig.json --noEmit"
        },
        "dependencies": {
            "@typescript-eslint/experimental-utils": "4.29.1",
            "@typescript-eslint/scope-manager": "4.29.1",
            "debug": "^4.3.1",
            "functional-red-black-tree": "^1.0.1",
            "regexpp": "^3.1.0",
            "semver": "^7.3.5",
            "tsutils": "^3.21.0"
        },
        "devDependencies": {
            "@types/debug": "*",
            "@types/marked": "*",
            "@types/prettier": "*",
            "chalk": "^4.1.1",
            "marked": "^2.0.6",
            "prettier": "*",
            "typescript": "*"
        },
        "peerDependencies": {
            "@typescript-eslint/parser": "^4.0.0",
            "eslint": "^5.0.0 || ^6.0.0 || ^7.0.0"
        },
        "peerDependenciesMeta": {
            "typescript": {
                "optional": true
            }
        },
        "funding": {
            "type": "opencollective",
            "url": "https://opencollective.com/typescript-eslint"
        },
        "gitHead": "db78642ea3ebd794e7deef767fa2c5195ef7c04c"
    };
    module.exports = json;
},
/* @typescript-eslint/eslint-plugin/dist/util/getFunctionHeadLoc.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    /**
     * Creates a report location for the given function.
     * The location only encompasses the "start" of the function, and not the body
     *
     * eg.
     *
     * ```
     * function foo(args) {}
     * ^^^^^^^^^^^^^^^^^^
     *
     * get y(args) {}
     * ^^^^^^^^^^^
     *
     * const x = (args) => {}
     *           ^^^^^^^^^
     * ```
     */
    function getFunctionHeadLoc(node, sourceCode) {
        function getLocStart() {
            if (node.parent && node.parent.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition) {
                // return the start location for class method
                if (node.parent.decorators && node.parent.decorators.length > 0) {
                    // exclude decorators
                    return sourceCode.getTokenAfter(node.parent.decorators[node.parent.decorators.length - 1]).loc.start;
                }
                return node.parent.loc.start;
            }
            if (node.parent &&
                node.parent.type === experimental_utils_1.AST_NODE_TYPES.Property &&
                node.parent.method) {
                // return the start location for object method shorthand
                return node.parent.loc.start;
            }
            // return the start location for a regular function
            return node.loc.start;
        }
        function getLocEnd() {
            if (node.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression) {
                // find the end location for arrow function expression
                return sourceCode.getTokenBefore(node.body, token => token.type === experimental_utils_1.AST_TOKEN_TYPES.Punctuator && token.value === '=>').loc.end;
            }
            // return the end location for a regular function
            return sourceCode.getTokenBefore(node.body).loc.end;
        }
        return {
            start: getLocStart(),
            end: getLocEnd(),
        };
    }
    exports.getFunctionHeadLoc = getFunctionHeadLoc;
},
/* @typescript-eslint/eslint-plugin/dist/util/getThisExpression.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    function getThisExpression(node) {
        while (node) {
            if (node.type === experimental_utils_1.AST_NODE_TYPES.CallExpression) {
                node = node.callee;
            }
            else if (node.type === experimental_utils_1.AST_NODE_TYPES.ThisExpression) {
                return node;
            }
            else if (node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression) {
                node = node.object;
            }
            else if (node.type === experimental_utils_1.AST_NODE_TYPES.ChainExpression) {
                node = node.expression;
            }
            else {
                break;
            }
        }
        return;
    }
    exports.getThisExpression = getThisExpression;
},
/* @typescript-eslint/eslint-plugin/dist/util/getWrappingFixer.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    /**
     * Wraps node with some code. Adds parenthesis as necessary.
     * @returns Fixer which adds the specified code and parens if necessary.
     */
    function getWrappingFixer(params) {
        const { sourceCode, node, innerNode = node, wrap } = params;
        const innerNodes = Array.isArray(innerNode) ? innerNode : [innerNode];
        return (fixer) => {
            const innerCodes = innerNodes.map(innerNode => {
                let code = sourceCode.getText(innerNode);
                // check the inner expression's precedence
                if (!isStrongPrecedenceNode(innerNode)) {
                    // the code we are adding might have stronger precedence than our wrapped node
                    // let's wrap our node in parens in case it has a weaker precedence than the code we are wrapping it in
                    code = `(${code})`;
                }
                return code;
            });
            // do the wrapping
            let code = wrap(...innerCodes);
            // check the outer expression's precedence
            if (isWeakPrecedenceParent(node)) {
                // we wrapped the node in some expression which very likely has a different precedence than original wrapped node
                // let's wrap the whole expression in parens just in case
                if (!util.isParenthesized(node, sourceCode)) {
                    code = `(${code})`;
                }
            }
            // check if we need to insert semicolon
            if (/^[`([]/.exec(code) && isMissingSemicolonBefore(node, sourceCode)) {
                code = `;${code}`;
            }
            return fixer.replaceText(node, code);
        };
    }
    exports.getWrappingFixer = getWrappingFixer;
    /**
     * Check if a node will always have the same precedence if it's parent changes.
     */
    function isStrongPrecedenceNode(innerNode) {
        return (innerNode.type === experimental_utils_1.AST_NODE_TYPES.Literal ||
            innerNode.type === experimental_utils_1.AST_NODE_TYPES.Identifier ||
            innerNode.type === experimental_utils_1.AST_NODE_TYPES.ArrayExpression ||
            innerNode.type === experimental_utils_1.AST_NODE_TYPES.ObjectExpression ||
            innerNode.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression ||
            innerNode.type === experimental_utils_1.AST_NODE_TYPES.CallExpression ||
            innerNode.type === experimental_utils_1.AST_NODE_TYPES.NewExpression ||
            innerNode.type === experimental_utils_1.AST_NODE_TYPES.TaggedTemplateExpression);
    }
    /**
     * Check if a node's parent could have different precedence if the node changes.
     */
    function isWeakPrecedenceParent(node) {
        const parent = node.parent;
        if (parent.type === experimental_utils_1.AST_NODE_TYPES.UpdateExpression ||
            parent.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression ||
            parent.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression ||
            parent.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression ||
            parent.type === experimental_utils_1.AST_NODE_TYPES.ConditionalExpression ||
            parent.type === experimental_utils_1.AST_NODE_TYPES.AwaitExpression) {
            return true;
        }
        if (parent.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
            parent.object === node) {
            return true;
        }
        if ((parent.type === experimental_utils_1.AST_NODE_TYPES.CallExpression ||
            parent.type === experimental_utils_1.AST_NODE_TYPES.NewExpression) &&
            parent.callee === node) {
            return true;
        }
        if (parent.type === experimental_utils_1.AST_NODE_TYPES.TaggedTemplateExpression &&
            parent.tag === node) {
            return true;
        }
        return false;
    }
    /**
     * Returns true if a node is at the beginning of expression statement and the statement above doesn't end with semicolon.
     * Doesn't check if the node begins with `(`, `[` or `` ` ``.
     */
    function isMissingSemicolonBefore(node, sourceCode) {
        for (;;) {
            const parent = node.parent;
            if (parent.type === experimental_utils_1.AST_NODE_TYPES.ExpressionStatement) {
                const block = parent.parent;
                if (block.type === experimental_utils_1.AST_NODE_TYPES.Program ||
                    block.type === experimental_utils_1.AST_NODE_TYPES.BlockStatement) {
                    // parent is an expression statement in a block
                    const statementIndex = block.body.indexOf(parent);
                    const previousStatement = block.body[statementIndex - 1];
                    if (statementIndex > 0 &&
                        sourceCode.getLastToken(previousStatement).value !== ';') {
                        return true;
                    }
                }
            }
            if (!isLeftHandSide(node)) {
                return false;
            }
            node = parent;
        }
    }
    /**
     * Checks if a node is LHS of an operator.
     */
    function isLeftHandSide(node) {
        const parent = node.parent;
        // a++
        if (parent.type === experimental_utils_1.AST_NODE_TYPES.UpdateExpression) {
            return true;
        }
        // a + b
        if ((parent.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression ||
            parent.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression ||
            parent.type === experimental_utils_1.AST_NODE_TYPES.AssignmentExpression) &&
            node === parent.left) {
            return true;
        }
        // a ? b : c
        if (parent.type === experimental_utils_1.AST_NODE_TYPES.ConditionalExpression &&
            node === parent.test) {
            return true;
        }
        // a(b)
        if (parent.type === experimental_utils_1.AST_NODE_TYPES.CallExpression && node === parent.callee) {
            return true;
        }
        // a`b`
        if (parent.type === experimental_utils_1.AST_NODE_TYPES.TaggedTemplateExpression &&
            node === parent.tag) {
            return true;
        }
        return false;
    }
},
/* @typescript-eslint/eslint-plugin/dist/util/isTypeReadonly.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const tsutils_1 = require(916) /* tsutils */;
    const ts = __importStar(require(21) /* typescript */);
    const _1 = require(906) /* . */;
    function isTypeReadonlyArrayOrTuple(checker, type, seenTypes) {
        function checkTypeArguments(arrayType) {
            const typeArguments = checker.getTypeArguments(arrayType);
            // this shouldn't happen in reality as:
            // - tuples require at least 1 type argument
            // - ReadonlyArray requires at least 1 type argument
            /* istanbul ignore if */ if (typeArguments.length === 0) {
                return 3 /* Readonly */;
            }
            // validate the element types are also readonly
            if (typeArguments.some(typeArg => isTypeReadonlyRecurser(checker, typeArg, seenTypes) ===
                2 /* Mutable */)) {
                return 2 /* Mutable */;
            }
            return 3 /* Readonly */;
        }
        if (checker.isArrayType(type)) {
            const symbol = _1.nullThrows(type.getSymbol(), _1.NullThrowsReasons.MissingToken('symbol', 'array type'));
            const escapedName = symbol.getEscapedName();
            if (escapedName === 'Array') {
                return 2 /* Mutable */;
            }
            return checkTypeArguments(type);
        }
        if (checker.isTupleType(type)) {
            if (!type.target.readonly) {
                return 2 /* Mutable */;
            }
            return checkTypeArguments(type);
        }
        return 1 /* UnknownType */;
    }
    function isTypeReadonlyObject(checker, type, seenTypes) {
        function checkIndexSignature(kind) {
            const indexInfo = checker.getIndexInfoOfType(type, kind);
            if (indexInfo) {
                return indexInfo.isReadonly
                    ? 3 /* Readonly */
                    : 2 /* Mutable */;
            }
            return 1 /* UnknownType */;
        }
        const properties = type.getProperties();
        if (properties.length) {
            // ensure the properties are marked as readonly
            for (const property of properties) {
                if (!tsutils_1.isPropertyReadonlyInType(type, property.getEscapedName(), checker)) {
                    return 2 /* Mutable */;
                }
            }
            // all properties were readonly
            // now ensure that all of the values are readonly also.
            // do this after checking property readonly-ness as a perf optimization,
            // as we might be able to bail out early due to a mutable property before
            // doing this deep, potentially expensive check.
            for (const property of properties) {
                const propertyType = _1.nullThrows(_1.getTypeOfPropertyOfType(checker, type, property), _1.NullThrowsReasons.MissingToken(`property "${property.name}"`, 'type'));
                // handle recursive types.
                // we only need this simple check, because a mutable recursive type will break via the above prop readonly check
                if (seenTypes.has(propertyType)) {
                    continue;
                }
                if (isTypeReadonlyRecurser(checker, propertyType, seenTypes) ===
                    2 /* Mutable */) {
                    return 2 /* Mutable */;
                }
            }
        }
        const isStringIndexSigReadonly = checkIndexSignature(ts.IndexKind.String);
        if (isStringIndexSigReadonly === 2 /* Mutable */) {
            return isStringIndexSigReadonly;
        }
        const isNumberIndexSigReadonly = checkIndexSignature(ts.IndexKind.Number);
        if (isNumberIndexSigReadonly === 2 /* Mutable */) {
            return isNumberIndexSigReadonly;
        }
        return 3 /* Readonly */;
    }
    // a helper function to ensure the seenTypes map is always passed down, except by the external caller
    function isTypeReadonlyRecurser(checker, type, seenTypes) {
        seenTypes.add(type);
        if (tsutils_1.isUnionType(type)) {
            // all types in the union must be readonly
            const result = tsutils_1.unionTypeParts(type).every(t => isTypeReadonlyRecurser(checker, t, seenTypes));
            const readonlyness = result ? 3 /* Readonly */ : 2 /* Mutable */;
            return readonlyness;
        }
        // all non-object, non-intersection types are readonly.
        // this should only be primitive types
        if (!tsutils_1.isObjectType(type) && !tsutils_1.isUnionOrIntersectionType(type)) {
            return 3 /* Readonly */;
        }
        // pure function types are readonly
        if (type.getCallSignatures().length > 0 &&
            type.getProperties().length === 0) {
            return 3 /* Readonly */;
        }
        const isReadonlyArray = isTypeReadonlyArrayOrTuple(checker, type, seenTypes);
        if (isReadonlyArray !== 1 /* UnknownType */) {
            return isReadonlyArray;
        }
        const isReadonlyObject = isTypeReadonlyObject(checker, type, seenTypes);
        /* istanbul ignore else */ if (isReadonlyObject !== 1 /* UnknownType */) {
            return isReadonlyObject;
        }
        throw new Error('Unhandled type');
    }
    /**
     * Checks if the given type is readonly
     */
    function isTypeReadonly(checker, type) {
        return (isTypeReadonlyRecurser(checker, type, new Set()) === 3 /* Readonly */);
    }
    exports.isTypeReadonly = isTypeReadonly;
},
/* tsutils/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(917) /* ./typeguard */, exports);
    tslib_1.__exportStar(require(928) /* ./util */, exports);
},
/* tsutils/typeguard/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(918) /* ./node */, exports);
    tslib_1.__exportStar(require(923) /* ./type */, exports);
},
/* tsutils/typeguard/node.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(919) /* ./3.2/node */, exports);
},
/* tsutils/typeguard/3.2/node.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(920) /* ../3.0/node */, exports);
    const ts = require(21) /* typescript */;
    function isBigIntLiteral(node) {
        return node.kind === ts.SyntaxKind.BigIntLiteral;
    }
    exports.isBigIntLiteral = isBigIntLiteral;
},
/* tsutils/typeguard/3.0/node.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(921) /* ../2.9/node */, exports);
    const ts = require(21) /* typescript */;
    function isOptionalTypeNode(node) {
        return node.kind === ts.SyntaxKind.OptionalType;
    }
    exports.isOptionalTypeNode = isOptionalTypeNode;
    function isRestTypeNode(node) {
        return node.kind === ts.SyntaxKind.RestType;
    }
    exports.isRestTypeNode = isRestTypeNode;
    function isSyntheticExpression(node) {
        return node.kind === ts.SyntaxKind.SyntheticExpression;
    }
    exports.isSyntheticExpression = isSyntheticExpression;
},
/* tsutils/typeguard/2.9/node.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(922) /* ../2.8/node */, exports);
    const ts = require(21) /* typescript */;
    function isImportTypeNode(node) {
        return node.kind === ts.SyntaxKind.ImportType;
    }
    exports.isImportTypeNode = isImportTypeNode;
},
/* tsutils/typeguard/2.8/node.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.isNamespaceImport = exports.isNamespaceDeclaration = exports.isNamedImports = exports.isNamedExports = exports.isModuleDeclaration = exports.isModuleBlock = exports.isMethodSignature = exports.isMethodDeclaration = exports.isMetaProperty = exports.isMappedTypeNode = exports.isLiteralTypeNode = exports.isLiteralExpression = exports.isLabeledStatement = exports.isJsxText = exports.isJsxSpreadAttribute = exports.isJsxSelfClosingElement = exports.isJsxOpeningLikeElement = exports.isJsxOpeningFragment = exports.isJsxOpeningElement = exports.isJsxFragment = exports.isJsxExpression = exports.isJsxElement = exports.isJsxClosingFragment = exports.isJsxClosingElement = exports.isJsxAttributes = exports.isJsxAttributeLike = exports.isJsxAttribute = exports.isJsDoc = exports.isIterationStatement = exports.isIntersectionTypeNode = exports.isInterfaceDeclaration = exports.isInferTypeNode = exports.isIndexSignatureDeclaration = exports.isIndexedAccessTypeNode = exports.isImportSpecifier = exports.isImportEqualsDeclaration = exports.isImportDeclaration = exports.isImportClause = exports.isIfStatement = exports.isIdentifier = exports.isGetAccessorDeclaration = exports.isFunctionTypeNode = exports.isFunctionExpression = exports.isFunctionDeclaration = exports.isForStatement = exports.isForOfStatement = exports.isForInOrOfStatement = exports.isForInStatement = exports.isExternalModuleReference = exports.isExpressionWithTypeArguments = void 0;
    exports.isVariableStatement = exports.isVariableDeclaration = exports.isUnionTypeNode = exports.isTypeQueryNode = exports.isTypeReferenceNode = exports.isTypePredicateNode = exports.isTypeParameterDeclaration = exports.isTypeOperatorNode = exports.isTypeOfExpression = exports.isTypeLiteralNode = exports.isTypeAssertion = exports.isTypeAliasDeclaration = exports.isTupleTypeNode = exports.isTryStatement = exports.isThrowStatement = exports.isTextualLiteral = exports.isTemplateLiteral = exports.isTemplateExpression = exports.isTaggedTemplateExpression = exports.isSyntaxList = exports.isSwitchStatement = exports.isStringLiteral = exports.isSpreadElement = exports.isSpreadAssignment = exports.isSourceFile = exports.isSignatureDeclaration = exports.isShorthandPropertyAssignment = exports.isSetAccessorDeclaration = exports.isReturnStatement = exports.isRegularExpressionLiteral = exports.isQualifiedName = exports.isPropertySignature = exports.isPropertyDeclaration = exports.isPropertyAssignment = exports.isPropertyAccessExpression = exports.isPrefixUnaryExpression = exports.isPostfixUnaryExpression = exports.isParenthesizedTypeNode = exports.isParenthesizedExpression = exports.isParameterDeclaration = exports.isOmittedExpression = exports.isObjectLiteralExpression = exports.isObjectBindingPattern = exports.isNumericOrStringLikeLiteral = exports.isNumericLiteral = exports.isNullLiteral = exports.isNoSubstitutionTemplateLiteral = exports.isNonNullExpression = exports.isNewExpression = exports.isNamespaceExportDeclaration = void 0;
    exports.isWithStatement = exports.isWhileStatement = exports.isVoidExpression = exports.isVariableDeclarationList = void 0;
    const ts = require(21) /* typescript */;
    function isAccessorDeclaration(node) {
        return node.kind === ts.SyntaxKind.GetAccessor ||
            node.kind === ts.SyntaxKind.SetAccessor;
    }
    exports.isAccessorDeclaration = isAccessorDeclaration;
    function isArrayBindingPattern(node) {
        return node.kind === ts.SyntaxKind.ArrayBindingPattern;
    }
    exports.isArrayBindingPattern = isArrayBindingPattern;
    function isArrayLiteralExpression(node) {
        return node.kind === ts.SyntaxKind.ArrayLiteralExpression;
    }
    exports.isArrayLiteralExpression = isArrayLiteralExpression;
    function isArrayTypeNode(node) {
        return node.kind === ts.SyntaxKind.ArrayType;
    }
    exports.isArrayTypeNode = isArrayTypeNode;
    function isArrowFunction(node) {
        return node.kind === ts.SyntaxKind.ArrowFunction;
    }
    exports.isArrowFunction = isArrowFunction;
    function isAsExpression(node) {
        return node.kind === ts.SyntaxKind.AsExpression;
    }
    exports.isAsExpression = isAsExpression;
    function isAssertionExpression(node) {
        return node.kind === ts.SyntaxKind.AsExpression ||
            node.kind === ts.SyntaxKind.TypeAssertionExpression;
    }
    exports.isAssertionExpression = isAssertionExpression;
    function isAwaitExpression(node) {
        return node.kind === ts.SyntaxKind.AwaitExpression;
    }
    exports.isAwaitExpression = isAwaitExpression;
    function isBinaryExpression(node) {
        return node.kind === ts.SyntaxKind.BinaryExpression;
    }
    exports.isBinaryExpression = isBinaryExpression;
    function isBindingElement(node) {
        return node.kind === ts.SyntaxKind.BindingElement;
    }
    exports.isBindingElement = isBindingElement;
    function isBindingPattern(node) {
        return node.kind === ts.SyntaxKind.ArrayBindingPattern ||
            node.kind === ts.SyntaxKind.ObjectBindingPattern;
    }
    exports.isBindingPattern = isBindingPattern;
    function isBlock(node) {
        return node.kind === ts.SyntaxKind.Block;
    }
    exports.isBlock = isBlock;
    function isBlockLike(node) {
        return node.statements !== undefined;
    }
    exports.isBlockLike = isBlockLike;
    function isBooleanLiteral(node) {
        return node.kind === ts.SyntaxKind.TrueKeyword || node.kind === ts.SyntaxKind.FalseKeyword;
    }
    exports.isBooleanLiteral = isBooleanLiteral;
    function isBreakOrContinueStatement(node) {
        return node.kind === ts.SyntaxKind.BreakStatement ||
            node.kind === ts.SyntaxKind.ContinueStatement;
    }
    exports.isBreakOrContinueStatement = isBreakOrContinueStatement;
    function isBreakStatement(node) {
        return node.kind === ts.SyntaxKind.BreakStatement;
    }
    exports.isBreakStatement = isBreakStatement;
    function isCallExpression(node) {
        return node.kind === ts.SyntaxKind.CallExpression;
    }
    exports.isCallExpression = isCallExpression;
    function isCallLikeExpression(node) {
        switch (node.kind) {
            case ts.SyntaxKind.CallExpression:
            case ts.SyntaxKind.Decorator:
            case ts.SyntaxKind.JsxOpeningElement:
            case ts.SyntaxKind.JsxSelfClosingElement:
            case ts.SyntaxKind.NewExpression:
            case ts.SyntaxKind.TaggedTemplateExpression:
                return true;
            default:
                return false;
        }
    }
    exports.isCallLikeExpression = isCallLikeExpression;
    function isCallSignatureDeclaration(node) {
        return node.kind === ts.SyntaxKind.CallSignature;
    }
    exports.isCallSignatureDeclaration = isCallSignatureDeclaration;
    function isCaseBlock(node) {
        return node.kind === ts.SyntaxKind.CaseBlock;
    }
    exports.isCaseBlock = isCaseBlock;
    function isCaseClause(node) {
        return node.kind === ts.SyntaxKind.CaseClause;
    }
    exports.isCaseClause = isCaseClause;
    function isCaseOrDefaultClause(node) {
        return node.kind === ts.SyntaxKind.CaseClause ||
            node.kind === ts.SyntaxKind.DefaultClause;
    }
    exports.isCaseOrDefaultClause = isCaseOrDefaultClause;
    function isCatchClause(node) {
        return node.kind === ts.SyntaxKind.CatchClause;
    }
    exports.isCatchClause = isCatchClause;
    function isClassDeclaration(node) {
        return node.kind === ts.SyntaxKind.ClassDeclaration;
    }
    exports.isClassDeclaration = isClassDeclaration;
    function isClassExpression(node) {
        return node.kind === ts.SyntaxKind.ClassExpression;
    }
    exports.isClassExpression = isClassExpression;
    function isClassLikeDeclaration(node) {
        return node.kind === ts.SyntaxKind.ClassDeclaration ||
            node.kind === ts.SyntaxKind.ClassExpression;
    }
    exports.isClassLikeDeclaration = isClassLikeDeclaration;
    function isCommaListExpression(node) {
        return node.kind === ts.SyntaxKind.CommaListExpression;
    }
    exports.isCommaListExpression = isCommaListExpression;
    function isConditionalExpression(node) {
        return node.kind === ts.SyntaxKind.ConditionalExpression;
    }
    exports.isConditionalExpression = isConditionalExpression;
    function isConditionalTypeNode(node) {
        return node.kind === ts.SyntaxKind.ConditionalType;
    }
    exports.isConditionalTypeNode = isConditionalTypeNode;
    function isConstructorDeclaration(node) {
        return node.kind === ts.SyntaxKind.Constructor;
    }
    exports.isConstructorDeclaration = isConstructorDeclaration;
    function isConstructorTypeNode(node) {
        return node.kind === ts.SyntaxKind.ConstructorType;
    }
    exports.isConstructorTypeNode = isConstructorTypeNode;
    function isConstructSignatureDeclaration(node) {
        return node.kind === ts.SyntaxKind.ConstructSignature;
    }
    exports.isConstructSignatureDeclaration = isConstructSignatureDeclaration;
    function isContinueStatement(node) {
        return node.kind === ts.SyntaxKind.ContinueStatement;
    }
    exports.isContinueStatement = isContinueStatement;
    function isComputedPropertyName(node) {
        return node.kind === ts.SyntaxKind.ComputedPropertyName;
    }
    exports.isComputedPropertyName = isComputedPropertyName;
    function isDebuggerStatement(node) {
        return node.kind === ts.SyntaxKind.DebuggerStatement;
    }
    exports.isDebuggerStatement = isDebuggerStatement;
    function isDecorator(node) {
        return node.kind === ts.SyntaxKind.Decorator;
    }
    exports.isDecorator = isDecorator;
    function isDefaultClause(node) {
        return node.kind === ts.SyntaxKind.DefaultClause;
    }
    exports.isDefaultClause = isDefaultClause;
    function isDeleteExpression(node) {
        return node.kind === ts.SyntaxKind.DeleteExpression;
    }
    exports.isDeleteExpression = isDeleteExpression;
    function isDoStatement(node) {
        return node.kind === ts.SyntaxKind.DoStatement;
    }
    exports.isDoStatement = isDoStatement;
    function isElementAccessExpression(node) {
        return node.kind === ts.SyntaxKind.ElementAccessExpression;
    }
    exports.isElementAccessExpression = isElementAccessExpression;
    function isEmptyStatement(node) {
        return node.kind === ts.SyntaxKind.EmptyStatement;
    }
    exports.isEmptyStatement = isEmptyStatement;
    function isEntityName(node) {
        return node.kind === ts.SyntaxKind.Identifier || isQualifiedName(node);
    }
    exports.isEntityName = isEntityName;
    function isEntityNameExpression(node) {
        return node.kind === ts.SyntaxKind.Identifier ||
            isPropertyAccessExpression(node) && isEntityNameExpression(node.expression);
    }
    exports.isEntityNameExpression = isEntityNameExpression;
    function isEnumDeclaration(node) {
        return node.kind === ts.SyntaxKind.EnumDeclaration;
    }
    exports.isEnumDeclaration = isEnumDeclaration;
    function isEnumMember(node) {
        return node.kind === ts.SyntaxKind.EnumMember;
    }
    exports.isEnumMember = isEnumMember;
    function isExportAssignment(node) {
        return node.kind === ts.SyntaxKind.ExportAssignment;
    }
    exports.isExportAssignment = isExportAssignment;
    function isExportDeclaration(node) {
        return node.kind === ts.SyntaxKind.ExportDeclaration;
    }
    exports.isExportDeclaration = isExportDeclaration;
    function isExportSpecifier(node) {
        return node.kind === ts.SyntaxKind.ExportSpecifier;
    }
    exports.isExportSpecifier = isExportSpecifier;
    function isExpression(node) {
        switch (node.kind) {
            case ts.SyntaxKind.ArrayLiteralExpression:
            case ts.SyntaxKind.ArrowFunction:
            case ts.SyntaxKind.AsExpression:
            case ts.SyntaxKind.AwaitExpression:
            case ts.SyntaxKind.BinaryExpression:
            case ts.SyntaxKind.CallExpression:
            case ts.SyntaxKind.ClassExpression:
            case ts.SyntaxKind.CommaListExpression:
            case ts.SyntaxKind.ConditionalExpression:
            case ts.SyntaxKind.DeleteExpression:
            case ts.SyntaxKind.ElementAccessExpression:
            case ts.SyntaxKind.FalseKeyword:
            case ts.SyntaxKind.FunctionExpression:
            case ts.SyntaxKind.Identifier:
            case ts.SyntaxKind.JsxElement:
            case ts.SyntaxKind.JsxFragment:
            case ts.SyntaxKind.JsxExpression:
            case ts.SyntaxKind.JsxOpeningElement:
            case ts.SyntaxKind.JsxOpeningFragment:
            case ts.SyntaxKind.JsxSelfClosingElement:
            case ts.SyntaxKind.MetaProperty:
            case ts.SyntaxKind.NewExpression:
            case ts.SyntaxKind.NonNullExpression:
            case ts.SyntaxKind.NoSubstitutionTemplateLiteral:
            case ts.SyntaxKind.NullKeyword:
            case ts.SyntaxKind.NumericLiteral:
            case ts.SyntaxKind.ObjectLiteralExpression:
            case ts.SyntaxKind.OmittedExpression:
            case ts.SyntaxKind.ParenthesizedExpression:
            case ts.SyntaxKind.PostfixUnaryExpression:
            case ts.SyntaxKind.PrefixUnaryExpression:
            case ts.SyntaxKind.PropertyAccessExpression:
            case ts.SyntaxKind.RegularExpressionLiteral:
            case ts.SyntaxKind.SpreadElement:
            case ts.SyntaxKind.StringLiteral:
            case ts.SyntaxKind.SuperKeyword:
            case ts.SyntaxKind.TaggedTemplateExpression:
            case ts.SyntaxKind.TemplateExpression:
            case ts.SyntaxKind.ThisKeyword:
            case ts.SyntaxKind.TrueKeyword:
            case ts.SyntaxKind.TypeAssertionExpression:
            case ts.SyntaxKind.TypeOfExpression:
            case ts.SyntaxKind.VoidExpression:
            case ts.SyntaxKind.YieldExpression:
                return true;
            default:
                return false;
        }
    }
    exports.isExpression = isExpression;
    function isExpressionStatement(node) {
        return node.kind === ts.SyntaxKind.ExpressionStatement;
    }
    exports.isExpressionStatement = isExpressionStatement;
    function isExpressionWithTypeArguments(node) {
        return node.kind === ts.SyntaxKind.ExpressionWithTypeArguments;
    }
    exports.isExpressionWithTypeArguments = isExpressionWithTypeArguments;
    function isExternalModuleReference(node) {
        return node.kind === ts.SyntaxKind.ExternalModuleReference;
    }
    exports.isExternalModuleReference = isExternalModuleReference;
    function isForInStatement(node) {
        return node.kind === ts.SyntaxKind.ForInStatement;
    }
    exports.isForInStatement = isForInStatement;
    function isForInOrOfStatement(node) {
        return node.kind === ts.SyntaxKind.ForOfStatement || node.kind === ts.SyntaxKind.ForInStatement;
    }
    exports.isForInOrOfStatement = isForInOrOfStatement;
    function isForOfStatement(node) {
        return node.kind === ts.SyntaxKind.ForOfStatement;
    }
    exports.isForOfStatement = isForOfStatement;
    function isForStatement(node) {
        return node.kind === ts.SyntaxKind.ForStatement;
    }
    exports.isForStatement = isForStatement;
    function isFunctionDeclaration(node) {
        return node.kind === ts.SyntaxKind.FunctionDeclaration;
    }
    exports.isFunctionDeclaration = isFunctionDeclaration;
    function isFunctionExpression(node) {
        return node.kind === ts.SyntaxKind.FunctionExpression;
    }
    exports.isFunctionExpression = isFunctionExpression;
    function isFunctionTypeNode(node) {
        return node.kind === ts.SyntaxKind.FunctionType;
    }
    exports.isFunctionTypeNode = isFunctionTypeNode;
    function isGetAccessorDeclaration(node) {
        return node.kind === ts.SyntaxKind.GetAccessor;
    }
    exports.isGetAccessorDeclaration = isGetAccessorDeclaration;
    function isIdentifier(node) {
        return node.kind === ts.SyntaxKind.Identifier;
    }
    exports.isIdentifier = isIdentifier;
    function isIfStatement(node) {
        return node.kind === ts.SyntaxKind.IfStatement;
    }
    exports.isIfStatement = isIfStatement;
    function isImportClause(node) {
        return node.kind === ts.SyntaxKind.ImportClause;
    }
    exports.isImportClause = isImportClause;
    function isImportDeclaration(node) {
        return node.kind === ts.SyntaxKind.ImportDeclaration;
    }
    exports.isImportDeclaration = isImportDeclaration;
    function isImportEqualsDeclaration(node) {
        return node.kind === ts.SyntaxKind.ImportEqualsDeclaration;
    }
    exports.isImportEqualsDeclaration = isImportEqualsDeclaration;
    function isImportSpecifier(node) {
        return node.kind === ts.SyntaxKind.ImportSpecifier;
    }
    exports.isImportSpecifier = isImportSpecifier;
    function isIndexedAccessTypeNode(node) {
        return node.kind === ts.SyntaxKind.IndexedAccessType;
    }
    exports.isIndexedAccessTypeNode = isIndexedAccessTypeNode;
    function isIndexSignatureDeclaration(node) {
        return node.kind === ts.SyntaxKind.IndexSignature;
    }
    exports.isIndexSignatureDeclaration = isIndexSignatureDeclaration;
    function isInferTypeNode(node) {
        return node.kind === ts.SyntaxKind.InferType;
    }
    exports.isInferTypeNode = isInferTypeNode;
    function isInterfaceDeclaration(node) {
        return node.kind === ts.SyntaxKind.InterfaceDeclaration;
    }
    exports.isInterfaceDeclaration = isInterfaceDeclaration;
    function isIntersectionTypeNode(node) {
        return node.kind === ts.SyntaxKind.IntersectionType;
    }
    exports.isIntersectionTypeNode = isIntersectionTypeNode;
    function isIterationStatement(node) {
        switch (node.kind) {
            case ts.SyntaxKind.ForStatement:
            case ts.SyntaxKind.ForOfStatement:
            case ts.SyntaxKind.ForInStatement:
            case ts.SyntaxKind.WhileStatement:
            case ts.SyntaxKind.DoStatement:
                return true;
            default:
                return false;
        }
    }
    exports.isIterationStatement = isIterationStatement;
    function isJsDoc(node) {
        return node.kind === ts.SyntaxKind.JSDocComment;
    }
    exports.isJsDoc = isJsDoc;
    function isJsxAttribute(node) {
        return node.kind === ts.SyntaxKind.JsxAttribute;
    }
    exports.isJsxAttribute = isJsxAttribute;
    function isJsxAttributeLike(node) {
        return node.kind === ts.SyntaxKind.JsxAttribute ||
            node.kind === ts.SyntaxKind.JsxSpreadAttribute;
    }
    exports.isJsxAttributeLike = isJsxAttributeLike;
    function isJsxAttributes(node) {
        return node.kind === ts.SyntaxKind.JsxAttributes;
    }
    exports.isJsxAttributes = isJsxAttributes;
    function isJsxClosingElement(node) {
        return node.kind === ts.SyntaxKind.JsxClosingElement;
    }
    exports.isJsxClosingElement = isJsxClosingElement;
    function isJsxClosingFragment(node) {
        return node.kind === ts.SyntaxKind.JsxClosingFragment;
    }
    exports.isJsxClosingFragment = isJsxClosingFragment;
    function isJsxElement(node) {
        return node.kind === ts.SyntaxKind.JsxElement;
    }
    exports.isJsxElement = isJsxElement;
    function isJsxExpression(node) {
        return node.kind === ts.SyntaxKind.JsxExpression;
    }
    exports.isJsxExpression = isJsxExpression;
    function isJsxFragment(node) {
        return node.kind === ts.SyntaxKind.JsxFragment;
    }
    exports.isJsxFragment = isJsxFragment;
    function isJsxOpeningElement(node) {
        return node.kind === ts.SyntaxKind.JsxOpeningElement;
    }
    exports.isJsxOpeningElement = isJsxOpeningElement;
    function isJsxOpeningFragment(node) {
        return node.kind === ts.SyntaxKind.JsxOpeningFragment;
    }
    exports.isJsxOpeningFragment = isJsxOpeningFragment;
    function isJsxOpeningLikeElement(node) {
        return node.kind === ts.SyntaxKind.JsxOpeningElement ||
            node.kind === ts.SyntaxKind.JsxSelfClosingElement;
    }
    exports.isJsxOpeningLikeElement = isJsxOpeningLikeElement;
    function isJsxSelfClosingElement(node) {
        return node.kind === ts.SyntaxKind.JsxSelfClosingElement;
    }
    exports.isJsxSelfClosingElement = isJsxSelfClosingElement;
    function isJsxSpreadAttribute(node) {
        return node.kind === ts.SyntaxKind.JsxSpreadAttribute;
    }
    exports.isJsxSpreadAttribute = isJsxSpreadAttribute;
    function isJsxText(node) {
        return node.kind === ts.SyntaxKind.JsxText;
    }
    exports.isJsxText = isJsxText;
    function isLabeledStatement(node) {
        return node.kind === ts.SyntaxKind.LabeledStatement;
    }
    exports.isLabeledStatement = isLabeledStatement;
    function isLiteralExpression(node) {
        return node.kind >= ts.SyntaxKind.FirstLiteralToken &&
            node.kind <= ts.SyntaxKind.LastLiteralToken;
    }
    exports.isLiteralExpression = isLiteralExpression;
    function isLiteralTypeNode(node) {
        return node.kind === ts.SyntaxKind.LiteralType;
    }
    exports.isLiteralTypeNode = isLiteralTypeNode;
    function isMappedTypeNode(node) {
        return node.kind === ts.SyntaxKind.MappedType;
    }
    exports.isMappedTypeNode = isMappedTypeNode;
    function isMetaProperty(node) {
        return node.kind === ts.SyntaxKind.MetaProperty;
    }
    exports.isMetaProperty = isMetaProperty;
    function isMethodDeclaration(node) {
        return node.kind === ts.SyntaxKind.MethodDeclaration;
    }
    exports.isMethodDeclaration = isMethodDeclaration;
    function isMethodSignature(node) {
        return node.kind === ts.SyntaxKind.MethodSignature;
    }
    exports.isMethodSignature = isMethodSignature;
    function isModuleBlock(node) {
        return node.kind === ts.SyntaxKind.ModuleBlock;
    }
    exports.isModuleBlock = isModuleBlock;
    function isModuleDeclaration(node) {
        return node.kind === ts.SyntaxKind.ModuleDeclaration;
    }
    exports.isModuleDeclaration = isModuleDeclaration;
    function isNamedExports(node) {
        return node.kind === ts.SyntaxKind.NamedExports;
    }
    exports.isNamedExports = isNamedExports;
    function isNamedImports(node) {
        return node.kind === ts.SyntaxKind.NamedImports;
    }
    exports.isNamedImports = isNamedImports;
    function isNamespaceDeclaration(node) {
        return isModuleDeclaration(node) &&
            node.name.kind === ts.SyntaxKind.Identifier &&
            node.body !== undefined &&
            (node.body.kind === ts.SyntaxKind.ModuleBlock ||
                isNamespaceDeclaration(node.body));
    }
    exports.isNamespaceDeclaration = isNamespaceDeclaration;
    function isNamespaceImport(node) {
        return node.kind === ts.SyntaxKind.NamespaceImport;
    }
    exports.isNamespaceImport = isNamespaceImport;
    function isNamespaceExportDeclaration(node) {
        return node.kind === ts.SyntaxKind.NamespaceExportDeclaration;
    }
    exports.isNamespaceExportDeclaration = isNamespaceExportDeclaration;
    function isNewExpression(node) {
        return node.kind === ts.SyntaxKind.NewExpression;
    }
    exports.isNewExpression = isNewExpression;
    function isNonNullExpression(node) {
        return node.kind === ts.SyntaxKind.NonNullExpression;
    }
    exports.isNonNullExpression = isNonNullExpression;
    function isNoSubstitutionTemplateLiteral(node) {
        return node.kind === ts.SyntaxKind.NoSubstitutionTemplateLiteral;
    }
    exports.isNoSubstitutionTemplateLiteral = isNoSubstitutionTemplateLiteral;
    function isNullLiteral(node) {
        return node.kind === ts.SyntaxKind.NullKeyword;
    }
    exports.isNullLiteral = isNullLiteral;
    function isNumericLiteral(node) {
        return node.kind === ts.SyntaxKind.NumericLiteral;
    }
    exports.isNumericLiteral = isNumericLiteral;
    function isNumericOrStringLikeLiteral(node) {
        switch (node.kind) {
            case ts.SyntaxKind.StringLiteral:
            case ts.SyntaxKind.NumericLiteral:
            case ts.SyntaxKind.NoSubstitutionTemplateLiteral:
                return true;
            default:
                return false;
        }
    }
    exports.isNumericOrStringLikeLiteral = isNumericOrStringLikeLiteral;
    function isObjectBindingPattern(node) {
        return node.kind === ts.SyntaxKind.ObjectBindingPattern;
    }
    exports.isObjectBindingPattern = isObjectBindingPattern;
    function isObjectLiteralExpression(node) {
        return node.kind === ts.SyntaxKind.ObjectLiteralExpression;
    }
    exports.isObjectLiteralExpression = isObjectLiteralExpression;
    function isOmittedExpression(node) {
        return node.kind === ts.SyntaxKind.OmittedExpression;
    }
    exports.isOmittedExpression = isOmittedExpression;
    function isParameterDeclaration(node) {
        return node.kind === ts.SyntaxKind.Parameter;
    }
    exports.isParameterDeclaration = isParameterDeclaration;
    function isParenthesizedExpression(node) {
        return node.kind === ts.SyntaxKind.ParenthesizedExpression;
    }
    exports.isParenthesizedExpression = isParenthesizedExpression;
    function isParenthesizedTypeNode(node) {
        return node.kind === ts.SyntaxKind.ParenthesizedType;
    }
    exports.isParenthesizedTypeNode = isParenthesizedTypeNode;
    function isPostfixUnaryExpression(node) {
        return node.kind === ts.SyntaxKind.PostfixUnaryExpression;
    }
    exports.isPostfixUnaryExpression = isPostfixUnaryExpression;
    function isPrefixUnaryExpression(node) {
        return node.kind === ts.SyntaxKind.PrefixUnaryExpression;
    }
    exports.isPrefixUnaryExpression = isPrefixUnaryExpression;
    function isPropertyAccessExpression(node) {
        return node.kind === ts.SyntaxKind.PropertyAccessExpression;
    }
    exports.isPropertyAccessExpression = isPropertyAccessExpression;
    function isPropertyAssignment(node) {
        return node.kind === ts.SyntaxKind.PropertyAssignment;
    }
    exports.isPropertyAssignment = isPropertyAssignment;
    function isPropertyDeclaration(node) {
        return node.kind === ts.SyntaxKind.PropertyDeclaration;
    }
    exports.isPropertyDeclaration = isPropertyDeclaration;
    function isPropertySignature(node) {
        return node.kind === ts.SyntaxKind.PropertySignature;
    }
    exports.isPropertySignature = isPropertySignature;
    function isQualifiedName(node) {
        return node.kind === ts.SyntaxKind.QualifiedName;
    }
    exports.isQualifiedName = isQualifiedName;
    function isRegularExpressionLiteral(node) {
        return node.kind === ts.SyntaxKind.RegularExpressionLiteral;
    }
    exports.isRegularExpressionLiteral = isRegularExpressionLiteral;
    function isReturnStatement(node) {
        return node.kind === ts.SyntaxKind.ReturnStatement;
    }
    exports.isReturnStatement = isReturnStatement;
    function isSetAccessorDeclaration(node) {
        return node.kind === ts.SyntaxKind.SetAccessor;
    }
    exports.isSetAccessorDeclaration = isSetAccessorDeclaration;
    function isShorthandPropertyAssignment(node) {
        return node.kind === ts.SyntaxKind.ShorthandPropertyAssignment;
    }
    exports.isShorthandPropertyAssignment = isShorthandPropertyAssignment;
    function isSignatureDeclaration(node) {
        return node.parameters !== undefined;
    }
    exports.isSignatureDeclaration = isSignatureDeclaration;
    function isSourceFile(node) {
        return node.kind === ts.SyntaxKind.SourceFile;
    }
    exports.isSourceFile = isSourceFile;
    function isSpreadAssignment(node) {
        return node.kind === ts.SyntaxKind.SpreadAssignment;
    }
    exports.isSpreadAssignment = isSpreadAssignment;
    function isSpreadElement(node) {
        return node.kind === ts.SyntaxKind.SpreadElement;
    }
    exports.isSpreadElement = isSpreadElement;
    function isStringLiteral(node) {
        return node.kind === ts.SyntaxKind.StringLiteral;
    }
    exports.isStringLiteral = isStringLiteral;
    function isSwitchStatement(node) {
        return node.kind === ts.SyntaxKind.SwitchStatement;
    }
    exports.isSwitchStatement = isSwitchStatement;
    function isSyntaxList(node) {
        return node.kind === ts.SyntaxKind.SyntaxList;
    }
    exports.isSyntaxList = isSyntaxList;
    function isTaggedTemplateExpression(node) {
        return node.kind === ts.SyntaxKind.TaggedTemplateExpression;
    }
    exports.isTaggedTemplateExpression = isTaggedTemplateExpression;
    function isTemplateExpression(node) {
        return node.kind === ts.SyntaxKind.TemplateExpression;
    }
    exports.isTemplateExpression = isTemplateExpression;
    function isTemplateLiteral(node) {
        return node.kind === ts.SyntaxKind.TemplateExpression ||
            node.kind === ts.SyntaxKind.NoSubstitutionTemplateLiteral;
    }
    exports.isTemplateLiteral = isTemplateLiteral;
    function isTextualLiteral(node) {
        return node.kind === ts.SyntaxKind.StringLiteral ||
            node.kind === ts.SyntaxKind.NoSubstitutionTemplateLiteral;
    }
    exports.isTextualLiteral = isTextualLiteral;
    function isThrowStatement(node) {
        return node.kind === ts.SyntaxKind.ThrowStatement;
    }
    exports.isThrowStatement = isThrowStatement;
    function isTryStatement(node) {
        return node.kind === ts.SyntaxKind.TryStatement;
    }
    exports.isTryStatement = isTryStatement;
    function isTupleTypeNode(node) {
        return node.kind === ts.SyntaxKind.TupleType;
    }
    exports.isTupleTypeNode = isTupleTypeNode;
    function isTypeAliasDeclaration(node) {
        return node.kind === ts.SyntaxKind.TypeAliasDeclaration;
    }
    exports.isTypeAliasDeclaration = isTypeAliasDeclaration;
    function isTypeAssertion(node) {
        return node.kind === ts.SyntaxKind.TypeAssertionExpression;
    }
    exports.isTypeAssertion = isTypeAssertion;
    function isTypeLiteralNode(node) {
        return node.kind === ts.SyntaxKind.TypeLiteral;
    }
    exports.isTypeLiteralNode = isTypeLiteralNode;
    function isTypeOfExpression(node) {
        return node.kind === ts.SyntaxKind.TypeOfExpression;
    }
    exports.isTypeOfExpression = isTypeOfExpression;
    function isTypeOperatorNode(node) {
        return node.kind === ts.SyntaxKind.TypeOperator;
    }
    exports.isTypeOperatorNode = isTypeOperatorNode;
    function isTypeParameterDeclaration(node) {
        return node.kind === ts.SyntaxKind.TypeParameter;
    }
    exports.isTypeParameterDeclaration = isTypeParameterDeclaration;
    function isTypePredicateNode(node) {
        return node.kind === ts.SyntaxKind.TypePredicate;
    }
    exports.isTypePredicateNode = isTypePredicateNode;
    function isTypeReferenceNode(node) {
        return node.kind === ts.SyntaxKind.TypeReference;
    }
    exports.isTypeReferenceNode = isTypeReferenceNode;
    function isTypeQueryNode(node) {
        return node.kind === ts.SyntaxKind.TypeQuery;
    }
    exports.isTypeQueryNode = isTypeQueryNode;
    function isUnionTypeNode(node) {
        return node.kind === ts.SyntaxKind.UnionType;
    }
    exports.isUnionTypeNode = isUnionTypeNode;
    function isVariableDeclaration(node) {
        return node.kind === ts.SyntaxKind.VariableDeclaration;
    }
    exports.isVariableDeclaration = isVariableDeclaration;
    function isVariableStatement(node) {
        return node.kind === ts.SyntaxKind.VariableStatement;
    }
    exports.isVariableStatement = isVariableStatement;
    function isVariableDeclarationList(node) {
        return node.kind === ts.SyntaxKind.VariableDeclarationList;
    }
    exports.isVariableDeclarationList = isVariableDeclarationList;
    function isVoidExpression(node) {
        return node.kind === ts.SyntaxKind.VoidExpression;
    }
    exports.isVoidExpression = isVoidExpression;
    function isWhileStatement(node) {
        return node.kind === ts.SyntaxKind.WhileStatement;
    }
    exports.isWhileStatement = isWhileStatement;
    function isWithStatement(node) {
        return node.kind === ts.SyntaxKind.WithStatement;
    }
    exports.isWithStatement = isWithStatement;
},
/* tsutils/typeguard/type.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(924) /* ./3.2/type */, exports);
},
/* tsutils/typeguard/3.2/type.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(925) /* ../3.0/type */, exports);
},
/* tsutils/typeguard/3.0/type.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(926) /* ../2.9/type */, exports);
    const ts = require(21) /* typescript */;
    const type_1 = require(926) /* ../2.9/type */;
    function isTupleType(type) {
        return (type.flags & ts.TypeFlags.Object && type.objectFlags & ts.ObjectFlags.Tuple) !== 0;
    }
    exports.isTupleType = isTupleType;
    function isTupleTypeReference(type) {
        return type_1.isTypeReference(type) && isTupleType(type.target);
    }
    exports.isTupleTypeReference = isTupleTypeReference;
},
/* tsutils/typeguard/2.9/type.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(927) /* ../2.8/type */, exports);
},
/* tsutils/typeguard/2.8/type.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ts = require(21) /* typescript */;
    function isConditionalType(type) {
        return (type.flags & ts.TypeFlags.Conditional) !== 0;
    }
    exports.isConditionalType = isConditionalType;
    function isEnumType(type) {
        return (type.flags & ts.TypeFlags.Enum) !== 0;
    }
    exports.isEnumType = isEnumType;
    function isGenericType(type) {
        return (type.flags & ts.TypeFlags.Object) !== 0 &&
            (type.objectFlags & ts.ObjectFlags.ClassOrInterface) !== 0 &&
            (type.objectFlags & ts.ObjectFlags.Reference) !== 0;
    }
    exports.isGenericType = isGenericType;
    function isIndexedAccessType(type) {
        return (type.flags & ts.TypeFlags.IndexedAccess) !== 0;
    }
    exports.isIndexedAccessType = isIndexedAccessType;
    function isIndexedAccessype(type) {
        return (type.flags & ts.TypeFlags.Index) !== 0;
    }
    exports.isIndexedAccessype = isIndexedAccessype;
    function isInstantiableType(type) {
        return (type.flags & ts.TypeFlags.Instantiable) !== 0;
    }
    exports.isInstantiableType = isInstantiableType;
    function isInterfaceType(type) {
        return (type.flags & ts.TypeFlags.Object) !== 0 &&
            (type.objectFlags & ts.ObjectFlags.ClassOrInterface) !== 0;
    }
    exports.isInterfaceType = isInterfaceType;
    function isIntersectionType(type) {
        return (type.flags & ts.TypeFlags.Intersection) !== 0;
    }
    exports.isIntersectionType = isIntersectionType;
    function isLiteralType(type) {
        return (type.flags & (ts.TypeFlags.StringOrNumberLiteral | ts.TypeFlags.BigIntLiteral)) !== 0;
    }
    exports.isLiteralType = isLiteralType;
    function isObjectType(type) {
        return (type.flags & ts.TypeFlags.Object) !== 0;
    }
    exports.isObjectType = isObjectType;
    function isSubstitutionType(type) {
        return (type.flags & ts.TypeFlags.Substitution) !== 0;
    }
    exports.isSubstitutionType = isSubstitutionType;
    function isTypeParameter(type) {
        return (type.flags & ts.TypeFlags.TypeParameter) !== 0;
    }
    exports.isTypeParameter = isTypeParameter;
    function isTypeReference(type) {
        return (type.flags & ts.TypeFlags.Object) !== 0 &&
            (type.objectFlags & ts.ObjectFlags.Reference) !== 0;
    }
    exports.isTypeReference = isTypeReference;
    function isTypeVariable(type) {
        return (type.flags & ts.TypeFlags.TypeVariable) !== 0;
    }
    exports.isTypeVariable = isTypeVariable;
    function isUnionOrIntersectionType(type) {
        return (type.flags & ts.TypeFlags.UnionOrIntersection) !== 0;
    }
    exports.isUnionOrIntersectionType = isUnionOrIntersectionType;
    function isUnionType(type) {
        return (type.flags & ts.TypeFlags.Union) !== 0;
    }
    exports.isUnionType = isUnionType;
    function isUniqueESSymbolType(type) {
        return (type.flags & ts.TypeFlags.UniqueESSymbol) !== 0;
    }
    exports.isUniqueESSymbolType = isUniqueESSymbolType;
},
/* tsutils/util/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(929) /* ./util */, exports);
    tslib_1.__exportStar(require(932) /* ./usage */, exports);
    tslib_1.__exportStar(require(933) /* ./control-flow */, exports);
    tslib_1.__exportStar(require(931) /* ./type */, exports);
    tslib_1.__exportStar(require(934) /* ./convert-ast */, exports);
},
/* tsutils/util/util.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.getBaseOfClassLikeExpression = exports.hasExhaustiveCaseClauses = exports.formatPseudoBigInt = exports.unwrapParentheses = exports.getSingleLateBoundPropertyNameOfPropertyName = exports.getLateBoundPropertyNamesOfPropertyName = exports.getLateBoundPropertyNames = exports.getPropertyNameOfWellKnownSymbol = exports.isWellKnownSymbolLiterally = exports.isBindableObjectDefinePropertyCall = exports.isReadonlyAssignmentDeclaration = exports.isInConstContext = exports.isConstAssertion = exports.getTsCheckDirective = exports.getCheckJsDirective = exports.isAmbientModule = exports.isCompilerOptionEnabled = exports.isStrictCompilerOptionEnabled = exports.getIIFE = exports.isAmbientModuleBlock = exports.isStatementInAmbientContext = exports.findImportLikeNodes = exports.findImports = exports.ImportKind = exports.parseJsDocOfNode = exports.getJsDoc = exports.canHaveJsDoc = exports.isReassignmentTarget = exports.getAccessKind = exports.AccessKind = exports.isExpressionValueUsed = exports.getDeclarationOfBindingElement = exports.hasSideEffects = exports.SideEffectOptions = exports.isSameLine = exports.isNumericPropertyName = exports.isValidJsxIdentifier = exports.isValidNumericLiteral = exports.isValidPropertyName = exports.isValidPropertyAccess = void 0;
    const ts = require(21) /* typescript */;
    const node_1 = require(918) /* ../typeguard/node */;
    const _3_2_1 = require(930) /* ../typeguard/3.2 */;
    const type_1 = require(931) /* ./type */;
    function getChildOfKind(node, kind, sourceFile) {
        for (const child of node.getChildren(sourceFile))
            if (child.kind === kind)
                return child;
    }
    exports.getChildOfKind = getChildOfKind;
    function isTokenKind(kind) {
        return kind >= ts.SyntaxKind.FirstToken && kind <= ts.SyntaxKind.LastToken;
    }
    exports.isTokenKind = isTokenKind;
    function isNodeKind(kind) {
        return kind >= ts.SyntaxKind.FirstNode;
    }
    exports.isNodeKind = isNodeKind;
    function isAssignmentKind(kind) {
        return kind >= ts.SyntaxKind.FirstAssignment && kind <= ts.SyntaxKind.LastAssignment;
    }
    exports.isAssignmentKind = isAssignmentKind;
    function isTypeNodeKind(kind) {
        return kind >= ts.SyntaxKind.FirstTypeNode && kind <= ts.SyntaxKind.LastTypeNode;
    }
    exports.isTypeNodeKind = isTypeNodeKind;
    function isJsDocKind(kind) {
        return kind >= ts.SyntaxKind.FirstJSDocNode && kind <= ts.SyntaxKind.LastJSDocNode;
    }
    exports.isJsDocKind = isJsDocKind;
    function isKeywordKind(kind) {
        return kind >= ts.SyntaxKind.FirstKeyword && kind <= ts.SyntaxKind.LastKeyword;
    }
    exports.isKeywordKind = isKeywordKind;
    function isThisParameter(parameter) {
        return parameter.name.kind === ts.SyntaxKind.Identifier && parameter.name.originalKeywordKind === ts.SyntaxKind.ThisKeyword;
    }
    exports.isThisParameter = isThisParameter;
    function getModifier(node, kind) {
        if (node.modifiers !== undefined)
            for (const modifier of node.modifiers)
                if (modifier.kind === kind)
                    return modifier;
    }
    exports.getModifier = getModifier;
    function hasModifier(modifiers, ...kinds) {
        if (modifiers === undefined)
            return false;
        for (const modifier of modifiers)
            if (kinds.includes(modifier.kind))
                return true;
        return false;
    }
    exports.hasModifier = hasModifier;
    function isParameterProperty(node) {
        return hasModifier(node.modifiers, ts.SyntaxKind.PublicKeyword, ts.SyntaxKind.ProtectedKeyword, ts.SyntaxKind.PrivateKeyword, ts.SyntaxKind.ReadonlyKeyword);
    }
    exports.isParameterProperty = isParameterProperty;
    function hasAccessModifier(node) {
        return isModifierFlagSet(node, ts.ModifierFlags.AccessibilityModifier);
    }
    exports.hasAccessModifier = hasAccessModifier;
    function isFlagSet(obj, flag) {
        return (obj.flags & flag) !== 0;
    }
    exports.isNodeFlagSet = isFlagSet;
    exports.isTypeFlagSet = isFlagSet;
    exports.isSymbolFlagSet = isFlagSet;
    function isObjectFlagSet(objectType, flag) {
        return (objectType.objectFlags & flag) !== 0;
    }
    exports.isObjectFlagSet = isObjectFlagSet;
    function isModifierFlagSet(node, flag) {
        return (ts.getCombinedModifierFlags(node) & flag) !== 0;
    }
    exports.isModifierFlagSet = isModifierFlagSet;
    function getPreviousStatement(statement) {
        const parent = statement.parent;
        if (node_1.isBlockLike(parent)) {
            const index = parent.statements.indexOf(statement);
            if (index > 0)
                return parent.statements[index - 1];
        }
    }
    exports.getPreviousStatement = getPreviousStatement;
    function getNextStatement(statement) {
        const parent = statement.parent;
        if (node_1.isBlockLike(parent)) {
            const index = parent.statements.indexOf(statement);
            if (index < parent.statements.length)
                return parent.statements[index + 1];
        }
    }
    exports.getNextStatement = getNextStatement;
    /** Returns the token before the start of `node` or `undefined` if there is none. */
    function getPreviousToken(node, sourceFile) {
        const { pos } = node;
        if (pos === 0)
            return;
        do
            node = node.parent;
        while (node.pos === pos);
        return getTokenAtPositionWorker(node, pos - 1, sourceFile !== null && sourceFile !== void 0 ? sourceFile : node.getSourceFile(), false);
    }
    exports.getPreviousToken = getPreviousToken;
    /** Returns the next token that begins after the end of `node`. Returns `undefined` for SourceFile and EndOfFileToken */
    function getNextToken(node, sourceFile) {
        if (node.kind === ts.SyntaxKind.SourceFile || node.kind === ts.SyntaxKind.EndOfFileToken)
            return;
        const end = node.end;
        node = node.parent;
        while (node.end === end) {
            if (node.parent === undefined)
                return node.endOfFileToken;
            node = node.parent;
        }
        return getTokenAtPositionWorker(node, end, sourceFile !== null && sourceFile !== void 0 ? sourceFile : node.getSourceFile(), false);
    }
    exports.getNextToken = getNextToken;
    /** Returns the token at or following the specified position or undefined if none is found inside `parent`. */
    function getTokenAtPosition(parent, pos, sourceFile, allowJsDoc) {
        if (pos < parent.pos || pos >= parent.end)
            return;
        if (isTokenKind(parent.kind))
            return parent;
        return getTokenAtPositionWorker(parent, pos, sourceFile !== null && sourceFile !== void 0 ? sourceFile : parent.getSourceFile(), allowJsDoc === true);
    }
    exports.getTokenAtPosition = getTokenAtPosition;
    function getTokenAtPositionWorker(node, pos, sourceFile, allowJsDoc) {
        if (!allowJsDoc) {
            // if we are not interested in JSDoc, we can skip to the deepest AST node at the given position
            node = getAstNodeAtPosition(node, pos);
            if (isTokenKind(node.kind))
                return node;
        }
        outer: while (true) {
            for (const child of node.getChildren(sourceFile)) {
                if (child.end > pos && (allowJsDoc || child.kind !== ts.SyntaxKind.JSDocComment)) {
                    if (isTokenKind(child.kind))
                        return child;
                    // next token is nested in another node
                    node = child;
                    continue outer;
                }
            }
            return;
        }
    }
    /**
     * Return the comment at the specified position.
     * You can pass an optional `parent` to avoid some work finding the corresponding token starting at `sourceFile`.
     * If the `parent` parameter is passed, `pos` must be between `parent.pos` and `parent.end`.
    */
    function getCommentAtPosition(sourceFile, pos, parent = sourceFile) {
        const token = getTokenAtPosition(parent, pos, sourceFile);
        if (token === undefined || token.kind === ts.SyntaxKind.JsxText || pos >= token.end - (ts.tokenToString(token.kind) || '').length)
            return;
        const startPos = token.pos === 0
            ? (ts.getShebang(sourceFile.text) || '').length
            : token.pos;
        return startPos !== 0 && ts.forEachTrailingCommentRange(sourceFile.text, startPos, commentAtPositionCallback, pos) ||
            ts.forEachLeadingCommentRange(sourceFile.text, startPos, commentAtPositionCallback, pos);
    }
    exports.getCommentAtPosition = getCommentAtPosition;
    function commentAtPositionCallback(pos, end, kind, _nl, at) {
        return at >= pos && at < end ? { pos, end, kind } : undefined;
    }
    /**
     * Returns whether the specified position is inside a comment.
     * You can pass an optional `parent` to avoid some work finding the corresponding token starting at `sourceFile`.
     * If the `parent` parameter is passed, `pos` must be between `parent.pos` and `parent.end`.
     */
    function isPositionInComment(sourceFile, pos, parent) {
        return getCommentAtPosition(sourceFile, pos, parent) !== undefined;
    }
    exports.isPositionInComment = isPositionInComment;
    function commentText(sourceText, comment) {
        return sourceText.substring(comment.pos + 2, comment.kind === ts.SyntaxKind.SingleLineCommentTrivia ? comment.end : comment.end - 2);
    }
    exports.commentText = commentText;
    /** Returns the deepest AST Node at `pos`. Returns undefined if `pos` is outside of the range of `node` */
    function getAstNodeAtPosition(node, pos) {
        if (node.pos > pos || node.end <= pos)
            return;
        while (isNodeKind(node.kind)) {
            const nested = ts.forEachChild(node, (child) => child.pos <= pos && child.end > pos ? child : undefined);
            if (nested === undefined)
                break;
            node = nested;
        }
        return node;
    }
    exports.getAstNodeAtPosition = getAstNodeAtPosition;
    /**
     * Returns the NodeWrap of deepest AST node that contains `pos` between its `pos` and `end`.
     * Only returns undefined if pos is outside of `wrap`
     */
    function getWrappedNodeAtPosition(wrap, pos) {
        if (wrap.node.pos > pos || wrap.node.end <= pos)
            return;
        outer: while (true) {
            for (const child of wrap.children) {
                if (child.node.pos > pos)
                    return wrap;
                if (child.node.end > pos) {
                    wrap = child;
                    continue outer;
                }
            }
            return wrap;
        }
    }
    exports.getWrappedNodeAtPosition = getWrappedNodeAtPosition;
    function getPropertyName(propertyName) {
        if (propertyName.kind === ts.SyntaxKind.ComputedPropertyName) {
            const expression = unwrapParentheses(propertyName.expression);
            if (node_1.isPrefixUnaryExpression(expression)) {
                let negate = false;
                switch (expression.operator) {
                    case ts.SyntaxKind.MinusToken:
                        negate = true;
                    // falls through
                    case ts.SyntaxKind.PlusToken:
                        return node_1.isNumericLiteral(expression.operand)
                            ? `${negate ? '-' : ''}${expression.operand.text}`
                            : _3_2_1.isBigIntLiteral(expression.operand)
                                ? `${negate ? '-' : ''}${expression.operand.text.slice(0, -1)}`
                                : undefined;
                    default:
                        return;
                }
            }
            if (_3_2_1.isBigIntLiteral(expression))
                // handle BigInt, even though TypeScript doesn't allow BigInt as computed property name
                return expression.text.slice(0, -1);
            if (node_1.isNumericOrStringLikeLiteral(expression))
                return expression.text;
            return;
        }
        return propertyName.kind === ts.SyntaxKind.PrivateIdentifier ? undefined : propertyName.text;
    }
    exports.getPropertyName = getPropertyName;
    function forEachDestructuringIdentifier(pattern, fn) {
        for (const element of pattern.elements) {
            if (element.kind !== ts.SyntaxKind.BindingElement)
                continue;
            let result;
            if (element.name.kind === ts.SyntaxKind.Identifier) {
                result = fn(element);
            }
            else {
                result = forEachDestructuringIdentifier(element.name, fn);
            }
            if (result)
                return result;
        }
    }
    exports.forEachDestructuringIdentifier = forEachDestructuringIdentifier;
    function forEachDeclaredVariable(declarationList, cb) {
        for (const declaration of declarationList.declarations) {
            let result;
            if (declaration.name.kind === ts.SyntaxKind.Identifier) {
                result = cb(declaration);
            }
            else {
                result = forEachDestructuringIdentifier(declaration.name, cb);
            }
            if (result)
                return result;
        }
    }
    exports.forEachDeclaredVariable = forEachDeclaredVariable;
    var VariableDeclarationKind;
    (function (VariableDeclarationKind) {
        VariableDeclarationKind[VariableDeclarationKind["Var"] = 0] = "Var";
        VariableDeclarationKind[VariableDeclarationKind["Let"] = 1] = "Let";
        VariableDeclarationKind[VariableDeclarationKind["Const"] = 2] = "Const";
    })(VariableDeclarationKind = exports.VariableDeclarationKind || (exports.VariableDeclarationKind = {}));
    function getVariableDeclarationKind(declarationList) {
        if (declarationList.flags & ts.NodeFlags.Let)
            return 1 /* Let */;
        if (declarationList.flags & ts.NodeFlags.Const)
            return 2 /* Const */;
        return 0 /* Var */;
    }
    exports.getVariableDeclarationKind = getVariableDeclarationKind;
    function isBlockScopedVariableDeclarationList(declarationList) {
        return (declarationList.flags & ts.NodeFlags.BlockScoped) !== 0;
    }
    exports.isBlockScopedVariableDeclarationList = isBlockScopedVariableDeclarationList;
    function isBlockScopedVariableDeclaration(declaration) {
        const parent = declaration.parent;
        return parent.kind === ts.SyntaxKind.CatchClause ||
            isBlockScopedVariableDeclarationList(parent);
    }
    exports.isBlockScopedVariableDeclaration = isBlockScopedVariableDeclaration;
    function isBlockScopedDeclarationStatement(statement) {
        switch (statement.kind) {
            case ts.SyntaxKind.VariableStatement:
                return isBlockScopedVariableDeclarationList(statement.declarationList);
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.EnumDeclaration:
            case ts.SyntaxKind.InterfaceDeclaration:
            case ts.SyntaxKind.TypeAliasDeclaration:
                return true;
            default:
                return false;
        }
    }
    exports.isBlockScopedDeclarationStatement = isBlockScopedDeclarationStatement;
    function isInSingleStatementContext(statement) {
        switch (statement.parent.kind) {
            case ts.SyntaxKind.ForStatement:
            case ts.SyntaxKind.ForInStatement:
            case ts.SyntaxKind.ForOfStatement:
            case ts.SyntaxKind.WhileStatement:
            case ts.SyntaxKind.DoStatement:
            case ts.SyntaxKind.IfStatement:
            case ts.SyntaxKind.WithStatement:
            case ts.SyntaxKind.LabeledStatement:
                return true;
            default:
                return false;
        }
    }
    exports.isInSingleStatementContext = isInSingleStatementContext;
    var ScopeBoundary;
    (function (ScopeBoundary) {
        ScopeBoundary[ScopeBoundary["None"] = 0] = "None";
        ScopeBoundary[ScopeBoundary["Function"] = 1] = "Function";
        ScopeBoundary[ScopeBoundary["Block"] = 2] = "Block";
        ScopeBoundary[ScopeBoundary["Type"] = 4] = "Type";
        ScopeBoundary[ScopeBoundary["ConditionalType"] = 8] = "ConditionalType";
    })(ScopeBoundary = exports.ScopeBoundary || (exports.ScopeBoundary = {}));
    var ScopeBoundarySelector;
    (function (ScopeBoundarySelector) {
        ScopeBoundarySelector[ScopeBoundarySelector["Function"] = 1] = "Function";
        ScopeBoundarySelector[ScopeBoundarySelector["Block"] = 3] = "Block";
        ScopeBoundarySelector[ScopeBoundarySelector["Type"] = 7] = "Type";
        ScopeBoundarySelector[ScopeBoundarySelector["InferType"] = 8] = "InferType";
    })(ScopeBoundarySelector = exports.ScopeBoundarySelector || (exports.ScopeBoundarySelector = {}));
    function isScopeBoundary(node) {
        return isFunctionScopeBoundary(node) || isBlockScopeBoundary(node) || isTypeScopeBoundary(node);
    }
    exports.isScopeBoundary = isScopeBoundary;
    function isTypeScopeBoundary(node) {
        switch (node.kind) {
            case ts.SyntaxKind.InterfaceDeclaration:
            case ts.SyntaxKind.TypeAliasDeclaration:
            case ts.SyntaxKind.MappedType:
                return 4 /* Type */;
            case ts.SyntaxKind.ConditionalType:
                return 8 /* ConditionalType */;
            default:
                return 0 /* None */;
        }
    }
    exports.isTypeScopeBoundary = isTypeScopeBoundary;
    function isFunctionScopeBoundary(node) {
        switch (node.kind) {
            case ts.SyntaxKind.FunctionExpression:
            case ts.SyntaxKind.ArrowFunction:
            case ts.SyntaxKind.Constructor:
            case ts.SyntaxKind.ModuleDeclaration:
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.ClassExpression:
            case ts.SyntaxKind.EnumDeclaration:
            case ts.SyntaxKind.MethodDeclaration:
            case ts.SyntaxKind.FunctionDeclaration:
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.SetAccessor:
            case ts.SyntaxKind.MethodSignature:
            case ts.SyntaxKind.CallSignature:
            case ts.SyntaxKind.ConstructSignature:
            case ts.SyntaxKind.ConstructorType:
            case ts.SyntaxKind.FunctionType:
                return 1 /* Function */;
            case ts.SyntaxKind.SourceFile:
                // if SourceFile is no module, it contributes to the global scope and is therefore no scope boundary
                return ts.isExternalModule(node) ? 1 /* Function */ : 0 /* None */;
            default:
                return 0 /* None */;
        }
    }
    exports.isFunctionScopeBoundary = isFunctionScopeBoundary;
    function isBlockScopeBoundary(node) {
        switch (node.kind) {
            case ts.SyntaxKind.Block:
                const parent = node.parent;
                return parent.kind !== ts.SyntaxKind.CatchClause &&
                    // blocks inside SourceFile are block scope boundaries
                    (parent.kind === ts.SyntaxKind.SourceFile ||
                        // blocks that are direct children of a function scope boundary are no scope boundary
                        // for example the FunctionBlock is part of the function scope of the containing function
                        !isFunctionScopeBoundary(parent))
                    ? 2 /* Block */
                    : 0 /* None */;
            case ts.SyntaxKind.ForStatement:
            case ts.SyntaxKind.ForInStatement:
            case ts.SyntaxKind.ForOfStatement:
            case ts.SyntaxKind.CaseBlock:
            case ts.SyntaxKind.CatchClause:
            case ts.SyntaxKind.WithStatement:
                return 2 /* Block */;
            default:
                return 0 /* None */;
        }
    }
    exports.isBlockScopeBoundary = isBlockScopeBoundary;
    /** Returns true for scope boundaries that have their own `this` reference instead of inheriting it from the containing scope */
    function hasOwnThisReference(node) {
        switch (node.kind) {
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.ClassExpression:
            case ts.SyntaxKind.FunctionExpression:
                return true;
            case ts.SyntaxKind.FunctionDeclaration:
                return node.body !== undefined;
            case ts.SyntaxKind.MethodDeclaration:
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.SetAccessor:
                return node.parent.kind === ts.SyntaxKind.ObjectLiteralExpression;
            default:
                return false;
        }
    }
    exports.hasOwnThisReference = hasOwnThisReference;
    function isFunctionWithBody(node) {
        switch (node.kind) {
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.SetAccessor:
            case ts.SyntaxKind.FunctionDeclaration:
            case ts.SyntaxKind.MethodDeclaration:
            case ts.SyntaxKind.Constructor:
                return node.body !== undefined;
            case ts.SyntaxKind.FunctionExpression:
            case ts.SyntaxKind.ArrowFunction:
                return true;
            default:
                return false;
        }
    }
    exports.isFunctionWithBody = isFunctionWithBody;
    /**
     * Iterate over all tokens of `node`
     *
     * @param node The node whose tokens should be visited
     * @param cb Is called for every token contained in `node`
     */
    function forEachToken(node, cb, sourceFile = node.getSourceFile()) {
        const queue = [];
        while (true) {
            if (isTokenKind(node.kind)) {
                cb(node);
            }
            else if (node.kind !== ts.SyntaxKind.JSDocComment) {
                const children = node.getChildren(sourceFile);
                if (children.length === 1) {
                    node = children[0];
                    continue;
                }
                for (let i = children.length - 1; i >= 0; --i)
                    queue.push(children[i]); // add children in reverse order, when we pop the next element from the queue, it's the first child
            }
            if (queue.length === 0)
                break;
            node = queue.pop();
        }
    }
    exports.forEachToken = forEachToken;
    /**
     * Iterate over all tokens and trivia of `node`
     *
     * @description JsDoc comments are treated like regular comments
     *
     * @param node The node whose tokens should be visited
     * @param cb Is called for every token contained in `node` and trivia before the token
     */
    function forEachTokenWithTrivia(node, cb, sourceFile = node.getSourceFile()) {
        const fullText = sourceFile.text;
        const scanner = ts.createScanner(sourceFile.languageVersion, false, sourceFile.languageVariant, fullText);
        return forEachToken(node, (token) => {
            const tokenStart = token.kind === ts.SyntaxKind.JsxText || token.pos === token.end ? token.pos : token.getStart(sourceFile);
            if (tokenStart !== token.pos) {
                // we only have to handle trivia before each token. whitespace at the end of the file is followed by EndOfFileToken
                scanner.setTextPos(token.pos);
                let kind = scanner.scan();
                let pos = scanner.getTokenPos();
                while (pos < tokenStart) {
                    const textPos = scanner.getTextPos();
                    cb(fullText, kind, { pos, end: textPos }, token.parent);
                    if (textPos === tokenStart)
                        break;
                    kind = scanner.scan();
                    pos = scanner.getTokenPos();
                }
            }
            return cb(fullText, token.kind, { end: token.end, pos: tokenStart }, token.parent);
        }, sourceFile);
    }
    exports.forEachTokenWithTrivia = forEachTokenWithTrivia;
    /** Iterate over all comments owned by `node` or its children */
    function forEachComment(node, cb, sourceFile = node.getSourceFile()) {
        /* Visit all tokens and skip trivia.
           Comment ranges between tokens are parsed without the need of a scanner.
           forEachTokenWithWhitespace does intentionally not pay attention to the correct comment ownership of nodes as it always
           scans all trivia before each token, which could include trailing comments of the previous token.
           Comment onwership is done right in this function*/
        const fullText = sourceFile.text;
        const notJsx = sourceFile.languageVariant !== ts.LanguageVariant.JSX;
        return forEachToken(node, (token) => {
            if (token.pos === token.end)
                return;
            if (token.kind !== ts.SyntaxKind.JsxText)
                ts.forEachLeadingCommentRange(fullText, 
                // skip shebang at position 0
                token.pos === 0 ? (ts.getShebang(fullText) || '').length : token.pos, commentCallback);
            if (notJsx || canHaveTrailingTrivia(token))
                return ts.forEachTrailingCommentRange(fullText, token.end, commentCallback);
        }, sourceFile);
        function commentCallback(pos, end, kind) {
            cb(fullText, { pos, end, kind });
        }
    }
    exports.forEachComment = forEachComment;
    /** Exclude trailing positions that would lead to scanning for trivia inside JsxText */
    function canHaveTrailingTrivia(token) {
        switch (token.kind) {
            case ts.SyntaxKind.CloseBraceToken:
                // after a JsxExpression inside a JsxElement's body can only be other JsxChild, but no trivia
                return token.parent.kind !== ts.SyntaxKind.JsxExpression || !isJsxElementOrFragment(token.parent.parent);
            case ts.SyntaxKind.GreaterThanToken:
                switch (token.parent.kind) {
                    case ts.SyntaxKind.JsxOpeningElement:
                        // if end is not equal, this is part of the type arguments list. in all other cases it would be inside the element body
                        return token.end !== token.parent.end;
                    case ts.SyntaxKind.JsxOpeningFragment:
                        return false; // would be inside the fragment
                    case ts.SyntaxKind.JsxSelfClosingElement:
                        return token.end !== token.parent.end || // if end is not equal, this is part of the type arguments list
                            !isJsxElementOrFragment(token.parent.parent); // there's only trailing trivia if it's the end of the top element
                    case ts.SyntaxKind.JsxClosingElement:
                    case ts.SyntaxKind.JsxClosingFragment:
                        // there's only trailing trivia if it's the end of the top element
                        return !isJsxElementOrFragment(token.parent.parent.parent);
                }
        }
        return true;
    }
    function isJsxElementOrFragment(node) {
        return node.kind === ts.SyntaxKind.JsxElement || node.kind === ts.SyntaxKind.JsxFragment;
    }
    function getLineRanges(sourceFile) {
        const lineStarts = sourceFile.getLineStarts();
        const result = [];
        const length = lineStarts.length;
        const sourceText = sourceFile.text;
        let pos = 0;
        for (let i = 1; i < length; ++i) {
            const end = lineStarts[i];
            let lineEnd = end;
            for (; lineEnd > pos; --lineEnd)
                if (!ts.isLineBreak(sourceText.charCodeAt(lineEnd - 1)))
                    break;
            result.push({
                pos,
                end,
                contentLength: lineEnd - pos,
            });
            pos = end;
        }
        result.push({
            pos,
            end: sourceFile.end,
            contentLength: sourceFile.end - pos,
        });
        return result;
    }
    exports.getLineRanges = getLineRanges;
    /** Get the line break style used in sourceFile. This function only looks at the first line break. If there is none, \n is assumed. */
    function getLineBreakStyle(sourceFile) {
        const lineStarts = sourceFile.getLineStarts();
        return lineStarts.length === 1 || lineStarts[1] < 2 || sourceFile.text[lineStarts[1] - 2] !== '\r'
            ? '\n'
            : '\r\n';
    }
    exports.getLineBreakStyle = getLineBreakStyle;
    let cachedScanner;
    function scanToken(text, languageVersion) {
        if (cachedScanner === undefined) {
            // cache scanner
            cachedScanner = ts.createScanner(languageVersion, false, undefined, text);
        }
        else {
            cachedScanner.setScriptTarget(languageVersion);
            cachedScanner.setText(text);
        }
        cachedScanner.scan();
        return cachedScanner;
    }
    /**
     * Determines whether the given text parses as a standalone identifier.
     * This is not a guarantee that it works in every context. The property name in PropertyAccessExpressions for example allows reserved words.
     * Depending on the context it could be parsed as contextual keyword or TypeScript keyword.
     */
    function isValidIdentifier(text, languageVersion = ts.ScriptTarget.Latest) {
        const scan = scanToken(text, languageVersion);
        return scan.isIdentifier() && scan.getTextPos() === text.length && scan.getTokenPos() === 0;
    }
    exports.isValidIdentifier = isValidIdentifier;
    function charSize(ch) {
        return ch >= 0x10000 ? 2 : 1;
    }
    /**
     * Determines whether the given text can be used to access a property with a PropertyAccessExpression while preserving the property's name.
     */
    function isValidPropertyAccess(text, languageVersion = ts.ScriptTarget.Latest) {
        if (text.length === 0)
            return false;
        let ch = text.codePointAt(0);
        if (!ts.isIdentifierStart(ch, languageVersion))
            return false;
        for (let i = charSize(ch); i < text.length; i += charSize(ch)) {
            ch = text.codePointAt(i);
            if (!ts.isIdentifierPart(ch, languageVersion))
                return false;
        }
        return true;
    }
    exports.isValidPropertyAccess = isValidPropertyAccess;
    /**
     * Determines whether the given text can be used as unquoted name of a property declaration while preserving the property's name.
     */
    function isValidPropertyName(text, languageVersion = ts.ScriptTarget.Latest) {
        if (isValidPropertyAccess(text, languageVersion))
            return true;
        const scan = scanToken(text, languageVersion);
        return scan.getTextPos() === text.length &&
            scan.getToken() === ts.SyntaxKind.NumericLiteral && scan.getTokenValue() === text; // ensure stringified number equals literal
    }
    exports.isValidPropertyName = isValidPropertyName;
    /**
     * Determines whether the given text can be parsed as a numeric literal.
     */
    function isValidNumericLiteral(text, languageVersion = ts.ScriptTarget.Latest) {
        const scan = scanToken(text, languageVersion);
        return scan.getToken() === ts.SyntaxKind.NumericLiteral && scan.getTextPos() === text.length && scan.getTokenPos() === 0;
    }
    exports.isValidNumericLiteral = isValidNumericLiteral;
    /**
     * Determines whether the given text can be used as JSX tag or attribute name while preserving the exact name.
     */
    function isValidJsxIdentifier(text, languageVersion = ts.ScriptTarget.Latest) {
        if (text.length === 0)
            return false;
        let seenNamespaceSeparator = false;
        let ch = text.codePointAt(0);
        if (!ts.isIdentifierStart(ch, languageVersion))
            return false;
        for (let i = charSize(ch); i < text.length; i += charSize(ch)) {
            ch = text.codePointAt(i);
            if (!ts.isIdentifierPart(ch, languageVersion) && ch !== 45 /* minus */) {
                if (!seenNamespaceSeparator && ch === 58 /* colon */ && i + charSize(ch) !== text.length) {
                    seenNamespaceSeparator = true;
                }
                else {
                    return false;
                }
            }
        }
        return true;
    }
    exports.isValidJsxIdentifier = isValidJsxIdentifier;
    function isNumericPropertyName(name) {
        return String(+name) === name;
    }
    exports.isNumericPropertyName = isNumericPropertyName;
    function isSameLine(sourceFile, pos1, pos2) {
        return ts.getLineAndCharacterOfPosition(sourceFile, pos1).line === ts.getLineAndCharacterOfPosition(sourceFile, pos2).line;
    }
    exports.isSameLine = isSameLine;
    var SideEffectOptions;
    (function (SideEffectOptions) {
        SideEffectOptions[SideEffectOptions["None"] = 0] = "None";
        SideEffectOptions[SideEffectOptions["TaggedTemplate"] = 1] = "TaggedTemplate";
        SideEffectOptions[SideEffectOptions["Constructor"] = 2] = "Constructor";
        SideEffectOptions[SideEffectOptions["JsxElement"] = 4] = "JsxElement";
    })(SideEffectOptions = exports.SideEffectOptions || (exports.SideEffectOptions = {}));
    function hasSideEffects(node, options) {
        var _a, _b;
        const queue = [];
        while (true) {
            switch (node.kind) {
                case ts.SyntaxKind.CallExpression:
                case ts.SyntaxKind.PostfixUnaryExpression:
                case ts.SyntaxKind.AwaitExpression:
                case ts.SyntaxKind.YieldExpression:
                case ts.SyntaxKind.DeleteExpression:
                    return true;
                case ts.SyntaxKind.TypeAssertionExpression:
                case ts.SyntaxKind.AsExpression:
                case ts.SyntaxKind.ParenthesizedExpression:
                case ts.SyntaxKind.NonNullExpression:
                case ts.SyntaxKind.VoidExpression:
                case ts.SyntaxKind.TypeOfExpression:
                case ts.SyntaxKind.PropertyAccessExpression:
                case ts.SyntaxKind.SpreadElement:
                case ts.SyntaxKind.PartiallyEmittedExpression:
                    node = node.expression;
                    continue;
                case ts.SyntaxKind.BinaryExpression:
                    if (isAssignmentKind(node.operatorToken.kind))
                        return true;
                    queue.push(node.right);
                    node = node.left;
                    continue;
                case ts.SyntaxKind.PrefixUnaryExpression:
                    switch (node.operator) {
                        case ts.SyntaxKind.PlusPlusToken:
                        case ts.SyntaxKind.MinusMinusToken:
                            return true;
                        default:
                            node = node.operand;
                            continue;
                    }
                case ts.SyntaxKind.ElementAccessExpression:
                    if (node.argumentExpression !== undefined) // for compatibility with typescript@<2.9.0
                        queue.push(node.argumentExpression);
                    node = node.expression;
                    continue;
                case ts.SyntaxKind.ConditionalExpression:
                    queue.push(node.whenTrue, node.whenFalse);
                    node = node.condition;
                    continue;
                case ts.SyntaxKind.NewExpression:
                    if (options & 2 /* Constructor */)
                        return true;
                    if (node.arguments !== undefined)
                        queue.push(...node.arguments);
                    node = node.expression;
                    continue;
                case ts.SyntaxKind.TaggedTemplateExpression:
                    if (options & 1 /* TaggedTemplate */)
                        return true;
                    queue.push(node.tag);
                    node = node.template;
                    if (node.kind === ts.SyntaxKind.NoSubstitutionTemplateLiteral)
                        break;
                // falls through
                case ts.SyntaxKind.TemplateExpression:
                    for (const child of node.templateSpans)
                        queue.push(child.expression);
                    break;
                case ts.SyntaxKind.ClassExpression: {
                    if (node.decorators !== undefined)
                        return true;
                    for (const child of node.members) {
                        if (child.decorators !== undefined)
                            return true;
                        if (!hasModifier(child.modifiers, ts.SyntaxKind.DeclareKeyword)) {
                            if (((_a = child.name) === null || _a === void 0 ? void 0 : _a.kind) === ts.SyntaxKind.ComputedPropertyName)
                                queue.push(child.name.expression);
                            if (node_1.isMethodDeclaration(child)) {
                                for (const p of child.parameters)
                                    if (p.decorators !== undefined)
                                        return true;
                            }
                            else if (node_1.isPropertyDeclaration(child) &&
                                child.initializer !== undefined &&
                                hasModifier(child.modifiers, ts.SyntaxKind.StaticKeyword)) {
                                queue.push(child.initializer);
                            }
                        }
                    }
                    const base = getBaseOfClassLikeExpression(node);
                    if (base === undefined)
                        break;
                    node = base.expression;
                    continue;
                }
                case ts.SyntaxKind.ArrayLiteralExpression:
                    queue.push(...node.elements);
                    break;
                case ts.SyntaxKind.ObjectLiteralExpression:
                    for (const child of node.properties) {
                        if (((_b = child.name) === null || _b === void 0 ? void 0 : _b.kind) === ts.SyntaxKind.ComputedPropertyName)
                            queue.push(child.name.expression);
                        switch (child.kind) {
                            case ts.SyntaxKind.PropertyAssignment:
                                queue.push(child.initializer);
                                break;
                            case ts.SyntaxKind.SpreadAssignment:
                                queue.push(child.expression);
                        }
                    }
                    break;
                case ts.SyntaxKind.JsxExpression:
                    if (node.expression === undefined)
                        break;
                    node = node.expression;
                    continue;
                case ts.SyntaxKind.JsxElement:
                case ts.SyntaxKind.JsxFragment:
                    for (const child of node.children)
                        if (child.kind !== ts.SyntaxKind.JsxText)
                            queue.push(child);
                    if (node.kind === ts.SyntaxKind.JsxFragment)
                        break;
                    node = node.openingElement;
                // falls through
                case ts.SyntaxKind.JsxSelfClosingElement:
                case ts.SyntaxKind.JsxOpeningElement:
                    if (options & 4 /* JsxElement */)
                        return true;
                    for (const child of node.attributes.properties) {
                        if (child.kind === ts.SyntaxKind.JsxSpreadAttribute) {
                            queue.push(child.expression);
                        }
                        else if (child.initializer !== undefined) {
                            queue.push(child.initializer);
                        }
                    }
                    break;
                case ts.SyntaxKind.CommaListExpression:
                    queue.push(...node.elements);
            }
            if (queue.length === 0)
                return false;
            node = queue.pop();
        }
    }
    exports.hasSideEffects = hasSideEffects;
    /** Returns the VariableDeclaration or ParameterDeclaration that contains the BindingElement */
    function getDeclarationOfBindingElement(node) {
        let parent = node.parent.parent;
        while (parent.kind === ts.SyntaxKind.BindingElement)
            parent = parent.parent.parent;
        return parent;
    }
    exports.getDeclarationOfBindingElement = getDeclarationOfBindingElement;
    function isExpressionValueUsed(node) {
        while (true) {
            const parent = node.parent;
            switch (parent.kind) {
                case ts.SyntaxKind.CallExpression:
                case ts.SyntaxKind.NewExpression:
                case ts.SyntaxKind.ElementAccessExpression:
                case ts.SyntaxKind.WhileStatement:
                case ts.SyntaxKind.DoStatement:
                case ts.SyntaxKind.WithStatement:
                case ts.SyntaxKind.ThrowStatement:
                case ts.SyntaxKind.ReturnStatement:
                case ts.SyntaxKind.JsxExpression:
                case ts.SyntaxKind.JsxSpreadAttribute:
                case ts.SyntaxKind.JsxElement:
                case ts.SyntaxKind.JsxFragment:
                case ts.SyntaxKind.JsxSelfClosingElement:
                case ts.SyntaxKind.ComputedPropertyName:
                case ts.SyntaxKind.ArrowFunction:
                case ts.SyntaxKind.ExportSpecifier:
                case ts.SyntaxKind.ExportAssignment:
                case ts.SyntaxKind.ImportDeclaration:
                case ts.SyntaxKind.ExternalModuleReference:
                case ts.SyntaxKind.Decorator:
                case ts.SyntaxKind.TaggedTemplateExpression:
                case ts.SyntaxKind.TemplateSpan:
                case ts.SyntaxKind.ExpressionWithTypeArguments:
                case ts.SyntaxKind.TypeOfExpression:
                case ts.SyntaxKind.AwaitExpression:
                case ts.SyntaxKind.YieldExpression:
                case ts.SyntaxKind.LiteralType:
                case ts.SyntaxKind.JsxAttributes:
                case ts.SyntaxKind.JsxOpeningElement:
                case ts.SyntaxKind.JsxClosingElement:
                case ts.SyntaxKind.IfStatement:
                case ts.SyntaxKind.CaseClause:
                case ts.SyntaxKind.SwitchStatement:
                    return true;
                case ts.SyntaxKind.PropertyAccessExpression:
                    return parent.expression === node;
                case ts.SyntaxKind.QualifiedName:
                    return parent.left === node;
                case ts.SyntaxKind.ShorthandPropertyAssignment:
                    return parent.objectAssignmentInitializer === node ||
                        !isInDestructuringAssignment(parent);
                case ts.SyntaxKind.PropertyAssignment:
                    return parent.initializer === node && !isInDestructuringAssignment(parent);
                case ts.SyntaxKind.SpreadAssignment:
                case ts.SyntaxKind.SpreadElement:
                case ts.SyntaxKind.ArrayLiteralExpression:
                    return !isInDestructuringAssignment(parent);
                case ts.SyntaxKind.ParenthesizedExpression:
                case ts.SyntaxKind.AsExpression:
                case ts.SyntaxKind.TypeAssertionExpression:
                case ts.SyntaxKind.PostfixUnaryExpression:
                case ts.SyntaxKind.PrefixUnaryExpression:
                case ts.SyntaxKind.NonNullExpression:
                    node = parent;
                    continue;
                case ts.SyntaxKind.ForStatement:
                    return parent.condition === node;
                case ts.SyntaxKind.ForInStatement:
                case ts.SyntaxKind.ForOfStatement:
                    return parent.expression === node;
                case ts.SyntaxKind.ConditionalExpression:
                    if (parent.condition === node)
                        return true;
                    node = parent;
                    break;
                case ts.SyntaxKind.PropertyDeclaration:
                case ts.SyntaxKind.BindingElement:
                case ts.SyntaxKind.VariableDeclaration:
                case ts.SyntaxKind.Parameter:
                case ts.SyntaxKind.EnumMember:
                    return parent.initializer === node;
                case ts.SyntaxKind.ImportEqualsDeclaration:
                    return parent.moduleReference === node;
                case ts.SyntaxKind.CommaListExpression:
                    if (parent.elements[parent.elements.length - 1] !== node)
                        return false;
                    node = parent;
                    break;
                case ts.SyntaxKind.BinaryExpression:
                    if (parent.right === node) {
                        if (parent.operatorToken.kind === ts.SyntaxKind.CommaToken) {
                            node = parent;
                            break;
                        }
                        return true;
                    }
                    switch (parent.operatorToken.kind) {
                        case ts.SyntaxKind.CommaToken:
                        case ts.SyntaxKind.EqualsToken:
                            return false;
                        case ts.SyntaxKind.EqualsEqualsEqualsToken:
                        case ts.SyntaxKind.EqualsEqualsToken:
                        case ts.SyntaxKind.ExclamationEqualsEqualsToken:
                        case ts.SyntaxKind.ExclamationEqualsToken:
                        case ts.SyntaxKind.InstanceOfKeyword:
                        case ts.SyntaxKind.PlusToken:
                        case ts.SyntaxKind.MinusToken:
                        case ts.SyntaxKind.AsteriskToken:
                        case ts.SyntaxKind.SlashToken:
                        case ts.SyntaxKind.PercentToken:
                        case ts.SyntaxKind.AsteriskAsteriskToken:
                        case ts.SyntaxKind.GreaterThanToken:
                        case ts.SyntaxKind.GreaterThanGreaterThanToken:
                        case ts.SyntaxKind.GreaterThanGreaterThanGreaterThanToken:
                        case ts.SyntaxKind.GreaterThanEqualsToken:
                        case ts.SyntaxKind.LessThanToken:
                        case ts.SyntaxKind.LessThanLessThanToken:
                        case ts.SyntaxKind.LessThanEqualsToken:
                        case ts.SyntaxKind.AmpersandToken:
                        case ts.SyntaxKind.BarToken:
                        case ts.SyntaxKind.CaretToken:
                        case ts.SyntaxKind.BarBarToken:
                        case ts.SyntaxKind.AmpersandAmpersandToken:
                        case ts.SyntaxKind.QuestionQuestionToken:
                        case ts.SyntaxKind.InKeyword:
                        case ts.SyntaxKind.QuestionQuestionEqualsToken:
                        case ts.SyntaxKind.AmpersandAmpersandEqualsToken:
                        case ts.SyntaxKind.BarBarEqualsToken:
                            return true;
                        default:
                            node = parent;
                    }
                    break;
                default:
                    return false;
            }
        }
    }
    exports.isExpressionValueUsed = isExpressionValueUsed;
    function isInDestructuringAssignment(node) {
        switch (node.kind) {
            case ts.SyntaxKind.ShorthandPropertyAssignment:
                if (node.objectAssignmentInitializer !== undefined)
                    return true;
            // falls through
            case ts.SyntaxKind.PropertyAssignment:
            case ts.SyntaxKind.SpreadAssignment:
                node = node.parent;
                break;
            case ts.SyntaxKind.SpreadElement:
                if (node.parent.kind !== ts.SyntaxKind.ArrayLiteralExpression)
                    return false;
                node = node.parent;
        }
        while (true) {
            switch (node.parent.kind) {
                case ts.SyntaxKind.BinaryExpression:
                    return node.parent.left === node &&
                        node.parent.operatorToken.kind === ts.SyntaxKind.EqualsToken;
                case ts.SyntaxKind.ForOfStatement:
                    return node.parent.initializer === node;
                case ts.SyntaxKind.ArrayLiteralExpression:
                case ts.SyntaxKind.ObjectLiteralExpression:
                    node = node.parent;
                    break;
                case ts.SyntaxKind.SpreadAssignment:
                case ts.SyntaxKind.PropertyAssignment:
                    node = node.parent.parent;
                    break;
                case ts.SyntaxKind.SpreadElement:
                    if (node.parent.parent.kind !== ts.SyntaxKind.ArrayLiteralExpression)
                        return false;
                    node = node.parent.parent;
                    break;
                default:
                    return false;
            }
        }
    }
    var AccessKind;
    (function (AccessKind) {
        AccessKind[AccessKind["None"] = 0] = "None";
        AccessKind[AccessKind["Read"] = 1] = "Read";
        AccessKind[AccessKind["Write"] = 2] = "Write";
        AccessKind[AccessKind["Delete"] = 4] = "Delete";
        AccessKind[AccessKind["ReadWrite"] = 3] = "ReadWrite";
        AccessKind[AccessKind["Modification"] = 6] = "Modification";
    })(AccessKind = exports.AccessKind || (exports.AccessKind = {}));
    function getAccessKind(node) {
        const parent = node.parent;
        switch (parent.kind) {
            case ts.SyntaxKind.DeleteExpression:
                return 4 /* Delete */;
            case ts.SyntaxKind.PostfixUnaryExpression:
                return 3 /* ReadWrite */;
            case ts.SyntaxKind.PrefixUnaryExpression:
                return parent.operator === ts.SyntaxKind.PlusPlusToken ||
                    parent.operator === ts.SyntaxKind.MinusMinusToken
                    ? 3 /* ReadWrite */
                    : 1 /* Read */;
            case ts.SyntaxKind.BinaryExpression:
                return parent.right === node
                    ? 1 /* Read */
                    : !isAssignmentKind(parent.operatorToken.kind)
                        ? 1 /* Read */
                        : parent.operatorToken.kind === ts.SyntaxKind.EqualsToken
                            ? 2 /* Write */
                            : 3 /* ReadWrite */;
            case ts.SyntaxKind.ShorthandPropertyAssignment:
                return parent.objectAssignmentInitializer === node
                    ? 1 /* Read */
                    : isInDestructuringAssignment(parent)
                        ? 2 /* Write */
                        : 1 /* Read */;
            case ts.SyntaxKind.PropertyAssignment:
                return parent.name === node
                    ? 0 /* None */
                    : isInDestructuringAssignment(parent)
                        ? 2 /* Write */
                        : 1 /* Read */;
            case ts.SyntaxKind.ArrayLiteralExpression:
            case ts.SyntaxKind.SpreadElement:
            case ts.SyntaxKind.SpreadAssignment:
                return isInDestructuringAssignment(parent)
                    ? 2 /* Write */
                    : 1 /* Read */;
            case ts.SyntaxKind.ParenthesizedExpression:
            case ts.SyntaxKind.NonNullExpression:
            case ts.SyntaxKind.TypeAssertionExpression:
            case ts.SyntaxKind.AsExpression:
                // (<number>foo! as {})++
                return getAccessKind(parent);
            case ts.SyntaxKind.ForOfStatement:
            case ts.SyntaxKind.ForInStatement:
                return parent.initializer === node
                    ? 2 /* Write */
                    : 1 /* Read */;
            case ts.SyntaxKind.ExpressionWithTypeArguments:
                return parent.parent.token === ts.SyntaxKind.ExtendsKeyword &&
                    parent.parent.parent.kind !== ts.SyntaxKind.InterfaceDeclaration
                    ? 1 /* Read */
                    : 0 /* None */;
            case ts.SyntaxKind.ComputedPropertyName:
            case ts.SyntaxKind.ExpressionStatement:
            case ts.SyntaxKind.TypeOfExpression:
            case ts.SyntaxKind.ElementAccessExpression:
            case ts.SyntaxKind.ForStatement:
            case ts.SyntaxKind.IfStatement:
            case ts.SyntaxKind.DoStatement:
            case ts.SyntaxKind.WhileStatement:
            case ts.SyntaxKind.SwitchStatement:
            case ts.SyntaxKind.WithStatement:
            case ts.SyntaxKind.ThrowStatement:
            case ts.SyntaxKind.CallExpression:
            case ts.SyntaxKind.NewExpression:
            case ts.SyntaxKind.TaggedTemplateExpression:
            case ts.SyntaxKind.JsxExpression:
            case ts.SyntaxKind.Decorator:
            case ts.SyntaxKind.TemplateSpan:
            case ts.SyntaxKind.JsxOpeningElement:
            case ts.SyntaxKind.JsxSelfClosingElement:
            case ts.SyntaxKind.JsxSpreadAttribute:
            case ts.SyntaxKind.VoidExpression:
            case ts.SyntaxKind.ReturnStatement:
            case ts.SyntaxKind.AwaitExpression:
            case ts.SyntaxKind.YieldExpression:
            case ts.SyntaxKind.ConditionalExpression:
            case ts.SyntaxKind.CaseClause:
            case ts.SyntaxKind.JsxElement:
                return 1 /* Read */;
            case ts.SyntaxKind.ArrowFunction:
                return parent.body === node
                    ? 1 /* Read */
                    : 2 /* Write */;
            case ts.SyntaxKind.PropertyDeclaration:
            case ts.SyntaxKind.VariableDeclaration:
            case ts.SyntaxKind.Parameter:
            case ts.SyntaxKind.EnumMember:
            case ts.SyntaxKind.BindingElement:
            case ts.SyntaxKind.JsxAttribute:
                return parent.initializer === node
                    ? 1 /* Read */
                    : 0 /* None */;
            case ts.SyntaxKind.PropertyAccessExpression:
                return parent.expression === node
                    ? 1 /* Read */
                    : 0 /* None */;
            case ts.SyntaxKind.ExportAssignment:
                return parent.isExportEquals
                    ? 1 /* Read */
                    : 0 /* None */;
        }
        return 0 /* None */;
    }
    exports.getAccessKind = getAccessKind;
    function isReassignmentTarget(node) {
        return (getAccessKind(node) & 2 /* Write */) !== 0;
    }
    exports.isReassignmentTarget = isReassignmentTarget;
    function canHaveJsDoc(node) {
        const kind = node.kind;
        switch (kind) {
            case ts.SyntaxKind.Parameter:
            case ts.SyntaxKind.CallSignature:
            case ts.SyntaxKind.ConstructSignature:
            case ts.SyntaxKind.MethodSignature:
            case ts.SyntaxKind.PropertySignature:
            case ts.SyntaxKind.ArrowFunction:
            case ts.SyntaxKind.ParenthesizedExpression:
            case ts.SyntaxKind.SpreadAssignment:
            case ts.SyntaxKind.ShorthandPropertyAssignment:
            case ts.SyntaxKind.PropertyAssignment:
            case ts.SyntaxKind.FunctionExpression:
            case ts.SyntaxKind.LabeledStatement:
            case ts.SyntaxKind.ExpressionStatement:
            case ts.SyntaxKind.VariableStatement:
            case ts.SyntaxKind.FunctionDeclaration:
            case ts.SyntaxKind.Constructor:
            case ts.SyntaxKind.MethodDeclaration:
            case ts.SyntaxKind.PropertyDeclaration:
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.SetAccessor:
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.ClassExpression:
            case ts.SyntaxKind.InterfaceDeclaration:
            case ts.SyntaxKind.TypeAliasDeclaration:
            case ts.SyntaxKind.EnumMember:
            case ts.SyntaxKind.EnumDeclaration:
            case ts.SyntaxKind.ModuleDeclaration:
            case ts.SyntaxKind.ImportEqualsDeclaration:
            case ts.SyntaxKind.ImportDeclaration:
            case ts.SyntaxKind.NamespaceExportDeclaration:
            case ts.SyntaxKind.ExportAssignment:
            case ts.SyntaxKind.IndexSignature:
            case ts.SyntaxKind.FunctionType:
            case ts.SyntaxKind.ConstructorType:
            case ts.SyntaxKind.JSDocFunctionType:
            case ts.SyntaxKind.ExportDeclaration:
            case ts.SyntaxKind.NamedTupleMember:
            case ts.SyntaxKind.EndOfFileToken:
                return true;
            default:
                return false;
        }
    }
    exports.canHaveJsDoc = canHaveJsDoc;
    /** Gets the JSDoc of a node. For performance reasons this function should only be called when `canHaveJsDoc` returns true. */
    function getJsDoc(node, sourceFile) {
        const result = [];
        for (const child of node.getChildren(sourceFile)) {
            if (!node_1.isJsDoc(child))
                break;
            result.push(child);
        }
        return result;
    }
    exports.getJsDoc = getJsDoc;
    /**
     * Parses the JsDoc of any node. This function is made for nodes that don't get their JsDoc parsed by the TypeScript parser.
     *
     * @param considerTrailingComments When set to `true` this function uses the trailing comments if the node starts on the same line
     *                                 as the previous node ends.
     */
    function parseJsDocOfNode(node, considerTrailingComments, sourceFile = node.getSourceFile()) {
        if (canHaveJsDoc(node) && node.kind !== ts.SyntaxKind.EndOfFileToken) {
            const result = getJsDoc(node, sourceFile);
            if (result.length !== 0 || !considerTrailingComments)
                return result;
        }
        return parseJsDocWorker(node, node.getStart(sourceFile), sourceFile, considerTrailingComments);
    }
    exports.parseJsDocOfNode = parseJsDocOfNode;
    function parseJsDocWorker(node, nodeStart, sourceFile, considerTrailingComments) {
        const start = ts[considerTrailingComments && isSameLine(sourceFile, node.pos, nodeStart)
            ? 'forEachTrailingCommentRange'
            : 'forEachLeadingCommentRange'](sourceFile.text, node.pos, 
        // return object to make `0` a truthy value
        (pos, _end, kind) => kind === ts.SyntaxKind.MultiLineCommentTrivia && sourceFile.text[pos + 2] === '*' ? { pos } : undefined);
        if (start === undefined)
            return [];
        const startPos = start.pos;
        const text = sourceFile.text.slice(startPos, nodeStart);
        const newSourceFile = ts.createSourceFile('jsdoc.ts', `${text}var a;`, sourceFile.languageVersion);
        const result = getJsDoc(newSourceFile.statements[0], newSourceFile);
        for (const doc of result)
            updateNode(doc, node);
        return result;
        function updateNode(n, parent) {
            n.pos += startPos;
            n.end += startPos;
            n.parent = parent;
            return ts.forEachChild(n, (child) => updateNode(child, n), (children) => {
                children.pos += startPos;
                children.end += startPos;
                for (const child of children)
                    updateNode(child, n);
            });
        }
    }
    var ImportKind;
    (function (ImportKind) {
        ImportKind[ImportKind["ImportDeclaration"] = 1] = "ImportDeclaration";
        ImportKind[ImportKind["ImportEquals"] = 2] = "ImportEquals";
        ImportKind[ImportKind["ExportFrom"] = 4] = "ExportFrom";
        ImportKind[ImportKind["DynamicImport"] = 8] = "DynamicImport";
        ImportKind[ImportKind["Require"] = 16] = "Require";
        ImportKind[ImportKind["ImportType"] = 32] = "ImportType";
        ImportKind[ImportKind["All"] = 63] = "All";
        ImportKind[ImportKind["AllImports"] = 59] = "AllImports";
        ImportKind[ImportKind["AllStaticImports"] = 3] = "AllStaticImports";
        ImportKind[ImportKind["AllImportExpressions"] = 24] = "AllImportExpressions";
        ImportKind[ImportKind["AllRequireLike"] = 18] = "AllRequireLike";
        // @internal
        ImportKind[ImportKind["AllNestedImports"] = 56] = "AllNestedImports";
        // @internal
        ImportKind[ImportKind["AllTopLevelImports"] = 7] = "AllTopLevelImports";
    })(ImportKind = exports.ImportKind || (exports.ImportKind = {}));
    function findImports(sourceFile, kinds, ignoreFileName = true) {
        const result = [];
        for (const node of findImportLikeNodes(sourceFile, kinds, ignoreFileName)) {
            switch (node.kind) {
                case ts.SyntaxKind.ImportDeclaration:
                    addIfTextualLiteral(node.moduleSpecifier);
                    break;
                case ts.SyntaxKind.ImportEqualsDeclaration:
                    addIfTextualLiteral(node.moduleReference.expression);
                    break;
                case ts.SyntaxKind.ExportDeclaration:
                    addIfTextualLiteral(node.moduleSpecifier);
                    break;
                case ts.SyntaxKind.CallExpression:
                    addIfTextualLiteral(node.arguments[0]);
                    break;
                case ts.SyntaxKind.ImportType:
                    if (node_1.isLiteralTypeNode(node.argument))
                        addIfTextualLiteral(node.argument.literal);
                    break;
                default:
                    throw new Error('unexpected node');
            }
        }
        return result;
        function addIfTextualLiteral(node) {
            if (node_1.isTextualLiteral(node))
                result.push(node);
        }
    }
    exports.findImports = findImports;
    function findImportLikeNodes(sourceFile, kinds, ignoreFileName = true) {
        return new ImportFinder(sourceFile, kinds, ignoreFileName).find();
    }
    exports.findImportLikeNodes = findImportLikeNodes;
    class ImportFinder {
        constructor(_sourceFile, _options, _ignoreFileName) {
            this._sourceFile = _sourceFile;
            this._options = _options;
            this._ignoreFileName = _ignoreFileName;
            this._result = [];
        }
        find() {
            if (this._sourceFile.isDeclarationFile)
                this._options &= ~24 /* AllImportExpressions */;
            if (this._options & 7 /* AllTopLevelImports */)
                this._findImports(this._sourceFile.statements);
            if (this._options & 56 /* AllNestedImports */)
                this._findNestedImports();
            return this._result;
        }
        _findImports(statements) {
            for (const statement of statements) {
                if (node_1.isImportDeclaration(statement)) {
                    if (this._options & 1 /* ImportDeclaration */)
                        this._result.push(statement);
                }
                else if (node_1.isImportEqualsDeclaration(statement)) {
                    if (this._options & 2 /* ImportEquals */ &&
                        statement.moduleReference.kind === ts.SyntaxKind.ExternalModuleReference)
                        this._result.push(statement);
                }
                else if (node_1.isExportDeclaration(statement)) {
                    if (statement.moduleSpecifier !== undefined && this._options & 4 /* ExportFrom */)
                        this._result.push(statement);
                }
                else if (node_1.isModuleDeclaration(statement)) {
                    this._findImportsInModule(statement);
                }
            }
        }
        _findImportsInModule(declaration) {
            if (declaration.body === undefined)
                return;
            if (declaration.body.kind === ts.SyntaxKind.ModuleDeclaration)
                return this._findImportsInModule(declaration.body);
            this._findImports(declaration.body.statements);
        }
        _findNestedImports() {
            const isJavaScriptFile = this._ignoreFileName || (this._sourceFile.flags & ts.NodeFlags.JavaScriptFile) !== 0;
            let re;
            let includeJsDoc;
            if ((this._options & 56 /* AllNestedImports */) === 16 /* Require */) {
                if (!isJavaScriptFile)
                    return; // don't look for 'require' in TS files
                re = /\brequire\s*[</(]/g;
                includeJsDoc = false;
            }
            else if (this._options & 16 /* Require */ && isJavaScriptFile) {
                re = /\b(?:import|require)\s*[</(]/g;
                includeJsDoc = (this._options & 32 /* ImportType */) !== 0;
            }
            else {
                re = /\bimport\s*[</(]/g;
                includeJsDoc = isJavaScriptFile && (this._options & 32 /* ImportType */) !== 0;
            }
            for (let match = re.exec(this._sourceFile.text); match !== null; match = re.exec(this._sourceFile.text)) {
                const token = getTokenAtPositionWorker(this._sourceFile, match.index, this._sourceFile, 
                // only look for ImportTypeNode within JSDoc in JS files
                match[0][0] === 'i' && includeJsDoc);
                if (token.kind === ts.SyntaxKind.ImportKeyword) {
                    if (token.end - 'import'.length !== match.index)
                        continue;
                    switch (token.parent.kind) {
                        case ts.SyntaxKind.ImportType:
                            this._result.push(token.parent);
                            break;
                        case ts.SyntaxKind.CallExpression:
                            if (token.parent.arguments.length > 1)
                                this._result.push(token.parent);
                    }
                }
                else if (token.kind === ts.SyntaxKind.Identifier &&
                    token.end - 'require'.length === match.index &&
                    token.parent.kind === ts.SyntaxKind.CallExpression &&
                    token.parent.expression === token &&
                    token.parent.arguments.length === 1) {
                    this._result.push(token.parent);
                }
            }
        }
    }
    /**
     * Ambient context means the statement itself has the `declare` keyword
     * or is inside a `declare namespace`,  `delcare module` or `declare global`.
     */
    function isStatementInAmbientContext(node) {
        while (node.flags & ts.NodeFlags.NestedNamespace)
            node = node.parent;
        return hasModifier(node.modifiers, ts.SyntaxKind.DeclareKeyword) || isAmbientModuleBlock(node.parent);
    }
    exports.isStatementInAmbientContext = isStatementInAmbientContext;
    /** Includes `declare namespace`, `declare module` and `declare global` and namespace nested in one of the aforementioned. */
    function isAmbientModuleBlock(node) {
        while (node.kind === ts.SyntaxKind.ModuleBlock) {
            do
                node = node.parent;
            while (node.flags & ts.NodeFlags.NestedNamespace);
            if (hasModifier(node.modifiers, ts.SyntaxKind.DeclareKeyword))
                return true;
            node = node.parent;
        }
        return false;
    }
    exports.isAmbientModuleBlock = isAmbientModuleBlock;
    function getIIFE(func) {
        let node = func.parent;
        while (node.kind === ts.SyntaxKind.ParenthesizedExpression)
            node = node.parent;
        return node_1.isCallExpression(node) && func.end <= node.expression.end ? node : undefined;
    }
    exports.getIIFE = getIIFE;
    function isStrictCompilerOptionEnabled(options, option) {
        return (options.strict ? options[option] !== false : options[option] === true) &&
            (option !== 'strictPropertyInitialization' || isStrictCompilerOptionEnabled(options, 'strictNullChecks'));
    }
    exports.isStrictCompilerOptionEnabled = isStrictCompilerOptionEnabled;
    // https://github.com/ajafff/tslint-consistent-codestyle/issues/85
    /**
     * Checks if a given compiler option is enabled.
     * It handles dependencies of options, e.g. `declaration` is implicitly enabled by `composite` or `strictNullChecks` is enabled by `strict`.
     * However, it does not check dependencies that are already checked and reported as errors, e.g. `checkJs` without `allowJs`.
     * This function only handles boolean flags.
     */
    function isCompilerOptionEnabled(options, option) {
        switch (option) {
            case 'stripInternal':
            case 'declarationMap':
            case 'emitDeclarationOnly':
                return options[option] === true && isCompilerOptionEnabled(options, 'declaration');
            case 'declaration':
                return options.declaration || isCompilerOptionEnabled(options, 'composite');
            case 'incremental':
                return options.incremental === undefined ? isCompilerOptionEnabled(options, 'composite') : options.incremental;
            case 'skipDefaultLibCheck':
                return options.skipDefaultLibCheck || isCompilerOptionEnabled(options, 'skipLibCheck');
            case 'suppressImplicitAnyIndexErrors':
                return options.suppressImplicitAnyIndexErrors === true && isCompilerOptionEnabled(options, 'noImplicitAny');
            case 'allowSyntheticDefaultImports':
                return options.allowSyntheticDefaultImports !== undefined
                    ? options.allowSyntheticDefaultImports
                    : isCompilerOptionEnabled(options, 'esModuleInterop') || options.module === ts.ModuleKind.System;
            case 'noUncheckedIndexedAccess':
                return options.noUncheckedIndexedAccess === true && isCompilerOptionEnabled(options, 'strictNullChecks');
            case 'allowJs':
                return options.allowJs === undefined ? isCompilerOptionEnabled(options, 'checkJs') : options.allowJs;
            case 'noImplicitAny':
            case 'noImplicitThis':
            case 'strictNullChecks':
            case 'strictFunctionTypes':
            case 'strictPropertyInitialization':
            case 'alwaysStrict':
            case 'strictBindCallApply':
                return isStrictCompilerOptionEnabled(options, option);
        }
        return options[option] === true;
    }
    exports.isCompilerOptionEnabled = isCompilerOptionEnabled;
    /**
     * Has nothing to do with `isAmbientModuleBlock`.
     *
     * @returns `true` if it's a global augmentation or has a string name.
     */
    function isAmbientModule(node) {
        return node.name.kind === ts.SyntaxKind.StringLiteral || (node.flags & ts.NodeFlags.GlobalAugmentation) !== 0;
    }
    exports.isAmbientModule = isAmbientModule;
    /**
     * @deprecated use `getTsCheckDirective` instead since `// @ts-nocheck` is no longer restricted to JS files.
     * @returns the last `// @ts-check` or `// @ts-nocheck` directive in the given file.
     */
    function getCheckJsDirective(source) {
        return getTsCheckDirective(source);
    }
    exports.getCheckJsDirective = getCheckJsDirective;
    /** @returns the last `// @ts-check` or `// @ts-nocheck` directive in the given file. */
    function getTsCheckDirective(source) {
        let directive;
        // needs to work around a shebang issue until https://github.com/Microsoft/TypeScript/issues/28477 is resolved
        ts.forEachLeadingCommentRange(source, (ts.getShebang(source) || '').length, (pos, end, kind) => {
            if (kind === ts.SyntaxKind.SingleLineCommentTrivia) {
                const text = source.slice(pos, end);
                const match = /^\/{2,3}\s*@ts-(no)?check(?:\s|$)/i.exec(text);
                if (match !== null)
                    directive = { pos, end, enabled: match[1] === undefined };
            }
        });
        return directive;
    }
    exports.getTsCheckDirective = getTsCheckDirective;
    function isConstAssertion(node) {
        return node_1.isTypeReferenceNode(node.type) &&
            node.type.typeName.kind === ts.SyntaxKind.Identifier &&
            node.type.typeName.escapedText === 'const';
    }
    exports.isConstAssertion = isConstAssertion;
    /** Detects whether an expression is affected by an enclosing 'as const' assertion and therefore treated literally. */
    function isInConstContext(node) {
        let current = node;
        while (true) {
            const parent = current.parent;
            outer: switch (parent.kind) {
                case ts.SyntaxKind.TypeAssertionExpression:
                case ts.SyntaxKind.AsExpression:
                    return isConstAssertion(parent);
                case ts.SyntaxKind.PrefixUnaryExpression:
                    if (current.kind !== ts.SyntaxKind.NumericLiteral)
                        return false;
                    switch (parent.operator) {
                        case ts.SyntaxKind.PlusToken:
                        case ts.SyntaxKind.MinusToken:
                            current = parent;
                            break outer;
                        default:
                            return false;
                    }
                case ts.SyntaxKind.PropertyAssignment:
                    if (parent.initializer !== current)
                        return false;
                    current = parent.parent;
                    break;
                case ts.SyntaxKind.ShorthandPropertyAssignment:
                    current = parent.parent;
                    break;
                case ts.SyntaxKind.ParenthesizedExpression:
                case ts.SyntaxKind.ArrayLiteralExpression:
                case ts.SyntaxKind.ObjectLiteralExpression:
                case ts.SyntaxKind.TemplateExpression:
                    current = parent;
                    break;
                default:
                    return false;
            }
        }
    }
    exports.isInConstContext = isInConstContext;
    /** Returns true for `Object.defineProperty(o, 'prop', {value, writable: false})` and  `Object.defineProperty(o, 'prop', {get: () => 1})`*/
    function isReadonlyAssignmentDeclaration(node, checker) {
        if (!isBindableObjectDefinePropertyCall(node))
            return false;
        const descriptorType = checker.getTypeAtLocation(node.arguments[2]);
        if (descriptorType.getProperty('value') === undefined)
            return descriptorType.getProperty('set') === undefined;
        const writableProp = descriptorType.getProperty('writable');
        if (writableProp === undefined)
            return false;
        const writableType = writableProp.valueDeclaration !== undefined && node_1.isPropertyAssignment(writableProp.valueDeclaration)
            ? checker.getTypeAtLocation(writableProp.valueDeclaration.initializer)
            : checker.getTypeOfSymbolAtLocation(writableProp, node.arguments[2]);
        return type_1.isBooleanLiteralType(writableType, false);
    }
    exports.isReadonlyAssignmentDeclaration = isReadonlyAssignmentDeclaration;
    /** Determines whether a call to `Object.defineProperty` is statically analyzable. */
    function isBindableObjectDefinePropertyCall(node) {
        return node.arguments.length === 3 &&
            node_1.isEntityNameExpression(node.arguments[0]) &&
            node_1.isNumericOrStringLikeLiteral(node.arguments[1]) &&
            node_1.isPropertyAccessExpression(node.expression) &&
            node.expression.name.escapedText === 'defineProperty' &&
            node_1.isIdentifier(node.expression.expression) &&
            node.expression.expression.escapedText === 'Object';
    }
    exports.isBindableObjectDefinePropertyCall = isBindableObjectDefinePropertyCall;
    function isWellKnownSymbolLiterally(node) {
        return ts.isPropertyAccessExpression(node) &&
            ts.isIdentifier(node.expression) &&
            node.expression.escapedText === 'Symbol';
    }
    exports.isWellKnownSymbolLiterally = isWellKnownSymbolLiterally;
    /** @deprecated typescript 4.3 removed the concept of literal well known symbols. Use `getPropertyNameFromType` instead. */
    function getPropertyNameOfWellKnownSymbol(node) {
        return {
            displayName: `[Symbol.${node.name.text}]`,
            symbolName: ('__@' + node.name.text),
        };
    }
    exports.getPropertyNameOfWellKnownSymbol = getPropertyNameOfWellKnownSymbol;
    const isTsBefore43 = (([major, minor]) => major < '4' || major === '4' && minor < '3')(ts.versionMajorMinor.split('.'));
    function getLateBoundPropertyNames(node, checker) {
        const result = {
            known: true,
            names: [],
        };
        node = unwrapParentheses(node);
        if (isTsBefore43 && isWellKnownSymbolLiterally(node)) {
            result.names.push(getPropertyNameOfWellKnownSymbol(node)); // wotan-disable-line no-unstable-api-use
        }
        else {
            const type = checker.getTypeAtLocation(node);
            for (const key of type_1.unionTypeParts(checker.getBaseConstraintOfType(type) || type)) {
                const propertyName = type_1.getPropertyNameFromType(key);
                if (propertyName) {
                    result.names.push(propertyName);
                }
                else {
                    result.known = false;
                }
            }
        }
        return result;
    }
    exports.getLateBoundPropertyNames = getLateBoundPropertyNames;
    function getLateBoundPropertyNamesOfPropertyName(node, checker) {
        const staticName = getPropertyName(node);
        return staticName !== undefined
            ? { known: true, names: [{ displayName: staticName, symbolName: ts.escapeLeadingUnderscores(staticName) }] }
            : node.kind === ts.SyntaxKind.PrivateIdentifier
                ? { known: true, names: [{ displayName: node.text, symbolName: checker.getSymbolAtLocation(node).escapedName }] }
                : getLateBoundPropertyNames(node.expression, checker);
    }
    exports.getLateBoundPropertyNamesOfPropertyName = getLateBoundPropertyNamesOfPropertyName;
    /** Most declarations demand there to be only one statically known name, e.g. class members with computed name. */
    function getSingleLateBoundPropertyNameOfPropertyName(node, checker) {
        const staticName = getPropertyName(node);
        if (staticName !== undefined)
            return { displayName: staticName, symbolName: ts.escapeLeadingUnderscores(staticName) };
        if (node.kind === ts.SyntaxKind.PrivateIdentifier)
            return { displayName: node.text, symbolName: checker.getSymbolAtLocation(node).escapedName };
        const { expression } = node;
        return isTsBefore43 && isWellKnownSymbolLiterally(expression)
            ? getPropertyNameOfWellKnownSymbol(expression) // wotan-disable-line no-unstable-api-use
            : type_1.getPropertyNameFromType(checker.getTypeAtLocation(expression));
    }
    exports.getSingleLateBoundPropertyNameOfPropertyName = getSingleLateBoundPropertyNameOfPropertyName;
    function unwrapParentheses(node) {
        while (node.kind === ts.SyntaxKind.ParenthesizedExpression)
            node = node.expression;
        return node;
    }
    exports.unwrapParentheses = unwrapParentheses;
    function formatPseudoBigInt(v) {
        return `${v.negative ? '-' : ''}${v.base10Value}n`;
    }
    exports.formatPseudoBigInt = formatPseudoBigInt;
    /**
     * Determines whether the given `SwitchStatement`'s `case` clauses cover every possible value of the switched expression.
     * The logic is the same as TypeScript's control flow analysis.
     * This does **not** check whether all `case` clauses do a certain action like assign a variable or return a value.
     * This function ignores the `default` clause if present.
     */
    function hasExhaustiveCaseClauses(node, checker) {
        const caseClauses = node.caseBlock.clauses.filter(node_1.isCaseClause);
        if (caseClauses.length === 0)
            return false;
        const typeParts = type_1.unionTypeParts(checker.getTypeAtLocation(node.expression));
        if (typeParts.length > caseClauses.length)
            return false;
        const types = new Set(typeParts.map(getPrimitiveLiteralFromType));
        if (types.has(undefined))
            return false;
        const seen = new Set();
        for (const clause of caseClauses) {
            const expressionType = checker.getTypeAtLocation(clause.expression);
            if (exports.isTypeFlagSet(expressionType, ts.TypeFlags.Never))
                continue; // additional case clause with 'never' is always allowed
            const type = getPrimitiveLiteralFromType(expressionType);
            if (types.has(type)) {
                seen.add(type);
            }
            else if (type !== 'null' && type !== 'undefined') { // additional case clauses with 'null' and 'undefined' are always allowed
                return false;
            }
        }
        return types.size === seen.size;
    }
    exports.hasExhaustiveCaseClauses = hasExhaustiveCaseClauses;
    function getPrimitiveLiteralFromType(t) {
        if (exports.isTypeFlagSet(t, ts.TypeFlags.Null))
            return 'null';
        if (exports.isTypeFlagSet(t, ts.TypeFlags.Undefined))
            return 'undefined';
        if (exports.isTypeFlagSet(t, ts.TypeFlags.NumberLiteral))
            return `${exports.isTypeFlagSet(t, ts.TypeFlags.EnumLiteral) ? 'enum:' : ''}${t.value}`;
        if (exports.isTypeFlagSet(t, ts.TypeFlags.StringLiteral))
            return `${exports.isTypeFlagSet(t, ts.TypeFlags.EnumLiteral) ? 'enum:' : ''}string:${t.value}`;
        if (exports.isTypeFlagSet(t, ts.TypeFlags.BigIntLiteral))
            return formatPseudoBigInt(t.value);
        if (_3_2_1.isUniqueESSymbolType(t))
            return t.escapedName;
        if (type_1.isBooleanLiteralType(t, true))
            return 'true';
        if (type_1.isBooleanLiteralType(t, false))
            return 'false';
    }
    function getBaseOfClassLikeExpression(node) {
        var _a;
        if (((_a = node.heritageClauses) === null || _a === void 0 ? void 0 : _a[0].token) === ts.SyntaxKind.ExtendsKeyword)
            return node.heritageClauses[0].types[0];
    }
    exports.getBaseOfClassLikeExpression = getBaseOfClassLikeExpression;
},
/* tsutils/typeguard/3.2/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const tslib_1 = require(58) /* tslib */;
    tslib_1.__exportStar(require(919) /* ./node */, exports);
    tslib_1.__exportStar(require(924) /* ./type */, exports);
},
/* tsutils/util/type.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ts = require(21) /* typescript */;
    const type_1 = require(923) /* ../typeguard/type */;
    const util_1 = require(929) /* ./util */;
    const node_1 = require(918) /* ../typeguard/node */;
    function isEmptyObjectType(type) {
        if (type_1.isObjectType(type) &&
            type.objectFlags & ts.ObjectFlags.Anonymous &&
            type.getProperties().length === 0 &&
            type.getCallSignatures().length === 0 &&
            type.getConstructSignatures().length === 0 &&
            type.getStringIndexType() === undefined &&
            type.getNumberIndexType() === undefined) {
            const baseTypes = type.getBaseTypes();
            return baseTypes === undefined || baseTypes.every(isEmptyObjectType);
        }
        return false;
    }
    exports.isEmptyObjectType = isEmptyObjectType;
    function removeOptionalityFromType(checker, type) {
        if (!containsTypeWithFlag(type, ts.TypeFlags.Undefined))
            return type;
        const allowsNull = containsTypeWithFlag(type, ts.TypeFlags.Null);
        type = checker.getNonNullableType(type);
        return allowsNull ? checker.getNullableType(type, ts.TypeFlags.Null) : type;
    }
    exports.removeOptionalityFromType = removeOptionalityFromType;
    function containsTypeWithFlag(type, flag) {
        for (const t of unionTypeParts(type))
            if (util_1.isTypeFlagSet(t, flag))
                return true;
        return false;
    }
    function removeOptionalChainingUndefinedMarkerType(checker, type) {
        if (!type_1.isUnionType(type))
            return isOptionalChainingUndefinedMarkerType(checker, type) ? type.getNonNullableType() : type;
        let flags = 0;
        let containsUndefinedMarker = false;
        for (const t of type.types) {
            if (isOptionalChainingUndefinedMarkerType(checker, t)) {
                containsUndefinedMarker = true;
            }
            else {
                flags |= t.flags;
            }
        }
        return containsUndefinedMarker
            ? checker.getNullableType(type.getNonNullableType(), flags)
            : type;
    }
    exports.removeOptionalChainingUndefinedMarkerType = removeOptionalChainingUndefinedMarkerType;
    function isOptionalChainingUndefinedMarkerType(checker, t) {
        return util_1.isTypeFlagSet(t, ts.TypeFlags.Undefined) && checker.getNullableType(t.getNonNullableType(), ts.TypeFlags.Undefined) !== t;
    }
    exports.isOptionalChainingUndefinedMarkerType = isOptionalChainingUndefinedMarkerType;
    function isTypeAssignableToNumber(checker, type) {
        return isTypeAssignableTo(checker, type, ts.TypeFlags.NumberLike);
    }
    exports.isTypeAssignableToNumber = isTypeAssignableToNumber;
    function isTypeAssignableToString(checker, type) {
        return isTypeAssignableTo(checker, type, ts.TypeFlags.StringLike);
    }
    exports.isTypeAssignableToString = isTypeAssignableToString;
    function isTypeAssignableTo(checker, type, flags) {
        flags |= ts.TypeFlags.Any;
        let typeParametersSeen;
        return (function check(t) {
            if (type_1.isTypeParameter(t) && t.symbol !== undefined && t.symbol.declarations !== undefined) {
                if (typeParametersSeen === undefined) {
                    typeParametersSeen = new Set([t]);
                }
                else if (!typeParametersSeen.has(t)) {
                    typeParametersSeen.add(t);
                }
                else {
                    return false;
                }
                const declaration = t.symbol.declarations[0];
                if (declaration.constraint === undefined)
                    return true; // TODO really?
                return check(checker.getTypeFromTypeNode(declaration.constraint));
            }
            if (type_1.isUnionType(t))
                return t.types.every(check);
            if (type_1.isIntersectionType(t))
                return t.types.some(check);
            return util_1.isTypeFlagSet(t, flags);
        })(type);
    }
    function getCallSignaturesOfType(type) {
        if (type_1.isUnionType(type)) {
            const signatures = [];
            for (const t of type.types)
                signatures.push(...getCallSignaturesOfType(t));
            return signatures;
        }
        if (type_1.isIntersectionType(type)) {
            let signatures;
            for (const t of type.types) {
                const sig = getCallSignaturesOfType(t);
                if (sig.length !== 0) {
                    if (signatures !== undefined)
                        return []; // if more than one type of the intersection has call signatures, none of them is useful for inference
                    signatures = sig;
                }
            }
            return signatures === undefined ? [] : signatures;
        }
        return type.getCallSignatures();
    }
    exports.getCallSignaturesOfType = getCallSignaturesOfType;
    /** Returns all types of a union type or an array containing `type` itself if it's no union type. */
    function unionTypeParts(type) {
        return type_1.isUnionType(type) ? type.types : [type];
    }
    exports.unionTypeParts = unionTypeParts;
    /** Returns all types of a intersection type or an array containing `type` itself if it's no intersection type. */
    function intersectionTypeParts(type) {
        return type_1.isIntersectionType(type) ? type.types : [type];
    }
    exports.intersectionTypeParts = intersectionTypeParts;
    function someTypePart(type, predicate, cb) {
        return predicate(type) ? type.types.some(cb) : cb(type);
    }
    exports.someTypePart = someTypePart;
    function isThenableType(checker, node, type = checker.getTypeAtLocation(node)) {
        for (const ty of unionTypeParts(checker.getApparentType(type))) {
            const then = ty.getProperty('then');
            if (then === undefined)
                continue;
            const thenType = checker.getTypeOfSymbolAtLocation(then, node);
            for (const t of unionTypeParts(thenType))
                for (const signature of t.getCallSignatures())
                    if (signature.parameters.length !== 0 && isCallback(checker, signature.parameters[0], node))
                        return true;
        }
        return false;
    }
    exports.isThenableType = isThenableType;
    function isCallback(checker, param, node) {
        let type = checker.getApparentType(checker.getTypeOfSymbolAtLocation(param, node));
        if (param.valueDeclaration.dotDotDotToken) {
            // unwrap array type of rest parameter
            type = type.getNumberIndexType();
            if (type === undefined)
                return false;
        }
        for (const t of unionTypeParts(type))
            if (t.getCallSignatures().length !== 0)
                return true;
        return false;
    }
    /** Determine if a type is definitely falsy. This function doesn't unwrap union types. */
    function isFalsyType(type) {
        if (type.flags & (ts.TypeFlags.Undefined | ts.TypeFlags.Null | ts.TypeFlags.Void))
            return true;
        if (type_1.isLiteralType(type))
            return !type.value;
        return isBooleanLiteralType(type, false);
    }
    exports.isFalsyType = isFalsyType;
    /** Determines whether the given type is a boolean literal type and matches the given boolean literal (true or false). */
    function isBooleanLiteralType(type, literal) {
        return util_1.isTypeFlagSet(type, ts.TypeFlags.BooleanLiteral) &&
            type.intrinsicName === (literal ? 'true' : 'false');
    }
    exports.isBooleanLiteralType = isBooleanLiteralType;
    function getPropertyOfType(type, name) {
        if (!name.startsWith('__'))
            return type.getProperty(name);
        return type.getProperties().find((s) => s.escapedName === name);
    }
    exports.getPropertyOfType = getPropertyOfType;
    function getWellKnownSymbolPropertyOfType(type, wellKnownSymbolName, checker) {
        const prefix = '__@' + wellKnownSymbolName;
        for (const prop of type.getProperties()) {
            if (!prop.name.startsWith(prefix))
                continue;
            const globalSymbol = checker.getApparentType(checker.getTypeAtLocation(prop.valueDeclaration.name.expression)).symbol;
            if (prop.escapedName === getPropertyNameOfWellKnownSymbol(checker, globalSymbol, wellKnownSymbolName))
                return prop;
        }
        return;
    }
    exports.getWellKnownSymbolPropertyOfType = getWellKnownSymbolPropertyOfType;
    function getPropertyNameOfWellKnownSymbol(checker, symbolConstructor, symbolName) {
        const knownSymbol = symbolConstructor &&
            checker.getTypeOfSymbolAtLocation(symbolConstructor, symbolConstructor.valueDeclaration).getProperty(symbolName);
        const knownSymbolType = knownSymbol && checker.getTypeOfSymbolAtLocation(knownSymbol, knownSymbol.valueDeclaration);
        if (knownSymbolType && type_1.isUniqueESSymbolType(knownSymbolType))
            return knownSymbolType.escapedName;
        return ('__@' + symbolName);
    }
    /** Determines if writing to a certain property of a given type is allowed. */
    function isPropertyReadonlyInType(type, name, checker) {
        let seenProperty = false;
        let seenReadonlySignature = false;
        for (const t of unionTypeParts(type)) {
            if (getPropertyOfType(t, name) === undefined) {
                // property is not present in this part of the union -> check for readonly index signature
                const index = (util_1.isNumericPropertyName(name) ? checker.getIndexInfoOfType(t, ts.IndexKind.Number) : undefined) ||
                    checker.getIndexInfoOfType(t, ts.IndexKind.String);
                if (index !== undefined && index.isReadonly) {
                    if (seenProperty)
                        return true;
                    seenReadonlySignature = true;
                }
            }
            else if (seenReadonlySignature || isReadonlyPropertyIntersection(t, name, checker)) {
                return true;
            }
            else {
                seenProperty = true;
            }
        }
        return false;
    }
    exports.isPropertyReadonlyInType = isPropertyReadonlyInType;
    function isReadonlyPropertyIntersection(type, name, checker) {
        return someTypePart(type, type_1.isIntersectionType, (t) => {
            const prop = getPropertyOfType(t, name);
            if (prop === undefined)
                return false;
            if (prop.flags & ts.SymbolFlags.Transient) {
                if (/^(?:[1-9]\d*|0)$/.test(name) && type_1.isTupleTypeReference(t))
                    return t.target.readonly;
                switch (isReadonlyPropertyFromMappedType(t, name, checker)) {
                    case true:
                        return true;
                    case false:
                        return false;
                    default:
                    // `undefined` falls through
                }
            }
            return (
            // members of namespace import
            util_1.isSymbolFlagSet(prop, ts.SymbolFlags.ValueModule) ||
                // we unwrapped every mapped type, now we can check the actual declarations
                symbolHasReadonlyDeclaration(prop, checker));
        });
    }
    function isReadonlyPropertyFromMappedType(type, name, checker) {
        if (!type_1.isObjectType(type) || !util_1.isObjectFlagSet(type, ts.ObjectFlags.Mapped))
            return;
        const declaration = type.symbol.declarations[0];
        // well-known symbols are not affected by mapped types
        if (declaration.readonlyToken !== undefined && !/^__@[^@]+$/.test(name))
            return declaration.readonlyToken.kind !== ts.SyntaxKind.MinusToken;
        return isPropertyReadonlyInType(type.modifiersType, name, checker);
    }
    function symbolHasReadonlyDeclaration(symbol, checker) {
        return (symbol.flags & ts.SymbolFlags.Accessor) === ts.SymbolFlags.GetAccessor ||
            symbol.declarations !== undefined &&
                symbol.declarations.some((node) => util_1.isModifierFlagSet(node, ts.ModifierFlags.Readonly) ||
                    node_1.isVariableDeclaration(node) && util_1.isNodeFlagSet(node.parent, ts.NodeFlags.Const) ||
                    node_1.isCallExpression(node) && util_1.isReadonlyAssignmentDeclaration(node, checker) ||
                    node_1.isEnumMember(node) ||
                    (node_1.isPropertyAssignment(node) || node_1.isShorthandPropertyAssignment(node)) && util_1.isInConstContext(node.parent));
    }
    exports.symbolHasReadonlyDeclaration = symbolHasReadonlyDeclaration;
    /** Returns the the literal name or unique symbol name from a given type. Doesn't unwrap union types. */
    function getPropertyNameFromType(type) {
        // string or number literal. bigint is intentionally excluded
        if (type.flags & (ts.TypeFlags.StringLiteral | ts.TypeFlags.NumberLiteral)) {
            const value = String(type.value);
            return { displayName: value, symbolName: ts.escapeLeadingUnderscores(value) };
        }
        if (type_1.isUniqueESSymbolType(type))
            return {
                displayName: `[${type.symbol
                    ? `${isKnownSymbol(type.symbol) ? 'Symbol.' : ''}${type.symbol.name}`
                    : type.escapedName.replace(/^__@|@\d+$/g, '')}]`,
                symbolName: type.escapedName,
            };
    }
    exports.getPropertyNameFromType = getPropertyNameFromType;
    function isKnownSymbol(symbol) {
        return util_1.isSymbolFlagSet(symbol, ts.SymbolFlags.Property) &&
            symbol.valueDeclaration !== undefined &&
            node_1.isInterfaceDeclaration(symbol.valueDeclaration.parent) &&
            symbol.valueDeclaration.parent.name.text === 'SymbolConstructor' &&
            isGlobalDeclaration(symbol.valueDeclaration.parent);
    }
    function isGlobalDeclaration(node) {
        return util_1.isNodeFlagSet(node.parent, ts.NodeFlags.GlobalAugmentation) || node_1.isSourceFile(node.parent) && !ts.isExternalModule(node.parent);
    }
    function getSymbolOfClassLikeDeclaration(node, checker) {
        var _a;
        return checker.getSymbolAtLocation((_a = node.name) !== null && _a !== void 0 ? _a : util_1.getChildOfKind(node, ts.SyntaxKind.ClassKeyword));
    }
    exports.getSymbolOfClassLikeDeclaration = getSymbolOfClassLikeDeclaration;
    function getConstructorTypeOfClassLikeDeclaration(node, checker) {
        return node.kind === ts.SyntaxKind.ClassExpression
            ? checker.getTypeAtLocation(node)
            : checker.getTypeOfSymbolAtLocation(getSymbolOfClassLikeDeclaration(node, checker), node);
    }
    exports.getConstructorTypeOfClassLikeDeclaration = getConstructorTypeOfClassLikeDeclaration;
    function getInstanceTypeOfClassLikeDeclaration(node, checker) {
        return node.kind === ts.SyntaxKind.ClassDeclaration
            ? checker.getTypeAtLocation(node)
            : checker.getDeclaredTypeOfSymbol(getSymbolOfClassLikeDeclaration(node, checker));
    }
    exports.getInstanceTypeOfClassLikeDeclaration = getInstanceTypeOfClassLikeDeclaration;
    function getIteratorYieldResultFromIteratorResult(type, node, checker) {
        return type_1.isUnionType(type) && type.types.find((t) => {
            const done = t.getProperty('done');
            return done !== undefined &&
                isBooleanLiteralType(removeOptionalityFromType(checker, checker.getTypeOfSymbolAtLocation(done, node)), false);
        }) || type;
    }
    exports.getIteratorYieldResultFromIteratorResult = getIteratorYieldResultFromIteratorResult;
    /** Lookup the declaration of a class member in the super class. */
    function getBaseClassMemberOfClassElement(node, checker) {
        if (!node_1.isClassLikeDeclaration(node.parent))
            return;
        const base = util_1.getBaseOfClassLikeExpression(node.parent);
        if (base === undefined)
            return;
        const name = util_1.getSingleLateBoundPropertyNameOfPropertyName(node.name, checker);
        if (name === undefined)
            return;
        const baseType = checker.getTypeAtLocation(util_1.hasModifier(node.modifiers, ts.SyntaxKind.StaticKeyword)
            ? base.expression
            : base);
        return getPropertyOfType(baseType, name.symbolName);
    }
    exports.getBaseClassMemberOfClassElement = getBaseClassMemberOfClassElement;
},
/* tsutils/util/usage.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const util_1 = require(929) /* ./util */;
    const ts = require(21) /* typescript */;
    var DeclarationDomain;
    (function (DeclarationDomain) {
        DeclarationDomain[DeclarationDomain["Namespace"] = 1] = "Namespace";
        DeclarationDomain[DeclarationDomain["Type"] = 2] = "Type";
        DeclarationDomain[DeclarationDomain["Value"] = 4] = "Value";
        DeclarationDomain[DeclarationDomain["Import"] = 8] = "Import";
        DeclarationDomain[DeclarationDomain["Any"] = 7] = "Any";
    })(DeclarationDomain = exports.DeclarationDomain || (exports.DeclarationDomain = {}));
    var UsageDomain;
    (function (UsageDomain) {
        UsageDomain[UsageDomain["Namespace"] = 1] = "Namespace";
        UsageDomain[UsageDomain["Type"] = 2] = "Type";
        UsageDomain[UsageDomain["Value"] = 4] = "Value";
        UsageDomain[UsageDomain["ValueOrNamespace"] = 5] = "ValueOrNamespace";
        UsageDomain[UsageDomain["Any"] = 7] = "Any";
        UsageDomain[UsageDomain["TypeQuery"] = 8] = "TypeQuery";
    })(UsageDomain = exports.UsageDomain || (exports.UsageDomain = {}));
    // TODO handle cases where values are used only for their types, e.g. `declare [propSymbol]: number`
    function getUsageDomain(node) {
        const parent = node.parent;
        switch (parent.kind) {
            case ts.SyntaxKind.TypeReference:
                return node.originalKeywordKind !== ts.SyntaxKind.ConstKeyword ? 2 /* Type */ : undefined;
            case ts.SyntaxKind.ExpressionWithTypeArguments:
                return parent.parent.token === ts.SyntaxKind.ImplementsKeyword ||
                    parent.parent.parent.kind === ts.SyntaxKind.InterfaceDeclaration
                    ? 2 /* Type */
                    : 4 /* Value */;
            case ts.SyntaxKind.TypeQuery:
                return 5 /* ValueOrNamespace */ | 8 /* TypeQuery */;
            case ts.SyntaxKind.QualifiedName:
                if (parent.left === node) {
                    if (getEntityNameParent(parent).kind === ts.SyntaxKind.TypeQuery)
                        return 1 /* Namespace */ | 8 /* TypeQuery */;
                    return 1 /* Namespace */;
                }
                break;
            case ts.SyntaxKind.ExportSpecifier:
                // either {name} or {propertyName as name}
                if (parent.propertyName === undefined ||
                    parent.propertyName === node)
                    return 7 /* Any */; // TODO handle type-only exports
                break;
            case ts.SyntaxKind.ExportAssignment:
                return 7 /* Any */;
            // Value
            case ts.SyntaxKind.BindingElement:
                if (parent.initializer === node)
                    return 5 /* ValueOrNamespace */;
                break;
            case ts.SyntaxKind.Parameter:
            case ts.SyntaxKind.EnumMember:
            case ts.SyntaxKind.PropertyDeclaration:
            case ts.SyntaxKind.VariableDeclaration:
            case ts.SyntaxKind.PropertyAssignment:
            case ts.SyntaxKind.PropertyAccessExpression:
            case ts.SyntaxKind.ImportEqualsDeclaration:
                if (parent.name !== node)
                    return 5 /* ValueOrNamespace */; // TODO handle type-only imports
                break;
            case ts.SyntaxKind.JsxAttribute:
            case ts.SyntaxKind.FunctionDeclaration:
            case ts.SyntaxKind.FunctionExpression:
            case ts.SyntaxKind.NamespaceImport:
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.ClassExpression:
            case ts.SyntaxKind.ModuleDeclaration:
            case ts.SyntaxKind.MethodDeclaration:
            case ts.SyntaxKind.EnumDeclaration:
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.SetAccessor:
            case ts.SyntaxKind.LabeledStatement:
            case ts.SyntaxKind.BreakStatement:
            case ts.SyntaxKind.ContinueStatement:
            case ts.SyntaxKind.ImportClause:
            case ts.SyntaxKind.ImportSpecifier:
            case ts.SyntaxKind.TypePredicate: // TODO this actually references a parameter
            case ts.SyntaxKind.MethodSignature:
            case ts.SyntaxKind.PropertySignature:
            case ts.SyntaxKind.NamespaceExportDeclaration:
            case ts.SyntaxKind.NamespaceExport:
            case ts.SyntaxKind.InterfaceDeclaration:
            case ts.SyntaxKind.TypeAliasDeclaration:
            case ts.SyntaxKind.TypeParameter:
            case ts.SyntaxKind.NamedTupleMember:
                break;
            default:
                return 5 /* ValueOrNamespace */;
        }
    }
    exports.getUsageDomain = getUsageDomain;
    function getDeclarationDomain(node) {
        switch (node.parent.kind) {
            case ts.SyntaxKind.TypeParameter:
            case ts.SyntaxKind.InterfaceDeclaration:
            case ts.SyntaxKind.TypeAliasDeclaration:
                return 2 /* Type */;
            case ts.SyntaxKind.ClassDeclaration:
            case ts.SyntaxKind.ClassExpression:
                return 2 /* Type */ | 4 /* Value */;
            case ts.SyntaxKind.EnumDeclaration:
                return 7 /* Any */;
            case ts.SyntaxKind.NamespaceImport:
            case ts.SyntaxKind.ImportClause:
                return 7 /* Any */ | 8 /* Import */; // TODO handle type-only imports
            case ts.SyntaxKind.ImportEqualsDeclaration:
            case ts.SyntaxKind.ImportSpecifier:
                return node.parent.name === node
                    ? 7 /* Any */ | 8 /* Import */ // TODO handle type-only imports
                    : undefined;
            case ts.SyntaxKind.ModuleDeclaration:
                return 1 /* Namespace */;
            case ts.SyntaxKind.Parameter:
                if (node.parent.parent.kind === ts.SyntaxKind.IndexSignature || node.originalKeywordKind === ts.SyntaxKind.ThisKeyword)
                    return;
            // falls through
            case ts.SyntaxKind.BindingElement:
            case ts.SyntaxKind.VariableDeclaration:
                return node.parent.name === node ? 4 /* Value */ : undefined;
            case ts.SyntaxKind.FunctionDeclaration:
            case ts.SyntaxKind.FunctionExpression:
                return 4 /* Value */;
        }
    }
    exports.getDeclarationDomain = getDeclarationDomain;
    function collectVariableUsage(sourceFile) {
        return new UsageWalker().getUsage(sourceFile);
    }
    exports.collectVariableUsage = collectVariableUsage;
    class AbstractScope {
        constructor(_global) {
            this._global = _global;
            this._variables = new Map();
            this._uses = [];
            this._namespaceScopes = undefined;
            this._enumScopes = undefined;
        }
        addVariable(identifier, name, selector, exported, domain) {
            const variables = this.getDestinationScope(selector).getVariables();
            const declaration = {
                domain,
                exported,
                declaration: name,
            };
            const variable = variables.get(identifier);
            if (variable === undefined) {
                variables.set(identifier, {
                    domain,
                    declarations: [declaration],
                    uses: [],
                });
            }
            else {
                variable.domain |= domain;
                variable.declarations.push(declaration);
            }
        }
        addUse(use) {
            this._uses.push(use);
        }
        getVariables() {
            return this._variables;
        }
        getFunctionScope() {
            return this;
        }
        end(cb) {
            if (this._namespaceScopes !== undefined)
                this._namespaceScopes.forEach((value) => value.finish(cb));
            this._namespaceScopes = this._enumScopes = undefined;
            this._applyUses();
            this._variables.forEach((variable) => {
                for (const declaration of variable.declarations) {
                    const result = {
                        declarations: [],
                        domain: declaration.domain,
                        exported: declaration.exported,
                        inGlobalScope: this._global,
                        uses: [],
                    };
                    for (const other of variable.declarations)
                        if (other.domain & declaration.domain)
                            result.declarations.push(other.declaration);
                    for (const use of variable.uses)
                        if (use.domain & declaration.domain)
                            result.uses.push(use);
                    cb(result, declaration.declaration, this);
                }
            });
        }
        // tslint:disable-next-line:prefer-function-over-method
        markExported(_name) { } // only relevant for the root scope
        createOrReuseNamespaceScope(name, _exported, ambient, hasExportStatement) {
            let scope;
            if (this._namespaceScopes === undefined) {
                this._namespaceScopes = new Map();
            }
            else {
                scope = this._namespaceScopes.get(name);
            }
            if (scope === undefined) {
                scope = new NamespaceScope(ambient, hasExportStatement, this);
                this._namespaceScopes.set(name, scope);
            }
            else {
                scope.refresh(ambient, hasExportStatement);
            }
            return scope;
        }
        createOrReuseEnumScope(name, _exported) {
            let scope;
            if (this._enumScopes === undefined) {
                this._enumScopes = new Map();
            }
            else {
                scope = this._enumScopes.get(name);
            }
            if (scope === undefined) {
                scope = new EnumScope(this);
                this._enumScopes.set(name, scope);
            }
            return scope;
        }
        _applyUses() {
            for (const use of this._uses)
                if (!this._applyUse(use))
                    this._addUseToParent(use);
            this._uses = [];
        }
        _applyUse(use, variables = this._variables) {
            const variable = variables.get(use.location.text);
            if (variable === undefined || (variable.domain & use.domain) === 0)
                return false;
            variable.uses.push(use);
            return true;
        }
        _addUseToParent(_use) { } // tslint:disable-line:prefer-function-over-method
    }
    class RootScope extends AbstractScope {
        constructor(_exportAll, global) {
            super(global);
            this._exportAll = _exportAll;
            this._exports = undefined;
            this._innerScope = new NonRootScope(this, 1 /* Function */);
        }
        addVariable(identifier, name, selector, exported, domain) {
            if (domain & 8 /* Import */)
                return super.addVariable(identifier, name, selector, exported, domain);
            return this._innerScope.addVariable(identifier, name, selector, exported, domain);
        }
        addUse(use, origin) {
            if (origin === this._innerScope)
                return super.addUse(use);
            return this._innerScope.addUse(use);
        }
        markExported(id) {
            if (this._exports === undefined) {
                this._exports = [id.text];
            }
            else {
                this._exports.push(id.text);
            }
        }
        end(cb) {
            this._innerScope.end((value, key) => {
                value.exported = value.exported || this._exportAll
                    || this._exports !== undefined && this._exports.includes(key.text);
                value.inGlobalScope = this._global;
                return cb(value, key, this);
            });
            return super.end((value, key, scope) => {
                value.exported = value.exported || scope === this
                    && this._exports !== undefined && this._exports.includes(key.text);
                return cb(value, key, scope);
            });
        }
        getDestinationScope() {
            return this;
        }
    }
    class NonRootScope extends AbstractScope {
        constructor(_parent, _boundary) {
            super(false);
            this._parent = _parent;
            this._boundary = _boundary;
        }
        _addUseToParent(use) {
            return this._parent.addUse(use, this);
        }
        getDestinationScope(selector) {
            return this._boundary & selector
                ? this
                : this._parent.getDestinationScope(selector);
        }
    }
    class EnumScope extends NonRootScope {
        constructor(parent) {
            super(parent, 1 /* Function */);
        }
        end() {
            this._applyUses();
        }
    }
    class ConditionalTypeScope extends NonRootScope {
        constructor(parent) {
            super(parent, 8 /* ConditionalType */);
            this._state = 0 /* Initial */;
        }
        updateState(newState) {
            this._state = newState;
        }
        addUse(use) {
            if (this._state === 2 /* TrueType */)
                return void this._uses.push(use);
            return this._parent.addUse(use, this);
        }
    }
    class FunctionScope extends NonRootScope {
        constructor(parent) {
            super(parent, 1 /* Function */);
        }
        beginBody() {
            this._applyUses();
        }
    }
    class AbstractNamedExpressionScope extends NonRootScope {
        constructor(_name, _domain, parent) {
            super(parent, 1 /* Function */);
            this._name = _name;
            this._domain = _domain;
        }
        end(cb) {
            this._innerScope.end(cb);
            return cb({
                declarations: [this._name],
                domain: this._domain,
                exported: false,
                uses: this._uses,
                inGlobalScope: false,
            }, this._name, this);
        }
        addUse(use, source) {
            if (source !== this._innerScope)
                return this._innerScope.addUse(use);
            if (use.domain & this._domain && use.location.text === this._name.text) {
                this._uses.push(use);
            }
            else {
                return this._parent.addUse(use, this);
            }
        }
        getFunctionScope() {
            return this._innerScope;
        }
        getDestinationScope() {
            return this._innerScope;
        }
    }
    class FunctionExpressionScope extends AbstractNamedExpressionScope {
        constructor(name, parent) {
            super(name, 4 /* Value */, parent);
            this._innerScope = new FunctionScope(this);
        }
        beginBody() {
            return this._innerScope.beginBody();
        }
    }
    class ClassExpressionScope extends AbstractNamedExpressionScope {
        constructor(name, parent) {
            super(name, 4 /* Value */ | 2 /* Type */, parent);
            this._innerScope = new NonRootScope(this, 1 /* Function */);
        }
    }
    class BlockScope extends NonRootScope {
        constructor(_functionScope, parent) {
            super(parent, 2 /* Block */);
            this._functionScope = _functionScope;
        }
        getFunctionScope() {
            return this._functionScope;
        }
    }
    function mapDeclaration(declaration) {
        return {
            declaration,
            exported: true,
            domain: getDeclarationDomain(declaration),
        };
    }
    class NamespaceScope extends NonRootScope {
        constructor(_ambient, _hasExport, parent) {
            super(parent, 1 /* Function */);
            this._ambient = _ambient;
            this._hasExport = _hasExport;
            this._innerScope = new NonRootScope(this, 1 /* Function */);
            this._exports = undefined;
        }
        finish(cb) {
            return super.end(cb);
        }
        end(cb) {
            this._innerScope.end((variable, key, scope) => {
                if (scope !== this._innerScope ||
                    !variable.exported && (!this._ambient || this._exports !== undefined && !this._exports.has(key.text)))
                    return cb(variable, key, scope);
                const namespaceVar = this._variables.get(key.text);
                if (namespaceVar === undefined) {
                    this._variables.set(key.text, {
                        declarations: variable.declarations.map(mapDeclaration),
                        domain: variable.domain,
                        uses: [...variable.uses],
                    });
                }
                else {
                    outer: for (const declaration of variable.declarations) {
                        for (const existing of namespaceVar.declarations)
                            if (existing.declaration === declaration)
                                continue outer;
                        namespaceVar.declarations.push(mapDeclaration(declaration));
                    }
                    namespaceVar.domain |= variable.domain;
                    for (const use of variable.uses) {
                        if (namespaceVar.uses.includes(use))
                            continue;
                        namespaceVar.uses.push(use);
                    }
                }
            });
            this._applyUses();
            this._innerScope = new NonRootScope(this, 1 /* Function */);
        }
        createOrReuseNamespaceScope(name, exported, ambient, hasExportStatement) {
            if (!exported && (!this._ambient || this._hasExport))
                return this._innerScope.createOrReuseNamespaceScope(name, exported, ambient || this._ambient, hasExportStatement);
            return super.createOrReuseNamespaceScope(name, exported, ambient || this._ambient, hasExportStatement);
        }
        createOrReuseEnumScope(name, exported) {
            if (!exported && (!this._ambient || this._hasExport))
                return this._innerScope.createOrReuseEnumScope(name, exported);
            return super.createOrReuseEnumScope(name, exported);
        }
        addUse(use, source) {
            if (source !== this._innerScope)
                return this._innerScope.addUse(use);
            this._uses.push(use);
        }
        refresh(ambient, hasExport) {
            this._ambient = ambient;
            this._hasExport = hasExport;
        }
        markExported(name, _as) {
            if (this._exports === undefined)
                this._exports = new Set();
            this._exports.add(name.text);
        }
        getDestinationScope() {
            return this._innerScope;
        }
    }
    function getEntityNameParent(name) {
        let parent = name.parent;
        while (parent.kind === ts.SyntaxKind.QualifiedName)
            parent = parent.parent;
        return parent;
    }
    // TODO class decorators resolve outside of class, element and parameter decorator resolve inside/at the class
    // TODO computed property name resolves inside/at the cass
    // TODO this and super in all of them are resolved outside of the class
    class UsageWalker {
        constructor() {
            this._result = new Map();
        }
        getUsage(sourceFile) {
            const variableCallback = (variable, key) => {
                this._result.set(key, variable);
            };
            const isModule = ts.isExternalModule(sourceFile);
            this._scope = new RootScope(sourceFile.isDeclarationFile && isModule && !containsExportStatement(sourceFile), !isModule);
            const cb = (node) => {
                if (util_1.isBlockScopeBoundary(node))
                    return continueWithScope(node, new BlockScope(this._scope.getFunctionScope(), this._scope), handleBlockScope);
                switch (node.kind) {
                    case ts.SyntaxKind.ClassExpression:
                        return continueWithScope(node, node.name !== undefined
                            ? new ClassExpressionScope(node.name, this._scope)
                            : new NonRootScope(this._scope, 1 /* Function */));
                    case ts.SyntaxKind.ClassDeclaration:
                        this._handleDeclaration(node, true, 4 /* Value */ | 2 /* Type */);
                        return continueWithScope(node, new NonRootScope(this._scope, 1 /* Function */));
                    case ts.SyntaxKind.InterfaceDeclaration:
                    case ts.SyntaxKind.TypeAliasDeclaration:
                        this._handleDeclaration(node, true, 2 /* Type */);
                        return continueWithScope(node, new NonRootScope(this._scope, 4 /* Type */));
                    case ts.SyntaxKind.EnumDeclaration:
                        this._handleDeclaration(node, true, 7 /* Any */);
                        return continueWithScope(node, this._scope.createOrReuseEnumScope(node.name.text, util_1.hasModifier(node.modifiers, ts.SyntaxKind.ExportKeyword)));
                    case ts.SyntaxKind.ModuleDeclaration:
                        return this._handleModule(node, continueWithScope);
                    case ts.SyntaxKind.MappedType:
                        return continueWithScope(node, new NonRootScope(this._scope, 4 /* Type */));
                    case ts.SyntaxKind.FunctionExpression:
                    case ts.SyntaxKind.ArrowFunction:
                    case ts.SyntaxKind.Constructor:
                    case ts.SyntaxKind.MethodDeclaration:
                    case ts.SyntaxKind.FunctionDeclaration:
                    case ts.SyntaxKind.GetAccessor:
                    case ts.SyntaxKind.SetAccessor:
                    case ts.SyntaxKind.MethodSignature:
                    case ts.SyntaxKind.CallSignature:
                    case ts.SyntaxKind.ConstructSignature:
                    case ts.SyntaxKind.ConstructorType:
                    case ts.SyntaxKind.FunctionType:
                        return this._handleFunctionLikeDeclaration(node, cb, variableCallback);
                    case ts.SyntaxKind.ConditionalType:
                        return this._handleConditionalType(node, cb, variableCallback);
                    // End of Scope specific handling
                    case ts.SyntaxKind.VariableDeclarationList:
                        this._handleVariableDeclaration(node);
                        break;
                    case ts.SyntaxKind.Parameter:
                        if (node.parent.kind !== ts.SyntaxKind.IndexSignature &&
                            (node.name.kind !== ts.SyntaxKind.Identifier ||
                                node.name.originalKeywordKind !== ts.SyntaxKind.ThisKeyword))
                            this._handleBindingName(node.name, false, false);
                        break;
                    case ts.SyntaxKind.EnumMember:
                        this._scope.addVariable(util_1.getPropertyName(node.name), node.name, 1 /* Function */, true, 4 /* Value */);
                        break;
                    case ts.SyntaxKind.ImportClause:
                    case ts.SyntaxKind.ImportSpecifier:
                    case ts.SyntaxKind.NamespaceImport:
                    case ts.SyntaxKind.ImportEqualsDeclaration:
                        this._handleDeclaration(node, false, 7 /* Any */ | 8 /* Import */);
                        break;
                    case ts.SyntaxKind.TypeParameter:
                        this._scope.addVariable(node.name.text, node.name, node.parent.kind === ts.SyntaxKind.InferType ? 8 /* InferType */ : 7 /* Type */, false, 2 /* Type */);
                        break;
                    case ts.SyntaxKind.ExportSpecifier:
                        if (node.propertyName !== undefined)
                            return this._scope.markExported(node.propertyName, node.name);
                        return this._scope.markExported(node.name);
                    case ts.SyntaxKind.ExportAssignment:
                        if (node.expression.kind === ts.SyntaxKind.Identifier)
                            return this._scope.markExported(node.expression);
                        break;
                    case ts.SyntaxKind.Identifier:
                        const domain = getUsageDomain(node);
                        if (domain !== undefined)
                            this._scope.addUse({ domain, location: node });
                        return;
                }
                return ts.forEachChild(node, cb);
            };
            const continueWithScope = (node, scope, next = forEachChild) => {
                const savedScope = this._scope;
                this._scope = scope;
                next(node);
                this._scope.end(variableCallback);
                this._scope = savedScope;
            };
            const handleBlockScope = (node) => {
                if (node.kind === ts.SyntaxKind.CatchClause && node.variableDeclaration !== undefined)
                    this._handleBindingName(node.variableDeclaration.name, true, false);
                return ts.forEachChild(node, cb);
            };
            ts.forEachChild(sourceFile, cb);
            this._scope.end(variableCallback);
            return this._result;
            function forEachChild(node) {
                return ts.forEachChild(node, cb);
            }
        }
        _handleConditionalType(node, cb, varCb) {
            const savedScope = this._scope;
            const scope = this._scope = new ConditionalTypeScope(savedScope);
            cb(node.checkType);
            scope.updateState(1 /* Extends */);
            cb(node.extendsType);
            scope.updateState(2 /* TrueType */);
            cb(node.trueType);
            scope.updateState(3 /* FalseType */);
            cb(node.falseType);
            scope.end(varCb);
            this._scope = savedScope;
        }
        _handleFunctionLikeDeclaration(node, cb, varCb) {
            if (node.decorators !== undefined)
                node.decorators.forEach(cb);
            const savedScope = this._scope;
            if (node.kind === ts.SyntaxKind.FunctionDeclaration)
                this._handleDeclaration(node, false, 4 /* Value */);
            const scope = this._scope = node.kind === ts.SyntaxKind.FunctionExpression && node.name !== undefined
                ? new FunctionExpressionScope(node.name, savedScope)
                : new FunctionScope(savedScope);
            if (node.name !== undefined)
                cb(node.name);
            if (node.typeParameters !== undefined)
                node.typeParameters.forEach(cb);
            node.parameters.forEach(cb);
            if (node.type !== undefined)
                cb(node.type);
            if (node.body !== undefined) {
                scope.beginBody();
                cb(node.body);
            }
            scope.end(varCb);
            this._scope = savedScope;
        }
        _handleModule(node, next) {
            if (node.flags & ts.NodeFlags.GlobalAugmentation)
                return next(node, this._scope.createOrReuseNamespaceScope('-global', false, true, false));
            if (node.name.kind === ts.SyntaxKind.Identifier) {
                const exported = isNamespaceExported(node);
                this._scope.addVariable(node.name.text, node.name, 1 /* Function */, exported, 1 /* Namespace */ | 4 /* Value */);
                const ambient = util_1.hasModifier(node.modifiers, ts.SyntaxKind.DeclareKeyword);
                return next(node, this._scope.createOrReuseNamespaceScope(node.name.text, exported, ambient, ambient && namespaceHasExportStatement(node)));
            }
            return next(node, this._scope.createOrReuseNamespaceScope(`"${node.name.text}"`, false, true, namespaceHasExportStatement(node)));
        }
        _handleDeclaration(node, blockScoped, domain) {
            if (node.name !== undefined)
                this._scope.addVariable(node.name.text, node.name, blockScoped ? 3 /* Block */ : 1 /* Function */, util_1.hasModifier(node.modifiers, ts.SyntaxKind.ExportKeyword), domain);
        }
        _handleBindingName(name, blockScoped, exported) {
            if (name.kind === ts.SyntaxKind.Identifier)
                return this._scope.addVariable(name.text, name, blockScoped ? 3 /* Block */ : 1 /* Function */, exported, 4 /* Value */);
            util_1.forEachDestructuringIdentifier(name, (declaration) => {
                this._scope.addVariable(declaration.name.text, declaration.name, blockScoped ? 3 /* Block */ : 1 /* Function */, exported, 4 /* Value */);
            });
        }
        _handleVariableDeclaration(declarationList) {
            const blockScoped = util_1.isBlockScopedVariableDeclarationList(declarationList);
            const exported = declarationList.parent.kind === ts.SyntaxKind.VariableStatement &&
                util_1.hasModifier(declarationList.parent.modifiers, ts.SyntaxKind.ExportKeyword);
            for (const declaration of declarationList.declarations)
                this._handleBindingName(declaration.name, blockScoped, exported);
        }
    }
    function isNamespaceExported(node) {
        return node.parent.kind === ts.SyntaxKind.ModuleDeclaration || util_1.hasModifier(node.modifiers, ts.SyntaxKind.ExportKeyword);
    }
    function namespaceHasExportStatement(ns) {
        if (ns.body === undefined || ns.body.kind !== ts.SyntaxKind.ModuleBlock)
            return false;
        return containsExportStatement(ns.body);
    }
    function containsExportStatement(block) {
        for (const statement of block.statements)
            if (statement.kind === ts.SyntaxKind.ExportDeclaration || statement.kind === ts.SyntaxKind.ExportAssignment)
                return true;
        return false;
    }
},
/* tsutils/util/control-flow.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ts = require(21) /* typescript */;
    const node_1 = require(918) /* ../typeguard/node */;
    const util_1 = require(929) /* ./util */;
    function endsControlFlow(statement, checker) {
        return getControlFlowEnd(statement, checker).end;
    }
    exports.endsControlFlow = endsControlFlow;
    const defaultControlFlowEnd = { statements: [], end: false };
    function getControlFlowEnd(statement, checker) {
        return node_1.isBlockLike(statement) ? handleBlock(statement, checker) : getControlFlowEndWorker(statement, checker);
    }
    exports.getControlFlowEnd = getControlFlowEnd;
    function getControlFlowEndWorker(statement, checker) {
        switch (statement.kind) {
            case ts.SyntaxKind.ReturnStatement:
            case ts.SyntaxKind.ThrowStatement:
            case ts.SyntaxKind.ContinueStatement:
            case ts.SyntaxKind.BreakStatement:
                return { statements: [statement], end: true };
            case ts.SyntaxKind.Block:
                return handleBlock(statement, checker);
            case ts.SyntaxKind.ForStatement:
            case ts.SyntaxKind.WhileStatement:
                return handleForAndWhileStatement(statement, checker);
            case ts.SyntaxKind.ForOfStatement:
            case ts.SyntaxKind.ForInStatement:
                return handleForInOrOfStatement(statement, checker);
            case ts.SyntaxKind.DoStatement:
                return matchBreakOrContinue(getControlFlowEndWorker(statement.statement, checker), node_1.isBreakOrContinueStatement);
            case ts.SyntaxKind.IfStatement:
                return handleIfStatement(statement, checker);
            case ts.SyntaxKind.SwitchStatement:
                return matchBreakOrContinue(handleSwitchStatement(statement, checker), node_1.isBreakStatement);
            case ts.SyntaxKind.TryStatement:
                return handleTryStatement(statement, checker);
            case ts.SyntaxKind.LabeledStatement:
                return matchLabel(getControlFlowEndWorker(statement.statement, checker), statement.label);
            case ts.SyntaxKind.WithStatement:
                return getControlFlowEndWorker(statement.statement, checker);
            case ts.SyntaxKind.ExpressionStatement:
                if (checker === undefined)
                    return defaultControlFlowEnd;
                return handleExpressionStatement(statement, checker);
            default:
                return defaultControlFlowEnd;
        }
    }
    function handleBlock(statement, checker) {
        const result = { statements: [], end: false };
        for (const s of statement.statements) {
            const current = getControlFlowEndWorker(s, checker);
            result.statements.push(...current.statements);
            if (current.end) {
                result.end = true;
                break;
            }
        }
        return result;
    }
    function handleForInOrOfStatement(statement, checker) {
        const end = matchBreakOrContinue(getControlFlowEndWorker(statement.statement, checker), node_1.isBreakOrContinueStatement);
        end.end = false; // loop body is guaranteed to be executed
        return end;
    }
    function handleForAndWhileStatement(statement, checker) {
        const constantCondition = statement.kind === ts.SyntaxKind.WhileStatement
            ? getConstantCondition(statement.expression)
            : statement.condition === undefined || getConstantCondition(statement.condition);
        if (constantCondition === false)
            return defaultControlFlowEnd; // loop body is never executed
        const end = matchBreakOrContinue(getControlFlowEndWorker(statement.statement, checker), node_1.isBreakOrContinueStatement);
        if (constantCondition === undefined)
            end.end = false; // can't be sure that loop body is executed at all
        return end;
    }
    /** Simply detects `true` and `false` in conditions. That matches TypeScript's behavior. */
    function getConstantCondition(node) {
        switch (node.kind) {
            case ts.SyntaxKind.TrueKeyword:
                return true;
            case ts.SyntaxKind.FalseKeyword:
                return false;
            default:
                return;
        }
    }
    function handleIfStatement(node, checker) {
        switch (getConstantCondition(node.expression)) {
            case true:
                // else branch is never executed
                return getControlFlowEndWorker(node.thenStatement, checker);
            case false:
                // then branch is never executed
                return node.elseStatement === undefined
                    ? defaultControlFlowEnd
                    : getControlFlowEndWorker(node.elseStatement, checker);
        }
        const then = getControlFlowEndWorker(node.thenStatement, checker);
        if (node.elseStatement === undefined)
            return {
                statements: then.statements,
                end: false,
            };
        const elze = getControlFlowEndWorker(node.elseStatement, checker);
        return {
            statements: [...then.statements, ...elze.statements],
            end: then.end && elze.end,
        };
    }
    function handleSwitchStatement(node, checker) {
        let hasDefault = false;
        const result = {
            statements: [],
            end: false,
        };
        for (const clause of node.caseBlock.clauses) {
            if (clause.kind === ts.SyntaxKind.DefaultClause)
                hasDefault = true;
            const current = handleBlock(clause, checker);
            result.end = current.end;
            result.statements.push(...current.statements);
        }
        result.end && (result.end = hasDefault || checker !== undefined && util_1.hasExhaustiveCaseClauses(node, checker));
        return result;
    }
    function handleTryStatement(node, checker) {
        let finallyResult;
        if (node.finallyBlock !== undefined) {
            finallyResult = handleBlock(node.finallyBlock, checker);
            // if 'finally' always ends control flow, we are not interested in any jump statements from 'try' or 'catch'
            if (finallyResult.end)
                return finallyResult;
        }
        const tryResult = handleBlock(node.tryBlock, checker);
        if (node.catchClause === undefined)
            return { statements: finallyResult.statements.concat(tryResult.statements), end: tryResult.end };
        const catchResult = handleBlock(node.catchClause.block, checker);
        return {
            statements: tryResult.statements
                // remove all throw statements and throwing function calls from the list of control flow statements inside tryBlock
                .filter((s) => s.kind !== ts.SyntaxKind.ThrowStatement && s.kind !== ts.SyntaxKind.ExpressionStatement)
                .concat(catchResult.statements, finallyResult === undefined ? [] : finallyResult.statements),
            end: tryResult.end && catchResult.end, // only ends control flow if try AND catch definitely end control flow
        };
    }
    /** Dotted name as TypeScript requires it for assertion signatures to affect control flow. */
    function isDottedNameWithExplicitTypeAnnotation(node, checker) {
        while (true) {
            switch (node.kind) {
                case ts.SyntaxKind.Identifier: {
                    const symbol = checker.getExportSymbolOfSymbol(checker.getSymbolAtLocation(node));
                    return isExplicitlyTypedSymbol(util_1.isSymbolFlagSet(symbol, ts.SymbolFlags.Alias) ? checker.getAliasedSymbol(symbol) : symbol, checker);
                }
                case ts.SyntaxKind.ThisKeyword:
                    return isExplicitlyTypedThis(node);
                case ts.SyntaxKind.SuperKeyword:
                    return true;
                case ts.SyntaxKind.PropertyAccessExpression:
                    if (!isExplicitlyTypedSymbol(checker.getSymbolAtLocation(node), checker))
                        return false;
                // falls through
                case ts.SyntaxKind.ParenthesizedExpression:
                    node = node.expression;
                    continue;
                default:
                    return false;
            }
        }
    }
    function isExplicitlyTypedSymbol(symbol, checker) {
        if (symbol === undefined)
            return false;
        if (util_1.isSymbolFlagSet(symbol, ts.SymbolFlags.Function | ts.SymbolFlags.Method | ts.SymbolFlags.Class | ts.SymbolFlags.ValueModule))
            return true;
        if (!util_1.isSymbolFlagSet(symbol, ts.SymbolFlags.Variable | ts.SymbolFlags.Property))
            return false;
        if (symbol.valueDeclaration === undefined)
            return false;
        if (declarationHasExplicitTypeAnnotation(symbol.valueDeclaration))
            return true;
        return node_1.isVariableDeclaration(symbol.valueDeclaration) &&
            symbol.valueDeclaration.parent.parent.kind === ts.SyntaxKind.ForOfStatement &&
            isDottedNameWithExplicitTypeAnnotation(symbol.valueDeclaration.parent.parent.expression, checker);
    }
    function declarationHasExplicitTypeAnnotation(node) {
        if (ts.isJSDocPropertyLikeTag(node))
            return node.typeExpression !== undefined;
        return (node_1.isVariableDeclaration(node) ||
            node_1.isParameterDeclaration(node) ||
            node_1.isPropertyDeclaration(node) ||
            node_1.isPropertySignature(node)) && (util_1.isNodeFlagSet(node, ts.NodeFlags.JavaScriptFile)
            ? ts.getJSDocType(node)
            : node.type) !== undefined;
    }
    function isExplicitlyTypedThis(node) {
        var _a;
        do {
            node = node.parent;
            if (node_1.isDecorator(node)) {
                // `this` in decorators always resolves outside of the containing class
                if (node.parent.kind === ts.SyntaxKind.Parameter && node_1.isClassLikeDeclaration(node.parent.parent.parent)) {
                    node = node.parent.parent.parent.parent;
                }
                else if (node_1.isClassLikeDeclaration(node.parent.parent)) {
                    node = node.parent.parent.parent;
                }
                else if (node_1.isClassLikeDeclaration(node.parent)) {
                    node = node.parent.parent;
                }
            }
        } while (util_1.isFunctionScopeBoundary(node) !== 1 /* Function */ || node.kind === ts.SyntaxKind.ArrowFunction);
        return util_1.isFunctionWithBody(node) &&
            (util_1.isNodeFlagSet(node, ts.NodeFlags.JavaScriptFile)
                ? ((_a = ts.getJSDocThisTag(node)) === null || _a === void 0 ? void 0 : _a.typeExpression) !== undefined
                : node.parameters.length !== 0 && util_1.isThisParameter(node.parameters[0]) && node.parameters[0].type !== undefined) ||
            node_1.isClassLikeDeclaration(node.parent);
    }
    var SignatureEffect;
    (function (SignatureEffect) {
        SignatureEffect[SignatureEffect["Never"] = 1] = "Never";
        SignatureEffect[SignatureEffect["Asserts"] = 2] = "Asserts";
    })(SignatureEffect = exports.SignatureEffect || (exports.SignatureEffect = {}));
    /**
     * Dermines whether a top level CallExpression has a control flow effect according to TypeScript's rules.
     * This handles functions returning `never` and `asserts`.
     */
    function callExpressionAffectsControlFlow(node, checker) {
        var _a, _b, _c;
        if (!node_1.isExpressionStatement(node.parent) ||
            ts.isOptionalChain(node) ||
            !isDottedNameWithExplicitTypeAnnotation(node.expression, checker))
            return;
        const signature = checker.getResolvedSignature(node);
        if ((signature === null || signature === void 0 ? void 0 : signature.declaration) === undefined)
            return;
        const typeNode = ts.isJSDocSignature(signature.declaration)
            ? (_b = (_a = signature.declaration.type) === null || _a === void 0 ? void 0 : _a.typeExpression) === null || _b === void 0 ? void 0 : _b.type
            : (_c = signature.declaration.type) !== null && _c !== void 0 ? _c : (util_1.isNodeFlagSet(signature.declaration, ts.NodeFlags.JavaScriptFile)
                ? ts.getJSDocReturnType(signature.declaration)
                : undefined);
        if (typeNode === undefined)
            return;
        if (node_1.isTypePredicateNode(typeNode) && typeNode.assertsModifier !== undefined)
            return 2 /* Asserts */;
        return util_1.isTypeFlagSet(checker.getTypeFromTypeNode(typeNode), ts.TypeFlags.Never) ? 1 /* Never */ : undefined;
    }
    exports.callExpressionAffectsControlFlow = callExpressionAffectsControlFlow;
    function handleExpressionStatement(node, checker) {
        if (!node_1.isCallExpression(node.expression))
            return defaultControlFlowEnd;
        switch (callExpressionAffectsControlFlow(node.expression, checker)) {
            case 2 /* Asserts */:
                return { statements: [node], end: false };
            case 1 /* Never */:
                return { statements: [node], end: true };
            case undefined:
                return defaultControlFlowEnd;
        }
    }
    function matchBreakOrContinue(current, pred) {
        const result = {
            statements: [],
            end: current.end,
        };
        for (const statement of current.statements) {
            if (pred(statement) && statement.label === undefined) {
                result.end = false;
                continue;
            }
            result.statements.push(statement);
        }
        return result;
    }
    function matchLabel(current, label) {
        const result = {
            statements: [],
            end: current.end,
        };
        const labelText = label.text;
        for (const statement of current.statements) {
            switch (statement.kind) {
                case ts.SyntaxKind.BreakStatement:
                case ts.SyntaxKind.ContinueStatement:
                    if (statement.label !== undefined && statement.label.text === labelText) {
                        result.end = false;
                        continue;
                    }
            }
            result.statements.push(statement);
        }
        return result;
    }
},
/* tsutils/util/convert-ast.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const ts = require(21) /* typescript */;
    const util_1 = require(929) /* ./util */;
    /**
     * Takes a `ts.SourceFile` and creates data structures that are easier (or more performant) to traverse.
     * Note that there is only a performance gain if you can reuse these structures. It's not recommended for one-time AST walks.
     */
    function convertAst(sourceFile) {
        const wrapped = {
            node: sourceFile,
            parent: undefined,
            kind: ts.SyntaxKind.SourceFile,
            children: [],
            next: undefined,
            skip: undefined,
        };
        const flat = [];
        let current = wrapped;
        function collectChildren(node) {
            current.children.push({
                node,
                parent: current,
                kind: node.kind,
                children: [],
                next: undefined,
                skip: undefined,
            });
        }
        const stack = [];
        while (true) {
            if (current.children.length === 0) {
                ts.forEachChild(current.node, collectChildren);
                if (current.children.length === 0) {
                    current = current.parent; // nothing to do here, go back to parent
                }
                else {
                    // recurse into first child
                    const firstChild = current.children[0];
                    current.next = firstChild;
                    flat.push(firstChild.node);
                    if (util_1.isNodeKind(firstChild.kind))
                        current = firstChild;
                    stack.push(1); // set index in stack so we know where to continue processing children
                }
            }
            else {
                const index = stack[stack.length - 1];
                if (index < current.children.length) { // handles 2nd child to the last
                    const currentChild = current.children[index];
                    flat.push(currentChild.node);
                    let previous = current.children[index - 1];
                    while (previous.children.length !== 0) {
                        previous.skip = currentChild;
                        previous = previous.children[previous.children.length - 1];
                    }
                    previous.skip = previous.next = currentChild;
                    ++stack[stack.length - 1];
                    if (util_1.isNodeKind(currentChild.kind))
                        current = currentChild; // recurse into child
                }
                else {
                    // done on this node
                    if (stack.length === 1)
                        break;
                    // remove index from stack and go back to parent
                    stack.pop();
                    current = current.parent;
                }
            }
        }
        return {
            wrapped,
            flat,
        };
    }
    exports.convertAst = convertAst;
},
/* @typescript-eslint/eslint-plugin/dist/util/misc.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    /**
     * Check if the context file name is *.d.ts or *.d.tsx
     */
    function isDefinitionFile(fileName) {
        return /\.d\.tsx?$/i.test(fileName || '');
    }
    exports.isDefinitionFile = isDefinitionFile;
    /**
     * Upper cases the first character or the string
     */
    function upperCaseFirst(str) {
        return str[0].toUpperCase() + str.slice(1);
    }
    exports.upperCaseFirst = upperCaseFirst;
    function arraysAreEqual(a, b, eq) {
        return (a === b ||
            (a !== undefined &&
                b !== undefined &&
                a.length === b.length &&
                a.every((x, idx) => eq(x, b[idx]))));
    }
    exports.arraysAreEqual = arraysAreEqual;
    /** Returns the first non-`undefined` result. */
    function findFirstResult(inputs, getResult) {
        for (const element of inputs) {
            const result = getResult(element);
            if (result !== undefined) {
                return result;
            }
        }
        return undefined;
    }
    exports.findFirstResult = findFirstResult;
    /**
     * Gets a string representation of the name of the index signature.
     */
    function getNameFromIndexSignature(node) {
        const propName = node.parameters.find((parameter) => parameter.type === experimental_utils_1.AST_NODE_TYPES.Identifier);
        return propName ? propName.name : '(index signature)';
    }
    exports.getNameFromIndexSignature = getNameFromIndexSignature;
    /**
     * Gets a string name representation of the name of the given MethodDefinition
     * or ClassProperty node, with handling for computed property names.
     */
    function getNameFromMember(member, sourceCode) {
        if (member.key.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
            return member.key.name;
        }
        if (member.key.type === experimental_utils_1.AST_NODE_TYPES.Literal) {
            return `${member.key.value}`;
        }
        return sourceCode.text.slice(...member.key.range);
    }
    exports.getNameFromMember = getNameFromMember;
    function getEnumNames(myEnum) {
        return Object.keys(myEnum).filter(x => isNaN(parseInt(x)));
    }
    exports.getEnumNames = getEnumNames;
},
/* @typescript-eslint/eslint-plugin/dist/util/nullThrows.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    /**
     * A set of common reasons for calling nullThrows
     */
    const NullThrowsReasons = {
        MissingParent: 'Expected node to have a parent.',
        MissingToken: (token, thing) => `Expected to find a ${token} for the ${thing}.`,
    };
    exports.NullThrowsReasons = NullThrowsReasons;
    /**
     * Assert that a value must not be null or undefined.
     * This is a nice explicit alternative to the non-null assertion operator.
     */
    function nullThrows(value, message) {
        // this function is primarily used to keep types happy in a safe way
        // i.e. is used when we expect that a value is never nullish
        // this means that it's pretty much impossible to test the below if...
        // so ignore it in coverage metrics.
        /* istanbul ignore if */
        if (value === null || value === undefined) {
            throw new Error(`Non-null Assertion Failed: ${message}`);
        }
        return value;
    }
    exports.nullThrows = nullThrows;
},
/* @typescript-eslint/eslint-plugin/dist/util/objectIterators.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function objectForEachKey(obj, callback) {
        const keys = Object.keys(obj);
        for (const key of keys) {
            callback(key);
        }
    }
    exports.objectForEachKey = objectForEachKey;
    function objectMapKey(obj, callback) {
        const values = [];
        objectForEachKey(obj, key => {
            values.push(callback(key));
        });
        return values;
    }
    exports.objectMapKey = objectMapKey;
    function objectReduceKey(obj, callback, initial) {
        let accumulator = initial;
        objectForEachKey(obj, key => {
            accumulator = callback(accumulator, key);
        });
        return accumulator;
    }
    exports.objectReduceKey = objectReduceKey;
},
/* @typescript-eslint/eslint-plugin/dist/util/propertyTypes.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function getTypeOfPropertyOfName(checker, type, name, escapedName) {
        // Most names are directly usable in the checker and aren't different from escaped names
        if (!escapedName || !name.startsWith('__')) {
            return checker.getTypeOfPropertyOfType(type, name);
        }
        // Symbolic names may differ in their escaped name compared to their human-readable name
        // https://github.com/typescript-eslint/typescript-eslint/issues/2143
        const escapedProperty = type
            .getProperties()
            .find(property => property.escapedName === escapedName);
        return escapedProperty
            ? checker.getDeclaredTypeOfSymbol(escapedProperty)
            : undefined;
    }
    exports.getTypeOfPropertyOfName = getTypeOfPropertyOfName;
    function getTypeOfPropertyOfType(checker, type, property) {
        return getTypeOfPropertyOfName(checker, type, property.getName(), property.getEscapedName());
    }
    exports.getTypeOfPropertyOfType = getTypeOfPropertyOfType;
},
/* @typescript-eslint/eslint-plugin/dist/util/requiresQuoting.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    function requiresQuoting(name, target = ts.ScriptTarget.ESNext) {
        if (name.length === 0) {
            return true;
        }
        if (!ts.isIdentifierStart(name.charCodeAt(0), target)) {
            return true;
        }
        for (let i = 1; i < name.length; i += 1) {
            if (!ts.isIdentifierPart(name.charCodeAt(i), target)) {
                return true;
            }
        }
        return false;
    }
    exports.requiresQuoting = requiresQuoting;
},
/* @typescript-eslint/eslint-plugin/dist/util/types.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const debug_1 = __importDefault(require(72) /* debug */);
    const tsutils_1 = require(916) /* tsutils */;
    const ts = __importStar(require(21) /* typescript */);
    const log = debug_1.default('typescript-eslint:eslint-plugin:utils:types');
    /**
     * Checks if the given type is either an array type,
     * or a union made up solely of array types.
     */
    function isTypeArrayTypeOrUnionOfArrayTypes(type, checker) {
        for (const t of tsutils_1.unionTypeParts(type)) {
            if (!checker.isArrayType(t)) {
                return false;
            }
        }
        return true;
    }
    exports.isTypeArrayTypeOrUnionOfArrayTypes = isTypeArrayTypeOrUnionOfArrayTypes;
    /**
     * @param type Type being checked by name.
     * @param allowedNames Symbol names checking on the type.
     * @returns Whether the type is, extends, or contains all of the allowed names.
     */
    function containsAllTypesByName(type, allowAny, allowedNames) {
        if (isTypeFlagSet(type, ts.TypeFlags.Any | ts.TypeFlags.Unknown)) {
            return !allowAny;
        }
        if (tsutils_1.isTypeReference(type)) {
            type = type.target;
        }
        const symbol = type.getSymbol();
        if (symbol && allowedNames.has(symbol.name)) {
            return true;
        }
        if (tsutils_1.isUnionOrIntersectionType(type)) {
            return type.types.every(t => containsAllTypesByName(t, allowAny, allowedNames));
        }
        const bases = type.getBaseTypes();
        return (typeof bases !== 'undefined' &&
            bases.length > 0 &&
            bases.every(t => containsAllTypesByName(t, allowAny, allowedNames)));
    }
    exports.containsAllTypesByName = containsAllTypesByName;
    /**
     * Get the type name of a given type.
     * @param typeChecker The context sensitive TypeScript TypeChecker.
     * @param type The type to get the name of.
     */
    function getTypeName(typeChecker, type) {
        // It handles `string` and string literal types as string.
        if ((type.flags & ts.TypeFlags.StringLike) !== 0) {
            return 'string';
        }
        // If the type is a type parameter which extends primitive string types,
        // but it was not recognized as a string like. So check the constraint
        // type of the type parameter.
        if ((type.flags & ts.TypeFlags.TypeParameter) !== 0) {
            // `type.getConstraint()` method doesn't return the constraint type of
            // the type parameter for some reason. So this gets the constraint type
            // via AST.
            const symbol = type.getSymbol();
            const decls = symbol === null || symbol === void 0 ? void 0 : symbol.getDeclarations();
            const typeParamDecl = decls === null || decls === void 0 ? void 0 : decls[0];
            if (ts.isTypeParameterDeclaration(typeParamDecl) &&
                typeParamDecl.constraint != null) {
                return getTypeName(typeChecker, typeChecker.getTypeFromTypeNode(typeParamDecl.constraint));
            }
        }
        // If the type is a union and all types in the union are string like,
        // return `string`. For example:
        // - `"a" | "b"` is string.
        // - `string | string[]` is not string.
        if (type.isUnion() &&
            type.types
                .map(value => getTypeName(typeChecker, value))
                .every(t => t === 'string')) {
            return 'string';
        }
        // If the type is an intersection and a type in the intersection is string
        // like, return `string`. For example: `string & {__htmlEscaped: void}`
        if (type.isIntersection() &&
            type.types
                .map(value => getTypeName(typeChecker, value))
                .some(t => t === 'string')) {
            return 'string';
        }
        return typeChecker.typeToString(type);
    }
    exports.getTypeName = getTypeName;
    /**
     * Resolves the given node's type. Will resolve to the type's generic constraint, if it has one.
     */
    function getConstrainedTypeAtLocation(checker, node) {
        const nodeType = checker.getTypeAtLocation(node);
        const constrained = checker.getBaseConstraintOfType(nodeType);
        return constrained !== null && constrained !== void 0 ? constrained : nodeType;
    }
    exports.getConstrainedTypeAtLocation = getConstrainedTypeAtLocation;
    /**
     * Checks if the given type is (or accepts) nullable
     * @param isReceiver true if the type is a receiving type (i.e. the type of a called function's parameter)
     */
    function isNullableType(type, { isReceiver = false, allowUndefined = true, } = {}) {
        const flags = getTypeFlags(type);
        if (isReceiver && flags & (ts.TypeFlags.Any | ts.TypeFlags.Unknown)) {
            return true;
        }
        if (allowUndefined) {
            return (flags & (ts.TypeFlags.Null | ts.TypeFlags.Undefined)) !== 0;
        }
        else {
            return (flags & ts.TypeFlags.Null) !== 0;
        }
    }
    exports.isNullableType = isNullableType;
    /**
     * Gets the declaration for the given variable
     */
    function getDeclaration(checker, node) {
        var _a;
        const symbol = checker.getSymbolAtLocation(node);
        if (!symbol) {
            return null;
        }
        const declarations = symbol.getDeclarations();
        return (_a = declarations === null || declarations === void 0 ? void 0 : declarations[0]) !== null && _a !== void 0 ? _a : null;
    }
    exports.getDeclaration = getDeclaration;
    /**
     * Gets all of the type flags in a type, iterating through unions automatically
     */
    function getTypeFlags(type) {
        let flags = 0;
        for (const t of tsutils_1.unionTypeParts(type)) {
            flags |= t.flags;
        }
        return flags;
    }
    exports.getTypeFlags = getTypeFlags;
    /**
     * Checks if the given type is (or accepts) the given flags
     * @param isReceiver true if the type is a receiving type (i.e. the type of a called function's parameter)
     */
    function isTypeFlagSet(type, flagsToCheck, isReceiver) {
        const flags = getTypeFlags(type);
        if (isReceiver && flags & (ts.TypeFlags.Any | ts.TypeFlags.Unknown)) {
            return true;
        }
        return (flags & flagsToCheck) !== 0;
    }
    exports.isTypeFlagSet = isTypeFlagSet;
    /**
     * @returns Whether a type is an instance of the parent type, including for the parent's base types.
     */
    function typeIsOrHasBaseType(type, parentType) {
        const parentSymbol = parentType.getSymbol();
        if (!type.getSymbol() || !parentSymbol) {
            return false;
        }
        const typeAndBaseTypes = [type];
        const ancestorTypes = type.getBaseTypes();
        if (ancestorTypes) {
            typeAndBaseTypes.push(...ancestorTypes);
        }
        for (const baseType of typeAndBaseTypes) {
            const baseSymbol = baseType.getSymbol();
            if (baseSymbol && baseSymbol.name === parentSymbol.name) {
                return true;
            }
        }
        return false;
    }
    exports.typeIsOrHasBaseType = typeIsOrHasBaseType;
    /**
     * Gets the source file for a given node
     */
    function getSourceFileOfNode(node) {
        while (node && node.kind !== ts.SyntaxKind.SourceFile) {
            node = node.parent;
        }
        return node;
    }
    exports.getSourceFileOfNode = getSourceFileOfNode;
    function getTokenAtPosition(sourceFile, position) {
        const queue = [sourceFile];
        let current;
        while (queue.length > 0) {
            current = queue.shift();
            // find the child that contains 'position'
            for (const child of current.getChildren(sourceFile)) {
                const start = child.getFullStart();
                if (start > position) {
                    // If this child begins after position, then all subsequent children will as well.
                    return current;
                }
                const end = child.getEnd();
                if (position < end ||
                    (position === end && child.kind === ts.SyntaxKind.EndOfFileToken)) {
                    queue.push(child);
                    break;
                }
            }
        }
        return current;
    }
    exports.getTokenAtPosition = getTokenAtPosition;
    function getEqualsKind(operator) {
        switch (operator) {
            case '==':
                return {
                    isPositive: true,
                    isStrict: false,
                };
            case '===':
                return {
                    isPositive: true,
                    isStrict: true,
                };
            case '!=':
                return {
                    isPositive: false,
                    isStrict: false,
                };
            case '!==':
                return {
                    isPositive: false,
                    isStrict: true,
                };
            default:
                return undefined;
        }
    }
    exports.getEqualsKind = getEqualsKind;
    function getTypeArguments(type, checker) {
        var _a;
        // getTypeArguments was only added in TS3.7
        if (checker.getTypeArguments) {
            return checker.getTypeArguments(type);
        }
        return (_a = type.typeArguments) !== null && _a !== void 0 ? _a : [];
    }
    exports.getTypeArguments = getTypeArguments;
    /**
     * @returns true if the type is `unknown`
     */
    function isTypeUnknownType(type) {
        return isTypeFlagSet(type, ts.TypeFlags.Unknown);
    }
    exports.isTypeUnknownType = isTypeUnknownType;
    /**
     * @returns true if the type is `any`
     */
    function isTypeAnyType(type) {
        if (isTypeFlagSet(type, ts.TypeFlags.Any)) {
            if (type.intrinsicName === 'error') {
                log('Found an "error" any type');
            }
            return true;
        }
        return false;
    }
    exports.isTypeAnyType = isTypeAnyType;
    /**
     * @returns true if the type is `any[]`
     */
    function isTypeAnyArrayType(type, checker) {
        return (checker.isArrayType(type) &&
            isTypeAnyType(
            // getTypeArguments was only added in TS3.7
            getTypeArguments(type, checker)[0]));
    }
    exports.isTypeAnyArrayType = isTypeAnyArrayType;
    /**
     * @returns true if the type is `unknown[]`
     */
    function isTypeUnknownArrayType(type, checker) {
        return (checker.isArrayType(type) &&
            isTypeUnknownType(
            // getTypeArguments was only added in TS3.7
            getTypeArguments(type, checker)[0]));
    }
    exports.isTypeUnknownArrayType = isTypeUnknownArrayType;
    /**
     * @returns `AnyType.Any` if the type is `any`, `AnyType.AnyArray` if the type is `any[]` or `readonly any[]`,
     *          otherwise it returns `AnyType.Safe`.
     */
    function isAnyOrAnyArrayTypeDiscriminated(node, checker) {
        const type = checker.getTypeAtLocation(node);
        if (isTypeAnyType(type)) {
            return 0 /* Any */;
        }
        if (isTypeAnyArrayType(type, checker)) {
            return 1 /* AnyArray */;
        }
        return 2 /* Safe */;
    }
    exports.isAnyOrAnyArrayTypeDiscriminated = isAnyOrAnyArrayTypeDiscriminated;
    /**
     * Does a simple check to see if there is an any being assigned to a non-any type.
     *
     * This also checks generic positions to ensure there's no unsafe sub-assignments.
     * Note: in the case of generic positions, it makes the assumption that the two types are the same.
     *
     * @example See tests for examples
     *
     * @returns false if it's safe, or an object with the two types if it's unsafe
     */
    function isUnsafeAssignment(type, receiver, checker, senderNode) {
        var _a, _b;
        if (isTypeAnyType(type)) {
            // Allow assignment of any ==> unknown.
            if (isTypeUnknownType(receiver)) {
                return false;
            }
            if (!isTypeAnyType(receiver)) {
                return { sender: type, receiver };
            }
        }
        if (tsutils_1.isTypeReference(type) && tsutils_1.isTypeReference(receiver)) {
            // TODO - figure out how to handle cases like this,
            // where the types are assignable, but not the same type
            /*
            function foo(): ReadonlySet<number> { return new Set<any>(); }
        
            // and
        
            type Test<T> = { prop: T }
            type Test2 = { prop: string }
            declare const a: Test<any>;
            const b: Test2 = a;
            */
            if (type.target !== receiver.target) {
                // if the type references are different, assume safe, as we won't know how to compare the two types
                // the generic positions might not be equivalent for both types
                return false;
            }
            if ((senderNode === null || senderNode === void 0 ? void 0 : senderNode.type) === experimental_utils_1.AST_NODE_TYPES.NewExpression &&
                senderNode.callee.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                senderNode.callee.name === 'Map' &&
                senderNode.arguments.length === 0 &&
                senderNode.typeParameters == null) {
                // special case to handle `new Map()`
                // unfortunately Map's default empty constructor is typed to return `Map<any, any>` :(
                // https://github.com/typescript-eslint/typescript-eslint/issues/2109#issuecomment-634144396
                return false;
            }
            const typeArguments = (_a = type.typeArguments) !== null && _a !== void 0 ? _a : [];
            const receiverTypeArguments = (_b = receiver.typeArguments) !== null && _b !== void 0 ? _b : [];
            for (let i = 0; i < typeArguments.length; i += 1) {
                const arg = typeArguments[i];
                const receiverArg = receiverTypeArguments[i];
                const unsafe = isUnsafeAssignment(arg, receiverArg, checker, senderNode);
                if (unsafe) {
                    return { sender: type, receiver };
                }
            }
            return false;
        }
        return false;
    }
    exports.isUnsafeAssignment = isUnsafeAssignment;
    /**
     * Returns the contextual type of a given node.
     * Contextual type is the type of the target the node is going into.
     * i.e. the type of a called function's parameter, or the defined type of a variable declaration
     */
    function getContextualType(checker, node) {
        const parent = node.parent;
        if (!parent) {
            return;
        }
        if (tsutils_1.isCallExpression(parent) || tsutils_1.isNewExpression(parent)) {
            if (node === parent.expression) {
                // is the callee, so has no contextual type
                return;
            }
        }
        else if (tsutils_1.isVariableDeclaration(parent) ||
            tsutils_1.isPropertyDeclaration(parent) ||
            tsutils_1.isParameterDeclaration(parent)) {
            return parent.type ? checker.getTypeFromTypeNode(parent.type) : undefined;
        }
        else if (tsutils_1.isJsxExpression(parent)) {
            return checker.getContextualType(parent);
        }
        else if (tsutils_1.isPropertyAssignment(parent) && tsutils_1.isIdentifier(node)) {
            return checker.getContextualType(node);
        }
        else if (tsutils_1.isBinaryExpression(parent) &&
            parent.operatorToken.kind === ts.SyntaxKind.EqualsToken &&
            parent.right === node) {
            // is RHS of assignment
            return checker.getTypeAtLocation(parent.left);
        }
        else if (![ts.SyntaxKind.TemplateSpan, ts.SyntaxKind.JsxExpression].includes(parent.kind)) {
            // parent is not something we know we can get the contextual type of
            return;
        }
        // TODO - support return statement checking
        return checker.getContextualType(node);
    }
    exports.getContextualType = getContextualType;
},
/* @typescript-eslint/eslint-plugin/dist/rules/array-type.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    /**
     * Check whatever node can be considered as simple
     * @param node the node to be evaluated.
     */
    function isSimpleType(node) {
        switch (node.type) {
            case experimental_utils_1.AST_NODE_TYPES.Identifier:
            case experimental_utils_1.AST_NODE_TYPES.TSAnyKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSBooleanKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSNeverKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSNumberKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSObjectKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSStringKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSSymbolKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSUnknownKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSVoidKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSNullKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSArrayType:
            case experimental_utils_1.AST_NODE_TYPES.TSUndefinedKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSThisType:
            case experimental_utils_1.AST_NODE_TYPES.TSQualifiedName:
                return true;
            case experimental_utils_1.AST_NODE_TYPES.TSTypeReference:
                if (node.typeName &&
                    node.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    node.typeName.name === 'Array') {
                    if (!node.typeParameters) {
                        return true;
                    }
                    if (node.typeParameters.params.length === 1) {
                        return isSimpleType(node.typeParameters.params[0]);
                    }
                }
                else {
                    if (node.typeParameters) {
                        return false;
                    }
                    return isSimpleType(node.typeName);
                }
                return false;
            default:
                return false;
        }
    }
    /**
     * Check if node needs parentheses
     * @param node the node to be evaluated.
     */
    function typeNeedsParentheses(node) {
        switch (node.type) {
            case experimental_utils_1.AST_NODE_TYPES.TSTypeReference:
                return typeNeedsParentheses(node.typeName);
            case experimental_utils_1.AST_NODE_TYPES.TSUnionType:
            case experimental_utils_1.AST_NODE_TYPES.TSFunctionType:
            case experimental_utils_1.AST_NODE_TYPES.TSIntersectionType:
            case experimental_utils_1.AST_NODE_TYPES.TSTypeOperator:
            case experimental_utils_1.AST_NODE_TYPES.TSInferType:
                return true;
            case experimental_utils_1.AST_NODE_TYPES.Identifier:
                return node.name === 'ReadonlyArray';
            default:
                return false;
        }
    }
    const arrayOption = { enum: ['array', 'generic', 'array-simple'] };
    exports.default = util.createRule({
        name: 'array-type',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Requires using either `T[]` or `Array<T>` for arrays',
                category: 'Stylistic Issues',
                // too opinionated to be recommended
                recommended: false,
            },
            fixable: 'code',
            messages: {
                errorStringGeneric: "Array type using '{{type}}[]' is forbidden. Use 'Array<{{type}}>' instead.",
                errorStringGenericSimple: "Array type using '{{type}}[]' is forbidden for non-simple types. Use 'Array<{{type}}>' instead.",
                errorStringArray: "Array type using 'Array<{{type}}>' is forbidden. Use '{{type}}[]' instead.",
                errorStringArraySimple: "Array type using 'Array<{{type}}>' is forbidden for simple types. Use '{{type}}[]' instead.",
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        default: arrayOption,
                        readonly: arrayOption,
                    },
                },
            ],
        },
        defaultOptions: [
            {
                default: 'array',
            },
        ],
        create(context, [options]) {
            var _a;
            const sourceCode = context.getSourceCode();
            const defaultOption = options.default;
            const readonlyOption = (_a = options.readonly) !== null && _a !== void 0 ? _a : defaultOption;
            /**
             * @param node the node to be evaluated.
             */
            function getMessageType(node) {
                if (node) {
                    if (node.type === experimental_utils_1.AST_NODE_TYPES.TSParenthesizedType) {
                        return getMessageType(node.typeAnnotation);
                    }
                    if (isSimpleType(node)) {
                        return sourceCode.getText(node);
                    }
                }
                return 'T';
            }
            return {
                TSArrayType(node) {
                    const isReadonly = node.parent &&
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.TSTypeOperator &&
                        node.parent.operator === 'readonly';
                    const currentOption = isReadonly ? readonlyOption : defaultOption;
                    if (currentOption === 'array' ||
                        (currentOption === 'array-simple' && isSimpleType(node.elementType))) {
                        return;
                    }
                    const messageId = currentOption === 'generic'
                        ? 'errorStringGeneric'
                        : 'errorStringGenericSimple';
                    const errorNode = isReadonly ? node.parent : node;
                    context.report({
                        node: errorNode,
                        messageId,
                        data: {
                            type: getMessageType(node.elementType),
                        },
                        fix(fixer) {
                            const typeNode = node.elementType.type === experimental_utils_1.AST_NODE_TYPES.TSParenthesizedType
                                ? node.elementType.typeAnnotation
                                : node.elementType;
                            const arrayType = isReadonly ? 'ReadonlyArray' : 'Array';
                            return [
                                fixer.replaceTextRange([errorNode.range[0], typeNode.range[0]], `${arrayType}<`),
                                fixer.replaceTextRange([typeNode.range[1], errorNode.range[1]], '>'),
                            ];
                        },
                    });
                },
                TSTypeReference(node) {
                    var _a, _b;
                    if (node.typeName.type !== experimental_utils_1.AST_NODE_TYPES.Identifier ||
                        !(node.typeName.name === 'Array' ||
                            node.typeName.name === 'ReadonlyArray')) {
                        return;
                    }
                    const isReadonlyArrayType = node.typeName.name === 'ReadonlyArray';
                    const currentOption = isReadonlyArrayType
                        ? readonlyOption
                        : defaultOption;
                    if (currentOption === 'generic') {
                        return;
                    }
                    const readonlyPrefix = isReadonlyArrayType ? 'readonly ' : '';
                    const typeParams = (_a = node.typeParameters) === null || _a === void 0 ? void 0 : _a.params;
                    const messageId = currentOption === 'array'
                        ? 'errorStringArray'
                        : 'errorStringArraySimple';
                    if (!typeParams || typeParams.length === 0) {
                        // Create an 'any' array
                        context.report({
                            node,
                            messageId,
                            data: {
                                type: 'any',
                            },
                            fix(fixer) {
                                return fixer.replaceText(node, `${readonlyPrefix}any[]`);
                            },
                        });
                        return;
                    }
                    if (typeParams.length !== 1 ||
                        (currentOption === 'array-simple' && !isSimpleType(typeParams[0]))) {
                        return;
                    }
                    const type = typeParams[0];
                    const typeParens = typeNeedsParentheses(type);
                    const parentParens = readonlyPrefix && ((_b = node.parent) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.TSArrayType;
                    const start = `${parentParens ? '(' : ''}${readonlyPrefix}${typeParens ? '(' : ''}`;
                    const end = `${typeParens ? ')' : ''}[]${parentParens ? ')' : ''}`;
                    context.report({
                        node,
                        messageId,
                        data: {
                            type: getMessageType(type),
                        },
                        fix(fixer) {
                            return [
                                fixer.replaceTextRange([node.range[0], type.range[0]], start),
                                fixer.replaceTextRange([type.range[1], node.range[1]], end),
                            ];
                        },
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/await-thenable.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const tsutils = __importStar(require(916) /* tsutils */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'await-thenable',
        meta: {
            docs: {
                description: 'Disallows awaiting a value that is not a Thenable',
                category: 'Best Practices',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                await: 'Unexpected `await` of a non-Promise (non-"Thenable") value.',
            },
            schema: [],
            type: 'problem',
        },
        defaultOptions: [],
        create(context) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            return {
                AwaitExpression(node) {
                    const originalNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                    const type = checker.getTypeAtLocation(originalNode.expression);
                    if (!util.isTypeAnyType(type) &&
                        !util.isTypeUnknownType(type) &&
                        !tsutils.isThenableType(checker, originalNode.expression, type)) {
                        context.report({
                            messageId: 'await',
                            node,
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/ban-ts-comment.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.defaultMinimumDescriptionLength = 3;
    exports.default = util.createRule({
        name: 'ban-ts-comment',
        meta: {
            type: 'problem',
            docs: {
                description: 'Bans `@ts-<directive>` comments from being used or requires descriptions after directive',
                category: 'Best Practices',
                recommended: 'error',
            },
            messages: {
                tsDirectiveComment: 'Do not use "@ts-{{directive}}" because it alters compilation errors.',
                tsDirectiveCommentRequiresDescription: 'Include a description after the "@ts-{{directive}}" directive to explain why the @ts-{{directive}} is necessary. The description must be {{minimumDescriptionLength}} characters or longer.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        'ts-expect-error': {
                            oneOf: [
                                {
                                    type: 'boolean',
                                    default: true,
                                },
                                {
                                    enum: ['allow-with-description'],
                                },
                            ],
                        },
                        'ts-ignore': {
                            oneOf: [
                                {
                                    type: 'boolean',
                                    default: true,
                                },
                                {
                                    enum: ['allow-with-description'],
                                },
                            ],
                        },
                        'ts-nocheck': {
                            oneOf: [
                                {
                                    type: 'boolean',
                                    default: true,
                                },
                                {
                                    enum: ['allow-with-description'],
                                },
                            ],
                        },
                        'ts-check': {
                            oneOf: [
                                {
                                    type: 'boolean',
                                    default: true,
                                },
                                {
                                    enum: ['allow-with-description'],
                                },
                            ],
                        },
                        minimumDescriptionLength: {
                            type: 'number',
                            default: exports.defaultMinimumDescriptionLength,
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                'ts-expect-error': 'allow-with-description',
                'ts-ignore': true,
                'ts-nocheck': true,
                'ts-check': false,
                minimumDescriptionLength: exports.defaultMinimumDescriptionLength,
            },
        ],
        create(context, [options]) {
            /*
              The regex used are taken from the ones used in the official TypeScript repo -
              https://github.com/microsoft/TypeScript/blob/master/src/compiler/scanner.ts#L281-L289
            */
            const commentDirectiveRegExSingleLine = /^\/*\s*@ts-(expect-error|ignore|check|nocheck)(.*)/;
            const commentDirectiveRegExMultiLine = /^\s*(?:\/|\*)*\s*@ts-(expect-error|ignore|check|nocheck)(.*)/;
            const sourceCode = context.getSourceCode();
            return {
                Program() {
                    const comments = sourceCode.getAllComments();
                    comments.forEach(comment => {
                        var _a;
                        let regExp = commentDirectiveRegExSingleLine;
                        if (comment.type !== experimental_utils_1.AST_TOKEN_TYPES.Line) {
                            regExp = commentDirectiveRegExMultiLine;
                        }
                        const [, directive, description] = (_a = regExp.exec(comment.value)) !== null && _a !== void 0 ? _a : [];
                        const fullDirective = `ts-${directive}`;
                        const option = options[fullDirective];
                        if (option === true) {
                            context.report({
                                data: { directive },
                                node: comment,
                                messageId: 'tsDirectiveComment',
                            });
                        }
                        if (option === 'allow-with-description') {
                            const { minimumDescriptionLength = exports.defaultMinimumDescriptionLength, } = options;
                            if (description.trim().length < minimumDescriptionLength) {
                                context.report({
                                    data: { directive, minimumDescriptionLength },
                                    node: comment,
                                    messageId: 'tsDirectiveCommentRequiresDescription',
                                });
                            }
                        }
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/ban-tslint-comment.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    // tslint regex
    // https://github.com/palantir/tslint/blob/95d9d958833fd9dc0002d18cbe34db20d0fbf437/src/enableDisableRules.ts#L32
    const ENABLE_DISABLE_REGEX = /^\s*tslint:(enable|disable)(?:-(line|next-line))?(:|\s|$)/;
    const toText = (text, type) => type === experimental_utils_1.AST_TOKEN_TYPES.Line
        ? ['//', text.trim()].join(' ')
        : ['/*', text.trim(), '*/'].join(' ');
    exports.default = util.createRule({
        name: 'ban-tslint-comment',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Bans `// tslint:<rule-flag>` comments from being used',
                category: 'Stylistic Issues',
                recommended: false,
            },
            messages: {
                commentDetected: 'tslint comment detected: "{{ text }}"',
            },
            schema: [],
            fixable: 'code',
        },
        defaultOptions: [],
        create: context => {
            const sourceCode = context.getSourceCode();
            return {
                Program() {
                    const comments = sourceCode.getAllComments();
                    comments.forEach(c => {
                        if (ENABLE_DISABLE_REGEX.test(c.value)) {
                            context.report({
                                data: { text: toText(c.value, c.type) },
                                node: c,
                                messageId: 'commentDetected',
                                fix(fixer) {
                                    const rangeStart = sourceCode.getIndexFromLoc({
                                        column: c.loc.start.column > 0 ? c.loc.start.column - 1 : 0,
                                        line: c.loc.start.line,
                                    });
                                    const rangeEnd = sourceCode.getIndexFromLoc({
                                        column: c.loc.end.column,
                                        line: c.loc.end.line,
                                    });
                                    return fixer.removeRange([rangeStart, rangeEnd + 1]);
                                },
                            });
                        }
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/ban-types.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    function removeSpaces(str) {
        return str.replace(/ /g, '');
    }
    function stringifyNode(node, sourceCode) {
        return removeSpaces(sourceCode.getText(node));
    }
    function getCustomMessage(bannedType) {
        if (bannedType === null) {
            return '';
        }
        if (typeof bannedType === 'string') {
            return ` ${bannedType}`;
        }
        if (bannedType.message) {
            return ` ${bannedType.message}`;
        }
        return '';
    }
    const defaultTypes = {
        String: {
            message: 'Use string instead',
            fixWith: 'string',
        },
        Boolean: {
            message: 'Use boolean instead',
            fixWith: 'boolean',
        },
        Number: {
            message: 'Use number instead',
            fixWith: 'number',
        },
        Symbol: {
            message: 'Use symbol instead',
            fixWith: 'symbol',
        },
        Function: {
            message: [
                'The `Function` type accepts any function-like value.',
                'It provides no type safety when calling the function, which can be a common source of bugs.',
                'It also accepts things like class declarations, which will throw at runtime as they will not be called with `new`.',
                'If you are expecting the function to accept certain arguments, you should explicitly define the function shape.',
            ].join('\n'),
        },
        // object typing
        Object: {
            message: [
                'The `Object` type actually means "any non-nullish value", so it is marginally better than `unknown`.',
                '- If you want a type meaning "any object", you probably want `Record<string, unknown>` instead.',
                '- If you want a type meaning "any value", you probably want `unknown` instead.',
            ].join('\n'),
        },
        '{}': {
            message: [
                '`{}` actually means "any non-nullish value".',
                '- If you want a type meaning "any object", you probably want `Record<string, unknown>` instead.',
                '- If you want a type meaning "any value", you probably want `unknown` instead.',
                '- If you want a type meaning "empty object", you probably want `Record<string, never>` instead.',
            ].join('\n'),
        },
        object: {
            message: [
                'The `object` type is currently hard to use ([see this issue](https://github.com/microsoft/TypeScript/issues/21732)).',
                'Consider using `Record<string, unknown>` instead, as it allows you to more easily inspect and use the keys.',
            ].join('\n'),
        },
    };
    exports.TYPE_KEYWORDS = {
        bigint: experimental_utils_1.AST_NODE_TYPES.TSBigIntKeyword,
        boolean: experimental_utils_1.AST_NODE_TYPES.TSBooleanKeyword,
        never: experimental_utils_1.AST_NODE_TYPES.TSNeverKeyword,
        null: experimental_utils_1.AST_NODE_TYPES.TSNullKeyword,
        number: experimental_utils_1.AST_NODE_TYPES.TSNumberKeyword,
        object: experimental_utils_1.AST_NODE_TYPES.TSObjectKeyword,
        string: experimental_utils_1.AST_NODE_TYPES.TSStringKeyword,
        symbol: experimental_utils_1.AST_NODE_TYPES.TSSymbolKeyword,
        undefined: experimental_utils_1.AST_NODE_TYPES.TSUndefinedKeyword,
        unknown: experimental_utils_1.AST_NODE_TYPES.TSUnknownKeyword,
        void: experimental_utils_1.AST_NODE_TYPES.TSVoidKeyword,
    };
    exports.default = util.createRule({
        name: 'ban-types',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Bans specific types from being used',
                category: 'Best Practices',
                recommended: 'error',
            },
            fixable: 'code',
            messages: {
                bannedTypeMessage: "Don't use `{{name}}` as a type.{{customMessage}}",
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        types: {
                            type: 'object',
                            additionalProperties: {
                                oneOf: [
                                    { type: 'null' },
                                    { type: 'boolean' },
                                    { type: 'string' },
                                    {
                                        type: 'object',
                                        properties: {
                                            message: { type: 'string' },
                                            fixWith: { type: 'string' },
                                        },
                                        additionalProperties: false,
                                    },
                                ],
                            },
                        },
                        extendDefaults: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [{}],
        create(context, [options]) {
            var _a, _b;
            const extendDefaults = (_a = options.extendDefaults) !== null && _a !== void 0 ? _a : true;
            const customTypes = (_b = options.types) !== null && _b !== void 0 ? _b : {};
            const types = Object.assign({}, extendDefaults ? defaultTypes : {}, customTypes);
            const bannedTypes = new Map(Object.entries(types).map(([type, data]) => [removeSpaces(type), data]));
            function checkBannedTypes(typeNode, name = stringifyNode(typeNode, context.getSourceCode())) {
                const bannedType = bannedTypes.get(name);
                if (bannedType === undefined || bannedType === false) {
                    return;
                }
                const customMessage = getCustomMessage(bannedType);
                const fixWith = bannedType && typeof bannedType === 'object' && bannedType.fixWith;
                context.report({
                    node: typeNode,
                    messageId: 'bannedTypeMessage',
                    data: {
                        name,
                        customMessage,
                    },
                    fix: fixWith
                        ? (fixer) => fixer.replaceText(typeNode, fixWith)
                        : null,
                });
            }
            const keywordSelectors = util.objectReduceKey(exports.TYPE_KEYWORDS, (acc, keyword) => {
                if (bannedTypes.has(keyword)) {
                    acc[exports.TYPE_KEYWORDS[keyword]] = (node) => checkBannedTypes(node, keyword);
                }
                return acc;
            }, {});
            return Object.assign(Object.assign({}, keywordSelectors), { TSTypeLiteral(node) {
                    if (node.members.length) {
                        return;
                    }
                    checkBannedTypes(node);
                },
                TSTupleType(node) {
                    if (node.elementTypes.length === 0) {
                        checkBannedTypes(node);
                    }
                },
                TSTypeReference(node) {
                    checkBannedTypes(node.typeName);
                    if (node.typeParameters) {
                        checkBannedTypes(node);
                    }
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/brace-style.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const brace_style_1 = __importDefault(require(408) /* eslint/lib/rules/brace-style */);
    const util_1 = require(906) /* ../util */;
    exports.default = util_1.createRule({
        name: 'brace-style',
        meta: {
            type: 'layout',
            docs: {
                description: 'Enforce consistent brace style for blocks',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            },
            messages: brace_style_1.default.meta.messages,
            fixable: brace_style_1.default.meta.fixable,
            schema: brace_style_1.default.meta.schema,
        },
        defaultOptions: ['1tbs'],
        create(context) {
            const [style, { allowSingleLine } = { allowSingleLine: false }] = context.options;
            const isAllmanStyle = style === 'allman';
            const sourceCode = context.getSourceCode();
            const rules = brace_style_1.default.create(context);
            /**
             * Checks a pair of curly brackets based on the user's config
             */
            function validateCurlyPair(openingCurlyToken, closingCurlyToken) {
                if (allowSingleLine &&
                    util_1.isTokenOnSameLine(openingCurlyToken, closingCurlyToken)) {
                    return;
                }
                const tokenBeforeOpeningCurly = sourceCode.getTokenBefore(openingCurlyToken);
                const tokenBeforeClosingCurly = sourceCode.getTokenBefore(closingCurlyToken);
                const tokenAfterOpeningCurly = sourceCode.getTokenAfter(openingCurlyToken);
                if (!isAllmanStyle &&
                    !util_1.isTokenOnSameLine(tokenBeforeOpeningCurly, openingCurlyToken)) {
                    context.report({
                        node: openingCurlyToken,
                        messageId: 'nextLineOpen',
                        fix: fixer => {
                            const textRange = [
                                tokenBeforeOpeningCurly.range[1],
                                openingCurlyToken.range[0],
                            ];
                            const textBetween = sourceCode.text.slice(textRange[0], textRange[1]);
                            if (textBetween.trim()) {
                                return null;
                            }
                            return fixer.replaceTextRange(textRange, ' ');
                        },
                    });
                }
                if (isAllmanStyle &&
                    util_1.isTokenOnSameLine(tokenBeforeOpeningCurly, openingCurlyToken)) {
                    context.report({
                        node: openingCurlyToken,
                        messageId: 'sameLineOpen',
                        fix: fixer => fixer.insertTextBefore(openingCurlyToken, '\n'),
                    });
                }
                if (util_1.isTokenOnSameLine(openingCurlyToken, tokenAfterOpeningCurly) &&
                    tokenAfterOpeningCurly !== closingCurlyToken) {
                    context.report({
                        node: openingCurlyToken,
                        messageId: 'blockSameLine',
                        fix: fixer => fixer.insertTextAfter(openingCurlyToken, '\n'),
                    });
                }
                if (util_1.isTokenOnSameLine(tokenBeforeClosingCurly, closingCurlyToken) &&
                    tokenBeforeClosingCurly !== openingCurlyToken) {
                    context.report({
                        node: closingCurlyToken,
                        messageId: 'singleLineClose',
                        fix: fixer => fixer.insertTextBefore(closingCurlyToken, '\n'),
                    });
                }
            }
            return Object.assign(Object.assign({}, rules), { 'TSInterfaceBody, TSModuleBlock'(node) {
                    const openingCurly = sourceCode.getFirstToken(node);
                    const closingCurly = sourceCode.getLastToken(node);
                    validateCurlyPair(openingCurly, closingCurly);
                },
                TSEnumDeclaration(node) {
                    const closingCurly = sourceCode.getLastToken(node);
                    const openingCurly = sourceCode.getTokenBefore(node.members.length ? node.members[0] : closingCurly);
                    validateCurlyPair(openingCurly, closingCurly);
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/class-literal-property-style.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const printNodeModifiers = (node, final) => {
        var _a;
        return `${(_a = node.accessibility) !== null && _a !== void 0 ? _a : ''}${node.static ? ' static' : ''} ${final} `.trimLeft();
    };
    const isSupportedLiteral = (node) => {
        if (node.type === experimental_utils_1.AST_NODE_TYPES.Literal) {
            return true;
        }
        if (node.type === experimental_utils_1.AST_NODE_TYPES.TaggedTemplateExpression ||
            node.type === experimental_utils_1.AST_NODE_TYPES.TemplateLiteral) {
            return ('quasi' in node ? node.quasi.quasis : node.quasis).length === 1;
        }
        return false;
    };
    exports.default = util.createRule({
        name: 'class-literal-property-style',
        meta: {
            type: 'problem',
            docs: {
                description: 'Ensures that literals on classes are exposed in a consistent style',
                category: 'Best Practices',
                recommended: false,
            },
            fixable: 'code',
            messages: {
                preferFieldStyle: 'Literals should be exposed using readonly fields.',
                preferGetterStyle: 'Literals should be exposed using getters.',
            },
            schema: [{ enum: ['fields', 'getters'] }],
        },
        defaultOptions: ['fields'],
        create(context, [style]) {
            if (style === 'fields') {
                return {
                    MethodDefinition(node) {
                        if (node.kind !== 'get' ||
                            !node.value.body ||
                            !node.value.body.body.length) {
                            return;
                        }
                        const [statement] = node.value.body.body;
                        if (statement.type !== experimental_utils_1.AST_NODE_TYPES.ReturnStatement) {
                            return;
                        }
                        const { argument } = statement;
                        if (!argument || !isSupportedLiteral(argument)) {
                            return;
                        }
                        context.report({
                            node: node.key,
                            messageId: 'preferFieldStyle',
                            fix(fixer) {
                                const sourceCode = context.getSourceCode();
                                const name = sourceCode.getText(node.key);
                                let text = '';
                                text += printNodeModifiers(node, 'readonly');
                                text += node.computed ? `[${name}]` : name;
                                text += ` = ${sourceCode.getText(argument)};`;
                                return fixer.replaceText(node, text);
                            },
                        });
                    },
                };
            }
            return {
                ClassProperty(node) {
                    if (!node.readonly || node.declare) {
                        return;
                    }
                    const { value } = node;
                    if (!value || !isSupportedLiteral(value)) {
                        return;
                    }
                    context.report({
                        node: node.key,
                        messageId: 'preferGetterStyle',
                        fix(fixer) {
                            const sourceCode = context.getSourceCode();
                            const name = sourceCode.getText(node.key);
                            let text = '';
                            text += printNodeModifiers(node, 'get');
                            text += node.computed ? `[${name}]` : name;
                            text += `() { return ${sourceCode.getText(value)}; }`;
                            return fixer.replaceText(node, text);
                        },
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/comma-dangle.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    const comma_dangle_1 = __importDefault(require(414) /* eslint/lib/rules/comma-dangle */);
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const OPTION_VALUE_SCHEME = [
        'always-multiline',
        'always',
        'never',
        'only-multiline',
    ];
    const DEFAULT_OPTION_VALUE = 'never';
    function normalizeOptions(options) {
        var _a, _b, _c;
        if (typeof options === 'string') {
            return {
                enums: options,
                generics: options,
                tuples: options,
            };
        }
        return {
            enums: (_a = options.enums) !== null && _a !== void 0 ? _a : DEFAULT_OPTION_VALUE,
            generics: (_b = options.generics) !== null && _b !== void 0 ? _b : DEFAULT_OPTION_VALUE,
            tuples: (_c = options.tuples) !== null && _c !== void 0 ? _c : DEFAULT_OPTION_VALUE,
        };
    }
    exports.default = util.createRule({
        name: 'comma-dangle',
        meta: {
            type: 'layout',
            docs: {
                description: 'Require or disallow trailing comma',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: {
                definitions: {
                    value: {
                        enum: OPTION_VALUE_SCHEME,
                    },
                    valueWithIgnore: {
                        enum: [...OPTION_VALUE_SCHEME, 'ignore'],
                    },
                },
                type: 'array',
                items: [
                    {
                        oneOf: [
                            {
                                $ref: '#/definitions/value',
                            },
                            {
                                type: 'object',
                                properties: {
                                    arrays: { $ref: '#/definitions/valueWithIgnore' },
                                    objects: { $ref: '#/definitions/valueWithIgnore' },
                                    imports: { $ref: '#/definitions/valueWithIgnore' },
                                    exports: { $ref: '#/definitions/valueWithIgnore' },
                                    functions: { $ref: '#/definitions/valueWithIgnore' },
                                    enums: { $ref: '#/definitions/valueWithIgnore' },
                                    generics: { $ref: '#/definitions/valueWithIgnore' },
                                    tuples: { $ref: '#/definitions/valueWithIgnore' },
                                },
                                additionalProperties: false,
                            },
                        ],
                    },
                ],
            },
            fixable: 'code',
            messages: comma_dangle_1.default.meta.messages,
        },
        defaultOptions: ['never'],
        create(context, [options]) {
            const rules = comma_dangle_1.default.create(context);
            const sourceCode = context.getSourceCode();
            const normalizedOptions = normalizeOptions(options);
            const predicate = {
                always: forceComma,
                'always-multiline': forceCommaIfMultiline,
                'only-multiline': allowCommaIfMultiline,
                never: forbidComma,
                ignore: () => { },
            };
            function last(nodes) {
                var _a;
                return (_a = nodes[nodes.length - 1]) !== null && _a !== void 0 ? _a : null;
            }
            function getLastItem(node) {
                switch (node.type) {
                    case experimental_utils_1.AST_NODE_TYPES.TSEnumDeclaration:
                        return last(node.members);
                    case experimental_utils_1.AST_NODE_TYPES.TSTypeParameterDeclaration:
                        return last(node.params);
                    case experimental_utils_1.AST_NODE_TYPES.TSTupleType:
                        return last(node.elementTypes);
                    default:
                        return null;
                }
            }
            function getTrailingToken(node) {
                const last = getLastItem(node);
                const trailing = last && sourceCode.getTokenAfter(last);
                return trailing;
            }
            function isMultiline(node) {
                const last = getLastItem(node);
                const lastToken = sourceCode.getLastToken(node);
                return (last === null || last === void 0 ? void 0 : last.loc.end.line) !== (lastToken === null || lastToken === void 0 ? void 0 : lastToken.loc.end.line);
            }
            function forbidComma(node) {
                const last = getLastItem(node);
                const trailing = getTrailingToken(node);
                if (last && trailing && util.isCommaToken(trailing)) {
                    context.report({
                        node,
                        messageId: 'unexpected',
                        fix(fixer) {
                            return fixer.remove(trailing);
                        },
                    });
                }
            }
            function forceComma(node) {
                const last = getLastItem(node);
                const trailing = getTrailingToken(node);
                if (last && trailing && !util.isCommaToken(trailing)) {
                    context.report({
                        node,
                        messageId: 'missing',
                        fix(fixer) {
                            return fixer.insertTextAfter(last, ',');
                        },
                    });
                }
            }
            function allowCommaIfMultiline(node) {
                if (!isMultiline(node)) {
                    forbidComma(node);
                }
            }
            function forceCommaIfMultiline(node) {
                if (isMultiline(node)) {
                    forceComma(node);
                }
                else {
                    forbidComma(node);
                }
            }
            return Object.assign(Object.assign({}, rules), { TSEnumDeclaration: predicate[normalizedOptions.enums], TSTypeParameterDeclaration: predicate[normalizedOptions.generics], TSTupleType: predicate[normalizedOptions.tuples] });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/comma-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util_1 = require(906) /* ../util */;
    exports.default = util_1.createRule({
        name: 'comma-spacing',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforces consistent spacing before and after commas',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: 'whitespace',
            schema: [
                {
                    type: 'object',
                    properties: {
                        before: {
                            type: 'boolean',
                            default: false,
                        },
                        after: {
                            type: 'boolean',
                            default: true,
                        },
                    },
                    additionalProperties: false,
                },
            ],
            messages: {
                unexpected: `There should be no space {{loc}} ','.`,
                missing: `A space is required {{loc}} ','.`,
            },
        },
        defaultOptions: [
            {
                before: false,
                after: true,
            },
        ],
        create(context, [{ before: spaceBefore, after: spaceAfter }]) {
            const sourceCode = context.getSourceCode();
            const tokensAndComments = sourceCode.tokensAndComments;
            const ignoredTokens = new Set();
            /**
             * Adds null elements of the ArrayExpression or ArrayPattern node to the ignore list
             * @param node node to evaluate
             */
            function addNullElementsToIgnoreList(node) {
                let previousToken = sourceCode.getFirstToken(node);
                for (const element of node.elements) {
                    let token;
                    if (element === null) {
                        token = sourceCode.getTokenAfter(previousToken);
                        if (token && util_1.isCommaToken(token)) {
                            ignoredTokens.add(token);
                        }
                    }
                    else {
                        token = sourceCode.getTokenAfter(element);
                    }
                    previousToken = token;
                }
            }
            /**
             * Adds type parameters trailing comma token to the ignore list
             * @param node node to evaluate
             */
            function addTypeParametersTrailingCommaToIgnoreList(node) {
                const paramLength = node.params.length;
                if (paramLength) {
                    const param = node.params[paramLength - 1];
                    const afterToken = sourceCode.getTokenAfter(param);
                    if (afterToken && util_1.isCommaToken(afterToken)) {
                        ignoredTokens.add(afterToken);
                    }
                }
            }
            /**
             * Validates the spacing around a comma token.
             * @param commaToken The token representing the comma
             * @param prevToken The last token before the comma
             * @param nextToken The first token after the comma
             */
            function validateCommaSpacing(commaToken, prevToken, nextToken) {
                if (prevToken &&
                    util_1.isTokenOnSameLine(prevToken, commaToken) &&
                    spaceBefore !== sourceCode.isSpaceBetweenTokens(prevToken, commaToken)) {
                    context.report({
                        node: commaToken,
                        data: {
                            loc: 'before',
                        },
                        messageId: spaceBefore ? 'missing' : 'unexpected',
                        fix: fixer => spaceBefore
                            ? fixer.insertTextBefore(commaToken, ' ')
                            : fixer.replaceTextRange([prevToken.range[1], commaToken.range[0]], ''),
                    });
                }
                if (nextToken && util_1.isClosingParenToken(nextToken)) {
                    return;
                }
                if (!spaceAfter && nextToken && nextToken.type === experimental_utils_1.AST_TOKEN_TYPES.Line) {
                    return;
                }
                if (nextToken &&
                    util_1.isTokenOnSameLine(commaToken, nextToken) &&
                    spaceAfter !== sourceCode.isSpaceBetweenTokens(commaToken, nextToken)) {
                    context.report({
                        node: commaToken,
                        data: {
                            loc: 'after',
                        },
                        messageId: spaceAfter ? 'missing' : 'unexpected',
                        fix: fixer => spaceAfter
                            ? fixer.insertTextAfter(commaToken, ' ')
                            : fixer.replaceTextRange([commaToken.range[1], nextToken.range[0]], ''),
                    });
                }
            }
            return {
                TSTypeParameterDeclaration: addTypeParametersTrailingCommaToIgnoreList,
                ArrayExpression: addNullElementsToIgnoreList,
                ArrayPattern: addNullElementsToIgnoreList,
                'Program:exit'() {
                    tokensAndComments.forEach((token, i) => {
                        if (!util_1.isCommaToken(token)) {
                            return;
                        }
                        const prevToken = tokensAndComments[i - 1];
                        const nextToken = tokensAndComments[i + 1];
                        validateCommaSpacing(token, util_1.isCommaToken(prevToken) || ignoredTokens.has(token)
                            ? null
                            : prevToken, util_1.isCommaToken(nextToken) || ignoredTokens.has(token)
                            ? null
                            : nextToken);
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/consistent-indexed-object-style.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const util_1 = require(906) /* ../util */;
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    exports.default = util_1.createRule({
        name: 'consistent-indexed-object-style',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforce or disallow the use of the record type',
                category: 'Stylistic Issues',
                // too opinionated to be recommended
                recommended: false,
            },
            messages: {
                preferRecord: 'A record is preferred over an index signature',
                preferIndexSignature: 'An index signature is preferred over a record.',
            },
            fixable: 'code',
            schema: [
                {
                    enum: ['record', 'index-signature'],
                },
            ],
        },
        defaultOptions: ['record'],
        create(context) {
            const sourceCode = context.getSourceCode();
            if (context.options[0] === 'index-signature') {
                return {
                    TSTypeReference(node) {
                        var _a;
                        const typeName = node.typeName;
                        if (typeName.type !== experimental_utils_1.AST_NODE_TYPES.Identifier) {
                            return;
                        }
                        if (typeName.name !== 'Record') {
                            return;
                        }
                        const params = (_a = node.typeParameters) === null || _a === void 0 ? void 0 : _a.params;
                        if ((params === null || params === void 0 ? void 0 : params.length) !== 2) {
                            return;
                        }
                        context.report({
                            node,
                            messageId: 'preferIndexSignature',
                            fix(fixer) {
                                const key = sourceCode.getText(params[0]);
                                const type = sourceCode.getText(params[1]);
                                return fixer.replaceText(node, `{ [key: ${key}]: ${type} }`);
                            },
                        });
                    },
                };
            }
            function checkMembers(members, node, prefix, postfix, safeFix = true) {
                if (members.length !== 1) {
                    return;
                }
                const [member] = members;
                if (member.type !== experimental_utils_1.AST_NODE_TYPES.TSIndexSignature) {
                    return;
                }
                const [parameter] = member.parameters;
                if (!parameter) {
                    return;
                }
                if (parameter.type !== experimental_utils_1.AST_NODE_TYPES.Identifier) {
                    return;
                }
                const keyType = parameter.typeAnnotation;
                if (!keyType) {
                    return;
                }
                const valueType = member.typeAnnotation;
                if (!valueType) {
                    return;
                }
                context.report({
                    node,
                    messageId: 'preferRecord',
                    fix: safeFix
                        ? (fixer) => {
                            const key = sourceCode.getText(keyType.typeAnnotation);
                            const value = sourceCode.getText(valueType.typeAnnotation);
                            const record = member.readonly
                                ? `Readonly<Record<${key}, ${value}>>`
                                : `Record<${key}, ${value}>`;
                            return fixer.replaceText(node, `${prefix}${record}${postfix}`);
                        }
                        : null,
                });
            }
            return {
                TSTypeLiteral(node) {
                    checkMembers(node.members, node, '', '');
                },
                TSInterfaceDeclaration(node) {
                    var _a, _b, _c, _d;
                    let genericTypes = '';
                    if (((_b = (_a = node.typeParameters) === null || _a === void 0 ? void 0 : _a.params) !== null && _b !== void 0 ? _b : []).length > 0) {
                        genericTypes = `<${(_c = node.typeParameters) === null || _c === void 0 ? void 0 : _c.params.map(p => p.name.name).join(', ')}>`;
                    }
                    checkMembers(node.body.body, node, `type ${node.id.name}${genericTypes} = `, ';', !((_d = node.extends) === null || _d === void 0 ? void 0 : _d.length));
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/consistent-type-assertions.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    exports.default = util.createRule({
        name: 'consistent-type-assertions',
        meta: {
            type: 'suggestion',
            docs: {
                category: 'Best Practices',
                description: 'Enforces consistent usage of type assertions',
                recommended: false,
            },
            messages: {
                as: "Use 'as {{cast}}' instead of '<{{cast}}>'.",
                'angle-bracket': "Use '<{{cast}}>' instead of 'as {{cast}}'.",
                never: 'Do not use any type assertions.',
                unexpectedObjectTypeAssertion: 'Always prefer const x: T = { ... }.',
            },
            schema: [
                {
                    oneOf: [
                        {
                            type: 'object',
                            properties: {
                                assertionStyle: {
                                    enum: ['never'],
                                },
                            },
                            additionalProperties: false,
                            required: ['assertionStyle'],
                        },
                        {
                            type: 'object',
                            properties: {
                                assertionStyle: {
                                    enum: ['as', 'angle-bracket'],
                                },
                                objectLiteralTypeAssertions: {
                                    enum: ['allow', 'allow-as-parameter', 'never'],
                                },
                            },
                            additionalProperties: false,
                            required: ['assertionStyle'],
                        },
                    ],
                },
            ],
        },
        defaultOptions: [
            {
                assertionStyle: 'as',
                objectLiteralTypeAssertions: 'allow',
            },
        ],
        create(context, [options]) {
            const sourceCode = context.getSourceCode();
            function isConst(node) {
                if (node.type !== experimental_utils_1.AST_NODE_TYPES.TSTypeReference) {
                    return false;
                }
                return (node.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    node.typeName.name === 'const');
            }
            function reportIncorrectAssertionType(node) {
                // If this node is `as const`, then don't report an error.
                if (isConst(node.typeAnnotation)) {
                    return;
                }
                const messageId = options.assertionStyle;
                context.report({
                    node,
                    messageId,
                    data: messageId !== 'never'
                        ? { cast: sourceCode.getText(node.typeAnnotation) }
                        : {},
                });
            }
            function checkType(node) {
                switch (node.type) {
                    case experimental_utils_1.AST_NODE_TYPES.TSAnyKeyword:
                    case experimental_utils_1.AST_NODE_TYPES.TSUnknownKeyword:
                        return false;
                    case experimental_utils_1.AST_NODE_TYPES.TSTypeReference:
                        return (
                        // Ignore `as const` and `<const>`
                        !isConst(node) ||
                            // Allow qualified names which have dots between identifiers, `Foo.Bar`
                            node.typeName.type === experimental_utils_1.AST_NODE_TYPES.TSQualifiedName);
                    default:
                        return true;
                }
            }
            function checkExpression(node) {
                if (options.assertionStyle === 'never' ||
                    options.objectLiteralTypeAssertions === 'allow' ||
                    node.expression.type !== experimental_utils_1.AST_NODE_TYPES.ObjectExpression) {
                    return;
                }
                if (options.objectLiteralTypeAssertions === 'allow-as-parameter' &&
                    node.parent &&
                    (node.parent.type === experimental_utils_1.AST_NODE_TYPES.NewExpression ||
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.CallExpression ||
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.ThrowStatement ||
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.AssignmentPattern ||
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.JSXExpressionContainer)) {
                    return;
                }
                if (checkType(node.typeAnnotation) &&
                    node.expression.type === experimental_utils_1.AST_NODE_TYPES.ObjectExpression) {
                    context.report({
                        node,
                        messageId: 'unexpectedObjectTypeAssertion',
                    });
                }
            }
            return {
                TSTypeAssertion(node) {
                    if (options.assertionStyle !== 'angle-bracket') {
                        reportIncorrectAssertionType(node);
                        return;
                    }
                    checkExpression(node);
                },
                TSAsExpression(node) {
                    if (options.assertionStyle !== 'as') {
                        reportIncorrectAssertionType(node);
                        return;
                    }
                    checkExpression(node);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/consistent-type-definitions.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'consistent-type-definitions',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Consistent with type definition either `interface` or `type`',
                category: 'Stylistic Issues',
                // too opinionated to be recommended
                recommended: false,
            },
            messages: {
                interfaceOverType: 'Use an `interface` instead of a `type`.',
                typeOverInterface: 'Use a `type` instead of an `interface`.',
            },
            schema: [
                {
                    enum: ['interface', 'type'],
                },
            ],
            fixable: 'code',
        },
        defaultOptions: ['interface'],
        create(context, [option]) {
            const sourceCode = context.getSourceCode();
            /**
             * Iterates from the highest parent to the currently traversed node
             * to determine whether any node in tree is globally declared module declaration
             */
            function isCurrentlyTraversedNodeWithinModuleDeclaration() {
                return context
                    .getAncestors()
                    .some(node => node.type === experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration &&
                    node.declare &&
                    node.global);
            }
            return {
                "TSTypeAliasDeclaration[typeAnnotation.type='TSTypeLiteral']"(node) {
                    if (option === 'interface') {
                        context.report({
                            node: node.id,
                            messageId: 'interfaceOverType',
                            fix(fixer) {
                                var _a;
                                const typeNode = (_a = node.typeParameters) !== null && _a !== void 0 ? _a : node.id;
                                const fixes = [];
                                const firstToken = sourceCode.getFirstToken(node);
                                if (firstToken) {
                                    fixes.push(fixer.replaceText(firstToken, 'interface'));
                                    fixes.push(fixer.replaceTextRange([typeNode.range[1], node.typeAnnotation.range[0]], ' '));
                                }
                                const afterToken = sourceCode.getTokenAfter(node.typeAnnotation);
                                if (afterToken &&
                                    afterToken.type === experimental_utils_1.AST_TOKEN_TYPES.Punctuator &&
                                    afterToken.value === ';') {
                                    fixes.push(fixer.remove(afterToken));
                                }
                                return fixes;
                            },
                        });
                    }
                },
                TSInterfaceDeclaration(node) {
                    if (option === 'type') {
                        context.report({
                            node: node.id,
                            messageId: 'typeOverInterface',
                            /**
                             * remove automatically fix when the interface is within a declare global
                             * @see {@link https://github.com/typescript-eslint/typescript-eslint/issues/2707}
                             */
                            fix: isCurrentlyTraversedNodeWithinModuleDeclaration()
                                ? null
                                : (fixer) => {
                                    var _a;
                                    const typeNode = (_a = node.typeParameters) !== null && _a !== void 0 ? _a : node.id;
                                    const fixes = [];
                                    const firstToken = sourceCode.getFirstToken(node);
                                    if (firstToken) {
                                        fixes.push(fixer.replaceText(firstToken, 'type'));
                                        fixes.push(fixer.replaceTextRange([typeNode.range[1], node.body.range[0]], ' = '));
                                    }
                                    if (node.extends) {
                                        node.extends.forEach(heritage => {
                                            const typeIdentifier = sourceCode.getText(heritage);
                                            fixes.push(fixer.insertTextAfter(node.body, ` & ${typeIdentifier}`));
                                        });
                                    }
                                    return fixes;
                                },
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/consistent-type-imports.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    function isImportToken(token) {
        return token.type === experimental_utils_1.AST_TOKEN_TYPES.Keyword && token.value === 'import';
    }
    function isTypeToken(token) {
        return token.type === experimental_utils_1.AST_TOKEN_TYPES.Identifier && token.value === 'type';
    }
    exports.default = util.createRule({
        name: 'consistent-type-imports',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforces consistent usage of type imports',
                category: 'Stylistic Issues',
                recommended: false,
            },
            messages: {
                typeOverValue: 'All imports in the declaration are only used as types. Use `import type`',
                someImportsAreOnlyTypes: 'Imports {{typeImports}} are only used as types',
                aImportIsOnlyTypes: 'Import {{typeImports}} is only used as types',
                someImportsInDecoMeta: 'Type imports {{typeImports}} are used by decorator metadata',
                aImportInDecoMeta: 'Type import {{typeImports}} is used by decorator metadata',
                valueOverType: 'Use an `import` instead of an `import type`.',
                noImportTypeAnnotations: '`import()` type annotations are forbidden.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        prefer: {
                            enum: ['type-imports', 'no-type-imports'],
                        },
                        disallowTypeAnnotations: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
            fixable: 'code',
        },
        defaultOptions: [
            {
                prefer: 'type-imports',
                disallowTypeAnnotations: true,
            },
        ],
        create(context, [option]) {
            var _a;
            const prefer = (_a = option.prefer) !== null && _a !== void 0 ? _a : 'type-imports';
            const disallowTypeAnnotations = option.disallowTypeAnnotations !== false;
            const sourceCode = context.getSourceCode();
            const sourceImportsMap = {};
            return Object.assign(Object.assign({}, (prefer === 'type-imports'
                ? {
                    // prefer type imports
                    ImportDeclaration(node) {
                        var _a;
                        const source = node.source.value;
                        const sourceImports = (_a = sourceImportsMap[source]) !== null && _a !== void 0 ? _a : (sourceImportsMap[source] = {
                            source,
                            reportValueImports: [],
                            typeOnlyNamedImport: null,
                            valueOnlyNamedImport: null,
                        });
                        if (node.importKind === 'type') {
                            if (!sourceImports.typeOnlyNamedImport &&
                                node.specifiers.every(specifier => specifier.type === experimental_utils_1.AST_NODE_TYPES.ImportSpecifier)) {
                                sourceImports.typeOnlyNamedImport = node;
                            }
                        }
                        else {
                            if (!sourceImports.valueOnlyNamedImport &&
                                node.specifiers.every(specifier => specifier.type === experimental_utils_1.AST_NODE_TYPES.ImportSpecifier)) {
                                sourceImports.valueOnlyNamedImport = node;
                            }
                        }
                        const typeSpecifiers = [];
                        const valueSpecifiers = [];
                        const unusedSpecifiers = [];
                        for (const specifier of node.specifiers) {
                            const [variable] = context.getDeclaredVariables(specifier);
                            if (variable.references.length === 0) {
                                unusedSpecifiers.push(specifier);
                            }
                            else {
                                const onlyHasTypeReferences = variable.references.every(ref => {
                                    var _a, _b;
                                    /**
                                     * keep origin import kind when export
                                     * export { Type }
                                     * export default Type;
                                     */
                                    if (((_a = ref.identifier.parent) === null || _a === void 0 ? void 0 : _a.type) ===
                                        experimental_utils_1.AST_NODE_TYPES.ExportSpecifier ||
                                        ((_b = ref.identifier.parent) === null || _b === void 0 ? void 0 : _b.type) ===
                                            experimental_utils_1.AST_NODE_TYPES.ExportDefaultDeclaration) {
                                        if (ref.isValueReference && ref.isTypeReference) {
                                            return node.importKind === 'type';
                                        }
                                    }
                                    if (ref.isValueReference) {
                                        let parent = ref.identifier.parent;
                                        let child = ref.identifier;
                                        while (parent) {
                                            switch (parent.type) {
                                                // CASE 1:
                                                // `type T = typeof foo` will create a value reference because "foo" must be a value type
                                                // however this value reference is safe to use with type-only imports
                                                case experimental_utils_1.AST_NODE_TYPES.TSTypeQuery:
                                                    return true;
                                                case experimental_utils_1.AST_NODE_TYPES.TSQualifiedName:
                                                    // TSTypeQuery must have a TSESTree.EntityName as its child, so we can filter here and break early
                                                    if (parent.left !== child) {
                                                        return false;
                                                    }
                                                    child = parent;
                                                    parent = parent.parent;
                                                    continue;
                                                // END CASE 1
                                                //////////////
                                                // CASE 2:
                                                // `type T = { [foo]: string }` will create a value reference because "foo" must be a value type
                                                // however this value reference is safe to use with type-only imports.
                                                // Also this is represented as a non-type AST - hence it uses MemberExpression
                                                case experimental_utils_1.AST_NODE_TYPES.TSPropertySignature:
                                                    return parent.key === child;
                                                case experimental_utils_1.AST_NODE_TYPES.MemberExpression:
                                                    if (parent.object !== child) {
                                                        return false;
                                                    }
                                                    child = parent;
                                                    parent = parent.parent;
                                                    continue;
                                                // END CASE 2
                                                default:
                                                    return false;
                                            }
                                        }
                                    }
                                    return ref.isTypeReference;
                                });
                                if (onlyHasTypeReferences) {
                                    typeSpecifiers.push(specifier);
                                }
                                else {
                                    valueSpecifiers.push(specifier);
                                }
                            }
                        }
                        if ((node.importKind === 'value' && typeSpecifiers.length) ||
                            (node.importKind === 'type' && valueSpecifiers.length)) {
                            sourceImports.reportValueImports.push({
                                node,
                                typeSpecifiers,
                                valueSpecifiers,
                                unusedSpecifiers,
                            });
                        }
                    },
                    'Program:exit'() {
                        for (const sourceImports of Object.values(sourceImportsMap)) {
                            if (sourceImports.reportValueImports.length === 0) {
                                continue;
                            }
                            for (const report of sourceImports.reportValueImports) {
                                if (report.valueSpecifiers.length === 0 &&
                                    report.unusedSpecifiers.length === 0) {
                                    // import is all type-only, convert the entire import to `import type`
                                    context.report({
                                        node: report.node,
                                        messageId: 'typeOverValue',
                                        *fix(fixer) {
                                            yield* fixToTypeImport(fixer, report, sourceImports);
                                        },
                                    });
                                }
                                else {
                                    const isTypeImport = report.node.importKind === 'type';
                                    // we have a mixed type/value import, so we need to split them out into multiple exports
                                    const importNames = (isTypeImport
                                        ? report.valueSpecifiers
                                        : report.typeSpecifiers).map(specifier => `"${specifier.local.name}"`);
                                    const message = (() => {
                                        if (importNames.length === 1) {
                                            const typeImports = importNames[0];
                                            if (isTypeImport) {
                                                return {
                                                    messageId: 'aImportInDecoMeta',
                                                    data: { typeImports },
                                                };
                                            }
                                            else {
                                                return {
                                                    messageId: 'aImportIsOnlyTypes',
                                                    data: { typeImports },
                                                };
                                            }
                                        }
                                        else {
                                            const typeImports = [
                                                importNames.slice(0, -1).join(', '),
                                                importNames.slice(-1)[0],
                                            ].join(' and ');
                                            if (isTypeImport) {
                                                return {
                                                    messageId: 'someImportsInDecoMeta',
                                                    data: { typeImports },
                                                };
                                            }
                                            else {
                                                return {
                                                    messageId: 'someImportsAreOnlyTypes',
                                                    data: { typeImports },
                                                };
                                            }
                                        }
                                    })();
                                    context.report(Object.assign(Object.assign({ node: report.node }, message), { *fix(fixer) {
                                            if (isTypeImport) {
                                                yield* fixToValueImportInDecoMeta(fixer, report, sourceImports);
                                            }
                                            else {
                                                yield* fixToTypeImport(fixer, report, sourceImports);
                                            }
                                        } }));
                                }
                            }
                        }
                    },
                }
                : {
                    // prefer no type imports
                    'ImportDeclaration[importKind = "type"]'(node) {
                        context.report({
                            node,
                            messageId: 'valueOverType',
                            fix(fixer) {
                                return fixToValueImport(fixer, node);
                            },
                        });
                    },
                })), (disallowTypeAnnotations
                ? {
                    // disallow `import()` type
                    TSImportType(node) {
                        context.report({
                            node,
                            messageId: 'noImportTypeAnnotations',
                        });
                    },
                }
                : {}));
            function classifySpecifier(node) {
                var _a;
                const defaultSpecifier = node.specifiers[0].type === experimental_utils_1.AST_NODE_TYPES.ImportDefaultSpecifier
                    ? node.specifiers[0]
                    : null;
                const namespaceSpecifier = (_a = node.specifiers.find((specifier) => specifier.type === experimental_utils_1.AST_NODE_TYPES.ImportNamespaceSpecifier)) !== null && _a !== void 0 ? _a : null;
                const namedSpecifiers = node.specifiers.filter((specifier) => specifier.type === experimental_utils_1.AST_NODE_TYPES.ImportSpecifier);
                return {
                    defaultSpecifier,
                    namespaceSpecifier,
                    namedSpecifiers,
                };
            }
            /**
             * Returns information for fixing named specifiers.
             */
            function getFixesNamedSpecifiers(fixer, node, typeNamedSpecifiers, allNamedSpecifiers) {
                if (allNamedSpecifiers.length === 0) {
                    return {
                        typeNamedSpecifiersText: '',
                        removeTypeNamedSpecifiers: [],
                    };
                }
                const typeNamedSpecifiersTexts = [];
                const removeTypeNamedSpecifiers = [];
                if (typeNamedSpecifiers.length === allNamedSpecifiers.length) {
                    // e.g.
                    // import Foo, {Type1, Type2} from 'foo'
                    // import DefType, {Type1, Type2} from 'foo'
                    const openingBraceToken = util.nullThrows(sourceCode.getTokenBefore(typeNamedSpecifiers[0], util.isOpeningBraceToken), util.NullThrowsReasons.MissingToken('{', node.type));
                    const commaToken = util.nullThrows(sourceCode.getTokenBefore(openingBraceToken, util.isCommaToken), util.NullThrowsReasons.MissingToken(',', node.type));
                    const closingBraceToken = util.nullThrows(sourceCode.getFirstTokenBetween(openingBraceToken, node.source, util.isClosingBraceToken), util.NullThrowsReasons.MissingToken('}', node.type));
                    // import DefType, {...} from 'foo'
                    //               ^^^^^^^ remove
                    removeTypeNamedSpecifiers.push(fixer.removeRange([commaToken.range[0], closingBraceToken.range[1]]));
                    typeNamedSpecifiersTexts.push(sourceCode.text.slice(openingBraceToken.range[1], closingBraceToken.range[0]));
                }
                else {
                    const typeNamedSpecifierGroups = [];
                    let group = [];
                    for (const namedSpecifier of allNamedSpecifiers) {
                        if (typeNamedSpecifiers.includes(namedSpecifier)) {
                            group.push(namedSpecifier);
                        }
                        else if (group.length) {
                            typeNamedSpecifierGroups.push(group);
                            group = [];
                        }
                    }
                    if (group.length) {
                        typeNamedSpecifierGroups.push(group);
                    }
                    for (const namedSpecifiers of typeNamedSpecifierGroups) {
                        const { removeRange, textRange } = getNamedSpecifierRanges(namedSpecifiers, allNamedSpecifiers);
                        removeTypeNamedSpecifiers.push(fixer.removeRange(removeRange));
                        typeNamedSpecifiersTexts.push(sourceCode.text.slice(...textRange));
                    }
                }
                return {
                    typeNamedSpecifiersText: typeNamedSpecifiersTexts.join(','),
                    removeTypeNamedSpecifiers,
                };
            }
            /**
             * Returns ranges for fixing named specifier.
             */
            function getNamedSpecifierRanges(namedSpecifierGroup, allNamedSpecifiers) {
                const first = namedSpecifierGroup[0];
                const last = namedSpecifierGroup[namedSpecifierGroup.length - 1];
                const removeRange = [first.range[0], last.range[1]];
                const textRange = [...removeRange];
                const before = sourceCode.getTokenBefore(first);
                textRange[0] = before.range[1];
                if (util.isCommaToken(before)) {
                    removeRange[0] = before.range[0];
                }
                else {
                    removeRange[0] = before.range[1];
                }
                const isFirst = allNamedSpecifiers[0] === first;
                const isLast = allNamedSpecifiers[allNamedSpecifiers.length - 1] === last;
                const after = sourceCode.getTokenAfter(last);
                textRange[1] = after.range[0];
                if (isFirst || isLast) {
                    if (util.isCommaToken(after)) {
                        removeRange[1] = after.range[1];
                    }
                }
                return {
                    textRange,
                    removeRange,
                };
            }
            /**
             * insert specifiers to named import node.
             * e.g.
             * import type { Already, Type1, Type2 } from 'foo'
             *                        ^^^^^^^^^^^^^ insert
             */
            function insertToNamedImport(fixer, target, insertText) {
                const closingBraceToken = util.nullThrows(sourceCode.getFirstTokenBetween(sourceCode.getFirstToken(target), target.source, util.isClosingBraceToken), util.NullThrowsReasons.MissingToken('}', target.type));
                const before = sourceCode.getTokenBefore(closingBraceToken);
                if (!util.isCommaToken(before) && !util.isOpeningBraceToken(before)) {
                    insertText = ',' + insertText;
                }
                return fixer.insertTextBefore(closingBraceToken, insertText);
            }
            function* fixToTypeImport(fixer, report, sourceImports) {
                const { node } = report;
                const { defaultSpecifier, namespaceSpecifier, namedSpecifiers } = classifySpecifier(node);
                if (namespaceSpecifier && !defaultSpecifier) {
                    // e.g.
                    // import * as types from 'foo'
                    yield* fixToTypeImportByInsertType(fixer, node, false);
                    return;
                }
                else if (defaultSpecifier) {
                    if (report.typeSpecifiers.includes(defaultSpecifier) &&
                        namedSpecifiers.length === 0 &&
                        !namespaceSpecifier) {
                        // e.g.
                        // import Type from 'foo'
                        yield* fixToTypeImportByInsertType(fixer, node, true);
                        return;
                    }
                }
                else {
                    if (namedSpecifiers.every(specifier => report.typeSpecifiers.includes(specifier)) &&
                        !namespaceSpecifier) {
                        // e.g.
                        // import {Type1, Type2} from 'foo'
                        yield* fixToTypeImportByInsertType(fixer, node, false);
                        return;
                    }
                }
                const typeNamedSpecifiers = namedSpecifiers.filter(specifier => report.typeSpecifiers.includes(specifier));
                const fixesNamedSpecifiers = getFixesNamedSpecifiers(fixer, node, typeNamedSpecifiers, namedSpecifiers);
                const afterFixes = [];
                if (typeNamedSpecifiers.length) {
                    if (sourceImports.typeOnlyNamedImport) {
                        const insertTypeNamedSpecifiers = insertToNamedImport(fixer, sourceImports.typeOnlyNamedImport, fixesNamedSpecifiers.typeNamedSpecifiersText);
                        if (sourceImports.typeOnlyNamedImport.range[1] <= node.range[0]) {
                            yield insertTypeNamedSpecifiers;
                        }
                        else {
                            afterFixes.push(insertTypeNamedSpecifiers);
                        }
                    }
                    else {
                        yield fixer.insertTextBefore(node, `import type {${fixesNamedSpecifiers.typeNamedSpecifiersText}} from ${sourceCode.getText(node.source)};\n`);
                    }
                }
                const fixesRemoveTypeNamespaceSpecifier = [];
                if (namespaceSpecifier &&
                    report.typeSpecifiers.includes(namespaceSpecifier)) {
                    // e.g.
                    // import Foo, * as Type from 'foo'
                    // import DefType, * as Type from 'foo'
                    // import DefType, * as Type from 'foo'
                    const commaToken = util.nullThrows(sourceCode.getTokenBefore(namespaceSpecifier, util.isCommaToken), util.NullThrowsReasons.MissingToken(',', node.type));
                    // import Def, * as Ns from 'foo'
                    //           ^^^^^^^^^ remove
                    fixesRemoveTypeNamespaceSpecifier.push(fixer.removeRange([commaToken.range[0], namespaceSpecifier.range[1]]));
                    // import type * as Ns from 'foo'
                    // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ insert
                    yield fixer.insertTextBefore(node, `import type ${sourceCode.getText(namespaceSpecifier)} from ${sourceCode.getText(node.source)};\n`);
                }
                if (defaultSpecifier &&
                    report.typeSpecifiers.includes(defaultSpecifier)) {
                    if (report.typeSpecifiers.length === node.specifiers.length) {
                        const importToken = util.nullThrows(sourceCode.getFirstToken(node, isImportToken), util.NullThrowsReasons.MissingToken('import', node.type));
                        // import type Type from 'foo'
                        //        ^^^^ insert
                        yield fixer.insertTextAfter(importToken, ' type');
                    }
                    else {
                        const commaToken = util.nullThrows(sourceCode.getTokenAfter(defaultSpecifier, util.isCommaToken), util.NullThrowsReasons.MissingToken(',', defaultSpecifier.type));
                        // import Type , {...} from 'foo'
                        //        ^^^^^ pick
                        const defaultText = sourceCode.text
                            .slice(defaultSpecifier.range[0], commaToken.range[0])
                            .trim();
                        yield fixer.insertTextBefore(node, `import type ${defaultText} from ${sourceCode.getText(node.source)};\n`);
                        const afterToken = util.nullThrows(sourceCode.getTokenAfter(commaToken, { includeComments: true }), util.NullThrowsReasons.MissingToken('any token', node.type));
                        // import Type , {...} from 'foo'
                        //        ^^^^^^^ remove
                        yield fixer.removeRange([
                            defaultSpecifier.range[0],
                            afterToken.range[0],
                        ]);
                    }
                }
                yield* fixesNamedSpecifiers.removeTypeNamedSpecifiers;
                yield* fixesRemoveTypeNamespaceSpecifier;
                yield* afterFixes;
            }
            function* fixToTypeImportByInsertType(fixer, node, isDefaultImport) {
                // import type Foo from 'foo'
                //       ^^^^^ insert
                const importToken = util.nullThrows(sourceCode.getFirstToken(node, isImportToken), util.NullThrowsReasons.MissingToken('import', node.type));
                yield fixer.insertTextAfter(importToken, ' type');
                if (isDefaultImport) {
                    // Has default import
                    const openingBraceToken = sourceCode.getFirstTokenBetween(importToken, node.source, util.isOpeningBraceToken);
                    if (openingBraceToken) {
                        // Only braces. e.g. import Foo, {} from 'foo'
                        const commaToken = util.nullThrows(sourceCode.getTokenBefore(openingBraceToken, util.isCommaToken), util.NullThrowsReasons.MissingToken(',', node.type));
                        const closingBraceToken = util.nullThrows(sourceCode.getFirstTokenBetween(openingBraceToken, node.source, util.isClosingBraceToken), util.NullThrowsReasons.MissingToken('}', node.type));
                        // import type Foo, {} from 'foo'
                        //                  ^^ remove
                        yield fixer.removeRange([
                            commaToken.range[0],
                            closingBraceToken.range[1],
                        ]);
                        const specifiersText = sourceCode.text.slice(commaToken.range[1], closingBraceToken.range[1]);
                        if (node.specifiers.length > 1) {
                            // import type Foo from 'foo'
                            // import type {...} from 'foo' // <- insert
                            yield fixer.insertTextAfter(node, `\nimport type${specifiersText} from ${sourceCode.getText(node.source)};`);
                        }
                    }
                }
            }
            function* fixToValueImportInDecoMeta(fixer, report, sourceImports) {
                const { node } = report;
                const { defaultSpecifier, namespaceSpecifier, namedSpecifiers } = classifySpecifier(node);
                if (namespaceSpecifier) {
                    // e.g.
                    // import type * as types from 'foo'
                    yield* fixToValueImport(fixer, node);
                    return;
                }
                else if (defaultSpecifier) {
                    if (report.valueSpecifiers.includes(defaultSpecifier) &&
                        namedSpecifiers.length === 0) {
                        // e.g.
                        // import type Type from 'foo'
                        yield* fixToValueImport(fixer, node);
                        return;
                    }
                }
                else {
                    if (namedSpecifiers.every(specifier => report.valueSpecifiers.includes(specifier))) {
                        // e.g.
                        // import type {Type1, Type2} from 'foo'
                        yield* fixToValueImport(fixer, node);
                        return;
                    }
                }
                const valueNamedSpecifiers = namedSpecifiers.filter(specifier => report.valueSpecifiers.includes(specifier));
                const fixesNamedSpecifiers = getFixesNamedSpecifiers(fixer, node, valueNamedSpecifiers, namedSpecifiers);
                const afterFixes = [];
                if (valueNamedSpecifiers.length) {
                    if (sourceImports.valueOnlyNamedImport) {
                        const insertTypeNamedSpecifiers = insertToNamedImport(fixer, sourceImports.valueOnlyNamedImport, fixesNamedSpecifiers.typeNamedSpecifiersText);
                        if (sourceImports.valueOnlyNamedImport.range[1] <= node.range[0]) {
                            yield insertTypeNamedSpecifiers;
                        }
                        else {
                            afterFixes.push(insertTypeNamedSpecifiers);
                        }
                    }
                    else {
                        yield fixer.insertTextBefore(node, `import {${fixesNamedSpecifiers.typeNamedSpecifiersText}} from ${sourceCode.getText(node.source)};\n`);
                    }
                }
                yield* fixesNamedSpecifiers.removeTypeNamedSpecifiers;
                yield* afterFixes;
            }
            function* fixToValueImport(fixer, node) {
                var _a, _b;
                // import type Foo from 'foo'
                //        ^^^^ remove
                const importToken = util.nullThrows(sourceCode.getFirstToken(node, isImportToken), util.NullThrowsReasons.MissingToken('import', node.type));
                const typeToken = util.nullThrows(sourceCode.getFirstTokenBetween(importToken, (_b = (_a = node.specifiers[0]) === null || _a === void 0 ? void 0 : _a.local) !== null && _b !== void 0 ? _b : node.source, isTypeToken), util.NullThrowsReasons.MissingToken('type', node.type));
                const afterToken = util.nullThrows(sourceCode.getTokenAfter(typeToken, { includeComments: true }), util.NullThrowsReasons.MissingToken('any token', node.type));
                yield fixer.removeRange([typeToken.range[0], afterToken.range[0]]);
            }
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/default-param-last.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const util_1 = require(906) /* ../util */;
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    exports.default = util_1.createRule({
        name: 'default-param-last',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforce default parameters to be last',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: [],
            messages: {
                shouldBeLast: 'Default parameters should be last.',
            },
        },
        defaultOptions: [],
        create(context) {
            /**
             * checks if node is optional parameter
             * @param node the node to be evaluated
             * @private
             */
            function isOptionalParam(node) {
                return 'optional' in node && node.optional === true;
            }
            /**
             * checks if node is plain parameter
             * @param node the node to be evaluated
             * @private
             */
            function isPlainParam(node) {
                return !(node.type === experimental_utils_1.AST_NODE_TYPES.AssignmentPattern ||
                    node.type === experimental_utils_1.AST_NODE_TYPES.RestElement ||
                    isOptionalParam(node));
            }
            function checkDefaultParamLast(node) {
                let hasSeenPlainParam = false;
                for (let i = node.params.length - 1; i >= 0; i--) {
                    const current = node.params[i];
                    const param = current.type === experimental_utils_1.AST_NODE_TYPES.TSParameterProperty
                        ? current.parameter
                        : current;
                    if (isPlainParam(param)) {
                        hasSeenPlainParam = true;
                        continue;
                    }
                    if (hasSeenPlainParam &&
                        (isOptionalParam(param) ||
                            param.type === experimental_utils_1.AST_NODE_TYPES.AssignmentPattern)) {
                        context.report({ node: current, messageId: 'shouldBeLast' });
                    }
                }
            }
            return {
                ArrowFunctionExpression: checkDefaultParamLast,
                FunctionDeclaration: checkDefaultParamLast,
                FunctionExpression: checkDefaultParamLast,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/dot-notation.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const tsutils = __importStar(require(916) /* tsutils */);
    const dot_notation_1 = __importDefault(require(428) /* eslint/lib/rules/dot-notation */);
    const util_1 = require(906) /* ../util */;
    exports.default = util_1.createRule({
        name: 'dot-notation',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'enforce dot notation whenever possible',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
                requiresTypeChecking: true,
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowKeywords: {
                            type: 'boolean',
                            default: true,
                        },
                        allowPattern: {
                            type: 'string',
                            default: '',
                        },
                        allowPrivateClassPropertyAccess: {
                            type: 'boolean',
                            default: false,
                        },
                        allowProtectedClassPropertyAccess: {
                            type: 'boolean',
                            default: false,
                        },
                        allowIndexSignaturePropertyAccess: {
                            type: 'boolean',
                            default: false,
                        },
                    },
                    additionalProperties: false,
                },
            ],
            fixable: dot_notation_1.default.meta.fixable,
            messages: dot_notation_1.default.meta.messages,
        },
        defaultOptions: [
            {
                allowPrivateClassPropertyAccess: false,
                allowProtectedClassPropertyAccess: false,
                allowIndexSignaturePropertyAccess: false,
                allowKeywords: true,
                allowPattern: '',
            },
        ],
        create(context, [options]) {
            var _a;
            const rules = dot_notation_1.default.create(context);
            const { program, esTreeNodeToTSNodeMap } = util_1.getParserServices(context);
            const typeChecker = program.getTypeChecker();
            const allowPrivateClassPropertyAccess = options.allowPrivateClassPropertyAccess;
            const allowProtectedClassPropertyAccess = options.allowProtectedClassPropertyAccess;
            const allowIndexSignaturePropertyAccess = ((_a = options.allowIndexSignaturePropertyAccess) !== null && _a !== void 0 ? _a : false) ||
                tsutils.isCompilerOptionEnabled(program.getCompilerOptions(), 
                // @ts-expect-error - TS is refining the type to never for some reason
                'noPropertyAccessFromIndexSignature');
            return {
                MemberExpression(node) {
                    var _a, _b, _c;
                    if ((allowPrivateClassPropertyAccess ||
                        allowProtectedClassPropertyAccess ||
                        allowIndexSignaturePropertyAccess) &&
                        node.computed) {
                        // for perf reasons - only fetch symbols if we have to
                        const propertySymbol = typeChecker.getSymbolAtLocation(esTreeNodeToTSNodeMap.get(node.property));
                        const modifierKind = (_c = (_b = (_a = propertySymbol === null || propertySymbol === void 0 ? void 0 : propertySymbol.getDeclarations()) === null || _a === void 0 ? void 0 : _a[0]) === null || _b === void 0 ? void 0 : _b.modifiers) === null || _c === void 0 ? void 0 : _c[0].kind;
                        if ((allowPrivateClassPropertyAccess &&
                            modifierKind == ts.SyntaxKind.PrivateKeyword) ||
                            (allowProtectedClassPropertyAccess &&
                                modifierKind == ts.SyntaxKind.ProtectedKeyword)) {
                            return;
                        }
                        if (propertySymbol === undefined &&
                            allowIndexSignaturePropertyAccess) {
                            const objectType = typeChecker.getTypeAtLocation(esTreeNodeToTSNodeMap.get(node.object));
                            const indexType = typeChecker.getIndexTypeOfType(objectType, ts.IndexKind.String);
                            if (indexType != undefined) {
                                return;
                            }
                        }
                    }
                    rules.MemberExpression(node);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/explicit-function-return-type.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const explicitReturnTypeUtils_1 = require(957) /* ../util/explicitReturnTypeUtils */;
    exports.default = util.createRule({
        name: 'explicit-function-return-type',
        meta: {
            type: 'problem',
            docs: {
                description: 'Require explicit return types on functions and class methods',
                category: 'Stylistic Issues',
                recommended: false,
            },
            messages: {
                missingReturnType: 'Missing return type on function.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowExpressions: {
                            type: 'boolean',
                        },
                        allowTypedFunctionExpressions: {
                            type: 'boolean',
                        },
                        allowHigherOrderFunctions: {
                            type: 'boolean',
                        },
                        allowDirectConstAssertionInArrowFunctions: {
                            type: 'boolean',
                        },
                        allowConciseArrowFunctionExpressionsStartingWithVoid: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                allowExpressions: false,
                allowTypedFunctionExpressions: true,
                allowHigherOrderFunctions: true,
                allowDirectConstAssertionInArrowFunctions: true,
                allowConciseArrowFunctionExpressionsStartingWithVoid: false,
            },
        ],
        create(context, [options]) {
            const sourceCode = context.getSourceCode();
            return {
                'ArrowFunctionExpression, FunctionExpression'(node) {
                    if (options.allowConciseArrowFunctionExpressionsStartingWithVoid &&
                        node.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression &&
                        node.expression &&
                        node.body.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                        node.body.operator === 'void') {
                        return;
                    }
                    explicitReturnTypeUtils_1.checkFunctionExpressionReturnType(node, options, sourceCode, loc => context.report({
                        node,
                        loc,
                        messageId: 'missingReturnType',
                    }));
                },
                FunctionDeclaration(node) {
                    explicitReturnTypeUtils_1.checkFunctionReturnType(node, options, sourceCode, loc => context.report({
                        node,
                        loc,
                        messageId: 'missingReturnType',
                    }));
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/util/explicitReturnTypeUtils.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const astUtils_1 = require(907) /* ./astUtils */;
    const getFunctionHeadLoc_1 = require(912) /* ./getFunctionHeadLoc */;
    const nullThrows_1 = require(936) /* ./nullThrows */;
    /**
     * Checks if a node is a variable declarator with a type annotation.
     * ```
     * const x: Foo = ...
     * ```
     */
    function isVariableDeclaratorWithTypeAnnotation(node) {
        return (node.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclarator && !!node.id.typeAnnotation);
    }
    /**
     * Checks if a node is a class property with a type annotation.
     * ```
     * public x: Foo = ...
     * ```
     */
    function isClassPropertyWithTypeAnnotation(node) {
        return node.type === experimental_utils_1.AST_NODE_TYPES.ClassProperty && !!node.typeAnnotation;
    }
    /**
     * Checks if a node belongs to:
     * ```
     * new Foo(() => {})
     *         ^^^^^^^^
     * ```
     */
    function isConstructorArgument(node) {
        return node.type === experimental_utils_1.AST_NODE_TYPES.NewExpression;
    }
    /**
     * Checks if a node is a property or a nested property of a typed object:
     * ```
     * const x: Foo = { prop: () => {} }
     * const x = { prop: () => {} } as Foo
     * const x = <Foo>{ prop: () => {} }
     * const x: Foo = { bar: { prop: () => {} } }
     * ```
     */
    function isPropertyOfObjectWithType(property) {
        if (!property || property.type !== experimental_utils_1.AST_NODE_TYPES.Property) {
            return false;
        }
        const objectExpr = property.parent; // this shouldn't happen, checking just in case
        /* istanbul ignore if */ if (!objectExpr ||
            objectExpr.type !== experimental_utils_1.AST_NODE_TYPES.ObjectExpression) {
            return false;
        }
        const parent = objectExpr.parent; // this shouldn't happen, checking just in case
        /* istanbul ignore if */ if (!parent) {
            return false;
        }
        return (astUtils_1.isTypeAssertion(parent) ||
            isClassPropertyWithTypeAnnotation(parent) ||
            isVariableDeclaratorWithTypeAnnotation(parent) ||
            isFunctionArgument(parent) ||
            isPropertyOfObjectWithType(parent));
    }
    /**
     * Checks if a function belongs to:
     * ```
     * () => () => ...
     * () => function () { ... }
     * () => { return () => ... }
     * () => { return function () { ... } }
     * function fn() { return () => ... }
     * function fn() { return function() { ... } }
     * ```
     */
    function doesImmediatelyReturnFunctionExpression({ body, }) {
        // Should always have a body; really checking just in case
        /* istanbul ignore if */ if (!body) {
            return false;
        }
        // Check if body is a block with a single statement
        if (body.type === experimental_utils_1.AST_NODE_TYPES.BlockStatement && body.body.length === 1) {
            const [statement] = body.body;
            // Check if that statement is a return statement with an argument
            if (statement.type === experimental_utils_1.AST_NODE_TYPES.ReturnStatement &&
                !!statement.argument) {
                // If so, check that returned argument as body
                body = statement.argument;
            }
        }
        // Check if the body being returned is a function expression
        return (body.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression ||
            body.type === experimental_utils_1.AST_NODE_TYPES.FunctionExpression);
    }
    exports.doesImmediatelyReturnFunctionExpression = doesImmediatelyReturnFunctionExpression;
    /**
     * Checks if a node belongs to:
     * ```
     * foo(() => 1)
     * ```
     */
    function isFunctionArgument(parent, callee) {
        return (parent.type === experimental_utils_1.AST_NODE_TYPES.CallExpression &&
            // make sure this isn't an IIFE
            parent.callee !== callee);
    }
    /**
     * Checks if a function belongs to:
     * ```
     * () => ({ action: 'xxx' } as const)
     * ```
     */
    function returnsConstAssertionDirectly(node) {
        const { body } = node;
        if (astUtils_1.isTypeAssertion(body)) {
            const { typeAnnotation } = body;
            if (typeAnnotation.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference) {
                const { typeName } = typeAnnotation;
                if (typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    typeName.name === 'const') {
                    return true;
                }
            }
        }
        return false;
    }
    /**
     * True when the provided function expression is typed.
     */
    function isTypedFunctionExpression(node, options) {
        const parent = nullThrows_1.nullThrows(node.parent, nullThrows_1.NullThrowsReasons.MissingParent);
        if (!options.allowTypedFunctionExpressions) {
            return false;
        }
        return (astUtils_1.isTypeAssertion(parent) ||
            isVariableDeclaratorWithTypeAnnotation(parent) ||
            isClassPropertyWithTypeAnnotation(parent) ||
            isPropertyOfObjectWithType(parent) ||
            isFunctionArgument(parent, node) ||
            isConstructorArgument(parent));
    }
    exports.isTypedFunctionExpression = isTypedFunctionExpression;
    /**
     * Check whether the function expression return type is either typed or valid
     * with the provided options.
     */
    function isValidFunctionExpressionReturnType(node, options) {
        if (isTypedFunctionExpression(node, options)) {
            return true;
        }
        const parent = nullThrows_1.nullThrows(node.parent, nullThrows_1.NullThrowsReasons.MissingParent);
        if (options.allowExpressions &&
            parent.type !== experimental_utils_1.AST_NODE_TYPES.VariableDeclarator &&
            parent.type !== experimental_utils_1.AST_NODE_TYPES.MethodDefinition &&
            parent.type !== experimental_utils_1.AST_NODE_TYPES.ExportDefaultDeclaration &&
            parent.type !== experimental_utils_1.AST_NODE_TYPES.ClassProperty) {
            return true;
        }
        // https://github.com/typescript-eslint/typescript-eslint/issues/653
        if (options.allowDirectConstAssertionInArrowFunctions &&
            node.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression &&
            returnsConstAssertionDirectly(node)) {
            return true;
        }
        return false;
    }
    /**
     * Check that the function expression or declaration is valid.
     */
    function isValidFunctionReturnType(node, options) {
        if (options.allowHigherOrderFunctions &&
            doesImmediatelyReturnFunctionExpression(node)) {
            return true;
        }
        if (node.returnType || astUtils_1.isConstructor(node.parent) || astUtils_1.isSetter(node.parent)) {
            return true;
        }
        return false;
    }
    /**
     * Checks if a function declaration/expression has a return type.
     */
    function checkFunctionReturnType(node, options, sourceCode, report) {
        if (isValidFunctionReturnType(node, options)) {
            return;
        }
        report(getFunctionHeadLoc_1.getFunctionHeadLoc(node, sourceCode));
    }
    exports.checkFunctionReturnType = checkFunctionReturnType;
    /**
     * Checks if a function declaration/expression has a return type.
     */
    function checkFunctionExpressionReturnType(node, options, sourceCode, report) {
        if (isValidFunctionExpressionReturnType(node, options)) {
            return;
        }
        checkFunctionReturnType(node, options, sourceCode, report);
    }
    exports.checkFunctionExpressionReturnType = checkFunctionExpressionReturnType;
},
/* @typescript-eslint/eslint-plugin/dist/rules/explicit-member-accessibility.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const accessibilityLevel = { enum: ['explicit', 'no-public', 'off'] };
    exports.default = util.createRule({
        name: 'explicit-member-accessibility',
        meta: {
            type: 'problem',
            docs: {
                description: 'Require explicit accessibility modifiers on class properties and methods',
                category: 'Stylistic Issues',
                // too opinionated to be recommended
                recommended: false,
            },
            fixable: 'code',
            messages: {
                missingAccessibility: 'Missing accessibility modifier on {{type}} {{name}}.',
                unwantedPublicAccessibility: 'Public accessibility modifier on {{type}} {{name}}.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        accessibility: accessibilityLevel,
                        overrides: {
                            type: 'object',
                            properties: {
                                accessors: accessibilityLevel,
                                constructors: accessibilityLevel,
                                methods: accessibilityLevel,
                                properties: accessibilityLevel,
                                parameterProperties: accessibilityLevel,
                            },
                            additionalProperties: false,
                        },
                        ignoredMethodNames: {
                            type: 'array',
                            items: {
                                type: 'string',
                            },
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [{ accessibility: 'explicit' }],
        create(context, [option]) {
            var _a, _b, _c, _d, _e, _f, _g, _h;
            const sourceCode = context.getSourceCode();
            const baseCheck = (_a = option.accessibility) !== null && _a !== void 0 ? _a : 'explicit';
            const overrides = (_b = option.overrides) !== null && _b !== void 0 ? _b : {};
            const ctorCheck = (_c = overrides.constructors) !== null && _c !== void 0 ? _c : baseCheck;
            const accessorCheck = (_d = overrides.accessors) !== null && _d !== void 0 ? _d : baseCheck;
            const methodCheck = (_e = overrides.methods) !== null && _e !== void 0 ? _e : baseCheck;
            const propCheck = (_f = overrides.properties) !== null && _f !== void 0 ? _f : baseCheck;
            const paramPropCheck = (_g = overrides.parameterProperties) !== null && _g !== void 0 ? _g : baseCheck;
            const ignoredMethodNames = new Set((_h = option.ignoredMethodNames) !== null && _h !== void 0 ? _h : []);
            /**
             * Generates the report for rule violations
             */
            function reportIssue(messageId, nodeType, node, nodeName, fix = null) {
                context.report({
                    node: node,
                    messageId: messageId,
                    data: {
                        type: nodeType,
                        name: nodeName,
                    },
                    fix: fix,
                });
            }
            /**
             * Checks if a method declaration has an accessibility modifier.
             * @param methodDefinition The node representing a MethodDefinition.
             */
            function checkMethodAccessibilityModifier(methodDefinition) {
                let nodeType = 'method definition';
                let check = baseCheck;
                switch (methodDefinition.kind) {
                    case 'method':
                        check = methodCheck;
                        break;
                    case 'constructor':
                        check = ctorCheck;
                        break;
                    case 'get':
                    case 'set':
                        check = accessorCheck;
                        nodeType = `${methodDefinition.kind} property accessor`;
                        break;
                }
                const methodName = util.getNameFromMember(methodDefinition, sourceCode);
                if (check === 'off' || ignoredMethodNames.has(methodName)) {
                    return;
                }
                if (check === 'no-public' &&
                    methodDefinition.accessibility === 'public') {
                    reportIssue('unwantedPublicAccessibility', nodeType, methodDefinition, methodName, getUnwantedPublicAccessibilityFixer(methodDefinition));
                }
                else if (check === 'explicit' && !methodDefinition.accessibility) {
                    reportIssue('missingAccessibility', nodeType, methodDefinition, methodName);
                }
            }
            /**
             * Creates a fixer that removes a "public" keyword with following spaces
             */
            function getUnwantedPublicAccessibilityFixer(node) {
                return function (fixer) {
                    const tokens = sourceCode.getTokens(node);
                    let rangeToRemove;
                    for (let i = 0; i < tokens.length; i++) {
                        const token = tokens[i];
                        if (token.type === experimental_utils_1.AST_TOKEN_TYPES.Keyword &&
                            token.value === 'public') {
                            const commensAfterPublicKeyword = sourceCode.getCommentsAfter(token);
                            if (commensAfterPublicKeyword.length) {
                                // public /* Hi there! */ static foo()
                                // ^^^^^^^
                                rangeToRemove = [
                                    token.range[0],
                                    commensAfterPublicKeyword[0].range[0],
                                ];
                                break;
                            }
                            else {
                                // public static foo()
                                // ^^^^^^^
                                rangeToRemove = [token.range[0], tokens[i + 1].range[0]];
                                break;
                            }
                        }
                    }
                    return fixer.removeRange(rangeToRemove);
                };
            }
            /**
             * Checks if property has an accessibility modifier.
             * @param classProperty The node representing a ClassProperty.
             */
            function checkPropertyAccessibilityModifier(classProperty) {
                const nodeType = 'class property';
                const propertyName = util.getNameFromMember(classProperty, sourceCode);
                if (propCheck === 'no-public' &&
                    classProperty.accessibility === 'public') {
                    reportIssue('unwantedPublicAccessibility', nodeType, classProperty, propertyName, getUnwantedPublicAccessibilityFixer(classProperty));
                }
                else if (propCheck === 'explicit' && !classProperty.accessibility) {
                    reportIssue('missingAccessibility', nodeType, classProperty, propertyName);
                }
            }
            /**
             * Checks that the parameter property has the desired accessibility modifiers set.
             * @param node The node representing a Parameter Property
             */
            function checkParameterPropertyAccessibilityModifier(node) {
                const nodeType = 'parameter property';
                // HAS to be an identifier or assignment or TSC will throw
                if (node.parameter.type !== experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    node.parameter.type !== experimental_utils_1.AST_NODE_TYPES.AssignmentPattern) {
                    return;
                }
                const nodeName = node.parameter.type === experimental_utils_1.AST_NODE_TYPES.Identifier
                    ? node.parameter.name
                    : // has to be an Identifier or TSC will throw an error
                        node.parameter.left.name;
                switch (paramPropCheck) {
                    case 'explicit': {
                        if (!node.accessibility) {
                            reportIssue('missingAccessibility', nodeType, node, nodeName);
                        }
                        break;
                    }
                    case 'no-public': {
                        if (node.accessibility === 'public' && node.readonly) {
                            reportIssue('unwantedPublicAccessibility', nodeType, node, nodeName, getUnwantedPublicAccessibilityFixer(node));
                        }
                        break;
                    }
                }
            }
            return {
                TSParameterProperty: checkParameterPropertyAccessibilityModifier,
                ClassProperty: checkPropertyAccessibilityModifier,
                MethodDefinition: checkMethodAccessibilityModifier,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/explicit-module-boundary-types.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const explicitReturnTypeUtils_1 = require(957) /* ../util/explicitReturnTypeUtils */;
    exports.default = util.createRule({
        name: 'explicit-module-boundary-types',
        meta: {
            type: 'problem',
            docs: {
                description: "Require explicit return and argument types on exported functions' and classes' public class methods",
                category: 'Stylistic Issues',
                recommended: 'warn',
            },
            messages: {
                missingReturnType: 'Missing return type on function.',
                missingArgType: "Argument '{{name}}' should be typed.",
                missingArgTypeUnnamed: '{{type}} argument should be typed.',
                anyTypedArg: "Argument '{{name}}' should be typed with a non-any type.",
                anyTypedArgUnnamed: '{{type}} argument should be typed with a non-any type.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowArgumentsExplicitlyTypedAsAny: {
                            type: 'boolean',
                        },
                        allowDirectConstAssertionInArrowFunctions: {
                            type: 'boolean',
                        },
                        allowedNames: {
                            type: 'array',
                            items: {
                                type: 'string',
                            },
                        },
                        allowHigherOrderFunctions: {
                            type: 'boolean',
                        },
                        allowTypedFunctionExpressions: {
                            type: 'boolean',
                        },
                        // DEPRECATED - To be removed in next major
                        shouldTrackReferences: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                allowArgumentsExplicitlyTypedAsAny: false,
                allowDirectConstAssertionInArrowFunctions: true,
                allowedNames: [],
                allowHigherOrderFunctions: true,
                allowTypedFunctionExpressions: true,
            },
        ],
        create(context, [options]) {
            const sourceCode = context.getSourceCode();
            // tracks all of the functions we've already checked
            const checkedFunctions = new Set();
            // tracks functions that were found whilst traversing
            const foundFunctions = [];
            // all nodes visited, avoids infinite recursion for cyclic references
            // (such as class member referring to itself)
            const alreadyVisited = new Set();
            /*
            # How the rule works:
        
            As the rule traverses the AST, it immediately checks every single function that it finds is exported.
            "exported" means that it is either directly exported, or that its name is exported.
        
            It also collects a list of every single function it finds on the way, but does not check them.
            After it's finished traversing the AST, it then iterates through the list of found functions, and checks to see if
            any of them are part of a higher-order function
            */
            return {
                ExportDefaultDeclaration(node) {
                    checkNode(node.declaration);
                },
                'ExportNamedDeclaration:not([source])'(node) {
                    if (node.declaration) {
                        checkNode(node.declaration);
                    }
                    else {
                        for (const specifier of node.specifiers) {
                            followReference(specifier.local);
                        }
                    }
                },
                TSExportAssignment(node) {
                    checkNode(node.expression);
                },
                'ArrowFunctionExpression, FunctionDeclaration, FunctionExpression'(node) {
                    foundFunctions.push(node);
                },
                'Program:exit'() {
                    for (const func of foundFunctions) {
                        if (isExportedHigherOrderFunction(func)) {
                            checkNode(func);
                        }
                    }
                },
            };
            function checkParameters(node) {
                function checkParameter(param) {
                    function report(namedMessageId, unnamedMessageId) {
                        if (param.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                            context.report({
                                node: param,
                                messageId: namedMessageId,
                                data: { name: param.name },
                            });
                        }
                        else if (param.type === experimental_utils_1.AST_NODE_TYPES.ArrayPattern) {
                            context.report({
                                node: param,
                                messageId: unnamedMessageId,
                                data: { type: 'Array pattern' },
                            });
                        }
                        else if (param.type === experimental_utils_1.AST_NODE_TYPES.ObjectPattern) {
                            context.report({
                                node: param,
                                messageId: unnamedMessageId,
                                data: { type: 'Object pattern' },
                            });
                        }
                        else if (param.type === experimental_utils_1.AST_NODE_TYPES.RestElement) {
                            if (param.argument.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                                context.report({
                                    node: param,
                                    messageId: namedMessageId,
                                    data: { name: param.argument.name },
                                });
                            }
                            else {
                                context.report({
                                    node: param,
                                    messageId: unnamedMessageId,
                                    data: { type: 'Rest' },
                                });
                            }
                        }
                    }
                    switch (param.type) {
                        case experimental_utils_1.AST_NODE_TYPES.ArrayPattern:
                        case experimental_utils_1.AST_NODE_TYPES.Identifier:
                        case experimental_utils_1.AST_NODE_TYPES.ObjectPattern:
                        case experimental_utils_1.AST_NODE_TYPES.RestElement:
                            if (!param.typeAnnotation) {
                                report('missingArgType', 'missingArgTypeUnnamed');
                            }
                            else if (options.allowArgumentsExplicitlyTypedAsAny !== true &&
                                param.typeAnnotation.typeAnnotation.type ===
                                    experimental_utils_1.AST_NODE_TYPES.TSAnyKeyword) {
                                report('anyTypedArg', 'anyTypedArgUnnamed');
                            }
                            return;
                        case experimental_utils_1.AST_NODE_TYPES.TSParameterProperty:
                            return checkParameter(param.parameter);
                        case experimental_utils_1.AST_NODE_TYPES.AssignmentPattern: // ignored as it has a type via its assignment
                            return;
                    }
                }
                for (const arg of node.params) {
                    checkParameter(arg);
                }
            }
            /**
             * Checks if a function name is allowed and should not be checked.
             */
            function isAllowedName(node) {
                var _a;
                if (!node || !options.allowedNames || !options.allowedNames.length) {
                    return false;
                }
                if (node.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclarator ||
                    node.type === experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration) {
                    return (((_a = node.id) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                        options.allowedNames.includes(node.id.name));
                }
                else if (node.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition ||
                    node.type === experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition ||
                    (node.type === experimental_utils_1.AST_NODE_TYPES.Property && node.method)) {
                    if (node.key.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                        typeof node.key.value === 'string') {
                        return options.allowedNames.includes(node.key.value);
                    }
                    if (node.key.type === experimental_utils_1.AST_NODE_TYPES.TemplateLiteral &&
                        node.key.expressions.length === 0) {
                        return options.allowedNames.includes(node.key.quasis[0].value.raw);
                    }
                    if (!node.computed && node.key.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                        return options.allowedNames.includes(node.key.name);
                    }
                }
                return false;
            }
            function isExportedHigherOrderFunction(node) {
                var _a;
                let current = node.parent;
                while (current) {
                    if (current.type === experimental_utils_1.AST_NODE_TYPES.ReturnStatement) {
                        // the parent of a return will always be a block statement, so we can skip over it
                        current = (_a = current.parent) === null || _a === void 0 ? void 0 : _a.parent;
                        continue;
                    }
                    if (!util.isFunction(current) ||
                        !explicitReturnTypeUtils_1.doesImmediatelyReturnFunctionExpression(current)) {
                        return false;
                    }
                    if (checkedFunctions.has(current)) {
                        return true;
                    }
                    current = current.parent;
                }
                return false;
            }
            function followReference(node) {
                const scope = context.getScope();
                const variable = scope.set.get(node.name);
                /* istanbul ignore if */ if (!variable) {
                    return;
                }
                // check all of the definitions
                for (const definition of variable.defs) {
                    // cases we don't care about in this rule
                    if (definition.type === 'ImplicitGlobalVariable' ||
                        definition.type === 'ImportBinding' ||
                        // eslint-disable-next-line @typescript-eslint/internal/prefer-ast-types-enum
                        definition.type === 'CatchClause' ||
                        definition.type === 'Parameter') {
                        continue;
                    }
                    checkNode(definition.node);
                }
                // follow references to find writes to the variable
                for (const reference of variable.references) {
                    if (
                    // we don't want to check the initialization ref, as this is handled by the declaration check
                    !reference.init &&
                        reference.writeExpr) {
                        checkNode(reference.writeExpr);
                    }
                }
            }
            function checkNode(node) {
                if (node == null || alreadyVisited.has(node)) {
                    return;
                }
                alreadyVisited.add(node);
                switch (node.type) {
                    case experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression:
                    case experimental_utils_1.AST_NODE_TYPES.FunctionExpression:
                        return checkFunctionExpression(node);
                    case experimental_utils_1.AST_NODE_TYPES.ArrayExpression:
                        for (const element of node.elements) {
                            checkNode(element);
                        }
                        return;
                    case experimental_utils_1.AST_NODE_TYPES.ClassProperty:
                    case experimental_utils_1.AST_NODE_TYPES.TSAbstractClassProperty:
                        if (node.accessibility === 'private') {
                            return;
                        }
                        return checkNode(node.value);
                    case experimental_utils_1.AST_NODE_TYPES.ClassDeclaration:
                    case experimental_utils_1.AST_NODE_TYPES.ClassExpression:
                        for (const element of node.body.body) {
                            checkNode(element);
                        }
                        return;
                    case experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration:
                        return checkFunction(node);
                    case experimental_utils_1.AST_NODE_TYPES.MethodDefinition:
                    case experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition:
                        if (node.accessibility === 'private') {
                            return;
                        }
                        return checkNode(node.value);
                    case experimental_utils_1.AST_NODE_TYPES.Identifier:
                        return followReference(node);
                    case experimental_utils_1.AST_NODE_TYPES.ObjectExpression:
                        for (const property of node.properties) {
                            checkNode(property);
                        }
                        return;
                    case experimental_utils_1.AST_NODE_TYPES.Property:
                        return checkNode(node.value);
                    case experimental_utils_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression:
                        return checkEmptyBodyFunctionExpression(node);
                    case experimental_utils_1.AST_NODE_TYPES.VariableDeclaration:
                        for (const declaration of node.declarations) {
                            checkNode(declaration);
                        }
                        return;
                    case experimental_utils_1.AST_NODE_TYPES.VariableDeclarator:
                        return checkNode(node.init);
                }
            }
            /**
             * Check whether any ancestor of the provided function has a valid return type.
             * This function assumes that the function either:
             * - belongs to an exported function chain validated by isExportedHigherOrderFunction
             * - is directly exported itself
             */
            function ancestorHasReturnType(node) {
                let ancestor = node.parent;
                if ((ancestor === null || ancestor === void 0 ? void 0 : ancestor.type) === experimental_utils_1.AST_NODE_TYPES.Property) {
                    ancestor = ancestor.value;
                }
                // if the ancestor is not a return, then this function was not returned at all, so we can exit early
                const isReturnStatement = (ancestor === null || ancestor === void 0 ? void 0 : ancestor.type) === experimental_utils_1.AST_NODE_TYPES.ReturnStatement;
                const isBodylessArrow = (ancestor === null || ancestor === void 0 ? void 0 : ancestor.type) === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression &&
                    ancestor.body.type !== experimental_utils_1.AST_NODE_TYPES.BlockStatement;
                if (!isReturnStatement && !isBodylessArrow) {
                    return false;
                }
                while (ancestor) {
                    switch (ancestor.type) {
                        case experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression:
                        case experimental_utils_1.AST_NODE_TYPES.FunctionExpression:
                        case experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration:
                            if (ancestor.returnType) {
                                return true;
                            }
                            // assume
                            break;
                        // const x: Foo = () => {};
                        // Assume that a typed variable types the function expression
                        case experimental_utils_1.AST_NODE_TYPES.VariableDeclarator:
                            if (ancestor.id.typeAnnotation) {
                                return true;
                            }
                            break;
                    }
                    ancestor = ancestor.parent;
                }
                return false;
            }
            function checkEmptyBodyFunctionExpression(node) {
                var _a, _b, _c;
                const isConstructor = ((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.MethodDefinition &&
                    node.parent.kind === 'constructor';
                const isSetAccessor = (((_b = node.parent) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition ||
                    ((_c = node.parent) === null || _c === void 0 ? void 0 : _c.type) === experimental_utils_1.AST_NODE_TYPES.MethodDefinition) &&
                    node.parent.kind === 'set';
                if (!isConstructor && !isSetAccessor && !node.returnType) {
                    context.report({
                        node,
                        messageId: 'missingReturnType',
                    });
                }
                checkParameters(node);
            }
            function checkFunctionExpression(node) {
                if (checkedFunctions.has(node)) {
                    return;
                }
                checkedFunctions.add(node);
                if (isAllowedName(node.parent) ||
                    explicitReturnTypeUtils_1.isTypedFunctionExpression(node, options) ||
                    ancestorHasReturnType(node)) {
                    return;
                }
                explicitReturnTypeUtils_1.checkFunctionExpressionReturnType(node, options, sourceCode, loc => {
                    context.report({
                        node,
                        loc,
                        messageId: 'missingReturnType',
                    });
                });
                checkParameters(node);
            }
            function checkFunction(node) {
                if (checkedFunctions.has(node)) {
                    return;
                }
                checkedFunctions.add(node);
                if (isAllowedName(node) || ancestorHasReturnType(node)) {
                    return;
                }
                explicitReturnTypeUtils_1.checkFunctionReturnType(node, options, sourceCode, loc => {
                    context.report({
                        node,
                        loc,
                        messageId: 'missingReturnType',
                    });
                });
                checkParameters(node);
            }
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/func-call-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'func-call-spacing',
        meta: {
            type: 'layout',
            docs: {
                description: 'Require or disallow spacing between function identifiers and their invocations',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: 'whitespace',
            schema: {
                anyOf: [
                    {
                        type: 'array',
                        items: [
                            {
                                enum: ['never'],
                            },
                        ],
                        minItems: 0,
                        maxItems: 1,
                    },
                    {
                        type: 'array',
                        items: [
                            {
                                enum: ['always'],
                            },
                            {
                                type: 'object',
                                properties: {
                                    allowNewlines: {
                                        type: 'boolean',
                                    },
                                },
                                additionalProperties: false,
                            },
                        ],
                        minItems: 0,
                        maxItems: 2,
                    },
                ],
            },
            messages: {
                unexpectedWhitespace: 'Unexpected whitespace between function name and paren.',
                unexpectedNewline: 'Unexpected newline between function name and paren.',
                missing: 'Missing space between function name and paren.',
            },
        },
        defaultOptions: ['never', {}],
        create(context, [option, config]) {
            const sourceCode = context.getSourceCode();
            const text = sourceCode.getText();
            /**
             * Check if open space is present in a function name
             * @param {ASTNode} node node to evaluate
             * @returns {void}
             * @private
             */
            function checkSpacing(node) {
                var _a;
                const isOptionalCall = util.isOptionalCallExpression(node);
                const closingParenToken = sourceCode.getLastToken(node);
                const lastCalleeTokenWithoutPossibleParens = sourceCode.getLastToken((_a = node.typeParameters) !== null && _a !== void 0 ? _a : node.callee);
                const openingParenToken = sourceCode.getFirstTokenBetween(lastCalleeTokenWithoutPossibleParens, closingParenToken, util.isOpeningParenToken);
                if (!openingParenToken || openingParenToken.range[1] >= node.range[1]) {
                    // new expression with no parens...
                    return;
                }
                const lastCalleeToken = sourceCode.getTokenBefore(openingParenToken, util.isNotOptionalChainPunctuator);
                const textBetweenTokens = text
                    .slice(lastCalleeToken.range[1], openingParenToken.range[0])
                    .replace(/\/\*.*?\*\//gu, '');
                const hasWhitespace = /\s/u.test(textBetweenTokens);
                const hasNewline = hasWhitespace && util.LINEBREAK_MATCHER.test(textBetweenTokens);
                if (option === 'never') {
                    if (hasWhitespace) {
                        return context.report({
                            node,
                            loc: lastCalleeToken.loc.start,
                            messageId: 'unexpectedWhitespace',
                            fix(fixer) {
                                /*
                                 * Only autofix if there is no newline
                                 * https://github.com/eslint/eslint/issues/7787
                                 */
                                if (!hasNewline &&
                                    // don't fix optional calls
                                    !isOptionalCall) {
                                    return fixer.removeRange([
                                        lastCalleeToken.range[1],
                                        openingParenToken.range[0],
                                    ]);
                                }
                                return null;
                            },
                        });
                    }
                }
                else if (isOptionalCall) {
                    // disallow:
                    // foo?. ();
                    // foo ?.();
                    // foo ?. ();
                    if (hasWhitespace || hasNewline) {
                        context.report({
                            node,
                            loc: lastCalleeToken.loc.start,
                            messageId: 'unexpectedWhitespace',
                        });
                    }
                }
                else {
                    if (!hasWhitespace) {
                        context.report({
                            node,
                            loc: lastCalleeToken.loc.start,
                            messageId: 'missing',
                            fix(fixer) {
                                return fixer.insertTextBefore(openingParenToken, ' ');
                            },
                        });
                    }
                    else if (!config.allowNewlines && hasNewline) {
                        context.report({
                            node,
                            loc: lastCalleeToken.loc.start,
                            messageId: 'unexpectedNewline',
                            fix(fixer) {
                                return fixer.replaceTextRange([lastCalleeToken.range[1], openingParenToken.range[0]], ' ');
                            },
                        });
                    }
                }
            }
            return {
                CallExpression: checkSpacing,
                NewExpression: checkSpacing,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/indent.js */ function _(require, module, exports, __esModule, __esExport) {
    /**
     * Note this file is rather type-unsafe in its current state.
     * This is due to some really funky type conversions between different node types.
     * This is done intentionally based on the internal implementation of the base indent rule.
     */
    /* eslint-disable @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-assignment  */
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const indent_1 = __importDefault(require(450) /* eslint/lib/rules/indent */);
    const util = __importStar(require(906) /* ../util */);
    const KNOWN_NODES = new Set([
        // Class properties aren't yet supported by eslint...
        experimental_utils_1.AST_NODE_TYPES.ClassProperty,
        // ts keywords
        experimental_utils_1.AST_NODE_TYPES.TSAbstractKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSAnyKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSBooleanKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSNeverKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSNumberKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSStringKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSSymbolKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSUndefinedKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSUnknownKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSVoidKeyword,
        experimental_utils_1.AST_NODE_TYPES.TSNullKeyword,
        // ts specific nodes we want to support
        experimental_utils_1.AST_NODE_TYPES.TSAbstractClassProperty,
        experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition,
        experimental_utils_1.AST_NODE_TYPES.TSArrayType,
        experimental_utils_1.AST_NODE_TYPES.TSAsExpression,
        experimental_utils_1.AST_NODE_TYPES.TSCallSignatureDeclaration,
        experimental_utils_1.AST_NODE_TYPES.TSConditionalType,
        experimental_utils_1.AST_NODE_TYPES.TSConstructorType,
        experimental_utils_1.AST_NODE_TYPES.TSConstructSignatureDeclaration,
        experimental_utils_1.AST_NODE_TYPES.TSDeclareFunction,
        experimental_utils_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression,
        experimental_utils_1.AST_NODE_TYPES.TSEnumDeclaration,
        experimental_utils_1.AST_NODE_TYPES.TSEnumMember,
        experimental_utils_1.AST_NODE_TYPES.TSExportAssignment,
        experimental_utils_1.AST_NODE_TYPES.TSExternalModuleReference,
        experimental_utils_1.AST_NODE_TYPES.TSFunctionType,
        experimental_utils_1.AST_NODE_TYPES.TSImportType,
        experimental_utils_1.AST_NODE_TYPES.TSIndexedAccessType,
        experimental_utils_1.AST_NODE_TYPES.TSIndexSignature,
        experimental_utils_1.AST_NODE_TYPES.TSInferType,
        experimental_utils_1.AST_NODE_TYPES.TSInterfaceBody,
        experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration,
        experimental_utils_1.AST_NODE_TYPES.TSInterfaceHeritage,
        experimental_utils_1.AST_NODE_TYPES.TSIntersectionType,
        experimental_utils_1.AST_NODE_TYPES.TSImportEqualsDeclaration,
        experimental_utils_1.AST_NODE_TYPES.TSLiteralType,
        experimental_utils_1.AST_NODE_TYPES.TSMappedType,
        experimental_utils_1.AST_NODE_TYPES.TSMethodSignature,
        'TSMinusToken',
        experimental_utils_1.AST_NODE_TYPES.TSModuleBlock,
        experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration,
        experimental_utils_1.AST_NODE_TYPES.TSNonNullExpression,
        experimental_utils_1.AST_NODE_TYPES.TSParameterProperty,
        experimental_utils_1.AST_NODE_TYPES.TSParenthesizedType,
        'TSPlusToken',
        experimental_utils_1.AST_NODE_TYPES.TSPropertySignature,
        experimental_utils_1.AST_NODE_TYPES.TSQualifiedName,
        'TSQuestionToken',
        experimental_utils_1.AST_NODE_TYPES.TSRestType,
        experimental_utils_1.AST_NODE_TYPES.TSThisType,
        experimental_utils_1.AST_NODE_TYPES.TSTupleType,
        experimental_utils_1.AST_NODE_TYPES.TSTypeAnnotation,
        experimental_utils_1.AST_NODE_TYPES.TSTypeLiteral,
        experimental_utils_1.AST_NODE_TYPES.TSTypeOperator,
        experimental_utils_1.AST_NODE_TYPES.TSTypeParameter,
        experimental_utils_1.AST_NODE_TYPES.TSTypeParameterDeclaration,
        experimental_utils_1.AST_NODE_TYPES.TSTypeParameterInstantiation,
        experimental_utils_1.AST_NODE_TYPES.TSTypeReference,
        experimental_utils_1.AST_NODE_TYPES.TSUnionType,
        experimental_utils_1.AST_NODE_TYPES.Decorator,
    ]);
    exports.default = util.createRule({
        name: 'indent',
        meta: {
            type: 'layout',
            docs: {
                description: 'Enforce consistent indentation',
                category: 'Stylistic Issues',
                // too opinionated to be recommended
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: 'whitespace',
            schema: indent_1.default.meta.schema,
            messages: (_a = indent_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                wrongIndentation: 'Expected indentation of {{expected}} but found {{actual}}.',
            },
        },
        defaultOptions: [
            // typescript docs and playground use 4 space indent
            4,
            {
                // typescript docs indent the case from the switch
                // https://www.typescriptlang.org/docs/handbook/release-notes/typescript-1-8.html#example-4
                SwitchCase: 1,
                flatTernaryExpressions: false,
                ignoredNodes: [],
            },
        ],
        create(context, optionsWithDefaults) {
            // because we extend the base rule, have to update opts on the context
            // the context defines options as readonly though...
            const contextWithDefaults = Object.create(context, {
                options: {
                    writable: false,
                    configurable: false,
                    value: optionsWithDefaults,
                },
            });
            const rules = indent_1.default.create(contextWithDefaults);
            /**
             * Converts from a TSPropertySignature to a Property
             * @param node a TSPropertySignature node
             * @param [type] the type to give the new node
             * @returns a Property node
             */
            function TSPropertySignatureToProperty(node, type = experimental_utils_1.AST_NODE_TYPES.Property) {
                const base = {
                    // indent doesn't actually use these
                    key: null,
                    value: null,
                    // Property flags
                    computed: false,
                    method: false,
                    kind: 'init',
                    // this will stop eslint from interrogating the type literal
                    shorthand: true,
                    // location data
                    parent: node.parent,
                    range: node.range,
                    loc: node.loc,
                };
                if (type === experimental_utils_1.AST_NODE_TYPES.Property) {
                    return Object.assign({ type }, base);
                }
                else {
                    return Object.assign({ type, static: false, readonly: false, declare: false }, base);
                }
            }
            return Object.assign({}, rules, {
                // overwrite the base rule here so we can use our KNOWN_NODES list instead
                '*:exit'(node) {
                    // For nodes we care about, skip the default handling, because it just marks the node as ignored...
                    if (!KNOWN_NODES.has(node.type)) {
                        rules['*:exit'](node);
                    }
                },
                VariableDeclaration(node) {
                    // https://github.com/typescript-eslint/typescript-eslint/issues/441
                    if (node.declarations.length === 0) {
                        return;
                    }
                    return rules.VariableDeclaration(node);
                },
                TSAsExpression(node) {
                    // transform it to a BinaryExpression
                    return rules['BinaryExpression, LogicalExpression']({
                        type: experimental_utils_1.AST_NODE_TYPES.BinaryExpression,
                        operator: 'as',
                        left: node.expression,
                        // the first typeAnnotation includes the as token
                        right: node.typeAnnotation,
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
                TSConditionalType(node) {
                    // transform it to a ConditionalExpression
                    return rules.ConditionalExpression({
                        type: experimental_utils_1.AST_NODE_TYPES.ConditionalExpression,
                        test: {
                            type: experimental_utils_1.AST_NODE_TYPES.BinaryExpression,
                            operator: 'extends',
                            left: node.checkType,
                            right: node.extendsType,
                            // location data
                            range: [node.checkType.range[0], node.extendsType.range[1]],
                            loc: {
                                start: node.checkType.loc.start,
                                end: node.extendsType.loc.end,
                            },
                        },
                        consequent: node.trueType,
                        alternate: node.falseType,
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
                'TSEnumDeclaration, TSTypeLiteral'(node) {
                    // transform it to an ObjectExpression
                    return rules['ObjectExpression, ObjectPattern']({
                        type: experimental_utils_1.AST_NODE_TYPES.ObjectExpression,
                        properties: node.members.map(member => TSPropertySignatureToProperty(member)),
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
                TSImportEqualsDeclaration(node) {
                    // transform it to an VariableDeclaration
                    // use VariableDeclaration instead of ImportDeclaration because it's essentially the same thing
                    const { id, moduleReference } = node;
                    return rules.VariableDeclaration({
                        type: experimental_utils_1.AST_NODE_TYPES.VariableDeclaration,
                        kind: 'const',
                        declarations: [
                            {
                                type: experimental_utils_1.AST_NODE_TYPES.VariableDeclarator,
                                range: [id.range[0], moduleReference.range[1]],
                                loc: {
                                    start: id.loc.start,
                                    end: moduleReference.loc.end,
                                },
                                id: id,
                                init: {
                                    type: experimental_utils_1.AST_NODE_TYPES.CallExpression,
                                    callee: {
                                        type: experimental_utils_1.AST_NODE_TYPES.Identifier,
                                        name: 'require',
                                        range: [
                                            moduleReference.range[0],
                                            moduleReference.range[0] + 'require'.length,
                                        ],
                                        loc: {
                                            start: moduleReference.loc.start,
                                            end: {
                                                line: moduleReference.loc.end.line,
                                                column: moduleReference.loc.start.line + 'require'.length,
                                            },
                                        },
                                    },
                                    arguments: 'expression' in moduleReference
                                        ? [moduleReference.expression]
                                        : [],
                                    // location data
                                    range: moduleReference.range,
                                    loc: moduleReference.loc,
                                },
                            },
                        ],
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
                TSIndexedAccessType(node) {
                    // convert to a MemberExpression
                    return rules['MemberExpression, JSXMemberExpression, MetaProperty']({
                        type: experimental_utils_1.AST_NODE_TYPES.MemberExpression,
                        object: node.objectType,
                        property: node.indexType,
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                        optional: false,
                        computed: true,
                    });
                },
                TSInterfaceBody(node) {
                    // transform it to an ClassBody
                    return rules['BlockStatement, ClassBody']({
                        type: experimental_utils_1.AST_NODE_TYPES.ClassBody,
                        body: node.body.map(p => TSPropertySignatureToProperty(p, experimental_utils_1.AST_NODE_TYPES.ClassProperty)),
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
                'TSInterfaceDeclaration[extends.length > 0]'(node) {
                    // transform it to a ClassDeclaration
                    return rules['ClassDeclaration[superClass], ClassExpression[superClass]']({
                        type: experimental_utils_1.AST_NODE_TYPES.ClassDeclaration,
                        body: node.body,
                        id: null,
                        // TODO: This is invalid, there can be more than one extends in interface
                        superClass: node.extends[0].expression,
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
                TSMappedType(node) {
                    const sourceCode = context.getSourceCode();
                    const squareBracketStart = sourceCode.getTokenBefore(node.typeParameter);
                    // transform it to an ObjectExpression
                    return rules['ObjectExpression, ObjectPattern']({
                        type: experimental_utils_1.AST_NODE_TYPES.ObjectExpression,
                        properties: [
                            {
                                type: experimental_utils_1.AST_NODE_TYPES.Property,
                                key: node.typeParameter,
                                value: node.typeAnnotation,
                                // location data
                                range: [
                                    squareBracketStart.range[0],
                                    node.typeAnnotation
                                        ? node.typeAnnotation.range[1]
                                        : squareBracketStart.range[0],
                                ],
                                loc: {
                                    start: squareBracketStart.loc.start,
                                    end: node.typeAnnotation
                                        ? node.typeAnnotation.loc.end
                                        : squareBracketStart.loc.end,
                                },
                                kind: 'init',
                                computed: false,
                                method: false,
                                shorthand: false,
                            },
                        ],
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
                TSModuleBlock(node) {
                    // transform it to a BlockStatement
                    return rules['BlockStatement, ClassBody']({
                        type: experimental_utils_1.AST_NODE_TYPES.BlockStatement,
                        body: node.body,
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
                TSQualifiedName(node) {
                    return rules['MemberExpression, JSXMemberExpression, MetaProperty']({
                        type: experimental_utils_1.AST_NODE_TYPES.MemberExpression,
                        object: node.left,
                        property: node.right,
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                        optional: false,
                        computed: false,
                    });
                },
                TSTupleType(node) {
                    // transform it to an ArrayExpression
                    return rules['ArrayExpression, ArrayPattern']({
                        type: experimental_utils_1.AST_NODE_TYPES.ArrayExpression,
                        elements: node.elementTypes,
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
                TSTypeParameterDeclaration(node) {
                    if (!node.params.length) {
                        return;
                    }
                    const [name, ...attributes] = node.params;
                    // JSX is about the closest we can get because the angle brackets
                    // it's not perfect but it works!
                    return rules.JSXOpeningElement({
                        type: experimental_utils_1.AST_NODE_TYPES.JSXOpeningElement,
                        selfClosing: false,
                        name: name,
                        attributes: attributes,
                        // location data
                        parent: node.parent,
                        range: node.range,
                        loc: node.loc,
                    });
                },
            });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/init-declarations.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const init_declarations_1 = __importDefault(require(453) /* eslint/lib/rules/init-declarations */);
    const util_1 = require(906) /* ../util */;
    exports.default = util_1.createRule({
        name: 'init-declarations',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'require or disallow initialization in variable declarations',
                category: 'Variables',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: init_declarations_1.default.meta.schema,
            messages: (_a = init_declarations_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                initialized: "Variable '{{idName}}' should be initialized on declaration.",
                notInitialized: "Variable '{{idName}}' should not be initialized on declaration.",
            },
        },
        defaultOptions: ['always'],
        create(context) {
            const rules = init_declarations_1.default.create(context);
            const mode = context.options[0] || 'always';
            return {
                'VariableDeclaration:exit'(node) {
                    var _a, _b, _c;
                    if (mode === 'always') {
                        if (node.declare) {
                            return;
                        }
                        if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.TSModuleBlock &&
                            ((_b = node.parent.parent) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration &&
                            ((_c = node.parent.parent) === null || _c === void 0 ? void 0 : _c.declare)) {
                            return;
                        }
                    }
                    rules['VariableDeclaration:exit'](node);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/keyword-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const keyword_spacing_1 = __importDefault(require(456) /* eslint/lib/rules/keyword-spacing */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'keyword-spacing',
        meta: {
            type: 'layout',
            docs: {
                description: 'Enforce consistent spacing before and after keywords',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: 'whitespace',
            schema: keyword_spacing_1.default.meta.schema,
            messages: (_a = keyword_spacing_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                expectedBefore: 'Expected space(s) before "{{value}}".',
                expectedAfter: 'Expected space(s) after "{{value}}".',
                unexpectedBefore: 'Unexpected space(s) before "{{value}}".',
                unexpectedAfter: 'Unexpected space(s) after "{{value}}".',
            },
        },
        defaultOptions: [{}],
        create(context) {
            const sourceCode = context.getSourceCode();
            const baseRules = keyword_spacing_1.default.create(context);
            return Object.assign(Object.assign({}, baseRules), { TSAsExpression(node) {
                    const asToken = util.nullThrows(sourceCode.getTokenAfter(node.expression, token => token.value === 'as'), util.NullThrowsReasons.MissingToken('as', node.type));
                    const oldTokenType = asToken.type;
                    // as is a contextual keyword, so it's always reported as an Identifier
                    // the rule looks for keyword tokens, so we temporarily override it
                    // we mutate it at the token level because the rule calls sourceCode.getFirstToken,
                    // so mutating a copy would not change the underlying copy returned by that method
                    asToken.type = experimental_utils_1.AST_TOKEN_TYPES.Keyword;
                    // use this selector just because it is just a call to `checkSpacingAroundFirstToken`
                    baseRules.DebuggerStatement(asToken);
                    // make sure to reset the type afterward so we don't permanently mutate the AST
                    asToken.type = oldTokenType;
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/lines-between-class-members.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const lines_between_class_members_1 = __importDefault(require(461) /* eslint/lib/rules/lines-between-class-members */);
    const util = __importStar(require(906) /* ../util */);
    const schema = util.deepMerge(Object.assign({}, lines_between_class_members_1.default.meta.schema), {
        1: {
            exceptAfterOverload: {
                type: 'boolean',
                default: true,
            },
        },
    });
    exports.default = util.createRule({
        name: 'lines-between-class-members',
        meta: {
            type: 'layout',
            docs: {
                description: 'Require or disallow an empty line between class members',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: 'whitespace',
            schema,
            messages: (_a = lines_between_class_members_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                never: 'Unexpected blank line between class members.',
                always: 'Expected blank line between class members.',
            },
        },
        defaultOptions: [
            'always',
            {
                exceptAfterOverload: true,
                exceptAfterSingleLine: false,
            },
        ],
        create(context, options) {
            var _a;
            const rules = lines_between_class_members_1.default.create(context);
            const exceptAfterOverload = ((_a = options[1]) === null || _a === void 0 ? void 0 : _a.exceptAfterOverload) && options[0] === 'always';
            function isOverload(node) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition &&
                    node.value.type === experimental_utils_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression);
            }
            return {
                ClassBody(node) {
                    const body = exceptAfterOverload
                        ? node.body.filter(node => !isOverload(node))
                        : node.body;
                    rules.ClassBody(Object.assign(Object.assign({}, node), { body }));
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/member-delimiter-style.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const definition = {
        type: 'object',
        properties: {
            multiline: {
                type: 'object',
                properties: {
                    delimiter: { enum: ['none', 'semi', 'comma'] },
                    requireLast: { type: 'boolean' },
                },
                additionalProperties: false,
            },
            singleline: {
                type: 'object',
                properties: {
                    // note can't have "none" for single line delimiter as it's invalid syntax
                    delimiter: { enum: ['semi', 'comma'] },
                    requireLast: { type: 'boolean' },
                },
                additionalProperties: false,
            },
        },
        additionalProperties: false,
    };
    const isLastTokenEndOfLine = (token, line) => {
        const positionInLine = line.indexOf(token);
        return positionInLine === line.length - 1;
    };
    const makeFixFunction = ({ optsNone, optsSemi, lastToken, missingDelimiter, lastTokenLine, isSingleLine, }) => {
        // if removing is the action but last token is not the end of the line
        if (optsNone &&
            !isLastTokenEndOfLine(lastToken.value, lastTokenLine) &&
            !isSingleLine) {
            return null;
        }
        return (fixer) => {
            if (optsNone) {
                // remove the unneeded token
                return fixer.remove(lastToken);
            }
            const token = optsSemi ? ';' : ',';
            if (missingDelimiter) {
                // add the missing delimiter
                return fixer.insertTextAfter(lastToken, token);
            }
            // correct the current delimiter
            return fixer.replaceText(lastToken, token);
        };
    };
    exports.default = util.createRule({
        name: 'member-delimiter-style',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Require a specific member delimiter style for interfaces and type literals',
                category: 'Stylistic Issues',
                recommended: false,
            },
            fixable: 'code',
            messages: {
                unexpectedComma: 'Unexpected separator (,).',
                unexpectedSemi: 'Unexpected separator (;).',
                expectedComma: 'Expected a comma.',
                expectedSemi: 'Expected a semicolon.',
            },
            schema: [
                {
                    type: 'object',
                    properties: Object.assign({}, definition.properties, {
                        overrides: {
                            type: 'object',
                            properties: {
                                interface: definition,
                                typeLiteral: definition,
                            },
                            additionalProperties: false,
                        },
                        multilineDetection: {
                            enum: ['brackets', 'last-member'],
                        },
                    }),
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                multiline: {
                    delimiter: 'semi',
                    requireLast: true,
                },
                singleline: {
                    delimiter: 'semi',
                    requireLast: false,
                },
                multilineDetection: 'brackets',
            },
        ],
        create(context, [options]) {
            var _a;
            const sourceCode = context.getSourceCode();
            // use the base options as the defaults for the cases
            const baseOptions = options;
            const overrides = (_a = baseOptions.overrides) !== null && _a !== void 0 ? _a : {};
            const interfaceOptions = util.deepMerge(baseOptions, overrides.interface);
            const typeLiteralOptions = util.deepMerge(baseOptions, overrides.typeLiteral);
            /**
             * Check the last token in the given member.
             * @param member the member to be evaluated.
             * @param opts the options to be validated.
             * @param isLast a flag indicating `member` is the last in the interface or type literal.
             */
            function checkLastToken(member, opts, isLast) {
                /**
                 * Resolves the boolean value for the given setting enum value
                 * @param type the option name
                 */
                function getOption(type) {
                    if (isLast && !opts.requireLast) {
                        // only turn the option on if its expecting no delimiter for the last member
                        return type === 'none';
                    }
                    return opts.delimiter === type;
                }
                let messageId = null;
                let missingDelimiter = false;
                const lastToken = sourceCode.getLastToken(member, {
                    includeComments: false,
                });
                if (!lastToken) {
                    return;
                }
                const sourceCodeLines = sourceCode.getLines();
                const lastTokenLine = sourceCodeLines[(lastToken === null || lastToken === void 0 ? void 0 : lastToken.loc.start.line) - 1];
                const optsSemi = getOption('semi');
                const optsComma = getOption('comma');
                const optsNone = getOption('none');
                if (lastToken.value === ';') {
                    if (optsComma) {
                        messageId = 'expectedComma';
                    }
                    else if (optsNone) {
                        missingDelimiter = true;
                        messageId = 'unexpectedSemi';
                    }
                }
                else if (lastToken.value === ',') {
                    if (optsSemi) {
                        messageId = 'expectedSemi';
                    }
                    else if (optsNone) {
                        missingDelimiter = true;
                        messageId = 'unexpectedComma';
                    }
                }
                else {
                    if (optsSemi) {
                        missingDelimiter = true;
                        messageId = 'expectedSemi';
                    }
                    else if (optsComma) {
                        missingDelimiter = true;
                        messageId = 'expectedComma';
                    }
                }
                if (messageId) {
                    context.report({
                        node: lastToken,
                        loc: {
                            start: {
                                line: lastToken.loc.end.line,
                                column: lastToken.loc.end.column,
                            },
                            end: {
                                line: lastToken.loc.end.line,
                                column: lastToken.loc.end.column,
                            },
                        },
                        messageId,
                        fix: makeFixFunction({
                            optsNone,
                            optsSemi,
                            lastToken,
                            missingDelimiter,
                            lastTokenLine,
                            isSingleLine: opts.type === 'single-line',
                        }),
                    });
                }
            }
            /**
             * Check the member separator being used matches the delimiter.
             * @param {ASTNode} node the node to be evaluated.
             */
            function checkMemberSeparatorStyle(node) {
                const members = node.type === experimental_utils_1.AST_NODE_TYPES.TSInterfaceBody ? node.body : node.members;
                let isSingleLine = node.loc.start.line === node.loc.end.line;
                if (options.multilineDetection === 'last-member' &&
                    !isSingleLine &&
                    members.length > 0) {
                    const lastMember = members[members.length - 1];
                    if (lastMember.loc.end.line === node.loc.end.line) {
                        isSingleLine = true;
                    }
                }
                const typeOpts = node.type === experimental_utils_1.AST_NODE_TYPES.TSInterfaceBody
                    ? interfaceOptions
                    : typeLiteralOptions;
                const opts = isSingleLine
                    ? Object.assign(Object.assign({}, typeOpts.singleline), { type: 'single-line' }) : Object.assign(Object.assign({}, typeOpts.multiline), { type: 'multi-line' });
                members.forEach((member, index) => {
                    checkLastToken(member, opts !== null && opts !== void 0 ? opts : {}, index === members.length - 1);
                });
            }
            return {
                TSInterfaceBody: checkMemberSeparatorStyle,
                TSTypeLiteral: checkMemberSeparatorStyle,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/member-ordering.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const neverConfig = {
        type: 'string',
        enum: ['never'],
    };
    const arrayConfig = (memberTypes) => ({
        type: 'array',
        items: {
            enum: memberTypes,
        },
    });
    const objectConfig = (memberTypes) => ({
        type: 'object',
        properties: {
            memberTypes: {
                oneOf: [arrayConfig(memberTypes), neverConfig],
            },
            order: {
                type: 'string',
                enum: ['alphabetically', 'as-written'],
            },
        },
        additionalProperties: false,
    });
    exports.defaultOrder = [
        // Index signature
        'signature',
        'call-signature',
        // Fields
        'public-static-field',
        'protected-static-field',
        'private-static-field',
        'public-decorated-field',
        'protected-decorated-field',
        'private-decorated-field',
        'public-instance-field',
        'protected-instance-field',
        'private-instance-field',
        'public-abstract-field',
        'protected-abstract-field',
        'private-abstract-field',
        'public-field',
        'protected-field',
        'private-field',
        'static-field',
        'instance-field',
        'abstract-field',
        'decorated-field',
        'field',
        // Constructors
        'public-constructor',
        'protected-constructor',
        'private-constructor',
        'constructor',
        // Methods
        'public-static-method',
        'protected-static-method',
        'private-static-method',
        'public-decorated-method',
        'protected-decorated-method',
        'private-decorated-method',
        'public-instance-method',
        'protected-instance-method',
        'private-instance-method',
        'public-abstract-method',
        'protected-abstract-method',
        'private-abstract-method',
        'public-method',
        'protected-method',
        'private-method',
        'static-method',
        'instance-method',
        'abstract-method',
        'decorated-method',
        'method',
    ];
    const allMemberTypes = [
        'signature',
        'field',
        'method',
        'call-signature',
        'constructor',
    ].reduce((all, type) => {
        all.push(type);
        ['public', 'protected', 'private'].forEach(accessibility => {
            if (type !== 'signature') {
                all.push(`${accessibility}-${type}`); // e.g. `public-field`
            }
            // Only class instance fields and methods can have decorators attached to them
            if (type === 'field' || type === 'method') {
                const decoratedMemberType = `${accessibility}-decorated-${type}`;
                const decoratedMemberTypeNoAccessibility = `decorated-${type}`;
                if (!all.includes(decoratedMemberType)) {
                    all.push(decoratedMemberType);
                }
                if (!all.includes(decoratedMemberTypeNoAccessibility)) {
                    all.push(decoratedMemberTypeNoAccessibility);
                }
            }
            if (type !== 'constructor' && type !== 'signature') {
                // There is no `static-constructor` or `instance-constructor` or `abstract-constructor`
                ['static', 'instance', 'abstract'].forEach(scope => {
                    if (!all.includes(`${scope}-${type}`)) {
                        all.push(`${scope}-${type}`);
                    }
                    all.push(`${accessibility}-${scope}-${type}`);
                });
            }
        });
        return all;
    }, []);
    const functionExpressions = [
        experimental_utils_1.AST_NODE_TYPES.FunctionExpression,
        experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression,
    ];
    /**
     * Gets the node type.
     *
     * @param node the node to be evaluated.
     */
    function getNodeType(node) {
        switch (node.type) {
            case experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition:
            case experimental_utils_1.AST_NODE_TYPES.MethodDefinition:
                return node.kind;
            case experimental_utils_1.AST_NODE_TYPES.TSMethodSignature:
                return 'method';
            case experimental_utils_1.AST_NODE_TYPES.TSCallSignatureDeclaration:
                return 'call-signature';
            case experimental_utils_1.AST_NODE_TYPES.TSConstructSignatureDeclaration:
                return 'constructor';
            case experimental_utils_1.AST_NODE_TYPES.TSAbstractClassProperty:
            case experimental_utils_1.AST_NODE_TYPES.ClassProperty:
                return node.value && functionExpressions.includes(node.value.type)
                    ? 'method'
                    : 'field';
            case experimental_utils_1.AST_NODE_TYPES.TSPropertySignature:
                return 'field';
            case experimental_utils_1.AST_NODE_TYPES.TSIndexSignature:
                return 'signature';
            default:
                return null;
        }
    }
    /**
     * Gets the member name based on the member type.
     *
     * @param node the node to be evaluated.
     * @param sourceCode
     */
    function getMemberName(node, sourceCode) {
        switch (node.type) {
            case experimental_utils_1.AST_NODE_TYPES.TSPropertySignature:
            case experimental_utils_1.AST_NODE_TYPES.TSMethodSignature:
            case experimental_utils_1.AST_NODE_TYPES.TSAbstractClassProperty:
            case experimental_utils_1.AST_NODE_TYPES.ClassProperty:
                return util.getNameFromMember(node, sourceCode);
            case experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition:
            case experimental_utils_1.AST_NODE_TYPES.MethodDefinition:
                return node.kind === 'constructor'
                    ? 'constructor'
                    : util.getNameFromMember(node, sourceCode);
            case experimental_utils_1.AST_NODE_TYPES.TSConstructSignatureDeclaration:
                return 'new';
            case experimental_utils_1.AST_NODE_TYPES.TSCallSignatureDeclaration:
                return 'call';
            case experimental_utils_1.AST_NODE_TYPES.TSIndexSignature:
                return util.getNameFromIndexSignature(node);
            default:
                return null;
        }
    }
    /**
     * Gets the calculated rank using the provided method definition.
     * The algorithm is as follows:
     * - Get the rank based on the accessibility-scope-type name, e.g. public-instance-field
     * - If there is no order for accessibility-scope-type, then strip out the accessibility.
     * - If there is no order for scope-type, then strip out the scope.
     * - If there is no order for type, then return -1
     * @param memberGroups the valid names to be validated.
     * @param orderConfig the current order to be validated.
     *
     * @return Index of the matching member type in the order configuration.
     */
    function getRankOrder(memberGroups, orderConfig) {
        let rank = -1;
        const stack = memberGroups.slice(); // Get a copy of the member groups
        while (stack.length > 0 && rank === -1) {
            rank = orderConfig.indexOf(stack.shift());
        }
        return rank;
    }
    /**
     * Gets the rank of the node given the order.
     * @param node the node to be evaluated.
     * @param orderConfig the current order to be validated.
     * @param supportsModifiers a flag indicating whether the type supports modifiers (scope or accessibility) or not.
     */
    function getRank(node, orderConfig, supportsModifiers) {
        const type = getNodeType(node);
        if (type === null) {
            // shouldn't happen but just in case, put it on the end
            return orderConfig.length - 1;
        }
        const abstract = node.type === experimental_utils_1.AST_NODE_TYPES.TSAbstractClassProperty ||
            node.type === experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition;
        const scope = 'static' in node && node.static
            ? 'static'
            : abstract
                ? 'abstract'
                : 'instance';
        const accessibility = 'accessibility' in node && node.accessibility
            ? node.accessibility
            : 'public';
        // Collect all existing member groups (e.g. 'public-instance-field', 'instance-field', 'public-field', 'constructor' etc.)
        const memberGroups = [];
        if (supportsModifiers) {
            const decorated = 'decorators' in node && node.decorators.length > 0;
            if (decorated && (type === 'field' || type === 'method')) {
                memberGroups.push(`${accessibility}-decorated-${type}`);
                memberGroups.push(`decorated-${type}`);
            }
            if (type !== 'constructor') {
                // Constructors have no scope
                memberGroups.push(`${accessibility}-${scope}-${type}`);
                memberGroups.push(`${scope}-${type}`);
            }
            memberGroups.push(`${accessibility}-${type}`);
        }
        memberGroups.push(type);
        return getRankOrder(memberGroups, orderConfig);
    }
    /**
     * Gets the lowest possible rank higher than target.
     * e.g. given the following order:
     *   ...
     *   public-static-method
     *   protected-static-method
     *   private-static-method
     *   public-instance-method
     *   protected-instance-method
     *   private-instance-method
     *   ...
     * and considering that a public-instance-method has already been declared, so ranks contains
     * public-instance-method, then the lowest possible rank for public-static-method is
     * public-instance-method.
     * @param ranks the existing ranks in the object.
     * @param target the target rank.
     * @param order the current order to be validated.
     * @returns the name of the lowest possible rank without dashes (-).
     */
    function getLowestRank(ranks, target, order) {
        let lowest = ranks[ranks.length - 1];
        ranks.forEach(rank => {
            if (rank > target) {
                lowest = Math.min(lowest, rank);
            }
        });
        return order[lowest].replace(/-/g, ' ');
    }
    exports.default = util.createRule({
        name: 'member-ordering',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Require a consistent member declaration order',
                category: 'Stylistic Issues',
                recommended: false,
            },
            messages: {
                incorrectOrder: 'Member "{{member}}" should be declared before member "{{beforeMember}}".',
                incorrectGroupOrder: 'Member {{name}} should be declared before all {{rank}} definitions.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        default: {
                            oneOf: [
                                neverConfig,
                                arrayConfig(allMemberTypes),
                                objectConfig(allMemberTypes),
                            ],
                        },
                        classes: {
                            oneOf: [
                                neverConfig,
                                arrayConfig(allMemberTypes),
                                objectConfig(allMemberTypes),
                            ],
                        },
                        classExpressions: {
                            oneOf: [
                                neverConfig,
                                arrayConfig(allMemberTypes),
                                objectConfig(allMemberTypes),
                            ],
                        },
                        interfaces: {
                            oneOf: [
                                neverConfig,
                                arrayConfig(['signature', 'field', 'method', 'constructor']),
                                objectConfig(['signature', 'field', 'method', 'constructor']),
                            ],
                        },
                        typeLiterals: {
                            oneOf: [
                                neverConfig,
                                arrayConfig(['signature', 'field', 'method', 'constructor']),
                                objectConfig(['signature', 'field', 'method', 'constructor']),
                            ],
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                default: exports.defaultOrder,
            },
        ],
        create(context, [options]) {
            /**
             * Checks if the member groups are correctly sorted.
             *
             * @param members Members to be validated.
             * @param groupOrder Group order to be validated.
             * @param supportsModifiers A flag indicating whether the type supports modifiers (scope or accessibility) or not.
             *
             * @return Array of member groups or null if one of the groups is not correctly sorted.
             */
            function checkGroupSort(members, groupOrder, supportsModifiers) {
                const previousRanks = [];
                const memberGroups = [];
                let isCorrectlySorted = true;
                // Find first member which isn't correctly sorted
                members.forEach(member => {
                    const rank = getRank(member, groupOrder, supportsModifiers);
                    const name = getMemberName(member, context.getSourceCode());
                    const rankLastMember = previousRanks[previousRanks.length - 1];
                    if (rank === -1) {
                        return;
                    }
                    // Works for 1st item because x < undefined === false for any x (typeof string)
                    if (rank < rankLastMember) {
                        context.report({
                            node: member,
                            messageId: 'incorrectGroupOrder',
                            data: {
                                name,
                                rank: getLowestRank(previousRanks, rank, groupOrder),
                            },
                        });
                        isCorrectlySorted = false;
                    }
                    else if (rank === rankLastMember) {
                        // Same member group --> Push to existing member group array
                        memberGroups[memberGroups.length - 1].push(member);
                    }
                    else {
                        // New member group --> Create new member group array
                        previousRanks.push(rank);
                        memberGroups.push([member]);
                    }
                });
                return isCorrectlySorted ? memberGroups : null;
            }
            /**
             * Checks if the members are alphabetically sorted.
             *
             * @param members Members to be validated.
             *
             * @return True if all members are correctly sorted.
             */
            function checkAlphaSort(members) {
                let previousName = '';
                let isCorrectlySorted = true;
                // Find first member which isn't correctly sorted
                members.forEach(member => {
                    const name = getMemberName(member, context.getSourceCode());
                    // Note: Not all members have names
                    if (name) {
                        if (name < previousName) {
                            context.report({
                                node: member,
                                messageId: 'incorrectOrder',
                                data: {
                                    member: name,
                                    beforeMember: previousName,
                                },
                            });
                            isCorrectlySorted = false;
                        }
                        previousName = name;
                    }
                });
                return isCorrectlySorted;
            }
            /**
             * Validates if all members are correctly sorted.
             *
             * @param members Members to be validated.
             * @param orderConfig Order config to be validated.
             * @param supportsModifiers A flag indicating whether the type supports modifiers (scope or accessibility) or not.
             */
            function validateMembersOrder(members, orderConfig, supportsModifiers) {
                if (orderConfig === 'never') {
                    return;
                }
                // Standardize config
                let order = null;
                let memberTypes;
                if (Array.isArray(orderConfig)) {
                    memberTypes = orderConfig;
                }
                else {
                    order = orderConfig.order;
                    memberTypes = orderConfig.memberTypes;
                }
                // Check order
                if (Array.isArray(memberTypes)) {
                    const grouped = checkGroupSort(members, memberTypes, supportsModifiers);
                    if (grouped === null) {
                        return;
                    }
                    if (order === 'alphabetically') {
                        grouped.some(groupMember => !checkAlphaSort(groupMember));
                    }
                }
                else if (order === 'alphabetically') {
                    checkAlphaSort(members);
                }
            }
            return {
                ClassDeclaration(node) {
                    var _a;
                    validateMembersOrder(node.body.body, (_a = options.classes) !== null && _a !== void 0 ? _a : options.default, true);
                },
                ClassExpression(node) {
                    var _a;
                    validateMembersOrder(node.body.body, (_a = options.classExpressions) !== null && _a !== void 0 ? _a : options.default, true);
                },
                TSInterfaceDeclaration(node) {
                    var _a;
                    validateMembersOrder(node.body.body, (_a = options.interfaces) !== null && _a !== void 0 ? _a : options.default, false);
                },
                TSTypeLiteral(node) {
                    var _a;
                    validateMembersOrder(node.members, (_a = options.typeLiterals) !== null && _a !== void 0 ? _a : options.default, false);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/method-signature-style.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'method-signature-style',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforces using a particular method signature syntax.',
                category: 'Best Practices',
                recommended: false,
            },
            fixable: 'code',
            messages: {
                errorMethod: 'Shorthand method signature is forbidden. Use a function property instead.',
                errorProperty: 'Function property signature is forbidden. Use a method shorthand instead.',
            },
            schema: [
                {
                    enum: ['property', 'method'],
                },
            ],
        },
        defaultOptions: ['property'],
        create(context, [mode]) {
            const sourceCode = context.getSourceCode();
            function getMethodKey(node) {
                let key = sourceCode.getText(node.key);
                if (node.computed) {
                    key = `[${key}]`;
                }
                if (node.optional) {
                    key = `${key}?`;
                }
                if (node.readonly) {
                    key = `readonly ${key}`;
                }
                return key;
            }
            function getMethodParams(node) {
                let params = '()';
                if (node.params.length > 0) {
                    const openingParen = util.nullThrows(sourceCode.getTokenBefore(node.params[0], util.isOpeningParenToken), 'Missing opening paren before first parameter');
                    const closingParen = util.nullThrows(sourceCode.getTokenAfter(node.params[node.params.length - 1], util.isClosingParenToken), 'Missing closing paren after last parameter');
                    params = sourceCode.text.substring(openingParen.range[0], closingParen.range[1]);
                }
                if (node.typeParameters != null) {
                    const typeParams = sourceCode.getText(node.typeParameters);
                    params = `${typeParams}${params}`;
                }
                return params;
            }
            function getMethodReturnType(node) {
                return node.returnType == null
                    ? // if the method has no return type, it implicitly has an `any` return type
                        // we just make it explicit here so we can do the fix
                        'any'
                    : sourceCode.getText(node.returnType.typeAnnotation);
            }
            function getDelimiter(node) {
                const lastToken = sourceCode.getLastToken(node);
                if (lastToken &&
                    (util.isSemicolonToken(lastToken) || util.isCommaToken(lastToken))) {
                    return lastToken.value;
                }
                return '';
            }
            function isNodeParentModuleDeclaration(node) {
                if (!node.parent) {
                    return false;
                }
                if (node.parent.type === experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration) {
                    return true;
                }
                if (node.parent.type === experimental_utils_1.AST_NODE_TYPES.Program) {
                    return false;
                }
                return isNodeParentModuleDeclaration(node.parent);
            }
            return {
                TSMethodSignature(methodNode) {
                    if (mode === 'method') {
                        return;
                    }
                    const parent = methodNode.parent;
                    const members = (parent === null || parent === void 0 ? void 0 : parent.type) === experimental_utils_1.AST_NODE_TYPES.TSInterfaceBody
                        ? parent.body
                        : (parent === null || parent === void 0 ? void 0 : parent.type) === experimental_utils_1.AST_NODE_TYPES.TSTypeLiteral
                            ? parent.members
                            : [];
                    const duplicatedKeyMethodNodes = members.filter((element) => element.type === experimental_utils_1.AST_NODE_TYPES.TSMethodSignature &&
                        element !== methodNode &&
                        getMethodKey(element) === getMethodKey(methodNode));
                    const isParentModule = isNodeParentModuleDeclaration(methodNode);
                    if (duplicatedKeyMethodNodes.length > 0) {
                        if (isParentModule) {
                            context.report({
                                node: methodNode,
                                messageId: 'errorMethod',
                            });
                        }
                        else {
                            context.report({
                                node: methodNode,
                                messageId: 'errorMethod',
                                *fix(fixer) {
                                    const methodNodes = [
                                        methodNode,
                                        ...duplicatedKeyMethodNodes,
                                    ].sort((a, b) => (a.range[0] < b.range[0] ? -1 : 1));
                                    const typeString = methodNodes
                                        .map(node => {
                                        const params = getMethodParams(node);
                                        const returnType = getMethodReturnType(node);
                                        return `(${params} => ${returnType})`;
                                    })
                                        .join(' & ');
                                    const key = getMethodKey(methodNode);
                                    const delimiter = getDelimiter(methodNode);
                                    yield fixer.replaceText(methodNode, `${key}: ${typeString}${delimiter}`);
                                    for (const node of duplicatedKeyMethodNodes) {
                                        const lastToken = sourceCode.getLastToken(node);
                                        if (lastToken) {
                                            const nextToken = sourceCode.getTokenAfter(lastToken);
                                            if (nextToken) {
                                                yield fixer.remove(node);
                                                yield fixer.replaceTextRange([lastToken.range[1], nextToken.range[0]], '');
                                            }
                                        }
                                    }
                                },
                            });
                        }
                        return;
                    }
                    if (isParentModule) {
                        context.report({
                            node: methodNode,
                            messageId: 'errorMethod',
                        });
                    }
                    else {
                        context.report({
                            node: methodNode,
                            messageId: 'errorMethod',
                            fix: fixer => {
                                const key = getMethodKey(methodNode);
                                const params = getMethodParams(methodNode);
                                const returnType = getMethodReturnType(methodNode);
                                const delimiter = getDelimiter(methodNode);
                                return fixer.replaceText(methodNode, `${key}: ${params} => ${returnType}${delimiter}`);
                            },
                        });
                    }
                },
                TSPropertySignature(propertyNode) {
                    var _a;
                    const typeNode = (_a = propertyNode.typeAnnotation) === null || _a === void 0 ? void 0 : _a.typeAnnotation;
                    if ((typeNode === null || typeNode === void 0 ? void 0 : typeNode.type) !== experimental_utils_1.AST_NODE_TYPES.TSFunctionType) {
                        return;
                    }
                    if (mode === 'property') {
                        return;
                    }
                    context.report({
                        node: propertyNode,
                        messageId: 'errorProperty',
                        fix: fixer => {
                            const key = getMethodKey(propertyNode);
                            const params = getMethodParams(typeNode);
                            const returnType = getMethodReturnType(typeNode);
                            const delimiter = getDelimiter(propertyNode);
                            return fixer.replaceText(propertyNode, `${key}${params}: ${returnType}${delimiter}`);
                        },
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/naming-convention.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const scope_manager_1 = require(768) /* @typescript-eslint/scope-manager */;
    const util = __importStar(require(906) /* ../util */);
    const naming_convention_utils_1 = require(969) /* ./naming-convention-utils */;
    // This essentially mirrors ESLint's `camelcase` rule
    // note that that rule ignores leading and trailing underscores and only checks those in the middle of a variable name
    const defaultCamelCaseAllTheThingsConfig = [
        {
            selector: 'default',
            format: ['camelCase'],
            leadingUnderscore: 'allow',
            trailingUnderscore: 'allow',
        },
        {
            selector: 'variable',
            format: ['camelCase', 'UPPER_CASE'],
            leadingUnderscore: 'allow',
            trailingUnderscore: 'allow',
        },
        {
            selector: 'typeLike',
            format: ['PascalCase'],
        },
    ];
    exports.default = util.createRule({
        name: 'naming-convention',
        meta: {
            docs: {
                category: 'Variables',
                description: 'Enforces naming conventions for everything across a codebase',
                recommended: false,
                // technically only requires type checking if the user uses "type" modifiers
                requiresTypeChecking: true,
            },
            type: 'suggestion',
            messages: {
                unexpectedUnderscore: '{{type}} name `{{name}}` must not have a {{position}} underscore.',
                missingUnderscore: '{{type}} name `{{name}}` must have {{count}} {{position}} underscore(s).',
                missingAffix: '{{type}} name `{{name}}` must have one of the following {{position}}es: {{affixes}}',
                satisfyCustom: '{{type}} name `{{name}}` must {{regexMatch}} the RegExp: {{regex}}',
                doesNotMatchFormat: '{{type}} name `{{name}}` must match one of the following formats: {{formats}}',
                doesNotMatchFormatTrimmed: '{{type}} name `{{name}}` trimmed as `{{processedName}}` must match one of the following formats: {{formats}}',
            },
            schema: naming_convention_utils_1.SCHEMA,
        },
        defaultOptions: defaultCamelCaseAllTheThingsConfig,
        create(contextWithoutDefaults) {
            const context = contextWithoutDefaults.options &&
                contextWithoutDefaults.options.length > 0
                ? contextWithoutDefaults
                : // only apply the defaults when the user provides no config
                    Object.setPrototypeOf({
                        options: defaultCamelCaseAllTheThingsConfig,
                    }, contextWithoutDefaults);
            const validators = naming_convention_utils_1.parseOptions(context);
            // getParserServices(context, false) -- dirty hack to work around the docs checker test...
            const compilerOptions = util
                .getParserServices(context, true)
                .program.getCompilerOptions();
            function handleMember(validator, node, modifiers) {
                if (!validator) {
                    return;
                }
                const key = node.key;
                if (requiresQuoting(key, compilerOptions.target)) {
                    modifiers.add(naming_convention_utils_1.Modifiers.requiresQuotes);
                }
                validator(key, modifiers);
            }
            function getMemberModifiers(node) {
                const modifiers = new Set();
                if (node.accessibility) {
                    modifiers.add(naming_convention_utils_1.Modifiers[node.accessibility]);
                }
                else {
                    modifiers.add(naming_convention_utils_1.Modifiers.public);
                }
                if (node.static) {
                    modifiers.add(naming_convention_utils_1.Modifiers.static);
                }
                if ('readonly' in node && node.readonly) {
                    modifiers.add(naming_convention_utils_1.Modifiers.readonly);
                }
                if (node.type === experimental_utils_1.AST_NODE_TYPES.TSAbstractClassProperty ||
                    node.type === experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition) {
                    modifiers.add(naming_convention_utils_1.Modifiers.abstract);
                }
                return modifiers;
            }
            const unusedVariables = util.collectUnusedVariables(context);
            function isUnused(name, initialScope = context.getScope()) {
                var _a;
                let variable = null;
                let scope = initialScope;
                while (scope) {
                    variable = (_a = scope.set.get(name)) !== null && _a !== void 0 ? _a : null;
                    if (variable) {
                        break;
                    }
                    scope = scope.upper;
                }
                if (!variable) {
                    return false;
                }
                return unusedVariables.has(variable);
            }
            function isDestructured(id) {
                var _a, _b, _c;
                return (
                // `const { x }`
                // does not match `const { x: y }`
                (((_a = id.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.Property && id.parent.shorthand) ||
                    // `const { x = 2 }`
                    // does not match const `{ x: y = 2 }`
                    (((_b = id.parent) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.AssignmentPattern &&
                        ((_c = id.parent.parent) === null || _c === void 0 ? void 0 : _c.type) === experimental_utils_1.AST_NODE_TYPES.Property &&
                        id.parent.parent.shorthand));
            }
            return {
                // #region variable
                VariableDeclarator(node) {
                    const validator = validators.variable;
                    if (!validator) {
                        return;
                    }
                    const identifiers = getIdentifiersFromPattern(node.id);
                    const baseModifiers = new Set();
                    const parent = node.parent;
                    if ((parent === null || parent === void 0 ? void 0 : parent.type) === experimental_utils_1.AST_NODE_TYPES.VariableDeclaration) {
                        if (parent.kind === 'const') {
                            baseModifiers.add(naming_convention_utils_1.Modifiers.const);
                        }
                        if (isGlobal(context.getScope())) {
                            baseModifiers.add(naming_convention_utils_1.Modifiers.global);
                        }
                    }
                    identifiers.forEach(id => {
                        const modifiers = new Set(baseModifiers);
                        if (isDestructured(id)) {
                            modifiers.add(naming_convention_utils_1.Modifiers.destructured);
                        }
                        if (isExported(parent, id.name, context.getScope())) {
                            modifiers.add(naming_convention_utils_1.Modifiers.exported);
                        }
                        if (isUnused(id.name)) {
                            modifiers.add(naming_convention_utils_1.Modifiers.unused);
                        }
                        validator(id, modifiers);
                    });
                },
                // #endregion
                // #region function
                'FunctionDeclaration, TSDeclareFunction, FunctionExpression'(node) {
                    const validator = validators.function;
                    if (!validator || node.id === null) {
                        return;
                    }
                    const modifiers = new Set();
                    // functions create their own nested scope
                    const scope = context.getScope().upper;
                    if (isGlobal(scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.global);
                    }
                    if (isExported(node, node.id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.exported);
                    }
                    if (isUnused(node.id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.unused);
                    }
                    validator(node.id, modifiers);
                },
                // #endregion function
                // #region parameter
                'FunctionDeclaration, TSDeclareFunction, TSEmptyBodyFunctionExpression, FunctionExpression, ArrowFunctionExpression'(node) {
                    const validator = validators.parameter;
                    if (!validator) {
                        return;
                    }
                    node.params.forEach(param => {
                        if (param.type === experimental_utils_1.AST_NODE_TYPES.TSParameterProperty) {
                            return;
                        }
                        const identifiers = getIdentifiersFromPattern(param);
                        identifiers.forEach(i => {
                            const modifiers = new Set();
                            if (isDestructured(i)) {
                                modifiers.add(naming_convention_utils_1.Modifiers.destructured);
                            }
                            if (isUnused(i.name)) {
                                modifiers.add(naming_convention_utils_1.Modifiers.unused);
                            }
                            validator(i, modifiers);
                        });
                    });
                },
                // #endregion parameter
                // #region parameterProperty
                TSParameterProperty(node) {
                    const validator = validators.parameterProperty;
                    if (!validator) {
                        return;
                    }
                    const modifiers = getMemberModifiers(node);
                    const identifiers = getIdentifiersFromPattern(node.parameter);
                    identifiers.forEach(i => {
                        validator(i, modifiers);
                    });
                },
                // #endregion parameterProperty
                // #region property
                ':not(ObjectPattern) > Property[computed = false][kind = "init"][value.type != "ArrowFunctionExpression"][value.type != "FunctionExpression"][value.type != "TSEmptyBodyFunctionExpression"]'(node) {
                    const modifiers = new Set([naming_convention_utils_1.Modifiers.public]);
                    handleMember(validators.objectLiteralProperty, node, modifiers);
                },
                ':matches(ClassProperty, TSAbstractClassProperty)[computed = false][value.type != "ArrowFunctionExpression"][value.type != "FunctionExpression"][value.type != "TSEmptyBodyFunctionExpression"]'(node) {
                    const modifiers = getMemberModifiers(node);
                    handleMember(validators.classProperty, node, modifiers);
                },
                'TSPropertySignature[computed = false]'(node) {
                    const modifiers = new Set([naming_convention_utils_1.Modifiers.public]);
                    if (node.readonly) {
                        modifiers.add(naming_convention_utils_1.Modifiers.readonly);
                    }
                    handleMember(validators.typeProperty, node, modifiers);
                },
                // #endregion property
                // #region method
                [[
                    'Property[computed = false][kind = "init"][value.type = "ArrowFunctionExpression"]',
                    'Property[computed = false][kind = "init"][value.type = "FunctionExpression"]',
                    'Property[computed = false][kind = "init"][value.type = "TSEmptyBodyFunctionExpression"]',
                ].join(', ')](node) {
                    const modifiers = new Set([naming_convention_utils_1.Modifiers.public]);
                    handleMember(validators.objectLiteralMethod, node, modifiers);
                },
                [[
                    ':matches(ClassProperty, TSAbstractClassProperty)[computed = false][value.type = "ArrowFunctionExpression"]',
                    ':matches(ClassProperty, TSAbstractClassProperty)[computed = false][value.type = "FunctionExpression"]',
                    ':matches(ClassProperty, TSAbstractClassProperty)[computed = false][value.type = "TSEmptyBodyFunctionExpression"]',
                    ':matches(MethodDefinition, TSAbstractMethodDefinition)[computed = false][kind = "method"]',
                ].join(', ')](node) {
                    const modifiers = getMemberModifiers(node);
                    handleMember(validators.classMethod, node, modifiers);
                },
                'TSMethodSignature[computed = false]'(node) {
                    const modifiers = new Set([naming_convention_utils_1.Modifiers.public]);
                    handleMember(validators.typeMethod, node, modifiers);
                },
                // #endregion method
                // #region accessor
                'Property[computed = false]:matches([kind = "get"], [kind = "set"])'(node) {
                    const modifiers = new Set([naming_convention_utils_1.Modifiers.public]);
                    handleMember(validators.accessor, node, modifiers);
                },
                'MethodDefinition[computed = false]:matches([kind = "get"], [kind = "set"])'(node) {
                    const modifiers = getMemberModifiers(node);
                    handleMember(validators.accessor, node, modifiers);
                },
                // #endregion accessor
                // #region enumMember
                // computed is optional, so can't do [computed = false]
                'TSEnumMember[computed != true]'(node) {
                    const validator = validators.enumMember;
                    if (!validator) {
                        return;
                    }
                    const id = node.id;
                    const modifiers = new Set();
                    if (requiresQuoting(id, compilerOptions.target)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.requiresQuotes);
                    }
                    validator(id, modifiers);
                },
                // #endregion enumMember
                // #region class
                'ClassDeclaration, ClassExpression'(node) {
                    const validator = validators.class;
                    if (!validator) {
                        return;
                    }
                    const id = node.id;
                    if (id === null) {
                        return;
                    }
                    const modifiers = new Set();
                    // classes create their own nested scope
                    const scope = context.getScope().upper;
                    if (node.abstract) {
                        modifiers.add(naming_convention_utils_1.Modifiers.abstract);
                    }
                    if (isExported(node, id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.exported);
                    }
                    if (isUnused(id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.unused);
                    }
                    validator(id, modifiers);
                },
                // #endregion class
                // #region interface
                TSInterfaceDeclaration(node) {
                    const validator = validators.interface;
                    if (!validator) {
                        return;
                    }
                    const modifiers = new Set();
                    const scope = context.getScope();
                    if (isExported(node, node.id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.exported);
                    }
                    if (isUnused(node.id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.unused);
                    }
                    validator(node.id, modifiers);
                },
                // #endregion interface
                // #region typeAlias
                TSTypeAliasDeclaration(node) {
                    const validator = validators.typeAlias;
                    if (!validator) {
                        return;
                    }
                    const modifiers = new Set();
                    const scope = context.getScope();
                    if (isExported(node, node.id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.exported);
                    }
                    if (isUnused(node.id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.unused);
                    }
                    validator(node.id, modifiers);
                },
                // #endregion typeAlias
                // #region enum
                TSEnumDeclaration(node) {
                    const validator = validators.enum;
                    if (!validator) {
                        return;
                    }
                    const modifiers = new Set();
                    // enums create their own nested scope
                    const scope = context.getScope().upper;
                    if (isExported(node, node.id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.exported);
                    }
                    if (isUnused(node.id.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.unused);
                    }
                    validator(node.id, modifiers);
                },
                // #endregion enum
                // #region typeParameter
                'TSTypeParameterDeclaration > TSTypeParameter'(node) {
                    const validator = validators.typeParameter;
                    if (!validator) {
                        return;
                    }
                    const modifiers = new Set();
                    const scope = context.getScope();
                    if (isUnused(node.name.name, scope)) {
                        modifiers.add(naming_convention_utils_1.Modifiers.unused);
                    }
                    validator(node.name, modifiers);
                },
                // #endregion typeParameter
            };
        },
    });
    function getIdentifiersFromPattern(pattern) {
        const identifiers = [];
        const visitor = new scope_manager_1.PatternVisitor({}, pattern, id => identifiers.push(id));
        visitor.visit(pattern);
        return identifiers;
    }
    function isExported(node, name, scope) {
        var _a, _b;
        if (((_a = node === null || node === void 0 ? void 0 : node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.ExportDefaultDeclaration ||
            ((_b = node === null || node === void 0 ? void 0 : node.parent) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.ExportNamedDeclaration) {
            return true;
        }
        if (scope == null) {
            return false;
        }
        const variable = scope.set.get(name);
        if (variable) {
            for (const ref of variable.references) {
                const refParent = ref.identifier.parent;
                if ((refParent === null || refParent === void 0 ? void 0 : refParent.type) === experimental_utils_1.AST_NODE_TYPES.ExportDefaultDeclaration ||
                    (refParent === null || refParent === void 0 ? void 0 : refParent.type) === experimental_utils_1.AST_NODE_TYPES.ExportSpecifier) {
                    return true;
                }
            }
        }
        return false;
    }
    function isGlobal(scope) {
        if (scope == null) {
            return false;
        }
        return (scope.type === experimental_utils_1.TSESLint.Scope.ScopeType.global ||
            scope.type === experimental_utils_1.TSESLint.Scope.ScopeType.module);
    }
    function requiresQuoting(node, target) {
        const name = node.type === experimental_utils_1.AST_NODE_TYPES.Identifier ? node.name : `${node.value}`;
        return util.requiresQuoting(name, target);
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/naming-convention-utils/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var enums_1 = require(970) /* ./enums */;
    __esExport("Modifiers", enums_1.Modifiers);
    var schema_1 = require(971) /* ./schema */;
    __esExport("SCHEMA", schema_1.SCHEMA);
    var shared_1 = require(972) /* ./shared */;
    __esExport("selectorTypeToMessageString", shared_1.selectorTypeToMessageString);
    var parse_options_1 = require(973) /* ./parse-options */;
    __esExport("parseOptions", parse_options_1.parseOptions);
},
/* @typescript-eslint/eslint-plugin/dist/rules/naming-convention-utils/enums.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var PredefinedFormats;
    (function (PredefinedFormats) {
        PredefinedFormats[PredefinedFormats["camelCase"] = 1] = "camelCase";
        PredefinedFormats[PredefinedFormats["strictCamelCase"] = 2] = "strictCamelCase";
        PredefinedFormats[PredefinedFormats["PascalCase"] = 3] = "PascalCase";
        PredefinedFormats[PredefinedFormats["StrictPascalCase"] = 4] = "StrictPascalCase";
        PredefinedFormats[PredefinedFormats["snake_case"] = 5] = "snake_case";
        PredefinedFormats[PredefinedFormats["UPPER_CASE"] = 6] = "UPPER_CASE";
    })(PredefinedFormats || (PredefinedFormats = {}));
    exports.PredefinedFormats = PredefinedFormats;
    var UnderscoreOptions;
    (function (UnderscoreOptions) {
        UnderscoreOptions[UnderscoreOptions["forbid"] = 1] = "forbid";
        UnderscoreOptions[UnderscoreOptions["allow"] = 2] = "allow";
        UnderscoreOptions[UnderscoreOptions["require"] = 3] = "require";
        // special cases as it's common practice to use double underscore
        UnderscoreOptions[UnderscoreOptions["requireDouble"] = 4] = "requireDouble";
        UnderscoreOptions[UnderscoreOptions["allowDouble"] = 5] = "allowDouble";
        UnderscoreOptions[UnderscoreOptions["allowSingleOrDouble"] = 6] = "allowSingleOrDouble";
    })(UnderscoreOptions || (UnderscoreOptions = {}));
    exports.UnderscoreOptions = UnderscoreOptions;
    var Selectors;
    (function (Selectors) {
        // variableLike
        Selectors[Selectors["variable"] = 1] = "variable";
        Selectors[Selectors["function"] = 2] = "function";
        Selectors[Selectors["parameter"] = 4] = "parameter";
        // memberLike
        Selectors[Selectors["parameterProperty"] = 8] = "parameterProperty";
        Selectors[Selectors["accessor"] = 16] = "accessor";
        Selectors[Selectors["enumMember"] = 32] = "enumMember";
        Selectors[Selectors["classMethod"] = 64] = "classMethod";
        Selectors[Selectors["objectLiteralMethod"] = 128] = "objectLiteralMethod";
        Selectors[Selectors["typeMethod"] = 256] = "typeMethod";
        Selectors[Selectors["classProperty"] = 512] = "classProperty";
        Selectors[Selectors["objectLiteralProperty"] = 1024] = "objectLiteralProperty";
        Selectors[Selectors["typeProperty"] = 2048] = "typeProperty";
        // typeLike
        Selectors[Selectors["class"] = 4096] = "class";
        Selectors[Selectors["interface"] = 8192] = "interface";
        Selectors[Selectors["typeAlias"] = 16384] = "typeAlias";
        Selectors[Selectors["enum"] = 32768] = "enum";
        Selectors[Selectors["typeParameter"] = 131072] = "typeParameter";
    })(Selectors || (Selectors = {}));
    exports.Selectors = Selectors;
    var MetaSelectors;
    (function (MetaSelectors) {
        MetaSelectors[MetaSelectors["default"] = -1] = "default";
        MetaSelectors[MetaSelectors["variableLike"] = 7] = "variableLike";
        MetaSelectors[MetaSelectors["memberLike"] = 4088] = "memberLike";
        MetaSelectors[MetaSelectors["typeLike"] = 192512] = "typeLike";
        MetaSelectors[MetaSelectors["method"] = 448] = "method";
        MetaSelectors[MetaSelectors["property"] = 3584] = "property";
    })(MetaSelectors || (MetaSelectors = {}));
    exports.MetaSelectors = MetaSelectors;
    var Modifiers;
    (function (Modifiers) {
        // const variable
        Modifiers[Modifiers["const"] = 1] = "const";
        // readonly members
        Modifiers[Modifiers["readonly"] = 2] = "readonly";
        // static members
        Modifiers[Modifiers["static"] = 4] = "static";
        // member accessibility
        Modifiers[Modifiers["public"] = 8] = "public";
        Modifiers[Modifiers["protected"] = 16] = "protected";
        Modifiers[Modifiers["private"] = 32] = "private";
        Modifiers[Modifiers["abstract"] = 64] = "abstract";
        // destructured variable
        Modifiers[Modifiers["destructured"] = 128] = "destructured";
        // variables declared in the top-level scope
        Modifiers[Modifiers["global"] = 256] = "global";
        // things that are exported
        Modifiers[Modifiers["exported"] = 512] = "exported";
        // things that are unused
        Modifiers[Modifiers["unused"] = 1024] = "unused";
        // properties that require quoting
        Modifiers[Modifiers["requiresQuotes"] = 2048] = "requiresQuotes";
        // make sure TypeModifiers starts at Modifiers + 1 or else sorting won't work
    })(Modifiers || (Modifiers = {}));
    exports.Modifiers = Modifiers;
    var TypeModifiers;
    (function (TypeModifiers) {
        TypeModifiers[TypeModifiers["boolean"] = 4096] = "boolean";
        TypeModifiers[TypeModifiers["string"] = 8192] = "string";
        TypeModifiers[TypeModifiers["number"] = 16384] = "number";
        TypeModifiers[TypeModifiers["function"] = 32768] = "function";
        TypeModifiers[TypeModifiers["array"] = 65536] = "array";
    })(TypeModifiers || (TypeModifiers = {}));
    exports.TypeModifiers = TypeModifiers;
},
/* @typescript-eslint/eslint-plugin/dist/rules/naming-convention-utils/schema.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const enums_1 = require(970) /* ./enums */;
    const util = __importStar(require(906) /* ../../util */);
    const UNDERSCORE_SCHEMA = {
        type: 'string',
        enum: util.getEnumNames(enums_1.UnderscoreOptions),
    };
    const PREFIX_SUFFIX_SCHEMA = {
        type: 'array',
        items: {
            type: 'string',
            minLength: 1,
        },
        additionalItems: false,
    };
    const MATCH_REGEX_SCHEMA = {
        type: 'object',
        properties: {
            match: { type: 'boolean' },
            regex: { type: 'string' },
        },
        required: ['match', 'regex'],
    };
    const FORMAT_OPTIONS_PROPERTIES = {
        format: {
            oneOf: [
                {
                    type: 'array',
                    items: {
                        type: 'string',
                        enum: util.getEnumNames(enums_1.PredefinedFormats),
                    },
                    additionalItems: false,
                },
                {
                    type: 'null',
                },
            ],
        },
        custom: MATCH_REGEX_SCHEMA,
        leadingUnderscore: UNDERSCORE_SCHEMA,
        trailingUnderscore: UNDERSCORE_SCHEMA,
        prefix: PREFIX_SUFFIX_SCHEMA,
        suffix: PREFIX_SUFFIX_SCHEMA,
        failureMessage: {
            type: 'string',
        },
    };
    function selectorSchema(selectorString, allowType, modifiers) {
        const selector = {
            filter: {
                oneOf: [
                    {
                        type: 'string',
                        minLength: 1,
                    },
                    MATCH_REGEX_SCHEMA,
                ],
            },
            selector: {
                type: 'string',
                enum: [selectorString],
            },
        };
        if (modifiers && modifiers.length > 0) {
            selector.modifiers = {
                type: 'array',
                items: {
                    type: 'string',
                    enum: modifiers,
                },
                additionalItems: false,
            };
        }
        if (allowType) {
            selector.types = {
                type: 'array',
                items: {
                    type: 'string',
                    enum: util.getEnumNames(enums_1.TypeModifiers),
                },
                additionalItems: false,
            };
        }
        return [
            {
                type: 'object',
                properties: Object.assign(Object.assign({}, FORMAT_OPTIONS_PROPERTIES), selector),
                required: ['selector', 'format'],
                additionalProperties: false,
            },
        ];
    }
    function selectorsSchema() {
        return {
            type: 'object',
            properties: Object.assign(Object.assign({}, FORMAT_OPTIONS_PROPERTIES), {
                filter: {
                    oneOf: [
                        {
                            type: 'string',
                            minLength: 1,
                        },
                        MATCH_REGEX_SCHEMA,
                    ],
                },
                selector: {
                    type: 'array',
                    items: {
                        type: 'string',
                        enum: [
                            ...util.getEnumNames(enums_1.MetaSelectors),
                            ...util.getEnumNames(enums_1.Selectors),
                        ],
                    },
                    additionalItems: false,
                },
                modifiers: {
                    type: 'array',
                    items: {
                        type: 'string',
                        enum: util.getEnumNames(enums_1.Modifiers),
                    },
                    additionalItems: false,
                },
                types: {
                    type: 'array',
                    items: {
                        type: 'string',
                        enum: util.getEnumNames(enums_1.TypeModifiers),
                    },
                    additionalItems: false,
                },
            }),
            required: ['selector', 'format'],
            additionalProperties: false,
        };
    }
    const SCHEMA = {
        type: 'array',
        items: {
            oneOf: [
                selectorsSchema(),
                ...selectorSchema('default', false, util.getEnumNames(enums_1.Modifiers)),
                ...selectorSchema('variableLike', false, ['unused']),
                ...selectorSchema('variable', true, [
                    'const',
                    'destructured',
                    'exported',
                    'global',
                    'unused',
                ]),
                ...selectorSchema('function', false, ['exported', 'global', 'unused']),
                ...selectorSchema('parameter', true, ['destructured', 'unused']),
                ...selectorSchema('memberLike', false, [
                    'abstract',
                    'private',
                    'protected',
                    'public',
                    'readonly',
                    'requiresQuotes',
                    'static',
                ]),
                ...selectorSchema('classProperty', true, [
                    'abstract',
                    'private',
                    'protected',
                    'public',
                    'readonly',
                    'requiresQuotes',
                    'static',
                ]),
                ...selectorSchema('objectLiteralProperty', true, [
                    'public',
                    'requiresQuotes',
                ]),
                ...selectorSchema('typeProperty', true, [
                    'public',
                    'readonly',
                    'requiresQuotes',
                ]),
                ...selectorSchema('parameterProperty', true, [
                    'private',
                    'protected',
                    'public',
                    'readonly',
                ]),
                ...selectorSchema('property', true, [
                    'abstract',
                    'private',
                    'protected',
                    'public',
                    'readonly',
                    'requiresQuotes',
                    'static',
                ]),
                ...selectorSchema('classMethod', false, [
                    'abstract',
                    'private',
                    'protected',
                    'public',
                    'requiresQuotes',
                    'static',
                ]),
                ...selectorSchema('objectLiteralMethod', false, [
                    'public',
                    'requiresQuotes',
                ]),
                ...selectorSchema('typeMethod', false, ['public', 'requiresQuotes']),
                ...selectorSchema('method', false, [
                    'abstract',
                    'private',
                    'protected',
                    'public',
                    'requiresQuotes',
                    'static',
                ]),
                ...selectorSchema('accessor', true, [
                    'abstract',
                    'private',
                    'protected',
                    'public',
                    'requiresQuotes',
                    'static',
                ]),
                ...selectorSchema('enumMember', false, ['requiresQuotes']),
                ...selectorSchema('typeLike', false, ['abstract', 'exported', 'unused']),
                ...selectorSchema('class', false, ['abstract', 'exported', 'unused']),
                ...selectorSchema('interface', false, ['exported', 'unused']),
                ...selectorSchema('typeAlias', false, ['exported', 'unused']),
                ...selectorSchema('enum', false, ['exported', 'unused']),
                ...selectorSchema('typeParameter', false, ['unused']),
            ],
        },
        additionalItems: false,
    };
    exports.SCHEMA = SCHEMA;
},
/* @typescript-eslint/eslint-plugin/dist/rules/naming-convention-utils/shared.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const enums_1 = require(970) /* ./enums */;
    function selectorTypeToMessageString(selectorType) {
        const notCamelCase = selectorType.replace(/([A-Z])/g, ' $1');
        return notCamelCase.charAt(0).toUpperCase() + notCamelCase.slice(1);
    }
    exports.selectorTypeToMessageString = selectorTypeToMessageString;
    function isMetaSelector(selector) {
        return selector in enums_1.MetaSelectors;
    }
    exports.isMetaSelector = isMetaSelector;
    function isMethodOrPropertySelector(selector) {
        return (selector === enums_1.MetaSelectors.method || selector === enums_1.MetaSelectors.property);
    }
    exports.isMethodOrPropertySelector = isMethodOrPropertySelector;
},
/* @typescript-eslint/eslint-plugin/dist/rules/naming-convention-utils/parse-options.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../../util */);
    const enums_1 = require(970) /* ./enums */;
    const shared_1 = require(972) /* ./shared */;
    const validator_1 = require(974) /* ./validator */;
    function normalizeOption(option) {
        var _a, _b, _c, _d, _e, _f;
        let weight = 0;
        (_a = option.modifiers) === null || _a === void 0 ? void 0 : _a.forEach(mod => {
            weight |= enums_1.Modifiers[mod];
        });
        (_b = option.types) === null || _b === void 0 ? void 0 : _b.forEach(mod => {
            weight |= enums_1.TypeModifiers[mod];
        });
        // give selectors with a filter the _highest_ priority
        if (option.filter) {
            weight |= 1 << 30;
        }
        const normalizedOption = {
            // format options
            format: option.format ? option.format.map(f => enums_1.PredefinedFormats[f]) : null,
            custom: option.custom
                ? {
                    regex: new RegExp(option.custom.regex, 'u'),
                    match: option.custom.match,
                }
                : null,
            leadingUnderscore: option.leadingUnderscore !== undefined
                ? enums_1.UnderscoreOptions[option.leadingUnderscore]
                : null,
            trailingUnderscore: option.trailingUnderscore !== undefined
                ? enums_1.UnderscoreOptions[option.trailingUnderscore]
                : null,
            prefix: option.prefix && option.prefix.length > 0 ? option.prefix : null,
            suffix: option.suffix && option.suffix.length > 0 ? option.suffix : null,
            modifiers: (_d = (_c = option.modifiers) === null || _c === void 0 ? void 0 : _c.map(m => enums_1.Modifiers[m])) !== null && _d !== void 0 ? _d : null,
            types: (_f = (_e = option.types) === null || _e === void 0 ? void 0 : _e.map(m => enums_1.TypeModifiers[m])) !== null && _f !== void 0 ? _f : null,
            filter: option.filter !== undefined
                ? typeof option.filter === 'string'
                    ? {
                        regex: new RegExp(option.filter, 'u'),
                        match: true,
                    }
                    : {
                        regex: new RegExp(option.filter.regex, 'u'),
                        match: option.filter.match,
                    }
                : null,
            // calculated ordering weight based on modifiers
            modifierWeight: weight,
        };
        const selectors = Array.isArray(option.selector)
            ? option.selector
            : [option.selector];
        return selectors.map(selector => (Object.assign({ selector: shared_1.isMetaSelector(selector)
                ? enums_1.MetaSelectors[selector]
                : enums_1.Selectors[selector] }, normalizedOption)));
    }
    function parseOptions(context) {
        const normalizedOptions = context.options
            .map(opt => normalizeOption(opt))
            .reduce((acc, val) => acc.concat(val), []);
        return util.getEnumNames(enums_1.Selectors).reduce((acc, k) => {
            acc[k] = validator_1.createValidator(k, context, normalizedOptions);
            return acc;
        }, {});
    }
    exports.parseOptions = parseOptions;
},
/* @typescript-eslint/eslint-plugin/dist/rules/naming-convention-utils/validator.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const enums_1 = require(970) /* ./enums */;
    const format_1 = require(975) /* ./format */;
    const shared_1 = require(972) /* ./shared */;
    const util = __importStar(require(906) /* ../../util */);
    function createValidator(type, context, allConfigs) {
        // make sure the "highest priority" configs are checked first
        const selectorType = enums_1.Selectors[type];
        const configs = allConfigs
            // gather all of the applicable selectors
            .filter(c => (c.selector & selectorType) !== 0 ||
            c.selector === enums_1.MetaSelectors.default)
            .sort((a, b) => {
            if (a.selector === b.selector) {
                // in the event of the same selector, order by modifier weight
                // sort descending - the type modifiers are "more important"
                return b.modifierWeight - a.modifierWeight;
            }
            const aIsMeta = shared_1.isMetaSelector(a.selector);
            const bIsMeta = shared_1.isMetaSelector(b.selector);
            // non-meta selectors should go ahead of meta selectors
            if (aIsMeta && !bIsMeta) {
                return 1;
            }
            if (!aIsMeta && bIsMeta) {
                return -1;
            }
            const aIsMethodOrProperty = shared_1.isMethodOrPropertySelector(a.selector);
            const bIsMethodOrProperty = shared_1.isMethodOrPropertySelector(b.selector);
            // for backward compatibility, method and property have higher precedence than other meta selectors
            if (aIsMethodOrProperty && !bIsMethodOrProperty) {
                return -1;
            }
            if (!aIsMethodOrProperty && bIsMethodOrProperty) {
                return 1;
            }
            // both aren't meta selectors
            // sort descending - the meta selectors are "least important"
            return b.selector - a.selector;
        });
        return (node, modifiers = new Set()) => {
            var _a, _b, _c;
            const originalName = node.type === experimental_utils_1.AST_NODE_TYPES.Identifier ? node.name : `${node.value}`;
            // return will break the loop and stop checking configs
            // it is only used when the name is known to have failed or succeeded a config.
            for (const config of configs) {
                if (((_a = config.filter) === null || _a === void 0 ? void 0 : _a.regex.test(originalName)) !== ((_b = config.filter) === null || _b === void 0 ? void 0 : _b.match)) {
                    // name does not match the filter
                    continue;
                }
                if ((_c = config.modifiers) === null || _c === void 0 ? void 0 : _c.some(modifier => !modifiers.has(modifier))) {
                    // does not have the required modifiers
                    continue;
                }
                if (!isCorrectType(node, config, context, selectorType)) {
                    // is not the correct type
                    continue;
                }
                let name = originalName;
                name = validateUnderscore('leading', config, name, node, originalName);
                if (name === null) {
                    // fail
                    return;
                }
                name = validateUnderscore('trailing', config, name, node, originalName);
                if (name === null) {
                    // fail
                    return;
                }
                name = validateAffix('prefix', config, name, node, originalName);
                if (name === null) {
                    // fail
                    return;
                }
                name = validateAffix('suffix', config, name, node, originalName);
                if (name === null) {
                    // fail
                    return;
                }
                if (!validateCustom(config, name, node, originalName)) {
                    // fail
                    return;
                }
                if (!validatePredefinedFormat(config, name, node, originalName)) {
                    // fail
                    return;
                }
                // it's valid for this config, so we don't need to check any more configs
                return;
            }
        };
        // centralizes the logic for formatting the report data
        function formatReportData({ affixes, formats, originalName, processedName, position, custom, count, }) {
            var _a;
            return {
                type: shared_1.selectorTypeToMessageString(type),
                name: originalName,
                processedName,
                position,
                count,
                affixes: affixes === null || affixes === void 0 ? void 0 : affixes.join(', '),
                formats: formats === null || formats === void 0 ? void 0 : formats.map(f => enums_1.PredefinedFormats[f]).join(', '),
                regex: (_a = custom === null || custom === void 0 ? void 0 : custom.regex) === null || _a === void 0 ? void 0 : _a.toString(),
                regexMatch: (custom === null || custom === void 0 ? void 0 : custom.match) === true
                    ? 'match'
                    : (custom === null || custom === void 0 ? void 0 : custom.match) === false
                        ? 'not match'
                        : null,
            };
        }
        /**
         * @returns the name with the underscore removed, if it is valid according to the specified underscore option, null otherwise
         */
        function validateUnderscore(position, config, name, node, originalName) {
            const option = position === 'leading'
                ? config.leadingUnderscore
                : config.trailingUnderscore;
            if (!option) {
                return name;
            }
            const hasSingleUnderscore = position === 'leading'
                ? () => name.startsWith('_')
                : () => name.endsWith('_');
            const trimSingleUnderscore = position === 'leading'
                ? () => name.slice(1)
                : () => name.slice(0, -1);
            const hasDoubleUnderscore = position === 'leading'
                ? () => name.startsWith('__')
                : () => name.endsWith('__');
            const trimDoubleUnderscore = position === 'leading'
                ? () => name.slice(2)
                : () => name.slice(0, -2);
            switch (option) {
                // ALLOW - no conditions as the user doesn't care if it's there or not
                case enums_1.UnderscoreOptions.allow: {
                    if (hasSingleUnderscore()) {
                        return trimSingleUnderscore();
                    }
                    return name;
                }
                case enums_1.UnderscoreOptions.allowDouble: {
                    if (hasDoubleUnderscore()) {
                        return trimDoubleUnderscore();
                    }
                    return name;
                }
                case enums_1.UnderscoreOptions.allowSingleOrDouble: {
                    if (hasDoubleUnderscore()) {
                        return trimDoubleUnderscore();
                    }
                    if (hasSingleUnderscore()) {
                        return trimSingleUnderscore();
                    }
                    return name;
                }
                // FORBID
                case enums_1.UnderscoreOptions.forbid: {
                    if (hasSingleUnderscore()) {
                        context.report({
                            node,
                            messageId: 'unexpectedUnderscore',
                            data: formatReportData({
                                originalName,
                                position,
                                count: 'one',
                            }),
                        });
                        return null;
                    }
                    return name;
                }
                // REQUIRE
                case enums_1.UnderscoreOptions.require: {
                    if (!hasSingleUnderscore()) {
                        context.report({
                            node,
                            messageId: 'missingUnderscore',
                            data: formatReportData({
                                originalName,
                                position,
                                count: 'one',
                            }),
                        });
                        return null;
                    }
                    return trimSingleUnderscore();
                }
                case enums_1.UnderscoreOptions.requireDouble: {
                    if (!hasDoubleUnderscore()) {
                        context.report({
                            node,
                            messageId: 'missingUnderscore',
                            data: formatReportData({
                                originalName,
                                position,
                                count: 'two',
                            }),
                        });
                        return null;
                    }
                    return trimDoubleUnderscore();
                }
            }
        }
        /**
         * @returns the name with the affix removed, if it is valid according to the specified affix option, null otherwise
         */
        function validateAffix(position, config, name, node, originalName) {
            const affixes = config[position];
            if (!affixes || affixes.length === 0) {
                return name;
            }
            for (const affix of affixes) {
                const hasAffix = position === 'prefix' ? name.startsWith(affix) : name.endsWith(affix);
                const trimAffix = position === 'prefix'
                    ? () => name.slice(affix.length)
                    : () => name.slice(0, -affix.length);
                if (hasAffix) {
                    // matches, so trim it and return
                    return trimAffix();
                }
            }
            context.report({
                node,
                messageId: 'missingAffix',
                data: formatReportData({
                    originalName,
                    position,
                    affixes,
                }),
            });
            return null;
        }
        /**
         * @returns true if the name is valid according to the `regex` option, false otherwise
         */
        function validateCustom(config, name, node, originalName) {
            const custom = config.custom;
            if (!custom) {
                return true;
            }
            const result = custom.regex.test(name);
            if (custom.match && result) {
                return true;
            }
            if (!custom.match && !result) {
                return true;
            }
            context.report({
                node,
                messageId: 'satisfyCustom',
                data: formatReportData({
                    originalName,
                    custom,
                }),
            });
            return false;
        }
        /**
         * @returns true if the name is valid according to the `format` option, false otherwise
         */
        function validatePredefinedFormat(config, name, node, originalName) {
            const formats = config.format;
            if (formats === null || formats.length === 0) {
                return true;
            }
            for (const format of formats) {
                const checker = format_1.PredefinedFormatToCheckFunction[format];
                if (checker(name)) {
                    return true;
                }
            }
            context.report({
                node,
                messageId: originalName === name
                    ? 'doesNotMatchFormat'
                    : 'doesNotMatchFormatTrimmed',
                data: formatReportData({
                    originalName,
                    processedName: name,
                    formats,
                }),
            });
            return false;
        }
    }
    exports.createValidator = createValidator;
    const SelectorsAllowedToHaveTypes = enums_1.Selectors.variable |
        enums_1.Selectors.parameter |
        enums_1.Selectors.classProperty |
        enums_1.Selectors.objectLiteralProperty |
        enums_1.Selectors.typeProperty |
        enums_1.Selectors.parameterProperty |
        enums_1.Selectors.accessor;
    function isCorrectType(node, config, context, selector) {
        if (config.types === null) {
            return true;
        }
        if ((SelectorsAllowedToHaveTypes & selector) === 0) {
            return true;
        }
        const { esTreeNodeToTSNodeMap, program } = util.getParserServices(context);
        const checker = program.getTypeChecker();
        const tsNode = esTreeNodeToTSNodeMap.get(node);
        const type = checker
            .getTypeAtLocation(tsNode)
            // remove null and undefined from the type, as we don't care about it here
            .getNonNullableType();
        for (const allowedType of config.types) {
            switch (allowedType) {
                case enums_1.TypeModifiers.array:
                    if (isAllTypesMatch(type, t => checker.isArrayType(t) || checker.isTupleType(t))) {
                        return true;
                    }
                    break;
                case enums_1.TypeModifiers.function:
                    if (isAllTypesMatch(type, t => t.getCallSignatures().length > 0)) {
                        return true;
                    }
                    break;
                case enums_1.TypeModifiers.boolean:
                case enums_1.TypeModifiers.number:
                case enums_1.TypeModifiers.string: {
                    const typeString = checker.typeToString(
                    // this will resolve things like true => boolean, 'a' => string and 1 => number
                    checker.getWidenedType(checker.getBaseTypeOfLiteralType(type)));
                    const allowedTypeString = enums_1.TypeModifiers[allowedType];
                    if (typeString === allowedTypeString) {
                        return true;
                    }
                    break;
                }
            }
        }
        return false;
    }
    /**
     * @returns `true` if the type (or all union types) in the given type return true for the callback
     */
    function isAllTypesMatch(type, cb) {
        if (type.isUnion()) {
            return type.types.every(t => cb(t));
        }
        return cb(type);
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/naming-convention-utils/format.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const enums_1 = require(970) /* ./enums */;
    /*
    These format functions are taken from `tslint-consistent-codestyle/naming-convention`:
    https://github.com/ajafff/tslint-consistent-codestyle/blob/ab156cc8881bcc401236d999f4ce034b59039e81/rules/namingConventionRule.ts#L603-L645
    
    The licence for the code can be viewed here:
    https://github.com/ajafff/tslint-consistent-codestyle/blob/ab156cc8881bcc401236d999f4ce034b59039e81/LICENSE
    */
    /*
    Why not regex here? Because it's actually really, really difficult to create a regex to handle
    all of the unicode cases, and we have many non-english users that use non-english characters.
    https://gist.github.com/mathiasbynens/6334847
    */
    function isPascalCase(name) {
        return (name.length === 0 ||
            (name[0] === name[0].toUpperCase() && !name.includes('_')));
    }
    function isStrictPascalCase(name) {
        return (name.length === 0 ||
            (name[0] === name[0].toUpperCase() && hasStrictCamelHumps(name, true)));
    }
    function isCamelCase(name) {
        return (name.length === 0 ||
            (name[0] === name[0].toLowerCase() && !name.includes('_')));
    }
    function isStrictCamelCase(name) {
        return (name.length === 0 ||
            (name[0] === name[0].toLowerCase() && hasStrictCamelHumps(name, false)));
    }
    function hasStrictCamelHumps(name, isUpper) {
        function isUppercaseChar(char) {
            return char === char.toUpperCase() && char !== char.toLowerCase();
        }
        if (name.startsWith('_')) {
            return false;
        }
        for (let i = 1; i < name.length; ++i) {
            if (name[i] === '_') {
                return false;
            }
            if (isUpper === isUppercaseChar(name[i])) {
                if (isUpper) {
                    return false;
                }
            }
            else {
                isUpper = !isUpper;
            }
        }
        return true;
    }
    function isSnakeCase(name) {
        return (name.length === 0 ||
            (name === name.toLowerCase() && validateUnderscores(name)));
    }
    function isUpperCase(name) {
        return (name.length === 0 ||
            (name === name.toUpperCase() && validateUnderscores(name)));
    }
    /** Check for leading trailing and adjacent underscores */
    function validateUnderscores(name) {
        if (name.startsWith('_')) {
            return false;
        }
        let wasUnderscore = false;
        for (let i = 1; i < name.length; ++i) {
            if (name[i] === '_') {
                if (wasUnderscore) {
                    return false;
                }
                wasUnderscore = true;
            }
            else {
                wasUnderscore = false;
            }
        }
        return !wasUnderscore;
    }
    const PredefinedFormatToCheckFunction = {
        [enums_1.PredefinedFormats.PascalCase]: isPascalCase,
        [enums_1.PredefinedFormats.StrictPascalCase]: isStrictPascalCase,
        [enums_1.PredefinedFormats.camelCase]: isCamelCase,
        [enums_1.PredefinedFormats.strictCamelCase]: isStrictCamelCase,
        [enums_1.PredefinedFormats.UPPER_CASE]: isUpperCase,
        [enums_1.PredefinedFormats.snake_case]: isSnakeCase,
    };
    exports.PredefinedFormatToCheckFunction = PredefinedFormatToCheckFunction;
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-array-constructor.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-array-constructor',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow generic `Array` constructors',
                category: 'Stylistic Issues',
                recommended: 'error',
                extendsBaseRule: true,
            },
            fixable: 'code',
            messages: {
                useLiteral: 'The array literal notation [] is preferable.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            /**
             * Disallow construction of dense arrays using the Array constructor
             * @param node node to evaluate
             */
            function check(node) {
                if (node.arguments.length !== 1 &&
                    node.callee.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    node.callee.name === 'Array' &&
                    !node.typeParameters &&
                    !util.isOptionalCallExpression(node)) {
                    context.report({
                        node,
                        messageId: 'useLiteral',
                        fix(fixer) {
                            if (node.arguments.length === 0) {
                                return fixer.replaceText(node, '[]');
                            }
                            const fullText = context.getSourceCode().getText(node);
                            const preambleLength = node.callee.range[1] - node.range[0];
                            return fixer.replaceText(node, `[${fullText.slice(preambleLength + 1, -1)}]`);
                        },
                    });
                }
            }
            return {
                CallExpression: check,
                NewExpression: check,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-base-to-string.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    var Usefulness;
    (function (Usefulness) {
        Usefulness[Usefulness["Always"] = 0] = "Always";
        Usefulness["Never"] = "will";
        Usefulness["Sometimes"] = "may";
    })(Usefulness || (Usefulness = {}));
    exports.default = util.createRule({
        name: 'no-base-to-string',
        meta: {
            docs: {
                description: 'Requires that `.toString()` is only called on objects which provide useful information when stringified',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            messages: {
                baseToString: "'{{name}} {{certainty}} evaluate to '[object Object]' when stringified.",
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        ignoredTypeNames: {
                            type: 'array',
                            items: {
                                type: 'string',
                            },
                        },
                    },
                    additionalProperties: false,
                },
            ],
            type: 'suggestion',
        },
        defaultOptions: [
            {
                ignoredTypeNames: ['RegExp'],
            },
        ],
        create(context, [option]) {
            var _a;
            const parserServices = util.getParserServices(context);
            const typeChecker = parserServices.program.getTypeChecker();
            const ignoredTypeNames = (_a = option.ignoredTypeNames) !== null && _a !== void 0 ? _a : [];
            function checkExpression(node, type) {
                if (node.type === experimental_utils_1.AST_NODE_TYPES.Literal) {
                    return;
                }
                const certainty = collectToStringCertainty(type !== null && type !== void 0 ? type : typeChecker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(node)));
                if (certainty === Usefulness.Always) {
                    return;
                }
                context.report({
                    data: {
                        certainty,
                        name: context.getSourceCode().getText(node),
                    },
                    messageId: 'baseToString',
                    node,
                });
            }
            function collectToStringCertainty(type) {
                const toString = typeChecker.getPropertyOfType(type, 'toString');
                const declarations = toString === null || toString === void 0 ? void 0 : toString.getDeclarations();
                if (!toString || !declarations || declarations.length === 0) {
                    return Usefulness.Always;
                }
                // Patch for old version TypeScript, the Boolean type definition missing toString()
                if (type.flags & ts.TypeFlags.Boolean ||
                    type.flags & ts.TypeFlags.BooleanLiteral) {
                    return Usefulness.Always;
                }
                if (ignoredTypeNames.includes(util.getTypeName(typeChecker, type))) {
                    return Usefulness.Always;
                }
                if (declarations.every(({ parent }) => !ts.isInterfaceDeclaration(parent) || parent.name.text !== 'Object')) {
                    return Usefulness.Always;
                }
                if (type.isIntersection()) {
                    for (const subType of type.types) {
                        const subtypeUsefulness = collectToStringCertainty(subType);
                        if (subtypeUsefulness === Usefulness.Always) {
                            return Usefulness.Always;
                        }
                    }
                    return Usefulness.Never;
                }
                if (!type.isUnion()) {
                    return Usefulness.Never;
                }
                let allSubtypesUseful = true;
                let someSubtypeUseful = false;
                for (const subType of type.types) {
                    const subtypeUsefulness = collectToStringCertainty(subType);
                    if (subtypeUsefulness !== Usefulness.Always && allSubtypesUseful) {
                        allSubtypesUseful = false;
                    }
                    if (subtypeUsefulness !== Usefulness.Never && !someSubtypeUseful) {
                        someSubtypeUseful = true;
                    }
                }
                if (allSubtypesUseful && someSubtypeUseful) {
                    return Usefulness.Always;
                }
                if (someSubtypeUseful) {
                    return Usefulness.Sometimes;
                }
                return Usefulness.Never;
            }
            return {
                'AssignmentExpression[operator = "+="], BinaryExpression[operator = "+"]'(node) {
                    const leftType = typeChecker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(node.left));
                    const rightType = typeChecker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(node.right));
                    if (util.getTypeName(typeChecker, leftType) === 'string') {
                        checkExpression(node.right, rightType);
                    }
                    else if (util.getTypeName(typeChecker, rightType) === 'string') {
                        checkExpression(node.left, leftType);
                    }
                },
                'CallExpression > MemberExpression.callee > Identifier[name = "toString"].property'(node) {
                    const memberExpr = node.parent;
                    checkExpression(memberExpr.object);
                },
                TemplateLiteral(node) {
                    if (node.parent &&
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.TaggedTemplateExpression) {
                        return;
                    }
                    for (const expression of node.expressions) {
                        checkExpression(expression);
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-confusing-non-null-assertion.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-confusing-non-null-assertion',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow non-null assertion in locations that may be confusing',
                category: 'Stylistic Issues',
                recommended: false,
            },
            fixable: 'code',
            messages: {
                confusingEqual: 'Confusing combinations of non-null assertion and equal test like "a! == b", which looks very similar to not equal "a !== b"',
                confusingAssign: 'Confusing combinations of non-null assertion and equal test like "a! = b", which looks very similar to not equal "a != b"',
                notNeedInEqualTest: 'Unnecessary non-null assertion (!) in equal test',
                notNeedInAssign: 'Unnecessary non-null assertion (!) in assignment left hand',
                wrapUpLeft: 'Wrap up left hand to avoid putting non-null assertion "!" and "=" together',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                'BinaryExpression, AssignmentExpression'(node) {
                    function isLeftHandPrimaryExpression(node) {
                        return node.type === experimental_utils_1.AST_NODE_TYPES.TSNonNullExpression;
                    }
                    if (node.operator === '==' ||
                        node.operator === '===' ||
                        node.operator === '=') {
                        const isAssign = node.operator === '=';
                        const leftHandFinalToken = sourceCode.getLastToken(node.left);
                        const tokenAfterLeft = sourceCode.getTokenAfter(node.left);
                        if ((leftHandFinalToken === null || leftHandFinalToken === void 0 ? void 0 : leftHandFinalToken.type) === experimental_utils_1.AST_TOKEN_TYPES.Punctuator &&
                            (leftHandFinalToken === null || leftHandFinalToken === void 0 ? void 0 : leftHandFinalToken.value) === '!' &&
                            (tokenAfterLeft === null || tokenAfterLeft === void 0 ? void 0 : tokenAfterLeft.value) !== ')') {
                            if (isLeftHandPrimaryExpression(node.left)) {
                                context.report({
                                    node,
                                    messageId: isAssign ? 'confusingAssign' : 'confusingEqual',
                                    suggest: [
                                        {
                                            messageId: isAssign
                                                ? 'notNeedInAssign'
                                                : 'notNeedInEqualTest',
                                            fix: (fixer) => [
                                                fixer.remove(leftHandFinalToken),
                                            ],
                                        },
                                    ],
                                });
                            }
                            else {
                                context.report({
                                    node,
                                    messageId: isAssign ? 'confusingAssign' : 'confusingEqual',
                                    suggest: [
                                        {
                                            messageId: 'wrapUpLeft',
                                            fix: (fixer) => [
                                                fixer.insertTextBefore(node.left, '('),
                                                fixer.insertTextAfter(node.left, ')'),
                                            ],
                                        },
                                    ],
                                });
                            }
                        }
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-confusing-void-expression.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-confusing-void-expression',
        meta: {
            docs: {
                description: 'Requires expressions of type void to appear in statement position',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            messages: {
                invalidVoidExpr: 'Placing a void expression inside another expression is forbidden. ' +
                    'Move it to its own statement instead.',
                invalidVoidExprWrapVoid: 'Void expressions used inside another expression ' +
                    'must be moved to its own statement ' +
                    'or marked explicitly with the `void` operator.',
                invalidVoidExprArrow: 'Returning a void expression from an arrow function shorthand is forbidden. ' +
                    'Please add braces to the arrow function.',
                invalidVoidExprArrowWrapVoid: 'Void expressions returned from an arrow function shorthand ' +
                    'must be marked explicitly with the `void` operator.',
                invalidVoidExprReturn: 'Returning a void expression from a function is forbidden. ' +
                    'Please move it before the `return` statement.',
                invalidVoidExprReturnLast: 'Returning a void expression from a function is forbidden. ' +
                    'Please remove the `return` statement.',
                invalidVoidExprReturnWrapVoid: 'Void expressions returned from a function ' +
                    'must be marked explicitly with the `void` operator.',
                voidExprWrapVoid: 'Mark with an explicit `void` operator',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        ignoreArrowShorthand: { type: 'boolean' },
                        ignoreVoidOperator: { type: 'boolean' },
                    },
                    additionalProperties: false,
                },
            ],
            type: 'problem',
            fixable: 'code',
        },
        defaultOptions: [{}],
        create(context, [options]) {
            return {
                'AwaitExpression, CallExpression, TaggedTemplateExpression'(node) {
                    const parserServices = util.getParserServices(context);
                    const checker = parserServices.program.getTypeChecker();
                    const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                    const type = util.getConstrainedTypeAtLocation(checker, tsNode);
                    if (!tsutils.isTypeFlagSet(type, ts.TypeFlags.VoidLike)) {
                        // not a void expression
                        return;
                    }
                    const invalidAncestor = findInvalidAncestor(node);
                    if (invalidAncestor == null) {
                        // void expression is in valid position
                        return;
                    }
                    const sourceCode = context.getSourceCode();
                    const wrapVoidFix = (fixer) => {
                        const nodeText = sourceCode.getText(node);
                        const newNodeText = `void ${nodeText}`;
                        return fixer.replaceText(node, newNodeText);
                    };
                    if (invalidAncestor.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression) {
                        // handle arrow function shorthand
                        if (options.ignoreVoidOperator) {
                            // handle wrapping with `void`
                            return context.report({
                                node,
                                messageId: 'invalidVoidExprArrowWrapVoid',
                                fix: wrapVoidFix,
                            });
                        }
                        // handle wrapping with braces
                        const arrowFunction = invalidAncestor;
                        return context.report({
                            node,
                            messageId: 'invalidVoidExprArrow',
                            fix(fixer) {
                                const arrowBody = arrowFunction.body;
                                const arrowBodyText = sourceCode.getText(arrowBody);
                                const newArrowBodyText = `{ ${arrowBodyText}; }`;
                                if (util.isParenthesized(arrowBody, sourceCode)) {
                                    const bodyOpeningParen = sourceCode.getTokenBefore(arrowBody, util.isOpeningParenToken);
                                    const bodyClosingParen = sourceCode.getTokenAfter(arrowBody, util.isClosingParenToken);
                                    return fixer.replaceTextRange([bodyOpeningParen.range[0], bodyClosingParen.range[1]], newArrowBodyText);
                                }
                                return fixer.replaceText(arrowBody, newArrowBodyText);
                            },
                        });
                    }
                    if (invalidAncestor.type === experimental_utils_1.AST_NODE_TYPES.ReturnStatement) {
                        // handle return statement
                        if (options.ignoreVoidOperator) {
                            // handle wrapping with `void`
                            return context.report({
                                node,
                                messageId: 'invalidVoidExprReturnWrapVoid',
                                fix: wrapVoidFix,
                            });
                        }
                        const returnStmt = invalidAncestor;
                        if (isFinalReturn(returnStmt)) {
                            // remove the `return` keyword
                            return context.report({
                                node,
                                messageId: 'invalidVoidExprReturnLast',
                                fix(fixer) {
                                    const returnValue = returnStmt.argument;
                                    const returnValueText = sourceCode.getText(returnValue);
                                    let newReturnStmtText = `${returnValueText};`;
                                    if (isPreventingASI(returnValue, sourceCode)) {
                                        // put a semicolon at the beginning of the line
                                        newReturnStmtText = `;${newReturnStmtText}`;
                                    }
                                    return fixer.replaceText(returnStmt, newReturnStmtText);
                                },
                            });
                        }
                        // move before the `return` keyword
                        return context.report({
                            node,
                            messageId: 'invalidVoidExprReturn',
                            fix(fixer) {
                                var _a;
                                const returnValue = returnStmt.argument;
                                const returnValueText = sourceCode.getText(returnValue);
                                let newReturnStmtText = `${returnValueText}; return;`;
                                if (isPreventingASI(returnValue, sourceCode)) {
                                    // put a semicolon at the beginning of the line
                                    newReturnStmtText = `;${newReturnStmtText}`;
                                }
                                if (((_a = returnStmt.parent) === null || _a === void 0 ? void 0 : _a.type) !== experimental_utils_1.AST_NODE_TYPES.BlockStatement) {
                                    // e.g. `if (cond) return console.error();`
                                    // add braces if not inside a block
                                    newReturnStmtText = `{ ${newReturnStmtText} }`;
                                }
                                return fixer.replaceText(returnStmt, newReturnStmtText);
                            },
                        });
                    }
                    // handle generic case
                    if (options.ignoreVoidOperator) {
                        // this would be reported by this rule btw. such irony
                        return context.report({
                            node,
                            messageId: 'invalidVoidExprWrapVoid',
                            suggest: [{ messageId: 'voidExprWrapVoid', fix: wrapVoidFix }],
                        });
                    }
                    context.report({
                        node,
                        messageId: 'invalidVoidExpr',
                    });
                },
            };
            /**
             * Inspects the void expression's ancestors and finds closest invalid one.
             * By default anything other than an ExpressionStatement is invalid.
             * Parent expressions which can be used for their short-circuiting behavior
             * are ignored and their parents are checked instead.
             * @param node The void expression node to check.
             * @returns Invalid ancestor node if it was found. `null` otherwise.
             */
            function findInvalidAncestor(node) {
                const parent = util.nullThrows(node.parent, util.NullThrowsReasons.MissingParent);
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.ExpressionStatement) {
                    // e.g. `{ console.log("foo"); }`
                    // this is always valid
                    return null;
                }
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression) {
                    if (parent.right === node) {
                        // e.g. `x && console.log(x)`
                        // this is valid only if the next ancestor is valid
                        return findInvalidAncestor(parent);
                    }
                }
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.ConditionalExpression) {
                    if (parent.consequent === node || parent.alternate === node) {
                        // e.g. `cond ? console.log(true) : console.log(false)`
                        // this is valid only if the next ancestor is valid
                        return findInvalidAncestor(parent);
                    }
                }
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression) {
                    // e.g. `() => console.log("foo")`
                    // this is valid with an appropriate option
                    if (options.ignoreArrowShorthand) {
                        return null;
                    }
                }
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression) {
                    if (parent.operator === 'void') {
                        // e.g. `void console.log("foo")`
                        // this is valid with an appropriate option
                        if (options.ignoreVoidOperator) {
                            return null;
                        }
                    }
                }
                // any other parent is invalid
                return parent;
            }
            /** Checks whether the return statement is the last statement in a function body. */
            function isFinalReturn(node) {
                // the parent must be a block
                const block = util.nullThrows(node.parent, util.NullThrowsReasons.MissingParent);
                if (block.type !== experimental_utils_1.AST_NODE_TYPES.BlockStatement) {
                    // e.g. `if (cond) return;` (not in a block)
                    return false;
                }
                // the block's parent must be a function
                const blockParent = util.nullThrows(block.parent, util.NullThrowsReasons.MissingParent);
                if (![
                    experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration,
                    experimental_utils_1.AST_NODE_TYPES.FunctionExpression,
                    experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression,
                ].includes(blockParent.type)) {
                    // e.g. `if (cond) { return; }`
                    // not in a top-level function block
                    return false;
                }
                // must be the last child of the block
                if (block.body.indexOf(node) < block.body.length - 1) {
                    // not the last statement in the block
                    return false;
                }
                return true;
            }
            /**
             * Checks whether the given node, if placed on its own line,
             * would prevent automatic semicolon insertion on the line before.
             *
             * This happens if the line begins with `(`, `[` or `` ` ``
             */
            function isPreventingASI(node, sourceCode) {
                const startToken = util.nullThrows(sourceCode.getFirstToken(node), util.NullThrowsReasons.MissingToken('first token', node.type));
                return ['(', '[', '`'].includes(startToken.value);
            }
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-dupe-class-members.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const no_dupe_class_members_1 = __importDefault(require(502) /* eslint/lib/rules/no-dupe-class-members */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-dupe-class-members',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow duplicate class members',
                category: 'Possible Errors',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: no_dupe_class_members_1.default.meta.schema,
            messages: no_dupe_class_members_1.default.meta.messages,
        },
        defaultOptions: [],
        create(context) {
            const rules = no_dupe_class_members_1.default.create(context);
            return Object.assign(Object.assign({}, rules), { MethodDefinition(node) {
                    if (node.computed) {
                        return;
                    }
                    if (node.value.type === experimental_utils_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression) {
                        return;
                    }
                    return rules.MethodDefinition(node);
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-duplicate-imports.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const no_duplicate_imports_1 = __importDefault(require(506) /* eslint/lib/rules/no-duplicate-imports */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-duplicate-imports',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow duplicate imports',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: no_duplicate_imports_1.default.meta.schema,
            messages: Object.assign(Object.assign({}, no_duplicate_imports_1.default.meta.messages), { importType: '{{module}} type import is duplicated', importTypeAs: '{{module}} type import is duplicated as type export', exportType: '{{module}} type export is duplicated', exportTypeAs: '{{module}} type export is duplicated as type import' }),
        },
        defaultOptions: [
            {
                includeExports: false,
            },
        ],
        create(context, [option]) {
            const rules = no_duplicate_imports_1.default.create(context);
            const includeExports = option.includeExports;
            const typeMemberImports = new Set();
            const typeDefaultImports = new Set();
            const typeExports = new Set();
            function report(messageId, node, module) {
                context.report({
                    messageId,
                    node,
                    data: {
                        module,
                    },
                });
            }
            function isStringLiteral(node) {
                return (!!node &&
                    node.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                    typeof node.value === 'string');
            }
            function isAllMemberImport(node) {
                return node.specifiers.every(specifier => specifier.type === experimental_utils_1.AST_NODE_TYPES.ImportSpecifier);
            }
            function checkTypeImport(node) {
                if (isStringLiteral(node.source)) {
                    const value = node.source.value;
                    const isMemberImport = isAllMemberImport(node);
                    if (isMemberImport
                        ? typeMemberImports.has(value)
                        : typeDefaultImports.has(value)) {
                        report('importType', node, value);
                    }
                    if (includeExports && typeExports.has(value)) {
                        report('importTypeAs', node, value);
                    }
                    if (isMemberImport) {
                        typeMemberImports.add(value);
                    }
                    else {
                        typeDefaultImports.add(value);
                    }
                }
            }
            function checkTypeExport(node) {
                if (isStringLiteral(node.source)) {
                    const value = node.source.value;
                    if (typeExports.has(value)) {
                        report('exportType', node, value);
                    }
                    if (typeMemberImports.has(value) || typeDefaultImports.has(value)) {
                        report('exportTypeAs', node, value);
                    }
                    typeExports.add(value);
                }
            }
            return Object.assign(Object.assign({}, rules), { ImportDeclaration(node) {
                    if (node.importKind === 'type') {
                        checkTypeImport(node);
                        return;
                    }
                    rules.ImportDeclaration(node);
                },
                ExportNamedDeclaration(node) {
                    var _a;
                    if (includeExports && node.exportKind === 'type') {
                        checkTypeExport(node);
                        return;
                    }
                    (_a = rules.ExportNamedDeclaration) === null || _a === void 0 ? void 0 : _a.call(rules, node);
                },
                ExportAllDeclaration(node) {
                    var _a;
                    if (includeExports && node.exportKind === 'type') {
                        checkTypeExport(node);
                        return;
                    }
                    (_a = rules.ExportAllDeclaration) === null || _a === void 0 ? void 0 : _a.call(rules, node);
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-dynamic-delete.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-dynamic-delete',
        meta: {
            docs: {
                category: 'Best Practices',
                description: 'Disallow the delete operator with computed key expressions',
                recommended: false,
            },
            fixable: 'code',
            messages: {
                dynamicDelete: 'Do not delete dynamically computed property keys.',
            },
            schema: [],
            type: 'suggestion',
        },
        defaultOptions: [],
        create(context) {
            function createFixer(member) {
                if (member.property.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                    typeof member.property.value === 'string') {
                    return createPropertyReplacement(member.property, `.${member.property.value}`);
                }
                return undefined;
            }
            return {
                'UnaryExpression[operator=delete]'(node) {
                    if (node.argument.type !== experimental_utils_1.AST_NODE_TYPES.MemberExpression ||
                        !node.argument.computed ||
                        isNecessaryDynamicAccess(diveIntoWrapperExpressions(node.argument.property))) {
                        return;
                    }
                    context.report({
                        fix: createFixer(node.argument),
                        messageId: 'dynamicDelete',
                        node: node.argument.property,
                    });
                },
            };
            function createPropertyReplacement(property, replacement) {
                return (fixer) => fixer.replaceTextRange(getTokenRange(property), replacement);
            }
            function getTokenRange(property) {
                const sourceCode = context.getSourceCode();
                return [
                    sourceCode.getTokenBefore(property).range[0],
                    sourceCode.getTokenAfter(property).range[1],
                ];
            }
        },
    });
    function diveIntoWrapperExpressions(node) {
        if (node.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression) {
            return diveIntoWrapperExpressions(node.argument);
        }
        return node;
    }
    function isNecessaryDynamicAccess(property) {
        if (property.type !== experimental_utils_1.AST_NODE_TYPES.Literal) {
            return false;
        }
        if (typeof property.value === 'number') {
            return true;
        }
        return (typeof property.value === 'string' &&
            !tsutils.isValidPropertyAccess(property.value));
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-empty-function.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const no_empty_function_1 = __importDefault(require(511) /* eslint/lib/rules/no-empty-function */);
    const util = __importStar(require(906) /* ../util */);
    const schema = util.deepMerge(Array.isArray(no_empty_function_1.default.meta.schema)
        ? no_empty_function_1.default.meta.schema[0]
        : no_empty_function_1.default.meta.schema, {
        properties: {
            allow: {
                items: {
                    enum: [
                        'functions',
                        'arrowFunctions',
                        'generatorFunctions',
                        'methods',
                        'generatorMethods',
                        'getters',
                        'setters',
                        'constructors',
                        'private-constructors',
                        'protected-constructors',
                        'asyncFunctions',
                        'asyncMethods',
                        'decoratedFunctions',
                    ],
                },
            },
        },
    });
    exports.default = util.createRule({
        name: 'no-empty-function',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow empty functions',
                category: 'Best Practices',
                recommended: 'error',
                extendsBaseRule: true,
            },
            schema: [schema],
            messages: no_empty_function_1.default.meta.messages,
        },
        defaultOptions: [
            {
                allow: [],
            },
        ],
        create(context, [{ allow = [] }]) {
            const rules = no_empty_function_1.default.create(context);
            const isAllowedProtectedConstructors = allow.includes('protected-constructors');
            const isAllowedPrivateConstructors = allow.includes('private-constructors');
            const isAllowedDecoratedFunctions = allow.includes('decoratedFunctions');
            /**
             * Check if the method body is empty
             * @param node the node to be validated
             * @returns true if the body is empty
             * @private
             */
            function isBodyEmpty(node) {
                return !node.body || node.body.body.length === 0;
            }
            /**
             * Check if method has parameter properties
             * @param node the node to be validated
             * @returns true if the body has parameter properties
             * @private
             */
            function hasParameterProperties(node) {
                var _a;
                return (_a = node.params) === null || _a === void 0 ? void 0 : _a.some(param => param.type === experimental_utils_1.AST_NODE_TYPES.TSParameterProperty);
            }
            /**
             * @param node the node to be validated
             * @returns true if the constructor is allowed to be empty
             * @private
             */
            function isAllowedEmptyConstructor(node) {
                const parent = node.parent;
                if (isBodyEmpty(node) &&
                    (parent === null || parent === void 0 ? void 0 : parent.type) === experimental_utils_1.AST_NODE_TYPES.MethodDefinition &&
                    parent.kind === 'constructor') {
                    const { accessibility } = parent;
                    return (
                    // allow protected constructors
                    (accessibility === 'protected' && isAllowedProtectedConstructors) ||
                        // allow private constructors
                        (accessibility === 'private' && isAllowedPrivateConstructors) ||
                        // allow constructors which have parameter properties
                        hasParameterProperties(node));
                }
                return false;
            }
            /**
             * @param node the node to be validated
             * @returns true if a function has decorators
             * @private
             */
            function isAllowedEmptyDecoratedFunctions(node) {
                var _a;
                if (isAllowedDecoratedFunctions && isBodyEmpty(node)) {
                    const decorators = ((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.MethodDefinition
                        ? node.parent.decorators
                        : undefined;
                    return !!decorators && !!decorators.length;
                }
                return false;
            }
            return Object.assign(Object.assign({}, rules), { FunctionExpression(node) {
                    if (isAllowedEmptyConstructor(node) ||
                        isAllowedEmptyDecoratedFunctions(node)) {
                        return;
                    }
                    rules.FunctionExpression(node);
                },
                FunctionDeclaration(node) {
                    if (isAllowedEmptyDecoratedFunctions(node)) {
                        return;
                    }
                    rules.FunctionDeclaration(node);
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-empty-interface.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-empty-interface',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow the declaration of empty interfaces',
                category: 'Best Practices',
                recommended: 'error',
                suggestion: true,
            },
            fixable: 'code',
            messages: {
                noEmpty: 'An empty interface is equivalent to `{}`.',
                noEmptyWithSuper: 'An interface declaring no members is equivalent to its supertype.',
            },
            schema: [
                {
                    type: 'object',
                    additionalProperties: false,
                    properties: {
                        allowSingleExtends: {
                            type: 'boolean',
                        },
                    },
                },
            ],
        },
        defaultOptions: [
            {
                allowSingleExtends: false,
            },
        ],
        create(context, [{ allowSingleExtends }]) {
            return {
                TSInterfaceDeclaration(node) {
                    const sourceCode = context.getSourceCode();
                    const filename = context.getFilename();
                    if (node.body.body.length !== 0) {
                        // interface contains members --> Nothing to report
                        return;
                    }
                    const extend = node.extends;
                    if (!extend || extend.length === 0) {
                        context.report({
                            node: node.id,
                            messageId: 'noEmpty',
                        });
                    }
                    else if (extend.length === 1) {
                        // interface extends exactly 1 interface --> Report depending on rule setting
                        if (!allowSingleExtends) {
                            const fix = (fixer) => {
                                let typeParam = '';
                                if (node.typeParameters) {
                                    typeParam = sourceCode.getText(node.typeParameters);
                                }
                                return fixer.replaceText(node, `type ${sourceCode.getText(node.id)}${typeParam} = ${sourceCode.getText(extend[0])}`);
                            };
                            // Check if interface is within ambient declaration
                            let useAutoFix = true;
                            if (util.isDefinitionFile(filename)) {
                                const scope = context.getScope();
                                if (scope.type === 'tsModule' && scope.block.declare) {
                                    useAutoFix = false;
                                }
                            }
                            context.report(Object.assign({ node: node.id, messageId: 'noEmptyWithSuper' }, (useAutoFix
                                ? { fix }
                                : {
                                    suggest: [
                                        {
                                            messageId: 'noEmptyWithSuper',
                                            fix,
                                        },
                                    ],
                                })));
                        }
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-explicit-any.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-explicit-any',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow usage of the `any` type',
                category: 'Best Practices',
                recommended: 'warn',
                suggestion: true,
            },
            fixable: 'code',
            messages: {
                unexpectedAny: 'Unexpected any. Specify a different type.',
                suggestUnknown: 'Use `unknown` instead, this will force you to explicitly, and safely assert the type is correct.',
                suggestNever: "Use `never` instead, this is useful when instantiating generic type parameters that you don't need to know the type of.",
            },
            schema: [
                {
                    type: 'object',
                    additionalProperties: false,
                    properties: {
                        fixToUnknown: {
                            type: 'boolean',
                        },
                        ignoreRestArgs: {
                            type: 'boolean',
                        },
                    },
                },
            ],
        },
        defaultOptions: [
            {
                fixToUnknown: false,
                ignoreRestArgs: false,
            },
        ],
        create(context, [{ ignoreRestArgs, fixToUnknown }]) {
            /**
             * Checks if the node is an arrow function, function/constructor declaration or function expression
             * @param node the node to be validated.
             * @returns true if the node is any kind of function declaration or expression
             * @private
             */
            function isNodeValidFunction(node) {
                return [
                    experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression,
                    experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration,
                    experimental_utils_1.AST_NODE_TYPES.FunctionExpression,
                    experimental_utils_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression,
                    experimental_utils_1.AST_NODE_TYPES.TSFunctionType,
                    experimental_utils_1.AST_NODE_TYPES.TSConstructorType,
                    experimental_utils_1.AST_NODE_TYPES.TSCallSignatureDeclaration,
                    experimental_utils_1.AST_NODE_TYPES.TSConstructSignatureDeclaration,
                    experimental_utils_1.AST_NODE_TYPES.TSMethodSignature,
                    experimental_utils_1.AST_NODE_TYPES.TSDeclareFunction, // declare function _8(...args: any[]): unknown;
                ].includes(node.type);
            }
            /**
             * Checks if the node is a rest element child node of a function
             * @param node the node to be validated.
             * @returns true if the node is a rest element child node of a function
             * @private
             */
            function isNodeRestElementInFunction(node) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.RestElement &&
                    typeof node.parent !== 'undefined' &&
                    isNodeValidFunction(node.parent));
            }
            /**
             * Checks if the node is a TSTypeOperator node with a readonly operator
             * @param node the node to be validated.
             * @returns true if the node is a TSTypeOperator node with a readonly operator
             * @private
             */
            function isNodeReadonlyTSTypeOperator(node) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.TSTypeOperator &&
                    node.operator === 'readonly');
            }
            /**
             * Checks if the node is a TSTypeReference node with an Array identifier
             * @param node the node to be validated.
             * @returns true if the node is a TSTypeReference node with an Array identifier
             * @private
             */
            function isNodeValidArrayTSTypeReference(node) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference &&
                    typeof node.typeName !== 'undefined' &&
                    node.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    ['Array', 'ReadonlyArray'].includes(node.typeName.name));
            }
            /**
             * Checks if the node is a valid TSTypeOperator or TSTypeReference node
             * @param node the node to be validated.
             * @returns true if the node is a valid TSTypeOperator or TSTypeReference node
             * @private
             */
            function isNodeValidTSType(node) {
                return (isNodeReadonlyTSTypeOperator(node) ||
                    isNodeValidArrayTSTypeReference(node));
            }
            /**
             * Checks if the great grand-parent node is a RestElement node in a function
             * @param node the node to be validated.
             * @returns true if the great grand-parent node is a RestElement node in a function
             * @private
             */
            function isGreatGrandparentRestElement(node) {
                var _a, _b;
                return (((_b = (_a = node === null || node === void 0 ? void 0 : node.parent) === null || _a === void 0 ? void 0 : _a.parent) === null || _b === void 0 ? void 0 : _b.parent) != null &&
                    isNodeRestElementInFunction(node.parent.parent.parent));
            }
            /**
             * Checks if the great great grand-parent node is a valid RestElement node in a function
             * @param node the node to be validated.
             * @returns true if the great great grand-parent node is a valid RestElement node in a function
             * @private
             */
            function isGreatGreatGrandparentRestElement(node) {
                var _a, _b, _c;
                return (((_c = (_b = (_a = node.parent) === null || _a === void 0 ? void 0 : _a.parent) === null || _b === void 0 ? void 0 : _b.parent) === null || _c === void 0 ? void 0 : _c.parent) != null &&
                    isNodeValidTSType(node.parent.parent) &&
                    isNodeRestElementInFunction(node.parent.parent.parent.parent));
            }
            /**
             * Checks if the great grand-parent or the great great grand-parent node is a RestElement node
             * @param node the node to be validated.
             * @returns true if the great grand-parent or the great great grand-parent node is a RestElement node
             * @private
             */
            function isNodeDescendantOfRestElementInFunction(node) {
                return (isGreatGrandparentRestElement(node) ||
                    isGreatGreatGrandparentRestElement(node));
            }
            return {
                TSAnyKeyword(node) {
                    if (ignoreRestArgs && isNodeDescendantOfRestElementInFunction(node)) {
                        return;
                    }
                    const fixOrSuggest = {
                        fix: null,
                        suggest: [
                            {
                                messageId: 'suggestUnknown',
                                fix(fixer) {
                                    return fixer.replaceText(node, 'unknown');
                                },
                            },
                            {
                                messageId: 'suggestNever',
                                fix(fixer) {
                                    return fixer.replaceText(node, 'never');
                                },
                            },
                        ],
                    };
                    if (fixToUnknown) {
                        fixOrSuggest.fix = (fixer => fixer.replaceText(node, 'unknown'));
                    }
                    context.report(Object.assign({ node, messageId: 'unexpectedAny' }, fixOrSuggest));
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-extra-non-null-assertion.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-extra-non-null-assertion',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow extra non-null assertion',
                category: 'Stylistic Issues',
                recommended: 'error',
            },
            fixable: 'code',
            schema: [],
            messages: {
                noExtraNonNullAssertion: 'Forbidden extra non-null assertion.',
            },
        },
        defaultOptions: [],
        create(context) {
            function checkExtraNonNullAssertion(node) {
                context.report({
                    node,
                    messageId: 'noExtraNonNullAssertion',
                    fix(fixer) {
                        return fixer.removeRange([node.range[1] - 1, node.range[1]]);
                    },
                });
            }
            return {
                'TSNonNullExpression > TSNonNullExpression': checkExtraNonNullAssertion,
                'MemberExpression[optional = true] > TSNonNullExpression.object': checkExtraNonNullAssertion,
                'CallExpression[optional = true] > TSNonNullExpression.callee': checkExtraNonNullAssertion,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-extra-parens.js */ function _(require, module, exports, __esModule, __esExport) {
    // any is required to work around manipulating the AST in weird ways
    /* eslint-disable @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-assignment */
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const no_extra_parens_1 = __importDefault(require(520) /* eslint/lib/rules/no-extra-parens */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-extra-parens',
        meta: {
            type: 'layout',
            docs: {
                description: 'Disallow unnecessary parentheses',
                category: 'Possible Errors',
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: 'code',
            schema: no_extra_parens_1.default.meta.schema,
            messages: no_extra_parens_1.default.meta.messages,
        },
        defaultOptions: ['all'],
        create(context) {
            const rules = no_extra_parens_1.default.create(context);
            function binaryExp(node) {
                const rule = rules.BinaryExpression;
                // makes the rule think it should skip the left or right
                const isLeftTypeAssertion = util.isTypeAssertion(node.left);
                const isRightTypeAssertion = util.isTypeAssertion(node.right);
                if (isLeftTypeAssertion && isRightTypeAssertion) {
                    return; // ignore
                }
                if (isLeftTypeAssertion) {
                    return rule(Object.assign(Object.assign({}, node), { left: Object.assign(Object.assign({}, node.left), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                }
                if (isRightTypeAssertion) {
                    return rule(Object.assign(Object.assign({}, node), { right: Object.assign(Object.assign({}, node.right), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                }
                return rule(node);
            }
            function callExp(node) {
                var _a;
                const rule = rules.CallExpression;
                if (util.isTypeAssertion(node.callee)) {
                    // reduces the precedence of the node so the rule thinks it needs to be wrapped
                    return rule(Object.assign(Object.assign({}, node), { callee: Object.assign(Object.assign({}, node.callee), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                }
                if (node.arguments.length === 1 &&
                    ((_a = node.typeParameters) === null || _a === void 0 ? void 0 : _a.params.some(param => param.type === experimental_utils_1.AST_NODE_TYPES.TSParenthesizedType ||
                        param.type === experimental_utils_1.AST_NODE_TYPES.TSImportType))) {
                    return rule(Object.assign(Object.assign({}, node), { arguments: [
                            Object.assign(Object.assign({}, node.arguments[0]), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }),
                        ] }));
                }
                return rule(node);
            }
            function unaryUpdateExpression(node) {
                const rule = rules.UnaryExpression;
                if (util.isTypeAssertion(node.argument)) {
                    // reduces the precedence of the node so the rule thinks it needs to be wrapped
                    return rule(Object.assign(Object.assign({}, node), { argument: Object.assign(Object.assign({}, node.argument), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                }
                return rule(node);
            }
            const overrides = {
                // ArrayExpression
                ArrowFunctionExpression(node) {
                    if (!util.isTypeAssertion(node.body)) {
                        return rules.ArrowFunctionExpression(node);
                    }
                },
                // AssignmentExpression
                // AwaitExpression
                BinaryExpression: binaryExp,
                CallExpression: callExp,
                // ClassDeclaration
                // ClassExpression
                ConditionalExpression(node) {
                    // reduces the precedence of the node so the rule thinks it needs to be wrapped
                    if (util.isTypeAssertion(node.test)) {
                        return rules.ConditionalExpression(Object.assign(Object.assign({}, node), { test: Object.assign(Object.assign({}, node.test), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                    }
                    if (util.isTypeAssertion(node.consequent)) {
                        return rules.ConditionalExpression(Object.assign(Object.assign({}, node), { consequent: Object.assign(Object.assign({}, node.consequent), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                    }
                    if (util.isTypeAssertion(node.alternate)) {
                        // reduces the precedence of the node so the rule thinks it needs to be rapped
                        return rules.ConditionalExpression(Object.assign(Object.assign({}, node), { alternate: Object.assign(Object.assign({}, node.alternate), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                    }
                    return rules.ConditionalExpression(node);
                },
                // DoWhileStatement
                // ForIn and ForOf are guarded by eslint version
                ForStatement(node) {
                    // make the rule skip the piece by removing it entirely
                    if (node.init && util.isTypeAssertion(node.init)) {
                        return rules.ForStatement(Object.assign(Object.assign({}, node), { init: null }));
                    }
                    if (node.test && util.isTypeAssertion(node.test)) {
                        return rules.ForStatement(Object.assign(Object.assign({}, node), { test: null }));
                    }
                    if (node.update && util.isTypeAssertion(node.update)) {
                        return rules.ForStatement(Object.assign(Object.assign({}, node), { update: null }));
                    }
                    return rules.ForStatement(node);
                },
                'ForStatement > *.init:exit'(node) {
                    if (!util.isTypeAssertion(node)) {
                        return rules['ForStatement > *.init:exit'](node);
                    }
                },
                // IfStatement
                LogicalExpression: binaryExp,
                MemberExpression(node) {
                    if (util.isTypeAssertion(node.object)) {
                        // reduces the precedence of the node so the rule thinks it needs to be wrapped
                        return rules.MemberExpression(Object.assign(Object.assign({}, node), { object: Object.assign(Object.assign({}, node.object), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                    }
                    return rules.MemberExpression(node);
                },
                NewExpression: callExp,
                // ObjectExpression
                // ReturnStatement
                // SequenceExpression
                SpreadElement(node) {
                    if (!util.isTypeAssertion(node.argument)) {
                        return rules.SpreadElement(node);
                    }
                },
                SwitchCase(node) {
                    if (node.test && !util.isTypeAssertion(node.test)) {
                        return rules.SwitchCase(node);
                    }
                },
                // SwitchStatement
                ThrowStatement(node) {
                    if (node.argument && !util.isTypeAssertion(node.argument)) {
                        return rules.ThrowStatement(node);
                    }
                },
                UnaryExpression: unaryUpdateExpression,
                UpdateExpression: unaryUpdateExpression,
                // VariableDeclarator
                // WhileStatement
                // WithStatement - i'm not going to even bother implementing this terrible and never used feature
                YieldExpression(node) {
                    if (node.argument && !util.isTypeAssertion(node.argument)) {
                        return rules.YieldExpression(node);
                    }
                },
            };
            if (rules.ForInStatement && rules.ForOfStatement) {
                overrides.ForInStatement = function (node) {
                    if (util.isTypeAssertion(node.right)) {
                        // as of 7.20.0 there's no way to skip checking the right of the ForIn
                        // so just don't validate it at all
                        return;
                    }
                    return rules.ForInStatement(node);
                };
                overrides.ForOfStatement = function (node) {
                    if (util.isTypeAssertion(node.right)) {
                        // makes the rule skip checking of the right
                        return rules.ForOfStatement(Object.assign(Object.assign({}, node), { type: experimental_utils_1.AST_NODE_TYPES.ForOfStatement, right: Object.assign(Object.assign({}, node.right), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                    }
                    return rules.ForOfStatement(node);
                };
            }
            else {
                overrides['ForInStatement, ForOfStatement'] = function (node) {
                    if (util.isTypeAssertion(node.right)) {
                        // makes the rule skip checking of the right
                        return rules['ForInStatement, ForOfStatement'](Object.assign(Object.assign({}, node), { type: experimental_utils_1.AST_NODE_TYPES.ForOfStatement, right: Object.assign(Object.assign({}, node.right), { type: experimental_utils_1.AST_NODE_TYPES.SequenceExpression }) }));
                    }
                    return rules['ForInStatement, ForOfStatement'](node);
                };
            }
            return Object.assign({}, rules, overrides);
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-extra-semi.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const no_extra_semi_1 = __importDefault(require(521) /* eslint/lib/rules/no-extra-semi */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-extra-semi',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow unnecessary semicolons',
                category: 'Possible Errors',
                recommended: 'error',
                extendsBaseRule: true,
            },
            fixable: 'code',
            schema: no_extra_semi_1.default.meta.schema,
            messages: no_extra_semi_1.default.meta.messages,
        },
        defaultOptions: [],
        create(context) {
            const rules = no_extra_semi_1.default.create(context);
            return Object.assign(Object.assign({}, rules), { ClassProperty(node) {
                    rules.MethodDefinition(node);
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-extraneous-class.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-extraneous-class',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Forbids the use of classes as namespaces',
                category: 'Best Practices',
                recommended: false,
            },
            schema: [
                {
                    type: 'object',
                    additionalProperties: false,
                    properties: {
                        allowConstructorOnly: {
                            type: 'boolean',
                        },
                        allowEmpty: {
                            type: 'boolean',
                        },
                        allowStaticOnly: {
                            type: 'boolean',
                        },
                        allowWithDecorator: {
                            type: 'boolean',
                        },
                    },
                },
            ],
            messages: {
                empty: 'Unexpected empty class.',
                onlyStatic: 'Unexpected class with only static properties.',
                onlyConstructor: 'Unexpected class with only a constructor.',
            },
        },
        defaultOptions: [
            {
                allowConstructorOnly: false,
                allowEmpty: false,
                allowStaticOnly: false,
                allowWithDecorator: false,
            },
        ],
        create(context, [{ allowConstructorOnly, allowEmpty, allowStaticOnly, allowWithDecorator }]) {
            const isAllowWithDecorator = (node) => {
                return !!(allowWithDecorator &&
                    node &&
                    node.decorators &&
                    node.decorators.length);
            };
            return {
                ClassBody(node) {
                    const parent = node.parent;
                    if (!parent || parent.superClass || isAllowWithDecorator(parent)) {
                        return;
                    }
                    const reportNode = 'id' in parent && parent.id ? parent.id : parent;
                    if (node.body.length === 0) {
                        if (allowEmpty) {
                            return;
                        }
                        context.report({
                            node: reportNode,
                            messageId: 'empty',
                        });
                        return;
                    }
                    let onlyStatic = true;
                    let onlyConstructor = true;
                    for (const prop of node.body) {
                        if ('kind' in prop && prop.kind === 'constructor') {
                            if (prop.value.params.some(param => param.type === experimental_utils_1.AST_NODE_TYPES.TSParameterProperty)) {
                                onlyConstructor = false;
                                onlyStatic = false;
                            }
                        }
                        else {
                            onlyConstructor = false;
                            if ('static' in prop && !prop.static) {
                                onlyStatic = false;
                            }
                        }
                        if (!(onlyStatic || onlyConstructor)) {
                            break;
                        }
                    }
                    if (onlyConstructor) {
                        if (!allowConstructorOnly) {
                            context.report({
                                node: reportNode,
                                messageId: 'onlyConstructor',
                            });
                        }
                        return;
                    }
                    if (onlyStatic && !allowStaticOnly) {
                        context.report({
                            node: reportNode,
                            messageId: 'onlyStatic',
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-floating-promises.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const tsutils = __importStar(require(916) /* tsutils */);
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-floating-promises',
        meta: {
            docs: {
                description: 'Requires Promise-like values to be handled appropriately',
                category: 'Best Practices',
                recommended: 'error',
                suggestion: true,
                requiresTypeChecking: true,
            },
            messages: {
                floating: 'Promises must be handled appropriately.',
                floatingVoid: 'Promises must be handled appropriately' +
                    ' or explicitly marked as ignored with the `void` operator.',
                floatingFixVoid: 'Add void operator to ignore.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        ignoreVoid: { type: 'boolean' },
                        ignoreIIFE: { type: 'boolean' },
                    },
                    additionalProperties: false,
                },
            ],
            type: 'problem',
        },
        defaultOptions: [
            {
                ignoreVoid: true,
                ignoreIIFE: false,
            },
        ],
        create(context, [options]) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const sourceCode = context.getSourceCode();
            return {
                ExpressionStatement(node) {
                    if (options.ignoreIIFE && isAsyncIife(node)) {
                        return;
                    }
                    if (isUnhandledPromise(checker, node.expression)) {
                        if (options.ignoreVoid) {
                            context.report({
                                node,
                                messageId: 'floatingVoid',
                                suggest: [
                                    {
                                        messageId: 'floatingFixVoid',
                                        fix(fixer) {
                                            let code = sourceCode.getText(node);
                                            code = `void ${code}`;
                                            return fixer.replaceText(node, code);
                                        },
                                    },
                                ],
                            });
                        }
                        else {
                            context.report({
                                node,
                                messageId: 'floating',
                            });
                        }
                    }
                },
            };
            function isAsyncIife(node) {
                if (node.expression.type !== experimental_utils_1.AST_NODE_TYPES.CallExpression) {
                    return false;
                }
                return (node.expression.type === experimental_utils_1.AST_NODE_TYPES.CallExpression &&
                    (node.expression.callee.type ===
                        experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression ||
                        node.expression.callee.type === experimental_utils_1.AST_NODE_TYPES.FunctionExpression));
            }
            function isUnhandledPromise(checker, node) {
                // First, check expressions whose resulting types may not be promise-like
                if (node.type === experimental_utils_1.AST_NODE_TYPES.SequenceExpression) {
                    // Any child in a comma expression could return a potentially unhandled
                    // promise, so we check them all regardless of whether the final returned
                    // value is promise-like.
                    return node.expressions.some(item => isUnhandledPromise(checker, item));
                }
                if (!options.ignoreVoid &&
                    node.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                    node.operator === 'void') {
                    // Similarly, a `void` expression always returns undefined, so we need to
                    // see what's inside it without checking the type of the overall expression.
                    return isUnhandledPromise(checker, node.argument);
                }
                // Check the type. At this point it can't be unhandled if it isn't a promise
                if (!isPromiseLike(checker, parserServices.esTreeNodeToTSNodeMap.get(node))) {
                    return false;
                }
                if (node.type === experimental_utils_1.AST_NODE_TYPES.CallExpression) {
                    // If the outer expression is a call, it must be either a `.then()` or
                    // `.catch()` that handles the promise.
                    return (!isPromiseCatchCallWithHandler(node) &&
                        !isPromiseThenCallWithRejectionHandler(node) &&
                        !isPromiseFinallyCallWithHandler(node));
                }
                else if (node.type === experimental_utils_1.AST_NODE_TYPES.ConditionalExpression) {
                    // We must be getting the promise-like value from one of the branches of the
                    // ternary. Check them directly.
                    return (isUnhandledPromise(checker, node.alternate) ||
                        isUnhandledPromise(checker, node.consequent));
                }
                else if (node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression ||
                    node.type === experimental_utils_1.AST_NODE_TYPES.Identifier ||
                    node.type === experimental_utils_1.AST_NODE_TYPES.NewExpression) {
                    // If it is just a property access chain or a `new` call (e.g. `foo.bar` or
                    // `new Promise()`), the promise is not handled because it doesn't have the
                    // necessary then/catch call at the end of the chain.
                    return true;
                }
                // We conservatively return false for all other types of expressions because
                // we don't want to accidentally fail if the promise is handled internally but
                // we just can't tell.
                return false;
            }
        },
    });
    // Modified from tsutils.isThenable() to only consider thenables which can be
    // rejected/caught via a second parameter. Original source (MIT licensed):
    //
    //   https://github.com/ajafff/tsutils/blob/49d0d31050b44b81e918eae4fbaf1dfe7b7286af/util/type.ts#L95-L125
    function isPromiseLike(checker, node) {
        const type = checker.getTypeAtLocation(node);
        for (const ty of tsutils.unionTypeParts(checker.getApparentType(type))) {
            const then = ty.getProperty('then');
            if (then === undefined) {
                continue;
            }
            const thenType = checker.getTypeOfSymbolAtLocation(then, node);
            if (hasMatchingSignature(thenType, signature => signature.parameters.length >= 2 &&
                isFunctionParam(checker, signature.parameters[0], node) &&
                isFunctionParam(checker, signature.parameters[1], node))) {
                return true;
            }
        }
        return false;
    }
    function hasMatchingSignature(type, matcher) {
        for (const t of tsutils.unionTypeParts(type)) {
            if (t.getCallSignatures().some(matcher)) {
                return true;
            }
        }
        return false;
    }
    function isFunctionParam(checker, param, node) {
        const type = checker.getApparentType(checker.getTypeOfSymbolAtLocation(param, node));
        for (const t of tsutils.unionTypeParts(type)) {
            if (t.getCallSignatures().length !== 0) {
                return true;
            }
        }
        return false;
    }
    function isPromiseCatchCallWithHandler(expression) {
        return (expression.callee.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
            expression.callee.property.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
            expression.callee.property.name === 'catch' &&
            expression.arguments.length >= 1);
    }
    function isPromiseThenCallWithRejectionHandler(expression) {
        return (expression.callee.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
            expression.callee.property.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
            expression.callee.property.name === 'then' &&
            expression.arguments.length >= 2);
    }
    function isPromiseFinallyCallWithHandler(expression) {
        return (expression.callee.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
            expression.callee.property.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
            expression.callee.property.name === 'finally' &&
            expression.arguments.length >= 1);
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-for-in-array.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-for-in-array',
        meta: {
            docs: {
                description: 'Disallow iterating over an array with a for-in loop',
                category: 'Best Practices',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                forInViolation: 'For-in loops over arrays are forbidden. Use for-of or array.forEach instead.',
            },
            schema: [],
            type: 'problem',
        },
        defaultOptions: [],
        create(context) {
            return {
                ForInStatement(node) {
                    const parserServices = util.getParserServices(context);
                    const checker = parserServices.program.getTypeChecker();
                    const originalNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                    const type = util.getConstrainedTypeAtLocation(checker, originalNode.expression);
                    if (util.isTypeArrayTypeOrUnionOfArrayTypes(type, checker) ||
                        (type.flags & ts.TypeFlags.StringLike) !== 0) {
                        context.report({
                            node,
                            messageId: 'forInViolation',
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-implicit-any-catch.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    exports.default = util.createRule({
        name: 'no-implicit-any-catch',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow usage of the implicit `any` type in catch clauses',
                category: 'Best Practices',
                recommended: false,
                suggestion: true,
            },
            fixable: 'code',
            messages: {
                implicitAnyInCatch: 'Implicit any in catch clause',
                explicitAnyInCatch: 'Explicit any in catch clause',
                suggestExplicitUnknown: 'Use `unknown` instead, this will force you to explicitly, and safely assert the type is correct.',
            },
            schema: [
                {
                    type: 'object',
                    additionalProperties: false,
                    properties: {
                        allowExplicitAny: {
                            type: 'boolean',
                        },
                    },
                },
            ],
        },
        defaultOptions: [
            {
                allowExplicitAny: false,
            },
        ],
        create(context, [{ allowExplicitAny }]) {
            return {
                CatchClause(node) {
                    if (!node.param) {
                        return; // ignore catch without variable
                    }
                    if (!node.param.typeAnnotation) {
                        context.report({
                            node,
                            messageId: 'implicitAnyInCatch',
                            suggest: [
                                {
                                    messageId: 'suggestExplicitUnknown',
                                    fix(fixer) {
                                        return fixer.insertTextAfter(node.param, ': unknown');
                                    },
                                },
                            ],
                        });
                    }
                    else if (!allowExplicitAny &&
                        node.param.typeAnnotation.typeAnnotation.type ===
                            experimental_utils_1.AST_NODE_TYPES.TSAnyKeyword) {
                        context.report({
                            node,
                            messageId: 'explicitAnyInCatch',
                            suggest: [
                                {
                                    messageId: 'suggestExplicitUnknown',
                                    fix(fixer) {
                                        return fixer.replaceText(node.param.typeAnnotation, ': unknown');
                                    },
                                },
                            ],
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-implied-eval.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const util = __importStar(require(906) /* ../util */);
    const FUNCTION_CONSTRUCTOR = 'Function';
    const GLOBAL_CANDIDATES = new Set(['global', 'window', 'globalThis']);
    const EVAL_LIKE_METHODS = new Set([
        'setImmediate',
        'setInterval',
        'setTimeout',
        'execScript',
    ]);
    exports.default = util.createRule({
        name: 'no-implied-eval',
        meta: {
            docs: {
                description: 'Disallow the use of `eval()`-like methods',
                category: 'Best Practices',
                recommended: 'error',
                extendsBaseRule: true,
                requiresTypeChecking: true,
            },
            messages: {
                noImpliedEvalError: 'Implied eval. Consider passing a function.',
                noFunctionConstructor: 'Implied eval. Do not use the Function constructor to create functions.',
            },
            schema: [],
            type: 'suggestion',
        },
        defaultOptions: [],
        create(context) {
            const parserServices = util.getParserServices(context);
            const program = parserServices.program;
            const checker = parserServices.program.getTypeChecker();
            function getCalleeName(node) {
                if (node.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                    return node.name;
                }
                if (node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                    node.object.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    GLOBAL_CANDIDATES.has(node.object.name)) {
                    if (node.property.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                        return node.property.name;
                    }
                    if (node.property.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                        typeof node.property.value === 'string') {
                        return node.property.value;
                    }
                }
                return null;
            }
            function isFunctionType(node) {
                var _a;
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                const type = checker.getTypeAtLocation(tsNode);
                const symbol = type.getSymbol();
                if (symbol &&
                    tsutils.isSymbolFlagSet(symbol, ts.SymbolFlags.Function | ts.SymbolFlags.Method)) {
                    return true;
                }
                if (symbol && symbol.escapedName === FUNCTION_CONSTRUCTOR) {
                    const declarations = (_a = symbol.getDeclarations()) !== null && _a !== void 0 ? _a : [];
                    for (const declaration of declarations) {
                        const sourceFile = declaration.getSourceFile();
                        if (program.isSourceFileDefaultLibrary(sourceFile)) {
                            return true;
                        }
                    }
                }
                const signatures = checker.getSignaturesOfType(type, ts.SignatureKind.Call);
                return signatures.length > 0;
            }
            function isBind(node) {
                return node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression
                    ? isBind(node.property)
                    : node.type === experimental_utils_1.AST_NODE_TYPES.Identifier && node.name === 'bind';
            }
            function isFunction(node) {
                switch (node.type) {
                    case experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression:
                    case experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration:
                    case experimental_utils_1.AST_NODE_TYPES.FunctionExpression:
                        return true;
                    case experimental_utils_1.AST_NODE_TYPES.Literal:
                    case experimental_utils_1.AST_NODE_TYPES.TemplateLiteral:
                        return false;
                    case experimental_utils_1.AST_NODE_TYPES.CallExpression:
                        return isBind(node.callee) || isFunctionType(node);
                    default:
                        return isFunctionType(node);
                }
            }
            function checkImpliedEval(node) {
                var _a;
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node.callee);
                const type = checker.getTypeAtLocation(tsNode);
                const calleeName = getCalleeName(node.callee);
                if (calleeName === null) {
                    return;
                }
                if (calleeName === FUNCTION_CONSTRUCTOR) {
                    const symbol = type.getSymbol();
                    if (symbol) {
                        const declarations = (_a = symbol.getDeclarations()) !== null && _a !== void 0 ? _a : [];
                        for (const declaration of declarations) {
                            const sourceFile = declaration.getSourceFile();
                            if (program.isSourceFileDefaultLibrary(sourceFile)) {
                                context.report({ node, messageId: 'noFunctionConstructor' });
                                return;
                            }
                        }
                    }
                    else {
                        context.report({ node, messageId: 'noFunctionConstructor' });
                        return;
                    }
                }
                if (node.arguments.length === 0) {
                    return;
                }
                const [handler] = node.arguments;
                if (EVAL_LIKE_METHODS.has(calleeName) && !isFunction(handler)) {
                    context.report({ node: handler, messageId: 'noImpliedEvalError' });
                }
            }
            return {
                NewExpression: checkImpliedEval,
                CallExpression: checkImpliedEval,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-inferrable-types.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-inferrable-types',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallows explicit type declarations for variables or parameters initialized to a number, string, or boolean',
                category: 'Best Practices',
                recommended: 'error',
            },
            fixable: 'code',
            messages: {
                noInferrableType: 'Type {{type}} trivially inferred from a {{type}} literal, remove type annotation.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        ignoreParameters: {
                            type: 'boolean',
                        },
                        ignoreProperties: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                ignoreParameters: false,
                ignoreProperties: false,
            },
        ],
        create(context, [{ ignoreParameters, ignoreProperties }]) {
            function isFunctionCall(init, callName) {
                if (init.type === experimental_utils_1.AST_NODE_TYPES.ChainExpression) {
                    return isFunctionCall(init.expression, callName);
                }
                return (init.type === experimental_utils_1.AST_NODE_TYPES.CallExpression &&
                    init.callee.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    init.callee.name === callName);
            }
            function isLiteral(init, typeName) {
                return (init.type === experimental_utils_1.AST_NODE_TYPES.Literal && typeof init.value === typeName);
            }
            function isIdentifier(init, ...names) {
                return (init.type === experimental_utils_1.AST_NODE_TYPES.Identifier && names.includes(init.name));
            }
            function hasUnaryPrefix(init, ...operators) {
                return (init.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                    operators.includes(init.operator));
            }
            const keywordMap = {
                [experimental_utils_1.AST_NODE_TYPES.TSBigIntKeyword]: 'bigint',
                [experimental_utils_1.AST_NODE_TYPES.TSBooleanKeyword]: 'boolean',
                [experimental_utils_1.AST_NODE_TYPES.TSNumberKeyword]: 'number',
                [experimental_utils_1.AST_NODE_TYPES.TSNullKeyword]: 'null',
                [experimental_utils_1.AST_NODE_TYPES.TSStringKeyword]: 'string',
                [experimental_utils_1.AST_NODE_TYPES.TSSymbolKeyword]: 'symbol',
                [experimental_utils_1.AST_NODE_TYPES.TSUndefinedKeyword]: 'undefined',
            };
            /**
             * Returns whether a node has an inferrable value or not
             */
            function isInferrable(annotation, init) {
                switch (annotation.type) {
                    case experimental_utils_1.AST_NODE_TYPES.TSBigIntKeyword: {
                        // note that bigint cannot have + prefixed to it
                        const unwrappedInit = hasUnaryPrefix(init, '-')
                            ? init.argument
                            : init;
                        return (isFunctionCall(unwrappedInit, 'BigInt') ||
                            (unwrappedInit.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                                'bigint' in unwrappedInit));
                    }
                    case experimental_utils_1.AST_NODE_TYPES.TSBooleanKeyword:
                        return (hasUnaryPrefix(init, '!') ||
                            // eslint-disable-next-line @typescript-eslint/internal/prefer-ast-types-enum
                            isFunctionCall(init, 'Boolean') ||
                            isLiteral(init, 'boolean'));
                    case experimental_utils_1.AST_NODE_TYPES.TSNumberKeyword: {
                        const unwrappedInit = hasUnaryPrefix(init, '+', '-')
                            ? init.argument
                            : init;
                        return (isIdentifier(unwrappedInit, 'Infinity', 'NaN') ||
                            isFunctionCall(unwrappedInit, 'Number') ||
                            isLiteral(unwrappedInit, 'number'));
                    }
                    case experimental_utils_1.AST_NODE_TYPES.TSNullKeyword:
                        return init.type === experimental_utils_1.AST_NODE_TYPES.Literal && init.value === null;
                    case experimental_utils_1.AST_NODE_TYPES.TSStringKeyword:
                        return (
                        // eslint-disable-next-line @typescript-eslint/internal/prefer-ast-types-enum
                        isFunctionCall(init, 'String') ||
                            isLiteral(init, 'string') ||
                            init.type === experimental_utils_1.AST_NODE_TYPES.TemplateLiteral);
                    case experimental_utils_1.AST_NODE_TYPES.TSSymbolKeyword:
                        return isFunctionCall(init, 'Symbol');
                    case experimental_utils_1.AST_NODE_TYPES.TSTypeReference: {
                        if (annotation.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                            annotation.typeName.name === 'RegExp') {
                            const isRegExpLiteral = init.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                                init.value instanceof RegExp;
                            const isRegExpNewCall = init.type === experimental_utils_1.AST_NODE_TYPES.NewExpression &&
                                init.callee.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                                init.callee.name === 'RegExp';
                            const isRegExpCall = isFunctionCall(init, 'RegExp');
                            return isRegExpLiteral || isRegExpCall || isRegExpNewCall;
                        }
                        return false;
                    }
                    case experimental_utils_1.AST_NODE_TYPES.TSUndefinedKeyword:
                        return (hasUnaryPrefix(init, 'void') || isIdentifier(init, 'undefined'));
                }
                return false;
            }
            /**
             * Reports an inferrable type declaration, if any
             */
            function reportInferrableType(node, typeNode, initNode) {
                if (!typeNode || !initNode || !typeNode.typeAnnotation) {
                    return;
                }
                if (!isInferrable(typeNode.typeAnnotation, initNode)) {
                    return;
                }
                const type = typeNode.typeAnnotation.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference
                    ? // TODO - if we add more references
                        'RegExp'
                    : keywordMap[typeNode.typeAnnotation.type];
                context.report({
                    node,
                    messageId: 'noInferrableType',
                    data: {
                        type,
                    },
                    fix: fixer => fixer.remove(typeNode),
                });
            }
            function inferrableVariableVisitor(node) {
                if (!node.id) {
                    return;
                }
                reportInferrableType(node, node.id.typeAnnotation, node.init);
            }
            function inferrableParameterVisitor(node) {
                if (ignoreParameters || !node.params) {
                    return;
                }
                node.params.filter(param => param.type === experimental_utils_1.AST_NODE_TYPES.AssignmentPattern &&
                    param.left &&
                    param.right).forEach(param => {
                    reportInferrableType(param, param.left.typeAnnotation, param.right);
                });
            }
            function inferrablePropertyVisitor(node) {
                // We ignore `readonly` because of Microsoft/TypeScript#14416
                // Essentially a readonly property without a type
                // will result in its value being the type, leading to
                // compile errors if the type is stripped.
                if (ignoreProperties || node.readonly || node.optional) {
                    return;
                }
                reportInferrableType(node, node.typeAnnotation, node.value);
            }
            return {
                VariableDeclarator: inferrableVariableVisitor,
                FunctionExpression: inferrableParameterVisitor,
                FunctionDeclaration: inferrableParameterVisitor,
                ArrowFunctionExpression: inferrableParameterVisitor,
                ClassProperty: inferrablePropertyVisitor,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-invalid-this.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const no_invalid_this_1 = __importDefault(require(533) /* eslint/lib/rules/no-invalid-this */);
    const util_1 = require(906) /* ../util */;
    exports.default = util_1.createRule({
        name: 'no-invalid-this',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow `this` keywords outside of classes or class-like objects',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
            },
            messages: (_a = no_invalid_this_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                unexpectedThis: "Unexpected 'this'.",
            },
            schema: no_invalid_this_1.default.meta.schema,
        },
        defaultOptions: [{ capIsConstructor: true }],
        create(context) {
            const rules = no_invalid_this_1.default.create(context);
            /**
             * Since function definitions can be nested we use a stack storing if "this" is valid in the current context.
             *
             * Example:
             *
             * function a(this: number) { // valid "this"
             *     function b() {
             *         console.log(this); // invalid "this"
             *     }
             * }
             *
             * When parsing the function declaration of "a" the stack will be: [true]
             * When parsing the function declaration of "b" the stack will be: [true, false]
             */
            const thisIsValidStack = [];
            return Object.assign(Object.assign({}, rules), { ClassProperty() {
                    thisIsValidStack.push(true);
                },
                'ClassProperty:exit'() {
                    thisIsValidStack.pop();
                },
                FunctionDeclaration(node) {
                    thisIsValidStack.push(node.params.some(param => param.type === experimental_utils_1.AST_NODE_TYPES.Identifier && param.name === 'this'));
                    // baseRule's work
                    rules.FunctionDeclaration(node);
                },
                'FunctionDeclaration:exit'(node) {
                    thisIsValidStack.pop();
                    // baseRule's work
                    rules['FunctionDeclaration:exit'](node);
                },
                FunctionExpression(node) {
                    thisIsValidStack.push(node.params.some(param => param.type === experimental_utils_1.AST_NODE_TYPES.Identifier && param.name === 'this'));
                    // baseRule's work
                    rules.FunctionExpression(node);
                },
                'FunctionExpression:exit'(node) {
                    thisIsValidStack.pop();
                    // baseRule's work
                    rules['FunctionExpression:exit'](node);
                },
                ThisExpression(node) {
                    const thisIsValidHere = thisIsValidStack[thisIsValidStack.length - 1];
                    if (thisIsValidHere) {
                        return;
                    }
                    // baseRule's work
                    rules.ThisExpression(node);
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-invalid-void-type.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-invalid-void-type',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows usage of `void` type outside of generic or return types',
                category: 'Best Practices',
                recommended: false,
            },
            messages: {
                invalidVoidForGeneric: '{{ generic }} may not have void as a type variable',
                invalidVoidNotReturnOrGeneric: 'void is only valid as a return type or generic type variable',
                invalidVoidNotReturn: 'void is only valid as a return type',
                invalidVoidNotReturnOrThisParam: 'void is only valid as return type or type of `this` parameter',
                invalidVoidNotReturnOrThisParamOrGeneric: 'void is only valid as a return type or generic type variable or the type of a `this` parameter',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowInGenericTypeArguments: {
                            oneOf: [
                                { type: 'boolean' },
                                {
                                    type: 'array',
                                    items: { type: 'string' },
                                    minLength: 1,
                                },
                            ],
                        },
                        allowAsThisParameter: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            { allowInGenericTypeArguments: true, allowAsThisParameter: false },
        ],
        create(context, [{ allowInGenericTypeArguments, allowAsThisParameter }]) {
            const validParents = [
                experimental_utils_1.AST_NODE_TYPES.TSTypeAnnotation, //
            ];
            const invalidGrandParents = [
                experimental_utils_1.AST_NODE_TYPES.TSPropertySignature,
                experimental_utils_1.AST_NODE_TYPES.CallExpression,
                experimental_utils_1.AST_NODE_TYPES.ClassProperty,
                experimental_utils_1.AST_NODE_TYPES.Identifier,
            ];
            const validUnionMembers = [
                experimental_utils_1.AST_NODE_TYPES.TSVoidKeyword,
                experimental_utils_1.AST_NODE_TYPES.TSNeverKeyword,
            ];
            if (allowInGenericTypeArguments === true) {
                validParents.push(experimental_utils_1.AST_NODE_TYPES.TSTypeParameterInstantiation);
            }
            /**
             * @brief check if the given void keyword is used as a valid generic type
             *
             * reports if the type parametrized by void is not in the whitelist, or
             * allowInGenericTypeArguments is false.
             * no-op if the given void keyword is not used as generic type
             */
            function checkGenericTypeArgument(node) {
                var _a, _b;
                // only matches T<..., void, ...>
                // extra check for precaution
                /* istanbul ignore next */
                if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) !== experimental_utils_1.AST_NODE_TYPES.TSTypeParameterInstantiation ||
                    ((_b = node.parent.parent) === null || _b === void 0 ? void 0 : _b.type) !== experimental_utils_1.AST_NODE_TYPES.TSTypeReference) {
                    return;
                }
                // check whitelist
                if (Array.isArray(allowInGenericTypeArguments)) {
                    const sourceCode = context.getSourceCode();
                    const fullyQualifiedName = sourceCode
                        .getText(node.parent.parent.typeName)
                        .replace(/ /gu, '');
                    if (!allowInGenericTypeArguments
                        .map(s => s.replace(/ /gu, ''))
                        .includes(fullyQualifiedName)) {
                        context.report({
                            messageId: 'invalidVoidForGeneric',
                            data: { generic: fullyQualifiedName },
                            node,
                        });
                    }
                    return;
                }
                if (!allowInGenericTypeArguments) {
                    context.report({
                        messageId: allowAsThisParameter
                            ? 'invalidVoidNotReturnOrThisParam'
                            : 'invalidVoidNotReturn',
                        node,
                    });
                }
            }
            /**
             * @brief checks that a union containing void is valid
             * @return true if every member of the union is specified as a valid type in
             * validUnionMembers, or is a valid generic type parametrized by void
             */
            function isValidUnionType(node) {
                return node.types.every(member => {
                    var _a, _b;
                    return validUnionMembers.includes(member.type) ||
                        // allows any T<..., void, ...> here, checked by checkGenericTypeArgument
                        (member.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference &&
                            ((_a = member.typeParameters) === null || _a === void 0 ? void 0 : _a.type) ===
                                experimental_utils_1.AST_NODE_TYPES.TSTypeParameterInstantiation &&
                            ((_b = member.typeParameters) === null || _b === void 0 ? void 0 : _b.params.map(param => param.type).includes(experimental_utils_1.AST_NODE_TYPES.TSVoidKeyword)));
                });
            }
            return {
                TSVoidKeyword(node) {
                    var _a;
                    /* istanbul ignore next */
                    if (!((_a = node.parent) === null || _a === void 0 ? void 0 : _a.parent)) {
                        return;
                    }
                    // checks T<..., void, ...> against specification of allowInGenericArguments option
                    if (node.parent.type === experimental_utils_1.AST_NODE_TYPES.TSTypeParameterInstantiation &&
                        node.parent.parent.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference) {
                        checkGenericTypeArgument(node);
                        return;
                    }
                    // union w/ void must contain types from validUnionMembers, or a valid generic void type
                    if (node.parent.type === experimental_utils_1.AST_NODE_TYPES.TSUnionType &&
                        isValidUnionType(node.parent)) {
                        return;
                    }
                    // this parameter is ok to be void.
                    if (allowAsThisParameter &&
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.TSTypeAnnotation &&
                        node.parent.parent.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                        node.parent.parent.name === 'this') {
                        return;
                    }
                    // default cases
                    if (validParents.includes(node.parent.type) &&
                        !invalidGrandParents.includes(node.parent.parent.type)) {
                        return;
                    }
                    context.report({
                        messageId: allowInGenericTypeArguments && allowAsThisParameter
                            ? 'invalidVoidNotReturnOrThisParamOrGeneric'
                            : allowInGenericTypeArguments
                                ? 'invalidVoidNotReturnOrGeneric'
                                : allowAsThisParameter
                                    ? 'invalidVoidNotReturnOrThisParam'
                                    : 'invalidVoidNotReturn',
                        node,
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-loop-func.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const no_loop_func_1 = __importDefault(require(540) /* eslint/lib/rules/no-loop-func */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-loop-func',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow function declarations that contain unsafe references inside loop statements',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: [],
            messages: (_a = no_loop_func_1.default === null || no_loop_func_1.default === void 0 ? void 0 : no_loop_func_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                unsafeRefs: 'Function declared in a loop contains unsafe references to variable(s) {{ varNames }}.',
            },
        },
        defaultOptions: [],
        create(context) {
            /**
             * Reports functions which match the following condition:
             * - has a loop node in ancestors.
             * - has any references which refers to an unsafe variable.
             *
             * @param node The AST node to check.
             * @returns Whether or not the node is within a loop.
             */
            function checkForLoops(node) {
                const loopNode = getContainingLoopNode(node);
                if (!loopNode) {
                    return;
                }
                const references = context.getScope().through;
                const unsafeRefs = references
                    .filter(r => !isSafe(loopNode, r))
                    .map(r => r.identifier.name);
                if (unsafeRefs.length > 0) {
                    context.report({
                        node,
                        messageId: 'unsafeRefs',
                        data: { varNames: `'${unsafeRefs.join("', '")}'` },
                    });
                }
            }
            return {
                ArrowFunctionExpression: checkForLoops,
                FunctionExpression: checkForLoops,
                FunctionDeclaration: checkForLoops,
            };
        },
    });
    /**
     * Gets the containing loop node of a specified node.
     *
     * We don't need to check nested functions, so this ignores those.
     * `Scope.through` contains references of nested functions.
     *
     * @param node An AST node to get.
     * @returns The containing loop node of the specified node, or `null`.
     */
    function getContainingLoopNode(node) {
        for (let currentNode = node; currentNode.parent; currentNode = currentNode.parent) {
            const parent = currentNode.parent;
            switch (parent.type) {
                case experimental_utils_1.AST_NODE_TYPES.WhileStatement:
                case experimental_utils_1.AST_NODE_TYPES.DoWhileStatement:
                    return parent;
                case experimental_utils_1.AST_NODE_TYPES.ForStatement:
                    // `init` is outside of the loop.
                    if (parent.init !== currentNode) {
                        return parent;
                    }
                    break;
                case experimental_utils_1.AST_NODE_TYPES.ForInStatement:
                case experimental_utils_1.AST_NODE_TYPES.ForOfStatement:
                    // `right` is outside of the loop.
                    if (parent.right !== currentNode) {
                        return parent;
                    }
                    break;
                case experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression:
                case experimental_utils_1.AST_NODE_TYPES.FunctionExpression:
                case experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration:
                    // We don't need to check nested functions.
                    return null;
                default:
                    break;
            }
        }
        return null;
    }
    /**
     * Gets the containing loop node of a given node.
     * If the loop was nested, this returns the most outer loop.
     * @param node A node to get. This is a loop node.
     * @param excludedNode A node that the result node should not include.
     * @returns The most outer loop node.
     */
    function getTopLoopNode(node, excludedNode) {
        const border = excludedNode ? excludedNode.range[1] : 0;
        let retv = node;
        let containingLoopNode = node;
        while (containingLoopNode && containingLoopNode.range[0] >= border) {
            retv = containingLoopNode;
            containingLoopNode = getContainingLoopNode(containingLoopNode);
        }
        return retv;
    }
    /**
     * Checks whether a given reference which refers to an upper scope's variable is
     * safe or not.
     * @param loopNode A containing loop node.
     * @param reference A reference to check.
     * @returns `true` if the reference is safe or not.
     */
    function isSafe(loopNode, reference) {
        var _a;
        const variable = reference.resolved;
        const definition = variable === null || variable === void 0 ? void 0 : variable.defs[0];
        const declaration = definition === null || definition === void 0 ? void 0 : definition.parent;
        const kind = (declaration === null || declaration === void 0 ? void 0 : declaration.type) === experimental_utils_1.AST_NODE_TYPES.VariableDeclaration
            ? declaration.kind
            : '';
        // type references are all safe
        // this only really matters for global types that haven't been configured
        if (reference.isTypeReference) {
            return true;
        }
        // Variables which are declared by `const` is safe.
        if (kind === 'const') {
            return true;
        }
        /*
         * Variables which are declared by `let` in the loop is safe.
         * It's a different instance from the next loop step's.
         */
        if (kind === 'let' &&
            declaration &&
            declaration.range[0] > loopNode.range[0] &&
            declaration.range[1] < loopNode.range[1]) {
            return true;
        }
        /*
         * WriteReferences which exist after this border are unsafe because those
         * can modify the variable.
         */
        const border = getTopLoopNode(loopNode, kind === 'let' ? declaration : null)
            .range[0];
        /**
         * Checks whether a given reference is safe or not.
         * The reference is every reference of the upper scope's variable we are
         * looking now.
         *
         * It's safe if the reference matches one of the following condition.
         * - is readonly.
         * - doesn't exist inside a local function and after the border.
         *
         * @param upperRef A reference to check.
         * @returns `true` if the reference is safe.
         */
        function isSafeReference(upperRef) {
            var _a;
            const id = upperRef.identifier;
            return (!upperRef.isWrite() ||
                (((_a = variable === null || variable === void 0 ? void 0 : variable.scope) === null || _a === void 0 ? void 0 : _a.variableScope) === upperRef.from.variableScope &&
                    id.range[0] < border));
        }
        return (_a = variable === null || variable === void 0 ? void 0 : variable.references.every(isSafeReference)) !== null && _a !== void 0 ? _a : false;
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-loss-of-precision.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var _a;
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    const baseRule = (() => {
        try {
            return require(541) /* eslint/lib/rules/no-loss-of-precision */;
        }
        catch (_a) {
            /* istanbul ignore next */
            return null;
        }
    })();
    exports.default = util.createRule({
        name: 'no-loss-of-precision',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow literal numbers that lose precision',
                category: 'Possible Errors',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: [],
            messages: (_a = baseRule === null || baseRule === void 0 ? void 0 : baseRule.meta.messages) !== null && _a !== void 0 ? _a : { noLossOfPrecision: '' },
        },
        defaultOptions: [],
        create(context) {
            /* istanbul ignore if */ if (baseRule === null) {
                throw new Error('@typescript-eslint/no-loss-of-precision requires at least ESLint v7.1.0');
            }
            // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-assertion
            const rules = baseRule.create(context);
            function isSeparatedNumeric(node) {
                return typeof node.value === 'number' && node.raw.includes('_');
            }
            return {
                Literal(node) {
                    rules.Literal(Object.assign(Object.assign({}, node), { raw: isSeparatedNumeric(node) ? node.raw.replace(/_/g, '') : node.raw }));
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-magic-numbers.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const no_magic_numbers_1 = __importDefault(require(542) /* eslint/lib/rules/no-magic-numbers */);
    const util = __importStar(require(906) /* ../util */);
    const baseRuleSchema = Array.isArray(no_magic_numbers_1.default.meta.schema)
        ? no_magic_numbers_1.default.meta.schema[0]
        : no_magic_numbers_1.default.meta.schema;
    exports.default = util.createRule({
        name: 'no-magic-numbers',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow magic numbers',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
            },
            // Extend base schema with additional property to ignore TS numeric literal types
            schema: [
                Object.assign(Object.assign({}, baseRuleSchema), { properties: Object.assign(Object.assign({}, baseRuleSchema.properties), { ignoreNumericLiteralTypes: {
                            type: 'boolean',
                        }, ignoreEnums: {
                            type: 'boolean',
                        }, ignoreReadonlyClassProperties: {
                            type: 'boolean',
                        } }) }),
            ],
            messages: (_a = no_magic_numbers_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                useConst: "Number constants declarations must use 'const'.",
                noMagic: 'No magic number: {{raw}}.',
            },
        },
        defaultOptions: [
            {
                ignore: [],
                ignoreArrayIndexes: false,
                enforceConst: false,
                detectObjects: false,
                ignoreNumericLiteralTypes: false,
                ignoreEnums: false,
                ignoreReadonlyClassProperties: false,
            },
        ],
        create(context, [options]) {
            const rules = no_magic_numbers_1.default.create(context);
            return {
                Literal(node) {
                    var _a;
                    // Check if the node is a TypeScript enum declaration
                    if (options.ignoreEnums && isParentTSEnumDeclaration(node)) {
                        return;
                    }
                    // Check TypeScript specific nodes for Numeric Literal
                    if (options.ignoreNumericLiteralTypes &&
                        typeof node.value === 'number' &&
                        isTSNumericLiteralType(node)) {
                        return;
                    }
                    // Check if the node is a readonly class property
                    if (typeof node.value === 'number' &&
                        isParentTSReadonlyClassProperty(node)) {
                        if (options.ignoreReadonlyClassProperties) {
                            return;
                        }
                        let fullNumberNode = node;
                        let raw = node.raw;
                        if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                            // the base rule only shows the operator for negative numbers
                            // https://github.com/eslint/eslint/blob/9dfc8501fb1956c90dc11e6377b4cb38a6bea65d/lib/rules/no-magic-numbers.js#L126
                            node.parent.operator === '-') {
                            fullNumberNode = node.parent;
                            raw = `${node.parent.operator}${node.raw}`;
                        }
                        context.report({
                            messageId: 'noMagic',
                            node: fullNumberNode,
                            data: { raw },
                        });
                        return;
                    }
                    // Let the base rule deal with the rest
                    rules.Literal(node);
                },
            };
        },
    });
    /**
     * Gets the true parent of the literal, handling prefixed numbers (-1 / +1)
     */
    function getLiteralParent(node) {
        var _a;
        if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
            ['-', '+'].includes(node.parent.operator)) {
            return node.parent.parent;
        }
        return node.parent;
    }
    /**
     * Checks if the node grandparent is a Typescript type alias declaration
     * @param node the node to be validated.
     * @returns true if the node grandparent is a Typescript type alias declaration
     * @private
     */
    function isGrandparentTSTypeAliasDeclaration(node) {
        var _a, _b;
        return ((_b = (_a = node.parent) === null || _a === void 0 ? void 0 : _a.parent) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.TSTypeAliasDeclaration;
    }
    /**
     * Checks if the node grandparent is a Typescript union type and its parent is a type alias declaration
     * @param node the node to be validated.
     * @returns true if the node grandparent is a Typescript union type and its parent is a type alias declaration
     * @private
     */
    function isGrandparentTSUnionType(node) {
        var _a, _b;
        if (((_b = (_a = node.parent) === null || _a === void 0 ? void 0 : _a.parent) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.TSUnionType) {
            return isGrandparentTSTypeAliasDeclaration(node.parent);
        }
        return false;
    }
    /**
     * Checks if the node parent is a Typescript enum member
     * @param node the node to be validated.
     * @returns true if the node parent is a Typescript enum member
     * @private
     */
    function isParentTSEnumDeclaration(node) {
        const parent = getLiteralParent(node);
        return (parent === null || parent === void 0 ? void 0 : parent.type) === experimental_utils_1.AST_NODE_TYPES.TSEnumMember;
    }
    /**
     * Checks if the node parent is a Typescript literal type
     * @param node the node to be validated.
     * @returns true if the node parent is a Typescript literal type
     * @private
     */
    function isParentTSLiteralType(node) {
        var _a;
        return ((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.TSLiteralType;
    }
    /**
     * Checks if the node is a valid TypeScript numeric literal type.
     * @param node the node to be validated.
     * @returns true if the node is a TypeScript numeric literal type.
     * @private
     */
    function isTSNumericLiteralType(node) {
        var _a;
        // For negative numbers, use the parent node
        if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
            node.parent.operator === '-') {
            node = node.parent;
        }
        // If the parent node is not a TSLiteralType, early return
        if (!isParentTSLiteralType(node)) {
            return false;
        }
        // If the grandparent is a TSTypeAliasDeclaration, ignore
        if (isGrandparentTSTypeAliasDeclaration(node)) {
            return true;
        }
        // If the grandparent is a TSUnionType and it's parent is a TSTypeAliasDeclaration, ignore
        if (isGrandparentTSUnionType(node)) {
            return true;
        }
        return false;
    }
    /**
     * Checks if the node parent is a readonly class property
     * @param node the node to be validated.
     * @returns true if the node parent is a readonly class property
     * @private
     */
    function isParentTSReadonlyClassProperty(node) {
        const parent = getLiteralParent(node);
        if ((parent === null || parent === void 0 ? void 0 : parent.type) === experimental_utils_1.AST_NODE_TYPES.ClassProperty && parent.readonly) {
            return true;
        }
        return false;
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-misused-new.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-misused-new',
        meta: {
            type: 'problem',
            docs: {
                description: 'Enforce valid definition of `new` and `constructor`',
                category: 'Best Practices',
                recommended: 'error',
            },
            schema: [],
            messages: {
                errorMessageInterface: 'Interfaces cannot be constructed, only classes.',
                errorMessageClass: 'Class cannot have method named `new`.',
            },
        },
        defaultOptions: [],
        create(context) {
            /**
             * @param node type to be inspected.
             * @returns name of simple type or null
             */
            function getTypeReferenceName(node) {
                if (node) {
                    switch (node.type) {
                        case experimental_utils_1.AST_NODE_TYPES.TSTypeAnnotation:
                            return getTypeReferenceName(node.typeAnnotation);
                        case experimental_utils_1.AST_NODE_TYPES.TSTypeReference:
                            return getTypeReferenceName(node.typeName);
                        case experimental_utils_1.AST_NODE_TYPES.Identifier:
                            return node.name;
                        default:
                            break;
                    }
                }
                return null;
            }
            /**
             * @param parent parent node.
             * @param returnType type to be compared
             */
            function isMatchingParentType(parent, returnType) {
                if (parent &&
                    'id' in parent &&
                    parent.id &&
                    parent.id.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                    return getTypeReferenceName(returnType) === parent.id.name;
                }
                return false;
            }
            return {
                'TSInterfaceBody > TSConstructSignatureDeclaration'(node) {
                    if (isMatchingParentType(node.parent.parent, node.returnType)) {
                        // constructor
                        context.report({
                            node,
                            messageId: 'errorMessageInterface',
                        });
                    }
                },
                "TSMethodSignature[key.name='constructor']"(node) {
                    context.report({
                        node,
                        messageId: 'errorMessageInterface',
                    });
                },
                "ClassBody > MethodDefinition[key.name='new']"(node) {
                    if (node.value.type === experimental_utils_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression) {
                        if (node.parent &&
                            isMatchingParentType(node.parent.parent, node.value.returnType)) {
                            context.report({
                                node,
                                messageId: 'errorMessageClass',
                            });
                        }
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-misused-promises.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-misused-promises',
        meta: {
            docs: {
                description: 'Avoid using promises in places not designed to handle them',
                category: 'Best Practices',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                voidReturn: 'Promise returned in function argument where a void return was expected.',
                conditional: 'Expected non-Promise value in a boolean conditional.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        checksConditionals: {
                            type: 'boolean',
                        },
                        checksVoidReturn: {
                            type: 'boolean',
                        },
                    },
                },
            ],
            type: 'problem',
        },
        defaultOptions: [
            {
                checksConditionals: true,
                checksVoidReturn: true,
            },
        ],
        create(context, [{ checksConditionals, checksVoidReturn }]) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const checkedNodes = new Set();
            const conditionalChecks = {
                ConditionalExpression: checkTestConditional,
                DoWhileStatement: checkTestConditional,
                ForStatement: checkTestConditional,
                IfStatement: checkTestConditional,
                LogicalExpression: checkConditional,
                'UnaryExpression[operator="!"]'(node) {
                    checkConditional(node.argument, true);
                },
                WhileStatement: checkTestConditional,
            };
            const voidReturnChecks = {
                CallExpression: checkArguments,
                NewExpression: checkArguments,
            };
            function checkTestConditional(node) {
                if (node.test) {
                    checkConditional(node.test, true);
                }
            }
            /**
             * This function analyzes the type of a node and checks if it is a Promise in a boolean conditional.
             * It uses recursion when checking nested logical operators.
             * @param node The AST node to check.
             * @param isTestExpr Whether the node is a descendant of a test expression.
             */
            function checkConditional(node, isTestExpr = false) {
                // prevent checking the same node multiple times
                if (checkedNodes.has(node)) {
                    return;
                }
                checkedNodes.add(node);
                if (node.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression) {
                    // ignore the left operand for nullish coalescing expressions not in a context of a test expression
                    if (node.operator !== '??' || isTestExpr) {
                        checkConditional(node.left, isTestExpr);
                    }
                    // we ignore the right operand when not in a context of a test expression
                    if (isTestExpr) {
                        checkConditional(node.right, isTestExpr);
                    }
                    return;
                }
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                if (isAlwaysThenable(checker, tsNode)) {
                    context.report({
                        messageId: 'conditional',
                        node,
                    });
                }
            }
            function checkArguments(node) {
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                const voidParams = voidFunctionParams(checker, tsNode);
                if (voidParams.size === 0) {
                    return;
                }
                for (const [index, argument] of node.arguments.entries()) {
                    if (!voidParams.has(index)) {
                        continue;
                    }
                    const tsNode = parserServices.esTreeNodeToTSNodeMap.get(argument);
                    if (returnsThenable(checker, tsNode)) {
                        context.report({
                            messageId: 'voidReturn',
                            node: argument,
                        });
                    }
                }
            }
            return Object.assign(Object.assign({}, (checksConditionals ? conditionalChecks : {})), (checksVoidReturn ? voidReturnChecks : {}));
        },
    });
    // Variation on the thenable check which requires all forms of the type (read:
    // alternates in a union) to be thenable. Otherwise, you might be trying to
    // check if something is defined or undefined and get caught because one of the
    // branches is thenable.
    function isAlwaysThenable(checker, node) {
        const type = checker.getTypeAtLocation(node);
        for (const subType of tsutils.unionTypeParts(checker.getApparentType(type))) {
            const thenProp = subType.getProperty('then');
            // If one of the alternates has no then property, it is not thenable in all
            // cases.
            if (thenProp === undefined) {
                return false;
            }
            // We walk through each variation of the then property. Since we know it
            // exists at this point, we just need at least one of the alternates to
            // be of the right form to consider it thenable.
            const thenType = checker.getTypeOfSymbolAtLocation(thenProp, node);
            let hasThenableSignature = false;
            for (const subType of tsutils.unionTypeParts(thenType)) {
                for (const signature of subType.getCallSignatures()) {
                    if (signature.parameters.length !== 0 &&
                        isFunctionParam(checker, signature.parameters[0], node)) {
                        hasThenableSignature = true;
                        break;
                    }
                }
                // We only need to find one variant of the then property that has a
                // function signature for it to be thenable.
                if (hasThenableSignature) {
                    break;
                }
            }
            // If no flavors of the then property are thenable, we don't consider the
            // overall type to be thenable
            if (!hasThenableSignature) {
                return false;
            }
        }
        // If all variants are considered thenable (i.e. haven't returned false), we
        // consider the overall type thenable
        return true;
    }
    function isFunctionParam(checker, param, node) {
        const type = checker.getApparentType(checker.getTypeOfSymbolAtLocation(param, node));
        for (const subType of tsutils.unionTypeParts(type)) {
            if (subType.getCallSignatures().length !== 0) {
                return true;
            }
        }
        return false;
    }
    // Get the positions of parameters which are void functions (and not also
    // thenable functions). These are the candidates for the void-return check at
    // the current call site.
    function voidFunctionParams(checker, node) {
        const voidReturnIndices = new Set();
        const thenableReturnIndices = new Set();
        const type = checker.getTypeAtLocation(node.expression);
        for (const subType of tsutils.unionTypeParts(type)) {
            // Standard function calls and `new` have two different types of signatures
            const signatures = ts.isCallExpression(node)
                ? subType.getCallSignatures()
                : subType.getConstructSignatures();
            for (const signature of signatures) {
                for (const [index, parameter] of signature.parameters.entries()) {
                    const type = checker.getTypeOfSymbolAtLocation(parameter, node.expression);
                    for (const subType of tsutils.unionTypeParts(type)) {
                        for (const signature of subType.getCallSignatures()) {
                            const returnType = signature.getReturnType();
                            if (tsutils.isTypeFlagSet(returnType, ts.TypeFlags.Void)) {
                                voidReturnIndices.add(index);
                            }
                            else if (tsutils.isThenableType(checker, node.expression, returnType)) {
                                thenableReturnIndices.add(index);
                            }
                        }
                    }
                }
            }
        }
        // If a certain positional argument accepts both thenable and void returns,
        // a promise-returning function is valid
        for (const thenable of thenableReturnIndices) {
            voidReturnIndices.delete(thenable);
        }
        return voidReturnIndices;
    }
    // Returns true if the expression is a function that returns a thenable
    function returnsThenable(checker, node) {
        const type = checker.getApparentType(checker.getTypeAtLocation(node));
        for (const subType of tsutils.unionTypeParts(type)) {
            for (const signature of subType.getCallSignatures()) {
                const returnType = signature.getReturnType();
                if (tsutils.isThenableType(checker, node, returnType)) {
                    return true;
                }
            }
        }
        return false;
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-namespace.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-namespace',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow the use of custom TypeScript modules and namespaces',
                category: 'Best Practices',
                recommended: 'error',
            },
            messages: {
                moduleSyntaxIsPreferred: 'ES2015 module syntax is preferred over custom TypeScript modules and namespaces.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowDeclarations: {
                            type: 'boolean',
                        },
                        allowDefinitionFiles: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                allowDeclarations: false,
                allowDefinitionFiles: true,
            },
        ],
        create(context, [{ allowDeclarations, allowDefinitionFiles }]) {
            const filename = context.getFilename();
            function isDeclaration(node) {
                var _a;
                return (node.declare === true ||
                    (((_a = node.parent.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration &&
                        isDeclaration(node.parent.parent)));
            }
            return {
                "TSModuleDeclaration[global!=true][id.type='Identifier']"(node) {
                    if ((node.parent &&
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration) ||
                        (allowDefinitionFiles && util.isDefinitionFile(filename)) ||
                        (allowDeclarations && isDeclaration(node))) {
                        return;
                    }
                    context.report({
                        node,
                        messageId: 'moduleSyntaxIsPreferred',
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-non-null-asserted-optional-chain.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const ts = __importStar(require(21) /* typescript */);
    const semver = __importStar(require(1004) /* semver */);
    const util = __importStar(require(906) /* ../util */);
    const is3dot9 = semver.satisfies(ts.version, `>= 3.9.0 || >= 3.9.1-rc || >= 3.9.0-beta`, {
        includePrerelease: true,
    });
    exports.default = util.createRule({
        name: 'no-non-null-asserted-optional-chain',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows using a non-null assertion after an optional chain expression',
                category: 'Possible Errors',
                recommended: 'error',
                suggestion: true,
            },
            messages: {
                noNonNullOptionalChain: 'Optional chain expressions can return undefined by design - using a non-null assertion is unsafe and wrong.',
                suggestRemovingNonNull: 'You should remove the non-null assertion.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            // TS3.9 made a breaking change to how non-null works with optional chains.
            // Pre-3.9,  `x?.y!.z` means `(x?.y).z` - i.e. it essentially scrubbed the optionality from the chain
            // Post-3.9, `x?.y!.z` means `x?.y!.z`  - i.e. it just asserts that the property `y` is non-null, not the result of `x?.y`.
            // This means that for > 3.9, x?.y!.z is valid!
            //
            // NOTE: these cases are still invalid for 3.9:
            // - x?.y.z!
            // - (x?.y)!.z
            const baseSelectors = {
                // non-nulling a wrapped chain will scrub all nulls introduced by the chain
                // (x?.y)!
                // (x?.())!
                'TSNonNullExpression > ChainExpression'(node) {
                    // selector guarantees this assertion
                    const parent = node.parent;
                    context.report({
                        node,
                        messageId: 'noNonNullOptionalChain',
                        // use a suggestion instead of a fixer, because this can obviously break type checks
                        suggest: [
                            {
                                messageId: 'suggestRemovingNonNull',
                                fix(fixer) {
                                    return fixer.removeRange([
                                        parent.range[1] - 1,
                                        parent.range[1],
                                    ]);
                                },
                            },
                        ],
                    });
                },
                // non-nulling at the end of a chain will scrub all nulls introduced by the chain
                // x?.y!
                // x?.()!
                'ChainExpression > TSNonNullExpression'(node) {
                    context.report({
                        node,
                        messageId: 'noNonNullOptionalChain',
                        // use a suggestion instead of a fixer, because this can obviously break type checks
                        suggest: [
                            {
                                messageId: 'suggestRemovingNonNull',
                                fix(fixer) {
                                    return fixer.removeRange([node.range[1] - 1, node.range[1]]);
                                },
                            },
                        ],
                    });
                },
            };
            if (is3dot9) {
                return baseSelectors;
            }
            return Object.assign(Object.assign({}, baseSelectors), { [[
                    // > :not(ChainExpression) because that case is handled by a previous selector
                    'MemberExpression > TSNonNullExpression.object > :not(ChainExpression)',
                    'CallExpression > TSNonNullExpression.callee > :not(ChainExpression)',
                ].join(', ')](child) {
                    // selector guarantees this assertion
                    const node = child.parent;
                    let current = child;
                    while (current) {
                        switch (current.type) {
                            case experimental_utils_1.AST_NODE_TYPES.MemberExpression:
                                if (current.optional) {
                                    // found an optional chain! stop traversing
                                    break;
                                }
                                current = current.object;
                                continue;
                            case experimental_utils_1.AST_NODE_TYPES.CallExpression:
                                if (current.optional) {
                                    // found an optional chain! stop traversing
                                    break;
                                }
                                current = current.callee;
                                continue;
                            default:
                                // something that's not a ChainElement, which means this is not an optional chain we want to check
                                return;
                        }
                    }
                    context.report({
                        node,
                        messageId: 'noNonNullOptionalChain',
                        // use a suggestion instead of a fixer, because this can obviously break type checks
                        suggest: [
                            {
                                messageId: 'suggestRemovingNonNull',
                                fix(fixer) {
                                    return fixer.removeRange([node.range[1] - 1, node.range[1]]);
                                },
                            },
                        ],
                    });
                } });
        },
    });
},
/* semver/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // just pre-load all the stuff that index.js lazily exports
    const internalRe = require(1005) /* ./internal/re */;
    module.exports = {
        re: internalRe.re,
        src: internalRe.src,
        tokens: internalRe.t,
        SEMVER_SPEC_VERSION: require(1006) /* ./internal/constants */.SEMVER_SPEC_VERSION,
        SemVer: require(1008) /* ./classes/semver */,
        compareIdentifiers: require(1010) /* ./internal/identifiers */.compareIdentifiers,
        rcompareIdentifiers: require(1010) /* ./internal/identifiers */.rcompareIdentifiers,
        parse: require(1011) /* ./functions/parse */,
        valid: require(1012) /* ./functions/valid */,
        clean: require(1013) /* ./functions/clean */,
        inc: require(1014) /* ./functions/inc */,
        diff: require(1015) /* ./functions/diff */,
        major: require(1018) /* ./functions/major */,
        minor: require(1019) /* ./functions/minor */,
        patch: require(1020) /* ./functions/patch */,
        prerelease: require(1021) /* ./functions/prerelease */,
        compare: require(1017) /* ./functions/compare */,
        rcompare: require(1022) /* ./functions/rcompare */,
        compareLoose: require(1023) /* ./functions/compare-loose */,
        compareBuild: require(1024) /* ./functions/compare-build */,
        sort: require(1025) /* ./functions/sort */,
        rsort: require(1026) /* ./functions/rsort */,
        gt: require(1027) /* ./functions/gt */,
        lt: require(1028) /* ./functions/lt */,
        eq: require(1016) /* ./functions/eq */,
        neq: require(1029) /* ./functions/neq */,
        gte: require(1030) /* ./functions/gte */,
        lte: require(1031) /* ./functions/lte */,
        cmp: require(1032) /* ./functions/cmp */,
        coerce: require(1033) /* ./functions/coerce */,
        Comparator: require(1034) /* ./classes/comparator */,
        Range: require(1035) /* ./classes/range */,
        satisfies: require(1039) /* ./functions/satisfies */,
        toComparators: require(1040) /* ./ranges/to-comparators */,
        maxSatisfying: require(1041) /* ./ranges/max-satisfying */,
        minSatisfying: require(1042) /* ./ranges/min-satisfying */,
        minVersion: require(1043) /* ./ranges/min-version */,
        validRange: require(1044) /* ./ranges/valid */,
        outside: require(1045) /* ./ranges/outside */,
        gtr: require(1046) /* ./ranges/gtr */,
        ltr: require(1047) /* ./ranges/ltr */,
        intersects: require(1048) /* ./ranges/intersects */,
        simplifyRange: require(1049) /* ./ranges/simplify */,
        subset: require(1050) /* ./ranges/subset */,
    };
},
/* semver/internal/re.js */ function _(require, module, exports, __esModule, __esExport) {
    const { MAX_SAFE_COMPONENT_LENGTH } = require(1006) /* ./constants */;
    const debug = require(1007) /* ./debug */;
    exports = module.exports = {};
    // The actual regexps go on exports.re
    const re = exports.re = [];
    const src = exports.src = [];
    const t = exports.t = {};
    let R = 0;
    const createToken = (name, value, isGlobal) => {
        const index = R++;
        debug(index, value);
        t[name] = index;
        src[index] = value;
        re[index] = new RegExp(value, isGlobal ? 'g' : undefined);
    };
    // The following Regular Expressions can be used for tokenizing,
    // validating, and parsing SemVer version strings.
    // ## Numeric Identifier
    // A single `0`, or a non-zero digit followed by zero or more digits.
    createToken('NUMERICIDENTIFIER', '0|[1-9]\\d*');
    createToken('NUMERICIDENTIFIERLOOSE', '[0-9]+');
    // ## Non-numeric Identifier
    // Zero or more digits, followed by a letter or hyphen, and then zero or
    // more letters, digits, or hyphens.
    createToken('NONNUMERICIDENTIFIER', '\\d*[a-zA-Z-][a-zA-Z0-9-]*');
    // ## Main Version
    // Three dot-separated numeric identifiers.
    createToken('MAINVERSION', `(${src[t.NUMERICIDENTIFIER]})\\.` +
        `(${src[t.NUMERICIDENTIFIER]})\\.` +
        `(${src[t.NUMERICIDENTIFIER]})`);
    createToken('MAINVERSIONLOOSE', `(${src[t.NUMERICIDENTIFIERLOOSE]})\\.` +
        `(${src[t.NUMERICIDENTIFIERLOOSE]})\\.` +
        `(${src[t.NUMERICIDENTIFIERLOOSE]})`);
    // ## Pre-release Version Identifier
    // A numeric identifier, or a non-numeric identifier.
    createToken('PRERELEASEIDENTIFIER', `(?:${src[t.NUMERICIDENTIFIER]}|${src[t.NONNUMERICIDENTIFIER]})`);
    createToken('PRERELEASEIDENTIFIERLOOSE', `(?:${src[t.NUMERICIDENTIFIERLOOSE]}|${src[t.NONNUMERICIDENTIFIER]})`);
    // ## Pre-release Version
    // Hyphen, followed by one or more dot-separated pre-release version
    // identifiers.
    createToken('PRERELEASE', `(?:-(${src[t.PRERELEASEIDENTIFIER]}(?:\\.${src[t.PRERELEASEIDENTIFIER]})*))`);
    createToken('PRERELEASELOOSE', `(?:-?(${src[t.PRERELEASEIDENTIFIERLOOSE]}(?:\\.${src[t.PRERELEASEIDENTIFIERLOOSE]})*))`);
    // ## Build Metadata Identifier
    // Any combination of digits, letters, or hyphens.
    createToken('BUILDIDENTIFIER', '[0-9A-Za-z-]+');
    // ## Build Metadata
    // Plus sign, followed by one or more period-separated build metadata
    // identifiers.
    createToken('BUILD', `(?:\\+(${src[t.BUILDIDENTIFIER]}(?:\\.${src[t.BUILDIDENTIFIER]})*))`);
    // ## Full Version String
    // A main version, followed optionally by a pre-release version and
    // build metadata.
    // Note that the only major, minor, patch, and pre-release sections of
    // the version string are capturing groups.  The build metadata is not a
    // capturing group, because it should not ever be used in version
    // comparison.
    createToken('FULLPLAIN', `v?${src[t.MAINVERSION]}${src[t.PRERELEASE]}?${src[t.BUILD]}?`);
    createToken('FULL', `^${src[t.FULLPLAIN]}$`);
    // like full, but allows v1.2.3 and =1.2.3, which people do sometimes.
    // also, 1.0.0alpha1 (prerelease without the hyphen) which is pretty
    // common in the npm registry.
    createToken('LOOSEPLAIN', `[v=\\s]*${src[t.MAINVERSIONLOOSE]}${src[t.PRERELEASELOOSE]}?${src[t.BUILD]}?`);
    createToken('LOOSE', `^${src[t.LOOSEPLAIN]}$`);
    createToken('GTLT', '((?:<|>)?=?)');
    // Something like "2.*" or "1.2.x".
    // Note that "x.x" is a valid xRange identifer, meaning "any version"
    // Only the first item is strictly required.
    createToken('XRANGEIDENTIFIERLOOSE', `${src[t.NUMERICIDENTIFIERLOOSE]}|x|X|\\*`);
    createToken('XRANGEIDENTIFIER', `${src[t.NUMERICIDENTIFIER]}|x|X|\\*`);
    createToken('XRANGEPLAIN', `[v=\\s]*(${src[t.XRANGEIDENTIFIER]})` +
        `(?:\\.(${src[t.XRANGEIDENTIFIER]})` +
        `(?:\\.(${src[t.XRANGEIDENTIFIER]})` +
        `(?:${src[t.PRERELEASE]})?${src[t.BUILD]}?` +
        `)?)?`);
    createToken('XRANGEPLAINLOOSE', `[v=\\s]*(${src[t.XRANGEIDENTIFIERLOOSE]})` +
        `(?:\\.(${src[t.XRANGEIDENTIFIERLOOSE]})` +
        `(?:\\.(${src[t.XRANGEIDENTIFIERLOOSE]})` +
        `(?:${src[t.PRERELEASELOOSE]})?${src[t.BUILD]}?` +
        `)?)?`);
    createToken('XRANGE', `^${src[t.GTLT]}\\s*${src[t.XRANGEPLAIN]}$`);
    createToken('XRANGELOOSE', `^${src[t.GTLT]}\\s*${src[t.XRANGEPLAINLOOSE]}$`);
    // Coercion.
    // Extract anything that could conceivably be a part of a valid semver
    createToken('COERCE', `${'(^|[^\\d])' +
        '(\\d{1,'}${MAX_SAFE_COMPONENT_LENGTH}})` +
        `(?:\\.(\\d{1,${MAX_SAFE_COMPONENT_LENGTH}}))?` +
        `(?:\\.(\\d{1,${MAX_SAFE_COMPONENT_LENGTH}}))?` +
        `(?:$|[^\\d])`);
    createToken('COERCERTL', src[t.COERCE], true);
    // Tilde ranges.
    // Meaning is "reasonably at or greater than"
    createToken('LONETILDE', '(?:~>?)');
    createToken('TILDETRIM', `(\\s*)${src[t.LONETILDE]}\\s+`, true);
    exports.tildeTrimReplace = '$1~';
    createToken('TILDE', `^${src[t.LONETILDE]}${src[t.XRANGEPLAIN]}$`);
    createToken('TILDELOOSE', `^${src[t.LONETILDE]}${src[t.XRANGEPLAINLOOSE]}$`);
    // Caret ranges.
    // Meaning is "at least and backwards compatible with"
    createToken('LONECARET', '(?:\\^)');
    createToken('CARETTRIM', `(\\s*)${src[t.LONECARET]}\\s+`, true);
    exports.caretTrimReplace = '$1^';
    createToken('CARET', `^${src[t.LONECARET]}${src[t.XRANGEPLAIN]}$`);
    createToken('CARETLOOSE', `^${src[t.LONECARET]}${src[t.XRANGEPLAINLOOSE]}$`);
    // A simple gt/lt/eq thing, or just "" to indicate "any version"
    createToken('COMPARATORLOOSE', `^${src[t.GTLT]}\\s*(${src[t.LOOSEPLAIN]})$|^$`);
    createToken('COMPARATOR', `^${src[t.GTLT]}\\s*(${src[t.FULLPLAIN]})$|^$`);
    // An expression to strip any whitespace between the gtlt and the thing
    // it modifies, so that `> 1.2.3` ==> `>1.2.3`
    createToken('COMPARATORTRIM', `(\\s*)${src[t.GTLT]}\\s*(${src[t.LOOSEPLAIN]}|${src[t.XRANGEPLAIN]})`, true);
    exports.comparatorTrimReplace = '$1$2$3';
    // Something like `1.2.3 - 1.2.4`
    // Note that these all use the loose form, because they'll be
    // checked against either the strict or loose comparator form
    // later.
    createToken('HYPHENRANGE', `^\\s*(${src[t.XRANGEPLAIN]})` +
        `\\s+-\\s+` +
        `(${src[t.XRANGEPLAIN]})` +
        `\\s*$`);
    createToken('HYPHENRANGELOOSE', `^\\s*(${src[t.XRANGEPLAINLOOSE]})` +
        `\\s+-\\s+` +
        `(${src[t.XRANGEPLAINLOOSE]})` +
        `\\s*$`);
    // Star ranges basically just allow anything at all.
    createToken('STAR', '(<|>)?=?\\s*\\*');
    // >=0.0.0 is like a star
    createToken('GTE0', '^\\s*>=\\s*0\.0\.0\\s*$');
    createToken('GTE0PRE', '^\\s*>=\\s*0\.0\.0-0\\s*$');
},
/* semver/internal/constants.js */ function _(require, module, exports, __esModule, __esExport) {
    // Note: this is the semver.org version of the spec that it implements
    // Not necessarily the package version of this code.
    const SEMVER_SPEC_VERSION = '2.0.0';
    const MAX_LENGTH = 256;
    const MAX_SAFE_INTEGER = Number.MAX_SAFE_INTEGER ||
        /* istanbul ignore next */ 9007199254740991;
    // Max safe segment length for coercion.
    const MAX_SAFE_COMPONENT_LENGTH = 16;
    module.exports = {
        SEMVER_SPEC_VERSION,
        MAX_LENGTH,
        MAX_SAFE_INTEGER,
        MAX_SAFE_COMPONENT_LENGTH
    };
},
/* semver/internal/debug.js */ function _(require, module, exports, __esModule, __esExport) {
    const debug = (typeof process === 'object' &&
        process.env &&
        process.env.NODE_DEBUG &&
        /\bsemver\b/i.test(process.env.NODE_DEBUG)) ? (...args) => console.error('SEMVER', ...args)
        : () => { };
    module.exports = debug;
},
/* semver/classes/semver.js */ function _(require, module, exports, __esModule, __esExport) {
    const debug = require(1007) /* ../internal/debug */;
    const { MAX_LENGTH, MAX_SAFE_INTEGER } = require(1006) /* ../internal/constants */;
    const { re, t } = require(1005) /* ../internal/re */;
    const parseOptions = require(1009) /* ../internal/parse-options */;
    const { compareIdentifiers } = require(1010) /* ../internal/identifiers */;
    class SemVer {
        constructor(version, options) {
            options = parseOptions(options);
            if (version instanceof SemVer) {
                if (version.loose === !!options.loose &&
                    version.includePrerelease === !!options.includePrerelease) {
                    return version;
                }
                else {
                    version = version.version;
                }
            }
            else if (typeof version !== 'string') {
                throw new TypeError(`Invalid Version: ${version}`);
            }
            if (version.length > MAX_LENGTH) {
                throw new TypeError(`version is longer than ${MAX_LENGTH} characters`);
            }
            debug('SemVer', version, options);
            this.options = options;
            this.loose = !!options.loose;
            // this isn't actually relevant for versions, but keep it so that we
            // don't run into trouble passing this.options around.
            this.includePrerelease = !!options.includePrerelease;
            const m = version.trim().match(options.loose ? re[t.LOOSE] : re[t.FULL]);
            if (!m) {
                throw new TypeError(`Invalid Version: ${version}`);
            }
            this.raw = version;
            // these are actually numbers
            this.major = +m[1];
            this.minor = +m[2];
            this.patch = +m[3];
            if (this.major > MAX_SAFE_INTEGER || this.major < 0) {
                throw new TypeError('Invalid major version');
            }
            if (this.minor > MAX_SAFE_INTEGER || this.minor < 0) {
                throw new TypeError('Invalid minor version');
            }
            if (this.patch > MAX_SAFE_INTEGER || this.patch < 0) {
                throw new TypeError('Invalid patch version');
            }
            // numberify any prerelease numeric ids
            if (!m[4]) {
                this.prerelease = [];
            }
            else {
                this.prerelease = m[4].split('.').map((id) => {
                    if (/^[0-9]+$/.test(id)) {
                        const num = +id;
                        if (num >= 0 && num < MAX_SAFE_INTEGER) {
                            return num;
                        }
                    }
                    return id;
                });
            }
            this.build = m[5] ? m[5].split('.') : [];
            this.format();
        }
        format() {
            this.version = `${this.major}.${this.minor}.${this.patch}`;
            if (this.prerelease.length) {
                this.version += `-${this.prerelease.join('.')}`;
            }
            return this.version;
        }
        toString() {
            return this.version;
        }
        compare(other) {
            debug('SemVer.compare', this.version, this.options, other);
            if (!(other instanceof SemVer)) {
                if (typeof other === 'string' && other === this.version) {
                    return 0;
                }
                other = new SemVer(other, this.options);
            }
            if (other.version === this.version) {
                return 0;
            }
            return this.compareMain(other) || this.comparePre(other);
        }
        compareMain(other) {
            if (!(other instanceof SemVer)) {
                other = new SemVer(other, this.options);
            }
            return (compareIdentifiers(this.major, other.major) ||
                compareIdentifiers(this.minor, other.minor) ||
                compareIdentifiers(this.patch, other.patch));
        }
        comparePre(other) {
            if (!(other instanceof SemVer)) {
                other = new SemVer(other, this.options);
            }
            // NOT having a prerelease is > having one
            if (this.prerelease.length && !other.prerelease.length) {
                return -1;
            }
            else if (!this.prerelease.length && other.prerelease.length) {
                return 1;
            }
            else if (!this.prerelease.length && !other.prerelease.length) {
                return 0;
            }
            let i = 0;
            do {
                const a = this.prerelease[i];
                const b = other.prerelease[i];
                debug('prerelease compare', i, a, b);
                if (a === undefined && b === undefined) {
                    return 0;
                }
                else if (b === undefined) {
                    return 1;
                }
                else if (a === undefined) {
                    return -1;
                }
                else if (a === b) {
                    continue;
                }
                else {
                    return compareIdentifiers(a, b);
                }
            } while (++i);
        }
        compareBuild(other) {
            if (!(other instanceof SemVer)) {
                other = new SemVer(other, this.options);
            }
            let i = 0;
            do {
                const a = this.build[i];
                const b = other.build[i];
                debug('prerelease compare', i, a, b);
                if (a === undefined && b === undefined) {
                    return 0;
                }
                else if (b === undefined) {
                    return 1;
                }
                else if (a === undefined) {
                    return -1;
                }
                else if (a === b) {
                    continue;
                }
                else {
                    return compareIdentifiers(a, b);
                }
            } while (++i);
        }
        // preminor will bump the version up to the next minor release, and immediately
        // down to pre-release. premajor and prepatch work the same way.
        inc(release, identifier) {
            switch (release) {
                case 'premajor':
                    this.prerelease.length = 0;
                    this.patch = 0;
                    this.minor = 0;
                    this.major++;
                    this.inc('pre', identifier);
                    break;
                case 'preminor':
                    this.prerelease.length = 0;
                    this.patch = 0;
                    this.minor++;
                    this.inc('pre', identifier);
                    break;
                case 'prepatch':
                    // If this is already a prerelease, it will bump to the next version
                    // drop any prereleases that might already exist, since they are not
                    // relevant at this point.
                    this.prerelease.length = 0;
                    this.inc('patch', identifier);
                    this.inc('pre', identifier);
                    break;
                // If the input is a non-prerelease version, this acts the same as
                // prepatch.
                case 'prerelease':
                    if (this.prerelease.length === 0) {
                        this.inc('patch', identifier);
                    }
                    this.inc('pre', identifier);
                    break;
                case 'major':
                    // If this is a pre-major version, bump up to the same major version.
                    // Otherwise increment major.
                    // 1.0.0-5 bumps to 1.0.0
                    // 1.1.0 bumps to 2.0.0
                    if (this.minor !== 0 ||
                        this.patch !== 0 ||
                        this.prerelease.length === 0) {
                        this.major++;
                    }
                    this.minor = 0;
                    this.patch = 0;
                    this.prerelease = [];
                    break;
                case 'minor':
                    // If this is a pre-minor version, bump up to the same minor version.
                    // Otherwise increment minor.
                    // 1.2.0-5 bumps to 1.2.0
                    // 1.2.1 bumps to 1.3.0
                    if (this.patch !== 0 || this.prerelease.length === 0) {
                        this.minor++;
                    }
                    this.patch = 0;
                    this.prerelease = [];
                    break;
                case 'patch':
                    // If this is not a pre-release version, it will increment the patch.
                    // If it is a pre-release it will bump up to the same patch version.
                    // 1.2.0-5 patches to 1.2.0
                    // 1.2.0 patches to 1.2.1
                    if (this.prerelease.length === 0) {
                        this.patch++;
                    }
                    this.prerelease = [];
                    break;
                // This probably shouldn't be used publicly.
                // 1.0.0 'pre' would become 1.0.0-0 which is the wrong direction.
                case 'pre':
                    if (this.prerelease.length === 0) {
                        this.prerelease = [0];
                    }
                    else {
                        let i = this.prerelease.length;
                        while (--i >= 0) {
                            if (typeof this.prerelease[i] === 'number') {
                                this.prerelease[i]++;
                                i = -2;
                            }
                        }
                        if (i === -1) {
                            // didn't increment anything
                            this.prerelease.push(0);
                        }
                    }
                    if (identifier) {
                        // 1.2.0-beta.1 bumps to 1.2.0-beta.2,
                        // 1.2.0-beta.fooblz or 1.2.0-beta bumps to 1.2.0-beta.0
                        if (this.prerelease[0] === identifier) {
                            if (isNaN(this.prerelease[1])) {
                                this.prerelease = [identifier, 0];
                            }
                        }
                        else {
                            this.prerelease = [identifier, 0];
                        }
                    }
                    break;
                default:
                    throw new Error(`invalid increment argument: ${release}`);
            }
            this.format();
            this.raw = this.version;
            return this;
        }
    }
    module.exports = SemVer;
},
/* semver/internal/parse-options.js */ function _(require, module, exports, __esModule, __esExport) {
    // parse out just the options we care about so we always get a consistent
    // obj with keys in a consistent order.
    const opts = ['includePrerelease', 'loose', 'rtl'];
    const parseOptions = options => !options ? {}
        : typeof options !== 'object' ? { loose: true }
            : opts.filter(k => options[k]).reduce((options, k) => {
                options[k] = true;
                return options;
            }, {});
    module.exports = parseOptions;
},
/* semver/internal/identifiers.js */ function _(require, module, exports, __esModule, __esExport) {
    const numeric = /^[0-9]+$/;
    const compareIdentifiers = (a, b) => {
        const anum = numeric.test(a);
        const bnum = numeric.test(b);
        if (anum && bnum) {
            a = +a;
            b = +b;
        }
        return a === b ? 0
            : (anum && !bnum) ? -1
                : (bnum && !anum) ? 1
                    : a < b ? -1
                        : 1;
    };
    const rcompareIdentifiers = (a, b) => compareIdentifiers(b, a);
    module.exports = {
        compareIdentifiers,
        rcompareIdentifiers
    };
},
/* semver/functions/parse.js */ function _(require, module, exports, __esModule, __esExport) {
    const { MAX_LENGTH } = require(1006) /* ../internal/constants */;
    const { re, t } = require(1005) /* ../internal/re */;
    const SemVer = require(1008) /* ../classes/semver */;
    const parseOptions = require(1009) /* ../internal/parse-options */;
    const parse = (version, options) => {
        options = parseOptions(options);
        if (version instanceof SemVer) {
            return version;
        }
        if (typeof version !== 'string') {
            return null;
        }
        if (version.length > MAX_LENGTH) {
            return null;
        }
        const r = options.loose ? re[t.LOOSE] : re[t.FULL];
        if (!r.test(version)) {
            return null;
        }
        try {
            return new SemVer(version, options);
        }
        catch (er) {
            return null;
        }
    };
    module.exports = parse;
},
/* semver/functions/valid.js */ function _(require, module, exports, __esModule, __esExport) {
    const parse = require(1011) /* ./parse */;
    const valid = (version, options) => {
        const v = parse(version, options);
        return v ? v.version : null;
    };
    module.exports = valid;
},
/* semver/functions/clean.js */ function _(require, module, exports, __esModule, __esExport) {
    const parse = require(1011) /* ./parse */;
    const clean = (version, options) => {
        const s = parse(version.trim().replace(/^[=v]+/, ''), options);
        return s ? s.version : null;
    };
    module.exports = clean;
},
/* semver/functions/inc.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const inc = (version, release, options, identifier) => {
        if (typeof (options) === 'string') {
            identifier = options;
            options = undefined;
        }
        try {
            return new SemVer(version, options).inc(release, identifier).version;
        }
        catch (er) {
            return null;
        }
    };
    module.exports = inc;
},
/* semver/functions/diff.js */ function _(require, module, exports, __esModule, __esExport) {
    const parse = require(1011) /* ./parse */;
    const eq = require(1016) /* ./eq */;
    const diff = (version1, version2) => {
        if (eq(version1, version2)) {
            return null;
        }
        else {
            const v1 = parse(version1);
            const v2 = parse(version2);
            const hasPre = v1.prerelease.length || v2.prerelease.length;
            const prefix = hasPre ? 'pre' : '';
            const defaultResult = hasPre ? 'prerelease' : '';
            for (const key in v1) {
                if (key === 'major' || key === 'minor' || key === 'patch') {
                    if (v1[key] !== v2[key]) {
                        return prefix + key;
                    }
                }
            }
            return defaultResult; // may be undefined
        }
    };
    module.exports = diff;
},
/* semver/functions/eq.js */ function _(require, module, exports, __esModule, __esExport) {
    const compare = require(1017) /* ./compare */;
    const eq = (a, b, loose) => compare(a, b, loose) === 0;
    module.exports = eq;
},
/* semver/functions/compare.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const compare = (a, b, loose) => new SemVer(a, loose).compare(new SemVer(b, loose));
    module.exports = compare;
},
/* semver/functions/major.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const major = (a, loose) => new SemVer(a, loose).major;
    module.exports = major;
},
/* semver/functions/minor.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const minor = (a, loose) => new SemVer(a, loose).minor;
    module.exports = minor;
},
/* semver/functions/patch.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const patch = (a, loose) => new SemVer(a, loose).patch;
    module.exports = patch;
},
/* semver/functions/prerelease.js */ function _(require, module, exports, __esModule, __esExport) {
    const parse = require(1011) /* ./parse */;
    const prerelease = (version, options) => {
        const parsed = parse(version, options);
        return (parsed && parsed.prerelease.length) ? parsed.prerelease : null;
    };
    module.exports = prerelease;
},
/* semver/functions/rcompare.js */ function _(require, module, exports, __esModule, __esExport) {
    const compare = require(1017) /* ./compare */;
    const rcompare = (a, b, loose) => compare(b, a, loose);
    module.exports = rcompare;
},
/* semver/functions/compare-loose.js */ function _(require, module, exports, __esModule, __esExport) {
    const compare = require(1017) /* ./compare */;
    const compareLoose = (a, b) => compare(a, b, true);
    module.exports = compareLoose;
},
/* semver/functions/compare-build.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const compareBuild = (a, b, loose) => {
        const versionA = new SemVer(a, loose);
        const versionB = new SemVer(b, loose);
        return versionA.compare(versionB) || versionA.compareBuild(versionB);
    };
    module.exports = compareBuild;
},
/* semver/functions/sort.js */ function _(require, module, exports, __esModule, __esExport) {
    const compareBuild = require(1024) /* ./compare-build */;
    const sort = (list, loose) => list.sort((a, b) => compareBuild(a, b, loose));
    module.exports = sort;
},
/* semver/functions/rsort.js */ function _(require, module, exports, __esModule, __esExport) {
    const compareBuild = require(1024) /* ./compare-build */;
    const rsort = (list, loose) => list.sort((a, b) => compareBuild(b, a, loose));
    module.exports = rsort;
},
/* semver/functions/gt.js */ function _(require, module, exports, __esModule, __esExport) {
    const compare = require(1017) /* ./compare */;
    const gt = (a, b, loose) => compare(a, b, loose) > 0;
    module.exports = gt;
},
/* semver/functions/lt.js */ function _(require, module, exports, __esModule, __esExport) {
    const compare = require(1017) /* ./compare */;
    const lt = (a, b, loose) => compare(a, b, loose) < 0;
    module.exports = lt;
},
/* semver/functions/neq.js */ function _(require, module, exports, __esModule, __esExport) {
    const compare = require(1017) /* ./compare */;
    const neq = (a, b, loose) => compare(a, b, loose) !== 0;
    module.exports = neq;
},
/* semver/functions/gte.js */ function _(require, module, exports, __esModule, __esExport) {
    const compare = require(1017) /* ./compare */;
    const gte = (a, b, loose) => compare(a, b, loose) >= 0;
    module.exports = gte;
},
/* semver/functions/lte.js */ function _(require, module, exports, __esModule, __esExport) {
    const compare = require(1017) /* ./compare */;
    const lte = (a, b, loose) => compare(a, b, loose) <= 0;
    module.exports = lte;
},
/* semver/functions/cmp.js */ function _(require, module, exports, __esModule, __esExport) {
    const eq = require(1016) /* ./eq */;
    const neq = require(1029) /* ./neq */;
    const gt = require(1027) /* ./gt */;
    const gte = require(1030) /* ./gte */;
    const lt = require(1028) /* ./lt */;
    const lte = require(1031) /* ./lte */;
    const cmp = (a, op, b, loose) => {
        switch (op) {
            case '===':
                if (typeof a === 'object')
                    a = a.version;
                if (typeof b === 'object')
                    b = b.version;
                return a === b;
            case '!==':
                if (typeof a === 'object')
                    a = a.version;
                if (typeof b === 'object')
                    b = b.version;
                return a !== b;
            case '':
            case '=':
            case '==':
                return eq(a, b, loose);
            case '!=':
                return neq(a, b, loose);
            case '>':
                return gt(a, b, loose);
            case '>=':
                return gte(a, b, loose);
            case '<':
                return lt(a, b, loose);
            case '<=':
                return lte(a, b, loose);
            default:
                throw new TypeError(`Invalid operator: ${op}`);
        }
    };
    module.exports = cmp;
},
/* semver/functions/coerce.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const parse = require(1011) /* ./parse */;
    const { re, t } = require(1005) /* ../internal/re */;
    const coerce = (version, options) => {
        if (version instanceof SemVer) {
            return version;
        }
        if (typeof version === 'number') {
            version = String(version);
        }
        if (typeof version !== 'string') {
            return null;
        }
        options = options || {};
        let match = null;
        if (!options.rtl) {
            match = version.match(re[t.COERCE]);
        }
        else {
            // Find the right-most coercible string that does not share
            // a terminus with a more left-ward coercible string.
            // Eg, '1.2.3.4' wants to coerce '2.3.4', not '3.4' or '4'
            //
            // Walk through the string checking with a /g regexp
            // Manually set the index so as to pick up overlapping matches.
            // Stop when we get a match that ends at the string end, since no
            // coercible string can be more right-ward without the same terminus.
            let next;
            while ((next = re[t.COERCERTL].exec(version)) &&
                (!match || match.index + match[0].length !== version.length)) {
                if (!match ||
                    next.index + next[0].length !== match.index + match[0].length) {
                    match = next;
                }
                re[t.COERCERTL].lastIndex = next.index + next[1].length + next[2].length;
            }
            // leave it in a clean state
            re[t.COERCERTL].lastIndex = -1;
        }
        if (match === null)
            return null;
        return parse(`${match[2]}.${match[3] || '0'}.${match[4] || '0'}`, options);
    };
    module.exports = coerce;
},
/* semver/classes/comparator.js */ function _(require, module, exports, __esModule, __esExport) {
    const ANY = Symbol('SemVer ANY');
    // hoisted class for cyclic dependency
    class Comparator {
        static get ANY() {
            return ANY;
        }
        constructor(comp, options) {
            options = parseOptions(options);
            if (comp instanceof Comparator) {
                if (comp.loose === !!options.loose) {
                    return comp;
                }
                else {
                    comp = comp.value;
                }
            }
            debug('comparator', comp, options);
            this.options = options;
            this.loose = !!options.loose;
            this.parse(comp);
            if (this.semver === ANY) {
                this.value = '';
            }
            else {
                this.value = this.operator + this.semver.version;
            }
            debug('comp', this);
        }
        parse(comp) {
            const r = this.options.loose ? re[t.COMPARATORLOOSE] : re[t.COMPARATOR];
            const m = comp.match(r);
            if (!m) {
                throw new TypeError(`Invalid comparator: ${comp}`);
            }
            this.operator = m[1] !== undefined ? m[1] : '';
            if (this.operator === '=') {
                this.operator = '';
            }
            // if it literally is just '>' or '' then allow anything.
            if (!m[2]) {
                this.semver = ANY;
            }
            else {
                this.semver = new SemVer(m[2], this.options.loose);
            }
        }
        toString() {
            return this.value;
        }
        test(version) {
            debug('Comparator.test', version, this.options.loose);
            if (this.semver === ANY || version === ANY) {
                return true;
            }
            if (typeof version === 'string') {
                try {
                    version = new SemVer(version, this.options);
                }
                catch (er) {
                    return false;
                }
            }
            return cmp(version, this.operator, this.semver, this.options);
        }
        intersects(comp, options) {
            if (!(comp instanceof Comparator)) {
                throw new TypeError('a Comparator is required');
            }
            if (!options || typeof options !== 'object') {
                options = {
                    loose: !!options,
                    includePrerelease: false
                };
            }
            if (this.operator === '') {
                if (this.value === '') {
                    return true;
                }
                return new Range(comp.value, options).test(this.value);
            }
            else if (comp.operator === '') {
                if (comp.value === '') {
                    return true;
                }
                return new Range(this.value, options).test(comp.semver);
            }
            const sameDirectionIncreasing = (this.operator === '>=' || this.operator === '>') &&
                (comp.operator === '>=' || comp.operator === '>');
            const sameDirectionDecreasing = (this.operator === '<=' || this.operator === '<') &&
                (comp.operator === '<=' || comp.operator === '<');
            const sameSemVer = this.semver.version === comp.semver.version;
            const differentDirectionsInclusive = (this.operator === '>=' || this.operator === '<=') &&
                (comp.operator === '>=' || comp.operator === '<=');
            const oppositeDirectionsLessThan = cmp(this.semver, '<', comp.semver, options) &&
                (this.operator === '>=' || this.operator === '>') &&
                (comp.operator === '<=' || comp.operator === '<');
            const oppositeDirectionsGreaterThan = cmp(this.semver, '>', comp.semver, options) &&
                (this.operator === '<=' || this.operator === '<') &&
                (comp.operator === '>=' || comp.operator === '>');
            return (sameDirectionIncreasing ||
                sameDirectionDecreasing ||
                (sameSemVer && differentDirectionsInclusive) ||
                oppositeDirectionsLessThan ||
                oppositeDirectionsGreaterThan);
        }
    }
    module.exports = Comparator;
    const parseOptions = require(1009) /* ../internal/parse-options */;
    const { re, t } = require(1005) /* ../internal/re */;
    const cmp = require(1032) /* ../functions/cmp */;
    const debug = require(1007) /* ../internal/debug */;
    const SemVer = require(1008) /* ./semver */;
    const Range = require(1035) /* ./range */;
},
/* semver/classes/range.js */ function _(require, module, exports, __esModule, __esExport) {
    // hoisted class for cyclic dependency
    class Range {
        constructor(range, options) {
            options = parseOptions(options);
            if (range instanceof Range) {
                if (range.loose === !!options.loose &&
                    range.includePrerelease === !!options.includePrerelease) {
                    return range;
                }
                else {
                    return new Range(range.raw, options);
                }
            }
            if (range instanceof Comparator) {
                // just put it in the set and return
                this.raw = range.value;
                this.set = [[range]];
                this.format();
                return this;
            }
            this.options = options;
            this.loose = !!options.loose;
            this.includePrerelease = !!options.includePrerelease;
            // First, split based on boolean or ||
            this.raw = range;
            this.set = range
                .split(/\s*\|\|\s*/)
                // map the range to a 2d array of comparators
                .map(range => this.parseRange(range.trim()))
                // throw out any comparator lists that are empty
                // this generally means that it was not a valid range, which is allowed
                // in loose mode, but will still throw if the WHOLE range is invalid.
                .filter(c => c.length);
            if (!this.set.length) {
                throw new TypeError(`Invalid SemVer Range: ${range}`);
            }
            // if we have any that are not the null set, throw out null sets.
            if (this.set.length > 1) {
                // keep the first one, in case they're all null sets
                const first = this.set[0];
                this.set = this.set.filter(c => !isNullSet(c[0]));
                if (this.set.length === 0)
                    this.set = [first];
                else if (this.set.length > 1) {
                    // if we have any that are *, then the range is just *
                    for (const c of this.set) {
                        if (c.length === 1 && isAny(c[0])) {
                            this.set = [c];
                            break;
                        }
                    }
                }
            }
            this.format();
        }
        format() {
            this.range = this.set
                .map((comps) => {
                return comps.join(' ').trim();
            })
                .join('||')
                .trim();
            return this.range;
        }
        toString() {
            return this.range;
        }
        parseRange(range) {
            range = range.trim();
            // memoize range parsing for performance.
            // this is a very hot path, and fully deterministic.
            const memoOpts = Object.keys(this.options).join(',');
            const memoKey = `parseRange:${memoOpts}:${range}`;
            const cached = cache.get(memoKey);
            if (cached)
                return cached;
            const loose = this.options.loose;
            // `1.2.3 - 1.2.4` => `>=1.2.3 <=1.2.4`
            const hr = loose ? re[t.HYPHENRANGELOOSE] : re[t.HYPHENRANGE];
            range = range.replace(hr, hyphenReplace(this.options.includePrerelease));
            debug('hyphen replace', range);
            // `> 1.2.3 < 1.2.5` => `>1.2.3 <1.2.5`
            range = range.replace(re[t.COMPARATORTRIM], comparatorTrimReplace);
            debug('comparator trim', range, re[t.COMPARATORTRIM]);
            // `~ 1.2.3` => `~1.2.3`
            range = range.replace(re[t.TILDETRIM], tildeTrimReplace);
            // `^ 1.2.3` => `^1.2.3`
            range = range.replace(re[t.CARETTRIM], caretTrimReplace);
            // normalize spaces
            range = range.split(/\s+/).join(' ');
            // At this point, the range is completely trimmed and
            // ready to be split into comparators.
            const compRe = loose ? re[t.COMPARATORLOOSE] : re[t.COMPARATOR];
            const rangeList = range
                .split(' ')
                .map(comp => parseComparator(comp, this.options))
                .join(' ')
                .split(/\s+/)
                // >=0.0.0 is equivalent to *
                .map(comp => replaceGTE0(comp, this.options))
                // in loose mode, throw out any that are not valid comparators
                .filter(this.options.loose ? comp => !!comp.match(compRe) : () => true)
                .map(comp => new Comparator(comp, this.options));
            // if any comparators are the null set, then replace with JUST null set
            // if more than one comparator, remove any * comparators
            // also, don't include the same comparator more than once
            const l = rangeList.length;
            const rangeMap = new Map();
            for (const comp of rangeList) {
                if (isNullSet(comp))
                    return [comp];
                rangeMap.set(comp.value, comp);
            }
            if (rangeMap.size > 1 && rangeMap.has(''))
                rangeMap.delete('');
            const result = [...rangeMap.values()];
            cache.set(memoKey, result);
            return result;
        }
        intersects(range, options) {
            if (!(range instanceof Range)) {
                throw new TypeError('a Range is required');
            }
            return this.set.some((thisComparators) => {
                return (isSatisfiable(thisComparators, options) &&
                    range.set.some((rangeComparators) => {
                        return (isSatisfiable(rangeComparators, options) &&
                            thisComparators.every((thisComparator) => {
                                return rangeComparators.every((rangeComparator) => {
                                    return thisComparator.intersects(rangeComparator, options);
                                });
                            }));
                    }));
            });
        }
        // if ANY of the sets match ALL of its comparators, then pass
        test(version) {
            if (!version) {
                return false;
            }
            if (typeof version === 'string') {
                try {
                    version = new SemVer(version, this.options);
                }
                catch (er) {
                    return false;
                }
            }
            for (let i = 0; i < this.set.length; i++) {
                if (testSet(this.set[i], version, this.options)) {
                    return true;
                }
            }
            return false;
        }
    }
    module.exports = Range;
    const LRU = require(1036) /* lru-cache */;
    const cache = new LRU({ max: 1000 });
    const parseOptions = require(1009) /* ../internal/parse-options */;
    const Comparator = require(1034) /* ./comparator */;
    const debug = require(1007) /* ../internal/debug */;
    const SemVer = require(1008) /* ./semver */;
    const { re, t, comparatorTrimReplace, tildeTrimReplace, caretTrimReplace } = require(1005) /* ../internal/re */;
    const isNullSet = c => c.value === '<0.0.0-0';
    const isAny = c => c.value === '';
    // take a set of comparators and determine whether there
    // exists a version which can satisfy it
    const isSatisfiable = (comparators, options) => {
        let result = true;
        const remainingComparators = comparators.slice();
        let testComparator = remainingComparators.pop();
        while (result && remainingComparators.length) {
            result = remainingComparators.every((otherComparator) => {
                return testComparator.intersects(otherComparator, options);
            });
            testComparator = remainingComparators.pop();
        }
        return result;
    };
    // comprised of xranges, tildes, stars, and gtlt's at this point.
    // already replaced the hyphen ranges
    // turn into a set of JUST comparators.
    const parseComparator = (comp, options) => {
        debug('comp', comp, options);
        comp = replaceCarets(comp, options);
        debug('caret', comp);
        comp = replaceTildes(comp, options);
        debug('tildes', comp);
        comp = replaceXRanges(comp, options);
        debug('xrange', comp);
        comp = replaceStars(comp, options);
        debug('stars', comp);
        return comp;
    };
    const isX = id => !id || id.toLowerCase() === 'x' || id === '*';
    // ~, ~> --> * (any, kinda silly)
    // ~2, ~2.x, ~2.x.x, ~>2, ~>2.x ~>2.x.x --> >=2.0.0 <3.0.0-0
    // ~2.0, ~2.0.x, ~>2.0, ~>2.0.x --> >=2.0.0 <2.1.0-0
    // ~1.2, ~1.2.x, ~>1.2, ~>1.2.x --> >=1.2.0 <1.3.0-0
    // ~1.2.3, ~>1.2.3 --> >=1.2.3 <1.3.0-0
    // ~1.2.0, ~>1.2.0 --> >=1.2.0 <1.3.0-0
    const replaceTildes = (comp, options) => comp.trim().split(/\s+/).map((comp) => {
        return replaceTilde(comp, options);
    }).join(' ');
    const replaceTilde = (comp, options) => {
        const r = options.loose ? re[t.TILDELOOSE] : re[t.TILDE];
        return comp.replace(r, (_, M, m, p, pr) => {
            debug('tilde', comp, _, M, m, p, pr);
            let ret;
            if (isX(M)) {
                ret = '';
            }
            else if (isX(m)) {
                ret = `>=${M}.0.0 <${+M + 1}.0.0-0`;
            }
            else if (isX(p)) {
                // ~1.2 == >=1.2.0 <1.3.0-0
                ret = `>=${M}.${m}.0 <${M}.${+m + 1}.0-0`;
            }
            else if (pr) {
                debug('replaceTilde pr', pr);
                ret = `>=${M}.${m}.${p}-${pr} <${M}.${+m + 1}.0-0`;
            }
            else {
                // ~1.2.3 == >=1.2.3 <1.3.0-0
                ret = `>=${M}.${m}.${p} <${M}.${+m + 1}.0-0`;
            }
            debug('tilde return', ret);
            return ret;
        });
    };
    // ^ --> * (any, kinda silly)
    // ^2, ^2.x, ^2.x.x --> >=2.0.0 <3.0.0-0
    // ^2.0, ^2.0.x --> >=2.0.0 <3.0.0-0
    // ^1.2, ^1.2.x --> >=1.2.0 <2.0.0-0
    // ^1.2.3 --> >=1.2.3 <2.0.0-0
    // ^1.2.0 --> >=1.2.0 <2.0.0-0
    const replaceCarets = (comp, options) => comp.trim().split(/\s+/).map((comp) => {
        return replaceCaret(comp, options);
    }).join(' ');
    const replaceCaret = (comp, options) => {
        debug('caret', comp, options);
        const r = options.loose ? re[t.CARETLOOSE] : re[t.CARET];
        const z = options.includePrerelease ? '-0' : '';
        return comp.replace(r, (_, M, m, p, pr) => {
            debug('caret', comp, _, M, m, p, pr);
            let ret;
            if (isX(M)) {
                ret = '';
            }
            else if (isX(m)) {
                ret = `>=${M}.0.0${z} <${+M + 1}.0.0-0`;
            }
            else if (isX(p)) {
                if (M === '0') {
                    ret = `>=${M}.${m}.0${z} <${M}.${+m + 1}.0-0`;
                }
                else {
                    ret = `>=${M}.${m}.0${z} <${+M + 1}.0.0-0`;
                }
            }
            else if (pr) {
                debug('replaceCaret pr', pr);
                if (M === '0') {
                    if (m === '0') {
                        ret = `>=${M}.${m}.${p}-${pr} <${M}.${m}.${+p + 1}-0`;
                    }
                    else {
                        ret = `>=${M}.${m}.${p}-${pr} <${M}.${+m + 1}.0-0`;
                    }
                }
                else {
                    ret = `>=${M}.${m}.${p}-${pr} <${+M + 1}.0.0-0`;
                }
            }
            else {
                debug('no pr');
                if (M === '0') {
                    if (m === '0') {
                        ret = `>=${M}.${m}.${p}${z} <${M}.${m}.${+p + 1}-0`;
                    }
                    else {
                        ret = `>=${M}.${m}.${p}${z} <${M}.${+m + 1}.0-0`;
                    }
                }
                else {
                    ret = `>=${M}.${m}.${p} <${+M + 1}.0.0-0`;
                }
            }
            debug('caret return', ret);
            return ret;
        });
    };
    const replaceXRanges = (comp, options) => {
        debug('replaceXRanges', comp, options);
        return comp.split(/\s+/).map((comp) => {
            return replaceXRange(comp, options);
        }).join(' ');
    };
    const replaceXRange = (comp, options) => {
        comp = comp.trim();
        const r = options.loose ? re[t.XRANGELOOSE] : re[t.XRANGE];
        return comp.replace(r, (ret, gtlt, M, m, p, pr) => {
            debug('xRange', comp, ret, gtlt, M, m, p, pr);
            const xM = isX(M);
            const xm = xM || isX(m);
            const xp = xm || isX(p);
            const anyX = xp;
            if (gtlt === '=' && anyX) {
                gtlt = '';
            }
            // if we're including prereleases in the match, then we need
            // to fix this to -0, the lowest possible prerelease value
            pr = options.includePrerelease ? '-0' : '';
            if (xM) {
                if (gtlt === '>' || gtlt === '<') {
                    // nothing is allowed
                    ret = '<0.0.0-0';
                }
                else {
                    // nothing is forbidden
                    ret = '*';
                }
            }
            else if (gtlt && anyX) {
                // we know patch is an x, because we have any x at all.
                // replace X with 0
                if (xm) {
                    m = 0;
                }
                p = 0;
                if (gtlt === '>') {
                    // >1 => >=2.0.0
                    // >1.2 => >=1.3.0
                    gtlt = '>=';
                    if (xm) {
                        M = +M + 1;
                        m = 0;
                        p = 0;
                    }
                    else {
                        m = +m + 1;
                        p = 0;
                    }
                }
                else if (gtlt === '<=') {
                    // <=0.7.x is actually <0.8.0, since any 0.7.x should
                    // pass.  Similarly, <=7.x is actually <8.0.0, etc.
                    gtlt = '<';
                    if (xm) {
                        M = +M + 1;
                    }
                    else {
                        m = +m + 1;
                    }
                }
                if (gtlt === '<')
                    pr = '-0';
                ret = `${gtlt + M}.${m}.${p}${pr}`;
            }
            else if (xm) {
                ret = `>=${M}.0.0${pr} <${+M + 1}.0.0-0`;
            }
            else if (xp) {
                ret = `>=${M}.${m}.0${pr} <${M}.${+m + 1}.0-0`;
            }
            debug('xRange return', ret);
            return ret;
        });
    };
    // Because * is AND-ed with everything else in the comparator,
    // and '' means "any version", just remove the *s entirely.
    const replaceStars = (comp, options) => {
        debug('replaceStars', comp, options);
        // Looseness is ignored here.  star is always as loose as it gets!
        return comp.trim().replace(re[t.STAR], '');
    };
    const replaceGTE0 = (comp, options) => {
        debug('replaceGTE0', comp, options);
        return comp.trim()
            .replace(re[options.includePrerelease ? t.GTE0PRE : t.GTE0], '');
    };
    // This function is passed to string.replace(re[t.HYPHENRANGE])
    // M, m, patch, prerelease, build
    // 1.2 - 3.4.5 => >=1.2.0 <=3.4.5
    // 1.2.3 - 3.4 => >=1.2.0 <3.5.0-0 Any 3.4.x will do
    // 1.2 - 3.4 => >=1.2.0 <3.5.0-0
    const hyphenReplace = incPr => ($0, from, fM, fm, fp, fpr, fb, to, tM, tm, tp, tpr, tb) => {
        if (isX(fM)) {
            from = '';
        }
        else if (isX(fm)) {
            from = `>=${fM}.0.0${incPr ? '-0' : ''}`;
        }
        else if (isX(fp)) {
            from = `>=${fM}.${fm}.0${incPr ? '-0' : ''}`;
        }
        else if (fpr) {
            from = `>=${from}`;
        }
        else {
            from = `>=${from}${incPr ? '-0' : ''}`;
        }
        if (isX(tM)) {
            to = '';
        }
        else if (isX(tm)) {
            to = `<${+tM + 1}.0.0-0`;
        }
        else if (isX(tp)) {
            to = `<${tM}.${+tm + 1}.0-0`;
        }
        else if (tpr) {
            to = `<=${tM}.${tm}.${tp}-${tpr}`;
        }
        else if (incPr) {
            to = `<${tM}.${tm}.${+tp + 1}-0`;
        }
        else {
            to = `<=${to}`;
        }
        return (`${from} ${to}`).trim();
    };
    const testSet = (set, version, options) => {
        for (let i = 0; i < set.length; i++) {
            if (!set[i].test(version)) {
                return false;
            }
        }
        if (version.prerelease.length && !options.includePrerelease) {
            // Find the set of versions that are allowed to have prereleases
            // For example, ^1.2.3-pr.1 desugars to >=1.2.3-pr.1 <2.0.0
            // That should allow `1.2.3-pr.2` to pass.
            // However, `1.2.4-alpha.notready` should NOT be allowed,
            // even though it's within the range set by the comparators.
            for (let i = 0; i < set.length; i++) {
                debug(set[i].semver);
                if (set[i].semver === Comparator.ANY) {
                    continue;
                }
                if (set[i].semver.prerelease.length > 0) {
                    const allowed = set[i].semver;
                    if (allowed.major === version.major &&
                        allowed.minor === version.minor &&
                        allowed.patch === version.patch) {
                        return true;
                    }
                }
            }
            // Version has a -pre, but it's not one of the ones we like.
            return false;
        }
        return true;
    };
},
/* lru-cache/index.js */ function _(require, module, exports, __esModule, __esExport) {
    // A linked list to keep track of recently-used-ness
    const Yallist = require(1037) /* yallist */;
    const MAX = Symbol('max');
    const LENGTH = Symbol('length');
    const LENGTH_CALCULATOR = Symbol('lengthCalculator');
    const ALLOW_STALE = Symbol('allowStale');
    const MAX_AGE = Symbol('maxAge');
    const DISPOSE = Symbol('dispose');
    const NO_DISPOSE_ON_SET = Symbol('noDisposeOnSet');
    const LRU_LIST = Symbol('lruList');
    const CACHE = Symbol('cache');
    const UPDATE_AGE_ON_GET = Symbol('updateAgeOnGet');
    const naiveLength = () => 1;
    // lruList is a yallist where the head is the youngest
    // item, and the tail is the oldest.  the list contains the Hit
    // objects as the entries.
    // Each Hit object has a reference to its Yallist.Node.  This
    // never changes.
    //
    // cache is a Map (or PseudoMap) that matches the keys to
    // the Yallist.Node object.
    class LRUCache {
        constructor(options) {
            if (typeof options === 'number')
                options = { max: options };
            if (!options)
                options = {};
            if (options.max && (typeof options.max !== 'number' || options.max < 0))
                throw new TypeError('max must be a non-negative number');
            // Kind of weird to have a default max of Infinity, but oh well.
            const max = this[MAX] = options.max || Infinity;
            const lc = options.length || naiveLength;
            this[LENGTH_CALCULATOR] = (typeof lc !== 'function') ? naiveLength : lc;
            this[ALLOW_STALE] = options.stale || false;
            if (options.maxAge && typeof options.maxAge !== 'number')
                throw new TypeError('maxAge must be a number');
            this[MAX_AGE] = options.maxAge || 0;
            this[DISPOSE] = options.dispose;
            this[NO_DISPOSE_ON_SET] = options.noDisposeOnSet || false;
            this[UPDATE_AGE_ON_GET] = options.updateAgeOnGet || false;
            this.reset();
        }
        // resize the cache when the max changes.
        set max(mL) {
            if (typeof mL !== 'number' || mL < 0)
                throw new TypeError('max must be a non-negative number');
            this[MAX] = mL || Infinity;
            trim(this);
        }
        get max() {
            return this[MAX];
        }
        set allowStale(allowStale) {
            this[ALLOW_STALE] = !!allowStale;
        }
        get allowStale() {
            return this[ALLOW_STALE];
        }
        set maxAge(mA) {
            if (typeof mA !== 'number')
                throw new TypeError('maxAge must be a non-negative number');
            this[MAX_AGE] = mA;
            trim(this);
        }
        get maxAge() {
            return this[MAX_AGE];
        }
        // resize the cache when the lengthCalculator changes.
        set lengthCalculator(lC) {
            if (typeof lC !== 'function')
                lC = naiveLength;
            if (lC !== this[LENGTH_CALCULATOR]) {
                this[LENGTH_CALCULATOR] = lC;
                this[LENGTH] = 0;
                this[LRU_LIST].forEach(hit => {
                    hit.length = this[LENGTH_CALCULATOR](hit.value, hit.key);
                    this[LENGTH] += hit.length;
                });
            }
            trim(this);
        }
        get lengthCalculator() { return this[LENGTH_CALCULATOR]; }
        get length() { return this[LENGTH]; }
        get itemCount() { return this[LRU_LIST].length; }
        rforEach(fn, thisp) {
            thisp = thisp || this;
            for (let walker = this[LRU_LIST].tail; walker !== null;) {
                const prev = walker.prev;
                forEachStep(this, fn, walker, thisp);
                walker = prev;
            }
        }
        forEach(fn, thisp) {
            thisp = thisp || this;
            for (let walker = this[LRU_LIST].head; walker !== null;) {
                const next = walker.next;
                forEachStep(this, fn, walker, thisp);
                walker = next;
            }
        }
        keys() {
            return this[LRU_LIST].toArray().map(k => k.key);
        }
        values() {
            return this[LRU_LIST].toArray().map(k => k.value);
        }
        reset() {
            if (this[DISPOSE] &&
                this[LRU_LIST] &&
                this[LRU_LIST].length) {
                this[LRU_LIST].forEach(hit => this[DISPOSE](hit.key, hit.value));
            }
            this[CACHE] = new Map(); // hash of items by key
            this[LRU_LIST] = new Yallist(); // list of items in order of use recency
            this[LENGTH] = 0; // length of items in the list
        }
        dump() {
            return this[LRU_LIST].map(hit => isStale(this, hit) ? false : {
                k: hit.key,
                v: hit.value,
                e: hit.now + (hit.maxAge || 0)
            }).toArray().filter(h => h);
        }
        dumpLru() {
            return this[LRU_LIST];
        }
        set(key, value, maxAge) {
            maxAge = maxAge || this[MAX_AGE];
            if (maxAge && typeof maxAge !== 'number')
                throw new TypeError('maxAge must be a number');
            const now = maxAge ? Date.now() : 0;
            const len = this[LENGTH_CALCULATOR](value, key);
            if (this[CACHE].has(key)) {
                if (len > this[MAX]) {
                    del(this, this[CACHE].get(key));
                    return false;
                }
                const node = this[CACHE].get(key);
                const item = node.value;
                // dispose of the old one before overwriting
                // split out into 2 ifs for better coverage tracking
                if (this[DISPOSE]) {
                    if (!this[NO_DISPOSE_ON_SET])
                        this[DISPOSE](key, item.value);
                }
                item.now = now;
                item.maxAge = maxAge;
                item.value = value;
                this[LENGTH] += len - item.length;
                item.length = len;
                this.get(key);
                trim(this);
                return true;
            }
            const hit = new Entry(key, value, len, now, maxAge);
            // oversized objects fall out of cache automatically.
            if (hit.length > this[MAX]) {
                if (this[DISPOSE])
                    this[DISPOSE](key, value);
                return false;
            }
            this[LENGTH] += hit.length;
            this[LRU_LIST].unshift(hit);
            this[CACHE].set(key, this[LRU_LIST].head);
            trim(this);
            return true;
        }
        has(key) {
            if (!this[CACHE].has(key))
                return false;
            const hit = this[CACHE].get(key).value;
            return !isStale(this, hit);
        }
        get(key) {
            return get(this, key, true);
        }
        peek(key) {
            return get(this, key, false);
        }
        pop() {
            const node = this[LRU_LIST].tail;
            if (!node)
                return null;
            del(this, node);
            return node.value;
        }
        del(key) {
            del(this, this[CACHE].get(key));
        }
        load(arr) {
            // reset the cache
            this.reset();
            const now = Date.now();
            // A previous serialized cache has the most recent items first
            for (let l = arr.length - 1; l >= 0; l--) {
                const hit = arr[l];
                const expiresAt = hit.e || 0;
                if (expiresAt === 0)
                    // the item was created without expiration in a non aged cache
                    this.set(hit.k, hit.v);
                else {
                    const maxAge = expiresAt - now;
                    // dont add already expired items
                    if (maxAge > 0) {
                        this.set(hit.k, hit.v, maxAge);
                    }
                }
            }
        }
        prune() {
            this[CACHE].forEach((value, key) => get(this, key, false));
        }
    }
    const get = (self, key, doUse) => {
        const node = self[CACHE].get(key);
        if (node) {
            const hit = node.value;
            if (isStale(self, hit)) {
                del(self, node);
                if (!self[ALLOW_STALE])
                    return undefined;
            }
            else {
                if (doUse) {
                    if (self[UPDATE_AGE_ON_GET])
                        node.value.now = Date.now();
                    self[LRU_LIST].unshiftNode(node);
                }
            }
            return hit.value;
        }
    };
    const isStale = (self, hit) => {
        if (!hit || (!hit.maxAge && !self[MAX_AGE]))
            return false;
        const diff = Date.now() - hit.now;
        return hit.maxAge ? diff > hit.maxAge
            : self[MAX_AGE] && (diff > self[MAX_AGE]);
    };
    const trim = self => {
        if (self[LENGTH] > self[MAX]) {
            for (let walker = self[LRU_LIST].tail; self[LENGTH] > self[MAX] && walker !== null;) {
                // We know that we're about to delete this one, and also
                // what the next least recently used key will be, so just
                // go ahead and set it now.
                const prev = walker.prev;
                del(self, walker);
                walker = prev;
            }
        }
    };
    const del = (self, node) => {
        if (node) {
            const hit = node.value;
            if (self[DISPOSE])
                self[DISPOSE](hit.key, hit.value);
            self[LENGTH] -= hit.length;
            self[CACHE].delete(hit.key);
            self[LRU_LIST].removeNode(node);
        }
    };
    class Entry {
        constructor(key, value, length, now, maxAge) {
            this.key = key;
            this.value = value;
            this.length = length;
            this.now = now;
            this.maxAge = maxAge || 0;
        }
    }
    const forEachStep = (self, fn, node, thisp) => {
        let hit = node.value;
        if (isStale(self, hit)) {
            del(self, node);
            if (!self[ALLOW_STALE])
                hit = undefined;
        }
        if (hit)
            fn.call(thisp, hit.value, hit.key, self);
    };
    module.exports = LRUCache;
},
/* yallist/yallist.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = Yallist;
    Yallist.Node = Node;
    Yallist.create = Yallist;
    function Yallist(list) {
        var self = this;
        if (!(self instanceof Yallist)) {
            self = new Yallist();
        }
        self.tail = null;
        self.head = null;
        self.length = 0;
        if (list && typeof list.forEach === 'function') {
            list.forEach(function (item) {
                self.push(item);
            });
        }
        else if (arguments.length > 0) {
            for (var i = 0, l = arguments.length; i < l; i++) {
                self.push(arguments[i]);
            }
        }
        return self;
    }
    Yallist.prototype.removeNode = function (node) {
        if (node.list !== this) {
            throw new Error('removing node which does not belong to this list');
        }
        var next = node.next;
        var prev = node.prev;
        if (next) {
            next.prev = prev;
        }
        if (prev) {
            prev.next = next;
        }
        if (node === this.head) {
            this.head = next;
        }
        if (node === this.tail) {
            this.tail = prev;
        }
        node.list.length--;
        node.next = null;
        node.prev = null;
        node.list = null;
        return next;
    };
    Yallist.prototype.unshiftNode = function (node) {
        if (node === this.head) {
            return;
        }
        if (node.list) {
            node.list.removeNode(node);
        }
        var head = this.head;
        node.list = this;
        node.next = head;
        if (head) {
            head.prev = node;
        }
        this.head = node;
        if (!this.tail) {
            this.tail = node;
        }
        this.length++;
    };
    Yallist.prototype.pushNode = function (node) {
        if (node === this.tail) {
            return;
        }
        if (node.list) {
            node.list.removeNode(node);
        }
        var tail = this.tail;
        node.list = this;
        node.prev = tail;
        if (tail) {
            tail.next = node;
        }
        this.tail = node;
        if (!this.head) {
            this.head = node;
        }
        this.length++;
    };
    Yallist.prototype.push = function () {
        for (var i = 0, l = arguments.length; i < l; i++) {
            push(this, arguments[i]);
        }
        return this.length;
    };
    Yallist.prototype.unshift = function () {
        for (var i = 0, l = arguments.length; i < l; i++) {
            unshift(this, arguments[i]);
        }
        return this.length;
    };
    Yallist.prototype.pop = function () {
        if (!this.tail) {
            return undefined;
        }
        var res = this.tail.value;
        this.tail = this.tail.prev;
        if (this.tail) {
            this.tail.next = null;
        }
        else {
            this.head = null;
        }
        this.length--;
        return res;
    };
    Yallist.prototype.shift = function () {
        if (!this.head) {
            return undefined;
        }
        var res = this.head.value;
        this.head = this.head.next;
        if (this.head) {
            this.head.prev = null;
        }
        else {
            this.tail = null;
        }
        this.length--;
        return res;
    };
    Yallist.prototype.forEach = function (fn, thisp) {
        thisp = thisp || this;
        for (var walker = this.head, i = 0; walker !== null; i++) {
            fn.call(thisp, walker.value, i, this);
            walker = walker.next;
        }
    };
    Yallist.prototype.forEachReverse = function (fn, thisp) {
        thisp = thisp || this;
        for (var walker = this.tail, i = this.length - 1; walker !== null; i--) {
            fn.call(thisp, walker.value, i, this);
            walker = walker.prev;
        }
    };
    Yallist.prototype.get = function (n) {
        for (var i = 0, walker = this.head; walker !== null && i < n; i++) {
            // abort out of the list early if we hit a cycle
            walker = walker.next;
        }
        if (i === n && walker !== null) {
            return walker.value;
        }
    };
    Yallist.prototype.getReverse = function (n) {
        for (var i = 0, walker = this.tail; walker !== null && i < n; i++) {
            // abort out of the list early if we hit a cycle
            walker = walker.prev;
        }
        if (i === n && walker !== null) {
            return walker.value;
        }
    };
    Yallist.prototype.map = function (fn, thisp) {
        thisp = thisp || this;
        var res = new Yallist();
        for (var walker = this.head; walker !== null;) {
            res.push(fn.call(thisp, walker.value, this));
            walker = walker.next;
        }
        return res;
    };
    Yallist.prototype.mapReverse = function (fn, thisp) {
        thisp = thisp || this;
        var res = new Yallist();
        for (var walker = this.tail; walker !== null;) {
            res.push(fn.call(thisp, walker.value, this));
            walker = walker.prev;
        }
        return res;
    };
    Yallist.prototype.reduce = function (fn, initial) {
        var acc;
        var walker = this.head;
        if (arguments.length > 1) {
            acc = initial;
        }
        else if (this.head) {
            walker = this.head.next;
            acc = this.head.value;
        }
        else {
            throw new TypeError('Reduce of empty list with no initial value');
        }
        for (var i = 0; walker !== null; i++) {
            acc = fn(acc, walker.value, i);
            walker = walker.next;
        }
        return acc;
    };
    Yallist.prototype.reduceReverse = function (fn, initial) {
        var acc;
        var walker = this.tail;
        if (arguments.length > 1) {
            acc = initial;
        }
        else if (this.tail) {
            walker = this.tail.prev;
            acc = this.tail.value;
        }
        else {
            throw new TypeError('Reduce of empty list with no initial value');
        }
        for (var i = this.length - 1; walker !== null; i--) {
            acc = fn(acc, walker.value, i);
            walker = walker.prev;
        }
        return acc;
    };
    Yallist.prototype.toArray = function () {
        var arr = new Array(this.length);
        for (var i = 0, walker = this.head; walker !== null; i++) {
            arr[i] = walker.value;
            walker = walker.next;
        }
        return arr;
    };
    Yallist.prototype.toArrayReverse = function () {
        var arr = new Array(this.length);
        for (var i = 0, walker = this.tail; walker !== null; i++) {
            arr[i] = walker.value;
            walker = walker.prev;
        }
        return arr;
    };
    Yallist.prototype.slice = function (from, to) {
        to = to || this.length;
        if (to < 0) {
            to += this.length;
        }
        from = from || 0;
        if (from < 0) {
            from += this.length;
        }
        var ret = new Yallist();
        if (to < from || to < 0) {
            return ret;
        }
        if (from < 0) {
            from = 0;
        }
        if (to > this.length) {
            to = this.length;
        }
        for (var i = 0, walker = this.head; walker !== null && i < from; i++) {
            walker = walker.next;
        }
        for (; walker !== null && i < to; i++, walker = walker.next) {
            ret.push(walker.value);
        }
        return ret;
    };
    Yallist.prototype.sliceReverse = function (from, to) {
        to = to || this.length;
        if (to < 0) {
            to += this.length;
        }
        from = from || 0;
        if (from < 0) {
            from += this.length;
        }
        var ret = new Yallist();
        if (to < from || to < 0) {
            return ret;
        }
        if (from < 0) {
            from = 0;
        }
        if (to > this.length) {
            to = this.length;
        }
        for (var i = this.length, walker = this.tail; walker !== null && i > to; i--) {
            walker = walker.prev;
        }
        for (; walker !== null && i > from; i--, walker = walker.prev) {
            ret.push(walker.value);
        }
        return ret;
    };
    Yallist.prototype.splice = function (start, deleteCount, ...nodes) {
        if (start > this.length) {
            start = this.length - 1;
        }
        if (start < 0) {
            start = this.length + start;
        }
        for (var i = 0, walker = this.head; walker !== null && i < start; i++) {
            walker = walker.next;
        }
        var ret = [];
        for (var i = 0; walker && i < deleteCount; i++) {
            ret.push(walker.value);
            walker = this.removeNode(walker);
        }
        if (walker === null) {
            walker = this.tail;
        }
        if (walker !== this.head && walker !== this.tail) {
            walker = walker.prev;
        }
        for (var i = 0; i < nodes.length; i++) {
            walker = insert(this, walker, nodes[i]);
        }
        return ret;
    };
    Yallist.prototype.reverse = function () {
        var head = this.head;
        var tail = this.tail;
        for (var walker = head; walker !== null; walker = walker.prev) {
            var p = walker.prev;
            walker.prev = walker.next;
            walker.next = p;
        }
        this.head = tail;
        this.tail = head;
        return this;
    };
    function insert(self, node, value) {
        var inserted = node === self.head ?
            new Node(value, null, node, self) :
            new Node(value, node, node.next, self);
        if (inserted.next === null) {
            self.tail = inserted;
        }
        if (inserted.prev === null) {
            self.head = inserted;
        }
        self.length++;
        return inserted;
    }
    function push(self, item) {
        self.tail = new Node(item, self.tail, null, self);
        if (!self.head) {
            self.head = self.tail;
        }
        self.length++;
    }
    function unshift(self, item) {
        self.head = new Node(item, null, self.head, self);
        if (!self.tail) {
            self.tail = self.head;
        }
        self.length++;
    }
    function Node(value, prev, next, list) {
        if (!(this instanceof Node)) {
            return new Node(value, prev, next, list);
        }
        this.list = list;
        this.value = value;
        if (prev) {
            prev.next = this;
            this.prev = prev;
        }
        else {
            this.prev = null;
        }
        if (next) {
            next.prev = this;
            this.next = next;
        }
        else {
            this.next = null;
        }
    }
    try {
        // add if support for Symbol.iterator is present
        require(1038) /* ./iterator.js */(Yallist);
    }
    catch (er) { }
},
/* yallist/iterator.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function (Yallist) {
        Yallist.prototype[Symbol.iterator] = function* () {
            for (let walker = this.head; walker; walker = walker.next) {
                yield walker.value;
            }
        };
    };
},
/* semver/functions/satisfies.js */ function _(require, module, exports, __esModule, __esExport) {
    const Range = require(1035) /* ../classes/range */;
    const satisfies = (version, range, options) => {
        try {
            range = new Range(range, options);
        }
        catch (er) {
            return false;
        }
        return range.test(version);
    };
    module.exports = satisfies;
},
/* semver/ranges/to-comparators.js */ function _(require, module, exports, __esModule, __esExport) {
    const Range = require(1035) /* ../classes/range */;
    // Mostly just for testing and legacy API reasons
    const toComparators = (range, options) => new Range(range, options).set
        .map(comp => comp.map(c => c.value).join(' ').trim().split(' '));
    module.exports = toComparators;
},
/* semver/ranges/max-satisfying.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const Range = require(1035) /* ../classes/range */;
    const maxSatisfying = (versions, range, options) => {
        let max = null;
        let maxSV = null;
        let rangeObj = null;
        try {
            rangeObj = new Range(range, options);
        }
        catch (er) {
            return null;
        }
        versions.forEach((v) => {
            if (rangeObj.test(v)) {
                // satisfies(v, range, options)
                if (!max || maxSV.compare(v) === -1) {
                    // compare(max, v, true)
                    max = v;
                    maxSV = new SemVer(max, options);
                }
            }
        });
        return max;
    };
    module.exports = maxSatisfying;
},
/* semver/ranges/min-satisfying.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const Range = require(1035) /* ../classes/range */;
    const minSatisfying = (versions, range, options) => {
        let min = null;
        let minSV = null;
        let rangeObj = null;
        try {
            rangeObj = new Range(range, options);
        }
        catch (er) {
            return null;
        }
        versions.forEach((v) => {
            if (rangeObj.test(v)) {
                // satisfies(v, range, options)
                if (!min || minSV.compare(v) === 1) {
                    // compare(min, v, true)
                    min = v;
                    minSV = new SemVer(min, options);
                }
            }
        });
        return min;
    };
    module.exports = minSatisfying;
},
/* semver/ranges/min-version.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const Range = require(1035) /* ../classes/range */;
    const gt = require(1027) /* ../functions/gt */;
    const minVersion = (range, loose) => {
        range = new Range(range, loose);
        let minver = new SemVer('0.0.0');
        if (range.test(minver)) {
            return minver;
        }
        minver = new SemVer('0.0.0-0');
        if (range.test(minver)) {
            return minver;
        }
        minver = null;
        for (let i = 0; i < range.set.length; ++i) {
            const comparators = range.set[i];
            let setMin = null;
            comparators.forEach((comparator) => {
                // Clone to avoid manipulating the comparator's semver object.
                const compver = new SemVer(comparator.semver.version);
                switch (comparator.operator) {
                    case '>':
                        if (compver.prerelease.length === 0) {
                            compver.patch++;
                        }
                        else {
                            compver.prerelease.push(0);
                        }
                        compver.raw = compver.format();
                    /* fallthrough */
                    case '':
                    case '>=':
                        if (!setMin || gt(compver, setMin)) {
                            setMin = compver;
                        }
                        break;
                    case '<':
                    case '<=':
                        /* Ignore maximum versions */
                        break;
                    /* istanbul ignore next */
                    default:
                        throw new Error(`Unexpected operation: ${comparator.operator}`);
                }
            });
            if (setMin && (!minver || gt(minver, setMin)))
                minver = setMin;
        }
        if (minver && range.test(minver)) {
            return minver;
        }
        return null;
    };
    module.exports = minVersion;
},
/* semver/ranges/valid.js */ function _(require, module, exports, __esModule, __esExport) {
    const Range = require(1035) /* ../classes/range */;
    const validRange = (range, options) => {
        try {
            // Return '*' instead of '' so that truthiness works.
            // This will throw if it's invalid anyway
            return new Range(range, options).range || '*';
        }
        catch (er) {
            return null;
        }
    };
    module.exports = validRange;
},
/* semver/ranges/outside.js */ function _(require, module, exports, __esModule, __esExport) {
    const SemVer = require(1008) /* ../classes/semver */;
    const Comparator = require(1034) /* ../classes/comparator */;
    const { ANY } = Comparator;
    const Range = require(1035) /* ../classes/range */;
    const satisfies = require(1039) /* ../functions/satisfies */;
    const gt = require(1027) /* ../functions/gt */;
    const lt = require(1028) /* ../functions/lt */;
    const lte = require(1031) /* ../functions/lte */;
    const gte = require(1030) /* ../functions/gte */;
    const outside = (version, range, hilo, options) => {
        version = new SemVer(version, options);
        range = new Range(range, options);
        let gtfn, ltefn, ltfn, comp, ecomp;
        switch (hilo) {
            case '>':
                gtfn = gt;
                ltefn = lte;
                ltfn = lt;
                comp = '>';
                ecomp = '>=';
                break;
            case '<':
                gtfn = lt;
                ltefn = gte;
                ltfn = gt;
                comp = '<';
                ecomp = '<=';
                break;
            default:
                throw new TypeError('Must provide a hilo val of "<" or ">"');
        }
        // If it satisfies the range it is not outside
        if (satisfies(version, range, options)) {
            return false;
        }
        // From now on, variable terms are as if we're in "gtr" mode.
        // but note that everything is flipped for the "ltr" function.
        for (let i = 0; i < range.set.length; ++i) {
            const comparators = range.set[i];
            let high = null;
            let low = null;
            comparators.forEach((comparator) => {
                if (comparator.semver === ANY) {
                    comparator = new Comparator('>=0.0.0');
                }
                high = high || comparator;
                low = low || comparator;
                if (gtfn(comparator.semver, high.semver, options)) {
                    high = comparator;
                }
                else if (ltfn(comparator.semver, low.semver, options)) {
                    low = comparator;
                }
            });
            // If the edge version comparator has a operator then our version
            // isn't outside it
            if (high.operator === comp || high.operator === ecomp) {
                return false;
            }
            // If the lowest version comparator has an operator and our version
            // is less than it then it isn't higher than the range
            if ((!low.operator || low.operator === comp) &&
                ltefn(version, low.semver)) {
                return false;
            }
            else if (low.operator === ecomp && ltfn(version, low.semver)) {
                return false;
            }
        }
        return true;
    };
    module.exports = outside;
},
/* semver/ranges/gtr.js */ function _(require, module, exports, __esModule, __esExport) {
    // Determine if version is greater than all the versions possible in the range.
    const outside = require(1045) /* ./outside */;
    const gtr = (version, range, options) => outside(version, range, '>', options);
    module.exports = gtr;
},
/* semver/ranges/ltr.js */ function _(require, module, exports, __esModule, __esExport) {
    const outside = require(1045) /* ./outside */;
    // Determine if version is less than all the versions possible in the range
    const ltr = (version, range, options) => outside(version, range, '<', options);
    module.exports = ltr;
},
/* semver/ranges/intersects.js */ function _(require, module, exports, __esModule, __esExport) {
    const Range = require(1035) /* ../classes/range */;
    const intersects = (r1, r2, options) => {
        r1 = new Range(r1, options);
        r2 = new Range(r2, options);
        return r1.intersects(r2);
    };
    module.exports = intersects;
},
/* semver/ranges/simplify.js */ function _(require, module, exports, __esModule, __esExport) {
    // given a set of versions and a range, create a "simplified" range
    // that includes the same versions that the original range does
    // If the original range is shorter than the simplified one, return that.
    const satisfies = require(1039) /* ../functions/satisfies.js */;
    const compare = require(1017) /* ../functions/compare.js */;
    module.exports = (versions, range, options) => {
        const set = [];
        let min = null;
        let prev = null;
        const v = versions.sort((a, b) => compare(a, b, options));
        for (const version of v) {
            const included = satisfies(version, range, options);
            if (included) {
                prev = version;
                if (!min)
                    min = version;
            }
            else {
                if (prev) {
                    set.push([min, prev]);
                }
                prev = null;
                min = null;
            }
        }
        if (min)
            set.push([min, null]);
        const ranges = [];
        for (const [min, max] of set) {
            if (min === max)
                ranges.push(min);
            else if (!max && min === v[0])
                ranges.push('*');
            else if (!max)
                ranges.push(`>=${min}`);
            else if (min === v[0])
                ranges.push(`<=${max}`);
            else
                ranges.push(`${min} - ${max}`);
        }
        const simplified = ranges.join(' || ');
        const original = typeof range.raw === 'string' ? range.raw : String(range);
        return simplified.length < original.length ? simplified : range;
    };
},
/* semver/ranges/subset.js */ function _(require, module, exports, __esModule, __esExport) {
    const Range = require(1035) /* ../classes/range.js */;
    const Comparator = require(1034) /* ../classes/comparator.js */;
    const { ANY } = Comparator;
    const satisfies = require(1039) /* ../functions/satisfies.js */;
    const compare = require(1017) /* ../functions/compare.js */;
    // Complex range `r1 || r2 || ...` is a subset of `R1 || R2 || ...` iff:
    // - Every simple range `r1, r2, ...` is a null set, OR
    // - Every simple range `r1, r2, ...` which is not a null set is a subset of
    //   some `R1, R2, ...`
    //
    // Simple range `c1 c2 ...` is a subset of simple range `C1 C2 ...` iff:
    // - If c is only the ANY comparator
    //   - If C is only the ANY comparator, return true
    //   - Else if in prerelease mode, return false
    //   - else replace c with `[>=0.0.0]`
    // - If C is only the ANY comparator
    //   - if in prerelease mode, return true
    //   - else replace C with `[>=0.0.0]`
    // - Let EQ be the set of = comparators in c
    // - If EQ is more than one, return true (null set)
    // - Let GT be the highest > or >= comparator in c
    // - Let LT be the lowest < or <= comparator in c
    // - If GT and LT, and GT.semver > LT.semver, return true (null set)
    // - If any C is a = range, and GT or LT are set, return false
    // - If EQ
    //   - If GT, and EQ does not satisfy GT, return true (null set)
    //   - If LT, and EQ does not satisfy LT, return true (null set)
    //   - If EQ satisfies every C, return true
    //   - Else return false
    // - If GT
    //   - If GT.semver is lower than any > or >= comp in C, return false
    //   - If GT is >=, and GT.semver does not satisfy every C, return false
    //   - If GT.semver has a prerelease, and not in prerelease mode
    //     - If no C has a prerelease and the GT.semver tuple, return false
    // - If LT
    //   - If LT.semver is greater than any < or <= comp in C, return false
    //   - If LT is <=, and LT.semver does not satisfy every C, return false
    //   - If GT.semver has a prerelease, and not in prerelease mode
    //     - If no C has a prerelease and the LT.semver tuple, return false
    // - Else return true
    const subset = (sub, dom, options = {}) => {
        if (sub === dom)
            return true;
        sub = new Range(sub, options);
        dom = new Range(dom, options);
        let sawNonNull = false;
        OUTER: for (const simpleSub of sub.set) {
            for (const simpleDom of dom.set) {
                const isSub = simpleSubset(simpleSub, simpleDom, options);
                sawNonNull = sawNonNull || isSub !== null;
                if (isSub)
                    continue OUTER;
            }
            // the null set is a subset of everything, but null simple ranges in
            // a complex range should be ignored.  so if we saw a non-null range,
            // then we know this isn't a subset, but if EVERY simple range was null,
            // then it is a subset.
            if (sawNonNull)
                return false;
        }
        return true;
    };
    const simpleSubset = (sub, dom, options) => {
        if (sub === dom)
            return true;
        if (sub.length === 1 && sub[0].semver === ANY) {
            if (dom.length === 1 && dom[0].semver === ANY)
                return true;
            else if (options.includePrerelease)
                sub = [new Comparator('>=0.0.0-0')];
            else
                sub = [new Comparator('>=0.0.0')];
        }
        if (dom.length === 1 && dom[0].semver === ANY) {
            if (options.includePrerelease)
                return true;
            else
                dom = [new Comparator('>=0.0.0')];
        }
        const eqSet = new Set();
        let gt, lt;
        for (const c of sub) {
            if (c.operator === '>' || c.operator === '>=')
                gt = higherGT(gt, c, options);
            else if (c.operator === '<' || c.operator === '<=')
                lt = lowerLT(lt, c, options);
            else
                eqSet.add(c.semver);
        }
        if (eqSet.size > 1)
            return null;
        let gtltComp;
        if (gt && lt) {
            gtltComp = compare(gt.semver, lt.semver, options);
            if (gtltComp > 0)
                return null;
            else if (gtltComp === 0 && (gt.operator !== '>=' || lt.operator !== '<='))
                return null;
        }
        // will iterate one or zero times
        for (const eq of eqSet) {
            if (gt && !satisfies(eq, String(gt), options))
                return null;
            if (lt && !satisfies(eq, String(lt), options))
                return null;
            for (const c of dom) {
                if (!satisfies(eq, String(c), options))
                    return false;
            }
            return true;
        }
        let higher, lower;
        let hasDomLT, hasDomGT;
        // if the subset has a prerelease, we need a comparator in the superset
        // with the same tuple and a prerelease, or it's not a subset
        let needDomLTPre = lt &&
            !options.includePrerelease &&
            lt.semver.prerelease.length ? lt.semver : false;
        let needDomGTPre = gt &&
            !options.includePrerelease &&
            gt.semver.prerelease.length ? gt.semver : false;
        // exception: <1.2.3-0 is the same as <1.2.3
        if (needDomLTPre && needDomLTPre.prerelease.length === 1 &&
            lt.operator === '<' && needDomLTPre.prerelease[0] === 0) {
            needDomLTPre = false;
        }
        for (const c of dom) {
            hasDomGT = hasDomGT || c.operator === '>' || c.operator === '>=';
            hasDomLT = hasDomLT || c.operator === '<' || c.operator === '<=';
            if (gt) {
                if (needDomGTPre) {
                    if (c.semver.prerelease && c.semver.prerelease.length &&
                        c.semver.major === needDomGTPre.major &&
                        c.semver.minor === needDomGTPre.minor &&
                        c.semver.patch === needDomGTPre.patch) {
                        needDomGTPre = false;
                    }
                }
                if (c.operator === '>' || c.operator === '>=') {
                    higher = higherGT(gt, c, options);
                    if (higher === c && higher !== gt)
                        return false;
                }
                else if (gt.operator === '>=' && !satisfies(gt.semver, String(c), options))
                    return false;
            }
            if (lt) {
                if (needDomLTPre) {
                    if (c.semver.prerelease && c.semver.prerelease.length &&
                        c.semver.major === needDomLTPre.major &&
                        c.semver.minor === needDomLTPre.minor &&
                        c.semver.patch === needDomLTPre.patch) {
                        needDomLTPre = false;
                    }
                }
                if (c.operator === '<' || c.operator === '<=') {
                    lower = lowerLT(lt, c, options);
                    if (lower === c && lower !== lt)
                        return false;
                }
                else if (lt.operator === '<=' && !satisfies(lt.semver, String(c), options))
                    return false;
            }
            if (!c.operator && (lt || gt) && gtltComp !== 0)
                return false;
        }
        // if there was a < or >, and nothing in the dom, then must be false
        // UNLESS it was limited by another range in the other direction.
        // Eg, >1.0.0 <1.0.1 is still a subset of <2.0.0
        if (gt && hasDomLT && !lt && gtltComp !== 0)
            return false;
        if (lt && hasDomGT && !gt && gtltComp !== 0)
            return false;
        // we needed a prerelease range in a specific tuple, but didn't get one
        // then this isn't a subset.  eg >=1.2.3-pre is not a subset of >=1.0.0,
        // because it includes prereleases in the 1.2.3 tuple
        if (needDomGTPre || needDomLTPre)
            return false;
        return true;
    };
    // >=1.2.3 is lower than >1.2.3
    const higherGT = (a, b, options) => {
        if (!a)
            return b;
        const comp = compare(a.semver, b.semver, options);
        return comp > 0 ? a
            : comp < 0 ? b
                : b.operator === '>' && a.operator === '>=' ? b
                    : a;
    };
    // <=1.2.3 is higher than <1.2.3
    const lowerLT = (a, b, options) => {
        if (!a)
            return b;
        const comp = compare(a.semver, b.semver, options);
        return comp < 0 ? a
            : comp > 0 ? b
                : b.operator === '<' && a.operator === '<=' ? b
                    : a;
    };
    module.exports = subset;
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-non-null-assertion.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-non-null-assertion',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows non-null assertions using the `!` postfix operator',
                category: 'Stylistic Issues',
                recommended: 'warn',
                suggestion: true,
            },
            messages: {
                noNonNull: 'Forbidden non-null assertion.',
                suggestOptionalChain: 'Consider using the optional chain operator `?.` instead. This operator includes runtime checks, so it is safer than the compile-only non-null assertion operator.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                TSNonNullExpression(node) {
                    var _a, _b;
                    const suggest = [];
                    function convertTokenToOptional(replacement) {
                        return (fixer) => {
                            const operator = sourceCode.getTokenAfter(node.expression, util.isNonNullAssertionPunctuator);
                            if (operator) {
                                return fixer.replaceText(operator, replacement);
                            }
                            return null;
                        };
                    }
                    function removeToken() {
                        return (fixer) => {
                            const operator = sourceCode.getTokenAfter(node.expression, util.isNonNullAssertionPunctuator);
                            if (operator) {
                                return fixer.remove(operator);
                            }
                            return null;
                        };
                    }
                    if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                        node.parent.object === node) {
                        if (!node.parent.optional) {
                            if (node.parent.computed) {
                                // it is x![y]?.z
                                suggest.push({
                                    messageId: 'suggestOptionalChain',
                                    fix: convertTokenToOptional('?.'),
                                });
                            }
                            else {
                                // it is x!.y?.z
                                suggest.push({
                                    messageId: 'suggestOptionalChain',
                                    fix: convertTokenToOptional('?'),
                                });
                            }
                        }
                        else {
                            if (node.parent.computed) {
                                // it is x!?.[y].z
                                suggest.push({
                                    messageId: 'suggestOptionalChain',
                                    fix: removeToken(),
                                });
                            }
                            else {
                                // it is x!?.y.z
                                suggest.push({
                                    messageId: 'suggestOptionalChain',
                                    fix: removeToken(),
                                });
                            }
                        }
                    }
                    else if (((_b = node.parent) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.CallExpression &&
                        node.parent.callee === node) {
                        if (!node.parent.optional) {
                            // it is x.y?.z!()
                            suggest.push({
                                messageId: 'suggestOptionalChain',
                                fix: convertTokenToOptional('?.'),
                            });
                        }
                        else {
                            // it is x.y.z!?.()
                            suggest.push({
                                messageId: 'suggestOptionalChain',
                                fix: removeToken(),
                            });
                        }
                    }
                    context.report({
                        node,
                        messageId: 'noNonNull',
                        suggest,
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-parameter-properties.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-parameter-properties',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow the use of parameter properties in class constructors',
                category: 'Stylistic Issues',
                // too opinionated to be recommended
                recommended: false,
            },
            messages: {
                noParamProp: 'Property {{parameter}} cannot be declared in the constructor.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allows: {
                            type: 'array',
                            items: {
                                enum: [
                                    'readonly',
                                    'private',
                                    'protected',
                                    'public',
                                    'private readonly',
                                    'protected readonly',
                                    'public readonly',
                                ],
                            },
                            minItems: 1,
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                allows: [],
            },
        ],
        create(context, [{ allows }]) {
            /**
             * Gets the modifiers of `node`.
             * @param node the node to be inspected.
             */
            function getModifiers(node) {
                const modifiers = [];
                if (node.accessibility) {
                    modifiers.push(node.accessibility);
                }
                if (node.readonly) {
                    modifiers.push('readonly');
                }
                return modifiers.filter(Boolean).join(' ');
            }
            return {
                TSParameterProperty(node) {
                    const modifiers = getModifiers(node);
                    if (!allows.includes(modifiers)) {
                        // HAS to be an identifier or assignment or TSC will throw
                        if (node.parameter.type !== experimental_utils_1.AST_NODE_TYPES.Identifier &&
                            node.parameter.type !== experimental_utils_1.AST_NODE_TYPES.AssignmentPattern) {
                            return;
                        }
                        const name = node.parameter.type === experimental_utils_1.AST_NODE_TYPES.Identifier
                            ? node.parameter.name
                            : // has to be an Identifier or TSC will throw an error
                                node.parameter.left.name;
                        context.report({
                            node,
                            messageId: 'noParamProp',
                            data: {
                                parameter: name,
                            },
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-redeclare.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-redeclare',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow variable redeclaration',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        builtinGlobals: {
                            type: 'boolean',
                        },
                        ignoreDeclarationMerge: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
            messages: {
                redeclared: "'{{id}}' is already defined.",
                redeclaredAsBuiltin: "'{{id}}' is already defined as a built-in global variable.",
                redeclaredBySyntax: "'{{id}}' is already defined by a variable declaration.",
            },
        },
        defaultOptions: [
            {
                builtinGlobals: true,
                ignoreDeclarationMerge: true,
            },
        ],
        create(context, [options]) {
            const sourceCode = context.getSourceCode();
            const CLASS_DECLARATION_MERGE_NODES = new Set([
                experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration,
                experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration,
                experimental_utils_1.AST_NODE_TYPES.ClassDeclaration,
            ]);
            const FUNCTION_DECLARATION_MERGE_NODES = new Set([
                experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration,
                experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration,
            ]);
            const ENUM_DECLARATION_MERGE_NODES = new Set([
                experimental_utils_1.AST_NODE_TYPES.TSEnumDeclaration,
                experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration,
            ]);
            function* iterateDeclarations(variable) {
                if ((options === null || options === void 0 ? void 0 : options.builtinGlobals) &&
                    'eslintImplicitGlobalSetting' in variable &&
                    (variable.eslintImplicitGlobalSetting === 'readonly' ||
                        variable.eslintImplicitGlobalSetting === 'writable')) {
                    yield { type: 'builtin' };
                }
                if ('eslintExplicitGlobalComments' in variable &&
                    variable.eslintExplicitGlobalComments) {
                    for (const comment of variable.eslintExplicitGlobalComments) {
                        yield {
                            type: 'comment',
                            node: comment,
                            loc: util.getNameLocationInGlobalDirectiveComment(sourceCode, comment, variable.name),
                        };
                    }
                }
                const identifiers = variable.identifiers
                    .map(id => ({
                    identifier: id,
                    parent: id.parent,
                }))
                    // ignore function declarations because TS will treat them as an overload
                    .filter(({ parent }) => parent.type !== experimental_utils_1.AST_NODE_TYPES.TSDeclareFunction);
                if (options.ignoreDeclarationMerge && identifiers.length > 1) {
                    if (
                    // interfaces merging
                    identifiers.every(({ parent }) => parent.type === experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration)) {
                        return;
                    }
                    if (
                    // namespace/module merging
                    identifiers.every(({ parent }) => parent.type === experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration)) {
                        return;
                    }
                    if (
                    // class + interface/namespace merging
                    identifiers.every(({ parent }) => CLASS_DECLARATION_MERGE_NODES.has(parent.type))) {
                        const classDecls = identifiers.filter(({ parent }) => parent.type === experimental_utils_1.AST_NODE_TYPES.ClassDeclaration);
                        if (classDecls.length === 1) {
                            // safe declaration merging
                            return;
                        }
                        // there's more than one class declaration, which needs to be reported
                        for (const { identifier } of classDecls) {
                            yield { type: 'syntax', node: identifier, loc: identifier.loc };
                        }
                        return;
                    }
                    if (
                    // class + interface/namespace merging
                    identifiers.every(({ parent }) => FUNCTION_DECLARATION_MERGE_NODES.has(parent.type))) {
                        const functionDecls = identifiers.filter(({ parent }) => parent.type === experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration);
                        if (functionDecls.length === 1) {
                            // safe declaration merging
                            return;
                        }
                        // there's more than one function declaration, which needs to be reported
                        for (const { identifier } of functionDecls) {
                            yield { type: 'syntax', node: identifier, loc: identifier.loc };
                        }
                        return;
                    }
                    if (
                    // enum + namespace merging
                    identifiers.every(({ parent }) => ENUM_DECLARATION_MERGE_NODES.has(parent.type))) {
                        const enumDecls = identifiers.filter(({ parent }) => parent.type === experimental_utils_1.AST_NODE_TYPES.TSEnumDeclaration);
                        if (enumDecls.length === 1) {
                            // safe declaration merging
                            return;
                        }
                        // there's more than one enum declaration, which needs to be reported
                        for (const { identifier } of enumDecls) {
                            yield { type: 'syntax', node: identifier, loc: identifier.loc };
                        }
                        return;
                    }
                }
                for (const { identifier } of identifiers) {
                    yield { type: 'syntax', node: identifier, loc: identifier.loc };
                }
            }
            function findVariablesInScope(scope) {
                for (const variable of scope.variables) {
                    const [declaration, ...extraDeclarations] = iterateDeclarations(variable);
                    if (extraDeclarations.length === 0) {
                        continue;
                    }
                    /*
                     * If the type of a declaration is different from the type of
                     * the first declaration, it shows the location of the first
                     * declaration.
                     */
                    const detailMessageId = declaration.type === 'builtin'
                        ? 'redeclaredAsBuiltin'
                        : 'redeclaredBySyntax';
                    const data = { id: variable.name };
                    // Report extra declarations.
                    for (const { type, node, loc } of extraDeclarations) {
                        const messageId = type === declaration.type ? 'redeclared' : detailMessageId;
                        if (node) {
                            context.report({ node, loc, messageId, data });
                        }
                        else if (loc) {
                            context.report({ loc, messageId, data });
                        }
                    }
                }
            }
            /**
             * Find variables in the current scope.
             */
            function checkForBlock(node) {
                const scope = context.getScope();
                /*
                 * In ES5, some node type such as `BlockStatement` doesn't have that scope.
                 * `scope.block` is a different node in such a case.
                 */
                if (scope.block === node) {
                    findVariablesInScope(scope);
                }
            }
            return {
                Program() {
                    const scope = context.getScope();
                    findVariablesInScope(scope);
                    // Node.js or ES modules has a special scope.
                    if (scope.type === 'global' &&
                        scope.childScopes[0] &&
                        // The special scope's block is the Program node.
                        scope.block === scope.childScopes[0].block) {
                        findVariablesInScope(scope.childScopes[0]);
                    }
                },
                FunctionDeclaration: checkForBlock,
                FunctionExpression: checkForBlock,
                ArrowFunctionExpression: checkForBlock,
                BlockStatement: checkForBlock,
                ForStatement: checkForBlock,
                ForInStatement: checkForBlock,
                ForOfStatement: checkForBlock,
                SwitchStatement: checkForBlock,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-require-imports.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-require-imports',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows invocation of `require()`',
                category: 'Best Practices',
                recommended: false,
            },
            schema: [],
            messages: {
                noRequireImports: 'A `require()` style import is forbidden.',
            },
        },
        defaultOptions: [],
        create(context) {
            return {
                'CallExpression > Identifier[name="require"]'(node) {
                    context.report({
                        node: node.parent,
                        messageId: 'noRequireImports',
                    });
                },
                TSExternalModuleReference(node) {
                    context.report({
                        node,
                        messageId: 'noRequireImports',
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-shadow.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const scope_manager_1 = require(768) /* @typescript-eslint/scope-manager */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-shadow',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow variable declarations from shadowing variables declared in the outer scope',
                category: 'Variables',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        builtinGlobals: {
                            type: 'boolean',
                        },
                        hoist: {
                            enum: ['all', 'functions', 'never'],
                        },
                        allow: {
                            type: 'array',
                            items: {
                                type: 'string',
                            },
                        },
                        ignoreTypeValueShadow: {
                            type: 'boolean',
                        },
                        ignoreFunctionTypeParameterNameValueShadow: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
            messages: {
                noShadow: "'{{name}}' is already declared in the upper scope.",
            },
        },
        defaultOptions: [
            {
                allow: [],
                builtinGlobals: false,
                hoist: 'functions',
                ignoreTypeValueShadow: true,
                ignoreFunctionTypeParameterNameValueShadow: true,
            },
        ],
        create(context, [options]) {
            /**
             * Check if a scope is a TypeScript module augmenting the global namespace.
             */
            function isGlobalAugmentation(scope) {
                return ((scope.type === scope_manager_1.ScopeType.tsModule && !!scope.block.global) ||
                    (!!scope.upper && isGlobalAugmentation(scope.upper)));
            }
            /**
             * Check if variable is a `this` parameter.
             */
            function isThisParam(variable) {
                return variable.defs[0].type === 'Parameter' && variable.name === 'this';
            }
            function isTypeValueShadow(variable, shadowed) {
                if (options.ignoreTypeValueShadow !== true) {
                    return false;
                }
                if (!('isValueVariable' in variable)) {
                    // this shouldn't happen...
                    return false;
                }
                const isShadowedValue = 'isValueVariable' in shadowed ? shadowed.isValueVariable : true;
                return variable.isValueVariable !== isShadowedValue;
            }
            function isFunctionTypeParameterNameValueShadow(variable, shadowed) {
                if (options.ignoreFunctionTypeParameterNameValueShadow !== true) {
                    return false;
                }
                if (!('isValueVariable' in variable)) {
                    // this shouldn't happen...
                    return false;
                }
                const isShadowedValue = 'isValueVariable' in shadowed ? shadowed.isValueVariable : true;
                if (!isShadowedValue) {
                    return false;
                }
                const id = variable.identifiers[0];
                return util.isFunctionType(id.parent);
            }
            function isGenericOfStaticMethod(variable) {
                if (!('isTypeVariable' in variable)) {
                    // this shouldn't happen...
                    return false;
                }
                if (!variable.isTypeVariable) {
                    return false;
                }
                if (variable.identifiers.length === 0) {
                    return false;
                }
                const typeParameter = variable.identifiers[0].parent;
                if ((typeParameter === null || typeParameter === void 0 ? void 0 : typeParameter.type) !== experimental_utils_1.AST_NODE_TYPES.TSTypeParameter) {
                    return false;
                }
                const typeParameterDecl = typeParameter.parent;
                if ((typeParameterDecl === null || typeParameterDecl === void 0 ? void 0 : typeParameterDecl.type) !== experimental_utils_1.AST_NODE_TYPES.TSTypeParameterDeclaration) {
                    return false;
                }
                const functionExpr = typeParameterDecl.parent;
                if (!functionExpr ||
                    (functionExpr.type !== experimental_utils_1.AST_NODE_TYPES.FunctionExpression &&
                        functionExpr.type !== experimental_utils_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression)) {
                    return false;
                }
                const methodDefinition = functionExpr.parent;
                if ((methodDefinition === null || methodDefinition === void 0 ? void 0 : methodDefinition.type) !== experimental_utils_1.AST_NODE_TYPES.MethodDefinition) {
                    return false;
                }
                return methodDefinition.static;
            }
            function isGenericOfClassDecl(variable) {
                if (!('isTypeVariable' in variable)) {
                    // this shouldn't happen...
                    return false;
                }
                if (!variable.isTypeVariable) {
                    return false;
                }
                if (variable.identifiers.length === 0) {
                    return false;
                }
                const typeParameter = variable.identifiers[0].parent;
                if ((typeParameter === null || typeParameter === void 0 ? void 0 : typeParameter.type) !== experimental_utils_1.AST_NODE_TYPES.TSTypeParameter) {
                    return false;
                }
                const typeParameterDecl = typeParameter.parent;
                if ((typeParameterDecl === null || typeParameterDecl === void 0 ? void 0 : typeParameterDecl.type) !== experimental_utils_1.AST_NODE_TYPES.TSTypeParameterDeclaration) {
                    return false;
                }
                const classDecl = typeParameterDecl.parent;
                return (classDecl === null || classDecl === void 0 ? void 0 : classDecl.type) === experimental_utils_1.AST_NODE_TYPES.ClassDeclaration;
            }
            function isGenericOfAStaticMethodShadow(variable, shadowed) {
                return (isGenericOfStaticMethod(variable) && isGenericOfClassDecl(shadowed));
            }
            /**
             * Check if variable name is allowed.
             * @param variable The variable to check.
             * @returns Whether or not the variable name is allowed.
             */
            function isAllowed(variable) {
                return options.allow.indexOf(variable.name) !== -1;
            }
            /**
             * Checks if a variable of the class name in the class scope of ClassDeclaration.
             *
             * ClassDeclaration creates two variables of its name into its outer scope and its class scope.
             * So we should ignore the variable in the class scope.
             * @param variable The variable to check.
             * @returns Whether or not the variable of the class name in the class scope of ClassDeclaration.
             */
            function isDuplicatedClassNameVariable(variable) {
                const block = variable.scope.block;
                return (block.type === experimental_utils_1.AST_NODE_TYPES.ClassDeclaration &&
                    block.id === variable.identifiers[0]);
            }
            /**
             * Checks if a variable of the class name in the class scope of TSEnumDeclaration.
             *
             * TSEnumDeclaration creates two variables of its name into its outer scope and its class scope.
             * So we should ignore the variable in the class scope.
             * @param variable The variable to check.
             * @returns Whether or not the variable of the class name in the class scope of TSEnumDeclaration.
             */
            function isDuplicatedEnumNameVariable(variable) {
                const block = variable.scope.block;
                return (block.type === experimental_utils_1.AST_NODE_TYPES.TSEnumDeclaration &&
                    block.id === variable.identifiers[0]);
            }
            /**
             * Checks if a variable is inside the initializer of scopeVar.
             *
             * To avoid reporting at declarations such as `var a = function a() {};`.
             * But it should report `var a = function(a) {};` or `var a = function() { function a() {} };`.
             * @param variable The variable to check.
             * @param scopeVar The scope variable to look for.
             * @returns Whether or not the variable is inside initializer of scopeVar.
             */
            function isOnInitializer(variable, scopeVar) {
                var _a;
                const outerScope = scopeVar.scope;
                const outerDef = scopeVar.defs[0];
                const outer = (_a = outerDef === null || outerDef === void 0 ? void 0 : outerDef.parent) === null || _a === void 0 ? void 0 : _a.range;
                const innerScope = variable.scope;
                const innerDef = variable.defs[0];
                const inner = innerDef === null || innerDef === void 0 ? void 0 : innerDef.name.range;
                return !!(outer &&
                    inner &&
                    outer[0] < inner[0] &&
                    inner[1] < outer[1] &&
                    ((innerDef.type === 'FunctionName' &&
                        innerDef.node.type === experimental_utils_1.AST_NODE_TYPES.FunctionExpression) ||
                        innerDef.node.type === experimental_utils_1.AST_NODE_TYPES.ClassExpression) &&
                    outerScope === innerScope.upper);
            }
            /**
             * Get a range of a variable's identifier node.
             * @param variable The variable to get.
             * @returns The range of the variable's identifier node.
             */
            function getNameRange(variable) {
                const def = variable.defs[0];
                return def === null || def === void 0 ? void 0 : def.name.range;
            }
            /**
             * Checks if a variable is in TDZ of scopeVar.
             * @param variable The variable to check.
             * @param scopeVar The variable of TDZ.
             * @returns Whether or not the variable is in TDZ of scopeVar.
             */
            function isInTdz(variable, scopeVar) {
                const outerDef = scopeVar.defs[0];
                const inner = getNameRange(variable);
                const outer = getNameRange(scopeVar);
                return !!(inner &&
                    outer &&
                    inner[1] < outer[0] &&
                    // Excepts FunctionDeclaration if is {"hoist":"function"}.
                    (options.hoist !== 'functions' ||
                        !outerDef ||
                        outerDef.node.type !== experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration));
            }
            /**
             * Finds the variable by a given name in a given scope and its upper scopes.
             * @param initScope A scope to start find.
             * @param name A variable name to find.
             * @returns A found variable or `null`.
             */
            function getVariableByName(initScope, name) {
                let scope = initScope;
                while (scope) {
                    const variable = scope.set.get(name);
                    if (variable) {
                        return variable;
                    }
                    scope = scope.upper;
                }
                return null;
            }
            /**
             * Checks the current context for shadowed variables.
             * @param {Scope} scope Fixme
             */
            function checkForShadows(scope) {
                // ignore global augmentation
                if (isGlobalAugmentation(scope)) {
                    return;
                }
                const variables = scope.variables;
                for (const variable of variables) {
                    // ignore "arguments"
                    if (variable.identifiers.length === 0) {
                        continue;
                    }
                    // this params are pseudo-params that cannot be shadowed
                    if (isThisParam(variable)) {
                        continue;
                    }
                    // ignore variables of a class name in the class scope of ClassDeclaration
                    if (isDuplicatedClassNameVariable(variable)) {
                        continue;
                    }
                    // ignore variables of a class name in the class scope of ClassDeclaration
                    if (isDuplicatedEnumNameVariable(variable)) {
                        continue;
                    }
                    // ignore configured allowed names
                    if (isAllowed(variable)) {
                        continue;
                    }
                    // Gets shadowed variable.
                    const shadowed = getVariableByName(scope.upper, variable.name);
                    if (!shadowed) {
                        continue;
                    }
                    // ignore type value variable shadowing if configured
                    if (isTypeValueShadow(variable, shadowed)) {
                        continue;
                    }
                    // ignore function type parameter name shadowing if configured
                    if (isFunctionTypeParameterNameValueShadow(variable, shadowed)) {
                        continue;
                    }
                    // ignore static class method generic shadowing class generic
                    // this is impossible for the scope analyser to understand
                    // so we have to handle this manually in this rule
                    if (isGenericOfAStaticMethodShadow(variable, shadowed)) {
                        continue;
                    }
                    const isESLintGlobal = 'writeable' in shadowed;
                    if ((shadowed.identifiers.length > 0 ||
                        (options.builtinGlobals && isESLintGlobal)) &&
                        !isOnInitializer(variable, shadowed) &&
                        !(options.hoist !== 'all' && isInTdz(variable, shadowed))) {
                        context.report({
                            node: variable.identifiers[0],
                            messageId: 'noShadow',
                            data: {
                                name: variable.name,
                            },
                        });
                    }
                }
            }
            return {
                'Program:exit'() {
                    const globalScope = context.getScope();
                    const stack = globalScope.childScopes.slice();
                    while (stack.length) {
                        const scope = stack.pop();
                        stack.push(...scope.childScopes);
                        checkForShadows(scope);
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-this-alias.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-this-alias',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow aliasing `this`',
                category: 'Best Practices',
                recommended: 'error',
            },
            schema: [
                {
                    type: 'object',
                    additionalProperties: false,
                    properties: {
                        allowDestructuring: {
                            type: 'boolean',
                        },
                        allowedNames: {
                            type: 'array',
                            items: {
                                type: 'string',
                            },
                        },
                    },
                },
            ],
            messages: {
                thisAssignment: "Unexpected aliasing of 'this' to local variable.",
                thisDestructure: "Unexpected aliasing of members of 'this' to local variables.",
            },
        },
        defaultOptions: [
            {
                allowDestructuring: true,
                allowedNames: [],
            },
        ],
        create(context, [{ allowDestructuring, allowedNames }]) {
            return {
                "VariableDeclarator[init.type='ThisExpression']"(node) {
                    const { id } = node;
                    if (allowDestructuring && id.type !== experimental_utils_1.AST_NODE_TYPES.Identifier) {
                        return;
                    }
                    const hasAllowedName = id.type === experimental_utils_1.AST_NODE_TYPES.Identifier
                        ? allowedNames.includes(id.name)
                        : false;
                    if (!hasAllowedName) {
                        context.report({
                            node: id,
                            messageId: id.type === experimental_utils_1.AST_NODE_TYPES.Identifier
                                ? 'thisAssignment'
                                : 'thisDestructure',
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-throw-literal.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    exports.default = util.createRule({
        name: 'no-throw-literal',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow throwing literals as exceptions',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
                requiresTypeChecking: true,
            },
            schema: [],
            messages: {
                object: 'Expected an error object to be thrown.',
                undef: 'Do not throw undefined.',
            },
        },
        defaultOptions: [],
        create(context) {
            const parserServices = util.getParserServices(context);
            const program = parserServices.program;
            const checker = program.getTypeChecker();
            function isErrorLike(type) {
                var _a;
                if (type.isIntersection()) {
                    return type.types.some(isErrorLike);
                }
                if (type.isUnion()) {
                    return type.types.every(isErrorLike);
                }
                const symbol = type.getSymbol();
                if (!symbol) {
                    return false;
                }
                if (symbol.getName() === 'Error') {
                    const declarations = (_a = symbol.getDeclarations()) !== null && _a !== void 0 ? _a : [];
                    for (const declaration of declarations) {
                        const sourceFile = declaration.getSourceFile();
                        if (program.isSourceFileDefaultLibrary(sourceFile)) {
                            return true;
                        }
                    }
                }
                if (symbol.flags & (ts.SymbolFlags.Class | ts.SymbolFlags.Interface)) {
                    for (const baseType of checker.getBaseTypes(type)) {
                        if (isErrorLike(baseType)) {
                            return true;
                        }
                    }
                }
                return false;
            }
            function checkThrowArgument(node) {
                if (node.type === experimental_utils_1.AST_NODE_TYPES.AwaitExpression ||
                    node.type === experimental_utils_1.AST_NODE_TYPES.YieldExpression) {
                    return;
                }
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                const type = checker.getTypeAtLocation(tsNode);
                if (type.flags & ts.TypeFlags.Undefined) {
                    context.report({ node, messageId: 'undef' });
                    return;
                }
                if (util.isTypeAnyType(type) ||
                    util.isTypeUnknownType(type) ||
                    isErrorLike(type)) {
                    return;
                }
                context.report({ node, messageId: 'object' });
            }
            return {
                ThrowStatement(node) {
                    if (node.argument) {
                        checkThrowArgument(node.argument);
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-type-alias.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const enumValues = [
        'always',
        'never',
        'in-unions',
        'in-intersections',
        'in-unions-and-intersections',
    ];
    exports.default = util.createRule({
        name: 'no-type-alias',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow the use of type aliases',
                category: 'Stylistic Issues',
                // too opinionated to be recommended
                recommended: false,
            },
            messages: {
                noTypeAlias: 'Type {{alias}} are not allowed.',
                noCompositionAlias: '{{typeName}} in {{compositionType}} types are not allowed.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowAliases: {
                            enum: enumValues,
                        },
                        allowCallbacks: {
                            enum: ['always', 'never'],
                        },
                        allowConditionalTypes: {
                            enum: ['always', 'never'],
                        },
                        allowConstructors: {
                            enum: ['always', 'never'],
                        },
                        allowLiterals: {
                            enum: enumValues,
                        },
                        allowMappedTypes: {
                            enum: enumValues,
                        },
                        allowTupleTypes: {
                            enum: enumValues,
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                allowAliases: 'never',
                allowCallbacks: 'never',
                allowConditionalTypes: 'never',
                allowConstructors: 'never',
                allowLiterals: 'never',
                allowMappedTypes: 'never',
                allowTupleTypes: 'never',
            },
        ],
        create(context, [{ allowAliases, allowCallbacks, allowConditionalTypes, allowConstructors, allowLiterals, allowMappedTypes, allowTupleTypes, },]) {
            const unions = ['always', 'in-unions', 'in-unions-and-intersections'];
            const intersections = [
                'always',
                'in-intersections',
                'in-unions-and-intersections',
            ];
            const compositions = [
                'in-unions',
                'in-intersections',
                'in-unions-and-intersections',
            ];
            const aliasTypes = new Set([
                experimental_utils_1.AST_NODE_TYPES.TSArrayType,
                experimental_utils_1.AST_NODE_TYPES.TSImportType,
                experimental_utils_1.AST_NODE_TYPES.TSTypeReference,
                experimental_utils_1.AST_NODE_TYPES.TSLiteralType,
                experimental_utils_1.AST_NODE_TYPES.TSTypeQuery,
                experimental_utils_1.AST_NODE_TYPES.TSIndexedAccessType,
            ]);
            /**
             * Determines if the composition type is supported by the allowed flags.
             * @param isTopLevel a flag indicating this is the top level node.
             * @param compositionType the composition type (either TSUnionType or TSIntersectionType)
             * @param allowed the currently allowed flags.
             */
            function isSupportedComposition(isTopLevel, compositionType, allowed) {
                return (!compositions.includes(allowed) ||
                    (!isTopLevel &&
                        ((compositionType === experimental_utils_1.AST_NODE_TYPES.TSUnionType &&
                            unions.includes(allowed)) ||
                            (compositionType === experimental_utils_1.AST_NODE_TYPES.TSIntersectionType &&
                                intersections.includes(allowed)))));
            }
            /**
             * Gets the message to be displayed based on the node type and whether the node is a top level declaration.
             * @param node the location
             * @param compositionType the type of composition this alias is part of (undefined if not
             *                                  part of a composition)
             * @param isRoot a flag indicating we are dealing with the top level declaration.
             * @param type the kind of type alias being validated.
             */
            function reportError(node, compositionType, isRoot, type) {
                if (isRoot) {
                    return context.report({
                        node,
                        messageId: 'noTypeAlias',
                        data: {
                            alias: type.toLowerCase(),
                        },
                    });
                }
                return context.report({
                    node,
                    messageId: 'noCompositionAlias',
                    data: {
                        compositionType: compositionType === experimental_utils_1.AST_NODE_TYPES.TSUnionType
                            ? 'union'
                            : 'intersection',
                        typeName: type,
                    },
                });
            }
            const isValidTupleType = (type) => {
                if (type.node.type === experimental_utils_1.AST_NODE_TYPES.TSTupleType) {
                    return true;
                }
                if (type.node.type === experimental_utils_1.AST_NODE_TYPES.TSTypeOperator) {
                    if (['keyof', 'readonly'].includes(type.node.operator) &&
                        type.node.typeAnnotation &&
                        type.node.typeAnnotation.type === experimental_utils_1.AST_NODE_TYPES.TSTupleType) {
                        return true;
                    }
                }
                return false;
            };
            const checkAndReport = (optionValue, isTopLevel, type, label) => {
                if (optionValue === 'never' ||
                    !isSupportedComposition(isTopLevel, type.compositionType, optionValue)) {
                    reportError(type.node, type.compositionType, isTopLevel, label);
                }
            };
            /**
             * Validates the node looking for aliases, callbacks and literals.
             * @param type the type of composition this alias is part of (null if not
             *                                  part of a composition)
             * @param isTopLevel a flag indicating this is the top level node.
             */
            function validateTypeAliases(type, isTopLevel = false) {
                if (type.node.type === experimental_utils_1.AST_NODE_TYPES.TSFunctionType) {
                    // callback
                    if (allowCallbacks === 'never') {
                        reportError(type.node, type.compositionType, isTopLevel, 'Callbacks');
                    }
                }
                else if (type.node.type === experimental_utils_1.AST_NODE_TYPES.TSConditionalType) {
                    // conditional type
                    if (allowConditionalTypes === 'never') {
                        reportError(type.node, type.compositionType, isTopLevel, 'Conditional types');
                    }
                }
                else if (type.node.type === experimental_utils_1.AST_NODE_TYPES.TSConstructorType) {
                    if (allowConstructors === 'never') {
                        reportError(type.node, type.compositionType, isTopLevel, 'Constructors');
                    }
                }
                else if (type.node.type === experimental_utils_1.AST_NODE_TYPES.TSTypeLiteral) {
                    // literal object type
                    checkAndReport(allowLiterals, isTopLevel, type, 'Literals');
                }
                else if (type.node.type === experimental_utils_1.AST_NODE_TYPES.TSMappedType) {
                    // mapped type
                    checkAndReport(allowMappedTypes, isTopLevel, type, 'Mapped types');
                }
                else if (isValidTupleType(type)) {
                    // tuple types
                    checkAndReport(allowTupleTypes, isTopLevel, type, 'Tuple Types');
                }
                else if (
                // eslint-disable-next-line @typescript-eslint/internal/prefer-ast-types-enum
                type.node.type.endsWith('Keyword') ||
                    aliasTypes.has(type.node.type) ||
                    (type.node.type === experimental_utils_1.AST_NODE_TYPES.TSTypeOperator &&
                        (type.node.operator === 'keyof' ||
                            (type.node.operator === 'readonly' &&
                                type.node.typeAnnotation &&
                                aliasTypes.has(type.node.typeAnnotation.type))))) {
                    // alias / keyword
                    checkAndReport(allowAliases, isTopLevel, type, 'Aliases');
                }
                else {
                    // unhandled type - shouldn't happen
                    reportError(type.node, type.compositionType, isTopLevel, 'Unhandled');
                }
            }
            /**
             * Flatten the given type into an array of its dependencies
             */
            function getTypes(node, compositionType = null) {
                if (node.type === experimental_utils_1.AST_NODE_TYPES.TSUnionType ||
                    node.type === experimental_utils_1.AST_NODE_TYPES.TSIntersectionType) {
                    return node.types.reduce((acc, type) => {
                        acc.push(...getTypes(type, node.type));
                        return acc;
                    }, []);
                }
                if (node.type === experimental_utils_1.AST_NODE_TYPES.TSParenthesizedType) {
                    return getTypes(node.typeAnnotation, compositionType);
                }
                return [{ node, compositionType }];
            }
            return {
                TSTypeAliasDeclaration(node) {
                    const types = getTypes(node.typeAnnotation);
                    if (types.length === 1) {
                        // is a top level type annotation
                        validateTypeAliases(types[0], true);
                    }
                    else {
                        // is a composition type
                        types.forEach(type => {
                            validateTypeAliases(type);
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unnecessary-boolean-literal-compare.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-unnecessary-boolean-literal-compare',
        meta: {
            docs: {
                description: 'Flags unnecessary equality comparisons against boolean literals',
                category: 'Stylistic Issues',
                recommended: false,
                requiresTypeChecking: true,
            },
            fixable: 'code',
            messages: {
                direct: 'This expression unnecessarily compares a boolean value to a boolean instead of using it directly.',
                negated: 'This expression unnecessarily compares a boolean value to a boolean instead of negating it.',
                comparingNullableToTrueDirect: 'This expression unnecessarily compares a nullable boolean value to true instead of using it directly.',
                comparingNullableToTrueNegated: 'This expression unnecessarily compares a nullable boolean value to true instead of negating it.',
                comparingNullableToFalse: 'This expression unnecessarily compares a nullable boolean value to false instead of using the ?? operator to provide a default.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowComparingNullableBooleansToTrue: {
                            type: 'boolean',
                        },
                        allowComparingNullableBooleansToFalse: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
            type: 'suggestion',
        },
        defaultOptions: [
            {
                allowComparingNullableBooleansToTrue: true,
                allowComparingNullableBooleansToFalse: true,
            },
        ],
        create(context, [options]) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            function getBooleanComparison(node) {
                const comparison = deconstructComparison(node);
                if (!comparison) {
                    return undefined;
                }
                const expressionType = checker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(comparison.expression));
                if (isBooleanType(expressionType)) {
                    return Object.assign(Object.assign({}, comparison), { expressionIsNullableBoolean: false });
                }
                if (isNullableBoolean(expressionType)) {
                    return Object.assign(Object.assign({}, comparison), { expressionIsNullableBoolean: true });
                }
                return undefined;
            }
            function isBooleanType(expressionType) {
                return tsutils.isTypeFlagSet(expressionType, ts.TypeFlags.Boolean | ts.TypeFlags.BooleanLiteral);
            }
            /**
             * checks if the expressionType is a union that
             *   1) contains at least one nullish type (null or undefined)
             *   2) contains at least once boolean type (true or false or boolean)
             *   3) does not contain any types besides nullish and boolean types
             */
            function isNullableBoolean(expressionType) {
                if (!expressionType.isUnion()) {
                    return false;
                }
                const { types } = expressionType;
                const nonNullishTypes = types.filter(type => !tsutils.isTypeFlagSet(type, ts.TypeFlags.Undefined | ts.TypeFlags.Null));
                const hasNonNullishType = nonNullishTypes.length > 0;
                if (!hasNonNullishType) {
                    return false;
                }
                const hasNullableType = nonNullishTypes.length < types.length;
                if (!hasNullableType) {
                    return false;
                }
                const allNonNullishTypesAreBoolean = nonNullishTypes.every(isBooleanType);
                if (!allNonNullishTypesAreBoolean) {
                    return false;
                }
                return true;
            }
            function deconstructComparison(node) {
                const comparisonType = util.getEqualsKind(node.operator);
                if (!comparisonType) {
                    return undefined;
                }
                for (const [against, expression] of [
                    [node.right, node.left],
                    [node.left, node.right],
                ]) {
                    if (against.type !== experimental_utils_1.AST_NODE_TYPES.Literal ||
                        typeof against.value !== 'boolean') {
                        continue;
                    }
                    const { value: literalBooleanInComparison } = against;
                    const negated = !comparisonType.isPositive;
                    return {
                        literalBooleanInComparison,
                        forTruthy: literalBooleanInComparison ? !negated : negated,
                        expression,
                        negated,
                        range: expression.range[0] < against.range[0]
                            ? [expression.range[1], against.range[1]]
                            : [against.range[0], expression.range[0]],
                    };
                }
                return undefined;
            }
            function nodeIsUnaryNegation(node) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                    node.prefix &&
                    node.operator === '!');
            }
            return {
                BinaryExpression(node) {
                    const comparison = getBooleanComparison(node);
                    if (comparison === undefined) {
                        return;
                    }
                    if (comparison.expressionIsNullableBoolean) {
                        if (comparison.literalBooleanInComparison &&
                            options.allowComparingNullableBooleansToTrue) {
                            return;
                        }
                        if (!comparison.literalBooleanInComparison &&
                            options.allowComparingNullableBooleansToFalse) {
                            return;
                        }
                    }
                    context.report({
                        fix: function* (fixer) {
                            yield fixer.removeRange(comparison.range);
                            // if the expression `exp` isn't nullable, or we're comparing to `true`,
                            // we can just replace the entire comparison with `exp` or `!exp`
                            if (!comparison.expressionIsNullableBoolean ||
                                comparison.literalBooleanInComparison) {
                                if (!comparison.forTruthy) {
                                    yield fixer.insertTextBefore(node, '!');
                                }
                                return;
                            }
                            // if we're here, then the expression is a nullable boolean and we're
                            // comparing to a literal `false`
                            // if we're doing `== false` or `=== false`, then we need to negate the expression
                            if (!comparison.negated) {
                                const { parent } = node;
                                // if the parent is a negation, we can instead just get rid of the parent's negation.
                                // i.e. instead of resulting in `!(!(exp))`, we can just result in `exp`
                                if (parent != null && nodeIsUnaryNegation(parent)) {
                                    // remove from the beginning of the parent to the beginning of this node
                                    yield fixer.removeRange([parent.range[0], node.range[0]]);
                                    // remove from the end of the node to the end of the parent
                                    yield fixer.removeRange([node.range[1], parent.range[1]]);
                                }
                                else {
                                    yield fixer.insertTextBefore(node, '!');
                                }
                            }
                            // provide the default `true`
                            yield fixer.insertTextBefore(node, '(');
                            yield fixer.insertTextAfter(node, ' ?? true)');
                        },
                        messageId: comparison.expressionIsNullableBoolean
                            ? comparison.literalBooleanInComparison
                                ? comparison.negated
                                    ? 'comparingNullableToTrueNegated'
                                    : 'comparingNullableToTrueDirect'
                                : 'comparingNullableToFalse'
                            : comparison.negated
                                ? 'negated'
                                : 'direct',
                        node,
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unnecessary-condition.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const ts = __importStar(require(21) /* typescript */);
    const tsutils_1 = require(916) /* tsutils */;
    const util_1 = require(906) /* ../util */;
    // Truthiness utilities
    // #region
    const isTruthyLiteral = (type) => tsutils_1.isBooleanLiteralType(type, true) || (tsutils_1.isLiteralType(type) && !!type.value);
    const isPossiblyFalsy = (type) => tsutils_1.unionTypeParts(type)
        // PossiblyFalsy flag includes literal values, so exclude ones that
        // are definitely truthy
        .filter(t => !isTruthyLiteral(t))
        .some(type => util_1.isTypeFlagSet(type, ts.TypeFlags.PossiblyFalsy));
    const isPossiblyTruthy = (type) => tsutils_1.unionTypeParts(type).some(type => !tsutils_1.isFalsyType(type));
    // Nullish utilities
    const nullishFlag = ts.TypeFlags.Undefined | ts.TypeFlags.Null;
    const isNullishType = (type) => util_1.isTypeFlagSet(type, nullishFlag);
    const isPossiblyNullish = (type) => tsutils_1.unionTypeParts(type).some(isNullishType);
    const isAlwaysNullish = (type) => tsutils_1.unionTypeParts(type).every(isNullishType);
    // isLiteralType only covers numbers and strings, this is a more exhaustive check.
    const isLiteral = (type) => tsutils_1.isBooleanLiteralType(type, true) ||
        tsutils_1.isBooleanLiteralType(type, false) ||
        type.flags === ts.TypeFlags.Undefined ||
        type.flags === ts.TypeFlags.Null ||
        type.flags === ts.TypeFlags.Void ||
        tsutils_1.isLiteralType(type);
    exports.default = util_1.createRule({
        name: 'no-unnecessary-condition',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Prevents conditionals where the type is always truthy or always falsy',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowConstantLoopConditions: {
                            type: 'boolean',
                        },
                        allowRuleToRunWithoutStrictNullChecksIKnowWhatIAmDoing: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
            fixable: 'code',
            messages: {
                alwaysTruthy: 'Unnecessary conditional, value is always truthy.',
                alwaysFalsy: 'Unnecessary conditional, value is always falsy.',
                alwaysTruthyFunc: 'This callback should return a conditional, but return is always truthy.',
                alwaysFalsyFunc: 'This callback should return a conditional, but return is always falsy.',
                neverNullish: 'Unnecessary conditional, expected left-hand side of `??` operator to be possibly null or undefined.',
                alwaysNullish: 'Unnecessary conditional, left-hand side of `??` operator is always `null` or `undefined`.',
                literalBooleanExpression: 'Unnecessary conditional, both sides of the expression are literal values',
                noOverlapBooleanExpression: 'Unnecessary conditional, the types have no overlap',
                never: 'Unnecessary conditional, value is `never`',
                neverOptionalChain: 'Unnecessary optional chain on a non-nullish value',
                noStrictNullCheck: 'This rule requires the `strictNullChecks` compiler option to be turned on to function correctly.',
            },
        },
        defaultOptions: [
            {
                allowConstantLoopConditions: false,
                allowRuleToRunWithoutStrictNullChecksIKnowWhatIAmDoing: false,
            },
        ],
        create(context, [{ allowConstantLoopConditions, allowRuleToRunWithoutStrictNullChecksIKnowWhatIAmDoing, },]) {
            const service = util_1.getParserServices(context);
            const checker = service.program.getTypeChecker();
            const sourceCode = context.getSourceCode();
            const compilerOptions = service.program.getCompilerOptions();
            const isStrictNullChecks = tsutils_1.isStrictCompilerOptionEnabled(compilerOptions, 'strictNullChecks');
            if (!isStrictNullChecks &&
                allowRuleToRunWithoutStrictNullChecksIKnowWhatIAmDoing !== true) {
                context.report({
                    loc: {
                        start: { line: 0, column: 0 },
                        end: { line: 0, column: 0 },
                    },
                    messageId: 'noStrictNullCheck',
                });
            }
            function getNodeType(node) {
                const tsNode = service.esTreeNodeToTSNodeMap.get(node);
                return util_1.getConstrainedTypeAtLocation(checker, tsNode);
            }
            function nodeIsArrayType(node) {
                const nodeType = getNodeType(node);
                return checker.isArrayType(nodeType);
            }
            function nodeIsTupleType(node) {
                const nodeType = getNodeType(node);
                return checker.isTupleType(nodeType);
            }
            function isArrayIndexExpression(node) {
                return (
                // Is an index signature
                node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                    node.computed &&
                    // ...into an array type
                    (nodeIsArrayType(node.object) ||
                        // ... or a tuple type
                        (nodeIsTupleType(node.object) &&
                            // Exception: literal index into a tuple - will have a sound type
                            node.property.type !== experimental_utils_1.AST_NODE_TYPES.Literal)));
            }
            /**
             * Checks if a conditional node is necessary:
             * if the type of the node is always true or always false, it's not necessary.
             */
            function checkNode(node, isUnaryNotArgument = false) {
                // Check if the node is Unary Negation expression and handle it
                if (node.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                    node.operator === '!') {
                    return checkNode(node.argument, true);
                }
                // Since typescript array index signature types don't represent the
                //  possibility of out-of-bounds access, if we're indexing into an array
                //  just skip the check, to avoid false positives
                if (isArrayIndexExpression(node)) {
                    return;
                }
                // When checking logical expressions, only check the right side
                //  as the left side has been checked by checkLogicalExpressionForUnnecessaryConditionals
                //
                // Unless the node is nullish coalescing, as it's common to use patterns like `nullBool ?? true` to to strict
                //  boolean checks if we inspect the right here, it'll usually be a constant condition on purpose.
                // In this case it's better to inspect the type of the expression as a whole.
                if (node.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression &&
                    node.operator !== '??') {
                    return checkNode(node.right);
                }
                const type = getNodeType(node);
                // Conditional is always necessary if it involves:
                //    `any` or `unknown` or a naked type parameter
                if (tsutils_1.unionTypeParts(type).some(part => util_1.isTypeAnyType(part) ||
                    util_1.isTypeUnknownType(part) ||
                    util_1.isTypeFlagSet(part, ts.TypeFlags.TypeParameter))) {
                    return;
                }
                let messageId = null;
                if (util_1.isTypeFlagSet(type, ts.TypeFlags.Never)) {
                    messageId = 'never';
                }
                else if (!isPossiblyTruthy(type)) {
                    messageId = !isUnaryNotArgument ? 'alwaysFalsy' : 'alwaysTruthy';
                }
                else if (!isPossiblyFalsy(type)) {
                    messageId = !isUnaryNotArgument ? 'alwaysTruthy' : 'alwaysFalsy';
                }
                if (messageId) {
                    context.report({ node, messageId });
                }
            }
            function checkNodeForNullish(node) {
                // Since typescript array index signature types don't represent the
                //  possibility of out-of-bounds access, if we're indexing into an array
                //  just skip the check, to avoid false positives
                if (isArrayIndexExpression(node)) {
                    return;
                }
                const type = getNodeType(node);
                // Conditional is always necessary if it involves `any` or `unknown`
                if (util_1.isTypeAnyType(type) || util_1.isTypeUnknownType(type)) {
                    return;
                }
                let messageId = null;
                if (util_1.isTypeFlagSet(type, ts.TypeFlags.Never)) {
                    messageId = 'never';
                }
                else if (!isPossiblyNullish(type)) {
                    messageId = 'neverNullish';
                }
                else if (isAlwaysNullish(type)) {
                    messageId = 'alwaysNullish';
                }
                if (messageId) {
                    context.report({ node, messageId });
                }
            }
            /**
             * Checks that a binary expression is necessarily conditional, reports otherwise.
             * If both sides of the binary expression are literal values, it's not a necessary condition.
             *
             * NOTE: It's also unnecessary if the types that don't overlap at all
             *    but that case is handled by the Typescript compiler itself.
             *    Known exceptions:
             *      * https://github.com/microsoft/TypeScript/issues/32627
             *      * https://github.com/microsoft/TypeScript/issues/37160 (handled)
             */
            const BOOL_OPERATORS = new Set([
                '<',
                '>',
                '<=',
                '>=',
                '==',
                '===',
                '!=',
                '!==',
            ]);
            function checkIfBinaryExpressionIsNecessaryConditional(node) {
                if (!BOOL_OPERATORS.has(node.operator)) {
                    return;
                }
                const leftType = getNodeType(node.left);
                const rightType = getNodeType(node.right);
                if (isLiteral(leftType) && isLiteral(rightType)) {
                    context.report({ node, messageId: 'literalBooleanExpression' });
                    return;
                }
                // Workaround for https://github.com/microsoft/TypeScript/issues/37160
                if (isStrictNullChecks) {
                    const UNDEFINED = ts.TypeFlags.Undefined;
                    const NULL = ts.TypeFlags.Null;
                    const isComparable = (type, flag) => {
                        // Allow comparison to `any`, `unknown` or a naked type parameter.
                        flag |=
                            ts.TypeFlags.Any |
                                ts.TypeFlags.Unknown |
                                ts.TypeFlags.TypeParameter;
                        // Allow loose comparison to nullish values.
                        if (node.operator === '==' || node.operator === '!=') {
                            flag |= NULL | UNDEFINED;
                        }
                        return util_1.isTypeFlagSet(type, flag);
                    };
                    if ((leftType.flags === UNDEFINED &&
                        !isComparable(rightType, UNDEFINED)) ||
                        (rightType.flags === UNDEFINED &&
                            !isComparable(leftType, UNDEFINED)) ||
                        (leftType.flags === NULL && !isComparable(rightType, NULL)) ||
                        (rightType.flags === NULL && !isComparable(leftType, NULL))) {
                        context.report({ node, messageId: 'noOverlapBooleanExpression' });
                        return;
                    }
                }
            }
            /**
             * Checks that a logical expression contains a boolean, reports otherwise.
             */
            function checkLogicalExpressionForUnnecessaryConditionals(node) {
                if (node.operator === '??') {
                    checkNodeForNullish(node.left);
                    return;
                }
                // Only checks the left side, since the right side might not be "conditional" at all.
                // The right side will be checked if the LogicalExpression is used in a conditional context
                checkNode(node.left);
            }
            /**
             * Checks that a testable expression of a loop is necessarily conditional, reports otherwise.
             */
            function checkIfLoopIsNecessaryConditional(node) {
                if (node.test === null) {
                    // e.g. `for(;;)`
                    return;
                }
                /**
                 * Allow:
                 *   while (true) {}
                 *   for (;true;) {}
                 *   do {} while (true)
                 */
                if (allowConstantLoopConditions &&
                    tsutils_1.isBooleanLiteralType(getNodeType(node.test), true)) {
                    return;
                }
                checkNode(node.test);
            }
            const ARRAY_PREDICATE_FUNCTIONS = new Set([
                'filter',
                'find',
                'some',
                'every',
            ]);
            function isArrayPredicateFunction(node) {
                const { callee } = node;
                return (
                // looks like `something.filter` or `something.find`
                callee.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                    callee.property.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    ARRAY_PREDICATE_FUNCTIONS.has(callee.property.name) &&
                    // and the left-hand side is an array, according to the types
                    (nodeIsArrayType(callee.object) || nodeIsTupleType(callee.object)));
            }
            function checkCallExpression(node) {
                // If this is something like arr.filter(x => /*condition*/), check `condition`
                if (isArrayPredicateFunction(node) && node.arguments.length) {
                    const callback = node.arguments[0];
                    // Inline defined functions
                    if ((callback.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression ||
                        callback.type === experimental_utils_1.AST_NODE_TYPES.FunctionExpression) &&
                        callback.body) {
                        // Two special cases, where we can directly check the node that's returned:
                        // () => something
                        if (callback.body.type !== experimental_utils_1.AST_NODE_TYPES.BlockStatement) {
                            return checkNode(callback.body);
                        }
                        // () => { return something; }
                        const callbackBody = callback.body.body;
                        if (callbackBody.length === 1 &&
                            callbackBody[0].type === experimental_utils_1.AST_NODE_TYPES.ReturnStatement &&
                            callbackBody[0].argument) {
                            return checkNode(callbackBody[0].argument);
                        }
                        // Potential enhancement: could use code-path analysis to check
                        //   any function with a single return statement
                        // (Value to complexity ratio is dubious however)
                    }
                    // Otherwise just do type analysis on the function as a whole.
                    const returnTypes = tsutils_1.getCallSignaturesOfType(getNodeType(callback)).map(sig => sig.getReturnType());
                    /* istanbul ignore if */ if (returnTypes.length === 0) {
                        // Not a callable function
                        return;
                    }
                    // Predicate is always necessary if it involves `any` or `unknown`
                    if (returnTypes.some(t => util_1.isTypeAnyType(t) || util_1.isTypeUnknownType(t))) {
                        return;
                    }
                    if (!returnTypes.some(isPossiblyFalsy)) {
                        return context.report({
                            node: callback,
                            messageId: 'alwaysTruthyFunc',
                        });
                    }
                    if (!returnTypes.some(isPossiblyTruthy)) {
                        return context.report({
                            node: callback,
                            messageId: 'alwaysFalsyFunc',
                        });
                    }
                }
            }
            // Recursively searches an optional chain for an array index expression
            //  Has to search the entire chain, because an array index will "infect" the rest of the types
            //  Example:
            //  ```
            //  [{x: {y: "z"} }][n] // type is {x: {y: "z"}}
            //    ?.x // type is {y: "z"}
            //    ?.y // This access is considered "unnecessary" according to the types
            //  ```
            function optionChainContainsArrayIndex(node) {
                const lhsNode = node.type === experimental_utils_1.AST_NODE_TYPES.CallExpression ? node.callee : node.object;
                if (isArrayIndexExpression(lhsNode)) {
                    return true;
                }
                if (lhsNode.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression ||
                    lhsNode.type === experimental_utils_1.AST_NODE_TYPES.CallExpression) {
                    return optionChainContainsArrayIndex(lhsNode);
                }
                return false;
            }
            function isNullablePropertyType(objType, propertyType) {
                if (propertyType.isUnion()) {
                    return propertyType.types.some(type => isNullablePropertyType(objType, type));
                }
                if (propertyType.isNumberLiteral() || propertyType.isStringLiteral()) {
                    const propType = util_1.getTypeOfPropertyOfName(checker, objType, propertyType.value.toString());
                    if (propType) {
                        return util_1.isNullableType(propType, { allowUndefined: true });
                    }
                }
                const typeName = util_1.getTypeName(checker, propertyType);
                return !!((typeName === 'string' &&
                    checker.getIndexInfoOfType(objType, ts.IndexKind.String)) ||
                    (typeName === 'number' &&
                        checker.getIndexInfoOfType(objType, ts.IndexKind.Number)));
            }
            // Checks whether a member expression is nullable or not regardless of it's previous node.
            //  Example:
            //  ```
            //  // 'bar' is nullable if 'foo' is null.
            //  // but this function checks regardless of 'foo' type, so returns 'true'.
            //  declare const foo: { bar : { baz: string } } | null
            //  foo?.bar;
            //  ```
            function isNullableOriginFromPrev(node) {
                const prevType = getNodeType(node.object);
                const property = node.property;
                if (prevType.isUnion() && util_1.isIdentifier(property)) {
                    const isOwnNullable = prevType.types.some(type => {
                        if (node.computed) {
                            const propertyType = getNodeType(node.property);
                            return isNullablePropertyType(type, propertyType);
                        }
                        const propType = util_1.getTypeOfPropertyOfName(checker, type, property.name);
                        return propType && util_1.isNullableType(propType, { allowUndefined: true });
                    });
                    return (!isOwnNullable && util_1.isNullableType(prevType, { allowUndefined: true }));
                }
                return false;
            }
            function isOptionableExpression(node) {
                const type = getNodeType(node);
                const isOwnNullable = node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression
                    ? !isNullableOriginFromPrev(node)
                    : true;
                return (util_1.isTypeAnyType(type) ||
                    util_1.isTypeUnknownType(type) ||
                    (util_1.isNullableType(type, { allowUndefined: true }) && isOwnNullable));
            }
            function checkOptionalChain(node, beforeOperator, fix) {
                // We only care if this step in the chain is optional. If just descend
                // from an optional chain, then that's fine.
                if (!node.optional) {
                    return;
                }
                // Since typescript array index signature types don't represent the
                //  possibility of out-of-bounds access, if we're indexing into an array
                //  just skip the check, to avoid false positives
                if (optionChainContainsArrayIndex(node)) {
                    return;
                }
                const nodeToCheck = node.type === experimental_utils_1.AST_NODE_TYPES.CallExpression ? node.callee : node.object;
                if (isOptionableExpression(nodeToCheck)) {
                    return;
                }
                const questionDotOperator = util_1.nullThrows(sourceCode.getTokenAfter(beforeOperator, token => token.type === experimental_utils_1.AST_TOKEN_TYPES.Punctuator && token.value === '?.'), util_1.NullThrowsReasons.MissingToken('operator', node.type));
                context.report({
                    node,
                    loc: questionDotOperator.loc,
                    messageId: 'neverOptionalChain',
                    fix(fixer) {
                        return fixer.replaceText(questionDotOperator, fix);
                    },
                });
            }
            function checkOptionalMemberExpression(node) {
                checkOptionalChain(node, node.object, node.computed ? '' : '.');
            }
            function checkOptionalCallExpression(node) {
                checkOptionalChain(node, node.callee, '');
            }
            return {
                BinaryExpression: checkIfBinaryExpressionIsNecessaryConditional,
                CallExpression: checkCallExpression,
                ConditionalExpression: (node) => checkNode(node.test),
                DoWhileStatement: checkIfLoopIsNecessaryConditional,
                ForStatement: checkIfLoopIsNecessaryConditional,
                IfStatement: (node) => checkNode(node.test),
                LogicalExpression: checkLogicalExpressionForUnnecessaryConditionals,
                WhileStatement: checkIfLoopIsNecessaryConditional,
                'MemberExpression[optional = true]': checkOptionalMemberExpression,
                'CallExpression[optional = true]': checkOptionalCallExpression,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unnecessary-qualifier.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const ts = __importStar(require(21) /* typescript */);
    const tsutils = __importStar(require(916) /* tsutils */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-unnecessary-qualifier',
        meta: {
            docs: {
                category: 'Best Practices',
                description: 'Warns when a namespace qualifier is unnecessary',
                recommended: false,
                requiresTypeChecking: true,
            },
            fixable: 'code',
            messages: {
                unnecessaryQualifier: "Qualifier is unnecessary since '{{ name }}' is in scope.",
            },
            schema: [],
            type: 'suggestion',
        },
        defaultOptions: [],
        create(context) {
            const namespacesInScope = [];
            let currentFailedNamespaceExpression = null;
            const parserServices = util.getParserServices(context);
            const esTreeNodeToTSNodeMap = parserServices.esTreeNodeToTSNodeMap;
            const program = parserServices.program;
            const checker = program.getTypeChecker();
            const sourceCode = context.getSourceCode();
            function tryGetAliasedSymbol(symbol, checker) {
                return tsutils.isSymbolFlagSet(symbol, ts.SymbolFlags.Alias)
                    ? checker.getAliasedSymbol(symbol)
                    : null;
            }
            function symbolIsNamespaceInScope(symbol) {
                var _a;
                const symbolDeclarations = (_a = symbol.getDeclarations()) !== null && _a !== void 0 ? _a : [];
                if (symbolDeclarations.some(decl => namespacesInScope.some(ns => ns === decl))) {
                    return true;
                }
                const alias = tryGetAliasedSymbol(symbol, checker);
                return alias !== null && symbolIsNamespaceInScope(alias);
            }
            function getSymbolInScope(node, flags, name) {
                // TODO:PERF `getSymbolsInScope` gets a long list. Is there a better way?
                const scope = checker.getSymbolsInScope(node, flags);
                return scope.find(scopeSymbol => scopeSymbol.name === name);
            }
            function symbolsAreEqual(accessed, inScope) {
                return accessed === checker.getExportSymbolOfSymbol(inScope);
            }
            function qualifierIsUnnecessary(qualifier, name) {
                const tsQualifier = esTreeNodeToTSNodeMap.get(qualifier);
                const tsName = esTreeNodeToTSNodeMap.get(name);
                const namespaceSymbol = checker.getSymbolAtLocation(tsQualifier);
                if (typeof namespaceSymbol === 'undefined' ||
                    !symbolIsNamespaceInScope(namespaceSymbol)) {
                    return false;
                }
                const accessedSymbol = checker.getSymbolAtLocation(tsName);
                if (typeof accessedSymbol === 'undefined') {
                    return false;
                }
                // If the symbol in scope is different, the qualifier is necessary.
                const fromScope = getSymbolInScope(tsQualifier, accessedSymbol.flags, sourceCode.getText(name));
                return (typeof fromScope === 'undefined' ||
                    symbolsAreEqual(accessedSymbol, fromScope));
            }
            function visitNamespaceAccess(node, qualifier, name) {
                // Only look for nested qualifier errors if we didn't already fail on the outer qualifier.
                if (!currentFailedNamespaceExpression &&
                    qualifierIsUnnecessary(qualifier, name)) {
                    currentFailedNamespaceExpression = node;
                    context.report({
                        node: qualifier,
                        messageId: 'unnecessaryQualifier',
                        data: {
                            name: sourceCode.getText(name),
                        },
                        fix(fixer) {
                            return fixer.removeRange([qualifier.range[0], name.range[0]]);
                        },
                    });
                }
            }
            function enterDeclaration(node) {
                namespacesInScope.push(esTreeNodeToTSNodeMap.get(node));
            }
            function exitDeclaration() {
                namespacesInScope.pop();
            }
            function resetCurrentNamespaceExpression(node) {
                if (node === currentFailedNamespaceExpression) {
                    currentFailedNamespaceExpression = null;
                }
            }
            function isPropertyAccessExpression(node) {
                return node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression && !node.computed;
            }
            function isEntityNameExpression(node) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.Identifier ||
                    (isPropertyAccessExpression(node) &&
                        isEntityNameExpression(node.object)));
            }
            return {
                TSModuleDeclaration: enterDeclaration,
                TSEnumDeclaration: enterDeclaration,
                'ExportNamedDeclaration[declaration.type="TSModuleDeclaration"]': enterDeclaration,
                'ExportNamedDeclaration[declaration.type="TSEnumDeclaration"]': enterDeclaration,
                'TSModuleDeclaration:exit': exitDeclaration,
                'TSEnumDeclaration:exit': exitDeclaration,
                'ExportNamedDeclaration[declaration.type="TSModuleDeclaration"]:exit': exitDeclaration,
                'ExportNamedDeclaration[declaration.type="TSEnumDeclaration"]:exit': exitDeclaration,
                TSQualifiedName(node) {
                    visitNamespaceAccess(node, node.left, node.right);
                },
                'MemberExpression[computed=false]': function (node) {
                    const property = node.property;
                    if (isEntityNameExpression(node.object)) {
                        visitNamespaceAccess(node, node.object, property);
                    }
                },
                'TSQualifiedName:exit': resetCurrentNamespaceExpression,
                'MemberExpression:exit': resetCurrentNamespaceExpression,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unnecessary-type-arguments.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const tsutils = __importStar(require(916) /* tsutils */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    const util_1 = require(906) /* ../util */;
    exports.default = util.createRule({
        name: 'no-unnecessary-type-arguments',
        meta: {
            docs: {
                description: 'Enforces that type arguments will not be used if not required',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            fixable: 'code',
            messages: {
                unnecessaryTypeParameter: 'This is the default value for this type parameter, so it can be omitted.',
            },
            schema: [],
            type: 'suggestion',
        },
        defaultOptions: [],
        create(context) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const sourceCode = context.getSourceCode();
            function checkTSArgsAndParameters(esParameters, typeParameters) {
                // Just check the last one. Must specify previous type parameters if the last one is specified.
                const i = esParameters.params.length - 1;
                const arg = esParameters.params[i];
                const param = typeParameters[i];
                // TODO: would like checker.areTypesEquivalent. https://github.com/Microsoft/TypeScript/issues/13502
                if (!(param === null || param === void 0 ? void 0 : param.default) ||
                    param.default.getText() !== sourceCode.getText(arg)) {
                    return;
                }
                context.report({
                    node: arg,
                    messageId: 'unnecessaryTypeParameter',
                    fix: fixer => fixer.removeRange(i === 0
                        ? esParameters.range
                        : [esParameters.params[i - 1].range[1], arg.range[1]]),
                });
            }
            return {
                TSTypeParameterInstantiation(node) {
                    const expression = parserServices.esTreeNodeToTSNodeMap.get(node);
                    const typeParameters = getTypeParametersFromNode(expression, checker);
                    if (typeParameters) {
                        checkTSArgsAndParameters(node, typeParameters);
                    }
                },
            };
        },
    });
    function getTypeParametersFromNode(node, checker) {
        if (ts.isExpressionWithTypeArguments(node)) {
            return getTypeParametersFromType(node.expression, checker);
        }
        if (ts.isTypeReferenceNode(node)) {
            return getTypeParametersFromType(node.typeName, checker);
        }
        if (ts.isCallExpression(node) || ts.isNewExpression(node)) {
            return getTypeParametersFromCall(node, checker);
        }
        return undefined;
    }
    function getTypeParametersFromType(type, checker) {
        const symAtLocation = checker.getSymbolAtLocation(type);
        if (!symAtLocation) {
            return undefined;
        }
        const sym = getAliasedSymbol(symAtLocation, checker);
        const declarations = sym.getDeclarations();
        if (!declarations) {
            return undefined;
        }
        return util_1.findFirstResult(declarations, decl => ts.isClassLike(decl) ||
            ts.isTypeAliasDeclaration(decl) ||
            ts.isInterfaceDeclaration(decl)
            ? decl.typeParameters
            : undefined);
    }
    function getTypeParametersFromCall(node, checker) {
        const sig = checker.getResolvedSignature(node);
        const sigDecl = sig === null || sig === void 0 ? void 0 : sig.getDeclaration();
        if (!sigDecl) {
            return ts.isNewExpression(node)
                ? getTypeParametersFromType(node.expression, checker)
                : undefined;
        }
        return sigDecl.typeParameters;
    }
    function getAliasedSymbol(symbol, checker) {
        return tsutils.isSymbolFlagSet(symbol, ts.SymbolFlags.Alias)
            ? checker.getAliasedSymbol(symbol)
            : symbol;
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unnecessary-type-assertion.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils_1 = require(916) /* tsutils */;
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-unnecessary-type-assertion',
        meta: {
            docs: {
                description: 'Warns if a type assertion does not change the type of an expression',
                category: 'Best Practices',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            fixable: 'code',
            messages: {
                unnecessaryAssertion: 'This assertion is unnecessary since it does not change the type of the expression.',
                contextuallyUnnecessary: 'This assertion is unnecessary since the receiver accepts the original type of the expression.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        typesToIgnore: {
                            type: 'array',
                            items: {
                                type: 'string',
                            },
                        },
                    },
                },
            ],
            type: 'suggestion',
        },
        defaultOptions: [{}],
        create(context, [options]) {
            const sourceCode = context.getSourceCode();
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const compilerOptions = parserServices.program.getCompilerOptions();
            /**
             * Sometimes tuple types don't have ObjectFlags.Tuple set, like when they're being matched against an inferred type.
             * So, in addition, check if there are integer properties 0..n and no other numeric keys
             */
            function couldBeTupleType(type) {
                const properties = type.getProperties();
                if (properties.length === 0) {
                    return false;
                }
                let i = 0;
                for (; i < properties.length; ++i) {
                    const name = properties[i].name;
                    if (String(i) !== name) {
                        if (i === 0) {
                            // if there are no integer properties, this is not a tuple
                            return false;
                        }
                        break;
                    }
                }
                for (; i < properties.length; ++i) {
                    if (String(+properties[i].name) === properties[i].name) {
                        return false; // if there are any other numeric properties, this is not a tuple
                    }
                }
                return true;
            }
            /**
             * Returns true if there's a chance the variable has been used before a value has been assigned to it
             */
            function isPossiblyUsedBeforeAssigned(node) {
                const declaration = util.getDeclaration(checker, node);
                if (!declaration) {
                    // don't know what the declaration is for some reason, so just assume the worst
                    return true;
                }
                if (
                // non-strict mode doesn't care about used before assigned errors
                tsutils_1.isStrictCompilerOptionEnabled(compilerOptions, 'strictNullChecks') &&
                    // ignore class properties as they are compile time guarded
                    // also ignore function arguments as they can't be used before defined
                    tsutils_1.isVariableDeclaration(declaration) &&
                    // is it `const x!: number`
                    declaration.initializer === undefined &&
                    declaration.exclamationToken === undefined &&
                    declaration.type !== undefined) {
                    // check if the defined variable type has changed since assignment
                    const declarationType = checker.getTypeFromTypeNode(declaration.type);
                    const type = util.getConstrainedTypeAtLocation(checker, node);
                    if (declarationType === type) {
                        // possibly used before assigned, so just skip it
                        // better to false negative and skip it, than false positive and fix to compile erroring code
                        //
                        // no better way to figure this out right now
                        // https://github.com/Microsoft/TypeScript/issues/31124
                        return true;
                    }
                }
                return false;
            }
            function isConstAssertion(node) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference &&
                    node.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    node.typeName.name === 'const');
            }
            return {
                TSNonNullExpression(node) {
                    var _a;
                    if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.AssignmentExpression &&
                        node.parent.operator === '=') {
                        if (node.parent.left === node) {
                            context.report({
                                node,
                                messageId: 'contextuallyUnnecessary',
                                fix(fixer) {
                                    return fixer.removeRange([
                                        node.expression.range[1],
                                        node.range[1],
                                    ]);
                                },
                            });
                        }
                        // for all other = assignments we ignore non-null checks
                        // this is because non-null assertions can change the type-flow of the code
                        // so whilst they might be unnecessary for the assignment - they are necessary
                        // for following code
                        return;
                    }
                    const originalNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                    const type = util.getConstrainedTypeAtLocation(checker, originalNode.expression);
                    if (!util.isNullableType(type)) {
                        if (isPossiblyUsedBeforeAssigned(originalNode.expression)) {
                            return;
                        }
                        context.report({
                            node,
                            messageId: 'unnecessaryAssertion',
                            fix(fixer) {
                                return fixer.removeRange([
                                    originalNode.expression.end,
                                    originalNode.end,
                                ]);
                            },
                        });
                    }
                    else {
                        // we know it's a nullable type
                        // so figure out if the variable is used in a place that accepts nullable types
                        const contextualType = util.getContextualType(checker, originalNode);
                        if (contextualType) {
                            // in strict mode you can't assign null to undefined, so we have to make sure that
                            // the two types share a nullable type
                            const typeIncludesUndefined = util.isTypeFlagSet(type, ts.TypeFlags.Undefined);
                            const typeIncludesNull = util.isTypeFlagSet(type, ts.TypeFlags.Null);
                            const contextualTypeIncludesUndefined = util.isTypeFlagSet(contextualType, ts.TypeFlags.Undefined);
                            const contextualTypeIncludesNull = util.isTypeFlagSet(contextualType, ts.TypeFlags.Null);
                            // make sure that the parent accepts the same types
                            // i.e. assigning `string | null | undefined` to `string | undefined` is invalid
                            const isValidUndefined = typeIncludesUndefined
                                ? contextualTypeIncludesUndefined
                                : true;
                            const isValidNull = typeIncludesNull
                                ? contextualTypeIncludesNull
                                : true;
                            if (isValidUndefined && isValidNull) {
                                context.report({
                                    node,
                                    messageId: 'contextuallyUnnecessary',
                                    fix(fixer) {
                                        return fixer.removeRange([
                                            originalNode.expression.end,
                                            originalNode.end,
                                        ]);
                                    },
                                });
                            }
                        }
                    }
                },
                'TSAsExpression, TSTypeAssertion'(node) {
                    var _a;
                    if (((_a = options.typesToIgnore) === null || _a === void 0 ? void 0 : _a.includes(sourceCode.getText(node.typeAnnotation))) ||
                        isConstAssertion(node.typeAnnotation)) {
                        return;
                    }
                    const originalNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                    const castType = checker.getTypeAtLocation(originalNode);
                    if (tsutils_1.isTypeFlagSet(castType, ts.TypeFlags.Literal) ||
                        (tsutils_1.isObjectType(castType) &&
                            (tsutils_1.isObjectFlagSet(castType, ts.ObjectFlags.Tuple) ||
                                couldBeTupleType(castType)))) {
                        // It's not always safe to remove a cast to a literal type or tuple
                        // type, as those types are sometimes widened without the cast.
                        return;
                    }
                    const uncastType = checker.getTypeAtLocation(originalNode.expression);
                    if (uncastType === castType) {
                        context.report({
                            node,
                            messageId: 'unnecessaryAssertion',
                            fix(fixer) {
                                if (originalNode.kind === ts.SyntaxKind.TypeAssertionExpression) {
                                    const closingAngleBracket = sourceCode.getTokenAfter(node.typeAnnotation);
                                    return (closingAngleBracket === null || closingAngleBracket === void 0 ? void 0 : closingAngleBracket.value) === '>'
                                        ? fixer.removeRange([
                                            node.range[0],
                                            closingAngleBracket.range[1],
                                        ])
                                        : null;
                                }
                                return fixer.removeRange([
                                    node.expression.range[1] + 1,
                                    node.range[1],
                                ]);
                            },
                        });
                    }
                    // TODO - add contextually unnecessary check for this
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unnecessary-type-constraint.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const semver = __importStar(require(1004) /* semver */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    const is3dot5 = semver.satisfies(ts.version, `>= 3.5.0 || >= 3.5.1-rc || >= 3.5.0-beta`, {
        includePrerelease: true,
    });
    const is3dot9 = is3dot5 &&
        semver.satisfies(ts.version, `>= 3.9.0 || >= 3.9.1-rc || >= 3.9.0-beta`, {
            includePrerelease: true,
        });
    exports.default = util.createRule({
        name: 'no-unnecessary-type-constraint',
        meta: {
            docs: {
                category: 'Best Practices',
                description: 'Disallows unnecessary constraints on generic types',
                recommended: false,
                suggestion: true,
            },
            fixable: 'code',
            messages: {
                unnecessaryConstraint: 'Constraining the generic type `{{name}}` to `{{constraint}}` does nothing and is unnecessary.',
            },
            schema: [],
            type: 'suggestion',
        },
        defaultOptions: [],
        create(context) {
            if (!is3dot5) {
                return {};
            }
            // In theory, we could use the type checker for more advanced constraint types...
            // ...but in practice, these types are rare, and likely not worth requiring type info.
            // https://github.com/typescript-eslint/typescript-eslint/pull/2516#discussion_r495731858
            const unnecessaryConstraints = is3dot9
                ? new Map([
                    [experimental_utils_1.AST_NODE_TYPES.TSAnyKeyword, 'any'],
                    [experimental_utils_1.AST_NODE_TYPES.TSUnknownKeyword, 'unknown'],
                ])
                : new Map([[experimental_utils_1.AST_NODE_TYPES.TSUnknownKeyword, 'unknown']]);
            const inJsx = context.getFilename().toLowerCase().endsWith('tsx');
            const checkNode = (node, inArrowFunction) => {
                const constraint = unnecessaryConstraints.get(node.constraint.type);
                if (constraint) {
                    context.report({
                        data: {
                            constraint,
                            name: node.name.name,
                        },
                        fix(fixer) {
                            return fixer.replaceTextRange([node.name.range[1], node.constraint.range[1]], inArrowFunction && inJsx ? ',' : '');
                        },
                        messageId: 'unnecessaryConstraint',
                        node,
                    });
                }
            };
            return {
                ':not(ArrowFunctionExpression) > TSTypeParameterDeclaration > TSTypeParameter[constraint]'(node) {
                    checkNode(node, false);
                },
                'ArrowFunctionExpression > TSTypeParameterDeclaration > TSTypeParameter[constraint]'(node) {
                    checkNode(node, true);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unsafe-argument.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    class FunctionSignature {
        constructor(paramTypes, restType) {
            this.paramTypes = paramTypes;
            this.restType = restType;
            this.parameterTypeIndex = 0;
            this.hasConsumedArguments = false;
        }
        static create(checker, tsNode) {
            var _a;
            const signature = checker.getResolvedSignature(tsNode);
            if (!signature) {
                return null;
            }
            const paramTypes = [];
            let restType = null;
            const parameters = signature.getParameters();
            for (let i = 0; i < parameters.length; i += 1) {
                const param = parameters[i];
                const type = checker.getTypeOfSymbolAtLocation(param, tsNode);
                const decl = (_a = param.getDeclarations()) === null || _a === void 0 ? void 0 : _a[0];
                if (decl && ts.isParameter(decl) && decl.dotDotDotToken) {
                    // is a rest param
                    if (checker.isArrayType(type)) {
                        restType = {
                            type: checker.getTypeArguments(type)[0],
                            kind: 0 /* Array */,
                            index: i,
                        };
                    }
                    else if (checker.isTupleType(type)) {
                        restType = {
                            typeArguments: checker.getTypeArguments(type),
                            kind: 1 /* Tuple */,
                            index: i,
                        };
                    }
                    else {
                        restType = {
                            type,
                            kind: 2 /* Other */,
                            index: i,
                        };
                    }
                    break;
                }
                paramTypes.push(type);
            }
            return new this(paramTypes, restType);
        }
        getNextParameterType() {
            const index = this.parameterTypeIndex;
            this.parameterTypeIndex += 1;
            if (index >= this.paramTypes.length || this.hasConsumedArguments) {
                if (this.restType == null) {
                    return null;
                }
                switch (this.restType.kind) {
                    case 1 /* Tuple */: {
                        const typeArguments = this.restType.typeArguments;
                        if (this.hasConsumedArguments) {
                            // all types consumed by a rest - just assume it's the last type
                            // there is one edge case where this is wrong, but we ignore it because
                            // it's rare and really complicated to handle
                            // eg: function foo(...a: [number, ...string[], number])
                            return typeArguments[typeArguments.length - 1];
                        }
                        const typeIndex = index - this.restType.index;
                        if (typeIndex >= typeArguments.length) {
                            return typeArguments[typeArguments.length - 1];
                        }
                        return typeArguments[typeIndex];
                    }
                    case 0 /* Array */:
                    case 2 /* Other */:
                        return this.restType.type;
                }
            }
            return this.paramTypes[index];
        }
        consumeRemainingArguments() {
            this.hasConsumedArguments = true;
        }
    }
    exports.default = util.createRule({
        name: 'no-unsafe-argument',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows calling an function with an any type value',
                category: 'Possible Errors',
                // TODO - enable this with next breaking
                recommended: false,
                requiresTypeChecking: true,
            },
            messages: {
                unsafeArgument: 'Unsafe argument of type `{{sender}}` assigned to a parameter of type `{{receiver}}`.',
                unsafeTupleSpread: 'Unsafe spread of a tuple type. The {{index}} element is of type `{{sender}}` and is assigned to a parameter of type `{{reciever}}`.',
                unsafeArraySpread: 'Unsafe spread of an `any` array type.',
                unsafeSpread: 'Unsafe spread of an `any` type.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const { program, esTreeNodeToTSNodeMap } = util.getParserServices(context);
            const checker = program.getTypeChecker();
            return {
                'CallExpression, NewExpression'(node) {
                    if (node.arguments.length === 0) {
                        return;
                    }
                    // ignore any-typed calls as these are caught by no-unsafe-call
                    if (util.isTypeAnyType(checker.getTypeAtLocation(esTreeNodeToTSNodeMap.get(node.callee)))) {
                        return;
                    }
                    const tsNode = esTreeNodeToTSNodeMap.get(node);
                    const signature = FunctionSignature.create(checker, tsNode);
                    if (!signature) {
                        return;
                    }
                    for (let i = 0; i < node.arguments.length; i += 1) {
                        const argument = node.arguments[i];
                        switch (argument.type) {
                            // spreads consume
                            case experimental_utils_1.AST_NODE_TYPES.SpreadElement: {
                                const spreadArgType = checker.getTypeAtLocation(esTreeNodeToTSNodeMap.get(argument.argument));
                                if (util.isTypeAnyType(spreadArgType)) {
                                    // foo(...any)
                                    context.report({
                                        node: argument,
                                        messageId: 'unsafeSpread',
                                    });
                                }
                                else if (util.isTypeAnyArrayType(spreadArgType, checker)) {
                                    // foo(...any[])
                                    // TODO - we could break down the spread and compare the array type against each argument
                                    context.report({
                                        node: argument,
                                        messageId: 'unsafeArraySpread',
                                    });
                                }
                                else if (checker.isTupleType(spreadArgType)) {
                                    // foo(...[tuple1, tuple2])
                                    const spreadTypeArguments = checker.getTypeArguments(spreadArgType);
                                    for (let j = 0; j < spreadTypeArguments.length; j += 1) {
                                        const tupleType = spreadTypeArguments[j];
                                        const parameterType = signature.getNextParameterType();
                                        if (parameterType == null) {
                                            continue;
                                        }
                                        const result = util.isUnsafeAssignment(tupleType, parameterType, checker, 
                                        // we can't pass the individual tuple members in here as this will most likely be a spread variable
                                        // not a spread array
                                        null);
                                        if (result) {
                                            context.report({
                                                node: argument,
                                                messageId: 'unsafeTupleSpread',
                                                data: {
                                                    sender: checker.typeToString(tupleType),
                                                    receiver: checker.typeToString(parameterType),
                                                },
                                            });
                                        }
                                    }
                                    if (spreadArgType.target.hasRestElement) {
                                        // the last element was a rest - so all remaining defined arguments can be considered "consumed"
                                        // all remaining arguments should be compared against the rest type (if one exists)
                                        signature.consumeRemainingArguments();
                                    }
                                }
                                else {
                                    // something that's iterable
                                    // handling this will be pretty complex - so we ignore it for now
                                    // TODO - handle generic iterable case
                                }
                                break;
                            }
                            default: {
                                const parameterType = signature.getNextParameterType();
                                if (parameterType == null) {
                                    continue;
                                }
                                const argumentType = checker.getTypeAtLocation(esTreeNodeToTSNodeMap.get(argument));
                                const result = util.isUnsafeAssignment(argumentType, parameterType, checker, argument);
                                if (result) {
                                    context.report({
                                        node: argument,
                                        messageId: 'unsafeArgument',
                                        data: {
                                            sender: checker.typeToString(argumentType),
                                            receiver: checker.typeToString(parameterType),
                                        },
                                    });
                                }
                            }
                        }
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unsafe-assignment.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const util = __importStar(require(906) /* ../util */);
    const util_1 = require(906) /* ../util */;
    exports.default = util.createRule({
        name: 'no-unsafe-assignment',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows assigning any to variables and properties',
                category: 'Possible Errors',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                anyAssignment: 'Unsafe assignment of an `any` value.',
                anyAssignmentThis: [
                    'Unsafe assignment of an `any` value. `this` is typed as `any`.',
                    'You can try to fix this by turning on the `noImplicitThis` compiler option, or adding a `this` parameter to the function.',
                ].join('\n'),
                unsafeArrayPattern: 'Unsafe array destructuring of an `any` array value.',
                unsafeArrayPatternFromTuple: 'Unsafe array destructuring of a tuple element with an `any` value.',
                unsafeAssignment: 'Unsafe assignment of type {{sender}} to a variable of type {{receiver}}.',
                unsafeArraySpread: 'Unsafe spread of an `any` value in an array.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const { program, esTreeNodeToTSNodeMap } = util.getParserServices(context);
            const checker = program.getTypeChecker();
            const compilerOptions = program.getCompilerOptions();
            const isNoImplicitThis = tsutils.isStrictCompilerOptionEnabled(compilerOptions, 'noImplicitThis');
            // returns true if the assignment reported
            function checkArrayDestructureHelper(receiverNode, senderNode) {
                if (receiverNode.type !== experimental_utils_1.AST_NODE_TYPES.ArrayPattern) {
                    return false;
                }
                const senderTsNode = esTreeNodeToTSNodeMap.get(senderNode);
                const senderType = checker.getTypeAtLocation(senderTsNode);
                return checkArrayDestructure(receiverNode, senderType, senderTsNode);
            }
            // returns true if the assignment reported
            function checkArrayDestructure(receiverNode, senderType, senderNode) {
                // any array
                // const [x] = ([] as any[]);
                if (util.isTypeAnyArrayType(senderType, checker)) {
                    context.report({
                        node: receiverNode,
                        messageId: 'unsafeArrayPattern',
                    });
                    return false;
                }
                if (!checker.isTupleType(senderType)) {
                    return true;
                }
                const tupleElements = util.getTypeArguments(senderType, checker);
                // tuple with any
                // const [x] = [1 as any];
                let didReport = false;
                for (let receiverIndex = 0; receiverIndex < receiverNode.elements.length; receiverIndex += 1) {
                    const receiverElement = receiverNode.elements[receiverIndex];
                    if (!receiverElement) {
                        continue;
                    }
                    if (receiverElement.type === experimental_utils_1.AST_NODE_TYPES.RestElement) {
                        // don't handle rests as they're not a 1:1 assignment
                        continue;
                    }
                    const senderType = tupleElements[receiverIndex];
                    if (!senderType) {
                        continue;
                    }
                    // check for the any type first so we can handle [[[x]]] = [any]
                    if (util.isTypeAnyType(senderType)) {
                        context.report({
                            node: receiverElement,
                            messageId: 'unsafeArrayPatternFromTuple',
                        });
                        // we want to report on every invalid element in the tuple
                        didReport = true;
                    }
                    else if (receiverElement.type === experimental_utils_1.AST_NODE_TYPES.ArrayPattern) {
                        didReport = checkArrayDestructure(receiverElement, senderType, senderNode);
                    }
                    else if (receiverElement.type === experimental_utils_1.AST_NODE_TYPES.ObjectPattern) {
                        didReport = checkObjectDestructure(receiverElement, senderType, senderNode);
                    }
                }
                return didReport;
            }
            // returns true if the assignment reported
            function checkObjectDestructureHelper(receiverNode, senderNode) {
                if (receiverNode.type !== experimental_utils_1.AST_NODE_TYPES.ObjectPattern) {
                    return false;
                }
                const senderTsNode = esTreeNodeToTSNodeMap.get(senderNode);
                const senderType = checker.getTypeAtLocation(senderTsNode);
                return checkObjectDestructure(receiverNode, senderType, senderTsNode);
            }
            // returns true if the assignment reported
            function checkObjectDestructure(receiverNode, senderType, senderNode) {
                const properties = new Map(senderType
                    .getProperties()
                    .map(property => [
                    property.getName(),
                    checker.getTypeOfSymbolAtLocation(property, senderNode),
                ]));
                let didReport = false;
                for (let receiverIndex = 0; receiverIndex < receiverNode.properties.length; receiverIndex += 1) {
                    const receiverProperty = receiverNode.properties[receiverIndex];
                    if (receiverProperty.type === experimental_utils_1.AST_NODE_TYPES.RestElement) {
                        // don't bother checking rest
                        continue;
                    }
                    let key;
                    if (receiverProperty.computed === false) {
                        key =
                            receiverProperty.key.type === experimental_utils_1.AST_NODE_TYPES.Identifier
                                ? receiverProperty.key.name
                                : String(receiverProperty.key.value);
                    }
                    else if (receiverProperty.key.type === experimental_utils_1.AST_NODE_TYPES.Literal) {
                        key = String(receiverProperty.key.value);
                    }
                    else if (receiverProperty.key.type === experimental_utils_1.AST_NODE_TYPES.TemplateLiteral &&
                        receiverProperty.key.quasis.length === 1) {
                        key = String(receiverProperty.key.quasis[0].value.cooked);
                    }
                    else {
                        // can't figure out the name, so skip it
                        continue;
                    }
                    const senderType = properties.get(key);
                    if (!senderType) {
                        continue;
                    }
                    // check for the any type first so we can handle {x: {y: z}} = {x: any}
                    if (util.isTypeAnyType(senderType)) {
                        context.report({
                            node: receiverProperty.value,
                            messageId: 'unsafeArrayPatternFromTuple',
                        });
                        didReport = true;
                    }
                    else if (receiverProperty.value.type === experimental_utils_1.AST_NODE_TYPES.ArrayPattern) {
                        didReport = checkArrayDestructure(receiverProperty.value, senderType, senderNode);
                    }
                    else if (receiverProperty.value.type === experimental_utils_1.AST_NODE_TYPES.ObjectPattern) {
                        didReport = checkObjectDestructure(receiverProperty.value, senderType, senderNode);
                    }
                }
                return didReport;
            }
            // returns true if the assignment reported
            function checkAssignment(receiverNode, senderNode, reportingNode, comparisonType) {
                var _a;
                const receiverTsNode = esTreeNodeToTSNodeMap.get(receiverNode);
                const receiverType = comparisonType === 2 /* Contextual */
                    ? (_a = util.getContextualType(checker, receiverTsNode)) !== null && _a !== void 0 ? _a : checker.getTypeAtLocation(receiverTsNode)
                    : checker.getTypeAtLocation(receiverTsNode);
                const senderType = checker.getTypeAtLocation(esTreeNodeToTSNodeMap.get(senderNode));
                if (util.isTypeAnyType(senderType)) {
                    // handle cases when we assign any ==> unknown.
                    if (util.isTypeUnknownType(receiverType)) {
                        return false;
                    }
                    let messageId = 'anyAssignment';
                    if (!isNoImplicitThis) {
                        // `var foo = this`
                        const thisExpression = util_1.getThisExpression(senderNode);
                        if (thisExpression &&
                            util.isTypeAnyType(util.getConstrainedTypeAtLocation(checker, esTreeNodeToTSNodeMap.get(thisExpression)))) {
                            messageId = 'anyAssignmentThis';
                        }
                    }
                    context.report({
                        node: reportingNode,
                        messageId,
                    });
                    return true;
                }
                if (comparisonType === 0 /* None */) {
                    return false;
                }
                const result = util.isUnsafeAssignment(senderType, receiverType, checker, senderNode);
                if (!result) {
                    return false;
                }
                const { sender, receiver } = result;
                context.report({
                    node: reportingNode,
                    messageId: 'unsafeAssignment',
                    data: {
                        sender: checker.typeToString(sender),
                        receiver: checker.typeToString(receiver),
                    },
                });
                return true;
            }
            function getComparisonType(typeAnnotation) {
                return typeAnnotation
                    ? // if there's a type annotation, we can do a comparison
                        1 /* Basic */
                    : // no type annotation means the variable's type will just be inferred, thus equal
                        0 /* None */;
            }
            return {
                'VariableDeclarator[init != null]'(node) {
                    const init = util.nullThrows(node.init, util.NullThrowsReasons.MissingToken(node.type, 'init'));
                    let didReport = checkAssignment(node.id, init, node, getComparisonType(node.id.typeAnnotation));
                    if (!didReport) {
                        didReport = checkArrayDestructureHelper(node.id, init);
                    }
                    if (!didReport) {
                        checkObjectDestructureHelper(node.id, init);
                    }
                },
                'ClassProperty[value != null]'(node) {
                    checkAssignment(node.key, node.value, node, getComparisonType(node.typeAnnotation));
                },
                'AssignmentExpression[operator = "="], AssignmentPattern'(node) {
                    let didReport = checkAssignment(node.left, node.right, node, 1 /* Basic */);
                    if (!didReport) {
                        didReport = checkArrayDestructureHelper(node.left, node.right);
                    }
                    if (!didReport) {
                        checkObjectDestructureHelper(node.left, node.right);
                    }
                },
                // object pattern props are checked via assignments
                ':not(ObjectPattern) > Property'(node) {
                    if (node.value.type === experimental_utils_1.AST_NODE_TYPES.AssignmentPattern ||
                        node.value.type === experimental_utils_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression) {
                        // handled by other selector
                        return;
                    }
                    checkAssignment(node.key, node.value, node, 2 /* Contextual */);
                },
                'ArrayExpression > SpreadElement'(node) {
                    const resetNode = esTreeNodeToTSNodeMap.get(node.argument);
                    const restType = checker.getTypeAtLocation(resetNode);
                    if (util.isTypeAnyType(restType) ||
                        util.isTypeAnyArrayType(restType, checker)) {
                        context.report({
                            node: node,
                            messageId: 'unsafeArraySpread',
                        });
                    }
                },
                'JSXAttribute[value != null]'(node) {
                    const value = util.nullThrows(node.value, util.NullThrowsReasons.MissingToken(node.type, 'value'));
                    if (value.type !== experimental_utils_1.AST_NODE_TYPES.JSXExpressionContainer ||
                        value.expression.type === experimental_utils_1.AST_NODE_TYPES.JSXEmptyExpression) {
                        return;
                    }
                    checkAssignment(node.name, value.expression, value.expression, 2 /* Contextual */);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unsafe-call.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const tsutils = __importStar(require(916) /* tsutils */);
    const util = __importStar(require(906) /* ../util */);
    const util_1 = require(906) /* ../util */;
    exports.default = util.createRule({
        name: 'no-unsafe-call',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows calling an any type value',
                category: 'Possible Errors',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                unsafeCall: 'Unsafe call of an `any` typed value.',
                unsafeCallThis: [
                    'Unsafe call of an `any` typed value. `this` is typed as `any`.',
                    'You can try to fix this by turning on the `noImplicitThis` compiler option, or adding a `this` parameter to the function.',
                ].join('\n'),
                unsafeNew: 'Unsafe construction of an any type value.',
                unsafeTemplateTag: 'Unsafe any typed template tag.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const { program, esTreeNodeToTSNodeMap } = util.getParserServices(context);
            const checker = program.getTypeChecker();
            const compilerOptions = program.getCompilerOptions();
            const isNoImplicitThis = tsutils.isStrictCompilerOptionEnabled(compilerOptions, 'noImplicitThis');
            function checkCall(node, reportingNode, messageId) {
                const tsNode = esTreeNodeToTSNodeMap.get(node);
                const type = util.getConstrainedTypeAtLocation(checker, tsNode);
                if (util.isTypeAnyType(type)) {
                    if (!isNoImplicitThis) {
                        // `this()` or `this.foo()` or `this.foo[bar]()`
                        const thisExpression = util_1.getThisExpression(node);
                        if (thisExpression &&
                            util.isTypeAnyType(util.getConstrainedTypeAtLocation(checker, esTreeNodeToTSNodeMap.get(thisExpression)))) {
                            messageId = 'unsafeCallThis';
                        }
                    }
                    context.report({
                        node: reportingNode,
                        messageId: messageId,
                    });
                }
            }
            return {
                'CallExpression > *.callee'(node) {
                    checkCall(node, node, 'unsafeCall');
                },
                NewExpression(node) {
                    checkCall(node.callee, node, 'unsafeNew');
                },
                'TaggedTemplateExpression > *.tag'(node) {
                    checkCall(node, node, 'unsafeTemplateTag');
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unsafe-member-access.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const util = __importStar(require(906) /* ../util */);
    const util_1 = require(906) /* ../util */;
    exports.default = util.createRule({
        name: 'no-unsafe-member-access',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows member access on any typed variables',
                category: 'Possible Errors',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                unsafeMemberExpression: 'Unsafe member access {{property}} on an `any` value.',
                unsafeThisMemberExpression: [
                    'Unsafe member access {{property}} on an `any` value. `this` is typed as `any`.',
                    'You can try to fix this by turning on the `noImplicitThis` compiler option, or adding a `this` parameter to the function.',
                ].join('\n'),
                unsafeComputedMemberAccess: 'Computed name {{property}} resolves to an any value.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const { program, esTreeNodeToTSNodeMap } = util.getParserServices(context);
            const checker = program.getTypeChecker();
            const compilerOptions = program.getCompilerOptions();
            const isNoImplicitThis = tsutils.isStrictCompilerOptionEnabled(compilerOptions, 'noImplicitThis');
            const sourceCode = context.getSourceCode();
            const stateCache = new Map();
            function checkMemberExpression(node) {
                const cachedState = stateCache.get(node);
                if (cachedState) {
                    return cachedState;
                }
                if (node.object.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression) {
                    const objectState = checkMemberExpression(node.object);
                    if (objectState === 1 /* Unsafe */) {
                        // if the object is unsafe, we know this will be unsafe as well
                        // we don't need to report, as we have already reported on the inner member expr
                        stateCache.set(node, objectState);
                        return objectState;
                    }
                }
                const tsNode = esTreeNodeToTSNodeMap.get(node.object);
                const type = checker.getTypeAtLocation(tsNode);
                const state = util.isTypeAnyType(type) ? 1 /* Unsafe */ : 2 /* Safe */;
                stateCache.set(node, state);
                if (state === 1 /* Unsafe */) {
                    const propertyName = sourceCode.getText(node.property);
                    let messageId = 'unsafeMemberExpression';
                    if (!isNoImplicitThis) {
                        // `this.foo` or `this.foo[bar]`
                        const thisExpression = util_1.getThisExpression(node);
                        if (thisExpression &&
                            util.isTypeAnyType(util.getConstrainedTypeAtLocation(checker, esTreeNodeToTSNodeMap.get(thisExpression)))) {
                            messageId = 'unsafeThisMemberExpression';
                        }
                    }
                    context.report({
                        node,
                        messageId,
                        data: {
                            property: node.computed ? `[${propertyName}]` : `.${propertyName}`,
                        },
                    });
                }
                return state;
            }
            return {
                // ignore MemberExpression if it's parent is TSClassImplements or TSInterfaceHeritage
                ':not(TSClassImplements, TSInterfaceHeritage) > MemberExpression': checkMemberExpression,
                'MemberExpression[computed = true] > *.property'(node) {
                    if (
                    // x[1]
                    node.type === experimental_utils_1.AST_NODE_TYPES.Literal ||
                        // x[1++] x[++x] etc
                        // FUN FACT - **all** update expressions return type number, regardless of the argument's type,
                        // because JS engines return NaN if there the argument is not a number.
                        node.type === experimental_utils_1.AST_NODE_TYPES.UpdateExpression) {
                        // perf optimizations - literals can obviously never be `any`
                        return;
                    }
                    const tsNode = esTreeNodeToTSNodeMap.get(node);
                    const type = checker.getTypeAtLocation(tsNode);
                    if (util.isTypeAnyType(type)) {
                        const propertyName = sourceCode.getText(node);
                        context.report({
                            node,
                            messageId: 'unsafeComputedMemberAccess',
                            data: {
                                property: `[${propertyName}]`,
                            },
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unsafe-return.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const util = __importStar(require(906) /* ../util */);
    const util_1 = require(906) /* ../util */;
    exports.default = util.createRule({
        name: 'no-unsafe-return',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows returning any from a function',
                category: 'Possible Errors',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                unsafeReturn: 'Unsafe return of an `{{type}}` typed value.',
                unsafeReturnThis: [
                    'Unsafe return of an `{{type}}` typed value. `this` is typed as `any`.',
                    'You can try to fix this by turning on the `noImplicitThis` compiler option, or adding a `this` parameter to the function.',
                ].join('\n'),
                unsafeReturnAssignment: 'Unsafe return of type `{{sender}}` from function with return type `{{receiver}}`.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const { program, esTreeNodeToTSNodeMap } = util.getParserServices(context);
            const checker = program.getTypeChecker();
            const compilerOptions = program.getCompilerOptions();
            const isNoImplicitThis = tsutils.isStrictCompilerOptionEnabled(compilerOptions, 'noImplicitThis');
            function getParentFunctionNode(node) {
                let current = node.parent;
                while (current) {
                    if (current.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression ||
                        current.type === experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration ||
                        current.type === experimental_utils_1.AST_NODE_TYPES.FunctionExpression) {
                        return current;
                    }
                    current = current.parent;
                }
                // this shouldn't happen in correct code, but someone may attempt to parse bad code
                // the parser won't error, so we shouldn't throw here
                /* istanbul ignore next */ return null;
            }
            function checkReturn(returnNode, reportingNode = returnNode) {
                const tsNode = esTreeNodeToTSNodeMap.get(returnNode);
                const anyType = util.isAnyOrAnyArrayTypeDiscriminated(tsNode, checker);
                const functionNode = getParentFunctionNode(returnNode);
                /* istanbul ignore if */ if (!functionNode) {
                    return;
                }
                // function has an explicit return type, so ensure it's a safe return
                const returnNodeType = util.getConstrainedTypeAtLocation(checker, esTreeNodeToTSNodeMap.get(returnNode));
                const functionTSNode = esTreeNodeToTSNodeMap.get(functionNode);
                // function expressions will not have their return type modified based on receiver typing
                // so we have to use the contextual typing in these cases, i.e.
                // const foo1: () => Set<string> = () => new Set<any>();
                // the return type of the arrow function is Set<any> even though the variable is typed as Set<string>
                let functionType = tsutils.isExpression(functionTSNode)
                    ? util.getContextualType(checker, functionTSNode)
                    : checker.getTypeAtLocation(functionTSNode);
                if (!functionType) {
                    functionType = checker.getTypeAtLocation(functionTSNode);
                }
                // If there is an explicit type annotation *and* that type matches the actual
                // function return type, we shouldn't complain (it's intentional, even if unsafe)
                if (functionTSNode.type) {
                    for (const signature of functionType.getCallSignatures()) {
                        if (returnNodeType === signature.getReturnType()) {
                            return;
                        }
                    }
                }
                if (anyType !== 2 /* Safe */) {
                    // Allow cases when the declared return type of the function is either unknown or unknown[]
                    // and the function is returning any or any[].
                    for (const signature of functionType.getCallSignatures()) {
                        const functionReturnType = signature.getReturnType();
                        if (anyType === 0 /* Any */ &&
                            util.isTypeUnknownType(functionReturnType)) {
                            return;
                        }
                        if (anyType === 1 /* AnyArray */ &&
                            util.isTypeUnknownArrayType(functionReturnType, checker)) {
                            return;
                        }
                    }
                    let messageId = 'unsafeReturn';
                    if (!isNoImplicitThis) {
                        // `return this`
                        const thisExpression = util_1.getThisExpression(returnNode);
                        if (thisExpression &&
                            util.isTypeAnyType(util.getConstrainedTypeAtLocation(checker, esTreeNodeToTSNodeMap.get(thisExpression)))) {
                            messageId = 'unsafeReturnThis';
                        }
                    }
                    // If the function return type was not unknown/unknown[], mark usage as unsafeReturn.
                    return context.report({
                        node: reportingNode,
                        messageId,
                        data: {
                            type: anyType === 0 /* Any */ ? 'any' : 'any[]',
                        },
                    });
                }
                for (const signature of functionType.getCallSignatures()) {
                    const functionReturnType = signature.getReturnType();
                    const result = util.isUnsafeAssignment(returnNodeType, functionReturnType, checker, returnNode);
                    if (!result) {
                        return;
                    }
                    const { sender, receiver } = result;
                    return context.report({
                        node: reportingNode,
                        messageId: 'unsafeReturnAssignment',
                        data: {
                            sender: checker.typeToString(sender),
                            receiver: checker.typeToString(receiver),
                        },
                    });
                }
            }
            return {
                ReturnStatement(node) {
                    const argument = node.argument;
                    if (!argument) {
                        return;
                    }
                    checkReturn(argument, node);
                },
                'ArrowFunctionExpression > :not(BlockStatement).body': checkReturn,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unused-expressions.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const no_unused_expressions_1 = __importDefault(require(616) /* eslint/lib/rules/no-unused-expressions */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-unused-expressions',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow unused expressions',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: no_unused_expressions_1.default.meta.schema,
            messages: (_a = no_unused_expressions_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                unusedExpression: 'Expected an assignment or function call and instead saw an expression.',
            },
        },
        defaultOptions: [
            {
                allowShortCircuit: false,
                allowTernary: false,
                allowTaggedTemplates: false,
            },
        ],
        create(context, options) {
            const rules = no_unused_expressions_1.default.create(context);
            const { allowShortCircuit = false, allowTernary = false } = options[0];
            function isValidExpression(node) {
                if (allowShortCircuit && node.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression) {
                    return isValidExpression(node.right);
                }
                if (allowTernary && node.type === experimental_utils_1.AST_NODE_TYPES.ConditionalExpression) {
                    return (isValidExpression(node.alternate) &&
                        isValidExpression(node.consequent));
                }
                return ((node.type === experimental_utils_1.AST_NODE_TYPES.ChainExpression &&
                    node.expression.type === experimental_utils_1.AST_NODE_TYPES.CallExpression) ||
                    node.type === experimental_utils_1.AST_NODE_TYPES.ImportExpression);
            }
            return {
                ExpressionStatement(node) {
                    if (node.directive || isValidExpression(node.expression)) {
                        return;
                    }
                    rules.ExpressionStatement(node);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unused-vars.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const scope_manager_1 = require(768) /* @typescript-eslint/scope-manager */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-unused-vars',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow unused variables',
                category: 'Variables',
                recommended: 'warn',
                extendsBaseRule: true,
            },
            schema: [
                {
                    oneOf: [
                        {
                            enum: ['all', 'local'],
                        },
                        {
                            type: 'object',
                            properties: {
                                vars: {
                                    enum: ['all', 'local'],
                                },
                                varsIgnorePattern: {
                                    type: 'string',
                                },
                                args: {
                                    enum: ['all', 'after-used', 'none'],
                                },
                                ignoreRestSiblings: {
                                    type: 'boolean',
                                },
                                argsIgnorePattern: {
                                    type: 'string',
                                },
                                caughtErrors: {
                                    enum: ['all', 'none'],
                                },
                                caughtErrorsIgnorePattern: {
                                    type: 'string',
                                },
                            },
                            additionalProperties: false,
                        },
                    ],
                },
            ],
            messages: {
                unusedVar: "'{{varName}}' is {{action}} but never used{{additional}}.",
            },
        },
        defaultOptions: [{}],
        create(context) {
            const filename = context.getFilename();
            const sourceCode = context.getSourceCode();
            const MODULE_DECL_CACHE = new Map();
            const options = (() => {
                var _a, _b, _c, _d;
                const options = {
                    vars: 'all',
                    args: 'after-used',
                    ignoreRestSiblings: false,
                    caughtErrors: 'none',
                };
                const firstOption = context.options[0];
                if (firstOption) {
                    if (typeof firstOption === 'string') {
                        options.vars = firstOption;
                    }
                    else {
                        options.vars = (_a = firstOption.vars) !== null && _a !== void 0 ? _a : options.vars;
                        options.args = (_b = firstOption.args) !== null && _b !== void 0 ? _b : options.args;
                        options.ignoreRestSiblings =
                            (_c = firstOption.ignoreRestSiblings) !== null && _c !== void 0 ? _c : options.ignoreRestSiblings;
                        options.caughtErrors =
                            (_d = firstOption.caughtErrors) !== null && _d !== void 0 ? _d : options.caughtErrors;
                        if (firstOption.varsIgnorePattern) {
                            options.varsIgnorePattern = new RegExp(firstOption.varsIgnorePattern, 'u');
                        }
                        if (firstOption.argsIgnorePattern) {
                            options.argsIgnorePattern = new RegExp(firstOption.argsIgnorePattern, 'u');
                        }
                        if (firstOption.caughtErrorsIgnorePattern) {
                            options.caughtErrorsIgnorePattern = new RegExp(firstOption.caughtErrorsIgnorePattern, 'u');
                        }
                    }
                }
                return options;
            })();
            function collectUnusedVariables() {
                var _a, _b, _c;
                /**
                 * Determines if a variable has a sibling rest property
                 * @param variable eslint-scope variable object.
                 * @returns True if the variable is exported, false if not.
                 */
                function hasRestSpreadSibling(variable) {
                    if (options.ignoreRestSiblings) {
                        return variable.defs.some(def => {
                            const propertyNode = def.name.parent;
                            const patternNode = propertyNode.parent;
                            return (propertyNode.type === experimental_utils_1.AST_NODE_TYPES.Property &&
                                patternNode.type === experimental_utils_1.AST_NODE_TYPES.ObjectPattern &&
                                patternNode.properties[patternNode.properties.length - 1].type ===
                                    experimental_utils_1.AST_NODE_TYPES.RestElement);
                        });
                    }
                    return false;
                }
                /**
                 * Checks whether the given variable is after the last used parameter.
                 * @param variable The variable to check.
                 * @returns `true` if the variable is defined after the last used parameter.
                 */
                function isAfterLastUsedArg(variable) {
                    const def = variable.defs[0];
                    const params = context.getDeclaredVariables(def.node);
                    const posteriorParams = params.slice(params.indexOf(variable) + 1);
                    // If any used parameters occur after this parameter, do not report.
                    return !posteriorParams.some(v => v.references.length > 0 || v.eslintUsed);
                }
                const unusedVariablesOriginal = util.collectUnusedVariables(context);
                const unusedVariablesReturn = [];
                for (const variable of unusedVariablesOriginal) {
                    // explicit global variables don't have definitions.
                    if (variable.defs.length === 0) {
                        unusedVariablesReturn.push(variable);
                        continue;
                    }
                    const def = variable.defs[0];
                    if (variable.scope.type === experimental_utils_1.TSESLint.Scope.ScopeType.global &&
                        options.vars === 'local') {
                        // skip variables in the global scope if configured to
                        continue;
                    }
                    // skip catch variables
                    if (def.type === experimental_utils_1.TSESLint.Scope.DefinitionType.CatchClause) {
                        if (options.caughtErrors === 'none') {
                            continue;
                        }
                        // skip ignored parameters
                        if ('name' in def.name &&
                            ((_a = options.caughtErrorsIgnorePattern) === null || _a === void 0 ? void 0 : _a.test(def.name.name))) {
                            continue;
                        }
                    }
                    if (def.type === experimental_utils_1.TSESLint.Scope.DefinitionType.Parameter) {
                        // if "args" option is "none", skip any parameter
                        if (options.args === 'none') {
                            continue;
                        }
                        // skip ignored parameters
                        if ('name' in def.name &&
                            ((_b = options.argsIgnorePattern) === null || _b === void 0 ? void 0 : _b.test(def.name.name))) {
                            continue;
                        }
                        // if "args" option is "after-used", skip used variables
                        if (options.args === 'after-used' &&
                            util.isFunction(def.name.parent) &&
                            !isAfterLastUsedArg(variable)) {
                            continue;
                        }
                    }
                    else {
                        // skip ignored variables
                        if ('name' in def.name &&
                            ((_c = options.varsIgnorePattern) === null || _c === void 0 ? void 0 : _c.test(def.name.name))) {
                            continue;
                        }
                    }
                    if (hasRestSpreadSibling(variable)) {
                        continue;
                    }
                    // in case another rule has run and used the collectUnusedVariables,
                    // we want to ensure our selectors that marked variables as used are respected
                    if (variable.eslintUsed) {
                        continue;
                    }
                    unusedVariablesReturn.push(variable);
                }
                return unusedVariablesReturn;
            }
            return {
                // declaration file handling
                [ambientDeclarationSelector(experimental_utils_1.AST_NODE_TYPES.Program, true)](node) {
                    if (!util.isDefinitionFile(filename)) {
                        return;
                    }
                    markDeclarationChildAsUsed(node);
                },
                // module declaration in module declaration should not report unused vars error
                // this is workaround as this change should be done in better way
                'TSModuleDeclaration > TSModuleDeclaration'(node) {
                    if (node.id.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                        let scope = context.getScope();
                        if (scope.upper) {
                            scope = scope.upper;
                        }
                        const superVar = scope.set.get(node.id.name);
                        if (superVar) {
                            superVar.eslintUsed = true;
                        }
                    }
                },
                // children of a namespace that is a child of a declared namespace are auto-exported
                [ambientDeclarationSelector('TSModuleDeclaration[declare = true] > TSModuleBlock TSModuleDeclaration > TSModuleBlock', false)](node) {
                    markDeclarationChildAsUsed(node);
                },
                // declared namespace handling
                [ambientDeclarationSelector('TSModuleDeclaration[declare = true] > TSModuleBlock', false)](node) {
                    var _a;
                    const moduleDecl = util.nullThrows((_a = node.parent) === null || _a === void 0 ? void 0 : _a.parent, util.NullThrowsReasons.MissingParent);
                    // declared ambient modules with an `export =` statement will only export that one thing
                    // all other statements are not automatically exported in this case
                    if (moduleDecl.id.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                        checkModuleDeclForExportEquals(moduleDecl)) {
                        return;
                    }
                    markDeclarationChildAsUsed(node);
                },
                // collect
                'Program:exit'(programNode) {
                    /**
                     * Generates the message data about the variable being defined and unused,
                     * including the ignore pattern if configured.
                     * @param unusedVar eslint-scope variable object.
                     * @returns The message data to be used with this unused variable.
                     */
                    function getDefinedMessageData(unusedVar) {
                        var _a;
                        const defType = (_a = unusedVar === null || unusedVar === void 0 ? void 0 : unusedVar.defs[0]) === null || _a === void 0 ? void 0 : _a.type;
                        let type;
                        let pattern;
                        if (defType === experimental_utils_1.TSESLint.Scope.DefinitionType.CatchClause &&
                            options.caughtErrorsIgnorePattern) {
                            type = 'args';
                            pattern = options.caughtErrorsIgnorePattern.toString();
                        }
                        else if (defType === experimental_utils_1.TSESLint.Scope.DefinitionType.Parameter &&
                            options.argsIgnorePattern) {
                            type = 'args';
                            pattern = options.argsIgnorePattern.toString();
                        }
                        else if (defType !== experimental_utils_1.TSESLint.Scope.DefinitionType.Parameter &&
                            options.varsIgnorePattern) {
                            type = 'vars';
                            pattern = options.varsIgnorePattern.toString();
                        }
                        const additional = type
                            ? `. Allowed unused ${type} must match ${pattern}`
                            : '';
                        return {
                            varName: unusedVar.name,
                            action: 'defined',
                            additional,
                        };
                    }
                    /**
                     * Generate the warning message about the variable being
                     * assigned and unused, including the ignore pattern if configured.
                     * @param unusedVar eslint-scope variable object.
                     * @returns The message data to be used with this unused variable.
                     */
                    function getAssignedMessageData(unusedVar) {
                        const additional = options.varsIgnorePattern
                            ? `. Allowed unused vars must match ${options.varsIgnorePattern.toString()}`
                            : '';
                        return {
                            varName: unusedVar.name,
                            action: 'assigned a value',
                            additional,
                        };
                    }
                    const unusedVars = collectUnusedVariables();
                    for (let i = 0, l = unusedVars.length; i < l; ++i) {
                        const unusedVar = unusedVars[i];
                        // Report the first declaration.
                        if (unusedVar.defs.length > 0) {
                            context.report({
                                node: unusedVar.references.length
                                    ? unusedVar.references[unusedVar.references.length - 1]
                                        .identifier
                                    : unusedVar.identifiers[0],
                                messageId: 'unusedVar',
                                data: unusedVar.references.some(ref => ref.isWrite())
                                    ? getAssignedMessageData(unusedVar)
                                    : getDefinedMessageData(unusedVar),
                            });
                            // If there are no regular declaration, report the first `/*globals*/` comment directive.
                        }
                        else if ('eslintExplicitGlobalComments' in unusedVar &&
                            unusedVar.eslintExplicitGlobalComments) {
                            const directiveComment = unusedVar.eslintExplicitGlobalComments[0];
                            context.report({
                                node: programNode,
                                loc: util.getNameLocationInGlobalDirectiveComment(sourceCode, directiveComment, unusedVar.name),
                                messageId: 'unusedVar',
                                data: getDefinedMessageData(unusedVar),
                            });
                        }
                    }
                },
            };
            function checkModuleDeclForExportEquals(node) {
                const cached = MODULE_DECL_CACHE.get(node);
                if (cached != null) {
                    return cached;
                }
                if (node.body && node.body.type === experimental_utils_1.AST_NODE_TYPES.TSModuleBlock) {
                    for (const statement of node.body.body) {
                        if (statement.type === experimental_utils_1.AST_NODE_TYPES.TSExportAssignment) {
                            MODULE_DECL_CACHE.set(node, true);
                            return true;
                        }
                    }
                }
                MODULE_DECL_CACHE.set(node, false);
                return false;
            }
            function ambientDeclarationSelector(parent, childDeclare) {
                return [
                    // Types are ambiently exported
                    `${parent} > :matches(${[
                        experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration,
                        experimental_utils_1.AST_NODE_TYPES.TSTypeAliasDeclaration,
                    ].join(', ')})`,
                    // Value things are ambiently exported if they are "declare"d
                    `${parent} > :matches(${[
                        experimental_utils_1.AST_NODE_TYPES.ClassDeclaration,
                        experimental_utils_1.AST_NODE_TYPES.TSDeclareFunction,
                        experimental_utils_1.AST_NODE_TYPES.TSEnumDeclaration,
                        experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration,
                        experimental_utils_1.AST_NODE_TYPES.VariableDeclaration,
                    ].join(', ')})${childDeclare ? '[declare = true]' : ''}`,
                ].join(', ');
            }
            function markDeclarationChildAsUsed(node) {
                var _a;
                const identifiers = [];
                switch (node.type) {
                    case experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration:
                    case experimental_utils_1.AST_NODE_TYPES.TSTypeAliasDeclaration:
                    case experimental_utils_1.AST_NODE_TYPES.ClassDeclaration:
                    case experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration:
                    case experimental_utils_1.AST_NODE_TYPES.TSDeclareFunction:
                    case experimental_utils_1.AST_NODE_TYPES.TSEnumDeclaration:
                    case experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration:
                        if (((_a = node.id) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                            identifiers.push(node.id);
                        }
                        break;
                    case experimental_utils_1.AST_NODE_TYPES.VariableDeclaration:
                        for (const declaration of node.declarations) {
                            visitPattern(declaration, pattern => {
                                identifiers.push(pattern);
                            });
                        }
                        break;
                }
                let scope = context.getScope();
                const shouldUseUpperScope = [
                    experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration,
                    experimental_utils_1.AST_NODE_TYPES.TSDeclareFunction,
                ].includes(node.type);
                if (scope.variableScope !== scope) {
                    scope = scope.variableScope;
                }
                else if (shouldUseUpperScope && scope.upper) {
                    scope = scope.upper;
                }
                for (const id of identifiers) {
                    const superVar = scope.set.get(id.name);
                    if (superVar) {
                        superVar.eslintUsed = true;
                    }
                }
            }
            function visitPattern(node, cb) {
                const visitor = new scope_manager_1.PatternVisitor({}, node, cb);
                visitor.visit(node);
            }
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-unused-vars-experimental.js */ function _(require, module, exports, __esModule, __esExport) {
    /* eslint-disable no-fallthrough */
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.DEFAULT_IGNORED_REGEX_STRING = '^_';
    exports.default = util.createRule({
        name: 'no-unused-vars-experimental',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow unused variables and arguments',
                category: 'Best Practices',
                recommended: false,
            },
            deprecated: true,
            replacedBy: ['no-unused-vars'],
            schema: [
                {
                    type: 'object',
                    properties: {
                        ignoredNamesRegex: {
                            oneOf: [
                                {
                                    type: 'string',
                                },
                                {
                                    type: 'boolean',
                                    enum: [false],
                                },
                            ],
                        },
                        ignoreArgsIfArgsAfterAreUsed: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
            messages: {
                unused: "{{type}} '{{name}}' is declared but its value is never read.",
                unusedWithIgnorePattern: "{{type}} '{{name}}' is declared but its value is never read. Allowed unused names must match {{pattern}}.",
                unusedImport: 'All imports in import declaration are unused.',
                unusedTypeParameters: 'All type parameters are unused.',
            },
        },
        defaultOptions: [
            {
                ignoredNamesRegex: exports.DEFAULT_IGNORED_REGEX_STRING,
                ignoreArgsIfArgsAfterAreUsed: false,
            },
        ],
        create(context, [userOptions]) {
            var _a;
            const parserServices = util.getParserServices(context, true);
            const tsProgram = parserServices.program;
            const afterAllDiagnosticsCallbacks = [];
            const options = {
                ignoredNames: userOptions && typeof userOptions.ignoredNamesRegex === 'string'
                    ? new RegExp(userOptions.ignoredNamesRegex)
                    : null,
                ignoreArgsIfArgsAfterAreUsed: (_a = userOptions.ignoreArgsIfArgsAfterAreUsed) !== null && _a !== void 0 ? _a : false,
            };
            function handleIdentifier(identifier) {
                function report(type) {
                    const node = parserServices.tsNodeToESTreeNodeMap.get(identifier);
                    const regex = options.ignoredNames;
                    const name = identifier.getText();
                    if (regex) {
                        if (!regex.test(name)) {
                            context.report({
                                node,
                                messageId: 'unusedWithIgnorePattern',
                                data: {
                                    name,
                                    type,
                                    pattern: regex.toString(),
                                },
                            });
                        }
                    }
                    else {
                        context.report({
                            node,
                            messageId: 'unused',
                            data: {
                                name,
                                type,
                            },
                        });
                    }
                }
                const parent = identifier.parent;
                // is a single variable diagnostic
                switch (parent.kind) {
                    case ts.SyntaxKind.BindingElement:
                    case ts.SyntaxKind.ObjectBindingPattern:
                        report('Destructured Variable');
                        break;
                    case ts.SyntaxKind.ClassDeclaration:
                        report('Class');
                        break;
                    case ts.SyntaxKind.EnumDeclaration:
                        report('Enum');
                        break;
                    case ts.SyntaxKind.FunctionDeclaration:
                        report('Function');
                        break;
                    // this won't happen because there are specific nodes that wrap up named/default import identifiers
                    // case ts.SyntaxKind.ImportDeclaration:
                    // import equals is always treated as a variable
                    case ts.SyntaxKind.ImportEqualsDeclaration:
                    // the default import is NOT used, but a named import is used
                    case ts.SyntaxKind.ImportClause:
                    // a named import is NOT used, but either another named import, or the default import is used
                    case ts.SyntaxKind.ImportSpecifier:
                    // a namespace import is NOT used, but the default import is used
                    case ts.SyntaxKind.NamespaceImport:
                        report('Import');
                        break;
                    case ts.SyntaxKind.InterfaceDeclaration:
                        report('Interface');
                        break;
                    case ts.SyntaxKind.MethodDeclaration:
                        report('Method');
                        break;
                    case ts.SyntaxKind.Parameter:
                        handleParameterDeclaration(identifier, parent);
                        break;
                    case ts.SyntaxKind.PropertyDeclaration:
                        // eslint-disable-next-line @typescript-eslint/internal/prefer-ast-types-enum
                        report('Property');
                        break;
                    case ts.SyntaxKind.TypeAliasDeclaration:
                        report('Type');
                        break;
                    case ts.SyntaxKind.TypeParameter:
                        handleTypeParam(identifier);
                        break;
                    case ts.SyntaxKind.VariableDeclaration:
                        report('Variable');
                        break;
                    default:
                        throw new Error(`Unknown node with kind ${parent.kind}.`);
                    // TODO - should we just handle this gracefully?
                    // report('Unknown Node');
                    // break;
                }
            }
            const unusedParameters = new Set();
            function handleParameterDeclaration(identifier, parent) {
                const name = identifier.getText();
                // regardless of if the parameter is ignored, track that it had a diagnostic fired on it
                unusedParameters.add(identifier);
                /*
                NOTE - Typescript will automatically ignore parameters that have a
                       leading underscore in their name. We cannot do anything about this.
                */
                function report() {
                    const node = parserServices.tsNodeToESTreeNodeMap.get(identifier);
                    context.report({
                        node,
                        messageId: 'unused',
                        data: {
                            name,
                            type: 'Parameter',
                        },
                    });
                }
                const isLastParameter = parent.parent.parameters.indexOf(parent) ===
                    parent.parent.parameters.length - 1;
                if (!isLastParameter && options.ignoreArgsIfArgsAfterAreUsed) {
                    // once all diagnostics are processed, we can check if the following args are unused
                    afterAllDiagnosticsCallbacks.push(() => {
                        for (const param of parent.parent.parameters) {
                            if (!unusedParameters.has(param.name)) {
                                return;
                            }
                        }
                        // none of the following params were unused, so report
                        report();
                    });
                }
                else {
                    report();
                }
            }
            function handleImportDeclaration(parent) {
                // the entire import statement is unused
                /*
                NOTE - Typescript will automatically ignore imports that have a
                       leading underscore in their name. We cannot do anything about this.
                */
                context.report({
                    messageId: 'unusedImport',
                    node: parserServices.tsNodeToESTreeNodeMap.get(parent),
                });
            }
            function handleDestructure(parent) {
                // the entire destructure is unused
                // note that this case only ever triggers for simple, single-level destructured objects
                // i.e. these will not trigger it:
                // - const {a:_a, b, c: {d}} = z;
                // - const [a, b] = c;
                parent.elements.forEach(element => {
                    if (element.kind === ts.SyntaxKind.BindingElement) {
                        const name = element.name;
                        if (name.kind === ts.SyntaxKind.Identifier) {
                            handleIdentifier(name);
                        }
                    }
                });
            }
            function handleTypeParamList(node) {
                // the entire generic decl list is unused
                /*
                NOTE - Typescript will automatically ignore generics that have a
                       leading underscore in their name. We cannot do anything about this.
                */
                const parent = parserServices.tsNodeToESTreeNodeMap.get(node);
                context.report({
                    messageId: 'unusedTypeParameters',
                    node: parent.typeParameters,
                });
            }
            function handleTypeParam(identifier) {
                context.report({
                    node: parserServices.tsNodeToESTreeNodeMap.get(identifier),
                    messageId: 'unused',
                    data: {
                        name: identifier.getText(),
                        type: 'Type Parameter',
                    },
                });
            }
            return {
                'Program:exit'(program) {
                    const tsNode = parserServices.esTreeNodeToTSNodeMap.get(program);
                    const sourceFile = util.getSourceFileOfNode(tsNode);
                    const diagnostics = tsProgram.getSemanticDiagnostics(sourceFile);
                    diagnostics.forEach(diag => {
                        if (isUnusedDiagnostic(diag.code)) {
                            if (diag.start !== undefined) {
                                const node = util.getTokenAtPosition(sourceFile, diag.start);
                                const parent = node.parent;
                                if (isIdentifier(node)) {
                                    handleIdentifier(node);
                                }
                                else if (isImport(parent)) {
                                    handleImportDeclaration(parent);
                                }
                                else if (isDestructure(parent)) {
                                    handleDestructure(parent);
                                }
                                else if (isGeneric(node, parent)) {
                                    handleTypeParamList(parent);
                                }
                            }
                        }
                    });
                    // trigger all the checks to be done after all the diagnostics have been evaluated
                    afterAllDiagnosticsCallbacks.forEach(cb => cb());
                },
            };
        },
    });
    /**
     * Checks if the diagnostic code is one of the expected "unused var" codes
     */
    function isUnusedDiagnostic(code) {
        return [
            6133,
            6138,
            6192,
            6196,
            6198,
            6199,
            6205, // All type parameters are unused.
        ].includes(code);
    }
    /**
     * Checks if the given node is a destructuring pattern
     */
    function isDestructure(node) {
        return (node.kind === ts.SyntaxKind.ObjectBindingPattern ||
            node.kind === ts.SyntaxKind.ArrayBindingPattern);
    }
    function isImport(node) {
        return node.kind === ts.SyntaxKind.ImportDeclaration;
    }
    function isIdentifier(node) {
        return node.kind === ts.SyntaxKind.Identifier;
    }
    function isGeneric(node, parent) {
        return (node.kind === ts.SyntaxKind.LessThanToken &&
            parent.typeParameters !== undefined);
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-use-before-define.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const SENTINEL_TYPE = /^(?:(?:Function|Class)(?:Declaration|Expression)|ArrowFunctionExpression|CatchClause|ImportDeclaration|ExportNamedDeclaration)$/;
    /**
     * Parses a given value as options.
     */
    function parseOptions(options) {
        let functions = true;
        let classes = true;
        let enums = true;
        let variables = true;
        let typedefs = true;
        let ignoreTypeReferences = true;
        if (typeof options === 'string') {
            functions = options !== 'nofunc';
        }
        else if (typeof options === 'object' && options !== null) {
            functions = options.functions !== false;
            classes = options.classes !== false;
            enums = options.enums !== false;
            variables = options.variables !== false;
            typedefs = options.typedefs !== false;
            ignoreTypeReferences = options.ignoreTypeReferences !== false;
        }
        return {
            functions,
            classes,
            enums,
            variables,
            typedefs,
            ignoreTypeReferences,
        };
    }
    /**
     * Checks whether or not a given variable is a function declaration.
     */
    function isFunction(variable) {
        return variable.defs[0].type === 'FunctionName';
    }
    /**
     * Checks whether or not a given variable is a type declaration.
     */
    function isTypedef(variable) {
        return variable.defs[0].type === 'Type';
    }
    /**
     * Checks whether or not a given variable is a enum declaration.
     */
    function isOuterEnum(variable, reference) {
        return (variable.defs[0].type == 'TSEnumName' &&
            variable.scope.variableScope !== reference.from.variableScope);
    }
    /**
     * Checks whether or not a given variable is a class declaration in an upper function scope.
     */
    function isOuterClass(variable, reference) {
        return (variable.defs[0].type === 'ClassName' &&
            variable.scope.variableScope !== reference.from.variableScope);
    }
    /**
     * Checks whether or not a given variable is a variable declaration in an upper function scope.
     */
    function isOuterVariable(variable, reference) {
        return (variable.defs[0].type === 'Variable' &&
            variable.scope.variableScope !== reference.from.variableScope);
    }
    /**
     * Recursively checks whether or not a given reference has a type query declaration among it's parents
     */
    function referenceContainsTypeQuery(node) {
        switch (node.type) {
            case experimental_utils_1.AST_NODE_TYPES.TSTypeQuery:
                return true;
            case experimental_utils_1.AST_NODE_TYPES.TSQualifiedName:
            case experimental_utils_1.AST_NODE_TYPES.Identifier:
                if (!node.parent) {
                    return false;
                }
                return referenceContainsTypeQuery(node.parent);
            default:
                // if we find a different node, there's no chance that we're in a TSTypeQuery
                return false;
        }
    }
    /**
     * Checks whether or not a given reference is a type reference.
     */
    function isTypeReference(reference) {
        return (reference.isTypeReference ||
            referenceContainsTypeQuery(reference.identifier));
    }
    /**
     * Checks whether or not a given location is inside of the range of a given node.
     */
    function isInRange(node, location) {
        return !!node && node.range[0] <= location && location <= node.range[1];
    }
    /**
     * Decorators are transpiled such that the decorator is placed after the class declaration
     * So it is considered safe
     */
    function isClassRefInClassDecorator(variable, reference) {
        if (variable.defs[0].type !== 'ClassName') {
            return false;
        }
        if (!variable.defs[0].node.decorators ||
            variable.defs[0].node.decorators.length === 0) {
            return false;
        }
        for (const deco of variable.defs[0].node.decorators) {
            if (reference.identifier.range[0] >= deco.range[0] &&
                reference.identifier.range[1] <= deco.range[1]) {
                return true;
            }
        }
        return false;
    }
    /**
     * Checks whether or not a given reference is inside of the initializers of a given variable.
     *
     * @returns `true` in the following cases:
     * - var a = a
     * - var [a = a] = list
     * - var {a = a} = obj
     * - for (var a in a) {}
     * - for (var a of a) {}
     */
    function isInInitializer(variable, reference) {
        var _a;
        if (variable.scope !== reference.from) {
            return false;
        }
        let node = variable.identifiers[0].parent;
        const location = reference.identifier.range[1];
        while (node) {
            if (node.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclarator) {
                if (isInRange(node.init, location)) {
                    return true;
                }
                if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.parent) &&
                    (node.parent.parent.type === experimental_utils_1.AST_NODE_TYPES.ForInStatement ||
                        node.parent.parent.type === experimental_utils_1.AST_NODE_TYPES.ForOfStatement) &&
                    isInRange(node.parent.parent.right, location)) {
                    return true;
                }
                break;
            }
            else if (node.type === experimental_utils_1.AST_NODE_TYPES.AssignmentPattern) {
                if (isInRange(node.right, location)) {
                    return true;
                }
            }
            else if (SENTINEL_TYPE.test(node.type)) {
                break;
            }
            node = node.parent;
        }
        return false;
    }
    exports.default = util.createRule({
        name: 'no-use-before-define',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow the use of variables before they are defined',
                category: 'Variables',
                recommended: false,
                extendsBaseRule: true,
            },
            messages: {
                noUseBeforeDefine: "'{{name}}' was used before it was defined.",
            },
            schema: [
                {
                    oneOf: [
                        {
                            enum: ['nofunc'],
                        },
                        {
                            type: 'object',
                            properties: {
                                functions: { type: 'boolean' },
                                classes: { type: 'boolean' },
                                enums: { type: 'boolean' },
                                variables: { type: 'boolean' },
                                typedefs: { type: 'boolean' },
                                ignoreTypeReferences: { type: 'boolean' },
                            },
                            additionalProperties: false,
                        },
                    ],
                },
            ],
        },
        defaultOptions: [
            {
                functions: true,
                classes: true,
                enums: true,
                variables: true,
                typedefs: true,
                ignoreTypeReferences: true,
            },
        ],
        create(context, optionsWithDefault) {
            const options = parseOptions(optionsWithDefault[0]);
            /**
             * Determines whether a given use-before-define case should be reported according to the options.
             * @param variable The variable that gets used before being defined
             * @param reference The reference to the variable
             */
            function isForbidden(variable, reference) {
                if (options.ignoreTypeReferences && isTypeReference(reference)) {
                    return false;
                }
                if (isFunction(variable)) {
                    return options.functions;
                }
                if (isOuterClass(variable, reference)) {
                    return options.classes;
                }
                if (isOuterVariable(variable, reference)) {
                    return options.variables;
                }
                if (isOuterEnum(variable, reference)) {
                    return options.enums;
                }
                if (isTypedef(variable)) {
                    return options.typedefs;
                }
                return true;
            }
            /**
             * Finds and validates all variables in a given scope.
             */
            function findVariablesInScope(scope) {
                scope.references.forEach(reference => {
                    const variable = reference.resolved;
                    // Skips when the reference is:
                    // - initializations.
                    // - referring to an undefined variable.
                    // - referring to a global environment variable (there're no identifiers).
                    // - located preceded by the variable (except in initializers).
                    // - allowed by options.
                    if (reference.init ||
                        !variable ||
                        variable.identifiers.length === 0 ||
                        (variable.identifiers[0].range[1] <= reference.identifier.range[1] &&
                            !isInInitializer(variable, reference)) ||
                        !isForbidden(variable, reference) ||
                        isClassRefInClassDecorator(variable, reference) ||
                        reference.from.type === experimental_utils_1.TSESLint.Scope.ScopeType.functionType) {
                        return;
                    }
                    // Reports.
                    context.report({
                        node: reference.identifier,
                        messageId: 'noUseBeforeDefine',
                        data: {
                            name: reference.identifier.name,
                        },
                    });
                });
                scope.childScopes.forEach(findVariablesInScope);
            }
            return {
                Program() {
                    findVariablesInScope(context.getScope());
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-useless-constructor.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const no_useless_constructor_1 = __importDefault(require(625) /* eslint/lib/rules/no-useless-constructor */);
    const util = __importStar(require(906) /* ../util */);
    /**
     * Check if method with accessibility is not useless
     */
    function checkAccessibility(node) {
        switch (node.accessibility) {
            case 'protected':
            case 'private':
                return false;
            case 'public':
                if (node.parent &&
                    node.parent.type === experimental_utils_1.AST_NODE_TYPES.ClassBody &&
                    node.parent.parent &&
                    'superClass' in node.parent.parent &&
                    node.parent.parent.superClass) {
                    return false;
                }
                break;
        }
        return true;
    }
    /**
     * Check if method is not unless due to typescript parameter properties
     */
    function checkParams(node) {
        return (!node.value.params ||
            !node.value.params.some(param => param.type === experimental_utils_1.AST_NODE_TYPES.TSParameterProperty));
    }
    exports.default = util.createRule({
        name: 'no-useless-constructor',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallow unnecessary constructors',
                category: 'Best Practices',
                recommended: false,
                extendsBaseRule: true,
            },
            schema: no_useless_constructor_1.default.meta.schema,
            messages: (_a = no_useless_constructor_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                noUselessConstructor: 'Useless constructor.',
            },
        },
        defaultOptions: [],
        create(context) {
            const rules = no_useless_constructor_1.default.create(context);
            return {
                MethodDefinition(node) {
                    if (node.value &&
                        node.value.type === experimental_utils_1.AST_NODE_TYPES.FunctionExpression &&
                        node.value.body &&
                        checkAccessibility(node) &&
                        checkParams(node)) {
                        rules.MethodDefinition(node);
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/no-var-requires.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'no-var-requires',
        meta: {
            type: 'problem',
            docs: {
                description: 'Disallows the use of require statements except in import statements',
                category: 'Best Practices',
                recommended: 'error',
            },
            messages: {
                noVarReqs: 'Require statement not part of import statement.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            return {
                CallExpression(node) {
                    var _a;
                    const parent = ((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.ChainExpression
                        ? node.parent.parent
                        : node.parent;
                    if (node.callee.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                        node.callee.name === 'require' &&
                        parent &&
                        (parent.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclarator ||
                            parent.type === experimental_utils_1.AST_NODE_TYPES.CallExpression ||
                            parent.type === experimental_utils_1.AST_NODE_TYPES.TSAsExpression ||
                            parent.type === experimental_utils_1.AST_NODE_TYPES.TSTypeAssertion ||
                            parent.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression)) {
                        context.report({
                            node,
                            messageId: 'noVarReqs',
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/non-nullable-type-assertion-style.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'non-nullable-type-assertion-style',
        meta: {
            docs: {
                category: 'Best Practices',
                description: 'Prefers a non-null assertion over explicit type cast when possible',
                recommended: false,
                requiresTypeChecking: true,
                suggestion: true,
            },
            fixable: 'code',
            messages: {
                preferNonNullAssertion: 'Use a ! assertion to more succintly remove null and undefined from the type.',
            },
            schema: [],
            type: 'suggestion',
        },
        defaultOptions: [],
        create(context) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const sourceCode = context.getSourceCode();
            const getTypesIfNotLoose = (node) => {
                const type = checker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(node));
                if (tsutils.isTypeFlagSet(type, ts.TypeFlags.Any | ts.TypeFlags.Unknown)) {
                    return undefined;
                }
                return tsutils.unionTypeParts(type);
            };
            const sameTypeWithoutNullish = (assertedTypes, originalTypes) => {
                const nonNullishOriginalTypes = originalTypes.filter(type => type.flags !== ts.TypeFlags.Null &&
                    type.flags !== ts.TypeFlags.Undefined);
                for (const assertedType of assertedTypes) {
                    if (!nonNullishOriginalTypes.includes(assertedType)) {
                        return false;
                    }
                }
                for (const originalType of nonNullishOriginalTypes) {
                    if (!assertedTypes.includes(originalType)) {
                        return false;
                    }
                }
                return true;
            };
            const isConstAssertion = (node) => {
                return (node.typeAnnotation.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference &&
                    node.typeAnnotation.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    node.typeAnnotation.typeName.name === 'const');
            };
            return {
                'TSAsExpression, TSTypeAssertion'(node) {
                    if (isConstAssertion(node)) {
                        return;
                    }
                    const originalTypes = getTypesIfNotLoose(node.expression);
                    if (!originalTypes) {
                        return;
                    }
                    const assertedTypes = getTypesIfNotLoose(node.typeAnnotation);
                    if (!assertedTypes) {
                        return;
                    }
                    if (sameTypeWithoutNullish(assertedTypes, originalTypes)) {
                        context.report({
                            fix(fixer) {
                                return fixer.replaceText(node, `${sourceCode.getText(node.expression)}!`);
                            },
                            messageId: 'preferNonNullAssertion',
                            node,
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/object-curly-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const object_curly_spacing_1 = __importDefault(require(636) /* eslint/lib/rules/object-curly-spacing */);
    const util_1 = require(906) /* ../util */;
    exports.default = util_1.createRule({
        name: 'object-curly-spacing',
        meta: Object.assign(Object.assign({}, object_curly_spacing_1.default.meta), { docs: {
                description: 'Enforce consistent spacing inside braces',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            } }),
        defaultOptions: ['never'],
        create(context) {
            const spaced = context.options[0] === 'always';
            const sourceCode = context.getSourceCode();
            /**
             * Determines whether an option is set, relative to the spacing option.
             * If spaced is "always", then check whether option is set to false.
             * If spaced is "never", then check whether option is set to true.
             * @param option The option to exclude.
             * @returns Whether or not the property is excluded.
             */
            function isOptionSet(option) {
                return context.options[1]
                    ? context.options[1][option] === !spaced
                    : false;
            }
            const options = {
                spaced,
                arraysInObjectsException: isOptionSet('arraysInObjects'),
                objectsInObjectsException: isOptionSet('objectsInObjects'),
            };
            //--------------------------------------------------------------------------
            // Helpers
            //--------------------------------------------------------------------------
            /**
             * Reports that there shouldn't be a space after the first token
             * @param node The node to report in the event of an error.
             * @param token The token to use for the report.
             */
            function reportNoBeginningSpace(node, token) {
                const nextToken = context
                    .getSourceCode()
                    .getTokenAfter(token, { includeComments: true });
                context.report({
                    node,
                    loc: { start: token.loc.end, end: nextToken.loc.start },
                    messageId: 'unexpectedSpaceAfter',
                    data: {
                        token: token.value,
                    },
                    fix(fixer) {
                        return fixer.removeRange([token.range[1], nextToken.range[0]]);
                    },
                });
            }
            /**
             * Reports that there shouldn't be a space before the last token
             * @param node The node to report in the event of an error.
             * @param token The token to use for the report.
             */
            function reportNoEndingSpace(node, token) {
                const previousToken = context
                    .getSourceCode()
                    .getTokenBefore(token, { includeComments: true });
                context.report({
                    node,
                    loc: { start: previousToken.loc.end, end: token.loc.start },
                    messageId: 'unexpectedSpaceBefore',
                    data: {
                        token: token.value,
                    },
                    fix(fixer) {
                        return fixer.removeRange([previousToken.range[1], token.range[0]]);
                    },
                });
            }
            /**
             * Reports that there should be a space after the first token
             * @param node The node to report in the event of an error.
             * @param token The token to use for the report.
             */
            function reportRequiredBeginningSpace(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: 'requireSpaceAfter',
                    data: {
                        token: token.value,
                    },
                    fix(fixer) {
                        return fixer.insertTextAfter(token, ' ');
                    },
                });
            }
            /**
             * Reports that there should be a space before the last token
             * @param node The node to report in the event of an error.
             * @param token The token to use for the report.
             */
            function reportRequiredEndingSpace(node, token) {
                context.report({
                    node,
                    loc: token.loc,
                    messageId: 'requireSpaceBefore',
                    data: {
                        token: token.value,
                    },
                    fix(fixer) {
                        return fixer.insertTextBefore(token, ' ');
                    },
                });
            }
            /**
             * Determines if spacing in curly braces is valid.
             * @param node The AST node to check.
             * @param first The first token to check (should be the opening brace)
             * @param second The second token to check (should be first after the opening brace)
             * @param penultimate The penultimate token to check (should be last before closing brace)
             * @param last The last token to check (should be closing brace)
             */
            function validateBraceSpacing(node, first, second, penultimate, last) {
                if (util_1.isTokenOnSameLine(first, second)) {
                    const firstSpaced = sourceCode.isSpaceBetween(first, second);
                    const secondType = sourceCode.getNodeByRangeIndex(second.range[0]).type;
                    const openingCurlyBraceMustBeSpaced = options.arraysInObjectsException &&
                        [
                            experimental_utils_1.AST_NODE_TYPES.TSMappedType,
                            experimental_utils_1.AST_NODE_TYPES.TSIndexSignature,
                        ].includes(secondType)
                        ? !options.spaced
                        : options.spaced;
                    if (openingCurlyBraceMustBeSpaced && !firstSpaced) {
                        reportRequiredBeginningSpace(node, first);
                    }
                    if (!openingCurlyBraceMustBeSpaced &&
                        firstSpaced &&
                        second.type !== experimental_utils_1.AST_TOKEN_TYPES.Line) {
                        reportNoBeginningSpace(node, first);
                    }
                }
                if (util_1.isTokenOnSameLine(penultimate, last)) {
                    const shouldCheckPenultimate = (options.arraysInObjectsException &&
                        util_1.isClosingBracketToken(penultimate)) ||
                        (options.objectsInObjectsException &&
                            util_1.isClosingBraceToken(penultimate));
                    const penultimateType = shouldCheckPenultimate
                        ? sourceCode.getNodeByRangeIndex(penultimate.range[0]).type
                        : undefined;
                    const closingCurlyBraceMustBeSpaced = (options.arraysInObjectsException &&
                        penultimateType === experimental_utils_1.AST_NODE_TYPES.TSTupleType) ||
                        (options.objectsInObjectsException &&
                            penultimateType !== undefined &&
                            [
                                experimental_utils_1.AST_NODE_TYPES.TSMappedType,
                                experimental_utils_1.AST_NODE_TYPES.TSTypeLiteral,
                            ].includes(penultimateType))
                        ? !options.spaced
                        : options.spaced;
                    const lastSpaced = sourceCode.isSpaceBetween(penultimate, last);
                    if (closingCurlyBraceMustBeSpaced && !lastSpaced) {
                        reportRequiredEndingSpace(node, last);
                    }
                    if (!closingCurlyBraceMustBeSpaced && lastSpaced) {
                        reportNoEndingSpace(node, last);
                    }
                }
            }
            /**
             * Gets '}' token of an object node.
             *
             * Because the last token of object patterns might be a type annotation,
             * this traverses tokens preceded by the last property, then returns the
             * first '}' token.
             * @param node The node to get. This node is an
             *      ObjectExpression or an ObjectPattern. And this node has one or
             *      more properties.
             * @returns '}' token.
             */
            function getClosingBraceOfObject(node) {
                const lastProperty = node.members[node.members.length - 1];
                return sourceCode.getTokenAfter(lastProperty, util_1.isClosingBraceToken);
            }
            //--------------------------------------------------------------------------
            // Public
            //--------------------------------------------------------------------------
            const rules = object_curly_spacing_1.default.create(context);
            return Object.assign(Object.assign({}, rules), { TSMappedType(node) {
                    const first = sourceCode.getFirstToken(node);
                    const last = sourceCode.getLastToken(node);
                    const second = sourceCode.getTokenAfter(first, {
                        includeComments: true,
                    });
                    const penultimate = sourceCode.getTokenBefore(last, {
                        includeComments: true,
                    });
                    validateBraceSpacing(node, first, second, penultimate, last);
                },
                TSTypeLiteral(node) {
                    if (node.members.length === 0) {
                        return;
                    }
                    const first = sourceCode.getFirstToken(node);
                    const last = getClosingBraceOfObject(node);
                    const second = sourceCode.getTokenAfter(first, {
                        includeComments: true,
                    });
                    const penultimate = sourceCode.getTokenBefore(last, {
                        includeComments: true,
                    });
                    validateBraceSpacing(node, first, second, penultimate, last);
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-as-const.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'prefer-as-const',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Prefer usage of `as const` over literal type',
                category: 'Best Practices',
                recommended: 'error',
                suggestion: true,
            },
            fixable: 'code',
            messages: {
                preferConstAssertion: 'Expected a `const` instead of a literal type assertion.',
                variableConstAssertion: 'Expected a `const` assertion instead of a literal type annotation.',
                variableSuggest: 'You should use `as const` instead of type annotation.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            function compareTypes(valueNode, typeNode, canFix) {
                if (valueNode.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                    typeNode.type === experimental_utils_1.AST_NODE_TYPES.TSLiteralType &&
                    'raw' in typeNode.literal &&
                    valueNode.raw === typeNode.literal.raw) {
                    if (canFix) {
                        context.report({
                            node: typeNode,
                            messageId: 'preferConstAssertion',
                            fix: fixer => fixer.replaceText(typeNode, 'const'),
                        });
                    }
                    else {
                        context.report({
                            node: typeNode,
                            messageId: 'variableConstAssertion',
                            suggest: [
                                {
                                    messageId: 'variableSuggest',
                                    fix: (fixer) => [
                                        fixer.remove(typeNode.parent),
                                        fixer.insertTextAfter(valueNode, ' as const'),
                                    ],
                                },
                            ],
                        });
                    }
                }
            }
            return {
                TSAsExpression(node) {
                    compareTypes(node.expression, node.typeAnnotation, true);
                },
                TSTypeAssertion(node) {
                    compareTypes(node.expression, node.typeAnnotation, true);
                },
                VariableDeclarator(node) {
                    if (node.init && node.id.typeAnnotation) {
                        compareTypes(node.init, node.id.typeAnnotation.typeAnnotation, false);
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-enum-initializers.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'prefer-enum-initializers',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Prefer initializing each enums member value',
                category: 'Best Practices',
                recommended: false,
                suggestion: true,
            },
            messages: {
                defineInitializer: "The value of the member '{{ name }}' should be explicitly defined",
                defineInitializerSuggestion: 'Can be fixed to {{ name }} = {{ suggested }}',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const sourceCode = context.getSourceCode();
            function TSEnumDeclaration(node) {
                const { members } = node;
                members.forEach((member, index) => {
                    if (member.initializer == null) {
                        const name = sourceCode.getText(member);
                        context.report({
                            node: member,
                            messageId: 'defineInitializer',
                            data: {
                                name,
                            },
                            suggest: [
                                {
                                    messageId: 'defineInitializerSuggestion',
                                    data: { name, suggested: index },
                                    fix: (fixer) => {
                                        return fixer.replaceText(member, `${name} = ${index}`);
                                    },
                                },
                                {
                                    messageId: 'defineInitializerSuggestion',
                                    data: { name, suggested: index + 1 },
                                    fix: (fixer) => {
                                        return fixer.replaceText(member, `${name} = ${index + 1}`);
                                    },
                                },
                                {
                                    messageId: 'defineInitializerSuggestion',
                                    data: { name, suggested: `'${name}'` },
                                    fix: (fixer) => {
                                        return fixer.replaceText(member, `${name} = '${name}'`);
                                    },
                                },
                            ],
                        });
                    }
                });
            }
            return {
                TSEnumDeclaration,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-for-of.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'prefer-for-of',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Prefer a ‘for-of’ loop over a standard ‘for’ loop if the index is only used to access the array being iterated',
                category: 'Stylistic Issues',
                recommended: false,
            },
            messages: {
                preferForOf: 'Expected a `for-of` loop instead of a `for` loop with this simple iteration.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            function isSingleVariableDeclaration(node) {
                return (node !== null &&
                    node.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclaration &&
                    node.kind !== 'const' &&
                    node.declarations.length === 1);
            }
            function isLiteral(node, value) {
                return node.type === experimental_utils_1.AST_NODE_TYPES.Literal && node.value === value;
            }
            function isZeroInitialized(node) {
                return node.init !== null && isLiteral(node.init, 0);
            }
            function isMatchingIdentifier(node, name) {
                return node.type === experimental_utils_1.AST_NODE_TYPES.Identifier && node.name === name;
            }
            function isLessThanLengthExpression(node, name) {
                if (node !== null &&
                    node.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression &&
                    node.operator === '<' &&
                    isMatchingIdentifier(node.left, name) &&
                    node.right.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                    isMatchingIdentifier(node.right.property, 'length')) {
                    return node.right.object;
                }
                return null;
            }
            function isIncrement(node, name) {
                if (!node) {
                    return false;
                }
                switch (node.type) {
                    case experimental_utils_1.AST_NODE_TYPES.UpdateExpression:
                        // x++ or ++x
                        return (node.operator === '++' && isMatchingIdentifier(node.argument, name));
                    case experimental_utils_1.AST_NODE_TYPES.AssignmentExpression:
                        if (isMatchingIdentifier(node.left, name)) {
                            if (node.operator === '+=') {
                                // x += 1
                                return isLiteral(node.right, 1);
                            }
                            else if (node.operator === '=') {
                                // x = x + 1 or x = 1 + x
                                const expr = node.right;
                                return (expr.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression &&
                                    expr.operator === '+' &&
                                    ((isMatchingIdentifier(expr.left, name) &&
                                        isLiteral(expr.right, 1)) ||
                                        (isLiteral(expr.left, 1) &&
                                            isMatchingIdentifier(expr.right, name))));
                            }
                        }
                }
                return false;
            }
            function contains(outer, inner) {
                return (outer.range[0] <= inner.range[0] && outer.range[1] >= inner.range[1]);
            }
            function isAssignee(node) {
                var _a;
                const parent = node.parent;
                if (!parent) {
                    return false;
                }
                // a[i] = 1, a[i] += 1, etc.
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.AssignmentExpression &&
                    parent.left === node) {
                    return true;
                }
                // delete a[i]
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                    parent.operator === 'delete' &&
                    parent.argument === node) {
                    return true;
                }
                // a[i]++, --a[i], etc.
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.UpdateExpression &&
                    parent.argument === node) {
                    return true;
                }
                // [a[i]] = [0]
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.ArrayPattern) {
                    return true;
                }
                // [...a[i]] = [0]
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.RestElement) {
                    return true;
                }
                // ({ foo: a[i] }) = { foo: 0 }
                if (parent.type === experimental_utils_1.AST_NODE_TYPES.Property &&
                    parent.value === node &&
                    ((_a = parent.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.ObjectExpression &&
                    isAssignee(parent.parent)) {
                    return true;
                }
                return false;
            }
            function isIndexOnlyUsedWithArray(body, indexVar, arrayExpression) {
                const sourceCode = context.getSourceCode();
                const arrayText = sourceCode.getText(arrayExpression);
                return indexVar.references.every(reference => {
                    const id = reference.identifier;
                    const node = id.parent;
                    return (!contains(body, id) ||
                        (node !== undefined &&
                            node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                            node.property === id &&
                            sourceCode.getText(node.object) === arrayText &&
                            !isAssignee(node)));
                });
            }
            return {
                'ForStatement:exit'(node) {
                    if (!isSingleVariableDeclaration(node.init)) {
                        return;
                    }
                    const declarator = node.init.declarations[0];
                    if (!declarator ||
                        !isZeroInitialized(declarator) ||
                        declarator.id.type !== experimental_utils_1.AST_NODE_TYPES.Identifier) {
                        return;
                    }
                    const indexName = declarator.id.name;
                    const arrayExpression = isLessThanLengthExpression(node.test, indexName);
                    if (!arrayExpression) {
                        return;
                    }
                    const [indexVar] = context.getDeclaredVariables(node.init);
                    if (isIncrement(node.update, indexName) &&
                        isIndexOnlyUsedWithArray(node.body, indexVar, arrayExpression)) {
                        context.report({
                            node,
                            messageId: 'preferForOf',
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-function-type.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.phrases = {
        [experimental_utils_1.AST_NODE_TYPES.TSTypeLiteral]: 'Type literal',
        [experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration]: 'Interface',
    };
    exports.default = util.createRule({
        name: 'prefer-function-type',
        meta: {
            docs: {
                description: 'Use function types instead of interfaces with call signatures',
                category: 'Best Practices',
                recommended: false,
            },
            fixable: 'code',
            messages: {
                functionTypeOverCallableType: '{{ literalOrInterface }} only has a call signature, you should use a function type instead.',
                unexpectedThisOnFunctionOnlyInterface: "`this` refers to the function type '{{ interfaceName }}', did you intend to use a generic `this` parameter like `<Self>(this: Self, ...) => Self` instead?",
            },
            schema: [],
            type: 'suggestion',
        },
        defaultOptions: [],
        create(context) {
            const sourceCode = context.getSourceCode();
            /**
             * Checks if there the interface has exactly one supertype that isn't named 'Function'
             * @param node The node being checked
             */
            function hasOneSupertype(node) {
                if (!node.extends || node.extends.length === 0) {
                    return false;
                }
                if (node.extends.length !== 1) {
                    return true;
                }
                const expr = node.extends[0].expression;
                return (expr.type !== experimental_utils_1.AST_NODE_TYPES.Identifier || expr.name !== 'Function');
            }
            /**
             * @param parent The parent of the call signature causing the diagnostic
             */
            function shouldWrapSuggestion(parent) {
                if (!parent) {
                    return false;
                }
                switch (parent.type) {
                    case experimental_utils_1.AST_NODE_TYPES.TSUnionType:
                    case experimental_utils_1.AST_NODE_TYPES.TSIntersectionType:
                    case experimental_utils_1.AST_NODE_TYPES.TSArrayType:
                        return true;
                    default:
                        return false;
                }
            }
            /**
             * @param member The TypeElement being checked
             * @param node The parent of member being checked
             * @param tsThisTypes
             */
            function checkMember(member, node, tsThisTypes = null) {
                if ((member.type === experimental_utils_1.AST_NODE_TYPES.TSCallSignatureDeclaration ||
                    member.type === experimental_utils_1.AST_NODE_TYPES.TSConstructSignatureDeclaration) &&
                    typeof member.returnType !== 'undefined') {
                    if (tsThisTypes !== null &&
                        tsThisTypes.length > 0 &&
                        node.type === experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration) {
                        // the message can be confusing if we don't point directly to the `this` node instead of the whole member
                        // and in favour of generating at most one error we'll only report the first occurrence of `this` if there are multiple
                        context.report({
                            node: tsThisTypes[0],
                            messageId: 'unexpectedThisOnFunctionOnlyInterface',
                            data: {
                                interfaceName: node.id.name,
                            },
                        });
                        return;
                    }
                    const fixable = node.parent &&
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.ExportDefaultDeclaration;
                    context.report({
                        node: member,
                        messageId: 'functionTypeOverCallableType',
                        data: {
                            literalOrInterface: exports.phrases[node.type],
                        },
                        fix: fixable
                            ? null
                            : (fixer) => {
                                const fixes = [];
                                const start = member.range[0];
                                const colonPos = member.returnType.range[0] - start;
                                const text = sourceCode.getText().slice(start, member.range[1]);
                                const comments = sourceCode
                                    .getCommentsBefore(member)
                                    .concat(sourceCode.getCommentsAfter(member));
                                let suggestion = `${text.slice(0, colonPos)} =>${text.slice(colonPos + 1)}`;
                                const lastChar = suggestion.endsWith(';') ? ';' : '';
                                if (lastChar) {
                                    suggestion = suggestion.slice(0, -1);
                                }
                                if (shouldWrapSuggestion(node.parent)) {
                                    suggestion = `(${suggestion})`;
                                }
                                if (node.type === experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration) {
                                    if (typeof node.typeParameters !== 'undefined') {
                                        suggestion = `type ${sourceCode
                                            .getText()
                                            .slice(node.id.range[0], node.typeParameters.range[1])} = ${suggestion}${lastChar}`;
                                    }
                                    else {
                                        suggestion = `type ${node.id.name} = ${suggestion}${lastChar}`;
                                    }
                                }
                                const isParentExported = node.parent &&
                                    node.parent.type === experimental_utils_1.AST_NODE_TYPES.ExportNamedDeclaration;
                                if (node.type === experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration &&
                                    isParentExported) {
                                    const commentsText = comments.reduce((text, comment) => {
                                        return (text +
                                            (comment.type === experimental_utils_1.AST_TOKEN_TYPES.Line
                                                ? `//${comment.value}`
                                                : `/*${comment.value}*/`) +
                                            '\n');
                                    }, '');
                                    // comments should move before export and not between export and interface declaration
                                    fixes.push(fixer.insertTextBefore(node.parent, commentsText));
                                }
                                else {
                                    comments.forEach(comment => {
                                        let commentText = comment.type === experimental_utils_1.AST_TOKEN_TYPES.Line
                                            ? `//${comment.value}`
                                            : `/*${comment.value}*/`;
                                        const isCommentOnTheSameLine = comment.loc.start.line === member.loc.start.line;
                                        if (!isCommentOnTheSameLine) {
                                            commentText += '\n';
                                        }
                                        else {
                                            commentText += ' ';
                                        }
                                        suggestion = commentText + suggestion;
                                    });
                                }
                                const fixStart = node.range[0];
                                fixes.push(fixer.replaceTextRange([fixStart, node.range[1]], suggestion));
                                return fixes;
                            },
                    });
                }
            }
            let tsThisTypes = null;
            let literalNesting = 0;
            return {
                TSInterfaceDeclaration() {
                    // when entering an interface reset the count of `this`s to empty.
                    tsThisTypes = [];
                },
                'TSInterfaceDeclaration TSThisType'(node) {
                    // inside an interface keep track of all ThisType references.
                    // unless it's inside a nested type literal in which case it's invalid code anyway
                    // we don't want to incorrectly say "it refers to name" while typescript says it's completely invalid.
                    if (literalNesting === 0 && tsThisTypes !== null) {
                        tsThisTypes.push(node);
                    }
                },
                'TSInterfaceDeclaration:exit'(node) {
                    if (!hasOneSupertype(node) && node.body.body.length === 1) {
                        checkMember(node.body.body[0], node, tsThisTypes);
                    }
                    // on exit check member and reset the array to nothing.
                    tsThisTypes = null;
                },
                // keep track of nested literals to avoid complaining about invalid `this` uses
                'TSInterfaceDeclaration TSTypeLiteral'() {
                    literalNesting += 1;
                },
                'TSInterfaceDeclaration TSTypeLiteral:exit'() {
                    literalNesting -= 1;
                },
                'TSTypeLiteral[members.length = 1]'(node) {
                    checkMember(node.members[0], node);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-includes.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const regexpp_1 = require(497) /* regexpp */;
    const ts = __importStar(require(21) /* typescript */);
    const util_1 = require(906) /* ../util */;
    exports.default = util_1.createRule({
        name: 'prefer-includes',
        defaultOptions: [],
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforce `includes` method over `indexOf` method',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            fixable: 'code',
            messages: {
                preferIncludes: "Use 'includes()' method instead.",
                preferStringIncludes: 'Use `String#includes()` method with a string instead.',
            },
            schema: [],
        },
        create(context) {
            const globalScope = context.getScope();
            const services = util_1.getParserServices(context);
            const types = services.program.getTypeChecker();
            function isNumber(node, value) {
                const evaluated = util_1.getStaticValue(node, globalScope);
                return evaluated !== null && evaluated.value === value;
            }
            function isPositiveCheck(node) {
                switch (node.operator) {
                    case '!==':
                    case '!=':
                    case '>':
                        return isNumber(node.right, -1);
                    case '>=':
                        return isNumber(node.right, 0);
                    default:
                        return false;
                }
            }
            function isNegativeCheck(node) {
                switch (node.operator) {
                    case '===':
                    case '==':
                    case '<=':
                        return isNumber(node.right, -1);
                    case '<':
                        return isNumber(node.right, 0);
                    default:
                        return false;
                }
            }
            function hasSameParameters(nodeA, nodeB) {
                if (!ts.isFunctionLike(nodeA) || !ts.isFunctionLike(nodeB)) {
                    return false;
                }
                const paramsA = nodeA.parameters;
                const paramsB = nodeB.parameters;
                if (paramsA.length !== paramsB.length) {
                    return false;
                }
                for (let i = 0; i < paramsA.length; ++i) {
                    const paramA = paramsA[i];
                    const paramB = paramsB[i];
                    // Check name, type, and question token once.
                    if (paramA.getText() !== paramB.getText()) {
                        return false;
                    }
                }
                return true;
            }
            /**
             * Parse a given node if it's a `RegExp` instance.
             * @param node The node to parse.
             */
            function parseRegExp(node) {
                const evaluated = util_1.getStaticValue(node, globalScope);
                if (evaluated == null || !(evaluated.value instanceof RegExp)) {
                    return null;
                }
                const { pattern, flags } = regexpp_1.parseRegExpLiteral(evaluated.value);
                if (pattern.alternatives.length !== 1 ||
                    flags.ignoreCase ||
                    flags.global) {
                    return null;
                }
                // Check if it can determine a unique string.
                const chars = pattern.alternatives[0].elements;
                if (!chars.every(c => c.type === 'Character')) {
                    return null;
                }
                // To string.
                return String.fromCodePoint(...chars.map(c => c.value));
            }
            function checkArrayIndexOf(node, allowFixing) {
                var _a, _b, _c;
                // Check if the comparison is equivalent to `includes()`.
                const callNode = node.parent;
                const compareNode = (((_a = callNode.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.ChainExpression
                    ? callNode.parent.parent
                    : callNode.parent);
                const negative = isNegativeCheck(compareNode);
                if (!negative && !isPositiveCheck(compareNode)) {
                    return;
                }
                // Get the symbol of `indexOf` method.
                const tsNode = services.esTreeNodeToTSNodeMap.get(node.property);
                const indexofMethodDeclarations = (_b = types
                    .getSymbolAtLocation(tsNode)) === null || _b === void 0 ? void 0 : _b.getDeclarations();
                if (indexofMethodDeclarations == null ||
                    indexofMethodDeclarations.length === 0) {
                    return;
                }
                // Check if every declaration of `indexOf` method has `includes` method
                // and the two methods have the same parameters.
                for (const instanceofMethodDecl of indexofMethodDeclarations) {
                    const typeDecl = instanceofMethodDecl.parent;
                    const type = types.getTypeAtLocation(typeDecl);
                    const includesMethodDecl = (_c = type
                        .getProperty('includes')) === null || _c === void 0 ? void 0 : _c.getDeclarations();
                    if (includesMethodDecl == null ||
                        !includesMethodDecl.some(includesMethodDecl => hasSameParameters(includesMethodDecl, instanceofMethodDecl))) {
                        return;
                    }
                }
                // Report it.
                context.report(Object.assign({ node: compareNode, messageId: 'preferIncludes' }, (allowFixing && {
                    *fix(fixer) {
                        if (negative) {
                            yield fixer.insertTextBefore(callNode, '!');
                        }
                        yield fixer.replaceText(node.property, 'includes');
                        yield fixer.removeRange([callNode.range[1], compareNode.range[1]]);
                    },
                })));
            }
            return {
                // a.indexOf(b) !== 1
                "BinaryExpression > CallExpression.left > MemberExpression.callee[property.name='indexOf'][computed=false]"(node) {
                    checkArrayIndexOf(node, /* allowFixing */ true);
                },
                // a?.indexOf(b) !== 1
                "BinaryExpression > ChainExpression.left > CallExpression > MemberExpression.callee[property.name='indexOf'][computed=false]"(node) {
                    checkArrayIndexOf(node, /* allowFixing */ false);
                },
                // /bar/.test(foo)
                'CallExpression > MemberExpression.callee[property.name="test"][computed=false]'(node) {
                    var _a;
                    const callNode = node.parent;
                    const text = callNode.arguments.length === 1 ? parseRegExp(node.object) : null;
                    if (text == null) {
                        return;
                    }
                    //check the argument type of test methods
                    const argument = callNode.arguments[0];
                    const tsNode = services.esTreeNodeToTSNodeMap.get(argument);
                    const type = util_1.getConstrainedTypeAtLocation(types, tsNode);
                    const includesMethodDecl = (_a = type
                        .getProperty('includes')) === null || _a === void 0 ? void 0 : _a.getDeclarations();
                    if (includesMethodDecl == null) {
                        return;
                    }
                    context.report({
                        node: callNode,
                        messageId: 'preferStringIncludes',
                        *fix(fixer) {
                            const argNode = callNode.arguments[0];
                            const needsParen = argNode.type !== experimental_utils_1.AST_NODE_TYPES.Literal &&
                                argNode.type !== experimental_utils_1.AST_NODE_TYPES.TemplateLiteral &&
                                argNode.type !== experimental_utils_1.AST_NODE_TYPES.Identifier &&
                                argNode.type !== experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                                argNode.type !== experimental_utils_1.AST_NODE_TYPES.CallExpression;
                            yield fixer.removeRange([callNode.range[0], argNode.range[0]]);
                            if (needsParen) {
                                yield fixer.insertTextBefore(argNode, '(');
                                yield fixer.insertTextAfter(argNode, ')');
                            }
                            yield fixer.insertTextAfter(argNode, `${node.optional ? '?.' : '.'}includes('${text}'`);
                        },
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-literal-enum-member.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util_1 = require(906) /* ../util */;
    exports.default = util_1.createRule({
        name: 'prefer-literal-enum-member',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Require that all enum members be literal values to prevent unintended enum member name shadow issues',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: false,
            },
            messages: {
                notLiteral: `Explicit enum value must only be a literal value (string, number, boolean, etc).`,
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowBitwiseExpressions: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                allowBitwiseExpressions: false,
            },
        ],
        create(context, [{ allowBitwiseExpressions }]) {
            return {
                TSEnumMember(node) {
                    // If there is no initializer, then this node is just the name of the member, so ignore.
                    if (node.initializer == null) {
                        return;
                    }
                    // any old literal
                    if (node.initializer.type === experimental_utils_1.AST_NODE_TYPES.Literal) {
                        return;
                    }
                    // TemplateLiteral without expressions
                    if (node.initializer.type === experimental_utils_1.AST_NODE_TYPES.TemplateLiteral &&
                        node.initializer.expressions.length === 0) {
                        return;
                    }
                    // -1 and +1
                    if (node.initializer.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                        node.initializer.argument.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                        (['+', '-'].includes(node.initializer.operator) ||
                            (allowBitwiseExpressions && node.initializer.operator === '~'))) {
                        return;
                    }
                    if (allowBitwiseExpressions &&
                        node.initializer.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression &&
                        ['|', '&', '^', '<<', '>>', '>>>'].includes(node.initializer.operator) &&
                        node.initializer.left.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                        node.initializer.right.type === experimental_utils_1.AST_NODE_TYPES.Literal) {
                        return;
                    }
                    context.report({
                        node: node.id,
                        messageId: 'notLiteral',
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-namespace-keyword.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'prefer-namespace-keyword',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Require the use of the `namespace` keyword instead of the `module` keyword to declare custom TypeScript modules',
                category: 'Best Practices',
                recommended: 'error',
            },
            fixable: 'code',
            messages: {
                useNamespace: "Use 'namespace' instead of 'module' to declare custom TypeScript modules.",
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                TSModuleDeclaration(node) {
                    // Do nothing if the name is a string.
                    if (!node.id || node.id.type === experimental_utils_1.AST_NODE_TYPES.Literal) {
                        return;
                    }
                    // Get tokens of the declaration header.
                    const moduleType = sourceCode.getTokenBefore(node.id);
                    if (moduleType &&
                        moduleType.type === experimental_utils_1.AST_TOKEN_TYPES.Identifier &&
                        moduleType.value === 'module') {
                        context.report({
                            node,
                            messageId: 'useNamespace',
                            fix(fixer) {
                                return fixer.replaceText(moduleType, 'namespace');
                            },
                        });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-nullish-coalescing.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'prefer-nullish-coalescing',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforce the usage of the nullish coalescing operator instead of logical chaining',
                category: 'Best Practices',
                recommended: false,
                suggestion: true,
                requiresTypeChecking: true,
            },
            messages: {
                preferNullish: 'Prefer using nullish coalescing operator (`??`) instead of a logical or (`||`), as it is a safer operator.',
                suggestNullish: 'Fix to nullish coalescing operator (`??`).',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        ignoreConditionalTests: {
                            type: 'boolean',
                        },
                        ignoreMixedLogicalExpressions: {
                            type: 'boolean',
                        },
                        forceSuggestionFixer: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                ignoreConditionalTests: true,
                ignoreMixedLogicalExpressions: true,
            },
        ],
        create(context, [{ ignoreConditionalTests, ignoreMixedLogicalExpressions }]) {
            const parserServices = util.getParserServices(context);
            const sourceCode = context.getSourceCode();
            const checker = parserServices.program.getTypeChecker();
            return {
                'LogicalExpression[operator = "||"]'(node) {
                    const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                    const type = checker.getTypeAtLocation(tsNode.left);
                    const isNullish = util.isNullableType(type, { allowUndefined: true });
                    if (!isNullish) {
                        return;
                    }
                    if (ignoreConditionalTests === true && isConditionalTest(node)) {
                        return;
                    }
                    const isMixedLogical = isMixedLogicalExpression(node);
                    if (ignoreMixedLogicalExpressions === true && isMixedLogical) {
                        return;
                    }
                    const barBarOperator = util.nullThrows(sourceCode.getTokenAfter(node.left, token => token.type === experimental_utils_1.AST_TOKEN_TYPES.Punctuator &&
                        token.value === node.operator), util.NullThrowsReasons.MissingToken('operator', node.type));
                    function* fix(fixer) {
                        if (node.parent && util.isLogicalOrOperator(node.parent)) {
                            // '&&' and '??' operations cannot be mixed without parentheses (e.g. a && b ?? c)
                            if (node.left.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression &&
                                !util.isLogicalOrOperator(node.left.left)) {
                                yield fixer.insertTextBefore(node.left.right, '(');
                            }
                            else {
                                yield fixer.insertTextBefore(node.left, '(');
                            }
                            yield fixer.insertTextAfter(node.right, ')');
                        }
                        yield fixer.replaceText(barBarOperator, '??');
                    }
                    context.report({
                        node: barBarOperator,
                        messageId: 'preferNullish',
                        suggest: [
                            {
                                messageId: 'suggestNullish',
                                fix,
                            },
                        ],
                    });
                },
            };
        },
    });
    function isConditionalTest(node) {
        const parents = new Set([node]);
        let current = node.parent;
        while (current) {
            parents.add(current);
            if ((current.type === experimental_utils_1.AST_NODE_TYPES.ConditionalExpression ||
                current.type === experimental_utils_1.AST_NODE_TYPES.DoWhileStatement ||
                current.type === experimental_utils_1.AST_NODE_TYPES.IfStatement ||
                current.type === experimental_utils_1.AST_NODE_TYPES.ForStatement ||
                current.type === experimental_utils_1.AST_NODE_TYPES.WhileStatement) &&
                parents.has(current.test)) {
                return true;
            }
            if ([
                experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression,
                experimental_utils_1.AST_NODE_TYPES.FunctionExpression,
            ].includes(current.type)) {
                /**
                 * This is a weird situation like:
                 * `if (() => a || b) {}`
                 * `if (function () { return a || b }) {}`
                 */
                return false;
            }
            current = current.parent;
        }
        return false;
    }
    function isMixedLogicalExpression(node) {
        const seen = new Set();
        const queue = [node.parent, node.left, node.right];
        for (const current of queue) {
            if (seen.has(current)) {
                continue;
            }
            seen.add(current);
            if (current && current.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression) {
                if (current.operator === '&&') {
                    return true;
                }
                else if (current.operator === '||') {
                    // check the pieces of the node to catch cases like `a || b || c && d`
                    queue.push(current.parent, current.left, current.right);
                }
            }
        }
        return false;
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-optional-chain.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    /*
    The AST is always constructed such the first element is always the deepest element.
    
    I.e. for this code: `foo && foo.bar && foo.bar.baz && foo.bar.baz.buzz`
    The AST will look like this:
    {
      left: {
        left: {
          left: foo
          right: foo.bar
        }
        right: foo.bar.baz
      }
      right: foo.bar.baz.buzz
    }
    */
    exports.default = util.createRule({
        name: 'prefer-optional-chain',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Prefer using concise optional chain expressions instead of chained logical ands',
                category: 'Best Practices',
                recommended: false,
                suggestion: true,
            },
            messages: {
                preferOptionalChain: "Prefer using an optional chain expression instead, as it's more concise and easier to read.",
                optionalChainSuggest: 'Change to an optional chain.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const sourceCode = context.getSourceCode();
            return {
                [[
                    'LogicalExpression[operator="&&"] > Identifier',
                    'LogicalExpression[operator="&&"] > MemberExpression',
                    'LogicalExpression[operator="&&"] > ChainExpression > MemberExpression',
                    'LogicalExpression[operator="&&"] > BinaryExpression[operator="!=="]',
                    'LogicalExpression[operator="&&"] > BinaryExpression[operator="!="]',
                ].join(',')](initialIdentifierOrNotEqualsExpr) {
                    var _a;
                    // selector guarantees this cast
                    const initialExpression = (((_a = initialIdentifierOrNotEqualsExpr.parent) === null || _a === void 0 ? void 0 : _a.type) ===
                        experimental_utils_1.AST_NODE_TYPES.ChainExpression
                        ? initialIdentifierOrNotEqualsExpr.parent.parent
                        : initialIdentifierOrNotEqualsExpr.parent);
                    if (initialExpression.left !== initialIdentifierOrNotEqualsExpr) {
                        // the node(identifier or member expression) is not the deepest left node
                        return;
                    }
                    if (!isValidChainTarget(initialIdentifierOrNotEqualsExpr, true)) {
                        return;
                    }
                    // walk up the tree to figure out how many logical expressions we can include
                    let previous = initialExpression;
                    let current = initialExpression;
                    let previousLeftText = getText(initialIdentifierOrNotEqualsExpr);
                    let optionallyChainedCode = previousLeftText;
                    let expressionCount = 1;
                    while (current.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression) {
                        if (!isValidChainTarget(current.right, 
                        // only allow identifiers for the first chain - foo && foo()
                        expressionCount === 1)) {
                            break;
                        }
                        const leftText = previousLeftText;
                        const rightText = getText(current.right);
                        // can't just use startsWith because of cases like foo && fooBar.baz;
                        const matchRegex = new RegExp(`^${
                        // escape regex characters
                        leftText.replace(/[.*+?^${}()|[\]\\]/g, '\\$&')}[^a-zA-Z0-9_$]`);
                        if (!matchRegex.test(rightText) &&
                            // handle redundant cases like foo.bar && foo.bar
                            leftText !== rightText) {
                            break;
                        }
                        // omit weird doubled up expression that make no sense like foo.bar && foo.bar
                        if (rightText !== leftText) {
                            expressionCount += 1;
                            previousLeftText = rightText;
                            /*
                            Diff the left and right text to construct the fix string
                            There are the following cases:
                
                            1)
                            rightText === 'foo.bar.baz.buzz'
                            leftText === 'foo.bar.baz'
                            diff === '.buzz'
                
                            2)
                            rightText === 'foo.bar.baz.buzz()'
                            leftText === 'foo.bar.baz'
                            diff === '.buzz()'
                
                            3)
                            rightText === 'foo.bar.baz.buzz()'
                            leftText === 'foo.bar.baz.buzz'
                            diff === '()'
                
                            4)
                            rightText === 'foo.bar.baz[buzz]'
                            leftText === 'foo.bar.baz'
                            diff === '[buzz]'
                
                            5)
                            rightText === 'foo.bar.baz?.buzz'
                            leftText === 'foo.bar.baz'
                            diff === '?.buzz'
                            */
                            const diff = rightText.replace(leftText, '');
                            if (diff.startsWith('?')) {
                                // item was "pre optional chained"
                                optionallyChainedCode += diff;
                            }
                            else {
                                const needsDot = diff.startsWith('(') || diff.startsWith('[');
                                optionallyChainedCode += `?${needsDot ? '.' : ''}${diff}`;
                            }
                        }
                        previous = current;
                        current = util.nullThrows(current.parent, util.NullThrowsReasons.MissingParent);
                    }
                    if (expressionCount > 1) {
                        if (previous.right.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression) {
                            // case like foo && foo.bar !== someValue
                            optionallyChainedCode += ` ${previous.right.operator} ${sourceCode.getText(previous.right.right)}`;
                        }
                        context.report({
                            node: previous,
                            messageId: 'preferOptionalChain',
                            suggest: [
                                {
                                    messageId: 'optionalChainSuggest',
                                    fix: (fixer) => [
                                        fixer.replaceText(previous, optionallyChainedCode),
                                    ],
                                },
                            ],
                        });
                    }
                },
            };
            function getText(node) {
                if (node.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression) {
                    return getText(
                    // isValidChainTarget ensures this is type safe
                    node.left);
                }
                if (node.type === experimental_utils_1.AST_NODE_TYPES.CallExpression) {
                    const calleeText = getText(
                    // isValidChainTarget ensures this is type safe
                    node.callee);
                    // ensure that the call arguments are left untouched, or else we can break cases that _need_ whitespace:
                    // - JSX: <Foo Needs Space Between Attrs />
                    // - Unary Operators: typeof foo, await bar, delete baz
                    const closingParenToken = util.nullThrows(sourceCode.getLastToken(node), util.NullThrowsReasons.MissingToken('closing parenthesis', node.type));
                    const openingParenToken = util.nullThrows(sourceCode.getFirstTokenBetween(node.callee, closingParenToken, util.isOpeningParenToken), util.NullThrowsReasons.MissingToken('opening parenthesis', node.type));
                    const argumentsText = sourceCode.text.substring(openingParenToken.range[0], closingParenToken.range[1]);
                    return `${calleeText}${argumentsText}`;
                }
                if (node.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                    return node.name;
                }
                if (node.type === experimental_utils_1.AST_NODE_TYPES.ThisExpression) {
                    return 'this';
                }
                if (node.type === experimental_utils_1.AST_NODE_TYPES.ChainExpression) {
                    /* istanbul ignore if */ if (node.expression.type === experimental_utils_1.AST_NODE_TYPES.TSNonNullExpression) {
                        // this shouldn't happen
                        return '';
                    }
                    return getText(node.expression);
                }
                return getMemberExpressionText(node);
            }
            /**
             * Gets a normalized representation of the given MemberExpression
             */
            function getMemberExpressionText(node) {
                let objectText;
                // cases should match the list in ALLOWED_MEMBER_OBJECT_TYPES
                switch (node.object.type) {
                    case experimental_utils_1.AST_NODE_TYPES.CallExpression:
                    case experimental_utils_1.AST_NODE_TYPES.Identifier:
                        objectText = getText(node.object);
                        break;
                    case experimental_utils_1.AST_NODE_TYPES.MemberExpression:
                        objectText = getMemberExpressionText(node.object);
                        break;
                    case experimental_utils_1.AST_NODE_TYPES.ThisExpression:
                        objectText = getText(node.object);
                        break;
                    /* istanbul ignore next */
                    default:
                        throw new Error(`Unexpected member object type: ${node.object.type}`);
                }
                let propertyText;
                if (node.computed) {
                    // cases should match the list in ALLOWED_COMPUTED_PROP_TYPES
                    switch (node.property.type) {
                        case experimental_utils_1.AST_NODE_TYPES.Identifier:
                            propertyText = getText(node.property);
                            break;
                        case experimental_utils_1.AST_NODE_TYPES.Literal:
                        case experimental_utils_1.AST_NODE_TYPES.TemplateLiteral:
                            propertyText = sourceCode.getText(node.property);
                            break;
                        case experimental_utils_1.AST_NODE_TYPES.MemberExpression:
                            propertyText = getMemberExpressionText(node.property);
                            break;
                        /* istanbul ignore next */
                        default:
                            throw new Error(`Unexpected member property type: ${node.object.type}`);
                    }
                    return `${objectText}${node.optional ? '?.' : ''}[${propertyText}]`;
                }
                else {
                    // cases should match the list in ALLOWED_NON_COMPUTED_PROP_TYPES
                    switch (node.property.type) {
                        case experimental_utils_1.AST_NODE_TYPES.Identifier:
                            propertyText = getText(node.property);
                            break;
                        /* istanbul ignore next */
                        default:
                            throw new Error(`Unexpected member property type: ${node.object.type}`);
                    }
                    return `${objectText}${node.optional ? '?.' : '.'}${propertyText}`;
                }
            }
        },
    });
    const ALLOWED_MEMBER_OBJECT_TYPES = new Set([
        experimental_utils_1.AST_NODE_TYPES.CallExpression,
        experimental_utils_1.AST_NODE_TYPES.Identifier,
        experimental_utils_1.AST_NODE_TYPES.MemberExpression,
        experimental_utils_1.AST_NODE_TYPES.ThisExpression,
    ]);
    const ALLOWED_COMPUTED_PROP_TYPES = new Set([
        experimental_utils_1.AST_NODE_TYPES.Identifier,
        experimental_utils_1.AST_NODE_TYPES.Literal,
        experimental_utils_1.AST_NODE_TYPES.MemberExpression,
        experimental_utils_1.AST_NODE_TYPES.TemplateLiteral,
    ]);
    const ALLOWED_NON_COMPUTED_PROP_TYPES = new Set([
        experimental_utils_1.AST_NODE_TYPES.Identifier,
    ]);
    function isValidChainTarget(node, allowIdentifier) {
        if (node.type === experimental_utils_1.AST_NODE_TYPES.ChainExpression) {
            return isValidChainTarget(node.expression, allowIdentifier);
        }
        if (node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression) {
            const isObjectValid = ALLOWED_MEMBER_OBJECT_TYPES.has(node.object.type) &&
                // make sure to validate the expression is of our expected structure
                isValidChainTarget(node.object, true);
            const isPropertyValid = node.computed
                ? ALLOWED_COMPUTED_PROP_TYPES.has(node.property.type) &&
                    // make sure to validate the member expression is of our expected structure
                    (node.property.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression
                        ? isValidChainTarget(node.property, allowIdentifier)
                        : true)
                : ALLOWED_NON_COMPUTED_PROP_TYPES.has(node.property.type);
            return isObjectValid && isPropertyValid;
        }
        if (node.type === experimental_utils_1.AST_NODE_TYPES.CallExpression) {
            return isValidChainTarget(node.callee, allowIdentifier);
        }
        if (allowIdentifier &&
            (node.type === experimental_utils_1.AST_NODE_TYPES.Identifier ||
                node.type === experimental_utils_1.AST_NODE_TYPES.ThisExpression)) {
            return true;
        }
        /*
        special case for the following, where we only want the left
        - foo !== null
        - foo != null
        - foo !== undefined
        - foo != undefined
        */
        if (node.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression &&
            ['!==', '!='].includes(node.operator) &&
            isValidChainTarget(node.left, allowIdentifier)) {
            if (node.right.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                node.right.name === 'undefined') {
                return true;
            }
            if (node.right.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                node.right.value === null) {
                return true;
            }
        }
        return false;
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-readonly.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const tsutils = __importStar(require(916) /* tsutils */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    const util_1 = require(906) /* ../util */;
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const functionScopeBoundaries = [
        experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression,
        experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration,
        experimental_utils_1.AST_NODE_TYPES.FunctionExpression,
        experimental_utils_1.AST_NODE_TYPES.MethodDefinition,
    ].join(', ');
    exports.default = util.createRule({
        name: 'prefer-readonly',
        meta: {
            docs: {
                description: "Requires that private members are marked as `readonly` if they're never modified outside of the constructor",
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            fixable: 'code',
            messages: {
                preferReadonly: "Member '{{name}}' is never reassigned; mark it as `readonly`.",
            },
            schema: [
                {
                    allowAdditionalProperties: false,
                    properties: {
                        onlyInlineLambdas: {
                            type: 'boolean',
                        },
                    },
                    type: 'object',
                },
            ],
            type: 'suggestion',
        },
        defaultOptions: [{ onlyInlineLambdas: false }],
        create(context, [{ onlyInlineLambdas }]) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const classScopeStack = [];
            function handlePropertyAccessExpression(node, parent, classScope) {
                if (ts.isBinaryExpression(parent)) {
                    handleParentBinaryExpression(node, parent, classScope);
                    return;
                }
                if (ts.isDeleteExpression(parent) || isDestructuringAssignment(node)) {
                    classScope.addVariableModification(node);
                    return;
                }
                if (ts.isPostfixUnaryExpression(parent) ||
                    ts.isPrefixUnaryExpression(parent)) {
                    handleParentPostfixOrPrefixUnaryExpression(parent, classScope);
                }
            }
            function handleParentBinaryExpression(node, parent, classScope) {
                if (parent.left === node &&
                    tsutils.isAssignmentKind(parent.operatorToken.kind)) {
                    classScope.addVariableModification(node);
                }
            }
            function handleParentPostfixOrPrefixUnaryExpression(node, classScope) {
                if (node.operator === ts.SyntaxKind.PlusPlusToken ||
                    node.operator === ts.SyntaxKind.MinusMinusToken) {
                    classScope.addVariableModification(node.operand);
                }
            }
            function isDestructuringAssignment(node) {
                let current = node.parent;
                while (current) {
                    const parent = current.parent;
                    if (ts.isObjectLiteralExpression(parent) ||
                        ts.isArrayLiteralExpression(parent) ||
                        ts.isSpreadAssignment(parent) ||
                        (ts.isSpreadElement(parent) &&
                            ts.isArrayLiteralExpression(parent.parent))) {
                        current = parent;
                    }
                    else if (ts.isBinaryExpression(parent)) {
                        return (parent.left === current &&
                            parent.operatorToken.kind === ts.SyntaxKind.EqualsToken);
                    }
                    else {
                        break;
                    }
                }
                return false;
            }
            function isConstructor(node) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition &&
                    node.kind === 'constructor');
            }
            function isFunctionScopeBoundaryInStack(node) {
                if (classScopeStack.length === 0) {
                    return false;
                }
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                if (ts.isConstructorDeclaration(tsNode)) {
                    return false;
                }
                return tsutils.isFunctionScopeBoundary(tsNode);
            }
            function getEsNodesFromViolatingNode(violatingNode) {
                if (ts.isParameterPropertyDeclaration(violatingNode, violatingNode.parent)) {
                    return {
                        esNode: parserServices.tsNodeToESTreeNodeMap.get(violatingNode.name),
                        nameNode: parserServices.tsNodeToESTreeNodeMap.get(violatingNode.name),
                    };
                }
                return {
                    esNode: parserServices.tsNodeToESTreeNodeMap.get(violatingNode),
                    nameNode: parserServices.tsNodeToESTreeNodeMap.get(violatingNode.name),
                };
            }
            return {
                'ClassDeclaration, ClassExpression'(node) {
                    classScopeStack.push(new ClassScope(checker, parserServices.esTreeNodeToTSNodeMap.get(node), onlyInlineLambdas));
                },
                'ClassDeclaration, ClassExpression:exit'() {
                    const finalizedClassScope = classScopeStack.pop();
                    const sourceCode = context.getSourceCode();
                    for (const violatingNode of finalizedClassScope.finalizeUnmodifiedPrivateNonReadonlys()) {
                        const { esNode, nameNode } = getEsNodesFromViolatingNode(violatingNode);
                        context.report({
                            data: {
                                name: sourceCode.getText(nameNode),
                            },
                            fix: fixer => fixer.insertTextBefore(nameNode, 'readonly '),
                            messageId: 'preferReadonly',
                            node: esNode,
                        });
                    }
                },
                MemberExpression(node) {
                    if (classScopeStack.length !== 0 && !node.computed) {
                        const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                        handlePropertyAccessExpression(tsNode, tsNode.parent, classScopeStack[classScopeStack.length - 1]);
                    }
                },
                [functionScopeBoundaries](node) {
                    if (isConstructor(node)) {
                        classScopeStack[classScopeStack.length - 1].enterConstructor(parserServices.esTreeNodeToTSNodeMap.get(node));
                    }
                    else if (isFunctionScopeBoundaryInStack(node)) {
                        classScopeStack[classScopeStack.length - 1].enterNonConstructor();
                    }
                },
                [`${functionScopeBoundaries}:exit`](node) {
                    if (isConstructor(node)) {
                        classScopeStack[classScopeStack.length - 1].exitConstructor();
                    }
                    else if (isFunctionScopeBoundaryInStack(node)) {
                        classScopeStack[classScopeStack.length - 1].exitNonConstructor();
                    }
                },
            };
        },
    });
    const OUTSIDE_CONSTRUCTOR = -1;
    const DIRECTLY_INSIDE_CONSTRUCTOR = 0;
    class ClassScope {
        constructor(checker, classNode, onlyInlineLambdas) {
            this.checker = checker;
            this.onlyInlineLambdas = onlyInlineLambdas;
            this.privateModifiableMembers = new Map();
            this.privateModifiableStatics = new Map();
            this.memberVariableModifications = new Set();
            this.staticVariableModifications = new Set();
            this.constructorScopeDepth = OUTSIDE_CONSTRUCTOR;
            this.checker = checker;
            this.classType = checker.getTypeAtLocation(classNode);
            for (const member of classNode.members) {
                if (ts.isPropertyDeclaration(member)) {
                    this.addDeclaredVariable(member);
                }
            }
        }
        addDeclaredVariable(node) {
            if (!tsutils.isModifierFlagSet(node, ts.ModifierFlags.Private) ||
                tsutils.isModifierFlagSet(node, ts.ModifierFlags.Readonly) ||
                ts.isComputedPropertyName(node.name)) {
                return;
            }
            if (this.onlyInlineLambdas &&
                node.initializer !== undefined &&
                !ts.isArrowFunction(node.initializer)) {
                return;
            }
            (tsutils.isModifierFlagSet(node, ts.ModifierFlags.Static)
                ? this.privateModifiableStatics
                : this.privateModifiableMembers).set(node.name.getText(), node);
        }
        addVariableModification(node) {
            const modifierType = this.checker.getTypeAtLocation(node.expression);
            if (!modifierType.getSymbol() ||
                !util_1.typeIsOrHasBaseType(modifierType, this.classType)) {
                return;
            }
            const modifyingStatic = tsutils.isObjectType(modifierType) &&
                tsutils.isObjectFlagSet(modifierType, ts.ObjectFlags.Anonymous);
            if (!modifyingStatic &&
                this.constructorScopeDepth === DIRECTLY_INSIDE_CONSTRUCTOR) {
                return;
            }
            (modifyingStatic
                ? this.staticVariableModifications
                : this.memberVariableModifications).add(node.name.text);
        }
        enterConstructor(node) {
            this.constructorScopeDepth = DIRECTLY_INSIDE_CONSTRUCTOR;
            for (const parameter of node.parameters) {
                if (tsutils.isModifierFlagSet(parameter, ts.ModifierFlags.Private)) {
                    this.addDeclaredVariable(parameter);
                }
            }
        }
        exitConstructor() {
            this.constructorScopeDepth = OUTSIDE_CONSTRUCTOR;
        }
        enterNonConstructor() {
            if (this.constructorScopeDepth !== OUTSIDE_CONSTRUCTOR) {
                this.constructorScopeDepth += 1;
            }
        }
        exitNonConstructor() {
            if (this.constructorScopeDepth !== OUTSIDE_CONSTRUCTOR) {
                this.constructorScopeDepth -= 1;
            }
        }
        finalizeUnmodifiedPrivateNonReadonlys() {
            this.memberVariableModifications.forEach(variableName => {
                this.privateModifiableMembers.delete(variableName);
            });
            this.staticVariableModifications.forEach(variableName => {
                this.privateModifiableStatics.delete(variableName);
            });
            return [
                ...Array.from(this.privateModifiableMembers.values()),
                ...Array.from(this.privateModifiableStatics.values()),
            ];
        }
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-readonly-parameter-types.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'prefer-readonly-parameter-types',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Requires that function parameters are typed as readonly to prevent accidental mutation of inputs',
                category: 'Possible Errors',
                recommended: false,
                requiresTypeChecking: true,
            },
            schema: [
                {
                    type: 'object',
                    additionalProperties: false,
                    properties: {
                        checkParameterProperties: {
                            type: 'boolean',
                        },
                        ignoreInferredTypes: {
                            type: 'boolean',
                        },
                    },
                },
            ],
            messages: {
                shouldBeReadonly: 'Parameter should be a read only type.',
            },
        },
        defaultOptions: [
            {
                checkParameterProperties: true,
                ignoreInferredTypes: false,
            },
        ],
        create(context, options) {
            const [{ checkParameterProperties, ignoreInferredTypes }] = options;
            const { esTreeNodeToTSNodeMap, program } = util.getParserServices(context);
            const checker = program.getTypeChecker();
            return {
                [[
                    experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression,
                    experimental_utils_1.AST_NODE_TYPES.FunctionDeclaration,
                    experimental_utils_1.AST_NODE_TYPES.FunctionExpression,
                    experimental_utils_1.AST_NODE_TYPES.TSCallSignatureDeclaration,
                    experimental_utils_1.AST_NODE_TYPES.TSConstructSignatureDeclaration,
                    experimental_utils_1.AST_NODE_TYPES.TSDeclareFunction,
                    experimental_utils_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression,
                    experimental_utils_1.AST_NODE_TYPES.TSFunctionType,
                    experimental_utils_1.AST_NODE_TYPES.TSMethodSignature,
                ].join(', ')](node) {
                    for (const param of node.params) {
                        if (!checkParameterProperties &&
                            param.type === experimental_utils_1.AST_NODE_TYPES.TSParameterProperty) {
                            continue;
                        }
                        const actualParam = param.type === experimental_utils_1.AST_NODE_TYPES.TSParameterProperty
                            ? param.parameter
                            : param;
                        if (ignoreInferredTypes && actualParam.typeAnnotation == null) {
                            continue;
                        }
                        const tsNode = esTreeNodeToTSNodeMap.get(actualParam);
                        const type = checker.getTypeAtLocation(tsNode);
                        const isReadOnly = util.isTypeReadonly(checker, type);
                        if (!isReadOnly) {
                            context.report({
                                node: actualParam,
                                messageId: 'shouldBeReadonly',
                            });
                        }
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-reduce-type-parameter.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const getMemberExpressionName = (member) => {
        if (!member.computed) {
            return member.property.name;
        }
        if (member.property.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
            typeof member.property.value === 'string') {
            return member.property.value;
        }
        return null;
    };
    exports.default = util.createRule({
        name: 'prefer-reduce-type-parameter',
        meta: {
            type: 'problem',
            docs: {
                category: 'Best Practices',
                recommended: false,
                description: 'Prefer using type parameter when calling `Array#reduce` instead of casting',
                requiresTypeChecking: true,
            },
            messages: {
                preferTypeParameter: 'Unnecessary cast: Array#reduce accepts a type parameter for the default value.',
            },
            fixable: 'code',
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const service = util.getParserServices(context);
            const checker = service.program.getTypeChecker();
            return {
                'CallExpression > MemberExpression.callee'(callee) {
                    if (getMemberExpressionName(callee) !== 'reduce') {
                        return;
                    }
                    const [, secondArg] = callee.parent.arguments;
                    if (callee.parent.arguments.length < 2 ||
                        !util.isTypeAssertion(secondArg)) {
                        return;
                    }
                    // Get the symbol of the `reduce` method.
                    const tsNode = service.esTreeNodeToTSNodeMap.get(callee.object);
                    const calleeObjType = util.getConstrainedTypeAtLocation(checker, tsNode);
                    // Check the owner type of the `reduce` method.
                    if (checker.isArrayType(calleeObjType)) {
                        context.report({
                            messageId: 'preferTypeParameter',
                            node: secondArg,
                            fix: fixer => [
                                fixer.removeRange([
                                    secondArg.range[0],
                                    secondArg.expression.range[0],
                                ]),
                                fixer.removeRange([
                                    secondArg.expression.range[1],
                                    secondArg.range[1],
                                ]),
                                fixer.insertTextAfter(callee, `<${context
                                    .getSourceCode()
                                    .getText(secondArg.typeAnnotation)}>`),
                            ],
                        });
                        return;
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-regexp-exec.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const util_1 = require(906) /* ../util */;
    var ArgumentType;
    (function (ArgumentType) {
        ArgumentType[ArgumentType["Other"] = 0] = "Other";
        ArgumentType[ArgumentType["String"] = 1] = "String";
        ArgumentType[ArgumentType["RegExp"] = 2] = "RegExp";
        ArgumentType[ArgumentType["Both"] = 3] = "Both";
    })(ArgumentType || (ArgumentType = {}));
    exports.default = util_1.createRule({
        name: 'prefer-regexp-exec',
        defaultOptions: [],
        meta: {
            type: 'suggestion',
            fixable: 'code',
            docs: {
                description: 'Enforce that `RegExp#exec` is used instead of `String#match` if no global flag is provided',
                category: 'Best Practices',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                regExpExecOverStringMatch: 'Use the `RegExp#exec()` method instead.',
            },
            schema: [],
        },
        create(context) {
            const globalScope = context.getScope();
            const parserServices = util_1.getParserServices(context);
            const typeChecker = parserServices.program.getTypeChecker();
            const sourceCode = context.getSourceCode();
            /**
             * Check if a given node type is a string.
             * @param node The node type to check.
             */
            function isStringType(type) {
                return util_1.getTypeName(typeChecker, type) === 'string';
            }
            /**
             * Check if a given node type is a RegExp.
             * @param node The node type to check.
             */
            function isRegExpType(type) {
                return util_1.getTypeName(typeChecker, type) === 'RegExp';
            }
            function collectArgumentTypes(types) {
                let result = ArgumentType.Other;
                for (const type of types) {
                    if (isRegExpType(type)) {
                        result |= ArgumentType.RegExp;
                    }
                    else if (isStringType(type)) {
                        result |= ArgumentType.String;
                    }
                }
                return result;
            }
            return {
                "CallExpression[arguments.length=1] > MemberExpression.callee[property.name='match'][computed=false]"(memberNode) {
                    const objectNode = memberNode.object;
                    const callNode = memberNode.parent;
                    const argumentNode = callNode.arguments[0];
                    const argumentValue = util_1.getStaticValue(argumentNode, globalScope);
                    if (!isStringType(typeChecker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(objectNode)))) {
                        return;
                    }
                    // Don't report regular expressions with global flag.
                    if (argumentValue &&
                        argumentValue.value instanceof RegExp &&
                        argumentValue.value.flags.includes('g')) {
                        return;
                    }
                    if (argumentNode.type === experimental_utils_1.AST_NODE_TYPES.Literal &&
                        typeof argumentNode.value == 'string') {
                        const regExp = RegExp(argumentNode.value);
                        return context.report({
                            node: memberNode.property,
                            messageId: 'regExpExecOverStringMatch',
                            fix: util_1.getWrappingFixer({
                                sourceCode,
                                node: callNode,
                                innerNode: [objectNode],
                                wrap: objectCode => `${regExp.toString()}.exec(${objectCode})`,
                            }),
                        });
                    }
                    const argumentType = typeChecker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(argumentNode));
                    const argumentTypes = collectArgumentTypes(tsutils.unionTypeParts(argumentType));
                    switch (argumentTypes) {
                        case ArgumentType.RegExp:
                            return context.report({
                                node: memberNode.property,
                                messageId: 'regExpExecOverStringMatch',
                                fix: util_1.getWrappingFixer({
                                    sourceCode,
                                    node: callNode,
                                    innerNode: [objectNode, argumentNode],
                                    wrap: (objectCode, argumentCode) => `${argumentCode}.exec(${objectCode})`,
                                }),
                            });
                        case ArgumentType.String:
                            return context.report({
                                node: memberNode.property,
                                messageId: 'regExpExecOverStringMatch',
                                fix: util_1.getWrappingFixer({
                                    sourceCode,
                                    node: callNode,
                                    innerNode: [objectNode, argumentNode],
                                    wrap: (objectCode, argumentCode) => `RegExp(${argumentCode}).exec(${objectCode})`,
                                }),
                            });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-return-this-type.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util_1 = require(906) /* ../util */;
    const ts = __importStar(require(21) /* typescript */);
    exports.default = util_1.createRule({
        name: 'prefer-return-this-type',
        defaultOptions: [],
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforce that `this` is used when only `this` type is returned',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            messages: {
                useThisType: 'use `this` type instead.',
            },
            schema: [],
            fixable: 'code',
        },
        create(context) {
            const parserServices = util_1.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            function tryGetNameInType(name, typeNode) {
                if (typeNode.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference &&
                    typeNode.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    typeNode.typeName.name === name) {
                    return typeNode.typeName;
                }
                if (typeNode.type === experimental_utils_1.AST_NODE_TYPES.TSUnionType) {
                    for (const type of typeNode.types) {
                        const found = tryGetNameInType(name, type);
                        if (found) {
                            return found;
                        }
                    }
                }
                return undefined;
            }
            function isThisSpecifiedInParameters(originalFunc) {
                const firstArg = originalFunc.params[0];
                return (firstArg &&
                    firstArg.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    firstArg.name === 'this');
            }
            function isFunctionReturningThis(originalFunc, originalClass) {
                if (isThisSpecifiedInParameters(originalFunc)) {
                    return false;
                }
                const func = parserServices.esTreeNodeToTSNodeMap.get(originalFunc);
                if (!func.body) {
                    return false;
                }
                const classType = checker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(originalClass));
                if (func.body.kind !== ts.SyntaxKind.Block) {
                    const type = checker.getTypeAtLocation(func.body);
                    return classType.thisType === type;
                }
                let hasReturnThis = false;
                let hasReturnClassType = false;
                util_1.forEachReturnStatement(func.body, stmt => {
                    const expr = stmt.expression;
                    if (!expr) {
                        return;
                    }
                    // fast check
                    if (expr.kind === ts.SyntaxKind.ThisKeyword) {
                        hasReturnThis = true;
                        return;
                    }
                    const type = checker.getTypeAtLocation(expr);
                    if (classType === type) {
                        hasReturnClassType = true;
                        return true;
                    }
                    if (classType.thisType === type) {
                        hasReturnThis = true;
                        return;
                    }
                    return;
                });
                return !hasReturnClassType && hasReturnThis;
            }
            function checkFunction(originalFunc, originalClass) {
                var _a;
                const className = (_a = originalClass.id) === null || _a === void 0 ? void 0 : _a.name;
                if (!className) {
                    return;
                }
                if (!originalFunc.returnType) {
                    return;
                }
                const classNameRef = tryGetNameInType(className, originalFunc.returnType.typeAnnotation);
                if (!classNameRef) {
                    return;
                }
                if (isFunctionReturningThis(originalFunc, originalClass)) {
                    context.report({
                        node: classNameRef,
                        messageId: 'useThisType',
                        fix(fixer) {
                            return fixer.replaceText(classNameRef, 'this');
                        },
                    });
                }
            }
            return {
                'ClassBody > MethodDefinition'(node) {
                    checkFunction(node.value, node.parent.parent);
                },
                'ClassBody > ClassProperty'(node) {
                    var _a, _b;
                    if (!(((_a = node.value) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.FunctionExpression ||
                        ((_b = node.value) === null || _b === void 0 ? void 0 : _b.type) === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression)) {
                        return;
                    }
                    checkFunction(node.value, node.parent.parent);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-string-starts-ends-with.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const regexpp_1 = require(497) /* regexpp */;
    const util_1 = require(906) /* ../util */;
    const EQ_OPERATORS = /^[=!]=/;
    const regexpp = new regexpp_1.RegExpParser();
    exports.default = util_1.createRule({
        name: 'prefer-string-starts-ends-with',
        defaultOptions: [],
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforce the use of `String#startsWith` and `String#endsWith` instead of other equivalent methods of checking substrings',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            messages: {
                preferStartsWith: "Use 'String#startsWith' method instead.",
                preferEndsWith: "Use the 'String#endsWith' method instead.",
            },
            schema: [],
            fixable: 'code',
        },
        create(context) {
            const globalScope = context.getScope();
            const sourceCode = context.getSourceCode();
            const service = util_1.getParserServices(context);
            const typeChecker = service.program.getTypeChecker();
            /**
             * Check if a given node is a string.
             * @param node The node to check.
             */
            function isStringType(node) {
                const objectType = typeChecker.getTypeAtLocation(service.esTreeNodeToTSNodeMap.get(node));
                return util_1.getTypeName(typeChecker, objectType) === 'string';
            }
            /**
             * Check if a given node is a `Literal` node that is null.
             * @param node The node to check.
             */
            function isNull(node) {
                const evaluated = util_1.getStaticValue(node, globalScope);
                return evaluated != null && evaluated.value === null;
            }
            /**
             * Check if a given node is a `Literal` node that is a given value.
             * @param node The node to check.
             * @param value The expected value of the `Literal` node.
             */
            function isNumber(node, value) {
                const evaluated = util_1.getStaticValue(node, globalScope);
                return evaluated != null && evaluated.value === value;
            }
            /**
             * Check if a given node is a `Literal` node that is a character.
             * @param node The node to check.
             * @param kind The method name to get a character.
             */
            function isCharacter(node) {
                const evaluated = util_1.getStaticValue(node, globalScope);
                return (evaluated != null &&
                    typeof evaluated.value === 'string' &&
                    // checks if the string is a character long
                    evaluated.value[0] === evaluated.value);
            }
            /**
             * Check if a given node is `==`, `===`, `!=`, or `!==`.
             * @param node The node to check.
             */
            function isEqualityComparison(node) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression &&
                    EQ_OPERATORS.test(node.operator));
            }
            /**
             * Check if two given nodes are the same meaning.
             * @param node1 A node to compare.
             * @param node2 Another node to compare.
             */
            function isSameTokens(node1, node2) {
                const tokens1 = sourceCode.getTokens(node1);
                const tokens2 = sourceCode.getTokens(node2);
                if (tokens1.length !== tokens2.length) {
                    return false;
                }
                for (let i = 0; i < tokens1.length; ++i) {
                    const token1 = tokens1[i];
                    const token2 = tokens2[i];
                    if (token1.type !== token2.type || token1.value !== token2.value) {
                        return false;
                    }
                }
                return true;
            }
            /**
             * Check if a given node is the expression of the length of a string.
             *
             * - If `length` property access of `expectedObjectNode`, it's `true`.
             *   E.g., `foo` → `foo.length` / `"foo"` → `"foo".length`
             * - If `expectedObjectNode` is a string literal, `node` can be a number.
             *   E.g., `"foo"` → `3`
             *
             * @param node The node to check.
             * @param expectedObjectNode The node which is expected as the receiver of `length` property.
             */
            function isLengthExpression(node, expectedObjectNode) {
                if (node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression) {
                    return (util_1.getPropertyName(node, globalScope) === 'length' &&
                        isSameTokens(node.object, expectedObjectNode));
                }
                const evaluatedLength = util_1.getStaticValue(node, globalScope);
                const evaluatedString = util_1.getStaticValue(expectedObjectNode, globalScope);
                return (evaluatedLength != null &&
                    evaluatedString != null &&
                    typeof evaluatedLength.value === 'number' &&
                    typeof evaluatedString.value === 'string' &&
                    evaluatedLength.value === evaluatedString.value.length);
            }
            /**
             * Check if a given node is a negative index expression
             *
             * E.g. `s.slice(- <expr>)`, `s.substring(s.length - <expr>)`
             *
             * @param node The node to check.
             * @param expectedIndexedNode The node which is expected as the receiver of index expression.
             */
            function isNegativeIndexExpression(node, expectedIndexedNode) {
                return ((node.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                    node.operator === '-') ||
                    (node.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression &&
                        node.operator === '-' &&
                        isLengthExpression(node.left, expectedIndexedNode)));
            }
            /**
             * Check if a given node is the expression of the last index.
             *
             * E.g. `foo.length - 1`
             *
             * @param node The node to check.
             * @param expectedObjectNode The node which is expected as the receiver of `length` property.
             */
            function isLastIndexExpression(node, expectedObjectNode) {
                return (node.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression &&
                    node.operator === '-' &&
                    isLengthExpression(node.left, expectedObjectNode) &&
                    isNumber(node.right, 1));
            }
            /**
             * Get the range of the property of a given `MemberExpression` node.
             *
             * - `obj[foo]` → the range of `[foo]`
             * - `obf.foo` → the range of `.foo`
             * - `(obj).foo` → the range of `.foo`
             *
             * @param node The member expression node to get.
             */
            function getPropertyRange(node) {
                const dotOrOpenBracket = sourceCode.getTokenAfter(node.object, util_1.isNotClosingParenToken);
                return [dotOrOpenBracket.range[0], node.range[1]];
            }
            /**
             * Parse a given `RegExp` pattern to that string if it's a static string.
             * @param pattern The RegExp pattern text to parse.
             * @param uFlag The Unicode flag of the RegExp.
             */
            function parseRegExpText(pattern, uFlag) {
                // Parse it.
                const ast = regexpp.parsePattern(pattern, undefined, undefined, uFlag);
                if (ast.alternatives.length !== 1) {
                    return null;
                }
                // Drop `^`/`$` assertion.
                const chars = ast.alternatives[0].elements;
                const first = chars[0];
                if (first.type === 'Assertion' && first.kind === 'start') {
                    chars.shift();
                }
                else {
                    chars.pop();
                }
                // Check if it can determine a unique string.
                if (!chars.every(c => c.type === 'Character')) {
                    return null;
                }
                // To string.
                return String.fromCodePoint(...chars.map(c => c.value));
            }
            /**
             * Parse a given node if it's a `RegExp` instance.
             * @param node The node to parse.
             */
            function parseRegExp(node) {
                const evaluated = util_1.getStaticValue(node, globalScope);
                if (evaluated == null || !(evaluated.value instanceof RegExp)) {
                    return null;
                }
                const { source, flags } = evaluated.value;
                const isStartsWith = source.startsWith('^');
                const isEndsWith = source.endsWith('$');
                if (isStartsWith === isEndsWith ||
                    flags.includes('i') ||
                    flags.includes('m')) {
                    return null;
                }
                const text = parseRegExpText(source, flags.includes('u'));
                if (text == null) {
                    return null;
                }
                return { isEndsWith, isStartsWith, text };
            }
            function getLeftNode(node) {
                if (node.type === experimental_utils_1.AST_NODE_TYPES.ChainExpression) {
                    return getLeftNode(node.expression);
                }
                let leftNode;
                if (node.type === experimental_utils_1.AST_NODE_TYPES.CallExpression) {
                    leftNode = node.callee;
                }
                else {
                    leftNode = node;
                }
                if (leftNode.type !== experimental_utils_1.AST_NODE_TYPES.MemberExpression) {
                    throw new Error(`Expected a MemberExpression, got ${leftNode.type}`);
                }
                return leftNode;
            }
            /**
             * Fix code with using the right operand as the search string.
             * For example: `foo.slice(0, 3) === 'bar'` → `foo.startsWith('bar')`
             * @param fixer The rule fixer.
             * @param node The node which was reported.
             * @param kind The kind of the report.
             * @param isNegative The flag to fix to negative condition.
             */
            function* fixWithRightOperand(fixer, node, kind, isNegative, isOptional) {
                // left is CallExpression or MemberExpression.
                const leftNode = getLeftNode(node.left);
                const propertyRange = getPropertyRange(leftNode);
                if (isNegative) {
                    yield fixer.insertTextBefore(node, '!');
                }
                yield fixer.replaceTextRange([propertyRange[0], node.right.range[0]], `${isOptional ? '?.' : '.'}${kind}sWith(`);
                yield fixer.replaceTextRange([node.right.range[1], node.range[1]], ')');
            }
            /**
             * Fix code with using the first argument as the search string.
             * For example: `foo.indexOf('bar') === 0` → `foo.startsWith('bar')`
             * @param fixer The rule fixer.
             * @param node The node which was reported.
             * @param kind The kind of the report.
             * @param negative The flag to fix to negative condition.
             */
            function* fixWithArgument(fixer, node, callNode, calleeNode, kind, negative, isOptional) {
                if (negative) {
                    yield fixer.insertTextBefore(node, '!');
                }
                yield fixer.replaceTextRange(getPropertyRange(calleeNode), `${isOptional ? '?.' : '.'}${kind}sWith`);
                yield fixer.removeRange([callNode.range[1], node.range[1]]);
            }
            function getParent(node) {
                var _a;
                return util_1.nullThrows(((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.ChainExpression
                    ? node.parent.parent
                    : node.parent, util_1.NullThrowsReasons.MissingParent);
            }
            return {
                // foo[0] === "a"
                // foo.charAt(0) === "a"
                // foo[foo.length - 1] === "a"
                // foo.charAt(foo.length - 1) === "a"
                [[
                    'BinaryExpression > MemberExpression.left[computed=true]',
                    'BinaryExpression > CallExpression.left > MemberExpression.callee[property.name="charAt"][computed=false]',
                    'BinaryExpression > ChainExpression.left > MemberExpression[computed=true]',
                    'BinaryExpression > ChainExpression.left > CallExpression > MemberExpression.callee[property.name="charAt"][computed=false]',
                ].join(', ')](node) {
                    let parentNode = getParent(node);
                    let indexNode = null;
                    if ((parentNode === null || parentNode === void 0 ? void 0 : parentNode.type) === experimental_utils_1.AST_NODE_TYPES.CallExpression) {
                        if (parentNode.arguments.length === 1) {
                            indexNode = parentNode.arguments[0];
                        }
                        parentNode = getParent(parentNode);
                    }
                    else {
                        indexNode = node.property;
                    }
                    if (indexNode == null ||
                        !isEqualityComparison(parentNode) ||
                        !isStringType(node.object)) {
                        return;
                    }
                    const isEndsWith = isLastIndexExpression(indexNode, node.object);
                    const isStartsWith = !isEndsWith && isNumber(indexNode, 0);
                    if (!isStartsWith && !isEndsWith) {
                        return;
                    }
                    const eqNode = parentNode;
                    context.report({
                        node: parentNode,
                        messageId: isStartsWith ? 'preferStartsWith' : 'preferEndsWith',
                        fix(fixer) {
                            // Don't fix if it can change the behavior.
                            if (!isCharacter(eqNode.right)) {
                                return null;
                            }
                            return fixWithRightOperand(fixer, eqNode, isStartsWith ? 'start' : 'end', eqNode.operator.startsWith('!'), node.optional);
                        },
                    });
                },
                // foo.indexOf('bar') === 0
                [[
                    'BinaryExpression > CallExpression.left > MemberExpression.callee[property.name="indexOf"][computed=false]',
                    'BinaryExpression > ChainExpression.left > CallExpression > MemberExpression.callee[property.name="indexOf"][computed=false]',
                ].join(', ')](node) {
                    const callNode = getParent(node);
                    const parentNode = getParent(callNode);
                    if (callNode.arguments.length !== 1 ||
                        !isEqualityComparison(parentNode) ||
                        !isNumber(parentNode.right, 0) ||
                        !isStringType(node.object)) {
                        return;
                    }
                    context.report({
                        node: parentNode,
                        messageId: 'preferStartsWith',
                        fix(fixer) {
                            return fixWithArgument(fixer, parentNode, callNode, node, 'start', parentNode.operator.startsWith('!'), node.optional);
                        },
                    });
                },
                // foo.lastIndexOf('bar') === foo.length - 3
                // foo.lastIndexOf(bar) === foo.length - bar.length
                [[
                    'BinaryExpression > CallExpression.left > MemberExpression.callee[property.name="lastIndexOf"][computed=false]',
                    'BinaryExpression > ChainExpression.left > CallExpression > MemberExpression.callee[property.name="lastIndexOf"][computed=false]',
                ].join(', ')](node) {
                    const callNode = getParent(node);
                    const parentNode = getParent(callNode);
                    if (callNode.arguments.length !== 1 ||
                        !isEqualityComparison(parentNode) ||
                        parentNode.right.type !== experimental_utils_1.AST_NODE_TYPES.BinaryExpression ||
                        parentNode.right.operator !== '-' ||
                        !isLengthExpression(parentNode.right.left, node.object) ||
                        !isLengthExpression(parentNode.right.right, callNode.arguments[0]) ||
                        !isStringType(node.object)) {
                        return;
                    }
                    context.report({
                        node: parentNode,
                        messageId: 'preferEndsWith',
                        fix(fixer) {
                            return fixWithArgument(fixer, parentNode, callNode, node, 'end', parentNode.operator.startsWith('!'), node.optional);
                        },
                    });
                },
                // foo.match(/^bar/) === null
                // foo.match(/bar$/) === null
                [[
                    'BinaryExpression > CallExpression.left > MemberExpression.callee[property.name="match"][computed=false]',
                    'BinaryExpression > ChainExpression.left > CallExpression > MemberExpression.callee[property.name="match"][computed=false]',
                ].join(', ')](node) {
                    const callNode = getParent(node);
                    const parentNode = getParent(callNode);
                    if (!isEqualityComparison(parentNode) ||
                        !isNull(parentNode.right) ||
                        !isStringType(node.object)) {
                        return;
                    }
                    const parsed = callNode.arguments.length === 1
                        ? parseRegExp(callNode.arguments[0])
                        : null;
                    if (parsed == null) {
                        return;
                    }
                    const { isStartsWith, text } = parsed;
                    context.report({
                        node: callNode,
                        messageId: isStartsWith ? 'preferStartsWith' : 'preferEndsWith',
                        *fix(fixer) {
                            if (!parentNode.operator.startsWith('!')) {
                                yield fixer.insertTextBefore(parentNode, '!');
                            }
                            yield fixer.replaceTextRange(getPropertyRange(node), `${node.optional ? '?.' : '.'}${isStartsWith ? 'start' : 'end'}sWith`);
                            yield fixer.replaceText(callNode.arguments[0], JSON.stringify(text));
                            yield fixer.removeRange([callNode.range[1], parentNode.range[1]]);
                        },
                    });
                },
                // foo.slice(0, 3) === 'bar'
                // foo.slice(-3) === 'bar'
                // foo.slice(-3, foo.length) === 'bar'
                // foo.substring(0, 3) === 'bar'
                // foo.substring(foo.length - 3) === 'bar'
                // foo.substring(foo.length - 3, foo.length) === 'bar'
                [[
                    'BinaryExpression > CallExpression.left > MemberExpression.callee[property.name="slice"][computed=false]',
                    'BinaryExpression > CallExpression.left > MemberExpression.callee[property.name="substring"][computed=false]',
                    'BinaryExpression > ChainExpression.left > CallExpression > MemberExpression.callee[property.name="slice"][computed=false]',
                    'BinaryExpression > ChainExpression.left > CallExpression > MemberExpression.callee[property.name="substring"][computed=false]',
                ].join(', ')](node) {
                    const callNode = getParent(node);
                    const parentNode = getParent(callNode);
                    if (!isEqualityComparison(parentNode) || !isStringType(node.object)) {
                        return;
                    }
                    const isEndsWith = (callNode.arguments.length === 1 ||
                        (callNode.arguments.length === 2 &&
                            isLengthExpression(callNode.arguments[1], node.object))) &&
                        isNegativeIndexExpression(callNode.arguments[0], node.object);
                    const isStartsWith = !isEndsWith &&
                        callNode.arguments.length === 2 &&
                        isNumber(callNode.arguments[0], 0) &&
                        !isNegativeIndexExpression(callNode.arguments[1], node.object);
                    if (!isStartsWith && !isEndsWith) {
                        return;
                    }
                    const eqNode = parentNode;
                    const negativeIndexSupported = node.property.name === 'slice';
                    context.report({
                        node: parentNode,
                        messageId: isStartsWith ? 'preferStartsWith' : 'preferEndsWith',
                        fix(fixer) {
                            // Don't fix if it can change the behavior.
                            if (eqNode.operator.length === 2 &&
                                (eqNode.right.type !== experimental_utils_1.AST_NODE_TYPES.Literal ||
                                    typeof eqNode.right.value !== 'string')) {
                                return null;
                            }
                            // code being checked is likely mistake:
                            // unequal length of strings being checked for equality
                            // or reliant on behavior of substring (negative indices interpreted as 0)
                            if (isStartsWith) {
                                if (!isLengthExpression(callNode.arguments[1], eqNode.right)) {
                                    return null;
                                }
                            }
                            else {
                                const posNode = callNode.arguments[0];
                                const posNodeIsAbsolutelyValid = (posNode.type === experimental_utils_1.AST_NODE_TYPES.BinaryExpression &&
                                    posNode.operator === '-' &&
                                    isLengthExpression(posNode.left, node.object) &&
                                    isLengthExpression(posNode.right, eqNode.right)) ||
                                    (negativeIndexSupported &&
                                        posNode.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression &&
                                        posNode.operator === '-' &&
                                        isLengthExpression(posNode.argument, eqNode.right));
                                if (!posNodeIsAbsolutelyValid) {
                                    return null;
                                }
                            }
                            return fixWithRightOperand(fixer, parentNode, isStartsWith ? 'start' : 'end', parentNode.operator.startsWith('!'), node.optional);
                        },
                    });
                },
                // /^bar/.test(foo)
                // /bar$/.test(foo)
                'CallExpression > MemberExpression.callee[property.name="test"][computed=false]'(node) {
                    const callNode = getParent(node);
                    const parsed = callNode.arguments.length === 1 ? parseRegExp(node.object) : null;
                    if (parsed == null) {
                        return;
                    }
                    const { isStartsWith, text } = parsed;
                    const messageId = isStartsWith ? 'preferStartsWith' : 'preferEndsWith';
                    const methodName = isStartsWith ? 'startsWith' : 'endsWith';
                    context.report({
                        node: callNode,
                        messageId,
                        *fix(fixer) {
                            const argNode = callNode.arguments[0];
                            const needsParen = argNode.type !== experimental_utils_1.AST_NODE_TYPES.Literal &&
                                argNode.type !== experimental_utils_1.AST_NODE_TYPES.TemplateLiteral &&
                                argNode.type !== experimental_utils_1.AST_NODE_TYPES.Identifier &&
                                argNode.type !== experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                                argNode.type !== experimental_utils_1.AST_NODE_TYPES.CallExpression;
                            yield fixer.removeRange([callNode.range[0], argNode.range[0]]);
                            if (needsParen) {
                                yield fixer.insertTextBefore(argNode, '(');
                                yield fixer.insertTextAfter(argNode, ')');
                            }
                            yield fixer.insertTextAfter(argNode, `${node.optional ? '?.' : '.'}${methodName}(${JSON.stringify(text)}`);
                        },
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/prefer-ts-expect-error.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    exports.default = util.createRule({
        name: 'prefer-ts-expect-error',
        meta: {
            type: 'problem',
            docs: {
                description: 'Recommends using `@ts-expect-error` over `@ts-ignore`',
                category: 'Best Practices',
                recommended: false,
            },
            fixable: 'code',
            messages: {
                preferExpectErrorComment: 'Use "@ts-expect-error" to ensure an error is actually being suppressed.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const tsIgnoreRegExpSingleLine = /^\s*\/?\s*@ts-ignore/;
            const tsIgnoreRegExpMultiLine = /^\s*(?:\/|\*)*\s*@ts-ignore/;
            const sourceCode = context.getSourceCode();
            function isLineComment(comment) {
                return comment.type === experimental_utils_1.AST_TOKEN_TYPES.Line;
            }
            function getLastCommentLine(comment) {
                if (isLineComment(comment)) {
                    return comment.value;
                }
                // For multiline comments - we look at only the last line.
                const commentlines = comment.value.split('\n');
                return commentlines[commentlines.length - 1];
            }
            function isValidTsIgnorePresent(comment) {
                const line = getLastCommentLine(comment);
                return isLineComment(comment)
                    ? tsIgnoreRegExpSingleLine.test(line)
                    : tsIgnoreRegExpMultiLine.test(line);
            }
            return {
                Program() {
                    const comments = sourceCode.getAllComments();
                    comments.forEach(comment => {
                        if (isValidTsIgnorePresent(comment)) {
                            const lineCommentRuleFixer = (fixer) => fixer.replaceText(comment, `//${comment.value.replace('@ts-ignore', '@ts-expect-error')}`);
                            const blockCommentRuleFixer = (fixer) => fixer.replaceText(comment, `/*${comment.value.replace('@ts-ignore', '@ts-expect-error')}*/`);
                            context.report({
                                node: comment,
                                messageId: 'preferExpectErrorComment',
                                fix: isLineComment(comment)
                                    ? lineCommentRuleFixer
                                    : blockCommentRuleFixer,
                            });
                        }
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/promise-function-async.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'promise-function-async',
        meta: {
            type: 'suggestion',
            fixable: 'code',
            docs: {
                description: 'Requires any function or method that returns a Promise to be marked async',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            messages: {
                missingAsync: 'Functions that return promises must be async.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowAny: {
                            type: 'boolean',
                        },
                        allowedPromiseNames: {
                            type: 'array',
                            items: {
                                type: 'string',
                            },
                        },
                        checkArrowFunctions: {
                            type: 'boolean',
                        },
                        checkFunctionDeclarations: {
                            type: 'boolean',
                        },
                        checkFunctionExpressions: {
                            type: 'boolean',
                        },
                        checkMethodDeclarations: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                allowAny: true,
                allowedPromiseNames: [],
                checkArrowFunctions: true,
                checkFunctionDeclarations: true,
                checkFunctionExpressions: true,
                checkMethodDeclarations: true,
            },
        ],
        create(context, [{ allowAny, allowedPromiseNames, checkArrowFunctions, checkFunctionDeclarations, checkFunctionExpressions, checkMethodDeclarations, },]) {
            const allAllowedPromiseNames = new Set([
                'Promise',
                ...allowedPromiseNames,
            ]);
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const sourceCode = context.getSourceCode();
            function validateNode(node) {
                var _a;
                const originalNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                const signatures = checker
                    .getTypeAtLocation(originalNode)
                    .getCallSignatures();
                if (!signatures.length) {
                    return;
                }
                const returnType = checker.getReturnTypeOfSignature(signatures[0]);
                if (!util.containsAllTypesByName(returnType, allowAny, allAllowedPromiseNames)) {
                    // Return type is not a promise
                    return;
                }
                if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition) {
                    // Abstract method can't be async
                    return;
                }
                if (node.parent &&
                    (node.parent.type === experimental_utils_1.AST_NODE_TYPES.Property ||
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition) &&
                    (node.parent.kind === 'get' || node.parent.kind === 'set')) {
                    // Getters and setters can't be async
                    return;
                }
                if (util.isTypeFlagSet(returnType, ts.TypeFlags.Any | ts.TypeFlags.Unknown)) {
                    // Report without auto fixer because the return type is unknown
                    return context.report({
                        messageId: 'missingAsync',
                        node,
                        loc: util.getFunctionHeadLoc(node, sourceCode),
                    });
                }
                context.report({
                    messageId: 'missingAsync',
                    node,
                    loc: util.getFunctionHeadLoc(node, sourceCode),
                    fix: fixer => {
                        if (node.parent &&
                            (node.parent.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition ||
                                (node.parent.type === experimental_utils_1.AST_NODE_TYPES.Property &&
                                    node.parent.method))) {
                            // this function is a class method or object function property shorthand
                            const method = node.parent;
                            // the token to put `async` before
                            let keyToken = sourceCode.getFirstToken(method);
                            // if there are decorators then skip past them
                            if (method.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition &&
                                method.decorators) {
                                const lastDecorator = method.decorators[method.decorators.length - 1];
                                keyToken = sourceCode.getTokenAfter(lastDecorator);
                            }
                            // if current token is a keyword like `static` or `public` then skip it
                            while (keyToken.type === experimental_utils_1.AST_TOKEN_TYPES.Keyword) {
                                keyToken = sourceCode.getTokenAfter(keyToken);
                            }
                            // check if there is a space between key and previous token
                            const insertSpace = !sourceCode.isSpaceBetween(sourceCode.getTokenBefore(keyToken), keyToken);
                            let code = 'async ';
                            if (insertSpace) {
                                code = ` ${code}`;
                            }
                            return fixer.insertTextBefore(keyToken, code);
                        }
                        return fixer.insertTextBefore(node, 'async ');
                    },
                });
            }
            return {
                'ArrowFunctionExpression[async = false]'(node) {
                    if (checkArrowFunctions) {
                        validateNode(node);
                    }
                },
                'FunctionDeclaration[async = false]'(node) {
                    if (checkFunctionDeclarations) {
                        validateNode(node);
                    }
                },
                'FunctionExpression[async = false]'(node) {
                    if (node.parent &&
                        node.parent.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition &&
                        node.parent.kind === 'method') {
                        if (checkMethodDeclarations) {
                            validateNode(node);
                        }
                        return;
                    }
                    if (checkFunctionExpressions) {
                        validateNode(node);
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/quotes.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const quotes_1 = __importDefault(require(659) /* eslint/lib/rules/quotes */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'quotes',
        meta: {
            type: 'layout',
            docs: {
                description: 'Enforce the consistent use of either backticks, double, or single quotes',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: 'code',
            messages: (_a = quotes_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                wrongQuotes: 'Strings must use {{description}}.',
            },
            schema: quotes_1.default.meta.schema,
        },
        defaultOptions: [
            'double',
            {
                allowTemplateLiterals: false,
                avoidEscape: false,
            },
        ],
        create(context, [option]) {
            const rules = quotes_1.default.create(context);
            function isAllowedAsNonBacktick(node) {
                const parent = node.parent;
                switch (parent === null || parent === void 0 ? void 0 : parent.type) {
                    case experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition:
                    case experimental_utils_1.AST_NODE_TYPES.TSMethodSignature:
                    case experimental_utils_1.AST_NODE_TYPES.TSPropertySignature:
                    case experimental_utils_1.AST_NODE_TYPES.TSModuleDeclaration:
                    case experimental_utils_1.AST_NODE_TYPES.TSLiteralType:
                    case experimental_utils_1.AST_NODE_TYPES.TSExternalModuleReference:
                        return true;
                    case experimental_utils_1.AST_NODE_TYPES.TSEnumMember:
                        return node === parent.id;
                    case experimental_utils_1.AST_NODE_TYPES.TSAbstractClassProperty:
                    case experimental_utils_1.AST_NODE_TYPES.ClassProperty:
                        return node === parent.key;
                    default:
                        return false;
                }
            }
            return {
                Literal(node) {
                    if (option === 'backtick' && isAllowedAsNonBacktick(node)) {
                        return;
                    }
                    rules.Literal(node);
                },
                TemplateLiteral(node) {
                    rules.TemplateLiteral(node);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/require-array-sort-compare.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'require-array-sort-compare',
        defaultOptions: [
            {
                ignoreStringArrays: false,
            },
        ],
        meta: {
            type: 'problem',
            docs: {
                description: 'Requires `Array#sort` calls to always provide a `compareFunction`',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            messages: {
                requireCompare: "Require 'compare' argument.",
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        ignoreStringArrays: {
                            type: 'boolean',
                        },
                    },
                },
            ],
        },
        create(context, [options]) {
            const service = util.getParserServices(context);
            const checker = service.program.getTypeChecker();
            /**
             * Check if a given node is an array which all elements are string.
             * @param node
             */
            function isStringArrayNode(node) {
                const type = checker.getTypeAtLocation(service.esTreeNodeToTSNodeMap.get(node));
                if (checker.isArrayType(type) || checker.isTupleType(type)) {
                    const typeArgs = checker.getTypeArguments(type);
                    return typeArgs.every(arg => util.getTypeName(checker, arg) === 'string');
                }
                return false;
            }
            return {
                "CallExpression[arguments.length=0] > MemberExpression[property.name='sort'][computed=false]"(callee) {
                    const tsNode = service.esTreeNodeToTSNodeMap.get(callee.object);
                    const calleeObjType = util.getConstrainedTypeAtLocation(checker, tsNode);
                    if (options.ignoreStringArrays && isStringArrayNode(callee.object)) {
                        return;
                    }
                    if (util.isTypeArrayTypeOrUnionOfArrayTypes(calleeObjType, checker)) {
                        context.report({ node: callee.parent, messageId: 'requireCompare' });
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/require-await.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'require-await',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Disallow async functions which have no `await` expression',
                category: 'Best Practices',
                recommended: 'error',
                requiresTypeChecking: true,
                extendsBaseRule: true,
            },
            schema: [],
            messages: {
                missingAwait: "{{name}} has no 'await' expression.",
            },
        },
        defaultOptions: [],
        create(context) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const sourceCode = context.getSourceCode();
            let scopeInfo = null;
            /**
             * Push the scope info object to the stack.
             */
            function enterFunction(node) {
                scopeInfo = {
                    upper: scopeInfo,
                    hasAwait: false,
                    hasAsync: node.async,
                    isGen: node.generator || false,
                    isAsyncYield: false,
                };
            }
            /**
             * Pop the top scope info object from the stack.
             * Also, it reports the function if needed.
             */
            function exitFunction(node) {
                /* istanbul ignore if */ if (!scopeInfo) {
                    // this shouldn't ever happen, as we have to exit a function after we enter it
                    return;
                }
                if (node.async &&
                    !scopeInfo.hasAwait &&
                    !isEmptyFunction(node) &&
                    !(scopeInfo.isGen && scopeInfo.isAsyncYield)) {
                    context.report({
                        node,
                        loc: getFunctionHeadLoc(node, sourceCode),
                        messageId: 'missingAwait',
                        data: {
                            name: util.upperCaseFirst(util.getFunctionNameWithKind(node)),
                        },
                    });
                }
                scopeInfo = scopeInfo.upper;
            }
            /**
             * Checks if the node returns a thenable type
             */
            function isThenableType(node) {
                const type = checker.getTypeAtLocation(node);
                return tsutils.isThenableType(checker, node, type);
            }
            /**
             * Marks the current scope as having an await
             */
            function markAsHasAwait() {
                if (!scopeInfo) {
                    return;
                }
                scopeInfo.hasAwait = true;
            }
            /**
             * mark `scopeInfo.isAsyncYield` to `true` if its a generator
             * function and the delegate is `true`
             */
            function markAsHasDelegateGen(node) {
                var _a;
                if (!scopeInfo || !scopeInfo.isGen || !node.argument) {
                    return;
                }
                if (((_a = node === null || node === void 0 ? void 0 : node.argument) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.Literal) {
                    // making this `false` as for literals we don't need to check the definition
                    // eg : async function* run() { yield* 1 }
                    scopeInfo.isAsyncYield = false;
                }
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node === null || node === void 0 ? void 0 : node.argument);
                const type = checker.getTypeAtLocation(tsNode);
                const symbol = type.getSymbol();
                // async function* test1() {yield* asyncGenerator() }
                if ((symbol === null || symbol === void 0 ? void 0 : symbol.getName()) === 'AsyncGenerator') {
                    scopeInfo.isAsyncYield = true;
                }
            }
            return {
                FunctionDeclaration: enterFunction,
                FunctionExpression: enterFunction,
                ArrowFunctionExpression: enterFunction,
                'FunctionDeclaration:exit': exitFunction,
                'FunctionExpression:exit': exitFunction,
                'ArrowFunctionExpression:exit': exitFunction,
                AwaitExpression: markAsHasAwait,
                'ForOfStatement[await = true]': markAsHasAwait,
                'YieldExpression[delegate = true]': markAsHasDelegateGen,
                // check body-less async arrow function.
                // ignore `async () => await foo` because it's obviously correct
                'ArrowFunctionExpression[async = true] > :not(BlockStatement, AwaitExpression)'(node) {
                    const expression = parserServices.esTreeNodeToTSNodeMap.get(node);
                    if (expression && isThenableType(expression)) {
                        markAsHasAwait();
                    }
                },
                ReturnStatement(node) {
                    // short circuit early to avoid unnecessary type checks
                    if (!scopeInfo || scopeInfo.hasAwait || !scopeInfo.hasAsync) {
                        return;
                    }
                    const { expression } = parserServices.esTreeNodeToTSNodeMap.get(node);
                    if (expression && isThenableType(expression)) {
                        markAsHasAwait();
                    }
                },
            };
        },
    });
    function isEmptyFunction(node) {
        var _a;
        return (((_a = node.body) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.BlockStatement &&
            node.body.body.length === 0);
    }
    // https://github.com/eslint/eslint/blob/03a69dbe86d5b5768a310105416ae726822e3c1c/lib/rules/utils/ast-utils.js#L382-L392
    /**
     * Gets the `(` token of the given function node.
     */
    function getOpeningParenOfParams(node, sourceCode) {
        return util.nullThrows(node.id
            ? sourceCode.getTokenAfter(node.id, util.isOpeningParenToken)
            : sourceCode.getFirstToken(node, util.isOpeningParenToken), util.NullThrowsReasons.MissingToken('(', node.type));
    }
    // https://github.com/eslint/eslint/blob/03a69dbe86d5b5768a310105416ae726822e3c1c/lib/rules/utils/ast-utils.js#L1220-L1242
    /**
     * Gets the location of the given function node for reporting.
     */
    function getFunctionHeadLoc(node, sourceCode) {
        const parent = util.nullThrows(node.parent, util.NullThrowsReasons.MissingParent);
        let start = null;
        let end = null;
        if (node.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression) {
            const arrowToken = util.nullThrows(sourceCode.getTokenBefore(node.body, util.isArrowToken), util.NullThrowsReasons.MissingToken('=>', node.type));
            start = arrowToken.loc.start;
            end = arrowToken.loc.end;
        }
        else if (parent.type === experimental_utils_1.AST_NODE_TYPES.Property ||
            parent.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition) {
            start = parent.loc.start;
            end = getOpeningParenOfParams(node, sourceCode).loc.start;
        }
        else {
            start = node.loc.start;
            end = getOpeningParenOfParams(node, sourceCode).loc.start;
        }
        return {
            start,
            end,
        };
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/restrict-plus-operands.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'restrict-plus-operands',
        meta: {
            type: 'problem',
            docs: {
                description: 'When adding two variables, operands must both be of type number or of type string',
                category: 'Best Practices',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                notNumbers: "Operands of '+' operation must either be both strings or both numbers.",
                notStrings: "Operands of '+' operation must either be both strings or both numbers. Consider using a template literal.",
                notBigInts: "Operands of '+' operation must be both bigints.",
            },
            schema: [
                {
                    type: 'object',
                    additionalProperties: false,
                    properties: {
                        checkCompoundAssignments: {
                            type: 'boolean',
                        },
                    },
                },
            ],
        },
        defaultOptions: [
            {
                checkCompoundAssignments: false,
            },
        ],
        create(context, [{ checkCompoundAssignments }]) {
            const service = util.getParserServices(context);
            const typeChecker = service.program.getTypeChecker();
            /**
             * Helper function to get base type of node
             */
            function getBaseTypeOfLiteralType(type) {
                if (type.isNumberLiteral()) {
                    return 'number';
                }
                if (type.isStringLiteral() ||
                    util.isTypeFlagSet(type, ts.TypeFlags.TemplateLiteral)) {
                    return 'string';
                }
                // is BigIntLiteral
                if (type.flags & ts.TypeFlags.BigIntLiteral) {
                    return 'bigint';
                }
                if (type.isUnion()) {
                    const types = type.types.map(getBaseTypeOfLiteralType);
                    return types.every(value => value === types[0]) ? types[0] : 'invalid';
                }
                if (type.isIntersection()) {
                    const types = type.types.map(getBaseTypeOfLiteralType);
                    return types.some(value => value === 'string') ? 'string' : 'invalid';
                }
                const stringType = typeChecker.typeToString(type);
                if (stringType === 'number' ||
                    stringType === 'string' ||
                    stringType === 'bigint') {
                    return stringType;
                }
                return 'invalid';
            }
            /**
             * Helper function to get base type of node
             * @param node the node to be evaluated.
             */
            function getNodeType(node) {
                const tsNode = service.esTreeNodeToTSNodeMap.get(node);
                const type = util.getConstrainedTypeAtLocation(typeChecker, tsNode);
                return getBaseTypeOfLiteralType(type);
            }
            function checkPlusOperands(node) {
                const leftType = getNodeType(node.left);
                const rightType = getNodeType(node.right);
                if (leftType === 'invalid' ||
                    rightType === 'invalid' ||
                    leftType !== rightType) {
                    if (leftType === 'string' || rightType === 'string') {
                        context.report({
                            node,
                            messageId: 'notStrings',
                        });
                    }
                    else if (leftType === 'bigint' || rightType === 'bigint') {
                        context.report({
                            node,
                            messageId: 'notBigInts',
                        });
                    }
                    else {
                        context.report({
                            node,
                            messageId: 'notNumbers',
                        });
                    }
                }
            }
            return {
                "BinaryExpression[operator='+']": checkPlusOperands,
                "AssignmentExpression[operator='+=']"(node) {
                    if (checkCompoundAssignments) {
                        checkPlusOperands(node);
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/restrict-template-expressions.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'restrict-template-expressions',
        meta: {
            type: 'problem',
            docs: {
                description: 'Enforce template literal expressions to be of string type',
                category: 'Best Practices',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                invalidType: 'Invalid type "{{type}}" of template literal expression.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowNumber: { type: 'boolean' },
                        allowBoolean: { type: 'boolean' },
                        allowAny: { type: 'boolean' },
                        allowNullish: { type: 'boolean' },
                    },
                },
            ],
        },
        defaultOptions: [
            {
                allowNumber: true,
            },
        ],
        create(context, [options]) {
            const service = util.getParserServices(context);
            const typeChecker = service.program.getTypeChecker();
            function isUnderlyingTypePrimitive(type) {
                if (util.isTypeFlagSet(type, ts.TypeFlags.StringLike)) {
                    return true;
                }
                if (options.allowNumber &&
                    util.isTypeFlagSet(type, ts.TypeFlags.NumberLike | ts.TypeFlags.BigIntLike)) {
                    return true;
                }
                if (options.allowBoolean &&
                    util.isTypeFlagSet(type, ts.TypeFlags.BooleanLike)) {
                    return true;
                }
                if (options.allowAny && util.isTypeAnyType(type)) {
                    return true;
                }
                if (options.allowNullish &&
                    util.isTypeFlagSet(type, ts.TypeFlags.Null | ts.TypeFlags.Undefined)) {
                    return true;
                }
                return false;
            }
            return {
                TemplateLiteral(node) {
                    // don't check tagged template literals
                    if (node.parent.type === experimental_utils_1.AST_NODE_TYPES.TaggedTemplateExpression) {
                        return;
                    }
                    for (const expression of node.expressions) {
                        const expressionType = util.getConstrainedTypeAtLocation(typeChecker, service.esTreeNodeToTSNodeMap.get(expression));
                        if (!isInnerUnionOrIntersectionConformingTo(expressionType, isUnderlyingTypePrimitive)) {
                            context.report({
                                node: expression,
                                messageId: 'invalidType',
                                data: { type: typeChecker.typeToString(expressionType) },
                            });
                        }
                    }
                },
            };
            function isInnerUnionOrIntersectionConformingTo(type, predicate) {
                return rec(type);
                function rec(innerType) {
                    if (innerType.isUnion()) {
                        return innerType.types.every(rec);
                    }
                    if (innerType.isIntersection()) {
                        return innerType.types.some(rec);
                    }
                    return predicate(innerType);
                }
            }
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/return-await.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'return-await',
        meta: {
            docs: {
                description: 'Enforces consistent returning of awaited values',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
                extendsBaseRule: 'no-return-await',
            },
            fixable: 'code',
            type: 'problem',
            messages: {
                nonPromiseAwait: 'Returning an awaited value that is not a promise is not allowed.',
                disallowedPromiseAwait: 'Returning an awaited promise is not allowed in this context.',
                requiredPromiseAwait: 'Returning an awaited promise is required in this context.',
            },
            schema: [
                {
                    enum: ['in-try-catch', 'always', 'never'],
                },
            ],
        },
        defaultOptions: ['in-try-catch'],
        create(context, [option]) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const sourceCode = context.getSourceCode();
            const scopeInfoStack = [];
            function enterFunction(node) {
                scopeInfoStack.push({
                    hasAsync: node.async,
                    owningFunc: node,
                });
            }
            function exitFunction() {
                scopeInfoStack.pop();
            }
            function inTry(node) {
                let ancestor = node.parent;
                while (ancestor && !ts.isFunctionLike(ancestor)) {
                    if (ts.isTryStatement(ancestor)) {
                        return true;
                    }
                    ancestor = ancestor.parent;
                }
                return false;
            }
            function inCatch(node) {
                let ancestor = node.parent;
                while (ancestor && !ts.isFunctionLike(ancestor)) {
                    if (ts.isCatchClause(ancestor)) {
                        return true;
                    }
                    ancestor = ancestor.parent;
                }
                return false;
            }
            function isReturnPromiseInFinally(node) {
                let ancestor = node.parent;
                while (ancestor && !ts.isFunctionLike(ancestor)) {
                    if (ts.isTryStatement(ancestor.parent) &&
                        ts.isBlock(ancestor) &&
                        ancestor.parent.end === ancestor.end) {
                        return true;
                    }
                    ancestor = ancestor.parent;
                }
                return false;
            }
            function hasFinallyBlock(node) {
                let ancestor = node.parent;
                while (ancestor && !ts.isFunctionLike(ancestor)) {
                    if (ts.isTryStatement(ancestor)) {
                        return !!ancestor.finallyBlock;
                    }
                    ancestor = ancestor.parent;
                }
                return false;
            }
            // function findTokensToRemove()
            function removeAwait(fixer, node) {
                // Should always be an await node; but let's be safe.
                /* istanbul ignore if */ if (!util.isAwaitExpression(node)) {
                    return null;
                }
                const awaitToken = sourceCode.getFirstToken(node, util.isAwaitKeyword);
                // Should always be the case; but let's be safe.
                /* istanbul ignore if */ if (!awaitToken) {
                    return null;
                }
                const startAt = awaitToken.range[0];
                let endAt = awaitToken.range[1];
                // Also remove any extraneous whitespace after `await`, if there is any.
                const nextToken = sourceCode.getTokenAfter(awaitToken, {
                    includeComments: true,
                });
                if (nextToken) {
                    endAt = nextToken.range[0];
                }
                return fixer.removeRange([startAt, endAt]);
            }
            function insertAwait(fixer, node) {
                if (node.type !== experimental_utils_1.AST_NODE_TYPES.TSAsExpression) {
                    return fixer.insertTextBefore(node, 'await ');
                }
                return [
                    fixer.insertTextBefore(node, 'await ('),
                    fixer.insertTextAfter(node, ')'),
                ];
            }
            function test(node, expression) {
                let child;
                const isAwait = ts.isAwaitExpression(expression);
                if (isAwait) {
                    child = expression.getChildAt(1);
                }
                else {
                    child = expression;
                }
                const type = checker.getTypeAtLocation(child);
                const isThenable = tsutils.isThenableType(checker, expression, type);
                if (!isAwait && !isThenable) {
                    return;
                }
                if (isAwait && !isThenable) {
                    // any/unknown could be thenable; do not auto-fix
                    const useAutoFix = !(util.isTypeAnyType(type) || util.isTypeUnknownType(type));
                    const fix = (fixer) => removeAwait(fixer, node);
                    context.report(Object.assign({ messageId: 'nonPromiseAwait', node }, (useAutoFix
                        ? { fix }
                        : {
                            suggest: [
                                {
                                    messageId: 'nonPromiseAwait',
                                    fix,
                                },
                            ],
                        })));
                    return;
                }
                if (option === 'always') {
                    if (!isAwait && isThenable) {
                        context.report({
                            messageId: 'requiredPromiseAwait',
                            node,
                            fix: fixer => insertAwait(fixer, node),
                        });
                    }
                    return;
                }
                if (option === 'never') {
                    if (isAwait) {
                        context.report({
                            messageId: 'disallowedPromiseAwait',
                            node,
                            fix: fixer => removeAwait(fixer, node),
                        });
                    }
                    return;
                }
                if (option === 'in-try-catch') {
                    const isInTryCatch = inTry(expression) || inCatch(expression);
                    if (isAwait && !isInTryCatch) {
                        context.report({
                            messageId: 'disallowedPromiseAwait',
                            node,
                            fix: fixer => removeAwait(fixer, node),
                        });
                    }
                    else if (!isAwait && isInTryCatch) {
                        if (inCatch(expression) && !hasFinallyBlock(expression)) {
                            return;
                        }
                        if (isReturnPromiseInFinally(expression)) {
                            return;
                        }
                        context.report({
                            messageId: 'requiredPromiseAwait',
                            node,
                            fix: fixer => insertAwait(fixer, node),
                        });
                    }
                    return;
                }
            }
            function findPossiblyReturnedNodes(node) {
                if (node.type === experimental_utils_1.AST_NODE_TYPES.ConditionalExpression) {
                    return [
                        ...findPossiblyReturnedNodes(node.alternate),
                        ...findPossiblyReturnedNodes(node.consequent),
                    ];
                }
                return [node];
            }
            return {
                FunctionDeclaration: enterFunction,
                FunctionExpression: enterFunction,
                ArrowFunctionExpression: enterFunction,
                'FunctionDeclaration:exit': exitFunction,
                'FunctionExpression:exit': exitFunction,
                'ArrowFunctionExpression:exit': exitFunction,
                // executes after less specific handler, so exitFunction is called
                'ArrowFunctionExpression[async = true]:exit'(node) {
                    if (node.body.type !== experimental_utils_1.AST_NODE_TYPES.BlockStatement) {
                        findPossiblyReturnedNodes(node.body).forEach(node => {
                            const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                            test(node, tsNode);
                        });
                    }
                },
                ReturnStatement(node) {
                    const scopeInfo = scopeInfoStack[scopeInfoStack.length - 1];
                    if (!scopeInfo || !scopeInfo.hasAsync || !node.argument) {
                        return;
                    }
                    findPossiblyReturnedNodes(node.argument).forEach(node => {
                        const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                        test(node, tsNode);
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/semi.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const semi_1 = __importDefault(require(667) /* eslint/lib/rules/semi */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'semi',
        meta: {
            type: 'layout',
            docs: {
                description: 'Require or disallow semicolons instead of ASI',
                category: 'Stylistic Issues',
                // too opinionated to be recommended
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: 'code',
            schema: semi_1.default.meta.schema,
            messages: (_a = semi_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                missingSemi: 'Missing semicolon.',
                extraSemi: 'Extra semicolon.',
            },
        },
        defaultOptions: [
            'always',
            {
                omitLastInOneLineBlock: false,
                beforeStatementContinuationChars: 'any',
            },
        ],
        create(context) {
            const rules = semi_1.default.create(context);
            const checkForSemicolon = rules.ExpressionStatement;
            /*
              The following nodes are handled by the member-delimiter-style rule
              AST_NODE_TYPES.TSCallSignatureDeclaration,
              AST_NODE_TYPES.TSConstructSignatureDeclaration,
              AST_NODE_TYPES.TSIndexSignature,
              AST_NODE_TYPES.TSMethodSignature,
              AST_NODE_TYPES.TSPropertySignature,
            */
            const nodesToCheck = [
                experimental_utils_1.AST_NODE_TYPES.ClassProperty,
                experimental_utils_1.AST_NODE_TYPES.TSAbstractClassProperty,
                experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition,
                experimental_utils_1.AST_NODE_TYPES.TSDeclareFunction,
                experimental_utils_1.AST_NODE_TYPES.TSExportAssignment,
                experimental_utils_1.AST_NODE_TYPES.TSImportEqualsDeclaration,
                experimental_utils_1.AST_NODE_TYPES.TSTypeAliasDeclaration,
            ].reduce((acc, node) => {
                acc[node] = checkForSemicolon;
                return acc;
            }, {});
            return Object.assign(Object.assign(Object.assign({}, rules), nodesToCheck), { ExportDefaultDeclaration(node) {
                    if (node.declaration.type !== experimental_utils_1.AST_NODE_TYPES.TSInterfaceDeclaration) {
                        rules.ExportDefaultDeclaration(node);
                    }
                } });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/sort-type-union-intersection-members.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    const util_1 = require(906) /* ../util */;
    var Group;
    (function (Group) {
        Group["conditional"] = "conditional";
        Group["function"] = "function";
        Group["import"] = "import";
        Group["intersection"] = "intersection";
        Group["keyword"] = "keyword";
        Group["nullish"] = "nullish";
        Group["literal"] = "literal";
        Group["named"] = "named";
        Group["object"] = "object";
        Group["operator"] = "operator";
        Group["tuple"] = "tuple";
        Group["union"] = "union";
    })(Group || (Group = {}));
    function getGroup(node) {
        switch (node.type) {
            case experimental_utils_1.AST_NODE_TYPES.TSParenthesizedType:
                return getGroup(node.typeAnnotation);
            case experimental_utils_1.AST_NODE_TYPES.TSConditionalType:
                return Group.conditional;
            case experimental_utils_1.AST_NODE_TYPES.TSConstructorType:
            case experimental_utils_1.AST_NODE_TYPES.TSFunctionType:
                return Group.function;
            case experimental_utils_1.AST_NODE_TYPES.TSImportType:
                return Group.import;
            case experimental_utils_1.AST_NODE_TYPES.TSIntersectionType:
                return Group.intersection;
            case experimental_utils_1.AST_NODE_TYPES.TSAnyKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSBigIntKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSBooleanKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSNeverKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSNumberKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSObjectKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSStringKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSSymbolKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSThisType:
            case experimental_utils_1.AST_NODE_TYPES.TSUnknownKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSIntrinsicKeyword:
                return Group.keyword;
            case experimental_utils_1.AST_NODE_TYPES.TSNullKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSUndefinedKeyword:
            case experimental_utils_1.AST_NODE_TYPES.TSVoidKeyword:
                return Group.nullish;
            case experimental_utils_1.AST_NODE_TYPES.TSLiteralType:
            case experimental_utils_1.AST_NODE_TYPES.TSTemplateLiteralType:
                return Group.literal;
            case experimental_utils_1.AST_NODE_TYPES.TSArrayType:
            case experimental_utils_1.AST_NODE_TYPES.TSIndexedAccessType:
            case experimental_utils_1.AST_NODE_TYPES.TSInferType:
            case experimental_utils_1.AST_NODE_TYPES.TSTypeReference:
                return Group.named;
            case experimental_utils_1.AST_NODE_TYPES.TSMappedType:
            case experimental_utils_1.AST_NODE_TYPES.TSTypeLiteral:
                return Group.object;
            case experimental_utils_1.AST_NODE_TYPES.TSTypeOperator:
            case experimental_utils_1.AST_NODE_TYPES.TSTypeQuery:
                return Group.operator;
            case experimental_utils_1.AST_NODE_TYPES.TSTupleType:
                return Group.tuple;
            case experimental_utils_1.AST_NODE_TYPES.TSUnionType:
                return Group.union;
            // These types should never occur as part of a union/intersection
            case experimental_utils_1.AST_NODE_TYPES.TSNamedTupleMember:
            case experimental_utils_1.AST_NODE_TYPES.TSOptionalType:
            case experimental_utils_1.AST_NODE_TYPES.TSRestType:
            case experimental_utils_1.AST_NODE_TYPES.TSTypePredicate:
                /* istanbul ignore next */
                throw new Error(`Unexpected Type ${node.type}`);
        }
    }
    exports.default = util.createRule({
        name: 'sort-type-union-intersection-members',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Enforces that members of a type union/intersection are sorted alphabetically',
                category: 'Stylistic Issues',
                recommended: false,
            },
            fixable: 'code',
            messages: {
                notSorted: '{{type}} type members must be sorted.',
                notSortedNamed: '{{type}} type {{name}} members must be sorted.',
                suggestFix: 'Sort members of type (removes all comments).',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        checkIntersections: {
                            type: 'boolean',
                        },
                        checkUnions: {
                            type: 'boolean',
                        },
                        groupOrder: {
                            type: 'array',
                            items: {
                                type: 'string',
                                enum: util_1.getEnumNames(Group),
                            },
                        },
                    },
                },
            ],
        },
        defaultOptions: [
            {
                checkIntersections: true,
                checkUnions: true,
                groupOrder: [
                    Group.named,
                    Group.keyword,
                    Group.operator,
                    Group.literal,
                    Group.function,
                    Group.import,
                    Group.conditional,
                    Group.object,
                    Group.tuple,
                    Group.intersection,
                    Group.union,
                    Group.nullish,
                ],
            },
        ],
        create(context, [{ checkIntersections, checkUnions, groupOrder }]) {
            const sourceCode = context.getSourceCode();
            const collator = new Intl.Collator('en', {
                sensitivity: 'base',
                numeric: true,
            });
            function checkSorting(node) {
                var _a;
                const sourceOrder = node.types.map(type => {
                    var _a;
                    const group = (_a = groupOrder === null || groupOrder === void 0 ? void 0 : groupOrder.indexOf(getGroup(type))) !== null && _a !== void 0 ? _a : -1;
                    return {
                        group: group === -1 ? Number.MAX_SAFE_INTEGER : group,
                        node: type,
                        text: sourceCode.getText(type),
                    };
                });
                const expectedOrder = [...sourceOrder].sort((a, b) => {
                    if (a.group !== b.group) {
                        return a.group - b.group;
                    }
                    return (collator.compare(a.text, b.text) ||
                        (a.text < b.text ? -1 : a.text > b.text ? 1 : 0));
                });
                const hasComments = node.types.some(type => {
                    const count = sourceCode.getCommentsBefore(type).length +
                        sourceCode.getCommentsAfter(type).length;
                    return count > 0;
                });
                for (let i = 0; i < expectedOrder.length; i += 1) {
                    if (expectedOrder[i].node !== sourceOrder[i].node) {
                        let messageId = 'notSorted';
                        const data = {
                            name: '',
                            type: node.type === experimental_utils_1.AST_NODE_TYPES.TSIntersectionType
                                ? 'Intersection'
                                : 'Union',
                        };
                        if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.TSTypeAliasDeclaration) {
                            messageId = 'notSortedNamed';
                            data.name = node.parent.id.name;
                        }
                        const fix = fixer => {
                            const sorted = expectedOrder
                                .map(t => t.text)
                                .join(node.type === experimental_utils_1.AST_NODE_TYPES.TSIntersectionType ? ' & ' : ' | ');
                            return fixer.replaceText(node, sorted);
                        };
                        return context.report(Object.assign({ node,
                            messageId,
                            data }, (hasComments
                            ? {
                                suggest: [
                                    {
                                        messageId: 'suggestFix',
                                        fix,
                                    },
                                ],
                            }
                            : { fix })));
                    }
                }
            }
            return {
                TSIntersectionType(node) {
                    if (checkIntersections === true) {
                        checkSorting(node);
                    }
                },
                TSUnionType(node) {
                    if (checkUnions === true) {
                        checkSorting(node);
                    }
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/space-before-function-paren.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'space-before-function-paren',
        meta: {
            type: 'layout',
            docs: {
                description: 'Enforces consistent spacing before function parenthesis',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: 'whitespace',
            schema: [
                {
                    oneOf: [
                        {
                            enum: ['always', 'never'],
                        },
                        {
                            type: 'object',
                            properties: {
                                anonymous: {
                                    enum: ['always', 'never', 'ignore'],
                                },
                                named: {
                                    enum: ['always', 'never', 'ignore'],
                                },
                                asyncArrow: {
                                    enum: ['always', 'never', 'ignore'],
                                },
                            },
                            additionalProperties: false,
                        },
                    ],
                },
            ],
            messages: {
                unexpected: 'Unexpected space before function parentheses.',
                missing: 'Missing space before function parentheses.',
            },
        },
        defaultOptions: ['always'],
        create(context) {
            const sourceCode = context.getSourceCode();
            const baseConfig = typeof context.options[0] === 'string' ? context.options[0] : 'always';
            const overrideConfig = typeof context.options[0] === 'object' ? context.options[0] : {};
            /**
             * Determines whether a function has a name.
             * @param {ASTNode} node The function node.
             * @returns {boolean} Whether the function has a name.
             */
            function isNamedFunction(node) {
                if (node.id != null) {
                    return true;
                }
                const parent = node.parent;
                return (parent.type === experimental_utils_1.AST_NODE_TYPES.MethodDefinition ||
                    parent.type === experimental_utils_1.AST_NODE_TYPES.TSAbstractMethodDefinition ||
                    (parent.type === experimental_utils_1.AST_NODE_TYPES.Property &&
                        (parent.kind === 'get' || parent.kind === 'set' || parent.method)));
            }
            /**
             * Gets the config for a given function
             * @param {ASTNode} node The function node
             * @returns {string} "always", "never", or "ignore"
             */
            function getConfigForFunction(node) {
                var _a, _b, _c;
                if (node.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression) {
                    // Always ignore non-async functions and arrow functions without parens, e.g. async foo => bar
                    if (node.async &&
                        util.isOpeningParenToken(sourceCode.getFirstToken(node, { skip: 1 }))) {
                        return (_a = overrideConfig.asyncArrow) !== null && _a !== void 0 ? _a : baseConfig;
                    }
                }
                else if (isNamedFunction(node)) {
                    return (_b = overrideConfig.named) !== null && _b !== void 0 ? _b : baseConfig;
                    // `generator-star-spacing` should warn anonymous generators. E.g. `function* () {}`
                }
                else if (!node.generator) {
                    return (_c = overrideConfig.anonymous) !== null && _c !== void 0 ? _c : baseConfig;
                }
                return 'ignore';
            }
            /**
             * Checks the parens of a function node
             * @param {ASTNode} node A function node
             * @returns {void}
             */
            function checkFunction(node) {
                const functionConfig = getConfigForFunction(node);
                if (functionConfig === 'ignore') {
                    return;
                }
                let leftToken, rightToken;
                if (node.typeParameters) {
                    leftToken = sourceCode.getLastToken(node.typeParameters);
                    rightToken = sourceCode.getTokenAfter(leftToken);
                }
                else {
                    rightToken = sourceCode.getFirstToken(node, util.isOpeningParenToken);
                    leftToken = sourceCode.getTokenBefore(rightToken);
                }
                const hasSpacing = sourceCode.isSpaceBetweenTokens(leftToken, rightToken);
                if (hasSpacing && functionConfig === 'never') {
                    context.report({
                        node,
                        loc: {
                            start: leftToken.loc.end,
                            end: rightToken.loc.start,
                        },
                        messageId: 'unexpected',
                        fix: fixer => fixer.removeRange([leftToken.range[1], rightToken.range[0]]),
                    });
                }
                else if (!hasSpacing &&
                    functionConfig === 'always' &&
                    (!node.typeParameters || node.id)) {
                    context.report({
                        node,
                        loc: rightToken.loc,
                        messageId: 'missing',
                        fix: fixer => fixer.insertTextAfter(leftToken, ' '),
                    });
                }
            }
            return {
                ArrowFunctionExpression: checkFunction,
                FunctionDeclaration: checkFunction,
                FunctionExpression: checkFunction,
                TSEmptyBodyFunctionExpression: checkFunction,
                TSDeclareFunction: checkFunction,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/space-infix-ops.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    var _a;
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const space_infix_ops_1 = __importDefault(require(677) /* eslint/lib/rules/space-infix-ops */);
    const util = __importStar(require(906) /* ../util */);
    const UNIONS = ['|', '&'];
    exports.default = util.createRule({
        name: 'space-infix-ops',
        meta: {
            type: 'layout',
            docs: {
                description: 'This rule is aimed at ensuring there are spaces around infix operators.',
                category: 'Stylistic Issues',
                recommended: false,
                extendsBaseRule: true,
            },
            fixable: space_infix_ops_1.default.meta.fixable,
            schema: space_infix_ops_1.default.meta.schema,
            messages: (_a = space_infix_ops_1.default.meta.messages) !== null && _a !== void 0 ? _a : {
                missingSpace: "Operator '{{operator}}' must be spaced.",
            },
        },
        defaultOptions: [
            {
                int32Hint: false,
            },
        ],
        create(context) {
            const rules = space_infix_ops_1.default.create(context);
            const sourceCode = context.getSourceCode();
            const report = (node, operator) => {
                context.report({
                    node: node,
                    loc: operator.loc,
                    messageId: 'missingSpace',
                    data: {
                        operator: operator.value,
                    },
                    fix(fixer) {
                        const previousToken = sourceCode.getTokenBefore(operator);
                        const afterToken = sourceCode.getTokenAfter(operator);
                        let fixString = '';
                        if (operator.range[0] - previousToken.range[1] === 0) {
                            fixString = ' ';
                        }
                        fixString += operator.value;
                        if (afterToken.range[0] - operator.range[1] === 0) {
                            fixString += ' ';
                        }
                        return fixer.replaceText(operator, fixString);
                    },
                });
            };
            function isSpaceChar(token) {
                return token.type === experimental_utils_1.AST_TOKEN_TYPES.Punctuator && token.value === '=';
            }
            function checkAndReportAssignmentSpace(node, leftNode, rightNode) {
                if (!rightNode) {
                    return;
                }
                const operator = sourceCode.getFirstTokenBetween(leftNode, rightNode, isSpaceChar);
                const prev = sourceCode.getTokenBefore(operator);
                const next = sourceCode.getTokenAfter(operator);
                if (!sourceCode.isSpaceBetween(prev, operator) ||
                    !sourceCode.isSpaceBetween(operator, next)) {
                    report(node, operator);
                }
            }
            /**
             * Check if it has an assignment char and report if it's faulty
             * @param node The node to report
             */
            function checkForEnumAssignmentSpace(node) {
                if (!node.initializer) {
                    return;
                }
                const leftNode = sourceCode.getTokenByRangeStart(node.id.range[0]);
                const rightNode = sourceCode.getTokenByRangeStart(node.initializer.range[0]);
                checkAndReportAssignmentSpace(node, leftNode, rightNode);
            }
            /**
             * Check if it has an assignment char and report if it's faulty
             * @param node The node to report
             */
            function checkForClassPropertyAssignmentSpace(node) {
                var _a, _b;
                const leftNode = sourceCode.getTokenByRangeStart((_b = (_a = node.typeAnnotation) === null || _a === void 0 ? void 0 : _a.range[0]) !== null && _b !== void 0 ? _b : node.range[0]);
                const rightNode = node.value
                    ? sourceCode.getTokenByRangeStart(node.value.range[0])
                    : undefined;
                checkAndReportAssignmentSpace(node, leftNode, rightNode);
            }
            /**
             * Check if it is missing spaces between type annotations chaining
             * @param typeAnnotation TypeAnnotations list
             */
            function checkForTypeAnnotationSpace(typeAnnotation) {
                const types = typeAnnotation.types;
                types.forEach(type => {
                    const operator = sourceCode.getTokenBefore(type);
                    if (operator != null && UNIONS.includes(operator.value)) {
                        const prev = sourceCode.getTokenBefore(operator);
                        const next = sourceCode.getTokenAfter(operator);
                        if (!sourceCode.isSpaceBetween(prev, operator) ||
                            !sourceCode.isSpaceBetween(operator, next)) {
                            report(typeAnnotation, operator);
                        }
                    }
                });
            }
            /**
             * Check if it has an assignment char and report if it's faulty
             * @param node The node to report
             */
            function checkForTypeAliasAssignment(node) {
                const leftNode = sourceCode.getTokenByRangeStart(node.id.range[0]);
                const rightNode = sourceCode.getTokenByRangeStart(node.typeAnnotation.range[0]);
                checkAndReportAssignmentSpace(node, leftNode, rightNode);
            }
            return Object.assign(Object.assign({}, rules), { TSEnumMember: checkForEnumAssignmentSpace, ClassProperty: checkForClassPropertyAssignmentSpace, TSTypeAliasDeclaration: checkForTypeAliasAssignment, TSUnionType: checkForTypeAnnotationSpace, TSIntersectionType: checkForTypeAnnotationSpace });
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/strict-boolean-expressions.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'strict-boolean-expressions',
        meta: {
            type: 'suggestion',
            fixable: 'code',
            docs: {
                description: 'Restricts the types allowed in boolean expressions',
                category: 'Best Practices',
                recommended: false,
                requiresTypeChecking: true,
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        allowString: { type: 'boolean' },
                        allowNumber: { type: 'boolean' },
                        allowNullableObject: { type: 'boolean' },
                        allowNullableBoolean: { type: 'boolean' },
                        allowNullableString: { type: 'boolean' },
                        allowNullableNumber: { type: 'boolean' },
                        allowAny: { type: 'boolean' },
                        allowRuleToRunWithoutStrictNullChecksIKnowWhatIAmDoing: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
            messages: {
                conditionErrorOther: 'Unexpected value in conditional. ' +
                    'A boolean expression is required.',
                conditionErrorAny: 'Unexpected any value in conditional. ' +
                    'An explicit comparison or type cast is required.',
                conditionErrorNullish: 'Unexpected nullish value in conditional. ' +
                    'The condition is always false.',
                conditionErrorNullableBoolean: 'Unexpected nullable boolean value in conditional. ' +
                    'Please handle the nullish case explicitly.',
                conditionErrorString: 'Unexpected string value in conditional. ' +
                    'An explicit empty string check is required.',
                conditionErrorNullableString: 'Unexpected nullable string value in conditional. ' +
                    'Please handle the nullish/empty cases explicitly.',
                conditionErrorNumber: 'Unexpected number value in conditional. ' +
                    'An explicit zero/NaN check is required.',
                conditionErrorNullableNumber: 'Unexpected nullable number value in conditional. ' +
                    'Please handle the nullish/zero/NaN cases explicitly.',
                conditionErrorObject: 'Unexpected object value in conditional. ' +
                    'The condition is always true.',
                conditionErrorNullableObject: 'Unexpected nullable object value in conditional. ' +
                    'An explicit null check is required.',
                noStrictNullCheck: 'This rule requires the `strictNullChecks` compiler option to be turned on to function correctly.',
                conditionFixDefaultFalse: 'Explicitly treat nullish value the same as false (`value ?? false`)',
                conditionFixDefaultEmptyString: 'Explicitly treat nullish value the same as an empty string (`value ?? ""`)',
                conditionFixDefaultZero: 'Explicitly treat nullish value the same as 0 (`value ?? 0`)',
                conditionFixCompareNullish: 'Change condition to check for null/undefined (`value != null`)',
                conditionFixCastBoolean: 'Explicitly cast value to a boolean (`Boolean(value)`)',
                conditionFixCompareTrue: 'Change condition to check if true (`value === true`)',
                conditionFixCompareFalse: 'Change condition to check if false (`value === false`)',
                conditionFixCompareStringLength: "Change condition to check string's length (`value.length !== 0`)",
                conditionFixCompareEmptyString: 'Change condition to check for empty string (`value !== ""`)',
                conditionFixCompareZero: 'Change condition to check for 0 (`value !== 0`)',
                conditionFixCompareNaN: 'Change condition to check for NaN (`!Number.isNaN(value)`)',
            },
        },
        defaultOptions: [
            {
                allowString: true,
                allowNumber: true,
                allowNullableObject: true,
                allowNullableBoolean: false,
                allowNullableString: false,
                allowNullableNumber: false,
                allowAny: false,
                allowRuleToRunWithoutStrictNullChecksIKnowWhatIAmDoing: false,
            },
        ],
        create(context, [options]) {
            const parserServices = util.getParserServices(context);
            const typeChecker = parserServices.program.getTypeChecker();
            const compilerOptions = parserServices.program.getCompilerOptions();
            const sourceCode = context.getSourceCode();
            const isStrictNullChecks = tsutils.isStrictCompilerOptionEnabled(compilerOptions, 'strictNullChecks');
            if (!isStrictNullChecks &&
                options.allowRuleToRunWithoutStrictNullChecksIKnowWhatIAmDoing !== true) {
                context.report({
                    loc: {
                        start: { line: 0, column: 0 },
                        end: { line: 0, column: 0 },
                    },
                    messageId: 'noStrictNullCheck',
                });
            }
            const checkedNodes = new Set();
            return {
                ConditionalExpression: checkTestExpression,
                DoWhileStatement: checkTestExpression,
                ForStatement: checkTestExpression,
                IfStatement: checkTestExpression,
                WhileStatement: checkTestExpression,
                'LogicalExpression[operator!="??"]': checkNode,
                'UnaryExpression[operator="!"]': checkUnaryLogicalExpression,
            };
            function checkTestExpression(node) {
                if (node.test == null) {
                    return;
                }
                checkNode(node.test, true);
            }
            function checkUnaryLogicalExpression(node) {
                checkNode(node.argument, true);
            }
            /**
             * This function analyzes the type of a node and checks if it is allowed in a boolean context.
             * It can recurse when checking nested logical operators, so that only the outermost operands are reported.
             * The right operand of a logical expression is ignored unless it's a part of a test expression (if/while/ternary/etc).
             * @param node The AST node to check.
             * @param isTestExpr Whether the node is a descendant of a test expression.
             */
            function checkNode(node, isTestExpr = false) {
                // prevent checking the same node multiple times
                if (checkedNodes.has(node)) {
                    return;
                }
                checkedNodes.add(node);
                // for logical operator, we check its operands
                if (node.type === experimental_utils_1.AST_NODE_TYPES.LogicalExpression &&
                    node.operator !== '??') {
                    checkNode(node.left, isTestExpr);
                    // we ignore the right operand when not in a context of a test expression
                    if (isTestExpr) {
                        checkNode(node.right, isTestExpr);
                    }
                    return;
                }
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                const type = util.getConstrainedTypeAtLocation(typeChecker, tsNode);
                const types = inspectVariantTypes(tsutils.unionTypeParts(type));
                const is = (...wantedTypes) => types.size === wantedTypes.length &&
                    wantedTypes.every(type => types.has(type));
                // boolean
                if (is('boolean')) {
                    // boolean is always okay
                    return;
                }
                // never
                if (is('never')) {
                    // never is always okay
                    return;
                }
                // nullish
                if (is('nullish')) {
                    // condition is always false
                    context.report({ node, messageId: 'conditionErrorNullish' });
                    return;
                }
                // nullable boolean
                if (is('nullish', 'boolean')) {
                    if (!options.allowNullableBoolean) {
                        if (isLogicalNegationExpression(node.parent)) {
                            // if (!nullableBoolean)
                            context.report({
                                node,
                                messageId: 'conditionErrorNullableBoolean',
                                suggest: [
                                    {
                                        messageId: 'conditionFixDefaultFalse',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} ?? false`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCompareFalse',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `${code} === false`,
                                        }),
                                    },
                                ],
                            });
                        }
                        else {
                            // if (nullableBoolean)
                            context.report({
                                node,
                                messageId: 'conditionErrorNullableBoolean',
                                suggest: [
                                    {
                                        messageId: 'conditionFixDefaultFalse',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} ?? false`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCompareTrue',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} === true`,
                                        }),
                                    },
                                ],
                            });
                        }
                    }
                    return;
                }
                // Known edge case: truthy primitives and nullish values are always valid boolean expressions
                if ((options.allowNumber && is('nullish', 'truthy number')) ||
                    (options.allowString && is('nullish', 'truthy string'))) {
                    return;
                }
                // string
                if (is('string') || is('truthy string')) {
                    if (!options.allowString) {
                        if (isLogicalNegationExpression(node.parent)) {
                            // if (!string)
                            context.report({
                                node,
                                messageId: 'conditionErrorString',
                                suggest: [
                                    {
                                        messageId: 'conditionFixCompareStringLength',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `${code}.length === 0`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCompareEmptyString',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `${code} === ""`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCastBoolean',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `!Boolean(${code})`,
                                        }),
                                    },
                                ],
                            });
                        }
                        else {
                            // if (string)
                            context.report({
                                node,
                                messageId: 'conditionErrorString',
                                suggest: [
                                    {
                                        messageId: 'conditionFixCompareStringLength',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code}.length > 0`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCompareEmptyString',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} !== ""`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCastBoolean',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `Boolean(${code})`,
                                        }),
                                    },
                                ],
                            });
                        }
                    }
                    return;
                }
                // nullable string
                if (is('nullish', 'string')) {
                    if (!options.allowNullableString) {
                        if (isLogicalNegationExpression(node.parent)) {
                            // if (!nullableString)
                            context.report({
                                node,
                                messageId: 'conditionErrorNullableString',
                                suggest: [
                                    {
                                        messageId: 'conditionFixCompareNullish',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `${code} == null`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixDefaultEmptyString',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} ?? ""`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCastBoolean',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `!Boolean(${code})`,
                                        }),
                                    },
                                ],
                            });
                        }
                        else {
                            // if (nullableString)
                            context.report({
                                node,
                                messageId: 'conditionErrorNullableString',
                                suggest: [
                                    {
                                        messageId: 'conditionFixCompareNullish',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} != null`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixDefaultEmptyString',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} ?? ""`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCastBoolean',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `Boolean(${code})`,
                                        }),
                                    },
                                ],
                            });
                        }
                    }
                    return;
                }
                // number
                if (is('number') || is('truthy number')) {
                    if (!options.allowNumber) {
                        if (isArrayLengthExpression(node, typeChecker, parserServices)) {
                            if (isLogicalNegationExpression(node.parent)) {
                                // if (!array.length)
                                context.report({
                                    node,
                                    messageId: 'conditionErrorNumber',
                                    fix: util.getWrappingFixer({
                                        sourceCode,
                                        node: node.parent,
                                        innerNode: node,
                                        wrap: code => `${code} === 0`,
                                    }),
                                });
                            }
                            else {
                                // if (array.length)
                                context.report({
                                    node,
                                    messageId: 'conditionErrorNumber',
                                    fix: util.getWrappingFixer({
                                        sourceCode,
                                        node,
                                        wrap: code => `${code} > 0`,
                                    }),
                                });
                            }
                        }
                        else if (isLogicalNegationExpression(node.parent)) {
                            // if (!number)
                            context.report({
                                node,
                                messageId: 'conditionErrorNumber',
                                suggest: [
                                    {
                                        messageId: 'conditionFixCompareZero',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            // TODO: we have to compare to 0n if the type is bigint
                                            wrap: code => `${code} === 0`,
                                        }),
                                    },
                                    {
                                        // TODO: don't suggest this for bigint because it can't be NaN
                                        messageId: 'conditionFixCompareNaN',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `Number.isNaN(${code})`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCastBoolean',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `!Boolean(${code})`,
                                        }),
                                    },
                                ],
                            });
                        }
                        else {
                            // if (number)
                            context.report({
                                node,
                                messageId: 'conditionErrorNumber',
                                suggest: [
                                    {
                                        messageId: 'conditionFixCompareZero',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} !== 0`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCompareNaN',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `!Number.isNaN(${code})`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCastBoolean',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `Boolean(${code})`,
                                        }),
                                    },
                                ],
                            });
                        }
                    }
                    return;
                }
                // nullable number
                if (is('nullish', 'number')) {
                    if (!options.allowNullableNumber) {
                        if (isLogicalNegationExpression(node.parent)) {
                            // if (!nullableNumber)
                            context.report({
                                node,
                                messageId: 'conditionErrorNullableNumber',
                                suggest: [
                                    {
                                        messageId: 'conditionFixCompareNullish',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `${code} == null`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixDefaultZero',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} ?? 0`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCastBoolean',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node: node.parent,
                                            innerNode: node,
                                            wrap: code => `!Boolean(${code})`,
                                        }),
                                    },
                                ],
                            });
                        }
                        else {
                            // if (nullableNumber)
                            context.report({
                                node,
                                messageId: 'conditionErrorNullableNumber',
                                suggest: [
                                    {
                                        messageId: 'conditionFixCompareNullish',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} != null`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixDefaultZero',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `${code} ?? 0`,
                                        }),
                                    },
                                    {
                                        messageId: 'conditionFixCastBoolean',
                                        fix: util.getWrappingFixer({
                                            sourceCode,
                                            node,
                                            wrap: code => `Boolean(${code})`,
                                        }),
                                    },
                                ],
                            });
                        }
                    }
                    return;
                }
                // object
                if (is('object')) {
                    // condition is always true
                    context.report({ node, messageId: 'conditionErrorObject' });
                    return;
                }
                // nullable object
                if (is('nullish', 'object')) {
                    if (!options.allowNullableObject) {
                        if (isLogicalNegationExpression(node.parent)) {
                            // if (!nullableObject)
                            context.report({
                                node,
                                messageId: 'conditionErrorNullableObject',
                                fix: util.getWrappingFixer({
                                    sourceCode,
                                    node: node.parent,
                                    innerNode: node,
                                    wrap: code => `${code} == null`,
                                }),
                            });
                        }
                        else {
                            // if (nullableObject)
                            context.report({
                                node,
                                messageId: 'conditionErrorNullableObject',
                                fix: util.getWrappingFixer({
                                    sourceCode,
                                    node,
                                    wrap: code => `${code} != null`,
                                }),
                            });
                        }
                    }
                    return;
                }
                // any
                if (is('any')) {
                    if (!options.allowAny) {
                        context.report({
                            node,
                            messageId: 'conditionErrorAny',
                            suggest: [
                                {
                                    messageId: 'conditionFixCastBoolean',
                                    fix: util.getWrappingFixer({
                                        sourceCode,
                                        node,
                                        wrap: code => `Boolean(${code})`,
                                    }),
                                },
                            ],
                        });
                    }
                    return;
                }
                // other
                context.report({ node, messageId: 'conditionErrorOther' });
            }
            /**
             * Check union variants for the types we care about
             */
            function inspectVariantTypes(types) {
                const variantTypes = new Set();
                if (types.some(type => tsutils.isTypeFlagSet(type, ts.TypeFlags.Null | ts.TypeFlags.Undefined | ts.TypeFlags.VoidLike))) {
                    variantTypes.add('nullish');
                }
                if (types.some(type => tsutils.isTypeFlagSet(type, ts.TypeFlags.BooleanLike))) {
                    variantTypes.add('boolean');
                }
                const strings = types.filter(type => tsutils.isTypeFlagSet(type, ts.TypeFlags.StringLike));
                if (strings.length) {
                    if (strings.some(type => type.isStringLiteral() && type.value !== '')) {
                        variantTypes.add('truthy string');
                    }
                    else {
                        variantTypes.add('string');
                    }
                }
                const numbers = types.filter(type => tsutils.isTypeFlagSet(type, ts.TypeFlags.NumberLike | ts.TypeFlags.BigIntLike));
                if (numbers.length) {
                    if (numbers.some(type => type.isNumberLiteral() && type.value !== 0)) {
                        variantTypes.add('truthy number');
                    }
                    else {
                        variantTypes.add('number');
                    }
                }
                if (types.some(type => !tsutils.isTypeFlagSet(type, ts.TypeFlags.Null |
                    ts.TypeFlags.Undefined |
                    ts.TypeFlags.VoidLike |
                    ts.TypeFlags.BooleanLike |
                    ts.TypeFlags.StringLike |
                    ts.TypeFlags.NumberLike |
                    ts.TypeFlags.BigIntLike |
                    ts.TypeFlags.Any |
                    ts.TypeFlags.Unknown |
                    ts.TypeFlags.Never))) {
                    variantTypes.add('object');
                }
                if (types.some(type => util.isTypeAnyType(type) || util.isTypeUnknownType(type))) {
                    variantTypes.add('any');
                }
                if (types.some(type => tsutils.isTypeFlagSet(type, ts.TypeFlags.Never))) {
                    variantTypes.add('never');
                }
                return variantTypes;
            }
        },
    });
    function isLogicalNegationExpression(node) {
        return node.type === experimental_utils_1.AST_NODE_TYPES.UnaryExpression && node.operator === '!';
    }
    function isArrayLengthExpression(node, typeChecker, parserServices) {
        if (node.type !== experimental_utils_1.AST_NODE_TYPES.MemberExpression) {
            return false;
        }
        if (node.computed) {
            return false;
        }
        if (node.property.name !== 'length') {
            return false;
        }
        const objectTsNode = parserServices.esTreeNodeToTSNodeMap.get(node.object);
        const objectType = util.getConstrainedTypeAtLocation(typeChecker, objectTsNode);
        return util.isTypeArrayTypeOrUnionOfArrayTypes(objectType, typeChecker);
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/switch-exhaustiveness-check.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const util_1 = require(906) /* ../util */;
    const tsutils_1 = require(916) /* tsutils */;
    exports.default = util_1.createRule({
        name: 'switch-exhaustiveness-check',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Exhaustiveness checking in switch with union type',
                category: 'Best Practices',
                recommended: false,
                suggestion: true,
                requiresTypeChecking: true,
            },
            schema: [],
            messages: {
                switchIsNotExhaustive: 'Switch is not exhaustive. Cases not matched: {{missingBranches}}',
                addMissingCases: 'Add branches for missing cases.',
            },
        },
        defaultOptions: [],
        create(context) {
            const sourceCode = context.getSourceCode();
            const service = util_1.getParserServices(context);
            const checker = service.program.getTypeChecker();
            const compilerOptions = service.program.getCompilerOptions();
            function getNodeType(node) {
                const tsNode = service.esTreeNodeToTSNodeMap.get(node);
                return util_1.getConstrainedTypeAtLocation(checker, tsNode);
            }
            function fixSwitch(fixer, node, missingBranchTypes, symbolName) {
                var _a;
                const lastCase = node.cases.length > 0 ? node.cases[node.cases.length - 1] : null;
                const caseIndent = lastCase
                    ? ' '.repeat(lastCase.loc.start.column)
                    : // if there are no cases, use indentation of the switch statement
                        // and leave it to user to format it correctly
                        ' '.repeat(node.loc.start.column);
                const missingCases = [];
                for (const missingBranchType of missingBranchTypes) {
                    // While running this rule on checker.ts of TypeScript project
                    // the fix introduced a compiler error due to:
                    //
                    // type __String = (string & {
                    //         __escapedIdentifier: void;
                    //     }) | (void & {
                    //         __escapedIdentifier: void;
                    //     }) | InternalSymbolName;
                    //
                    // The following check fixes it.
                    if (missingBranchType.isIntersection()) {
                        continue;
                    }
                    const missingBranchName = (_a = missingBranchType.getSymbol()) === null || _a === void 0 ? void 0 : _a.escapedName;
                    let caseTest = checker.typeToString(missingBranchType);
                    if (symbolName &&
                        (missingBranchName || missingBranchName === '') &&
                        util_1.requiresQuoting(missingBranchName.toString(), compilerOptions.target)) {
                        caseTest = `${symbolName}['${missingBranchName}']`;
                    }
                    const errorMessage = `Not implemented yet: ${caseTest} case`;
                    missingCases.push(`case ${caseTest}: { throw new Error('${errorMessage}') }`);
                }
                const fixString = missingCases
                    .map(code => `${caseIndent}${code}`)
                    .join('\n');
                if (lastCase) {
                    return fixer.insertTextAfter(lastCase, `\n${fixString}`);
                }
                // there were no existing cases
                const openingBrace = sourceCode.getTokenAfter(node.discriminant, util_1.isOpeningBraceToken);
                const closingBrace = sourceCode.getTokenAfter(node.discriminant, util_1.isClosingBraceToken);
                return fixer.replaceTextRange([openingBrace.range[0], closingBrace.range[1]], ['{', fixString, `${caseIndent}}`].join('\n'));
            }
            function checkSwitchExhaustive(node) {
                var _a;
                const discriminantType = getNodeType(node.discriminant);
                const symbolName = (_a = discriminantType.getSymbol()) === null || _a === void 0 ? void 0 : _a.escapedName;
                if (discriminantType.isUnion()) {
                    const unionTypes = tsutils_1.unionTypeParts(discriminantType);
                    const caseTypes = new Set();
                    for (const switchCase of node.cases) {
                        if (switchCase.test === null) {
                            // Switch has 'default' branch - do nothing.
                            return;
                        }
                        caseTypes.add(getNodeType(switchCase.test));
                    }
                    const missingBranchTypes = unionTypes.filter(unionType => !caseTypes.has(unionType));
                    if (missingBranchTypes.length === 0) {
                        // All cases matched - do nothing.
                        return;
                    }
                    context.report({
                        node: node.discriminant,
                        messageId: 'switchIsNotExhaustive',
                        data: {
                            missingBranches: missingBranchTypes
                                .map(missingType => {
                                var _a;
                                return tsutils_1.isTypeFlagSet(missingType, ts.TypeFlags.ESSymbolLike)
                                    ? `typeof ${(_a = missingType.getSymbol()) === null || _a === void 0 ? void 0 : _a.escapedName}`
                                    : checker.typeToString(missingType);
                            })
                                .join(' | '),
                        },
                        suggest: [
                            {
                                messageId: 'addMissingCases',
                                fix(fixer) {
                                    return fixSwitch(fixer, node, missingBranchTypes, symbolName === null || symbolName === void 0 ? void 0 : symbolName.toString());
                                },
                            },
                        ],
                    });
                }
            }
            return {
                SwitchStatement: checkSwitchExhaustive,
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/triple-slash-reference.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'triple-slash-reference',
        meta: {
            type: 'suggestion',
            docs: {
                description: 'Sets preference level for triple slash directives versus ES6-style import declarations',
                category: 'Best Practices',
                recommended: 'error',
            },
            messages: {
                tripleSlashReference: 'Do not use a triple slash reference for {{module}}, use `import` style instead.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        lib: {
                            enum: ['always', 'never'],
                        },
                        path: {
                            enum: ['always', 'never'],
                        },
                        types: {
                            enum: ['always', 'never', 'prefer-import'],
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            {
                lib: 'always',
                path: 'never',
                types: 'prefer-import',
            },
        ],
        create(context, [{ lib, path, types }]) {
            let programNode;
            const sourceCode = context.getSourceCode();
            const references = [];
            function hasMatchingReference(source) {
                references.forEach(reference => {
                    if (reference.importName === source.value) {
                        context.report({
                            node: reference.comment,
                            messageId: 'tripleSlashReference',
                            data: {
                                module: reference.importName,
                            },
                        });
                    }
                });
            }
            return {
                ImportDeclaration(node) {
                    if (programNode) {
                        hasMatchingReference(node.source);
                    }
                },
                TSImportEqualsDeclaration(node) {
                    if (programNode) {
                        const reference = node.moduleReference;
                        if (reference.type === experimental_utils_1.AST_NODE_TYPES.TSExternalModuleReference) {
                            hasMatchingReference(reference.expression);
                        }
                    }
                },
                Program(node) {
                    if (lib === 'always' && path === 'always' && types == 'always') {
                        return;
                    }
                    programNode = node;
                    const referenceRegExp = /^\/\s*<reference\s*(types|path|lib)\s*=\s*["|'](.*)["|']/;
                    const commentsBefore = sourceCode.getCommentsBefore(programNode);
                    commentsBefore.forEach(comment => {
                        if (comment.type !== experimental_utils_1.AST_TOKEN_TYPES.Line) {
                            return;
                        }
                        const referenceResult = referenceRegExp.exec(comment.value);
                        if (referenceResult) {
                            if ((referenceResult[1] === 'types' && types === 'never') ||
                                (referenceResult[1] === 'path' && path === 'never') ||
                                (referenceResult[1] === 'lib' && lib === 'never')) {
                                context.report({
                                    node: comment,
                                    messageId: 'tripleSlashReference',
                                    data: {
                                        module: referenceResult[2],
                                    },
                                });
                                return;
                            }
                            if (referenceResult[1] === 'types' && types === 'prefer-import') {
                                references.push({ comment, importName: referenceResult[2] });
                            }
                        }
                    });
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/type-annotation-spacing.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const util = __importStar(require(906) /* ../util */);
    const util_1 = require(906) /* ../util */;
    const definition = {
        type: 'object',
        properties: {
            before: { type: 'boolean' },
            after: { type: 'boolean' },
        },
        additionalProperties: false,
    };
    function createRules(options) {
        var _a;
        const globals = Object.assign(Object.assign({}, ((options === null || options === void 0 ? void 0 : options.before) !== undefined ? { before: options.before } : {})), ((options === null || options === void 0 ? void 0 : options.after) !== undefined ? { after: options.after } : {}));
        const override = (_a = options === null || options === void 0 ? void 0 : options.overrides) !== null && _a !== void 0 ? _a : {};
        const colon = Object.assign(Object.assign({ before: false, after: true }, globals), override === null || override === void 0 ? void 0 : override.colon);
        const arrow = Object.assign(Object.assign({ before: true, after: true }, globals), override === null || override === void 0 ? void 0 : override.arrow);
        return {
            colon: colon,
            arrow: arrow,
            variable: Object.assign(Object.assign({}, colon), override === null || override === void 0 ? void 0 : override.variable),
            property: Object.assign(Object.assign({}, colon), override === null || override === void 0 ? void 0 : override.property),
            parameter: Object.assign(Object.assign({}, colon), override === null || override === void 0 ? void 0 : override.parameter),
            returnType: Object.assign(Object.assign({}, colon), override === null || override === void 0 ? void 0 : override.returnType),
        };
    }
    function getIdentifierRules(rules, node) {
        const scope = node === null || node === void 0 ? void 0 : node.parent;
        if (util_1.isVariableDeclarator(scope)) {
            return rules.variable;
        }
        else if (util_1.isFunctionOrFunctionType(scope)) {
            return rules.parameter;
        }
        else {
            return rules.colon;
        }
    }
    function getRules(rules, node) {
        var _a;
        const scope = (_a = node === null || node === void 0 ? void 0 : node.parent) === null || _a === void 0 ? void 0 : _a.parent;
        if (util_1.isTSFunctionType(scope) || util_1.isTSConstructorType(scope)) {
            return rules.arrow;
        }
        else if (util_1.isIdentifier(scope)) {
            return getIdentifierRules(rules, scope);
        }
        else if (util_1.isClassOrTypeElement(scope)) {
            return rules.property;
        }
        else if (util_1.isFunction(scope)) {
            return rules.returnType;
        }
        else {
            return rules.colon;
        }
    }
    exports.default = util.createRule({
        name: 'type-annotation-spacing',
        meta: {
            type: 'layout',
            docs: {
                description: 'Require consistent spacing around type annotations',
                category: 'Stylistic Issues',
                recommended: false,
            },
            fixable: 'whitespace',
            messages: {
                expectedSpaceAfter: "Expected a space after the '{{type}}'.",
                expectedSpaceBefore: "Expected a space before the '{{type}}'.",
                unexpectedSpaceAfter: "Unexpected space after the '{{type}}'.",
                unexpectedSpaceBefore: "Unexpected space before the '{{type}}'.",
                unexpectedSpaceBetween: "Unexpected space between the '{{previousToken}}' and the '{{type}}'.",
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        before: { type: 'boolean' },
                        after: { type: 'boolean' },
                        overrides: {
                            type: 'object',
                            properties: {
                                colon: definition,
                                arrow: definition,
                                variable: definition,
                                parameter: definition,
                                property: definition,
                                returnType: definition,
                            },
                            additionalProperties: false,
                        },
                    },
                    additionalProperties: false,
                },
            ],
        },
        defaultOptions: [
            // technically there is a default, but the overrides mean
            // that if we apply them here, it will break the no override case.
            {},
        ],
        create(context, [options]) {
            const punctuators = [':', '=>'];
            const sourceCode = context.getSourceCode();
            const ruleSet = createRules(options);
            /**
             * Checks if there's proper spacing around type annotations (no space
             * before colon, one space after).
             */
            function checkTypeAnnotationSpacing(typeAnnotation) {
                const nextToken = typeAnnotation;
                const punctuatorTokenEnd = sourceCode.getTokenBefore(nextToken);
                let punctuatorTokenStart = punctuatorTokenEnd;
                let previousToken = sourceCode.getTokenBefore(punctuatorTokenEnd);
                let type = punctuatorTokenEnd.value;
                if (!punctuators.includes(type)) {
                    return;
                }
                const { before, after } = getRules(ruleSet, typeAnnotation);
                if (type === ':' && previousToken.value === '?') {
                    if (sourceCode.isSpaceBetweenTokens(previousToken, punctuatorTokenStart)) {
                        context.report({
                            node: punctuatorTokenStart,
                            messageId: 'unexpectedSpaceBetween',
                            data: {
                                type,
                                previousToken: previousToken.value,
                            },
                            fix(fixer) {
                                return fixer.removeRange([
                                    previousToken.range[1],
                                    punctuatorTokenStart.range[0],
                                ]);
                            },
                        });
                    }
                    // shift the start to the ?
                    type = '?:';
                    punctuatorTokenStart = previousToken;
                    previousToken = sourceCode.getTokenBefore(previousToken);
                    // handle the +/- modifiers for optional modification operators
                    if (previousToken.value === '+' || previousToken.value === '-') {
                        type = `${previousToken.value}?:`;
                        punctuatorTokenStart = previousToken;
                        previousToken = sourceCode.getTokenBefore(previousToken);
                    }
                }
                const previousDelta = punctuatorTokenStart.range[0] - previousToken.range[1];
                const nextDelta = nextToken.range[0] - punctuatorTokenEnd.range[1];
                if (after && nextDelta === 0) {
                    context.report({
                        node: punctuatorTokenEnd,
                        messageId: 'expectedSpaceAfter',
                        data: {
                            type,
                        },
                        fix(fixer) {
                            return fixer.insertTextAfter(punctuatorTokenEnd, ' ');
                        },
                    });
                }
                else if (!after && nextDelta > 0) {
                    context.report({
                        node: punctuatorTokenEnd,
                        messageId: 'unexpectedSpaceAfter',
                        data: {
                            type,
                        },
                        fix(fixer) {
                            return fixer.removeRange([
                                punctuatorTokenEnd.range[1],
                                nextToken.range[0],
                            ]);
                        },
                    });
                }
                if (before && previousDelta === 0) {
                    context.report({
                        node: punctuatorTokenStart,
                        messageId: 'expectedSpaceBefore',
                        data: {
                            type,
                        },
                        fix(fixer) {
                            return fixer.insertTextAfter(previousToken, ' ');
                        },
                    });
                }
                else if (!before && previousDelta > 0) {
                    context.report({
                        node: punctuatorTokenStart,
                        messageId: 'unexpectedSpaceBefore',
                        data: {
                            type,
                        },
                        fix(fixer) {
                            return fixer.removeRange([
                                previousToken.range[1],
                                punctuatorTokenStart.range[0],
                            ]);
                        },
                    });
                }
            }
            return {
                TSMappedType(node) {
                    if (node.typeAnnotation) {
                        checkTypeAnnotationSpacing(node.typeAnnotation);
                    }
                },
                TSTypeAnnotation(node) {
                    checkTypeAnnotationSpacing(node.typeAnnotation);
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/typedef.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'typedef',
        meta: {
            docs: {
                description: 'Requires type annotations to exist',
                category: 'Stylistic Issues',
                recommended: false,
            },
            messages: {
                expectedTypedef: 'Expected a type annotation.',
                expectedTypedefNamed: 'Expected {{name}} to have a type annotation.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        ["arrayDestructuring" /* ArrayDestructuring */]: { type: 'boolean' },
                        ["arrowParameter" /* ArrowParameter */]: { type: 'boolean' },
                        ["memberVariableDeclaration" /* MemberVariableDeclaration */]: { type: 'boolean' },
                        ["objectDestructuring" /* ObjectDestructuring */]: { type: 'boolean' },
                        ["parameter" /* Parameter */]: { type: 'boolean' },
                        ["propertyDeclaration" /* PropertyDeclaration */]: { type: 'boolean' },
                        ["variableDeclaration" /* VariableDeclaration */]: { type: 'boolean' },
                        ["variableDeclarationIgnoreFunction" /* VariableDeclarationIgnoreFunction */]: { type: 'boolean' },
                    },
                },
            ],
            type: 'suggestion',
        },
        defaultOptions: [
            {
                ["arrayDestructuring" /* ArrayDestructuring */]: false,
                ["arrowParameter" /* ArrowParameter */]: false,
                ["memberVariableDeclaration" /* MemberVariableDeclaration */]: false,
                ["objectDestructuring" /* ObjectDestructuring */]: false,
                ["parameter" /* Parameter */]: false,
                ["propertyDeclaration" /* PropertyDeclaration */]: false,
                ["variableDeclaration" /* VariableDeclaration */]: false,
                ["variableDeclarationIgnoreFunction" /* VariableDeclarationIgnoreFunction */]: false,
            },
        ],
        create(context, [options]) {
            function report(location, name) {
                context.report({
                    node: location,
                    messageId: name ? 'expectedTypedefNamed' : 'expectedTypedef',
                    data: { name },
                });
            }
            function getNodeName(node) {
                return node.type === experimental_utils_1.AST_NODE_TYPES.Identifier ? node.name : undefined;
            }
            function isForOfStatementContext(node) {
                let current = node.parent;
                while (current) {
                    switch (current.type) {
                        case experimental_utils_1.AST_NODE_TYPES.VariableDeclarator:
                        case experimental_utils_1.AST_NODE_TYPES.VariableDeclaration:
                        case experimental_utils_1.AST_NODE_TYPES.ObjectPattern:
                        case experimental_utils_1.AST_NODE_TYPES.ArrayPattern:
                        case experimental_utils_1.AST_NODE_TYPES.Property:
                            current = current.parent;
                            break;
                        case experimental_utils_1.AST_NODE_TYPES.ForOfStatement:
                            return true;
                        default:
                            current = undefined;
                    }
                }
                return false;
            }
            function checkParameters(params) {
                for (const param of params) {
                    let annotationNode;
                    switch (param.type) {
                        case experimental_utils_1.AST_NODE_TYPES.AssignmentPattern:
                            annotationNode = param.left;
                            break;
                        case experimental_utils_1.AST_NODE_TYPES.TSParameterProperty:
                            annotationNode = param.parameter;
                            // Check TS parameter property with default value like `constructor(private param: string = 'something') {}`
                            if (annotationNode &&
                                annotationNode.type === experimental_utils_1.AST_NODE_TYPES.AssignmentPattern) {
                                annotationNode = annotationNode.left;
                            }
                            break;
                        default:
                            annotationNode = param;
                            break;
                    }
                    if (annotationNode !== undefined && !annotationNode.typeAnnotation) {
                        report(param, getNodeName(param));
                    }
                }
            }
            function isVariableDeclarationIgnoreFunction(node) {
                return (!!options["variableDeclarationIgnoreFunction" /* VariableDeclarationIgnoreFunction */] &&
                    (node.type === experimental_utils_1.AST_NODE_TYPES.FunctionExpression ||
                        node.type === experimental_utils_1.AST_NODE_TYPES.ArrowFunctionExpression));
            }
            return {
                ArrayPattern(node) {
                    var _a;
                    if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === experimental_utils_1.AST_NODE_TYPES.RestElement &&
                        node.parent.typeAnnotation) {
                        return;
                    }
                    if (options["arrayDestructuring" /* ArrayDestructuring */] &&
                        !node.typeAnnotation &&
                        !isForOfStatementContext(node)) {
                        report(node);
                    }
                },
                ArrowFunctionExpression(node) {
                    if (options["arrowParameter" /* ArrowParameter */]) {
                        checkParameters(node.params);
                    }
                },
                ClassProperty(node) {
                    if (node.value && isVariableDeclarationIgnoreFunction(node.value)) {
                        return;
                    }
                    if (options["memberVariableDeclaration" /* MemberVariableDeclaration */] &&
                        !node.typeAnnotation) {
                        report(node, node.key.type === experimental_utils_1.AST_NODE_TYPES.Identifier
                            ? node.key.name
                            : undefined);
                    }
                },
                'FunctionDeclaration, FunctionExpression'(node) {
                    if (options["parameter" /* Parameter */]) {
                        checkParameters(node.params);
                    }
                },
                ObjectPattern(node) {
                    if (options["objectDestructuring" /* ObjectDestructuring */] &&
                        !node.typeAnnotation &&
                        !isForOfStatementContext(node)) {
                        report(node);
                    }
                },
                'TSIndexSignature, TSPropertySignature'(node) {
                    if (options["propertyDeclaration" /* PropertyDeclaration */] && !node.typeAnnotation) {
                        report(node, node.type === experimental_utils_1.AST_NODE_TYPES.TSPropertySignature
                            ? getNodeName(node.key)
                            : undefined);
                    }
                },
                VariableDeclarator(node) {
                    if (!options["variableDeclaration" /* VariableDeclaration */] ||
                        node.id.typeAnnotation ||
                        (node.id.type === experimental_utils_1.AST_NODE_TYPES.ArrayPattern &&
                            !options["arrayDestructuring" /* ArrayDestructuring */]) ||
                        (node.id.type === experimental_utils_1.AST_NODE_TYPES.ObjectPattern &&
                            !options["objectDestructuring" /* ObjectDestructuring */]) ||
                        (node.init && isVariableDeclarationIgnoreFunction(node.init))) {
                        return;
                    }
                    let current = node.parent;
                    while (current) {
                        switch (current.type) {
                            case experimental_utils_1.AST_NODE_TYPES.VariableDeclaration:
                                // Keep looking upwards
                                current = current.parent;
                                break;
                            case experimental_utils_1.AST_NODE_TYPES.ForOfStatement:
                            case experimental_utils_1.AST_NODE_TYPES.ForInStatement:
                                // Stop traversing and don't report an error
                                return;
                            default:
                                // Stop traversing
                                current = undefined;
                                break;
                        }
                    }
                    report(node, getNodeName(node.id));
                },
            };
        },
    });
},
/* @typescript-eslint/eslint-plugin/dist/rules/unbound-method.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const tsutils = __importStar(require(916) /* tsutils */);
    const ts = __importStar(require(21) /* typescript */);
    const util = __importStar(require(906) /* ../util */);
    /**
     * The following is a list of exceptions to the rule
     * Generated via the following script.
     * This is statically defined to save making purposely invalid calls every lint run
     * ```
    SUPPORTED_GLOBALS.flatMap(namespace => {
      const object = window[namespace];
        return Object.getOwnPropertyNames(object)
          .filter(
            name =>
              !name.startsWith('_') &&
              typeof object[name] === 'function',
          )
          .map(name => {
            try {
              const x = object[name];
              x();
            } catch (e) {
              if (e.message.includes("called on non-object")) {
                return `${namespace}.${name}`;
              }
            }
          });
    }).filter(Boolean);
       * ```
     */
    const nativelyNotBoundMembers = new Set([
        'Promise.all',
        'Promise.race',
        'Promise.resolve',
        'Promise.reject',
        'Promise.allSettled',
        'Object.defineProperties',
        'Object.defineProperty',
        'Reflect.defineProperty',
        'Reflect.deleteProperty',
        'Reflect.get',
        'Reflect.getOwnPropertyDescriptor',
        'Reflect.getPrototypeOf',
        'Reflect.has',
        'Reflect.isExtensible',
        'Reflect.ownKeys',
        'Reflect.preventExtensions',
        'Reflect.set',
        'Reflect.setPrototypeOf',
    ]);
    const SUPPORTED_GLOBALS = [
        'Number',
        'Object',
        'String',
        'RegExp',
        'Symbol',
        'Array',
        'Proxy',
        'Date',
        'Infinity',
        'Atomics',
        'Reflect',
        'console',
        'Math',
        'JSON',
        'Intl',
    ];
    const nativelyBoundMembers = SUPPORTED_GLOBALS.map(namespace => {
        if (!(namespace in global)) {
            // node.js might not have namespaces like Intl depending on compilation options
            // https://nodejs.org/api/intl.html#intl_options_for_building_node_js
            return [];
        }
        const object = global[namespace];
        return Object.getOwnPropertyNames(object)
            .filter(name => !name.startsWith('_') &&
            typeof object[name] === 'function')
            .map(name => `${namespace}.${name}`);
    })
        .reduce((arr, names) => arr.concat(names), [])
        .filter(name => !nativelyNotBoundMembers.has(name));
    const isNotImported = (symbol, currentSourceFile) => {
        const { valueDeclaration } = symbol;
        if (!valueDeclaration) {
            // working around https://github.com/microsoft/TypeScript/issues/31294
            return false;
        }
        return (!!currentSourceFile &&
            currentSourceFile !== valueDeclaration.getSourceFile());
    };
    const getNodeName = (node) => node.type === experimental_utils_1.AST_NODE_TYPES.Identifier ? node.name : null;
    const getMemberFullName = (node) => `${getNodeName(node.object)}.${getNodeName(node.property)}`;
    const BASE_MESSAGE = 'Avoid referencing unbound methods which may cause unintentional scoping of `this`.';
    exports.default = util.createRule({
        name: 'unbound-method',
        meta: {
            docs: {
                category: 'Best Practices',
                description: 'Enforces unbound methods are called with their expected scope',
                recommended: 'error',
                requiresTypeChecking: true,
            },
            messages: {
                unbound: BASE_MESSAGE,
                unboundWithoutThisAnnotation: BASE_MESSAGE +
                    '\n' +
                    'If your function does not access `this`, you can annotate it with `this: void`, or consider using an arrow function instead.',
            },
            schema: [
                {
                    type: 'object',
                    properties: {
                        ignoreStatic: {
                            type: 'boolean',
                        },
                    },
                    additionalProperties: false,
                },
            ],
            type: 'problem',
        },
        defaultOptions: [
            {
                ignoreStatic: false,
            },
        ],
        create(context, [{ ignoreStatic }]) {
            const parserServices = util.getParserServices(context);
            const checker = parserServices.program.getTypeChecker();
            const currentSourceFile = parserServices.program.getSourceFile(context.getFilename());
            function checkMethodAndReport(node, symbol) {
                if (!symbol) {
                    return;
                }
                const { dangerous, firstParamIsThis } = checkMethod(symbol, ignoreStatic);
                if (dangerous) {
                    context.report({
                        messageId: firstParamIsThis === false
                            ? 'unboundWithoutThisAnnotation'
                            : 'unbound',
                        node,
                    });
                }
            }
            return {
                MemberExpression(node) {
                    if (isSafeUse(node)) {
                        return;
                    }
                    const objectSymbol = checker.getSymbolAtLocation(parserServices.esTreeNodeToTSNodeMap.get(node.object));
                    if (objectSymbol &&
                        nativelyBoundMembers.includes(getMemberFullName(node)) &&
                        isNotImported(objectSymbol, currentSourceFile)) {
                        return;
                    }
                    const originalNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                    checkMethodAndReport(node, checker.getSymbolAtLocation(originalNode));
                },
                'VariableDeclarator, AssignmentExpression'(node) {
                    const [idNode, initNode] = node.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclarator
                        ? [node.id, node.init]
                        : [node.left, node.right];
                    if (initNode && idNode.type === experimental_utils_1.AST_NODE_TYPES.ObjectPattern) {
                        const tsNode = parserServices.esTreeNodeToTSNodeMap.get(initNode);
                        const rightSymbol = checker.getSymbolAtLocation(tsNode);
                        const initTypes = checker.getTypeAtLocation(tsNode);
                        const notImported = rightSymbol && isNotImported(rightSymbol, currentSourceFile);
                        idNode.properties.forEach(property => {
                            if (property.type === experimental_utils_1.AST_NODE_TYPES.Property &&
                                property.key.type === experimental_utils_1.AST_NODE_TYPES.Identifier) {
                                if (notImported &&
                                    util.isIdentifier(initNode) &&
                                    nativelyBoundMembers.includes(`${initNode.name}.${property.key.name}`)) {
                                    return;
                                }
                                checkMethodAndReport(node, initTypes.getProperty(property.key.name));
                            }
                        });
                    }
                },
            };
        },
    });
    function checkMethod(symbol, ignoreStatic) {
        var _a, _b;
        const { valueDeclaration } = symbol;
        if (!valueDeclaration) {
            // working around https://github.com/microsoft/TypeScript/issues/31294
            return { dangerous: false };
        }
        switch (valueDeclaration.kind) {
            case ts.SyntaxKind.PropertyDeclaration:
                return {
                    dangerous: ((_a = valueDeclaration.initializer) === null || _a === void 0 ? void 0 : _a.kind) ===
                        ts.SyntaxKind.FunctionExpression,
                };
            case ts.SyntaxKind.MethodDeclaration:
            case ts.SyntaxKind.MethodSignature: {
                const decl = valueDeclaration;
                const firstParam = decl.parameters[0];
                const firstParamIsThis = (firstParam === null || firstParam === void 0 ? void 0 : firstParam.name.kind) === ts.SyntaxKind.Identifier &&
                    (firstParam === null || firstParam === void 0 ? void 0 : firstParam.name.escapedText) === 'this';
                const thisArgIsVoid = firstParamIsThis &&
                    ((_b = firstParam === null || firstParam === void 0 ? void 0 : firstParam.type) === null || _b === void 0 ? void 0 : _b.kind) === ts.SyntaxKind.VoidKeyword;
                return {
                    dangerous: !thisArgIsVoid &&
                        !(ignoreStatic &&
                            tsutils.hasModifier(valueDeclaration.modifiers, ts.SyntaxKind.StaticKeyword)),
                    firstParamIsThis,
                };
            }
        }
        return { dangerous: false };
    }
    function isSafeUse(node) {
        const parent = node.parent;
        switch (parent === null || parent === void 0 ? void 0 : parent.type) {
            case experimental_utils_1.AST_NODE_TYPES.IfStatement:
            case experimental_utils_1.AST_NODE_TYPES.ForStatement:
            case experimental_utils_1.AST_NODE_TYPES.MemberExpression:
            case experimental_utils_1.AST_NODE_TYPES.SwitchStatement:
            case experimental_utils_1.AST_NODE_TYPES.UpdateExpression:
            case experimental_utils_1.AST_NODE_TYPES.WhileStatement:
                return true;
            case experimental_utils_1.AST_NODE_TYPES.CallExpression:
                return parent.callee === node;
            case experimental_utils_1.AST_NODE_TYPES.ConditionalExpression:
                return parent.test === node;
            case experimental_utils_1.AST_NODE_TYPES.TaggedTemplateExpression:
                return parent.tag === node;
            case experimental_utils_1.AST_NODE_TYPES.UnaryExpression:
                // the first case is safe for obvious
                // reasons. The second one is also fine
                // since we're returning something falsy
                return ['typeof', '!', 'void', 'delete'].includes(parent.operator);
            case experimental_utils_1.AST_NODE_TYPES.BinaryExpression:
                return ['instanceof', '==', '!=', '===', '!=='].includes(parent.operator);
            case experimental_utils_1.AST_NODE_TYPES.AssignmentExpression:
                return (parent.operator === '=' &&
                    (node === parent.left ||
                        (node.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                            node.object.type === experimental_utils_1.AST_NODE_TYPES.Super &&
                            parent.left.type === experimental_utils_1.AST_NODE_TYPES.MemberExpression &&
                            parent.left.object.type === experimental_utils_1.AST_NODE_TYPES.ThisExpression)));
            case experimental_utils_1.AST_NODE_TYPES.ChainExpression:
            case experimental_utils_1.AST_NODE_TYPES.TSNonNullExpression:
            case experimental_utils_1.AST_NODE_TYPES.TSAsExpression:
            case experimental_utils_1.AST_NODE_TYPES.TSTypeAssertion:
                return isSafeUse(parent);
            case experimental_utils_1.AST_NODE_TYPES.LogicalExpression:
                if (parent.operator === '&&' && parent.left === node) {
                    // this is safe, as && will return the left if and only if it's falsy
                    return true;
                }
                // in all other cases, it's likely the logical expression will return the method ref
                // so make sure the parent is a safe usage
                return isSafeUse(parent);
        }
        return false;
    }
},
/* @typescript-eslint/eslint-plugin/dist/rules/unified-signatures.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const experimental_utils_1 = require(735) /* @typescript-eslint/experimental-utils */;
    const util = __importStar(require(906) /* ../util */);
    exports.default = util.createRule({
        name: 'unified-signatures',
        meta: {
            docs: {
                description: 'Warns for any two overloads that could be unified into one by using a union or an optional/rest parameter',
                category: 'Variables',
                // too opinionated to be recommended
                recommended: false,
            },
            type: 'suggestion',
            messages: {
                omittingRestParameter: '{{failureStringStart}} with a rest parameter.',
                omittingSingleParameter: '{{failureStringStart}} with an optional parameter.',
                singleParameterDifference: '{{failureStringStart}} taking `{{type1}} | {{type2}}`.',
            },
            schema: [],
        },
        defaultOptions: [],
        create(context) {
            const sourceCode = context.getSourceCode();
            //----------------------------------------------------------------------
            // Helpers
            //----------------------------------------------------------------------
            function failureStringStart(otherLine) {
                // For only 2 overloads we don't need to specify which is the other one.
                const overloads = otherLine === undefined
                    ? 'These overloads'
                    : `This overload and the one on line ${otherLine}`;
                return `${overloads} can be combined into one signature`;
            }
            function addFailures(failures) {
                for (const failure of failures) {
                    const { unify, only2 } = failure;
                    switch (unify.kind) {
                        case 'single-parameter-difference': {
                            const { p0, p1 } = unify;
                            const lineOfOtherOverload = only2 ? undefined : p0.loc.start.line;
                            const typeAnnotation0 = isTSParameterProperty(p0)
                                ? p0.parameter.typeAnnotation
                                : p0.typeAnnotation;
                            const typeAnnotation1 = isTSParameterProperty(p1)
                                ? p1.parameter.typeAnnotation
                                : p1.typeAnnotation;
                            context.report({
                                loc: p1.loc,
                                messageId: 'singleParameterDifference',
                                data: {
                                    failureStringStart: failureStringStart(lineOfOtherOverload),
                                    type1: sourceCode.getText(typeAnnotation0 === null || typeAnnotation0 === void 0 ? void 0 : typeAnnotation0.typeAnnotation),
                                    type2: sourceCode.getText(typeAnnotation1 === null || typeAnnotation1 === void 0 ? void 0 : typeAnnotation1.typeAnnotation),
                                },
                                node: p1,
                            });
                            break;
                        }
                        case 'extra-parameter': {
                            const { extraParameter, otherSignature } = unify;
                            const lineOfOtherOverload = only2
                                ? undefined
                                : otherSignature.loc.start.line;
                            context.report({
                                loc: extraParameter.loc,
                                messageId: extraParameter.type === experimental_utils_1.AST_NODE_TYPES.RestElement
                                    ? 'omittingRestParameter'
                                    : 'omittingSingleParameter',
                                data: {
                                    failureStringStart: failureStringStart(lineOfOtherOverload),
                                },
                                node: extraParameter,
                            });
                        }
                    }
                }
            }
            function checkOverloads(signatures, typeParameters) {
                const result = [];
                const isTypeParameter = getIsTypeParameter(typeParameters);
                for (const overloads of signatures) {
                    if (overloads.length === 2) {
                        const signature0 = overloads[0].value || overloads[0];
                        const signature1 = overloads[1].value || overloads[1];
                        const unify = compareSignatures(signature0, signature1, isTypeParameter);
                        if (unify !== undefined) {
                            result.push({ unify, only2: true });
                        }
                    }
                    else {
                        forEachPair(overloads, (a, b) => {
                            const signature0 = a.value || a;
                            const signature1 = b.value || b;
                            const unify = compareSignatures(signature0, signature1, isTypeParameter);
                            if (unify !== undefined) {
                                result.push({ unify, only2: false });
                            }
                        });
                    }
                }
                return result;
            }
            function compareSignatures(a, b, isTypeParameter) {
                if (!signaturesCanBeUnified(a, b, isTypeParameter)) {
                    return undefined;
                }
                return a.params.length === b.params.length
                    ? signaturesDifferBySingleParameter(a.params, b.params)
                    : signaturesDifferByOptionalOrRestParameter(a, b);
            }
            function signaturesCanBeUnified(a, b, isTypeParameter) {
                // Must return the same type.
                const aTypeParams = a.typeParameters !== undefined ? a.typeParameters.params : undefined;
                const bTypeParams = b.typeParameters !== undefined ? b.typeParameters.params : undefined;
                return (typesAreEqual(a.returnType, b.returnType) &&
                    // Must take the same type parameters.
                    // If one uses a type parameter (from outside) and the other doesn't, they shouldn't be joined.
                    util.arraysAreEqual(aTypeParams, bTypeParams, typeParametersAreEqual) &&
                    signatureUsesTypeParameter(a, isTypeParameter) ===
                        signatureUsesTypeParameter(b, isTypeParameter));
            }
            /** Detect `a(x: number, y: number, z: number)` and `a(x: number, y: string, z: number)`. */
            function signaturesDifferBySingleParameter(types1, types2) {
                const index = getIndexOfFirstDifference(types1, types2, parametersAreEqual);
                if (index === undefined) {
                    return undefined;
                }
                // If remaining arrays are equal, the signatures differ by just one parameter type
                if (!util.arraysAreEqual(types1.slice(index + 1), types2.slice(index + 1), parametersAreEqual)) {
                    return undefined;
                }
                const a = types1[index];
                const b = types2[index];
                // Can unify `a?: string` and `b?: number`. Can't unify `...args: string[]` and `...args: number[]`.
                // See https://github.com/Microsoft/TypeScript/issues/5077
                return parametersHaveEqualSigils(a, b) &&
                    a.type !== experimental_utils_1.AST_NODE_TYPES.RestElement
                    ? { kind: 'single-parameter-difference', p0: a, p1: b }
                    : undefined;
            }
            /**
             * Detect `a(): void` and `a(x: number): void`.
             * Returns the parameter declaration (`x: number` in this example) that should be optional/rest, and overload it's a part of.
             */
            function signaturesDifferByOptionalOrRestParameter(a, b) {
                const sig1 = a.params;
                const sig2 = b.params;
                const minLength = Math.min(sig1.length, sig2.length);
                const longer = sig1.length < sig2.length ? sig2 : sig1;
                const shorter = sig1.length < sig2.length ? sig1 : sig2;
                const shorterSig = sig1.length < sig2.length ? a : b;
                // If one is has 2+ parameters more than the other, they must all be optional/rest.
                // Differ by optional parameters: f() and f(x), f() and f(x, ?y, ...z)
                // Not allowed: f() and f(x, y)
                for (let i = minLength + 1; i < longer.length; i++) {
                    if (!parameterMayBeMissing(longer[i])) {
                        return undefined;
                    }
                }
                for (let i = 0; i < minLength; i++) {
                    const sig1i = sig1[i];
                    const sig2i = sig2[i];
                    const typeAnnotation1 = isTSParameterProperty(sig1i)
                        ? sig1i.parameter.typeAnnotation
                        : sig1i.typeAnnotation;
                    const typeAnnotation2 = isTSParameterProperty(sig2i)
                        ? sig2i.parameter.typeAnnotation
                        : sig2i.typeAnnotation;
                    if (!typesAreEqual(typeAnnotation1, typeAnnotation2)) {
                        return undefined;
                    }
                }
                if (minLength > 0 &&
                    shorter[minLength - 1].type === experimental_utils_1.AST_NODE_TYPES.RestElement) {
                    return undefined;
                }
                return {
                    extraParameter: longer[longer.length - 1],
                    kind: 'extra-parameter',
                    otherSignature: shorterSig,
                };
            }
            /** Given type parameters, returns a function to test whether a type is one of those parameters. */
            function getIsTypeParameter(typeParameters) {
                if (typeParameters === undefined) {
                    return (() => false);
                }
                const set = new Set();
                for (const t of typeParameters.params) {
                    set.add(t.name.name);
                }
                return (typeName => set.has(typeName));
            }
            /** True if any of the outer type parameters are used in a signature. */
            function signatureUsesTypeParameter(sig, isTypeParameter) {
                return sig.params.some((p) => typeContainsTypeParameter(isTSParameterProperty(p)
                    ? p.parameter.typeAnnotation
                    : p.typeAnnotation));
                function typeContainsTypeParameter(type) {
                    if (!type) {
                        return false;
                    }
                    if (type.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference) {
                        const typeName = type.typeName;
                        if (isIdentifier(typeName) && isTypeParameter(typeName.name)) {
                            return true;
                        }
                    }
                    return typeContainsTypeParameter(type.typeAnnotation ||
                        type.elementType);
                }
            }
            function isTSParameterProperty(node) {
                return (node.type ===
                    experimental_utils_1.AST_NODE_TYPES.TSParameterProperty);
            }
            function parametersAreEqual(a, b) {
                const typeAnnotationA = isTSParameterProperty(a)
                    ? a.parameter.typeAnnotation
                    : a.typeAnnotation;
                const typeAnnotationB = isTSParameterProperty(b)
                    ? b.parameter.typeAnnotation
                    : b.typeAnnotation;
                return (parametersHaveEqualSigils(a, b) &&
                    typesAreEqual(typeAnnotationA, typeAnnotationB));
            }
            /** True for optional/rest parameters. */
            function parameterMayBeMissing(p) {
                const optional = isTSParameterProperty(p)
                    ? p.parameter.optional
                    : p.optional;
                return p.type === experimental_utils_1.AST_NODE_TYPES.RestElement || optional;
            }
            /** False if one is optional and the other isn't, or one is a rest parameter and the other isn't. */
            function parametersHaveEqualSigils(a, b) {
                const optionalA = isTSParameterProperty(a)
                    ? a.parameter.optional
                    : a.optional;
                const optionalB = isTSParameterProperty(b)
                    ? b.parameter.optional
                    : b.optional;
                return ((a.type === experimental_utils_1.AST_NODE_TYPES.RestElement) ===
                    (b.type === experimental_utils_1.AST_NODE_TYPES.RestElement) &&
                    (optionalA !== undefined) === (optionalB !== undefined));
            }
            function typeParametersAreEqual(a, b) {
                return (a.name.name === b.name.name &&
                    constraintsAreEqual(a.constraint, b.constraint));
            }
            function typesAreEqual(a, b) {
                return (a === b ||
                    (a !== undefined &&
                        b !== undefined &&
                        sourceCode.getText(a.typeAnnotation) ===
                            sourceCode.getText(b.typeAnnotation)));
            }
            function constraintsAreEqual(a, b) {
                return (a === b || (a !== undefined && b !== undefined && a.type === b.type));
            }
            /* Returns the first index where `a` and `b` differ. */
            function getIndexOfFirstDifference(a, b, equal) {
                for (let i = 0; i < a.length && i < b.length; i++) {
                    if (!equal(a[i], b[i])) {
                        return i;
                    }
                }
                return undefined;
            }
            /** Calls `action` for every pair of values in `values`. */
            function forEachPair(values, action) {
                for (let i = 0; i < values.length; i++) {
                    for (let j = i + 1; j < values.length; j++) {
                        action(values[i], values[j]);
                    }
                }
            }
            const scopes = [];
            let currentScope = {
                overloads: new Map(),
            };
            function createScope(parent, typeParameters) {
                currentScope && scopes.push(currentScope);
                currentScope = {
                    overloads: new Map(),
                    parent,
                    typeParameters,
                };
            }
            function checkScope() {
                const failures = checkOverloads(Array.from(currentScope.overloads.values()), currentScope.typeParameters);
                addFailures(failures);
                currentScope = scopes.pop();
            }
            function addOverload(signature, key, containingNode) {
                key = key !== null && key !== void 0 ? key : getOverloadKey(signature);
                if (currentScope &&
                    (containingNode || signature).parent === currentScope.parent) {
                    const overloads = currentScope.overloads.get(key);
                    if (overloads !== undefined) {
                        overloads.push(signature);
                    }
                    else {
                        currentScope.overloads.set(key, [signature]);
                    }
                }
            }
            //----------------------------------------------------------------------
            // Public
            //----------------------------------------------------------------------
            return {
                Program: createScope,
                TSModuleBlock: createScope,
                TSInterfaceDeclaration(node) {
                    createScope(node.body, node.typeParameters);
                },
                ClassDeclaration(node) {
                    createScope(node.body, node.typeParameters);
                },
                TSTypeLiteral: createScope,
                // collect overloads
                TSDeclareFunction(node) {
                    var _a, _b;
                    const exportingNode = getExportingNode(node);
                    addOverload(node, (_b = (_a = node.id) === null || _a === void 0 ? void 0 : _a.name) !== null && _b !== void 0 ? _b : exportingNode === null || exportingNode === void 0 ? void 0 : exportingNode.type, exportingNode);
                },
                TSCallSignatureDeclaration: addOverload,
                TSConstructSignatureDeclaration: addOverload,
                TSMethodSignature: addOverload,
                TSAbstractMethodDefinition(node) {
                    if (!node.value.body) {
                        addOverload(node);
                    }
                },
                MethodDefinition(node) {
                    if (!node.value.body) {
                        addOverload(node);
                    }
                },
                // validate scopes
                'Program:exit': checkScope,
                'TSModuleBlock:exit': checkScope,
                'TSInterfaceDeclaration:exit': checkScope,
                'ClassDeclaration:exit': checkScope,
                'TSTypeLiteral:exit': checkScope,
            };
        },
    });
    function getExportingNode(node) {
        return node.parent &&
            (node.parent.type === experimental_utils_1.AST_NODE_TYPES.ExportNamedDeclaration ||
                node.parent.type === experimental_utils_1.AST_NODE_TYPES.ExportDefaultDeclaration)
            ? node.parent
            : undefined;
    }
    function getOverloadKey(node) {
        const info = getOverloadInfo(node);
        return ((node.computed ? '0' : '1') +
            (node.static ? '0' : '1') +
            info);
    }
    function getOverloadInfo(node) {
        switch (node.type) {
            case experimental_utils_1.AST_NODE_TYPES.TSConstructSignatureDeclaration:
                return 'constructor';
            case experimental_utils_1.AST_NODE_TYPES.TSCallSignatureDeclaration:
                return '()';
            default: {
                const { key } = node;
                return isIdentifier(key) ? key.name : key.raw;
            }
        }
    }
    function isIdentifier(node) {
        return node.type === experimental_utils_1.AST_NODE_TYPES.Identifier;
    }
},
/* @typescript-eslint/eslint-plugin/dist/configs/all.js */ function _(require, module, exports, __esModule, __esExport) {
    // THIS CODE WAS AUTOMATICALLY GENERATED
    // DO NOT EDIT THIS CODE BY HAND
    // YOU CAN REGENERATE IT USING yarn generate:configs
    module.exports = {
        extends: ['./configs/base', './configs/eslint-recommended'],
        rules: {
            '@typescript-eslint/adjacent-overload-signatures': 'error',
            '@typescript-eslint/array-type': 'error',
            '@typescript-eslint/await-thenable': 'error',
            '@typescript-eslint/ban-ts-comment': 'error',
            '@typescript-eslint/ban-tslint-comment': 'error',
            '@typescript-eslint/ban-types': 'error',
            'brace-style': 'off',
            '@typescript-eslint/brace-style': 'error',
            '@typescript-eslint/class-literal-property-style': 'error',
            'comma-dangle': 'off',
            '@typescript-eslint/comma-dangle': 'error',
            'comma-spacing': 'off',
            '@typescript-eslint/comma-spacing': 'error',
            '@typescript-eslint/consistent-indexed-object-style': 'error',
            '@typescript-eslint/consistent-type-assertions': 'error',
            '@typescript-eslint/consistent-type-definitions': 'error',
            '@typescript-eslint/consistent-type-imports': 'error',
            'default-param-last': 'off',
            '@typescript-eslint/default-param-last': 'error',
            'dot-notation': 'off',
            '@typescript-eslint/dot-notation': 'error',
            '@typescript-eslint/explicit-function-return-type': 'error',
            '@typescript-eslint/explicit-member-accessibility': 'error',
            '@typescript-eslint/explicit-module-boundary-types': 'error',
            'func-call-spacing': 'off',
            '@typescript-eslint/func-call-spacing': 'error',
            indent: 'off',
            '@typescript-eslint/indent': 'error',
            'init-declarations': 'off',
            '@typescript-eslint/init-declarations': 'error',
            'keyword-spacing': 'off',
            '@typescript-eslint/keyword-spacing': 'error',
            'lines-between-class-members': 'off',
            '@typescript-eslint/lines-between-class-members': 'error',
            '@typescript-eslint/member-delimiter-style': 'error',
            '@typescript-eslint/member-ordering': 'error',
            '@typescript-eslint/method-signature-style': 'error',
            '@typescript-eslint/naming-convention': 'error',
            'no-array-constructor': 'off',
            '@typescript-eslint/no-array-constructor': 'error',
            '@typescript-eslint/no-base-to-string': 'error',
            '@typescript-eslint/no-confusing-non-null-assertion': 'error',
            '@typescript-eslint/no-confusing-void-expression': 'error',
            'no-dupe-class-members': 'off',
            '@typescript-eslint/no-dupe-class-members': 'error',
            'no-duplicate-imports': 'off',
            '@typescript-eslint/no-duplicate-imports': 'error',
            '@typescript-eslint/no-dynamic-delete': 'error',
            'no-empty-function': 'off',
            '@typescript-eslint/no-empty-function': 'error',
            '@typescript-eslint/no-empty-interface': 'error',
            '@typescript-eslint/no-explicit-any': 'error',
            '@typescript-eslint/no-extra-non-null-assertion': 'error',
            'no-extra-parens': 'off',
            '@typescript-eslint/no-extra-parens': 'error',
            'no-extra-semi': 'off',
            '@typescript-eslint/no-extra-semi': 'error',
            '@typescript-eslint/no-extraneous-class': 'error',
            '@typescript-eslint/no-floating-promises': 'error',
            '@typescript-eslint/no-for-in-array': 'error',
            '@typescript-eslint/no-implicit-any-catch': 'error',
            'no-implied-eval': 'off',
            '@typescript-eslint/no-implied-eval': 'error',
            '@typescript-eslint/no-inferrable-types': 'error',
            'no-invalid-this': 'off',
            '@typescript-eslint/no-invalid-this': 'error',
            '@typescript-eslint/no-invalid-void-type': 'error',
            'no-loop-func': 'off',
            '@typescript-eslint/no-loop-func': 'error',
            'no-loss-of-precision': 'off',
            '@typescript-eslint/no-loss-of-precision': 'error',
            'no-magic-numbers': 'off',
            '@typescript-eslint/no-magic-numbers': 'error',
            '@typescript-eslint/no-misused-new': 'error',
            '@typescript-eslint/no-misused-promises': 'error',
            '@typescript-eslint/no-namespace': 'error',
            '@typescript-eslint/no-non-null-asserted-optional-chain': 'error',
            '@typescript-eslint/no-non-null-assertion': 'error',
            '@typescript-eslint/no-parameter-properties': 'error',
            'no-redeclare': 'off',
            '@typescript-eslint/no-redeclare': 'error',
            '@typescript-eslint/no-require-imports': 'error',
            'no-shadow': 'off',
            '@typescript-eslint/no-shadow': 'error',
            '@typescript-eslint/no-this-alias': 'error',
            'no-throw-literal': 'off',
            '@typescript-eslint/no-throw-literal': 'error',
            '@typescript-eslint/no-type-alias': 'error',
            '@typescript-eslint/no-unnecessary-boolean-literal-compare': 'error',
            '@typescript-eslint/no-unnecessary-condition': 'error',
            '@typescript-eslint/no-unnecessary-qualifier': 'error',
            '@typescript-eslint/no-unnecessary-type-arguments': 'error',
            '@typescript-eslint/no-unnecessary-type-assertion': 'error',
            '@typescript-eslint/no-unnecessary-type-constraint': 'error',
            '@typescript-eslint/no-unsafe-argument': 'error',
            '@typescript-eslint/no-unsafe-assignment': 'error',
            '@typescript-eslint/no-unsafe-call': 'error',
            '@typescript-eslint/no-unsafe-member-access': 'error',
            '@typescript-eslint/no-unsafe-return': 'error',
            'no-unused-expressions': 'off',
            '@typescript-eslint/no-unused-expressions': 'error',
            'no-unused-vars': 'off',
            '@typescript-eslint/no-unused-vars': 'error',
            'no-use-before-define': 'off',
            '@typescript-eslint/no-use-before-define': 'error',
            'no-useless-constructor': 'off',
            '@typescript-eslint/no-useless-constructor': 'error',
            '@typescript-eslint/no-var-requires': 'error',
            '@typescript-eslint/non-nullable-type-assertion-style': 'error',
            'object-curly-spacing': 'off',
            '@typescript-eslint/object-curly-spacing': 'error',
            '@typescript-eslint/prefer-as-const': 'error',
            '@typescript-eslint/prefer-enum-initializers': 'error',
            '@typescript-eslint/prefer-for-of': 'error',
            '@typescript-eslint/prefer-function-type': 'error',
            '@typescript-eslint/prefer-includes': 'error',
            '@typescript-eslint/prefer-literal-enum-member': 'error',
            '@typescript-eslint/prefer-namespace-keyword': 'error',
            '@typescript-eslint/prefer-nullish-coalescing': 'error',
            '@typescript-eslint/prefer-optional-chain': 'error',
            '@typescript-eslint/prefer-readonly': 'error',
            '@typescript-eslint/prefer-readonly-parameter-types': 'error',
            '@typescript-eslint/prefer-reduce-type-parameter': 'error',
            '@typescript-eslint/prefer-regexp-exec': 'error',
            '@typescript-eslint/prefer-return-this-type': 'error',
            '@typescript-eslint/prefer-string-starts-ends-with': 'error',
            '@typescript-eslint/prefer-ts-expect-error': 'error',
            '@typescript-eslint/promise-function-async': 'error',
            quotes: 'off',
            '@typescript-eslint/quotes': 'error',
            '@typescript-eslint/require-array-sort-compare': 'error',
            'require-await': 'off',
            '@typescript-eslint/require-await': 'error',
            '@typescript-eslint/restrict-plus-operands': 'error',
            '@typescript-eslint/restrict-template-expressions': 'error',
            'no-return-await': 'off',
            '@typescript-eslint/return-await': 'error',
            semi: 'off',
            '@typescript-eslint/semi': 'error',
            '@typescript-eslint/sort-type-union-intersection-members': 'error',
            'space-before-function-paren': 'off',
            '@typescript-eslint/space-before-function-paren': 'error',
            'space-infix-ops': 'off',
            '@typescript-eslint/space-infix-ops': 'error',
            '@typescript-eslint/strict-boolean-expressions': 'error',
            '@typescript-eslint/switch-exhaustiveness-check': 'error',
            '@typescript-eslint/triple-slash-reference': 'error',
            '@typescript-eslint/type-annotation-spacing': 'error',
            '@typescript-eslint/typedef': 'error',
            '@typescript-eslint/unbound-method': 'error',
            '@typescript-eslint/unified-signatures': 'error',
        },
    };
},
/* @typescript-eslint/eslint-plugin/dist/configs/base.js */ function _(require, module, exports, __esModule, __esExport) {
    // THIS CODE WAS AUTOMATICALLY GENERATED
    // DO NOT EDIT THIS CODE BY HAND
    // YOU CAN REGENERATE IT USING yarn generate:configs
    module.exports = {
        parser: '@typescript-eslint/parser',
        parserOptions: { sourceType: 'module' },
        plugins: ['@typescript-eslint'],
    };
},
/* @typescript-eslint/eslint-plugin/dist/configs/recommended.js */ function _(require, module, exports, __esModule, __esExport) {
    // THIS CODE WAS AUTOMATICALLY GENERATED
    // DO NOT EDIT THIS CODE BY HAND
    // YOU CAN REGENERATE IT USING yarn generate:configs
    module.exports = {
        extends: ['./configs/base', './configs/eslint-recommended'],
        rules: {
            '@typescript-eslint/adjacent-overload-signatures': 'error',
            '@typescript-eslint/ban-ts-comment': 'error',
            '@typescript-eslint/ban-types': 'error',
            '@typescript-eslint/explicit-module-boundary-types': 'warn',
            'no-array-constructor': 'off',
            '@typescript-eslint/no-array-constructor': 'error',
            'no-empty-function': 'off',
            '@typescript-eslint/no-empty-function': 'error',
            '@typescript-eslint/no-empty-interface': 'error',
            '@typescript-eslint/no-explicit-any': 'warn',
            '@typescript-eslint/no-extra-non-null-assertion': 'error',
            'no-extra-semi': 'off',
            '@typescript-eslint/no-extra-semi': 'error',
            '@typescript-eslint/no-inferrable-types': 'error',
            '@typescript-eslint/no-misused-new': 'error',
            '@typescript-eslint/no-namespace': 'error',
            '@typescript-eslint/no-non-null-asserted-optional-chain': 'error',
            '@typescript-eslint/no-non-null-assertion': 'warn',
            '@typescript-eslint/no-this-alias': 'error',
            'no-unused-vars': 'off',
            '@typescript-eslint/no-unused-vars': 'warn',
            '@typescript-eslint/no-var-requires': 'error',
            '@typescript-eslint/prefer-as-const': 'error',
            '@typescript-eslint/prefer-namespace-keyword': 'error',
            '@typescript-eslint/triple-slash-reference': 'error',
        },
    };
},
/* @typescript-eslint/eslint-plugin/dist/configs/recommended-requiring-type-checking.js */ function _(require, module, exports, __esModule, __esExport) {
    // THIS CODE WAS AUTOMATICALLY GENERATED
    // DO NOT EDIT THIS CODE BY HAND
    // YOU CAN REGENERATE IT USING yarn generate:configs
    module.exports = {
        extends: ['./configs/base', './configs/eslint-recommended'],
        rules: {
            '@typescript-eslint/await-thenable': 'error',
            '@typescript-eslint/no-floating-promises': 'error',
            '@typescript-eslint/no-for-in-array': 'error',
            'no-implied-eval': 'off',
            '@typescript-eslint/no-implied-eval': 'error',
            '@typescript-eslint/no-misused-promises': 'error',
            '@typescript-eslint/no-unnecessary-type-assertion': 'error',
            '@typescript-eslint/no-unsafe-assignment': 'error',
            '@typescript-eslint/no-unsafe-call': 'error',
            '@typescript-eslint/no-unsafe-member-access': 'error',
            '@typescript-eslint/no-unsafe-return': 'error',
            '@typescript-eslint/prefer-regexp-exec': 'error',
            'require-await': 'off',
            '@typescript-eslint/require-await': 'error',
            '@typescript-eslint/restrict-plus-operands': 'error',
            '@typescript-eslint/restrict-template-expressions': 'error',
            '@typescript-eslint/unbound-method': 'error',
        },
    };
},
/* @typescript-eslint/eslint-plugin/dist/configs/eslint-recommended.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        overrides: [
            {
                files: ['*.ts', '*.tsx'],
                rules: {
                    'constructor-super': 'off',
                    'getter-return': 'off',
                    'no-const-assign': 'off',
                    'no-dupe-args': 'off',
                    'no-dupe-class-members': 'off',
                    'no-dupe-keys': 'off',
                    'no-func-assign': 'off',
                    'no-import-assign': 'off',
                    'no-new-symbol': 'off',
                    'no-obj-calls': 'off',
                    'no-redeclare': 'off',
                    'no-setter-return': 'off',
                    'no-this-before-super': 'off',
                    'no-undef': 'off',
                    'no-unreachable': 'off',
                    'no-unsafe-negation': 'off',
                    'no-var': 'error',
                    'prefer-const': 'error',
                    'prefer-rest-params': 'error',
                    'prefer-spread': 'error',
                    'valid-typeof': 'off', // ts(2367)
                },
            },
        ],
    };
},
/* @typescript-eslint/parser/dist/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    var parser_1 = require(1118) /* ./parser */;
    __esExport("parse", parser_1.parse);
    __esExport("parseForESLint", parser_1.parseForESLint);
    var typescript_estree_1 = require(1119) /* @typescript-eslint/typescript-estree */;
    __esExport("clearCaches", typescript_estree_1.clearCaches);
    __esExport("createProgram", typescript_estree_1.createProgram);
    // note - cannot migrate this to an import statement because it will make TSC copy the package.json to the dist folder
    // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-member-access
    exports.version = require(1216) /* ../package.json */.version;
},
/* @typescript-eslint/parser/dist/parser.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const typescript_estree_1 = require(1119) /* @typescript-eslint/typescript-estree */;
    const scope_manager_1 = require(768) /* @typescript-eslint/scope-manager */;
    const debug_1 = __importDefault(require(72) /* debug */);
    const typescript_1 = require(21) /* typescript */;
    const log = debug_1.default('typescript-eslint:parser:parser');
    function validateBoolean(value, fallback = false) {
        if (typeof value !== 'boolean') {
            return fallback;
        }
        return value;
    }
    const LIB_FILENAME_REGEX = /lib\.(.+)\.d\.ts$/;
    function getLib(compilerOptions) {
        var _a;
        if (compilerOptions.lib) {
            return compilerOptions.lib.reduce((acc, lib) => {
                const match = LIB_FILENAME_REGEX.exec(lib.toLowerCase());
                if (match) {
                    acc.push(match[1]);
                }
                return acc;
            }, []);
        }
        const target = (_a = compilerOptions.target) !== null && _a !== void 0 ? _a : typescript_1.ScriptTarget.ES5;
        // https://github.com/Microsoft/TypeScript/blob/59ad375234dc2efe38d8ee0ba58414474c1d5169/src/compiler/utilitiesPublic.ts#L13-L32
        switch (target) {
            case typescript_1.ScriptTarget.ESNext:
                return ['esnext.full'];
            case typescript_1.ScriptTarget.ES2020:
                return ['es2020.full'];
            case typescript_1.ScriptTarget.ES2019:
                return ['es2019.full'];
            case typescript_1.ScriptTarget.ES2018:
                return ['es2018.full'];
            case typescript_1.ScriptTarget.ES2017:
                return ['es2017.full'];
            case typescript_1.ScriptTarget.ES2016:
                return ['es2016.full'];
            case typescript_1.ScriptTarget.ES2015:
                return ['es6'];
            default:
                return ['lib'];
        }
    }
    function parse(code, options) {
        return parseForESLint(code, options).ast;
    }
    exports.parse = parse;
    function parseForESLint(code, options) {
        if (!options || typeof options !== 'object') {
            options = {};
        }
        else {
            options = Object.assign({}, options);
        }
        // https://eslint.org/docs/user-guide/configuring#specifying-parser-options
        // if sourceType is not provided by default eslint expect that it will be set to "script"
        if (options.sourceType !== 'module' && options.sourceType !== 'script') {
            options.sourceType = 'script';
        }
        if (typeof options.ecmaFeatures !== 'object') {
            options.ecmaFeatures = {};
        }
        const parserOptions = {};
        Object.assign(parserOptions, options, {
            useJSXTextNode: validateBoolean(options.useJSXTextNode, true),
            jsx: validateBoolean(options.ecmaFeatures.jsx),
        });
        const analyzeOptions = {
            ecmaVersion: options.ecmaVersion,
            globalReturn: options.ecmaFeatures.globalReturn,
            jsxPragma: options.jsxPragma,
            jsxFragmentName: options.jsxFragmentName,
            lib: options.lib,
            sourceType: options.sourceType,
        };
        if (typeof options.filePath === 'string') {
            const tsx = options.filePath.endsWith('.tsx');
            if (tsx || options.filePath.endsWith('.ts')) {
                parserOptions.jsx = tsx;
            }
        }
        /**
         * Allow the user to suppress the warning from typescript-estree if they are using an unsupported
         * version of TypeScript
         */
        const warnOnUnsupportedTypeScriptVersion = validateBoolean(options.warnOnUnsupportedTypeScriptVersion, true);
        if (!warnOnUnsupportedTypeScriptVersion) {
            parserOptions.loggerFn = false;
        }
        const { ast, services } = typescript_estree_1.parseAndGenerateServices(code, parserOptions);
        ast.sourceType = options.sourceType;
        if (services.hasFullTypeInformation) {
            // automatically apply the options configured for the program
            const compilerOptions = services.program.getCompilerOptions();
            if (analyzeOptions.lib == null) {
                analyzeOptions.lib = getLib(compilerOptions);
                log('Resolved libs from program: %o', analyzeOptions.lib);
            }
            if (parserOptions.jsx === true) {
                if (analyzeOptions.jsxPragma === undefined &&
                    compilerOptions.jsxFactory != null) {
                    // in case the user has specified something like "preact.h"
                    const factory = compilerOptions.jsxFactory.split('.')[0].trim();
                    analyzeOptions.jsxPragma = factory;
                    log('Resolved jsxPragma from program: %s', analyzeOptions.jsxPragma);
                }
                if (analyzeOptions.jsxFragmentName === undefined &&
                    compilerOptions.jsxFragmentFactory != null) {
                    // in case the user has specified something like "preact.Fragment"
                    const fragFactory = compilerOptions.jsxFragmentFactory
                        .split('.')[0]
                        .trim();
                    analyzeOptions.jsxFragmentName = fragFactory;
                    log('Resolved jsxFragmentName from program: %s', analyzeOptions.jsxFragmentName);
                }
            }
            if (compilerOptions.emitDecoratorMetadata === true) {
                analyzeOptions.emitDecoratorMetadata =
                    compilerOptions.emitDecoratorMetadata;
            }
        }
        const scopeManager = scope_manager_1.analyze(ast, analyzeOptions);
        return { ast, services, scopeManager, visitorKeys: typescript_estree_1.visitorKeys };
    }
    exports.parseForESLint = parseForESLint;
},
/* @typescript-eslint/typescript-estree/dist/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    var parser_1 = require(1120) /* ./parser */;
    __esExport("parse", parser_1.parse);
    __esExport("parseAndGenerateServices", parser_1.parseAndGenerateServices);
    __esExport("parseWithNodeMaps", parser_1.parseWithNodeMaps);
    __esExport("clearProgramCache", parser_1.clearProgramCache);
    var simple_traverse_1 = require(1206) /* ./simple-traverse */;
    __esExport("simpleTraverse", simple_traverse_1.simpleTraverse);
    __exportStar(require(1200) /* ./ts-estree */, exports);
    var createWatchProgram_1 = require(1211) /* ./create-program/createWatchProgram */;
    __esExport("clearCaches", createWatchProgram_1.clearWatchCaches);
    var useProvidedPrograms_1 = require(1214) /* ./create-program/useProvidedPrograms */;
    __esExport("createProgram", useProvidedPrograms_1.createProgramFromConfigFile);
    // re-export for backwards-compat
    var visitor_keys_1 = require(770) /* @typescript-eslint/visitor-keys */;
    __esExport("visitorKeys", visitor_keys_1.visitorKeys);
    // note - cannot migrate this to an import statement because it will make TSC copy the package.json to the dist folder
    // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-member-access
    exports.version = require(1215) /* ../package.json */.version;
},
/* @typescript-eslint/typescript-estree/dist/parser.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const debug_1 = __importDefault(require(72) /* debug */);
    const globby_1 = require(1121) /* globby */;
    const is_glob_1 = __importDefault(require(327) /* is-glob */);
    const semver_1 = __importDefault(require(1004) /* semver */);
    const path_1 = require("path");
    const ts = __importStar(require(21) /* typescript */);
    const ast_converter_1 = require(1197) /* ./ast-converter */;
    const convert_1 = require(1198) /* ./convert */;
    const createDefaultProgram_1 = require(1207) /* ./create-program/createDefaultProgram */;
    const createIsolatedProgram_1 = require(1209) /* ./create-program/createIsolatedProgram */;
    const createProjectProgram_1 = require(1210) /* ./create-program/createProjectProgram */;
    const createSourceFile_1 = require(1212) /* ./create-program/createSourceFile */;
    const semantic_or_syntactic_errors_1 = require(1213) /* ./semantic-or-syntactic-errors */;
    const shared_1 = require(1208) /* ./create-program/shared */;
    const useProvidedPrograms_1 = require(1214) /* ./create-program/useProvidedPrograms */;
    const log = debug_1.default('typescript-eslint:typescript-estree:parser');
    /**
     * This needs to be kept in sync with the top-level README.md in the
     * typescript-eslint monorepo
     */
    const SUPPORTED_TYPESCRIPT_VERSIONS = '>=3.3.1 <4.4.0';
    /*
     * The semver package will ignore prerelease ranges, and we don't want to explicitly document every one
     * List them all separately here, so we can automatically create the full string
     */
    const SUPPORTED_PRERELEASE_RANGES = ['4.3.0-beta', '4.3.1-rc'];
    const ACTIVE_TYPESCRIPT_VERSION = ts.version;
    const isRunningSupportedTypeScriptVersion = semver_1.default.satisfies(ACTIVE_TYPESCRIPT_VERSION, [SUPPORTED_TYPESCRIPT_VERSIONS]
        .concat(SUPPORTED_PRERELEASE_RANGES)
        .join(' || '));
    let extra;
    let warnedAboutTSVersion = false;
    /**
     * Cache existing programs for the single run use-case.
     *
     * clearProgramCache() is only intended to be used in testing to ensure the parser is clean between tests.
     */
    const existingPrograms = new Map();
    function clearProgramCache() {
        existingPrograms.clear();
    }
    exports.clearProgramCache = clearProgramCache;
    function enforceString(code) {
        /**
         * Ensure the source code is a string
         */
        if (typeof code !== 'string') {
            return String(code);
        }
        return code;
    }
    /**
     * @param code The code of the file being linted
     * @param programInstances One or more (potentially lazily constructed) existing programs to use
     * @param shouldProvideParserServices True if the program should be attempted to be calculated from provided tsconfig files
     * @param shouldCreateDefaultProgram True if the program should be created from compiler host
     * @returns Returns a source file and program corresponding to the linted code
     */
    function getProgramAndAST(code, programInstances, shouldProvideParserServices, shouldCreateDefaultProgram) {
        return ((programInstances && useProvidedPrograms_1.useProvidedPrograms(programInstances, extra)) ||
            (shouldProvideParserServices &&
                createProjectProgram_1.createProjectProgram(code, shouldCreateDefaultProgram, extra)) ||
            (shouldProvideParserServices &&
                shouldCreateDefaultProgram &&
                createDefaultProgram_1.createDefaultProgram(code, extra)) ||
            createIsolatedProgram_1.createIsolatedProgram(code, extra));
    }
    /**
     * Compute the filename based on the parser options.
     *
     * Even if jsx option is set in typescript compiler, filename still has to
     * contain .tsx file extension.
     *
     * @param options Parser options
     */
    function getFileName({ jsx } = {}) {
        return jsx ? 'estree.tsx' : 'estree.ts';
    }
    /**
     * Resets the extra config object
     */
    function resetExtra() {
        extra = {
            code: '',
            comment: false,
            comments: [],
            createDefaultProgram: false,
            debugLevel: new Set(),
            errorOnTypeScriptSyntacticAndSemanticIssues: false,
            errorOnUnknownASTType: false,
            EXPERIMENTAL_useSourceOfProjectReferenceRedirect: false,
            extraFileExtensions: [],
            filePath: getFileName(),
            jsx: false,
            loc: false,
            log: console.log,
            preserveNodeMaps: true,
            programs: null,
            projects: [],
            range: false,
            strict: false,
            tokens: null,
            tsconfigRootDir: process.cwd(),
            useJSXTextNode: false,
            /**
             * Unless we can reliably infer otherwise, we default to assuming that this run could be part
             * of a long-running session (e.g. in an IDE) and watch programs will therefore be required
             */
            singleRun: false,
            moduleResolver: '',
        };
    }
    function getTsconfigPath(tsconfigPath, extra) {
        return shared_1.getCanonicalFileName(shared_1.ensureAbsolutePath(tsconfigPath, extra));
    }
    /**
     * Normalizes, sanitizes, resolves and filters the provided project paths
     */
    function prepareAndTransformProjects(projectsInput, ignoreListInput) {
        const sanitizedProjects = [];
        // Normalize and sanitize the project paths
        if (typeof projectsInput === 'string') {
            sanitizedProjects.push(projectsInput);
        }
        else if (Array.isArray(projectsInput)) {
            for (const project of projectsInput) {
                if (typeof project === 'string') {
                    sanitizedProjects.push(project);
                }
            }
        }
        if (sanitizedProjects.length === 0) {
            return [];
        }
        // Transform glob patterns into paths
        const nonGlobProjects = sanitizedProjects.filter(project => !is_glob_1.default(project));
        const globProjects = sanitizedProjects.filter(project => is_glob_1.default(project));
        const uniqueCanonicalProjectPaths = new Set(nonGlobProjects
            .concat(globby_1.sync([...globProjects, ...ignoreListInput], {
            cwd: extra.tsconfigRootDir,
        }))
            .map(project => getTsconfigPath(project, extra)));
        log('parserOptions.project (excluding ignored) matched projects: %s', uniqueCanonicalProjectPaths);
        return Array.from(uniqueCanonicalProjectPaths);
    }
    function applyParserOptionsToExtra(options) {
        var _a;
        /**
         * Configure Debug logging
         */
        if (options.debugLevel === true) {
            extra.debugLevel = new Set(['typescript-eslint']);
        }
        else if (Array.isArray(options.debugLevel)) {
            extra.debugLevel = new Set(options.debugLevel);
        }
        if (extra.debugLevel.size > 0) {
            // debug doesn't support multiple `enable` calls, so have to do it all at once
            const namespaces = [];
            if (extra.debugLevel.has('typescript-eslint')) {
                namespaces.push('typescript-eslint:*');
            }
            if (extra.debugLevel.has('eslint') ||
                // make sure we don't turn off the eslint debug if it was enabled via --debug
                debug_1.default.enabled('eslint:*,-eslint:code-path')) {
                // https://github.com/eslint/eslint/blob/9dfc8501fb1956c90dc11e6377b4cb38a6bea65d/bin/eslint.js#L25
                namespaces.push('eslint:*,-eslint:code-path');
            }
            debug_1.default.enable(namespaces.join(','));
        }
        /**
         * Track range information in the AST
         */
        extra.range = typeof options.range === 'boolean' && options.range;
        extra.loc = typeof options.loc === 'boolean' && options.loc;
        /**
         * Track tokens in the AST
         */
        if (typeof options.tokens === 'boolean' && options.tokens) {
            extra.tokens = [];
        }
        /**
         * Track comments in the AST
         */
        if (typeof options.comment === 'boolean' && options.comment) {
            extra.comment = true;
            extra.comments = [];
        }
        /**
         * Enable JSX - note the applicable file extension is still required
         */
        if (typeof options.jsx === 'boolean' && options.jsx) {
            extra.jsx = true;
        }
        /**
         * Get the file path
         */
        if (typeof options.filePath === 'string' && options.filePath !== '<input>') {
            extra.filePath = options.filePath;
        }
        else {
            extra.filePath = getFileName(extra);
        }
        /**
         * The JSX AST changed the node type for string literals
         * inside a JSX Element from `Literal` to `JSXText`.
         *
         * When value is `true`, these nodes will be parsed as type `JSXText`.
         * When value is `false`, these nodes will be parsed as type `Literal`.
         */
        if (typeof options.useJSXTextNode === 'boolean' && options.useJSXTextNode) {
            extra.useJSXTextNode = true;
        }
        /**
         * Allow the user to cause the parser to error if it encounters an unknown AST Node Type
         * (used in testing)
         */
        if (typeof options.errorOnUnknownASTType === 'boolean' &&
            options.errorOnUnknownASTType) {
            extra.errorOnUnknownASTType = true;
        }
        /**
         * Allow the user to override the function used for logging
         */
        if (typeof options.loggerFn === 'function') {
            extra.log = options.loggerFn;
        }
        else if (options.loggerFn === false) {
            extra.log = () => { };
        }
        if (typeof options.tsconfigRootDir === 'string') {
            extra.tsconfigRootDir = options.tsconfigRootDir;
        }
        // NOTE - ensureAbsolutePath relies upon having the correct tsconfigRootDir in extra
        extra.filePath = shared_1.ensureAbsolutePath(extra.filePath, extra);
        if (Array.isArray(options.programs)) {
            if (!options.programs.length) {
                throw new Error(`You have set parserOptions.programs to an empty array. This will cause all files to not be found in existing programs. Either provide one or more existing TypeScript Program instances in the array, or remove the parserOptions.programs setting.`);
            }
            extra.programs = options.programs;
            log('parserOptions.programs was provided, so parserOptions.project will be ignored.');
        }
        if (!extra.programs) {
            // providing a program overrides project resolution
            const projectFolderIgnoreList = ((_a = options.projectFolderIgnoreList) !== null && _a !== void 0 ? _a : ['**/node_modules/**'])
                .reduce((acc, folder) => {
                if (typeof folder === 'string') {
                    acc.push(folder);
                }
                return acc;
            }, [])
                // prefix with a ! for not match glob
                .map(folder => (folder.startsWith('!') ? folder : `!${folder}`));
            // NOTE - prepareAndTransformProjects relies upon having the correct tsconfigRootDir in extra
            extra.projects = prepareAndTransformProjects(options.project, projectFolderIgnoreList);
        }
        if (Array.isArray(options.extraFileExtensions) &&
            options.extraFileExtensions.every(ext => typeof ext === 'string')) {
            extra.extraFileExtensions = options.extraFileExtensions;
        }
        /**
         * Allow the user to enable or disable the preservation of the AST node maps
         * during the conversion process.
         */
        if (typeof options.preserveNodeMaps === 'boolean') {
            extra.preserveNodeMaps = options.preserveNodeMaps;
        }
        extra.createDefaultProgram =
            typeof options.createDefaultProgram === 'boolean' &&
                options.createDefaultProgram;
        extra.EXPERIMENTAL_useSourceOfProjectReferenceRedirect =
            typeof options.EXPERIMENTAL_useSourceOfProjectReferenceRedirect ===
                'boolean' && options.EXPERIMENTAL_useSourceOfProjectReferenceRedirect;
        if (typeof options.moduleResolver === 'string') {
            extra.moduleResolver = options.moduleResolver;
        }
    }
    function warnAboutTSVersion() {
        var _a;
        if (!isRunningSupportedTypeScriptVersion && !warnedAboutTSVersion) {
            const isTTY = typeof process === 'undefined' ? false : (_a = process.stdout) === null || _a === void 0 ? void 0 : _a.isTTY;
            if (isTTY) {
                const border = '=============';
                const versionWarning = [
                    border,
                    'WARNING: You are currently running a version of TypeScript which is not officially supported by @typescript-eslint/typescript-estree.',
                    'You may find that it works just fine, or you may not.',
                    `SUPPORTED TYPESCRIPT VERSIONS: ${SUPPORTED_TYPESCRIPT_VERSIONS}`,
                    `YOUR TYPESCRIPT VERSION: ${ACTIVE_TYPESCRIPT_VERSION}`,
                    'Please only submit bug reports when using the officially supported version.',
                    border,
                ];
                extra.log(versionWarning.join('\n\n'));
            }
            warnedAboutTSVersion = true;
        }
    }
    /**
     * ESLint (and therefore typescript-eslint) is used in both "single run"/one-time contexts,
     * such as an ESLint CLI invocation, and long-running sessions (such as continuous feedback
     * on a file in an IDE).
     *
     * When typescript-eslint handles TypeScript Program management behind the scenes, this distinction
     * is important because there is significant overhead to managing the so called Watch Programs
     * needed for the long-running use-case. We therefore use the following logic to figure out which
     * of these contexts applies to the current execution.
     */
    function inferSingleRun(options) {
        // Allow users to explicitly inform us of their intent to perform a single run (or not) with TSESTREE_SINGLE_RUN
        if (process.env.TSESTREE_SINGLE_RUN === 'false') {
            extra.singleRun = false;
            return;
        }
        if (process.env.TSESTREE_SINGLE_RUN === 'true') {
            extra.singleRun = true;
            return;
        }
        // Currently behind a flag while we gather real-world feedback
        if (options === null || options === void 0 ? void 0 : options.allowAutomaticSingleRunInference) {
            if (
            // Default to single runs for CI processes. CI=true is set by most CI providers by default.
            process.env.CI === 'true' ||
                // This will be true for invocations such as `npx eslint ...` and `./node_modules/.bin/eslint ...`
                process.argv[1].endsWith(path_1.normalize('node_modules/.bin/eslint'))) {
                extra.singleRun = true;
                return;
            }
        }
        /**
         * We default to assuming that this run could be part of a long-running session (e.g. in an IDE)
         * and watch programs will therefore be required
         */
        extra.singleRun = false;
    }
    function parse(code, options) {
        const { ast } = parseWithNodeMapsInternal(code, options, false);
        return ast;
    }
    exports.parse = parse;
    function parseWithNodeMapsInternal(code, options, shouldPreserveNodeMaps) {
        /**
         * Reset the parse configuration
         */
        resetExtra();
        /**
         * Ensure users do not attempt to use parse() when they need parseAndGenerateServices()
         */
        if (options === null || options === void 0 ? void 0 : options.errorOnTypeScriptSyntacticAndSemanticIssues) {
            throw new Error(`"errorOnTypeScriptSyntacticAndSemanticIssues" is only supported for parseAndGenerateServices()`);
        }
        /**
         * Ensure the source code is a string, and store a reference to it
         */
        code = enforceString(code);
        extra.code = code;
        /**
         * Apply the given parser options
         */
        if (typeof options !== 'undefined') {
            applyParserOptionsToExtra(options);
        }
        /**
         * Warn if the user is using an unsupported version of TypeScript
         */
        warnAboutTSVersion();
        /**
         * Figure out whether this is a single run or part of a long-running process
         */
        inferSingleRun(options);
        /**
         * Create a ts.SourceFile directly, no ts.Program is needed for a simple
         * parse
         */
        const ast = createSourceFile_1.createSourceFile(code, extra);
        /**
         * Convert the TypeScript AST to an ESTree-compatible one
         */
        const { estree, astMaps } = ast_converter_1.astConverter(ast, extra, shouldPreserveNodeMaps);
        return {
            ast: estree,
            esTreeNodeToTSNodeMap: astMaps.esTreeNodeToTSNodeMap,
            tsNodeToESTreeNodeMap: astMaps.tsNodeToESTreeNodeMap,
        };
    }
    function parseWithNodeMaps(code, options) {
        return parseWithNodeMapsInternal(code, options, true);
    }
    exports.parseWithNodeMaps = parseWithNodeMaps;
    let parseAndGenerateServicesCalls = {};
    // Privately exported utility intended for use in typescript-eslint unit tests only
    function clearParseAndGenerateServicesCalls() {
        parseAndGenerateServicesCalls = {};
    }
    exports.clearParseAndGenerateServicesCalls = clearParseAndGenerateServicesCalls;
    function parseAndGenerateServices(code, options) {
        var _a;
        /**
         * Reset the parse configuration
         */
        resetExtra();
        /**
         * Ensure the source code is a string, and store a reference to it
         */
        code = enforceString(code);
        extra.code = code;
        /**
         * Apply the given parser options
         */
        if (typeof options !== 'undefined') {
            applyParserOptionsToExtra(options);
            if (typeof options.errorOnTypeScriptSyntacticAndSemanticIssues ===
                'boolean' &&
                options.errorOnTypeScriptSyntacticAndSemanticIssues) {
                extra.errorOnTypeScriptSyntacticAndSemanticIssues = true;
            }
        }
        /**
         * Warn if the user is using an unsupported version of TypeScript
         */
        warnAboutTSVersion();
        /**
         * Figure out whether this is a single run or part of a long-running process
         */
        inferSingleRun(options);
        /**
         * If this is a single run in which the user has not provided any existing programs but there
         * are programs which need to be created from the provided "project" option,
         * create an Iterable which will lazily create the programs as needed by the iteration logic
         */
        if (extra.singleRun && !extra.programs && ((_a = extra.projects) === null || _a === void 0 ? void 0 : _a.length) > 0) {
            extra.programs = {
                *[Symbol.iterator]() {
                    for (const configFile of extra.projects) {
                        const existingProgram = existingPrograms.get(configFile);
                        if (existingProgram) {
                            yield existingProgram;
                        }
                        else {
                            log('Detected single-run/CLI usage, creating Program once ahead of time for project: %s', configFile);
                            const newProgram = useProvidedPrograms_1.createProgramFromConfigFile(configFile);
                            existingPrograms.set(configFile, newProgram);
                            yield newProgram;
                        }
                    }
                },
            };
        }
        /**
         * Generate a full ts.Program or offer provided instances in order to be able to provide parser services, such as type-checking
         */
        const shouldProvideParserServices = extra.programs != null || (extra.projects && extra.projects.length > 0);
        /**
         * If we are in singleRun mode but the parseAndGenerateServices() function has been called more than once for the current file,
         * it must mean that we are in the middle of an ESLint automated fix cycle (in which parsing can be performed up to an additional
         * 10 times in order to apply all possible fixes for the file).
         *
         * In this scenario we cannot rely upon the singleRun AOT compiled programs because the SourceFiles will not contain the source
         * with the latest fixes applied. Therefore we fallback to creating the quickest possible isolated program from the updated source.
         */
        let ast;
        let program;
        if (extra.singleRun && options.filePath) {
            parseAndGenerateServicesCalls[options.filePath] =
                (parseAndGenerateServicesCalls[options.filePath] || 0) + 1;
        }
        if (extra.singleRun &&
            options.filePath &&
            parseAndGenerateServicesCalls[options.filePath] > 1) {
            const isolatedAstAndProgram = createIsolatedProgram_1.createIsolatedProgram(code, extra);
            ast = isolatedAstAndProgram.ast;
            program = isolatedAstAndProgram.program;
        }
        else {
            const astAndProgram = getProgramAndAST(code, extra.programs, shouldProvideParserServices, extra.createDefaultProgram);
            ast = astAndProgram.ast;
            program = astAndProgram.program;
        }
        /**
         * Convert the TypeScript AST to an ESTree-compatible one, and optionally preserve
         * mappings between converted and original AST nodes
         */
        const preserveNodeMaps = typeof extra.preserveNodeMaps === 'boolean' ? extra.preserveNodeMaps : true;
        const { estree, astMaps } = ast_converter_1.astConverter(ast, extra, preserveNodeMaps);
        /**
         * Even if TypeScript parsed the source code ok, and we had no problems converting the AST,
         * there may be other syntactic or semantic issues in the code that we can optionally report on.
         */
        if (program && extra.errorOnTypeScriptSyntacticAndSemanticIssues) {
            const error = semantic_or_syntactic_errors_1.getFirstSemanticOrSyntacticError(program, ast);
            if (error) {
                throw convert_1.convertError(error);
            }
        }
        /**
         * Return the converted AST and additional parser services
         */
        return {
            ast: estree,
            services: {
                hasFullTypeInformation: shouldProvideParserServices,
                program,
                esTreeNodeToTSNodeMap: astMaps.esTreeNodeToTSNodeMap,
                tsNodeToESTreeNodeMap: astMaps.tsNodeToESTreeNodeMap,
            },
        };
    }
    exports.parseAndGenerateServices = parseAndGenerateServices;
},
/* globby/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const fs = require('fs');
    const arrayUnion = require(1122) /* array-union */;
    const merge2 = require(1123) /* merge2 */;
    const fastGlob = require(1124) /* fast-glob */;
    const dirGlob = require(1192) /* dir-glob */;
    const gitignore = require(1194) /* ./gitignore */;
    const { FilterStream, UniqueStream } = require(1196) /* ./stream-utils */;
    const DEFAULT_FILTER = () => false;
    const isNegative = pattern => pattern[0] === '!';
    const assertPatternsInput = patterns => {
        if (!patterns.every(pattern => typeof pattern === 'string')) {
            throw new TypeError('Patterns must be a string or an array of strings');
        }
    };
    const checkCwdOption = (options = {}) => {
        if (!options.cwd) {
            return;
        }
        let stat;
        try {
            stat = fs.statSync(options.cwd);
        }
        catch {
            return;
        }
        if (!stat.isDirectory()) {
            throw new Error('The `cwd` option must be a path to a directory');
        }
    };
    const getPathString = p => p.stats instanceof fs.Stats ? p.path : p;
    const generateGlobTasks = (patterns, taskOptions) => {
        patterns = arrayUnion([].concat(patterns));
        assertPatternsInput(patterns);
        checkCwdOption(taskOptions);
        const globTasks = [];
        taskOptions = {
            ignore: [],
            expandDirectories: true,
            ...taskOptions
        };
        for (const [index, pattern] of patterns.entries()) {
            if (isNegative(pattern)) {
                continue;
            }
            const ignore = patterns
                .slice(index)
                .filter(pattern => isNegative(pattern))
                .map(pattern => pattern.slice(1));
            const options = {
                ...taskOptions,
                ignore: taskOptions.ignore.concat(ignore)
            };
            globTasks.push({ pattern, options });
        }
        return globTasks;
    };
    const globDirs = (task, fn) => {
        let options = {};
        if (task.options.cwd) {
            options.cwd = task.options.cwd;
        }
        if (Array.isArray(task.options.expandDirectories)) {
            options = {
                ...options,
                files: task.options.expandDirectories
            };
        }
        else if (typeof task.options.expandDirectories === 'object') {
            options = {
                ...options,
                ...task.options.expandDirectories
            };
        }
        return fn(task.pattern, options);
    };
    const getPattern = (task, fn) => task.options.expandDirectories ? globDirs(task, fn) : [task.pattern];
    const getFilterSync = options => {
        return options && options.gitignore ?
            gitignore.sync({ cwd: options.cwd, ignore: options.ignore }) :
            DEFAULT_FILTER;
    };
    const globToTask = task => glob => {
        const { options } = task;
        if (options.ignore && Array.isArray(options.ignore) && options.expandDirectories) {
            options.ignore = dirGlob.sync(options.ignore);
        }
        return {
            pattern: glob,
            options
        };
    };
    module.exports = async (patterns, options) => {
        const globTasks = generateGlobTasks(patterns, options);
        const getFilter = async () => {
            return options && options.gitignore ?
                gitignore({ cwd: options.cwd, ignore: options.ignore }) :
                DEFAULT_FILTER;
        };
        const getTasks = async () => {
            const tasks = await Promise.all(globTasks.map(async (task) => {
                const globs = await getPattern(task, dirGlob);
                return Promise.all(globs.map(globToTask(task)));
            }));
            return arrayUnion(...tasks);
        };
        const [filter, tasks] = await Promise.all([getFilter(), getTasks()]);
        const paths = await Promise.all(tasks.map(task => fastGlob(task.pattern, task.options)));
        return arrayUnion(...paths).filter(path_ => !filter(getPathString(path_)));
    };
    module.exports.sync = (patterns, options) => {
        const globTasks = generateGlobTasks(patterns, options);
        const tasks = [];
        for (const task of globTasks) {
            const newTask = getPattern(task, dirGlob.sync).map(globToTask(task));
            tasks.push(...newTask);
        }
        const filter = getFilterSync(options);
        let matches = [];
        for (const task of tasks) {
            matches = arrayUnion(matches, fastGlob.sync(task.pattern, task.options));
        }
        return matches.filter(path_ => !filter(path_));
    };
    module.exports.stream = (patterns, options) => {
        const globTasks = generateGlobTasks(patterns, options);
        const tasks = [];
        for (const task of globTasks) {
            const newTask = getPattern(task, dirGlob.sync).map(globToTask(task));
            tasks.push(...newTask);
        }
        const filter = getFilterSync(options);
        const filterStream = new FilterStream(p => !filter(p));
        const uniqueStream = new UniqueStream();
        return merge2(tasks.map(task => fastGlob.stream(task.pattern, task.options)))
            .pipe(filterStream)
            .pipe(uniqueStream);
    };
    module.exports.generateGlobTasks = generateGlobTasks;
    module.exports.hasMagic = (patterns, options) => []
        .concat(patterns)
        .some(pattern => fastGlob.isDynamicPattern(pattern, options));
    module.exports.gitignore = gitignore;
},
/* array-union/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = (...arguments_) => {
        return [...new Set([].concat(...arguments_))];
    };
},
/* merge2/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /*
     * merge2
     * https://github.com/teambition/merge2
     *
     * Copyright (c) 2014-2020 Teambition
     * Licensed under the MIT license.
     */
    const Stream = require('stream');
    const PassThrough = Stream.PassThrough;
    const slice = Array.prototype.slice;
    module.exports = merge2;
    function merge2() {
        const streamsQueue = [];
        const args = slice.call(arguments);
        let merging = false;
        let options = args[args.length - 1];
        if (options && !Array.isArray(options) && options.pipe == null) {
            args.pop();
        }
        else {
            options = {};
        }
        const doEnd = options.end !== false;
        const doPipeError = options.pipeError === true;
        if (options.objectMode == null) {
            options.objectMode = true;
        }
        if (options.highWaterMark == null) {
            options.highWaterMark = 64 * 1024;
        }
        const mergedStream = PassThrough(options);
        function addStream() {
            for (let i = 0, len = arguments.length; i < len; i++) {
                streamsQueue.push(pauseStreams(arguments[i], options));
            }
            mergeStream();
            return this;
        }
        function mergeStream() {
            if (merging) {
                return;
            }
            merging = true;
            let streams = streamsQueue.shift();
            if (!streams) {
                process.nextTick(endStream);
                return;
            }
            if (!Array.isArray(streams)) {
                streams = [streams];
            }
            let pipesCount = streams.length + 1;
            function next() {
                if (--pipesCount > 0) {
                    return;
                }
                merging = false;
                mergeStream();
            }
            function pipe(stream) {
                function onend() {
                    stream.removeListener('merge2UnpipeEnd', onend);
                    stream.removeListener('end', onend);
                    if (doPipeError) {
                        stream.removeListener('error', onerror);
                    }
                    next();
                }
                function onerror(err) {
                    mergedStream.emit('error', err);
                }
                // skip ended stream
                if (stream._readableState.endEmitted) {
                    return next();
                }
                stream.on('merge2UnpipeEnd', onend);
                stream.on('end', onend);
                if (doPipeError) {
                    stream.on('error', onerror);
                }
                stream.pipe(mergedStream, { end: false });
                // compatible for old stream
                stream.resume();
            }
            for (let i = 0; i < streams.length; i++) {
                pipe(streams[i]);
            }
            next();
        }
        function endStream() {
            merging = false;
            // emit 'queueDrain' when all streams merged.
            mergedStream.emit('queueDrain');
            if (doEnd) {
                mergedStream.end();
            }
        }
        mergedStream.setMaxListeners(0);
        mergedStream.add = addStream;
        mergedStream.on('unpipe', function (stream) {
            stream.emit('merge2UnpipeEnd');
        });
        if (args.length) {
            addStream.apply(null, args);
        }
        return mergedStream;
    }
    // check and pause streams for pipe.
    function pauseStreams(streams, options) {
        if (!Array.isArray(streams)) {
            // Backwards-compat with old-style streams
            if (!streams._readableState && streams.pipe) {
                streams = streams.pipe(PassThrough(options));
            }
            if (!streams._readableState || !streams.pause || !streams.pipe) {
                throw new Error('Only readable stream can be merged.');
            }
            streams.pause();
        }
        else {
            for (let i = 0, len = streams.length; i < len; i++) {
                streams[i] = pauseStreams(streams[i], options);
            }
        }
        return streams;
    }
},
/* fast-glob/out/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const taskManager = require(1125) /* ./managers/tasks */;
    const async_1 = require(1151) /* ./providers/async */;
    const stream_1 = require(1188) /* ./providers/stream */;
    const sync_1 = require(1189) /* ./providers/sync */;
    const settings_1 = require(1191) /* ./settings */;
    const utils = require(1126) /* ./utils */;
    async function FastGlob(source, options) {
        assertPatternsInput(source);
        const works = getWorks(source, async_1.default, options);
        const result = await Promise.all(works);
        return utils.array.flatten(result);
    }
    // https://github.com/typescript-eslint/typescript-eslint/issues/60
    // eslint-disable-next-line no-redeclare
    (function (FastGlob) {
        function sync(source, options) {
            assertPatternsInput(source);
            const works = getWorks(source, sync_1.default, options);
            return utils.array.flatten(works);
        }
        FastGlob.sync = sync;
        function stream(source, options) {
            assertPatternsInput(source);
            const works = getWorks(source, stream_1.default, options);
            /**
             * The stream returned by the provider cannot work with an asynchronous iterator.
             * To support asynchronous iterators, regardless of the number of tasks, we always multiplex streams.
             * This affects performance (+25%). I don't see best solution right now.
             */
            return utils.stream.merge(works);
        }
        FastGlob.stream = stream;
        function generateTasks(source, options) {
            assertPatternsInput(source);
            const patterns = [].concat(source);
            const settings = new settings_1.default(options);
            return taskManager.generate(patterns, settings);
        }
        FastGlob.generateTasks = generateTasks;
        function isDynamicPattern(source, options) {
            assertPatternsInput(source);
            const settings = new settings_1.default(options);
            return utils.pattern.isDynamicPattern(source, settings);
        }
        FastGlob.isDynamicPattern = isDynamicPattern;
        function escapePath(source) {
            assertPatternsInput(source);
            return utils.path.escape(source);
        }
        FastGlob.escapePath = escapePath;
    })(FastGlob || (FastGlob = {}));
    function getWorks(source, _Provider, options) {
        const patterns = [].concat(source);
        const settings = new settings_1.default(options);
        const tasks = taskManager.generate(patterns, settings);
        const provider = new _Provider(settings);
        return tasks.map(provider.read, provider);
    }
    function assertPatternsInput(input) {
        const source = [].concat(input);
        const isValidSource = source.every((item) => utils.string.isString(item) && !utils.string.isEmpty(item));
        if (!isValidSource) {
            throw new TypeError('Patterns must be a string (non empty) or an array of strings');
        }
    }
    module.exports = FastGlob;
},
/* fast-glob/out/managers/tasks.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const utils = require(1126) /* ../utils */;
    function generate(patterns, settings) {
        const positivePatterns = getPositivePatterns(patterns);
        const negativePatterns = getNegativePatternsAsPositive(patterns, settings.ignore);
        const staticPatterns = positivePatterns.filter((pattern) => utils.pattern.isStaticPattern(pattern, settings));
        const dynamicPatterns = positivePatterns.filter((pattern) => utils.pattern.isDynamicPattern(pattern, settings));
        const staticTasks = convertPatternsToTasks(staticPatterns, negativePatterns, /* dynamic */ false);
        const dynamicTasks = convertPatternsToTasks(dynamicPatterns, negativePatterns, /* dynamic */ true);
        return staticTasks.concat(dynamicTasks);
    }
    exports.generate = generate;
    function convertPatternsToTasks(positive, negative, dynamic) {
        const positivePatternsGroup = groupPatternsByBaseDirectory(positive);
        // When we have a global group – there is no reason to divide the patterns into independent tasks.
        // In this case, the global task covers the rest.
        if ('.' in positivePatternsGroup) {
            const task = convertPatternGroupToTask('.', positive, negative, dynamic);
            return [task];
        }
        return convertPatternGroupsToTasks(positivePatternsGroup, negative, dynamic);
    }
    exports.convertPatternsToTasks = convertPatternsToTasks;
    function getPositivePatterns(patterns) {
        return utils.pattern.getPositivePatterns(patterns);
    }
    exports.getPositivePatterns = getPositivePatterns;
    function getNegativePatternsAsPositive(patterns, ignore) {
        const negative = utils.pattern.getNegativePatterns(patterns).concat(ignore);
        const positive = negative.map(utils.pattern.convertToPositivePattern);
        return positive;
    }
    exports.getNegativePatternsAsPositive = getNegativePatternsAsPositive;
    function groupPatternsByBaseDirectory(patterns) {
        const group = {};
        return patterns.reduce((collection, pattern) => {
            const base = utils.pattern.getBaseDirectory(pattern);
            if (base in collection) {
                collection[base].push(pattern);
            }
            else {
                collection[base] = [pattern];
            }
            return collection;
        }, group);
    }
    exports.groupPatternsByBaseDirectory = groupPatternsByBaseDirectory;
    function convertPatternGroupsToTasks(positive, negative, dynamic) {
        return Object.keys(positive).map((base) => {
            return convertPatternGroupToTask(base, positive[base], negative, dynamic);
        });
    }
    exports.convertPatternGroupsToTasks = convertPatternGroupsToTasks;
    function convertPatternGroupToTask(base, positive, negative, dynamic) {
        return {
            dynamic,
            positive,
            negative,
            base,
            patterns: [].concat(positive, negative.map(utils.pattern.convertToNegativePattern))
        };
    }
    exports.convertPatternGroupToTask = convertPatternGroupToTask;
},
/* fast-glob/out/utils/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const array = require(1127) /* ./array */;
    exports.array = array;
    const errno = require(1128) /* ./errno */;
    exports.errno = errno;
    const fs = require(1129) /* ./fs */;
    exports.fs = fs;
    const path = require(1130) /* ./path */;
    exports.path = path;
    const pattern = require(1131) /* ./pattern */;
    exports.pattern = pattern;
    const stream = require(1149) /* ./stream */;
    exports.stream = stream;
    const string = require(1150) /* ./string */;
    exports.string = string;
},
/* fast-glob/out/utils/array.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function flatten(items) {
        return items.reduce((collection, item) => [].concat(collection, item), []);
    }
    exports.flatten = flatten;
    function splitWhen(items, predicate) {
        const result = [[]];
        let groupIndex = 0;
        for (const item of items) {
            if (predicate(item)) {
                groupIndex++;
                result[groupIndex] = [];
            }
            else {
                result[groupIndex].push(item);
            }
        }
        return result;
    }
    exports.splitWhen = splitWhen;
},
/* fast-glob/out/utils/errno.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function isEnoentCodeError(error) {
        return error.code === 'ENOENT';
    }
    exports.isEnoentCodeError = isEnoentCodeError;
},
/* fast-glob/out/utils/fs.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    class DirentFromStats {
        constructor(name, stats) {
            this.name = name;
            this.isBlockDevice = stats.isBlockDevice.bind(stats);
            this.isCharacterDevice = stats.isCharacterDevice.bind(stats);
            this.isDirectory = stats.isDirectory.bind(stats);
            this.isFIFO = stats.isFIFO.bind(stats);
            this.isFile = stats.isFile.bind(stats);
            this.isSocket = stats.isSocket.bind(stats);
            this.isSymbolicLink = stats.isSymbolicLink.bind(stats);
        }
    }
    function createDirentFromStats(name, stats) {
        return new DirentFromStats(name, stats);
    }
    exports.createDirentFromStats = createDirentFromStats;
},
/* fast-glob/out/utils/path.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const path = require("path");
    const LEADING_DOT_SEGMENT_CHARACTERS_COUNT = 2; // ./ or .\\
    const UNESCAPED_GLOB_SYMBOLS_RE = /(\\?)([()*?[\]{|}]|^!|[!+@](?=\())/g;
    /**
     * Designed to work only with simple paths: `dir\\file`.
     */
    function unixify(filepath) {
        return filepath.replace(/\\/g, '/');
    }
    exports.unixify = unixify;
    function makeAbsolute(cwd, filepath) {
        return path.resolve(cwd, filepath);
    }
    exports.makeAbsolute = makeAbsolute;
    function escape(pattern) {
        return pattern.replace(UNESCAPED_GLOB_SYMBOLS_RE, '\\$2');
    }
    exports.escape = escape;
    function removeLeadingDotSegment(entry) {
        // We do not use `startsWith` because this is 10x slower than current implementation for some cases.
        // eslint-disable-next-line @typescript-eslint/prefer-string-starts-ends-with
        if (entry.charAt(0) === '.') {
            const secondCharactery = entry.charAt(1);
            if (secondCharactery === '/' || secondCharactery === '\\') {
                return entry.slice(LEADING_DOT_SEGMENT_CHARACTERS_COUNT);
            }
        }
        return entry;
    }
    exports.removeLeadingDotSegment = removeLeadingDotSegment;
},
/* fast-glob/out/utils/pattern.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const path = require("path");
    const globParent = require(326) /* glob-parent */;
    const micromatch = require(1132) /* micromatch */;
    const picomatch = require(1143) /* picomatch */;
    const GLOBSTAR = '**';
    const ESCAPE_SYMBOL = '\\';
    const COMMON_GLOB_SYMBOLS_RE = /[*?]|^!/;
    const REGEX_CHARACTER_CLASS_SYMBOLS_RE = /\[.*]/;
    const REGEX_GROUP_SYMBOLS_RE = /(?:^|[^!*+?@])\(.*\|.*\)/;
    const GLOB_EXTENSION_SYMBOLS_RE = /[!*+?@]\(.*\)/;
    const BRACE_EXPANSIONS_SYMBOLS_RE = /{.*(?:,|\.\.).*}/;
    function isStaticPattern(pattern, options = {}) {
        return !isDynamicPattern(pattern, options);
    }
    exports.isStaticPattern = isStaticPattern;
    function isDynamicPattern(pattern, options = {}) {
        /**
         * A special case with an empty string is necessary for matching patterns that start with a forward slash.
         * An empty string cannot be a dynamic pattern.
         * For example, the pattern `/lib/*` will be spread into parts: '', 'lib', '*'.
         */
        if (pattern === '') {
            return false;
        }
        /**
         * When the `caseSensitiveMatch` option is disabled, all patterns must be marked as dynamic, because we cannot check
         * filepath directly (without read directory).
         */
        if (options.caseSensitiveMatch === false || pattern.includes(ESCAPE_SYMBOL)) {
            return true;
        }
        if (COMMON_GLOB_SYMBOLS_RE.test(pattern) || REGEX_CHARACTER_CLASS_SYMBOLS_RE.test(pattern) || REGEX_GROUP_SYMBOLS_RE.test(pattern)) {
            return true;
        }
        if (options.extglob !== false && GLOB_EXTENSION_SYMBOLS_RE.test(pattern)) {
            return true;
        }
        if (options.braceExpansion !== false && BRACE_EXPANSIONS_SYMBOLS_RE.test(pattern)) {
            return true;
        }
        return false;
    }
    exports.isDynamicPattern = isDynamicPattern;
    function convertToPositivePattern(pattern) {
        return isNegativePattern(pattern) ? pattern.slice(1) : pattern;
    }
    exports.convertToPositivePattern = convertToPositivePattern;
    function convertToNegativePattern(pattern) {
        return '!' + pattern;
    }
    exports.convertToNegativePattern = convertToNegativePattern;
    function isNegativePattern(pattern) {
        return pattern.startsWith('!') && pattern[1] !== '(';
    }
    exports.isNegativePattern = isNegativePattern;
    function isPositivePattern(pattern) {
        return !isNegativePattern(pattern);
    }
    exports.isPositivePattern = isPositivePattern;
    function getNegativePatterns(patterns) {
        return patterns.filter(isNegativePattern);
    }
    exports.getNegativePatterns = getNegativePatterns;
    function getPositivePatterns(patterns) {
        return patterns.filter(isPositivePattern);
    }
    exports.getPositivePatterns = getPositivePatterns;
    function getBaseDirectory(pattern) {
        return globParent(pattern, { flipBackslashes: false });
    }
    exports.getBaseDirectory = getBaseDirectory;
    function hasGlobStar(pattern) {
        return pattern.includes(GLOBSTAR);
    }
    exports.hasGlobStar = hasGlobStar;
    function endsWithSlashGlobStar(pattern) {
        return pattern.endsWith('/' + GLOBSTAR);
    }
    exports.endsWithSlashGlobStar = endsWithSlashGlobStar;
    function isAffectDepthOfReadingPattern(pattern) {
        const basename = path.basename(pattern);
        return endsWithSlashGlobStar(pattern) || isStaticPattern(basename);
    }
    exports.isAffectDepthOfReadingPattern = isAffectDepthOfReadingPattern;
    function expandPatternsWithBraceExpansion(patterns) {
        return patterns.reduce((collection, pattern) => {
            return collection.concat(expandBraceExpansion(pattern));
        }, []);
    }
    exports.expandPatternsWithBraceExpansion = expandPatternsWithBraceExpansion;
    function expandBraceExpansion(pattern) {
        return micromatch.braces(pattern, {
            expand: true,
            nodupes: true
        });
    }
    exports.expandBraceExpansion = expandBraceExpansion;
    function getPatternParts(pattern, options) {
        let { parts } = picomatch.scan(pattern, Object.assign(Object.assign({}, options), { parts: true }));
        /**
         * The scan method returns an empty array in some cases.
         * See micromatch/picomatch#58 for more details.
         */
        if (parts.length === 0) {
            parts = [pattern];
        }
        /**
         * The scan method does not return an empty part for the pattern with a forward slash.
         * This is another part of micromatch/picomatch#58.
         */
        if (parts[0].startsWith('/')) {
            parts[0] = parts[0].slice(1);
            parts.unshift('');
        }
        return parts;
    }
    exports.getPatternParts = getPatternParts;
    function makeRe(pattern, options) {
        return micromatch.makeRe(pattern, options);
    }
    exports.makeRe = makeRe;
    function convertPatternsToRe(patterns, options) {
        return patterns.map((pattern) => makeRe(pattern, options));
    }
    exports.convertPatternsToRe = convertPatternsToRe;
    function matchAny(entry, patternsRe) {
        return patternsRe.some((patternRe) => patternRe.test(entry));
    }
    exports.matchAny = matchAny;
},
/* micromatch/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const util = require('util');
    const braces = require(1133) /* braces */;
    const picomatch = require(1143) /* picomatch */;
    const utils = require(1146) /* picomatch/lib/utils */;
    const isEmptyString = val => val === '' || val === './';
    /**
     * Returns an array of strings that match one or more glob patterns.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm(list, patterns[, options]);
     *
     * console.log(mm(['a.js', 'a.txt'], ['*.js']));
     * //=> [ 'a.js' ]
     * ```
     * @param {String|Array<string>} `list` List of strings to match.
     * @param {String|Array<string>} `patterns` One or more glob patterns to use for matching.
     * @param {Object} `options` See available [options](#options)
     * @return {Array} Returns an array of matches
     * @summary false
     * @api public
     */
    const micromatch = (list, patterns, options) => {
        patterns = [].concat(patterns);
        list = [].concat(list);
        let omit = new Set();
        let keep = new Set();
        let items = new Set();
        let negatives = 0;
        let onResult = state => {
            items.add(state.output);
            if (options && options.onResult) {
                options.onResult(state);
            }
        };
        for (let i = 0; i < patterns.length; i++) {
            let isMatch = picomatch(String(patterns[i]), { ...options, onResult }, true);
            let negated = isMatch.state.negated || isMatch.state.negatedExtglob;
            if (negated)
                negatives++;
            for (let item of list) {
                let matched = isMatch(item, true);
                let match = negated ? !matched.isMatch : matched.isMatch;
                if (!match)
                    continue;
                if (negated) {
                    omit.add(matched.output);
                }
                else {
                    omit.delete(matched.output);
                    keep.add(matched.output);
                }
            }
        }
        let result = negatives === patterns.length ? [...items] : [...keep];
        let matches = result.filter(item => !omit.has(item));
        if (options && matches.length === 0) {
            if (options.failglob === true) {
                throw new Error(`No matches found for "${patterns.join(', ')}"`);
            }
            if (options.nonull === true || options.nullglob === true) {
                return options.unescape ? patterns.map(p => p.replace(/\\/g, '')) : patterns;
            }
        }
        return matches;
    };
    /**
     * Backwards compatibility
     */
    micromatch.match = micromatch;
    /**
     * Returns a matcher function from the given glob `pattern` and `options`.
     * The returned function takes a string to match as its only argument and returns
     * true if the string is a match.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm.matcher(pattern[, options]);
     *
     * const isMatch = mm.matcher('*.!(*a)');
     * console.log(isMatch('a.a')); //=> false
     * console.log(isMatch('a.b')); //=> true
     * ```
     * @param {String} `pattern` Glob pattern
     * @param {Object} `options`
     * @return {Function} Returns a matcher function.
     * @api public
     */
    micromatch.matcher = (pattern, options) => picomatch(pattern, options);
    /**
     * Returns true if **any** of the given glob `patterns` match the specified `string`.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm.isMatch(string, patterns[, options]);
     *
     * console.log(mm.isMatch('a.a', ['b.*', '*.a'])); //=> true
     * console.log(mm.isMatch('a.a', 'b.*')); //=> false
     * ```
     * @param {String} `str` The string to test.
     * @param {String|Array} `patterns` One or more glob patterns to use for matching.
     * @param {Object} `[options]` See available [options](#options).
     * @return {Boolean} Returns true if any patterns match `str`
     * @api public
     */
    micromatch.isMatch = (str, patterns, options) => picomatch(patterns, options)(str);
    /**
     * Backwards compatibility
     */
    micromatch.any = micromatch.isMatch;
    /**
     * Returns a list of strings that _**do not match any**_ of the given `patterns`.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm.not(list, patterns[, options]);
     *
     * console.log(mm.not(['a.a', 'b.b', 'c.c'], '*.a'));
     * //=> ['b.b', 'c.c']
     * ```
     * @param {Array} `list` Array of strings to match.
     * @param {String|Array} `patterns` One or more glob pattern to use for matching.
     * @param {Object} `options` See available [options](#options) for changing how matches are performed
     * @return {Array} Returns an array of strings that **do not match** the given patterns.
     * @api public
     */
    micromatch.not = (list, patterns, options = {}) => {
        patterns = [].concat(patterns).map(String);
        let result = new Set();
        let items = [];
        let onResult = state => {
            if (options.onResult)
                options.onResult(state);
            items.push(state.output);
        };
        let matches = micromatch(list, patterns, { ...options, onResult });
        for (let item of items) {
            if (!matches.includes(item)) {
                result.add(item);
            }
        }
        return [...result];
    };
    /**
     * Returns true if the given `string` contains the given pattern. Similar
     * to [.isMatch](#isMatch) but the pattern can match any part of the string.
     *
     * ```js
     * var mm = require('micromatch');
     * // mm.contains(string, pattern[, options]);
     *
     * console.log(mm.contains('aa/bb/cc', '*b'));
     * //=> true
     * console.log(mm.contains('aa/bb/cc', '*d'));
     * //=> false
     * ```
     * @param {String} `str` The string to match.
     * @param {String|Array} `patterns` Glob pattern to use for matching.
     * @param {Object} `options` See available [options](#options) for changing how matches are performed
     * @return {Boolean} Returns true if any of the patterns matches any part of `str`.
     * @api public
     */
    micromatch.contains = (str, pattern, options) => {
        if (typeof str !== 'string') {
            throw new TypeError(`Expected a string: "${util.inspect(str)}"`);
        }
        if (Array.isArray(pattern)) {
            return pattern.some(p => micromatch.contains(str, p, options));
        }
        if (typeof pattern === 'string') {
            if (isEmptyString(str) || isEmptyString(pattern)) {
                return false;
            }
            if (str.includes(pattern) || (str.startsWith('./') && str.slice(2).includes(pattern))) {
                return true;
            }
        }
        return micromatch.isMatch(str, pattern, { ...options, contains: true });
    };
    /**
     * Filter the keys of the given object with the given `glob` pattern
     * and `options`. Does not attempt to match nested keys. If you need this feature,
     * use [glob-object][] instead.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm.matchKeys(object, patterns[, options]);
     *
     * const obj = { aa: 'a', ab: 'b', ac: 'c' };
     * console.log(mm.matchKeys(obj, '*b'));
     * //=> { ab: 'b' }
     * ```
     * @param {Object} `object` The object with keys to filter.
     * @param {String|Array} `patterns` One or more glob patterns to use for matching.
     * @param {Object} `options` See available [options](#options) for changing how matches are performed
     * @return {Object} Returns an object with only keys that match the given patterns.
     * @api public
     */
    micromatch.matchKeys = (obj, patterns, options) => {
        if (!utils.isObject(obj)) {
            throw new TypeError('Expected the first argument to be an object');
        }
        let keys = micromatch(Object.keys(obj), patterns, options);
        let res = {};
        for (let key of keys)
            res[key] = obj[key];
        return res;
    };
    /**
     * Returns true if some of the strings in the given `list` match any of the given glob `patterns`.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm.some(list, patterns[, options]);
     *
     * console.log(mm.some(['foo.js', 'bar.js'], ['*.js', '!foo.js']));
     * // true
     * console.log(mm.some(['foo.js'], ['*.js', '!foo.js']));
     * // false
     * ```
     * @param {String|Array} `list` The string or array of strings to test. Returns as soon as the first match is found.
     * @param {String|Array} `patterns` One or more glob patterns to use for matching.
     * @param {Object} `options` See available [options](#options) for changing how matches are performed
     * @return {Boolean} Returns true if any `patterns` matches any of the strings in `list`
     * @api public
     */
    micromatch.some = (list, patterns, options) => {
        let items = [].concat(list);
        for (let pattern of [].concat(patterns)) {
            let isMatch = picomatch(String(pattern), options);
            if (items.some(item => isMatch(item))) {
                return true;
            }
        }
        return false;
    };
    /**
     * Returns true if every string in the given `list` matches
     * any of the given glob `patterns`.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm.every(list, patterns[, options]);
     *
     * console.log(mm.every('foo.js', ['foo.js']));
     * // true
     * console.log(mm.every(['foo.js', 'bar.js'], ['*.js']));
     * // true
     * console.log(mm.every(['foo.js', 'bar.js'], ['*.js', '!foo.js']));
     * // false
     * console.log(mm.every(['foo.js'], ['*.js', '!foo.js']));
     * // false
     * ```
     * @param {String|Array} `list` The string or array of strings to test.
     * @param {String|Array} `patterns` One or more glob patterns to use for matching.
     * @param {Object} `options` See available [options](#options) for changing how matches are performed
     * @return {Boolean} Returns true if all `patterns` matches all of the strings in `list`
     * @api public
     */
    micromatch.every = (list, patterns, options) => {
        let items = [].concat(list);
        for (let pattern of [].concat(patterns)) {
            let isMatch = picomatch(String(pattern), options);
            if (!items.every(item => isMatch(item))) {
                return false;
            }
        }
        return true;
    };
    /**
     * Returns true if **all** of the given `patterns` match
     * the specified string.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm.all(string, patterns[, options]);
     *
     * console.log(mm.all('foo.js', ['foo.js']));
     * // true
     *
     * console.log(mm.all('foo.js', ['*.js', '!foo.js']));
     * // false
     *
     * console.log(mm.all('foo.js', ['*.js', 'foo.js']));
     * // true
     *
     * console.log(mm.all('foo.js', ['*.js', 'f*', '*o*', '*o.js']));
     * // true
     * ```
     * @param {String|Array} `str` The string to test.
     * @param {String|Array} `patterns` One or more glob patterns to use for matching.
     * @param {Object} `options` See available [options](#options) for changing how matches are performed
     * @return {Boolean} Returns true if any patterns match `str`
     * @api public
     */
    micromatch.all = (str, patterns, options) => {
        if (typeof str !== 'string') {
            throw new TypeError(`Expected a string: "${util.inspect(str)}"`);
        }
        return [].concat(patterns).every(p => picomatch(p, options)(str));
    };
    /**
     * Returns an array of matches captured by `pattern` in `string, or `null` if the pattern did not match.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm.capture(pattern, string[, options]);
     *
     * console.log(mm.capture('test/*.js', 'test/foo.js'));
     * //=> ['foo']
     * console.log(mm.capture('test/*.js', 'foo/bar.css'));
     * //=> null
     * ```
     * @param {String} `glob` Glob pattern to use for matching.
     * @param {String} `input` String to match
     * @param {Object} `options` See available [options](#options) for changing how matches are performed
     * @return {Array|null} Returns an array of captures if the input matches the glob pattern, otherwise `null`.
     * @api public
     */
    micromatch.capture = (glob, input, options) => {
        let posix = utils.isWindows(options);
        let regex = picomatch.makeRe(String(glob), { ...options, capture: true });
        let match = regex.exec(posix ? utils.toPosixSlashes(input) : input);
        if (match) {
            return match.slice(1).map(v => v === void 0 ? '' : v);
        }
    };
    /**
     * Create a regular expression from the given glob `pattern`.
     *
     * ```js
     * const mm = require('micromatch');
     * // mm.makeRe(pattern[, options]);
     *
     * console.log(mm.makeRe('*.js'));
     * //=> /^(?:(\.[\\\/])?(?!\.)(?=.)[^\/]*?\.js)$/
     * ```
     * @param {String} `pattern` A glob pattern to convert to regex.
     * @param {Object} `options`
     * @return {RegExp} Returns a regex created from the given pattern.
     * @api public
     */
    micromatch.makeRe = (...args) => picomatch.makeRe(...args);
    /**
     * Scan a glob pattern to separate the pattern into segments. Used
     * by the [split](#split) method.
     *
     * ```js
     * const mm = require('micromatch');
     * const state = mm.scan(pattern[, options]);
     * ```
     * @param {String} `pattern`
     * @param {Object} `options`
     * @return {Object} Returns an object with
     * @api public
     */
    micromatch.scan = (...args) => picomatch.scan(...args);
    /**
     * Parse a glob pattern to create the source string for a regular
     * expression.
     *
     * ```js
     * const mm = require('micromatch');
     * const state = mm(pattern[, options]);
     * ```
     * @param {String} `glob`
     * @param {Object} `options`
     * @return {Object} Returns an object with useful properties and output to be used as regex source string.
     * @api public
     */
    micromatch.parse = (patterns, options) => {
        let res = [];
        for (let pattern of [].concat(patterns || [])) {
            for (let str of braces(String(pattern), options)) {
                res.push(picomatch.parse(str, options));
            }
        }
        return res;
    };
    /**
     * Process the given brace `pattern`.
     *
     * ```js
     * const { braces } = require('micromatch');
     * console.log(braces('foo/{a,b,c}/bar'));
     * //=> [ 'foo/(a|b|c)/bar' ]
     *
     * console.log(braces('foo/{a,b,c}/bar', { expand: true }));
     * //=> [ 'foo/a/bar', 'foo/b/bar', 'foo/c/bar' ]
     * ```
     * @param {String} `pattern` String with brace pattern to process.
     * @param {Object} `options` Any [options](#options) to change how expansion is performed. See the [braces][] library for all available options.
     * @return {Array}
     * @api public
     */
    micromatch.braces = (pattern, options) => {
        if (typeof pattern !== 'string')
            throw new TypeError('Expected a string');
        if ((options && options.nobrace === true) || !/\{.*\}/.test(pattern)) {
            return [pattern];
        }
        return braces(pattern, options);
    };
    /**
     * Expand braces
     */
    micromatch.braceExpand = (pattern, options) => {
        if (typeof pattern !== 'string')
            throw new TypeError('Expected a string');
        return micromatch.braces(pattern, { ...options, expand: true });
    };
    /**
     * Expose micromatch
     */
    module.exports = micromatch;
},
/* braces/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const stringify = require(1134) /* ./lib/stringify */;
    const compile = require(1136) /* ./lib/compile */;
    const expand = require(1140) /* ./lib/expand */;
    const parse = require(1141) /* ./lib/parse */;
    /**
     * Expand the given pattern or create a regex-compatible string.
     *
     * ```js
     * const braces = require('braces');
     * console.log(braces('{a,b,c}', { compile: true })); //=> ['(a|b|c)']
     * console.log(braces('{a,b,c}')); //=> ['a', 'b', 'c']
     * ```
     * @param {String} `str`
     * @param {Object} `options`
     * @return {String}
     * @api public
     */
    const braces = (input, options = {}) => {
        let output = [];
        if (Array.isArray(input)) {
            for (let pattern of input) {
                let result = braces.create(pattern, options);
                if (Array.isArray(result)) {
                    output.push(...result);
                }
                else {
                    output.push(result);
                }
            }
        }
        else {
            output = [].concat(braces.create(input, options));
        }
        if (options && options.expand === true && options.nodupes === true) {
            output = [...new Set(output)];
        }
        return output;
    };
    /**
     * Parse the given `str` with the given `options`.
     *
     * ```js
     * // braces.parse(pattern, [, options]);
     * const ast = braces.parse('a/{b,c}/d');
     * console.log(ast);
     * ```
     * @param {String} pattern Brace pattern to parse
     * @param {Object} options
     * @return {Object} Returns an AST
     * @api public
     */
    braces.parse = (input, options = {}) => parse(input, options);
    /**
     * Creates a braces string from an AST, or an AST node.
     *
     * ```js
     * const braces = require('braces');
     * let ast = braces.parse('foo/{a,b}/bar');
     * console.log(stringify(ast.nodes[2])); //=> '{a,b}'
     * ```
     * @param {String} `input` Brace pattern or AST.
     * @param {Object} `options`
     * @return {Array} Returns an array of expanded values.
     * @api public
     */
    braces.stringify = (input, options = {}) => {
        if (typeof input === 'string') {
            return stringify(braces.parse(input, options), options);
        }
        return stringify(input, options);
    };
    /**
     * Compiles a brace pattern into a regex-compatible, optimized string.
     * This method is called by the main [braces](#braces) function by default.
     *
     * ```js
     * const braces = require('braces');
     * console.log(braces.compile('a/{b,c}/d'));
     * //=> ['a/(b|c)/d']
     * ```
     * @param {String} `input` Brace pattern or AST.
     * @param {Object} `options`
     * @return {Array} Returns an array of expanded values.
     * @api public
     */
    braces.compile = (input, options = {}) => {
        if (typeof input === 'string') {
            input = braces.parse(input, options);
        }
        return compile(input, options);
    };
    /**
     * Expands a brace pattern into an array. This method is called by the
     * main [braces](#braces) function when `options.expand` is true. Before
     * using this method it's recommended that you read the [performance notes](#performance))
     * and advantages of using [.compile](#compile) instead.
     *
     * ```js
     * const braces = require('braces');
     * console.log(braces.expand('a/{b,c}/d'));
     * //=> ['a/b/d', 'a/c/d'];
     * ```
     * @param {String} `pattern` Brace pattern
     * @param {Object} `options`
     * @return {Array} Returns an array of expanded values.
     * @api public
     */
    braces.expand = (input, options = {}) => {
        if (typeof input === 'string') {
            input = braces.parse(input, options);
        }
        let result = expand(input, options);
        // filter out empty strings if specified
        if (options.noempty === true) {
            result = result.filter(Boolean);
        }
        // filter out duplicates if specified
        if (options.nodupes === true) {
            result = [...new Set(result)];
        }
        return result;
    };
    /**
     * Processes a brace pattern and returns either an expanded array
     * (if `options.expand` is true), a highly optimized regex-compatible string.
     * This method is called by the main [braces](#braces) function.
     *
     * ```js
     * const braces = require('braces');
     * console.log(braces.create('user-{200..300}/project-{a,b,c}-{1..10}'))
     * //=> 'user-(20[0-9]|2[1-9][0-9]|300)/project-(a|b|c)-([1-9]|10)'
     * ```
     * @param {String} `pattern` Brace pattern
     * @param {Object} `options`
     * @return {Array} Returns an array of expanded values.
     * @api public
     */
    braces.create = (input, options = {}) => {
        if (input === '' || input.length < 3) {
            return [input];
        }
        return options.expand !== true
            ? braces.compile(input, options)
            : braces.expand(input, options);
    };
    /**
     * Expose "braces"
     */
    module.exports = braces;
},
/* braces/lib/stringify.js */ function _(require, module, exports, __esModule, __esExport) {
    const utils = require(1135) /* ./utils */;
    module.exports = (ast, options = {}) => {
        let stringify = (node, parent = {}) => {
            let invalidBlock = options.escapeInvalid && utils.isInvalidBrace(parent);
            let invalidNode = node.invalid === true && options.escapeInvalid === true;
            let output = '';
            if (node.value) {
                if ((invalidBlock || invalidNode) && utils.isOpenOrClose(node)) {
                    return '\\' + node.value;
                }
                return node.value;
            }
            if (node.value) {
                return node.value;
            }
            if (node.nodes) {
                for (let child of node.nodes) {
                    output += stringify(child);
                }
            }
            return output;
        };
        return stringify(ast);
    };
},
/* braces/lib/utils.js */ function _(require, module, exports, __esModule, __esExport) {
    exports.isInteger = num => {
        if (typeof num === 'number') {
            return Number.isInteger(num);
        }
        if (typeof num === 'string' && num.trim() !== '') {
            return Number.isInteger(Number(num));
        }
        return false;
    };
    /**
     * Find a node of the given type
     */
    exports.find = (node, type) => node.nodes.find(node => node.type === type);
    /**
     * Find a node of the given type
     */
    exports.exceedsLimit = (min, max, step = 1, limit) => {
        if (limit === false)
            return false;
        if (!exports.isInteger(min) || !exports.isInteger(max))
            return false;
        return ((Number(max) - Number(min)) / Number(step)) >= limit;
    };
    /**
     * Escape the given node with '\\' before node.value
     */
    exports.escapeNode = (block, n = 0, type) => {
        let node = block.nodes[n];
        if (!node)
            return;
        if ((type && node.type === type) || node.type === 'open' || node.type === 'close') {
            if (node.escaped !== true) {
                node.value = '\\' + node.value;
                node.escaped = true;
            }
        }
    };
    /**
     * Returns true if the given brace node should be enclosed in literal braces
     */
    exports.encloseBrace = node => {
        if (node.type !== 'brace')
            return false;
        if ((node.commas >> 0 + node.ranges >> 0) === 0) {
            node.invalid = true;
            return true;
        }
        return false;
    };
    /**
     * Returns true if a brace node is invalid.
     */
    exports.isInvalidBrace = block => {
        if (block.type !== 'brace')
            return false;
        if (block.invalid === true || block.dollar)
            return true;
        if ((block.commas >> 0 + block.ranges >> 0) === 0) {
            block.invalid = true;
            return true;
        }
        if (block.open !== true || block.close !== true) {
            block.invalid = true;
            return true;
        }
        return false;
    };
    /**
     * Returns true if a node is an open or close node
     */
    exports.isOpenOrClose = node => {
        if (node.type === 'open' || node.type === 'close') {
            return true;
        }
        return node.open === true || node.close === true;
    };
    /**
     * Reduce an array of text nodes.
     */
    exports.reduce = nodes => nodes.reduce((acc, node) => {
        if (node.type === 'text')
            acc.push(node.value);
        if (node.type === 'range')
            node.type = 'text';
        return acc;
    }, []);
    /**
     * Flatten an array
     */
    exports.flatten = (...args) => {
        const result = [];
        const flat = arr => {
            for (let i = 0; i < arr.length; i++) {
                let ele = arr[i];
                Array.isArray(ele) ? flat(ele, result) : ele !== void 0 && result.push(ele);
            }
            return result;
        };
        flat(args);
        return result;
    };
},
/* braces/lib/compile.js */ function _(require, module, exports, __esModule, __esExport) {
    const fill = require(1137) /* fill-range */;
    const utils = require(1135) /* ./utils */;
    const compile = (ast, options = {}) => {
        let walk = (node, parent = {}) => {
            let invalidBlock = utils.isInvalidBrace(parent);
            let invalidNode = node.invalid === true && options.escapeInvalid === true;
            let invalid = invalidBlock === true || invalidNode === true;
            let prefix = options.escapeInvalid === true ? '\\' : '';
            let output = '';
            if (node.isOpen === true) {
                return prefix + node.value;
            }
            if (node.isClose === true) {
                return prefix + node.value;
            }
            if (node.type === 'open') {
                return invalid ? (prefix + node.value) : '(';
            }
            if (node.type === 'close') {
                return invalid ? (prefix + node.value) : ')';
            }
            if (node.type === 'comma') {
                return node.prev.type === 'comma' ? '' : (invalid ? node.value : '|');
            }
            if (node.value) {
                return node.value;
            }
            if (node.nodes && node.ranges > 0) {
                let args = utils.reduce(node.nodes);
                let range = fill(...args, { ...options, wrap: false, toRegex: true });
                if (range.length !== 0) {
                    return args.length > 1 && range.length > 1 ? `(${range})` : range;
                }
            }
            if (node.nodes) {
                for (let child of node.nodes) {
                    output += walk(child, node);
                }
            }
            return output;
        };
        return walk(ast);
    };
    module.exports = compile;
},
/* fill-range/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const util = require('util');
    const toRegexRange = require(1138) /* to-regex-range */;
    const isObject = val => val !== null && typeof val === 'object' && !Array.isArray(val);
    const transform = toNumber => {
        return value => toNumber === true ? Number(value) : String(value);
    };
    const isValidValue = value => {
        return typeof value === 'number' || (typeof value === 'string' && value !== '');
    };
    const isNumber = num => Number.isInteger(+num);
    const zeros = input => {
        let value = `${input}`;
        let index = -1;
        if (value[0] === '-')
            value = value.slice(1);
        if (value === '0')
            return false;
        while (value[++index] === '0')
            ;
        return index > 0;
    };
    const stringify = (start, end, options) => {
        if (typeof start === 'string' || typeof end === 'string') {
            return true;
        }
        return options.stringify === true;
    };
    const pad = (input, maxLength, toNumber) => {
        if (maxLength > 0) {
            let dash = input[0] === '-' ? '-' : '';
            if (dash)
                input = input.slice(1);
            input = (dash + input.padStart(dash ? maxLength - 1 : maxLength, '0'));
        }
        if (toNumber === false) {
            return String(input);
        }
        return input;
    };
    const toMaxLen = (input, maxLength) => {
        let negative = input[0] === '-' ? '-' : '';
        if (negative) {
            input = input.slice(1);
            maxLength--;
        }
        while (input.length < maxLength)
            input = '0' + input;
        return negative ? ('-' + input) : input;
    };
    const toSequence = (parts, options) => {
        parts.negatives.sort((a, b) => a < b ? -1 : a > b ? 1 : 0);
        parts.positives.sort((a, b) => a < b ? -1 : a > b ? 1 : 0);
        let prefix = options.capture ? '' : '?:';
        let positives = '';
        let negatives = '';
        let result;
        if (parts.positives.length) {
            positives = parts.positives.join('|');
        }
        if (parts.negatives.length) {
            negatives = `-(${prefix}${parts.negatives.join('|')})`;
        }
        if (positives && negatives) {
            result = `${positives}|${negatives}`;
        }
        else {
            result = positives || negatives;
        }
        if (options.wrap) {
            return `(${prefix}${result})`;
        }
        return result;
    };
    const toRange = (a, b, isNumbers, options) => {
        if (isNumbers) {
            return toRegexRange(a, b, { wrap: false, ...options });
        }
        let start = String.fromCharCode(a);
        if (a === b)
            return start;
        let stop = String.fromCharCode(b);
        return `[${start}-${stop}]`;
    };
    const toRegex = (start, end, options) => {
        if (Array.isArray(start)) {
            let wrap = options.wrap === true;
            let prefix = options.capture ? '' : '?:';
            return wrap ? `(${prefix}${start.join('|')})` : start.join('|');
        }
        return toRegexRange(start, end, options);
    };
    const rangeError = (...args) => {
        return new RangeError('Invalid range arguments: ' + util.inspect(...args));
    };
    const invalidRange = (start, end, options) => {
        if (options.strictRanges === true)
            throw rangeError([start, end]);
        return [];
    };
    const invalidStep = (step, options) => {
        if (options.strictRanges === true) {
            throw new TypeError(`Expected step "${step}" to be a number`);
        }
        return [];
    };
    const fillNumbers = (start, end, step = 1, options = {}) => {
        let a = Number(start);
        let b = Number(end);
        if (!Number.isInteger(a) || !Number.isInteger(b)) {
            if (options.strictRanges === true)
                throw rangeError([start, end]);
            return [];
        }
        // fix negative zero
        if (a === 0)
            a = 0;
        if (b === 0)
            b = 0;
        let descending = a > b;
        let startString = String(start);
        let endString = String(end);
        let stepString = String(step);
        step = Math.max(Math.abs(step), 1);
        let padded = zeros(startString) || zeros(endString) || zeros(stepString);
        let maxLen = padded ? Math.max(startString.length, endString.length, stepString.length) : 0;
        let toNumber = padded === false && stringify(start, end, options) === false;
        let format = options.transform || transform(toNumber);
        if (options.toRegex && step === 1) {
            return toRange(toMaxLen(start, maxLen), toMaxLen(end, maxLen), true, options);
        }
        let parts = { negatives: [], positives: [] };
        let push = num => parts[num < 0 ? 'negatives' : 'positives'].push(Math.abs(num));
        let range = [];
        let index = 0;
        while (descending ? a >= b : a <= b) {
            if (options.toRegex === true && step > 1) {
                push(a);
            }
            else {
                range.push(pad(format(a, index), maxLen, toNumber));
            }
            a = descending ? a - step : a + step;
            index++;
        }
        if (options.toRegex === true) {
            return step > 1
                ? toSequence(parts, options)
                : toRegex(range, null, { wrap: false, ...options });
        }
        return range;
    };
    const fillLetters = (start, end, step = 1, options = {}) => {
        if ((!isNumber(start) && start.length > 1) || (!isNumber(end) && end.length > 1)) {
            return invalidRange(start, end, options);
        }
        let format = options.transform || (val => String.fromCharCode(val));
        let a = `${start}`.charCodeAt(0);
        let b = `${end}`.charCodeAt(0);
        let descending = a > b;
        let min = Math.min(a, b);
        let max = Math.max(a, b);
        if (options.toRegex && step === 1) {
            return toRange(min, max, false, options);
        }
        let range = [];
        let index = 0;
        while (descending ? a >= b : a <= b) {
            range.push(format(a, index));
            a = descending ? a - step : a + step;
            index++;
        }
        if (options.toRegex === true) {
            return toRegex(range, null, { wrap: false, options });
        }
        return range;
    };
    const fill = (start, end, step, options = {}) => {
        if (end == null && isValidValue(start)) {
            return [start];
        }
        if (!isValidValue(start) || !isValidValue(end)) {
            return invalidRange(start, end, options);
        }
        if (typeof step === 'function') {
            return fill(start, end, 1, { transform: step });
        }
        if (isObject(step)) {
            return fill(start, end, 0, step);
        }
        let opts = { ...options };
        if (opts.capture === true)
            opts.wrap = true;
        step = step || opts.step || 1;
        if (!isNumber(step)) {
            if (step != null && !isObject(step))
                return invalidStep(step, opts);
            return fill(start, end, 1, step);
        }
        if (isNumber(start) && isNumber(end)) {
            return fillNumbers(start, end, step, opts);
        }
        return fillLetters(start, end, Math.max(Math.abs(step), 1), opts);
    };
    module.exports = fill;
},
/* to-regex-range/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const isNumber = require(1139) /* is-number */;
    const toRegexRange = (min, max, options) => {
        if (isNumber(min) === false) {
            throw new TypeError('toRegexRange: expected the first argument to be a number');
        }
        if (max === void 0 || min === max) {
            return String(min);
        }
        if (isNumber(max) === false) {
            throw new TypeError('toRegexRange: expected the second argument to be a number.');
        }
        let opts = { relaxZeros: true, ...options };
        if (typeof opts.strictZeros === 'boolean') {
            opts.relaxZeros = opts.strictZeros === false;
        }
        let relax = String(opts.relaxZeros);
        let shorthand = String(opts.shorthand);
        let capture = String(opts.capture);
        let wrap = String(opts.wrap);
        let cacheKey = min + ':' + max + '=' + relax + shorthand + capture + wrap;
        if (toRegexRange.cache.hasOwnProperty(cacheKey)) {
            return toRegexRange.cache[cacheKey].result;
        }
        let a = Math.min(min, max);
        let b = Math.max(min, max);
        if (Math.abs(a - b) === 1) {
            let result = min + '|' + max;
            if (opts.capture) {
                return `(${result})`;
            }
            if (opts.wrap === false) {
                return result;
            }
            return `(?:${result})`;
        }
        let isPadded = hasPadding(min) || hasPadding(max);
        let state = { min, max, a, b };
        let positives = [];
        let negatives = [];
        if (isPadded) {
            state.isPadded = isPadded;
            state.maxLen = String(state.max).length;
        }
        if (a < 0) {
            let newMin = b < 0 ? Math.abs(b) : 1;
            negatives = splitToPatterns(newMin, Math.abs(a), state, opts);
            a = state.a = 0;
        }
        if (b >= 0) {
            positives = splitToPatterns(a, b, state, opts);
        }
        state.negatives = negatives;
        state.positives = positives;
        state.result = collatePatterns(negatives, positives, opts);
        if (opts.capture === true) {
            state.result = `(${state.result})`;
        }
        else if (opts.wrap !== false && (positives.length + negatives.length) > 1) {
            state.result = `(?:${state.result})`;
        }
        toRegexRange.cache[cacheKey] = state;
        return state.result;
    };
    function collatePatterns(neg, pos, options) {
        let onlyNegative = filterPatterns(neg, pos, '-', false, options) || [];
        let onlyPositive = filterPatterns(pos, neg, '', false, options) || [];
        let intersected = filterPatterns(neg, pos, '-?', true, options) || [];
        let subpatterns = onlyNegative.concat(intersected).concat(onlyPositive);
        return subpatterns.join('|');
    }
    function splitToRanges(min, max) {
        let nines = 1;
        let zeros = 1;
        let stop = countNines(min, nines);
        let stops = new Set([max]);
        while (min <= stop && stop <= max) {
            stops.add(stop);
            nines += 1;
            stop = countNines(min, nines);
        }
        stop = countZeros(max + 1, zeros) - 1;
        while (min < stop && stop <= max) {
            stops.add(stop);
            zeros += 1;
            stop = countZeros(max + 1, zeros) - 1;
        }
        stops = [...stops];
        stops.sort(compare);
        return stops;
    }
    /**
     * Convert a range to a regex pattern
     * @param {Number} `start`
     * @param {Number} `stop`
     * @return {String}
     */
    function rangeToPattern(start, stop, options) {
        if (start === stop) {
            return { pattern: start, count: [], digits: 0 };
        }
        let zipped = zip(start, stop);
        let digits = zipped.length;
        let pattern = '';
        let count = 0;
        for (let i = 0; i < digits; i++) {
            let [startDigit, stopDigit] = zipped[i];
            if (startDigit === stopDigit) {
                pattern += startDigit;
            }
            else if (startDigit !== '0' || stopDigit !== '9') {
                pattern += toCharacterClass(startDigit, stopDigit, options);
            }
            else {
                count++;
            }
        }
        if (count) {
            pattern += options.shorthand === true ? '\\d' : '[0-9]';
        }
        return { pattern, count: [count], digits };
    }
    function splitToPatterns(min, max, tok, options) {
        let ranges = splitToRanges(min, max);
        let tokens = [];
        let start = min;
        let prev;
        for (let i = 0; i < ranges.length; i++) {
            let max = ranges[i];
            let obj = rangeToPattern(String(start), String(max), options);
            let zeros = '';
            if (!tok.isPadded && prev && prev.pattern === obj.pattern) {
                if (prev.count.length > 1) {
                    prev.count.pop();
                }
                prev.count.push(obj.count[0]);
                prev.string = prev.pattern + toQuantifier(prev.count);
                start = max + 1;
                continue;
            }
            if (tok.isPadded) {
                zeros = padZeros(max, tok, options);
            }
            obj.string = zeros + obj.pattern + toQuantifier(obj.count);
            tokens.push(obj);
            start = max + 1;
            prev = obj;
        }
        return tokens;
    }
    function filterPatterns(arr, comparison, prefix, intersection, options) {
        let result = [];
        for (let ele of arr) {
            let { string } = ele;
            // only push if _both_ are negative...
            if (!intersection && !contains(comparison, 'string', string)) {
                result.push(prefix + string);
            }
            // or _both_ are positive
            if (intersection && contains(comparison, 'string', string)) {
                result.push(prefix + string);
            }
        }
        return result;
    }
    /**
     * Zip strings
     */
    function zip(a, b) {
        let arr = [];
        for (let i = 0; i < a.length; i++)
            arr.push([a[i], b[i]]);
        return arr;
    }
    function compare(a, b) {
        return a > b ? 1 : b > a ? -1 : 0;
    }
    function contains(arr, key, val) {
        return arr.some(ele => ele[key] === val);
    }
    function countNines(min, len) {
        return Number(String(min).slice(0, -len) + '9'.repeat(len));
    }
    function countZeros(integer, zeros) {
        return integer - (integer % Math.pow(10, zeros));
    }
    function toQuantifier(digits) {
        let [start = 0, stop = ''] = digits;
        if (stop || start > 1) {
            return `{${start + (stop ? ',' + stop : '')}}`;
        }
        return '';
    }
    function toCharacterClass(a, b, options) {
        return `[${a}${(b - a === 1) ? '' : '-'}${b}]`;
    }
    function hasPadding(str) {
        return /^-?(0+)\d/.test(str);
    }
    function padZeros(value, tok, options) {
        if (!tok.isPadded) {
            return value;
        }
        let diff = Math.abs(tok.maxLen - String(value).length);
        let relax = options.relaxZeros !== false;
        switch (diff) {
            case 0:
                return '';
            case 1:
                return relax ? '0?' : '0';
            case 2:
                return relax ? '0{0,2}' : '00';
            default: {
                return relax ? `0{0,${diff}}` : `0{${diff}}`;
            }
        }
    }
    /**
     * Cache
     */
    toRegexRange.cache = {};
    toRegexRange.clearCache = () => (toRegexRange.cache = {});
    /**
     * Expose `toRegexRange`
     */
    module.exports = toRegexRange;
},
/* is-number/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = function (num) {
        if (typeof num === 'number') {
            return num - num === 0;
        }
        if (typeof num === 'string' && num.trim() !== '') {
            return Number.isFinite ? Number.isFinite(+num) : isFinite(+num);
        }
        return false;
    };
},
/* braces/lib/expand.js */ function _(require, module, exports, __esModule, __esExport) {
    const fill = require(1137) /* fill-range */;
    const stringify = require(1134) /* ./stringify */;
    const utils = require(1135) /* ./utils */;
    const append = (queue = '', stash = '', enclose = false) => {
        let result = [];
        queue = [].concat(queue);
        stash = [].concat(stash);
        if (!stash.length)
            return queue;
        if (!queue.length) {
            return enclose ? utils.flatten(stash).map(ele => `{${ele}}`) : stash;
        }
        for (let item of queue) {
            if (Array.isArray(item)) {
                for (let value of item) {
                    result.push(append(value, stash, enclose));
                }
            }
            else {
                for (let ele of stash) {
                    if (enclose === true && typeof ele === 'string')
                        ele = `{${ele}}`;
                    result.push(Array.isArray(ele) ? append(item, ele, enclose) : (item + ele));
                }
            }
        }
        return utils.flatten(result);
    };
    const expand = (ast, options = {}) => {
        let rangeLimit = options.rangeLimit === void 0 ? 1000 : options.rangeLimit;
        let walk = (node, parent = {}) => {
            node.queue = [];
            let p = parent;
            let q = parent.queue;
            while (p.type !== 'brace' && p.type !== 'root' && p.parent) {
                p = p.parent;
                q = p.queue;
            }
            if (node.invalid || node.dollar) {
                q.push(append(q.pop(), stringify(node, options)));
                return;
            }
            if (node.type === 'brace' && node.invalid !== true && node.nodes.length === 2) {
                q.push(append(q.pop(), ['{}']));
                return;
            }
            if (node.nodes && node.ranges > 0) {
                let args = utils.reduce(node.nodes);
                if (utils.exceedsLimit(...args, options.step, rangeLimit)) {
                    throw new RangeError('expanded array length exceeds range limit. Use options.rangeLimit to increase or disable the limit.');
                }
                let range = fill(...args, options);
                if (range.length === 0) {
                    range = stringify(node, options);
                }
                q.push(append(q.pop(), range));
                node.nodes = [];
                return;
            }
            let enclose = utils.encloseBrace(node);
            let queue = node.queue;
            let block = node;
            while (block.type !== 'brace' && block.type !== 'root' && block.parent) {
                block = block.parent;
                queue = block.queue;
            }
            for (let i = 0; i < node.nodes.length; i++) {
                let child = node.nodes[i];
                if (child.type === 'comma' && node.type === 'brace') {
                    if (i === 1)
                        queue.push('');
                    queue.push('');
                    continue;
                }
                if (child.type === 'close') {
                    q.push(append(q.pop(), queue, enclose));
                    continue;
                }
                if (child.value && child.type !== 'open') {
                    queue.push(append(queue.pop(), child.value));
                    continue;
                }
                if (child.nodes) {
                    walk(child, node);
                }
            }
            return queue;
        };
        return utils.flatten(walk(ast));
    };
    module.exports = expand;
},
/* braces/lib/parse.js */ function _(require, module, exports, __esModule, __esExport) {
    const stringify = require(1134) /* ./stringify */;
    /**
     * Constants
     */
    const { MAX_LENGTH, CHAR_BACKSLASH, /* \ */ CHAR_BACKTICK, /* ` */ CHAR_COMMA, /* , */ CHAR_DOT, /* . */ CHAR_LEFT_PARENTHESES, /* ( */ CHAR_RIGHT_PARENTHESES, /* ) */ CHAR_LEFT_CURLY_BRACE, /* { */ CHAR_RIGHT_CURLY_BRACE, /* } */ CHAR_LEFT_SQUARE_BRACKET, /* [ */ CHAR_RIGHT_SQUARE_BRACKET, /* ] */ CHAR_DOUBLE_QUOTE, /* " */ CHAR_SINGLE_QUOTE, /* ' */ CHAR_NO_BREAK_SPACE, CHAR_ZERO_WIDTH_NOBREAK_SPACE } = require(1142) /* ./constants */;
    /**
     * parse
     */
    const parse = (input, options = {}) => {
        if (typeof input !== 'string') {
            throw new TypeError('Expected a string');
        }
        let opts = options || {};
        let max = typeof opts.maxLength === 'number' ? Math.min(MAX_LENGTH, opts.maxLength) : MAX_LENGTH;
        if (input.length > max) {
            throw new SyntaxError(`Input length (${input.length}), exceeds max characters (${max})`);
        }
        let ast = { type: 'root', input, nodes: [] };
        let stack = [ast];
        let block = ast;
        let prev = ast;
        let brackets = 0;
        let length = input.length;
        let index = 0;
        let depth = 0;
        let value;
        let memo = {};
        /**
         * Helpers
         */
        const advance = () => input[index++];
        const push = node => {
            if (node.type === 'text' && prev.type === 'dot') {
                prev.type = 'text';
            }
            if (prev && prev.type === 'text' && node.type === 'text') {
                prev.value += node.value;
                return;
            }
            block.nodes.push(node);
            node.parent = block;
            node.prev = prev;
            prev = node;
            return node;
        };
        push({ type: 'bos' });
        while (index < length) {
            block = stack[stack.length - 1];
            value = advance();
            /**
             * Invalid chars
             */
            if (value === CHAR_ZERO_WIDTH_NOBREAK_SPACE || value === CHAR_NO_BREAK_SPACE) {
                continue;
            }
            /**
             * Escaped chars
             */
            if (value === CHAR_BACKSLASH) {
                push({ type: 'text', value: (options.keepEscaping ? value : '') + advance() });
                continue;
            }
            /**
             * Right square bracket (literal): ']'
             */
            if (value === CHAR_RIGHT_SQUARE_BRACKET) {
                push({ type: 'text', value: '\\' + value });
                continue;
            }
            /**
             * Left square bracket: '['
             */
            if (value === CHAR_LEFT_SQUARE_BRACKET) {
                brackets++;
                let closed = true;
                let next;
                while (index < length && (next = advance())) {
                    value += next;
                    if (next === CHAR_LEFT_SQUARE_BRACKET) {
                        brackets++;
                        continue;
                    }
                    if (next === CHAR_BACKSLASH) {
                        value += advance();
                        continue;
                    }
                    if (next === CHAR_RIGHT_SQUARE_BRACKET) {
                        brackets--;
                        if (brackets === 0) {
                            break;
                        }
                    }
                }
                push({ type: 'text', value });
                continue;
            }
            /**
             * Parentheses
             */
            if (value === CHAR_LEFT_PARENTHESES) {
                block = push({ type: 'paren', nodes: [] });
                stack.push(block);
                push({ type: 'text', value });
                continue;
            }
            if (value === CHAR_RIGHT_PARENTHESES) {
                if (block.type !== 'paren') {
                    push({ type: 'text', value });
                    continue;
                }
                block = stack.pop();
                push({ type: 'text', value });
                block = stack[stack.length - 1];
                continue;
            }
            /**
             * Quotes: '|"|`
             */
            if (value === CHAR_DOUBLE_QUOTE || value === CHAR_SINGLE_QUOTE || value === CHAR_BACKTICK) {
                let open = value;
                let next;
                if (options.keepQuotes !== true) {
                    value = '';
                }
                while (index < length && (next = advance())) {
                    if (next === CHAR_BACKSLASH) {
                        value += next + advance();
                        continue;
                    }
                    if (next === open) {
                        if (options.keepQuotes === true)
                            value += next;
                        break;
                    }
                    value += next;
                }
                push({ type: 'text', value });
                continue;
            }
            /**
             * Left curly brace: '{'
             */
            if (value === CHAR_LEFT_CURLY_BRACE) {
                depth++;
                let dollar = prev.value && prev.value.slice(-1) === '$' || block.dollar === true;
                let brace = {
                    type: 'brace',
                    open: true,
                    close: false,
                    dollar,
                    depth,
                    commas: 0,
                    ranges: 0,
                    nodes: []
                };
                block = push(brace);
                stack.push(block);
                push({ type: 'open', value });
                continue;
            }
            /**
             * Right curly brace: '}'
             */
            if (value === CHAR_RIGHT_CURLY_BRACE) {
                if (block.type !== 'brace') {
                    push({ type: 'text', value });
                    continue;
                }
                let type = 'close';
                block = stack.pop();
                block.close = true;
                push({ type, value });
                depth--;
                block = stack[stack.length - 1];
                continue;
            }
            /**
             * Comma: ','
             */
            if (value === CHAR_COMMA && depth > 0) {
                if (block.ranges > 0) {
                    block.ranges = 0;
                    let open = block.nodes.shift();
                    block.nodes = [open, { type: 'text', value: stringify(block) }];
                }
                push({ type: 'comma', value });
                block.commas++;
                continue;
            }
            /**
             * Dot: '.'
             */
            if (value === CHAR_DOT && depth > 0 && block.commas === 0) {
                let siblings = block.nodes;
                if (depth === 0 || siblings.length === 0) {
                    push({ type: 'text', value });
                    continue;
                }
                if (prev.type === 'dot') {
                    block.range = [];
                    prev.value += value;
                    prev.type = 'range';
                    if (block.nodes.length !== 3 && block.nodes.length !== 5) {
                        block.invalid = true;
                        block.ranges = 0;
                        prev.type = 'text';
                        continue;
                    }
                    block.ranges++;
                    block.args = [];
                    continue;
                }
                if (prev.type === 'range') {
                    siblings.pop();
                    let before = siblings[siblings.length - 1];
                    before.value += prev.value + value;
                    prev = before;
                    block.ranges--;
                    continue;
                }
                push({ type: 'dot', value });
                continue;
            }
            /**
             * Text
             */
            push({ type: 'text', value });
        }
        // Mark imbalanced braces and brackets as invalid
        do {
            block = stack.pop();
            if (block.type !== 'root') {
                block.nodes.forEach(node => {
                    if (!node.nodes) {
                        if (node.type === 'open')
                            node.isOpen = true;
                        if (node.type === 'close')
                            node.isClose = true;
                        if (!node.nodes)
                            node.type = 'text';
                        node.invalid = true;
                    }
                });
                // get the location of the block on parent.nodes (block's siblings)
                let parent = stack[stack.length - 1];
                let index = parent.nodes.indexOf(block);
                // replace the (invalid) block with it's nodes
                parent.nodes.splice(index, 1, ...block.nodes);
            }
        } while (stack.length > 0);
        push({ type: 'eos' });
        return ast;
    };
    module.exports = parse;
},
/* braces/lib/constants.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = {
        MAX_LENGTH: 1024 * 64,
        // Digits
        CHAR_0: '0',
        CHAR_9: '9',
        // Alphabet chars.
        CHAR_UPPERCASE_A: 'A',
        CHAR_LOWERCASE_A: 'a',
        CHAR_UPPERCASE_Z: 'Z',
        CHAR_LOWERCASE_Z: 'z',
        CHAR_LEFT_PARENTHESES: '(',
        CHAR_RIGHT_PARENTHESES: ')',
        CHAR_ASTERISK: '*',
        // Non-alphabetic chars.
        CHAR_AMPERSAND: '&',
        CHAR_AT: '@',
        CHAR_BACKSLASH: '\\',
        CHAR_BACKTICK: '`',
        CHAR_CARRIAGE_RETURN: '\r',
        CHAR_CIRCUMFLEX_ACCENT: '^',
        CHAR_COLON: ':',
        CHAR_COMMA: ',',
        CHAR_DOLLAR: '$',
        CHAR_DOT: '.',
        CHAR_DOUBLE_QUOTE: '"',
        CHAR_EQUAL: '=',
        CHAR_EXCLAMATION_MARK: '!',
        CHAR_FORM_FEED: '\f',
        CHAR_FORWARD_SLASH: '/',
        CHAR_HASH: '#',
        CHAR_HYPHEN_MINUS: '-',
        CHAR_LEFT_ANGLE_BRACKET: '<',
        CHAR_LEFT_CURLY_BRACE: '{',
        CHAR_LEFT_SQUARE_BRACKET: '[',
        CHAR_LINE_FEED: '\n',
        CHAR_NO_BREAK_SPACE: '\u00A0',
        CHAR_PERCENT: '%',
        CHAR_PLUS: '+',
        CHAR_QUESTION_MARK: '?',
        CHAR_RIGHT_ANGLE_BRACKET: '>',
        CHAR_RIGHT_CURLY_BRACE: '}',
        CHAR_RIGHT_SQUARE_BRACKET: ']',
        CHAR_SEMICOLON: ';',
        CHAR_SINGLE_QUOTE: '\'',
        CHAR_SPACE: ' ',
        CHAR_TAB: '\t',
        CHAR_UNDERSCORE: '_',
        CHAR_VERTICAL_LINE: '|',
        CHAR_ZERO_WIDTH_NOBREAK_SPACE: '\uFEFF' /* \uFEFF */
    };
},
/* picomatch/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = require(1144) /* ./lib/picomatch */;
},
/* picomatch/lib/picomatch.js */ function _(require, module, exports, __esModule, __esExport) {
    const path = require('path');
    const scan = require(1145) /* ./scan */;
    const parse = require(1148) /* ./parse */;
    const utils = require(1146) /* ./utils */;
    const constants = require(1147) /* ./constants */;
    const isObject = val => val && typeof val === 'object' && !Array.isArray(val);
    /**
     * Creates a matcher function from one or more glob patterns. The
     * returned function takes a string to match as its first argument,
     * and returns true if the string is a match. The returned matcher
     * function also takes a boolean as the second argument that, when true,
     * returns an object with additional information.
     *
     * ```js
     * const picomatch = require('picomatch');
     * // picomatch(glob[, options]);
     *
     * const isMatch = picomatch('*.!(*a)');
     * console.log(isMatch('a.a')); //=> false
     * console.log(isMatch('a.b')); //=> true
     * ```
     * @name picomatch
     * @param {String|Array} `globs` One or more glob patterns.
     * @param {Object=} `options`
     * @return {Function=} Returns a matcher function.
     * @api public
     */
    const picomatch = (glob, options, returnState = false) => {
        if (Array.isArray(glob)) {
            const fns = glob.map(input => picomatch(input, options, returnState));
            const arrayMatcher = str => {
                for (const isMatch of fns) {
                    const state = isMatch(str);
                    if (state)
                        return state;
                }
                return false;
            };
            return arrayMatcher;
        }
        const isState = isObject(glob) && glob.tokens && glob.input;
        if (glob === '' || (typeof glob !== 'string' && !isState)) {
            throw new TypeError('Expected pattern to be a non-empty string');
        }
        const opts = options || {};
        const posix = utils.isWindows(options);
        const regex = isState
            ? picomatch.compileRe(glob, options)
            : picomatch.makeRe(glob, options, false, true);
        const state = regex.state;
        delete regex.state;
        let isIgnored = () => false;
        if (opts.ignore) {
            const ignoreOpts = { ...options, ignore: null, onMatch: null, onResult: null };
            isIgnored = picomatch(opts.ignore, ignoreOpts, returnState);
        }
        const matcher = (input, returnObject = false) => {
            const { isMatch, match, output } = picomatch.test(input, regex, options, { glob, posix });
            const result = { glob, state, regex, posix, input, output, match, isMatch };
            if (typeof opts.onResult === 'function') {
                opts.onResult(result);
            }
            if (isMatch === false) {
                result.isMatch = false;
                return returnObject ? result : false;
            }
            if (isIgnored(input)) {
                if (typeof opts.onIgnore === 'function') {
                    opts.onIgnore(result);
                }
                result.isMatch = false;
                return returnObject ? result : false;
            }
            if (typeof opts.onMatch === 'function') {
                opts.onMatch(result);
            }
            return returnObject ? result : true;
        };
        if (returnState) {
            matcher.state = state;
        }
        return matcher;
    };
    /**
     * Test `input` with the given `regex`. This is used by the main
     * `picomatch()` function to test the input string.
     *
     * ```js
     * const picomatch = require('picomatch');
     * // picomatch.test(input, regex[, options]);
     *
     * console.log(picomatch.test('foo/bar', /^(?:([^/]*?)\/([^/]*?))$/));
     * // { isMatch: true, match: [ 'foo/', 'foo', 'bar' ], output: 'foo/bar' }
     * ```
     * @param {String} `input` String to test.
     * @param {RegExp} `regex`
     * @return {Object} Returns an object with matching info.
     * @api public
     */
    picomatch.test = (input, regex, options, { glob, posix } = {}) => {
        if (typeof input !== 'string') {
            throw new TypeError('Expected input to be a string');
        }
        if (input === '') {
            return { isMatch: false, output: '' };
        }
        const opts = options || {};
        const format = opts.format || (posix ? utils.toPosixSlashes : null);
        let match = input === glob;
        let output = (match && format) ? format(input) : input;
        if (match === false) {
            output = format ? format(input) : input;
            match = output === glob;
        }
        if (match === false || opts.capture === true) {
            if (opts.matchBase === true || opts.basename === true) {
                match = picomatch.matchBase(input, regex, options, posix);
            }
            else {
                match = regex.exec(output);
            }
        }
        return { isMatch: Boolean(match), match, output };
    };
    /**
     * Match the basename of a filepath.
     *
     * ```js
     * const picomatch = require('picomatch');
     * // picomatch.matchBase(input, glob[, options]);
     * console.log(picomatch.matchBase('foo/bar.js', '*.js'); // true
     * ```
     * @param {String} `input` String to test.
     * @param {RegExp|String} `glob` Glob pattern or regex created by [.makeRe](#makeRe).
     * @return {Boolean}
     * @api public
     */
    picomatch.matchBase = (input, glob, options, posix = utils.isWindows(options)) => {
        const regex = glob instanceof RegExp ? glob : picomatch.makeRe(glob, options);
        return regex.test(path.basename(input));
    };
    /**
     * Returns true if **any** of the given glob `patterns` match the specified `string`.
     *
     * ```js
     * const picomatch = require('picomatch');
     * // picomatch.isMatch(string, patterns[, options]);
     *
     * console.log(picomatch.isMatch('a.a', ['b.*', '*.a'])); //=> true
     * console.log(picomatch.isMatch('a.a', 'b.*')); //=> false
     * ```
     * @param {String|Array} str The string to test.
     * @param {String|Array} patterns One or more glob patterns to use for matching.
     * @param {Object} [options] See available [options](#options).
     * @return {Boolean} Returns true if any patterns match `str`
     * @api public
     */
    picomatch.isMatch = (str, patterns, options) => picomatch(patterns, options)(str);
    /**
     * Parse a glob pattern to create the source string for a regular
     * expression.
     *
     * ```js
     * const picomatch = require('picomatch');
     * const result = picomatch.parse(pattern[, options]);
     * ```
     * @param {String} `pattern`
     * @param {Object} `options`
     * @return {Object} Returns an object with useful properties and output to be used as a regex source string.
     * @api public
     */
    picomatch.parse = (pattern, options) => {
        if (Array.isArray(pattern))
            return pattern.map(p => picomatch.parse(p, options));
        return parse(pattern, { ...options, fastpaths: false });
    };
    /**
     * Scan a glob pattern to separate the pattern into segments.
     *
     * ```js
     * const picomatch = require('picomatch');
     * // picomatch.scan(input[, options]);
     *
     * const result = picomatch.scan('!./foo/*.js');
     * console.log(result);
     * { prefix: '!./',
     *   input: '!./foo/*.js',
     *   start: 3,
     *   base: 'foo',
     *   glob: '*.js',
     *   isBrace: false,
     *   isBracket: false,
     *   isGlob: true,
     *   isExtglob: false,
     *   isGlobstar: false,
     *   negated: true }
     * ```
     * @param {String} `input` Glob pattern to scan.
     * @param {Object} `options`
     * @return {Object} Returns an object with
     * @api public
     */
    picomatch.scan = (input, options) => scan(input, options);
    /**
     * Compile a regular expression from the `state` object returned by the
     * [parse()](#parse) method.
     *
     * @param {Object} `state`
     * @param {Object} `options`
     * @param {Boolean} `returnOutput` Intended for implementors, this argument allows you to return the raw output from the parser.
     * @param {Boolean} `returnState` Adds the state to a `state` property on the returned regex. Useful for implementors and debugging.
     * @return {RegExp}
     * @api public
     */
    picomatch.compileRe = (state, options, returnOutput = false, returnState = false) => {
        if (returnOutput === true) {
            return state.output;
        }
        const opts = options || {};
        const prepend = opts.contains ? '' : '^';
        const append = opts.contains ? '' : '$';
        let source = `${prepend}(?:${state.output})${append}`;
        if (state && state.negated === true) {
            source = `^(?!${source}).*$`;
        }
        const regex = picomatch.toRegex(source, options);
        if (returnState === true) {
            regex.state = state;
        }
        return regex;
    };
    /**
     * Create a regular expression from a parsed glob pattern.
     *
     * ```js
     * const picomatch = require('picomatch');
     * const state = picomatch.parse('*.js');
     * // picomatch.compileRe(state[, options]);
     *
     * console.log(picomatch.compileRe(state));
     * //=> /^(?:(?!\.)(?=.)[^/]*?\.js)$/
     * ```
     * @param {String} `state` The object returned from the `.parse` method.
     * @param {Object} `options`
     * @param {Boolean} `returnOutput` Implementors may use this argument to return the compiled output, instead of a regular expression. This is not exposed on the options to prevent end-users from mutating the result.
     * @param {Boolean} `returnState` Implementors may use this argument to return the state from the parsed glob with the returned regular expression.
     * @return {RegExp} Returns a regex created from the given pattern.
     * @api public
     */
    picomatch.makeRe = (input, options = {}, returnOutput = false, returnState = false) => {
        if (!input || typeof input !== 'string') {
            throw new TypeError('Expected a non-empty string');
        }
        let parsed = { negated: false, fastpaths: true };
        if (options.fastpaths !== false && (input[0] === '.' || input[0] === '*')) {
            parsed.output = parse.fastpaths(input, options);
        }
        if (!parsed.output) {
            parsed = parse(input, options);
        }
        return picomatch.compileRe(parsed, options, returnOutput, returnState);
    };
    /**
     * Create a regular expression from the given regex source string.
     *
     * ```js
     * const picomatch = require('picomatch');
     * // picomatch.toRegex(source[, options]);
     *
     * const { output } = picomatch.parse('*.js');
     * console.log(picomatch.toRegex(output));
     * //=> /^(?:(?!\.)(?=.)[^/]*?\.js)$/
     * ```
     * @param {String} `source` Regular expression source string.
     * @param {Object} `options`
     * @return {RegExp}
     * @api public
     */
    picomatch.toRegex = (source, options) => {
        try {
            const opts = options || {};
            return new RegExp(source, opts.flags || (opts.nocase ? 'i' : ''));
        }
        catch (err) {
            if (options && options.debug === true)
                throw err;
            return /$^/;
        }
    };
    /**
     * Picomatch constants.
     * @return {Object}
     */
    picomatch.constants = constants;
    /**
     * Expose "picomatch"
     */
    module.exports = picomatch;
},
/* picomatch/lib/scan.js */ function _(require, module, exports, __esModule, __esExport) {
    const utils = require(1146) /* ./utils */;
    const { CHAR_ASTERISK, /* * */ CHAR_AT, /* @ */ CHAR_BACKWARD_SLASH, /* \ */ CHAR_COMMA, /* , */ CHAR_DOT, /* . */ CHAR_EXCLAMATION_MARK, /* ! */ CHAR_FORWARD_SLASH, /* / */ CHAR_LEFT_CURLY_BRACE, /* { */ CHAR_LEFT_PARENTHESES, /* ( */ CHAR_LEFT_SQUARE_BRACKET, /* [ */ CHAR_PLUS, /* + */ CHAR_QUESTION_MARK, /* ? */ CHAR_RIGHT_CURLY_BRACE, /* } */ CHAR_RIGHT_PARENTHESES, /* ) */ CHAR_RIGHT_SQUARE_BRACKET /* ] */ } = require(1147) /* ./constants */;
    const isPathSeparator = code => {
        return code === CHAR_FORWARD_SLASH || code === CHAR_BACKWARD_SLASH;
    };
    const depth = token => {
        if (token.isPrefix !== true) {
            token.depth = token.isGlobstar ? Infinity : 1;
        }
    };
    /**
     * Quickly scans a glob pattern and returns an object with a handful of
     * useful properties, like `isGlob`, `path` (the leading non-glob, if it exists),
     * `glob` (the actual pattern), `negated` (true if the path starts with `!` but not
     * with `!(`) and `negatedExtglob` (true if the path starts with `!(`).
     *
     * ```js
     * const pm = require('picomatch');
     * console.log(pm.scan('foo/bar/*.js'));
     * { isGlob: true, input: 'foo/bar/*.js', base: 'foo/bar', glob: '*.js' }
     * ```
     * @param {String} `str`
     * @param {Object} `options`
     * @return {Object} Returns an object with tokens and regex source string.
     * @api public
     */
    const scan = (input, options) => {
        const opts = options || {};
        const length = input.length - 1;
        const scanToEnd = opts.parts === true || opts.scanToEnd === true;
        const slashes = [];
        const tokens = [];
        const parts = [];
        let str = input;
        let index = -1;
        let start = 0;
        let lastIndex = 0;
        let isBrace = false;
        let isBracket = false;
        let isGlob = false;
        let isExtglob = false;
        let isGlobstar = false;
        let braceEscaped = false;
        let backslashes = false;
        let negated = false;
        let negatedExtglob = false;
        let finished = false;
        let braces = 0;
        let prev;
        let code;
        let token = { value: '', depth: 0, isGlob: false };
        const eos = () => index >= length;
        const peek = () => str.charCodeAt(index + 1);
        const advance = () => {
            prev = code;
            return str.charCodeAt(++index);
        };
        while (index < length) {
            code = advance();
            let next;
            if (code === CHAR_BACKWARD_SLASH) {
                backslashes = token.backslashes = true;
                code = advance();
                if (code === CHAR_LEFT_CURLY_BRACE) {
                    braceEscaped = true;
                }
                continue;
            }
            if (braceEscaped === true || code === CHAR_LEFT_CURLY_BRACE) {
                braces++;
                while (eos() !== true && (code = advance())) {
                    if (code === CHAR_BACKWARD_SLASH) {
                        backslashes = token.backslashes = true;
                        advance();
                        continue;
                    }
                    if (code === CHAR_LEFT_CURLY_BRACE) {
                        braces++;
                        continue;
                    }
                    if (braceEscaped !== true && code === CHAR_DOT && (code = advance()) === CHAR_DOT) {
                        isBrace = token.isBrace = true;
                        isGlob = token.isGlob = true;
                        finished = true;
                        if (scanToEnd === true) {
                            continue;
                        }
                        break;
                    }
                    if (braceEscaped !== true && code === CHAR_COMMA) {
                        isBrace = token.isBrace = true;
                        isGlob = token.isGlob = true;
                        finished = true;
                        if (scanToEnd === true) {
                            continue;
                        }
                        break;
                    }
                    if (code === CHAR_RIGHT_CURLY_BRACE) {
                        braces--;
                        if (braces === 0) {
                            braceEscaped = false;
                            isBrace = token.isBrace = true;
                            finished = true;
                            break;
                        }
                    }
                }
                if (scanToEnd === true) {
                    continue;
                }
                break;
            }
            if (code === CHAR_FORWARD_SLASH) {
                slashes.push(index);
                tokens.push(token);
                token = { value: '', depth: 0, isGlob: false };
                if (finished === true)
                    continue;
                if (prev === CHAR_DOT && index === (start + 1)) {
                    start += 2;
                    continue;
                }
                lastIndex = index + 1;
                continue;
            }
            if (opts.noext !== true) {
                const isExtglobChar = code === CHAR_PLUS
                    || code === CHAR_AT
                    || code === CHAR_ASTERISK
                    || code === CHAR_QUESTION_MARK
                    || code === CHAR_EXCLAMATION_MARK;
                if (isExtglobChar === true && peek() === CHAR_LEFT_PARENTHESES) {
                    isGlob = token.isGlob = true;
                    isExtglob = token.isExtglob = true;
                    finished = true;
                    if (code === CHAR_EXCLAMATION_MARK && index === start) {
                        negatedExtglob = true;
                    }
                    if (scanToEnd === true) {
                        while (eos() !== true && (code = advance())) {
                            if (code === CHAR_BACKWARD_SLASH) {
                                backslashes = token.backslashes = true;
                                code = advance();
                                continue;
                            }
                            if (code === CHAR_RIGHT_PARENTHESES) {
                                isGlob = token.isGlob = true;
                                finished = true;
                                break;
                            }
                        }
                        continue;
                    }
                    break;
                }
            }
            if (code === CHAR_ASTERISK) {
                if (prev === CHAR_ASTERISK)
                    isGlobstar = token.isGlobstar = true;
                isGlob = token.isGlob = true;
                finished = true;
                if (scanToEnd === true) {
                    continue;
                }
                break;
            }
            if (code === CHAR_QUESTION_MARK) {
                isGlob = token.isGlob = true;
                finished = true;
                if (scanToEnd === true) {
                    continue;
                }
                break;
            }
            if (code === CHAR_LEFT_SQUARE_BRACKET) {
                while (eos() !== true && (next = advance())) {
                    if (next === CHAR_BACKWARD_SLASH) {
                        backslashes = token.backslashes = true;
                        advance();
                        continue;
                    }
                    if (next === CHAR_RIGHT_SQUARE_BRACKET) {
                        isBracket = token.isBracket = true;
                        isGlob = token.isGlob = true;
                        finished = true;
                        break;
                    }
                }
                if (scanToEnd === true) {
                    continue;
                }
                break;
            }
            if (opts.nonegate !== true && code === CHAR_EXCLAMATION_MARK && index === start) {
                negated = token.negated = true;
                start++;
                continue;
            }
            if (opts.noparen !== true && code === CHAR_LEFT_PARENTHESES) {
                isGlob = token.isGlob = true;
                if (scanToEnd === true) {
                    while (eos() !== true && (code = advance())) {
                        if (code === CHAR_LEFT_PARENTHESES) {
                            backslashes = token.backslashes = true;
                            code = advance();
                            continue;
                        }
                        if (code === CHAR_RIGHT_PARENTHESES) {
                            finished = true;
                            break;
                        }
                    }
                    continue;
                }
                break;
            }
            if (isGlob === true) {
                finished = true;
                if (scanToEnd === true) {
                    continue;
                }
                break;
            }
        }
        if (opts.noext === true) {
            isExtglob = false;
            isGlob = false;
        }
        let base = str;
        let prefix = '';
        let glob = '';
        if (start > 0) {
            prefix = str.slice(0, start);
            str = str.slice(start);
            lastIndex -= start;
        }
        if (base && isGlob === true && lastIndex > 0) {
            base = str.slice(0, lastIndex);
            glob = str.slice(lastIndex);
        }
        else if (isGlob === true) {
            base = '';
            glob = str;
        }
        else {
            base = str;
        }
        if (base && base !== '' && base !== '/' && base !== str) {
            if (isPathSeparator(base.charCodeAt(base.length - 1))) {
                base = base.slice(0, -1);
            }
        }
        if (opts.unescape === true) {
            if (glob)
                glob = utils.removeBackslashes(glob);
            if (base && backslashes === true) {
                base = utils.removeBackslashes(base);
            }
        }
        const state = {
            prefix,
            input,
            start,
            base,
            glob,
            isBrace,
            isBracket,
            isGlob,
            isExtglob,
            isGlobstar,
            negated,
            negatedExtglob
        };
        if (opts.tokens === true) {
            state.maxDepth = 0;
            if (!isPathSeparator(code)) {
                tokens.push(token);
            }
            state.tokens = tokens;
        }
        if (opts.parts === true || opts.tokens === true) {
            let prevIndex;
            for (let idx = 0; idx < slashes.length; idx++) {
                const n = prevIndex ? prevIndex + 1 : start;
                const i = slashes[idx];
                const value = input.slice(n, i);
                if (opts.tokens) {
                    if (idx === 0 && start !== 0) {
                        tokens[idx].isPrefix = true;
                        tokens[idx].value = prefix;
                    }
                    else {
                        tokens[idx].value = value;
                    }
                    depth(tokens[idx]);
                    state.maxDepth += tokens[idx].depth;
                }
                if (idx !== 0 || value !== '') {
                    parts.push(value);
                }
                prevIndex = i;
            }
            if (prevIndex && prevIndex + 1 < input.length) {
                const value = input.slice(prevIndex + 1);
                parts.push(value);
                if (opts.tokens) {
                    tokens[tokens.length - 1].value = value;
                    depth(tokens[tokens.length - 1]);
                    state.maxDepth += tokens[tokens.length - 1].depth;
                }
            }
            state.slashes = slashes;
            state.parts = parts;
        }
        return state;
    };
    module.exports = scan;
},
/* picomatch/lib/utils.js */ function _(require, module, exports, __esModule, __esExport) {
    const path = require('path');
    const win32 = process.platform === 'win32';
    const { REGEX_BACKSLASH, REGEX_REMOVE_BACKSLASH, REGEX_SPECIAL_CHARS, REGEX_SPECIAL_CHARS_GLOBAL } = require(1147) /* ./constants */;
    exports.isObject = val => val !== null && typeof val === 'object' && !Array.isArray(val);
    exports.hasRegexChars = str => REGEX_SPECIAL_CHARS.test(str);
    exports.isRegexChar = str => str.length === 1 && exports.hasRegexChars(str);
    exports.escapeRegex = str => str.replace(REGEX_SPECIAL_CHARS_GLOBAL, '\\$1');
    exports.toPosixSlashes = str => str.replace(REGEX_BACKSLASH, '/');
    exports.removeBackslashes = str => {
        return str.replace(REGEX_REMOVE_BACKSLASH, match => {
            return match === '\\' ? '' : match;
        });
    };
    exports.supportsLookbehinds = () => {
        const segs = process.version.slice(1).split('.').map(Number);
        if (segs.length === 3 && segs[0] >= 9 || (segs[0] === 8 && segs[1] >= 10)) {
            return true;
        }
        return false;
    };
    exports.isWindows = options => {
        if (options && typeof options.windows === 'boolean') {
            return options.windows;
        }
        return win32 === true || path.sep === '\\';
    };
    exports.escapeLast = (input, char, lastIdx) => {
        const idx = input.lastIndexOf(char, lastIdx);
        if (idx === -1)
            return input;
        if (input[idx - 1] === '\\')
            return exports.escapeLast(input, char, idx - 1);
        return `${input.slice(0, idx)}\\${input.slice(idx)}`;
    };
    exports.removePrefix = (input, state = {}) => {
        let output = input;
        if (output.startsWith('./')) {
            output = output.slice(2);
            state.prefix = './';
        }
        return output;
    };
    exports.wrapOutput = (input, state = {}, options = {}) => {
        const prepend = options.contains ? '' : '^';
        const append = options.contains ? '' : '$';
        let output = `${prepend}(?:${input})${append}`;
        if (state.negated === true) {
            output = `(?:^(?!${output}).*$)`;
        }
        return output;
    };
},
/* picomatch/lib/constants.js */ function _(require, module, exports, __esModule, __esExport) {
    const path = require('path');
    const WIN_SLASH = '\\\\/';
    const WIN_NO_SLASH = `[^${WIN_SLASH}]`;
    /**
     * Posix glob regex
     */
    const DOT_LITERAL = '\\.';
    const PLUS_LITERAL = '\\+';
    const QMARK_LITERAL = '\\?';
    const SLASH_LITERAL = '\\/';
    const ONE_CHAR = '(?=.)';
    const QMARK = '[^/]';
    const END_ANCHOR = `(?:${SLASH_LITERAL}|$)`;
    const START_ANCHOR = `(?:^|${SLASH_LITERAL})`;
    const DOTS_SLASH = `${DOT_LITERAL}{1,2}${END_ANCHOR}`;
    const NO_DOT = `(?!${DOT_LITERAL})`;
    const NO_DOTS = `(?!${START_ANCHOR}${DOTS_SLASH})`;
    const NO_DOT_SLASH = `(?!${DOT_LITERAL}{0,1}${END_ANCHOR})`;
    const NO_DOTS_SLASH = `(?!${DOTS_SLASH})`;
    const QMARK_NO_DOT = `[^.${SLASH_LITERAL}]`;
    const STAR = `${QMARK}*?`;
    const POSIX_CHARS = {
        DOT_LITERAL,
        PLUS_LITERAL,
        QMARK_LITERAL,
        SLASH_LITERAL,
        ONE_CHAR,
        QMARK,
        END_ANCHOR,
        DOTS_SLASH,
        NO_DOT,
        NO_DOTS,
        NO_DOT_SLASH,
        NO_DOTS_SLASH,
        QMARK_NO_DOT,
        STAR,
        START_ANCHOR
    };
    /**
     * Windows glob regex
     */
    const WINDOWS_CHARS = {
        ...POSIX_CHARS,
        SLASH_LITERAL: `[${WIN_SLASH}]`,
        QMARK: WIN_NO_SLASH,
        STAR: `${WIN_NO_SLASH}*?`,
        DOTS_SLASH: `${DOT_LITERAL}{1,2}(?:[${WIN_SLASH}]|$)`,
        NO_DOT: `(?!${DOT_LITERAL})`,
        NO_DOTS: `(?!(?:^|[${WIN_SLASH}])${DOT_LITERAL}{1,2}(?:[${WIN_SLASH}]|$))`,
        NO_DOT_SLASH: `(?!${DOT_LITERAL}{0,1}(?:[${WIN_SLASH}]|$))`,
        NO_DOTS_SLASH: `(?!${DOT_LITERAL}{1,2}(?:[${WIN_SLASH}]|$))`,
        QMARK_NO_DOT: `[^.${WIN_SLASH}]`,
        START_ANCHOR: `(?:^|[${WIN_SLASH}])`,
        END_ANCHOR: `(?:[${WIN_SLASH}]|$)`
    };
    /**
     * POSIX Bracket Regex
     */
    const POSIX_REGEX_SOURCE = {
        alnum: 'a-zA-Z0-9',
        alpha: 'a-zA-Z',
        ascii: '\\x00-\\x7F',
        blank: ' \\t',
        cntrl: '\\x00-\\x1F\\x7F',
        digit: '0-9',
        graph: '\\x21-\\x7E',
        lower: 'a-z',
        print: '\\x20-\\x7E ',
        punct: '\\-!"#$%&\'()\\*+,./:;<=>?@[\\]^_`{|}~',
        space: ' \\t\\r\\n\\v\\f',
        upper: 'A-Z',
        word: 'A-Za-z0-9_',
        xdigit: 'A-Fa-f0-9'
    };
    module.exports = {
        MAX_LENGTH: 1024 * 64,
        POSIX_REGEX_SOURCE,
        // regular expressions
        REGEX_BACKSLASH: /\\(?![*+?^${}(|)[\]])/g,
        REGEX_NON_SPECIAL_CHARS: /^[^@![\].,$*+?^{}()|\\/]+/,
        REGEX_SPECIAL_CHARS: /[-*+?.^${}(|)[\]]/,
        REGEX_SPECIAL_CHARS_BACKREF: /(\\?)((\W)(\3*))/g,
        REGEX_SPECIAL_CHARS_GLOBAL: /([-*+?.^${}(|)[\]])/g,
        REGEX_REMOVE_BACKSLASH: /(?:\[.*?[^\\]\]|\\(?=.))/g,
        // Replace globs with equivalent patterns to reduce parsing time.
        REPLACEMENTS: {
            '***': '*',
            '**/**': '**',
            '**/**/**': '**'
        },
        // Digits
        CHAR_0: 48,
        CHAR_9: 57,
        // Alphabet chars.
        CHAR_UPPERCASE_A: 65,
        CHAR_LOWERCASE_A: 97,
        CHAR_UPPERCASE_Z: 90,
        CHAR_LOWERCASE_Z: 122,
        CHAR_LEFT_PARENTHESES: 40,
        CHAR_RIGHT_PARENTHESES: 41,
        CHAR_ASTERISK: 42,
        // Non-alphabetic chars.
        CHAR_AMPERSAND: 38,
        CHAR_AT: 64,
        CHAR_BACKWARD_SLASH: 92,
        CHAR_CARRIAGE_RETURN: 13,
        CHAR_CIRCUMFLEX_ACCENT: 94,
        CHAR_COLON: 58,
        CHAR_COMMA: 44,
        CHAR_DOT: 46,
        CHAR_DOUBLE_QUOTE: 34,
        CHAR_EQUAL: 61,
        CHAR_EXCLAMATION_MARK: 33,
        CHAR_FORM_FEED: 12,
        CHAR_FORWARD_SLASH: 47,
        CHAR_GRAVE_ACCENT: 96,
        CHAR_HASH: 35,
        CHAR_HYPHEN_MINUS: 45,
        CHAR_LEFT_ANGLE_BRACKET: 60,
        CHAR_LEFT_CURLY_BRACE: 123,
        CHAR_LEFT_SQUARE_BRACKET: 91,
        CHAR_LINE_FEED: 10,
        CHAR_NO_BREAK_SPACE: 160,
        CHAR_PERCENT: 37,
        CHAR_PLUS: 43,
        CHAR_QUESTION_MARK: 63,
        CHAR_RIGHT_ANGLE_BRACKET: 62,
        CHAR_RIGHT_CURLY_BRACE: 125,
        CHAR_RIGHT_SQUARE_BRACKET: 93,
        CHAR_SEMICOLON: 59,
        CHAR_SINGLE_QUOTE: 39,
        CHAR_SPACE: 32,
        CHAR_TAB: 9,
        CHAR_UNDERSCORE: 95,
        CHAR_VERTICAL_LINE: 124,
        CHAR_ZERO_WIDTH_NOBREAK_SPACE: 65279,
        SEP: path.sep,
        /**
         * Create EXTGLOB_CHARS
         */
        extglobChars(chars) {
            return {
                '!': { type: 'negate', open: '(?:(?!(?:', close: `))${chars.STAR})` },
                '?': { type: 'qmark', open: '(?:', close: ')?' },
                '+': { type: 'plus', open: '(?:', close: ')+' },
                '*': { type: 'star', open: '(?:', close: ')*' },
                '@': { type: 'at', open: '(?:', close: ')' }
            };
        },
        /**
         * Create GLOB_CHARS
         */
        globChars(win32) {
            return win32 === true ? WINDOWS_CHARS : POSIX_CHARS;
        }
    };
},
/* picomatch/lib/parse.js */ function _(require, module, exports, __esModule, __esExport) {
    const constants = require(1147) /* ./constants */;
    const utils = require(1146) /* ./utils */;
    /**
     * Constants
     */
    const { MAX_LENGTH, POSIX_REGEX_SOURCE, REGEX_NON_SPECIAL_CHARS, REGEX_SPECIAL_CHARS_BACKREF, REPLACEMENTS } = constants;
    /**
     * Helpers
     */
    const expandRange = (args, options) => {
        if (typeof options.expandRange === 'function') {
            return options.expandRange(...args, options);
        }
        args.sort();
        const value = `[${args.join('-')}]`;
        try {
            /* eslint-disable-next-line no-new */
            new RegExp(value);
        }
        catch (ex) {
            return args.map(v => utils.escapeRegex(v)).join('..');
        }
        return value;
    };
    /**
     * Create the message for a syntax error
     */
    const syntaxError = (type, char) => {
        return `Missing ${type}: "${char}" - use "\\\\${char}" to match literal characters`;
    };
    /**
     * Parse the given input string.
     * @param {String} input
     * @param {Object} options
     * @return {Object}
     */
    const parse = (input, options) => {
        if (typeof input !== 'string') {
            throw new TypeError('Expected a string');
        }
        input = REPLACEMENTS[input] || input;
        const opts = { ...options };
        const max = typeof opts.maxLength === 'number' ? Math.min(MAX_LENGTH, opts.maxLength) : MAX_LENGTH;
        let len = input.length;
        if (len > max) {
            throw new SyntaxError(`Input length: ${len}, exceeds maximum allowed length: ${max}`);
        }
        const bos = { type: 'bos', value: '', output: opts.prepend || '' };
        const tokens = [bos];
        const capture = opts.capture ? '' : '?:';
        const win32 = utils.isWindows(options);
        // create constants based on platform, for windows or posix
        const PLATFORM_CHARS = constants.globChars(win32);
        const EXTGLOB_CHARS = constants.extglobChars(PLATFORM_CHARS);
        const { DOT_LITERAL, PLUS_LITERAL, SLASH_LITERAL, ONE_CHAR, DOTS_SLASH, NO_DOT, NO_DOT_SLASH, NO_DOTS_SLASH, QMARK, QMARK_NO_DOT, STAR, START_ANCHOR } = PLATFORM_CHARS;
        const globstar = opts => {
            return `(${capture}(?:(?!${START_ANCHOR}${opts.dot ? DOTS_SLASH : DOT_LITERAL}).)*?)`;
        };
        const nodot = opts.dot ? '' : NO_DOT;
        const qmarkNoDot = opts.dot ? QMARK : QMARK_NO_DOT;
        let star = opts.bash === true ? globstar(opts) : STAR;
        if (opts.capture) {
            star = `(${star})`;
        }
        // minimatch options support
        if (typeof opts.noext === 'boolean') {
            opts.noextglob = opts.noext;
        }
        const state = {
            input,
            index: -1,
            start: 0,
            dot: opts.dot === true,
            consumed: '',
            output: '',
            prefix: '',
            backtrack: false,
            negated: false,
            brackets: 0,
            braces: 0,
            parens: 0,
            quotes: 0,
            globstar: false,
            tokens
        };
        input = utils.removePrefix(input, state);
        len = input.length;
        const extglobs = [];
        const braces = [];
        const stack = [];
        let prev = bos;
        let value;
        /**
         * Tokenizing helpers
         */
        const eos = () => state.index === len - 1;
        const peek = state.peek = (n = 1) => input[state.index + n];
        const advance = state.advance = () => input[++state.index] || '';
        const remaining = () => input.slice(state.index + 1);
        const consume = (value = '', num = 0) => {
            state.consumed += value;
            state.index += num;
        };
        const append = token => {
            state.output += token.output != null ? token.output : token.value;
            consume(token.value);
        };
        const negate = () => {
            let count = 1;
            while (peek() === '!' && (peek(2) !== '(' || peek(3) === '?')) {
                advance();
                state.start++;
                count++;
            }
            if (count % 2 === 0) {
                return false;
            }
            state.negated = true;
            state.start++;
            return true;
        };
        const increment = type => {
            state[type]++;
            stack.push(type);
        };
        const decrement = type => {
            state[type]--;
            stack.pop();
        };
        /**
         * Push tokens onto the tokens array. This helper speeds up
         * tokenizing by 1) helping us avoid backtracking as much as possible,
         * and 2) helping us avoid creating extra tokens when consecutive
         * characters are plain text. This improves performance and simplifies
         * lookbehinds.
         */
        const push = tok => {
            if (prev.type === 'globstar') {
                const isBrace = state.braces > 0 && (tok.type === 'comma' || tok.type === 'brace');
                const isExtglob = tok.extglob === true || (extglobs.length && (tok.type === 'pipe' || tok.type === 'paren'));
                if (tok.type !== 'slash' && tok.type !== 'paren' && !isBrace && !isExtglob) {
                    state.output = state.output.slice(0, -prev.output.length);
                    prev.type = 'star';
                    prev.value = '*';
                    prev.output = star;
                    state.output += prev.output;
                }
            }
            if (extglobs.length && tok.type !== 'paren') {
                extglobs[extglobs.length - 1].inner += tok.value;
            }
            if (tok.value || tok.output)
                append(tok);
            if (prev && prev.type === 'text' && tok.type === 'text') {
                prev.value += tok.value;
                prev.output = (prev.output || '') + tok.value;
                return;
            }
            tok.prev = prev;
            tokens.push(tok);
            prev = tok;
        };
        const extglobOpen = (type, value) => {
            const token = { ...EXTGLOB_CHARS[value], conditions: 1, inner: '' };
            token.prev = prev;
            token.parens = state.parens;
            token.output = state.output;
            const output = (opts.capture ? '(' : '') + token.open;
            increment('parens');
            push({ type, value, output: state.output ? '' : ONE_CHAR });
            push({ type: 'paren', extglob: true, value: advance(), output });
            extglobs.push(token);
        };
        const extglobClose = token => {
            let output = token.close + (opts.capture ? ')' : '');
            let rest;
            if (token.type === 'negate') {
                let extglobStar = star;
                if (token.inner && token.inner.length > 1 && token.inner.includes('/')) {
                    extglobStar = globstar(opts);
                }
                if (extglobStar !== star || eos() || /^\)+$/.test(remaining())) {
                    output = token.close = `)$))${extglobStar}`;
                }
                if (token.inner.includes('*') && (rest = remaining()) && /^\.[^\\/.]+$/.test(rest)) {
                    output = token.close = `)${rest})${extglobStar})`;
                }
                if (token.prev.type === 'bos') {
                    state.negatedExtglob = true;
                }
            }
            push({ type: 'paren', extglob: true, value, output });
            decrement('parens');
        };
        /**
         * Fast paths
         */
        if (opts.fastpaths !== false && !/(^[*!]|[/()[\]{}"])/.test(input)) {
            let backslashes = false;
            let output = input.replace(REGEX_SPECIAL_CHARS_BACKREF, (m, esc, chars, first, rest, index) => {
                if (first === '\\') {
                    backslashes = true;
                    return m;
                }
                if (first === '?') {
                    if (esc) {
                        return esc + first + (rest ? QMARK.repeat(rest.length) : '');
                    }
                    if (index === 0) {
                        return qmarkNoDot + (rest ? QMARK.repeat(rest.length) : '');
                    }
                    return QMARK.repeat(chars.length);
                }
                if (first === '.') {
                    return DOT_LITERAL.repeat(chars.length);
                }
                if (first === '*') {
                    if (esc) {
                        return esc + first + (rest ? star : '');
                    }
                    return star;
                }
                return esc ? m : `\\${m}`;
            });
            if (backslashes === true) {
                if (opts.unescape === true) {
                    output = output.replace(/\\/g, '');
                }
                else {
                    output = output.replace(/\\+/g, m => {
                        return m.length % 2 === 0 ? '\\\\' : (m ? '\\' : '');
                    });
                }
            }
            if (output === input && opts.contains === true) {
                state.output = input;
                return state;
            }
            state.output = utils.wrapOutput(output, state, options);
            return state;
        }
        /**
         * Tokenize input until we reach end-of-string
         */
        while (!eos()) {
            value = advance();
            if (value === '\u0000') {
                continue;
            }
            /**
             * Escaped characters
             */
            if (value === '\\') {
                const next = peek();
                if (next === '/' && opts.bash !== true) {
                    continue;
                }
                if (next === '.' || next === ';') {
                    continue;
                }
                if (!next) {
                    value += '\\';
                    push({ type: 'text', value });
                    continue;
                }
                // collapse slashes to reduce potential for exploits
                const match = /^\\+/.exec(remaining());
                let slashes = 0;
                if (match && match[0].length > 2) {
                    slashes = match[0].length;
                    state.index += slashes;
                    if (slashes % 2 !== 0) {
                        value += '\\';
                    }
                }
                if (opts.unescape === true) {
                    value = advance();
                }
                else {
                    value += advance();
                }
                if (state.brackets === 0) {
                    push({ type: 'text', value });
                    continue;
                }
            }
            /**
             * If we're inside a regex character class, continue
             * until we reach the closing bracket.
             */
            if (state.brackets > 0 && (value !== ']' || prev.value === '[' || prev.value === '[^')) {
                if (opts.posix !== false && value === ':') {
                    const inner = prev.value.slice(1);
                    if (inner.includes('[')) {
                        prev.posix = true;
                        if (inner.includes(':')) {
                            const idx = prev.value.lastIndexOf('[');
                            const pre = prev.value.slice(0, idx);
                            const rest = prev.value.slice(idx + 2);
                            const posix = POSIX_REGEX_SOURCE[rest];
                            if (posix) {
                                prev.value = pre + posix;
                                state.backtrack = true;
                                advance();
                                if (!bos.output && tokens.indexOf(prev) === 1) {
                                    bos.output = ONE_CHAR;
                                }
                                continue;
                            }
                        }
                    }
                }
                if ((value === '[' && peek() !== ':') || (value === '-' && peek() === ']')) {
                    value = `\\${value}`;
                }
                if (value === ']' && (prev.value === '[' || prev.value === '[^')) {
                    value = `\\${value}`;
                }
                if (opts.posix === true && value === '!' && prev.value === '[') {
                    value = '^';
                }
                prev.value += value;
                append({ value });
                continue;
            }
            /**
             * If we're inside a quoted string, continue
             * until we reach the closing double quote.
             */
            if (state.quotes === 1 && value !== '"') {
                value = utils.escapeRegex(value);
                prev.value += value;
                append({ value });
                continue;
            }
            /**
             * Double quotes
             */
            if (value === '"') {
                state.quotes = state.quotes === 1 ? 0 : 1;
                if (opts.keepQuotes === true) {
                    push({ type: 'text', value });
                }
                continue;
            }
            /**
             * Parentheses
             */
            if (value === '(') {
                increment('parens');
                push({ type: 'paren', value });
                continue;
            }
            if (value === ')') {
                if (state.parens === 0 && opts.strictBrackets === true) {
                    throw new SyntaxError(syntaxError('opening', '('));
                }
                const extglob = extglobs[extglobs.length - 1];
                if (extglob && state.parens === extglob.parens + 1) {
                    extglobClose(extglobs.pop());
                    continue;
                }
                push({ type: 'paren', value, output: state.parens ? ')' : '\\)' });
                decrement('parens');
                continue;
            }
            /**
             * Square brackets
             */
            if (value === '[') {
                if (opts.nobracket === true || !remaining().includes(']')) {
                    if (opts.nobracket !== true && opts.strictBrackets === true) {
                        throw new SyntaxError(syntaxError('closing', ']'));
                    }
                    value = `\\${value}`;
                }
                else {
                    increment('brackets');
                }
                push({ type: 'bracket', value });
                continue;
            }
            if (value === ']') {
                if (opts.nobracket === true || (prev && prev.type === 'bracket' && prev.value.length === 1)) {
                    push({ type: 'text', value, output: `\\${value}` });
                    continue;
                }
                if (state.brackets === 0) {
                    if (opts.strictBrackets === true) {
                        throw new SyntaxError(syntaxError('opening', '['));
                    }
                    push({ type: 'text', value, output: `\\${value}` });
                    continue;
                }
                decrement('brackets');
                const prevValue = prev.value.slice(1);
                if (prev.posix !== true && prevValue[0] === '^' && !prevValue.includes('/')) {
                    value = `/${value}`;
                }
                prev.value += value;
                append({ value });
                // when literal brackets are explicitly disabled
                // assume we should match with a regex character class
                if (opts.literalBrackets === false || utils.hasRegexChars(prevValue)) {
                    continue;
                }
                const escaped = utils.escapeRegex(prev.value);
                state.output = state.output.slice(0, -prev.value.length);
                // when literal brackets are explicitly enabled
                // assume we should escape the brackets to match literal characters
                if (opts.literalBrackets === true) {
                    state.output += escaped;
                    prev.value = escaped;
                    continue;
                }
                // when the user specifies nothing, try to match both
                prev.value = `(${capture}${escaped}|${prev.value})`;
                state.output += prev.value;
                continue;
            }
            /**
             * Braces
             */
            if (value === '{' && opts.nobrace !== true) {
                increment('braces');
                const open = {
                    type: 'brace',
                    value,
                    output: '(',
                    outputIndex: state.output.length,
                    tokensIndex: state.tokens.length
                };
                braces.push(open);
                push(open);
                continue;
            }
            if (value === '}') {
                const brace = braces[braces.length - 1];
                if (opts.nobrace === true || !brace) {
                    push({ type: 'text', value, output: value });
                    continue;
                }
                let output = ')';
                if (brace.dots === true) {
                    const arr = tokens.slice();
                    const range = [];
                    for (let i = arr.length - 1; i >= 0; i--) {
                        tokens.pop();
                        if (arr[i].type === 'brace') {
                            break;
                        }
                        if (arr[i].type !== 'dots') {
                            range.unshift(arr[i].value);
                        }
                    }
                    output = expandRange(range, opts);
                    state.backtrack = true;
                }
                if (brace.comma !== true && brace.dots !== true) {
                    const out = state.output.slice(0, brace.outputIndex);
                    const toks = state.tokens.slice(brace.tokensIndex);
                    brace.value = brace.output = '\\{';
                    value = output = '\\}';
                    state.output = out;
                    for (const t of toks) {
                        state.output += (t.output || t.value);
                    }
                }
                push({ type: 'brace', value, output });
                decrement('braces');
                braces.pop();
                continue;
            }
            /**
             * Pipes
             */
            if (value === '|') {
                if (extglobs.length > 0) {
                    extglobs[extglobs.length - 1].conditions++;
                }
                push({ type: 'text', value });
                continue;
            }
            /**
             * Commas
             */
            if (value === ',') {
                let output = value;
                const brace = braces[braces.length - 1];
                if (brace && stack[stack.length - 1] === 'braces') {
                    brace.comma = true;
                    output = '|';
                }
                push({ type: 'comma', value, output });
                continue;
            }
            /**
             * Slashes
             */
            if (value === '/') {
                // if the beginning of the glob is "./", advance the start
                // to the current index, and don't add the "./" characters
                // to the state. This greatly simplifies lookbehinds when
                // checking for BOS characters like "!" and "." (not "./")
                if (prev.type === 'dot' && state.index === state.start + 1) {
                    state.start = state.index + 1;
                    state.consumed = '';
                    state.output = '';
                    tokens.pop();
                    prev = bos; // reset "prev" to the first token
                    continue;
                }
                push({ type: 'slash', value, output: SLASH_LITERAL });
                continue;
            }
            /**
             * Dots
             */
            if (value === '.') {
                if (state.braces > 0 && prev.type === 'dot') {
                    if (prev.value === '.')
                        prev.output = DOT_LITERAL;
                    const brace = braces[braces.length - 1];
                    prev.type = 'dots';
                    prev.output += value;
                    prev.value += value;
                    brace.dots = true;
                    continue;
                }
                if ((state.braces + state.parens) === 0 && prev.type !== 'bos' && prev.type !== 'slash') {
                    push({ type: 'text', value, output: DOT_LITERAL });
                    continue;
                }
                push({ type: 'dot', value, output: DOT_LITERAL });
                continue;
            }
            /**
             * Question marks
             */
            if (value === '?') {
                const isGroup = prev && prev.value === '(';
                if (!isGroup && opts.noextglob !== true && peek() === '(' && peek(2) !== '?') {
                    extglobOpen('qmark', value);
                    continue;
                }
                if (prev && prev.type === 'paren') {
                    const next = peek();
                    let output = value;
                    if (next === '<' && !utils.supportsLookbehinds()) {
                        throw new Error('Node.js v10 or higher is required for regex lookbehinds');
                    }
                    if ((prev.value === '(' && !/[!=<:]/.test(next)) || (next === '<' && !/<([!=]|\w+>)/.test(remaining()))) {
                        output = `\\${value}`;
                    }
                    push({ type: 'text', value, output });
                    continue;
                }
                if (opts.dot !== true && (prev.type === 'slash' || prev.type === 'bos')) {
                    push({ type: 'qmark', value, output: QMARK_NO_DOT });
                    continue;
                }
                push({ type: 'qmark', value, output: QMARK });
                continue;
            }
            /**
             * Exclamation
             */
            if (value === '!') {
                if (opts.noextglob !== true && peek() === '(') {
                    if (peek(2) !== '?' || !/[!=<:]/.test(peek(3))) {
                        extglobOpen('negate', value);
                        continue;
                    }
                }
                if (opts.nonegate !== true && state.index === 0) {
                    negate();
                    continue;
                }
            }
            /**
             * Plus
             */
            if (value === '+') {
                if (opts.noextglob !== true && peek() === '(' && peek(2) !== '?') {
                    extglobOpen('plus', value);
                    continue;
                }
                if ((prev && prev.value === '(') || opts.regex === false) {
                    push({ type: 'plus', value, output: PLUS_LITERAL });
                    continue;
                }
                if ((prev && (prev.type === 'bracket' || prev.type === 'paren' || prev.type === 'brace')) || state.parens > 0) {
                    push({ type: 'plus', value });
                    continue;
                }
                push({ type: 'plus', value: PLUS_LITERAL });
                continue;
            }
            /**
             * Plain text
             */
            if (value === '@') {
                if (opts.noextglob !== true && peek() === '(' && peek(2) !== '?') {
                    push({ type: 'at', extglob: true, value, output: '' });
                    continue;
                }
                push({ type: 'text', value });
                continue;
            }
            /**
             * Plain text
             */
            if (value !== '*') {
                if (value === '$' || value === '^') {
                    value = `\\${value}`;
                }
                const match = REGEX_NON_SPECIAL_CHARS.exec(remaining());
                if (match) {
                    value += match[0];
                    state.index += match[0].length;
                }
                push({ type: 'text', value });
                continue;
            }
            /**
             * Stars
             */
            if (prev && (prev.type === 'globstar' || prev.star === true)) {
                prev.type = 'star';
                prev.star = true;
                prev.value += value;
                prev.output = star;
                state.backtrack = true;
                state.globstar = true;
                consume(value);
                continue;
            }
            let rest = remaining();
            if (opts.noextglob !== true && /^\([^?]/.test(rest)) {
                extglobOpen('star', value);
                continue;
            }
            if (prev.type === 'star') {
                if (opts.noglobstar === true) {
                    consume(value);
                    continue;
                }
                const prior = prev.prev;
                const before = prior.prev;
                const isStart = prior.type === 'slash' || prior.type === 'bos';
                const afterStar = before && (before.type === 'star' || before.type === 'globstar');
                if (opts.bash === true && (!isStart || (rest[0] && rest[0] !== '/'))) {
                    push({ type: 'star', value, output: '' });
                    continue;
                }
                const isBrace = state.braces > 0 && (prior.type === 'comma' || prior.type === 'brace');
                const isExtglob = extglobs.length && (prior.type === 'pipe' || prior.type === 'paren');
                if (!isStart && prior.type !== 'paren' && !isBrace && !isExtglob) {
                    push({ type: 'star', value, output: '' });
                    continue;
                }
                // strip consecutive `/**/`
                while (rest.slice(0, 3) === '/**') {
                    const after = input[state.index + 4];
                    if (after && after !== '/') {
                        break;
                    }
                    rest = rest.slice(3);
                    consume('/**', 3);
                }
                if (prior.type === 'bos' && eos()) {
                    prev.type = 'globstar';
                    prev.value += value;
                    prev.output = globstar(opts);
                    state.output = prev.output;
                    state.globstar = true;
                    consume(value);
                    continue;
                }
                if (prior.type === 'slash' && prior.prev.type !== 'bos' && !afterStar && eos()) {
                    state.output = state.output.slice(0, -(prior.output + prev.output).length);
                    prior.output = `(?:${prior.output}`;
                    prev.type = 'globstar';
                    prev.output = globstar(opts) + (opts.strictSlashes ? ')' : '|$)');
                    prev.value += value;
                    state.globstar = true;
                    state.output += prior.output + prev.output;
                    consume(value);
                    continue;
                }
                if (prior.type === 'slash' && prior.prev.type !== 'bos' && rest[0] === '/') {
                    const end = rest[1] !== void 0 ? '|$' : '';
                    state.output = state.output.slice(0, -(prior.output + prev.output).length);
                    prior.output = `(?:${prior.output}`;
                    prev.type = 'globstar';
                    prev.output = `${globstar(opts)}${SLASH_LITERAL}|${SLASH_LITERAL}${end})`;
                    prev.value += value;
                    state.output += prior.output + prev.output;
                    state.globstar = true;
                    consume(value + advance());
                    push({ type: 'slash', value: '/', output: '' });
                    continue;
                }
                if (prior.type === 'bos' && rest[0] === '/') {
                    prev.type = 'globstar';
                    prev.value += value;
                    prev.output = `(?:^|${SLASH_LITERAL}|${globstar(opts)}${SLASH_LITERAL})`;
                    state.output = prev.output;
                    state.globstar = true;
                    consume(value + advance());
                    push({ type: 'slash', value: '/', output: '' });
                    continue;
                }
                // remove single star from output
                state.output = state.output.slice(0, -prev.output.length);
                // reset previous token to globstar
                prev.type = 'globstar';
                prev.output = globstar(opts);
                prev.value += value;
                // reset output with globstar
                state.output += prev.output;
                state.globstar = true;
                consume(value);
                continue;
            }
            const token = { type: 'star', value, output: star };
            if (opts.bash === true) {
                token.output = '.*?';
                if (prev.type === 'bos' || prev.type === 'slash') {
                    token.output = nodot + token.output;
                }
                push(token);
                continue;
            }
            if (prev && (prev.type === 'bracket' || prev.type === 'paren') && opts.regex === true) {
                token.output = value;
                push(token);
                continue;
            }
            if (state.index === state.start || prev.type === 'slash' || prev.type === 'dot') {
                if (prev.type === 'dot') {
                    state.output += NO_DOT_SLASH;
                    prev.output += NO_DOT_SLASH;
                }
                else if (opts.dot === true) {
                    state.output += NO_DOTS_SLASH;
                    prev.output += NO_DOTS_SLASH;
                }
                else {
                    state.output += nodot;
                    prev.output += nodot;
                }
                if (peek() !== '*') {
                    state.output += ONE_CHAR;
                    prev.output += ONE_CHAR;
                }
            }
            push(token);
        }
        while (state.brackets > 0) {
            if (opts.strictBrackets === true)
                throw new SyntaxError(syntaxError('closing', ']'));
            state.output = utils.escapeLast(state.output, '[');
            decrement('brackets');
        }
        while (state.parens > 0) {
            if (opts.strictBrackets === true)
                throw new SyntaxError(syntaxError('closing', ')'));
            state.output = utils.escapeLast(state.output, '(');
            decrement('parens');
        }
        while (state.braces > 0) {
            if (opts.strictBrackets === true)
                throw new SyntaxError(syntaxError('closing', '}'));
            state.output = utils.escapeLast(state.output, '{');
            decrement('braces');
        }
        if (opts.strictSlashes !== true && (prev.type === 'star' || prev.type === 'bracket')) {
            push({ type: 'maybe_slash', value: '', output: `${SLASH_LITERAL}?` });
        }
        // rebuild the output if we had to backtrack at any point
        if (state.backtrack === true) {
            state.output = '';
            for (const token of state.tokens) {
                state.output += token.output != null ? token.output : token.value;
                if (token.suffix) {
                    state.output += token.suffix;
                }
            }
        }
        return state;
    };
    /**
     * Fast paths for creating regular expressions for common glob patterns.
     * This can significantly speed up processing and has very little downside
     * impact when none of the fast paths match.
     */
    parse.fastpaths = (input, options) => {
        const opts = { ...options };
        const max = typeof opts.maxLength === 'number' ? Math.min(MAX_LENGTH, opts.maxLength) : MAX_LENGTH;
        const len = input.length;
        if (len > max) {
            throw new SyntaxError(`Input length: ${len}, exceeds maximum allowed length: ${max}`);
        }
        input = REPLACEMENTS[input] || input;
        const win32 = utils.isWindows(options);
        // create constants based on platform, for windows or posix
        const { DOT_LITERAL, SLASH_LITERAL, ONE_CHAR, DOTS_SLASH, NO_DOT, NO_DOTS, NO_DOTS_SLASH, STAR, START_ANCHOR } = constants.globChars(win32);
        const nodot = opts.dot ? NO_DOTS : NO_DOT;
        const slashDot = opts.dot ? NO_DOTS_SLASH : NO_DOT;
        const capture = opts.capture ? '' : '?:';
        const state = { negated: false, prefix: '' };
        let star = opts.bash === true ? '.*?' : STAR;
        if (opts.capture) {
            star = `(${star})`;
        }
        const globstar = opts => {
            if (opts.noglobstar === true)
                return star;
            return `(${capture}(?:(?!${START_ANCHOR}${opts.dot ? DOTS_SLASH : DOT_LITERAL}).)*?)`;
        };
        const create = str => {
            switch (str) {
                case '*':
                    return `${nodot}${ONE_CHAR}${star}`;
                case '.*':
                    return `${DOT_LITERAL}${ONE_CHAR}${star}`;
                case '*.*':
                    return `${nodot}${star}${DOT_LITERAL}${ONE_CHAR}${star}`;
                case '*/*':
                    return `${nodot}${star}${SLASH_LITERAL}${ONE_CHAR}${slashDot}${star}`;
                case '**':
                    return nodot + globstar(opts);
                case '**/*':
                    return `(?:${nodot}${globstar(opts)}${SLASH_LITERAL})?${slashDot}${ONE_CHAR}${star}`;
                case '**/*.*':
                    return `(?:${nodot}${globstar(opts)}${SLASH_LITERAL})?${slashDot}${star}${DOT_LITERAL}${ONE_CHAR}${star}`;
                case '**/.*':
                    return `(?:${nodot}${globstar(opts)}${SLASH_LITERAL})?${DOT_LITERAL}${ONE_CHAR}${star}`;
                default: {
                    const match = /^(.*?)\.(\w+)$/.exec(str);
                    if (!match)
                        return;
                    const source = create(match[1]);
                    if (!source)
                        return;
                    return source + DOT_LITERAL + match[2];
                }
            }
        };
        const output = utils.removePrefix(input, state);
        let source = create(output);
        if (source && opts.strictSlashes !== true) {
            source += `${SLASH_LITERAL}?`;
        }
        return source;
    };
    module.exports = parse;
},
/* fast-glob/out/utils/stream.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const merge2 = require(1123) /* merge2 */;
    function merge(streams) {
        const mergedStream = merge2(streams);
        streams.forEach((stream) => {
            stream.once('error', (error) => mergedStream.emit('error', error));
        });
        mergedStream.once('close', () => propagateCloseEventToSources(streams));
        mergedStream.once('end', () => propagateCloseEventToSources(streams));
        return mergedStream;
    }
    exports.merge = merge;
    function propagateCloseEventToSources(streams) {
        streams.forEach((stream) => stream.emit('close'));
    }
},
/* fast-glob/out/utils/string.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function isString(input) {
        return typeof input === 'string';
    }
    exports.isString = isString;
    function isEmpty(input) {
        return input === '';
    }
    exports.isEmpty = isEmpty;
},
/* fast-glob/out/providers/async.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const stream_1 = require(1152) /* ../readers/stream */;
    const provider_1 = require(1181) /* ./provider */;
    class ProviderAsync extends provider_1.default {
        constructor() {
            super(...arguments);
            this._reader = new stream_1.default(this._settings);
        }
        read(task) {
            const root = this._getRootDirectory(task);
            const options = this._getReaderOptions(task);
            const entries = [];
            return new Promise((resolve, reject) => {
                const stream = this.api(root, task, options);
                stream.once('error', reject);
                stream.on('data', (entry) => entries.push(options.transform(entry)));
                stream.once('end', () => resolve(entries));
            });
        }
        api(root, task, options) {
            if (task.dynamic) {
                return this._reader.dynamic(root, options);
            }
            return this._reader.static(task.patterns, options);
        }
    }
    exports.default = ProviderAsync;
},
/* fast-glob/out/readers/stream.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const stream_1 = require("stream");
    const fsStat = require(1153) /* @nodelib/fs.stat */;
    const fsWalk = require(1158) /* @nodelib/fs.walk */;
    const reader_1 = require(1180) /* ./reader */;
    class ReaderStream extends reader_1.default {
        constructor() {
            super(...arguments);
            this._walkStream = fsWalk.walkStream;
            this._stat = fsStat.stat;
        }
        dynamic(root, options) {
            return this._walkStream(root, options);
        }
        static(patterns, options) {
            const filepaths = patterns.map(this._getFullEntryPath, this);
            const stream = new stream_1.PassThrough({ objectMode: true });
            stream._write = (index, _enc, done) => {
                return this._getEntry(filepaths[index], patterns[index], options)
                    .then((entry) => {
                    if (entry !== null && options.entryFilter(entry)) {
                        stream.push(entry);
                    }
                    if (index === filepaths.length - 1) {
                        stream.end();
                    }
                    done();
                })
                    .catch(done);
            };
            for (let i = 0; i < filepaths.length; i++) {
                stream.write(i);
            }
            return stream;
        }
        _getEntry(filepath, pattern, options) {
            return this._getStat(filepath)
                .then((stats) => this._makeEntry(stats, pattern))
                .catch((error) => {
                if (options.errorFilter(error)) {
                    return null;
                }
                throw error;
            });
        }
        _getStat(filepath) {
            return new Promise((resolve, reject) => {
                this._stat(filepath, this._fsStatSettings, (error, stats) => {
                    return error === null ? resolve(stats) : reject(error);
                });
            });
        }
    }
    exports.default = ReaderStream;
},
/* @nodelib/fs.stat/out/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const async = require(1154) /* ./providers/async */;
    const sync = require(1155) /* ./providers/sync */;
    const settings_1 = require(1156) /* ./settings */;
    exports.Settings = settings_1.default;
    function stat(path, optionsOrSettingsOrCallback, callback) {
        if (typeof optionsOrSettingsOrCallback === 'function') {
            async.read(path, getSettings(), optionsOrSettingsOrCallback);
            return;
        }
        async.read(path, getSettings(optionsOrSettingsOrCallback), callback);
    }
    exports.stat = stat;
    function statSync(path, optionsOrSettings) {
        const settings = getSettings(optionsOrSettings);
        return sync.read(path, settings);
    }
    exports.statSync = statSync;
    function getSettings(settingsOrOptions = {}) {
        if (settingsOrOptions instanceof settings_1.default) {
            return settingsOrOptions;
        }
        return new settings_1.default(settingsOrOptions);
    }
},
/* @nodelib/fs.stat/out/providers/async.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function read(path, settings, callback) {
        settings.fs.lstat(path, (lstatError, lstat) => {
            if (lstatError !== null) {
                callFailureCallback(callback, lstatError);
                return;
            }
            if (!lstat.isSymbolicLink() || !settings.followSymbolicLink) {
                callSuccessCallback(callback, lstat);
                return;
            }
            settings.fs.stat(path, (statError, stat) => {
                if (statError !== null) {
                    if (settings.throwErrorOnBrokenSymbolicLink) {
                        callFailureCallback(callback, statError);
                        return;
                    }
                    callSuccessCallback(callback, lstat);
                    return;
                }
                if (settings.markSymbolicLink) {
                    stat.isSymbolicLink = () => true;
                }
                callSuccessCallback(callback, stat);
            });
        });
    }
    exports.read = read;
    function callFailureCallback(callback, error) {
        callback(error);
    }
    function callSuccessCallback(callback, result) {
        callback(null, result);
    }
},
/* @nodelib/fs.stat/out/providers/sync.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function read(path, settings) {
        const lstat = settings.fs.lstatSync(path);
        if (!lstat.isSymbolicLink() || !settings.followSymbolicLink) {
            return lstat;
        }
        try {
            const stat = settings.fs.statSync(path);
            if (settings.markSymbolicLink) {
                stat.isSymbolicLink = () => true;
            }
            return stat;
        }
        catch (error) {
            if (!settings.throwErrorOnBrokenSymbolicLink) {
                return lstat;
            }
            throw error;
        }
    }
    exports.read = read;
},
/* @nodelib/fs.stat/out/settings.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const fs = require(1157) /* ./adapters/fs */;
    class Settings {
        constructor(_options = {}) {
            this._options = _options;
            this.followSymbolicLink = this._getValue(this._options.followSymbolicLink, true);
            this.fs = fs.createFileSystemAdapter(this._options.fs);
            this.markSymbolicLink = this._getValue(this._options.markSymbolicLink, false);
            this.throwErrorOnBrokenSymbolicLink = this._getValue(this._options.throwErrorOnBrokenSymbolicLink, true);
        }
        _getValue(option, value) {
            return option !== null && option !== void 0 ? option : value;
        }
    }
    exports.default = Settings;
},
/* @nodelib/fs.stat/out/adapters/fs.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const fs = require("fs");
    exports.FILE_SYSTEM_ADAPTER = {
        lstat: fs.lstat,
        stat: fs.stat,
        lstatSync: fs.lstatSync,
        statSync: fs.statSync
    };
    function createFileSystemAdapter(fsMethods) {
        if (fsMethods === undefined) {
            return exports.FILE_SYSTEM_ADAPTER;
        }
        return Object.assign(Object.assign({}, exports.FILE_SYSTEM_ADAPTER), fsMethods);
    }
    exports.createFileSystemAdapter = createFileSystemAdapter;
},
/* @nodelib/fs.walk/out/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const async_1 = require(1159) /* ./providers/async */;
    const stream_1 = require(1176) /* ./providers/stream */;
    const sync_1 = require(1177) /* ./providers/sync */;
    const settings_1 = require(1179) /* ./settings */;
    exports.Settings = settings_1.default;
    function walk(directory, optionsOrSettingsOrCallback, callback) {
        if (typeof optionsOrSettingsOrCallback === 'function') {
            new async_1.default(directory, getSettings()).read(optionsOrSettingsOrCallback);
            return;
        }
        new async_1.default(directory, getSettings(optionsOrSettingsOrCallback)).read(callback);
    }
    exports.walk = walk;
    function walkSync(directory, optionsOrSettings) {
        const settings = getSettings(optionsOrSettings);
        const provider = new sync_1.default(directory, settings);
        return provider.read();
    }
    exports.walkSync = walkSync;
    function walkStream(directory, optionsOrSettings) {
        const settings = getSettings(optionsOrSettings);
        const provider = new stream_1.default(directory, settings);
        return provider.read();
    }
    exports.walkStream = walkStream;
    function getSettings(settingsOrOptions = {}) {
        if (settingsOrOptions instanceof settings_1.default) {
            return settingsOrOptions;
        }
        return new settings_1.default(settingsOrOptions);
    }
},
/* @nodelib/fs.walk/out/providers/async.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const async_1 = require(1160) /* ../readers/async */;
    class AsyncProvider {
        constructor(_root, _settings) {
            this._root = _root;
            this._settings = _settings;
            this._reader = new async_1.default(this._root, this._settings);
            this._storage = new Set();
        }
        read(callback) {
            this._reader.onError((error) => {
                callFailureCallback(callback, error);
            });
            this._reader.onEntry((entry) => {
                this._storage.add(entry);
            });
            this._reader.onEnd(() => {
                callSuccessCallback(callback, [...this._storage]);
            });
            this._reader.read();
        }
    }
    exports.default = AsyncProvider;
    function callFailureCallback(callback, error) {
        callback(error);
    }
    function callSuccessCallback(callback, entries) {
        callback(null, entries);
    }
},
/* @nodelib/fs.walk/out/readers/async.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const events_1 = require("events");
    const fsScandir = require(1161) /* @nodelib/fs.scandir */;
    const fastq = require(1172) /* fastq */;
    const common = require(1174) /* ./common */;
    const reader_1 = require(1175) /* ./reader */;
    class AsyncReader extends reader_1.default {
        constructor(_root, _settings) {
            super(_root, _settings);
            this._settings = _settings;
            this._scandir = fsScandir.scandir;
            this._emitter = new events_1.EventEmitter();
            this._queue = fastq(this._worker.bind(this), this._settings.concurrency);
            this._isFatalError = false;
            this._isDestroyed = false;
            this._queue.drain = () => {
                if (!this._isFatalError) {
                    this._emitter.emit('end');
                }
            };
        }
        read() {
            this._isFatalError = false;
            this._isDestroyed = false;
            setImmediate(() => {
                this._pushToQueue(this._root, this._settings.basePath);
            });
            return this._emitter;
        }
        get isDestroyed() {
            return this._isDestroyed;
        }
        destroy() {
            if (this._isDestroyed) {
                throw new Error('The reader is already destroyed');
            }
            this._isDestroyed = true;
            this._queue.killAndDrain();
        }
        onEntry(callback) {
            this._emitter.on('entry', callback);
        }
        onError(callback) {
            this._emitter.once('error', callback);
        }
        onEnd(callback) {
            this._emitter.once('end', callback);
        }
        _pushToQueue(directory, base) {
            const queueItem = { directory, base };
            this._queue.push(queueItem, (error) => {
                if (error !== null) {
                    this._handleError(error);
                }
            });
        }
        _worker(item, done) {
            this._scandir(item.directory, this._settings.fsScandirSettings, (error, entries) => {
                if (error !== null) {
                    done(error, undefined);
                    return;
                }
                for (const entry of entries) {
                    this._handleEntry(entry, item.base);
                }
                done(null, undefined);
            });
        }
        _handleError(error) {
            if (this._isDestroyed || !common.isFatalError(this._settings, error)) {
                return;
            }
            this._isFatalError = true;
            this._isDestroyed = true;
            this._emitter.emit('error', error);
        }
        _handleEntry(entry, base) {
            if (this._isDestroyed || this._isFatalError) {
                return;
            }
            const fullpath = entry.path;
            if (base !== undefined) {
                entry.path = common.joinPathSegments(base, entry.name, this._settings.pathSegmentSeparator);
            }
            if (common.isAppliedFilter(this._settings.entryFilter, entry)) {
                this._emitEntry(entry);
            }
            if (entry.dirent.isDirectory() && common.isAppliedFilter(this._settings.deepFilter, entry)) {
                this._pushToQueue(fullpath, entry.path);
            }
        }
        _emitEntry(entry) {
            this._emitter.emit('entry', entry);
        }
    }
    exports.default = AsyncReader;
},
/* @nodelib/fs.scandir/out/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const async = require(1162) /* ./providers/async */;
    const sync = require(1169) /* ./providers/sync */;
    const settings_1 = require(1170) /* ./settings */;
    exports.Settings = settings_1.default;
    function scandir(path, optionsOrSettingsOrCallback, callback) {
        if (typeof optionsOrSettingsOrCallback === 'function') {
            async.read(path, getSettings(), optionsOrSettingsOrCallback);
            return;
        }
        async.read(path, getSettings(optionsOrSettingsOrCallback), callback);
    }
    exports.scandir = scandir;
    function scandirSync(path, optionsOrSettings) {
        const settings = getSettings(optionsOrSettings);
        return sync.read(path, settings);
    }
    exports.scandirSync = scandirSync;
    function getSettings(settingsOrOptions = {}) {
        if (settingsOrOptions instanceof settings_1.default) {
            return settingsOrOptions;
        }
        return new settings_1.default(settingsOrOptions);
    }
},
/* @nodelib/fs.scandir/out/providers/async.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const fsStat = require(1153) /* @nodelib/fs.stat */;
    const rpl = require(1163) /* run-parallel */;
    const constants_1 = require(1165) /* ../constants */;
    const utils = require(1166) /* ../utils */;
    const common = require(1168) /* ./common */;
    function read(directory, settings, callback) {
        if (!settings.stats && constants_1.IS_SUPPORT_READDIR_WITH_FILE_TYPES) {
            readdirWithFileTypes(directory, settings, callback);
            return;
        }
        readdir(directory, settings, callback);
    }
    exports.read = read;
    function readdirWithFileTypes(directory, settings, callback) {
        settings.fs.readdir(directory, { withFileTypes: true }, (readdirError, dirents) => {
            if (readdirError !== null) {
                callFailureCallback(callback, readdirError);
                return;
            }
            const entries = dirents.map((dirent) => ({
                dirent,
                name: dirent.name,
                path: common.joinPathSegments(directory, dirent.name, settings.pathSegmentSeparator)
            }));
            if (!settings.followSymbolicLinks) {
                callSuccessCallback(callback, entries);
                return;
            }
            const tasks = entries.map((entry) => makeRplTaskEntry(entry, settings));
            rpl(tasks, (rplError, rplEntries) => {
                if (rplError !== null) {
                    callFailureCallback(callback, rplError);
                    return;
                }
                callSuccessCallback(callback, rplEntries);
            });
        });
    }
    exports.readdirWithFileTypes = readdirWithFileTypes;
    function makeRplTaskEntry(entry, settings) {
        return (done) => {
            if (!entry.dirent.isSymbolicLink()) {
                done(null, entry);
                return;
            }
            settings.fs.stat(entry.path, (statError, stats) => {
                if (statError !== null) {
                    if (settings.throwErrorOnBrokenSymbolicLink) {
                        done(statError);
                        return;
                    }
                    done(null, entry);
                    return;
                }
                entry.dirent = utils.fs.createDirentFromStats(entry.name, stats);
                done(null, entry);
            });
        };
    }
    function readdir(directory, settings, callback) {
        settings.fs.readdir(directory, (readdirError, names) => {
            if (readdirError !== null) {
                callFailureCallback(callback, readdirError);
                return;
            }
            const tasks = names.map((name) => {
                const path = common.joinPathSegments(directory, name, settings.pathSegmentSeparator);
                return (done) => {
                    fsStat.stat(path, settings.fsStatSettings, (error, stats) => {
                        if (error !== null) {
                            done(error);
                            return;
                        }
                        const entry = {
                            name,
                            path,
                            dirent: utils.fs.createDirentFromStats(name, stats)
                        };
                        if (settings.stats) {
                            entry.stats = stats;
                        }
                        done(null, entry);
                    });
                };
            });
            rpl(tasks, (rplError, entries) => {
                if (rplError !== null) {
                    callFailureCallback(callback, rplError);
                    return;
                }
                callSuccessCallback(callback, entries);
            });
        });
    }
    exports.readdir = readdir;
    function callFailureCallback(callback, error) {
        callback(error);
    }
    function callSuccessCallback(callback, result) {
        callback(null, result);
    }
},
/* run-parallel/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /*! run-parallel. MIT License. Feross Aboukhadijeh <https://feross.org/opensource> */
    module.exports = runParallel;
    const queueMicrotask = require(1164) /* queue-microtask */;
    function runParallel(tasks, cb) {
        let results, pending, keys;
        let isSync = true;
        if (Array.isArray(tasks)) {
            results = [];
            pending = tasks.length;
        }
        else {
            keys = Object.keys(tasks);
            results = {};
            pending = keys.length;
        }
        function done(err) {
            function end() {
                if (cb)
                    cb(err, results);
                cb = null;
            }
            if (isSync)
                queueMicrotask(end);
            else
                end();
        }
        function each(i, err, result) {
            results[i] = result;
            if (--pending === 0 || err) {
                done(err);
            }
        }
        if (!pending) {
            // empty
            done(null);
        }
        else if (keys) {
            // object
            keys.forEach(function (key) {
                tasks[key](function (err, result) { each(key, err, result); });
            });
        }
        else {
            // array
            tasks.forEach(function (task, i) {
                task(function (err, result) { each(i, err, result); });
            });
        }
        isSync = false;
    }
},
/* queue-microtask/index.js */ function _(require, module, exports, __esModule, __esExport) {
    /*! queue-microtask. MIT License. Feross Aboukhadijeh <https://feross.org/opensource> */
    let promise;
    module.exports = typeof queueMicrotask === 'function'
        ? queueMicrotask.bind(typeof window !== 'undefined' ? window : global)
        // reuse resolved promise, and allocate it lazily
        : cb => (promise || (promise = Promise.resolve()))
            .then(cb)
            .catch(err => setTimeout(() => { throw err; }, 0));
},
/* @nodelib/fs.scandir/out/constants.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const NODE_PROCESS_VERSION_PARTS = process.versions.node.split('.');
    if (NODE_PROCESS_VERSION_PARTS[0] === undefined || NODE_PROCESS_VERSION_PARTS[1] === undefined) {
        throw new Error(`Unexpected behavior. The 'process.versions.node' variable has invalid value: ${process.versions.node}`);
    }
    const MAJOR_VERSION = Number.parseInt(NODE_PROCESS_VERSION_PARTS[0], 10);
    const MINOR_VERSION = Number.parseInt(NODE_PROCESS_VERSION_PARTS[1], 10);
    const SUPPORTED_MAJOR_VERSION = 10;
    const SUPPORTED_MINOR_VERSION = 10;
    const IS_MATCHED_BY_MAJOR = MAJOR_VERSION > SUPPORTED_MAJOR_VERSION;
    const IS_MATCHED_BY_MAJOR_AND_MINOR = MAJOR_VERSION === SUPPORTED_MAJOR_VERSION && MINOR_VERSION >= SUPPORTED_MINOR_VERSION;
    /**
     * IS `true` for Node.js 10.10 and greater.
     */
    exports.IS_SUPPORT_READDIR_WITH_FILE_TYPES = IS_MATCHED_BY_MAJOR || IS_MATCHED_BY_MAJOR_AND_MINOR;
},
/* @nodelib/fs.scandir/out/utils/index.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const fs = require(1167) /* ./fs */;
    exports.fs = fs;
},
/* @nodelib/fs.scandir/out/utils/fs.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    class DirentFromStats {
        constructor(name, stats) {
            this.name = name;
            this.isBlockDevice = stats.isBlockDevice.bind(stats);
            this.isCharacterDevice = stats.isCharacterDevice.bind(stats);
            this.isDirectory = stats.isDirectory.bind(stats);
            this.isFIFO = stats.isFIFO.bind(stats);
            this.isFile = stats.isFile.bind(stats);
            this.isSocket = stats.isSocket.bind(stats);
            this.isSymbolicLink = stats.isSymbolicLink.bind(stats);
        }
    }
    function createDirentFromStats(name, stats) {
        return new DirentFromStats(name, stats);
    }
    exports.createDirentFromStats = createDirentFromStats;
},
/* @nodelib/fs.scandir/out/providers/common.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function joinPathSegments(a, b, separator) {
        /**
         * The correct handling of cases when the first segment is a root (`/`, `C:/`) or UNC path (`//?/C:/`).
         */
        if (a.endsWith(separator)) {
            return a + b;
        }
        return a + separator + b;
    }
    exports.joinPathSegments = joinPathSegments;
},
/* @nodelib/fs.scandir/out/providers/sync.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const fsStat = require(1153) /* @nodelib/fs.stat */;
    const constants_1 = require(1165) /* ../constants */;
    const utils = require(1166) /* ../utils */;
    const common = require(1168) /* ./common */;
    function read(directory, settings) {
        if (!settings.stats && constants_1.IS_SUPPORT_READDIR_WITH_FILE_TYPES) {
            return readdirWithFileTypes(directory, settings);
        }
        return readdir(directory, settings);
    }
    exports.read = read;
    function readdirWithFileTypes(directory, settings) {
        const dirents = settings.fs.readdirSync(directory, { withFileTypes: true });
        return dirents.map((dirent) => {
            const entry = {
                dirent,
                name: dirent.name,
                path: common.joinPathSegments(directory, dirent.name, settings.pathSegmentSeparator)
            };
            if (entry.dirent.isSymbolicLink() && settings.followSymbolicLinks) {
                try {
                    const stats = settings.fs.statSync(entry.path);
                    entry.dirent = utils.fs.createDirentFromStats(entry.name, stats);
                }
                catch (error) {
                    if (settings.throwErrorOnBrokenSymbolicLink) {
                        throw error;
                    }
                }
            }
            return entry;
        });
    }
    exports.readdirWithFileTypes = readdirWithFileTypes;
    function readdir(directory, settings) {
        const names = settings.fs.readdirSync(directory);
        return names.map((name) => {
            const entryPath = common.joinPathSegments(directory, name, settings.pathSegmentSeparator);
            const stats = fsStat.statSync(entryPath, settings.fsStatSettings);
            const entry = {
                name,
                path: entryPath,
                dirent: utils.fs.createDirentFromStats(name, stats)
            };
            if (settings.stats) {
                entry.stats = stats;
            }
            return entry;
        });
    }
    exports.readdir = readdir;
},
/* @nodelib/fs.scandir/out/settings.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const path = require("path");
    const fsStat = require(1153) /* @nodelib/fs.stat */;
    const fs = require(1171) /* ./adapters/fs */;
    class Settings {
        constructor(_options = {}) {
            this._options = _options;
            this.followSymbolicLinks = this._getValue(this._options.followSymbolicLinks, false);
            this.fs = fs.createFileSystemAdapter(this._options.fs);
            this.pathSegmentSeparator = this._getValue(this._options.pathSegmentSeparator, path.sep);
            this.stats = this._getValue(this._options.stats, false);
            this.throwErrorOnBrokenSymbolicLink = this._getValue(this._options.throwErrorOnBrokenSymbolicLink, true);
            this.fsStatSettings = new fsStat.Settings({
                followSymbolicLink: this.followSymbolicLinks,
                fs: this.fs,
                throwErrorOnBrokenSymbolicLink: this.throwErrorOnBrokenSymbolicLink
            });
        }
        _getValue(option, value) {
            return option !== null && option !== void 0 ? option : value;
        }
    }
    exports.default = Settings;
},
/* @nodelib/fs.scandir/out/adapters/fs.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const fs = require("fs");
    exports.FILE_SYSTEM_ADAPTER = {
        lstat: fs.lstat,
        stat: fs.stat,
        lstatSync: fs.lstatSync,
        statSync: fs.statSync,
        readdir: fs.readdir,
        readdirSync: fs.readdirSync
    };
    function createFileSystemAdapter(fsMethods) {
        if (fsMethods === undefined) {
            return exports.FILE_SYSTEM_ADAPTER;
        }
        return Object.assign(Object.assign({}, exports.FILE_SYSTEM_ADAPTER), fsMethods);
    }
    exports.createFileSystemAdapter = createFileSystemAdapter;
},
/* fastq/queue.js */ function _(require, module, exports, __esModule, __esExport) {
    var reusify = require(1173) /* reusify */;
    function fastqueue(context, worker, concurrency) {
        if (typeof context === 'function') {
            concurrency = worker;
            worker = context;
            context = null;
        }
        if (concurrency < 1) {
            throw new Error('fastqueue concurrency must be greater than 1');
        }
        var cache = reusify(Task);
        var queueHead = null;
        var queueTail = null;
        var _running = 0;
        var errorHandler = null;
        var self = {
            push: push,
            drain: noop,
            saturated: noop,
            pause: pause,
            paused: false,
            concurrency: concurrency,
            running: running,
            resume: resume,
            idle: idle,
            length: length,
            getQueue: getQueue,
            unshift: unshift,
            empty: noop,
            kill: kill,
            killAndDrain: killAndDrain,
            error: error
        };
        return self;
        function running() {
            return _running;
        }
        function pause() {
            self.paused = true;
        }
        function length() {
            var current = queueHead;
            var counter = 0;
            while (current) {
                current = current.next;
                counter++;
            }
            return counter;
        }
        function getQueue() {
            var current = queueHead;
            var tasks = [];
            while (current) {
                tasks.push(current.value);
                current = current.next;
            }
            return tasks;
        }
        function resume() {
            if (!self.paused)
                return;
            self.paused = false;
            for (var i = 0; i < self.concurrency; i++) {
                _running++;
                release();
            }
        }
        function idle() {
            return _running === 0 && self.length() === 0;
        }
        function push(value, done) {
            var current = cache.get();
            current.context = context;
            current.release = release;
            current.value = value;
            current.callback = done || noop;
            current.errorHandler = errorHandler;
            if (_running === self.concurrency || self.paused) {
                if (queueTail) {
                    queueTail.next = current;
                    queueTail = current;
                }
                else {
                    queueHead = current;
                    queueTail = current;
                    self.saturated();
                }
            }
            else {
                _running++;
                worker.call(context, current.value, current.worked);
            }
        }
        function unshift(value, done) {
            var current = cache.get();
            current.context = context;
            current.release = release;
            current.value = value;
            current.callback = done || noop;
            if (_running === self.concurrency || self.paused) {
                if (queueHead) {
                    current.next = queueHead;
                    queueHead = current;
                }
                else {
                    queueHead = current;
                    queueTail = current;
                    self.saturated();
                }
            }
            else {
                _running++;
                worker.call(context, current.value, current.worked);
            }
        }
        function release(holder) {
            if (holder) {
                cache.release(holder);
            }
            var next = queueHead;
            if (next) {
                if (!self.paused) {
                    if (queueTail === queueHead) {
                        queueTail = null;
                    }
                    queueHead = next.next;
                    next.next = null;
                    worker.call(context, next.value, next.worked);
                    if (queueTail === null) {
                        self.empty();
                    }
                }
                else {
                    _running--;
                }
            }
            else if (--_running === 0) {
                self.drain();
            }
        }
        function kill() {
            queueHead = null;
            queueTail = null;
            self.drain = noop;
        }
        function killAndDrain() {
            queueHead = null;
            queueTail = null;
            self.drain();
            self.drain = noop;
        }
        function error(handler) {
            errorHandler = handler;
        }
    }
    function noop() { }
    function Task() {
        this.value = null;
        this.callback = noop;
        this.next = null;
        this.release = noop;
        this.context = null;
        this.errorHandler = null;
        var self = this;
        this.worked = function worked(err, result) {
            var callback = self.callback;
            var errorHandler = self.errorHandler;
            var val = self.value;
            self.value = null;
            self.callback = noop;
            if (self.errorHandler) {
                errorHandler(err, val);
            }
            callback.call(self.context, err, result);
            self.release(self);
        };
    }
    function queueAsPromised(context, worker, concurrency) {
        if (typeof context === 'function') {
            concurrency = worker;
            worker = context;
            context = null;
        }
        function asyncWrapper(arg, cb) {
            worker.call(this, arg)
                .then(function (res) {
                cb(null, res);
            }, cb);
        }
        var queue = fastqueue(context, asyncWrapper, concurrency);
        var pushCb = queue.push;
        var unshiftCb = queue.unshift;
        queue.push = push;
        queue.unshift = unshift;
        return queue;
        function push(value) {
            return new Promise(function (resolve, reject) {
                pushCb(value, function (err, result) {
                    if (err) {
                        reject(err);
                        return;
                    }
                    resolve(result);
                });
            });
        }
        function unshift(value) {
            return new Promise(function (resolve, reject) {
                unshiftCb(value, function (err, result) {
                    if (err) {
                        reject(err);
                        return;
                    }
                    resolve(result);
                });
            });
        }
    }
    module.exports = fastqueue;
    module.exports.promise = queueAsPromised;
},
/* reusify/reusify.js */ function _(require, module, exports, __esModule, __esExport) {
    function reusify(Constructor) {
        var head = new Constructor();
        var tail = head;
        function get() {
            var current = head;
            if (current.next) {
                head = current.next;
            }
            else {
                head = new Constructor();
                tail = head;
            }
            current.next = null;
            return current;
        }
        function release(obj) {
            tail.next = obj;
            tail = obj;
        }
        return {
            get: get,
            release: release
        };
    }
    module.exports = reusify;
},
/* @nodelib/fs.walk/out/readers/common.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    function isFatalError(settings, error) {
        if (settings.errorFilter === null) {
            return true;
        }
        return !settings.errorFilter(error);
    }
    exports.isFatalError = isFatalError;
    function isAppliedFilter(filter, value) {
        return filter === null || filter(value);
    }
    exports.isAppliedFilter = isAppliedFilter;
    function replacePathSegmentSeparator(filepath, separator) {
        return filepath.split(/[/\\]/).join(separator);
    }
    exports.replacePathSegmentSeparator = replacePathSegmentSeparator;
    function joinPathSegments(a, b, separator) {
        if (a === '') {
            return b;
        }
        /**
         * The correct handling of cases when the first segment is a root (`/`, `C:/`) or UNC path (`//?/C:/`).
         */
        if (a.endsWith(separator)) {
            return a + b;
        }
        return a + separator + b;
    }
    exports.joinPathSegments = joinPathSegments;
},
/* @nodelib/fs.walk/out/readers/reader.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const common = require(1174) /* ./common */;
    class Reader {
        constructor(_root, _settings) {
            this._root = _root;
            this._settings = _settings;
            this._root = common.replacePathSegmentSeparator(_root, _settings.pathSegmentSeparator);
        }
    }
    exports.default = Reader;
},
/* @nodelib/fs.walk/out/providers/stream.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const stream_1 = require("stream");
    const async_1 = require(1160) /* ../readers/async */;
    class StreamProvider {
        constructor(_root, _settings) {
            this._root = _root;
            this._settings = _settings;
            this._reader = new async_1.default(this._root, this._settings);
            this._stream = new stream_1.Readable({
                objectMode: true,
                read: () => { },
                destroy: () => {
                    if (!this._reader.isDestroyed) {
                        this._reader.destroy();
                    }
                }
            });
        }
        read() {
            this._reader.onError((error) => {
                this._stream.emit('error', error);
            });
            this._reader.onEntry((entry) => {
                this._stream.push(entry);
            });
            this._reader.onEnd(() => {
                this._stream.push(null);
            });
            this._reader.read();
            return this._stream;
        }
    }
    exports.default = StreamProvider;
},
/* @nodelib/fs.walk/out/providers/sync.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const sync_1 = require(1178) /* ../readers/sync */;
    class SyncProvider {
        constructor(_root, _settings) {
            this._root = _root;
            this._settings = _settings;
            this._reader = new sync_1.default(this._root, this._settings);
        }
        read() {
            return this._reader.read();
        }
    }
    exports.default = SyncProvider;
},
/* @nodelib/fs.walk/out/readers/sync.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const fsScandir = require(1161) /* @nodelib/fs.scandir */;
    const common = require(1174) /* ./common */;
    const reader_1 = require(1175) /* ./reader */;
    class SyncReader extends reader_1.default {
        constructor() {
            super(...arguments);
            this._scandir = fsScandir.scandirSync;
            this._storage = new Set();
            this._queue = new Set();
        }
        read() {
            this._pushToQueue(this._root, this._settings.basePath);
            this._handleQueue();
            return [...this._storage];
        }
        _pushToQueue(directory, base) {
            this._queue.add({ directory, base });
        }
        _handleQueue() {
            for (const item of this._queue.values()) {
                this._handleDirectory(item.directory, item.base);
            }
        }
        _handleDirectory(directory, base) {
            try {
                const entries = this._scandir(directory, this._settings.fsScandirSettings);
                for (const entry of entries) {
                    this._handleEntry(entry, base);
                }
            }
            catch (error) {
                this._handleError(error);
            }
        }
        _handleError(error) {
            if (!common.isFatalError(this._settings, error)) {
                return;
            }
            throw error;
        }
        _handleEntry(entry, base) {
            const fullpath = entry.path;
            if (base !== undefined) {
                entry.path = common.joinPathSegments(base, entry.name, this._settings.pathSegmentSeparator);
            }
            if (common.isAppliedFilter(this._settings.entryFilter, entry)) {
                this._pushToStorage(entry);
            }
            if (entry.dirent.isDirectory() && common.isAppliedFilter(this._settings.deepFilter, entry)) {
                this._pushToQueue(fullpath, entry.path);
            }
        }
        _pushToStorage(entry) {
            this._storage.add(entry);
        }
    }
    exports.default = SyncReader;
},
/* @nodelib/fs.walk/out/settings.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const path = require("path");
    const fsScandir = require(1161) /* @nodelib/fs.scandir */;
    class Settings {
        constructor(_options = {}) {
            this._options = _options;
            this.basePath = this._getValue(this._options.basePath, undefined);
            this.concurrency = this._getValue(this._options.concurrency, Number.POSITIVE_INFINITY);
            this.deepFilter = this._getValue(this._options.deepFilter, null);
            this.entryFilter = this._getValue(this._options.entryFilter, null);
            this.errorFilter = this._getValue(this._options.errorFilter, null);
            this.pathSegmentSeparator = this._getValue(this._options.pathSegmentSeparator, path.sep);
            this.fsScandirSettings = new fsScandir.Settings({
                followSymbolicLinks: this._options.followSymbolicLinks,
                fs: this._options.fs,
                pathSegmentSeparator: this._options.pathSegmentSeparator,
                stats: this._options.stats,
                throwErrorOnBrokenSymbolicLink: this._options.throwErrorOnBrokenSymbolicLink
            });
        }
        _getValue(option, value) {
            return option !== null && option !== void 0 ? option : value;
        }
    }
    exports.default = Settings;
},
/* fast-glob/out/readers/reader.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const path = require("path");
    const fsStat = require(1153) /* @nodelib/fs.stat */;
    const utils = require(1126) /* ../utils */;
    class Reader {
        constructor(_settings) {
            this._settings = _settings;
            this._fsStatSettings = new fsStat.Settings({
                followSymbolicLink: this._settings.followSymbolicLinks,
                fs: this._settings.fs,
                throwErrorOnBrokenSymbolicLink: this._settings.followSymbolicLinks
            });
        }
        _getFullEntryPath(filepath) {
            return path.resolve(this._settings.cwd, filepath);
        }
        _makeEntry(stats, pattern) {
            const entry = {
                name: pattern,
                path: pattern,
                dirent: utils.fs.createDirentFromStats(pattern, stats)
            };
            if (this._settings.stats) {
                entry.stats = stats;
            }
            return entry;
        }
        _isFatalError(error) {
            return !utils.errno.isEnoentCodeError(error) && !this._settings.suppressErrors;
        }
    }
    exports.default = Reader;
},
/* fast-glob/out/providers/provider.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const path = require("path");
    const deep_1 = require(1182) /* ./filters/deep */;
    const entry_1 = require(1185) /* ./filters/entry */;
    const error_1 = require(1186) /* ./filters/error */;
    const entry_2 = require(1187) /* ./transformers/entry */;
    class Provider {
        constructor(_settings) {
            this._settings = _settings;
            this.errorFilter = new error_1.default(this._settings);
            this.entryFilter = new entry_1.default(this._settings, this._getMicromatchOptions());
            this.deepFilter = new deep_1.default(this._settings, this._getMicromatchOptions());
            this.entryTransformer = new entry_2.default(this._settings);
        }
        _getRootDirectory(task) {
            return path.resolve(this._settings.cwd, task.base);
        }
        _getReaderOptions(task) {
            const basePath = task.base === '.' ? '' : task.base;
            return {
                basePath,
                pathSegmentSeparator: '/',
                concurrency: this._settings.concurrency,
                deepFilter: this.deepFilter.getFilter(basePath, task.positive, task.negative),
                entryFilter: this.entryFilter.getFilter(task.positive, task.negative),
                errorFilter: this.errorFilter.getFilter(),
                followSymbolicLinks: this._settings.followSymbolicLinks,
                fs: this._settings.fs,
                stats: this._settings.stats,
                throwErrorOnBrokenSymbolicLink: this._settings.throwErrorOnBrokenSymbolicLink,
                transform: this.entryTransformer.getTransformer()
            };
        }
        _getMicromatchOptions() {
            return {
                dot: this._settings.dot,
                matchBase: this._settings.baseNameMatch,
                nobrace: !this._settings.braceExpansion,
                nocase: !this._settings.caseSensitiveMatch,
                noext: !this._settings.extglob,
                noglobstar: !this._settings.globstar,
                posix: true,
                strictSlashes: false
            };
        }
    }
    exports.default = Provider;
},
/* fast-glob/out/providers/filters/deep.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const utils = require(1126) /* ../../utils */;
    const partial_1 = require(1183) /* ../matchers/partial */;
    class DeepFilter {
        constructor(_settings, _micromatchOptions) {
            this._settings = _settings;
            this._micromatchOptions = _micromatchOptions;
        }
        getFilter(basePath, positive, negative) {
            const matcher = this._getMatcher(positive);
            const negativeRe = this._getNegativePatternsRe(negative);
            return (entry) => this._filter(basePath, entry, matcher, negativeRe);
        }
        _getMatcher(patterns) {
            return new partial_1.default(patterns, this._settings, this._micromatchOptions);
        }
        _getNegativePatternsRe(patterns) {
            const affectDepthOfReadingPatterns = patterns.filter(utils.pattern.isAffectDepthOfReadingPattern);
            return utils.pattern.convertPatternsToRe(affectDepthOfReadingPatterns, this._micromatchOptions);
        }
        _filter(basePath, entry, matcher, negativeRe) {
            if (this._isSkippedByDeep(basePath, entry.path)) {
                return false;
            }
            if (this._isSkippedSymbolicLink(entry)) {
                return false;
            }
            const filepath = utils.path.removeLeadingDotSegment(entry.path);
            if (this._isSkippedByPositivePatterns(filepath, matcher)) {
                return false;
            }
            return this._isSkippedByNegativePatterns(filepath, negativeRe);
        }
        _isSkippedByDeep(basePath, entryPath) {
            /**
             * Avoid unnecessary depth calculations when it doesn't matter.
             */
            if (this._settings.deep === Infinity) {
                return false;
            }
            return this._getEntryLevel(basePath, entryPath) >= this._settings.deep;
        }
        _getEntryLevel(basePath, entryPath) {
            const entryPathDepth = entryPath.split('/').length;
            if (basePath === '') {
                return entryPathDepth;
            }
            const basePathDepth = basePath.split('/').length;
            return entryPathDepth - basePathDepth;
        }
        _isSkippedSymbolicLink(entry) {
            return !this._settings.followSymbolicLinks && entry.dirent.isSymbolicLink();
        }
        _isSkippedByPositivePatterns(entryPath, matcher) {
            return !this._settings.baseNameMatch && !matcher.match(entryPath);
        }
        _isSkippedByNegativePatterns(entryPath, patternsRe) {
            return !utils.pattern.matchAny(entryPath, patternsRe);
        }
    }
    exports.default = DeepFilter;
},
/* fast-glob/out/providers/matchers/partial.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const matcher_1 = require(1184) /* ./matcher */;
    class PartialMatcher extends matcher_1.default {
        match(filepath) {
            const parts = filepath.split('/');
            const levels = parts.length;
            const patterns = this._storage.filter((info) => !info.complete || info.segments.length > levels);
            for (const pattern of patterns) {
                const section = pattern.sections[0];
                /**
                 * In this case, the pattern has a globstar and we must read all directories unconditionally,
                 * but only if the level has reached the end of the first group.
                 *
                 * fixtures/{a,b}/**
                 *  ^ true/false  ^ always true
                */
                if (!pattern.complete && levels > section.length) {
                    return true;
                }
                const match = parts.every((part, index) => {
                    const segment = pattern.segments[index];
                    if (segment.dynamic && segment.patternRe.test(part)) {
                        return true;
                    }
                    if (!segment.dynamic && segment.pattern === part) {
                        return true;
                    }
                    return false;
                });
                if (match) {
                    return true;
                }
            }
            return false;
        }
    }
    exports.default = PartialMatcher;
},
/* fast-glob/out/providers/matchers/matcher.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const utils = require(1126) /* ../../utils */;
    class Matcher {
        constructor(_patterns, _settings, _micromatchOptions) {
            this._patterns = _patterns;
            this._settings = _settings;
            this._micromatchOptions = _micromatchOptions;
            this._storage = [];
            this._fillStorage();
        }
        _fillStorage() {
            /**
             * The original pattern may include `{,*,**,a/*}`, which will lead to problems with matching (unresolved level).
             * So, before expand patterns with brace expansion into separated patterns.
             */
            const patterns = utils.pattern.expandPatternsWithBraceExpansion(this._patterns);
            for (const pattern of patterns) {
                const segments = this._getPatternSegments(pattern);
                const sections = this._splitSegmentsIntoSections(segments);
                this._storage.push({
                    complete: sections.length <= 1,
                    pattern,
                    segments,
                    sections
                });
            }
        }
        _getPatternSegments(pattern) {
            const parts = utils.pattern.getPatternParts(pattern, this._micromatchOptions);
            return parts.map((part) => {
                const dynamic = utils.pattern.isDynamicPattern(part, this._settings);
                if (!dynamic) {
                    return {
                        dynamic: false,
                        pattern: part
                    };
                }
                return {
                    dynamic: true,
                    pattern: part,
                    patternRe: utils.pattern.makeRe(part, this._micromatchOptions)
                };
            });
        }
        _splitSegmentsIntoSections(segments) {
            return utils.array.splitWhen(segments, (segment) => segment.dynamic && utils.pattern.hasGlobStar(segment.pattern));
        }
    }
    exports.default = Matcher;
},
/* fast-glob/out/providers/filters/entry.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const utils = require(1126) /* ../../utils */;
    class EntryFilter {
        constructor(_settings, _micromatchOptions) {
            this._settings = _settings;
            this._micromatchOptions = _micromatchOptions;
            this.index = new Map();
        }
        getFilter(positive, negative) {
            const positiveRe = utils.pattern.convertPatternsToRe(positive, this._micromatchOptions);
            const negativeRe = utils.pattern.convertPatternsToRe(negative, this._micromatchOptions);
            return (entry) => this._filter(entry, positiveRe, negativeRe);
        }
        _filter(entry, positiveRe, negativeRe) {
            if (this._settings.unique && this._isDuplicateEntry(entry)) {
                return false;
            }
            if (this._onlyFileFilter(entry) || this._onlyDirectoryFilter(entry)) {
                return false;
            }
            if (this._isSkippedByAbsoluteNegativePatterns(entry.path, negativeRe)) {
                return false;
            }
            const filepath = this._settings.baseNameMatch ? entry.name : entry.path;
            const isMatched = this._isMatchToPatterns(filepath, positiveRe) && !this._isMatchToPatterns(entry.path, negativeRe);
            if (this._settings.unique && isMatched) {
                this._createIndexRecord(entry);
            }
            return isMatched;
        }
        _isDuplicateEntry(entry) {
            return this.index.has(entry.path);
        }
        _createIndexRecord(entry) {
            this.index.set(entry.path, undefined);
        }
        _onlyFileFilter(entry) {
            return this._settings.onlyFiles && !entry.dirent.isFile();
        }
        _onlyDirectoryFilter(entry) {
            return this._settings.onlyDirectories && !entry.dirent.isDirectory();
        }
        _isSkippedByAbsoluteNegativePatterns(entryPath, patternsRe) {
            if (!this._settings.absolute) {
                return false;
            }
            const fullpath = utils.path.makeAbsolute(this._settings.cwd, entryPath);
            return utils.pattern.matchAny(fullpath, patternsRe);
        }
        _isMatchToPatterns(entryPath, patternsRe) {
            const filepath = utils.path.removeLeadingDotSegment(entryPath);
            return utils.pattern.matchAny(filepath, patternsRe);
        }
    }
    exports.default = EntryFilter;
},
/* fast-glob/out/providers/filters/error.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const utils = require(1126) /* ../../utils */;
    class ErrorFilter {
        constructor(_settings) {
            this._settings = _settings;
        }
        getFilter() {
            return (error) => this._isNonFatalError(error);
        }
        _isNonFatalError(error) {
            return utils.errno.isEnoentCodeError(error) || this._settings.suppressErrors;
        }
    }
    exports.default = ErrorFilter;
},
/* fast-glob/out/providers/transformers/entry.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const utils = require(1126) /* ../../utils */;
    class EntryTransformer {
        constructor(_settings) {
            this._settings = _settings;
        }
        getTransformer() {
            return (entry) => this._transform(entry);
        }
        _transform(entry) {
            let filepath = entry.path;
            if (this._settings.absolute) {
                filepath = utils.path.makeAbsolute(this._settings.cwd, filepath);
                filepath = utils.path.unixify(filepath);
            }
            if (this._settings.markDirectories && entry.dirent.isDirectory()) {
                filepath += '/';
            }
            if (!this._settings.objectMode) {
                return filepath;
            }
            return Object.assign(Object.assign({}, entry), { path: filepath });
        }
    }
    exports.default = EntryTransformer;
},
/* fast-glob/out/providers/stream.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const stream_1 = require("stream");
    const stream_2 = require(1152) /* ../readers/stream */;
    const provider_1 = require(1181) /* ./provider */;
    class ProviderStream extends provider_1.default {
        constructor() {
            super(...arguments);
            this._reader = new stream_2.default(this._settings);
        }
        read(task) {
            const root = this._getRootDirectory(task);
            const options = this._getReaderOptions(task);
            const source = this.api(root, task, options);
            const destination = new stream_1.Readable({ objectMode: true, read: () => { } });
            source
                .once('error', (error) => destination.emit('error', error))
                .on('data', (entry) => destination.emit('data', options.transform(entry)))
                .once('end', () => destination.emit('end'));
            destination
                .once('close', () => source.destroy());
            return destination;
        }
        api(root, task, options) {
            if (task.dynamic) {
                return this._reader.dynamic(root, options);
            }
            return this._reader.static(task.patterns, options);
        }
    }
    exports.default = ProviderStream;
},
/* fast-glob/out/providers/sync.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const sync_1 = require(1190) /* ../readers/sync */;
    const provider_1 = require(1181) /* ./provider */;
    class ProviderSync extends provider_1.default {
        constructor() {
            super(...arguments);
            this._reader = new sync_1.default(this._settings);
        }
        read(task) {
            const root = this._getRootDirectory(task);
            const options = this._getReaderOptions(task);
            const entries = this.api(root, task, options);
            return entries.map(options.transform);
        }
        api(root, task, options) {
            if (task.dynamic) {
                return this._reader.dynamic(root, options);
            }
            return this._reader.static(task.patterns, options);
        }
    }
    exports.default = ProviderSync;
},
/* fast-glob/out/readers/sync.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const fsStat = require(1153) /* @nodelib/fs.stat */;
    const fsWalk = require(1158) /* @nodelib/fs.walk */;
    const reader_1 = require(1180) /* ./reader */;
    class ReaderSync extends reader_1.default {
        constructor() {
            super(...arguments);
            this._walkSync = fsWalk.walkSync;
            this._statSync = fsStat.statSync;
        }
        dynamic(root, options) {
            return this._walkSync(root, options);
        }
        static(patterns, options) {
            const entries = [];
            for (const pattern of patterns) {
                const filepath = this._getFullEntryPath(pattern);
                const entry = this._getEntry(filepath, pattern, options);
                if (entry === null || !options.entryFilter(entry)) {
                    continue;
                }
                entries.push(entry);
            }
            return entries;
        }
        _getEntry(filepath, pattern, options) {
            try {
                const stats = this._getStat(filepath);
                return this._makeEntry(stats, pattern);
            }
            catch (error) {
                if (options.errorFilter(error)) {
                    return null;
                }
                throw error;
            }
        }
        _getStat(filepath) {
            return this._statSync(filepath, this._fsStatSettings);
        }
    }
    exports.default = ReaderSync;
},
/* fast-glob/out/settings.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const fs = require("fs");
    const os = require("os");
    /**
     * The `os.cpus` method can return zero. We expect the number of cores to be greater than zero.
     * https://github.com/nodejs/node/blob/7faeddf23a98c53896f8b574a6e66589e8fb1eb8/lib/os.js#L106-L107
     */
    const CPU_COUNT = Math.max(os.cpus().length, 1);
    exports.DEFAULT_FILE_SYSTEM_ADAPTER = {
        lstat: fs.lstat,
        lstatSync: fs.lstatSync,
        stat: fs.stat,
        statSync: fs.statSync,
        readdir: fs.readdir,
        readdirSync: fs.readdirSync
    };
    class Settings {
        constructor(_options = {}) {
            this._options = _options;
            this.absolute = this._getValue(this._options.absolute, false);
            this.baseNameMatch = this._getValue(this._options.baseNameMatch, false);
            this.braceExpansion = this._getValue(this._options.braceExpansion, true);
            this.caseSensitiveMatch = this._getValue(this._options.caseSensitiveMatch, true);
            this.concurrency = this._getValue(this._options.concurrency, CPU_COUNT);
            this.cwd = this._getValue(this._options.cwd, process.cwd());
            this.deep = this._getValue(this._options.deep, Infinity);
            this.dot = this._getValue(this._options.dot, false);
            this.extglob = this._getValue(this._options.extglob, true);
            this.followSymbolicLinks = this._getValue(this._options.followSymbolicLinks, true);
            this.fs = this._getFileSystemMethods(this._options.fs);
            this.globstar = this._getValue(this._options.globstar, true);
            this.ignore = this._getValue(this._options.ignore, []);
            this.markDirectories = this._getValue(this._options.markDirectories, false);
            this.objectMode = this._getValue(this._options.objectMode, false);
            this.onlyDirectories = this._getValue(this._options.onlyDirectories, false);
            this.onlyFiles = this._getValue(this._options.onlyFiles, true);
            this.stats = this._getValue(this._options.stats, false);
            this.suppressErrors = this._getValue(this._options.suppressErrors, false);
            this.throwErrorOnBrokenSymbolicLink = this._getValue(this._options.throwErrorOnBrokenSymbolicLink, false);
            this.unique = this._getValue(this._options.unique, true);
            if (this.onlyDirectories) {
                this.onlyFiles = false;
            }
            if (this.stats) {
                this.objectMode = true;
            }
        }
        _getValue(option, value) {
            return option === undefined ? value : option;
        }
        _getFileSystemMethods(methods = {}) {
            return Object.assign(Object.assign({}, exports.DEFAULT_FILE_SYSTEM_ADAPTER), methods);
        }
    }
    exports.default = Settings;
},
/* dir-glob/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const path = require('path');
    const pathType = require(1193) /* path-type */;
    const getExtensions = extensions => extensions.length > 1 ? `{${extensions.join(',')}}` : extensions[0];
    const getPath = (filepath, cwd) => {
        const pth = filepath[0] === '!' ? filepath.slice(1) : filepath;
        return path.isAbsolute(pth) ? pth : path.join(cwd, pth);
    };
    const addExtensions = (file, extensions) => {
        if (path.extname(file)) {
            return `**/${file}`;
        }
        return `**/${file}.${getExtensions(extensions)}`;
    };
    const getGlob = (directory, options) => {
        if (options.files && !Array.isArray(options.files)) {
            throw new TypeError(`Expected \`files\` to be of type \`Array\` but received type \`${typeof options.files}\``);
        }
        if (options.extensions && !Array.isArray(options.extensions)) {
            throw new TypeError(`Expected \`extensions\` to be of type \`Array\` but received type \`${typeof options.extensions}\``);
        }
        if (options.files && options.extensions) {
            return options.files.map(x => path.posix.join(directory, addExtensions(x, options.extensions)));
        }
        if (options.files) {
            return options.files.map(x => path.posix.join(directory, `**/${x}`));
        }
        if (options.extensions) {
            return [path.posix.join(directory, `**/*.${getExtensions(options.extensions)}`)];
        }
        return [path.posix.join(directory, '**')];
    };
    module.exports = async (input, options) => {
        options = {
            cwd: process.cwd(),
            ...options
        };
        if (typeof options.cwd !== 'string') {
            throw new TypeError(`Expected \`cwd\` to be of type \`string\` but received type \`${typeof options.cwd}\``);
        }
        const globs = await Promise.all([].concat(input).map(async (x) => {
            const isDirectory = await pathType.isDirectory(getPath(x, options.cwd));
            return isDirectory ? getGlob(x, options) : x;
        }));
        return [].concat.apply([], globs); // eslint-disable-line prefer-spread
    };
    module.exports.sync = (input, options) => {
        options = {
            cwd: process.cwd(),
            ...options
        };
        if (typeof options.cwd !== 'string') {
            throw new TypeError(`Expected \`cwd\` to be of type \`string\` but received type \`${typeof options.cwd}\``);
        }
        const globs = [].concat(input).map(x => pathType.isDirectorySync(getPath(x, options.cwd)) ? getGlob(x, options) : x);
        return [].concat.apply([], globs); // eslint-disable-line prefer-spread
    };
},
/* path-type/index.js */ function _(require, module, exports, __esModule, __esExport) {
    const { promisify } = require('util');
    const fs = require('fs');
    async function isType(fsStatType, statsMethodName, filePath) {
        if (typeof filePath !== 'string') {
            throw new TypeError(`Expected a string, got ${typeof filePath}`);
        }
        try {
            const stats = await promisify(fs[fsStatType])(filePath);
            return stats[statsMethodName]();
        }
        catch (error) {
            if (error.code === 'ENOENT') {
                return false;
            }
            throw error;
        }
    }
    function isTypeSync(fsStatType, statsMethodName, filePath) {
        if (typeof filePath !== 'string') {
            throw new TypeError(`Expected a string, got ${typeof filePath}`);
        }
        try {
            return fs[fsStatType](filePath)[statsMethodName]();
        }
        catch (error) {
            if (error.code === 'ENOENT') {
                return false;
            }
            throw error;
        }
    }
    exports.isFile = isType.bind(null, 'stat', 'isFile');
    exports.isDirectory = isType.bind(null, 'stat', 'isDirectory');
    exports.isSymlink = isType.bind(null, 'lstat', 'isSymbolicLink');
    exports.isFileSync = isTypeSync.bind(null, 'statSync', 'isFile');
    exports.isDirectorySync = isTypeSync.bind(null, 'statSync', 'isDirectory');
    exports.isSymlinkSync = isTypeSync.bind(null, 'lstatSync', 'isSymbolicLink');
},
/* globby/gitignore.js */ function _(require, module, exports, __esModule, __esExport) {
    const { promisify } = require('util');
    const fs = require('fs');
    const path = require('path');
    const fastGlob = require(1124) /* fast-glob */;
    const gitIgnore = require(276) /* ignore */;
    const slash = require(1195) /* slash */;
    const DEFAULT_IGNORE = [
        '**/node_modules/**',
        '**/flow-typed/**',
        '**/coverage/**',
        '**/.git'
    ];
    const readFileP = promisify(fs.readFile);
    const mapGitIgnorePatternTo = base => ignore => {
        if (ignore.startsWith('!')) {
            return '!' + path.posix.join(base, ignore.slice(1));
        }
        return path.posix.join(base, ignore);
    };
    const parseGitIgnore = (content, options) => {
        const base = slash(path.relative(options.cwd, path.dirname(options.fileName)));
        return content
            .split(/\r?\n/)
            .filter(Boolean)
            .filter(line => !line.startsWith('#'))
            .map(mapGitIgnorePatternTo(base));
    };
    const reduceIgnore = files => {
        const ignores = gitIgnore();
        for (const file of files) {
            ignores.add(parseGitIgnore(file.content, {
                cwd: file.cwd,
                fileName: file.filePath
            }));
        }
        return ignores;
    };
    const ensureAbsolutePathForCwd = (cwd, p) => {
        cwd = slash(cwd);
        if (path.isAbsolute(p)) {
            if (slash(p).startsWith(cwd)) {
                return p;
            }
            throw new Error(`Path ${p} is not in cwd ${cwd}`);
        }
        return path.join(cwd, p);
    };
    const getIsIgnoredPredecate = (ignores, cwd) => {
        return p => ignores.ignores(slash(path.relative(cwd, ensureAbsolutePathForCwd(cwd, p.path || p))));
    };
    const getFile = async (file, cwd) => {
        const filePath = path.join(cwd, file);
        const content = await readFileP(filePath, 'utf8');
        return {
            cwd,
            filePath,
            content
        };
    };
    const getFileSync = (file, cwd) => {
        const filePath = path.join(cwd, file);
        const content = fs.readFileSync(filePath, 'utf8');
        return {
            cwd,
            filePath,
            content
        };
    };
    const normalizeOptions = ({ ignore = [], cwd = slash(process.cwd()) } = {}) => {
        return { ignore, cwd };
    };
    module.exports = async (options) => {
        options = normalizeOptions(options);
        const paths = await fastGlob('**/.gitignore', {
            ignore: DEFAULT_IGNORE.concat(options.ignore),
            cwd: options.cwd
        });
        const files = await Promise.all(paths.map(file => getFile(file, options.cwd)));
        const ignores = reduceIgnore(files);
        return getIsIgnoredPredecate(ignores, options.cwd);
    };
    module.exports.sync = options => {
        options = normalizeOptions(options);
        const paths = fastGlob.sync('**/.gitignore', {
            ignore: DEFAULT_IGNORE.concat(options.ignore),
            cwd: options.cwd
        });
        const files = paths.map(file => getFileSync(file, options.cwd));
        const ignores = reduceIgnore(files);
        return getIsIgnoredPredecate(ignores, options.cwd);
    };
},
/* slash/index.js */ function _(require, module, exports, __esModule, __esExport) {
    module.exports = path => {
        const isExtendedLengthPath = /^\\\\\?\\/.test(path);
        const hasNonAscii = /[^\u0000-\u0080]+/.test(path); // eslint-disable-line no-control-regex
        if (isExtendedLengthPath || hasNonAscii) {
            return path;
        }
        return path.replace(/\\/g, '/');
    };
},
/* globby/stream-utils.js */ function _(require, module, exports, __esModule, __esExport) {
    const { Transform } = require('stream');
    class ObjectTransform extends Transform {
        constructor() {
            super({
                objectMode: true
            });
        }
    }
    class FilterStream extends ObjectTransform {
        constructor(filter) {
            super();
            this._filter = filter;
        }
        _transform(data, encoding, callback) {
            if (this._filter(data)) {
                this.push(data);
            }
            callback();
        }
    }
    class UniqueStream extends ObjectTransform {
        constructor() {
            super();
            this._pushed = new Set();
        }
        _transform(data, encoding, callback) {
            if (!this._pushed.has(data)) {
                this.push(data);
                this._pushed.add(data);
            }
            callback();
        }
    }
    module.exports = {
        FilterStream,
        UniqueStream
    };
},
/* @typescript-eslint/typescript-estree/dist/ast-converter.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const convert_1 = require(1198) /* ./convert */;
    const convert_comments_1 = require(1205) /* ./convert-comments */;
    const node_utils_1 = require(1199) /* ./node-utils */;
    const simple_traverse_1 = require(1206) /* ./simple-traverse */;
    function astConverter(ast, extra, shouldPreserveNodeMaps) {
        /**
         * The TypeScript compiler produced fundamental parse errors when parsing the
         * source.
         */
        const { parseDiagnostics } = ast;
        if (parseDiagnostics.length) {
            throw convert_1.convertError(parseDiagnostics[0]);
        }
        /**
         * Recursively convert the TypeScript AST into an ESTree-compatible AST
         */
        const instance = new convert_1.Converter(ast, {
            errorOnUnknownASTType: extra.errorOnUnknownASTType || false,
            useJSXTextNode: extra.useJSXTextNode || false,
            shouldPreserveNodeMaps,
        });
        const estree = instance.convertProgram();
        /**
         * Optionally remove range and loc if specified
         */
        if (!extra.range || !extra.loc) {
            simple_traverse_1.simpleTraverse(estree, {
                enter: node => {
                    if (!extra.range) {
                        // eslint-disable-next-line @typescript-eslint/ban-ts-comment -- TS 4.0 made this an error because the types aren't optional
                        // @ts-expect-error
                        delete node.range;
                    }
                    if (!extra.loc) {
                        // eslint-disable-next-line @typescript-eslint/ban-ts-comment -- TS 4.0 made this an error because the types aren't optional
                        // @ts-expect-error
                        delete node.loc;
                    }
                },
            });
        }
        /**
         * Optionally convert and include all tokens in the AST
         */
        if (extra.tokens) {
            estree.tokens = node_utils_1.convertTokens(ast);
        }
        /**
         * Optionally convert and include all comments in the AST
         */
        if (extra.comment) {
            estree.comments = convert_comments_1.convertComments(ast, extra.code);
        }
        const astMaps = instance.getASTMaps();
        return { estree, astMaps };
    }
    exports.astConverter = astConverter;
},
/* @typescript-eslint/typescript-estree/dist/convert.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    // There's lots of funny stuff due to the typing of ts.Node
    /* eslint-disable @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-call, @typescript-eslint/no-unsafe-return, @typescript-eslint/no-unsafe-member-access */
    const ts = __importStar(require(21) /* typescript */);
    const node_utils_1 = require(1199) /* ./node-utils */;
    const ts_estree_1 = require(1200) /* ./ts-estree */;
    const version_check_1 = require(1204) /* ./version-check */;
    const SyntaxKind = ts.SyntaxKind;
    /**
     * Extends and formats a given error object
     * @param error the error object
     * @returns converted error object
     */
    function convertError(error) {
        return node_utils_1.createError(error.file, error.start, error.message || error.messageText);
    }
    exports.convertError = convertError;
    class Converter {
        /**
         * Converts a TypeScript node into an ESTree node
         * @param ast the full TypeScript AST
         * @param options additional options for the conversion
         * @returns the converted ESTreeNode
         */
        constructor(ast, options) {
            this.esTreeNodeToTSNodeMap = new WeakMap();
            this.tsNodeToESTreeNodeMap = new WeakMap();
            this.allowPattern = false;
            this.inTypeMode = false;
            this.ast = ast;
            this.options = Object.assign({}, options);
        }
        getASTMaps() {
            return {
                esTreeNodeToTSNodeMap: this.esTreeNodeToTSNodeMap,
                tsNodeToESTreeNodeMap: this.tsNodeToESTreeNodeMap,
            };
        }
        convertProgram() {
            return this.converter(this.ast);
        }
        /**
         * Converts a TypeScript node into an ESTree node.
         * @param node the child ts.Node
         * @param parent parentNode
         * @param inTypeMode flag to determine if we are in typeMode
         * @param allowPattern flag to determine if patterns are allowed
         * @returns the converted ESTree node
         */
        converter(node, parent, inTypeMode, allowPattern) {
            /**
             * Exit early for null and undefined
             */
            if (!node) {
                return null;
            }
            const typeMode = this.inTypeMode;
            const pattern = this.allowPattern;
            if (inTypeMode !== undefined) {
                this.inTypeMode = inTypeMode;
            }
            if (allowPattern !== undefined) {
                this.allowPattern = allowPattern;
            }
            const result = this.convertNode(node, (parent !== null && parent !== void 0 ? parent : node.parent));
            this.registerTSNodeInNodeMap(node, result);
            this.inTypeMode = typeMode;
            this.allowPattern = pattern;
            return result;
        }
        /**
         * Fixes the exports of the given ts.Node
         * @param node the ts.Node
         * @param result result
         * @returns the ESTreeNode with fixed exports
         */
        fixExports(node, result) {
            // check for exports
            if (node.modifiers && node.modifiers[0].kind === SyntaxKind.ExportKeyword) {
                /**
                 * Make sure that original node is registered instead of export
                 */
                this.registerTSNodeInNodeMap(node, result);
                const exportKeyword = node.modifiers[0];
                const nextModifier = node.modifiers[1];
                const declarationIsDefault = nextModifier && nextModifier.kind === SyntaxKind.DefaultKeyword;
                const varToken = declarationIsDefault
                    ? node_utils_1.findNextToken(nextModifier, this.ast, this.ast)
                    : node_utils_1.findNextToken(exportKeyword, this.ast, this.ast);
                result.range[0] = varToken.getStart(this.ast);
                result.loc = node_utils_1.getLocFor(result.range[0], result.range[1], this.ast);
                if (declarationIsDefault) {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ExportDefaultDeclaration,
                        declaration: result,
                        range: [exportKeyword.getStart(this.ast), result.range[1]],
                        exportKind: 'value',
                    });
                }
                else {
                    const isType = result.type === ts_estree_1.AST_NODE_TYPES.TSInterfaceDeclaration ||
                        result.type === ts_estree_1.AST_NODE_TYPES.TSTypeAliasDeclaration;
                    const isDeclare = result.declare === true;
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ExportNamedDeclaration,
                        declaration: result,
                        specifiers: [],
                        source: null,
                        exportKind: isType || isDeclare ? 'type' : 'value',
                        range: [exportKeyword.getStart(this.ast), result.range[1]],
                    });
                }
            }
            return result;
        }
        /**
         * Register specific TypeScript node into map with first ESTree node provided
         */
        registerTSNodeInNodeMap(node, result) {
            if (result && this.options.shouldPreserveNodeMaps) {
                if (!this.tsNodeToESTreeNodeMap.has(node)) {
                    this.tsNodeToESTreeNodeMap.set(node, result);
                }
            }
        }
        /**
         * Converts a TypeScript node into an ESTree node.
         * @param child the child ts.Node
         * @param parent parentNode
         * @returns the converted ESTree node
         */
        convertPattern(child, parent) {
            return this.converter(child, parent, this.inTypeMode, true);
        }
        /**
         * Converts a TypeScript node into an ESTree node.
         * @param child the child ts.Node
         * @param parent parentNode
         * @returns the converted ESTree node
         */
        convertChild(child, parent) {
            return this.converter(child, parent, this.inTypeMode, false);
        }
        /**
         * Converts a TypeScript node into an ESTree node.
         * @param child the child ts.Node
         * @param parent parentNode
         * @returns the converted ESTree node
         */
        convertType(child, parent) {
            return this.converter(child, parent, true, false);
        }
        createNode(node, data) {
            const result = data;
            if (!result.range) {
                result.range = node_utils_1.getRange(
                // this is completely valid, but TS hates it
                node, this.ast);
            }
            if (!result.loc) {
                result.loc = node_utils_1.getLocFor(result.range[0], result.range[1], this.ast);
            }
            if (result && this.options.shouldPreserveNodeMaps) {
                this.esTreeNodeToTSNodeMap.set(result, node);
            }
            return result;
        }
        convertBindingNameWithTypeAnnotation(name, tsType, parent) {
            const id = this.convertPattern(name);
            if (tsType) {
                id.typeAnnotation = this.convertTypeAnnotation(tsType, parent);
                this.fixParentLocation(id, id.typeAnnotation.range);
            }
            return id;
        }
        /**
         * Converts a child into a type annotation. This creates an intermediary
         * TypeAnnotation node to match what Flow does.
         * @param child The TypeScript AST node to convert.
         * @param parent parentNode
         * @returns The type annotation node.
         */
        convertTypeAnnotation(child, parent) {
            // in FunctionType and ConstructorType typeAnnotation has 2 characters `=>` and in other places is just colon
            const offset = (parent === null || parent === void 0 ? void 0 : parent.kind) === SyntaxKind.FunctionType ||
                (parent === null || parent === void 0 ? void 0 : parent.kind) === SyntaxKind.ConstructorType
                ? 2
                : 1;
            const annotationStartCol = child.getFullStart() - offset;
            const loc = node_utils_1.getLocFor(annotationStartCol, child.end, this.ast);
            return {
                type: ts_estree_1.AST_NODE_TYPES.TSTypeAnnotation,
                loc,
                range: [annotationStartCol, child.end],
                typeAnnotation: this.convertType(child),
            };
        }
        /**
         * Coverts body Nodes and add a directive field to StringLiterals
         * @param nodes of ts.Node
         * @param parent parentNode
         * @returns Array of body statements
         */
        convertBodyExpressions(nodes, parent) {
            let allowDirectives = node_utils_1.canContainDirective(parent);
            return (nodes
                .map(statement => {
                const child = this.convertChild(statement);
                if (allowDirectives) {
                    if ((child === null || child === void 0 ? void 0 : child.expression) &&
                        ts.isExpressionStatement(statement) &&
                        ts.isStringLiteral(statement.expression)) {
                        const raw = child.expression.raw;
                        child.directive = raw.slice(1, -1);
                        return child; // child can be null, but it's filtered below
                    }
                    else {
                        allowDirectives = false;
                    }
                }
                return child; // child can be null, but it's filtered below
            })
                // filter out unknown nodes for now
                .filter(statement => statement));
        }
        /**
         * Converts a ts.Node's typeArguments to TSTypeParameterInstantiation node
         * @param typeArguments ts.NodeArray typeArguments
         * @param node parent used to create this node
         * @returns TypeParameterInstantiation node
         */
        convertTypeArgumentsToTypeParameters(typeArguments, node) {
            const greaterThanToken = node_utils_1.findNextToken(typeArguments, this.ast, this.ast);
            return this.createNode(node, {
                type: ts_estree_1.AST_NODE_TYPES.TSTypeParameterInstantiation,
                range: [typeArguments.pos - 1, greaterThanToken.end],
                params: typeArguments.map(typeArgument => this.convertType(typeArgument)),
            });
        }
        /**
         * Converts a ts.Node's typeParameters to TSTypeParameterDeclaration node
         * @param typeParameters ts.Node typeParameters
         * @returns TypeParameterDeclaration node
         */
        convertTSTypeParametersToTypeParametersDeclaration(typeParameters) {
            const greaterThanToken = node_utils_1.findNextToken(typeParameters, this.ast, this.ast);
            return {
                type: ts_estree_1.AST_NODE_TYPES.TSTypeParameterDeclaration,
                range: [typeParameters.pos - 1, greaterThanToken.end],
                loc: node_utils_1.getLocFor(typeParameters.pos - 1, greaterThanToken.end, this.ast),
                params: typeParameters.map(typeParameter => this.convertType(typeParameter)),
            };
        }
        /**
         * Converts an array of ts.Node parameters into an array of ESTreeNode params
         * @param parameters An array of ts.Node params to be converted
         * @returns an array of converted ESTreeNode params
         */
        convertParameters(parameters) {
            if (!parameters || !parameters.length) {
                return [];
            }
            return parameters.map(param => {
                var _a;
                const convertedParam = this.convertChild(param);
                if ((_a = param.decorators) === null || _a === void 0 ? void 0 : _a.length) {
                    convertedParam.decorators = param.decorators.map(el => this.convertChild(el));
                }
                return convertedParam;
            });
        }
        convertChainExpression(node, tsNode) {
            const { child, isOptional } = (() => {
                if (node.type === ts_estree_1.AST_NODE_TYPES.MemberExpression) {
                    return { child: node.object, isOptional: node.optional };
                }
                if (node.type === ts_estree_1.AST_NODE_TYPES.CallExpression) {
                    return { child: node.callee, isOptional: node.optional };
                }
                return { child: node.expression, isOptional: false };
            })();
            const isChildUnwrappable = node_utils_1.isChildUnwrappableOptionalChain(tsNode, child);
            if (!isChildUnwrappable && !isOptional) {
                return node;
            }
            if (isChildUnwrappable && node_utils_1.isChainExpression(child)) {
                // unwrap the chain expression child
                const newChild = child.expression;
                if (node.type === ts_estree_1.AST_NODE_TYPES.MemberExpression) {
                    node.object = newChild;
                }
                else if (node.type === ts_estree_1.AST_NODE_TYPES.CallExpression) {
                    node.callee = newChild;
                }
                else {
                    node.expression = newChild;
                }
            }
            return this.createNode(tsNode, {
                type: ts_estree_1.AST_NODE_TYPES.ChainExpression,
                expression: node,
            });
        }
        /**
         * For nodes that are copied directly from the TypeScript AST into
         * ESTree mostly as-is. The only difference is the addition of a type
         * property instead of a kind property. Recursively copies all children.
         */
        deeplyCopy(node) {
            if (node.kind === ts.SyntaxKind.JSDocFunctionType) {
                throw node_utils_1.createError(this.ast, node.pos, 'JSDoc types can only be used inside documentation comments.');
            }
            const customType = `TS${SyntaxKind[node.kind]}`;
            /**
             * If the "errorOnUnknownASTType" option is set to true, throw an error,
             * otherwise fallback to just including the unknown type as-is.
             */
            if (this.options.errorOnUnknownASTType && !ts_estree_1.AST_NODE_TYPES[customType]) {
                throw new Error(`Unknown AST_NODE_TYPE: "${customType}"`);
            }
            const result = this.createNode(node, {
                type: customType,
            });
            if ('type' in node) {
                result.typeAnnotation =
                    node.type && 'kind' in node.type && ts.isTypeNode(node.type)
                        ? this.convertTypeAnnotation(node.type, node)
                        : null;
            }
            if ('typeArguments' in node) {
                result.typeParameters =
                    node.typeArguments && 'pos' in node.typeArguments
                        ? this.convertTypeArgumentsToTypeParameters(node.typeArguments, node)
                        : null;
            }
            if ('typeParameters' in node) {
                result.typeParameters =
                    node.typeParameters && 'pos' in node.typeParameters
                        ? this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters)
                        : null;
            }
            if ('decorators' in node && node.decorators && node.decorators.length) {
                result.decorators = node.decorators.map(el => this.convertChild(el));
            }
            Object.entries(node)
                .filter(([key]) => !/^(?:_children|kind|parent|pos|end|flags|modifierFlagsCache|jsDoc|type|typeArguments|typeParameters|decorators|transformFlags)$/.test(key))
                .forEach(([key, value]) => {
                if (Array.isArray(value)) {
                    result[key] = value.map(el => this.convertChild(el));
                }
                else if (value && typeof value === 'object' && value.kind) {
                    // need to check node[key].kind to ensure we don't try to convert a symbol
                    result[key] = this.convertChild(value);
                }
                else {
                    result[key] = value;
                }
            });
            return result;
        }
        convertJSXIdentifier(node) {
            const result = this.createNode(node, {
                type: ts_estree_1.AST_NODE_TYPES.JSXIdentifier,
                name: node.getText(),
            });
            this.registerTSNodeInNodeMap(node, result);
            return result;
        }
        convertJSXNamespaceOrIdentifier(node) {
            const text = node.getText();
            const colonIndex = text.indexOf(':');
            // this is intentional we can ignore conversion if `:` is in first character
            if (colonIndex > 0) {
                const range = node_utils_1.getRange(node, this.ast);
                const result = this.createNode(node, {
                    type: ts_estree_1.AST_NODE_TYPES.JSXNamespacedName,
                    namespace: this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXIdentifier,
                        name: text.slice(0, colonIndex),
                        range: [range[0], range[0] + colonIndex],
                    }),
                    name: this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXIdentifier,
                        name: text.slice(colonIndex + 1),
                        range: [range[0] + colonIndex + 1, range[1]],
                    }),
                    range,
                });
                this.registerTSNodeInNodeMap(node, result);
                return result;
            }
            return this.convertJSXIdentifier(node);
        }
        /**
         * Converts a TypeScript JSX node.tagName into an ESTree node.name
         * @param node the tagName object from a JSX ts.Node
         * @param parent
         * @returns the converted ESTree name object
         */
        convertJSXTagName(node, parent) {
            let result;
            switch (node.kind) {
                case SyntaxKind.PropertyAccessExpression:
                    if (node.name.kind === SyntaxKind.PrivateIdentifier) {
                        // This is one of the few times where TS explicitly errors, and doesn't even gracefully handle the syntax.
                        // So we shouldn't ever get into this state to begin with.
                        throw new Error('Non-private identifier expected.');
                    }
                    result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXMemberExpression,
                        object: this.convertJSXTagName(node.expression, parent),
                        property: this.convertJSXIdentifier(node.name),
                    });
                    break;
                case SyntaxKind.ThisKeyword:
                case SyntaxKind.Identifier:
                default:
                    return this.convertJSXNamespaceOrIdentifier(node);
            }
            this.registerTSNodeInNodeMap(node, result);
            return result;
        }
        convertMethodSignature(node) {
            const result = this.createNode(node, {
                type: ts_estree_1.AST_NODE_TYPES.TSMethodSignature,
                computed: node_utils_1.isComputedProperty(node.name),
                key: this.convertChild(node.name),
                params: this.convertParameters(node.parameters),
                kind: (() => {
                    switch (node.kind) {
                        case SyntaxKind.GetAccessor:
                            return 'get';
                        case SyntaxKind.SetAccessor:
                            return 'set';
                        case SyntaxKind.MethodSignature:
                            return 'method';
                    }
                })(),
            });
            if (node_utils_1.isOptional(node)) {
                result.optional = true;
            }
            if (node.type) {
                result.returnType = this.convertTypeAnnotation(node.type, node);
            }
            if (node_utils_1.hasModifier(SyntaxKind.ReadonlyKeyword, node)) {
                result.readonly = true;
            }
            if (node.typeParameters) {
                result.typeParameters =
                    this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
            }
            const accessibility = node_utils_1.getTSNodeAccessibility(node);
            if (accessibility) {
                result.accessibility = accessibility;
            }
            if (node_utils_1.hasModifier(SyntaxKind.ExportKeyword, node)) {
                result.export = true;
            }
            if (node_utils_1.hasModifier(SyntaxKind.StaticKeyword, node)) {
                result.static = true;
            }
            return result;
        }
        /**
         * Applies the given TS modifiers to the given result object.
         * @param result
         * @param modifiers original ts.Nodes from the node.modifiers array
         * @returns the current result object will be mutated
         * @deprecated This method adds not standardized `modifiers` property in nodes
         */
        applyModifiersToResult(result, modifiers) {
            if (!modifiers || !modifiers.length) {
                return;
            }
            const remainingModifiers = [];
            /**
             * Some modifiers are explicitly handled by applying them as
             * boolean values on the result node. As well as adding them
             * to the result, we remove them from the array, so that they
             * are not handled twice.
             */
            for (let i = 0; i < modifiers.length; i++) {
                const modifier = modifiers[i];
                switch (modifier.kind) {
                    /**
                     * Ignore ExportKeyword and DefaultKeyword, they are handled
                     * via the fixExports utility function
                     */
                    case SyntaxKind.ExportKeyword:
                    case SyntaxKind.DefaultKeyword:
                        break;
                    case SyntaxKind.ConstKeyword:
                        result.const = true;
                        break;
                    case SyntaxKind.DeclareKeyword:
                        result.declare = true;
                        break;
                    default:
                        remainingModifiers.push(this.convertChild(modifier));
                        break;
                }
            }
            /**
             * If there are still valid modifiers available which have
             * not been explicitly handled above, we just convert and
             * add the modifiers array to the result node.
             */
            if (remainingModifiers.length) {
                result.modifiers = remainingModifiers;
            }
        }
        /**
         * Uses the provided range location to adjust the location data of the given Node
         * @param result The node that will have its location data mutated
         * @param childRange The child node range used to expand location
         */
        fixParentLocation(result, childRange) {
            if (childRange[0] < result.range[0]) {
                result.range[0] = childRange[0];
                result.loc.start = node_utils_1.getLineAndCharacterFor(result.range[0], this.ast);
            }
            if (childRange[1] > result.range[1]) {
                result.range[1] = childRange[1];
                result.loc.end = node_utils_1.getLineAndCharacterFor(result.range[1], this.ast);
            }
        }
        /**
         * Converts a TypeScript node into an ESTree node.
         * The core of the conversion logic:
         * Identify and convert each relevant TypeScript SyntaxKind
         * @param node the child ts.Node
         * @param parent parentNode
         * @returns the converted ESTree node
         */
        convertNode(node, parent) {
            var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
            switch (node.kind) {
                case SyntaxKind.SourceFile: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Program,
                        body: this.convertBodyExpressions(node.statements, node),
                        sourceType: node.externalModuleIndicator ? 'module' : 'script',
                        range: [node.getStart(this.ast), node.endOfFileToken.end],
                    });
                }
                case SyntaxKind.Block: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.BlockStatement,
                        body: this.convertBodyExpressions(node.statements, node),
                    });
                }
                case SyntaxKind.Identifier: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Identifier,
                        name: node.text,
                    });
                }
                case SyntaxKind.WithStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.WithStatement,
                        object: this.convertChild(node.expression),
                        body: this.convertChild(node.statement),
                    });
                // Control Flow
                case SyntaxKind.ReturnStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ReturnStatement,
                        argument: this.convertChild(node.expression),
                    });
                case SyntaxKind.LabeledStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.LabeledStatement,
                        label: this.convertChild(node.label),
                        body: this.convertChild(node.statement),
                    });
                case SyntaxKind.ContinueStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ContinueStatement,
                        label: this.convertChild(node.label),
                    });
                case SyntaxKind.BreakStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.BreakStatement,
                        label: this.convertChild(node.label),
                    });
                // Choice
                case SyntaxKind.IfStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.IfStatement,
                        test: this.convertChild(node.expression),
                        consequent: this.convertChild(node.thenStatement),
                        alternate: this.convertChild(node.elseStatement),
                    });
                case SyntaxKind.SwitchStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.SwitchStatement,
                        discriminant: this.convertChild(node.expression),
                        cases: node.caseBlock.clauses.map(el => this.convertChild(el)),
                    });
                case SyntaxKind.CaseClause:
                case SyntaxKind.DefaultClause:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.SwitchCase,
                        // expression is present in case only
                        test: node.kind === SyntaxKind.CaseClause
                            ? this.convertChild(node.expression)
                            : null,
                        consequent: node.statements.map(el => this.convertChild(el)),
                    });
                // Exceptions
                case SyntaxKind.ThrowStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ThrowStatement,
                        argument: this.convertChild(node.expression),
                    });
                case SyntaxKind.TryStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TryStatement,
                        block: this.convertChild(node.tryBlock),
                        handler: this.convertChild(node.catchClause),
                        finalizer: this.convertChild(node.finallyBlock),
                    });
                case SyntaxKind.CatchClause:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.CatchClause,
                        param: node.variableDeclaration
                            ? this.convertBindingNameWithTypeAnnotation(node.variableDeclaration.name, node.variableDeclaration.type)
                            : null,
                        body: this.convertChild(node.block),
                    });
                // Loops
                case SyntaxKind.WhileStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.WhileStatement,
                        test: this.convertChild(node.expression),
                        body: this.convertChild(node.statement),
                    });
                /**
                 * Unlike other parsers, TypeScript calls a "DoWhileStatement"
                 * a "DoStatement"
                 */
                case SyntaxKind.DoStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.DoWhileStatement,
                        test: this.convertChild(node.expression),
                        body: this.convertChild(node.statement),
                    });
                case SyntaxKind.ForStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ForStatement,
                        init: this.convertChild(node.initializer),
                        test: this.convertChild(node.condition),
                        update: this.convertChild(node.incrementor),
                        body: this.convertChild(node.statement),
                    });
                case SyntaxKind.ForInStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ForInStatement,
                        left: this.convertPattern(node.initializer),
                        right: this.convertChild(node.expression),
                        body: this.convertChild(node.statement),
                    });
                case SyntaxKind.ForOfStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ForOfStatement,
                        left: this.convertPattern(node.initializer),
                        right: this.convertChild(node.expression),
                        body: this.convertChild(node.statement),
                        await: Boolean(node.awaitModifier &&
                            node.awaitModifier.kind === SyntaxKind.AwaitKeyword),
                    });
                // Declarations
                case SyntaxKind.FunctionDeclaration: {
                    const isDeclare = node_utils_1.hasModifier(SyntaxKind.DeclareKeyword, node);
                    const result = this.createNode(node, {
                        type: isDeclare || !node.body
                            ? ts_estree_1.AST_NODE_TYPES.TSDeclareFunction
                            : ts_estree_1.AST_NODE_TYPES.FunctionDeclaration,
                        id: this.convertChild(node.name),
                        generator: !!node.asteriskToken,
                        expression: false,
                        async: node_utils_1.hasModifier(SyntaxKind.AsyncKeyword, node),
                        params: this.convertParameters(node.parameters),
                        body: this.convertChild(node.body) || undefined,
                    });
                    // Process returnType
                    if (node.type) {
                        result.returnType = this.convertTypeAnnotation(node.type, node);
                    }
                    if (isDeclare) {
                        result.declare = true;
                    }
                    // Process typeParameters
                    if (node.typeParameters) {
                        result.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                    }
                    // check for exports
                    return this.fixExports(node, result);
                }
                case SyntaxKind.VariableDeclaration: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.VariableDeclarator,
                        id: this.convertBindingNameWithTypeAnnotation(node.name, node.type, node),
                        init: this.convertChild(node.initializer),
                    });
                    if (node.exclamationToken) {
                        result.definite = true;
                    }
                    return result;
                }
                case SyntaxKind.VariableStatement: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.VariableDeclaration,
                        declarations: node.declarationList.declarations.map(el => this.convertChild(el)),
                        kind: node_utils_1.getDeclarationKind(node.declarationList),
                    });
                    /**
                     * Semantically, decorators are not allowed on variable declarations,
                     * but the TypeScript compiler will parse them and produce a valid AST,
                     * so we handle them here too.
                     */
                    if (node.decorators) {
                        result.decorators = node.decorators.map(el => this.convertChild(el));
                    }
                    if (node_utils_1.hasModifier(SyntaxKind.DeclareKeyword, node)) {
                        result.declare = true;
                    }
                    // check for exports
                    return this.fixExports(node, result);
                }
                // mostly for for-of, for-in
                case SyntaxKind.VariableDeclarationList:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.VariableDeclaration,
                        declarations: node.declarations.map(el => this.convertChild(el)),
                        kind: node_utils_1.getDeclarationKind(node),
                    });
                // Expressions
                case SyntaxKind.ExpressionStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ExpressionStatement,
                        expression: this.convertChild(node.expression),
                    });
                case SyntaxKind.ThisKeyword:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ThisExpression,
                    });
                case SyntaxKind.ArrayLiteralExpression: {
                    // TypeScript uses ArrayLiteralExpression in destructuring assignment, too
                    if (this.allowPattern) {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.ArrayPattern,
                            elements: node.elements.map(el => this.convertPattern(el)),
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.ArrayExpression,
                            elements: node.elements.map(el => this.convertChild(el)),
                        });
                    }
                }
                case SyntaxKind.ObjectLiteralExpression: {
                    // TypeScript uses ObjectLiteralExpression in destructuring assignment, too
                    if (this.allowPattern) {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.ObjectPattern,
                            properties: node.properties.map(el => this.convertPattern(el)),
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.ObjectExpression,
                            properties: node.properties.map(el => this.convertChild(el)),
                        });
                    }
                }
                case SyntaxKind.PropertyAssignment:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Property,
                        key: this.convertChild(node.name),
                        value: this.converter(node.initializer, node, this.inTypeMode, this.allowPattern),
                        computed: node_utils_1.isComputedProperty(node.name),
                        method: false,
                        shorthand: false,
                        kind: 'init',
                    });
                case SyntaxKind.ShorthandPropertyAssignment: {
                    if (node.objectAssignmentInitializer) {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.Property,
                            key: this.convertChild(node.name),
                            value: this.createNode(node, {
                                type: ts_estree_1.AST_NODE_TYPES.AssignmentPattern,
                                left: this.convertPattern(node.name),
                                right: this.convertChild(node.objectAssignmentInitializer),
                            }),
                            computed: false,
                            method: false,
                            shorthand: true,
                            kind: 'init',
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.Property,
                            key: this.convertChild(node.name),
                            value: this.convertChild(node.name),
                            computed: false,
                            method: false,
                            shorthand: true,
                            kind: 'init',
                        });
                    }
                }
                case SyntaxKind.ComputedPropertyName:
                    return this.convertChild(node.expression);
                case SyntaxKind.PropertyDeclaration: {
                    const isAbstract = node_utils_1.hasModifier(SyntaxKind.AbstractKeyword, node);
                    const result = this.createNode(node, {
                        type: isAbstract
                            ? ts_estree_1.AST_NODE_TYPES.TSAbstractClassProperty
                            : ts_estree_1.AST_NODE_TYPES.ClassProperty,
                        key: this.convertChild(node.name),
                        value: this.convertChild(node.initializer),
                        computed: node_utils_1.isComputedProperty(node.name),
                        static: node_utils_1.hasModifier(SyntaxKind.StaticKeyword, node),
                        readonly: node_utils_1.hasModifier(SyntaxKind.ReadonlyKeyword, node) || undefined,
                        declare: node_utils_1.hasModifier(SyntaxKind.DeclareKeyword, node),
                        override: node_utils_1.hasModifier(SyntaxKind.OverrideKeyword, node),
                    });
                    if (node.type) {
                        result.typeAnnotation = this.convertTypeAnnotation(node.type, node);
                    }
                    if (node.decorators) {
                        result.decorators = node.decorators.map(el => this.convertChild(el));
                    }
                    const accessibility = node_utils_1.getTSNodeAccessibility(node);
                    if (accessibility) {
                        result.accessibility = accessibility;
                    }
                    if ((node.name.kind === SyntaxKind.Identifier ||
                        node.name.kind === SyntaxKind.ComputedPropertyName) &&
                        node.questionToken) {
                        result.optional = true;
                    }
                    if (node.exclamationToken) {
                        result.definite = true;
                    }
                    if (result.key.type === ts_estree_1.AST_NODE_TYPES.Literal && node.questionToken) {
                        result.optional = true;
                    }
                    return result;
                }
                case SyntaxKind.GetAccessor:
                case SyntaxKind.SetAccessor: {
                    if (node.parent.kind === SyntaxKind.InterfaceDeclaration ||
                        node.parent.kind === SyntaxKind.TypeLiteral) {
                        return this.convertMethodSignature(node);
                    }
                }
                // otherwise, it is a non-type accessor - intentional fallthrough
                case SyntaxKind.MethodDeclaration: {
                    const method = this.createNode(node, {
                        type: !node.body
                            ? ts_estree_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression
                            : ts_estree_1.AST_NODE_TYPES.FunctionExpression,
                        id: null,
                        generator: !!node.asteriskToken,
                        expression: false,
                        async: node_utils_1.hasModifier(SyntaxKind.AsyncKeyword, node),
                        body: this.convertChild(node.body),
                        range: [node.parameters.pos - 1, node.end],
                        params: [],
                    });
                    if (node.type) {
                        method.returnType = this.convertTypeAnnotation(node.type, node);
                    }
                    // Process typeParameters
                    if (node.typeParameters) {
                        method.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                        this.fixParentLocation(method, method.typeParameters.range);
                    }
                    let result;
                    if (parent.kind === SyntaxKind.ObjectLiteralExpression) {
                        method.params = node.parameters.map(el => this.convertChild(el));
                        result = this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.Property,
                            key: this.convertChild(node.name),
                            value: method,
                            computed: node_utils_1.isComputedProperty(node.name),
                            method: node.kind === SyntaxKind.MethodDeclaration,
                            shorthand: false,
                            kind: 'init',
                        });
                    }
                    else {
                        // class
                        /**
                         * Unlike in object literal methods, class method params can have decorators
                         */
                        method.params = this.convertParameters(node.parameters);
                        /**
                         * TypeScript class methods can be defined as "abstract"
                         */
                        const methodDefinitionType = node_utils_1.hasModifier(SyntaxKind.AbstractKeyword, node)
                            ? ts_estree_1.AST_NODE_TYPES.TSAbstractMethodDefinition
                            : ts_estree_1.AST_NODE_TYPES.MethodDefinition;
                        result = this.createNode(node, {
                            type: methodDefinitionType,
                            key: this.convertChild(node.name),
                            value: method,
                            computed: node_utils_1.isComputedProperty(node.name),
                            static: node_utils_1.hasModifier(SyntaxKind.StaticKeyword, node),
                            kind: 'method',
                            override: node_utils_1.hasModifier(SyntaxKind.OverrideKeyword, node),
                        });
                        if (node.decorators) {
                            result.decorators = node.decorators.map(el => this.convertChild(el));
                        }
                        const accessibility = node_utils_1.getTSNodeAccessibility(node);
                        if (accessibility) {
                            result.accessibility = accessibility;
                        }
                    }
                    if (node.questionToken) {
                        result.optional = true;
                    }
                    if (node.kind === SyntaxKind.GetAccessor) {
                        result.kind = 'get';
                    }
                    else if (node.kind === SyntaxKind.SetAccessor) {
                        result.kind = 'set';
                    }
                    else if (!result.static &&
                        node.name.kind === SyntaxKind.StringLiteral &&
                        node.name.text === 'constructor' &&
                        result.type !== ts_estree_1.AST_NODE_TYPES.Property) {
                        result.kind = 'constructor';
                    }
                    return result;
                }
                // TypeScript uses this even for static methods named "constructor"
                case SyntaxKind.Constructor: {
                    const lastModifier = node_utils_1.getLastModifier(node);
                    const constructorToken = (lastModifier && node_utils_1.findNextToken(lastModifier, node, this.ast)) ||
                        node.getFirstToken();
                    const constructor = this.createNode(node, {
                        type: !node.body
                            ? ts_estree_1.AST_NODE_TYPES.TSEmptyBodyFunctionExpression
                            : ts_estree_1.AST_NODE_TYPES.FunctionExpression,
                        id: null,
                        params: this.convertParameters(node.parameters),
                        generator: false,
                        expression: false,
                        async: false,
                        body: this.convertChild(node.body),
                        range: [node.parameters.pos - 1, node.end],
                    });
                    // Process typeParameters
                    if (node.typeParameters) {
                        constructor.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                        this.fixParentLocation(constructor, constructor.typeParameters.range);
                    }
                    // Process returnType
                    if (node.type) {
                        constructor.returnType = this.convertTypeAnnotation(node.type, node);
                    }
                    const constructorKey = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Identifier,
                        name: 'constructor',
                        range: [constructorToken.getStart(this.ast), constructorToken.end],
                    });
                    const isStatic = node_utils_1.hasModifier(SyntaxKind.StaticKeyword, node);
                    const result = this.createNode(node, {
                        type: node_utils_1.hasModifier(SyntaxKind.AbstractKeyword, node)
                            ? ts_estree_1.AST_NODE_TYPES.TSAbstractMethodDefinition
                            : ts_estree_1.AST_NODE_TYPES.MethodDefinition,
                        key: constructorKey,
                        value: constructor,
                        computed: false,
                        static: isStatic,
                        kind: isStatic ? 'method' : 'constructor',
                        override: false,
                    });
                    const accessibility = node_utils_1.getTSNodeAccessibility(node);
                    if (accessibility) {
                        result.accessibility = accessibility;
                    }
                    return result;
                }
                case SyntaxKind.FunctionExpression: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.FunctionExpression,
                        id: this.convertChild(node.name),
                        generator: !!node.asteriskToken,
                        params: this.convertParameters(node.parameters),
                        body: this.convertChild(node.body),
                        async: node_utils_1.hasModifier(SyntaxKind.AsyncKeyword, node),
                        expression: false,
                    });
                    // Process returnType
                    if (node.type) {
                        result.returnType = this.convertTypeAnnotation(node.type, node);
                    }
                    // Process typeParameters
                    if (node.typeParameters) {
                        result.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                    }
                    return result;
                }
                case SyntaxKind.SuperKeyword:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Super,
                    });
                case SyntaxKind.ArrayBindingPattern:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ArrayPattern,
                        elements: node.elements.map(el => this.convertPattern(el)),
                    });
                // occurs with missing array elements like [,]
                case SyntaxKind.OmittedExpression:
                    return null;
                case SyntaxKind.ObjectBindingPattern:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ObjectPattern,
                        properties: node.elements.map(el => this.convertPattern(el)),
                    });
                case SyntaxKind.BindingElement: {
                    if (parent.kind === SyntaxKind.ArrayBindingPattern) {
                        const arrayItem = this.convertChild(node.name, parent);
                        if (node.initializer) {
                            return this.createNode(node, {
                                type: ts_estree_1.AST_NODE_TYPES.AssignmentPattern,
                                left: arrayItem,
                                right: this.convertChild(node.initializer),
                            });
                        }
                        else if (node.dotDotDotToken) {
                            return this.createNode(node, {
                                type: ts_estree_1.AST_NODE_TYPES.RestElement,
                                argument: arrayItem,
                            });
                        }
                        else {
                            return arrayItem;
                        }
                    }
                    else {
                        let result;
                        if (node.dotDotDotToken) {
                            result = this.createNode(node, {
                                type: ts_estree_1.AST_NODE_TYPES.RestElement,
                                argument: this.convertChild((_a = node.propertyName) !== null && _a !== void 0 ? _a : node.name),
                            });
                        }
                        else {
                            result = this.createNode(node, {
                                type: ts_estree_1.AST_NODE_TYPES.Property,
                                key: this.convertChild((_b = node.propertyName) !== null && _b !== void 0 ? _b : node.name),
                                value: this.convertChild(node.name),
                                computed: Boolean(node.propertyName &&
                                    node.propertyName.kind === SyntaxKind.ComputedPropertyName),
                                method: false,
                                shorthand: !node.propertyName,
                                kind: 'init',
                            });
                        }
                        if (node.initializer) {
                            result.value = this.createNode(node, {
                                type: ts_estree_1.AST_NODE_TYPES.AssignmentPattern,
                                left: this.convertChild(node.name),
                                right: this.convertChild(node.initializer),
                                range: [node.name.getStart(this.ast), node.initializer.end],
                            });
                        }
                        return result;
                    }
                }
                case SyntaxKind.ArrowFunction: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ArrowFunctionExpression,
                        generator: false,
                        id: null,
                        params: this.convertParameters(node.parameters),
                        body: this.convertChild(node.body),
                        async: node_utils_1.hasModifier(SyntaxKind.AsyncKeyword, node),
                        expression: node.body.kind !== SyntaxKind.Block,
                    });
                    // Process returnType
                    if (node.type) {
                        result.returnType = this.convertTypeAnnotation(node.type, node);
                    }
                    // Process typeParameters
                    if (node.typeParameters) {
                        result.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                    }
                    return result;
                }
                case SyntaxKind.YieldExpression:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.YieldExpression,
                        delegate: !!node.asteriskToken,
                        argument: this.convertChild(node.expression),
                    });
                case SyntaxKind.AwaitExpression:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.AwaitExpression,
                        argument: this.convertChild(node.expression),
                    });
                // Template Literals
                case SyntaxKind.NoSubstitutionTemplateLiteral:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TemplateLiteral,
                        quasis: [
                            this.createNode(node, {
                                type: ts_estree_1.AST_NODE_TYPES.TemplateElement,
                                value: {
                                    raw: this.ast.text.slice(node.getStart(this.ast) + 1, node.end - 1),
                                    cooked: node.text,
                                },
                                tail: true,
                            }),
                        ],
                        expressions: [],
                    });
                case SyntaxKind.TemplateExpression: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TemplateLiteral,
                        quasis: [this.convertChild(node.head)],
                        expressions: [],
                    });
                    node.templateSpans.forEach(templateSpan => {
                        result.expressions.push(this.convertChild(templateSpan.expression));
                        result.quasis.push(this.convertChild(templateSpan.literal));
                    });
                    return result;
                }
                case SyntaxKind.TaggedTemplateExpression:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TaggedTemplateExpression,
                        typeParameters: node.typeArguments
                            ? this.convertTypeArgumentsToTypeParameters(node.typeArguments, node)
                            : undefined,
                        tag: this.convertChild(node.tag),
                        quasi: this.convertChild(node.template),
                    });
                case SyntaxKind.TemplateHead:
                case SyntaxKind.TemplateMiddle:
                case SyntaxKind.TemplateTail: {
                    const tail = node.kind === SyntaxKind.TemplateTail;
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TemplateElement,
                        value: {
                            raw: this.ast.text.slice(node.getStart(this.ast) + 1, node.end - (tail ? 1 : 2)),
                            cooked: node.text,
                        },
                        tail,
                    });
                }
                // Patterns
                case SyntaxKind.SpreadAssignment:
                case SyntaxKind.SpreadElement: {
                    if (this.allowPattern) {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.RestElement,
                            argument: this.convertPattern(node.expression),
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.SpreadElement,
                            argument: this.convertChild(node.expression),
                        });
                    }
                }
                case SyntaxKind.Parameter: {
                    let parameter;
                    let result;
                    if (node.dotDotDotToken) {
                        parameter = result = this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.RestElement,
                            argument: this.convertChild(node.name),
                        });
                    }
                    else if (node.initializer) {
                        parameter = this.convertChild(node.name);
                        result = this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.AssignmentPattern,
                            left: parameter,
                            right: this.convertChild(node.initializer),
                        });
                        if (node.modifiers) {
                            // AssignmentPattern should not contain modifiers in range
                            result.range[0] = parameter.range[0];
                            result.loc = node_utils_1.getLocFor(result.range[0], result.range[1], this.ast);
                        }
                    }
                    else {
                        parameter = result = this.convertChild(node.name, parent);
                    }
                    if (node.type) {
                        parameter.typeAnnotation = this.convertTypeAnnotation(node.type, node);
                        this.fixParentLocation(parameter, parameter.typeAnnotation.range);
                    }
                    if (node.questionToken) {
                        if (node.questionToken.end > parameter.range[1]) {
                            parameter.range[1] = node.questionToken.end;
                            parameter.loc.end = node_utils_1.getLineAndCharacterFor(parameter.range[1], this.ast);
                        }
                        parameter.optional = true;
                    }
                    if (node.modifiers) {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.TSParameterProperty,
                            accessibility: (_c = node_utils_1.getTSNodeAccessibility(node)) !== null && _c !== void 0 ? _c : undefined,
                            readonly: node_utils_1.hasModifier(SyntaxKind.ReadonlyKeyword, node) || undefined,
                            static: node_utils_1.hasModifier(SyntaxKind.StaticKeyword, node) || undefined,
                            export: node_utils_1.hasModifier(SyntaxKind.ExportKeyword, node) || undefined,
                            override: node_utils_1.hasModifier(SyntaxKind.OverrideKeyword, node) || undefined,
                            parameter: result,
                        });
                    }
                    return result;
                }
                // Classes
                case SyntaxKind.ClassDeclaration:
                case SyntaxKind.ClassExpression: {
                    const heritageClauses = (_d = node.heritageClauses) !== null && _d !== void 0 ? _d : [];
                    const classNodeType = node.kind === SyntaxKind.ClassDeclaration
                        ? ts_estree_1.AST_NODE_TYPES.ClassDeclaration
                        : ts_estree_1.AST_NODE_TYPES.ClassExpression;
                    const superClass = heritageClauses.find(clause => clause.token === SyntaxKind.ExtendsKeyword);
                    const implementsClause = heritageClauses.find(clause => clause.token === SyntaxKind.ImplementsKeyword);
                    const result = this.createNode(node, {
                        type: classNodeType,
                        id: this.convertChild(node.name),
                        body: this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.ClassBody,
                            body: [],
                            range: [node.members.pos - 1, node.end],
                        }),
                        superClass: (superClass === null || superClass === void 0 ? void 0 : superClass.types[0])
                            ? this.convertChild(superClass.types[0].expression)
                            : null,
                    });
                    if (superClass) {
                        if (superClass.types.length > 1) {
                            throw node_utils_1.createError(this.ast, superClass.types[1].pos, 'Classes can only extend a single class.');
                        }
                        if ((_e = superClass.types[0]) === null || _e === void 0 ? void 0 : _e.typeArguments) {
                            result.superTypeParameters =
                                this.convertTypeArgumentsToTypeParameters(superClass.types[0].typeArguments, superClass.types[0]);
                        }
                    }
                    if (node.typeParameters) {
                        result.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                    }
                    if (implementsClause) {
                        result.implements = implementsClause.types.map(el => this.convertChild(el));
                    }
                    /**
                     * TypeScript class declarations can be defined as "abstract"
                     */
                    if (node_utils_1.hasModifier(SyntaxKind.AbstractKeyword, node)) {
                        result.abstract = true;
                    }
                    if (node_utils_1.hasModifier(SyntaxKind.DeclareKeyword, node)) {
                        result.declare = true;
                    }
                    if (node.decorators) {
                        result.decorators = node.decorators.map(el => this.convertChild(el));
                    }
                    const filteredMembers = node.members.filter(node_utils_1.isESTreeClassMember);
                    if (filteredMembers.length) {
                        result.body.body = filteredMembers.map(el => this.convertChild(el));
                    }
                    // check for exports
                    return this.fixExports(node, result);
                }
                // Modules
                case SyntaxKind.ModuleBlock:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSModuleBlock,
                        body: this.convertBodyExpressions(node.statements, node),
                    });
                case SyntaxKind.ImportDeclaration: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ImportDeclaration,
                        source: this.convertChild(node.moduleSpecifier),
                        specifiers: [],
                        importKind: 'value',
                    });
                    if (node.importClause) {
                        if (node.importClause.isTypeOnly) {
                            result.importKind = 'type';
                        }
                        if (node.importClause.name) {
                            result.specifiers.push(this.convertChild(node.importClause));
                        }
                        if (node.importClause.namedBindings) {
                            switch (node.importClause.namedBindings.kind) {
                                case SyntaxKind.NamespaceImport:
                                    result.specifiers.push(this.convertChild(node.importClause.namedBindings));
                                    break;
                                case SyntaxKind.NamedImports:
                                    result.specifiers = result.specifiers.concat(node.importClause.namedBindings.elements.map(el => this.convertChild(el)));
                                    break;
                            }
                        }
                    }
                    return result;
                }
                case SyntaxKind.NamespaceImport:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ImportNamespaceSpecifier,
                        local: this.convertChild(node.name),
                    });
                case SyntaxKind.ImportSpecifier:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ImportSpecifier,
                        local: this.convertChild(node.name),
                        imported: this.convertChild((_f = node.propertyName) !== null && _f !== void 0 ? _f : node.name),
                    });
                case SyntaxKind.ImportClause: {
                    const local = this.convertChild(node.name);
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ImportDefaultSpecifier,
                        local,
                        range: local.range,
                    });
                }
                case SyntaxKind.ExportDeclaration:
                    if (((_g = node.exportClause) === null || _g === void 0 ? void 0 : _g.kind) === SyntaxKind.NamedExports) {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.ExportNamedDeclaration,
                            source: this.convertChild(node.moduleSpecifier),
                            specifiers: node.exportClause.elements.map(el => this.convertChild(el)),
                            exportKind: node.isTypeOnly ? 'type' : 'value',
                            declaration: null,
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.ExportAllDeclaration,
                            source: this.convertChild(node.moduleSpecifier),
                            exportKind: node.isTypeOnly ? 'type' : 'value',
                            exported: 
                            // note - for compat with 3.7.x, where node.exportClause is always undefined and
                            //        SyntaxKind.NamespaceExport does not exist yet (i.e. is undefined), this
                            //        cannot be shortened to an optional chain, or else you end up with
                            //        undefined === undefined, and the true path will hard error at runtime
                            node.exportClause &&
                                node.exportClause.kind === SyntaxKind.NamespaceExport
                                ? this.convertChild(node.exportClause.name)
                                : null,
                        });
                    }
                case SyntaxKind.ExportSpecifier:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ExportSpecifier,
                        local: this.convertChild((_h = node.propertyName) !== null && _h !== void 0 ? _h : node.name),
                        exported: this.convertChild(node.name),
                    });
                case SyntaxKind.ExportAssignment:
                    if (node.isExportEquals) {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.TSExportAssignment,
                            expression: this.convertChild(node.expression),
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.ExportDefaultDeclaration,
                            declaration: this.convertChild(node.expression),
                            exportKind: 'value',
                        });
                    }
                // Unary Operations
                case SyntaxKind.PrefixUnaryExpression:
                case SyntaxKind.PostfixUnaryExpression: {
                    const operator = node_utils_1.getTextForTokenKind(node.operator);
                    /**
                     * ESTree uses UpdateExpression for ++/--
                     */
                    if (operator === '++' || operator === '--') {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.UpdateExpression,
                            operator,
                            prefix: node.kind === SyntaxKind.PrefixUnaryExpression,
                            argument: this.convertChild(node.operand),
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.UnaryExpression,
                            operator,
                            prefix: node.kind === SyntaxKind.PrefixUnaryExpression,
                            argument: this.convertChild(node.operand),
                        });
                    }
                }
                case SyntaxKind.DeleteExpression:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.UnaryExpression,
                        operator: 'delete',
                        prefix: true,
                        argument: this.convertChild(node.expression),
                    });
                case SyntaxKind.VoidExpression:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.UnaryExpression,
                        operator: 'void',
                        prefix: true,
                        argument: this.convertChild(node.expression),
                    });
                case SyntaxKind.TypeOfExpression:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.UnaryExpression,
                        operator: 'typeof',
                        prefix: true,
                        argument: this.convertChild(node.expression),
                    });
                case SyntaxKind.TypeOperator:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTypeOperator,
                        operator: node_utils_1.getTextForTokenKind(node.operator),
                        typeAnnotation: this.convertChild(node.type),
                    });
                // Binary Operations
                case SyntaxKind.BinaryExpression: {
                    // TypeScript uses BinaryExpression for sequences as well
                    if (node_utils_1.isComma(node.operatorToken)) {
                        const result = this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.SequenceExpression,
                            expressions: [],
                        });
                        const left = this.convertChild(node.left);
                        if (left.type === ts_estree_1.AST_NODE_TYPES.SequenceExpression &&
                            node.left.kind !== SyntaxKind.ParenthesizedExpression) {
                            result.expressions = result.expressions.concat(left.expressions);
                        }
                        else {
                            result.expressions.push(left);
                        }
                        result.expressions.push(this.convertChild(node.right));
                        return result;
                    }
                    else {
                        const type = node_utils_1.getBinaryExpressionType(node.operatorToken);
                        if (this.allowPattern &&
                            type === ts_estree_1.AST_NODE_TYPES.AssignmentExpression) {
                            return this.createNode(node, {
                                type: ts_estree_1.AST_NODE_TYPES.AssignmentPattern,
                                left: this.convertPattern(node.left, node),
                                right: this.convertChild(node.right),
                            });
                        }
                        return this.createNode(node, {
                            type,
                            operator: node_utils_1.getTextForTokenKind(node.operatorToken.kind),
                            left: this.converter(node.left, node, this.inTypeMode, type === ts_estree_1.AST_NODE_TYPES.AssignmentExpression),
                            right: this.convertChild(node.right),
                        });
                    }
                }
                case SyntaxKind.PropertyAccessExpression: {
                    const object = this.convertChild(node.expression);
                    const property = this.convertChild(node.name);
                    const computed = false;
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.MemberExpression,
                        object,
                        property,
                        computed,
                        optional: node.questionDotToken !== undefined,
                    });
                    return this.convertChainExpression(result, node);
                }
                case SyntaxKind.ElementAccessExpression: {
                    const object = this.convertChild(node.expression);
                    const property = this.convertChild(node.argumentExpression);
                    const computed = true;
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.MemberExpression,
                        object,
                        property,
                        computed,
                        optional: node.questionDotToken !== undefined,
                    });
                    return this.convertChainExpression(result, node);
                }
                case SyntaxKind.CallExpression: {
                    if (node.expression.kind === SyntaxKind.ImportKeyword) {
                        if (node.arguments.length !== 1) {
                            throw node_utils_1.createError(this.ast, node.arguments.pos, 'Dynamic import must have one specifier as an argument.');
                        }
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.ImportExpression,
                            source: this.convertChild(node.arguments[0]),
                        });
                    }
                    const callee = this.convertChild(node.expression);
                    const args = node.arguments.map(el => this.convertChild(el));
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.CallExpression,
                        callee,
                        arguments: args,
                        optional: node.questionDotToken !== undefined,
                    });
                    if (node.typeArguments) {
                        result.typeParameters = this.convertTypeArgumentsToTypeParameters(node.typeArguments, node);
                    }
                    return this.convertChainExpression(result, node);
                }
                case SyntaxKind.NewExpression: {
                    // NOTE - NewExpression cannot have an optional chain in it
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.NewExpression,
                        callee: this.convertChild(node.expression),
                        arguments: node.arguments
                            ? node.arguments.map(el => this.convertChild(el))
                            : [],
                    });
                    if (node.typeArguments) {
                        result.typeParameters = this.convertTypeArgumentsToTypeParameters(node.typeArguments, node);
                    }
                    return result;
                }
                case SyntaxKind.ConditionalExpression:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.ConditionalExpression,
                        test: this.convertChild(node.condition),
                        consequent: this.convertChild(node.whenTrue),
                        alternate: this.convertChild(node.whenFalse),
                    });
                case SyntaxKind.MetaProperty: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.MetaProperty,
                        meta: this.createNode(
                        // TODO: do we really want to convert it to Token?
                        node.getFirstToken(), {
                            type: ts_estree_1.AST_NODE_TYPES.Identifier,
                            name: node_utils_1.getTextForTokenKind(node.keywordToken),
                        }),
                        property: this.convertChild(node.name),
                    });
                }
                case SyntaxKind.Decorator: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Decorator,
                        expression: this.convertChild(node.expression),
                    });
                }
                // Literals
                case SyntaxKind.StringLiteral: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Literal,
                        value: parent.kind === SyntaxKind.JsxAttribute
                            ? node_utils_1.unescapeStringLiteralText(node.text)
                            : node.text,
                        raw: node.getText(),
                    });
                }
                case SyntaxKind.NumericLiteral: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Literal,
                        value: Number(node.text),
                        raw: node.getText(),
                    });
                }
                case SyntaxKind.BigIntLiteral: {
                    const range = node_utils_1.getRange(node, this.ast);
                    const rawValue = this.ast.text.slice(range[0], range[1]);
                    const bigint = rawValue
                        // remove suffix `n`
                        .slice(0, -1)
                        // `BigInt` doesn't accept numeric separator
                        // and `bigint` property should not include numeric separator
                        .replace(/_/g, '');
                    const value = typeof BigInt !== 'undefined' ? BigInt(bigint) : null;
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Literal,
                        raw: rawValue,
                        value: value,
                        bigint: value === null ? bigint : String(value),
                        range,
                    });
                }
                case SyntaxKind.RegularExpressionLiteral: {
                    const pattern = node.text.slice(1, node.text.lastIndexOf('/'));
                    const flags = node.text.slice(node.text.lastIndexOf('/') + 1);
                    let regex = null;
                    try {
                        regex = new RegExp(pattern, flags);
                    }
                    catch (exception) {
                        regex = null;
                    }
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Literal,
                        value: regex,
                        raw: node.text,
                        regex: {
                            pattern,
                            flags,
                        },
                    });
                }
                case SyntaxKind.TrueKeyword:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Literal,
                        value: true,
                        raw: 'true',
                    });
                case SyntaxKind.FalseKeyword:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Literal,
                        value: false,
                        raw: 'false',
                    });
                case SyntaxKind.NullKeyword: {
                    if (!version_check_1.typescriptVersionIsAtLeast['4.0'] && this.inTypeMode) {
                        // 4.0 started nesting null types inside a LiteralType node, but we still need to support pre-4.0
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.TSNullKeyword,
                        });
                    }
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.Literal,
                        value: null,
                        raw: 'null',
                    });
                }
                case SyntaxKind.EmptyStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.EmptyStatement,
                    });
                case SyntaxKind.DebuggerStatement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.DebuggerStatement,
                    });
                // JSX
                case SyntaxKind.JsxElement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXElement,
                        openingElement: this.convertChild(node.openingElement),
                        closingElement: this.convertChild(node.closingElement),
                        children: node.children.map(el => this.convertChild(el)),
                    });
                case SyntaxKind.JsxFragment:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXFragment,
                        openingFragment: this.convertChild(node.openingFragment),
                        closingFragment: this.convertChild(node.closingFragment),
                        children: node.children.map(el => this.convertChild(el)),
                    });
                case SyntaxKind.JsxSelfClosingElement: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXElement,
                        /**
                         * Convert SyntaxKind.JsxSelfClosingElement to SyntaxKind.JsxOpeningElement,
                         * TypeScript does not seem to have the idea of openingElement when tag is self-closing
                         */
                        openingElement: this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.JSXOpeningElement,
                            typeParameters: node.typeArguments
                                ? this.convertTypeArgumentsToTypeParameters(node.typeArguments, node)
                                : undefined,
                            selfClosing: true,
                            name: this.convertJSXTagName(node.tagName, node),
                            attributes: node.attributes.properties.map(el => this.convertChild(el)),
                            range: node_utils_1.getRange(node, this.ast),
                        }),
                        closingElement: null,
                        children: [],
                    });
                }
                case SyntaxKind.JsxOpeningElement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXOpeningElement,
                        typeParameters: node.typeArguments
                            ? this.convertTypeArgumentsToTypeParameters(node.typeArguments, node)
                            : undefined,
                        selfClosing: false,
                        name: this.convertJSXTagName(node.tagName, node),
                        attributes: node.attributes.properties.map(el => this.convertChild(el)),
                    });
                case SyntaxKind.JsxClosingElement:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXClosingElement,
                        name: this.convertJSXTagName(node.tagName, node),
                    });
                case SyntaxKind.JsxOpeningFragment:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXOpeningFragment,
                    });
                case SyntaxKind.JsxClosingFragment:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXClosingFragment,
                    });
                case SyntaxKind.JsxExpression: {
                    const expression = node.expression
                        ? this.convertChild(node.expression)
                        : this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.JSXEmptyExpression,
                            range: [node.getStart(this.ast) + 1, node.getEnd() - 1],
                        });
                    if (node.dotDotDotToken) {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.JSXSpreadChild,
                            expression,
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.JSXExpressionContainer,
                            expression,
                        });
                    }
                }
                case SyntaxKind.JsxAttribute: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXAttribute,
                        name: this.convertJSXNamespaceOrIdentifier(node.name),
                        value: this.convertChild(node.initializer),
                    });
                }
                /**
                 * The JSX AST changed the node type for string literals
                 * inside a JSX Element from `Literal` to `JSXText`. We
                 * provide a flag to support both types until `Literal`
                 * node type is deprecated in ESLint v5.
                 */
                case SyntaxKind.JsxText: {
                    const start = node.getFullStart();
                    const end = node.getEnd();
                    const text = this.ast.text.slice(start, end);
                    if (this.options.useJSXTextNode) {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.JSXText,
                            value: node_utils_1.unescapeStringLiteralText(text),
                            raw: text,
                            range: [start, end],
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.Literal,
                            value: node_utils_1.unescapeStringLiteralText(text),
                            raw: text,
                            range: [start, end],
                        });
                    }
                }
                case SyntaxKind.JsxSpreadAttribute:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.JSXSpreadAttribute,
                        argument: this.convertChild(node.expression),
                    });
                case SyntaxKind.QualifiedName: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSQualifiedName,
                        left: this.convertChild(node.left),
                        right: this.convertChild(node.right),
                    });
                }
                // TypeScript specific
                case SyntaxKind.TypeReference: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTypeReference,
                        typeName: this.convertType(node.typeName),
                        typeParameters: node.typeArguments
                            ? this.convertTypeArgumentsToTypeParameters(node.typeArguments, node)
                            : undefined,
                    });
                }
                case SyntaxKind.TypeParameter: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTypeParameter,
                        name: this.convertType(node.name),
                        constraint: node.constraint
                            ? this.convertType(node.constraint)
                            : undefined,
                        default: node.default ? this.convertType(node.default) : undefined,
                    });
                }
                case SyntaxKind.ThisType:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSThisType,
                    });
                case SyntaxKind.AnyKeyword:
                case SyntaxKind.BigIntKeyword:
                case SyntaxKind.BooleanKeyword:
                case SyntaxKind.NeverKeyword:
                case SyntaxKind.NumberKeyword:
                case SyntaxKind.ObjectKeyword:
                case SyntaxKind.StringKeyword:
                case SyntaxKind.SymbolKeyword:
                case SyntaxKind.UnknownKeyword:
                case SyntaxKind.VoidKeyword:
                case SyntaxKind.UndefinedKeyword:
                case SyntaxKind.IntrinsicKeyword: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES[`TS${SyntaxKind[node.kind]}`],
                    });
                }
                case SyntaxKind.NonNullExpression: {
                    const nnExpr = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSNonNullExpression,
                        expression: this.convertChild(node.expression),
                    });
                    return this.convertChainExpression(nnExpr, node);
                }
                case SyntaxKind.TypeLiteral: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTypeLiteral,
                        members: node.members.map(el => this.convertChild(el)),
                    });
                }
                case SyntaxKind.ArrayType: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSArrayType,
                        elementType: this.convertType(node.elementType),
                    });
                }
                case SyntaxKind.IndexedAccessType: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSIndexedAccessType,
                        objectType: this.convertType(node.objectType),
                        indexType: this.convertType(node.indexType),
                    });
                }
                case SyntaxKind.ConditionalType: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSConditionalType,
                        checkType: this.convertType(node.checkType),
                        extendsType: this.convertType(node.extendsType),
                        trueType: this.convertType(node.trueType),
                        falseType: this.convertType(node.falseType),
                    });
                }
                case SyntaxKind.TypeQuery: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTypeQuery,
                        exprName: this.convertType(node.exprName),
                    });
                }
                case SyntaxKind.MappedType: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSMappedType,
                        typeParameter: this.convertType(node.typeParameter),
                        nameType: (_j = this.convertType(node.nameType)) !== null && _j !== void 0 ? _j : null,
                    });
                    if (node.readonlyToken) {
                        if (node.readonlyToken.kind === SyntaxKind.ReadonlyKeyword) {
                            result.readonly = true;
                        }
                        else {
                            result.readonly = node_utils_1.getTextForTokenKind(node.readonlyToken.kind);
                        }
                    }
                    if (node.questionToken) {
                        if (node.questionToken.kind === SyntaxKind.QuestionToken) {
                            result.optional = true;
                        }
                        else {
                            result.optional = node_utils_1.getTextForTokenKind(node.questionToken.kind);
                        }
                    }
                    if (node.type) {
                        result.typeAnnotation = this.convertType(node.type);
                    }
                    return result;
                }
                case SyntaxKind.ParenthesizedExpression:
                    return this.convertChild(node.expression, parent);
                case SyntaxKind.TypeAliasDeclaration: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTypeAliasDeclaration,
                        id: this.convertChild(node.name),
                        typeAnnotation: this.convertType(node.type),
                    });
                    if (node_utils_1.hasModifier(SyntaxKind.DeclareKeyword, node)) {
                        result.declare = true;
                    }
                    // Process typeParameters
                    if (node.typeParameters) {
                        result.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                    }
                    // check for exports
                    return this.fixExports(node, result);
                }
                case SyntaxKind.MethodSignature: {
                    return this.convertMethodSignature(node);
                }
                case SyntaxKind.PropertySignature: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSPropertySignature,
                        optional: node_utils_1.isOptional(node) || undefined,
                        computed: node_utils_1.isComputedProperty(node.name),
                        key: this.convertChild(node.name),
                        typeAnnotation: node.type
                            ? this.convertTypeAnnotation(node.type, node)
                            : undefined,
                        initializer: this.convertChild(node.initializer) || undefined,
                        readonly: node_utils_1.hasModifier(SyntaxKind.ReadonlyKeyword, node) || undefined,
                        static: node_utils_1.hasModifier(SyntaxKind.StaticKeyword, node) || undefined,
                        export: node_utils_1.hasModifier(SyntaxKind.ExportKeyword, node) || undefined,
                    });
                    const accessibility = node_utils_1.getTSNodeAccessibility(node);
                    if (accessibility) {
                        result.accessibility = accessibility;
                    }
                    return result;
                }
                case SyntaxKind.IndexSignature: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSIndexSignature,
                        parameters: node.parameters.map(el => this.convertChild(el)),
                    });
                    if (node.type) {
                        result.typeAnnotation = this.convertTypeAnnotation(node.type, node);
                    }
                    if (node_utils_1.hasModifier(SyntaxKind.ReadonlyKeyword, node)) {
                        result.readonly = true;
                    }
                    const accessibility = node_utils_1.getTSNodeAccessibility(node);
                    if (accessibility) {
                        result.accessibility = accessibility;
                    }
                    if (node_utils_1.hasModifier(SyntaxKind.ExportKeyword, node)) {
                        result.export = true;
                    }
                    if (node_utils_1.hasModifier(SyntaxKind.StaticKeyword, node)) {
                        result.static = true;
                    }
                    return result;
                }
                case SyntaxKind.ConstructorType: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSConstructorType,
                        params: this.convertParameters(node.parameters),
                        abstract: node_utils_1.hasModifier(SyntaxKind.AbstractKeyword, node),
                    });
                    if (node.type) {
                        result.returnType = this.convertTypeAnnotation(node.type, node);
                    }
                    if (node.typeParameters) {
                        result.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                    }
                    return result;
                }
                case SyntaxKind.FunctionType:
                case SyntaxKind.ConstructSignature:
                case SyntaxKind.CallSignature: {
                    const type = node.kind === SyntaxKind.ConstructSignature
                        ? ts_estree_1.AST_NODE_TYPES.TSConstructSignatureDeclaration
                        : node.kind === SyntaxKind.CallSignature
                            ? ts_estree_1.AST_NODE_TYPES.TSCallSignatureDeclaration
                            : ts_estree_1.AST_NODE_TYPES.TSFunctionType;
                    const result = this.createNode(node, {
                        type: type,
                        params: this.convertParameters(node.parameters),
                    });
                    if (node.type) {
                        result.returnType = this.convertTypeAnnotation(node.type, node);
                    }
                    if (node.typeParameters) {
                        result.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                    }
                    return result;
                }
                case SyntaxKind.ExpressionWithTypeArguments: {
                    const result = this.createNode(node, {
                        type: parent && parent.kind === SyntaxKind.InterfaceDeclaration
                            ? ts_estree_1.AST_NODE_TYPES.TSInterfaceHeritage
                            : ts_estree_1.AST_NODE_TYPES.TSClassImplements,
                        expression: this.convertChild(node.expression),
                    });
                    if (node.typeArguments) {
                        result.typeParameters = this.convertTypeArgumentsToTypeParameters(node.typeArguments, node);
                    }
                    return result;
                }
                case SyntaxKind.InterfaceDeclaration: {
                    const interfaceHeritageClauses = (_k = node.heritageClauses) !== null && _k !== void 0 ? _k : [];
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSInterfaceDeclaration,
                        body: this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.TSInterfaceBody,
                            body: node.members.map(member => this.convertChild(member)),
                            range: [node.members.pos - 1, node.end],
                        }),
                        id: this.convertChild(node.name),
                    });
                    if (node.typeParameters) {
                        result.typeParameters =
                            this.convertTSTypeParametersToTypeParametersDeclaration(node.typeParameters);
                    }
                    if (interfaceHeritageClauses.length > 0) {
                        const interfaceExtends = [];
                        const interfaceImplements = [];
                        for (const heritageClause of interfaceHeritageClauses) {
                            if (heritageClause.token === SyntaxKind.ExtendsKeyword) {
                                for (const n of heritageClause.types) {
                                    interfaceExtends.push(this.convertChild(n, node));
                                }
                            }
                            else {
                                for (const n of heritageClause.types) {
                                    interfaceImplements.push(this.convertChild(n, node));
                                }
                            }
                        }
                        if (interfaceExtends.length) {
                            result.extends = interfaceExtends;
                        }
                        if (interfaceImplements.length) {
                            result.implements = interfaceImplements;
                        }
                    }
                    if (node_utils_1.hasModifier(SyntaxKind.AbstractKeyword, node)) {
                        result.abstract = true;
                    }
                    if (node_utils_1.hasModifier(SyntaxKind.DeclareKeyword, node)) {
                        result.declare = true;
                    }
                    // check for exports
                    return this.fixExports(node, result);
                }
                case SyntaxKind.TypePredicate: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTypePredicate,
                        asserts: node.assertsModifier !== undefined,
                        parameterName: this.convertChild(node.parameterName),
                        typeAnnotation: null,
                    });
                    /**
                     * Specific fix for type-guard location data
                     */
                    if (node.type) {
                        result.typeAnnotation = this.convertTypeAnnotation(node.type, node);
                        result.typeAnnotation.loc = result.typeAnnotation.typeAnnotation.loc;
                        result.typeAnnotation.range =
                            result.typeAnnotation.typeAnnotation.range;
                    }
                    return result;
                }
                case SyntaxKind.ImportType:
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSImportType,
                        isTypeOf: !!node.isTypeOf,
                        parameter: this.convertChild(node.argument),
                        qualifier: this.convertChild(node.qualifier),
                        typeParameters: node.typeArguments
                            ? this.convertTypeArgumentsToTypeParameters(node.typeArguments, node)
                            : null,
                    });
                case SyntaxKind.EnumDeclaration: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSEnumDeclaration,
                        id: this.convertChild(node.name),
                        members: node.members.map(el => this.convertChild(el)),
                    });
                    // apply modifiers first...
                    this.applyModifiersToResult(result, node.modifiers);
                    // ...then check for exports
                    return this.fixExports(node, result);
                }
                case SyntaxKind.EnumMember: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSEnumMember,
                        id: this.convertChild(node.name),
                    });
                    if (node.initializer) {
                        result.initializer = this.convertChild(node.initializer);
                    }
                    if (node.name.kind === ts.SyntaxKind.ComputedPropertyName) {
                        result.computed = true;
                    }
                    return result;
                }
                case SyntaxKind.ModuleDeclaration: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSModuleDeclaration,
                        id: this.convertChild(node.name),
                    });
                    if (node.body) {
                        result.body = this.convertChild(node.body);
                    }
                    // apply modifiers first...
                    this.applyModifiersToResult(result, node.modifiers);
                    if (node.flags & ts.NodeFlags.GlobalAugmentation) {
                        result.global = true;
                    }
                    // ...then check for exports
                    return this.fixExports(node, result);
                }
                // TypeScript specific types
                case SyntaxKind.ParenthesizedType: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSParenthesizedType,
                        typeAnnotation: this.convertType(node.type),
                    });
                }
                case SyntaxKind.UnionType: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSUnionType,
                        types: node.types.map(el => this.convertType(el)),
                    });
                }
                case SyntaxKind.IntersectionType: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSIntersectionType,
                        types: node.types.map(el => this.convertType(el)),
                    });
                }
                case SyntaxKind.AsExpression: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSAsExpression,
                        expression: this.convertChild(node.expression),
                        typeAnnotation: this.convertType(node.type),
                    });
                }
                case SyntaxKind.InferType: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSInferType,
                        typeParameter: this.convertType(node.typeParameter),
                    });
                }
                case SyntaxKind.LiteralType: {
                    if (version_check_1.typescriptVersionIsAtLeast['4.0'] &&
                        node.literal.kind === SyntaxKind.NullKeyword) {
                        // 4.0 started nesting null types inside a LiteralType node
                        // but our AST is designed around the old way of null being a keyword
                        return this.createNode(node.literal, {
                            type: ts_estree_1.AST_NODE_TYPES.TSNullKeyword,
                        });
                    }
                    else {
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.TSLiteralType,
                            literal: this.convertType(node.literal),
                        });
                    }
                }
                case SyntaxKind.TypeAssertionExpression: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTypeAssertion,
                        typeAnnotation: this.convertType(node.type),
                        expression: this.convertChild(node.expression),
                    });
                }
                case SyntaxKind.ImportEqualsDeclaration: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSImportEqualsDeclaration,
                        id: this.convertChild(node.name),
                        moduleReference: this.convertChild(node.moduleReference),
                        importKind: node.isTypeOnly ? 'type' : 'value',
                        isExport: node_utils_1.hasModifier(SyntaxKind.ExportKeyword, node),
                    });
                }
                case SyntaxKind.ExternalModuleReference: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSExternalModuleReference,
                        expression: this.convertChild(node.expression),
                    });
                }
                case SyntaxKind.NamespaceExportDeclaration: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSNamespaceExportDeclaration,
                        id: this.convertChild(node.name),
                    });
                }
                case SyntaxKind.AbstractKeyword: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSAbstractKeyword,
                    });
                }
                // Tuple
                case SyntaxKind.TupleType: {
                    // In TS 4.0, the `elementTypes` property was changed to `elements`.
                    // To support both at compile time, we cast to access the newer version
                    // if the former does not exist.
                    const elementTypes = 'elementTypes' in node
                        ? node.elementTypes.map((el) => this.convertType(el))
                        : node.elements.map(el => this.convertType(el));
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTupleType,
                        elementTypes,
                    });
                }
                case SyntaxKind.NamedTupleMember: {
                    const member = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSNamedTupleMember,
                        elementType: this.convertType(node.type, node),
                        label: this.convertChild(node.name, node),
                        optional: node.questionToken != null,
                    });
                    if (node.dotDotDotToken) {
                        // adjust the start to account for the "..."
                        member.range[0] = member.label.range[0];
                        member.loc.start = member.label.loc.start;
                        return this.createNode(node, {
                            type: ts_estree_1.AST_NODE_TYPES.TSRestType,
                            typeAnnotation: member,
                        });
                    }
                    return member;
                }
                case SyntaxKind.OptionalType: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSOptionalType,
                        typeAnnotation: this.convertType(node.type),
                    });
                }
                case SyntaxKind.RestType: {
                    return this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSRestType,
                        typeAnnotation: this.convertType(node.type),
                    });
                }
                // Template Literal Types
                case SyntaxKind.TemplateLiteralType: {
                    const result = this.createNode(node, {
                        type: ts_estree_1.AST_NODE_TYPES.TSTemplateLiteralType,
                        quasis: [this.convertChild(node.head)],
                        types: [],
                    });
                    node.templateSpans.forEach(templateSpan => {
                        result.types.push(this.convertChild(templateSpan.type));
                        result.quasis.push(this.convertChild(templateSpan.literal));
                    });
                    return result;
                }
                default:
                    return this.deeplyCopy(node);
            }
        }
    }
    exports.Converter = Converter;
},
/* @typescript-eslint/typescript-estree/dist/node-utils.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const ts_estree_1 = require(1200) /* ./ts-estree */;
    const xhtml_entities_1 = require(1203) /* ./jsx/xhtml-entities */;
    const SyntaxKind = ts.SyntaxKind;
    const LOGICAL_OPERATORS = [
        SyntaxKind.BarBarToken,
        SyntaxKind.AmpersandAmpersandToken,
        SyntaxKind.QuestionQuestionToken,
    ];
    /**
     * Returns true if the given ts.Token is the assignment operator
     * @param operator the operator token
     * @returns is assignment
     */
    function isAssignmentOperator(operator) {
        return (operator.kind >= SyntaxKind.FirstAssignment &&
            operator.kind <= SyntaxKind.LastAssignment);
    }
    exports.isAssignmentOperator = isAssignmentOperator;
    /**
     * Returns true if the given ts.Token is a logical operator
     * @param operator the operator token
     * @returns is a logical operator
     */
    function isLogicalOperator(operator) {
        return LOGICAL_OPERATORS.includes(operator.kind);
    }
    exports.isLogicalOperator = isLogicalOperator;
    /**
     * Returns the string form of the given TSToken SyntaxKind
     * @param kind the token's SyntaxKind
     * @returns the token applicable token as a string
     */
    function getTextForTokenKind(kind) {
        return ts.tokenToString(kind);
    }
    exports.getTextForTokenKind = getTextForTokenKind;
    /**
     * Returns true if the given ts.Node is a valid ESTree class member
     * @param node TypeScript AST node
     * @returns is valid ESTree class member
     */
    function isESTreeClassMember(node) {
        return node.kind !== SyntaxKind.SemicolonClassElement;
    }
    exports.isESTreeClassMember = isESTreeClassMember;
    /**
     * Checks if a ts.Node has a modifier
     * @param modifierKind TypeScript SyntaxKind modifier
     * @param node TypeScript AST node
     * @returns has the modifier specified
     */
    function hasModifier(modifierKind, node) {
        return (!!node.modifiers &&
            !!node.modifiers.length &&
            node.modifiers.some(modifier => modifier.kind === modifierKind));
    }
    exports.hasModifier = hasModifier;
    /**
     * Get last last modifier in ast
     * @param node TypeScript AST node
     * @returns returns last modifier if present or null
     */
    function getLastModifier(node) {
        return ((!!node.modifiers &&
            !!node.modifiers.length &&
            node.modifiers[node.modifiers.length - 1]) ||
            null);
    }
    exports.getLastModifier = getLastModifier;
    /**
     * Returns true if the given ts.Token is a comma
     * @param token the TypeScript token
     * @returns is comma
     */
    function isComma(token) {
        return token.kind === SyntaxKind.CommaToken;
    }
    exports.isComma = isComma;
    /**
     * Returns true if the given ts.Node is a comment
     * @param node the TypeScript node
     * @returns is comment
     */
    function isComment(node) {
        return (node.kind === SyntaxKind.SingleLineCommentTrivia ||
            node.kind === SyntaxKind.MultiLineCommentTrivia);
    }
    exports.isComment = isComment;
    /**
     * Returns true if the given ts.Node is a JSDoc comment
     * @param node the TypeScript node
     * @returns is JSDoc comment
     */
    function isJSDocComment(node) {
        return node.kind === SyntaxKind.JSDocComment;
    }
    exports.isJSDocComment = isJSDocComment;
    /**
     * Returns the binary expression type of the given ts.Token
     * @param operator the operator token
     * @returns the binary expression type
     */
    function getBinaryExpressionType(operator) {
        if (isAssignmentOperator(operator)) {
            return ts_estree_1.AST_NODE_TYPES.AssignmentExpression;
        }
        else if (isLogicalOperator(operator)) {
            return ts_estree_1.AST_NODE_TYPES.LogicalExpression;
        }
        return ts_estree_1.AST_NODE_TYPES.BinaryExpression;
    }
    exports.getBinaryExpressionType = getBinaryExpressionType;
    /**
     * Returns line and column data for the given positions,
     * @param pos position to check
     * @param ast the AST object
     * @returns line and column
     */
    function getLineAndCharacterFor(pos, ast) {
        const loc = ast.getLineAndCharacterOfPosition(pos);
        return {
            line: loc.line + 1,
            column: loc.character,
        };
    }
    exports.getLineAndCharacterFor = getLineAndCharacterFor;
    /**
     * Returns line and column data for the given start and end positions,
     * for the given AST
     * @param start start data
     * @param end   end data
     * @param ast   the AST object
     * @returns the loc data
     */
    function getLocFor(start, end, ast) {
        return {
            start: getLineAndCharacterFor(start, ast),
            end: getLineAndCharacterFor(end, ast),
        };
    }
    exports.getLocFor = getLocFor;
    /**
     * Check whatever node can contain directive
     * @param node
     * @returns returns true if node can contain directive
     */
    function canContainDirective(node) {
        if (node.kind === ts.SyntaxKind.Block) {
            switch (node.parent.kind) {
                case ts.SyntaxKind.Constructor:
                case ts.SyntaxKind.GetAccessor:
                case ts.SyntaxKind.SetAccessor:
                case ts.SyntaxKind.ArrowFunction:
                case ts.SyntaxKind.FunctionExpression:
                case ts.SyntaxKind.FunctionDeclaration:
                case ts.SyntaxKind.MethodDeclaration:
                    return true;
                default:
                    return false;
            }
        }
        return true;
    }
    exports.canContainDirective = canContainDirective;
    /**
     * Returns range for the given ts.Node
     * @param node the ts.Node or ts.Token
     * @param ast the AST object
     * @returns the range data
     */
    function getRange(node, ast) {
        return [node.getStart(ast), node.getEnd()];
    }
    exports.getRange = getRange;
    /**
     * Returns true if a given ts.Node is a token
     * @param node the ts.Node
     * @returns is a token
     */
    function isToken(node) {
        return (node.kind >= SyntaxKind.FirstToken && node.kind <= SyntaxKind.LastToken);
    }
    exports.isToken = isToken;
    /**
     * Returns true if a given ts.Node is a JSX token
     * @param node ts.Node to be checked
     * @returns is a JSX token
     */
    function isJSXToken(node) {
        return (node.kind >= SyntaxKind.JsxElement && node.kind <= SyntaxKind.JsxAttribute);
    }
    exports.isJSXToken = isJSXToken;
    /**
     * Returns the declaration kind of the given ts.Node
     * @param node TypeScript AST node
     * @returns declaration kind
     */
    function getDeclarationKind(node) {
        if (node.flags & ts.NodeFlags.Let) {
            return 'let';
        }
        if (node.flags & ts.NodeFlags.Const) {
            return 'const';
        }
        return 'var';
    }
    exports.getDeclarationKind = getDeclarationKind;
    /**
     * Gets a ts.Node's accessibility level
     * @param node The ts.Node
     * @returns accessibility "public", "protected", "private", or null
     */
    function getTSNodeAccessibility(node) {
        const modifiers = node.modifiers;
        if (!modifiers) {
            return null;
        }
        for (let i = 0; i < modifiers.length; i++) {
            const modifier = modifiers[i];
            switch (modifier.kind) {
                case SyntaxKind.PublicKeyword:
                    return 'public';
                case SyntaxKind.ProtectedKeyword:
                    return 'protected';
                case SyntaxKind.PrivateKeyword:
                    return 'private';
                default:
                    break;
            }
        }
        return null;
    }
    exports.getTSNodeAccessibility = getTSNodeAccessibility;
    /**
     * Finds the next token based on the previous one and its parent
     * Had to copy this from TS instead of using TS's version because theirs doesn't pass the ast to getChildren
     * @param previousToken The previous TSToken
     * @param parent The parent TSNode
     * @param ast The TS AST
     * @returns the next TSToken
     */
    function findNextToken(previousToken, parent, ast) {
        return find(parent);
        function find(n) {
            if (ts.isToken(n) && n.pos === previousToken.end) {
                // this is token that starts at the end of previous token - return it
                return n;
            }
            return firstDefined(n.getChildren(ast), (child) => {
                const shouldDiveInChildNode = 
                // previous token is enclosed somewhere in the child
                (child.pos <= previousToken.pos && child.end > previousToken.end) ||
                    // previous token ends exactly at the beginning of child
                    child.pos === previousToken.end;
                return shouldDiveInChildNode && nodeHasTokens(child, ast)
                    ? find(child)
                    : undefined;
            });
        }
    }
    exports.findNextToken = findNextToken;
    /**
     * Find the first matching ancestor based on the given predicate function.
     * @param node The current ts.Node
     * @param predicate The predicate function to apply to each checked ancestor
     * @returns a matching parent ts.Node
     */
    function findFirstMatchingAncestor(node, predicate) {
        while (node) {
            if (predicate(node)) {
                return node;
            }
            node = node.parent;
        }
        return undefined;
    }
    exports.findFirstMatchingAncestor = findFirstMatchingAncestor;
    /**
     * Returns true if a given ts.Node has a JSX token within its hierarchy
     * @param node ts.Node to be checked
     * @returns has JSX ancestor
     */
    function hasJSXAncestor(node) {
        return !!findFirstMatchingAncestor(node, isJSXToken);
    }
    exports.hasJSXAncestor = hasJSXAncestor;
    /**
     * Unescape the text content of string literals, e.g. &amp; -> &
     * @param text The escaped string literal text.
     * @returns The unescaped string literal text.
     */
    function unescapeStringLiteralText(text) {
        return text.replace(/&(?:#\d+|#x[\da-fA-F]+|[0-9a-zA-Z]+);/g, entity => {
            const item = entity.slice(1, -1);
            if (item[0] === '#') {
                const codePoint = item[1] === 'x'
                    ? parseInt(item.slice(2), 16)
                    : parseInt(item.slice(1), 10);
                return codePoint > 0x10ffff // RangeError: Invalid code point
                    ? entity
                    : String.fromCodePoint(codePoint);
            }
            return xhtml_entities_1.xhtmlEntities[item] || entity;
        });
    }
    exports.unescapeStringLiteralText = unescapeStringLiteralText;
    /**
     * Returns true if a given ts.Node is a computed property
     * @param node ts.Node to be checked
     * @returns is Computed Property
     */
    function isComputedProperty(node) {
        return node.kind === SyntaxKind.ComputedPropertyName;
    }
    exports.isComputedProperty = isComputedProperty;
    /**
     * Returns true if a given ts.Node is optional (has QuestionToken)
     * @param node ts.Node to be checked
     * @returns is Optional
     */
    function isOptional(node) {
        return node.questionToken
            ? node.questionToken.kind === SyntaxKind.QuestionToken
            : false;
    }
    exports.isOptional = isOptional;
    /**
     * Returns true if the node is an optional chain node
     */
    function isChainExpression(node) {
        return node.type === ts_estree_1.AST_NODE_TYPES.ChainExpression;
    }
    exports.isChainExpression = isChainExpression;
    /**
     * Returns true of the child of property access expression is an optional chain
     */
    function isChildUnwrappableOptionalChain(node, child) {
        return (isChainExpression(child) &&
            // (x?.y).z is semantically different, and as such .z is no longer optional
            node.expression.kind !== ts.SyntaxKind.ParenthesizedExpression);
    }
    exports.isChildUnwrappableOptionalChain = isChildUnwrappableOptionalChain;
    /**
     * Returns the type of a given ts.Token
     * @param token the ts.Token
     * @returns the token type
     */
    function getTokenType(token) {
        if ('originalKeywordKind' in token && token.originalKeywordKind) {
            if (token.originalKeywordKind === SyntaxKind.NullKeyword) {
                return ts_estree_1.AST_TOKEN_TYPES.Null;
            }
            else if (token.originalKeywordKind >= SyntaxKind.FirstFutureReservedWord &&
                token.originalKeywordKind <= SyntaxKind.LastKeyword) {
                return ts_estree_1.AST_TOKEN_TYPES.Identifier;
            }
            return ts_estree_1.AST_TOKEN_TYPES.Keyword;
        }
        if (token.kind >= SyntaxKind.FirstKeyword &&
            token.kind <= SyntaxKind.LastFutureReservedWord) {
            if (token.kind === SyntaxKind.FalseKeyword ||
                token.kind === SyntaxKind.TrueKeyword) {
                return ts_estree_1.AST_TOKEN_TYPES.Boolean;
            }
            return ts_estree_1.AST_TOKEN_TYPES.Keyword;
        }
        if (token.kind >= SyntaxKind.FirstPunctuation &&
            token.kind <= SyntaxKind.LastPunctuation) {
            return ts_estree_1.AST_TOKEN_TYPES.Punctuator;
        }
        if (token.kind >= SyntaxKind.NoSubstitutionTemplateLiteral &&
            token.kind <= SyntaxKind.TemplateTail) {
            return ts_estree_1.AST_TOKEN_TYPES.Template;
        }
        switch (token.kind) {
            case SyntaxKind.NumericLiteral:
                return ts_estree_1.AST_TOKEN_TYPES.Numeric;
            case SyntaxKind.JsxText:
                return ts_estree_1.AST_TOKEN_TYPES.JSXText;
            case SyntaxKind.StringLiteral:
                // A TypeScript-StringLiteral token with a TypeScript-JsxAttribute or TypeScript-JsxElement parent,
                // must actually be an ESTree-JSXText token
                if (token.parent &&
                    (token.parent.kind === SyntaxKind.JsxAttribute ||
                        token.parent.kind === SyntaxKind.JsxElement)) {
                    return ts_estree_1.AST_TOKEN_TYPES.JSXText;
                }
                return ts_estree_1.AST_TOKEN_TYPES.String;
            case SyntaxKind.RegularExpressionLiteral:
                return ts_estree_1.AST_TOKEN_TYPES.RegularExpression;
            case SyntaxKind.Identifier:
            case SyntaxKind.ConstructorKeyword:
            case SyntaxKind.GetKeyword:
            case SyntaxKind.SetKeyword:
            // intentional fallthrough
            default:
        }
        // Some JSX tokens have to be determined based on their parent
        if (token.parent && token.kind === SyntaxKind.Identifier) {
            if (isJSXToken(token.parent)) {
                return ts_estree_1.AST_TOKEN_TYPES.JSXIdentifier;
            }
            if (token.parent.kind === SyntaxKind.PropertyAccessExpression &&
                hasJSXAncestor(token)) {
                return ts_estree_1.AST_TOKEN_TYPES.JSXIdentifier;
            }
        }
        return ts_estree_1.AST_TOKEN_TYPES.Identifier;
    }
    exports.getTokenType = getTokenType;
    /**
     * Extends and formats a given ts.Token, for a given AST
     * @param token the ts.Token
     * @param ast   the AST object
     * @returns the converted Token
     */
    function convertToken(token, ast) {
        const start = token.kind === SyntaxKind.JsxText
            ? token.getFullStart()
            : token.getStart(ast);
        const end = token.getEnd();
        const value = ast.text.slice(start, end);
        const tokenType = getTokenType(token);
        if (tokenType === ts_estree_1.AST_TOKEN_TYPES.RegularExpression) {
            return {
                type: tokenType,
                value,
                range: [start, end],
                loc: getLocFor(start, end, ast),
                regex: {
                    pattern: value.slice(1, value.lastIndexOf('/')),
                    flags: value.slice(value.lastIndexOf('/') + 1),
                },
            };
        }
        else {
            // @ts-expect-error TS is complaining about `value` not being the correct
            // type but it is
            return {
                type: tokenType,
                value,
                range: [start, end],
                loc: getLocFor(start, end, ast),
            };
        }
    }
    exports.convertToken = convertToken;
    /**
     * Converts all tokens for the given AST
     * @param ast the AST object
     * @returns the converted Tokens
     */
    function convertTokens(ast) {
        const result = [];
        /**
         * @param node the ts.Node
         */
        function walk(node) {
            // TypeScript generates tokens for types in JSDoc blocks. Comment tokens
            // and their children should not be walked or added to the resulting tokens list.
            if (isComment(node) || isJSDocComment(node)) {
                return;
            }
            if (isToken(node) && node.kind !== SyntaxKind.EndOfFileToken) {
                const converted = convertToken(node, ast);
                if (converted) {
                    result.push(converted);
                }
            }
            else {
                node.getChildren(ast).forEach(walk);
            }
        }
        walk(ast);
        return result;
    }
    exports.convertTokens = convertTokens;
    class TSError extends Error {
        constructor(message, fileName, index, lineNumber, column) {
            super(message);
            this.fileName = fileName;
            this.index = index;
            this.lineNumber = lineNumber;
            this.column = column;
            Object.defineProperty(this, 'name', {
                value: new.target.name,
                enumerable: false,
                configurable: true,
            });
        }
    }
    exports.TSError = TSError;
    /**
     * @param ast     the AST object
     * @param start   the index at which the error starts
     * @param message the error message
     * @returns converted error object
     */
    function createError(ast, start, message) {
        const loc = ast.getLineAndCharacterOfPosition(start);
        return new TSError(message, ast.fileName, start, loc.line + 1, loc.character);
    }
    exports.createError = createError;
    /**
     * @param n the TSNode
     * @param ast the TS AST
     */
    function nodeHasTokens(n, ast) {
        // If we have a token or node that has a non-zero width, it must have tokens.
        // Note: getWidth() does not take trivia into account.
        return n.kind === SyntaxKind.EndOfFileToken
            ? !!n.jsDoc
            : n.getWidth(ast) !== 0;
    }
    exports.nodeHasTokens = nodeHasTokens;
    /**
     * Like `forEach`, but suitable for use with numbers and strings (which may be falsy).
     * @template T
     * @template U
     * @param array
     * @param callback
     */
    function firstDefined(array, callback) {
        if (array === undefined) {
            return undefined;
        }
        for (let i = 0; i < array.length; i++) {
            const result = callback(array[i], i);
            if (result !== undefined) {
                return result;
            }
        }
        return undefined;
    }
    exports.firstDefined = firstDefined;
},
/* @typescript-eslint/typescript-estree/dist/ts-estree/index.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __exportStar = (this && this.__exportStar) || function (m, exports) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p))
                __createBinding(exports, m, p);
    };
    __esModule();
    // for simplicity and backwards-compatibility
    var types_1 = require(740) /* @typescript-eslint/types */;
    __esExport("AST_NODE_TYPES", types_1.AST_NODE_TYPES);
    __esExport("AST_TOKEN_TYPES", types_1.AST_TOKEN_TYPES);
    __esExport("TSESTree", types_1.TSESTree);
    __exportStar(require(1201) /* ./ts-nodes */, exports);
    __exportStar(require(1202) /* ./estree-to-ts-node-types */, exports);
},
/* @typescript-eslint/typescript-estree/dist/ts-estree/ts-nodes.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
},
/* @typescript-eslint/typescript-estree/dist/ts-estree/estree-to-ts-node-types.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const types_1 = require(740) /* @typescript-eslint/types */;
},
/* @typescript-eslint/typescript-estree/dist/jsx/xhtml-entities.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    exports.xhtmlEntities = {
        quot: '\u0022',
        amp: '&',
        apos: '\u0027',
        lt: '<',
        gt: '>',
        nbsp: '\u00A0',
        iexcl: '\u00A1',
        cent: '\u00A2',
        pound: '\u00A3',
        curren: '\u00A4',
        yen: '\u00A5',
        brvbar: '\u00A6',
        sect: '\u00A7',
        uml: '\u00A8',
        copy: '\u00A9',
        ordf: '\u00AA',
        laquo: '\u00AB',
        not: '\u00AC',
        shy: '\u00AD',
        reg: '\u00AE',
        macr: '\u00AF',
        deg: '\u00B0',
        plusmn: '\u00B1',
        sup2: '\u00B2',
        sup3: '\u00B3',
        acute: '\u00B4',
        micro: '\u00B5',
        para: '\u00B6',
        middot: '\u00B7',
        cedil: '\u00B8',
        sup1: '\u00B9',
        ordm: '\u00BA',
        raquo: '\u00BB',
        frac14: '\u00BC',
        frac12: '\u00BD',
        frac34: '\u00BE',
        iquest: '\u00BF',
        Agrave: '\u00C0',
        Aacute: '\u00C1',
        Acirc: '\u00C2',
        Atilde: '\u00C3',
        Auml: '\u00C4',
        Aring: '\u00C5',
        AElig: '\u00C6',
        Ccedil: '\u00C7',
        Egrave: '\u00C8',
        Eacute: '\u00C9',
        Ecirc: '\u00CA',
        Euml: '\u00CB',
        Igrave: '\u00CC',
        Iacute: '\u00CD',
        Icirc: '\u00CE',
        Iuml: '\u00CF',
        ETH: '\u00D0',
        Ntilde: '\u00D1',
        Ograve: '\u00D2',
        Oacute: '\u00D3',
        Ocirc: '\u00D4',
        Otilde: '\u00D5',
        Ouml: '\u00D6',
        times: '\u00D7',
        Oslash: '\u00D8',
        Ugrave: '\u00D9',
        Uacute: '\u00DA',
        Ucirc: '\u00DB',
        Uuml: '\u00DC',
        Yacute: '\u00DD',
        THORN: '\u00DE',
        szlig: '\u00DF',
        agrave: '\u00E0',
        aacute: '\u00E1',
        acirc: '\u00E2',
        atilde: '\u00E3',
        auml: '\u00E4',
        aring: '\u00E5',
        aelig: '\u00E6',
        ccedil: '\u00E7',
        egrave: '\u00E8',
        eacute: '\u00E9',
        ecirc: '\u00EA',
        euml: '\u00EB',
        igrave: '\u00EC',
        iacute: '\u00ED',
        icirc: '\u00EE',
        iuml: '\u00EF',
        eth: '\u00F0',
        ntilde: '\u00F1',
        ograve: '\u00F2',
        oacute: '\u00F3',
        ocirc: '\u00F4',
        otilde: '\u00F5',
        ouml: '\u00F6',
        divide: '\u00F7',
        oslash: '\u00F8',
        ugrave: '\u00F9',
        uacute: '\u00FA',
        ucirc: '\u00FB',
        uuml: '\u00FC',
        yacute: '\u00FD',
        thorn: '\u00FE',
        yuml: '\u00FF',
        OElig: '\u0152',
        oelig: '\u0153',
        Scaron: '\u0160',
        scaron: '\u0161',
        Yuml: '\u0178',
        fnof: '\u0192',
        circ: '\u02C6',
        tilde: '\u02DC',
        Alpha: '\u0391',
        Beta: '\u0392',
        Gamma: '\u0393',
        Delta: '\u0394',
        Epsilon: '\u0395',
        Zeta: '\u0396',
        Eta: '\u0397',
        Theta: '\u0398',
        Iota: '\u0399',
        Kappa: '\u039A',
        Lambda: '\u039B',
        Mu: '\u039C',
        Nu: '\u039D',
        Xi: '\u039E',
        Omicron: '\u039F',
        Pi: '\u03A0',
        Rho: '\u03A1',
        Sigma: '\u03A3',
        Tau: '\u03A4',
        Upsilon: '\u03A5',
        Phi: '\u03A6',
        Chi: '\u03A7',
        Psi: '\u03A8',
        Omega: '\u03A9',
        alpha: '\u03B1',
        beta: '\u03B2',
        gamma: '\u03B3',
        delta: '\u03B4',
        epsilon: '\u03B5',
        zeta: '\u03B6',
        eta: '\u03B7',
        theta: '\u03B8',
        iota: '\u03B9',
        kappa: '\u03BA',
        lambda: '\u03BB',
        mu: '\u03BC',
        nu: '\u03BD',
        xi: '\u03BE',
        omicron: '\u03BF',
        pi: '\u03C0',
        rho: '\u03C1',
        sigmaf: '\u03C2',
        sigma: '\u03C3',
        tau: '\u03C4',
        upsilon: '\u03C5',
        phi: '\u03C6',
        chi: '\u03C7',
        psi: '\u03C8',
        omega: '\u03C9',
        thetasym: '\u03D1',
        upsih: '\u03D2',
        piv: '\u03D6',
        ensp: '\u2002',
        emsp: '\u2003',
        thinsp: '\u2009',
        zwnj: '\u200C',
        zwj: '\u200D',
        lrm: '\u200E',
        rlm: '\u200F',
        ndash: '\u2013',
        mdash: '\u2014',
        lsquo: '\u2018',
        rsquo: '\u2019',
        sbquo: '\u201A',
        ldquo: '\u201C',
        rdquo: '\u201D',
        bdquo: '\u201E',
        dagger: '\u2020',
        Dagger: '\u2021',
        bull: '\u2022',
        hellip: '\u2026',
        permil: '\u2030',
        prime: '\u2032',
        Prime: '\u2033',
        lsaquo: '\u2039',
        rsaquo: '\u203A',
        oline: '\u203E',
        frasl: '\u2044',
        euro: '\u20AC',
        image: '\u2111',
        weierp: '\u2118',
        real: '\u211C',
        trade: '\u2122',
        alefsym: '\u2135',
        larr: '\u2190',
        uarr: '\u2191',
        rarr: '\u2192',
        darr: '\u2193',
        harr: '\u2194',
        crarr: '\u21B5',
        lArr: '\u21D0',
        uArr: '\u21D1',
        rArr: '\u21D2',
        dArr: '\u21D3',
        hArr: '\u21D4',
        forall: '\u2200',
        part: '\u2202',
        exist: '\u2203',
        empty: '\u2205',
        nabla: '\u2207',
        isin: '\u2208',
        notin: '\u2209',
        ni: '\u220B',
        prod: '\u220F',
        sum: '\u2211',
        minus: '\u2212',
        lowast: '\u2217',
        radic: '\u221A',
        prop: '\u221D',
        infin: '\u221E',
        ang: '\u2220',
        and: '\u2227',
        or: '\u2228',
        cap: '\u2229',
        cup: '\u222A',
        int: '\u222B',
        there4: '\u2234',
        sim: '\u223C',
        cong: '\u2245',
        asymp: '\u2248',
        ne: '\u2260',
        equiv: '\u2261',
        le: '\u2264',
        ge: '\u2265',
        sub: '\u2282',
        sup: '\u2283',
        nsub: '\u2284',
        sube: '\u2286',
        supe: '\u2287',
        oplus: '\u2295',
        otimes: '\u2297',
        perp: '\u22A5',
        sdot: '\u22C5',
        lceil: '\u2308',
        rceil: '\u2309',
        lfloor: '\u230A',
        rfloor: '\u230B',
        lang: '\u2329',
        rang: '\u232A',
        loz: '\u25CA',
        spades: '\u2660',
        clubs: '\u2663',
        hearts: '\u2665',
        diams: '\u2666',
    };
},
/* @typescript-eslint/typescript-estree/dist/version-check.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const semver = __importStar(require(1004) /* semver */);
    const ts = __importStar(require(21) /* typescript */);
    function semverCheck(version) {
        return semver.satisfies(ts.version, `>= ${version}.0 || >= ${version}.1-rc || >= ${version}.0-beta`, {
            includePrerelease: true,
        });
    }
    const versions = [
        //
        '3.7',
        '3.8',
        '3.9',
        '4.0',
    ];
    const typescriptVersionIsAtLeast = {};
    exports.typescriptVersionIsAtLeast = typescriptVersionIsAtLeast;
    for (const version of versions) {
        typescriptVersionIsAtLeast[version] = semverCheck(version);
    }
},
/* @typescript-eslint/typescript-estree/dist/convert-comments.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    const util_1 = require(929) /* tsutils/util/util */;
    const node_utils_1 = require(1199) /* ./node-utils */;
    const ts_estree_1 = require(1200) /* ./ts-estree */;
    /**
     * Convert all comments for the given AST.
     * @param ast the AST object
     * @param code the TypeScript code
     * @returns the converted ESTreeComment
     * @private
     */
    function convertComments(ast, code) {
        const comments = [];
        util_1.forEachComment(ast, (_, comment) => {
            const type = comment.kind == ts.SyntaxKind.SingleLineCommentTrivia
                ? ts_estree_1.AST_TOKEN_TYPES.Line
                : ts_estree_1.AST_TOKEN_TYPES.Block;
            const range = [comment.pos, comment.end];
            const loc = node_utils_1.getLocFor(range[0], range[1], ast);
            // both comments start with 2 characters - /* or //
            const textStart = range[0] + 2;
            const textEnd = comment.kind === ts.SyntaxKind.SingleLineCommentTrivia
                ? // single line comments end at the end
                    range[1] - textStart
                : // multiline comments end 2 characters early
                    range[1] - textStart - 2;
            comments.push({
                type,
                value: code.substr(textStart, textEnd),
                range,
                loc,
            });
        }, ast);
        return comments;
    }
    exports.convertComments = convertComments;
},
/* @typescript-eslint/typescript-estree/dist/simple-traverse.js */ function _(require, module, exports, __esModule, __esExport) {
    __esModule();
    const visitor_keys_1 = require(770) /* @typescript-eslint/visitor-keys */;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    function isValidNode(x) {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
        return x !== null && typeof x === 'object' && typeof x.type === 'string';
    }
    function getVisitorKeysForNode(allVisitorKeys, node) {
        const keys = allVisitorKeys[node.type];
        return (keys !== null && keys !== void 0 ? keys : []);
    }
    class SimpleTraverser {
        constructor(selectors, setParentPointers = false) {
            this.allVisitorKeys = visitor_keys_1.visitorKeys;
            this.selectors = selectors;
            this.setParentPointers = setParentPointers;
        }
        traverse(node, parent) {
            if (!isValidNode(node)) {
                return;
            }
            if (this.setParentPointers) {
                node.parent = parent;
            }
            if ('enter' in this.selectors) {
                this.selectors.enter(node, parent);
            }
            else if (node.type in this.selectors) {
                this.selectors[node.type](node, parent);
            }
            const keys = getVisitorKeysForNode(this.allVisitorKeys, node);
            if (keys.length < 1) {
                return;
            }
            for (const key of keys) {
                const childOrChildren = node[key];
                if (Array.isArray(childOrChildren)) {
                    for (const child of childOrChildren) {
                        this.traverse(child, node);
                    }
                }
                else {
                    this.traverse(childOrChildren, node);
                }
            }
        }
    }
    function simpleTraverse(startingNode, options, setParentPointers = false) {
        new SimpleTraverser(options, setParentPointers).traverse(startingNode, undefined);
    }
    exports.simpleTraverse = simpleTraverse;
},
/* @typescript-eslint/typescript-estree/dist/create-program/createDefaultProgram.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const debug_1 = __importDefault(require(72) /* debug */);
    const path_1 = __importDefault(require("path"));
    const ts = __importStar(require(21) /* typescript */);
    const shared_1 = require(1208) /* ./shared */;
    const log = debug_1.default('typescript-eslint:typescript-estree:createDefaultProgram');
    /**
     * @param code The code of the file being linted
     * @param extra The config object
     * @param extra.tsconfigRootDir The root directory for relative tsconfig paths
     * @param extra.projects Provided tsconfig paths
     * @returns If found, returns the source file corresponding to the code and the containing program
     */
    function createDefaultProgram(code, extra) {
        log('Getting default program for: %s', extra.filePath || 'unnamed file');
        if (!extra.projects || extra.projects.length !== 1) {
            return undefined;
        }
        const tsconfigPath = extra.projects[0];
        const commandLine = ts.getParsedCommandLineOfConfigFile(tsconfigPath, shared_1.createDefaultCompilerOptionsFromExtra(extra), Object.assign(Object.assign({}, ts.sys), { onUnRecoverableConfigFileDiagnostic: () => { } }));
        if (!commandLine) {
            return undefined;
        }
        const compilerHost = ts.createCompilerHost(commandLine.options, 
        /* setParentNodes */ true);
        if (extra.moduleResolver) {
            compilerHost.resolveModuleNames = shared_1.getModuleResolver(extra.moduleResolver).resolveModuleNames;
        }
        const oldReadFile = compilerHost.readFile;
        compilerHost.readFile = (fileName) => path_1.default.normalize(fileName) === path_1.default.normalize(extra.filePath)
            ? code
            : oldReadFile(fileName);
        const program = ts.createProgram([extra.filePath], commandLine.options, compilerHost);
        const ast = program.getSourceFile(extra.filePath);
        return ast && { ast, program };
    }
    exports.createDefaultProgram = createDefaultProgram;
},
/* @typescript-eslint/typescript-estree/dist/create-program/shared.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const path_1 = __importDefault(require("path"));
    const ts = __importStar(require(21) /* typescript */);
    /**
     * Compiler options required to avoid critical functionality issues
     */
    const CORE_COMPILER_OPTIONS = {
        noEmit: true,
        /**
         * Flags required to make no-unused-vars work
         */
        noUnusedLocals: true,
        noUnusedParameters: true,
    };
    exports.CORE_COMPILER_OPTIONS = CORE_COMPILER_OPTIONS;
    /**
     * Default compiler options for program generation
     */
    const DEFAULT_COMPILER_OPTIONS = Object.assign(Object.assign({}, CORE_COMPILER_OPTIONS), { allowNonTsExtensions: true, allowJs: true, checkJs: true });
    function createDefaultCompilerOptionsFromExtra(extra) {
        if (extra.debugLevel.has('typescript')) {
            return Object.assign(Object.assign({}, DEFAULT_COMPILER_OPTIONS), { extendedDiagnostics: true });
        }
        return DEFAULT_COMPILER_OPTIONS;
    }
    exports.createDefaultCompilerOptionsFromExtra = createDefaultCompilerOptionsFromExtra;
    // typescript doesn't provide a ts.sys implementation for browser environments
    const useCaseSensitiveFileNames = ts.sys !== undefined ? ts.sys.useCaseSensitiveFileNames : true;
    const correctPathCasing = useCaseSensitiveFileNames
        ? (filePath) => filePath
        : (filePath) => filePath.toLowerCase();
    function getCanonicalFileName(filePath) {
        let normalized = path_1.default.normalize(filePath);
        if (normalized.endsWith(path_1.default.sep)) {
            normalized = normalized.substr(0, normalized.length - 1);
        }
        return correctPathCasing(normalized);
    }
    exports.getCanonicalFileName = getCanonicalFileName;
    function ensureAbsolutePath(p, extra) {
        return path_1.default.isAbsolute(p)
            ? p
            : path_1.default.join(extra.tsconfigRootDir || process.cwd(), p);
    }
    exports.ensureAbsolutePath = ensureAbsolutePath;
    function canonicalDirname(p) {
        return path_1.default.dirname(p);
    }
    exports.canonicalDirname = canonicalDirname;
    function getScriptKind(extra, filePath = extra.filePath) {
        const extension = path_1.default.extname(filePath).toLowerCase();
        // note - we respect the user's extension when it is known  we could override it and force it to match their
        // jsx setting, but that could create weird situations where we throw parse errors when TSC doesn't
        switch (extension) {
            case '.ts':
                return ts.ScriptKind.TS;
            case '.tsx':
                return ts.ScriptKind.TSX;
            case '.js':
                return ts.ScriptKind.JS;
            case '.jsx':
                return ts.ScriptKind.JSX;
            case '.json':
                return ts.ScriptKind.JSON;
            default:
                // unknown extension, force typescript to ignore the file extension, and respect the user's setting
                return extra.jsx ? ts.ScriptKind.TSX : ts.ScriptKind.TS;
        }
    }
    exports.getScriptKind = getScriptKind;
    function getExtension(fileName) {
        if (!fileName) {
            return null;
        }
        return fileName.endsWith('.d.ts') ? '.d.ts' : path_1.default.extname(fileName);
    }
    function getAstFromProgram(currentProgram, extra) {
        const ast = currentProgram.getSourceFile(extra.filePath);
        // working around https://github.com/typescript-eslint/typescript-eslint/issues/1573
        const expectedExt = getExtension(extra.filePath);
        const returnedExt = getExtension(ast === null || ast === void 0 ? void 0 : ast.fileName);
        if (expectedExt !== returnedExt) {
            return undefined;
        }
        return ast && { ast, program: currentProgram };
    }
    exports.getAstFromProgram = getAstFromProgram;
    function getModuleResolver(moduleResolverPath) {
        let moduleResolver;
        try {
            moduleResolver = require(moduleResolverPath);
        }
        catch (error) {
            const errorLines = [
                'Could not find the provided parserOptions.moduleResolver.',
                'Hint: use an absolute path if you are not in control over where the ESLint instance runs.',
            ];
            throw new Error(errorLines.join('\n'));
        }
        return moduleResolver;
    }
    exports.getModuleResolver = getModuleResolver;
},
/* @typescript-eslint/typescript-estree/dist/create-program/createIsolatedProgram.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const debug_1 = __importDefault(require(72) /* debug */);
    const ts = __importStar(require(21) /* typescript */);
    const shared_1 = require(1208) /* ./shared */;
    const log = debug_1.default('typescript-eslint:typescript-estree:createIsolatedProgram');
    /**
     * @param code The code of the file being linted
     * @returns Returns a new source file and program corresponding to the linted code
     */
    function createIsolatedProgram(code, extra) {
        log('Getting isolated program in %s mode for: %s', extra.jsx ? 'TSX' : 'TS', extra.filePath);
        const compilerHost = {
            fileExists() {
                return true;
            },
            getCanonicalFileName() {
                return extra.filePath;
            },
            getCurrentDirectory() {
                return '';
            },
            getDirectories() {
                return [];
            },
            getDefaultLibFileName() {
                return 'lib.d.ts';
            },
            // TODO: Support Windows CRLF
            getNewLine() {
                return '\n';
            },
            getSourceFile(filename) {
                return ts.createSourceFile(filename, code, ts.ScriptTarget.Latest, 
                /* setParentNodes */ true, shared_1.getScriptKind(extra, filename));
            },
            readFile() {
                return undefined;
            },
            useCaseSensitiveFileNames() {
                return true;
            },
            writeFile() {
                return null;
            },
        };
        const program = ts.createProgram([extra.filePath], Object.assign({ noResolve: true, target: ts.ScriptTarget.Latest, jsx: extra.jsx ? ts.JsxEmit.Preserve : undefined }, shared_1.createDefaultCompilerOptionsFromExtra(extra)), compilerHost);
        const ast = program.getSourceFile(extra.filePath);
        if (!ast) {
            throw new Error('Expected an ast to be returned for the single-file isolated program.');
        }
        return { ast, program };
    }
    exports.createIsolatedProgram = createIsolatedProgram;
},
/* @typescript-eslint/typescript-estree/dist/create-program/createProjectProgram.js */ function _(require, module, exports, __esModule, __esExport) {
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const debug_1 = __importDefault(require(72) /* debug */);
    const path_1 = __importDefault(require("path"));
    const createWatchProgram_1 = require(1211) /* ./createWatchProgram */;
    const node_utils_1 = require(1199) /* ../node-utils */;
    const shared_1 = require(1208) /* ./shared */;
    const log = debug_1.default('typescript-eslint:typescript-estree:createProjectProgram');
    const DEFAULT_EXTRA_FILE_EXTENSIONS = ['.ts', '.tsx', '.js', '.jsx'];
    /**
     * @param code The code of the file being linted
     * @param createDefaultProgram True if the default program should be created
     * @param extra The config object
     * @returns If found, returns the source file corresponding to the code and the containing program
     */
    function createProjectProgram(code, createDefaultProgram, extra) {
        log('Creating project program for: %s', extra.filePath);
        const astAndProgram = node_utils_1.firstDefined(createWatchProgram_1.getProgramsForProjects(code, extra.filePath, extra), currentProgram => shared_1.getAstFromProgram(currentProgram, extra));
        if (!astAndProgram && !createDefaultProgram) {
            // the file was either not matched within the tsconfig, or the extension wasn't expected
            const errorLines = [
                '"parserOptions.project" has been set for @typescript-eslint/parser.',
                `The file does not match your project config: ${path_1.default.relative(extra.tsconfigRootDir || process.cwd(), extra.filePath)}.`,
            ];
            let hasMatchedAnError = false;
            const extraFileExtensions = extra.extraFileExtensions || [];
            extraFileExtensions.forEach(extraExtension => {
                if (!extraExtension.startsWith('.')) {
                    errorLines.push(`Found unexpected extension "${extraExtension}" specified with the "extraFileExtensions" option. Did you mean ".${extraExtension}"?`);
                }
                if (DEFAULT_EXTRA_FILE_EXTENSIONS.includes(extraExtension)) {
                    errorLines.push(`You unnecessarily included the extension "${extraExtension}" with the "extraFileExtensions" option. This extension is already handled by the parser by default.`);
                }
            });
            const fileExtension = path_1.default.extname(extra.filePath);
            if (!DEFAULT_EXTRA_FILE_EXTENSIONS.includes(fileExtension)) {
                const nonStandardExt = `The extension for the file (${fileExtension}) is non-standard`;
                if (extraFileExtensions.length > 0) {
                    if (!extraFileExtensions.includes(fileExtension)) {
                        errorLines.push(`${nonStandardExt}. It should be added to your existing "parserOptions.extraFileExtensions".`);
                        hasMatchedAnError = true;
                    }
                }
                else {
                    errorLines.push(`${nonStandardExt}. You should add "parserOptions.extraFileExtensions" to your config.`);
                    hasMatchedAnError = true;
                }
            }
            if (!hasMatchedAnError) {
                errorLines.push('The file must be included in at least one of the projects provided.');
            }
            throw new Error(errorLines.join('\n'));
        }
        return astAndProgram;
    }
    exports.createProjectProgram = createProjectProgram;
},
/* @typescript-eslint/typescript-estree/dist/create-program/createWatchProgram.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const debug_1 = __importDefault(require(72) /* debug */);
    const fs_1 = __importDefault(require("fs"));
    const semver_1 = __importDefault(require(1004) /* semver */);
    const ts = __importStar(require(21) /* typescript */);
    const shared_1 = require(1208) /* ./shared */;
    const log = debug_1.default('typescript-eslint:typescript-estree:createWatchProgram');
    /**
     * Maps tsconfig paths to their corresponding file contents and resulting watches
     */
    const knownWatchProgramMap = new Map();
    /**
     * Maps file/folder paths to their set of corresponding watch callbacks
     * There may be more than one per file/folder if a file/folder is shared between projects
     */
    const fileWatchCallbackTrackingMap = new Map();
    const folderWatchCallbackTrackingMap = new Map();
    /**
     * Stores the list of known files for each program
     */
    const programFileListCache = new Map();
    /**
     * Caches the last modified time of the tsconfig files
     */
    const tsconfigLastModifiedTimestampCache = new Map();
    const parsedFilesSeenHash = new Map();
    /**
     * Clear all of the parser caches.
     * This should only be used in testing to ensure the parser is clean between tests.
     */
    function clearWatchCaches() {
        knownWatchProgramMap.clear();
        fileWatchCallbackTrackingMap.clear();
        folderWatchCallbackTrackingMap.clear();
        parsedFilesSeenHash.clear();
        programFileListCache.clear();
        tsconfigLastModifiedTimestampCache.clear();
    }
    exports.clearWatchCaches = clearWatchCaches;
    function saveWatchCallback(trackingMap) {
        return (fileName, callback) => {
            const normalizedFileName = shared_1.getCanonicalFileName(fileName);
            const watchers = (() => {
                let watchers = trackingMap.get(normalizedFileName);
                if (!watchers) {
                    watchers = new Set();
                    trackingMap.set(normalizedFileName, watchers);
                }
                return watchers;
            })();
            watchers.add(callback);
            return {
                close: () => {
                    watchers.delete(callback);
                },
            };
        };
    }
    /**
     * Holds information about the file currently being linted
     */
    const currentLintOperationState = {
        code: '',
        filePath: '',
    };
    /**
     * Appropriately report issues found when reading a config file
     * @param diagnostic The diagnostic raised when creating a program
     */
    function diagnosticReporter(diagnostic) {
        throw new Error(ts.flattenDiagnosticMessageText(diagnostic.messageText, ts.sys.newLine));
    }
    /**
     * Hash content for compare content.
     * @param content hashed contend
     * @returns hashed result
     */
    function createHash(content) {
        var _a;
        // No ts.sys in browser environments.
        if ((_a = ts.sys) === null || _a === void 0 ? void 0 : _a.createHash) {
            return ts.sys.createHash(content);
        }
        return content;
    }
    function updateCachedFileList(tsconfigPath, program, extra) {
        const fileList = extra.EXPERIMENTAL_useSourceOfProjectReferenceRedirect
            ? new Set(program.getSourceFiles().map(sf => shared_1.getCanonicalFileName(sf.fileName)))
            : new Set(program.getRootFileNames().map(f => shared_1.getCanonicalFileName(f)));
        programFileListCache.set(tsconfigPath, fileList);
        return fileList;
    }
    /**
     * Calculate project environments using options provided by consumer and paths from config
     * @param code The code being linted
     * @param filePathIn The path of the file being parsed
     * @param extra.tsconfigRootDir The root directory for relative tsconfig paths
     * @param extra.projects Provided tsconfig paths
     * @returns The programs corresponding to the supplied tsconfig paths
     */
    function getProgramsForProjects(code, filePathIn, extra) {
        const filePath = shared_1.getCanonicalFileName(filePathIn);
        const results = [];
        // preserve reference to code and file being linted
        currentLintOperationState.code = code;
        currentLintOperationState.filePath = filePath;
        // Update file version if necessary
        const fileWatchCallbacks = fileWatchCallbackTrackingMap.get(filePath);
        const codeHash = createHash(code);
        if (parsedFilesSeenHash.get(filePath) !== codeHash &&
            fileWatchCallbacks &&
            fileWatchCallbacks.size > 0) {
            fileWatchCallbacks.forEach(cb => cb(filePath, ts.FileWatcherEventKind.Changed));
        }
        /*
         * before we go into the process of attempting to find and update every program
         * see if we know of a program that contains this file
         */
        for (const [tsconfigPath, existingWatch] of knownWatchProgramMap.entries()) {
            let fileList = programFileListCache.get(tsconfigPath);
            let updatedProgram = null;
            if (!fileList) {
                updatedProgram = existingWatch.getProgram().getProgram();
                fileList = updateCachedFileList(tsconfigPath, updatedProgram, extra);
            }
            if (fileList.has(filePath)) {
                log('Found existing program for file. %s', filePath);
                updatedProgram =
                    updatedProgram !== null && updatedProgram !== void 0 ? updatedProgram : existingWatch.getProgram().getProgram();
                // sets parent pointers in source files
                updatedProgram.getTypeChecker();
                return [updatedProgram];
            }
        }
        log('File did not belong to any existing programs, moving to create/update. %s', filePath);
        /*
         * We don't know of a program that contains the file, this means that either:
         * - the required program hasn't been created yet, or
         * - the file is new/renamed, and the program hasn't been updated.
         */
        for (const tsconfigPath of extra.projects) {
            const existingWatch = knownWatchProgramMap.get(tsconfigPath);
            if (existingWatch) {
                const updatedProgram = maybeInvalidateProgram(existingWatch, filePath, tsconfigPath);
                if (!updatedProgram) {
                    continue;
                }
                // sets parent pointers in source files
                updatedProgram.getTypeChecker();
                // cache and check the file list
                const fileList = updateCachedFileList(tsconfigPath, updatedProgram, extra);
                if (fileList.has(filePath)) {
                    log('Found updated program for file. %s', filePath);
                    // we can return early because we know this program contains the file
                    return [updatedProgram];
                }
                results.push(updatedProgram);
                continue;
            }
            const programWatch = createWatchProgram(tsconfigPath, extra);
            knownWatchProgramMap.set(tsconfigPath, programWatch);
            const program = programWatch.getProgram().getProgram();
            // sets parent pointers in source files
            program.getTypeChecker();
            // cache and check the file list
            const fileList = updateCachedFileList(tsconfigPath, program, extra);
            if (fileList.has(filePath)) {
                log('Found program for file. %s', filePath);
                // we can return early because we know this program contains the file
                return [program];
            }
            results.push(program);
        }
        return results;
    }
    exports.getProgramsForProjects = getProgramsForProjects;
    const isRunningNoTimeoutFix = semver_1.default.satisfies(ts.version, '>=3.9.0-beta', {
        includePrerelease: true,
    });
    function createWatchProgram(tsconfigPath, extra) {
        log('Creating watch program for %s.', tsconfigPath);
        // create compiler host
        const watchCompilerHost = ts.createWatchCompilerHost(tsconfigPath, shared_1.createDefaultCompilerOptionsFromExtra(extra), ts.sys, ts.createAbstractBuilder, diagnosticReporter, 
        /*reportWatchStatus*/ () => { });
        if (extra.moduleResolver) {
            watchCompilerHost.resolveModuleNames = shared_1.getModuleResolver(extra.moduleResolver).resolveModuleNames;
        }
        // ensure readFile reads the code being linted instead of the copy on disk
        const oldReadFile = watchCompilerHost.readFile;
        watchCompilerHost.readFile = (filePathIn, encoding) => {
            const filePath = shared_1.getCanonicalFileName(filePathIn);
            const fileContent = filePath === currentLintOperationState.filePath
                ? currentLintOperationState.code
                : oldReadFile(filePath, encoding);
            if (fileContent !== undefined) {
                parsedFilesSeenHash.set(filePath, createHash(fileContent));
            }
            return fileContent;
        };
        // ensure process reports error on failure instead of exiting process immediately
        watchCompilerHost.onUnRecoverableConfigFileDiagnostic = diagnosticReporter;
        // ensure process doesn't emit programs
        watchCompilerHost.afterProgramCreate = (program) => {
            // report error if there are any errors in the config file
            const configFileDiagnostics = program
                .getConfigFileParsingDiagnostics()
                .filter(diag => diag.category === ts.DiagnosticCategory.Error && diag.code !== 18003);
            if (configFileDiagnostics.length > 0) {
                diagnosticReporter(configFileDiagnostics[0]);
            }
        };
        /*
         * From the CLI, the file watchers won't matter, as the files will be parsed once and then forgotten.
         * When running from an IDE, these watchers will let us tell typescript about changes.
         *
         * ESLint IDE plugins will send us unfinished file content as the user types (before it's saved to disk).
         * We use the file watchers to tell typescript about this latest file content.
         *
         * When files are created (or renamed), we won't know about them because we have no filesystem watchers attached.
         * We use the folder watchers to tell typescript it needs to go and find new files in the project folders.
         */
        watchCompilerHost.watchFile = saveWatchCallback(fileWatchCallbackTrackingMap);
        watchCompilerHost.watchDirectory = saveWatchCallback(folderWatchCallbackTrackingMap);
        // allow files with custom extensions to be included in program (uses internal ts api)
        const oldOnDirectoryStructureHostCreate = watchCompilerHost.onCachedDirectoryStructureHostCreate;
        watchCompilerHost.onCachedDirectoryStructureHostCreate = (host) => {
            const oldReadDirectory = host.readDirectory;
            host.readDirectory = (path, extensions, exclude, include, depth) => oldReadDirectory(path, !extensions ? undefined : extensions.concat(extra.extraFileExtensions), exclude, include, depth);
            oldOnDirectoryStructureHostCreate(host);
        };
        // This works only on 3.9
        watchCompilerHost.extraFileExtensions = extra.extraFileExtensions.map(extension => ({
            extension,
            isMixedContent: true,
            scriptKind: ts.ScriptKind.Deferred,
        }));
        watchCompilerHost.trace = log;
        /**
         * TODO: this needs refinement and development, but we're allowing users to opt-in to this for now for testing and feedback.
         * See https://github.com/typescript-eslint/typescript-eslint/issues/2094
         */
        watchCompilerHost.useSourceOfProjectReferenceRedirect = () => extra.EXPERIMENTAL_useSourceOfProjectReferenceRedirect;
        // Since we don't want to asynchronously update program we want to disable timeout methods
        // So any changes in the program will be delayed and updated when getProgram is called on watch
        let callback;
        if (isRunningNoTimeoutFix) {
            watchCompilerHost.setTimeout = undefined;
            watchCompilerHost.clearTimeout = undefined;
        }
        else {
            log('Running without timeout fix');
            // But because of https://github.com/microsoft/TypeScript/pull/37308 we cannot just set it to undefined
            // instead save it and call before getProgram is called
            watchCompilerHost.setTimeout = (cb, _ms, ...args) => {
                callback = cb.bind(/*this*/ undefined, ...args);
                return callback;
            };
            watchCompilerHost.clearTimeout = () => {
                callback = undefined;
            };
        }
        const watch = ts.createWatchProgram(watchCompilerHost);
        if (!isRunningNoTimeoutFix) {
            const originalGetProgram = watch.getProgram;
            watch.getProgram = () => {
                if (callback) {
                    callback();
                }
                callback = undefined;
                return originalGetProgram.call(watch);
            };
        }
        return watch;
    }
    exports.createWatchProgram = createWatchProgram;
    function hasTSConfigChanged(tsconfigPath) {
        const stat = fs_1.default.statSync(tsconfigPath);
        const lastModifiedAt = stat.mtimeMs;
        const cachedLastModifiedAt = tsconfigLastModifiedTimestampCache.get(tsconfigPath);
        tsconfigLastModifiedTimestampCache.set(tsconfigPath, lastModifiedAt);
        if (cachedLastModifiedAt === undefined) {
            return false;
        }
        return Math.abs(cachedLastModifiedAt - lastModifiedAt) > Number.EPSILON;
    }
    function maybeInvalidateProgram(existingWatch, filePath, tsconfigPath) {
        /*
         * By calling watchProgram.getProgram(), it will trigger a resync of the program based on
         * whatever new file content we've given it from our input.
         */
        let updatedProgram = existingWatch.getProgram().getProgram();
        // In case this change causes problems in larger real world codebases
        // Provide an escape hatch so people don't _have_ to revert to an older version
        if (process.env.TSESTREE_NO_INVALIDATION === 'true') {
            return updatedProgram;
        }
        if (hasTSConfigChanged(tsconfigPath)) {
            /*
             * If the stat of the tsconfig has changed, that could mean the include/exclude/files lists has changed
             * We need to make sure typescript knows this so it can update appropriately
             */
            log('tsconfig has changed - triggering program update. %s', tsconfigPath);
            fileWatchCallbackTrackingMap
                .get(tsconfigPath)
                .forEach(cb => cb(tsconfigPath, ts.FileWatcherEventKind.Changed));
            // tsconfig change means that the file list more than likely changed, so clear the cache
            programFileListCache.delete(tsconfigPath);
        }
        let sourceFile = updatedProgram.getSourceFile(filePath);
        if (sourceFile) {
            return updatedProgram;
        }
        /*
         * Missing source file means our program's folder structure might be out of date.
         * So we need to tell typescript it needs to update the correct folder.
         */
        log('File was not found in program - triggering folder update. %s', filePath);
        // Find the correct directory callback by climbing the folder tree
        const currentDir = shared_1.canonicalDirname(filePath);
        let current = null;
        let next = currentDir;
        let hasCallback = false;
        while (current !== next) {
            current = next;
            const folderWatchCallbacks = folderWatchCallbackTrackingMap.get(current);
            if (folderWatchCallbacks) {
                folderWatchCallbacks.forEach(cb => {
                    if (currentDir !== current) {
                        cb(currentDir, ts.FileWatcherEventKind.Changed);
                    }
                    cb(current, ts.FileWatcherEventKind.Changed);
                });
                hasCallback = true;
            }
            next = shared_1.canonicalDirname(current);
        }
        if (!hasCallback) {
            /*
             * No callback means the paths don't matchup - so no point returning any program
             * this will signal to the caller to skip this program
             */
            log('No callback found for file, not part of this program. %s', filePath);
            return null;
        }
        // directory update means that the file list more than likely changed, so clear the cache
        programFileListCache.delete(tsconfigPath);
        // force the immediate resync
        updatedProgram = existingWatch.getProgram().getProgram();
        sourceFile = updatedProgram.getSourceFile(filePath);
        if (sourceFile) {
            return updatedProgram;
        }
        /*
         * At this point we're in one of two states:
         * - The file isn't supposed to be in this program due to exclusions
         * - The file is new, and was renamed from an old, included filename
         *
         * For the latter case, we need to tell typescript that the old filename is now deleted
         */
        log('File was still not found in program after directory update - checking file deletions. %s', filePath);
        const rootFilenames = updatedProgram.getRootFileNames();
        // use find because we only need to "delete" one file to cause typescript to do a full resync
        const deletedFile = rootFilenames.find(file => !fs_1.default.existsSync(file));
        if (!deletedFile) {
            // There are no deleted files, so it must be the former case of the file not belonging to this program
            return null;
        }
        const fileWatchCallbacks = fileWatchCallbackTrackingMap.get(shared_1.getCanonicalFileName(deletedFile));
        if (!fileWatchCallbacks) {
            // shouldn't happen, but just in case
            log('Could not find watch callbacks for root file. %s', deletedFile);
            return updatedProgram;
        }
        log('Marking file as deleted. %s', deletedFile);
        fileWatchCallbacks.forEach(cb => cb(deletedFile, ts.FileWatcherEventKind.Deleted));
        // deleted files means that the file list _has_ changed, so clear the cache
        programFileListCache.delete(tsconfigPath);
        updatedProgram = existingWatch.getProgram().getProgram();
        sourceFile = updatedProgram.getSourceFile(filePath);
        if (sourceFile) {
            return updatedProgram;
        }
        log('File was still not found in program after deletion check, assuming it is not part of this program. %s', filePath);
        return null;
    }
},
/* @typescript-eslint/typescript-estree/dist/create-program/createSourceFile.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const debug_1 = __importDefault(require(72) /* debug */);
    const ts = __importStar(require(21) /* typescript */);
    const shared_1 = require(1208) /* ./shared */;
    const log = debug_1.default('typescript-eslint:typescript-estree:createSourceFile');
    function createSourceFile(code, extra) {
        log('Getting AST without type information in %s mode for: %s', extra.jsx ? 'TSX' : 'TS', extra.filePath);
        return ts.createSourceFile(extra.filePath, code, ts.ScriptTarget.Latest, 
        /* setParentNodes */ true, shared_1.getScriptKind(extra));
    }
    exports.createSourceFile = createSourceFile;
},
/* @typescript-eslint/typescript-estree/dist/semantic-or-syntactic-errors.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    __esModule();
    const ts = __importStar(require(21) /* typescript */);
    /**
     * By default, diagnostics from the TypeScript compiler contain all errors - regardless of whether
     * they are related to generic ECMAScript standards, or TypeScript-specific constructs.
     *
     * Therefore, we filter out all diagnostics, except for the ones we explicitly want to consider when
     * the user opts in to throwing errors on semantic issues.
     */
    function getFirstSemanticOrSyntacticError(program, ast) {
        try {
            const supportedSyntacticDiagnostics = whitelistSupportedDiagnostics(program.getSyntacticDiagnostics(ast));
            if (supportedSyntacticDiagnostics.length) {
                return convertDiagnosticToSemanticOrSyntacticError(supportedSyntacticDiagnostics[0]);
            }
            const supportedSemanticDiagnostics = whitelistSupportedDiagnostics(program.getSemanticDiagnostics(ast));
            if (supportedSemanticDiagnostics.length) {
                return convertDiagnosticToSemanticOrSyntacticError(supportedSemanticDiagnostics[0]);
            }
            return undefined;
        }
        catch (e) {
            /**
             * TypeScript compiler has certain Debug.fail() statements in, which will cause the diagnostics
             * retrieval above to throw.
             *
             * E.g. from ast-alignment-tests
             * "Debug Failure. Shouldn't ever directly check a JsxOpeningElement"
             *
             * For our current use-cases this is undesired behavior, so we just suppress it
             * and log a a warning.
             */
            /* istanbul ignore next */
            console.warn(`Warning From TSC: "${e.message}`); // eslint-disable-line no-console
            /* istanbul ignore next */
            return undefined;
        }
    }
    exports.getFirstSemanticOrSyntacticError = getFirstSemanticOrSyntacticError;
    function whitelistSupportedDiagnostics(diagnostics) {
        return diagnostics.filter(diagnostic => {
            switch (diagnostic.code) {
                case 1013: // "A rest parameter or binding pattern may not have a trailing comma."
                case 1014: // "A rest parameter must be last in a parameter list."
                case 1044: // "'{0}' modifier cannot appear on a module or namespace element."
                case 1045: // "A '{0}' modifier cannot be used with an interface declaration."
                case 1048: // "A rest parameter cannot have an initializer."
                case 1049: // "A 'set' accessor must have exactly one parameter."
                case 1070: // "'{0}' modifier cannot appear on a type member."
                case 1071: // "'{0}' modifier cannot appear on an index signature."
                case 1085: // "Octal literals are not available when targeting ECMAScript 5 and higher. Use the syntax '{0}'."
                case 1090: // "'{0}' modifier cannot appear on a parameter."
                case 1096: // "An index signature must have exactly one parameter."
                case 1097: // "'{0}' list cannot be empty."
                case 1098: // "Type parameter list cannot be empty."
                case 1099: // "Type argument list cannot be empty."
                case 1117: // "An object literal cannot have multiple properties with the same name in strict mode."
                case 1121: // "Octal literals are not allowed in strict mode."
                case 1123: //  "Variable declaration list cannot be empty."
                case 1141: // "String literal expected."
                case 1162: // "An object member cannot be declared optional."
                case 1164: // "Computed property names are not allowed in enums."
                case 1172: // "'extends' clause already seen."
                case 1173: // "'extends' clause must precede 'implements' clause."
                case 1175: // "'implements' clause already seen."
                case 1176: // "Interface declaration cannot have 'implements' clause."
                case 1190: // "The variable declaration of a 'for...of' statement cannot have an initializer."
                case 1196: // "Catch clause variable type annotation must be 'any' or 'unknown' if specified."
                case 1200: // "Line terminator not permitted before arrow."
                case 1206: // "Decorators are not valid here."
                case 1211: // "A class declaration without the 'default' modifier must have a name."
                case 1242: // "'abstract' modifier can only appear on a class, method, or property declaration."
                case 1246: // "An interface property cannot have an initializer."
                case 1255: // "A definite assignment assertion '!' is not permitted in this context."
                case 1308: // "'await' expression is only allowed within an async function."
                case 2364: // "The left-hand side of an assignment expression must be a variable or a property access."
                case 2369: // "A parameter property is only allowed in a constructor implementation."
                case 2452: // "An enum member cannot have a numeric name."
                case 2462: // "A rest element must be last in a destructuring pattern."
                case 8017: // "Octal literal types must use ES2015 syntax. Use the syntax '{0}'."
                case 17012: // "'{0}' is not a valid meta-property for keyword '{1}'. Did you mean '{2}'?"
                case 17013: // "Meta-property '{0}' is only allowed in the body of a function declaration, function expression, or constructor."
                    return true;
            }
            return false;
        });
    }
    function convertDiagnosticToSemanticOrSyntacticError(diagnostic) {
        return Object.assign(Object.assign({}, diagnostic), { message: ts.flattenDiagnosticMessageText(diagnostic.messageText, ts.sys.newLine) });
    }
},
/* @typescript-eslint/typescript-estree/dist/create-program/useProvidedPrograms.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const debug_1 = __importDefault(require(72) /* debug */);
    const fs = __importStar(require("fs"));
    const path = __importStar(require("path"));
    const ts = __importStar(require(21) /* typescript */);
    const shared_1 = require(1208) /* ./shared */;
    const log = debug_1.default('typescript-eslint:typescript-estree:useProvidedProgram');
    function useProvidedPrograms(programInstances, extra) {
        log('Retrieving ast for %s from provided program instance(s)', extra.filePath);
        let astAndProgram;
        for (const programInstance of programInstances) {
            astAndProgram = shared_1.getAstFromProgram(programInstance, extra);
            // Stop at the first applicable program instance
            if (astAndProgram) {
                break;
            }
        }
        if (!astAndProgram) {
            const relativeFilePath = path.relative(extra.tsconfigRootDir || process.cwd(), extra.filePath);
            const errorLines = [
                '"parserOptions.programs" has been provided for @typescript-eslint/parser.',
                `The file was not found in any of the provided program instance(s): ${relativeFilePath}`,
            ];
            throw new Error(errorLines.join('\n'));
        }
        astAndProgram.program.getTypeChecker(); // ensure parent pointers are set in source files
        return astAndProgram;
    }
    exports.useProvidedPrograms = useProvidedPrograms;
    /**
     * Utility offered by parser to help consumers construct their own program instance.
     *
     * @param configFile the path to the tsconfig.json file, relative to `projectDirectory`
     * @param projectDirectory the project directory to use as the CWD, defaults to `process.cwd()`
     */
    function createProgramFromConfigFile(configFile, projectDirectory) {
        if (ts.sys === undefined) {
            throw new Error('`createProgramFromConfigFile` is only supported in a Node-like environment.');
        }
        const parsed = ts.getParsedCommandLineOfConfigFile(configFile, shared_1.CORE_COMPILER_OPTIONS, {
            onUnRecoverableConfigFileDiagnostic: diag => {
                throw new Error(formatDiagnostics([diag])); // ensures that `parsed` is defined.
            },
            fileExists: fs.existsSync,
            getCurrentDirectory: () => (projectDirectory && path.resolve(projectDirectory)) || process.cwd(),
            readDirectory: ts.sys.readDirectory,
            readFile: file => fs.readFileSync(file, 'utf-8'),
            useCaseSensitiveFileNames: ts.sys.useCaseSensitiveFileNames,
        });
        const result = parsed; // parsed is not undefined, since we throw on failure.
        if (result.errors.length) {
            throw new Error(formatDiagnostics(result.errors));
        }
        const host = ts.createCompilerHost(result.options, true);
        return ts.createProgram(result.fileNames, result.options, host);
    }
    exports.createProgramFromConfigFile = createProgramFromConfigFile;
    function formatDiagnostics(diagnostics) {
        return ts.formatDiagnostics(diagnostics, {
            getCanonicalFileName: f => f,
            getCurrentDirectory: process.cwd,
            getNewLine: () => '\n',
        });
    }
},
/* @typescript-eslint/typescript-estree/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "@typescript-eslint/typescript-estree",
        "version": "4.29.1",
        "description": "A parser that converts TypeScript source code into an ESTree compatible form",
        "main": "dist/index.js",
        "types": "dist/index.d.ts",
        "files": [
            "dist",
            "README.md",
            "LICENSE"
        ],
        "engines": {
            "node": "^10.12.0 || >=12.0.0"
        },
        "repository": {
            "type": "git",
            "url": "https://github.com/typescript-eslint/typescript-eslint.git",
            "directory": "packages/typescript-estree"
        },
        "bugs": {
            "url": "https://github.com/typescript-eslint/typescript-eslint/issues"
        },
        "license": "BSD-2-Clause",
        "keywords": [
            "ast",
            "estree",
            "ecmascript",
            "javascript",
            "typescript",
            "parser",
            "syntax"
        ],
        "scripts": {
            "build": "tsc -b tsconfig.build.json",
            "postbuild": "downlevel-dts dist _ts3.4/dist",
            "clean": "tsc -b tsconfig.build.json --clean",
            "postclean": "rimraf dist && rimraf _ts3.4 && rimraf coverage",
            "format": "prettier --write \"./**/*.{ts,js,json,md}\" --ignore-path ../../.prettierignore",
            "lint": "eslint . --ext .js,.ts --ignore-path='../../.eslintignore'",
            "test": "jest --coverage",
            "typecheck": "tsc -p tsconfig.json --noEmit"
        },
        "dependencies": {
            "@typescript-eslint/types": "4.29.1",
            "@typescript-eslint/visitor-keys": "4.29.1",
            "debug": "^4.3.1",
            "globby": "^11.0.3",
            "is-glob": "^4.0.1",
            "semver": "^7.3.5",
            "tsutils": "^3.21.0"
        },
        "devDependencies": {
            "@babel/code-frame": "*",
            "@babel/parser": "*",
            "@babel/types": "*",
            "@types/babel__code-frame": "*",
            "@types/debug": "*",
            "@types/glob": "*",
            "@types/is-glob": "*",
            "@types/semver": "*",
            "@types/tmp": "*",
            "@typescript-eslint/shared-fixtures": "4.29.1",
            "glob": "*",
            "jest-specific-snapshot": "*",
            "make-dir": "*",
            "tmp": "^0.2.1",
            "typescript": "*"
        },
        "peerDependenciesMeta": {
            "typescript": {
                "optional": true
            }
        },
        "funding": {
            "type": "opencollective",
            "url": "https://opencollective.com/typescript-eslint"
        },
        "typesVersions": {
            "<3.8": {
                "*": [
                    "_ts3.4/*"
                ]
            }
        },
        "gitHead": "db78642ea3ebd794e7deef767fa2c5195ef7c04c"
    };
    module.exports = json;
},
/* @typescript-eslint/parser/package.json */ function _(require, module, exports, __esModule, __esExport) {
    var json = {
        "name": "@typescript-eslint/parser",
        "version": "4.29.1",
        "description": "An ESLint custom parser which leverages TypeScript ESTree",
        "main": "dist/index.js",
        "types": "dist/index.d.ts",
        "files": [
            "dist",
            "README.md",
            "LICENSE"
        ],
        "engines": {
            "node": "^10.12.0 || >=12.0.0"
        },
        "repository": {
            "type": "git",
            "url": "https://github.com/typescript-eslint/typescript-eslint.git",
            "directory": "packages/parser"
        },
        "bugs": {
            "url": "https://github.com/typescript-eslint/typescript-eslint/issues"
        },
        "license": "BSD-2-Clause",
        "keywords": [
            "ast",
            "ecmascript",
            "javascript",
            "typescript",
            "parser",
            "syntax",
            "eslint"
        ],
        "scripts": {
            "build": "tsc -b tsconfig.build.json",
            "postbuild": "downlevel-dts dist _ts3.4/dist",
            "clean": "tsc -b tsconfig.build.json --clean",
            "postclean": "rimraf dist && rimraf _ts3.4 && rimraf coverage",
            "format": "prettier --write \"./**/*.{ts,js,json,md}\" --ignore-path ../../.prettierignore",
            "lint": "eslint . --ext .js,.ts --ignore-path='../../.eslintignore'",
            "test": "jest --coverage",
            "typecheck": "tsc -p tsconfig.json --noEmit"
        },
        "peerDependencies": {
            "eslint": "^5.0.0 || ^6.0.0 || ^7.0.0"
        },
        "dependencies": {
            "@typescript-eslint/scope-manager": "4.29.1",
            "@typescript-eslint/types": "4.29.1",
            "@typescript-eslint/typescript-estree": "4.29.1",
            "debug": "^4.3.1"
        },
        "devDependencies": {
            "@types/glob": "*",
            "@typescript-eslint/experimental-utils": "4.29.1",
            "glob": "*",
            "typescript": "*"
        },
        "peerDependenciesMeta": {
            "typescript": {
                "optional": true
            }
        },
        "funding": {
            "type": "opencollective",
            "url": "https://opencollective.com/typescript-eslint"
        },
        "typesVersions": {
            "<3.8": {
                "*": [
                    "_ts3.4/*"
                ]
            }
        },
        "gitHead": "db78642ea3ebd794e7deef767fa2c5195ef7c04c"
    };
    module.exports = json;
},
/* compile.js */ function _(require, module, exports, __esModule, __esExport) {
    var __createBinding = (this && this.__createBinding) || (Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    }));
    var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    });
    var __importStar = (this && this.__importStar) || function (mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    };
    var __importDefault = (this && this.__importDefault) || function (mod) {
        return (mod && mod.__esModule) ? mod : { "default": mod };
    };
    __esModule();
    const path = __importStar(require("path"));
    const ts = __importStar(require(21) /* typescript */);
    const less_1 = __importDefault(require(56) /* less */);
    const compiler_1 = require(36) /* ./compiler */;
    const sys_1 = require(20) /* ./sys */;
    const transforms = __importStar(require(42) /* ./transforms */);
    const tsconfig_json = __importStar(require(202) /* ./tsconfig.ext.json */);
    function parse_patched_tsconfig(base_dir, preconfigure) {
        // XXX: silence the config validator. We are providing inputs through `inputs` argument anyway.
        const json = { ...tsconfig_json, include: undefined, files: ["dummy.ts"] };
        return (0, compiler_1.parse_tsconfig)(json, base_dir, preconfigure);
    }
    function compile_typescript(base_dir, inputs, bokehjs_dir) {
        const preconfigure = {
            module: ts.ModuleKind.CommonJS,
            paths: {
                "*": [
                    path.join(bokehjs_dir, "js/lib/*"),
                    path.join(bokehjs_dir, "js/types/*"),
                ],
            },
            outDir: undefined,
        };
        const tsconfig = parse_patched_tsconfig(base_dir, preconfigure);
        if (tsconfig.diagnostics != null)
            return { diagnostics: tsconfig.diagnostics };
        const host = (0, compiler_1.compiler_host)(inputs, tsconfig.options, bokehjs_dir);
        const transformers = (0, compiler_1.default_transformers)(tsconfig.options);
        const outputs = new Map();
        host.writeFile = (name, data) => {
            outputs.set(name, data);
        };
        const files = [...inputs.keys()];
        return { outputs, ...(0, compiler_1.compile_files)(files, tsconfig.options, transformers, host) };
    }
    exports.compile_typescript = compile_typescript;
    function compile_javascript(base_dir, file, code) {
        const tsconfig = parse_patched_tsconfig(base_dir, {});
        if (tsconfig.diagnostics != null)
            return { diagnostics: tsconfig.diagnostics };
        const { outputText, diagnostics } = ts.transpileModule(code, {
            fileName: file,
            reportDiagnostics: true,
            compilerOptions: {
                target: tsconfig.options.target,
                module: ts.ModuleKind.CommonJS,
            },
        });
        return { output: outputText, diagnostics };
    }
    function normalize(path) {
        return path.replace(/\\/g, "/");
    }
    async function compile_and_resolve_deps(input) {
        const { file, lang, bokehjs_dir } = input;
        const { code } = input;
        let output;
        switch (lang) {
            case "typescript":
                const inputs = new Map([[normalize(file), code]]);
                const { outputs, diagnostics } = compile_typescript(".", inputs, bokehjs_dir);
                if (diagnostics != null && diagnostics.length != 0) {
                    const failure = (0, compiler_1.report_diagnostics)(diagnostics);
                    return { error: failure.text };
                }
                else {
                    const js_file = normalize((0, sys_1.rename)(file, { ext: ".js" }));
                    output = outputs.get(js_file);
                }
                break;
            case "javascript": {
                const result = compile_javascript(".", file, code);
                if (result.diagnostics != null && result.diagnostics.length != 0) {
                    const failure = (0, compiler_1.report_diagnostics)(result.diagnostics);
                    return { error: failure.text };
                }
                else {
                    output = result.output;
                }
                break;
            }
            case "less":
                try {
                    const { css } = await less_1.default.render(code, { filename: file, compress: true });
                    return { code: css };
                }
                catch (error) {
                    return { error: `${error}` };
                }
            default:
                throw new Error(`unsupported input type: ${lang}`);
        }
        const source = ts.createSourceFile(file, output, ts.ScriptTarget.ES5, true, ts.ScriptKind.JS);
        const deps = transforms.collect_deps(source);
        return { code: output, deps };
    }
    exports.compile_and_resolve_deps = compile_and_resolve_deps;
},
], 0, {"main":0,"sys":20,"build":35,"compiler":36,"transforms":42,"error":43,"linker":44,"graph":54,"styles":55,"prelude":201,"tsconfig.ext.json":202,"compile":1217}, {"path":true,"assert":true,"fs":true,"util":true,"crypto":true,"perf_hooks":true,"os":true,"buffer":true,"inspector":true,"child_process":true,"readline":true,"tty":true,"module":true,"constants":true,"stream":true,"url":true,"http":true,"https":true,"zlib":true,"querystring":true,"string_decoder":true,"events":true});});
//# sourceMappingURL=compiler.js.map
